!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of spin orbit integrals over Cartesian Gaussian-type functions
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      none
!> \par Parameters
!>       - ax,ay,az    : Angular momentum index numbers of orbital a.
!>       - bx,by,bz    : Angular momentum index numbers of orbital b.
!>       - coset       : Cartesian orbital set pointer.
!>       - dab         : Distance between the atomic centers a and b.
!>       - dac         : Distance between the atomic centers a and c.
!>       - dbc         : Distance between the atomic centers b and c.
!>       - l{a,b,c}    : Angular momentum quantum number of shell a, b or c.
!>       - l{a,b,c}_max: Maximum angular momentum quantum number of shell a, b or c.
!>       - l{a,b,c}_min: Minimum angular momentum quantum number of shell a, b or c.
!>       - ncoset      : Number of orbitals in a Cartesian orbital set.
!>       - npgf{a,b}   : Degree of contraction of shell a or b.
!>       - rab         : Distance vector between the atomic centers a and b.
!>       - rab2        : Square of the distance between the atomic centers a and b.
!>       - rac         : Distance vector between the atomic centers a and c.
!>       - rac2        : Square of the distance between the atomic centers a and c.
!>       - rbc         : Distance vector between the atomic centers b and c.
!>       - rbc2        : Square of the distance between the atomic centers b and c.
!>       - rpgf{a,b,c} : Radius of the primitive Gaussian-type function a, b or c.
!>       - zet{a,b,c}  : Exponents of the Gaussian-type functions a, b or c.
!>       - zetp        : Reciprocal of the sum of the exponents of orbital a and b.
!> \author VW
! **************************************************************************************************
MODULE ai_spin_orbit
   USE ai_os_rr,                        ONLY: os_rr_coul
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_spin_orbit'
   PRIVATE

   ! *** Public subroutines ***

   PUBLIC :: pso

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the primitive paramagnetic spin orbit integrals over
!>          Cartesian Gaussian-type functions.
!> \param la_max ...
!> \param la_min ...
!> \param npgfa ...
!> \param rpgfa ...
!> \param zeta ...
!> \param lb_max ...
!> \param lb_min ...
!> \param npgfb ...
!> \param rpgfb ...
!> \param zetb ...
!> \param rac ...
!> \param rbc ...
!> \param rab ...
!> \param vab ...
!> \param ldrr1 ...
!> \param ldrr2 ...
!> \param rr ...
!> \date    02.03.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE pso(la_max, la_min, npgfa, rpgfa, zeta, lb_max, lb_min, npgfb, rpgfb, zetb, &
                  rac, rbc, rab, vab, ldrr1, ldrr2, rr)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac, rbc, rab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: vab
      INTEGER, INTENT(IN)                                :: ldrr1, ldrr2
      REAL(dp), DIMENSION(0:ldrr1-1, ldrr2, *), &
         INTENT(INOUT)                                   :: rr

      CHARACTER(len=*), PARAMETER :: routineN = 'pso', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, coa, coam1x, coam1y, coam1z, coap1x, coap1y, coap1z, cob, &
         cobm1x, cobm1y, cobm1z, cobp1x, cobp1y, cobp1z, i, ipgf, j, jpgf, la, lb, ma, mb, na, nb
      REAL(dp)                                           :: dab, dum1, dum2, f0, rab2, xhi, zet, &
                                                            zetab
      REAL(dp), DIMENSION(3)                             :: rap, rbp, rcp

! *** Calculate the distance of the centers a and c ***

      rab2 = rab(1)**2+rab(2)**2+rab(3)**2
      dab = SQRT(rab2)

      ! *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0

      DO ipgf = 1, npgfa

         nb = 0

         DO jpgf = 1, npgfb

            ! *** Screening ***

            IF (rpgfa(ipgf)+rpgfb(jpgf) < dab) THEN
               DO j = nb+1, nb+ncoset(lb_max)
                  DO i = na+1, na+ncoset(la_max)
                     vab(i, j, 1) = 0.0_dp
                     vab(i, j, 2) = 0.0_dp
                     vab(i, j, 3) = 0.0_dp
                  ENDDO
               ENDDO
               nb = nb+ncoset(lb_max)
               CYCLE
            ENDIF

            ! *** Calculate some prefactors ***

            zetab = zeta(ipgf)*zetb(jpgf)
            zet = zeta(ipgf)+zetb(jpgf)
            xhi = zetab/zet
            rap = zetb(jpgf)*rab/zet
            rbp = -zeta(ipgf)*rab/zet
            rcp = -(zeta(ipgf)*rac+zetb(jpgf)*rbc)/zet

            f0 = 2.0_dp*SQRT(zet/pi)*(pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

            ! *** Calculate the reccurence relation ***

            CALL os_rr_coul(rap, la_max+1, rbp, lb_max+1, rcp, zet, ldrr1, ldrr2, rr)

            ! *** Calculate the primitive Fermi contact integrals ***

            DO lb = lb_min, lb_max
            DO bx = 0, lb
            DO by = 0, lb-bx
               bz = lb-bx-by
               cob = coset(bx, by, bz)
               cobm1x = coset(MAX(bx-1, 0), by, bz)
               cobm1y = coset(bx, MAX(by-1, 0), bz)
               cobm1z = coset(bx, by, MAX(bz-1, 0))
               cobp1x = coset(bx+1, by, bz)
               cobp1y = coset(bx, by+1, bz)
               cobp1z = coset(bx, by, bz+1)
               mb = nb+cob
               DO la = la_min, la_max
               DO ax = 0, la
               DO ay = 0, la-ax
                  az = la-ax-ay
                  coa = coset(ax, ay, az)
                  coam1x = coset(MAX(ax-1, 0), ay, az)
                  coam1y = coset(ax, MAX(ay-1, 0), az)
                  coam1z = coset(ax, ay, MAX(az-1, 0))
                  coap1x = coset(ax+1, ay, az)
                  coap1y = coset(ax, ay+1, az)
                  coap1z = coset(ax, ay, az+1)
                  ma = na+coa
                  !
                  !
                  ! (a|pso_x|b) = (4*zeta*zetb*(a+y||b+z)
                  !               -2*zeta*Nz(b)*(a+y||b-z)-2*zetb*Ny(a)*(a-y||b+z)
                  !               +Ny(a)*Nz(b)*(a-y||b-z))
                  !              -(4*zeta*zetb*(a+z||b+y)
                  !               -2*zeta*Ny(b)*(a+z||b-y)-2*zetb*Nz(a)*(a-z||b+y)
                  !               +Nz(a)*Ny(b)*(a-z||b-y))
                  dum1 = 4.0_dp*zeta(ipgf)*zetb(jpgf)*rr(0, coap1y, cobp1z)
                  IF (bz .GT. 0) dum1 = dum1-2.0_dp*zeta(ipgf)*REAL(bz, dp)*rr(0, coap1y, cobm1z)
                  IF (ay .GT. 0) dum1 = dum1-2.0_dp*zetb(jpgf)*REAL(ay, dp)*rr(0, coam1y, cobp1z)
                  IF (ay .GT. 0 .AND. bz .GT. 0) dum1 = dum1+REAL(ay, dp)*REAL(bz, dp)*rr(0, coam1y, cobm1z)
                  !
                  dum2 = 4.0_dp*zeta(ipgf)*zetb(jpgf)*rr(0, coap1z, cobp1y)
                  IF (by .GT. 0) dum2 = dum2-2.0_dp*zeta(ipgf)*REAL(by, dp)*rr(0, coap1z, cobm1y)
                  IF (az .GT. 0) dum2 = dum2-2.0_dp*zetb(jpgf)*REAL(az, dp)*rr(0, coam1z, cobp1y)
                  IF (az .GT. 0 .AND. by .GT. 0) dum2 = dum2+REAL(az, dp)*REAL(by, dp)*rr(0, coam1z, cobm1y)
                  vab(ma, mb, 1) = f0*(dum1-dum2)
                  !
                  !
                  ! (a|pso_y|b) = (4*zeta*zetb*(a+z||b+x)
                  !               -2*zeta*Nx(b)*(a+z||b-x)-2*zetb*Nz(a)*(a-z||b+x)
                  !               +Nz(a)*Nx(b)*(a-z||b-x))
                  !              -(4*zeta*zetb*(a+x||b+z)
                  !               -2*zeta*Nz(b)*(a+x||b-z)-2*zetb*Nx(a)*(a-x||b+z)
                  !               +Nx(a)*Nz(b)*(a-x||b-z))
                  dum1 = 4.0_dp*zeta(ipgf)*zetb(jpgf)*rr(0, coap1z, cobp1x)
                  IF (bx .GT. 0) dum1 = dum1-2.0_dp*zeta(ipgf)*REAL(bx, dp)*rr(0, coap1z, cobm1x)
                  IF (az .GT. 0) dum1 = dum1-2.0_dp*zetb(jpgf)*REAL(az, dp)*rr(0, coam1z, cobp1x)
                  IF (az .GT. 0 .AND. bx .GT. 0) dum1 = dum1+REAL(az, dp)*REAL(bx, dp)*rr(0, coam1z, cobm1x)
                  !
                  dum2 = 4.0_dp*zeta(ipgf)*zetb(jpgf)*rr(0, coap1x, cobp1z)
                  IF (bz .GT. 0) dum2 = dum2-2.0_dp*zeta(ipgf)*REAL(bz, dp)*rr(0, coap1x, cobm1z)
                  IF (ax .GT. 0) dum2 = dum2-2.0_dp*zetb(jpgf)*REAL(ax, dp)*rr(0, coam1x, cobp1z)
                  IF (ax .GT. 0 .AND. bz .GT. 0) dum2 = dum2+REAL(ax, dp)*REAL(bz, dp)*rr(0, coam1x, cobm1z)
                  vab(ma, mb, 2) = f0*(dum1-dum2)
                  !
                  !
                  ! (a|pso_z|b) = (4*zeta*zetb*(a+x||b+y)
                  !               -2*zeta*Ny(b)*(a+x||b-y)-2*zetb*Nx(a)*(a-x||b+y)
                  !               +Nx(a)*Ny(b)*(a-x||b-y))
                  !              -(4*zeta*zetb*(a+y||b+x)
                  !               -2*zeta*Nx(b)*(a+y||b-x)-2*zetb*Ny(a)*(a-y||b+x)
                  !               +Ny(a)*Nx(b)*(a-y||b-x))
                  dum1 = 4.0_dp*zeta(ipgf)*zetb(jpgf)*rr(0, coap1x, cobp1y)
                  IF (by .GT. 0) dum1 = dum1-2.0_dp*zeta(ipgf)*REAL(by, dp)*rr(0, coap1x, cobm1y)
                  IF (ax .GT. 0) dum1 = dum1-2.0_dp*zetb(jpgf)*REAL(ax, dp)*rr(0, coam1x, cobp1y)
                  IF (ax .GT. 0 .AND. by .GT. 0) dum1 = dum1+REAL(ax, dp)*REAL(by, dp)*rr(0, coam1x, cobm1y)
                  !
                  dum2 = 4.0_dp*zeta(ipgf)*zetb(jpgf)*rr(0, coap1y, cobp1x)
                  IF (bx .GT. 0) dum2 = dum2-2.0_dp*zeta(ipgf)*REAL(bx, dp)*rr(0, coap1y, cobm1x)
                  IF (ay .GT. 0) dum2 = dum2-2.0_dp*zetb(jpgf)*REAL(ay, dp)*rr(0, coam1y, cobp1x)
                  IF (ay .GT. 0 .AND. bx .GT. 0) dum2 = dum2+REAL(ay, dp)*REAL(bx, dp)*rr(0, coam1y, cobm1x)
                  vab(ma, mb, 3) = f0*(dum1-dum2)
                  !
               ENDDO
               ENDDO
               ENDDO !la

            ENDDO
            ENDDO
            ENDDO !lb

            nb = nb+ncoset(lb_max)

         ENDDO

         na = na+ncoset(la_max)

      ENDDO

   END SUBROUTINE pso

END MODULE ai_spin_orbit
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Distribution of the spin orbit integral matrix.
!> \par History
!> \author VW (27.02.2009)
! **************************************************************************************************
MODULE qs_spin_orbit
   USE ai_spin_orbit,                   ONLY: pso
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE block_p_types,                   ONLY: block_p_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_spin_orbit'

! *** Public subroutines ***

   PUBLIC :: build_pso_matrix

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the paramagnetic spin orbit matrix over
!>          Cartesian Gaussian functions.
!> \param qs_env ...
!> \param matrix_so ...
!> \param rc ...
!> \date    27.02.2009
!> \author  VW
!> \version 1.0
! **************************************************************************************************

   SUBROUTINE build_pso_matrix(qs_env, matrix_so, rc)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_so
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rc

      CHARACTER(len=*), PARAMETER :: routineN = 'build_pso_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER :: after, handle, i, iatom, icol, ikind, inode, irow, iset, istat, iw, jatom, jkind, &
         jset, last_jatom, ldai, maxco, maxlgto, maxsgf, natom, ncoa, ncob, neighbor_list_id, &
         nkind, nseta, nsetb, sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: found, new_atom_b, omit_headers
      REAL(KIND=dp)                                      :: dab, rab2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rr_work, soab
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rac, rb, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb
      TYPE(block_p_type), ALLOCATABLE, DIMENSION(:)      :: soint
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind

      CALL timeset(routineN, handle)

      NULLIFY (cell, sab_orb, qs_kind_set, particle_set, para_env)
      NULLIFY (logger)

      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      neighbor_list_id=neighbor_list_id, &
                      para_env=para_env, &
                      sab_orb=sab_orb, &
                      cell=cell)

      nkind = SIZE(qs_kind_set)
      natom = SIZE(particle_set)

!   *** Allocate work storage ***

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxlgto=maxlgto, &
                           maxsgf=maxsgf)

      ldai = ncoset(maxlgto+1)
      CALL init_orbital_pointers(ldai)

      ALLOCATE (rr_work(0:2*maxlgto+2, ldai, ldai), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 139, &
                                       "ff", (2*maxlgto+2)*dp_size)

      ALLOCATE (soab(maxco, maxco, 3), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 143, &
                                       "soab", maxco*maxco*3*dp_size)

      ALLOCATE (work(maxco, maxsgf), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 147, &
                                       "work", maxco*maxsgf*dp_size)

      ALLOCATE (soint(3), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 151, &
                                       "soint", 3*int_size)

      rr_work(:, :, :) = 0.0_dp
      soab(:, :, :) = 0.0_dp
      work(:, :) = 0.0_dp

      ALLOCATE (basis_set_list(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 159, &
                                       "basis_set_list", nkind)
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         IF (inode == 1) last_jatom = 0

         rb = rab+ra
         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         dab = SQRT(rab2)
         rac = pbc(ra, rc, cell)
         rbc = rac-rab

         IF (jatom /= last_jatom) THEN
            new_atom_b = .TRUE.
            last_jatom = jatom
         ELSE
            new_atom_b = .FALSE.
         END IF

         IF (new_atom_b) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF

            DO i = 1, 3
               NULLIFY (soint(i)%block)
               CALL cp_dbcsr_get_block_p(matrix=matrix_so(i)%matrix, &
                                         row=irow, col=icol, BLOCK=soint(i)%block, found=found)
            ENDDO
         END IF

         DO iset = 1, nseta

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               ! *** Calculate the primitive fermi contact integrals ***

               CALL pso(la_max(iset), la_min(iset), npgfa(iset), &
                        rpgfa(:, iset), zeta(:, iset), &
                        lb_max(jset), lb_min(jset), npgfb(jset), &
                        rpgfb(:, jset), zetb(:, jset), &
                        rac, rbc, rab, soab, SIZE(rr_work, 1), SIZE(rr_work, 2), rr_work)

               ! *** Contraction step ***

               DO i = 1, 3

                  CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                             1.0_dp, soab(1, 1, i), SIZE(soab, 1), &
                             sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             0.0_dp, work(1, 1), SIZE(work, 1))

                  IF (iatom <= jatom) THEN

                     CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                                1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                work(1, 1), SIZE(work, 1), &
                                1.0_dp, soint(i)%block(sgfa, sgfb), &
                                SIZE(soint(i)%block, 1))

                  ELSE

                     CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                                -1.0_dp, work(1, 1), SIZE(work, 1), &
                                sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                1.0_dp, soint(i)%block(sgfb, sgfa), &
                                SIZE(soint(i)%block, 1))
                  ENDIF

               ENDDO

            ENDDO

         ENDDO

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ! *** Release work storage ***

      DEALLOCATE (basis_set_list)

      DEALLOCATE (soab)

      DEALLOCATE (work)

      DEALLOCATE (soint)

!   *** Print the spin orbit matrix, if requested ***

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "DFT%PRINT%AO_MATRICES/PSO"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/PSO", &
                                   extension=".Log")
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
         CALL cp_dbcsr_write_sparse_matrix(matrix_so(1)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_dbcsr_write_sparse_matrix(matrix_so(2)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_dbcsr_write_sparse_matrix(matrix_so(3)%matrix, 4, after, qs_env, &
                                           para_env, output_unit=iw, omit_headers=omit_headers)
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "DFT%PRINT%AO_MATRICES/PSO")
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_pso_matrix

! **************************************************************************************************

END MODULE qs_spin_orbit

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the operators p rxp and D needed in the optimization
!>      of the different contribution of the firs order response orbitals
!>      in a epr calculation
!> \note
!>      The interactions are considered only within the minimum image convention
!> \par History
!>       created 07-2005 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_op
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_array_utils_i,                ONLY: cp_2d_i_p_type
   USE cp_array_utils_r,                ONLY: cp_2d_r_p_type
   USE cp_cfm_basic_linalg,             ONLY: cp_cfm_solve
   USE cp_cfm_types,                    ONLY: cp_cfm_create,&
                                              cp_cfm_get_info,&
                                              cp_cfm_p_type,&
                                              cp_cfm_release,&
                                              cp_cfm_set_all
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        convert_offsets_to_sizes, cp_dbcsr_allocate_matrix_set, cp_dbcsr_checksum, cp_dbcsr_copy, &
        cp_dbcsr_create, cp_dbcsr_deallocate_matrix, cp_dbcsr_deallocate_matrix_set, &
        cp_dbcsr_init, cp_dbcsr_p_type, cp_dbcsr_set, cp_dbcsr_type, dbcsr_distribution_obj, &
        dbcsr_type_antisymmetric, dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_info, cp_fm_get_submatrix, cp_fm_p_type, cp_fm_release, &
        cp_fm_set_all, cp_fm_set_submatrix, cp_fm_to_fm, cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: twopi
   USE orbital_pointers,                ONLY: coset
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE qs_elec_field,                   ONLY: build_efg_matrix
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_fermi_contact,                ONLY: build_fermi_contact_matrix
   USE qs_kind_types,                   ONLY: get_qs_kind_set,&
                                              qs_kind_type
   USE qs_linres_types,                 ONLY: current_env_type,&
                                              get_current_env,&
                                              get_issc_env,&
                                              get_polar_env,&
                                              issc_env_type,&
                                              linres_control_type,&
                                              polar_env_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_moments,                      ONLY: build_berry_moment_matrix
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_operators_ao,                 ONLY: build_ang_mom_matrix,&
                                              build_lin_mom_matrix,&
                                              rRc_xyz_ao
   USE qs_spin_orbit,                   ONLY: build_pso_matrix
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: current_operators, issc_operators, fac_vecp, ind_m2, set_vecp, set_vecp_rev, &
             fm_scale_by_pbc_AC, polar_operators

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_op'

CONTAINS

! **************************************************************************************************
!> \brief Calculate the first order hamiltonian applied to the ao
!>      and then apply them to the ground state orbitals,
!>      the h1_psi1 full matrices are then ready to solve the
!>      non-homogeneous linear equations that give the psi1
!>      linear response orbitals.
!> \param current_env ...
!> \param qs_env ...
!> \par History
!>      07.2005 created [MI]
!> \author MI
!> \note
!>      For the operators rxp and D the h1 depends on the psi0 to which
!>      is applied, or better the center of charge of the psi0 is
!>      used to define the position operator
!>      The centers of the orbitals result form the orbital localization procedure
!>      that typicaly uses the berry phase operator to define the Wannier centers.
! **************************************************************************************************
   SUBROUTINE current_operators(current_env, qs_env)

      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'current_operators', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iao, icenter, idir, ii, iii, &
                                                            ispin, istate, j, nao, natom, &
                                                            nbr_center(2), nmo, nsgf, nspins, &
                                                            nstates(2), output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, last_sgf
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_sizes
      REAL(dp)                                           :: chk(3), ck(3), ckdk(3), dk(3)
      REAL(dp), DIMENSION(:, :), POINTER                 :: basisfun_center, vecbuf_c0
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_2d_i_p_type), DIMENSION(:), POINTER        :: center_list
      TYPE(cp_2d_r_p_type), DIMENSION(3)                 :: vecbuf_RmdC0
      TYPE(cp_2d_r_p_type), DIMENSION(:), POINTER        :: centers_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op_ao
      TYPE(cp_fm_p_type), DIMENSION(3)                   :: fm_Rmd_mos
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: psi0_order
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: p_psi0, rxp_psi0
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_work1, mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_all, sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: lr_section

      CALL timeset(routineN, handle)

      NULLIFY (qs_kind_set, cell, dft_control, linres_control, &
               logger, particle_set, lr_section, fm_work1, &
               basisfun_center, centers_set, center_list, p_psi0, &
               rxp_psi0, vecbuf_c0, psi0_order, &
               mo_coeff, op_ao, sab_all)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, &
                                               "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")
      IF (output_unit > 0) THEN
         WRITE (output_unit, FMT="(T2,A,/)") &
            "CURRENT| Calculation of the p and (r-d)xp operators applied to psi0"
      ENDIF

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      linres_control=linres_control, &
                      para_env=para_env, &
                      particle_set=particle_set, &
                      sab_all=sab_all, &
                      sab_orb=sab_orb, &
                      dbcsr_dist=dbcsr_dist)

      nspins = dft_control%nspins

      CALL get_current_env(current_env=current_env, nao=nao, centers_set=centers_set, &
                           center_list=center_list, basisfun_center=basisfun_center, &
                           nbr_center=nbr_center, p_psi0=p_psi0, rxp_psi0=rxp_psi0, &
                           psi0_order=psi0_order, &
                           nstates=nstates)

      ALLOCATE (vecbuf_c0(1, nao))
      DO idir = 1, 3
         NULLIFY (vecbuf_Rmdc0(idir)%array)
         ALLOCATE (vecbuf_Rmdc0(idir)%array(1, nao))
      ENDDO

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, nsgf=nsgf)

      natom = SIZE(particle_set, 1)
      ALLOCATE (first_sgf(natom))
      ALLOCATE (last_sgf(natom))

      CALL get_particle_set(particle_set, qs_kind_set, &
                            first_sgf=first_sgf, &
                            last_sgf=last_sgf)

      ! Calculate the (r - dk)xp operator applied to psi0k
      ! One possible way to go is to use the distributive property of the vector product and calculatr
      ! (r-c)xp + (c-d)xp
      ! where c depends on the contracted functions and not on the states
      ! d is the center of a specific state and a loop over states is needed
      ! the second term can be added in a second moment as a correction
      ! notice: (r-c) and p are operators, whereas (c-d) is a multiplicative factor

      !    !First term: operator matrix elements
      !    CALL rmc_x_p_xyz_ao(op_rmd_ao,qs_env,minimum_image=.FALSE.)
      !************************************************************
      !
      ! Since many psi0 vector can have the same center, depending on how the center is selected,
      ! the (r - dk)xp operator matrix is computed Ncenter times,
      ! where Ncenter is the total number of different centers
      ! and each time it is multiplied by all the psi0 with center dk to get the rxp_psi0 matrix

      !
      ! prepare for allocation
      ALLOCATE (row_blk_sizes(natom))
      CALL convert_offsets_to_sizes(first_sgf, row_blk_sizes, last_sgf)
      !
      !
      CALL cp_dbcsr_allocate_matrix_set(op_ao, 3)
      ALLOCATE (op_ao(1)%matrix, op_ao(2)%matrix, op_ao(3)%matrix)
      CALL cp_dbcsr_init(op_ao(1)%matrix)
      CALL cp_dbcsr_init(op_ao(2)%matrix)
      CALL cp_dbcsr_init(op_ao(3)%matrix)

      CALL cp_dbcsr_create(matrix=op_ao(1)%matrix, &
                           name="op_ao", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_no_symmetry, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0, mutable_work=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(op_ao(1)%matrix, sab_all)
      CALL cp_dbcsr_set(op_ao(1)%matrix, 0.0_dp)

      DO idir = 2, 3
         CALL cp_dbcsr_copy(op_ao(idir)%matrix, op_ao(1)%matrix, &
                            "op_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir))))
         CALL cp_dbcsr_set(op_ao(idir)%matrix, 0.0_dp)
      ENDDO

      chk(:) = 0.0_dp
      DO ispin = 1, nspins
         mo_coeff => psi0_order(ispin)%matrix
         nmo = nstates(ispin)
         CALL cp_fm_set_all(p_psi0(ispin, 1)%matrix, 0.0_dp)
         CALL cp_fm_set_all(p_psi0(ispin, 2)%matrix, 0.0_dp)
         CALL cp_fm_set_all(p_psi0(ispin, 3)%matrix, 0.0_dp)
         DO icenter = 1, nbr_center(ispin)
            CALL cp_dbcsr_set(op_ao(1)%matrix, 0.0_dp)
            CALL cp_dbcsr_set(op_ao(2)%matrix, 0.0_dp)
            CALL cp_dbcsr_set(op_ao(3)%matrix, 0.0_dp)
            !CALL rmc_x_p_xyz_ao(op_ao,qs_env,minimum_image=.FALSE.,&
            !     &              wancen=centers_set(ispin)%array(1:3,icenter))
            !     &
            CALL build_ang_mom_matrix(qs_env, op_ao, centers_set(ispin)%array(1:3, icenter))
            !
            ! accumulate checksums
            chk(1) = chk(1)+cp_dbcsr_checksum(op_ao(1)%matrix)
            chk(2) = chk(2)+cp_dbcsr_checksum(op_ao(2)%matrix)
            chk(3) = chk(3)+cp_dbcsr_checksum(op_ao(3)%matrix)
            DO idir = 1, 3
               CALL cp_fm_set_all(rxp_psi0(ispin, idir)%matrix, 0.0_dp)
               CALL cp_dbcsr_sm_fm_multiply(op_ao(idir)%matrix, mo_coeff, &
                                            rxp_psi0(ispin, idir)%matrix, ncol=nmo, &
                                            alpha=-1.0_dp)
               DO j = center_list(ispin)%array(1, icenter), center_list(ispin)%array(1, icenter+1)-1
                  istate = center_list(ispin)%array(2, j)
                  ! the p_psi0 fm is used as temporary matrix to store the results for the psi0 centered in dk
                  CALL cp_fm_to_fm(rxp_psi0(ispin, idir)%matrix, &
                                   p_psi0(ispin, idir)%matrix, 1, istate, istate)
               ENDDO
            END DO
         ENDDO
         CALL cp_fm_to_fm(p_psi0(ispin, 1)%matrix, rxp_psi0(ispin, 1)%matrix)
         CALL cp_fm_to_fm(p_psi0(ispin, 2)%matrix, rxp_psi0(ispin, 2)%matrix)
         CALL cp_fm_to_fm(p_psi0(ispin, 3)%matrix, rxp_psi0(ispin, 3)%matrix)
      ENDDO
      !
      CALL cp_dbcsr_deallocate_matrix_set(op_ao)
      !
      ! print checksums
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E23.16)') 'CURRENT| current_operators: CheckSum L_x =', chk(1)
         WRITE (output_unit, '(T2,A,E23.16)') 'CURRENT| current_operators: CheckSum L_y =', chk(2)
         WRITE (output_unit, '(T2,A,E23.16)') 'CURRENT| current_operators: CheckSum L_z =', chk(3)
      ENDIF
      !
      ! Calculate the px py pz operators
      CALL cp_dbcsr_allocate_matrix_set(op_ao, 3)
      ALLOCATE (op_ao(1)%matrix, op_ao(2)%matrix, op_ao(3)%matrix)
      CALL cp_dbcsr_init(op_ao(1)%matrix)
      CALL cp_dbcsr_init(op_ao(2)%matrix)
      CALL cp_dbcsr_init(op_ao(3)%matrix)

      CALL cp_dbcsr_create(matrix=op_ao(1)%matrix, &
                           name="op_ao", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_antisymmetric, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0, mutable_work=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(op_ao(1)%matrix, sab_orb)
      CALL cp_dbcsr_set(op_ao(1)%matrix, 0.0_dp)

      DO idir = 2, 3
         CALL cp_dbcsr_copy(op_ao(idir)%matrix, op_ao(1)%matrix, &
                            "op_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir))))
         CALL cp_dbcsr_set(op_ao(idir)%matrix, 0.0_dp)
      ENDDO
      !
      CALL build_lin_mom_matrix(qs_env, op_ao)
      !CALL p_xyz_ao(op_ao,qs_env,minimum_image=.FALSE.)
      !
      ! print checksums
      chk(1) = cp_dbcsr_checksum(op_ao(1)%matrix)
      chk(2) = cp_dbcsr_checksum(op_ao(2)%matrix)
      chk(3) = cp_dbcsr_checksum(op_ao(3)%matrix)
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E23.16)') 'CURRENT| current_operators: CheckSum P_x =', chk(1)
         WRITE (output_unit, '(T2,A,E23.16)') 'CURRENT| current_operators: CheckSum P_y =', chk(2)
         WRITE (output_unit, '(T2,A,E23.16)') 'CURRENT| current_operators: CheckSum P_z =', chk(3)
      ENDIF
      ! Apply the p operator to the psi0
      DO idir = 1, 3
         DO ispin = 1, nspins
            mo_coeff => psi0_order(ispin)%matrix
            nmo = nstates(ispin)
            CALL cp_fm_set_all(p_psi0(ispin, idir)%matrix, 0.0_dp)
            CALL cp_dbcsr_sm_fm_multiply(op_ao(idir)%matrix, mo_coeff, &
                                         p_psi0(ispin, idir)%matrix, ncol=nmo, &
                                         alpha=-1.0_dp)
         END DO
      END DO
      !
      CALL cp_dbcsr_deallocate_matrix_set(op_ao)
      !
      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      !  This part is not necessary with the present implementation
      !  the angular momentum operator is computed directly for each dk independently
      !  and multiplied by the proper psi0 (i.e. those centered in dk)
      !  If Wannier centers are used, and no grouping of states with close centers is applied
      !  the (r-dk)xp operator is computed Nstate times and each time applied to only one vector psi0
      !
      ! Apply the (r-c)xp operator to the psi0
      !DO ispin = 1,nspins
      !  CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo, homo=homo)
      !  DO idir = 1,3
      !     CALL cp_fm_set_all(rxp_psi0(ispin,idir)%matrix,0.0_dp)
      !     CALL cp_sm_fm_multiply(op_rmd_ao(idir)%matrix,mo_coeff,&
      !            rxp_psi0(ispin,idir)%matrix,ncol=nmo,alpha=-1.0_dp)
      !  END DO
      !END DO

      !Calculate the second term of the operator state by state
      !!!! what follows is a way to avoid calculating the L matrix for each centers.
      !!!! not tested
      IF (.FALSE.) THEN
         DO ispin = 1, nspins
            !   Allocate full matrices as working storage in the calculation
            !   of the rxp operator matrix. 3 matrices for the 3 Cartesian direction
            !   plus one to apply the momentum oprator to the modified mos fm
            mo_coeff => psi0_order(ispin)%matrix
            nmo = nstates(ispin)
            NULLIFY (tmp_fm_struct)
            CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                     ncol_global=nmo, para_env=para_env, &
                                     context=mo_coeff%matrix_struct%context)
            DO idir = 1, 3
               NULLIFY (fm_Rmd_mos(idir)%matrix)
               CALL cp_fm_create(fm_Rmd_mos(idir)%matrix, tmp_fm_struct)
            END DO
            CALL cp_fm_create(fm_work1, tmp_fm_struct)
            CALL cp_fm_struct_release(tmp_fm_struct)

            ! This part should be done better, using the full matrix distribution
            DO istate = 1, nmo
               CALL cp_fm_get_submatrix(psi0_order(ispin)%matrix, vecbuf_c0, 1, istate, nao, 1, &
                                        transpose=.TRUE.)
               !center of the localized psi0 state istate
               dk(1:3) = centers_set(ispin)%array(1:3, istate)
               DO idir = 1, 3
                  !  This loop should be distributed over the processors
                  DO iao = 1, nao
                     ck(1:3) = basisfun_center(1:3, iao)
                     ckdk = pbc(dk, ck, cell)
                     vecbuf_Rmdc0(idir)%array(1, iao) = vecbuf_c0(1, iao)*ckdk(idir)
                  END DO ! iao
                  CALL cp_fm_set_submatrix(fm_Rmd_mos(idir)%matrix, vecbuf_Rmdc0(idir)%array, &
                                           1, istate, nao, 1, transpose=.TRUE.)
               END DO ! idir
            END DO ! istate

            DO idir = 1, 3
               CALL set_vecp(idir, ii, iii)

               !Add the second term to the idir component
               CALL cp_fm_set_all(fm_work1, 0.0_dp)
               CALL cp_dbcsr_sm_fm_multiply(op_ao(iii)%matrix, fm_Rmd_mos(ii)%matrix, &
                                            fm_work1, ncol=nmo, alpha=-1.0_dp)
               CALL cp_fm_scale_and_add(1.0_dp, rxp_psi0(ispin, idir)%matrix, &
                                        1.0_dp, fm_work1)

               CALL cp_fm_set_all(fm_work1, 0.0_dp)
               CALL cp_dbcsr_sm_fm_multiply(op_ao(ii)%matrix, fm_Rmd_mos(iii)%matrix, &
                                            fm_work1, ncol=nmo, alpha=-1.0_dp)
               CALL cp_fm_scale_and_add(1.0_dp, rxp_psi0(ispin, idir)%matrix, &
                                        -1.0_dp, fm_work1)

            END DO ! idir

            DO idir = 1, 3
               CALL cp_fm_release(fm_Rmd_mos(idir)%matrix)
            END DO
            CALL cp_fm_release(fm_work1)

         END DO ! ispin
      ENDIF

      DEALLOCATE (row_blk_sizes)

      DEALLOCATE (first_sgf, last_sgf)

      DEALLOCATE (vecbuf_c0)
      DO idir = 1, 3
         DEALLOCATE (vecbuf_Rmdc0(idir)%array)
      END DO

      CALL timestop(handle)

   END SUBROUTINE current_operators

! **************************************************************************************************
!> \brief ...
!> \param issc_env ...
!> \param qs_env ...
!> \param iatom ...
! **************************************************************************************************
   SUBROUTINE issc_operators(issc_env, qs_env, iatom)

      TYPE(issc_env_type)                                :: issc_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iatom

      CHARACTER(LEN=*), PARAMETER :: routineN = 'issc_operators', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idir, ispin, nmo, nspins, &
                                                            output_unit
      LOGICAL                                            :: do_dso, do_fc, do_pso, do_sd
      REAL(dp)                                           :: chk(20), r_i(3)
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_dso, matrix_efg, matrix_fc, &
                                                            matrix_pso
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fc_psi0
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: dso_psi0, efg_psi0, pso_psi0
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: lr_section

      CALL timeset(routineN, handle)

      NULLIFY (matrix_fc, matrix_pso, matrix_efg)
      NULLIFY (efg_psi0, pso_psi0, fc_psi0)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, &
                                               "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      linres_control=linres_control, &
                      para_env=para_env, &
                      mos=mos, &
                      particle_set=particle_set)

      nspins = dft_control%nspins

      CALL get_issc_env(issc_env=issc_env, &
                        matrix_efg=matrix_efg, & !this is used only here alloc/dealloc here???
                        matrix_pso=matrix_pso, & !this is used only here alloc/dealloc here???
                        matrix_fc=matrix_fc, & !this is used only here alloc/dealloc here???
                        matrix_dso=matrix_dso, & !this is used only here alloc/dealloc here???
                        efg_psi0=efg_psi0, &
                        pso_psi0=pso_psi0, &
                        dso_psi0=dso_psi0, &
                        fc_psi0=fc_psi0, &
                        do_fc=do_fc, &
                        do_sd=do_sd, &
                        do_pso=do_pso, &
                        do_dso=do_dso)
      !
      !
      r_i = particle_set(iatom)%r !pbc(particle_set(iatom)%r,cell)
      !write(*,*) 'issc_operators iatom=',iatom,' r_i=',r_i
      chk = 0.0_dp
      !
      !
      !
      ! Fermi contact integral
      !IF(do_fc) THEN
      IF (.TRUE.) THEN ! for the moment we build it (regs)
         CALL cp_dbcsr_set(matrix_fc(1)%matrix, 0.0_dp)
         CALL build_fermi_contact_matrix(qs_env, matrix_fc, r_i)

         chk(1) = cp_dbcsr_checksum(matrix_fc(1)%matrix)

         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| fermi_contact: CheckSum =', chk(1)
         ENDIF
      ENDIF
      !
      ! spin-orbit integral
      !IF(do_pso) THEN
      IF (.TRUE.) THEN ! for the moment we build it (regs)
         CALL cp_dbcsr_set(matrix_pso(1)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_pso(2)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_pso(3)%matrix, 0.0_dp)
         CALL build_pso_matrix(qs_env, matrix_pso, r_i)

         chk(2) = cp_dbcsr_checksum(matrix_pso(1)%matrix)
         chk(3) = cp_dbcsr_checksum(matrix_pso(2)%matrix)
         chk(4) = cp_dbcsr_checksum(matrix_pso(3)%matrix)

         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| pso_x: CheckSum =', chk(2)
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| pso_y: CheckSum =', chk(3)
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| pso_z: CheckSum =', chk(4)
         ENDIF
      ENDIF
      !
      ! electric field integral
      !IF(do_sd) THEN
      IF (.TRUE.) THEN ! for the moment we build it (regs)
         CALL cp_dbcsr_set(matrix_efg(1)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_efg(2)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_efg(3)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_efg(4)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_efg(5)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_efg(6)%matrix, 0.0_dp)
         CALL build_efg_matrix(qs_env, matrix_efg, r_i)

         chk(5) = cp_dbcsr_checksum(matrix_efg(1)%matrix)
         chk(6) = cp_dbcsr_checksum(matrix_efg(2)%matrix)
         chk(7) = cp_dbcsr_checksum(matrix_efg(3)%matrix)
         chk(8) = cp_dbcsr_checksum(matrix_efg(4)%matrix)
         chk(9) = cp_dbcsr_checksum(matrix_efg(5)%matrix)
         chk(10) = cp_dbcsr_checksum(matrix_efg(6)%matrix)

         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| efg (3xx-rr)/3: CheckSum =', chk(5)
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| efg (3yy-rr)/3: CheckSum =', chk(6)
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| efg (3zz-rr)/3: CheckSum =', chk(7)
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| efg xy: CheckSum =', chk(8)
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| efg xz: CheckSum =', chk(9)
            WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| efg yz: CheckSum =', chk(10)
         ENDIF
      ENDIF
      !
      !
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| all operator: CheckSum =', SUM(chk(1:10))
      ENDIF
      !
      !>>> debugging only  here we build the dipole matrix... debugging the kernel...
      IF (do_dso) THEN
         CALL cp_dbcsr_set(matrix_dso(1)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_dso(2)%matrix, 0.0_dp)
         CALL cp_dbcsr_set(matrix_dso(3)%matrix, 0.0_dp)
         CALL rRc_xyz_ao(matrix_dso, qs_env, (/0.0_dp, 0.0_dp, 0.0_dp/), 1)
      ENDIF
      !
      ! multiply by the mos
      DO ispin = 1, nspins
         !
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, ncol_global=nmo)
         !
         ! EFG
         IF (do_sd) THEN
            DO idir = 1, 6
               CALL cp_dbcsr_sm_fm_multiply(matrix_efg(idir)%matrix, mo_coeff, &
                                            efg_psi0(ispin, idir)%matrix, ncol=nmo, &
                                            alpha=1.0_dp)
            ENDDO
         ENDIF
         !
         ! PSO
         IF (do_pso) THEN
            DO idir = 1, 3
               CALL cp_dbcsr_sm_fm_multiply(matrix_pso(idir)%matrix, mo_coeff, &
                                            pso_psi0(ispin, idir)%matrix, ncol=nmo, &
                                            alpha=-1.0_dp)
            ENDDO
         ENDIF
         !
         ! FC
         IF (do_fc) THEN
            CALL cp_dbcsr_sm_fm_multiply(matrix_fc(1)%matrix, mo_coeff, &
                                         fc_psi0(ispin)%matrix, ncol=nmo, &
                                         alpha=1.0_dp)
         ENDIF
         !
         !>>> for debugging only
         IF (do_dso) THEN
            DO idir = 1, 3
               CALL cp_dbcsr_sm_fm_multiply(matrix_dso(idir)%matrix, mo_coeff, &
                                            dso_psi0(ispin, idir)%matrix, ncol=nmo, &
                                            alpha=-1.0_dp)
            ENDDO
         ENDIF
         !<<< for debugging only
      ENDDO

      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE issc_operators

! **************************************************************************************************
!> \brief Calculate the Berry phase operator in the AO basis and
!>         then the derivative of the Berry phase operator with respect to
!>         the ground state wave function (see paper Putrino et al., JCP, 13, 7102) for the AOs;
!>        afterwards multiply with the ground state MO coefficients
!>
!> \param polar_env ...
!> \param qs_env ...
!> \par History
!>      01.2013 created [SL]
!> \author SL
! **************************************************************************************************

   SUBROUTINE polar_operators(polar_env, qs_env)

      TYPE(polar_env_type)                               :: polar_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'polar_operators', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: one = (1.0_dp, 0.0_dp), &
                                                            zero = (0.0_dp, 0.0_dp)

      COMPLEX(DP)                                        :: zdet, zdeta
      INTEGER                                            :: handle, i, idim, ispin, nao, nmo, &
                                                            nspins, output_unit, tmp_dim, z
      LOGICAL                                            :: do_raman
      REAL(dp)                                           :: kvec(3), maxocc
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_cfm_p_type), DIMENSION(:), POINTER         :: eigrmat
      TYPE(cp_cfm_p_type), DIMENSION(:, :), POINTER      :: inv_mat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: cosmat, sinmat
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: dBerry_psi0, mo_derivs, op_fm_set, opvec
      TYPE(cp_fm_p_type), DIMENSION(:, :, :), POINTER    :: inv_work
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(section_vals_type), POINTER                   :: lr_section

      CALL timeset(routineN, handle)

      NULLIFY (dBerry_psi0, inv_work, mo_derivs, &
               op_fm_set, opvec, sinmat, cosmat)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, &
                                               "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")
      NULLIFY (cell, dft_control, mos, matrix_s)
      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      mos=mos, &
                      matrix_s=matrix_s)

      nspins = dft_control%nspins

      CALL get_polar_env(polar_env=polar_env, &
                         do_raman=do_raman, &
                         dBerry_psi0=dBerry_psi0, &
                         mo_derivs=mo_derivs)
!SL calculate dipole berry phase
      IF (do_raman) THEN

         DO i = 1, 3
            DO ispin = 1, nspins
               CALL cp_fm_set_all(dBerry_psi0(i, ispin)%matrix, 0.0_dp)
               CALL cp_fm_set_all(mo_derivs(i, ispin)%matrix, 0.0_dp)
            ENDDO
         END DO

         !
         !

         ! initialize all work matrices needed
         ALLOCATE (opvec(2, dft_control%nspins))
         ALLOCATE (op_fm_set(2, dft_control%nspins))
         ALLOCATE (eigrmat(dft_control%nspins))
         ALLOCATE (inv_mat(3, dft_control%nspins))
         ALLOCATE (inv_work(2, 3, dft_control%nspins))

         ! A bit to allocate for the wavefunction
         DO ispin = 1, dft_control%nspins
            NULLIFY (tmp_fm_struct, mo_coeff)
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo)
            CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nmo, &
                                     ncol_global=nmo, para_env=para_env, context=mo_coeff%matrix_struct%context)
            DO i = 1, SIZE(op_fm_set, 1)
               CALL cp_fm_create(opvec(i, ispin)%matrix, mo_coeff%matrix_struct)
               NULLIFY (op_fm_set(i, ispin)%matrix)
               CALL cp_fm_create(op_fm_set(i, ispin)%matrix, tmp_fm_struct)
            END DO
            CALL cp_cfm_create(eigrmat(ispin)%matrix, op_fm_set(1, ispin)%matrix%matrix_struct)
            CALL cp_fm_struct_release(tmp_fm_struct)
            DO i = 1, 3
               CALL cp_cfm_create(inv_mat(i, ispin)%matrix, op_fm_set(1, ispin)%matrix%matrix_struct)
               CALL cp_fm_create(inv_work(2, i, ispin)%matrix, op_fm_set(2, ispin)%matrix%matrix_struct)
               CALL cp_fm_create(inv_work(1, i, ispin)%matrix, op_fm_set(1, ispin)%matrix%matrix_struct)

            END DO
         END DO

         !
         NULLIFY (cosmat, sinmat)
         ALLOCATE (cosmat, sinmat)
         CALL cp_dbcsr_init(cosmat)
         CALL cp_dbcsr_init(sinmat)
         CALL cp_dbcsr_copy(cosmat, matrix_s(1)%matrix, 'COS MOM')
         CALL cp_dbcsr_copy(sinmat, matrix_s(1)%matrix, 'SIN MOM')
         CALL cp_dbcsr_set(cosmat, 0.0_dp)
         CALL cp_dbcsr_set(sinmat, 0.0_dp)

         DO i = 1, 3
            kvec(:) = twopi*cell%h_inv(i, :)
            CALL cp_dbcsr_set(cosmat, 0.0_dp)
            CALL cp_dbcsr_set(sinmat, 0.0_dp)
            CALL build_berry_moment_matrix(qs_env, cosmat, sinmat, kvec)

            DO ispin = 1, dft_control%nspins ! spin
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, mo_coeff=mo_coeff, nmo=nmo)

               CALL cp_dbcsr_sm_fm_multiply(cosmat, mo_coeff, opvec(1, ispin)%matrix, ncol=nmo)
               CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mo_coeff, opvec(1, ispin)%matrix, 0.0_dp, &
                            op_fm_set(1, ispin)%matrix)
               CALL cp_dbcsr_sm_fm_multiply(sinmat, mo_coeff, opvec(2, ispin)%matrix, ncol=nmo)
               CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mo_coeff, opvec(2, ispin)%matrix, 0.0_dp, &
                            op_fm_set(2, ispin)%matrix)

            ENDDO

            !second step invert C^T S_berry C
            zdet = one
            DO ispin = 1, dft_control%nspins
               CALL cp_cfm_get_info(eigrmat(ispin)%matrix, ncol_local=tmp_dim)
               DO idim = 1, tmp_dim
                  eigrmat(ispin)%matrix%local_data(:, idim) = &
                     CMPLX(op_fm_set(1, ispin)%matrix%local_data(:, idim), &
                           -op_fm_set(2, ispin)%matrix%local_data(:, idim), dp)
               END DO
               CALL cp_cfm_set_all(inv_mat(i, ispin)%matrix, zero, one)
               CALL cp_cfm_solve(eigrmat(ispin)%matrix, inv_mat(i, ispin)%matrix, zdeta)
            END DO
            !
            !
            !compute the derivative and add the result to mo_derivatives
            DO ispin = 1, dft_control%nspins
               CALL cp_cfm_get_info(eigrmat(ispin)%matrix, ncol_local=tmp_dim)
               CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, nmo=nmo, maxocc=maxocc)
               DO z = 1, tmp_dim
                  inv_work(1, i, ispin)%matrix%local_data(:, z) = REAL(inv_mat(i, ispin)%matrix%local_data(:, z), dp)
                  inv_work(2, i, ispin)%matrix%local_data(:, z) = AIMAG(inv_mat(i, ispin)%matrix%local_data(:, z))
               END DO
               CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, opvec(1, ispin)%matrix, inv_work(2, i, ispin)%matrix, &
                            0.0_dp, mo_derivs(i, ispin)%matrix)
               CALL cp_gemm("N", "N", nao, nmo, nmo, -1.0_dp, opvec(2, ispin)%matrix, inv_work(1, i, ispin)%matrix, &
                            1.0_dp, mo_derivs(i, ispin)%matrix)

               CALL cp_gemm("N", "N", nao, nmo, nmo, -1.0_dp, opvec(1, ispin)%matrix, inv_work(2, i, ispin)%matrix, &
                            0.0_dp, dBerry_psi0(i, ispin)%matrix)
               CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, opvec(2, ispin)%matrix, inv_work(1, i, ispin)%matrix, &
                            1.0_dp, dBerry_psi0(i, ispin)%matrix)

            END DO
         END DO !x/y/z-direction
!SL we omit here the multiplication with hmat (this scaling back done at the end of the response calc)

         DO ispin = 1, dft_control%nspins
            CALL cp_cfm_release(eigrmat(ispin)%matrix)
            DO i = 1, 3
               CALL cp_cfm_release(inv_mat(i, ispin)%matrix)
               CALL cp_fm_release(inv_work(1, i, ispin)%matrix)
               CALL cp_fm_release(inv_work(2, i, ispin)%matrix)
            END DO
            CALL cp_fm_release(opvec(1, ispin)%matrix)
            CALL cp_fm_release(opvec(2, ispin)%matrix)
            CALL cp_fm_release(op_fm_set(2, ispin)%matrix)
            CALL cp_fm_release(op_fm_set(1, ispin)%matrix)
         END DO

         CALL cp_dbcsr_deallocate_matrix(cosmat)
         CALL cp_dbcsr_deallocate_matrix(sinmat)

         DEALLOCATE (inv_work)
         DEALLOCATE (opvec)
         DEALLOCATE (op_fm_set)
         DEALLOCATE (eigrmat)
         DEALLOCATE (inv_mat)

      ENDIF ! raman
      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE polar_operators

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param c ...
!> \retval factor ...
! **************************************************************************************************
   FUNCTION fac_vecp(a, b, c) RESULT(factor)

      INTEGER                                            :: a, b, c
      REAL(dp)                                           :: factor

      factor = 0.0_dp

      IF ((b .EQ. a+1 .OR. b .EQ. a-2) .AND. (c .EQ. b+1 .OR. c .EQ. b-2)) THEN
         factor = 1.0_dp
      ELSEIF ((b .EQ. a-1 .OR. b .EQ. a+2) .AND. (c .EQ. b-1 .OR. c .EQ. b+2)) THEN
         factor = -1.0_dp
      END IF

   END FUNCTION fac_vecp

! **************************************************************************************************
!> \brief ...
!> \param ii ...
!> \param iii ...
!> \retval i ...
! **************************************************************************************************
   FUNCTION ind_m2(ii, iii) RESULT(i)

      INTEGER                                            :: ii, iii, i

      INTEGER                                            :: l(3)

      i = 0
      l(1:3) = 0
      IF (ii == 0) THEN
         l(iii) = 1
      ELSEIF (iii == 0) THEN
         l(ii) = 1
      ELSEIF (ii == iii) THEN
         l(ii) = 2
         i = coset(l(1), l(2), l(3))-1
      ELSE
         l(ii) = 1
         l(iii) = 1
      ENDIF
      i = coset(l(1), l(2), l(3))-1
   END FUNCTION ind_m2

! **************************************************************************************************
!> \brief ...
!> \param i1 ...
!> \param i2 ...
!> \param i3 ...
! **************************************************************************************************
   SUBROUTINE set_vecp(i1, i2, i3)

      INTEGER, INTENT(IN)                                :: i1
      INTEGER, INTENT(OUT)                               :: i2, i3

      IF (i1 == 1) THEN
         i2 = 2
         i3 = 3
      ELSEIF (i1 == 2) THEN
         i2 = 3
         i3 = 1
      ELSEIF (i1 == 3) THEN
         i2 = 1
         i3 = 2
      ELSE
      END IF

   END SUBROUTINE set_vecp
! **************************************************************************************************
!> \brief ...
!> \param i1 ...
!> \param i2 ...
!> \param i3 ...
! **************************************************************************************************
   SUBROUTINE set_vecp_rev(i1, i2, i3)

      INTEGER, INTENT(IN)                                :: i1, i2
      INTEGER, INTENT(OUT)                               :: i3

      IF ((i1+i2) == 3) THEN
         i3 = 3
      ELSEIF ((i1+i2) == 4) THEN
         i3 = 2
      ELSEIF ((i1+i2) == 5) THEN
         i3 = 1
      ELSE
      END IF

   END SUBROUTINE set_vecp_rev

! **************************************************************************************************
!> \brief scale a matrix as a_ij = a_ij * pbc(rc(:,j),ra(:,i))(ixyz)
!> \param matrix ...
!> \param ra ...
!> \param rc ...
!> \param cell ...
!> \param ixyz ...
!> \author vw
! **************************************************************************************************
   SUBROUTINE fm_scale_by_pbc_AC(matrix, ra, rc, cell, ixyz)
      TYPE(cp_fm_type), POINTER                :: matrix
      REAL(KIND=dp), DIMENSION(:, :), &
         INTENT(in)                             :: ra, rc
      TYPE(cell_type), POINTER                 :: cell
      INTEGER, INTENT(IN)                      :: ixyz

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fm_scale_by_pbc_AC', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: handle, icol_local, irow_local, m, mypcol, myprow, n, &
                 ncol_block, ncol_global, ncol_local, npcol, nprow, nrow_block, &
                 nrow_global, nrow_local
      REAL(KIND=dp)                            :: dist(3), rra(3), rrc(3)
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a

      CALL timeset(routineN, handle)

      myprow = matrix%matrix_struct%context%mepos(1)
      mypcol = matrix%matrix_struct%context%mepos(2)
      nprow = matrix%matrix_struct%context%num_pe(1)
      npcol = matrix%matrix_struct%context%num_pe(2)

      nrow_block = matrix%matrix_struct%nrow_block
      ncol_block = matrix%matrix_struct%ncol_block
      nrow_global = matrix%matrix_struct%nrow_global
      ncol_global = matrix%matrix_struct%ncol_global
      nrow_local = matrix%matrix_struct%nrow_locals(myprow)
      ncol_local = matrix%matrix_struct%ncol_locals(mypcol)

      n = SIZE(rc, 2)
      m = SIZE(ra, 2)

      a => matrix%local_data
      DO icol_local = 1, MIN(n, ncol_global)
         rrc = rc(:, icol_local)
         DO irow_local = 1, MIN(m, nrow_global)
            rra = ra(:, irow_local)
            dist = pbc(rrc, rra, cell)
            a(irow_local, icol_local) = a(irow_local, icol_local)*dist(ixyz)
         ENDDO
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE fm_scale_by_pbc_AC

END MODULE qs_linres_op

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module that builds the second order perturbation kernel
!>      kpp1 = delta_rho|_P delta_rho|_P E drho(P1) drho
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qs_kpp1_env_methods
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_scale,&
                                              cp_dbcsr_set
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE hartree_local_methods,           ONLY: Vh_1c_gg_integrals
   USE input_constants,                 ONLY: do_method_gapw,&
                                              do_method_gapw_xc,&
                                              tddfpt_excitations,&
                                              tddfpt_singlet,&
                                              tddfpt_triplet
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_get_rval,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_integrate_function,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_create,&
                                              pw_p_type,&
                                              pw_release,&
                                              pw_retain
   USE qs_energy_types,                 ONLY: allocate_qs_energy,&
                                              deallocate_qs_energy,&
                                              qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gapw_densities,               ONLY: prepare_gapw_den
   USE qs_integrate_potential,          ONLY: integrate_v_rspace
   USE qs_kpp1_env_types,               ONLY: qs_kpp1_env_type
   USE qs_ks_atom,                      ONLY: update_ks_atom
   USE qs_ks_methods,                   ONLY: qs_ks_build_kohn_sham_matrix
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_rho0_ggrid,                   ONLY: integrate_vhg0_rspace
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_vxc_atom,                     ONLY: calculate_xc_2nd_deriv_atom
   USE xc,                              ONLY: xc_calc_2nd_deriv,&
                                              xc_prep_2nd_deriv
   USE xc_derivative_set_types,         ONLY: xc_dset_release
   USE xc_derivatives,                  ONLY: xc_functionals_get_needs
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_create,&
                                              xc_rho_set_release,&
                                              xc_rho_set_type,&
                                              xc_rho_set_update
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_kpp1_env_methods'
   INTEGER, PRIVATE, SAVE :: last_kpp1_id_nr = 0

   PUBLIC :: kpp1_create, &
             kpp1_calc_k_p_p1, &
             kpp1_calc_k_p_p1_fdiff, &
             kpp1_did_change

CONTAINS

! **************************************************************************************************
!> \brief allocates and initializes a kpp1_env
!> \param kpp1_env the environement to initialize
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE kpp1_create(kpp1_env)
      TYPE(qs_kpp1_env_type), POINTER                    :: kpp1_env

      CHARACTER(len=*), PARAMETER :: routineN = 'kpp1_create', routineP = moduleN//':'//routineN

      ALLOCATE (kpp1_env)
      NULLIFY (kpp1_env%v_rspace, kpp1_env%v_ao, kpp1_env%drho_r, &
               kpp1_env%rho_set, &
               kpp1_env%deriv_set, kpp1_env%spin_pot, kpp1_env%grad_pot, &
               kpp1_env%ndiag_term)
      kpp1_env%ref_count = 1
      last_kpp1_id_nr = last_kpp1_id_nr+1
      kpp1_env%id_nr = last_kpp1_id_nr
      kpp1_env%iter = 0
      kpp1_env%print_count = 0
   END SUBROUTINE kpp1_create

! **************************************************************************************************
!> \brief calcualtes the k_p_p1 kernel of the pertubation theory
!> \param kpp1_env the kpp1 environement in which perform the calculation
!> \param p_env ...
!> \param qs_env kpp1's qs_env
!> \param k_p_p1 the sparse matrix that will contain the kernel k_p_p1
!> \param rho the density where to evaluate the derivatives (i.e. p along
!>        with with its grid representations, that must be valid)
!> \param rho1 the density that represent the first direction along which
!>        you should evaluate the derivatives
!> \param rho1_xc ...
! **************************************************************************************************
   SUBROUTINE kpp1_calc_k_p_p1(kpp1_env, p_env, qs_env, k_p_p1, rho, rho1, rho1_xc)

      TYPE(qs_kpp1_env_type), POINTER                    :: kpp1_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: k_p_p1
      TYPE(qs_rho_type), POINTER                         :: rho, rho1
      TYPE(qs_rho_type), OPTIONAL, POINTER               :: rho1_xc

      CHARACTER(len=*), PARAMETER :: routineN = 'kpp1_calc_k_p_p1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: excitations, handle, ispin, ns, nspins, &
                                                            output_unit, res_etype
      INTEGER, DIMENSION(2, 3)                           :: bo
      LOGICAL                                            :: explicit, gapw, gapw_xc, ionode, lsd, &
                                                            lsd_singlets
      REAL(KIND=dp)                                      :: energy_hartree, energy_hartree_1c, fac
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ksmat, psmat
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rho1_tot_gspace, v_hartree_gspace, &
                                                            v_hartree_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho1_g, rho1_g_pw, rho1_r, rho1_r_pw, &
                                                            tau_pw, v_rspace_new, v_xc
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(section_vals_type), POINTER                   :: input, scf_section, xc_fun_section, &
                                                            xc_section
      TYPE(xc_rho_cflags_type)                           :: needs
      TYPE(xc_rho_set_type), POINTER                     :: rho1_set

      CALL timeset(routineN, handle)

      NULLIFY (auxbas_pw_pool, pw_pools, pw_env, v_rspace_new, &
               rho1_r, rho1_g_pw, tau_pw, v_xc, rho1_set, &
               poisson_env, input, scf_section, rho1_g, rho_ao)
      logger => cp_get_default_logger()
      ionode = logger%para_env%mepos == logger%para_env%source

      IF(.NOT.(ASSOCIATED(kpp1_env)))CALL cp__a("qs_kpp1_env_methods.F",182)
      IF(.NOT.(ASSOCIATED(k_p_p1)))CALL cp__a("qs_kpp1_env_methods.F",183)
      IF(.NOT.(ASSOCIATED(rho)))CALL cp__a("qs_kpp1_env_methods.F",184)
      IF(.NOT.(ASSOCIATED(rho1)))CALL cp__a("qs_kpp1_env_methods.F",185)

      IF(.NOT.(kpp1_env%ref_count > 0))CALL cp__a("qs_kpp1_env_methods.F",187)
      CALL kpp1_check_i_alloc(kpp1_env, qs_env=qs_env)
      CALL get_qs_env(qs_env=qs_env, &
                      pw_env=pw_env, &
                      input=input)

      CALL qs_rho_get(rho, rho_ao=rho_ao)
      CALL qs_rho_get(rho1, rho_g=rho1_g)

      gapw = (section_get_ival(input, "DFT%QS%METHOD") == do_method_gapw)
      gapw_xc = (section_get_ival(input, "DFT%QS%METHOD") == do_method_gapw_xc)
      IF (gapw_xc) THEN
         IF(.NOT.(ASSOCIATED(rho1_xc)))CALL cp__a("qs_kpp1_env_methods.F",199)
      END IF

      nspins = SIZE(k_p_p1)
      lsd = (nspins == 2)

      xc_section => section_vals_get_subs_vals(input, "DFT%XC")
      scf_section => section_vals_get_subs_vals(input, "DFT%SCF")
      CALL section_vals_val_get(input, "DFT%EXCITATIONS", &
                                i_val=excitations)
      IF (excitations == tddfpt_excitations) THEN
         xc_section => section_vals_get_subs_vals(input, "DFT%TDDFPT%XC")
         !FM this check should already had happened and section made explicit, give an error?
         CALL section_vals_get(xc_section, explicit=explicit)
         IF (.NOT. explicit) THEN
            xc_section => section_vals_get_subs_vals(input, "DFT%XC")
         END IF
      END IF

      CALL section_vals_val_get(input, "DFT%TDDFPT%LSD_SINGLETS", &
                                l_val=lsd_singlets)
      CALL section_vals_val_get(input, "DFT%TDDFPT%RES_ETYPE", &
                                i_val=res_etype)

      kpp1_env%iter = kpp1_env%iter+1

! gets the tmp grids
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_kpp1_env_methods.F",226)
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools, poisson_env=poisson_env)
      ALLOCATE (v_rspace_new(nspins))
      CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, v_hartree_rspace%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      IF (gapw .OR. gapw_xc) &
         CALL prepare_gapw_den(qs_env, p_env%local_rho_set, do_rho0=(.NOT. gapw_xc))

! *** calculate the hartree potential on the total density ***

      CALL pw_pool_create_pw(auxbas_pw_pool, rho1_tot_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      CALL pw_copy(rho1_g(1)%pw, rho1_tot_gspace%pw)
      DO ispin = 2, nspins
         CALL pw_axpy(rho1_g(ispin)%pw, rho1_tot_gspace%pw)
      END DO
      IF (gapw) &
         CALL pw_axpy(p_env%local_rho_set%rho0_mpole%rho0_s_gs%pw, rho1_tot_gspace%pw)

      IF (cp_print_key_should_output(logger%iter_info, scf_section, "PRINT%TOTAL_DENSITIES") &
          /= 0) THEN
         output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%TOTAL_DENSITIES", &
                                            extension=".scfLog")
         CALL print_densities(rho1, rho1_tot_gspace, output_unit)
         CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                           "PRINT%TOTAL_DENSITIES")
      END IF

      IF (.NOT. (nspins == 1 .AND. excitations == tddfpt_excitations .AND. &
                 res_etype /= tddfpt_singlet)) THEN
         CALL pw_poisson_solve(poisson_env, rho1_tot_gspace%pw, &
                               energy_hartree, &
                               v_hartree_gspace%pw)
         CALL pw_transfer(v_hartree_gspace%pw, v_hartree_rspace%pw)
      END IF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho1_tot_gspace%pw)

! *** calculate the xc potential ***
      IF (gapw_xc) THEN
         CALL qs_rho_get(rho1_xc, rho_r=rho1_r)
      ELSE
         CALL qs_rho_get(rho1, rho_r=rho1_r)
      END IF

      IF (nspins == 1 .AND. excitations == tddfpt_excitations .AND. &
          (lsd_singlets .OR. res_etype == tddfpt_triplet)) THEN

         lsd = .TRUE.
         ALLOCATE (rho1_r_pw(2))
         DO ispin = 1, 2
            NULLIFY (rho1_r_pw(ispin)%pw)
            CALL pw_create(rho1_r_pw(ispin)%pw, rho1_r(1)%pw%pw_grid, &
                           rho1_r(1)%pw%in_use, rho1_r(1)%pw%in_space)
            CALL pw_transfer(rho1_r(1)%pw, rho1_r_pw(ispin)%pw)
         END DO

      ELSE

         ALLOCATE (rho1_r_pw(nspins))
         DO ispin = 1, nspins
            rho1_r_pw(ispin)%pw => rho1_r(ispin)%pw
            CALL pw_retain(rho1_r_pw(ispin)%pw)
         END DO

      END IF

      NULLIFY (tau_pw)

      !------!
      ! rho1 !
      !------!
      bo = rho1_r(1)%pw%pw_grid%bounds_local
      ! create the place where to store the argument for the functionals
      CALL xc_rho_set_create(rho1_set, bo, &
                             rho_cutoff=section_get_rval(xc_section, "DENSITY_CUTOFF"), &
                             drho_cutoff=section_get_rval(xc_section, "GRADIENT_CUTOFF"), &
                             tau_cutoff=section_get_rval(xc_section, "TAU_CUTOFF"))

      xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")
      needs = xc_functionals_get_needs(xc_fun_section, lsd, .TRUE.)

      ! calculate the arguments needed by the functionals
      CALL xc_rho_set_update(rho1_set, rho1_r_pw, rho1_g_pw, tau_pw, needs, &
                             section_get_ival(xc_section, "XC_GRID%XC_DERIV"), &
                             section_get_ival(xc_section, "XC_GRID%XC_SMOOTH_RHO"), &
                             auxbas_pw_pool)

      ALLOCATE (v_xc(nspins))
      DO ispin = 1, nspins
         NULLIFY (v_xc(ispin)%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, v_xc(ispin)%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_zero(v_xc(ispin)%pw)
      END DO

      fac = 0._dp
      IF (nspins == 1 .AND. excitations == tddfpt_excitations) THEN
         IF (lsd_singlets) fac = 1.0_dp
         IF (res_etype == tddfpt_triplet) fac = -1.0_dp
      END IF

      CALL xc_calc_2nd_deriv(v_xc, kpp1_env%deriv_set, kpp1_env%rho_set, &
                             rho1_set, auxbas_pw_pool, xc_section=xc_section, &
                             tddfpt_fac=fac)

      DO ispin = 1, nspins
         v_rspace_new(ispin)%pw => v_xc(ispin)%pw
      END DO
      DEALLOCATE (v_xc)

      IF (gapw) CALL calculate_xc_2nd_deriv_atom(p_env, qs_env, xc_section)

      CALL xc_rho_set_release(rho1_set)

      DO ispin = 1, SIZE(rho1_r_pw)
         CALL pw_release(rho1_r_pw(ispin)%pw)
      END DO
      DEALLOCATE (rho1_r_pw)

      !-------------------------------!
      ! Add both hartree and xc terms !
      !-------------------------------!
      DO ispin = 1, nspins

         IF (gapw_xc) THEN
            ! XC and Hartree are integrated separatedly
            ! XC uses the sofft basis set only
            v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d* &
                                          v_rspace_new(ispin)%pw%pw_grid%dvol

            IF (excitations == tddfpt_excitations .AND. nspins == 1) THEN

               IF (.NOT. (lsd_singlets .OR. &
                          res_etype == tddfpt_triplet)) THEN

                  v_rspace_new(1)%pw%cr3d = 2.0_dp*v_rspace_new(1)%pw%cr3d

               END IF
               ! remove kpp1_env%v_ao and work directly on k_p_p1 ?
               CALL cp_dbcsr_set(kpp1_env%v_ao(ispin)%matrix, 0.0_dp)
               CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                       pmat=rho_ao(ispin), &
                                       hmat=kpp1_env%v_ao(ispin), &
                                       qs_env=qs_env, &
                                       calculate_forces=.FALSE., gapw=gapw_xc)

               ! add hartree only for SINGLETS
               IF (res_etype == tddfpt_singlet) THEN
                  v_hartree_rspace%pw%cr3d = v_hartree_rspace%pw%cr3d* &
                                             v_hartree_rspace%pw%pw_grid%dvol
                  v_rspace_new(1)%pw%cr3d = 2.0_dp*v_hartree_rspace%pw%cr3d

                  CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                          pmat=rho_ao(ispin), &
                                          hmat=kpp1_env%v_ao(ispin), &
                                          qs_env=qs_env, &
                                          calculate_forces=.FALSE., gapw=gapw)
               END IF
            ELSE
               ! remove kpp1_env%v_ao and work directly on k_p_p1 ?
               CALL cp_dbcsr_set(kpp1_env%v_ao(ispin)%matrix, 0.0_dp)
               CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                       pmat=rho_ao(ispin), &
                                       hmat=kpp1_env%v_ao(ispin), &
                                       qs_env=qs_env, &
                                       calculate_forces=.FALSE., gapw=gapw_xc)

               IF (ispin == 1) THEN
                  v_hartree_rspace%pw%cr3d = v_hartree_rspace%pw%cr3d* &
                                             v_hartree_rspace%pw%pw_grid%dvol
               END IF
               v_rspace_new(ispin)%pw%cr3d = v_hartree_rspace%pw%cr3d
               CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                       pmat=rho_ao(ispin), &
                                       hmat=kpp1_env%v_ao(ispin), &
                                       qs_env=qs_env, &
                                       calculate_forces=.FALSE., gapw=gapw)
            END IF

         ELSE
            v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d* &
                                          v_rspace_new(ispin)%pw%pw_grid%dvol

            IF (excitations == tddfpt_excitations .AND. nspins == 1) THEN

               IF (.NOT. (lsd_singlets .OR. &
                          res_etype == tddfpt_triplet)) THEN

                  v_rspace_new(1)%pw%cr3d = 2.0_dp*v_rspace_new(1)%pw%cr3d

               END IF

               ! add hartree only for SINGLETS
               IF (res_etype == tddfpt_singlet) THEN
                  v_hartree_rspace%pw%cr3d = v_hartree_rspace%pw%cr3d* &
                                             v_hartree_rspace%pw%pw_grid%dvol
                  v_rspace_new(1)%pw%cr3d = v_rspace_new(1)%pw%cr3d+ &
                                            2.0_dp*v_hartree_rspace%pw%cr3d
               END IF
            ELSE
               IF (ispin == 1) THEN
                  v_hartree_rspace%pw%cr3d = v_hartree_rspace%pw%cr3d* &
                                             v_hartree_rspace%pw%pw_grid%dvol
               END IF
               v_rspace_new(ispin)%pw%cr3d = v_rspace_new(ispin)%pw%cr3d+ &
                                             v_hartree_rspace%pw%cr3d
            END IF

            ! remove kpp1_env%v_ao and work directly on k_p_p1 ?
            CALL cp_dbcsr_set(kpp1_env%v_ao(ispin)%matrix, 0.0_dp)
            CALL integrate_v_rspace(v_rspace=v_rspace_new(ispin), &
                                    pmat=rho_ao(ispin), &
                                    hmat=kpp1_env%v_ao(ispin), &
                                    qs_env=qs_env, &
                                    calculate_forces=.FALSE., gapw=gapw)

         END IF

         CALL cp_dbcsr_copy(k_p_p1(ispin)%matrix, kpp1_env%v_ao(ispin)%matrix)
      END DO

      IF (gapw) THEN
         IF (.NOT. (excitations == tddfpt_excitations .AND. (nspins == 1 .AND. &
                                                             res_etype == tddfpt_triplet))) THEN
            CALL Vh_1c_gg_integrals(qs_env, energy_hartree_1c, .TRUE., p_env=p_env)
            CALL integrate_vhg0_rspace(qs_env, v_hartree_rspace, &
                                       .FALSE., .TRUE., p_env=p_env)
         END IF
!         ***  Add single atom contributions to the KS matrix ***
         ! remap pointer
         ns = SIZE(p_env%kpp1)
         ksmat(1:ns, 1:1) => k_p_p1(1:ns)
         ns = SIZE(rho_ao)
         psmat(1:ns, 1:1) => rho_ao(1:ns)
         CALL update_ks_atom(qs_env, ksmat, psmat, .FALSE., .TRUE., p_env)
      END IF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_gspace%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, v_hartree_rspace%pw)
      DO ispin = 1, nspins
         CALL pw_pool_give_back_pw(auxbas_pw_pool, v_rspace_new(ispin)%pw)
      END DO
      DEALLOCATE (v_rspace_new)

      CALL timestop(handle)
   END SUBROUTINE kpp1_calc_k_p_p1

! **************************************************************************************************
!> \brief calcualtes the k_p_p1 kernel of the pertubation theory with finite
!>      differences
!> \param qs_env kpp1's qs_env
!> \param k_p_p1 the sparse matrix that will contain the kernel k_p_p1
!> \param rho the density where to evaluate the derivatives (i.e. p along
!>        with with its grid representations, that must be valid)
!> \param rho1 the density that represent the first direction along which
!>        you should evaluate the derivatives
!> \param diff the amount of the finite difference step
!> \par History
!>      01.2003 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      useful for testing purposes.
!>      rescale my_diff depending on the norm of rho1?
! **************************************************************************************************
   SUBROUTINE kpp1_calc_k_p_p1_fdiff(qs_env, k_p_p1, rho, rho1, &
                                     diff)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: k_p_p1
      TYPE(qs_rho_type), POINTER                         :: rho, rho1
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: diff

      CHARACTER(len=*), PARAMETER :: routineN = 'kpp1_calc_k_p_p1_fdiff', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nspins
      REAL(KIND=dp)                                      :: my_diff
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_2, matrix_s, rho1_ao, rho_ao
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho1_g, rho1_r, rho_g, rho_r
      TYPE(qs_energy_type), POINTER                      :: qs_energy

      NULLIFY (ks_2, matrix_s, qs_energy, rho_ao, rho1_ao, rho_r, rho1_r, rho_g, rho1_g)
      nspins = SIZE(k_p_p1)
      my_diff = 1.0e-6_dp
      IF (PRESENT(diff)) my_diff = diff
      CALL allocate_qs_energy(qs_energy)

      CALL qs_rho_get(rho, rho_ao=rho_ao, rho_r=rho_r, rho_g=rho_g)
      CALL qs_rho_get(rho1, rho_ao=rho1_ao, rho_r=rho1_r, rho_g=rho1_g)
      CALL get_qs_env(qs_env, matrix_s=matrix_s)

      ! rho = rho0+h/2*rho1
      my_diff = my_diff/2.0_dp
      DO ispin = 1, SIZE(k_p_p1)
         CALL cp_dbcsr_add(rho_ao(ispin)%matrix, rho1_ao(ispin)%matrix, &
                           alpha_scalar=1.0_dp, beta_scalar=my_diff)
         rho_r(ispin)%pw%cr3d = rho_r(ispin)%pw%cr3d+my_diff*rho1_r(ispin)%pw%cr3d
         rho_g(ispin)%pw%cc = rho_g(ispin)%pw%cc+my_diff*rho1_g(ispin)%pw%cc
      END DO

      CALL qs_ks_build_kohn_sham_matrix(qs_env, &
                                        ext_ks_matrix=k_p_p1, &
                                        calculate_forces=.FALSE., &
                                        just_energy=.FALSE.)

      CALL cp_dbcsr_allocate_matrix_set(ks_2, nspins)
      DO ispin = 1, nspins
         ALLOCATE (ks_2(ispin)%matrix)
         CALL cp_dbcsr_init(ks_2(ispin)%matrix)
         CALL cp_dbcsr_copy(ks_2(ispin)%matrix, matrix_s(1)%matrix, &
                            name="tmp_ks2-"//ADJUSTL(cp_to_string(ispin)))
      END DO

      ! rho = rho0-h/2*rho1
      my_diff = -2.0_dp*my_diff
      DO ispin = 1, nspins
         CALL cp_dbcsr_add(rho_ao(ispin)%matrix, rho1_ao(ispin)%matrix, &
                           alpha_scalar=1.0_dp, beta_scalar=my_diff)
         rho_r(ispin)%pw%cr3d = rho_r(ispin)%pw%cr3d+my_diff*rho1_r(ispin)%pw%cr3d
         rho_g(ispin)%pw%cc = rho_g(ispin)%pw%cc+my_diff*rho1_g(ispin)%pw%cc
      END DO

      CALL qs_ks_build_kohn_sham_matrix(qs_env, &
                                        ext_ks_matrix=ks_2, &
                                        calculate_forces=.FALSE., &
                                        just_energy=.FALSE.)

      ! rho = rho0
      my_diff = -0.5_dp*my_diff
      DO ispin = 1, nspins
         CALL cp_dbcsr_add(rho_ao(ispin)%matrix, rho1_ao(ispin)%matrix, &
                           alpha_scalar=1.0_dp, beta_scalar=my_diff)
         rho_r(ispin)%pw%cr3d = rho_r(ispin)%pw%cr3d+my_diff*rho1_r(ispin)%pw%cr3d
         rho_g(ispin)%pw%cc = rho_g(ispin)%pw%cc+my_diff*rho1_g(ispin)%pw%cc
      END DO

      ! k_p_p1=(H(rho0+h/2 rho1)-H(rho0-h/2 rho1))/h
      DO ispin = 1, nspins
         CALL cp_dbcsr_add(k_p_p1(ispin)%matrix, ks_2(ispin)%matrix, &
                           alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
         CALL cp_dbcsr_scale(k_p_p1(ispin)%matrix, alpha_scalar=0.5_dp/my_diff)
      END DO

      CALL cp_dbcsr_deallocate_matrix_set(ks_2)
      CALL deallocate_qs_energy(qs_energy)
   END SUBROUTINE kpp1_calc_k_p_p1_fdiff

! **************************************************************************************************
!> \brief checks that the intenal storage is allocated, and allocs it if needed
!> \param kpp1_env the environment to check
!> \param qs_env the qs environment this kpp1_env lives in
!> \author Fawzi Mohamed
!> \note
!>      private routine
! **************************************************************************************************
   SUBROUTINE kpp1_check_i_alloc(kpp1_env, qs_env)

      TYPE(qs_kpp1_env_type), POINTER                    :: kpp1_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'kpp1_check_i_alloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: excitations, ispin, nspins, res_etype
      LOGICAL                                            :: explicit, lsd_singlets
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: my_rho_r, rho_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: input, xc_section

! ------------------------------------------------------------------

      NULLIFY (pw_env, auxbas_pw_pool, matrix_s, rho, rho_r, input)

      IF(.NOT.(ASSOCIATED(kpp1_env)))CALL cp__a("qs_kpp1_env_methods.F",611)
      IF(.NOT.(kpp1_env%ref_count > 0))CALL cp__a("qs_kpp1_env_methods.F",612)

      CALL get_qs_env(qs_env, pw_env=pw_env, &
                      matrix_s=matrix_s, input=input, rho=rho)

      CALL qs_rho_get(rho, rho_r=rho_r)
      nspins = SIZE(rho_r)

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)

      IF (.NOT. ASSOCIATED(kpp1_env%v_rspace)) THEN
         ALLOCATE (kpp1_env%v_rspace(nspins))
         DO ispin = 1, nspins
            CALL pw_pool_create_pw(auxbas_pw_pool, &
                                   kpp1_env%v_rspace(ispin)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
         END DO
      END IF

      IF (.NOT. ASSOCIATED(kpp1_env%v_ao)) THEN
         CALL cp_dbcsr_allocate_matrix_set(kpp1_env%v_ao, nspins)
         DO ispin = 1, nspins
            ALLOCATE (kpp1_env%v_ao(ispin)%matrix)
            CALL cp_dbcsr_init(kpp1_env%v_ao(ispin)%matrix)
            CALL cp_dbcsr_copy(kpp1_env%v_ao(ispin)%matrix, matrix_s(1)%matrix, &
                               name="kpp1%v_ao-"//ADJUSTL(cp_to_string(ispin)))
         END DO
      END IF

      IF (.NOT. ASSOCIATED(kpp1_env%deriv_set)) THEN

         CALL section_vals_val_get(input, "DFT%EXCITATIONS", &
                                   i_val=excitations)
         CALL section_vals_val_get(input, "DFT%TDDFPT%LSD_SINGLETS", &
                                   l_val=lsd_singlets)
         CALL section_vals_val_get(input, "DFT%TDDFPT%RES_ETYPE", &
                                   i_val=res_etype)
         IF (nspins == 1 .AND. (excitations == tddfpt_excitations .AND. &
                                (lsd_singlets .OR. res_etype == tddfpt_triplet))) THEN
            ALLOCATE (my_rho_r(2))
            DO ispin = 1, 2
               CALL pw_pool_create_pw(auxbas_pw_pool, my_rho_r(ispin)%pw, &
                                      use_data=rho_r(1)%pw%in_use, in_space=rho_r(1)%pw%in_space)
               my_rho_r(ispin)%pw%cr3d = 0.5_dp*rho_r(1)%pw%cr3d
            END DO
         ELSE
            ALLOCATE (my_rho_r(SIZE(rho_r)))
            DO ispin = 1, SIZE(rho_r)
               my_rho_r(ispin)%pw => rho_r(ispin)%pw
               CALL pw_retain(my_rho_r(ispin)%pw)
            END DO
         END IF

         xc_section => section_vals_get_subs_vals(input, "DFT%XC")
         CALL section_vals_val_get(input, "DFT%EXCITATIONS", &
                                   i_val=excitations)
         IF (excitations == tddfpt_excitations) THEN
            xc_section => section_vals_get_subs_vals(input, "DFT%TDDFPT%XC")
!FM this check should already had happened and section made explicit, give an error?
            CALL section_vals_get(xc_section, explicit=explicit)
            IF (.NOT. explicit) THEN
               xc_section => section_vals_get_subs_vals(input, "DFT%XC")
            END IF
         END IF

         CALL xc_prep_2nd_deriv(kpp1_env%deriv_set, kpp1_env%rho_set, &
                                my_rho_r, auxbas_pw_pool, &
                                xc_section=xc_section)

         DO ispin = 1, SIZE(my_rho_r)
            CALL pw_release(my_rho_r(ispin)%pw)
         END DO
         DEALLOCATE (my_rho_r)
      END IF

   END SUBROUTINE kpp1_check_i_alloc

! **************************************************************************************************
!> \brief function to advise of changes either in the grids or in the the
!>      matrix structure (to update kpp1_env internal cache)
!> \param kpp1_env the kpp1_env
!> \param s_struct_changed true if the structure of the s matrix has changed
!> \param grid_changed true if the grids have changed and have to be rebuilt
!> \param psi0_changed true if the value of psi0 (qs_env%rho) has changed
!> \par History
!>      11.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE kpp1_did_change(kpp1_env, s_struct_changed, grid_changed, &
                              psi0_changed)
      TYPE(qs_kpp1_env_type), POINTER                    :: kpp1_env
      LOGICAL, INTENT(in), OPTIONAL                      :: s_struct_changed, grid_changed, &
                                                            psi0_changed

      CHARACTER(len=*), PARAMETER :: routineN = 'kpp1_did_change', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_psi0_changed, my_s_struct_changed

      my_s_struct_changed = .FALSE.
      my_psi0_changed = .FALSE.

      IF (PRESENT(s_struct_changed)) my_s_struct_changed = s_struct_changed
      IF (PRESENT(psi0_changed)) my_psi0_changed = psi0_changed

      IF(.NOT.(ASSOCIATED(kpp1_env)))CALL cp__a("qs_kpp1_env_methods.F",718)
      IF(.NOT.(kpp1_env%ref_count > 0))CALL cp__a("qs_kpp1_env_methods.F",719)
      IF (my_s_struct_changed) THEN
         IF (ASSOCIATED(kpp1_env%v_ao)) THEN
            CALL cp_dbcsr_deallocate_matrix_set(kpp1_env%v_ao)
         END IF
      END IF
      IF (my_s_struct_changed .OR. my_psi0_changed) THEN
         IF (ASSOCIATED(kpp1_env%drho_r)) THEN
            DEALLOCATE (kpp1_env%drho_r)
         END IF
         IF (ASSOCIATED(kpp1_env%deriv_set)) THEN
            CALL xc_dset_release(kpp1_env%deriv_set)
            NULLIFY (kpp1_env%deriv_set)
         END IF
         IF (ASSOCIATED(kpp1_env%spin_pot)) THEN
            DEALLOCATE (kpp1_env%spin_pot)
         END IF
         IF (ASSOCIATED(kpp1_env%grad_pot)) THEN
            DEALLOCATE (kpp1_env%grad_pot)
         END IF
         IF (ASSOCIATED(kpp1_env%ndiag_term)) THEN
            DEALLOCATE (kpp1_env%ndiag_term)
         END IF
         CALL xc_rho_set_release(kpp1_env%rho_set) ! it would be better to pass a pw pool
      END IF
      IF (PRESENT(grid_changed)) THEN
         IF (grid_changed) THEN
            IF (ASSOCIATED(kpp1_env%v_rspace)) THEN
               DO i = 1, SIZE(kpp1_env%v_rspace)
                  CALL pw_release(kpp1_env%v_rspace(i)%pw)
               END DO
               DEALLOCATE (kpp1_env%v_rspace)
            END IF
         END IF
      END IF
   END SUBROUTINE kpp1_did_change

! **************************************************************************************************
!> \brief ...
!> \param rho1 ...
!> \param rho1_tot_gspace ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE print_densities(rho1, rho1_tot_gspace, output_unit)

      TYPE(qs_rho_type), POINTER                         :: rho1
      TYPE(pw_p_type), INTENT(IN)                        :: rho1_tot_gspace
      INTEGER                                            :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'print_densities', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: total_rho_gspace
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho1_r

      NULLIFY (tot_rho1_r)

      total_rho_gspace = pw_integrate_function(rho1_tot_gspace%pw, isign=-1)
      IF (output_unit > 0) THEN
         CALL qs_rho_get(rho1, tot_rho_r=tot_rho1_r)
         WRITE (UNIT=output_unit, FMT="(T3,A,T60,F20.10)") &
            "KPP1 total charge density (r-space):", &
            accurate_sum(tot_rho1_r), &
            "KPP1 total charge density (g-space):", &
            total_rho_gspace
      END IF

   END SUBROUTINE print_densities

END MODULE qs_kpp1_env_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility functions for the perturbation calculations.
!> \note
!>      - routines are programmed with spins in mind
!>        but are as of now not tested with them
!> \par History
!>      22-08-2002, TCH, started development
! **************************************************************************************************
MODULE qs_p_env_methods
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_scale,&
                                              cp_dbcsr_set,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_plus_fm_fm_t,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add,&
                                              cp_fm_symm,&
                                              cp_fm_triangular_multiply
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              cp_fm_pool_type,&
                                              fm_pool_create_fm,&
                                              fm_pool_get_el_struct,&
                                              fm_pool_give_back_fm,&
                                              fm_pools_create_fm_vect
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_get,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_retain,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_fm_vect,                      ONLY: cp_fm_vect_copy,&
                                              cp_fm_vect_dealloc
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE hartree_local_methods,           ONLY: init_coulomb_local
   USE hartree_local_types,             ONLY: hartree_local_create
   USE input_constants,                 ONLY: ot_precond_none
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE preconditioner_types,            ONLY: init_preconditioner
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kpp1_env_methods,             ONLY: kpp1_calc_k_p_p1,&
                                              kpp1_calc_k_p_p1_fdiff,&
                                              kpp1_create,&
                                              kpp1_did_change
   USE qs_kpp1_env_types,               ONLY: qs_kpp1_env_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type
   USE qs_linres_types,                 ONLY: linres_control_type
   USE qs_local_rho_types,              ONLY: local_rho_set_create
   USE qs_matrix_pools,                 ONLY: mpools_get
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_rho0_methods,                 ONLY: init_rho0
   USE qs_rho_atom_methods,             ONLY: allocate_rho_atom_internals
   USE qs_rho_methods,                  ONLY: qs_rho_rebuild,&
                                              qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_create,&
                                              qs_rho_get,&
                                              qs_rho_type
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_p_env_methods'
   INTEGER, PRIVATE, SAVE :: last_p_env_id = 0
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   PRIVATE
   PUBLIC :: p_env_create, p_env_psi0_changed
   PUBLIC :: p_op_l1, p_op_l2, p_preortho, p_postortho

CONTAINS

! **************************************************************************************************
!> \brief allocates and initializes the perturbation environment (no setup)
!> \param p_env the environment to initialize
!> \param qs_env the qs_environment for the system
!> \param kpp1_env the environment that builds the second order
!>        perturbation kernel
!> \param p1_option ...
!> \param psi0d ...
!> \param orthogonal_orbitals if the orbitals are orthogonal
!> \param linres_control ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE p_env_create(p_env, qs_env, kpp1_env, p1_option, &
                           psi0d, orthogonal_orbitals, linres_control)

      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_kpp1_env_type), OPTIONAL, POINTER          :: kpp1_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: p1_option
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: psi0d
      LOGICAL, INTENT(in), OPTIONAL                      :: orthogonal_orbitals
      TYPE(linres_control_type), OPTIONAL, POINTER       :: linres_control

      CHARACTER(len=*), PARAMETER :: routineN = 'p_env_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, n_ao, n_mo, n_spins, natom, spin
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools, mo_mo_fm_pools
      TYPE(cp_fm_type), POINTER                          :: qs_env_c
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control

! code

      CALL timeset(routineN, handle)
      NULLIFY (ao_mo_fm_pools, mo_mo_fm_pools, matrix_s, dft_control, para_env, blacs_env)
      CALL get_qs_env(qs_env, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      blacs_env=blacs_env)

      n_spins = dft_control%nspins

      ALLOCATE (p_env)
      NULLIFY (p_env%kpp1, &
               p_env%p1, &
               p_env%m_epsilon, &
               p_env%psi0d, &
               p_env%S_psi0, &
               p_env%kpp1_env, &
               p_env%rho1, &
               p_env%rho1_xc, &
               p_env%Smo_inv, &
               p_env%local_rho_set, &
               p_env%hartree_local, &
               p_env%PS_psi0, &
               p_env%preconditioner, &
               p_env%ev_h0)

      p_env%ref_count = 1
      last_p_env_id = last_p_env_id+1
      p_env%id_nr = last_p_env_id
      p_env%iter = 0

      p_env%new_preconditioner = .TRUE.
      p_env%only_energy = .FALSE.
      p_env%os_valid = .FALSE.
      p_env%ls_count = 0
      p_env%delta = 0.0_dp
      p_env%gnorm = 0.0_dp
      p_env%gnorm_old = 0.0_dp
      p_env%etotal = 0.0_dp
      p_env%gradient = 0.0_dp

      CALL qs_rho_create(p_env%rho1)
      CALL qs_rho_create(p_env%rho1_xc)

      IF (PRESENT(kpp1_env)) THEN
         p_env%kpp1_env => kpp1_env
      ELSE
         CALL kpp1_create(p_env%kpp1_env)
      END IF

      IF (PRESENT(p1_option)) THEN
         p_env%p1 => p1_option
      ELSE

         CALL cp_dbcsr_allocate_matrix_set(p_env%p1, n_spins)
         DO spin = 1, n_spins
            ALLOCATE (p_env%p1(spin)%matrix)
            CALL cp_dbcsr_init(p_env%p1(spin)%matrix)
            CALL cp_dbcsr_copy(p_env%p1(spin)%matrix, matrix_s(1)%matrix, &
                               name="p_env"//TRIM(ADJUSTL(cp_to_string(p_env%id_nr)))// &
                               "%p1-"//TRIM(ADJUSTL(cp_to_string(spin))))
            CALL cp_dbcsr_set(p_env%p1(spin)%matrix, 0.0_dp)
         END DO
      END IF

      CALL mpools_get(qs_env%mpools, ao_mo_fm_pools=ao_mo_fm_pools, &
                      mo_mo_fm_pools=mo_mo_fm_pools)

      p_env%n_mo = 0
      p_env%n_ao = 0
      DO spin = 1, n_spins
         IF (PRESENT(psi0d)) THEN
            CALL cp_fm_get_info(psi0d(spin)%matrix, &
                                ncol_global=n_mo, nrow_global=n_ao)
         ELSE
            CALL get_mo_set(qs_env%mos(spin)%mo_set, mo_coeff=qs_env_c)
            CALL cp_fm_get_info(qs_env_c, &
                                ncol_global=n_mo, nrow_global=n_ao)
         END IF
         p_env%n_mo(spin) = n_mo
         p_env%n_ao(spin) = n_ao
      END DO

      p_env%orthogonal_orbitals = .FALSE.
      IF (PRESENT(orthogonal_orbitals)) &
         p_env%orthogonal_orbitals = orthogonal_orbitals

      CALL fm_pools_create_fm_vect(ao_mo_fm_pools, elements=p_env%S_psi0, &
                                   name="p_env"//TRIM(ADJUSTL(cp_to_string(p_env%id_nr)))//"%S_psi0")

      ! alloc m_epsilon
      CALL fm_pools_create_fm_vect(mo_mo_fm_pools, elements=p_env%m_epsilon, &
                                   name="p_env"//TRIM(ADJUSTL(cp_to_string(p_env%id_nr)))// &
                                   "%m_epsilon")

      ! alloc Smo_inv
      IF (.NOT. p_env%orthogonal_orbitals) THEN
         CALL fm_pools_create_fm_vect(mo_mo_fm_pools, elements=p_env%Smo_inv, &
                                      name="p_env"//TRIM(ADJUSTL(cp_to_string(p_env%id_nr)))// &
                                      "%Smo_inv")
      END IF

      IF (PRESENT(psi0d)) THEN
         IF (ASSOCIATED(psi0d)) THEN
            CALL cp_fm_vect_copy(psi0d, p_env%psi0d)
         END IF
      ELSE IF (.NOT. p_env%orthogonal_orbitals) THEN
         CALL fm_pools_create_fm_vect(ao_mo_fm_pools, &
                                      elements=p_env%psi0d, &
                                      name="p_env"//TRIM(ADJUSTL(cp_to_string(p_env%id_nr)))// &
                                      "%psi0d")
      END IF

      !----------------------!
      ! GAPW initializations !
      !----------------------!
      IF (dft_control%qs_control%gapw) THEN
         CALL local_rho_set_create(p_env%local_rho_set)
         CALL allocate_rho_atom_internals(qs_env, p_env%local_rho_set%rho_atom_set)
         CALL init_rho0(qs_env, dft_control%qs_control%gapw_control, &
                        .TRUE., p_env%local_rho_set)
         CALL hartree_local_create(p_env%hartree_local)
         CALL get_qs_env(qs_env=qs_env, natom=natom)
         CALL init_coulomb_local(p_env%hartree_local, natom)
      END IF

      !------------------------!
      ! LINRES initializations !
      !------------------------!
      IF (PRESENT(linres_control)) THEN

         IF (linres_control%preconditioner_type /= ot_precond_none) THEN
            ! Initialize the preconditioner matrix
            IF (.NOT. ASSOCIATED(p_env%preconditioner)) THEN

               ALLOCATE (p_env%preconditioner(n_spins))
               DO spin = 1, n_spins
                  CALL init_preconditioner(p_env%preconditioner(spin), &
                                           para_env=para_env, blacs_env=blacs_env)
               END DO
               p_env%os_valid = .FALSE.

               CALL fm_pools_create_fm_vect(ao_mo_fm_pools, elements=p_env%PS_psi0, &
                                            name="p_env"//TRIM(ADJUSTL(cp_to_string(p_env%id_nr)))//"%PS_psi0")

               ALLOCATE (p_env%ev_h0(n_spins))
!             CALL fm_pools_create_fm_vect(ao_mo_fm_pools,elements=p_env%ev_h0,&
!                  name="p_env"//TRIM(ADJUSTL(cp_to_string(p_env%id_nr)))//"%ev_h0")
            END IF
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE p_env_create

! **************************************************************************************************
!> \brief checks that the intenal storage is allocated, and allocs it if needed
!> \param p_env the environment to check
!> \param qs_env the qs environment this p_env lives in
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      private routine
! **************************************************************************************************
   SUBROUTINE p_env_check_i_alloc(p_env, qs_env)
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'p_env_check_i_alloc', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=25)                                  :: name
      INTEGER                                            :: handle, ispin, nspins
      LOGICAL                                            :: gapw_xc
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(dft_control_type), POINTER                    :: dft_control

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, matrix_s)

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_p_env_methods.F",327)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_methods.F",328)
      CALL get_qs_env(qs_env, dft_control=dft_control)
      gapw_xc = dft_control%qs_control%gapw_xc
      IF (.NOT. ASSOCIATED(p_env%kpp1)) THEN
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         nspins = dft_control%nspins

         CALL cp_dbcsr_allocate_matrix_set(p_env%kpp1, nspins)
         name = "p_env"//cp_to_string(p_env%id_nr)//"%kpp1-"
         CALL compress(name, full=.TRUE.)
         DO ispin = 1, nspins
            ALLOCATE (p_env%kpp1(ispin)%matrix)
            CALL cp_dbcsr_init(p_env%kpp1(ispin)%matrix)
            CALL cp_dbcsr_copy(p_env%kpp1(ispin)%matrix, matrix_s(1)%matrix, &
                               name=TRIM(name)//ADJUSTL(cp_to_string(ispin)))
         END DO

         CALL qs_rho_rebuild(p_env%rho1, qs_env=qs_env)
         IF (gapw_xc) THEN
            CALL qs_rho_rebuild(p_env%rho1_xc, qs_env=qs_env)
         END IF

      END IF
      CALL timestop(handle)
   END SUBROUTINE p_env_check_i_alloc

! **************************************************************************************************
!> \brief To be called after the value of psi0 has changed.
!>      Recalculates the quantities S_psi0 and m_epsilon.
!> \param p_env the perturbation environment to set
!> \param qs_env ...
!> \param psi0 the value of psi0, if not given defaults to the qs_env mos
!> \param Hrho_psi0d is given, then the partial result Hrho_psi0d is stored in
!>        that vector
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE p_env_psi0_changed(p_env, qs_env, psi0, Hrho_psi0d)

      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: psi0
      TYPE(cp_fm_p_type), DIMENSION(:), INTENT(inout), &
         OPTIONAL                                        :: Hrho_psi0d

      CHARACTER(len=*), PARAMETER :: routineN = 'p_env_psi0_changed', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, lfomo, n_spins, nmo, &
                                                            output_unit, spin
      LOGICAL                                            :: was_present
      REAL(KIND=dp)                                      :: maxocc
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s, rho_ao
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: my_psi0
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: input, lr_section

      CALL timeset(routineN, handle)

      NULLIFY (ao_mo_fm_pools, mos, my_psi0, matrix_s, mos, para_env, ks_env, rho, &
               logger, input, lr_section, energy, matrix_ks, dft_control, rho_ao)
      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_p_env_methods.F",400)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_methods.F",401)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      mos=mos, &
                      matrix_s=matrix_s, &
                      matrix_ks=matrix_ks, &
                      para_env=para_env, &
                      rho=rho, &
                      input=input, &
                      energy=energy, &
                      dft_control=dft_control)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      n_spins = dft_control%nspins
      p_env%iter = p_env%iter+1
      CALL mpools_get(qs_env%mpools, &
                      ao_mo_fm_pools=ao_mo_fm_pools)
      ! def my_psi0
      IF (PRESENT(psi0)) THEN
         CALL cp_fm_vect_copy(psi0, my_psi0)
      ELSE
         ALLOCATE (my_psi0(n_spins))
         DO spin = 1, n_spins
            NULLIFY (my_psi0(spin)%matrix)
            CALL get_mo_set(mos(spin)%mo_set, &
                            mo_coeff=my_psi0(spin)%matrix)
            CALL cp_fm_retain(my_psi0(spin)%matrix)
         END DO
      END IF

      lr_section => section_vals_get_subs_vals(input, "PROPERTIES%LINRES")
      ! def psi0d
      IF (p_env%orthogonal_orbitals) THEN
         IF (ASSOCIATED(p_env%psi0d)) THEN
            CALL cp_fm_vect_dealloc(p_env%psi0d)
         END IF
         p_env%psi0d => my_psi0
      ELSE

         DO spin = 1, n_spins
            ! m_epsilon=choleski_decomposition(my_psi0^T S my_psi0)^-1
            ! could be optimized by combining next two calls
            CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, &
                                         my_psi0(spin)%matrix, &
                                         p_env%S_psi0(spin)%matrix, &
                                         ncol=p_env%n_mo(spin), alpha=1.0_dp)
            CALL cp_gemm(transa='T', transb='N', n=p_env%n_mo(spin), &
                         m=p_env%n_mo(spin), k=p_env%n_ao(spin), alpha=1.0_dp, &
                         matrix_a=my_psi0(spin)%matrix, &
                         matrix_b=p_env%S_psi0(spin)%matrix, &
                         beta=0.0_dp, matrix_c=p_env%m_epsilon(spin)%matrix)
            CALL cp_fm_cholesky_decompose(p_env%m_epsilon(spin)%matrix, &
                                          n=p_env%n_mo(spin))

            ! Smo_inv= (my_psi0^T S my_psi0)^-1
            CALL cp_fm_set_all(p_env%Smo_inv(spin)%matrix, 0.0_dp, 1.0_dp)
            ! faster using cp_fm_cholesky_invert ?
            CALL cp_fm_triangular_multiply( &
               triangular_matrix=p_env%m_epsilon(spin)%matrix, &
               matrix_b=p_env%Smo_inv(spin)%matrix, side='R', &
               invert_tr=.TRUE., n_rows=p_env%n_mo(spin), &
               n_cols=p_env%n_mo(spin))
            CALL cp_fm_triangular_multiply( &
               triangular_matrix=p_env%m_epsilon(spin)%matrix, &
               matrix_b=p_env%Smo_inv(spin)%matrix, side='R', &
               transpose_tr=.TRUE., &
               invert_tr=.TRUE., n_rows=p_env%n_mo(spin), &
               n_cols=p_env%n_mo(spin))

            ! psi0d=my_psi0 (my_psi0^T S my_psi0)^-1
            ! faster using cp_fm_cholesky_invert ?
            CALL cp_fm_to_fm(my_psi0(spin)%matrix, &
                             p_env%psi0d(spin)%matrix)
            CALL cp_fm_triangular_multiply( &
               triangular_matrix=p_env%m_epsilon(spin)%matrix, &
               matrix_b=p_env%psi0d(spin)%matrix, side='R', &
               invert_tr=.TRUE., n_rows=p_env%n_ao(spin), &
               n_cols=p_env%n_mo(spin))
            CALL cp_fm_triangular_multiply( &
               triangular_matrix=p_env%m_epsilon(spin)%matrix, &
               matrix_b=p_env%psi0d(spin)%matrix, side='R', &
               transpose_tr=.TRUE., &
               invert_tr=.TRUE., n_rows=p_env%n_ao(spin), &
               n_cols=p_env%n_mo(spin))

            ! updates P
            CALL get_mo_set(mos(spin)%mo_set, lfomo=lfomo, &
                            nmo=nmo, maxocc=maxocc)
            IF (lfomo > nmo) THEN
               CALL cp_dbcsr_set(rho_ao(spin)%matrix, 0.0_dp)
               CALL cp_dbcsr_plus_fm_fm_t(rho_ao(spin)%matrix, &
                                          matrix_v=my_psi0(spin)%matrix, &
                                          matrix_g=p_env%psi0d(spin)%matrix, &
                                          ncol=p_env%n_mo(spin))
               CALL cp_dbcsr_scale(rho_ao(spin)%matrix, alpha_scalar=maxocc)
            ELSE
               CALL cp__b("qs_p_env_methods.F",499,"symmetrized onesided smearing to do")
            END IF
         END DO

         ! updates rho
         CALL qs_rho_update_rho(rho_struct=rho, qs_env=qs_env)

         ! tells ks_env that p changed
         CALL qs_ks_did_change(ks_env=ks_env, rho_changed=.TRUE.)

      END IF

      ! updates K (if necessary)
      CALL qs_ks_update_qs_env(qs_env)
      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")
      IF (output_unit > 0) THEN
         CALL section_vals_get(lr_section, explicit=was_present)
         IF (was_present) THEN
            WRITE (UNIT=output_unit, FMT="(/,(T3,A,T55,F25.14))") &
               "Total energy ground state:                     ", energy%total
         END IF
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")
      !-----------------------------------------------------------------------|
      ! calculates                                                            |
      ! m_epsilon = - psi0d^T times K times psi0d                             |
      !           = - [K times psi0d]^T times psi0d (because K is symmetric)  |
      !-----------------------------------------------------------------------|
      DO spin = 1, n_spins
         ! S_psi0 = k times psi0d
         CALL cp_dbcsr_sm_fm_multiply(matrix_ks(spin)%matrix, &
                                      p_env%psi0d(spin)%matrix, &
                                      p_env%S_psi0(spin)%matrix, p_env%n_mo(spin))
         IF (PRESENT(Hrho_psi0d)) THEN
            CALL cp_fm_scale_and_add(alpha=0.0_dp, matrix_a=Hrho_psi0d(spin)%matrix, &
                                     beta=1.0_dp, matrix_b=p_env%S_psi0(spin)%matrix)
         END IF
         ! m_epsilon = -1 times S_psi0^T times psi0d
         CALL cp_gemm('T', 'N', &
                      p_env%n_mo(spin), p_env%n_mo(spin), p_env%n_ao(spin), &
                      -1.0_dp, p_env%S_psi0(spin)%matrix, p_env%psi0d(spin)%matrix, &
                      0.0_dp, p_env%m_epsilon(spin)%matrix)
!   DO i =1,size(p_env%m_epsilon(spin)%matrix%local_data,1)
!    write(*,'(I4,4f12.6)') i,(p_env%m_epsilon(spin)%matrix%local_data(i,j), j=1,4)
!   end do
! stop 'o'
      END DO

      !----------------------------------|
      ! calculates S_psi0 = S * my_psi0  |
      !----------------------------------|
      ! calculating this reduces the mat mult without storing a full aoxao
      ! matrix (for P). If nspin>1 you might consider calculating it on the
      ! fly to spare some memory
      CALL get_qs_env(qs_env, matrix_s=matrix_s)
      DO spin = 1, n_spins
         CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, &
                                      my_psi0(spin)%matrix, &
                                      p_env%S_psi0(spin)%matrix, &
                                      p_env%n_mo(spin))
      END DO

      ! releases my_psi0
      IF (p_env%orthogonal_orbitals) THEN
         NULLIFY (my_psi0)
      ELSE
         CALL cp_fm_vect_dealloc(my_psi0)
      END IF

      ! tells kpp1_env about the change of psi0
      CALL kpp1_did_change(p_env%kpp1_env, psi0_changed=.TRUE.)

      CALL timestop(handle)

   END SUBROUTINE p_env_psi0_changed

! **************************************************************************************************
!> \brief Evaluates Fv (S_mo)^-1 - Sv(epsilon) and stores it in res
!> \param p_env perturbation calculation environment
!> \param qs_env the qs_env that is perturbed by this p_env
!> \param v the matrix to operate on
!> \param res the result
!> \par History
!>      10.2002, TCH, extracted single spin calculation
!> \author Thomas Chassaing
! **************************************************************************************************
   SUBROUTINE p_op_l1(p_env, qs_env, v, res)

      ! argument
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_p_type), DIMENSION(:), INTENT(in)       :: v
      TYPE(cp_fm_p_type), DIMENSION(:), INTENT(inout)    :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'p_op_l1', routineP = moduleN//':'//routineN

      INTEGER                                            :: n_spins, spin
      TYPE(dft_control_type), POINTER                    :: dft_control

! code

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_p_env_methods.F",602)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_methods.F",603)
      NULLIFY (dft_control)
      CALL get_qs_env(qs_env, dft_control=dft_control)

      n_spins = dft_control%nspins
      DO spin = 1, n_spins
         CALL p_op_l1_spin(p_env, qs_env, spin, v(spin)%matrix, &
                           res(spin)%matrix)
      END DO

   END SUBROUTINE p_op_l1

! **************************************************************************************************
!> \brief Evaluates Fv (S_mo)^-1 - Sv(epsilon) and stores it in res
!>      for a given spin
!> \param p_env perturbation calculation environment
!> \param qs_env the qs_env that is perturbed by this p_env
!> \param spin the spin to calculate (1 or 2 normally)
!> \param v the matrix to operate on
!> \param res the result
!> \par History
!>      10.2002, TCH, created
!> \author Thomas Chassaing
!> \note
!>      Same as p_op_l1 but takes a spin as additional argument.
! **************************************************************************************************
   SUBROUTINE p_op_l1_spin(p_env, qs_env, spin, v, res)

      ! argument
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: spin
      TYPE(cp_fm_type), POINTER                          :: v, res

      CHARACTER(len=*), PARAMETER :: routineN = 'p_op_l1_spin', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ncol
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: k_p
      TYPE(cp_fm_pool_type), POINTER                     :: maxao_maxmo_fm_pool
      TYPE(cp_fm_type), POINTER                          :: tmp
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control

! code

      CALL timeset(routineN, handle)
      NULLIFY (tmp, matrix_ks, matrix_s, dft_control)

      CALL get_qs_env(qs_env, &
                      para_env=para_env, &
                      matrix_s=matrix_s, &
                      matrix_ks=matrix_ks, &
                      dft_control=dft_control)

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_p_env_methods.F",658)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_methods.F",659)
      IF(.NOT.(0 < spin))CALL cp__a("qs_p_env_methods.F",660)
      IF(.NOT.(spin <= dft_control%nspins))CALL cp__a("qs_p_env_methods.F",661)

      CALL mpools_get(qs_env%mpools, maxao_maxmo_fm_pool=maxao_maxmo_fm_pool)
      CALL fm_pool_create_fm(maxao_maxmo_fm_pool, tmp)

      k_p => matrix_ks(spin)%matrix
      CALL cp_fm_get_info(v, ncol_global=ncol)

      IF (p_env%orthogonal_orbitals) THEN
         CALL cp_dbcsr_sm_fm_multiply(k_p, v, res, ncol)
      ELSE
         CALL cp_dbcsr_sm_fm_multiply(k_p, v, tmp, ncol)
         CALL cp_fm_symm('R', 'U', p_env%n_ao(spin), p_env%n_mo(spin), 1.0_dp, &
                         p_env%Smo_inv(spin)%matrix, tmp, 0.0_dp, res)
      END IF

      CALL cp_fm_symm('R', 'U', p_env%n_ao(spin), p_env%n_mo(spin), 1.0_dp, &
                      p_env%m_epsilon(spin)%matrix, v, 0.0_dp, tmp)
      CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, tmp, &
                                   res, p_env%n_mo(spin), alpha=1.0_dp, beta=1.0_dp)
      CALL fm_pool_give_back_fm(maxao_maxmo_fm_pool, tmp); NULLIFY (tmp)

      CALL timestop(handle)

   END SUBROUTINE p_op_l1_spin

! **************************************************************************************************
!> \brief evaluates res = alpha kpp1(v)*psi0 + beta res
!>      with kpp1 evaluated with p=qs_env%rho%rho_ao, p1=p1
!> \param p_env the perturbation environment
!> \param qs_env the qs_env that is perturbed by this p_env
!> \param p1 direction in which evaluate the second derivative
!> \param res place where to store the result
!> \param alpha scale factor of the result (defaults to 1.0)
!> \param beta scale factor of the old values (defaults to 0.0)
!> \par History
!>      02.09.2002 adapted for new qs_p_env_type (TC)
!>      03.2003 extended for p1 not taken from v (TC)
!> \author fawzi
!> \note
!>      qs_env%rho must be up to date
!>      it would be better to pass rho1, not p1
! **************************************************************************************************
   SUBROUTINE p_op_l2(p_env, qs_env, p1, res, alpha, beta)

      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p1
      TYPE(cp_fm_p_type), DIMENSION(:), INTENT(INOUT)    :: res
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: alpha, beta

      CHARACTER(len=*), PARAMETER :: routineN = 'p_op_l2', routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: fdiff = .FALSE.

      INTEGER                                            :: handle, ispin, n_spins
      INTEGER, SAVE                                      :: iter = 0
      LOGICAL                                            :: gapw, gapw_xc
      REAL(KIND=dp)                                      :: my_alpha, my_beta
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho1_ao
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, rho, rho1_ao)

      CALL get_qs_env(qs_env, dft_control=dft_control, rho=rho)

      gapw = dft_control%qs_control%gapw
      gapw_xc = dft_control%qs_control%gapw_xc
      my_alpha = 1.0_dp
      IF (PRESENT(alpha)) my_alpha = alpha
      my_beta = 0.0_dp
      IF (PRESENT(beta)) my_beta = beta

      iter = iter+1

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_p_env_methods.F",737)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_methods.F",738)

      CALL p_env_check_i_alloc(p_env, qs_env=qs_env)
      n_spins = dft_control%nspins

      CALL qs_rho_get(p_env%rho1, rho_ao=rho1_ao)
      DO ispin = 1, SIZE(p1)
         ! hack to avoid crashes in ep regs
         IF (.NOT. ASSOCIATED(rho1_ao(ispin)%matrix, p1(ispin)%matrix)) THEN
            CALL cp_dbcsr_copy(rho1_ao(ispin)%matrix, p1(ispin)%matrix)
         ENDIF
      ENDDO
      CALL qs_rho_update_rho(rho_struct=p_env%rho1, qs_env=qs_env)

      IF (fdiff) THEN
         CALL kpp1_calc_k_p_p1_fdiff(qs_env=qs_env, &
                                     k_p_p1=p_env%kpp1, rho=rho, rho1=p_env%rho1)
      ELSE
         CALL kpp1_calc_k_p_p1(kpp1_env=p_env%kpp1_env, p_env=p_env, qs_env=qs_env, &
                               k_p_p1=p_env%kpp1, rho=rho, rho1=p_env%rho1, rho1_xc=p_env%rho1)
      END IF

      DO ispin = 1, n_spins
         CALL cp_dbcsr_sm_fm_multiply(p_env%kpp1(ispin)%matrix, &
                                      p_env%psi0d(ispin)%matrix, res(ispin)%matrix, &
                                      ncol=p_env%n_mo(ispin), &
                                      alpha=my_alpha, beta=my_beta)
      END DO

      CALL timestop(handle)

   END SUBROUTINE p_op_l2

! **************************************************************************************************
!> \brief does a preorthogonalization of the given matrix:
!>      v = (I-PS)v
!> \param p_env the perturbation environment
!> \param qs_env the qs_env that is perturbed by this p_env
!> \param v matrix to orthogonalize
!> \param n_cols the number of columns of C to multiply (defaults to size(v,2))
!> \par History
!>      02.09.2002 adapted for new qs_p_env_type (TC)
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE p_preortho(p_env, qs_env, v, n_cols)

      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_p_type), DIMENSION(:), INTENT(inout)    :: v
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: n_cols

      CHARACTER(len=*), PARAMETER :: routineN = 'p_preortho', routineP = moduleN//':'//routineN

      INTEGER                                            :: cols, handle, max_cols, maxnmo, n_spins, &
                                                            nmo2, spin, v_cols, v_rows
      TYPE(cp_fm_pool_type), POINTER                     :: maxmo_maxmo_fm_pool
      TYPE(cp_fm_struct_type), POINTER                   :: maxmo_maxmo_fmstruct, tmp_fmstruct
      TYPE(cp_fm_type), POINTER                          :: tmp_matrix
      TYPE(dft_control_type), POINTER                    :: dft_control

! code

      CALL timeset(routineN, handle)

      NULLIFY (tmp_matrix, maxmo_maxmo_fm_pool, maxmo_maxmo_fmstruct, tmp_fmstruct, &
               dft_control)

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_p_env_methods.F",805)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_methods.F",806)

      CALL get_qs_env(qs_env, dft_control=dft_control)
      CALL mpools_get(qs_env%mpools, maxmo_maxmo_fm_pool=maxmo_maxmo_fm_pool)
      n_spins = dft_control%nspins
      maxmo_maxmo_fmstruct => fm_pool_get_el_struct(maxmo_maxmo_fm_pool)
      CALL cp_fm_struct_get(maxmo_maxmo_fmstruct, nrow_global=nmo2, ncol_global=maxnmo)
      IF(.NOT.(SIZE(v) >= n_spins))CALL cp__a("qs_p_env_methods.F",813)
      ! alloc tmp storage
      IF (PRESENT(n_cols)) THEN
         max_cols = MAXVAL(n_cols(1:n_spins))
      ELSE
         max_cols = 0
         DO spin = 1, n_spins
            CALL cp_fm_get_info(v(spin)%matrix, ncol_global=v_cols)
            max_cols = MAX(max_cols, v_cols)
         END DO
      END IF
      IF (max_cols <= nmo2) THEN
         CALL fm_pool_create_fm(maxmo_maxmo_fm_pool, tmp_matrix)
      ELSE
         CALL cp_fm_struct_create(tmp_fmstruct, nrow_global=max_cols, &
                                  ncol_global=maxnmo, template_fmstruct=maxmo_maxmo_fmstruct)
         CALL cp_fm_create(tmp_matrix, matrix_struct=tmp_fmstruct)
         CALL cp_fm_struct_release(tmp_fmstruct)
      END IF

      DO spin = 1, n_spins

         CALL cp_fm_get_info(v(spin)%matrix, &
                             nrow_global=v_rows, ncol_global=v_cols)
         IF(.NOT.(v_rows >= p_env%n_ao(spin)))CALL cp__a("qs_p_env_methods.F",837)
         cols = v_cols
         IF (PRESENT(n_cols)) THEN
            IF(.NOT.(n_cols(spin) <= cols))CALL cp__a("qs_p_env_methods.F",840)
            cols = n_cols(spin)
         END IF
         IF(.NOT.(cols <= max_cols))CALL cp__a("qs_p_env_methods.F",843)

         ! tmp_matrix = v^T (S psi0)
         CALL cp_gemm(transa='T', transb='N', m=cols, n=p_env%n_mo(spin), &
                      k=p_env%n_ao(spin), alpha=1.0_dp, matrix_a=v(spin)%matrix, &
                      matrix_b=p_env%S_psi0(spin)%matrix, beta=0.0_dp, &
                      matrix_c=tmp_matrix)
         ! v = v- psi0d tmp_matrix^T = v - psi0d psi0^T S v
         CALL cp_gemm(transa='N', transb='T', m=p_env%n_ao(spin), n=cols, &
                      k=p_env%n_mo(spin), alpha=-1.0_dp, &
                      matrix_a=p_env%psi0d(spin)%matrix, matrix_b=tmp_matrix, &
                      beta=1.0_dp, matrix_c=v(spin)%matrix)

      END DO

      IF (max_cols <= nmo2) THEN
         CALL fm_pool_give_back_fm(maxmo_maxmo_fm_pool, tmp_matrix)
      ELSE
         CALL cp_fm_release(tmp_matrix)
      END IF

      CALL timestop(handle)

   END SUBROUTINE p_preortho

! **************************************************************************************************
!> \brief does a postorthogonalization on the given matrix vector:
!>      v = (I-SP) v
!> \param p_env the perturbation environment
!> \param qs_env the qs_env that is perturbed by this p_env
!> \param v matrix to orthogonalize
!> \param n_cols the number of columns of C to multiply (defaults to size(v,2))
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE p_postortho(p_env, qs_env, v, n_cols)

      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_p_type), DIMENSION(:), INTENT(inout)    :: v
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: n_cols

      CHARACTER(len=*), PARAMETER :: routineN = 'p_postortho', routineP = moduleN//':'//routineN

      INTEGER                                            :: cols, handle, max_cols, maxnmo, n_spins, &
                                                            nmo2, spin, v_cols, v_rows
      TYPE(cp_fm_pool_type), POINTER                     :: maxmo_maxmo_fm_pool
      TYPE(cp_fm_struct_type), POINTER                   :: maxmo_maxmo_fmstruct, tmp_fmstruct
      TYPE(cp_fm_type), POINTER                          :: tmp_matrix
      TYPE(dft_control_type), POINTER                    :: dft_control

! code

      CALL timeset(routineN, handle)

      NULLIFY (tmp_matrix, maxmo_maxmo_fm_pool, maxmo_maxmo_fmstruct, tmp_fmstruct, &
               dft_control)

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_p_env_methods.F",902)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_methods.F",903)

      CALL get_qs_env(qs_env, dft_control=dft_control)
      CALL mpools_get(qs_env%mpools, maxmo_maxmo_fm_pool=maxmo_maxmo_fm_pool)
      n_spins = dft_control%nspins
      maxmo_maxmo_fmstruct => fm_pool_get_el_struct(maxmo_maxmo_fm_pool)
      CALL cp_fm_struct_get(maxmo_maxmo_fmstruct, nrow_global=nmo2, ncol_global=maxnmo)
      IF(.NOT.(SIZE(v) >= n_spins))CALL cp__a("qs_p_env_methods.F",910)
      ! alloc tmp storage
      IF (PRESENT(n_cols)) THEN
         max_cols = MAXVAL(n_cols(1:n_spins))
      ELSE
         max_cols = 0
         DO spin = 1, n_spins
            CALL cp_fm_get_info(v(spin)%matrix, ncol_global=v_cols)
            max_cols = MAX(max_cols, v_cols)
         END DO
      END IF
      IF (max_cols <= nmo2) THEN
         CALL fm_pool_create_fm(maxmo_maxmo_fm_pool, tmp_matrix)
      ELSE
         CALL cp_fm_struct_create(tmp_fmstruct, nrow_global=max_cols, &
                                  ncol_global=maxnmo, template_fmstruct=maxmo_maxmo_fmstruct)
         CALL cp_fm_create(tmp_matrix, matrix_struct=tmp_fmstruct)
         CALL cp_fm_struct_release(tmp_fmstruct)
      END IF

      DO spin = 1, n_spins

         CALL cp_fm_get_info(v(spin)%matrix, &
                             nrow_global=v_rows, ncol_global=v_cols)
         IF(.NOT.(v_rows >= p_env%n_ao(spin)))CALL cp__a("qs_p_env_methods.F",934)
         cols = v_cols
         IF (PRESENT(n_cols)) THEN
            IF(.NOT.(n_cols(spin) <= cols))CALL cp__a("qs_p_env_methods.F",937)
            cols = n_cols(spin)
         END IF
         IF(.NOT.(cols <= max_cols))CALL cp__a("qs_p_env_methods.F",940)

         ! tmp_matrix = v^T psi0d
         CALL cp_gemm(transa='T', transb='N', m=cols, n=p_env%n_mo(spin), &
                      k=p_env%n_ao(spin), alpha=1.0_dp, matrix_a=v(spin)%matrix, &
                      matrix_b=p_env%psi0d(spin)%matrix, beta=0.0_dp, &
                      matrix_c=tmp_matrix)
         ! v = v- (S psi0) tmp_matrix^T = v - S psi0 psi0d^T v
         CALL cp_gemm(transa='N', transb='T', m=p_env%n_ao(spin), n=cols, &
                      k=p_env%n_mo(spin), alpha=-1.0_dp, &
                      matrix_a=p_env%S_psi0(spin)%matrix, matrix_b=tmp_matrix, &
                      beta=1.0_dp, matrix_c=v(spin)%matrix)

      END DO

      IF (max_cols <= nmo2) THEN
         CALL fm_pool_give_back_fm(maxmo_maxmo_fm_pool, tmp_matrix)
      ELSE
         CALL cp_fm_release(tmp_matrix)
      END IF

      CALL timestop(handle)

   END SUBROUTINE p_postortho

END MODULE qs_p_env_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Chemical shift calculation by dfpt
!>      Initialization of the nmr_env, creation of the special neighbor lists
!>      Perturbation Hamiltonians by application of the p and rxp oprtators to  psi0
!>      Write output
!>      Deallocate everything
!> \note
!>      The psi0 should be localized
!>      the Sebastiani method works within the assumption that the orbitals are
!>      completely contained in the simulation box
!> \par History
!>       created 07-2005 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_current_utils
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_array_utils_i,                ONLY: cp_2d_i_p_type
   USE cp_array_utils_r,                ONLY: cp_2d_r_p_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        convert_offsets_to_sizes, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_init, cp_dbcsr_p_type, cp_dbcsr_set, &
        dbcsr_distribution_obj, dbcsr_type_antisymmetric
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_scale_and_add
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: current_gauge_atom,&
                                              current_gauge_r,&
                                              current_gauge_r_and_step_func,&
                                              current_orb_center_atom,&
                                              current_orb_center_box,&
                                              current_orb_center_common,&
                                              current_orb_center_wannier,&
                                              ot_precond_full_all
   USE input_section_types,             ONLY: section_get_lval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_linres_methods,               ONLY: linres_read_restart,&
                                              linres_solver,&
                                              linres_write_restart
   USE qs_linres_op,                    ONLY: set_vecp
   USE qs_linres_types,                 ONLY: current_env_create,&
                                              current_env_type,&
                                              deallocate_jrho_atom_set,&
                                              get_current_env,&
                                              init_jrho_atom_set,&
                                              jrho_atom_type,&
                                              linres_control_type,&
                                              set_current_env
   USE qs_loc_methods,                  ONLY: qs_print_cubes
   USE qs_loc_types,                    ONLY: get_qs_loc_env,&
                                              localized_wfn_control_type,&
                                              qs_loc_env_new_type
   USE qs_matrix_pools,                 ONLY: qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_operators_ao,                 ONLY: build_lin_mom_matrix
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_rho_types,                    ONLY: qs_rho_clear,&
                                              qs_rho_create,&
                                              qs_rho_retain,&
                                              qs_rho_set
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              rs_grid_release
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: current_response, current_env_cleanup, current_env_init

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_current_utils'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param p_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE current_response(current_env, p_env, qs_env)
      !
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'current_response', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: my_pos
      INTEGER :: first_center, handle, i, icenter, idir, ii, iii, ispin, ist_true, istate, j, &
         jcenter, jstate, max_nbr_center, max_states, nao, natom, nbr_center(2), ncubes, nmo, &
         nspins, nstates(2), output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, last_sgf
      INTEGER, DIMENSION(:), POINTER                     :: list_cubes, row_blk_sizes
      INTEGER, DIMENSION(:, :, :), POINTER               :: statetrueindex
      LOGICAL                                            :: append_cube, ionode, should_stop
      REAL(dp)                                           :: dk(3), dkl(3), dl(3)
      REAL(dp), DIMENSION(:), POINTER                    :: dkl_vec_ii, dkl_vec_iii
      REAL(dp), DIMENSION(:, :), POINTER                 :: vecbuf_dklxp0
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_2d_i_p_type), DIMENSION(:), POINTER        :: center_list
      TYPE(cp_2d_r_p_type), DIMENSION(:), POINTER        :: centers_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op_p_ao
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fm_work_ii, fm_work_iii, h1_psi0, &
                                                            hpsi0_ptr, psi0_order, psi1, psi1_ptr
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: p_psi0, psi1_D, psi1_p, psi1_rxp, &
                                                            rxp_psi0
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(section_vals_type), POINTER                   :: current_section, lr_section, print_key

      CALL timeset(routineN, handle)
      !
      NULLIFY (cell, dft_control, linres_control, lr_section, current_section, &
               logger, mpools, psi1, h1_psi0, vecbuf_dklxp0, mo_coeff, para_env, &
               fm_work_ii, fm_work_iii, tmp_fm_struct, dkl_vec_ii, dkl_vec_iii, &
               list_cubes, statetrueindex, centers_set, center_list, psi1_p, psi1_rxp, psi1_D, &
               p_psi0, rxp_psi0, psi0_order, op_p_ao, sab_orb, particle_set)

      logger => cp_get_default_logger()
      ionode = logger%para_env%mepos == logger%para_env%source
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")
      current_section => section_vals_get_subs_vals(qs_env%input, &
                                                    "PROPERTIES%LINRES%CURRENT")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(T10,A,/)") &
            "*** Self consistent optimization of the response wavefunctions ***"
      ENDIF

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      mpools=mpools, cell=cell, &
                      linres_control=linres_control, &
                      sab_orb=sab_orb, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      dbcsr_dist=dbcsr_dist, &
                      para_env=para_env)

      nspins = dft_control%nspins

      CALL get_current_env(current_env=current_env, &
                           nao=nao, &
                           nstates=nstates, &
                           centers_set=centers_set, &
                           nbr_center=nbr_center, &
                           center_list=center_list, &
                           list_cubes=list_cubes, &
                           statetrueindex=statetrueindex, &
                           psi1_p=psi1_p, &
                           psi1_rxp=psi1_rxp, &
                           psi1_D=psi1_D, &
                           p_psi0=p_psi0, &
                           rxp_psi0=rxp_psi0, &
                           psi0_order=psi0_order)
      !
      ! allocate the vectors
      IF (current_env%full) THEN
         ALLOCATE (psi1(nspins), h1_psi0(nspins))
         DO ispin = 1, nspins
            mo_coeff => psi0_order(ispin)%matrix
            NULLIFY (tmp_fm_struct, psi1(ispin)%matrix, h1_psi0(ispin)%matrix)
            CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                     ncol_global=nstates(ispin), &
                                     context=mo_coeff%matrix_struct%context)
            CALL cp_fm_create(psi1(ispin)%matrix, tmp_fm_struct)
            CALL cp_fm_create(h1_psi0(ispin)%matrix, tmp_fm_struct)
            CALL cp_fm_struct_release(tmp_fm_struct)
         ENDDO
         !
         ! prepare for allocation
         natom = SIZE(particle_set, 1)
         ALLOCATE (first_sgf(natom))
         ALLOCATE (last_sgf(natom))
         CALL get_particle_set(particle_set, qs_kind_set, &
                               first_sgf=first_sgf, &
                               last_sgf=last_sgf)
         ALLOCATE (row_blk_sizes(natom))
         CALL convert_offsets_to_sizes(first_sgf, row_blk_sizes, last_sgf)
         DEALLOCATE (first_sgf, last_sgf)
         !
         ! rebuild the linear momentum matrices
         CALL cp_dbcsr_allocate_matrix_set(op_p_ao, 3)
         ALLOCATE (op_p_ao(1)%matrix)
         CALL cp_dbcsr_init(op_p_ao(1)%matrix)
         CALL cp_dbcsr_create(matrix=op_p_ao(1)%matrix, &
                              name="OP_P", &
                              dist=dbcsr_dist, matrix_type=dbcsr_type_antisymmetric, &
                              row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                              nze=0, mutable_work=.TRUE.)
         CALL cp_dbcsr_alloc_block_from_nbl(op_p_ao(1)%matrix, sab_orb)
         !
         !
         DEALLOCATE (row_blk_sizes)
         !
         !
         CALL cp_dbcsr_set(op_p_ao(1)%matrix, 0.0_dp)
         DO idir = 2, 3
            ALLOCATE (op_p_ao(idir)%matrix)
            CALL cp_dbcsr_init(op_p_ao(idir)%matrix)
            CALL cp_dbcsr_copy(op_p_ao(idir)%matrix, op_p_ao(1)%matrix, &
                               "current_env%op_p_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir))))
            CALL cp_dbcsr_set(op_p_ao(idir)%matrix, 0.0_dp)
         ENDDO
         !
         !CALL p_xyz_ao(op_p_ao,qs_env,minimum_image=.FALSE.)
         CALL build_lin_mom_matrix(qs_env, op_p_ao)
         !
      ENDIF
      !
      ! set response to zero
      !
      DO idir = 1, 3
         DO ispin = 1, nspins
            CALL cp_fm_set_all(psi1_p(ispin, idir)%matrix, 0.0_dp)
            CALL cp_fm_set_all(psi1_rxp(ispin, idir)%matrix, 0.0_dp)
            IF (current_env%full) CALL cp_fm_set_all(psi1_D(ispin, idir)%matrix, 0.0_dp)
         ENDDO
      ENDDO
      !
      !
      !
      ! load restart file
      !
      first_center = 0
      IF (linres_control%linres_restart) THEN
         DO idir = 1, 3
            ! operator p
            psi1_ptr => psi1_p(:, idir)
            CALL linres_read_restart(qs_env, lr_section, psi1_ptr, idir, "nmr_p")
            ! operator rxp
            psi1_ptr => psi1_rxp(:, idir)
            CALL linres_read_restart(qs_env, lr_section, psi1_ptr, idir, "nmr_rxp")
         ENDDO
      ENDIF
      !
      !
      !
      !
      should_stop = .FALSE.
      current_env%all_pert_op_done = .FALSE.
      ! operator p
      DO idir = 1, 3
         IF (should_stop) EXIT
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(T10,A)") "Response to the perturbation operator P_"//ACHAR(idir+119)
         ENDIF
         !
         ! Initial guess for psi1
         psi1_ptr => psi1_p(:, idir)
         hpsi0_ptr => p_psi0(:, idir)
         !
         !
         linres_control%converged = .FALSE.
         CALL linres_solver(p_env, qs_env, psi1_ptr, hpsi0_ptr, psi0_order, should_stop)
         !
         !
         ! print response functions
         IF (BTEST(cp_print_key_should_output(logger%iter_info, current_section,&
              &   "PRINT%RESPONSE_FUNCTION_CUBES"), cp_p_file)) THEN
            ncubes = SIZE(list_cubes, 1)
            print_key => section_vals_get_subs_vals(current_section, "PRINT%RESPONSE_FUNCTION_CUBES")
            append_cube = section_get_lval(current_section, "PRINT%RESPONSE_FUNCTION_CUBES%APPEND")
            my_pos = "REWIND"
            IF (append_cube) THEN
               my_pos = "APPEND"
            END IF
            !

            DO ispin = 1, nspins
               CALL qs_print_cubes(qs_env, psi1_ptr(ispin)%matrix, ncubes, list_cubes, &
                                   centers_set(ispin)%array, print_key, 'psi1_p', &
                                   idir=idir, ispin=ispin, file_position=my_pos)
            ENDDO ! ispin
         ENDIF ! print response functions
         !
         ! write restart file
         CALL linres_write_restart(qs_env, lr_section, psi1_ptr, idir, "nmr_p")
      ENDDO ! idir
      !
      ! operator rxp
      DO idir = 1, 3
         IF (should_stop) EXIT
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(T10,A)") "Response to the perturbation operator L_"//ACHAR(idir+119)
         ENDIF
         !
         ! Initial guess for psi1
         psi1_ptr => psi1_rxp(:, idir)
         hpsi0_ptr => rxp_psi0(:, idir)
         !
         !
         linres_control%converged = .FALSE.
         CALL linres_solver(p_env, qs_env, psi1_ptr, hpsi0_ptr, psi0_order, should_stop)
         !
         ! print response functions
         IF (BTEST(cp_print_key_should_output(logger%iter_info, current_section,&
              &   "PRINT%RESPONSE_FUNCTION_CUBES"), cp_p_file)) THEN
            ncubes = SIZE(list_cubes, 1)
            print_key => section_vals_get_subs_vals(current_section, "PRINT%RESPONSE_FUNCTION_CUBES")

            DO ispin = 1, nspins
               CALL qs_print_cubes(qs_env, psi1_ptr(ispin)%matrix, ncubes, list_cubes, &
                                   centers_set(ispin)%array, print_key, 'psi1_rxp', &
                                   idir=idir, ispin=ispin, file_position=my_pos)
            ENDDO ! ispin
         ENDIF ! print response functions
         !
         ! write restart file
         CALL linres_write_restart(qs_env, lr_section, psi1_ptr, idir, "nmr_rxp")
      ENDDO ! idir
      IF (.NOT. should_stop) current_env%all_pert_op_done = .TRUE.
      !
      ! operator D
      IF (current_env%full) THEN
         current_env%all_pert_op_done = .FALSE.
         !
         !
         DO ispin = 1, nspins
            CALL cp_fm_set_all(psi1(ispin)%matrix, 0.0_dp)
         ENDDO
         !
         ! The correction is state depedent a loop over the states is necessary
         max_nbr_center = MAXVAL(nbr_center(1:nspins))
         max_states = MAXVAL(nstates(1:nspins))
         !
         ALLOCATE (vecbuf_dklxp0(1, nao), fm_work_ii(nspins), fm_work_iii(nspins), &
                   dkl_vec_ii(max_states), dkl_vec_iii(max_states))
         vecbuf_dklxp0(1, nao) = 0.0_dp
         !
         DO ispin = 1, nspins
            nmo = nstates(ispin)
            mo_coeff => psi0_order(ispin)%matrix
            NULLIFY (tmp_fm_struct, fm_work_ii(ispin)%matrix, fm_work_iii(ispin)%matrix)
            CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                     ncol_global=nmo, para_env=para_env, &
                                     context=mo_coeff%matrix_struct%context)

            CALL cp_fm_create(fm_work_ii(ispin)%matrix, tmp_fm_struct)
            CALL cp_fm_set_all(fm_work_ii(ispin)%matrix, 0.0_dp)
            CALL cp_fm_create(fm_work_iii(ispin)%matrix, tmp_fm_struct)
            CALL cp_fm_set_all(fm_work_iii(ispin)%matrix, 0.0_dp)
            CALL cp_fm_struct_release(tmp_fm_struct)
         ENDDO ! ispin
         !
         DO idir = 1, 3
            IF (should_stop) EXIT
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1_D(ispin, idir)%matrix, 0.0_dp)
            END DO
            first_center = 0
            IF (linres_control%linres_restart) THEN
               psi1_ptr => psi1_D(:, idir)
               CALL linres_read_restart(qs_env, lr_section, psi1_ptr, idir, "nmr_dxp-", ind=first_center)
            END IF
            IF (first_center > 0) THEN
               IF (output_unit > 0) THEN
                  WRITE (output_unit, "(T10,A,I4,A)")&
                     & "Response to the perturbation operators (dk-dl)xp up to state ", &
                       first_center, " have been read from restart"
               ENDIF
            END IF
            !
            ! here we run over the max number of states, then we need
            ! to be careful with overflow while doing uks calculations.
            DO icenter = 1, max_nbr_center
               !
               IF (should_stop) EXIT
               IF (icenter > first_center) THEN
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, "(T10,A,I4,A)")&
                          & "Response to the perturbation operator (dk-dl)xp for -state- ", &
                            icenter, " in dir. "//ACHAR(idir+119)
                  ENDIF
                  !
                  DO ispin = 1, nspins
                     nmo = nstates(ispin)
                     mo_coeff => psi0_order(ispin)%matrix
                     !
                     ! take care that no overflow can occur for uks
                     IF (icenter .GT. nbr_center(ispin)) THEN
                        !
                        ! set h1_psi0 and psi1 to zero to avoid problems in linres_scf
                        CALL cp_fm_set_all(h1_psi0(ispin)%matrix, 0.0_dp)
                        CALL cp_fm_set_all(psi1(ispin)%matrix, 0.0_dp)
                        CYCLE
                     ENDIF
                     !
                     dkl_vec_ii(:) = 0.0_dp
                     dkl_vec_iii(:) = 0.0_dp
                     !
                     ist_true = statetrueindex(idir, icenter, ispin)
                     !
                     ! the initial guess is the previous set of psi1, just optimized
                     CALL set_vecp(idir, ii, iii)
                     dk(1:3) = centers_set(ispin)%array(1:3, ist_true)
                     !
                     DO jcenter = 1, nbr_center(ispin)
                        dl(1:3) = centers_set(ispin)%array(1:3, jcenter)
                        dkl = pbc(dl, dk, cell)
                        DO j = center_list(ispin)%array(1, jcenter), center_list(ispin)%array(1, jcenter+1)-1
                           jstate = center_list(ispin)%array(2, j)
                           dkl_vec_ii(jstate) = dkl(ii)
                           dkl_vec_iii(jstate) = dkl(iii)
                        ENDDO
                     ENDDO
                     !
                     ! First term
                     ! Rescale the ground state orbitals by (dk-dl)_ii
                     CALL cp_fm_to_fm(mo_coeff, fm_work_ii(ispin)%matrix)
                     CALL cp_fm_column_scale(fm_work_ii(ispin)%matrix, dkl_vec_ii(1:nmo))
                     !
                     ! Apply the p_iii operator
                     ! fm_work_iii = -p_iii * (dk-dl)_ii * C0
                     CALL cp_dbcsr_sm_fm_multiply(op_p_ao(iii)%matrix, fm_work_ii(ispin)%matrix, &
                                                  fm_work_iii(ispin)%matrix, ncol=nmo, alpha=-1.0_dp)
                     !
                     ! Copy in h1_psi0
                     ! h1_psi0_i = fm_work_iii
                     CALL cp_fm_to_fm(fm_work_iii(ispin)%matrix, h1_psi0(ispin)%matrix)
                     !
                     ! Second term
                     ! Rescale the ground state orbitals by (dk-dl)_iii
                     CALL cp_fm_to_fm(mo_coeff, fm_work_iii(ispin)%matrix)
                     CALL cp_fm_column_scale(fm_work_iii(ispin)%matrix, dkl_vec_iii(1:nmo))
                     !
                     ! Apply the p_ii operator
                     ! fm_work_ii = -p_ii * (dk-dl)_iii * C0
                     CALL cp_dbcsr_sm_fm_multiply(op_p_ao(ii)%matrix, fm_work_iii(ispin)%matrix, &
                                                  fm_work_ii(ispin)%matrix, ncol=nmo, alpha=-1.0_dp)
                     !
                     ! Copy in h1_psi0
                     ! h1_psi0_i = fm_work_iii - fm_work_ii
                     CALL cp_fm_scale_and_add(1.0_dp, h1_psi0(ispin)%matrix,&
                          &                  -1.0_dp, fm_work_ii(ispin)%matrix)
                  ENDDO

                  !
                  ! Optimize the response wavefunctions
                  CALL linres_solver(p_env, qs_env, psi1, h1_psi0, psi0_order, should_stop)
                  !
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, "(T10,A,/)")&
                          & "Store the psi1 vector for the calculation of the response current density "
                  ENDIF
                  !
                  DO ispin = 1, nspins
                     !
                     ! take care that no overflow can occur for uks
                     IF (icenter .GT. nbr_center(ispin)) CYCLE
                     !
                     ! need to reset those guys
                     ist_true = statetrueindex(idir, icenter, ispin)
                     DO i = center_list(ispin)%array(1, ist_true), center_list(ispin)%array(1, ist_true+1)-1
                        istate = center_list(ispin)%array(2, i)
                        !
                        ! the optimized wfns are copied in the fm
                        CALL cp_fm_to_fm(psi1(ispin)%matrix, psi1_D(ispin, idir)%matrix, 1, istate, istate)
                     ENDDO
                     current_env%full_done(idir*ispin, icenter) = .TRUE.
                  ENDDO ! ispin
                  !
               ELSE
                  DO ispin = 1, nspins
                     current_env%full_done(idir*ispin, icenter) = .TRUE.
                  ENDDO ! ispin
               END IF
               psi1_ptr => psi1_D(:, idir)
               CALL linres_write_restart(qs_env, lr_section, psi1_ptr, idir, "nmr_dxp-", ind=icenter)

            ENDDO ! center
            !
            ! print response functions
            IF (BTEST(cp_print_key_should_output(logger%iter_info, current_section,&
                 &   "PRINT%RESPONSE_FUNCTION_CUBES"), cp_p_file)) THEN
               ncubes = SIZE(list_cubes, 1)
               print_key => section_vals_get_subs_vals(current_section, "PRINT%RESPONSE_FUNCTION_CUBES")
               DO ispin = 1, nspins
                  CALL qs_print_cubes(qs_env, psi1_D(ispin, idir)%matrix, &
                                      ncubes, list_cubes, centers_set(ispin)%array, print_key, 'psi1_D', &
                                      idir=idir, ispin=ispin, file_position=my_pos)
               ENDDO
            ENDIF ! print response functions
            !
         ENDDO ! idir
         IF (.NOT. should_stop) current_env%all_pert_op_done = .TRUE.
         !
         ! clean up
         DO ispin = 1, nspins
            CALL cp_fm_release(fm_work_ii(ispin)%matrix)
            CALL cp_fm_release(fm_work_iii(ispin)%matrix)
         ENDDO
         DEALLOCATE (fm_work_ii, fm_work_iii, dkl_vec_ii, dkl_vec_iii, vecbuf_dklxp0)

      ENDIF
      !
      ! clean up
      IF (current_env%full) THEN
         CALL cp_dbcsr_deallocate_matrix_set(op_p_ao)
         DO ispin = 1, nspins
            CALL cp_fm_release(psi1(ispin)%matrix)
            CALL cp_fm_release(h1_psi0(ispin)%matrix)
         ENDDO
         DEALLOCATE (psi1, h1_psi0)
      ENDIF
      !
      CALL cp_print_key_finished_output(output_unit, logger, lr_section,&
           &                            "PRINT%PROGRAM_RUN_INFO")
      !
      CALL timestop(handle)
      !
   END SUBROUTINE current_response

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE current_env_init(current_env, qs_env)
      !
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'current_env_init', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, homo, i, iao, iatom, ibox, icenter, icount, idir, ii, ini, ir, is, ispin, &
         istate, istate2, istate_next, ix, iy, iz, j, jstate, k, max_nbr_center, max_states, n, &
         n_rep, nao, natom, nbr_box, ncubes, nmo, nspins, nstate, nstate_list(2), output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: buff, first_sgf, last_sgf
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: state_list
      INTEGER, DIMENSION(:), POINTER                     :: bounds, list, nbox, &
                                                            selected_states_on_atom_list
      LOGICAL                                            :: gapw, is0, uniform_occupation
      LOGICAL, ALLOCATABLE, DIMENSION(:, :)              :: state_done
      REAL(dp)                                           :: center(3), center2(3), dist, mdist, &
                                                            r(3), rab(3)
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: rbuff
      REAL(dp), DIMENSION(:), POINTER                    :: common_center
      REAL(dp), DIMENSION(:, :), POINTER                 :: center_array
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: current_section, lr_section

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, qs_kind_set, cell, dft_control, linres_control, scf_control, &
               logger, mos, mpools, current_section, particle_set, mo_coeff, &
               auxbas_pw_pool, pw_env, jrho1_atom_set, common_center, tmp_fm_struct, &
               para_env, qs_loc_env, localized_wfn_control, rho_g, rho_r)
      !
      !
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      linres_control=linres_control, &
                      mos=mos, &
                      mpools=mpools, &
                      particle_set=particle_set, &
                      pw_env=pw_env, &
                      scf_control=scf_control, &
                      para_env=para_env)
      !
      gapw = dft_control%qs_control%gapw
      nspins = dft_control%nspins
      natom = SIZE(particle_set, 1)
      CALL get_mo_set(mo_set=mos(1)%mo_set, nao=nao)
      !
      max_states = 0
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, nmo=nmo)
         max_states = MAX(max_states, nmo)
      ENDDO
      !
      !
      !
      ! some checks
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, nmo=nmo, homo=homo, &
                         uniform_occupation=uniform_occupation)
         !
         ! check that homo=nmo
         IF (nmo .NE. homo) CALL cp__b("qs_linres_current_utils.F",665,"nmo != homo")
         !
         ! check that the nbr of localized states is equal to the nbr of states
!       nmoloc = SIZE(linres_control%localized_wfn_control%centers_set(ispin)%array,2)
!       IF (nmoloc.NE.nmo) CALL cp_abort(cp__l("qs_linres_current_utils.F",669),&
!                                            "The nbr of localized functions "//&
!            &                               "is not equal to the nbr of states.")
         !
         ! check that the occupation is uniform
         IF (.NOT. uniform_occupation) CALL cp__b("qs_linres_current_utils.F",674,"nmo != homo")
      ENDDO
      !
      !
      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")
      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(/,T20,A,/)") "*** Start current Calculation ***"
         WRITE (output_unit, "(T10,A,/)") "Inizialization of the current environment"
      ENDIF

      IF (current_env%ref_count /= 0) CALL current_env_cleanup(current_env, qs_env)

      CALL current_env_create(current_env)
      !
      !
      current_env%chi_tensor(:, :, :) = 0.0_dp
      current_env%chi_tensor_loc(:, :, :) = 0.0_dp
      current_env%nao = nao
      current_env%full = .TRUE.
      current_env%do_selected_states = .FALSE.
      !
      ! If current_density or full_nmr different allocations are required
      current_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES%CURRENT")
      !
      ! Select the gauge
      CALL section_vals_val_get(current_section, "GAUGE", i_val=current_env%gauge)
      SELECT CASE (current_env%gauge)
      CASE (current_gauge_r)
         current_env%gauge_name = "R"
      CASE (current_gauge_r_and_step_func)
         current_env%gauge_name = "R_AND_STEP_FUNCTION"
      CASE (current_gauge_atom)
         current_env%gauge_name = "ATOM"
      CASE DEFAULT
         CALL cp__b("qs_linres_current_utils.F",712,"Unknown gauge, try again...")
      END SELECT
      !
      ! maximal radius to build the atom gauge
      CALL section_vals_val_get(current_section, "GAUGE_ATOM_RADIUS", &
                                r_val=current_env%gauge_atom_radius)
      ! use old gauge atom
      CALL section_vals_val_get(current_section, "USE_OLD_GAUGE_ATOM", &
                                l_val=current_env%use_old_gauge_atom)
      ! chi for pbc
      CALL section_vals_val_get(current_section, "CHI_PBC", l_val=current_env%chi_pbc)
      !
      ! use old gauge atom
      CALL section_vals_val_get(current_section, "USE_OLD_GAUGE_ATOM", &
                                l_val=current_env%use_old_gauge_atom)
      !
      ! chi for pbc
      CALL section_vals_val_get(current_section, "CHI_PBC", l_val=current_env%chi_pbc)
      !
      ! which center for the orbitals shall we use
      CALL section_vals_val_get(current_section, "ORBITAL_CENTER", i_val=current_env%orb_center)
      SELECT CASE (current_env%orb_center)
      CASE (current_orb_center_wannier)
         !
         current_env%orb_center_name = "WANNIER"
      CASE (current_orb_center_common)
         !
         current_env%orb_center_name = "COMMON"
         current_env%full = .FALSE.
         !
         ! Is there a user specified common_center?
         CALL section_vals_val_get(current_section, "COMMON_CENTER", r_vals=common_center)
      CASE (current_orb_center_atom)
         !
         current_env%orb_center_name = "ATOM"
      CASE (current_orb_center_box)
         !
         current_env%orb_center_name = "BOX"
         !
         ! Is there a user specified nbox?
         CALL section_vals_val_get(current_section, "NBOX", i_vals=nbox)
      CASE DEFAULT
         CALL cp__b("qs_linres_current_utils.F",754,"Unknown orbital center, try again...")
      END SELECT
      !
      ! Check if restat also psi0 should be restarted
      !IF(current_env%restart_current .AND. scf_control%density_guess/=restart_guess)THEN
      !   CALL cp__b("qs_linres_current_utils.F",759,"restart_nmr requires density_guess=restart")
      !ENDIF
      !
      ! check that the psi0 are localized and you have all the centers
      IF(.NOT.(linres_control%localized_psi0))CALL cp__a("qs_linres_current_utils.F",763)
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(A)') &
            ' To get CURRENT parameters within PBC you need localized zero order orbitals '
      ENDIF
      qs_loc_env => linres_control%qs_loc_env
      CALL get_qs_loc_env(qs_loc_env, localized_wfn_control=localized_wfn_control)
      !
      !
      ALLOCATE (current_env%centers_set(nspins), current_env%center_list(nspins), &
                state_list(max_states, nspins))
      state_list(:, :) = HUGE(0)
      nstate_list(:) = HUGE(0)
      !
      !
      !
      ! if qmmm is selected change the definition of the center of the box 0 -> L/2
      IF (current_env%do_qmmm) THEN
         DO ispin = 1, nspins
            DO istate = 1, SIZE(localized_wfn_control%centers_set(ispin)%array, 2)
               ! just to be sure...
               r(:) = pbc(localized_wfn_control%centers_set(ispin)%array(:, istate), cell)
               IF (r(1) .LT. 0.0_dp) THEN
                  localized_wfn_control%centers_set(ispin)%array(1, istate) = &
                     r(1)+cell%hmat(1, 1)
               ENDIF
               IF (r(2) .LT. 0.0_dp) THEN
                  localized_wfn_control%centers_set(ispin)%array(2, istate) = &
                     r(2)+cell%hmat(2, 2)
               ENDIF
               IF (r(3) .LT. 0.0_dp) THEN
                  localized_wfn_control%centers_set(ispin)%array(3, istate) = &
                     r(3)+cell%hmat(3, 3)
               ENDIF
            ENDDO
         ENDDO
      ENDIF
      !
      !
      !
      ! if the user has requested to compute the response for a subset of the states
      ! we collect them here. it requies the states to be localized!
      CALL section_vals_val_get(current_section, "SELECTED_STATES_ATOM_RADIUS", &
                                r_val=current_env%selected_states_atom_radius)
      CALL section_vals_val_get(current_section, "SELECTED_STATES_ON_ATOM_LIST", n_rep_val=n_rep)
      !
      current_env%do_selected_states = n_rep .GT. 0
      !
      ! for the moment selected states doesnt work with the preconditioner FULL_ALL
      IF (linres_control%preconditioner_type .EQ. ot_precond_full_all .AND. current_env%do_selected_states) &
         CALL cp__b("qs_linres_current_utils.F",813,"Selected states doesnt work with the preconditioner FULL_ALL")
      !
      !
      NULLIFY (current_env%selected_states_on_atom_list)
      n = 0
      DO ir = 1, n_rep
         NULLIFY (list)
         CALL section_vals_val_get(current_section, "SELECTED_STATES_ON_ATOM_LIST", &
                                   i_rep_val=ir, i_vals=list)
         IF (ASSOCIATED(list)) THEN
            CALL reallocate(current_env%selected_states_on_atom_list, 1, n+SIZE(list))
            DO ini = 1, SIZE(list)
               current_env%selected_states_on_atom_list(ini+n) = list(ini)
            ENDDO
            n = n+SIZE(list)
         ENDIF
      ENDDO
      !
      ! build the subset
      IF (current_env%do_selected_states) THEN
         selected_states_on_atom_list => current_env%selected_states_on_atom_list
         DO ispin = 1, nspins
            center_array => localized_wfn_control%centers_set(ispin)%array
            nstate = 0
            DO istate = 1, SIZE(center_array, 2)
               DO i = 1, SIZE(selected_states_on_atom_list, 1)
                  iatom = selected_states_on_atom_list(i)
                  r(:) = pbc(center_array(1:3, istate)-particle_set(iatom)%r(:), cell)
                  IF (SQRT(DOT_PRODUCT(r, r)) .LE. current_env%selected_states_atom_radius) THEN
                     !
                     ! add the state to the list
                     nstate = nstate+1
                     state_list(nstate, ispin) = istate
                     EXIT
                  ENDIF
               ENDDO
            ENDDO
            nstate_list(ispin) = nstate
         ENDDO
      ELSE
         DO ispin = 1, nspins
            center_array => localized_wfn_control%centers_set(ispin)%array
            nstate = 0
            DO istate = 1, SIZE(center_array, 2)
               nstate = nstate+1
               state_list(nstate, ispin) = istate
            ENDDO
            nstate_list(ispin) = nstate
         ENDDO
      ENDIF
      !
      !
      !
      ! clustering the states
      DO ispin = 1, nspins
         nstate = nstate_list(ispin)
         current_env%nstates(ispin) = nstate
         !
         ALLOCATE (current_env%center_list(ispin)%array(2, nstate+1), &
                   current_env%centers_set(ispin)%array(3, nstate))
         current_env%center_list(ispin)%array(:, :) = HUGE(0)
         current_env%centers_set(ispin)%array(:, :) = HUGE(0.0_dp)
         !
         center_array => localized_wfn_control%centers_set(ispin)%array
         !
         ! point to the psi0 centers
         SELECT CASE (current_env%orb_center)
         CASE (current_orb_center_wannier)
            !
            ! use the wannier center as -center-
            current_env%nbr_center(ispin) = nstate
            DO is = 1, nstate
               istate = state_list(is, ispin)
               current_env%centers_set(ispin)%array(1:3, is) = center_array(1:3, istate)
               current_env%center_list(ispin)%array(1, is) = is
               current_env%center_list(ispin)%array(2, is) = istate
            ENDDO
            current_env%center_list(ispin)%array(1, nstate+1) = nstate+1
            !
         CASE (current_orb_center_common)
            !
            ! use a common -center-
            current_env%centers_set(ispin)%array(:, 1) = common_center(:)
            current_env%nbr_center(ispin) = 1
            current_env%center_list(ispin)%array(1, 1) = 1
            current_env%center_list(ispin)%array(1, 2) = nstate+1
            DO is = 1, nstate
               istate = state_list(is, ispin)
               current_env%center_list(ispin)%array(2, is) = istate
            ENDDO
            !
         CASE (current_orb_center_atom)
            !
            ! use the atom as -center-
            ALLOCATE (buff(nstate_list(ispin)))
            buff(:) = 0
            !
            DO is = 1, nstate
               istate = state_list(is, ispin)
               mdist = HUGE(0.0_dp)
               DO iatom = 1, natom
                  r = pbc(particle_set(iatom)%r(:), cell)
                  rab = pbc(r, center_array(1:3, istate), cell)
                  dist = SQRT(rab(1)**2+rab(2)**2+rab(3)**2)
                  IF (dist .LT. mdist) THEN
                     buff(is) = iatom
                     mdist = dist
                  ENDIF
               ENDDO
            ENDDO
            !
            i = 0
            ii = 1
            current_env%center_list(ispin)%array(1, 1) = 1
            DO iatom = 1, natom
               j = 0
               is0 = .TRUE.
               DO is = 1, nstate
                  istate = state_list(is, ispin)
                  IF (buff(is) .EQ. iatom) THEN
                     j = j+1
                     i = i+1
                     is0 = .FALSE.
                     current_env%center_list(ispin)%array(2, i) = istate
                  ENDIF
               ENDDO
               IF (.NOT. is0) THEN
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, '(T2,A,I6,A,I6)') 'clustering ', j, ' center(s) on atom ', iatom
                  ENDIF
                  current_env%center_list(ispin)%array(1, ii+1) = &
                     current_env%center_list(ispin)%array(1, ii)+j
                  current_env%centers_set(ispin)%array(:, ii) = &
                     pbc(particle_set(iatom)%r, cell)
                  ii = ii+1
               ENDIF
            ENDDO
            current_env%nbr_center(ispin) = ii-1
            !
            DEALLOCATE (buff)
         CASE (current_orb_center_box)
            !
            ! use boxes as -center-
            nbr_box = nbox(1)*nbox(2)*nbox(3)
            ALLOCATE (rbuff(3, nbr_box), buff(nstate))
            rbuff(:, :) = HUGE(0.0_dp)
            buff(:) = 0
            !
            ibox = 1
            DO iz = 1, nbox(3)
               DO iy = 1, nbox(2)
                  DO ix = 1, nbox(1)
                     rbuff(1, ibox) = cell%hmat(1, 1)*((REAL(ix, dp)-0.5_dp)/REAL(nbox(1), dp)-0.5_dp)
                     rbuff(2, ibox) = cell%hmat(2, 2)*((REAL(iy, dp)-0.5_dp)/REAL(nbox(2), dp)-0.5_dp)
                     rbuff(3, ibox) = cell%hmat(3, 3)*((REAL(iz, dp)-0.5_dp)/REAL(nbox(3), dp)-0.5_dp)
                     ibox = ibox+1
                  ENDDO
               ENDDO
            ENDDO
            !
            DO is = 1, nstate
               istate = state_list(is, ispin)
               mdist = HUGE(0.0_dp)
               DO ibox = 1, nbr_box
                  rab(:) = pbc(rbuff(:, ibox), center_array(1:3, istate), cell)
                  dist = SQRT(rab(1)**2+rab(2)**2+rab(3)**2)
                  IF (dist .LT. mdist) THEN
                     buff(is) = ibox
                     mdist = dist
                  ENDIF
               ENDDO
            ENDDO
            !
            i = 0
            ii = 1
            current_env%center_list(ispin)%array(1, 1) = 1
            DO ibox = 1, nbr_box
               j = 0
               is0 = .TRUE.
               DO is = 1, nstate
                  istate = state_list(is, ispin)
                  IF (buff(is) .EQ. ibox) THEN
                     j = j+1
                     i = i+1
                     is0 = .FALSE.
                     current_env%center_list(ispin)%array(2, i) = istate
                  ENDIF
               ENDDO
               IF (.NOT. is0) THEN
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, '(T2,A,I6,A,I6)') 'clustering ', j, ' center(s) on box ', ibox
                  ENDIF
                  current_env%center_list(ispin)%array(1, ii+1) = &
                     current_env%center_list(ispin)%array(1, ii)+j
                  current_env%centers_set(ispin)%array(:, ii) = rbuff(:, ibox)
                  ii = ii+1
               ENDIF
            ENDDO
            current_env%nbr_center(ispin) = ii-1
            !
            DEALLOCATE (buff, rbuff)
         CASE DEFAULT
            CALL cp__b("qs_linres_current_utils.F",1015,"Unknown orbital center, try again...")
         END SELECT
      ENDDO
      !
      !
      !
      ! block the states for each center
      ALLOCATE (current_env%psi0_order(nspins))
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         NULLIFY (tmp_fm_struct)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                  ncol_global=nstate_list(ispin), para_env=para_env, &
                                  context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(current_env%psi0_order(ispin)%matrix, tmp_fm_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
         CALL cp_fm_set_all(current_env%psi0_order(ispin)%matrix, 0.0_dp)
      ENDDO
      !
      !
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         jstate = 0
         DO icenter = 1, current_env%nbr_center(ispin)
            DO i = current_env%center_list(ispin)%array(1, icenter), &
                   current_env%center_list(ispin)%array(1, icenter+1)-1
               !
               istate = current_env%center_list(ispin)%array(2, i)
               jstate = jstate+1
               !
               IF (current_env%do_selected_states) THEN ! this should be removed. always reorder the states
                  ! the blocking works (so far) with all the precond except FULL_ALL
                  ! the center_list(ispin)%array(2,i) array is obsolete then
                  current_env%center_list(ispin)%array(2, i) = jstate
                  CALL cp_fm_to_fm(mo_coeff, current_env%psi0_order(ispin)%matrix, 1, istate, jstate)
               ELSE
                  ! for the moment we just copy them
                  CALL cp_fm_to_fm(mo_coeff, current_env%psi0_order(ispin)%matrix, 1, istate, istate)
               ENDIF
            ENDDO
         ENDDO
      ENDDO
      !
      !
      !
      !
      IF (current_env%do_qmmm .AND.&
           & (current_env%orb_center .EQ. current_orb_center_wannier .OR. &
              current_env%orb_center .EQ. current_orb_center_atom .OR. &
              current_env%orb_center .EQ. current_orb_center_box)) THEN
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) 'WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING '
            WRITE (output_unit, *) 'orbital center shifted to match the '
            WRITE (output_unit, *) 'center of the box (L/2 L/2 L/2) (superdirty...)'
            WRITE (output_unit, *) 'WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING '
         ENDIF
         DO ispin = 1, nspins
            DO istate = 1, current_env%nbr_center(ispin)
               IF (current_env%centers_set(ispin)%array(1, istate) .LE. 0.0_dp) THEN
                  current_env%centers_set(ispin)%array(1, istate) = &
                     current_env%centers_set(ispin)%array(1, istate)+cell%hmat(1, 1)
               ENDIF
               IF (current_env%centers_set(ispin)%array(2, istate) .LE. 0.0_dp) THEN
                  current_env%centers_set(ispin)%array(2, istate) = &
                     current_env%centers_set(ispin)%array(2, istate)+cell%hmat(2, 2)
               ENDIF
               IF (current_env%centers_set(ispin)%array(3, istate) .LE. 0.0_dp) THEN
                  current_env%centers_set(ispin)%array(3, istate) = &
                     current_env%centers_set(ispin)%array(3, istate)+cell%hmat(3, 3)
               ENDIF
            ENDDO
         ENDDO
         ! printing
         DO ispin = 1, nspins
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(/,T2,A,I2)') "WANNIER CENTERS for spin ", ispin
               WRITE (output_unit, '(/,T18,A)') "--------------- Shifted Centers --------------- "
               DO istate = 1, current_env%nbr_center(ispin)
                  WRITE (output_unit, '(T5,A,I6,3F12.6)') &
                     'state ', istate, current_env%centers_set(ispin)%array(1:3, istate)
               ENDDO
            ENDIF
         ENDDO
      ENDIF
      !
      !
      !
      ! reorder the center dependent responses
      max_nbr_center = MAXVAL(current_env%nbr_center(1:nspins))
      current_env%nao = nao
      ALLOCATE (current_env%statetrueindex(3, max_nbr_center, nspins))
      current_env%statetrueindex(:, :, :) = 0
      IF (.TRUE.) THEN
         ALLOCATE (state_done(3, max_nbr_center))
         DO ispin = 1, nspins
            state_done(:, :) = .FALSE.
            current_env%statetrueindex(1, 1, ispin) = 1
            center(1) = current_env%centers_set(ispin)%array(1, 1)
            center(2) = current_env%centers_set(ispin)%array(2, 1)
            center(3) = current_env%centers_set(ispin)%array(3, 1)
            state_done(1, 1) = .TRUE.
            icount = 1
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, nmo=nmo)
            !
            DO idir = 1, 3
               ini = 1
               IF (idir == 1) ini = 2
               DO istate = ini, current_env%nbr_center(ispin)
                  mdist = HUGE(0.0_dp)
                  !
                  DO istate2 = 1, current_env%nbr_center(ispin)
                     IF (.NOT. state_done(idir, istate2)) THEN
                        center2(1) = current_env%centers_set(ispin)%array(1, istate2)
                        center2(2) = current_env%centers_set(ispin)%array(2, istate2)
                        center2(3) = current_env%centers_set(ispin)%array(3, istate2)
                        !
                        rab = pbc(center, center2, cell)
                        CALL set_vecp(idir, j, k)
                        dist = SQRT(rab(j)*rab(j)+rab(k)*rab(k))
                        !
                        IF (dist .LT. mdist) THEN
                           mdist = dist
                           istate_next = istate2
                        ENDIF
                     ENDIF
                  ENDDO ! istate2
                  !
                  icount = icount+1
                  state_done(idir, istate_next) = .TRUE.
                  current_env%statetrueindex(idir, icount, ispin) = istate_next
                  center(1) = current_env%centers_set(ispin)%array(1, istate_next)
                  center(2) = current_env%centers_set(ispin)%array(2, istate_next)
                  center(3) = current_env%centers_set(ispin)%array(3, istate_next)
               ENDDO ! istate
               icount = 0
            ENDDO ! idir
         ENDDO
         DEALLOCATE (state_done)
      ELSE
         DO ispin = 1, nspins
            DO icenter = 1, current_env%nbr_center(ispin)
               current_env%statetrueindex(:, icenter, ispin) = icenter
            ENDDO
         ENDDO
      ENDIF
      !
      !
      IF (output_unit > 0) THEN
         WRITE (output_unit, "(T2,A,T60,A)") "CURRENT| Gauge used", &
            REPEAT(' ', 20-LEN_TRIM(current_env%gauge_name))//TRIM(current_env%gauge_name)
         WRITE (output_unit, "(T2,A,T79,L1)") "CURRENT| Use old gauge code", current_env%use_old_gauge_atom
         WRITE (output_unit, "(T2,A,T79,L1)") "CURRENT| Compute chi for PBC calculation", current_env%chi_pbc
         IF (current_env%gauge .EQ. current_gauge_atom) THEN
            WRITE (output_unit, "(T2,A,T68,E12.6)") "CURRENT| Radius for building the gauge", &
               current_env%gauge_atom_radius
         ENDIF
         WRITE (output_unit, "(T2,A,T60,A)") "CURRENT| Orbital center used", &
            REPEAT(' ', 20-LEN_TRIM(current_env%orb_center_name))//TRIM(current_env%orb_center_name)
         IF (current_env%orb_center .EQ. current_orb_center_common) THEN
            WRITE (output_unit, "(T2,A,T50,3F10.6)") "CURRENT| Common center", common_center(1:3)
         ELSEIF (current_env%orb_center .EQ. current_orb_center_box) THEN
            WRITE (output_unit, "(T2,A,T60,3I5)") "CURRENT| Nbr boxes in each direction", nbox(1:3)
         ENDIF
         !
         DO ispin = 1, nspins
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, nmo=nmo)
            WRITE (output_unit, '(T2,A,I6,A,I6,A,I2)') 'CURRENT| Compute', &
               nstate_list(ispin), ' selected response functions out of', &
               nmo, ' for spin ', ispin
            !
            WRITE (output_unit, '(T2,A,I6,A,I2)') 'CURRENT| There is a total of ', &
               current_env%nbr_center(ispin), ' (clustered) center(s) for spin ', ispin
         ENDDO
      ENDIF

      IF (BTEST(cp_print_key_should_output(logger%iter_info, current_section,&
           &    "PRINT%RESPONSE_FUNCTION_CUBES"), cp_p_file)) THEN

         NULLIFY (bounds, list)
         ncubes = 0
         CALL section_vals_val_get(current_section,&
              &                    "PRINT%RESPONSE_FUNCTION_CUBES%CUBES_LU_BOUNDS", &
                                   i_vals=bounds)
         ncubes = bounds(2)-bounds(1)+1
         IF (ncubes > 0) THEN
            ALLOCATE (current_env%list_cubes(ncubes))
            DO ir = 1, ncubes
               current_env%list_cubes(ir) = bounds(1)+(ir-1)
            ENDDO
         ENDIF
         IF (.NOT. ASSOCIATED(current_env%list_cubes)) THEN
            CALL section_vals_val_get(current_section, "PRINT%RESPONSE_FUNCTION_CUBES%CUBES_LIST", &
                                      n_rep_val=n_rep)
            ncubes = 0
            DO ir = 1, n_rep
               NULLIFY (list)
               CALL section_vals_val_get(current_section, "PRINT%RESPONSE_FUNCTION_CUBES%CUBES_LIST", &
                                         i_rep_val=ir, i_vals=list)
               IF (ASSOCIATED(list)) THEN
                  CALL reallocate(current_env%list_cubes, 1, ncubes+SIZE(list))
                  DO ini = 1, SIZE(list)
                     current_env%list_cubes(ini+ncubes) = list(ini)
                  ENDDO
                  ncubes = ncubes+SIZE(list)
               ENDIF
            ENDDO ! ir
         ENDIF
         IF (.NOT. ASSOCIATED(current_env%list_cubes)) THEN
            ALLOCATE (current_env%list_cubes(max_states))
            DO ir = 1, max_states
               current_env%list_cubes(ir) = ir
            ENDDO
         ENDIF
      ENDIF
      IF (BTEST(cp_print_key_should_output(logger%iter_info, current_section,&
           &    "PRINT%CURRENT_CUBES"), cp_p_file)) THEN
      ENDIF

      ! for the chemical shift we need 6 psi1, i.e. 6 optimization procedures
      ! They become 9 if full nmr is calculated, i.e. with the correction term too
      ! All of them are required at the end of the optimization procedure
      ! if the current density and the induced fiels have to be calculated
      ! If instead only the shift is needed, only one psi1 should be enough, providing
      ! that after every optimization the corresponding shift contribution is calculated
      ! prepare the psi1
      !
      ! for the rxp we cannot calculate it a priori because it is in facts (r-dk)xp
      ! where dk is the center of the orbital it is applied to. We would need nstate operators
      ! What we can store is (r-dk)xp|psi0k> for each k, which is a full matrix only
      ! Therefore we prepare here the full matrix p_psi0 and rxp_psi0
      ! We also need a temporary sparse matrix where to store the integrals during the calculation
      ALLOCATE (current_env%p_psi0(nspins, 3), current_env%rxp_psi0(nspins, 3), &
                current_env%psi1_p(nspins, 3), current_env%psi1_rxp(nspins, 3))
      IF (current_env%full) THEN
         ALLOCATE (current_env%psi1_D(nspins, 3))
      ENDIF
      DO ispin = 1, nspins
         mo_coeff => current_env%psi0_order(ispin)%matrix
         NULLIFY (tmp_fm_struct)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                  ncol_global=current_env%nstates(ispin), &
                                  context=mo_coeff%matrix_struct%context)
         DO idir = 1, 3
            NULLIFY (current_env%psi1_p(ispin, idir)%matrix, current_env%psi1_rxp(ispin, idir)%matrix)
            NULLIFY (current_env%p_psi0(ispin, idir)%matrix, current_env%rxp_psi0(ispin, idir)%matrix)
            CALL cp_fm_create(current_env%psi1_p(ispin, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(current_env%psi1_rxp(ispin, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(current_env%p_psi0(ispin, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(current_env%rxp_psi0(ispin, idir)%matrix, tmp_fm_struct)
            IF (current_env%full) THEN
               NULLIFY (current_env%psi1_D(ispin, idir)%matrix)
               CALL cp_fm_create(current_env%psi1_D(ispin, idir)%matrix, tmp_fm_struct)
            ENDIF
         ENDDO
         CALL cp_fm_struct_release(tmp_fm_struct)
      ENDDO
      !
      ! If the current density on the grid needs to be stored
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
      ALLOCATE (current_env%jrho1_set(3))
      DO idir = 1, 3
         NULLIFY (rho_r, rho_g)
         ALLOCATE (rho_r(nspins), rho_g(nspins))
         DO ispin = 1, nspins
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_r(ispin)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(rho_r(ispin)%pw)
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_g(ispin)%pw, &
                                   use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            CALL pw_zero(rho_g(ispin)%pw)
         ENDDO
         NULLIFY (current_env%jrho1_set(idir)%rho)
         CALL qs_rho_create(current_env%jrho1_set(idir)%rho)
         CALL qs_rho_set(current_env%jrho1_set(idir)%rho, &
                         rho_r=rho_r, rho_g=rho_g)
         CALL qs_rho_retain(current_env%jrho1_set(idir)%rho)
      ENDDO
      !
      ! Initialize local current density if GAPW calculation
      IF (gapw) THEN
         CALL init_jrho_atom_set(jrho1_atom_set, atomic_kind_set, nspins)
         CALL set_current_env(current_env=current_env, jrho1_atom_set=jrho1_atom_set)
      ENDIF
      !
      ALLOCATE (current_env%basisfun_center(3, current_env%nao), &
                first_sgf(natom), last_sgf(natom))
      current_env%basisfun_center = 0.0_dp
      CALL get_particle_set(particle_set, qs_kind_set, &
                            first_sgf=first_sgf, &
                            last_sgf=last_sgf)
      !Build 3 arrays where for each contracted basis function
      !the x y and z coordinates of the center are given
      DO iatom = 1, natom
         DO iao = first_sgf(iatom), last_sgf(iatom)
            DO idir = 1, 3
               current_env%basisfun_center(idir, iao) = particle_set(iatom)%r(idir)
            END DO
         END DO
      END DO

      DEALLOCATE (first_sgf, last_sgf, state_list)

      current_env%simple_done(1:6) = .FALSE.

      ALLOCATE (current_env%full_done(3*nspins, max_nbr_center))
      current_env%full_done = .FALSE.
      !
      ! allocate pointer for the gauge
      ALLOCATE (current_env%rs_gauge(3))
      NULLIFY (current_env%rs_gauge(1)%rs, current_env%rs_gauge(2)%rs, current_env%rs_gauge(3)%rs)

      CALL cp_print_key_finished_output(output_unit, logger, lr_section, "PRINT%PROGRAM_RUN_INFO")
      CALL timestop(handle)

   END SUBROUTINE current_env_init

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE current_env_cleanup(current_env, qs_env)

      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'current_env_cleanup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idir, ispin
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs

      NULLIFY (pw_env, rs_descs)

      current_env%ref_count = current_env%ref_count-1
      IF (current_env%ref_count == 0) THEN
         !
         !psi0_order
         IF (ASSOCIATED(current_env%psi0_order)) THEN
            DO ispin = 1, SIZE(current_env%psi0_order, 1)
               CALL cp_fm_release(current_env%psi0_order(ispin)%matrix)
            ENDDO
            DEALLOCATE (current_env%psi0_order)
         ENDIF
         !
         !psi1_p
         IF (ASSOCIATED(current_env%psi1_p)) THEN
            DO idir = 1, SIZE(current_env%psi1_p, 2)
               DO ispin = 1, SIZE(current_env%psi1_p, 1)
                  CALL cp_fm_release(current_env%psi1_p(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (current_env%psi1_p)
         ENDIF
         !
         !psi1_rxp
         IF (ASSOCIATED(current_env%psi1_rxp)) THEN
            DO idir = 1, SIZE(current_env%psi1_rxp, 2)
               DO ispin = 1, SIZE(current_env%psi1_rxp, 1)
                  CALL cp_fm_release(current_env%psi1_rxp(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (current_env%psi1_rxp)
         ENDIF
         !
         !psi1_D
         IF (ASSOCIATED(current_env%psi1_D)) THEN
            DO idir = 1, SIZE(current_env%psi1_D, 2)
               DO ispin = 1, SIZE(current_env%psi1_D, 1)
                  CALL cp_fm_release(current_env%psi1_D(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (current_env%psi1_D)
         END IF
         !
         !p_psi0
         IF (ASSOCIATED(current_env%p_psi0)) THEN
            DO idir = 1, SIZE(current_env%p_psi0, 2)
               DO ispin = 1, SIZE(current_env%p_psi0, 1)
                  CALL cp_fm_release(current_env%p_psi0(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (current_env%p_psi0)
         ENDIF
         !
         !rxp_psi0
         IF (ASSOCIATED(current_env%rxp_psi0)) THEN
            DO idir = 1, SIZE(current_env%rxp_psi0, 2)
               DO ispin = 1, SIZE(current_env%rxp_psi0, 1)
                  CALL cp_fm_release(current_env%rxp_psi0(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (current_env%rxp_psi0)
         ENDIF
         !
         !
         DO ispin = 1, SIZE(current_env%centers_set, 1)
            DEALLOCATE (current_env%centers_set(ispin)%array)
         END DO
         DEALLOCATE (current_env%centers_set)

         DO ispin = 1, SIZE(current_env%center_list, 1)
            DEALLOCATE (current_env%center_list(ispin)%array)
         END DO
         DEALLOCATE (current_env%center_list)

         IF (ASSOCIATED(current_env%list_cubes)) THEN
            DEALLOCATE (current_env%list_cubes)
         END IF
         ! Current density on the grid
         IF (ASSOCIATED(current_env%jrho1_set)) THEN
            DO idir = 1, 3
               CALL qs_rho_clear(current_env%jrho1_set(idir)%rho)
               DEALLOCATE (current_env%jrho1_set(idir)%rho)
            END DO
            DEALLOCATE (current_env%jrho1_set)
         END IF
         !
         ! Local current density, atom by atom (only gapw)
         IF (ASSOCIATED(current_env%jrho1_atom_set)) THEN
            CALL deallocate_jrho_atom_set(current_env%jrho1_atom_set)
         END IF

         !fullnmr_done
         IF (ASSOCIATED(current_env%full_done)) THEN
            DEALLOCATE (current_env%full_done)
         END IF
         IF (ASSOCIATED(current_env%basisfun_center)) THEN
            DEALLOCATE (current_env%basisfun_center)
         END IF
         IF (ASSOCIATED(current_env%statetrueindex)) THEN
            DEALLOCATE (current_env%statetrueindex)
         END IF
         IF (ASSOCIATED(current_env%selected_states_on_atom_list)) THEN
            DEALLOCATE (current_env%selected_states_on_atom_list)
         END IF
         ! give back the gauge
         CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
         CALL pw_env_get(pw_env, rs_descs=rs_descs)
         DO idir = 1, 3
            IF (ASSOCIATED(current_env%rs_gauge(idir)%rs)) THEN
               DO i = 1, SIZE(current_env%rs_gauge(idir)%rs)
                  CALL rs_grid_release(current_env%rs_gauge(idir)%rs(i)%rs_grid)
               END DO
               DEALLOCATE (current_env%rs_gauge(idir)%rs)
            ENDIF
         ENDDO
         !
         DEALLOCATE (current_env%rs_gauge)

         ! give back the buf
         IF (ASSOCIATED(current_env%rs_buf)) THEN
            DO i = 1, SIZE(current_env%rs_buf)
               CALL rs_grid_release(current_env%rs_buf(i)%rs_grid)
            END DO
            DEALLOCATE (current_env%rs_buf)
         ENDIF

      END IF ! ref count

   END SUBROUTINE current_env_cleanup

END MODULE qs_linres_current_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief given the response wavefunctions obtained by the application
!>      of the (rxp), p, and ((dk-dl)xp) operators,
!>      here the current density vector (jx, jy, jz)
!>      is computed for the 3 directions of the magnetic field (Bx, By, Bz)
!> \par History
!>      created 02-2006 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_nmr_epr_common_utils
   USE cell_types,                      ONLY: cell_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: gaussi
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_transfer
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Public subroutines ***
   PUBLIC :: mult_G_ov_G2_grid

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_nmr_epr_common_utils'

CONTAINS

! **************************************************************************************************
!> \brief Given the current density on the PW grid in reciprcal space
!>       (obtained by FFT), calculate the integral
!>         \int_{r}[ ((r-r') x j(r))/|r-r'|^3 ] = Bind(r')
!>       which in reciprcal space reads  (for G/=0)
!>          i G/|G|^2 x J(G)
!> \param cell ...
!> \param pw_pool ...
!> \param rho_gspace ...
!> \param funcG_times_rho ...
!> \param idir ...
!> \param my_chi ...
!> \author MI
!> \note
!>      The G=0 component is not comnputed here, but can be evaluated
!>      through the susceptibility and added to the shift in a second time
!>
!>      This method would not work for a non periodic system
!>      It should be generalized like the calculation of Hartree
! **************************************************************************************************
   SUBROUTINE mult_G_ov_G2_grid(cell, pw_pool, rho_gspace, funcG_times_rho, idir, my_chi)

      TYPE(cell_type), POINTER                           :: cell
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_p_type), POINTER                           :: rho_gspace
      TYPE(pw_p_type)                                    :: funcG_times_rho
      INTEGER, INTENT(IN)                                :: idir
      REAL(dp), INTENT(IN)                               :: my_chi

      INTEGER                                            :: handle, ig, ng
      REAL(dp)                                           :: g2
      TYPE(pw_grid_type), POINTER                        :: grid
      CHARACTER(len=*), PARAMETER :: routineN = 'mult_G_ov_G2_grid', &
         routineP = moduleN//':'//routineN

      TYPE(pw_type), POINTER                             :: frho, influence_fn

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("qs_linres_nmr_epr_common_utils.F",80)

      CALL pw_pool_create_pw(pw_pool, influence_fn, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      grid => influence_fn%pw_grid
      DO ig = grid%first_gne0, grid%ngpts_cut_local
         g2 = grid%gsq(ig)
         influence_fn%cc(ig) = gaussi*grid%g(idir, ig)/g2
      END DO ! ig
      IF (grid%have_g0) influence_fn%cc(1) = 0.0_dp

      frho => funcG_times_rho%pw
      CALL pw_transfer(rho_gspace%pw, frho)

      ng = SIZE(grid%gsq)
      frho%cc(1:ng) = frho%cc(1:ng)*influence_fn%cc(1:ng)
      IF (grid%have_g0) frho%cc(1) = my_chi

      CALL pw_pool_give_back_pw(pw_pool, influence_fn, &
                                accept_non_compatible=.TRUE.)

      CALL timestop(handle)

   END SUBROUTINE mult_G_ov_G2_grid

END MODULE qs_linres_nmr_epr_common_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>       created 06-2006 [RD]
!> \author RD
! **************************************************************************************************
MODULE qs_linres_epr_ownutils
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathlib,                         ONLY: diamat_all
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_integral_ab,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_spline_utils,                 ONLY: Eval_Interp_Spl3_pbc,&
                                              find_coeffs,&
                                              pw_spline_do_precond,&
                                              pw_spline_precond_create,&
                                              pw_spline_precond_release,&
                                              pw_spline_precond_set_kind,&
                                              pw_spline_precond_type,&
                                              spl3_pbc
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_core_energies,                ONLY: calculate_ptrace
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_linres_nmr_epr_common_utils,  ONLY: mult_G_ov_G2_grid
   USE qs_linres_op,                    ONLY: fac_vecp,&
                                              set_vecp,&
                                              set_vecp_rev
   USE qs_linres_types,                 ONLY: current_env_type,&
                                              epr_env_type,&
                                              get_current_env,&
                                              get_epr_env,&
                                              jrho_atom_type,&
                                              nablavks_atom_type
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_p_type,&
                                              qs_rho_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: epr_g_print, epr_g_zke, epr_g_so, epr_g_soo, epr_ind_magnetic_field

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_epr_ownutils'

CONTAINS

! **************************************************************************************************
!> \brief Prints the g tensor
!> \param epr_env ...
!> \param qs_env ...
!> \par History
!>      06.2006 created [RD]
!> \author RD
! **************************************************************************************************
   SUBROUTINE epr_g_print(epr_env, qs_env)

      TYPE(epr_env_type)                                 :: epr_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'epr_g_print', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: title
      INTEGER                                            :: idir1, idir2, output_unit, unit_nr
      REAL(KIND=dp)                                      :: eigenv_g(3), g_sym(3, 3), gsum
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: lr_section

      NULLIFY (logger, lr_section)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      gsum = 0.0_dp

      DO idir1 = 1, 3
         DO idir2 = 1, 3
            gsum = gsum+epr_env%g_total(idir1, idir2)
         END DO
      END DO

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(T2,A,T56,E14.6)") &
            "epr|TOT:checksum", gsum
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      IF (BTEST(cp_print_key_should_output(logger%iter_info, lr_section, &
                                           "EPR%PRINT%G_TENSOR"), cp_p_file)) THEN

         unit_nr = cp_print_key_unit_nr(logger, lr_section, "EPR%PRINT%G_TENSOR", &
                                        extension=".data", middle_name="GTENSOR", &
                                        log_filename=.FALSE.)

         IF (unit_nr > 0) THEN

            WRITE (title, "(A)") "G tensor "
            WRITE (unit_nr, "(T2,A)") title

            WRITE (unit_nr, "(T2,A)") "gmatrix_zke"
            WRITE (unit_nr, "(3(A,f15.10))") " XX=", epr_env%g_zke, &
               " XY=", 0.0_dp, " XZ=", 0.0_dp
            WRITE (unit_nr, "(3(A,f15.10))") " YX=", 0.0_dp, &
               " YY=", epr_env%g_zke, " YZ=", 0.0_dp
            WRITE (unit_nr, "(3(A,f15.10))") " ZX=", 0.0_dp, &
               " ZY=", 0.0_dp, " ZZ=", epr_env%g_zke

            WRITE (unit_nr, "(T2,A)") "gmatrix_so"
            WRITE (unit_nr, "(3(A,f15.10))") " XX=", epr_env%g_so(1, 1), &
               " XY=", epr_env%g_so(1, 2), " XZ=", epr_env%g_so(1, 3)
            WRITE (unit_nr, "(3(A,f15.10))") " YX=", epr_env%g_so(2, 1), &
               " YY=", epr_env%g_so(2, 2), " YZ=", epr_env%g_so(2, 3)
            WRITE (unit_nr, "(3(A,f15.10))") " ZX=", epr_env%g_so(3, 1), &
               " ZY=", epr_env%g_so(3, 2), " ZZ=", epr_env%g_so(3, 3)

            WRITE (unit_nr, "(T2,A)") "gmatrix_soo"
            WRITE (unit_nr, "(3(A,f15.10))") " XX=", epr_env%g_soo(1, 1), &
               " XY=", epr_env%g_soo(1, 2), " XZ=", epr_env%g_soo(1, 3)
            WRITE (unit_nr, "(3(A,f15.10))") " YX=", epr_env%g_soo(2, 1), &
               " YY=", epr_env%g_soo(2, 2), " YZ=", epr_env%g_soo(2, 3)
            WRITE (unit_nr, "(3(A,f15.10))") " ZX=", epr_env%g_soo(3, 1), &
               " ZY=", epr_env%g_soo(3, 2), " ZZ=", epr_env%g_soo(3, 3)

            WRITE (unit_nr, "(T2,A)") "gmatrix_total"
            WRITE (unit_nr, "(3(A,f15.10))") " XX=", epr_env%g_total(1, 1)+epr_env%g_free_factor, &
               " XY=", epr_env%g_total(1, 2), " XZ=", epr_env%g_total(1, 3)
            WRITE (unit_nr, "(3(A,f15.10))") " YX=", epr_env%g_total(2, 1), &
               " YY=", epr_env%g_total(2, 2)+epr_env%g_free_factor, " YZ=", epr_env%g_total(2, 3)
            WRITE (unit_nr, "(3(A,f15.10))") " ZX=", epr_env%g_total(3, 1), &
               " ZY=", epr_env%g_total(3, 2), " ZZ=", epr_env%g_total(3, 3)+epr_env%g_free_factor

            DO idir1 = 1, 3
               DO idir2 = 1, 3
                  g_sym(idir1, idir2) = (epr_env%g_total(idir1, idir2)+ &
                                         epr_env%g_total(idir2, idir1))/2.0_dp
               END DO
            END DO

            WRITE (unit_nr, "(T2,A)") "gtensor_total"
            WRITE (unit_nr, "(3(A,f15.10))") " XX=", g_sym(1, 1)+epr_env%g_free_factor, &
               " XY=", g_sym(1, 2), " XZ=", g_sym(1, 3)
            WRITE (unit_nr, "(3(A,f15.10))") " YX=", g_sym(2, 1), &
               " YY=", g_sym(2, 2)+epr_env%g_free_factor, " YZ=", g_sym(2, 3)
            WRITE (unit_nr, "(3(A,f15.10))") " ZX=", g_sym(3, 1), &
               " ZY=", g_sym(3, 2), " ZZ=", g_sym(3, 3)+epr_env%g_free_factor

            CALL diamat_all(g_sym, eigenv_g)
            eigenv_g(:) = eigenv_g(:)*1.0e6_dp

            WRITE (unit_nr, "(T2,A)") "delta_g principal values in ppm"
            WRITE (unit_nr, "(f15.3,3(A,f15.10))") eigenv_g(1), " X=", g_sym(1, 1), &
               " Y=", g_sym(2, 1), " Z=", g_sym(3, 1)
            WRITE (unit_nr, "(f15.3,3(A,f15.10))") eigenv_g(2), " X=", g_sym(1, 2), &
               " Y=", g_sym(2, 2), " Z=", g_sym(3, 2)
            WRITE (unit_nr, "(f15.3,3(A,f15.10))") eigenv_g(3), " X=", g_sym(1, 3), &
               " Y=", g_sym(2, 3), " Z=", g_sym(3, 3)

         END IF

         CALL cp_print_key_finished_output(unit_nr, logger, lr_section,&
              &                            "EPR%PRINT%G_TENSOR")

      ENDIF

   END SUBROUTINE epr_g_print

! **************************************************************************************************
!> \brief Calculate zke part of the g tensor
!> \param epr_env ...
!> \param qs_env ...
!> \par History
!>      06.2006 created [RD]
!> \author RD
! **************************************************************************************************
   SUBROUTINE epr_g_zke(epr_env, qs_env)

      TYPE(epr_env_type)                                 :: epr_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'epr_g_zke', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, ispin, output_unit
      REAL(KIND=dp)                                      :: epr_g_zke_temp(2)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: kinetic, rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: lr_section

      NULLIFY (dft_control, logger, lr_section, rho, kinetic, para_env, rho_ao)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      kinetic=kinetic, rho=rho, para_env=para_env)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      DO ispin = 1, dft_control%nspins
         CALL calculate_ptrace(kinetic(1)%matrix, rho_ao(ispin)%matrix, &
                               ecore=epr_g_zke_temp(ispin))
      END DO

      epr_env%g_zke = epr_env%g_zke_factor*(epr_g_zke_temp(1)-epr_g_zke_temp(2))
      DO i1 = 1, 3
         epr_env%g_total(i1, i1) = epr_env%g_total(i1, i1)+epr_env%g_zke
      END DO

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(T2,A,T56,E24.16)") &
            "epr|ZKE:g_zke", epr_env%g_zke
      END IF

      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE epr_g_zke

! **************************************************************************************************
!> \brief Calculates g_so
!> \param epr_env ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
!> \par History
!>      06.2006 created [RD]
!> \author RD
! **************************************************************************************************
   SUBROUTINE epr_g_so(epr_env, current_env, qs_env, iB)

      TYPE(epr_env_type)                                 :: epr_env
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB

      CHARACTER(LEN=*), PARAMETER :: routineN = 'epr_g_so', routineP = moduleN//':'//routineN

      INTEGER                                            :: aint_precond, ia, iat, iatom, idir1, &
                                                            idir2, idir3, ikind, ir, ispin, &
                                                            max_iter, natom, nkind, nspins, &
                                                            output_unit, precond_kind
      INTEGER, DIMENSION(2)                              :: bo
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: gapw, paw_atom, success
      REAL(dp)                                           :: eps_r, eps_x, hard_radius, temp_so_soft, &
                                                            vks_ra_idir2, vks_ra_idir3
      REAL(dp), DIMENSION(3, 3)                          :: temp_so_gapw
      REAL(dp), DIMENSION(:, :), POINTER                 :: g_so, g_total
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      TYPE(jrho_atom_type), POINTER                      :: jrho1_atom
      TYPE(nablavks_atom_type), DIMENSION(:), POINTER    :: nablavks_atom_set
      TYPE(nablavks_atom_type), POINTER                  :: nablavks_atom
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: jrho2_r, jrho3_r, nrho1_r, nrho2_r, &
                                                            nrho3_r
      TYPE(pw_p_type), DIMENSION(:, :), POINTER          :: vks_pw_spline
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_spline_precond_type), POINTER              :: precond
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_p_type), DIMENSION(:), POINTER         :: jrho1_set
      TYPE(qs_rho_p_type), DIMENSION(:, :), POINTER      :: nablavks_set
      TYPE(section_vals_type), POINTER                   :: interp_section, lr_section

      NULLIFY (atomic_kind_set, qs_kind_set, atom_list, dft_control, &
               grid_atom, g_so, g_total, harmonics, interp_section, jrho1_atom_set, &
               jrho1_set, logger, lr_section, nablavks_atom, nablavks_atom_set, &
               nablavks_set, para_env, particle_set, jrho2_r, jrho3_r, nrho1_r, nrho2_r, nrho3_r)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      para_env=para_env, pw_env=pw_env, &
                      particle_set=particle_set)

      CALL get_epr_env(epr_env=epr_env, &
                       nablavks_set=nablavks_set, &
                       nablavks_atom_set=nablavks_atom_set, &
                       g_total=g_total, g_so=g_so)

      CALL get_current_env(current_env=current_env, &
                           jrho1_set=jrho1_set, jrho1_atom_set=jrho1_atom_set)

      gapw = dft_control%qs_control%gapw
      nkind = SIZE(qs_kind_set, 1)
      nspins = dft_control%nspins

      DO idir1 = 1, 3
         CALL set_vecp(idir1, idir2, idir3)
         ! j_pw x nabla_vks_pw
         temp_so_soft = 0.0_dp
         DO ispin = 1, nspins
            CALL qs_rho_get(jrho1_set(idir2)%rho, rho_r=jrho2_r)
            CALL qs_rho_get(jrho1_set(idir3)%rho, rho_r=jrho3_r)
            CALL qs_rho_get(nablavks_set(idir2, ispin)%rho, rho_r=nrho2_r)
            CALL qs_rho_get(nablavks_set(idir3, ispin)%rho, rho_r=nrho3_r)
            temp_so_soft = temp_so_soft+(-1.0_dp)**(1+ispin)*( &
                           pw_integral_ab(jrho2_r(ispin)%pw, nrho3_r(1)%pw)- &
                           pw_integral_ab(jrho3_r(ispin)%pw, nrho2_r(1)%pw))
         END DO
         temp_so_soft = -1.0_dp*epr_env%g_so_factor*temp_so_soft
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T18,I1,I1,T56,E24.16)") &
               "epr|SOX:soft", iB, idir1, temp_so_soft
         END IF
         g_so(iB, idir1) = temp_so_soft
      END DO !idir1

      IF (gapw) THEN

         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
         ALLOCATE (vks_pw_spline(3, nspins))

         interp_section => section_vals_get_subs_vals(lr_section, &
                                                      "EPR%INTERPOLATOR")
         CALL section_vals_val_get(interp_section, "aint_precond", &
                                   i_val=aint_precond)
         CALL section_vals_val_get(interp_section, "precond", i_val=precond_kind)
         CALL section_vals_val_get(interp_section, "max_iter", i_val=max_iter)
         CALL section_vals_val_get(interp_section, "eps_r", r_val=eps_r)
         CALL section_vals_val_get(interp_section, "eps_x", r_val=eps_x)

         DO ispin = 1, nspins
            DO idir1 = 1, 3
               CALL pw_pool_create_pw(auxbas_pw_pool, &
                                      vks_pw_spline(idir1, ispin)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
               ! calculate spline coefficients
               CALL pw_spline_precond_create(precond, precond_kind=aint_precond, &
                                             pool=auxbas_pw_pool, pbc=.TRUE., transpose=.FALSE.)
               CALL qs_rho_get(nablavks_set(idir1, ispin)%rho, rho_r=nrho1_r)
               CALL pw_spline_do_precond(precond, nrho1_r(1)%pw, &
                                         vks_pw_spline(idir1, ispin)%pw)
               CALL pw_spline_precond_set_kind(precond, precond_kind)
               success = find_coeffs(values=nrho1_r(1)%pw, &
                                     coeffs=vks_pw_spline(idir1, ispin)%pw, linOp=spl3_pbc, &
                                     preconditioner=precond, pool=auxbas_pw_pool, &
                                     eps_r=eps_r, eps_x=eps_x, max_iter=max_iter)
               IF(.NOT.(success))CALL cp__a("qs_linres_epr_ownutils.F",406)
               CALL pw_spline_precond_release(precond)
            END DO ! idir1
         END DO ! ispin

         temp_so_gapw = 0.0_dp

         DO ikind = 1, nkind
            NULLIFY (atom_list, grid_atom, harmonics)
            CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             hard_radius=hard_radius, &
                             grid_atom=grid_atom, &
                             harmonics=harmonics, &
                             paw_atom=paw_atom)

            IF (.NOT. paw_atom) CYCLE

            ! Distribute the atoms of this kind

            bo = get_limit(natom, para_env%num_pe, para_env%mepos)

            DO iat = 1, natom !bo(1),bo(2)! this partitioning blocks the interpolation
               !                         ! routines (i.e. waiting for mp_sum)
               iatom = atom_list(iat)
               NULLIFY (jrho1_atom, nablavks_atom)
               jrho1_atom => jrho1_atom_set(iatom)
               nablavks_atom => nablavks_atom_set(iatom)
               DO idir1 = 1, 3
                  CALL set_vecp(idir1, idir2, idir3)
                  DO ispin = 1, nspins
                     DO ir = 1, grid_atom%nr

                        IF (grid_atom%rad(ir) >= hard_radius) CYCLE

                        DO ia = 1, grid_atom%ng_sphere

                           ra = particle_set(iatom)%r
                           ra(:) = ra(:)+grid_atom%rad(ir)*harmonics%a(:, ia)
                           vks_ra_idir2 = Eval_Interp_Spl3_pbc(ra, &
                                                               vks_pw_spline(idir2, ispin)%pw)
                           vks_ra_idir3 = Eval_Interp_Spl3_pbc(ra, &
                                                               vks_pw_spline(idir3, ispin)%pw)

                           IF (iat .LT. bo(1) .OR. iat .GT. bo(2)) CYCLE !quick and dirty:
                           !                                    !here take care of the partition

                           ! + sum_A j_loc_h_A x nabla_vks_s_A
                           temp_so_gapw(iB, idir1) = temp_so_gapw(iB, idir1)+ &
                                                     (-1.0_dp)**(1+ispin)*( &
                                                     jrho1_atom%jrho_vec_rad_h(idir2, ispin)%r_coef(ir, ia)* &
                                                     vks_ra_idir3- &
                                                     jrho1_atom%jrho_vec_rad_h(idir3, ispin)%r_coef(ir, ia)* &
                                                     vks_ra_idir2 &
                                                     )*grid_atom%wr(ir)*grid_atom%wa(ia)

                           ! - sum_A j_loc_s_A x nabla_vks_s_A
                           temp_so_gapw(iB, idir1) = temp_so_gapw(iB, idir1)- &
                                                     (-1.0_dp)**(1+ispin)*( &
                                                     jrho1_atom%jrho_vec_rad_s(idir2, ispin)%r_coef(ir, ia)* &
                                                     vks_ra_idir3- &
                                                     jrho1_atom%jrho_vec_rad_s(idir3, ispin)%r_coef(ir, ia)* &
                                                     vks_ra_idir2 &
                                                     )*grid_atom%wr(ir)*grid_atom%wa(ia)

                           ! + sum_A j_loc_h_A x nabla_vks_loc_h_A
                           temp_so_gapw(iB, idir1) = temp_so_gapw(iB, idir1)+ &
                                                     (-1.0_dp)**(1+ispin)*( &
                                                     jrho1_atom%jrho_vec_rad_h(idir2, ispin)%r_coef(ir, ia)* &
                                                     nablavks_atom%nablavks_vec_rad_h(idir3, ispin)%r_coef(ir, ia)- &
                                                     jrho1_atom%jrho_vec_rad_h(idir3, ispin)%r_coef(ir, ia)* &
                                                     nablavks_atom%nablavks_vec_rad_h(idir2, ispin)%r_coef(ir, ia) &
                                                     )*grid_atom%wr(ir)*grid_atom%wa(ia)

                           ! - sum_A j_loc_h_A x nabla_vks_loc_s_A
                           temp_so_gapw(iB, idir1) = temp_so_gapw(iB, idir1)- &
                                                     (-1.0_dp)**(1+ispin)*( &
                                                     jrho1_atom%jrho_vec_rad_h(idir2, ispin)%r_coef(ir, ia)* &
                                                     nablavks_atom%nablavks_vec_rad_s(idir3, ispin)%r_coef(ir, ia)- &
                                                     jrho1_atom%jrho_vec_rad_h(idir3, ispin)%r_coef(ir, ia)* &
                                                     nablavks_atom%nablavks_vec_rad_s(idir2, ispin)%r_coef(ir, ia) &
                                                     )*grid_atom%wr(ir)*grid_atom%wa(ia)

! ORIGINAL
!                            ! + sum_A j_loc_h_A x nabla_vks_loc_h_A
!                            temp_so_gapw(iB,idir1) = temp_so_gapw(iB,idir1) + &
!                               (-1.0_dp)**(1.0_dp + REAL(ispin,KIND=dp)) * ( &
!                               jrho1_atom%jrho_vec_rad_h(idir2,iB,ispin)%r_coef(ir,ia) * &
!                               nablavks_atom%nablavks_vec_rad_h(idir3,ispin)%r_coef(ir,ia) - &
!                               jrho1_atom%jrho_vec_rad_h(idir3,iB,ispin)%r_coef(ir,ia) * &
!                               nablavks_atom%nablavks_vec_rad_h(idir2,ispin)%r_coef(ir,ia) &
!                               ) * grid_atom%wr(ir)*grid_atom%wa(ia)
!                            ! - sum_A j_loc_s_A x nabla_vks_loc_s_A
!                            temp_so_gapw(iB,idir1) = temp_so_gapw(iB,idir1) - &
!                               (-1.0_dp)**(1.0_dp + REAL(ispin,KIND=dp)) * ( &
!                               jrho1_atom%jrho_vec_rad_s(idir2,iB,ispin)%r_coef(ir,ia) * &
!                               nablavks_atom%nablavks_vec_rad_s(idir3,ispin)%r_coef(ir,ia) - &
!                               jrho1_atom%jrho_vec_rad_s(idir3,iB,ispin)%r_coef(ir,ia) * &
!                               nablavks_atom%nablavks_vec_rad_s(idir2,ispin)%r_coef(ir,ia) &
!                               ) * grid_atom%wr(ir)*grid_atom%wa(ia)
                        END DO !ia
                     END DO !ir
                  END DO !ispin
               END DO !idir1
            END DO !iat
         END DO !ikind

         CALL mp_sum(temp_so_gapw, para_env%group)
         temp_so_gapw(:, :) = -1.0_dp*epr_env%g_so_factor_gapw*temp_so_gapw(:, :)

         IF (output_unit > 0) THEN
            DO idir1 = 1, 3
               WRITE (UNIT=output_unit, FMT="(T2,A,T18,I1,I1,T56,E24.16)") &
                  "epr|SOX:gapw", iB, idir1, temp_so_gapw(iB, idir1)
            END DO
         END IF

         g_so(iB, :) = g_so(iB, :)+temp_so_gapw(iB, :)

         DO ispin = 1, nspins
            DO idir1 = 1, 3
               CALL pw_pool_give_back_pw(auxbas_pw_pool, vks_pw_spline(idir1, ispin)%pw)
            END DO
         END DO

         DEALLOCATE (vks_pw_spline)

      END IF ! gapw

      g_total(iB, :) = g_total(iB, :)+g_so(iB, :)

      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE epr_g_so

! **************************************************************************************************
!> \brief Calculates g_soo (soft part only for now)
!> \param epr_env ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
!> \par History
!>      06.2006 created [RD]
!> \author RD
! **************************************************************************************************
   SUBROUTINE epr_g_soo(epr_env, current_env, qs_env, iB)

      TYPE(epr_env_type)                                 :: epr_env
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB

      CHARACTER(LEN=*), PARAMETER :: routineN = 'epr_g_soo', routineP = moduleN//':'//routineN

      INTEGER                                            :: aint_precond, ia, iat, iatom, idir1, &
                                                            ikind, ir, iso, ispin, max_iter, &
                                                            natom, nkind, nspins, output_unit, &
                                                            precond_kind
      INTEGER, DIMENSION(2)                              :: bo
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: gapw, paw_atom, soo_rho_hard, success
      REAL(dp)                                           :: bind_ra_idir1, chi_tensor(3, 3, 2), &
                                                            eps_r, eps_x, hard_radius, rho_spin, &
                                                            temp_soo_soft
      REAL(dp), DIMENSION(3, 3)                          :: temp_soo_gapw
      REAL(dp), DIMENSION(:, :), POINTER                 :: g_soo, g_total
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: brho1_r, rho_r
      TYPE(pw_p_type), DIMENSION(:, :), POINTER          :: bind_pw_spline
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_spline_precond_type), POINTER              :: precond
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_p_type), DIMENSION(:, :), POINTER      :: bind_set
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: rho_rad_h, rho_rad_s
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      TYPE(section_vals_type), POINTER                   :: g_section, interp_section, lr_section

      NULLIFY (atomic_kind_set, qs_kind_set, atom_list, bind_set, dft_control, &
               grid_atom, g_section, g_soo, g_total, harmonics, interp_section, &
               logger, lr_section, para_env, particle_set, rho, rho_atom, &
               rho_atom_set, rho_r, brho1_r)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      g_section => section_vals_get_subs_vals(lr_section, &
                                              "EPR%PRINT%G_TENSOR")

      CALL section_vals_val_get(g_section, "soo_rho_hard", l_val=soo_rho_hard)

      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, &
                      dft_control=dft_control, para_env=para_env, particle_set=particle_set, &
                      pw_env=pw_env, rho=rho, rho_atom_set=rho_atom_set)

      CALL get_epr_env(epr_env=epr_env, bind_set=bind_set, &
                       g_soo=g_soo, g_total=g_total)

      CALL get_current_env(current_env=current_env, &
                           chi_tensor=chi_tensor)
      CALL qs_rho_get(rho, rho_r=rho_r)

      gapw = dft_control%qs_control%gapw
      nkind = SIZE(qs_kind_set, 1)
      nspins = dft_control%nspins

      DO idir1 = 1, 3
         temp_soo_soft = 0.0_dp
         DO ispin = 1, nspins
            CALL qs_rho_get(bind_set(idir1, iB)%rho, rho_r=brho1_r)
            temp_soo_soft = temp_soo_soft+(-1.0_dp)**(1+ispin)* &
                            pw_integral_ab(brho1_r(1)%pw, rho_r(ispin)%pw)
         END DO
         temp_soo_soft = 1.0_dp*epr_env%g_soo_factor*temp_soo_soft
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T18,i1,i1,T56,E24.16)") &
               "epr|SOO:soft", iB, idir1, temp_soo_soft
         END IF
         g_soo(iB, idir1) = temp_soo_soft
      END DO

      DO idir1 = 1, 3
         temp_soo_soft = 1.0_dp*epr_env%g_soo_chicorr_factor*chi_tensor(idir1, iB, 2)* &
                         (REAL(dft_control%multiplicity, KIND=dp)-1.0_dp)
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T18,i1,i1,T56,E24.16)") &
               "epr|SOO:soft_g0", iB, idir1, temp_soo_soft
         END IF
         g_soo(iB, idir1) = g_soo(iB, idir1)+temp_soo_soft
      END DO

      IF (gapw .AND. soo_rho_hard) THEN

         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
         ALLOCATE (bind_pw_spline(3, 3))

         interp_section => section_vals_get_subs_vals(lr_section, &
                                                      "EPR%INTERPOLATOR")
         CALL section_vals_val_get(interp_section, "aint_precond", &
                                   i_val=aint_precond)
         CALL section_vals_val_get(interp_section, "precond", i_val=precond_kind)
         CALL section_vals_val_get(interp_section, "max_iter", i_val=max_iter)
         CALL section_vals_val_get(interp_section, "eps_r", r_val=eps_r)
         CALL section_vals_val_get(interp_section, "eps_x", r_val=eps_x)

         DO idir1 = 1, 3
            CALL pw_pool_create_pw(auxbas_pw_pool, &
                                   bind_pw_spline(idir1, iB)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            ! calculate spline coefficients
            CALL pw_spline_precond_create(precond, precond_kind=aint_precond, &
                                          pool=auxbas_pw_pool, pbc=.TRUE., transpose=.FALSE.)
            CALL qs_rho_get(bind_set(idir1, iB)%rho, rho_r=brho1_r)
            CALL pw_spline_do_precond(precond, brho1_r(1)%pw, &
                                      bind_pw_spline(idir1, iB)%pw)
            CALL pw_spline_precond_set_kind(precond, precond_kind)
            success = find_coeffs(values=brho1_r(1)%pw, &
                                  coeffs=bind_pw_spline(idir1, iB)%pw, linOp=spl3_pbc, &
                                  preconditioner=precond, pool=auxbas_pw_pool, &
                                  eps_r=eps_r, eps_x=eps_x, max_iter=max_iter)
            IF(.NOT.(success))CALL cp__a("qs_linres_epr_ownutils.F",679)
            CALL pw_spline_precond_release(precond)
         END DO ! idir1

         temp_soo_gapw = 0.0_dp

         DO ikind = 1, nkind
            NULLIFY (atom_list, grid_atom, harmonics)
            CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             hard_radius=hard_radius, &
                             grid_atom=grid_atom, &
                             harmonics=harmonics, &
                             paw_atom=paw_atom)

            IF (.NOT. paw_atom) CYCLE

            ! Distribute the atoms of this kind

            bo = get_limit(natom, para_env%num_pe, para_env%mepos)

            DO iat = 1, natom !bo(1),bo(2)! this partitioning blocks the interpolation
               !                         ! routines (i.e. waiting for mp_sum)
               iatom = atom_list(iat)
               rho_atom => rho_atom_set(iatom)
               NULLIFY (rho_rad_h, rho_rad_s)
               CALL get_rho_atom(rho_atom=rho_atom, rho_rad_h=rho_rad_h, &
                                 rho_rad_s=rho_rad_s)
               DO idir1 = 1, 3
                  DO ispin = 1, nspins
                     DO ir = 1, grid_atom%nr

                        IF (grid_atom%rad(ir) >= hard_radius) CYCLE

                        DO ia = 1, grid_atom%ng_sphere

                           ra = particle_set(iatom)%r
                           ra(:) = ra(:)+grid_atom%rad(ir)*harmonics%a(:, ia)
                           bind_ra_idir1 = Eval_Interp_Spl3_pbc(ra, &
                                                                bind_pw_spline(idir1, iB)%pw)

                           IF (iat .LT. bo(1) .OR. iat .GT. bo(2)) CYCLE !quick and dirty:
                           !                                    !here take care of the partition

                           rho_spin = 0.0_dp

                           DO iso = 1, harmonics%max_iso_not0
                              rho_spin = rho_spin+ &
                                         (rho_rad_h(ispin)%r_coef(ir, iso)- &
                                          rho_rad_s(ispin)%r_coef(ir, iso))* &
                                         harmonics%slm(ia, iso)
                           END DO

                           temp_soo_gapw(iB, idir1) = temp_soo_gapw(iB, idir1)+ &
                                                      (-1.0_dp)**(1+ispin)*( &
                                                      bind_ra_idir1*rho_spin &
                                                      )*grid_atom%wr(ir)*grid_atom%wa(ia)

                        END DO !ia
                     END DO !ir
                  END DO ! ispin
               END DO !idir1
            END DO !iat
         END DO !ikind

         CALL mp_sum(temp_soo_gapw, para_env%group)
         temp_soo_gapw(:, :) = 1.0_dp*epr_env%g_soo_factor*temp_soo_gapw(:, :)

         IF (output_unit > 0) THEN
            DO idir1 = 1, 3
               WRITE (UNIT=output_unit, FMT="(T2,A,T18,I1,I1,T56,E24.16)") &
                  "epr|SOO:gapw", iB, idir1, temp_soo_gapw(iB, idir1)
            END DO
         END IF

         g_soo(iB, :) = g_soo(iB, :)+temp_soo_gapw(iB, :)

         DO idir1 = 1, 3
            CALL pw_pool_give_back_pw(auxbas_pw_pool, bind_pw_spline(idir1, iB)%pw)
         END DO

         DEALLOCATE (bind_pw_spline)

      END IF ! gapw

      g_total(iB, :) = g_total(iB, :)+g_soo(iB, :)

      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE epr_g_soo

! **************************************************************************************************
!> \brief ...
!> \param epr_env ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
! **************************************************************************************************
   SUBROUTINE epr_ind_magnetic_field(epr_env, current_env, qs_env, iB)

      TYPE(epr_env_type)                                 :: epr_env
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB

      CHARACTER(LEN=*), PARAMETER :: routineN = 'epr_ind_magnetic_field', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idir, idir2, idir3, iiB, iiiB, &
                                                            ispin, natom, nspins
      LOGICAL                                            :: gapw
      REAL(dp)                                           :: scale_fac
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: pw_gspace_work, shift_pw_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: epr_rho_r, jrho1_g
      TYPE(pw_p_type), DIMENSION(:, :), POINTER          :: shift_pw_gspace
      TYPE(pw_p_type), POINTER                           :: rho_gspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(realspace_grid_desc_type), POINTER            :: auxbas_rs_desc

      CALL timeset(routineN, handle)

      NULLIFY (cell, dft_control, pw_env, auxbas_rs_desc, auxbas_pw_pool, &
               rho_gspace, pw_pools, particle_set, jrho1_g, epr_rho_r)

      CALL get_qs_env(qs_env=qs_env, cell=cell, dft_control=dft_control, &
                      particle_set=particle_set)

      gapw = dft_control%qs_control%gapw
      natom = SIZE(particle_set, 1)
      nspins = dft_control%nspins

      CALL get_epr_env(epr_env=epr_env)

      CALL get_current_env(current_env=current_env)

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_desc=auxbas_rs_desc, &
                      auxbas_pw_pool=auxbas_pw_pool, pw_pools=pw_pools)
      !
      ! Initialize
      ! Allocate grids for the calculation of jrho and the shift
      ALLOCATE (shift_pw_gspace(3, nspins))
      DO ispin = 1, nspins
         DO idir = 1, 3
            CALL pw_pool_create_pw(auxbas_pw_pool, shift_pw_gspace(idir, ispin)%pw, &
                                   use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
            CALL pw_zero(shift_pw_gspace(idir, ispin)%pw)
         ENDDO
      ENDDO
      CALL pw_pool_create_pw(auxbas_pw_pool, shift_pw_rspace%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_zero(shift_pw_rspace%pw)
      CALL pw_pool_create_pw(auxbas_pw_pool, pw_gspace_work%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      CALL pw_zero(pw_gspace_work%pw)
      !
      CALL set_vecp(iB, iiB, iiiB)
      !
      DO ispin = 1, nspins
         !
         DO idir3 = 1, 3
            ! set to zero for the calculation of the shift
            CALL pw_zero(shift_pw_gspace(idir3, ispin)%pw)
         ENDDO
         DO idir = 1, 3
            CALL qs_rho_get(current_env%jrho1_set(idir)%rho, rho_g=jrho1_g)
            rho_gspace => jrho1_g(ispin)
            ! Field gradient
            ! loop over the Gvec  components: x,y,z
            DO idir2 = 1, 3
               IF (idir /= idir2) THEN
                  ! in reciprocal space multiply (G_idir2(i)/G(i)^2)J_(idir)(G(i))
                  CALL mult_G_ov_G2_grid(cell, auxbas_pw_pool, rho_gspace, &
                                         pw_gspace_work, idir2, 0.0_dp)
                  !
                  ! scale and add to the correct component of the shift column
                  CALL set_vecp_rev(idir, idir2, idir3)
                  scale_fac = fac_vecp(idir3, idir2, idir)
                  CALL pw_scale(pw_gspace_work%pw, scale_fac)
                  CALL pw_axpy(pw_gspace_work%pw, shift_pw_gspace(idir3, ispin)%pw)
               ENDIF
            ENDDO
            !
         ENDDO ! idir
      ENDDO ! ispin

      ! Store the total soft induced magnetic field (corrected for sic)
      IF (dft_control%nspins == 2) THEN
         DO idir = 1, 3
            CALL qs_rho_get(epr_env%bind_set(idir, iB)%rho, rho_r=epr_rho_r)
            CALL pw_transfer(shift_pw_gspace(idir, 2)%pw, epr_rho_r(1)%pw)
         ENDDO
      END IF
      !
      ! Dellocate grids for the calculation of jrho and the shift
      CALL pw_pool_give_back_pw(auxbas_pw_pool, pw_gspace_work%pw)
      DO ispin = 1, dft_control%nspins
         DO idir = 1, 3
            CALL pw_pool_give_back_pw(auxbas_pw_pool, shift_pw_gspace(idir, ispin)%pw)
         ENDDO
      ENDDO
      DEALLOCATE (shift_pw_gspace)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, shift_pw_rspace%pw)
      !
      ! Finalize
      CALL timestop(handle)
      !
   END SUBROUTINE epr_ind_magnetic_field

END MODULE qs_linres_epr_ownutils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief given the response wavefunctions obtained by the application
!>      of the (rxp), p, and ((dk-dl)xp) operators,
!>      here the current density vector (jx, jy, jz)
!>      is computed for the 3 directions of the magnetic field (Bx, By, Bz)
!> \par History
!>      created 02-2006 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_atom_current
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: current_gauge_atom,&
                                              current_gauge_r,&
                                              current_gauge_r_and_step_func
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: indso,&
                                              nsoset
   USE particle_types,                  ONLY: particle_type
   USE paw_proj_set_types,              ONLY: get_paw_proj_set,&
                                              paw_proj_set_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: get_none0_cg_list,&
                                              harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_linres_op,                    ONLY: fac_vecp,&
                                              set_vecp,&
                                              set_vecp_rev
   USE qs_linres_types,                 ONLY: allocate_jrho_atom_rad,&
                                              allocate_jrho_coeff,&
                                              current_env_type,&
                                              get_current_env,&
                                              jrho_atom_type,&
                                              set2zero_jrho_atom_rad
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_oce_methods,                  ONLY: proj_blk
   USE qs_oce_types,                    ONLY: oce_matrix_type
   USE qs_rho_atom_types,               ONLY: rho_atom_coeff
   USE sap_kind_types,                  ONLY: alist_pre_align_blk,&
                                              alist_type,&
                                              get_alist
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Public subroutines ***
   PUBLIC :: calculate_jrho_atom_rad, calculate_jrho_atom, calculate_jrho_atom_coeff

   ! *** Global parameters (only in this module)
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_atom_current'

CONTAINS

! **************************************************************************************************
!> \brief Calculate the expansion coefficients for the atomic terms
!>      of the current densitiy in GAPW
!> \param qs_env ...
!> \param current_env ...
!> \param mat_d0 ...
!> \param mat_jp ...
!> \param mat_jp_rii ...
!> \param mat_jp_riii ...
!> \param iB ...
!> \param idir ...
!> \par History
!>      07.2006 created [MI]
!>      02.2009 using new setup of projector-basis overlap [jgh]
!> \author MI
! **************************************************************************************************
   SUBROUTINE calculate_jrho_atom_coeff(qs_env, current_env, mat_d0, mat_jp, mat_jp_rii, &
                                        mat_jp_riii, iB, idir)
      !
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(current_env_type)                             :: current_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_d0, mat_jp, mat_jp_rii, mat_jp_riii
      INTEGER, INTENT(IN)                                :: iB, idir

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_jrho_atom_coeff', &
         routineP = moduleN//':'//routineN

      INTEGER :: bo(2), handle, iac, iat, iatom, ibc, idir2, ii, iii, ikind, ispin, jatom, jkind, &
         kac, katom, kbc, kkind, max_gau, max_nsgf, mepos, n_cont_a, n_cont_b, nat, natom, nkind, &
         nsatbas, nsgfa, nsgfb, nso, nsoctot, nspins, num_pe, output_unit
      INTEGER, DIMENSION(3)                              :: cell_b
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, list_a, list_b
      LOGICAL                                            :: den_found, dista, distab, distb, &
                                                            is_not_associated, paw_atom, &
                                                            sgf_soft_only_a, sgf_soft_only_b
      REAL(dp)                                           :: eps_cpc, hard_radius_c, jmax, nbr_dbl, &
                                                            rab(3), rbc(3)
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: a_matrix, b_matrix, c_matrix, d_matrix
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: CPC, PC1
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: C_coeff_hh_a, C_coeff_hh_b, &
                                                            C_coeff_ss_a, C_coeff_ss_b, r_coef_h, &
                                                            r_coef_s, tmp_coeff, zero_coeff
      TYPE(alist_type), POINTER                          :: alist_ac, alist_bc
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_a, mat_b, mat_c, mat_d
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b, orb_basis_set
      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_all
      TYPE(oce_matrix_type), POINTER                     :: oce
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: a_block, b_block, c_block, d_block, &
                                                            jp2_RARnu, jp_RARnu

      CALL timeset(routineN, handle)
      !
      NULLIFY (atom_list, atomic_kind_set, qs_kind_set, dft_control, orb_basis_set, sab_all, &
               jrho1_atom_set, oce, a_block, b_block, c_block, &
               d_block, jp_RARnu, jp2_RARnu, C_coeff_hh_a, C_coeff_hh_b, C_coeff_ss_a, &
               C_coeff_ss_b, para_env, zero_coeff, tmp_coeff)
      !
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control, &
                      oce=oce, &
                      sab_all=sab_all, &
                      para_env=para_env)

      !
      IF(.NOT.(ASSOCIATED(oce)))CALL cp__a("qs_linres_atom_current.F",158)
      !
      CALL get_current_env(current_env=current_env, &
                           jrho1_atom_set=jrho1_atom_set)
      !
      IF(.NOT.(ASSOCIATED(jrho1_atom_set)))CALL cp__a("qs_linres_atom_current.F",163)
      !
      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxsgf=max_nsgf, &
                           maxgtops=max_gau)

      eps_cpc = dft_control%qs_control%gapw_control%eps_cpc

      idir2 = 1
      IF (idir .NE. iB) THEN
         CALL set_vecp_rev(idir, iB, idir2)
      ENDIF
      CALL set_vecp(iB, ii, iii)
      !
      !
      ! Set pointers for the different gauge
      mat_a => mat_d0
      mat_b => mat_jp
      mat_c => mat_jp_rii
      mat_d => mat_jp_riii
      !
      ! Density-like matrices
      nkind = SIZE(qs_kind_set)
      natom = SIZE(jrho1_atom_set)
      nspins = dft_control%nspins
      !
      ! Allocate some arrays
      ALLOCATE (a_matrix(max_nsgf, max_nsgf), b_matrix(max_nsgf, max_nsgf), &
                c_matrix(max_nsgf, max_nsgf), d_matrix(max_nsgf, max_nsgf), &
                a_block(nspins), b_block(nspins), c_block(nspins), d_block(nspins), &
                jp_RARnu(nspins), jp2_RARnu(nspins), PC1(max_nsgf*max_gau), &
                CPC(max_gau*max_gau))
      !
      ! Reset CJC coefficients and local density arrays
      DO ikind = 1, nkind
         NULLIFY (atom_list)
         CALL get_atomic_kind(atomic_kind_set(ikind), &
                              atom_list=atom_list, &
                              natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom)
         !
         ! Quick cycle if needed.
         IF (.NOT. paw_atom) CYCLE
         !
         ! Initialize the density matrix-like arrays.
         DO iat = 1, nat
            iatom = atom_list(iat)
            DO ispin = 1, nspins
               IF (ASSOCIATED(jrho1_atom_set(iatom)%cjc0_h(1)%r_coef)) THEN
                  jrho1_atom_set(iatom)%cjc0_h(ispin)%r_coef = 0.0_dp
                  jrho1_atom_set(iatom)%cjc0_s(ispin)%r_coef = 0.0_dp
                  jrho1_atom_set(iatom)%cjc_h(ispin)%r_coef = 0.0_dp
                  jrho1_atom_set(iatom)%cjc_s(ispin)%r_coef = 0.0_dp
                  jrho1_atom_set(iatom)%cjc_ii_h(ispin)%r_coef = 0.0_dp
                  jrho1_atom_set(iatom)%cjc_ii_s(ispin)%r_coef = 0.0_dp
                  jrho1_atom_set(iatom)%cjc_iii_h(ispin)%r_coef = 0.0_dp
                  jrho1_atom_set(iatom)%cjc_iii_s(ispin)%r_coef = 0.0_dp
               ENDIF
            ENDDO ! ispin
         ENDDO ! iat
      ENDDO ! ikind
      !
      ! Three centers
      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_all)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                iatom=iatom, jatom=jatom, cell=cell_b, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         nsgfa = basis_set_a%nsgf
         nsgfb = basis_set_b%nsgf
         DO ispin = 1, nspins
            NULLIFY (jp_RARnu(ispin)%r_coef, jp2_RARnu(ispin)%r_coef)
            ALLOCATE (jp_RARnu(ispin)%r_coef(nsgfa, nsgfb), &
                      jp2_RARnu(ispin)%r_coef(nsgfa, nsgfb))
         ENDDO

         ! Take the block \mu\nu of jpab, jpab_ii and jpab_iii
         jmax = 0._dp
         DO ispin = 1, nspins
            NULLIFY (a_block(ispin)%r_coef)
            NULLIFY (b_block(ispin)%r_coef)
            NULLIFY (c_block(ispin)%r_coef)
            NULLIFY (d_block(ispin)%r_coef)
            CALL cp_dbcsr_get_block_p(matrix=mat_a(ispin)%matrix, &
                                      row=iatom, col=jatom, block=a_block(ispin)%r_coef, &
                                      found=den_found)
            jmax = jmax+MAXVAL(ABS(a_block(ispin)%r_coef))
            CALL cp_dbcsr_get_block_p(matrix=mat_b(ispin)%matrix, &
                                      row=iatom, col=jatom, block=b_block(ispin)%r_coef, &
                                      found=den_found)
            jmax = jmax+MAXVAL(ABS(b_block(ispin)%r_coef))
            CALL cp_dbcsr_get_block_p(matrix=mat_c(ispin)%matrix, &
                                      row=iatom, col=jatom, block=c_block(ispin)%r_coef, &
                                      found=den_found)
            jmax = jmax+MAXVAL(ABS(c_block(ispin)%r_coef))
            CALL cp_dbcsr_get_block_p(matrix=mat_d(ispin)%matrix, &
                                      row=iatom, col=jatom, block=d_block(ispin)%r_coef, &
                                      found=den_found)
            jmax = jmax+MAXVAL(ABS(d_block(ispin)%r_coef))
         ENDDO
         !
         ! Loop over atoms
         DO kkind = 1, nkind
            NULLIFY (paw_proj)
            CALL get_qs_kind(qs_kind_set(kkind), &
                             basis_set=orb_basis_set, &
                             hard_radius=hard_radius_c, &
                             paw_proj_set=paw_proj, &
                             paw_atom=paw_atom)
            !
            ! Quick cycle if needed.
            IF (.NOT. paw_atom) CYCLE

            CALL get_paw_proj_set(paw_proj_set=paw_proj, nsatbas=nsatbas)
            nsoctot = nsatbas

            iac = ikind+nkind*(kkind-1)
            ibc = jkind+nkind*(kkind-1)

            IF (.NOT. ASSOCIATED(oce%intac(iac)%alist)) CYCLE
            IF (.NOT. ASSOCIATED(oce%intac(ibc)%alist)) CYCLE
            CALL get_alist(oce%intac(iac), alist_ac, iatom)
            CALL get_alist(oce%intac(ibc), alist_bc, jatom)
            IF (.NOT. ASSOCIATED(alist_ac)) CYCLE
            IF (.NOT. ASSOCIATED(alist_bc)) CYCLE

            DO kac = 1, alist_ac%nclist
               DO kbc = 1, alist_bc%nclist
                  IF (alist_ac%clist(kac)%catom /= alist_bc%clist(kbc)%catom) CYCLE
                  IF (ALL(cell_b+alist_bc%clist(kbc)%cell-alist_ac%clist(kac)%cell == 0)) THEN
                     IF (jmax*alist_bc%clist(kbc)%maxac*alist_ac%clist(kac)%maxac < eps_cpc) CYCLE
                     !
                     n_cont_a = alist_ac%clist(kac)%nsgf_cnt
                     n_cont_b = alist_bc%clist(kbc)%nsgf_cnt
                     sgf_soft_only_a = alist_ac%clist(kac)%sgf_soft_only
                     sgf_soft_only_b = alist_bc%clist(kbc)%sgf_soft_only
                     IF (n_cont_a .EQ. 0 .OR. n_cont_b .EQ. 0) CYCLE
                     !
                     ! thanks to the linearity of the response, we
                     ! can avoid computing soft-soft interations.
                     ! those terms are already included in the
                     ! regular grid.
                     IF (sgf_soft_only_a .AND. sgf_soft_only_b) CYCLE
                     !
                     list_a => alist_ac%clist(kac)%sgf_list
                     list_b => alist_bc%clist(kbc)%sgf_list
                     !
                     katom = alist_ac%clist(kac)%catom

                     IF (.NOT. ASSOCIATED(jrho1_atom_set(katom)%cjc0_h(1)%r_coef)) THEN
                        CALL allocate_jrho_coeff(jrho1_atom_set, katom, nsoctot)
                     ENDIF
                     !
                     ! Compute the modified Qai matrix as
                     ! mQai_\mu\nu = Qai_\mu\nu - Qbi_\mu\nu * (R_A-R_\nu)_ii
                     !             + Qci_\mu\nu * (R_A-R_\nu)_iii
                     rbc = alist_bc%clist(kbc)%rac
                     DO ispin = 1, nspins
                        CALL DCOPY(nsgfa*nsgfb, b_block(ispin)%r_coef(1, 1), 1, &
                                   jp_RARnu(ispin)%r_coef(1, 1), 1)
                        CALL DAXPY(nsgfa*nsgfb, -rbc(ii), d_block(ispin)%r_coef(1, 1), 1, &
                                   jp_RARnu(ispin)%r_coef(1, 1), 1)
                        CALL DAXPY(nsgfa*nsgfb, rbc(iii), c_block(ispin)%r_coef(1, 1), 1, &
                                   jp_RARnu(ispin)%r_coef(1, 1), 1)
                     ENDDO
                     !
                     ! Get the d_A's for the hard and soft densities.
                     IF (iatom == katom .AND. ALL(alist_ac%clist(kac)%cell == 0)) THEN
                        C_coeff_hh_a => alist_ac%clist(kac)%achint(:, :, 1)
                        C_coeff_ss_a => alist_ac%clist(kac)%acint(:, :, 1)
                        dista = .FALSE.
                     ELSE
                        C_coeff_hh_a => alist_ac%clist(kac)%acint(:, :, 1)
                        C_coeff_ss_a => alist_ac%clist(kac)%acint(:, :, 1)
                        dista = .TRUE.
                     END IF
                     ! Get the d_B's for the hard and soft densities.
                     IF (jatom == katom .AND. ALL(alist_bc%clist(kbc)%cell == 0)) THEN
                        C_coeff_hh_b => alist_bc%clist(kbc)%achint(:, :, 1)
                        C_coeff_ss_b => alist_bc%clist(kbc)%acint(:, :, 1)
                        distb = .FALSE.
                     ELSE
                        C_coeff_hh_b => alist_bc%clist(kbc)%acint(:, :, 1)
                        C_coeff_ss_b => alist_bc%clist(kbc)%acint(:, :, 1)
                        distb = .TRUE.
                     END IF

                     distab = dista .AND. distb

                     nso = nsoctot
                     !
                     DO ispin = 1, nspins
                        !
                        ! align the blocks
                        CALL alist_pre_align_blk(a_block(ispin)%r_coef, SIZE(a_block(ispin)%r_coef, 1), &
                                                 a_matrix, SIZE(a_matrix, 1), list_a, n_cont_a, list_b, n_cont_b)

                        CALL alist_pre_align_blk(jp_RARnu(ispin)%r_coef, SIZE(jp_RARnu(ispin)%r_coef, 1), &
                                                 b_matrix, SIZE(b_matrix, 1), list_a, n_cont_a, list_b, n_cont_b)

                        CALL alist_pre_align_blk(c_block(ispin)%r_coef, SIZE(c_block(ispin)%r_coef, 1), &
                                                 c_matrix, SIZE(c_matrix, 1), list_a, n_cont_a, list_b, n_cont_b)

                        CALL alist_pre_align_blk(d_block(ispin)%r_coef, SIZE(d_block(ispin)%r_coef, 1), &
                                                 d_matrix, SIZE(d_matrix, 1), list_a, n_cont_a, list_b, n_cont_b)
                        !------------------------------------------------------------------
                        ! P_\alpha\alpha'
                        r_coef_h => jrho1_atom_set(katom)%cjc0_h(ispin)%r_coef
                        r_coef_s => jrho1_atom_set(katom)%cjc0_s(ispin)%r_coef
                        CALL proj_blk(C_coeff_hh_a, C_coeff_ss_a, n_cont_a, &
                                      C_coeff_hh_b, C_coeff_ss_b, n_cont_b, &
                                      a_matrix, max_nsgf, r_coef_h, r_coef_s, nso, &
                                      PC1, CPC, 1.0_dp, distab)
                        !------------------------------------------------------------------
                        ! mQai_\alpha\alpha'
                        r_coef_h => jrho1_atom_set(katom)%cjc_h(ispin)%r_coef
                        r_coef_s => jrho1_atom_set(katom)%cjc_s(ispin)%r_coef
                        CALL proj_blk(C_coeff_hh_a, C_coeff_ss_a, n_cont_a, &
                                      C_coeff_hh_b, C_coeff_ss_b, n_cont_b, &
                                      b_matrix, max_nsgf, r_coef_h, r_coef_s, nso, &
                                      PC1, CPC, 1.0_dp, distab)
                        !------------------------------------------------------------------
                        ! Qci_\alpha\alpha'
                        r_coef_h => jrho1_atom_set(katom)%cjc_ii_h(ispin)%r_coef
                        r_coef_s => jrho1_atom_set(katom)%cjc_ii_s(ispin)%r_coef
                        CALL proj_blk(C_coeff_hh_a, C_coeff_ss_a, n_cont_a, &
                                      C_coeff_hh_b, C_coeff_ss_b, n_cont_b, &
                                      c_matrix, max_nsgf, r_coef_h, r_coef_s, nso, &
                                      PC1, CPC, 1.0_dp, distab)
                        !------------------------------------------------------------------
                        ! Qbi_\alpha\alpha'
                        r_coef_h => jrho1_atom_set(katom)%cjc_iii_h(ispin)%r_coef
                        r_coef_s => jrho1_atom_set(katom)%cjc_iii_s(ispin)%r_coef
                        CALL proj_blk(C_coeff_hh_a, C_coeff_ss_a, n_cont_a, &
                                      C_coeff_hh_b, C_coeff_ss_b, n_cont_b, &
                                      d_matrix, max_nsgf, r_coef_h, r_coef_s, nso, &
                                      PC1, CPC, 1.0_dp, distab)
                        !------------------------------------------------------------------
                     ENDDO ! ispin
                     !
                     EXIT !search loop over jatom-katom list
                  END IF
               END DO
            END DO

         ENDDO ! kkind
         DO ispin = 1, nspins
            DEALLOCATE (jp_RARnu(ispin)%r_coef, jp2_RARnu(ispin)%r_coef)
         ENDDO
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (basis_set_list)
      !
      ! parallel sum up
      nbr_dbl = 0.0_dp
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), &
                              atom_list=atom_list, &
                              natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          basis_set=orb_basis_set, &
                          paw_proj_set=paw_proj, &
                          paw_atom=paw_atom)

         IF (.NOT. paw_atom) CYCLE

         CALL get_paw_proj_set(paw_proj_set=paw_proj, nsatbas=nsatbas)
         nsoctot = nsatbas
         !
         num_pe = para_env%num_pe
         mepos = para_env%mepos
         bo = get_limit(nat, num_pe, mepos)
         !
         ALLOCATE (zero_coeff(nsoctot, nsoctot))
         DO iat = 1, nat
            iatom = atom_list(iat)
            is_not_associated = .NOT. ASSOCIATED(jrho1_atom_set(iatom)%cjc0_h(1)%r_coef)
            !
            IF (iat .GE. bo(1) .AND. iat .LE. bo(2) .AND. is_not_associated) THEN
               CALL allocate_jrho_coeff(jrho1_atom_set, iatom, nsoctot)
            ENDIF
            !
            DO ispin = 1, nspins
               !
               tmp_coeff => jrho1_atom_set(iatom)%cjc0_h(ispin)%r_coef
               IF (is_not_associated) THEN
                  zero_coeff = 0.0_dp; tmp_coeff => zero_coeff
               ENDIF
               CALL mp_sum(tmp_coeff, para_env%group)
               !
               tmp_coeff => jrho1_atom_set(iatom)%cjc0_s(ispin)%r_coef
               IF (is_not_associated) THEN
                  zero_coeff = 0.0_dp; tmp_coeff => zero_coeff
               ENDIF
               CALL mp_sum(tmp_coeff, para_env%group)
               !
               tmp_coeff => jrho1_atom_set(iatom)%cjc_h(ispin)%r_coef
               IF (is_not_associated) THEN
                  zero_coeff = 0.0_dp; tmp_coeff => zero_coeff
               ENDIF
               !
               CALL mp_sum(tmp_coeff, para_env%group)
               tmp_coeff => jrho1_atom_set(iatom)%cjc_s(ispin)%r_coef
               IF (is_not_associated) THEN
                  zero_coeff = 0.0_dp; tmp_coeff => zero_coeff
               ENDIF
               CALL mp_sum(tmp_coeff, para_env%group)
               !
               tmp_coeff => jrho1_atom_set(iatom)%cjc_ii_h(ispin)%r_coef
               IF (is_not_associated) THEN
                  zero_coeff = 0.0_dp; tmp_coeff => zero_coeff
               ENDIF
               CALL mp_sum(tmp_coeff, para_env%group)
               !
               tmp_coeff => jrho1_atom_set(iatom)%cjc_ii_s(ispin)%r_coef
               IF (is_not_associated) THEN
                  zero_coeff = 0.0_dp; tmp_coeff => zero_coeff
               ENDIF
               CALL mp_sum(tmp_coeff, para_env%group)
               !
               tmp_coeff => jrho1_atom_set(iatom)%cjc_iii_h(ispin)%r_coef
               IF (is_not_associated) THEN
                  zero_coeff = 0.0_dp; tmp_coeff => zero_coeff
               ENDIF
               CALL mp_sum(tmp_coeff, para_env%group)
               !
               tmp_coeff => jrho1_atom_set(iatom)%cjc_iii_s(ispin)%r_coef
               IF (is_not_associated) THEN
                  zero_coeff = 0.0_dp; tmp_coeff => zero_coeff
               ENDIF
               CALL mp_sum(tmp_coeff, para_env%group)
               IF (ASSOCIATED(jrho1_atom_set(iatom)%cjc0_h(ispin)%r_coef)) &
                  nbr_dbl = nbr_dbl+8.0_dp*REAL(SIZE(jrho1_atom_set(iatom)%cjc0_h(ispin)%r_coef), dp)
            ENDDO ! ispin
         ENDDO ! iat
         !
         DEALLOCATE (zero_coeff)
         !
      ENDDO ! ikind
      !
      !
      output_unit = cp_logger_get_default_io_unit()
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(A,E8.2)') 'calculate_jrho_atom_coeff: nbr_dbl=', nbr_dbl
      ENDIF
      !
      ! clean up
      DEALLOCATE (a_matrix, b_matrix, c_matrix, d_matrix, PC1, CPC, a_block, b_block, c_block, d_block, &
                  jp_RARnu, jp2_RARnu)
      !
      CALL timestop(handle)
      !
   END SUBROUTINE calculate_jrho_atom_coeff

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param current_env ...
!> \param idir ...
! **************************************************************************************************
   SUBROUTINE calculate_jrho_atom_rad(qs_env, current_env, idir)
      !
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(current_env_type)                             :: current_env
      INTEGER, INTENT(IN)                                :: idir

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_jrho_atom_rad', &
         routineP = moduleN//':'//routineN

      INTEGER :: damax_iso_not0, damax_iso_not0_local, handle, i1, i2, iat, iatom, icg, ikind, &
         ipgf1, ipgf2, ir, iset1, iset2, iso, iso1, iso1_first, iso1_last, iso2, iso2_first, &
         iso2_last, ispin, l, l_iso, llmax, lmax12, lmax_expansion, lmin12, m1s, m2s, m_iso, &
         max_iso_not0, max_iso_not0_local, max_max_iso_not0, max_nso, max_s_harm, maxl, maxlgto, &
         maxso, mepos, n1s, n2s, na, natom, natom_tot, nkind, nr, nset, nspins, num_pe, size1, &
         size2
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list, dacg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list, dacg_list
      INTEGER, DIMENSION(2)                              :: bo
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, lmax, lmin, npgf, o2nindex
      LOGICAL                                            :: paw_atom
      LOGICAL, ALLOCATABLE, DIMENSION(:, :)              :: is_set_to_0
      REAL(dp)                                           :: hard_radius
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: g1, g2, gauge_h, gauge_s
      REAL(dp), ALLOCATABLE, DIMENSION(:, :) :: cjc0_h_block, cjc0_s_block, cjc_h_block, &
         cjc_ii_h_block, cjc_ii_s_block, cjc_iii_h_block, cjc_iii_s_block, cjc_s_block, dgg_1, gg, &
         gg_lm1
      REAL(dp), DIMENSION(:, :), POINTER :: coeff, Fr_a_h, Fr_a_h_ii, Fr_a_h_iii, Fr_a_s, &
         Fr_a_s_ii, Fr_a_s_iii, Fr_b_h, Fr_b_h_ii, Fr_b_h_iii, Fr_b_s, Fr_b_s_ii, Fr_b_s_iii, &
         Fr_h, Fr_s, zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: my_CG_dxyz_asym
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      TYPE(jrho_atom_type), POINTER                      :: jrho1_atom
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

!
!

      CALL timeset(routineN, handle)
      !
      NULLIFY (atomic_kind_set, qs_kind_set, dft_control, para_env, &
               coeff, Fr_h, Fr_s, Fr_a_h, Fr_a_s, Fr_a_h_ii, Fr_a_s_ii, &
               Fr_a_h_iii, Fr_a_s_iii, Fr_b_h, Fr_b_s, Fr_b_h_ii, &
               Fr_b_s_ii, Fr_b_h_iii, Fr_b_s_iii, jrho1_atom_set, &
               jrho1_atom)
      !
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control, &
                      para_env=para_env)

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, maxlgto=maxlgto)

      !
      CALL get_current_env(current_env=current_env, &
                           jrho1_atom_set=jrho1_atom_set)
      !

      nkind = SIZE(qs_kind_set)
      nspins = dft_control%nspins
      !
      natom_tot = SIZE(jrho1_atom_set, 1)
      ALLOCATE (is_set_to_0(natom_tot, nspins))
      is_set_to_0(:, :) = .FALSE.

      !
      DO ikind = 1, nkind
         NULLIFY (atom_list, grid_atom, harmonics, orb_basis_set, &
                  lmax, lmin, npgf, zet, grid_atom, harmonics, my_CG, my_CG_dxyz_asym)
         !
         CALL get_atomic_kind(atomic_kind_set(ikind), &
                              atom_list=atom_list, &
                              natom=natom)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          grid_atom=grid_atom, &
                          paw_proj_set=paw_proj, &
                          paw_atom=paw_atom, &
                          harmonics=harmonics, &
                          hard_radius=hard_radius, &
                          basis_set=orb_basis_set)
         !
         ! Quick cycle if needed.
         IF (.NOT. paw_atom) CYCLE
         !
         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                lmax=lmax, lmin=lmin, &
                                maxl=maxl, npgf=npgf, &
                                nset=nset, zet=zet, &
                                maxso=maxso)
         CALL get_paw_proj_set(paw_proj_set=paw_proj, o2nindex=o2nindex)
         !
         nr = grid_atom%nr
         na = grid_atom%ng_sphere
         max_iso_not0 = harmonics%max_iso_not0
         damax_iso_not0 = harmonics%damax_iso_not0
         max_max_iso_not0 = MAX(max_iso_not0, damax_iso_not0)
         lmax_expansion = indso(1, max_max_iso_not0)
         max_s_harm = harmonics%max_s_harm
         llmax = harmonics%llmax
         !
         ! Distribute the atoms of this kind
         num_pe = para_env%num_pe
         mepos = para_env%mepos
         bo = get_limit(natom, num_pe, mepos)
         !
         my_CG => harmonics%my_CG
         my_CG_dxyz_asym => harmonics%my_CG_dxyz_asym
         !
         ! Allocate some arrays.
         max_nso = nsoset(maxl)
         ALLOCATE (g1(nr), g2(nr), gg(nr, 0:2*maxl), gg_lm1(nr, 0:2*maxl), dgg_1(nr, 0:2*maxl), &
                   cjc0_h_block(max_nso, max_nso), cjc0_s_block(max_nso, max_nso), &
                   cjc_h_block(max_nso, max_nso), cjc_s_block(max_nso, max_nso), &
                   cjc_ii_h_block(max_nso, max_nso), cjc_ii_s_block(max_nso, max_nso), &
                   cjc_iii_h_block(max_nso, max_nso), cjc_iii_s_block(max_nso, max_nso), &
                   cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm), &
                   dacg_list(2, nsoset(maxl)**2, max_s_harm), dacg_n_list(max_s_harm), &
                   gauge_h(nr), gauge_s(nr))
         !
         ! Compute the gauge
         SELECT CASE (current_env%gauge)
         CASE (current_gauge_r)
            ! d(r)=r
            gauge_h(1:nr) = grid_atom%rad(1:nr)
            gauge_s(1:nr) = grid_atom%rad(1:nr)
         CASE (current_gauge_r_and_step_func)
            ! step function
            gauge_h(1:nr) = 0e0_dp
            DO ir = 1, nr
               IF (grid_atom%rad(ir) .LE. hard_radius) THEN
                  gauge_s(ir) = grid_atom%rad(ir)
               ELSE
                  gauge_s(ir) = gauge_h(ir)
               ENDIF
            ENDDO
         CASE (current_gauge_atom)
            ! d(r)=A
            gauge_h(1:nr) = HUGE(0e0_dp) !0e0_dp
            gauge_s(1:nr) = HUGE(0e0_dp) !0e0_dp
         CASE DEFAULT
            CALL cp__b("qs_linres_atom_current.F",684,"Unknown gauge, try again...")
         END SELECT
         !
         !
         m1s = 0
         DO iset1 = 1, nset
            m2s = 0
            DO iset2 = 1, nset
               CALL get_none0_cg_list(my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                      max_s_harm, lmax_expansion, cg_list, cg_n_list, max_iso_not0_local)
               IF(.NOT.(max_iso_not0_local .LE. max_iso_not0))CALL cp__a("qs_linres_atom_current.F",694)
               CALL get_none0_cg_list(my_CG_dxyz_asym, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                      max_s_harm, lmax_expansion, dacg_list, dacg_n_list, damax_iso_not0_local)
               IF(.NOT.(damax_iso_not0_local .LE. damax_iso_not0))CALL cp__a("qs_linres_atom_current.F",697)

               n1s = nsoset(lmax(iset1))
               DO ipgf1 = 1, npgf(iset1)
                  iso1_first = nsoset(lmin(iset1)-1)+1+n1s*(ipgf1-1)+m1s
                  iso1_last = nsoset(lmax(iset1))+n1s*(ipgf1-1)+m1s
                  size1 = iso1_last-iso1_first+1
                  iso1_first = o2nindex(iso1_first)
                  iso1_last = o2nindex(iso1_last)
                  i1 = iso1_last-iso1_first+1
                  IF(.NOT.(size1 == i1))CALL cp__a("qs_linres_atom_current.F",707)
                  i1 = nsoset(lmin(iset1)-1)+1
                  !
                  g1(1:nr) = EXP(-zet(ipgf1, iset1)*grid_atom%rad2(1:nr))
                  !
                  n2s = nsoset(lmax(iset2))
                  DO ipgf2 = 1, npgf(iset2)
                     iso2_first = nsoset(lmin(iset2)-1)+1+n2s*(ipgf2-1)+m2s
                     iso2_last = nsoset(lmax(iset2))+n2s*(ipgf2-1)+m2s
                     size2 = iso2_last-iso2_first+1
                     iso2_first = o2nindex(iso2_first)
                     iso2_last = o2nindex(iso2_last)
                     i2 = iso2_last-iso2_first+1
                     IF(.NOT.(size2 == i2))CALL cp__a("qs_linres_atom_current.F",720)
                     i2 = nsoset(lmin(iset2)-1)+1
                     !
                     g2(1:nr) = EXP(-zet(ipgf2, iset2)*grid_atom%rad2(1:nr))
                     !
                     lmin12 = lmin(iset1)+lmin(iset2)
                     lmax12 = lmax(iset1)+lmax(iset2)
                     !
                     gg = 0.0_dp
                     gg_lm1 = 0.0_dp
                     dgg_1 = 0.0_dp
                     !
                     ! Take only the terms of expansion for L < lmax_expansion
                     IF (lmin12 .LE. lmax_expansion) THEN
                        !
                        IF (lmax12 .GT. lmax_expansion) lmax12 = lmax_expansion
                        !
                        IF (lmin12 == 0) THEN
                           gg(1:nr, lmin12) = g1(1:nr)*g2(1:nr)
                           gg_lm1(1:nr, lmin12) = 0.0_dp
                        ELSE
                           gg(1:nr, lmin12) = grid_atom%rad2l(1:nr, lmin12)*g1(1:nr)*g2(1:nr)
                           gg_lm1(1:nr, lmin12) = grid_atom%rad2l(1:nr, lmin12-1)*g1(1:nr)*g2(1:nr)
                        ENDIF
                        !
                        DO l = lmin12+1, lmax12
                           gg(1:nr, l) = grid_atom%rad(1:nr)*gg(1:nr, l-1)
                           gg_lm1(1:nr, l) = gg(1:nr, l-1)
                        ENDDO
                        !
                        DO l = lmin12, lmax12
                           dgg_1(1:nr, l) = 2.0_dp*(zet(ipgf1, iset1)-zet(ipgf2, iset2))&
                                           &              *gg(1:nr, l)*grid_atom%rad(1:nr)
                        ENDDO
                     ELSE
                        CYCLE
                     ENDIF ! lmin12
                     !
                     DO iat = bo(1), bo(2)
                        iatom = atom_list(iat)
                        !
                        DO ispin = 1, nspins
                           !------------------------------------------------------------------
                           ! P_\alpha\alpha'
                           cjc0_h_block = HUGE(1.0_dp)
                           cjc0_s_block = HUGE(1.0_dp)
                           !
                           ! Hard term
                           coeff => jrho1_atom_set(iatom)%cjc0_h(ispin)%r_coef
                           cjc0_h_block(i1:i1+size1-1, i2:i2+size2-1) = &
                              coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                           !
                           ! Soft term
                           coeff => jrho1_atom_set(iatom)%cjc0_s(ispin)%r_coef
                           cjc0_s_block(i1:i1+size1-1, i2:i2+size2-1) = &
                              coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                           !------------------------------------------------------------------
                           ! mQai_\alpha\alpha'
                           cjc_h_block = HUGE(1.0_dp)
                           cjc_s_block = HUGE(1.0_dp)
                           !
                           ! Hard term
                           coeff => jrho1_atom_set(iatom)%cjc_h(ispin)%r_coef
                           cjc_h_block(i1:i1+size1-1, i2:i2+size2-1) = &
                              coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                           !
                           ! Soft term
                           coeff => jrho1_atom_set(iatom)%cjc_s(ispin)%r_coef
                           cjc_s_block(i1:i1+size1-1, i2:i2+size2-1) = &
                              coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                           !------------------------------------------------------------------
                           ! Qci_\alpha\alpha'
                           cjc_ii_h_block = HUGE(1.0_dp)
                           cjc_ii_s_block = HUGE(1.0_dp)
                           !
                           ! Hard term
                           coeff => jrho1_atom_set(iatom)%cjc_ii_h(ispin)%r_coef
                           cjc_ii_h_block(i1:i1+size1-1, i2:i2+size2-1) = &
                              coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                           !
                           ! Soft term
                           coeff => jrho1_atom_set(iatom)%cjc_ii_s(ispin)%r_coef
                           cjc_ii_s_block(i1:i1+size1-1, i2:i2+size2-1) = &
                              coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                           !------------------------------------------------------------------
                           ! Qbi_\alpha\alpha'
                           cjc_iii_h_block = HUGE(1.0_dp)
                           cjc_iii_s_block = HUGE(1.0_dp)
                           !
                           !
                           ! Hard term
                           coeff => jrho1_atom_set(iatom)%cjc_iii_h(ispin)%r_coef
                           cjc_iii_h_block(i1:i1+size1-1, i2:i2+size2-1) = &
                              coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                           !
                           ! Soft term
                           coeff => jrho1_atom_set(iatom)%cjc_iii_s(ispin)%r_coef
                           cjc_iii_s_block(i1:i1+size1-1, i2:i2+size2-1) = &
                              coeff(iso1_first:iso1_last, iso2_first:iso2_last)
                           !------------------------------------------------------------------
                           !
                           ! Allocation radial functions
                           jrho1_atom => jrho1_atom_set(iatom)
                           IF (.NOT. ASSOCIATED(jrho1_atom%jrho_a_h(ispin)%r_coef)) THEN
                              CALL allocate_jrho_atom_rad(jrho1_atom, ispin, nr, na, &
                                                          max_max_iso_not0)
                              is_set_to_0(iatom, ispin) = .TRUE.
                           ELSE
                              IF (.NOT. is_set_to_0(iatom, ispin)) THEN
                                 CALL set2zero_jrho_atom_rad(jrho1_atom, ispin)
                                 is_set_to_0(iatom, ispin) = .TRUE.
                              ENDIF
                           ENDIF
                           !------------------------------------------------------------------
                           !
                           Fr_h => jrho1_atom%jrho_h(ispin)%r_coef
                           Fr_s => jrho1_atom%jrho_s(ispin)%r_coef
                           !------------------------------------------------------------------
                           !
                           Fr_a_h => jrho1_atom%jrho_a_h(ispin)%r_coef
                           Fr_a_s => jrho1_atom%jrho_a_s(ispin)%r_coef
                           Fr_b_h => jrho1_atom%jrho_b_h(ispin)%r_coef
                           Fr_b_s => jrho1_atom%jrho_b_s(ispin)%r_coef
                           !------------------------------------------------------------------
                           !
                           Fr_a_h_ii => jrho1_atom%jrho_a_h_ii(ispin)%r_coef
                           Fr_a_s_ii => jrho1_atom%jrho_a_s_ii(ispin)%r_coef
                           Fr_b_h_ii => jrho1_atom%jrho_b_h_ii(ispin)%r_coef
                           Fr_b_s_ii => jrho1_atom%jrho_b_s_ii(ispin)%r_coef
                           !------------------------------------------------------------------
                           !
                           Fr_a_h_iii => jrho1_atom%jrho_a_h_iii(ispin)%r_coef
                           Fr_a_s_iii => jrho1_atom%jrho_a_s_iii(ispin)%r_coef
                           Fr_b_h_iii => jrho1_atom%jrho_b_h_iii(ispin)%r_coef
                           Fr_b_s_iii => jrho1_atom%jrho_b_s_iii(ispin)%r_coef
                           !------------------------------------------------------------------
                           !
                           DO iso = 1, max_iso_not0_local
                              l_iso = indso(1, iso) ! not needed
                              m_iso = indso(2, iso) ! not needed
                              !
                              DO icg = 1, cg_n_list(iso)
                                 !
                                 iso1 = cg_list(1, icg, iso)
                                 iso2 = cg_list(2, icg, iso)
                                 !
                                 IF (.NOT. (iso2 > 0 .AND. iso1 > 0)) THEN
                                    WRITE (*, *) 'iso1=', iso1, ' iso2=', iso2, ' iso=', iso, ' icg=', icg
                                    WRITE (*, *) '.... will stop!'
                                 ENDIF
                                 IF(.NOT.(iso2 > 0 .AND. iso1 > 0))CALL cp__a("qs_linres_atom_current.F",870)
                                 !
                                 l = indso(1, iso1)+indso(1, iso2)
                                 IF (l .GT. lmax_expansion .OR. l .LT. .0) THEN
                                    WRITE (*, *) 'calculate_jrho_atom_rad: 1 l', l
                                    WRITE (*, *) 'calculate_jrho_atom_rad: 1 lmax_expansion', lmax_expansion
                                    WRITE (*, *) '.... will stop!'
                                 ENDIF
                                 IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_linres_atom_current.F",878)
                                 !------------------------------------------------------------------
                                 ! P0
                                 !
                                 IF (current_env%gauge .EQ. current_gauge_atom) THEN
                                    ! Hard term
                                    Fr_h(1:nr, iso) = Fr_h(1:nr, iso)+ &
                                                      gg(1:nr, l)*cjc0_h_block(iso1, iso2)* &
                                                      my_CG(iso1, iso2, iso)
                                    ! Soft term
                                    Fr_s(1:nr, iso) = Fr_s(1:nr, iso)+ &
                                                      gg(1:nr, l)*cjc0_s_block(iso1, iso2)* &
                                                      my_CG(iso1, iso2, iso)
                                 ELSE
                                    ! Hard term
                                    Fr_h(1:nr, iso) = Fr_h(1:nr, iso)+ &
                                                      gg(1:nr, l)*cjc0_h_block(iso1, iso2)* &
                                                      my_CG(iso1, iso2, iso)*(grid_atom%rad(1:nr)-gauge_h(1:nr))
                                    ! Soft term
                                    Fr_s(1:nr, iso) = Fr_s(1:nr, iso)+ &
                                                      gg(1:nr, l)*cjc0_s_block(iso1, iso2)* &
                                                      my_CG(iso1, iso2, iso)*(grid_atom%rad(1:nr)-gauge_s(1:nr))
                                 ENDIF
                                 !------------------------------------------------------------------
                                 ! Rai
                                 !
                                 ! Hard term
                                 Fr_a_h(1:nr, iso) = Fr_a_h(1:nr, iso)+ &
                                                     dgg_1(1:nr, l)*cjc_h_block(iso1, iso2)* &
                                                     my_CG(iso1, iso2, iso)
                                 !
                                 ! Soft term
                                 Fr_a_s(1:nr, iso) = Fr_a_s(1:nr, iso)+ &
                                                     dgg_1(1:nr, l)*cjc_s_block(iso1, iso2)* &
                                                     my_CG(iso1, iso2, iso)
                                 !------------------------------------------------------------------
                                 ! Rci
                                 !
                                 IF (current_env%gauge .EQ. current_gauge_atom) THEN
                                    ! Hard term
                                    Fr_a_h_ii(1:nr, iso) = Fr_a_h_ii(1:nr, iso)+ &
                                                           dgg_1(1:nr, l)* &
                                                           cjc_ii_h_block(iso1, iso2)* &
                                                           my_CG(iso1, iso2, iso)
                                    ! Soft term
                                    Fr_a_s_ii(1:nr, iso) = Fr_a_s_ii(1:nr, iso)+ &
                                                           dgg_1(1:nr, l)* &
                                                           cjc_ii_s_block(iso1, iso2)* &
                                                           my_CG(iso1, iso2, iso)
                                 ELSE
                                    ! Hard term
                                    Fr_a_h_ii(1:nr, iso) = Fr_a_h_ii(1:nr, iso)+ &
                                                           dgg_1(1:nr, l)*gauge_h(1:nr)* &
                                                           cjc_ii_h_block(iso1, iso2)* &
                                                           my_CG(iso1, iso2, iso)
                                    ! Soft term
                                    Fr_a_s_ii(1:nr, iso) = Fr_a_s_ii(1:nr, iso)+ &
                                                           dgg_1(1:nr, l)*gauge_s(1:nr)* &
                                                           cjc_ii_s_block(iso1, iso2)* &
                                                           my_CG(iso1, iso2, iso)
                                 ENDIF
                                 !------------------------------------------------------------------
                                 ! Rbi
                                 !
                                 IF (current_env%gauge .EQ. current_gauge_atom) THEN
                                    ! Hard term
                                    Fr_a_h_iii(1:nr, iso) = Fr_a_h_iii(1:nr, iso)+ &
                                                            dgg_1(1:nr, l)* &
                                                            cjc_iii_h_block(iso1, iso2)* &
                                                            my_CG(iso1, iso2, iso)
                                    ! Soft term
                                    Fr_a_s_iii(1:nr, iso) = Fr_a_s_iii(1:nr, iso)+ &
                                                            dgg_1(1:nr, l)* &
                                                            cjc_iii_s_block(iso1, iso2)* &
                                                            my_CG(iso1, iso2, iso)
                                 ELSE
                                    ! Hard term
                                    Fr_a_h_iii(1:nr, iso) = Fr_a_h_iii(1:nr, iso)+ &
                                                            dgg_1(1:nr, l)*gauge_h(1:nr)* &
                                                            cjc_iii_h_block(iso1, iso2)* &
                                                            my_CG(iso1, iso2, iso)
                                    ! Soft term
                                    Fr_a_s_iii(1:nr, iso) = Fr_a_s_iii(1:nr, iso)+ &
                                                            dgg_1(1:nr, l)*gauge_s(1:nr)* &
                                                            cjc_iii_s_block(iso1, iso2)* &
                                                            my_CG(iso1, iso2, iso)
                                 ENDIF
                                 !------------------------------------------------------------------
                              ENDDO !icg
                              !
                           ENDDO ! iso
                           !
                           !
                           DO iso = 1, damax_iso_not0_local
                              l_iso = indso(1, iso) ! not needed
                              m_iso = indso(2, iso) ! not needed
                              !
                              DO icg = 1, dacg_n_list(iso)
                                 !
                                 iso1 = dacg_list(1, icg, iso)
                                 iso2 = dacg_list(2, icg, iso)
                                 !
                                 IF (.NOT. (iso2 > 0 .AND. iso1 > 0)) THEN
                                    WRITE (*, *) 'iso1=', iso1, ' iso2=', iso2, ' iso=', iso, ' icg=', icg
                                    WRITE (*, *) '.... will stop!'
                                 ENDIF
                                 IF(.NOT.(iso2 > 0 .AND. iso1 > 0))CALL cp__a("qs_linres_atom_current.F",984)
                                 !
                                 l = indso(1, iso1)+indso(1, iso2)
                                 IF (l .GT. lmax_expansion) THEN
                                    WRITE (*, *) 'calculate_jrho_atom_rad: 1 l', l
                                    WRITE (*, *) 'calculate_jrho_atom_rad: 1 lmax_expansion', lmax_expansion
                                    WRITE (*, *) '.... will stop!'
                                 ENDIF
                                 IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_linres_atom_current.F",992)
                                 !------------------------------------------------------------------
                                 ! Daij
                                 !
                                 ! Hard term
                                 Fr_b_h(1:nr, iso) = Fr_b_h(1:nr, iso)+ &
                                                     gg_lm1(1:nr, l)*cjc_h_block(iso1, iso2)* &
                                                     my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                 !
                                 ! Soft term
                                 Fr_b_s(1:nr, iso) = Fr_b_s(1:nr, iso)+ &
                                                     gg_lm1(1:nr, l)*cjc_s_block(iso1, iso2)* &
                                                     my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                 !
                                 !------------------------------------------------------------------
                                 ! Dcij
                                 !
                                 IF (current_env%gauge .EQ. current_gauge_atom) THEN
                                    ! Hard term
                                    Fr_b_h_ii(1:nr, iso) = Fr_b_h_ii(1:nr, iso)+ &
                                                           gg_lm1(1:nr, l)* &
                                                           cjc_ii_h_block(iso1, iso2)* &
                                                           my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                    ! Soft term
                                    Fr_b_s_ii(1:nr, iso) = Fr_b_s_ii(1:nr, iso)+ &
                                                           gg_lm1(1:nr, l)* &
                                                           cjc_ii_s_block(iso1, iso2)* &
                                                           my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                 ELSE
                                    ! Hard term
                                    Fr_b_h_ii(1:nr, iso) = Fr_b_h_ii(1:nr, iso)+ &
                                                           gg_lm1(1:nr, l)*gauge_h(1:nr)* &
                                                           cjc_ii_h_block(iso1, iso2)* &
                                                           my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                    ! Soft term
                                    Fr_b_s_ii(1:nr, iso) = Fr_b_s_ii(1:nr, iso)+ &
                                                           gg_lm1(1:nr, l)*gauge_s(1:nr)* &
                                                           cjc_ii_s_block(iso1, iso2)* &
                                                           my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                 ENDIF
                                 !------------------------------------------------------------------
                                 ! Dbij
                                 !
                                 IF (current_env%gauge .EQ. current_gauge_atom) THEN
                                    ! Hard term
                                    Fr_b_h_iii(1:nr, iso) = Fr_b_h_iii(1:nr, iso)+ &
                                                            gg_lm1(1:nr, l)* &
                                                            cjc_iii_h_block(iso1, iso2)* &
                                                            my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                    ! Soft term
                                    Fr_b_s_iii(1:nr, iso) = Fr_b_s_iii(1:nr, iso)+ &
                                                            gg_lm1(1:nr, l)* &
                                                            cjc_iii_s_block(iso1, iso2)* &
                                                            my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                 ELSE
                                    ! Hard term
                                    Fr_b_h_iii(1:nr, iso) = Fr_b_h_iii(1:nr, iso)+ &
                                                            gg_lm1(1:nr, l)*gauge_h(1:nr)* &
                                                            cjc_iii_h_block(iso1, iso2)* &
                                                            my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                    ! Soft term
                                    Fr_b_s_iii(1:nr, iso) = Fr_b_s_iii(1:nr, iso)+ &
                                                            gg_lm1(1:nr, l)*gauge_s(1:nr)* &
                                                            cjc_iii_s_block(iso1, iso2)* &
                                                            my_CG_dxyz_asym(idir, iso1, iso2, iso)
                                 ENDIF
                                 !------------------------------------------------------------------
                              ENDDO ! icg
                           ENDDO ! iso
                           !
                        ENDDO ! ispin
                     ENDDO ! iat
                     !
                     !------------------------------------------------------------------
                     !
                  ENDDO !ipgf2
               ENDDO ! ipgf1
               m2s = m2s+maxso
            ENDDO ! iset2
            m1s = m1s+maxso
         ENDDO ! iset1
         !
         DEALLOCATE (cjc0_h_block, cjc0_s_block, cjc_h_block, cjc_s_block, cjc_ii_h_block, cjc_ii_s_block, &
                     cjc_iii_h_block, cjc_iii_s_block, g1, g2, gg, gg_lm1, dgg_1, gauge_h, gauge_s, &
                     cg_list, cg_n_list, dacg_list, dacg_n_list)
      ENDDO ! ikind
      !
      !
      DEALLOCATE (is_set_to_0)
      !
      CALL timestop(handle)
      !
   END SUBROUTINE calculate_jrho_atom_rad

! **************************************************************************************************
!> \brief ...
!> \param jrho1_atom ...
!> \param jrho_h ...
!> \param jrho_s ...
!> \param grid_atom ...
!> \param harmonics ...
!> \param do_igaim ...
!> \param ratom ...
!> \param natm_gauge ...
!> \param iB ...
!> \param idir ...
!> \param ispin ...
! **************************************************************************************************
   SUBROUTINE calculate_jrho_atom_ang(jrho1_atom, jrho_h, jrho_s, grid_atom, &
                                      harmonics, do_igaim, ratom, natm_gauge, &
                                      iB, idir, ispin)
      !
      TYPE(jrho_atom_type), POINTER            :: jrho1_atom
      REAL(dp), DIMENSION(:, :), POINTER       :: jrho_h, jrho_s
      TYPE(grid_atom_type), POINTER            :: grid_atom
      TYPE(harmonics_atom_type), POINTER       :: harmonics
      LOGICAL, INTENT(IN)                      :: do_igaim
      INTEGER, INTENT(IN)                      :: iB, idir, ispin, natm_gauge
      REAL(dp), INTENT(IN) :: ratom(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_jrho_atom_ang', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: ia, idir2, iiB, iiiB, ir, &
                                                  iso, max_max_iso_not0, na, nr
      REAL(dp)                                 :: rad_part, scale
      REAL(dp), DIMENSION(:, :), POINTER :: a, Fr_a_h, Fr_a_h_ii, Fr_a_h_iii, &
                                            Fr_a_s, Fr_a_s_ii, Fr_a_s_iii, Fr_b_h, Fr_b_h_ii, Fr_b_h_iii, Fr_b_s, &
                                            Fr_b_s_ii, Fr_b_s_iii, Fr_h, Fr_s, slm
      REAL(dp), DIMENSION(:), POINTER :: r
      REAL(dp), DIMENSION(:, :, :), ALLOCATABLE :: g
!
!
      NULLIFY (Fr_h, Fr_s, Fr_a_h, Fr_a_s, Fr_a_h_ii, Fr_a_s_ii, Fr_a_h_iii, Fr_a_s_iii, &
               Fr_b_h, Fr_b_s, Fr_b_h_ii, Fr_b_s_ii, Fr_b_h_iii, Fr_b_s_iii, &
               a, slm)
      !
      IF(.NOT.(ASSOCIATED(jrho_h)))CALL cp__a("qs_linres_atom_current.F",1129)
      IF(.NOT.(ASSOCIATED(jrho_s)))CALL cp__a("qs_linres_atom_current.F",1130)
      IF(.NOT.(ASSOCIATED(jrho1_atom)))CALL cp__a("qs_linres_atom_current.F",1131)
      ! just to be sure...
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_h)))CALL cp__a("qs_linres_atom_current.F",1133)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_s)))CALL cp__a("qs_linres_atom_current.F",1134)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_h)))CALL cp__a("qs_linres_atom_current.F",1135)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_s)))CALL cp__a("qs_linres_atom_current.F",1136)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_h(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1137)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_s(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1138)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_h(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1139)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_s(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1140)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_h_ii)))CALL cp__a("qs_linres_atom_current.F",1141)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_s_ii)))CALL cp__a("qs_linres_atom_current.F",1142)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_h_ii)))CALL cp__a("qs_linres_atom_current.F",1143)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_s_ii)))CALL cp__a("qs_linres_atom_current.F",1144)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_h_ii(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1145)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_s_ii(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1146)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_h_ii(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1147)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_s_ii(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1148)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_h_iii)))CALL cp__a("qs_linres_atom_current.F",1149)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_s_iii)))CALL cp__a("qs_linres_atom_current.F",1150)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_h_iii)))CALL cp__a("qs_linres_atom_current.F",1151)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_s_iii)))CALL cp__a("qs_linres_atom_current.F",1152)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_h_iii(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1153)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_a_s_iii(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1154)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_h_iii(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1155)
      IF(.NOT.(ASSOCIATED(jrho1_atom%jrho_b_s_iii(ispin)%r_coef)))CALL cp__a("qs_linres_atom_current.F",1156)
      !
      !
      nr = grid_atom%nr
      na = grid_atom%ng_sphere
      max_max_iso_not0 = MAX(harmonics%max_iso_not0, harmonics%damax_iso_not0)
      ALLOCATE (g(3, nr, na))
      !------------------------------------------------------------------
      !
      Fr_h => jrho1_atom%jrho_h(ispin)%r_coef
      Fr_s => jrho1_atom%jrho_s(ispin)%r_coef
      !------------------------------------------------------------------
      !
      Fr_a_h => jrho1_atom%jrho_a_h(ispin)%r_coef !Rai
      Fr_a_s => jrho1_atom%jrho_a_s(ispin)%r_coef
      Fr_b_h => jrho1_atom%jrho_b_h(ispin)%r_coef !Daij
      Fr_b_s => jrho1_atom%jrho_b_s(ispin)%r_coef
      !------------------------------------------------------------------
      !
      Fr_a_h_ii => jrho1_atom%jrho_a_h_ii(ispin)%r_coef !Rci
      Fr_a_s_ii => jrho1_atom%jrho_a_s_ii(ispin)%r_coef
      Fr_b_h_ii => jrho1_atom%jrho_b_h_ii(ispin)%r_coef !Dcij
      Fr_b_s_ii => jrho1_atom%jrho_b_s_ii(ispin)%r_coef
      !------------------------------------------------------------------
      !
      Fr_a_h_iii => jrho1_atom%jrho_a_h_iii(ispin)%r_coef !Rbi
      Fr_a_s_iii => jrho1_atom%jrho_a_s_iii(ispin)%r_coef
      Fr_b_h_iii => jrho1_atom%jrho_b_h_iii(ispin)%r_coef !Dbij
      Fr_b_s_iii => jrho1_atom%jrho_b_s_iii(ispin)%r_coef
      !------------------------------------------------------------------
      !
      a => harmonics%a
      slm => harmonics%slm
      r => grid_atom%rad
      !
      CALL set_vecp(iB, iiB, iiiB)
      !
      scale = 0.0_dp
      idir2 = 1
      IF (idir .NE. iB) THEN
         CALL set_vecp_rev(idir, iB, idir2)
         scale = fac_vecp(idir, iB, idir2)
      ENDIF
      !
      ! Set the gauge
      CALL get_gauge()
      !
      DO ir = 1, nr
         DO iso = 1, max_max_iso_not0
            DO ia = 1, na
               IF (do_igaim) THEN
                  !------------------------------------------------------------------
                  ! Hard current density response
                  ! radial(ia,ir) = (               aj(ia) * Rai(ir,iso) + Daij
                  !                  -  aii(ia) * ( aj(ia) * Rbi(ir,iso) + Dbij )
                  !                  + aiii(ia) * ( aj(ia) * Rci(ir,iso) + Dcij )
                  !                 ) * Ylm(ia)
                  rad_part = a(idir, ia)*Fr_a_h(ir, iso)+Fr_b_h(ir, iso) &
                       &   -g(iiB, ir, ia)*(a(idir, ia)*Fr_a_h_iii(ir, iso)+Fr_b_h_iii(ir, iso))&
                       &   +g(iiiB, ir, ia)*(a(idir, ia)*Fr_a_h_ii(ir, iso)+Fr_b_h_ii(ir, iso))&
                       &   +scale*(a(idir2, ia)*r(ir)-g(idir2, ir, ia))*Fr_h(ir, iso)
                  !
                  jrho_h(ir, ia) = jrho_h(ir, ia)+rad_part*slm(ia, iso)
                  !------------------------------------------------------------------
                  ! Soft current density response
                  rad_part = a(idir, ia)*Fr_a_s(ir, iso)+Fr_b_s(ir, iso) &
                       &   -g(iiB, ir, ia)*(a(idir, ia)*Fr_a_s_iii(ir, iso)+Fr_b_s_iii(ir, iso))&
                       &   +g(iiiB, ir, ia)*(a(idir, ia)*Fr_a_s_ii(ir, iso)+Fr_b_s_ii(ir, iso))&
                       &   +scale*(a(idir2, ia)*r(ir)-g(idir2, ir, ia))*Fr_s(ir, iso)
                  !
                  jrho_s(ir, ia) = jrho_s(ir, ia)+rad_part*slm(ia, iso)
                  !------------------------------------------------------------------
               ELSE
                  !------------------------------------------------------------------
                  ! Hard current density response
                  ! radial(ia,ir) = (               aj(ia) * Rai(ir,iso) + Daij
                  !                  -  aii(ia) * ( aj(ia) * Rbi(ir,iso) + Dbij )
                  !                  + aiii(ia) * ( aj(ia) * Rci(ir,iso) + Dcij )
                  !                 ) * Ylm(ia)
                  rad_part = a(idir, ia)*Fr_a_h(ir, iso)+Fr_b_h(ir, iso) &
                       &   -a(iiB, ia)*(a(idir, ia)*Fr_a_h_iii(ir, iso)+Fr_b_h_iii(ir, iso))&
                       &   +a(iiiB, ia)*(a(idir, ia)*Fr_a_h_ii(ir, iso)+Fr_b_h_ii(ir, iso))&
                       &   +scale*a(idir2, ia)*Fr_h(ir, iso)
                  !
                  jrho_h(ir, ia) = jrho_h(ir, ia)+rad_part*slm(ia, iso)
                  !------------------------------------------------------------------
                  ! Soft current density response
                  rad_part = a(idir, ia)*Fr_a_s(ir, iso)+Fr_b_s(ir, iso) &
                       &   -a(iiB, ia)*(a(idir, ia)*Fr_a_s_iii(ir, iso)+Fr_b_s_iii(ir, iso))&
                       &   +a(iiiB, ia)*(a(idir, ia)*Fr_a_s_ii(ir, iso)+Fr_b_s_ii(ir, iso))&
                       &   +scale*a(idir2, ia)*Fr_s(ir, iso)
                  !
                  jrho_s(ir, ia) = jrho_s(ir, ia)+rad_part*slm(ia, iso)
                  !------------------------------------------------------------------
               ENDIF
            ENDDO ! ia
         ENDDO ! iso
      ENDDO ! ir
      !
      DEALLOCATE (g)
      !
   CONTAINS
      !
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE get_gauge()
      INTEGER                                            :: iatom, ixyz, jatom
      REAL(dp)                                           :: ab, pa, pb, point(3), pra(3), prb(3), &
                                                            res, tmp
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: buf

         ALLOCATE (buf(natm_gauge))
         DO ir = 1, nr
         DO ia = 1, na
            DO ixyz = 1, 3
               g(ixyz, ir, ia) = 0.0_dp
            ENDDO
            point(1) = r(ir)*a(1, ia)
            point(2) = r(ir)*a(2, ia)
            point(3) = r(ir)*a(3, ia)
            DO iatom = 1, natm_gauge
               buf(iatom) = 1.0_dp
               pra = point-ratom(:, iatom)
               pa = SQRT(pra(1)**2+pra(2)**2+pra(3)**2)
               DO jatom = 1, natm_gauge
                  IF (iatom .EQ. jatom) CYCLE
                  prb = point-ratom(:, jatom)
                  pb = SQRT(prb(1)**2+prb(2)**2+prb(3)**2)
                  ab = SQRT((pra(1)-prb(1))**2+(pra(2)-prb(2))**2+(pra(3)-prb(3))**2)
                  !
                  tmp = (pa-pb)/ab
                  tmp = 0.5_dp*(3.0_dp-tmp**2)*tmp
                  tmp = 0.5_dp*(3.0_dp-tmp**2)*tmp
                  tmp = 0.5_dp*(3.0_dp-tmp**2)*tmp
                  buf(iatom) = buf(iatom)*0.5_dp*(1.0_dp-tmp)
               ENDDO
            ENDDO
            DO ixyz = 1, 3
               res = 0.0_dp
               DO iatom = 1, natm_gauge
                  res = res+ratom(ixyz, iatom)*buf(iatom)
               ENDDO
               res = res/SUM(buf(1:natm_gauge))
               !
               g(ixyz, ir, ia) = res
            ENDDO
         ENDDO
         ENDDO
         DEALLOCATE (buf)
      END SUBROUTINE get_gauge
   END SUBROUTINE calculate_jrho_atom_ang

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
!> \param idir ...
! **************************************************************************************************
   SUBROUTINE calculate_jrho_atom(current_env, qs_env, iB, idir)
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB, idir

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_jrho_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat, iatom, ikind, ispin, jatom, &
                                                            natm_gauge, natm_tot, natom, nkind, &
                                                            nspins
      INTEGER, DIMENSION(2)                              :: bo
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: do_igaim, gapw, paw_atom
      REAL(dp)                                           :: hard_radius, r(3)
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ratom
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      TYPE(jrho_atom_type), POINTER                      :: jrho1_atom
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      NULLIFY (para_env, dft_control)
      NULLIFY (jrho1_atom_set, grid_atom, harmonics)
      NULLIFY (atomic_kind_set, qs_kind_set, atom_list)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      cell=cell, &
                      para_env=para_env)

      CALL get_current_env(current_env=current_env, &
                           jrho1_atom_set=jrho1_atom_set)

      do_igaim = .FALSE.
      IF (current_env%gauge .EQ. current_gauge_atom) do_igaim = .TRUE.

      gapw = dft_control%qs_control%gapw
      nkind = SIZE(qs_kind_set, 1)
      nspins = dft_control%nspins

      natm_tot = SIZE(particle_set)
      ALLOCATE (ratom(3, natm_tot))

      IF (gapw) THEN
         DO ikind = 1, nkind
            NULLIFY (atom_list, grid_atom, harmonics)
            CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)
            CALL get_qs_kind(qs_kind_set(ikind), &
                             grid_atom=grid_atom, &
                             harmonics=harmonics, &
                             hard_radius=hard_radius, &
                             paw_atom=paw_atom)

            IF (.NOT. paw_atom) CYCLE

            ! Distribute the atoms of this kind

            bo = get_limit(natom, para_env%num_pe, para_env%mepos)

            DO iat = bo(1), bo(2)
               iatom = atom_list(iat)
               NULLIFY (jrho1_atom)
               jrho1_atom => jrho1_atom_set(iatom)

               natm_gauge = 0
               DO jatom = 1, natm_tot
                  r(:) = pbc(particle_set(jatom)%r(:)-particle_set(iatom)%r(:), cell)
                  IF (SQRT(SUM(r(:)**2)) .LE. 2.0_dp*hard_radius) THEN
                     natm_gauge = natm_gauge+1
                     ratom(:, natm_gauge) = r(:)
                  ENDIF
               ENDDO

               DO ispin = 1, nspins
                  jrho1_atom%jrho_vec_rad_h(idir, ispin)%r_coef = 0.0_dp
                  jrho1_atom%jrho_vec_rad_s(idir, ispin)%r_coef = 0.0_dp
                  CALL calculate_jrho_atom_ang(jrho1_atom, &
                                               jrho1_atom%jrho_vec_rad_h(idir, ispin)%r_coef, &
                                               jrho1_atom%jrho_vec_rad_s(idir, ispin)%r_coef, &
                                               grid_atom, harmonics, &
                                               do_igaim, &
                                               ratom, natm_gauge, iB, idir, ispin)
               END DO !ispin
            END DO !iat
         END DO !ikind
      END IF

      DEALLOCATE (ratom)

   END SUBROUTINE calculate_jrho_atom

END MODULE qs_linres_atom_current
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief given the response wavefunctions obtained by the application
!>      of the (rxp), p, and ((dk-dl)xp) operators,
!>      here the current density vector (jx, jy, jz)
!>      is computed for the 3 directions of the magnetic field (Bx, By, Bz)
!> \par History
!>      created 02-2006 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_current
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_array_utils_i,                ONLY: cp_2d_i_p_type
   USE cp_array_utils_r,                ONLY: cp_2d_r_p_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        convert_offsets_to_sizes, cp_dbcsr_add_block_node, cp_dbcsr_allocate_matrix_set, &
        cp_dbcsr_col_block_sizes, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_deallocate_matrix, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_distribution, cp_dbcsr_finalize, &
        cp_dbcsr_get_block_p, cp_dbcsr_get_data_size, cp_dbcsr_get_matrix_type, cp_dbcsr_init, &
        cp_dbcsr_p_type, cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, &
        dbcsr_distribution_obj, dbcsr_type_antisymmetric, dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_plus_fm_fm_t,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add,&
                                              cp_fm_trace
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE cube_utils,                      ONLY: cube_info_type
   USE gaussian_gridlevels,             ONLY: gridlevel_info_type
   USE input_constants,                 ONLY: current_gauge_atom
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_get_lval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              dp,&
                                              int_8
   USE lgrid_types,                     ONLY: lgrid_type
   USE mathconstants,                   ONLY: twopi
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: ncoset
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_integrate_function,&
                                              pw_scale,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: REALDATA3D,&
                                              REALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: collocate_pgf_product_rspace
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_linres_atom_current,          ONLY: calculate_jrho_atom,&
                                              calculate_jrho_atom_coeff,&
                                              calculate_jrho_atom_rad
   USE qs_linres_op,                    ONLY: fac_vecp,&
                                              fm_scale_by_pbc_AC,&
                                              ind_m2,&
                                              set_vecp,&
                                              set_vecp_rev
   USE qs_linres_types,                 ONLY: current_env_type,&
                                              get_current_env,&
                                              realspaces_grid_p_type
   USE qs_matrix_pools,                 ONLY: qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_modify_pab_block,             ONLY: FUNC_AB,&
                                              FUNC_ADBmDAB,&
                                              FUNC_ARDBmDARB
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_operators_ao,                 ONLY: build_lin_mom_matrix,&
                                              rRc_xyz_der_ao
   USE qs_rho_types,                    ONLY: qs_rho_get
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
   USE realspace_grid_types,            ONLY: &
        realspace_grid_desc_p_type, realspace_grid_desc_type, realspace_grid_p_type, &
        realspace_grid_type, rs_grid_create, rs_grid_mult_and_add, rs_grid_release, &
        rs_grid_retain, rs_grid_zero
   USE rs_pw_interface,                 ONLY: density_rs2pw
   USE task_list_methods,               ONLY: distribute_tasks,&
                                              int2pair,&
                                              rs_distribute_matrix,&
                                              task_list_inner_loop
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Public subroutines ***
   PUBLIC :: current_build_current, current_build_chi, calculate_jrho_resp

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_current'

   TYPE box_type
      INTEGER :: n
      REAL(dp), POINTER, DIMENSION(:, :) :: r
   END TYPE box_type
   REAL(dp), DIMENSION(3, 3, 3), PARAMETER  :: Levi_Civita = RESHAPE((/ &
                                                          0.0_dp, 0.0_dp, 0.0_dp, 0.0_dp, 0.0_dp, -1.0_dp, 0.0_dp, 1.0_dp, 0.0_dp, &
                                                          0.0_dp, 0.0_dp, 1.0_dp, 0.0_dp, 0.0_dp, 0.0_dp, -1.0_dp, 0.0_dp, 0.0_dp, &
                                             0.0_dp, -1.0_dp, 0.0_dp, 1.0_dp, 0.0_dp, 0.0_dp, 0.0_dp, 0.0_dp, 0.0_dp/), (/3, 3, 3/))

CONTAINS

! **************************************************************************************************
!> \brief First calculate the density matrixes, for each component of the current
!>       they are 3 because of the r dependent terms
!>       Next it collocates on the grid to have J(r)
!>       In the GAPW case one need to collocate on the PW grid only the soft part
!>       while the rest goes on Lebedev grids
!>       The contributions to the shift and to the susceptibility will be
!>       calulated separately and added only at the end
!>       The calculation of the shift tensor is performed on the position of the atoms
!>       and on other selected points in real space summing up the contributions
!>       from the PW grid current density and the local densities
!>       Spline interpolation is used
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
!> \author MI
!> \note
!>       The susceptibility is needed to compute the G=0 term of the shift
!>       in reciprocal space. \chi_{ij} = \int (r x Jj)_i
!>       (where Jj id the current density generated by the field in direction j)
!>       To calculate the susceptibility on the PW grids it is necessary to apply
!>       the position operator yet another time.
!>       This cannot be done on directly on the full J(r) because it is not localized
!>       Therefore it is done state by state (see linres_nmr_shift)
! **************************************************************************************************
   SUBROUTINE current_build_current(current_env, qs_env, iB)
      !
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB

      CHARACTER(LEN=*), PARAMETER :: routineN = 'current_build_current', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: ext, filename, my_pos
      INTEGER                                            :: handle, idir, iiB, iiiB, ispin, istate, &
                                                            j, jstate, nao, natom, nmo, nspins, &
                                                            nstates(2), output_unit, unit_nr
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, last_sgf
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_sizes
      LOGICAL                                            :: append_cube, gapw
      REAL(dp)                                           :: dk(3), jrho_tot_G(3, 3), &
                                                            jrho_tot_R(3, 3), maxocc, scale_fac
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ddk
      REAL(dp), EXTERNAL                                 :: DDOT
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_2d_i_p_type), DIMENSION(:), POINTER        :: center_list
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: density_matrix0, density_matrix_a, &
                                                            density_matrix_ii, density_matrix_iii
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: p_psi1, psi0_order, psi1
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: psi1_D, psi1_p, psi1_rxp
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, psi_a_iB, psi_buf
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_all
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: wf_r
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: jrho1_g, jrho1_r
      TYPE(pw_p_type), POINTER                           :: jrho_gspace, jrho_rspace
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(realspace_grid_desc_type), POINTER            :: auxbas_rs_desc
      TYPE(section_vals_type), POINTER                   :: current_section

      CALL timeset(routineN, handle)
      !
      NULLIFY (jrho_rspace, jrho_gspace, logger, current_section, density_matrix0, density_matrix_a, &
               density_matrix_ii, density_matrix_iii, cell, dft_control, mos, &
               particle_set, pw_env, auxbas_rs_desc, auxbas_pw_pool, &
               para_env, center_list, mo_coeff, psi_a_iB, jrho1_r, jrho1_g, &
               psi1, p_psi1, psi1_p, psi1_D, psi1_rxp, psi0_order, sab_all, qs_kind_set)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      !
      !
      CALL get_current_env(current_env=current_env, &
                           center_list=center_list, &
                           psi1_rxp=psi1_rxp, &
                           psi1_D=psi1_D, &
                           psi1_p=psi1_p, &
                           psi0_order=psi0_order, &
                           nstates=nstates, &
                           nao=nao)
      !
      !
      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      mos=mos, &
                      mpools=mpools, &
                      pw_env=pw_env, &
                      para_env=para_env, &
                      subsys=subsys, &
                      sab_all=sab_all, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      dbcsr_dist=dbcsr_dist)

      CALL qs_subsys_get(subsys, particles=particles)

      gapw = dft_control%qs_control%gapw
      nspins = dft_control%nspins
      natom = SIZE(particle_set, 1)
      !
      ! allocate temporary arrays
      ALLOCATE (psi1(nspins), p_psi1(nspins))
      DO ispin = 1, nspins
         CALL cp_fm_create(psi1(ispin)%matrix, psi0_order(ispin)%matrix%matrix_struct)
         CALL cp_fm_create(p_psi1(ispin)%matrix, psi0_order(ispin)%matrix%matrix_struct)
         CALL cp_fm_set_all(psi1(ispin)%matrix, 0.0_dp)
         CALL cp_fm_set_all(p_psi1(ispin)%matrix, 0.0_dp)
      ENDDO
      !
      !
      CALL cp_dbcsr_allocate_matrix_set(density_matrix0, nspins)
      CALL cp_dbcsr_allocate_matrix_set(density_matrix_a, nspins)
      CALL cp_dbcsr_allocate_matrix_set(density_matrix_ii, nspins)
      CALL cp_dbcsr_allocate_matrix_set(density_matrix_iii, nspins)
      !
      ! prepare for allocation
      ALLOCATE (first_sgf(natom))
      ALLOCATE (last_sgf(natom))
      CALL get_particle_set(particle_set, qs_kind_set, &
                            first_sgf=first_sgf, &
                            last_sgf=last_sgf)
      ALLOCATE (row_blk_sizes(natom))
      CALL convert_offsets_to_sizes(first_sgf, row_blk_sizes, last_sgf)
      DEALLOCATE (first_sgf)
      DEALLOCATE (last_sgf)
      !
      !
      DO ispin = 1, nspins
         !
         !density_matrix0
         ALLOCATE (density_matrix0(ispin)%matrix)
         CALL cp_dbcsr_init(density_matrix0(ispin)%matrix)
         CALL cp_dbcsr_create(matrix=density_matrix0(ispin)%matrix, &
                              name="density_matrix0", &
                              dist=dbcsr_dist, matrix_type=dbcsr_type_no_symmetry, &
                              row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                              nze=0, mutable_work=.TRUE.)
         CALL cp_dbcsr_alloc_block_from_nbl(density_matrix0(ispin)%matrix, sab_all)
         !
         !density_matrix_a
         ALLOCATE (density_matrix_a(ispin)%matrix)
         CALL cp_dbcsr_init(density_matrix_a(ispin)%matrix)
         CALL cp_dbcsr_copy(density_matrix_a(ispin)%matrix, density_matrix0(ispin)%matrix, &
                            name="density_matrix_a")
         !
         !density_matrix_ii
         ALLOCATE (density_matrix_ii(ispin)%matrix)
         CALL cp_dbcsr_init(density_matrix_ii(ispin)%matrix)
         CALL cp_dbcsr_copy(density_matrix_ii(ispin)%matrix, density_matrix0(ispin)%matrix, &
                            name="density_matrix_ii")
         !
         !density_matrix_iii
         ALLOCATE (density_matrix_iii(ispin)%matrix)
         CALL cp_dbcsr_init(density_matrix_iii(ispin)%matrix)
         CALL cp_dbcsr_copy(density_matrix_iii(ispin)%matrix, density_matrix0(ispin)%matrix, &
                            name="density_matrix_iii")
      ENDDO
      !
      DEALLOCATE (row_blk_sizes)
      !
      !
      current_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES%CURRENT")
      !
      !
      jrho_tot_G = 0.0_dp
      jrho_tot_R = 0.0_dp
      !
      ! Lets go!
      CALL set_vecp(iB, iiB, iiiB)
      DO ispin = 1, nspins
         nmo = nstates(ispin)
         mo_coeff => psi0_order(ispin)%matrix
         !maxocc = max_occ(ispin)
         !
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, maxocc=maxocc)
         !
         !
         ! Build the first density matrix
         CALL cp_dbcsr_set(density_matrix0(ispin)%matrix, 0.0_dp)
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=density_matrix0(ispin)%matrix, &
                                    matrix_v=mo_coeff, matrix_g=mo_coeff, &
                                    ncol=nmo, alpha=maxocc)
         !
         ! Allocate buffer vectors
         ALLOCATE (ddk(3, nmo))
         !
         ! Construct the 3 density matrices for the field in direction iB
         !
         ! First the full matrix psi_a_iB
         psi_a_iB => psi1(ispin)%matrix
         psi_buf => p_psi1(ispin)%matrix
         CALL cp_fm_set_all(psi_a_iB, 0.0_dp)
         CALL cp_fm_set_all(psi_buf, 0.0_dp)
         ! psi_a_iB = - (R_\nu-dk)_ii psi1_piiiB + (R_\nu-dk)_iii psi1_piiB
         !
         ! contributions from the response psi1_p_ii and psi1_p_iii
         DO istate = 1, current_env%nbr_center(ispin)
            dk(1:3) = current_env%centers_set(ispin)%array(1:3, istate)
            !
            ! Copy the vector in the full matrix psi1
            !nstate_loc = center_list(ispin)%array(1,icenter+1)-center_list(ispin)%array(1,icenter)
            DO j = center_list(ispin)%array(1, istate), center_list(ispin)%array(1, istate+1)-1
               jstate = center_list(ispin)%array(2, j)
               CALL cp_fm_to_fm(psi1_p(ispin, iiB)%matrix, psi_a_iB, 1, jstate, jstate)
               CALL cp_fm_to_fm(psi1_p(ispin, iiiB)%matrix, psi_buf, 1, jstate, jstate)
               ddk(:, jstate) = dk(1:3)
            ENDDO
         ENDDO ! istate
         CALL fm_scale_by_pbc_AC(psi_a_iB, current_env%basisfun_center, ddk, cell, iiiB)
         CALL fm_scale_by_pbc_AC(psi_buf, current_env%basisfun_center, ddk, cell, iiB)
         CALL cp_fm_scale_and_add(-1.0_dp, psi_a_iB, 1.0_dp, psi_buf)
         !
         !psi_a_iB = psi_a_iB + psi1_rxp
         !
         ! contribution from the response psi1_rxp
         CALL cp_fm_scale_and_add(-1.0_dp, psi_a_iB, 1.0_dp, psi1_rxp(ispin, iB)%matrix)
         !
         !psi_a_iB = psi_a_iB - psi1_D
         IF (current_env%full) THEN
            !
            ! contribution from the response psi1_D
            CALL cp_fm_scale_and_add(1.0_dp, psi_a_iB, -1.0_dp, psi1_D(ispin, iB)%matrix)
         ENDIF
         !
         ! Multiply by the occupation number for the density matrix
         !
         ! Build the first density matrix
         CALL cp_dbcsr_set(density_matrix_a(ispin)%matrix, 0.0_dp)
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=density_matrix_a(ispin)%matrix, &
                                    matrix_v=mo_coeff, matrix_g=psi_a_iB, &
                                    ncol=nmo, alpha=maxocc)
         !
         ! Build the second density matrix
         CALL cp_dbcsr_set(density_matrix_iii(ispin)%matrix, 0.0_dp)
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=density_matrix_iii(ispin)%matrix, &
                                    matrix_v=mo_coeff, matrix_g=psi1_p(ispin, iiiB)%matrix, &
                                    ncol=nmo, alpha=maxocc)
         !
         ! Build the third density matrix
         CALL cp_dbcsr_set(density_matrix_ii(ispin)%matrix, 0.0_dp)
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=density_matrix_ii(ispin)%matrix, &
                                    matrix_v=mo_coeff, matrix_g=psi1_p(ispin, iiB)%matrix, &
                                    ncol=nmo, alpha=maxocc)
         DO idir = 1, 3
            !
            ! Calculate the current density on the pw grid (only soft if GAPW)
            ! idir is the cartesian component of the response current density
            ! generated by the magnetic field pointing in cartesian direction iB
            ! Use the qs_rho_type already  used for rho during the scf
            CALL qs_rho_get(current_env%jrho1_set(idir)%rho, rho_r=jrho1_r)
            CALL qs_rho_get(current_env%jrho1_set(idir)%rho, rho_g=jrho1_g)
            jrho_rspace => jrho1_r(ispin)
            jrho_gspace => jrho1_g(ispin)
            CALL pw_zero(jrho_rspace%pw)
            CALL pw_zero(jrho_gspace%pw)
            CALL calculate_jrho_resp(density_matrix0(ispin)%matrix, &
                                     density_matrix_a(ispin)%matrix, &
                                     density_matrix_ii(ispin)%matrix, &
                                     density_matrix_iii(ispin)%matrix, &
                                     iB, idir, jrho_rspace, jrho_gspace, qs_env, &
                                     current_env, gapw)

            scale_fac = cell%deth/twopi
            CALL pw_scale(jrho_rspace%pw, scale_fac)
            CALL pw_scale(jrho_gspace%pw, scale_fac)

            jrho_tot_G(idir, iB) = pw_integrate_function(jrho_gspace%pw, isign=-1)
            jrho_tot_R(idir, iB) = pw_integrate_function(jrho_rspace%pw, isign=-1)

            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T2,2(A,E24.16))') 'Integrated j_'&
                    &//ACHAR(idir+119)//ACHAR(iB+119)//'(r): G-space=', &
                     jrho_tot_G(idir, iB), ' R-space=', jrho_tot_R(idir, iB)
            ENDIF
            !
         ENDDO ! idir
         !
         ! Deallocate buffer vectors
         DEALLOCATE (ddk)
         !
      ENDDO ! ispin

      IF (gapw) THEN
         DO idir = 1, 3
            !
            ! compute the atomic response current densities on the spherical grids
            ! First the sparse matrices are multiplied by the expansion coefficients
            ! this is the equivalent of the CPC for the charge density
            CALL calculate_jrho_atom_coeff(qs_env, current_env, &
                                           density_matrix0, &
                                           density_matrix_a, &
                                           density_matrix_ii, &
                                           density_matrix_iii, &
                                           iB, idir)
            !
            ! Then the radial parts are computed on the local radial grid, atom by atom
            ! 8 functions are computed for each atom, per grid point
            ! and per LM angular momentum. The multiplication by the Clebsh-Gordon
            ! coefficients or they correspondent for the derivatives, is also done here
            CALL calculate_jrho_atom_rad(qs_env, current_env, idir)
            !
            ! The current on the atomic grids
            CALL calculate_jrho_atom(current_env, qs_env, iB, idir)
         ENDDO ! idir
      ENDIF
      !
      ! Cube files
      IF (BTEST(cp_print_key_should_output(logger%iter_info, current_section,&
           &   "PRINT%CURRENT_CUBES"), cp_p_file)) THEN
         append_cube = section_get_lval(current_section, "PRINT%CURRENT_CUBES%APPEND")
         my_pos = "REWIND"
         IF (append_cube) THEN
            my_pos = "APPEND"
         END IF
         !
         CALL pw_env_get(pw_env, auxbas_rs_desc=auxbas_rs_desc, &
                         auxbas_pw_pool=auxbas_pw_pool)
         !
         CALL pw_pool_create_pw(auxbas_pw_pool, wf_r%pw, use_data=REALDATA3D, &
                                in_space=REALSPACE)
         !
         DO idir = 1, 3
            CALL pw_zero(wf_r%pw)
            CALL qs_rho_get(current_env%jrho1_set(idir)%rho, rho_r=jrho1_r)
            DO ispin = 1, nspins
               CALL pw_axpy(jrho1_r(ispin)%pw, wf_r%pw, 1.0_dp)
            ENDDO
            !
            IF (gapw) THEN
               ! Add the local hard and soft contributions
               ! This can be done atom by atom by a spline extrapolation of the  values
               ! on the spherical grid to the grid points.
               CALL cp__b("qs_linres_current.F",494,"GAPW needs to be finalized")
            ENDIF
            filename = "jresp"
            WRITE (ext, '(a2,I1,a2,I1,a5)') "iB", iB, "_d", idir, ".cube"
            WRITE (ext, '(a2,a1,a2,a1,a5)') "iB", ACHAR(iB+119), "_d", ACHAR(idir+119), ".cube"
            unit_nr = cp_print_key_unit_nr(logger, current_section, "PRINT%CURRENT_CUBES", &
                                           extension=TRIM(ext), middle_name=TRIM(filename), &
                                           log_filename=.FALSE., file_position=my_pos)

            CALL cp_pw_to_cube(wf_r%pw, unit_nr, "RESPONSE CURRENT DENSITY ", &
                               particles=particles, &
                               stride=section_get_ivals(current_section, "PRINT%CURRENT_CUBES%STRIDE"))
            CALL cp_print_key_finished_output(unit_nr, logger, current_section,&
                 &                            "PRINT%CURRENT_CUBES")
         ENDDO
         !
         CALL pw_pool_give_back_pw(auxbas_pw_pool, wf_r%pw)
      ENDIF ! current cube
      !
      ! Integrated current response checksum
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E24.16)') 'CheckSum R-integrated j=', &
            SQRT(DDOT(9, jrho_tot_R(1, 1), 1, jrho_tot_R(1, 1), 1))
      ENDIF
      !
      !
      ! Dellocate grids for the calculation of jrho and the shift
      DO ispin = 1, nspins
         CALL cp_fm_release(psi1(ispin)%matrix)
         CALL cp_fm_release(p_psi1(ispin)%matrix)
      ENDDO
      DEALLOCATE (psi1, p_psi1)

      CALL cp_dbcsr_deallocate_matrix_set(density_matrix0)
      CALL cp_dbcsr_deallocate_matrix_set(density_matrix_a)
      CALL cp_dbcsr_deallocate_matrix_set(density_matrix_ii)
      CALL cp_dbcsr_deallocate_matrix_set(density_matrix_iii)
      !
      ! Finalize
      CALL timestop(handle)
      !
   END SUBROUTINE current_build_current

! **************************************************************************************************
!> \brief Calculation of the idir component of the response current density
!>       in the presence of a constant magnetic field in direction iB
!>       the current density is collocated on the pw grid in real space
!> \param mat_d0 ...
!> \param mat_jp ...
!> \param mat_jp_rii ...
!> \param mat_jp_riii ...
!> \param iB ...
!> \param idir ...
!> \param current_rs ...
!> \param current_gs ...
!> \param qs_env ...
!> \param current_env ...
!> \param soft_valid ...
!> \param retain_rsgrid ...
!> \note
!>       The collocate is done in three parts, one for each density matrix
!>       In all cases the density matrices and therefore the collocation
!>       are not symmetric, that means that all the pairs (ab and ba) have
!>       to be considered separately
!>
!>       mat_jp_{\mu\nu} is multiplied by
!>           f_{\mu\nu} = \phi_{\mu} (d\phi_{\nu}/dr)_{idir} -
!>                        (d\phi_{\mu}/dr)_{idir} \phi_{\nu}
!>
!>       mat_jp_rii_{\mu\nu} is multiplied by
!>           f_{\mu\nu} = \phi_{\mu} (r - R_{\nu})_{iiiB} (d\phi_{\nu}/dr)_{idir} -
!>                        (d\phi_{\mu}/dr)_{idir} (r - R_{\nu})_{iiiB} \phi_{\nu} +
!>                         \phi_{\mu} \phi_{\nu}  (last term only if iiiB=idir)
!>
!>       mat_jp_riii_{\mu\nu} is multiplied by
!>                             (be careful: change in sign with respect to previous)
!>           f_{\mu\nu} = -\phi_{\mu} (r - R_{\nu})_{iiB} (d\phi_{\nu}/dr)_{idir} +
!>                        (d\phi_{\mu}/dr)_{idir} (r - R_{\nu})_{iiB} \phi_{\nu} -
!>                         \phi_{\mu} \phi_{\nu}  (last term only if iiB=idir)
!>
!>       All the terms sum up to the same grid
! **************************************************************************************************
   SUBROUTINE calculate_jrho_resp(mat_d0, mat_jp, mat_jp_rii, mat_jp_riii, iB, idir, &
                                  current_rs, current_gs, qs_env, current_env, soft_valid, retain_rsgrid)

      TYPE(cp_dbcsr_type), POINTER                       :: mat_d0, mat_jp, mat_jp_rii, mat_jp_riii
      INTEGER, INTENT(IN)                                :: iB, idir
      TYPE(pw_p_type), INTENT(INOUT)                     :: current_rs, current_gs
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(current_env_type)                             :: current_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: soft_valid, retain_rsgrid

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_jrho_resp', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: max_tasks = 2000

      INTEGER :: bcol, brow, cindex, curr_tasks, handle, i, iatom, iatom_old, idir2, igrid_level, &
         iiB, iiiB, ikind, ikind_old, ipgf, iset, iset_old, itask, ithread, jatom, jatom_old, &
         jkind, jkind_old, jpgf, jset, jset_old, lmax_global, maxco, maxpgf, maxset, maxsgf, &
         maxsgf_set, na1, na2, natom, nb1, nb2, ncoa, ncob, ngpts, nimages, nkind, nseta, nsetb, &
         ntasks, nthread, sgfa, sgfb
      INTEGER(kind=int_8), DIMENSION(:), POINTER         :: atom_pair_recv, atom_pair_send
      INTEGER(kind=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, mylmax, &
                                                            npgfa, npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL :: atom_pair_changed, den_found, den_found_a, distributed_rs_grids, do_igaim, &
         map_consistent, my_retain_rsgrid, my_soft
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_current, my_gauge, my_rho
      REAL(KIND=dp)                                      :: eps_rho_rspace, kind_radius_a, &
                                                            kind_radius_b, Lxo2, Lyo2, Lzo2, rab2, &
                                                            scale, scale2, zetp
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rb
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: dist_ab, jp_block_a, jp_block_b, jp_block_c, &
         jp_block_d, jpab_a, jpab_b, jpab_c, jpab_d, jpblock_a, jpblock_b, jpblock_c, jpblock_d, &
         rpgfa, rpgfb, sphi_a, sphi_b, work, zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: jpabt_a, jpabt_b, jpabt_c, jpabt_d, workt
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: deltajp_a, deltajp_b, deltajp_c, &
                                                            deltajp_d
      TYPE(cp_dbcsr_type), POINTER                       :: mat_a, mat_b, mat_c, mat_d
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b, orb_basis_set
      TYPE(lgrid_type)                                   :: lgrid
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_current, rs_rho
      TYPE(realspaces_grid_p_type), DIMENSION(:), &
         POINTER                                         :: rs_gauge

      NULLIFY (qs_kind, cell, dft_control, orb_basis_set, rs_rho, &
               qs_kind_set, sab_orb, particle_set, rs_current, pw_env, &
               rs_descs, para_env, dist_ab, set_radius_a, set_radius_b, la_max, &
               la_min, lb_max, lb_min, npgfa, npgfb, nsgfa, nsgfb, rpgfa, rpgfb, &
               sphi_a, sphi_b, zeta, zetb, first_sgfa, first_sgfb, dist_ab, &
               tasks, workt, mat_a, mat_b, mat_c, mat_d, rs_gauge, mylmax)
      NULLIFY (deltajp_a, deltajp_b, deltajp_c, deltajp_d)
      NULLIFY (jp_block_a, jp_block_b, jp_block_c, jp_block_d)
      NULLIFY (jpblock_a, jpblock_b, jpblock_c, jpblock_d)
      NULLIFY (jpabt_a, jpabt_b, jpabt_c, jpabt_d)
      NULLIFY (lgrid%r)

      CALL timeset(routineN, handle)

      !
      ! Set pointers for the different gauge
      do_igaim = current_env%gauge .EQ. current_gauge_atom ! If do_igaim is False the current_env is never needed

      mat_a => mat_jp
      mat_b => mat_jp_rii
      mat_c => mat_jp_riii
      IF (do_igaim) mat_d => mat_d0

      my_retain_rsgrid = .FALSE.
      IF (PRESENT(retain_rsgrid)) my_retain_rsgrid = retain_rsgrid

      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      sab_all=sab_orb, &
                      para_env=para_env, &
                      pw_env=pw_env)

      IF (do_igaim) CALL get_current_env(current_env=current_env, rs_gauge=rs_gauge)

      ! Component of appearing in the vector product rxp, iiB and iiiB
      CALL set_vecp(iB, iiB, iiiB)
      !
      !
      scale2 = 0.0_dp
      idir2 = 1
      IF (idir .NE. iB) THEN
         CALL set_vecp_rev(idir, iB, idir2)
         scale2 = fac_vecp(idir, iB, idir2)
      ENDIF
      !
      ! *** assign from pw_env
      gridlevel_info => pw_env%gridlevel_info
      cube_info => pw_env%cube_info

      !   Check that the neighbor list with all the pairs is associated
      IF(.NOT.(ASSOCIATED(sab_orb)))CALL cp__a("qs_linres_current.F",690)
      ! *** set up the pw multi-grids
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_linres_current.F",692)
      CALL pw_env_get(pw_env, rs_descs=rs_descs, rs_grids=rs_rho)

      distributed_rs_grids = .FALSE.
      DO igrid_level = 1, gridlevel_info%ngrid_levels
         IF (.NOT. ALL(rs_descs(igrid_level)%rs_desc%perd == 1)) THEN
            distributed_rs_grids = .TRUE.
         ENDIF
      ENDDO
      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      map_consistent = dft_control%qs_control%map_consistent
      nthread = 1

      !   *** Allocate work storage ***
      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxsgf=maxsgf, &
                           maxsgf_set=maxsgf_set)

      Lxo2 = SQRT(SUM(cell%hmat(:, 1)**2))/2.0_dp
      Lyo2 = SQRT(SUM(cell%hmat(:, 2)**2))/2.0_dp
      Lzo2 = SQRT(SUM(cell%hmat(:, 3)**2))/2.0_dp

      my_soft = .FALSE.
      IF (PRESENT(soft_valid)) my_soft = soft_valid

      nkind = SIZE(qs_kind_set)

      CALL reallocate(jpabt_a, 1, maxco, 1, maxco, 0, nthread-1)
      CALL reallocate(jpabt_b, 1, maxco, 1, maxco, 0, nthread-1)
      CALL reallocate(jpabt_c, 1, maxco, 1, maxco, 0, nthread-1)
      CALL reallocate(jpabt_d, 1, maxco, 1, maxco, 0, nthread-1)
      CALL reallocate(workt, 1, maxco, 1, maxsgf_set, 0, nthread-1)
      CALL reallocate(tasks, 1, 6, 1, max_tasks)
      CALL reallocate(dist_ab, 1, 3, 1, max_tasks)

      tasks = 0
      ntasks = 0
      curr_tasks = SIZE(tasks, 2)

      !   get maximum numbers
      natom = SIZE(particle_set)
      maxset = 0
      maxpgf = 0

      ! hard code matrix index (no kpoints)
      nimages = dft_control%nimages
      IF(.NOT.(nimages == 1))CALL cp__a("qs_linres_current.F",739)
      cindex = 1

      lmax_global = 0
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)

         CALL get_qs_kind(qs_kind=qs_kind, &
                          basis_set=orb_basis_set)

         IF (.NOT. ASSOCIATED(orb_basis_set)) CYCLE

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                npgf=npgfa, nset=nseta, lmax=mylmax)

         maxset = MAX(nseta, maxset)
         maxpgf = MAX(MAXVAL(npgfa), maxpgf)
         ! Find maximum l, for collocate_pgf_product_rspace
         lmax_global = MAX(MAXVAL(mylmax), lmax_global)
      END DO

      ! ga_gb_function can be one of FUNC_AB, FUNC_ADBmDAB or FUNC_ARDBm_DARB
      ! take worst case and add 2 to lmax_global
      lmax_global = lmax_global+2

      !   *** Initialize working density matrix ***

      ! distributed rs grids require a matrix that will be changed (distribute_tasks)
      ! whereas this is not the case for replicated grids
      ALLOCATE (deltajp_a(1), deltajp_b(1), deltajp_c(1), deltajp_d(1))
      IF (distributed_rs_grids) THEN
         ALLOCATE (deltajp_a(1)%matrix, deltajp_b(1)%matrix, deltajp_c(1)%matrix)
         CALL cp_dbcsr_init(deltajp_a(1)%matrix)
         CALL cp_dbcsr_init(deltajp_b(1)%matrix)
         CALL cp_dbcsr_init(deltajp_c(1)%matrix)
         IF (do_igaim) THEN
            ALLOCATE (deltajp_d(1)%matrix)
            CALL cp_dbcsr_init(deltajp_d(1)%matrix)
         ENDIF

         CALL cp_dbcsr_create(deltajp_a(1)%matrix, ' deltajp_a ', cp_dbcsr_distribution(mat_a), cp_dbcsr_get_matrix_type(mat_a), &
                              cp_dbcsr_row_block_sizes(mat_a), cp_dbcsr_col_block_sizes(mat_a), &
                              cp_dbcsr_get_data_size(mat_a))

         CALL cp_dbcsr_create(deltajp_b(1)%matrix, ' deltajp_b ', cp_dbcsr_distribution(mat_a), cp_dbcsr_get_matrix_type(mat_a), &
                              cp_dbcsr_row_block_sizes(mat_a), cp_dbcsr_col_block_sizes(mat_a), &
                              cp_dbcsr_get_data_size(mat_a))

         CALL cp_dbcsr_create(deltajp_c(1)%matrix, ' deltajp_c ', cp_dbcsr_distribution(mat_a), cp_dbcsr_get_matrix_type(mat_a), &
                              cp_dbcsr_row_block_sizes(mat_a), cp_dbcsr_col_block_sizes(mat_a), &
                              cp_dbcsr_get_data_size(mat_a))

         IF (do_igaim) CALL cp_dbcsr_create( &
            deltajp_d(1)%matrix, ' deltajp_d ', cp_dbcsr_distribution(mat_a), &
            cp_dbcsr_get_matrix_type(mat_a), cp_dbcsr_row_block_sizes(mat_a), cp_dbcsr_col_block_sizes(mat_a), &
            cp_dbcsr_get_data_size(mat_a))
      ELSE
         deltajp_a(1)%matrix => mat_a !mat_jp
         deltajp_b(1)%matrix => mat_b !mat_jp_rii
         deltajp_c(1)%matrix => mat_c !mat_jp_riii
         IF (do_igaim) deltajp_d(1)%matrix => mat_d !mat_d0
      ENDIF

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         qs_kind => qs_kind_set(ikind)
         CALL get_qs_kind(qs_kind=qs_kind, softb=my_soft, basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, r=rab)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ra(:) = pbc(particle_set(iatom)%r, cell)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         kind_radius_a = basis_set_a%kind_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         kind_radius_b = basis_set_b%kind_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         IF (ABS(rab(1)) > Lxo2 .OR. ABS(rab(2)) > Lyo2 .OR. ABS(rab(3)) > Lzo2) THEN
            CYCLE
         END IF

         brow = iatom
         bcol = jatom

         CALL cp_dbcsr_get_block_p(matrix=mat_a, row=brow, col=bcol, &
                                   block=jp_block_a, found=den_found_a)
         CALL cp_dbcsr_get_block_p(matrix=mat_b, row=brow, col=bcol, &
                                   block=jp_block_b, found=den_found)
         CALL cp_dbcsr_get_block_p(matrix=mat_c, row=brow, col=bcol, &
                                   block=jp_block_c, found=den_found)
         IF (do_igaim) CALL cp_dbcsr_get_block_p(matrix=mat_d, row=brow, col=bcol, &
                                                 block=jp_block_d, found=den_found)

         IF (.NOT. ASSOCIATED(jp_block_a)) CYCLE

         IF (distributed_rs_grids) THEN
            NULLIFY (jpblock_a, jpblock_b, jpblock_c, jpblock_d)
            CALL cp_dbcsr_add_block_node(deltajp_a(1)%matrix, brow, bcol, jpblock_a)
            jpblock_a = jp_block_a
            CALL cp_dbcsr_add_block_node(deltajp_b(1)%matrix, brow, bcol, jpblock_b)
            jpblock_b = jp_block_b
            CALL cp_dbcsr_add_block_node(deltajp_c(1)%matrix, brow, bcol, jpblock_c)
            jpblock_c = jp_block_c
            IF (do_igaim) THEN
               CALL cp_dbcsr_add_block_node(deltajp_d(1)%matrix, brow, bcol, jpblock_d)
               jpblock_d = jp_block_d
            END IF
         ELSE
            jpblock_a => jp_block_a
            jpblock_b => jp_block_b
            jpblock_c => jp_block_c
            IF (do_igaim) jpblock_d => jp_block_d
         ENDIF

         IF (.NOT. map_consistent) THEN
            IF (ALL(100.0_dp*ABS(jpblock_a) < eps_rho_rspace) .AND. &
                ALL(100.0_dp*ABS(jpblock_b) < eps_rho_rspace) .AND. &
                ALL(100.0_dp*ABS(jpblock_c) < eps_rho_rspace)) THEN
               CYCLE
            END IF
         END IF

         CALL task_list_inner_loop(tasks, dist_ab, ntasks, curr_tasks, rs_descs, &
                                   dft_control, cube_info, gridlevel_info, cindex, &
                                   iatom, jatom, rpgfa, rpgfb, zeta, zetb, kind_radius_b, set_radius_a, set_radius_b, ra, rab, &
                                   la_max, la_min, lb_max, lb_min, npgfa, npgfb, maxpgf, maxset, natom, nimages, nseta, nsetb)

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (basis_set_list)

      IF (distributed_rs_grids) THEN
         CALL cp_dbcsr_finalize(deltajp_a(1)%matrix)
         CALL cp_dbcsr_finalize(deltajp_b(1)%matrix)
         CALL cp_dbcsr_finalize(deltajp_c(1)%matrix)
         IF (do_igaim) CALL cp_dbcsr_finalize(deltajp_d(1)%matrix)
      ENDIF

      ! sorts / redistributes the task list
      CALL distribute_tasks(rs_descs, ntasks, natom, maxset, maxpgf, nimages, &
                            tasks, dist_ab, atom_pair_send, atom_pair_recv, &
                            symmetric=.FALSE., reorder_rs_grid_ranks=.TRUE., &
                            skip_load_balance_distributed=.FALSE.)

      ALLOCATE (rs_current(gridlevel_info%ngrid_levels))

      DO igrid_level = 1, gridlevel_info%ngrid_levels
         ! Here we need to reallocate the distributed rs_grids, which may have been reordered
         ! by distribute_tasks
         IF (rs_descs(igrid_level)%rs_desc%distributed .AND. .NOT. my_retain_rsgrid) THEN
            CALL rs_grid_release(rs_rho(igrid_level)%rs_grid)
            NULLIFY (rs_rho(igrid_level)%rs_grid)
            CALL rs_grid_create(rs_rho(igrid_level)%rs_grid, rs_descs(igrid_level)%rs_desc)
         ELSE
            IF (.NOT. my_retain_rsgrid) CALL rs_grid_retain(rs_rho(igrid_level)%rs_grid)
         ENDIF
         CALL rs_grid_zero(rs_rho(igrid_level)%rs_grid)
         CALL rs_grid_create(rs_current(igrid_level)%rs_grid, rs_descs(igrid_level)%rs_desc)
         CALL rs_grid_zero(rs_current(igrid_level)%rs_grid)
      ENDDO

      ngpts = 0
      IF (nthread > 1) THEN
         DO igrid_level = 1, gridlevel_info%ngrid_levels
            ngpts = MAX(ngpts, rs_rho(igrid_level)%rs_grid%ngpts_local)
         END DO
         ngpts = ngpts*nthread
         CALL reallocate(lgrid%r, 1, ngpts)
      END IF

      !
      ! we need to build the gauge here
      IF (.NOT. current_env%gauge_init .AND. do_igaim) THEN
         CALL current_set_gauge(current_env, qs_env)
         current_env%gauge_init = .TRUE.
      ENDIF
      !
      ! for any case double check the bounds !
      IF (do_igaim) THEN
         DO igrid_level = 1, gridlevel_info%ngrid_levels
            my_rho => rs_rho(igrid_level)%rs_grid%r
            my_current => rs_current(igrid_level)%rs_grid%r
            IF (LBOUND(my_rho, 3) .NE. LBOUND(my_current, 3) .OR. &
                LBOUND(my_rho, 2) .NE. LBOUND(my_current, 2) .OR. &
                LBOUND(my_rho, 1) .NE. LBOUND(my_current, 1) .OR. &
                UBOUND(my_rho, 3) .NE. UBOUND(my_current, 3) .OR. &
                UBOUND(my_rho, 2) .NE. UBOUND(my_current, 2) .OR. &
                UBOUND(my_rho, 1) .NE. UBOUND(my_current, 1)) THEN
               WRITE (*, *) 'LBOUND(my_rho,3),LBOUND(my_current,3)', LBOUND(my_rho, 3), LBOUND(my_current, 3)
               WRITE (*, *) 'LBOUND(my_rho,2),LBOUND(my_current,2)', LBOUND(my_rho, 2), LBOUND(my_current, 2)
               WRITE (*, *) 'LBOUND(my_rho,1),LBOUND(my_current,1)', LBOUND(my_rho, 1), LBOUND(my_current, 1)
               WRITE (*, *) 'UBOUND(my_rho,3),UBOUND(my_current,3)', UBOUND(my_rho, 3), UBOUND(my_current, 3)
               WRITE (*, *) 'UBOUND(my_rho,2),UBOUND(my_current,2)', UBOUND(my_rho, 2), UBOUND(my_current, 2)
               WRITE (*, *) 'UBOUND(my_rho,1),UBOUND(my_current,1)', UBOUND(my_rho, 1), UBOUND(my_current, 1)
               CALL cp__b("qs_linres_current.F",963,"Bug")
            ENDIF

            my_gauge => rs_gauge(1)%rs(igrid_level)%rs_grid%r
            IF (LBOUND(my_rho, 3) .NE. LBOUND(my_gauge, 3) .OR. &
                LBOUND(my_rho, 2) .NE. LBOUND(my_gauge, 2) .OR. &
                LBOUND(my_rho, 1) .NE. LBOUND(my_gauge, 1) .OR. &
                UBOUND(my_rho, 3) .NE. UBOUND(my_gauge, 3) .OR. &
                UBOUND(my_rho, 2) .NE. UBOUND(my_gauge, 2) .OR. &
                UBOUND(my_rho, 1) .NE. UBOUND(my_gauge, 1)) THEN
               WRITE (*, *) 'LBOUND(my_rho,3),LBOUND(my_gauge,3)', LBOUND(my_rho, 3), LBOUND(my_gauge, 3)
               WRITE (*, *) 'LBOUND(my_rho,2),LBOUND(my_gauge,2)', LBOUND(my_rho, 2), LBOUND(my_gauge, 2)
               WRITE (*, *) 'LBOUND(my_rho,1),LBOUND(my_gauge,1)', LBOUND(my_rho, 1), LBOUND(my_gauge, 1)
               WRITE (*, *) 'UBOUND(my_rho,3),UbOUND(my_gauge,3)', UBOUND(my_rho, 3), UBOUND(my_gauge, 3)
               WRITE (*, *) 'UBOUND(my_rho,2),UBOUND(my_gauge,2)', UBOUND(my_rho, 2), UBOUND(my_gauge, 2)
               WRITE (*, *) 'UBOUND(my_rho,1),UBOUND(my_gauge,1)', UBOUND(my_rho, 1), UBOUND(my_gauge, 1)
               CALL cp__b("qs_linres_current.F",979,"Bug")
            ENDIF
         ENDDO
      ENDIF
      !
      !-------------------------------------------------------------

      IF (distributed_rs_grids) THEN
         CALL rs_distribute_matrix(rs_descs, deltajp_a, atom_pair_send, atom_pair_recv, &
                                   natom, nimages, scatter=.TRUE.)
         CALL rs_distribute_matrix(rs_descs, deltajp_b, atom_pair_send, atom_pair_recv, &
                                   natom, nimages, scatter=.TRUE.)
         CALL rs_distribute_matrix(rs_descs, deltajp_c, atom_pair_send, atom_pair_recv, &
                                   natom, nimages, scatter=.TRUE.)
         IF (do_igaim) CALL rs_distribute_matrix(rs_descs, deltajp_d, atom_pair_send, atom_pair_recv, &
                                                 natom, nimages, scatter=.TRUE.)
      ENDIF

      ithread = 0
      jpab_a => jpabt_a(:, :, ithread)
      jpab_b => jpabt_b(:, :, ithread)
      jpab_c => jpabt_c(:, :, ithread)
      IF (do_igaim) jpab_d => jpabt_d(:, :, ithread)
      work => workt(:, :, ithread)

      iatom_old = -1; jatom_old = -1; iset_old = -1; jset_old = -1
      ikind_old = -1; jkind_old = -1

      loop_tasks: DO itask = 1, ntasks

         CALL int2pair(tasks(3, itask), igrid_level, cindex, iatom, jatom, iset, jset, ipgf, jpgf, &
                       nimages, natom, maxset, maxpgf)

         ! apparently generalised collocation not implemented correctly yet
         IF(.NOT.(tasks(4, itask) .NE. 2))CALL cp__a("qs_linres_current.F",1013)

         IF (iatom .NE. iatom_old .OR. jatom .NE. jatom_old) THEN

            ikind = particle_set(iatom)%atomic_kind%kind_number
            jkind = particle_set(jatom)%atomic_kind%kind_number

            IF (iatom .NE. iatom_old) ra(:) = pbc(particle_set(iatom)%r, cell)

            brow = iatom
            bcol = jatom

            IF (ikind .NE. ikind_old) THEN
               CALL get_qs_kind(qs_kind_set(ikind), &
                                softb=my_soft, &
                                basis_set=orb_basis_set)

               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      first_sgf=first_sgfa, &
                                      lmax=la_max, &
                                      lmin=la_min, &
                                      npgf=npgfa, &
                                      nset=nseta, &
                                      nsgf_set=nsgfa, &
                                      pgf_radius=rpgfa, &
                                      set_radius=set_radius_a, &
                                      sphi=sphi_a, &
                                      zet=zeta)
            ENDIF

            IF (jkind .NE. jkind_old) THEN

               CALL get_qs_kind(qs_kind_set(jkind), &
                                softb=my_soft, &
                                basis_set=orb_basis_set)

               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                      first_sgf=first_sgfb, &
                                      kind_radius=kind_radius_b, &
                                      lmax=lb_max, &
                                      lmin=lb_min, &
                                      npgf=npgfb, &
                                      nset=nsetb, &
                                      nsgf_set=nsgfb, &
                                      pgf_radius=rpgfb, &
                                      set_radius=set_radius_b, &
                                      sphi=sphi_b, &
                                      zet=zetb)

            ENDIF

            CALL cp_dbcsr_get_block_p(matrix=deltajp_a(1)%matrix, row=brow, col=bcol, &
                                      block=jp_block_a, found=den_found)
            CALL cp_dbcsr_get_block_p(matrix=deltajp_b(1)%matrix, row=brow, col=bcol, &
                                      block=jp_block_b, found=den_found)
            CALL cp_dbcsr_get_block_p(matrix=deltajp_c(1)%matrix, row=brow, col=bcol, &
                                      block=jp_block_c, found=den_found)
            IF (do_igaim) CALL cp_dbcsr_get_block_p(matrix=deltajp_d(1)%matrix, row=brow, col=bcol, &
                                                    block=jp_block_d, found=den_found)

            IF (.NOT. ASSOCIATED(jp_block_a)) &
               CALL cp__b("qs_linres_current.F",1074,"p_block not associated in deltap")

            iatom_old = iatom
            jatom_old = jatom
            ikind_old = ikind
            jkind_old = jkind

            atom_pair_changed = .TRUE.

         ELSE

            atom_pair_changed = .FALSE.

         ENDIF

         IF (atom_pair_changed .OR. iset_old .NE. iset .OR. jset_old .NE. jset) THEN

            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)
            ncob = npgfb(jset)*ncoset(lb_max(jset))
            sgfb = first_sgfb(1, jset)
            ! Decontraction step for the selected blocks of the 3 density matrices

            CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                       1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                       jp_block_a(sgfa, sgfb), SIZE(jp_block_a, 1), &
                       0.0_dp, work(1, 1), maxco)
            CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                       1.0_dp, work(1, 1), maxco, &
                       sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                       0.0_dp, jpab_a(1, 1), maxco)

            CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                       1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                       jp_block_b(sgfa, sgfb), SIZE(jp_block_b, 1), &
                       0.0_dp, work(1, 1), maxco)
            CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                       1.0_dp, work(1, 1), maxco, &
                       sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                       0.0_dp, jpab_b(1, 1), maxco)

            CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                       1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                       jp_block_c(sgfa, sgfb), SIZE(jp_block_c, 1), &
                       0.0_dp, work(1, 1), maxco)
            CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                       1.0_dp, work(1, 1), maxco, &
                       sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                       0.0_dp, jpab_c(1, 1), maxco)

            IF (do_igaim) THEN
               CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                          1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                          jp_block_d(sgfa, sgfb), SIZE(jp_block_d, 1), &
                          0.0_dp, work(1, 1), maxco)
               CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                          1.0_dp, work(1, 1), maxco, &
                          sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                          0.0_dp, jpab_d(1, 1), maxco)
            ENDIF

            iset_old = iset
            jset_old = jset

         ENDIF

         rab(:) = dist_ab(:, itask)
         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         rb(:) = ra(:)+rab(:)
         zetp = zeta(ipgf, iset)+zetb(jpgf, jset)

         na1 = (ipgf-1)*ncoset(la_max(iset))+1
         na2 = ipgf*ncoset(la_max(iset))
         nb1 = (jpgf-1)*ncoset(lb_max(jset))+1
         nb2 = jpgf*ncoset(lb_max(jset))

         ! Four calls to the general collocate density, to multply the correct function
         ! to each density matrix

         !
         ! here the decontracted mat_jp_{ab} is multiplied by
         !     f_{ab} = g_{a} (dg_{b}/dr)_{idir} - (dg_{a}/dr)_{idir} g_{b}
         scale = 1.0_dp
         CALL collocate_pgf_product_rspace(la_max(iset), zeta(ipgf, iset), &
                                           la_min(iset), lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                                           ra, rab, rab2, scale, jpab_a, na1-1, nb1-1, &
                                           rs_current(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                                           eps_rho_rspace, &
                                           ga_gb_function=FUNC_ADBmDAB, &
                                           idir=idir, &
                                           map_consistent=map_consistent, lmax_global=lmax_global)
         IF (do_igaim) THEN
            ! here the decontracted mat_jb_{ab} is multiplied by
            !     f_{ab} = g_{a} * g_{b} ! THIS GOES OUTSIDE THE LOOP !
            IF (scale2 .NE. 0.0_dp) THEN
               CALL collocate_pgf_product_rspace(la_max(iset), zeta(ipgf, iset), &
                                                 la_min(iset), lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                                                 ra, rab, rab2, scale2, jpab_d, na1-1, nb1-1, &
                                                 rs_rho(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                                                 eps_rho_rspace, &
                                                 ga_gb_function=FUNC_AB, &
                                                 map_consistent=map_consistent, lmax_global=lmax_global)
            ENDIF !rm
            ! here the decontracted mat_jp_rii{ab} is multiplied by
            !     f_{ab} = g_{a} (d(r) - R_{b})_{iiB} (dg_{b}/dr)_{idir} -
            !             (dg_{a}/dr)_{idir} (d(r) - R_{b})_{iiB} g_{b}
            scale = 1.0_dp
            CALL collocate_pgf_product_rspace(la_max(iset), zeta(ipgf, iset), &
                                              la_min(iset), lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                                              ra, rab, rab2, scale, jpab_b, na1-1, nb1-1, &
                                              rs_current(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                                              eps_rho_rspace, &
                                              ga_gb_function=FUNC_ADBmDAB, &
                                              idir=idir, ir=iiiB, &
                                              rsgauge=rs_gauge(iiiB)%rs(igrid_level)%rs_grid, &
                                              rsbuf=current_env%rs_buf(igrid_level)%rs_grid, &
                                              map_consistent=map_consistent, lmax_global=lmax_global)
            ! here the decontracted mat_jp_riii{ab} is multiplied by
            !     f_{ab} = -g_{a} (d(r) - R_{b})_{iiB} (dg_{b}/dr)_{idir} +
            !             (dg_{a}/dr)_{idir} (d(r) - R_{b})_{iiB} g_{b}
            scale = -1.0_dp
            CALL collocate_pgf_product_rspace(la_max(iset), zeta(ipgf, iset), &
                                              la_min(iset), lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                                              ra, rab, rab2, scale, jpab_c, na1-1, nb1-1, &
                                              rs_current(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                                              eps_rho_rspace, &
                                              ga_gb_function=FUNC_ADBmDAB, &
                                              idir=idir, ir=iiB, &
                                              rsgauge=rs_gauge(iiB)%rs(igrid_level)%rs_grid, &
                                              rsbuf=current_env%rs_buf(igrid_level)%rs_grid, &
                                              map_consistent=map_consistent, lmax_global=lmax_global)
         ELSE
            ! here the decontracted mat_jp_rii{ab} is multiplied by
            !     f_{ab} = g_{a} (r - R_{b})_{iiB} (dg_{b}/dr)_{idir} -
            !             (dg_{a}/dr)_{idir} (r - R_{b})_{iiB} g_{b}
            scale = 1.0_dp
            CALL collocate_pgf_product_rspace(la_max(iset), zeta(ipgf, iset), &
                                              la_min(iset), lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                                              ra, rab, rab2, scale, jpab_b, na1-1, nb1-1, &
                                              rs_current(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                                              eps_rho_rspace, &
                                              ga_gb_function=FUNC_ARDBmDARB, &
                                              idir=idir, ir=iiiB, &
                                              map_consistent=map_consistent, lmax_global=lmax_global)
            ! here the decontracted mat_jp_riii{ab} is multiplied by
            !     f_{ab} = -g_{a} (r - R_{b})_{iiB} (dg_{b}/dr)_{idir} +
            !             (dg_{a}/dr)_{idir} (r - R_{b})_{iiB} g_{b}
            scale = -1.0_dp
            CALL collocate_pgf_product_rspace(la_max(iset), zeta(ipgf, iset), &
                                              la_min(iset), lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                                              ra, rab, rab2, scale, jpab_c, na1-1, nb1-1, &
                                              rs_current(igrid_level)%rs_grid, cell, cube_info(igrid_level), &
                                              eps_rho_rspace, &
                                              ga_gb_function=FUNC_ARDBmDARB, &
                                              idir=idir, ir=iiB, &
                                              map_consistent=map_consistent, lmax_global=lmax_global)
         ENDIF

      END DO loop_tasks
      !
      ! Scale the density with the gauge rho * ( r - d(r) ) if needed
      IF (do_igaim) THEN
         DO igrid_level = 1, gridlevel_info%ngrid_levels
            CALL rs_grid_mult_and_add(rs_current(igrid_level)%rs_grid, rs_rho(igrid_level)%rs_grid, &
                                      rs_gauge(idir2)%rs(igrid_level)%rs_grid, 1.0_dp)
         ENDDO
      ENDIF
      !   *** Release work storage ***

      IF (distributed_rs_grids) THEN
         CALL cp_dbcsr_deallocate_matrix(deltajp_a(1)%matrix)
         CALL cp_dbcsr_deallocate_matrix(deltajp_b(1)%matrix)
         CALL cp_dbcsr_deallocate_matrix(deltajp_c(1)%matrix)
         IF (do_igaim) CALL cp_dbcsr_deallocate_matrix(deltajp_d(1)%matrix)
      END IF
      DEALLOCATE (deltajp_a, deltajp_b, deltajp_c, deltajp_d)

      IF (nthread > 1) THEN
         DEALLOCATE (lgrid%r)
      END IF

      DEALLOCATE (jpabt_a, jpabt_b, jpabt_c, jpabt_d, workt, tasks, dist_ab)

      IF (distributed_rs_grids) THEN
         DEALLOCATE (atom_pair_send, atom_pair_recv)
      ENDIF

      CALL density_rs2pw(pw_env, rs_current, current_rs, current_gs)

      IF (ASSOCIATED(rs_rho) .AND. .NOT. my_retain_rsgrid) THEN
         DO i = 1, SIZE(rs_rho)
            CALL rs_grid_release(rs_rho(i)%rs_grid)
         END DO
      END IF

      ! Free the array of grids (grids themselves are released in density_rs2pw)
      DEALLOCATE (rs_current)

      CALL timestop(handle)

   END SUBROUTINE calculate_jrho_resp

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE current_set_gauge(current_env, qs_env)
      !
      TYPE(current_env_type)                   :: current_env
      TYPE(qs_environment_type), POINTER       :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'current_set_gauge', &
                                     routineP = moduleN//':'//routineN

      REAL(dp)                                 :: dbox(3)
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)    :: box_data
      INTEGER                                  :: handle, igrid_level, nbox(3), gauge
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :) :: box_ptr
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                :: rs_descs
      TYPE(pw_env_type), POINTER               :: pw_env
      TYPE(realspaces_grid_p_type), DIMENSION(:), POINTER :: rs_gauge

      TYPE(box_type), DIMENSION(:, :, :), POINTER :: box
      LOGICAL                                   :: use_old_gauge_atom

      NULLIFY (rs_gauge, box)

      CALL timeset(routineN, handle)

      CALL get_current_env(current_env=current_env, &
                           use_old_gauge_atom=use_old_gauge_atom, &
                           rs_gauge=rs_gauge, &
                           gauge=gauge)

      IF (gauge .EQ. current_gauge_atom) THEN
         CALL get_qs_env(qs_env=qs_env, &
                         pw_env=pw_env)
         CALL pw_env_get(pw_env=pw_env, &
                         rs_descs=rs_descs)
         !
         ! box the atoms
         IF (use_old_gauge_atom) THEN
            CALL box_atoms(qs_env)
         ELSE
            CALL box_atoms_new(current_env, qs_env, box)
         ENDIF
         !
         ! allocate and build the gauge
         ALLOCATE (rs_gauge(1)%rs(pw_env%gridlevel_info%ngrid_levels))
         ALLOCATE (rs_gauge(2)%rs(pw_env%gridlevel_info%ngrid_levels))
         ALLOCATE (rs_gauge(3)%rs(pw_env%gridlevel_info%ngrid_levels))
         DO igrid_level = pw_env%gridlevel_info%ngrid_levels, 1, -1

            CALL rs_grid_create(rs_gauge(1)%rs(igrid_level)%rs_grid, rs_descs(igrid_level)%rs_desc)
            CALL rs_grid_create(rs_gauge(2)%rs(igrid_level)%rs_grid, rs_descs(igrid_level)%rs_desc)
            CALL rs_grid_create(rs_gauge(3)%rs(igrid_level)%rs_grid, rs_descs(igrid_level)%rs_desc)

            IF (use_old_gauge_atom) THEN
               CALL collocate_gauge(current_env, qs_env, &
                                    rs_gauge(1)%rs(igrid_level)%rs_grid, &
                                    rs_gauge(2)%rs(igrid_level)%rs_grid, &
                                    rs_gauge(3)%rs(igrid_level)%rs_grid)
            ELSE
               CALL collocate_gauge_new(current_env, qs_env, &
                                        rs_gauge(1)%rs(igrid_level)%rs_grid, &
                                        rs_gauge(2)%rs(igrid_level)%rs_grid, &
                                        rs_gauge(3)%rs(igrid_level)%rs_grid, &
                                        box)
            ENDIF
         ENDDO
         !
         ! allocate the buf
         ALLOCATE (current_env%rs_buf(pw_env%gridlevel_info%ngrid_levels))
         DO igrid_level = 1, pw_env%gridlevel_info%ngrid_levels
            CALL rs_grid_create(current_env%rs_buf(igrid_level)%rs_grid, rs_descs(igrid_level)%rs_desc)
         END DO
         !
         DEALLOCATE (box_ptr, box_data)
         CALL deallocate_box(box)
      ENDIF

      CALL timestop(handle)

   CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
      SUBROUTINE box_atoms(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      REAL(kind=dp), PARAMETER                           :: box_size_guess = 5.0_dp

      INTEGER                                            :: i, iatom, ibox, ii, jbox, kbox, natms
      REAL(dp)                                           :: offset(3)
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ratom
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

         CALL get_qs_env(qs_env=qs_env, &
                         qs_kind_set=qs_kind_set, &
                         cell=cell, &
                         particle_set=particle_set)

         natms = SIZE(particle_set, 1)
         ALLOCATE (ratom(3, natms))
         !
         ! box the atoms
         nbox(1) = CEILING(cell%hmat(1, 1)/box_size_guess)
         nbox(2) = CEILING(cell%hmat(2, 2)/box_size_guess)
         nbox(3) = CEILING(cell%hmat(3, 3)/box_size_guess)
         !write(*,*) 'nbox',nbox
         dbox(1) = cell%hmat(1, 1)/REAL(nbox(1), dp)
         dbox(2) = cell%hmat(2, 2)/REAL(nbox(2), dp)
         dbox(3) = cell%hmat(3, 3)/REAL(nbox(3), dp)
         !write(*,*) 'dbox',dbox
         ALLOCATE (box_ptr(0:nbox(1), 0:nbox(2)-1, 0:nbox(3)-1), box_data(3, natms))
         box_data(:, :) = HUGE(0.0_dp)
         box_ptr(:, :, :) = HUGE(0)
         !
         offset(1) = cell%hmat(1, 1)*0.5_dp
         offset(2) = cell%hmat(2, 2)*0.5_dp
         offset(3) = cell%hmat(3, 3)*0.5_dp
         DO iatom = 1, natms
            ratom(:, iatom) = pbc(particle_set(iatom)%r(:), cell)+offset(:)
         ENDDO
         !
         i = 1
         DO kbox = 0, nbox(3)-1
         DO jbox = 0, nbox(2)-1
            box_ptr(0, jbox, kbox) = i
            DO ibox = 0, nbox(1)-1
               ii = 0
               DO iatom = 1, natms
                  IF (INT(ratom(1, iatom)/dbox(1)) .EQ. ibox .AND. &
                      INT(ratom(2, iatom)/dbox(2)) .EQ. jbox .AND. &
                      INT(ratom(3, iatom)/dbox(3)) .EQ. kbox) THEN
                     box_data(:, i) = ratom(:, iatom)-offset(:)
                     i = i+1
                     ii = ii+1
                  ENDIF
               ENDDO
               box_ptr(ibox+1, jbox, kbox) = box_ptr(ibox, jbox, kbox)+ii
            ENDDO
         ENDDO
         ENDDO
         !
         IF (.FALSE.) THEN
            DO kbox = 0, nbox(3)-1
            DO jbox = 0, nbox(2)-1
            DO ibox = 0, nbox(1)-1
               WRITE (*, *) 'box=', ibox, jbox, kbox
               WRITE (*, *) 'nbr atom=', box_ptr(ibox+1, jbox, kbox)-box_ptr(ibox, jbox, kbox)
               DO iatom = box_ptr(ibox, jbox, kbox), box_ptr(ibox+1, jbox, kbox)-1
                  WRITE (*, *) 'iatom=', iatom
                  WRITE (*, '(A,3E14.6)') 'coor=', box_data(:, iatom)
               ENDDO
            ENDDO
            ENDDO
            ENDDO
         ENDIF
         DEALLOCATE (ratom)
      END SUBROUTINE box_atoms

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
!> \param rs_grid_x ...
!> \param rs_grid_y ...
!> \param rs_grid_z ...
! **************************************************************************************************
      SUBROUTINE collocate_gauge(current_env, qs_env, rs_grid_x, rs_grid_y, rs_grid_z)
         !
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(realspace_grid_type), POINTER                 :: rs_grid_x, rs_grid_y, rs_grid_z

      INTEGER                                            :: i, iatom, ibeg, ibox, iend, imax, imin, &
                                                            j, jatom, jbox, jmax, jmin, k, kbox, &
                                                            kmax, kmin, lb(3), lb_local(3), natms, &
                                                            natms_local, ng(3)
      REAL(KIND=dp)                                      :: ab, buf_tmp, dist, dr(3), &
                                                            gauge_atom_radius, offset(3), pa, pb, &
                                                            point(3), pra(3), r(3), res(3), summe, &
                                                            tmp, x, y, z
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: buf, nrm_atms_pnt
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: atms_pnt, ratom
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: grid_x, grid_y, grid_z
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

!

         CALL get_current_env(current_env=current_env, &
                              gauge_atom_radius=gauge_atom_radius)
         !
         CALL get_qs_env(qs_env=qs_env, &
                         qs_kind_set=qs_kind_set, &
                         cell=cell, &
                         particle_set=particle_set)
         !
         natms = SIZE(particle_set, 1)
         dr(1) = rs_grid_x%desc%dh(1, 1)
         dr(2) = rs_grid_x%desc%dh(2, 2)
         dr(3) = rs_grid_x%desc%dh(3, 3)
         lb(:) = rs_grid_x%desc%lb(:)
         lb_local(:) = rs_grid_x%lb_local(:)
         grid_x => rs_grid_x%r(:, :, :)
         grid_y => rs_grid_y%r(:, :, :)
         grid_z => rs_grid_z%r(:, :, :)
         ng(:) = UBOUND(grid_x)
         offset(1) = cell%hmat(1, 1)*0.5_dp
         offset(2) = cell%hmat(2, 2)*0.5_dp
         offset(3) = cell%hmat(3, 3)*0.5_dp
         ALLOCATE (buf(natms), ratom(3, natms), atms_pnt(3, natms), nrm_atms_pnt(natms))
         !
         ! go over the grid
         DO k = 1, ng(3)
            DO j = 1, ng(2)
               DO i = 1, ng(1)
                  !
                  point(3) = REAL(k-1+lb_local(3)-lb(3), dp)*dr(3)
                  point(2) = REAL(j-1+lb_local(2)-lb(2), dp)*dr(2)
                  point(1) = REAL(i-1+lb_local(1)-lb(1), dp)*dr(1)
                  point = pbc(point, cell)
                  !
                  ! run over the overlaping boxes
                  natms_local = 0
                  kmin = INT((point(3)+offset(3)-gauge_atom_radius)/dbox(3))
                  kmax = INT((point(3)+offset(3)+gauge_atom_radius)/dbox(3))
                  IF (kmax-kmin+1 .GT. nbox(3)) THEN
                     kmin = 0
                     kmax = nbox(3)-1
                  ENDIF
                  DO kbox = kmin, kmax
                     jmin = INT((point(2)+offset(2)-gauge_atom_radius)/dbox(2))
                     jmax = INT((point(2)+offset(2)+gauge_atom_radius)/dbox(2))
                     IF (jmax-jmin+1 .GT. nbox(2)) THEN
                        jmin = 0
                        jmax = nbox(2)-1
                     ENDIF
                     DO jbox = jmin, jmax
                        imin = INT((point(1)+offset(1)-gauge_atom_radius)/dbox(1))
                        imax = INT((point(1)+offset(1)+gauge_atom_radius)/dbox(1))
                        IF (imax-imin+1 .GT. nbox(1)) THEN
                           imin = 0
                           imax = nbox(1)-1
                        ENDIF
                        DO ibox = imin, imax
                           ibeg = box_ptr(MODULO(ibox, nbox(1)), MODULO(jbox, nbox(2)), MODULO(kbox, nbox(3)))
                           iend = box_ptr(MODULO(ibox, nbox(1))+1, MODULO(jbox, nbox(2)), MODULO(kbox, nbox(3)))-1
                           DO iatom = ibeg, iend
                              r(:) = pbc(box_data(:, iatom)-point(:), cell)+point(:)
                              dist = (r(1)-point(1))**2+(r(2)-point(2))**2+(r(3)-point(3))**2
                              IF (dist .LT. gauge_atom_radius**2) THEN
                                 natms_local = natms_local+1
                                 ratom(:, natms_local) = r(:)
                                 !
                                 ! compute the distance atoms-point
                                 x = point(1)-r(1)
                                 y = point(2)-r(2)
                                 z = point(3)-r(3)
                                 atms_pnt(1, natms_local) = x
                                 atms_pnt(2, natms_local) = y
                                 atms_pnt(3, natms_local) = z
                                 nrm_atms_pnt(natms_local) = SQRT(x*x+y*y+z*z)
                              ENDIF
                           ENDDO
                        ENDDO
                     ENDDO
                  ENDDO
                  !
                  IF (natms_local .GT. 0) THEN
                     !
                     !
                     DO iatom = 1, natms_local
                        buf_tmp = 1.0_dp
                        pra(1) = atms_pnt(1, iatom)
                        pra(2) = atms_pnt(2, iatom)
                        pra(3) = atms_pnt(3, iatom)
                        pa = nrm_atms_pnt(iatom)
                        DO jatom = 1, natms_local
                           IF (iatom .EQ. jatom) CYCLE
                           pb = nrm_atms_pnt(jatom)
                           x = pra(1)-atms_pnt(1, jatom)
                           y = pra(2)-atms_pnt(2, jatom)
                           z = pra(3)-atms_pnt(3, jatom)
                           ab = SQRT(x*x+y*y+z*z)
                           !
                           tmp = (pa-pb)/ab
                           tmp = 0.5_dp*(3.0_dp-tmp*tmp)*tmp
                           tmp = 0.5_dp*(3.0_dp-tmp*tmp)*tmp
                           tmp = 0.5_dp*(3.0_dp-tmp*tmp)*tmp
                           buf_tmp = buf_tmp*0.5_dp*(1.0_dp-tmp)
                        ENDDO
                        buf(iatom) = buf_tmp
                     ENDDO
                     res(1) = 0.0_dp
                     res(2) = 0.0_dp
                     res(3) = 0.0_dp
                     summe = 0.0_dp
                     DO iatom = 1, natms_local
                        res(1) = res(1)+ratom(1, iatom)*buf(iatom)
                        res(2) = res(2)+ratom(2, iatom)*buf(iatom)
                        res(3) = res(3)+ratom(3, iatom)*buf(iatom)
                        summe = summe+buf(iatom)
                     ENDDO
                     res(1) = res(1)/summe
                     res(2) = res(2)/summe
                     res(3) = res(3)/summe
                     grid_x(i, j, k) = point(1)-res(1)
                     grid_y(i, j, k) = point(2)-res(2)
                     grid_z(i, j, k) = point(3)-res(3)
                  ELSE
                     grid_x(i, j, k) = 0.0_dp
                     grid_y(i, j, k) = 0.0_dp
                     grid_z(i, j, k) = 0.0_dp
                  ENDIF
               ENDDO
            ENDDO
         ENDDO

         DEALLOCATE (buf, ratom, atms_pnt, nrm_atms_pnt)

      END SUBROUTINE collocate_gauge

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
!> \param box ...
! **************************************************************************************************
      SUBROUTINE box_atoms_new(current_env, qs_env, box)
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(box_type), DIMENSION(:, :, :), POINTER        :: box

      CHARACTER(LEN=*), PARAMETER :: routineN = 'box_atoms_new', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, ibeg, ibox, iend, &
                                                            ifind, ii, imax, imin, j, jatom, jbox, &
                                                            jmax, jmin, k, kbox, kmax, kmin, &
                                                            natms, natms_local
      REAL(dp)                                           :: gauge_atom_radius, offset(3), scale
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: ratom
      REAL(dp), DIMENSION(:, :), POINTER                 :: r_ptr
      REAL(kind=dp)                                      :: box_center(3), box_center_wrap(3), &
                                                            box_size_guess, r(3)
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

         CALL timeset(routineN, handle)

         CALL get_qs_env(qs_env=qs_env, &
                         qs_kind_set=qs_kind_set, &
                         cell=cell, &
                         particle_set=particle_set)

         CALL get_current_env(current_env=current_env, &
                              gauge_atom_radius=gauge_atom_radius)

         scale = 2.0_dp

         box_size_guess = gauge_atom_radius/scale

         natms = SIZE(particle_set, 1)
         ALLOCATE (ratom(3, natms))

         !
         ! box the atoms
         nbox(1) = CEILING(cell%hmat(1, 1)/box_size_guess)
         nbox(2) = CEILING(cell%hmat(2, 2)/box_size_guess)
         nbox(3) = CEILING(cell%hmat(3, 3)/box_size_guess)
         dbox(1) = cell%hmat(1, 1)/REAL(nbox(1), dp)
         dbox(2) = cell%hmat(2, 2)/REAL(nbox(2), dp)
         dbox(3) = cell%hmat(3, 3)/REAL(nbox(3), dp)
         ALLOCATE (box_ptr(0:nbox(1), 0:nbox(2)-1, 0:nbox(3)-1), box_data(3, natms))
         box_data(:, :) = HUGE(0.0_dp)
         box_ptr(:, :, :) = HUGE(0)
         !
         offset(1) = cell%hmat(1, 1)*0.5_dp
         offset(2) = cell%hmat(2, 2)*0.5_dp
         offset(3) = cell%hmat(3, 3)*0.5_dp
         DO iatom = 1, natms
            ratom(:, iatom) = pbc(particle_set(iatom)%r(:), cell)
         ENDDO
         !
         i = 1
         DO kbox = 0, nbox(3)-1
         DO jbox = 0, nbox(2)-1
            box_ptr(0, jbox, kbox) = i
            DO ibox = 0, nbox(1)-1
               ii = 0
               DO iatom = 1, natms
                  IF (MODULO(FLOOR(ratom(1, iatom)/dbox(1)), nbox(1)) .EQ. ibox .AND. &
                      MODULO(FLOOR(ratom(2, iatom)/dbox(2)), nbox(2)) .EQ. jbox .AND. &
                      MODULO(FLOOR(ratom(3, iatom)/dbox(3)), nbox(3)) .EQ. kbox) THEN
                     box_data(:, i) = ratom(:, iatom)
                     i = i+1
                     ii = ii+1
                  ENDIF
               ENDDO
               box_ptr(ibox+1, jbox, kbox) = box_ptr(ibox, jbox, kbox)+ii
            ENDDO
         ENDDO
         ENDDO
         !
         IF (.FALSE.) THEN
            DO kbox = 0, nbox(3)-1
            DO jbox = 0, nbox(2)-1
            DO ibox = 0, nbox(1)-1
               IF (box_ptr(ibox+1, jbox, kbox)-box_ptr(ibox, jbox, kbox) .GT. 0) THEN
                  WRITE (*, *) 'box=', ibox, jbox, kbox
                  WRITE (*, *) 'nbr atom=', box_ptr(ibox+1, jbox, kbox)-box_ptr(ibox, jbox, kbox)
                  DO iatom = box_ptr(ibox, jbox, kbox), box_ptr(ibox+1, jbox, kbox)-1
                     WRITE (*, '(A,I3,3E14.6)') 'coor=', iatom, box_data(:, iatom)
                  ENDDO
               ENDIF
            ENDDO
            ENDDO
            ENDDO
         ENDIF
         !
         NULLIFY (box)
         ALLOCATE (box(0:nbox(1)-1, 0:nbox(2)-1, 0:nbox(3)-1))
         !
         ! build the list
         DO k = 0, nbox(3)-1
         DO j = 0, nbox(2)-1
         DO i = 0, nbox(1)-1
            !
            box_center(1) = (REAL(i, dp)+0.5_dp)*dbox(1)
            box_center(2) = (REAL(j, dp)+0.5_dp)*dbox(2)
            box_center(3) = (REAL(k, dp)+0.5_dp)*dbox(3)
            box_center_wrap = pbc(box_center, cell)
            !
            ! find the atoms that are in the overlaping boxes
            natms_local = 0
            kmin = FLOOR((box_center(3)-gauge_atom_radius)/dbox(3))
            kmax = FLOOR((box_center(3)+gauge_atom_radius)/dbox(3))
            IF (kmax-kmin+1 .GT. nbox(3)) THEN
               kmin = 0
               kmax = nbox(3)-1
            ENDIF
            DO kbox = kmin, kmax
               jmin = FLOOR((box_center(2)-gauge_atom_radius)/dbox(2))
               jmax = FLOOR((box_center(2)+gauge_atom_radius)/dbox(2))
               IF (jmax-jmin+1 .GT. nbox(2)) THEN
                  jmin = 0
                  jmax = nbox(2)-1
               ENDIF
               DO jbox = jmin, jmax
                  imin = FLOOR((box_center(1)-gauge_atom_radius)/dbox(1))
                  imax = FLOOR((box_center(1)+gauge_atom_radius)/dbox(1))
                  IF (imax-imin+1 .GT. nbox(1)) THEN
                     imin = 0
                     imax = nbox(1)-1
                  ENDIF
                  DO ibox = imin, imax
                     ibeg = box_ptr(MODULO(ibox, nbox(1)), MODULO(jbox, nbox(2)), MODULO(kbox, nbox(3)))
                     iend = box_ptr(MODULO(ibox, nbox(1))+1, MODULO(jbox, nbox(2)), MODULO(kbox, nbox(3)))-1
                     DO iatom = ibeg, iend
                        r = pbc(box_center_wrap(:)-box_data(:, iatom), cell)
                        IF (ABS(r(1)) .LE. (scale+0.5_dp)*dbox(1) .AND. &
                            ABS(r(2)) .LE. (scale+0.5_dp)*dbox(2) .AND. &
                            ABS(r(3)) .LE. (scale+0.5_dp)*dbox(3)) THEN
                           natms_local = natms_local+1
                           ratom(:, natms_local) = box_data(:, iatom)
                        ENDIF
                     ENDDO
                  ENDDO ! box
               ENDDO
            ENDDO
            !
            ! set the list
            box(i, j, k)%n = natms_local
            NULLIFY (box(i, j, k)%r)
            IF (natms_local .GT. 0) THEN
               ALLOCATE (box(i, j, k)%r(3, natms_local))
               r_ptr => box(i, j, k)%r
               CALL dcopy(3*natms_local, ratom(1, 1), 1, r_ptr(1, 1), 1)
            ENDIF
         ENDDO ! list
         ENDDO
         ENDDO

         IF (.FALSE.) THEN
            DO k = 0, nbox(3)-1
            DO j = 0, nbox(2)-1
            DO i = 0, nbox(1)-1
               IF (box(i, j, k)%n .GT. 0) THEN
                  WRITE (*, *)
                  WRITE (*, *) 'box=', i, j, k
                  box_center(1) = (REAL(i, dp)+0.5_dp)*dbox(1)
                  box_center(2) = (REAL(j, dp)+0.5_dp)*dbox(2)
                  box_center(3) = (REAL(k, dp)+0.5_dp)*dbox(3)
                  box_center = pbc(box_center, cell)
                  WRITE (*, '(A,3E14.6)') 'box_center=', box_center
                  WRITE (*, *) 'nbr atom=', box(i, j, k)%n
                  r_ptr => box(i, j, k)%r
                  DO iatom = 1, box(i, j, k)%n
                     WRITE (*, '(A,I3,3E14.6)') 'coor=', iatom, r_ptr(:, iatom)
                     r(:) = pbc(box_center(:)-r_ptr(:, iatom), cell)
                     IF (ABS(r(1)) .GT. (scale+0.5_dp)*dbox(1) .OR. &
                         ABS(r(2)) .GT. (scale+0.5_dp)*dbox(2) .OR. &
                         ABS(r(3)) .GT. (scale+0.5_dp)*dbox(3)) THEN
                        WRITE (*, *) 'error too many atoms'
                        WRITE (*, *) 'dist=', ABS(r(:))
                        WRITE (*, *) 'large_dist=', (scale+0.5_dp)*dbox
                        CALL cp__b("qs_linres_current.F",1790,"")
                     ENDIF
                  ENDDO
               ENDIF
            ENDDO ! list
            ENDDO
            ENDDO
         ENDIF

         IF (.TRUE.) THEN
            DO k = 0, nbox(3)-1
            DO j = 0, nbox(2)-1
            DO i = 0, nbox(1)-1
               box_center(1) = (REAL(i, dp)+0.5_dp)*dbox(1)
               box_center(2) = (REAL(j, dp)+0.5_dp)*dbox(2)
               box_center(3) = (REAL(k, dp)+0.5_dp)*dbox(3)
               box_center = pbc(box_center, cell)
               r_ptr => box(i, j, k)%r
               DO iatom = 1, natms
                  r(:) = pbc(box_center(:)-ratom(:, iatom), cell)
                  ifind = 0
                  DO jatom = 1, box(i, j, k)%n
                     IF (SUM(ABS(ratom(:, iatom)-r_ptr(:, jatom))) .LT. 1E-10_dp) ifind = 1
                  ENDDO

                  IF (ifind .EQ. 0) THEN
                     IF (SQRT(DOT_PRODUCT(r, r)) .LT. gauge_atom_radius) THEN
                        WRITE (*, *) 'error atom too close'
                        WRITE (*, *) 'iatom', iatom
                        WRITE (*, *) 'box_center', box_center
                        WRITE (*, *) 'ratom', ratom(:, iatom)
                        WRITE (*, *) 'gauge_atom_radius', gauge_atom_radius
                        CALL cp__b("qs_linres_current.F",1822,"")
                     ENDIF
                  ENDIF
               ENDDO
            ENDDO ! list
            ENDDO
            ENDDO
         ENDIF

         DEALLOCATE (ratom)

         CALL timestop(handle)

      END SUBROUTINE box_atoms_new

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
!> \param rs_grid_x ...
!> \param rs_grid_y ...
!> \param rs_grid_z ...
!> \param box ...
! **************************************************************************************************
      SUBROUTINE collocate_gauge_new(current_env, qs_env, rs_grid_x, rs_grid_y, rs_grid_z, box)
         !
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(realspace_grid_type), POINTER                 :: rs_grid_x, rs_grid_y, rs_grid_z
      TYPE(box_type), DIMENSION(:, :, :), POINTER        :: box

      CHARACTER(LEN=*), PARAMETER :: routineN = 'collocate_gauge_new', &
         routineP = moduleN//':'//routineN

      INTEGER :: delta_lb(3), handle, i, iatom, ib, ibe, ibox, ibs, ie, is, j, jatom, jb, jbe, &
         jbox, jbs, je, js, k, kb, kbe, kbox, kbs, ke, ks, lb(3), lb_local(3), natms, &
         natms_local0, natms_local1, ng(3)
      REAL(dp), DIMENSION(:, :), POINTER                 :: r_ptr
      REAL(KIND=dp)                                      :: ab, box_center(3), buf_tmp, dist, dr(3), &
                                                            gauge_atom_radius, offset(3), pa, pb, &
                                                            point(3), pra(3), r(3), res(3), summe, &
                                                            tmp, x, y, z
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: buf, nrm_atms_pnt
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: atms_pnt, ratom
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: grid_x, grid_y, grid_z
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

         CALL timeset(routineN, handle)

!
         CALL get_current_env(current_env=current_env, &
                              gauge_atom_radius=gauge_atom_radius)
         !
         CALL get_qs_env(qs_env=qs_env, &
                         qs_kind_set=qs_kind_set, &
                         cell=cell, &
                         particle_set=particle_set)
         !
         natms = SIZE(particle_set, 1)
         dr(1) = rs_grid_x%desc%dh(1, 1)
         dr(2) = rs_grid_x%desc%dh(2, 2)
         dr(3) = rs_grid_x%desc%dh(3, 3)
         lb(:) = rs_grid_x%desc%lb(:)
         lb_local(:) = rs_grid_x%lb_local(:)
         grid_x => rs_grid_x%r(:, :, :)
         grid_y => rs_grid_y%r(:, :, :)
         grid_z => rs_grid_z%r(:, :, :)
         ng(:) = UBOUND(grid_x)
         delta_lb(:) = lb_local(:)-lb(:)
         offset(1) = cell%hmat(1, 1)*0.5_dp
         offset(2) = cell%hmat(2, 2)*0.5_dp
         offset(3) = cell%hmat(3, 3)*0.5_dp
         ALLOCATE (buf(natms), ratom(3, natms), atms_pnt(3, natms), nrm_atms_pnt(natms))
         !
         ! find the boxes that match the grid
         ibs = FLOOR(REAL(delta_lb(1), dp)*dr(1)/dbox(1))
         ibe = FLOOR(REAL(ng(1)-1+delta_lb(1), dp)*dr(1)/dbox(1))
         jbs = FLOOR(REAL(delta_lb(2), dp)*dr(2)/dbox(2))
         jbe = FLOOR(REAL(ng(2)-1+delta_lb(2), dp)*dr(2)/dbox(2))
         kbs = FLOOR(REAL(delta_lb(3), dp)*dr(3)/dbox(3))
         kbe = FLOOR(REAL(ng(3)-1+delta_lb(3), dp)*dr(3)/dbox(3))
         !
         ! go over the box-list
         DO kb = kbs, kbe
         DO jb = jbs, jbe
         DO ib = ibs, ibe
            ibox = MODULO(ib, nbox(1))
            jbox = MODULO(jb, nbox(2))
            kbox = MODULO(kb, nbox(3))
            !
            is = MAX(CEILING(REAL(ib, dp)*dbox(1)/dr(1)), delta_lb(1))-delta_lb(1)+1
            ie = MIN(FLOOR(REAL(ib+1, dp)*dbox(1)/dr(1)), ng(1)-1+delta_lb(1))-delta_lb(1)+1
            js = MAX(CEILING(REAL(jb, dp)*dbox(2)/dr(2)), delta_lb(2))-delta_lb(2)+1
            je = MIN(FLOOR(REAL(jb+1, dp)*dbox(2)/dr(2)), ng(2)-1+delta_lb(2))-delta_lb(2)+1
            ks = MAX(CEILING(REAL(kb, dp)*dbox(3)/dr(3)), delta_lb(3))-delta_lb(3)+1
            ke = MIN(FLOOR(REAL(kb+1, dp)*dbox(3)/dr(3)), ng(3)-1+delta_lb(3))-delta_lb(3)+1
            !
            ! sanity checks
            IF (.TRUE.) THEN
               IF (REAL(ks-1+delta_lb(3), dp)*dr(3) .LT. REAL(kb, dp)*dbox(3) .OR. &
                   REAL(ke-1+delta_lb(3), dp)*dr(3) .GT. REAL(kb+1, dp)*dbox(3)) THEN
                  WRITE (*, *) 'box_k', REAL(kb, dp)*dbox(3), REAL(kb+1, dp)*dbox(3)
                  WRITE (*, *) 'point_k', REAL(ks-1+delta_lb(3), dp)*dr(3), REAL(ke-1+delta_lb(3), dp)*dr(3)
                  WRITE (*, *) 'ibox', ibox, 'jbox', jbox, 'kbox', kbox
                  WRITE (*, *) 'is,ie', is, ie, ' js,je', js, je, ' ks,ke', ks, ke
                  WRITE (*, *) 'ibs,ibe', ibs, ibe, ' jbs,jbe', jbs, jbe, ' kbs,kbe', kbs, kbe
                  CALL cp__b("qs_linres_current.F",1930,"we stop_k")
               ENDIF
               IF (REAL(js-1+delta_lb(2), dp)*dr(2) .LT. REAL(jb, dp)*dbox(2) .OR. &
                   REAL(je-1+delta_lb(2), dp)*dr(2) .GT. REAL(jb+1, dp)*dbox(2)) THEN
                  WRITE (*, *) 'box_j', REAL(jb, dp)*dbox(2), REAL(jb+1, dp)*dbox(2)
                  WRITE (*, *) 'point_j', REAL(js-1+delta_lb(2), dp)*dr(2), REAL(je-1+delta_lb(2), dp)*dr(2)
                  WRITE (*, *) 'is,ie', is, ie, ' js,je', js, je, ' ks,ke', ks, ke
                  WRITE (*, *) 'ibs,ibe', ibs, ibe, ' jbs,jbe', jbs, jbe, ' kbs,kbe', kbs, kbe
                  CALL cp__b("qs_linres_current.F",1938,"we stop_j")
               ENDIF
               IF (REAL(is-1+delta_lb(1), dp)*dr(1) .LT. REAL(ib, dp)*dbox(1) .OR. &
                   REAL(ie-1+delta_lb(1), dp)*dr(1) .GT. REAL(ib+1, dp)*dbox(1)) THEN
                  WRITE (*, *) 'box_i', REAL(ib, dp)*dbox(1), REAL(ib+1, dp)*dbox(1)
                  WRITE (*, *) 'point_i', REAL(is-1+delta_lb(1), dp)*dr(1), REAL(ie-1+delta_lb(1), dp)*dr(1)
                  WRITE (*, *) 'is,ie', is, ie, ' js,je', js, je, ' ks,ke', ks, ke
                  WRITE (*, *) 'ibs,ibe', ibs, ibe, ' jbs,jbe', jbs, jbe, ' kbs,kbe', kbs, kbe
                  CALL cp__b("qs_linres_current.F",1946,"we stop_i")
               ENDIF
            ENDIF
            !
            ! the center of the box
            box_center(1) = (REAL(ibox, dp)+0.5_dp)*dbox(1)
            box_center(2) = (REAL(jbox, dp)+0.5_dp)*dbox(2)
            box_center(3) = (REAL(kbox, dp)+0.5_dp)*dbox(3)
            !
            ! find the atoms that are in the overlaping boxes
            natms_local0 = box(ibox, jbox, kbox)%n
            r_ptr => box(ibox, jbox, kbox)%r
            !
            ! go over the grid inside the box
            IF (natms_local0 .GT. 0) THEN
               !
               ! here there are some atoms...
               DO k = ks, ke
               DO j = js, je
               DO i = is, ie
                  point(1) = REAL(i-1+delta_lb(1), dp)*dr(1)
                  point(2) = REAL(j-1+delta_lb(2), dp)*dr(2)
                  point(3) = REAL(k-1+delta_lb(3), dp)*dr(3)
                  point = pbc(point, cell)
                  !
                  ! compute atom-point distances
                  natms_local1 = 0
                  DO iatom = 1, natms_local0
                     r(:) = pbc(r_ptr(:, iatom)-point(:), cell)+point(:) !needed?
                     dist = (r(1)-point(1))**2+(r(2)-point(2))**2+(r(3)-point(3))**2
                     IF (dist .LT. gauge_atom_radius**2) THEN
                        natms_local1 = natms_local1+1
                        ratom(:, natms_local1) = r(:)
                        !
                        ! compute the distance atoms-point
                        x = point(1)-r(1)
                        y = point(2)-r(2)
                        z = point(3)-r(3)
                        atms_pnt(1, natms_local1) = x
                        atms_pnt(2, natms_local1) = y
                        atms_pnt(3, natms_local1) = z
                        nrm_atms_pnt(natms_local1) = SQRT(x*x+y*y+z*z)
                     ENDIF
                  ENDDO
                  !
                  !
                  IF (natms_local1 .GT. 0) THEN
                     !
                     ! build the step
                     DO iatom = 1, natms_local1
                        buf_tmp = 1.0_dp
                        pra(1) = atms_pnt(1, iatom)
                        pra(2) = atms_pnt(2, iatom)
                        pra(3) = atms_pnt(3, iatom)
                        pa = nrm_atms_pnt(iatom)
                        DO jatom = 1, natms_local1
                           IF (iatom .EQ. jatom) CYCLE
                           pb = nrm_atms_pnt(jatom)
                           x = pra(1)-atms_pnt(1, jatom)
                           y = pra(2)-atms_pnt(2, jatom)
                           z = pra(3)-atms_pnt(3, jatom)
                           ab = SQRT(x*x+y*y+z*z)
                           !
                           tmp = (pa-pb)/ab
                           tmp = 0.5_dp*(3.0_dp-tmp*tmp)*tmp
                           tmp = 0.5_dp*(3.0_dp-tmp*tmp)*tmp
                           tmp = 0.5_dp*(3.0_dp-tmp*tmp)*tmp
                           buf_tmp = buf_tmp*0.5_dp*(1.0_dp-tmp)
                        ENDDO
                        buf(iatom) = buf_tmp
                     ENDDO
                     res(1) = 0.0_dp
                     res(2) = 0.0_dp
                     res(3) = 0.0_dp
                     summe = 0.0_dp
                     DO iatom = 1, natms_local1
                        res(1) = res(1)+ratom(1, iatom)*buf(iatom)
                        res(2) = res(2)+ratom(2, iatom)*buf(iatom)
                        res(3) = res(3)+ratom(3, iatom)*buf(iatom)
                        summe = summe+buf(iatom)
                     ENDDO
                     res(1) = res(1)/summe
                     res(2) = res(2)/summe
                     res(3) = res(3)/summe
                     grid_x(i, j, k) = point(1)-res(1)
                     grid_y(i, j, k) = point(2)-res(2)
                     grid_z(i, j, k) = point(3)-res(3)
                  ELSE
                     grid_x(i, j, k) = 0.0_dp
                     grid_y(i, j, k) = 0.0_dp
                     grid_z(i, j, k) = 0.0_dp
                  ENDIF
               ENDDO ! grid
               ENDDO
               ENDDO
               !
            ELSE
               !
               ! here there is no atom
               DO k = ks, ke
               DO j = js, je
               DO i = is, ie
                  grid_x(i, j, k) = 0.0_dp
                  grid_y(i, j, k) = 0.0_dp
                  grid_z(i, j, k) = 0.0_dp
               ENDDO ! grid
               ENDDO
               ENDDO
               !
            ENDIF
            !
         ENDDO ! list
         ENDDO
         ENDDO

         DEALLOCATE (buf, ratom, atms_pnt, nrm_atms_pnt)

         CALL timestop(handle)

      END SUBROUTINE collocate_gauge_new

! **************************************************************************************************
!> \brief ...
!> \param box ...
! **************************************************************************************************
      SUBROUTINE deallocate_box(box)
      TYPE(box_type), DIMENSION(:, :, :), POINTER        :: box

      INTEGER                                            :: i, j, k

         IF (ASSOCIATED(box)) THEN
            DO k = LBOUND(box, 3), UBOUND(box, 3)
            DO j = LBOUND(box, 2), UBOUND(box, 2)
            DO i = LBOUND(box, 1), UBOUND(box, 1)
               IF (ASSOCIATED(box(i, j, k)%r)) THEN
                  DEALLOCATE (box(i, j, k)%r)
               ENDIF
            ENDDO
            ENDDO
            ENDDO
            DEALLOCATE (box)
         ENDIF
      END SUBROUTINE deallocate_box
   END SUBROUTINE current_set_gauge

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
! **************************************************************************************************
   SUBROUTINE current_build_chi(current_env, qs_env, iB)
      !
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB

      IF (current_env%full) THEN
         CALL current_build_chi_many_centers(current_env, qs_env, iB)
      ELSE
         CALL current_build_chi_one_center(current_env, qs_env, iB)
      ENDIF

   END SUBROUTINE current_build_chi

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
! **************************************************************************************************
   SUBROUTINE current_build_chi_many_centers(current_env, qs_env, iB)
      !
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB

      CHARACTER(LEN=*), PARAMETER :: routineN = 'current_build_chi_many_centers', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, icenter, idir, idir2, ii, iiB, iii, iiiB, ispin, istate, j, jstate, &
         max_states, nao, natom, nbr_center(2), nmo, nspins, nstate_loc, nstates(2), output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, last_sgf
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_sizes
      LOGICAL                                            :: chi_pbc, gapw
      REAL(dp)                                           :: chi(3), chi_tmp, contrib, contrib2, &
                                                            dk(3), int_current(3), &
                                                            int_current_tmp, maxocc
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_2d_i_p_type), DIMENSION(:), POINTER        :: center_list
      TYPE(cp_2d_r_p_type), DIMENSION(:), POINTER        :: centers_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op_mom_ao, op_p_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: op_mom_der_ao
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: p_rxp, psi0_order, r_p1, r_p2
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: psi1_D, psi1_p, psi1_rxp, rr_p1, rr_p2, &
                                                            rr_rxp
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, psi0, psi_D, psi_p1, psi_p2, &
                                                            psi_rxp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_all, sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

!

      CALL timeset(routineN, handle)
      !
      NULLIFY (dft_control, mos, para_env, mo_coeff, op_mom_ao, &
               op_mom_der_ao, center_list, centers_set, psi0, psi_rxp, psi_D, psi_p1, psi_p2, &
               op_p_ao, psi1_p, psi1_rxp, psi1_D, p_rxp, r_p1, r_p2, rr_rxp, rr_p1, rr_p2, &
               cell, psi0_order, particle_set, qs_kind_set)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      mos=mos, &
                      para_env=para_env, &
                      cell=cell, &
                      dbcsr_dist=dbcsr_dist, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      sab_all=sab_all, &
                      sab_orb=sab_orb)

      nspins = dft_control%nspins
      gapw = dft_control%qs_control%gapw

      CALL get_current_env(current_env=current_env, &
                           chi_pbc=chi_pbc, &
                           nao=nao, &
                           nbr_center=nbr_center, &
                           center_list=center_list, &
                           centers_set=centers_set, &
                           psi1_p=psi1_p, &
                           psi1_rxp=psi1_rxp, &
                           psi1_D=psi1_D, &
                           nstates=nstates, &
                           psi0_order=psi0_order)
      !
      ! get max nbr of states per center
      max_states = 0
      DO ispin = 1, nspins
         DO icenter = 1, nbr_center(ispin)
            max_states = MAX(max_states, center_list(ispin)%array(1, icenter+1)&
                 &                     -center_list(ispin)%array(1, icenter))
         ENDDO
      ENDDO
      !
      ! Allocate sparse matrices for dipole, quadrupole and their derivatives => 9x3
      ! Remember the derivatives are antisymmetric
      CALL cp_dbcsr_allocate_matrix_set(op_mom_ao, 9)
      CALL cp_dbcsr_allocate_matrix_set(op_mom_der_ao, 9, 3)
      !
      ! prepare for allocation
      natom = SIZE(particle_set, 1)
      ALLOCATE (first_sgf(natom))
      ALLOCATE (last_sgf(natom))
      CALL get_particle_set(particle_set, qs_kind_set, &
                            first_sgf=first_sgf, &
                            last_sgf=last_sgf)
      ALLOCATE (row_blk_sizes(natom))
      CALL convert_offsets_to_sizes(first_sgf, row_blk_sizes, last_sgf)
      DEALLOCATE (first_sgf)
      DEALLOCATE (last_sgf)
      !
      !
      ALLOCATE (op_mom_ao(1)%matrix)
      CALL cp_dbcsr_init(op_mom_ao(1)%matrix)
      CALL cp_dbcsr_create(matrix=op_mom_ao(1)%matrix, &
                           name="op_mom", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_no_symmetry, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0, mutable_work=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(op_mom_ao(1)%matrix, sab_all)

      DO idir2 = 1, 3
         ALLOCATE (op_mom_der_ao(1, idir2)%matrix)
         CALL cp_dbcsr_init(op_mom_der_ao(1, idir2)%matrix)
         CALL cp_dbcsr_copy(op_mom_der_ao(1, idir2)%matrix, op_mom_ao(1)%matrix, &
                            "op_mom_der_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir2))))
      ENDDO

      DO idir = 2, SIZE(op_mom_ao, 1)
         ALLOCATE (op_mom_ao(idir)%matrix)
         CALL cp_dbcsr_init(op_mom_ao(idir)%matrix)
         CALL cp_dbcsr_copy(op_mom_ao(idir)%matrix, op_mom_ao(1)%matrix, &
                            "op_mom_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir))))
         DO idir2 = 1, 3
            ALLOCATE (op_mom_der_ao(idir, idir2)%matrix)
            CALL cp_dbcsr_init(op_mom_der_ao(idir, idir2)%matrix)
            CALL cp_dbcsr_copy(op_mom_der_ao(idir, idir2)%matrix, op_mom_ao(1)%matrix, &
                               "op_mom_der_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir*idir2))))
         ENDDO
      ENDDO
      !
      CALL cp_dbcsr_allocate_matrix_set(op_p_ao, 3)
      ALLOCATE (op_p_ao(1)%matrix)
      CALL cp_dbcsr_init(op_p_ao(1)%matrix)
      CALL cp_dbcsr_create(matrix=op_p_ao(1)%matrix, &
                           name="op_p_ao", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_antisymmetric, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0, mutable_work=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(op_p_ao(1)%matrix, sab_orb)

      DO idir = 2, 3
         ALLOCATE (op_p_ao(idir)%matrix)
         CALL cp_dbcsr_init(op_p_ao(idir)%matrix)
         CALL cp_dbcsr_copy(op_p_ao(idir)%matrix, op_p_ao(1)%matrix, &
                            "op_p_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir))))
      ENDDO
      !
      !
      DEALLOCATE (row_blk_sizes)
      !
      !
      ! Allocate full matrices for only one vector
      mo_coeff => psi0_order(1)%matrix
      NULLIFY (tmp_fm_struct)
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                               ncol_global=max_states, para_env=para_env, &
                               context=mo_coeff%matrix_struct%context)
      CALL cp_fm_create(psi0, tmp_fm_struct)
      CALL cp_fm_create(psi_D, tmp_fm_struct)
      CALL cp_fm_create(psi_rxp, tmp_fm_struct)
      CALL cp_fm_create(psi_p1, tmp_fm_struct)
      CALL cp_fm_create(psi_p2, tmp_fm_struct)
      CALL cp_fm_struct_release(tmp_fm_struct)
      !
      ALLOCATE (p_rxp(3), r_p1(3), r_p2(3), rr_rxp(9, 3), rr_p1(9, 3), rr_p2(9, 3))
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                               ncol_global=max_states, para_env=para_env, &
                               context=mo_coeff%matrix_struct%context)
      DO idir = 1, 3
         CALL cp_fm_create(p_rxp(idir)%matrix, tmp_fm_struct)
         CALL cp_fm_create(r_p1(idir)%matrix, tmp_fm_struct)
         CALL cp_fm_create(r_p2(idir)%matrix, tmp_fm_struct)
         DO idir2 = 1, 9
            CALL cp_fm_create(rr_rxp(idir2, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(rr_p1(idir2, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(rr_p2(idir2, idir)%matrix, tmp_fm_struct)
         ENDDO
      ENDDO
      CALL cp_fm_struct_release(tmp_fm_struct)
      !
      !
      !
      ! recompute the linear momentum matrices
      CALL build_lin_mom_matrix(qs_env, op_p_ao)
      !CALL p_xyz_ao(op_p_ao,qs_env,minimum_image=.FALSE.)
      !
      !
      ! get iiB and iiiB
      CALL set_vecp(iB, iiB, iiiB)
      DO ispin = 1, nspins
         !
         ! get ground state MOS
         nmo = nstates(ispin)
         mo_coeff => psi0_order(ispin)%matrix
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, maxocc=maxocc)
         !
         ! Initialize the temporary vector chi
         chi = 0.0_dp
         int_current = 0.0_dp
         !
         ! Start loop over the occupied  states
         DO icenter = 1, nbr_center(ispin)
            !
            ! Get the Wannier center of the istate-th ground state orbital
            dk(1:3) = centers_set(ispin)%array(1:3, icenter)
            !
            ! Compute the multipole integrals for the state istate,
            ! using as reference center the corresponding Wannier center
            DO idir = 1, 9
               CALL cp_dbcsr_set(op_mom_ao(idir)%matrix, 0.0_dp)
               DO idir2 = 1, 3
                  CALL cp_dbcsr_set(op_mom_der_ao(idir, idir2)%matrix, 0.0_dp)
               ENDDO
            ENDDO
            CALL rRc_xyz_der_ao(op_mom_ao, op_mom_der_ao, qs_env, dk, order=2, &
                                minimum_image=.FALSE., soft=gapw)
            !
            ! collecte the states that belong to a given center
            CALL cp_fm_set_all(psi0, 0.0_dp)
            CALL cp_fm_set_all(psi_rxp, 0.0_dp)
            CALL cp_fm_set_all(psi_D, 0.0_dp)
            CALL cp_fm_set_all(psi_p1, 0.0_dp)
            CALL cp_fm_set_all(psi_p2, 0.0_dp)
            nstate_loc = center_list(ispin)%array(1, icenter+1)-center_list(ispin)%array(1, icenter)
            jstate = 1
            DO j = center_list(ispin)%array(1, icenter), center_list(ispin)%array(1, icenter+1)-1
               istate = center_list(ispin)%array(2, j)
               !
               ! block the states that belong to this center
               CALL cp_fm_to_fm(mo_coeff, psi0, 1, istate, jstate)
               !
               CALL cp_fm_to_fm(psi1_rxp(ispin, iB)%matrix, psi_rxp, 1, istate, jstate)
               IF (current_env%full) CALL cp_fm_to_fm(psi1_D(ispin, iB)%matrix, psi_D, 1, istate, jstate)
               !
               ! psi1_p_iiB_istate and psi1_p_iiiB_istate
               CALL cp_fm_to_fm(psi1_p(ispin, iiB)%matrix, psi_p1, 1, istate, jstate)
               CALL cp_fm_to_fm(psi1_p(ispin, iiiB)%matrix, psi_p2, 1, istate, jstate)
               !
               jstate = jstate+1
            ENDDO ! istate
            !
            ! scale the ordered mos
            IF (current_env%full) CALL cp_fm_scale_and_add(1.0_dp, psi_rxp, -1.0_dp, psi_D)
            !
            DO idir = 1, 3
               CALL set_vecp(idir, ii, iii)
               CALL cp_dbcsr_sm_fm_multiply(op_p_ao(idir)%matrix, psi_rxp, &
                                            p_rxp(idir)%matrix, ncol=nstate_loc, alpha=1.e0_dp)
               IF (iiiB .EQ. iii .OR. iiiB .EQ. ii) THEN
                  CALL cp_dbcsr_sm_fm_multiply(op_mom_ao(idir)%matrix, psi_p1, &
                                               r_p1(idir)%matrix, ncol=nstate_loc, alpha=1.e0_dp)
               ENDIF
               IF (iiB .EQ. iii .OR. iiB .EQ. ii) THEN
                  CALL cp_dbcsr_sm_fm_multiply(op_mom_ao(idir)%matrix, psi_p2, &
                                               r_p2(idir)%matrix, ncol=nstate_loc, alpha=1.e0_dp)
               ENDIF
               DO idir2 = 1, 9
                  IF (idir2 .EQ. ii .OR. idir2 .EQ. iii) THEN
                     CALL cp_dbcsr_sm_fm_multiply(op_mom_der_ao(idir2, idir)%matrix, psi_rxp, &
                                                  rr_rxp(idir2, idir)%matrix, ncol=nstate_loc, alpha=1.e0_dp)
                  ENDIF
                  !
                  IF (idir2 .EQ. ind_m2(ii, iiiB) .OR. idir2 .EQ. ind_m2(iii, iiiB) .OR. idir2 .EQ. iiiB) THEN
                     CALL cp_dbcsr_sm_fm_multiply(op_mom_der_ao(idir2, idir)%matrix, psi_p1, &
                                                  rr_p1(idir2, idir)%matrix, ncol=nstate_loc, alpha=1.e0_dp)
                  ENDIF
                  !
                  IF (idir2 .EQ. ind_m2(ii, iiB) .OR. idir2 .EQ. ind_m2(iii, iiB) .OR. idir2 .EQ. iiB) THEN
                     CALL cp_dbcsr_sm_fm_multiply(op_mom_der_ao(idir2, idir)%matrix, psi_p2, &
                                                  rr_p2(idir2, idir)%matrix, ncol=nstate_loc, alpha=1.e0_dp)
                  ENDIF
               ENDDO
            ENDDO
            !
            ! Multuply left and right by the appropriate coefficients and sum into the
            ! correct component of the chi tensor using the appropriate multiplicative factor
            ! (don't forget the occupation number)
            ! Loop over the cartesian components of the tensor
            ! The loop over the components of the external field is external, thereby
            ! only one column of the chi tensor is computed here
            DO idir = 1, 3
               chi_tmp = 0.0_dp
               int_current_tmp = 0.0_dp
               !
               ! get ii and iii
               CALL set_vecp(idir, ii, iii)
               !
               ! term: 2[C0| (r-dk)_ii |d_iii(C1(rxp-D))]-2[C0| (r-dk)_iii |d_ii(C1(rxp-D))]
               ! the factor 2 should be already included in the matrix elements
               contrib = 0.0_dp
               CALL cp_fm_trace(psi0, rr_rxp(ii, iii)%matrix, contrib)
               chi_tmp = chi_tmp+2.0_dp*contrib
               !
               contrib = 0.0_dp
               CALL cp_fm_trace(psi0, rr_rxp(iii, ii)%matrix, contrib)
               chi_tmp = chi_tmp-2.0_dp*contrib
               !
               ! correction: dk_ii*2[C0| d_iii(C1(rxp-D))] - dk_iii*2[C0| d_ii(C1(rxp-D))]
               ! factor 2 not included in the matrix elements
               contrib = 0.0_dp
               CALL cp_fm_trace(psi0, p_rxp(iii)%matrix, contrib)
               IF (.NOT. chi_pbc) chi_tmp = chi_tmp+2.0_dp*dk(ii)*contrib
               int_current_tmp = int_current_tmp+2.0_dp*contrib
               !
               contrib2 = 0.0_dp
               CALL cp_fm_trace(psi0, p_rxp(ii)%matrix, contrib2)
               IF (.NOT. chi_pbc) chi_tmp = chi_tmp-2.0_dp*dk(iii)*contrib2
               !
               ! term: -2[C0| (r-dk)_ii  (r-dk)_iiB | d_iii(C1(piiiB))]                !       +2[C0| (r-dk)_iii (r-dk)_iiB | d_ii(C1(piiiB))]
               ! the factor 2 should be already included in the matrix elements
               contrib = 0.0_dp
               idir2 = ind_m2(ii, iiB)
               CALL cp_fm_trace(psi0, rr_p2(idir2, iii)%matrix, contrib)
               chi_tmp = chi_tmp-2.0_dp*contrib
               contrib2 = 0.0_dp
               IF (iiB == iii) THEN
                  CALL cp_fm_trace(psi0, r_p2(ii)%matrix, contrib2)
                  chi_tmp = chi_tmp-contrib2
               ENDIF
               !
               contrib = 0.0_dp
               idir2 = ind_m2(iii, iiB)
               CALL cp_fm_trace(psi0, rr_p2(idir2, ii)%matrix, contrib)
               chi_tmp = chi_tmp+2.0_dp*contrib
               contrib2 = 0.0_dp
               IF (iiB == ii) THEN
                  CALL cp_fm_trace(psi0, r_p2(iii)%matrix, contrib2)
                  chi_tmp = chi_tmp+contrib2
               ENDIF
               !
               ! correction: -dk_ii * 2[C0|(r-dk)_iiB | d_iii(C1(piiiB))]                !             +dk_iii * 2[C0|(r-dk)_iiB | d_ii(C1(piiiB))]
               ! the factor 2 should be already included in the matrix elements
               ! no additional correction terms because of the orthogonality between C0 and C1
               contrib = 0.0_dp
               CALL cp_fm_trace(psi0, rr_p2(iiB, iii)%matrix, contrib)
               IF (.NOT. chi_pbc) chi_tmp = chi_tmp-2.0_dp*dk(ii)*contrib
               int_current_tmp = int_current_tmp-2.0_dp*contrib
               !
               contrib2 = 0.0_dp
               CALL cp_fm_trace(psi0, rr_p2(iiB, ii)%matrix, contrib2)
               IF (.NOT. chi_pbc) chi_tmp = chi_tmp+2.0_dp*dk(iii)*contrib2
               !
               ! term: +2[C0| (r-dk)_ii  (r-dk)_iiiB | d_iii(C1(piiB))]                !       -2[C0| (r-dk)_iii (r-dk)_iiiB | d_ii(C1(piiB))]
               ! the factor 2 should be already included in the matrix elements
               contrib = 0.0_dp
               idir2 = ind_m2(ii, iiiB)
               CALL cp_fm_trace(psi0, rr_p1(idir2, iii)%matrix, contrib)
               chi_tmp = chi_tmp+2.0_dp*contrib
               contrib2 = 0.0_dp
               IF (iiiB == iii) THEN
                  CALL cp_fm_trace(psi0, r_p1(ii)%matrix, contrib2)
                  chi_tmp = chi_tmp+contrib2
               ENDIF
               !
               contrib = 0.0_dp
               idir2 = ind_m2(iii, iiiB)
               CALL cp_fm_trace(psi0, rr_p1(idir2, ii)%matrix, contrib)
               chi_tmp = chi_tmp-2.0_dp*contrib
               contrib2 = 0.0_dp
               IF (iiiB == ii) THEN
                  CALL cp_fm_trace(psi0, r_p1(iii)%matrix, contrib2)
                  chi_tmp = chi_tmp-contrib2
               ENDIF
               !
               ! correction: +dk_ii * 2[C0|(r-dk)_iiiB | d_iii(C1(piiB))] +               !             -dk_iii * 2[C0|(r-dk)_iiiB | d_ii(C1(piiB))]
               ! the factor 2 should be already included in the matrix elements
               contrib = 0.0_dp
               CALL cp_fm_trace(psi0, rr_p1(iiiB, iii)%matrix, contrib)
               IF (.NOT. chi_pbc) chi_tmp = chi_tmp+2.0_dp*dk(ii)*contrib
               int_current_tmp = int_current_tmp+2.0_dp*contrib
               !
               contrib2 = 0.0_dp
               CALL cp_fm_trace(psi0, rr_p1(iiiB, ii)%matrix, contrib2)
               IF (.NOT. chi_pbc) chi_tmp = chi_tmp-2.0_dp*dk(iii)*contrib2
               !
               ! accumulate
               chi(idir) = chi(idir)+maxocc*chi_tmp
               int_current(iii) = int_current(iii)+int_current_tmp
            ENDDO ! idir

         ENDDO ! icenter
         !
         DO idir = 1, 3
            current_env%chi_tensor(idir, iB, ispin) = current_env%chi_tensor(idir, iB, ispin)+ &
                                                      chi(idir)
            IF (output_unit > 0) THEN
               !WRITE(output_unit,'(A,E12.6)') ' chi_'//ACHAR(119+idir)//ACHAR(119+iB)//&
               !     &                         ' = ',chi(idir)
               !WRITE(output_unit,'(A,E12.6)') ' analytic \int j_'//ACHAR(119+idir)//ACHAR(119+iB)//&
               !     &                         '(r) d^3r = ',int_current(idir)
            ENDIF
         ENDDO
         !
      ENDDO ! ispin
      !
      ! deallocate the sparse matrices
      CALL cp_dbcsr_deallocate_matrix_set(op_mom_ao)
      CALL cp_dbcsr_deallocate_matrix_set(op_mom_der_ao)
      CALL cp_dbcsr_deallocate_matrix_set(op_p_ao)

      CALL cp_fm_release(psi0)
      CALL cp_fm_release(psi_rxp)
      CALL cp_fm_release(psi_D)
      CALL cp_fm_release(psi_p1)
      CALL cp_fm_release(psi_p2)
      DO idir = 1, 3
         CALL cp_fm_release(p_rxp(idir)%matrix)
         CALL cp_fm_release(r_p1(idir)%matrix)
         CALL cp_fm_release(r_p2(idir)%matrix)
         DO idir2 = 1, 9
            CALL cp_fm_release(rr_rxp(idir2, idir)%matrix)
            CALL cp_fm_release(rr_p1(idir2, idir)%matrix)
            CALL cp_fm_release(rr_p2(idir2, idir)%matrix)
         ENDDO
      ENDDO
      DEALLOCATE (p_rxp, r_p1, r_p2, rr_rxp, rr_p1, rr_p2)

      CALL timestop(handle)

   END SUBROUTINE current_build_chi_many_centers

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
! **************************************************************************************************
   SUBROUTINE current_build_chi_one_center(current_env, qs_env, iB)
      !
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB

      CHARACTER(LEN=*), PARAMETER :: routineN = 'current_build_chi_one_center', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, idir, idir2, iiB, iiiB, ispin, jdir, jjdir, kdir, max_states, nao, natom, &
         nbr_center(2), nmo, nspins, nstates(2), output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, last_sgf
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_sizes
      LOGICAL                                            :: chi_pbc, gapw
      REAL(dp)                                           :: chi(3), contrib, dk(3), int_current(3), &
                                                            maxocc
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_2d_i_p_type), DIMENSION(:), POINTER        :: center_list
      TYPE(cp_2d_r_p_type), DIMENSION(:), POINTER        :: centers_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op_mom_ao, op_p_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: op_mom_der_ao
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: psi0_order
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: psi1_D, psi1_p, psi1_rxp
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: buf, mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_all, sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

!

      CALL timeset(routineN, handle)
      !
      NULLIFY (dft_control, mos, para_env, mo_coeff, op_mom_ao, &
               op_mom_der_ao, center_list, centers_set, &
               op_p_ao, psi1_p, psi1_rxp, psi1_D, buf, cell, psi0_order)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      mos=mos, &
                      para_env=para_env, &
                      cell=cell, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      sab_all=sab_all, &
                      sab_orb=sab_orb, &
                      dbcsr_dist=dbcsr_dist)

      nspins = dft_control%nspins
      gapw = dft_control%qs_control%gapw

      CALL get_current_env(current_env=current_env, &
                           chi_pbc=chi_pbc, &
                           nao=nao, &
                           nbr_center=nbr_center, &
                           center_list=center_list, &
                           centers_set=centers_set, &
                           psi1_p=psi1_p, &
                           psi1_rxp=psi1_rxp, &
                           psi1_D=psi1_D, &
                           nstates=nstates, &
                           psi0_order=psi0_order)
      !
      max_states = MAXVAL(nstates(1:nspins))
      !
      ! Allocate sparse matrices for dipole, quadrupole and their derivatives => 9x3
      ! Remember the derivatives are antisymmetric
      CALL cp_dbcsr_allocate_matrix_set(op_mom_ao, 9)
      CALL cp_dbcsr_allocate_matrix_set(op_mom_der_ao, 9, 3)
      !
      ! prepare for allocation
      natom = SIZE(particle_set, 1)
      ALLOCATE (first_sgf(natom))
      ALLOCATE (last_sgf(natom))
      CALL get_particle_set(particle_set, qs_kind_set, &
                            first_sgf=first_sgf, &
                            last_sgf=last_sgf)
      ALLOCATE (row_blk_sizes(natom))
      CALL convert_offsets_to_sizes(first_sgf, row_blk_sizes, last_sgf)
      DEALLOCATE (first_sgf)
      DEALLOCATE (last_sgf)
      !
      !
      ALLOCATE (op_mom_ao(1)%matrix)
      CALL cp_dbcsr_init(op_mom_ao(1)%matrix)
      CALL cp_dbcsr_create(matrix=op_mom_ao(1)%matrix, &
                           name="op_mom", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_no_symmetry, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0, mutable_work=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(op_mom_ao(1)%matrix, sab_all)

      DO idir2 = 1, 3
         ALLOCATE (op_mom_der_ao(1, idir2)%matrix)
         CALL cp_dbcsr_init(op_mom_der_ao(1, idir2)%matrix)
         CALL cp_dbcsr_copy(op_mom_der_ao(1, idir2)%matrix, op_mom_ao(1)%matrix, &
                            "op_mom_der_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir2))))
      ENDDO

      DO idir = 2, SIZE(op_mom_ao, 1)
         ALLOCATE (op_mom_ao(idir)%matrix)
         CALL cp_dbcsr_init(op_mom_ao(idir)%matrix)
         CALL cp_dbcsr_copy(op_mom_ao(idir)%matrix, op_mom_ao(1)%matrix, &
                            "op_mom_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir))))
         DO idir2 = 1, 3
            ALLOCATE (op_mom_der_ao(idir, idir2)%matrix)
            CALL cp_dbcsr_init(op_mom_der_ao(idir, idir2)%matrix)
            CALL cp_dbcsr_copy(op_mom_der_ao(idir, idir2)%matrix, op_mom_ao(1)%matrix, &
                               "op_mom_der_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir*idir2))))
         ENDDO
      ENDDO
      !
      CALL cp_dbcsr_allocate_matrix_set(op_p_ao, 3)
      ALLOCATE (op_p_ao(1)%matrix)
      CALL cp_dbcsr_init(op_p_ao(1)%matrix)
      CALL cp_dbcsr_create(matrix=op_p_ao(1)%matrix, &
                           name="op_p_ao", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_antisymmetric, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0, mutable_work=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(op_p_ao(1)%matrix, sab_orb)

      DO idir = 2, 3
         ALLOCATE (op_p_ao(idir)%matrix)
         CALL cp_dbcsr_init(op_p_ao(idir)%matrix)
         CALL cp_dbcsr_copy(op_p_ao(idir)%matrix, op_p_ao(1)%matrix, &
                            "op_p_ao"//"-"//TRIM(ADJUSTL(cp_to_string(idir))))
      ENDDO
      !
      !
      DEALLOCATE (row_blk_sizes)
      !
      !
      ! Allocate full matrices for only one vector
      mo_coeff => psi0_order(1)%matrix
      NULLIFY (tmp_fm_struct)
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                               ncol_global=max_states, para_env=para_env, &
                               context=mo_coeff%matrix_struct%context)
      CALL cp_fm_create(buf, tmp_fm_struct)
      CALL cp_fm_struct_release(tmp_fm_struct)
      !
      !
      !
      ! recompute the linear momentum matrices
      CALL build_lin_mom_matrix(qs_env, op_p_ao)
      !CALL p_xyz_ao(op_p_ao,qs_env,minimum_image=.FALSE.)
      !
      !
      ! get iiB and iiiB
      CALL set_vecp(iB, iiB, iiiB)
      DO ispin = 1, nspins
         !
         IF(.NOT.(nbr_center(ispin) == 1))CALL cp__a("qs_linres_current.F",2712)
         !
         ! get ground state MOS
         nmo = nstates(ispin)
         mo_coeff => psi0_order(ispin)%matrix
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, maxocc=maxocc)
         !
         ! Initialize the temporary vector chi
         chi = 0.0_dp
         int_current = 0.0_dp
         !
         !
         ! Get the Wannier center of the istate-th ground state orbital
         dk(1:3) = centers_set(ispin)%array(1:3, 1)
         !
         ! Compute the multipole integrals for the state istate,
         ! using as reference center the corresponding Wannier center
         DO idir = 1, 9
            CALL cp_dbcsr_set(op_mom_ao(idir)%matrix, 0.0_dp)
            DO idir2 = 1, 3
               CALL cp_dbcsr_set(op_mom_der_ao(idir, idir2)%matrix, 0.0_dp)
            ENDDO
         ENDDO
         CALL rRc_xyz_der_ao(op_mom_ao, op_mom_der_ao, qs_env, dk, order=2, &
                             minimum_image=.FALSE., soft=gapw)
         !
         !
         ! Multuply left and right by the appropriate coefficients and sum into the
         ! correct component of the chi tensor using the appropriate multiplicative factor
         ! (don't forget the occupation number)
         ! Loop over the cartesian components of the tensor
         ! The loop over the components of the external field is external, thereby
         ! only one column of the chi tensor is computed here
         DO idir = 1, 3
            !
            !
            !
            ! term: dk_ii*2[C0| d_iii(C1(rxp-D))] - dk_iii*2[C0| d_ii(C1(rxp-D))]
            IF (.NOT. chi_pbc) THEN
               CALL cp_dbcsr_sm_fm_multiply(op_p_ao(idir)%matrix, mo_coeff, &
                                            buf, ncol=nmo, alpha=1.e0_dp)
               DO jdir = 1, 3
                  DO kdir = 1, 3
                     IF (Levi_Civita(kdir, jdir, idir) .EQ. 0.0_dp) CYCLE
                     CALL cp_fm_trace(buf, psi1_rxp(ispin, iB)%matrix, contrib)
                     chi(kdir) = chi(kdir)-Levi_Civita(kdir, jdir, idir)*2.0_dp*dk(jdir)*contrib
                  ENDDO
               ENDDO
            ENDIF
            !
            !
            !
            ! term: 2[C0| (r-dk)_ii |d_iii(C1(rxp-D))]-2[C0| (r-dk)_iii |d_ii(C1(rxp-D))]
            ! and
            ! term: -dk_ii * 2[C0|(r-dk)_iiB | d_iii(C1(piiiB))] +
            !       +dk_iii * 2[C0|(r-dk)_iiB | d_ii(C1(piiiB))]
            ! and
            ! term: +dk_ii * 2[C0|(r-dk)_iiiB | d_iii(C1(piiB))] +
            !       -dk_iii * 2[C0|(r-dk)_iiiB | d_ii(C1(piiB))]
            DO jdir = 1, 3
               CALL cp_dbcsr_sm_fm_multiply(op_mom_der_ao(jdir, idir)%matrix, mo_coeff, &
                                            buf, ncol=nmo, alpha=1.e0_dp)
               DO kdir = 1, 3
                  IF (Levi_Civita(kdir, jdir, idir) .EQ. 0.0_dp) CYCLE
                  CALL cp_fm_trace(buf, psi1_rxp(ispin, iB)%matrix, contrib)
                  chi(kdir) = chi(kdir)-Levi_Civita(kdir, jdir, idir)*2.0_dp*contrib
               ENDDO
               !
               IF (.NOT. chi_pbc) THEN
                  IF (jdir .EQ. iiB) THEN
                     DO jjdir = 1, 3
                        DO kdir = 1, 3
                           IF (Levi_Civita(kdir, jjdir, idir) .EQ. 0.0_dp) CYCLE
                           CALL cp_fm_trace(buf, psi1_p(ispin, iiiB)%matrix, contrib)
                           chi(kdir) = chi(kdir)+Levi_Civita(kdir, jjdir, idir)*2.0_dp*dk(jjdir)*contrib
                        ENDDO
                     ENDDO
                  ENDIF
                  !
                  IF (jdir .EQ. iiiB) THEN
                     DO jjdir = 1, 3
                        DO kdir = 1, 3
                           IF (Levi_Civita(kdir, jjdir, idir) .EQ. 0.0_dp) CYCLE
                           CALL cp_fm_trace(buf, psi1_p(ispin, iiB)%matrix, contrib)
                           chi(kdir) = chi(kdir)-Levi_Civita(kdir, jjdir, idir)*2.0_dp*dk(jjdir)*contrib
                        ENDDO
                     ENDDO
                  ENDIF
               ENDIF
            ENDDO
            !
            !
            !
            ! term1: -2[C0| (r-dk)_ii  (r-dk)_iiB | d_iii(C1(piiiB))] +
            !        +2[C0| (r-dk)_iii (r-dk)_iiB | d_ii(C1(piiiB))]
            ! and
            ! term1: +2[C0| (r-dk)_ii  (r-dk)_iiiB | d_iii(C1(piiB))] +
            !        -2[C0| (r-dk)_iii (r-dk)_iiiB | d_ii(C1(piiB))]
            ! HERE THERE IS ONE EXTRA MULTIPLY
            DO jdir = 1, 3
               CALL cp_dbcsr_sm_fm_multiply(op_mom_der_ao(ind_m2(jdir, iiB), idir)%matrix, mo_coeff, &
                                            buf, ncol=nmo, alpha=1.e0_dp)
               DO kdir = 1, 3
                  IF (Levi_Civita(kdir, jdir, idir) .EQ. 0.0_dp) CYCLE
                  CALL cp_fm_trace(buf, psi1_p(ispin, iiiB)%matrix, contrib)
                  chi(kdir) = chi(kdir)+Levi_Civita(kdir, jdir, idir)*2.0_dp*contrib
               ENDDO
               !
               CALL cp_dbcsr_sm_fm_multiply(op_mom_der_ao(ind_m2(jdir, iiiB), idir)%matrix, mo_coeff, &
                                            buf, ncol=nmo, alpha=1.e0_dp)
               DO kdir = 1, 3
                  IF (Levi_Civita(kdir, jdir, idir) .EQ. 0.0_dp) CYCLE
                  CALL cp_fm_trace(buf, psi1_p(ispin, iiB)%matrix, contrib)
                  chi(kdir) = chi(kdir)-Levi_Civita(kdir, jdir, idir)*2.0_dp*contrib
               ENDDO
            ENDDO
            !
            !
            !
            ! term2: -2[C0| (r-dk)_ii  (r-dk)_iiB | d_iii(C1(piiiB))] +
            !        +2[C0| (r-dk)_iii (r-dk)_iiB | d_ii(C1(piiiB))]
            ! and
            ! term2: +2[C0| (r-dk)_ii  (r-dk)_iiiB | d_iii(C1(piiB))] +
            !        -2[C0| (r-dk)_iii (r-dk)_iiiB | d_ii(C1(piiB))]
            CALL cp_dbcsr_sm_fm_multiply(op_mom_ao(idir)%matrix, mo_coeff, &
                                         buf, ncol=nmo, alpha=1.e0_dp)
            DO jdir = 1, 3
               DO kdir = 1, 3
                  IF (Levi_Civita(kdir, idir, jdir) .EQ. 0.0_dp) CYCLE
                  IF (iiB == jdir) THEN
                     CALL cp_fm_trace(buf, psi1_p(ispin, iiiB)%matrix, contrib)
                     chi(kdir) = chi(kdir)+Levi_Civita(kdir, idir, jdir)*contrib
                  ENDIF
               ENDDO
            ENDDO
            !
            DO jdir = 1, 3
               DO kdir = 1, 3
                  IF (Levi_Civita(kdir, idir, jdir) .EQ. 0.0_dp) CYCLE
                  IF (iiiB == jdir) THEN
                     CALL cp_fm_trace(buf, psi1_p(ispin, iiB)%matrix, contrib)
                     chi(kdir) = chi(kdir)-Levi_Civita(kdir, idir, jdir)*contrib
                  ENDIF
                  !
               ENDDO
            ENDDO
            !
            !
            !
            !
         ENDDO ! idir
         !
         DO idir = 1, 3
            current_env%chi_tensor(idir, iB, ispin) = current_env%chi_tensor(idir, iB, ispin)+ &
                                                      maxocc*chi(idir)
            IF (output_unit > 0) THEN
               !WRITE(output_unit,'(A,E12.6)') ' chi_'//ACHAR(119+idir)//ACHAR(119+iB)//&
               !     &                         ' = ',maxocc * chi(idir)
            ENDIF
         ENDDO
         !
      ENDDO ! ispin
      !
      ! deallocate the sparse matrices
      CALL cp_dbcsr_deallocate_matrix_set(op_mom_ao)
      CALL cp_dbcsr_deallocate_matrix_set(op_mom_der_ao)
      CALL cp_dbcsr_deallocate_matrix_set(op_p_ao)
      CALL cp_fm_release(buf)

      CALL timestop(handle)

   END SUBROUTINE current_build_chi_one_center

END MODULE qs_linres_current
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief from the response current density calculates the shift tensor
!>      and the susceptibility
!> \par History
!>      created 02-2006 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_nmr_shift
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: gaussi,&
                                              twopi
   USE mathlib,                         ONLY: diamat_all
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_spline_utils,                 ONLY: Eval_Interp_Spl3_pbc,&
                                              find_coeffs,&
                                              pw_spline_do_precond,&
                                              pw_spline_precond_create,&
                                              pw_spline_precond_release,&
                                              pw_spline_precond_set_kind,&
                                              pw_spline_precond_type,&
                                              spl3_pbc
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_linres_nmr_epr_common_utils,  ONLY: mult_G_ov_G2_grid
   USE qs_linres_op,                    ONLY: fac_vecp,&
                                              set_vecp,&
                                              set_vecp_rev
   USE qs_linres_types,                 ONLY: current_env_type,&
                                              get_current_env,&
                                              get_nmr_env,&
                                              jrho_atom_type,&
                                              nmr_env_type
   USE qs_rho_types,                    ONLY: qs_rho_get
   USE realspace_grid_types,            ONLY: realspace_grid_desc_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Public subroutines ***
   PUBLIC :: nmr_shift_print, &
             nmr_shift

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_nmr_shift'

! **************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param nmr_env ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
! **************************************************************************************************
   SUBROUTINE nmr_shift(nmr_env, current_env, qs_env, iB)

      TYPE(nmr_env_type)                                 :: nmr_env
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iB

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nmr_shift', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idir, idir2, idir3, iiB, iiiB, &
                                                            ispin, natom, nspins
      LOGICAL                                            :: gapw, interpolate_shift
      REAL(dp)                                           :: scale_fac
      REAL(dp), DIMENSION(:, :, :), POINTER              :: chemical_shift, chemical_shift_loc, &
                                                            chemical_shift_nics, &
                                                            chemical_shift_nics_loc
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: pw_gspace_work, shift_pw_rspace
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: jrho1_g
      TYPE(pw_p_type), DIMENSION(:, :), POINTER          :: shift_pw_gspace
      TYPE(pw_p_type), POINTER                           :: jrho_gspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(realspace_grid_desc_type), POINTER            :: auxbas_rs_desc
      TYPE(section_vals_type), POINTER                   :: nmr_section

      CALL timeset(routineN, handle)

      NULLIFY (chemical_shift, chemical_shift_loc, chemical_shift_nics, chemical_shift_nics_loc, &
               cell, dft_control, pw_env, auxbas_rs_desc, auxbas_pw_pool, jrho_gspace, &
               pw_pools, particle_set, jrho1_g)

      CALL get_qs_env(qs_env=qs_env, cell=cell, dft_control=dft_control, &
                      particle_set=particle_set)

      gapw = dft_control%qs_control%gapw
      natom = SIZE(particle_set, 1)
      nspins = dft_control%nspins

      CALL get_nmr_env(nmr_env=nmr_env, chemical_shift=chemical_shift, &
                       chemical_shift_loc=chemical_shift_loc, &
                       chemical_shift_nics=chemical_shift_nics, &
                       chemical_shift_nics_loc=chemical_shift_nics_loc, &
                       interpolate_shift=interpolate_shift)

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      CALL pw_env_get(pw_env, auxbas_rs_desc=auxbas_rs_desc, &
                      auxbas_pw_pool=auxbas_pw_pool, pw_pools=pw_pools)
      !
      !
      nmr_section => section_vals_get_subs_vals(qs_env%input, &
           & "PROPERTIES%LINRES%NMR")
      !
      ! Initialize
      ! Allocate grids for the calculation of jrho and the shift
      ALLOCATE (shift_pw_gspace(3, nspins))
      DO ispin = 1, nspins
         DO idir = 1, 3
            CALL pw_pool_create_pw(auxbas_pw_pool, shift_pw_gspace(idir, ispin)%pw, &
                                   use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            CALL pw_zero(shift_pw_gspace(idir, ispin)%pw)
         ENDDO
      ENDDO
      !
      !
      CALL set_vecp(iB, iiB, iiiB)
      !
      CALL pw_pool_create_pw(auxbas_pw_pool, pw_gspace_work%pw, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL pw_zero(pw_gspace_work%pw)
      DO ispin = 1, nspins
         !
         DO idir = 1, 3
            CALL qs_rho_get(current_env%jrho1_set(idir)%rho, rho_g=jrho1_g)
            jrho_gspace => jrho1_g(ispin)
            ! Field gradient
            ! loop over the Gvec  components: x,y,z
            DO idir2 = 1, 3
               IF (idir /= idir2) THEN
                  ! in reciprocal space multiply (G_idir2(i)/G(i)^2)J_(idir)(G(i))
                  CALL mult_G_ov_G2_grid(cell, auxbas_pw_pool, jrho_gspace, &
                                         pw_gspace_work, idir2, 0.0_dp)
                  !
                  ! scale and add to the correct component of the shift column
                  CALL set_vecp_rev(idir, idir2, idir3)
                  scale_fac = fac_vecp(idir3, idir2, idir)
                  CALL pw_axpy(pw_gspace_work%pw, shift_pw_gspace(idir3, ispin)%pw, scale_fac)
               ENDIF
            ENDDO
            !
         ENDDO ! idir
      ENDDO ! ispin
      !
      CALL pw_pool_give_back_pw(auxbas_pw_pool, pw_gspace_work%pw)
      !
      ! compute shildings
      IF (interpolate_shift) THEN
         CALL pw_pool_create_pw(auxbas_pw_pool, shift_pw_rspace%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)
         DO ispin = 1, nspins
            DO idir = 1, 3
               ! Here first G->R and then interpolation to get the shifts.
               ! The interpolation doesnt work in parallel yet.
               ! The choice between both methods should be left to the user.
               CALL pw_transfer(shift_pw_gspace(idir, ispin)%pw, shift_pw_rspace%pw)
               CALL interpolate_shift_pwgrid(nmr_env, pw_env, particle_set, cell, shift_pw_rspace, &
                                             iB, idir, nmr_section)
            ENDDO
         ENDDO
         CALL pw_pool_give_back_pw(auxbas_pw_pool, shift_pw_rspace%pw)
      ELSE
         DO ispin = 1, nspins
            DO idir = 1, 3
               ! Here the shifts are computed from summation of the coeff on the G-grip .
               CALL gsum_shift_pwgrid(nmr_env, particle_set, cell, &
                                      shift_pw_gspace(idir, ispin), iB, idir)
            ENDDO
         ENDDO
      ENDIF
      !
      IF (gapw) THEN
         DO idir = 1, 3
            ! Finally the radial functions are multiplied by the YLM and properly summed
            ! The resulting array is J on the local grid. One array per atom.
            ! Local contributions by numerical integration over the spherical grids
            CALL nmr_shift_gapw(nmr_env, current_env, qs_env, iB, idir)
         ENDDO ! idir
      ENDIF
      !
      ! Dellocate grids for the calculation of jrho and the shift
      DO ispin = 1, nspins
         DO idir = 1, 3
            CALL pw_pool_give_back_pw(auxbas_pw_pool, shift_pw_gspace(idir, ispin)%pw)
         ENDDO
      ENDDO
      DEALLOCATE (shift_pw_gspace)
      !
      ! Finalize
      CALL timestop(handle)
      !
   END SUBROUTINE nmr_shift

! **************************************************************************************************
!> \brief ...
!> \param nmr_env ...
!> \param current_env ...
!> \param qs_env ...
!> \param iB ...
!> \param idir ...
! **************************************************************************************************
   SUBROUTINE nmr_shift_gapw(nmr_env, current_env, qs_env, iB, idir)

      TYPE(nmr_env_type)                                 :: nmr_env
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: IB, idir

      CHARACTER(len=*), PARAMETER :: routineN = 'nmr_shift_gapw', routineP = moduleN//':'//routineN

      INTEGER :: handle, ia, iat, iatom, idir2_1, idir3_1, ikind, ir, ira, ispin, j, jatom, mepos, &
         n_nics, na, natom, natom_local, natom_tot, nkind, nnics_local, nr, nra, nspins, num_pe, &
         output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: list_j, list_nics_j
      INTEGER, DIMENSION(2)                              :: bo
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: do_nics, paw_atom
      REAL(dp)                                           :: ddiff, dist, dum, itegrated_jrho, &
                                                            r_jatom(3), rdiff(3), rij(3), s_1, &
                                                            s_2, scale_fac_1, shift_gapw_radius
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: j_grid
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: cs_loc_tmp, cs_nics_loc_tmp, dist_ij, &
                                                            dist_nics_ij, r_grid
      REAL(dp), DIMENSION(:, :), POINTER                 :: jrho_h_grid, jrho_s_grid, r_nics
      REAL(dp), DIMENSION(:, :, :), POINTER              :: chemical_shift_loc, &
                                                            chemical_shift_nics_loc
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      TYPE(jrho_atom_type), POINTER                      :: jrho1_atom
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)
      !
      NULLIFY (atomic_kind_set, qs_kind_set, cell, dft_control, para_env, particle_set, &
               chemical_shift_loc, chemical_shift_nics_loc, jrho1_atom_set, &
               jrho1_atom, r_nics, jrho_h_grid, jrho_s_grid, &
               atom_list, grid_atom, harmonics, logger)
      !
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      !
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      particle_set=particle_set)

      CALL get_nmr_env(nmr_env=nmr_env, &
                       chemical_shift_loc=chemical_shift_loc, &
                       chemical_shift_nics_loc=chemical_shift_nics_loc, &
                       shift_gapw_radius=shift_gapw_radius, &
                       n_nics=n_nics, &
                       r_nics=r_nics, &
                       do_nics=do_nics)

      CALL get_current_env(current_env=current_env, &
                           jrho1_atom_set=jrho1_atom_set)
      !
      nkind = SIZE(atomic_kind_set, 1)
      natom_tot = SIZE(particle_set, 1)
      nspins = dft_control%nspins
      itegrated_jrho = 0.0_dp
      !
      idir2_1 = MODULO(idir, 3)+1
      idir3_1 = MODULO(idir+1, 3)+1
      scale_fac_1 = fac_vecp(idir3_1, idir2_1, idir)
      !
      ALLOCATE (cs_loc_tmp(3, natom_tot), list_j(natom_tot), &
                dist_ij(3, natom_tot))
      cs_loc_tmp = 0.0_dp
      IF (do_nics) THEN
         ALLOCATE (cs_nics_loc_tmp(3, n_nics), list_nics_j(n_nics), &
                   dist_nics_ij(3, n_nics))
         cs_nics_loc_tmp = 0.0_dp
      ENDIF
      !
      ! Loop over atoms to collocate the current on each atomic grid, JA
      ! Per each JA, loop over the points where the shift needs to be computed
      DO ikind = 1, nkind

         NULLIFY (atom_list, grid_atom, harmonics)
         CALL get_atomic_kind(atomic_kind_set(ikind), &
                              atom_list=atom_list, &
                              natom=natom)

         CALL get_qs_kind(qs_kind_set(ikind), &
                          paw_atom=paw_atom, &
                          harmonics=harmonics, &
                          grid_atom=grid_atom)
         !
         na = grid_atom%ng_sphere
         nr = grid_atom%nr
         nra = nr*na
         ALLOCATE (r_grid(3, nra), j_grid(nra))
         ira = 1
         DO ia = 1, na
            DO ir = 1, nr
               r_grid(:, ira) = grid_atom%rad(ir)*harmonics%a(:, ia)
               ira = ira+1
            ENDDO
         ENDDO
         !
         ! Quick cycle if needed
         IF (paw_atom) THEN
            !
            ! Distribute the atoms of this kind
            num_pe = para_env%num_pe
            mepos = para_env%mepos
            bo = get_limit(natom, num_pe, mepos)
            !
            DO iat = bo(1), bo(2)
               iatom = atom_list(iat)
               !
               ! find all the atoms within the radius
               natom_local = 0
               DO jatom = 1, natom_tot
                  rij(:) = pbc(particle_set(iatom)%r, particle_set(jatom)%r, cell)
                  dist = SQRT(rij(1)*rij(1)+rij(2)*rij(2)+rij(3)*rij(3))
                  IF (dist .LE. shift_gapw_radius) THEN
                     natom_local = natom_local+1
                     list_j(natom_local) = jatom
                     dist_ij(:, natom_local) = rij(:)
                  ENDIF
               ENDDO
               !
               ! ... also for nics
               IF (do_nics) THEN
                  nnics_local = 0
                  DO jatom = 1, n_nics
                     r_jatom(:) = r_nics(:, jatom)
                     rij(:) = pbc(particle_set(iatom)%r, r_jatom, cell)
                     dist = SQRT(rij(1)*rij(1)+rij(2)*rij(2)+rij(3)*rij(3))
                     IF (dist .LE. shift_gapw_radius) THEN
                        nnics_local = nnics_local+1
                        list_nics_j(nnics_local) = jatom
                        dist_nics_ij(:, nnics_local) = rij(:)
                     ENDIF
                  ENDDO
               ENDIF
               !
               NULLIFY (jrho1_atom, jrho_h_grid, jrho_s_grid)
               jrho1_atom => jrho1_atom_set(iatom)
               !
               DO ispin = 1, nspins
                  jrho_h_grid => jrho1_atom%jrho_vec_rad_h(idir, ispin)%r_coef
                  jrho_s_grid => jrho1_atom%jrho_vec_rad_s(idir, ispin)%r_coef
                  !
                  ! loop over the atoms neighbors of iatom in terms of the current density
                  ! for each compute the contribution to the shift coming from the
                  ! local current density at iatom
                  ira = 1
                  DO ia = 1, na
                     DO ir = 1, nr
                        j_grid(ira) = (jrho_h_grid(ir, ia)-jrho_s_grid(ir, ia))*grid_atom%weight(ia, ir)
                        itegrated_jrho = itegrated_jrho+j_grid(ira)
                        ira = ira+1
                     ENDDO
                  ENDDO
                  !
                  DO j = 1, natom_local
                     jatom = list_j(j)
                     rij(:) = dist_ij(:, j)
                     !
                     s_1 = 0.0_dp
                     s_2 = 0.0_dp
                     DO ira = 1, nra
                        !
                        rdiff(:) = rij(:)-r_grid(:, ira)
                        ddiff = SQRT(rdiff(1)*rdiff(1)+rdiff(2)*rdiff(2)+rdiff(3)*rdiff(3))
                        IF (ddiff .GT. 1.0E-12_dp) THEN
                           dum = scale_fac_1*j_grid(ira)/(ddiff*ddiff*ddiff)
                           s_1 = s_1+rdiff(idir2_1)*dum
                           s_2 = s_2+rdiff(idir3_1)*dum
                        ENDIF ! ddiff
                     ENDDO ! ira
                     cs_loc_tmp(idir3_1, jatom) = cs_loc_tmp(idir3_1, jatom)+s_1
                     cs_loc_tmp(idir2_1, jatom) = cs_loc_tmp(idir2_1, jatom)-s_2
                  ENDDO ! j
                  !
                  IF (do_nics) THEN
                     DO j = 1, nnics_local
                        jatom = list_nics_j(j)
                        rij(:) = dist_nics_ij(:, j)
                        !
                        s_1 = 0.0_dp
                        s_2 = 0.0_dp
                        DO ira = 1, nra
                           !
                           rdiff(:) = rij(:)-r_grid(:, ira)
                           ddiff = SQRT(rdiff(1)*rdiff(1)+rdiff(2)*rdiff(2)+rdiff(3)*rdiff(3))
                           IF (ddiff .GT. 1.0E-12_dp) THEN
                              dum = scale_fac_1*j_grid(ira)/(ddiff*ddiff*ddiff)
                              s_1 = s_1+rdiff(idir2_1)*dum
                              s_2 = s_2+rdiff(idir3_1)*dum
                           ENDIF ! ddiff
                        ENDDO ! ira
                        cs_nics_loc_tmp(idir3_1, jatom) = cs_nics_loc_tmp(idir3_1, jatom)+s_1
                        cs_nics_loc_tmp(idir2_1, jatom) = cs_nics_loc_tmp(idir2_1, jatom)-s_2
                     ENDDO ! j
                  ENDIF ! do_nics
               ENDDO ! ispin
            ENDDO ! iat
         ENDIF
         DEALLOCATE (r_grid, j_grid)
      ENDDO ! ikind
      !
      !
      CALL mp_sum(itegrated_jrho, para_env%group)
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E24.16)') 'Integrated local j_'&
              &//ACHAR(idir+119)//ACHAR(iB+119)//'(r)=', itegrated_jrho
      ENDIF
      !
      CALL mp_sum(cs_loc_tmp, para_env%group)
      chemical_shift_loc(:, iB, :) = chemical_shift_loc(:, iB, :) &
           & -nmr_env%shift_factor_gapw*cs_loc_tmp(:, :)/2.0_dp
      !
      DEALLOCATE (cs_loc_tmp, list_j, dist_ij)
      !
      IF (do_nics) THEN
         CALL mp_sum(cs_nics_loc_tmp, para_env%group)
         chemical_shift_nics_loc(:, iB, :) = chemical_shift_nics_loc(:, iB, :) &
              & -nmr_env%shift_factor_gapw*cs_nics_loc_tmp(:, :)/2.0_dp
         !
         DEALLOCATE (cs_nics_loc_tmp, list_nics_j, dist_nics_ij)
      ENDIF
      !
      CALL timestop(handle)
      !
   END SUBROUTINE nmr_shift_gapw

! **************************************************************************************************
!> \brief interpolate the shift calculated on the PW grid in order to ger
!>       the value on arbitrary points in real space
!> \param nmr_env to get the shift tensor and the list of additional points
!> \param pw_env ...
!> \param particle_set for the atomic position
!> \param cell to take into account the pbs, and to have the volume
!> \param shift_pw_rspace specific component of the shift tensor on the pw grid
!> \param i_B component of the magnetic field for which the shift is calculated (row)
!> \param idir component of the vector \int_{r}[ ((r-r') x j(r))/|r-r'|^3 ] = Bind(r')
!> \param nmr_section ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE interpolate_shift_pwgrid(nmr_env, pw_env, particle_set, cell, shift_pw_rspace, &
                                       i_B, idir, nmr_section)

      TYPE(nmr_env_type)                                 :: nmr_env
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(pw_p_type)                                    :: shift_pw_rspace
      INTEGER, INTENT(IN)                                :: i_B, idir
      TYPE(section_vals_type), POINTER                   :: nmr_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'interpolate_shift_pwgrid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: aint_precond, handle, iat, iatom, &
                                                            max_iter, n_nics, natom, precond_kind
      INTEGER, DIMENSION(:), POINTER                     :: cs_atom_list
      LOGICAL                                            :: do_nics, success
      REAL(dp)                                           :: eps_r, eps_x, R_iatom(3), ra(3), &
                                                            shift_val
      REAL(dp), DIMENSION(:, :), POINTER                 :: r_nics
      REAL(dp), DIMENSION(:, :, :), POINTER              :: chemical_shift, chemical_shift_nics
      TYPE(pw_p_type)                                    :: shiftspl
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(pw_spline_precond_type), POINTER              :: precond
      TYPE(section_vals_type), POINTER                   :: interp_section

!

      CALL timeset(routineN, handle)
      !
      NULLIFY (interp_section)
      NULLIFY (auxbas_pw_pool, precond)
      NULLIFY (cs_atom_list, chemical_shift, chemical_shift_nics, r_nics)

      IF(.NOT.(ASSOCIATED(shift_pw_rspace%pw)))CALL cp__a("qs_linres_nmr_shift.F",542)

      interp_section => section_vals_get_subs_vals(nmr_section, &
                                                   "INTERPOLATOR")
      CALL section_vals_val_get(interp_section, "aint_precond", &
                                i_val=aint_precond)
      CALL section_vals_val_get(interp_section, "precond", i_val=precond_kind)
      CALL section_vals_val_get(interp_section, "max_iter", i_val=max_iter)
      CALL section_vals_val_get(interp_section, "eps_r", r_val=eps_r)
      CALL section_vals_val_get(interp_section, "eps_x", r_val=eps_x)

      ! calculate spline coefficients
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
      CALL pw_pool_create_pw(auxbas_pw_pool, shiftspl%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL pw_spline_precond_create(precond, precond_kind=aint_precond, &
                                    pool=auxbas_pw_pool, pbc=.TRUE., transpose=.FALSE.)
      CALL pw_spline_do_precond(precond, shift_pw_rspace%pw, shiftspl%pw)
      CALL pw_spline_precond_set_kind(precond, precond_kind)
      success = find_coeffs(values=shift_pw_rspace%pw, coeffs=shiftspl%pw, &
                            linOp=spl3_pbc, preconditioner=precond, pool=auxbas_pw_pool, &
                            eps_r=eps_r, eps_x=eps_x, max_iter=max_iter)
      IF(.NOT.(success))CALL cp__a("qs_linres_nmr_shift.F",565)
      CALL pw_spline_precond_release(precond)

      CALL get_nmr_env(nmr_env=nmr_env, cs_atom_list=cs_atom_list, &
                       chemical_shift=chemical_shift, &
                       chemical_shift_nics=chemical_shift_nics, &
                       n_nics=n_nics, r_nics=r_nics, &
                       do_nics=do_nics)

      IF (ASSOCIATED(cs_atom_list)) THEN
         natom = SIZE(cs_atom_list, 1)
      ELSE
         natom = -1
      ENDIF

      DO iat = 1, natom
         iatom = cs_atom_list(iat)
         R_iatom = pbc(particle_set(iatom)%r, cell)
         shift_val = Eval_Interp_Spl3_pbc(R_iatom, shiftspl%pw)
         chemical_shift(idir, i_B, iatom) = chemical_shift(idir, i_B, iatom)+ &
                                            nmr_env%shift_factor*twopi**2*shift_val
      END DO

      IF (do_nics) THEN
         DO iatom = 1, n_nics
            ra(:) = r_nics(:, iatom)
            R_iatom = pbc(ra, cell)
            shift_val = Eval_Interp_Spl3_pbc(R_iatom, shiftspl%pw)
            chemical_shift_nics(idir, i_B, iatom) = chemical_shift_nics(idir, i_B, iatom)+ &
                                                    nmr_env%shift_factor*twopi**2*shift_val
         END DO
      END IF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, shiftspl%pw)

      !
      CALL timestop(handle)
      !
   END SUBROUTINE interpolate_shift_pwgrid

! **************************************************************************************************
!> \brief ...
!> \param nmr_env ...
!> \param particle_set ...
!> \param cell ...
!> \param shift_pw_gspace ...
!> \param i_B ...
!> \param idir ...
! **************************************************************************************************
   SUBROUTINE gsum_shift_pwgrid(nmr_env, particle_set, cell, shift_pw_gspace, &
                                i_B, idir)
      TYPE(nmr_env_type)                                 :: nmr_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(pw_p_type)                                    :: shift_pw_gspace
      INTEGER, INTENT(IN)                                :: i_B, idir

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gsum_shift_pwgrid', &
         routineP = moduleN//':'//routineN

      COMPLEX(dp)                                        :: cplx
      INTEGER                                            :: handle, iat, iatom, n_nics, natom
      INTEGER, DIMENSION(:), POINTER                     :: cs_atom_list
      LOGICAL                                            :: do_nics
      REAL(dp)                                           :: R_iatom(3), ra(3)
      REAL(dp), DIMENSION(:, :), POINTER                 :: r_nics
      REAL(dp), DIMENSION(:, :, :), POINTER              :: chemical_shift, chemical_shift_nics

!
!

      CALL timeset(routineN, handle)
      !
      NULLIFY (cs_atom_list, chemical_shift, chemical_shift_nics, r_nics)
      IF(.NOT.(ASSOCIATED(shift_pw_gspace%pw)))CALL cp__a("qs_linres_nmr_shift.F",639)
      !
      CALL get_nmr_env(nmr_env=nmr_env, cs_atom_list=cs_atom_list, &
                       chemical_shift=chemical_shift, &
                       chemical_shift_nics=chemical_shift_nics, &
                       n_nics=n_nics, r_nics=r_nics, do_nics=do_nics)
      !
      IF (ASSOCIATED(cs_atom_list)) THEN
         natom = SIZE(cs_atom_list, 1)
      ELSE
         natom = -1
      ENDIF
      !
      ! compute the chemical shift
      DO iat = 1, natom
         iatom = cs_atom_list(iat)
         R_iatom = pbc(particle_set(iatom)%r, cell)
         CALL gsumr(R_iatom, shift_pw_gspace%pw, cplx)
         chemical_shift(idir, i_B, iatom) = chemical_shift(idir, i_B, iatom)+ &
                                            nmr_env%shift_factor*twopi**2*REAL(cplx, dp)
      ENDDO
      !
      ! compute nics
      IF (do_nics) THEN
         DO iat = 1, n_nics
            ra = pbc(r_nics(:, iat), cell)
            CALL gsumr(ra, shift_pw_gspace%pw, cplx)
            chemical_shift_nics(idir, i_B, iat) = chemical_shift_nics(idir, i_B, iat)+ &
                                                  nmr_env%shift_factor*twopi**2*REAL(cplx, dp)
         ENDDO
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE gsum_shift_pwgrid

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param pw ...
!> \param cplx ...
! **************************************************************************************************
   SUBROUTINE gsumr(r, pw, cplx)
      REAL(dp), INTENT(IN)                               :: r(3)
      TYPE(pw_type), POINTER                             :: pw
      COMPLEX(dp)                                        :: cplx

      COMPLEX(dp)                                        :: rg
      INTEGER                                            :: ig
      TYPE(pw_grid_type), POINTER                        :: grid

      grid => pw%pw_grid
      cplx = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      DO ig = grid%first_gne0, grid%ngpts_cut_local
         rg = (grid%g(1, ig)*r(1)+grid%g(2, ig)*r(2)+grid%g(3, ig)*r(3))*gaussi
         cplx = cplx+pw%cc(ig)*EXP(rg)
      ENDDO
      IF (grid%have_g0) cplx = cplx+pw%cc(1)
      CALL mp_sum(cplx, grid%para%group)
   END SUBROUTINE gsumr

! **************************************************************************************************
!> \brief Shielding tensor and Chi are printed into a file
!>       if required from input
!>       It is possible to print only for a subset of atoms or
!>       or points in non-ionic positions
!> \param nmr_env ...
!> \param current_env ...
!> \param qs_env ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE nmr_shift_print(nmr_env, current_env, qs_env)
      TYPE(nmr_env_type)                                 :: nmr_env
      TYPE(current_env_type)                             :: current_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'nmr_shift_print', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=default_string_length)               :: name, title
      INTEGER                                            :: iatom, ir, n_nics, nat_print, natom, &
                                                            output_unit, unit_atoms, unit_nics
      INTEGER, DIMENSION(:), POINTER                     :: cs_atom_list
      LOGICAL                                            :: do_nics, gapw
      REAL(dp) :: chi_aniso, chi_iso, chi_sym_tot(3, 3), chi_tensor(3, 3, 2), &
         chi_tensor_loc(3, 3, 2), chi_tensor_loc_tmp(3, 3), chi_tensor_tmp(3, 3), chi_tmp(3, 3), &
         eig(3), rpos(3), shift_aniso, shift_iso, shift_sym_tot(3, 3)
      REAL(dp), DIMENSION(:, :), POINTER                 :: r_nics
      REAL(dp), DIMENSION(:, :, :), POINTER              :: cs, cs_loc, cs_nics, cs_nics_loc, &
                                                            cs_nics_tot, cs_tot
      REAL(dp), EXTERNAL                                 :: DDOT
      TYPE(atomic_kind_type), POINTER                    :: atom_kind
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: nmr_section

      NULLIFY (cs, cs_nics, r_nics, cs_loc, cs_nics_loc, logger, particle_set, atom_kind, dft_control)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      nmr_section => section_vals_get_subs_vals(qs_env%input, &
                                                "PROPERTIES%LINRES%NMR")

      CALL get_nmr_env(nmr_env=nmr_env, &
                       chemical_shift=cs, &
                       chemical_shift_nics=cs_nics, &
                       chemical_shift_loc=cs_loc, &
                       chemical_shift_nics_loc=cs_nics_loc, &
                       cs_atom_list=cs_atom_list, &
                       n_nics=n_nics, &
                       r_nics=r_nics, &
                       do_nics=do_nics)
      !
      CALL get_current_env(current_env=current_env, &
                           chi_tensor=chi_tensor, &
                           chi_tensor_loc=chi_tensor_loc)
      !
      ! multiply by the appropriate factor
      chi_tensor_tmp(:, :) = 0.0_dp
      chi_tensor_loc_tmp(:, :) = 0.0_dp
      chi_tensor_tmp(:, :) = (chi_tensor(:, :, 1)+chi_tensor(:, :, 2))*nmr_env%chi_factor
      !chi_tensor_loc_tmp(:,:) = (chi_tensor_loc(:,:,1) + chi_tensor_loc(:,:,2)) * here there is another factor
      !
      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      particle_set=particle_set)

      natom = SIZE(particle_set, 1)
      gapw = dft_control%qs_control%gapw
      nat_print = SIZE(cs_atom_list, 1)

      ALLOCATE (cs_tot(3, 3, nat_print))
      IF (do_nics) THEN
         ALLOCATE (cs_nics_tot(3, 3, n_nics))
      ENDIF
      ! Finalize Chi calculation
      ! Symmetrize
      chi_sym_tot(:, :) = (chi_tensor_tmp(:, :)+TRANSPOSE(chi_tensor_tmp(:, :)))/2.0_dp
      IF (gapw) THEN
         chi_sym_tot(:, :) = chi_sym_tot(:, :) &
              & +(chi_tensor_loc_tmp(:, :)+TRANSPOSE(chi_tensor_loc_tmp(:, :)))/2.0_dp
      ENDIF
      chi_tmp(:, :) = chi_sym_tot(:, :)
      CALL diamat_all(chi_tmp, eig)
      chi_iso = (eig(1)+eig(2)+eig(3))/3.0_dp
      chi_aniso = eig(3)-(eig(2)+eig(1))/2.0_dp
      !
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E14.6)') 'CheckSum Chi =', &
            SQRT(DDOT(9, chi_tensor_tmp(1, 1), 1, chi_tensor_tmp(1, 1), 1))
      ENDIF
      !
      IF (BTEST(cp_print_key_should_output(logger%iter_info, nmr_section, &
                                           "PRINT%CHI_TENSOR"), cp_p_file)) THEN

         unit_atoms = cp_print_key_unit_nr(logger, nmr_section, "PRINT%CHI_TENSOR", &
                                           extension=".data", middle_name="CHI", log_filename=.FALSE.)

         WRITE (title, '(A)') "Magnetic Susceptibility Tensor "
         IF (unit_atoms > 0) THEN
            WRITE (unit_atoms, '(T2,A)') title
            WRITE (unit_atoms, '(T1,A)') " CHI from SOFT J in 10^-30 J/T^2 units"
            WRITE (unit_atoms, '(3(A,f10.4))') '  XX = ', chi_tensor_tmp(1, 1),&
                 &                           '  XY = ', chi_tensor_tmp(1, 2),&
                 &                           '  XZ = ', chi_tensor_tmp(1, 3)
            WRITE (unit_atoms, '(3(A,f10.4))') '  YX = ', chi_tensor_tmp(2, 1),&
                 &                           '  YY = ', chi_tensor_tmp(2, 2),&
                 &                           '  YZ = ', chi_tensor_tmp(2, 3)
            WRITE (unit_atoms, '(3(A,f10.4))') '  ZX = ', chi_tensor_tmp(3, 1),&
                 &                           '  ZY = ', chi_tensor_tmp(3, 2),&
                 &                           '  ZZ = ', chi_tensor_tmp(3, 3)
            IF (gapw) THEN
               WRITE (unit_atoms, '(T1,A)') " CHI from LOCAL J in 10^-30 J/T^2 units"
               WRITE (unit_atoms, '(3(A,f10.4))') '  XX = ', chi_tensor_loc_tmp(1, 1),&
                    &                           '  XY = ', chi_tensor_loc_tmp(1, 2),&
                    &                           '  XZ = ', chi_tensor_loc_tmp(1, 3)
               WRITE (unit_atoms, '(3(A,f10.4))') '  YX = ', chi_tensor_loc_tmp(2, 1),&
                    &                           '  YY = ', chi_tensor_loc_tmp(2, 2),&
                    &                           '  YZ = ', chi_tensor_loc_tmp(2, 3)
               WRITE (unit_atoms, '(3(A,f10.4))') '  ZX = ', chi_tensor_loc_tmp(3, 1),&
                    &                           '  ZY = ', chi_tensor_loc_tmp(3, 2),&
                    &                           '  ZZ = ', chi_tensor_loc_tmp(3, 3)
            ENDIF
            WRITE (unit_atoms, '(T1,A)') " Total CHI in 10^-30 J/T^2 units"
            WRITE (unit_atoms, '(3(A,f10.4))') '  XX = ', chi_sym_tot(1, 1),&
                 &                          '  XY = ', chi_sym_tot(1, 2),&
                 &                          '  XZ = ', chi_sym_tot(1, 3)
            WRITE (unit_atoms, '(3(A,f10.4))') '  YX = ', chi_sym_tot(2, 1),&
                 &                          '  YY = ', chi_sym_tot(2, 2),&
                 &                          '  YZ = ', chi_sym_tot(2, 3)
            WRITE (unit_atoms, '(3(A,f10.4))') '  ZX = ', chi_sym_tot(3, 1),&
                 &                          '  ZY = ', chi_sym_tot(3, 2),&
                 &                          '  ZZ = ', chi_sym_tot(3, 3)
            chi_sym_tot(:, :) = chi_sym_tot(:, :)*nmr_env%chi_SI2ppmcgs
            WRITE (unit_atoms, '(T1,A)') " Total CHI in ppm-cgs units"
            WRITE (unit_atoms, '(3(A,f10.4))') '  XX = ', chi_sym_tot(1, 1),&
                 &                          '  XY = ', chi_sym_tot(1, 2),&
                 &                          '  XZ = ', chi_sym_tot(1, 3)
            WRITE (unit_atoms, '(3(A,f10.4))') '  YX = ', chi_sym_tot(2, 1),&
                 &                          '  YY = ', chi_sym_tot(2, 2),&
                 &                          '  YZ = ', chi_sym_tot(2, 3)
            WRITE (unit_atoms, '(3(A,f10.4))') '  ZX = ', chi_sym_tot(3, 1),&
                 &                          '  ZY = ', chi_sym_tot(3, 2),&
                 &                          '  ZZ = ', chi_sym_tot(3, 3)
            WRITE (unit_atoms, '(/T1,3(A,f10.4))') &
               '  PV1=', nmr_env%chi_SI2ppmcgs*eig(1), &
               '  PV2=', nmr_env%chi_SI2ppmcgs*eig(2), &
               '  PV3=', nmr_env%chi_SI2ppmcgs*eig(3)
            WRITE (unit_atoms, '(T1,A,F10.4,10X,A,F10.4)') &
               '  ISO=', nmr_env%chi_SI2ppmcgs*chi_iso, &
               'ANISO=', nmr_env%chi_SI2ppmcgs*chi_aniso
         ENDIF

         CALL cp_print_key_finished_output(unit_atoms, logger, nmr_section,&
              &                            "PRINT%CHI_TENSOR")
      ENDIF ! print chi
      !
      ! Add the chi part to the shifts
      cs_tot(:, :, :) = 0.0_dp
      DO ir = 1, nat_print
         iatom = cs_atom_list(ir)
         rpos(1:3) = particle_set(iatom)%r(1:3)
         atom_kind => particle_set(iatom)%atomic_kind
         CALL get_atomic_kind(atom_kind, name=name, element_symbol=element_symbol)
         cs_tot(:, :, ir) = chi_tensor_tmp(:, :)*nmr_env%chi_SI2shiftppm+cs(:, :, iatom)
         IF (gapw) cs_tot(:, :, ir) = cs_tot(:, :, ir)+cs_loc(:, :, iatom)
      ENDDO ! ir
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E14.6)') 'CheckSum Shifts =', &
            SQRT(DDOT(9*SIZE(cs_tot, 3), cs_tot(1, 1, 1), 1, cs_tot(1, 1, 1), 1))
      ENDIF
      !
      ! print shifts
      IF (BTEST(cp_print_key_should_output(logger%iter_info, nmr_section, &
                                           "PRINT%SHIELDING_TENSOR"), cp_p_file)) THEN

         unit_atoms = cp_print_key_unit_nr(logger, nmr_section, "PRINT%SHIELDING_TENSOR", &
                                           extension=".data", middle_name="SHIFT", &
                                           log_filename=.FALSE.)

         nat_print = SIZE(cs_atom_list, 1)
         IF (unit_atoms > 0) THEN
            WRITE (title, '(A,1X,I5)') "Shielding atom at atomic positions. # tensors printed ", nat_print
            WRITE (unit_atoms, '(T2,A)') title
            DO ir = 1, nat_print
               iatom = cs_atom_list(ir)
               rpos(1:3) = particle_set(iatom)%r(1:3)
               atom_kind => particle_set(iatom)%atomic_kind
               CALL get_atomic_kind(atom_kind, name=name, element_symbol=element_symbol)
               shift_sym_tot(:, :) = 0.5_dp*(cs_tot(:, :, ir)+TRANSPOSE(cs_tot(:, :, ir)))
               CALL diamat_all(shift_sym_tot, eig)
               shift_iso = (eig(1)+eig(2)+eig(3))/3.0_dp
               shift_aniso = eig(3)-(eig(2)+eig(1))/2.0_dp
               !
               WRITE (unit_atoms, '(T2,I5,A,2X,A2,2X,3f15.6)') iatom, TRIM(name), element_symbol, rpos(1:3)
               !
               IF (gapw) THEN
                  WRITE (unit_atoms, '(T1,A)') " SIGMA from SOFT J"
                  WRITE (unit_atoms, '(3(A,f10.4))') '  XX = ', cs(1, 1, iatom),&
                       &                           '  XY = ', cs(1, 2, iatom),&
                       &                           '  XZ = ', cs(1, 3, iatom)
                  WRITE (unit_atoms, '(3(A,f10.4))') '  YX = ', cs(2, 1, iatom),&
                       &                           '  YY = ', cs(2, 2, iatom),&
                       &                           '  YZ = ', cs(2, 3, iatom)
                  WRITE (unit_atoms, '(3(A,f10.4))') '  ZX = ', cs(3, 1, iatom),&
                       &                           '  ZY = ', cs(3, 2, iatom),&
                       &                           '  ZZ = ', cs(3, 3, iatom)
                  WRITE (unit_atoms, '(T1,A)') " SIGMA from LOCAL J"
                  WRITE (unit_atoms, '(3(A,f10.4))') '  XX = ', cs_loc(1, 1, iatom),&
                       &                           '  XY = ', cs_loc(1, 2, iatom),&
                       &                           '  XZ = ', cs_loc(1, 3, iatom)
                  WRITE (unit_atoms, '(3(A,f10.4))') '  YX = ', cs_loc(2, 1, iatom),&
                       &                           '  YY = ', cs_loc(2, 2, iatom),&
                       &                           '  YZ = ', cs_loc(2, 3, iatom)
                  WRITE (unit_atoms, '(3(A,f10.4))') '  ZX = ', cs_loc(3, 1, iatom),&
                       &                           '  ZY = ', cs_loc(3, 2, iatom),&
                       &                           '  ZZ = ', cs_loc(3, 3, iatom)
               ENDIF
               WRITE (unit_atoms, '(T1,A)') " SIGMA TOTAL"
               WRITE (unit_atoms, '(3(A,f10.4))') '  XX = ', cs_tot(1, 1, ir),&
                    &                           '  XY = ', cs_tot(1, 2, ir),&
                    &                           '  XZ = ', cs_tot(1, 3, ir)
               WRITE (unit_atoms, '(3(A,f10.4))') '  YX = ', cs_tot(2, 1, ir),&
                    &                           '  YY = ', cs_tot(2, 2, ir),&
                    &                           '  YZ = ', cs_tot(2, 3, ir)
               WRITE (unit_atoms, '(3(A,f10.4))') '  ZX = ', cs_tot(3, 1, ir),&
                    &                           '  ZY = ', cs_tot(3, 2, ir),&
                    &                           '  ZZ = ', cs_tot(3, 3, ir)
               WRITE (unit_atoms, '(T1,2(A,f12.4))') '  ISOTROPY = ', shift_iso,&
                    &                            '  ANISOTROPY = ', shift_aniso
            ENDDO ! ir
         ENDIF
         CALL cp_print_key_finished_output(unit_atoms, logger, nmr_section,&
              &                            "PRINT%SHIELDING_TENSOR")

         IF (do_nics) THEN
            !
            ! Add the chi part to the nics
            cs_nics_tot(:, :, :) = 0.0_dp
            DO ir = 1, n_nics
               cs_nics_tot(:, :, ir) = chi_tensor_tmp(:, :)*nmr_env%chi_SI2shiftppm+cs_nics(:, :, ir)
               IF (gapw) cs_nics_tot(:, :, ir) = cs_nics_tot(:, :, ir)+cs_nics_loc(:, :, ir)
            ENDDO ! ir
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T2,A,E14.6)') 'CheckSum NICS =', &
                  SQRT(DDOT(9*SIZE(cs_nics_tot, 3), cs_nics_tot(1, 1, 1), 1, cs_nics_tot(1, 1, 1), 1))
            ENDIF
            !
            unit_nics = cp_print_key_unit_nr(logger, nmr_section, "PRINT%SHIELDING_TENSOR", &
                                             extension=".data", middle_name="NICS", &
                                             log_filename=.FALSE.)
            IF (unit_nics > 0) THEN
               WRITE (title, '(A,1X,I5)') "Shielding at nics positions. # tensors printed ", n_nics
               WRITE (unit_nics, '(T2,A)') title
               DO ir = 1, n_nics
                  shift_sym_tot(:, :) = 0.5_dp*(cs_nics_tot(:, :, ir)+TRANSPOSE(cs_nics_tot(:, :, ir)))
                  CALL diamat_all(shift_sym_tot, eig)
                  shift_iso = (eig(1)+eig(2)+eig(3))/3.0_dp
                  shift_aniso = eig(3)-(eig(2)+eig(1))/2.0_dp
                  !
                  WRITE (unit_nics, '(T2,I5,2X,3f15.6)') ir, r_nics(1:3, ir)
                  !
                  IF (gapw) THEN
                     WRITE (unit_nics, '(T1,A)') " SIGMA from SOFT J"
                     WRITE (unit_nics, '(3(A,f10.4))') '  XX = ', cs_nics(1, 1, ir),&
                          &                          '  XY = ', cs_nics(1, 2, ir),&
                          &                          '  XZ = ', cs_nics(1, 3, ir)
                     WRITE (unit_nics, '(3(A,f10.4))') '  YX = ', cs_nics(2, 1, ir),&
                          &                          '  YY = ', cs_nics(2, 2, ir),&
                          &                          '  YZ = ', cs_nics(2, 3, ir)
                     WRITE (unit_nics, '(3(A,f10.4))') '  ZX = ', cs_nics(3, 1, ir),&
                          &                          '  ZY = ', cs_nics(3, 2, ir),&
                          &                          '  ZZ = ', cs_nics(3, 3, ir)
                     WRITE (unit_nics, '(T1,A)') " SIGMA from LOCAL J"
                     WRITE (unit_nics, '(3(A,f10.4))') '  XX = ', cs_nics_loc(1, 1, ir),&
                          &                          '  XY = ', cs_nics_loc(1, 2, ir),&
                          &                          '  XZ = ', cs_nics_loc(1, 3, ir)
                     WRITE (unit_nics, '(3(A,f10.4))') '  YX = ', cs_nics_loc(2, 1, ir),&
                          &                          '  YY = ', cs_nics_loc(2, 2, ir),&
                          &                          '  YZ = ', cs_nics_loc(2, 3, ir)
                     WRITE (unit_nics, '(3(A,f10.4))') '  ZX = ', cs_nics_loc(3, 1, ir),&
                          &                          '  ZY = ', cs_nics_loc(3, 2, ir),&
                          &                          '  ZZ = ', cs_nics_loc(3, 3, ir)
                  ENDIF
                  WRITE (unit_nics, '(T1,A)') " SIGMA TOTAL"
                  WRITE (unit_nics, '(3(A,f10.4))') '  XX = ', cs_nics_tot(1, 1, ir),&
                       &                          '  XY = ', cs_nics_tot(1, 2, ir),&
                       &                          '  XZ = ', cs_nics_tot(1, 3, ir)
                  WRITE (unit_nics, '(3(A,f10.4))') '  YX = ', cs_nics_tot(2, 1, ir),&
                       &                          '  YY = ', cs_nics_tot(2, 2, ir),&
                       &                          '  YZ = ', cs_nics_tot(2, 3, ir)
                  WRITE (unit_nics, '(3(A,f10.4))') '  ZX = ', cs_nics_tot(3, 1, ir),&
                       &                          '  ZY = ', cs_nics_tot(3, 2, ir),&
                       &                          '  ZZ = ', cs_nics_tot(3, 3, ir)
                  WRITE (unit_nics, '(T1,2(A,f12.4))') '  ISOTROPY = ', shift_iso,&
                       &                           '  ANISOTROPY = ', shift_aniso
               ENDDO
            ENDIF
            CALL cp_print_key_finished_output(unit_nics, logger, nmr_section,&
                 &                            "PRINT%SHIELDING_TENSOR")
         ENDIF
      ENDIF ! print shift
      !
      ! clean up
      DEALLOCATE (cs_tot)
      IF (do_nics) THEN
         DEALLOCATE (cs_nics_tot)
      ENDIF
      !
!100 FORMAT(A,1X,I5)
!101 FORMAT(T2,A)
!102 FORMAT(T2,I5,A,2X,A2,2X,3f15.6)
!103 FORMAT(T2,I5,2X,3f15.6)
!104 FORMAT(T1,A)
!105 FORMAT(3(A,f10.4))
!106 FORMAT(T1,2(A,f12.4))
   END SUBROUTINE nmr_shift_print

END MODULE qs_linres_nmr_shift

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief g tensor calculation by dfpt
!>      Initialization of the epr_env, creation of the special neighbor lists
!>      Perturbation Hamiltonians by application of the p and rxp oprtators to  psi0
!>      Write output
!>      Deallocate everything
!> \note
!>      The psi0 should be localized
!>      the Sebastiani method works within the assumption that the orbitals are
!>      completely contained in the simulation box
!> \par History
!>       created 07-2005 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_epr_utils
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              twopi
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: a_fine,&
                                              e_gfactor
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_linres_types,                 ONLY: deallocate_nablavks_atom_set,&
                                              epr_env_create,&
                                              epr_env_type,&
                                              init_nablavks_atom_set,&
                                              linres_control_type,&
                                              nablavks_atom_type,&
                                              set_epr_env
   USE qs_matrix_pools,                 ONLY: qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: mo_set_p_type
   USE qs_rho_atom_types,               ONLY: deallocate_rho_atom_set
   USE qs_rho_types,                    ONLY: qs_rho_clear,&
                                              qs_rho_create,&
                                              qs_rho_set
   USE scf_control_types,               ONLY: scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: epr_env_cleanup, epr_env_init

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_epr_utils'

CONTAINS

! **************************************************************************************************
!> \brief Initialize the epr environment
!> \param epr_env ...
!> \param qs_env ...
!> \par History
!>      07.2006 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE epr_env_init(epr_env, qs_env)
      !
      TYPE(epr_env_type)                                 :: epr_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'epr_env_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_B, idir, ispin, n_mo(2), nao, &
                                                            natom, nmoloc, nspins, output_unit
      LOGICAL                                            :: gapw
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(nablavks_atom_type), DIMENSION(:), POINTER    :: nablavks_atom_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: lr_section

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, qs_kind_set, cell, dft_control, linres_control, scf_control)
      NULLIFY (logger, mos, mpools, particle_set)
      NULLIFY (auxbas_pw_pool, pw_env)
      NULLIFY (nablavks_atom_set)

      n_mo(1:2) = 0
      nao = 0
      nmoloc = 0

      logger => cp_get_default_logger()
      !ionode = logger%para_env%mepos==logger%para_env%source
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      IF (epr_env%ref_count /= 0) THEN
         CALL epr_env_cleanup(epr_env)
      END IF

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(/,T20,A,/)") "*** Start EPR g tensor calculation ***"
         WRITE (output_unit, "(T10,A,/)") "Initialization of the EPR environment"
      ENDIF

      CALL epr_env_create(epr_env)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      linres_control=linres_control, &
                      mos=mos, &
                      mpools=mpools, &
                      particle_set=particle_set, &
                      pw_env=pw_env, &
                      scf_control=scf_control)
      !
      ! Check if restat also psi0 should be restarted
      !IF(epr_env%restart_epr .AND. scf_control%density_guess/=restart_guess)THEN
      !   CALL cp__b("qs_linres_epr_utils.F",151,"restart_epr requires density_guess=restart")
      !ENDIF
      !
      ! check that the psi0 are localized and you have all the centers
      IF(.NOT.(linres_control%localized_psi0))CALL cp__a("qs_linres_epr_utils.F",155)
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(A)') &
            ' To get EPR parameters within PBC you need localized zero order orbitals '
      ENDIF
      gapw = dft_control%qs_control%gapw
      nspins = dft_control%nspins
      natom = SIZE(particle_set, 1)
      !
      ! Conversion factors
      ! Magical constant twopi/cell%deth just like in NMR shift (basically undo scale_fac in qs_linres_nmr_current.F)
      epr_env%g_free_factor = -1.0_dp*e_gfactor
      epr_env%g_zke_factor = e_gfactor*(a_fine)**2
      epr_env%g_so_factor = (a_fine)**2*(-1.0_dp*e_gfactor-1.0_dp)/2.0_dp*twopi/cell%deth
      epr_env%g_so_factor_gapw = (a_fine)**2*(-1.0_dp*e_gfactor-1.0_dp)/2.0_dp
      ! * 2 because B_ind = 2 * B_beta
      epr_env%g_soo_factor = 2.0_dp*fourpi*(a_fine)**2*twopi/cell%deth
      ! 2 * 2 * 1/4 * e^2 / m * a_0^2 * 2/3 * mu_0 / (omega * 1e-30 )
      epr_env%g_soo_chicorr_factor = 2.0/3.0_dp*fourpi*(a_fine)**2/cell%deth
      !
      ! If the current density on the grid needs to be stored
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
      !
      ! Initialize local current density if GAPW calculation
      IF (gapw) THEN
         CALL init_nablavks_atom_set(nablavks_atom_set, atomic_kind_set, qs_kind_set, nspins)
         CALL set_epr_env(epr_env=epr_env, &
                          nablavks_atom_set=nablavks_atom_set)
      ENDIF
      !
      ! Bind
      ALLOCATE (epr_env%bind_set(3, 3))
      DO i_B = 1, 3
         DO idir = 1, 3
            NULLIFY (epr_env%bind_set(idir, i_B)%rho, rho_r, rho_g)
            CALL qs_rho_create(epr_env%bind_set(idir, i_B)%rho)
            ALLOCATE (rho_r(1), rho_g(1))
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_r(1)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_g(1)%pw, &
                                   use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            CALL qs_rho_set(epr_env%bind_set(idir, i_B)%rho, rho_r=rho_r, rho_g=rho_g)
         END DO
      END DO

      ! Nabla_V_ks
      ALLOCATE (epr_env%nablavks_set(3, dft_control%nspins))
      DO idir = 1, 3
         DO ispin = 1, nspins
            NULLIFY (epr_env%nablavks_set(idir, ispin)%rho, rho_r, rho_g)
            CALL qs_rho_create(epr_env%nablavks_set(idir, ispin)%rho)
            ALLOCATE (rho_r(1), rho_g(1))
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_r(1)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_pool_create_pw(auxbas_pw_pool, rho_g(1)%pw, &
                                   use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            CALL qs_rho_set(epr_env%nablavks_set(idir, ispin)%rho, &
                            rho_r=rho_r, rho_g=rho_g)
         END DO
      END DO

      ! Initialize the g tensor components
      ALLOCATE (epr_env%g_total(3, 3))
      ALLOCATE (epr_env%g_so(3, 3))
      ALLOCATE (epr_env%g_soo(3, 3))
      epr_env%g_total = 0.0_dp
      epr_env%g_zke = 0.0_dp
      epr_env%g_so = 0.0_dp
      epr_env%g_soo = 0.0_dp

      CALL cp_print_key_finished_output(output_unit, logger, lr_section,&
           &                            "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE epr_env_init

! **************************************************************************************************
!> \brief Deallocate the epr environment
!> \param epr_env ...
!> \par History
!>      07.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE epr_env_cleanup(epr_env)

      TYPE(epr_env_type)                                 :: epr_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'epr_env_cleanup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_B, idir, ispin

      epr_env%ref_count = epr_env%ref_count-1
      IF (epr_env%ref_count == 0) THEN
         ! nablavks_set
         IF (ASSOCIATED(epr_env%nablavks_set)) THEN
            DO ispin = 1, SIZE(epr_env%nablavks_set, 2)
            DO idir = 1, SIZE(epr_env%nablavks_set, 1)
               CALL qs_rho_clear(epr_env%nablavks_set(idir, ispin)%rho)
               DEALLOCATE (epr_env%nablavks_set(idir, ispin)%rho)
            ENDDO
            ENDDO
            DEALLOCATE (epr_env%nablavks_set)
         END IF
         ! nablavks_atom_set
         IF (ASSOCIATED(epr_env%nablavks_atom_set)) THEN
            CALL deallocate_nablavks_atom_set(epr_env%nablavks_atom_set)
         END IF
         ! vks_atom_set
         IF (ASSOCIATED(epr_env%vks_atom_set)) THEN
            CALL deallocate_rho_atom_set(epr_env%vks_atom_set)
         END IF
         ! bind_set
         IF (ASSOCIATED(epr_env%bind_set)) THEN
            DO i_B = 1, SIZE(epr_env%bind_set, 2)
            DO idir = 1, SIZE(epr_env%bind_set, 1)
               CALL qs_rho_clear(epr_env%bind_set(idir, i_B)%rho)
               DEALLOCATE (epr_env%bind_set(idir, i_B)%rho)
            ENDDO
            ENDDO
            DEALLOCATE (epr_env%bind_set)
         END IF
         ! bind_atom_set
         IF (ASSOCIATED(epr_env%bind_atom_set)) THEN
            DEALLOCATE (epr_env%bind_atom_set)
         END IF
         ! g_total
         IF (ASSOCIATED(epr_env%g_total)) THEN
            DEALLOCATE (epr_env%g_total)
         END IF
         ! g_so
         IF (ASSOCIATED(epr_env%g_so)) THEN
            DEALLOCATE (epr_env%g_so)
         END IF
         ! g_soo
         IF (ASSOCIATED(epr_env%g_soo)) THEN
            DEALLOCATE (epr_env%g_soo)
         END IF
      END IF ! ref count

   END SUBROUTINE epr_env_cleanup

END MODULE qs_linres_epr_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Chemical shift calculation by dfpt
!>      Initialization of the issc_env, creation of the special neighbor lists
!>      Perturbation Hamiltonians by application of the p and rxp oprtators to  psi0
!>      Write output
!>      Deallocate everything
!> \note
!>      The psi0 should be localized
!>      the Sebastiani method works within the assumption that the orbitals are
!>      completely contained in the simulation box
! **************************************************************************************************
MODULE qs_linres_issc_utils
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        convert_offsets_to_sizes, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_init, cp_dbcsr_p_type, cp_dbcsr_set, &
        dbcsr_distribution_obj, dbcsr_type_antisymmetric, dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_frobenius_norm,&
                                              cp_fm_trace
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathlib,                         ONLY: diamat_all
   USE memory_utilities,                ONLY: reallocate
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: a_fine,&
                                              e_mass,&
                                              hertz,&
                                              p_mass
   USE qs_elec_field,                   ONLY: build_efg_matrix
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_fermi_contact,                ONLY: build_fermi_contact_matrix
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_linres_methods,               ONLY: linres_solver
   USE qs_linres_types,                 ONLY: get_issc_env,&
                                              issc_env_create,&
                                              issc_env_type,&
                                              linres_control_type
   USE qs_matrix_pools,                 ONLY: qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_spin_orbit,                   ONLY: build_pso_matrix
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: issc_env_cleanup, issc_env_init, issc_response, issc_issc, issc_print

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_issc_utils'

CONTAINS

! **************************************************************************************************
!> \brief Initialize the issc environment
!> \param issc_env ...
!> \param p_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE issc_response(issc_env, p_env, qs_env)
      !
      TYPE(issc_env_type)                                :: issc_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'issc_response', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idir, ijdir, ispin, jdir, nao, &
                                                            nmo, nspins, output_unit
      LOGICAL                                            :: do_dso, do_fc, do_pso, do_sd, should_stop
      REAL(dp)                                           :: chk, fro
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fc_psi0, h1_psi0, psi0_order, psi1, &
                                                            psi1_fc
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: dso_psi0, efg_psi0, psi1_dso, psi1_efg, &
                                                            psi1_pso, pso_psi0
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(section_vals_type), POINTER                   :: issc_section, lr_section

      CALL timeset(routineN, handle)
      !
      NULLIFY (dft_control, linres_control, lr_section, issc_section)
      NULLIFY (logger, mpools, psi1, h1_psi0, mo_coeff, para_env)
      NULLIFY (tmp_fm_struct, psi1_fc, psi1_efg, psi1_pso, pso_psi0, fc_psi0, efg_psi0, psi0_order)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")
      issc_section => section_vals_get_subs_vals(qs_env%input, &
                                                 "PROPERTIES%LINRES%SPINSPIN")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(T10,A,/)") &
            "*** Self consistent optimization of the response wavefunctions ***"
      ENDIF

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      mpools=mpools, &
                      linres_control=linres_control, &
                      mos=mos, &
                      para_env=para_env)

      nspins = dft_control%nspins

      CALL get_issc_env(issc_env=issc_env, &
                        !list_cubes=list_cubes, &
                        psi1_efg=psi1_efg, &
                        psi1_pso=psi1_pso, &
                        psi1_dso=psi1_dso, &
                        psi1_fc=psi1_fc, &
                        efg_psi0=efg_psi0, &
                        pso_psi0=pso_psi0, &
                        dso_psi0=dso_psi0, &
                        fc_psi0=fc_psi0, &
                        do_fc=do_fc, &
                        do_sd=do_sd, &
                        do_pso=do_pso, &
                        do_dso=do_dso)
      !
      ! allocate the vectors
      ALLOCATE (psi0_order(nspins))
      ALLOCATE (psi1(nspins), h1_psi0(nspins))
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         psi0_order(ispin)%matrix => mo_coeff
         CALL cp_fm_get_info(mo_coeff, ncol_global=nmo, nrow_global=nao)
         NULLIFY (tmp_fm_struct, psi1(ispin)%matrix, h1_psi0(ispin)%matrix)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                  ncol_global=nmo, &
                                  context=mo_coeff%matrix_struct%context)
         CALL cp_fm_create(psi1(ispin)%matrix, tmp_fm_struct)
         CALL cp_fm_create(h1_psi0(ispin)%matrix, tmp_fm_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
      ENDDO
      chk = 0.0_dp
      should_stop = .FALSE.
      !
      ! operator efg
      IF (do_sd) THEN
         ijdir = 0
         DO idir = 1, 3
         DO jdir = idir, 3
            ijdir = ijdir+1
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1_efg(ispin, ijdir)%matrix, 0.0_dp)
            ENDDO
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T10,A)") "Response to the perturbation operator efg_"//ACHAR(idir+119)//ACHAR(jdir+119)
            ENDIF
            !
            !Initial guess for psi1
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1(ispin)%matrix, 0.0_dp)
               !CALL cp_fm_to_fm(p_psi0(ispin,ijdir)%matrix, psi1(ispin)%matrix)
               !CALL cp_fm_scale(-1.0_dp,psi1(ispin)%matrix)
            ENDDO
            !
            !DO scf cycle to optimize psi1
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(efg_psi0(ispin, ijdir)%matrix, h1_psi0(ispin)%matrix)
            ENDDO
            !
            !
            linres_control%lr_triplet = .FALSE.
            linres_control%do_kernel = .FALSE.
            linres_control%converged = .FALSE.
            CALL linres_solver(p_env, qs_env, psi1, h1_psi0, psi0_order, should_stop)
            !
            !
            ! copy the response
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(psi1(ispin)%matrix, psi1_efg(ispin, ijdir)%matrix)
               CALL cp_fm_frobenius_norm(psi1(ispin)%matrix, fro)
               chk = chk+fro
            ENDDO
            !
            ! print response functions
            !IF(BTEST(cp_print_key_should_output(logger%iter_info,issc_section,&
            !     &   "PRINT%RESPONSE_FUNCTION_CUBES"),cp_p_file)) THEN
            !   ncubes = SIZE(list_cubes,1)
            !   print_key => section_vals_get_subs_vals(issc_section,"PRINT%RESPONSE_FUNCTION_CUBES")
            !   DO ispin = 1,nspins
            !      CALL qs_print_cubes(qs_env,psi1(ispin)%matrix,ncubes,list_cubes,&
            !            centers_set(ispin)%array,print_key,'psi1_efg',&
            !            idir=ijdir,ispin=ispin)
            !   ENDDO ! ispin
            !ENDIF ! print response functions
            !
            !
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T10,A)") "Write the resulting psi1 in restart file... not implemented yet"
            ENDIF
            !
            ! Write the result in the restart file
         ENDDO ! jdir
         ENDDO ! idir
      ENDIF
      !
      ! operator pso
      IF (do_pso) THEN
         DO idir = 1, 3
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1_pso(ispin, idir)%matrix, 0.0_dp)
            ENDDO
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T10,A)") "Response to the perturbation operator pso_"//ACHAR(idir+119)
            ENDIF
            !
            !Initial guess for psi1
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1(ispin)%matrix, 0.0_dp)
               !CALL cp_fm_to_fm(rxp_psi0(ispin,idir)%matrix, psi1(ispin)%matrix)
               !CALL cp_fm_scale(-1.0_dp,psi1(ispin)%matrix)
            ENDDO
            !
            !DO scf cycle to optimize psi1
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(pso_psi0(ispin, idir)%matrix, h1_psi0(ispin)%matrix)
            ENDDO
            !
            !
            linres_control%lr_triplet = .FALSE. ! we do singlet response
            linres_control%do_kernel = .FALSE. ! we do uncoupled response
            linres_control%converged = .FALSE.
            CALL linres_solver(p_env, qs_env, psi1, h1_psi0, psi0_order, should_stop)
            !
            !
            ! copy the response
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(psi1(ispin)%matrix, psi1_pso(ispin, idir)%matrix)
               CALL cp_fm_frobenius_norm(psi1(ispin)%matrix, fro)
               chk = chk+fro
            ENDDO
            !
            ! print response functions
            !IF(BTEST(cp_print_key_should_output(logger%iter_info,issc_section,&
            !     &   "PRINT%RESPONSE_FUNCTION_CUBES"),cp_p_file)) THEN
            !   ncubes = SIZE(list_cubes,1)
            !   print_key => section_vals_get_subs_vals(issc_section,"PRINT%RESPONSE_FUNCTION_CUBES")
            !   DO ispin = 1,nspins
            !      CALL qs_print_cubes(qs_env,psi1(ispin)%matrix,ncubes,list_cubes,&
            !           centers_set(ispin)%array,print_key,'psi1_pso',&
            !           idir=idir,ispin=ispin)
            !   ENDDO ! ispin
            !ENDIF ! print response functions
            !
            !
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T10,A)") "Write the resulting psi1 in restart file... not implemented yet"
            ENDIF
            !
            ! Write the result in the restart file
         ENDDO ! idir
      ENDIF
      !
      ! operator fc
      IF (do_fc) THEN
         DO ispin = 1, nspins
            CALL cp_fm_set_all(psi1_fc(ispin)%matrix, 0.0_dp)
         ENDDO
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(T10,A)") "Response to the perturbation operator fc"
         ENDIF
         !
         !Initial guess for psi1
         DO ispin = 1, nspins
            CALL cp_fm_set_all(psi1(ispin)%matrix, 0.0_dp)
            !CALL cp_fm_to_fm(rxp_psi0(ispin,idir)%matrix, psi1(ispin)%matrix)
            !CALL cp_fm_scale(-1.0_dp,psi1(ispin)%matrix)
         ENDDO
         !
         !DO scf cycle to optimize psi1
         DO ispin = 1, nspins
            CALL cp_fm_to_fm(fc_psi0(ispin)%matrix, h1_psi0(ispin)%matrix)
         ENDDO
         !
         !
         linres_control%lr_triplet = .TRUE. ! we do triplet response
         linres_control%do_kernel = .TRUE. ! we do coupled response
         linres_control%converged = .FALSE.
         CALL linres_solver(p_env, qs_env, psi1, h1_psi0, psi0_order, should_stop)
         !
         !
         ! copy the response
         DO ispin = 1, nspins
            CALL cp_fm_to_fm(psi1(ispin)%matrix, psi1_fc(ispin)%matrix)
            CALL cp_fm_frobenius_norm(psi1(ispin)%matrix, fro)
            chk = chk+fro
         ENDDO
         !
         ! print response functions
         !IF(BTEST(cp_print_key_should_output(logger%iter_info,issc_section,&
         !     &   "PRINT%RESPONSE_FUNCTION_CUBES"),cp_p_file)) THEN
         !   ncubes = SIZE(list_cubes,1)
         !   print_key => section_vals_get_subs_vals(issc_section,"PRINT%RESPONSE_FUNCTION_CUBES")
         !   DO ispin = 1,nspins
         !      CALL qs_print_cubes(qs_env,psi1(ispin)%matrix,ncubes,list_cubes,&
         !           centers_set(ispin)%array,print_key,'psi1_pso',&
         !           idir=idir,ispin=ispin)
         !   ENDDO ! ispin
         !ENDIF ! print response functions
         !
         !
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(T10,A)") "Write the resulting psi1 in restart file... not implemented yet"
         ENDIF
         !
         ! Write the result in the restart file
      ENDIF

      !>>>> debugging only
      !
      ! here we have the operator r and compute the polarizability for debugging the kernel only
      IF (do_dso) THEN
         DO idir = 1, 3
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1_dso(ispin, idir)%matrix, 0.0_dp)
            ENDDO
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T10,A)") "Response to the perturbation operator r_"//ACHAR(idir+119)
            ENDIF
            !
            !Initial guess for psi1
            DO ispin = 1, nspins
               CALL cp_fm_set_all(psi1(ispin)%matrix, 0.0_dp)
               !CALL cp_fm_to_fm(rxp_psi0(ispin,idir)%matrix, psi1(ispin)%matrix)
               !CALL cp_fm_scale(-1.0_dp,psi1(ispin)%matrix)
            ENDDO
            !
            !DO scf cycle to optimize psi1
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(dso_psi0(ispin, idir)%matrix, h1_psi0(ispin)%matrix)
            ENDDO
            !
            !
            linres_control%lr_triplet = .FALSE. ! we do singlet response
            linres_control%do_kernel = .TRUE. ! we do uncoupled response
            linres_control%converged = .FALSE.
            CALL linres_solver(p_env, qs_env, psi1, h1_psi0, psi0_order, should_stop)
            !
            !
            ! copy the response
            DO ispin = 1, nspins
               CALL cp_fm_to_fm(psi1(ispin)%matrix, psi1_dso(ispin, idir)%matrix)
               CALL cp_fm_frobenius_norm(psi1(ispin)%matrix, fro)
               chk = chk+fro
            ENDDO
            !
            ! print response functions
            !IF(BTEST(cp_print_key_should_output(logger%iter_info,issc_section,&
            !     &   "PRINT%RESPONSE_FUNCTION_CUBES"),cp_p_file)) THEN
            !   ncubes = SIZE(list_cubes,1)
            !   print_key => section_vals_get_subs_vals(issc_section,"PRINT%RESPONSE_FUNCTION_CUBES")
            !   DO ispin = 1,nspins
            !      CALL qs_print_cubes(qs_env,psi1(ispin)%matrix,ncubes,list_cubes,&
            !           centers_set(ispin)%array,print_key,'psi1_pso',&
            !           idir=idir,ispin=ispin)
            !   ENDDO ! ispin
            !ENDIF ! print response functions
            !
            !
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(T10,A)") "Write the resulting psi1 in restart file... not implemented yet"
            ENDIF
            !
            ! Write the result in the restart file
         ENDDO ! idir
      ENDIF
      !<<<< debugging only

      !
      !
      ! print the checksum
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E23.16)') 'ISSC| response: CheckSum =', chk
      ENDIF
      !
      !
      ! clean up
      DO ispin = 1, nspins
         CALL cp_fm_release(psi1(ispin)%matrix)
         CALL cp_fm_release(h1_psi0(ispin)%matrix)
      ENDDO
      DEALLOCATE (psi1, h1_psi0, psi0_order)
      !
      CALL cp_print_key_finished_output(output_unit, logger, lr_section,&
           &                            "PRINT%PROGRAM_RUN_INFO")
      !
      CALL timestop(handle)
      !
   END SUBROUTINE issc_response

! **************************************************************************************************
!> \brief ...
!> \param issc_env ...
!> \param qs_env ...
!> \param iatom ...
! **************************************************************************************************
   SUBROUTINE issc_issc(issc_env, qs_env, iatom)

      TYPE(issc_env_type)                                :: issc_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iatom

      CHARACTER(LEN=*), PARAMETER :: routineN = 'issc_issc', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, ixyz, jatom, jxyz, natom, &
                                                            nmo, nspins
      LOGICAL                                            :: do_dso, do_fc, do_pso, do_sd, gapw
      REAL(dp)                                           :: buf, facdso, facfc, facpso, facsd, g, &
                                                            issc_dso, issc_fc, issc_pso, issc_sd, &
                                                            maxocc
      REAL(dp), DIMENSION(3)                             :: r_i, r_j
      REAL(dp), DIMENSION(:, :, :, :, :), POINTER        :: issc
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_dso, matrix_efg, matrix_fc, &
                                                            matrix_pso
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fc_psi0, psi1_fc
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: psi1_dso, psi1_efg, psi1_pso
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: issc_section

      CALL timeset(routineN, handle)

      NULLIFY (cell, dft_control, particle_set, issc, psi1_fc, psi1_efg, psi1_pso)
      NULLIFY (matrix_efg, matrix_fc, matrix_pso, mos, mo_coeff, fc_psi0)

      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      mos=mos)

      gapw = dft_control%qs_control%gapw
      natom = SIZE(particle_set, 1)
      nspins = dft_control%nspins

      CALL get_issc_env(issc_env=issc_env, &
                        matrix_efg=matrix_efg, &
                        matrix_pso=matrix_pso, &
                        matrix_fc=matrix_fc, &
                        matrix_dso=matrix_dso, &
                        psi1_fc=psi1_fc, &
                        psi1_efg=psi1_efg, &
                        psi1_pso=psi1_pso, &
                        psi1_dso=psi1_dso, &
                        fc_psi0=fc_psi0, &
                        issc=issc, &
                        do_fc=do_fc, &
                        do_sd=do_sd, &
                        do_pso=do_pso, &
                        do_dso=do_dso)

      g = e_mass/(2.0_dp*p_mass)
      facfc = hertz*g**2*a_fine**4
      facpso = hertz*g**2*a_fine**4
      facsd = hertz*g**2*a_fine**4
      facdso = hertz*g**2*a_fine**4

      !
      !
      issc_section => section_vals_get_subs_vals(qs_env%input, &
           & "PROPERTIES%LINRES%SPINSPIN")
      !
      ! Initialize
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff, maxocc=maxocc)
         CALL cp_fm_get_info(mo_coeff, ncol_global=nmo)

         DO jatom = 1, natom
            r_i = particle_set(iatom)%r
            r_j = particle_set(jatom)%r
            r_j = pbc(r_i, r_j, cell)+r_i
            !
            !
            !
            !write(*,*) 'iatom =',iatom,' r_i=',r_i
            !write(*,*) 'jatom =',jatom,' r_j=',r_j
            !
            ! FC term
            !
            IF (do_fc .AND. iatom .NE. jatom) THEN
               !
               ! build the integral for the jatom
               CALL cp_dbcsr_set(matrix_fc(1)%matrix, 0.0_dp)
               CALL build_fermi_contact_matrix(qs_env, matrix_fc, r_j)
               CALL cp_dbcsr_sm_fm_multiply(matrix_fc(1)%matrix, mo_coeff, &
                                      fc_psi0(ispin)%matrix, ncol=nmo,& ! fc_psi0 a buffer
                    &                 alpha=1.0_dp)

               CALL cp_fm_trace(fc_psi0(ispin)%matrix, mo_coeff, buf)
               WRITE (*, *) ' jatom', jatom, 'tr(P*fc)=', buf

               CALL cp_fm_trace(fc_psi0(ispin)%matrix, psi1_fc(ispin)%matrix, buf)
               issc_fc = 2.0_dp*2.0_dp*maxocc*facfc*buf
               issc(1, 1, iatom, jatom, 1) = issc(1, 1, iatom, jatom, 1)+issc_fc
               issc(2, 2, iatom, jatom, 1) = issc(2, 2, iatom, jatom, 1)+issc_fc
               issc(3, 3, iatom, jatom, 1) = issc(3, 3, iatom, jatom, 1)+issc_fc
            ENDIF
            !
            ! SD term
            !
            IF (do_sd .AND. iatom .NE. jatom) THEN
               !
               ! build the integral for the jatom
               CALL cp_dbcsr_set(matrix_efg(1)%matrix, 0.0_dp)
               CALL cp_dbcsr_set(matrix_efg(2)%matrix, 0.0_dp)
               CALL cp_dbcsr_set(matrix_efg(3)%matrix, 0.0_dp)
               CALL cp_dbcsr_set(matrix_efg(4)%matrix, 0.0_dp)
               CALL cp_dbcsr_set(matrix_efg(5)%matrix, 0.0_dp)
               CALL cp_dbcsr_set(matrix_efg(6)%matrix, 0.0_dp)
               CALL build_efg_matrix(qs_env, matrix_efg, r_j)
               DO ixyz = 1, 6
                  CALL cp_dbcsr_sm_fm_multiply(matrix_efg(ixyz)%matrix, mo_coeff, &
                                         fc_psi0(ispin)%matrix, ncol=nmo,& ! fc_psi0 a buffer
                       &                 alpha=1.0_dp, beta=0.0_dp)
                  CALL cp_fm_trace(fc_psi0(ispin)%matrix, mo_coeff, buf)
                  WRITE (*, *) ' jatom', jatom, ixyz, 'tr(P*efg)=', buf
                  DO jxyz = 1, 6
                     CALL cp_fm_trace(fc_psi0(ispin)%matrix, psi1_efg(ispin, jxyz)%matrix, buf)
                     issc_sd = 2.0_dp*maxocc*facsd*buf
                     !issc(ixyz,jxyz,iatom,jatom) = issc_sd
                     !write(*,*) 'pso_',ixyz,jxyz,' iatom',iatom,'jatom',jatom,issc_pso
                  ENDDO
               ENDDO
            ENDIF
            !
            ! PSO term
            !
            IF (do_pso .AND. iatom .NE. jatom) THEN
               !
               ! build the integral for the jatom
               CALL cp_dbcsr_set(matrix_pso(1)%matrix, 0.0_dp)
               CALL cp_dbcsr_set(matrix_pso(2)%matrix, 0.0_dp)
               CALL cp_dbcsr_set(matrix_pso(3)%matrix, 0.0_dp)
               CALL build_pso_matrix(qs_env, matrix_pso, r_j)
               DO ixyz = 1, 3
                  CALL cp_dbcsr_sm_fm_multiply(matrix_pso(ixyz)%matrix, mo_coeff, &
                                         fc_psi0(ispin)%matrix, ncol=nmo,& ! fc_psi0 a buffer
                       &                 alpha=1.0_dp, beta=0.0_dp)
                  DO jxyz = 1, 3
                     CALL cp_fm_trace(fc_psi0(ispin)%matrix, psi1_pso(ispin, jxyz)%matrix, buf)
                     issc_pso = -2.0_dp*maxocc*facpso*buf
                     issc(ixyz, jxyz, iatom, jatom, 3) = issc(ixyz, jxyz, iatom, jatom, 3)+issc_pso
                  ENDDO
               ENDDO
            ENDIF
            !
            ! DSO term
            !
            !>>>>> for debugging we compute here the polarizability and NOT the DSO term!
            IF (do_dso .AND. iatom .EQ. natom .AND. jatom .EQ. natom) THEN
               !
               ! build the integral for the jatom
               !CALL cp_dbcsr_set(matrix_dso(1)%matrix,0.0_dp)
               !CALL cp_dbcsr_set(matrix_dso(2)%matrix,0.0_dp)
               !CALL cp_dbcsr_set(matrix_dso(3)%matrix,0.0_dp)
               !CALL cp_dbcsr_set(matrix_dso(4)%matrix,0.0_dp)
               !CALL cp_dbcsr_set(matrix_dso(5)%matrix,0.0_dp)
               !CALL cp_dbcsr_set(matrix_dso(6)%matrix,0.0_dp)
               !CALL build_dso_matrix(qs_env,matrix_dso,r_i,r_j)
               !DO ixyz = 1,6
               !   CALL cp_sm_fm_multiply(matrix_dso(ixyz)%matrix,mo_coeff,&
               !        &                 fc_psi0(ispin)%matrix,ncol=nmo,& ! fc_psi0 a buffer
               !        &                 alpha=1.0_dp,beta=0.0_dp)
               !   CALL cp_fm_trace(fc_psi0(ispin)%matrix,mo_coeff,buf)
               !   issc_dso = 2.0_dp * maxocc * facdso * buf
               !   issc(ixyz,jxyz,iatom,jatom,4) = issc_dso
               !ENDDO
               !CALL cp_dbcsr_set(matrix_dso(1)%matrix,0.0_dp)
               !CALL cp_dbcsr_set(matrix_dso(2)%matrix,0.0_dp)
               !CALL cp_dbcsr_set(matrix_dso(3)%matrix,0.0_dp)
               !CALL rRc_xyz_ao(matrix_dso,qs_env,(/0.0_dp,0.0_dp,0.0_dp/),1)
               DO ixyz = 1, 3
                  CALL cp_dbcsr_sm_fm_multiply(matrix_dso(ixyz)%matrix, mo_coeff, &
                                         fc_psi0(ispin)%matrix, ncol=nmo,& ! fc_psi0 a buffer
                       &                 alpha=1.0_dp, beta=0.0_dp)
                  DO jxyz = 1, 3
                     CALL cp_fm_trace(psi1_dso(ispin, jxyz)%matrix, fc_psi0(ispin)%matrix, buf)
                     ! we save the polarizability for a checksum later on !
                     issc_dso = 2.0_dp*maxocc*buf
                     !WRITE(*,*) ixyz,jxyz,'tr(P_r*r)=',2.0_dp * maxocc * buf
                     issc(ixyz, jxyz, iatom, jatom, 4) = issc(ixyz, jxyz, iatom, jatom, 4)+issc_dso
                  ENDDO
               ENDDO

            ENDIF
            !
         ENDDO ! jatom
      ENDDO ! ispin
      !
      !
      ! Finalize
      CALL timestop(handle)
      !
   END SUBROUTINE issc_issc

! **************************************************************************************************
!> \brief ...
!> \param issc_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE issc_print(issc_env, qs_env)
      TYPE(issc_env_type)                                :: issc_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'issc_print', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol_i, element_symbol_j
      CHARACTER(LEN=default_string_length)               :: name_i, name_j, title
      INTEGER                                            :: iatom, jatom, natom, output_unit, &
                                                            unit_atoms
      LOGICAL                                            :: do_dso, do_fc, do_pso, do_sd, gapw
      REAL(dp)                                           :: eig(3), issc_iso_dso, issc_iso_fc, &
                                                            issc_iso_pso, issc_iso_sd, &
                                                            issc_iso_tot, issc_tmp(3, 3)
      REAL(dp), DIMENSION(:, :, :, :, :), POINTER        :: issc
      REAL(dp), EXTERNAL                                 :: DDOT
      TYPE(atomic_kind_type), POINTER                    :: atom_kind_i, atom_kind_j
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: issc_section

      NULLIFY (logger, particle_set, atom_kind_i, atom_kind_j, dft_control)

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      issc_section => section_vals_get_subs_vals(qs_env%input, &
                                                 "PROPERTIES%LINRES%SPINSPIN")

      CALL get_issc_env(issc_env=issc_env, &
                        issc=issc, &
                        do_fc=do_fc, &
                        do_sd=do_sd, &
                        do_pso=do_pso, &
                        do_dso=do_dso)
      !
      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      particle_set=particle_set)

      natom = SIZE(particle_set, 1)
      gapw = dft_control%qs_control%gapw

      !
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E14.6)') 'ISSC| CheckSum K =', &
            SQRT(DDOT(SIZE(issc), issc, 1, issc, 1))
      ENDIF
      !
      IF (BTEST(cp_print_key_should_output(logger%iter_info, issc_section, &
                                           "PRINT%K_MATRIX"), cp_p_file)) THEN

         unit_atoms = cp_print_key_unit_nr(logger, issc_section, "PRINT%K_MATRIX", &
                                           extension=".data", middle_name="K", log_filename=.FALSE.)

         IF (unit_atoms > 0) THEN
            WRITE (unit_atoms, *)
            WRITE (unit_atoms, *)
            WRITE (title, '(A)') "Indirect spin-spin coupling matrix"
            WRITE (unit_atoms, '(T2,A)') title
            DO iatom = 1, natom
               atom_kind_i => particle_set(iatom)%atomic_kind
               CALL get_atomic_kind(atom_kind_i, name=name_i, element_symbol=element_symbol_i)
               DO jatom = 1, natom
                  atom_kind_j => particle_set(jatom)%atomic_kind
                  CALL get_atomic_kind(atom_kind_j, name=name_j, element_symbol=element_symbol_j)
                  !
                  IF (iatom .EQ. jatom .AND. .NOT. do_dso) CYCLE
                  !
                  !
                  ! FC
                  issc_tmp(:, :) = issc(:, :, iatom, jatom, 1)
                  issc_tmp(:, :) = 0.5_dp*(issc_tmp(:, :)+TRANSPOSE(issc_tmp(:, :)))
                  CALL diamat_all(issc_tmp, eig)
                  issc_iso_fc = (eig(1)+eig(2)+eig(3))/3.0_dp
                  !
                  ! SD
                  issc_tmp(:, :) = issc(:, :, iatom, jatom, 2)
                  issc_tmp(:, :) = 0.5_dp*(issc_tmp(:, :)+TRANSPOSE(issc_tmp(:, :)))
                  CALL diamat_all(issc_tmp, eig)
                  issc_iso_sd = (eig(1)+eig(2)+eig(3))/3.0_dp
                  !
                  ! PSO
                  issc_tmp(:, :) = issc(:, :, iatom, jatom, 3)
                  issc_tmp(:, :) = 0.5_dp*(issc_tmp(:, :)+TRANSPOSE(issc_tmp(:, :)))
                  CALL diamat_all(issc_tmp, eig)
                  issc_iso_pso = (eig(1)+eig(2)+eig(3))/3.0_dp
                  !
                  ! DSO
                  issc_tmp(:, :) = issc(:, :, iatom, jatom, 4)
                  issc_tmp(:, :) = 0.5_dp*(issc_tmp(:, :)+TRANSPOSE(issc_tmp(:, :)))
                  CALL diamat_all(issc_tmp, eig)
                  issc_iso_dso = (eig(1)+eig(2)+eig(3))/3.0_dp
                  !
                  ! TOT
                  issc_iso_tot = issc_iso_fc+issc_iso_sd+issc_iso_dso+issc_iso_pso
                  !
                  !
                  WRITE (unit_atoms, *)
                  WRITE (unit_atoms, '(T2,2(A,I5,A,2X,A2))') 'Indirect spin-spin coupling between ', &
                     iatom, TRIM(name_i), element_symbol_i, ' and ', &
                     jatom, TRIM(name_j), element_symbol_j
                  !
                  IF (do_fc) WRITE (unit_atoms, '(T1,A,f12.4,A)') ' Isotropic FC contribution  = ', issc_iso_fc, ' Hz'
                  IF (do_sd) WRITE (unit_atoms, '(T1,A,f12.4,A)') ' Isotropic SD contribution  = ', issc_iso_sd, ' Hz'
                  IF (do_pso) WRITE (unit_atoms, '(T1,A,f12.4,A)') ' Isotropic PSO contribution = ', issc_iso_pso, ' Hz'
                  !IF(do_dso) WRITE(unit_atoms,'(T1,A,f12.4,A)') ' Isotropic DSO contribution = ',issc_iso_dso,' Hz'
                  IF (do_dso) WRITE (unit_atoms, '(T1,A,f12.4,A)') ' !!! POLARIZABILITY (for the moment) = ', issc_iso_dso, ' Hz'
                  IF (.NOT. do_dso) WRITE (unit_atoms, '(T1,A,f12.4,A)') ' Isotropic coupling         = ', issc_iso_tot, ' Hz'
               ENDDO
            ENDDO
         ENDIF
         CALL cp_print_key_finished_output(unit_atoms, logger, issc_section,&
              &                            "PRINT%K_MATRIX")
      ENDIF
      !
      !
   END SUBROUTINE issc_print

! **************************************************************************************************
!> \brief Initialize the issc environment
!> \param issc_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE issc_env_init(issc_env, qs_env)
      !
      TYPE(issc_env_type)                                :: issc_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'issc_env_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, idir, ini, ir, ispin, &
                                                            istat, m, n, n_rep, nao, natom, &
                                                            nspins, output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, last_sgf
      INTEGER, DIMENSION(:), POINTER                     :: list, row_blk_sizes
      LOGICAL                                            :: gapw
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: issc_section, lr_section

!

      CALL timeset(routineN, handle)

      NULLIFY (linres_control)
      NULLIFY (logger, issc_section)
      NULLIFY (tmp_fm_struct)
      NULLIFY (particle_set, qs_kind_set)
      NULLIFY (sab_orb)

      logger => cp_get_default_logger()
      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")

      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")

      IF (issc_env%ref_count /= 0) THEN
         CALL issc_env_cleanup(issc_env)
      ENDIF

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(/,T20,A,/)") "*** Start indirect spin-spin coupling Calculation ***"
         WRITE (output_unit, "(T10,A,/)") "Inizialization of the ISSC environment"
      ENDIF

      CALL issc_env_create(issc_env)
      !
      issc_section => section_vals_get_subs_vals(qs_env%input, &
           &          "PROPERTIES%LINRES%SPINSPIN")
      !CALL section_vals_val_get(nmr_section,"INTERPOLATE_SHIFT",l_val=nmr_env%interpolate_shift)
      !CALL section_vals_val_get(nmr_section,"SHIFT_GAPW_RADIUS",r_val=nmr_env%shift_gapw_radius)

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      linres_control=linres_control, &
                      mos=mos, &
                      sab_orb=sab_orb, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      dbcsr_dist=dbcsr_dist)
      !
      !
      gapw = dft_control%qs_control%gapw
      nspins = dft_control%nspins
      natom = SIZE(particle_set, 1)
      !
      ! check that the psi0 are localized and you have all the centers
      IF (.NOT. linres_control%localized_psi0) &
         CALL cp_warn(cp__l("qs_linres_issc_utils.F",848), 'To get indirect spin-spin coupling parameters within '// &
                      'PBC you need to localize zero order orbitals')
      !
      !
      ! read terms need to be calculated
      ! FC
      CALL section_vals_val_get(issc_section, "DO_FC", l_val=issc_env%do_fc)
      ! SD
      CALL section_vals_val_get(issc_section, "DO_SD", l_val=issc_env%do_sd)
      ! PSO
      CALL section_vals_val_get(issc_section, "DO_PSO", l_val=issc_env%do_pso)
      ! DSO
      CALL section_vals_val_get(issc_section, "DO_DSO", l_val=issc_env%do_dso)
      !
      !
      ! read the list of atoms on which the issc need to be calculated
      CALL section_vals_val_get(issc_section, "ISSC_ON_ATOM_LIST", n_rep_val=n_rep)
      !
      !
      NULLIFY (issc_env%issc_on_atom_list)
      n = 0
      DO ir = 1, n_rep
         NULLIFY (list)
         CALL section_vals_val_get(issc_section, "ISSC_ON_ATOM_LIST", i_rep_val=ir, i_vals=list)
         IF (ASSOCIATED(list)) THEN
            CALL reallocate(issc_env%issc_on_atom_list, 1, n+SIZE(list))
            DO ini = 1, SIZE(list)
               issc_env%issc_on_atom_list(ini+n) = list(ini)
            ENDDO
            n = n+SIZE(list)
         ENDIF
      ENDDO
      !
      IF (.NOT. ASSOCIATED(issc_env%issc_on_atom_list)) THEN
         ALLOCATE (issc_env%issc_on_atom_list(natom), STAT=istat)
         IF(.NOT.(istat .EQ. 0))CALL cp__a("qs_linres_issc_utils.F",883)
         DO iatom = 1, natom
            issc_env%issc_on_atom_list(iatom) = iatom
         ENDDO
      ENDIF
      issc_env%issc_natms = SIZE(issc_env%issc_on_atom_list)
      !
      !
      ! Initialize the issc tensor
      ALLOCATE (issc_env%issc(3, 3, issc_env%issc_natms, issc_env%issc_natms, 4), &
                issc_env%issc_loc(3, 3, issc_env%issc_natms, issc_env%issc_natms, 4), &
                STAT=istat)
      IF(.NOT.(istat == 0))CALL cp__a("qs_linres_issc_utils.F",895)
      issc_env%issc(:, :, :, :, :) = 0.0_dp
      issc_env%issc_loc(:, :, :, :, :) = 0.0_dp
      !
      ! allocation
      ALLOCATE (issc_env%efg_psi0(nspins, 6), issc_env%pso_psi0(nspins, 3), issc_env%fc_psi0(nspins), &
                issc_env%psi1_efg(nspins, 6), issc_env%psi1_pso(nspins, 3), issc_env%psi1_fc(nspins), &
                issc_env%dso_psi0(nspins, 3), issc_env%psi1_dso(nspins, 3), &
                STAT=istat)
      IF(.NOT.(istat == 0))CALL cp__a("qs_linres_issc_utils.F",904)
      DO ispin = 1, nspins
         !mo_coeff => current_env%psi0_order(ispin)%matrix
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_info(mo_coeff, ncol_global=m, nrow_global=nao)

         NULLIFY (tmp_fm_struct)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                  ncol_global=m, &
                                  context=mo_coeff%matrix_struct%context)
         DO idir = 1, 6
            NULLIFY (issc_env%psi1_efg(ispin, idir)%matrix, issc_env%efg_psi0(ispin, idir)%matrix)
            CALL cp_fm_create(issc_env%psi1_efg(ispin, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(issc_env%efg_psi0(ispin, idir)%matrix, tmp_fm_struct)
         ENDDO
         DO idir = 1, 3
            NULLIFY (issc_env%psi1_pso(ispin, idir)%matrix, issc_env%pso_psi0(ispin, idir)%matrix, &
                     issc_env%dso_psi0(ispin, idir)%matrix)
            CALL cp_fm_create(issc_env%psi1_pso(ispin, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(issc_env%pso_psi0(ispin, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(issc_env%psi1_dso(ispin, idir)%matrix, tmp_fm_struct)
            CALL cp_fm_create(issc_env%dso_psi0(ispin, idir)%matrix, tmp_fm_struct)
         ENDDO
         NULLIFY (issc_env%psi1_fc(ispin)%matrix, issc_env%fc_psi0(ispin)%matrix)
         CALL cp_fm_create(issc_env%psi1_fc(ispin)%matrix, tmp_fm_struct)
         CALL cp_fm_create(issc_env%fc_psi0(ispin)%matrix, tmp_fm_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
      ENDDO
      !
      ! prepare for allocation
      ALLOCATE (first_sgf(natom))
      ALLOCATE (last_sgf(natom))
      CALL get_particle_set(particle_set, qs_kind_set, &
                            first_sgf=first_sgf, &
                            last_sgf=last_sgf)
      ALLOCATE (row_blk_sizes(natom))
      CALL convert_offsets_to_sizes(first_sgf, row_blk_sizes, last_sgf)
      DEALLOCATE (first_sgf)
      DEALLOCATE (last_sgf)

      !
      ! efg, pso and fc operators
      CALL cp_dbcsr_allocate_matrix_set(issc_env%matrix_efg, 6)
      ALLOCATE (issc_env%matrix_efg(1)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_efg(1)%matrix)
      CALL cp_dbcsr_create(matrix=issc_env%matrix_efg(1)%matrix, &
                           name="efg (3xx-rr)/3", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_symmetric, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0, mutable_work=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(issc_env%matrix_efg(1)%matrix, sab_orb)

      ALLOCATE (issc_env%matrix_efg(2)%matrix, &
                issc_env%matrix_efg(3)%matrix, issc_env%matrix_efg(4)%matrix, &
                issc_env%matrix_efg(5)%matrix, issc_env%matrix_efg(6)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_efg(2)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_efg(3)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_efg(4)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_efg(5)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_efg(6)%matrix)
      CALL cp_dbcsr_copy(issc_env%matrix_efg(2)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'efg xy')
      CALL cp_dbcsr_copy(issc_env%matrix_efg(3)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'efg xz')
      CALL cp_dbcsr_copy(issc_env%matrix_efg(4)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'efg (3yy-rr)/3')
      CALL cp_dbcsr_copy(issc_env%matrix_efg(5)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'efg yz')
      CALL cp_dbcsr_copy(issc_env%matrix_efg(6)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'efg (3zz-rr)/3')

      CALL cp_dbcsr_set(issc_env%matrix_efg(1)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_efg(2)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_efg(3)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_efg(4)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_efg(5)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_efg(6)%matrix, 0.0_dp)
      !
      ! PSO
      CALL cp_dbcsr_allocate_matrix_set(issc_env%matrix_pso, 3)
      ALLOCATE (issc_env%matrix_pso(1)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_pso(1)%matrix)
      CALL cp_dbcsr_create(matrix=issc_env%matrix_pso(1)%matrix, &
                           name="pso x", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_antisymmetric, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0, mutable_work=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(issc_env%matrix_pso(1)%matrix, sab_orb)

      ALLOCATE (issc_env%matrix_pso(2)%matrix, issc_env%matrix_pso(3)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_pso(2)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_pso(3)%matrix)
      CALL cp_dbcsr_copy(issc_env%matrix_pso(2)%matrix, issc_env%matrix_pso(1)%matrix, &
                         'pso y')
      CALL cp_dbcsr_copy(issc_env%matrix_pso(3)%matrix, issc_env%matrix_pso(1)%matrix, &
                         'pso z')
      CALL cp_dbcsr_set(issc_env%matrix_pso(1)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_pso(2)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_pso(3)%matrix, 0.0_dp)
      !
      ! DSO
      CALL cp_dbcsr_allocate_matrix_set(issc_env%matrix_dso, 3)
      ALLOCATE (issc_env%matrix_dso(1)%matrix, issc_env%matrix_dso(2)%matrix, issc_env%matrix_dso(3)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_dso(1)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_dso(2)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_dso(3)%matrix)
      CALL cp_dbcsr_copy(issc_env%matrix_dso(1)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'dso x')
      CALL cp_dbcsr_copy(issc_env%matrix_dso(2)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'dso y')
      CALL cp_dbcsr_copy(issc_env%matrix_dso(3)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'dso z')
      CALL cp_dbcsr_set(issc_env%matrix_dso(1)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_dso(2)%matrix, 0.0_dp)
      CALL cp_dbcsr_set(issc_env%matrix_dso(3)%matrix, 0.0_dp)
      !
      ! FC
      CALL cp_dbcsr_allocate_matrix_set(issc_env%matrix_fc, 1)
      ALLOCATE (issc_env%matrix_fc(1)%matrix)
      CALL cp_dbcsr_init(issc_env%matrix_fc(1)%matrix)
      CALL cp_dbcsr_copy(issc_env%matrix_fc(1)%matrix, issc_env%matrix_efg(1)%matrix, &
                         'fc')
      CALL cp_dbcsr_set(issc_env%matrix_fc(1)%matrix, 0.0_dp)

      DEALLOCATE (row_blk_sizes)
      !
      ! Conversion factors
      IF (output_unit > 0) THEN
         WRITE (output_unit, "(T2,A,T60,I4,A)")&
              & "ISSC| spin-spin coupling computed for ", issc_env%issc_natms, ' atoms'
      ENDIF

      CALL cp_print_key_finished_output(output_unit, logger, lr_section,&
           &                            "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE issc_env_init

! **************************************************************************************************
!> \brief Deallocate the issc environment
!> \param issc_env ...
!> \par History
! **************************************************************************************************
   SUBROUTINE issc_env_cleanup(issc_env)

      TYPE(issc_env_type)                                :: issc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'issc_env_cleanup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idir, ispin

      issc_env%ref_count = issc_env%ref_count-1
      IF (issc_env%ref_count == 0) THEN
         IF (ASSOCIATED(issc_env%issc_on_atom_list)) THEN
            DEALLOCATE (issc_env%issc_on_atom_list)
         ENDIF
         IF (ASSOCIATED(issc_env%issc)) THEN
            DEALLOCATE (issc_env%issc)
         ENDIF
         IF (ASSOCIATED(issc_env%issc_loc)) THEN
            DEALLOCATE (issc_env%issc_loc)
         ENDIF
         !
         !efg_psi0
         IF (ASSOCIATED(issc_env%efg_psi0)) THEN
            DO idir = 1, SIZE(issc_env%efg_psi0, 2)
               DO ispin = 1, SIZE(issc_env%efg_psi0, 1)
                  CALL cp_fm_release(issc_env%efg_psi0(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (issc_env%efg_psi0)
         ENDIF
         !
         !pso_psi0
         IF (ASSOCIATED(issc_env%pso_psi0)) THEN
            DO idir = 1, SIZE(issc_env%pso_psi0, 2)
               DO ispin = 1, SIZE(issc_env%pso_psi0, 1)
                  CALL cp_fm_release(issc_env%pso_psi0(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (issc_env%pso_psi0)
         ENDIF
         !
         !dso_psi0
         IF (ASSOCIATED(issc_env%dso_psi0)) THEN
            DO idir = 1, SIZE(issc_env%dso_psi0, 2)
               DO ispin = 1, SIZE(issc_env%dso_psi0, 1)
                  CALL cp_fm_release(issc_env%dso_psi0(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (issc_env%dso_psi0)
         ENDIF
         !
         !fc_psi0
         IF (ASSOCIATED(issc_env%fc_psi0)) THEN
            DO ispin = 1, SIZE(issc_env%fc_psi0, 1)
               CALL cp_fm_release(issc_env%fc_psi0(ispin)%matrix)
            ENDDO
            DEALLOCATE (issc_env%fc_psi0)
         ENDIF
         !
         !psi1_efg
         IF (ASSOCIATED(issc_env%psi1_efg)) THEN
            DO idir = 1, SIZE(issc_env%psi1_efg, 2)
               DO ispin = 1, SIZE(issc_env%psi1_efg, 1)
                  CALL cp_fm_release(issc_env%psi1_efg(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (issc_env%psi1_efg)
         ENDIF
         !
         !psi1_pso
         IF (ASSOCIATED(issc_env%psi1_pso)) THEN
            DO idir = 1, SIZE(issc_env%psi1_pso, 2)
               DO ispin = 1, SIZE(issc_env%psi1_pso, 1)
                  CALL cp_fm_release(issc_env%psi1_pso(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (issc_env%psi1_pso)
         ENDIF
         !
         !psi1_dso
         IF (ASSOCIATED(issc_env%psi1_dso)) THEN
            DO idir = 1, SIZE(issc_env%psi1_dso, 2)
               DO ispin = 1, SIZE(issc_env%psi1_dso, 1)
                  CALL cp_fm_release(issc_env%psi1_dso(ispin, idir)%matrix)
               ENDDO
            ENDDO
            DEALLOCATE (issc_env%psi1_dso)
         ENDIF
         !
         !psi1_fc
         IF (ASSOCIATED(issc_env%psi1_fc)) THEN
            DO ispin = 1, SIZE(issc_env%psi1_fc, 1)
               CALL cp_fm_release(issc_env%psi1_fc(ispin)%matrix)
            ENDDO
            DEALLOCATE (issc_env%psi1_fc)
         ENDIF
         !
         ! cubes
         !IF(ASSOCIATED(issc_env%list_cubes)) THEN
         !   DEALLOCATE(issc_env%list_cubes,STAT=istat)
         !   CPPostcondition(istat==0,cp_failure_level,routineP,failure)
         !ENDIF
         !
         !matrix_efg
         IF (ASSOCIATED(issc_env%matrix_efg)) THEN
            CALL cp_dbcsr_deallocate_matrix_set(issc_env%matrix_efg)
         ENDIF
         !
         !matrix_pso
         IF (ASSOCIATED(issc_env%matrix_pso)) THEN
            CALL cp_dbcsr_deallocate_matrix_set(issc_env%matrix_pso)
         ENDIF
         !
         !matrix_dso
         IF (ASSOCIATED(issc_env%matrix_dso)) THEN
            CALL cp_dbcsr_deallocate_matrix_set(issc_env%matrix_dso)
         ENDIF
         !
         !matrix_fc
         IF (ASSOCIATED(issc_env%matrix_fc)) THEN
            CALL cp_dbcsr_deallocate_matrix_set(issc_env%matrix_fc)
         ENDIF

      ENDIF ! ref count

   END SUBROUTINE issc_env_cleanup

END MODULE qs_linres_issc_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains the setup for  the calculation of properties by linear response
!>      by the application of second order density functional perturbation theory.
!>      The knowledge of the ground state energy, density and wavefunctions is assumed.
!>      Uses the self consistent approach.
!>      Properties that can be calculated : none
!> \par History
!>       created 06-2005 [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_linres_module
   USE bibliography,                    ONLY: Weber2009,&
                                              cite_reference
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type,&
                                              use_qmmm,&
                                              use_qs_force
   USE input_constants,                 ONLY: lr_current,&
                                              lr_none,&
                                              ot_precond_full_all,&
                                              ot_precond_full_kinetic,&
                                              ot_precond_full_single,&
                                              ot_precond_full_single_inverse,&
                                              ot_precond_none,&
                                              ot_precond_s_inverse
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_linres_current,               ONLY: current_build_chi,&
                                              current_build_current
   USE qs_linres_current_utils,         ONLY: current_env_cleanup,&
                                              current_env_init,&
                                              current_response
   USE qs_linres_epr_nablavks,          ONLY: epr_nablavks
   USE qs_linres_epr_ownutils,          ONLY: epr_g_print,&
                                              epr_g_so,&
                                              epr_g_soo,&
                                              epr_g_zke,&
                                              epr_ind_magnetic_field
   USE qs_linres_epr_utils,             ONLY: epr_env_cleanup,&
                                              epr_env_init
   USE qs_linres_issc_utils,            ONLY: issc_env_cleanup,&
                                              issc_env_init,&
                                              issc_issc,&
                                              issc_print,&
                                              issc_response
   USE qs_linres_methods,               ONLY: linres_localize
   USE qs_linres_nmr_shift,             ONLY: nmr_shift,&
                                              nmr_shift_print
   USE qs_linres_nmr_utils,             ONLY: nmr_env_cleanup,&
                                              nmr_env_init
   USE qs_linres_op,                    ONLY: current_operators,&
                                              issc_operators,&
                                              polar_operators
   USE qs_linres_polar_utils,           ONLY: polar_env_cleanup,&
                                              polar_env_init,&
                                              polar_polar,&
                                              polar_print,&
                                              polar_response
   USE qs_linres_types,                 ONLY: current_env_type,&
                                              epr_env_type,&
                                              issc_env_type,&
                                              linres_control_create,&
                                              linres_control_release,&
                                              linres_control_type,&
                                              nmr_env_type,&
                                              polar_env_type
   USE qs_mo_methods,                   ONLY: calculate_density_matrix
   USE qs_mo_types,                     ONLY: mo_set_p_type
   USE qs_p_env_methods,                ONLY: p_env_create,&
                                              p_env_psi0_changed
   USE qs_p_env_types,                  ONLY: p_env_release,&
                                              qs_p_env_type
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: linres_calculation, linres_calculation_low

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_module'

CONTAINS

! **************************************************************************************************
!> \brief Driver for the linear response calculatios
!> \param force_env ...
!> \par History
!>      06.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE linres_calculation(force_env)

      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'linres_calculation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CALL timeset(routineN, handle)

      NULLIFY (qs_env)

      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("qs_linres_module.F",124)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("qs_linres_module.F",125)

      SELECT CASE (force_env%in_use)
      CASE (use_qs_force)
         CALL force_env_get(force_env, qs_env=qs_env)
      CASE (use_qmmm)
         qs_env => force_env%qmmm_env%qs_env
      CASE DEFAULT
         CALL cp__b("qs_linres_module.F",133,"Doesnt recognize this force_env.")
      END SELECT

      qs_env%linres_run = .TRUE.

      CALL linres_calculation_low(qs_env)

      CALL timestop(handle)

   END SUBROUTINE linres_calculation

! **************************************************************************************************
!> \brief Linear response can be called as run type or as post scf calculation
!>      Initialize the perturbation environment
!>      Define which properties is to be calculated
!>      Start up the optimization of the response density and wfn
!> \param qs_env ...
!> \par History
!>      06.2005 created [MI]
!>      02.2013 added polarizability section [SL]
!> \author MI
! **************************************************************************************************
   SUBROUTINE linres_calculation_low(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'linres_calculation_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, output_unit
      LOGICAL                                            :: epr_present, issc_present, &
                                                            lr_calculation, nmr_present, &
                                                            polar_present
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(section_vals_type), POINTER                   :: lr_section, prop_section

      CALL timeset(routineN, handle)
      lr_calculation = .FALSE.
      nmr_present = .FALSE.
      epr_present = .FALSE.
      issc_present = .FALSE.
      polar_present = .FALSE.

      NULLIFY (dft_control, p_env, linres_control, logger, prop_section, lr_section)
      logger => cp_get_default_logger()

      lr_section => section_vals_get_subs_vals(qs_env%input, "PROPERTIES%LINRES")
      CALL section_vals_get(lr_section, explicit=lr_calculation)

      IF (lr_calculation) THEN
         CALL linres_init(lr_section, p_env, qs_env)
         output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                            extension=".linresLog")
         CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                         linres_control=linres_control)

         !The type of perturbation has not been defined yet
         linres_control%property = lr_none
         !
         ! We do NMR or EPR, then compute the current response
         prop_section => section_vals_get_subs_vals(lr_section, "NMR")
         CALL section_vals_get(prop_section, explicit=nmr_present)
         prop_section => section_vals_get_subs_vals(lr_section, "EPR")
         CALL section_vals_get(prop_section, explicit=epr_present)

         IF (nmr_present .OR. epr_present) THEN
            CALL nmr_epr_linres(linres_control, qs_env, p_env, dft_control, &
                                nmr_present, epr_present, output_unit)
         ENDIF
         !
         ! We do the indirect spin-spin coupling calculation
         prop_section => section_vals_get_subs_vals(lr_section, "SPINSPIN")
         CALL section_vals_get(prop_section, explicit=issc_present)

         IF (issc_present) THEN
            CALL issc_linres(linres_control, qs_env, p_env, dft_control)
         ENDIF

         !
         ! We do the polarizability calculation
         prop_section => section_vals_get_subs_vals(lr_section, "POLAR")
         CALL section_vals_get(prop_section, explicit=polar_present)

         IF (polar_present) THEN
            CALL polar_linres(qs_env, p_env)
         END IF

         ! Other possible LR calculations can be introduced here

         CALL p_env_release(p_env)

         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,/,T25,A,/,T3,A,/)") &
               REPEAT("=", 77), &
               "ENDED LINRES CALCULATION", &
               REPEAT("=", 77)
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                           "PRINT%PROGRAM_RUN_INFO")
!!    ELSE
!       output_unit = cp_logger_get_default_io_unit(logger)
!       IF(output_unit>0) THEN
!          WRITE (output_unit, "(2X,A)") "",&
!               "-----------------------------------------------------------------------------",&
!               "-              No LR calculation has been specified in the input            -",&
!               "                        cp2k is going to stop, bye bye                       ",&
!               "-----------------------------------------------------------------------------",&
!               ""
!       END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE linres_calculation_low

! **************************************************************************************************
!> \brief Initialize some general settings like the p_env
!>      Localize the psi0 if required
!> \param lr_section ...
!> \param p_env ...
!> \param qs_env ...
!> \par History
!>      06.2005 created [MI]
!> \author MI
!> \note
!>      - The localization should probably be always for all the occupied states
! **************************************************************************************************
   SUBROUTINE linres_init(lr_section, p_env, qs_env)

      TYPE(section_vals_type), POINTER                   :: lr_section
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'linres_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, output_unit
      LOGICAL                                            :: do_it
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: loc_section

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, lr_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".linresLog")
      NULLIFY (dft_control, linres_control, loc_section, rho, mos, matrix_ks, rho_ao)

      IF(.NOT.(.NOT. ASSOCIATED(p_env)))CALL cp__a("qs_linres_module.F",287)

      CALL linres_control_create(linres_control)
      CALL set_qs_env(qs_env=qs_env, linres_control=linres_control)
      CALL linres_control_release(linres_control)
      CALL get_qs_env(qs_env=qs_env, linres_control=linres_control, &
                      dft_control=dft_control, matrix_ks=matrix_ks, mos=mos, rho=rho)
      CALL qs_rho_get(rho, rho_ao=rho_ao)

      ! Localized Psi0 are required when the position operator has to be defined (nmr)
      loc_section => section_vals_get_subs_vals(lr_section, "LOCALIZE")
      CALL section_vals_val_get(loc_section, "_SECTION_PARAMETERS_", &
                                l_val=linres_control%localized_psi0)
      IF (linres_control%localized_psi0) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T3,A,A)") &
               "Localization of the ground state orbitals", &
               " before starting the linear response calculation"
         END IF

         CALL linres_localize(qs_env, linres_control, dft_control%nspins)

         DO ispin = 1, dft_control%nspins
            CALL calculate_density_matrix(mos(ispin)%mo_set, rho_ao(ispin)%matrix)
         ENDDO
! ** update qs_env%rho
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
      END IF

      CALL section_vals_val_get(lr_section, "RESTART", l_val=linres_control%linres_restart)
      CALL section_vals_val_get(lr_section, "MAX_ITER", i_val=linres_control%max_iter)
      CALL section_vals_val_get(lr_section, "EPS", r_val=linres_control%eps)
      CALL section_vals_val_get(lr_section, "RESTART_EVERY", i_val=linres_control%restart_every)
      CALL section_vals_val_get(lr_section, "PRECONDITIONER", i_val=linres_control%preconditioner_type)
      CALL section_vals_val_get(lr_section, "ENERGY_GAP", r_val=linres_control%energy_gap)

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T3,A,/,T25,A,/,T3,A,/)") &
            REPEAT("=", 77), &
            "START LINRES CALCULATION", &
            REPEAT("=", 77)

         WRITE (UNIT=output_unit, FMT="(/,T10,A,/)") &
            "Properties to be Calulated:"
         CALL section_vals_val_get(lr_section, "NMR%_SECTION_PARAMETERS_", &
                                   l_val=do_it)
         IF (do_it) WRITE (UNIT=output_unit, FMT="(T45,A)") &
            "NMR Chemical Shift"

         IF (linres_control%localized_psi0) WRITE (UNIT=output_unit, FMT="(T2,A,T65,A)") &
            "LINRES|", " LOCALIZED PSI0"

         WRITE (UNIT=output_unit, FMT="(T2,A,T60,A)") &
            "LINRES| Optimization algorithm", " Conjugate Gradients"

         SELECT CASE (linres_control%preconditioner_type)
         CASE (ot_precond_none)
            WRITE (UNIT=output_unit, FMT="(T2,A,T60,A)") &
               "LINRES| Preconditioner", "                NONE"
         CASE (ot_precond_full_single)
            WRITE (UNIT=output_unit, FMT="(T2,A,T60,A)") &
               "LINRES| Preconditioner", "         FULL_SINGLE"
         CASE (ot_precond_full_kinetic)
            WRITE (UNIT=output_unit, FMT="(T2,A,T60,A)") &
               "LINRES| Preconditioner", "        FULL_KINETIC"
         CASE (ot_precond_s_inverse)
            WRITE (UNIT=output_unit, FMT="(T2,A,T60,A)") &
               "LINRES| Preconditioner", "      FULL_S_INVERSE"
         CASE (ot_precond_full_single_inverse)
            WRITE (UNIT=output_unit, FMT="(T2,A,T60,A)") &
               "LINRES| Preconditioner", " FULL_SINGLE_INVERSE"
         CASE (ot_precond_full_all)
            WRITE (UNIT=output_unit, FMT="(T2,A,T60,A)") &
               "LINRES| Preconditioner", "            FULL_ALL"
         CASE DEFAULT
            CALL cp__b("qs_linres_module.F",362,"Preconditioner NYI")
         END SELECT

         WRITE (UNIT=output_unit, FMT="(T2,A,T72,ES8.1)") &
            "LINRES| EPS", linres_control%eps
         WRITE (UNIT=output_unit, FMT="(T2,A,T72,I8)") &
            "LINRES| MAX_ITER", linres_control%max_iter
      END IF

      !------------------!
      ! create the p_env !
      !------------------!
      CALL p_env_create(p_env, qs_env, orthogonal_orbitals=.TRUE., linres_control=linres_control)

      ! update the m_epsilon matrix
      CALL p_env_psi0_changed(p_env, qs_env)

      ! calculate eigenvectros and eigenvalues of K
      p_env%os_valid = .FALSE.
      p_env%new_preconditioner = .TRUE.
      CALL cp_print_key_finished_output(output_unit, logger, lr_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE linres_init

! **************************************************************************************************
!> \brief ...
!> \param linres_control ...
!> \param qs_env ...
!> \param p_env ...
!> \param dft_control ...
!> \param nmr_present ...
!> \param epr_present ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE nmr_epr_linres(linres_control, qs_env, p_env, dft_control, nmr_present, epr_present, output_unit)

      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      LOGICAL                                            :: nmr_present, epr_present
      INTEGER                                            :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nmr_epr_linres', routineP = moduleN//':'//routineN

      INTEGER                                            :: iB
      LOGICAL                                            :: do_qmmm
      TYPE(current_env_type)                             :: current_env
      TYPE(epr_env_type)                                 :: epr_env
      TYPE(nmr_env_type)                                 :: nmr_env

      linres_control%property = lr_current

      CALL cite_reference(Weber2009)

      IF (.NOT. linres_control%localized_psi0) THEN
         CALL cp_abort(cp__l("qs_linres_module.F",419), &
                       "Are you sure that you want to calculate the chemical "// &
                       "shift without localized psi0?")
         CALL linres_localize(qs_env, linres_control, &
                              dft_control%nspins, centers_only=.TRUE.)
      ENDIF
      IF (dft_control%nspins /= 2 .AND. epr_present) THEN
         CALL cp__b("qs_linres_module.F",426,"LSD is needed to perform a g tensor calculation!")
      ENDIF
      !
      !Initialize the current environment
      do_qmmm = .FALSE.
      current_env%ref_count = 0
      IF (qs_env%qmmm) do_qmmm = .TRUE.
      current_env%do_qmmm = do_qmmm
      !current_env%prop='nmr'
      CALL current_env_init(current_env, qs_env)
      CALL current_operators(current_env, qs_env)
      CALL current_response(current_env, p_env, qs_env)
      !
      IF (current_env%all_pert_op_done) THEN
         !Initialize the nmr environment
         IF (nmr_present) THEN
            nmr_env%ref_count = 0
            CALL nmr_env_init(nmr_env, qs_env)
         ENDIF
         !
         !Initialize the epr environment
         IF (epr_present) THEN
            epr_env%ref_count = 0
            CALL epr_env_init(epr_env, qs_env)
            CALL epr_g_zke(epr_env, qs_env)
            CALL epr_nablavks(epr_env, qs_env)
         ENDIF
         !
         ! Build the rs_gauge if needed
         !CALL current_set_gauge(current_env,qs_env)
         !
         ! Loop over field direction
         DO iB = 1, 3
            !
            ! Build current response and succeptibility
            CALL current_build_current(current_env, qs_env, iB)
            CALL current_build_chi(current_env, qs_env, iB)
            !
            ! Compute NMR shift
            IF (nmr_present) THEN
               CALL nmr_shift(nmr_env, current_env, qs_env, iB)
            ENDIF
            !
            ! Compute EPR
            IF (epr_present) THEN
               CALL epr_ind_magnetic_field(epr_env, current_env, qs_env, iB)
               CALL epr_g_so(epr_env, current_env, qs_env, iB)
               CALL epr_g_soo(epr_env, current_env, qs_env, iB)
            ENDIF
         ENDDO
         !
         ! Finalized the nmr environment
         IF (nmr_present) THEN
            CALL nmr_shift_print(nmr_env, current_env, qs_env)
            CALL nmr_env_cleanup(nmr_env)
         ENDIF
         !
         ! Finalized the epr environment
         IF (epr_present) THEN
            CALL epr_g_print(epr_env, qs_env)
            CALL epr_env_cleanup(epr_env)
         ENDIF
         !
      ELSE
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(T10,A,/T20,A,/)") &
               "CURRENT: Not all responses to perturbation operators could be calculated.", &
               " Hence: NO nmr and NO epr possible."
         END IF
      END IF
      ! Finalized the current environment
      CALL current_env_cleanup(current_env, qs_env)

   END SUBROUTINE nmr_epr_linres

! **************************************************************************************************
!> \brief ...
!> \param linres_control ...
!> \param qs_env ...
!> \param p_env ...
!> \param dft_control ...
! **************************************************************************************************
   SUBROUTINE issc_linres(linres_control, qs_env, p_env, dft_control)

      TYPE(linres_control_type), POINTER                 :: linres_control
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'issc_linres', routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom
      LOGICAL                                            :: do_qmmm
      TYPE(current_env_type)                             :: current_env
      TYPE(issc_env_type)                                :: issc_env

      linres_control%property = lr_current
      IF (.NOT. linres_control%localized_psi0) THEN
         CALL cp_abort(cp__l("qs_linres_module.F",524), &
                       "Are you sure that you want to calculate the chemical "// &
                       "shift without localized psi0?")
         CALL linres_localize(qs_env, linres_control, &
                              dft_control%nspins, centers_only=.TRUE.)
      ENDIF
      !
      !Initialize the current environment
      do_qmmm = .FALSE.
      current_env%ref_count = 0
      IF (qs_env%qmmm) do_qmmm = .TRUE.
      current_env%do_qmmm = do_qmmm
      !current_env%prop='issc'
      !CALL current_env_init(current_env,qs_env)
      !CALL current_response(current_env,p_env,qs_env)
      !
      !Initialize the issc environment
      issc_env%ref_count = 0
      CALL issc_env_init(issc_env, qs_env)
      !
      ! Loop over atoms
      DO iatom = 1, issc_env%issc_natms
         CALL issc_operators(issc_env, qs_env, iatom)
         CALL issc_response(issc_env, p_env, qs_env)
         CALL issc_issc(issc_env, qs_env, iatom)
      ENDDO
      !
      ! Finalized the issc environment
      CALL issc_print(issc_env, qs_env)
      CALL issc_env_cleanup(issc_env)

   END SUBROUTINE issc_linres

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param p_env ...
! **************************************************************************************************
   SUBROUTINE polar_linres(qs_env, p_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_p_env_type), POINTER                       :: p_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'polar_linres', routineP = moduleN//':'//routineN

      TYPE(polar_env_type)                               :: polar_env

      polar_env%ref_count = 0
      CALL polar_env_init(polar_env, qs_env)
      CALL polar_operators(polar_env, qs_env)
      CALL polar_response(polar_env, p_env, qs_env)
      CALL polar_polar(polar_env, qs_env)
      !
      CALL polar_print(polar_env, qs_env)
      CALL polar_env_cleanup(polar_env)

   END SUBROUTINE polar_linres

END MODULE qs_linres_module

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the kinetic energy integrals over Cartesian
!>      Gaussian-type functions.
!>
!>      [a|T|b] = [a|-nabla**2/2|b]
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      - Derivatives added (10.05.2002,MK)
!>      - Fully refactored (07.07.2014,JGH)
!> \author Matthias Krack (31.07.2000)
! **************************************************************************************************
MODULE ai_kinetic
   USE ai_os_rr,                        ONLY: os_rr_ovlp
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_kinetic'

! *** Public subroutines ***

   PUBLIC :: kinetic

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the two-center kinetic energy integrals [a|T|b] over
!>          Cartesian Gaussian-type functions.
!> \param la_max Maximum L of basis on A
!> \param la_min Minimum L of basis on A
!> \param npgfa  Number of primitive functions in set of basis on A
!> \param rpgfa  Range of functions on A (used for prescreening)
!> \param zeta   Exponents of basis on center A
!> \param lb_max Maximum L of basis on A
!> \param lb_min Minimum L of basis on A
!> \param npgfb  Number of primitive functions in set of basis on B
!> \param rpgfb  Range of functions on B (used for prescreening)
!> \param zetb   Exponents of basis on center B
!> \param rab    Distance vector between centers A and B
!> \param kab    Kinetic energy integrals, optional
!> \param dab    First derivatives of Kinetic energy integrals, optional
!> \date    07.07.2014
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE kinetic(la_max, la_min, npgfa, rpgfa, zeta, &
                      lb_max, lb_min, npgfb, rpgfb, zetb, &
                      rab, kab, dab)
      INTEGER, INTENT(IN)                                :: la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: kab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT), &
         OPTIONAL                                        :: dab

      CHARACTER(len=*), PARAMETER :: routineN = 'kinetic', routineP = moduleN//':'//routineN

      INTEGER                                            :: ax, ay, az, bx, by, bz, coa, cob, ia, &
                                                            ib, idx, idy, idz, ipgf, jpgf, la, lb, &
                                                            ldrr, lma, lmb, ma, mb, na, nb, ofa, &
                                                            ofb
      REAL(KIND=dp)                                      :: a, b, dsx, dsy, dsz, dtx, dty, dtz, f0, &
                                                            rab2, tab, xhi, zet
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rr, tt
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp

! Distance of the centers a and b

      rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
      tab = SQRT(rab2)

      ! Maximum l for auxiliary integrals
      IF (PRESENT(kab)) THEN
         lma = la_max+1
         lmb = lb_max+1
      END IF
      IF (PRESENT(dab)) THEN
         lma = la_max+2
         lmb = lb_max+1
         idx = coset(1, 0, 0)-coset(0, 0, 0)
         idy = coset(0, 1, 0)-coset(0, 0, 0)
         idz = coset(0, 0, 1)-coset(0, 0, 0)
      END IF
      ldrr = MAX(lma, lmb)+1

      ! Allocate space for auxiliary integrals
      ALLOCATE (rr(0:ldrr-1, 0:ldrr-1, 3), tt(0:ldrr-1, 0:ldrr-1, 3))

      ! Number of integrals, check size of arrays
      ofa = ncoset(la_min-1)
      ofb = ncoset(lb_min-1)
      na = ncoset(la_max)-ofa
      nb = ncoset(lb_max)-ofb
      IF (PRESENT(kab)) THEN
         IF(.NOT.((SIZE(kab, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_kinetic.F",109)
         IF(.NOT.((SIZE(kab, 2) >= nb*npgfb)))CALL cp__a("aobasis/ai_kinetic.F",110)
      END IF
      IF (PRESENT(dab)) THEN
         IF(.NOT.((SIZE(dab, 1) >= na*npgfa)))CALL cp__a("aobasis/ai_kinetic.F",113)
         IF(.NOT.((SIZE(dab, 2) >= nb*npgfb)))CALL cp__a("aobasis/ai_kinetic.F",114)
         IF(.NOT.((SIZE(dab, 3) >= 3)))CALL cp__a("aobasis/ai_kinetic.F",115)
      END IF

      ! Loops over all pairs of primitive Gaussian-type functions
      ma = 0
      DO ipgf = 1, npgfa
         mb = 0
         DO jpgf = 1, npgfb
            ! Distance Screening
            IF (rpgfa(ipgf)+rpgfb(jpgf) < tab) THEN
               IF (PRESENT(kab)) kab(ma+1:ma+na, mb+1:mb+nb) = 0.0_dp
               IF (PRESENT(dab)) dab(ma+1:ma+na, mb+1:mb+nb, 1:3) = 0.0_dp
               mb = mb+nb
               CYCLE
            ENDIF

            ! Calculate some prefactors
            a = zeta(ipgf)
            b = zetb(jpgf)
            zet = a+b
            xhi = a*b/zet
            rap = b*rab/zet
            rbp = -a*rab/zet

            ! [s|s] integral
            f0 = 0.5_dp*(pi/zet)**(1.5_dp)*EXP(-xhi*rab2)

            ! Calculate the reccurence relation, overlap
            CALL os_rr_ovlp(rap, lma, rbp, lmb, zet, ldrr, rr)

            ! kinetic energy auxiliary integrals, overlap of [da/dx|db/dx]
            DO la = 0, lma-1
               DO lb = 0, lmb-1
                  tt(la, lb, 1) = 4.0_dp*a*b*rr(la+1, lb+1, 1)
                  tt(la, lb, 2) = 4.0_dp*a*b*rr(la+1, lb+1, 2)
                  tt(la, lb, 3) = 4.0_dp*a*b*rr(la+1, lb+1, 3)
                  IF (la > 0 .AND. lb > 0) THEN
                     tt(la, lb, 1) = tt(la, lb, 1)+REAL(la*lb, dp)*rr(la-1, lb-1, 1)
                     tt(la, lb, 2) = tt(la, lb, 2)+REAL(la*lb, dp)*rr(la-1, lb-1, 2)
                     tt(la, lb, 3) = tt(la, lb, 3)+REAL(la*lb, dp)*rr(la-1, lb-1, 3)
                  END IF
                  IF (la > 0) THEN
                     tt(la, lb, 1) = tt(la, lb, 1)-2.0_dp*REAL(la, dp)*b*rr(la-1, lb+1, 1)
                     tt(la, lb, 2) = tt(la, lb, 2)-2.0_dp*REAL(la, dp)*b*rr(la-1, lb+1, 2)
                     tt(la, lb, 3) = tt(la, lb, 3)-2.0_dp*REAL(la, dp)*b*rr(la-1, lb+1, 3)
                  END IF
                  IF (lb > 0) THEN
                     tt(la, lb, 1) = tt(la, lb, 1)-2.0_dp*REAL(lb, dp)*a*rr(la+1, lb-1, 1)
                     tt(la, lb, 2) = tt(la, lb, 2)-2.0_dp*REAL(lb, dp)*a*rr(la+1, lb-1, 2)
                     tt(la, lb, 3) = tt(la, lb, 3)-2.0_dp*REAL(lb, dp)*a*rr(la+1, lb-1, 3)
                  END IF
               END DO
            END DO

            DO lb = lb_min, lb_max
            DO bx = 0, lb
            DO by = 0, lb-bx
               bz = lb-bx-by
               cob = coset(bx, by, bz)-ofb
               ib = mb+cob
               DO la = la_min, la_max
               DO ax = 0, la
               DO ay = 0, la-ax
                  az = la-ax-ay
                  coa = coset(ax, ay, az)-ofa
                  ia = ma+coa
                  ! integrals
                  IF (PRESENT(kab)) THEN
                     kab(ia, ib) = f0*(tt(ax, bx, 1)*rr(ay, by, 2)*rr(az, bz, 3)+ &
                                       rr(ax, bx, 1)*tt(ay, by, 2)*rr(az, bz, 3)+ &
                                       rr(ax, bx, 1)*rr(ay, by, 2)*tt(az, bz, 3))
                  END IF
                  ! first derivatives
                  IF (PRESENT(dab)) THEN
                     ! dx
                     dsx = 2.0_dp*a*rr(ax+1, bx, 1)
                     IF (ax > 0) dsx = dsx-REAL(ax, dp)*rr(ax-1, bx, 1)
                     dtx = 2.0_dp*a*tt(ax+1, bx, 1)
                     IF (ax > 0) dtx = dtx-REAL(ax, dp)*tt(ax-1, bx, 1)
                     dab(ia, ib, idx) = dtx*rr(ay, by, 2)*rr(az, bz, 3)+ &
                                        dsx*(tt(ay, by, 2)*rr(az, bz, 3)+rr(ay, by, 2)*tt(az, bz, 3))
                     ! dy
                     dsy = 2.0_dp*a*rr(ay+1, by, 2)
                     IF (ay > 0) dsy = dsy-REAL(ay, dp)*rr(ay-1, by, 2)
                     dty = 2.0_dp*a*tt(ay+1, by, 2)
                     IF (ay > 0) dty = dty-REAL(ay, dp)*tt(ay-1, by, 2)
                     dab(ia, ib, idy) = dty*rr(ax, bx, 1)*rr(az, bz, 3)+ &
                                        dsy*(tt(ax, bx, 1)*rr(az, bz, 3)+rr(ax, bx, 1)*tt(az, bz, 3))
                     ! dz
                     dsz = 2.0_dp*a*rr(az+1, bz, 3)
                     IF (az > 0) dsz = dsz-REAL(az, dp)*rr(az-1, bz, 3)
                     dtz = 2.0_dp*a*tt(az+1, bz, 3)
                     IF (az > 0) dtz = dtz-REAL(az, dp)*tt(az-1, bz, 3)
                     dab(ia, ib, idz) = dtz*rr(ax, bx, 1)*rr(ay, by, 2)+ &
                                        dsz*(tt(ax, bx, 1)*rr(ay, by, 2)+rr(ax, bx, 1)*tt(ay, by, 2))
                     ! scale
                     dab(ia, ib, 1:3) = f0*dab(ia, ib, 1:3)
                  END IF
                  !
               ENDDO
               ENDDO
               ENDDO !la
            ENDDO
            ENDDO
            ENDDO !lb

            mb = mb+nb
         END DO
         ma = ma+na
      END DO

      DEALLOCATE (rr, tt)

   END SUBROUTINE kinetic

END MODULE ai_kinetic
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of kinetic energy matrix and forces
!> \par History
!>      JGH: from core_hamiltonian
!>      simplify further [7.2014]
!> \author Juerg Hutter
! **************************************************************************************************
MODULE qs_kinetic
   USE ai_contraction,                  ONLY: block_add,&
                                              contraction,&
                                              decontraction,&
                                              force_trace
   USE ai_kinetic,                      ONLY: kinetic
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_filter,&
                                              cp_dbcsr_finalize,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type,&
                                              dbcsr_distribution_obj
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE orbital_pointers,                ONLY: ncoset
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_integral_utils,               ONLY: basis_set_list_setup,&
                                              get_memory_usage
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              get_neighbor_list_set_p,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_overlap,                      ONLY: create_sab_matrix
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_kinetic'

! *** Public subroutines ***

   PUBLIC :: build_kinetic_matrix

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the kinetic energy matrix over Cartesian Gaussian functions.
!> \param   ks_env the QS environment
!> \param   matrix_t The kinetic energy matrix to be calculated (optional)
!> \param   matrixkp_t The kinetic energy matrices to be calculated (kpoints,optional)
!> \param   matrix_name The name of the matrix (i.e. for output)
!> \param   basis_type basis set to be used
!> \param   sab_nl pair list (must be consistent with basis sets!)
!> \param   calculate_forces (optional)
!> \param   matrix_p density matrix for force calculation (optional)
!> \param   matrixkp_p density matrix for force calculation with kpoints (optional)
!> \param   eps_filter Filter final matrix (optional)
!> \date    11.10.2010
!> \par     History
!>          Ported from qs_overlap, replaces code in build_core_hamiltonian
!>          Refactoring [07.2014] JGH
!>          Simplify options and use new kinetic energy integral routine
!>          kpoints [08.2014] JGH
!> \author  JGH
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE build_kinetic_matrix(ks_env, matrix_t, matrixkp_t, matrix_name, &
                                   basis_type, sab_nl, calculate_forces, matrix_p, matrixkp_p, &
                                   eps_filter)

      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: matrix_t
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: matrixkp_t
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: matrix_name
      CHARACTER(LEN=*), INTENT(IN)                       :: basis_type
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_nl
      LOGICAL, INTENT(IN), OPTIONAL                      :: calculate_forces
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_p
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: matrixkp_p
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: eps_filter

      CHARACTER(len=*), PARAMETER :: routineN = 'build_kinetic_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, handle, iatom, ic, icol, ikind, inode, irow, iset, jatom, jkind, &
         jset, ldsab, mepos, natom, ncoa, ncob, nimg, nkind, nseta, nsetb, nthread, sgfa, sgfb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, DIMENSION(3)                              :: cell
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: do_forces, do_symmetric, dokp, found, &
                                                            trans, use_cell_mapping, use_virial
      REAL(KIND=dp)                                      :: f0, ff, rab2, tab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: kab, pab, qab
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, rab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: k_block, p_block, rpgfa, rpgfb, scon_a, &
                                                            scon_b, zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      ! test for matrices (kpoints or standard gamma point)
      IF (PRESENT(matrix_t)) THEN
         dokp = .FALSE.
         use_cell_mapping = .FALSE.
      ELSEIF (PRESENT(matrixkp_t)) THEN
         dokp = .TRUE.
         CALL get_ks_env(ks_env=ks_env, kpoints=kpoints)
         CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
         use_cell_mapping = (SIZE(cell_to_index) > 1)
      ELSE
         CALL cp__b("qs_kinetic.F",153,"")
      END IF

      NULLIFY (atomic_kind_set, qs_kind_set, p_block, dft_control)
      CALL get_ks_env(ks_env, &
                      dft_control=dft_control, &
                      atomic_kind_set=atomic_kind_set, &
                      natom=natom, &
                      qs_kind_set=qs_kind_set, &
                      dbcsr_dist=dbcsr_dist)

      nimg = dft_control%nimages
      nkind = SIZE(atomic_kind_set)

      ALLOCATE (atom_of_kind(natom))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)

      do_forces = .FALSE.
      IF (PRESENT(calculate_forces)) do_forces = calculate_forces

      ! check for symmetry
      IF(.NOT.(SIZE(sab_nl) > 0))CALL cp__a("qs_kinetic.F",174)
      CALL get_neighbor_list_set_p(neighbor_list_sets=sab_nl, symmetric=do_symmetric)

      ! prepare basis set
      ALLOCATE (basis_set_list(nkind))
      CALL basis_set_list_setup(basis_set_list, basis_type, qs_kind_set)

      IF (dokp) THEN
         CALL cp_dbcsr_allocate_matrix_set(matrixkp_t, 1, nimg)
         CALL create_sab_matrix(ks_env, matrixkp_t, matrix_name, basis_set_list, basis_set_list, &
                                sab_nl, do_symmetric)
      ELSE
         CALL cp_dbcsr_allocate_matrix_set(matrix_t, 1)
         CALL create_sab_matrix(ks_env, matrix_t, matrix_name, basis_set_list, basis_set_list, &
                                sab_nl, do_symmetric)
      END IF

      IF (do_forces) THEN
         ! if forces -> maybe virial too
         CALL get_ks_env(ks_env=ks_env, force=force, virial=virial)
         use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
         ! we need density matrix for forces
         IF (dokp) THEN
            IF(.NOT.(PRESENT(matrixkp_p)))CALL cp__a("qs_kinetic.F",197)
         ELSE
            IF(.NOT.(PRESENT(matrix_p)))CALL cp__a("qs_kinetic.F",199)
         END IF
      END IF
      ! *** Allocate work storage ***
      ldsab = get_memory_usage(qs_kind_set, basis_type)

      nthread = 1
!$    nthread = omp_get_max_threads()
      ! Iterate of neighbor list
      CALL neighbor_list_iterator_create(nl_iterator, sab_nl, nthread=nthread)

!$OMP PARALLEL DEFAULT(NONE) &
!$OMP SHARED (nthread,do_forces,ldsab,nl_iterator, use_cell_mapping, do_symmetric,dokp,&
!$OMP         ncoset,use_virial,force,virial,matrix_t,matrixkp_t,&
!$OMP         matrix_p,basis_set_list, atom_of_kind, cell_to_index, matrixkp_p)  &
!$OMP PRIVATE (k_block,mepos,kab,qab,pab,ikind,jkind,inode,iatom,jatom,rab,cell,basis_set_a,basis_set_b,atom_a,atom_b,&
!$OMP          first_sgfa, la_max, la_min, npgfa, nsgfa, nseta, rpgfa, set_radius_a, ncoa, ncob, force_a, &
!$OMP          zeta, first_sgfb, lb_max, lb_min, npgfb, nsetb, rpgfb, set_radius_b, nsgfb, p_block, dab, tab, &
!$OMP          zetb, scon_a, scon_b, ic, irow, icol, f0, ff, found, trans, rab2, sgfa, sgfb, iset, jset )

      mepos = 0
!$    mepos = omp_get_thread_num()

      ALLOCATE (kab(ldsab, ldsab), qab(ldsab, ldsab))
      IF (do_forces) THEN
         ALLOCATE (dab(ldsab, ldsab, 3), pab(ldsab, ldsab))
      END IF

      DO WHILE (neighbor_list_iterate(nl_iterator, mepos=mepos) == 0)
         CALL get_iterator_info(nl_iterator, mepos=mepos, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab, cell=cell)
         atom_a = atom_of_kind(iatom)
         atom_b = atom_of_kind(jatom)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         scon_a => basis_set_a%scon
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         scon_b => basis_set_b%scon
         zetb => basis_set_b%zet

         IF (use_cell_mapping) THEN
            ic = cell_to_index(cell(1), cell(2), cell(3))
            IF(.NOT.(ic > 0))CALL cp__a("qs_kinetic.F",261)
         ELSE
            ic = 1
         END IF

         IF (do_symmetric) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF
            f0 = 2.0_dp
            IF (iatom == jatom) f0 = 1.0_dp
            ff = 2.0_dp
         ELSE
            irow = iatom
            icol = jatom
            f0 = 1.0_dp
            ff = 1.0_dp
         END IF
         NULLIFY (k_block)
         IF (dokp) THEN
            CALL cp_dbcsr_get_block_p(matrix=matrixkp_t(1, ic)%matrix, &
                                      row=irow, col=icol, BLOCK=k_block, found=found)
            IF(.NOT.(found))CALL cp__a("qs_kinetic.F",287)
         ELSE
            CALL cp_dbcsr_get_block_p(matrix=matrix_t(1)%matrix, &
                                      row=irow, col=icol, BLOCK=k_block, found=found)
            IF(.NOT.(found))CALL cp__a("qs_kinetic.F",291)
         END IF

         IF (do_forces) THEN
            NULLIFY (p_block)
            IF (dokp) THEN
               CALL cp_dbcsr_get_block_p(matrix=matrixkp_p(1, ic)%matrix, &
                                         row=irow, col=icol, block=p_block, found=found)
               IF(.NOT.(found))CALL cp__a("qs_kinetic.F",299)
            ELSE
               CALL cp_dbcsr_get_block_p(matrix=matrix_p, row=irow, col=icol, &
                                         block=p_block, found=found)
               IF(.NOT.(found))CALL cp__a("qs_kinetic.F",303)
            END IF
         END IF

         rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
         tab = SQRT(rab2)
         trans = do_symmetric .AND. (iatom > jatom)

         DO iset = 1, nseta

            ncoa = npgfa(iset)*(ncoset(la_max(iset))-ncoset(la_min(iset)-1))
            sgfa = first_sgfa(1, iset)

            DO jset = 1, nsetb

               IF (set_radius_a(iset)+set_radius_b(jset) < tab) CYCLE

               ncob = npgfb(jset)*(ncoset(lb_max(jset))-ncoset(lb_min(jset)-1))
               sgfb = first_sgfb(1, jset)

               IF (do_forces .AND. ASSOCIATED(p_block) .AND. ((iatom /= jatom) .OR. use_virial)) THEN
                  ! Decontract P matrix block
                  kab = 0.0_dp
                  CALL block_add("OUT", kab, nsgfa(iset), nsgfb(jset), p_block, sgfa, sgfb, trans=trans)
                  CALL decontraction(kab, pab, scon_a(:, sgfa:), ncoa, nsgfa(iset), scon_b(:, sgfb:), ncob, nsgfb(jset), &
                                     trans=trans)
                  ! calculate integrals and derivatives
                  CALL kinetic(la_max(iset), la_min(iset), npgfa(iset), rpgfa(:, iset), zeta(:, iset), &
                               lb_max(jset), lb_min(jset), npgfb(jset), rpgfb(:, jset), zetb(:, jset), &
                               rab, kab, dab)
                  CALL force_trace(force_a, dab, pab, ncoa, ncob, 3)
!$OMP CRITICAL(forceupate)
                  force(ikind)%kinetic(:, atom_a) = force(ikind)%kinetic(:, atom_a)+ff*force_a(:)
                  force(jkind)%kinetic(:, atom_b) = force(jkind)%kinetic(:, atom_b)-ff*force_a(:)
                  IF (use_virial) THEN
                     CALL virial_pair_force(virial%pv_virial, f0, force_a, rab)
                  END IF
!$OMP END CRITICAL(forceupate)
               ELSE
                  ! calclulate integrals
                  CALL kinetic(la_max(iset), la_min(iset), npgfa(iset), rpgfa(:, iset), zeta(:, iset), &
                               lb_max(jset), lb_min(jset), npgfb(jset), rpgfb(:, jset), zetb(:, jset), &
                               rab, kab)
               END IF
               ! Contraction step
               CALL contraction(kab, qab, ca=scon_a(:, sgfa:), na=ncoa, ma=nsgfa(iset), &
                                cb=scon_b(:, sgfb:), nb=ncob, mb=nsgfb(jset), &
                                trans=trans)
!$OMP CRITICAL(blockadd)
               CALL block_add("IN", qab, nsgfa(iset), nsgfb(jset), k_block, sgfa, sgfb, trans=trans)
!$OMP END CRITICAL(blockadd)

            END DO
         END DO

      END DO
      DEALLOCATE (kab, qab)
      IF (do_forces) DEALLOCATE (pab, dab)
!$OMP END PARALLEL
      CALL neighbor_list_iterator_release(nl_iterator)

      IF (dokp) THEN
         DO ic = 1, nimg
            CALL cp_dbcsr_finalize(matrixkp_t(1, ic)%matrix)
            IF (PRESENT(eps_filter)) THEN
               CALL cp_dbcsr_filter(matrixkp_t(1, ic)%matrix, eps_filter)
            END IF
         END DO
      ELSE
         CALL cp_dbcsr_finalize(matrix_t(1)%matrix)
         IF (PRESENT(eps_filter)) THEN
            CALL cp_dbcsr_filter(matrix_t(1)%matrix, eps_filter)
         END IF
      END IF

      ! Release work storage
      DEALLOCATE (atom_of_kind)
      DEALLOCATE (basis_set_list)

      CALL timestop(handle)

   END SUBROUTINE build_kinetic_matrix

END MODULE qs_kinetic

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of general three-center integrals over Cartesian
!>      Gaussian-type functions and a spherical operator centered at position C
!>
!>      <a|V(local)|b> = <a|F(|r-C|)|b>
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      - Based in part on code by MK
!> \par Parameters
!>      -  ax,ay,az   : Angular momentum index numbers of orbital a.
!>      -  bx,by,bz   : Angular momentum index numbers of orbital b.
!>      -  coset      : Cartesian orbital set pointer.
!>      -  dab        : Distance between the atomic centers a and b.
!>      -  dac        : Distance between the atomic centers a and c.
!>      -  dbc        : Distance between the atomic centers b and c.
!>      -  l{a,b}     : Angular momentum quantum number of shell a or b.
!>      -  l{a,b}_max : Maximum angular momentum quantum number of shell a or b.
!>      -  ncoset     : Number of Cartesian orbitals up to l.
!>      -  rab        : Distance vector between the atomic centers a and b.
!>      -  rac        : Distance vector between the atomic centers a and c.
!>      -  rbc        : Distance vector between the atomic centers b and c.
!>      -  rpgf{a,b,c}: Radius of the primitive Gaussian-type function a or b.
!>      -  zet{a,b}   : Exponents of the Gaussian-type functions a or b.
!> \author jhu (05.2011)
! **************************************************************************************************
MODULE ai_oneelectron

   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_oneelectron'

! *** Public subroutines ***

   PUBLIC :: os_3center

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of three-center integrals <a|c|b> over
!>           Cartesian Gaussian functions and a spherical potential
!>
!> \param la_max_set ...
!> \param la_min_set ...
!> \param npgfa ...
!> \param rpgfa ...
!> \param zeta ...
!> \param lb_max_set ...
!> \param lb_min_set ...
!> \param npgfb ...
!> \param rpgfb ...
!> \param zetb ...
!> \param auxint ...
!> \param rpgfc ...
!> \param rab ...
!> \param dab ...
!> \param rac ...
!> \param dac ...
!> \param rbc ...
!> \param dbc ...
!> \param vab ...
!> \param s ...
!> \param pab ...
!> \param force_a ...
!> \param force_b ...
!> \param fs ...
!> \date    May 2011
!> \author  Juerg Hutter
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE os_3center(la_max_set, la_min_set, npgfa, rpgfa, zeta, &
                         lb_max_set, lb_min_set, npgfb, rpgfb, zetb, auxint, rpgfc, &
                         rab, dab, rac, dac, rbc, dbc, vab, s, pab, force_a, force_b, fs)
      INTEGER, INTENT(IN)                                :: la_max_set, la_min_set, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max_set, lb_min_set, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      REAL(KIND=dp), DIMENSION(0:, :), INTENT(IN)        :: auxint
      REAL(KIND=dp), INTENT(IN)                          :: rpgfc
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: dab
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac
      REAL(KIND=dp), INTENT(IN)                          :: dac
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rbc
      REAL(KIND=dp), INTENT(IN)                          :: dbc
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: vab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: pab
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT), OPTIONAL :: force_a, force_b
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT), &
         OPTIONAL                                        :: fs

      CHARACTER(len=*), PARAMETER :: routineN = 'os_3center', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, cda, cdax, cday, cdaz, cdb, cdbx, cdby, cdbz, coa, coamx, &
         coamy, coamz, coapx, coapy, coapz, cob, cobmx, cobmy, cobmz, cobpx, cobpy, cobpz, da, &
         da_max, dax, day, daz, db, db_max, dbx, dby, dbz, i, ia, iap, iax, iay, iaz, ib, ibm, &
         ibx, iby, ibz, ii(3), iim(3), ij, ipgf, ir, ir1, ir2, irm(3), irr(3), irx, iry, irz, ix, &
         ixx(1), j, jj(3), jjp(3), jpgf, la, la_max, la_min, lb, lb_max, lb_min, llr, m, ma, mb, &
         mmax, na, nb
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: iiap
      LOGICAL                                            :: calculate_force_a, calculate_force_b
      REAL(KIND=dp)                                      :: aai, abx, fax, fay, faz, fbx, fby, fbz, &
                                                            ftz, orho, rho, s1, s2
      REAL(KIND=dp), DIMENSION(3)                        :: pai, pbi, pci

      IF (PRESENT(pab)) THEN
         IF(.NOT.(PRESENT(fs)))CALL cp__a("aobasis/ai_oneelectron.F",121)
         IF (PRESENT(force_a)) THEN
            calculate_force_a = .TRUE.
         ELSE
            calculate_force_a = .FALSE.
         END IF
         IF (PRESENT(force_b)) THEN
            calculate_force_b = .TRUE.
         ELSE
            calculate_force_b = .FALSE.
         END IF
      ELSE
         calculate_force_a = .FALSE.
         calculate_force_b = .FALSE.
      END IF

      IF (calculate_force_a) THEN
         da_max = 1
         force_a = 0.0_dp
      ELSE
         da_max = 0
      END IF

      IF (calculate_force_b) THEN
         db_max = 1
         force_b = 0.0_dp
      ELSE
         db_max = 0
      END IF

      la_max = la_max_set+da_max
      la_min = MAX(0, la_min_set-da_max)

      lb_max = lb_max_set+db_max
      lb_min = MAX(0, lb_min_set-db_max)

      mmax = la_max+lb_max

      ! precalculate indices for horizontal recursion
      ALLOCATE (iiap(ncoset(mmax), 3))
      DO ma = 0, mmax
         DO iax = 0, ma
            DO iay = 0, ma-iax
               iaz = ma-iax-iay
               ia = coset(iax, iay, iaz)
               jj(1) = iax; jj(2) = iay; jj(3) = iaz
               jjp = jj
               jjp(1) = jjp(1)+1
               iap = coset(jjp(1), jjp(2), jjp(3))
               iiap(ia, 1) = iap
               jjp = jj
               jjp(2) = jjp(2)+1
               iap = coset(jjp(1), jjp(2), jjp(3))
               iiap(ia, 2) = iap
               jjp = jj
               jjp(3) = jjp(3)+1
               iap = coset(jjp(1), jjp(2), jjp(3))
               iiap(ia, 3) = iap
            END DO
         END DO
      END DO

!   *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0

      DO ipgf = 1, npgfa

!     *** Screening ***
         IF (rpgfa(ipgf)+rpgfc < dac) THEN
            na = na+ncoset(la_max_set)
            CYCLE
         END IF

         nb = 0

         DO jpgf = 1, npgfb

!       *** Screening ***
            IF ((rpgfb(jpgf)+rpgfc < dbc) .OR. &
                (rpgfa(ipgf)+rpgfb(jpgf) < dab)) THEN
               nb = nb+ncoset(lb_max_set)
               CYCLE
            END IF

!       *** Calculate some prefactors ***
            rho = zeta(ipgf)+zetb(jpgf)
            pai(:) = zetb(jpgf)/rho*rab(:)
            pbi(:) = -zeta(ipgf)/rho*rab(:)
            pci(:) = -(zeta(ipgf)*rac(:)+zetb(jpgf)*rbc(:))/rho
            orho = 0.5_dp/rho

            ij = (ipgf-1)*npgfb+jpgf
            s(1, 1, 1:mmax+1) = auxint(0:mmax, ij)

            IF (la_max > 0) THEN
!         *** Recurrence steps: [s|c|s] -> [a|c|s]               ***
!         *** [a|c|s](m) = (Pi - Ai)*[a-1i|c|s](m) -             ***
!         ***              (Pi - Ci)*[a-1i|c|s](m+1)) +          ***
!         ***              Ni(a-1i)/2(a+b)*[a-2i|c|s](m) -       ***
!         ***              Ni(a-1i)/2(a+b)*[a-2i|c|s](m+1)       ***
               DO llr = 1, mmax
                  IF (llr == 1) THEN
                     DO m = 0, mmax-llr
                        s1 = s(1, 1, m+1)
                        s2 = s(1, 1, m+2)
                        s(2, 1, m+1) = pai(1)*s1-pci(1)*s2 ! [px|o|s]
                        s(3, 1, m+1) = pai(2)*s1-pci(2)*s2 ! [py|o|s]
                        s(4, 1, m+1) = pai(3)*s1-pci(3)*s2 ! [pz|o|s]
                     END DO
                  ELSE IF (llr == 2) THEN
                     DO m = 0, mmax-llr
                        s1 = s(1, 1, m+1)-s(1, 1, m+2)
                        s(5, 1, m+1) = pai(1)*s(2, 1, m+1)-pci(1)*s(2, 1, m+2)+orho*s1 ! [dx2|o|s]
                        s(6, 1, m+1) = pai(1)*s(3, 1, m+1)-pci(1)*s(3, 1, m+2) ! [dxy|o|s]
                        s(7, 1, m+1) = pai(1)*s(4, 1, m+1)-pci(1)*s(4, 1, m+2) ! [dxz|o|s]
                        s(8, 1, m+1) = pai(2)*s(3, 1, m+1)-pci(2)*s(3, 1, m+2)+orho*s1 ! [dy2|o|s]
                        s(9, 1, m+1) = pai(2)*s(4, 1, m+1)-pci(2)*s(4, 1, m+2) ! [dyz|o|s]
                        s(10, 1, m+1) = pai(3)*s(4, 1, m+1)-pci(3)*s(4, 1, m+2)+orho*s1 ! [dz2|o|s]
                     END DO
                  ELSE IF (llr == 3) THEN
                     DO m = 0, mmax-llr
                        s(11, 1, m+1) = pai(1)*s(5, 1, m+1)-pci(1)*s(5, 1, m+2) & ! [fx3 |o|s]
                                        +2._dp*orho*(s(2, 1, m+1)-s(2, 1, m+2))
                        s(12, 1, m+1) = pai(1)*s(6, 1, m+1)-pci(1)*s(6, 1, m+2) & ! [fx2y|o|s]
                                        +orho*(s(3, 1, m+1)-s(3, 1, m+2))
                        s(13, 1, m+1) = pai(1)*s(7, 1, m+1)-pci(1)*s(7, 1, m+2) & ! [fx2z|o|s]
                                        +orho*(s(4, 1, m+1)-s(4, 1, m+2))
                        s(14, 1, m+1) = pai(2)*s(6, 1, m+1)-pci(2)*s(6, 1, m+2) & ! [fxy2|o|s]
                                        +orho*(s(2, 1, m+1)-s(2, 1, m+2))
                        s(15, 1, m+1) = pai(1)*s(9, 1, m+1)-pci(1)*s(9, 1, m+2) ! [fxyz|o|s]
                        s(16, 1, m+1) = pai(3)*s(7, 1, m+1)-pci(3)*s(7, 1, m+2) & ! [fxz2|o|s]
                                        +orho*(s(2, 1, m+1)-s(2, 1, m+2))
                        s(17, 1, m+1) = pai(2)*s(8, 1, m+1)-pci(2)*s(8, 1, m+2) & ! [fy3 |o|s]
                                        +2._dp*orho*(s(3, 1, m+1)-s(3, 1, m+2))
                        s(18, 1, m+1) = pai(2)*s(9, 1, m+1)-pci(2)*s(9, 1, m+2) & ! [fy2z|o|s]
                                        +orho*(s(4, 1, m+1)-s(4, 1, m+2))
                        s(19, 1, m+1) = pai(3)*s(9, 1, m+1)-pci(3)*s(9, 1, m+2) & ! [fyz2|o|s]
                                        +orho*(s(3, 1, m+1)-s(3, 1, m+2))
                        s(20, 1, m+1) = pai(3)*s(10, 1, m+1)-pci(3)*s(10, 1, m+2) & ! [fz3 |o|s]
                                        +2._dp*orho*(s(4, 1, m+1)-s(4, 1, m+2))
                     END DO
                  ELSE IF (llr == 4) THEN
                     DO m = 0, mmax-llr
                        s(21, 1, m+1) = pai(1)*s(11, 1, m+1)-pci(1)*s(11, 1, m+2) & ! [gx4  |s|s]
                                        +3._dp*orho*(s(5, 1, m+1)-s(5, 1, m+2))
                        s(22, 1, m+1) = pai(1)*s(12, 1, m+1)-pci(1)*s(12, 1, m+2) & ! [gx3y |s|s]
                                        +2._dp*orho*(s(6, 1, m+1)-s(6, 1, m+2))
                        s(23, 1, m+1) = pai(1)*s(13, 1, m+1)-pci(1)*s(13, 1, m+2) & ! [gx3z |s|s]
                                        +2._dp*orho*(s(7, 1, m+1)-s(7, 1, m+2))
                        s(24, 1, m+1) = pai(1)*s(14, 1, m+1)-pci(1)*s(14, 1, m+2) & ! [gx2y2|s|s]
                                        +orho*(s(8, 1, m+1)-s(8, 1, m+2))
                        s(25, 1, m+1) = pai(1)*s(15, 1, m+1)-pci(1)*s(15, 1, m+2) & ! [gx2yz|s|s]
                                        +orho*(s(9, 1, m+1)-s(9, 1, m+2))
                        s(26, 1, m+1) = pai(1)*s(16, 1, m+1)-pci(1)*s(16, 1, m+2) & ! [gx2z2|s|s]
                                        +orho*(s(10, 1, m+1)-s(10, 1, m+2))
                        s(27, 1, m+1) = pai(1)*s(17, 1, m+1)-pci(1)*s(17, 1, m+2) ! [gxy3 |s|s]
                        s(28, 1, m+1) = pai(1)*s(18, 1, m+1)-pci(1)*s(18, 1, m+2) ! [gxy2z|s|s]
                        s(29, 1, m+1) = pai(1)*s(19, 1, m+1)-pci(1)*s(19, 1, m+2) ! [gxyz2|s|s]
                        s(30, 1, m+1) = pai(1)*s(20, 1, m+1)-pci(1)*s(20, 1, m+2) ! [gxz3 |s|s]
                        s(31, 1, m+1) = pai(2)*s(17, 1, m+1)-pci(2)*s(17, 1, m+2) & ! [gy4  |s|s]
                                        +3._dp*orho*(s(8, 1, m+1)-s(8, 1, m+2))
                        s(32, 1, m+1) = pai(2)*s(18, 1, m+1)-pci(2)*s(18, 1, m+2) & ! [gy3z |s|s]
                                        +2._dp*orho*(s(9, 1, m+1)-s(9, 1, m+2))
                        s(33, 1, m+1) = pai(2)*s(19, 1, m+1)-pci(2)*s(19, 1, m+2) & ! [gy2z2|s|s]
                                        +orho*(s(10, 1, m+1)-s(10, 1, m+2))
                        s(34, 1, m+1) = pai(2)*s(20, 1, m+1)-pci(2)*s(20, 1, m+2) ! [gyz3 |s|s]
                        s(35, 1, m+1) = pai(3)*s(20, 1, m+1)-pci(3)*s(20, 1, m+2) & ! [gz4  |s|s]
                                        +3._dp*orho*(s(10, 1, m+1)-s(10, 1, m+2))
                     END DO
                  ELSE
                     DO irx = 0, llr
                        DO iry = 0, llr-irx
                           irz = llr-irx-iry
                           irr(1) = irx; irr(2) = iry; irr(3) = irz
                           ixx = MAXLOC(irr)
                           ix = ixx(1)
                           ir = coset(irx, iry, irz)
                           irm = irr
                           irm(ix) = irm(ix)-1
                           aai = REAL(MAX(irm(ix), 0), dp)*orho
                           ir1 = coset(irm(1), irm(2), irm(3))
                           irm(ix) = irm(ix)-1
                           ir2 = coset(irm(1), irm(2), irm(3))
                           DO m = 0, mmax-llr
                              s(ir, 1, m+1) = pai(ix)*s(ir1, 1, m+1)-pci(ix)*s(ir1, 1, m+2) &
                                              +aai*(s(ir2, 1, m+1)-s(ir2, 1, m+2))
                           END DO
                        END DO
                     END DO
                  END IF
               END DO

!         *** Horizontal recurrence steps ***
!         *** [a|c|b+1i] = [a+1i|c|b] + (Ai - Bi)*[a|c|b] ***

               DO mb = 1, lb_max
                  DO ibx = 0, mb
                     DO iby = 0, mb-ibx
                        ibz = mb-ibx-iby
                        ib = coset(ibx, iby, ibz)
                        ii(1) = ibx; ii(2) = iby; ii(3) = ibz
                        ixx = MAXLOC(ii)
                        ix = ixx(1)
                        abx = -rab(ix)
                        iim = ii
                        iim(ix) = iim(ix)-1
                        ibm = coset(iim(1), iim(2), iim(3))
                        DO ia = 1, ncoset(mmax-mb)
                           iap = iiap(ia, ix)
                           s(ia, ib, 1) = s(iap, ibm, 1)+abx*s(ia, ibm, 1)
                        END DO
                     END DO
                  END DO
               END DO

            ELSE IF (lb_max > 0) THEN

!         *** Recurrence steps: [s|c|s] -> [s|c|b]               ***
!         *** [s|c|b](m) = (Pi - Bi)*[s|c|b-1i](m) -             ***
!         ***              (Pi - Ci)*[s|c|b-1i](m+1)) +          ***
!         ***              Ni(b-1i)/2(a+b)*[s|c|b-2i](m) -       ***
!         ***              Ni(b-1i)/2(a+b)*[s|c|b-2i](m+1)       ***
               DO llr = 1, lb_max
                  IF (llr == 1) THEN
                     DO m = 0, lb_max-llr
                        s1 = s(1, 1, m+1)
                        s2 = s(1, 1, m+2)
                        s(1, 2, m+1) = pbi(1)*s1-pci(1)*s2 ! [px|o|s]
                        s(1, 3, m+1) = pbi(2)*s1-pci(2)*s2 ! [py|o|s]
                        s(1, 4, m+1) = pbi(3)*s1-pci(3)*s2 ! [pz|o|s]
                     END DO
                  ELSE IF (llr == 2) THEN
                     DO m = 0, lb_max-llr
                        s1 = s(1, 1, m+1)-s(1, 1, m+2)
                        s(1, 5, m+1) = pbi(1)*s(1, 2, m+1)-pci(1)*s(1, 2, m+2)+orho*s1 ! [dx2|o|s]
                        s(1, 6, m+1) = pbi(1)*s(1, 3, m+1)-pci(1)*s(1, 3, m+2) ! [dxy|o|s]
                        s(1, 7, m+1) = pbi(1)*s(1, 4, m+1)-pci(1)*s(1, 4, m+2) ! [dxz|o|s]
                        s(1, 8, m+1) = pbi(2)*s(1, 3, m+1)-pci(2)*s(1, 3, m+2)+orho*s1 ! [dy2|o|s]
                        s(1, 9, m+1) = pbi(2)*s(1, 4, m+1)-pci(2)*s(1, 4, m+2) ! [dyz|o|s]
                        s(1, 10, m+1) = pbi(3)*s(1, 4, m+1)-pci(3)*s(1, 4, m+2)+orho*s1 ! [dz2|o|s]
                     END DO
                  ELSE IF (llr == 3) THEN
                     DO m = 0, lb_max-llr
                        s(1, 11, m+1) = pbi(1)*s(1, 5, m+1)-pci(1)*s(1, 5, m+2) & ! [fx3 |o|s]
                                        +2._dp*orho*(s(1, 2, m+1)-s(1, 2, m+2))
                        s(1, 12, m+1) = pbi(1)*s(1, 6, m+1)-pci(1)*s(1, 6, m+2) & ! [fx2y|o|s]
                                        +orho*(s(1, 3, m+1)-s(1, 3, m+2))
                        s(1, 13, m+1) = pbi(1)*s(1, 7, m+1)-pci(1)*s(1, 7, m+2) & ! [fx2z|o|s]
                                        +orho*(s(1, 4, m+1)-s(1, 4, m+2))
                        s(1, 14, m+1) = pbi(2)*s(1, 6, m+1)-pci(2)*s(1, 6, m+2) & ! [fxy2|o|s]
                                        +orho*(s(1, 2, m+1)-s(1, 2, m+2))
                        s(1, 15, m+1) = pbi(1)*s(1, 9, m+1)-pci(1)*s(1, 9, m+2) ! [fxyz|o|s]
                        s(1, 16, m+1) = pbi(3)*s(1, 7, m+1)-pci(3)*s(1, 7, m+2) & ! [fxz2|o|s]
                                        +orho*(s(1, 2, m+1)-s(1, 2, m+2))
                        s(1, 17, m+1) = pbi(2)*s(1, 8, m+1)-pci(2)*s(1, 8, m+2) & ! [fy3 |o|s]
                                        +2._dp*orho*(s(1, 3, m+1)-s(1, 3, m+2))
                        s(1, 18, m+1) = pbi(2)*s(1, 9, m+1)-pci(2)*s(1, 9, m+2) & ! [fy2z|o|s]
                                        +orho*(s(1, 4, m+1)-s(1, 4, m+2))
                        s(1, 19, m+1) = pbi(3)*s(1, 9, m+1)-pci(3)*s(1, 9, m+2) & ! [fyz2|o|s]
                                        +orho*(s(1, 3, m+1)-s(1, 3, m+2))
                        s(1, 20, m+1) = pbi(3)*s(1, 10, m+1)-pci(3)*s(1, 10, m+2) & ! [fz3 |o|s]
                                        +2._dp*orho*(s(1, 4, m+1)-s(1, 4, m+2))
                     END DO
                  ELSE
                     DO irx = 0, llr
                        DO iry = 0, llr-irx
                           irz = llr-irx-iry
                           irr(1) = irx; irr(2) = iry; irr(3) = irz
                           ixx = MAXLOC(irr)
                           ix = ixx(1)
                           ir = coset(irx, iry, irz)
                           irm = irr
                           irm(ix) = irm(ix)-1
                           aai = REAL(MAX(irm(ix), 0), dp)
                           ir1 = coset(irm(1), irm(2), irm(3))
                           irm(ix) = irm(ix)-1
                           ir2 = coset(irm(1), irm(2), irm(3))
                           DO m = 0, lb_max-llr
                              s(1, ir, m+1) = pbi(ix)*s(1, ir1, m+1)-pci(ix)*s(1, ir1, m+2) &
                                              +aai*orho*(s(1, ir2, m+1)-s(1, ir2, m+2))
                           END DO
                        END DO
                     END DO
                  END IF
               END DO

            END IF

!       *** Store the primitive three-center overlap integrals ***
            DO j = ncoset(lb_min_set-1)+1, ncoset(lb_max_set)
               DO i = ncoset(la_min_set-1)+1, ncoset(la_max_set)
                  vab(na+i, nb+j) = vab(na+i, nb+j)+s(i, j, 1)
               END DO
            END DO

!       *** Calculate the requested derivatives with respect  ***
!       *** to the nuclear coordinates of the atomic center a ***

            DO da = 0, da_max-1
               ftz = 2.0_dp*zeta(ipgf)
               DO dax = 0, da
                  DO day = 0, da-dax
                     daz = da-dax-day
                     cda = coset(dax, day, daz)
                     cdax = coset(dax+1, day, daz)
                     cday = coset(dax, day+1, daz)
                     cdaz = coset(dax, day, daz+1)

!             *** [da/dAi|c|b] = 2*zeta*[a+1i|c|b] - Ni(a)[a-1i|c|b] ***

                     DO la = la_min_set, la_max-da-1
                        DO ax = 0, la
                           fax = REAL(ax, dp)
                           DO ay = 0, la-ax
                              fay = REAL(ay, dp)
                              az = la-ax-ay
                              faz = REAL(az, dp)
                              coa = coset(ax, ay, az)
                              coamx = coset(ax-1, ay, az)
                              coamy = coset(ax, ay-1, az)
                              coamz = coset(ax, ay, az-1)
                              coapx = coset(ax+1, ay, az)
                              coapy = coset(ax, ay+1, az)
                              coapz = coset(ax, ay, az+1)
                              DO cob = ncoset(lb_min_set-1)+1, ncoset(lb_max_set)
                                 fs(coa, cob, cdax) = ftz*s(coapx, cob, cda)-fax*s(coamx, cob, cda)
                                 fs(coa, cob, cday) = ftz*s(coapy, cob, cda)-fay*s(coamy, cob, cda)
                                 fs(coa, cob, cdaz) = ftz*s(coapz, cob, cda)-faz*s(coamz, cob, cda)
                              END DO
                           END DO
                        END DO
                     END DO
                  END DO

               END DO
            END DO

!       *** Calculate the force contribution for the atomic center a ***

            IF (calculate_force_a) THEN
               DO j = ncoset(lb_min_set-1)+1, ncoset(lb_max_set)
                  DO i = ncoset(la_min_set-1)+1, ncoset(la_max_set)
                     force_a(1) = force_a(1)+pab(na+i, nb+j)*fs(i, j, 2)
                     force_a(2) = force_a(2)+pab(na+i, nb+j)*fs(i, j, 3)
                     force_a(3) = force_a(3)+pab(na+i, nb+j)*fs(i, j, 4)
                  END DO
               END DO
            END IF

!       *** Calculate the requested derivatives with respect  ***
!       *** to the nuclear coordinates of the atomic center b ***

            DO db = 0, db_max-1
               ftz = 2.0_dp*zetb(jpgf)
               DO dbx = 0, db
                  DO dby = 0, db-dbx
                     dbz = db-dbx-dby
                     cdb = coset(dbx, dby, dbz)
                     cdbx = coset(dbx+1, dby, dbz)
                     cdby = coset(dbx, dby+1, dbz)
                     cdbz = coset(dbx, dby, dbz+1)

!             *** [a|c|db/dBi] = 2*zetb*[a|c|b+1i] - Ni(b)[a|c|b-1i] ***

                     DO lb = lb_min_set, lb_max-db-1
                        DO bx = 0, lb
                           fbx = REAL(bx, dp)
                           DO by = 0, lb-bx
                              fby = REAL(by, dp)
                              bz = lb-bx-by
                              fbz = REAL(bz, dp)
                              cob = coset(bx, by, bz)
                              cobmx = coset(bx-1, by, bz)
                              cobmy = coset(bx, by-1, bz)
                              cobmz = coset(bx, by, bz-1)
                              cobpx = coset(bx+1, by, bz)
                              cobpy = coset(bx, by+1, bz)
                              cobpz = coset(bx, by, bz+1)
                              DO coa = ncoset(la_min_set-1)+1, ncoset(la_max_set)
                                 fs(coa, cob, cdbx) = ftz*s(coa, cobpx, cdb)-fbx*s(coa, cobmx, cdb)
                                 fs(coa, cob, cdby) = ftz*s(coa, cobpy, cdb)-fby*s(coa, cobmy, cdb)
                                 fs(coa, cob, cdbz) = ftz*s(coa, cobpz, cdb)-fbz*s(coa, cobmz, cdb)
                              END DO
                           END DO
                        END DO
                     END DO

                  END DO
               END DO
            END DO

!       *** Calculate the force contribution for the atomic center b ***

            IF (calculate_force_b) THEN
               DO j = ncoset(lb_min_set-1)+1, ncoset(lb_max_set)
                  DO i = ncoset(la_min_set-1)+1, ncoset(la_max_set)
                     force_b(1) = force_b(1)+pab(na+i, nb+j)*fs(i, j, 2)
                     force_b(2) = force_b(2)+pab(na+i, nb+j)*fs(i, j, 3)
                     force_b(3) = force_b(3)+pab(na+i, nb+j)*fs(i, j, 4)
                  END DO
               END DO
            END IF

            nb = nb+ncoset(lb_max_set)

         END DO

         na = na+ncoset(la_max_set)

      END DO

      DEALLOCATE (iiap)

   END SUBROUTINE os_3center
! **************************************************************************************************

END MODULE ai_oneelectron
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of three-center overlap integrals over Cartesian
!>      Gaussian-type functions for the second term V(ppl) of the local
!>      part of the Goedecker pseudopotential (GTH):
!>
!>      <a|V(local)|b> = <a|V(erf) + V(ppl)|b>
!>                     = <a|V(erf)|b> + <a|V(ppl)|b>
!>                     = <a|-Z(eff)*erf(SQRT(2)*alpha*r)/r +
!>                       (C1 + C2*(alpha*r)**2 + C3*(alpha*r)**4 +
!>                        C4*(alpha*r)**6)*exp(-(alpha*r)**2/2))|b>
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!>      S. Goedecker, M. Teter and J. Hutter, Phys. Rev. B 54, 1703 (1996)
!>      C. Hartwigsen, S. Goedecker and J. Hutter, Phys. Rev. B 58, 3641 (1998)
!> \par History
!>      - Derivatives added (17.05.2002,MK)
!>      - Complete refactoring (05.2011,jhu)
!> \author Matthias Krack (04.10.2000)
! **************************************************************************************************
MODULE ai_overlap_ppl
   USE ai_oneelectron,                  ONLY: os_3center
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fac,&
                                              pi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_overlap_ppl'

! *** Public subroutines ***

   PUBLIC :: ppl_integral

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of three-center overlap integrals <a|c|b> over
!>           Cartesian Gaussian functions for the local part of the Goedecker
!>           pseudopotential (GTH). c is a primitive Gaussian-type function
!>           with a set of even angular momentum indices.
!>
!>           <a|V(ppl)|b> = <a| (C1 + C2*(alpha*r)**2 + C3*(alpha*r)**4 +
!>                               C4*(alpha*r)**6)*exp(-(alpha*r)**2/2))|b>
!>           zetc = alpha**2/2
!>
!> \param la_max_set ...
!> \param la_min_set ...
!> \param npgfa ...
!> \param rpgfa ...
!> \param zeta ...
!> \param lb_max_set ...
!> \param lb_min_set ...
!> \param npgfb ...
!> \param rpgfb ...
!> \param zetb ...
!> \param nexp_ppl ...
!> \param alpha_ppl ...
!> \param nct_ppl ...
!> \param cexp_ppl ...
!> \param rpgfc ...
!> \param rab ...
!> \param dab ...
!> \param rac ...
!> \param dac ...
!> \param rbc ...
!> \param dbc ...
!> \param vab ...
!> \param s ...
!> \param pab ...
!> \param force_a ...
!> \param force_b ...
!> \param fs ...
!> \date    May 2011
!> \author  Juerg Hutter
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE ppl_integral(la_max_set, la_min_set, npgfa, rpgfa, zeta, &
                           lb_max_set, lb_min_set, npgfb, rpgfb, zetb, nexp_ppl, alpha_ppl, nct_ppl, cexp_ppl, rpgfc, &
                           rab, dab, rac, dac, rbc, dbc, vab, s, pab, force_a, force_b, fs)
      INTEGER, INTENT(IN)                                :: la_max_set, la_min_set, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfa, zeta
      INTEGER, INTENT(IN)                                :: lb_max_set, lb_min_set, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rpgfb, zetb
      INTEGER, INTENT(IN)                                :: nexp_ppl
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: alpha_ppl
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nct_ppl
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: cexp_ppl
      REAL(KIND=dp), INTENT(IN)                          :: rpgfc
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: dab
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac
      REAL(KIND=dp), INTENT(IN)                          :: dac
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rbc
      REAL(KIND=dp), INTENT(IN)                          :: dbc
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: vab
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: pab
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT), OPTIONAL :: force_a, force_b
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT), &
         OPTIONAL                                        :: fs

      CHARACTER(len=*), PARAMETER :: routineN = 'ppl_integral', routineP = moduleN//':'//routineN

      INTEGER                                            :: iexp, ij, ipgf, jpgf, mmax, nexp
      REAL(KIND=dp)                                      :: rho, sab, t, zetc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: auxint
      REAL(KIND=dp), DIMENSION(3)                        :: pci

      IF (PRESENT(pab)) THEN
         IF(.NOT.(PRESENT(force_a)))CALL cp__a("aobasis/ai_overlap_ppl.F",119)
         IF(.NOT.(PRESENT(force_b)))CALL cp__a("aobasis/ai_overlap_ppl.F",120)
         IF(.NOT.(PRESENT(fs)))CALL cp__a("aobasis/ai_overlap_ppl.F",121)
         mmax = la_max_set+lb_max_set+2
         force_a(:) = 0.0_dp
         force_b(:) = 0.0_dp
      ELSE
         mmax = la_max_set+lb_max_set
      END IF

      ALLOCATE (auxint(0:mmax, npgfa*npgfb))
      auxint = 0._dp

      ! *** Calculate auxiliary integrals ***

      DO ipgf = 1, npgfa
         ! *** Screening ***
         IF (rpgfa(ipgf)+rpgfc < dac) CYCLE
         DO jpgf = 1, npgfb
            ! *** Screening ***
            IF ((rpgfb(jpgf)+rpgfc < dbc) .OR. &
                (rpgfa(ipgf)+rpgfb(jpgf) < dab)) CYCLE
            ij = (ipgf-1)*npgfb+jpgf
            rho = zeta(ipgf)+zetb(jpgf)
            pci(:) = -(zeta(ipgf)*rac(:)+zetb(jpgf)*rbc(:))/rho
            sab = EXP(-(zeta(ipgf)*zetb(jpgf)/rho*dab*dab))
            t = rho*SUM(pci(:)*pci(:))

            DO iexp = 1, nexp_ppl
               nexp = nct_ppl(iexp)
               zetc = alpha_ppl(iexp)
               CALL ppl_aux(auxint(0:mmax, ij), mmax, t, rho, nexp, cexp_ppl(:, iexp), zetc)
            END DO

            auxint(0:mmax, ij) = sab*auxint(0:mmax, ij)

         END DO
      END DO

      CALL os_3center(la_max_set, la_min_set, npgfa, rpgfa, zeta, &
                      lb_max_set, lb_min_set, npgfb, rpgfb, zetb, auxint, rpgfc, &
                      rab, dab, rac, dac, rbc, dbc, vab, s, pab, force_a, force_b, fs)

      DEALLOCATE (auxint)

   END SUBROUTINE ppl_integral
! **************************************************************************************************
!> \brief ...
!> \param auxint ...
!> \param mmax ...
!> \param t ...
!> \param rho ...
!> \param nexp_ppl ...
!> \param cexp_ppl ...
!> \param zetc ...
! **************************************************************************************************
   SUBROUTINE ppl_aux(auxint, mmax, t, rho, nexp_ppl, cexp_ppl, zetc)
      INTEGER, INTENT(IN)                                :: mmax
      REAL(KIND=dp), DIMENSION(0:mmax)                   :: auxint
      REAL(KIND=dp), INTENT(IN)                          :: t, rho
      INTEGER, INTENT(IN)                                :: nexp_ppl
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: cexp_ppl
      REAL(KIND=dp), INTENT(IN)                          :: zetc

      CHARACTER(len=*), PARAMETER :: routineN = 'ppl_aux', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, ke, kp, pmax
      REAL(KIND=dp)                                      :: a2, a3, a4, cc, f, q, q2, q4, q6, rho2, &
                                                            rho3, t2, t3
      REAL(KIND=dp), DIMENSION(0:6)                      :: polder
      REAL(KIND=dp), DIMENSION(0:mmax)                   :: expder

      q = rho+zetc
      polder = 0._dp
      pmax = 0
      IF (nexp_ppl > 0) THEN
         polder(0) = polder(0)+cexp_ppl(1)
         pmax = 0
      END IF
      IF (nexp_ppl > 1) THEN
         q2 = q*q
         a2 = 0.5_dp/q2*cexp_ppl(2)
         polder(0) = polder(0)+a2*(2._dp*rho*t+3._dp*q)
         polder(1) = polder(1)-a2*2._dp*rho
         pmax = 1
      END IF
      IF (nexp_ppl > 2) THEN
         q4 = q2*q2
         rho2 = rho*rho
         t2 = t*t
         a3 = 0.25_dp/q4*cexp_ppl(3)
         polder(0) = polder(0)+a3*(4._dp*rho2*t2+20._dp*rho*t*q+15._dp*q2)
         polder(1) = polder(1)-a3*(8._dp*rho2*t+20._dp*rho*q)
         polder(2) = polder(2)+a3*8._dp*rho2
         pmax = 2
      END IF
      IF (nexp_ppl > 3) THEN
         q6 = q4*q2
         rho3 = rho2*rho
         t3 = t2*t
         a4 = 0.125_dp/q6*cexp_ppl(4)
         polder(0) = polder(0)+a4*(8._dp*rho3*t3+84._dp*rho2*t2*q+210._dp*rho*t*q2+105._dp*q*q2)
         polder(1) = polder(1)-a4*(24._dp*rho3*t2+168._dp*rho2*t*q+210._dp*rho*q2)
         polder(2) = polder(2)+a4*(48._dp*rho3*t+168._dp*rho2*q)
         polder(3) = polder(3)-a4*48_dp*rho3
         pmax = 3
      END IF
      IF (nexp_ppl > 4) THEN
         CALL cp__b("aobasis/ai_overlap_ppl.F",227,"nexp_ppl > 4")
      END IF

      f = zetc/q
      cc = (pi/q)**1.5_dp*EXP(-t*f)

      IF (mmax >= 0) expder(0) = cc
      DO i = 1, mmax
         expder(i) = f*expder(i-1)
      END DO

      DO i = 0, mmax
         DO j = 0, MIN(i, pmax)
            kp = j
            ke = i-j
            auxint(i) = auxint(i)+expder(ke)*polder(kp)*choose(i, j)
         END DO
      END DO

   END SUBROUTINE ppl_aux
! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param k ...
!> \retval choose ...
! **************************************************************************************************
   FUNCTION choose(n, k)

      INTEGER, INTENT(IN)                                :: n, k
      REAL(KIND=dp)                                      :: choose

      IF (n >= k) THEN
         choose = REAL(NINT(fac(n)/(fac(k)*fac(n-k))), KIND=dp)
      ELSE
         choose = 0.0_dp
      ENDIF

   END FUNCTION choose
! **************************************************************************************************

END MODULE ai_overlap_ppl
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Calculation of the local pseudopotential contribution to the core Hamiltonian
!>         <a|V(local)|b> = <a|Sum e^a*rc**2|b>
!> \par History
!>      - core_ppnl refactored from qs_core_hamiltonian [Joost VandeVondele, 2008-11-01]
!>      - adapted for PPL [jhu, 2009-02-23]
!>      - OpenMP added [Iain Bethune, Fiona Reid, 2013-11-13]
!>      - Bug fix: correct orbital pointer range [07.2014,JGH]
!>      - k-point aware [07.2015,JGH]
! **************************************************************************************************
MODULE core_ppl

   USE ai_overlap_ppl,                  ONLY: ppl_integral
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE external_potential_types,        ONLY: get_potential,&
                                              gth_potential_type
   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type,&
                                              nl_set_sub_iterator,&
                                              nl_sub_iterate
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'core_ppl'

   PUBLIC :: build_core_ppl

CONTAINS

!==========================================================================================================

! **************************************************************************************************
!> \brief ...
!> \param matrix_h ...
!> \param matrix_p ...
!> \param force ...
!> \param virial ...
!> \param calculate_forces ...
!> \param use_virial ...
!> \param nder ...
!> \param qs_kind_set ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param sab_orb ...
!> \param sac_ppl ...
!> \param nimages ...
!> \param cell_to_index ...
! **************************************************************************************************
   SUBROUTINE build_core_ppl(matrix_h, matrix_p, force, virial, calculate_forces, use_virial, nder, &
                             qs_kind_set, atomic_kind_set, particle_set, sab_orb, sac_ppl, nimages, cell_to_index)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_h, matrix_p
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(virial_type), POINTER                         :: virial
      LOGICAL, INTENT(IN)                                :: calculate_forces
      LOGICAL                                            :: use_virial
      INTEGER                                            :: nder
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb, sac_ppl
      INTEGER, INTENT(IN)                                :: nimages
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_core_ppl', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nexp_max = 10

      INTEGER :: atom_a, atom_b, atom_c, handle, i, iatom, icol, ikind, img, inode, irow, iset, &
         jatom, jkind, jset, katom, kkind, ldai, ldsab, maxco, maxder, maxl, maxlgto, maxlppl, &
         maxnset, maxsgf, mepos, natom, ncoa, ncob, nexp_lpot, nexp_ppl, nkind, nseta, nsetb, &
         nthread, sgfa, sgfb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, DIMENSION(3)                              :: cellind
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, &
                                                            nct_lpot, npgfa, npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      INTEGER, DIMENSION(nexp_max)                       :: nct_ppl
      LOGICAL                                            :: dokp, found, lpotextended
      REAL(KIND=dp)                                      :: alpha, dab, dac, dbc, f0, ppl_radius
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: ppl_fwork, ppl_work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: hab, pab
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, rab, rac, rbc
      REAL(KIND=dp), DIMENSION(4, nexp_max)              :: cval_ppl
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha_lpot, cexp_ppl, set_radius_a, &
                                                            set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cval_lpot, h_block, p_block, rpgfa, &
                                                            rpgfb, sphi_a, sphi_b, zeta, zetb
      REAL(KIND=dp), DIMENSION(nexp_max)                 :: alpha_ppl
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: ap_iterator, nl_iterator

      IF (calculate_forces) THEN
         CALL timeset(routineN//"_forces", handle)
      ELSE
         CALL timeset(routineN, handle)
      ENDIF

      nkind = SIZE(atomic_kind_set)
      natom = SIZE(particle_set)

      dokp = (nimages > 1)

      ALLOCATE (atom_of_kind(natom))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)

      IF (calculate_forces) THEN
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, nimages
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
               CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                 alpha_scalar=-2.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      END IF

      maxder = ncoset(nder)

      CALL get_qs_kind_set(qs_kind_set, maxco=maxco, maxlgto=maxlgto, &
                           maxsgf=maxsgf, maxnset=maxnset, maxlppl=maxlppl)

      maxl = MAX(maxlgto, maxlppl)
      CALL init_orbital_pointers(2*maxl+2*nder+1)

      ldsab = MAX(maxco, ncoset(maxlppl), maxsgf, maxlppl)
      ldai = ncoset(maxl+nder+1)

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO

      nthread = 1
!$    nthread = omp_get_max_threads()

      CALL neighbor_list_iterator_create(nl_iterator, sab_orb, nthread=nthread)

!$OMP PARALLEL &
!$OMP DEFAULT (NONE) &
!$OMP SHARED  (nl_iterator, basis_set_list, calculate_forces, force, use_virial, &
!$OMP          matrix_h, matrix_p, virial, atomic_kind_set, qs_kind_set, particle_set, &
!$OMP          sab_orb, sac_ppl, nthread, ncoset, nkind, cell_to_index, &
!$OMP          atom_of_kind, ldsab,  maxnset, maxder, &
!$OMP          maxlgto, nder, maxco, dokp) &
!$OMP PRIVATE (ikind, jkind, inode, iatom, jatom, rab, basis_set_a, basis_set_b, atom_a, &
!$OMP          atom_b, first_sgfa, la_max, la_min, npgfa, nsgfa, sphi_a, &
!$OMP          zeta, first_sgfb, lb_max, lb_min, npgfb, nsetb, rpgfb, set_radius_b, sphi_b, &
!$OMP          zetb, dab, irow, icol, h_block, found, iset, ncoa, &
!$OMP          sgfa, jset, ncob, sgfb, nsgfb, p_block, work, pab, hab, kkind, nseta, &
!$OMP          gth_potential, alpha, cexp_ppl, lpotextended, &
!$OMP          ppl_radius, nexp_lpot, nexp_ppl, alpha_ppl, alpha_lpot, nct_ppl, &
!$OMP          nct_lpot, cval_ppl, cval_lpot, ap_iterator, rac, dac, rbc, dbc, &
!$OMP          set_radius_a,  rpgfa, force_a, force_b, ppl_fwork, mepos, &
!$OMP          f0, katom, ppl_work, atom_c, cellind, img, &
!$OMP          ldai)

      mepos = 0
!$    mepos = omp_get_thread_num()

      ! iterator for basis/potential list
      CALL neighbor_list_iterator_create(ap_iterator, sac_ppl, search=.TRUE.)

      ALLOCATE (hab(ldsab, ldsab, maxnset, maxnset), work(ldsab, ldsab*maxder))
      ldai = ncoset(2*maxlgto+2*nder)
      ALLOCATE (ppl_work(ldai, ldai, MAX(maxder, 2*maxlgto+2*nder+1)))
      IF (calculate_forces) THEN
         ALLOCATE (pab(maxco, maxco, maxnset, maxnset))
         ldai = ncoset(maxlgto)
         ALLOCATE (ppl_fwork(ldai, ldai, maxder))
      END IF

      DO WHILE (neighbor_list_iterate(nl_iterator, mepos=mepos) == 0)

         CALL get_iterator_info(nl_iterator, mepos=mepos, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab, cell=cellind)

         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE

         atom_a = atom_of_kind(iatom)
         atom_b = atom_of_kind(jatom)

         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         dab = SQRT(SUM(rab*rab))

         IF (dokp) THEN
            img = cell_to_index(cellind(1), cellind(2), cellind(3))
         ELSE
            img = 1
         END IF

         ! *** Use the symmetry of the first derivatives ***
         IF (iatom == jatom) THEN
            f0 = 1.0_dp
         ELSE
            f0 = 2.0_dp
         END IF

         ! *** Create matrix blocks for a new matrix block column ***
         IF (iatom <= jatom) THEN
            irow = iatom
            icol = jatom
         ELSE
            irow = jatom
            icol = iatom
         END IF
         NULLIFY (h_block)
         CALL cp_dbcsr_get_block_p(matrix_h(1, img)%matrix, irow, icol, h_block, found)
         IF(.NOT.(found))CALL cp__a("core_ppl.F",273)
         IF (calculate_forces) THEN
            NULLIFY (p_block)
            CALL cp_dbcsr_get_block_p(matrix_p(1, img)%matrix, irow, icol, p_block, found)
            IF (ASSOCIATED(p_block)) THEN
               DO iset = 1, nseta
                  ncoa = npgfa(iset)*ncoset(la_max(iset))
                  sgfa = first_sgfa(1, iset)
                  DO jset = 1, nsetb
                     ncob = npgfb(jset)*ncoset(lb_max(jset))
                     sgfb = first_sgfb(1, jset)

                     ! *** Decontract density matrix block ***
                     IF (iatom <= jatom) THEN
                        CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                                   1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                   p_block(sgfa, sgfb), SIZE(p_block, 1), &
                                   0.0_dp, work(1, 1), SIZE(work, 1))
                     ELSE
                        CALL dgemm("N", "T", ncoa, nsgfb(jset), nsgfa(iset), &
                                   1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                   p_block(sgfb, sgfa), SIZE(p_block, 1), &
                                   0.0_dp, work(1, 1), SIZE(work, 1))
                     END IF

                     CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                                1.0_dp, work(1, 1), SIZE(work, 1), &
                                sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                                0.0_dp, pab(1, 1, iset, jset), SIZE(pab, 1))
                  END DO
               END DO
            ENDIF
         END IF

         hab = 0._dp

         ! loop over all kinds for pseudopotential atoms
         DO kkind = 1, nkind

            CALL get_qs_kind(qs_kind_set(kkind), gth_potential=gth_potential)
            IF (.NOT. ASSOCIATED(gth_potential)) CYCLE
            CALL get_potential(potential=gth_potential, &
                               alpha_ppl=alpha, cexp_ppl=cexp_ppl, &
                               lpot_present=lpotextended, ppl_radius=ppl_radius)
            nexp_ppl = 1
            alpha_ppl(1) = alpha
            nct_ppl(1) = SIZE(cexp_ppl)
            cval_ppl(1:nct_ppl(1), 1) = cexp_ppl(1:nct_ppl(1))
            IF (lpotextended) THEN
               CALL get_potential(potential=gth_potential, &
                                  nexp_lpot=nexp_lpot, alpha_lpot=alpha_lpot, nct_lpot=nct_lpot, cval_lpot=cval_lpot)
               IF(.NOT.(nexp_lpot < nexp_max))CALL cp__a("core_ppl.F",324)
               nexp_ppl = nexp_lpot+1
               alpha_ppl(2:nexp_lpot+1) = alpha_lpot(1:nexp_lpot)
               nct_ppl(2:nexp_lpot+1) = nct_lpot(1:nexp_lpot)
               DO i = 1, nexp_lpot
                  cval_ppl(1:nct_lpot(i), i+1) = cval_lpot(1:nct_lpot(i), i)
               END DO
            ENDIF

            CALL nl_set_sub_iterator(ap_iterator, ikind, kkind, iatom)

            DO WHILE (nl_sub_iterate(ap_iterator) == 0)

               CALL get_iterator_info(ap_iterator, jatom=katom, r=rac)

               dac = SQRT(SUM(rac*rac))
               rbc(:) = rac(:)-rab(:)
               dbc = SQRT(SUM(rbc*rbc))
               IF ((MAXVAL(set_radius_a(:))+ppl_radius < dac) .OR. &
                   (MAXVAL(set_radius_b(:))+ppl_radius < dbc)) THEN
                  CYCLE
               END IF

               DO iset = 1, nseta
                  IF (set_radius_a(iset)+ppl_radius < dac) CYCLE
                  ncoa = npgfa(iset)*ncoset(la_max(iset))
                  sgfa = first_sgfa(1, iset)
                  DO jset = 1, nsetb
                     IF (set_radius_b(jset)+ppl_radius < dbc) CYCLE
                     ncob = npgfb(jset)*ncoset(lb_max(jset))
                     sgfb = first_sgfb(1, jset)
                     IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE
                     ! *** Calculate the GTH pseudo potential forces ***
                     IF (calculate_forces) THEN

                        CALL ppl_integral( &
                           la_max(iset), la_min(iset), npgfa(iset), &
                           rpgfa(:, iset), zeta(:, iset), &
                           lb_max(jset), lb_min(jset), npgfb(jset), &
                           rpgfb(:, jset), zetb(:, jset), &
                           nexp_ppl, alpha_ppl, nct_ppl, cval_ppl, ppl_radius, &
                           rab, dab, rac, dac, rbc, dbc, &
                           hab(:, :, iset, jset), ppl_work, pab(:, :, iset, jset), &
                           force_a, force_b, ppl_fwork)
                        ! *** The derivatives w.r.t. atomic center c are    ***
                        ! *** calculated using the translational invariance ***
                        ! *** of the first derivatives                      ***
                        atom_c = atom_of_kind(katom)

!$OMP CRITICAL(force_critical)
                        force(ikind)%gth_ppl(1, atom_a) = force(ikind)%gth_ppl(1, atom_a)+f0*force_a(1)
                        force(ikind)%gth_ppl(2, atom_a) = force(ikind)%gth_ppl(2, atom_a)+f0*force_a(2)
                        force(ikind)%gth_ppl(3, atom_a) = force(ikind)%gth_ppl(3, atom_a)+f0*force_a(3)
                        force(kkind)%gth_ppl(1, atom_c) = force(kkind)%gth_ppl(1, atom_c)-f0*force_a(1)
                        force(kkind)%gth_ppl(2, atom_c) = force(kkind)%gth_ppl(2, atom_c)-f0*force_a(2)
                        force(kkind)%gth_ppl(3, atom_c) = force(kkind)%gth_ppl(3, atom_c)-f0*force_a(3)

                        force(jkind)%gth_ppl(1, atom_b) = force(jkind)%gth_ppl(1, atom_b)+f0*force_b(1)
                        force(jkind)%gth_ppl(2, atom_b) = force(jkind)%gth_ppl(2, atom_b)+f0*force_b(2)
                        force(jkind)%gth_ppl(3, atom_b) = force(jkind)%gth_ppl(3, atom_b)+f0*force_b(3)
                        force(kkind)%gth_ppl(1, atom_c) = force(kkind)%gth_ppl(1, atom_c)-f0*force_b(1)
                        force(kkind)%gth_ppl(2, atom_c) = force(kkind)%gth_ppl(2, atom_c)-f0*force_b(2)
                        force(kkind)%gth_ppl(3, atom_c) = force(kkind)%gth_ppl(3, atom_c)-f0*force_b(3)

                        IF (use_virial) THEN
                           CALL virial_pair_force(virial%pv_virial, f0, force_a, rac)
                           CALL virial_pair_force(virial%pv_virial, f0, force_b, rbc)
                        END IF
!$OMP END CRITICAL(force_critical)

                     ELSE
                        CALL ppl_integral( &
                           la_max(iset), la_min(iset), npgfa(iset), &
                           rpgfa(:, iset), zeta(:, iset), &
                           lb_max(jset), lb_min(jset), npgfb(jset), &
                           rpgfb(:, jset), zetb(:, jset), &
                           nexp_ppl, alpha_ppl, nct_ppl, cval_ppl, ppl_radius, &
                           rab, dab, rac, dac, rbc, dbc, hab(:, :, iset, jset), ppl_work)
                     END IF
                  END DO
               END DO
            END DO
         END DO

         ! *** Contract PPL integrals
         DO iset = 1, nseta
            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)
            DO jset = 1, nsetb
               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                          1.0_dp, hab(1, 1, iset, jset), SIZE(hab, 1), &
                          sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                          0.0_dp, work(1, 1), SIZE(work, 1))

!$OMP CRITICAL(h_block_critical)
               IF (iatom <= jatom) THEN
                  CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             work(1, 1), SIZE(work, 1), &
                             1.0_dp, h_block(sgfa, sgfb), SIZE(h_block, 1))
               ELSE
                  CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                             1.0_dp, work(1, 1), SIZE(work, 1), &
                             sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             1.0_dp, h_block(sgfb, sgfa), SIZE(h_block, 1))
               END IF
!$OMP END CRITICAL(h_block_critical)

            END DO
         END DO
      END DO

      CALL neighbor_list_iterator_release(ap_iterator)

      DEALLOCATE (hab, work, ppl_work)

      IF (calculate_forces) THEN
         DEALLOCATE (pab, ppl_fwork)
      END IF

!$OMP END PARALLEL

      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (atom_of_kind, basis_set_list)

      IF (calculate_forces) THEN
         ! *** If LSD, then recover alpha density and beta density     ***
         ! *** from the total density (1) and the spin density (2)     ***
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, nimages
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=0.5_dp, beta_scalar=0.5_dp)
               CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                 alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_core_ppl

!==========================================================================================================

END MODULE core_ppl
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Calculation of the non-local pseudopotential contribution to the core Hamiltonian
!>         <a|V(non-local)|b> = <a|p(l,i)>*h(i,j)*<p(l,j)|b>
!> \par History
!>      - refactered from qs_core_hamiltian [Joost VandeVondele, 2008-11-01]
!>      - full rewrite [jhu, 2009-01-23]
! **************************************************************************************************
MODULE core_ppnl
   USE ai_overlap,                      ONLY: overlap
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE external_potential_types,        ONLY: gth_potential_p_type,&
                                              gth_potential_type
   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              nco,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE sap_kind_types,                  ONLY: alist_type,&
                                              clist_type,&
                                              get_alist,&
                                              release_sap_int,&
                                              sap_int_type,&
                                              sap_sort
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'core_ppnl'

   PUBLIC :: build_core_ppnl

CONTAINS

!==========================================================================================================

! **************************************************************************************************
!> \brief ...
!> \param matrix_h ...
!> \param matrix_p ...
!> \param force ...
!> \param virial ...
!> \param calculate_forces ...
!> \param use_virial ...
!> \param nder ...
!> \param qs_kind_set ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param sab_orb ...
!> \param sap_ppnl ...
!> \param eps_ppnl ...
!> \param nimages ...
!> \param cell_to_index ...
! **************************************************************************************************
   SUBROUTINE build_core_ppnl(matrix_h, matrix_p, force, virial, calculate_forces, use_virial, nder, &
                              qs_kind_set, atomic_kind_set, particle_set, sab_orb, sap_ppnl, eps_ppnl, &
                              nimages, cell_to_index)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_h, matrix_p
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(virial_type), POINTER                         :: virial
      LOGICAL, INTENT(IN)                                :: calculate_forces
      LOGICAL                                            :: use_virial
      INTEGER                                            :: nder
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb, sap_ppnl
      REAL(KIND=dp), INTENT(IN)                          :: eps_ppnl
      INTEGER, INTENT(IN)                                :: nimages
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_core_ppnl', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, atom_c, first_col, handle, i, iab, iac, iatom, ibc, icol, ikind, &
         ilist, img, inode, irow, iset, j, jatom, jkind, jneighbor, kac, katom, kbc, kkind, l, &
         lc_max, lc_min, ldai, ldsab, lppnl, maxco, maxder, maxl, maxlgto, maxlppnl, maxppnl, &
         maxsgf, mepos, na, natom, nb, ncoa, ncoc, nkind, nlist, nneighbor, nnode, np, nppnl, &
         nprjc, nseta, nsgfa, nthread, prjc, sgfa
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, DIMENSION(3)                              :: cell_b, cell_c
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, npgfa, nprj_ppnl, &
                                                            nsgf_seta
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa
      LOGICAL                                            :: dokp, found, ppnl_present
      REAL(KIND=dp)                                      :: dac, f0, ppnl_radius
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: sab, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: ai_work
      REAL(KIND=dp), DIMENSION(1)                        :: rprjc, zetc
      REAL(KIND=dp), DIMENSION(3)                        :: fa, fb, rab, rac, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha_ppnl, set_radius_a
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cprj, h_block, p_block, rpgfa, sphi_a, &
                                                            vprj_ppnl, zeta
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: achint, acint, bchint, bcint
      TYPE(alist_type), POINTER                          :: alist_ac, alist_bc
      TYPE(clist_type), POINTER                          :: clist
      TYPE(gth_potential_p_type), DIMENSION(:), POINTER  :: potential
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(sap_int_type), DIMENSION(:), POINTER          :: sap_int

      IF (calculate_forces) THEN
         CALL timeset(routineN//"_forces", handle)
      ELSE
         CALL timeset(routineN, handle)
      ENDIF

      ppnl_present = ASSOCIATED(sap_ppnl)

      IF (ppnl_present) THEN

         nkind = SIZE(atomic_kind_set)
         natom = SIZE(particle_set)

         dokp = (nimages > 1)

         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind)

         IF (calculate_forces) THEN
            IF (SIZE(matrix_p, 1) == 2) THEN
               DO img = 1, nimages
                  CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                    alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
                  CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                    alpha_scalar=-2.0_dp, beta_scalar=1.0_dp)
               END DO
            END IF
         END IF

         maxder = ncoset(nder)

         CALL get_qs_kind_set(qs_kind_set, &
                              maxco=maxco, &
                              maxlgto=maxlgto, &
                              maxsgf=maxsgf, &
                              maxlppnl=maxlppnl, &
                              maxppnl=maxppnl)

         maxl = MAX(maxlgto, maxlppnl)
         CALL init_orbital_pointers(maxl+nder+1)

         ldsab = MAX(maxco, ncoset(maxlppnl), maxsgf, maxppnl)
         ldai = ncoset(maxl+nder+1)

         !sap_int needs to be shared as multiple threads need to access this
         ALLOCATE (sap_int(nkind*nkind))
         DO i = 1, nkind*nkind
            NULLIFY (sap_int(i)%alist, sap_int(i)%asort, sap_int(i)%aindex)
            sap_int(i)%nalist = 0
         END DO

         !set up direct access to basis and potential
         ALLOCATE (basis_set(nkind), potential(nkind))
         DO ikind = 1, nkind
            CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
            IF (ASSOCIATED(orb_basis_set)) THEN
               basis_set(ikind)%gto_basis_set => orb_basis_set
            ELSE
               NULLIFY (basis_set(ikind)%gto_basis_set)
            END IF
            CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)
            IF (ASSOCIATED(gth_potential)) THEN
               potential(ikind)%gth_potential => gth_potential
            ELSE
               NULLIFY (potential(ikind)%gth_potential)
            END IF
         END DO

         nthread = 1
!$       nthread = omp_get_max_threads()

         !calculate the overlap integrals <a|p>
         CALL neighbor_list_iterator_create(nl_iterator, sap_ppnl, nthread=nthread)
!$OMP PARALLEL &
!$OMP DEFAULT (NONE) &
!$OMP SHARED  (nl_iterator, basis_set, potential, maxder, ncoset, &
!$OMP          sap_int, nkind, ldsab,  ldai, nder, nco ) &
!$OMP PRIVATE (mepos, ikind, kkind, iatom, katom, nlist, ilist, nneighbor, jneighbor, &
!$OMP          cell_c, rac, iac, first_sgfa, la_max, la_min, npgfa, nseta, nsgfa, nsgf_seta, &
!$OMP          sphi_a, zeta, cprj, lppnl, nppnl, nprj_ppnl, &
!$OMP          clist, iset, ncoa, sgfa, prjc, work, sab, ai_work, nprjc,  ppnl_radius, &
!$OMP          ncoc, rpgfa, first_col, vprj_ppnl, i, l, &
!$OMP          set_radius_a, rprjc, dac, lc_max, lc_min, zetc, alpha_ppnl)
         mepos = 0
!$       mepos = omp_get_thread_num()

         ALLOCATE (sab(ldsab, ldsab*maxder), work(ldsab, ldsab*maxder))
         sab = 0.0_dp
         ALLOCATE (ai_work(ldai, ldai, ncoset(nder+1)))
         ai_work = 0.0_dp

         DO WHILE (neighbor_list_iterate(nl_iterator, mepos=mepos) == 0)
            CALL get_iterator_info(nl_iterator, mepos=mepos, ikind=ikind, jkind=kkind, iatom=iatom, &
                                   jatom=katom, nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor, cell=cell_c, r=rac)
            iac = ikind+nkind*(kkind-1)
            IF (.NOT. ASSOCIATED(basis_set(ikind)%gto_basis_set)) CYCLE
            IF (.NOT. ASSOCIATED(potential(kkind)%gth_potential)) CYCLE
            ! get definition of basis set
            first_sgfa => basis_set(ikind)%gto_basis_set%first_sgf
            la_max => basis_set(ikind)%gto_basis_set%lmax
            la_min => basis_set(ikind)%gto_basis_set%lmin
            npgfa => basis_set(ikind)%gto_basis_set%npgf
            nseta = basis_set(ikind)%gto_basis_set%nset
            nsgfa = basis_set(ikind)%gto_basis_set%nsgf
            nsgf_seta => basis_set(ikind)%gto_basis_set%nsgf_set
            rpgfa => basis_set(ikind)%gto_basis_set%pgf_radius
            set_radius_a => basis_set(ikind)%gto_basis_set%set_radius
            sphi_a => basis_set(ikind)%gto_basis_set%sphi
            zeta => basis_set(ikind)%gto_basis_set%zet
            ! get definition of PP projectors
            alpha_ppnl => potential(kkind)%gth_potential%alpha_ppnl
            cprj => potential(kkind)%gth_potential%cprj
            lppnl = potential(kkind)%gth_potential%lppnl
            nppnl = potential(kkind)%gth_potential%nppnl
            nprj_ppnl => potential(kkind)%gth_potential%nprj_ppnl
            ppnl_radius = potential(kkind)%gth_potential%ppnl_radius
            vprj_ppnl => potential(kkind)%gth_potential%vprj_ppnl
!$OMP CRITICAL(sap_int_critical)
            IF (.NOT. ASSOCIATED(sap_int(iac)%alist)) THEN
               sap_int(iac)%a_kind = ikind
               sap_int(iac)%p_kind = kkind
               sap_int(iac)%nalist = nlist
               ALLOCATE (sap_int(iac)%alist(nlist))
               DO i = 1, nlist
                  NULLIFY (sap_int(iac)%alist(i)%clist)
                  sap_int(iac)%alist(i)%aatom = 0
                  sap_int(iac)%alist(i)%nclist = 0
               END DO
            END IF
            IF (.NOT. ASSOCIATED(sap_int(iac)%alist(ilist)%clist)) THEN
               sap_int(iac)%alist(ilist)%aatom = iatom
               sap_int(iac)%alist(ilist)%nclist = nneighbor
               ALLOCATE (sap_int(iac)%alist(ilist)%clist(nneighbor))
               DO i = 1, nneighbor
                  sap_int(iac)%alist(ilist)%clist(i)%catom = 0
               END DO
            END IF
!$OMP END CRITICAL(sap_int_critical)
            dac = SQRT(SUM(rac*rac))
            clist => sap_int(iac)%alist(ilist)%clist(jneighbor)
            clist%catom = katom
            clist%cell = cell_c
            clist%rac = rac
            ALLOCATE (clist%acint(nsgfa, nppnl, maxder), &
                      clist%achint(nsgfa, nppnl, maxder))
            clist%acint = 0._dp
            clist%achint = 0._dp
            clist%nsgf_cnt = 0
            NULLIFY (clist%sgf_list)
            DO iset = 1, nseta
               ncoa = npgfa(iset)*ncoset(la_max(iset))
               sgfa = first_sgfa(1, iset)
               prjc = 1
               ! XXX fix, use correct bounds
               work = 0._dp
               DO l = 0, lppnl
                  nprjc = nprj_ppnl(l)*nco(l)
                  IF (nprjc == 0) CYCLE
                  rprjc(1) = ppnl_radius
                  IF (set_radius_a(iset)+rprjc(1) < dac) CYCLE
                  lc_max = l+2*(nprj_ppnl(l)-1)
                  lc_min = l
                  zetc(1) = alpha_ppnl(l)
                  ncoc = ncoset(lc_max)

                  ! *** Calculate the primitive overlap integrals ***
                  CALL overlap(la_max(iset), la_min(iset), npgfa(iset), rpgfa(:, iset), zeta(:, iset), &
                               lc_max, lc_min, 1, rprjc, zetc, rac, dac, sab, nder, .TRUE., ai_work, ldai)
                  ! *** Transformation step projector functions (cartesian->spherical) ***
                  DO i = 1, maxder
                     first_col = (i-1)*SIZE(work, 1)
                     CALL dgemm("N", "N", ncoa, nprjc, ncoc, 1.0_dp, sab(1, first_col+1), SIZE(sab, 1), &
                                cprj(1, prjc), SIZE(cprj, 1), 0.0_dp, work(1, first_col+prjc), SIZE(work, 1))
                  END DO
                  prjc = prjc+nprjc
               END DO
               DO i = 1, maxder
                  first_col = (i-1)*SIZE(work, 1)+1
                  ! *** Contraction step (basis functions) ***
                  CALL dgemm("T", "N", nsgf_seta(iset), nppnl, ncoa, 1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             work(1, first_col), SIZE(work, 1), 0.0_dp, clist%acint(sgfa, 1, i), nsgfa)
                  ! *** Multiply with interaction matrix(h) ***
                  CALL dgemm("N", "N", nsgf_seta(iset), nppnl, nppnl, 1.0_dp, clist%acint(sgfa, 1, i), nsgfa, &
                             vprj_ppnl(1, 1), SIZE(vprj_ppnl, 1), 0.0_dp, clist%achint(sgfa, 1, i), nsgfa)
               END DO
            END DO
            clist%maxac = MAXVAL(ABS(clist%acint(:, :, 1)))
            clist%maxach = MAXVAL(ABS(clist%achint(:, :, 1)))
         END DO

         DEALLOCATE (sab, ai_work, work)
!$OMP END PARALLEL
         CALL neighbor_list_iterator_release(nl_iterator)

         ! *** Set up a sorting index
         CALL sap_sort(sap_int)
         ! *** All integrals needed have been calculated and stored in sap_int
         ! *** We now calculate the Hamiltonian matrix elements
         CALL neighbor_list_iterator_create(nl_iterator, sab_orb, nthread=nthread)

!$OMP PARALLEL &
!$OMP DEFAULT (NONE) &
!$OMP SHARED  (nl_iterator, dokp, basis_set, atom_of_kind, matrix_h, cell_to_index,&
!$OMP          matrix_p, sap_int, nkind, eps_ppnl, force, virial, use_virial, calculate_forces) &
!$OMP PRIVATE (mepos, ikind, jkind, iatom, jatom, nlist, ilist, nnode, inode, cell_b, rab, &
!$OMP          iab, atom_a, atom_b, f0, irow, icol, h_block, &
!$OMP          found,p_block, iac, ibc, alist_ac, alist_bc, acint, bcint, &
!$OMP          achint, bchint, na, np, nb, katom, atom_c, j, fa, fb, rbc, rac, &
!$OMP          kkind, kac, kbc, i, img)

         mepos = 0
!$       mepos = omp_get_thread_num()

         DO WHILE (neighbor_list_iterate(nl_iterator, mepos=mepos) == 0)
            CALL get_iterator_info(nl_iterator, mepos=mepos, ikind=ikind, jkind=jkind, iatom=iatom, &
                                   jatom=jatom, nlist=nlist, ilist=ilist, nnode=nnode, inode=inode, cell=cell_b, r=rab)
            IF (.NOT. ASSOCIATED(basis_set(ikind)%gto_basis_set)) CYCLE
            IF (.NOT. ASSOCIATED(basis_set(jkind)%gto_basis_set)) CYCLE
            iab = ikind+nkind*(jkind-1)
            atom_a = atom_of_kind(iatom)
            atom_b = atom_of_kind(jatom)

            ! *** Use the symmetry of the first derivatives ***
            IF (iatom == jatom) THEN
               f0 = 1.0_dp
            ELSE
               f0 = 2.0_dp
            END IF

            IF (dokp) THEN
               img = cell_to_index(cell_b(1), cell_b(2), cell_b(3))
            ELSE
               img = 1
            END IF

            ! *** Create matrix blocks for a new matrix block column ***
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF
            NULLIFY (h_block)
            CALL cp_dbcsr_get_block_p(matrix_h(1, img)%matrix, irow, icol, h_block, found)
            IF (calculate_forces) THEN
               NULLIFY (p_block)
               CALL cp_dbcsr_get_block_p(matrix_p(1, img)%matrix, irow, icol, p_block, found)
            END IF

            ! loop over all kinds for projector atom
            IF (ASSOCIATED(h_block)) THEN
               DO kkind = 1, nkind
                  iac = ikind+nkind*(kkind-1)
                  ibc = jkind+nkind*(kkind-1)
                  IF (.NOT. ASSOCIATED(sap_int(iac)%alist)) CYCLE
                  IF (.NOT. ASSOCIATED(sap_int(ibc)%alist)) CYCLE
                  CALL get_alist(sap_int(iac), alist_ac, iatom)
                  CALL get_alist(sap_int(ibc), alist_bc, jatom)
                  IF (.NOT. ASSOCIATED(alist_ac)) CYCLE
                  IF (.NOT. ASSOCIATED(alist_bc)) CYCLE
                  DO kac = 1, alist_ac%nclist
                     DO kbc = 1, alist_bc%nclist
                        IF (alist_ac%clist(kac)%catom /= alist_bc%clist(kbc)%catom) CYCLE
                        IF (ALL(cell_b+alist_bc%clist(kbc)%cell-alist_ac%clist(kac)%cell == 0)) THEN
                           IF (alist_ac%clist(kac)%maxac*alist_bc%clist(kbc)%maxach < eps_ppnl) CYCLE
                           acint => alist_ac%clist(kac)%acint
                           bcint => alist_bc%clist(kbc)%acint
                           achint => alist_ac%clist(kac)%achint
                           bchint => alist_bc%clist(kbc)%achint
                           na = SIZE(acint, 1)
                           np = SIZE(acint, 2)
                           nb = SIZE(bcint, 1)
!$OMP CRITICAL(h_block_critical)
                           IF (iatom <= jatom) THEN
                              CALL dgemm("N", "T", na, nb, np, 1._dp, achint(1, 1, 1), na, &
                                         bcint(1, 1, 1), nb, 1.0_dp, h_block, SIZE(h_block, 1))
                           ELSE
                              CALL dgemm("N", "T", nb, na, np, 1.0_dp, bchint(1, 1, 1), nb, &
                                         acint(1, 1, 1), na, 1.0_dp, h_block, SIZE(h_block, 1))
                           END IF
!$OMP END CRITICAL(h_block_critical)
                           IF (calculate_forces) THEN
                              IF (ASSOCIATED(p_block)) THEN
                                 katom = alist_ac%clist(kac)%catom
                                 atom_c = atom_of_kind(katom)
                                 DO i = 1, 3
                                    j = i+1
                                    IF (iatom <= jatom) THEN
                                       fa(i) = SUM(p_block(1:na, 1:nb)* &
                                                   MATMUL(acint(1:na, 1:np, j), TRANSPOSE(bchint(1:nb, 1:np, 1))))
                                       fb(i) = SUM(p_block(1:na, 1:nb)* &
                                                   MATMUL(achint(1:na, 1:np, 1), TRANSPOSE(bcint(1:nb, 1:np, j))))
                                    ELSE
                                       fa(i) = SUM(p_block(1:nb, 1:na)* &
                                                   MATMUL(bchint(1:nb, 1:np, 1), TRANSPOSE(acint(1:na, 1:np, j))))
                                       fb(i) = SUM(p_block(1:nb, 1:na)* &
                                                   MATMUL(bcint(1:nb, 1:np, j), TRANSPOSE(achint(1:na, 1:np, 1))))
                                    END IF
!$OMP CRITICAL(force_critical)
                                    force(ikind)%gth_ppnl(i, atom_a) = force(ikind)%gth_ppnl(i, atom_a)+f0*fa(i)
                                    force(kkind)%gth_ppnl(i, atom_c) = force(kkind)%gth_ppnl(i, atom_c)-f0*fa(i)
                                    force(jkind)%gth_ppnl(i, atom_b) = force(jkind)%gth_ppnl(i, atom_b)+f0*fb(i)
                                    force(kkind)%gth_ppnl(i, atom_c) = force(kkind)%gth_ppnl(i, atom_c)-f0*fb(i)
!$OMP END CRITICAL(force_critical)
                                 END DO

                                 IF (use_virial) THEN
                                    rac = alist_ac%clist(kac)%rac
                                    rbc = alist_bc%clist(kbc)%rac
!$OMP CRITICAL(virial_critical)
                                    CALL virial_pair_force(virial%pv_virial, f0, fa, rac)
                                    CALL virial_pair_force(virial%pv_virial, f0, fb, rbc)
!$OMP END CRITICAL(virial_critical)
                                 END IF
                              ENDIF
                           END IF
                           EXIT ! We have found a match and there can be only one single match
                        END IF
                     END DO
                  END DO
               END DO
            ENDIF
         END DO
!$OMP END PARALLEL
         CALL neighbor_list_iterator_release(nl_iterator)

         CALL release_sap_int(sap_int)

         DEALLOCATE (atom_of_kind)
         DEALLOCATE (basis_set, potential)

         IF (calculate_forces) THEN
            ! *** If LSD, then recover alpha density and beta density     ***
            ! *** from the total density (1) and the spin density (2)     ***
            IF (SIZE(matrix_p, 1) == 2) THEN
               DO img = 1, nimages
                  CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                    alpha_scalar=0.5_dp, beta_scalar=0.5_dp)
                  CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                    alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
               END DO
            END IF
         END IF

      END IF !ppnl_present

      CALL timestop(handle)

   END SUBROUTINE build_core_ppnl

!==========================================================================================================

END MODULE core_ppnl
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Build up the nuclear potential part of the core Hamiltonian matrix
!>      in the case of an allelectron calculation by computing the nuclear
!>      attraction integral matrix <a|-Z/r|b> and the integral matrix
!>      <a|Z*erf(r)/r|b>.  The integrals <a|Z*erf(r)/r|b> can be rewritten
!>      as the three-center Coulomb integrals <ab||c> with a  primitive
!>      s-type Gaussian function c multiplied by a factor N.
!>
!>                       erfc(r)
!>      <a|V|b> = -Z*<a|---------|b>
!>                          r
!>
!>                       1 - erf(r)
!>              = -Z*<a|------------|b>
!>                           r
!>
!>                        1           erf(r)
!>              = -Z*(<a|---|b> - <a|--------|b>)
!>                        r             r
!>
!>                        1
!>              = -Z*(<a|---|b> - N*<ab||c>)
!>                        r
!>
!>                       1
!>              = -Z*<a|---|b> + Z*N*<ab||c>
!>                       r
!>                   \_______/       \_____/
!>                       |              |
!>                    nuclear        coulomb
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par Parameters
!>      -  ax,ay,az  : Angular momentum index numbers of orbital a.
!>      -  bx,by,bz  : Angular momentum index numbers of orbital b.
!>      -  coset     : Cartesian orbital set pointer.
!>      -  dab       : Distance between the atomic centers a and b.
!>      -  dac       : Distance between the atomic centers a and c.
!>      -  dbc       : Distance between the atomic centers b and c.
!>      -  l{a,b}    : Angular momentum quantum number of shell a or b.
!>      -  l{a,b}_max: Maximum angular momentum quantum number of shell a or b.
!>      -  l{a,b}_min: Minimum angular momentum quantum number of shell a or b.
!>      -  ncoset    : Number of orbitals in a Cartesian orbital set.
!>      -  npgf{a,b} : Degree of contraction of shell a or b.
!>      -  rab       : Distance vector between the atomic centers a and b.
!>      -  rab2      : Square of the distance between the atomic centers a and b.
!>      -  rac       : Distance vector between the atomic centers a and c.
!>      -  rac2      : Square of the distance between the atomic centers a and c.
!>      -  rbc2      : Square of the distance between the atomic centers b and c.
!>      -  rpgf{a,b} : Radius of the primitive Gaussian-type function a or b.
!>      -  zet{a,b}  : Exponents of the Gaussian-type functions a or b.
!>      -  zetp      : Reciprocal of the sum of the exponents of orbital a and b.
!>      -  zetw        : Reciprocal of the sum of the exponents of orbital a, b and c.
!> Indices for pVp matrices
!>      -  ax,ay,az                : Angular momentum index numbers of orbital a.
!>      -  n{x1,y1,z1}{x2,y2,z2}   : indices for the pVp matrix elements
!>                                   <p_{x1,y1,z1}a|V|p_{x2,y2,z2}>
!>      -  co{a,b}{m,p}{x,y,z}     : coefficients of the pVp matrix elements: co == coset(lx,ly,lz)
!>                                   m == li - 1; p == li +1; li= x,y,z (which l quantum number is modified); l = a,b
!>                                   coset ( ... -1 ...) = 1  , should be zero, lmi = 1.0 for li >=1 and lmi = 0.0 for li = 0
!>                                  guarantees this as a prefactor
!>      -  f{a,b}{x,y,z},ft{a,b},z : f(t)li: prefactors as a result of derivations of l with respect to i
!>      -  pVpout                  : pVp matrix elements
!>      -  pVp                     : local pVp matrix elements
!>      -  lamin,lbmin             : minimal angular quantum number used in pVp matrices
!>      _  vnucp                   : potential of the nuclei
!>      -  vnuc                    : potential of the electrons
!>      _  pVpa, pVpb              : pVpl=coset(lx,ly,lz)
!>      -  na_pgf,nb_pgf           : indice for primitive gaussian functions
!> \par History
!>      10.2008 added pVp matrix elements (jens)
!> \author Matthias Krack (04.10.2000)
! **************************************************************************************************

MODULE ai_verfc

   USE gamma,                           ONLY: fgamma=>fgamma_0
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: coset,&
                                              ncoset
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_verfc'
   PRIVATE

! *** Public subroutines ***

   PUBLIC :: verfc

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the primitive three-center nuclear potential
!>          integrals <a|Z*erfc(r)/r|b> over Cartesian Gaussian-type
!>          functions.
!> \param la_max1 ...
!> \param npgfa ...
!> \param zeta ...
!> \param rpgfa ...
!> \param la_min1 ...
!> \param lb_max1 ...
!> \param npgfb ...
!> \param zetb ...
!> \param rpgfb ...
!> \param lb_min1 ...
!> \param zetc ...
!> \param rpgfc ...
!> \param zc ...
!> \param cerf ...
!> \param rab ...
!> \param rab2 ...
!> \param rac ...
!> \param rac2 ...
!> \param rbc2 ...
!> \param vabc ...
!> \param verf ...
!> \param vnuc ...
!> \param f ...
!> \param maxder ...
!> \param vabc_plus ...
!> \param vnabc ...
!> \param pVp_sum ...
!> \param pVp ...
!> \param dkh_erfc ...
!> \date    04.10.2000
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE verfc(la_max1, npgfa, zeta, rpgfa, la_min1, lb_max1, npgfb, zetb, rpgfb, lb_min1, &
                    zetc, rpgfc, zc, cerf, rab, rab2, rac, rac2, rbc2, vabc, verf, vnuc, f, &
                    maxder, vabc_plus, vnabc, pVp_sum, pVp, dkh_erfc) !JT
      INTEGER, INTENT(IN)                                :: la_max1, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta, rpgfa
      INTEGER, INTENT(IN)                                :: la_min1, lb_max1, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb, rpgfb
      INTEGER, INTENT(IN)                                :: lb_min1
      REAL(KIND=dp), INTENT(IN)                          :: zetc, rpgfc, zc, cerf
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: rab2
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rac
      REAL(KIND=dp), INTENT(IN)                          :: rac2, rbc2
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: vabc
      REAL(KIND=dp), DIMENSION(:, :, :)                  :: verf, vnuc
      REAL(KIND=dp), DIMENSION(0:)                       :: f
      INTEGER, INTENT(IN), OPTIONAL                      :: maxder
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL           :: vabc_plus, vnabc, pVp_sum, pVp
      LOGICAL, OPTIONAL                                  :: dkh_erfc

      CHARACTER(len=*), PARAMETER :: routineN = 'verfc', routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, coamx, coamy, coamz, coapx, coapy, coapz, cobmx, cobmy, &
         cobmz, cobpx, cobpy, cobpz, i, ipgf, j, jpgf, la, la_max, la_min, la_start, lb, lb_max, &
         lb_min, maxder_local, n, na, nap, nb, nmax, nxx, nxy, nxz, nyx, nyy, nyz, nzx, nzy, nzz, &
         pVpa, pVpb
      LOGICAL                                            :: do_dkh
      REAL(KIND=dp)                                      :: dab, dac, dbc, f0, f1, f2, f3, f4, fax, &
                                                            fay, faz, fbx, fby, fbz, ferf, fnuc, &
                                                            ftaz, ftbz, fx, fy, fz, rcp2, t, zetp, &
                                                            zetq, zetw
      REAL(KIND=dp), DIMENSION(3)                        :: rap, rbp, rcp, rpw

!   ---------------------------------------------------------------------------

      IF (PRESENT(maxder)) THEN
         verf(1:ncoset(la_max1), 1:ncoset(lb_max1), 1) = 0._dp
         vnuc(1:ncoset(la_max1), 1:ncoset(lb_max1), 1) = 0._dp
      END IF

      do_dkh = .FALSE.

      IF (PRESENT(pVp)) THEN
         do_dkh = .TRUE.
         pVp = 0.0_dp
         ! pVp matrices requires angular momentum quantum numbers l+1 and l-1

         la_max = la_max1+1
         IF (PRESENT(maxder)) THEN
            IF (maxder > 0) THEN
               la_max = la_max1
            END IF
         END IF
         lb_max = lb_max1+1

         la_min = MAX(0, la_min1-1)
         lb_min = MAX(0, lb_min1-1)
      ELSE
         do_dkh = .FALSE.
         la_max = la_max1
         lb_max = lb_max1
         la_min = la_min1
         lb_min = lb_min1
      END IF

      nmax = la_max+lb_max+1

      maxder_local = 0
      IF (PRESENT(maxder)) maxder_local = maxder

!   *** Calculate the distances of the centers a, b and c ***

      dab = SQRT(rab2)
      dac = SQRT(rac2)
      dbc = SQRT(rbc2)

!   *** Loop over all pairs of primitive Gaussian-type functions ***

      na = 0
      nap = 0

      DO ipgf = 1, npgfa

!     *** Screening ***

         IF (do_dkh) THEN
            IF (dkh_erfc) THEN
               IF (rpgfa(ipgf)+rpgfc < dac) THEN
                  na = na+ncoset(la_max1-maxder_local) !JT
                  nap = nap+ncoset(la_max1) !JT
                  CYCLE
               END IF
            END IF
         ELSE
            IF (rpgfa(ipgf)+rpgfc < dac) THEN
               na = na+ncoset(la_max1-maxder_local) !JT
               nap = nap+ncoset(la_max1) !JT
               CYCLE
            END IF
         END IF

         nb = 0

         DO jpgf = 1, npgfb

!       *** Screening ***
            IF (do_dkh) THEN
               IF (dkh_erfc) THEN
                  IF ((rpgfb(jpgf)+rpgfc < dbc) .OR. &
                      (rpgfa(ipgf)+rpgfb(jpgf) < dab)) THEN
                     nb = nb+ncoset(lb_max1) !JT
                     CYCLE
                  END IF
               END IF
            ELSE
               IF ((rpgfb(jpgf)+rpgfc < dbc) .OR. &
                   (rpgfa(ipgf)+rpgfb(jpgf) < dab)) THEN
                  nb = nb+ncoset(lb_max1) !JT
                  CYCLE
               END IF
            END IF
!       *** Calculate some prefactors ***

            zetp = 1.0_dp/(zeta(ipgf)+zetb(jpgf))
            zetq = 1.0_dp/zetc
            zetw = 1.0_dp/(zeta(ipgf)+zetb(jpgf)+zetc)

            f1 = zetb(jpgf)*zetp
            f2 = 0.5_dp*zetp
            f4 = -zetc*zetw

            f0 = EXP(-zeta(ipgf)*f1*rab2)

            fnuc = 2.0_dp*pi*zetp*f0
            ferf = 2.0_dp*SQRT(pi**5*zetw)*zetp*zetq*f0

            rap(:) = f1*rab(:)
            rcp(:) = rap(:)-rac(:)
            rpw(:) = f4*rcp(:)

!       *** Calculate the incomplete Gamma function values ***

            rcp2 = rcp(1)*rcp(1)+rcp(2)*rcp(2)+rcp(3)*rcp(3)

            t = rcp2/zetp

            CALL fgamma(nmax-1, t, f)

!       *** Calculate the basic nuclear attraction integrals [s|A(0)|s]{n} ***

            DO n = 1, nmax
               vnuc(1, 1, n) = fnuc*f(n-1)
            END DO

!       *** Calculate the incomplete Gamma function values ***

            t = -f4*rcp2/zetp

            CALL fgamma(nmax-1, t, f)

!       *** Calculate the basic three-center Coulomb integrals [ss||s]{n} ***

            DO n = 1, nmax
               verf(1, 1, n) = ferf*f(n-1)
            END DO

!       *** Recurrence steps: [s|A(0)|s] -> [a|A(0)|b] ***
!       ***                   [ss||s]    -> [ab||s]    ***

            IF (la_max > 0) THEN

!         *** Vertical recurrence steps: [s|A(0)|s] -> [a|A(0)|s] ***
!         ***                            [ss||s]    -> [as||s]    ***

!         *** [p|A(0)|s]{n} = (Pi - Ai)*[s|A(0)|s]{n} -           ***
!         ***                 (Pi - Ci)*[s|A(0)|s]{n+1}           ***
!         *** [ps||s]{n}    = (Pi - Ai)*[ss||s]{n} +              ***
!         ***                 (Wi - Pi)*[ss||s]{n+1}  (i = x,y,z) ***

               DO n = 1, nmax-1
                  vnuc(2, 1, n) = rap(1)*vnuc(1, 1, n)-rcp(1)*vnuc(1, 1, n+1)
                  verf(2, 1, n) = rap(1)*verf(1, 1, n)+rpw(1)*verf(1, 1, n+1)
                  vnuc(3, 1, n) = rap(2)*vnuc(1, 1, n)-rcp(2)*vnuc(1, 1, n+1)
                  verf(3, 1, n) = rap(2)*verf(1, 1, n)+rpw(2)*verf(1, 1, n+1)
                  vnuc(4, 1, n) = rap(3)*vnuc(1, 1, n)-rcp(3)*vnuc(1, 1, n+1)
                  verf(4, 1, n) = rap(3)*verf(1, 1, n)+rpw(3)*verf(1, 1, n+1)
               END DO

!         *** [a|A(0)|s]{n} = (Pi - Ai)*[a-1i|A(0)|s]{n} -       ***
!         ***                 (Pi - Ci)*[a-1i|A(0)|s]{n+1} +     ***
!         ***                 f2*Ni(a-1i)*([a-2i|A(0)|s]{n} -    ***
!         ***                              [a-2i|A(0)|s]{n+1}    ***
!         *** [as||s]{n}    = (Pi - Ai)*[(a-1i)s||s]{n} +        ***
!         ***                 (Wi - Pi)*[(a-1i)s||s]{n+1} +      ***
!         ***                 f2*Ni(a-1i)*(   [(a-2i)s||s]{n} +  ***
!         ***                              f4*[(a-2i)s||s]{n+1}) ***

               DO la = 2, la_max

                  DO n = 1, nmax-la

!             *** Increase the angular momentum component z of function a ***

                     vnuc(coset(0, 0, la), 1, n) = &
                        rap(3)*vnuc(coset(0, 0, la-1), 1, n)- &
                        rcp(3)*vnuc(coset(0, 0, la-1), 1, n+1)+ &
                        f2*REAL(la-1, dp)*(vnuc(coset(0, 0, la-2), 1, n)- &
                                           vnuc(coset(0, 0, la-2), 1, n+1))
                     verf(coset(0, 0, la), 1, n) = &
                        rap(3)*verf(coset(0, 0, la-1), 1, n)+ &
                        rpw(3)*verf(coset(0, 0, la-1), 1, n+1)+ &
                        f2*REAL(la-1, dp)*(verf(coset(0, 0, la-2), 1, n)+ &
                                           f4*verf(coset(0, 0, la-2), 1, n+1))

!             *** Increase the angular momentum component y of function a ***

                     az = la-1
                     vnuc(coset(0, 1, az), 1, n) = &
                        rap(2)*vnuc(coset(0, 0, az), 1, n)- &
                        rcp(2)*vnuc(coset(0, 0, az), 1, n+1)
                     verf(coset(0, 1, az), 1, n) = &
                        rap(2)*verf(coset(0, 0, az), 1, n)+ &
                        rpw(2)*verf(coset(0, 0, az), 1, n+1)

                     DO ay = 2, la
                        az = la-ay
                        vnuc(coset(0, ay, az), 1, n) = &
                           rap(2)*vnuc(coset(0, ay-1, az), 1, n)- &
                           rcp(2)*vnuc(coset(0, ay-1, az), 1, n+1)+ &
                           f2*REAL(ay-1, dp)*(vnuc(coset(0, ay-2, az), 1, n)- &
                                              vnuc(coset(0, ay-2, az), 1, n+1))
                        verf(coset(0, ay, az), 1, n) = &
                           rap(2)*verf(coset(0, ay-1, az), 1, n)+ &
                           rpw(2)*verf(coset(0, ay-1, az), 1, n+1)+ &
                           f2*REAL(ay-1, dp)*(verf(coset(0, ay-2, az), 1, n)+ &
                                              f4*verf(coset(0, ay-2, az), 1, n+1))
                     END DO

!             *** Increase the angular momentum component x of function a ***

                     DO ay = 0, la-1
                        az = la-1-ay
                        vnuc(coset(1, ay, az), 1, n) = &
                           rap(1)*vnuc(coset(0, ay, az), 1, n)- &
                           rcp(1)*vnuc(coset(0, ay, az), 1, n+1)
                        verf(coset(1, ay, az), 1, n) = &
                           rap(1)*verf(coset(0, ay, az), 1, n)+ &
                           rpw(1)*verf(coset(0, ay, az), 1, n+1)
                     END DO

                     DO ax = 2, la
                        f3 = f2*REAL(ax-1, dp)
                        DO ay = 0, la-ax
                           az = la-ax-ay
                           vnuc(coset(ax, ay, az), 1, n) = &
                              rap(1)*vnuc(coset(ax-1, ay, az), 1, n)- &
                              rcp(1)*vnuc(coset(ax-1, ay, az), 1, n+1)+ &
                              f3*(vnuc(coset(ax-2, ay, az), 1, n)- &
                                  vnuc(coset(ax-2, ay, az), 1, n+1))
                           verf(coset(ax, ay, az), 1, n) = &
                              rap(1)*verf(coset(ax-1, ay, az), 1, n)+ &
                              rpw(1)*verf(coset(ax-1, ay, az), 1, n+1)+ &
                              f3*(verf(coset(ax-2, ay, az), 1, n)+ &
                                  f4*verf(coset(ax-2, ay, az), 1, n+1))
                        END DO
                     END DO

                  END DO

               END DO

!         *** Recurrence steps: [a|A(0)|s] -> [a|A(0)|b] ***
!         ***                   [as||s]    -> [ab||s]    ***

               IF (lb_max > 0) THEN

!           *** Horizontal recurrence steps ***

                  rbp(:) = rap(:)-rab(:)

!           *** [a||A(0)|p]{n} = [a+1i|A(0)|s]{n} - (Bi - Ai)*[a|A(0)|s]{n} ***
!           *** [ap||s]{n}     = [(a+1i)s||s]{n}  - (Bi - Ai)*[as||s]{n}    ***

                  la_start = MAX(0, la_min-1)

                  DO la = la_start, la_max-1
                     DO n = 1, nmax-la-1
                        DO ax = 0, la
                           DO ay = 0, la-ax
                              az = la-ax-ay
                              vnuc(coset(ax, ay, az), 2, n) = &
                                 vnuc(coset(ax+1, ay, az), 1, n)- &
                                 rab(1)*vnuc(coset(ax, ay, az), 1, n)
                              verf(coset(ax, ay, az), 2, n) = &
                                 verf(coset(ax+1, ay, az), 1, n)- &
                                 rab(1)*verf(coset(ax, ay, az), 1, n)
                              vnuc(coset(ax, ay, az), 3, n) = &
                                 vnuc(coset(ax, ay+1, az), 1, n)- &
                                 rab(2)*vnuc(coset(ax, ay, az), 1, n)
                              verf(coset(ax, ay, az), 3, n) = &
                                 verf(coset(ax, ay+1, az), 1, n)- &
                                 rab(2)*verf(coset(ax, ay, az), 1, n)
                              vnuc(coset(ax, ay, az), 4, n) = &
                                 vnuc(coset(ax, ay, az+1), 1, n)- &
                                 rab(3)*vnuc(coset(ax, ay, az), 1, n)
                              verf(coset(ax, ay, az), 4, n) = &
                                 verf(coset(ax, ay, az+1), 1, n)- &
                                 rab(3)*verf(coset(ax, ay, az), 1, n)
                           END DO
                        END DO
                     END DO
                  END DO

!           *** Vertical recurrence step ***

!           *** [a|A(0)|p]{n} = (Pi - Bi)*[a|A(0)|s]{n} -        ***
!           ***                 (Pi - Ci)*[a|A(0)|s]{n+1} +      ***
!           ***                 f2*Ni(a)*([a-1i|A(0)|s]{n} -     ***
!           ***                           [a-1i|A(0)|s]{n+1})    ***
!           *** [ap||s]{n}    = (Pi - Bi)*[as||s]{n} +           ***
!           ***                 (Wi - Pi)*[as||s]{n+1} +         ***
!           ***                  f2*Ni(a)*(   [(a-1i)s||s]{n} +  ***
!           ***                            f4*[(a-1i)s||s]{n+1}) ***

                  DO n = 1, nmax-la_max-1
                     DO ax = 0, la_max
                        fx = f2*REAL(ax, dp)
                        DO ay = 0, la_max-ax
                           fy = f2*REAL(ay, dp)
                           az = la_max-ax-ay
                           fz = f2*REAL(az, dp)

                           IF (ax == 0) THEN
                              vnuc(coset(ax, ay, az), 2, n) = &
                                 rbp(1)*vnuc(coset(ax, ay, az), 1, n)- &
                                 rcp(1)*vnuc(coset(ax, ay, az), 1, n+1)
                              verf(coset(ax, ay, az), 2, n) = &
                                 rbp(1)*verf(coset(ax, ay, az), 1, n)+ &
                                 rpw(1)*verf(coset(ax, ay, az), 1, n+1)
                           ELSE
                              vnuc(coset(ax, ay, az), 2, n) = &
                                 rbp(1)*vnuc(coset(ax, ay, az), 1, n)- &
                                 rcp(1)*vnuc(coset(ax, ay, az), 1, n+1)+ &
                                 fx*(vnuc(coset(ax-1, ay, az), 1, n)- &
                                     vnuc(coset(ax-1, ay, az), 1, n+1))
                              verf(coset(ax, ay, az), 2, n) = &
                                 rbp(1)*verf(coset(ax, ay, az), 1, n)+ &
                                 rpw(1)*verf(coset(ax, ay, az), 1, n+1)+ &
                                 fx*(verf(coset(ax-1, ay, az), 1, n)+ &
                                     f4*verf(coset(ax-1, ay, az), 1, n+1))
                           END IF

                           IF (ay == 0) THEN
                              vnuc(coset(ax, ay, az), 3, n) = &
                                 rbp(2)*vnuc(coset(ax, ay, az), 1, n)- &
                                 rcp(2)*vnuc(coset(ax, ay, az), 1, n+1)
                              verf(coset(ax, ay, az), 3, n) = &
                                 rbp(2)*verf(coset(ax, ay, az), 1, n)+ &
                                 rpw(2)*verf(coset(ax, ay, az), 1, n+1)
                           ELSE
                              vnuc(coset(ax, ay, az), 3, n) = &
                                 rbp(2)*vnuc(coset(ax, ay, az), 1, n)- &
                                 rcp(2)*vnuc(coset(ax, ay, az), 1, n+1)+ &
                                 fy*(vnuc(coset(ax, ay-1, az), 1, n)- &
                                     vnuc(coset(ax, ay-1, az), 1, n+1))
                              verf(coset(ax, ay, az), 3, n) = &
                                 rbp(2)*verf(coset(ax, ay, az), 1, n)+ &
                                 rpw(2)*verf(coset(ax, ay, az), 1, n+1)+ &
                                 fy*(verf(coset(ax, ay-1, az), 1, n)+ &
                                     f4*verf(coset(ax, ay-1, az), 1, n+1))
                           END IF

                           IF (az == 0) THEN
                              vnuc(coset(ax, ay, az), 4, n) = &
                                 rbp(3)*vnuc(coset(ax, ay, az), 1, n)- &
                                 rcp(3)*vnuc(coset(ax, ay, az), 1, n+1)
                              verf(coset(ax, ay, az), 4, n) = &
                                 rbp(3)*verf(coset(ax, ay, az), 1, n)+ &
                                 rpw(3)*verf(coset(ax, ay, az), 1, n+1)
                           ELSE
                              vnuc(coset(ax, ay, az), 4, n) = &
                                 rbp(3)*vnuc(coset(ax, ay, az), 1, n)- &
                                 rcp(3)*vnuc(coset(ax, ay, az), 1, n+1)+ &
                                 fz*(vnuc(coset(ax, ay, az-1), 1, n)- &
                                     vnuc(coset(ax, ay, az-1), 1, n+1))
                              verf(coset(ax, ay, az), 4, n) = &
                                 rbp(3)*verf(coset(ax, ay, az), 1, n)+ &
                                 rpw(3)*verf(coset(ax, ay, az), 1, n+1)+ &
                                 fz*(verf(coset(ax, ay, az-1), 1, n)+ &
                                     f4*verf(coset(ax, ay, az-1), 1, n+1))
                           END IF

                        END DO
                     END DO
                  END DO

!           *** Recurrence steps: [a|A(0)|p] -> [a|A(0)|b] ***
!           ***                   [ap||s]    -> [ab||s]    ***

                  DO lb = 2, lb_max

!             *** Horizontal recurrence steps ***

!             *** [a||A(0)|b]{n} = [a+1i|A(0)|b-1i]{n} -      ***
!             ***                  (Bi - Ai)*[a|A(0)|b-1i]{n} ***
!             *** [ab||s]{n}     = [(a+1i)(b-1i)||s]{n} -     ***
!             ***                  (Bi - Ai)*[a(b-1i)||s]{n}  ***

                     la_start = MAX(0, la_min-1)

                     DO la = la_start, la_max-1
                        DO n = 1, nmax-la-lb
                           DO ax = 0, la
                              DO ay = 0, la-ax
                                 az = la-ax-ay

!                     *** Shift of angular momentum component z from a to b ***

                                 vnuc(coset(ax, ay, az), coset(0, 0, lb), n) = &
                                    vnuc(coset(ax, ay, az+1), coset(0, 0, lb-1), n)- &
                                    rab(3)*vnuc(coset(ax, ay, az), coset(0, 0, lb-1), n)
                                 verf(coset(ax, ay, az), coset(0, 0, lb), n) = &
                                    verf(coset(ax, ay, az+1), coset(0, 0, lb-1), n)- &
                                    rab(3)*verf(coset(ax, ay, az), coset(0, 0, lb-1), n)

!                     *** Shift of angular momentum component y from a to b ***

                                 DO by = 1, lb
                                    bz = lb-by
                                    vnuc(coset(ax, ay, az), coset(0, by, bz), n) = &
                                       vnuc(coset(ax, ay+1, az), coset(0, by-1, bz), n)- &
                                       rab(2)*vnuc(coset(ax, ay, az), coset(0, by-1, bz), n)
                                    verf(coset(ax, ay, az), coset(0, by, bz), n) = &
                                       verf(coset(ax, ay+1, az), coset(0, by-1, bz), n)- &
                                       rab(2)*verf(coset(ax, ay, az), coset(0, by-1, bz), n)
                                 END DO

!                     *** Shift of angular momentum component x from a to b ***

                                 DO bx = 1, lb
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       vnuc(coset(ax, ay, az), coset(bx, by, bz), n) = &
                                          vnuc(coset(ax+1, ay, az), coset(bx-1, by, bz), n)- &
                                          rab(1)*vnuc(coset(ax, ay, az), coset(bx-1, by, bz), n)
                                       verf(coset(ax, ay, az), coset(bx, by, bz), n) = &
                                          verf(coset(ax+1, ay, az), coset(bx-1, by, bz), n)- &
                                          rab(1)*verf(coset(ax, ay, az), coset(bx-1, by, bz), n)
                                    END DO
                                 END DO

                              END DO
                           END DO
                        END DO
                     END DO

!             *** Vertical recurrence step ***

!             *** [a|A(0)|b]{n} = (Pi - Bi)*[a|A(0)|b-1i]{n} -         ***
!             ***                 (Pi - Ci)*[a|A(0)|b-1i]{n+1} +       ***
!             ***                 f2*Ni(a)*([a-1i|A(0)|b-1i]{n} -      ***
!             ***                           [a-1i|A(0)|b-1i]{n+1}) +   ***
!             ***                 f2*Ni(b-1i)*([a|A(0)|b-2i]{n} -      ***
!             ***                              [a|A(0)|b-2i]{n+1})     ***
!             *** [ab||s]{n}    = (Pi - Bi)*[a(b-1i)||s]{n} +          ***
!             ***                 (Wi - Pi)*[a(b-1i)||s]{n+1} +        ***
!             ***                 f2*Ni(a)*(   [(a-1i)(b-1i)||s]{n} +  ***
!             ***                           f4*[(a-1i)(b-1i)||s]{n+1}) ***
!             ***                 f2*Ni(b-1i)*(   [a(b-2i)||s]{n} +    ***
!             ***                              f4*[a(b-2i)||s]{n+1})   ***

                     DO n = 1, nmax-la_max-lb
                        DO ax = 0, la_max
                           fx = f2*REAL(ax, dp)
                           DO ay = 0, la_max-ax
                              fy = f2*REAL(ay, dp)
                              az = la_max-ax-ay
                              fz = f2*REAL(az, dp)

!                   *** Shift of angular momentum component z from a to b ***

                              f3 = f2*REAL(lb-1, dp)

                              IF (az == 0) THEN
                                 vnuc(coset(ax, ay, az), coset(0, 0, lb), n) = &
                                    rbp(3)*vnuc(coset(ax, ay, az), coset(0, 0, lb-1), n)- &
                                    rcp(3)*vnuc(coset(ax, ay, az), coset(0, 0, lb-1), n+1)+ &
                                    f3*(vnuc(coset(ax, ay, az), coset(0, 0, lb-2), n)- &
                                        vnuc(coset(ax, ay, az), coset(0, 0, lb-2), n+1))
                                 verf(coset(ax, ay, az), coset(0, 0, lb), n) = &
                                    rbp(3)*verf(coset(ax, ay, az), coset(0, 0, lb-1), n)+ &
                                    rpw(3)*verf(coset(ax, ay, az), coset(0, 0, lb-1), n+1)+ &
                                    f3*(verf(coset(ax, ay, az), coset(0, 0, lb-2), n)+ &
                                        f4*verf(coset(ax, ay, az), coset(0, 0, lb-2), n+1))
                              ELSE
                                 vnuc(coset(ax, ay, az), coset(0, 0, lb), n) = &
                                    rbp(3)*vnuc(coset(ax, ay, az), coset(0, 0, lb-1), n)- &
                                    rcp(3)*vnuc(coset(ax, ay, az), coset(0, 0, lb-1), n+1)+ &
                                    fz*(vnuc(coset(ax, ay, az-1), coset(0, 0, lb-1), n)- &
                                        vnuc(coset(ax, ay, az-1), coset(0, 0, lb-1), n+1))+ &
                                    f3*(vnuc(coset(ax, ay, az), coset(0, 0, lb-2), n)- &
                                        vnuc(coset(ax, ay, az), coset(0, 0, lb-2), n+1))
                                 verf(coset(ax, ay, az), coset(0, 0, lb), n) = &
                                    rbp(3)*verf(coset(ax, ay, az), coset(0, 0, lb-1), n)+ &
                                    rpw(3)*verf(coset(ax, ay, az), coset(0, 0, lb-1), n+1)+ &
                                    fz*(verf(coset(ax, ay, az-1), coset(0, 0, lb-1), n)+ &
                                        f4*verf(coset(ax, ay, az-1), coset(0, 0, lb-1), n+1))+ &
                                    f3*(verf(coset(ax, ay, az), coset(0, 0, lb-2), n)+ &
                                        f4*verf(coset(ax, ay, az), coset(0, 0, lb-2), n+1))
                              END IF

!                   *** Shift of angular momentum component y from a to b ***

                              IF (ay == 0) THEN
                                 bz = lb-1
                                 vnuc(coset(ax, ay, az), coset(0, 1, bz), n) = &
                                    rbp(2)*vnuc(coset(ax, ay, az), coset(0, 0, bz), n)- &
                                    rcp(2)*vnuc(coset(ax, ay, az), coset(0, 0, bz), n+1)
                                 verf(coset(ax, ay, az), coset(0, 1, bz), n) = &
                                    rbp(2)*verf(coset(ax, ay, az), coset(0, 0, bz), n)+ &
                                    rpw(2)*verf(coset(ax, ay, az), coset(0, 0, bz), n+1)
                                 DO by = 2, lb
                                    bz = lb-by
                                    f3 = f2*REAL(by-1, dp)
                                    vnuc(coset(ax, ay, az), coset(0, by, bz), n) = &
                                       rbp(2)*vnuc(coset(ax, ay, az), coset(0, by-1, bz), n)- &
                                       rcp(2)*vnuc(coset(ax, ay, az), coset(0, by-1, bz), n+1)+ &
                                       f3*(vnuc(coset(ax, ay, az), coset(0, by-2, bz), n)- &
                                           vnuc(coset(ax, ay, az), coset(0, by-2, bz), n+1))
                                    verf(coset(ax, ay, az), coset(0, by, bz), n) = &
                                       rbp(2)*verf(coset(ax, ay, az), coset(0, by-1, bz), n)+ &
                                       rpw(2)*verf(coset(ax, ay, az), coset(0, by-1, bz), n+1)+ &
                                       f3*(verf(coset(ax, ay, az), coset(0, by-2, bz), n)+ &
                                           f4*verf(coset(ax, ay, az), coset(0, by-2, bz), n+1))
                                 END DO
                              ELSE
                                 bz = lb-1
                                 vnuc(coset(ax, ay, az), coset(0, 1, bz), n) = &
                                    rbp(2)*vnuc(coset(ax, ay, az), coset(0, 0, bz), n)- &
                                    rcp(2)*vnuc(coset(ax, ay, az), coset(0, 0, bz), n+1)+ &
                                    fy*(vnuc(coset(ax, ay-1, az), coset(0, 0, bz), n)- &
                                        vnuc(coset(ax, ay-1, az), coset(0, 0, bz), n+1))
                                 verf(coset(ax, ay, az), coset(0, 1, bz), n) = &
                                    rbp(2)*verf(coset(ax, ay, az), coset(0, 0, bz), n)+ &
                                    rpw(2)*verf(coset(ax, ay, az), coset(0, 0, bz), n+1)+ &
                                    fy*(verf(coset(ax, ay-1, az), coset(0, 0, bz), n)+ &
                                        f4*verf(coset(ax, ay-1, az), coset(0, 0, bz), n+1))
                                 DO by = 2, lb
                                    bz = lb-by
                                    f3 = f2*REAL(by-1, dp)
                                    vnuc(coset(ax, ay, az), coset(0, by, bz), n) = &
                                       rbp(2)*vnuc(coset(ax, ay, az), coset(0, by-1, bz), n)- &
                                       rcp(2)*vnuc(coset(ax, ay, az), coset(0, by-1, bz), n+1)+ &
                                       fy*(vnuc(coset(ax, ay-1, az), coset(0, by-1, bz), n)- &
                                           vnuc(coset(ax, ay-1, az), coset(0, by-1, bz), n+1))+ &
                                       f3*(vnuc(coset(ax, ay, az), coset(0, by-2, bz), n)- &
                                           vnuc(coset(ax, ay, az), coset(0, by-2, bz), n+1))
                                    verf(coset(ax, ay, az), coset(0, by, bz), n) = &
                                       rbp(2)*verf(coset(ax, ay, az), coset(0, by-1, bz), n)+ &
                                       rpw(2)*verf(coset(ax, ay, az), coset(0, by-1, bz), n+1)+ &
                                       fy*(verf(coset(ax, ay-1, az), coset(0, by-1, bz), n)+ &
                                           f4*verf(coset(ax, ay-1, az), &
                                                   coset(0, by-1, bz), n+1))+ &
                                       f3*(verf(coset(ax, ay, az), coset(0, by-2, bz), n)+ &
                                           f4*verf(coset(ax, ay, az), coset(0, by-2, bz), n+1))
                                 END DO
                              END IF

!                   *** Shift of angular momentum component x from a to b ***

                              IF (ax == 0) THEN
                                 DO by = 0, lb-1
                                    bz = lb-1-by
                                    vnuc(coset(ax, ay, az), coset(1, by, bz), n) = &
                                       rbp(1)*vnuc(coset(ax, ay, az), coset(0, by, bz), n)- &
                                       rcp(1)*vnuc(coset(ax, ay, az), coset(0, by, bz), n+1)
                                    verf(coset(ax, ay, az), coset(1, by, bz), n) = &
                                       rbp(1)*verf(coset(ax, ay, az), coset(0, by, bz), n)+ &
                                       rpw(1)*verf(coset(ax, ay, az), coset(0, by, bz), n+1)
                                 END DO
                                 DO bx = 2, lb
                                    f3 = f2*REAL(bx-1, dp)
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       vnuc(coset(ax, ay, az), coset(bx, by, bz), n) = &
                                          rbp(1)*vnuc(coset(ax, ay, az), coset(bx-1, by, bz), n)- &
                                          rcp(1)*vnuc(coset(ax, ay, az), &
                                                      coset(bx-1, by, bz), n+1)+ &
                                          f3*(vnuc(coset(ax, ay, az), coset(bx-2, by, bz), n)- &
                                              vnuc(coset(ax, ay, az), coset(bx-2, by, bz), n+1))
                                       verf(coset(ax, ay, az), coset(bx, by, bz), n) = &
                                          rbp(1)*verf(coset(ax, ay, az), coset(bx-1, by, bz), n)+ &
                                          rpw(1)*verf(coset(ax, ay, az), &
                                                      coset(bx-1, by, bz), n+1)+ &
                                          f3*(verf(coset(ax, ay, az), coset(bx-2, by, bz), n)+ &
                                              f4*verf(coset(ax, ay, az), coset(bx-2, by, bz), n+1))
                                    END DO
                                 END DO
                              ELSE
                                 DO by = 0, lb-1
                                    bz = lb-1-by
                                    vnuc(coset(ax, ay, az), coset(1, by, bz), n) = &
                                       rbp(1)*vnuc(coset(ax, ay, az), coset(0, by, bz), n)- &
                                       rcp(1)*vnuc(coset(ax, ay, az), coset(0, by, bz), n+1)+ &
                                       fx*(vnuc(coset(ax-1, ay, az), coset(0, by, bz), n)- &
                                           vnuc(coset(ax-1, ay, az), coset(0, by, bz), n+1))
                                    verf(coset(ax, ay, az), coset(1, by, bz), n) = &
                                       rbp(1)*verf(coset(ax, ay, az), coset(0, by, bz), n)+ &
                                       rpw(1)*verf(coset(ax, ay, az), coset(0, by, bz), n+1)+ &
                                       fx*(verf(coset(ax-1, ay, az), coset(0, by, bz), n)+ &
                                           f4*verf(coset(ax-1, ay, az), coset(0, by, bz), n+1))
                                 END DO
                                 DO bx = 2, lb
                                    f3 = f2*REAL(bx-1, dp)
                                    DO by = 0, lb-bx
                                       bz = lb-bx-by
                                       vnuc(coset(ax, ay, az), coset(bx, by, bz), n) = &
                                          rbp(1)*vnuc(coset(ax, ay, az), coset(bx-1, by, bz), n)- &
                                          rcp(1)*vnuc(coset(ax, ay, az), &
                                                      coset(bx-1, by, bz), n+1)+ &
                                          fx*(vnuc(coset(ax-1, ay, az), coset(bx-1, by, bz), n)- &
                                              vnuc(coset(ax-1, ay, az), &
                                                   coset(bx-1, by, bz), n+1))+ &
                                          f3*(vnuc(coset(ax, ay, az), coset(bx-2, by, bz), n)- &
                                              vnuc(coset(ax, ay, az), coset(bx-2, by, bz), n+1))
                                       verf(coset(ax, ay, az), coset(bx, by, bz), n) = &
                                          rbp(1)*verf(coset(ax, ay, az), coset(bx-1, by, bz), n)+ &
                                          rpw(1)*verf(coset(ax, ay, az), &
                                                      coset(bx-1, by, bz), n+1)+ &
                                          fx*(verf(coset(ax-1, ay, az), &
                                                   coset(bx-1, by, bz), n)+ &
                                              f4*verf(coset(ax-1, ay, az), &
                                                      coset(bx-1, by, bz), n+1))+ &
                                          f3*(verf(coset(ax, ay, az), coset(bx-2, by, bz), n)+ &
                                              f4*verf(coset(ax, ay, az), coset(bx-2, by, bz), n+1))
                                    END DO
                                 END DO
                              END IF

                           END DO
                        END DO
                     END DO

                  END DO

               END IF

            ELSE

               IF (lb_max > 0) THEN

!           *** Vertical recurrence steps: [s|A(0)|s] -> [s|A(0)|b] ***
!           ***                            [ss||s]    -> [sb||s]    ***

                  rbp(:) = rap(:)-rab(:)

!           *** [s|A(0)|p]{n} = (Pi - Bi)*[s|A(0)|s]{n} - ***
!           ***                 (Pi - Ci)*[s|A(0)|s]{n+1} ***
!           *** [sp||s]{n}    = (Pi - Bi)*[ss||s]{n} +    ***
!           ***                 (Wi - Pi)*[ss||s]{n+1}    ***

                  DO n = 1, nmax-1
                     vnuc(1, 2, n) = rbp(1)*vnuc(1, 1, n)-rcp(1)*vnuc(1, 1, n+1)
                     verf(1, 2, n) = rbp(1)*verf(1, 1, n)+rpw(1)*verf(1, 1, n+1)
                     vnuc(1, 3, n) = rbp(2)*vnuc(1, 1, n)-rcp(2)*vnuc(1, 1, n+1)
                     verf(1, 3, n) = rbp(2)*verf(1, 1, n)+rpw(2)*verf(1, 1, n+1)
                     vnuc(1, 4, n) = rbp(3)*vnuc(1, 1, n)-rcp(3)*vnuc(1, 1, n+1)
                     verf(1, 4, n) = rbp(3)*verf(1, 1, n)+rpw(3)*verf(1, 1, n+1)
                  END DO

!           *** [s|A(0)|b]{n} = (Pi - Bi)*[s|A(0)|b-1i]{n} -       ***
!           ***                 (Pi - Ci)*[s|A(0)|b-1i]{n+1} +     ***
!           ***                 f2*Ni(b-1i)*([s|A(0)|b-2i]{n} -    ***
!           ***                              [s|A(0)|b-2i]{n+1}    ***
!           *** [sb||s]{n}    = (Pi - Bi)*[s(b-1i)||s]{n} +        ***
!           ***                 (Wi - Pi)*[s(b-1i)||s]{n+1} +      ***
!           ***                 f2*Ni(b-1i)*(   [s(b-2i)||s]{n} +  ***
!           ***                              f4*[s(b-2i)||s]{n+1}) ***

                  DO lb = 2, lb_max

                     DO n = 1, nmax-lb

!               *** Increase the angular momentum component z of function b ***

                        vnuc(1, coset(0, 0, lb), n) = &
                           rbp(3)*vnuc(1, coset(0, 0, lb-1), n)- &
                           rcp(3)*vnuc(1, coset(0, 0, lb-1), n+1)+ &
                           f2*REAL(lb-1, dp)*(vnuc(1, coset(0, 0, lb-2), n)- &
                                              vnuc(1, coset(0, 0, lb-2), n+1))
                        verf(1, coset(0, 0, lb), n) = &
                           rbp(3)*verf(1, coset(0, 0, lb-1), n)+ &
                           rpw(3)*verf(1, coset(0, 0, lb-1), n+1)+ &
                           f2*REAL(lb-1, dp)*(verf(1, coset(0, 0, lb-2), n)+ &
                                              f4*verf(1, coset(0, 0, lb-2), n+1))

!               *** Increase the angular momentum component y of function b ***

                        bz = lb-1
                        vnuc(1, coset(0, 1, bz), n) = &
                           rbp(2)*vnuc(1, coset(0, 0, bz), n)- &
                           rcp(2)*vnuc(1, coset(0, 0, bz), n+1)
                        verf(1, coset(0, 1, bz), n) = &
                           rbp(2)*verf(1, coset(0, 0, bz), n)+ &
                           rpw(2)*verf(1, coset(0, 0, bz), n+1)

                        DO by = 2, lb
                           bz = lb-by
                           vnuc(1, coset(0, by, bz), n) = &
                              rbp(2)*vnuc(1, coset(0, by-1, bz), n)- &
                              rcp(2)*vnuc(1, coset(0, by-1, bz), n+1)+ &
                              f2*REAL(by-1, dp)*(vnuc(1, coset(0, by-2, bz), n)- &
                                                 vnuc(1, coset(0, by-2, bz), n+1))
                           verf(1, coset(0, by, bz), n) = &
                              rbp(2)*verf(1, coset(0, by-1, bz), n)+ &
                              rpw(2)*verf(1, coset(0, by-1, bz), n+1)+ &
                              f2*REAL(by-1, dp)*(verf(1, coset(0, by-2, bz), n)+ &
                                                 f4*verf(1, coset(0, by-2, bz), n+1))
                        END DO

!               *** Increase the angular momentum component x of function b ***

                        DO by = 0, lb-1
                           bz = lb-1-by
                           vnuc(1, coset(1, by, bz), n) = &
                              rbp(1)*vnuc(1, coset(0, by, bz), n)- &
                              rcp(1)*vnuc(1, coset(0, by, bz), n+1)
                           verf(1, coset(1, by, bz), n) = &
                              rbp(1)*verf(1, coset(0, by, bz), n)+ &
                              rpw(1)*verf(1, coset(0, by, bz), n+1)
                        END DO

                        DO bx = 2, lb
                           f3 = f2*REAL(bx-1, dp)
                           DO by = 0, lb-bx
                              bz = lb-bx-by
                              vnuc(1, coset(bx, by, bz), n) = &
                                 rbp(1)*vnuc(1, coset(bx-1, by, bz), n)- &
                                 rcp(1)*vnuc(1, coset(bx-1, by, bz), n+1)+ &
                                 f3*(vnuc(1, coset(bx-2, by, bz), n)- &
                                     vnuc(1, coset(bx-2, by, bz), n+1))
                              verf(1, coset(bx, by, bz), n) = &
                                 rbp(1)*verf(1, coset(bx-1, by, bz), n)+ &
                                 rpw(1)*verf(1, coset(bx-1, by, bz), n+1)+ &
                                 f3*(verf(1, coset(bx-2, by, bz), n)+ &
                                     f4*verf(1, coset(bx-2, by, bz), n+1))
                           END DO
                        END DO

                     END DO

                  END DO

               END IF

            END IF

!       *** Add the contribution of the current pair ***
!       *** of primitive Gaussian-type functions     ***

!JT
            IF (do_dkh) THEN
               DO j = 1, ncoset(lb_max1)
                  DO i = ncoset(la_min1-1)+1, ncoset(la_max1-maxder_local)
                     vnabc(na+i, nb+j) = vnabc(na+i, nb+j)+cerf*verf(i, j, 1)
                  END DO
               END DO
               DO j = 1, ncoset(lb_max1)
                  DO i = ncoset(la_min1-1)+1, ncoset(la_max1-maxder_local)
                     vabc(na+i, nb+j) = vabc(na+i, nb+j)-zc*vnuc(i, j, 1)
                  END DO
               END DO
            ELSE
               DO j = 1, ncoset(lb_max1) !JT
                  DO i = ncoset(la_min1-1)+1, ncoset(la_max1-maxder_local) !JT
                     vabc(na+i, nb+j) = vabc(na+i, nb+j)- &
                                        zc*vnuc(i, j, 1)+cerf*verf(i, j, 1)
                  END DO
               END DO
            END IF
!JT

            IF (PRESENT(maxder)) THEN
               DO j = 1, ncoset(lb_max1) !JT
                  DO i = 1, ncoset(la_max1) !JT
                     vabc_plus(nap+i, nb+j) = vabc_plus(nap+i, nb+j)- &
                                              zc*vnuc(i, j, 1)+cerf*verf(i, j, 1)
                  END DO
               END DO
            END IF
!JTs
            IF (do_dkh) THEN
               ! *********** Calculate pVp matrix **************
               ! [pa|V|pb]=4*zeta*zetb*[a+i|V|b+i]-2*zetb*Ni(a)*[a-i|V|b+i]-2*zeta*Ni(b)*[a+i|V|b-i]+Ni(a)*Ni(b)*[a-i|V|b-i]
               ! Every integral has been calculated before (except [-1|V|-1],[a|V|-1] and [-1|V|b],
               ! which do not contribute, because their prefactor Ni(a) or Ni(b) is zero)
               ! and is given by verf/vnuc(coset(ax,ay,az),coset(bx,by,bz),1)
               ! ***********************************************
               ftaz = 2.0_dp*zeta(ipgf)
               ftbz = 2.0_dp*zetb(jpgf)
               nxx = 1
               nyy = 2
               nzz = 3
               nxy = 4
               nxz = 5
               nyx = 6
               nyz = 7
               nzx = 8
               nzy = 9
               DO la = 0, la_max1
                  DO ax = 0, la
                     fax = REAL(ax, dp)
                     DO ay = 0, la-ax
                        fay = REAL(ay, dp)
                        az = la-ax-ay
                        faz = REAL(az, dp)
                        pVpa = coset(ax, ay, az)
                        coamx = coset(ax-1, ay, az)
                        coamy = coset(ax, ay-1, az)
                        coamz = coset(ax, ay, az-1)
                        coapx = coset(ax+1, ay, az)
                        coapy = coset(ax, ay+1, az)
                        coapz = coset(ax, ay, az+1)
                        DO lb = 0, lb_max1
                           DO bx = 0, lb
                              fbx = REAL(bx, dp)
                              DO by = 0, lb-bx
                                 fby = REAL(by, dp)
                                 bz = lb-bx-by
                                 fbz = REAL(bz, dp)
                                 pVpb = coset(bx, by, bz)
                                 cobmx = coset(bx-1, by, bz)
                                 cobmy = coset(bx, by-1, bz)
                                 cobmz = coset(bx, by, bz-1)
                                 cobpx = coset(bx+1, by, bz)
                                 cobpy = coset(bx, by+1, bz)
                                 cobpz = coset(bx, by, bz+1)
                                 IF (dkh_erfc) THEN
                                    pVp(pVpa, pVpb) = ftaz*ftbz*(-zc*vnuc(coapx, cobpx, 1)+cerf*verf(coapx, cobpx, 1))- &
                                                      ftaz*fbx*(-zc*vnuc(coapx, cobmx, 1)+cerf*verf(coapx, cobmx, 1))- &
                                                      ftbz*fax*(-zc*vnuc(coamx, cobpx, 1)+cerf*verf(coamx, cobpx, 1))+ &
                                                      fax*fbx*(-zc*vnuc(coamx, cobmx, 1)+cerf*verf(coamx, cobmx, 1))+ &
                                                      ftaz*ftbz*(-zc*vnuc(coapy, cobpy, 1)+cerf*verf(coapy, cobpy, 1))- &
                                                      ftaz*fby*(-zc*vnuc(coapy, cobmy, 1)+cerf*verf(coapy, cobmy, 1))- &
                                                      ftbz*fay*(-zc*vnuc(coamy, cobpy, 1)+cerf*verf(coamy, cobpy, 1))+ &
                                                      fay*fby*(-zc*vnuc(coamy, cobmy, 1)+cerf*verf(coamy, cobmy, 1))+ &
                                                      ftaz*ftbz*(-zc*vnuc(coapz, cobpz, 1)+cerf*verf(coapz, cobpz, 1))- &
                                                      ftaz*fbz*(-zc*vnuc(coapz, cobmz, 1)+cerf*verf(coapz, cobmz, 1))- &
                                                      ftbz*faz*(-zc*vnuc(coamz, cobpz, 1)+cerf*verf(coamz, cobpz, 1))+ &
                                                      faz*fbz*(-zc*vnuc(coamz, cobmz, 1)+cerf*verf(coamz, cobmz, 1))
                                 ELSE
                                    pVp(pVpa, pVpb) = ftaz*ftbz*(-zc*vnuc(coapx, cobpx, 1))- &
                                                      ftaz*fbx*(-zc*vnuc(coapx, cobmx, 1))- &
                                                      ftbz*fax*(-zc*vnuc(coamx, cobpx, 1))+ &
                                                      fax*fbx*(-zc*vnuc(coamx, cobmx, 1))+ &
                                                      ftaz*ftbz*(-zc*vnuc(coapy, cobpy, 1))- &
                                                      ftaz*fby*(-zc*vnuc(coapy, cobmy, 1))- &
                                                      ftbz*fay*(-zc*vnuc(coamy, cobpy, 1))+ &
                                                      fay*fby*(-zc*vnuc(coamy, cobmy, 1))+ &
                                                      ftaz*ftbz*(-zc*vnuc(coapz, cobpz, 1))- &
                                                      ftaz*fbz*(-zc*vnuc(coapz, cobmz, 1))- &
                                                      ftbz*faz*(-zc*vnuc(coamz, cobpz, 1))+ &
                                                      faz*fbz*(-zc*vnuc(coamz, cobmz, 1))
                                 END IF
                              END DO
                           END DO
                        END DO
                     END DO
                  END DO
               END DO

               DO j = 1, ncoset(lb_max1)
                  DO i = ncoset(la_min1-1)+1, ncoset(la_max1-maxder_local)
                     pVp_sum(na+i, nb+j) = pVp_sum(na+i, nb+j)+pVp(i, j)
                  END DO
               END DO
            END IF
!JTe
            nb = nb+ncoset(lb_max1) !JT

         END DO

         na = na+ncoset(la_max1-maxder_local) !JT
         nap = nap+ncoset(la_max1) !JT

      END DO

   END SUBROUTINE verfc

END MODULE ai_verfc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Calculation of the nuclear attraction contribution to the core Hamiltonian
!>         <a|erfc|b> :we only calculate the non-screened part
!> \par History
!>      - core_ppnl refactored from qs_core_hamiltonian [Joost VandeVondele, 2008-11-01]
!>      - adapted for nuclear attraction [jhu, 2009-02-24]
! **************************************************************************************************
MODULE core_ae
   USE ai_verfc,                        ONLY: verfc
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE external_potential_types,        ONLY: all_potential_type,&
                                              get_potential
   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: coset,&
                                              indco,&
                                              init_orbital_pointers,&
                                              ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type,&
                                              nl_set_sub_iterator,&
                                              nl_sub_iterate
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'core_ae'

   PUBLIC :: build_core_ae

CONTAINS

!==========================================================================================================

! **************************************************************************************************
!> \brief ...
!> \param matrix_h ...
!> \param matrix_p ...
!> \param force ...
!> \param virial ...
!> \param calculate_forces ...
!> \param use_virial ...
!> \param nder ...
!> \param qs_kind_set ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param sab_orb ...
!> \param sac_ae ...
!> \param nimages ...
!> \param cell_to_index ...
! **************************************************************************************************
   SUBROUTINE build_core_ae(matrix_h, matrix_p, force, virial, calculate_forces, use_virial, nder, &
                            qs_kind_set, atomic_kind_set, particle_set, sab_orb, sac_ae, nimages, cell_to_index)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_h, matrix_p
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(virial_type), POINTER                         :: virial
      LOGICAL, INTENT(IN)                                :: calculate_forces
      LOGICAL                                            :: use_virial
      INTEGER                                            :: nder
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb, sac_ae
      INTEGER, INTENT(IN)                                :: nimages
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_core_ae', routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, atom_c, handle, iatom, icol, ikind, img, inode, irow, iset, &
         jatom, jkind, jset, katom, kkind, ldai, ldsab, maxco, maxder, maxl, maxnset, maxsgf, &
         na_plus, natom, nb_plus, ncoa, ncob, nij, nkind, nseta, nsetb, sgfa, sgfb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, DIMENSION(3)                              :: cellind
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: dokp, found
      REAL(KIND=dp)                                      :: alpha_c, core_charge, core_radius, dab, &
                                                            dac, dbc, f0, rab2, rac2, rbc2, zeta_c
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: habd, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: hab, pab, verf, vnuc
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, rab, rac, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: h_block, p_block, rpgfa, rpgfb, sphi_a, &
                                                            sphi_b, zeta, zetb
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: ap_iterator, nl_iterator

      IF (calculate_forces) THEN
         CALL timeset(routineN//"_forces", handle)
      ELSE
         CALL timeset(routineN, handle)
      ENDIF

      nkind = SIZE(atomic_kind_set)
      natom = SIZE(particle_set)

      dokp = (nimages > 1)

      ALLOCATE (atom_of_kind(natom))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)

      IF (calculate_forces) THEN
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, nimages
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
               CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                 alpha_scalar=-2.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      END IF

      maxder = ncoset(nder)

      CALL get_qs_kind_set(qs_kind_set, &
                           maxco=maxco, maxlgto=maxl, maxsgf=maxsgf, maxnset=maxnset)

      CALL init_orbital_pointers(maxl+nder+1)

      ldsab = MAX(maxco, maxsgf)
      ldai = ncoset(maxl+nder+1)
      ALLOCATE (hab(ldsab, ldsab, maxnset*maxnset), work(ldsab, ldsab))
      ALLOCATE (verf(ldai, ldai, 2*maxl+nder+1), vnuc(ldai, ldai, 2*maxl+nder+1), ff(0:2*maxl+nder))
      IF (calculate_forces) THEN
         ALLOCATE (pab(maxco, maxco, maxnset*maxnset))
      END IF

      ! iterator for basis/potential list
      CALL neighbor_list_iterator_create(ap_iterator, sac_ae, search=.TRUE.)

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab, cell=cellind)
         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE
         atom_a = atom_of_kind(iatom)
         atom_b = atom_of_kind(jatom)
         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         dab = SQRT(SUM(rab*rab))

         IF (dokp) THEN
            img = cell_to_index(cellind(1), cellind(2), cellind(3))
         ELSE
            img = 1
         END IF

         ! *** Use the symmetry of the first derivatives ***
         IF (iatom == jatom) THEN
            f0 = 1.0_dp
         ELSE
            f0 = 2.0_dp
         END IF

         ! *** Create matrix blocks for a new matrix block column ***
         IF (iatom <= jatom) THEN
            irow = iatom
            icol = jatom
         ELSE
            irow = jatom
            icol = iatom
         END IF
         NULLIFY (h_block)
         CALL cp_dbcsr_get_block_p(matrix=matrix_h(1, img)%matrix, &
                                   row=irow, col=icol, BLOCK=h_block, found=found)
         IF (calculate_forces) THEN
            NULLIFY (p_block)
            CALL cp_dbcsr_get_block_p(matrix=matrix_p(1, img)%matrix, &
                                      row=irow, col=icol, BLOCK=p_block, found=found)
            IF(.NOT.(ASSOCIATED(p_block)))CALL cp__a("core_ae.F",232)
            ! *** Decontract density matrix block ***
            DO iset = 1, nseta
               ncoa = npgfa(iset)*ncoset(la_max(iset))
               sgfa = first_sgfa(1, iset)
               DO jset = 1, nsetb
                  ncob = npgfb(jset)*ncoset(lb_max(jset))
                  sgfb = first_sgfb(1, jset)
                  nij = jset+(iset-1)*maxnset
                  IF (iatom <= jatom) THEN
                     CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                                1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                p_block(sgfa, sgfb), SIZE(p_block, 1), &
                                0.0_dp, work(1, 1), SIZE(work, 1))
                  ELSE
                     CALL dgemm("N", "T", ncoa, nsgfb(jset), nsgfa(iset), &
                                1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                p_block(sgfb, sgfa), SIZE(p_block, 1), &
                                0.0_dp, work(1, 1), SIZE(work, 1))
                  END IF
                  CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                             1.0_dp, work(1, 1), SIZE(work, 1), &
                             sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             0.0_dp, pab(1, 1, nij), SIZE(pab, 1))
               END DO
            END DO
         END IF

         ! loop over all kinds for pseudopotential  atoms
         hab = 0._dp
         DO kkind = 1, nkind
            CALL get_qs_kind(qs_kind_set(kkind), all_potential=all_potential)
            IF (.NOT. ASSOCIATED(all_potential)) CYCLE
            CALL get_potential(potential=all_potential, &
                               alpha_core_charge=alpha_c, zeff=zeta_c, &
                               ccore_charge=core_charge, core_charge_radius=core_radius)

            CALL nl_set_sub_iterator(ap_iterator, ikind, kkind, iatom)
            DO WHILE (nl_sub_iterate(ap_iterator) == 0)
               CALL get_iterator_info(ap_iterator, jatom=katom, r=rac)
               dac = SQRT(SUM(rac*rac))
               rbc(:) = rac(:)-rab(:)
               dbc = SQRT(SUM(rbc*rbc))
               IF ((MAXVAL(set_radius_a(:))+core_radius < dac) .OR. &
                   (MAXVAL(set_radius_b(:))+core_radius < dbc)) THEN
                  CYCLE
               END IF

               DO iset = 1, nseta
                  IF (set_radius_a(iset)+core_radius < dac) CYCLE
                  ncoa = npgfa(iset)*ncoset(la_max(iset))
                  sgfa = first_sgfa(1, iset)
                  DO jset = 1, nsetb
                     IF (set_radius_b(jset)+core_radius < dbc) CYCLE
                     ncob = npgfb(jset)*ncoset(lb_max(jset))
                     sgfb = first_sgfb(1, jset)
                     IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE
                     rab2 = dab*dab
                     rac2 = dac*dac
                     rbc2 = dbc*dbc
                     nij = jset+(iset-1)*maxnset
                     ! *** Calculate the GTH pseudo potential forces ***
                     IF (calculate_forces) THEN
                        na_plus = npgfa(iset)*ncoset(la_max(iset)+nder)
                        nb_plus = npgfb(jset)*ncoset(lb_max(jset))
                        ALLOCATE (habd(na_plus, nb_plus))
                        habd = 0._dp
                        CALL verfc( &
                           la_max(iset)+nder, npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                           lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                           alpha_c, core_radius, zeta_c, core_charge, &
                           rab, rab2, rac, rac2, rbc2, hab(:, :, nij), verf, vnuc, ff(0:), &
                           nder, habd)

                        ! *** The derivatives w.r.t. atomic center c are    ***
                        ! *** calculated using the translational invariance ***
                        ! *** of the first derivatives                      ***
                        CALL verfc_force(habd, pab(:, :, nij), force_a, force_b, nder, &
                                         la_max(iset), la_min(iset), npgfa(iset), zeta(:, iset), &
                                         lb_max(jset), lb_min(jset), npgfb(jset), zetb(:, jset), rab)

                        DEALLOCATE (habd)

                        atom_c = atom_of_kind(katom)
                        force(ikind)%all_potential(1, atom_a) = force(ikind)%all_potential(1, atom_a)+f0*force_a(1)
                        force(ikind)%all_potential(2, atom_a) = force(ikind)%all_potential(2, atom_a)+f0*force_a(2)
                        force(ikind)%all_potential(3, atom_a) = force(ikind)%all_potential(3, atom_a)+f0*force_a(3)

                        force(jkind)%all_potential(1, atom_b) = force(jkind)%all_potential(1, atom_b)+f0*force_b(1)
                        force(jkind)%all_potential(2, atom_b) = force(jkind)%all_potential(2, atom_b)+f0*force_b(2)
                        force(jkind)%all_potential(3, atom_b) = force(jkind)%all_potential(3, atom_b)+f0*force_b(3)

                        force(kkind)%all_potential(1, atom_c) = force(kkind)%all_potential(1, atom_c) &
                                                                -f0*force_a(1)-f0*force_b(1)
                        force(kkind)%all_potential(2, atom_c) = force(kkind)%all_potential(2, atom_c) &
                                                                -f0*force_a(2)-f0*force_b(2)
                        force(kkind)%all_potential(3, atom_c) = force(kkind)%all_potential(3, atom_c) &
                                                                -f0*force_a(3)-f0*force_b(3)

                        IF (use_virial) THEN
                           CALL virial_pair_force(virial%pv_virial, f0, force_a, rac)
                           CALL virial_pair_force(virial%pv_virial, f0, force_b, rbc)
                        END IF
                     ELSE
                        CALL verfc( &
                           la_max(iset), npgfa(iset), zeta(:, iset), rpgfa(:, iset), la_min(iset), &
                           lb_max(jset), npgfb(jset), zetb(:, jset), rpgfb(:, jset), lb_min(jset), &
                           alpha_c, core_radius, zeta_c, core_charge, &
                           rab, rab2, rac, rac2, rbc2, hab(:, :, nij), verf, vnuc, ff(0:))
                     END IF
                  END DO
               END DO
            END DO
         END DO
         ! *** Contract nuclear attraction integrals
         DO iset = 1, nseta
            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)
            DO jset = 1, nsetb
               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)
               nij = jset+(iset-1)*maxnset
               CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                          1.0_dp, hab(1, 1, nij), SIZE(hab, 1), &
                          sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                          0.0_dp, work(1, 1), SIZE(work, 1))
               IF (iatom <= jatom) THEN
                  CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             work(1, 1), SIZE(work, 1), &
                             1.0_dp, h_block(sgfa, sgfb), SIZE(h_block, 1))
               ELSE
                  CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                             1.0_dp, work(1, 1), SIZE(work, 1), &
                             sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             1.0_dp, h_block(sgfb, sgfa), SIZE(h_block, 1))
               END IF
            END DO
         END DO

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      CALL neighbor_list_iterator_release(ap_iterator)

      DEALLOCATE (atom_of_kind, basis_set_list)
      DEALLOCATE (hab, work, verf, vnuc, ff)

      IF (calculate_forces) THEN
         DEALLOCATE (pab)
      END IF
      IF (calculate_forces) THEN
         ! *** If LSD, then recover alpha density and beta density     ***
         ! *** from the total density (1) and the spin density (2)     ***
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, nimages
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=0.5_dp, beta_scalar=0.5_dp)
               CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                 alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_core_ae

!==========================================================================================================
! **************************************************************************************************
!> \brief ...
!> \param habd ...
!> \param pab ...
!> \param fa ...
!> \param fb ...
!> \param nder ...
!> \param la_max ...
!> \param la_min ...
!> \param npgfa ...
!> \param zeta ...
!> \param lb_max ...
!> \param lb_min ...
!> \param npgfb ...
!> \param zetb ...
!> \param rab ...
! **************************************************************************************************
   SUBROUTINE verfc_force(habd, pab, fa, fb, nder, la_max, la_min, npgfa, zeta, lb_max, lb_min, npgfb, zetb, rab)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: habd, pab
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: fa, fb
      INTEGER, INTENT(IN)                                :: nder, la_max, la_min, npgfa
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zeta
      INTEGER, INTENT(IN)                                :: lb_max, lb_min, npgfb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: zetb
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab

      CHARACTER(LEN=*), PARAMETER :: routineN = 'verfc_force', routineP = moduleN//':'//routineN

      INTEGER                                            :: ic_a, ic_b, icam1, icam2, icam3, icap1, &
                                                            icap2, icap3, icax, icbm1, icbm2, &
                                                            icbm3, icbx, icoa, icob, ipgfa, ipgfb, &
                                                            na, nap, nb
      INTEGER, DIMENSION(3)                              :: la, lb
      REAL(KIND=dp)                                      :: zax2, zbx2

      fa = 0.0_dp
      fb = 0.0_dp

      na = ncoset(la_max)
      nap = ncoset(la_max+nder)
      nb = ncoset(lb_max)
      DO ipgfa = 1, npgfa
         zax2 = zeta(ipgfa)*2.0_dp
         DO ipgfb = 1, npgfb
            zbx2 = zetb(ipgfb)*2.0_dp
            DO ic_a = ncoset(la_min-1)+1, ncoset(la_max)
               la(1:3) = indco(1:3, ic_a)
               icap1 = coset(la(1)+1, la(2), la(3))
               icap2 = coset(la(1), la(2)+1, la(3))
               icap3 = coset(la(1), la(2), la(3)+1)
               icam1 = coset(la(1)-1, la(2), la(3))
               icam2 = coset(la(1), la(2)-1, la(3))
               icam3 = coset(la(1), la(2), la(3)-1)
               icoa = ic_a+(ipgfa-1)*na
               icax = (ipgfa-1)*nap

               DO ic_b = ncoset(lb_min-1)+1, ncoset(lb_max)
                  lb(1:3) = indco(1:3, ic_b)
                  icbm1 = coset(lb(1)-1, lb(2), lb(3))
                  icbm2 = coset(lb(1), lb(2)-1, lb(3))
                  icbm3 = coset(lb(1), lb(2), lb(3)-1)
                  icob = ic_b+(ipgfb-1)*nb
                  icbx = (ipgfb-1)*nb

                  fa(1) = fa(1)-pab(icoa, icob)*(-zax2*habd(icap1+icax, icob)+ &
                                                 REAL(la(1), KIND=dp)*habd(icam1+icax, icob))
                  fa(2) = fa(2)-pab(icoa, icob)*(-zax2*habd(icap2+icax, icob)+ &
                                                 REAL(la(2), KIND=dp)*habd(icam2+icax, icob))
                  fa(3) = fa(3)-pab(icoa, icob)*(-zax2*habd(icap3+icax, icob)+ &
                                                 REAL(la(3), KIND=dp)*habd(icam3+icax, icob))

                  fb(1) = fb(1)-pab(icoa, icob)*( &
                          -zbx2*(habd(icap1+icax, icob)-rab(1)*habd(ic_a+icax, icob))+ &
                          REAL(lb(1), KIND=dp)*habd(ic_a+icax, icbm1+icbx))
                  fb(2) = fb(2)-pab(icoa, icob)*( &
                          -zbx2*(habd(icap2+icax, icob)-rab(2)*habd(ic_a+icax, icob))+ &
                          REAL(lb(2), KIND=dp)*habd(ic_a+icax, icbm2+icbx))
                  fb(3) = fb(3)-pab(icoa, icob)*( &
                          -zbx2*(habd(icap3+icax, icob)-rab(3)*habd(ic_a+icax, icob))+ &
                          REAL(lb(3), KIND=dp)*habd(ic_a+icax, icbm3+icbx))

               END DO ! ic_b
            END DO ! ic_a
         END DO ! ipgfb
      END DO ! ipgfa

   END SUBROUTINE verfc_force

!==========================================================================================================

END MODULE core_ae
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Calculation of the local potential contribution of the nonadditive kinetic energy
!>         <a|V(local)|b> = <a|Sum e^a*rc**2|b>
!> \par History
!>      - adapted from core_ppl
! **************************************************************************************************
MODULE kg_tnadd_mat
   USE ai_overlap_ppl,                  ONLY: ppl_integral
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_finalize, cp_dbcsr_get_block_p, cp_dbcsr_init, &
        cp_dbcsr_p_type, cp_dbcsr_set, dbcsr_distribution_obj, dbcsr_type_symmetric
   USE external_potential_types,        ONLY: get_potential,&
                                              local_potential_type
   USE kg_environment_types,            ONLY: kg_environment_type
   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: init_orbital_pointers,&
                                              ncoset
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type,&
                                              nl_set_sub_iterator,&
                                              nl_sub_iterate
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'kg_tnadd_mat'

   PUBLIC :: build_tnadd_mat

CONTAINS

!==========================================================================================================

! **************************************************************************************************
!> \brief ...
!> \param kg_env ...
!> \param matrix_p ...
!> \param force ...
!> \param virial ...
!> \param calculate_forces ...
!> \param use_virial ...
!> \param qs_kind_set ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param sab_orb ...
!> \param dbcsr_dist ...
! **************************************************************************************************
   SUBROUTINE build_tnadd_mat(kg_env, matrix_p, force, virial, calculate_forces, use_virial, &
                              qs_kind_set, atomic_kind_set, particle_set, sab_orb, dbcsr_dist)

      TYPE(kg_environment_type), POINTER                 :: kg_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_p
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(virial_type), POINTER                         :: virial
      LOGICAL, INTENT(IN)                                :: calculate_forces
      LOGICAL                                            :: use_virial
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_tnadd_mat', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nexp_max = 10

      INTEGER :: atom_a, atom_b, atom_c, handle, i, iatom, icol, ikind, img, imol, inode, irow, &
         iset, jatom, jkind, jmol, jset, katom, kkind, kmol, last_iatom, last_jatom, ldai, ldsab, &
         maxco, maxder, maxl, maxlgto, maxnset, maxpol, maxsgf, mepos, natom, ncoa, ncob, nder, &
         ngau, nkind, npol, nseta, nsetb, nthread, sgfa, sgfb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, DIMENSION(:), POINTER                     :: atom_to_molecule, col_blk_sizes, la_max, &
                                                            la_min, lb_max, lb_min, npgfa, npgfb, &
                                                            nsgfa, nsgfb, row_blk_sizes
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      INTEGER, DIMENSION(nexp_max)                       :: nct
      LOGICAL                                            :: found, new_atom_pair
      REAL(KIND=dp)                                      :: dab, dac, dbc, f0, radius
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: ppl_fwork, ppl_work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: hab, pab
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, rab, rac, rbc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha, set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ccval, cval, h_block, p_block, rpgfa, &
                                                            rpgfb, sphi_a, sphi_b, zeta, zetb
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_kg
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(local_potential_type), POINTER                :: tnadd_potential
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: ap_iterator, nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sac_kin

      IF (calculate_forces) THEN
         CALL timeset(routineN//"_forces", handle)
      ELSE
         CALL timeset(routineN, handle)
      ENDIF

      NULLIFY (matrix_kg)
      IF (ASSOCIATED(kg_env%tnadd_mat)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(kg_env%tnadd_mat)
      ENDIF
      sac_kin => kg_env%sac_kin
      atom_to_molecule => kg_env%atom_to_molecule

      nkind = SIZE(atomic_kind_set)
      natom = SIZE(particle_set)

      ALLOCATE (atom_of_kind(natom))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)

      IF (calculate_forces) THEN
         nder = 1
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, SIZE(matrix_p, 2)
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
               CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                 alpha_scalar=-2.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      ELSE
         nder = 0
      END IF

      maxder = ncoset(nder)

      CALL get_qs_kind_set(qs_kind_set, maxpol=maxpol, maxco=maxco, maxlgto=maxlgto, &
                           maxsgf=maxsgf, maxnset=maxnset)

      maxl = MAX(maxlgto, maxpol)
      CALL init_orbital_pointers(maxl+nder+1)

      ldsab = MAX(maxco, ncoset(maxpol), maxsgf, maxpol)
      ldai = ncoset(maxl+nder+1)

      ALLOCATE (basis_set_list(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=basis_set_a)
         IF (ASSOCIATED(basis_set_a)) THEN
            basis_set_list(ikind)%gto_basis_set => basis_set_a
         ELSE
            NULLIFY (basis_set_list(ikind)%gto_basis_set)
         END IF
      END DO

      ! build the matrix
      ALLOCATE (row_blk_sizes(natom), col_blk_sizes(natom))

      CALL get_particle_set(particle_set, qs_kind_set, nsgf=row_blk_sizes)
      CALL get_particle_set(particle_set, qs_kind_set, nsgf=col_blk_sizes)

      CALL cp_dbcsr_allocate_matrix_set(matrix_kg, 1)

      ALLOCATE (matrix_kg(1)%matrix)
      CALL cp_dbcsr_init(matrix_kg(1)%matrix)
      CALL cp_dbcsr_create(matrix=matrix_kg(1)%matrix, &
                           name="Nonadditive kinetic energy potential", &
                           dist=dbcsr_dist, matrix_type=dbcsr_type_symmetric, &
                           row_blk_size=row_blk_sizes, col_blk_size=col_blk_sizes, &
                           nze=0, reuse_arrays=.TRUE.)
      CALL cp_dbcsr_alloc_block_from_nbl(matrix_kg(1)%matrix, sab_orb)
      CALL cp_dbcsr_set(matrix_kg(1)%matrix, 0.0_dp)

      nthread = 1
!$    nthread = omp_get_max_threads()

      CALL neighbor_list_iterator_create(nl_iterator, sab_orb, nthread=nthread)

!$OMP PARALLEL &
!$OMP DEFAULT (NONE) &
!$OMP SHARED  (nl_iterator, basis_set_list, calculate_forces, force, use_virial,&
!$OMP          matrix_kg, matrix_p,virial, atomic_kind_set, qs_kind_set, particle_set,&
!$OMP          sab_orb, sac_kin, nthread, ncoset, nkind,&
!$OMP          atom_of_kind, ldsab,  maxnset, maxder, &
!$OMP          maxlgto, nder, maxco, atom_to_molecule) &
!$OMP PRIVATE (ikind, jkind, inode, iatom, jatom, rab, basis_set_a, basis_set_b, atom_a, &
!$OMP          atom_b, first_sgfa, la_max, la_min, npgfa, nsgfa, sphi_a, &
!$OMP          zeta, first_sgfb, lb_max, lb_min, npgfb, nsetb, rpgfb, set_radius_b, sphi_b, &
!$OMP          zetb, last_iatom, last_jatom, new_atom_pair, dab, irow, icol, h_block, found, iset, ncoa, &
!$OMP          sgfa, jset, ncob, sgfb, nsgfb, p_block, work, pab, hab, kkind, nseta, &
!$OMP          radius, alpha, nct, imol, jmol, kmol,&
!$OMP          npol, ngau, cval, ccval, ap_iterator, rac, dac, rbc, dbc, &
!$OMP          set_radius_a,  rpgfa, force_a, force_b, ppl_fwork, mepos, &
!$OMP          f0, katom, ppl_work, atom_c,&
!$OMP          ldai,tnadd_potential)

      mepos = 0
!$    mepos = omp_get_thread_num()

      ! iterator for basis/potential list
      CALL neighbor_list_iterator_create(ap_iterator, sac_kin, search=.TRUE.)

      ALLOCATE (hab(ldsab, ldsab, maxnset, maxnset), work(ldsab, ldsab*maxder))
      ldai = ncoset(2*maxlgto+2*nder)
      ALLOCATE (ppl_work(ldai, ldai, MAX(maxder, 2*maxlgto+2*nder+1)))
      IF (calculate_forces) THEN
         ALLOCATE (pab(maxco, maxco, maxnset, maxnset))
         ldai = ncoset(maxlgto)
         ALLOCATE (ppl_fwork(ldai, ldai, maxder))
      END IF

      last_iatom = 0
      last_jatom = 0
      DO WHILE (neighbor_list_iterate(nl_iterator, mepos=mepos) == 0)

         CALL get_iterator_info(nl_iterator, mepos=mepos, ikind=ikind, jkind=jkind, inode=inode, &
                                iatom=iatom, jatom=jatom, r=rab)

         basis_set_a => basis_set_list(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE

         atom_a = atom_of_kind(iatom)
         atom_b = atom_of_kind(jatom)
         imol = atom_to_molecule(iatom)
         jmol = atom_to_molecule(jatom)
         IF(.NOT.(imol == jmol))CALL cp__a("kg_tnadd_mat.F",249)

         ! basis ikind
         first_sgfa => basis_set_a%first_sgf
         la_max => basis_set_a%lmax
         la_min => basis_set_a%lmin
         npgfa => basis_set_a%npgf
         nseta = basis_set_a%nset
         nsgfa => basis_set_a%nsgf_set
         rpgfa => basis_set_a%pgf_radius
         set_radius_a => basis_set_a%set_radius
         sphi_a => basis_set_a%sphi
         zeta => basis_set_a%zet
         ! basis jkind
         first_sgfb => basis_set_b%first_sgf
         lb_max => basis_set_b%lmax
         lb_min => basis_set_b%lmin
         npgfb => basis_set_b%npgf
         nsetb = basis_set_b%nset
         nsgfb => basis_set_b%nsgf_set
         rpgfb => basis_set_b%pgf_radius
         set_radius_b => basis_set_b%set_radius
         sphi_b => basis_set_b%sphi
         zetb => basis_set_b%zet

         dab = SQRT(SUM(rab*rab))

         IF (iatom == last_iatom .AND. jatom == last_jatom) THEN
            new_atom_pair = .FALSE.
         ELSE
            new_atom_pair = .TRUE.
            last_jatom = jatom
            last_iatom = iatom
         END IF

         ! *** Use the symmetry of the first derivatives ***
         IF (iatom == jatom) THEN
            f0 = 1.0_dp
         ELSE
            f0 = 2.0_dp
         END IF

         ! *** Create matrix blocks for a new matrix block column ***
         IF (new_atom_pair) THEN
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF
            NULLIFY (h_block)
            CALL cp_dbcsr_get_block_p(matrix_kg(1)%matrix, irow, icol, h_block, found)
            IF (ASSOCIATED(h_block)) THEN
            IF (calculate_forces) THEN
               IF(.NOT.(SIZE(matrix_p, 2) == 1))CALL cp__a("kg_tnadd_mat.F",304)
               NULLIFY (p_block)
               CALL cp_dbcsr_get_block_p(matrix_p(1, 1)%matrix, irow, icol, p_block, found)
               IF (ASSOCIATED(p_block)) THEN
                  DO iset = 1, nseta
                     ncoa = npgfa(iset)*ncoset(la_max(iset))
                     sgfa = first_sgfa(1, iset)
                     DO jset = 1, nsetb
                        ncob = npgfb(jset)*ncoset(lb_max(jset))
                        sgfb = first_sgfb(1, jset)

                        ! *** Decontract density matrix block ***
                        IF (iatom <= jatom) THEN
                           CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                                      1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                      p_block(sgfa, sgfb), SIZE(p_block, 1), &
                                      0.0_dp, work(1, 1), SIZE(work, 1))
                        ELSE
                           CALL dgemm("N", "T", ncoa, nsgfb(jset), nsgfa(iset), &
                                      1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                      p_block(sgfb, sgfa), SIZE(p_block, 1), &
                                      0.0_dp, work(1, 1), SIZE(work, 1))
                        END IF

                        CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                                   1.0_dp, work(1, 1), SIZE(work, 1), &
                                   sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                                   0.0_dp, pab(1, 1, iset, jset), SIZE(pab, 1))
                     END DO
                  END DO
               ENDIF
            END IF
            END IF
         END IF

         hab = 0._dp

         ! loop over all kinds for nonadditive kinetic energy potential atoms
         DO kkind = 1, nkind

            CALL get_qs_kind(qs_kind_set(kkind), tnadd_potential=tnadd_potential)
            IF (.NOT. ASSOCIATED(tnadd_potential)) CYCLE
            CALL get_potential(potential=tnadd_potential, &
                               alpha=alpha, cval=cval, ngau=ngau, npol=npol, radius=radius)
            nct = npol
            ALLOCATE (ccval(npol, ngau))
            ccval(1:npol, 1:ngau) = TRANSPOSE(cval(1:ngau, 1:npol))

            CALL nl_set_sub_iterator(ap_iterator, ikind, kkind, iatom)

            DO WHILE (nl_sub_iterate(ap_iterator) == 0)

               CALL get_iterator_info(ap_iterator, jatom=katom, r=rac)

               ! this potential only acts on other moleclules
               kmol = atom_to_molecule(katom)
               IF (kmol == imol) CYCLE

               dac = SQRT(SUM(rac*rac))
               rbc(:) = rac(:)-rab(:)
               dbc = SQRT(SUM(rbc*rbc))
               IF ((MAXVAL(set_radius_a(:))+radius < dac) .OR. &
                   (MAXVAL(set_radius_b(:))+radius < dbc)) THEN
                  CYCLE
               END IF

               DO iset = 1, nseta
                  IF (set_radius_a(iset)+radius < dac) CYCLE
                  ncoa = npgfa(iset)*ncoset(la_max(iset))
                  sgfa = first_sgfa(1, iset)
                  DO jset = 1, nsetb
                     IF (set_radius_b(jset)+radius < dbc) CYCLE
                     ncob = npgfb(jset)*ncoset(lb_max(jset))
                     sgfb = first_sgfb(1, jset)
                     IF (set_radius_a(iset)+set_radius_b(jset) < dab) CYCLE
                     ! *** Calculate the GTH pseudo potential forces ***
                     IF (calculate_forces) THEN

                        CALL ppl_integral( &
                           la_max(iset), la_min(iset), npgfa(iset), &
                           rpgfa(:, iset), zeta(:, iset), &
                           lb_max(jset), lb_min(jset), npgfb(jset), &
                           rpgfb(:, jset), zetb(:, jset), &
                           ngau, alpha, nct, ccval, radius, &
                           rab, dab, rac, dac, rbc, dbc, &
                           hab(:, :, iset, jset), ppl_work, pab(:, :, iset, jset), &
                           force_a, force_b, ppl_fwork)
                        ! *** The derivatives w.r.t. atomic center c are    ***
                        ! *** calculated using the translational invariance ***
                        ! *** of the first derivatives                      ***
                        atom_c = atom_of_kind(katom)

!$OMP CRITICAL(force_critical)
                        force(ikind)%kinetic(1, atom_a) = force(ikind)%kinetic(1, atom_a)+f0*force_a(1)
                        force(ikind)%kinetic(2, atom_a) = force(ikind)%kinetic(2, atom_a)+f0*force_a(2)
                        force(ikind)%kinetic(3, atom_a) = force(ikind)%kinetic(3, atom_a)+f0*force_a(3)
                        force(kkind)%kinetic(1, atom_c) = force(kkind)%kinetic(1, atom_c)-f0*force_a(1)
                        force(kkind)%kinetic(2, atom_c) = force(kkind)%kinetic(2, atom_c)-f0*force_a(2)
                        force(kkind)%kinetic(3, atom_c) = force(kkind)%kinetic(3, atom_c)-f0*force_a(3)

                        force(jkind)%kinetic(1, atom_b) = force(jkind)%kinetic(1, atom_b)+f0*force_b(1)
                        force(jkind)%kinetic(2, atom_b) = force(jkind)%kinetic(2, atom_b)+f0*force_b(2)
                        force(jkind)%kinetic(3, atom_b) = force(jkind)%kinetic(3, atom_b)+f0*force_b(3)
                        force(kkind)%kinetic(1, atom_c) = force(kkind)%kinetic(1, atom_c)-f0*force_b(1)
                        force(kkind)%kinetic(2, atom_c) = force(kkind)%kinetic(2, atom_c)-f0*force_b(2)
                        force(kkind)%kinetic(3, atom_c) = force(kkind)%kinetic(3, atom_c)-f0*force_b(3)

                        IF (use_virial) THEN
                           CALL virial_pair_force(virial%pv_virial, f0, force_a, rac)
                           CALL virial_pair_force(virial%pv_virial, f0, force_b, rbc)
                        END IF
!$OMP END CRITICAL(force_critical)

                     ELSE
                        CALL ppl_integral( &
                           la_max(iset), la_min(iset), npgfa(iset), &
                           rpgfa(:, iset), zeta(:, iset), &
                           lb_max(jset), lb_min(jset), npgfb(jset), &
                           rpgfb(:, jset), zetb(:, jset), &
                           ngau, alpha, nct, ccval, radius, &
                           rab, dab, rac, dac, rbc, dbc, hab(:, :, iset, jset), ppl_work)
                     END IF
                  END DO
               END DO
            END DO
            DEALLOCATE (ccval)
         END DO

         ! *** Contract integrals
         DO iset = 1, nseta
            ncoa = npgfa(iset)*ncoset(la_max(iset))
            sgfa = first_sgfa(1, iset)
            DO jset = 1, nsetb
               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)

               CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                          1.0_dp, hab(1, 1, iset, jset), SIZE(hab, 1), &
                          sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                          0.0_dp, work(1, 1), SIZE(work, 1))

!$OMP CRITICAL(h_block_critical)
               IF (iatom <= jatom) THEN
                  CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             work(1, 1), SIZE(work, 1), &
                             1.0_dp, h_block(sgfa, sgfb), SIZE(h_block, 1))
               ELSE
                  CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncoa, &
                             1.0_dp, work(1, 1), SIZE(work, 1), &
                             sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             1.0_dp, h_block(sgfb, sgfa), SIZE(h_block, 1))
               END IF
!$OMP END CRITICAL(h_block_critical)

            END DO
         END DO
      END DO

      CALL neighbor_list_iterator_release(ap_iterator)

      DEALLOCATE (hab, work, ppl_work)

      IF (calculate_forces) THEN
         DEALLOCATE (pab, ppl_fwork)
      END IF

!$OMP END PARALLEL

      CALL neighbor_list_iterator_release(nl_iterator)

      DO i = 1, SIZE(matrix_kg)
         CALL cp_dbcsr_finalize(matrix_kg(i)%matrix)
      ENDDO
      kg_env%tnadd_mat => matrix_kg

      DEALLOCATE (atom_of_kind, basis_set_list)

      IF (calculate_forces) THEN
         ! *** If LSD, then recover alpha density and beta density     ***
         ! *** from the total density (1) and the spin density (2)     ***
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, SIZE(matrix_p, 2)
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=0.5_dp, beta_scalar=0.5_dp)
               CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                 alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_tnadd_mat

!==========================================================================================================

END MODULE kg_tnadd_mat
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the core Hamiltonian integral matrix <a|H|b> over
!>      Cartesian Gaussian-type functions.
!>
!>      <a|H|b> = <a|T|b> + <a|V|b>
!>
!>      Kinetic energy:
!>
!>      <a|T|b> = <a|-nabla**2/2|b>
!>                \_______________/
!>                        |
!>                     kinetic
!>
!>      Nuclear potential energy:
!>
!>      a) Allelectron calculation:
!>
!>                          erfc(r)
!>         <a|V|b> = -Z*<a|---------|b>
!>                             r
!>
!>                          1 - erf(r)
!>                 = -Z*<a|------------|b>
!>                              r
!>
!>                           1           erf(r)
!>                 = -Z*(<a|---|b> - <a|--------|b>)
!>                           r             r
!>
!>                           1
!>                 = -Z*(<a|---|b> - N*<ab||c>)
!>                           r
!>
!>                      -Z
!>                 = <a|---|b> + Z*N*<ab||c>
!>                       r
!>                   \_______/       \_____/
!>                       |              |
!>                    nuclear        coulomb
!>
!>      b) Pseudopotential calculation (Goedecker, Teter and Hutter; GTH):
!>
!>         <a|V|b> = <a|(V(local) + V(non-local))|b>
!>
!>                 = <a|(V(local)|b> + <a|V(non-local))|b>
!>
!>         <a|V(local)|b> = <a|-Z(eff)*erf(SQRT(2)*alpha*r)/r +
!>                             (C1 + C2*(alpha*r)**2 + C3*(alpha*r)**4 +
!>                              C4*(alpha*r)**6)*exp(-(alpha*r)**2/2))|b>
!>
!>         <a|V(non-local)|b> = <a|p(l,i)>*h(i,j)*<p(l,j)|b>
!> \par Literature
!>      S. Goedecker, M. Teter and J. Hutter, Phys. Rev. B 54, 1703 (1996)
!>      C. Hartwigsen, S. Goedecker and J. Hutter, Phys. Rev. B 58, 3641 (1998)
!>      M. Krack and M. Parrinello, Phys. Chem. Chem. Phys. 2, 2105 (2000)
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      - Joost VandeVondele (April 2003) : added LSD forces
!>      - Non-redundant calculation of the non-local part of the GTH PP
!>        (22.05.2003,MK)
!>      - New parallelization scheme (27.06.2003,MK)
!>      - OpenMP version (07.12.2003,JGH)
!>      - Binary search loop for VPPNL operators (09.01.2004,JGH,MK)
!>      - Refactoring of pseudopotential and nuclear attraction integrals (25.02.2009,JGH)
!>      - General refactoring (01.10.2010,JGH)
!>      - Refactoring related to the new kinetic energy and overlap routines (07.2014,JGH)
!>      - k-point functionality (07.2015,JGH)
!> \author Matthias Krack (14.09.2000,21.03.02)
! **************************************************************************************************
MODULE qs_core_hamiltonian
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE core_ae,                         ONLY: build_core_ae
   USE core_ppl,                        ONLY: build_core_ppl
   USE core_ppnl,                       ONLY: build_core_ppnl
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_init, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_p_type, cp_dbcsr_type, dbcsr_distribution_obj
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_matrix_dist,&
                                              cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: do_ppl_analytic,&
                                              kg_tnadd_atomic,&
                                              rel_none,&
                                              rel_trans_atom
   USE input_section_types,             ONLY: section_vals_val_get
   USE kg_environment_types,            ONLY: kg_environment_type
   USE kg_tnadd_mat,                    ONLY: build_tnadd_mat
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE lri_environment_methods,         ONLY: build_lri_matrices
   USE lri_environment_types,           ONLY: lri_environment_type
   USE orbital_pointers,                ONLY: ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_kinetic,                      ONLY: build_kinetic_matrix
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_oce_methods,                  ONLY: build_oce_matrices
   USE qs_oce_types,                    ONLY: allocate_oce_set,&
                                              create_oce_set,&
                                              oce_matrix_type
   USE qs_overlap,                      ONLY: build_overlap_matrix
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_core_hamiltonian'

   PUBLIC :: build_core_hamiltonian_matrix
   PUBLIC :: dump_info_core_hamiltonian

CONTAINS

! **************************************************************************************************
!> \brief Cosntruction of the QS Core Hamiltonian Matrix
!> \param qs_env ...
!> \param calculate_forces ...
!> \author Creation (11.03.2002,MK)
!>      Non-redundant calculation of the non-local part of the GTH PP (22.05.2003,MK)
!>      New parallelization scheme (27.06.2003,MK)
! **************************************************************************************************
   SUBROUTINE build_core_hamiltonian_matrix(qs_env, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_core_hamiltonian_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, img, maxder, natom, nder, nders, &
                                                            nimages, nkind
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: all_potential_present, kgpot, lrigpw, &
                                                            ofdft, ppl_present, use_virial
      REAL(KIND=dp)                                      :: eps_filter, eps_fit, eps_ppnl
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s_aux_fit, matrix_s_aux_fit_vs_orb
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_h, matrix_p, matrix_s, matrix_t, &
                                                            matrix_w
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kg_environment_type), POINTER                 :: kg_env
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_aux_fit, sab_aux_fit_vs_orb, &
                                                            sab_orb, sac_ae, sac_ppl, sap_oce, &
                                                            sap_ppnl
      TYPE(oce_matrix_type), POINTER                     :: oce
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(virial_type), POINTER                         :: virial

      IF (calculate_forces) THEN
         CALL timeset(routineN//"_forces", handle)
      ELSE
         CALL timeset(routineN, handle)
      ENDIF

      NULLIFY (logger)
      logger => cp_get_default_logger()

      NULLIFY (atomic_kind_set)
      NULLIFY (qs_kind_set)
      NULLIFY (particle_set)
      NULLIFY (para_env)
      NULLIFY (ks_env)
      NULLIFY (dft_control)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      virial=virial, &
                      para_env=para_env, &
                      ks_env=ks_env, &
                      dft_control=dft_control)

      ! is this a orbital-free method calculation
      ofdft = dft_control%qs_control%ofgpw
      ! is this a LRI method calculation
      lrigpw = dft_control%qs_control%lrigpw
      ! is this a KG method calculation using atomic TNADD potentials
      IF (dft_control%qs_control%do_kg) THEN
         kgpot = (qs_env%kg_env%tnadd_method == kg_tnadd_atomic)
      ELSE
         kgpot = .FALSE.
      END IF

      nkind = SIZE(atomic_kind_set)
      natom = SIZE(particle_set)
      nimages = dft_control%nimages
      IF (ofdft) THEN
         IF(.NOT.(nimages == 1))CALL cp__a("qs_core_hamiltonian.F",229)
      END IF

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      NULLIFY (rho, force, matrix_p, matrix_w)
      nders = 0
      IF (calculate_forces) THEN
         nder = 1
         CALL get_qs_env(qs_env=qs_env, force=force, matrix_w_kp=matrix_w)
         CALL get_qs_env(qs_env=qs_env, rho=rho)
         CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
         !     *** If LSD, then combine alpha density and beta density to
         !     *** total density: alpha <- alpha + beta   and
         !     *** spin density:   beta <- alpha - beta
         !     (since all things can be computed based on the sum of these matrices anyway)
         !     (matrix_p is restored at the end of the run, matrix_w is left in its modified state
         !     (as it should not be needed afterwards)
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, nimages
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
               CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                 alpha_scalar=-2.0_dp, beta_scalar=1.0_dp)
               CALL cp_dbcsr_add(matrix_w(1, img)%matrix, matrix_w(2, img)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      ELSE
         IF (cp_print_key_should_output(logger%iter_info, qs_env%input, &
                                        "DFT%PRINT%AO_MATRICES/DERIVATIVES") /= 0) THEN
            nder = 1
         ELSE
            nder = 0
         END IF
      END IF

      IF ((cp_print_key_should_output(logger%iter_info, qs_env%input, &
                                      "DFT%PRINT%AO_MATRICES/OVERLAP") /= 0 .AND. &
           BTEST(cp_print_key_should_output(logger%iter_info, qs_env%input, &
                                            "DFT%PRINT%AO_MATRICES/DERIVATIVES"), cp_p_file))) THEN
         nders = 1
      END IF

      ! the delta pulse in the periodic case needs the momentum operator,
      ! which is equivalent to the derivative of the overlap matrix
      IF (ASSOCIATED(dft_control%rtp_control)) THEN
         IF (dft_control%rtp_control%apply_delta_pulse .AND. &
             dft_control%rtp_control%periodic) THEN
            nders = 1
         ENDIF
      ENDIF

      ! filter for new matrices
      eps_filter = dft_control%qs_control%eps_filter_matrix

      maxder = ncoset(nder)

      NULLIFY (matrix_s, matrix_t)
      CALL get_qs_env(qs_env=qs_env, kinetic_kp=matrix_t, matrix_s_kp=matrix_s)
      NULLIFY (sab_orb)
      CALL get_qs_env(qs_env=qs_env, sab_orb=sab_orb)
      IF (calculate_forces) THEN
         ! S matrix
         CALL build_overlap_matrix(ks_env, nderivative=nders, matrixkp_s=matrix_s, &
                                   matrix_name="OVERLAP MATRIX", &
                                   basis_type_a="ORB", &
                                   basis_type_b="ORB", &
                                   sab_nl=sab_orb, calculate_forces=.TRUE., &
                                   matrixkp_p=matrix_w, qs_input=qs_env%input)
         ! T matrix
         IF (.NOT. ofdft) &
            CALL build_kinetic_matrix(ks_env, matrixkp_t=matrix_t, &
                                      matrix_name="KINETIC ENERGY MATRIX", &
                                      basis_type="ORB", &
                                      sab_nl=sab_orb, calculate_forces=.TRUE., &
                                      matrixkp_p=matrix_p, &
                                      eps_filter=eps_filter)
      ELSE
         ! S matrix
         CALL build_overlap_matrix(ks_env, nderivative=nders, matrixkp_s=matrix_s, &
                                   matrix_name="OVERLAP MATRIX", &
                                   basis_type_a="ORB", &
                                   basis_type_b="ORB", &
                                   sab_nl=sab_orb, qs_input=qs_env%input)
         ! T matrix
         IF (.NOT. ofdft) &
            CALL build_kinetic_matrix(ks_env, matrixkp_t=matrix_t, &
                                      matrix_name="KINETIC ENERGY MATRIX", &
                                      basis_type="ORB", &
                                      sab_nl=sab_orb, &
                                      eps_filter=eps_filter)

         IF (dft_control%do_admm) THEN
            NULLIFY (matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, &
                     sab_aux_fit, sab_aux_fit_vs_orb)
            CALL get_qs_env(qs_env=qs_env, matrix_s_aux_fit=matrix_s_aux_fit, &
                            sab_aux_fit=sab_aux_fit)
            CALL build_overlap_matrix(ks_env, matrix_s=matrix_s_aux_fit, &
                                      matrix_name="AUX_FIT_OVERLAP", &
                                      basis_type_a="AUX_FIT", &
                                      basis_type_b="AUX_FIT", &
                                      sab_nl=sab_aux_fit, qs_input=qs_env%input)
            CALL set_ks_env(ks_env, matrix_s_aux_fit=matrix_s_aux_fit)
            CALL get_qs_env(qs_env=qs_env, matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb, &
                            sab_aux_fit_vs_orb=sab_aux_fit_vs_orb)
            CALL build_overlap_matrix(ks_env, matrix_s=matrix_s_aux_fit_vs_orb, &
                                      matrix_name="MIXED_OVERLAP", &
                                      basis_type_a="AUX_FIT", &
                                      basis_type_b="ORB", &
                                      sab_nl=sab_aux_fit_vs_orb, qs_input=qs_env%input)
            CALL set_ks_env(ks_env, matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb)
         END IF
      END IF

      ! initialize H matrix
      NULLIFY (matrix_h)
      CALL get_qs_env(qs_env=qs_env, matrix_h_kp=matrix_h)
      CALL cp_dbcsr_allocate_matrix_set(matrix_h, 1, nimages)
      DO img = 1, nimages
         ALLOCATE (matrix_h(1, img)%matrix)
         CALL cp_dbcsr_init(matrix_h(1, img)%matrix)
         CALL cp_dbcsr_create(matrix_h(1, img)%matrix, template=matrix_s(1, 1)%matrix)
         CALL cp_dbcsr_alloc_block_from_nbl(matrix_h(1, img)%matrix, sab_orb)
         IF (.NOT. ofdft) THEN
            CALL cp_dbcsr_copy(matrix_h(1, img)%matrix, matrix_t(1, img)%matrix, &
                               keep_sparsity=.TRUE., name="CORE HAMILTONIAN MATRIX")
         END IF
      END DO
      IF (.NOT. ofdft) THEN
         ! relativistic atomic correction to kinetic energy
         IF (qs_env%rel_control%rel_method /= rel_none) THEN
            IF(.NOT.(nimages == 1))CALL cp__a("qs_core_hamiltonian.F",361)
            IF (qs_env%rel_control%rel_transformation == rel_trans_atom) THEN
               CALL build_atomic_relmat(matrix_h(1, 1)%matrix, &
                                        atomic_kind_set, qs_kind_set, particle_set)
            END IF
         END IF
      END IF

      !   *** Allocate the matrix of coefficients for one center expansions
      NULLIFY (oce)
      IF (dft_control%qs_control%gapw .OR. dft_control%qs_control%gapw_xc) THEN
         CALL get_qs_env(qs_env=qs_env, oce=oce)
         CALL create_oce_set(oce)
         CALL allocate_oce_set(oce, nkind)
         ! force analytic ppl calcuation for GAPW methods
         dft_control%qs_control%do_ppl_method = do_ppl_analytic
      ENDIF

      NULLIFY (sac_ae, sac_ppl, sap_ppnl, sap_oce)
      CALL get_qs_env(qs_env=qs_env, &
                      sac_ae=sac_ae, &
                      sac_ppl=sac_ppl, &
                      sap_ppnl=sap_ppnl, &
                      sap_oce=sap_oce)
      CALL get_qs_kind_set(qs_kind_set, &
                           all_potential_present=all_potential_present)
      ppl_present = ASSOCIATED(sac_ppl)

      IF (calculate_forces) THEN
         ! *** If LSD, then recover alpha density and beta density     ***
         ! *** from the total density (1) and the spin density (2)     ***
         ! *** The W matrix is neglected, since it will be destroyed   ***
         ! *** in the calling force routine after leaving this routine ***
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, nimages
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=0.5_dp, beta_scalar=0.5_dp)
               CALL cp_dbcsr_add(matrix_p(2, img)%matrix, matrix_p(1, img)%matrix, &
                                 alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      END IF

      ! prepare for k-points
      NULLIFY (cell_to_index)
      IF (nimages > 1) THEN
         CALL get_ks_env(ks_env=ks_env, kpoints=kpoints)
         CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
      END IF

      ! *** compute the ppl contribution to the core hamiltonian ***
      IF (ppl_present) THEN
         IF (dft_control%qs_control%do_ppl_method == do_ppl_analytic) THEN
            CALL build_core_ppl(matrix_h, matrix_p, force, virial, calculate_forces, use_virial, nder, &
                                qs_kind_set, atomic_kind_set, particle_set, sab_orb, sac_ppl, nimages, cell_to_index)
         END IF
      END IF

      ! *** compute the nuclear attraction contribution to the core hamiltonian ***
      IF (all_potential_present) THEN
         CALL build_core_ae(matrix_h, matrix_p, force, virial, calculate_forces, use_virial, nder, &
                            qs_kind_set, atomic_kind_set, particle_set, sab_orb, sac_ae, nimages, cell_to_index)
      END IF

      ! *** compute the ppnl contribution to the core hamiltonian ***
      eps_ppnl = dft_control%qs_control%eps_ppnl
      CALL build_core_ppnl(matrix_h, matrix_p, force, virial, calculate_forces, use_virial, nder, &
                           qs_kind_set, atomic_kind_set, particle_set, sab_orb, sap_ppnl, eps_ppnl, nimages, cell_to_index)

      ! *** GAPW one-center-expansion (oce) matrices
      IF (dft_control%qs_control%gapw .OR. dft_control%qs_control%gapw_xc) THEN
         eps_fit = dft_control%qs_control%gapw_control%eps_fit
         IF (ASSOCIATED(sap_oce)) &
            CALL build_oce_matrices(oce%intac, calculate_forces, nder, qs_kind_set, particle_set, &
                                    sap_oce, eps_fit)
      END IF

      ! *** LRIGPW  matrices
      IF (lrigpw) THEN
         CALL get_qs_env(qs_env=qs_env, lri_env=lri_env)
         CALL build_lri_matrices(lri_env, qs_env, calculate_forces)
      END IF

      ! *** KG atomic potentials for nonadditive kinetic energy
      IF (kgpot) THEN
         CALL get_qs_env(qs_env=qs_env, kg_env=kg_env, dbcsr_dist=dbcsr_dist)
         CALL build_tnadd_mat(kg_env, matrix_p, force, virial, calculate_forces, use_virial, &
                              qs_kind_set, atomic_kind_set, particle_set, sab_orb, dbcsr_dist)
      END IF

      ! *** Put the core Hamiltonian matrix in the QS environment ***
      CALL set_qs_env(qs_env, oce=oce)
      CALL set_ks_env(ks_env, &
                      matrix_s_kp=matrix_s, &
                      kinetic_kp=matrix_t, &
                      matrix_h_kp=matrix_h)

      IF (qs_env%rel_control%rel_method /= rel_none) THEN
         IF (qs_env%rel_control%rel_transformation /= rel_trans_atom) THEN
            CALL cp__b("qs_core_hamiltonian.F",460,"Relativistic corrections of this type are currently not implemented")
         END IF
      END IF

      ! Print matrices if requested
      CALL dump_info_core_hamiltonian(qs_env, calculate_forces)
      CALL timestop(handle)

   END SUBROUTINE build_core_hamiltonian_matrix

! **************************************************************************************************
!> \brief Adds atomic blocks of relativistic correction for the kinetic energy
!> \param matrix_h ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param particle_set ...
! **************************************************************************************************
   SUBROUTINE build_atomic_relmat(matrix_h, atomic_kind_set, qs_kind_set, particle_set)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_atomic_relmat', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, iatom, ikind, jatom, natom
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: hblock, reltmat
      TYPE(cp_dbcsr_iterator)                            :: iter

      natom = SIZE(particle_set)
      ALLOCATE (kind_of(natom))

      CALL get_atomic_kind_set(atomic_kind_set, kind_of=kind_of)

      CALL cp_dbcsr_iterator_start(iter, matrix_h)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, hblock, blk)
         IF (iatom == jatom) THEN
            ikind = kind_of(iatom)
            CALL get_qs_kind(qs_kind_set(ikind), reltmat=reltmat)
            hblock = hblock+reltmat
         END IF
      END DO
      CALL cp_dbcsr_iterator_stop(iter)

      DEALLOCATE (kind_of)

   END SUBROUTINE build_atomic_relmat

! **************************************************************************************************
!> \brief Possibly prints matrices after the construction of the Core
!>     Hamiltonian Matrix
!> \param qs_env ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE dump_info_core_hamiltonian(qs_env, calculate_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_info_core_hamiltonian', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: after, handle, i, ic, iw, output_unit
      LOGICAL                                            :: omit_headers
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s, matrix_v
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrixkp_h, matrixkp_s, matrixkp_t
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)

      NULLIFY (logger, matrix_v, matrix_s, para_env)
      logger => cp_get_default_logger()
      CALL get_qs_env(qs_env, para_env=para_env)

      ! Print the distribution of the overlap matrix blocks
      ! this duplicates causes duplicate printing at the force calc
      IF (.NOT. calculate_forces) THEN
         IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                              qs_env%input, "PRINT%DISTRIBUTION"), cp_p_file)) THEN
            output_unit = cp_print_key_unit_nr(logger, qs_env%input, "PRINT%DISTRIBUTION", &
                                               extension=".distribution")
            CALL get_qs_env(qs_env, matrix_s_kp=matrixkp_s)
            CALL cp_dbcsr_write_matrix_dist(matrixkp_s(1, 1)%matrix, output_unit, para_env)
            CALL cp_print_key_finished_output(output_unit, logger, qs_env%input, "PRINT%DISTRIBUTION")
         END IF
      END IF

      CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
      ! Print the overlap integral matrix, if requested
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "DFT%PRINT%AO_MATRICES/OVERLAP"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/OVERLAP", &
                                   extension=".Log")
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         CALL get_qs_env(qs_env, matrix_s_kp=matrixkp_s)
         IF (ASSOCIATED(matrixkp_s)) THEN
            DO ic = 1, SIZE(matrixkp_s, 2)
               CALL cp_dbcsr_write_sparse_matrix(matrixkp_s(1, ic)%matrix, 4, after, qs_env, para_env, &
                                                 output_unit=iw, omit_headers=omit_headers)
            END DO
            IF (BTEST(cp_print_key_should_output(logger%iter_info, qs_env%input, &
                                                 "DFT%PRINT%AO_MATRICES/DERIVATIVES"), cp_p_file) &
                .AND. ASSOCIATED(matrix_s)) THEN
               DO ic = 1, SIZE(matrixkp_s, 2)
                  DO i = 2, SIZE(matrix_s)
                     CALL cp_dbcsr_write_sparse_matrix(matrixkp_s(i, ic)%matrix, 4, after, qs_env, para_env, &
                                                       output_unit=iw, omit_headers=omit_headers)
                  END DO
               END DO
            END IF
         END IF
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "DFT%PRINT%AO_MATRICES/OVERLAP")
      END IF

      ! Print the kinetic energy integral matrix, if requested
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "DFT%PRINT%AO_MATRICES/KINETIC_ENERGY"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/KINETIC_ENERGY", &
                                   extension=".Log")
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         CALL get_qs_env(qs_env, kinetic_kp=matrixkp_t)
         IF (ASSOCIATED(matrixkp_t)) THEN
            DO ic = 1, SIZE(matrixkp_t, 2)
               CALL cp_dbcsr_write_sparse_matrix(matrixkp_t(1, ic)%matrix, 4, after, qs_env, para_env, &
                                                 output_unit=iw, omit_headers=omit_headers)
            END DO
         END IF
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "DFT%PRINT%AO_MATRICES/KINETIC_ENERGY")
      END IF

      ! Print the potential energy matrix, if requested
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "DFT%PRINT%AO_MATRICES/POTENTIAL_ENERGY"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/POTENTIAL_ENERGY", &
                                   extension=".Log")
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         CALL get_qs_env(qs_env, matrix_h_kp=matrixkp_h, kinetic_kp=matrixkp_t)
         IF (ASSOCIATED(matrixkp_h)) THEN
            IF (SIZE(matrixkp_h, 2) == 1) THEN
               CALL cp_dbcsr_allocate_matrix_set(matrix_v, 1)
               ALLOCATE (matrix_v(1)%matrix)
               CALL cp_dbcsr_init(matrix_v(1)%matrix)
               CALL cp_dbcsr_copy(matrix_v(1)%matrix, matrixkp_h(1, 1)%matrix, name="POTENTIAL ENERGY MATRIX")
               CALL cp_dbcsr_add(matrix_v(1)%matrix, matrixkp_t(1, 1)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
               CALL cp_dbcsr_write_sparse_matrix(matrix_v(1)%matrix, 4, after, qs_env, &
                                                 para_env, output_unit=iw, omit_headers=omit_headers)
               CALL cp_dbcsr_deallocate_matrix_set(matrix_v)
            ELSE
               CALL cp__w("qs_core_hamiltonian.F",617,"Printing of potential energy matrix not implemented for k-points")
            END IF
         END IF
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "DFT%PRINT%AO_MATRICES/POTENTIAL_ENERGY")
      END IF

      ! Print the core Hamiltonian matrix, if requested
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "DFT%PRINT%AO_MATRICES/CORE_HAMILTONIAN"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/CORE_HAMILTONIAN", &
                                   extension=".Log")
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         after = MIN(MAX(after, 1), 16)
         CALL get_qs_env(qs_env, matrix_h_kp=matrixkp_h)
         IF (ASSOCIATED(matrixkp_h)) THEN
            DO ic = 1, SIZE(matrixkp_h, 2)
               CALL cp_dbcsr_write_sparse_matrix(matrixkp_h(1, ic)%matrix, 4, after, qs_env, para_env, &
                                                 output_unit=iw, omit_headers=omit_headers)
            END DO
         END IF
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "DFT%PRINT%AO_MATRICES/CORE_HAMILTONIAN")
      END IF

      CALL timestop(handle)

   END SUBROUTINE dump_info_core_hamiltonian

END MODULE qs_core_hamiltonian
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the Hamiltonian integral and Overlap matrices
!>        <a|b> and <a|T+V|b> for SCP-TB methods
!> \par History  Some refactoring related to new overlap and
!>               kinetic energy integrals [07.2014, JGH]
!>               Kpoints [08.2014, JGH]
!>       2014.11 unified k-point and gamma-point code [Ole Schuett]
!> \author JGH
! **************************************************************************************************
MODULE scptb_core_matrix
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_should_output
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE particle_types,                  ONLY: particle_type
   USE qs_core_hamiltonian,             ONLY: dump_info_core_hamiltonian
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_kinetic,                      ONLY: build_kinetic_matrix
   USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_overlap,                      ONLY: build_overlap_matrix
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE scptb_types,                     ONLY: get_scptb_parameter,&
                                              scptb_parameter_p_type,&
                                              scptb_parameter_type
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'scptb_core_matrix'

   PUBLIC :: build_scptb_core_matrix

CONTAINS

! **************************************************************************************************
!> \brief Builds the SCPTB core Hamiltonian matrix and its derivatives
!> \param qs_env           The QS environment
!> \param calculate_forces Flag to toggle force calculation
! **************************************************************************************************
   SUBROUTINE build_scptb_core_matrix(qs_env, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_scptb_core_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, handle, i, iatom, ic, &
                                                            icol, icor, ikind, inode, irow, j, &
                                                            jatom, jkind, l, li, ll, natom, nder, &
                                                            nimg, nkind, nspins
      INTEGER, DIMENSION(3)                              :: cell
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: defined, found, use_cell_mapping, &
                                                            use_virial
      LOGICAL, DIMENSION(:), POINTER                     :: scptb_defined
      REAL(KIND=dp)                                      :: dr, eps_filter, f0, kh, kij
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: hmt, kk, umt
      REAL(KIND=dp), DIMENSION(3)                        :: force_ab, rij
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ds_block, h_block, p_block, s_block
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_h, matrix_p, matrix_s, matrix_t, &
                                                            matrix_w
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(scptb_parameter_p_type), DIMENSION(:), &
         POINTER                                         :: scptb_kind_param
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind_a
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      NULLIFY (rho, force, atomic_kind_set, qs_kind_set, sab_orb, matrix_p, ks_env, &
               dft_control, matrix_h, matrix_s, matrix_t, kpoints)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      dft_control=dft_control, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set, &
                      sab_orb=sab_orb, &
                      virial=virial, &
                      para_env=para_env, &
                      kpoints=kpoints, &
                      matrix_s_kp=matrix_s, &
                      kinetic_kp=matrix_t, &
                      matrix_h_kp=matrix_h)

      CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
      use_cell_mapping = (SIZE(cell_to_index) > 1)

      nkind = SIZE(atomic_kind_set)
      natom = SIZE(particle_set)
      nimg = dft_control%nimages
      nspins = dft_control%nspins

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      ! filter for new matrices
      eps_filter = dft_control%qs_control%eps_filter_matrix

      IF (calculate_forces) THEN
         ! forces are not yet supported for k-points
         nder = 1
         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)
         CALL get_qs_env(qs_env, &
                         rho=rho, &
                         force=force, &
                         matrix_w_kp=matrix_w)

         CALL qs_rho_get(rho, rho_ao_kp=matrix_p)

         IF (nspins == 2) THEN
            DO ic = 1, nimg
               CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
               CALL cp_dbcsr_add(matrix_p(2, ic)%matrix, matrix_p(1, ic)%matrix, &
                                 alpha_scalar=-2.0_dp, beta_scalar=1.0_dp)
               CALL cp_dbcsr_add(matrix_w(1, ic)%matrix, matrix_w(2, ic)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF

         ! S matrix
         CALL build_overlap_matrix(ks_env, nderivative=nder, matrixkp_s=matrix_s, &
                                   matrix_name="OVERLAP MATRIX", &
                                   basis_type_a="ORB", &
                                   basis_type_b="ORB", &
                                   sab_nl=sab_orb, calculate_forces=.TRUE., &
                                   matrixkp_p=matrix_w, &
                                   qs_input=qs_env%input)
         ! T matrix
         CALL build_kinetic_matrix(ks_env, matrixkp_t=matrix_t, &
                                   matrix_name="KINETIC ENERGY MATRIX", &
                                   basis_type="ORB", &
                                   sab_nl=sab_orb, calculate_forces=.TRUE., &
                                   matrixkp_p=matrix_p, &
                                   eps_filter=eps_filter)
      ELSE
         IF (cp_print_key_should_output(logger%iter_info, qs_env%input, &
                                        "DFT%PRINT%AO_MATRICES/DERIVATIVES") /= 0) THEN
            nder = 1
         ELSE
            nder = 0
         END IF

         ! S matrix
         CALL build_overlap_matrix(ks_env, nderivative=nder, matrixkp_s=matrix_s, &
                                   matrix_name="OVERLAP MATRIX", &
                                   basis_type_a="ORB", &
                                   basis_type_b="ORB", &
                                   sab_nl=sab_orb, &
                                   qs_input=qs_env%input)
         ! T matrix
         CALL build_kinetic_matrix(ks_env, matrixkp_t=matrix_t, &
                                   matrix_name="KINETIC ENERGY MATRIX", &
                                   basis_type="ORB", &
                                   sab_nl=sab_orb, &
                                   eps_filter=eps_filter)
      END IF

      CALL set_ks_env(ks_env, matrix_s_kp=matrix_s, kinetic_kp=matrix_t)

      ! initialize H matrix
      CALL cp_dbcsr_allocate_matrix_set(matrix_h, 1, nimg)
      DO ic = 1, nimg
         ALLOCATE (matrix_h(1, ic)%matrix)
         CALL cp_dbcsr_init(matrix_h(1, ic)%matrix)
         CALL cp_dbcsr_copy(matrix_h(1, ic)%matrix, matrix_t(1, ic)%matrix, &
                            name="CORE HAMILTONIAN MATRIX")
      END DO

      !The band energy elements will be added to the kinetic energy.

      ALLOCATE (scptb_kind_param(nkind), scptb_defined(nkind))
      ALLOCATE (umt(100, nkind), hmt(100, nkind), kk(nkind, nkind))
      umt = 0._dp
      hmt = 0._dp
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind_a)
         scptb_kind_param(ikind)%scptb_param => scptb_kind_a
         CALL get_scptb_parameter(scptb_kind_a, defined=defined)
         scptb_defined(ikind) = defined
         li = 0
         DO l = 0, scptb_kind_a%lmaxorb
            ll = ((l+1)*(l+2))/2
            DO i = 1, scptb_kind_a%norb(l)
               hmt(li+1:li+ll, ikind) = scptb_kind_a%hcore(i, l)
               li = li+ll
            END DO
         END DO
      END DO
      !
      kk = 0.5_dp

      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                iatom=iatom, jatom=jatom, inode=inode, r=rij, cell=cell)
         IF (.NOT. scptb_defined(ikind)) CYCLE
         IF (.NOT. scptb_defined(jkind)) CYCLE

         kh = kk(ikind, jkind)

         IF (use_cell_mapping) THEN
            ic = cell_to_index(cell(1), cell(2), cell(3))
            IF(.NOT.(ic > 0))CALL cp__a("scptb_core_matrix.F",256)
         ELSE
            ic = 1
         END IF

         NULLIFY (h_block)
         IF (iatom <= jatom) THEN
            irow = iatom
            icol = jatom
         ELSE
            irow = jatom
            icol = iatom
         END IF
         CALL cp_dbcsr_get_block_p(matrix_h(1, ic)%matrix, irow, icol, h_block, found)
         IF(.NOT.(ASSOCIATED(h_block)))CALL cp__a("scptb_core_matrix.F",270)

         dr = SUM(rij(:)**2)
         !actual core energy determination , eq. 2.17 in my notes ???
         IF (iatom == jatom .AND. dr < 0.0001_dp) THEN
            DO i = 1, SIZE(h_block, 1)
               h_block(i, i) = h_block(i, i)+hmt(i, ikind)
            END DO
         ELSE
            NULLIFY (s_block) !s_block is overlap [sub]matrix ???
            CALL cp_dbcsr_get_block_p(matrix_s(1, ic)%matrix, irow, icol, s_block, found)
            IF(.NOT.(ASSOCIATED(s_block)))CALL cp__a("scptb_core_matrix.F",281)
            IF (irow == iatom) THEN
               DO i = 1, SIZE(h_block, 1)
                  DO j = 1, SIZE(h_block, 2)
                     kij = kh*(hmt(i, ikind)+hmt(j, jkind))
                     h_block(i, j) = h_block(i, j)+kij*s_block(i, j)
                  END DO
               END DO
            ELSE
               DO i = 1, SIZE(h_block, 1)
                  DO j = 1, SIZE(h_block, 2)
                     kij = kh*(hmt(j, ikind)+hmt(i, jkind))
                     h_block(i, j) = h_block(i, j)+kij*s_block(i, j)
                  END DO
               END DO
            END IF
            IF (calculate_forces) THEN
               atom_a = atom_of_kind(iatom)
               atom_b = atom_of_kind(jatom)
               CALL cp_dbcsr_get_block_p(matrix_p(1, ic)%matrix, irow, icol, p_block, found)
               IF(.NOT.(ASSOCIATED(p_block)))CALL cp__a("scptb_core_matrix.F",301)
               DO icor = 1, 3
                  force_ab(icor) = 0._dp
                  CALL cp_dbcsr_get_block_p(matrix_s(icor+1, ic)%matrix, irow, icol, ds_block, found)
                  IF(.NOT.(ASSOCIATED(ds_block)))CALL cp__a("scptb_core_matrix.F",305)
                  IF (irow == iatom) THEN
                     DO i = 1, SIZE(h_block, 1)
                        DO j = 1, SIZE(h_block, 2)
                           kij = 2._dp*kh*(hmt(i, ikind)+hmt(j, jkind))
                           force_ab(icor) = force_ab(icor)+kij*ds_block(i, j)*p_block(i, j)
                        END DO
                     END DO
                  ELSE
                     DO i = 1, SIZE(h_block, 1)
                        DO j = 1, SIZE(h_block, 2)
                           kij = 2._dp*kh*(hmt(j, ikind)+hmt(i, jkind))
                           force_ab(icor) = force_ab(icor)+kij*ds_block(i, j)*p_block(i, j)
                        END DO
                     END DO
                  END IF
               END DO
               IF (irow == iatom) force_ab = -force_ab
               force(ikind)%all_potential(:, atom_a) = force(ikind)%all_potential(:, atom_a)-force_ab(:)
               force(jkind)%all_potential(:, atom_b) = force(jkind)%all_potential(:, atom_b)+force_ab(:)
               IF (use_virial) THEN
                  f0 = -1.0_dp
                  IF (iatom == jatom) f0 = -0.5_dp
                  CALL virial_pair_force(virial%pv_virial, f0, force_ab, rij)
               END IF
            END IF

         END IF

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (scptb_kind_param, scptb_defined)
      DEALLOCATE (umt, hmt)

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind)
         IF (nspins == 2) THEN
            DO ic = 1, nimg
               CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                 alpha_scalar=0.5_dp, beta_scalar=0.5_dp)
               CALL cp_dbcsr_add(matrix_p(2, ic)%matrix, matrix_p(1, ic)%matrix, &
                                 alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
      END IF

      ! *** Put the core Hamiltonian matrix in the QS environment ***
      CALL set_ks_env(ks_env, matrix_h_kp=matrix_h)

      ! Print matrices if requested
      CALL dump_info_core_hamiltonian(qs_env, calculate_forces)
      CALL timestop(handle)

   END SUBROUTINE build_scptb_core_matrix

END MODULE scptb_core_matrix

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate MP2 energy
!> \par History
!>      06.2011 created [Mauro Del Ben]
!> \author Mauro Del Ben
! **************************************************************************************************
MODULE mp2_direct_method
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE gamma,                           ONLY: init_md_ftable
   USE hfx_energy_potential,            ONLY: coulomb4
   USE hfx_libint_wrapper_types,        ONLY: lib_int
   USE hfx_load_balance_methods,        ONLY: cost_model,&
                                              p1_energy,&
                                              p2_energy,&
                                              p3_energy
   USE hfx_pair_list_methods,           ONLY: build_pair_list_mp2
   USE hfx_screening_methods,           ONLY: calc_pair_dist_radii,&
                                              calc_screening_functions
   USE hfx_types,                       ONLY: &
        hfx_basis_info_type, hfx_basis_type, hfx_create_neighbor_cells, hfx_general_type, &
        hfx_load_balance_type, hfx_memory_type, hfx_pgf_list, hfx_pgf_product_list, &
        hfx_potential_type, hfx_screen_coeff_type, hfx_screening_type, hfx_type, log_zero, &
        pair_set_list_type
   USE input_constants,                 ONLY: do_mp2_potential_TShPSC
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE mathconstants,                   ONLY: zero
   USE message_passing,                 ONLY: mp_max,&
                                              mp_sendrecv,&
                                              mp_sum
   USE mp2_types,                       ONLY: init_TShPSC_lmax,&
                                              mp2_biel_type,&
                                              mp2_type,&
                                              pair_list_type_mp2
   USE orbital_pointers,                ONLY: ncoset
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE t_sh_p_s_c,                      ONLY: free_C0,&
                                              init
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC ::  mp2_canonical_direct_single_batch

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_direct_method'

!***

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param Emp2 ...
!> \param Emp2_Cou ...
!> \param Emp2_ex ...
!> \param mp2_env ...
!> \param qs_env ...
!> \param rho density matrix
!> \param para_env ...
!> \param mp2_biel ...
!> \param dimen ...
!> \param C ...
!> \param Auto ...
!> \param i_batch_start ...
!> \param Ni_occupied ...
!> \param occupied ...
!> \param elements_ij_proc ...
!> \param ij_list_proc ...
!> \param Nj_occupied ...
!> \param j_batch_start ...
!> \param occupied_beta ...
!> \param C_beta ...
!> \param Auto_beta ...
!> \param Integ_MP2 ...
!> \par History
!>      06.2011 created [Mauro Del Ben]
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE mp2_canonical_direct_single_batch(Emp2, Emp2_Cou, Emp2_ex, mp2_env, qs_env, rho, para_env, &
                                                mp2_biel, dimen, C, Auto, i_batch_start, Ni_occupied, &
                                                occupied, elements_ij_proc, ij_list_proc, Nj_occupied, j_batch_start, &
                                                occupied_beta, C_beta, Auto_beta, Integ_MP2)

      REAL(KIND=dp)                                      :: Emp2, Emp2_Cou, Emp2_ex
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(mp2_biel_type)                                :: mp2_biel
      INTEGER                                            :: dimen
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C
      REAL(KIND=dp), DIMENSION(dimen)                    :: Auto
      INTEGER                                            :: i_batch_start, Ni_occupied, occupied, &
                                                            elements_ij_proc
      INTEGER, DIMENSION(elements_ij_proc, 2)            :: ij_list_proc
      INTEGER                                            :: Nj_occupied, j_batch_start
      INTEGER, OPTIONAL                                  :: occupied_beta
      REAL(KIND=dp), DIMENSION(dimen, dimen), OPTIONAL   :: C_beta
      REAL(KIND=dp), DIMENSION(dimen), OPTIONAL          :: Auto_beta
      REAL(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :), OPTIONAL                 :: Integ_MP2

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_canonical_direct_single_batch', &
         routineP = moduleN//':'//routineN

      INTEGER :: bits_max_val, cache_size, case_index, case_send_receive, counter_proc, &
         elements_ij_proc_rec, elements_kl_proc, global_counter, handle, i, i_list_ij, i_list_kl, &
         i_set_list_ij, i_set_list_ij_start, i_set_list_ij_stop, i_set_list_kl, &
         i_set_list_kl_start, i_set_list_kl_stop, i_start, i_thread, iatom, iatom_end, &
         iatom_start, iiB, ij_elem_max, ikind, index_ij_rec, index_ij_send, index_kl, &
         index_proc_ij, index_proc_shift, irep, iset, jatom, jatom_end, jatom_start, jjB, jkind, &
         jset, katom, katom_end, katom_start, kkB, kkind, kset, l_max, latom, latom_end, &
         latom_start, lkind, llB, lset
      INTEGER :: max_am, max_num_call_sec_transf, max_pgf, max_set, multiple, &
         my_num_call_sec_transf, n_threads, natom, ncob, ncos_max, nints, nkind, nneighbors, &
         nseta, nsetb, nsgf_max, nspins, primitive_counter, proc_num, proc_receive, proc_send, &
         R_offset_rec, Rsize_rec, S_offset_rec, same_size_kl_index, sgfb, sphi_a_u1, sphi_a_u2, &
         sphi_a_u3, sphi_b_u1, sphi_b_u2, sphi_b_u3, sphi_c_u1, sphi_c_u2, sphi_c_u3, sphi_d_u1, &
         sphi_d_u2, sphi_d_u3, Ssize_rec, step_size, total_num_RS_task, unit_id
      INTEGER(int_8) :: estimate_to_store_int, max_val_memory, mem_eris, mem_eris_disk, &
         mem_max_val, n_processes, ncpu, neris_disk, neris_incore, neris_onthefly, neris_tmp, &
         neris_total, nprim_ints, shm_neris_disk, shm_neris_incore, shm_neris_onthefly, &
         shm_neris_total, shm_nprim_ints, shm_stor_count_int_disk, shm_stor_count_max_val, &
         shm_storage_counter_integrals
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of, last_sgf_global, nimages, &
                                                            proc_map, proc_num_task, &
                                                            same_size_kl_elements_counter
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: kl_list_proc, task_counter_RS, &
                                                            task_counter_RS_temp
      INTEGER, DIMENSION(4)                              :: RS_counter_temp
      INTEGER, DIMENSION(5)                              :: size_parameter_rec, size_parameter_send
      INTEGER, DIMENSION(:), POINTER :: la_max, la_min, lb_max, lb_min, lc_max, lc_min, ld_max, &
         ld_min, npgfa, npgfb, npgfc, npgfd, nsgfa, nsgfb, nsgfc, nsgfd, shm_block_offset
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfb, nsgfl_a, nsgfl_b, nsgfl_c, &
                                                            nsgfl_d, shm_atomic_block_offset
      INTEGER, DIMENSION(:, :), POINTER, SAVE            :: shm_is_assoc_atomic_block
      INTEGER, DIMENSION(:, :, :, :), POINTER            :: shm_set_offset
      INTEGER, SAVE                                      :: shm_number_of_p_entries
      LOGICAL                                            :: alpha_beta_case, copy_integrals, &
                                                            do_periodic, treat_lsd_in_core
      LOGICAL, DIMENSION(:, :), POINTER                  :: shm_atomic_pair_list
      REAL(KIND=dp) :: cartesian_estimate, coeffs_kind_max0, compression_factor, &
         compression_factor_disk, cost_tmp, eps_schwarz, eps_storage, hf_fraction, ln_10, &
         log10_eps_schwarz, log10_pmax, max_contraction_val, max_val1, max_val2, max_val2_set, &
         pmax_atom, pmax_entry, ra(3), rab2, rb(3), rc(3), rcd2, rd(3), screen_kind_ij, &
         screen_kind_kl
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cost_RS, cost_RS_temp, ee_buffer1, &
                                                            ee_buffer2, ee_primitives_tmp, &
                                                            ee_work, ee_work2, primitive_integrals
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIb_RS_mat_rec, C_beta_T
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb, BIb_RS_mat_rec_big, zero_mat_big
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: BI1, MNRS
      REAL(KIND=dp), DIMENSION(:), POINTER               :: p_work
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: max_contraction, shm_pmax_block, sphi_b, &
                                                            zeta, zetb, zetc, zetd
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: sphi_a_ext_set, sphi_b_ext_set, &
                                                            sphi_c_ext_set, sphi_d_ext_set
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: sphi_a_ext, sphi_b_ext, sphi_c_ext, &
                                                            sphi_d_ext
      REAL(KIND=dp), DIMENSION(dimen, 2)                 :: zero_mat
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C_T
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(hfx_basis_info_type), POINTER                 :: basis_info
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_parameter
      TYPE(hfx_general_type)                             :: general_parameter
      TYPE(hfx_load_balance_type), POINTER               :: load_balance_parameter
      TYPE(hfx_memory_type), POINTER                     :: memory_parameter
      TYPE(hfx_pgf_list), ALLOCATABLE, DIMENSION(:)      :: pgf_list_ij, pgf_list_kl
      TYPE(hfx_pgf_product_list), ALLOCATABLE, &
         DIMENSION(:)                                    :: pgf_product_list
      TYPE(hfx_potential_type)                           :: mp2_potential_parameter, &
                                                            potential_parameter
      TYPE(hfx_screen_coeff_type), DIMENSION(:, :), &
         POINTER                                         :: screen_coeffs_kind, tmp_R_1, tmp_R_2, &
                                                            tmp_screen_pgf1, tmp_screen_pgf2
      TYPE(hfx_screen_coeff_type), &
         DIMENSION(:, :, :, :), POINTER                  :: screen_coeffs_set
      TYPE(hfx_screen_coeff_type), &
         DIMENSION(:, :, :, :, :, :), POINTER            :: radii_pgf, screen_coeffs_pgf
      TYPE(hfx_screening_type)                           :: screening_parameter
      TYPE(hfx_type), POINTER                            :: actual_x_data, shm_master_x_data
      TYPE(lib_int)                                      :: private_lib
      TYPE(pair_list_type_mp2)                           :: list_ij, list_kl
      TYPE(pair_set_list_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: set_list_ij, set_list_kl
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CALL timeset(routineN, handle)
      NULLIFY (dft_control)

      ! The Integ_MP2 will contain the (ia|jb) integrals, necessary for example
      ! for the RI-MP2 basis optimization. In this case the number of ij batches
      ! has to be equal to 1 (all integrals over molecular orbitals are compituted
      ! in a sigle step).
      copy_integrals = .FALSE.
      IF (PRESENT(Integ_MP2)) copy_integrals = .TRUE.

      alpha_beta_case = .FALSE.

      irep = 1

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      cell=cell, &
                      dft_control=dft_control)

      !! Calculate l_max used in fgamma , because init_md_ftable is definitely not thread safe
      nkind = SIZE(atomic_kind_set, 1)
      l_max = 0
      DO ikind = 1, nkind
         l_max = MAX(l_max, MAXVAL(qs_env%x_data(1, 1)%basis_parameter(ikind)%lmax))
      ENDDO
      l_max = 4*l_max
      CALL init_md_ftable(l_max)

      IF (mp2_env%potential_parameter%potential_type == do_mp2_potential_TShPSC) THEN
         IF (l_max > init_TShPSC_lmax) THEN
            IF (para_env%mepos == 0) THEN
               CALL open_file(unit_number=unit_id, file_name=mp2_env%potential_parameter%filename)
            END IF
            CALL init(l_max, unit_id, para_env%mepos, para_env%group)
            IF (para_env%mepos == 0) THEN
               CALL close_file(unit_id)
            END IF
            init_TShPSC_lmax = l_max
         END IF
         mp2_potential_parameter%cutoff_radius = mp2_env%potential_parameter%truncation_radius/2.0_dp
      END IF

      mp2_potential_parameter%potential_type = mp2_env%potential_parameter%potential_type

      n_threads = 1

      shm_neris_total = 0
      shm_nprim_ints = 0
      shm_neris_onthefly = 0
      shm_storage_counter_integrals = 0
      shm_stor_count_int_disk = 0
      shm_neris_incore = 0
      shm_neris_disk = 0
      shm_stor_count_max_val = 0

      ln_10 = LOG(10.0_dp)
      i_thread = 0

      actual_x_data => qs_env%x_data(irep, i_thread+1)

      shm_master_x_data => qs_env%x_data(irep, 1)

      do_periodic = actual_x_data%periodic_parameter%do_periodic

      IF (do_periodic) THEN
         ! ** Rebuild neighbor lists in case the cell has changed (i.e. NPT MD)
         actual_x_data%periodic_parameter%number_of_shells = actual_x_data%periodic_parameter%mode
         CALL hfx_create_neighbor_cells(actual_x_data, actual_x_data%periodic_parameter%number_of_shells_from_input, &
                                        cell, i_thread)
      END IF

      screening_parameter = actual_x_data%screening_parameter
      potential_parameter = actual_x_data%potential_parameter

      general_parameter = actual_x_data%general_parameter
      load_balance_parameter => actual_x_data%load_balance_parameter
      memory_parameter => actual_x_data%memory_parameter

      cache_size = memory_parameter%cache_size
      bits_max_val = memory_parameter%bits_max_val

      basis_parameter => actual_x_data%basis_parameter
      basis_info => actual_x_data%basis_info

      treat_lsd_in_core = general_parameter%treat_lsd_in_core

      ncpu = para_env%num_pe
      n_processes = ncpu*n_threads

      !! initalize some counters
      neris_total = 0_int_8
      neris_incore = 0_int_8
      neris_disk = 0_int_8
      neris_onthefly = 0_int_8
      mem_eris = 0_int_8
      mem_eris_disk = 0_int_8
      mem_max_val = 0_int_8
      compression_factor = 0.0_dp
      compression_factor_disk = 0.0_dp
      nprim_ints = 0_int_8
      neris_tmp = 0_int_8
      max_val_memory = 1_int_8

      max_am = basis_info%max_am

      max_set = basis_info%max_set
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      particle_set=particle_set)

      natom = SIZE(particle_set, 1)

      ALLOCATE (kind_of(natom))

      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               kind_of=kind_of)

      !! precompute maximum nco and allocate scratch
      ncos_max = 0
      nsgf_max = 0
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         nseta = basis_parameter(ikind)%nset
         npgfa => basis_parameter(ikind)%npgf
         la_max => basis_parameter(ikind)%lmax
         nsgfa => basis_parameter(ikind)%nsgf
         DO iset = 1, nseta
            ncos_max = MAX(ncos_max, ncoset(la_max(iset)))
            nsgf_max = MAX(nsgf_max, nsgfa(iset))
         ENDDO
      ENDDO
      !! Allocate the arrays for the integrals.
      ALLOCATE (primitive_integrals(nsgf_max**4))
      primitive_integrals = 0.0_dp

      ALLOCATE (ee_work(ncos_max**4))
      ALLOCATE (ee_work2(ncos_max**4))
      ALLOCATE (ee_buffer1(ncos_max**4))
      ALLOCATE (ee_buffer2(ncos_max**4))
      ALLOCATE (ee_primitives_tmp(nsgf_max**4))

      nspins = dft_control%nspins

      ALLOCATE (max_contraction(max_set, natom))

      max_contraction = 0.0_dp
      max_pgf = 0
      DO jatom = 1, natom
         jkind = kind_of(jatom)
         lb_max => basis_parameter(jkind)%lmax
         nsetb = basis_parameter(jkind)%nset
         npgfb => basis_parameter(jkind)%npgf
         first_sgfb => basis_parameter(jkind)%first_sgf
         sphi_b => basis_parameter(jkind)%sphi
         nsgfb => basis_parameter(jkind)%nsgf
         DO jset = 1, nsetb
            ! takes the primitive to contracted transformation into account
            ncob = npgfb(jset)*ncoset(lb_max(jset))
            sgfb = first_sgfb(1, jset)
            ! if the primitives are assumed to be all of max_val2, max_val2*p2s_b becomes
            ! the maximum value after multiplication with sphi_b
            max_contraction(jset, jatom) = MAXVAL((/(SUM(ABS(sphi_b(1:ncob, i))), i=sgfb, sgfb+nsgfb(jset)-1)/))
            max_pgf = MAX(max_pgf, npgfb(jset))
         ENDDO
      ENDDO

      ! ** Allocate buffers for pgf_lists
      nneighbors = SIZE(actual_x_data%neighbor_cells)
      ALLOCATE (pgf_list_ij(max_pgf**2))
      ALLOCATE (pgf_list_kl(max_pgf**2))
      ALLOCATE (pgf_product_list(nneighbors**3))
      ALLOCATE (nimages(max_pgf**2))

      DO i = 1, max_pgf**2
         ALLOCATE (pgf_list_ij(i)%image_list(nneighbors))
         ALLOCATE (pgf_list_kl(i)%image_list(nneighbors))
      END DO

      ! ** Set pointers
      shm_number_of_p_entries = shm_master_x_data%number_of_p_entries
      shm_is_assoc_atomic_block => shm_master_x_data%is_assoc_atomic_block
      shm_atomic_block_offset => shm_master_x_data%atomic_block_offset
      shm_set_offset => shm_master_x_data%set_offset
      shm_block_offset => shm_master_x_data%block_offset

      !!! Skipped part

      !! Get screening parameter
      eps_schwarz = screening_parameter%eps_schwarz
      IF (eps_schwarz <= 0.0_dp) THEN
         log10_eps_schwarz = log_zero
      ELSE
         log10_eps_schwarz = LOG10(eps_schwarz)
      END IF
      !! get storage epsilon
      eps_storage = eps_schwarz*memory_parameter%eps_storage_scaling

      !! If we have a hybrid functional, we may need only a fraction of exact exchange
      hf_fraction = general_parameter%fraction

      !! The number of integrals that fit into the given MAX_MEMORY

      !! Parameters related to the potential 1/r, erf(wr)/r, erfc(wr/r)
      potential_parameter = actual_x_data%potential_parameter

      logger => cp_get_default_logger()

      private_lib = actual_x_data%lib

      !! Helper array to map local basis function indeces to global ones
      ALLOCATE (last_sgf_global(0:natom))
      last_sgf_global(0) = 0
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         last_sgf_global(iatom) = last_sgf_global(iatom-1)+basis_parameter(ikind)%nsgf_total
      END DO

      !!!!!!!! Missing part on the density matrix

      !! Initialize schwarz screening matrices for near field estimates and boxing screening matrices
      !! for far field estimates. The update is only performed if the geomtry of the system changed.
      !! If the system is periodic, then the corresponding routines are called and some variables
      !! are initialized

      IF (.NOT. shm_master_x_data%screen_funct_is_initialized) THEN
         CALL calc_pair_dist_radii(qs_env, basis_parameter, &
                                   shm_master_x_data%pair_dist_radii_pgf, max_set, max_pgf, eps_schwarz, &
                                   n_threads, i_thread)
         CALL calc_screening_functions(qs_env, basis_parameter, private_lib, shm_master_x_data%potential_parameter, &
                                       shm_master_x_data%screen_funct_coeffs_set, &
                                       shm_master_x_data%screen_funct_coeffs_kind, &
                                       shm_master_x_data%screen_funct_coeffs_pgf, &
                                       shm_master_x_data%pair_dist_radii_pgf, &
                                       max_set, max_pgf, n_threads, i_thread, p_work)

         shm_master_x_data%screen_funct_is_initialized = .TRUE.
      END IF

      screen_coeffs_set => shm_master_x_data%screen_funct_coeffs_set
      screen_coeffs_kind => shm_master_x_data%screen_funct_coeffs_kind
      screen_coeffs_pgf => shm_master_x_data%screen_funct_coeffs_pgf
      radii_pgf => shm_master_x_data%pair_dist_radii_pgf

!!!!!!!!!
      ALLOCATE (list_ij%elements(natom**2))
      ALLOCATE (list_kl%elements(natom**2))
!!!!!!!!!

      coeffs_kind_max0 = MAXVAL(screen_coeffs_kind(:, :)%x(2))
      ALLOCATE (set_list_ij((max_set*natom)**2))
      ALLOCATE (set_list_kl((max_set*natom)**2))

      !! precalculate maximum density matrix elements in blocks
      actual_x_data%pmax_block = 0.0_dp
      shm_pmax_block => actual_x_data%pmax_block

      shm_atomic_pair_list => actual_x_data%atomic_pair_list

      iatom_start = 1
      iatom_end = natom
      jatom_start = 1
      jatom_end = natom
      katom_start = 1
      katom_end = natom
      latom_start = 1
      latom_end = natom

      CALL build_pair_list_mp2(natom, list_ij, set_list_ij, iatom_start, iatom_end, &
                               jatom_start, jatom_end, &
                               kind_of, basis_parameter, particle_set, &
                               do_periodic, screen_coeffs_set, screen_coeffs_kind, &
                               coeffs_kind_max0, log10_eps_schwarz, cell, 0.D+00, &
                               shm_atomic_pair_list)

      CALL build_pair_list_mp2(natom, list_kl, set_list_kl, katom_start, katom_end, &
                               latom_start, latom_end, &
                               kind_of, basis_parameter, particle_set, &
                               do_periodic, screen_coeffs_set, screen_coeffs_kind, &
                               coeffs_kind_max0, log10_eps_schwarz, cell, 0.D+00, &
                               shm_atomic_pair_list)

      ALLOCATE (BIb(dimen, dimen, elements_ij_proc))
      BIb = 0.0D+00
      C_T = TRANSPOSE(C)

      IF (PRESENT(occupied_beta) .AND. PRESENT(C_beta) .AND. PRESENT(Auto_beta)) THEN
         ALLOCATE (C_beta_T(dimen, dimen))
         C_beta_T(:, :) = TRANSPOSE(C_beta)
         alpha_beta_case = .TRUE.
      END IF

      ij_elem_max = elements_ij_proc
      CALL mp_max(ij_elem_max, para_env%group)

      ! proc_map, vector that replicate the processor numbers also
      ! for negative and positive number > num_pe
      ! needed to know which is the processor, to respect to another one,
      ! for a given shift
      ALLOCATE (proc_map(-para_env%num_pe:2*para_env%num_pe-1))
      DO iiB = 0, para_env%num_pe-1
         proc_map(iiB) = iiB
         proc_map(-iiB-1) = para_env%num_pe-iiB-1
         proc_map(para_env%num_pe+iiB) = iiB
      END DO

      ! calculate the minimum multiple of num_pe >= to Ni_occupied*occupied, in such a way
      ! that the i, j loop is performed exactly the same number of time for each procewssor
      multiple = 0
      DO
         multiple = multiple+para_env%num_pe
         IF (multiple >= Ni_occupied*Nj_occupied) EXIT
      END DO

      ! proc_num_task contains the numer of time second occupied
      ! orbital transformation is called for each processor, needs for balancing
      ! the point to point send
      ALLOCATE (proc_num_task(0:para_env%num_pe-1))

      proc_num_task = 0

      ! step_size=nsgf_max

      counter_proc = 0
      !!!! DO iiB=nsgf_max**2+1, 0, -step_size
      DO i_list_ij = 1, list_ij%n_element
         iatom = list_ij%elements(i_list_ij)%pair(1)
         jatom = list_ij%elements(i_list_ij)%pair(2)
         i_set_list_ij_start = list_ij%elements(i_list_ij)%set_bounds(1)
         i_set_list_ij_stop = list_ij%elements(i_list_ij)%set_bounds(2)
         ikind = list_ij%elements(i_list_ij)%kind_pair(1)
         jkind = list_ij%elements(i_list_ij)%kind_pair(2)

         nsgfb => basis_parameter(jkind)%nsgf
         nsgfa => basis_parameter(ikind)%nsgf

         ! proc_num=MOD(i_list_ij,para_env%num_pe)

         DO i_set_list_ij = i_set_list_ij_start, i_set_list_ij_stop
            iset = set_list_ij(i_set_list_ij)%pair(1)
            jset = set_list_ij(i_set_list_ij)%pair(2)
            IF (iatom == jatom .AND. jset < iset) CYCLE

            !!!! IF(nsgfa(iset)*nsgfb(jset)>iiB-step_size .AND. nsgfa(iset)*nsgfb(jset)<=iiB) THEN
            counter_proc = counter_proc+1
            proc_num = MOD(counter_proc, para_env%num_pe)

            proc_num_task(proc_num) = proc_num_task(proc_num)+1
            !!!! END IF

         END DO
      END DO
      !!!! END DO
      ! calculate the exact maximum  number of call to the second occupied
      ! orbital transformation
      ! max_num_call_sec_transf=MAXVAL(proc_num_task)

      ! distribute the RS pair over all processor
      ALLOCATE (kl_list_proc(proc_num_task(para_env%mepos), 3))

      kl_list_proc = 0

      !!!! ALLOCATE(same_size_kl_elements_counter((nsgf_max**2+1)/step_size+1))
      !!!! same_size_kl_elements_counter=0

      !!!! same_size_kl_index=0
      counter_proc = 0
      elements_kl_proc = 0
      !!!! DO iiB=nsgf_max**2+1, 0, -step_size
      DO i_list_ij = 1, list_ij%n_element
         iatom = list_ij%elements(i_list_ij)%pair(1)
         jatom = list_ij%elements(i_list_ij)%pair(2)
         i_set_list_ij_start = list_ij%elements(i_list_ij)%set_bounds(1)
         i_set_list_ij_stop = list_ij%elements(i_list_ij)%set_bounds(2)
         ikind = list_ij%elements(i_list_ij)%kind_pair(1)
         jkind = list_ij%elements(i_list_ij)%kind_pair(2)

         nsgfb => basis_parameter(jkind)%nsgf
         nsgfa => basis_parameter(ikind)%nsgf

         ! proc_num=MOD(i_list_ij,para_env%num_pe)

         DO i_set_list_ij = i_set_list_ij_start, i_set_list_ij_stop
            iset = set_list_ij(i_set_list_ij)%pair(1)
            jset = set_list_ij(i_set_list_ij)%pair(2)
            IF (iatom == jatom .AND. jset < iset) CYCLE

            !!!! IF(nsgfa(iset)*nsgfb(jset)>iiB-step_size .AND. nsgfa(iset)*nsgfb(jset)<=iiB) THEN
            counter_proc = counter_proc+1
            proc_num = MOD(counter_proc, para_env%num_pe)

            IF (proc_num == para_env%mepos) THEN
               elements_kl_proc = elements_kl_proc+1
               kl_list_proc(elements_kl_proc, 1) = i_list_ij
               kl_list_proc(elements_kl_proc, 2) = i_set_list_ij
               kl_list_proc(elements_kl_proc, 3) = counter_proc
            END IF
            !!!! END IF

         END DO
      END DO
      !!!! same_size_kl_index=same_size_kl_index+1
      !!!! same_size_kl_elements_counter(same_size_kl_index)=elements_kl_proc
      !!!! END DO

      total_num_RS_task = SUM(proc_num_task)
      ALLOCATE (task_counter_RS(total_num_RS_task, 4))

      ALLOCATE (cost_RS(total_num_RS_task))

      task_counter_RS = 0
      cost_RS = 0.0_dp

      DO case_index = 1, 2

         my_num_call_sec_transf = 0
         DO index_kl = 1, elements_kl_proc

            i_list_ij = kl_list_proc(index_kl, 1)
            i_set_list_ij = kl_list_proc(index_kl, 2)

            iatom = list_ij%elements(i_list_ij)%pair(1)
            jatom = list_ij%elements(i_list_ij)%pair(2)
            i_set_list_ij_start = list_ij%elements(i_list_ij)%set_bounds(1)
            i_set_list_ij_stop = list_ij%elements(i_list_ij)%set_bounds(2)
            ikind = list_ij%elements(i_list_ij)%kind_pair(1)
            jkind = list_ij%elements(i_list_ij)%kind_pair(2)
            ra = list_ij%elements(i_list_ij)%r1
            rb = list_ij%elements(i_list_ij)%r2
            rab2 = list_ij%elements(i_list_ij)%dist2

            la_max => basis_parameter(ikind)%lmax
            la_min => basis_parameter(ikind)%lmin
            npgfa => basis_parameter(ikind)%npgf
            nseta = basis_parameter(ikind)%nset
            zeta => basis_parameter(ikind)%zet
            nsgfa => basis_parameter(ikind)%nsgf
            sphi_a_ext => basis_parameter(ikind)%sphi_ext(:, :, :, :)
            nsgfl_a => basis_parameter(ikind)%nsgfl
            sphi_a_u1 = UBOUND(sphi_a_ext, 1)
            sphi_a_u2 = UBOUND(sphi_a_ext, 2)
            sphi_a_u3 = UBOUND(sphi_a_ext, 3)

            lb_max => basis_parameter(jkind)%lmax
            lb_min => basis_parameter(jkind)%lmin
            npgfb => basis_parameter(jkind)%npgf
            nsetb = basis_parameter(jkind)%nset
            zetb => basis_parameter(jkind)%zet
            nsgfb => basis_parameter(jkind)%nsgf
            sphi_b_ext => basis_parameter(jkind)%sphi_ext(:, :, :, :)
            nsgfl_b => basis_parameter(jkind)%nsgfl
            sphi_b_u1 = UBOUND(sphi_b_ext, 1)
            sphi_b_u2 = UBOUND(sphi_b_ext, 2)
            sphi_b_u3 = UBOUND(sphi_b_ext, 3)

            !!!!!     DO i_set_list_ij=i_set_list_ij_start, i_set_list_ij_stop
            iset = set_list_ij(i_set_list_ij)%pair(1)
            jset = set_list_ij(i_set_list_ij)%pair(2)

            !!!!!         IF(iatom==jatom .AND. jset<iset) CYCLE

            ncob = npgfb(jset)*ncoset(lb_max(jset))
            max_val1 = screen_coeffs_set(jset, iset, jkind, ikind)%x(1)*rab2+ &
                       screen_coeffs_set(jset, iset, jkind, ikind)%x(2)

            sphi_a_ext_set => sphi_a_ext(:, :, :, iset)
            sphi_b_ext_set => sphi_b_ext(:, :, :, jset)

            IF (case_index == 1) THEN
               global_counter = kl_list_proc(index_kl, 3)
               task_counter_RS(global_counter, 1) = i_list_ij
               task_counter_RS(global_counter, 2) = i_set_list_ij
               task_counter_RS(global_counter, 3) = nsgfb(jset)*nsgfa(iset)
            END IF

            IF (ALLOCATED(BI1)) DEALLOCATE (BI1)
            ALLOCATE (BI1(dimen, Ni_occupied, nsgfb(jset), nsgfa(iset)))

            BI1 = 0.D+00

            !!!t1=m_walltime()

            DO i_list_kl = 1, list_kl%n_element
               !!!!!DO i_list_kl=1+para_env%mepos,list_kl%n_element,para_env%num_pe

               katom = list_kl%elements(i_list_kl)%pair(1)
               latom = list_kl%elements(i_list_kl)%pair(2)

               i_set_list_kl_start = list_kl%elements(i_list_kl)%set_bounds(1)
               i_set_list_kl_stop = list_kl%elements(i_list_kl)%set_bounds(2)
               kkind = list_kl%elements(i_list_kl)%kind_pair(1)
               lkind = list_kl%elements(i_list_kl)%kind_pair(2)
               rc = list_kl%elements(i_list_kl)%r1
               rd = list_kl%elements(i_list_kl)%r2
               rcd2 = list_kl%elements(i_list_kl)%dist2

               pmax_atom = 0.0_dp

               screen_kind_ij = screen_coeffs_kind(jkind, ikind)%x(1)*rab2+ &
                                screen_coeffs_kind(jkind, ikind)%x(2)
               screen_kind_kl = screen_coeffs_kind(lkind, kkind)%x(1)*rcd2+ &
                                screen_coeffs_kind(lkind, kkind)%x(2)

               !!!!! Change the loop order
               IF (max_val1+screen_kind_kl+pmax_atom < log10_eps_schwarz) CYCLE
               !!!!!
               IF (screen_kind_ij+screen_kind_kl+pmax_atom < log10_eps_schwarz) CYCLE

               lc_max => basis_parameter(kkind)%lmax
               lc_min => basis_parameter(kkind)%lmin
               npgfc => basis_parameter(kkind)%npgf
               zetc => basis_parameter(kkind)%zet
               nsgfc => basis_parameter(kkind)%nsgf
               sphi_c_ext => basis_parameter(kkind)%sphi_ext(:, :, :, :)
               nsgfl_c => basis_parameter(kkind)%nsgfl
               sphi_c_u1 = UBOUND(sphi_c_ext, 1)
               sphi_c_u2 = UBOUND(sphi_c_ext, 2)
               sphi_c_u3 = UBOUND(sphi_c_ext, 3)

               ld_max => basis_parameter(lkind)%lmax
               ld_min => basis_parameter(lkind)%lmin
               npgfd => basis_parameter(lkind)%npgf
               zetd => basis_parameter(lkind)%zet
               nsgfd => basis_parameter(lkind)%nsgf
               sphi_d_ext => basis_parameter(lkind)%sphi_ext(:, :, :, :)
               nsgfl_d => basis_parameter(lkind)%nsgfl
               sphi_d_u1 = UBOUND(sphi_d_ext, 1)
               sphi_d_u2 = UBOUND(sphi_d_ext, 2)
               sphi_d_u3 = UBOUND(sphi_d_ext, 3)

               DO i_set_list_kl = i_set_list_kl_start, i_set_list_kl_stop
                  kset = set_list_kl(i_set_list_kl)%pair(1)
                  lset = set_list_kl(i_set_list_kl)%pair(2)

                  IF (katom == latom .AND. lset < kset) CYCLE

                  max_val2_set = (screen_coeffs_set(lset, kset, lkind, kkind)%x(1)*rcd2+ &
                                  screen_coeffs_set(lset, kset, lkind, kkind)%x(2))
                  max_val2 = max_val1+max_val2_set

                  !! Near field screening
                  IF (max_val2+pmax_atom < log10_eps_schwarz) CYCLE
                  sphi_c_ext_set => sphi_c_ext(:, :, :, kset)
                  sphi_d_ext_set => sphi_d_ext(:, :, :, lset)
                  !! get max_vals if we screen on initial density
                  pmax_entry = 0.0_dp

                  log10_pmax = pmax_entry
                  max_val2 = max_val2+log10_pmax
                  IF (max_val2 < log10_eps_schwarz) CYCLE
                  pmax_entry = EXP(log10_pmax*ln_10)

                  IF (case_index == 2) THEN
                     IF (ALLOCATED(MNRS)) DEALLOCATE (MNRS)
                     ALLOCATE (MNRS(nsgfd(lset), nsgfc(kset), nsgfb(jset), nsgfa(iset)))

                     MNRS = 0.D+00

                     max_contraction_val = max_contraction(iset, iatom)* &
                                           max_contraction(jset, jatom)* &
                                           max_contraction(kset, katom)* &
                                           max_contraction(lset, latom)*pmax_entry
                     tmp_R_1 => radii_pgf(:, :, jset, iset, jkind, ikind)
                     tmp_R_2 => radii_pgf(:, :, lset, kset, lkind, kkind)
                     tmp_screen_pgf1 => screen_coeffs_pgf(:, :, jset, iset, jkind, ikind)
                     tmp_screen_pgf2 => screen_coeffs_pgf(:, :, lset, kset, lkind, kkind)

                     CALL coulomb4(private_lib, ra, rb, rc, rd, npgfa(iset), npgfb(jset), npgfc(kset), npgfd(lset), &
                                   la_min(iset), la_max(iset), lb_min(jset), lb_max(jset), &
                                   lc_min(kset), lc_max(kset), ld_min(lset), ld_max(lset), &
                                   nsgfa(iset), nsgfb(jset), nsgfc(kset), nsgfd(lset), &
                                   sphi_a_u1, sphi_a_u2, sphi_a_u3, &
                                   sphi_b_u1, sphi_b_u2, sphi_b_u3, &
                                   sphi_c_u1, sphi_c_u2, sphi_c_u3, &
                                   sphi_d_u1, sphi_d_u2, sphi_d_u3, &
                                   zeta(1:npgfa(iset), iset), zetb(1:npgfb(jset), jset), &
                                   zetc(1:npgfc(kset), kset), zetd(1:npgfd(lset), lset), &
                                   primitive_integrals, &
                                   mp2_potential_parameter, &
                                   actual_x_data%neighbor_cells, screen_coeffs_set(jset, iset, jkind, ikind)%x, &
                                   screen_coeffs_set(lset, kset, lkind, kkind)%x, eps_schwarz, &
                                   max_contraction_val, cartesian_estimate, cell, neris_tmp, &
                                   log10_pmax, log10_eps_schwarz, &
                                   tmp_R_1, tmp_R_2, tmp_screen_pgf1, tmp_screen_pgf2, &
                                   pgf_list_ij, pgf_list_kl, pgf_product_list, &
                                   nsgfl_a(:, iset), nsgfl_b(:, jset), &
                                   nsgfl_c(:, kset), nsgfl_d(:, lset), &
                                   sphi_a_ext_set, &
                                   sphi_b_ext_set, &
                                   sphi_c_ext_set, &
                                   sphi_d_ext_set, &
                                   ee_work, ee_work2, ee_buffer1, ee_buffer2, ee_primitives_tmp, &
                                   nimages, do_periodic, p_work)
                     !END IF

                     nints = nsgfa(iset)*nsgfb(jset)*nsgfc(kset)*nsgfd(lset)
                     neris_total = neris_total+nints
                     nprim_ints = nprim_ints+neris_tmp
                     IF (cartesian_estimate == 0.0_dp) cartesian_estimate = TINY(cartesian_estimate)
                     estimate_to_store_int = EXPONENT(cartesian_estimate)
                     estimate_to_store_int = MAX(estimate_to_store_int, -15_int_8)
                     cartesian_estimate = SET_EXPONENT(1.0_dp, estimate_to_store_int+1)

                     IF (cartesian_estimate < eps_schwarz) CYCLE

                     !IF(case_index==2) THEN
                     primitive_counter = 0
                     DO llB = 1, nsgfd(lset)
                        DO kkB = 1, nsgfc(kset)
                           DO jjB = 1, nsgfb(jset)
                              DO iiB = 1, nsgfa(iset)
                                 primitive_counter = primitive_counter+1
                                 MNRS(llB, kkB, jjB, iiB) = primitive_integrals(primitive_counter)
                              END DO
                           END DO
                        END DO
                     END DO

                     CALL transform_occupied_orbitals_first(dimen, iatom, jatom, katom, latom, &
                                                            iset, jset, kset, lset, &
                                                            nsgfa(iset), nsgfb(jset), nsgfc(kset), nsgfd(lset), &
                                                            i_batch_start, Ni_occupied, &
                                                            MNRS, C_T, mp2_biel, BI1)
                  ELSE
                     task_counter_RS(global_counter, 4) = task_counter_RS(global_counter, 4)+1

                     cost_tmp = 0.0_dp
                     cost_tmp = cost_model(nsgfd(lset), nsgfc(kset), nsgfb(jset), nsgfa(iset), &
                                           npgfd(lset), npgfc(kset), npgfb(jset), npgfa(iset), &
                                           max_val2/log10_eps_schwarz, &
                                           p1_energy, p2_energy, p3_energy)
                     cost_RS(global_counter) = cost_RS(global_counter)+cost_tmp
                  END IF

               END DO ! i_set_list_kl
            END DO ! i_list_kl

            IF (case_index == 2) THEN
               my_num_call_sec_transf = my_num_call_sec_transf+1
               IF (.NOT. alpha_beta_case) THEN
                  IF (.NOT. mp2_env%direct_canonical%big_send) THEN
                     CALL transform_occupied_orbitals_second(dimen, iatom, jatom, iset, jset, &
                                                             nsgfa(iset), nsgfb(jset), Ni_occupied, Nj_occupied, j_batch_start, &
                                                             BI1, C_T, mp2_biel, para_env, elements_ij_proc, &
                                                             multiple, proc_map, BIb)
                  ELSE
                     CALL transform_occupied_orbitals_second_big( &
                        dimen, iatom, jatom, iset, jset, &
                        nsgfa(iset), nsgfb(jset), Ni_occupied, Nj_occupied, j_batch_start, &
                        ij_elem_max, BI1, C_T, mp2_biel, para_env, elements_ij_proc, &
                        proc_map, BIb)
                  END IF
               ELSE
                  IF (.NOT. mp2_env%direct_canonical%big_send) THEN
                     CALL transform_occupied_orbitals_second(dimen, iatom, jatom, iset, jset, &
                                                             nsgfa(iset), nsgfb(jset), Ni_occupied, Nj_occupied, j_batch_start, &
                                                             BI1, C_beta_T, mp2_biel, para_env, elements_ij_proc, &
                                                             multiple, proc_map, BIb)
                  ELSE
                     CALL transform_occupied_orbitals_second_big( &
                        dimen, iatom, jatom, iset, jset, &
                        nsgfa(iset), nsgfb(jset), Ni_occupied, Nj_occupied, j_batch_start, &
                        ij_elem_max, BI1, C_beta_T, mp2_biel, para_env, elements_ij_proc, &
                        proc_map, BIb)
                  END IF
               END IF
            END IF

            !!!!!!! END DO !i_set_list_ij
         END DO !i_list_ij

         IF (case_index == 1) THEN
            CALL mp_sum(task_counter_RS, para_env%group)
            CALL mp_sum(cost_RS, para_env%group)
            ALLOCATE (task_counter_RS_temp(total_num_RS_task, 4))

            ALLOCATE (cost_RS_temp(total_num_RS_task))

            step_size = 1
            ALLOCATE (same_size_kl_elements_counter((nsgf_max**2+1)/step_size+1))

            same_size_kl_elements_counter = 0

            same_size_kl_index = 0
            global_counter = 0
            DO iiB = nsgf_max**2+1, 0, -step_size
               DO jjB = 1, total_num_RS_task
                  IF (task_counter_RS(jjB, 3) > iiB-step_size .AND. task_counter_RS(jjB, 3) <= iiB) THEN
                     global_counter = global_counter+1
                     task_counter_RS_temp(global_counter, 1:4) = task_counter_RS(jjB, 1:4)
                     cost_RS_temp(global_counter) = cost_RS(jjB)
                  END IF
               END DO
               same_size_kl_index = same_size_kl_index+1
               same_size_kl_elements_counter(same_size_kl_index) = global_counter
            END DO

            DEALLOCATE (task_counter_RS)
            DEALLOCATE (cost_RS)

            i_start = 1
            DO same_size_kl_index = 1, SIZE(same_size_kl_elements_counter)
               DO iiB = i_start, same_size_kl_elements_counter(same_size_kl_index)
                  DO jjB = iiB+1, same_size_kl_elements_counter(same_size_kl_index)

                     IF (cost_RS_temp(jjB) >= cost_RS_temp(iiB)) THEN
                        RS_counter_temp = task_counter_RS_temp(iiB, 1:4)
                        task_counter_RS_temp(iiB, 1:4) = task_counter_RS_temp(jjB, 1:4)
                        task_counter_RS_temp(jjB, 1:4) = RS_counter_temp

                        cost_tmp = cost_RS_temp(iiB)
                        cost_RS_temp(iiB) = cost_RS_temp(jjB)
                        cost_RS_temp(jjB) = cost_tmp
                     END IF
                  END DO
               END DO
               i_start = same_size_kl_elements_counter(same_size_kl_index)+1
            END DO

            proc_num_task = 0
            DO counter_proc = 1, total_num_RS_task
               proc_num = MOD(counter_proc, para_env%num_pe)
               proc_num_task(proc_num) = proc_num_task(proc_num)+1
            END DO

            max_num_call_sec_transf = MAXVAL(proc_num_task)

            DEALLOCATE (kl_list_proc)
            ALLOCATE (kl_list_proc(proc_num_task(para_env%mepos), 2))

            kl_list_proc = 0

            elements_kl_proc = 0
            DO counter_proc = 1, total_num_RS_task
               proc_num = MOD(counter_proc, para_env%num_pe)
               IF (proc_num == para_env%mepos) THEN
                  elements_kl_proc = elements_kl_proc+1
                  kl_list_proc(elements_kl_proc, 1) = task_counter_RS_temp(counter_proc, 1)
                  kl_list_proc(elements_kl_proc, 2) = task_counter_RS_temp(counter_proc, 2)
               END IF
            END DO

            DEALLOCATE (task_counter_RS_temp)
            DEALLOCATE (cost_RS_temp)
         END IF
      END DO ! case_index

      size_parameter_send(1) = 1
      size_parameter_send(2) = 1
      size_parameter_send(3) = 0
      size_parameter_send(4) = 0
      size_parameter_send(5) = elements_ij_proc

      IF (mp2_env%direct_canonical%big_send) THEN
         ALLOCATE (zero_mat_big(dimen, 2, ij_elem_max))

      END IF

      DO iiB = my_num_call_sec_transf+1, max_num_call_sec_transf
         DO index_proc_shift = 0, para_env%num_pe-1

            proc_send = proc_map(para_env%mepos+index_proc_shift)
            proc_receive = proc_map(para_env%mepos-index_proc_shift)

            !DO index_proc_shift=0, para_env%num_pe-1
            !  proc_send=proc_map(0-para_env%mepos+index_proc_shift)
            !  proc_receive=proc_send

            IF (proc_send /= para_env%mepos) THEN
               ! the processor starts to send (and receive?)

               CALL mp_sendrecv(size_parameter_send, proc_send, size_parameter_rec, proc_receive, para_env%group)

               Rsize_rec = size_parameter_rec(1)
               Ssize_rec = size_parameter_rec(2)
               R_offset_rec = size_parameter_rec(3)
               S_offset_rec = size_parameter_rec(4)
               elements_ij_proc_rec = size_parameter_rec(5)
               IF (.NOT. mp2_env%direct_canonical%big_send) THEN
                  ALLOCATE (BIb_RS_mat_rec(dimen, Rsize_rec+Ssize_rec))
               ELSE
                  ALLOCATE (BIb_RS_mat_rec_big(dimen, Rsize_rec+Ssize_rec, ij_elem_max))
               END IF
            ELSE
               elements_ij_proc_rec = elements_ij_proc
            END IF

            IF (.NOT. mp2_env%direct_canonical%big_send) THEN
               index_ij_send = 0
               index_ij_rec = 0
               DO index_proc_ij = proc_send+1, multiple, para_env%num_pe

                  zero_mat = 0.D+00
                  case_send_receive = -1
                  IF (proc_send == para_env%mepos) THEN
                     case_send_receive = 0
                  ELSE
                     case_send_receive = 1
                  END IF

                  SELECT CASE (case_send_receive)
                  CASE (0)
                     ! do nothing
                  CASE (1)

                     CALL mp_sendrecv(zero_mat, proc_send, BIb_RS_mat_rec, proc_receive, para_env%group)

                     index_ij_rec = index_ij_rec+1
                     IF (index_ij_rec <= elements_ij_proc .AND. elements_ij_proc > 0) THEN

                        BIb(1:dimen, R_offset_rec+1:R_offset_rec+Rsize_rec, index_ij_rec) = &
                           BIb(1:dimen, R_offset_rec+1:R_offset_rec+Rsize_rec, index_ij_rec)+ &
                           BIb_RS_mat_rec(1:dimen, 1:Rsize_rec)

                        BIb(1:dimen, S_offset_rec+1:S_offset_rec+Ssize_rec, index_ij_rec) = &
                           BIb(1:dimen, S_offset_rec+1:S_offset_rec+Ssize_rec, index_ij_rec)+ &
                           BIb_RS_mat_rec(1:dimen, Rsize_rec+1:Rsize_rec+Ssize_rec)

                     END IF
                  CASE DEFAULT
                     CALL cp__b("mp2_direct_method.F",1034,"")
                  END SELECT

               END DO
            ELSE
               zero_mat_big = 0.D+00
               case_send_receive = -1
               IF (proc_send == para_env%mepos) THEN
                  case_send_receive = 0
               ELSE
                  case_send_receive = 1
               END IF

               SELECT CASE (case_send_receive)
               CASE (0)
                  ! do nothing
               CASE (1)

                  CALL mp_sendrecv(zero_mat_big, proc_send, BIb_RS_mat_rec_big, proc_receive, para_env%group)

                  BIb(1:dimen, R_offset_rec+1:R_offset_rec+Rsize_rec, 1:elements_ij_proc) = &
                     BIb(1:dimen, R_offset_rec+1:R_offset_rec+Rsize_rec, 1:elements_ij_proc)+ &
                     BIb_RS_mat_rec_big(1:dimen, 1:Rsize_rec, 1:elements_ij_proc)

                  BIb(1:dimen, S_offset_rec+1:S_offset_rec+Ssize_rec, 1:elements_ij_proc) = &
                     BIb(1:dimen, S_offset_rec+1:S_offset_rec+Ssize_rec, 1:elements_ij_proc)+ &
                     BIb_RS_mat_rec_big(1:dimen, Rsize_rec+1:Rsize_rec+Ssize_rec, 1:elements_ij_proc)

               CASE DEFAULT
                  CALL cp__b("mp2_direct_method.F",1063,"")
               END SELECT
            END IF

            IF (proc_send /= para_env%mepos) THEN
               IF (.NOT. mp2_env%direct_canonical%big_send) THEN
                  DEALLOCATE (BIb_RS_mat_rec)
               ELSE
                  DEALLOCATE (BIb_RS_mat_rec_big)
               END IF
            END IF

         END DO
      END DO

      IF (mp2_env%direct_canonical%big_send) THEN
         DEALLOCATE (zero_mat_big)
      END IF

      logger => cp_get_default_logger()

      DEALLOCATE (primitive_integrals)

      IF (.NOT. alpha_beta_case) THEN
         CALL transform_virtual_orbitals_and_accumulate(dimen, occupied, dimen-occupied, i_batch_start, &
                                                        j_batch_start, BIb, C, Auto, elements_ij_proc, ij_list_proc, &
                                                        nspins, Emp2, Emp2_Cou, Emp2_ex)
      ELSE
         CALL transform_virtual_orbitals_and_accumulate_ABcase( &
            dimen, occupied, occupied_beta, dimen-occupied, dimen-occupied_beta, &
            i_batch_start, j_batch_start, &
            BIb, C, C_beta, Auto, Auto_beta, &
            elements_ij_proc, ij_list_proc, Emp2, Emp2_Cou)
         DEALLOCATE (C_beta_T)
      END IF

      IF (copy_integrals) THEN
         IF (.NOT. alpha_beta_case) THEN
            ALLOCATE (Integ_MP2(dimen-occupied, dimen-occupied, occupied, occupied))
            Integ_MP2 = 0.0_dp
            DO i = 1, elements_ij_proc
               iiB = ij_list_proc(i, 1)
               jjB = ij_list_proc(i, 2)
               Integ_MP2(:, :, iiB+i_batch_start, jjB+j_batch_start) = BIb(1:dimen-occupied, 1:dimen-occupied, i)
            END DO
         ELSE
            ALLOCATE (Integ_MP2(dimen-occupied, dimen-occupied_beta, occupied, occupied_beta))
            Integ_MP2 = 0.0_dp
            DO i = 1, elements_ij_proc
               iiB = ij_list_proc(i, 1)
               jjB = ij_list_proc(i, 2)
               Integ_MP2(:, :, iiB+i_batch_start, jjB+j_batch_start) = BIb(1:dimen-occupied, 1:dimen-occupied_beta, i)
            END DO
         END IF
      END IF
      DEALLOCATE (BIb)

      DEALLOCATE (set_list_ij, set_list_kl)

      DO i = 1, max_pgf**2
         DEALLOCATE (pgf_list_ij(i)%image_list)
         DEALLOCATE (pgf_list_kl(i)%image_list)
      END DO

      DEALLOCATE (pgf_list_ij)
      DEALLOCATE (pgf_list_kl)
      DEALLOCATE (pgf_product_list)

      DEALLOCATE (max_contraction, kind_of)

      DEALLOCATE (ee_work, ee_work2, ee_buffer1, ee_buffer2, ee_primitives_tmp)

      DEALLOCATE (nimages)

      IF (mp2_env%potential_parameter%potential_type == do_mp2_potential_TShPSC) THEN
         init_TShPSC_lmax = -1
         CALL free_C0()
      END IF

      CALL timestop(handle)

   END SUBROUTINE mp2_canonical_direct_single_batch

! **************************************************************************************************
!> \brief ...
!> \param dimen ...
!> \param latom ...
!> \param katom ...
!> \param jatom ...
!> \param iatom ...
!> \param lset ...
!> \param kset ...
!> \param jset ...
!> \param iset ...
!> \param Ssize ...
!> \param Rsize ...
!> \param Nsize ...
!> \param Msize ...
!> \param i_batch_start ...
!> \param Ni_occupied ...
!> \param MNRS ...
!> \param C_T ...
!> \param mp2_biel ...
!> \param BI1 ...
! **************************************************************************************************
   SUBROUTINE transform_occupied_orbitals_first(dimen, latom, katom, jatom, iatom, &
                                                lset, kset, jset, iset, &
                                                Ssize, Rsize, Nsize, Msize, &
                                                i_batch_start, Ni_occupied, &
                                                MNRS, C_T, mp2_biel, BI1)

      INTEGER                                            :: dimen, latom, katom, jatom, iatom, lset, &
                                                            kset, jset, iset, Ssize, Rsize, Nsize, &
                                                            Msize, i_batch_start, Ni_occupied
      REAL(KIND=dp), &
         DIMENSION(Msize, Nsize, Rsize, Ssize)           :: MNRS
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C_T
      TYPE(mp2_biel_type)                                :: mp2_biel
      REAL(KIND=dp), &
         DIMENSION(dimen, Ni_occupied, Rsize, Ssize)     :: BI1

      CHARACTER(LEN=*), PARAMETER :: routineN = 'transform_occupied_orbitals_first', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i_global, m, M_global, M_offset, &
                                                            M_start, n, N_global, N_offset, r, &
                                                            R_offset, R_start, s, S_offset
      REAL(KIND=dp)                                      :: MNRS_element

      N_offset = mp2_biel%index_table(jatom, jset)-1
      M_offset = mp2_biel%index_table(iatom, iset)-1
      S_offset = mp2_biel%index_table(latom, lset)-1
      R_offset = mp2_biel%index_table(katom, kset)-1

      DO S = 1, Ssize
         R_start = 1
         IF (katom == latom .AND. kset == lset) R_start = S
         DO R = R_start, Rsize

            ! fast i don't know why
            DO N = 1, Nsize
               N_global = N+N_offset
               M_start = 1
               IF (iatom == jatom .AND. iset == jset) THEN
                  M = N
                  M_global = M+M_offset
                  MNRS_element = MNRS(M, N, R, S)
                  DO i = 1, Ni_occupied
                     i_global = i+i_batch_start
                     BI1(N_global, i, R, S) = BI1(N_global, i, R, S)+C_T(i_global, M_global)*MNRS_element
                  END DO
                  M_start = N+1
               END IF

               DO M = M_start, Msize
                  M_global = M+M_offset
                  MNRS_element = MNRS(M, N, R, S)
                  DO i = 1, Ni_occupied
                     i_global = i+i_batch_start
                     BI1(N_global, i, R, S) = BI1(N_global, i, R, S)+C_T(i_global, M_global)*MNRS_element
                     BI1(M_global, i, R, S) = BI1(M_global, i, R, S)+C_T(i_global, N_global)*MNRS_element
                  END DO
               END DO
            END DO

         END DO
      END DO

   END SUBROUTINE transform_occupied_orbitals_first

! **************************************************************************************************
!> \brief ...
!> \param dimen ...
!> \param latom ...
!> \param katom ...
!> \param lset ...
!> \param kset ...
!> \param Ssize ...
!> \param Rsize ...
!> \param Ni_occupied ...
!> \param Nj_occupied ...
!> \param j_batch_start ...
!> \param BI1 ...
!> \param C_T ...
!> \param mp2_biel ...
!> \param para_env ...
!> \param elements_ij_proc ...
!> \param multiple ...
!> \param proc_map ...
!> \param BIb ...
! **************************************************************************************************
   SUBROUTINE transform_occupied_orbitals_second(dimen, latom, katom, lset, kset, &
                                                 Ssize, Rsize, Ni_occupied, Nj_occupied, j_batch_start, &
                                                 BI1, C_T, mp2_biel, para_env, &
                                                 elements_ij_proc, &
                                                 multiple, proc_map, BIb)

      INTEGER                                            :: dimen, latom, katom, lset, kset, Ssize, &
                                                            Rsize, Ni_occupied, Nj_occupied, &
                                                            j_batch_start
      REAL(KIND=dp), &
         DIMENSION(dimen, Ni_occupied, Rsize, Ssize)     :: BI1
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C_T
      TYPE(mp2_biel_type)                                :: mp2_biel
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: elements_ij_proc, multiple
      INTEGER, DIMENSION((-para_env%num_pe):(2*para_env%&
         num_pe-1))                                      :: proc_map
      REAL(KIND=dp), &
         DIMENSION(dimen, dimen, elements_ij_proc)       :: BIb

      CHARACTER(LEN=*), PARAMETER :: routineN = 'transform_occupied_orbitals_second', &
         routineP = moduleN//':'//routineN

      INTEGER :: case_send_receive, elements_ij_proc_rec, handle, i, index_ij_rec, index_ij_send, &
         index_proc_ij, index_proc_shift, j, n, proc_receive, proc_send, r, R_global, R_offset, &
         R_offset_rec, R_start, Rsize_rec, s, S_global, S_offset, S_offset_rec, Ssize_rec
      REAL(KIND=dp), DIMENSION(dimen, Rsize+Ssize)       :: BIb_RS_mat
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIb_RS_mat_rec
      REAL(KIND=dp)                                      :: C_T_R, C_T_S
      INTEGER, DIMENSION(5)                              :: size_parameter_rec, size_parameter_send

      CALL timeset(routineN, handle)

      S_offset = mp2_biel%index_table(latom, lset)-1
      R_offset = mp2_biel%index_table(katom, kset)-1

      size_parameter_send(1) = Rsize
      size_parameter_send(2) = Ssize
      size_parameter_send(3) = R_offset
      size_parameter_send(4) = S_offset
      size_parameter_send(5) = elements_ij_proc

      DO index_proc_shift = 0, para_env%num_pe-1

         proc_send = proc_map(para_env%mepos+index_proc_shift)
         proc_receive = proc_map(para_env%mepos-index_proc_shift)

         IF (proc_send /= para_env%mepos) THEN
            ! the processor starts to send (and receive?)

            CALL mp_sendrecv(size_parameter_send, proc_send, size_parameter_rec, proc_receive, para_env%group)

            Rsize_rec = size_parameter_rec(1)
            Ssize_rec = size_parameter_rec(2)
            R_offset_rec = size_parameter_rec(3)
            S_offset_rec = size_parameter_rec(4)
            elements_ij_proc_rec = size_parameter_rec(5)
            ALLOCATE (BIb_RS_mat_rec(dimen, Rsize_rec+Ssize_rec))

         ELSE
            elements_ij_proc_rec = elements_ij_proc
         END IF

         case_send_receive = -1
         IF (proc_send == para_env%mepos) THEN
            case_send_receive = 0
         ELSE
            case_send_receive = 1
         END IF

         index_ij_send = 0
         index_ij_rec = 0
         DO index_proc_ij = proc_send+1, multiple, para_env%num_pe

            BIb_RS_mat = zero
            IF (index_proc_ij <= Ni_occupied*Nj_occupied) THEN

               index_ij_send = index_ij_send+1

               i = (index_proc_ij-1)/Nj_occupied+1
               j = index_proc_ij-(i-1)*Nj_occupied+j_batch_start

               DO S = 1, Ssize
                  S_global = S+S_offset
                  R_start = 1
                  IF (katom == latom .AND. kset == lset) R_start = S
                  DO R = R_start, Rsize
                     R_global = R+R_offset

                     IF (R_global /= S_global) THEN
                        C_T_R = C_T(j, R_global)
                        C_T_S = C_T(j, S_global)
                        DO N = 1, dimen
                           BIb_RS_mat(N, R) = BIb_RS_mat(N, R)+C_T_S*BI1(N, i, R, S)
                        END DO
                        DO N = 1, dimen
                           BIb_RS_mat(N, Rsize+S) = BIb_RS_mat(N, Rsize+S)+C_T_R*BI1(N, i, R, S)
                        END DO
                     ELSE
                        C_T_S = C_T(j, S_global)
                        DO N = 1, dimen
                           BIb_RS_mat(N, R) = BIb_RS_mat(N, R)+C_T_S*BI1(N, i, R, S)
                        END DO
                     END IF

                  END DO
               END DO

            END IF

            SELECT CASE (case_send_receive)
            CASE (0)
               ! the processor is the sender and receiver itself
               IF (index_ij_send <= elements_ij_proc .AND. elements_ij_proc > 0) THEN

                  BIb(1:dimen, R_offset+1:R_offset+Rsize, index_ij_send) = &
                     BIb(1:dimen, R_offset+1:R_offset+Rsize, index_ij_send)+BIb_RS_mat(1:dimen, 1:Rsize)

                  BIb(1:dimen, S_offset+1:S_offset+Ssize, index_ij_send) = &
                     BIb(1:dimen, S_offset+1:S_offset+Ssize, index_ij_send)+BIb_RS_mat(1:dimen, Rsize+1:Rsize+Ssize)

               END IF
            CASE (1)

               CALL mp_sendrecv(BIb_RS_mat, proc_send, BIb_RS_mat_rec, proc_receive, para_env%group)

               index_ij_rec = index_ij_rec+1
               IF (index_ij_rec <= elements_ij_proc .AND. elements_ij_proc > 0) THEN

                  BIb(1:dimen, R_offset_rec+1:R_offset_rec+Rsize_rec, index_ij_rec) = &
                     BIb(1:dimen, R_offset_rec+1:R_offset_rec+Rsize_rec, index_ij_rec)+ &
                     BIb_RS_mat_rec(1:dimen, 1:Rsize_rec)

                  BIb(1:dimen, S_offset_rec+1:S_offset_rec+Ssize_rec, index_ij_rec) = &
                     BIb(1:dimen, S_offset_rec+1:S_offset_rec+Ssize_rec, index_ij_rec)+ &
                     BIb_RS_mat_rec(1:dimen, Rsize_rec+1:Rsize_rec+Ssize_rec)

               END IF
            CASE DEFAULT
               CALL cp__b("mp2_direct_method.F",1393,"")
            END SELECT

         END DO ! loop over the ij of the processor

         IF (proc_send /= para_env%mepos) THEN
            DEALLOCATE (BIb_RS_mat_rec)
         END IF

      END DO ! loop over the processor starting from itself

      CALL timestop(handle)

   END SUBROUTINE transform_occupied_orbitals_second

! **************************************************************************************************
!> \brief ...
!> \param dimen ...
!> \param latom ...
!> \param katom ...
!> \param lset ...
!> \param kset ...
!> \param Ssize ...
!> \param Rsize ...
!> \param Ni_occupied ...
!> \param Nj_occupied ...
!> \param j_batch_start ...
!> \param ij_elem_max ...
!> \param BI1 ...
!> \param C_T ...
!> \param mp2_biel ...
!> \param para_env ...
!> \param elements_ij_proc ...
!> \param proc_map ...
!> \param BIb ...
! **************************************************************************************************
   SUBROUTINE transform_occupied_orbitals_second_big(dimen, latom, katom, lset, kset, &
                                                     Ssize, Rsize, Ni_occupied, Nj_occupied, j_batch_start, &
                                                     ij_elem_max, BI1, C_T, mp2_biel, para_env, &
                                                     elements_ij_proc, &
                                                     proc_map, BIb)

      INTEGER                                            :: dimen, latom, katom, lset, kset, Ssize, &
                                                            Rsize, Ni_occupied, Nj_occupied, &
                                                            j_batch_start, ij_elem_max
      REAL(KIND=dp), &
         DIMENSION(dimen, Ni_occupied, Rsize, Ssize)     :: BI1
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C_T
      TYPE(mp2_biel_type)                                :: mp2_biel
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: elements_ij_proc
      INTEGER, DIMENSION((-para_env%num_pe):(2*para_env%&
         num_pe-1))                                      :: proc_map
      REAL(KIND=dp), &
         DIMENSION(dimen, dimen, elements_ij_proc)       :: BIb

      CHARACTER(LEN=*), PARAMETER :: routineN = 'transform_occupied_orbitals_second_big', &
         routineP = moduleN//':'//routineN

      INTEGER :: case_send_receive, elements_ij_proc_rec, handle, i, index_ij_rec, index_ij_send, &
         index_proc_ij, index_proc_shift, j, n, proc_receive, proc_send, r, R_global, R_offset, &
         R_offset_rec, R_start, Rsize_rec, s, S_global, S_offset, S_offset_rec, Ssize_rec
      REAL(KIND=dp), &
         DIMENSION(dimen, Rsize+Ssize, ij_elem_max)      :: BIb_RS_mat
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_RS_mat_rec
      REAL(KIND=dp)                                      :: C_T_R, C_T_S
      INTEGER, DIMENSION(5)                              :: size_parameter_rec, size_parameter_send

      CALL timeset(routineN, handle)

      S_offset = mp2_biel%index_table(latom, lset)-1
      R_offset = mp2_biel%index_table(katom, kset)-1

      size_parameter_send(1) = Rsize
      size_parameter_send(2) = Ssize
      size_parameter_send(3) = R_offset
      size_parameter_send(4) = S_offset
      size_parameter_send(5) = elements_ij_proc

      DO index_proc_shift = 0, para_env%num_pe-1

         proc_send = proc_map(para_env%mepos+index_proc_shift)
         proc_receive = proc_map(para_env%mepos-index_proc_shift)

         IF (proc_send /= para_env%mepos) THEN
            ! the processor starts to send (and receive?)

            CALL mp_sendrecv(size_parameter_send, proc_send, size_parameter_rec, proc_receive, para_env%group)

            Rsize_rec = size_parameter_rec(1)
            Ssize_rec = size_parameter_rec(2)
            R_offset_rec = size_parameter_rec(3)
            S_offset_rec = size_parameter_rec(4)
            elements_ij_proc_rec = size_parameter_rec(5)
            ALLOCATE (BIb_RS_mat_rec(dimen, Rsize_rec+Ssize_rec, ij_elem_max))
         ELSE
            elements_ij_proc_rec = elements_ij_proc
         END IF

         index_ij_send = 0
         index_ij_rec = 0
         BIb_RS_mat = zero

         case_send_receive = -1
         IF (proc_send == para_env%mepos) THEN
            case_send_receive = 0
         ELSE
            case_send_receive = 1
         END IF

         DO index_proc_ij = proc_send+1, Ni_occupied*Nj_occupied, para_env%num_pe

            index_ij_send = index_ij_send+1

            i = (index_proc_ij-1)/Nj_occupied+1
            j = index_proc_ij-(i-1)*Nj_occupied+j_batch_start

            DO S = 1, Ssize
               S_global = S+S_offset
               R_start = 1
               IF (katom == latom .AND. kset == lset) R_start = S
               DO R = R_start, Rsize
                  R_global = R+R_offset

                  IF (R_global /= S_global) THEN
                     C_T_R = C_T(j, R_global)
                     C_T_S = C_T(j, S_global)
                     DO N = 1, dimen
                        BIb_RS_mat(N, R, index_ij_send) = BIb_RS_mat(N, R, index_ij_send)+C_T_S*BI1(N, i, R, S)
                     END DO
                     DO N = 1, dimen
                        BIb_RS_mat(N, Rsize+S, index_ij_send) = BIb_RS_mat(N, Rsize+S, index_ij_send)+C_T_R*BI1(N, i, R, S)
                     END DO
                  ELSE
                     C_T_S = C_T(j, S_global)
                     DO N = 1, dimen
                        BIb_RS_mat(N, R, index_ij_send) = BIb_RS_mat(N, R, index_ij_send)+C_T_S*BI1(N, i, R, S)
                     END DO
                  END IF

               END DO
            END DO

         END DO

         SELECT CASE (case_send_receive)
         CASE (0)
            ! the processor is the sender and receiver itself
            BIb(1:dimen, R_offset+1:R_offset+Rsize, 1:elements_ij_proc) = &
               BIb(1:dimen, R_offset+1:R_offset+Rsize, 1:elements_ij_proc)+ &
               BIb_RS_mat(1:dimen, 1:Rsize, 1:elements_ij_proc)

            BIb(1:dimen, S_offset+1:S_offset+Ssize, 1:elements_ij_proc) = &
               BIb(1:dimen, S_offset+1:S_offset+Ssize, 1:elements_ij_proc)+ &
               BIb_RS_mat(1:dimen, Rsize+1:Rsize+Ssize, 1:elements_ij_proc)

         CASE (1)

            CALL mp_sendrecv(BIb_RS_mat, proc_send, BIb_RS_mat_rec, proc_receive, para_env%group)

            BIb(1:dimen, R_offset_rec+1:R_offset_rec+Rsize_rec, 1:elements_ij_proc) = &
               BIb(1:dimen, R_offset_rec+1:R_offset_rec+Rsize_rec, 1:elements_ij_proc)+ &
               BIb_RS_mat_rec(1:dimen, 1:Rsize_rec, 1:elements_ij_proc)

            BIb(1:dimen, S_offset_rec+1:S_offset_rec+Ssize_rec, 1:elements_ij_proc) = &
               BIb(1:dimen, S_offset_rec+1:S_offset_rec+Ssize_rec, 1:elements_ij_proc)+ &
               BIb_RS_mat_rec(1:dimen, Rsize_rec+1:Rsize_rec+Ssize_rec, 1:elements_ij_proc)

         CASE DEFAULT
            CALL cp__b("mp2_direct_method.F",1562,"")
         END SELECT

         IF (proc_send /= para_env%mepos) THEN
            DEALLOCATE (BIb_RS_mat_rec)
         END IF

      END DO ! loop over the processor starting from itself

      CALL timestop(handle)

   END SUBROUTINE transform_occupied_orbitals_second_big

! **************************************************************************************************
!> \brief ...
!> \param dimen ...
!> \param occupied ...
!> \param virtual ...
!> \param i_batch_start ...
!> \param j_batch_start ...
!> \param BIb ...
!> \param C ...
!> \param Auto ...
!> \param elements_ij_proc ...
!> \param ij_list_proc ...
!> \param nspins ...
!> \param Emp2 ...
!> \param Emp2_Cou ...
!> \param Emp2_ex ...
! **************************************************************************************************
   SUBROUTINE transform_virtual_orbitals_and_accumulate(dimen, occupied, virtual, i_batch_start, &
                                                        j_batch_start, BIb, C, Auto, elements_ij_proc, &
                                                        ij_list_proc, nspins, Emp2, Emp2_Cou, Emp2_ex)

      INTEGER                                            :: dimen, occupied, virtual, i_batch_start, &
                                                            j_batch_start
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C
      REAL(KIND=dp), DIMENSION(dimen)                    :: Auto
      INTEGER                                            :: elements_ij_proc
      INTEGER, DIMENSION(elements_ij_proc, 2)            :: ij_list_proc
      INTEGER                                            :: nspins
      REAL(KIND=dp)                                      :: Emp2, Emp2_Cou, Emp2_ex

      CHARACTER(LEN=*), PARAMETER :: routineN = 'transform_virtual_orbitals_and_accumulate', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      INTEGER                                            :: a, a_global, b, b_global, handle, i, &
                                                            i_global, index_ij, j, j_global
      REAL(KIND=dp)                                      :: iajb, ibja, parz, two
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIa

      CALL timeset(routineN, handle)

      ALLOCATE (BIa(dimen, virtual))

      BIa = zero
      DO index_ij = 1, elements_ij_proc

         CALL DGEMM('T', 'N', dimen, virtual, dimen, 1.0_dp, Bib(1, 1, index_ij), &
                    dimen, C(1, occupied+1), dimen, 0.0_dp, Bia(1, 1), dimen)
         Bib(1:dimen, 1:virtual, index_ij) = Bia(1:dimen, 1:virtual)

      END DO

      DEALLOCATE (BIa)
      ALLOCATE (BIa(virtual, virtual))

      BIa = zero
      DO index_ij = 1, elements_ij_proc

         CALL DGEMM('T', 'N', virtual, virtual, dimen, 1.0_dp, Bib(1, 1, index_ij), dimen, C(1, occupied+1), dimen, 0.0_dp, &
                    BIa(1, 1), virtual)
         BIb(1:virtual, 1:virtual, index_ij) = BIa(1:virtual, 1:virtual)

      END DO

      two = 2.0_dp/nspins
      DO index_ij = 1, elements_ij_proc
         i = ij_list_proc(index_ij, 1)
         j = ij_list_proc(index_ij, 2)
         i_global = i+i_batch_start
         j_global = j+j_batch_start
         DO a = 1, virtual
            a_global = a+occupied
            DO b = 1, virtual
               b_global = b+occupied
               iajb = BIb(a, b, index_ij)
               ibja = BIb(b, a, index_ij)
               parz = iajb/(Auto(i_global)+Auto(j_global)-Auto(a_global)-Auto(b_global))
               ! parz=parz*(two*iajb-ibja)   !Full
               ! parz=parz*(iajb)            !Coulomb
               ! parz=parz*(ibja)            !Coulomb
               ! Emp2=Emp2+parz/nspins
               Emp2_Cou = Emp2_Cou+parz*two*(iajb)/nspins
               Emp2_ex = Emp2_ex-parz*(ibja)/nspins
               Emp2 = Emp2+parz*(two*iajb-ibja)/nspins
            END DO
         END DO
      END DO

      DEALLOCATE (BIa)

      CALL timestop(handle)

   END SUBROUTINE transform_virtual_orbitals_and_accumulate

! **************************************************************************************************
!> \brief ...
!> \param dimen ...
!> \param occ_i ...
!> \param occ_j ...
!> \param virt_i ...
!> \param virt_j ...
!> \param i_batch_start ...
!> \param j_batch_start ...
!> \param BIb ...
!> \param C_i ...
!> \param C_j ...
!> \param Auto_i ...
!> \param Auto_j ...
!> \param elements_ij_proc ...
!> \param ij_list_proc ...
!> \param Emp2 ...
!> \param Emp2_Cou ...
! **************************************************************************************************
   SUBROUTINE transform_virtual_orbitals_and_accumulate_ABcase(dimen, occ_i, occ_j, virt_i, virt_j, i_batch_start, &
                                                               j_batch_start, BIb, C_i, C_j, Auto_i, Auto_j, elements_ij_proc, &
                                                               ij_list_proc, Emp2, Emp2_Cou)

      INTEGER                                            :: dimen, occ_i, occ_j, virt_i, virt_j, &
                                                            i_batch_start, j_batch_start
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C_i, C_j
      REAL(KIND=dp), DIMENSION(dimen)                    :: Auto_i, Auto_j
      INTEGER                                            :: elements_ij_proc
      INTEGER, DIMENSION(elements_ij_proc, 2)            :: ij_list_proc
      REAL(KIND=dp)                                      :: Emp2, Emp2_Cou

      CHARACTER(LEN=*), PARAMETER :: &
         routineN = 'transform_virtual_orbitals_and_accumulate_ABcase', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: two = 2.D+00, zero = 0.D+00

      INTEGER                                            :: a, a_global, b, b_global, handle, i, &
                                                            i_global, index_ij, j, j_global, n, s
      REAL(KIND=dp)                                      :: iajb, parz
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIa

      CALL timeset(routineN, handle)

      ALLOCATE (BIa(dimen, virt_i))

      DO index_ij = 1, elements_ij_proc

         DO a = 1, virt_i
            a_global = a+occ_i
            DO S = 1, dimen
               parz = zero
               DO N = 1, dimen
                  parz = parz+C_i(N, a_global)*BIb(N, S, index_ij)
               END DO
               BIa(S, a) = parz
            END DO
         END DO

         BIb(1:dimen, 1:virt_i, index_ij) = BIa

      END DO

      DEALLOCATE (BIa)
      ALLOCATE (BIa(virt_i, virt_j))

      DO index_ij = 1, elements_ij_proc

         DO a = 1, virt_i
            DO b = 1, virt_j
               b_global = b+occ_j
               parz = zero
               DO S = 1, dimen
                  parz = parz+C_j(S, b_global)*BIb(S, a, index_ij)
               END DO
               BIa(a, b) = parz
            END DO
         END DO

         BIb(1:virt_i, 1:virt_j, index_ij) = BIa

      END DO

      DO index_ij = 1, elements_ij_proc
         i = ij_list_proc(index_ij, 1)
         j = ij_list_proc(index_ij, 2)
         i_global = i+i_batch_start
         j_global = j+j_batch_start
         DO a = 1, virt_i
            a_global = a+occ_i
            DO b = 1, virt_j
               b_global = b+occ_j
               iajb = BIb(a, b, index_ij)
               parz = iajb*iajb/(Auto_i(i_global)+Auto_j(j_global)-Auto_i(a_global)-Auto_j(b_global))
               ! Emp2=Emp2+parz/two
               Emp2_Cou = Emp2_Cou+parz/two
               Emp2 = Emp2+parz/two
            END DO
         END DO
      END DO

      DEALLOCATE (BIa)

      CALL timestop(handle)

   END SUBROUTINE transform_virtual_orbitals_and_accumulate_ABcase

END MODULE mp2_direct_method
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate the 3 and 2 center ERI's needed in the RI
!>        approximation using libint
!> \par History
!>      08.2013 created [Mauro Del Ben]
!> \author Mauro Del Ben
! **************************************************************************************************
MODULE mp2_ri_libint
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type,&
                                              init_aux_basis_set
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                              cp_blacs_env_release,&
                                              cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_fm_basic_linalg,              ONLY: cp_fm_triangular_invert
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_release,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE gamma,                           ONLY: init_md_ftable
   USE hfx_energy_potential,            ONLY: coulomb4
   USE hfx_libint_wrapper_types,        ONLY: lib_int
   USE hfx_pair_list_methods,           ONLY: build_pair_list_mp2
   USE hfx_screening_methods,           ONLY: calc_pair_dist_radii,&
                                              calc_screening_functions
   USE hfx_types,                       ONLY: &
        hfx_basis_info_type, hfx_basis_type, hfx_create_neighbor_cells, hfx_general_type, &
        hfx_load_balance_type, hfx_memory_type, hfx_pgf_list, hfx_pgf_product_list, &
        hfx_potential_type, hfx_screen_coeff_type, hfx_screening_type, hfx_type, log_zero, &
        pair_set_list_type
   USE input_constants,                 ONLY: do_mp2_potential_TShPSC
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE message_passing,                 ONLY: mp_sum
   USE mp2_types,                       ONLY: init_TShPSC_lmax,&
                                              mp2_biel_type,&
                                              mp2_type,&
                                              pair_list_type_mp2
   USE orbital_pointers,                ONLY: nco,&
                                              ncoset,&
                                              nso
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE t_sh_p_s_c,                      ONLY: free_C0,&
                                              init
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC ::  libint_ri_mp2, read_RI_basis_set, release_RI_basis_set

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_ri_libint'

!***

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param dimen ...
!> \param RI_dimen ...
!> \param occupied ...
!> \param natom ...
!> \param mp2_biel ...
!> \param mp2_env ...
!> \param C ...
!> \param kind_of ...
!> \param RI_basis_parameter ...
!> \param RI_basis_info ...
!> \param basis_S0 ...
!> \param RI_index_table ...
!> \param qs_env ...
!> \param para_env ...
!> \param Lai ...
! **************************************************************************************************
   SUBROUTINE libint_ri_mp2(dimen, RI_dimen, occupied, natom, mp2_biel, mp2_env, C, &
                            kind_of, &
                            RI_basis_parameter, RI_basis_info, basis_S0, RI_index_table, &
                            qs_env, para_env, &
                            Lai)
      INTEGER                                            :: dimen, RI_dimen, occupied, natom
      TYPE(mp2_biel_type)                                :: mp2_biel
      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C
      INTEGER, DIMENSION(:)                              :: kind_of
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter
      TYPE(hfx_basis_info_type)                          :: RI_basis_info
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_S0
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: RI_index_table
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: Lai

      CHARACTER(LEN=*), PARAMETER :: routineN = 'libint_ri_mp2', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nkind

      CALL timeset(routineN, handle)

      ! Get the RI basis set and store in to a nice form
      IF (.NOT. (ASSOCIATED(RI_basis_parameter))) THEN
         IF (ALLOCATED(RI_index_table)) DEALLOCATE (RI_index_table)
         IF (ASSOCIATED(basis_S0)) DEALLOCATE (basis_S0)
         CALL read_RI_basis_set(qs_env, RI_basis_parameter, RI_basis_info, &
                                natom, nkind, kind_of, RI_index_table, RI_dimen, &
                                basis_S0)
      END IF

      CALL calc_lai_libint(mp2_env, qs_env, para_env, &
                           mp2_biel, dimen, C, occupied, &
                           RI_basis_parameter, RI_basis_info, RI_index_table, RI_dimen, basis_S0, &
                           Lai)

      CALL timestop(handle)

   END SUBROUTINE libint_ri_mp2

! **************************************************************************************************
!> \brief Read the auxiliary basis set for RI approxiamtion
!> \param qs_env ...
!> \param RI_basis_parameter ...
!> \param RI_basis_info ...
!> \param natom ...
!> \param nkind ...
!> \param kind_of ...
!> \param RI_index_table ...
!> \param RI_dimen ...
!> \param basis_S0 ...
!> \par History
!>      08.2013 created [Mauro Del Ben]
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE read_RI_basis_set(qs_env, RI_basis_parameter, RI_basis_info, &
                                natom, nkind, kind_of, RI_index_table, RI_dimen, &
                                basis_S0)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter
      TYPE(hfx_basis_info_type)                          :: RI_basis_info
      INTEGER                                            :: natom, nkind
      INTEGER, DIMENSION(:)                              :: kind_of
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: RI_index_table
      INTEGER                                            :: RI_dimen
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_S0

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_RI_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER :: co_counter, counter, i, iatom, ikind, ipgf, iset, j, k, la, max_am_kind, &
         max_coeff, max_nsgfl, max_pgf, max_pgf_kind, max_set, nl_count, nset, nseta, offset_a, &
         offset_a1, s_offset_nl_a, sgfa, so_counter
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, npgfa, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, nl_a
      REAL(dp), DIMENSION(:, :), POINTER                 :: sphi_a
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_a
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: atom_kind

      NULLIFY (RI_basis_parameter)

      NULLIFY (qs_kind_set)
      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set)

      nkind = SIZE(qs_kind_set, 1)
      ALLOCATE (RI_basis_parameter(nkind))
      ALLOCATE (basis_S0(nkind))
      max_set = 0
      DO ikind = 1, nkind
         NULLIFY (atom_kind)
         atom_kind => qs_kind_set(ikind)
         ! here we reset the initial RI basis such that we can
         ! work with non-normalized auxiliary basis functions
         CALL get_qs_kind(qs_kind=atom_kind, &
                          basis_set=orb_basis_a, basis_type="RI_AUX")
         IF (.NOT. (ASSOCIATED(orb_basis_a))) THEN
            CALL cp__b("mp2_ri_libint.F",199,"Initial RI auxiliary basis not specified.")
         END IF
         orb_basis_a%gcc = 1.0_dp
         orb_basis_a%norm_type = 1
         CALL init_aux_basis_set(orb_basis_a)

         CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                              maxsgf=RI_basis_info%max_sgf, &
                              maxnset=RI_basis_info%max_set, &
                              maxlgto=RI_basis_info%max_am, &
                              basis_type="RI_AUX")
         CALL get_gto_basis_set(gto_basis_set=orb_basis_a, &
                                lmax=RI_basis_parameter(ikind)%lmax, &
                                lmin=RI_basis_parameter(ikind)%lmin, &
                                npgf=RI_basis_parameter(ikind)%npgf, &
                                nset=RI_basis_parameter(ikind)%nset, &
                                zet=RI_basis_parameter(ikind)%zet, &
                                nsgf_set=RI_basis_parameter(ikind)%nsgf, &
                                first_sgf=RI_basis_parameter(ikind)%first_sgf, &
                                sphi=RI_basis_parameter(ikind)%sphi, &
                                nsgf=RI_basis_parameter(ikind)%nsgf_total, &
                                l=RI_basis_parameter(ikind)%nl, &
                                nshell=RI_basis_parameter(ikind)%nshell, &
                                set_radius=RI_basis_parameter(ikind)%set_radius, &
                                pgf_radius=RI_basis_parameter(ikind)%pgf_radius, &
                                kind_radius=RI_basis_parameter(ikind)%kind_radius)

         max_set = MAX(max_set, RI_basis_parameter(ikind)%nset)

         basis_S0(ikind)%kind_radius = RI_basis_parameter(ikind)%kind_radius
         basis_S0(ikind)%nset = 1
         basis_S0(ikind)%nsgf_total = 1
         ALLOCATE (basis_S0(ikind)%set_radius(1))
         basis_S0(ikind)%set_radius(1) = RI_basis_parameter(ikind)%kind_radius
         ALLOCATE (basis_S0(ikind)%lmax(1))
         basis_S0(ikind)%lmax(1) = 0
         ALLOCATE (basis_S0(ikind)%lmin(1))
         basis_S0(ikind)%lmin(1) = 0
         ALLOCATE (basis_S0(ikind)%npgf(1))
         basis_S0(ikind)%npgf(1) = 1
         ALLOCATE (basis_S0(ikind)%nsgf(1))
         basis_S0(ikind)%nsgf(1) = 1
         ALLOCATE (basis_S0(ikind)%nshell(1))
         basis_S0(ikind)%nshell(1) = 1
         ALLOCATE (basis_S0(ikind)%pgf_radius(1, 1))
         basis_S0(ikind)%pgf_radius(1, 1) = RI_basis_parameter(ikind)%kind_radius
         ALLOCATE (basis_S0(ikind)%sphi(1, 1))
         basis_S0(ikind)%sphi(1, 1) = 1.0_dp
         ALLOCATE (basis_S0(ikind)%zet(1, 1))
         basis_S0(ikind)%zet(1, 1) = 0.0_dp
         ALLOCATE (basis_S0(ikind)%first_sgf(1, 1))
         basis_S0(ikind)%first_sgf(1, 1) = 1
         ALLOCATE (basis_S0(ikind)%nl(1, 1))
         basis_S0(ikind)%nl(1, 1) = 0

         ALLOCATE (basis_S0(ikind)%nsgfl(0:0, 1))
         basis_S0(ikind)%nsgfl = 1
         ALLOCATE (basis_S0(ikind)%sphi_ext(1, 0:0, 1, 1))
         basis_S0(ikind)%sphi_ext(1, 0, 1, 1) = 1.0_dp

      END DO
      RI_basis_info%max_set = max_set

      DO ikind = 1, nkind
         ALLOCATE (RI_basis_parameter(ikind)%nsgfl(0:RI_basis_info%max_am, max_set))
         RI_basis_parameter(ikind)%nsgfl = 0
         nset = RI_basis_parameter(ikind)%nset
         nshell => RI_basis_parameter(ikind)%nshell
         DO iset = 1, nset
            DO i = 0, RI_basis_info%max_am
               nl_count = 0
               DO j = 1, nshell(iset)
                  IF (RI_basis_parameter(ikind)%nl(j, iset) == i) nl_count = nl_count+1
               END DO
               RI_basis_parameter(ikind)%nsgfl(i, iset) = nl_count
            END DO
         END DO
      END DO

      max_nsgfl = 0
      max_pgf = 0
      DO ikind = 1, nkind
         max_coeff = 0
         max_am_kind = 0
         max_pgf_kind = 0
         npgfa => RI_basis_parameter(ikind)%npgf
         nseta = RI_basis_parameter(ikind)%nset
         nl_a => RI_basis_parameter(ikind)%nsgfl
         la_max => RI_basis_parameter(ikind)%lmax
         la_min => RI_basis_parameter(ikind)%lmin
         DO iset = 1, nseta
            max_pgf_kind = MAX(max_pgf_kind, npgfa(iset))
            max_pgf = MAX(max_pgf, npgfa(iset))
            DO la = la_min(iset), la_max(iset)
               max_nsgfl = MAX(max_nsgfl, nl_a(la, iset))
               max_coeff = MAX(max_coeff, nso(la)*nl_a(la, iset)*nco(la))
               max_am_kind = MAX(max_am_kind, la)
            END DO
         END DO
         ALLOCATE (RI_basis_parameter(ikind)%sphi_ext(max_coeff, 0:max_am_kind, max_pgf_kind, nseta))
         RI_basis_parameter(ikind)%sphi_ext = 0.0_dp
      END DO

      DO ikind = 1, nkind
         sphi_a => RI_basis_parameter(ikind)%sphi
         nseta = RI_basis_parameter(ikind)%nset
         la_max => RI_basis_parameter(ikind)%lmax
         la_min => RI_basis_parameter(ikind)%lmin
         npgfa => RI_basis_parameter(ikind)%npgf
         first_sgfa => RI_basis_parameter(ikind)%first_sgf
         nl_a => RI_basis_parameter(ikind)%nsgfl
         DO iset = 1, nseta
            sgfa = first_sgfa(1, iset)
            DO ipgf = 1, npgfa(iset)
               offset_a1 = (ipgf-1)*ncoset(la_max(iset))
               s_offset_nl_a = 0
               DO la = la_min(iset), la_max(iset)
                  offset_a = offset_a1+ncoset(la-1)
                  co_counter = 0
                  co_counter = co_counter+1
                  so_counter = 0
                  DO k = sgfa+s_offset_nl_a, sgfa+s_offset_nl_a+nso(la)*nl_a(la, iset)-1
                     DO i = offset_a+1, offset_a+nco(la)
                        so_counter = so_counter+1
                        RI_basis_parameter(ikind)%sphi_ext(so_counter, la, ipgf, iset) = sphi_a(i, k)
                     END DO
                  END DO
                  s_offset_nl_a = s_offset_nl_a+nso(la)*(nl_a(la, iset))
               END DO
            END DO
         END DO
      END DO

      ALLOCATE (RI_index_table(natom, max_set))
      RI_index_table = -HUGE(0)
      counter = 0
      RI_dimen = 0
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         nset = RI_basis_parameter(ikind)%nset
         DO iset = 1, nset
            RI_index_table(iatom, iset) = counter+1
            counter = counter+RI_basis_parameter(ikind)%nsgf(iset)
            RI_dimen = RI_dimen+RI_basis_parameter(ikind)%nsgf(iset)
         END DO
      END DO

   END SUBROUTINE read_RI_basis_set

! **************************************************************************************************
!> \brief Release the auxiliary basis set for RI approxiamtion (to be used
!>        only in the case of basis optimization)
!> \param RI_basis_parameter ...
!> \param basis_S0 ...
!> \par History
!>      08.2013 created [Mauro Del Ben]
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE release_RI_basis_set(RI_basis_parameter, basis_S0)
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter, basis_S0

      CHARACTER(LEN=*), PARAMETER :: routineN = 'release_RI_basis_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

! RI basis

      DO i = 1, SIZE(RI_basis_parameter)
         DEALLOCATE (RI_basis_parameter(i)%nsgfl)
         DEALLOCATE (RI_basis_parameter(i)%sphi_ext)
      END DO
      DEALLOCATE (RI_basis_parameter)

      ! S0 basis
      DO i = 1, SIZE(basis_S0)
         DEALLOCATE (basis_S0(i)%set_radius)
         DEALLOCATE (basis_S0(i)%lmax)
         DEALLOCATE (basis_S0(i)%lmin)
         DEALLOCATE (basis_S0(i)%npgf)
         DEALLOCATE (basis_S0(i)%nsgf)
         DEALLOCATE (basis_S0(i)%nshell)
         DEALLOCATE (basis_S0(i)%pgf_radius)
         DEALLOCATE (basis_S0(i)%sphi)
         DEALLOCATE (basis_S0(i)%zet)
         DEALLOCATE (basis_S0(i)%first_sgf)
         DEALLOCATE (basis_S0(i)%nl)
         DEALLOCATE (basis_S0(i)%nsgfl)
         DEALLOCATE (basis_S0(i)%sphi_ext)
      END DO
      DEALLOCATE (basis_S0)

   END SUBROUTINE release_RI_basis_set

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
!> \param qs_env   ...
!> \param para_env ...
!> \param mp2_biel ...
!> \param dimen ...
!> \param C ...
!> \param occupied ...
!> \param RI_basis_parameter ...
!> \param RI_basis_info ...
!> \param RI_index_table ...
!> \param RI_dimen ...
!> \param basis_S0 ...
!> \param Lai ...
!> \par History
!>      08.2013 created [Mauro Del Ben]
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE calc_lai_libint(mp2_env, qs_env, para_env, &
                              mp2_biel, dimen, C, occupied, &
                              RI_basis_parameter, RI_basis_info, RI_index_table, RI_dimen, basis_S0, &
                              Lai)

      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(mp2_biel_type)                                :: mp2_biel
      INTEGER                                            :: dimen
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C
      INTEGER                                            :: occupied
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter
      TYPE(hfx_basis_info_type)                          :: RI_basis_info
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: RI_index_table
      INTEGER                                            :: RI_dimen
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_S0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: Lai

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_lai_libint', &
         routineP = moduleN//':'//routineN

      INTEGER :: bits_max_val, cache_size, counter_L_blocks, handle, i, i_list_kl, i_set_list_kl, &
         i_set_list_kl_start, i_set_list_kl_stop, i_thread, iatom, iatom_end, iatom_start, iiB, &
         ikind, info_chol, irep, iset, jatom, jatom_end, jatom_start, jjB, jkind, jset, katom, &
         katom_end, katom_start, kkB, kkind, kset, kset_start, L_B_i_end, L_B_i_start, L_B_k_end, &
         L_B_k_start, l_max, latom, latom_end, latom_start, lkind, llB, lset, max_am, max_pgf, &
         max_set, n_threads, natom, ncob, ncos_max, nints, nkind, nneighbors, nseta, nsetb, nsetc, &
         nsetd, nsgf_max, nspins, orb_k_end, orb_k_start, orb_l_end
      INTEGER :: orb_l_start, primitive_counter, sgfb, sphi_a_u1, sphi_a_u2, sphi_a_u3, sphi_b_u1, &
         sphi_b_u2, sphi_b_u3, sphi_c_u1, sphi_c_u2, sphi_c_u3, sphi_d_u1, sphi_d_u2, sphi_d_u3, &
         unit_id, virtual
      INTEGER(int_8) :: estimate_to_store_int, max_val_memory, mem_eris, mem_eris_disk, &
         mem_max_val, n_processes, ncpu, neris_disk, neris_incore, neris_onthefly, neris_tmp, &
         neris_total, nprim_ints, shm_neris_disk, shm_neris_incore, shm_neris_onthefly, &
         shm_neris_total, shm_nprim_ints, shm_stor_count_int_disk, shm_stor_count_max_val, &
         shm_storage_counter_integrals
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of, last_sgf_global, nimages
      INTEGER, DIMENSION(:), POINTER :: la_max, la_min, lb_max, lb_min, lc_max, lc_min, ld_max, &
         ld_min, npgfa, npgfb, npgfc, npgfd, nsgfa, nsgfb, nsgfc, nsgfd, shm_block_offset
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfb, nsgfl_a, nsgfl_b, nsgfl_c, &
                                                            nsgfl_d, shm_atomic_block_offset
      INTEGER, DIMENSION(:, :), POINTER, SAVE            :: shm_is_assoc_atomic_block
      INTEGER, DIMENSION(:, :, :, :), POINTER            :: shm_set_offset
      INTEGER, SAVE                                      :: shm_number_of_p_entries
      LOGICAL                                            :: do_periodic, treat_lsd_in_core
      LOGICAL, DIMENSION(:, :), POINTER                  :: shm_atomic_pair_list
      REAL(KIND=dp) :: cartesian_estimate, coeffs_kind_max0, compression_factor, &
         compression_factor_disk, eps_schwarz, eps_storage, hf_fraction, ln_10, log10_eps_schwarz, &
         log10_pmax, max_contraction_val, pmax_atom, pmax_entry, ra(3), rab2, rb(3), rc(3), rcd2, &
         rd(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ee_buffer1, ee_buffer2, &
                                                            ee_primitives_tmp, ee_work, ee_work2, &
                                                            primitive_integrals
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: L_block, L_full_matrix
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BI1, MNRS
      REAL(KIND=dp), DIMENSION(:), POINTER               :: p_work
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: max_contraction, shm_pmax_block, sphi_b, &
                                                            zeta, zetb, zetc, zetd
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: sphi_a_ext_set, sphi_b_ext_set, &
                                                            sphi_c_ext_set, sphi_d_ext_set
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: sphi_a_ext, sphi_b_ext, sphi_c_ext, &
                                                            sphi_d_ext
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_L
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_L
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(hfx_basis_info_type), POINTER                 :: basis_info
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_parameter
      TYPE(hfx_general_type)                             :: general_parameter
      TYPE(hfx_load_balance_type), POINTER               :: load_balance_parameter
      TYPE(hfx_memory_type), POINTER                     :: memory_parameter
      TYPE(hfx_pgf_list), ALLOCATABLE, DIMENSION(:)      :: pgf_list_ij, pgf_list_kl
      TYPE(hfx_pgf_product_list), ALLOCATABLE, &
         DIMENSION(:)                                    :: pgf_product_list
      TYPE(hfx_potential_type)                           :: mp2_potential_parameter, &
                                                            potential_parameter
      TYPE(hfx_screen_coeff_type), ALLOCATABLE, &
         DIMENSION(:, :), TARGET                         :: radii_pgf_large, screen_coeffs_pgf_large
      TYPE(hfx_screen_coeff_type), DIMENSION(:, :), &
         POINTER                                         :: screen_coeffs_kind, tmp_R_1, tmp_R_2, &
                                                            tmp_screen_pgf1, tmp_screen_pgf2
      TYPE(hfx_screen_coeff_type), &
         DIMENSION(:, :, :, :), POINTER                  :: screen_coeffs_set
      TYPE(hfx_screen_coeff_type), &
         DIMENSION(:, :, :, :, :, :), POINTER            :: radii_pgf, screen_coeffs_pgf
      TYPE(hfx_screening_type)                           :: screening_parameter
      TYPE(hfx_type), POINTER                            :: actual_x_data, shm_master_x_data
      TYPE(lib_int)                                      :: private_lib
      TYPE(pair_list_type_mp2)                           :: list_kl
      TYPE(pair_set_list_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: set_list_kl
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      NULLIFY (dft_control)
      CALL timeset(routineN, handle)

      log10_eps_schwarz = log_zero

      irep = 1

      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control)

      !! Calculate l_max used in fgamma , because init_md_ftable is definitely not thread safe
      nkind = SIZE(qs_kind_set, 1)
      l_max = 0
      DO ikind = 1, nkind
         l_max = MAX(l_max, MAXVAL(qs_env%x_data(1, 1)%basis_parameter(ikind)%lmax))
         l_max = MAX(l_max, MAXVAL(RI_basis_parameter(ikind)%lmax))
      ENDDO
      l_max = 4*l_max
      CALL init_md_ftable(l_max)

      IF (mp2_env%potential_parameter%potential_type == do_mp2_potential_TShPSC) THEN
         IF (l_max > init_TShPSC_lmax) THEN
            IF (para_env%mepos == 0) THEN
               CALL open_file(unit_number=unit_id, file_name=mp2_env%potential_parameter%filename)
            END IF
            CALL init(l_max, unit_id, para_env%mepos, para_env%group)
            IF (para_env%mepos == 0) THEN
               CALL close_file(unit_id)
            END IF
            init_TShPSC_lmax = l_max
         END IF
         mp2_potential_parameter%cutoff_radius = mp2_env%potential_parameter%truncation_radius/2.0_dp
      END IF

      mp2_potential_parameter%potential_type = mp2_env%potential_parameter%potential_type

      n_threads = 1

      shm_neris_total = 0
      shm_nprim_ints = 0
      shm_neris_onthefly = 0
      shm_storage_counter_integrals = 0
      shm_stor_count_int_disk = 0
      shm_neris_incore = 0
      shm_neris_disk = 0
      shm_stor_count_max_val = 0

      ln_10 = LOG(10.0_dp)
      i_thread = 0

      actual_x_data => qs_env%x_data(irep, i_thread+1)

      shm_master_x_data => qs_env%x_data(irep, 1)

      do_periodic = actual_x_data%periodic_parameter%do_periodic

      IF (do_periodic) THEN
         ! ** Rebuild neighbor lists in case the cell has changed (i.e. NPT MD)
         actual_x_data%periodic_parameter%number_of_shells = actual_x_data%periodic_parameter%mode
         CALL hfx_create_neighbor_cells(actual_x_data, actual_x_data%periodic_parameter%number_of_shells_from_input, &
                                        cell, i_thread)
      END IF

      screening_parameter = actual_x_data%screening_parameter
      potential_parameter = actual_x_data%potential_parameter

      general_parameter = actual_x_data%general_parameter
      load_balance_parameter => actual_x_data%load_balance_parameter
      memory_parameter => actual_x_data%memory_parameter

      cache_size = memory_parameter%cache_size
      bits_max_val = memory_parameter%bits_max_val

      basis_parameter => actual_x_data%basis_parameter
      basis_info => actual_x_data%basis_info

      treat_lsd_in_core = general_parameter%treat_lsd_in_core

      ncpu = para_env%num_pe
      n_processes = ncpu*n_threads

      !! initalize some counters
      neris_total = 0_int_8
      neris_incore = 0_int_8
      neris_disk = 0_int_8
      neris_onthefly = 0_int_8
      mem_eris = 0_int_8
      mem_eris_disk = 0_int_8
      mem_max_val = 0_int_8
      compression_factor = 0.0_dp
      compression_factor_disk = 0.0_dp
      nprim_ints = 0_int_8
      neris_tmp = 0_int_8
      max_val_memory = 1_int_8

      max_am = MAX(basis_info%max_am, RI_basis_info%max_am)

      max_set = MAX(basis_info%max_set, RI_basis_info%max_set)
      CALL get_qs_env(qs_env=qs_env, &
                      qs_kind_set=qs_kind_set, &
                      particle_set=particle_set)

      natom = SIZE(particle_set, 1)

      ALLOCATE (kind_of(natom))

      CALL get_atomic_kind_set(atomic_kind_set, kind_of=kind_of)

      !! precompute maximum nco and allocate scratch
      ncos_max = 0
      nsgf_max = 0
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         nseta = basis_parameter(ikind)%nset
         npgfa => basis_parameter(ikind)%npgf
         la_max => basis_parameter(ikind)%lmax
         nsgfa => basis_parameter(ikind)%nsgf
         DO iset = 1, nseta
            ncos_max = MAX(ncos_max, ncoset(la_max(iset)))
            nsgf_max = MAX(nsgf_max, nsgfa(iset))
         ENDDO
         nseta = RI_basis_parameter(ikind)%nset
         npgfa => RI_basis_parameter(ikind)%npgf
         la_max => RI_basis_parameter(ikind)%lmax
         nsgfa => RI_basis_parameter(ikind)%nsgf
         DO iset = 1, nseta
            ncos_max = MAX(ncos_max, ncoset(la_max(iset)))
            nsgf_max = MAX(nsgf_max, nsgfa(iset))
         ENDDO
      ENDDO
      !! Allocate the arrays for the integrals.
      ALLOCATE (primitive_integrals(nsgf_max**4))
      primitive_integrals = 0.0_dp

      ALLOCATE (ee_work(ncos_max**4))
      ALLOCATE (ee_work2(ncos_max**4))
      ALLOCATE (ee_buffer1(ncos_max**4))
      ALLOCATE (ee_buffer2(ncos_max**4))
      ALLOCATE (ee_primitives_tmp(nsgf_max**4))

      nspins = dft_control%nspins

      ALLOCATE (max_contraction(max_set, natom))

      max_contraction = 0.0_dp
      max_pgf = 0
      DO jatom = 1, natom
         jkind = kind_of(jatom)
         lb_max => basis_parameter(jkind)%lmax
         nsetb = basis_parameter(jkind)%nset
         npgfb => basis_parameter(jkind)%npgf
         first_sgfb => basis_parameter(jkind)%first_sgf
         sphi_b => basis_parameter(jkind)%sphi
         nsgfb => basis_parameter(jkind)%nsgf
         DO jset = 1, nsetb
            ! takes the primitive to contracted transformation into account
            ncob = npgfb(jset)*ncoset(lb_max(jset))
            sgfb = first_sgfb(1, jset)
            ! if the primitives are assumed to be all of max_val2, max_val2*p2s_b becomes
            ! the maximum value after multiplication with sphi_b
            max_contraction(jset, jatom) = MAXVAL((/(SUM(ABS(sphi_b(1:ncob, i))), i=sgfb, sgfb+nsgfb(jset)-1)/))
            max_pgf = MAX(max_pgf, npgfb(jset))
         ENDDO
      ENDDO

      ! ** Allocate buffers for pgf_lists
      nneighbors = SIZE(actual_x_data%neighbor_cells)
      ALLOCATE (pgf_list_ij(max_pgf**2))
      ALLOCATE (pgf_list_kl(max_pgf**2))
      ALLOCATE (pgf_product_list(nneighbors**3))
      ALLOCATE (nimages(max_pgf**2))

      DO i = 1, max_pgf**2
         ALLOCATE (pgf_list_ij(i)%image_list(nneighbors))
         ALLOCATE (pgf_list_kl(i)%image_list(nneighbors))
      END DO

      ! ** Set pointers
      shm_number_of_p_entries = shm_master_x_data%number_of_p_entries
      shm_is_assoc_atomic_block => shm_master_x_data%is_assoc_atomic_block
      shm_atomic_block_offset => shm_master_x_data%atomic_block_offset
      shm_set_offset => shm_master_x_data%set_offset
      shm_block_offset => shm_master_x_data%block_offset

      !!! Skipped part

      !! Get screening parameter
      eps_schwarz = screening_parameter%eps_schwarz
      IF (eps_schwarz <= 0.0_dp) THEN
         log10_eps_schwarz = log_zero
      ELSE
         log10_eps_schwarz = LOG10(eps_schwarz)
      END IF
      !! get storage epsilon
      eps_storage = eps_schwarz*memory_parameter%eps_storage_scaling

      !! If we have a hybrid functional, we may need only a fraction of exact exchange
      hf_fraction = general_parameter%fraction

      !! The number of integrals that fit into the given MAX_MEMORY

      !! Parameters related to the potential 1/r, erf(wr)/r, erfc(wr/r)
      potential_parameter = actual_x_data%potential_parameter

      logger => cp_get_default_logger()

      private_lib = actual_x_data%lib

      !!! Helper array to map local basis function indeces to global ones
      ALLOCATE (last_sgf_global(0:natom))
      last_sgf_global(0) = 0
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         last_sgf_global(iatom) = last_sgf_global(iatom-1)+basis_parameter(ikind)%nsgf_total
      END DO

      IF (.NOT. shm_master_x_data%screen_funct_is_initialized) THEN
         CALL calc_pair_dist_radii(qs_env, basis_parameter, &
                                   shm_master_x_data%pair_dist_radii_pgf, max_set, max_pgf, eps_schwarz, &
                                   n_threads, i_thread)
         CALL calc_screening_functions(qs_env, basis_parameter, private_lib, shm_master_x_data%potential_parameter, &
                                       shm_master_x_data%screen_funct_coeffs_set, &
                                       shm_master_x_data%screen_funct_coeffs_kind, &
                                       shm_master_x_data%screen_funct_coeffs_pgf, &
                                       shm_master_x_data%pair_dist_radii_pgf, &
                                       max_set, max_pgf, n_threads, i_thread, p_work)

         shm_master_x_data%screen_funct_is_initialized = .TRUE.
      END IF

      screen_coeffs_set => shm_master_x_data%screen_funct_coeffs_set
      screen_coeffs_kind => shm_master_x_data%screen_funct_coeffs_kind
      screen_coeffs_pgf => shm_master_x_data%screen_funct_coeffs_pgf
      radii_pgf => shm_master_x_data%pair_dist_radii_pgf

      ALLOCATE (radii_pgf_large(SIZE(radii_pgf, 1), SIZE(radii_pgf, 2)))
      ALLOCATE (screen_coeffs_pgf_large(SIZE(screen_coeffs_pgf, 1), SIZE(screen_coeffs_pgf, 2)))
      DO iiB = 1, SIZE(radii_pgf, 1)
         DO jjB = 1, SIZE(radii_pgf, 2)
            radii_pgf_large(iiB, jjB)%x = 100_dp
         END DO
      END DO
      DO iiB = 1, SIZE(screen_coeffs_pgf, 1)
         DO jjB = 1, SIZE(screen_coeffs_pgf, 2)
            screen_coeffs_pgf_large(iiB, jjB)%x = 5000_dp
         END DO
      END DO
      tmp_R_1 => radii_pgf_large(:, :)
      tmp_R_2 => radii_pgf_large(:, :)
      tmp_screen_pgf1 => screen_coeffs_pgf_large(:, :)
      tmp_screen_pgf2 => screen_coeffs_pgf_large(:, :)

      ! start computing the L matrix
      ALLOCATE (L_full_matrix(RI_dimen, RI_dimen))
      L_full_matrix = 0.0_dp

      counter_L_blocks = 0
      DO iatom = 1, natom
         jatom = iatom

         ikind = kind_of(iatom)
         jkind = kind_of(jatom)
         ra = particle_set(iatom)%r(:)
         rb = particle_set(jatom)%r(:)
         !rab2=(ra(1)-rb(1))**2+(ra(2)-rb(2))**2+(ra(3)-rb(3))**2
         rab2 = 0.0_dp

         la_max => RI_basis_parameter(ikind)%lmax
         la_min => RI_basis_parameter(ikind)%lmin
         npgfa => RI_basis_parameter(ikind)%npgf
         nseta = RI_basis_parameter(ikind)%nset
         zeta => RI_basis_parameter(ikind)%zet
         nsgfa => RI_basis_parameter(ikind)%nsgf
         sphi_a_ext => RI_basis_parameter(ikind)%sphi_ext(:, :, :, :)
         nsgfl_a => RI_basis_parameter(ikind)%nsgfl
         sphi_a_u1 = UBOUND(sphi_a_ext, 1)
         sphi_a_u2 = UBOUND(sphi_a_ext, 2)
         sphi_a_u3 = UBOUND(sphi_a_ext, 3)

         lb_max => basis_S0(jkind)%lmax
         lb_min => basis_S0(jkind)%lmin
         npgfb => basis_S0(jkind)%npgf
         nsetb = basis_S0(jkind)%nset
         zetb => basis_S0(jkind)%zet
         nsgfb => basis_S0(jkind)%nsgf
         sphi_b_ext => basis_S0(jkind)%sphi_ext(:, :, :, :)
         nsgfl_b => basis_S0(jkind)%nsgfl
         sphi_b_u1 = UBOUND(sphi_b_ext, 1)
         sphi_b_u2 = UBOUND(sphi_b_ext, 2)
         sphi_b_u3 = UBOUND(sphi_b_ext, 3)

         DO katom = iatom, natom
            latom = katom

            kkind = kind_of(katom)
            lkind = kind_of(latom)
            rc = particle_set(katom)%r(:)
            rd = particle_set(latom)%r(:)
            !rcd2=(rc(1)-rd(1))**2+(rc(2)-rd(2))**2+(rc(3)-rd(3))**2
            rcd2 = 0.0_dp

            !pmax_atom = 0.0_dp
            !screen_kind_ij = screen_coeffs_kind(jkind,ikind)%x(1)*rab2+&
            !                 screen_coeffs_kind(jkind,ikind)%x(2)
            !screen_kind_kl = screen_coeffs_kind(lkind,kkind)%x(1)*rcd2+&
            !                 screen_coeffs_kind(lkind,kkind)%x(2)
            !IF( max_val1 + screen_kind_kl + pmax_atom < log10_eps_schwarz) CYCLE
            !IF( screen_kind_ij + screen_kind_kl + pmax_atom < log10_eps_schwarz ) CYCLE

            lc_max => RI_basis_parameter(kkind)%lmax
            lc_min => RI_basis_parameter(kkind)%lmin
            npgfc => RI_basis_parameter(kkind)%npgf
            nsetc = RI_basis_parameter(kkind)%nset
            zetc => RI_basis_parameter(kkind)%zet
            nsgfc => RI_basis_parameter(kkind)%nsgf
            sphi_c_ext => RI_basis_parameter(kkind)%sphi_ext(:, :, :, :)
            nsgfl_c => RI_basis_parameter(kkind)%nsgfl
            sphi_c_u1 = UBOUND(sphi_c_ext, 1)
            sphi_c_u2 = UBOUND(sphi_c_ext, 2)
            sphi_c_u3 = UBOUND(sphi_c_ext, 3)

            ld_max => basis_S0(lkind)%lmax
            ld_min => basis_S0(lkind)%lmin
            npgfd => basis_S0(lkind)%npgf
            nsetd = RI_basis_parameter(lkind)%nset
            zetd => basis_S0(lkind)%zet
            nsgfd => basis_S0(lkind)%nsgf
            sphi_d_ext => basis_S0(lkind)%sphi_ext(:, :, :, :)
            nsgfl_d => basis_S0(lkind)%nsgfl
            sphi_d_u1 = UBOUND(sphi_d_ext, 1)
            sphi_d_u2 = UBOUND(sphi_d_ext, 2)
            sphi_d_u3 = UBOUND(sphi_d_ext, 3)

            jset = 1
            lset = 1
            DO iset = 1, nseta

               ncob = npgfb(jset)*ncoset(lb_max(jset))
               !max_val1 = screen_coeffs_set(jset,iset,jkind,ikind)%x(1)*rab2 + &
               !           screen_coeffs_set(jset,iset,jkind,ikind)%x(2)
               sphi_a_ext_set => sphi_a_ext(:, :, :, iset)
               sphi_b_ext_set => sphi_b_ext(:, :, :, jset)

               L_B_i_start = RI_index_table(iatom, iset)
               L_B_i_end = RI_index_table(iatom, iset)+nsgfa(iset)-1

               kset_start = 1
               IF (iatom == katom) kset_start = iset
               DO kset = kset_start, nsetc

                  counter_L_blocks = counter_L_blocks+1
                  IF (MOD(counter_L_blocks, para_env%num_pe) /= para_env%mepos) CYCLE
                  !max_val2_set = (screen_coeffs_set(lset,kset,lkind,kkind)%x(1)*rcd2 + &
                  !                screen_coeffs_set(lset,kset,lkind,kkind)%x(2) )
                  !max_val2 = max_val1 + max_val2_set
                  !!! Near field screening
                  !IF(max_val2 + pmax_atom <log10_eps_schwarz) CYCLE

                  sphi_c_ext_set => sphi_c_ext(:, :, :, kset)
                  sphi_d_ext_set => sphi_d_ext(:, :, :, lset)

                  L_B_k_start = RI_index_table(katom, kset)
                  L_B_k_end = RI_index_table(katom, kset)+nsgfc(kset)-1

                  pmax_entry = 0.0_dp
                  log10_pmax = pmax_entry
                  log10_eps_schwarz = log_zero
                  !max_val2 = max_val2 + log10_pmax
                  !IF(max_val2<log10_eps_schwarz) CYCLE
                  !pmax_entry = EXP(log10_pmax*ln_10)

                  max_contraction_val = 1.0000_dp
                  ! tmp_R_1 => radii_pgf(:,:,1,1,1,1)
                  ! tmp_R_2 => radii_pgf(:,:,1,1,1,1)
                  ! tmp_screen_pgf1 => screen_coeffs_pgf(:,:,1,1,1,1)
                  ! tmp_screen_pgf2 => screen_coeffs_pgf(:,:,1,1,1,1)

                  ALLOCATE (L_block(nsgfa(iset), nsgfc(kset)))

                  CALL coulomb4(private_lib, ra, rb, rc, rd, npgfa(iset), npgfb(jset), npgfc(kset), npgfd(lset), &
                                la_min(iset), la_max(iset), lb_min(jset), lb_max(jset), &
                                lc_min(kset), lc_max(kset), ld_min(lset), ld_max(lset), &
                                nsgfa(iset), nsgfb(jset), nsgfc(kset), nsgfd(lset), &
                                sphi_a_u1, sphi_a_u2, sphi_a_u3, &
                                sphi_b_u1, sphi_b_u2, sphi_b_u3, &
                                sphi_c_u1, sphi_c_u2, sphi_c_u3, &
                                sphi_d_u1, sphi_d_u2, sphi_d_u3, &
                                zeta(1:npgfa(iset), iset), zetb(1:npgfb(jset), jset), &
                                zetc(1:npgfc(kset), kset), zetd(1:npgfd(lset), lset), &
                                primitive_integrals, &
                                mp2_potential_parameter, &
                                actual_x_data%neighbor_cells, screen_coeffs_set(1, 1, 1, 1)%x, &
                                screen_coeffs_set(1, 1, 1, 1)%x, eps_schwarz, &
                                max_contraction_val, cartesian_estimate, cell, neris_tmp, &
                                log10_pmax, log10_eps_schwarz, &
                                tmp_R_1, tmp_R_2, tmp_screen_pgf1, tmp_screen_pgf2, &
                                pgf_list_ij, pgf_list_kl, pgf_product_list, &
                                nsgfl_a(:, iset), nsgfl_b(:, jset), &
                                nsgfl_c(:, kset), nsgfl_d(:, lset), &
                                sphi_a_ext_set, &
                                sphi_b_ext_set, &
                                sphi_c_ext_set, &
                                sphi_d_ext_set, &
                                ee_work, ee_work2, ee_buffer1, ee_buffer2, ee_primitives_tmp, &
                                nimages, do_periodic, p_work)

                  primitive_counter = 0
                  DO llB = 1, nsgfd(lset)
                     DO kkB = 1, nsgfc(kset)
                        DO jjB = 1, nsgfb(jset)
                           DO iiB = 1, nsgfa(iset)
                              primitive_counter = primitive_counter+1
                              L_block(iiB, kkB) = primitive_integrals(primitive_counter)
                              ! WRITE(1001,*) L_B_i_start+iiB-1, L_B_k_start+kkB-1, L_block(iiB,kkB)
                           END DO
                        END DO
                     END DO
                  END DO

                  L_full_matrix(L_B_i_start:L_B_i_end, L_B_k_start:L_B_k_end) = L_block
                  L_full_matrix(L_B_k_start:L_B_k_end, L_B_i_start:L_B_i_end) = TRANSPOSE(L_block)

                  DEALLOCATE (L_block)

               END DO ! kset
            END DO ! iset

         END DO !katom
      END DO !iatom

      ! now create a fm_matrix for the L matrix
      CALL mp_sum(L_full_matrix, para_env%group)

      ! create a sub blacs_env
      NULLIFY (blacs_env)
      CALL cp_blacs_env_create(blacs_env=blacs_env, para_env=para_env)

      NULLIFY (fm_matrix_L)
      NULLIFY (fm_struct_L)
      CALL cp_fm_struct_create(fm_struct_L, context=blacs_env, nrow_global=RI_dimen, &
                               ncol_global=RI_dimen, para_env=para_env)
      CALL cp_fm_create(fm_matrix_L, fm_struct_L, name="fm_matrix_L")
      CALL cp_fm_struct_release(fm_struct_L)
      CALL cp_blacs_env_release(blacs_env)

      CALL cp_fm_set_submatrix(fm=fm_matrix_L, new_values=L_full_matrix, start_row=1, start_col=1, &
                               n_rows=RI_dimen, n_cols=RI_dimen)

      info_chol = 0
      CALL cp_fm_cholesky_decompose(matrix=fm_matrix_L, n=RI_dimen, info_out=info_chol)
      IF(.NOT.(info_chol == 0))CALL cp__a("mp2_ri_libint.F",965)

      ! triangual invert
      CALL cp_fm_triangular_invert(matrix_a=fm_matrix_L, uplo_tr='U')

      ! replicate L matrix to each proc
      L_full_matrix = 0.0_dp
      CALL cp_fm_get_submatrix(fm_matrix_L, L_full_matrix, 1, 1, RI_dimen, RI_dimen, .FALSE.)
      CALL cp_fm_release(fm_matrix_L)

      ! clean lower part
      DO iiB = 1, RI_dimen
         L_full_matrix(iiB+1:RI_dimen, iiB) = 0.0_dp
      END DO

      ALLOCATE (list_kl%elements(natom**2))

      coeffs_kind_max0 = MAXVAL(screen_coeffs_kind(:, :)%x(2))
      ALLOCATE (set_list_kl((max_set*natom)**2))

      !! precalculate maximum density matrix elements in blocks
      actual_x_data%pmax_block = 0.0_dp
      shm_pmax_block => actual_x_data%pmax_block

      shm_atomic_pair_list => actual_x_data%atomic_pair_list

      iatom_start = 1
      iatom_end = natom
      jatom_start = 1
      jatom_end = natom
      katom_start = 1
      katom_end = natom
      latom_start = 1
      latom_end = natom

      CALL build_pair_list_mp2(natom, list_kl, set_list_kl, katom_start, katom_end, &
                               latom_start, latom_end, &
                               kind_of, basis_parameter, particle_set, &
                               do_periodic, screen_coeffs_set, screen_coeffs_kind, &
                               coeffs_kind_max0, log10_eps_schwarz, cell, 0.D+00, &
                               shm_atomic_pair_list)

      virtual = dimen-occupied

      ALLOCATE (Lai(RI_dimen, virtual, occupied))
      Lai = 0.0_dp

      DO iatom = 1, natom
         jatom = iatom

         ikind = kind_of(iatom)
         jkind = kind_of(jatom)
         ra = particle_set(iatom)%r(:)
         rb = particle_set(jatom)%r(:)
         !rab2=(ra(1)-rb(1))**2+(ra(2)-rb(2))**2+(ra(3)-rb(3))**2
         rab2 = 0.0_dp

         la_max => RI_basis_parameter(ikind)%lmax
         la_min => RI_basis_parameter(ikind)%lmin
         npgfa => RI_basis_parameter(ikind)%npgf
         nseta = RI_basis_parameter(ikind)%nset
         zeta => RI_basis_parameter(ikind)%zet
         nsgfa => RI_basis_parameter(ikind)%nsgf
         sphi_a_ext => RI_basis_parameter(ikind)%sphi_ext(:, :, :, :)
         nsgfl_a => RI_basis_parameter(ikind)%nsgfl
         sphi_a_u1 = UBOUND(sphi_a_ext, 1)
         sphi_a_u2 = UBOUND(sphi_a_ext, 2)
         sphi_a_u3 = UBOUND(sphi_a_ext, 3)

         lb_max => basis_S0(jkind)%lmax
         lb_min => basis_S0(jkind)%lmin
         npgfb => basis_S0(jkind)%npgf
         nsetb = basis_S0(jkind)%nset
         zetb => basis_S0(jkind)%zet
         nsgfb => basis_S0(jkind)%nsgf
         sphi_b_ext => basis_S0(jkind)%sphi_ext(:, :, :, :)
         nsgfl_b => basis_S0(jkind)%nsgfl
         sphi_b_u1 = UBOUND(sphi_b_ext, 1)
         sphi_b_u2 = UBOUND(sphi_b_ext, 2)
         sphi_b_u3 = UBOUND(sphi_b_ext, 3)

         jset = 1
         DO iset = 1, nseta

            counter_L_blocks = counter_L_blocks+1
            IF (MOD(counter_L_blocks, para_env%num_pe) /= para_env%mepos) CYCLE

            ncob = npgfb(jset)*ncoset(lb_max(jset))
            !max_val1 = screen_coeffs_set(jset,iset,jkind,ikind)%x(1)*rab2 + &
            !           screen_coeffs_set(jset,iset,jkind,ikind)%x(2)
            sphi_a_ext_set => sphi_a_ext(:, :, :, iset)
            sphi_b_ext_set => sphi_b_ext(:, :, :, jset)

            L_B_i_start = RI_index_table(iatom, iset)
            L_B_i_end = RI_index_table(iatom, iset)+nsgfa(iset)-1

            ALLOCATE (BI1(dimen, dimen, nsgfa(iset)))
            BI1 = 0.0_dp

            DO i_list_kl = 1, list_kl%n_element

               katom = list_kl%elements(i_list_kl)%pair(1)
               latom = list_kl%elements(i_list_kl)%pair(2)

               i_set_list_kl_start = list_kl%elements(i_list_kl)%set_bounds(1)
               i_set_list_kl_stop = list_kl%elements(i_list_kl)%set_bounds(2)
               kkind = list_kl%elements(i_list_kl)%kind_pair(1)
               lkind = list_kl%elements(i_list_kl)%kind_pair(2)
               rc = list_kl%elements(i_list_kl)%r1
               rd = list_kl%elements(i_list_kl)%r2
               rcd2 = list_kl%elements(i_list_kl)%dist2

               pmax_atom = 0.0_dp

               !screen_kind_ij = screen_coeffs_kind(jkind,ikind)%x(1)*rab2+&
               !                 screen_coeffs_kind(jkind,ikind)%x(2)
               !screen_kind_kl = screen_coeffs_kind(lkind,kkind)%x(1)*rcd2+&
               !                 screen_coeffs_kind(lkind,kkind)%x(2)

               !!!!!! Change the loop order
               !IF( max_val1 + screen_kind_kl + pmax_atom < log10_eps_schwarz) CYCLE
               !!!!!!
               !IF( screen_kind_ij + screen_kind_kl + pmax_atom < log10_eps_schwarz ) CYCLE

               lc_max => basis_parameter(kkind)%lmax
               lc_min => basis_parameter(kkind)%lmin
               npgfc => basis_parameter(kkind)%npgf
               zetc => basis_parameter(kkind)%zet
               nsgfc => basis_parameter(kkind)%nsgf
               sphi_c_ext => basis_parameter(kkind)%sphi_ext(:, :, :, :)
               nsgfl_c => basis_parameter(kkind)%nsgfl
               sphi_c_u1 = UBOUND(sphi_c_ext, 1)
               sphi_c_u2 = UBOUND(sphi_c_ext, 2)
               sphi_c_u3 = UBOUND(sphi_c_ext, 3)

               ld_max => basis_parameter(lkind)%lmax
               ld_min => basis_parameter(lkind)%lmin
               npgfd => basis_parameter(lkind)%npgf
               zetd => basis_parameter(lkind)%zet
               nsgfd => basis_parameter(lkind)%nsgf
               sphi_d_ext => basis_parameter(lkind)%sphi_ext(:, :, :, :)
               nsgfl_d => basis_parameter(lkind)%nsgfl
               sphi_d_u1 = UBOUND(sphi_d_ext, 1)
               sphi_d_u2 = UBOUND(sphi_d_ext, 2)
               sphi_d_u3 = UBOUND(sphi_d_ext, 3)

               DO i_set_list_kl = i_set_list_kl_start, i_set_list_kl_stop
                  kset = set_list_kl(i_set_list_kl)%pair(1)
                  lset = set_list_kl(i_set_list_kl)%pair(2)

                  IF (katom == latom .AND. lset < kset) CYCLE

                  orb_k_start = mp2_biel%index_table(katom, kset)
                  orb_k_end = orb_k_start+nsgfc(kset)-1
                  orb_l_start = mp2_biel%index_table(latom, lset)
                  orb_l_end = orb_l_start+nsgfd(lset)-1

                  !max_val2_set = (screen_coeffs_set(lset,kset,lkind,kkind)%x(1)*rcd2 + &
                  !                screen_coeffs_set(lset,kset,lkind,kkind)%x(2) )
                  !max_val2 = max_val1 + max_val2_set

                  !!! Near field screening
                  !IF(max_val2 + pmax_atom <log10_eps_schwarz) CYCLE
                  !! get max_vals if we screen on initial density
                  pmax_entry = 0.0_dp

                  sphi_c_ext_set => sphi_c_ext(:, :, :, kset)
                  sphi_d_ext_set => sphi_d_ext(:, :, :, lset)

                  log10_pmax = pmax_entry
                  log10_eps_schwarz = log_zero
                  !max_val2 = max_val2 + log10_pmax
                  !!IF(max_val2<log10_eps_schwarz) CYCLE
                  !pmax_entry = EXP(log10_pmax*ln_10)

                  IF (ALLOCATED(MNRS)) DEALLOCATE (MNRS)
                  ALLOCATE (MNRS(nsgfd(lset), nsgfc(kset), nsgfa(iset)))

                  MNRS = 0.0_dp

                  !max_contraction_val =  max_contraction(iset,iatom) * &
                  !                       max_contraction(jset,jatom) * &
                  !                       max_contraction(kset,katom) * &
                  !                       max_contraction(lset,latom) * pmax_entry
                  !tmp_R_1 => radii_pgf(:,:,jset,iset,jkind,ikind)
                  !tmp_R_2 => radii_pgf(:,:,lset,kset,lkind,kkind)
                  !tmp_screen_pgf1 => screen_coeffs_pgf(:,:,jset,iset,jkind,ikind)
                  !tmp_screen_pgf2 => screen_coeffs_pgf(:,:,lset,kset,lkind,kkind)

                  max_contraction_val = max_contraction(kset, katom)*max_contraction(lset, latom)
                  ! tmp_R_1 => radii_pgf(:,:,kset,kset,kkind,kkind)
                  ! tmp_R_2 => radii_pgf(:,:,kset,kset,kkind,kkind)
                  ! tmp_screen_pgf1 => screen_coeffs_pgf(:,:,kset,kset,kkind,kkind)
                  ! tmp_screen_pgf2 => screen_coeffs_pgf(:,:,kset,kset,kkind,kkind)

                  CALL coulomb4(private_lib, ra, rb, rc, rd, npgfa(iset), npgfb(jset), npgfc(kset), npgfd(lset), &
                                la_min(iset), la_max(iset), lb_min(jset), lb_max(jset), &
                                lc_min(kset), lc_max(kset), ld_min(lset), ld_max(lset), &
                                nsgfa(iset), nsgfb(jset), nsgfc(kset), nsgfd(lset), &
                                sphi_a_u1, sphi_a_u2, sphi_a_u3, &
                                sphi_b_u1, sphi_b_u2, sphi_b_u3, &
                                sphi_c_u1, sphi_c_u2, sphi_c_u3, &
                                sphi_d_u1, sphi_d_u2, sphi_d_u3, &
                                zeta(1:npgfa(iset), iset), zetb(1:npgfb(jset), jset), &
                                zetc(1:npgfc(kset), kset), zetd(1:npgfd(lset), lset), &
                                primitive_integrals, &
                                mp2_potential_parameter, &
                                actual_x_data%neighbor_cells, screen_coeffs_set(kset, kset, kkind, kkind)%x, &
                                screen_coeffs_set(kset, kset, kkind, kkind)%x, eps_schwarz, &
                                max_contraction_val, cartesian_estimate, cell, neris_tmp, &
                                log10_pmax, log10_eps_schwarz, &
                                tmp_R_1, tmp_R_2, tmp_screen_pgf1, tmp_screen_pgf2, &
                                pgf_list_ij, pgf_list_kl, pgf_product_list, &
                                nsgfl_a(:, iset), nsgfl_b(:, jset), &
                                nsgfl_c(:, kset), nsgfl_d(:, lset), &
                                sphi_a_ext_set, &
                                sphi_b_ext_set, &
                                sphi_c_ext_set, &
                                sphi_d_ext_set, &
                                ee_work, ee_work2, ee_buffer1, ee_buffer2, ee_primitives_tmp, &
                                nimages, do_periodic, p_work)

                  nints = nsgfa(iset)*nsgfb(jset)*nsgfc(kset)*nsgfd(lset)
                  neris_total = neris_total+nints
                  nprim_ints = nprim_ints+neris_tmp
                  IF (cartesian_estimate == 0.0_dp) cartesian_estimate = TINY(cartesian_estimate)
                  estimate_to_store_int = EXPONENT(cartesian_estimate)
                  estimate_to_store_int = MAX(estimate_to_store_int, -15_int_8)
                  cartesian_estimate = SET_EXPONENT(1.0_dp, estimate_to_store_int+1)

                  !IF(cartesian_estimate<eps_schwarz) CYCLE

                  ! WRITE(1111,*) nsgfd(lset),nsgfc(kset),nsgfa(iset)

                  primitive_counter = 0
                  DO llB = 1, nsgfd(lset)
                     DO kkB = 1, nsgfc(kset)
                        DO jjB = 1, nsgfb(jset)
                           DO iiB = 1, nsgfa(iset)
                              primitive_counter = primitive_counter+1
                              MNRS(llB, kkB, iiB) = primitive_integrals(primitive_counter)
                              ! WRITE(1111,*) orb_l_start+llB-1,orb_k_start+kkB-1,L_B_i_start+iiB-1,MNRS(llB,kkB,iiB)
                           END DO
                        END DO
                     END DO
                  END DO

                  DO iiB = 1, nsgfa(iset)
                     BI1(orb_l_start:orb_l_end, orb_k_start:orb_k_end, iiB) = MNRS(:, :, iiB)
                     BI1(orb_k_start:orb_k_end, orb_l_start:orb_l_end, iiB) = TRANSPOSE(MNRS(:, :, iiB))
                  END DO

               END DO ! i_set_list_kl
            END DO ! i_list_kl

            DO iiB = 1, nsgfa(iset)
               BI1(1:virtual, 1:occupied, iiB) = MATMUL(TRANSPOSE(C(1:dimen, occupied+1:dimen)), &
                                                        MATMUL(BI1(1:dimen, 1:dimen, iiB), C(1:dimen, 1:occupied)))
               Lai(L_B_i_start+iiB-1, 1:virtual, 1:occupied) = BI1(1:virtual, 1:occupied, iiB)
            END DO

            DEALLOCATE (BI1)

         END DO
      END DO

      CALL mp_sum(Lai, para_env%group)

      DO iiB = 1, occupied
         IF (MOD(iiB, para_env%num_pe) == para_env%mepos) THEN
            Lai(1:RI_dimen, 1:virtual, iiB) = MATMUL(TRANSPOSE(L_full_matrix), Lai(1:RI_dimen, 1:virtual, iiB))
         ELSE
            Lai(:, :, iiB) = 0.0_dp
         END IF
      END DO

      CALL mp_sum(Lai, para_env%group)

      ! DO iiB=1, occupied
      !   DO jjB=1, virtual
      !     DO kkB=1, RI_dimen
      !       WRITE(3011,*) kkB,jjB,iiB,Lai(kkB,jjB,iiB)
      !     END DO
      !   END DO
      ! END DO

      DEALLOCATE (set_list_kl)

      DO i = 1, max_pgf**2
         DEALLOCATE (pgf_list_ij(i)%image_list)
         DEALLOCATE (pgf_list_kl(i)%image_list)
      END DO

      DEALLOCATE (pgf_list_ij)
      DEALLOCATE (pgf_list_kl)
      DEALLOCATE (pgf_product_list)

      DEALLOCATE (max_contraction, kind_of)

      DEALLOCATE (ee_work, ee_work2, ee_buffer1, ee_buffer2, ee_primitives_tmp)

      DEALLOCATE (nimages)

      IF (mp2_env%potential_parameter%potential_type == do_mp2_potential_TShPSC) THEN
         init_TShPSC_lmax = -1
         CALL free_C0()
      END IF

      CALL timestop(handle)

   END SUBROUTINE calc_lai_libint

END MODULE mp2_ri_libint
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to optimize the RI-MP2 basis. Only exponents of
!>        non-contracted auxiliary basis basis are optimized. The derivative
!>        of the MP2 energy with respect to the exponents of the basis
!>        are calculated numerically.
!> \par History
!>      08.2013 created [Mauro Del Ben]
!> \author Mauro Del Ben
! **************************************************************************************************
MODULE mp2_optimize_ri_basis
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE basis_set_container_types,       ONLY: add_basis_set_to_container,&
                                              remove_basis_from_container
   USE basis_set_types,                 ONLY: allocate_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_log_handling,                 ONLY: cp_add_default_logger,&
                                              cp_get_default_logger,&
                                              cp_logger_create,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_release,&
                                              cp_logger_set,&
                                              cp_logger_type,&
                                              cp_rm_default_logger
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
!
   USE hfx_libint_wrapper_types,        ONLY: build_eri_size
   USE hfx_types,                       ONLY: hfx_basis_info_type,&
                                              hfx_basis_type
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: default_output_unit
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_comm_split_direct,&
                                              mp_sum
   USE mp2_direct_method,               ONLY: mp2_canonical_direct_single_batch
   USE mp2_ri_libint,                   ONLY: libint_ri_mp2,&
                                              read_RI_basis_set,&
                                              release_RI_basis_set
   USE mp2_types,                       ONLY: mp2_biel_type,&
                                              mp2_type
   USE orbital_pointers,                ONLY: indco,&
                                              init_orbital_pointers,&
                                              nco,&
                                              ncoset,&
                                              nso
   USE orbital_symbols,                 ONLY: cgf_symbol,&
                                              sgf_symbol
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_rho_types,                    ONLY: qs_rho_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_optimize_ri_basis'

   PUBLIC :: optimize_ri_basis_main

CONTAINS

! **************************************************************************************************
!> \brief optimize RI-MP2 basis set
!> \param Emp2 ...
!> \param Emp2_Cou ...
!> \param Emp2_ex ...
!> \param Emp2_S ...
!> \param Emp2_T ...
!> \param dimen ...
!> \param natom ...
!> \param homo ...
!> \param mp2_biel ...
!> \param mp2_env ...
!> \param C ...
!> \param Auto ...
!> \param kind_of ...
!> \param qs_env ...
!> \param rho ...
!> \param para_env ...
!> \param unit_nr ...
!> \param homo_beta ...
!> \param C_beta ...
!> \param Auto_beta ...
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE optimize_ri_basis_main(Emp2, Emp2_Cou, Emp2_ex, Emp2_S, Emp2_T, dimen, natom, homo, &
                                     mp2_biel, mp2_env, C, Auto, kind_of, &
                                     qs_env, rho, para_env, &
                                     unit_nr, homo_beta, C_beta, Auto_beta)

      REAL(KIND=dp)                                      :: Emp2, Emp2_Cou, Emp2_ex, Emp2_S, Emp2_T
      INTEGER                                            :: dimen, natom, homo
      TYPE(mp2_biel_type)                                :: mp2_biel
      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: C
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Auto
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: unit_nr
      INTEGER, OPTIONAL                                  :: homo_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :), &
         OPTIONAL                                        :: C_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), OPTIONAL :: Auto_beta

      CHARACTER(len=*), PARAMETER :: routineN = 'optimize_ri_basis_main', &
         routineP = moduleN//':'//routineN

      INTEGER :: color_sub, comm_sub, dimen_RI, elements_ij_proc, handle, i, iiter, ikind, ipgf, &
         iset, ishell, j, local_unit_nr, max_l_am, max_num_iter, ndof, nkind, number_groups, &
         virtual, virtual_beta
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ij_list_proc, index_table_RI
      LOGICAL                                            :: hess_up, open_shell_case, reset_boundary
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: basis_was_assoc
      REAL(KIND=dp) :: DI, DI_new, DRI, DRI_new, Emp2_AA, Emp2_AA_Cou, Emp2_AA_ex, Emp2_AB, &
         Emp2_AB_Cou, Emp2_AB_ex, Emp2_BB, Emp2_BB_Cou, Emp2_BB_ex, Emp2_RI, Emp2_RI_new, &
         eps_DI_rel, eps_DRI, eps_step, expon, fac, fad, fae, reset_edge, sumdg, sumxi
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: deriv, dg, g, hdg, lower_B, max_dev, &
                                                            max_rel_dev, p, pnew, xi
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: exp_limits, hessin
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: Integ_MP2, Integ_MP2_AA, Integ_MP2_AB, &
                                                            Integ_MP2_BB
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger, logger_sub
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      TYPE(gto_basis_set_type), POINTER                  :: ri_aux_basis
      TYPE(hfx_basis_info_type)                          :: RI_basis_info
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_S0, RI_basis_parameter
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      open_shell_case = .FALSE.
      IF (PRESENT(homo_beta) .AND. PRESENT(C_beta) .AND. PRESENT(Auto_beta)) open_shell_case = .TRUE.

      virtual = dimen-homo

      eps_DRI = mp2_env%ri_opt_param%DRI
      eps_DI_rel = mp2_env%ri_opt_param%DI_rel
      eps_step = mp2_env%ri_opt_param%eps_step
      max_num_iter = mp2_env%ri_opt_param%max_num_iter

      ! calculate the ERI's over molecular integrals
      Emp2 = 0.0_dp
      Emp2_Cou = 0.0_dp
      Emp2_ex = 0.0_dp
      Emp2_S = 0.0_dp
      Emp2_T = 0.0_dp
      IF (open_shell_case) THEN
         ! open shell case
         virtual_beta = dimen-homo_beta

         ! alpha-aplha case
         Emp2_AA = 0.0_dp
         Emp2_AA_Cou = 0.0_dp
         Emp2_AA_ex = 0.0_dp
         CALL calc_elem_ij_proc(homo, homo, para_env, elements_ij_proc, ij_list_proc)
         CALL mp2_canonical_direct_single_batch(Emp2_AA, Emp2_AA_Cou, Emp2_AA_ex, mp2_env, qs_env, rho, para_env, &
                                                mp2_biel, dimen, C, Auto, 0, homo, homo, &
                                                elements_ij_proc, ij_list_proc, homo, 0, &
                                                Integ_MP2=Integ_MP2_AA)
         CALL mp_sum(Emp2_AA_Cou, para_env%group)
         CALL mp_sum(Emp2_AA_Ex, para_env%group)
         CALL mp_sum(Emp2_AA, para_env%group)
         DEALLOCATE (ij_list_proc)

         ! beta-beta case
         Emp2_BB = 0.0_dp
         Emp2_BB_Cou = 0.0_dp
         Emp2_BB_ex = 0.0_dp
         CALL calc_elem_ij_proc(homo_beta, homo_beta, para_env, elements_ij_proc, ij_list_proc)
         CALL mp2_canonical_direct_single_batch(Emp2_BB, Emp2_BB_Cou, Emp2_BB_ex, mp2_env, qs_env, rho, para_env, &
                                                mp2_biel, dimen, C_beta, Auto_beta, 0, homo_beta, homo_beta, &
                                                elements_ij_proc, ij_list_proc, homo_beta, 0, &
                                                Integ_MP2=Integ_MP2_BB)
         CALL mp_sum(Emp2_BB_Cou, para_env%group)
         CALL mp_sum(Emp2_BB_Ex, para_env%group)
         CALL mp_sum(Emp2_BB, para_env%group)
         DEALLOCATE (ij_list_proc)

         ! aplha-beta case
         Emp2_AB = 0.0_dp
         Emp2_AB_Cou = 0.0_dp
         Emp2_AB_ex = 0.0_dp
         CALL calc_elem_ij_proc(homo, homo_beta, para_env, elements_ij_proc, ij_list_proc)
         CALL mp2_canonical_direct_single_batch(Emp2_AB, Emp2_AB_Cou, Emp2_AB_ex, mp2_env, qs_env, rho, para_env, &
                                                mp2_biel, dimen, C, Auto, 0, homo, homo, &
                                                elements_ij_proc, ij_list_proc, homo_beta, 0, &
                                                homo_beta, C_beta, Auto_beta, Integ_MP2=Integ_MP2_AB)
         CALL mp_sum(Emp2_AB_Cou, para_env%group)
         CALL mp_sum(Emp2_AB_Ex, para_env%group)
         CALL mp_sum(Emp2_AB, para_env%group)
         DEALLOCATE (ij_list_proc)

         ! IF (unit_nr>0) WRITE(unit_nr,'(T3,A,T56,F25.14)') 'MP2 Energy Alpha-Alpha = ', Emp2_AA
         ! IF (unit_nr>0) WRITE(unit_nr,'(T3,A,T56,F25.14)') 'MP2 Energy Beta-Beta   = ', Emp2_BB
         ! IF (unit_nr>0) WRITE(unit_nr,'(T3,A,T56,F25.14)') 'MP2 Energy Alpha-Beta  = ', Emp2_AB

         Emp2 = Emp2_AA+Emp2_BB+Emp2_AB*2.0_dp !+Emp2_BA
         Emp2_Cou = Emp2_AA_Cou+Emp2_BB_Cou+Emp2_AB_Cou*2.0_dp !+Emp2_BA
         Emp2_ex = Emp2_AA_ex+Emp2_BB_ex+Emp2_AB_ex*2.0_dp !+Emp2_BA

         Emp2_S = Emp2_AB*2.0_dp
         Emp2_T = Emp2_AA+Emp2_BB

         ! Replicate the MO-ERI's over all processes
         CALL mp_sum(Integ_MP2_AA, para_env%group)
         CALL mp_sum(Integ_MP2_BB, para_env%group)
         CALL mp_sum(Integ_MP2_AB, para_env%group)

      ELSE
         ! close shell case
         CALL calc_elem_ij_proc(homo, homo, para_env, elements_ij_proc, ij_list_proc)
         CALL mp2_canonical_direct_single_batch(Emp2, Emp2_Cou, Emp2_ex, mp2_env, qs_env, rho, para_env, &
                                                mp2_biel, dimen, C, Auto, 0, homo, homo, &
                                                elements_ij_proc, ij_list_proc, homo, 0, &
                                                Integ_MP2=Integ_MP2)
         CALL mp_sum(Emp2_Cou, para_env%group)
         CALL mp_sum(Emp2_Ex, para_env%group)
         CALL mp_sum(Emp2, para_env%group)
         DEALLOCATE (ij_list_proc)

         ! Replicate the MO-ERI's over all processes
         CALL mp_sum(Integ_MP2, para_env%group)

      END IF

      ! create the para_env_sub
      number_groups = para_env%num_pe/mp2_env%mp2_num_proc
      color_sub = para_env%mepos/mp2_env%mp2_num_proc
      CALL mp_comm_split_direct(para_env%group, comm_sub, color_sub)
      NULLIFY (para_env_sub)
      CALL cp_para_env_create(para_env_sub, comm_sub)

      IF (para_env%mepos == para_env%source) THEN
         local_unit_nr = cp_logger_get_default_unit_nr(logger, local=.FALSE.)
      ELSE
         local_unit_nr = default_output_unit
      ENDIF
      NULLIFY (logger_sub)
      CALL cp_logger_create(logger_sub, para_env=para_env_sub, &
                            default_global_unit_nr=local_unit_nr, close_global_unit_on_dealloc=.FALSE.)
      CALL cp_logger_set(logger_sub, local_filename="opt_RI_basis_localLog")
      CALL cp_add_default_logger(logger_sub)

      CALL generate_RI_init_basis(qs_env, mp2_env, nkind, max_rel_dev, basis_was_assoc)

      CALL read_RI_basis_set(qs_env, RI_basis_parameter, RI_basis_info, &
                             natom, nkind, kind_of, index_table_RI, dimen_RI, &
                             basis_S0)

      ndof = 0
      max_l_am = 0
      DO ikind = 1, nkind
         DO iset = 1, RI_basis_parameter(ikind)%nset
            ndof = ndof+1
            max_l_am = MAX(max_l_am, MAXVAL(RI_basis_parameter(ikind)%lmax))
         END DO
      END DO

      ALLOCATE (exp_limits(2, nkind))
      exp_limits(1, :) = HUGE(0)
      exp_limits(2, :) = -HUGE(0)
      DO ikind = 1, nkind
         DO iset = 1, RI_basis_parameter(ikind)%nset
            expon = RI_basis_parameter(ikind)%zet(1, iset)
            IF (expon <= exp_limits(1, ikind)) exp_limits(1, ikind) = expon
            IF (expon >= exp_limits(2, ikind)) exp_limits(2, ikind) = expon
         END DO
         IF (basis_was_assoc(ikind)) THEN
            exp_limits(1, ikind) = exp_limits(1, ikind)*0.5_dp
            exp_limits(2, ikind) = exp_limits(2, ikind)*1.5_dp
         ELSE
            exp_limits(1, ikind) = exp_limits(1, ikind)*0.8_dp*0.5_dp
            exp_limits(2, ikind) = exp_limits(2, ikind)*1.2_dp*1.5_dp
         END IF
      END DO
      DEALLOCATE (basis_was_assoc)

      ! check if the max angular momentum exceed the libint one
      IF (max_l_am > build_eri_size) THEN
         CALL cp__b("mp2_optimize_ri_basis.F",294,"The angular momentum needed exceeds the value assumed when configuring libint.")
      END IF

      ! Allocate stuff
      ALLOCATE (p(ndof))
      p = 0.0_dp
      ALLOCATE (xi(ndof))
      xi = 0.0_dp
      ALLOCATE (g(ndof))
      g = 0.0_dp
      ALLOCATE (dg(ndof))
      dg = 0.0_dp
      ALLOCATE (hdg(ndof))
      hdg = 0.0_dp
      ALLOCATE (pnew(ndof))
      pnew = 0.0_dp
      ALLOCATE (deriv(ndof))
      deriv = 0.0_dp
      ALLOCATE (hessin(ndof, ndof))
      hessin = 0.0_dp
      DO i = 1, ndof
         hessin(i, i) = 1.0_dp
      END DO

      ! initialize trasformation function
      ALLOCATE (lower_B(ndof))
      lower_B = 0.0_dp
      ALLOCATE (max_dev(ndof))
      max_dev = 0.0_dp

      ! Initialize the transformation function
      CALL init_transf(nkind, RI_basis_parameter, lower_B, max_dev, max_rel_dev)

      ! get the atomic kind set for writing the basis
      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set)

      ! Calculate RI-MO-ERI's
      CALL calc_energy_func(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, Emp2_RI, DRI, DI, &
                            Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, &
                            qs_env, natom, dimen, dimen_RI, homo, virtual, &
                            kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                            RI_basis_parameter, RI_basis_info, basis_S0, &
                            open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, para_env, unit_nr, &
                            .TRUE.)

      ! ! Calculate function (DI) derivatives with respect to the RI basis exponent
      CALL calc_energy_func_der(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, DI, &
                                Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, eps_step, &
                                qs_env, nkind, natom, dimen, dimen_RI, homo, virtual, &
                                kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                                RI_basis_parameter, RI_basis_info, basis_S0, &
                                open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, &
                                para_env, para_env_sub, number_groups, color_sub, unit_nr, &
                                p, lower_B, max_dev, &
                                deriv)

      g(:) = deriv
      xi(:) = -g

      reset_edge = 1.5_dp
      DO iiter = 1, max_num_iter
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,I5)') 'OPTIMIZATION STEP NUMBER', iiter

         ! perform step
         pnew(:) = p+xi
         CALL p2basis(nkind, RI_basis_parameter, lower_B, max_dev, pnew)

         ! check if we have to reset boundaries
         reset_boundary = .FALSE.
         ! nreset=0
         i = 0
         DO ikind = 1, nkind
            DO iset = 1, RI_basis_parameter(ikind)%nset
               i = i+1
               CALL transf_val(lower_B(i), max_dev(i), pnew(i), expon)
               IF (ABS(pnew(i)) > reset_edge .OR. expon < exp_limits(1, ikind) .OR. expon > exp_limits(2, ikind)) THEN
                  reset_boundary = .TRUE.
                  EXIT
               END IF
            END DO
         END DO
         ! IF(nreset>1) reset_boundary=.TRUE.

         IF (reset_boundary) THEN
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'RESET BASIS: one of the exponent hits the boundary'
            CALL reset_basis(nkind, ndof, RI_basis_parameter, reset_edge, &
                             pnew, lower_B, max_dev, max_rel_dev, exp_limits)
            p(:) = pnew
            xi = 0.0_dp
            g = 0.0_dp
            dg = 0.0_dp
            hdg = 0.0_dp
            pnew = 0.0_dp
            hessin = 0.0_dp
            DO i = 1, ndof
               hessin(i, i) = 1.0_dp
            END DO
            deriv = 0.0_dp
            ! Calculate RI-MO-ERI's
            CALL calc_energy_func(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, Emp2_RI, DRI, DI, &
                                  Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, &
                                  qs_env, natom, dimen, dimen_RI, homo, virtual, &
                                  kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                                  RI_basis_parameter, RI_basis_info, basis_S0, &
                                  open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, para_env, unit_nr, &
                                  .FALSE.)
            ! ! Calculate function (DI) derivatives with respect to the RI basis exponent
            CALL calc_energy_func_der(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, DI, &
                                      Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, eps_step, &
                                      qs_env, nkind, natom, dimen, dimen_RI, homo, virtual, &
                                      kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                                      RI_basis_parameter, RI_basis_info, basis_S0, &
                                      open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, &
                                      para_env, para_env_sub, number_groups, color_sub, unit_nr, &
                                      p, lower_B, max_dev, &
                                      deriv)

            g(:) = deriv
            xi(:) = -g
            pnew(:) = p+xi
            CALL p2basis(nkind, RI_basis_parameter, lower_B, max_dev, pnew)
         END IF

         ! calculate energy at the new point
         CALL calc_energy_func(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, Emp2_RI_new, DRI_new, DI_new, &
                               Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, &
                               qs_env, natom, dimen, dimen_RI, homo, virtual, &
                               kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                               RI_basis_parameter, RI_basis_info, basis_S0, &
                               open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, para_env, unit_nr, &
                               .FALSE.)

         ! update energy and direction
         DI = DI_new
         xi(:) = pnew-p
         p(:) = pnew

         ! check for convergence
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, *)
            DO ikind = 1, nkind
               CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=ri_aux_basis, &
                                basis_type="RI_AUX")
               WRITE (unit_nr, '(T3,A,A)') atomic_kind_set(ikind)%element_symbol, '   RI_opt_basis'
               WRITE (unit_nr, '(T3,I3)') RI_basis_parameter(ikind)%nset
               DO iset = 1, RI_basis_parameter(ikind)%nset
                  WRITE (unit_nr, '(T3,10I4)') iset, &
                     RI_basis_parameter(ikind)%lmin(iset), &
                     RI_basis_parameter(ikind)%lmax(iset), &
                     RI_basis_parameter(ikind)%npgf(iset), &
                     (1, ishell=1, RI_basis_parameter(ikind)%nshell(iset))
                  DO ipgf = 1, RI_basis_parameter(ikind)%npgf(iset)
                     WRITE (unit_nr, '(T3,10F16.10)') RI_basis_parameter(ikind)%zet(ipgf, iset), &
                        (ri_aux_basis%gcc(ipgf, ishell, iset), &
                         ishell=1, ri_aux_basis%nshell(iset))
                  END DO
               END DO
               WRITE (unit_nr, *)
            END DO
            WRITE (unit_nr, *)
         END IF
         IF (DI/ABS(Emp2) <= eps_DI_rel .AND. ABS(DRI_new) <= eps_DRI) THEN
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'OPTIMIZATION CONVERGED'
            IF (unit_nr > 0) WRITE (unit_nr, *)
            EXIT
         END IF

         ! calculate gradients
         CALL calc_energy_func_der(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, DI, &
                                   Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, eps_step, &
                                   qs_env, nkind, natom, dimen, dimen_RI, homo, virtual, &
                                   kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                                   RI_basis_parameter, RI_basis_info, basis_S0, &
                                   open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, &
                                   para_env, para_env_sub, number_groups, color_sub, unit_nr, &
                                   p, lower_B, max_dev, &
                                   deriv)

         ! g is the vector containing the old gradient
         dg(:) = deriv-g
         g(:) = deriv
         hdg(:) = MATMUL(hessin, dg)

         fac = SUM(dg*xi)
         fae = SUM(dg*hdg)
         sumdg = SUM(dg*dg)
         sumxi = SUM(xi*xi)

         hess_up = .TRUE.
         IF (fac**2 > sumdg*sumxi*3.0E-8_dp) THEN
            fac = 1.0_dp/fac
            fad = 1.0_dp/fae
            dg(:) = fac*xi-fad*hdg
            DO i = 1, ndof
               DO j = 1, ndof
                  hessin(i, j) = hessin(i, j)+fac*xi(i)*xi(j) &
                                 -fad*hdg(i)*hdg(j) &
                                 +fae*dg(i)*dg(j)
               END DO
            END DO
         ELSE
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'Skip Hessian Update'
            hess_up = .FALSE.
         END IF

         !XXXXXXXXXXXXX
         ! DO i=1, ndof
         !   IF(g(i)<0.0D+00) THEN
         !     xi(i)=MAX(0.05,ABS(g(i)))
         !   ELSE
         !     xi(i)=-MAX(0.05,ABS(g(i)))
         !   END IF
         ! END DO
         !XXXXXXXXXXXXX

         ! new direction
         xi(:) = -MATMUL(hessin, g)

      END DO

      IF (.NOT. (DI/ABS(Emp2) <= eps_DI_rel .AND. ABS(DRI_new) <= eps_DRI)) THEN
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,I5,A)') 'OPTIMIZATION NOT CONVERGED IN', max_num_iter, ' STEPS.'
         IF (unit_nr > 0) WRITE (unit_nr, *)
      END IF

      DEALLOCATE (max_rel_dev)

      DEALLOCATE (p)
      DEALLOCATE (xi)
      DEALLOCATE (g)
      DEALLOCATE (pnew)
      DEALLOCATE (dg)
      DEALLOCATE (hdg)
      DEALLOCATE (deriv)
      DEALLOCATE (Hessin)
      DEALLOCATE (lower_B)
      DEALLOCATE (max_dev)
      DEALLOCATE (exp_limits)

      IF (open_shell_case) THEN
         DEALLOCATE (Integ_MP2_AA)
         DEALLOCATE (Integ_MP2_BB)
         DEALLOCATE (Integ_MP2_AB)
      ELSE
         DEALLOCATE (Integ_MP2)
      END IF
      DEALLOCATE (index_table_RI)

      ! Release RI basis set
      CALL release_RI_basis_set(RI_basis_parameter, basis_S0)

      CALL cp_para_env_release(para_env_sub)
      CALL cp_rm_default_logger()
      CALL cp_logger_release(logger_sub)

      CALL timestop(handle)

   END SUBROUTINE optimize_ri_basis_main

! **************************************************************************************************
!> \brief ...
!> \param Emp2 ...
!> \param Emp2_AA ...
!> \param Emp2_BB ...
!> \param Emp2_AB ...
!> \param DI_ref ...
!> \param Integ_MP2 ...
!> \param Integ_MP2_AA ...
!> \param Integ_MP2_BB ...
!> \param Integ_MP2_AB ...
!> \param eps ...
!> \param qs_env ...
!> \param nkind ...
!> \param natom ...
!> \param dimen ...
!> \param dimen_RI ...
!> \param homo ...
!> \param virtual ...
!> \param kind_of ...
!> \param index_table_RI ...
!> \param mp2_biel ...
!> \param mp2_env ...
!> \param Auto ...
!> \param C ...
!> \param RI_basis_parameter ...
!> \param RI_basis_info ...
!> \param basis_S0 ...
!> \param open_shell_case ...
!> \param homo_beta ...
!> \param virtual_beta ...
!> \param Auto_beta ...
!> \param C_beta ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param number_groups ...
!> \param color_sub ...
!> \param unit_nr ...
!> \param p ...
!> \param lower_B ...
!> \param max_dev ...
!> \param deriv ...
! **************************************************************************************************
   SUBROUTINE calc_energy_func_der(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, DI_ref, &
                                   Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, eps, &
                                   qs_env, nkind, natom, dimen, dimen_RI, homo, virtual, &
                                   kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                                   RI_basis_parameter, RI_basis_info, basis_S0, &
                                   open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, &
                                   para_env, para_env_sub, number_groups, color_sub, unit_nr, &
                                   p, lower_B, max_dev, &
                                   deriv)
      REAL(KIND=dp)                                      :: Emp2, Emp2_AA, Emp2_BB, Emp2_AB, DI_ref
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, &
                                                            Integ_MP2_AB
      REAL(KIND=dp)                                      :: eps
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER                                            :: nkind, natom, dimen, dimen_RI, homo, &
                                                            virtual
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: index_table_RI
      TYPE(mp2_biel_type)                                :: mp2_biel
      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Auto
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: C
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter
      TYPE(hfx_basis_info_type)                          :: RI_basis_info
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_S0
      LOGICAL                                            :: open_shell_case
      INTEGER                                            :: homo_beta, virtual_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Auto_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: C_beta
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER                                            :: number_groups, color_sub, unit_nr
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: p, lower_B, max_dev, deriv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_energy_func_der', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ideriv, ikind, iset, nseta
      REAL(KIND=dp)                                      :: DI, DRI, Emp2_RI, new_basis_val, &
                                                            orig_basis_val
      REAL(KIND=dp), VOLATILE                            :: step, temp

      CALL timeset(routineN, handle)

      step = eps

      ! cycle over the RI basis set exponent
      deriv = 0.0_dp
      ideriv = 0
      DO ikind = 1, nkind
         nseta = RI_basis_parameter(ikind)%nset
         DO iset = 1, nseta
            ! for now only uncontracted aux basis set
            ideriv = ideriv+1
            IF (MOD(ideriv, number_groups) /= color_sub) CYCLE

            ! ! calculate the numerical derivative
            ! ! The eps is the relative change of the exponent for the
            ! ! calculation of the numerical derivative
            ! CPPostcondition(RI_basis_parameter(ikind)%npgf(iset)==1,cp_failure_level,routineP,failure)
            ! step=eps*RI_basis_parameter(ikind)%zet(1,iset)
            ! temp=RI_basis_parameter(ikind)%zet(1,iset)+step
            ! step=temp-RI_basis_parameter(ikind)%zet(1,iset)
            ! RI_basis_parameter(ikind)%zet(1,iset)=RI_basis_parameter(ikind)%zet(1,iset)+step

            ! in the new case eps is just the step length for calculating the numerical derivative

            IF(.NOT.(RI_basis_parameter(ikind)%npgf(iset) == 1))CALL cp__a("mp2_optimize_ri_basis.F",662)
            orig_basis_val = RI_basis_parameter(ikind)%zet(1, iset)
            temp = p(ideriv)+step
            CALL transf_val(lower_B(ideriv), max_dev(ideriv), temp, new_basis_val)
            RI_basis_parameter(ikind)%zet(1, iset) = new_basis_val

            CALL calc_energy_func(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, Emp2_RI, DRI, DI, &
                                  Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, &
                                  qs_env, natom, dimen, dimen_RI, homo, virtual, &
                                  kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                                  RI_basis_parameter, RI_basis_info, basis_S0, &
                                  open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, &
                                  para_env_sub, unit_nr, .TRUE.)

            RI_basis_parameter(ikind)%zet(1, iset) = orig_basis_val

            IF (para_env_sub%mepos == 0) THEN
               temp = EXP(DI)
               temp = temp/EXP(DI_ref)
               deriv(ideriv) = LOG(temp)/step
            END IF

         END DO
      END DO

      CALL mp_sum(deriv, para_env%group)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param Emp2 ...
!> \param Emp2_AA ...
!> \param Emp2_BB ...
!> \param Emp2_AB ...
!> \param Emp2_RI ...
!> \param DRI ...
!> \param DI ...
!> \param Integ_MP2 ...
!> \param Integ_MP2_AA ...
!> \param Integ_MP2_BB ...
!> \param Integ_MP2_AB ...
!> \param qs_env ...
!> \param natom ...
!> \param dimen ...
!> \param dimen_RI ...
!> \param homo ...
!> \param virtual ...
!> \param kind_of ...
!> \param index_table_RI ...
!> \param mp2_biel ...
!> \param mp2_env ...
!> \param Auto ...
!> \param C ...
!> \param RI_basis_parameter ...
!> \param RI_basis_info ...
!> \param basis_S0 ...
!> \param open_shell_case ...
!> \param homo_beta ...
!> \param virtual_beta ...
!> \param Auto_beta ...
!> \param C_beta ...
!> \param para_env ...
!> \param unit_nr ...
!> \param no_write ...
! **************************************************************************************************
   SUBROUTINE calc_energy_func(Emp2, Emp2_AA, Emp2_BB, Emp2_AB, Emp2_RI, DRI, DI, &
                               Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, Integ_MP2_AB, &
                               qs_env, natom, dimen, dimen_RI, homo, virtual, &
                               kind_of, index_table_RI, mp2_biel, mp2_env, Auto, C, &
                               RI_basis_parameter, RI_basis_info, basis_S0, &
                               open_shell_case, homo_beta, virtual_beta, Auto_beta, C_beta, para_env, unit_nr, &
                               no_write)
      REAL(KIND=dp)                                      :: Emp2, Emp2_AA, Emp2_BB, Emp2_AB, &
                                                            Emp2_RI, DRI, DI
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: Integ_MP2, Integ_MP2_AA, Integ_MP2_BB, &
                                                            Integ_MP2_AB
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER                                            :: natom, dimen, dimen_RI, homo, virtual
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: index_table_RI
      TYPE(mp2_biel_type)                                :: mp2_biel
      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Auto
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: C
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter
      TYPE(hfx_basis_info_type)                          :: RI_basis_info
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_S0
      LOGICAL                                            :: open_shell_case
      INTEGER                                            :: homo_beta, virtual_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Auto_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: C_beta
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: unit_nr
      LOGICAL                                            :: no_write

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_energy_func', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: DI_AA, DI_AB, DI_BB, DRI_AA, DRI_AB, &
                                                            DRI_BB, Emp2_RI_AA, Emp2_RI_AB, &
                                                            Emp2_RI_BB
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: Lai, Lai_beta

      CALL timeset(routineN, handle)

      CALL libint_ri_mp2(dimen, dimen_RI, homo, natom, mp2_biel, mp2_env, C, &
                         kind_of, RI_basis_parameter, RI_basis_info, basis_S0, index_table_RI, &
                         qs_env, para_env, Lai)
      IF (open_shell_case) THEN
         CALL libint_ri_mp2(dimen, dimen_RI, homo_beta, natom, mp2_biel, mp2_env, C_beta, &
                            kind_of, RI_basis_parameter, RI_basis_info, basis_S0, index_table_RI, &
                            qs_env, para_env, Lai_beta)
      END IF

      ! Contract integrals into energy
      IF (open_shell_case) THEN
         ! alpha-alpha
         CALL contract_integrals(DI_AA, Emp2_RI_AA, DRI_AA, Emp2_AA, homo, homo, virtual, virtual, &
                                 1.0_dp, 0.5_dp, .TRUE., &
                                 Auto, Auto, Integ_MP2_AA, &
                                 Lai, Lai, para_env)

         ! beta-beta
         CALL contract_integrals(DI_BB, Emp2_RI_BB, DRI_BB, Emp2_BB, homo_beta, homo_beta, virtual_beta, virtual_beta, &
                                 1.0_dp, 0.5_dp, .TRUE., &
                                 Auto_beta, Auto_beta, Integ_MP2_BB, &
                                 Lai_beta, Lai_beta, para_env)

         ! alpha-beta
         CALL contract_integrals(DI_AB, Emp2_RI_AB, DRI_AB, Emp2_AB*2.0_dp, homo, homo_beta, virtual, virtual_beta, &
                                 1.0_dp, 1.0_dp, .FALSE., &
                                 Auto, Auto_beta, Integ_MP2_AB, &
                                 Lai, Lai_beta, para_env)

         Emp2_RI = Emp2_RI_AA+Emp2_RI_BB+Emp2_RI_AB
         DRI = DRI_AA+DRI_BB+DRI_AB
         DI = DI_AA+DI_BB+DI_AB
      ELSE
         CALL contract_integrals(DI, Emp2_RI, DRI, Emp2, homo, homo, virtual, virtual, &
                                 2.0_dp, 1.0_dp, .TRUE., &
                                 Auto, Auto, Integ_MP2, &
                                 Lai, Lai, para_env)
      END IF

      IF (.NOT. no_write) THEN
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)')
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'Emp2 =   ', Emp2
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'Emp2-RI =', Emp2_RI
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,ES25.10)') 'DRI =    ', DRI
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,ES25.10)') 'DI =     ', DI
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,ES25.10)') 'DI/|Emp2| =     ', DI/ABS(Emp2)
      END IF

      DEALLOCATE (Lai)
      IF (open_shell_case) DEALLOCATE (Lai_beta)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param nkind ...
!> \param RI_basis_parameter ...
!> \param lower_B ...
!> \param max_dev ...
!> \param max_rel_dev ...
! **************************************************************************************************
   SUBROUTINE init_transf(nkind, RI_basis_parameter, lower_B, max_dev, max_rel_dev)
      INTEGER                                            :: nkind
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: lower_B, max_dev, max_rel_dev

      INTEGER                                            :: ikind, ipos, iset

      ipos = 0
      DO ikind = 1, nkind
         DO iset = 1, RI_basis_parameter(ikind)%nset
            ipos = ipos+1
            lower_B(ipos) = RI_basis_parameter(ikind)%zet(1, iset)*(1.0_dp-max_rel_dev(ipos))
            max_dev(ipos) = RI_basis_parameter(ikind)%zet(1, iset)*2.0_dp*max_rel_dev(ipos)
         END DO
      END DO

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param nkind ...
!> \param RI_basis_parameter ...
!> \param Lower_B ...
!> \param max_dev ...
!> \param p ...
! **************************************************************************************************
   SUBROUTINE p2basis(nkind, RI_basis_parameter, Lower_B, max_dev, p)
      INTEGER                                            :: nkind
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Lower_B, max_dev, p

      INTEGER                                            :: ikind, ipos, iset
      REAL(KIND=dp)                                      :: valout

      ipos = 0
      DO ikind = 1, nkind
         DO iset = 1, RI_basis_parameter(ikind)%nset
            ipos = ipos+1
            CALL transf_val(lower_B(ipos), max_dev(ipos), p(ipos), valout)
            RI_basis_parameter(ikind)%zet(1, iset) = valout
         END DO
      END DO

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param nkind ...
!> \param ndof ...
!> \param RI_basis_parameter ...
!> \param reset_edge ...
!> \param pnew ...
!> \param lower_B ...
!> \param max_dev ...
!> \param max_rel_dev ...
!> \param exp_limits ...
! **************************************************************************************************
   SUBROUTINE reset_basis(nkind, ndof, RI_basis_parameter, reset_edge, &
                          pnew, lower_B, max_dev, max_rel_dev, exp_limits)
      INTEGER                                            :: nkind, ndof
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: RI_basis_parameter
      REAL(KIND=dp)                                      :: reset_edge
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: pnew, Lower_B, max_dev, max_rel_dev
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: exp_limits

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reset_basis', routineP = moduleN//':'//routineN

      INTEGER                                            :: am_max, iexpo, ikind, ipos, ipos_p, &
                                                            iset, la
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nf_per_l
      INTEGER, DIMENSION(ndof)                           :: change_expo
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: has_to_be_changed
      REAL(KIND=dp)                                      :: expo, geom_fact, pmax
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: max_min_exp_per_l
      REAL(KIND=dp), DIMENSION(ndof)                     :: new_expo, old_expo, old_lower_B, &
                                                            old_max_dev, old_max_rel_dev, old_pnew

! make a copy of the original parameters

      old_pnew = pnew
      old_lower_B = lower_B
      old_max_dev = max_dev
      old_max_rel_dev = max_rel_dev
      old_expo = 0.0_dp
      ipos = 0
      DO ikind = 1, nkind
         DO iset = 1, RI_basis_parameter(ikind)%nset
            ipos = ipos+1
            old_expo(ipos) = RI_basis_parameter(ikind)%zet(1, iset)
         END DO
      END DO

      pnew = 0.0_dp
      lower_B = 0.0_dp
      max_dev = 0.0_dp
      max_rel_dev = 0.0_dp

      change_expo = 0

      new_expo = 0.0_dp
      ipos = 0
      ipos_p = 0
      DO ikind = 1, nkind
         am_max = MAXVAL(RI_basis_parameter(ikind)%lmax(:))
         ALLOCATE (nf_per_l(0:am_max))
         nf_per_l = 0
         ALLOCATE (max_min_exp_per_l(2, 0:am_max))
         max_min_exp_per_l(1, :) = HUGE(0)
         max_min_exp_per_l(2, :) = -HUGE(0)

         DO iset = 1, RI_basis_parameter(ikind)%nset
            la = RI_basis_parameter(ikind)%lmax(iset)
            expo = RI_basis_parameter(ikind)%zet(1, iset)
            nf_per_l(la) = nf_per_l(la)+1
            IF (expo <= max_min_exp_per_l(1, la)) max_min_exp_per_l(1, la) = expo
            IF (expo >= max_min_exp_per_l(2, la)) max_min_exp_per_l(2, la) = expo
         END DO

         max_min_exp_per_l(1, la) = MAX(max_min_exp_per_l(1, la), exp_limits(1, ikind))
         max_min_exp_per_l(2, la) = MIN(max_min_exp_per_l(2, la), exp_limits(2, ikind))

         ! always s exponents as maximum and minimu
         ! expo=MAXVAL(max_min_exp_per_l(2,:))
         ! max_min_exp_per_l(2,0)=expo
         ! expo=MINVAL(max_min_exp_per_l(1,:))
         ! max_min_exp_per_l(1,0)=expo

         ALLOCATE (has_to_be_changed(0:am_max))
         has_to_be_changed = .FALSE.
         DO la = 0, am_max
            pmax = -HUGE(0)
            DO iexpo = 1, nf_per_l(la)
               ipos_p = ipos_p+1
               IF (ABS(old_pnew(ipos_p)) >= pmax) pmax = ABS(old_pnew(ipos_p))
               ! check if any of the exponents go out of range
               CALL transf_val(old_lower_B(ipos_p), old_max_dev(ipos_p), old_pnew(ipos_p), expo)
               IF (expo < exp_limits(1, ikind) .OR. expo > exp_limits(2, ikind)) has_to_be_changed(la) = .TRUE.
            END DO
            IF (pmax > reset_edge) has_to_be_changed(la) = .TRUE.
         END DO

         DO la = 0, am_max
            IF (nf_per_l(la) == 1) THEN
               ipos = ipos+1
               new_expo(ipos) = max_min_exp_per_l(1, la)
               IF (new_expo(ipos) >= exp_limits(1, ikind) .AND. new_expo(ipos) <= exp_limits(2, ikind)) THEN
                  max_rel_dev(ipos) = (new_expo(ipos)-old_lower_B(ipos))/new_expo(ipos)
                  IF (max_rel_dev(ipos) <= 0.1_dp) max_rel_dev(ipos) = 0.8_dp
               ELSE
                  new_expo(ipos) = (exp_limits(2, ikind)+exp_limits(1, ikind))/2.0_dp
                  max_rel_dev(ipos) = (new_expo(ipos)-exp_limits(1, ikind))/new_expo(ipos)
               END IF
               IF (has_to_be_changed(la)) change_expo(ipos) = 1
            ELSE
               IF (max_min_exp_per_l(2, la)/max_min_exp_per_l(1, la) < 2.0_dp) THEN
                  max_min_exp_per_l(1, la) = max_min_exp_per_l(1, la)*0.5
                  max_min_exp_per_l(2, la) = max_min_exp_per_l(2, la)*1.5
               END IF
               geom_fact = (max_min_exp_per_l(2, la)/max_min_exp_per_l(1, la))**(1.0_dp/REAL(nf_per_l(la)-1, dp))
               DO iexpo = 1, nf_per_l(la)
                  ipos = ipos+1
                  new_expo(ipos) = max_min_exp_per_l(1, la)*(geom_fact**(iexpo-1))
                  max_rel_dev(ipos) = (geom_fact-1.0_dp)/(geom_fact+1.0_dp)*0.9_dp
                  IF (has_to_be_changed(la)) change_expo(ipos) = 1
               END DO
            END IF
         END DO

         DEALLOCATE (has_to_be_changed)

         DEALLOCATE (nf_per_l)
         DEALLOCATE (max_min_exp_per_l)
      END DO

      ipos = 0
      DO ikind = 1, nkind
         DO iset = 1, RI_basis_parameter(ikind)%nset
            ipos = ipos+1
            RI_basis_parameter(ikind)%zet(1, iset) = new_expo(ipos)
         END DO
      END DO

      CALL init_transf(nkind, RI_basis_parameter, lower_B, max_dev, max_rel_dev)

      ipos = 0
      DO ikind = 1, nkind
         DO iset = 1, RI_basis_parameter(ikind)%nset
            ipos = ipos+1
            IF (change_expo(ipos) == 0) THEN
               ! restore original
               pnew(ipos) = old_pnew(ipos)
               lower_B(ipos) = old_lower_B(ipos)
               max_dev(ipos) = old_max_dev(ipos)
               max_rel_dev(ipos) = old_max_rel_dev(ipos)
               RI_basis_parameter(ikind)%zet(1, iset) = old_expo(ipos)
            END IF
         END DO
      END DO

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param DI ...
!> \param Emp2_RI ...
!> \param DRI ...
!> \param Emp2 ...
!> \param homo ...
!> \param homo_beta ...
!> \param virtual ...
!> \param virtual_beta ...
!> \param fact ...
!> \param fact2 ...
!> \param calc_ex ...
!> \param MOenerg ...
!> \param MOenerg_beta ...
!> \param abij ...
!> \param Lai ...
!> \param Lai_beta ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE contract_integrals(DI, Emp2_RI, DRI, Emp2, homo, homo_beta, virtual, virtual_beta, &
                                 fact, fact2, calc_ex, &
                                 MOenerg, MOenerg_beta, abij, &
                                 Lai, Lai_beta, para_env)
      REAL(KIND=dp)                                      :: DI, Emp2_RI, DRI, Emp2
      INTEGER                                            :: homo, homo_beta, virtual, virtual_beta
      REAL(KIND=dp)                                      :: fact, fact2
      LOGICAL                                            :: calc_ex
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: MOenerg, MOenerg_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: abij
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: Lai, Lai_beta
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'contract_integrals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: a, b, i, ij_counter, j
      REAL(KIND=dp)                                      :: t_iajb, t_iajb_RI
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: mat_ab

      ALLOCATE (mat_ab(virtual, virtual_beta))

      DI = 0.0_dp
      Emp2_RI = 0.0_dp
      ij_counter = 0
      DO j = 1, homo_beta
         DO i = 1, homo
            ij_counter = ij_counter+1
            IF (MOD(ij_counter, para_env%num_pe) /= para_env%mepos) CYCLE
            mat_ab = 0.0_dp
            mat_ab(:, :) = MATMUL(TRANSPOSE(Lai(:, :, i)), Lai_beta(:, :, j))
            DO b = 1, virtual_beta
               DO a = 1, virtual
                  IF (calc_ex) THEN
                     t_iajb = fact*abij(a, b, i, j)-abij(b, a, i, j)
                     t_iajb_RI = fact*mat_ab(a, b)-mat_ab(b, a)
                  ELSE
                     t_iajb = fact*abij(a, b, i, j)
                     t_iajb_RI = fact*mat_ab(a, b)
                  END IF
                  t_iajb = t_iajb/(MOenerg(a+homo)+MOenerg_beta(b+homo_beta)-MOenerg(i)-MOenerg_beta(j))
                  t_iajb_RI = t_iajb_RI/(MOenerg(a+homo)+MOenerg_beta(b+homo_beta)-MOenerg(i)-MOenerg_beta(j))

                  Emp2_RI = Emp2_RI-t_iajb_RI*mat_ab(a, b)*fact2

                  DI = DI-t_iajb*mat_ab(a, b)*fact2

               END DO
            END DO
         END DO
      END DO
      CALL mp_sum(DI, para_env%group)
      CALL mp_sum(Emp2_RI, para_env%group)

      DRI = Emp2-Emp2_RI
      DI = 2.0D+00*DI-Emp2-Emp2_RI

      DEALLOCATE (mat_ab)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param homo ...
!> \param homo_beta ...
!> \param para_env ...
!> \param elements_ij_proc ...
!> \param ij_list_proc ...
! **************************************************************************************************
   SUBROUTINE calc_elem_ij_proc(homo, homo_beta, para_env, elements_ij_proc, ij_list_proc)
      INTEGER                                            :: homo, homo_beta
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: elements_ij_proc
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ij_list_proc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_elem_ij_proc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij_counter, j

      elements_ij_proc = 0
      ij_counter = -1
      DO i = 1, homo
         DO j = 1, homo_beta
            ij_counter = ij_counter+1
            IF (MOD(ij_counter, para_env%num_pe) == para_env%mepos) elements_ij_proc = elements_ij_proc+1
         END DO
      END DO

      ALLOCATE (ij_list_proc(elements_ij_proc, 2))
      ij_list_proc = 0
      ij_counter = -1
      elements_ij_proc = 0
      DO i = 1, homo
         DO j = 1, homo_beta
            ij_counter = ij_counter+1
            IF (MOD(ij_counter, para_env%num_pe) == para_env%mepos) THEN
               elements_ij_proc = elements_ij_proc+1
               ij_list_proc(elements_ij_proc, 1) = i
               ij_list_proc(elements_ij_proc, 2) = j
            END IF
         END DO
      END DO

   END SUBROUTINE calc_elem_ij_proc

! **************************************************************************************************
!> \brief ...
!> \param lower_B ...
!> \param max_dev ...
!> \param valin ...
!> \param valout ...
! **************************************************************************************************
   SUBROUTINE transf_val(lower_B, max_dev, valin, valout)
      REAL(KIND=dp)                                      :: lower_B, max_dev, valin, valout

      REAL(KIND=dp), PARAMETER                           :: alpha = 2.633915794_dp

      valout = 0.0_dp
      valout = lower_B+max_dev/(1.0_dp+EXP(-alpha*valin))

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param mp2_env ...
!> \param nkind ...
!> \param max_rel_dev_output ...
!> \param basis_was_assoc ...
! **************************************************************************************************
   SUBROUTINE generate_RI_init_basis(qs_env, mp2_env, nkind, max_rel_dev_output, basis_was_assoc)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mp2_type), POINTER                            :: mp2_env
      INTEGER                                            :: nkind
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: max_rel_dev_output
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: basis_was_assoc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'generate_RI_init_basis', &
         routineP = moduleN//':'//routineN

      INTEGER :: basis_quality, handle, iexpo, iii, ikind, ipgf, iset, ishell, jexpo, jjj, la, &
         max_am, nexpo_shell, nseta, nsgfa, nsgfa_RI, prog_func, prog_l, RI_max_am, RI_nset, &
         RI_prev_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: l_expo, num_sgf_per_l, ordered_pos, &
                                                            RI_l_expo, RI_num_sgf_per_l, &
                                                            tot_num_exp_per_l
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: l_tab
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, npgfa, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, nl
      LOGICAL                                            :: external_num_of_func
      REAL(dp)                                           :: geom_fact
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: exponents, RI_exponents
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: exp_tab, max_min_exp_l
      REAL(dp), DIMENSION(:, :), POINTER                 :: sphi_a, zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: gcc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: max_rel_dev, max_rel_dev_prev
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_a, tmp_basis
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: atom_kind

      CALL timeset(routineN, handle)

      basis_quality = mp2_env%ri_opt_param%basis_quality
      external_num_of_func = .FALSE.
      IF (ALLOCATED(mp2_env%ri_opt_param%RI_nset_per_l)) external_num_of_func = .TRUE.

      NULLIFY (qs_kind_set)
      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set)
      nkind = SIZE(qs_kind_set, 1)

      ALLOCATE (basis_was_assoc(nkind))
      basis_was_assoc = .FALSE.

      IF (external_num_of_func .AND. nkind > 1) THEN
         CALL cp_warn(cp__l("mp2_optimize_ri_basis.F",1230), &
                      "There are more than one kind of atom. The same pattern of functions, "// &
                      "as specified by NUM_FUNC, will be used for all kinds.")
      END IF

      DO ikind = 1, nkind
         NULLIFY (atom_kind)
         atom_kind => qs_kind_set(ikind)

         CALL get_qs_kind(qs_kind=atom_kind, basis_set=orb_basis_a, basis_type="RI_AUX")
         ! save info if the basis was or not associated
         basis_was_assoc(ikind) = ASSOCIATED(orb_basis_a)

         ! skip the generation of the initial guess if the RI basis is
         ! provided in input
         IF (basis_was_assoc(ikind)) THEN
            nseta = orb_basis_a%nset
            la_max => orb_basis_a%lmax
            la_min => orb_basis_a%lmin
            npgfa => orb_basis_a%npgf
            nshell => orb_basis_a%nshell
            zet => orb_basis_a%zet
            nl => orb_basis_a%l

            max_am = MAXVAL(la_max)

            RI_max_am = max_am
            ALLOCATE (RI_num_sgf_per_l(0:RI_max_am))
            RI_num_sgf_per_l = 0
            RI_nset = 0
            DO iset = 1, nseta
               DO la = la_min(iset), la_max(iset)
                  RI_nset = RI_nset+1
                  RI_num_sgf_per_l(la) = RI_num_sgf_per_l(la)+1
                  IF (npgfa(iset) > 1) THEN
                     CALL cp_warn(cp__l("mp2_optimize_ri_basis.F",1265), &
                                  "The RI basis set optimizer can not handle contracted Gaussian. "// &
                                  "Calculation continue with only uncontracted functions.")
                  END IF
               END DO
            END DO

            ALLOCATE (exp_tab(MAXVAL(RI_num_sgf_per_l), 0:RI_max_am))
            exp_tab = 0.0_dp
            DO iii = 0, RI_max_am
               iexpo = 0
               DO iset = 1, nseta
                  DO la = la_min(iset), la_max(iset)
                     IF (la /= iii) CYCLE
                     iexpo = iexpo+1
                     exp_tab(iexpo, iii) = zet(1, iset)
                  END DO
               END DO
            END DO

            ! sort exponents
            DO iii = 0, RI_max_am
               ALLOCATE (ordered_pos(RI_num_sgf_per_l(iii)))
               ordered_pos = 0
               CALL sort(exp_tab(1:RI_num_sgf_per_l(iii), iii), RI_num_sgf_per_l(iii), ordered_pos)
               DEALLOCATE (ordered_pos)
            END DO

            ALLOCATE (RI_l_expo(RI_nset))
            RI_l_expo = 0
            ALLOCATE (RI_exponents(RI_nset))
            RI_exponents = 0.0_dp

            iset = 0
            DO iii = 0, RI_max_am
               DO iexpo = 1, RI_num_sgf_per_l(iii)
                  iset = iset+1
                  RI_l_expo(iset) = iii
                  RI_exponents(iset) = exp_tab(iexpo, iii)
               END DO
            END DO
            DEALLOCATE (exp_tab)

            ALLOCATE (max_rel_dev(RI_nset))
            max_rel_dev = 1.0_dp
            iset = 0
            DO iii = 0, RI_max_am
               IF (RI_num_sgf_per_l(iii) == 1) THEN
                  iset = iset+1
                  max_rel_dev(iset) = 0.35_dp
               ELSE
                  iset = iset+1
                  max_rel_dev(iset) = (RI_exponents(iset+1)+RI_exponents(iset))/2.0_dp
                  max_rel_dev(iset) = max_rel_dev(iset)/RI_exponents(iset)-1.0_dp
                  DO iexpo = 2, RI_num_sgf_per_l(iii)
                     iset = iset+1
                     max_rel_dev(iset) = (RI_exponents(iset)+RI_exponents(iset-1))/2.0_dp
                     max_rel_dev(iset) = 1.0_dp-max_rel_dev(iset)/RI_exponents(iset)
                  END DO
               END IF
            END DO
            max_rel_dev(:) = max_rel_dev(:)*0.9_dp
            DEALLOCATE (RI_num_sgf_per_l)

            ! deallocate the old basis before moving on
            CALL remove_basis_from_container(qs_kind_set(ikind)%basis_sets, basis_type="RI_AUX")

         ELSE

            CALL get_qs_kind(qs_kind=atom_kind, basis_set=orb_basis_a)

            sphi_a => orb_basis_a%sphi
            nseta = orb_basis_a%nset
            la_max => orb_basis_a%lmax
            la_min => orb_basis_a%lmin
            npgfa => orb_basis_a%npgf
            first_sgfa => orb_basis_a%first_sgf
            nshell => orb_basis_a%nshell
            zet => orb_basis_a%zet
            gcc => orb_basis_a%gcc
            nl => orb_basis_a%l
            nsgfa = orb_basis_a%nsgf

            max_am = MAXVAL(la_max)

            nexpo_shell = 0
            DO iset = 1, nseta
               DO ishell = 1, nshell(iset)
                  DO la = la_min(iset), la_max(iset)
                     IF (ishell > 1) THEN
                        IF (nl(ishell, iset) == nl(ishell-1, iset)) CYCLE
                     END IF
                     IF (la /= nl(ishell, iset)) CYCLE
                     nexpo_shell = nexpo_shell+npgfa(iset)
                  END DO
               END DO
            END DO

            ALLOCATE (exponents(nexpo_shell))
            exponents = 0.0_dp
            ALLOCATE (l_expo(nexpo_shell))
            l_expo = 0
            ALLOCATE (num_sgf_per_l(0:max_am))
            num_sgf_per_l = 0
            iexpo = 0
            DO iset = 1, nseta
               DO ishell = 1, nshell(iset)
                  DO la = la_min(iset), la_max(iset)
                     IF (ishell > 1) THEN
                        IF (nl(ishell, iset) == nl(ishell-1, iset)) CYCLE
                     END IF
                     IF (la /= nl(ishell, iset)) CYCLE
                     DO ipgf = 1, npgfa(iset)
                        iexpo = iexpo+1
                        exponents(iexpo) = zet(ipgf, iset)
                        l_expo(iexpo) = la
                     END DO
                     num_sgf_per_l(la) = num_sgf_per_l(la)+1
                  END DO
               END DO
            END DO

            ALLOCATE (exp_tab(nexpo_shell, nexpo_shell))
            exp_tab = 0.0_dp
            ALLOCATE (l_tab(nexpo_shell, nexpo_shell))
            l_tab = 0
            ALLOCATE (tot_num_exp_per_l(0:max_am*2))
            tot_num_exp_per_l = 0
            DO iexpo = 1, nexpo_shell
               DO jexpo = iexpo, nexpo_shell
                  exp_tab(jexpo, iexpo) = exponents(jexpo)+exponents(iexpo)
                  exp_tab(iexpo, jexpo) = exp_tab(jexpo, iexpo)
                  l_tab(jexpo, iexpo) = l_expo(jexpo)+l_expo(iexpo)
                  l_tab(iexpo, jexpo) = l_tab(jexpo, iexpo)
                  tot_num_exp_per_l(l_tab(jexpo, iexpo)) = tot_num_exp_per_l(l_tab(jexpo, iexpo))+1
               END DO
            END DO
            DEALLOCATE (l_expo)
            DEALLOCATE (exponents)

            ALLOCATE (max_min_exp_l(2, 0:max_am*2))
            max_min_exp_l(1, :) = HUGE(0)
            max_min_exp_l(2, :) = -HUGE(0)

            DO la = 0, max_am*2
               DO iexpo = 1, nexpo_shell
                  DO jexpo = iexpo, nexpo_shell
                     IF (la /= l_tab(jexpo, iexpo)) CYCLE
                     IF (exp_tab(jexpo, iexpo) <= max_min_exp_l(1, la)) max_min_exp_l(1, la) = exp_tab(jexpo, iexpo)
                     IF (exp_tab(jexpo, iexpo) >= max_min_exp_l(2, la)) max_min_exp_l(2, la) = exp_tab(jexpo, iexpo)
                  END DO
               END DO
            END DO
            DEALLOCATE (exp_tab)
            DEALLOCATE (l_tab)

            ! ! scale the limits of the exponents to avoid the optimizer to go out-of-range
            ! ! (0.2 just empirical parameter)
            max_min_exp_l(1, :) = max_min_exp_l(1, :)/0.80_dp
            max_min_exp_l(2, :) = max_min_exp_l(2, :)/1.20_dp

            ALLOCATE (RI_num_sgf_per_l(0:max_am*2))
            RI_num_sgf_per_l = 0

            SELECT CASE (basis_quality)
            CASE (0)
               ! normal
               prog_func = 0
               prog_l = 2
            CASE (1)
               ! large
               prog_func = 1
               prog_l = 2
            CASE (2)
               prog_func = 2
               prog_l = 2
            CASE DEFAULT
               prog_func = 0
               prog_l = 2
            END SELECT

            IF (external_num_of_func) THEN
               ! cp2k can not exceed angular momentum 7
               RI_max_am = MIN(SIZE(mp2_env%ri_opt_param%RI_nset_per_l)-1, 7)
               IF (RI_max_am > max_am*2) THEN
                  DEALLOCATE (RI_num_sgf_per_l)
                  ALLOCATE (RI_num_sgf_per_l(0:RI_max_am))
                  RI_num_sgf_per_l = 0
               END IF
               DO la = 0, RI_max_am
                  RI_num_sgf_per_l(la) = mp2_env%ri_opt_param%RI_nset_per_l(la)
               END DO
            ELSE
               RI_num_sgf_per_l(0) = num_sgf_per_l(0)*2+prog_func
               DO la = 1, max_am
                  RI_num_sgf_per_l(la) = RI_num_sgf_per_l(la-1)-1
                  IF (RI_num_sgf_per_l(la) == 0) THEN
                     RI_num_sgf_per_l(la) = 1
                     EXIT
                  END IF
               END DO
               DO la = max_am+1, max_am*2
                  RI_num_sgf_per_l(la) = RI_num_sgf_per_l(la-1)-prog_l
                  IF (RI_num_sgf_per_l(la) == 0) THEN
                     RI_num_sgf_per_l(la) = 1
                  END IF
                  IF (RI_num_sgf_per_l(la) == 1) EXIT
               END DO
               RI_max_am = MIN(max_am*2, 7)
               DO la = 0, MIN(max_am*2, 7)
                  IF (RI_num_sgf_per_l(la) == 0) THEN
                     RI_max_am = la-1
                     EXIT
                  END IF
               END DO

               iii = 0
               jjj = 0
               nsgfa_RI = 0
               DO la = 1, max_am*2
                  IF (tot_num_exp_per_l(la) >= iii) THEN
                     iii = tot_num_exp_per_l(la)
                     jjj = la
                  END IF
                  nsgfa_RI = nsgfa_RI+RI_num_sgf_per_l(la)*(la*2+1)
               END DO
               DEALLOCATE (tot_num_exp_per_l)
               IF (REAL(nsgfa_RI, KIND=dp)/REAL(nsgfa, KIND=dp) <= 2.5_dp) THEN
                  RI_num_sgf_per_l(jjj) = RI_num_sgf_per_l(jjj)+1
               END IF
            END IF

            RI_nset = SUM(RI_num_sgf_per_l)

            ALLOCATE (RI_exponents(RI_nset))
            RI_exponents = 0.0_dp

            ALLOCATE (RI_l_expo(RI_nset))
            RI_l_expo = 0

            ALLOCATE (max_rel_dev(RI_nset))
            max_rel_dev = 1.0_dp

            iset = 0
            DO la = 0, RI_max_am
               IF (RI_num_sgf_per_l(la) == 1) THEN
                  iset = iset+1
                  RI_exponents(iset) = (max_min_exp_l(2, la)+max_min_exp_l(1, la))/2.0_dp
                  RI_l_expo(iset) = la
                  geom_fact = max_min_exp_l(2, la)/max_min_exp_l(1, la)

                  max_rel_dev(iset) = (geom_fact-1.0_dp)/(geom_fact+1.0_dp)*0.9_dp
               ELSE
                  geom_fact = (max_min_exp_l(2, la)/max_min_exp_l(1, la))**(1.0_dp/REAL(RI_num_sgf_per_l(la)-1, dp))
                  DO iexpo = 1, RI_num_sgf_per_l(la)
                     iset = iset+1
                     RI_exponents(iset) = max_min_exp_l(1, la)*(geom_fact**(iexpo-1))
                     RI_l_expo(iset) = la
                     max_rel_dev(iset) = (geom_fact-1.0_dp)/(geom_fact+1.0_dp)*0.9_dp
                  END DO
               END IF
            END DO
            DEALLOCATE (num_sgf_per_l)
            DEALLOCATE (max_min_exp_l)
            DEALLOCATE (RI_num_sgf_per_l)

            ! IF(qs_env%para_env%mepos==0) THEN
            !   WRITE(*,*) RI_nset, RI_max_am, geom_fact
            !   iset=0
            !   DO la=0, RI_max_am
            !     DO iexpo=1, RI_num_sgf_per_l(la)
            !       iset=iset+1
            !       WRITE(*,*) RI_l_expo(iset), RI_exponents(iset)
            !     END DO
            !   END DO
            !   WRITE(*,*) nsgfa, nsgfa_RI
            ! END IF

         END IF ! RI basis not associated

         ! create the new basis
         NULLIFY (tmp_basis)
         CALL create_ri_basis(tmp_basis, RI_nset, RI_l_expo, RI_exponents)
         CALL add_basis_set_to_container(qs_kind_set(ikind)%basis_sets, tmp_basis, "RI_AUX")
!d    CALL copy_gto_basis_set(tmp_basis,qs_kind_set(ikind)%ri_aux_basis_set)

         DEALLOCATE (RI_exponents)
         DEALLOCATE (RI_l_expo)

         IF (.NOT. ALLOCATED(max_rel_dev_output)) THEN
            ALLOCATE (max_rel_dev_output(RI_nset))
            max_rel_dev_output(:) = max_rel_dev
         ELSE
            ! make a copy
            RI_prev_size = SIZE(max_rel_dev_output)
            ALLOCATE (max_rel_dev_prev(RI_prev_size))
            max_rel_dev_prev(:) = max_rel_dev_output
            DEALLOCATE (max_rel_dev_output)
            ! reallocate and copy
            ALLOCATE (max_rel_dev_output(RI_prev_size+RI_nset))
            max_rel_dev_output(1:RI_prev_size) = max_rel_dev_prev
            max_rel_dev_output(RI_prev_size+1:RI_prev_size+RI_nset) = max_rel_dev
            DEALLOCATE (max_rel_dev_prev)
         END IF
         DEALLOCATE (max_rel_dev)

      END DO ! ikind

      IF (external_num_of_func) DEALLOCATE (mp2_env%ri_opt_param%RI_nset_per_l)

      CALL timestop(handle)

   END SUBROUTINE generate_RI_init_basis

! **************************************************************************************************
!> \brief ...
!> \param gto_basis_set ...
!> \param RI_nset ...
!> \param RI_l_expo ...
!> \param RI_exponents ...
! **************************************************************************************************
   SUBROUTINE create_ri_basis(gto_basis_set, RI_nset, RI_l_expo, RI_exponents)
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis_set
      INTEGER                                            :: RI_nset
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: RI_l_expo
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: RI_exponents

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_ri_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ico, ipgf, iset, ishell, &
                                                            lshell, m, maxco, maxl, maxpgf, &
                                                            maxshell, ncgf, nmin, nset, nsgf
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: l, n
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc

      CALL timeset(routineN, handle)
      NULLIFY (lmax, lmin, npgf, nshell, l, n, zet, gcc)

      ! allocate the basis
      CALL allocate_gto_basis_set(gto_basis_set)

      ! brute force
      nset = 0
      maxshell = 0
      maxpgf = 0
      maxco = 0
      ncgf = 0
      nsgf = 0
      gto_basis_set%nset = nset
      gto_basis_set%ncgf = ncgf
      gto_basis_set%nsgf = nsgf
      CALL reallocate(gto_basis_set%lmax, 1, nset)
      CALL reallocate(gto_basis_set%lmin, 1, nset)
      CALL reallocate(gto_basis_set%npgf, 1, nset)
      CALL reallocate(gto_basis_set%nshell, 1, nset)
      CALL reallocate(gto_basis_set%n, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%l, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%zet, 1, maxpgf, 1, nset)
      CALL reallocate(gto_basis_set%gcc, 1, maxpgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%set_radius, 1, nset)
      CALL reallocate(gto_basis_set%pgf_radius, 1, maxpgf, 1, nset)
      CALL reallocate(gto_basis_set%first_cgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%first_sgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%last_cgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%last_sgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%ncgf_set, 1, nset)
      CALL reallocate(gto_basis_set%nsgf_set, 1, nset)
      CALL reallocate(gto_basis_set%cphi, 1, maxco, 1, ncgf)
      CALL reallocate(gto_basis_set%sphi, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%scon, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%lx, 1, ncgf)
      CALL reallocate(gto_basis_set%ly, 1, ncgf)
      CALL reallocate(gto_basis_set%lz, 1, ncgf)
      CALL reallocate(gto_basis_set%m, 1, nsgf)
      CALL reallocate(gto_basis_set%norm_cgf, 1, ncgf)

      nset = RI_nset

      ! locals
      CALL reallocate(npgf, 1, nset)
      CALL reallocate(nshell, 1, nset)
      CALL reallocate(lmax, 1, nset)
      CALL reallocate(lmin, 1, nset)
      CALL reallocate(n, 1, 1, 1, nset)

      maxl = 0
      maxpgf = 0
      maxshell = 0
      DO iset = 1, nset
         n(1, iset) = iset
         lmin(iset) = RI_l_expo(iset)
         lmax(iset) = RI_l_expo(iset)
         npgf(iset) = 1

         maxl = MAX(maxl, lmax(iset))

         IF (npgf(iset) > maxpgf) THEN
            maxpgf = npgf(iset)
            CALL reallocate(zet, 1, maxpgf, 1, nset)
            CALL reallocate(gcc, 1, maxpgf, 1, maxshell, 1, nset)
         END IF
         nshell(iset) = 0
         DO lshell = lmin(iset), lmax(iset)
            nmin = n(1, iset)+lshell-lmin(iset)
            ishell = 1
            nshell(iset) = nshell(iset)+ishell
            IF (nshell(iset) > maxshell) THEN
               maxshell = nshell(iset)
               CALL reallocate(n, 1, maxshell, 1, nset)
               CALL reallocate(l, 1, maxshell, 1, nset)
               CALL reallocate(gcc, 1, maxpgf, 1, maxshell, 1, nset)
            END IF
            DO i = 1, ishell
               n(nshell(iset)-ishell+i, iset) = nmin+i-1
               l(nshell(iset)-ishell+i, iset) = lshell
            END DO
         END DO

         DO ipgf = 1, npgf(iset)
            zet(ipgf, iset) = RI_exponents(iset)
            DO ishell = 1, nshell(iset)
               gcc(ipgf, ishell, iset) = 1.0_dp
            END DO
         END DO
      END DO

      CALL init_orbital_pointers(maxl)

      gto_basis_set%nset = nset
      CALL reallocate(gto_basis_set%lmax, 1, nset)
      CALL reallocate(gto_basis_set%lmin, 1, nset)
      CALL reallocate(gto_basis_set%npgf, 1, nset)
      CALL reallocate(gto_basis_set%nshell, 1, nset)
      CALL reallocate(gto_basis_set%n, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%l, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%zet, 1, maxpgf, 1, nset)
      CALL reallocate(gto_basis_set%gcc, 1, maxpgf, 1, maxshell, 1, nset)

      ! Copy the basis set information into the data structure

      DO iset = 1, nset
         gto_basis_set%lmax(iset) = lmax(iset)
         gto_basis_set%lmin(iset) = lmin(iset)
         gto_basis_set%npgf(iset) = npgf(iset)
         gto_basis_set%nshell(iset) = nshell(iset)
         DO ishell = 1, nshell(iset)
            gto_basis_set%n(ishell, iset) = n(ishell, iset)
            gto_basis_set%l(ishell, iset) = l(ishell, iset)
            DO ipgf = 1, npgf(iset)
               gto_basis_set%gcc(ipgf, ishell, iset) = gcc(ipgf, ishell, iset)
            END DO
         END DO
         DO ipgf = 1, npgf(iset)
            gto_basis_set%zet(ipgf, iset) = zet(ipgf, iset)
         END DO
      END DO

      ! Initialise the depending atomic kind information

      CALL reallocate(gto_basis_set%set_radius, 1, nset)
      CALL reallocate(gto_basis_set%pgf_radius, 1, maxpgf, 1, nset)
      CALL reallocate(gto_basis_set%first_cgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%first_sgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%last_cgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%last_sgf, 1, maxshell, 1, nset)
      CALL reallocate(gto_basis_set%ncgf_set, 1, nset)
      CALL reallocate(gto_basis_set%nsgf_set, 1, nset)

      maxco = 0
      ncgf = 0
      nsgf = 0

      DO iset = 1, nset
         gto_basis_set%ncgf_set(iset) = 0
         gto_basis_set%nsgf_set(iset) = 0
         DO ishell = 1, nshell(iset)
            lshell = gto_basis_set%l(ishell, iset)
            gto_basis_set%first_cgf(ishell, iset) = ncgf+1
            ncgf = ncgf+nco(lshell)
            gto_basis_set%last_cgf(ishell, iset) = ncgf
            gto_basis_set%ncgf_set(iset) = &
               gto_basis_set%ncgf_set(iset)+nco(lshell)
            gto_basis_set%first_sgf(ishell, iset) = nsgf+1
            nsgf = nsgf+nso(lshell)
            gto_basis_set%last_sgf(ishell, iset) = nsgf
            gto_basis_set%nsgf_set(iset) = &
               gto_basis_set%nsgf_set(iset)+nso(lshell)
         END DO
         maxco = MAX(maxco, npgf(iset)*ncoset(lmax(iset)))
      END DO

      gto_basis_set%ncgf = ncgf
      gto_basis_set%nsgf = nsgf

      CALL reallocate(gto_basis_set%cphi, 1, maxco, 1, ncgf)
      CALL reallocate(gto_basis_set%sphi, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%scon, 1, maxco, 1, nsgf)
      CALL reallocate(gto_basis_set%lx, 1, ncgf)
      CALL reallocate(gto_basis_set%ly, 1, ncgf)
      CALL reallocate(gto_basis_set%lz, 1, ncgf)
      CALL reallocate(gto_basis_set%m, 1, nsgf)
      CALL reallocate(gto_basis_set%norm_cgf, 1, ncgf)
      ALLOCATE (gto_basis_set%cgf_symbol(ncgf))

      ALLOCATE (gto_basis_set%sgf_symbol(nsgf))

      ncgf = 0
      nsgf = 0

      DO iset = 1, nset
         DO ishell = 1, nshell(iset)
            lshell = gto_basis_set%l(ishell, iset)
            DO ico = ncoset(lshell-1)+1, ncoset(lshell)
               ncgf = ncgf+1
               gto_basis_set%lx(ncgf) = indco(1, ico)
               gto_basis_set%ly(ncgf) = indco(2, ico)
               gto_basis_set%lz(ncgf) = indco(3, ico)
               gto_basis_set%cgf_symbol(ncgf) = &
                  cgf_symbol(n(ishell, iset), (/gto_basis_set%lx(ncgf), &
                                                gto_basis_set%ly(ncgf), &
                                                gto_basis_set%lz(ncgf)/))
            END DO
            DO m = -lshell, lshell
               nsgf = nsgf+1
               gto_basis_set%m(nsgf) = m
               gto_basis_set%sgf_symbol(nsgf) = &
                  sgf_symbol(n(ishell, iset), lshell, m)
            END DO
         END DO
      END DO

      DEALLOCATE (gcc, l, lmax, lmin, n, npgf, nshell, zet)

      CALL timestop(handle)

   END SUBROUTINE

END MODULE mp2_optimize_ri_basis

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines for calculating RI-MP2 gradients
!> \par History
!>      10.2013 created [Mauro Del Ben]
! **************************************************************************************************
MODULE mp2_ri_grad_util
!
   USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                              cp_blacs_env_release,&
                                              cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_iterator,&
                                              cp_dbcsr_iterator_blocks_left,&
                                              cp_dbcsr_iterator_next_block,&
                                              cp_dbcsr_iterator_start,&
                                              cp_dbcsr_iterator_stop,&
                                              dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: copy_fm_to_dbcsr,&
                                              cp_dbcsr_m_by_n_from_template
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_indxg2l,&
                                              cp_fm_indxg2p,&
                                              cp_fm_indxl2g,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_irecv,&
                                              mp_isend,&
                                              mp_sendrecv,&
                                              mp_sum,&
                                              mp_wait,&
                                              mp_waitall
   USE mp2_types,                       ONLY: integ_mat_buffer_type,&
                                              mp2_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_ri_grad_util'

   PUBLIC :: complete_gamma

   TYPE index_map
      INTEGER, DIMENSION(:, :), ALLOCATABLE :: map
   END TYPE

CONTAINS

! **************************************************************************************************
!> \brief complete the calculation of the Gamma matrices
!> \param mp2_env ...
!> \param B_ia_Q ...
!> \param dimen_RI ...
!> \param homo ...
!> \param virtual ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param ngroup ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
!> \param my_group_L_end ...
!> \param my_B_size ...
!> \param my_B_virtual_start ...
!> \param ends_array ...
!> \param ends_B_virtual ...
!> \param sizes_array ...
!> \param sizes_B_virtual ...
!> \param starts_array ...
!> \param starts_B_virtual ...
!> \param sub_proc_map ...
!> \param alpha_case ...
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE complete_gamma(mp2_env, B_ia_Q, dimen_RI, homo, virtual, para_env, para_env_sub, ngroup, &
                             my_group_L_size, my_group_L_start, my_group_L_end, &
                             my_B_size, my_B_virtual_start, &
                             ends_array, ends_B_virtual, sizes_array, sizes_B_virtual, starts_array, starts_B_virtual, &
                             sub_proc_map, alpha_case)

      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: B_ia_Q
      INTEGER                                            :: dimen_RI, homo, virtual
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER                                            :: ngroup, my_group_L_size, &
                                                            my_group_L_start, my_group_L_end, &
                                                            my_B_size, my_B_virtual_start
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_array, ends_B_virtual, sizes_array, &
                                                            sizes_B_virtual, starts_array, &
                                                            starts_B_virtual, sub_proc_map
      LOGICAL                                            :: alpha_case

      CHARACTER(LEN=*), PARAMETER :: routineN = 'complete_gamma', routineP = moduleN//':'//routineN

      INTEGER :: blk, col, col_offset, col_size, dimen_ia, handle, i, i_global, ia_global, iiB, &
         iproc, itmp(2), j_global, jjB, kkB, my_ia_end, my_ia_size, my_ia_start, my_P_end, &
         my_P_size, my_P_start, ncol_local, nrow_local, pos_group, pos_sub, proc_receive, &
         proc_send, proc_shift, rec_B_size, rec_B_virtual_end, rec_B_virtual_start, row, &
         row_offset, row_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_ia, ends_P, pos_info, proc_map, &
                                                            sizes_ia, sizes_P, starts_ia, starts_P
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ends, group_grid_2_mepos, &
                                                            mepos_2_grid_group, sizes, starts
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIb_C_2D, Gamma_2D
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C_rec
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_block
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_Gamma, fm_Gamma_PQ, fm_Gamma_PQ_temp, &
                                                            fm_ia_P, fm_Y, PQ_half

      CALL timeset(routineN, handle)

      ! reshape the data into a 2D array
      ! reorder the local data in such a way to help the next stage of matrix creation
      ! now the data inside the group are divided into a ia x K matrix
      dimen_ia = homo*virtual
      ALLOCATE (sizes_ia(0:para_env_sub%num_pe-1))
      sizes_ia = 0
      ALLOCATE (starts_ia(0:para_env_sub%num_pe-1))
      starts_ia = 0
      ALLOCATE (ends_ia(0:para_env_sub%num_pe-1))
      ends_ia = 0

      DO iproc = 0, para_env_sub%num_pe-1
         itmp = get_limit(dimen_ia, para_env_sub%num_pe, iproc)
         starts_ia(iproc) = itmp(1)
         ends_ia(iproc) = itmp(2)
         sizes_ia(iproc) = itmp(2)-itmp(1)+1
      END DO

      my_ia_start = starts_ia(para_env_sub%mepos)
      my_ia_end = ends_ia(para_env_sub%mepos)
      my_ia_size = sizes_ia(para_env_sub%mepos)

      ! reorder data (first (ia|K))
      ALLOCATE (BIb_C_2D(my_ia_size, my_group_L_size))
      BIb_C_2D = 0.0_dp

      DO iiB = 1, homo
         DO jjB = 1, my_B_size
            ia_global = (iiB-1)*virtual+my_B_virtual_start+jjB-1
            IF (ia_global >= my_ia_start .AND. ia_global <= my_ia_end) THEN
               BIb_C_2D(ia_global-my_ia_start+1, 1:my_group_L_size) = B_ia_Q(iiB, jjB, 1:my_group_L_size)
            END IF
         END DO
      END DO

      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

         rec_B_size = sizes_B_virtual(proc_receive)
         rec_B_virtual_end = ends_B_virtual(proc_receive)
         rec_B_virtual_start = starts_B_virtual(proc_receive)

         ALLOCATE (BIb_C_rec(homo, rec_B_size, my_group_L_size))
         BIb_C_rec = 0.0_dp

         CALL mp_sendrecv(B_ia_Q, proc_send, &
                          BIb_C_rec, proc_receive, &
                          para_env_sub%group)

         DO iiB = 1, homo
            DO jjB = 1, rec_B_size
               ia_global = (iiB-1)*virtual+rec_B_virtual_start+jjB-1
               IF (ia_global >= my_ia_start .AND. ia_global <= my_ia_end) THEN
                  BIb_C_2D(ia_global-my_ia_start+1, 1:my_group_L_size) = BIb_C_rec(iiB, jjB, 1:my_group_L_size)
               END IF
            END DO
         END DO

         DEALLOCATE (BIb_C_rec)
      END DO
      DEALLOCATE (B_ia_Q)

      ! reorder data (Gamma)
      ALLOCATE (Gamma_2D(my_ia_size, my_group_L_size))
      Gamma_2D = 0.0_dp

      DO iiB = 1, homo
         DO jjB = 1, my_B_size
            ia_global = (iiB-1)*virtual+my_B_virtual_start+jjB-1
            IF (ia_global >= my_ia_start .AND. ia_global <= my_ia_end) THEN
               ! Closed-shell, open-shell alpha-alpha component
               IF (alpha_case) THEN
                  Gamma_2D(ia_global-my_ia_start+1, 1:my_group_L_size) = &
                     mp2_env%ri_grad%Gamma_P_ia(iiB, jjB, 1:my_group_L_size)
               ELSE ! Open-shell beta-beta component
                  Gamma_2D(ia_global-my_ia_start+1, 1:my_group_L_size) = &
                     mp2_env%ri_grad%Gamma_P_ia_beta(iiB, jjB, 1:my_group_L_size)
               ENDIF
            END IF
         END DO
      END DO

      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

         rec_B_size = sizes_B_virtual(proc_receive)
         rec_B_virtual_end = ends_B_virtual(proc_receive)
         rec_B_virtual_start = starts_B_virtual(proc_receive)

         ALLOCATE (BIb_C_rec(homo, rec_B_size, my_group_L_size))
         BIb_C_rec = 0.0_dp

         IF (alpha_case) THEN
            CALL mp_sendrecv(mp2_env%ri_grad%Gamma_P_ia, proc_send, &
                             BIb_C_rec, proc_receive, &
                             para_env_sub%group)
         ELSE
            CALL mp_sendrecv(mp2_env%ri_grad%Gamma_P_ia_beta, proc_send, &
                             BIb_C_rec, proc_receive, &
                             para_env_sub%group)
         ENDIF

         DO iiB = 1, homo
            DO jjB = 1, rec_B_size
               ia_global = (iiB-1)*virtual+rec_B_virtual_start+jjB-1
               IF (ia_global >= my_ia_start .AND. ia_global <= my_ia_end) THEN
                  Gamma_2D(ia_global-my_ia_start+1, 1:my_group_L_size) = BIb_C_rec(iiB, jjB, 1:my_group_L_size)
               END IF
            END DO
         END DO

         DEALLOCATE (BIb_C_rec)
      END DO
      IF (alpha_case) THEN
         DEALLOCATE (mp2_env%ri_grad%Gamma_P_ia)
      ELSE
         DEALLOCATE (mp2_env%ri_grad%Gamma_P_ia_beta)
      ENDIF

      ! create the processor map and size arrays
      ALLOCATE (proc_map(-para_env%num_pe:2*para_env%num_pe-1))
      proc_map = 0
      ALLOCATE (sizes(2, 0:para_env%num_pe-1))
      sizes = 0
      ALLOCATE (starts(2, 0:para_env%num_pe-1))
      starts = 0
      ALLOCATE (ends(2, 0:para_env%num_pe-1))
      ends = 0

      ALLOCATE (pos_info(0:para_env%num_pe-1))
      pos_info = 0
      pos_info(para_env%mepos) = para_env_sub%mepos
      CALL mp_sum(pos_info, para_env%group)

      ALLOCATE (group_grid_2_mepos(0:para_env_sub%num_pe-1, 0:ngroup-1))
      group_grid_2_mepos = 0
      ALLOCATE (mepos_2_grid_group(0:para_env%num_pe-1, 2))
      mepos_2_grid_group = 0

      DO i = 0, para_env%num_pe-1
         proc_map(i) = i
         proc_map(-i-1) = para_env%num_pe-i-1
         proc_map(para_env%num_pe+i) = i
         ! calculate postition of the group
         pos_group = i/para_env_sub%num_pe
         ! calculate postition in the subgroup
         pos_sub = pos_info(i)
         ! 1 -> rows, 2 -> cols
         ! fill sizes array
         sizes(1, i) = sizes_ia(pos_sub)
         sizes(2, i) = sizes_array(pos_group)
         ! fill starts array
         starts(1, i) = starts_ia(pos_sub)
         starts(2, i) = starts_array(pos_group)
         ! fill ends array
         ends(1, i) = ends_ia(pos_sub)
         ends(2, i) = ends_array(pos_group)
         ! fill the map from the grid of groups to process
         group_grid_2_mepos(pos_sub, pos_group) = i
         ! and the opposite, from the global pos to the grid pos
         mepos_2_grid_group(i, 1) = pos_sub
         mepos_2_grid_group(i, 2) = pos_group
      END DO

      ! WRITE(*,*) mepos_2_grid_group(:,1)
      ! WRITE(*,*) mepos_2_grid_group(:,2)

      ! create the blacs env
      NULLIFY (blacs_env)
      CALL cp_blacs_env_create(blacs_env=blacs_env, para_env=para_env)

      ! create the fm matrix Gamma
      CALL array2fm(Gamma_2D, dimen_ia, dimen_RI, para_env, blacs_env, proc_map, &
                    my_ia_start, my_ia_end, &
                    my_group_L_start, my_group_L_end, &
                    sizes, starts, ends, &
                    group_grid_2_mepos, para_env_sub%num_pe, ngroup, &
                    fm_Y)
      ! create the fm matrix B_ia_P
      CALL array2fm(BIb_C_2D, dimen_ia, dimen_RI, para_env, blacs_env, proc_map, &
                    my_ia_start, my_ia_end, &
                    my_group_L_start, my_group_L_end, &
                    sizes, starts, ends, &
                    group_grid_2_mepos, para_env_sub%num_pe, ngroup, &
                    fm_ia_P)

      ! create PQ_half
      ! get the ranges and info for the (P|Q)^(-1/2) matrix
      ALLOCATE (sizes_P(0:para_env_sub%num_pe-1))
      sizes_P = 0
      ALLOCATE (starts_P(0:para_env_sub%num_pe-1))
      starts_P = 0
      ALLOCATE (ends_P(0:para_env_sub%num_pe-1))
      ends_P = 0
      DO iproc = 0, para_env_sub%num_pe-1
         itmp = get_limit(dimen_RI, para_env_sub%num_pe, iproc)
         starts_P(iproc) = itmp(1)
         ends_P(iproc) = itmp(2)
         sizes_P(iproc) = itmp(2)-itmp(1)+1
      END DO
      my_P_start = starts_P(para_env_sub%mepos)
      my_P_end = ends_P(para_env_sub%mepos)
      my_P_size = sizes_P(para_env_sub%mepos)
      sizes = 0
      starts = 0
      ends = 0
      DO i = 0, para_env%num_pe-1
         ! calculate postition of the group
         pos_group = i/para_env_sub%num_pe
         ! calculate postition in the subgroup
         pos_sub = pos_info(i)
         ! 1 -> rows, 2 -> cols
         ! fill sizes array
         sizes(1, i) = sizes_P(pos_sub)
         sizes(2, i) = sizes_array(pos_group)
         ! fill starts array
         starts(1, i) = starts_P(pos_sub)
         starts(2, i) = starts_array(pos_group)
         ! fill ends array
         ends(1, i) = ends_P(pos_sub)
         ends(2, i) = ends_array(pos_group)
      END DO
      ! since we will need (P|Q)^(-1/2) in the future, make a copy
      IF (alpha_case) THEN
         ALLOCATE (mp2_env%ri_grad%Gamma_PQ(my_P_size, my_group_L_size))
         mp2_env%ri_grad%Gamma_PQ(:, :) = mp2_env%ri_grad%PQ_half
         CALL array2fm(mp2_env%ri_grad%Gamma_PQ, dimen_RI, dimen_RI, para_env, blacs_env, proc_map, &
                       my_P_start, my_P_end, &
                       my_group_L_start, my_group_L_end, &
                       sizes, starts, ends, &
                       group_grid_2_mepos, para_env_sub%num_pe, ngroup, &
                       PQ_half)
      ELSE
         ALLOCATE (mp2_env%ri_grad%Gamma_PQ_beta(my_P_size, my_group_L_size))
         mp2_env%ri_grad%Gamma_PQ_beta(:, :) = mp2_env%ri_grad%PQ_half
         CALL array2fm(mp2_env%ri_grad%Gamma_PQ_beta, dimen_RI, dimen_RI, para_env, blacs_env, proc_map, &
                       my_P_start, my_P_end, &
                       my_group_L_start, my_group_L_end, &
                       sizes, starts, ends, &
                       group_grid_2_mepos, para_env_sub%num_pe, ngroup, &
                       PQ_half)
      ENDIF

      ! complete the gamma matrix Gamma_ia^P = sum_Q (Y_ia^Q * (Q|P)^(-1/2))
      NULLIFY (fm_Gamma)
      NULLIFY (fm_struct)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=dimen_ia, &
                               ncol_global=dimen_RI, para_env=para_env)
      ! for now we don't force the blocks
      ! CALL cp_fm_struct_create(fm_struct,context=blacs_env,nrow_global=num_rows,&
      !                          ncol_global=num_cols,para_env=para_env,&
      !                          nrow_block=nrow_block_mat, ncol_block=ncol_block_mat, force_block=.TRUE.)
      CALL cp_fm_create(fm_Gamma, fm_struct, name="fm_Gamma")
      CALL cp_fm_set_all(matrix=fm_Gamma, alpha=0.0_dp)
      CALL cp_fm_struct_release(fm_struct)
      ! perform the matrix multiplication
      CALL cp_gemm(transa="N", transb="T", m=dimen_ia, n=dimen_RI, k=dimen_RI, alpha=1.0_dp, &
                   matrix_a=fm_Y, matrix_b=PQ_half, beta=0.0_dp, &
                   matrix_c=fm_Gamma)
      ! release the Y matrix
      CALL cp_fm_release(fm_Y)

      ! complete gamma small (fm_Gamma_PQ)
      NULLIFY (fm_Gamma_PQ_temp)
      NULLIFY (fm_struct)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=dimen_RI, &
                               ncol_global=dimen_RI, para_env=para_env)
      ! create temp matrix
      CALL cp_fm_create(fm_Gamma_PQ_temp, fm_struct, name="fm_Gamma_PQ_temp")
      CALL cp_fm_set_all(matrix=fm_Gamma_PQ_temp, alpha=0.0_dp)
      CALL cp_gemm(transa="T", transb="N", m=dimen_RI, n=dimen_RI, k=dimen_ia, alpha=1.0_dp, &
                   matrix_a=fm_Gamma, matrix_b=fm_ia_P, beta=0.0_dp, &
                   matrix_c=fm_Gamma_PQ_temp)
      CALL cp_fm_release(fm_ia_P)
      ! create fm_Gamma_PQ matrix
      CALL cp_fm_create(fm_Gamma_PQ, fm_struct, name="fm_Gamma_PQ")
      ! release the structure
      CALL cp_fm_struct_release(fm_struct)
      CALL cp_fm_set_all(matrix=fm_Gamma_PQ, alpha=0.0_dp)
      ! perfome matrix multiplication
      CALL cp_gemm(transa="N", transb="T", m=dimen_RI, n=dimen_RI, k=dimen_RI, alpha=1.0_dp, &
                   matrix_a=fm_Gamma_PQ_temp, matrix_b=PQ_half, beta=0.0_dp, &
                   matrix_c=fm_Gamma_PQ)
      CALL cp_fm_release(fm_Gamma_PQ_temp)
      CALL cp_fm_release(PQ_half)

      IF (.FALSE.) THEN
         ALLOCATE (Gamma_2D(dimen_ia, dimen_RI))
         Gamma_2D = 0.0_dp
         CALL cp_fm_get_info(matrix=fm_Gamma, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               Gamma_2D(i_global, j_global) = fm_Gamma%local_data(iiB, jjB)
            END DO
         END DO
         CALL mp_sum(Gamma_2D, para_env%group)
         IF (para_env%mepos == 0) CALL write_array(Gamma_2D(1:dimen_ia, 1:dimen_RI))
         DEALLOCATE (Gamma_2D)

         ALLOCATE (Gamma_2D(dimen_RI, dimen_RI))
         Gamma_2D = 0.0_dp
         CALL cp_fm_get_info(matrix=fm_Gamma_PQ, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               Gamma_2D(i_global, j_global) = fm_Gamma_PQ%local_data(iiB, jjB)
            END DO
         END DO
         CALL mp_sum(Gamma_2D, para_env%group)
         IF (para_env%mepos == 0) CALL write_array(Gamma_2D(1:dimen_RI, 1:dimen_RI))
         DEALLOCATE (Gamma_2D)
      END IF

      ! restore original sizes information (dimen_ia x dimen_RI)
      ! DO i=0,para_env%num_pe-1
      !   ! calculate postition of the group
      !   pos_group=i/para_env_sub%num_pe
      !   ! calculate postition in the subgroup
      !   pos_sub=pos_info(i)
      !   ! fill the map from the grid of groups to process
      !   group_grid_2_mepos(pos_sub,pos_group)=i
      ! END DO
      CALL cp_fm_get_info(matrix=fm_Gamma, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local)
      sizes = 0
      sizes(1, para_env%mepos) = nrow_local
      sizes(2, para_env%mepos) = ncol_local
      CALL mp_sum(sizes, para_env%group)

      ! now redistribute the data, in this case we go back
      ! to the original way the integrals were distributed
      CALL fm2array(Gamma_2D, dimen_ia, dimen_RI, para_env, proc_map, &
                    my_ia_size, my_ia_start, my_ia_end, &
                    my_group_L_size, my_group_L_start, my_group_L_end, &
                    sizes, group_grid_2_mepos, mepos_2_grid_group, &
                    para_env_sub%num_pe, ngroup, &
                    fm_Gamma)

      ! the same for the Gamma_PQ matrix
      CALL cp_fm_get_info(matrix=fm_Gamma_PQ, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local)
      sizes = 0
      sizes(1, para_env%mepos) = nrow_local
      sizes(2, para_env%mepos) = ncol_local
      CALL mp_sum(sizes, para_env%group)
      IF (alpha_case) THEN
         CALL fm2array(mp2_env%ri_grad%Gamma_PQ, dimen_RI, dimen_RI, para_env, proc_map, &
                       my_P_size, my_P_start, my_P_end, &
                       my_group_L_size, my_group_L_start, my_group_L_end, &
                       sizes, group_grid_2_mepos, mepos_2_grid_group, &
                       para_env_sub%num_pe, ngroup, &
                       fm_Gamma_PQ)
      ELSE
         CALL fm2array(mp2_env%ri_grad%Gamma_PQ_beta, dimen_RI, dimen_RI, para_env, proc_map, &
                       my_P_size, my_P_start, my_P_end, &
                       my_group_L_size, my_group_L_start, my_group_L_end, &
                       sizes, group_grid_2_mepos, mepos_2_grid_group, &
                       para_env_sub%num_pe, ngroup, &
                       fm_Gamma_PQ)
      ENDIF

      IF (.FALSE.) THEN
         ALLOCATE (BIb_C_2D(dimen_ia, dimen_RI))
         BIb_C_2D = 0.0_dp
         BIb_C_2D(my_ia_start:my_ia_end, my_group_L_start:my_group_L_end) = Gamma_2D
         CALL mp_sum(BIb_C_2D, para_env%group)
         IF (para_env%mepos == 0) CALL write_array(BIb_C_2D(1:dimen_ia, 1:dimen_RI))
         DEALLOCATE (BIb_C_2D)

         ALLOCATE (BIb_C_2D(dimen_RI, dimen_RI))
         BIb_C_2D = 0.0_dp
         IF (alpha_case) THEN
            BIb_C_2D(my_P_start:my_P_end, my_group_L_start:my_group_L_end) = &
               mp2_env%ri_grad%Gamma_PQ
         ELSE
            BIb_C_2D(my_P_start:my_P_end, my_group_L_start:my_group_L_end) = &
               mp2_env%ri_grad%Gamma_PQ_beta
         ENDIF
         CALL mp_sum(BIb_C_2D, para_env%group)
         IF (para_env%mepos == 0) THEN
            WRITE (*, *) 'AAAAAAAAAAAAAAAAAA'
            WRITE (*, *)
            CALL write_array(BIb_C_2D(1:dimen_RI, 1:dimen_RI))
         END IF
         DEALLOCATE (BIb_C_2D)
      END IF

      ! create the Gamma_ia_P in DBCSR style
      CALL create_dbcsr_gamma(Gamma_2D, homo, virtual, dimen_ia, para_env_sub, sub_proc_map, &
                              my_ia_start, my_ia_end, my_group_L_size, &
                              sizes_ia, starts_ia, ends_ia, &
                              mp2_env, alpha_case)

      IF (.FALSE.) THEN
         ALLOCATE (Gamma_2D(dimen_ia, dimen_RI))
         ALLOCATE (BIb_C_rec(virtual, homo, dimen_RI))
         Gamma_2D = 0.0_dp
         BIb_C_rec = 0.0_dp
         DO kkB = 1, my_group_L_size
            IF (alpha_case) THEN
               CALL cp_dbcsr_iterator_start(iter, mp2_env%ri_grad%G_P_ia(kkB)%matrix)
            ELSE
               CALL cp_dbcsr_iterator_start(iter, mp2_env%ri_grad%G_P_ia_beta(kkB)%matrix)
            ENDIF
            DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
               CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                                 row_size=row_size, col_size=col_size, &
                                                 row_offset=row_offset, col_offset=col_offset)
               DO iiB = 1, row_size
                  i_global = row_offset+iiB-1
                  DO jjB = 1, col_size
                     j_global = col_offset+jjB-1
                     ia_global = (i_global-1)*virtual+j_global
                     Gamma_2D(ia_global, my_group_L_start+kkB-1) = data_block(iiB, jjB)
                     BIb_C_rec(j_global, i_global, my_group_L_start+kkB-1) = data_block(iiB, jjB)
                  ENDDO
               ENDDO
            ENDDO
            CALL cp_dbcsr_iterator_stop(iter)
         END DO
         CALL mp_sum(Gamma_2D, para_env%group)
         CALL mp_sum(BIb_C_rec, para_env%group)
         ! IF(para_env%mepos==0) CALL write_array(Gamma_2D(1:dimen_ia,1:dimen_RI))
         IF (para_env%mepos == 0) THEN
            DO iiB = 1, dimen_RI
               WRITE (*, *) 'AAAAAAAAAA', iiB
               CALL write_array(BIb_C_rec(1:virtual, 1:homo, iiB))
            END DO
         END IF
         DEALLOCATE (Gamma_2D)
         DEALLOCATE (BIb_C_rec)
      END IF

      DEALLOCATE (pos_info)
      DEALLOCATE (group_grid_2_mepos, mepos_2_grid_group)
      DEALLOCATE (sizes_ia, starts_ia, ends_ia)
      DEALLOCATE (sizes_P, starts_P, ends_P)
      DEALLOCATE (proc_map)
      DEALLOCATE (sizes, starts, ends)

      ! release fm_matrices
      ! CALL cp_fm_release(fm_Gamma)
      ! CALL cp_fm_release(fm_Gamma_PQ)

      ! release blacs_env
      CALL cp_blacs_env_release(blacs_env)

      CALL timestop(handle)

   END SUBROUTINE complete_gamma

! **************************************************************************************************
!> \brief ...
!> \param mat2D ...
!> \param num_rows ...
!> \param num_cols ...
!> \param para_env ...
!> \param blacs_env ...
!> \param proc_map ...
!> \param my_start_row ...
!> \param my_end_row ...
!> \param my_start_col ...
!> \param my_end_col ...
!> \param sizes ...
!> \param starts ...
!> \param ends ...
!> \param group_grid_2_mepos ...
!> \param ngroup_row ...
!> \param ngroup_col ...
!> \param fm_mat ...
! **************************************************************************************************
   SUBROUTINE array2fm(mat2D, num_rows, num_cols, para_env, blacs_env, proc_map, &
                       my_start_row, my_end_row, &
                       my_start_col, my_end_col, &
                       sizes, starts, ends, &
                       group_grid_2_mepos, ngroup_row, ngroup_col, &
                       fm_mat)

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: mat2D
      INTEGER                                            :: num_rows, num_cols
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: proc_map
      INTEGER                                            :: my_start_row, my_end_row, my_start_col, &
                                                            my_end_col
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: sizes, starts, ends, group_grid_2_mepos
      INTEGER                                            :: ngroup_row, ngroup_col
      TYPE(cp_fm_type), POINTER                          :: fm_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'array2fm', routineP = moduleN//':'//routineN

      INTEGER :: dummy_proc, end_col_block, end_row_block, handle, handle2, i_global, i_local, &
         i_sub, iiB, iii, itmp(2), j_global, j_local, j_sub, jjB, my_num_col_blocks, &
         my_num_row_blocks, mypcol, myprow, ncol_block, ncol_local, npcol, nprow, nrow_block, &
         nrow_local, num_rec_cols, number_of_rec, number_of_send, proc_receive, proc_send, &
         proc_shift, rec_col_end, rec_col_size, rec_col_start, rec_counter, rec_pcol, rec_prow, &
         rec_row_end, rec_row_size, rec_row_start, ref_send_pcol, ref_send_prow, send_counter, &
         send_pcol, send_prow, size_rec_buffer, size_send_buffer, start_col_block, start_row_block
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iii_vet, index_col_rec, map_rec_size, &
                                                            map_send_size, req_send
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: blocks_ranges_col, blocks_ranges_row, &
                                                            grid_2_mepos, grid_ref_2_send_pos, &
                                                            mepos_2_grid
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: part_col, part_row
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_send

      CALL timeset(routineN, handle)

      ! create the full matrix, (num_rows x num_cols)
      NULLIFY (fm_mat)
      NULLIFY (fm_struct)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=num_rows, &
                               ncol_global=num_cols, para_env=para_env)
      ! for now we don't force the blocks
      ! CALL cp_fm_struct_create(fm_struct,context=blacs_env,nrow_global=num_rows,&
      !                          ncol_global=num_cols,para_env=para_env,&
      !                          nrow_block=nrow_block_mat, ncol_block=ncol_block_mat, force_block=.TRUE.)
      CALL cp_fm_create(fm_mat, fm_struct, name="fm_mat")
      CALL cp_fm_set_all(matrix=fm_mat, alpha=0.0_dp)
      CALL cp_fm_struct_release(fm_struct)

      ! start filling procedure
      ! fill the matrix
      CALL cp_fm_get_info(matrix=fm_mat, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          nrow_block=nrow_block, &
                          ncol_block=ncol_block)
      myprow = fm_mat%matrix_struct%context%mepos(1)
      mypcol = fm_mat%matrix_struct%context%mepos(2)
      nprow = fm_mat%matrix_struct%context%num_pe(1)
      npcol = fm_mat%matrix_struct%context%num_pe(2)

      ! start the communication part
      ! 0) create array containing the processes position
      !    and supporting infos
      CALL timeset(routineN//"_info", handle2)
      ALLOCATE (grid_2_mepos(0:nprow-1, 0:npcol-1))
      grid_2_mepos = 0
      ALLOCATE (mepos_2_grid(0:para_env%num_pe-1, 2))
      mepos_2_grid = 0
      ! fill the info array
      grid_2_mepos(myprow, mypcol) = para_env%mepos
      mepos_2_grid(para_env%mepos, 1) = myprow
      mepos_2_grid(para_env%mepos, 2) = mypcol
      ! sum infos
      CALL mp_sum(grid_2_mepos, para_env%group)
      CALL mp_sum(mepos_2_grid, para_env%group)

      ! 1) loop over my local data and define a map for the proc to send data
      ALLOCATE (map_send_size(0:para_env%num_pe-1))
      map_send_size = 0
      DO jjB = my_start_col, my_end_col
         send_pcol = cp_fm_indxg2p(jjB, ncol_block, dummy_proc, &
                                   fm_mat%matrix_struct%first_p_pos(2), npcol)
         DO iiB = my_start_row, my_end_row
            send_prow = cp_fm_indxg2p(iiB, nrow_block, dummy_proc, &
                                      fm_mat%matrix_struct%first_p_pos(1), nprow)
            proc_send = grid_2_mepos(send_prow, send_pcol)
            map_send_size(proc_send) = map_send_size(proc_send)+1
         END DO
      END DO

      ! 2) loop over my local data of fm_mat and define a map for the proc from which rec data
      ALLOCATE (map_rec_size(0:para_env%num_pe-1))
      map_rec_size = 0
      part_row = FLOAT(num_rows)/FLOAT(ngroup_row)
      part_col = FLOAT(num_cols)/FLOAT(ngroup_col)

      DO jjB = 1, ncol_local
         j_global = col_indices(jjB)
         ! check the group holding this element
         ! dirty way, if someone has a better idea ...
         rec_pcol = INT(FLOAT(j_global-1)/part_col)
         rec_pcol = MAX(0, rec_pcol)
         rec_pcol = MIN(rec_pcol, ngroup_col-1)
         DO
            itmp = get_limit(num_cols, ngroup_col, rec_pcol)
            IF (j_global >= itmp(1) .AND. j_global <= itmp(2)) EXIT
            IF (j_global < itmp(1)) rec_pcol = rec_pcol-1
            IF (j_global > itmp(2)) rec_pcol = rec_pcol+1
         END DO

         DO iiB = 1, nrow_local
            i_global = row_indices(iiB)
            ! check the process in the group holding this element
            rec_prow = INT(FLOAT(i_global-1)/part_row)
            rec_prow = MAX(0, rec_prow)
            rec_prow = MIN(rec_prow, ngroup_row-1)
            DO
               itmp = get_limit(num_rows, ngroup_row, rec_prow)
               IF (i_global >= itmp(1) .AND. i_global <= itmp(2)) EXIT
               IF (i_global < itmp(1)) rec_prow = rec_prow-1
               IF (i_global > itmp(2)) rec_prow = rec_prow+1
            END DO

            proc_receive = group_grid_2_mepos(rec_prow, rec_pcol)

            map_rec_size(proc_receive) = map_rec_size(proc_receive)+1

         END DO ! i_global
      END DO ! j_global

      ! 3) check if the local data has to be stored in the new fm_mat
      IF (map_rec_size(para_env%mepos) > 0) THEN
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            IF (j_global >= my_start_col .AND. j_global <= my_end_col) THEN
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  IF (i_global >= my_start_row .AND. i_global <= my_end_row) THEN
                     fm_mat%local_data(iiB, jjB) = mat2D(i_global-my_start_row+1, j_global-my_start_col+1)
                  END IF
               END DO
            END IF
         END DO
      END IF
      CALL timestop(handle2)

      ! 4) reorder data in the send_buffer
      CALL timeset(routineN//"_buffer_send", handle2)
      ! count the number of messages to send
      number_of_send = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         IF (map_send_size(proc_send) > 0) THEN
            number_of_send = number_of_send+1
         END IF
      END DO
      ! allocate the structure that will hold the messages to be sent
      ALLOCATE (buffer_send(number_of_send))
      ! grid_ref_2_send_pos is an array (map) that given a pair
      ! (ref_send_prow,ref_send_pcol) returns
      ! the position in the buffer_send associated to that process
      ALLOCATE (grid_ref_2_send_pos(0:nprow-1, 0:npcol-1))
      grid_ref_2_send_pos = 0
      ! finalize the allocation of buffer_send with the actual size
      ! of each message (actual size is size_send_buffer)
      send_counter = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         size_send_buffer = map_send_size(proc_send)
         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            ! allocate the sending buffer (msg)
            ALLOCATE (buffer_send(send_counter)%msg(size_send_buffer))
            buffer_send(send_counter)%msg = 0.0_dp
            buffer_send(send_counter)%proc = proc_send
            ! get the pointer to prow, pcol of the process that has
            ! to receive this message
            ref_send_prow = mepos_2_grid(proc_send, 1)
            ref_send_pcol = mepos_2_grid(proc_send, 2)
            ! save the rank of the process that has to receive this message
            grid_ref_2_send_pos(ref_send_prow, ref_send_pcol) = send_counter
         END IF
      END DO

      ! loop over the locally held data and fill the buffer_send
      ! for doing that we need an array that keep track if the
      ! sequential increase of the index for each message
      ALLOCATE (iii_vet(number_of_send))
      iii_vet = 0
      DO iiB = my_start_row, my_end_row
         send_prow = cp_fm_indxg2p(iiB, nrow_block, dummy_proc, &
                                   fm_mat%matrix_struct%first_p_pos(1), nprow)
         DO jjB = my_start_col, my_end_col
            send_pcol = cp_fm_indxg2p(jjB, ncol_block, dummy_proc, &
                                      fm_mat%matrix_struct%first_p_pos(2), npcol)
            ! we don't need to send to ourselves
            IF (grid_2_mepos(send_prow, send_pcol) == para_env%mepos) CYCLE

            send_counter = grid_ref_2_send_pos(send_prow, send_pcol)
            iii_vet(send_counter) = iii_vet(send_counter)+1
            iii = iii_vet(send_counter)
            buffer_send(send_counter)%msg(iii) = mat2D(iiB-my_start_row+1, jjB-my_start_col+1)
         END DO
      END DO

      DEALLOCATE (iii_vet)
      DEALLOCATE (grid_ref_2_send_pos)
      DEALLOCATE (mat2D)
      CALL timestop(handle2)

      ! 5) similarly to what done for the buffer_send
      !    create the buffer for receive, post the message with irecv
      !    and send the messages with mp_isend
      CALL timeset(routineN//"_isendrecv", handle2)
      ! count the number of messages to be received
      number_of_rec = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_receive = proc_map(para_env%mepos-proc_shift)
         IF (map_rec_size(proc_receive) > 0) THEN
            number_of_rec = number_of_rec+1
         END IF
      END DO

      ALLOCATE (buffer_rec(number_of_rec))

      rec_counter = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_receive = proc_map(para_env%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)
         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1
            ! prepare the buffer for receive
            ALLOCATE (buffer_rec(rec_counter)%msg(size_rec_buffer))
            buffer_rec(rec_counter)%msg = 0.0_dp
            buffer_rec(rec_counter)%proc = proc_receive
            ! post the message to be received
            CALL mp_irecv(buffer_rec(rec_counter)%msg, proc_receive, para_env%group, &
                          buffer_rec(rec_counter)%msg_req)
         END IF
      END DO

      ! send messages
      ALLOCATE (req_send(number_of_send))
      send_counter = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            CALL mp_isend(buffer_send(send_counter)%msg, proc_send, para_env%group, &
                          buffer_send(send_counter)%msg_req)
            req_send(send_counter) = buffer_send(send_counter)%msg_req
         END IF
      END DO
      CALL timestop(handle2)

      ! 6) fill the fm_mat matrix with the messages received from the
      !    other processes
      CALL timeset(routineN//"_fill", handle2)
      ! In order to perform this step efficiently first we have to know the
      ! ranges of the blocks over which a given process hold its local data.
      ! Start with the rows ...
      my_num_row_blocks = 1
      DO iiB = 1, nrow_local-1
         IF (ABS(row_indices(iiB+1)-row_indices(iiB)) == 1) CYCLE
         my_num_row_blocks = my_num_row_blocks+1
      END DO
      ALLOCATE (blocks_ranges_row(2, my_num_row_blocks))
      blocks_ranges_row = 0
      blocks_ranges_row(1, 1) = row_indices(1)
      iii = 1
      DO iiB = 1, nrow_local-1
         IF (ABS(row_indices(iiB+1)-row_indices(iiB)) == 1) CYCLE
         iii = iii+1
         blocks_ranges_row(2, iii-1) = row_indices(iiB)
         blocks_ranges_row(1, iii) = row_indices(iiB+1)
      END DO
      blocks_ranges_row(2, my_num_row_blocks) = row_indices(MAX(nrow_local, 1))
      ! ... and columns
      my_num_col_blocks = 1
      DO jjB = 1, ncol_local-1
         IF (ABS(col_indices(jjB+1)-col_indices(jjB)) == 1) CYCLE
         my_num_col_blocks = my_num_col_blocks+1
      END DO
      ALLOCATE (blocks_ranges_col(2, my_num_col_blocks))
      blocks_ranges_col = 0
      blocks_ranges_col(1, 1) = col_indices(1)
      iii = 1
      DO jjB = 1, ncol_local-1
         IF (ABS(col_indices(jjB+1)-col_indices(jjB)) == 1) CYCLE
         iii = iii+1
         blocks_ranges_col(2, iii-1) = col_indices(jjB)
         blocks_ranges_col(1, iii) = col_indices(jjB+1)
      END DO
      blocks_ranges_col(2, my_num_col_blocks) = col_indices(MAX(ncol_local, 1))

      rec_counter = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_receive = proc_map(para_env%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)

         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1

            rec_col_size = sizes(2, proc_receive)
            rec_col_start = starts(2, proc_receive)
            rec_col_end = ends(2, proc_receive)

            ! precompute the number of received columns and relative index
            num_rec_cols = 0
            DO jjB = 1, my_num_col_blocks
               start_col_block = MAX(blocks_ranges_col(1, jjB), rec_col_start)
               end_col_block = MIN(blocks_ranges_col(2, jjB), rec_col_end)
               DO j_sub = start_col_block, end_col_block
                  num_rec_cols = num_rec_cols+1
               END DO
            END DO
            ALLOCATE (index_col_rec(num_rec_cols))
            index_col_rec = 0
            iii = 0
            DO jjB = 1, my_num_col_blocks
               start_col_block = MAX(blocks_ranges_col(1, jjB), rec_col_start)
               end_col_block = MIN(blocks_ranges_col(2, jjB), rec_col_end)
               DO j_sub = start_col_block, end_col_block
                  iii = iii+1
                  j_local = cp_fm_indxg2l(j_sub, ncol_block, dummy_proc, &
                                          fm_mat%matrix_struct%first_p_pos(2), npcol)
                  index_col_rec(iii) = j_local
               END DO
            END DO

            rec_row_size = sizes(1, proc_receive)
            rec_row_start = starts(1, proc_receive)
            rec_row_end = ends(1, proc_receive)

            ! wait for the message
            CALL mp_wait(buffer_rec(rec_counter)%msg_req)

            ! fill the local data in fm_mat
            iii = 0
            DO iiB = 1, my_num_row_blocks
               start_row_block = MAX(blocks_ranges_row(1, iiB), rec_row_start)
               end_row_block = MIN(blocks_ranges_row(2, iiB), rec_row_end)
               DO i_sub = start_row_block, end_row_block
                  i_local = cp_fm_indxg2l(i_sub, nrow_block, dummy_proc, &
                                          fm_mat%matrix_struct%first_p_pos(1), nprow)
                  DO jjB = 1, num_rec_cols
                     iii = iii+1
                     j_local = index_col_rec(jjB)
                     fm_mat%local_data(i_local, j_local) = buffer_rec(rec_counter)%msg(iii)
                  END DO
               END DO
            END DO

            DEALLOCATE (buffer_rec(rec_counter)%msg)
            DEALLOCATE (index_col_rec)
         END IF
      END DO
      DEALLOCATE (buffer_rec)

      DEALLOCATE (blocks_ranges_row)
      DEALLOCATE (blocks_ranges_col)

      CALL timestop(handle2)

      ! 7) Finally wait for all messeges to be sent
      CALL timeset(routineN//"_waitall", handle2)
      CALL mp_waitall(req_send(:))
      DO send_counter = 1, number_of_send
         DEALLOCATE (buffer_send(send_counter)%msg)
      END DO
      DEALLOCATE (buffer_send)
      CALL timestop(handle2)

      DEALLOCATE (map_send_size)
      DEALLOCATE (map_rec_size)
      DEALLOCATE (grid_2_mepos)
      DEALLOCATE (mepos_2_grid)

      CALL timestop(handle)

   END SUBROUTINE array2fm

! **************************************************************************************************
!> \brief ...
!> \param mat2D ...
!> \param num_rows ...
!> \param num_cols ...
!> \param para_env ...
!> \param proc_map ...
!> \param my_rows ...
!> \param my_start_row ...
!> \param my_end_row ...
!> \param my_cols ...
!> \param my_start_col ...
!> \param my_end_col ...
!> \param sizes ...
!> \param group_grid_2_mepos ...
!> \param mepos_2_grid_group ...
!> \param ngroup_row ...
!> \param ngroup_col ...
!> \param fm_mat ...
! **************************************************************************************************
   SUBROUTINE fm2array(mat2D, num_rows, num_cols, para_env, proc_map, &
                       my_rows, my_start_row, my_end_row, &
                       my_cols, my_start_col, my_end_col, &
                       sizes, group_grid_2_mepos, mepos_2_grid_group, &
                       ngroup_row, ngroup_col, &
                       fm_mat)

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: mat2D
      INTEGER                                            :: num_rows, num_cols
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: proc_map
      INTEGER                                            :: my_rows, my_start_row, my_end_row, &
                                                            my_cols, my_start_col, my_end_col
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: sizes, group_grid_2_mepos, &
                                                            mepos_2_grid_group
      INTEGER                                            :: ngroup_row, ngroup_col
      TYPE(cp_fm_type), POINTER                          :: fm_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fm2array', routineP = moduleN//':'//routineN

      INTEGER :: dummy_proc, handle, handle2, i_global, iiB, iii, itmp(2), j_global, jjB, mypcol, &
         myprow, ncol_block, ncol_local, npcol, nprow, nrow_block, nrow_local, num_rec_rows, &
         number_of_rec, number_of_send, proc_receive, proc_send, proc_shift, rec_col_size, &
         rec_counter, rec_pcol, rec_prow, rec_row_size, ref_send_pcol, ref_send_prow, &
         send_counter, send_pcol, send_prow, size_rec_buffer, size_send_buffer
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iii_vet, index_row_rec, map_rec_size, &
                                                            map_send_size, req_send
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: grid_2_mepos, grid_ref_2_send_pos, &
                                                            mepos_2_grid
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: part_col, part_row
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_send

      CALL timeset(routineN, handle)

      ! allocate the array
      ALLOCATE (mat2D(my_rows, my_cols))
      mat2D = 0.0_dp

      ! start procedure
      ! get information of from the full matrix
      CALL cp_fm_get_info(matrix=fm_mat, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          nrow_block=nrow_block, &
                          ncol_block=ncol_block)
      myprow = fm_mat%matrix_struct%context%mepos(1)
      mypcol = fm_mat%matrix_struct%context%mepos(2)
      nprow = fm_mat%matrix_struct%context%num_pe(1)
      npcol = fm_mat%matrix_struct%context%num_pe(2)

      ! start the communication part
      ! 0) create array containing the processes position
      !    and supporting infos
      CALL timeset(routineN//"_info", handle2)
      ALLOCATE (grid_2_mepos(0:nprow-1, 0:npcol-1))
      grid_2_mepos = 0
      ALLOCATE (mepos_2_grid(0:para_env%num_pe-1, 2))
      mepos_2_grid = 0
      ! ALLOCATE(mepos_2_grid_group(0:para_env%num_pe-1,2),STAT=stat)
      ! CPPostcondition(stat==0,cp_failure_level,routineP,failure)
      ! mepos_2_grid_group=0
      ! fill the info array
      grid_2_mepos(myprow, mypcol) = para_env%mepos
      mepos_2_grid(para_env%mepos, 1) = myprow
      mepos_2_grid(para_env%mepos, 2) = mypcol

      ! sum infos
      CALL mp_sum(grid_2_mepos, para_env%group)
      CALL mp_sum(mepos_2_grid, para_env%group)

      ! 1) loop over my local data and define a map for the proc to send data
      ALLOCATE (map_send_size(0:para_env%num_pe-1))
      map_send_size = 0
      part_row = FLOAT(num_rows)/FLOAT(ngroup_row)
      part_col = FLOAT(num_cols)/FLOAT(ngroup_col)

      DO jjB = 1, ncol_local
         j_global = col_indices(jjB)
         ! check the group holding this element
         ! dirty way, if someone has a better idea ...
         send_pcol = INT(FLOAT(j_global-1)/part_col)
         send_pcol = MAX(0, send_pcol)
         send_pcol = MIN(send_pcol, ngroup_col-1)
         DO
            itmp = get_limit(num_cols, ngroup_col, send_pcol)
            IF (j_global >= itmp(1) .AND. j_global <= itmp(2)) EXIT
            IF (j_global < itmp(1)) send_pcol = send_pcol-1
            IF (j_global > itmp(2)) send_pcol = send_pcol+1
         END DO

         DO iiB = 1, nrow_local
            i_global = row_indices(iiB)
            ! check the process in the group holding this element
            send_prow = INT(FLOAT(i_global-1)/part_row)
            send_prow = MAX(0, send_prow)
            send_prow = MIN(send_prow, ngroup_row-1)
            DO
               itmp = get_limit(num_rows, ngroup_row, send_prow)
               IF (i_global >= itmp(1) .AND. i_global <= itmp(2)) EXIT
               IF (i_global < itmp(1)) send_prow = send_prow-1
               IF (i_global > itmp(2)) send_prow = send_prow+1
            END DO

            proc_send = group_grid_2_mepos(send_prow, send_pcol)

            map_send_size(proc_send) = map_send_size(proc_send)+1

         END DO ! i_global
      END DO ! j_global

      ! WRITE(*,*) para_env%mepos,'AAAA',map_send_size

      ! 2) loop over my local data of the array and define a map for the proc from which rec data
      ALLOCATE (map_rec_size(0:para_env%num_pe-1))
      map_rec_size = 0
      DO jjB = my_start_col, my_end_col
         rec_pcol = cp_fm_indxg2p(jjB, ncol_block, dummy_proc, &
                                  fm_mat%matrix_struct%first_p_pos(2), npcol)
         DO iiB = my_start_row, my_end_row
            rec_prow = cp_fm_indxg2p(iiB, nrow_block, dummy_proc, &
                                     fm_mat%matrix_struct%first_p_pos(1), nprow)
            proc_receive = grid_2_mepos(rec_prow, rec_pcol)
            map_rec_size(proc_receive) = map_rec_size(proc_receive)+1
         END DO
      END DO

      ! 3) check if the local data in the fm_mat has to be stored in the new array
      IF (map_rec_size(para_env%mepos) > 0) THEN
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            IF (j_global >= my_start_col .AND. j_global <= my_end_col) THEN
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  IF (i_global >= my_start_row .AND. i_global <= my_end_row) THEN
                     mat2D(i_global-my_start_row+1, j_global-my_start_col+1) = fm_mat%local_data(iiB, jjB)
                  END IF
               END DO
            END IF
         END DO
      END IF
      CALL timestop(handle2)

      ! 4) reorder data in the send_buffer
      CALL timeset(routineN//"_buffer_send", handle2)
      ! count the number of messages to send
      number_of_send = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         IF (map_send_size(proc_send) > 0) THEN
            number_of_send = number_of_send+1
         END IF
      END DO
      ! allocate the structure that will hold the messages to be sent
      ALLOCATE (buffer_send(number_of_send))
      ! grid_ref_2_send_pos is an array (map) that given a pair
      ! (ref_send_prow,ref_send_pcol) returns
      ! the position in the buffer_send associated to that process

      ! ALLOCATE(grid_ref_2_send_pos(0:nprow-1,0:npcol-1),STAT=stat)
      ALLOCATE (grid_ref_2_send_pos(0:ngroup_row-1, 0:ngroup_col-1))
      grid_ref_2_send_pos = 0

      ! finalize the allocation of buffer_send with the actual size
      ! of each message (actual size is size_send_buffer)
      send_counter = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         size_send_buffer = map_send_size(proc_send)
         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            ! allocate the sending buffer (msg)
            ALLOCATE (buffer_send(send_counter)%msg(size_send_buffer))
            buffer_send(send_counter)%msg = 0.0_dp
            buffer_send(send_counter)%proc = proc_send
            ! get the pointer to prow, pcol of the process that has
            ! to receive this message
            ref_send_prow = mepos_2_grid_group(proc_send, 1)
            ref_send_pcol = mepos_2_grid_group(proc_send, 2)
            ! save the rank of the process that has to receive this message
            grid_ref_2_send_pos(ref_send_prow, ref_send_pcol) = send_counter
         END IF
      END DO

      ! loop over the locally held data and fill the buffer_send
      ! for doing that we need an array that keep track if the
      ! sequential increase of the index for each message
      ALLOCATE (iii_vet(number_of_send))
      iii_vet = 0
      DO jjB = 1, ncol_local
         j_global = col_indices(jjB)
         ! check the group holding this element
         ! dirty way, if someone has a better idea ...
         send_pcol = INT(FLOAT(j_global-1)/part_col)
         send_pcol = MAX(0, send_pcol)
         send_pcol = MIN(send_pcol, ngroup_col-1)
         DO
            itmp = get_limit(num_cols, ngroup_col, send_pcol)
            IF (j_global >= itmp(1) .AND. j_global <= itmp(2)) EXIT
            IF (j_global < itmp(1)) send_pcol = send_pcol-1
            IF (j_global > itmp(2)) send_pcol = send_pcol+1
         END DO

         DO iiB = 1, nrow_local
            i_global = row_indices(iiB)
            ! check the process in the group holding this element
            send_prow = INT(FLOAT(i_global-1)/part_row)
            send_prow = MAX(0, send_prow)
            send_prow = MIN(send_prow, ngroup_row-1)
            DO
               itmp = get_limit(num_rows, ngroup_row, send_prow)
               IF (i_global >= itmp(1) .AND. i_global <= itmp(2)) EXIT
               IF (i_global < itmp(1)) send_prow = send_prow-1
               IF (i_global > itmp(2)) send_prow = send_prow+1
            END DO
            ! we don't need to send to ourselves
            IF (group_grid_2_mepos(send_prow, send_pcol) == para_env%mepos) CYCLE

            send_counter = grid_ref_2_send_pos(send_prow, send_pcol)
            iii_vet(send_counter) = iii_vet(send_counter)+1
            iii = iii_vet(send_counter)
            buffer_send(send_counter)%msg(iii) = fm_mat%local_data(iiB, jjB)
         END DO
      END DO

      DEALLOCATE (iii_vet)
      DEALLOCATE (grid_ref_2_send_pos)
      ! CALL cp_fm_release(fm_mat)
      CALL timestop(handle2)

      ! 5) similarly to what done for the buffer_send
      !    create the buffer for receive, post the message with irecv
      !    and send the messages with mp_isend
      CALL timeset(routineN//"_isendrecv", handle2)
      ! count the number of messages to be received
      number_of_rec = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_receive = proc_map(para_env%mepos-proc_shift)
         IF (map_rec_size(proc_receive) > 0) THEN
            number_of_rec = number_of_rec+1
         END IF
      END DO

      ALLOCATE (buffer_rec(number_of_rec))

      rec_counter = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_receive = proc_map(para_env%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)
         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1
            ! prepare the buffer for receive
            ALLOCATE (buffer_rec(rec_counter)%msg(size_rec_buffer))
            buffer_rec(rec_counter)%msg = 0.0_dp
            buffer_rec(rec_counter)%proc = proc_receive
            ! post the message to be received
            CALL mp_irecv(buffer_rec(rec_counter)%msg, proc_receive, para_env%group, &
                          buffer_rec(rec_counter)%msg_req)
         END IF
      END DO

      ! send messages
      ALLOCATE (req_send(number_of_send))
      send_counter = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            CALL mp_isend(buffer_send(send_counter)%msg, proc_send, para_env%group, &
                          buffer_send(send_counter)%msg_req)
            req_send(send_counter) = buffer_send(send_counter)%msg_req
         END IF
      END DO
      CALL timestop(handle2)

      ! 6) fill the fm_mat matrix with the messages received from the
      !    other processes
      CALL timeset(routineN//"_fill", handle2)
      iiB = MAXVAL(sizes(1, :))
      ALLOCATE (index_row_rec(iiB))
      index_row_rec = 0
      rec_counter = 0
      DO proc_shift = 1, para_env%num_pe-1
         proc_receive = proc_map(para_env%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)

         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1

            rec_col_size = sizes(2, proc_receive)
            rec_row_size = sizes(1, proc_receive)

            ! precompute the indeces of the rows
            num_rec_rows = 0
            DO iiB = 1, rec_row_size
               i_global = cp_fm_indxl2g(iiB, nrow_block, mepos_2_grid(proc_receive, 1), &
                                        fm_mat%matrix_struct%first_p_pos(1), nprow)
               IF (i_global >= my_start_row .AND. i_global <= my_end_row) THEN
                  num_rec_rows = num_rec_rows+1
                  index_row_rec(num_rec_rows) = i_global
               END IF
            END DO

            CALL mp_wait(buffer_rec(rec_counter)%msg_req)

            iii = 0
            DO jjB = 1, rec_col_size
               j_global = cp_fm_indxl2g(jjB, ncol_block, mepos_2_grid(proc_receive, 2), &
                                        fm_mat%matrix_struct%first_p_pos(2), npcol)
               IF (j_global >= my_start_col .AND. j_global <= my_end_col) THEN
                  DO iiB = 1, num_rec_rows
                     i_global = index_row_rec(iiB)
                     iii = iii+1
                     mat2D(i_global-my_start_row+1, j_global-my_start_col+1) = buffer_rec(rec_counter)%msg(iii)
                  END DO
               END IF
            END DO

            DEALLOCATE (buffer_rec(rec_counter)%msg)
         END IF
      END DO
      DEALLOCATE (buffer_rec)
      DEALLOCATE (index_row_rec)
      CALL cp_fm_release(fm_mat)
      CALL timestop(handle2)

      ! 7) Finally wait for all messeges to be sent
      CALL timeset(routineN//"_waitall", handle2)
      CALL mp_waitall(req_send(:))
      DO send_counter = 1, number_of_send
         DEALLOCATE (buffer_send(send_counter)%msg)
      END DO
      DEALLOCATE (buffer_send)
      CALL timestop(handle2)

      CALL timestop(handle)

   END SUBROUTINE fm2array

! **************************************************************************************************
!> \brief ...
!> \param Gamma_2D ...
!> \param homo ...
!> \param virtual ...
!> \param dimen_ia ...
!> \param para_env_sub ...
!> \param sub_proc_map ...
!> \param my_ia_start ...
!> \param my_ia_end ...
!> \param my_group_L_size ...
!> \param sizes_ia ...
!> \param starts_ia ...
!> \param ends_ia ...
!> \param mp2_env ...
!> \param alpha_case ...
! **************************************************************************************************
   SUBROUTINE create_dbcsr_gamma(Gamma_2D, homo, virtual, dimen_ia, para_env_sub, sub_proc_map, &
                                 my_ia_start, my_ia_end, my_group_L_size, &
                                 sizes_ia, starts_ia, ends_ia, &
                                 mp2_env, alpha_case)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Gamma_2D
      INTEGER                                            :: homo, virtual, dimen_ia
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sub_proc_map
      INTEGER                                            :: my_ia_start, my_ia_end, my_group_L_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sizes_ia, starts_ia, ends_ia
      TYPE(mp2_type), POINTER                            :: mp2_env
      LOGICAL                                            :: alpha_case

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_dbcsr_gamma', &
         routineP = moduleN//':'//routineN

      INTEGER :: dummy_proc, handle, i_global, i_local, iaia, iiB, iii, itmp(2), j_global, &
         j_local, jjB, jjj, kkB, mypcol, myprow, ncol_block, ncol_local, npcol, nprow, nrow_block, &
         nrow_local, number_of_rec, number_of_send, proc_receive, proc_send, proc_shift, &
         rec_counter, rec_iaia_end, rec_iaia_size, rec_iaia_start, rec_pcol, rec_prow, &
         ref_send_pcol, ref_send_prow, send_counter, send_pcol, send_prow, size_rec_buffer, &
         size_send_buffer
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iii_vet, map_rec_size, map_send_size, &
                                                            req_send
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: grid_2_mepos, grid_ref_2_send_pos, &
                                                            indeces_map_my, mepos_2_grid
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: part_ia
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_ia
      TYPE(index_map), ALLOCATABLE, DIMENSION(:)         :: indeces_rec
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_send

! TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: G_P_ia

      CALL timeset(routineN, handle)

      ! allocate G_P_ia (DBCSR)
      IF (alpha_case) THEN
         NULLIFY (mp2_env%ri_grad%G_P_ia)
         CALL cp_dbcsr_allocate_matrix_set(mp2_env%ri_grad%G_P_ia, &
                                           my_group_L_size)
      ELSE
         NULLIFY (mp2_env%ri_grad%G_P_ia_beta)
         CALL cp_dbcsr_allocate_matrix_set(mp2_env%ri_grad%G_P_ia_beta, &
                                           my_group_L_size)
      ENDIF

      ! create sub blacs env
      NULLIFY (blacs_env)
      CALL cp_blacs_env_create(blacs_env=blacs_env, para_env=para_env_sub)

      ! create the fm_ia buffer matrix
      NULLIFY (fm_ia)
      NULLIFY (fm_struct)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=homo, &
                               ncol_global=virtual, para_env=para_env_sub)
      CALL cp_fm_create(fm_ia, fm_struct, name="fm_ia")
      ! release structure
      CALL cp_fm_struct_release(fm_struct)
      ! release blacs_env
      CALL cp_blacs_env_release(blacs_env)

      ! get array information
      CALL cp_fm_get_info(matrix=fm_ia, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          nrow_block=nrow_block, &
                          ncol_block=ncol_block)
      myprow = fm_ia%matrix_struct%context%mepos(1)
      mypcol = fm_ia%matrix_struct%context%mepos(2)
      nprow = fm_ia%matrix_struct%context%num_pe(1)
      npcol = fm_ia%matrix_struct%context%num_pe(2)

      ! 0) create array containing the processes position and supporting infos
      ALLOCATE (grid_2_mepos(0:nprow-1, 0:npcol-1))
      grid_2_mepos = 0
      ALLOCATE (mepos_2_grid(0:para_env_sub%num_pe-1, 2))
      mepos_2_grid = 0
      ! fill the info array
      grid_2_mepos(myprow, mypcol) = para_env_sub%mepos
      mepos_2_grid(para_env_sub%mepos, 1) = myprow
      mepos_2_grid(para_env_sub%mepos, 2) = mypcol
      ! sum infos
      CALL mp_sum(grid_2_mepos, para_env_sub%group)
      CALL mp_sum(mepos_2_grid, para_env_sub%group)

      ! loop over local index range and define the sending map
      ALLOCATE (map_send_size(0:para_env_sub%num_pe-1))
      map_send_size = 0
      DO iaia = my_ia_start, my_ia_end
         i_global = (iaia-1)/virtual+1
         j_global = MOD(iaia-1, virtual)+1
         send_prow = cp_fm_indxg2p(i_global, nrow_block, dummy_proc, &
                                   fm_ia%matrix_struct%first_p_pos(1), nprow)
         send_pcol = cp_fm_indxg2p(j_global, ncol_block, dummy_proc, &
                                   fm_ia%matrix_struct%first_p_pos(2), npcol)
         proc_send = grid_2_mepos(send_prow, send_pcol)
         map_send_size(proc_send) = map_send_size(proc_send)+1
      END DO

      ! loop over local data of fm_ia and define the receiving map
      ALLOCATE (map_rec_size(0:para_env_sub%num_pe-1))
      map_rec_size = 0
      part_ia = FLOAT(dimen_ia)/FLOAT(para_env_sub%num_pe)

      DO iiB = 1, nrow_local
         i_global = row_indices(iiB)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            iaia = (i_global-1)*virtual+j_global
            proc_receive = INT(FLOAT(iaia-1)/part_ia)
            proc_receive = MAX(0, proc_receive)
            proc_receive = MIN(proc_receive, para_env_sub%num_pe-1)
            DO
               itmp = get_limit(dimen_ia, para_env_sub%num_pe, proc_receive)
               IF (iaia >= itmp(1) .AND. iaia <= itmp(2)) EXIT
               IF (iaia < itmp(1)) proc_receive = proc_receive-1
               IF (iaia > itmp(2)) proc_receive = proc_receive+1
            END DO
            map_rec_size(proc_receive) = map_rec_size(proc_receive)+1
         END DO
      END DO

      ! allocate the buffer for sending data
      number_of_send = 0
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         IF (map_send_size(proc_send) > 0) THEN
            number_of_send = number_of_send+1
         END IF
      END DO
      ! allocate the structure that will hold the messages to be sent
      ALLOCATE (buffer_send(number_of_send))
      ! and the map from the grid of processess to the message position
      ALLOCATE (grid_ref_2_send_pos(0:nprow-1, 0:npcol-1))
      grid_ref_2_send_pos = 0
      ! finally allocate each message
      send_counter = 0
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         size_send_buffer = map_send_size(proc_send)
         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            ! allocate the sending buffer (msg)
            ALLOCATE (buffer_send(send_counter)%msg(size_send_buffer))
            buffer_send(send_counter)%proc = proc_send
            ! get the pointer to prow, pcol of the process that has
            ! to receive this message
            ref_send_prow = mepos_2_grid(proc_send, 1)
            ref_send_pcol = mepos_2_grid(proc_send, 2)
            ! save the rank of the process that has to receive this message
            grid_ref_2_send_pos(ref_send_prow, ref_send_pcol) = send_counter
         END IF
      END DO

      ! allocate the buffer for receiving data
      number_of_rec = 0
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
         IF (map_rec_size(proc_receive) > 0) THEN
            number_of_rec = number_of_rec+1
         END IF
      END DO
      ! allocate the structure that will hold the messages to be received
      ! and relative indeces
      ALLOCATE (buffer_rec(number_of_rec))
      ALLOCATE (indeces_rec(number_of_rec))
      ! finally allocate each message and fill the array of indeces
      rec_counter = 0
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)
         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1
            ! prepare the buffer for receive
            ALLOCATE (buffer_rec(rec_counter)%msg(size_rec_buffer))
            buffer_rec(rec_counter)%proc = proc_receive
            ! create the indeces array
            ALLOCATE (indeces_rec(rec_counter)%map(2, size_rec_buffer))
            indeces_rec(rec_counter)%map = 0
            rec_iaia_start = starts_ia(proc_receive)
            rec_iaia_end = ends_ia(proc_receive)
            rec_iaia_size = sizes_ia(proc_receive)
            iii = 0
            DO iaia = rec_iaia_start, rec_iaia_end
               i_global = (iaia-1)/virtual+1
               j_global = MOD(iaia-1, virtual)+1
               rec_prow = cp_fm_indxg2p(i_global, nrow_block, dummy_proc, &
                                        fm_ia%matrix_struct%first_p_pos(1), nprow)
               rec_pcol = cp_fm_indxg2p(j_global, ncol_block, dummy_proc, &
                                        fm_ia%matrix_struct%first_p_pos(2), npcol)
               IF (grid_2_mepos(rec_prow, rec_pcol) /= para_env_sub%mepos) CYCLE
               iii = iii+1
               i_local = cp_fm_indxg2l(i_global, nrow_block, dummy_proc, &
                                       fm_ia%matrix_struct%first_p_pos(1), nprow)
               j_local = cp_fm_indxg2l(j_global, ncol_block, dummy_proc, &
                                       fm_ia%matrix_struct%first_p_pos(2), npcol)
               indeces_rec(rec_counter)%map(1, iii) = i_local
               indeces_rec(rec_counter)%map(2, iii) = j_local
            END DO
         END IF
      END DO
      ! and create the index map for my local data
      IF (map_rec_size(para_env_sub%mepos) > 0) THEN
         size_rec_buffer = map_rec_size(para_env_sub%mepos)
         ALLOCATE (indeces_map_my(2, size_rec_buffer))
         indeces_map_my = 0
         iii = 0
         DO iaia = my_ia_start, my_ia_end
            i_global = (iaia-1)/virtual+1
            j_global = MOD(iaia-1, virtual)+1
            rec_prow = cp_fm_indxg2p(i_global, nrow_block, dummy_proc, &
                                     fm_ia%matrix_struct%first_p_pos(1), nprow)
            rec_pcol = cp_fm_indxg2p(j_global, ncol_block, dummy_proc, &
                                     fm_ia%matrix_struct%first_p_pos(2), npcol)
            IF (grid_2_mepos(rec_prow, rec_pcol) /= para_env_sub%mepos) CYCLE
            iii = iii+1
            i_local = cp_fm_indxg2l(i_global, nrow_block, dummy_proc, &
                                    fm_ia%matrix_struct%first_p_pos(1), nprow)
            j_local = cp_fm_indxg2l(j_global, ncol_block, dummy_proc, &
                                    fm_ia%matrix_struct%first_p_pos(2), npcol)
            indeces_map_my(1, iii) = i_local
            indeces_map_my(2, iii) = j_local
         END DO
      END IF

      ! auxiliary vector of indeces for the send buffer
      ALLOCATE (iii_vet(number_of_send))
      ! vector for the send requests
      ALLOCATE (req_send(number_of_send))
      ! loop over auxiliary basis function and redistribute into a fm
      ! and then compy the fm into a dbcsr matrix
      DO kkB = 1, my_group_L_size
         ! zero the matries of the buffers and post the messages to be received
         CALL cp_fm_set_all(matrix=fm_ia, alpha=0.0_dp)
         rec_counter = 0
         DO proc_shift = 1, para_env_sub%num_pe-1
            proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
            IF (map_rec_size(proc_receive) > 0) THEN
               rec_counter = rec_counter+1
               buffer_rec(rec_counter)%msg = 0.0_dp
               CALL mp_irecv(buffer_rec(rec_counter)%msg, proc_receive, para_env_sub%group, &
                             buffer_rec(rec_counter)%msg_req)
            END IF
         END DO
         ! fill the sending buffer and send the messages
         DO send_counter = 1, number_of_send
            buffer_send(send_counter)%msg = 0.0_dp
         END DO
         iii_vet = 0
         jjj = 0
         DO iaia = my_ia_start, my_ia_end
            i_global = (iaia-1)/virtual+1
            j_global = MOD(iaia-1, virtual)+1
            send_prow = cp_fm_indxg2p(i_global, nrow_block, dummy_proc, &
                                      fm_ia%matrix_struct%first_p_pos(1), nprow)
            send_pcol = cp_fm_indxg2p(j_global, ncol_block, dummy_proc, &
                                      fm_ia%matrix_struct%first_p_pos(2), npcol)
            proc_send = grid_2_mepos(send_prow, send_pcol)
            ! we don't need to send to ourselves
            IF (grid_2_mepos(send_prow, send_pcol) == para_env_sub%mepos) THEN
               ! filling fm_ia with local data
               jjj = jjj+1
               i_local = indeces_map_my(1, jjj)
               j_local = indeces_map_my(2, jjj)
               fm_ia%local_data(i_local, j_local) = Gamma_2D(iaia-my_ia_start+1, kkB)
            ELSE
               send_counter = grid_ref_2_send_pos(send_prow, send_pcol)
               iii_vet(send_counter) = iii_vet(send_counter)+1
               iii = iii_vet(send_counter)
               buffer_send(send_counter)%msg(iii) = Gamma_2D(iaia-my_ia_start+1, kkB)
            END IF
         END DO
         req_send = 0
         send_counter = 0
         DO proc_shift = 1, para_env_sub%num_pe-1
            proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
            IF (map_send_size(proc_send) > 0) THEN
               send_counter = send_counter+1
               CALL mp_isend(buffer_send(send_counter)%msg, proc_send, para_env_sub%group, &
                             buffer_send(send_counter)%msg_req)
               req_send(send_counter) = buffer_send(send_counter)%msg_req
            END IF
         END DO

         ! receive the massages and fill the fm_ia
         rec_counter = 0
         DO proc_shift = 1, para_env_sub%num_pe-1
            proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
            size_rec_buffer = map_rec_size(proc_receive)
            IF (map_rec_size(proc_receive) > 0) THEN
               rec_counter = rec_counter+1
               ! wait for the message
               CALL mp_wait(buffer_rec(rec_counter)%msg_req)
               DO iii = 1, size_rec_buffer
                  i_local = indeces_rec(rec_counter)%map(1, iii)
                  j_local = indeces_rec(rec_counter)%map(2, iii)
                  fm_ia%local_data(i_local, j_local) = buffer_rec(rec_counter)%msg(iii)
               END DO
            END IF
         END DO

         ! wait all
         CALL mp_waitall(req_send(:))

         ! now create the DBCSR matrix and copy fm_ia into it
         IF (alpha_case) THEN
            ALLOCATE (mp2_env%ri_grad%G_P_ia(kkB)%matrix)
            CALL cp_dbcsr_init(mp2_env%ri_grad%G_P_ia(kkB)%matrix)
            CALL cp_dbcsr_m_by_n_from_template(mp2_env%ri_grad%G_P_ia(kkB)%matrix, &
                                               template=mp2_env%ri_grad%mo_coeff_o, &
                                               m=homo, n=virtual, sym=dbcsr_type_no_symmetry)
            CALL copy_fm_to_dbcsr(fm_ia, mp2_env%ri_grad%G_P_ia(kkB)%matrix, keep_sparsity=.FALSE.)
         ELSE
            ALLOCATE (mp2_env%ri_grad%G_P_ia_beta(kkB)%matrix)
            CALL cp_dbcsr_init(mp2_env%ri_grad%G_P_ia_beta(kkB)%matrix)
            CALL cp_dbcsr_m_by_n_from_template(mp2_env%ri_grad%G_P_ia_beta(kkB)%matrix, &
                                               template=mp2_env%ri_grad%mo_coeff_o_beta, &
                                               m=homo, n=virtual, sym=dbcsr_type_no_symmetry)
            CALL copy_fm_to_dbcsr(fm_ia, mp2_env%ri_grad%G_P_ia_beta(kkB)%matrix, keep_sparsity=.FALSE.)
         ENDIF

      END DO

      ! deallocate stuff
      DEALLOCATE (Gamma_2D)
      DEALLOCATE (iii_vet)
      DEALLOCATE (req_send)
      IF (map_rec_size(para_env_sub%mepos) > 0) THEN
         DEALLOCATE (indeces_map_my)
      END IF
      DO rec_counter = 1, number_of_rec
         DEALLOCATE (indeces_rec(rec_counter)%map)
         DEALLOCATE (buffer_rec(rec_counter)%msg)
      END DO
      DEALLOCATE (indeces_rec)
      DEALLOCATE (buffer_rec)
      DO send_counter = 1, number_of_send
         DEALLOCATE (buffer_send(send_counter)%msg)
      END DO
      DEALLOCATE (buffer_send)
      DEALLOCATE (map_send_size)
      DEALLOCATE (map_rec_size)
      DEALLOCATE (grid_2_mepos)
      DEALLOCATE (mepos_2_grid)

      ! release buffer matrix
      CALL cp_fm_release(fm_ia)

      CALL timestop(handle)

   END SUBROUTINE create_dbcsr_gamma

! **************************************************************************************************
!> \brief ...
!> \param mat ...
! **************************************************************************************************
   SUBROUTINE write_array(mat)
      REAL(KIND=dp), DIMENSION(:, :)                     :: mat

      INTEGER                                            :: iii, jjj

      WRITE (*, *)
      DO iii = 1, SIZE(mat, 1)
         WRITE (*, *) iii
         DO jjj = 1, SIZE(mat, 2), 20
            WRITE (*, '(1000F10.5)') mat(iii, jjj:MIN(SIZE(mat, 2), jjj+19))
         END DO
         WRITE (*, *)
      END DO
      WRITE (*, *)
   END SUBROUTINE

END MODULE mp2_ri_grad_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate RI-GPW-MP2 energy using pw
!> \par History
!>      06.2012 created [Mauro Del Ben]
! **************************************************************************************************
MODULE mp2_ri_gpw
   USE ai_coulomb,                      ONLY: coulomb2_new
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE bibliography,                    ONLY: DelBen2013,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                              cp_blacs_env_release,&
                                              cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_filter, &
        cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, &
        cp_dbcsr_set, cp_dbcsr_type, dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_m_by_n_from_template
   USE cp_eri_mme_interface,            ONLY: eri_mme_2c_integrate,&
                                              eri_mme_set_params
   USE cp_fm_basic_linalg,              ONLY: cp_fm_triangular_invert
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose
   USE cp_fm_diag,                      ONLY: cp_fm_syevx
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE eri_mme_types,                   ONLY: eri_mme_param
   USE gaussian_gridlevels,             ONLY: gaussian_gridlevel
   USE input_constants,                 ONLY: do_eri_gpw,&
                                              do_eri_mme,&
                                              ri_coulomb,&
                                              ri_overlap
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE machine,                         ONLY: m_flush,&
                                              m_memory,&
                                              m_walltime
   USE mathconstants,                   ONLY: pi
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_comm_split_direct,&
                                              mp_max,&
                                              mp_min,&
                                              mp_sendrecv,&
                                              mp_sum,&
                                              mp_sync
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE mp2_ri_grad_util,                ONLY: complete_gamma
   USE mp2_types,                       ONLY: mp2_type
   USE orbital_pointers,                ONLY: ncoset
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_copy,&
                                              pw_scale,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: REALDATA3D,&
                                              REALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: calculate_wavefunction
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_integrate_potential,          ONLY: integrate_pgf_product_rspace,&
                                              integrate_v_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              realspace_grid_p_type,&
                                              rs_grid_release,&
                                              rs_grid_retain
   USE rs_pw_interface,                 ONLY: potential_pw2rs
   USE task_list_types,                 ONLY: task_list_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_ri_gpw'

   PUBLIC :: mp2_ri_gpw_compute_in, &
             mp2_ri_gpw_compute_en

CONTAINS

! **************************************************************************************************
!> \brief with ri mp2 gpw
!> \param BIb_C ...
!> \param BIb_C_gw ...
!> \param ends_array ...
!> \param ends_B_virtual ...
!> \param sizes_array ...
!> \param sizes_B_virtual ...
!> \param starts_array ...
!> \param starts_B_virtual ...
!> \param dimen_RI ...
!> \param qs_env ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param color_sub ...
!> \param dft_control ...
!> \param cell ...
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param mo_coeff ...
!> \param nmo ...
!> \param homo ...
!> \param rho_r ...
!> \param rho_g ...
!> \param pot_g ...
!> \param mat_munu ...
!> \param sab_orb_sub ...
!> \param pw_env_sub ...
!> \param poisson_env ...
!> \param auxbas_pw_pool ...
!> \param task_list_sub ...
!> \param mo_coeff_o ...
!> \param mo_coeff_v ...
!> \param mo_coeff_all ...
!> \param mo_coeff_gw ...
!> \param eps_filter ...
!> \param unit_nr ...
!> \param mp2_memory ...
!> \param calc_PQ_cond_num ...
!> \param calc_forces ...
!> \param blacs_env_sub ...
!> \param my_do_gw ...
!> \param starts_B_all ...
!> \param sizes_B_all ...
!> \param ends_B_all ...
!> \param gw_corr_lev_occ ...
!> \param gw_corr_lev_virt ...
!> \param BIb_C_beta ...
!> \param BIb_C_gw_beta ...
!> \param ends_B_virtual_beta ...
!> \param sizes_B_virtual_beta ...
!> \param starts_B_virtual_beta ...
!> \param homo_beta ...
!> \param mo_coeff_o_beta ...
!> \param mo_coeff_v_beta ...
!> \param mo_coeff_all_beta ...
!> \param mo_coeff_gw_beta ...
!> \param do_im_time ...
!> \param ri_metric ...
!> \param ri_metric_gw ...
!> \param mat_B_munu ...
!> \param my_Lrows ...
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE mp2_ri_gpw_compute_in(BIb_C, BIb_C_gw, ends_array, ends_B_virtual, &
                                    sizes_array, sizes_B_virtual, starts_array, starts_B_virtual, &
                                    dimen_RI, qs_env, para_env, para_env_sub, color_sub, dft_control, &
                                    cell, particle_set, atomic_kind_set, qs_kind_set, mo_coeff, nmo, homo, &
                                    rho_r, rho_g, pot_g, mat_munu, sab_orb_sub, pw_env_sub, &
                                    poisson_env, auxbas_pw_pool, task_list_sub, mo_coeff_o, mo_coeff_v, mo_coeff_all, &
                                    mo_coeff_gw, eps_filter, unit_nr, &
                                    mp2_memory, calc_PQ_cond_num, calc_forces, blacs_env_sub, my_do_gw, &
                                    starts_B_all, sizes_B_all, ends_B_all, gw_corr_lev_occ, gw_corr_lev_virt, &
                                    BIb_C_beta, BIb_C_gw_beta, ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta, &
                                    homo_beta, mo_coeff_o_beta, mo_coeff_v_beta, mo_coeff_all_beta, mo_coeff_gw_beta, do_im_time, &
                                    ri_metric, ri_metric_gw, mat_B_munu, my_Lrows)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C, BIb_C_gw
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_array, ends_B_virtual, sizes_array, &
                                                            sizes_B_virtual, starts_array, &
                                                            starts_B_virtual
      INTEGER                                            :: dimen_RI
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER                                            :: color_sub
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      INTEGER                                            :: nmo, homo
      TYPE(pw_p_type)                                    :: rho_r, rho_g, pot_g
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb_sub
      TYPE(pw_env_type), POINTER                         :: pw_env_sub
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(task_list_type), POINTER                      :: task_list_sub
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_o, mo_coeff_v, mo_coeff_all, &
                                                            mo_coeff_gw
      REAL(KIND=dp)                                      :: eps_filter
      INTEGER                                            :: unit_nr
      REAL(KIND=dp)                                      :: mp2_memory
      LOGICAL                                            :: calc_PQ_cond_num, calc_forces
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env_sub
      LOGICAL                                            :: my_do_gw
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: starts_B_all, sizes_B_all, ends_B_all
      INTEGER                                            :: gw_corr_lev_occ, gw_corr_lev_virt
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :), &
         OPTIONAL                                        :: BIb_C_beta, BIb_C_gw_beta
      INTEGER, ALLOCATABLE, DIMENSION(:), OPTIONAL       :: ends_B_virtual_beta, &
                                                            sizes_B_virtual_beta, &
                                                            starts_B_virtual_beta
      INTEGER, OPTIONAL                                  :: homo_beta
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: mo_coeff_o_beta, mo_coeff_v_beta, &
                                                            mo_coeff_all_beta, mo_coeff_gw_beta
      LOGICAL, OPTIONAL                                  :: do_im_time
      INTEGER, OPTIONAL                                  :: ri_metric, ri_metric_gw
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mat_B_munu
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :), &
         OPTIONAL                                        :: my_Lrows

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_ri_gpw_compute_in', &
         routineP = moduleN//':'//routineN

      INTEGER :: eri_method, gw_corr_lev_total, handle, handle2, handle3, i, i_counter, itmp(2), &
         LLL, max_row_col_local, max_row_col_local_beta, max_row_col_local_gw, my_B_all_end, &
         my_B_all_size, my_B_all_start, my_B_size, my_B_size_beta, my_B_virtual_end, &
         my_B_virtual_end_beta, my_B_virtual_start, my_B_virtual_start_beta, my_group_L_end, &
         my_group_L_size, my_group_L_start, my_ri_metric, my_ri_metric_gw, ncol_local, &
         nfullcols_total, nfullrows_total, ngroup, nrow_local, num_small_eigen, virtual, &
         virtual_beta
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sub_proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_col_row_info, &
                                                            local_col_row_info_beta, &
                                                            local_col_row_info_gw
      LOGICAL                                            :: do_alpha_beta, do_mult_2c, my_do_im_time
      REAL(KIND=dp)                                      :: cond_num, mem_for_iaK, pair_energy, &
                                                            wfn_size
      TYPE(cp_dbcsr_type) :: matrix_ia_jb, matrix_ia_jb_beta, matrix_ia_jnu, matrix_ia_jnu_beta, &
         matrix_in_jm, matrix_in_jm_beta, matrix_in_jnu, matrix_in_jnu_beta
      TYPE(cp_fm_type), POINTER                          :: fm_BIb_gw, fm_BIb_gw_beta, fm_BIb_jb, &
                                                            fm_BIb_jb_beta, fm_matrix_L, &
                                                            fm_matrix_L_RI_metric
      TYPE(cp_para_env_type), POINTER                    :: para_env_L
      TYPE(eri_mme_param), POINTER                       :: eri_param
      TYPE(pw_p_type)                                    :: psi_L

      CALL timeset(routineN, handle)

      CALL cite_reference(DelBen2013)

      do_alpha_beta = .FALSE.
      IF (PRESENT(BIb_C_beta) .AND. &
          PRESENT(ends_B_virtual_beta) .AND. &
          PRESENT(sizes_B_virtual_beta) .AND. &
          PRESENT(starts_B_virtual_beta) .AND. &
          PRESENT(homo_beta) .AND. &
          PRESENT(mo_coeff_o_beta) .AND. &
          PRESENT(mo_coeff_v_beta)) do_alpha_beta = .TRUE.

      my_do_im_time = .FALSE.
      IF (PRESENT(do_im_time)) my_do_im_time = do_im_time

      ! default: Coulomb metric
      my_ri_metric = ri_coulomb
      IF (PRESENT(ri_metric)) my_ri_metric = ri_metric

      ! default: Coulomb metric
      my_ri_metric_gw = ri_coulomb
      IF (PRESENT(ri_metric_gw)) my_ri_metric_gw = ri_metric_gw

      ! imag. time only without Coulomb metric implemented
      IF(.NOT.(.NOT. (ri_metric == ri_coulomb .AND. my_do_im_time)))CALL cp__a("mp2_ri_gpw.F",295)

      virtual = nmo-homo

      ! in case of imaginary time, we do not need the intermediate matrices
      IF (.NOT. my_do_im_time) THEN

         CALL create_intermediate_matrices(matrix_ia_jnu, matrix_ia_jb, mo_coeff_o, virtual, homo, &
                                           fm_BIb_jb, "fm_BIb_jb", max_row_col_local, &
                                           nfullrows_total, nfullcols_total, nrow_local, &
                                           ncol_local, blacs_env_sub, &
                                           para_env_sub, local_col_row_info)

         CALL create_parallelization_arrays(para_env_sub, ends_B_virtual, sizes_B_virtual, &
                                            starts_B_virtual, my_B_virtual_start, &
                                            my_B_virtual_end, my_B_size, &
                                            virtual)

         IF (do_alpha_beta) THEN

            virtual_beta = nmo-homo_beta

            CALL create_intermediate_matrices(matrix_ia_jnu_beta, matrix_ia_jb_beta, mo_coeff_o_beta, &
                                              virtual_beta, homo_beta, &
                                              fm_BIb_jb_beta, "fm_BIb_jb_beta", &
                                              max_row_col_local_beta, &
                                              nfullrows_total, nfullcols_total, nrow_local, &
                                              ncol_local, blacs_env_sub, &
                                              para_env_sub, local_col_row_info_beta)

            CALL create_parallelization_arrays(para_env_sub, ends_B_virtual_beta, sizes_B_virtual_beta, &
                                               starts_B_virtual_beta, my_B_virtual_start_beta, &
                                               my_B_virtual_end_beta, my_B_size_beta, &
                                               virtual_beta)

         END IF

      END IF ! not for imaginary time

      ! in the case of G0W0, we need (K|nm), n,m may be occ or virt (m restricted to corrected levels)
      IF (my_do_gw) THEN

         gw_corr_lev_total = gw_corr_lev_virt+gw_corr_lev_occ

         CALL create_intermediate_matrices(matrix_in_jnu, matrix_in_jm, mo_coeff_gw, &
                                           nmo, gw_corr_lev_total, &
                                           fm_BIb_gw, "fm_BIb_gw", &
                                           max_row_col_local_gw, &
                                           nfullrows_total, nfullcols_total, nrow_local, &
                                           ncol_local, blacs_env_sub, &
                                           para_env_sub, local_col_row_info_gw)

         CALL create_parallelization_arrays(para_env_sub, ends_B_all, sizes_B_all, &
                                            starts_B_all, my_B_all_start, &
                                            my_B_all_end, my_B_all_size, &
                                            nmo)

         IF (do_alpha_beta) THEN
            ! deallocate local_col_row_info_gw, otherwise it gets twice allocated in create_intermediate_m
            DEALLOCATE (local_col_row_info_gw)
            CALL create_intermediate_matrices(matrix_in_jnu_beta, matrix_in_jm_beta, mo_coeff_gw_beta, &
                                              nmo, gw_corr_lev_total, &
                                              fm_BIb_gw_beta, "fm_BIb_gw_beta", &
                                              max_row_col_local_gw, &
                                              nfullrows_total, nfullcols_total, nrow_local, &
                                              ncol_local, blacs_env_sub, &
                                              para_env_sub, local_col_row_info_gw)

            ! we don"t need parallelization arrays for beta since the matrix sizes of B_nm^P is the same
            ! for the beta case and therefore the parallelization of beta is the same than for alpha

         END IF
      END IF

      wfn_size = REAL(SIZE(rho_r%pw%cr3d), KIND=dp)
      CALL mp_max(wfn_size, para_env%group)

      ngroup = para_env%num_pe/para_env_sub%num_pe

      ALLOCATE (sub_proc_map(-para_env_sub%num_pe:2*para_env_sub%num_pe-1))
      DO i = 0, para_env_sub%num_pe-1
         sub_proc_map(i) = i
         sub_proc_map(-i-1) = para_env_sub%num_pe-i-1
         sub_proc_map(para_env_sub%num_pe+i) = i
      END DO

      ! start real calculation
      NULLIFY (psi_L%pw)
      CALL pw_pool_create_pw(auxbas_pw_pool, psi_L%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      do_mult_2c = qs_env%mp2_env%mp2_gpw%do_mult_2c
      eri_method = qs_env%mp2_env%eri_method
      eri_param => qs_env%mp2_env%eri_mme_param

      ! Preparations for ERI MME method to compute 2c ERIs
      IF (qs_env%mp2_env%eri_method .EQ. do_eri_mme) THEN
         CALL eri_mme_set_params(eri_param, cell, qs_kind_set, basis_type="RI_AUX", para_env=para_env)
      ENDIF

      CALL calculate_Lmin1(qs_env, eri_method, eri_param, para_env, para_env_sub, para_env_L, mp2_memory, &
                           fm_matrix_L, ngroup, color_sub, dimen_RI, &
                           mo_coeff, dft_control, psi_L, rho_r, rho_g, pot_g, pw_env_sub, poisson_env, &
                           my_group_L_size, my_group_L_start, my_group_L_end, sab_orb_sub, &
                           sizes_array, starts_array, ends_array, calc_PQ_cond_num, cond_num, &
                           num_small_eigen, auxbas_pw_pool, do_mult_2c, my_ri_metric, &
                           fm_matrix_L_RI_metric, my_do_im_time)

      IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
         "RI_INFO| Cholesky decomposition group size:", para_env_L%num_pe
      IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
         "RI_INFO| Number of groups for auxiliary basis functions", ngroup
      IF (calc_PQ_cond_num) THEN
         IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="(T3,A,T67,ES14.5)") &
            "RI_INFO| Condition number of the (P|Q):", cond_num
         IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
            "RI_INFO| Number of Eigenvalue of (P|Q) smaller than 10^(-3):", num_small_eigen
      END IF

      ! replicate the necessary row of the L^{-1} matrix on each proc
      CALL grep_Lcols(para_env_L, dimen_RI, fm_matrix_L, &
                      my_group_L_start, my_group_L_end, my_group_L_size, my_Lrows)
      ! clean the L^{-1} matrix
      CALL cp_fm_release(fm_matrix_L)

      ! in case of imag. time we need the para_env_L later
      IF (.NOT. my_do_im_time) THEN
         CALL cp_para_env_release(para_env_L)
      END IF

      IF (calc_forces) THEN
         ! we need (P|Q)^(-1/2) for future use, just save it
         ! in a fully (home made) distributed way
         itmp = get_limit(dimen_RI, para_env_sub%num_pe, para_env_sub%mepos)
         lll = itmp(2)-itmp(1)+1
         ALLOCATE (qs_env%mp2_env%ri_grad%PQ_half(lll, my_group_L_size))
         qs_env%mp2_env%ri_grad%PQ_half(:, :) = my_Lrows(itmp(1):itmp(2), 1:my_group_L_size)
      END IF

      IF (unit_nr > 0) THEN
         WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
            "RI_INFO| Occupied  basis set size:", homo, &
            "RI_INFO| Virtual   basis set size:", virtual, &
            "RI_INFO| Auxiliary basis set size:", dimen_RI

         mem_for_iaK = dimen_RI*REAL(homo, KIND=dp)*virtual*8.0_dp/(1024_dp**2)
         IF (do_alpha_beta) mem_for_iaK = mem_for_iaK+ &
                                          dimen_RI*REAL(homo_beta, KIND=dp)*(nmo-homo_beta)*8.0_dp/(1024_dp**2)

         WRITE (unit_nr, '(T3,A,T66,F11.2,A4)') 'RI_INFO| Total memory for (ia|K) integrals:', &
            mem_for_iaK, ' MiB'
         IF (my_do_gw) THEN
            mem_for_iaK = dimen_RI*REAL(nmo, KIND=dp)*gw_corr_lev_total*8.0_dp/(1024_dp**2)

            WRITE (unit_nr, '(T3,A,T66,F11.2,A4)') 'RI_INFO| Total memory for G0W0-(nm|K) integrals:', &
               mem_for_iaK, ' MiB'
         END IF
         CALL m_flush(unit_nr)
      ENDIF

      CALL timeset(routineN//"_alloc", handle2)
      CALL mp_sync(para_env%group) ! sync to see memory output

      ! in case we do imaginary time, we need the matrix B_alphabeta^P
      ! B has size N_molecular_orb x N_molecular_orb and contains N_RI_basis_functions
      ! copies (one for every RI basis function)
      IF (my_do_im_time) THEN

         NULLIFY (mat_B_munu)
         CALL cp_dbcsr_allocate_matrix_set(mat_B_munu, my_group_L_size)

      ELSE

         ! array that will store the (ia|K) integrals
         ALLOCATE (BIb_C(my_group_L_size, my_B_size, homo))
         BIb_C = 0.0_dp

         IF (do_alpha_beta) THEN
            ALLOCATE (BIb_C_beta(my_group_L_size, my_B_size_beta, homo_beta))
            BIb_C_beta = 0.0_dp
         END IF

      END IF ! imag. time

      ! in the case of GW, we also need (nm|K)
      IF (my_do_gw) THEN

         ALLOCATE (BIb_C_gw(my_group_L_size, my_B_all_size, gw_corr_lev_total))
         BIb_C_gw = 0.0_dp
         IF (do_alpha_beta) THEN
            ALLOCATE (BIb_C_gw_beta(my_group_L_size, my_B_all_size, gw_corr_lev_total))
            BIb_C_gw_beta = 0.0_dp
         END IF

      END IF

      CALL timestop(handle2)

      CALL timeset(routineN//"_loop", handle2)
      i_counter = 0
      DO LLL = my_group_L_start, my_group_L_end
         i_counter = i_counter+1

         ! pseudo psi_L
         CALL calculate_wavefunction(mo_coeff, 1, psi_L, rho_g, atomic_kind_set, &
                                     qs_kind_set, cell, dft_control, particle_set, pw_env_sub, &
                                     basis_type="RI_AUX", &
                                     external_vector=my_Lrows(:, LLL-my_group_L_start+1))

         CALL timeset(routineN//"_pot", handle3)
         rho_r%pw%cr3d = psi_L%pw%cr3d

         ! in case we do Coulomb metric for RI, we need the Coulomb operator, but for RI with the
         ! overlap metric, we do not need the Coulomb operator
         IF (my_ri_metric /= ri_overlap) THEN
            CALL pw_transfer(rho_r%pw, rho_g%pw)
            CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
            CALL pw_transfer(pot_g%pw, rho_r%pw)
         END IF

         CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)
         CALL timestop(handle3)

         ! and finally (K|mu nu)
         CALL timeset(routineN//"_int", handle3)
         CALL cp_dbcsr_set(mat_munu%matrix, 0.0_dp)
         CALL integrate_v_rspace(rho_r, hmat=mat_munu, qs_env=qs_env, calculate_forces=.FALSE., compute_tau=.FALSE., gapw=.FALSE., &
                                 pw_env_external=pw_env_sub, task_list_external=task_list_sub)
         CALL timestop(handle3)

         ! in case of imaginary time, we do not transform from AOs to MOs
         IF (my_do_im_time) THEN

            ALLOCATE (mat_B_munu(i_counter)%matrix)

            CALL cp_dbcsr_init(mat_B_munu(i_counter)%matrix)
!JW 2del        CALL cp_dbcsr_create(matrix=mat_B_munu(i_counter)%matrix,template=mat_munu%matrix)

            CALL cp_dbcsr_copy(mat_B_munu(i_counter)%matrix, mat_munu%matrix)

            ! filter to save memory but do not filter the first matrix because we need the matrix structure
            CALL cp_dbcsr_filter(mat_B_munu(i_counter)%matrix, eps_filter)

         ELSE

            CALL ao_to_mo_and_store_B(para_env_sub, mat_munu, matrix_ia_jnu, matrix_ia_jb, &
                                      fm_BIb_jb, BIb_C(i_counter, 1:my_B_size, 1:homo), &
                                      mo_coeff_o, mo_coeff_v, eps_filter, max_row_col_local, sub_proc_map, &
                                      local_col_row_info, my_B_virtual_end, my_B_virtual_start, "alpha")

            IF (do_alpha_beta) THEN
               CALL ao_to_mo_and_store_B(para_env_sub, mat_munu, matrix_ia_jnu_beta, matrix_ia_jb_beta, &
                                         fm_BIb_jb_beta, BIb_C_beta(i_counter, 1:my_B_size_beta, 1:homo_beta), &
                                         mo_coeff_o_beta, mo_coeff_v_beta, eps_filter, max_row_col_local_beta, sub_proc_map, &
                                         local_col_row_info_beta, my_B_virtual_end_beta, my_B_virtual_start_beta, "beta")

            ENDIF

            IF (my_do_gw .AND. (my_ri_metric_gw == my_ri_metric)) THEN
               CALL ao_to_mo_and_store_B(para_env_sub, mat_munu, matrix_in_jnu, matrix_in_jm, &
                                         fm_BIb_gw, BIb_C_gw(i_counter, 1:my_B_all_size, 1:gw_corr_lev_total), &
                                         mo_coeff_gw, mo_coeff_all, eps_filter, max_row_col_local_gw, sub_proc_map, &
                                         local_col_row_info_gw, my_B_all_end, my_B_all_start, "gw_alpha")

               IF (do_alpha_beta) THEN
                  CALL ao_to_mo_and_store_B(para_env_sub, mat_munu, matrix_in_jnu_beta, matrix_in_jm_beta, &
                                            fm_BIb_gw_beta, BIb_c_gw_beta(i_counter, 1:my_B_all_size, 1:gw_corr_lev_total), &
                                            mo_coeff_gw_beta, mo_coeff_all_beta, eps_filter, max_row_col_local_gw, sub_proc_map, &
                                            local_col_row_info_gw, my_B_all_end, my_B_all_start, "gw_beta")
               ENDIF
            ENDIF
         END IF ! imag. time

         IF (my_do_gw .AND. (my_ri_metric_gw == my_ri_metric)) THEN
            ! transform (K|mu nu) to (K|nm), n corresponds to corrected GW levels, m is in nmo
            CALL ao_to_mo_and_store_B(para_env_sub, mat_munu, matrix_in_jnu, matrix_in_jm, &
                                      fm_BIb_gw, BIb_C_gw(i_counter, 1:my_B_all_size, 1:gw_corr_lev_total), &
                                      mo_coeff_gw, mo_coeff_all, eps_filter, max_row_col_local_gw, sub_proc_map, &
                                      local_col_row_info_gw, my_B_all_end, my_B_all_start, "gw_alpha")

            ! the same for beta
            IF (do_alpha_beta) THEN
               CALL ao_to_mo_and_store_B(para_env_sub, mat_munu, matrix_in_jnu_beta, matrix_in_jm_beta, &
                                         fm_BIb_gw_beta, BIb_C_gw_beta(i_counter, 1:my_B_all_size, 1:gw_corr_lev_total), &
                                         mo_coeff_gw_beta, mo_coeff_all_beta, eps_filter, max_row_col_local_gw, sub_proc_map, &
                                         local_col_row_info_gw, my_B_all_end, my_B_all_start, "gw_beta")

            END IF
         END IF

      END DO

      ! use possibly different RI metric for GW matrix elements B^nn_P
      IF (my_do_gw .AND. my_ri_metric_gw /= my_ri_metric) THEN

         DEALLOCATE (my_Lrows, sizes_array, starts_array, ends_array)

         ! difference to call of calculate_Lmin1 above: other RI metric for GW
         CALL calculate_Lmin1(qs_env, eri_method, eri_param, para_env, para_env_sub, para_env_L, mp2_memory, &
                              fm_matrix_L, ngroup, color_sub, dimen_RI, &
                              mo_coeff, dft_control, psi_L, rho_r, rho_g, pot_g, pw_env_sub, poisson_env, &
                              my_group_L_size, my_group_L_start, my_group_L_end, sab_orb_sub, &
                              sizes_array, starts_array, ends_array, calc_PQ_cond_num, cond_num, &
                              num_small_eigen, auxbas_pw_pool, do_mult_2c, my_ri_metric_gw, &
                              fm_matrix_L_RI_metric, my_do_im_time)

         ! replicate the necessary row of the L^{-1} matrix on each proc
         CALL grep_Lcols(para_env_L, dimen_RI, fm_matrix_L, &
                         my_group_L_start, my_group_L_end, my_group_L_size, my_Lrows)

         ! clean the L^{-1} matrix
         CALL cp_fm_release(fm_matrix_L)
         CALL cp_para_env_release(para_env_L)

         i_counter = 0
         DO LLL = my_group_L_start, my_group_L_end
            i_counter = i_counter+1

            ! pseudo psi_L
            CALL calculate_wavefunction(mo_coeff, 1, psi_L, rho_g, atomic_kind_set, &
                                        qs_kind_set, cell, dft_control, particle_set, pw_env_sub, &
                                        basis_type="RI_AUX", &
                                        external_vector=my_Lrows(:, LLL-my_group_L_start+1))

            CALL timeset(routineN//"_pot_trunc", handle3)
            rho_r%pw%cr3d = psi_L%pw%cr3d

            ! in case we do Coulomb metric for RI, we need the Coulomb operator, but for RI with the
            ! overlap metric, we do not need the Coulomb operator
            IF (my_ri_metric_gw /= ri_overlap) THEN
               CALL pw_transfer(rho_r%pw, rho_g%pw)
               CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
               CALL pw_transfer(pot_g%pw, rho_r%pw)
            END IF

            CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)

            CALL timestop(handle3)

            ! and finally (K|mu nu)
            CALL timeset(routineN//"_int_trunc", handle3)
            CALL cp_dbcsr_set(mat_munu%matrix, 0.0_dp)
            CALL integrate_v_rspace(rho_r, hmat=mat_munu, qs_env=qs_env, &
                                    calculate_forces=.FALSE., compute_tau=.FALSE., gapw=.FALSE., &
                                    pw_env_external=pw_env_sub, task_list_external=task_list_sub)
            CALL timestop(handle3)

            CALL ao_to_mo_and_store_B(para_env_sub, mat_munu, matrix_in_jnu, matrix_in_jm, &
                                      fm_BIb_gw, BIb_C_gw(i_counter, 1:my_B_all_size, 1:gw_corr_lev_total), &
                                      mo_coeff_gw, mo_coeff_all, eps_filter, max_row_col_local_gw, sub_proc_map, &
                                      local_col_row_info_gw, my_B_all_end, my_B_all_start, "gw_trunc")

         END DO

         ! set this flag so that the truncation of above will be overwritten the next time the poisson_env is used
         IF (my_group_L_start <= my_group_L_end) &
            poisson_env%rebuild = .TRUE.

      END IF ! use truncated operator for GW
      CALL timestop(handle2)

      DEALLOCATE (my_Lrows)

      IF (my_do_im_time) THEN

         ! transfer entries of fm_matrix_L_RI_metric to RPA
         CALL grep_Lcols(para_env_L, dimen_RI, fm_matrix_L_RI_metric, &
                         my_group_L_start, my_group_L_end, my_group_L_size, my_Lrows)
         CALL cp_fm_release(fm_matrix_L_RI_metric)
         CALL cp_para_env_release(para_env_L)

      ELSE

         CALL cp_fm_release(fm_BIb_jb)
         DEALLOCATE (local_col_row_info)

         CALL cp_dbcsr_release(matrix_ia_jnu)
         CALL cp_dbcsr_release(matrix_ia_jb)
         IF (do_alpha_beta) THEN
            CALL cp_dbcsr_release(matrix_ia_jnu_beta)
            CALL cp_dbcsr_release(matrix_ia_jb_beta)
            CALL cp_fm_release(fm_BIb_jb_beta)
            DEALLOCATE (local_col_row_info_beta)
         END IF

      END IF

      IF (my_do_gw) THEN
         CALL cp_dbcsr_release(matrix_in_jnu)
         CALL cp_dbcsr_release(matrix_in_jm)
         CALL cp_fm_release(fm_BIb_gw)
         DEALLOCATE (local_col_row_info_gw)
         IF (do_alpha_beta) THEN
            CALL cp_dbcsr_release(matrix_in_jnu_beta)
            CALL cp_dbcsr_release(matrix_in_jm_beta)
            CALL cp_fm_release(fm_BIb_gw_beta)
         END IF
      END IF

      CALL pw_pool_give_back_pw(auxbas_pw_pool, psi_L%pw)

      DEALLOCATE (sub_proc_map)

      CALL timestop(handle)

   END SUBROUTINE mp2_ri_gpw_compute_in

! **************************************************************************************************
!> \brief Encapsulate building of intermediate matrices matrix_ia_jnu(_beta
!>         matrix_ia_jb(_beta),fm_BIb_jb(_beta),matrix_in_jnu(for G0W0) and
!>         fm_BIb_all(for G0W0)
!> \param matrix_ia_jnu ...
!> \param matrix_ia_jb ...
!> \param mo_coeff_templ ...
!> \param size_1 ...
!> \param size_2 ...
!> \param fm_BIb_jb ...
!> \param matrix_name_2 ...
!> \param max_row_col_local ...
!> \param nfullrows_total ...
!> \param nfullcols_total ...
!> \param nrow_local ...
!> \param ncol_local ...
!> \param blacs_env_sub ...
!> \param para_env_sub ...
!> \param local_col_row_info ...
!> \author Jan Wilhelm
! **************************************************************************************************
   SUBROUTINE create_intermediate_matrices(matrix_ia_jnu, matrix_ia_jb, mo_coeff_templ, size_1, size_2, &
                                           fm_BIb_jb, matrix_name_2, max_row_col_local, &
                                           nfullrows_total, nfullcols_total, nrow_local, &
                                           ncol_local, blacs_env_sub, &
                                           para_env_sub, local_col_row_info)

      TYPE(cp_dbcsr_type)                                :: matrix_ia_jnu, matrix_ia_jb, &
                                                            mo_coeff_templ
      INTEGER                                            :: size_1, size_2
      TYPE(cp_fm_type), POINTER                          :: fm_BIb_jb
      CHARACTER(LEN=*)                                   :: matrix_name_2
      INTEGER                                            :: max_row_col_local, nfullrows_total, &
                                                            nfullcols_total, nrow_local, ncol_local
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env_sub
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_col_row_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_intermediate_matrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct

      CALL timeset(routineN, handle)

      ! initialize and create the matrix (K|jnu)
      CALL cp_dbcsr_init(matrix_ia_jnu)
      CALL cp_dbcsr_create(matrix_ia_jnu, template=mo_coeff_templ)

      ! Allocate Sparse matrices: (K|jb)
      CALL cp_dbcsr_init(matrix_ia_jb)
      CALL cp_dbcsr_m_by_n_from_template(matrix_ia_jb, template=mo_coeff_templ, m=size_2, n=size_1, &
                                         sym=dbcsr_type_no_symmetry)

      ! set all to zero in such a way that the memory is actually allocated
      CALL cp_dbcsr_set(matrix_ia_jnu, 0.0_dp)
      CALL cp_dbcsr_set(matrix_ia_jb, 0.0_dp)

      ! create the analogous of matrix_ia_jb in fm type
      NULLIFY (fm_BIb_jb)
      NULLIFY (fm_struct)
      CALL cp_dbcsr_get_info(matrix_ia_jb, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env_sub, nrow_global=nfullrows_total, &
                               ncol_global=nfullcols_total, para_env=para_env_sub)
      CALL cp_fm_create(fm_BIb_jb, fm_struct, name=matrix_name_2)

      CALL copy_dbcsr_to_fm(matrix_ia_jb, fm_BIb_jb)
      CALL cp_fm_struct_release(fm_struct)

      CALL cp_fm_get_info(matrix=fm_BIb_jb, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)

      max_row_col_local = MAX(nrow_local, ncol_local)
      CALL mp_max(max_row_col_local, para_env_sub%group)

      ALLOCATE (local_col_row_info(0:max_row_col_local, 2))
      local_col_row_info = 0
      ! 0,1 nrows
      local_col_row_info(0, 1) = nrow_local
      local_col_row_info(1:nrow_local, 1) = row_indices(1:nrow_local)
      ! 0,2 ncols
      local_col_row_info(0, 2) = ncol_local
      local_col_row_info(1:ncol_local, 2) = col_indices(1:ncol_local)

      CALL timestop(handle)

   END SUBROUTINE create_intermediate_matrices

! **************************************************************************************************
!> \brief Encapsulate building of parallelization arrays
!> \param para_env_sub ...
!> \param ends_B ...
!> \param sizes_B ...
!> \param starts_B ...
!> \param my_B_start ...
!> \param my_B_end ...
!> \param my_B_size ...
!> \param size_1 ...
!> \author Jan Wilhelm
! **************************************************************************************************
   SUBROUTINE create_parallelization_arrays(para_env_sub, ends_B, sizes_B, starts_B, &
                                            my_B_start, my_B_end, my_B_size, size_1)

      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_B, sizes_B, starts_B
      INTEGER                                            :: my_B_start, my_B_end, my_B_size, size_1

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_parallelization_arrays', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iproc, itmp(2)

      CALL timeset(routineN, handle)

      ! divide the b states in the sub_group in such a way to create
      ! b_start and b_end for each proc inside the sub_group
      ALLOCATE (sizes_B(0:para_env_sub%num_pe-1))
      sizes_B = 0
      ALLOCATE (starts_B(0:para_env_sub%num_pe-1))
      starts_B = 0
      ALLOCATE (ends_B(0:para_env_sub%num_pe-1))
      ends_B = 0

      DO iproc = 0, para_env_sub%num_pe-1
         itmp = get_limit(size_1, para_env_sub%num_pe, iproc)
         !IF (itmp(2) < itmp(1)) THEN
         !   itmp(2) = itmp(1)
         !ENDIF
         starts_B(iproc) = itmp(1)
         ends_B(iproc) = itmp(2)
         sizes_B(iproc) = itmp(2)-itmp(1)+1
      END DO

      my_B_start = starts_B(para_env_sub%mepos)
      my_B_end = ends_B(para_env_sub%mepos)
      my_B_size = sizes_B(para_env_sub%mepos)

      CALL timestop(handle)

   END SUBROUTINE create_parallelization_arrays

! **************************************************************************************************
!> \brief Encapsulate ERI postprocessing: AO to MO transformation and store in B matrix.
!> \param para_env ...
!> \param mat_munu ...
!> \param matrix_ia_jnu ...
!> \param matrix_ia_jb ...
!> \param fm_BIb_jb ...
!> \param BIb_jb ...
!> \param mo_coeff_o ...
!> \param mo_coeff_v ...
!> \param eps_filter ...
!> \param max_row_col_local ...
!> \param sub_proc_map ...
!> \param local_col_row_info ...
!> \param my_B_end ...
!> \param my_B_start ...
!> \param descr ...
! **************************************************************************************************
   SUBROUTINE ao_to_mo_and_store_B(para_env, mat_munu, matrix_ia_jnu, matrix_ia_jb, fm_BIb_jb, BIb_jb, &
                                   mo_coeff_o, mo_coeff_v, eps_filter, &
                                   max_row_col_local, sub_proc_map, local_col_row_info, &
                                   my_B_end, my_B_start, descr)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(cp_dbcsr_type)                                :: matrix_ia_jnu, matrix_ia_jb
      TYPE(cp_fm_type), POINTER                          :: fm_BIb_jb
      REAL(KIND=dp), DIMENSION(:, :)                     :: BIb_jb
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_o, mo_coeff_v
      REAL(KIND=dp)                                      :: eps_filter
      INTEGER                                            :: max_row_col_local
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sub_proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_col_row_info
      INTEGER                                            :: my_B_end, my_B_start
      CHARACTER(LEN=*)                                   :: descr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ao_to_mo_and_store_B'

      INTEGER                                            :: handle1, handle2

      CALL timeset(routineN//"_mult_"//descr, handle1)

      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mat_munu%matrix, mo_coeff_o, &
                             0.0_dp, matrix_ia_jnu, filter_eps=eps_filter)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, matrix_ia_jnu, mo_coeff_v, &
                             0.0_dp, matrix_ia_jb, filter_eps=eps_filter)
      CALL timestop(handle1)

      CALL timeset(routineN//"_E_Ex_"//descr, handle2)
      CALL copy_dbcsr_to_fm(matrix_ia_jb, fm_BIb_jb)

      CALL grep_my_integrals(para_env, fm_BIb_jb, BIb_jb, max_row_col_local, &
                             sub_proc_map, local_col_row_info, &
                             my_B_end, my_B_start)

      CALL timestop(handle2)
   END SUBROUTINE ao_to_mo_and_store_B

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param eri_method ...
!> \param eri_param ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param para_env_L ...
!> \param mp2_memory ...
!> \param fm_matrix_L ...
!> \param ngroup ...
!> \param color_sub ...
!> \param dimen_RI ...
!> \param mo_coeff ...
!> \param dft_control ...
!> \param psi_L ...
!> \param rho_r ...
!> \param rho_g ...
!> \param pot_g ...
!> \param pw_env_sub ...
!> \param poisson_env ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
!> \param my_group_L_end ...
!> \param sab_orb_sub ...
!> \param sizes_array ...
!> \param starts_array ...
!> \param ends_array ...
!> \param calc_PQ_cond_num ...
!> \param cond_num ...
!> \param num_small_eigen ...
!> \param auxbas_pw_pool ...
!> \param do_mult_2c ...
!> \param ri_metric ...
!> \param fm_matrix_L_RI_metric ...
!> \param do_im_time ...
! **************************************************************************************************
   SUBROUTINE calculate_Lmin1(qs_env, eri_method, eri_param, para_env, para_env_sub, para_env_L, mp2_memory, &
                              fm_matrix_L, ngroup, color_sub, dimen_RI, &
                              mo_coeff, dft_control, psi_L, rho_r, rho_g, pot_g, pw_env_sub, poisson_env, &
                              my_group_L_size, my_group_L_start, my_group_L_end, sab_orb_sub, &
                              sizes_array, starts_array, ends_array, calc_PQ_cond_num, cond_num, &
                              num_small_eigen, auxbas_pw_pool, do_mult_2c, &
                              ri_metric, fm_matrix_L_RI_metric, do_im_time)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER                                            :: eri_method
      TYPE(eri_mme_param), POINTER                       :: eri_param
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub, para_env_L
      REAL(KIND=dp)                                      :: mp2_memory
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_L
      INTEGER                                            :: ngroup, color_sub, dimen_RI
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type)                                    :: psi_L, rho_r, rho_g, pot_g
      TYPE(pw_env_type), POINTER                         :: pw_env_sub
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      INTEGER                                            :: my_group_L_size, my_group_L_start, &
                                                            my_group_L_end
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sizes_array, starts_array, ends_array
      LOGICAL                                            :: calc_PQ_cond_num
      REAL(KIND=dp)                                      :: cond_num
      INTEGER                                            :: num_small_eigen
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      LOGICAL                                            :: do_mult_2c
      INTEGER                                            :: ri_metric
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_L_RI_metric
      LOGICAL                                            :: do_im_time

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_Lmin1', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: do_inversion
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_S_inv_work

! we have to invert the Coulomb matrix V_PQ only when doing RI with the Coulomb metric

      IF (ri_metric == ri_coulomb) THEN
         do_inversion = .TRUE.
      ELSE
         do_inversion = .FALSE.
      END IF

      ! calculate L^{-1} or L (for overlap and trunc. Coulomb metric)
      CALL compute_2c_in_and_cholesky(qs_env, eri_method, eri_param, para_env, para_env_sub, para_env_L, mp2_memory, &
                                      fm_matrix_L, ngroup, color_sub, dimen_RI, &
                                      mo_coeff, dft_control, psi_L, rho_r, rho_g, pot_g, pw_env_sub, poisson_env, &
                                      my_group_L_size, my_group_L_start, my_group_L_end, sab_orb_sub, &
                                      sizes_array, starts_array, ends_array, calc_PQ_cond_num, cond_num, &
                                      num_small_eigen, auxbas_pw_pool, do_mult_2c, do_inversion)

      IF (ri_metric /= ri_coulomb) THEN

         CALL cp_para_env_release(para_env_L)
         DEALLOCATE (sizes_array, starts_array, ends_array)

         ! Calculate Chol. decomp. of matrix of RI operator, store it in fm_matrix_L_RI_metric
         CALL compute_2c_in_and_cholesky(qs_env, eri_method, eri_param, para_env, para_env_sub, para_env_L, mp2_memory, &
                                         fm_matrix_L_RI_metric, ngroup, color_sub, dimen_RI, &
                                         mo_coeff, dft_control, psi_L, rho_r, rho_g, pot_g, pw_env_sub, poisson_env, &
                                         my_group_L_size, my_group_L_start, my_group_L_end, sab_orb_sub, &
                                         sizes_array, starts_array, ends_array, calc_PQ_cond_num, cond_num, &
                                         num_small_eigen, auxbas_pw_pool, do_mult_2c=.FALSE., do_inversion=.TRUE., &
                                         external_matrix_struc=.TRUE., fm_matrix_L_extern=fm_matrix_L, &
                                         ri_metric=ri_metric)

         NULLIFY (fm_matrix_S_inv_work)
         CALL cp_fm_create(fm_matrix_S_inv_work, fm_matrix_L_RI_metric%matrix_struct)
         CALL cp_fm_set_all(fm_matrix_S_inv_work, 0.0_dp)

         CALL cp_gemm('N', 'T', dimen_RI, dimen_RI, dimen_RI, 1.0_dp, fm_matrix_L_RI_metric, fm_matrix_L_RI_metric, &
                      0.0_dp, fm_matrix_S_inv_work)

         ! L (=fm_matrix_L_RI_metric) = S^(-1)*chol_dec(V), V: matrix of RI operator
         CALL cp_gemm('T', 'T', dimen_RI, dimen_RI, dimen_RI, 1.0_dp, fm_matrix_S_inv_work, fm_matrix_L, &
                      0.0_dp, fm_matrix_L_RI_metric)

         IF (do_im_time) THEN

            ! set L to identity matrix to compute the 3-center matrix elements (alpha beta | aux) with
            ! a single auxiliary basis function aux
            CALL set_identity(fm_matrix_L)

         ELSE

            CALL cp_fm_to_fm(fm_matrix_L_RI_metric, fm_matrix_L)
            CALL cp_fm_release(fm_matrix_L_RI_metric)

         END IF

         ! clean the S_inv matrix
         CALL cp_fm_release(fm_matrix_S_inv_work)

      END IF

   END SUBROUTINE calculate_Lmin1

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param eri_method ...
!> \param eri_param ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param para_env_L ...
!> \param mp2_memory ...
!> \param fm_matrix_L ...
!> \param ngroup ...
!> \param color_sub ...
!> \param dimen_RI ...
!> \param mo_coeff ...
!> \param dft_control ...
!> \param psi_L ...
!> \param rho_r ...
!> \param rho_g ...
!> \param pot_g ...
!> \param pw_env_sub ...
!> \param poisson_env ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
!> \param my_group_L_end ...
!> \param sab_orb_sub ...
!> \param sizes_array ...
!> \param starts_array ...
!> \param ends_array ...
!> \param calc_PQ_cond_num ...
!> \param cond_num ...
!> \param num_small_eigen ...
!> \param auxbas_pw_pool ...
!> \param do_mult_2c ...
!> \param do_inversion ...
!> \param external_matrix_struc ...
!> \param fm_matrix_L_extern ...
!> \param ri_metric ...
! **************************************************************************************************
   SUBROUTINE compute_2c_in_and_cholesky(qs_env, eri_method, eri_param, para_env, para_env_sub, para_env_L, mp2_memory, &
                                         fm_matrix_L, ngroup, color_sub, dimen_RI, &
                                         mo_coeff, dft_control, psi_L, rho_r, rho_g, pot_g, pw_env_sub, poisson_env, &
                                         my_group_L_size, my_group_L_start, my_group_L_end, sab_orb_sub, &
                                         sizes_array, starts_array, ends_array, calc_PQ_cond_num, cond_num, &
                                         num_small_eigen, auxbas_pw_pool, do_mult_2c, do_inversion, &
                                         external_matrix_struc, fm_matrix_L_extern, ri_metric)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER                                            :: eri_method
      TYPE(eri_mme_param), POINTER                       :: eri_param
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub, para_env_L
      REAL(KIND=dp)                                      :: mp2_memory
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_L
      INTEGER                                            :: ngroup, color_sub, dimen_RI
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type)                                    :: psi_L, rho_r, rho_g, pot_g
      TYPE(pw_env_type), POINTER                         :: pw_env_sub
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      INTEGER                                            :: my_group_L_size, my_group_L_start, &
                                                            my_group_L_end
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sizes_array, starts_array, ends_array
      LOGICAL                                            :: calc_PQ_cond_num
      REAL(KIND=dp)                                      :: cond_num
      INTEGER                                            :: num_small_eigen
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      LOGICAL                                            :: do_mult_2c, do_inversion
      LOGICAL, OPTIONAL                                  :: external_matrix_struc
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_matrix_L_extern
      INTEGER, OPTIONAL                                  :: ri_metric

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_2c_in_and_cholesky', &
         routineP = moduleN//':'//routineN

      INTEGER :: best_group_size, color_exc, color_L, comm_exc, comm_exchange, comm_L, dir, &
         group_size, handle, handle2, handle3, i, i_bf, i_counter, i_counter_diff_bf, &
         i_counter_diff_bf_at_any_atom, i_counter_diff_sets, i_global, iatom, igrid_level, igroup, &
         iiB, ikind, info_chol, inode, ipgf, iproc, iset, itmp(2), j, j_global, j_ref_atom, &
         j_ref_bf, j_ref_kind, j_ref_set, jatom, jjB, jkind, jset, lb(3), LLL, location(3), m, &
         my_group_L_q_end, my_group_L_q_size, my_group_L_q_start, my_group_pot_end, &
         my_group_pot_size, my_group_pot_start, my_ri_metric, na1, na2, natom, ncoa, ncoa_a, ncoa_b
      INTEGER :: ncol_local, nkind, nrow_local, nseta, nsetb, nsgf, offset, offsetb, proc_receive, &
         proc_receive_static, proc_send, proc_send_static, proc_shift, rec_L_end, rec_L_size, &
         rec_L_start, sgfa, sgfb, strat_group_size, sub_sub_color, tp(3), ub(3)
      INTEGER, ALLOCATABLE, DIMENSION(:) :: diffbf_atom_arbitr_atom, diffbf_noinbas, &
         diffbfatanyatom_noinbas, ends_array_pot, kind_of, proc_map, sizes_array_pot, &
         starts_array_pot, sub_ends_array, sub_sizes_array, sub_starts_array
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: atom_set_afo_diffbf, &
                                                            atom_set_afo_diffbf_atanyatom, &
                                                            diffsets_atom_set_bf, ref_matrix
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: naolimits, ref_at_se_sgf_ofs
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, la_max, la_min, lb_max, &
                                                            lb_min, npgfa, npgfb, nsgfa, nsgfb, &
                                                            row_indices
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL                                            :: check_i, check_j, map_it_here, &
                                                            my_external_matrix_struc
      REAL(KIND=dp)                                      :: min_mem_for_QK, pair_energy, rab2, rac2, &
                                                            scaling, zet_tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: egen_L, f, wf_vector
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: I_ab, I_tmp, L_external_col, &
                                                            L_local_col, L_local_col_q, vac
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: v
      REAL(KIND=dp), DIMENSION(1)                        :: exp_q
      REAL(KIND=dp), DIMENSION(3)                        :: ra, rab, rac
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: I_tmp2, rpgfa, rpgfb, sphi_a, sphi_b, &
                                                            zeta, zetb
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env_L
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_L_diag
      TYPE(cp_para_env_type), POINTER                    :: para_env_exc, para_env_exchange
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_p_type), ALLOCATABLE, DIMENSION(:)         :: rho_r_array
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_v

      CALL timeset(routineN, handle)

      ! default: Coulomb metric
      my_ri_metric = ri_coulomb
      IF (PRESENT(ri_metric)) my_ri_metric = ri_metric

      my_external_matrix_struc = .FALSE.
      IF (PRESENT(external_matrix_struc)) THEN
         my_external_matrix_struc = external_matrix_struc
      END IF

      ! get stuff
      CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, &
                      cell=cell, molecule_set=molecule_set, particle_set=particle_set, &
                      molecule_kind_set=molecule_kind_set)

      ! blacs_env => qs_env%blacs_env

      nkind = SIZE(qs_kind_set)
      natom = SIZE(particle_set)

      ALLOCATE (kind_of(natom))

      CALL get_atomic_kind_set(atomic_kind_set, kind_of=kind_of)

      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=basis_set_a, basis_type="RI_AUX")
         IF(.NOT.(ASSOCIATED(basis_set_a)))CALL cp__a("mp2_ri_gpw.F",1198)
      END DO

      dimen_RI = 0
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         CALL get_qs_kind(qs_kind=qs_kind_set(ikind), nsgf=nsgf, basis_type="RI_AUX")
         dimen_RI = dimen_RI+nsgf
      END DO

      ! calculate which rows of L^{-1} to have
      ALLOCATE (sizes_array(0:ngroup-1))
      sizes_array = 0
      ALLOCATE (starts_array(0:ngroup-1))
      starts_array = 0
      ALLOCATE (ends_array(0:ngroup-1))
      ends_array = 0

      DO igroup = 0, ngroup-1
         itmp = get_limit(dimen_RI, ngroup, igroup)
         starts_array(igroup) = itmp(1)
         ends_array(igroup) = itmp(2)
         sizes_array(igroup) = itmp(2)-itmp(1)+1
      ENDDO

      my_group_L_size = sizes_array(color_sub)
      my_group_L_start = starts_array(color_sub)
      my_group_L_end = ends_array(color_sub)

      CALL timeset(routineN//"_loop_lm", handle2)

      IF (do_mult_2c) THEN

         ! write arrays with number(=label) of every basis function, its quantum number (l,m)
         ! and the reference to the first basis function with (l,m)
         ALLOCATE (ref_matrix(dimen_RI, 10))

         ! naolimits(iatom,iset,1/2)=begin/end reference entries in L_local_q
         ALLOCATE (naolimits(natom, 100, 2))

         ! array with reference (atom,set) for arbitrary (atom,set);
         ALLOCATE (ref_at_se_sgf_ofs(natom, 100, 4))

         i_counter = 0
         i_counter_diff_bf = 0
         i_counter_diff_sets = 0
         i_counter_diff_bf_at_any_atom = 0

         ! 1) build up reference matrices to get the reference q_alpha for every basis function alpha

         offset = 0

         DO iatom = 1, natom
            ikind = kind_of(iatom)
            CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=basis_set_a, basis_type="RI_AUX")

            la_max => basis_set_a%lmax
            la_min => basis_set_a%lmin
            npgfa => basis_set_a%npgf
            nseta = basis_set_a%nset
            nsgfa => basis_set_a%nsgf_set
            zeta => basis_set_a%zet
            first_sgfa => basis_set_a%first_sgf

            DO iset = 1, nseta

               DO ipgf = 1, npgfa(iset)

                  !pseudo-m quantum number runs from 1 to 2l+1
                  DO m = 1, nsgfa(iset)
                     i_counter = i_counter+1
                     !first entry: l; only in case of la_max(iset)==la_min(iset) the shortcut will work, see IF below
                     ref_matrix(i_counter, 1) = la_max(iset)
                     !second entry: pseudo-m quantum number runs from 1 to 2l+1
                     ref_matrix(i_counter, 2) = m
                     ref_matrix(i_counter, 3) = iatom
                     ref_matrix(i_counter, 4) = iset
                     ref_matrix(i_counter, 5) = ipgf
                     ref_matrix(i_counter, 8) = first_sgfa(1, iset)

                     !Check, whether (l,m) combination already exists
                     IF (i_counter >= 2) THEN
                        DO i_bf = 1, i_counter-1
                           ! 6th row of ref_matrix contains number of first basis function with identical (l,m)
                           ! for set with la_min=/la_max at the same atom
                           IF (ref_matrix(i_bf, 1) == la_max(iset) .AND. ref_matrix(i_bf, 2) == m .AND. &
                               la_max(iset) == la_min(iset) .AND. ref_matrix(i_bf, 3) == iatom) THEN
                              ref_matrix(i_counter, 6) = i_bf

                              ! we are done with checking the basis functions in DO when a (l,m) match was found.
                              EXIT

                           END IF
                           ! 0 if no basis function has this (l,m) combination
                           ref_matrix(i_counter, 6) = 0

                        END DO !loop already existing basis functions

                        DO i_bf = 1, i_counter-1
                           ! 6th row of ref_matrix contains number of first basis function with identical (l,m)
                           ! for set with la_min=/la_max AT ANY ATOM (!!DIFFERENCE!!)
                           IF (ref_matrix(i_bf, 1) == la_max(iset) .AND. ref_matrix(i_bf, 2) == m .AND. &
                               la_max(iset) == la_min(iset)) THEN
                              ref_matrix(i_counter, 9) = i_bf

                              ! we are done with checking the basis functions in DO when a (l,m) match was found.
                              EXIT

                           END IF
                           ! 0 if no basis function has this (l,m) combination
                           ref_matrix(i_counter, 9) = 0

                        END DO !loop already existing basis functions

                     ELSE
                        ref_matrix(1, 6) = 0
                        ref_matrix(1, 9) = 0

                     END IF
                     ! for new (l,m) combination increase i_counter_diff_bf by 1 and set reference to itself
                     IF (ref_matrix(i_counter, 6) == 0) THEN
                        i_counter_diff_bf = i_counter_diff_bf+1
                        ref_matrix(i_counter, 7) = i_counter_diff_bf
                        ! for already existing (l,m) combination, 7th entry refers to number/label of first (l,m) combination
                     ELSE
                        ref_matrix(i_counter, 7) = &
                           ref_matrix(ref_matrix(i_counter, 6), 7)
                     END IF

                     ! for new (l,m) at any atom combination increase i_counter_diff_bf_at_any_atom by 1
                     !  and set reference to itself
                     IF (ref_matrix(i_counter, 9) == 0) THEN
                        i_counter_diff_bf_at_any_atom = i_counter_diff_bf_at_any_atom+1
                        ref_matrix(i_counter, 10) = i_counter_diff_bf_at_any_atom
                        ! for already existing (l,m) combination, 7th entry refers to number/label of first (l,m) combination
                     ELSE
                        ref_matrix(i_counter, 10) = &
                           ref_matrix(ref_matrix(i_counter, 9), 10)
                     END IF

                  END DO !m

               END DO !prim. gf.

               IF (ref_matrix(i_counter, 6) == 0) THEN
                  i_counter_diff_sets = i_counter_diff_sets+1
                  ref_at_se_sgf_ofs(iatom, iset, 1) = iatom
                  ref_at_se_sgf_ofs(iatom, iset, 2) = iset
                  ref_at_se_sgf_ofs(iatom, iset, 3) = first_sgfa(1, iset)

               ELSE
                  ref_at_se_sgf_ofs(iatom, iset, 1) = ref_matrix(ref_matrix(i_counter, 6), 3)
                  ref_at_se_sgf_ofs(iatom, iset, 2) = ref_matrix(ref_matrix(i_counter, 6), 4)
                  ref_at_se_sgf_ofs(iatom, iset, 3) = ref_matrix(ref_matrix(i_counter, 6), 8)
               END IF

               naolimits(iatom, iset, 1) = ref_matrix(i_counter, 7)-nsgfa(iset)+1
               naolimits(iatom, iset, 2) = ref_matrix(i_counter, 7)

               ref_at_se_sgf_ofs(iatom, iset, 4) = offset
               offset = offset+nsgfa(iset)

            END DO !sets

         END DO !atoms

         ! array with number of q_a basis function at an atom
         ALLOCATE (diffbf_noinbas(i_counter_diff_bf))

         ! array with number of a basis function with first unique (l,m) combination
         ALLOCATE (diffbfatanyatom_noinbas(i_counter_diff_bf_at_any_atom))

         ! array with number of q_a (at arbitrary atom) with q_a at specific atom
         ALLOCATE (diffbf_atom_arbitr_atom(i_counter_diff_bf))

         ! array with atom and set number for every q_a bf
         ALLOCATE (atom_set_afo_diffbf(i_counter_diff_bf, 2))

         ! array with atom and set number for every q_a bf
         ALLOCATE (atom_set_afo_diffbf_atanyatom(i_counter_diff_bf, 2))

         j = 1
         DO i = 1, dimen_RI
            IF (ref_matrix(i, 9) == 0) THEN
               diffbfatanyatom_noinbas(j) = i
               atom_set_afo_diffbf_atanyatom(j, 1) = ref_matrix(i, 3)
               atom_set_afo_diffbf_atanyatom(j, 2) = ref_matrix(i, 4)
               j = j+1
            END IF
         END DO

         j = 1
         DO i = 1, dimen_RI
            IF (ref_matrix(i, 6) == 0) THEN
               diffbf_noinbas(j) = i
               atom_set_afo_diffbf(j, 1) = ref_matrix(i, 3)
               atom_set_afo_diffbf(j, 2) = ref_matrix(i, 4)
               diffbf_atom_arbitr_atom(j) = ref_matrix(i, 10)
               j = j+1
            END IF
         END DO

         ! array with number of first set with quantum number l
         ALLOCATE (diffsets_atom_set_bf(i_counter_diff_sets, 3))

         ! first set with new quantum number l is on atom 1, set 1
         diffsets_atom_set_bf(1, 1) = 1
         diffsets_atom_set_bf(1, 2) = 1
         diffsets_atom_set_bf(1, 3) = 1

         j = 2
         DO i = 2, dimen_RI
            IF (ref_matrix(i, 6) == 0 .AND. (diffsets_atom_set_bf(j-1, 1) /= ref_matrix(i, 3) .OR. &
                                             diffsets_atom_set_bf(j-1, 2) /= ref_matrix(i, 4))) THEN
               diffsets_atom_set_bf(j, 1) = ref_matrix(i, 3)
               diffsets_atom_set_bf(j, 2) = ref_matrix(i, 4)
               diffsets_atom_set_bf(j, 3) = ref_matrix(i, 7)
               j = j+1
            END IF
         END DO

         ! For parallelization: Check, which elements of L_local_col_q
         ! are needed for setting up L_local_q

         my_group_L_q_start = naolimits(ref_matrix(my_group_L_start, 3), 1, 1)
         IF (ref_matrix(my_group_L_end, 3) /= natom) THEN
            my_group_L_q_end = naolimits(ref_matrix(my_group_L_end, 3)+1, 1, 1)-1
         ELSE
            nkind = kind_of(natom)
            CALL get_qs_kind(qs_kind=qs_kind_set(nkind), basis_set=basis_set_a, basis_type="RI_AUX")
            nseta = basis_set_a%nset
            my_group_L_q_end = naolimits(ref_matrix(my_group_L_end, 3), nseta, 2)
         END IF
         my_group_L_q_size = my_group_L_q_end-my_group_L_q_start+1

         ! calculate which potentials to calculate
         ALLOCATE (sizes_array_pot(0:ngroup-1))
         sizes_array_pot = 0
         ALLOCATE (starts_array_pot(0:ngroup-1))
         starts_array_pot = 0
         ALLOCATE (ends_array_pot(0:ngroup-1))
         ends_array_pot = 0

         DO igroup = 0, ngroup-1
            itmp = get_limit(i_counter_diff_bf_at_any_atom, ngroup, igroup)
            starts_array_pot(igroup) = itmp(1)
            ends_array_pot(igroup) = itmp(2)
            sizes_array_pot(igroup) = itmp(2)-itmp(1)+1
         ENDDO

         my_group_pot_size = sizes_array_pot(color_sub)
         my_group_pot_start = starts_array_pot(color_sub)
         my_group_pot_end = ends_array_pot(color_sub)

         DEALLOCATE (ref_matrix)

         ! 2.0) calculate potential (q_alpha|_p for first (l,m) combination at an atom
         i_counter = 0
         exp_q = 0.5_dp

         ALLOCATE (rho_r_array(i_counter_diff_bf_at_any_atom))

         ALLOCATE (wf_vector(dimen_RI))

         DO j = my_group_pot_start, my_group_pot_end
            wf_vector = 0.0_dp
            ! only take one basis function unique l,m-combination at any atom
            wf_vector(diffbfatanyatom_noinbas(j)) = 1.0_dp

            ! set exponent of aux basis set temporarily to exp_q since reference wavefunction
            ! q_a/b should have exponent exp_q. Then integration (q_a|q_b) is numerically
            ! feasible in reciprocal space
            jatom = atom_set_afo_diffbf_atanyatom(j, 1)
            jset = atom_set_afo_diffbf_atanyatom(j, 2)
            jkind = kind_of(jatom)
            CALL get_qs_kind(qs_kind=qs_kind_set(jkind), basis_set=basis_set_b, basis_type="RI_AUX")
            zetb => basis_set_b%zet
            lb_max => basis_set_b%lmax

            zet_tmp = zetb(1, jset)
            zetb(1, jset) = exp_q(1)

            ! pseudo psi_L
            CALL calculate_wavefunction(mo_coeff, 1, psi_L, rho_g, atomic_kind_set, &
                                        qs_kind_set, cell, dft_control, particle_set, pw_env_sub, &
                                        basis_type="RI_AUX", &
                                        external_vector=wf_vector)

            ! set back the exponent
            zetb(1, jset) = zet_tmp

            rho_r%pw%cr3d = psi_L%pw%cr3d
            CALL pw_transfer(rho_r%pw, rho_g%pw)
            CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
            CALL pw_transfer(pot_g%pw, rho_r%pw)
            CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)

            ! Writing array of potential copied from xc/xc_rho_set_types.F

            NULLIFY (rho_r_array(j)%pw)

            rho_r_array(j)%pw => rho_r%pw

            CALL pw_pool_create_pw(auxbas_pw_pool, rho_r_array(j)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_copy(rho_r%pw, rho_r_array(j)%pw)

         END DO

         ! Fill pw_grid and zero-components cr3d, aso. into rho_r_array for parallization
         ! If no potential was calculated because of my_group_pot_start>my_group_pot_end,
         ! calculate potential of my_group_pot_start-component to get a pw_grid
         IF (my_group_pot_start > my_group_pot_end) THEN
            wf_vector = 0.0_dp
            ! pseudo psi_L
            CALL calculate_wavefunction(mo_coeff, 1, psi_L, rho_g, atomic_kind_set, &
                                        qs_kind_set, cell, dft_control, particle_set, pw_env_sub, &
                                        basis_type="RI_AUX", &
                                        external_vector=wf_vector)

            rho_r%pw%cr3d = psi_L%pw%cr3d
            CALL pw_transfer(rho_r%pw, rho_g%pw)
            CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
            CALL pw_transfer(pot_g%pw, rho_r%pw)
            CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)

         END IF

         DEALLOCATE (wf_vector)

         ! Fill pw_grid and 0.0_dp potential in every component of rho_r_array
         DO j = 1, i_counter_diff_bf_at_any_atom
            IF (j < my_group_pot_start .OR. j > my_group_pot_end) THEN

               NULLIFY (rho_r_array(j)%pw)
               rho_r%pw%cr3d = 0.0_dp
               rho_r_array(j)%pw => rho_r%pw

               CALL pw_pool_create_pw(auxbas_pw_pool, rho_r_array(j)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_copy(rho_r%pw, rho_r_array(j)%pw)

            END IF
         END DO

         color_exc = para_env_sub%mepos
         CALL mp_comm_split_direct(para_env%group, comm_exc, color_exc)
         NULLIFY (para_env_exc)
         CALL cp_para_env_create(para_env_exc, comm_exc)

         ! collect all calculated potentials
         DO j = 1, i_counter_diff_bf_at_any_atom
            CALL mp_sum(rho_r_array(j)%pw%cr3d, para_env_exc%group)
         END DO

         CALL cp_para_env_release(para_env_exc)

         ALLOCATE (L_local_col_q(i_counter_diff_bf, my_group_L_q_size))

         L_local_col_q = 0.0_dp

         i_counter = 0

         ! 2.1) calculate (q_alpha|q_beta)_p for every (l_alpha,m_alpha), (l_beta,m_beta)
         !      at every atom. Use the potential (q_alpha|_p calculated in 2.0)

         DO j = my_group_L_q_start, my_group_L_q_end
            i_counter = i_counter+1

            jatom = atom_set_afo_diffbf(j, 1)
            jset = atom_set_afo_diffbf(j, 2)

            j_ref_bf = diffbf_atom_arbitr_atom(j)
            j_ref_atom = atom_set_afo_diffbf_atanyatom(j_ref_bf, 1)
            j_ref_set = atom_set_afo_diffbf_atanyatom(j_ref_bf, 2)

            j_ref_kind = kind_of(j_ref_atom)
            CALL get_qs_kind(qs_kind=qs_kind_set(j_ref_kind), basis_set=basis_set_b, basis_type="RI_AUX")
            zetb => basis_set_b%zet
            lb_max => basis_set_b%lmax

            NULLIFY (rs_v)
            NULLIFY (rs_descs)
            CALL pw_env_get(pw_env_sub, rs_descs=rs_descs, rs_grids=rs_v)
            DO i = 1, SIZE(rs_v)
               ! allocation
               CALL rs_grid_retain(rs_v(i)%rs_grid)
            END DO
            CALL potential_pw2rs(rs_v, rho_r_array(j_ref_bf), pw_env_sub)

            ! integrate the little bastards
            offset = 0
            ! only take one single set once for every (l,m) combination
            DO i = 1, i_counter_diff_sets
               iatom = diffsets_atom_set_bf(i, 1)
               iset = diffsets_atom_set_bf(i, 2)

               ikind = kind_of(iatom)
               CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=basis_set_a, basis_type="RI_AUX")

               first_sgfa => basis_set_a%first_sgf
               la_max => basis_set_a%lmax
               la_min => basis_set_a%lmin
               npgfa => basis_set_a%npgf
               nseta = basis_set_a%nset
               nsgfa => basis_set_a%nsgf_set
               rpgfa => basis_set_a%pgf_radius
               set_radius_a => basis_set_a%set_radius
               sphi_a => basis_set_a%sphi
               zeta => basis_set_a%zet

               ra(:) = pbc(particle_set(iatom)%r, cell)- &
                       pbc(particle_set(jatom)%r, cell)+ &
                       pbc(particle_set(j_ref_atom)%r, cell)
               rab = 0.0_dp
               rab2 = 0.0_dp

               ncoa = npgfa(iset)*ncoset(la_max(iset))
               sgfa = first_sgfa(1, iset)

               ALLOCATE (I_tmp2(ncoa, 1))
               I_tmp2 = 0.0_dp
               ALLOCATE (I_ab(nsgfa(iset), 1))
               I_ab = 0.0_dp

               igrid_level = gaussian_gridlevel(pw_env_sub%gridlevel_info, MINVAL(exp_q(:)))
               map_it_here = .FALSE.
               IF (.NOT. ALL(rs_v(igrid_level)%rs_grid%desc%perd == 1)) THEN
                  DO dir = 1, 3
                     ! bounds of local grid (i.e. removing the 'wings'), if periodic
                     tp(dir) = FLOOR(DOT_PRODUCT(cell%h_inv(dir, :), ra)*rs_v(igrid_level)%rs_grid%desc%npts(dir))
                     tp(dir) = MODULO(tp(dir), rs_v(igrid_level)%rs_grid%desc%npts(dir))
                     IF (rs_v(igrid_level)%rs_grid%desc%perd(dir) .NE. 1) THEN
                        lb(dir) = rs_v(igrid_level)%rs_grid%lb_local(dir)+rs_v(igrid_level)%rs_grid%desc%border
                        ub(dir) = rs_v(igrid_level)%rs_grid%ub_local(dir)-rs_v(igrid_level)%rs_grid%desc%border
                     ELSE
                        lb(dir) = rs_v(igrid_level)%rs_grid%lb_local(dir)
                        ub(dir) = rs_v(igrid_level)%rs_grid%ub_local(dir)
                     ENDIF
                     ! distributed grid, only map if it is local to the grid
                     location(dir) = tp(dir)+rs_v(igrid_level)%rs_grid%desc%lb(dir)
                  ENDDO
                  IF (lb(1) <= location(1) .AND. location(1) <= ub(1) .AND. &
                      lb(2) <= location(2) .AND. location(2) <= ub(2) .AND. &
                      lb(3) <= location(3) .AND. location(3) <= ub(3)) THEN
                     map_it_here = .TRUE.
                  ENDIF
               ELSE
                  ! not distributed, just a round-robin distribution over the full set of CPUs
                  IF (MODULO(offset, para_env_sub%num_pe) == para_env_sub%mepos) map_it_here = .TRUE.
               ENDIF

               offset = offset+nsgfa(iset)

               IF (map_it_here) THEN
                  DO ipgf = 1, npgfa(iset)
                     na1 = (ipgf-1)*ncoset(la_max(iset))+1
                     na2 = ipgf*ncoset(la_max(iset))
                     igrid_level = gaussian_gridlevel(pw_env_sub%gridlevel_info, exp_q(1))

                     CALL integrate_pgf_product_rspace(la_max=la_max(iset), zeta=exp_q(1)/2.0_dp, la_min=la_min(iset), &
                                                       lb_max=0, zetb=exp_q(1)/2.0_dp, lb_min=0, &
                                                       ra=ra, rab=rab, rab2=rab2, &
                                                       rsgrid=rs_v(igrid_level)%rs_grid, &
                                                       cell=cell, &
                                                       cube_info=pw_env_sub%cube_info(igrid_level), &
                                                       hab=I_tmp2, &
                                                       o1=na1-1, &
                                                       o2=0, &
                                                       map_consistent=.TRUE., &
                                                       eps_gvg_rspace=dft_control%qs_control%eps_gvg_rspace, &
                                                       calculate_forces=.FALSE.)
                  END DO

                  CALL dgemm("T", "N", nsgfa(iset), 1, ncoa, &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             I_tmp2(1, 1), SIZE(I_tmp2, 1), &
                             1.0_dp, I_ab(1, 1), SIZE(I_ab, 1))

                  ! integration result has to be scaled since from basis set differing exponent exp_q
                  ! causes different normalization
                  L_local_col_q(offset-nsgfa(iset)+1:offset, i_counter) = &
                     I_ab(1:nsgfa(iset), 1)* &
                     (exp_q(1)/zeta(1, iset))**(la_max(iset)*0.5_dp+3.0_dp/4.0_dp)* &
                     (exp_q(1)/zetb(1, j_ref_set))**(lb_max(j_ref_set)*0.5_dp+3.0_dp/4.0_dp)
               END IF

               DEALLOCATE (I_tmp2)
               DEALLOCATE (I_ab)

            END DO ! q_beta)-set loop

            ! deallocation
            DO i = 1, SIZE(rs_v)
               CALL rs_grid_release(rs_v(i)%rs_grid)
            END DO

         END DO !(q_alpha| loop

         ! post cleanup
         DO j = 1, i_counter_diff_bf_at_any_atom
            CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_r_array(j)%pw, &
                                      accept_non_compatible=.TRUE.)
         END DO

         DEALLOCATE (rho_r_array)

         DEALLOCATE (diffbf_noinbas)
         DEALLOCATE (diffbfatanyatom_noinbas)

         !  3) Distribute (q_alpha|q_beta)_p to L_local_col

         ALLOCATE (L_local_col(dimen_RI, my_group_L_size))
         L_local_col = 0.0_dp

         DO iatom = 1, natom
            ikind = kind_of(iatom)
            CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=basis_set_a, basis_type="RI_AUX")
            first_sgfa => basis_set_a%first_sgf
            la_max => basis_set_a%lmax
            la_min => basis_set_a%lmin
            npgfa => basis_set_a%npgf
            nseta = basis_set_a%nset
            nsgfa => basis_set_a%nsgf_set
            rpgfa => basis_set_a%pgf_radius
            set_radius_a => basis_set_a%set_radius
            sphi_a => basis_set_a%sphi
            zeta => basis_set_a%zet

            DO iset = 1, nseta

               offsetb = 0

               DO jatom = 1, natom
                  jkind = kind_of(jatom)
                  CALL get_qs_kind(qs_kind=qs_kind_set(jkind), basis_set=basis_set_b, basis_type="RI_AUX")
                  first_sgfb => basis_set_b%first_sgf
                  lb_max => basis_set_b%lmax
                  lb_min => basis_set_b%lmin
                  npgfb => basis_set_b%npgf
                  nsetb = basis_set_b%nset
                  nsgfb => basis_set_b%nsgf_set
                  rpgfb => basis_set_b%pgf_radius
                  set_radius_b => basis_set_b%set_radius
                  sphi_b => basis_set_b%sphi
                  zetb => basis_set_b%zet

                  DO jset = 1, nsetb

                     ! for parallelization: check, if jset is in paralleliz. range
                     IF (ref_at_se_sgf_ofs(jatom, jset, 4)+nsgfb(jset) >= my_group_L_start .AND. &
                         ref_at_se_sgf_ofs(jatom, jset, 4)+1 <= my_group_L_end) THEN

                        ! add contribution from (q_a|q_b)_p
                        scaling = 1.0_dp/(zeta(1, iset))**(la_max(iset)*0.5_dp+3.0_dp/4.0_dp)/ &
                                  (zetb(1, jset))**(lb_max(jset)*0.5_dp+3.0_dp/4.0_dp)* &
                                  exp_q(1)**(la_max(iset)*0.5_dp+3.0_dp/4.0_dp)* &
                                  exp_q(1)**(lb_max(jset)*0.5_dp+3.0_dp/4.0_dp)

                        DO i = 1, nsgfa(iset)
                           DO j = 1, nsgfb(jset)
                              ! for parallelization: check, if sgf_j is in paralleliz. range
                              IF (ref_at_se_sgf_ofs(jatom, jset, 4)+j >= my_group_L_start .AND. &
                                  ref_at_se_sgf_ofs(jatom, jset, 4)+j <= my_group_L_end) THEN

                                 ! Scale (q_a|q_b) by c_a*c_b, only implemented for uncontracted Gaussians and lmin=lmax
                                 L_local_col(ref_at_se_sgf_ofs(iatom, iset, 4)+i, &
                                             ref_at_se_sgf_ofs(jatom, jset, 4)+j-my_group_L_start+1) = &
                                    L_local_col_q(naolimits(iatom, iset, 1)+i-1, &
                                                  naolimits(jatom, jset, 1)+j-1-my_group_L_q_start+1)* &
                                    scaling

                              END IF
                           END DO
                        END DO

                        ! for s-functions alpha and beta, we need additional correction V_a(G=0)-V_q(G=0))*beta(G=0)
                        ! and V_b(G=0)-V_q(G=0))*alpha(G=0), only add once for every subgroup (for mepos==0)
                        IF (la_max(iset) == 0 .AND. lb_max(jset) == 0 .AND. para_env_sub%mepos == 0) THEN
                           L_local_col(ref_at_se_sgf_ofs(iatom, iset, 4)+1, &
                                       ref_at_se_sgf_ofs(jatom, jset, 4)-my_group_L_start+2) = &
                              L_local_col(ref_at_se_sgf_ofs(iatom, iset, 4)+1, &
                                          ref_at_se_sgf_ofs(jatom, jset, 4)-my_group_L_start+2)- &
                              Pi/(rho_r%pw%pw_grid%vol)*(4.0_dp*Pi**2.0_dp/(zeta(1, iset)*zetb(1, jset)))**0.75_dp* &
                              (2.0_dp/exp_q(1)-1.0_dp/zeta(1, iset)-1.0_dp/zetb(1, jset))
                        END IF

                     END IF

                  END DO !jset
               END DO !jatom
            END DO !iset
         END DO !iatom

         DEALLOCATE (naolimits)
         DEALLOCATE (L_local_col_q)

         ! End of calculation (q_alpha|q_beta) for every (l_alpha,m_alpha), (l_beta,m_beta) combination

         ! 4) Calculate (a|b)_s - (q_alpha|q_beta)_s and add them to (q_a|q_b)_p

         ! Loop over neighbor list
         CALL neighbor_list_iterator_create(nl_iterator, sab_orb_sub)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, &
                                   iatom=iatom, jatom=jatom, r=rac)

            ikind = kind_of(iatom)
            CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=basis_set_a, basis_type="RI_AUX")
            first_sgfa => basis_set_a%first_sgf
            la_max => basis_set_a%lmax
            la_min => basis_set_a%lmin
            npgfa => basis_set_a%npgf
            nseta = basis_set_a%nset
            nsgfa => basis_set_a%nsgf_set
            rpgfa => basis_set_a%pgf_radius
            set_radius_a => basis_set_a%set_radius
            sphi_a => basis_set_a%sphi
            zeta => basis_set_a%zet

            DO iset = 1, nseta

               jkind = kind_of(jatom)
               CALL get_qs_kind(qs_kind=qs_kind_set(jkind), basis_set=basis_set_b, basis_type="RI_AUX")
               first_sgfb => basis_set_b%first_sgf
               lb_max => basis_set_b%lmax
               lb_min => basis_set_b%lmin
               npgfb => basis_set_b%npgf
               nsetb = basis_set_b%nset
               nsgfb => basis_set_b%nsgf_set
               rpgfb => basis_set_b%pgf_radius
               set_radius_b => basis_set_b%set_radius
               sphi_b => basis_set_b%sphi
               zetb => basis_set_b%zet

               DO jset = 1, nsetb

                  ! for parallelization: check, if (jatom,jset) is in paralleliz. range
                  check_j = ref_at_se_sgf_ofs(jatom, jset, 4)+nsgfb(jset) >= my_group_L_start .AND. &
                            ref_at_se_sgf_ofs(jatom, jset, 4)+1 <= my_group_L_end

                  check_i = ref_at_se_sgf_ofs(iatom, iset, 4)+nsgfa(iset) >= my_group_L_start .AND. &
                            ref_at_se_sgf_ofs(iatom, iset, 4)+1 <= my_group_L_end .AND. &
                            iatom /= jatom

                  IF (check_j .OR. check_i) THEN

                     ncoa_a = npgfa(iset)*ncoset(la_max(iset))
                     ncoa_b = npgfb(jset)*ncoset(lb_max(jset))
                     sgfa = first_sgfa(1, iset)
                     sgfb = first_sgfb(1, jset)

                     ALLOCATE (I_tmp(ncoa_a, nsgfb(jset)))
                     ALLOCATE (I_tmp2(nsgfa(iset), nsgfb(jset)))

                     ALLOCATE (f(0:la_max(iset)+lb_max(jset)+2), &
                               v(ncoa_a, ncoa_b, la_max(iset)+lb_max(jset)+1), &
                               vac(ncoa_a, ncoa_b))
                     f = 0.0_dp
                     v = 0.0_dp
                     vac = 0.0_dp
                     I_tmp = 0.0_dp
                     I_tmp2 = 0.0_dp

                     rac2 = rac(1)*rac(1)+rac(2)*rac(2)+rac(3)*rac(3)

                     ! calculate spherical matrix elements (a|b)_s
                     CALL coulomb2_new(la_max=la_max(iset), npgfa=npgfa(iset), zeta=zeta(:, iset), &
                                       la_min=la_min(iset), &
                                       lc_max=lb_max(jset), npgfc=npgfb(jset), zetc=zetb(:, jset), &
                                       lc_min=lb_min(jset), rac=rac, &
                                       rac2=rac2, vac=vac, v=v, f=f)

                     ! trafo to spherical and add (a|b)_s to L_local_col
                     I_tmp(1:ncoa_a, 1:nsgfb(jset)) = MATMUL(vac(1:ncoa_a, 1:ncoa_b), sphi_b(1:ncoa_b, sgfb:(sgfb+nsgfb(jset)-1)))
                     I_tmp2(1:nsgfa(iset), 1:nsgfb(jset)) = MATMUL(TRANSPOSE(sphi_a(1:ncoa_a, sgfa:(sgfa+nsgfa(iset)-1))), &
                                                                   I_tmp(1:ncoa_a, 1:nsgfb(jset)))

                     f = 0.0_dp
                     v = 0.0_dp
                     vac = 0.0_dp

                     ! calculate -(q_a|q_b)_s
                     CALL coulomb2_new(la_max=la_max(iset), npgfa=npgfa(iset), zeta=exp_q(:), &
                                       la_min=la_min(iset), &
                                       lc_max=lb_max(jset), npgfc=npgfb(jset), zetc=exp_q(:), &
                                       lc_min=lb_min(jset), rac=rac, &
                                       rac2=rac2, vac=vac, v=v, f=f)

                     ! trafo to spherical and scale by c_a*c_b, only implemented for uncontracted Gaussians and lmin=lmax
                     I_tmp(1:ncoa_a, 1:nsgfb(jset)) = MATMUL(vac(1:ncoa_a, 1:ncoa_b), &
                                                             sphi_b(1:ncoa_b, ref_at_se_sgf_ofs(jatom, jset, 3): &
                                                                    (ref_at_se_sgf_ofs(jatom, jset, 3)+nsgfb(jset)-1)))
                     I_tmp2(1:nsgfa(iset), 1:nsgfb(jset)) = &
                        I_tmp2(1:nsgfa(iset), 1:nsgfb(jset))- &
                        MATMUL(TRANSPOSE(sphi_a(1:ncoa_a, ref_at_se_sgf_ofs(iatom, iset, 3): &
                                                (ref_at_se_sgf_ofs(iatom, iset, 3)+nsgfa(iset)-1))), &
                               I_tmp(1:ncoa_a, 1:nsgfb(jset)))/ &
                        (zeta(1, iset))**(la_max(iset)*0.5_dp+3.0_dp/4.0_dp)/ &
                        (zetb(1, jset))**(lb_max(jset)*0.5_dp+3.0_dp/4.0_dp)* &
                        exp_q(1)**(la_max(iset)*1.0_dp+3.0_dp/2.0_dp)* &
                        exp_q(1)**(lb_max(jset)*1.0_dp+3.0_dp/2.0_dp)/ &
                        (zeta(1, ref_at_se_sgf_ofs(iatom, iset, 2)))**(la_max(iset)*0.5_dp+3.0_dp/4.0_dp)/ &
                        (zetb(1, ref_at_se_sgf_ofs(jatom, jset, 2)))**(lb_max(jset)*0.5_dp+3.0_dp/4.0_dp)

                     ! add contribution from (a|b)_s-(q_a|q_b)_s
                     DO i = 1, nsgfa(iset)
                        DO j = 1, nsgfb(jset)
                           ! for parallelization: check, if sgf_j is in paralleliz. range
                           IF (ref_at_se_sgf_ofs(jatom, jset, 4)+j >= my_group_L_start .AND. &
                               ref_at_se_sgf_ofs(jatom, jset, 4)+j <= my_group_L_end .AND. &
                               check_j) THEN

                              L_local_col(ref_at_se_sgf_ofs(iatom, iset, 4)+i, &
                                          ref_at_se_sgf_ofs(jatom, jset, 4)+j-my_group_L_start+1) = &
                                 L_local_col(ref_at_se_sgf_ofs(iatom, iset, 4)+i, &
                                             ref_at_se_sgf_ofs(jatom, jset, 4)+j-my_group_L_start+1)+ &
                                 I_tmp2(i, j)
                           END IF

                           IF (ref_at_se_sgf_ofs(iatom, iset, 4)+i >= my_group_L_start .AND. &
                               ref_at_se_sgf_ofs(iatom, iset, 4)+i <= my_group_L_end .AND. &
                               check_i) THEN
                              L_local_col(ref_at_se_sgf_ofs(jatom, jset, 4)+j, &
                                          ref_at_se_sgf_ofs(iatom, iset, 4)+i-my_group_L_start+1) = &
                                 L_local_col(ref_at_se_sgf_ofs(jatom, jset, 4)+j, &
                                             ref_at_se_sgf_ofs(iatom, iset, 4)+i-my_group_L_start+1)+ &
                                 I_tmp2(i, j)
                           END IF
                        END DO
                     END DO

                     DEALLOCATE (f, v, vac, I_tmp, I_tmp2)

                  END IF !parallelization

               END DO !jset
            END DO !iset

         END DO

         CALL neighbor_list_iterator_release(nl_iterator)

         DEALLOCATE (diffsets_atom_set_bf, ref_at_se_sgf_ofs)
         DEALLOCATE (atom_set_afo_diffbf)
         DEALLOCATE (sizes_array_pot, starts_array_pot, ends_array_pot)

         CALL mp_sum(L_local_col, para_env_sub%group)

         ! end if of new algorithm
      ELSE
         ! old algorithm

         ALLOCATE (L_local_col(dimen_RI, my_group_L_size))
         L_local_col = 0.0_dp

         IF (eri_method .EQ. do_eri_mme .AND. my_ri_metric .EQ. ri_coulomb) THEN
            CALL eri_mme_2c_integrate(eri_param, para_env_sub, qs_env, &
                                      basis_type_a="RI_AUX", basis_type_b="RI_AUX", &
                                      hab=L_local_col, first_b=my_group_L_start, last_b=my_group_L_end)

         ELSEIF (eri_method .EQ. do_eri_gpw .OR. my_ri_metric .NE. ri_coulomb) THEN

            ALLOCATE (wf_vector(dimen_RI))

            i_counter = 0
            DO LLL = my_group_L_start, my_group_L_end
               i_counter = i_counter+1

               wf_vector = 0.0_dp
               wf_vector(LLL) = 1.0_dp

               ! pseudo psi_L
               CALL calculate_wavefunction(mo_coeff, 1, psi_L, rho_g, atomic_kind_set, &
                                           qs_kind_set, cell, dft_control, particle_set, pw_env_sub, &
                                           basis_type="RI_AUX", &
                                           external_vector=wf_vector)

               CALL timeset(routineN//"_pot_lm", handle3)
               rho_r%pw%cr3d = psi_L%pw%cr3d

               ! in case we do the overlap metric, we have no potential and hence no FT has to be carried out
               IF (my_ri_metric /= ri_overlap) THEN
                  CALL pw_transfer(rho_r%pw, rho_g%pw)
                  CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
                  CALL pw_transfer(pot_g%pw, rho_r%pw)
               END IF

               CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)

               NULLIFY (rs_v)
               NULLIFY (rs_descs)
               CALL pw_env_get(pw_env_sub, rs_descs=rs_descs, rs_grids=rs_v)
               DO i = 1, SIZE(rs_v)
                  CALL rs_grid_retain(rs_v(i)%rs_grid)
               END DO
               CALL potential_pw2rs(rs_v, rho_r, pw_env_sub)

               CALL timestop(handle3)

               ! integrate the little bastards
               offset = 0
               DO iatom = 1, natom
                  ikind = kind_of(iatom)
                  CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=basis_set_a, basis_type="RI_AUX")

                  first_sgfa => basis_set_a%first_sgf
                  la_max => basis_set_a%lmax
                  la_min => basis_set_a%lmin
                  npgfa => basis_set_a%npgf
                  nseta = basis_set_a%nset
                  nsgfa => basis_set_a%nsgf_set
                  rpgfa => basis_set_a%pgf_radius
                  set_radius_a => basis_set_a%set_radius
                  sphi_a => basis_set_a%sphi
                  zeta => basis_set_a%zet

                  ra(:) = pbc(particle_set(iatom)%r, cell)
                  rab = 0.0_dp
                  rab2 = 0.0_dp

                  DO iset = 1, nseta
                     ncoa = npgfa(iset)*ncoset(la_max(iset))
                     sgfa = first_sgfa(1, iset)

                     ALLOCATE (I_tmp2(ncoa, 1))
                     I_tmp2 = 0.0_dp
                     ALLOCATE (I_ab(nsgfa(iset), 1))
                     I_ab = 0.0_dp

                     igrid_level = gaussian_gridlevel(pw_env_sub%gridlevel_info, MINVAL(zeta(:, iset)))
                     map_it_here = .FALSE.
                     IF (.NOT. ALL(rs_v(igrid_level)%rs_grid%desc%perd == 1)) THEN
                        DO dir = 1, 3
                           ! bounds of local grid (i.e. removing the 'wings'), if periodic
                           tp(dir) = FLOOR(DOT_PRODUCT(cell%h_inv(dir, :), ra)*rs_v(igrid_level)%rs_grid%desc%npts(dir))
                           tp(dir) = MODULO(tp(dir), rs_v(igrid_level)%rs_grid%desc%npts(dir))
                           IF (rs_v(igrid_level)%rs_grid%desc%perd(dir) .NE. 1) THEN
                              lb(dir) = rs_v(igrid_level)%rs_grid%lb_local(dir)+rs_v(igrid_level)%rs_grid%desc%border
                              ub(dir) = rs_v(igrid_level)%rs_grid%ub_local(dir)-rs_v(igrid_level)%rs_grid%desc%border
                           ELSE
                              lb(dir) = rs_v(igrid_level)%rs_grid%lb_local(dir)
                              ub(dir) = rs_v(igrid_level)%rs_grid%ub_local(dir)
                           ENDIF
                           ! distributed grid, only map if it is local to the grid
                           location(dir) = tp(dir)+rs_v(igrid_level)%rs_grid%desc%lb(dir)
                        ENDDO
                        IF (lb(1) <= location(1) .AND. location(1) <= ub(1) .AND. &
                            lb(2) <= location(2) .AND. location(2) <= ub(2) .AND. &
                            lb(3) <= location(3) .AND. location(3) <= ub(3)) THEN
                           map_it_here = .TRUE.
                        ENDIF
                     ELSE
                        ! not distributed, just a round-robin distribution over the full set of CPUs
                        IF (MODULO(offset, para_env_sub%num_pe) == para_env_sub%mepos) map_it_here = .TRUE.
                     ENDIF

                     offset = offset+nsgfa(iset)

                     IF (map_it_here) THEN
                        DO ipgf = 1, npgfa(iset)
                           sgfa = first_sgfa(1, iset)

                           na1 = (ipgf-1)*ncoset(la_max(iset))+1
                           na2 = ipgf*ncoset(la_max(iset))
                           igrid_level = gaussian_gridlevel(pw_env_sub%gridlevel_info, zeta(ipgf, iset))

                           CALL integrate_pgf_product_rspace( &
                              la_max=la_max(iset), zeta=zeta(ipgf, iset)/2.0_dp, la_min=la_min(iset), &
                              lb_max=0, zetb=zeta(ipgf, iset)/2.0_dp, lb_min=0, &
                              ra=ra, rab=rab, rab2=rab2, &
                              rsgrid=rs_v(igrid_level)%rs_grid, &
                              cell=cell, &
                              cube_info=pw_env_sub%cube_info(igrid_level), &
                              hab=I_tmp2, &
                              o1=na1-1, &
                              o2=0, &
                              map_consistent=.TRUE., &
                              eps_gvg_rspace=dft_control%qs_control%eps_gvg_rspace, &
                              calculate_forces=.FALSE.)
                        END DO

                        CALL dgemm("T", "N", nsgfa(iset), 1, ncoa, &
                                   1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                   I_tmp2(1, 1), SIZE(I_tmp2, 1), &
                                   1.0_dp, I_ab(1, 1), SIZE(I_ab, 1))

                        L_local_col(offset-nsgfa(iset)+1:offset, i_counter) = I_ab(1:nsgfa(iset), 1)
                     END IF

                     DEALLOCATE (I_tmp2)
                     DEALLOCATE (I_ab)

                  END DO
               END DO

               DO i = 1, SIZE(rs_v)
                  CALL rs_grid_release(rs_v(i)%rs_grid)
               END DO

            END DO

            DEALLOCATE (wf_vector)

            CALL mp_sum(L_local_col, para_env_sub%group)

         ELSE
            CALL cp__b("mp2_ri_gpw.F",2105,"unknown ERI method")
         ENDIF
      END IF !end old code

      CALL timestop(handle2)

      ! split the total number of proc in a subgroup of the size of ~1/10 of the
      ! total num of proc
      best_group_size = para_env%num_pe

      strat_group_size = MAX(1, para_env%num_pe/10)

      min_mem_for_QK = REAL(dimen_RI, KIND=dp)*dimen_RI*3.0_dp*8.0_dp/1024_dp/1024_dp

      group_size = strat_group_size-1
      DO iproc = strat_group_size, para_env%num_pe
         group_size = group_size+1
         ! check that group_size is a multiple of sub_group_size and a divisor of
         ! the total num of proc
         IF (MOD(para_env%num_pe, group_size) /= 0 .OR. MOD(group_size, para_env_sub%num_pe) /= 0) CYCLE

         ! check for memory
         IF (REAL(group_size, KIND=dp)*mp2_memory < min_mem_for_QK) CYCLE

         best_group_size = group_size
         EXIT
      END DO

      ! create the L group
      color_L = para_env%mepos/best_group_size
      CALL mp_comm_split_direct(para_env%group, comm_L, color_L)
      NULLIFY (para_env_L)
      CALL cp_para_env_create(para_env_L, comm_L)

      ! create the blacs_L
      NULLIFY (blacs_env_L)
      CALL cp_blacs_env_create(blacs_env=blacs_env_L, para_env=para_env_L)

      ! now create the exchange group (for communication only between members not belonging to the
      ! same group
      sub_sub_color = para_env_sub%mepos
      CALL mp_comm_split_direct(para_env_L%group, comm_exchange, sub_sub_color)
      NULLIFY (para_env_exchange)
      CALL cp_para_env_create(para_env_exchange, comm_exchange)

      ! crate the proc maps
      ALLOCATE (proc_map(-para_env_exchange%num_pe:2*para_env_exchange%num_pe-1))
      DO i = 0, para_env_exchange%num_pe-1
         proc_map(i) = i
         proc_map(-i-1) = para_env_exchange%num_pe-i-1
         proc_map(para_env_exchange%num_pe+i) = i
      END DO

      ! create the information array
      ALLOCATE (sub_sizes_array(0:para_env_exchange%num_pe-1))
      sub_sizes_array = 0
      ALLOCATE (sub_starts_array(0:para_env_exchange%num_pe-1))
      sub_starts_array = 0
      ALLOCATE (sub_ends_array(0:para_env_exchange%num_pe-1))
      sub_ends_array = 0

      sub_sizes_array(para_env_exchange%mepos) = my_group_L_size
      sub_starts_array(para_env_exchange%mepos) = my_group_L_start
      sub_ends_array(para_env_exchange%mepos) = my_group_L_end

      CALL mp_sum(sub_sizes_array, para_env_exchange%group)
      CALL mp_sum(sub_starts_array, para_env_exchange%group)
      CALL mp_sum(sub_ends_array, para_env_exchange%group)

      ! create the full matrix L defined in the L group
      NULLIFY (fm_matrix_L)
      NULLIFY (fm_struct)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env_L, nrow_global=dimen_RI, &
                               ncol_global=dimen_RI, para_env=para_env_L)
      IF (my_external_matrix_struc) THEN
         CALL cp_fm_create(fm_matrix_L, fm_matrix_L_extern%matrix_struct, name="fm_matrix_L")
      ELSE
         CALL cp_fm_create(fm_matrix_L, fm_struct, name="fm_matrix_L")
      END IF

      CALL cp_fm_struct_release(fm_struct)

      CALL cp_fm_set_all(matrix=fm_matrix_L, alpha=0.0_dp)

      CALL cp_fm_get_info(matrix=fm_matrix_L, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)

      DO jjB = 1, ncol_local
         j_global = col_indices(jjB)
         IF (j_global >= my_group_L_start .AND. j_global <= my_group_L_end) THEN
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               fm_matrix_L%local_data(iiB, jjB) = L_local_col(i_global, j_global-my_group_L_start+1)
            END DO
         END IF
      END DO

      proc_send_static = proc_map(para_env_exchange%mepos+1)
      proc_receive_static = proc_map(para_env_exchange%mepos-1)

      DO proc_shift = 1, para_env_exchange%num_pe-1
         proc_send = proc_map(para_env_exchange%mepos+proc_shift)
         proc_receive = proc_map(para_env_exchange%mepos-proc_shift)

         rec_L_size = sub_sizes_array(proc_receive)
         rec_L_start = sub_starts_array(proc_receive)
         rec_L_end = sub_ends_array(proc_receive)

         ALLOCATE (L_external_col(dimen_RI, rec_L_size))
         L_external_col = 0.0_dp

         CALL mp_sendrecv(L_local_col, proc_send_static, L_external_col, proc_receive_static, para_env_exchange%group)

         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            IF (j_global >= rec_L_start .AND. j_global <= rec_L_end) THEN
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  fm_matrix_L%local_data(iiB, jjB) = L_external_col(i_global, j_global-rec_L_start+1)
               END DO
            END IF
         END DO

         DEALLOCATE (L_local_col)
         ALLOCATE (L_local_col(dimen_RI, rec_L_size))
         L_local_col(:, :) = L_external_col

         DEALLOCATE (L_external_col)

      END DO

      DEALLOCATE (L_local_col)

      ! free the old exchange group stuff
      DEALLOCATE (proc_map)
      CALL cp_para_env_release(para_env_exchange)

      DEALLOCATE (sub_sizes_array)
      DEALLOCATE (sub_starts_array)
      DEALLOCATE (sub_ends_array)

      ! create the new group for the mp_sum of the local data
      sub_sub_color = para_env_L%mepos
      CALL mp_comm_split_direct(para_env%group, comm_exchange, sub_sub_color)
      NULLIFY (para_env_exchange)
      CALL cp_para_env_create(para_env_exchange, comm_exchange)

      CALL mp_sum(fm_matrix_L%local_data, para_env_exchange%group)

      CALL cp_para_env_release(para_env_exchange)

      cond_num = 1.0_dp
      num_small_eigen = 0
      IF (calc_PQ_cond_num) THEN
         ! calculate the condition number of the (P|Q) matrix
         ! create a copy of the matrix
         NULLIFY (fm_matrix_L_diag)
         NULLIFY (fm_struct)
         CALL cp_fm_struct_create(fm_struct, context=blacs_env_L, nrow_global=dimen_RI, &
                                  ncol_global=dimen_RI, para_env=para_env_L)
         IF (my_external_matrix_struc) THEN
            CALL cp_fm_create(fm_matrix_L_diag, fm_matrix_L_extern%matrix_struct, name="fm_matrix_L_diag")
         ELSE
            CALL cp_fm_create(fm_matrix_L_diag, fm_struct, name="fm_matrix_L_diag")
         END IF

         CALL cp_fm_struct_release(fm_struct)

         CALL cp_fm_set_all(matrix=fm_matrix_L_diag, alpha=0.0_dp)

         CALL cp_fm_to_fm(source=fm_matrix_L, destination=fm_matrix_L_diag)

         ALLOCATE (egen_L(dimen_RI))

         egen_L = 0.0_dp
         CALL cp_fm_syevx(matrix=fm_matrix_L_diag, eigenvalues=egen_L)

         num_small_eigen = 0
         DO iiB = 1, dimen_RI
            IF (ABS(egen_L(iiB)) < 0.001_dp) num_small_eigen = num_small_eigen+1
         END DO

         cond_num = MAXVAL(ABS(egen_L))/MINVAL(ABS(egen_L))

         CALL cp_fm_release(fm_matrix_L_diag)

         DEALLOCATE (egen_L)
      END IF

      ! do cholesky decomposition
      CALL cp_fm_cholesky_decompose(matrix=fm_matrix_L, n=dimen_RI, info_out=info_chol)
      IF(.NOT.(info_chol == 0))CALL cp__a("mp2_ri_gpw.F",2299)

      IF (do_inversion) THEN
         CALL cp_fm_triangular_invert(matrix_a=fm_matrix_L, uplo_tr='U')
      END IF

      ! clean the lower part of the L^{-1} matrix (just to not have surprises afterwards)
      CALL cp_fm_get_info(matrix=fm_matrix_L, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)
      DO iiB = 1, nrow_local
         i_global = row_indices(iiB)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            IF (j_global < i_global) fm_matrix_L%local_data(iiB, jjB) = 0.0_dp
         END DO
      END DO

      ! release blacs_env
      CALL cp_blacs_env_release(blacs_env_L)

      CALL timestop(handle)

   END SUBROUTINE compute_2c_in_and_cholesky

! **************************************************************************************************
!> \brief ...
!> \param para_env ...
!> \param dimen_RI ...
!> \param fm_matrix_L ...
!> \param my_group_L_start ...
!> \param my_group_L_end ...
!> \param my_group_L_size ...
!> \param my_Lrows ...
! **************************************************************************************************
   SUBROUTINE grep_Lcols(para_env, dimen_RI, fm_matrix_L, &
                         my_group_L_start, my_group_L_end, my_group_L_size, my_Lrows)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: dimen_RI
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_L
      INTEGER                                            :: my_group_L_start, my_group_L_end, &
                                                            my_group_L_size
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: my_Lrows

      CHARACTER(LEN=*), PARAMETER :: routineN = 'grep_Lcols', routineP = moduleN//':'//routineN

      INTEGER :: handle, handle2, i_global, iiB, j_global, jjB, max_row_col_local, ncol_local, &
         ncol_rec, nrow_local, nrow_rec, proc_receive, proc_receive_static, proc_send, &
         proc_send_static, proc_shift
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_col_row_info, rec_col_row_info
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, col_indices_rec, &
                                                            row_indices, row_indices_rec
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_L, local_L_internal, rec_L

      CALL timeset(routineN, handle)

      ALLOCATE (my_Lrows(dimen_RI, my_group_L_size))
      my_Lrows = 0.0_dp

      ! proc_map, vector that replicate the processor numbers also
      ! for negative and positive number > num_pe
      ! needed to know which is the processor, to respect to another one,
      ! for a given shift
      ALLOCATE (proc_map(-para_env%num_pe:2*para_env%num_pe-1))
      DO iiB = 0, para_env%num_pe-1
         proc_map(iiB) = iiB
         proc_map(-iiB-1) = para_env%num_pe-iiB-1
         proc_map(para_env%num_pe+iiB) = iiB
      END DO

      CALL cp_fm_get_info(matrix=fm_matrix_L, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          local_data=local_L_internal)

      ALLOCATE (local_L(nrow_local, ncol_local))
      local_L = local_L_internal(1:nrow_local, 1:ncol_local)

      max_row_col_local = MAX(nrow_local, ncol_local)
      CALL mp_max(max_row_col_local, para_env%group)

      ALLOCATE (local_col_row_info(0:max_row_col_local, 2))
      local_col_row_info = 0
      ! 0,1 nrows
      local_col_row_info(0, 1) = nrow_local
      local_col_row_info(1:nrow_local, 1) = row_indices(1:nrow_local)
      ! 0,2 ncols
      local_col_row_info(0, 2) = ncol_local
      local_col_row_info(1:ncol_local, 2) = col_indices(1:ncol_local)

      ALLOCATE (rec_col_row_info(0:max_row_col_local, 2))

      ! accumulate data on my_Lrows starting from myself
      DO jjB = 1, ncol_local
         j_global = col_indices(jjB)
         IF (j_global >= my_group_L_start .AND. j_global <= my_group_L_end) THEN
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               my_Lrows(i_global, j_global-my_group_L_start+1) = local_L(iiB, jjB)
            END DO
         END IF
      END DO

      proc_send_static = proc_map(para_env%mepos+1)
      proc_receive_static = proc_map(para_env%mepos-1)

      CALL timeset(routineN//"_comm", handle2)

      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         proc_receive = proc_map(para_env%mepos-proc_shift)

         ! first exchange information on the local data
         rec_col_row_info = 0
         CALL mp_sendrecv(local_col_row_info, proc_send_static, rec_col_row_info, proc_receive_static, para_env%group)
         nrow_rec = rec_col_row_info(0, 1)
         ncol_rec = rec_col_row_info(0, 2)

         ALLOCATE (row_indices_rec(nrow_rec))
         row_indices_rec = rec_col_row_info(1:nrow_rec, 1)

         ALLOCATE (col_indices_rec(ncol_rec))
         col_indices_rec = rec_col_row_info(1:ncol_rec, 2)

         ALLOCATE (rec_L(nrow_rec, ncol_rec))
         rec_L = 0.0_dp

         ! then send and receive the real data
         CALL mp_sendrecv(local_L, proc_send_static, rec_L, proc_receive_static, para_env%group)

         ! accumulate the received data on my_Lrows
         DO jjB = 1, ncol_rec
            j_global = col_indices_rec(jjB)
            IF (j_global >= my_group_L_start .AND. j_global <= my_group_L_end) THEN
               DO iiB = 1, nrow_rec
                  i_global = row_indices_rec(iiB)
                  my_Lrows(i_global, j_global-my_group_L_start+1) = rec_L(iiB, jjB)
               END DO
            END IF
         END DO

         local_col_row_info(:, :) = rec_col_row_info
         DEALLOCATE (local_L)
         ALLOCATE (local_L(nrow_rec, ncol_rec))
         local_L = rec_L

         DEALLOCATE (col_indices_rec)
         DEALLOCATE (row_indices_rec)
         DEALLOCATE (rec_L)
      END DO
      CALL timestop(handle2)

      DEALLOCATE (local_col_row_info)
      DEALLOCATE (rec_col_row_info)
      DEALLOCATE (proc_map)
      DEALLOCATE (local_L)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param fm_matrix_L ...
! **************************************************************************************************
   SUBROUTINE set_identity(fm_matrix_L)
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_L

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_identity', routineP = moduleN//':'//routineN

      INTEGER                                            :: i_global, iiB, j_global, jjB, &
                                                            ncol_local, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices

      CALL cp_fm_get_info(matrix=fm_matrix_L, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)
      DO iiB = 1, nrow_local
         i_global = row_indices(iiB)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            IF (j_global == i_global) THEN
               fm_matrix_L%local_data(iiB, jjB) = 1.0_dp
            ELSE
               fm_matrix_L%local_data(iiB, jjB) = 0.0_dp
            END IF
         END DO
      END DO

   END SUBROUTINE set_identity

! **************************************************************************************************
!> \brief ...
!> \param para_env_sub ...
!> \param fm_BIb_jb ...
!> \param BIb_jb ...
!> \param max_row_col_local ...
!> \param proc_map ...
!> \param local_col_row_info ...
!> \param my_B_virtual_end ...
!> \param my_B_virtual_start ...
! **************************************************************************************************
   SUBROUTINE grep_my_integrals(para_env_sub, fm_BIb_jb, BIb_jb, max_row_col_local, &
                                proc_map, local_col_row_info, &
                                my_B_virtual_end, my_B_virtual_start)
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      TYPE(cp_fm_type), POINTER                          :: fm_BIb_jb
      REAL(KIND=dp), DIMENSION(:, :)                     :: BIb_jb
      INTEGER                                            :: max_row_col_local
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_col_row_info
      INTEGER                                            :: my_B_virtual_end, my_B_virtual_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'grep_my_integrals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_global, iiB, j_global, jjB, ncol_rec, &
                                                            nrow_rec, proc_receive, proc_send, &
                                                            proc_shift
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: rec_col_row_info
      INTEGER, DIMENSION(:), POINTER                     :: col_indices_rec, row_indices_rec
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_BI, rec_BI

      ALLOCATE (rec_col_row_info(0:max_row_col_local, 2))

      rec_col_row_info(:, :) = local_col_row_info

      nrow_rec = rec_col_row_info(0, 1)
      ncol_rec = rec_col_row_info(0, 2)

      ALLOCATE (row_indices_rec(nrow_rec))
      row_indices_rec = rec_col_row_info(1:nrow_rec, 1)

      ALLOCATE (col_indices_rec(ncol_rec))
      col_indices_rec = rec_col_row_info(1:ncol_rec, 2)

      ! accumulate data on BIb_jb buffer starting from myself
      DO jjB = 1, ncol_rec
         j_global = col_indices_rec(jjB)
         IF (j_global >= my_B_virtual_start .AND. j_global <= my_B_virtual_end) THEN
            DO iiB = 1, nrow_rec
               i_global = row_indices_rec(iiB)
               BIb_jb(j_global-my_B_virtual_start+1, i_global) = fm_BIb_jb%local_data(iiB, jjB)
            END DO
         END IF
      END DO

      DEALLOCATE (row_indices_rec)
      DEALLOCATE (col_indices_rec)

      IF (para_env_sub%num_pe > 1) THEN
         ALLOCATE (local_BI(nrow_rec, ncol_rec))
         local_BI(1:nrow_rec, 1:ncol_rec) = fm_BIb_jb%local_data(1:nrow_rec, 1:ncol_rec)

         DO proc_shift = 1, para_env_sub%num_pe-1
            proc_send = proc_map(para_env_sub%mepos+proc_shift)
            proc_receive = proc_map(para_env_sub%mepos-proc_shift)

            ! first exchange information on the local data
            rec_col_row_info = 0
            CALL mp_sendrecv(local_col_row_info, proc_send, rec_col_row_info, proc_receive, para_env_sub%group)
            nrow_rec = rec_col_row_info(0, 1)
            ncol_rec = rec_col_row_info(0, 2)

            ALLOCATE (row_indices_rec(nrow_rec))
            row_indices_rec = rec_col_row_info(1:nrow_rec, 1)

            ALLOCATE (col_indices_rec(ncol_rec))
            col_indices_rec = rec_col_row_info(1:ncol_rec, 2)

            ALLOCATE (rec_BI(nrow_rec, ncol_rec))
            rec_BI = 0.0_dp

            ! then send and receive the real data
            CALL mp_sendrecv(local_BI, proc_send, rec_BI, proc_receive, para_env_sub%group)

            ! accumulate the received data on BIb_jb buffer
            DO jjB = 1, ncol_rec
               j_global = col_indices_rec(jjB)
               IF (j_global >= my_B_virtual_start .AND. j_global <= my_B_virtual_end) THEN
                  DO iiB = 1, nrow_rec
                     i_global = row_indices_rec(iiB)
                     BIb_jb(j_global-my_B_virtual_start+1, i_global) = rec_BI(iiB, jjB)
                  END DO
               END IF
            END DO

            DEALLOCATE (col_indices_rec)
            DEALLOCATE (row_indices_rec)
            DEALLOCATE (rec_BI)
         END DO

         DEALLOCATE (local_BI)
      END IF

      DEALLOCATE (rec_col_row_info)

   END SUBROUTINE grep_my_integrals

! **************************************************************************************************
!> \brief ...
!> \param Emp2 ...
!> \param Emp2_Cou ...
!> \param Emp2_EX ...
!> \param BIb_C ...
!> \param mp2_env ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param color_sub ...
!> \param ends_array ...
!> \param ends_B_virtual ...
!> \param sizes_array ...
!> \param sizes_B_virtual ...
!> \param starts_array ...
!> \param starts_B_virtual ...
!> \param Eigenval ...
!> \param nmo ...
!> \param homo ...
!> \param dimen_RI ...
!> \param unit_nr ...
!> \param calc_forces ...
!> \param calc_ex ...
!> \param open_shell_SS ...
!> \param BIb_C_beta ...
!> \param homo_beta ...
!> \param Eigenval_beta ...
!> \param ends_B_virtual_beta ...
!> \param sizes_B_virtual_beta ...
!> \param starts_B_virtual_beta ...
! **************************************************************************************************
   SUBROUTINE mp2_ri_gpw_compute_en(Emp2, Emp2_Cou, Emp2_EX, BIb_C, mp2_env, para_env, para_env_sub, color_sub, &
                                    ends_array, ends_B_virtual, sizes_array, sizes_B_virtual, starts_array, starts_B_virtual, &
                                    Eigenval, nmo, homo, dimen_RI, unit_nr, calc_forces, calc_ex, &
                                    open_shell_SS, BIb_C_beta, homo_beta, Eigenval_beta, &
                                    ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta)
      REAL(KIND=dp)                                      :: Emp2, Emp2_Cou, Emp2_EX
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER                                            :: color_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_array, ends_B_virtual, sizes_array, &
                                                            sizes_B_virtual, starts_array, &
                                                            starts_B_virtual
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: nmo, homo, dimen_RI, unit_nr
      LOGICAL, INTENT(IN)                                :: calc_forces
      LOGICAL                                            :: calc_ex
      LOGICAL, OPTIONAL                                  :: open_shell_SS
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :), &
         OPTIONAL                                        :: BIb_C_beta
      INTEGER, OPTIONAL                                  :: homo_beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      INTEGER, ALLOCATABLE, DIMENSION(:), OPTIONAL       :: ends_B_virtual_beta, &
                                                            sizes_B_virtual_beta, &
                                                            starts_B_virtual_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_ri_gpw_compute_en', &
         routineP = moduleN//':'//routineN

      INTEGER :: a, a_global, b, b_global, best_block_size, best_integ_group_size, block_size, &
         comm_exchange, comm_P, end_point, handle, handle2, handle3, iiB, ij_counter, &
         ij_counter_send, ij_index, integ_group_size, irep, jjB, Lend_pos, Lstart_pos, &
         max_ij_pairs, min_integ_group_size, my_B_size, my_B_size_beta, my_B_virtual_end, &
         my_B_virtual_end_beta, my_B_virtual_start, my_B_virtual_start_beta, my_block_size, &
         my_group_L_end, my_group_L_size, my_group_L_size_orig, my_group_L_start, my_homo_beta, &
         my_i, my_ij_pairs, my_j, my_new_group_L_size, my_num_dgemm_call, ngroup, num_IJ_blocks
      INTEGER :: num_integ_group, pos_integ_group, proc_receive, proc_send, proc_shift, &
         rec_B_size, rec_B_virtual_end, rec_B_virtual_start, rec_L_size, send_B_size, &
         send_B_virtual_end, send_B_virtual_start, send_block_size, send_i, send_ij_index, send_j, &
         start_point, sub_P_color, sub_sub_color, total_ij_pairs, virtual, virtual_beta
      INTEGER, ALLOCATABLE, DIMENSION(:) :: integ_group_pos2color_sub, new_sizes_array, &
         num_ij_pairs, proc_map, proc_map_rep, sizes_array_orig, sub_proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ij_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: ranges_info_array
      LOGICAL                                            :: my_alpha_alpha_case, my_alpha_beta_case, &
                                                            my_beta_beta_case, my_open_shell_SS
      REAL(KIND=dp) :: actual_flop_rate, amp_fac, mem_for_aK, mem_for_comm, mem_for_iaK, &
         mem_for_rep, mem_min, mem_per_group, mem_real, my_flop_rate, null_mat_rec(2, 2, 2), &
         null_mat_send(2, 2, 2), sym_fac, t_end, t_start
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: external_ab, external_i_aL, local_ab, &
                                                            local_ba, t_ab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: B_ia_Q, B_ia_Q_beta, BI_C_rec, &
                                                            local_i_aL, local_j_aL, Y_i_aP, &
                                                            Y_j_aP, Y_j_aP_beta
      TYPE(cp_para_env_type), POINTER                    :: para_env_exchange, para_env_P, &
                                                            para_env_rep

      CALL timeset(routineN, handle)

      my_open_shell_SS = .FALSE.
      IF (PRESENT(open_shell_SS)) my_open_shell_SS = open_shell_SS

      ! t_ab = amp_fac*(:,a|:,b)-(:,b|:,a)
      IF (calc_forces) amp_fac = 2.0_dp
      ! If we calculate the gradient we need to distinguish
      ! between alpha-alpha and beta-beta cases for UMP2

      my_alpha_alpha_case = .FALSE.
      my_beta_beta_case = .FALSE.
      my_alpha_beta_case = .FALSE.
      IF (calc_forces) THEN
         IF (my_open_shell_SS) THEN
            amp_fac = 1.0_dp
            IF ((.NOT. ALLOCATED(mp2_env%ri_grad%P_ij)) &
                .AND. (.NOT. ALLOCATED(mp2_env%ri_grad%P_ab))) THEN
               my_alpha_alpha_case = .TRUE.
               amp_fac = 1.0_dp
            ELSE
               IF ((.NOT. ALLOCATED(mp2_env%ri_grad%P_ij_beta)) &
                   .AND. (.NOT. ALLOCATED(mp2_env%ri_grad%P_ab_beta))) THEN
                  my_beta_beta_case = .TRUE.
               ENDIF
            ENDIF
         ENDIF
      ENDIF

      my_alpha_beta_case = .FALSE.
      IF (PRESENT(BIb_C_beta) .AND. &
          PRESENT(ends_B_virtual_beta) .AND. &
          PRESENT(sizes_B_virtual_beta) .AND. &
          PRESENT(starts_B_virtual_beta) .AND. &
          PRESENT(homo_beta) .AND. &
          PRESENT(Eigenval_beta)) THEN
         my_alpha_beta_case = .TRUE.
         my_alpha_alpha_case = .FALSE.
      ENDIF

      IF (my_alpha_beta_case) amp_fac = 1.0_dp

      virtual = nmo-homo
      IF (my_alpha_beta_case) virtual_beta = nmo-homo_beta

      CALL mp2_ri_get_sizes( &
         mp2_env, para_env, para_env_sub, ends_array, ends_B_virtual, sizes_array, sizes_B_virtual, &
         starts_array, starts_B_virtual, homo, dimen_RI, unit_nr, color_sub, best_block_size, best_integ_group_size, block_size, &
         integ_group_size, min_integ_group_size, my_B_size, my_B_virtual_end, my_B_virtual_start, my_group_L_size, &
         my_group_L_start, my_group_L_end, ngroup, num_IJ_blocks, num_integ_group, pos_integ_group, virtual, my_alpha_beta_case, &
         my_open_shell_SS, mem_for_aK, mem_for_comm, mem_for_iaK, mem_for_rep, mem_min, mem_per_group, mem_real)

      IF (my_alpha_beta_case) THEN
         my_B_virtual_start_beta = starts_B_virtual_beta(para_env_sub%mepos)
         my_B_virtual_end_beta = ends_B_virtual_beta(para_env_sub%mepos)
         my_B_size_beta = sizes_B_virtual_beta(para_env_sub%mepos)
         my_homo_beta = homo_beta
      ELSE
         my_B_virtual_start_beta = my_B_virtual_start
         my_B_virtual_end_beta = my_B_virtual_end
         my_B_size_beta = my_B_size
         my_homo_beta = homo
      END IF

      ! now create a group that contains all the proc that have the same virtual starting point
      ! in the integ group
      ! sub_sub_color=para_env_sub%mepos
      CALL mp2_ri_create_group( &
         BIb_C, para_env, para_env_sub, homo, color_sub, &
         sizes_array, calc_forces, &
         comm_exchange, integ_group_size, my_B_size, iiB, my_group_L_end, &
         my_group_L_size, my_group_L_size_orig, my_group_L_start, my_new_group_L_size, &
         sub_sub_color, integ_group_pos2color_sub, new_sizes_array, proc_map, proc_map_rep, sizes_array_orig, &
         sub_proc_map, ranges_info_array, para_env_exchange, para_env_rep, num_integ_group)

      ! *****************************************************************
      ! **********  REPLICATION-BLOCKED COMMUNICATION SCHEME  ***********
      ! *****************************************************************
      ! introduce block size, the number of occupied orbitals has to be a
      ! multiple of the block size

      ! Calculate the maximum number of ij pairs that have to be computed
      ! among groups
      CALL mp2_ri_communication(my_alpha_beta_case, total_ij_pairs, homo, homo_beta, num_IJ_blocks, &
                                block_size, ngroup, ij_map, color_sub, my_ij_pairs, my_open_shell_SS, unit_nr)

      ALLOCATE (num_ij_pairs(0:para_env_exchange%num_pe-1))
      num_ij_pairs = 0
      num_ij_pairs(para_env_exchange%mepos) = my_ij_pairs
      CALL mp_sum(num_ij_pairs, para_env_exchange%group)

      max_ij_pairs = MAXVAL(num_ij_pairs)

      ! start real stuff
      IF (.NOT. my_alpha_beta_case) THEN
         CALL mp2_ri_allocate(local_ab, t_ab, mp2_env, homo, virtual, dimen_RI, my_B_size, &
                              block_size, my_B_size_beta, my_group_L_size, local_i_aL, &
                              local_j_aL, calc_forces, Y_i_aP, Y_j_aP, &
                              my_alpha_beta_case, &
                              my_beta_beta_case)
      ELSE
         CALL mp2_ri_allocate(local_ab, t_ab, mp2_env, homo, virtual, dimen_RI, my_B_size, &
                              block_size, my_B_size_beta, my_group_L_size, local_i_aL, &
                              local_j_aL, calc_forces, Y_i_aP, Y_j_aP_beta, &
                              my_alpha_beta_case, &
                              my_beta_beta_case, local_ba, virtual_beta)
      ENDIF

      CALL timeset(routineN//"_RI_loop", handle2)
      null_mat_rec = 0.0_dp
      null_mat_send = 0.0_dp
      Emp2 = 0.0_dp
      Emp2_Cou = 0.0_dp
      Emp2_EX = 0.0_dp
      my_num_dgemm_call = 0
      my_flop_rate = 0.0_dp
      DO ij_index = 1, max_ij_pairs

         IF (ij_index <= my_ij_pairs) THEN
            ! We have work to do
            ij_counter = (ij_index-MIN(1, color_sub))*ngroup+color_sub
            my_i = ij_map(ij_counter, 1)
            my_j = ij_map(ij_counter, 2)
            my_block_size = ij_map(ij_counter, 3)

            local_i_aL = 0.0_dp
            ! local_i_aL(my_group_L_start:my_group_L_end,1:my_B_size)=BIb_C(1:my_group_L_size,1:my_B_size,my_i)
            DO irep = 0, num_integ_group-1
               Lstart_pos = ranges_info_array(1, irep, para_env_exchange%mepos)
               Lend_pos = ranges_info_array(2, irep, para_env_exchange%mepos)
               start_point = ranges_info_array(3, irep, para_env_exchange%mepos)
               end_point = ranges_info_array(4, irep, para_env_exchange%mepos)

               local_i_aL(Lstart_pos:Lend_pos, 1:my_B_size, 1:my_block_size) = &
                  BIb_C(start_point:end_point, 1:my_B_size, my_i:my_i+my_block_size-1)
            END DO

            local_j_aL = 0.0_dp
            ! local_j_aL(my_group_L_start:my_group_L_end,1:my_B_size)=BIb_C(1:my_group_L_size,1:my_B_size,my_j)
            DO irep = 0, num_integ_group-1
               Lstart_pos = ranges_info_array(1, irep, para_env_exchange%mepos)
               Lend_pos = ranges_info_array(2, irep, para_env_exchange%mepos)
               start_point = ranges_info_array(3, irep, para_env_exchange%mepos)
               end_point = ranges_info_array(4, irep, para_env_exchange%mepos)

               IF (.NOT. my_alpha_beta_case) THEN
                  local_j_aL(Lstart_pos:Lend_pos, 1:my_B_size, 1:my_block_size) = &
                     BIb_C(start_point:end_point, 1:my_B_size, my_j:my_j+my_block_size-1)
               ELSE
                  local_j_aL(Lstart_pos:Lend_pos, 1:my_B_size_beta, 1:my_block_size) = &
                     BIb_C_beta(start_point:end_point, 1:my_B_size_beta, my_j:my_j+my_block_size-1)
               END IF
            END DO

            ! collect data from other proc
            CALL timeset(routineN//"_comm", handle3)
            DO proc_shift = 1, para_env_exchange%num_pe-1
               proc_send = proc_map(para_env_exchange%mepos+proc_shift)
               proc_receive = proc_map(para_env_exchange%mepos-proc_shift)

               send_ij_index = num_ij_pairs(proc_send)

               rec_L_size = sizes_array(proc_receive)
               ALLOCATE (BI_C_rec(rec_L_size, MAX(my_B_size, my_B_size_beta), my_block_size))

               IF (ij_index <= send_ij_index) THEN
                  ! ij_counter_send=(ij_index-MIN(1,proc_send))*ngroup+proc_send
                  ij_counter_send = (ij_index-MIN(1, integ_group_pos2color_sub(proc_send)))*ngroup+ &
                                    integ_group_pos2color_sub(proc_send)
                  send_i = ij_map(ij_counter_send, 1)
                  send_j = ij_map(ij_counter_send, 2)
                  send_block_size = ij_map(ij_counter_send, 3)

                  ! occupied i
                  BI_C_rec = 0.0_dp
                  CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:my_B_size, send_i:send_i+send_block_size-1), proc_send, &
                                   BI_C_rec(1:rec_L_size, 1:my_B_size, 1:my_block_size), proc_receive, &
                                   para_env_exchange%group)
                  DO irep = 0, num_integ_group-1
                     Lstart_pos = ranges_info_array(1, irep, proc_receive)
                     Lend_pos = ranges_info_array(2, irep, proc_receive)
                     start_point = ranges_info_array(3, irep, proc_receive)
                     end_point = ranges_info_array(4, irep, proc_receive)

                     local_i_aL(Lstart_pos:Lend_pos, 1:my_B_size, 1:my_block_size) = &
                        BI_C_rec(start_point:end_point, 1:my_B_size, 1:my_block_size)

                  END DO

                  ! occupied j
                  BI_C_rec = 0.0_dp
                  IF (.NOT. my_alpha_beta_case) THEN
                     CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:my_B_size, send_j:send_j+send_block_size-1), proc_send, &
                                      BI_C_rec(1:rec_L_size, 1:my_B_size, 1:my_block_size), proc_receive, &
                                      para_env_exchange%group)
                  ELSE
                     CALL mp_sendrecv(BIb_C_beta(1:my_group_L_size, 1:my_B_size_beta, send_j:send_j+send_block_size-1), proc_send, &
                                      BI_C_rec(1:rec_L_size, 1:my_B_size_beta, 1:my_block_size), proc_receive, &
                                      para_env_exchange%group)
                  END IF

                  DO irep = 0, num_integ_group-1
                     Lstart_pos = ranges_info_array(1, irep, proc_receive)
                     Lend_pos = ranges_info_array(2, irep, proc_receive)
                     start_point = ranges_info_array(3, irep, proc_receive)
                     end_point = ranges_info_array(4, irep, proc_receive)

                     IF (.NOT. my_alpha_beta_case) THEN
                        local_j_aL(Lstart_pos:Lend_pos, 1:my_B_size, 1:my_block_size) = &
                           BI_C_rec(start_point:end_point, 1:my_B_size, 1:my_block_size)
                     ELSE
                        local_j_aL(Lstart_pos:Lend_pos, 1:my_B_size_beta, 1:my_block_size) = &
                           BI_C_rec(start_point:end_point, 1:my_B_size_beta, 1:my_block_size)
                     END IF

                  END DO

               ELSE
                  ! we send the null matrix while we know that we have to receive something

                  ! occupied i
                  BI_C_rec = 0.0_dp
                  CALL mp_sendrecv(null_mat_send, proc_send, &
                                   BI_C_rec(1:rec_L_size, 1:my_B_size, 1:my_block_size), proc_receive, &
                                   para_env_exchange%group)

                  DO irep = 0, num_integ_group-1
                     Lstart_pos = ranges_info_array(1, irep, proc_receive)
                     Lend_pos = ranges_info_array(2, irep, proc_receive)
                     start_point = ranges_info_array(3, irep, proc_receive)
                     end_point = ranges_info_array(4, irep, proc_receive)

                     local_i_aL(Lstart_pos:Lend_pos, 1:my_B_size, 1:my_block_size) = &
                        BI_C_rec(start_point:end_point, 1:my_B_size, 1:my_block_size)

                  END DO

                  ! occupied j
                  BI_C_rec = 0.0_dp
                  IF (.NOT. my_alpha_beta_case) THEN
                     CALL mp_sendrecv(null_mat_send, proc_send, &
                                      BI_C_rec(1:rec_L_size, 1:my_B_size, 1:my_block_size), proc_receive, &
                                      para_env_exchange%group)
                  ELSE
                     CALL mp_sendrecv(null_mat_send, proc_send, &
                                      BI_C_rec(1:rec_L_size, 1:my_B_size_beta, 1:my_block_size), proc_receive, &
                                      para_env_exchange%group)
                  END IF
                  DO irep = 0, num_integ_group-1
                     Lstart_pos = ranges_info_array(1, irep, proc_receive)
                     Lend_pos = ranges_info_array(2, irep, proc_receive)
                     start_point = ranges_info_array(3, irep, proc_receive)
                     end_point = ranges_info_array(4, irep, proc_receive)

                     IF (.NOT. my_alpha_beta_case) THEN
                        local_j_aL(Lstart_pos:Lend_pos, 1:my_B_size, 1:my_block_size) = &
                           BI_C_rec(start_point:end_point, 1:my_B_size, 1:my_block_size)
                     ELSE
                        local_j_aL(Lstart_pos:Lend_pos, 1:my_B_size_beta, 1:my_block_size) = &
                           BI_C_rec(start_point:end_point, 1:my_B_size_beta, 1:my_block_size)
                     END IF

                  END DO

               END IF

               DEALLOCATE (BI_C_rec)

            END DO
            CALL timestop(handle3)

            ! loop over the block elements
            DO iiB = 1, my_block_size
               DO jjB = 1, my_block_size
                  CALL timeset(routineN//"_expansion", handle3)
                  ! calculate the integrals (ia|jb) strating from my local data ...
                  local_ab = 0.0_dp
                  IF ((my_alpha_beta_case) .AND. (calc_forces)) THEN
                     local_ba = 0.0_dp
                  ENDIF
                  t_start = m_walltime()
                  CALL dgemm('T', 'N', my_B_size, my_B_size_beta, dimen_RI, 1.0_dp, &
                             local_i_aL(:, :, iiB), dimen_RI, local_j_aL(:, :, jjB), dimen_RI, &
                             0.0_dp, local_ab(my_B_virtual_start:my_B_virtual_end, 1:my_B_size_beta), my_B_size)
                  t_end = m_walltime()
                  actual_flop_rate = 2.0_dp*my_B_size*my_B_size_beta*REAL(dimen_RI, KIND=dp)/(MAX(0.01_dp, t_end-t_start))
                  my_flop_rate = my_flop_rate+actual_flop_rate
                  my_num_dgemm_call = my_num_dgemm_call+1
                  ! Additional integrals only for alpha_beta case and forces
                  IF ((my_alpha_beta_case) .AND. (calc_forces)) THEN
                     t_start = m_walltime()
                     CALL dgemm('T', 'N', my_B_size_beta, my_B_size, dimen_RI, 1.0_dp, &
                                local_j_aL(:, :, iiB), dimen_RI, local_i_aL(:, :, jjB), dimen_RI, &
                                0.0_dp, local_ba(my_B_virtual_start_beta:my_B_virtual_end_beta, 1:my_B_size), my_B_size_beta)
                     t_end = m_walltime()
                     actual_flop_rate = 2.0_dp*my_B_size*my_B_size_beta*REAL(dimen_RI, KIND=dp)/(MAX(0.01_dp, t_end-t_start))
                     my_flop_rate = my_flop_rate+actual_flop_rate
                     my_num_dgemm_call = my_num_dgemm_call+1
                  ENDIF
                  ! ... and from the other of my subgroup
                  DO proc_shift = 1, para_env_sub%num_pe-1
                     proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
                     proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

                     rec_B_size = sizes_B_virtual(proc_receive)
                     rec_B_virtual_end = ends_B_virtual(proc_receive)
                     rec_B_virtual_start = starts_B_virtual(proc_receive)

                     ALLOCATE (external_i_aL(dimen_RI, rec_B_size))
                     external_i_aL = 0.0_dp

                     CALL mp_sendrecv(local_i_aL(:, :, iiB), proc_send, &
                                      external_i_aL, proc_receive, &
                                      para_env_sub%group)

                     ! local_ab(rec_B_virtual_start:rec_B_virtual_end,1:my_B_size)=MATMUL(TRANSPOSE(external_i_aL),local_j_aL)
                     t_start = m_walltime()
                     CALL dgemm('T', 'N', rec_B_size, my_B_size_beta, dimen_RI, 1.0_dp, &
                                external_i_aL, dimen_RI, local_j_aL(:, :, jjB), dimen_RI, &
                                0.0_dp, local_ab(rec_B_virtual_start:rec_B_virtual_end, 1:my_B_size_beta), rec_B_size)

                     t_end = m_walltime()
                     actual_flop_rate = 2.0_dp*rec_B_size*my_B_size_beta*REAL(dimen_RI, KIND=dp)/(MAX(0.01_dp, t_end-t_start))
                     my_flop_rate = my_flop_rate+actual_flop_rate
                     my_num_dgemm_call = my_num_dgemm_call+1

                     DEALLOCATE (external_i_aL)
                     ! Additional integrals only for alpha_beta case and forces
                     IF ((my_alpha_beta_case) .AND. (calc_forces)) THEN

                        rec_B_size = sizes_B_virtual_beta(proc_receive)
                        rec_B_virtual_end = ends_B_virtual_beta(proc_receive)
                        rec_B_virtual_start = starts_B_virtual_beta(proc_receive)

                        ALLOCATE (external_i_aL(dimen_RI, rec_B_size))
                        external_i_aL = 0.0_dp

                        CALL mp_sendrecv(local_j_aL(:, :, jjB), proc_send, &
                                         external_i_aL, proc_receive, &
                                         para_env_sub%group)

                        t_start = m_walltime()
                        CALL dgemm('T', 'N', rec_B_size, my_B_size, dimen_RI, 1.0_dp, &
                                   external_i_aL, dimen_RI, local_i_aL(:, :, iiB), dimen_RI, &
                                   0.0_dp, local_ba(rec_B_virtual_start:rec_B_virtual_end, 1:my_B_size), rec_B_size)
                        t_end = m_walltime()
                        actual_flop_rate = 2.0_dp*rec_B_size*my_B_size*REAL(dimen_RI, KIND=dp)/(MAX(0.01_dp, t_end-t_start))
                        my_flop_rate = my_flop_rate+actual_flop_rate
                        my_num_dgemm_call = my_num_dgemm_call+1

                        DEALLOCATE (external_i_aL)
                     ENDIF

                  END DO
                  CALL timestop(handle3)

                  !sample peak memory
                  CALL m_memory()

                  CALL timeset(routineN//"_ener", handle3)
                  ! calculate coulomb only MP2
                  sym_fac = 2.0_dp
                  IF (my_i == my_j) sym_fac = 1.0_dp
                  IF (.NOT. my_alpha_beta_case) THEN
                     ! IF(my_open_shell_SS) sym_fac=sym_fac/2.0_dp
                     DO b = 1, my_B_size
                        b_global = b+my_B_virtual_start-1
                        DO a = 1, virtual
                           Emp2_Cou = Emp2_Cou-sym_fac*2.0_dp*local_ab(a, b)**2/ &
                                      (Eigenval(homo+a)+Eigenval(homo+b_global)-Eigenval(my_i+iiB-1)-Eigenval(my_j+jjB-1))
                        END DO
                     END DO
                  ELSE
                     DO b = 1, my_B_size_beta
                        b_global = b+my_B_virtual_start_beta-1
                        DO a = 1, virtual
                           Emp2_Cou = Emp2_Cou-local_ab(a, b)**2/ &
                                      (Eigenval(homo+a)+Eigenval_beta(homo_beta+b_global)- &
                                       Eigenval(my_i+iiB-1)-Eigenval_beta(my_j+jjB-1))
                        END DO
                     END DO
                  END IF

                  IF (calc_ex) THEN
                     ! contract integrals with orbital energies for exchange MP2 energy
                     ! starting with local ...
                     ! IF(my_open_shell_SS) sym_fac=sym_fac*2.0_dp
                     IF (calc_forces .AND. (.NOT. my_alpha_beta_case)) t_ab = 0.0_dp
                     DO b = 1, my_B_size
                        b_global = b+my_B_virtual_start-1
                        DO a = 1, my_B_size
                           a_global = a+my_B_virtual_start-1
                           Emp2_Ex = Emp2_Ex+sym_fac*local_ab(a_global, b)*local_ab(b_global, a)/ &
                                     (Eigenval(homo+a_global)+Eigenval(homo+b_global)-Eigenval(my_i+iiB-1)-Eigenval(my_j+jjB-1))
                           IF (calc_forces .AND. (.NOT. my_alpha_beta_case)) &
                              t_ab(a_global, b) = -(amp_fac*local_ab(a_global, b)-local_ab(b_global, a))/ &
                                                  (Eigenval(homo+a_global)+Eigenval(homo+b_global)- &
                                                   Eigenval(my_i+iiB-1)-Eigenval(my_j+jjB-1))
                        END DO
                     END DO
                     ! ... and then with external data
                     DO proc_shift = 1, para_env_sub%num_pe-1
                        proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
                        proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

                        rec_B_size = sizes_B_virtual(proc_receive)
                        rec_B_virtual_end = ends_B_virtual(proc_receive)
                        rec_B_virtual_start = starts_B_virtual(proc_receive)

                        send_B_size = sizes_B_virtual(proc_send)
                        send_B_virtual_end = ends_B_virtual(proc_send)
                        send_B_virtual_start = starts_B_virtual(proc_send)

                        ALLOCATE (external_ab(my_B_size, rec_B_size))
                        external_ab = 0.0_dp

                        CALL mp_sendrecv(local_ab(send_B_virtual_start:send_B_virtual_end, 1:my_B_size), proc_send, &
                                         external_ab(1:my_B_size, 1:rec_B_size), proc_receive, &
                                         para_env_sub%group)

                        DO b = 1, my_B_size
                           b_global = b+my_B_virtual_start-1
                           DO a = 1, rec_B_size
                              a_global = a+rec_B_virtual_start-1
                              Emp2_Ex = Emp2_Ex+sym_fac*local_ab(a_global, b)*external_ab(b, a)/ &
                                        (Eigenval(homo+a_global)+Eigenval(homo+b_global)-Eigenval(my_i+iiB-1)-Eigenval(my_j+jjB-1))
                              IF (calc_forces .AND. (.NOT. my_alpha_beta_case)) &
                                 t_ab(a_global, b) = -(amp_fac*local_ab(a_global, b)-external_ab(b, a))/ &
                                                     (Eigenval(homo+a_global)+Eigenval(homo+b_global)- &
                                                      Eigenval(my_i+iiB-1)-Eigenval(my_j+jjB-1))
                           END DO
                        END DO

                        DEALLOCATE (external_ab)
                     END DO
                  END IF
                  CALL timestop(handle3)

                  IF (calc_forces) THEN
                     ! update P_ab, Gamma_P_ia
                     IF (.NOT. my_alpha_beta_case) THEN
                        Y_i_aP = 0.0_dp
                        Y_j_aP = 0.0_dp
                        CALL mp2_update_P_gamma(mp2_env, para_env_sub, ends_B_virtual, sizes_B_virtual, &
                                                starts_B_virtual, Eigenval, homo, dimen_RI, iiB, jjB, my_B_size, &
                                                my_B_virtual_end, my_B_virtual_start, my_i, my_j, virtual, &
                                                sub_proc_map, local_ab, t_ab, local_i_aL, local_j_aL, &
                                                my_open_shell_ss, my_alpha_alpha_case, my_beta_beta_case, Y_i_aP, Y_j_aP)
                     ELSE
                        Y_i_aP = 0.0_dp
                        Y_j_aP_beta = 0.0_dp
                        CALL mp2_update_P_gamma(mp2_env, para_env_sub, ends_B_virtual, sizes_B_virtual, &
                                                starts_B_virtual, Eigenval, homo, dimen_RI, iiB, jjB, my_B_size, &
                                                my_B_virtual_end, my_B_virtual_start, my_i, my_j, virtual, sub_proc_map, &
                                                local_ab, t_ab, local_i_aL, local_j_aL, my_open_shell_ss, my_alpha_alpha_case, &
                                                my_beta_beta_case, Y_i_aP, Y_j_aP_beta, Eigenval_beta, homo_beta, my_B_size_beta, &
                                                ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta, &
                                                my_B_virtual_start_beta, my_B_virtual_end_beta, virtual_beta, local_ba)
                     ENDIF

                  END IF

               END DO ! jjB
            END DO ! iiB

         ELSE
            ! No work to do and we know that we have to receive nothing, but send something
            ! send data to other proc
            DO proc_shift = 1, para_env_exchange%num_pe-1
               proc_send = proc_map(para_env_exchange%mepos+proc_shift)
               proc_receive = proc_map(para_env_exchange%mepos-proc_shift)

               send_ij_index = num_ij_pairs(proc_send)

               IF (ij_index <= send_ij_index) THEN
                  ! something to send
                  ! ij_counter_send=(ij_index-MIN(1,proc_send))*ngroup+proc_send
                  ij_counter_send = (ij_index-MIN(1, integ_group_pos2color_sub(proc_send)))*ngroup+ &
                                    integ_group_pos2color_sub(proc_send)
                  send_i = ij_map(ij_counter_send, 1)
                  send_j = ij_map(ij_counter_send, 2)
                  send_block_size = ij_map(ij_counter_send, 3)

                  ! occupied i
                  CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:my_B_size, send_i:send_i+send_block_size-1), proc_send, &
                                   null_mat_rec, proc_receive, &
                                   para_env_exchange%group)
                  ! occupied j
                  IF (.NOT. my_alpha_beta_case) THEN
                     CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:my_B_size, send_j:send_j+send_block_size-1), proc_send, &
                                      null_mat_rec, proc_receive, &
                                      para_env_exchange%group)
                  ELSE
                     CALL mp_sendrecv(BIb_C_beta(1:my_group_L_size, 1:my_B_size_beta, send_j:send_j+send_block_size-1), proc_send, &
                                      null_mat_rec, proc_receive, &
                                      para_env_exchange%group)
                  END IF

               ELSE
                  ! nothing to send
                  ! occupied i
                  CALL mp_sendrecv(null_mat_send, proc_send, &
                                   null_mat_rec, proc_receive, &
                                   para_env_exchange%group)
                  ! occupied j
                  CALL mp_sendrecv(null_mat_send, proc_send, &
                                   null_mat_rec, proc_receive, &
                                   para_env_exchange%group)

               END IF
            END DO
         END IF

         ! redistribute gamma
         IF (calc_forces) THEN
            ! Closed shell, alpha-alpha or beta-beta case
            IF ((.NOT. my_alpha_beta_case)) THEN
               CALL mp2_redistribute_gamma(mp2_env, ij_index, my_B_size, &
                                           my_block_size, my_group_L_size, my_i, my_ij_pairs, my_j, ngroup, &
                                           num_integ_group, integ_group_pos2color_sub, num_ij_pairs, proc_map, &
                                           ij_map, ranges_info_array, Y_i_aP, Y_j_aP, para_env_exchange, &
                                           null_mat_rec, null_mat_send, sizes_array, my_alpha_alpha_case, &
                                           my_beta_beta_case, my_alpha_beta_case, my_open_shell_ss)
            ELSE
               ! Alpha-beta case
               CALL mp2_redistribute_gamma(mp2_env, ij_index, my_B_size, &
                                           my_block_size, my_group_L_size, my_i, my_ij_pairs, my_j, ngroup, &
                                           num_integ_group, integ_group_pos2color_sub, num_ij_pairs, proc_map, &
                                           ij_map, ranges_info_array, Y_i_aP, Y_j_aP_beta, para_env_exchange, &
                                           null_mat_rec, null_mat_send, sizes_array, my_alpha_alpha_case, my_beta_beta_case, &
                                           my_alpha_beta_case, my_open_shell_ss, my_B_size_beta)
            ENDIF
         END IF

      END DO
      CALL timestop(handle2)

      DEALLOCATE (local_i_aL)
      DEALLOCATE (local_j_aL)
      DEALLOCATE (ij_map)
      DEALLOCATE (num_ij_pairs)

      IF (calc_forces) THEN
         DEALLOCATE (Y_i_aP)
         IF (.NOT. my_alpha_beta_case) THEN
            DEALLOCATE (Y_j_aP)
         ELSE
            DEALLOCATE (Y_j_aP_beta)
         ENDIF
         IF (ALLOCATED(t_ab)) THEN
            DEALLOCATE (t_ab)
         ENDIF
         ! Deallocate additional integrals: alpha_beta case with forces
         IF (ALLOCATED(local_ba)) THEN
            DEALLOCATE (local_ba)
         ENDIF

         ! here we check if there are almost degenerate ij
         ! pairs and we update P_ij with these contribution.
         ! If all pairs are degenerate with each other this step will scale O(N^6),
         ! if the number of degenerate pairs scales linearly with the system size
         ! this step will scale O(N^5).
         ! Start counting the number of almost degenerate ij pairs according
         ! to eps_canonical
         IF (.NOT. my_alpha_beta_case) THEN
            CALL quasi_degenerate_P_ij( &
               mp2_env, Eigenval, homo, virtual, my_open_shell_ss, &
               my_beta_beta_case, my_alpha_beta_case, Bib_C, unit_nr, dimen_RI, &
               my_B_size, ngroup, num_integ_group, my_group_L_size, &
               color_sub, ranges_info_array, para_env_exchange, para_env_sub, proc_map, &
               my_B_virtual_start, my_B_virtual_end, sizes_array, ends_B_virtual, sizes_B_virtual, &
               starts_B_virtual, sub_proc_map, integ_group_pos2color_sub, local_ab)
         ELSE
            CALL quasi_degenerate_P_ij( &
               mp2_env, Eigenval, homo, virtual, my_open_shell_ss, &
               my_beta_beta_case, my_alpha_beta_case, Bib_C, unit_nr, dimen_RI, &
               my_B_size, ngroup, num_integ_group, my_group_L_size, &
               color_sub, ranges_info_array, para_env_exchange, para_env_sub, proc_map, &
               my_B_virtual_start, my_B_virtual_end, sizes_array, ends_B_virtual, sizes_B_virtual, &
               starts_B_virtual, sub_proc_map, integ_group_pos2color_sub, local_ab, BIb_C_beta, my_B_size_beta, &
               ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta, my_B_virtual_start_beta, &
               virtual_beta, homo_beta, Eigenval_beta, my_B_virtual_end_beta)
         ENDIF

      END IF

      DEALLOCATE (integ_group_pos2color_sub)
      DEALLOCATE (local_ab)

      CALL mp_sum(Emp2_Cou, para_env%group)
      CALL mp_sum(Emp2_Ex, para_env%group)

      IF (calc_forces) THEN
         ! sum P_ab
         IF (.NOT. my_open_shell_ss) THEN
            mp2_env%ri_grad%P_ab(:, :) = mp2_env%ri_grad%P_ab(:, :)*amp_fac
            IF (my_alpha_beta_case) mp2_env%ri_grad%P_ab_beta(:, :) = &
               mp2_env%ri_grad%P_ab_beta(:, :)*amp_fac
            sub_P_color = para_env_sub%mepos
            CALL mp_comm_split_direct(para_env%group, comm_P, sub_P_color)
            NULLIFY (para_env_P)
            CALL cp_para_env_create(para_env_P, comm_P)
            CALL mp_sum(mp2_env%ri_grad%P_ab, para_env_P%group)
            IF (my_alpha_beta_case) CALL mp_sum(mp2_env%ri_grad%P_ab_beta, para_env_P%group)
            ! release para_env_P
            CALL cp_para_env_release(para_env_P)
         ENDIF

         ! sum P_ij (later)
         ! mp2_env%ri_grad%P_ij=mp2_env%ri_grad%P_ij*2.0_dp
         ! CALL mp_sum(mp2_env%ri_grad%P_ij,para_env%group)
         ! WRITE(*,*) mp2_env%ri_mp2%eps_canonical

         ! XXXXXXXXXXXXXX
         ! Write P_ab
         ! ALLOCATE(external_ab(virtual,virtual))
         ! external_ab=0.0_dp
         ! external_ab(my_B_virtual_start:my_B_virtual_end,1:virtual)=mp2_env%ri_grad%P_ab
         ! CALL mp_sum(external_ab,para_env_sub%group)
         ! IF(para_env%mepos==0) CALL write_array(external_ab)
         ! DEALLOCATE(external_ab)
         ! ! Write P_ij
         ! IF(para_env%mepos==0) CALL write_array(mp2_env%ri_grad%P_ij)
         ! XXXXXXXXXXXXXX

         ! recover original information (before replication)
         DEALLOCATE (sizes_array)
         iiB = SIZE(sizes_array_orig)
         ALLOCATE (sizes_array(0:iiB-1))
         sizes_array(:) = sizes_array_orig
         DEALLOCATE (sizes_array_orig)

         ! make a copy of the original integrals (ia|Q)
         my_group_L_size = my_group_L_size_orig
         ALLOCATE (B_ia_Q(homo, my_B_size, my_group_L_size))
         B_ia_Q = 0.0_dp
         DO jjB = 1, homo
            DO iiB = 1, my_B_size
               B_ia_Q(jjB, iiB, 1:my_group_L_size) = BIb_C(1:my_group_L_size, iiB, jjB)
            END DO
         END DO
         DEALLOCATE (BIb_C)
         IF (my_alpha_beta_case) THEN
            ALLOCATE (B_ia_Q_beta(homo_beta, my_B_size_beta, my_group_L_size))
            B_ia_Q_beta = 0.0_dp
            DO jjB = 1, homo_beta
               DO iiB = 1, my_B_size_beta
                  B_ia_Q_beta(jjB, iiB, 1:my_group_L_size) = &
                     BIb_C_beta(1:my_group_L_size, iiB, jjB)
               END DO
            END DO
            DEALLOCATE (BIb_C_beta)
         ENDIF

         ! sum Gamma and dereplicate
         ALLOCATE (BIb_C(homo, my_B_size, my_group_L_size))
         IF (my_alpha_beta_case) ALLOCATE (BIb_C_beta(homo_beta, my_B_size_beta, my_group_L_size))
         DO proc_shift = 1, para_env_rep%num_pe-1
            ! invert order
            proc_send = proc_map_rep(para_env_rep%mepos-proc_shift)
            proc_receive = proc_map_rep(para_env_rep%mepos+proc_shift)

            start_point = ranges_info_array(3, proc_shift, para_env_exchange%mepos)
            end_point = ranges_info_array(4, proc_shift, para_env_exchange%mepos)

            BIb_C = 0.0_dp
            ! Closed shell, alpha-alpha, and alpha-alpha part in alpha-beta case
            IF (my_alpha_alpha_case .OR. (.NOT. my_open_shell_ss)) THEN
               CALL mp_sendrecv(mp2_env%ri_grad%Gamma_P_ia(1:homo, 1:my_B_size, start_point:end_point), &
                                proc_send, BIb_C, proc_receive, para_env_rep%group)
               mp2_env%ri_grad%Gamma_P_ia(1:homo, 1:my_B_size, 1:my_group_L_size) = &
                  mp2_env%ri_grad%Gamma_P_ia(1:homo, 1:my_B_size, 1:my_group_L_size)+BIb_C
            ENDIF
            ! Beta-beta
            IF (my_beta_beta_case) THEN
               CALL mp_sendrecv(mp2_env%ri_grad%Gamma_P_ia_beta(1:homo, 1:my_B_size, start_point:end_point), &
                                proc_send, BIb_C, proc_receive, para_env_rep%group)
               mp2_env%ri_grad%Gamma_P_ia_beta(1:homo, 1:my_B_size, 1:my_group_L_size) = &
                  mp2_env%ri_grad%Gamma_P_ia_beta(1:homo, 1:my_B_size, 1:my_group_L_size)+BIb_C
            ENDIF
            IF (my_alpha_beta_case) THEN ! Beta-beta part of alpha-beta case
               CALL mp_sendrecv(mp2_env%ri_grad%Gamma_P_ia_beta(1:homo_beta, 1:my_B_size_beta, start_point:end_point), &
                                proc_send, BIb_C_beta, proc_receive, para_env_rep%group)
               mp2_env%ri_grad%Gamma_P_ia_beta(1:homo_beta, 1:my_B_size_beta, 1:my_group_L_size) = &
                  mp2_env%ri_grad%Gamma_P_ia_beta(1:homo_beta, 1:my_B_size_beta, 1:my_group_L_size)+BIb_C_beta
            ENDIF

         END DO
         IF (.NOT. my_open_shell_ss) THEN
            BIb_C(:, :, :) = mp2_env%ri_grad%Gamma_P_ia(1:homo, 1:my_B_size, 1:my_group_L_size)
            DEALLOCATE (mp2_env%ri_grad%Gamma_P_ia)
            ALLOCATE (mp2_env%ri_grad%Gamma_P_ia(homo, my_B_size, my_group_L_size))
            mp2_env%ri_grad%Gamma_P_ia(:, :, :) = BIb_C
            DEALLOCATE (BIb_C)
            IF (my_alpha_beta_case) THEN
               BIb_C_beta(:, :, :) = mp2_env%ri_grad%Gamma_P_ia_beta(1:homo_beta, 1:my_B_size_beta, 1:my_group_L_size)
               DEALLOCATE (mp2_env%ri_grad%Gamma_P_ia_beta)
               ALLOCATE (mp2_env%ri_grad%Gamma_P_ia_beta(homo_beta, my_B_size_beta, my_group_L_size))
               mp2_env%ri_grad%Gamma_P_ia_beta(:, :, :) = BIb_C_beta
               DEALLOCATE (BIb_C_beta)
            ENDIF
         ENDIF
         ! For open shell systems, we need to pass Bib_C through the subroutine in alpha-alpha and beta-beta case.
         ! Only for forces, as IF suggests! Here we deallocate it, but restore after complete_gamma
         IF (my_open_shell_ss) THEN
            DEALLOCATE (BIb_C)
         ENDIF

         IF (.NOT. my_open_shell_ss) THEN
            CALL complete_gamma(mp2_env, B_ia_Q, dimen_RI, homo, virtual, para_env, para_env_sub, ngroup, &
                                my_group_L_size, my_group_L_start, my_group_L_end, &
                                my_B_size, my_B_virtual_start, &
                                ends_array, ends_B_virtual, sizes_array, sizes_B_virtual, starts_array, starts_B_virtual, &
                                sub_proc_map, .TRUE.)
            IF (my_alpha_beta_case) THEN
               CALL complete_gamma(mp2_env, B_ia_Q_beta, dimen_RI, homo_beta, virtual_beta, para_env, para_env_sub, &
                                   ngroup, my_group_L_size, my_group_L_start, my_group_L_end, &
                                   my_B_size_beta, my_B_virtual_start_beta, &
                                   ends_array, ends_B_virtual_beta, sizes_array, sizes_B_virtual_beta, starts_array, &
                                   starts_B_virtual_beta, sub_proc_map, .FALSE.)
            ENDIF
         ENDIF
         ! Here we restore BIb_C
         IF (my_open_shell_ss) THEN
            ALLOCATE (BIb_C(my_group_L_size, my_B_size, homo))
            BIb_C = 0.0_dp
            ! copy the integrals (ia|Q) back
            DO jjB = 1, homo
               DO iiB = 1, my_B_size
                  BIb_C(1:my_group_L_size, iiB, jjB) = &
                     B_ia_Q(jjB, iiB, 1:my_group_L_size)
               END DO
            END DO
         ENDIF

         ! ALLOCATE(Y_i_aP(homo,virtual,dimen_RI))
         ! Y_i_aP=0.0_dp
         ! Y_i_aP(1:homo,my_B_virtual_start:my_B_virtual_end,my_group_L_start:my_group_L_end)=mp2_env%ri_grad%Gamma_P_ia
         ! CALL mp_sum(Y_i_aP,para_env%group)
         ! IF(para_env%mepos==0) THEN
         !   DO iiB=1, dimen_RI
         !     WRITE(*,*) iiB
         !     CALL write_array(Y_i_aP(1:homo,1:virtual,iiB))
         !     WRITE(*,*)
         !   END DO
         ! END IF
         ! DEALLOCATE(Y_i_aP)

      END IF

      Emp2 = Emp2_Cou+Emp2_EX

      DEALLOCATE (proc_map)
      DEALLOCATE (sub_proc_map)
      DEALLOCATE (proc_map_rep)
      DEALLOCATE (ranges_info_array)

      IF (.NOT. my_open_shell_SS) THEN
         ! keep the array for the next calculations
         IF (ALLOCATED(BIb_C)) DEALLOCATE (BIb_C)
         DEALLOCATE (sizes_array)
         DEALLOCATE (starts_array)
         DEALLOCATE (ends_array)
         DEALLOCATE (starts_B_virtual)
         DEALLOCATE (ends_B_virtual)
         DEALLOCATE (sizes_B_virtual)
         IF (my_alpha_beta_case) THEN
            DEALLOCATE (starts_B_virtual_beta)
            DEALLOCATE (ends_B_virtual_beta)
            DEALLOCATE (sizes_B_virtual_beta)
         END IF
      END IF

      CALL cp_para_env_release(para_env_exchange)
      CALL cp_para_env_release(para_env_rep)

      my_flop_rate = my_flop_rate/REAL(MAX(my_num_dgemm_call, 1), KIND=dp)/1.0E9_dp
      CALL mp_sum(my_flop_rate, para_env%group)
      my_flop_rate = my_flop_rate/para_env%num_pe
      IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="(T3,A,T66,F15.2)") &
         "PERFORMANCE| DGEMM flop rate (Gflops / MPI rank):", my_flop_rate

      CALL timestop(handle)

   END SUBROUTINE mp2_ri_gpw_compute_en

! **************************************************************************************************
!> \brief ...
!> \param BIb_C ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param para_env_exchange ...
!> \param para_env_rep ...
!> \param homo ...
!> \param proc_map_rep ...
!> \param sizes_array ...
!> \param my_B_size ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
!> \param my_group_L_end ...
!> \param my_new_group_L_size ...
!> \param new_sizes_array ...
!> \param ranges_info_array ...
! **************************************************************************************************
   SUBROUTINE replicate_iaK_2intgroup(BIb_C, para_env, para_env_sub, para_env_exchange, para_env_rep, &
                                      homo, proc_map_rep, &
                                      sizes_array, &
                                      my_B_size, &
                                      my_group_L_size, my_group_L_start, my_group_L_end, &
                                      my_new_group_L_size, new_sizes_array, ranges_info_array)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub, &
                                                            para_env_exchange, para_env_rep
      INTEGER                                            :: homo
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: proc_map_rep, sizes_array
      INTEGER                                            :: my_B_size, my_group_L_size, &
                                                            my_group_L_start, my_group_L_end, &
                                                            my_new_group_L_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: new_sizes_array
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: ranges_info_array

      CHARACTER(LEN=*), PARAMETER :: routineN = 'replicate_iaK_2intgroup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: comm_rep, end_point, handle, i, &
                                                            max_L_size, proc_receive, proc_send, &
                                                            proc_shift, start_point, sub_sub_color
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rep_ends_array, rep_sizes_array, &
                                                            rep_starts_array
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C_copy
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: BIb_C_gather

      CALL timeset(routineN, handle)

      ! create the replication group
      sub_sub_color = para_env_sub%mepos*para_env_exchange%num_pe+para_env_exchange%mepos
      CALL mp_comm_split_direct(para_env%group, comm_rep, sub_sub_color)
      NULLIFY (para_env_rep)
      CALL cp_para_env_create(para_env_rep, comm_rep)

      ! crate the proc maps
      ALLOCATE (proc_map_rep(-para_env_rep%num_pe:2*para_env_rep%num_pe-1))
      DO i = 0, para_env_rep%num_pe-1
         proc_map_rep(i) = i
         proc_map_rep(-i-1) = para_env_rep%num_pe-i-1
         proc_map_rep(para_env_rep%num_pe+i) = i
      END DO

      ! create the new limits for K according to the size
      ! of the integral group
      ALLOCATE (new_sizes_array(0:para_env_exchange%num_pe-1))
      new_sizes_array = 0
      ALLOCATE (ranges_info_array(4, 0:para_env_rep%num_pe-1, 0:para_env_exchange%num_pe-1))
      ranges_info_array = 0

      ! info array for replication
      ALLOCATE (rep_ends_array(0:para_env_rep%num_pe-1))
      rep_ends_array = 0
      ALLOCATE (rep_starts_array(0:para_env_rep%num_pe-1))
      rep_starts_array = 0
      ALLOCATE (rep_sizes_array(0:para_env_rep%num_pe-1))
      rep_sizes_array = 0

      rep_sizes_array(para_env_rep%mepos) = my_group_L_size
      rep_starts_array(para_env_rep%mepos) = my_group_L_start
      rep_ends_array(para_env_rep%mepos) = my_group_L_end

      CALL mp_sum(rep_sizes_array, para_env_rep%group)
      CALL mp_sum(rep_starts_array, para_env_rep%group)
      CALL mp_sum(rep_ends_array, para_env_rep%group)

      ! calculate my_new_group_L_size according to sizes_array
      my_new_group_L_size = my_group_L_size
      ranges_info_array(1, 0, para_env_exchange%mepos) = my_group_L_start
      ranges_info_array(2, 0, para_env_exchange%mepos) = my_group_L_end
      ranges_info_array(3, 0, para_env_exchange%mepos) = 1
      ranges_info_array(4, 0, para_env_exchange%mepos) = my_group_L_size

      DO proc_shift = 1, para_env_rep%num_pe-1
         proc_send = proc_map_rep(para_env_rep%mepos+proc_shift)
         proc_receive = proc_map_rep(para_env_rep%mepos-proc_shift)

         my_new_group_L_size = my_new_group_L_size+rep_sizes_array(proc_receive)

         ranges_info_array(1, proc_shift, para_env_exchange%mepos) = rep_starts_array(proc_receive)
         ranges_info_array(2, proc_shift, para_env_exchange%mepos) = rep_ends_array(proc_receive)
         ranges_info_array(3, proc_shift, para_env_exchange%mepos) = ranges_info_array(4, proc_shift-1, para_env_exchange%mepos)+1
         ranges_info_array(4, proc_shift, para_env_exchange%mepos) = my_new_group_L_size

      END DO
      new_sizes_array(para_env_exchange%mepos) = my_new_group_L_size

      CALL mp_sum(new_sizes_array, para_env_exchange%group)
      CALL mp_sum(ranges_info_array, para_env_exchange%group)

      IF (.FALSE.) THEN
         ! replication scheme using mp_sendrecv
         ALLOCATE (BIb_C_copy(my_group_L_size, my_B_size, homo))
         BIb_C_copy(:, :, :) = BIb_C

         DEALLOCATE (BIb_C)

         ALLOCATE (BIb_C(my_new_group_L_size, my_B_size, homo))
         BIb_C = 0.0_dp

         start_point = ranges_info_array(3, 0, para_env_exchange%mepos)
         end_point = ranges_info_array(4, 0, para_env_exchange%mepos)

         BIb_C(start_point:end_point, 1:my_B_size, 1:homo) = BIb_C_copy

         DO proc_shift = 1, para_env_rep%num_pe-1
            proc_send = proc_map_rep(para_env_rep%mepos+proc_shift)
            proc_receive = proc_map_rep(para_env_rep%mepos-proc_shift)

            start_point = ranges_info_array(3, proc_shift, para_env_exchange%mepos)
            end_point = ranges_info_array(4, proc_shift, para_env_exchange%mepos)

            CALL mp_sendrecv(BIb_C_copy(1:my_group_L_size, 1:my_B_size, 1:homo), proc_send, &
                             BIb_C(start_point:end_point, 1:my_B_size, 1:homo), proc_receive, &
                             para_env_rep%group)

         END DO

         DEALLOCATE (BIb_C_copy)

      ELSE
         ! replication scheme using mp_allgather
         ! get the max L size of the
         max_L_size = MAXVAL(sizes_array)

         ALLOCATE (BIb_C_copy(max_L_size, my_B_size, homo))
         BIb_C_copy = 0.0_dp
         BIb_C_copy(1:my_group_L_size, 1:my_B_size, 1:homo) = BIb_C

         DEALLOCATE (BIb_C)

         ALLOCATE (BIb_C_gather(max_L_size, my_B_size, homo, 0:para_env_rep%num_pe-1))
         BIb_C_gather = 0.0_dp

         CALL mp_allgather(BIb_C_copy, BIb_C_gather, para_env_rep%group)

         DEALLOCATE (BIb_C_copy)

         ALLOCATE (BIb_C(my_new_group_L_size, my_B_size, homo))
         BIb_C = 0.0_dp

         ! reorder data
         DO proc_shift = 0, para_env_rep%num_pe-1
            proc_send = proc_map_rep(para_env_rep%mepos+proc_shift)
            proc_receive = proc_map_rep(para_env_rep%mepos-proc_shift)

            start_point = ranges_info_array(3, proc_shift, para_env_exchange%mepos)
            end_point = ranges_info_array(4, proc_shift, para_env_exchange%mepos)

            BIb_C(start_point:end_point, 1:my_B_size, 1:homo) = &
               BIb_C_gather(1:end_point-start_point+1, 1:my_B_size, 1:homo, proc_receive)

         END DO

         DEALLOCATE (BIb_C_gather)

      END IF

      ! DEALLOCATE(proc_map_rep)
      DEALLOCATE (rep_sizes_array)
      DEALLOCATE (rep_starts_array)
      DEALLOCATE (rep_ends_array)

      ! CALL  cp_para_env_release(para_env_rep)

      CALL timestop(handle)

   END SUBROUTINE replicate_iaK_2intgroup

! **************************************************************************************************
!> \brief ...
!> \param mat ...
! **************************************************************************************************
   SUBROUTINE write_array(mat)
      REAL(KIND=dp), DIMENSION(:, :)                     :: mat

      INTEGER                                            :: iii, jjj

      WRITE (*, *)
      DO iii = 1, SIZE(mat, 1)
         DO jjj = 1, SIZE(mat, 2), 20
            WRITE (*, '(1000F10.5)') mat(iii, jjj:MIN(SIZE(mat, 2), jjj+19))
         END DO
         WRITE (*, *)
      END DO
      WRITE (*, *)
   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param local_ab ...
!> \param t_ab ...
!> \param mp2_env ...
!> \param homo ...
!> \param virtual ...
!> \param dimen_RI ...
!> \param my_B_size ...
!> \param block_size ...
!> \param my_B_size_beta ...
!> \param my_group_L_size ...
!> \param local_i_aL ...
!> \param local_j_aL ...
!> \param calc_forces ...
!> \param Y_i_aP ...
!> \param Y_j_aP ...
!> \param alpha_beta ...
!> \param beta_beta ...
!> \param local_ba ...
!> \param virtual_beta ...
! **************************************************************************************************
   SUBROUTINE mp2_ri_allocate(local_ab, t_ab, mp2_env, homo, virtual, dimen_RI, my_B_size, &
                              block_size, my_B_size_beta, my_group_L_size, &
                              local_i_aL, local_j_aL, calc_forces, &
                              Y_i_aP, Y_j_aP, alpha_beta, &
                              beta_beta, local_ba, virtual_beta)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: local_ab, t_ab
      TYPE(mp2_type), POINTER                            :: mp2_env
      INTEGER                                            :: homo, virtual, dimen_RI, my_B_size, &
                                                            block_size, my_B_size_beta, &
                                                            my_group_L_size
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: local_i_aL, local_j_aL
      LOGICAL                                            :: calc_forces
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: Y_i_aP, Y_j_aP
      LOGICAL                                            :: alpha_beta, beta_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :), &
         OPTIONAL                                        :: local_ba
      INTEGER, OPTIONAL                                  :: virtual_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_ri_allocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      ALLOCATE (local_i_aL(dimen_RI, my_B_size, block_size))
      ALLOCATE (local_j_aL(dimen_RI, my_B_size_beta, block_size))
      ALLOCATE (local_ab(virtual, my_B_size_beta))

      IF (calc_forces) THEN
         ALLOCATE (Y_i_aP(my_B_size, dimen_RI, block_size))
         Y_i_aP = 0.0_dp
         ! For  closed-shell, alpha-alpha and beta-beta my_B_size_beta=my_b_size
         ! Not for alpha-beta case: Y_j_aP_beta is sent and received as Y_j_aP
         ALLOCATE (Y_j_aP(my_B_size_beta, dimen_RI, block_size))
         Y_j_aP = 0.0_dp
         ! Closed shell or alpha-alpha case
         IF (.NOT. (beta_beta .OR. alpha_beta)) THEN
            ALLOCATE (mp2_env%ri_grad%P_ij(homo, homo))
            ALLOCATE (mp2_env%ri_grad%P_ab(my_B_size, virtual))
            mp2_env%ri_grad%P_ij = 0.0_dp
            mp2_env%ri_grad%P_ab = 0.0_dp
            ALLOCATE (mp2_env%ri_grad%Gamma_P_ia(homo, my_B_size, my_group_L_size))
            mp2_env%ri_grad%Gamma_P_ia = 0.0_dp
         ELSE
            IF (beta_beta) THEN
               ALLOCATE (mp2_env%ri_grad%P_ij_beta(homo, homo))
               ALLOCATE (mp2_env%ri_grad%P_ab_beta(my_B_size, virtual))
               mp2_env%ri_grad%P_ij_beta = 0.0_dp
               mp2_env%ri_grad%P_ab_beta = 0.0_dp
               ALLOCATE (mp2_env%ri_grad%Gamma_P_ia_beta(homo, my_B_size_beta, my_group_L_size))
               mp2_env%ri_grad%Gamma_P_ia_beta = 0.0_dp
            ENDIF
         ENDIF
         IF (.NOT. alpha_beta) THEN
            ! For non-alpha-beta case we need amplitudes
            ALLOCATE (t_ab(virtual, my_B_size_beta))
         ELSE
            ! We need more integrals
            ALLOCATE (local_ba(virtual_beta, my_B_size))
         ENDIF
      END IF
      !

      CALL timestop(handle)

   END SUBROUTINE mp2_ri_allocate

! **************************************************************************************************
!> \brief ...
!> \param my_alpha_beta_case ...
!> \param total_ij_pairs ...
!> \param homo ...
!> \param homo_beta ...
!> \param num_IJ_blocks ...
!> \param block_size ...
!> \param ngroup ...
!> \param ij_map ...
!> \param color_sub ...
!> \param my_ij_pairs ...
!> \param my_open_shell_SS ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE mp2_ri_communication(my_alpha_beta_case, total_ij_pairs, homo, homo_beta, num_IJ_blocks, &
                                   block_size, ngroup, ij_map, color_sub, my_ij_pairs, my_open_shell_SS, unit_nr)
      LOGICAL                                            :: my_alpha_beta_case
      INTEGER                                            :: total_ij_pairs, homo, homo_beta, &
                                                            num_IJ_blocks, block_size, ngroup
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ij_map
      INTEGER                                            :: color_sub, my_ij_pairs
      LOGICAL                                            :: my_open_shell_SS
      INTEGER                                            :: unit_nr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_ri_communication', &
         routineP = moduleN//':'//routineN

      INTEGER :: assigned_blocks, first_I_block, first_J_block, handle, iiB, ij_block_counter, &
         ij_counter, jjB, last_i_block, last_J_block, num_block_per_group, total_ij_block, &
         total_ij_pairs_blocks
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ij_marker

! Calculate the maximum number of ij pairs that have to be computed
! among groups

      CALL timeset(routineN, handle)

      IF (.NOT. my_alpha_beta_case) THEN
         total_ij_pairs = homo*(1+homo)/2
         num_IJ_blocks = homo/block_size-1

         first_I_block = 1
         last_i_block = block_size*(num_IJ_blocks-1)

         first_J_block = block_size+1
         last_J_block = block_size*(num_IJ_blocks+1)

         ij_block_counter = 0
         DO iiB = first_I_block, last_i_block, block_size
            DO jjB = iiB+block_size, last_J_block, block_size
               ij_block_counter = ij_block_counter+1
            END DO
         END DO

         total_ij_block = ij_block_counter
         num_block_per_group = total_ij_block/ngroup
         assigned_blocks = num_block_per_group*ngroup

         total_ij_pairs_blocks = assigned_blocks+(total_ij_pairs-assigned_blocks*(block_size**2))

         ALLOCATE (ij_marker(homo, homo))
         ij_marker = 0
         ALLOCATE (ij_map(total_ij_pairs_blocks, 3))
         ij_map = 0
         ij_counter = 0
         my_ij_pairs = 0
         DO iiB = first_I_block, last_i_block, block_size
            DO jjB = iiB+block_size, last_J_block, block_size
               IF (ij_counter+1 > assigned_blocks) EXIT
               ij_counter = ij_counter+1
               ij_marker(iiB:iiB+block_size-1, jjB:jjB+block_size-1) = 1
               ij_map(ij_counter, 1) = iiB
               ij_map(ij_counter, 2) = jjB
               ij_map(ij_counter, 3) = block_size
               IF (MOD(ij_counter, ngroup) == color_sub) my_ij_pairs = my_ij_pairs+1
            END DO
         END DO
         DO iiB = 1, homo
            DO jjB = iiB, homo
               IF (ij_marker(iiB, jjB) == 0) THEN
                  ij_counter = ij_counter+1
                  ij_map(ij_counter, 1) = iiB
                  ij_map(ij_counter, 2) = jjB
                  ij_map(ij_counter, 3) = 1
                  IF (MOD(ij_counter, ngroup) == color_sub) my_ij_pairs = my_ij_pairs+1
               END IF
            END DO
         END DO
         DEALLOCATE (ij_marker)

         IF ((.NOT. my_open_shell_SS)) THEN
            IF (unit_nr > 0) THEN
               IF (block_size == 1) THEN
                  WRITE (UNIT=unit_nr, FMT="(T3,A,T66,F15.1)") &
                     "RI_INFO| Percentage of ij pairs communicated with block size 1:", 100.0_dp
               ELSE
                  WRITE (UNIT=unit_nr, FMT="(T3,A,T66,F15.1)") &
                     "RI_INFO| Percentage of ij pairs communicated with block size 1:", &
                     100.0_dp*REAL((total_ij_pairs-assigned_blocks*(block_size**2)), KIND=dp)/REAL(total_ij_pairs, KIND=dp)
               END IF
               CALL m_flush(unit_nr)
            END IF
         END IF

      ELSE
         ! alpha-beta case no index symmetry
         total_ij_pairs = homo*homo_beta
         ALLOCATE (ij_map(total_ij_pairs, 3))
         ij_map = 0
         ij_counter = 0
         my_ij_pairs = 0
         DO iiB = 1, homo
            DO jjB = 1, homo_beta
               ij_counter = ij_counter+1
               ij_map(ij_counter, 1) = iiB
               ij_map(ij_counter, 2) = jjB
               ij_map(ij_counter, 3) = 1
               IF (MOD(ij_counter, ngroup) == color_sub) my_ij_pairs = my_ij_pairs+1
            END DO
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE mp2_ri_communication

! **************************************************************************************************
!> \brief ...
!> \param BIb_C ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param homo ...
!> \param color_sub ...
!> \param sizes_array ...
!> \param calc_forces ...
!> \param comm_exchange ...
!> \param integ_group_size ...
!> \param my_B_size ...
!> \param iiB ...
!> \param my_group_L_end ...
!> \param my_group_L_size ...
!> \param my_group_L_size_orig ...
!> \param my_group_L_start ...
!> \param my_new_group_L_size ...
!> \param sub_sub_color ...
!> \param integ_group_pos2color_sub ...
!> \param new_sizes_array ...
!> \param proc_map ...
!> \param proc_map_rep ...
!> \param sizes_array_orig ...
!> \param sub_proc_map ...
!> \param ranges_info_array ...
!> \param para_env_exchange ...
!> \param para_env_rep ...
!> \param num_integ_group ...
! **************************************************************************************************
   SUBROUTINE mp2_ri_create_group(BIb_C, para_env, para_env_sub, homo, color_sub, &
                                  sizes_array, calc_forces, &
                                  comm_exchange, integ_group_size, my_B_size, iiB, my_group_L_end, &
                                  my_group_L_size, my_group_L_size_orig, my_group_L_start, my_new_group_L_size, &
                                  sub_sub_color, integ_group_pos2color_sub, new_sizes_array, &
                                  proc_map, proc_map_rep, sizes_array_orig, &
                                  sub_proc_map, ranges_info_array, para_env_exchange, para_env_rep, num_integ_group)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER                                            :: homo, color_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sizes_array
      LOGICAL, INTENT(IN)                                :: calc_forces
      INTEGER :: comm_exchange, integ_group_size, my_B_size, iiB, my_group_L_end, my_group_L_size, &
         my_group_L_size_orig, my_group_L_start, my_new_group_L_size, sub_sub_color
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: integ_group_pos2color_sub, &
                                                            new_sizes_array, proc_map, &
                                                            proc_map_rep, sizes_array_orig, &
                                                            sub_proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: ranges_info_array
      TYPE(cp_para_env_type), POINTER                    :: para_env_exchange, para_env_rep
      INTEGER                                            :: num_integ_group

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_ri_create_group', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

!

      CALL timeset(routineN, handle)
      !
      sub_sub_color = para_env_sub%mepos*num_integ_group+color_sub/integ_group_size
      CALL mp_comm_split_direct(para_env%group, comm_exchange, sub_sub_color)
      NULLIFY (para_env_exchange)
      CALL cp_para_env_create(para_env_exchange, comm_exchange)

      ! create the proc maps
      ALLOCATE (proc_map(-para_env_exchange%num_pe:2*para_env_exchange%num_pe-1))
      DO i = 0, para_env_exchange%num_pe-1
         proc_map(i) = i
         proc_map(-i-1) = para_env_exchange%num_pe-i-1
         proc_map(para_env_exchange%num_pe+i) = i
      END DO

      ALLOCATE (sub_proc_map(-para_env_sub%num_pe:2*para_env_sub%num_pe-1))
      DO i = 0, para_env_sub%num_pe-1
         sub_proc_map(i) = i
         sub_proc_map(-i-1) = para_env_sub%num_pe-i-1
         sub_proc_map(para_env_sub%num_pe+i) = i
      END DO

      CALL replicate_iaK_2intgroup(BIb_C, para_env, para_env_sub, para_env_exchange, para_env_rep, &
                                   homo, proc_map_rep, &
                                   sizes_array, &
                                   my_B_size, &
                                   my_group_L_size, my_group_L_start, my_group_L_end, &
                                   my_new_group_L_size, new_sizes_array, ranges_info_array)

      ALLOCATE (integ_group_pos2color_sub(0:para_env_exchange%num_pe-1))
      integ_group_pos2color_sub = 0
      integ_group_pos2color_sub(para_env_exchange%mepos) = color_sub
      CALL mp_sum(integ_group_pos2color_sub, para_env_exchange%group)

      IF (calc_forces) THEN
         iiB = SIZE(sizes_array)
         ALLOCATE (sizes_array_orig(0:iiB-1))
         sizes_array_orig(:) = sizes_array
      END IF

      my_group_L_size_orig = my_group_L_size
      my_group_L_size = my_new_group_L_size
      DEALLOCATE (sizes_array)

      ALLOCATE (sizes_array(0:integ_group_size-1))
      sizes_array(:) = new_sizes_array

      DEALLOCATE (new_sizes_array)
      !
      CALL timestop(handle)

   END SUBROUTINE mp2_ri_create_group

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param ends_array ...
!> \param ends_B_virtual ...
!> \param sizes_array ...
!> \param sizes_B_virtual ...
!> \param starts_array ...
!> \param starts_B_virtual ...
!> \param homo ...
!> \param dimen_RI ...
!> \param unit_nr ...
!> \param color_sub ...
!> \param best_block_size ...
!> \param best_integ_group_size ...
!> \param block_size ...
!> \param integ_group_size ...
!> \param min_integ_group_size ...
!> \param my_B_size ...
!> \param my_B_virtual_end ...
!> \param my_B_virtual_start ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
!> \param my_group_L_end ...
!> \param ngroup ...
!> \param num_IJ_blocks ...
!> \param num_integ_group ...
!> \param pos_integ_group ...
!> \param virtual ...
!> \param my_alpha_beta_case ...
!> \param my_open_shell_SS ...
!> \param mem_for_aK ...
!> \param mem_for_comm ...
!> \param mem_for_iaK ...
!> \param mem_for_rep ...
!> \param mem_min ...
!> \param mem_per_group ...
!> \param mem_real ...
! **************************************************************************************************
   SUBROUTINE mp2_ri_get_sizes(mp2_env, para_env, para_env_sub, ends_array, ends_B_virtual, sizes_array, sizes_B_virtual, &
                               starts_array, starts_B_virtual, homo, dimen_RI, unit_nr, color_sub, &
                               best_block_size, best_integ_group_size, block_size, &
                               integ_group_size, min_integ_group_size, my_B_size, &
                               my_B_virtual_end, my_B_virtual_start, my_group_L_size, &
                               my_group_L_start, my_group_L_end, ngroup, num_IJ_blocks, num_integ_group, &
                               pos_integ_group, virtual, my_alpha_beta_case, &
                               my_open_shell_SS, mem_for_aK, mem_for_comm, &
                               mem_for_iaK, mem_for_rep, mem_min, mem_per_group, mem_real)
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_array, ends_B_virtual, sizes_array, &
                                                            sizes_B_virtual, starts_array, &
                                                            starts_B_virtual
      INTEGER :: homo, dimen_RI, unit_nr, color_sub, best_block_size, best_integ_group_size, &
         block_size, integ_group_size, min_integ_group_size, my_B_size, my_B_virtual_end, &
         my_B_virtual_start, my_group_L_size, my_group_L_start, my_group_L_end, ngroup, &
         num_IJ_blocks, num_integ_group, pos_integ_group, virtual
      LOGICAL                                            :: my_alpha_beta_case, my_open_shell_SS
      REAL(KIND=dp)                                      :: mem_for_aK, mem_for_comm, mem_for_iaK, &
                                                            mem_for_rep, mem_min, mem_per_group, &
                                                            mem_real

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_ri_get_sizes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iiB
      INTEGER(KIND=int_8)                                :: mem

      CALL timeset(routineN, handle)

      ngroup = para_env%num_pe/para_env_sub%num_pe

      ! Calculate available memory and create integral group according to that
      ! mem_for_iaK is the memory needed for storing the 3 centre integrals
      mem_for_iaK = REAL(homo, KIND=dp)*virtual*dimen_RI*8.0_dp/(1024_dp**2)
      mem_for_aK = REAL(virtual, KIND=dp)*dimen_RI*8.0_dp/(1024_dp**2)

      CALL m_memory(mem)
      mem_real = (mem+1024*1024-1)/(1024*1024)
      ! mp_min .... a hack.. it should be mp_max, but as it turns out, on some processes the previously freed memory (hfx)
      ! has not been given back to the OS yet.
      CALL mp_min(mem_real, para_env%group)

      mem_min = 2.0_dp*REAL(homo, KIND=dp)*MAXVAL(sizes_B_virtual)*MAXVAL(sizes_array)*8.0_dp/(1024**2)
      mem_min = mem_min+3.0_dp*MAXVAL(sizes_B_virtual)*REAL(dimen_RI, KIND=dp)*8.0_dp/(1024**2)

      IF ((.NOT. my_open_shell_SS) .AND. (.NOT. my_alpha_beta_case)) THEN
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T68,F9.2,A4)') 'RI_INFO| Minimum required memory per MPI process:', &
            mem_min, ' MiB'
      END IF

      mem_real = mp2_env%mp2_memory

      mem_per_group = mem_real*para_env_sub%num_pe

      ! here we try to find the best block_size and integ_group_size
      best_integ_group_size = ngroup
      best_block_size = 1

      ! in the open shell case no replication and no block communication is done
      IF ((.NOT. my_open_shell_SS) .AND. (.NOT. my_alpha_beta_case)) THEN
         ! Here we split the memory half for the communication, half for replication
         IF (mp2_env%ri_mp2%block_size > 0) THEN
            best_block_size = mp2_env%ri_mp2%block_size
            mem_for_rep = MAX(mem_min, mem_per_group-2.0_dp*mem_for_aK*best_block_size)
         ELSE
            mem_for_rep = mem_per_group/2.0_dp
         END IF
         ! calculate the minimum replication group size according to the available memory
         min_integ_group_size = CEILING(2.0_dp*mem_for_iaK/mem_for_rep)

         integ_group_size = MIN(min_integ_group_size, ngroup)-1
         DO iiB = min_integ_group_size+1, ngroup
            integ_group_size = integ_group_size+1
            ! check that the ngroup is a multiple of  integ_group_size
            IF (MOD(ngroup, integ_group_size) /= 0) CYCLE
            ! check that the integ group size is not too small (10% is empirical for now)
            IF (REAL(integ_group_size, KIND=dp)/REAL(ngroup, KIND=dp) < 0.1_dp) CYCLE

            best_integ_group_size = integ_group_size
            EXIT
         END DO

         IF (.NOT. (mp2_env%ri_mp2%block_size > 0)) THEN
            mem_for_comm = mem_per_group-2.0_dp*mem_for_iaK/best_integ_group_size
            !k_size=MAX(INT(mem_for_comm/(2.0_dp*mem_for_aK)),1)
            DO
               num_IJ_blocks = (homo/best_block_size)
               num_IJ_blocks = (num_IJ_blocks*num_IJ_blocks-num_IJ_blocks)/2
               IF (num_IJ_blocks > ngroup .OR. best_block_size == 1) THEN
                  EXIT
               ELSE
                  best_block_size = best_block_size-1
               END IF
            END DO
         END IF

         ! check that best_block_size is not bigger than homo/2-1
         best_block_size = MIN(MAX(homo/2-1+MOD(homo, 2), 1), best_block_size)
      END IF

      integ_group_size = best_integ_group_size
      block_size = best_block_size

      IF ((.NOT. my_open_shell_SS) .AND. (.NOT. my_alpha_beta_case)) THEN
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "RI_INFO| Group size for integral replication:", integ_group_size*para_env_sub%num_pe
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "RI_INFO| Block size:", block_size
            CALL m_flush(unit_nr)
         END IF
      END IF

      num_integ_group = ngroup/integ_group_size

      pos_integ_group = MOD(color_sub, integ_group_size)

      my_group_L_size = sizes_array(color_sub)
      my_group_L_start = starts_array(color_sub)
      my_group_L_end = ends_array(color_sub)

      my_B_virtual_start = starts_B_virtual(para_env_sub%mepos)
      my_B_virtual_end = ends_B_virtual(para_env_sub%mepos)
      my_B_size = sizes_B_virtual(para_env_sub%mepos)

      CALL timestop(handle)

   END SUBROUTINE mp2_ri_get_sizes

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
!> \param para_env_sub ...
!> \param ends_B_virtual ...
!> \param sizes_B_virtual ...
!> \param starts_B_virtual ...
!> \param Eigenval ...
!> \param homo ...
!> \param dimen_RI ...
!> \param iiB ...
!> \param jjB ...
!> \param my_B_size ...
!> \param my_B_virtual_end ...
!> \param my_B_virtual_start ...
!> \param my_i ...
!> \param my_j ...
!> \param virtual ...
!> \param sub_proc_map ...
!> \param local_ab ...
!> \param t_ab ...
!> \param local_i_aL ...
!> \param local_j_aL ...
!> \param open_ss ...
!> \param alpha_alpha ...
!> \param beta_beta ...
!> \param Y_i_aP ...
!> \param Y_j_aP ...
!> \param eigenval_beta ...
!> \param homo_beta ...
!> \param my_B_size_beta ...
!> \param ends_B_virtual_beta ...
!> \param sizes_B_virtual_beta ...
!> \param starts_B_virtual_beta ...
!> \param my_B_virtual_start_beta ...
!> \param my_B_virtual_end_beta ...
!> \param virtual_beta ...
!> \param local_ba ...
! **************************************************************************************************
   SUBROUTINE mp2_update_P_gamma(mp2_env, para_env_sub, ends_B_virtual, sizes_B_virtual, starts_B_virtual, &
                                 Eigenval, homo, dimen_RI, iiB, jjB, my_B_size, &
                                 my_B_virtual_end, my_B_virtual_start, my_i, my_j, virtual, sub_proc_map, local_ab, &
                                 t_ab, local_i_aL, local_j_aL, open_ss, alpha_alpha, beta_beta, Y_i_aP, Y_j_aP, &
                                 eigenval_beta, homo_beta, my_B_size_beta, ends_B_virtual_beta, sizes_B_virtual_beta, &
                                 starts_B_virtual_beta, my_B_virtual_start_beta, my_B_virtual_end_beta, virtual_beta, local_ba)
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_B_virtual, sizes_B_virtual, &
                                                            starts_B_virtual
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: homo, dimen_RI, iiB, jjB, my_B_size, &
                                                            my_B_virtual_end, my_B_virtual_start, &
                                                            my_i, my_j, virtual
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sub_proc_map
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: local_ab, t_ab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: local_i_aL, local_j_aL
      LOGICAL                                            :: open_ss, alpha_alpha, beta_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: Y_i_aP, Y_j_aP
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      INTEGER, OPTIONAL                                  :: homo_beta, my_B_size_beta
      INTEGER, ALLOCATABLE, DIMENSION(:), OPTIONAL       :: ends_B_virtual_beta, &
                                                            sizes_B_virtual_beta, &
                                                            starts_B_virtual_beta
      INTEGER, OPTIONAL                                  :: my_B_virtual_start_beta, &
                                                            my_B_virtual_end_beta, virtual_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :), &
         OPTIONAL                                        :: local_ba

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_update_P_gamma', &
         routineP = moduleN//':'//routineN

      INTEGER :: a, b, b_global, handle, proc_receive, proc_send, proc_shift, rec_B_size, &
         rec_B_virtual_end, rec_B_virtual_start, send_B_size, send_B_virtual_end, &
         send_B_virtual_start
      LOGICAL                                            :: alpha_beta
      REAL(KIND=dp)                                      :: factor, P_ij_diag
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: external_ab, send_ab

!
!  In alpha-beta case Y_j_aP_beta is sent and is received as Y_j_aP
!

      CALL timeset(routineN//"_Pia", handle)
! Find out whether we have an alpha-beta case
      alpha_beta = .FALSE.
      IF (PRESENT(Eigenval_beta) .AND. PRESENT(homo_beta) .AND. PRESENT(my_B_size_beta)) &
         alpha_beta = .TRUE.
! update P_ab, Gamma_P_ia
! First, P_ab
      IF (open_ss) THEN
         factor = 1.0_dp
      ELSE
         factor = 2.0_dp
      ENDIF
      ! divide the (ia|jb) integrals by Delta_ij^ab
      IF (.NOT. alpha_beta) THEN
         DO b = 1, my_B_size
            b_global = b+my_B_virtual_start-1
            DO a = 1, virtual
               local_ab(a, b) = -local_ab(a, b)/ &
                                (Eigenval(homo+a)+Eigenval(homo+b_global)-Eigenval(my_i+iiB-1)-Eigenval(my_j+jjB-1))
            END DO
         END DO
         ! update diagonal part of P_ij
         P_ij_diag = -SUM(local_ab*t_ab)*factor
      ELSE
         DO b = 1, my_B_size_beta
            b_global = b+my_B_virtual_start_beta-1
            DO a = 1, virtual
               local_ab(a, b) = -local_ab(a, b)/ &
                                (Eigenval(homo+a)+Eigenval_beta(homo_beta+b_global)-Eigenval(my_i+iiB-1)-Eigenval_beta(my_j+jjB-1))
            END DO
         END DO
         ! update diagonal part of P_ij
         P_ij_diag = -SUM(local_ab*local_ab)
         ! More integrals needed only for alpha-beta case: local_ba
         DO b = 1, my_B_size
            b_global = b+my_B_virtual_start-1
            DO a = 1, virtual_beta
               local_ba(a, b) = -local_ba(a, b)/ &
                                (Eigenval_beta(homo_beta+a)+Eigenval(homo+b_global)-Eigenval(my_i+iiB-1)-Eigenval_beta(my_j+jjB-1))
            END DO
         END DO
      ENDIF

      ! P_ab and add diagonal part of P_ij

      ! Alpha_alpha or closed-shell case
      IF (((.NOT. open_ss) .AND. (.NOT. alpha_beta)) .OR. alpha_alpha) THEN
         CALL dgemm('T', 'N', my_B_size, my_B_size, virtual, 1.0_dp, &
                    t_ab(:, :), virtual, local_ab(:, :), virtual, &
                    1.0_dp, mp2_env%ri_grad%P_ab(1:my_B_size, my_B_virtual_start:my_B_virtual_end), my_B_size)
         mp2_env%ri_grad%P_ij(my_i+iiB-1, my_i+iiB-1) = &
            mp2_env%ri_grad%P_ij(my_i+iiB-1, my_i+iiB-1)+P_ij_diag
      ENDIF
      ! Beta_beta case
      IF (beta_beta) THEN
         CALL dgemm('T', 'N', my_B_size, my_B_size, virtual, 1.0_dp, &
                    t_ab(:, :), virtual, local_ab(:, :), virtual, &
                    1.0_dp, mp2_env%ri_grad%P_ab_beta(1:my_B_size, my_B_virtual_start:my_B_virtual_end), my_B_size)
         mp2_env%ri_grad%P_ij_beta(my_i+iiB-1, my_i+iiB-1) = &
            mp2_env%ri_grad%P_ij_beta(my_i+iiB-1, my_i+iiB-1)+P_ij_diag
      ENDIF
      ! Alpha_beta case
      IF (alpha_beta) THEN
         CALL dgemm('T', 'N', my_B_size, my_B_size, virtual_beta, 1.0_dp, &
                    local_ba(:, :), virtual_beta, local_ba(:, :), virtual_beta, 1.0_dp, &
                    mp2_env%ri_grad%P_ab(1:my_B_size, my_B_virtual_start: &
                                         my_B_virtual_end), my_B_size)
         mp2_env%ri_grad%P_ij(my_i+iiB-1, my_i+iiB-1) = &
            mp2_env%ri_grad%P_ij(my_i+iiB-1, my_i+iiB-1)+P_ij_diag
         CALL dgemm('T', 'N', my_B_size_beta, my_B_size_beta, virtual, 1.0_dp, &
                    local_ab(:, :), virtual, local_ab(:, :), virtual, 1.0_dp, &
                    mp2_env%ri_grad%P_ab_beta(1:my_B_size_beta, my_B_virtual_start_beta: &
                                              my_B_virtual_end_beta), my_B_size_beta)
         mp2_env%ri_grad%P_ij_beta(my_j+jjB-1, my_j+jjB-1) = &
            mp2_env%ri_grad%P_ij_beta(my_j+jjB-1, my_j+jjB-1)+P_ij_diag
      ENDIF
      ! The summation is over unique pairs. In alpha-beta case, all pairs are unique: subroutine is called for
      ! both i^alpha,j^beta and i^beta,j^alpha. Formally, my_i can be equal to my_j, but they are different
      ! due to spin in alpha-beta case.
      IF ((my_i /= my_j) .AND. (.NOT. alpha_beta)) THEN
         ! CALL dgemm('N','T',virtual,virtual,my_B_size,1.0_dp,&
         !                t_ab(:,:),virtual,local_ab(:,:),virtual,&
         !                1.0_dp,mp2_env%ri_grad%P_ab(:,:),virtual)

         ! Alpha_alpha or closed-shell case
         IF (((.NOT. open_ss) .AND. (.NOT. alpha_beta)) .OR. alpha_alpha) THEN
            CALL dgemm('N', 'T', my_B_size, virtual, my_B_size, 1.0_dp, &
                       t_ab(my_B_virtual_start:my_B_virtual_end, 1:my_B_size), my_B_size, &
                       local_ab(:, :), virtual, &
                       1.0_dp, mp2_env%ri_grad%P_ab(1:my_B_size, 1:virtual), my_B_size)
            ! CHECK CORRECTNESS IF BLOCK SIZE greater than 1
            mp2_env%ri_grad%P_ij(my_j+jjB-1, my_j+jjB-1) = &
               mp2_env%ri_grad%P_ij(my_j+jjB-1, my_j+jjB-1)+P_ij_diag
         ENDIF
         ! Beta_beta_case
         IF (beta_beta) THEN
            CALL dgemm('N', 'T', my_B_size, virtual, my_B_size, 1.0_dp, &
                       t_ab(my_B_virtual_start:my_B_virtual_end, 1:my_B_size), my_B_size, &
                       local_ab(:, :), virtual, &
                       1.0_dp, mp2_env%ri_grad%P_ab_beta(1:my_B_size, 1:virtual), my_B_size)
            ! CHECK CORRECTNESS IF BLOCK SIZE greater than 1
            mp2_env%ri_grad%P_ij_beta(my_j+jjB-1, my_j+jjB-1) = &
               mp2_env%ri_grad%P_ij_beta(my_j+jjB-1, my_j+jjB-1)+P_ij_diag
         ENDIF
      END IF
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
         ! Alpha-alpha, beta-beta, closed shell
         IF (.NOT. alpha_beta) THEN
            rec_B_size = sizes_B_virtual(proc_receive)
            rec_B_virtual_end = ends_B_virtual(proc_receive)
            rec_B_virtual_start = starts_B_virtual(proc_receive)
            send_B_size = sizes_B_virtual(proc_send)
            send_B_virtual_end = ends_B_virtual(proc_send)
            send_B_virtual_start = starts_B_virtual(proc_send)
         ELSE ! Alpha-beta case
            rec_B_size = sizes_B_virtual_beta(proc_receive)
            rec_B_virtual_end = ends_B_virtual_beta(proc_receive)
            rec_B_virtual_start = starts_B_virtual_beta(proc_receive)
            send_B_size = sizes_B_virtual_beta(proc_send)
            send_B_virtual_end = ends_B_virtual_beta(proc_send)
            send_B_virtual_start = starts_B_virtual_beta(proc_send)
         ENDIF

         ALLOCATE (external_ab(virtual, rec_B_size))
         external_ab = 0.0_dp

         IF (.NOT. alpha_beta) THEN
            CALL mp_sendrecv(local_ab(1:virtual, 1:my_B_size), proc_send, &
                             external_ab(1:virtual, 1:rec_B_size), proc_receive, &
                             para_env_sub%group)
         ELSE
            CALL mp_sendrecv(local_ab(1:virtual, 1:my_B_size_beta), proc_send, &
                             external_ab(1:virtual, 1:rec_B_size), proc_receive, &
                             para_env_sub%group)
         ENDIF

         ! Alpha-alpha or closed-shell case
         IF (((.NOT. open_ss) .AND. (.NOT. alpha_beta)) .OR. alpha_alpha) &
            CALL dgemm('T', 'N', my_B_size, rec_B_size, virtual, 1.0_dp, &
                       t_ab(:, :), virtual, external_ab(:, :), virtual, &
                       1.0_dp, mp2_env%ri_grad%P_ab(1:my_B_size, rec_B_virtual_start:rec_B_virtual_end), my_B_size)
         ! Beta-beta case
         IF (beta_beta) &
            CALL dgemm('T', 'N', my_B_size, rec_B_size, virtual, 1.0_dp, &
                       t_ab(:, :), virtual, external_ab(:, :), virtual, &
                       1.0_dp, mp2_env%ri_grad%P_ab_beta(1:my_B_size, rec_B_virtual_start:rec_B_virtual_end), my_B_size)
         ! Alpha-beta case
         IF (alpha_beta) THEN
            !   CALL dgemm('N','T',my_B_size,virtual,rec_B_size,1.0_dp,&
            !             local_ab(1:my_B_size,rec_B_virtual_start:rec_B_virtual_end),my_B_size,external_ab(:,:),rec_B_size,&
            !             1.0_dp,mp2_env%ri_grad%P_ab(1:my_B_size,1:virtual),my_B_size)

            ! Alpha-beta part of beta-beta density
            CALL dgemm('T', 'N', my_B_size_beta, rec_B_size, virtual, 1.0_dp, &
                       local_ab(:, :), virtual, external_ab(:, :), virtual, &
                       1.0_dp, mp2_env%ri_grad%P_ab_beta(1:my_B_size_beta, rec_B_virtual_start:rec_B_virtual_end), &
                       my_B_size_beta)

            ! For alpha-beta part of alpha-density we need a new parallel code
            DEALLOCATE (external_ab)
            ! And new external_ab (of a different size)
            rec_B_size = sizes_B_virtual(proc_receive)
            rec_B_virtual_end = ends_B_virtual(proc_receive)
            rec_B_virtual_start = starts_B_virtual(proc_receive)
            send_B_size = sizes_B_virtual(proc_send)
            send_B_virtual_end = ends_B_virtual(proc_send)
            send_B_virtual_start = starts_B_virtual(proc_send)
            ALLOCATE (external_ab(virtual_beta, rec_B_size))
            external_ab = 0.0_dp
            CALL mp_sendrecv(local_ba(1:virtual_beta, 1:my_B_size), proc_send, &
                             external_ab(1:virtual_beta, 1:rec_B_size), proc_receive, &
                             para_env_sub%group)
            CALL dgemm('T', 'N', my_B_size, rec_B_size, virtual_beta, 1.0_dp, &
                       local_ba(:, :), virtual_beta, external_ab(:, :), virtual_beta, &
                       1.0_dp, mp2_env%ri_grad%P_ab(1:my_B_size, rec_B_virtual_start:rec_B_virtual_end), my_B_size)
         ENDIF

         DEALLOCATE (external_ab)

         IF ((my_i /= my_j) .AND. (.NOT. alpha_beta)) THEN
            ALLOCATE (external_ab(my_B_size, virtual))
            external_ab = 0.0_dp

            ALLOCATE (send_ab(send_B_size, virtual))
            send_ab = 0.0_dp

            CALL dgemm('N', 'T', send_B_size, virtual, my_B_size, 1.0_dp, &
                       t_ab(send_B_virtual_start:send_B_virtual_end, 1:my_B_size), send_B_size, &
                       local_ab(:, :), virtual, &
                       0.0_dp, send_ab(1:send_B_size, 1:virtual), send_B_size)

            CALL mp_sendrecv(send_ab, proc_send, &
                             external_ab, proc_receive, &
                             para_env_sub%group)

            ! Alpha_alpha or closed-shell case
            IF (((.NOT. open_ss) .AND. (.NOT. alpha_beta)) .OR. alpha_alpha) &
               mp2_env%ri_grad%P_ab(:, :) = mp2_env%ri_grad%P_ab+external_ab
            ! Beta_beta case
            IF (beta_beta) &
               mp2_env%ri_grad%P_ab_beta(:, :) = mp2_env%ri_grad%P_ab_beta+external_ab

            DEALLOCATE (external_ab)
            DEALLOCATE (send_ab)
         END IF

      END DO
      CALL timestop(handle)

      ! Now, Gamma_P_ia (made of Y_ia_P)

      CALL timeset(routineN//"_Gamma", handle)
      ! CALL dgemm('T','N',my_B_size,my_B_size_beta,dimen_RI,1.0_dp,&
      !                local_i_aL(:,:,iiB),dimen_RI,local_j_aL(:,:,jjB),dimen_RI,&
      !                0.0_dp,local_ab(my_B_virtual_start:my_B_virtual_end,1:my_B_size_beta),my_B_size)
      IF (.NOT. alpha_beta) THEN
         ! Alpha-alpha, beta-beta and closed shell
         CALL dgemm('N', 'T', my_B_size, dimen_RI, my_B_size, 1.0_dp, &
                    t_ab(my_B_virtual_start:my_B_virtual_end, 1:my_B_size), my_B_size, &
                    local_j_aL(1:dimen_RI, 1:my_B_size, jjB), dimen_RI, &
                    1.0_dp, Y_i_aP(1:my_B_size, 1:dimen_RI, iiB), my_B_size)
      ELSE ! Alpha-beta
         CALL dgemm('N', 'T', my_B_size, dimen_RI, my_B_size_beta, 1.0_dp, &
                    local_ab(my_B_virtual_start:my_B_virtual_end, 1:my_B_size_beta), my_B_size, &
                    local_j_aL(1:dimen_RI, 1:my_B_size_beta, jjB), dimen_RI, &
                    1.0_dp, Y_i_aP(1:my_B_size, 1:dimen_RI, iiB), my_B_size)
         CALL dgemm('T', 'T', my_B_size_beta, dimen_RI, my_B_size, 1.0_dp, &
                    local_ab(my_B_virtual_start:my_B_virtual_end, 1:my_B_size_beta), my_B_size, &
                    local_i_aL(1:dimen_RI, 1:my_B_size, jjB), dimen_RI, &
                    1.0_dp, Y_j_aP(1:my_B_size_beta, 1:dimen_RI, jjB), my_B_size_beta)

      ENDIF

      ALLOCATE (external_ab(my_B_size, dimen_RI))
      external_ab = 0.0_dp
      !
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

         rec_B_size = sizes_B_virtual(proc_receive)
         rec_B_virtual_end = ends_B_virtual(proc_receive)
         rec_B_virtual_start = starts_B_virtual(proc_receive)

         send_B_size = sizes_B_virtual(proc_send)
         send_B_virtual_end = ends_B_virtual(proc_send)
         send_B_virtual_start = starts_B_virtual(proc_send)

         IF (.NOT. alpha_beta) THEN
            ALLOCATE (send_ab(send_B_size, dimen_RI))
            send_ab = 0.0_dp
            CALL dgemm('N', 'T', send_B_size, dimen_RI, my_B_size, 1.0_dp, &
                       t_ab(send_B_virtual_start:send_B_virtual_end, 1:my_B_size), send_B_size, &
                       local_j_aL(1:dimen_RI, 1:my_B_size, jjB), dimen_RI, &
                       0.0_dp, send_ab(1:send_B_size, 1:dimen_RI), send_B_size)
            CALL mp_sendrecv(send_ab, proc_send, external_ab, proc_receive, para_env_sub%group)
            ! Alpha-alpha, beta-beta and closed shell

            Y_i_aP(1:my_B_size, 1:dimen_RI, iiB) = &
               Y_i_aP(1:my_B_size, 1:dimen_RI, iiB)+external_ab

            DEALLOCATE (send_ab)
         ELSE ! Alpha-beta case
            ! Alpha-alpha part
            ALLOCATE (send_ab(send_B_size, dimen_RI))
            send_ab = 0.0_dp
            CALL dgemm('N', 'T', send_B_size, dimen_RI, my_B_size_beta, 1.0_dp, &
                       local_ab(send_B_virtual_start:send_B_virtual_end, 1:my_B_size_beta), send_B_size, &
                       local_j_aL(1:dimen_RI, 1:my_B_size_beta, jjB), dimen_RI, &
                       0.0_dp, send_ab(1:send_B_size, 1:dimen_RI), send_B_size)
            CALL mp_sendrecv(send_ab, proc_send, external_ab, proc_receive, para_env_sub%group)
            Y_i_aP(1:my_B_size, 1:dimen_RI, iiB) = &
               Y_i_aP(1:my_B_size, 1:dimen_RI, iiB)+external_ab
            DEALLOCATE (send_ab)
         ENDIF
      END DO
      DEALLOCATE (external_ab)

      IF (alpha_beta) THEN
         ! For beta-beta part (in alpha-beta case) we need a new parallel code
         ALLOCATE (external_ab(my_B_size_beta, dimen_RI))
         external_ab = 0.0_dp
         DO proc_shift = 1, para_env_sub%num_pe-1
            proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
            proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

            send_B_size = sizes_B_virtual_beta(proc_send)
            send_B_virtual_end = ends_B_virtual_beta(proc_send)
            send_B_virtual_start = starts_B_virtual_beta(proc_send)
            ALLOCATE (send_ab(send_B_size, dimen_RI))
            send_ab = 0.0_dp

            CALL dgemm('N', 'T', send_B_size, dimen_RI, my_B_size, 1.0_dp, &
                       local_ba(send_B_virtual_start:send_B_virtual_end, 1:my_B_size), send_B_size, &
                       local_i_aL(1:dimen_RI, 1:my_B_size, jjB), dimen_RI, &
                       0.0_dp, send_ab(1:send_B_size, 1:dimen_RI), send_B_size)
            CALL mp_sendrecv(send_ab, proc_send, external_ab, proc_receive, para_env_sub%group)
            Y_j_aP(1:my_B_size_beta, 1:dimen_RI, jjB) = &
               Y_j_aP(1:my_B_size_beta, 1:dimen_RI, jjB)+external_ab
            DEALLOCATE (send_ab)

         END DO
         DEALLOCATE (external_ab)
      ENDIF

      IF ((my_i /= my_j) .AND. (.NOT. alpha_beta)) THEN
         ! Alpha-alpha, beta-beta and closed shell
         CALL dgemm('T', 'T', my_B_size, dimen_RI, my_B_size, 1.0_dp, &
                    t_ab(my_B_virtual_start:my_B_virtual_end, 1:my_B_size), my_B_size, &
                    local_i_aL(1:dimen_RI, 1:my_B_size, iiB), dimen_RI, &
                    1.0_dp, Y_j_aP(1:my_B_size, 1:dimen_RI, jjB), my_B_size)

         DO proc_shift = 1, para_env_sub%num_pe-1
            proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
            proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

            rec_B_size = sizes_B_virtual(proc_receive)
            rec_B_virtual_end = ends_B_virtual(proc_receive)
            rec_B_virtual_start = starts_B_virtual(proc_receive)

            ALLOCATE (external_ab(dimen_RI, rec_B_size))
            external_ab = 0.0_dp

            CALL mp_sendrecv(local_i_aL(1:dimen_RI, 1:my_B_size, iiB), proc_send, &
                             external_ab, proc_receive, para_env_sub%group)

            ! Alpha-alpha, beta-beta and closed shell
            CALL dgemm('T', 'T', my_B_size, dimen_RI, rec_B_size, 1.0_dp, &
                       t_ab(rec_B_virtual_start:rec_B_virtual_end, 1:my_B_size), rec_B_size, &
                       external_ab(1:dimen_RI, 1:rec_B_size), dimen_RI, &
                       1.0_dp, Y_j_aP(1:my_B_size, 1:dimen_RI, jjB), my_B_size)

            DEALLOCATE (external_ab)
         END DO
      END IF

      ! Y_i_aP=0.0_dp
      ! Y_j_aP=0.0_dp

      CALL timestop(handle)

   END SUBROUTINE mp2_update_P_gamma

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
!> \param ij_index ...
!> \param my_B_size ...
!> \param my_block_size ...
!> \param my_group_L_size ...
!> \param my_i ...
!> \param my_ij_pairs ...
!> \param my_j ...
!> \param ngroup ...
!> \param num_integ_group ...
!> \param integ_group_pos2color_sub ...
!> \param num_ij_pairs ...
!> \param proc_map ...
!> \param ij_map ...
!> \param ranges_info_array ...
!> \param Y_i_aP ...
!> \param Y_j_aP ...
!> \param para_env_exchange ...
!> \param null_mat_rec ...
!> \param null_mat_send ...
!> \param sizes_array ...
!> \param alpha_alpha ...
!> \param beta_beta ...
!> \param alpha_beta ...
!> \param open_shell ...
!> \param my_b_size_beta ...
! **************************************************************************************************
   SUBROUTINE mp2_redistribute_gamma(mp2_env, ij_index, my_B_size, &
                                     my_block_size, my_group_L_size, my_i, my_ij_pairs, my_j, ngroup, &
                                     num_integ_group, integ_group_pos2color_sub, num_ij_pairs, proc_map, &
                                     ij_map, ranges_info_array, Y_i_aP, Y_j_aP, para_env_exchange, &
                                     null_mat_rec, null_mat_send, sizes_array, alpha_alpha, beta_beta, &
                                     alpha_beta, open_shell, my_b_size_beta)

      TYPE(mp2_type), POINTER                            :: mp2_env
      INTEGER                                            :: ij_index, my_B_size, my_block_size, &
                                                            my_group_L_size, my_i, my_ij_pairs, &
                                                            my_j, ngroup, num_integ_group
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: integ_group_pos2color_sub, num_ij_pairs, &
                                                            proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ij_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: ranges_info_array
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: Y_i_aP, Y_j_aP
      TYPE(cp_para_env_type), POINTER                    :: para_env_exchange
      REAL(KIND=dp)                                      :: null_mat_rec(:, :, :), &
                                                            null_mat_send(:, :, :)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sizes_array
      LOGICAL                                            :: alpha_alpha, beta_beta, alpha_beta, &
                                                            open_shell
      INTEGER, OPTIONAL                                  :: my_B_size_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_redistribute_gamma', &
         routineP = moduleN//':'//routineN

      INTEGER :: end_point, handle, handle2, iiB, ij_counter_rec, irep, jjb, kkk, Lend_pos, lll, &
         Lstart_pos, proc_receive, proc_send, proc_shift, rec_block_size, rec_i, rec_ij_index, &
         rec_j, send_L_size, start_point
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BI_C_rec, BI_C_rec_beta, BI_C_send, &
                                                            BI_C_send_beta

! In alpha-beta case Y_i_aP_beta is sent as Y_j_aP

      CALL timeset(routineN//"_comm2", handle)
      IF (ij_index <= my_ij_pairs) THEN
         ! somethig to send
         ! start with myself
         CALL timeset(routineN//"_comm2_w", handle2)
         DO irep = 0, num_integ_group-1
            Lstart_pos = ranges_info_array(1, irep, para_env_exchange%mepos)
            Lend_pos = ranges_info_array(2, irep, para_env_exchange%mepos)
            start_point = ranges_info_array(3, irep, para_env_exchange%mepos)
            end_point = ranges_info_array(4, irep, para_env_exchange%mepos)
            DO iiB = 1, my_block_size
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(kkk,lll) &
!$OMP                                     SHARED(start_point,end_point,Lstart_pos,Lend_pos,&
!$OMP                                            mp2_env,my_i,iiB,my_B_size,Y_i_aP,&
!$OMP                                     alpha_alpha,beta_beta,open_shell)
               DO kkk = start_point, end_point
                  lll = kkk-start_point+Lstart_pos
                  IF (alpha_alpha .OR. (.NOT. open_shell)) THEN
                     mp2_env%ri_grad%Gamma_P_ia(my_i+iiB-1, 1:my_B_size, kkk) = &
                        mp2_env%ri_grad%Gamma_P_ia(my_i+iiB-1, 1:my_B_size, kkk)+ &
                        Y_i_aP(1:my_B_size, lll, iiB)
                  ENDIF
                  IF (beta_beta) THEN
                     mp2_env%ri_grad%Gamma_P_ia_beta(my_i+iiB-1, 1:my_B_size, kkk) = &
                        mp2_env%ri_grad%Gamma_P_ia_beta(my_i+iiB-1, 1:my_B_size, kkk)+ &
                        Y_i_aP(1:my_B_size, lll, iiB)
                  ENDIF
               END DO
!$OMP           END PARALLEL DO
            END DO
            DO jjB = 1, my_block_size
               IF (.NOT. alpha_beta) THEN
!$OMP              PARALLEL DO DEFAULT(NONE) PRIVATE(kkk,lll) &
!$OMP                                        SHARED(start_point,end_point,Lstart_pos,Lend_pos,&
!$OMP                                               mp2_env,my_j,jjB,my_B_size,Y_j_aP,&
!$OMP                                        alpha_alpha,beta_beta,open_shell)
                  DO kkk = start_point, end_point
                     lll = kkk-start_point+Lstart_pos
                     IF (alpha_alpha .OR. (.NOT. open_shell)) THEN
                        mp2_env%ri_grad%Gamma_P_ia(my_j+jjB-1, 1:my_B_size, kkk) = &
                           mp2_env%ri_grad%Gamma_P_ia(my_j+jjB-1, 1:my_B_size, kkk)+ &
                           Y_j_aP(1:my_B_size, lll, jjB)
                     ENDIF
                     IF (beta_beta) THEN
                        mp2_env%ri_grad%Gamma_P_ia_beta(my_j+jjB-1, 1:my_B_size, kkk) = &
                           mp2_env%ri_grad%Gamma_P_ia_beta(my_j+jjB-1, 1:my_B_size, kkk)+ &
                           Y_j_aP(1:my_B_size, lll, jjB)
                     ENDIF
                  END DO
!$OMP              END PARALLEL DO
               ELSE
!$OMP              PARALLEL DO DEFAULT(NONE) PRIVATE(kkk,lll) &
!$OMP                                        SHARED(start_point,end_point,Lstart_pos,Lend_pos,&
!$OMP                                               mp2_env,my_j,jjB,my_B_size_beta,Y_j_aP)
                  DO kkk = start_point, end_point
                     lll = kkk-start_point+Lstart_pos
                     mp2_env%ri_grad%Gamma_P_ia_beta(my_j+jjB-1, 1:my_B_size_beta, kkk) = &
                        mp2_env%ri_grad%Gamma_P_ia_beta(my_j+jjB-1, 1:my_B_size_beta, kkk)+ &
                        Y_j_aP(1:my_B_size_beta, lll, jjB)
                  ENDDO
!$OMP              END PARALLEL DO
               ENDIF
            END DO
            ! DO iiB=1, my_block_size
            !   mp2_env%ri_grad%Gamma_P_ia(my_i+iiB-1,1:my_B_size,start_point:end_point)=&
            !   mp2_env%ri_grad%Gamma_P_ia(my_i+iiB-1,1:my_B_size,start_point:end_point)+&
            !   Y_i_aP(1:my_B_size,Lstart_pos:Lend_pos,iiB)
            ! END DO
            ! DO jjB=1, my_block_size
            !   mp2_env%ri_grad%Gamma_P_ia(my_j+jjB-1,1:my_B_size,start_point:end_point)=&
            !   mp2_env%ri_grad%Gamma_P_ia(my_j+jjB-1,1:my_B_size,start_point:end_point)+&
            !   Y_j_aP(1:my_B_size,Lstart_pos:Lend_pos,jjB)
            ! END DO
         END DO
         CALL timestop(handle2)

         ! Y_i_aP(my_B_size,dimen_RI,block_size)

         DO proc_shift = 1, para_env_exchange%num_pe-1
            proc_send = proc_map(para_env_exchange%mepos+proc_shift)
            proc_receive = proc_map(para_env_exchange%mepos-proc_shift)

            send_L_size = sizes_array(proc_send)
            IF (.NOT. alpha_beta) THEN
               ALLOCATE (BI_C_send(2*my_block_size, my_B_size, send_L_size))
            ELSE
               ALLOCATE (BI_C_send(my_block_size, my_B_size, send_L_size))
               ALLOCATE (BI_C_send_beta(my_block_size, my_B_size_beta, send_L_size))
            ENDIF
            CALL timeset(routineN//"_comm2_w", handle2)
            BI_C_send = 0.0_dp
            IF (alpha_beta) BI_C_send_beta = 0.0_dp
            DO irep = 0, num_integ_group-1
               Lstart_pos = ranges_info_array(1, irep, proc_send)
               Lend_pos = ranges_info_array(2, irep, proc_send)
               start_point = ranges_info_array(3, irep, proc_send)
               end_point = ranges_info_array(4, irep, proc_send)
               DO iiB = 1, my_block_size
!$OMP             PARALLEL DO DEFAULT(NONE) PRIVATE(kkk,lll) &
!$OMP                                       SHARED(start_point,end_point,Lstart_pos,Lend_pos,&
!$OMP                                              BI_C_send,iiB,my_B_size,Y_i_aP)
                  DO kkk = start_point, end_point
                     lll = kkk-start_point+Lstart_pos
                     BI_C_send(iiB, 1:my_B_size, kkk) = Y_i_aP(1:my_B_size, lll, iiB)
                  END DO
!$OMP             END PARALLEL DO
               END DO
               DO jjB = 1, my_block_size
                  IF (.NOT. alpha_beta) THEN
!$OMP                PARALLEL DO DEFAULT(NONE) PRIVATE(kkk,lll) &
!$OMP                                          SHARED(start_point,end_point,Lstart_pos,Lend_pos,my_block_size,&
!$OMP                                                 BI_C_send,jjB,my_B_size,Y_j_aP)
                     DO kkk = start_point, end_point
                        lll = kkk-start_point+Lstart_pos
                        BI_C_send(jjB+my_block_size, 1:my_B_size, kkk) = Y_j_aP(1:my_B_size, lll, jjB)
                     END DO
!$OMP                END PARALLEL DO
                  ELSE
!$OMP                PARALLEL DO DEFAULT(NONE) PRIVATE(kkk,lll) &
!$OMP                                          SHARED(start_point,end_point,Lstart_pos,Lend_pos,&
!$OMP                                                 BI_C_send_beta,jjB,my_B_size_beta,Y_j_aP)
                     DO kkk = start_point, end_point
                        lll = kkk-start_point+Lstart_pos
                        BI_C_send_beta(jjB, 1:my_B_size_beta, kkk) = Y_j_aP(1:my_B_size_beta, lll, jjB)
                     END DO
!$OMP                END PARALLEL DO
                  ENDIF
               END DO
               ! DO iiB=1, my_block_size
               !   BI_C_send(iiB,1:my_B_size,start_point:end_point)=&
               !      Y_i_aP(1:my_B_size,Lstart_pos:Lend_pos,iiB)
               ! END DO
               ! DO jjB=1, my_block_size
               !   BI_C_send(jjB+my_block_size,1:my_B_size,start_point:end_point)=&
               !      Y_j_aP(1:my_B_size,Lstart_pos:Lend_pos,jjB)
               ! END DO
            END DO
            CALL timestop(handle2)

            rec_ij_index = num_ij_pairs(proc_receive)

            IF (ij_index <= rec_ij_index) THEN
               ! we know that proc_receive has something to send for us, let's see what
               ij_counter_rec = &
                  (ij_index-MIN(1, integ_group_pos2color_sub(proc_receive)))*ngroup+integ_group_pos2color_sub(proc_receive)

               rec_i = ij_map(ij_counter_rec, 1)
               rec_j = ij_map(ij_counter_rec, 2)
               rec_block_size = ij_map(ij_counter_rec, 3)

               IF (.NOT. alpha_beta) THEN
                  ALLOCATE (BI_C_rec(2*rec_block_size, my_B_size, my_group_L_size))
               ELSE
                  ALLOCATE (BI_C_rec(rec_block_size, my_B_size, my_group_L_size))
                  ALLOCATE (BI_C_rec_beta(rec_block_size, my_B_size_beta, my_group_L_size))
               ENDIF

               BI_C_rec = 0.0_dp
               IF (alpha_beta) BI_C_rec_beta = 0.0_dp

               CALL mp_sendrecv(BI_C_send, proc_send, &
                                BI_C_rec, proc_receive, &
                                para_env_exchange%group)
               IF (alpha_beta) THEN
                  CALL mp_sendrecv(BI_C_send_beta, proc_send, &
                                   BI_C_rec_beta, proc_receive, &
                                   para_env_exchange%group)
               ENDIF

               CALL timeset(routineN//"_comm2_w", handle2)
               DO irep = 0, num_integ_group-1
                  Lstart_pos = ranges_info_array(1, irep, para_env_exchange%mepos)
                  Lend_pos = ranges_info_array(2, irep, para_env_exchange%mepos)
                  start_point = ranges_info_array(3, irep, para_env_exchange%mepos)
                  end_point = ranges_info_array(4, irep, para_env_exchange%mepos)
                  DO iiB = 1, rec_block_size
!$OMP               PARALLEL DO DEFAULT(NONE) PRIVATE(kkk) &
!$OMP                                         SHARED(start_point,end_point,&
!$OMP                                                mp2_env,rec_i,iiB,my_B_size,BI_C_rec,&
!$OMP                                         alpha_alpha,beta_beta,open_shell)
                     DO kkk = start_point, end_point
                        IF (alpha_alpha .OR. (.NOT. open_shell)) THEN
                           mp2_env%ri_grad%Gamma_P_ia(rec_i+iiB-1, 1:my_B_size, kkk) = &
                              mp2_env%ri_grad%Gamma_P_ia(rec_i+iiB-1, 1:my_B_size, kkk)+ &
                              BI_C_rec(iiB, 1:my_B_size, kkk)
                        ENDIF
                        IF (beta_beta) THEN
                           mp2_env%ri_grad%Gamma_P_ia_beta(rec_i+iiB-1, 1:my_B_size, kkk) = &
                              mp2_env%ri_grad%Gamma_P_ia_beta(rec_i+iiB-1, 1:my_B_size, kkk)+ &
                              BI_C_rec(iiB, 1:my_B_size, kkk)
                        ENDIF
                     END DO
!$OMP               END PARALLEL DO
                  END DO
                  DO jjB = 1, rec_block_size
                     IF (.NOT. alpha_beta) THEN
!$OMP                   PARALLEL DO DEFAULT(NONE) PRIVATE(kkk) &
!$OMP                                             SHARED(start_point,end_point,rec_block_size,&
!$OMP                                                    mp2_env,rec_j,jjB,my_B_size,BI_C_rec,&
!$OMP                                             alpha_alpha,beta_beta,open_shell)
                        DO kkk = start_point, end_point
                           IF (alpha_alpha .OR. (.NOT. open_shell)) THEN
                              mp2_env%ri_grad%Gamma_P_ia(rec_j+jjB-1, 1:my_B_size, kkk) = &
                                 mp2_env%ri_grad%Gamma_P_ia(rec_j+jjB-1, 1:my_B_size, kkk)+ &
                                 BI_C_rec(jjB+rec_block_size, 1:my_B_size, kkk)
                           ENDIF
                           IF (beta_beta) THEN
                              mp2_env%ri_grad%Gamma_P_ia_beta(rec_j+jjB-1, 1:my_B_size, kkk) = &
                                 mp2_env%ri_grad%Gamma_P_ia_beta(rec_j+jjB-1, 1:my_B_size, kkk)+ &
                                 BI_C_rec(jjB+rec_block_size, 1:my_B_size, kkk)
                           ENDIF
                        END DO
!$OMP                   END PARALLEL DO
                     ELSE
!$OMP                   PARALLEL DO DEFAULT(NONE) PRIVATE(kkk) &
!$OMP                                             SHARED(start_point,end_point,&
!$OMP                                           mp2_env,rec_j,jjB,my_B_size_beta,BI_C_rec_beta)
                        DO kkk = start_point, end_point
                           mp2_env%ri_grad%Gamma_P_ia_beta(rec_j+jjB-1, 1:my_B_size_beta, kkk) = &
                              mp2_env%ri_grad%Gamma_P_ia_beta(rec_j+jjB-1, 1:my_B_size_beta, kkk)+ &
                              BI_C_rec_beta(jjB, 1:my_B_size_beta, kkk)
                        END DO
!$OMP                   END PARALLEL DO
                     ENDIF
                  END DO
                  ! DO iiB=1, rec_block_size
                  !   mp2_env%ri_grad%Gamma_P_ia(rec_i+iiB-1,1:my_B_size,start_point:end_point)=&
                  !   mp2_env%ri_grad%Gamma_P_ia(rec_i+iiB-1,1:my_B_size,start_point:end_point)+&
                  !   BI_C_rec(iiB,1:my_B_size,start_point:end_point)
                  ! END DO
                  ! DO jjB=1, rec_block_size
                  !   mp2_env%ri_grad%Gamma_P_ia(rec_j+jjB-1,1:my_B_size,start_point:end_point)=&
                  !   mp2_env%ri_grad%Gamma_P_ia(rec_j+jjB-1,1:my_B_size,start_point:end_point)+&
                  !   BI_C_rec(jjB+rec_block_size,1:my_B_size,start_point:end_point)
                  ! END DO
               END DO
               CALL timestop(handle2)

               DEALLOCATE (BI_C_rec)
               IF (alpha_beta) DEALLOCATE (BI_C_rec_beta)

            ELSE
               ! we have something to send but nothing to receive

               CALL mp_sendrecv(BI_C_send, proc_send, &
                                null_mat_rec, proc_receive, &
                                para_env_exchange%group)
               IF (alpha_beta) THEN
                  CALL mp_sendrecv(BI_C_send_beta, proc_send, &
                                   null_mat_rec, proc_receive, &
                                   para_env_exchange%group)
               ENDIF

            END IF

            DEALLOCATE (BI_C_send)
            IF (alpha_beta) DEALLOCATE (BI_C_send_beta)
         END DO

      ELSE
         ! noting to send check if we have to receive
         DO proc_shift = 1, para_env_exchange%num_pe-1
            proc_send = proc_map(para_env_exchange%mepos+proc_shift)
            proc_receive = proc_map(para_env_exchange%mepos-proc_shift)
            rec_ij_index = num_ij_pairs(proc_receive)

            IF (ij_index <= rec_ij_index) THEN
               ! we know that proc_receive has something to send for us, let's see what
               ij_counter_rec = &
                  (ij_index-MIN(1, integ_group_pos2color_sub(proc_receive)))*ngroup+integ_group_pos2color_sub(proc_receive)

               rec_i = ij_map(ij_counter_rec, 1)
               rec_j = ij_map(ij_counter_rec, 2)
               rec_block_size = ij_map(ij_counter_rec, 3)

               IF (.NOT. alpha_beta) THEN
                  ALLOCATE (BI_C_rec(2*rec_block_size, my_B_size, my_group_L_size))
               ELSE
                  ALLOCATE (BI_C_rec(rec_block_size, my_B_size, my_group_L_size))
                  ALLOCATE (BI_C_rec_beta(rec_block_size, my_B_size_beta, my_group_L_size))
               ENDIF

               BI_C_rec = 0.0_dp
               IF (alpha_beta) BI_C_rec_beta = 0.0_dp

               CALL mp_sendrecv(null_mat_send, proc_send, &
                                BI_C_rec, proc_receive, &
                                para_env_exchange%group)
               IF (alpha_beta) THEN
                  CALL mp_sendrecv(null_mat_send, proc_send, &
                                   BI_C_rec_beta, proc_receive, &
                                   para_env_exchange%group)
               ENDIF

               CALL timeset(routineN//"_comm2_w", handle2)
               DO irep = 0, num_integ_group-1
                  Lstart_pos = ranges_info_array(1, irep, para_env_exchange%mepos)
                  Lend_pos = ranges_info_array(2, irep, para_env_exchange%mepos)
                  start_point = ranges_info_array(3, irep, para_env_exchange%mepos)
                  end_point = ranges_info_array(4, irep, para_env_exchange%mepos)
                  DO iiB = 1, rec_block_size
!$OMP               PARALLEL DO DEFAULT(NONE) PRIVATE(kkk) &
!$OMP                                         SHARED(start_point,end_point,&
!$OMP                                                mp2_env,rec_i,iiB,my_B_size,BI_C_rec,&
!$OMP                                         alpha_alpha,beta_beta,open_shell)
                     DO kkk = start_point, end_point
                        IF (alpha_alpha .OR. (.NOT. open_shell)) THEN
                           mp2_env%ri_grad%Gamma_P_ia(rec_i+iiB-1, 1:my_B_size, kkk) = &
                              mp2_env%ri_grad%Gamma_P_ia(rec_i+iiB-1, 1:my_B_size, kkk)+ &
                              BI_C_rec(iiB, 1:my_B_size, kkk)
                        ENDIF
                        IF (beta_beta) THEN
                           mp2_env%ri_grad%Gamma_P_ia_beta(rec_i+iiB-1, 1:my_B_size, kkk) = &
                              mp2_env%ri_grad%Gamma_P_ia_beta(rec_i+iiB-1, 1:my_B_size, kkk)+ &
                              BI_C_rec(iiB, 1:my_B_size, kkk)
                        ENDIF
                     END DO
!$OMP               END PARALLEL DO
                  END DO
                  DO jjB = 1, rec_block_size
                     IF (.NOT. alpha_beta) THEN
!$OMP                  PARALLEL DO DEFAULT(NONE) PRIVATE(kkk) &
!$OMP                                         SHARED(start_point,end_point,rec_block_size,&
!$OMP                                                mp2_env,rec_j,jjB,my_B_size,BI_C_rec,&
!$OMP                                            alpha_alpha,beta_beta,open_shell)
                        DO kkk = start_point, end_point
                           IF (alpha_alpha .OR. (.NOT. open_shell)) THEN
                              mp2_env%ri_grad%Gamma_P_ia(rec_j+jjB-1, 1:my_B_size, kkk) = &
                                 mp2_env%ri_grad%Gamma_P_ia(rec_j+jjB-1, 1:my_B_size, kkk)+ &
                                 BI_C_rec(jjB+rec_block_size, 1:my_B_size, kkk)
                           ENDIF
                           IF (beta_beta) THEN
                              mp2_env%ri_grad%Gamma_P_ia_beta(rec_j+jjB-1, 1:my_B_size, kkk) = &
                                 mp2_env%ri_grad%Gamma_P_ia_beta(rec_j+jjB-1, 1:my_B_size, kkk)+ &
                                 BI_C_rec(jjB+rec_block_size, 1:my_B_size, kkk)
                           ENDIF
                        END DO
!$OMP                  END PARALLEL DO
                     ELSE
!$OMP                  PARALLEL DO DEFAULT(NONE) PRIVATE(kkk) &
!$OMP                                         SHARED(start_point,end_point,&
!$OMP                                       mp2_env,rec_j,jjB,my_B_size_beta,BI_C_rec_beta)
                        DO kkk = start_point, end_point
                           mp2_env%ri_grad%Gamma_P_ia_beta(rec_j+jjB-1, 1:my_B_size_beta, kkk) = &
                              mp2_env%ri_grad%Gamma_P_ia_beta(rec_j+jjB-1, 1:my_B_size_beta, kkk)+ &
                              BI_C_rec_beta(jjB, 1:my_B_size_beta, kkk)
                        END DO
!$OMP                  END PARALLEL DO
                     ENDIF
                  END DO
                  ! DO iiB=1, rec_block_size
                  !   mp2_env%ri_grad%Gamma_P_ia(rec_i+iiB-1,1:my_B_size,start_point:end_point)=&
                  !   mp2_env%ri_grad%Gamma_P_ia(rec_i+iiB-1,1:my_B_size,start_point:end_point)+&
                  !   BI_C_rec(iiB,1:my_B_size,start_point:end_point)
                  ! END DO
                  ! DO jjB=1, rec_block_size
                  !   mp2_env%ri_grad%Gamma_P_ia(rec_j+jjB-1,1:my_B_size,start_point:end_point)=&
                  !   mp2_env%ri_grad%Gamma_P_ia(rec_j+jjB-1,1:my_B_size,start_point:end_point)+&
                  !   BI_C_rec(jjB+rec_block_size,1:my_B_size,start_point:end_point)
                  ! END DO
               END DO
               CALL timestop(handle2)

               DEALLOCATE (BI_C_rec)
               IF (alpha_beta) DEALLOCATE (BI_C_rec_beta)

            ELSE
               ! nothing to send nothing to receive
               CALL mp_sendrecv(null_mat_send, proc_send, &
                                null_mat_rec, proc_receive, &
                                para_env_exchange%group)
               IF (alpha_beta) THEN
                  CALL mp_sendrecv(null_mat_send, proc_send, &
                                   null_mat_rec, proc_receive, &
                                   para_env_exchange%group)
               ENDIF

            END IF
         END DO

      END IF
      CALL timestop(handle)

   END SUBROUTINE mp2_redistribute_gamma

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
!> \param Eigenval ...
!> \param homo ...
!> \param virtual ...
!> \param open_shell ...
!> \param beta_beta ...
!> \param alpha_beta ...
!> \param Bib_C ...
!> \param unit_nr ...
!> \param dimen_RI ...
!> \param my_B_size ...
!> \param ngroup ...
!> \param num_integ_group ...
!> \param my_group_L_size ...
!> \param color_sub ...
!> \param ranges_info_array ...
!> \param para_env_exchange ...
!> \param para_env_sub ...
!> \param proc_map ...
!> \param my_B_virtual_start ...
!> \param my_B_virtual_end ...
!> \param sizes_array ...
!> \param ends_B_virtual ...
!> \param sizes_B_virtual ...
!> \param starts_B_virtual ...
!> \param sub_proc_map ...
!> \param integ_group_pos2color_sub ...
!> \param local_ab ...
!> \param BIb_C_beta ...
!> \param my_B_size_beta ...
!> \param ends_B_virtual_beta ...
!> \param sizes_B_virtual_beta ...
!> \param starts_B_virtual_beta ...
!> \param my_B_virtual_start_beta ...
!> \param virtual_beta ...
!> \param homo_beta ...
!> \param Eigenval_beta ...
!> \param my_B_virtual_end_beta ...
! **************************************************************************************************
   SUBROUTINE quasi_degenerate_P_ij(mp2_env, Eigenval, homo, virtual, open_shell, &
                                    beta_beta, alpha_beta, Bib_C, unit_nr, dimen_RI, &
                                    my_B_size, ngroup, num_integ_group, my_group_L_size, &
                                    color_sub, ranges_info_array, para_env_exchange, para_env_sub, proc_map, &
                                    my_B_virtual_start, my_B_virtual_end, sizes_array, ends_B_virtual, sizes_B_virtual, &
                                    starts_B_virtual, sub_proc_map, integ_group_pos2color_sub, &
                                    local_ab, BIb_C_beta, my_B_size_beta, &
                                    ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta, my_B_virtual_start_beta, &
                                    virtual_beta, homo_beta, Eigenval_beta, my_B_virtual_end_beta)
      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: homo, virtual
      LOGICAL                                            :: open_shell, beta_beta, alpha_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C
      INTEGER                                            :: unit_nr, dimen_RI, my_B_size, ngroup, &
                                                            num_integ_group, my_group_L_size, &
                                                            color_sub
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: ranges_info_array
      TYPE(cp_para_env_type), POINTER                    :: para_env_exchange, para_env_sub
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: proc_map
      INTEGER                                            :: my_B_virtual_start, my_B_virtual_end
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sizes_array, ends_B_virtual, &
                                                            sizes_B_virtual, starts_B_virtual, &
                                                            sub_proc_map, integ_group_pos2color_sub
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: local_ab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :), &
         OPTIONAL                                        :: BIb_C_beta
      INTEGER, OPTIONAL                                  :: my_B_size_beta
      INTEGER, ALLOCATABLE, DIMENSION(:), OPTIONAL       :: ends_B_virtual_beta, &
                                                            sizes_B_virtual_beta, &
                                                            starts_B_virtual_beta
      INTEGER, OPTIONAL                                  :: my_B_virtual_start_beta, virtual_beta, &
                                                            homo_beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      INTEGER, OPTIONAL                                  :: my_B_virtual_end_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'quasi_degenerate_P_ij', &
         routineP = moduleN//':'//routineN

      INTEGER :: a, a_global, b, b_global, end_point, handle, handle2, ijk_counter, &
         ijk_counter_send, ijk_index, iloops, irep, Lend_pos, Lstart_pos, max_ijk, max_ijk_beta, &
         max_ijk_loop, my_i, my_ijk, my_ijk_beta, my_j, my_k, my_virtual, nloops, proc_receive, &
         proc_send, proc_shift, rec_B_size, rec_B_virtual_end, rec_B_virtual_start, rec_L_size, &
         send_B_size, send_B_virtual_end, send_B_virtual_start, send_i, send_ijk_index, send_j, &
         send_k, size_B_i, size_B_j, size_B_k, start_point
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: num_ijk, num_ijk_beta
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ijk_map, ijk_map_beta
      REAL(KIND=dp)                                      :: amp_fac, null_mat_rec(2, 2, 2), &
                                                            null_mat_send(2, 2, 2), P_ij_elem
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: external_ab, external_i_aL, t_ab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BI_C_rec, local_i_aL, local_j_aL, &
                                                            local_k_aL

!

      CALL timeset(routineN//"_ij_sing", handle)
! Define the number of loops over orbital triplets

      nloops = 1
      IF (alpha_beta) nloops = 2

      ! Find the number of quasi-degenerate orbitals and orbital triplets

      IF (.NOT. alpha_beta) THEN
         CALL Find_quasi_degenerate_ij(my_ijk, homo, Eigenval, mp2_env, ijk_map, unit_nr, ngroup, &
                                       beta_beta, alpha_beta, para_env_exchange, num_ijk, max_ijk, color_sub)
      ELSE
         CALL Find_quasi_degenerate_ij(my_ijk, homo, Eigenval, mp2_env, ijk_map, unit_nr, ngroup, &
                                       beta_beta, alpha_beta, para_env_exchange, num_ijk, max_ijk, color_sub, &
                                       Eigenval_beta, homo_beta, ijk_map_beta, num_ijk_beta, max_ijk_beta, my_ijk_beta)
      ENDIF

      ! Set amplitude factor
      amp_fac = 2.0_dp
      IF (open_shell) amp_fac = 1.0_dp

      ! Loop(s) over orbital triplets
      DO iloops = 1, nloops
         IF (iloops .EQ. 1) THEN
            size_B_i = my_B_size
            size_B_j = my_B_size
            max_ijk_loop = max_ijk
            my_virtual = virtual
            IF (alpha_beta) THEN
               size_B_k = my_B_size_beta
            ELSE
               size_B_k = my_B_size
            ENDIF
         ELSE
            size_B_i = my_B_size_beta
            size_B_j = my_B_size_beta
            size_B_k = my_B_size
            my_virtual = virtual_beta
            max_ijk_loop = max_ijk_beta
         ENDIF

         ALLOCATE (local_i_aL(dimen_RI, size_B_i, 1))
         ALLOCATE (local_j_aL(dimen_RI, size_B_j, 1))
         ALLOCATE (local_k_aL(dimen_RI, size_B_k, 1))
         ALLOCATE (t_ab(my_virtual, size_B_k))

         DO ijk_index = 1, max_ijk_loop
            IF (iloops .EQ. 2) my_ijk = my_ijk_beta
            IF (ijk_index <= my_ijk) THEN
               ! work to be done
               ijk_counter = (ijk_index-MIN(1, color_sub))*ngroup+color_sub
               IF (iloops .EQ. 1) THEN
                  my_i = ijk_map(ijk_counter, 1)
                  my_j = ijk_map(ijk_counter, 2)
                  my_k = ijk_map(ijk_counter, 3)
               ELSE
                  my_i = ijk_map_beta(ijk_counter, 1)
                  my_j = ijk_map_beta(ijk_counter, 2)
                  my_k = ijk_map_beta(ijk_counter, 3)
               ENDIF

               local_i_aL = 0.0_dp
               local_j_al = 0.0_dp
               local_k_al = 0.0_dp
               DO irep = 0, num_integ_group-1
                  Lstart_pos = ranges_info_array(1, irep, para_env_exchange%mepos)
                  Lend_pos = ranges_info_array(2, irep, para_env_exchange%mepos)
                  start_point = ranges_info_array(3, irep, para_env_exchange%mepos)
                  end_point = ranges_info_array(4, irep, para_env_exchange%mepos)

                  IF (.NOT. alpha_beta) THEN
                     local_i_aL(Lstart_pos:Lend_pos, 1:size_B_i, 1) = BIb_C(start_point:end_point, 1:size_B_i, my_i)
                     local_j_aL(Lstart_pos:Lend_pos, 1:size_B_j, 1) = BIb_C(start_point:end_point, 1:size_B_j, my_j)
                     local_k_aL(Lstart_pos:Lend_pos, 1:size_B_k, 1) = BIb_C(start_point:end_point, 1:size_B_k, my_k)
                  ELSE
                     IF (iloops .EQ. 1) THEN ! For alpha-alpha density
                        local_i_aL(Lstart_pos:Lend_pos, 1:size_B_i, 1) = BIb_C(start_point:end_point, 1:size_B_i, my_i)
                        local_j_aL(Lstart_pos:Lend_pos, 1:size_B_j, 1) = BIb_C(start_point:end_point, 1:size_B_j, my_j)
                        local_k_aL(Lstart_pos:Lend_pos, 1:size_B_k, 1) = BIb_C_beta(start_point:end_point, 1:size_B_k, my_k)
                     ELSE ! For beta-beta density
                        local_i_aL(Lstart_pos:Lend_pos, 1:size_B_i, 1) = BIb_C_beta(start_point:end_point, 1:size_B_i, my_i)
                        local_j_aL(Lstart_pos:Lend_pos, 1:size_B_j, 1) = BIb_C_beta(start_point:end_point, 1:size_B_j, my_j)
                        local_k_aL(Lstart_pos:Lend_pos, 1:size_B_k, 1) = BIb_C(start_point:end_point, 1:size_B_k, my_k)
                     ENDIF
                  ENDIF
               END DO

               DO proc_shift = 1, para_env_exchange%num_pe-1
                  proc_send = proc_map(para_env_exchange%mepos+proc_shift)
                  proc_receive = proc_map(para_env_exchange%mepos-proc_shift)

                  send_ijk_index = num_ijk(proc_send)
                  IF (iloops .EQ. 2) send_ijk_index = num_ijk_beta(proc_send)

                  rec_L_size = sizes_array(proc_receive)
                  ALLOCATE (BI_C_rec(rec_L_size, size_B_i, 1))

                  IF (ijk_index <= send_ijk_index) THEN
                     ! something to send
                     ijk_counter_send = (ijk_index-MIN(1, integ_group_pos2color_sub(proc_send)))* &
                                        ngroup+integ_group_pos2color_sub(proc_send)
                     IF (iloops .EQ. 1) THEN
                        send_i = ijk_map(ijk_counter_send, 1)
                        send_j = ijk_map(ijk_counter_send, 2)
                        send_k = ijk_map(ijk_counter_send, 3)
                     ELSE
                        send_i = ijk_map_beta(ijk_counter_send, 1)
                        send_j = ijk_map_beta(ijk_counter_send, 2)
                        send_k = ijk_map_beta(ijk_counter_send, 3)
                     ENDIF
                  END IF

                  ! occupied i
                  BI_C_rec = 0.0_dp
                  IF (ijk_index <= send_ijk_index) THEN
                     IF (iloops .EQ. 1) THEN
                        CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:size_B_i, send_i), proc_send, &
                                         BI_C_rec(1:rec_L_size, 1:size_B_i, 1), proc_receive, &
                                         para_env_exchange%group)
                     ELSE
                        CALL mp_sendrecv(BIb_C_beta(1:my_group_L_size, 1:size_B_i, send_i), proc_send, &
                                         BI_C_rec(1:rec_L_size, 1:size_B_i, 1), proc_receive, &
                                         para_env_exchange%group)
                     ENDIF
                  ELSE
                     ! nothing to send
                     CALL mp_sendrecv(null_mat_send, proc_send, &
                                      BI_C_rec(1:rec_L_size, 1:size_B_i, 1:1), proc_receive, &
                                      para_env_exchange%group)
                  END IF
                  DO irep = 0, num_integ_group-1
                     Lstart_pos = ranges_info_array(1, irep, proc_receive)
                     Lend_pos = ranges_info_array(2, irep, proc_receive)
                     start_point = ranges_info_array(3, irep, proc_receive)
                     end_point = ranges_info_array(4, irep, proc_receive)

                     local_i_aL(Lstart_pos:Lend_pos, 1:size_B_i, 1) = BI_C_rec(start_point:end_point, 1:size_B_i, 1)
                  END DO

                  ! occupied j
                  BI_C_rec = 0.0_dp
                  IF (ijk_index <= send_ijk_index) THEN
                     IF (iloops .EQ. 1) THEN
                        CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:size_B_j, send_j), proc_send, &
                                         BI_C_rec(1:rec_L_size, 1:size_B_j, 1), proc_receive, &
                                         para_env_exchange%group)
                     ELSE ! For beta_beta density, the size is different now
                        !DEALLOCATE(BI_C_rec)
                        !ALLOCATE(BI_C_rec(rec_L_size,size_B_j,1))
                        !BI_C_rec=0.0_dp
                        CALL mp_sendrecv(BIb_C_beta(1:my_group_L_size, 1:size_B_j, send_j), proc_send, &
                                         BI_C_rec(1:rec_L_size, 1:size_B_j, 1), proc_receive, &
                                         para_env_exchange%group)
                     ENDIF
                  ELSE
                     ! nothing to send
                     CALL mp_sendrecv(null_mat_send, proc_send, &
                                      BI_C_rec(1:rec_L_size, 1:size_B_j, 1:1), proc_receive, &
                                      para_env_exchange%group)
                  END IF
                  DO irep = 0, num_integ_group-1
                     Lstart_pos = ranges_info_array(1, irep, proc_receive)
                     Lend_pos = ranges_info_array(2, irep, proc_receive)
                     start_point = ranges_info_array(3, irep, proc_receive)
                     end_point = ranges_info_array(4, irep, proc_receive)

                     local_j_aL(Lstart_pos:Lend_pos, 1:size_B_j, 1) = BI_C_rec(start_point:end_point, 1:size_B_j, 1)
                  END DO

                  ! occupied k
                  BI_C_rec = 0.0_dp
                  DEALLOCATE (BI_C_rec)
                  ALLOCATE (BI_C_rec(rec_L_size, size_B_k, 1))
                  BI_C_rec = 0.0_dp
                  IF (ijk_index <= send_ijk_index) THEN
                     IF (iloops .EQ. 1) THEN
                        IF (.NOT. alpha_beta) THEN
                           CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:size_B_k, send_k), proc_send, &
                                            BI_C_rec(1:rec_L_size, 1:size_B_k, 1), proc_receive, &
                                            para_env_exchange%group)
                        ELSE
                           CALL mp_sendrecv(BIb_C_beta(1:my_group_L_size, 1:size_B_k, send_k), proc_send, &
                                            BI_C_rec(1:rec_L_size, 1:size_B_k, 1), proc_receive, &
                                            para_env_exchange%group)
                        ENDIF
                     ELSE ! For beta_beta density, the size is different now
                        BI_C_rec = 0.0_dp
                        CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:size_B_k, send_k), proc_send, &
                                         BI_C_rec(1:rec_L_size, 1:size_B_k, 1), proc_receive, &
                                         para_env_exchange%group)
                     ENDIF
                  ELSE
                     ! nothing to send
                     BI_C_rec = 0.0_dp
                     CALL mp_sendrecv(null_mat_send, proc_send, &
                                      BI_C_rec(1:rec_L_size, 1:size_B_k, 1:1), proc_receive, &
                                      para_env_exchange%group)
                  END IF
                  DO irep = 0, num_integ_group-1
                     Lstart_pos = ranges_info_array(1, irep, proc_receive)
                     Lend_pos = ranges_info_array(2, irep, proc_receive)
                     start_point = ranges_info_array(3, irep, proc_receive)
                     end_point = ranges_info_array(4, irep, proc_receive)

                     local_k_aL(Lstart_pos:Lend_pos, 1:size_B_k, 1) = BI_C_rec(start_point:end_point, 1:size_B_k, 1)
                  END DO

                  DEALLOCATE (BI_C_rec)
               END DO

               ! expand integrals
               CALL timeset(routineN//"_exp_ik", handle2)
               local_ab = 0.0_dp
               IF (iloops .EQ. 2) THEN ! For alpha-beta case for beta-beta density the dimensions are different
                  DEALLOCATE (local_ab)
                  ALLOCATE (local_ab(virtual_beta, size_B_k))
                  local_ab = 0.0_dp
                  CALL dgemm('T', 'N', size_B_i, size_B_k, dimen_RI, 1.0_dp, &
                             local_i_aL(:, :, 1), dimen_RI, local_k_aL(:, :, 1), dimen_RI, &
                             0.0_dp, local_ab(my_B_virtual_start_beta:my_B_virtual_end_beta, 1:size_B_k), size_B_i)
               ELSE
                  CALL dgemm('T', 'N', size_B_i, size_B_k, dimen_RI, 1.0_dp, &
                             local_i_aL(:, :, 1), dimen_RI, local_k_aL(:, :, 1), dimen_RI, &
                             0.0_dp, local_ab(my_B_virtual_start:my_B_virtual_end, 1:size_B_k), size_B_i)
               ENDIF
               DO proc_shift = 1, para_env_sub%num_pe-1
                  proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
                  proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

                  IF (iloops .EQ. 1) THEN
                     rec_B_size = sizes_B_virtual(proc_receive)
                     rec_B_virtual_end = ends_B_virtual(proc_receive)
                     rec_B_virtual_start = starts_B_virtual(proc_receive)
                  ELSE
                     rec_B_size = sizes_B_virtual_beta(proc_receive)
                     rec_B_virtual_end = ends_B_virtual_beta(proc_receive)
                     rec_B_virtual_start = starts_B_virtual_beta(proc_receive)
                  ENDIF

                  ALLOCATE (external_i_aL(dimen_RI, rec_B_size))
                  external_i_aL = 0.0_dp

                  CALL mp_sendrecv(local_i_aL(:, :, 1), proc_send, &
                                   external_i_aL, proc_receive, &
                                   para_env_sub%group)

                  CALL dgemm('T', 'N', rec_B_size, size_B_k, dimen_RI, 1.0_dp, &
                             external_i_aL, dimen_RI, local_k_aL(:, :, 1), dimen_RI, &
                             0.0_dp, local_ab(rec_B_virtual_start:rec_B_virtual_end, 1:size_B_k), rec_B_size)

                  DEALLOCATE (external_i_aL)
               END DO
               CALL timestop(handle2)

               ! Amplitudes
               CALL timeset(routineN//"_tab", handle2)
               t_ab = 0.0_dp
               ! Alpha-alpha, beta-beta and closed shell
               IF (.NOT. alpha_beta) THEN
                  DO b = 1, size_B_k
                     b_global = b+my_B_virtual_start-1
                     DO a = 1, my_B_size
                        a_global = a+my_B_virtual_start-1
                        t_ab(a_global, b) = (amp_fac*local_ab(a_global, b)-local_ab(b_global, a))/ &
                                            (Eigenval(my_i)+Eigenval(my_k)-Eigenval(homo+a_global)-Eigenval(homo+b_global))
                     END DO
                  END DO
               ELSE
                  IF (iloops .EQ. 1) THEN ! Alpha-beta for alpha-alpha density
                     DO b = 1, size_B_k
                        b_global = b+my_B_virtual_start_beta-1
                        DO a = 1, my_B_size
                           a_global = a+my_B_virtual_start-1
                           t_ab(a_global, b) = local_ab(a_global, b)/ &
                                               (Eigenval(my_i)+Eigenval_beta(my_k)-Eigenval(homo+a_global)- &
                                                Eigenval_beta(homo_beta+b_global))
                        END DO
                     END DO
                  ELSE ! Alpha-beta for beta-beta density
                     DO b = 1, size_B_k
                        b_global = b+my_B_virtual_start-1
                        DO a = 1, my_B_size_beta
                           a_global = a+my_B_virtual_start_beta-1
                           t_ab(a_global, b) = local_ab(a_global, b)/ &
                                               (Eigenval_beta(my_i)+Eigenval(my_k)-Eigenval_beta(homo_beta+a_global)- &
                                                Eigenval(homo+b_global))
                        END DO
                     END DO
                  ENDIF
               ENDIF

               IF (.NOT. alpha_beta) THEN
                  DO proc_shift = 1, para_env_sub%num_pe-1
                     proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
                     proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
                     rec_B_size = sizes_B_virtual(proc_receive)
                     rec_B_virtual_end = ends_B_virtual(proc_receive)
                     rec_B_virtual_start = starts_B_virtual(proc_receive)
                     send_B_size = sizes_B_virtual(proc_send)
                     send_B_virtual_end = ends_B_virtual(proc_send)
                     send_B_virtual_start = starts_B_virtual(proc_send)

                     ALLOCATE (external_ab(size_B_i, rec_B_size))
                     external_ab = 0.0_dp
                     CALL mp_sendrecv(local_ab(send_B_virtual_start:send_B_virtual_end, 1:size_B_k), proc_send, &
                                      external_ab(1:size_B_i, 1:rec_B_size), proc_receive, para_env_sub%group)

                     DO b = 1, my_B_size
                        b_global = b+my_B_virtual_start-1
                        DO a = 1, rec_B_size
                           a_global = a+rec_B_virtual_start-1
                           t_ab(a_global, b) = (amp_fac*local_ab(a_global, b)-external_ab(b, a))/ &
                                               (Eigenval(my_i)+Eigenval(my_k)-Eigenval(homo+a_global)-Eigenval(homo+b_global))
                        END DO
                     END DO

                     DEALLOCATE (external_ab)
                  END DO
               ENDIF
               CALL timestop(handle2)

               ! Expand the second set of integrals
               CALL timeset(routineN//"_exp_jk", handle2)
               local_ab = 0.0_dp

               IF (iloops .EQ. 2) THEN ! In alpha-beta case for beta-beta density the dimensions are different
                  CALL dgemm('T', 'N', size_B_j, size_B_k, dimen_RI, 1.0_dp, &
                             local_j_aL(:, :, 1), dimen_RI, local_k_aL(:, :, 1), dimen_RI, &
                             0.0_dp, local_ab(my_B_virtual_start_beta:my_B_virtual_end_beta, 1:size_B_k), size_B_j)
               ELSE
                  CALL dgemm('T', 'N', size_B_j, size_B_k, dimen_RI, 1.0_dp, &
                             local_j_aL(:, :, 1), dimen_RI, local_k_aL(:, :, 1), dimen_RI, &
                             0.0_dp, local_ab(my_B_virtual_start:my_B_virtual_end, 1:size_B_k), size_B_j)
               ENDIF

               DO proc_shift = 1, para_env_sub%num_pe-1
                  proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
                  proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

                  IF (iloops .EQ. 1) THEN
                     rec_B_size = sizes_B_virtual(proc_receive)
                     rec_B_virtual_end = ends_B_virtual(proc_receive)
                     rec_B_virtual_start = starts_B_virtual(proc_receive)
                  ELSE
                     rec_B_size = sizes_B_virtual_beta(proc_receive)
                     rec_B_virtual_end = ends_B_virtual_beta(proc_receive)
                     rec_B_virtual_start = starts_B_virtual_beta(proc_receive)
                  ENDIF

                  ALLOCATE (external_i_aL(dimen_RI, rec_B_size))
                  external_i_aL = 0.0_dp

                  CALL mp_sendrecv(local_j_aL(:, :, 1), proc_send, &
                                   external_i_aL, proc_receive, &
                                   para_env_sub%group)

                  CALL dgemm('T', 'N', rec_B_size, size_B_k, dimen_RI, 1.0_dp, &
                             external_i_aL, dimen_RI, local_k_aL(:, :, 1), dimen_RI, &
                             0.0_dp, local_ab(rec_B_virtual_start:rec_B_virtual_end, 1:size_B_k), rec_B_size)

                  DEALLOCATE (external_i_aL)
               END DO
               CALL timestop(handle2)

               CALL timeset(routineN//"_Pij", handle2)
               ! Alpha-alpha, beta-beta and closed shell
               IF (.NOT. alpha_beta) THEN
                  DO b = 1, size_B_k
                     b_global = b+my_B_virtual_start-1
                     DO a = 1, my_B_size
                        a_global = a+my_B_virtual_start-1
                        local_ab(a_global, b) = local_ab(a_global, b)/ &
                                                (Eigenval(my_j)+Eigenval(my_k)-Eigenval(homo+a_global)-Eigenval(homo+b_global))
                     END DO
                  END DO
               ELSE
                  IF (iloops .EQ. 1) THEN ! Alpha-beta for alpha-alpha density
                     DO b = 1, size_B_k
                        b_global = b+my_B_virtual_start_beta-1
                        DO a = 1, my_B_size
                           a_global = a+my_B_virtual_start-1
                           local_ab(a_global, b) = local_ab(a_global, b)/ &
                                                   (Eigenval(my_j)+Eigenval_beta(my_k)-Eigenval(homo+a_global)- &
                                                    Eigenval_beta(homo_beta+b_global))
                        END DO
                     END DO
                  ELSE ! Alpha-beta for beta-beta density
                     DO b = 1, size_B_k
                        b_global = b+my_B_virtual_start-1
                        DO a = 1, my_B_size_beta
                           a_global = a+my_B_virtual_start_beta-1
                           local_ab(a_global, b) = local_ab(a_global, b)/ &
                                                   (Eigenval_beta(my_j)+Eigenval(my_k)-Eigenval_beta(homo_beta+a_global)- &
                                                    Eigenval(homo+b_global))
                        END DO
                     END DO
                  ENDIF
               ENDIF
               !
               P_ij_elem = SUM(local_ab*t_ab)
               IF ((.NOT. open_shell) .AND. (.NOT. alpha_beta)) THEN
                  P_ij_elem = P_ij_elem*2.0_dp
               ENDIF
               IF ((beta_beta) .OR. (iloops .EQ. 2)) THEN
                  mp2_env%ri_grad%P_ij_beta(my_i, my_j) = mp2_env%ri_grad%P_ij_beta(my_i, my_j)-P_ij_elem
                  mp2_env%ri_grad%P_ij_beta(my_j, my_i) = mp2_env%ri_grad%P_ij_beta(my_j, my_i)-P_ij_elem
               ELSE
                  mp2_env%ri_grad%P_ij(my_i, my_j) = mp2_env%ri_grad%P_ij(my_i, my_j)-P_ij_elem
                  mp2_env%ri_grad%P_ij(my_j, my_i) = mp2_env%ri_grad%P_ij(my_j, my_i)-P_ij_elem
               ENDIF
               CALL timestop(handle2)
            ELSE
               ! no work to be done, possible messeges to be exchanged
               DO proc_shift = 1, para_env_exchange%num_pe-1
                  proc_send = proc_map(para_env_exchange%mepos+proc_shift)
                  proc_receive = proc_map(para_env_exchange%mepos-proc_shift)

                  send_ijk_index = num_ijk(proc_send)
                  IF (iloops .EQ. 2) send_ijk_index = num_ijk_beta(proc_send)

                  IF (ijk_index <= send_ijk_index) THEN
                     ! somethig to send
                     ijk_counter_send = (ijk_index-MIN(1, integ_group_pos2color_sub(proc_send)))*ngroup+ &
                                        integ_group_pos2color_sub(proc_send)
                     IF (iloops .EQ. 1) THEN
                        send_i = ijk_map(ijk_counter_send, 1)
                        send_j = ijk_map(ijk_counter_send, 2)
                        send_k = ijk_map(ijk_counter_send, 3)
                     ELSE
                        send_i = ijk_map_beta(ijk_counter_send, 1)
                        send_j = ijk_map_beta(ijk_counter_send, 2)
                        send_k = ijk_map_beta(ijk_counter_send, 3)
                     ENDIF
                     ! occupied i
                     IF (iloops .EQ. 1) THEN
                        CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:size_B_i, send_i:send_i), proc_send, &
                                         null_mat_rec, proc_receive, para_env_exchange%group)
                     ELSE
                        CALL mp_sendrecv(BIb_C_beta(1:my_group_L_size, 1:size_B_i, send_i:send_i), proc_send, &
                                         null_mat_rec, proc_receive, para_env_exchange%group)
                     ENDIF
                     ! occupied j
                     IF (iloops .EQ. 1) THEN
                        CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:size_B_j, send_j:send_j), proc_send, &
                                         null_mat_rec, proc_receive, para_env_exchange%group)
                     ELSE ! For beta_beta density, the size is different now
                        CALL mp_sendrecv(BIb_C_beta(1:my_group_L_size, 1:size_B_j, send_j:send_j), proc_send, &
                                         null_mat_rec, proc_receive, para_env_exchange%group)
                     ENDIF
                     ! occupied k
                     IF (iloops .EQ. 1) THEN
                        IF (.NOT. alpha_beta) THEN
                           CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:size_B_k, send_k:send_k), proc_send, &
                                            null_mat_rec, proc_receive, para_env_exchange%group)
                        ELSE
                           CALL mp_sendrecv(BIb_C_beta(1:my_group_L_size, 1:size_B_k, send_k:send_k), proc_send, &
                                            null_mat_rec, proc_receive, para_env_exchange%group)
                        ENDIF
                     ELSE ! For beta_beta density, the size is different now
                        CALL mp_sendrecv(BIb_C(1:my_group_L_size, 1:size_B_k, send_k:send_k), proc_send, &
                                         null_mat_rec, proc_receive, para_env_exchange%group)
                     ENDIF

                  ELSE
                     ! nothing to send
                     ! occupied i
                     CALL mp_sendrecv(null_mat_send, proc_send, &
                                      null_mat_rec, proc_receive, &
                                      para_env_exchange%group)
                     ! occupied j
                     CALL mp_sendrecv(null_mat_send, proc_send, &
                                      null_mat_rec, proc_receive, &
                                      para_env_exchange%group)
                     ! occupied k
                     CALL mp_sendrecv(null_mat_send, proc_send, &
                                      null_mat_rec, proc_receive, &
                                      para_env_exchange%group)
                  END IF

               END DO ! proc loop
            END IF
         END DO ! ijk_index loop
         DEALLOCATE (local_i_aL)
         DEALLOCATE (local_j_aL)
         DEALLOCATE (local_k_aL)
         DEALLOCATE (t_ab)
      ENDDO ! over number of loops (iloop)
      !
      DEALLOCATE (ijk_map)
      DEALLOCATE (num_ijk)
      IF (alpha_beta) THEN
         DEALLOCATE (ijk_map_beta, num_ijk_beta)
      ENDIF
      CALL timestop(handle)

   END SUBROUTINE Quasi_degenerate_P_ij

! **************************************************************************************************
!> \brief ...
!> \param my_ijk ...
!> \param homo ...
!> \param Eigenval ...
!> \param mp2_env ...
!> \param ijk_map ...
!> \param unit_nr ...
!> \param ngroup ...
!> \param beta_beta ...
!> \param alpha_beta ...
!> \param para_env_exchange ...
!> \param num_ijk ...
!> \param max_ijk ...
!> \param color_sub ...
!> \param Eigenval_beta ...
!> \param homo_beta ...
!> \param ijk_map_beta ...
!> \param num_ijk_beta ...
!> \param max_ijk_beta ...
!> \param my_ijk_beta ...
! **************************************************************************************************
   SUBROUTINE Find_quasi_degenerate_ij(my_ijk, homo, Eigenval, mp2_env, ijk_map, unit_nr, ngroup, &
                                       beta_beta, alpha_beta, para_env_exchange, num_ijk, max_ijk, color_sub, Eigenval_beta, &
                                       homo_beta, ijk_map_beta, num_ijk_beta, max_ijk_beta, my_ijk_beta)

      INTEGER                                            :: my_ijk, homo
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      TYPE(mp2_type), POINTER                            :: mp2_env
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ijk_map
      INTEGER                                            :: unit_nr, ngroup
      LOGICAL                                            :: beta_beta, alpha_beta
      TYPE(cp_para_env_type), POINTER                    :: para_env_exchange
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: num_ijk
      INTEGER                                            :: max_ijk, color_sub
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      INTEGER, OPTIONAL                                  :: homo_beta
      INTEGER, ALLOCATABLE, DIMENSION(:, :), OPTIONAL    :: ijk_map_beta
      INTEGER, ALLOCATABLE, DIMENSION(:), OPTIONAL       :: num_ijk_beta
      INTEGER, OPTIONAL                                  :: max_ijk_beta, my_ijk_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'Find_quasi_degenerate_ij', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iib, ijk_counter, jjb, kkb, my_homo, &
                                                            num_sing_ij, total_ijk

      IF (alpha_beta) THEN
         my_homo = homo_beta
      ELSE
         my_homo = homo
      ENDIF

      ! General case
      num_sing_ij = 0
      DO iiB = 1, homo
         ! diagonal elements already updated
         DO jjB = iiB+1, homo
            IF (ABS(Eigenval(jjB)-Eigenval(iiB)) < mp2_env%ri_mp2%eps_canonical) &
               num_sing_ij = num_sing_ij+1
         END DO
      END DO
      IF (.NOT. beta_beta) THEN
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "MO_INFO| Number of ij pairs below EPS_CANONICAL:", num_sing_ij
         END IF
      ELSE
         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "MO_INFO| Number of ij pairs (spin beta) below EPS_CANONICAL:", num_sing_ij
         END IF
      ENDIF
      total_ijk = my_homo*num_sing_ij
      ALLOCATE (ijk_map(total_ijk, 3))
      ijk_map = 0

      my_ijk = 0
      ijk_counter = 0
      DO iiB = 1, homo
         ! diagonal elements already updated
         DO jjB = iiB+1, homo
            IF (ABS(Eigenval(jjB)-Eigenval(iiB)) >= mp2_env%ri_mp2%eps_canonical) CYCLE
            DO kkB = 1, my_homo
               ijk_counter = ijk_counter+1
               ijk_map(ijk_counter, 1) = iiB
               ijk_map(ijk_counter, 2) = jjB
               ijk_map(ijk_counter, 3) = kkB
               IF (MOD(ijk_counter, ngroup) == color_sub) my_ijk = my_ijk+1
            END DO
         END DO
      END DO

      ALLOCATE (num_ijk(0:para_env_exchange%num_pe-1))
      num_ijk = 0
      num_ijk(para_env_exchange%mepos) = my_ijk
      CALL mp_sum(num_ijk, para_env_exchange%group)
      max_ijk = MAXVAL(num_ijk)

      ! Alpha-beta case: we need a second map
      IF (alpha_beta) THEN
         num_sing_ij = 0
         DO iiB = 1, homo_beta
            ! diagonal elements already updated
            DO jjB = iiB+1, homo_beta
               IF (ABS(Eigenval_beta(jjB)-Eigenval_beta(iiB)) < mp2_env%ri_mp2%eps_canonical) &
                  num_sing_ij = num_sing_ij+1
            END DO
         END DO
         ! total number of elemets that have to be computed
         total_ijk = homo*num_sing_ij
         ALLOCATE (ijk_map_beta(total_ijk, 3))
         ijk_map_beta = 0
         my_ijk_beta = 0
         ijk_counter = 0
         DO iiB = 1, homo_beta
            ! diagonal elements already updated
            DO jjB = iiB+1, homo_beta
               IF (ABS(Eigenval_beta(jjB)-Eigenval_beta(iiB)) >= mp2_env%ri_mp2%eps_canonical) CYCLE
               DO kkB = 1, homo
                  ijk_counter = ijk_counter+1
                  ijk_map_beta(ijk_counter, 1) = iiB
                  ijk_map_beta(ijk_counter, 2) = jjB
                  ijk_map_beta(ijk_counter, 3) = kkB
                  IF (MOD(ijk_counter, ngroup) == color_sub) my_ijk_beta = my_ijk_beta+1
               END DO
            END DO
         END DO
         ALLOCATE (num_ijk_beta(0:para_env_exchange%num_pe-1))
         num_ijk_beta = 0
         num_ijk_beta(para_env_exchange%mepos) = my_ijk_beta
         CALL mp_sum(num_ijk_beta, para_env_exchange%group)
         max_ijk_beta = MAXVAL(num_ijk_beta)
      ENDIF

   END SUBROUTINE Find_quasi_degenerate_ij

END MODULE mp2_ri_gpw
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate CPHF like update and solve Z-vector equation
!>        for MP2 gradients (only GPW)
!> \par History
!>      11.2013 created [Mauro Del Ben]
! **************************************************************************************************
MODULE mp2_cphf
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_col_block_sizes, cp_dbcsr_copy, &
        cp_dbcsr_create, cp_dbcsr_distribution, cp_dbcsr_init, cp_dbcsr_p_type, cp_dbcsr_release, &
        cp_dbcsr_row_block_sizes, cp_dbcsr_scale, cp_dbcsr_set, dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr
   USE cp_fm_basic_linalg,              ONLY: cp_fm_upper_to_full
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm_submat,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE hfx_energy_potential,            ONLY: integrate_four_center
   USE hfx_types,                       ONLY: alloc_containers,&
                                              hfx_container_type,&
                                              hfx_init_container,&
                                              hfx_type
   USE input_constants,                 ONLY: hfx_do_eval_energy
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp
   USE linear_systems,                  ONLY: solve_system
   USE machine,                         ONLY: m_walltime
   USE mathconstants,                   ONLY: fourpi
   USE message_passing,                 ONLY: mp_sum
   USE mp2_types,                       ONLY: mp2_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_derive,&
                                              pw_integral_ab,&
                                              pw_scale,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: calculate_rho_elec
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_integrate_potential,          ONLY: integrate_v_core_rspace,&
                                              integrate_v_rspace
   USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_rho_methods,                  ONLY: qs_rho_rebuild
   USE qs_rho_types,                    ONLY: qs_rho_create,&
                                              qs_rho_get,&
                                              qs_rho_release,&
                                              qs_rho_type
   USE virial_types,                    ONLY: virial_type

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_cphf'

   PUBLIC :: solve_z_vector_eq

CONTAINS

! **************************************************************************************************
!> \brief Solve Z-vector equations necessary for the calculation of the MP2
!>        gradients, in order to be consistent here the parameters for the
!>        calculation of the CPHF like updats have to be exactly equal to the
!>        SCF case
!> \param qs_env ...
!> \param mp2_env ...
!> \param para_env ...
!> \param dft_control ...
!> \param atomic_kind_set ...
!> \param mo_coeff ...
!> \param nmo ...
!> \param homo ...
!> \param Eigenval ...
!> \param unit_nr ...
!> \param Eigenval_beta ...
!> \param homo_beta ...
!> \param mo_coeff_beta ...
!> \author Mauro Del Ben, Vladimir Rybkin
! **************************************************************************************************
   SUBROUTINE solve_z_vector_eq(qs_env, mp2_env, para_env, dft_control, &
                                atomic_kind_set, mo_coeff, nmo, homo, Eigenval, unit_nr, &
                                Eigenval_beta, homo_beta, mo_coeff_beta)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      INTEGER                                            :: nmo, homo
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: unit_nr
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      INTEGER, OPTIONAL                                  :: homo_beta
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: mo_coeff_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'solve_z_vector_eq', &
         routineP = moduleN//':'//routineN

      INTEGER :: alpha, beta, bin, dimen, handle, handle2, i, i_global, i_thread, iiB, ikind, &
         irep, ispin, j_global, jjB, my_bin_size, n_rep_hf, n_threads, ncol_local, nrow_local, &
         transf_type_in, transf_type_out, virtual, virtual_beta
      INTEGER, DIMENSION(3)                              :: comp
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: alpha_beta, do_dynamic_load_balancing, &
                                                            do_hfx, hfx_treat_lsd_in_core, &
                                                            restore_p_screen, use_virial
      REAL(KIND=dp)                                      :: e_hartree, factor, out_alpha, &
                                                            pair_energy, tot_rho_r
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: mat_deb
      REAL(KIND=dp), DIMENSION(3, 3)                     :: h_stress
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type)                              :: P_mu_nu
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_mu_nu, matrix_ks, matrix_p_mp2, &
                                                            matrix_s, matrix_w_mp2, rho_ao
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER :: fm_back, fm_G_mu_nu, L_jb, L_jb_beta, mo_coeff_o, &
         mo_coeff_o_beta, mo_coeff_v, mo_coeff_v_beta, P_ia, P_ia_beta, P_mo, P_mo_beta, W_mo, &
         W_mo_beta
      TYPE(hfx_container_type), DIMENSION(:), POINTER    :: integral_containers
      TYPE(hfx_container_type), POINTER                  :: maxval_container
      TYPE(hfx_type), POINTER                            :: actual_x_data
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: dvg(3), pot_g, rho_g, rho_r, temp_pw_g
      TYPE(pw_p_type), POINTER                           :: rho_core
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho, rho_work
      TYPE(section_vals_type), POINTER                   :: hfx_sections, input
      TYPE(virial_type), POINTER                         :: virial

! Set alpha-beta

      alpha_beta = .FALSE.
      IF (PRESENT(homo_beta) .AND. PRESENT(Eigenval_beta) &
          .AND. PRESENT(mo_coeff_beta)) alpha_beta = .TRUE.

      CALL timeset(routineN, handle)

      ! start collecting stuff
      dimen = nmo
      virtual = dimen-homo
      IF (alpha_beta) virtual_beta = dimen-homo_beta
      NULLIFY (input, pw_env, matrix_s, blacs_env, rho, energy, force, virial, matrix_w_mp2, &
               matrix_p_mp2, matrix_ks, rho_core, sab_orb)
      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      pw_env=pw_env, &
                      input=input, &
                      matrix_s=matrix_s, &
                      matrix_ks=matrix_ks, &
                      matrix_p_mp2=matrix_p_mp2, &
                      matrix_w_mp2=matrix_w_mp2, &
                      blacs_env=blacs_env, &
                      rho=rho, &
                      energy=energy, &
                      force=force, &
                      virial=virial, &
                      rho_core=rho_core, &
                      sab_orb=sab_orb)

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      ! check if we have to calculate the virial
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      ! mp2 matrices
      NULLIFY (P_mo, W_mo, L_jb)
      P_mo => mp2_env%ri_grad%P_mo
      W_mo => mp2_env%ri_grad%W_mo
      L_jb => mp2_env%ri_grad%L_jb
      IF (alpha_beta) THEN
         NULLIFY (P_mo_beta, W_mo_beta, L_jb_beta)
         P_mo_beta => mp2_env%ri_grad%P_mo_beta
         W_mo_beta => mp2_env%ri_grad%W_mo_beta
         L_jb_beta => mp2_env%ri_grad%L_jb_beta
      ENDIF

      ! pw stuff
      NULLIFY (poisson_env, pw_pools, auxbas_pw_pool)
      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                      pw_pools=pw_pools, poisson_env=poisson_env)

      ! get some of the grids ready
      NULLIFY (rho_r%pw, rho_g%pw, pot_g%pw)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_r%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, pot_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      ! hfx section
      NULLIFY (hfx_sections)
      hfx_sections => section_vals_get_subs_vals(input, "DFT%XC%HF")
      CALL section_vals_get(hfx_sections, explicit=do_hfx, n_repetition=n_rep_hf)
      IF (do_hfx) THEN
         CALL section_vals_val_get(hfx_sections, "TREAT_LSD_IN_CORE", l_val=hfx_treat_lsd_in_core, &
                                   i_rep_section=1)
      END IF

      ! create work array
      NULLIFY (mat_mu_nu)
      CALL cp_dbcsr_allocate_matrix_set(mat_mu_nu, dft_control%nspins)
      DO ispin = 1, dft_control%nspins
         ALLOCATE (mat_mu_nu(ispin)%matrix)
         CALL cp_dbcsr_init(mat_mu_nu(ispin)%matrix)
         CALL cp_dbcsr_create(matrix=mat_mu_nu(ispin)%matrix, &
                              name="T_mu_nu", &
                              dist=cp_dbcsr_distribution(matrix_s(1)%matrix), matrix_type=dbcsr_type_symmetric, &
                              row_blk_size=cp_dbcsr_row_block_sizes(matrix_s(1)%matrix), &
                              col_blk_size=cp_dbcsr_col_block_sizes(matrix_s(1)%matrix), &
                              nze=0)
         CALL cp_dbcsr_alloc_block_from_nbl(mat_mu_nu(ispin)%matrix, sab_orb)
         CALL cp_dbcsr_set(mat_mu_nu(ispin)%matrix, 0.0_dp)
      END DO

      ALLOCATE (P_mu_nu%matrix)
      CALL cp_dbcsr_init(P_mu_nu%matrix)
      ! CALL cp_dbcsr_create(P_mu_nu%matrix,template=mat_mu_nu(1)%matrix)
      ! CALL cp_dbcsr_copy(P_mu_nu%matrix,mat_mu_nu(1)%matrix,name="P_mu_nu")
      CALL cp_dbcsr_copy(P_mu_nu%matrix, rho_ao(1)%matrix, name="P_mu_nu")
      CALL cp_dbcsr_set(P_mu_nu%matrix, 0.0_dp)

      NULLIFY (fm_G_mu_nu, fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                               nrow_global=dimen, ncol_global=dimen)
      CALL cp_fm_create(fm_G_mu_nu, fm_struct_tmp, name="G_mu_nu")
      CALL cp_fm_create(fm_back, fm_struct_tmp, name="fm_back")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_set_all(fm_G_mu_nu, 0.0_dp)
      CALL cp_fm_set_all(fm_back, 0.0_dp)

      NULLIFY (mo_coeff_o, fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                               nrow_global=dimen, ncol_global=homo)
      CALL cp_fm_create(mo_coeff_o, fm_struct_tmp, name="mo_coeff_o")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_set_all(mo_coeff_o, 0.0_dp)
      CALL cp_fm_to_fm_submat(msource=mo_coeff, mtarget=mo_coeff_o, &
                              nrow=dimen, ncol=homo, &
                              s_firstrow=1, s_firstcol=1, &
                              t_firstrow=1, t_firstcol=1)

      NULLIFY (mo_coeff_v, fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                               nrow_global=dimen, ncol_global=virtual)
      CALL cp_fm_create(mo_coeff_v, fm_struct_tmp, name="mo_coeff_v")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_set_all(mo_coeff_v, 0.0_dp)
      CALL cp_fm_to_fm_submat(msource=mo_coeff, mtarget=mo_coeff_v, &
                              nrow=dimen, ncol=virtual, &
                              s_firstrow=1, s_firstcol=homo+1, &
                              t_firstrow=1, t_firstcol=1)

      IF (alpha_beta) THEN
         NULLIFY (mo_coeff_o_beta, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=dimen, ncol_global=homo_beta)
         CALL cp_fm_create(mo_coeff_o_beta, fm_struct_tmp, name="mo_coeff_o_beta")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_set_all(mo_coeff_o_beta, 0.0_dp)
         CALL cp_fm_to_fm_submat(msource=mo_coeff_beta, mtarget=mo_coeff_o_beta, &
                                 nrow=dimen, ncol=homo_beta, &
                                 s_firstrow=1, s_firstcol=1, &
                                 t_firstrow=1, t_firstcol=1)

         NULLIFY (mo_coeff_v_beta, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=dimen, ncol_global=virtual_beta)
         CALL cp_fm_create(mo_coeff_v_beta, fm_struct_tmp, name="mo_coeff_v_beta")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_set_all(mo_coeff_v_beta, 0.0_dp)
         CALL cp_fm_to_fm_submat(msource=mo_coeff_beta, mtarget=mo_coeff_v_beta, &
                                 nrow=dimen, ncol=virtual_beta, &
                                 s_firstrow=1, s_firstcol=homo_beta+1, &
                                 t_firstrow=1, t_firstcol=1)
      ENDIF

      ! create a working rho environment
      NULLIFY (rho_work)
      CALL qs_rho_create(rho_work)
      CALL qs_rho_rebuild(rho=rho_work, qs_env=qs_env, rebuild_ao=.TRUE., rebuild_grids=.FALSE.)

      ! here we check if we have to reallocate the HFX container
      IF (mp2_env%ri_mp2%free_hfx_buffer) THEN
         CALL timeset(routineN//"_alloc_hfx", handle2)
         n_threads = 1
!$       n_threads = omp_get_max_threads()

         DO irep = 1, n_rep_hf
            DO i_thread = 0, n_threads-1
               actual_x_data => qs_env%x_data(irep, i_thread+1)

               do_dynamic_load_balancing = .TRUE.
               IF (n_threads == 1 .OR. actual_x_data%memory_parameter%do_disk_storage) do_dynamic_load_balancing = .FALSE.

               IF (do_dynamic_load_balancing) THEN
                  my_bin_size = SIZE(actual_x_data%distribution_energy)
               ELSE
                  my_bin_size = 1
               END IF

               IF (.NOT. actual_x_data%memory_parameter%do_all_on_the_fly) THEN
                  !  CALL dealloc_containers(actual_x_data, hfx_do_eval_energy)
                  CALL alloc_containers(actual_x_data, my_bin_size, hfx_do_eval_energy)

                  DO bin = 1, my_bin_size
                     maxval_container => actual_x_data%maxval_container(bin)
                     integral_containers => actual_x_data%integral_containers(:, bin)
                     CALL hfx_init_container(maxval_container, actual_x_data%memory_parameter%actual_memory_usage, .FALSE.)
                     DO i = 1, 64
                        CALL hfx_init_container(integral_containers(i), actual_x_data%memory_parameter%actual_memory_usage, .FALSE.)
                     END DO
                  END DO
               END IF
            END DO
         END DO
         CALL timestop(handle2)
      END IF

      ! set up parameters for P_screening
      restore_p_screen = qs_env%x_data(1, 1)%screening_parameter%do_initial_p_screening
      IF (qs_env%x_data(1, 1)%screening_parameter%do_initial_p_screening) THEN
         IF (mp2_env%ri_mp2%free_hfx_buffer) THEN
            mp2_env%p_screen = .FALSE.
         ELSE
            mp2_env%p_screen = .TRUE.
         ENDIF
      END IF

      ! update Lagrangian with the CPHF like update, occ-occ block, first call (recompute hfx integrals if needed)
      transf_type_in = 1
      transf_type_out = 1
      out_alpha = 0.5_dp
      ! In alpha-beta case, L_bj_alpha has Coulomb and XC alpha-alpha part
      ! and (only) Coulomb alpha-beta part and vice versa.

      ! Complete in closed shell case, alpha-alpha (Coulomb and XC)
      ! part of L_bj(alpha) for open shell
      factor = 4.0_dp
      IF (alpha_beta) factor = 2.0_dp
      CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                            mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                            hfx_sections, energy, n_rep_hf, poisson_env, &
                            rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                            P_mo, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                            L_jb, transf_type_out, &
                            recalc_hfx_integrals=mp2_env%ri_mp2%free_hfx_buffer, &
                            factor=factor)
      IF (alpha_beta) THEN
         ! Alpha-beta (Coulomb) part of L_jb(alpha)
         CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                               mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                               hfx_sections, energy, n_rep_hf, poisson_env, &
                               rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                               P_mo_beta, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                               L_jb, transf_type_out, &
                               mo_coeff_beta=mo_coeff_beta, &
                               mo_coeff_o_beta=mo_coeff_o_beta, &
                               mo_coeff_v_beta=mo_coeff_v_beta, &
                               homo_beta=homo_beta, virtual_beta=virtual_beta, &
                               factor=factor)
         ! Beta-beta (Coulomb and XC) part of L_jb(beta)
         CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                               mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                               Eigenval_beta, hfx_sections, energy, n_rep_hf, &
                               poisson_env, rho_work, pot_g, rho_g, rho_r, mat_mu_nu, &
                               P_mu_nu, P_mo_beta, fm_G_mu_nu, fm_back, transf_type_in, &
                               out_alpha, L_jb_beta, transf_type_out, factor=factor)
         ! Alpha-beta (Coulomb) part of L_jb(beta)
         CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                               mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                               Eigenval_beta, &
                               hfx_sections, energy, n_rep_hf, poisson_env, &
                               rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                               P_mo, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                               L_jb_beta, transf_type_out, &
                               mo_coeff_beta=mo_coeff, &
                               mo_coeff_o_beta=mo_coeff_o, &
                               mo_coeff_v_beta=mo_coeff_v, &
                               homo_beta=homo, virtual_beta=virtual, factor=factor)
      ENDIF

      ! update Lagrangian with the CPHF like update, virt-virt block
      transf_type_in = 2
      transf_type_out = 1
      out_alpha = 0.5_dp
      ! Complete in closed shell case, alpha-alpha (Coulomb and XC)
      ! part of L_bj(alpha) for open shell
      CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                            mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                            hfx_sections, energy, n_rep_hf, poisson_env, &
                            rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                            P_mo, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                            L_jb, transf_type_out, factor=factor)
      IF (alpha_beta) THEN
         ! Alpha-beta (Coulomb) part of L_jb(alpha)
         CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                               mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                               hfx_sections, energy, n_rep_hf, poisson_env, &
                               rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                               P_mo_beta, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                               L_jb, transf_type_out, &
                               mo_coeff_beta=mo_coeff_beta, &
                               mo_coeff_o_beta=mo_coeff_o_beta, &
                               mo_coeff_v_beta=mo_coeff_v_beta, &
                               homo_beta=homo_beta, virtual_beta=virtual_beta, &
                               factor=factor)
         ! Beta-beta (Coulomb and XC) part of L_jb(beta)
         CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                               mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                               Eigenval_beta, hfx_sections, energy, n_rep_hf, &
                               poisson_env, rho_work, pot_g, rho_g, rho_r, mat_mu_nu, &
                               P_mu_nu, P_mo_beta, fm_G_mu_nu, fm_back, transf_type_in, &
                               out_alpha, L_jb_beta, transf_type_out, factor=factor)
         ! Alpha-beta (Coulomb) part of L_jb(beta)
         CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                               mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                               Eigenval_beta, &
                               hfx_sections, energy, n_rep_hf, poisson_env, &
                               rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                               P_mo, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                               L_jb_beta, transf_type_out, &
                               mo_coeff_beta=mo_coeff, &
                               mo_coeff_o_beta=mo_coeff_o, &
                               mo_coeff_v_beta=mo_coeff_v, &
                               homo_beta=homo, virtual_beta=virtual, factor=factor)
      ENDIF
      ! at this point Lagrnagian is completed ready to solve the Z-vector equations
      ! P_ia will contain the solution of these equations
      NULLIFY (P_ia, fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                               nrow_global=homo, ncol_global=virtual)
      CALL cp_fm_create(P_ia, fm_struct_tmp, name="P_ia")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_set_all(P_ia, 0.0_dp)
      IF (alpha_beta) THEN
         NULLIFY (P_ia_beta, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=homo_beta, ncol_global=virtual_beta)
         CALL cp_fm_create(P_ia_beta, fm_struct_tmp, name="P_ia_beta")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_set_all(P_ia_beta, 0.0_dp)
      ENDIF

      IF (.NOT. alpha_beta) THEN
         CALL solve_z_vector_eq_low(qs_env, mp2_env, para_env, homo, virtual, dimen, unit_nr, &
                                    mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, blacs_env, &
                                    hfx_sections, energy, n_rep_hf, &
                                    poisson_env, rho_work, pot_g, rho_g, rho_r, mat_mu_nu, &
                                    P_mu_nu, L_jb, fm_G_mu_nu, fm_back, P_ia)
      ELSE
         CALL solve_z_vector_eq_low(qs_env, mp2_env, para_env, homo, virtual, dimen, unit_nr, &
                                    mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, blacs_env, &
                                    hfx_sections, energy, n_rep_hf, &
                                    poisson_env, rho_work, pot_g, rho_g, rho_r, mat_mu_nu, &
                                    P_mu_nu, L_jb, fm_G_mu_nu, fm_back, P_ia, &
                                    homo_beta, Eigenval_beta, P_ia_beta, mo_coeff_beta, &
                                    mo_coeff_o_beta, mo_coeff_v_beta, L_jb_beta)
      ENDIF

      ! release Lagrangian
      CALL cp_fm_release(L_jb)
      IF (alpha_beta) CALL cp_fm_release(L_jb_beta)

      ! update the MP2-MO density matrix with the occ-virt block
      CALL cp_fm_to_fm_submat(msource=P_ia, mtarget=P_mo, &
                              nrow=homo, ncol=virtual, &
                              s_firstrow=1, s_firstcol=1, &
                              t_firstrow=1, t_firstcol=homo+1)
      CALL cp_fm_release(P_ia)
      ! transpose P_MO matrix (easy way to symmetrize)
      CALL cp_fm_set_all(fm_back, 0.0_dp)
      ! P_mo now is ready
      CALL cp_fm_upper_to_full(matrix=P_mo, work=fm_back)
      IF (alpha_beta) THEN
         ! update the MP2-MO density matrix with the occ-virt block
         CALL cp_fm_to_fm_submat(msource=P_ia_beta, mtarget=P_mo_beta, &
                                 nrow=homo_beta, ncol=virtual_beta, &
                                 s_firstrow=1, s_firstcol=1, &
                                 t_firstrow=1, t_firstcol=homo_beta+1)
         CALL cp_fm_release(P_ia_beta)
         ! transpose P_MO matrix (easy way to symmetrize)
         CALL cp_fm_set_all(fm_back, 0.0_dp)
         ! P_mo now is ready
         CALL cp_fm_upper_to_full(matrix=P_mo_beta, work=fm_back)
      ENDIF

      ! do the final update to MP2 energy weighted matrix W_MO
      CALL cp_fm_get_info(matrix=W_mo, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)
      DO jjB = 1, ncol_local
         j_global = col_indices(jjB)
         IF (j_global <= homo) THEN
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               W_mo%local_data(iiB, jjB) = W_mo%local_data(iiB, jjB)-P_mo%local_data(iiB, jjB)*Eigenval(j_global)
            END DO
         ELSE
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               IF (i_global <= homo) THEN
                  ! virt-occ
                  W_mo%local_data(iiB, jjB) = W_mo%local_data(iiB, jjB)-P_mo%local_data(iiB, jjB)*Eigenval(i_global)
               ELSE
                  ! virt-virt
                  W_mo%local_data(iiB, jjB) = W_mo%local_data(iiB, jjB)-P_mo%local_data(iiB, jjB)*Eigenval(j_global)
               END IF
            END DO
         END IF
      END DO
      IF (alpha_beta) THEN
         CALL cp_fm_get_info(matrix=W_mo_beta, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            IF (j_global <= homo_beta) THEN
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  W_mo_beta%local_data(iiB, jjB) = &
                     W_mo_beta%local_data(iiB, jjB)-P_mo_beta%local_data(iiB, jjB)*Eigenval_beta(j_global)
               END DO
            ELSE
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  IF (i_global <= homo_beta) THEN
                     ! virt-occ
                     W_mo_beta%local_data(iiB, jjB) = &
                        W_mo_beta%local_data(iiB, jjB)-P_mo_beta%local_data(iiB, jjB)*Eigenval_beta(i_global)
                  ELSE
                     ! virt-virt
                     W_mo_beta%local_data(iiB, jjB) = &
                        W_mo_beta%local_data(iiB, jjB)-P_mo_beta%local_data(iiB, jjB)*Eigenval_beta(j_global)
                  END IF
               END DO
            END IF
         END DO
      ENDIF

      ! complete the occ-occ block of W_mo with a CPHF like update
      transf_type_in = 4
      transf_type_out = 2
      out_alpha = -0.5_dp
      ! Complete in closed shell case, alpha-alpha (Coulomb and XC)
      ! part of W_mo(III)(alpha) for open shell
      CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                            mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                            hfx_sections, energy, n_rep_hf, poisson_env, &
                            rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                            P_mo, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                            W_mo, transf_type_out, factor=factor)
      IF (alpha_beta) THEN
         ! Alpha-beta (Coulomb) part of W_mo(III)(alpha)
         CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                               mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                               hfx_sections, energy, n_rep_hf, poisson_env, &
                               rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                               P_mo_beta, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                               W_mo, transf_type_out, &
                               mo_coeff_beta=mo_coeff_beta, &
                               mo_coeff_o_beta=mo_coeff_o_beta, &
                               mo_coeff_v_beta=mo_coeff_v_beta, &
                               homo_beta=homo_beta, virtual_beta=virtual_beta, &
                               factor=factor)
         ! Beta-beta (Coulomb and XC) part of W_mo(III)(beta)
         CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                               mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                               Eigenval_beta, hfx_sections, energy, n_rep_hf, &
                               poisson_env, rho_work, pot_g, rho_g, rho_r, mat_mu_nu, &
                               P_mu_nu, P_mo_beta, fm_G_mu_nu, fm_back, transf_type_in, &
                               out_alpha, W_mo_beta, transf_type_out, factor=factor)
         ! Alpha-beta (Coulomb) part of W_mo(III)(beta)
         CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                               mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                               Eigenval_beta, &
                               hfx_sections, energy, n_rep_hf, poisson_env, &
                               rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                               P_mo, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                               W_mo_beta, transf_type_out, &
                               mo_coeff_beta=mo_coeff, &
                               mo_coeff_o_beta=mo_coeff_o, &
                               mo_coeff_v_beta=mo_coeff_v, &
                               homo_beta=homo, virtual_beta=virtual, factor=factor)
      ENDIF

      ! release DBCSR stuff
      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_release(mat_mu_nu(ispin)%matrix)
         DEALLOCATE (mat_mu_nu(ispin)%matrix)
      END DO
      DEALLOCATE (mat_mu_nu)
      CALL cp_dbcsr_release(P_mu_nu%matrix)
      DEALLOCATE (P_mu_nu%matrix)
      ! release fm stuff
      CALL cp_fm_release(fm_G_mu_nu)
      CALL cp_fm_release(mo_coeff_o)
      CALL cp_fm_release(mo_coeff_v)
      IF (alpha_beta) THEN
         CALL cp_fm_release(mo_coeff_o_beta)
         CALL cp_fm_release(mo_coeff_v_beta)
      ENDIF
      ! release rho stuff
      CALL qs_rho_release(rho_struct=rho_work)

      IF (.FALSE.) THEN
         ALLOCATE (mat_deb(dimen, dimen))
         mat_deb = 0.0_dp
         CALL cp_fm_get_info(matrix=P_mo, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               mat_deb(i_global, j_global) = P_mo%local_data(iiB, jjB)
            END DO
         END DO
         CALL mp_sum(mat_deb, para_env%group)
         IF (para_env%mepos == 0) CALL write_array(mat_deb(1:dimen, 1:dimen))
         mat_deb = 0.0_dp
         CALL cp_fm_get_info(matrix=W_mo, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               mat_deb(i_global, j_global) = W_mo%local_data(iiB, jjB)
            END DO
         END DO
         CALL mp_sum(mat_deb, para_env%group)
         IF (para_env%mepos == 0) CALL write_array(mat_deb(1:dimen, 1:dimen))
         DEALLOCATE (mat_deb)
         IF (alpha_beta) THEN
            ALLOCATE (mat_deb(dimen, dimen))
            mat_deb = 0.0_dp
            CALL cp_fm_get_info(matrix=P_mo_beta, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  mat_deb(i_global, j_global) = P_mo_beta%local_data(iiB, jjB)
               END DO
            END DO
            CALL mp_sum(mat_deb, para_env%group)
            IF (para_env%mepos == 0) CALL write_array(mat_deb(1:dimen, 1:dimen))
            mat_deb = 0.0_dp
            CALL cp_fm_get_info(matrix=W_mo_beta, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  mat_deb(i_global, j_global) = W_mo_beta%local_data(iiB, jjB)
               END DO
            END DO
            CALL mp_sum(mat_deb, para_env%group)
            IF (para_env%mepos == 0) CALL write_array(mat_deb(1:dimen, 1:dimen))
            DEALLOCATE (mat_deb)
         ENDIF
      END IF

      ! backtransform into AO basis, since P_mo and W_mo
      ! are symmetric (in principle), no need to symmetrize
      ! first W_mo
      CALL cp_fm_set_all(fm_back, 0.0_dp)
      CALL cp_gemm('N', 'N', dimen, dimen, dimen, 1.0_dp, &
                   mo_coeff, W_mo, 0.0_dp, fm_back, &
                   a_first_col=1, &
                   a_first_row=1, &
                   b_first_col=1, &
                   b_first_row=1, &
                   c_first_col=1, &
                   c_first_row=1)
      CALL cp_fm_set_all(W_mo, 0.0_dp)
      CALL cp_gemm('N', 'T', dimen, dimen, dimen, 1.0_dp, &
                   fm_back, mo_coeff, 0.0_dp, W_mo, &
                   a_first_col=1, &
                   a_first_row=1, &
                   b_first_col=1, &
                   b_first_row=1, &
                   c_first_col=1, &
                   c_first_row=1)
      ! and P_mo
      CALL cp_fm_set_all(fm_back, 0.0_dp)
      CALL cp_gemm('N', 'N', dimen, dimen, dimen, 1.0_dp, &
                   mo_coeff, P_mo, 0.0_dp, fm_back, &
                   a_first_col=1, &
                   a_first_row=1, &
                   b_first_col=1, &
                   b_first_row=1, &
                   c_first_col=1, &
                   c_first_row=1)
      CALL cp_fm_set_all(P_mo, 0.0_dp)
      CALL cp_gemm('N', 'T', dimen, dimen, dimen, 1.0_dp, &
                   fm_back, mo_coeff, 0.0_dp, P_mo, &
                   a_first_col=1, &
                   a_first_row=1, &
                   b_first_col=1, &
                   b_first_row=1, &
                   c_first_col=1, &
                   c_first_row=1)
      IF (alpha_beta) THEN
         CALL cp_fm_set_all(fm_back, 0.0_dp)
         CALL cp_gemm('N', 'N', dimen, dimen, dimen, 1.0_dp, &
                      mo_coeff_beta, W_mo_beta, 0.0_dp, fm_back, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
         CALL cp_fm_set_all(W_mo_beta, 0.0_dp)
         CALL cp_gemm('N', 'T', dimen, dimen, dimen, 1.0_dp, &
                      fm_back, mo_coeff_beta, 0.0_dp, W_mo_beta, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
         ! and P_mo
         CALL cp_fm_set_all(fm_back, 0.0_dp)
         CALL cp_gemm('N', 'N', dimen, dimen, dimen, 1.0_dp, &
                      mo_coeff_beta, P_mo_beta, 0.0_dp, fm_back, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
         CALL cp_fm_set_all(P_mo_beta, 0.0_dp)
         CALL cp_gemm('N', 'T', dimen, dimen, dimen, 1.0_dp, &
                      fm_back, mo_coeff_beta, 0.0_dp, P_mo_beta, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
      ENDIF

      ! copy W_mo into dbcsr
      CALL copy_fm_to_dbcsr(W_mo, matrix_w_mp2(1)%matrix, keep_sparsity=.TRUE.)
      IF (alpha_beta) THEN
         CALL copy_fm_to_dbcsr(W_mo_beta, matrix_w_mp2(2)%matrix, keep_sparsity=.TRUE.)
      ENDIF

      ! create mp2 DBCSR density
      CALL cp_dbcsr_allocate_matrix_set(matrix_p_mp2, dft_control%nspins)
      DO ispin = 1, dft_control%nspins
         ALLOCATE (matrix_p_mp2(ispin)%matrix)
         CALL cp_dbcsr_init(matrix_p_mp2(ispin)%matrix)
         CALL cp_dbcsr_copy(matrix_p_mp2(ispin)%matrix, rho_ao(ispin)%matrix, &
                            name="P MATRIX MP2")
         CALL cp_dbcsr_set(matrix_p_mp2(ispin)%matrix, 0.0_dp)
         IF (ispin == 1) &
            CALL copy_fm_to_dbcsr(P_mo, matrix_p_mp2(ispin)%matrix, keep_sparsity=.TRUE.)
         IF (ispin == 2) &
            CALL copy_fm_to_dbcsr(P_mo_beta, matrix_p_mp2(ispin)%matrix, keep_sparsity=.TRUE.)
      END DO
      CALL set_ks_env(ks_env, matrix_p_mp2=matrix_p_mp2)

      CALL cp_fm_release(fm_back)

      ! release remaining fm stuff
      CALL cp_fm_release(W_mo)
      CALL cp_fm_release(P_mo)
      IF (alpha_beta) THEN
         CALL cp_fm_release(W_mo_beta)
         CALL cp_fm_release(P_mo_beta)
      ENDIF

      ! update the core-forces with the MP2-density contribution
      ! put MP2 density on the grid
      IF (alpha_beta) THEN ! In alpha_beta case, get the joint density
         CALL cp_dbcsr_add(matrix_p_mp2(1)%matrix, matrix_p_mp2(2)%matrix, 1.0_dp, 1.0_dp)
      ENDIF
      CALL calculate_rho_elec(matrix_p=matrix_p_mp2(1)%matrix, &
                              rho=rho_r, &
                              rho_gspace=rho_g, &
                              total_rho=tot_rho_r, &
                              ks_env=ks_env, &
                              soft_valid=.FALSE.)
      ! calculate the MP2 potential
      CALL pw_transfer(rho_r%pw, rho_g%pw)
      CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
      CALL pw_transfer(pot_g%pw, rho_r%pw)
      CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)

      ! calculate core forces
      CALL integrate_v_core_rspace(rho_r, qs_env)
      DO ikind = 1, SIZE(atomic_kind_set)
         force(ikind)%mp2_sep = force(ikind)%rho_core
         force(ikind)%rho_core = 0.0_dp
      ENDDO
      ! right contribution
      IF (alpha_beta) THEN ! In alpha_beta case, get the joint density
         CALL cp_dbcsr_add(rho_ao(1)%matrix, rho_ao(2)%matrix, 1.0_dp, 1.0_dp)
         CALL cp_dbcsr_add(matrix_ks(1)%matrix, matrix_ks(2)%matrix, 1.0_dp, 1.0_dp)
      ENDIF
      CALL integrate_v_rspace(v_rspace=rho_r, pmat=rho_ao(1), hmat=matrix_ks(1), &
                              qs_env=qs_env, calculate_forces=.TRUE.)

      IF (use_virial) THEN
         ! update virial if necessary with the volume term
         ! first create pw auxilliary stuff
         CALL timeset(routineN//"_Virial", handle2)
         NULLIFY (temp_pw_g%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, temp_pw_g%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
         DO i = 1, 3
            NULLIFY (dvg(i)%pw)
            CALL pw_pool_create_pw(auxbas_pw_pool, dvg(i)%pw, &
                                   use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
         END DO

         ! make a copy of the MP2 density in G space
         CALL pw_copy(rho_g%pw, temp_pw_g%pw)
         ! calculate MP2-like-hartree potential derivatives
         DO i = 1, 3
            comp = 0
            comp(i) = 1
            CALL pw_copy(pot_g%pw, dvg(i)%pw)
            CALL pw_derive(dvg(i)%pw, comp)
         END DO

         ! calculate total SCF density and potential
         CALL calculate_rho_elec(matrix_p=rho_ao(1)%matrix, &
                                 rho=rho_r, &
                                 rho_gspace=rho_g, &
                                 total_rho=tot_rho_r, &
                                 ks_env=ks_env, &
                                 soft_valid=.FALSE.)
         ! and associated potential
         CALL pw_transfer(rho_r%pw, rho_g%pw)
         ! don't forget the core density
         CALL pw_axpy(rho_core%pw, rho_g%pw)
         CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)

         ! finally update virial with the volume contribution
         e_hartree = pw_integral_ab(temp_pw_g%pw, pot_g%pw)
         h_stress = 0.0_dp
         DO alpha = 1, 3
            comp = 0
            comp(alpha) = 1
            CALL pw_copy(pot_g%pw, rho_g%pw)
            CALL pw_derive(rho_g%pw, comp)
            h_stress(alpha, alpha) = -e_hartree
            DO beta = alpha, 3
               h_stress(alpha, beta) = h_stress(alpha, beta) &
                                       -2.0_dp*pw_integral_ab(rho_g%pw, dvg(beta)%pw)/fourpi
               h_stress(beta, alpha) = h_stress(alpha, beta)
            END DO
         END DO
         virial%pv_virial = virial%pv_virial+h_stress/REAL(para_env%num_pe, dp)

         ! free stuff
         CALL pw_pool_give_back_pw(auxbas_pw_pool, temp_pw_g%pw)
         DO i = 1, 3
            CALL pw_pool_give_back_pw(auxbas_pw_pool, dvg(i)%pw)
         END DO
         CALL timestop(handle2)
      END IF

      IF (alpha_beta) THEN ! In alpha_beta case, get the initial densities back
         CALL cp_dbcsr_add(rho_ao(1)%matrix, rho_ao(2)%matrix, 1.0_dp, -1.0_dp)
         CALL cp_dbcsr_add(matrix_p_mp2(1)%matrix, matrix_p_mp2(2)%matrix, 1.0_dp, -1.0_dp)
         CALL cp_dbcsr_add(matrix_ks(1)%matrix, matrix_ks(2)%matrix, 1.0_dp, -1.0_dp)
      ENDIF

      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_add(rho_ao(ispin)%matrix, matrix_p_mp2(ispin)%matrix, 1.0_dp, 1.0_dp)
      END DO
      ! We will need one more hfx calculation for HF gradient part
      mp2_env%not_last_hfx = .FALSE.
      mp2_env%p_screen = restore_p_screen

      ! release stuff
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_g%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, pot_g%pw)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Here we performe the CPHF like update using GPW,
!>        transf_type_in  defines the type of transformation for the matrix in input
!>        transf_type_in = 1 -> occ-occ back transformation
!>        transf_type_in = 2 -> virt-virt back transformation
!>        transf_type_in = 3 -> occ-virt back transformation including the
!>                              eigenvalues energy differences for the diagonal elements
!>        transf_type_in = 4 -> full range
!>        transf_type_out defines the type of transformation for the matrix in output
!>        transf_type_out = 1 -> occ-vit transformation
!>        transf_type_out = 2 -> occ-occ transformation
!> \param qs_env ...
!> \param para_env ...
!> \param homo ...
!> \param virtual ...
!> \param dimen ...
!> \param mo_coeff ...
!> \param mo_coeff_o ...
!> \param mo_coeff_v ...
!> \param Eigenval ...
!> \param hfx_sections ...
!> \param energy ...
!> \param n_rep_hf ...
!> \param poisson_env ...
!> \param rho_work ...
!> \param pot_g ...
!> \param rho_g ...
!> \param rho_r ...
!> \param mat_mu_nu ...
!> \param P_mu_nu ...
!> \param fm_mo ...
!> \param fm_ao ...
!> \param fm_back ...
!> \param transf_type_in ...
!> \param out_alpha ...
!> \param fm_mo_out ...
!> \param transf_type_out ...
!> \param recalc_hfx_integrals ...
!> \param mo_coeff_beta ...
!> \param mo_coeff_o_beta ...
!> \param mo_coeff_v_beta ...
!> \param homo_beta ...
!> \param virtual_beta ...
!> \param factor ...
!> \author Mauro Del Ben, Vladimir Rybkin
! **************************************************************************************************
   SUBROUTINE cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                               mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                               hfx_sections, energy, n_rep_hf, poisson_env, &
                               rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                               fm_mo, fm_ao, fm_back, transf_type_in, out_alpha, &
                               fm_mo_out, transf_type_out, recalc_hfx_integrals, &
                               mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                               homo_beta, virtual_beta, factor)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: homo, virtual, dimen
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_o, mo_coeff_v
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      TYPE(section_vals_type), POINTER                   :: hfx_sections
      TYPE(qs_energy_type), POINTER                      :: energy
      INTEGER                                            :: n_rep_hf
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(qs_rho_type), POINTER                         :: rho_work
      TYPE(pw_p_type)                                    :: pot_g, rho_g, rho_r
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_mu_nu
      TYPE(cp_dbcsr_p_type)                              :: P_mu_nu
      TYPE(cp_fm_type), POINTER                          :: fm_mo, fm_ao, fm_back
      INTEGER                                            :: transf_type_in
      REAL(KIND=dp)                                      :: out_alpha
      TYPE(cp_fm_type), POINTER                          :: fm_mo_out
      INTEGER                                            :: transf_type_out
      LOGICAL, OPTIONAL                                  :: recalc_hfx_integrals
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: mo_coeff_beta, mo_coeff_o_beta, &
                                                            mo_coeff_v_beta
      INTEGER, OPTIONAL                                  :: homo_beta, virtual_beta
      REAL(KIND=dp), OPTIONAL                            :: factor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cphf_like_update', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_global, iiB, irep, j_global, &
                                                            jjB, ncol_local, nrow_local, ns
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: alpha_beta, my_recalc_hfx_integrals
      REAL(KIND=dp)                                      :: ex_energy, pair_energy, total_rho
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_work_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: mat_2d, rho_ao_2d
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      alpha_beta = .FALSE.
      IF (PRESENT(mo_coeff_beta)) alpha_beta = .TRUE.

      NULLIFY (ks_env, rho_work_ao)
      CALL timeset(routineN, handle)

      my_recalc_hfx_integrals = .FALSE.
      IF (PRESENT(recalc_hfx_integrals)) my_recalc_hfx_integrals = recalc_hfx_integrals

      CALL get_qs_env(qs_env, ks_env=ks_env)
      ! perform back transformation
      SELECT CASE (transf_type_in)
      CASE (1)
         IF (.NOT. alpha_beta) THEN
            ! occ-occ block
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'N', dimen, homo, homo, 1.0_dp, &
                         mo_coeff_o, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_fm_set_all(fm_ao, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, dimen, homo, 1.0_dp, &
                         fm_back, mo_coeff_o, 0.0_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
         ELSE
            ! occ-occ block
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'N', dimen, homo_beta, homo_beta, 1.0_dp, &
                         mo_coeff_o_beta, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_fm_set_all(fm_ao, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, dimen, homo_beta, 1.0_dp, &
                         fm_back, mo_coeff_o_beta, 0.0_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
         ENDIF

      CASE (2)
         IF (.NOT. alpha_beta) THEN
            ! virt-virt block
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'N', dimen, virtual, virtual, 1.0_dp, &
                         mo_coeff_v, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=homo+1, &
                         b_first_row=homo+1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_fm_set_all(fm_ao, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, dimen, virtual, 1.0_dp, &
                         fm_back, mo_coeff_v, 0.0_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
         ELSE
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'N', dimen, virtual_beta, virtual_beta, 1.0_dp, &
                         mo_coeff_v_beta, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=homo_beta+1, &
                         b_first_row=homo_beta+1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_fm_set_all(fm_ao, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, dimen, virtual_beta, 1.0_dp, &
                         fm_back, mo_coeff_v_beta, 0.0_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
         ENDIF

      CASE (3)
         IF (.NOT. alpha_beta) THEN
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'N', dimen, virtual, homo, 1.0_dp, &
                         mo_coeff_o, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_fm_set_all(fm_ao, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, dimen, virtual, 1.0_dp, &
                         fm_back, mo_coeff_v, 0.0_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            ! and symmetrize (here again multiply instead of transposing)
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, homo, virtual, 1.0_dp, &
                         mo_coeff_v, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_gemm('N', 'T', dimen, dimen, homo, 0.5_dp, &
                         fm_back, mo_coeff_o, 0.5_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            ! scale for the orbital energy differences for the diagonal elements
            fm_mo_out%local_data(:, :) = fm_mo%local_data(:, :)
            CALL cp_fm_get_info(matrix=fm_mo_out, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  fm_mo_out%local_data(iiB, jjB) = fm_mo_out%local_data(iiB, jjB)* &
                                                   (Eigenval(j_global+homo)-Eigenval(i_global))
               END DO
            END DO
         ELSE
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'N', dimen, virtual_beta, homo_beta, 1.0_dp, &
                         mo_coeff_o_beta, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_fm_set_all(fm_ao, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, dimen, virtual_beta, 1.0_dp, &
                         fm_back, mo_coeff_v_beta, 0.0_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            ! and symmetrize (here again multiply instead of transposing)
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, homo_beta, virtual_beta, 1.0_dp, &
                         mo_coeff_v_beta, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_gemm('N', 'T', dimen, dimen, homo_beta, 0.5_dp, &
                         fm_back, mo_coeff_o_beta, 0.5_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
         ENDIF

      CASE (4)
         ! all-all block
         IF (.NOT. alpha_beta) THEN
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'N', dimen, dimen, dimen, 1.0_dp, &
                         mo_coeff, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_fm_set_all(fm_ao, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, dimen, dimen, 1.0_dp, &
                         fm_back, mo_coeff, 0.0_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
         ELSE
            CALL cp_fm_set_all(fm_back, 0.0_dp)
            CALL cp_gemm('N', 'N', dimen, dimen, dimen, 1.0_dp, &
                         mo_coeff_beta, fm_mo, 0.0_dp, fm_back, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
            CALL cp_fm_set_all(fm_ao, 0.0_dp)
            CALL cp_gemm('N', 'T', dimen, dimen, dimen, 1.0_dp, &
                         fm_back, mo_coeff_beta, 0.0_dp, fm_ao, &
                         a_first_col=1, &
                         a_first_row=1, &
                         b_first_col=1, &
                         b_first_row=1, &
                         c_first_col=1, &
                         c_first_row=1)
         ENDIF

      CASE DEFAULT
         ! nothing
      END SELECT

      ! copy fm into DBCSR
      CALL cp_dbcsr_set(P_mu_nu%matrix, 0.0_dp)
      CALL copy_fm_to_dbcsr(fm_ao, P_mu_nu%matrix, keep_sparsity=.TRUE.)

      ! calculate associated density
      CALL calculate_rho_elec(matrix_p=P_mu_nu%matrix, &
                              rho=rho_r, &
                              rho_gspace=rho_g, &
                              total_rho=total_rho, &
                              ks_env=ks_env)
      ! and calculate potential
      CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
      CALL pw_transfer(pot_g%pw, rho_r%pw)
      CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)
      ! integrate the potential
      CALL cp_dbcsr_set(mat_mu_nu(1)%matrix, 0.0_dp)
      CALL integrate_v_rspace(rho_r, hmat=mat_mu_nu(1), &
                              qs_env=qs_env, calculate_forces=.FALSE., compute_tau=.FALSE., gapw=.FALSE.)

      ! update with the exchange like contributions
      ! copy mat_mu_nu into rho_ao work

      ! Only for alpha-alpha and beta-beta
      IF (.NOT. alpha_beta) THEN
         CALL qs_rho_get(rho_work, rho_ao=rho_work_ao)
         CALL cp_dbcsr_set(rho_work_ao(1)%matrix, 0.0_dp)
         CALL cp_dbcsr_copy(rho_work_ao(1)%matrix, P_mu_nu%matrix)
         ! save old EX energy
         ex_energy = energy%ex
         DO irep = 1, n_rep_hf
            ns = SIZE(rho_work_ao)
            rho_ao_2d(1:ns, 1:1) => rho_work_ao(1:ns)
            ns = SIZE(mat_mu_nu)
            mat_2d(1:ns, 1:1) => mat_mu_nu(1:ns)
            CALL integrate_four_center(qs_env, mat_2d, energy, rho_ao_2d, hfx_sections, &
                                       para_env, my_recalc_hfx_integrals, irep, .TRUE., &
                                       ispin=1)
         END DO
         ! restore original EX energy
         energy%ex = ex_energy
      ENDIF

      ! scale by a factor 4.0 (closed shell) or 2.0 (open-shell)
      CALL cp_dbcsr_scale(mat_mu_nu(1)%matrix, factor)

      ! copy back to fm
      CALL cp_fm_set_all(fm_ao, 0.0_dp)
      CALL copy_dbcsr_to_fm(matrix=mat_mu_nu(1)%matrix, fm=fm_ao)
      CALL cp_fm_set_all(fm_back, 0.0_dp)
      CALL cp_fm_upper_to_full(fm_ao, fm_back)

      ! transform to MO basis, here we always sum the result into the input matrix
      SELECT CASE (transf_type_out)
         ! In alpha-beta case the density is contracted with the orbitals of different
         ! spin. Thus, there's branching at forward trasform and no branching here,
         ! at back transform.
      CASE (1)
         ! occ-virt block
         CALL cp_fm_set_all(fm_back, 0.0_dp)
         CALL cp_gemm('T', 'N', homo, dimen, dimen, 1.0_dp, &
                      mo_coeff_o, fm_ao, 0.0_dp, fm_back, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
         CALL cp_gemm('N', 'N', homo, virtual, dimen, out_alpha, &
                      fm_back, mo_coeff_v, 1.0_dp, fm_mo_out, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)

      CASE (2)
         ! occ-occ block
         CALL cp_fm_set_all(fm_back, 0.0_dp)
         CALL cp_gemm('T', 'N', homo, dimen, dimen, 1.0_dp, &
                      mo_coeff_o, fm_ao, 0.0_dp, fm_back, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
         CALL cp_gemm('N', 'N', homo, homo, dimen, out_alpha, &
                      fm_back, mo_coeff_o, 1.0_dp, fm_mo_out, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)

      CASE DEFAULT
         ! nothing
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE cphf_like_update

! **************************************************************************************************
!> \brief Low level subroutine for the iterative solution of a large
!>        system of linear equation
!> \param qs_env ...
!> \param mp2_env ...
!> \param para_env ...
!> \param homo ...
!> \param virtual ...
!> \param dimen ...
!> \param unit_nr ...
!> \param mo_coeff ...
!> \param mo_coeff_o ...
!> \param mo_coeff_v ...
!> \param Eigenval ...
!> \param blacs_env ...
!> \param hfx_sections ...
!> \param energy ...
!> \param n_rep_hf ...
!> \param poisson_env ...
!> \param rho_work ...
!> \param pot_g ...
!> \param rho_g ...
!> \param rho_r ...
!> \param mat_mu_nu ...
!> \param P_mu_nu ...
!> \param L_jb ...
!> \param fm_G_mu_nu ...
!> \param fm_back ...
!> \param P_ia ...
!> \param homo_beta ...
!> \param Eigenval_beta ...
!> \param P_ia_beta ...
!> \param mo_coeff_beta ...
!> \param mo_coeff_o_beta ...
!> \param mo_coeff_v_beta ...
!> \param L_jb_beta ...
!> \author Mauro Del Ben, Vladimir Rybkin
! **************************************************************************************************
   SUBROUTINE solve_z_vector_eq_low(qs_env, mp2_env, para_env, homo, virtual, dimen, unit_nr, &
                                    mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, blacs_env, &
                                    hfx_sections, energy, n_rep_hf, poisson_env, &
                                    rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                                    L_jb, fm_G_mu_nu, fm_back, P_ia, homo_beta, Eigenval_beta, &
                                    P_ia_beta, mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                                    L_jb_beta)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: homo, virtual, dimen, unit_nr
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_o, mo_coeff_v
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(section_vals_type), POINTER                   :: hfx_sections
      TYPE(qs_energy_type), POINTER                      :: energy
      INTEGER                                            :: n_rep_hf
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(qs_rho_type), POINTER                         :: rho_work
      TYPE(pw_p_type)                                    :: pot_g, rho_g, rho_r
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_mu_nu
      TYPE(cp_dbcsr_p_type)                              :: P_mu_nu
      TYPE(cp_fm_type), POINTER                          :: L_jb, fm_G_mu_nu, fm_back, P_ia
      INTEGER, OPTIONAL                                  :: homo_beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: P_ia_beta, mo_coeff_beta, &
                                                            mo_coeff_o_beta, mo_coeff_v_beta, &
                                                            L_jb_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'solve_z_vector_eq_low', &
         routineP = moduleN//':'//routineN

      INTEGER :: cycle_counter, handle, i_global, iiB, iiter, j_global, jjB, max_num_iter, &
         ncol_local, ncol_local_b, nrow_local, nrow_local_b, transf_type_in, transf_type_out, &
         virtual_beta
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: alpha_beta, converged
      REAL(KIND=dp) :: alpha, beta, conv, conv_b, eps_conv, factor, norm_b, norm_b_beta, norms(3), &
         norms_beta(3), out_alpha, rkrk, rkrk_beta, t1, t2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: proj_bi_xj, temp_vals, x_norm, x_norm_b, &
                                                            xi_b
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: A_small, b_small, xi_Axi
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: Ax, Ax_ab, Ax_ba, Ax_bb, xn, xn_b
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp, fm_struct_tmp_b
      TYPE(cp_fm_type), POINTER                          :: Ap, Ap_b, b_i, b_i_b, pk, pk_b, precond, &
                                                            precond_b, residual, residual_b, rk, &
                                                            rk_b, xk, xk_b

      CALL timeset(routineN, handle)

      alpha_beta = .FALSE.
      factor = 4.0_dp
      IF (PRESENT(homo_beta) .AND. PRESENT(Eigenval_beta)) alpha_beta = .TRUE.
      IF (alpha_beta) THEN
         virtual_beta = dimen-homo_beta
         factor = 2.0_dp
      ENDIF

      max_num_iter = mp2_env%ri_grad%cphf_max_num_iter
      eps_conv = mp2_env%ri_grad%cphf_eps_conv

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, *)
         WRITE (unit_nr, '(T3,A)') 'MP2_CPHF| Iterative solution of Z-Vector equations'
         WRITE (unit_nr, '(T3,A,T45,ES8.1)') 'MP2_CPHF| Convergence threshold:', eps_conv
         WRITE (unit_nr, '(T3,A,T45,I8)') 'MP2_CPHF| Maximum number of iterations: ', max_num_iter
         WRITE (unit_nr, '(T4,A)') REPEAT("-", 40)
         WRITE (unit_nr, '(T4,A,T15,A,T33,A)') 'Step', 'Time', 'Convergence'
         WRITE (unit_nr, '(T4,A)') REPEAT("-", 40)
      END IF

      ! set the transformation type (equal for all methods all updates)
      transf_type_in = 3
      transf_type_out = 1
      out_alpha = 1.0_dp

      ! set convergece flag
      converged = .FALSE.

      IF ((.FALSE.)) THEN
         ! CG algorithm
         ! Used for open-shell systems
         ! create some work array
         NULLIFY (xk, pk, rk, Ap, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=homo, ncol_global=virtual)
         CALL cp_fm_create(xk, fm_struct_tmp, name="xk")
         CALL cp_fm_create(pk, fm_struct_tmp, name="pk")
         CALL cp_fm_create(rk, fm_struct_tmp, name="rk")
         CALL cp_fm_create(Ap, fm_struct_tmp, name="Ap")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_set_all(xk, 0.0_dp)
         CALL cp_fm_set_all(pk, 0.0_dp)
         CALL cp_fm_set_all(rk, 0.0_dp)
         CALL cp_fm_set_all(Ap, 0.0_dp)

         ! copy -L_jb into pk and rk
         pk%local_data(:, :) = -L_jb%local_data(:, :)
         rk%local_data(:, :) = -L_jb%local_data(:, :)
         norm_b = 0.0_dp
         norm_b = SUM(L_jb%local_data(:, :)*L_jb%local_data(:, :))
         CALL mp_sum(norm_b, para_env%group)

         IF (alpha_beta) THEN
            NULLIFY (xk_b, pk_b, rk_b, Ap_b, fm_struct_tmp)
            CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                     nrow_global=homo_beta, ncol_global=virtual_beta)
            CALL cp_fm_create(xk_b, fm_struct_tmp, name="xk")
            CALL cp_fm_create(pk_b, fm_struct_tmp, name="pk")
            CALL cp_fm_create(rk_b, fm_struct_tmp, name="rk")
            CALL cp_fm_create(Ap_b, fm_struct_tmp, name="Ap")
            CALL cp_fm_struct_release(fm_struct_tmp)
            CALL cp_fm_set_all(xk_b, 0.0_dp)
            CALL cp_fm_set_all(pk_b, 0.0_dp)
            CALL cp_fm_set_all(rk_b, 0.0_dp)
            CALL cp_fm_set_all(Ap_b, 0.0_dp)

            ! copy -L_jb_beta into pk_b and rk_b
            pk_b%local_data(:, :) = -L_jb_beta%local_data(:, :)
            rk_b%local_data(:, :) = -L_jb_beta%local_data(:, :)
            norm_b_beta = 0.0_dp
            norm_b_beta = SUM(L_jb_beta%local_data(:, :)*L_jb_beta%local_data(:, :))
            CALL mp_sum(norm_b_beta, para_env%group)
            norm_b = norm_b+norm_b_beta
         ENDIF
         ! Set the norm
         norm_b = SQRT(norm_b)

         cycle_counter = 0
         DO iiter = 1, max_num_iter
            cycle_counter = cycle_counter+1
            t1 = m_walltime()

            ! calculate matrix-vector product
            CALL cp_fm_set_all(Ap, 0.0_dp)
            ! Full for closed shell. Alpha-alpha part of alpha for open shell.
            CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                                  mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                                  hfx_sections, energy, n_rep_hf, poisson_env, &
                                  rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                                  pk, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                                  Ap, transf_type_out, factor=factor)
            IF (alpha_beta) THEN
               ! Alpha-beta part of alpha.
               CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                                     mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                                     hfx_sections, energy, n_rep_hf, poisson_env, &
                                     rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                                     pk_b, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                                     Ap, transf_type_out, factor=factor, &
                                     mo_coeff_beta=mo_coeff_beta, &
                                     mo_coeff_o_beta=mo_coeff_o_beta, &
                                     mo_coeff_v_beta=mo_coeff_v_beta, &
                                     homo_beta=homo_beta, virtual_beta=virtual_beta)
               ! Beta-beta part (Coulomb and XC) of beta.
               CALL cp_fm_set_all(Ap_b, 0.0_dp)
               CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                                     mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                                     Eigenval_beta, hfx_sections, energy, n_rep_hf, &
                                     poisson_env, rho_work, pot_g, rho_g, rho_r, mat_mu_nu, &
                                     P_mu_nu, Pk_b, fm_G_mu_nu, fm_back, transf_type_in, &
                                     out_alpha, Ap_b, transf_type_out, factor=factor)
               ! Beta-alpha part of beta.
               CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                                     mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                                     Eigenval_beta, &
                                     hfx_sections, energy, n_rep_hf, poisson_env, &
                                     rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                                     pk, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                                     Ap_b, transf_type_out, factor=factor, &
                                     mo_coeff_beta=mo_coeff, &
                                     mo_coeff_o_beta=mo_coeff_o, &
                                     mo_coeff_v_beta=mo_coeff_v, &
                                     homo_beta=homo, virtual_beta=virtual)
            ENDIF

            norms = 0.0_dp
            norms(1) = SUM(rk%local_data(:, :)*rk%local_data(:, :))
            norms(2) = SUM(rk%local_data(:, :)*pk%local_data(:, :))
            norms(3) = SUM(pk%local_data(:, :)*Ap%local_data(:, :))
            CALL mp_sum(norms, para_env%group)
            IF (alpha_beta) THEN
               norms_beta = 0.0_dp
               norms_beta(1) = SUM(rk_b%local_data(:, :)*rk_b%local_data(:, :))
               norms_beta(2) = SUM(rk_b%local_data(:, :)*pk_b%local_data(:, :))
               norms_beta(3) = SUM(pk_b%local_data(:, :)*Ap_b%local_data(:, :))
               CALL mp_sum(norms_beta, para_env%group)
               norms = norms+norms_beta
            ENDIF
            alpha = norms(1)/norms(3)

            xk%local_data(:, :) = xk%local_data(:, :)+alpha*pk%local_data(:, :)
            rk%local_data(:, :) = rk%local_data(:, :)-alpha*Ap%local_data(:, :)
            IF (alpha_beta) THEN
               xk_b%local_data(:, :) = xk_b%local_data(:, :)+alpha*pk_b%local_data(:, :)
               rk_b%local_data(:, :) = rk_b%local_data(:, :)-alpha*Ap_b%local_data(:, :)
            ENDIF

            rkrk = 0.0_dp
            rkrk = SUM(rk%local_data(:, :)*rk%local_data(:, :))
            CALL mp_sum(rkrk, para_env%group)

            IF (alpha_beta) THEN
               rkrk_beta = 0.0_dp
               rkrk_beta = SUM(rk_b%local_data(:, :)*rk_b%local_data(:, :))
               CALL mp_sum(rkrk_beta, para_env%group)
               rkrk = rkrk+rkrk_beta
            ENDIF
            beta = rkrk/norms(2)

            pk%local_data(:, :) = rk%local_data(:, :)+beta*pk%local_data(:, :)
            IF (alpha_beta) THEN
               pk_b%local_data(:, :) = rk_b%local_data(:, :)+beta*pk_b%local_data(:, :)
            ENDIF

            conv = SQRT(rkrk)/norm_b

            t2 = m_walltime()

            IF (unit_nr > 0) THEN
               WRITE (unit_nr, '(T3,I5,T13,F6.1,11X,F14.8)') iiter, t2-t1, conv
            END IF

            IF (conv <= eps_conv) THEN
               converged = .TRUE.
               EXIT
            END IF

         END DO

         P_ia%local_data(:, :) = xk%local_data(:, :)
         IF (alpha_beta) THEN
            P_ia_beta%local_data(:, :) = xk_b%local_data(:, :)
         ENDIF

         CALL cp_fm_release(xk)
         CALL cp_fm_release(pk)
         CALL cp_fm_release(rk)
         CALL cp_fm_release(Ap)
         IF (alpha_beta) THEN
            CALL cp_fm_release(xk_b)
            CALL cp_fm_release(pk_b)
            CALL cp_fm_release(rk_b)
            CALL cp_fm_release(Ap_b)
         ENDIF

      ELSE
         ! Pople method
         ! change sign to L_jb
         L_jb%local_data(:, :) = -L_jb%local_data(:, :)

         ! allocate stuff
         ALLOCATE (xn(1:max_num_iter))
         ALLOCATE (Ax(1:max_num_iter))

         ! create fm structure
         NULLIFY (fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=homo, ncol_global=virtual)

         ! create preconditioner (for now only orbital energy differences)
         NULLIFY (precond)
         CALL cp_fm_create(precond, fm_struct_tmp, name="precond")
         CALL cp_fm_set_all(precond, 1.0_dp)
         CALL cp_fm_get_info(matrix=precond, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               precond%local_data(iiB, jjB) = precond%local_data(iiB, jjB)/ &
                                              (Eigenval(j_global+homo)-Eigenval(i_global))
            END DO
         END DO

         ! x(iiter) vector
         NULLIFY (b_i)
         CALL cp_fm_create(b_i, fm_struct_tmp, name="b_i")
         CALL cp_fm_set_all(b_i, 0.0_dp)
         b_i%local_data(:, :) = precond%local_data(:, :)*L_jb%local_data(:, :)

         ! create the residual vector (r), we check convergence on the norm of
         ! this vector r=(Ax-b)
         NULLIFY (residual)
         CALL cp_fm_create(residual, fm_struct_tmp, name="residual")
         CALL cp_fm_set_all(residual, 0.0_dp)

         ! allocate array containing the various scalar products
         ALLOCATE (x_norm(1:max_num_iter))
         ALLOCATE (xi_b(1:max_num_iter))
         ALLOCATE (xi_Axi(1:max_num_iter, 0:max_num_iter))
         x_norm = 0.0_dp
         xi_b = 0.0_dp
         xi_Axi = 0.0_dp

         ! The same and more for alpha-beta case
         IF (alpha_beta) THEN
            L_jb_beta%local_data(:, :) = -L_jb_beta%local_data(:, :)

            ! allocate stuff
            ALLOCATE (xn_b(1:max_num_iter))
            ALLOCATE (Ax_bb(1:max_num_iter))
            ALLOCATE (Ax_ab(1:max_num_iter))
            ALLOCATE (Ax_ba(1:max_num_iter))

            ! create fm structure
            NULLIFY (fm_struct_tmp_b)
            CALL cp_fm_struct_create(fm_struct_tmp_b, para_env=para_env, context=blacs_env, &
                                     nrow_global=homo_beta, ncol_global=virtual_beta)

            ! create preconditioner (for now only orbital energy differences)
            NULLIFY (precond_b)
            CALL cp_fm_create(precond_b, fm_struct_tmp_b, name="precond_b")
            CALL cp_fm_set_all(precond_b, 1.0_dp)
            CALL cp_fm_get_info(matrix=precond_b, &
                                nrow_local=nrow_local_b, &
                                ncol_local=ncol_local_b, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
            DO jjB = 1, ncol_local_b
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local_b
                  i_global = row_indices(iiB)
                  precond_b%local_data(iiB, jjB) = precond_b%local_data(iiB, jjB)/ &
                                                   (Eigenval_beta(j_global+homo_beta)-Eigenval_beta(i_global))
               END DO
            END DO

            ! x(iiter) vector
            NULLIFY (b_i_b)
            CALL cp_fm_create(b_i_b, fm_struct_tmp_b, name="b_i_b")
            CALL cp_fm_set_all(b_i_b, 0.0_dp)
            b_i_b%local_data(:, :) = precond_b%local_data(:, :)*L_jb_beta%local_data(:, :)

            ! create the residual vector (r), we check convergence on the norm of
            ! this vector r=(Ax-b)
            NULLIFY (residual_b)
            CALL cp_fm_create(residual_b, fm_struct_tmp_b, name="residual")
            CALL cp_fm_set_all(residual_b, 0.0_dp)

            ! allocate array containing the various scalar products
            ALLOCATE (x_norm_b(1:max_num_iter))
            x_norm_b = 0.0_dp

         ENDIF

         cycle_counter = 0
         DO iiter = 1, max_num_iter
            cycle_counter = cycle_counter+1

            t1 = m_walltime()

            ! create and update x_i (orthogonalization with previous vectors)
            NULLIFY (xn(iiter)%matrix)
            CALL cp_fm_create(xn(iiter)%matrix, fm_struct_tmp, name="xi")
            CALL cp_fm_set_all(xn(iiter)%matrix, 0.0_dp)

            IF (.NOT. alpha_beta) THEN
               ! first compute the projection of the actual b_i into all previous x_i
               ! already scaled with the norm of each x_i
               ALLOCATE (proj_bi_xj(iiter-1))
               DO iiB = 1, iiter-1
                  proj_bi_xj(iiB) = 0.0_dp
                  proj_bi_xj(iiB) = accurate_sum(b_i%local_data(1:nrow_local, 1:ncol_local)* &
                                                 xn(iiB)%matrix%local_data(1:nrow_local, 1:ncol_local))
                  proj_bi_xj(iiB) = proj_bi_xj(iiB)/x_norm(iiB)
               END DO
               CALL mp_sum(proj_bi_xj, para_env%group)

               ! update actual x_i
               xn(iiter)%matrix%local_data(:, :) = b_i%local_data(:, :)
               DO iiB = 1, iiter-1
                  xn(iiter)%matrix%local_data(:, :) = xn(iiter)%matrix%local_data(:, :)- &
                                                      xn(iiB)%matrix%local_data(:, :)*proj_bi_xj(iiB)
               END DO
               DEALLOCATE (proj_bi_xj)

            ELSE
               ! create and update x_i (orthogonalization with previous vectors)
               NULLIFY (xn_b(iiter)%matrix)
               CALL cp_fm_create(xn_b(iiter)%matrix, fm_struct_tmp_b, name="xi_b")
               CALL cp_fm_set_all(xn_b(iiter)%matrix, 0.0_dp)

               ! first compute the projection of the actual b_i into all previous x_i
               ! already scaled with the norm of each x_i
               ALLOCATE (proj_bi_xj(iiter-1))
               DO iiB = 1, iiter-1
                  proj_bi_xj(iiB) = 0.0_dp
                  proj_bi_xj(iiB) = &
                     accurate_sum(b_i%local_data(1:nrow_local, 1:ncol_local)* &
                                  xn(iiB)%matrix%local_data(1:nrow_local, 1:ncol_local))+ &
                     accurate_sum(b_i_b%local_data(1:nrow_local_b, 1:ncol_local_b)* &
                                  xn_b(iiB)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b))
                  proj_bi_xj(iiB) = proj_bi_xj(iiB)/(x_norm(iib)+x_norm_b(iib))
               END DO
               CALL mp_sum(proj_bi_xj, para_env%group)

               ! update actual x_i
               xn(iiter)%matrix%local_data(:, :) = b_i%local_data(:, :)
               xn_b(iiter)%matrix%local_data(:, :) = b_i_b%local_data(:, :)
               DO iiB = 1, iiter-1
                  xn(iiter)%matrix%local_data(:, :) = xn(iiter)%matrix%local_data(:, :)- &
                                                      xn(iiB)%matrix%local_data(:, :)*proj_bi_xj(iiB)
                  xn_b(iiter)%matrix%local_data(:, :) = xn_b(iiter)%matrix%local_data(:, :)- &
                                                        xn_b(iiB)%matrix%local_data(:, :)*proj_bi_xj(iiB)
               END DO
               DEALLOCATE (proj_bi_xj)
            ENDIF

            ! create Ax(iiter) that will store the matrix vector product for this cycle
            NULLIFY (Ax(iiter)%matrix)
            CALL cp_fm_create(Ax(iiter)%matrix, fm_struct_tmp, name="Ai")
            CALL cp_fm_set_all(Ax(iiter)%matrix, 0.0_dp)
            ! perform the matrix-vector product (CPHF like update)
            CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                                  mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                                  hfx_sections, energy, n_rep_hf, poisson_env, &
                                  rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                                  xn(iiter)%matrix, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                                  Ax(iiter)%matrix, transf_type_out, factor=factor)

            IF (alpha_beta) THEN
               NULLIFY (Ax_ab(iiter)%matrix)
               CALL cp_fm_create(Ax_ab(iiter)%matrix, fm_struct_tmp, name="Ai_ab")
               CALL cp_fm_set_all(Ax_ab(iiter)%matrix, 0.0_dp)
               ! Alpha-beta part of alpha.
               CALL cphf_like_update(qs_env, para_env, homo, virtual, dimen, &
                                     mo_coeff, mo_coeff_o, mo_coeff_v, Eigenval, &
                                     hfx_sections, energy, n_rep_hf, poisson_env, &
                                     rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                                     xn_b(iiter)%matrix, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                                     Ax_ab(iiter)%matrix, transf_type_out, factor=factor, &
                                     mo_coeff_beta=mo_coeff_beta, &
                                     mo_coeff_o_beta=mo_coeff_o_beta, &
                                     mo_coeff_v_beta=mo_coeff_v_beta, &
                                     homo_beta=homo_beta, virtual_beta=virtual_beta)
               ! Beta-beta part (Coulomb and XC) of beta.
               NULLIFY (Ax_bb(iiter)%matrix)
               CALL cp_fm_create(Ax_bb(iiter)%matrix, fm_struct_tmp_b, name="Ai_bb")
               CALL cp_fm_set_all(Ax_bb(iiter)%matrix, 0.0_dp)
               CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                                     mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                                     Eigenval_beta, hfx_sections, energy, n_rep_hf, &
                                     poisson_env, rho_work, pot_g, rho_g, rho_r, mat_mu_nu, &
                                     P_mu_nu, xn_b(iiter)%matrix, fm_G_mu_nu, fm_back, transf_type_in, &
                                     out_alpha, Ax_bb(iiter)%matrix, transf_type_out, factor=factor)
               ! Beta-alpha part of beta.
               NULLIFY (Ax_ba(iiter)%matrix)
               CALL cp_fm_create(Ax_ba(iiter)%matrix, fm_struct_tmp_b, name="Ai_ba")
               CALL cp_fm_set_all(Ax_ba(iiter)%matrix, 0.0_dp)
               CALL cphf_like_update(qs_env, para_env, homo_beta, virtual_beta, dimen, &
                                     mo_coeff_beta, mo_coeff_o_beta, mo_coeff_v_beta, &
                                     Eigenval_beta, &
                                     hfx_sections, energy, n_rep_hf, poisson_env, &
                                     rho_work, pot_g, rho_g, rho_r, mat_mu_nu, P_mu_nu, &
                                     xn(iiter)%matrix, fm_G_mu_nu, fm_back, transf_type_in, out_alpha, &
                                     Ax_ba(iiter)%matrix, transf_type_out, factor=factor, &
                                     mo_coeff_beta=mo_coeff, &
                                     mo_coeff_o_beta=mo_coeff_o, &
                                     mo_coeff_v_beta=mo_coeff_v, &
                                     homo_beta=homo, virtual_beta=virtual)
            ENDIF

            ! in order to reduce the number of calls to mp_sum here we
            ! cluster all necessary scalar products into a sigle vector
            ! temp_vals contains:
            ! 1:iiter -> <Ax_i|x_j>
            ! iiter+1 -> <x_i|b>
            ! iiter+2 -> <x_i|x_i>

            ALLOCATE (temp_vals(iiter+2))
            temp_vals = 0.0_dp
            ! <Ax_i|x_j>
            DO iiB = 1, iiter
               temp_vals(iiB) = accurate_sum(Ax(iiter)%matrix%local_data(1:nrow_local, 1:ncol_local)* &
                                             xn(iiB)%matrix%local_data(1:nrow_local, 1:ncol_local))
               IF (alpha_beta) THEN
                  temp_vals(iiB) = temp_vals(iib)+ &
                                   accurate_sum(Ax_ab(iiter)%matrix%local_data(1:nrow_local, 1:ncol_local)* &
                                                xn(iiB)%matrix%local_data(1:nrow_local, 1:ncol_local))+ &
                                   accurate_sum(Ax_bb(iiter)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b)* &
                                                xn_b(iiB)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b))+ &
                                   accurate_sum(Ax_ba(iiter)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b)* &
                                                xn_b(iiB)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b))
               ENDIF
            END DO
            ! <x_i|b>
            temp_vals(iiter+1) = accurate_sum(xn(iiter)%matrix%local_data(1:nrow_local, 1:ncol_local)* &
                                              L_jb%local_data(1:nrow_local, 1:ncol_local))
            ! norm
            temp_vals(iiter+2) = accurate_sum(xn(iiter)%matrix%local_data(1:nrow_local, 1:ncol_local)* &
                                              xn(iiter)%matrix%local_data(1:nrow_local, 1:ncol_local))
            IF (alpha_beta) THEN
               ! <x_i_b|b_b>
               temp_vals(iiter+1) = temp_vals(iiter+1)+ &
                                    accurate_sum(xn_b(iiter)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b)* &
                                                 L_jb_beta%local_data(1:nrow_local_b, 1:ncol_local_b))
               ! norm
               temp_vals(iiter+2) = temp_vals(iiter+2)+ &
                                    accurate_sum(xn_b(iiter)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b)* &
                                                 xn_b(iiter)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b))
            ENDIF
            CALL mp_sum(temp_vals, para_env%group)
            ! update <Ax_i|x_j>,  <x_i|b> and norm <x_i|x_i>
            xi_Axi(iiter, 1:iiter) = temp_vals(1:iiter)
            xi_Axi(1:iiter, iiter) = temp_vals(1:iiter)
            xi_b(iiter) = temp_vals(iiter+1)
            x_norm(iiter) = temp_vals(iiter+2)
            DEALLOCATE (temp_vals)

            ! solve reduced system
            IF (ALLOCATED(A_small)) DEALLOCATE (A_small)
            IF (ALLOCATED(b_small)) DEALLOCATE (b_small)
            ALLOCATE (A_small(iiter, iiter))
            ALLOCATE (b_small(iiter, 1))
            A_small(1:iiter, 1:iiter) = xi_Axi(1:iiter, 1:iiter)
            b_small(1:iiter, 1) = xi_b(1:iiter)

            CALL solve_system(matrix=A_small, mysize=iiter, eigenvectors=b_small)

            ! check for convergence
            CALL cp_fm_set_all(residual, 0.0_dp)
            IF (.NOT. alpha_beta) THEN
               DO iiB = 1, iiter
                  residual%local_data(1:nrow_local, 1:ncol_local) = &
                     residual%local_data(1:nrow_local, 1:ncol_local)+ &
                     b_small(iiB, 1)*Ax(iiB)%matrix%local_data(1:nrow_local, 1:ncol_local)
               END DO
               residual%local_data(1:nrow_local, 1:ncol_local) = residual%local_data(1:nrow_local, 1:ncol_local)- &
                                                                 L_jb%local_data(1:nrow_local, 1:ncol_local)
               conv = 0.0_dp
               conv = accurate_sum(residual%local_data(1:nrow_local, 1:ncol_local)* &
                                   residual%local_data(1:nrow_local, 1:ncol_local))
               CALL mp_sum(conv, para_env%group)
               conv = SQRT(conv)
            ELSE
               ! The same for beta
               CALL cp_fm_set_all(residual_b, 0.0_dp)
               DO iiB = 1, iiter
                  residual%local_data(1:nrow_local, 1:ncol_local) = &
                     residual%local_data(1:nrow_local, 1:ncol_local)+ &
                     b_small(iiB, 1)*Ax(iiB)%matrix%local_data(1:nrow_local, 1:ncol_local)+ &
                     b_small(iiB, 1)*Ax_ab(iiB)%matrix%local_data(1:nrow_local, 1:ncol_local)
                  residual_b%local_data(1:nrow_local_b, 1:ncol_local_b) = &
                     residual_b%local_data(1:nrow_local_b, 1:ncol_local_b)+ &
                     b_small(iiB, 1)*Ax_bb(iiB)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b)+ &
                     b_small(iiB, 1)*Ax_ba(iiB)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b)
               END DO
               residual%local_data(1:nrow_local, 1:ncol_local) = residual%local_data(1:nrow_local, 1:ncol_local)- &
                                                                 L_jb%local_data(1:nrow_local, 1:ncol_local)
               conv = 0.0_dp
               conv = accurate_sum(residual%local_data(1:nrow_local, 1:ncol_local)* &
                                   residual%local_data(1:nrow_local, 1:ncol_local))
               CALL mp_sum(conv, para_env%group)
               conv = SQRT(conv)
               residual_b%local_data(1:nrow_local_b, 1:ncol_local_b) = residual_b%local_data(1:nrow_local_b, 1:ncol_local_b)- &
                                                                       L_jb_beta%local_data(1:nrow_local_b, 1:ncol_local_b)
               conv_b = 0.0_dp
               conv_b = accurate_sum(residual_b%local_data(1:nrow_local_b, 1:ncol_local_b)* &
                                     residual_b%local_data(1:nrow_local_b, 1:ncol_local_b))
               CALL mp_sum(conv_b, para_env%group)
               conv = conv+SQRT(conv_b)
            ENDIF

            t2 = m_walltime()

            IF (unit_nr > 0) THEN
               WRITE (unit_nr, '(T3,I5,T13,F6.1,11X,F14.8)') iiter, t2-t1, conv
            END IF

            IF (conv <= eps_conv) THEN
               converged = .TRUE.
               EXIT
            END IF

            ! update b_i for the next round
            IF (.NOT. alpha_beta) THEN
               b_i%local_data(:, :) = precond%local_data(:, :)*Ax(iiter)%matrix%local_data(:, :)
            ELSE
               b_i%local_data(:, :) = precond%local_data(:, :)* &
                                      (Ax(iiter)%matrix%local_data(:, :) &
                                       +Ax_ab(iiter)%matrix%local_data(:, :))
               b_i_b%local_data(:, :) = precond_b%local_data(:, :)* &
                                        (Ax_bb(iiter)%matrix%local_data(:, :)+ &
                                         Ax_ba(iiter)%matrix%local_data(:, :))
            ENDIF

         END DO

         ! store solution into P_ia
         DO iiter = 1, cycle_counter
            P_ia%local_data(1:nrow_local, 1:ncol_local) = P_ia%local_data(1:nrow_local, 1:ncol_local)+ &
                                                          b_small(iiter, 1)*xn(iiter)%matrix%local_data(1:nrow_local, 1:ncol_local)
            ! The same for beta
            IF (alpha_beta) THEN
               P_ia_beta%local_data(1:nrow_local_b, 1:ncol_local_b) = &
                  P_ia_beta%local_data(1:nrow_local_b, 1:ncol_local_b)+ &
                  b_small(iiter, 1)*xn_b(iiter)%matrix%local_data(1:nrow_local_b, 1:ncol_local_b)
            ENDIF
         END DO

         DEALLOCATE (x_norm)
         DEALLOCATE (xi_b)
         DEALLOCATE (xi_Axi)

         CALL cp_fm_release(precond)
         CALL cp_fm_release(b_i)
         CALL cp_fm_release(residual)
         CALL cp_fm_struct_release(fm_struct_tmp)

         ! release Ax, xn
         DO iiter = 1, cycle_counter
            CALL cp_fm_release(Ax(iiter)%matrix)
            CALL cp_fm_release(xn(iiter)%matrix)
         END DO
         DEALLOCATE (xn)
         DEALLOCATE (Ax)

         ! The same for beta
         IF (alpha_beta) THEN
            DEALLOCATE (x_norm_b)

            CALL cp_fm_release(precond_b)
            CALL cp_fm_release(b_i_b)
            CALL cp_fm_release(residual_b)
            CALL cp_fm_struct_release(fm_struct_tmp_b)

            ! release Ax, xn and more
            DO iiter = 1, cycle_counter
               CALL cp_fm_release(Ax_bb(iiter)%matrix)
               CALL cp_fm_release(Ax_ab(iiter)%matrix)
               CALL cp_fm_release(Ax_ba(iiter)%matrix)
               CALL cp_fm_release(xn_b(iiter)%matrix)
            END DO
            DEALLOCATE (xn_b)
            DEALLOCATE (Ax_bb)
            DEALLOCATE (Ax_ab)
            DEALLOCATE (Ax_ba)
         ENDIF

      END IF

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T4,A)') REPEAT("-", 40)
         IF (converged) THEN
            WRITE (unit_nr, '(T3,A,I5,A)') 'Z-Vector equations converged in', cycle_counter, ' steps'
         ELSE
            WRITE (unit_nr, '(T3,A,I5,A)') 'Z-Vector equations NOT converged in', cycle_counter, ' steps'
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE solve_z_vector_eq_low

! **************************************************************************************************
!> \brief ...
!> \param mat ...
!> \param unitout ...
! **************************************************************************************************
   SUBROUTINE write_array(mat, unitout)
      REAL(KIND=dp), DIMENSION(:, :)                     :: mat
      INTEGER, OPTIONAL                                  :: unitout

      INTEGER                                            :: iii, jjj

      WRITE (*, *)
      DO iii = 1, SIZE(mat, 1)
         WRITE (*, *) iii
         DO jjj = 1, SIZE(mat, 2), 20
            IF (PRESENT(unitout)) THEN
               WRITE (1000+unitout, '(1000F10.5)') mat(iii, jjj:MIN(SIZE(mat, 2), jjj+19))
            ELSE
               WRITE (*, '(1000F10.5)') mat(iii, jjj:MIN(SIZE(mat, 2), jjj+19))
            END IF
         END DO
         WRITE (*, *)
      END DO
      WRITE (*, *)
   END SUBROUTINE

END MODULE mp2_cphf
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate gradients of RI-GPW-MP2 energy using pw
!> \par History
!>      10.2013 created [Mauro Del Ben]
! **************************************************************************************************
MODULE mp2_ri_grad
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_copy, cp_dbcsr_copy_into_existing, cp_dbcsr_create, cp_dbcsr_init, &
        cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_set, cp_dbcsr_transposed, &
        cp_dbcsr_type, dbcsr_type_no_symmetry, dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_indxg2l,&
                                              cp_fm_indxg2p,&
                                              cp_fm_indxl2g,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE gaussian_gridlevels,             ONLY: gaussian_gridlevel
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi
   USE message_passing,                 ONLY: &
        mp_alltoall, mp_comm_split_direct, mp_irecv, mp_isend, mp_request_null, mp_sendrecv, &
        mp_sum, mp_wait, mp_waitall
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE mp2_types,                       ONLY: integ_mat_buffer_type,&
                                              integ_mat_buffer_type_2D,&
                                              mp2_type
   USE orbital_pointers,                ONLY: ncoset
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_copy,&
                                              pw_derive,&
                                              pw_integral_ab,&
                                              pw_scale,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: calculate_rho_elec,&
                                              calculate_wavefunction
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_integrate_potential,          ONLY: integrate_pgf_product_rspace,&
                                              integrate_v_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              realspace_grid_p_type,&
                                              rs_grid_release,&
                                              rs_grid_retain
   USE rs_pw_interface,                 ONLY: potential_pw2rs
   USE task_list_types,                 ONLY: task_list_type
   USE util,                            ONLY: get_limit
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_ri_grad'

   PUBLIC :: calc_ri_mp2_nonsep

CONTAINS

! **************************************************************************************************
!> \brief Calcualte the non-separable part of the gradients and update the
!>        Lagrangian
!> \param qs_env ...
!> \param mp2_env ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param dft_control ...
!> \param cell ...
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param mo_coeff ...
!> \param nmo ...
!> \param homo ...
!> \param dimen_RI ...
!> \param Eigenval ...
!> \param my_group_L_start ...
!> \param my_group_L_end ...
!> \param my_group_L_size ...
!> \param rho_r ...
!> \param rho_g ...
!> \param pot_g ...
!> \param mat_munu ...
!> \param pw_env_sub ...
!> \param poisson_env ...
!> \param auxbas_pw_pool ...
!> \param task_list_sub ...
!> \param blacs_env_sub ...
!> \param Eigenval_beta ...
!> \param homo_beta ...
!> \param mo_coeff_beta ...
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE calc_ri_mp2_nonsep(qs_env, mp2_env, para_env, para_env_sub, dft_control, cell, particle_set, &
                                 atomic_kind_set, qs_kind_set, mo_coeff, nmo, homo, dimen_RI, Eigenval, &
                                 my_group_L_start, my_group_L_end, my_group_L_size, rho_r, rho_g, pot_g, &
                                 mat_munu, pw_env_sub, poisson_env, auxbas_pw_pool, task_list_sub, &
                                 blacs_env_sub, Eigenval_beta, homo_beta, mo_coeff_beta)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      INTEGER                                            :: nmo, homo, dimen_RI
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: my_group_L_start, my_group_L_end, &
                                                            my_group_L_size
      TYPE(pw_p_type)                                    :: rho_r, rho_g, pot_g
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(pw_env_type), POINTER                         :: pw_env_sub
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(task_list_type), POINTER                      :: task_list_sub
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env_sub
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      INTEGER, OPTIONAL                                  :: homo_beta
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: mo_coeff_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_ri_mp2_nonsep', &
         routineP = moduleN//':'//routineN

      INTEGER :: alpha, atom_a, beta, dimen, dir, handle, handle2, handle3, i, iatom, igrid_level, &
         ikind, iorb, ipgf, iset, itmp(2), L_counter, lb(3), LLL, location(3), my_P_end, &
         my_P_size, my_P_start, na1, na2, natom, ncoa, nkind, nseta, offset, sgfa, tp(3), ub(3), &
         virtual, virtual_beta
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of
      INTEGER, DIMENSION(3)                              :: comp
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, npgfa, nsgfa
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa
      LOGICAL                                            :: alpha_beta, map_it_here, skip_shell, &
                                                            use_virial
      REAL(KIND=dp)                                      :: e_hartree, eps_filter, factor, &
                                                            pair_energy, rab2, total_rho
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: wf_vector
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: G_PQ_local, I_ab
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, ra, rab
      REAL(KIND=dp), DIMENSION(3, 3)                     :: h_stress, my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: I_tmp2, pab, rpgfa, sphi_a, zeta
      TYPE(cp_dbcsr_p_type) :: Lag_mu_i_1, Lag_mu_i_1_beta, Lag_nu_a_2, Lag_nu_a_2_beta, &
         matrix_P_inu, matrix_P_inu_beta, matrix_P_munu, matrix_P_munu_nosym
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: G_P_ia, G_P_ia_beta
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_o, mo_coeff_o_beta, mo_coeff_v, &
                                                            mo_coeff_v_beta
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: L1_mu_i, L1_mu_i_beta, L2_nu_a, &
                                                            L2_nu_a_beta
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(pw_p_type)                                    :: dvg(3), psi_L, psi_L_beta, temp_pw_g
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_v
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      ! Find out whether we have a closed or open shell
      alpha_beta = .FALSE.
      IF (PRESENT(homo_beta) .AND. PRESENT(Eigenval_beta)) alpha_beta = .TRUE.

      dimen = nmo
      virtual = dimen-homo
      IF (alpha_beta) virtual_beta = dimen-homo_beta
      eps_filter = mp2_env%mp2_gpw%eps_filter
      NULLIFY (mo_coeff_o, mo_coeff_v, G_P_ia, ks_env)
      mo_coeff_o => mp2_env%ri_grad%mo_coeff_o
      mo_coeff_v => mp2_env%ri_grad%mo_coeff_v
      G_P_ia => mp2_env%ri_grad%G_P_ia

      IF (alpha_beta) THEN
         NULLIFY (mo_coeff_o_beta, mo_coeff_v_beta, G_P_ia_beta)
         mo_coeff_o_beta => mp2_env%ri_grad%mo_coeff_o_beta
         mo_coeff_v_beta => mp2_env%ri_grad%mo_coeff_v_beta
         G_P_ia_beta => mp2_env%ri_grad%G_P_ia_beta
      ENDIF

      CALL get_qs_env(qs_env, ks_env=ks_env, molecule_set=molecule_set, &
                      molecule_kind_set=molecule_kind_set)

      nkind = SIZE(atomic_kind_set)
      natom = SIZE(particle_set)

      ALLOCATE (kind_of(natom))
      ALLOCATE (atom_of_kind(natom))
      CALL get_atomic_kind_set(atomic_kind_set, kind_of=kind_of, atom_of_kind=atom_of_kind)

      itmp = get_limit(dimen_RI, para_env_sub%num_pe, para_env_sub%mepos)
      my_P_start = itmp(1)
      my_P_end = itmp(2)
      my_P_size = itmp(2)-itmp(1)+1
      ALLOCATE (G_PQ_local(dimen_RI, my_group_L_size))

      G_PQ_local = 0.0_dp
      IF (.NOT. alpha_beta) THEN
         G_PQ_local(my_P_start:my_P_end, 1:my_group_L_size) = mp2_env%ri_grad%Gamma_PQ
      ELSE
         G_PQ_local(my_P_start:my_P_end, 1:my_group_L_size) = &
            0.50_dp*(mp2_env%ri_grad%Gamma_PQ+mp2_env%ri_grad%Gamma_PQ_beta)
      ENDIF
      DEALLOCATE (mp2_env%ri_grad%Gamma_PQ)
      IF (alpha_beta) THEN
         DEALLOCATE (mp2_env%ri_grad%Gamma_PQ_beta)
      ENDIF
      CALL mp_sum(G_PQ_local, para_env_sub%group)

      ! deallocate here PQ_half, maybe usefull in the future
      DEALLOCATE (mp2_env%ri_grad%PQ_half)

      ! create matrix holding the back transformation (G_P_inu)
      ALLOCATE (matrix_P_inu%matrix)
      CALL cp_dbcsr_init(matrix_P_inu%matrix)
      CALL cp_dbcsr_create(matrix_P_inu%matrix, template=mo_coeff_o)
      IF (alpha_beta) THEN
         ALLOCATE (matrix_P_inu_beta%matrix)
         CALL cp_dbcsr_init(matrix_P_inu_beta%matrix)
         CALL cp_dbcsr_create(matrix_P_inu_beta%matrix, template=mo_coeff_o_beta)
      ENDIF

      ! non symmetric matrix
      ALLOCATE (matrix_P_munu_nosym%matrix)
      CALL cp_dbcsr_init(matrix_P_munu_nosym%matrix)
      CALL cp_dbcsr_create(matrix_P_munu_nosym%matrix, template=mat_munu%matrix, &
                           matrix_type=dbcsr_type_no_symmetry)
      ! CALL cp_dbcsr_create(matrix_P_munu_nosym%matrix,template=mat_munu%matrix)
      ! CALL cp_dbcsr_copy(matrix_P_munu_nosym%matrix,mat_munu%matrix,&
      !                    name="matrix_P_munu_nosym")
      !XXXXXX
      ! CALL cp_dbcsr_desymmetrize(mat_munu%matrix,matrix_P_munu_nosym%matrix)

      ! create Lagrangian matrices in mixed AO/MO formalism
      ALLOCATE (Lag_mu_i_1%matrix)
      CALL cp_dbcsr_init(Lag_mu_i_1%matrix)
      CALL cp_dbcsr_create(Lag_mu_i_1%matrix, template=mo_coeff_o)
      CALL cp_dbcsr_set(Lag_mu_i_1%matrix, 0.0_dp)
      IF (alpha_beta) THEN
         ALLOCATE (Lag_mu_i_1_beta%matrix)
         CALL cp_dbcsr_init(Lag_mu_i_1_beta%matrix)
         CALL cp_dbcsr_create(Lag_mu_i_1_beta%matrix, template=mo_coeff_o_beta)
         CALL cp_dbcsr_set(Lag_mu_i_1_beta%matrix, 0.0_dp)
      ENDIF

      ALLOCATE (Lag_nu_a_2%matrix)
      CALL cp_dbcsr_init(Lag_nu_a_2%matrix)
      CALL cp_dbcsr_create(Lag_nu_a_2%matrix, template=mo_coeff_v)
      CALL cp_dbcsr_set(Lag_nu_a_2%matrix, 0.0_dp)
      IF (alpha_beta) THEN
         ALLOCATE (Lag_nu_a_2_beta%matrix)
         CALL cp_dbcsr_init(Lag_nu_a_2_beta%matrix)
         CALL cp_dbcsr_create(Lag_nu_a_2_beta%matrix, template=mo_coeff_v_beta)
         CALL cp_dbcsr_set(Lag_nu_a_2_beta%matrix, 0.0_dp)
      ENDIF

      ALLOCATE (matrix_P_munu%matrix)
      CALL cp_dbcsr_init(matrix_P_munu%matrix)
      CALL cp_dbcsr_create(matrix_P_munu%matrix, template=mat_munu%matrix, &
                           matrix_type=dbcsr_type_symmetric)
      ! CALL cp_dbcsr_set(mat_munu%matrix,0.0_dp)
      ! CALL cp_dbcsr_copy(matrix_P_munu%matrix,mat_munu%matrix,&
      !                    name="matrix_P_munu")

      ! wave function vector and supporting stuff
      ALLOCATE (wf_vector(dimen_RI))
      NULLIFY (psi_L%pw)
      CALL pw_pool_create_pw(auxbas_pw_pool, psi_L%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      IF (alpha_beta) THEN
         NULLIFY (psi_L_beta%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, psi_L_beta%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
      ENDIF

      ! get forces
      NULLIFY (force, virial)
      CALL get_qs_env(qs_env=qs_env, force=force, virial=virial)

      ! check if we want to calculate the virial
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      ! in case virial is required we need auxilliary pw
      ! for calculate the MP2-volume contribution to the virial
      ! (hartree potential derivatives)
      IF (use_virial) THEN
         NULLIFY (temp_pw_g%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, temp_pw_g%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
         DO i = 1, 3
            NULLIFY (dvg(i)%pw)
            CALL pw_pool_create_pw(auxbas_pw_pool, dvg(i)%pw, &
                                   use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
         END DO
      END IF

      ! start main loop over auxiliary basis functions
      CALL timeset(routineN//"_loop", handle2)
      L_counter = 0
      DO LLL = my_group_L_start, my_group_L_end
         L_counter = L_counter+1

         ! first back-transformation a->nu
         CALL timeset(routineN//"_back_v", handle3)
         CALL cp_dbcsr_set(matrix_P_inu%matrix, 0.0_dp)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, mo_coeff_v, G_P_ia(L_counter)%matrix, &
                                0.0_dp, matrix_P_inu%matrix, filter_eps=eps_filter)
         IF (alpha_beta) THEN
            CALL cp_dbcsr_set(matrix_P_inu_beta%matrix, 0.0_dp)
            CALL cp_dbcsr_multiply("N", "T", 1.0_dp, mo_coeff_v_beta, G_P_ia_beta(L_counter)%matrix, &
                                   0.0_dp, matrix_P_inu_beta%matrix, filter_eps=eps_filter)
         ENDIF
         CALL timestop(handle3)

         ! second back-transformation i->mu
         CALL timeset(routineN//"_back_o", handle3)
         CALL cp_dbcsr_set(matrix_P_munu_nosym%matrix, 0.0_dp)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_P_inu%matrix, mo_coeff_o, &
                                0.0_dp, matrix_P_munu_nosym%matrix, filter_eps=eps_filter)
         IF (alpha_beta) THEN
            !CALL cp_dbcsr_set(matrix_P_munu_nosym%matrix,0.0_dp)
            CALL cp_dbcsr_multiply("N", "T", 0.5_dp, matrix_P_inu_beta%matrix, mo_coeff_o_beta, &
                                   0.5_dp, matrix_P_munu_nosym%matrix, filter_eps=eps_filter)
         ENDIF

         ! ! first back-transformation a->nu
         ! CALL timeset(routineN//"_back_v",handle3)
         ! CALL cp_dbcsr_set(matrix_P_inu%matrix,0.0_dp)
         ! CALL cp_dbcsr_multiply("N", "T", 1.0_dp, G_P_ia(L_counter)%matrix, mo_coeff_v, &
         !                         0.0_dp, matrix_P_inu%matrix, filter_eps=eps_filter)
         ! CALL timestop(handle3)

         ! ! second back-transformation i->mu
         ! CALL timeset(routineN//"_back_o",handle3)
         ! CALL cp_dbcsr_set(matrix_P_munu_nosym%matrix,0.0_dp)
         ! CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mo_coeff_o, matrix_P_inu%matrix, &
         !                         0.0_dp, matrix_P_munu_nosym%matrix, filter_eps=eps_filter)
         ! symmetrize
         CALL cp_dbcsr_set(matrix_P_munu%matrix, 0.0_dp)
         CALL cp_dbcsr_transposed(matrix_P_munu%matrix, matrix_P_munu_nosym%matrix)
         CALL cp_dbcsr_add(matrix_P_munu%matrix, matrix_P_munu_nosym%matrix, &
                           alpha_scalar=2.0_dp, beta_scalar=2.0_dp)
         ! this is a trick to avoid that integrate_v_rspace starts to cry
         CALL cp_dbcsr_copy_into_existing(mat_munu%matrix, matrix_P_munu%matrix)
         CALL cp_dbcsr_copy(matrix_P_munu%matrix, mat_munu%matrix)

         CALL timestop(handle3)

         ! calculate potential associted to the single aux function
         CALL timeset(routineN//"_wf_pot", handle3)
         wf_vector = 0.0_dp
         wf_vector(LLL) = 1.0_dp
         ! pseudo psi_L
         CALL calculate_wavefunction(mo_coeff, 1, psi_L, rho_g, atomic_kind_set, &
                                     qs_kind_set, cell, dft_control, particle_set, pw_env_sub, &
                                     basis_type="RI_AUX", &
                                     external_vector=wf_vector)

         rho_r%pw%cr3d = psi_L%pw%cr3d
         IF (alpha_beta) THEN
            CALL calculate_wavefunction(mo_coeff_beta, 1, psi_L_beta, rho_g, atomic_kind_set, &
                                        qs_kind_set, cell, dft_control, particle_set, &
                                        pw_env_sub, basis_type='RI_AUX', &
                                        external_vector=wf_vector)
            rho_r%pw%cr3d = 0.50_dp*(rho_r%pw%cr3d+psi_L_beta%pw%cr3d)
         ENDIF
         CALL pw_transfer(rho_r%pw, rho_g%pw)
         CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
         CALL pw_transfer(pot_g%pw, rho_r%pw)
         CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)
         CALL timestop(handle3)

         IF (use_virial) THEN
            ! make a copy of the density in G space
            ! calculate the potential derivatives in G space
            CALL timeset(routineN//"_Virial", handle3)
            CALL pw_copy(rho_g%pw, temp_pw_g%pw)
            DO i = 1, 3
               comp = 0
               comp(i) = 1
               CALL pw_copy(pot_g%pw, dvg(i)%pw)
               CALL pw_derive(dvg(i)%pw, comp)
            END DO
            CALL timestop(handle3)
         END IF

         ! integrate the potential of the single gaussian and update
         ! 2-center forces with Gamma_PQ
         CALL timeset(routineN//"_int_PQ", handle3)
         NULLIFY (rs_v)
         NULLIFY (rs_descs)
         CALL pw_env_get(pw_env_sub, rs_descs=rs_descs, rs_grids=rs_v)
         DO i = 1, SIZE(rs_v)
            CALL rs_grid_retain(rs_v(i)%rs_grid)
         END DO
         CALL potential_pw2rs(rs_v, rho_r, pw_env_sub)

         offset = 0
         DO iatom = 1, natom
            ikind = kind_of(iatom)
            atom_a = atom_of_kind(iatom)
            CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=basis_set_a, &
                             basis_type="RI_AUX")

            first_sgfa => basis_set_a%first_sgf
            la_max => basis_set_a%lmax
            la_min => basis_set_a%lmin
            npgfa => basis_set_a%npgf
            nseta = basis_set_a%nset
            nsgfa => basis_set_a%nsgf_set
            rpgfa => basis_set_a%pgf_radius
            set_radius_a => basis_set_a%set_radius
            sphi_a => basis_set_a%sphi
            zeta => basis_set_a%zet

            ra(:) = pbc(particle_set(iatom)%r, cell)
            rab = 0.0_dp
            rab2 = 0.0_dp

            force_a(:) = 0.0_dp
            force_b(:) = 0.0_dp
            IF (use_virial) THEN
               my_virial_a = 0.0_dp
               my_virial_b = 0.0_dp
            END IF

            DO iset = 1, nseta
               ncoa = npgfa(iset)*ncoset(la_max(iset))
               sgfa = first_sgfa(1, iset)

               ALLOCATE (I_tmp2(ncoa, 1))
               I_tmp2 = 0.0_dp
               ALLOCATE (I_ab(nsgfa(iset), 1))
               I_ab = 0.0_dp
               ALLOCATE (pab(ncoa, 1))
               pab = 0.0_dp

               I_ab(1:nsgfa(iset), 1) = -4.0_dp*G_PQ_local(offset+1:offset+nsgfa(iset), L_counter)

               CALL dgemm("N", "N", ncoa, 1, nsgfa(iset), &
                          1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                          I_ab(1, 1), SIZE(I_ab, 1), &
                          0.0_dp, pab(1, 1), SIZE(pab, 1))

               I_ab = 0.0_dp

               igrid_level = gaussian_gridlevel(pw_env_sub%gridlevel_info, MINVAL(zeta(:, iset)))
               map_it_here = .FALSE.
               IF (.NOT. ALL(rs_v(igrid_level)%rs_grid%desc%perd == 1)) THEN
                  DO dir = 1, 3
                     ! bounds of local grid (i.e. removing the 'wings'), if periodic
                     tp(dir) = FLOOR(DOT_PRODUCT(cell%h_inv(dir, :), ra)*rs_v(igrid_level)%rs_grid%desc%npts(dir))
                     tp(dir) = MODULO(tp(dir), rs_v(igrid_level)%rs_grid%desc%npts(dir))
                     IF (rs_v(igrid_level)%rs_grid%desc%perd(dir) .NE. 1) THEN
                        lb(dir) = rs_v(igrid_level)%rs_grid%lb_local(dir)+rs_v(igrid_level)%rs_grid%desc%border
                        ub(dir) = rs_v(igrid_level)%rs_grid%ub_local(dir)-rs_v(igrid_level)%rs_grid%desc%border
                     ELSE
                        lb(dir) = rs_v(igrid_level)%rs_grid%lb_local(dir)
                        ub(dir) = rs_v(igrid_level)%rs_grid%ub_local(dir)
                     ENDIF
                     ! distributed grid, only map if it is local to the grid
                     location(dir) = tp(dir)+rs_v(igrid_level)%rs_grid%desc%lb(dir)
                  ENDDO
                  IF (lb(1) <= location(1) .AND. location(1) <= ub(1) .AND. &
                      lb(2) <= location(2) .AND. location(2) <= ub(2) .AND. &
                      lb(3) <= location(3) .AND. location(3) <= ub(3)) THEN
                     map_it_here = .TRUE.
                  ENDIF
               ELSE
                  ! not distributed, just a round-robin distribution over the full set of CPUs
                  IF (MODULO(offset, para_env_sub%num_pe) == para_env_sub%mepos) map_it_here = .TRUE.
               ENDIF

               offset = offset+nsgfa(iset)

               IF (map_it_here) THEN
                  DO ipgf = 1, npgfa(iset)
                     na1 = (ipgf-1)*ncoset(la_max(iset))+1
                     na2 = ipgf*ncoset(la_max(iset))
                     igrid_level = gaussian_gridlevel(pw_env_sub%gridlevel_info, zeta(ipgf, iset))

                     CALL integrate_pgf_product_rspace(la_max=la_max(iset), zeta=zeta(ipgf, iset)/2.0_dp, la_min=la_min(iset), &
                                                       lb_max=0, zetb=zeta(ipgf, iset)/2.0_dp, lb_min=0, &
                                                       ra=ra, rab=rab, rab2=rab2, &
                                                       rsgrid=rs_v(igrid_level)%rs_grid, &
                                                       cell=cell, &
                                                       cube_info=pw_env_sub%cube_info(igrid_level), &
                                                       hab=I_tmp2, &
                                                       pab=pab, &
                                                       o1=na1-1, &
                                                       o2=0, &
                                                       map_consistent=.TRUE., &
                                                       eps_gvg_rspace=dft_control%qs_control%eps_gvg_rspace, &
                                                       calculate_forces=.TRUE., &
                                                       force_a=force_a, force_b=force_b, &
                                                       use_virial=use_virial, my_virial_a=my_virial_a, my_virial_b=my_virial_b)

                  END DO

                  ! CALL dgemm("T","N",nsgfa(iset),1,ncoa,&
                  !             1.0_dp,sphi_a(1,sgfa),SIZE(sphi_a,1),&
                  !             I_tmp2(1,1),SIZE(I_tmp2,1),&
                  !             1.0_dp,I_ab(1,1),SIZE(I_ab,1))
                  ! L_local_col(offset-nsgfa(iset)+1:offset,i_counter)=I_ab(1:nsgfa(iset),1)
               END IF

               DEALLOCATE (I_tmp2)
               DEALLOCATE (I_ab)
               DEALLOCATE (pab)

            END DO

            force(ikind)%rho_elec(:, atom_a) = &
               force(ikind)%rho_elec(:, atom_a)+force_a(:)+force_b
            IF (use_virial) THEN
               virial%pv_virial = virial%pv_virial+my_virial_a+my_virial_b
            END IF
         END DO

         DO i = 1, SIZE(rs_v)
            CALL rs_grid_release(rs_v(i)%rs_grid)
         END DO
         CALL timestop(handle3)
         ! here we are done with the 2 centers

         ! CALL cp_dbcsr_write_sparse_matrix(matrix_P_munu%matrix,4,12,qs_env,para_env_sub,&
         !                                   output_unit=unit_nr)

         ! integrate the potential of the single gaussian and update
         ! 3-center forces
         CALL timeset(routineN//"_int", handle3)
         CALL cp_dbcsr_set(mat_munu%matrix, 0.0_dp)
         CALL integrate_v_rspace(rho_r, hmat=mat_munu, pmat=matrix_P_munu, &
                                 qs_env=qs_env, calculate_forces=.TRUE., compute_tau=.FALSE., gapw=.FALSE., &
                                 pw_env_external=pw_env_sub, &
                                 task_list_external=task_list_sub)
         CALL timestop(handle3)

         ! update lagrangian
         CALL timeset(routineN//"_Lag", handle3)
         ! first contract mat_munu with the half back transformed Gamma_i_nu
         ! in order to update Lag_mu_i_1
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mat_munu%matrix, matrix_P_inu%matrix, &
                                1.0_dp, Lag_mu_i_1%matrix, filter_eps=eps_filter)

         IF (alpha_beta) THEN
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mat_munu%matrix, matrix_P_inu_beta%matrix, &
                                   1.0_dp, Lag_mu_i_1_beta%matrix, filter_eps=eps_filter)
         ENDIF

         ! transform first index of mat_munu and store the result into matrix_P_inu
         CALL cp_dbcsr_set(matrix_P_inu%matrix, 0.0_dp)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mat_munu%matrix, mo_coeff_o, &
                                0.0_dp, matrix_P_inu%matrix, filter_eps=eps_filter)

         IF (alpha_beta) THEN
            CALL cp_dbcsr_set(matrix_P_inu_beta%matrix, 0.0_dp)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mat_munu%matrix, mo_coeff_o_beta, &
                                   0.0_dp, matrix_P_inu_beta%matrix, filter_eps=eps_filter)
         ENDIF

         ! contract the transformend matrix_P_inu with the untransformend Gamma_i_a
         ! in order to update Lag_nu_a_2
         CALL cp_dbcsr_multiply("N", "N", -1.0_dp, matrix_P_inu%matrix, G_P_ia(L_counter)%matrix, &
                                1.0_dp, Lag_nu_a_2%matrix, filter_eps=eps_filter)

         IF (alpha_beta) THEN
            CALL cp_dbcsr_multiply("N", "N", -1.0_dp, matrix_P_inu_beta%matrix, &
                                   G_P_ia_beta(L_counter)%matrix, 1.0_dp, Lag_nu_a_2_beta%matrix, &
                                   filter_eps=eps_filter)
         ENDIF

         ! release the actual gamma_P_ia
         CALL cp_dbcsr_release(G_P_ia(L_counter)%matrix)
         DEALLOCATE (G_P_ia(L_counter)%matrix)
         IF (alpha_beta) THEN
            CALL cp_dbcsr_release(G_P_ia_beta(L_counter)%matrix)
            DEALLOCATE (G_P_ia_beta(L_counter)%matrix)
         ENDIF
         CALL timestop(handle3)

         IF (use_virial) THEN
            ! add the volume contribution to the virial due to
            ! the (P|Q) integrals, first we put the full gamme_PQ
            ! pseudo wave-function into grid in order to calculate the
            ! hartree potential derivatives
            CALL timeset(routineN//"_Virial", handle3)
            wf_vector = 0.0_dp
            wf_vector(:) = -2.0_dp*G_PQ_local(:, L_counter)
            CALL calculate_wavefunction(mo_coeff, 1, psi_L, rho_g, atomic_kind_set, &
                                        qs_kind_set, cell, dft_control, particle_set, pw_env_sub, &
                                        basis_type="RI_AUX", &
                                        external_vector=wf_vector)
            ! transfer to reciprocal space and calculate potential
            rho_r%pw%cr3d = psi_L%pw%cr3d
            CALL pw_transfer(rho_r%pw, rho_g%pw)
            CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
            ! update virial with volume term (first calculate hartree like energy (diagonal part of the virial))
            e_hartree = 0.0_dp
            h_stress = 0.0_dp
            e_hartree = pw_integral_ab(temp_pw_g%pw, pot_g%pw)
            DO alpha = 1, 3
               comp = 0
               comp(alpha) = 1
               CALL pw_copy(pot_g%pw, rho_g%pw)
               CALL pw_derive(rho_g%pw, comp)
               h_stress(alpha, alpha) = -e_hartree
               DO beta = alpha, 3
                  h_stress(alpha, beta) = h_stress(alpha, beta) &
                                          -2.0_dp*pw_integral_ab(rho_g%pw, dvg(beta)%pw)/fourpi
                  h_stress(beta, alpha) = h_stress(alpha, beta)
               END DO
            END DO
            virial%pv_virial = virial%pv_virial+h_stress/REAL(para_env_sub%num_pe, dp)
            CALL timestop(handle3)
         END IF

         ! put the gamma density on grid
         CALL timeset(routineN//"_Gpot", handle3)
         CALL calculate_rho_elec(matrix_p=matrix_P_munu%matrix, &
                                 rho=rho_r, &
                                 rho_gspace=rho_g, &
                                 total_rho=total_rho, &
                                 task_list_external=task_list_sub, &
                                 pw_env_external=pw_env_sub, &
                                 ks_env=ks_env)
         ! calculate associated hartree potential
         ! CALL pw_transfer(rho_r%pw, rho_g%pw)
         !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
         CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
         CALL pw_transfer(pot_g%pw, rho_r%pw)
         CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)
         CALL timestop(handle3)

         IF (use_virial) THEN
            ! add the volume contribution to the virial coming from
            ! the 3-center integrals (mu nu|P)
            CALL timeset(routineN//"_Virial", handle3)
            e_hartree = 0.0_dp
            h_stress = 0.0_dp
            e_hartree = pw_integral_ab(temp_pw_g%pw, pot_g%pw)
            DO alpha = 1, 3
               comp = 0
               comp(alpha) = 1
               CALL pw_copy(pot_g%pw, rho_g%pw)
               CALL pw_derive(rho_g%pw, comp)
               h_stress(alpha, alpha) = -e_hartree
               DO beta = alpha, 3
                  h_stress(alpha, beta) = h_stress(alpha, beta) &
                                          -2.0_dp*pw_integral_ab(rho_g%pw, dvg(beta)%pw)/fourpi
                  h_stress(beta, alpha) = h_stress(alpha, beta)
               END DO
            END DO
            virial%pv_virial = virial%pv_virial+h_stress/REAL(para_env_sub%num_pe, dp)
            CALL timestop(handle3)
         END IF

         ! integrate potential with auxiliary basis function derivatives
         NULLIFY (rs_v)
         NULLIFY (rs_descs)
         CALL pw_env_get(pw_env_sub, rs_descs=rs_descs, rs_grids=rs_v)
         DO i = 1, SIZE(rs_v)
            CALL rs_grid_retain(rs_v(i)%rs_grid)
         END DO
         CALL potential_pw2rs(rs_v, rho_r, pw_env_sub)

         offset = 0
         DO iatom = 1, natom
            ikind = kind_of(iatom)
            atom_a = atom_of_kind(iatom)
            CALL get_qs_kind(qs_kind=qs_kind_set(ikind), basis_set=basis_set_a, &
                             basis_type="RI_AUX")

            first_sgfa => basis_set_a%first_sgf
            la_max => basis_set_a%lmax
            la_min => basis_set_a%lmin
            npgfa => basis_set_a%npgf
            nseta = basis_set_a%nset
            nsgfa => basis_set_a%nsgf_set
            rpgfa => basis_set_a%pgf_radius
            set_radius_a => basis_set_a%set_radius
            sphi_a => basis_set_a%sphi
            zeta => basis_set_a%zet

            ra(:) = pbc(particle_set(iatom)%r, cell)
            rab = 0.0_dp
            rab2 = 0.0_dp

            force_a(:) = 0.0_dp
            force_b(:) = 0.0_dp
            IF (use_virial) THEN
               my_virial_a = 0.0_dp
               my_virial_b = 0.0_dp
            END IF

            DO iset = 1, nseta
               ncoa = npgfa(iset)*ncoset(la_max(iset))
               sgfa = first_sgfa(1, iset)

               ALLOCATE (I_tmp2(ncoa, 1))
               I_tmp2 = 0.0_dp
               ALLOCATE (I_ab(nsgfa(iset), 1))
               I_ab = 0.0_dp
               ALLOCATE (pab(ncoa, 1))
               pab = 0.0_dp

               skip_shell = .TRUE.
               DO iorb = 1, nsgfa(iset)
                  IF (iorb+offset == LLL) THEN
                     I_ab(iorb, 1) = 1.0_dp
                     skip_shell = .FALSE.
                  END IF
               END DO

               IF (skip_shell) THEN
                  offset = offset+nsgfa(iset)
                  DEALLOCATE (I_tmp2)
                  DEALLOCATE (I_ab)
                  DEALLOCATE (pab)
                  CYCLE
               END IF

               CALL dgemm("N", "N", ncoa, 1, nsgfa(iset), &
                          1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                          I_ab(1, 1), SIZE(I_ab, 1), &
                          0.0_dp, pab(1, 1), SIZE(pab, 1))
               I_ab = 0.0_dp

               igrid_level = gaussian_gridlevel(pw_env_sub%gridlevel_info, MINVAL(zeta(:, iset)))
               map_it_here = .FALSE.
               IF (.NOT. ALL(rs_v(igrid_level)%rs_grid%desc%perd == 1)) THEN
                  DO dir = 1, 3
                     ! bounds of local grid (i.e. removing the 'wings'), if periodic
                     tp(dir) = FLOOR(DOT_PRODUCT(cell%h_inv(dir, :), ra)*rs_v(igrid_level)%rs_grid%desc%npts(dir))
                     tp(dir) = MODULO(tp(dir), rs_v(igrid_level)%rs_grid%desc%npts(dir))
                     IF (rs_v(igrid_level)%rs_grid%desc%perd(dir) .NE. 1) THEN
                        lb(dir) = rs_v(igrid_level)%rs_grid%lb_local(dir)+rs_v(igrid_level)%rs_grid%desc%border
                        ub(dir) = rs_v(igrid_level)%rs_grid%ub_local(dir)-rs_v(igrid_level)%rs_grid%desc%border
                     ELSE
                        lb(dir) = rs_v(igrid_level)%rs_grid%lb_local(dir)
                        ub(dir) = rs_v(igrid_level)%rs_grid%ub_local(dir)
                     ENDIF
                     ! distributed grid, only map if it is local to the grid
                     location(dir) = tp(dir)+rs_v(igrid_level)%rs_grid%desc%lb(dir)
                  ENDDO
                  IF (lb(1) <= location(1) .AND. location(1) <= ub(1) .AND. &
                      lb(2) <= location(2) .AND. location(2) <= ub(2) .AND. &
                      lb(3) <= location(3) .AND. location(3) <= ub(3)) THEN
                     map_it_here = .TRUE.
                  ENDIF
               ELSE
                  ! not distributed, just a round-robin distribution over the full set of CPUs
                  IF (MODULO(offset, para_env_sub%num_pe) == para_env_sub%mepos) map_it_here = .TRUE.
               ENDIF

               offset = offset+nsgfa(iset)

               IF (map_it_here) THEN
                  DO ipgf = 1, npgfa(iset)
                     na1 = (ipgf-1)*ncoset(la_max(iset))+1
                     na2 = ipgf*ncoset(la_max(iset))
                     igrid_level = gaussian_gridlevel(pw_env_sub%gridlevel_info, zeta(ipgf, iset))

                     CALL integrate_pgf_product_rspace(la_max=la_max(iset), zeta=zeta(ipgf, iset)/2.0_dp, la_min=la_min(iset), &
                                                       lb_max=0, zetb=zeta(ipgf, iset)/2.0_dp, lb_min=0, &
                                                       ra=ra, rab=rab, rab2=rab2, &
                                                       rsgrid=rs_v(igrid_level)%rs_grid, &
                                                       cell=cell, &
                                                       cube_info=pw_env_sub%cube_info(igrid_level), &
                                                       hab=I_tmp2, &
                                                       pab=pab, &
                                                       o1=na1-1, &
                                                       o2=0, &
                                                       map_consistent=.TRUE., &
                                                       eps_gvg_rspace=dft_control%qs_control%eps_gvg_rspace, &
                                                       calculate_forces=.TRUE., &
                                                       force_a=force_a, force_b=force_b, &
                                                       use_virial=use_virial, my_virial_a=my_virial_a, my_virial_b=my_virial_b)

                  END DO

                  ! CALL dgemm("T","N".0_dp,sphi_a(1,sgfa),SIZE(sphi_a,1),&
                  !             I_tmp2(1,1),SIZE(I_tmp2,1),&
                  !             1.0_dp,I_ab(1,1),SIZE(I_ab,1))
                  ! L_local_col(offset-nsgfa(iset)+1:offset,i_counter)=I_ab(1:nsgfa(iset),1)
               END IF

               DEALLOCATE (I_tmp2)
               DEALLOCATE (I_ab)
               DEALLOCATE (pab)

            END DO

            force(ikind)%rho_elec(:, atom_a) = &
               force(ikind)%rho_elec(:, atom_a)+force_a(:)+force_b(:)
            IF (use_virial) THEN
               virial%pv_virial = virial%pv_virial+my_virial_a+my_virial_b
            END IF
         END DO

         DO i = 1, SIZE(rs_v)
            CALL rs_grid_release(rs_v(i)%rs_grid)
         END DO

      END DO
      CALL timestop(handle2)

      DEALLOCATE (G_PQ_local)

      DEALLOCATE (wf_vector)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, psi_L%pw)

      IF (alpha_beta) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, psi_L_beta%pw)
      ENDIF

      IF (use_virial) THEN
         CALL pw_pool_give_back_pw(auxbas_pw_pool, temp_pw_g%pw)
         DO i = 1, 3
            CALL pw_pool_give_back_pw(auxbas_pw_pool, dvg(i)%pw)
         END DO
      END IF

      CALL cp_dbcsr_release(matrix_P_inu%matrix)
      DEALLOCATE (matrix_P_inu%matrix)

      CALL cp_dbcsr_release(matrix_P_munu_nosym%matrix)
      DEALLOCATE (matrix_P_munu_nosym%matrix)

      CALL cp_dbcsr_release(matrix_P_munu%matrix)
      DEALLOCATE (matrix_P_munu%matrix)

      ! release the full gamma_P_ia structure
      DEALLOCATE (G_P_ia)

      ! Release G_P_ia_beta and and matrix_P_inu_beta
      IF (alpha_beta) THEN
         CALL cp_dbcsr_release(matrix_P_inu_beta%matrix)
         DEALLOCATE (matrix_P_inu_beta%matrix)
         DEALLOCATE (G_P_ia_beta)
      ENDIF

      DEALLOCATE (kind_of)
      DEALLOCATE (atom_of_kind)

      ! move the forces in the correct place
      DO ikind = 1, SIZE(force)
         force(ikind)%mp2_non_sep(:, :) = force(ikind)%rho_elec(:, :)
         force(ikind)%rho_elec(:, :) = 0.0_dp
      END DO

      ! Now we move from the local matrices to the global ones
      ! defined over all MPI tasks
      ! Start with moving from the DBCSR to FM for the lagrangians
      NULLIFY (L1_mu_i, fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env_sub, context=blacs_env_sub, &
                               nrow_global=dimen, ncol_global=homo)
      CALL cp_fm_create(L1_mu_i, fm_struct_tmp, name="Lag_mu_i")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_set_all(L1_mu_i, 0.0_dp)
      CALL copy_dbcsr_to_fm(matrix=Lag_mu_i_1%matrix, fm=L1_mu_i)

      ! release Lag_mu_i_1
      CALL cp_dbcsr_release(Lag_mu_i_1%matrix)
      DEALLOCATE (Lag_mu_i_1%matrix)

      NULLIFY (L2_nu_a, fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env_sub, context=blacs_env_sub, &
                               nrow_global=dimen, ncol_global=virtual)
      CALL cp_fm_create(L2_nu_a, fm_struct_tmp, name="Lag_nu_a")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_set_all(L2_nu_a, 0.0_dp)
      CALL copy_dbcsr_to_fm(matrix=Lag_nu_a_2%matrix, fm=L2_nu_a)

      ! release Lag_nu_a_2
      CALL cp_dbcsr_release(Lag_nu_a_2%matrix)
      DEALLOCATE (Lag_nu_a_2%matrix)

      ! The same for the beta Lagrangians
      IF (alpha_beta) THEN
         ! Now we move from the local matrices to the global ones
         ! defined over all MPI tasks
         ! Start with moving from the DBCSR to FM for the lagrangians
         NULLIFY (L1_mu_i_beta, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env_sub, context=blacs_env_sub, &
                                  nrow_global=dimen, ncol_global=homo_beta)
         CALL cp_fm_create(L1_mu_i_beta, fm_struct_tmp, name="Lag_mu_i_beta")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_set_all(L1_mu_i_beta, 0.0_dp)
         CALL copy_dbcsr_to_fm(matrix=Lag_mu_i_1_beta%matrix, fm=L1_mu_i_beta)

         ! release Lag_mu_i_1
         CALL cp_dbcsr_release(Lag_mu_i_1_beta%matrix)
         DEALLOCATE (Lag_mu_i_1_beta%matrix)

         NULLIFY (L2_nu_a_beta, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env_sub, context=blacs_env_sub, &
                                  nrow_global=dimen, ncol_global=virtual_beta)
         CALL cp_fm_create(L2_nu_a_beta, fm_struct_tmp, name="Lag_nu_a_beta")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_set_all(L2_nu_a_beta, 0.0_dp)
         CALL copy_dbcsr_to_fm(matrix=Lag_nu_a_2_beta%matrix, fm=L2_nu_a_beta)

         ! release Lag_nu_a_2
         CALL cp_dbcsr_release(Lag_nu_a_2_beta%matrix)
         DEALLOCATE (Lag_nu_a_2_beta%matrix)
      ENDIF

      ! Set the factor to multiply P_ij (depends on the open or closed shell)
      factor = 1.0_dp
      IF (alpha_beta) factor = 0.50_dp

      CALL create_W_P(qs_env, mp2_env, mo_coeff, homo, virtual, dimen, para_env, para_env_sub, &
                      Eigenval, L1_mu_i, L2_nu_a, factor, .FALSE.)
      ! Alpha_beta_case
      IF (alpha_beta) THEN
         CALL create_W_P(qs_env, mp2_env, mo_coeff_beta, homo_beta, virtual_beta, dimen, para_env, &
                         para_env_sub, Eigenval_beta, L1_mu_i_beta, L2_nu_a_beta, &
                         factor, .TRUE.)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE calc_ri_mp2_nonsep

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param mp2_env ...
!> \param mo_coeff ...
!> \param homo ...
!> \param virtual ...
!> \param dimen ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param Eigenval ...
!> \param L1_mu_i ...
!> \param L2_nu_a ...
!> \param factor ...
!> \param alpha_beta ...
! **************************************************************************************************
   SUBROUTINE create_W_P(qs_env, mp2_env, mo_coeff, homo, virtual, dimen, para_env, para_env_sub, &
                         Eigenval, L1_mu_i, L2_nu_a, factor, alpha_beta)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      INTEGER                                            :: homo, virtual, dimen
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      TYPE(cp_fm_type), POINTER                          :: L1_mu_i, L2_nu_a
      REAL(KIND=dp)                                      :: factor
      LOGICAL                                            :: alpha_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_W_P', routineP = moduleN//':'//routineN

      INTEGER :: color_exchange, comm_exchange, dummy_proc, handle, handle2, handle3, i, i_global, &
         i_local, iiB, iii, iproc, itmp(2), j_global, j_local, jjB, max_col_size, max_row_size, &
         my_B_virtual_end, my_B_virtual_start, mypcol, mypcol_1i, mypcol_2a, myprow, myprow_1i, &
         myprow_2a, ncol_block, ncol_block_1i, ncol_block_2a, ncol_local, ncol_local_1i, &
         ncol_local_2a, npcol, npcol_1i, npcol_2a, nprow, nprow_1i, nprow_2a, nrow_block, &
         nrow_block_1i, nrow_block_2a, nrow_local, nrow_local_1i, nrow_local_2a, number_of_rec, &
         number_of_send, proc_receive, proc_receive_static, proc_send, proc_send_ex
      INTEGER :: proc_send_static, proc_send_sub, proc_shift, rec_col_size, rec_counter, &
         rec_row_size, send_col_size, send_counter, send_pcol, send_prow, send_row_size, &
         size_rec_buffer, size_send_buffer
      INTEGER, ALLOCATABLE, DIMENSION(:) :: iii_vet, map_rec_size, map_send_size, pos_info, &
         pos_info_ex, proc_2_send_pos, proc_map_ex, req_send, sub_proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: grid_2_mepos, mepos_2_grid, &
                                                            mepos_2_grid_1i, mepos_2_grid_2a, &
                                                            sizes, sizes_1i, sizes_2a
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: col_indeces_info_1i, &
                                                            col_indeces_info_2a, &
                                                            row_indeces_info_1i, &
                                                            row_indeces_info_2a
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, col_indices_1i, &
                                                            col_indices_2a, row_indices, &
                                                            row_indices_1i, row_indices_2a
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: ab_rec, ab_send, mat_deb, mat_rec, &
                                                            mat_send
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: fm_P_ij, L_mu_q
      TYPE(cp_para_env_type), POINTER                    :: para_env_exchange
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_send
      TYPE(integ_mat_buffer_type_2D), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_cyclic

      CALL timeset(routineN, handle)

      ! create the globally distributed mixed lagrangian
      NULLIFY (blacs_env)
      CALL get_qs_env(qs_env, blacs_env=blacs_env)

      NULLIFY (L_mu_q, fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                               nrow_global=dimen, ncol_global=dimen)
      CALL cp_fm_create(L_mu_q, fm_struct_tmp, name="Lag_mu_q")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_set_all(L_mu_q, 0.0_dp)

      ! create all information array
      ALLOCATE (pos_info(0:para_env%num_pe-1))
      pos_info = 0
      pos_info(para_env%mepos) = para_env_sub%mepos
      CALL mp_sum(pos_info, para_env%group)

      ALLOCATE (sub_proc_map(-para_env_sub%num_pe:2*para_env_sub%num_pe-1))
      sub_proc_map = 0
      DO i = 0, para_env_sub%num_pe-1
         sub_proc_map(i) = i
         sub_proc_map(-i-1) = para_env_sub%num_pe-i-1
         sub_proc_map(para_env_sub%num_pe+i) = i
      END DO

      ! get matrix information for the global
      CALL cp_fm_get_info(matrix=L_mu_q, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          nrow_block=nrow_block, &
                          ncol_block=ncol_block)
      myprow = L_mu_q%matrix_struct%context%mepos(1)
      mypcol = L_mu_q%matrix_struct%context%mepos(2)
      nprow = L_mu_q%matrix_struct%context%num_pe(1)
      npcol = L_mu_q%matrix_struct%context%num_pe(2)

      ALLOCATE (grid_2_mepos(0:nprow-1, 0:npcol-1))
      grid_2_mepos = 0
      grid_2_mepos(myprow, mypcol) = para_env%mepos
      CALL mp_sum(grid_2_mepos, para_env%group)

      ! get matrix information for L1_mu_i
      CALL cp_fm_get_info(matrix=L1_mu_i, &
                          nrow_local=nrow_local_1i, &
                          ncol_local=ncol_local_1i, &
                          row_indices=row_indices_1i, &
                          col_indices=col_indices_1i, &
                          nrow_block=nrow_block_1i, &
                          ncol_block=ncol_block_1i)
      myprow_1i = L1_mu_i%matrix_struct%context%mepos(1)
      mypcol_1i = L1_mu_i%matrix_struct%context%mepos(2)
      nprow_1i = L1_mu_i%matrix_struct%context%num_pe(1)
      npcol_1i = L1_mu_i%matrix_struct%context%num_pe(2)

      ALLOCATE (mepos_2_grid_1i(0:para_env_sub%num_pe-1, 2))
      mepos_2_grid_1i = 0
      mepos_2_grid_1i(para_env_sub%mepos, 1) = myprow_1i
      mepos_2_grid_1i(para_env_sub%mepos, 2) = mypcol_1i
      CALL mp_sum(mepos_2_grid_1i, para_env_sub%group)

      ALLOCATE (sizes_1i(2, 0:para_env_sub%num_pe-1))
      sizes_1i = 0
      sizes_1i(1, para_env_sub%mepos) = nrow_local_1i
      sizes_1i(2, para_env_sub%mepos) = ncol_local_1i
      CALL mp_sum(sizes_1i, para_env_sub%group)

      ! get matrix information for L2_nu_a
      CALL cp_fm_get_info(matrix=L2_nu_a, &
                          nrow_local=nrow_local_2a, &
                          ncol_local=ncol_local_2a, &
                          row_indices=row_indices_2a, &
                          col_indices=col_indices_2a, &
                          nrow_block=nrow_block_2a, &
                          ncol_block=ncol_block_2a)
      myprow_2a = L2_nu_a%matrix_struct%context%mepos(1)
      mypcol_2a = L2_nu_a%matrix_struct%context%mepos(2)
      nprow_2a = L2_nu_a%matrix_struct%context%num_pe(1)
      npcol_2a = L2_nu_a%matrix_struct%context%num_pe(2)

      ALLOCATE (mepos_2_grid_2a(0:para_env_sub%num_pe-1, 2))
      mepos_2_grid_2a = 0
      mepos_2_grid_2a(para_env_sub%mepos, 1) = myprow_2a
      mepos_2_grid_2a(para_env_sub%mepos, 2) = mypcol_2a
      CALL mp_sum(mepos_2_grid_2a, para_env_sub%group)

      ALLOCATE (sizes_2a(2, 0:para_env_sub%num_pe-1))
      sizes_2a = 0
      sizes_2a(1, para_env_sub%mepos) = nrow_local_2a
      sizes_2a(2, para_env_sub%mepos) = ncol_local_2a
      CALL mp_sum(sizes_2a, para_env_sub%group)

      ! Here we perform a ring communication scheme taking into account
      ! for the sub-group distribution of the source matrices.
      ! as a first step we need to redistribute the data within
      ! the subgroup.
      ! In order to do so we have to allocate the structure
      ! that will hold the local data involved in communication, this
      ! structure will be the same for processes in different subgroups
      ! sharing the same position in the subgroup.
      ! -1) create the exchange para_env
      color_exchange = para_env_sub%mepos
      CALL mp_comm_split_direct(para_env%group, comm_exchange, color_exchange)
      NULLIFY (para_env_exchange)
      CALL cp_para_env_create(para_env_exchange, comm_exchange)
      ! crate the proc maps exchange and info
      ALLOCATE (proc_map_ex(-para_env_exchange%num_pe:2*para_env_exchange%num_pe-1))
      DO i = 0, para_env_exchange%num_pe-1
         proc_map_ex(i) = i
         proc_map_ex(-i-1) = para_env_exchange%num_pe-i-1
         proc_map_ex(para_env_exchange%num_pe+i) = i
      END DO
      ALLOCATE (pos_info_ex(0:para_env%num_pe-1))
      pos_info_ex = 0
      pos_info_ex(para_env%mepos) = para_env_exchange%mepos
      CALL mp_sum(pos_info_ex, para_env%group)
      ALLOCATE (sizes(2, 0:para_env_exchange%num_pe-1))
      sizes = 0
      sizes(1, para_env_exchange%mepos) = nrow_local
      sizes(2, para_env_exchange%mepos) = ncol_local
      CALL mp_sum(sizes, para_env_exchange%group)

      ! 0) store some info about indeces of the fm matrices (subgroup)
      CALL timeset(routineN//"_inx", handle2)
      ! matrix L1_mu_i
      max_row_size = MAXVAL(sizes_1i(1, :))
      max_col_size = MAXVAL(sizes_1i(2, :))
      ALLOCATE (row_indeces_info_1i(2, max_row_size, 0:para_env_sub%num_pe-1))
      ALLOCATE (col_indeces_info_1i(2, max_col_size, 0:para_env_sub%num_pe-1))
      row_indeces_info_1i = 0
      col_indeces_info_1i = 0
      ! row
      DO iiB = 1, nrow_local_1i
         i_global = row_indices_1i(iiB)
         send_prow = cp_fm_indxg2p(i_global, nrow_block, dummy_proc, &
                                   L_mu_q%matrix_struct%first_p_pos(1), nprow)
         i_local = cp_fm_indxg2l(i_global, nrow_block, dummy_proc, &
                                 L_mu_q%matrix_struct%first_p_pos(1), nprow)
         row_indeces_info_1i(1, iiB, para_env_sub%mepos) = send_prow
         row_indeces_info_1i(2, iiB, para_env_sub%mepos) = i_local
      END DO
      ! col
      DO jjB = 1, ncol_local_1i
         j_global = col_indices_1i(jjB)
         send_pcol = cp_fm_indxg2p(j_global, ncol_block, dummy_proc, &
                                   L_mu_q%matrix_struct%first_p_pos(2), npcol)
         j_local = cp_fm_indxg2l(j_global, ncol_block, dummy_proc, &
                                 L_mu_q%matrix_struct%first_p_pos(2), npcol)
         col_indeces_info_1i(1, jjB, para_env_sub%mepos) = send_pcol
         col_indeces_info_1i(2, jjB, para_env_sub%mepos) = j_local
      END DO
      CALL mp_sum(row_indeces_info_1i, para_env_sub%group)
      CALL mp_sum(col_indeces_info_1i, para_env_sub%group)

      ! matrix L2_nu_a
      max_row_size = MAXVAL(sizes_2a(1, :))
      max_col_size = MAXVAL(sizes_2a(2, :))
      ALLOCATE (row_indeces_info_2a(2, max_row_size, 0:para_env_sub%num_pe-1))
      ALLOCATE (col_indeces_info_2a(2, max_col_size, 0:para_env_sub%num_pe-1))
      row_indeces_info_2a = 0
      col_indeces_info_2a = 0
      ! row
      DO iiB = 1, nrow_local_2a
         i_global = row_indices_2a(iiB)
         send_prow = cp_fm_indxg2p(i_global, nrow_block, dummy_proc, &
                                   L_mu_q%matrix_struct%first_p_pos(1), nprow)
         i_local = cp_fm_indxg2l(i_global, nrow_block, dummy_proc, &
                                 L_mu_q%matrix_struct%first_p_pos(1), nprow)
         row_indeces_info_2a(1, iiB, para_env_sub%mepos) = send_prow
         row_indeces_info_2a(2, iiB, para_env_sub%mepos) = i_local
      END DO
      ! col
      DO jjB = 1, ncol_local_2a
         j_global = col_indices_2a(jjB)+homo
         send_pcol = cp_fm_indxg2p(j_global, ncol_block, dummy_proc, &
                                   L_mu_q%matrix_struct%first_p_pos(2), npcol)
         j_local = cp_fm_indxg2l(j_global, ncol_block, dummy_proc, &
                                 L_mu_q%matrix_struct%first_p_pos(2), npcol)
         col_indeces_info_2a(1, jjB, para_env_sub%mepos) = send_pcol
         col_indeces_info_2a(2, jjB, para_env_sub%mepos) = j_local
      END DO
      CALL mp_sum(row_indeces_info_2a, para_env_sub%group)
      CALL mp_sum(col_indeces_info_2a, para_env_sub%group)
      CALL timestop(handle2)

      ! 1) define the map for sending data in the subgroup starting with L1_mu_i
      CALL timeset(routineN//"_subinfo", handle2)
      ALLOCATE (map_send_size(0:para_env_sub%num_pe-1))
      map_send_size = 0
      DO jjB = 1, ncol_local_1i
         ! j_global=col_indices_1i(jjB)
         ! send_pcol=cp_fm_indxg2p(j_global,ncol_block,dummy_proc,&
         !                         L_mu_q%matrix_struct%first_p_pos(2),npcol)
         send_pcol = col_indeces_info_1i(1, jjB, para_env_sub%mepos)
         DO iiB = 1, nrow_local_1i
            ! i_global=row_indices_1i(iiB)
            ! send_prow=cp_fm_indxg2p(i_global,nrow_block,dummy_proc,&
            !                         L_mu_q%matrix_struct%first_p_pos(1),nprow)
            send_prow = row_indeces_info_1i(1, iiB, para_env_sub%mepos)
            proc_send = grid_2_mepos(send_prow, send_pcol)
            proc_send_sub = pos_info(proc_send)
            map_send_size(proc_send_sub) = map_send_size(proc_send_sub)+1
         END DO
      END DO
      ! and the same for L2_nu_a
      DO jjB = 1, ncol_local_2a
         ! j_global=col_indices_2a(jjB)+homo
         ! send_pcol=cp_fm_indxg2p(j_global,ncol_block,dummy_proc,&
         !                         L_mu_q%matrix_struct%first_p_pos(2),npcol)
         send_pcol = col_indeces_info_2a(1, jjB, para_env_sub%mepos)
         DO iiB = 1, nrow_local_2a
            ! i_global=row_indices_2a(iiB)
            ! send_prow=cp_fm_indxg2p(i_global,nrow_block,dummy_proc,&
            !                         L_mu_q%matrix_struct%first_p_pos(1),nprow)
            send_prow = row_indeces_info_2a(1, iiB, para_env_sub%mepos)
            proc_send = grid_2_mepos(send_prow, send_pcol)
            proc_send_sub = pos_info(proc_send)
            map_send_size(proc_send_sub) = map_send_size(proc_send_sub)+1
         END DO
      END DO
      ! and exchange data in order to create map_rec_size
      ALLOCATE (map_rec_size(0:para_env_sub%num_pe-1))
      map_rec_size = 0
      CALL mp_alltoall(map_send_size, map_rec_size, 1, para_env_sub%group)
      CALL timestop(handle2)

      ! 2) reorder data in sending buffer
      CALL timeset(routineN//"_sub_Bsend", handle2)
      ! count the number of messages (include myself)
      number_of_send = 0
      DO proc_shift = 0, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         IF (map_send_size(proc_send) > 0) THEN
            number_of_send = number_of_send+1
         END IF
      END DO
      ! allocate the structure that will hold the messages to be sent
      ALLOCATE (buffer_send(number_of_send))
      send_counter = 0
      ALLOCATE (proc_2_send_pos(0:para_env_sub%num_pe-1))
      proc_2_send_pos = 0
      DO proc_shift = 0, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         size_send_buffer = map_send_size(proc_send)
         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            ! allocate the sending buffer (msg)
            ALLOCATE (buffer_send(send_counter)%msg(size_send_buffer))
            buffer_send(send_counter)%msg = 0.0_dp
            buffer_send(send_counter)%proc = proc_send
            proc_2_send_pos(proc_send) = send_counter
         END IF
      END DO
      ! loop over the locally held data and fill the buffer_send
      ! for doing that we need an array that keep track if the
      ! sequential increase of the index for each message
      ALLOCATE (iii_vet(number_of_send))
      iii_vet = 0
      DO jjB = 1, ncol_local_1i
         ! j_global=col_indices_1i(jjB)
         ! send_pcol=cp_fm_indxg2p(j_global,ncol_block,dummy_proc,&
         !                         L_mu_q%matrix_struct%first_p_pos(2),npcol)
         send_pcol = col_indeces_info_1i(1, jjB, para_env_sub%mepos)
         DO iiB = 1, nrow_local_1i
            ! i_global=row_indices_1i(iiB)
            ! send_prow=cp_fm_indxg2p(i_global,nrow_block,dummy_proc,&
            !                         L_mu_q%matrix_struct%first_p_pos(1),nprow)
            send_prow = row_indeces_info_1i(1, iiB, para_env_sub%mepos)
            proc_send = grid_2_mepos(send_prow, send_pcol)
            proc_send_sub = pos_info(proc_send)
            send_counter = proc_2_send_pos(proc_send_sub)
            iii_vet(send_counter) = iii_vet(send_counter)+1
            iii = iii_vet(send_counter)
            buffer_send(send_counter)%msg(iii) = L1_mu_i%local_data(iiB, jjB)
         END DO
      END DO
      ! release the local data of L1_mu_i
      DEALLOCATE (L1_mu_i%local_data)
      ! and the same for L2_nu_a
      DO jjB = 1, ncol_local_2a
         ! j_global=col_indices_2a(jjB)+homo
         ! send_pcol=cp_fm_indxg2p(j_global,ncol_block,dummy_proc,&
         !                         L_mu_q%matrix_struct%first_p_pos(2),npcol)
         send_pcol = col_indeces_info_2a(1, jjB, para_env_sub%mepos)
         DO iiB = 1, nrow_local_2a
            ! i_global=row_indices_2a(iiB)
            ! send_prow=cp_fm_indxg2p(i_global,nrow_block,dummy_proc,&
            !                         L_mu_q%matrix_struct%first_p_pos(1),nprow)
            send_prow = row_indeces_info_2a(1, iiB, para_env_sub%mepos)
            proc_send = grid_2_mepos(send_prow, send_pcol)
            proc_send_sub = pos_info(proc_send)
            send_counter = proc_2_send_pos(proc_send_sub)
            iii_vet(send_counter) = iii_vet(send_counter)+1
            iii = iii_vet(send_counter)
            buffer_send(send_counter)%msg(iii) = L2_nu_a%local_data(iiB, jjB)
         END DO
      END DO
      DEALLOCATE (L2_nu_a%local_data)
      DEALLOCATE (proc_2_send_pos)
      DEALLOCATE (iii_vet)
      CALL timestop(handle2)

      ! 3) create the buffer for receive, post the message with irecv
      !    and send the messages with mp_isend
      CALL timeset(routineN//"_sub_isendrecv", handle2)
      ! count the number of messages to be received
      number_of_rec = 0
      DO proc_shift = 0, para_env_sub%num_pe-1
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
         IF (map_rec_size(proc_receive) > 0) THEN
            number_of_rec = number_of_rec+1
         END IF
      END DO
      ALLOCATE (buffer_rec(number_of_rec))
      rec_counter = 0
      DO proc_shift = 0, para_env_sub%num_pe-1
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)
         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1
            ! prepare the buffer for receive
            ALLOCATE (buffer_rec(rec_counter)%msg(size_rec_buffer))
            buffer_rec(rec_counter)%msg = 0.0_dp
            buffer_rec(rec_counter)%proc = proc_receive
            ! post the message to be received (not need to send to myself)
            IF (proc_receive /= para_env_sub%mepos) THEN
               CALL mp_irecv(buffer_rec(rec_counter)%msg, proc_receive, para_env_sub%group, &
                             buffer_rec(rec_counter)%msg_req)
            END IF
         END IF
      END DO
      ! send messages
      ALLOCATE (req_send(number_of_send))
      req_send = mp_request_null
      send_counter = 0
      DO proc_shift = 0, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            IF (proc_send == para_env_sub%mepos) THEN
               buffer_rec(send_counter)%msg = buffer_send(send_counter)%msg
            ELSE
               CALL mp_isend(buffer_send(send_counter)%msg, proc_send, para_env_sub%group, &
                             buffer_send(send_counter)%msg_req)
               req_send(send_counter) = buffer_send(send_counter)%msg_req
            END IF
         END IF
      END DO
      DEALLOCATE (map_send_size)
      CALL timestop(handle2)

      ! 4) (if memory is a problem we should move this part after point 5)
      !    Here we create the new buffer for cyclic(ring) communication and
      !    we fill it with the data received from the other member of the
      !    subgroup
      CALL timeset(routineN//"_Bcyclic", handle2)
      ! first allocata new structure
      ALLOCATE (buffer_cyclic(0:para_env_exchange%num_pe-1))
      DO iproc = 0, para_env_exchange%num_pe-1
         rec_row_size = sizes(1, iproc)
         rec_col_size = sizes(2, iproc)
         ALLOCATE (buffer_cyclic(iproc)%msg(rec_row_size, rec_col_size))
         buffer_cyclic(iproc)%msg = 0.0_dp
      END DO
      ! now collect data from other member of the subgroup and fill
      ! buffer_cyclic
      rec_counter = 0
      DO proc_shift = 0, para_env_sub%num_pe-1
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)
         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1

            ! wait for the message
            IF (proc_receive /= para_env_sub%mepos) CALL mp_wait(buffer_rec(rec_counter)%msg_req)

            CALL timeset(routineN//"_fill", handle3)
            iii = 0
            DO jjB = 1, sizes_1i(2, proc_receive)
               ! j_global=cp_fm_indxl2g(jjB,ncol_block_1i,mepos_2_grid_1i(proc_receive,2),&
               !                        L1_mu_i%matrix_struct%first_p_pos(2),npcol_1i)
               ! send_pcol=cp_fm_indxg2p(j_global,ncol_block,dummy_proc,&
               !                         L_mu_q%matrix_struct%first_p_pos(2),npcol)
               ! j_local=cp_fm_indxg2l(j_global,ncol_block,dummy_proc,&
               !                       L_mu_q%matrix_struct%first_p_pos(2),npcol)
               send_pcol = col_indeces_info_1i(1, jjB, proc_receive)
               j_local = col_indeces_info_1i(2, jjB, proc_receive)
               DO iiB = 1, sizes_1i(1, proc_receive)
                  ! i_global=cp_fm_indxl2g(iiB,nrow_block_1i,mepos_2_grid_1i(proc_receive,1),&
                  !                        L1_mu_i%matrix_struct%first_p_pos(1),nprow_1i)
                  ! send_prow=cp_fm_indxg2p(i_global,nrow_block,dummy_proc,&
                  !                         L_mu_q%matrix_struct%first_p_pos(1),nprow)
                  send_prow = row_indeces_info_1i(1, iiB, proc_receive)
                  proc_send = grid_2_mepos(send_prow, send_pcol)
                  proc_send_sub = pos_info(proc_send)
                  IF (proc_send_sub /= para_env_sub%mepos) CYCLE
                  iii = iii+1
                  ! i_local=cp_fm_indxg2l(i_global,nrow_block,dummy_proc,&
                  !                       L_mu_q%matrix_struct%first_p_pos(1),nprow)
                  i_local = row_indeces_info_1i(2, iiB, proc_receive)
                  proc_send_ex = pos_info_ex(proc_send)
                  buffer_cyclic(proc_send_ex)%msg(i_local, j_local) = buffer_rec(rec_counter)%msg(iii)
               END DO
            END DO
            ! and the same for L2_nu_a
            DO jjB = 1, sizes_2a(2, proc_receive)
               ! j_global=cp_fm_indxl2g(jjB,ncol_block_2a,mepos_2_grid_2a(proc_receive,2),&
               !                        L2_nu_a%matrix_struct%first_p_pos(2),npcol_2a)+homo
               ! send_pcol=cp_fm_indxg2p(j_global,ncol_block,dummy_proc,&
               !                         L_mu_q%matrix_struct%first_p_pos(2),npcol)
               ! j_local=cp_fm_indxg2l(j_global,ncol_block,dummy_proc,&
               !                       L_mu_q%matrix_struct%first_p_pos(2),npcol)
               send_pcol = col_indeces_info_2a(1, jjB, proc_receive)
               j_local = col_indeces_info_2a(2, jjB, proc_receive)
               DO iiB = 1, sizes_2a(1, proc_receive)
                  ! i_global=cp_fm_indxl2g(iiB,nrow_block_2a,mepos_2_grid_2a(proc_receive,1),&
                  !                        L2_nu_a%matrix_struct%first_p_pos(1),nprow_2a)
                  ! send_prow=cp_fm_indxg2p(i_global,nrow_block,dummy_proc,&
                  !                         L_mu_q%matrix_struct%first_p_pos(1),nprow)
                  send_prow = row_indeces_info_2a(1, iiB, proc_receive)
                  proc_send = grid_2_mepos(send_prow, send_pcol)
                  proc_send_sub = pos_info(proc_send)
                  IF (proc_send_sub /= para_env_sub%mepos) CYCLE
                  iii = iii+1
                  ! i_local=cp_fm_indxg2l(i_global,nrow_block,dummy_proc,&
                  !                       L_mu_q%matrix_struct%first_p_pos(1),nprow)
                  i_local = row_indeces_info_2a(2, iiB, proc_receive)
                  proc_send_ex = pos_info_ex(proc_send)
                  buffer_cyclic(proc_send_ex)%msg(i_local, j_local) = buffer_rec(rec_counter)%msg(iii)
               END DO
            END DO
            CALL timestop(handle3)

            ! deallocate the received message
            DEALLOCATE (buffer_rec(rec_counter)%msg)
         END IF
      END DO
      DEALLOCATE (row_indeces_info_1i)
      DEALLOCATE (col_indeces_info_1i)
      DEALLOCATE (row_indeces_info_2a)
      DEALLOCATE (col_indeces_info_2a)
      DEALLOCATE (buffer_rec)
      DEALLOCATE (map_rec_size)
      CALL timestop(handle2)

      ! 5)  Wait for all messeges to be sent in the subgroup
      CALL timeset(routineN//"_sub_waitall", handle2)
      CALL mp_waitall(req_send(:))
      DO send_counter = 1, number_of_send
         DEALLOCATE (buffer_send(send_counter)%msg)
      END DO
      DEALLOCATE (buffer_send)
      DEALLOCATE (req_send)
      CALL timestop(handle2)

      ! 6) Start with ring communication
      CALL timeset(routineN//"_ring", handle2)
      proc_send_static = proc_map_ex(para_env_exchange%mepos+1)
      proc_receive_static = proc_map_ex(para_env_exchange%mepos-1)
      max_row_size = MAXVAL(sizes(1, :))
      max_col_size = MAXVAL(sizes(2, :))
      ALLOCATE (mat_send(max_row_size, max_col_size))
      ALLOCATE (mat_rec(max_row_size, max_col_size))
      mat_send = 0.0_dp
      mat_send(1:nrow_local, 1:ncol_local) = buffer_cyclic(para_env_exchange%mepos)%msg(:, :)
      DEALLOCATE (buffer_cyclic(para_env_exchange%mepos)%msg)
      DO proc_shift = 1, para_env_exchange%num_pe-1
         proc_send = proc_map_ex(para_env_exchange%mepos+proc_shift)
         proc_receive = proc_map_ex(para_env_exchange%mepos-proc_shift)

         rec_row_size = sizes(1, proc_receive)
         rec_col_size = sizes(2, proc_receive)

         mat_rec = 0.0_dp
         CALL mp_sendrecv(mat_send, proc_send_static, &
                          mat_rec, proc_receive_static, &
                          para_env_exchange%group)

         mat_send = 0.0_dp
         mat_send(1:rec_row_size, 1:rec_col_size) = mat_rec(1:rec_row_size, 1:rec_col_size)+ &
                                                    buffer_cyclic(proc_receive)%msg(:, :)

         DEALLOCATE (buffer_cyclic(proc_receive)%msg)
      END DO
      ! and finaly
      CALL mp_sendrecv(mat_send, proc_send_static, &
                       mat_rec, proc_receive_static, &
                       para_env_exchange%group)
      L_mu_q%local_data(1:nrow_local, 1:ncol_local) = mat_rec(1:nrow_local, 1:ncol_local)
      DEALLOCATE (buffer_cyclic)
      DEALLOCATE (mat_send)
      DEALLOCATE (mat_rec)
      CALL timestop(handle2)

      DEALLOCATE (proc_map_ex)
      ! release para_env_exchange
      CALL cp_para_env_release(para_env_exchange)

      CALL cp_fm_release(L1_mu_i)
      CALL cp_fm_release(L2_nu_a)
      DEALLOCATE (pos_info_ex)
      DEALLOCATE (grid_2_mepos)
      DEALLOCATE (mepos_2_grid_1i)
      DEALLOCATE (mepos_2_grid_2a)
      DEALLOCATE (sizes)
      DEALLOCATE (sizes_1i)
      DEALLOCATE (sizes_2a)

      ! update the P_ij block of P_MP2 with the
      ! non-singular ij pairs
      CALL timeset(routineN//"_Pij", handle2)
      NULLIFY (fm_P_ij, fm_struct_tmp)
      CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                               nrow_global=homo, ncol_global=homo)
      CALL cp_fm_create(fm_P_ij, fm_struct_tmp, name="fm_P_ij")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_fm_set_all(fm_P_ij, 0.0_dp)

      CALL cp_gemm('T', 'N', homo, homo, dimen, 1.0_dp, &
                   mo_coeff, L_mu_q, 0.0_dp, fm_P_ij, &
                   a_first_col=1, &
                   a_first_row=1, &
                   b_first_col=1, &
                   b_first_row=1, &
                   c_first_col=1, &
                   c_first_row=1)
      ! don't know if it is better to transpose (for communication reasons)
      ! or just recompute the transposed matrix
      CALL cp_gemm('T', 'N', homo, homo, dimen, -2.0_dp, &
                   L_mu_q, mo_coeff, 2.0_dp, fm_P_ij, &
                   a_first_col=1, &
                   a_first_row=1, &
                   b_first_col=1, &
                   b_first_row=1, &
                   c_first_col=1, &
                   c_first_row=1)
      ! we have it, update P_ij local
      CALL cp_fm_get_info(matrix=fm_P_ij, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)
      DO jjB = 1, ncol_local
         j_global = col_indices(jjB)
         DO iiB = 1, nrow_local
            i_global = row_indices(iiB)
            ! diagonal elements alreasy updated
            IF (j_global == i_global) CYCLE
            ! check if the given element is above the threshold
            IF (ABS(Eigenval(j_global)-Eigenval(i_global)) < mp2_env%ri_mp2%eps_canonical) CYCLE
            IF (.NOT. alpha_beta) THEN
               mp2_env%ri_grad%P_ij(i_global, j_global) = &
                  factor*fm_P_ij%local_data(iiB, jjB)/(Eigenval(j_global)-Eigenval(i_global))
            ELSE
               mp2_env%ri_grad%P_ij_beta(i_global, j_global) = &
                  factor*fm_P_ij%local_data(iiB, jjB)/(Eigenval(j_global)-Eigenval(i_global))
            ENDIF
         END DO
      END DO
      ! release fm_P_ij
      CALL cp_fm_release(fm_P_ij)
      ! mp_sum it (we can avoid mp_sum, but for now let's keep it easy)
      IF (.NOT. alpha_beta) THEN
         CALL mp_sum(mp2_env%ri_grad%P_ij, para_env%group)
      ELSE
         CALL mp_sum(mp2_env%ri_grad%P_ij_beta, para_env%group)
      ENDIF
      CALL timestop(handle2)

      ! Now create and fill the P matrix (MO)
      ! FOR NOW WE ASSUME P_ab AND P_ij ARE REPLICATED OVER EACH MPI RANK
      IF (.NOT. alpha_beta) THEN
         CALL timeset(routineN//"_PMO", handle2)
         NULLIFY (mp2_env%ri_grad%P_mo, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=dimen, ncol_global=dimen)
         CALL cp_fm_create(mp2_env%ri_grad%P_mo, fm_struct_tmp, name="P_MP2_MO")
         CALL cp_fm_set_all(mp2_env%ri_grad%P_mo, 0.0_dp)
      ELSE
         CALL timeset(routineN//"_PMO", handle2)
         NULLIFY (mp2_env%ri_grad%P_mo_beta, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=dimen, ncol_global=dimen)
         CALL cp_fm_create(mp2_env%ri_grad%P_mo_beta, fm_struct_tmp, name="P_MP2_MO")
         CALL cp_fm_set_all(mp2_env%ri_grad%P_mo_beta, 0.0_dp)
      ENDIF

      ! start with the (easy) occ-occ block and locally held P_ab elements
      itmp = get_limit(virtual, para_env_sub%num_pe, para_env_sub%mepos)
      my_B_virtual_start = itmp(1)
      my_B_virtual_end = itmp(2)

      IF (.NOT. alpha_beta) THEN
         CALL cp_fm_get_info(matrix=mp2_env%ri_grad%P_mo, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices, &
                             nrow_block=nrow_block, &
                             ncol_block=ncol_block)
         myprow = mp2_env%ri_grad%P_mo%matrix_struct%context%mepos(1)
         mypcol = mp2_env%ri_grad%P_mo%matrix_struct%context%mepos(2)
         nprow = mp2_env%ri_grad%P_mo%matrix_struct%context%num_pe(1)
         npcol = mp2_env%ri_grad%P_mo%matrix_struct%context%num_pe(2)
      ELSE
         CALL cp_fm_get_info(matrix=mp2_env%ri_grad%P_mo_beta, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices, &
                             nrow_block=nrow_block, &
                             ncol_block=ncol_block)
         myprow = mp2_env%ri_grad%P_mo_beta%matrix_struct%context%mepos(1)
         mypcol = mp2_env%ri_grad%P_mo_beta%matrix_struct%context%mepos(2)
         nprow = mp2_env%ri_grad%P_mo_beta%matrix_struct%context%num_pe(1)
         npcol = mp2_env%ri_grad%P_mo_beta%matrix_struct%context%num_pe(2)
      ENDIF

      DO jjB = 1, ncol_local
         j_global = col_indices(jjB)
         DO iiB = 1, nrow_local
            i_global = row_indices(iiB)
            IF (.NOT. alpha_beta) THEN
               IF (i_global <= homo .AND. j_global <= homo) THEN
                  mp2_env%ri_grad%P_mo%local_data(iiB, jjB) = mp2_env%ri_grad%P_ij(i_global, j_global)
               END IF
               IF ((my_B_virtual_start <= i_global-homo .AND. i_global-homo <= my_B_virtual_end) .AND. (j_global > homo)) THEN
                  mp2_env%ri_grad%P_mo%local_data(iiB, jjB) = &
                     mp2_env%ri_grad%P_ab(i_global-homo-my_B_virtual_start+1, j_global-homo)
               END IF
            ELSE
               IF (i_global <= homo .AND. j_global <= homo) THEN
                  mp2_env%ri_grad%P_mo_beta%local_data(iiB, jjB) = mp2_env%ri_grad%P_ij_beta(i_global, j_global)
               END IF
               IF ((my_B_virtual_start <= i_global-homo .AND. i_global-homo <= my_B_virtual_end) .AND. (j_global > homo)) THEN
                  mp2_env%ri_grad%P_mo_beta%local_data(iiB, jjB) = &
                     mp2_env%ri_grad%P_ab_beta(i_global-homo-my_B_virtual_start+1, j_global-homo)
               END IF
            ENDIF
         END DO
      END DO
      ! deallocate the local P_ij
      IF (.NOT. alpha_beta) THEN
         DEALLOCATE (mp2_env%ri_grad%P_ij)
      ELSE
         DEALLOCATE (mp2_env%ri_grad%P_ij_beta)
      ENDIF

      ! send around the sub_group the local data and check if we
      ! have to update our block with external elements
      ALLOCATE (mepos_2_grid(0:para_env_sub%num_pe-1, 2))
      mepos_2_grid = 0
      mepos_2_grid(para_env_sub%mepos, 1) = myprow
      mepos_2_grid(para_env_sub%mepos, 2) = mypcol
      CALL mp_sum(mepos_2_grid, para_env_sub%group)

      ALLOCATE (sizes(2, 0:para_env_sub%num_pe-1))
      sizes = 0
      sizes(1, para_env_sub%mepos) = nrow_local
      sizes(2, para_env_sub%mepos) = ncol_local
      CALL mp_sum(sizes, para_env_sub%group)

      ALLOCATE (ab_rec(nrow_local, ncol_local))
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

         send_prow = mepos_2_grid(proc_send, 1)
         send_pcol = mepos_2_grid(proc_send, 2)

         send_row_size = sizes(1, proc_send)
         send_col_size = sizes(2, proc_send)

         ALLOCATE (ab_send(send_row_size, send_col_size))
         ab_send = 0.0_dp

         ! first loop over row since in this way we can cycle
         DO iiB = 1, send_row_size
            i_global = cp_fm_indxl2g(iiB, nrow_block, send_prow, &
                                     mp2_env%ri_grad%P_mo%matrix_struct%first_p_pos(1), nprow)
            IF (i_global <= homo) CYCLE
            i_global = i_global-homo
            IF (.NOT. (my_B_virtual_start <= i_global .AND. i_global <= my_B_virtual_end)) CYCLE
            DO jjB = 1, send_col_size
               IF (.NOT. alpha_beta) THEN
                  j_global = cp_fm_indxl2g(jjB, ncol_block, send_pcol, &
                                           mp2_env%ri_grad%P_mo%matrix_struct%first_p_pos(2), npcol)
               ELSE
                  j_global = cp_fm_indxl2g(jjB, ncol_block, send_pcol, &
                                           mp2_env%ri_grad%P_mo_beta%matrix_struct%first_p_pos(2), npcol)
               ENDIF
               IF (j_global <= homo) CYCLE
               j_global = j_global-homo
               IF (.NOT. alpha_beta) THEN
                  ab_send(iiB, jjB) = mp2_env%ri_grad%P_ab(i_global-my_B_virtual_start+1, j_global)
               ELSE
                  ab_send(iiB, jjB) = mp2_env%ri_grad%P_ab_beta(i_global-my_B_virtual_start+1, j_global)
               ENDIF
            END DO
         END DO

         ab_rec = 0.0_dp
         CALL mp_sendrecv(ab_send, proc_send, &
                          ab_rec, proc_receive, &
                          para_env_sub%group)
         IF (.NOT. alpha_beta) THEN
            mp2_env%ri_grad%P_mo%local_data(1:nrow_local, 1:ncol_local) = &
               mp2_env%ri_grad%P_mo%local_data(1:nrow_local, 1:ncol_local)+ &
               ab_rec(1:nrow_local, 1:ncol_local)
         ELSE
            mp2_env%ri_grad%P_mo_beta%local_data(1:nrow_local, 1:ncol_local) = &
               mp2_env%ri_grad%P_mo_beta%local_data(1:nrow_local, 1:ncol_local)+ &
               ab_rec(1:nrow_local, 1:ncol_local)
         ENDIF

         DEALLOCATE (ab_send)
      END DO
      DEALLOCATE (ab_rec)
      DEALLOCATE (mepos_2_grid)
      DEALLOCATE (sizes)

      ! deallocate the local P_ab'
      IF (.NOT. alpha_beta) THEN
         DEALLOCATE (mp2_env%ri_grad%P_ab)
      ELSE
         DEALLOCATE (mp2_env%ri_grad%P_ab_beta)
      ENDIF
      CALL timestop(handle2)

      ! create also W_MP2_MO
      CALL timeset(routineN//"_WMO", handle2)
      IF (.NOT. alpha_beta) THEN
         NULLIFY (mp2_env%ri_grad%W_mo)
         CALL cp_fm_create(mp2_env%ri_grad%W_mo, fm_struct_tmp, name="W_MP2_MO")
         CALL cp_fm_set_all(mp2_env%ri_grad%W_mo, 0.0_dp)
         CALL cp_fm_struct_release(fm_struct_tmp)

         ! all block
         CALL cp_gemm('T', 'N', dimen, dimen, dimen, 2.0_dp*factor, &
                      L_mu_q, mo_coeff, 0.0_dp, mp2_env%ri_grad%W_mo, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)

         ! occ-occ block
         CALL cp_gemm('T', 'N', homo, homo, dimen, -2.0_dp*factor, &
                      L_mu_q, mo_coeff, 0.0_dp, mp2_env%ri_grad%W_mo, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)

         ! occ-virt block
         CALL cp_gemm('T', 'N', homo, virtual, dimen, 2.0_dp*factor, &
                      mo_coeff, L_mu_q, 0.0_dp, mp2_env%ri_grad%W_mo, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=homo+1, &
                      b_first_row=1, &
                      c_first_col=homo+1, &
                      c_first_row=1)
      ELSE
         NULLIFY (mp2_env%ri_grad%W_mo_beta)
         CALL cp_fm_create(mp2_env%ri_grad%W_mo_beta, fm_struct_tmp, name="W_MP2_MO")
         CALL cp_fm_set_all(mp2_env%ri_grad%W_mo_beta, 0.0_dp)
         CALL cp_fm_struct_release(fm_struct_tmp)

         ! all block
         CALL cp_gemm('T', 'N', dimen, dimen, dimen, 2.0_dp*factor, &
                      L_mu_q, mo_coeff, 0.0_dp, mp2_env%ri_grad%W_mo_beta, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)

         ! occ-occ block
         CALL cp_gemm('T', 'N', homo, homo, dimen, -2.0_dp*factor, &
                      L_mu_q, mo_coeff, 0.0_dp, mp2_env%ri_grad%W_mo_beta, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)

         ! occ-virt block
         CALL cp_gemm('T', 'N', homo, virtual, dimen, 2.0_dp*factor, &
                      mo_coeff, L_mu_q, 0.0_dp, mp2_env%ri_grad%W_mo_beta, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=homo+1, &
                      b_first_row=1, &
                      c_first_col=homo+1, &
                      c_first_row=1)
      ENDIF
      CALL timestop(handle2)

      ! IF(para_env%mepos==0) CALL write_array(mp2_env%ri_grad%P_ij(1:homo,1:homo))

      ! Calculate occ-virt block of the lagrangian in MO
      CALL timeset(routineN//"_Ljb", handle2)
      IF (.NOT. alpha_beta) THEN
         NULLIFY (mp2_env%ri_grad%L_jb, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=homo, ncol_global=virtual)
         CALL cp_fm_create(mp2_env%ri_grad%L_jb, fm_struct_tmp, name="fm_L_jb")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_set_all(mp2_env%ri_grad%L_jb, 0.0_dp)

         ! first Virtual
         CALL cp_gemm('T', 'N', homo, virtual, dimen, 2.0_dp*factor, &
                      L_mu_q, mo_coeff, 0.0_dp, mp2_env%ri_grad%L_jb, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=homo+1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
         ! then occupied
         CALL cp_gemm('T', 'N', homo, virtual, dimen, 2.0_dp*factor, &
                      mo_coeff, L_mu_q, 1.0_dp, mp2_env%ri_grad%L_jb, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=homo+1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
      ELSE
         NULLIFY (mp2_env%ri_grad%L_jb_beta, fm_struct_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=para_env, context=blacs_env, &
                                  nrow_global=homo, ncol_global=virtual)
         CALL cp_fm_create(mp2_env%ri_grad%L_jb_beta, fm_struct_tmp, name="fm_L_jb")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_set_all(mp2_env%ri_grad%L_jb_beta, 0.0_dp)

         ! first Virtual
         CALL cp_gemm('T', 'N', homo, virtual, dimen, 2.0_dp*factor, &
                      L_mu_q, mo_coeff, 0.0_dp, mp2_env%ri_grad%L_jb_beta, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=homo+1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
         ! then occupied
         CALL cp_gemm('T', 'N', homo, virtual, dimen, 2.0_dp*factor, &
                      mo_coeff, L_mu_q, 1.0_dp, mp2_env%ri_grad%L_jb_beta, &
                      a_first_col=1, &
                      a_first_row=1, &
                      b_first_col=homo+1, &
                      b_first_row=1, &
                      c_first_col=1, &
                      c_first_row=1)
      ENDIF
      ! finally release L_mu_q
      CALL cp_fm_release(L_mu_q)
      CALL timestop(handle2)

      ! here we should be done next CPHF

      IF (.FALSE.) THEN
         ALLOCATE (ab_send(homo, virtual))
         ab_send = 0.0_dp
         IF (.NOT. alpha_beta) THEN
            CALL cp_fm_get_info(matrix=mp2_env%ri_grad%L_jb, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
         ELSE
            CALL cp_fm_get_info(matrix=mp2_env%ri_grad%L_jb_beta, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
         ENDIF
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               IF (.NOT. alpha_beta) THEN
                  ab_send(i_global, j_global) = mp2_env%ri_grad%L_jb%local_data(iiB, jjB)
               ELSE
                  ab_send(i_global, j_global) = mp2_env%ri_grad%L_jb_beta%local_data(iiB, jjB)
               ENDIF
            END DO
         END DO
         CALL mp_sum(ab_send, para_env%group)
         IF (para_env%mepos == 0) THEN
            WRITE (*, *) 'Ljb'
            CALL write_array(ab_send(1:homo, 1:virtual))
         END IF
         DEALLOCATE (ab_send)
         ALLOCATE (ab_send(homo, homo))
         ALLOCATE (ab_rec(virtual, virtual))
         ALLOCATE (mat_deb(homo, virtual))
         ab_send = 0.0_dp
         ab_rec = 0.0_dp
         mat_deb = 0.0_dp
         IF (.NOT. alpha_beta) THEN
            CALL cp_fm_get_info(matrix=mp2_env%ri_grad%P_mo, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
         ELSE
            CALL cp_fm_get_info(matrix=mp2_env%ri_grad%P_mo_beta, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
         ENDIF
         IF (.NOT. alpha_beta) THEN
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  IF (i_global <= homo .AND. j_global <= homo) THEN
                     ab_send(i_global, j_global) = mp2_env%ri_grad%P_mo%local_data(iiB, jjB)
                  ELSE
                     IF (i_global > homo .AND. j_global > homo) THEN
                        ab_rec(i_global-homo, j_global-homo) = mp2_env%ri_grad%P_mo%local_data(iiB, jjB)
                     ELSE
                        mat_deb(MIN(i_global, j_global), MAX(i_global, j_global)-homo) = &
                           mat_deb(MIN(i_global, j_global), MAX(i_global, j_global)-homo)+ &
                           mp2_env%ri_grad%P_mo%local_data(iiB, jjB)
                     END IF
                  END IF
               END DO
            END DO
         ELSE
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  IF (i_global <= homo .AND. j_global <= homo) THEN
                     ab_send(i_global, j_global) = mp2_env%ri_grad%P_mo_beta%local_data(iiB, jjB)
                  ELSE
                     IF (i_global > homo .AND. j_global > homo) THEN
                        ab_rec(i_global-homo, j_global-homo) = mp2_env%ri_grad%P_mo_beta%local_data(iiB, jjB)
                     ELSE
                        mat_deb(MIN(i_global, j_global), MAX(i_global, j_global)-homo) = &
                           mat_deb(MIN(i_global, j_global), MAX(i_global, j_global)-homo)+ &
                           mp2_env%ri_grad%P_mo_beta%local_data(iiB, jjB)
                     END IF
                  END IF
               END DO
            END DO
         ENDIF

         CALL mp_sum(ab_send, para_env%group)
         CALL mp_sum(ab_rec, para_env%group)
         CALL mp_sum(mat_deb, para_env%group)
         IF (para_env%mepos == 0) THEN
            WRITE (*, *) 'Pij'
            CALL write_array(ab_send(1:homo, 1:homo))
            WRITE (*, *) 'Pab'
            CALL write_array(ab_rec(1:virtual, 1:virtual))
            WRITE (*, *) 'Pia'
            CALL write_array(mat_deb(1:homo, 1:virtual))
         END IF
         DEALLOCATE (ab_send, ab_rec, mat_deb)
         ALLOCATE (ab_send(dimen, dimen))
         ab_send = 0.0_dp
         IF (.NOT. alpha_beta) THEN
            CALL cp_fm_get_info(matrix=mp2_env%ri_grad%W_mo, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
         ELSE
            CALL cp_fm_get_info(matrix=mp2_env%ri_grad%W_mo_beta, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
         ENDIF

         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               IF (.NOT. alpha_beta) THEN
                  ab_send(i_global, j_global) = &
                     mp2_env%ri_grad%W_mo%local_data(iiB, jjB)
               ELSE
                  ab_send(i_global, j_global) = &
                     mp2_env%ri_grad%W_mo_beta%local_data(iiB, jjB)
               ENDIF
            END DO
         END DO
         CALL mp_sum(ab_send, para_env%group)
         IF (para_env%mepos == 0) THEN
            WRITE (*, *) 'Wij'
            CALL write_array(ab_send(1:dimen, 1:dimen))
         END IF
         DEALLOCATE (ab_send)
      END IF

      DEALLOCATE (sub_proc_map)
      DEALLOCATE (pos_info)

      CALL timestop(handle)

   END SUBROUTINE create_W_P

! **************************************************************************************************
!> \brief ...
!> \param mat ...
!> \param unitout ...
! **************************************************************************************************
   SUBROUTINE write_array(mat, unitout)
      REAL(KIND=dp), DIMENSION(:, :)                     :: mat
      INTEGER, OPTIONAL                                  :: unitout

      INTEGER                                            :: iii, jjj

      WRITE (*, *)
      DO iii = 1, SIZE(mat, 1)
         WRITE (*, *) iii
         DO jjj = 1, SIZE(mat, 2), 20
            IF (PRESENT(unitout)) THEN
               WRITE (1000+unitout, '(1000F10.5)') mat(iii, jjj:MIN(SIZE(mat, 2), jjj+19))
            ELSE
               WRITE (*, '(1000F10.5)') mat(iii, jjj:MIN(SIZE(mat, 2), jjj+19))
            END IF
         END DO
         WRITE (*, *)
      END DO
      WRITE (*, *)
   END SUBROUTINE

END MODULE mp2_ri_grad
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines for RPA with imaginary time
!> \par History
!>      10.2015 created [Jan Wilhelm]
! **************************************************************************************************
MODULE rpa_im_time
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_copy, cp_dbcsr_filter, cp_dbcsr_get_num_blocks, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_set, cp_dbcsr_type
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale,&
                                              cp_fm_scale_and_add
   USE cp_fm_struct,                    ONLY: cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_comm_split_direct,&
                                              mp_irecv,&
                                              mp_isend,&
                                              mp_max,&
                                              mp_sendrecv,&
                                              mp_sum,&
                                              mp_sync,&
                                              mp_waitall
   USE mp2_types,                       ONLY: integ_mat_buffer_type,&
                                              mp2_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rpa_im_time'

   PUBLIC :: fill_fm_with_2d_array, &
             compute_fm_mat_Q_omega
CONTAINS

! **************************************************************************************************
!> \brief compute the matrix Q(it) (intermediate) and Fourier transform it
!>        directly to fm_mat_Q_omega(iw) (output)
!> \param fm_mat_Q_omega ...
!> \param fm_scaled_dm_occ ...
!> \param fm_scaled_dm_virt ...
!> \param fm_mo_coeff_occ ...
!> \param fm_mo_coeff_virt ...
!> \param fm_mat_Q ...
!> \param mat_M_munu_occ ...
!> \param mat_M_munu_virt ...
!> \param mat_M_munu_P_occ ...
!> \param mat_M_munu_P_virt ...
!> \param mat_B_munu ...
!> \param mat_munu ...
!> \param trace_2d ...
!> \param tau_wj_2d ...
!> \param tj ...
!> \param wj ...
!> \param first_cycle_im_time ...
!> \param do_minimax_quad ...
!> \param tau ...
!> \param tau_old ...
!> \param e_fermi ...
!> \param eps_filter ...
!> \param a_scaling ...
!> \param alpha ...
!> \param Eigenval ...
!> \param nmo ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
!> \param dimen_RI ...
!> \param max_size_RI_group ...
!> \param ngroup ...
!> \param num_integ_points ...
!> \param jquad ...
!> \param mp2_env ...
!> \param para_env ...
!> \param para_env_sub ...
! **************************************************************************************************
   SUBROUTINE compute_fm_mat_Q_omega(fm_mat_Q_omega, fm_scaled_dm_occ, &
                                     fm_scaled_dm_virt, fm_mo_coeff_occ, fm_mo_coeff_virt, &
                                     fm_mat_Q, mat_M_munu_occ, mat_M_munu_virt, mat_M_munu_P_occ, &
                                     mat_M_munu_P_virt, mat_B_munu, mat_munu, trace_2d, tau_wj_2d, tj, wj, &
                                     first_cycle_im_time, do_minimax_quad, &
                                     tau, tau_old, e_fermi, eps_filter, a_scaling, alpha, &
                                     Eigenval, nmo, my_group_L_size, &
                                     my_group_L_start, dimen_RI, max_size_RI_group, ngroup, &
                                     num_integ_points, jquad, mp2_env, para_env, para_env_sub)

      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fm_mat_Q_omega
      TYPE(cp_fm_type), POINTER                          :: fm_scaled_dm_occ, fm_scaled_dm_virt, &
                                                            fm_mo_coeff_occ, fm_mo_coeff_virt, &
                                                            fm_mat_Q
      TYPE(cp_dbcsr_type), POINTER                       :: mat_M_munu_occ, mat_M_munu_virt
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_M_munu_P_occ, mat_M_munu_P_virt, &
                                                            mat_B_munu
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: trace_2d, tau_wj_2d
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tj, wj
      LOGICAL                                            :: first_cycle_im_time, do_minimax_quad
      REAL(KIND=dp)                                      :: tau, tau_old, e_fermi, eps_filter, &
                                                            a_scaling, alpha
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: nmo, my_group_L_size, my_group_L_start, &
                                                            dimen_RI, max_size_RI_group, ngroup, &
                                                            num_integ_points, jquad
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_fm_mat_Q_omega', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle4, handle5, i_global, iiB, &
                                                            iquad, jjB, LLL, ncol_local, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: first_cycle_omega_loop
      REAL(KIND=dp)                                      :: omega, omega_old, weight, weight_old

      CALL timeset(routineN, handle)

      ! Multiply the occupied and the virtual MO coefficients with the factor exp((-e_i-e_F)*tau/2).
      ! Then, we simply get the sum over all occ states and virt. states by a simple matrix-matrix
      ! multiplication.

      ! get info of fm_mo_coeff_occ
      CALL cp_fm_get_info(matrix=fm_mo_coeff_occ, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)

      CALL timeset(routineN//"_im_time_calc_scaled_dm", handle4)

      ! first, the occ
      IF (first_cycle_im_time) THEN

         DO jjB = 1, nrow_local
            DO iiB = 1, ncol_local
               i_global = col_indices(iiB)

               fm_mo_coeff_occ%local_data(jjB, iiB) = &
                  fm_mo_coeff_occ%local_data(jjB, iiB)*EXP(tau*0.5_dp*(Eigenval(i_global)-e_fermi))
            END DO
         END DO

      ELSE

         DO jjB = 1, nrow_local
            DO iiB = 1, ncol_local
               i_global = col_indices(iiB)

               fm_mo_coeff_occ%local_data(jjB, iiB) = &
                  fm_mo_coeff_occ%local_data(jjB, iiB)*EXP((tau-tau_old)*0.5_dp*(Eigenval(i_global)-e_fermi))
            END DO
         END DO

      END IF

      ! get info of fm_mo_coeff_virt
      CALL cp_fm_get_info(matrix=fm_mo_coeff_virt, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)

      ! the same for virt
      IF (first_cycle_im_time) THEN

         DO jjB = 1, nrow_local
            DO iiB = 1, ncol_local
               i_global = col_indices(iiB)

               fm_mo_coeff_virt%local_data(jjB, iiB) = &
                  fm_mo_coeff_virt%local_data(jjB, iiB)*EXP(-tau*0.5_dp*(Eigenval(i_global)-e_fermi))
            END DO
         END DO

      ELSE

         DO jjB = 1, nrow_local
            DO iiB = 1, ncol_local
               i_global = col_indices(iiB)

               fm_mo_coeff_virt%local_data(jjB, iiB) = &
                  fm_mo_coeff_virt%local_data(jjB, iiB)*EXP(-(tau-tau_old)*0.5_dp*(Eigenval(i_global)-e_fermi))
            END DO
         END DO

      END IF

      CALL cp_gemm(transa="N", transb="T", m=nmo, n=nmo, k=nmo, alpha=1.0_dp, &
                   matrix_a=fm_mo_coeff_occ, matrix_b=fm_mo_coeff_occ, beta=0.0_dp, &
                   matrix_c=fm_scaled_dm_occ)

      CALL cp_gemm(transa="N", transb="T", m=nmo, n=nmo, k=nmo, alpha=1.0_dp, &
                   matrix_a=fm_mo_coeff_virt, matrix_b=fm_mo_coeff_virt, beta=0.0_dp, &
                   matrix_c=fm_scaled_dm_virt)

      CALL timestop(handle4)

      CALL timeset(routineN//"_im_time_repl_subgr", handle4)

      ! replicate fm_scaled_dm_occ to every subgroup into sparse matrix mat_M_munu_occ
      CALL replicate_mat_to_subgroup_im_time(mp2_env, para_env, para_env_sub, fm_scaled_dm_occ, nmo, &
                                             mat_munu, mat_M_munu_occ)

      CALL cp_dbcsr_filter(mat_M_munu_occ, eps_filter)

      ! the same for the virtual scaled density matrix
      CALL replicate_mat_to_subgroup_im_time(mp2_env, para_env, para_env_sub, fm_scaled_dm_virt, nmo, &
                                             mat_munu, mat_M_munu_virt)

      CALL cp_dbcsr_filter(mat_M_munu_virt, eps_filter)

      CALL timestop(handle4)

      CALL timeset(routineN//"_mult_3cERI_sc_DM", handle5)

      DO LLL = 1, my_group_L_size

         ! transposed or not does not matter because both matrices being multiplied with each other
         ! are symmetric (other order as with occ, because for the trace we actually need one matrix
         ! transposed (and transposing is easily done in the previous multiplication)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, mat_M_munu_occ, mat_B_munu(LLL)%matrix, &
                                0.0_dp, mat_M_munu_P_occ(LLL)%matrix, filter_eps=eps_filter)

         CALL cp_dbcsr_filter(mat_M_munu_P_occ(LLL)%matrix, eps_filter)

         ! alpha=2.0_dp only for closed-shell (for imag. frequency it is a factor 4.0)
         CALL cp_dbcsr_multiply("T", "N", alpha/2.0_dp, mat_B_munu(LLL)%matrix, mat_M_munu_virt, &
                                0.0_dp, mat_M_munu_P_virt(LLL)%matrix, filter_eps=eps_filter)

         CALL cp_dbcsr_filter(mat_M_munu_P_virt(LLL)%matrix, eps_filter)

      END DO

      ! in case we have different my_group_L_size for different groups, we generate the
      ! the dbcsr matrix format for all mat_M_munu_P_virt matrices

      ! JW should check whether this is really necessary
      IF (my_group_L_size < max_size_RI_group) THEN

         DO LLL = my_group_L_size+1, max_size_RI_group

            CALL cp_dbcsr_copy(mat_M_munu_P_virt(LLL)%matrix, mat_M_munu_P_virt(LLL-1)%matrix)

            CALL cp_dbcsr_set(mat_M_munu_P_virt(LLL)%matrix, 0.0_dp)

            CALL cp_dbcsr_filter(mat_M_munu_P_virt(LLL)%matrix, eps_filter)

         END DO

      END IF

      CALL timestop(handle5)

      trace_2d = 0.0_dp

      CALL dbcsr_3index_trace(trace_2d, mat_M_munu_P_occ, mat_M_munu_P_virt, para_env, para_env_sub, &
                              ngroup, my_group_L_size, my_group_L_start)

      ! fill fm_mat_Q
      CALL fill_fm_with_2d_array(fm_mat_Q, trace_2d, para_env, my_group_L_start, my_group_L_size, &
                                 dimen_RI, ngroup)

      CALL timeset(routineN//"_Fourier_transform", handle5)

      ! Fourier transform of Q(it) to Q(iw) (fm_mat_Q_omega)
      first_cycle_omega_loop = .TRUE.

      DO iquad = 1, num_integ_points

         IF (do_minimax_quad) THEN
            omega = tj(iquad)
            weight = tau_wj_2d(iquad, jquad)
         ELSE
            omega = a_scaling/TAN(tj(iquad))
            weight = wj(jquad)
         END IF

         IF (first_cycle_omega_loop) THEN
            ! no multiplication with 2.0 as in Kresses paper (Kaltak, JCTC 10, 2498 (2014), Eq. 12)
            ! because this factor is already absorbed in the weight w_j
            CALL cp_fm_scale(COS(omega*tau)*weight, fm_mat_Q)
         ELSE
            CALL cp_fm_scale(COS(omega*tau)/COS(omega_old*tau)*weight/weight_old, fm_mat_Q)
         END IF

         CALL cp_fm_scale_and_add(1.0_dp, fm_mat_Q_omega(iquad)%matrix, 1.0_dp, fm_mat_Q)

         first_cycle_omega_loop = .FALSE.

         omega_old = omega
         weight_old = weight

      END DO

      CALL timestop(handle5)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param fm_scaled_dm ...
!> \param nmo ...
!> \param mat_munu ...
!> \param mat_M_munu ...
! **************************************************************************************************
   SUBROUTINE replicate_mat_to_subgroup_im_time(mp2_env, para_env, para_env_sub, fm_scaled_dm, &
                                                nmo, mat_munu, mat_M_munu)
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      TYPE(cp_fm_type), POINTER                          :: fm_scaled_dm
      INTEGER                                            :: nmo
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(cp_dbcsr_type), POINTER                       :: mat_M_munu

      CHARACTER(LEN=*), PARAMETER :: routineN = 'replicate_mat_to_subgroup_im_time', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk, col, col_offset, col_size, handle, i_global, iiB, iproc, itmp(2), j_global, &
         jjB, max_row_col_local, my_mu_end, my_mu_size, my_mu_start, ncol_local, ncol_rec, &
         nrow_local, nrow_rec, proc_receive, proc_receive_static, proc_send, proc_send_static, &
         proc_shift, row, row_offset, row_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_array, proc_map, sizes_array, &
                                                            starts_array
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_col_row_info, rec_col_row_info
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, col_indices_rec, &
                                                            row_indices, row_indices_rec
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: C, rec_C
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_block, local_C, local_C_internal
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_coeff

      CALL timeset(routineN, handle)

      ALLOCATE (sizes_array(0:para_env_sub%num_pe-1))
      ALLOCATE (starts_array(0:para_env_sub%num_pe-1))
      starts_array = 0
      ALLOCATE (ends_array(0:para_env_sub%num_pe-1))
      ends_array = 0

      DO iproc = 0, para_env_sub%num_pe-1
         itmp = get_limit(nmo, para_env_sub%num_pe, iproc)
         starts_array(iproc) = itmp(1)
         ends_array(iproc) = itmp(2)
         sizes_array(iproc) = itmp(2)-itmp(1)+1
      ENDDO

      my_mu_size = sizes_array(para_env_sub%mepos)
      my_mu_start = starts_array(para_env_sub%mepos)
      my_mu_end = ends_array(para_env_sub%mepos)

      ! local storage for the C matrix
      ALLOCATE (C(my_mu_size, nmo))
      C = 0.0_dp

      ! proc_map, vector that replicate the processor numbers also
      ! for negative and positive number > num_pe
      ! needed to know which is the processor, to respect to another one,
      ! for a given shift
      ALLOCATE (proc_map(-para_env%num_pe:2*para_env%num_pe-1))
      DO iiB = 0, para_env%num_pe-1
         proc_map(iiB) = iiB
         proc_map(-iiB-1) = para_env%num_pe-iiB-1
         proc_map(para_env%num_pe+iiB) = iiB
      END DO

      CALL cp_fm_get_info(matrix=fm_scaled_dm, &
                          matrix_struct=fm_struct_coeff, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          local_data=local_C_internal)

      ALLOCATE (local_C(nrow_local, ncol_local))
      local_C = local_C_internal(1:nrow_local, 1:ncol_local)
      NULLIFY (local_C_internal)

      max_row_col_local = MAX(nrow_local, ncol_local)
      CALL mp_max(max_row_col_local, para_env%group)

      ALLOCATE (local_col_row_info(0:max_row_col_local, 2))
      local_col_row_info = 0
      ! 0,1 nrows
      local_col_row_info(0, 1) = nrow_local
      local_col_row_info(1:nrow_local, 1) = row_indices(1:nrow_local)
      ! 0,2 ncols
      local_col_row_info(0, 2) = ncol_local
      local_col_row_info(1:ncol_local, 2) = col_indices(1:ncol_local)

      ALLOCATE (rec_col_row_info(0:max_row_col_local, 2))

      ! accumulate data on C buffer starting from myself
      DO iiB = 1, nrow_local
         i_global = row_indices(iiB)
         IF (i_global >= my_mu_start .AND. i_global <= my_mu_end) THEN
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               C(i_global-my_mu_start+1, j_global) = local_C(iiB, jjB)
            END DO
         END IF
      END DO

      ! start ring communication for collecting the data from the other
      proc_send_static = proc_map(para_env%mepos+1)
      proc_receive_static = proc_map(para_env%mepos-1)
      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         proc_receive = proc_map(para_env%mepos-proc_shift)

         ! first exchange information on the local data
         rec_col_row_info = 0
         CALL mp_sendrecv(local_col_row_info, proc_send_static, rec_col_row_info, proc_receive_static, para_env%group)
         nrow_rec = rec_col_row_info(0, 1)
         ncol_rec = rec_col_row_info(0, 2)

         ALLOCATE (row_indices_rec(nrow_rec))
         row_indices_rec = rec_col_row_info(1:nrow_rec, 1)

         ALLOCATE (col_indices_rec(ncol_rec))
         col_indices_rec = rec_col_row_info(1:ncol_rec, 2)

         ALLOCATE (rec_C(nrow_rec, ncol_rec))
         rec_C = 0.0_dp

         ! then send and receive the real data
         CALL mp_sendrecv(local_C, proc_send_static, rec_C, proc_receive_static, para_env%group)

         ! accumulate the received data on C buffer
         DO iiB = 1, nrow_rec
            i_global = row_indices_rec(iiB)
            IF (i_global >= my_mu_start .AND. i_global <= my_mu_end) THEN
               DO jjB = 1, ncol_rec
                  j_global = col_indices_rec(jjB)
                  C(i_global-my_mu_start+1, j_global) = rec_C(iiB, jjB)
               END DO
            END IF
         END DO

         local_col_row_info(:, :) = rec_col_row_info
         DEALLOCATE (local_C)
         ALLOCATE (local_C(nrow_rec, ncol_rec))
         local_C = rec_C

         DEALLOCATE (col_indices_rec)
         DEALLOCATE (row_indices_rec)
         DEALLOCATE (rec_C)
      END DO

      DEALLOCATE (local_col_row_info)
      DEALLOCATE (rec_col_row_info)
      DEALLOCATE (proc_map)

      ! proc_map, for the sub_group
      ALLOCATE (proc_map(-para_env_sub%num_pe:2*para_env_sub%num_pe-1))
      DO iiB = 0, para_env_sub%num_pe-1
         proc_map(iiB) = iiB
         proc_map(-iiB-1) = para_env_sub%num_pe-iiB-1
         proc_map(para_env_sub%num_pe+iiB) = iiB
      END DO

      ! all levels
      CALL build_mat_M_munu(mp2_env, para_env_sub, mat_M_munu, C, &
                            nmo, blk, row, col, row_size, col_size, row_offset, &
                            col_offset, i_global, j_global, my_mu_start, my_mu_end, &
                            mat_munu, iter, &
                            data_block, ends_array, proc_map, &
                            sizes_array, starts_array)

      DEALLOCATE (proc_map)
      DEALLOCATE (sizes_array)
      DEALLOCATE (starts_array)
      DEALLOCATE (ends_array)
      DEALLOCATE (local_C)

      CALL timestop(handle)

   END SUBROUTINE replicate_mat_to_subgroup_im_time

! **************************************************************************************************
!> \brief fill dbcsr matrix mat_M_munu
!> \param mp2_env ...
!> \param para_env_sub ...
!> \param mat_M_munu ...
!> \param Cread ...
!> \param number_of_level ...
!> \param blk ...
!> \param row ...
!> \param col ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
!> \param i_global ...
!> \param j_global ...
!> \param my_mu_start ...
!> \param my_mu_end ...
!> \param mat_munu ...
!> \param iter ...
!> \param data_block ...
!> \param ends_array ...
!> \param proc_map ...
!> \param sizes_array ...
!> \param starts_array ...
!> \author Jan Wilhelm
! **************************************************************************************************
   SUBROUTINE build_mat_M_munu(mp2_env, para_env_sub, mat_M_munu, Cread, &
                               number_of_level, blk, row, col, row_size, col_size, row_offset, &
                               col_offset, i_global, j_global, my_mu_start, my_mu_end, &
                               mat_munu, iter, &
                               data_block, ends_array, proc_map, &
                               sizes_array, starts_array)
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      TYPE(cp_dbcsr_type), POINTER                       :: mat_M_munu
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Cread
      INTEGER                                            :: number_of_level, blk, row, col, &
                                                            row_size, col_size, row_offset, &
                                                            col_offset, i_global, j_global, &
                                                            my_mu_start, my_mu_end
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(cp_dbcsr_iterator)                            :: iter
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_block
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_array, proc_map, sizes_array, &
                                                            starts_array

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_mat_M_munu', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, proc_receive, proc_send, &
                                                            proc_shift, rec_mu_end, rec_mu_size, &
                                                            rec_mu_start
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rec_C

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_copy(mat_M_munu, mat_munu%matrix)
      CALL cp_dbcsr_set(mat_M_munu, 0.0_dp)

      ! accumulate data on mat_M_munu starting from myself
      CALL cp_dbcsr_iterator_start(iter, mat_M_munu)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                           row_size=row_size, col_size=col_size, &
                                           row_offset=row_offset, col_offset=col_offset)
         DO i = 1, row_size
            i_global = row_offset+i-1
            IF (i_global >= my_mu_start .AND. i_global <= my_mu_end) THEN
               DO j = 1, col_size
                  j_global = col_offset+j-1
                  data_block(i, j) = Cread(i_global-my_mu_start+1, col_offset+j-1)
               ENDDO
            END IF
         ENDDO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      ! start ring communication in the subgroup for collecting the data from the other
      ! proc (occupied)
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = proc_map(para_env_sub%mepos+proc_shift)
         proc_receive = proc_map(para_env_sub%mepos-proc_shift)

         rec_mu_start = starts_array(proc_receive)
         rec_mu_end = ends_array(proc_receive)
         rec_mu_size = sizes_array(proc_receive)

         ALLOCATE (rec_C(rec_mu_size, number_of_level))
         rec_C = 0.0_dp

         ! then send and receive the real data
         CALL mp_sendrecv(Cread, proc_send, rec_C, proc_receive, para_env_sub%group)

         ! accumulate data on mat_M_munu the data received from proc_rec
         CALL cp_dbcsr_iterator_start(iter, mat_M_munu)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                              row_size=row_size, col_size=col_size, &
                                              row_offset=row_offset, col_offset=col_offset)
            DO i = 1, row_size
               i_global = row_offset+i-1
               IF (i_global >= rec_mu_start .AND. i_global <= rec_mu_end) THEN
                  DO j = 1, col_size
                     j_global = col_offset+j-1
                     data_block(i, j) = rec_C(i_global-rec_mu_start+1, col_offset+j-1)
                  ENDDO
               END IF
            ENDDO
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)

         DEALLOCATE (rec_C)

      END DO
      CALL cp_dbcsr_filter(mat_M_munu, mp2_env%mp2_gpw%eps_filter)

      DEALLOCATE (Cread)

      CALL timestop(handle)

   END SUBROUTINE build_mat_M_munu

! **************************************************************************************************
!> \brief dbcsr_3index_trace
!> \param trace_2d ...
!> \param mat_occ ...
!> \param mat_virt ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param ngroup ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
! **************************************************************************************************
   SUBROUTINE dbcsr_3index_trace(trace_2d, mat_occ, mat_virt, para_env, para_env_sub, &
                                 ngroup, my_group_L_size, my_group_L_start)

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: trace_2d
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_occ, mat_virt
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER                                            :: ngroup, my_group_L_size, my_group_L_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_3index_trace', &
         routineP = moduleN//':'//routineN

      INTEGER :: col, col_occ, col_size, col_virt, comm_exch, exch_color, group_iter, handle, &
         handle2, i_block, i_block_occ, i_block_virt, KKK, LLL, my_group_K_start, &
         n_blocks_local_occ, n_blocks_local_virt, n_entries_local_occ, n_entries_local_virt, &
         offset, offset_occ, offset_virt, proc_rec_static, proc_send_static, row, row_occ, &
         row_size, row_virt, size_occ, size_virt
      INTEGER, DIMENSION(:), POINTER                     :: proc_rec, proc_send
      INTEGER, DIMENSION(:, :), POINTER                  :: buffer_info_occ, buffer_info_virt, &
                                                            req_array
      REAL(KIND=dp), DIMENSION(:), POINTER               :: buffer_data_occ, buffer_data_virt
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_block
      REAL(KIND=dp), EXTERNAL                            :: DDOT
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_para_env_type), POINTER                    :: para_env_exch
      TYPE(integ_mat_buffer_type), DIMENSION(:), POINTER :: buffer_rec
      TYPE(integ_mat_buffer_type), POINTER               :: buffer_send

      CALL timeset(routineN, handle)

      NULLIFY (data_block, para_env_exch)

      n_blocks_local_occ = 0
      n_blocks_local_virt = 0

      ! get number of local blocks of mat_M_occ and mat_M_virt
      DO LLL = 1, my_group_L_size

         n_blocks_local_occ = n_blocks_local_occ+cp_dbcsr_get_num_blocks(mat_occ(LLL)%matrix)
         n_blocks_local_virt = n_blocks_local_virt+cp_dbcsr_get_num_blocks(mat_virt(LLL)%matrix)

      END DO

      ! get number of entries in mat_M_occ and mat_M_virt
      n_entries_local_occ = 0
      n_entries_local_virt = 0

      DO LLL = 1, my_group_L_size

         ! n_entries_local_occ
         CALL cp_dbcsr_iterator_start(iter, mat_occ(LLL)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, &
                                              row_size=row_size, col_size=col_size)

            n_entries_local_occ = n_entries_local_occ+row_size*col_size

         END DO
         CALL cp_dbcsr_iterator_stop(iter)

         ! n_entries_local_virt
         CALL cp_dbcsr_iterator_start(iter, mat_virt(LLL)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, &
                                              row_size=row_size, col_size=col_size)

            n_entries_local_virt = n_entries_local_virt+row_size*col_size

         END DO
         CALL cp_dbcsr_iterator_stop(iter)

      END DO

      ALLOCATE (buffer_info_occ(n_blocks_local_occ, 5))

      ALLOCATE (buffer_data_occ(n_entries_local_occ))

      ! 1) fill buffer_data_occ with blocks from mat_occ(LLL)%matrix
      ! 2) fill block_info_occ with LLL-counter (1), row/col in original dbcsr matrix (2/3),
      !    block_size = row_size*col_size (4) and offset in buffer (5)

      offset = 0
      i_block = 1

      DO LLL = 1, my_group_L_size
         CALL cp_dbcsr_iterator_start(iter, mat_occ(LLL)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, &
                                              row_size=row_size, col_size=col_size)

            ! write 1D array of data_block into buffer_data_occ
            buffer_data_occ(offset+1:offset+row_size*col_size) = RESHAPE(data_block, (/row_size*col_size/))

            buffer_info_occ(i_block, 1) = LLL
            buffer_info_occ(i_block, 2) = row
            buffer_info_occ(i_block, 3) = col
            buffer_info_occ(i_block, 4) = row_size*col_size
            buffer_info_occ(i_block, 5) = offset

            offset = offset+row_size*col_size
            i_block = i_block+1

         END DO ! dbcsr iterator
         CALL cp_dbcsr_iterator_stop(iter)

         CALL cp_dbcsr_set(mat_occ(LLL)%matrix, 0.0_dp)
         ! deallocate the blocks in mat_occ
         CALL cp_dbcsr_filter(mat_occ(LLL)%matrix, 1.0_dp)

      END DO ! LLL

      ! the same for the virtuals
      ALLOCATE (buffer_info_virt(n_blocks_local_virt, 6))
      ALLOCATE (buffer_data_virt(n_entries_local_virt))

      offset = 0
      i_block = 1

      DO LLL = 1, my_group_L_size
         CALL cp_dbcsr_iterator_start(iter, mat_virt(LLL)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, &
                                              row_size=row_size, col_size=col_size)

            ! write 1D array of data_block into buffer_data_virt
            buffer_data_virt(offset+1:offset+row_size*col_size) = RESHAPE(data_block, (/row_size*col_size/))

            buffer_info_virt(i_block, 1) = LLL
            buffer_info_virt(i_block, 2) = row
            buffer_info_virt(i_block, 3) = col
            buffer_info_virt(i_block, 4) = row_size*col_size
            buffer_info_virt(i_block, 5) = offset
            buffer_info_virt(i_block, 6) = my_group_L_start

            offset = offset+row_size*col_size
            i_block = i_block+1

         END DO ! dbcsr iterator
         CALL cp_dbcsr_iterator_stop(iter)

         CALL cp_dbcsr_set(mat_virt(LLL)%matrix, 0.0_dp)
         ! deallocate the blocks in mat_virt
         CALL cp_dbcsr_filter(mat_virt(LLL)%matrix, 1.0_dp)

      END DO ! LLL

      CALL timeset(routineN//"_local_DDOT", handle2)

      ! check locally which data is there to multiply
      ! JW write a subroutine for that!!!
      DO i_block_virt = 1, n_blocks_local_virt

         KKK = buffer_info_virt(i_block_virt, 1)
         row_virt = buffer_info_virt(i_block_virt, 2)
         col_virt = buffer_info_virt(i_block_virt, 3)
         size_virt = buffer_info_virt(i_block_virt, 4)
         offset_virt = buffer_info_virt(i_block_virt, 5)
         my_group_K_start = buffer_info_virt(i_block_virt, 6)

         DO i_block_occ = 1, n_blocks_local_occ

            row_occ = buffer_info_occ(i_block_occ, 2)
            IF (row_occ /= row_virt) CYCLE

            col_occ = buffer_info_occ(i_block_occ, 3)
            IF (col_occ /= col_virt) CYCLE

            LLL = buffer_info_occ(i_block_occ, 1)
            size_occ = buffer_info_occ(i_block_occ, 4)
            offset_occ = buffer_info_occ(i_block_occ, 5)

            IF(.NOT.(size_occ == size_virt))CALL cp__a("rpa_im_time.F",806)

            trace_2d(LLL, KKK+my_group_K_start-1) = trace_2d(LLL, KKK+my_group_K_start-1)+ &
                                                    DDOT(size_occ, buffer_data_occ(offset_occ+1:offset_occ+size_occ), 1, &
                                                         buffer_data_virt(offset_virt+1:offset_virt+size_virt), 1)

         END DO ! i_block_occ

      END DO ! i_block_virt

      CALL timestop(handle2)

      ! Get process, from which we get and to whom we send the blocks
      exch_color = MODULO(para_env%mepos, para_env%num_pe/ngroup)
      CALL mp_comm_split_direct(para_env%group, comm_exch, exch_color)
      NULLIFY (para_env_exch)
      CALL cp_para_env_create(para_env_exch, comm_exch)

      proc_rec_static = para_env_exch%mepos+1
      IF (proc_rec_static >= para_env_exch%num_pe) proc_rec_static = proc_rec_static-para_env_exch%num_pe

      proc_send_static = para_env_exch%mepos-1
      IF (proc_send_static < 0) proc_send_static = proc_send_static+para_env_exch%num_pe

      ALLOCATE (proc_rec(ngroup-1))
      ALLOCATE (proc_send(ngroup-1))

      DO group_iter = 1, ngroup-1

         proc_rec(group_iter) = para_env_exch%mepos+group_iter
         IF (proc_rec(group_iter) >= para_env_exch%num_pe) THEN
            proc_rec(group_iter) = proc_rec(group_iter)-para_env_exch%num_pe
         END IF

         proc_send(group_iter) = para_env_exch%mepos-group_iter
         IF (proc_send(group_iter) < 0) THEN
            proc_send(group_iter) = proc_send(group_iter)+para_env_exch%num_pe
         END IF

      END DO

      CALL timeset(routineN//"_comm_1", handle2)

      ALLOCATE (req_array(ngroup-1, 2))
      ALLOCATE (buffer_rec(ngroup-1))
      ALLOCATE (buffer_send)

      DO group_iter = 1, ngroup-1
         ALLOCATE (buffer_rec(group_iter)%sizes(2))
      END DO
      ALLOCATE (buffer_send%sizes(2))

      ! get and send number of local virtual blocks in advance with  non-blocking communication
      buffer_send%sizes(1) = n_blocks_local_virt
      buffer_send%sizes(2) = n_entries_local_virt

      DO group_iter = 1, ngroup-1
         CALL mp_irecv(buffer_rec(group_iter)%sizes, proc_rec(group_iter), para_env_exch%group, &
                       req_array(group_iter, 2), tag=group_iter)
         CALL mp_isend(buffer_send%sizes, proc_send(group_iter), para_env_exch%group, req_array(group_iter, 1), tag=group_iter)
      END DO
      CALL mp_waitall(req_array(:, :))

      CALL timestop(handle2)

      DO group_iter = 1, ngroup-1

         ALLOCATE (buffer_rec(group_iter)%indx(buffer_rec(group_iter)%sizes(1), 6))
         ALLOCATE (buffer_rec(group_iter)%msg(buffer_rec(group_iter)%sizes(2)))

         ! get infos
         CALL timeset(routineN//"_comm_2", handle2)
         CALL mp_irecv(buffer_rec(group_iter)%indx, proc_rec(group_iter), para_env_exch%group, &
                       req_array(group_iter, 2), tag=group_iter)
         CALL mp_isend(buffer_info_virt, proc_send(group_iter), para_env_exch%group, req_array(group_iter, 1), tag=group_iter)
         CALL timestop(handle2)

         ! wait (makes it blocking send/recv)
         CALL timeset(routineN//"_wait", handle2)
         CALL mp_waitall(req_array(group_iter, :))
         CALL timestop(handle2)

         ! get actual data
         CALL timeset(routineN//"_comm_3", handle2)
         CALL mp_irecv(buffer_rec(group_iter)%msg, proc_rec(group_iter), para_env_exch%group, &
                       req_array(group_iter, 2), tag=group_iter)
         CALL mp_isend(buffer_data_virt, proc_send(group_iter), para_env_exch%group, req_array(group_iter, 1), tag=group_iter)
         CALL timestop(handle2)

         ! wait (makes it blocking send/recv)
         CALL timeset(routineN//"_wait", handle2)
         CALL mp_waitall(req_array(group_iter, :))
         CALL timestop(handle2)

         CALL timeset(routineN//"_DDOT", handle2)

         ! perform local multiplications
         DO i_block_virt = 1, buffer_rec(group_iter)%sizes(1)

            KKK = buffer_rec(group_iter)%indx(i_block_virt, 1)
            row_virt = buffer_rec(group_iter)%indx(i_block_virt, 2)
            col_virt = buffer_rec(group_iter)%indx(i_block_virt, 3)
            size_virt = buffer_rec(group_iter)%indx(i_block_virt, 4)
            offset_virt = buffer_rec(group_iter)%indx(i_block_virt, 5)
            my_group_K_start = buffer_rec(group_iter)%indx(i_block_virt, 6)

            DO i_block_occ = 1, n_blocks_local_occ

               row_occ = buffer_info_occ(i_block_occ, 2)
               IF (row_occ /= row_virt) CYCLE

               col_occ = buffer_info_occ(i_block_occ, 3)
               IF (col_occ /= col_virt) CYCLE

               LLL = buffer_info_occ(i_block_occ, 1)
               size_occ = buffer_info_occ(i_block_occ, 4)
               offset_occ = buffer_info_occ(i_block_occ, 5)

               IF(.NOT.(size_occ == size_virt))CALL cp__a("rpa_im_time.F",924)

               trace_2d(LLL, KKK+my_group_K_start-1) = trace_2d(LLL, KKK+my_group_K_start-1)+ &
                                                       DDOT(size_occ, buffer_data_occ(offset_occ+1:offset_occ+size_occ), 1, &
                                                            buffer_rec(group_iter)%msg(offset_virt+1:offset_virt+size_virt), 1)
            END DO ! i_block_occ

         END DO ! i_block_virt

         DEALLOCATE (buffer_rec(group_iter)%indx)

         DEALLOCATE (buffer_rec(group_iter)%msg)

         CALL timestop(handle2)

      END DO ! communication in groups

      CALL timeset(routineN//"_comm_trace", handle2)

      ! para_env_sub is para env of the underlying dbcsr matrix
      CALL mp_sum(trace_2d(:, :), para_env_sub%group)

      CALL timestop(handle2)

      CALL timeset(routineN//"_sync", handle2)

      CALL mp_sync(para_env_exch%group)

      CALL timestop(handle2)

      CALL cp_para_env_release(para_env_exch)

      DEALLOCATE (buffer_info_occ, buffer_info_virt)
      DEALLOCATE (buffer_data_occ, buffer_data_virt)
      DEALLOCATE (proc_send, proc_rec)
      DEALLOCATE (req_array)
      DO group_iter = 1, ngroup-1
         DEALLOCATE (buffer_rec(group_iter)%sizes)
      END DO
      DEALLOCATE (buffer_send%sizes)
      DEALLOCATE (buffer_rec, buffer_send)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief fill_fm_with_2d_array
!> \param fm_mat_Q ...
!> \param trace_2d ...
!> \param para_env ...
!> \param my_group_L_start ...
!> \param my_group_L_size ...
!> \param dimen_RI ...
!> \param ngroup ...
! **************************************************************************************************
   SUBROUTINE fill_fm_with_2d_array(fm_mat_Q, trace_2d, para_env, my_group_L_start, my_group_L_size, &
                                    dimen_RI, ngroup)

      TYPE(cp_fm_type), POINTER                          :: fm_mat_Q
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: trace_2d
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: my_group_L_start, my_group_L_size, &
                                                            dimen_RI, ngroup

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fill_fm_with_2d_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: igroup, my_group_K_size, &
                                                            my_group_K_start, proc_rec_M, &
                                                            proc_send_M
      INTEGER, DIMENSION(:), POINTER                     :: my_group_K_rec, my_group_K_send
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: trace_2d_rec, trace_2d_send

! set the local data

      CALL cp_fm_set_submatrix(fm=fm_mat_Q, new_values=trace_2d, start_row=my_group_L_start, start_col=1, &
                               n_rows=my_group_L_size, n_cols=dimen_RI)

      ! Get process, from which we get and to whom we send trace2d

      proc_send_M = para_env%mepos+para_env%num_pe/ngroup

      IF (proc_send_M >= para_env%num_pe) THEN
         proc_send_M = proc_send_M-para_env%num_pe
      END IF

      proc_rec_M = para_env%mepos-para_env%num_pe/ngroup

      IF (proc_rec_M < 0) THEN
         proc_rec_M = proc_rec_M+para_env%num_pe
      END IF

      ALLOCATE (my_group_K_send(2))
      ALLOCATE (my_group_K_rec(2))
      my_group_K_send(1) = my_group_L_start
      my_group_K_send(2) = my_group_L_size

      ALLOCATE (trace_2d_send(my_group_L_size, dimen_RI))

      trace_2d_send = trace_2d

      DO igroup = 1, ngroup

         ! send and receive my_group_K_start and my_group_K_size
         CALL mp_sendrecv(my_group_K_send, proc_send_M, my_group_K_rec, proc_rec_M, para_env%group)

         my_group_K_start = my_group_K_rec(1)
         my_group_K_size = my_group_K_rec(2)

         ALLOCATE (trace_2d_rec(my_group_K_size, dimen_RI))

         CALL mp_sendrecv(trace_2d_send, proc_send_M, trace_2d_rec, proc_rec_M, para_env%group)

         ! send and receive trace_2d
         CALL cp_fm_set_submatrix(fm=fm_mat_Q, new_values=trace_2d_rec, start_row=my_group_K_start, start_col=1, &
                                  n_rows=my_group_K_size, n_cols=dimen_RI)

         DEALLOCATE (trace_2d_send)

         ALLOCATE (trace_2d_send(my_group_K_size, dimen_RI))

         trace_2d_send = trace_2d_rec

         DEALLOCATE (trace_2d_rec)

         my_group_K_send(1) = my_group_K_start
         my_group_K_send(2) = my_group_K_size

      END DO

      DEALLOCATE (trace_2d_send)

      DEALLOCATE (my_group_K_send, my_group_K_rec)

   END SUBROUTINE fill_fm_with_2d_array

END MODULE rpa_im_time
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Auxiliary rountines necessary to redistribute an fm_matrix from a
!>        given blacs_env to another
!> \par History
!>      12.2012 created [Mauro Del Ben]
! **************************************************************************************************
MODULE rpa_communication
   USE cp_fm_types,                     ONLY: cp_fm_get_info,&
                                              cp_fm_indxg2l,&
                                              cp_fm_indxg2p,&
                                              cp_fm_indxl2g,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_irecv,&
                                              mp_isend,&
                                              mp_sum,&
                                              mp_wait,&
                                              mp_waitall
   USE mp2_types,                       ONLY: integ_mat_buffer_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rpa_communication'

   PUBLIC :: initialize_buffer, &
             fm_redistribute, &
             release_buffer

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param fm_mat_source ...
!> \param fm_mat_dest ...
!> \param RPA_proc_map ...
!> \param buffer_rec ...
!> \param buffer_send ...
!> \param number_of_rec ...
!> \param number_of_send ...
!> \param map_send_size ...
!> \param map_rec_size ...
!> \param local_size_source ...
!> \param para_env_RPA ...
! **************************************************************************************************
   SUBROUTINE initialize_buffer(fm_mat_source, fm_mat_dest, RPA_proc_map, buffer_rec, buffer_send, &
                                number_of_rec, number_of_send, &
                                map_send_size, map_rec_size, local_size_source, para_env_RPA)
      TYPE(cp_fm_type), POINTER                          :: fm_mat_source, fm_mat_dest
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: RPA_proc_map
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_send
      INTEGER                                            :: number_of_rec, number_of_send
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: map_send_size, map_rec_size
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_size_source
      TYPE(cp_para_env_type), POINTER                    :: para_env_RPA

      CHARACTER(LEN=*), PARAMETER :: routineN = 'initialize_buffer', &
         routineP = moduleN//':'//routineN

      INTEGER :: dummy_proc, handle, handle2, i, i_global, i_local, iiB, iii, j_global, j_local, &
         jjB, mypcol_d, mypcol_s, myprow_d, myprow_s, ncol_block_d, ncol_block_s, ncol_local_d, &
         ncol_local_s, npcol_d, npcol_s, nprow_d, nprow_s, nrow_block_d, nrow_block_s, &
         nrow_local_d, nrow_local_s, proc_receive, proc_send, proc_shift, rec_counter, &
         rec_local_col, rec_local_row, rec_pcol, rec_prow, ref_rec_pcol, ref_rec_prow, &
         send_counter, send_pcol, send_prow, size_rec_buffer, size_send_buffer
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: index_counter, proc2counter
      INTEGER, DIMENSION(:), POINTER                     :: col_indices_d, col_indices_s, &
                                                            row_indices_d, row_indices_s

      CALL timeset(routineN, handle)

      ! create the RPA proc_map
      IF (.NOT. (ALLOCATED(RPA_proc_map))) THEN
         ALLOCATE (RPA_proc_map(-para_env_RPA%num_pe:2*para_env_RPA%num_pe-1))
         RPA_proc_map = 0
         DO i = 0, para_env_RPA%num_pe-1
            RPA_proc_map(i) = i
            RPA_proc_map(-i-1) = para_env_RPA%num_pe-i-1
            RPA_proc_map(para_env_RPA%num_pe+i) = i
         END DO
      END IF

      ! get info for the source
      CALL cp_fm_get_info(matrix=fm_mat_source, &
                          nrow_local=nrow_local_s, &
                          ncol_local=ncol_local_s, &
                          row_indices=row_indices_s, &
                          col_indices=col_indices_s, &
                          nrow_block=nrow_block_s, &
                          ncol_block=ncol_block_s)
      myprow_s = fm_mat_source%matrix_struct%context%mepos(1)
      mypcol_s = fm_mat_source%matrix_struct%context%mepos(2)
      nprow_s = fm_mat_source%matrix_struct%context%num_pe(1)
      npcol_s = fm_mat_source%matrix_struct%context%num_pe(2)

      ! get info for the dest
      CALL cp_fm_get_info(matrix=fm_mat_dest, &
                          nrow_local=nrow_local_d, &
                          ncol_local=ncol_local_d, &
                          row_indices=row_indices_d, &
                          col_indices=col_indices_d, &
                          nrow_block=nrow_block_d, &
                          ncol_block=ncol_block_d)
      myprow_d = fm_mat_dest%matrix_struct%context%mepos(1)
      mypcol_d = fm_mat_dest%matrix_struct%context%mepos(2)
      nprow_d = fm_mat_dest%matrix_struct%context%num_pe(1)
      npcol_d = fm_mat_dest%matrix_struct%context%num_pe(2)

      ! 0) create the map for the local sizes
      ALLOCATE (local_size_source(2, 0:para_env_RPA%num_pe-1))
      local_size_source = 0
      local_size_source(1, para_env_RPA%mepos) = nrow_local_s
      local_size_source(2, para_env_RPA%mepos) = ncol_local_s
      CALL mp_sum(local_size_source, para_env_RPA%group)

      ! 1) loop over my local data and define a map for the proc to send data
      ALLOCATE (map_send_size(0:para_env_RPA%num_pe-1))
      map_send_size = 0
      DO jjB = 1, ncol_local_s
         j_global = col_indices_s(jjB)
         send_pcol = cp_fm_indxg2p(j_global, ncol_block_d, dummy_proc, &
                                   fm_mat_dest%matrix_struct%first_p_pos(2), npcol_d)
         DO iiB = 1, nrow_local_s
            i_global = row_indices_s(iiB)
            send_prow = cp_fm_indxg2p(i_global, nrow_block_d, dummy_proc, &
                                      fm_mat_dest%matrix_struct%first_p_pos(1), nprow_d)
            proc_send = fm_mat_dest%matrix_struct%context%blacs2mpi(send_prow, send_pcol)
            map_send_size(proc_send) = map_send_size(proc_send)+1
         END DO
      END DO

      ! 2) loop over my local data of fm_mat_S and define a map for the proc from which rec data
      ALLOCATE (map_rec_size(0:para_env_RPA%num_pe-1))
      map_rec_size = 0
      DO jjB = 1, ncol_local_d
         j_global = col_indices_d(jjB)
         rec_pcol = cp_fm_indxg2p(j_global, ncol_block_s, dummy_proc, &
                                  fm_mat_source%matrix_struct%first_p_pos(2), npcol_s)
         DO iiB = 1, nrow_local_d
            i_global = row_indices_d(iiB)
            rec_prow = cp_fm_indxg2p(i_global, nrow_block_s, dummy_proc, &
                                     fm_mat_source%matrix_struct%first_p_pos(1), nprow_s)
            proc_receive = fm_mat_source%matrix_struct%context%blacs2mpi(rec_prow, rec_pcol)
            map_rec_size(proc_receive) = map_rec_size(proc_receive)+1
         END DO
      END DO

      ! 3) calculate the number of messages to send and allocate the send buffer
      number_of_send = 0
      DO proc_shift = 1, para_env_RPA%num_pe-1
         proc_send = RPA_proc_map(para_env_RPA%mepos+proc_shift)
         IF (map_send_size(proc_send) > 0) THEN
            number_of_send = number_of_send+1
         END IF
      END DO

      ALLOCATE (buffer_send(number_of_send))

      ! 3.5) prepare the index map
      CALL timeset(routineN//"_bS", handle2)
      ALLOCATE (proc2counter(0:para_env_RPA%num_pe-1))
      proc2counter = 0
      ! allocate buffer for sending
      send_counter = 0
      DO proc_shift = 1, para_env_RPA%num_pe-1
         proc_send = RPA_proc_map(para_env_RPA%mepos+proc_shift)
         size_send_buffer = map_send_size(proc_send)

         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            ! prepare the sending buffer
            ALLOCATE (buffer_send(send_counter)%indx(2, size_send_buffer))
            buffer_send(send_counter)%indx = 0

            proc2counter(proc_send) = send_counter
         END IF
      END DO

      ALLOCATE (index_counter(0:para_env_RPA%num_pe-1))
      index_counter = 0
      DO iiB = 1, nrow_local_s
         i_global = row_indices_s(iiB)
         send_prow = cp_fm_indxg2p(i_global, nrow_block_d, dummy_proc, &
                                   fm_mat_dest%matrix_struct%first_p_pos(1), nprow_d)
         DO jjB = 1, ncol_local_s
            j_global = col_indices_s(jjB)
            send_pcol = cp_fm_indxg2p(j_global, ncol_block_d, dummy_proc, &
                                      fm_mat_dest%matrix_struct%first_p_pos(2), npcol_d)
            iii = fm_mat_dest%matrix_struct%context%blacs2mpi(send_prow, send_pcol)
            IF (iii == para_env_RPA%mepos) CYCLE
            index_counter(iii) = index_counter(iii)+1
            send_counter = proc2counter(iii)
            buffer_send(send_counter)%indx(1, index_counter(iii)) = iiB
            buffer_send(send_counter)%indx(2, index_counter(iii)) = jjB
            ! buffer_send(send_counter)%msg(index_counter(iii))=fm_mat_source%local_data(iiB,jjB)
         END DO
      END DO

      DEALLOCATE (index_counter)
      CALL timestop(handle2)

      ! 4) calculate the number of messages to receive and allocate the rec buffer
      number_of_rec = 0
      DO proc_shift = 1, para_env_RPA%num_pe-1
         proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)
         IF (map_rec_size(proc_receive) > 0) THEN
            number_of_rec = number_of_rec+1
         END IF
      END DO

      ALLOCATE (buffer_rec(number_of_rec))

      ! 4.5) prepare the index map
      CALL timeset(routineN//"_bR", handle2)
      rec_counter = 0
      proc2counter = 0
      DO proc_shift = 1, para_env_RPA%num_pe-1
         proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)

         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1
            ! allocate the auxilliary index structure
            ALLOCATE (buffer_rec(rec_counter)%indx(2, size_rec_buffer))
            buffer_rec(rec_counter)%indx = 0

            proc2counter(proc_receive) = rec_counter
         END IF
      END DO

!$OMP     PARALLEL DO DEFAULT(NONE) SHARED(para_env_RPA,RPA_proc_map,map_rec_size,&
!$OMP                                         local_size_source,fm_mat_source,fm_mat_dest,&
!$OMP                                         nrow_block_s,nprow_s,nrow_block_d,nprow_d,dummy_proc,&
!$OMP                                         ncol_block_s,npcol_s,ncol_block_d,npcol_d,&
!$OMP                                         myprow_d,mypcol_d,&
!$OMP                                         buffer_rec,proc2counter)&
!$OMP     PRIVATE(proc_receive,size_rec_buffer,rec_local_row,rec_local_col,ref_rec_prow,ref_rec_pcol,&
!$OMP             rec_counter,iii,i_global,rec_prow,i_local,j_global,rec_pcol,j_local)
      DO proc_shift = 1, para_env_RPA%num_pe-1
         proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)

         rec_local_row = local_size_source(1, proc_receive)
         rec_local_col = local_size_source(2, proc_receive)

         ref_rec_prow = fm_mat_source%matrix_struct%context%mpi2blacs(1, proc_receive)
         ref_rec_pcol = fm_mat_source%matrix_struct%context%mpi2blacs(2, proc_receive)

         IF (map_rec_size(proc_receive) > 0) THEN

            rec_counter = proc2counter(proc_receive)
            iii = 0
            DO iiB = 1, rec_local_row
               i_global = cp_fm_indxl2g(iiB, nrow_block_s, ref_rec_prow, &
                                        fm_mat_source%matrix_struct%first_p_pos(1), nprow_s)
               rec_prow = cp_fm_indxg2p(i_global, nrow_block_d, dummy_proc, &
                                        fm_mat_dest%matrix_struct%first_p_pos(1), nprow_d)
               IF (rec_prow == myprow_d) THEN
                  i_local = cp_fm_indxg2l(i_global, nrow_block_d, dummy_proc, &
                                          fm_mat_dest%matrix_struct%first_p_pos(1), nprow_d)
                  DO jjB = 1, rec_local_col
                     j_global = cp_fm_indxl2g(jjB, ncol_block_s, ref_rec_pcol, &
                                              fm_mat_source%matrix_struct%first_p_pos(2), npcol_s)
                     rec_pcol = cp_fm_indxg2p(j_global, ncol_block_d, dummy_proc, &
                                              fm_mat_dest%matrix_struct%first_p_pos(2), npcol_d)
                     IF (rec_pcol == mypcol_d) THEN
                        j_local = cp_fm_indxg2l(j_global, ncol_block_d, dummy_proc, &
                                                fm_mat_dest%matrix_struct%first_p_pos(2), npcol_d)

                        iii = iii+1
                        buffer_rec(rec_counter)%indx(1, iii) = i_local
                        buffer_rec(rec_counter)%indx(2, iii) = j_local
                     END IF
                  END DO
               END IF
            END DO

         END IF
      END DO
!$OMP     END PARALLEL DO

      DEALLOCATE (proc2counter)
      CALL timestop(handle2)

      CALL timestop(handle)

   END SUBROUTINE initialize_buffer

! **************************************************************************************************
!> \brief ...
!> \param fm_mat_source ...
!> \param fm_mat_dest ...
!> \param RPA_proc_map ...
!> \param buffer_rec ...
!> \param buffer_send ...
!> \param number_of_send ...
!> \param map_send_size ...
!> \param map_rec_size ...
!> \param local_size_source ...
!> \param para_env_RPA ...
! **************************************************************************************************
   SUBROUTINE fm_redistribute(fm_mat_source, fm_mat_dest, RPA_proc_map, buffer_rec, buffer_send, &
                              number_of_send, &
                              map_send_size, map_rec_size, local_size_source, para_env_RPA)
      TYPE(cp_fm_type), POINTER                          :: fm_mat_source, fm_mat_dest
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: RPA_proc_map
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_send
      INTEGER                                            :: number_of_send
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: map_send_size, map_rec_size
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_size_source
      TYPE(cp_para_env_type), POINTER                    :: para_env_RPA

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fm_redistribute', &
         routineP = moduleN//':'//routineN

      INTEGER :: dummy_proc, handle, handle2, i_global, i_local, iiB, iii, j_global, j_local, jjB, &
         mypcol_d, mypcol_s, myprow_d, myprow_s, ncol_block_d, ncol_block_s, ncol_local_d, &
         ncol_local_s, npcol_d, npcol_s, nprow_d, nprow_s, nrow_block_d, nrow_block_s, &
         nrow_local_d, nrow_local_s, proc_receive, proc_send, proc_shift, rec_counter, &
         rec_local_col, rec_local_row, rec_pcol, rec_prow, ref_rec_pcol, ref_rec_prow, &
         send_counter, size_rec_buffer, size_send_buffer
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: req_send
      INTEGER, DIMENSION(:), POINTER                     :: col_indices_d, col_indices_s, &
                                                            row_indices_d, row_indices_s

      CALL timeset(routineN, handle)

      ! get info for the source
      CALL cp_fm_get_info(matrix=fm_mat_source, &
                          nrow_local=nrow_local_s, &
                          ncol_local=ncol_local_s, &
                          row_indices=row_indices_s, &
                          col_indices=col_indices_s, &
                          nrow_block=nrow_block_s, &
                          ncol_block=ncol_block_s)
      myprow_s = fm_mat_source%matrix_struct%context%mepos(1)
      mypcol_s = fm_mat_source%matrix_struct%context%mepos(2)
      nprow_s = fm_mat_source%matrix_struct%context%num_pe(1)
      npcol_s = fm_mat_source%matrix_struct%context%num_pe(2)

      ! get info for the dest
      CALL cp_fm_get_info(matrix=fm_mat_dest, &
                          nrow_local=nrow_local_d, &
                          ncol_local=ncol_local_d, &
                          row_indices=row_indices_d, &
                          col_indices=col_indices_d, &
                          nrow_block=nrow_block_d, &
                          ncol_block=ncol_block_d)
      myprow_d = fm_mat_dest%matrix_struct%context%mepos(1)
      mypcol_d = fm_mat_dest%matrix_struct%context%mepos(2)
      nprow_d = fm_mat_dest%matrix_struct%context%num_pe(1)
      npcol_d = fm_mat_dest%matrix_struct%context%num_pe(2)

      ! 0) check if the local data has to be stored in the new fm_mat_S
      CALL timeset(routineN//"_loc", handle2)
      IF (map_rec_size(para_env_RPA%mepos) > 0) THEN
         DO jjB = 1, ncol_local_s
            j_global = col_indices_s(jjB)
            rec_pcol = cp_fm_indxg2p(j_global, ncol_block_d, dummy_proc, &
                                     fm_mat_dest%matrix_struct%first_p_pos(2), npcol_d)
            IF (rec_pcol == mypcol_d) THEN
               j_local = cp_fm_indxg2l(j_global, ncol_block_d, dummy_proc, &
                                       fm_mat_dest%matrix_struct%first_p_pos(2), npcol_d)
               DO iiB = 1, nrow_local_s
                  i_global = row_indices_s(iiB)
                  rec_prow = cp_fm_indxg2p(i_global, nrow_block_d, dummy_proc, &
                                           fm_mat_dest%matrix_struct%first_p_pos(1), nprow_d)
                  IF (rec_prow == myprow_d) THEN
                     i_local = cp_fm_indxg2l(i_global, nrow_block_d, dummy_proc, &
                                             fm_mat_dest%matrix_struct%first_p_pos(1), nprow_d)
                     fm_mat_dest%local_data(i_local, j_local) = fm_mat_source%local_data(iiB, jjB)
                  END IF
               END DO
            END IF
         END DO
      END IF
      CALL timestop(handle2)

      ! 1) prepare receiving buffer
      CALL timeset(routineN//"_post", handle2)
      rec_counter = 0
      DO proc_shift = 1, para_env_RPA%num_pe-1
         proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)

         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1
            ! prepare the buffer for receive
            ALLOCATE (buffer_rec(rec_counter)%msg(size_rec_buffer))
            buffer_rec(rec_counter)%msg = 0.0_dp
            buffer_rec(rec_counter)%proc = proc_receive

            ! post the receiving message
            CALL mp_irecv(buffer_rec(rec_counter)%msg, proc_receive, para_env_RPA%group, buffer_rec(rec_counter)%msg_req)

         END IF
      END DO
      CALL timestop(handle2)

      ! 2) prepare sending buffer
      CALL timeset(routineN//"_bS", handle2)
      ! allocate buffer for sending, fill the buffer, send the message
      ALLOCATE (req_send(number_of_send))
      send_counter = 0
      DO proc_shift = 1, para_env_RPA%num_pe-1
         proc_send = RPA_proc_map(para_env_RPA%mepos+proc_shift)
         size_send_buffer = map_send_size(proc_send)

         IF (map_send_size(proc_send) > 0) THEN
            send_counter = send_counter+1
            ! allocate
            ALLOCATE (buffer_send(send_counter)%msg(size_send_buffer))
            buffer_send(send_counter)%msg = 0.0_dp
            buffer_send(send_counter)%proc = proc_send

            ! fill
!$OMP         PARALLEL DO DEFAULT(NONE) PRIVATE(iii,iiB,jjB) &
!$OMP                                   SHARED(size_send_buffer,buffer_send,&
!$OMP                                          send_counter,fm_mat_source)
            DO iii = 1, size_send_buffer
               iiB = buffer_send(send_counter)%indx(1, iii)
               jjB = buffer_send(send_counter)%indx(2, iii)
               buffer_send(send_counter)%msg(iii) = fm_mat_source%local_data(iiB, jjB)
            END DO
!$OMP         END PARALLEL DO

            ! send
            CALL mp_isend(buffer_send(send_counter)%msg, proc_send, para_env_RPA%group, buffer_send(send_counter)%msg_req)
            req_send(send_counter) = buffer_send(send_counter)%msg_req

         END IF
      END DO
      CALL timestop(handle2)

      ! 3) fill the fm_mat_dest matrix with the received data
      CALL timeset(routineN//"_fill", handle2)
      rec_counter = 0
      DO proc_shift = 1, para_env_RPA%num_pe-1
         proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)
         size_rec_buffer = map_rec_size(proc_receive)

         rec_local_row = local_size_source(1, proc_receive)
         rec_local_col = local_size_source(2, proc_receive)

         ref_rec_prow = fm_mat_source%matrix_struct%context%mpi2blacs(1, proc_receive)
         ref_rec_pcol = fm_mat_source%matrix_struct%context%mpi2blacs(2, proc_receive)

         IF (map_rec_size(proc_receive) > 0) THEN
            rec_counter = rec_counter+1

            CALL mp_wait(buffer_rec(rec_counter)%msg_req)

            ! fill the destination matrix
!$OMP         PARALLEL DO DEFAULT(NONE) PRIVATE(iii,i_local,j_local) &
!$OMP                                   SHARED(size_rec_buffer,proc_receive,&
!$OMP                                          rec_counter,fm_mat_dest,buffer_rec)
            DO iii = 1, size_rec_buffer
               i_local = buffer_rec(rec_counter)%indx(1, iii)
               j_local = buffer_rec(rec_counter)%indx(2, iii)
               fm_mat_dest%local_data(i_local, j_local) = buffer_rec(rec_counter)%msg(iii)
            END DO
!$OMP         END PARALLEL DO

            DEALLOCATE (buffer_rec(rec_counter)%msg)

         END IF
      END DO
      CALL timestop(handle2)

      ! 5) wait all
      CALL timeset(routineN//"_wA", handle2)
      CALL mp_waitall(req_send(:))
      DEALLOCATE (req_send)
      DO send_counter = 1, number_of_send
         DEALLOCATE (buffer_send(send_counter)%msg)
      END DO
      CALL timestop(handle2)

      CALL timestop(handle)

   END SUBROUTINE fm_redistribute

! **************************************************************************************************
!> \brief ...
!> \param RPA_proc_map ...
!> \param buffer_rec ...
!> \param buffer_send ...
!> \param number_of_rec ...
!> \param number_of_send ...
!> \param map_send_size ...
!> \param map_rec_size ...
!> \param local_size_source ...
! **************************************************************************************************
   SUBROUTINE release_buffer(RPA_proc_map, buffer_rec, buffer_send, &
                             number_of_rec, number_of_send, &
                             map_send_size, map_rec_size, local_size_source)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: RPA_proc_map
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_send
      INTEGER                                            :: number_of_rec, number_of_send
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: map_send_size, map_rec_size
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_size_source

      CHARACTER(LEN=*), PARAMETER :: routineN = 'release_buffer', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, rec_counter, send_counter

      CALL timeset(routineN, handle)

      IF (ALLOCATED(RPA_proc_map)) DEALLOCATE (RPA_proc_map)
      IF (ALLOCATED(map_send_size)) DEALLOCATE (map_send_size)
      IF (ALLOCATED(map_rec_size)) DEALLOCATE (map_rec_size)
      IF (ALLOCATED(local_size_source)) DEALLOCATE (local_size_source)

      IF (ALLOCATED(buffer_send)) THEN
         DO send_counter = 1, number_of_send
            IF (ASSOCIATED(buffer_send(send_counter)%msg)) DEALLOCATE (buffer_send(send_counter)%msg)
            IF (ASSOCIATED(buffer_send(send_counter)%indx)) DEALLOCATE (buffer_send(send_counter)%indx)
         END DO
         DEALLOCATE (buffer_send)
      END IF

      IF (ALLOCATED(buffer_rec)) THEN
         DO rec_counter = 1, number_of_rec
            IF (ASSOCIATED(buffer_rec(rec_counter)%msg)) DEALLOCATE (buffer_rec(rec_counter)%msg)
            IF (ASSOCIATED(buffer_rec(rec_counter)%indx)) DEALLOCATE (buffer_rec(rec_counter)%indx)
         END DO
         DEALLOCATE (buffer_rec)
      END IF

      CALL timestop(handle)

   END SUBROUTINE release_buffer

END MODULE rpa_communication
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to calculate the minimax coefficients for approximating 1/x as
!>        1/x ~ 1/pi SUM_{i}^{K} w_i x^2/(x^2+a_i^2)^2  for x belonging to [1:Rc].
!> \par History
!>      06.2014 created [Mauro Del Ben]
! **************************************************************************************************
MODULE minimax_rpa
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'minimax_rpa'

   PUBLIC get_rpa_minimax_coeff

CONTAINS

! **************************************************************************************************
!> \brief The a_i and w_i coefficient are stored in aw such that the first 1:K
!>        elements correspond to a_i and the K+1:2k correspond to w_i.
!>        All a_i and w_i have been fitted with a 12th order polynomial as
!>        a function of Rc for each value of K.
!> \param k ...
!> \param E_range ...
!> \param aw ...
!> \param ierr ...
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE get_rpa_minimax_coeff(k, E_range, aw, ierr)
      INTEGER                                            :: k
      REAL(KIND=dp)                                      :: E_range
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: aw
      INTEGER                                            :: ierr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_rpa_minimax_coeff', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=512)                                 :: error_msg
      CHARACTER(LEN=64)                                  :: Rc_cha
      INTEGER                                            :: handle, icoef, kkk
      REAL(KIND=dp)                                      :: L_b, R_transf, Rc, U_b
      REAL(KIND=dp), DIMENSION(13, 2*k)                  :: fit_coef

      CALL timeset(routineN, handle)
      Rc = E_range
      fit_coef = 0.0_dp
      ierr = 0

      SELECT CASE (k)
      CASE (1)
         CALL get_fit_coef_k01(Rc, L_b, U_b, fit_coef)
      CASE (2)
         CALL get_fit_coef_k02(Rc, L_b, U_b, fit_coef)
      CASE (3)
         CALL get_fit_coef_k03(Rc, L_b, U_b, fit_coef)
      CASE (4)
         CALL get_fit_coef_k04(Rc, L_b, U_b, fit_coef)
      CASE (5)
         CALL get_fit_coef_k05(Rc, L_b, U_b, fit_coef)
      CASE (6)
         CALL get_fit_coef_k06(Rc, L_b, U_b, fit_coef)
      CASE (7)
         CALL get_fit_coef_k07(Rc, L_b, U_b, fit_coef)
      CASE (8)
         CALL get_fit_coef_k08(Rc, L_b, U_b, fit_coef)
      CASE (9)
         CALL get_fit_coef_k09(Rc, L_b, U_b, fit_coef)
      CASE (10)
         CALL get_fit_coef_k10(Rc, L_b, U_b, fit_coef)
      CASE (11)
         CALL get_fit_coef_k11(Rc, L_b, U_b, fit_coef)
      CASE (12)
         CALL get_fit_coef_k12(Rc, L_b, U_b, fit_coef)
      CASE (13)
         CALL get_fit_coef_k13(Rc, L_b, U_b, fit_coef)
      CASE (14)
         CALL get_fit_coef_k14(Rc, L_b, U_b, fit_coef)
      CASE (15)
         CALL get_fit_coef_k15(Rc, L_b, U_b, fit_coef)
      CASE (16)
         CALL get_fit_coef_k16(Rc, L_b, U_b, fit_coef)
      CASE (17)
         CALL get_fit_coef_k17(Rc, L_b, U_b, fit_coef)
      CASE (18)
         CALL get_fit_coef_k18(Rc, L_b, U_b, fit_coef)
      CASE (19)
         CALL get_fit_coef_k19(Rc, L_b, U_b, fit_coef)
      CASE (20)
         CALL get_fit_coef_k20(Rc, L_b, U_b, fit_coef)
      CASE DEFAULT
         k = 20
         CALL get_fit_coef_k20(Rc, L_b, U_b, fit_coef(1:13, 1:40))
         ierr = -1
      END SELECT

      WRITE (Rc_cha, '(F16.2)') Rc
      Rc_cha = ADJUSTL(Rc_cha)
      IF (E_range < Rc) THEN
         WRITE (error_msg, '(A,A)') " The range for the Minimax approximation is smaller than the lower bound "// &
            "for the required number of quadrature points. The Minimax parameters "// &
            "have been optimized for the lower bound of: ", TRIM(Rc_cha)
         CALL cp__w("minimax_rpa.F",110,error_msg)
      END IF
      IF (E_range > Rc) THEN
         WRITE (error_msg, '(A,A)') " The range for the Minimax approximation exceeds the upper bound "// &
            "for the required number of quadrature points. The Minimax parameters "// &
            "have been optimized for the upper bound of: ", TRIM(Rc_cha)
         CALL cp__w("minimax_rpa.F",116,error_msg)
      END IF
      IF (ierr < 0) THEN
         error_msg = "The required number of quadrature point exceeds the maximum available. "// &
                     "The number of quadrature point has been reset to 20."
         CALL cp__w("minimax_rpa.F",121,error_msg)
      END IF

      R_transf = SQRT(L_b/Rc)
      R_transf = SQRT(R_transf)*(SQRT(U_b/(U_b-L_b))-SQRT(Rc/(U_b-L_b)))
      DO kkk = 1, 2*k
         aw(kkk) = fit_coef(1, kkk)
         DO icoef = 1, 12
            aw(kkk) = aw(kkk)+fit_coef(icoef+1, kkk)*(R_transf**icoef)
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE get_rpa_minimax_coeff

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k01(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 2)                    :: fit_coef

      REAL(KIND=dp), DIMENSION(52), PARAMETER :: c01 = (/8.4569134345088148E-01_dp, &
         -2.3864746901255809E-01_dp, -1.1361819501552294E-01_dp, 6.7690505100738471E-02_dp, &
         5.3985186361341728E-03_dp, -1.9612317325166117E-02_dp, 7.3513074383591715E-03_dp, &
         1.9996243815975012E-03_dp, -3.1205386442664557E-03_dp, 6.2573451848435199E-04_dp, &
         1.0159339561927395E-03_dp, -2.0336924331972007E-04_dp, -8.0157986968086049E-04_dp, &
         8.7743290162972425E+00_dp, -3.4658285658553871E+00_dp, -1.0159575820898459E+00_dp, &
         1.5316366676023252E+00_dp, -2.2374031893904067E-01_dp, -4.9079952564718116E-01_dp, &
         2.8542459603969011E-01_dp, 4.0445340834152646E-02_dp, -1.1213157710848982E-01_dp, &
         2.5807944540072045E-02_dp, 3.6871815925478856E-02_dp, -7.5013131351734158E-03_dp, &
         -2.8645535700322102E-02_dp, 9.3922532274951809E-01_dp, -2.1747417837053222E-07_dp, &
         -4.6907902157368980E-01_dp, -2.1616033822571426E-01_dp, 4.1319893871687463E-01_dp, &
         2.9123071569339970E-01_dp, -4.4281181896713878E-01_dp, -5.5512445336312510E-01_dp, &
         1.1432764546204668E+00_dp, -3.4020236095345024E-01_dp, -4.4151758350979708E-01_dp, &
         1.8627895375666054E-01_dp, 1.0229706519679303E-01_dp, 1.0233271450911223E+01_dp, &
         -3.6844989012010113E-06_dp, -7.8361514387479545E+00_dp, -3.6112788486477720E+00_dp, &
         9.5851351388967405E+00_dp, 7.2698001012821694E+00_dp, -1.1403856909523945E+01_dp, &
         -1.7651082087203267E+01_dp, 3.2669706643838275E+01_dp, -5.4176678145626020E+00_dp, &
         -1.4771604007512861E+01_dp, 1.1600808336065933E+00_dp, 6.4627594951385223E+00_dp/)
      REAL(KIND=dp), DIMENSION(13, 2, 2), PARAMETER :: coefdata = RESHAPE((/c01/), (/13, 2, 2/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+00_dp) Rc = 2.0000000000000000E+00_dp
      IF (Rc >= 5.8879280090332031E+00_dp) Rc = 5.8879280090332031E+00_dp
      IF (Rc >= 2.0000000000000000E+00_dp .AND. Rc <= 2.6646270991316325E+00_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+00_dp
         U_b = 2.6646270991316325E+00_dp
      END IF
      IF (Rc >= 2.6646270991316325E+00_dp .AND. Rc <= 5.8879280090332031E+00_dp) THEN
         irange = 2
         L_b = 2.6646270991316325E+00_dp
         U_b = 5.8879280090332031E+00_dp
      END IF

      fit_coef(1:13, 1:2) = coefdata(1:13, 1:2, irange)

   END SUBROUTINE get_fit_coef_k01

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k02(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 4)                    :: fit_coef

      REAL(KIND=dp), DIMENSION(156), PARAMETER :: c01 = (/4.9649110596357299E-01_dp, &
         -1.8397680763708085E-01_dp, -2.6764777836094898E-02_dp, 1.4685597048762128E-02_dp, &
         3.8551475599549168E-03_dp, -6.6371384965829621E-04_dp, -8.2857349546675376E-04_dp, &
         3.7111617653054724E-05_dp, 1.2814660801267966E-04_dp, 6.1538340479780230E-06_dp, &
         -2.4584409073123429E-05_dp, -5.0991779586331699E-06_dp, 1.0669432321911992E-05_dp, &
         2.3206994094733933E+00_dp, -1.3514563041382210E+00_dp, 3.1921448263800534E-01_dp, &
         3.5958872361872854E-02_dp, -5.3831249908297907E-02_dp, 1.7289732546709874E-02_dp, &
         -2.4440065554605303E-03_dp, -1.2724078521801746E-05_dp, 2.2546039620310139E-04_dp, &
         -9.7860415748580936E-05_dp, -6.5067080496765577E-05_dp, 2.6089206215138224E-05_dp, &
         5.2757851532030513E-05_dp, 4.3667474519629623E+00_dp, -1.8617368732105988E+00_dp, &
         -7.4376289802548262E-02_dp, 1.8267819450392606E-01_dp, -3.2026231603982562E-03_dp, &
         -1.8945310128742759E-02_dp, -3.2473576905305453E-03_dp, 4.1419261548186481E-03_dp, &
         8.8569143253757507E-04_dp, -6.7268426110531814E-04_dp, -3.8199070267735632E-04_dp, &
         6.2942305485704747E-05_dp, 2.1244794526782270E-04_dp, 1.3299400044019976E+01_dp, &
         -1.0102857731754517E+01_dp, 4.5167812366146727E+00_dp, -7.7404161077659717E-01_dp, &
         -3.4951153894568415E-01_dp, 3.1655455535081256E-01_dp, -1.2883665080828036E-01_dp, &
         2.4210630196288028E-02_dp, 6.5829796220587648E-03_dp, -4.6936987882240733E-03_dp, &
         -8.5247269973254237E-04_dp, 1.0535642322120099E-03_dp, 4.2267254389369188E-04_dp, &
         6.5760438758196227E-01_dp, -1.7918847011385985E-01_dp, -2.2673006578030033E-01_dp, &
         1.0051289981916312E-01_dp, -3.0133666064475630E-02_dp, -4.8262232970595134E-03_dp, &
         1.1378620206724364E-01_dp, -1.8105019665171873E-01_dp, 1.9273690986546238E-01_dp, &
         -1.6962737938522135E-01_dp, 9.8332759521266175E-02_dp, -2.6789710748995204E-02_dp, &
         1.0387096874176875E-03_dp, 4.1246654556789606E+00_dp, -3.0716392263747014E+00_dp, &
         -2.0785962463279271E+00_dp, 5.2664445183974742E+00_dp, -3.0074972977105965E+00_dp, &
         -2.7503179147598340E+00_dp, 7.6392314052684709E+00_dp, -8.4073849148263342E+00_dp, &
         4.7943508114021212E+00_dp, 2.6737517312394893E-01_dp, -2.9722441934551838E+00_dp, &
         2.3158844939360632E+00_dp, -6.5764037707338063E-01_dp, 6.1608320892349369E+00_dp, &
         -2.1898229650048577E+00_dp, -2.5422729209340225E+00_dp, 1.7935819542593845E+00_dp, &
         -3.0756618643446343E-01_dp, -3.4608361155905393E-01_dp, 1.6002204372920785E+00_dp, &
         -2.6003746540622035E+00_dp, 2.6278080511803972E+00_dp, -2.2341338029312703E+00_dp, &
         1.4684729639545540E+00_dp, -5.4384672282947255E-01_dp, 7.0214867320780153E-02_dp, &
         3.0690007708797335E+01_dp, -3.6434557745981337E+01_dp, -1.2802331409904522E+01_dp, &
         7.8237428089364300E+01_dp, -7.1628441704020958E+01_dp, -3.6767672386188337E+01_dp, &
         1.6598068739496907E+02_dp, -1.9500321193457228E+02_dp, 8.7558312074511292E+01_dp, &
         6.3428744200583431E+01_dp, -1.2995121398289083E+02_dp, 8.7747827655552456E+01_dp, &
         -2.3421029630375354E+01_dp, 6.8832717879511052E-01_dp, 8.9994611434700178E-08_dp, &
         -1.7447802413631236E-01_dp, 1.1128373934782229E-02_dp, 1.0288810881396471E-01_dp, &
         -9.6952440957114305E-02_dp, -6.4905803226054201E-02_dp, 2.4883468266829842E-01_dp, &
         -2.3908071572971198E-01_dp, 4.2672150253429388E-02_dp, 8.4512682635884184E-02_dp, &
         -3.6754750174471884E-02_dp, -1.2870785796384022E-02_dp, 4.7101965082399921E+00_dp, &
         1.9264636868860040E-06_dp, -3.6955911120984810E+00_dp, 2.3579843361005315E-01_dp, &
         4.5049726325907349E+00_dp, -2.3256735181630925E+00_dp, -5.7088231586772453E+00_dp, &
         9.2251939039827988E+00_dp, -2.1118379768244511E+00_dp, -6.1276676279296476E+00_dp, &
         2.9985533393084327E+00_dp, 4.9045918358347871E+00_dp, -4.1266597706732586E+00_dp, &
         6.5430742560184436E+00_dp, 1.1412075979871461E-06_dp, -2.2097196426847456E+00_dp, &
         1.4094456985563819E-01_dp, 1.5262198438277494E+00_dp, -1.2544592957128169E+00_dp, &
         -1.1128799568053420E+00_dp, 3.4976285641773011E+00_dp, -3.0348111780709122E+00_dp, &
         2.3892287319584160E-01_dp, 1.1788607699039138E+00_dp, -2.2391998953368625E-01_dp, &
         -3.5101044491985517E-01_dp, 3.8067685920148641E+01_dp, 2.5902565425035982E-05_dp, &
         -4.9301857791232557E+01_dp, 3.1465928358831281E+00_dp, 7.7404606802819202E+01_dp, &
         -3.3013072565887242E+01_dp, -1.2019494685884978E+02_dp, 1.5486223363142261E+02_dp, &
         3.2337501402758228E+01_dp, -1.8488455295931200E+02_dp, 5.0234562973130487E+01_dp, &
         1.4928893567718609E+02_dp, -1.0852276420140453E+02_dp/)
      REAL(KIND=dp), DIMENSION(13, 4, 3), PARAMETER :: coefdata = RESHAPE((/c01/), (/13, 4, 3/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+00_dp) Rc = 2.0000000000000000E+00_dp
      IF (Rc >= 2.5266985893249512E+01_dp) Rc = 2.5266985893249512E+01_dp
      IF (Rc >= 2.0000000000000000E+00_dp .AND. Rc <= 2.8845195225564151E+00_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+00_dp
         U_b = 2.8845195225564151E+00_dp
      END IF
      IF (Rc >= 2.8845195225564151E+00_dp .AND. Rc <= 1.1473793111539274E+01_dp) THEN
         irange = 2
         L_b = 2.8845195225564151E+00_dp
         U_b = 1.1473793111539274E+01_dp
      END IF
      IF (Rc >= 1.1473793111539274E+01_dp .AND. Rc <= 2.5266985893249512E+01_dp) THEN
         irange = 3
         L_b = 1.1473793111539274E+01_dp
         U_b = 2.5266985893249512E+01_dp
      END IF

      fit_coef(1:13, 1:4) = coefdata(1:13, 1:4, irange)

   END SUBROUTINE get_fit_coef_k02

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k03(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 6)                    :: fit_coef

      REAL(KIND=dp), DIMENSION(234), PARAMETER :: c01 = (/4.7223379816403283E-01_dp, &
         -2.2042504935487547E-01_dp, -1.0700855642800766E-01_dp, 2.9282041797865813E-02_dp, &
         3.3026836310057026E-02_dp, 1.1386996102844459E-02_dp, -1.2818345627763791E-02_dp, &
         -4.0229827105483680E-03_dp, 5.2714678853526230E-03_dp, -5.7896018199156346E-03_dp, &
         2.6848950232278156E-03_dp, 1.9016456475500060E-03_dp, -1.3675117740777492E-03_dp, &
         2.0032594434499789E+00_dp, -1.6049856670510629E+00_dp, -7.5084218460525085E-02_dp, &
         6.0247881175491358E-01_dp, -1.0588835881424769E-01_dp, -1.5334719856063742E-01_dp, &
         -3.6117192085523851E-02_dp, 1.1475935854850117E-01_dp, 4.7173969318522357E-02_dp, &
         -1.5024338659366171E-01_dp, 9.3080665778357816E-02_dp, -1.7678496124707836E-02_dp, &
         -1.7574693551064798E-03_dp, 6.9852771279712096E+00_dp, -9.8678652716743986E+00_dp, &
         5.8034120673487637E+00_dp, 8.4032860298957390E-01_dp, -3.9344819362623489E+00_dp, &
         2.9352624815927943E+00_dp, -1.0295714123401773E+00_dp, 6.4840906026874964E-02_dp, &
         2.3541965450457605E-01_dp, -3.8563728580228640E-01_dp, 3.7881278851716366E-01_dp, &
         -2.0274209183322051E-01_dp, 4.5981510089675416E-02_dp, 4.1012174536431987E+00_dp, &
         -2.2240356017273251E+00_dp, -8.3949352914134823E-01_dp, 4.9927067503647976E-01_dp, &
         2.8387636248232850E-01_dp, 4.7904669834267075E-04_dp, -1.7161769150318792E-01_dp, &
         -7.6990836877090770E-03_dp, 9.7592444267693049E-02_dp, -9.9847132214505516E-02_dp, &
         9.1265071496279340E-02_dp, -4.8072775002788476E-02_dp, 9.2276548198524724E-03_dp, &
         9.5513399432547033E+00_dp, -1.1675840721046807E+01_dp, 3.0385953635634566E+00_dp, &
         5.1268986462936210E+00_dp, -3.3964946209348490E+00_dp, -1.3155690145638068E+00_dp, &
         9.3593509658323282E-01_dp, 1.1892612191185217E+00_dp, -1.0436920428437203E-01_dp, &
         -2.2056329972056523E+00_dp, 2.3677798945356323E+00_dp, -1.0414299374786686E+00_dp, &
         1.7342693495440725E-01_dp, 3.9609801714084597E+01_dp, -7.3889749869401939E+01_dp, &
         6.9592887675897757E+01_dp, -2.4811522219671406E+01_dp, -2.2841578322462372E+01_dp, &
         4.4418611403245741E+01_dp, -3.7852497542173246E+01_dp, 1.6302785218464713E+01_dp, &
         4.4126861094794840E+00_dp, -1.3635441404324794E+01_dp, 1.1138794870115078E+01_dp, &
         -4.7156831629367568E+00_dp, 8.7213397592194231E-01_dp, 5.5875221039160261E-01_dp, &
         -8.9468055007220743E-02_dp, -1.2065995777914583E-01_dp, 4.8471007034375492E-02_dp, &
         -3.8301904720482650E-02_dp, 2.4275678600046053E-02_dp, 2.8136983363421911E-02_dp, &
         -8.4794476923930245E-02_dp, 1.2805871374430119E-01_dp, -1.3799846162209611E-01_dp, &
         1.0242711526473426E-01_dp, -4.6555343095993788E-02_dp, 9.7559415471718224E-03_dp, &
         2.7624820993735941E+00_dp, -9.4404189731579302E-01_dp, -1.0778440090092487E+00_dp, &
         1.0000508959158847E+00_dp, -4.0720645947670364E-01_dp, 3.2143562912951618E-02_dp, &
         5.5974630350026500E-01_dp, -1.3059019163765901E+00_dp, 1.7808779108085686E+00_dp, &
         -1.7587309793812396E+00_dp, 1.2489411338126502E+00_dp, -5.6093609251530230E-01_dp, &
         1.1760304859958214E-01_dp, 1.3544129923931036E+01_dp, -1.1316468354856699E+01_dp, &
         -7.6541274135369042E+00_dp, 2.1813286094573890E+01_dp, -1.5817891923183874E+01_dp, &
         -5.7933843562996516E+00_dp, 2.9001150775153093E+01_dp, -4.1936816889961740E+01_dp, &
         3.9707552859267892E+01_dp, -2.6000900774396619E+01_dp, 1.0755546253602940E+01_dp, &
         -2.0944265775352795E+00_dp, -1.5875486677710954E-02_dp, 5.0129963552083687E+00_dp, &
         -9.8497955022805650E-01_dp, -1.2826931706396834E+00_dp, 6.5452164227503740E-01_dp, &
         -3.9738951291723318E-01_dp, 2.3136357394360504E-01_dp, 3.5477656525902451E-01_dp, &
         -1.0231998800842284E+00_dp, 1.5032961384142487E+00_dp, -1.5914516297167578E+00_dp, &
         1.1742544052689099E+00_dp, -5.3334314176404851E-01_dp, 1.1180896150674231E-01_dp, &
         1.5835060026431352E+01_dp, -8.7111880093588692E+00_dp, -8.8826285808589898E+00_dp, &
         1.1712175066481947E+01_dp, -4.4172083122828187E+00_dp, -1.2587072349970747E+00_dp, &
         7.1497450368592714E+00_dp, -1.4524138295210200E+01_dp, 1.8912025585426967E+01_dp, &
         -1.7743638731961465E+01_dp, 1.2115533439508143E+01_dp, -5.3562991919747693E+00_dp, &
         1.1231498116960723E+00_dp, 1.0046853530042941E+02_dp, -1.2562130523877248E+02_dp, &
         -4.8899698987716903E+01_dp, 2.9033084145860795E+02_dp, -2.9969216846999444E+02_dp, &
         -5.7160567223254283E+01_dp, 5.7249435821985480E+02_dp, -8.6440457177677740E+02_dp, &
         7.2603165823538518E+02_dp, -3.1482726992615238E+02_dp, -1.7030029333405142E+01_dp, &
         9.2288380720163971E+01_dp, -3.3395722174722863E+01_dp, 5.7310488938023063E-01_dp, &
         4.7901259633936180E-08_dp, -9.1651209265819295E-02_dp, 2.7732065459135504E-02_dp, &
         2.8645481150354719E-02_dp, -6.7359425012262830E-02_dp, 3.8584259620296449E-02_dp, &
         6.3446129710153454E-02_dp, -1.5358151939100778E-01_dp, 1.2187793055913763E-01_dp, &
         1.5086978285231711E-02_dp, -9.6662929399273831E-02_dp, 5.2098208706298513E-02_dp, &
         2.9191177943658442E+00_dp, 5.4133537461405717E-07_dp, -1.0344448907244952E+00_dp, &
         3.1300849735874009E-01_dp, 5.4891616747359162E-01_dp, -8.9562127361783095E-01_dp, &
         2.5683436290728645E-01_dp, 1.1981010184653673E+00_dp, -2.2140209166896678E+00_dp, &
         1.5279502299230125E+00_dp, 3.0429149987602905E-01_dp, -1.2173426926379263E+00_dp, &
         6.0860526435749707E-01_dp, 1.5577506560483336E+01_dp, 7.6602384911749927E-06_dp, &
         -1.4533497939231582E+01_dp, 4.3979051512530276E+00_dp, 1.5541107768555671E+01_dp, &
         -1.7236024655372692E+01_dp, -8.5725482253998937E+00_dp, 4.0651917108261443E+01_dp, &
         -4.5625286541426540E+01_dp, 1.7082545781556608E+01_dp, 1.1810832859818145E+01_dp, &
         -1.3783312630992485E+01_dp, 3.3847888256586232E+00_dp, 5.1721946579885314E+00_dp, &
         5.3542232317209938E-07_dp, -1.0242183084092691E+00_dp, 3.0991127039090627E-01_dp, &
         3.6925347779677709E-01_dp, -7.8228043802145897E-01_dp, 4.0037631487810599E-01_dp, &
         8.0471691788497290E-01_dp, -1.8201969969209009E+00_dp, 1.4022833800509154E+00_dp, &
         1.9468516183415735E-01_dp, -1.1133025066398865E+00_dp, 5.9138813613689389E-01_dp, &
         1.7309446051118943E+01_dp, 5.1997518715207899E-06_dp, -9.9271747598821172E+00_dp, &
         3.0038475175646426E+00_dp, 6.5128609543997671E+00_dp, -9.3405057766655286E+00_dp, &
         1.2450381441251936E+00_dp, 1.4427779610011438E+01_dp, -2.3898845190857141E+01_dp, &
         1.5265437336149422E+01_dp, 3.7526247706518707E+00_dp, -1.2192107238730179E+01_dp, &
         5.8009575841189420E+00_dp, 1.2420803433458813E+02_dp, 9.4529083484548541E-05_dp, &
         -1.7804086701584447E+02_dp, 5.3879411878296139E+01_dp, 2.5019726837316304E+02_dp, &
         -2.4632876610675547E+02_dp, -2.3728096521661567E+02_dp, 7.2832435058031899E+02_dp, &
         -6.3406381009820018E+02_dp, 7.2166814993792187E+01_dp, 2.2865645970566064E+02_dp, &
         -5.9497992164644941E+01_dp, -5.9908598748248629E+01_dp/)
      REAL(KIND=dp), DIMENSION(13, 6, 3), PARAMETER :: coefdata = RESHAPE((/c01/), (/13, 6, 3/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+00_dp) Rc = 2.0000000000000000E+00_dp
      IF (Rc >= 8.3026037216186523E+01_dp) Rc = 8.3026037216186523E+01_dp
      IF (Rc >= 2.0000000000000000E+00_dp .AND. Rc <= 9.5121852168680867E+00_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+00_dp
         U_b = 9.5121852168680867E+00_dp
      END IF
      IF (Rc >= 9.5121852168680867E+00_dp .AND. Rc <= 3.9527872132017364E+01_dp) THEN
         irange = 2
         L_b = 9.5121852168680867E+00_dp
         U_b = 3.9527872132017364E+01_dp
      END IF
      IF (Rc >= 3.9527872132017364E+01_dp .AND. Rc <= 8.3026037216186523E+01_dp) THEN
         irange = 3
         L_b = 3.9527872132017364E+01_dp
         U_b = 8.3026037216186523E+01_dp
      END IF

      fit_coef(1:13, 1:6) = coefdata(1:13, 1:6, irange)

   END SUBROUTINE get_fit_coef_k03

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k04(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 8)                    :: fit_coef

      REAL(KIND=dp), DIMENSION(312), PARAMETER :: c01 = (/4.0437679455062153E-01_dp, &
         -1.9968645981826424E-01_dp, -9.6669556285432015E-02_dp, 2.6839132012201752E-02_dp, &
         3.6190132805626725E-02_dp, 4.1166010414531336E-03_dp, -1.0769456290470580E-02_dp, &
         -8.7529874160434775E-05_dp, 5.7309585756926467E-03_dp, -1.0860752534826048E-02_dp, &
         4.3544925730678802E-03_dp, 3.1889806346690661E-03_dp, -1.9657212784353890E-03_dp, &
         1.5529811167323646E+00_dp, -1.1621893493531430E+00_dp, -1.7795567981281610E-01_dp, &
         3.9842720393802583E-01_dp, 4.8033159486636778E-02_dp, -1.4469416247317868E-01_dp, &
         -3.6520970201122292E-02_dp, 8.8008094722564351E-02_dp, 5.9318119282124994E-03_dp, &
         -7.6009276630089620E-02_dp, 6.7040879385407398E-02_dp, -2.8473998439286250E-02_dp, &
         5.2089090903696678E-03_dp, 4.2289764691942366E+00_dp, -5.2559714863096954E+00_dp, &
         1.6753291726988750E+00_dp, 1.8504840635800728E+00_dp, -1.6044237761904363E+00_dp, &
         -4.1733772536083369E-01_dp, 7.9444215167607390E-01_dp, 3.1495572908490888E-01_dp, &
         -7.9016995059826667E-01_dp, 1.9066260373671118E-01_dp, 3.5297731186965675E-01_dp, &
         -3.0243198972383212E-01_dp, 7.7756449578611248E-02_dp, 1.2967616994753087E+01_dp, &
         -2.4042024222254298E+01_dp, 2.0845923141347605E+01_dp, -5.1419824074547673E+00_dp, &
         -7.6388284418830947E+00_dp, 8.8813058337816244E+00_dp, -3.7007126133769321E+00_dp, &
         2.3382249105105274E-01_dp, -3.6597180661269557E-01_dp, 1.2892925753313280E+00_dp, &
         -1.1365833385148225E+00_dp, 4.5209303506995352E-01_dp, -7.1102161040361975E-02_dp, &
         3.4348988639807971E+00_dp, -1.9004725480240556E+00_dp, -7.5786793916739725E-01_dp, &
         3.8852279314766375E-01_dp, 3.0614786999111232E-01_dp, -4.5572384093187306E-02_dp, &
         -1.1543847219093645E-01_dp, 1.9867871621586992E-02_dp, 9.2786670262761134E-02_dp, &
         -1.8405662264456446E-01_dp, 1.5762768904523788E-01_dp, -5.6388852934436663E-02_dp, &
         5.3681094755664528E-03_dp, 6.4373196501777841E+00_dp, -7.1885129098744072E+00_dp, &
         8.4206205249439814E-01_dp, 3.2712289706274404E+00_dp, -8.7181606322780358E-01_dp, &
         -1.4986726477160628E+00_dp, 2.5136694217250172E-01_dp, 9.7470575226088230E-01_dp, &
         -3.3001228959270695E-01_dp, -3.8632103708308441E-01_dp, 3.1100472103485055E-01_dp, &
         -7.3188169703122946E-02_dp, 3.0493082567784697E-03_dp, 1.7166455411776617E+01_dp, &
         -3.1333371912090001E+01_dp, 2.1039285623747343E+01_dp, 6.3380935910509359E+00_dp, &
         -1.7459609199004912E+01_dp, 4.3115360437865684E+00_dp, 7.8590901947406229E+00_dp, &
         -2.1493300987092994E+00_dp, -9.7183726596572662E+00_dp, 1.1703427571259667E+01_dp, &
         -5.4936928390403912E+00_dp, 7.6914546328373024E-01_dp, 1.3439917053669492E-01_dp, &
         6.8475451857998991E+01_dp, -1.5591045907250185E+02_dp, 1.8464815785166243E+02_dp, &
         -1.1596150427703571E+02_dp, 7.1241317727507019E+00_dp, 6.6321778045233330E+01_dp, &
         -8.0699134248162082E+01_dp, 5.8884469891028630E+01_dp, -2.9896108613598663E+01_dp, &
         8.3528206096836914E+00_dp, 1.7944119950838084E+00_dp, -2.7711044075604732E+00_dp, &
         8.2189970277835223E-01_dp, 4.8928026378102546E-01_dp, -7.7395534889906831E-02_dp, &
         -9.4804748738418779E-02_dp, 3.4028552814568652E-02_dp, -3.7372824741901886E-02_dp, &
         5.0155168087250734E-02_dp, -2.5580726927850888E-02_dp, -1.2765405686004021E-02_dp, &
         4.7823767294427916E-02_dp, -7.0139763686717996E-02_dp, 6.3966662234848093E-02_dp, &
         -3.3470391316181090E-02_dp, 7.6464138985924456E-03_dp, 2.1282655054474775E+00_dp, &
         -6.1070711898886576E-01_dp, -6.5624939488951939E-01_dp, 4.7938223840521010E-01_dp, &
         -2.8747099836990569E-01_dp, 3.5208496407539869E-01_dp, -1.8503707524330915E-01_dp, &
         -2.0984499620126879E-01_dp, 5.4309316443457201E-01_dp, -6.7503677537568874E-01_dp, &
         5.6360705631090735E-01_dp, -2.8697201451927273E-01_dp, 6.5788018908140136E-02_dp, &
         7.5361243129007143E+00_dp, -4.3792129089618763E+00_dp, -3.6353821872855629E+00_dp, &
         5.5509223987118927E+00_dp, -3.0694183036306772E+00_dp, 1.6845171689522207E+00_dp, &
         -3.0607322949855914E-01_dp, -3.1489329672562265E+00_dp, 6.5649521591318027E+00_dp, &
         -7.2571558002312946E+00_dp, 5.2325909153168650E+00_dp, -2.3623770342134143E+00_dp, &
         5.0394953774065110E-01_dp, 3.5135862296987604E+01_dp, -4.1984978325667221E+01_dp, &
         -1.2198457583034720E+01_dp, 8.0268759850780640E+01_dp, -9.2774677162666293E+01_dp, &
         3.2220400417767237E+01_dp, 6.9167997126742932E+01_dp, -1.6719365094500833E+02_dp, &
         2.2313470021774773E+02_dp, -2.1138538881579083E+02_dp, 1.4038632740328094E+02_dp, &
         -5.8475121575654534E+01_dp, 1.1417246912012875E+01_dp, 4.2741823086253019E+00_dp, &
         -7.9536270802455533E-01_dp, -9.4491706673580489E-01_dp, 4.1996484386720750E-01_dp, &
         -3.7185002444196380E-01_dp, 5.0738873025610154E-01_dp, -2.6146306289710541E-01_dp, &
         -1.6591733635017125E-01_dp, 5.3934898406905241E-01_dp, -7.5567887713707849E-01_dp, &
         6.7814453825724363E-01_dp, -3.5449909952088476E-01_dp, 8.1351842070653560E-02_dp, &
         1.0448175241518934E+01_dp, -4.7162392903479757E+00_dp, -4.6011899797114211E+00_dp, &
         4.7226102598841244E+00_dp, -2.3542529316399179E+00_dp, 2.4099345435619655E+00_dp, &
         -1.2513655298312776E+00_dp, -2.1934790172813501E+00_dp, 5.1389071993822633E+00_dp, &
         -5.9476078886883359E+00_dp, 4.6727698853554340E+00_dp, -2.3040973762603332E+00_dp, &
         5.2314098593306413E-01_dp, 4.0791877448869535E+01_dp, -3.5969613616045955E+01_dp, &
         -2.4171436973533854E+01_dp, 5.5569664985979848E+01_dp, -3.4214469221683196E+01_dp, &
         9.5381575304633053E+00_dp, 6.8341242106838243E+00_dp, -3.7624304752009998E+01_dp, &
         7.1464033968563371E+01_dp, -7.7194048734399516E+01_dp, 5.2693863162066592E+01_dp, &
         -2.2021176330374324E+01_dp, 4.3695027993673685E+00_dp, 2.5377039095893605E+02_dp, &
         -4.2994677187007676E+02_dp, 2.5923064187804403E+01_dp, 9.0781913943732923E+02_dp, &
         -1.5171598707454250E+03_dp, 8.3438365693987998E+02_dp, 1.2503155206476356E+03_dp, &
         -3.6979114649792905E+03_dp, 5.0518426675233222E+03_dp, -4.5767744436753173E+03_dp, &
         2.8251444928014480E+03_dp, -1.0881627886949789E+03_dp, 1.9770830037818715E+02_dp, &
         5.0267168969343001E-01_dp, -7.4520589209002717E-09_dp, -7.5842890294191626E-02_dp, &
         3.8582667911904310E-02_dp, 9.9191590198104990E-03_dp, -7.7610198141607994E-02_dp, &
         1.0186993606336588E-01_dp, 4.8879008495791680E-03_dp, -2.5794860820246435E-01_dp, &
         5.1563319707313904E-01_dp, -5.5779404977642033E-01_dp, 3.4264223722618931E-01_dp, &
         -9.4245905218378417E-02_dp, 2.2367579356890843E+00_dp, -6.1826581961609364E-08_dp, &
         -6.3088494762561431E-01_dp, 3.2094379025394543E-01_dp, 1.7810679428530926E-01_dp, &
         -7.4243963130673429E-01_dp, 8.2776076323167247E-01_dp, 3.0319271394860581E-01_dp, &
         -2.6573611449279531E+00_dp, 4.8575612228055789E+00_dp, -5.0211444947565047E+00_dp, &
         2.9860321004621575E+00_dp, -8.0025459339884109E-01_dp, 8.3489662287757334E+00_dp, &
         -4.8081262326306912E-07_dp, -4.9360724580789306E+00_dp, 2.5110974244370783E+00_dp, &
         2.6505734507780212E+00_dp, -7.0808033521004949E+00_dp, 5.7735097501417201E+00_dp, &
         6.5854985317005275E+00_dp, -2.7913575624582830E+01_dp, 4.4816273679035426E+01_dp, &
         -4.2776722427843723E+01_dp, 2.3856755737116387E+01_dp, -6.0412032353147502E+00_dp, &
         4.3783125829708453E+01_dp, -5.5075146143120555E-06_dp, -5.8188143370443861E+01_dp, &
         2.9602672801274529E+01_dp, 6.8935349052196329E+01_dp, -1.2146548315255620E+02_dp, &
         1.2141505510027159E+01_dp, 2.6316188098445042E+02_dp, -5.3756800699873463E+02_dp, &
         5.8351355045248351E+02_dp, -3.6666352432161307E+02_dp, 1.1185697561839862E+02_dp, &
         -6.4042001817970036E+00_dp, 4.4126879375879327E+00_dp, -7.7530076630345340E-08_dp, &
         -7.8950662310582564E-01_dp, 4.0163677540945897E-01_dp, 1.3226474718849238E-01_dp, &
         -8.3730503258614408E-01_dp, 1.0575301512207400E+00_dp, 1.2535085289938416E-01_dp, &
         -2.8368981498982810E+00_dp, 5.5425059181428447E+00_dp, -5.9300191913604117E+00_dp, &
         3.6151514581350099E+00_dp, -9.8848229054852588E-01_dp, 1.1300654351102278E+01_dp, &
         -4.9316510953177751E-07_dp, -5.0417664922059764E+00_dp, 2.5648534509539416E+00_dp, &
         1.9121547495752944E+00_dp, -6.4282832469653375E+00_dp, 6.4567195634727446E+00_dp, &
         3.8645778438672509E+00_dp, -2.3915493684299207E+01_dp, 4.1667602971700134E+01_dp, &
         -4.1912306733231269E+01_dp, 2.4412186136448963E+01_dp, -6.4287151562910099E+00_dp, &
         4.7661122084367847E+01_dp, -4.1563904102160877E-06_dp, -4.2866809130358867E+01_dp, &
         2.1807486009438179E+01_dp, 2.9911195911583174E+01_dp, -6.8460053271712155E+01_dp, &
         4.4618313870507741E+01_dp, 8.3146438950249362E+01_dp, -2.8242592797658313E+02_dp, &
         4.2273976065299058E+02_dp, -3.8330562005148283E+02_dp, 2.0396556823058276E+02_dp, &
         -4.9315105307208164E+01_dp, 3.4873416553466677E+02_dp, -6.2812514911455712E-05_dp, &
         -6.8223088256704716E+02_dp, 3.4708903230411840E+02_dp, 1.0990846081358511E+03_dp, &
         -1.7164437218305100E+03_dp, -5.1026413431245879E+02_dp, 4.8843366036040652E+03_dp, &
         -7.6063698448775094E+03_dp, 5.5268501397810141E+03_dp, -6.4043799549226162E+02_dp, &
         -1.8963185698837842E+03_dp, 1.0098848671258462E+03_dp/)
      REAL(KIND=dp), DIMENSION(13, 8, 3), PARAMETER :: coefdata = RESHAPE((/c01/), (/13, 8, 3/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+00_dp) Rc = 2.0000000000000000E+00_dp
      IF (Rc >= 2.3323979377746582E+02_dp) Rc = 2.3323979377746582E+02_dp
      IF (Rc >= 2.0000000000000000E+00_dp .AND. Rc <= 1.4737051046890191E+01_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+00_dp
         U_b = 1.4737051046890191E+01_dp
      END IF
      IF (Rc >= 1.4737051046890191E+01_dp .AND. Rc <= 9.3388181952559151E+01_dp) THEN
         irange = 2
         L_b = 1.4737051046890191E+01_dp
         U_b = 9.3388181952559151E+01_dp
      END IF
      IF (Rc >= 9.3388181952559151E+01_dp .AND. Rc <= 2.3323979377746582E+02_dp) THEN
         irange = 3
         L_b = 9.3388181952559151E+01_dp
         U_b = 2.3323979377746582E+02_dp
      END IF

      fit_coef(1:13, 1:8) = coefdata(1:13, 1:8, irange)

   END SUBROUTINE get_fit_coef_k04

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k05(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 10)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(120), PARAMETER :: c02 = (/-4.8858613834160547E-01_dp, &
         3.4033685149450971E-01_dp, -1.0142642654121627E-01_dp, 1.8656576613307461E+00_dp, &
         -2.2611249862234847E-07_dp, -4.2139600757794465E-01_dp, 2.7923611356976985E-01_dp, &
         4.7524652113269663E-03_dp, -5.0664309914469052E-01_dp, 9.4151166751519222E-01_dp, &
         -6.3706758785225281E-01_dp, -8.8091156763512435E-01_dp, 2.9544585972346740E+00_dp, &
         -3.8184952299416528E+00_dp, 2.5987562420872985E+00_dp, -7.6383091805613368E-01_dp, &
         5.8307532704464835E+00_dp, -1.3235163063855027E-06_dp, -2.4668718803223433E+00_dp, &
         1.6346632087317283E+00_dp, 4.2793395990681971E-01_dp, -3.4953188430530364E+00_dp, &
         5.6218711785562494E+00_dp, -2.5804721986537875E+00_dp, -8.1487301965478487E+00_dp, &
         2.1579397272924947E+01_dp, -2.6243755352892087E+01_dp, 1.7322843340400933E+01_dp, &
         -4.9952855649733046E+00_dp, 2.1137358448719954E+01_dp, -8.8571011614827196E-06_dp, &
         -1.6522540002420385E+01_dp, 1.0948652835693094E+01_dp, 7.3116158949398784E+00_dp, &
         -2.9281521088202449E+01_dp, 3.7027168556990446E+01_dp, -3.0044062555390699E-01_dp, &
         -9.2699397470543445E+01_dp, 1.9478454491936060E+02_dp, -2.1832575096605436E+02_dp, &
         1.3760074533870568E+02_dp, -3.8467640857581380E+01_dp, 1.1063436781801329E+02_dp, &
         -9.4105195497773129E-05_dp, -1.7704451196700302E+02_dp, 1.1732328617291050E+02_dp, &
         2.0294758259689306E+02_dp, -4.7724061601861166E+02_dp, 2.4811244866542938E+02_dp, &
         7.7992320599735456E+02_dp, -2.3222278304245528E+03_dp, 3.3978948507388491E+03_dp, &
         -3.0865392333615719E+03_dp, 1.6634416402126349E+03_dp, -4.0958960918663729E+02_dp, &
         3.9147839650080969E+00_dp, -3.2520105963657300E-07_dp, -6.0607587797083018E-01_dp, &
         4.0161328249353928E-01_dp, -4.2547612515729416E-02_dp, -6.6328981617512195E-01_dp, &
         1.3305448184666258E+00_dp, -1.0356092213054167E+00_dp, -9.2632217066791589E-01_dp, &
         3.7416671458058000E+00_dp, -5.0171843083604752E+00_dp, 3.4735112435596691E+00_dp, &
         -1.0314560294834274E+00_dp, 8.5066988542163120E+00_dp, -1.5956969892592431E-06_dp, &
         -2.9738314062541940E+00_dp, 1.9705962320072299E+00_dp, 2.6378420045856443E-01_dp, &
         -3.8802561477551043E+00_dp, 6.7413347741113512E+00_dp, -3.9103035750119401E+00_dp, &
         -7.8429956431410091E+00_dp, 2.3246376149210917E+01_dp, -2.9171150509747868E+01_dp, &
         1.9568524918391812E+01_dp, -5.7008266943227328E+00_dp, 2.7615861977495285E+01_dp, &
         -9.5262930517419120E-06_dp, -1.7758829666390575E+01_dp, 1.1767834951864666E+01_dp, &
         5.0226252636691662E+00_dp, -2.7730364017515328E+01_dp, 4.0708746346053992E+01_dp, &
         -1.2326936467644122E+01_dp, -7.4007384787068389E+01_dp, 1.7668831197880644E+02_dp, &
         -2.0786024226383526E+02_dp, 1.3475739031543512E+02_dp, -3.8406427466935618E+01_dp, &
         1.2015405393414257E+02_dp, -7.2621559578354175E-05_dp, -1.3559498479069612E+02_dp, &
         8.9852390667028018E+01_dp, 8.4938472949204964E+01_dp, -2.7318759226639492E+02_dp, &
         2.9372517367957340E+02_dp, 1.0805522341970379E+02_dp, -9.9062857988965527E+02_dp, &
         1.8854424898786242E+03_dp, -2.0220366613923659E+03_dp, 1.2394263774198967E+03_dp, &
         -3.3964650795183775E+02_dp, 8.8110212941579391E+02_dp, -1.0631056167459761E-03_dp, &
         -2.0215919999407249E+03_dp, 1.3397278899112848E+03_dp, 3.2785602402023524E+03_dp, &
         -6.7015638377241066E+03_dp, 8.2607523725594240E+02_dp, 1.7011468237133256E+04_dp, &
         -3.7898634628790373E+04_dp, 4.6113041763967398E+04_dp, -3.4882603781213867E+04_dp, &
         1.5359465882590917E+04_dp, -2.9911038365687018E+03_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/2.6550926316288959E-01_dp, &
         -1.4194808279243129E-01_dp, -1.8340243943233556E-02_dp, 1.5594859644926301E-02_dp, &
         6.5213609288545414E-03_dp, -7.2580394496045918E-04_dp, -1.4367869259991610E-03_dp, &
         -5.1441935266499043E-04_dp, 1.1018516863846910E-04_dp, 1.3857094404802231E-04_dp, &
         2.1996788631150751E-04_dp, -1.7152748784100114E-04_dp, 2.0272591048711931E-05_dp, &
         8.8501620116612245E-01_dp, -5.6405190932498062E-01_dp, 2.1373574602775065E-02_dp, &
         6.3146135511270907E-02_dp, -2.3147542901442224E-05_dp, -7.8175619763017744E-03_dp, &
         -9.0807036857844356E-04_dp, 5.6214855418821154E-04_dp, 2.4232956169489725E-04_dp, &
         7.6709494480488226E-05_dp, 7.0466487248201543E-05_dp, -1.6311093828886886E-04_dp, &
         5.2007073293462518E-05_dp, 1.8263499497356539E+00_dp, -1.5009396517669864E+00_dp, &
         4.1870586931086229E-01_dp, 8.1727208946953692E-02_dp, -7.9985490073436638E-02_dp, &
         -2.8386179694723992E-04_dp, 1.3714799393655325E-02_dp, -1.2765799618254241E-03_dp, &
         -2.3532630322310429E-03_dp, 2.9144423884117456E-04_dp, 5.3898879161715227E-04_dp, &
         -2.1117200296074146E-04_dp, 1.1890367495914655E-05_dp, 3.6068126097193955E+00_dp, &
         -3.7478336902205442E+00_dp, 1.9251356093214684E+00_dp, -3.3446998446816661E-01_dp, &
         -1.8448451339597130E-01_dp, 1.0688477101804973E-01_dp, 2.8699422521969551E-03_dp, &
         -1.7054194854644773E-02_dp, 2.7799971291916320E-03_dp, 2.1897310530225645E-03_dp, &
         -9.6910398641202955E-04_dp, -1.7014948966298835E-05_dp, 7.0764797933466401E-05_dp, &
         8.5076801150728940E+00_dp, -1.0347947968427668E+01_dp, 7.0922698151187697E+00_dp, &
         -2.6797820059032738E+00_dp, 2.9467919296352824E-01_dp, 2.0165045409973004E-01_dp, &
         -8.9988805784438325E-02_dp, 2.4637727661109806E-03_dp, 8.1179838192678364E-03_dp, &
         -1.1425709074878842E-03_dp, -1.4540082710946707E-03_dp, 7.4743096963710671E-04_dp, &
         -9.5059826033180061E-05_dp, 2.1799143945696855E+00_dp, -1.2215729848174013E+00_dp, &
         -1.0378540265907392E-01_dp, 1.3956467575170556E-01_dp, 4.1207538207079744E-02_dp, &
         -1.1282311598005327E-02_dp, -1.0171856224520444E-02_dp, -2.3373290671285817E-03_dp, &
         1.2653652219253339E-03_dp, 9.5707490422460495E-04_dp, 1.2101366427858700E-03_dp, &
         -1.2780291066208068E-03_dp, 2.5756088060468888E-04_dp, 2.9208363903622483E+00_dp, &
         -2.3899201322394812E+00_dp, 5.8884810920058861E-01_dp, 2.2762106025416384E-01_dp, &
         -1.3362043169847870E-01_dp, -3.7501085007398918E-02_dp, 2.4747513399946530E-02_dp, &
         9.5199228229788154E-03_dp, -3.2039915558658432E-03_dp, -1.9527625509728832E-03_dp, &
         -1.2995234139512991E-03_dp, 1.6838354170501096E-03_dp, -3.6264295088535243E-04_dp, &
         4.9061852059580007E+00_dp, -5.5772589654145772E+00_dp, 3.0458942185873026E+00_dp, &
         -3.1151471012072757E-01_dp, -5.0962295355233656E-01_dp, 1.7119353498272677E-01_dp, &
         7.3421207250380538E-02_dp, -4.2415114572035145E-02_dp, -1.0449306726232923E-02_dp, &
         8.8592606635370576E-03_dp, 2.3422146446452656E-03_dp, -3.0644537695472701E-03_dp, &
         7.3510771583872308E-04_dp, 1.0361057376012624E+01_dp, -1.3801723359447648E+01_dp, &
         1.0269637159792943E+01_dp, -3.8517372257931135E+00_dp, 4.8210238729628187E-02_dp, &
         6.4093439772823868E-01_dp, -2.4036195455750892E-01_dp, -2.8318831513976336E-02_dp, &
         5.0063938591181703E-02_dp, -9.0143432978221204E-03_dp, -7.6402101435511958E-03_dp, &
         4.7996517140455634E-03_dp, -8.8688820948506499E-04_dp, 3.7821289404628558E+01_dp, &
         -5.0486826995622586E+01_dp, 3.9699236632662554E+01_dp, -1.8952398193101370E+01_dp, &
         5.0163848584714357E+00_dp, -2.0867251269953616E-01_dp, -3.6024073310994120E-01_dp, &
         1.2257744971762999E-01_dp, -2.0556888444551520E-03_dp, -4.3391049691802507E-03_dp, &
         -5.8398844655175447E-03_dp, 4.3389323693662314E-03_dp, -7.8818019920401015E-04_dp, &
         3.9278921854245491E-01_dp, -1.3880097398618357E-01_dp, -7.8179351611203124E-02_dp, &
         1.1692302652801807E-02_dp, 2.2561926975071613E-02_dp, 2.7058132466622827E-03_dp, &
         -1.1199299265664485E-02_dp, 2.3864268449642063E-03_dp, 4.1154746187549900E-03_dp, &
         -6.1786230594345682E-04_dp, -3.7868033504497873E-03_dp, 2.7672863440031232E-03_dp, &
         -5.9500399325799157E-04_dp, 1.4855053290241629E+00_dp, -7.8227620557414324E-01_dp, &
         -2.7024820829841489E-01_dp, 2.1644799093884431E-01_dp, 9.2348870972741803E-02_dp, &
         -6.6068295446994954E-02_dp, -6.4291229844158415E-02_dp, 5.0728492961798738E-02_dp, &
         4.4472290549627404E-02_dp, -7.4201442830987882E-02_dp, 3.5972879001982185E-02_dp, &
         -4.4915225522013684E-03_dp, -1.0083365147057061E-03_dp, 3.9016131418911306E+00_dp, &
         -3.4134865404724506E+00_dp, -1.3191399671771223E-01_dp, 1.4894435669800952E+00_dp, &
         -1.9976923375832278E-01_dp, -6.1585252828530235E-01_dp, 7.5074460723037759E-03_dp, &
         4.4772359113701282E-01_dp, 4.1917618732983360E-03_dp, -5.4159522447435060E-01_dp, &
         5.2585946916715653E-01_dp, -2.2522710312470215E-01_dp, 3.9131815966317804E-02_dp, &
         1.0618774133100274E+01_dp, -1.4832497853376831E+01_dp, 5.4089669906270323E+00_dp, &
         5.8728749535356988E+00_dp, -5.9311544312086957E+00_dp, -8.0009436905790932E-01_dp, &
         2.6888418432666157E+00_dp, 1.2359800658100402E+00_dp, -3.5501238739964611E+00_dp, &
         1.5596171190481971E+00_dp, 8.2306292720997121E-01_dp, -1.0016591564251089E+00_dp, &
         2.8042093815884245E-01_dp, 3.4412759856983151E+01_dp, -7.0327105581163224E+01_dp, &
         6.3665816591943127E+01_dp, -1.3949266108790304E+01_dp, -3.0883863109385448E+01_dp, &
         3.7009631013886363E+01_dp, -1.7390259933678596E+01_dp, 2.3656018006829531E+00_dp, &
         -1.7787709312289635E+00_dp, 5.3590592455391164E+00_dp, -4.8736190529851529E+00_dp, &
         2.0252305602012526E+00_dp, -3.3791044802812337E-01_dp, 3.3248343266591256E+00_dp, &
         -1.3094675160912539E+00_dp, -6.6328990690593814E-01_dp, 1.8491505826004428E-01_dp, &
         2.0917162881033141E-01_dp, -1.0803587412967225E-02_dp, -1.1437161193809799E-01_dp, &
         3.7418912128605243E-02_dp, 5.1391607560251427E-02_dp, -3.0429928256627381E-02_dp, &
         -2.3127059109442100E-02_dp, 2.5736365267549230E-02_dp, -6.6907054068503953E-03_dp, &
         6.0158175654245101E+00_dp, -4.7207061941701625E+00_dp, -7.7455852362166566E-01_dp, &
         1.9487679367187645E+00_dp, 2.4649244194611070E-01_dp, -7.8079051035147518E-01_dp, &
         -2.9522118467223896E-01_dp, 5.0896330357227537E-01_dp, 2.9773062721204507E-01_dp, &
         -8.0505532840347449E-01_dp, 5.8712415328747258E-01_dp, -1.9752002341908578E-01_dp, &
         2.5075590818523352E-02_dp, 1.4995996415485168E+01_dp, -1.9905726450058165E+01_dp, &
         3.9266337752980962E+00_dp, 9.9560813158678823E+00_dp, -4.7405279236906184E+00_dp, &
         -4.1496599853648704E+00_dp, 2.0613032230603539E+00_dp, 3.2488688171619757E+00_dp, &
         -1.7390067336108939E+00_dp, -3.0683317274378470E+00_dp, 4.2625062495968713E+00_dp, &
         -2.1726002323140921E+00_dp, 4.3089670881740993E-01_dp, 4.4932892872124903E+01_dp, &
         -8.8545093801744059E+01_dp, 6.0520129702721192E+01_dp, 2.1122643416448547E+01_dp, &
         -5.7886283556557316E+01_dp, 1.8213156609123580E+01_dp, 2.3077349912808526E+01_dp, &
         -5.3877102833312573E+00_dp, -3.7896180227046528E+01_dp, 4.8693135565506630E+01_dp, &
         -2.6834347784939805E+01_dp, 6.5383813741214230E+00_dp, -3.5691234747562894E-01_dp, &
         1.8944775478692549E+02_dp, -4.7251287821524852E+02_dp, 5.8393508267544064E+02_dp, &
         -3.6596986190399576E+02_dp, -2.7795511500841403E+01_dp, 3.2140628060441372E+02_dp, &
         -3.8370262121429585E+02_dp, 2.7692519369868364E+02_dp, -1.2701896028677757E+02_dp, &
         1.8998010412055702E+01_dp, 2.1870301973287983E+01_dp, -1.7120461042981180E+01_dp, &
         4.2347422391201750E+00_dp, 4.4257797087053102E-01_dp, -5.2788253001961889E-02_dp, &
         -5.2533965745094266E-02_dp, 1.4778463051880717E-02_dp, -1.5137288357210552E-02_dp, &
         2.1322924325981523E-02_dp, -1.4854865556221506E-02_dp, 5.1232024991161550E-03_dp, &
         2.8565216174117648E-03_dp, -8.8401575833623354E-03_dp, 9.9499800976253043E-03_dp, &
         -5.7899840382429997E-03_dp, 1.4210619672786226E-03_dp, 1.7901185397017931E+00_dp, &
         -3.5083028235258551E-01_dp, -3.1680230121654573E-01_dp, 1.5935023709180390E-01_dp, &
         -9.6024129986843676E-02_dp, 1.3660790479315130E-01_dp, -1.0337866658231480E-01_dp, &
         2.3138570375758987E-02_dp, 3.9352695535177187E-02_dp, -7.4818829308294185E-02_dp, &
         7.5902068752193866E-02_dp, -4.3371504574516717E-02_dp, 1.0713355872630637E-02_dp, &
         5.4011480843282529E+00_dp, -1.9378498606319501E+00_dp, -1.4885710416432727E+00_dp, &
         1.3285785197655227E+00_dp, -6.1573768211511359E-01_dp, 6.6799377473279897E-01_dp, &
         -5.6560044387820196E-01_dp, 4.4739939170913824E-02_dp, 4.1235540457510034E-01_dp, &
         -5.8360743029848561E-01_dp, 5.0454137563363111E-01_dp, -2.6911125304703654E-01_dp, &
         6.5360795178081008E-02_dp, 1.8390811469886259E+01_dp, -1.1816135218340939E+01_dp, &
         -6.5672253964212350E+00_dp, 1.1633062919350717E+01_dp, -6.3986980710522809E+00_dp, &
         3.2939209809513676E+00_dp, -2.2272935468482853E+00_dp, -1.0294340901252237E+00_dp, &
         4.9393479720617810E+00_dp, -6.2342624050806759E+00_dp, 4.6837849201767963E+00_dp, &
         -2.1673488253493325E+00_dp, 4.7646725156583597E-01_dp, 8.4336157557514952E+01_dp, &
         -1.0089617503672321E+02_dp, -8.2208590050504053E+00_dp, 1.3366974137866799E+02_dp, &
         -1.5820903503567655E+02_dp, 7.9337305769534240E+01_dp, 3.5616012388753418E+01_dp, &
         -1.3130781106536887E+02_dp, 1.8168276052903283E+02_dp, -1.7517829713272582E+02_dp, &
         1.1971260321867801E+02_dp, -5.1913306577605510E+01_dp, 1.0642365990915014E+01_dp, &
         3.8045299867771694E+00_dp, -5.1963750597518243E-01_dp, -5.0493129539283388E-01_dp, &
         1.6923327934559071E-01_dp, -1.4533839571881732E-01_dp, 2.0900682452821240E-01_dp, &
         -1.4833602186795575E-01_dp, 4.6115749291176553E-02_dp, 3.5317361550774348E-02_dp, &
         -9.2726757854478412E-02_dp, 1.0182710534344803E-01_dp, -5.9085836061311556E-02_dp, &
         1.4532428085371502E-02_dp, 7.9810439938759341E+00_dp, -2.4066556149498681E+00_dp, &
         -2.0114738728086969E+00_dp, 1.3887125725182221E+00_dp, -6.6411710313452954E-01_dp, &
         8.9805661552767102E-01_dp, -7.2555023164908816E-01_dp, 1.0674885228225427E-01_dp, &
         3.7883539402094696E-01_dp, -6.0015012013448121E-01_dp, 5.6777546719923455E-01_dp, &
         -3.1846433701583721E-01_dp, 7.8788285144172049E-02_dp, 2.4582796909232854E+01_dp, &
         -1.3409035847628800E+01_dp, -9.0550767935687659E+00_dp, 1.1183625388029842E+01_dp, &
         -5.0365240125360282E+00_dp, 4.1413731325516521E+00_dp, -3.6960181281327560E+00_dp, &
         -1.1306341316679099E-01_dp, 3.8775217524474308E+00_dp, -5.0313030646474921E+00_dp, &
         3.9973018771671991E+00_dp, -2.0116602935864174E+00_dp, 4.7586858968669676E-01_dp, &
         9.8317340117791673E+01_dp, -9.0854955135839106E+01_dp, -3.6899617942823220E+01_dp, &
         1.0594926320423745E+02_dp, -6.9019058089648595E+01_dp, 2.3438541656567132E+01_dp, &
         -6.0730297290477884E+00_dp, -1.8851461187993809E+01_dp, 5.3716963536506903E+01_dp, &
         -6.6369119452540346E+01_dp, 4.8465181956859681E+01_dp, -2.1165852568624782E+01_dp, &
         4.3585743760971170E+00_dp, 6.0207029114733461E+02_dp, -9.8896334810595192E+02_dp, &
         2.3242568533798712E+02_dp, 1.3611103974253565E+03_dp, -2.3785984539197157E+03_dp, &
         1.6929413675263338E+03_dp, 4.9725086152063358E+02_dp, -2.8910968561531386E+03_dp, &
         4.1515560821383906E+03_dp, -3.8021899986365097E+03_dp, 2.3881906025704566E+03_dp, &
         -9.5025460048391619E+02_dp, 1.8084927413816368E+02_dp, 4.5372254459206868E-01_dp, &
         -3.2706671239427972E-08_dp, -6.0956447616697843E-02_dp, 4.0392497654627497E-02_dp, &
         -5.9611390145588245E-03_dp, -6.4482812112108656E-02_dp, 1.3285507045769324E-01_dp, &
         -1.0786011793963135E-01_dp, -8.2044143611323866E-02_dp, 3.5942368229798438E-01_dp/)
      REAL(KIND=dp), DIMENSION(13, 10, 4), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02/), (/13, 10, 4/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+00_dp) Rc = 2.0000000000000000E+00_dp
      IF (Rc >= 5.8933372497558594E+02_dp) Rc = 5.8933372497558594E+02_dp
      IF (Rc >= 2.0000000000000000E+00_dp .AND. Rc <= 5.0183126784059855E+00_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+00_dp
         U_b = 5.0183126784059855E+00_dp
      END IF
      IF (Rc >= 5.0183126784059855E+00_dp .AND. Rc <= 4.3973147761154188E+01_dp) THEN
         irange = 2
         L_b = 5.0183126784059855E+00_dp
         U_b = 4.3973147761154188E+01_dp
      END IF
      IF (Rc >= 4.3973147761154188E+01_dp .AND. Rc <= 2.1570029942124913E+02_dp) THEN
         irange = 3
         L_b = 4.3973147761154188E+01_dp
         U_b = 2.1570029942124913E+02_dp
      END IF
      IF (Rc >= 2.1570029942124913E+02_dp .AND. Rc <= 5.8933372497558594E+02_dp) THEN
         irange = 4
         L_b = 2.1570029942124913E+02_dp
         U_b = 5.8933372497558594E+02_dp
      END IF

      fit_coef(1:13, 1:10) = coefdata(1:13, 1:10, irange)

   END SUBROUTINE get_fit_coef_k05

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k06(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 12)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(224), PARAMETER :: c02 = (/-2.1886099081761100E-02_dp, &
         1.6493591518278907E-03_dp, 1.1789604920345118E-03_dp, 6.4862232853117350E+00_dp, &
         -1.7754848782963699E+00_dp, -1.3982954414549702E+00_dp, 8.0870739791892121E-01_dp, &
         -3.2912209721480506E-01_dp, 6.6541689936850468E-01_dp, -7.1706760669376735E-01_dp, &
         4.1574320605711801E-01_dp, -2.3388393290566312E-01_dp, 1.8531370116520027E-01_dp, &
         -1.0473625529229623E-01_dp, 2.3517641650793227E-02_dp, 8.6244301390663231E-04_dp, &
         1.7041857699471105E+01_dp, -8.0569049462408699E+00_dp, -5.3193377730235101E+00_dp, &
         5.2407625518133418E+00_dp, -1.8916937701142056E+00_dp, 2.6635338859946827E+00_dp, &
         -3.4546849156995960E+00_dp, 1.9859939673472693E+00_dp, -6.3042912150903241E-01_dp, &
         4.6651519138633141E-01_dp, -4.6544250709211266E-01_dp, 2.0725559386709488E-01_dp, &
         -2.9916484811181372E-02_dp, 5.4109098148865449E+01_dp, -4.0117846296551058E+01_dp, &
         -1.9343735087203584E+01_dp, 3.5163214199743337E+01_dp, -1.6224366437588795E+01_dp, &
         1.0967650063875425E+01_dp, -1.6086523192511546E+01_dp, 1.0289145829533357E+01_dp, &
         7.0780220688975037E-01_dp, -3.9012608853545987E+00_dp, 1.3711604453275446E+00_dp, &
         1.9280292939737056E-01_dp, -1.3952390127271364E-01_dp, 2.1301179294983942E+02_dp, &
         -2.4882116172344013E+02_dp, -4.4576343993020110E+01_dp, 2.8297777525326859E+02_dp, &
         -2.2475060506667668E+02_dp, 8.3513335806543239E+01_dp, -4.8903037857891732E+01_dp, &
         3.0142938850263359E+01_dp, 5.6123720416754203E+01_dp, -1.2447276961108798E+02_dp, &
         1.0366026212412571E+02_dp, -4.3281169846400822E+01_dp, 7.6490691991820876E+00_dp, &
         1.2550066552494284E+03_dp, -2.4322243152990673E+03_dp, 1.2301224175493214E+03_dp, &
         2.5239221182153856E+03_dp, -6.0423251531521428E+03_dp, 6.0570404483243292E+03_dp, &
         -1.7810143991349457E+03_dp, -4.4038002244714671E+03_dp, 8.8124430795611534E+03_dp, &
         -9.1651316571527568E+03_dp, 6.1357943020836110E+03_dp, -2.5081424703717635E+03_dp, &
         4.7811463385949389E+02_dp, 4.1708346282835440E-01_dp, 5.4402119511920706E-09_dp, &
         -5.5526039243043029E-02_dp, 4.5875452433465184E-02_dp, -1.9952652014140874E-02_dp, &
         -6.2561720018590866E-02_dp, 1.9429518598408954E-01_dp, -2.9279948367686742E-01_dp, &
         2.3756176667253856E-01_dp, -1.4901115793698551E-02_dp, -1.8840975031156562E-01_dp, &
         1.8774227251450593E-01_dp, -6.3735493877313892E-02_dp, 1.6278649020334413E+00_dp, &
         3.2838198887433358E-08_dp, -3.3846296868812692E-01_dp, 2.7963563823781828E-01_dp, &
         -9.0630995340932871E-02_dp, -4.3293262954703110E-01_dp, 1.2280220838385034E+00_dp, &
         -1.7442094645170365E+00_dp, 1.2336530163787787E+00_dp, 2.9572964716091310E-01_dp, &
         -1.5535951629300453E+00_dp, 1.3874995299222836E+00_dp, -4.5339014253773563E-01_dp, &
         4.5582381667386960E+00_dp, 1.5806304897319024E-07_dp, -1.6513291934494874E+00_dp, &
         1.3643063680863621E+00_dp, -2.2797406279030716E-01_dp, -2.4685898874588244E+00_dp, &
         6.2562167208053836E+00_dp, -8.1279599723834828E+00_dp, 4.3636702189236134E+00_dp, &
         4.3186587764282933E+00_dp, -1.0540942234481506E+01_dp, 8.5267333193930419E+00_dp, &
         -2.6774459759737059E+00_dp, 1.3685608614427798E+01_dp, 7.8953218257276503E-07_dp, &
         -8.3945440967525986E+00_dp, 6.9353993597959400E+00_dp, 3.6407979682346647E-01_dp, &
         -1.5080151073493695E+01_dp, 3.3241166265120924E+01_dp, -3.7369343631286654E+01_dp, &
         8.3569027854846123E+00_dp, 4.4753024708589386E+01_dp, -7.6432229501076378E+01_dp, &
         5.6664416689962970E+01_dp, -1.7092781886452087E+01_dp, 4.9412897930577003E+01_dp, &
         4.7277509167796719E-06_dp, -5.1468102651658292E+01_dp, 4.2521417495234374E+01_dp, &
         1.8361899440516524E+01_dp, -1.1920260676251814E+02_dp, 2.1112678563410637E+02_dp, &
         -1.6483281126514837E+02_dp, -1.3110713864031490E+02_dp, 5.5418022612192499E+02_dp, &
         -7.3738040876829530E+02_dp, 4.9946323905482791E+02_dp, -1.4364581567818436E+02_dp, &
         2.5856094754727951E+02_dp, 4.9537890342657514E-05_dp, -5.1982003609948777E+02_dp, &
         4.2947182873531199E+02_dp, 6.1345329880719396E+02_dp, -1.9062470900920243E+03_dp, &
         1.7662921114675391E+03_dp, 1.7231825404801061E+03_dp, -8.6457455493579109E+03_dp, &
         1.5120925743819069E+04_dp, -1.5451074708996110E+04_dp, 9.0782260203870264E+03_dp, &
         -2.3814784812948374E+03_dp, 3.5563510586400620E+00_dp, 5.2243732431620397E-08_dp, &
         -5.3458995179895974E-01_dp, 4.4167605244136621E-01_dp, -1.7944761997460976E-01_dp, &
         -6.2338748927321674E-01_dp, 1.8891718769031944E+00_dp, -2.8043855484821694E+00_dp, &
         2.2030189466232226E+00_dp, 1.0320235800834847E-02_dp, -1.9761291843643918E+00_dp, &
         1.9052121189045435E+00_dp, -6.3976729377541963E-01_dp, 6.9028517698471328E+00_dp, &
         2.0949242746734285E-07_dp, -2.1756673311275518E+00_dp, 1.7975131975615710E+00_dp, &
         -4.2778160177964319E-01_dp, -3.0405366789532842E+00_dp, 8.1022253001293034E+00_dp, &
         -1.0982617302565613E+01_dp, 6.8131825397372605E+00_dp, 3.8862264333809295E+00_dp, &
         -1.2053532500608455E+01_dp, 1.0154164646180618E+01_dp, -3.2431802300014412E+00_dp, &
         1.8989552788955329E+01_dp, 9.9286067106856817E-07_dp, -1.0472400361079698E+01_dp, &
         8.6521158401880101E+00_dp, -4.6627567381311841E-01_dp, -1.7284175855799315E+01_dp, &
         4.0771265171966490E+01_dp, -4.9482292403614558E+01_dp, 1.9569212507327620E+01_dp, &
         4.1160600907466055E+01_dp, -8.0872171406820684E+01_dp, 6.2337361427142859E+01_dp, &
         -1.9156859827683505E+01_dp, 6.4230168958280117E+01_dp, 5.2702432513283289E-06_dp, &
         -5.6724000801554709E+01_dp, 4.6863917998784160E+01_dp, 1.0053412110284610E+01_dp, &
         -1.1451045308948315E+02_dp, 2.3044585885631165E+02_dp, -2.2906289039834832E+02_dp, &
         -1.8606184753175402E+01_dp, 4.2289320906156615E+02_dp, -6.3539698307177753E+02_dp, &
         4.5150351227550135E+02_dp, -1.3329696991326111E+02_dp, 2.8071766112627870E+02_dp, &
         3.6997307117075781E-05_dp, -4.0789941535711910E+02_dp, 3.3699261109303779E+02_dp, &
         2.3647894788939269E+02_dp, -1.0952495362197274E+03_dp, 1.6854111697437818E+03_dp, &
         -8.5957784789276764E+02_dp, -2.1979269738574790E+03_dp, 6.1001514329566426E+03_dp, &
         -7.4416162018202294E+03_dp, 4.8441672159087721E+03_dp, -1.3611139403438265E+03_dp, &
         2.0591686087846747E+03_dp, 6.2055330463178533E-04_dp, -5.8307099335577632E+03_dp, &
         4.8176388108956071E+03_dp, 1.0175276953355078E+04_dp, -2.6714845272387549E+04_dp, &
         1.3141868511851815E+04_dp, 5.7600064337445598E+04_dp, -1.7044732018512717E+05_dp, &
         2.5361567867653081E+05_dp, -2.3362423558935185E+05_dp, 1.2691185251979054E+05_dp, &
         -3.1278542676771558E+04_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/2.3880318322659616E-01_dp, &
         -1.3504254699048457E-01_dp, -1.8598173825632135E-02_dp, 1.5951626606880761E-02_dp, &
         7.2684956676590529E-03_dp, -5.7141087071240480E-04_dp, -1.5395949213734461E-03_dp, &
         -7.1572363929054415E-04_dp, 3.6877254806601694E-05_dp, 5.4320682223059976E-06_dp, &
         5.5227348644639883E-04_dp, -3.0433080098958792E-04_dp, 2.0054857421109012E-05_dp, &
         7.7890990470111321E-01_dp, -5.1037664112865433E-01_dp, 2.2683669399025620E-03_dp, &
         6.3137195324291309E-02_dp, 6.2052527538001894E-03_dp, -7.4341644045277456E-03_dp, &
         -2.1210008138554367E-03_dp, -1.4297915693002166E-04_dp, 2.3675140917504573E-04_dp, &
         7.8277224401520325E-05_dp, 7.7539910614835689E-04_dp, -6.6808270119695333E-04_dp, &
         1.4504128675696130E-04_dp, 1.5338400887947281E+00_dp, -1.2608929319268196E+00_dp, &
         2.7752546229559455E-01_dp, 1.1281909197533520E-01_dp, -5.7715437974610692E-02_dp, &
         -1.2667929676877377E-02_dp, 1.1252151548690961E-02_dp, 1.3715704256258806E-03_dp, &
         -2.1997078310386719E-03_dp, -1.9740866319126966E-04_dp, 6.4159705514259713E-04_dp, &
         -2.0892816180705209E-04_dp, 1.2980332628359802E-05_dp, 2.7673985137655222E+00_dp, &
         -2.8693997600463321E+00_dp, 1.2857991829002497E+00_dp, -5.2344086831360041E-02_dp, &
         -2.1561376228191623E-01_dp, 5.8188217877604753E-02_dp, 3.2625544363151460E-02_dp, &
         -1.6261879788195335E-02_dp, -4.9150534412067178E-03_dp, 3.7784034995137794E-03_dp, &
         1.0172896355501188E-03_dp, -1.3544723088620958E-03_dp, 3.3791427042144639E-04_dp, &
         5.1404992525605566E+00_dp, -6.4443570834347890E+00_dp, 4.1957808536165739E+00_dp, &
         -1.2275923471635759E+00_dp, -2.0396556460478615E-01_dp, 2.7559304730791950E-01_dp, &
         -4.1449669923351355E-02_dp, -3.6410768056789396E-02_dp, 1.4347460568477105E-02_dp, &
         4.4454675528836068E-03_dp, -4.5378784097625498E-03_dp, 9.9961733190086135E-04_dp, &
         3.6523493717956677E-05_dp, 1.1734352150699646E+01_dp, -1.6617308812671567E+01_dp, &
         1.3192363740983607E+01_dp, -5.9719748231036816E+00_dp, 1.0980419528166845E+00_dp, &
         3.2645028974174173E-01_dp, -2.2662577335842013E-01_dp, 1.5473932548189148E-02_dp, &
         2.4554441955295301E-02_dp, -3.0077324938441938E-03_dp, -6.6466692299168616E-03_dp, &
         3.3841806314929717E-03_dp, -4.5363653651646422E-04_dp, 1.9504561037810109E+00_dp, &
         -1.1467433477431543E+00_dp, -1.1509042021220600E-01_dp, 1.4002884228893558E-01_dp, &
         4.9229414420382964E-02_dp, -9.3425227737155385E-03_dp, -1.1325413873739049E-02_dp, &
         -4.2563268902206224E-03_dp, 6.7116883229536705E-04_dp, 1.5906634439756071E-04_dp, &
         3.8770231995856335E-03_dp, -2.5391926650602447E-03_dp, 3.3891223789822106E-04_dp, &
         2.4673029044535064E+00_dp, -2.0268552313475112E+00_dp, 4.0046771139949433E-01_dp, &
         2.3814029693552785E-01_dp, -8.9396133850395965E-02_dp, -4.4527776658762855E-02_dp, &
         1.4866152379554987E-02_dp, 9.7091069081123305E-03_dp, -1.0458116448709301E-03_dp, &
         -1.6304626515434921E-04_dp, -2.5676032129145179E-03_dp, 1.0030055918766197E-03_dp, &
         1.2679235174542378E-04_dp, 3.7347941735050534E+00_dp, -4.2796614963173818E+00_dp, &
         2.0984473158827268E+00_dp, 3.8759752258727312E-02_dp, -4.5990959564031419E-01_dp, &
         5.1967372176975800E-02_dp, 9.7141975244401893E-02_dp, -1.3178648045339187E-02_dp, &
         -2.0209423274504174E-02_dp, 1.0105807719139734E-03_dp, 5.4186563920573787E-03_dp, &
         -1.2039467701267226E-03_dp, -2.3204020322593582E-04_dp, 6.5077723320734462E+00_dp, &
         -9.2352437198316721E+00_dp, 6.6750942739970016E+00_dp, -1.8455859476736007E+00_dp, &
         -6.6861629653657528E-01_dp, 5.9764889487189854E-01_dp, -5.0747709526665586E-03_dp, &
         -1.2913515441578532E-01_dp, 2.3272560810134339E-02_dp, 2.9135236898039468E-02_dp, &
         -1.4574834005683180E-02_dp, -5.2715455080038322E-04_dp, 1.3890615806649285E-03_dp, &
         1.3850913871979026E+01_dp, -2.1458713896727076E+01_dp, 1.8585391600120367E+01_dp, &
         -8.8983459146558150E+00_dp, 1.4489567962728551E+00_dp, 9.2702628293193634E-01_dp, &
         -6.2616397632266363E-01_dp, 5.7790606569577026E-02_dp, 9.4342841805570538E-02_dp, &
         -3.4918226526611050E-02_dp, -1.1666274195182996E-02_dp, 1.1468596370156781E-02_dp, &
         -2.5023897917183561E-03_dp, 5.1146534060966360E+01_dp, -7.7941926275268443E+01_dp, &
         6.8596107997848165E+01_dp, -3.6890210242625017E+01_dp, 1.1357177081042030E+01_dp, &
         -9.2047899141260558E-01_dp, -7.4474117122587935E-01_dp, 2.5322565001955349E-01_dp, &
         3.2248532258476223E-02_dp, -1.7095085335450899E-02_dp, -2.2598156948147496E-02_dp, &
         1.5156344013831745E-02_dp, -2.5379823676707301E-03_dp, 3.5751798194464240E-01_dp, &
         -1.2600309325537651E-01_dp, -6.8004084173955454E-02_dp, 1.2157899234698420E-02_dp, &
         2.1943110510575771E-02_dp, -7.9427632594316407E-04_dp, -1.0361087641281600E-02_dp, &
         2.4821596699121824E-03_dp, 4.0348366680994473E-03_dp, 2.8135401244154632E-04_dp, &
         -4.6877219120401567E-03_dp, 3.1546476763206766E-03_dp, -6.7677457007812901E-04_dp, &
         1.2968177885798766E+00_dp, -6.3941693810976563E-01_dp, -2.2959973185660049E-01_dp, &
         1.6087434354085126E-01_dp, 8.6486743214774675E-02_dp, -6.1614892715247518E-02_dp, &
         -4.8028034899867122E-02_dp, 3.8425445859627622E-02_dp, 3.8605581267093236E-02_dp, &
         -5.6164847560257060E-02_dp, 2.0910077215322051E-02_dp, 1.2964252900134450E-03_dp, &
         -1.8446256582284172E-03_dp, 3.1324438674766699E+00_dp, -2.4151854273598716E+00_dp, &
         -2.6047168751114452E-01_dp, 9.5966522342849570E-01_dp, 2.2893370477180448E-02_dp, &
         -4.6081953292291444E-01_dp, -1.8520616685596340E-02_dp, 2.7887399359474097E-01_dp, &
         7.6020644245194932E-02_dp, -4.2088644819457605E-01_dp, 3.6460600516358160E-01_dp, &
         -1.4357909360196239E-01_dp, 2.2753321647943275E-02_dp, 7.4899427697510124E+00_dp, &
         -8.8280335952168212E+00_dp, 1.7123462104755773E+00_dp, 3.8707486333542227E+00_dp, &
         -1.9815357349695950E+00_dp, -1.6832510627073594E+00_dp, 1.2202327185548325E+00_dp, &
         1.1503961401223624E+00_dp, -1.1691367635695169E+00_dp, -6.2515758649445019E-01_dp, &
         1.4260929313914967E+00_dp, -8.3005376615383542E-01_dp, 1.7732045352944734E-01_dp, &
         1.9233621800739694E+01_dp, -3.2998946276611100E+01_dp, 1.9380091371032101E+01_dp, &
         8.0399253608428722E+00_dp, -1.6904953829368541E+01_dp, 3.0927597143955228E+00_dp, &
         8.0994471956382856E+00_dp, -1.6782211346692346E+00_dp, -1.0394704303491993E+01_dp, &
         1.2830229456737785E+01_dp, -6.9393548763620077E+00_dp, 1.7427344551235533E+00_dp, &
         -1.2946538236454874E-01_dp, 5.8780095440425775E+01_dp, -1.3609530737303956E+02_dp, &
         1.4571588736412139E+02_dp, -6.0814926940405925E+01_dp, -4.0021974482403188E+01_dp, &
         7.1891281980526486E+01_dp, -3.7611826478981364E+01_dp, 3.7915945110802157E-01_dp, &
         3.5968397055679930E+00_dp, 8.9772162495814740E+00_dp, -1.2728332178915190E+01_dp, &
         6.6326513236414772E+00_dp, -1.3383759658374235E+00_dp, 2.9966529964603850E+00_dp, &
         -1.1564049434132586E+00_dp, -5.6912440531538755E-01_dp, 1.6403335898067262E-01_dp, &
         1.9593329346060476E-01_dp, -3.5619845228756358E-02_dp, -9.7999140185112146E-02_dp, &
         3.4747319410711512E-02_dp, 4.7910901871321382E-02_dp, -2.0772008642700124E-02_dp, &
         -2.9956272785111183E-02_dp, 2.7218123458554580E-02_dp, -6.6820638736595937E-03_dp, &
         4.9257007168397831E+00_dp, -3.5217910316373517E+00_dp, -6.7964650224305845E-01_dp, &
         1.3277332421011259E+00_dp, 2.8052383481804360E-01_dp, -6.1434742292954359E-01_dp, &
         -1.8565572658205229E-01_dp, 3.4561046137822643E-01_dp, 2.6278957514639034E-01_dp, &
         -6.1511718055603581E-01_dp, 4.0794606897589142E-01_dp, -1.1640640393281354E-01_dp, &
         1.0168781521691091E-02_dp, 1.0726490162771071E+01_dp, -1.2573726945225451E+01_dp, &
         1.2724164436038299E+00_dp, 6.0207845401403643E+00_dp, -1.6036784790091536E+00_dp, &
         -3.0437006266607063E+00_dp, 9.3830897885006914E-01_dp, 1.9452809069891837E+00_dp, &
         -5.0977128821033357E-01_dp, -2.3520680072775599E+00_dp, 2.7623531351073121E+00_dp, &
         -1.3250450645956215E+00_dp, 2.5196945929743420E-01_dp, 2.6913661603602964E+01_dp, &
         -4.5878078735449954E+01_dp, 2.0881697200238932E+01_dp, 1.8610130455880881E+01_dp, &
         -2.0410996419992742E+01_dp, -5.3120031731806314E+00_dp, 1.2504024396654433E+01_dp, &
         4.3865502206845264E+00_dp, -1.4269112044025723E+01_dp, 5.9137624753890998E+00_dp, &
         3.7147114162202497E+00_dp, -4.1086230435170954E+00_dp, 1.0998347157428636E+00_dp, &
         7.6888910996094467E+01_dp, -1.7775327311229657E+02_dp, 1.6411788739480767E+02_dp, &
         -9.6173540032258611E+00_dp, -1.1891208667143621E+02_dp, 8.3816879097966989E+01_dp, &
         3.2893073687014486E+01_dp, -5.7269860830743504E+01_dp, -3.4020895019064923E+01_dp, &
         1.1350920564093998E+02_dp, -1.0130944425436361E+02_dp, 4.3976474202829252E+01_dp, &
         -7.9786030819662486E+00_dp, 3.1097179840531845E+02_dp, -8.4703919274622899E+02_dp, &
         1.1584149634487351E+03_dp, -8.8086527965485436E+02_dp, 1.9720816629417556E+02_dp, &
         3.8303494310936651E+02_dp, -5.6094825875548361E+02_dp, 4.3778287553188028E+02_dp, &
         -2.5506476530794743E+02_dp, 1.3048998666847970E+02_dp, -5.8900820248922564E+01_dp, &
         1.9052077316170529E+01_dp, -2.9873042920703918E+00_dp, 4.0610663028040472E-01_dp, &
         -4.8301399895610123E-02_dp, -4.4786160630258885E-02_dp, 1.0238364112541655E-02_dp, &
         -9.8866021882278981E-03_dp, 1.9164217647153457E-02_dp, -1.7059481339925454E-02_dp, &
         1.1318694351430274E-02_dp, -8.4011787568469918E-03_dp, 5.7439095201463966E-03_dp, &
         -2.2151424764168225E-03_dp, 1.2458395701843222E-04_dp, 1.3096563994294966E-04_dp, &
         1.5621404690580485E+00_dp, -2.8409503600939162E-01_dp, -2.4137581977533118E-01_dp, &
         9.9141439607062593E-02_dp, -5.3835288846892100E-02_dp, 1.0987263584304191E-01_dp, &
         -1.0856682700179419E-01_dp, 6.6251217219820219E-02_dp, -4.3024826957833544E-02_dp, &
         3.1864939035592607E-02_dp, -1.5056601946291184E-02_dp, 2.2374108930218049E-03_dp, &
         4.7878419253682188E-04_dp, 4.2455143126678196E+00_dp, -1.3179848214982610E+00_dp, &
         -9.7615902340430949E-01_dp, 6.9415355017497327E-01_dp, -2.7067833562287136E-01_dp, &
         4.7087097135136075E-01_dp, -5.4436814808071254E-01_dp, 3.1543314880009077E-01_dp, &
         -1.4717023929479195E-01_dp, 1.1496028816623026E-01_dp, -7.8220850515749757E-02_dp, &
         2.4387786200859570E-02_dp, -1.6180203432776962E-03_dp, 1.2149461500713047E+01_dp, &
         -6.2522971044566518E+00_dp, -3.7120029221243351E+00_dp, 4.5879681385727071E+00_dp, &
         -1.8836178860735326E+00_dp, 1.9211532966550233E+00_dp, -2.6104498488332846E+00_dp, &
         1.5732843933959362E+00_dp, -2.5885107195900686E-01_dp, -1.4630825125608142E-02_dp, &
         -1.4387123907875027E-01_dp, 1.1644713286818151E-01_dp, -2.4631995705449012E-02_dp, &
         4.0517086212606905E+01_dp, -3.4146852052539010E+01_dp, -1.2144458368633835E+01_dp, &
         3.3721864721439921E+01_dp, -2.0892740591813197E+01_dp, 9.9040393481778999E+00_dp, &
         -1.0775109641610795E+01_dp, 7.1502845688098153E+00_dp, 3.4412821970099596E+00_dp, &
         -9.1170213254974826E+00_dp, 6.5769358742518946E+00_dp, -2.2907036885136565E+00_dp, &
         3.3510039300298694E-01_dp, 1.8014990635510685E+02_dp, -2.6175603836231346E+02_dp, &
         4.1474134976042123E+01_dp, 2.9935351196164999E+02_dp, -4.4486623371446325E+02_dp, &
         3.1189315679201121E+02_dp, -4.5857776124608712E+01_dp, -1.9606266882144880E+02_dp, &
         3.3823280092634457E+02_dp, -3.5026730689155517E+02_dp, 2.4561561076464852E+02_dp, &
         -1.0634490523999061E+02_dp, 2.1350371455168180E+01_dp, 3.4511581153937105E+00_dp, &
         -4.6074986504725718E-01_dp, -4.1793306232973060E-01_dp, 1.1447565909224683E-01_dp, &
         -9.1392964448968397E-02_dp, 1.8219557370192899E-01_dp, -1.6634354182093192E-01_dp, &
         1.0757299759560938E-01_dp, -7.7683124155527070E-02_dp, 5.4245972534987336E-02_dp/)
      REAL(KIND=dp), DIMENSION(13, 12, 4), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02/), (/13, 12, 4/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+00_dp) Rc = 2.0000000000000000E+00_dp
      IF (Rc >= 1.3773078918457031E+03_dp) Rc = 1.3773078918457031E+03_dp
      IF (Rc >= 2.0000000000000000E+00_dp .AND. Rc <= 6.1993729896731562E+00_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+00_dp
         U_b = 6.1993729896731562E+00_dp
      END IF
      IF (Rc >= 6.1993729896731562E+00_dp .AND. Rc <= 6.7432603912821875E+01_dp) THEN
         irange = 2
         L_b = 6.1993729896731562E+00_dp
         U_b = 6.7432603912821875E+01_dp
      END IF
      IF (Rc >= 6.7432603912821875E+01_dp .AND. Rc <= 4.2684148594526539E+02_dp) THEN
         irange = 3
         L_b = 6.7432603912821875E+01_dp
         U_b = 4.2684148594526539E+02_dp
      END IF
      IF (Rc >= 4.2684148594526539E+02_dp .AND. Rc <= 1.3773078918457031E+03_dp) THEN
         irange = 4
         L_b = 4.2684148594526539E+02_dp
         U_b = 1.3773078918457031E+03_dp
      END IF

      fit_coef(1:13, 1:12) = coefdata(1:13, 1:12, irange)

   END SUBROUTINE get_fit_coef_k06

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k07(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 14)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(328), PARAMETER :: c02 = (/-4.1754731782680632E-02_dp, &
         6.9637423606893978E-03_dp, 9.8317123038661812E-04_dp, 9.3141247748056060E+00_dp, &
         -3.3034187378384154E+00_dp, -2.3099777024265484E+00_dp, 1.9795050912537626E+00_dp, &
         -8.5251482053787397E-01_dp, 1.2153039011546896E+00_dp, -1.4541564892554977E+00_dp, &
         9.3955684867600964E-01_dp, -4.4317426042703595E-01_dp, 2.5918306441105204E-01_dp, &
         -1.4364447330590904E-01_dp, 3.6774610849770517E-02_dp, 3.7984486570491994E-05_dp, &
         2.6566070447108867E+01_dp, -1.4519054730075402E+01_dp, -8.2332223680110719E+00_dp, &
         1.1291093707590436E+01_dp, -5.2875112740441601E+00_dp, 4.8057488315402921E+00_dp, &
         -6.2898759890537734E+00_dp, 4.2232285921976249E+00_dp, -1.0596997995690820E+00_dp, &
         -1.4117925460764269E-01_dp, 6.8927735632697190E-02_dp, 1.2553388326253440E-02_dp, &
         1.7986555726404723E-03_dp, 8.9097724809361893E+01_dp, -7.5812556261643451E+01_dp, &
         -2.6441251265412269E+01_dp, 7.6120519989047565E+01_dp, -5.0397684172251800E+01_dp, &
         2.5525680254727302E+01_dp, -2.5376314020646312E+01_dp, 1.7957148357612915E+01_dp, &
         3.9328335707666944E+00_dp, -1.7750527732846916E+01_dp, 1.4718739531687156E+01_dp, &
         -6.0048241368590292E+00_dp, 1.0761088627700097E+00_dp, 4.0097466825142169E+02_dp, &
         -5.7103639425422421E+02_dp, 7.4021376601505452E+01_dp, 6.6329196339694579E+02_dp, &
         -9.6191980645397257E+02_dp, 6.7184174656773678E+02_dp, -1.1693581450498789E+02_dp, &
         -3.7933371221780379E+02_dp, 6.7374230959220836E+02_dp, -7.0914904191998016E+02_dp, &
         5.0560984367040209E+02_dp, -2.2277301564215111E+02_dp, 4.5529917350575040E+01_dp, &
         3.2067597521812030E+00_dp, -3.3444886966378912E-01_dp, -3.0363926355787030E-01_dp, &
         8.5884885141460449E-02_dp, -7.6865383844994903E-02_dp, 1.3715123754384689E-01_dp, &
         -1.2901078408577635E-01_dp, 9.2959485070991041E-02_dp, -6.8861591314818962E-02_dp, &
         4.5033224293921377E-02_dp, -1.7988390790948459E-02_dp, 2.2599225511603246E-03_dp, &
         5.2747450486550899E-04_dp, 5.6016069269588744E+00_dp, -1.1356436841662141E+00_dp, &
         -9.3290577892707627E-01_dp, 4.6398361532553123E-01_dp, -2.4608924526924877E-01_dp, &
         4.5627545506491052E-01_dp, -4.7134496624771294E-01_dp, 3.1401691722393993E-01_dp, &
         -2.0538932052729497E-01_dp, 1.3700800483099954E-01_dp, -5.9637767470295158E-02_dp, &
         8.4168325815202547E-03_dp, 1.8074844968448427E-03_dp, 1.3276336253382626E+01_dp, &
         -4.5209338962192493E+00_dp, -3.3197028028590223E+00_dp, 2.4854654336765201E+00_dp, &
         -1.0573849001152691E+00_dp, 1.7194404658483133E+00_dp, -1.9805547924035773E+00_dp, &
         1.2692361653686275E+00_dp, -6.7728308285586858E-01_dp, 4.3924148750555142E-01_dp, &
         -2.2613532847246806E-01_dp, 4.7779438733642737E-02_dp, 2.8573307488057543E-03_dp, &
         3.6866579094980089E+01_dp, -1.8921333920638745E+01_dp, -1.1800048250915063E+01_dp, &
         1.3436395189426809E+01_dp, -5.6772945221557212E+00_dp, 6.5245530820654025E+00_dp, &
         -8.4720689804565641E+00_dp, 5.4939590398921556E+00_dp, -1.9352146435742505E+00_dp, &
         7.9538222420964921E-01_dp, -5.7375841633563951E-01_dp, 2.1793383428406723E-01_dp, &
         -1.7879956499238674E-02_dp, 1.1861871394764727E+02_dp, -8.9749853349748633E+01_dp, &
         -4.1428506082732056E+01_dp, 8.1467781253253762E+01_dp, -4.2508182713086420E+01_dp, &
         2.7648470829113410E+01_dp, -3.6879292077457670E+01_dp, 2.5999309174132279E+01_dp, &
         -1.8359836135406524E+00_dp, -8.8433081142871490E+00_dp, 5.8531631602299123E+00_dp, &
         -1.6443587665309087E+00_dp, 2.0950921411586954E-01_dp, 4.7136803623790843E+02_dp, &
         -5.4268938167235092E+02_dp, -9.9473364877013339E+01_dp, 6.2005028905297843E+02_dp, &
         -5.1239886834398976E+02_dp, 2.1442818406738263E+02_dp, -1.2285032837677923E+02_dp, &
         7.7219547317156128E+01_dp, 9.0809838346683208E+01_dp, -2.3266758546949035E+02_dp, &
         2.0624116217858705E+02_dp, -9.2082040963793673E+01_dp, 1.7627448485288181E+01_dp, &
         2.8205675067555067E+03_dp, -5.3084226900396279E+03_dp, 2.4278553483206338E+03_dp, &
         5.7509013750262075E+03_dp, -1.2945226216946787E+04_dp, 1.2536477012596763E+04_dp, &
         -3.5209404505394255E+03_dp, -8.8941447689752422E+03_dp, 1.7502570418682619E+04_dp, &
         -1.8136983461264608E+04_dp, 1.2199553060977927E+04_dp, -5.0398280865362967E+03_dp, &
         9.7440960962504107E+02_dp, 3.8829047042903236E-01_dp, -2.0369951393868561E-08_dp, &
         -4.2760469054619193E-02_dp, 3.8145842955685037E-02_dp, -2.3458707726265494E-02_dp, &
         -3.4029805428294910E-02_dp, 1.4008388453470724E-01_dp, -2.5114912377498655E-01_dp, &
         2.8125815091946693E-01_dp, -1.8225356246487157E-01_dp, 3.1654580106274942E-02_dp, &
         3.9862347606661845E-02_dp, -2.1625916286328855E-02_dp, 1.4602532069412997E+00_dp, &
         -1.1343456594384706E-07_dp, -2.3779548855070412E-01_dp, 2.1213231958215728E-01_dp, &
         -1.1477625538424836E-01_dp, -2.1742813069917741E-01_dp, 8.0954389183928199E-01_dp, &
         -1.3959867402492177E+00_dp, 1.4874447805393780E+00_dp, -8.5261195435033799E-01_dp, &
         -4.8377391586045737E-03_dp, 3.3544213557537367E-01_dp, -1.5177913624259395E-01_dp, &
         3.7908369875064492E+00_dp, -4.8755632999722838E-07_dp, -1.0201467453462310E+00_dp, &
         9.1004690794757048E-01_dp, -3.9887236631988443E-01_dp, -1.1008475868239735E+00_dp, &
         3.6443712251015832E+00_dp, -5.9545892934730142E+00_dp, 5.8697667029529539E+00_dp, &
         -2.6303342326824413E+00_dp, -1.1564907871129693E+00_dp, 2.1468635303100569E+00_dp, &
         -8.4605567363145651E-01_dp, 1.0098487510998055E+01_dp, -2.0928481135752870E-06_dp, &
         -4.3686670908442755E+00_dp, 3.8971531298032169E+00_dp, -1.1974685442041848E+00_dp, &
         -5.6317683508763983E+00_dp, 1.6448696769067997E+01_dp, -2.5044320019174183E+01_dp, &
         2.1867904128895628E+01_dp, -5.0681337141096767E+00_dp, -1.1640616685797564E+01_dp, &
         1.3304985336172242E+01_dp, -4.7450643250946332E+00_dp, 3.0122997452208132E+01_dp, &
         -9.8187954584462908E-06_dp, -2.0428668438494554E+01_dp, 1.8223634263690474E+01_dp, &
         -2.1117881276751187E+00_dp, -3.2598057750235725E+01_dp, 8.1733300202487357E+01_dp, &
         -1.1128759913299137E+02_dp, 7.5228856030747480E+01_dp, 2.4248090620667043E+01_dp, &
         -1.0465160278140920E+02_dp, 9.2524007787235988E+01_dp, -3.0355195426460511E+01_dp, &
         1.0868874503660004E+02_dp, -5.7245956587297789E-05_dp, -1.1851742957082871E+02_dp, &
         1.0572360394609667E+02_dp, 2.3152742557056847E+01_dp, -2.5259994940936937E+02_dp, &
         5.0697463610979355E+02_dp, -5.3877942879889804E+02_dp, 8.2107193792851831E+01_dp, &
         7.1782122188306062E+02_dp, -1.1851552842814824E+03_dp, 8.7548914212122679E+02_dp, &
         -2.6543645341234554E+02_dp, 5.6870599977970653E+02_dp, -5.5564345141122099E-04_dp, &
         -1.1496133162535530E+03_dp, 1.0255182144253201E+03_dp, 1.1270281983242282E+03_dp, &
         -4.0567389638844206E+03_dp, 4.6523831299513085E+03_dp, 9.3327881111406134E+02_dp, &
         -1.3616969369427763E+04_dp, 2.6755200078585309E+04_dp, -2.9000034659790475E+04_dp, &
         1.7777250492541923E+04_dp, -4.8299772192891851E+03_dp, 3.2824512335777354E+00_dp, &
         -1.9156078888589085E-07_dp, -4.0197908004871374E-01_dp, 3.5859795441503123E-01_dp, &
         -2.1359817735657924E-01_dp, -3.3236257745885156E-01_dp, 1.3306435836390915E+00_dp, &
         -2.3614269209662528E+00_dp, 2.6114918833282528E+00_dp, -1.6438816802045650E+00_dp, &
         2.1906671723756974E-01_dp, 4.2427429641952058E-01_dp, -2.1705100617360901E-01_dp, &
         5.8637162691723583E+00_dp, -6.7787456845426279E-07_dp, -1.4194162820318337E+00_dp, &
         1.2662274814594010E+00_dp, -6.0649584934093481E-01_dp, -1.4391376533593900E+00_dp, &
         4.9819389045221047E+00_dp, -8.3199774717421811E+00_dp, 8.4774621253096996E+00_dp, &
         -4.2614106061434578E+00_dp, -9.5354015733314812E-01_dp, 2.5818521234785710E+00_dp, &
         -1.0661797920012148E+00_dp, 1.4340890738155760E+01_dp, -2.8142212154235703E-06_dp, &
         -5.8799168063088727E+00_dp, 5.2453046203869578E+00_dp, -1.8855227779627646E+00_dp, &
         -7.0881291395155026E+00_dp, 2.1731282085407450E+01_dp, -3.4078471876229919E+01_dp, &
         3.1408179116155015E+01_dp, -1.0416315289376390E+01_dp, -1.1856223667791461E+01_dp, &
         1.5589813689555845E+01_dp, -5.7586314676712993E+00_dp, 4.1438693013073269E+01_dp, &
         -1.2431014094133345E-05_dp, -2.5901135444548817E+01_dp, 2.3105496129919860E+01_dp, &
         -4.7050262856885112E+00_dp, -3.7691356222507558E+01_dp, 1.0121215124632137E+02_dp, &
         -1.4560344087059264E+02_dp, 1.1301980711089153E+02_dp, 5.8816837789371679E-01_dp, &
         -1.0167168279243205E+02_dp, 9.8807325076452315E+01_dp, -3.3542146164453939E+01_dp, &
         1.4116111924805350E+02_dp, -6.3883119768052977E-05_dp, -1.3258808163442185E+02_dp, &
         1.1827603870174343E+02_dp, 3.8519512221529113E+00_dp, -2.4308579826673557E+02_dp, &
         5.5187570135781425E+02_dp, -6.8463116272698392E+02_dp, 3.3797300275487532E+02_dp, &
         4.1534125658831630E+02_dp, -9.4503311947133648E+02_dp, 7.5927098696013218E+02_dp, &
         -2.3947738391807886E+02_dp, 6.1740638530300282E+02_dp, -4.4451581514866301E-04_dp, &
         -9.1747212798222540E+02_dp, 8.1842617649364138E+02_dp, 3.7908245119295702E+02_dp, &
         -2.3133450738754764E+03_dp, 4.0499698509594637E+03_dp, -3.3872426487702342E+03_dp, &
         -1.6589887942666303E+03_dp, 9.1305340362382631E+03_dp, -1.2658983382111219E+04_dp, &
         8.7808323210291001E+03_dp, -2.5756185231754689E+03_dp, 4.5291383930690972E+03_dp, &
         -6.0796509451681840E-03_dp, -1.2729055327721569E+04_dp, 1.1355442435730931E+04_dp, &
         1.9420434975201191E+04_dp, -5.7157101053784776E+04_dp, 4.1961441009360387E+04_dp, &
         8.2301170473632315E+04_dp, -2.9848869860082364E+05_dp, 4.7636758440468600E+05_dp, &
         -4.6114819297381287E+05_dp, 2.6149542649940349E+05_dp, -6.7032773576863750E+04_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/2.6444089699692058E-01_dp, &
         -1.5470504999202458E-01_dp, -4.6856578935403119E-02_dp, 2.5042624726789489E-02_dp, &
         1.9131665599864848E-02_dp, -2.9696778036559586E-03_dp, -3.7556543742271655E-03_dp, &
         1.6266657874186815E-04_dp, 4.6360340975112090E-03_dp, -7.8332396642386976E-03_dp, &
         1.9496480970880425E-03_dp, 3.0433369974490975E-03_dp, -1.5191543419188006E-03_dp, &
         8.7800372636622848E-01_dp, -6.2037241128343046E-01_dp, -8.1566349888960304E-02_dp, &
         1.3641008969898111E-01_dp, 3.1936419084862153E-02_dp, -3.7386849280010785E-02_dp, &
         -1.7482245191148163E-03_dp, 9.2077758125164444E-03_dp, 1.5212614028576979E-02_dp, &
         -4.4026339324134335E-02_dp, 3.6890098749264175E-02_dp, -1.1864460175613026E-02_dp, &
         8.4517974895171439E-04_dp, 1.7847969899964811E+00_dp, -1.6779481459284062E+00_dp, &
         2.0639156665955180E-01_dp, 4.1809341533061167E-01_dp, -1.2015867619368460E-01_dp, &
         -1.3848385742735578E-01_dp, 7.1455146740269884E-02_dp, 4.7031832850392787E-02_dp, &
         -2.3224446381251464E-02_dp, -6.5420768433947313E-02_dp, 8.5668141830795880E-02_dp, &
         -4.2757220494944660E-02_dp, 8.2228259261456181E-03_dp, 3.3330019221471758E+00_dp, &
         -4.2118312167575729E+00_dp, 1.6976220198254712E+00_dp, 7.5697277244499039E-01_dp, &
         -8.6691818383277741E-01_dp, -1.3679549434400173E-01_dp, 4.0739753850692434E-01_dp, &
         1.3289608237476967E-02_dp, -2.0557302626129631E-01_dp, 1.3788895525784092E-02_dp, &
         1.3789054262037703E-01_dp, -9.7837925769756875E-02_dp, 2.2630562602574813E-02_dp, &
         6.2205176319490292E+00_dp, -1.0220190906809847E+01_dp, 7.1086559724074778E+00_dp, &
         -3.5476394951841800E-01_dp, -2.7983732776043846E+00_dp, 1.1308103572081498E+00_dp, &
         9.1862692160779713E-01_dp, -6.5116275075965147E-01_dp, -5.4845428679228692E-01_dp, &
         8.5393419777208546E-01_dp, -4.0098327169829756E-01_dp, 5.2146858458048319E-02_dp, &
         9.6927653738469108E-03_dp, 1.2217423515701739E+01_dp, -2.4726297120084297E+01_dp, &
         2.4311598260589175E+01_dp, -1.0091505606342285E+01_dp, -3.4422060817821842E+00_dp, &
         5.9067475855100486E+00_dp, -1.1792990447941114E+00_dp, -1.9379521455346627E+00_dp, &
         8.8478642486008252E-01_dp, 1.0620423409368160E+00_dp, -1.3724468274695665E+00_dp, &
         6.3778052609253422E-01_dp, -1.1561528817453086E-01_dp, 2.9227809843211485E+01_dp, &
         -6.7776067894817842E+01_dp, 8.1967717455976384E+01_dp, -5.5738618729830407E+01_dp, &
         1.4735306844058032E+01_dp, 8.5839415419898337E+00_dp, -8.4661714956724126E+00_dp, &
         6.0685765490240084E-01_dp, 2.1941720347394860E+00_dp, -2.9568623894341339E-01_dp, &
         -1.1485475801649907E+00_dp, 8.0239930035514562E-01_dp, -1.7857991241725557E-01_dp, &
         2.1695269471571570E+00_dp, -1.3341984989631275E+00_dp, -3.4389129585042044E-01_dp, &
         2.4095334703936172E-01_dp, 1.4150794439393821E-01_dp, -4.3049372222392836E-02_dp, &
         -2.6067000342438331E-02_dp, 5.5273699894245716E-03_dp, 4.3837090204711077E-02_dp, &
         -8.5402788417645176E-02_dp, 4.5304897368381669E-02_dp, 5.5864340640867909E-03_dp, &
         -7.5526934617299851E-03_dp, 2.8719411494837974E+00_dp, -2.6619899142189949E+00_dp, &
         2.1544205537663841E-01_dp, 7.3077140572600108E-01_dp, -1.1791677396871071E-01_dp, &
         -2.6552626743236452E-01_dp, 7.5373709525561841E-02_dp, 8.7066019221526805E-02_dp, &
         3.0676066663530050E-03_dp, -1.7129353323714722E-01_dp, 2.1354379443237309E-01_dp, &
         -1.1821773827373153E-01_dp, 2.5969541945324099E-02_dp, 4.6027595145407618E+00_dp, &
         -6.3492168966396259E+00_dp, 2.6044413599352203E+00_dp, 1.5061915998878004E+00_dp, &
         -1.3747926425853620E+00_dp, -4.7941746387017842E-01_dp, 6.4441114269605160E-01_dp, &
         1.7317222818454989E-01_dp, -3.6457960453035143E-01_dp, 2.6412345558134549E-02_dp, &
         1.7751675692818278E-01_dp, -1.2546750431443543E-01_dp, 3.0158242944704584E-02_dp, &
         8.2140137621645088E+00_dp, -1.5156353495817562E+01_dp, 1.0947240829307352E+01_dp, &
         4.4625153473055351E-01_dp, -5.1300450604063039E+00_dp, 1.1684074353510967E+00_dp, &
         2.1518827109078962E+00_dp, -7.8884326600074284E-01_dp, -1.2573516650555800E+00_dp, &
         1.0661112404200725E+00_dp, -2.4342415056009679E-02_dp, -2.7206985827086660E-01_dp, &
         9.0849357725157689E-02_dp, 1.5901440158546173E+01_dp, -3.5870927834752429E+01_dp, &
         3.7033971510712249E+01_dp, -1.3426281591978212E+01_dp, -9.4793547969558798E+00_dp, &
         1.1009953605607825E+01_dp, 2.7169499451768997E-01_dp, -5.1938101275141140E+00_dp, &
         5.4398901692303922E-02_dp, 5.2939817053415910E+00_dp, -4.9957868625718627E+00_dp, &
         2.0763539253174690E+00_dp, -3.4840813951986949E-01_dp, 3.5630385206267455E+01_dp, &
         -8.9470008341766714E+01_dp, 1.1468454951859270E+02_dp, -7.9123579819767599E+01_dp, &
         1.4519114506197292E+01_dp, 2.4567610944439050E+01_dp, -2.0756718052404963E+01_dp, &
         -2.2859415630517427E-01_dp, 1.0041311285714274E+01_dp, -5.9487245509897209E+00_dp, &
         1.8350813953697698E-01_dp, 1.1110319845170011E+00_dp, -3.5005274800008057E-01_dp, &
         1.3147420367988465E+02_dp, -3.3027880006216265E+02_dp, 4.4490187241967470E+02_dp, &
         -3.6553068151822038E+02_dp, 1.7159265183702419E+02_dp, -1.6947279016681499E+01_dp, &
         -3.1006819193390971E+01_dp, 1.4450351194202961E+01_dp, 3.0038030298188554E+00_dp, &
         -2.7207016900807592E+00_dp, -1.8280868626869793E+00_dp, 2.0065403080939705E+00_dp, &
         -5.1840780008942378E-01_dp, 3.4505624259329698E-01_dp, -8.8324180899744836E-02_dp, &
         -4.6110672571305959E-02_dp, 6.0246988024288736E-03_dp, 1.1793333275048282E-02_dp, &
         -8.0133449953961704E-05_dp, -5.3349402958927537E-03_dp, 8.7754046282992803E-04_dp, &
         1.8165729377018357E-03_dp, -1.2767142351947607E-05_dp, -1.4303468914768861E-03_dp, &
         9.3816945542221544E-04_dp, -2.0328540292083126E-04_dp, 1.2346521989036530E+00_dp, &
         -4.3285647708194658E-01_dp, -1.7305878733158908E-01_dp, 7.6461063018067821E-02_dp, &
         5.3040050433052502E-02_dp, -2.1338549422926928E-02_dp, -2.7841143266162320E-02_dp, &
         1.3809276004079121E-02_dp, 1.2835193248531657E-02_dp, -1.0209736277395076E-02_dp, &
         -2.3155639359263330E-03_dp, 4.4602013911147409E-03_dp, -1.2940505553950435E-03_dp, &
         2.9028045455595000E+00_dp, -1.5585305439459514E+00_dp, -3.5847481830556316E-01_dp, &
         4.6103449323803292E-01_dp, 1.1117795227399657E-01_dp, -1.7456928361614407E-01_dp, &
         -7.1484101037929793E-02_dp, 9.9321509433804411E-02_dp, 5.0234375484831069E-02_dp, &
         -1.0773630683561888E-01_dp, 5.3437356912495085E-02_dp, -6.4215993294399389E-03_dp, &
         -1.6157910177500984E-03_dp, 6.6668490392315718E+00_dp, -5.3944853379232089E+00_dp, &
         -1.7333398494242408E-01_dp, 2.0362320110314558E+00_dp, -2.0552463518558933E-01_dp, &
         -8.4623002001901582E-01_dp, 6.6284256825966503E-02_dp, 4.8869806045351949E-01_dp, &
         9.5173683686716991E-03_dp, -5.5847228905099633E-01_dp, 5.0444583260989662E-01_dp, &
         -1.9696925294729772E-01_dp, 3.0305983725112742E-02_dp, 1.6122832821281406E+01_dp, &
         -1.8919243474316684E+01_dp, 4.0995726503647489E+00_dp, 7.3012627484186252E+00_dp, &
         -4.1313958700040532E+00_dp, -2.5095528548141255E+00_dp, 2.1272122489935632E+00_dp, &
         1.5307151716278145E+00_dp, -1.6669783074449143E+00_dp, -8.7588270967422077E-01_dp, &
         1.9886177370819973E+00_dp, -1.1470721637269370E+00_dp, 2.4363731014453902E-01_dp, &
         4.2621032980490320E+01_dp, -7.0376198214321079E+01_dp, 3.9171087201361452E+01_dp, &
         1.5272113738363034E+01_dp, -3.0728448719316823E+01_dp, 6.1171990910524778E+00_dp, &
         1.2221131447314844E+01_dp, -3.0894759011925577E+00_dp, -1.2362510396133640E+01_dp, &
         1.3678769893431953E+01_dp, -5.7528463567135004E+00_dp, 5.7726933215022425E-01_dp, &
         1.8772765681804807E-01_dp, 1.3491831737342127E+02_dp, -2.9806012349757901E+02_dp, &
         2.9841543160447191E+02_dp, -1.1170387310363532E+02_dp, -8.2203315408406809E+01_dp, &
         1.3165976851638894E+02_dp, -6.6112701554090933E+01_dp, 3.9695001548314552E+00_dp, &
         2.4844454458504051E+00_dp, 1.4202168620692383E+01_dp, -1.7455344455966326E+01_dp, &
         8.4672335324952464E+00_dp, -1.6173714685616287E+00_dp, 2.8828052690184736E+00_dp, &
         -8.0317040808415729E-01_dp, -3.9338075068484524E-01_dp, 7.9426195162603297E-02_dp, &
         1.0701666183071350E-01_dp, -1.0894403804546973E-02_dp, -5.0613235156913752E-02_dp, &
         1.2364103801681792E-02_dp, 1.8715420448839691E-02_dp, -3.9552027806400514E-03_dp, &
         -1.2146377166990881E-02_dp, 9.1425474616888842E-03_dp, -2.1144232652572591E-03_dp, &
         4.5886049198840508E+00_dp, -2.3077795883390579E+00_dp, -6.5366859610516070E-01_dp, &
         6.2765642178924896E-01_dp, 2.3574184645147922E-01_dp, -2.1996460901076612E-01_dp, &
         -1.4269945382294638E-01_dp, 1.2419991538546622E-01_dp, 8.4407863785150652E-02_dp, &
         -1.1735824544474278E-01_dp, 2.7653602309322248E-02_dp, 1.4931997736797792E-02_dp, &
         -6.8733985726516520E-03_dp, 9.5514395283650053E+00_dp, -7.7961380998560772E+00_dp, &
         -6.5864480154455607E-01_dp, 2.9590017002588471E+00_dp, 7.0736288843595091E-02_dp, &
         -1.2427968660363538E+00_dp, -1.3092708858258839E-01_dp, 6.9777177541844704E-01_dp, &
         1.9860466096286125E-01_dp, -8.4883928953606491E-01_dp, 6.1233781394136422E-01_dp, &
         -1.8682451466359717E-01_dp, 1.8956260629113059E-02_dp, 2.2681504235495581E+01_dp, &
         -2.6769556369521098E+01_dp, 3.8421035220956412E+00_dp, 1.1264519302530468E+01_dp, &
         -4.1146418985136082E+00_dp, -4.6910373462028936E+00_dp, 2.0744686686393647E+00_dp, &
         2.7897820660673656E+00_dp, -1.2053952427752199E+00_dp, -2.8742770555552286E+00_dp, &
         3.6128431894307571E+00_dp, -1.7420891025487704E+00_dp, 3.2880124209784622E-01_dp, &
         5.9399762273351683E+01_dp, -9.6606057719920145E+01_dp, 4.2613160864105112E+01_dp, &
         3.3344679979439910E+01_dp, -3.7404106629648304E+01_dp, -5.9153100847618516E+00_dp, &
         1.9306590740934482E+01_dp, 4.0805389526686682E+00_dp, -1.7468356988844324E+01_dp, &
         6.0010093546906882E+00_dp, 6.3096748376840859E+00_dp, -6.0062912235746344E+00_dp, &
         1.5620124014024448E+00_dp, 1.7618641024115868E+02_dp, -3.8264513931144631E+02_dp, &
         3.2569027453836890E+02_dp, -1.0595168890069804E+01_dp, -2.1449524316195871E+02_dp, &
         1.4019517892817902E+02_dp, 4.6943857262423244E+01_dp, -7.6432309148664586E+01_dp, &
         -4.8967291294613943E+01_dp, 1.4017406628314595E+02_dp, -1.1415196135040772E+02_dp, &
         4.5334643929331783E+01_dp, -7.5061186786457128E+00_dp, 7.3128032527974494E+02_dp, &
         -1.8965334349732912E+03_dp, 2.4281313649625095E+03_dp, -1.6963332283179839E+03_dp, &
         2.7274665872075781E+02_dp, 7.8978365574188206E+02_dp, -1.0283798449446413E+03_dp, &
         7.3549301335607697E+02_dp, -3.7290787690851232E+02_dp, 1.4467625159798999E+02_dp, &
         -4.0346760405500355E+01_dp, 5.3828624895447126E+00_dp, 3.3309708393995346E-01_dp, &
         3.8021272284375796E-01_dp, -3.5807245218190974E-02_dp, -3.3016407213619321E-02_dp, &
         8.2764560291011275E-03_dp, -8.3775496696670056E-03_dp, 1.4694568348773565E-02_dp, &
         -1.3633406730491519E-02_dp, 9.9748664766613742E-03_dp, -7.5058053308662068E-03_dp, &
         4.8898215686446114E-03_dp, -1.9376750420256020E-03_dp, 2.4578500454910593E-04_dp, &
         5.4135419394418854E-05_dp, 1.4158828018382148E+00_dp, -1.9428018520695578E-01_dp, &
         -1.6862783721434119E-01_dp, 6.3605882624844068E-02_dp, -4.3227688915331915E-02_dp, &
         7.9049443390899574E-02_dp, -7.7603914841223490E-02_dp, 5.3822690965089216E-02_dp, &
         -3.7801490871684225E-02_dp, 2.4957462270798092E-02_dp, -1.0325755989873857E-02_dp, &
         1.3457500127744759E-03_dp, 3.1573603989801897E-04_dp, 3.6037055399201039E+00_dp, &
         -8.0542869559173069E-01_dp, -6.3866068626924610E-01_dp, 3.6581611775536099E-01_dp, &
         -1.8032325305523136E-01_dp, 3.1726528871551424E-01_dp, -3.4024223010999666E-01_dp, &
         2.2409139887689597E-01_dp, -1.3723892898385098E-01_dp, 9.0749009387187926E-02_dp/)
      REAL(KIND=dp), DIMENSION(13, 14, 4), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02/), (/13, 14, 4/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+00_dp) Rc = 2.0000000000000000E+00_dp
      IF (Rc >= 3.0293891906738281E+03_dp) Rc = 3.0293891906738281E+03_dp
      IF (Rc >= 2.0000000000000000E+00_dp .AND. Rc <= 1.8737646720921180E+01_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+00_dp
         U_b = 1.8737646720921180E+01_dp
      END IF
      IF (Rc >= 1.8737646720921180E+01_dp .AND. Rc <= 1.6546209696020762E+02_dp) THEN
         irange = 2
         L_b = 1.8737646720921180E+01_dp
         U_b = 1.6546209696020762E+02_dp
      END IF
      IF (Rc >= 1.6546209696020762E+02_dp .AND. Rc <= 9.7857476979193905E+02_dp) THEN
         irange = 3
         L_b = 1.6546209696020762E+02_dp
         U_b = 9.7857476979193905E+02_dp
      END IF
      IF (Rc >= 9.7857476979193905E+02_dp .AND. Rc <= 3.0293891906738281E+03_dp) THEN
         irange = 4
         L_b = 9.7857476979193905E+02_dp
         U_b = 3.0293891906738281E+03_dp
      END IF

      fit_coef(1:13, 1:14) = coefdata(1:13, 1:14, irange)

   END SUBROUTINE get_fit_coef_k07

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k08(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 16)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(32), PARAMETER :: c03 = (/-8.9013840273255062E+02_dp, &
         4.4450456221108550E+02_dp, 5.3174323393059228E+02_dp, -1.2592882857295644E+03_dp, &
         1.0557554278716216E+03_dp, -3.5046238641220043E+02_dp, 1.2930757301899466E+03_dp, &
         1.1516939780618657E-03_dp, -1.6608759395740256E+03_dp, 1.4462154462223984E+03_dp, &
         3.8507708508937276E+02_dp, -3.0516713114026320E+03_dp, 5.3203711059143534E+03_dp, &
         -4.6213330077407009E+03_dp, -1.3283580700058483E+03_dp, 1.0430200674761059E+04_dp, &
         -1.5296967080846491E+04_dp, 1.1151529076955769E+04_dp, -3.4549406522575855E+03_dp, &
         9.4857715069098995E+03_dp, 1.5676078832282545E-02_dp, -2.2530834004834767E+04_dp, &
         1.9619149559132464E+04_dp, 2.5759373028729122E+04_dp, -7.7020575840804217E+04_dp, &
         6.3201727988014049E+04_dp, 6.8163632238120859E+04_dp, -2.9053341169350612E+05_dp, &
         4.7359098571404989E+05_dp, -4.6694609866564447E+05_dp, 2.7229709401159501E+05_dp, &
         -7.2557173401722219E+04_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/2.7310961221813451E-01_dp, &
         -9.5711109831144323E-02_dp, -2.5960297888544426E-02_dp, 7.7258590328211849E-03_dp, &
         5.1368496261163018E-03_dp, -1.1234374052788891E-03_dp, -1.0608869644587510E-03_dp, &
         4.0213920388401469E-04_dp, 3.0775150301006290E-04_dp, -5.8915574825377335E-05_dp, &
         -2.3576009852965474E-04_dp, 1.4943422000932038E-04_dp, -2.7193575609055488E-05_dp, &
         9.1301533296989179E-01_dp, -3.9165490470431008E-01_dp, -6.4063511602689208E-02_dp, &
         4.7237118624819781E-02_dp, 1.2334980874650106E-02_dp, -9.9115847862352204E-03_dp, &
         -2.3398525854494041E-03_dp, 3.1276613598798316E-03_dp, 6.9920456208351215E-04_dp, &
         -1.0397269514394823E-03_dp, -3.7227644159783696E-04_dp, 5.7251368824157297E-04_dp, &
         -1.5641106130487642E-04_dp, 1.8800921994120694E+00_dp, -1.0923258017805402E+00_dp, &
         -6.4592903048945596E-03_dp, 1.7145991202300620E-01_dp, -1.0258641827216298E-02_dp, &
         -4.0650011661850526E-02_dp, 5.9145330731061310E-03_dp, 1.1814015431998023E-02_dp, &
         -1.7645864693727417E-03_dp, -5.1968514483175078E-03_dp, 2.5254767142020877E-03_dp, &
         1.9008839648663282E-04_dp, -2.7310037630009072E-04_dp, 3.5700906437015720E+00_dp, &
         -2.8410095229290957E+00_dp, 4.7566776292824731E-01_dp, 4.5078537203369262E-01_dp, &
         -1.7662327214122509E-01_dp, -9.8111714854984447E-02_dp, 5.9791489912834436E-02_dp, &
         2.3988463192324526E-02_dp, -1.8060958844165663E-02_dp, -1.2105234356603977E-02_dp, &
         1.5101056055278140E-02_dp, -5.3449828416008731E-03_dp, 5.7306130912794712E-04_dp, &
         6.7581167823431345E+00_dp, -7.1699027258228432E+00_dp, 2.5445844349889786E+00_dp, &
         7.6672634398993644E-01_dp, -8.3067733243108621E-01_dp, -5.7644011834428951E-02_dp, &
         2.4822641367814308E-01_dp, -1.4162606641062165E-02_dp, -7.1554288045698092E-02_dp, &
         1.1874431773050711E-03_dp, 4.0456779863544762E-02_dp, -2.4380100288226139E-02_dp, &
         4.9075045448962002E-03_dp, 1.3121698258325543E+01_dp, -1.7881394250826862E+01_dp, &
         1.0058691417480512E+01_dp, -3.9782510556929929E-01_dp, -2.5566824525318088E+00_dp, &
         8.4735029392079875E-01_dp, 5.3977975375645593E-01_dp, -3.5357733173281275E-01_dp, &
         -1.1643964121149164E-01_dp, 1.4323415182866206E-01_dp, 2.7894270359389755E-03_dp, &
         -4.1885479536085533E-02_dp, 1.3600151683161683E-02_dp, 2.6983172825423299E+01_dp, &
         -4.5136785690700066E+01_dp, 3.5347241038093827E+01_dp, -1.1322977696396405E+01_dp, &
         -3.6432371690369023E+00_dp, 4.5653163916195441E+00_dp, -6.2527689198743497E-01_dp, &
         -1.0316108714043186E+00_dp, 3.9002390962416106E-01_dp, 2.7867998432914626E-01_dp, &
         -2.7830279098452576E-01_dp, 8.6406944254303217E-02_dp, -7.7988695959731675E-03_dp, &
         6.7020985153330088E+01_dp, -1.2895350676499106E+02_dp, 1.2515988271865132E+02_dp, &
         -6.7730535979917050E+01_dp, 1.3509022726826249E+01_dp, 7.5870813693372865E+00_dp, &
         -5.6551478844811518E+00_dp, 3.1592362611458347E-01_dp, 8.9353909446106417E-01_dp, &
         -5.1419383112248918E-02_dp, -3.4896277335001258E-01_dp, 1.9242088216158001E-01_dp, &
         -3.4930665201446769E-02_dp, 2.2444409953889672E+00_dp, -8.2981004194755881E-01_dp, &
         -2.0149587021835375E-01_dp, 7.6862324986017819E-02_dp, 4.0212219357266481E-02_dp, &
         -1.3088417493763247E-02_dp, -8.3346391656963330E-03_dp, 4.4376518893953151E-03_dp, &
         2.4395035166568041E-03_dp, -9.7210852177202040E-04_dp, -1.8091432752667988E-03_dp, &
         1.3510459080231529E-03_dp, -2.7978062458459800E-04_dp, 3.0233566510549710E+00_dp, &
         -1.7229666153085565E+00_dp, -5.9154969723083692E-02_dp, 2.7883404191218236E-01_dp, &
         3.3814752466955864E-03_dp, -6.7761866419578015E-02_dp, 2.6604108325230498E-03_dp, &
         2.0289299159528842E-02_dp, -6.7365152413148034E-04_dp, -8.5272902970696760E-03_dp, &
         2.3135442605826443E-03_dp, 1.5922697497338322E-03_dp, -7.4271387725927010E-04_dp, &
         4.9641303680359998E+00_dp, -4.2791867262695549E+00_dp, 7.0701632981021489E-01_dp, &
         7.6974697844881379E-01_dp, -2.5050637135588621E-01_dp, -1.8674240141839671E-01_dp, &
         8.5004325417316742E-02_dp, 4.9721733708245786E-02_dp, -2.5721746300345582E-02_dp, &
         -2.5577126368071773E-02_dp, 2.5694288079317958E-02_dp, -7.5350930197050558E-03_dp, &
         4.1016625473918824E-04_dp, 9.0428782023867402E+00_dp, -1.0651609613739524E+01_dp, &
         3.8210533639574078E+00_dp, 1.4421513328672826E+00_dp, -1.2962088562797218E+00_dp, &
         -2.1856889727324386E-01_dp, 4.0769349474573302E-01_dp, 2.1778197894256581E-02_dp, &
         -1.2217423363358416E-01_dp, -1.8092159326321852E-02_dp, 7.8958894324477696E-02_dp, &
         -4.2323525382300678E-02_dp, 7.8988389506311936E-03_dp, 1.7494968543499823E+01_dp, &
         -2.6299261098424981E+01_dp, 1.4981092153446502E+01_dp, 3.5566014675930263E-01_dp, &
         -4.4078347265838902E+00_dp, 9.6789361825124498E-01_dp, 1.1548302407139202E+00_dp, &
         -4.8295531452813190E-01_dp, -3.0638711477524511E-01_dp, 1.9053711104052087E-01_dp, &
         9.8530328249180302E-02_dp, -1.1445401656369447E-01_dp, 3.0271989036667647E-02_dp, &
         3.5912642657867430E+01_dp, -6.5373617495328972E+01_dp, 5.2339797574661134E+01_dp, &
         -1.4153218765640851E+01_dp, -9.0572445032884144E+00_dp, 7.8281386285875207E+00_dp, &
         3.6341723094328376E-01_dp, -2.5273798563351466E+00_dp, 3.1250363833851924E-01_dp, &
         1.0318273705347971E+00_dp, -6.6189452520406300E-01_dp, 1.1809550079304761E-01_dp, &
         1.0806151322463294E-02_dp, 8.3477139318733393E+01_dp, -1.7156836453569142E+02_dp, &
         1.7336798683819080E+02_dp, -9.2344665032728145E+01_dp, 9.9138773578404322E+00_dp, &
         2.1111102738858747E+01_dp, -1.2703114305540435E+01_dp, -6.8491988017260363E-01_dp, &
         3.7238467452567474E+00_dp, -9.6853869036389029E-01_dp, -8.2076794251227392E-01_dp, &
         6.5200581790559620E-01_dp, -1.4892747075362087E-01_dp, 3.0901806252841146E+02_dp, &
         -6.4535118104996093E+02_dp, 7.0059421765500099E+02_dp, -4.6233013645500171E+02_dp, &
         1.7229734254873787E+02_dp, -1.0791044680098347E+01_dp, -2.2384185132904548E+01_dp, &
         8.6880396981906074E+00_dp, 4.8992454334398755E-01_dp, -2.1707249656643718E-01_dp, &
         -1.0500007845801604E+00_dp, 7.1848860832775263E-01_dp, -1.5485197147846710E-01_dp, &
         3.3508729161701295E-01_dp, -6.5967375280714266E-02_dp, -3.6671191882035471E-02_dp, &
         3.7463544686884482E-03_dp, 8.5622563668911605E-03_dp, 6.8363383574831626E-04_dp, &
         -4.5035979562529508E-03_dp, 6.6279905673241254E-04_dp, 1.5024716048307220E-03_dp, &
         -1.2576817257599221E-04_dp, -1.0433416780076630E-03_dp, 7.2426679548123546E-04_dp, &
         -1.6273288963037326E-04_dp, 1.1864553104914806E+00_dp, -3.1461997524782426E-01_dp, &
         -1.4693892529281682E-01_dp, 4.5646442517946223E-02_dp, 4.1088276571797845E-02_dp, &
         -7.6528941123103732E-03_dp, -2.4112522507619763E-02_dp, 8.1014486563994372E-03_dp, &
         9.9635648532678794E-03_dp, -5.5731398359707141E-03_dp, -3.2969694635874730E-03_dp, &
         3.7600953456707166E-03_dp, -9.9587115962167325E-04_dp, 2.7324626612556053E+00_dp, &
         -1.0908886663262836E+00_dp, -3.7452568962784560E-01_dp, 2.7448650188755308E-01_dp, &
         1.1947510898542109E-01_dp, -8.1512972680725215E-02_dp, -8.2669082784466011E-02_dp, &
         5.5843294460943604E-02_dp, 4.4965144102735111E-02_dp, -5.5709333452207237E-02_dp, &
         1.2128490666523937E-02_dp, 7.2438183397116164E-03_dp, -3.2102466121958495E-03_dp, &
         6.0892254495015932E+00_dp, -3.6163036866884251E+00_dp, -7.2954415660089122E-01_dp, &
         1.2449867999159732E+00_dp, 1.8627074862998558E-01_dp, -4.5836716242884368E-01_dp, &
         -1.8726612737773107E-01_dp, 2.9412452507008080E-01_dp, 1.4464334222064024E-01_dp, &
         -3.5797443880437330E-01_dp, 2.1243449070641837E-01_dp, -4.7979547413800436E-02_dp, &
         1.1430581477758399E-03_dp, 1.4139026562869708E+01_dp, -1.2041501426557486E+01_dp, &
         -3.5767048473142754E-01_dp, 4.9535087968389506E+00_dp, -6.9859714731516531E-01_dp, &
         -1.9604651106042494E+00_dp, 8.7491811796409316E-02_dp, 1.3013744792257471E+00_dp, &
         5.7319425555596631E-02_dp, -1.5953469275658381E+00_dp, 1.4745984158804955E+00_dp, &
         -5.9602490347273818E-01_dp, 9.6024900648760458E-02_dp, 3.5251429422184806E+01_dp, &
         -4.1992637583344994E+01_dp, 8.2569840030412731E+00_dp, 1.7766017038241735E+01_dp, &
         -9.9409207406392266E+00_dp, -5.7296196204934073E+00_dp, 4.4447226214233817E+00_dp, &
         4.3695801257258298E+00_dp, -4.0443214386948814E+00_dp, -2.8548551132553746E+00_dp, &
         5.6567089023402772E+00_dp, -3.2060726305631433E+00_dp, 6.7793233648081430E-01_dp, &
         9.7332609213038637E+01_dp, -1.6055678095075575E+02_dp, 8.3326641397475115E+01_dp, &
         4.5224931903958627E+01_dp, -7.6146167520984406E+01_dp, 1.3010701305137788E+01_dp, &
         2.8386463320181605E+01_dp, -1.6702334571147881E+00_dp, -3.5079103558906922E+01_dp, &
         3.3344889931611391E+01_dp, -1.1144196235639770E+01_dp, -6.2677207691225156E-01_dp, &
         9.4490594623257140E-01_dp, 3.2494740113105559E+02_dp, -7.2364022403862566E+02_dp, &
         7.0784514785353440E+02_dp, -2.1766148619497469E+02_dp, -2.8011046892812595E+02_dp, &
         3.9157870616328319E+02_dp, -2.0111436008133452E+02_dp, 2.6746934497255534E+01_dp, &
         -5.1250779800289745E+00_dp, 4.5207247760493551E+01_dp, -4.9344510991644917E+01_dp, &
         2.2933203884729071E+01_dp, -4.2463467974507454E+00_dp, 2.7924787903755193E+00_dp, &
         -5.9557486085327938E-01_dp, -3.1708788737155785E-01_dp, 4.8363147353357339E-02_dp, &
         7.8353097117355233E-02_dp, 9.0057144008685636E-04_dp, -4.2436219791212565E-02_dp, &
         8.2581734545659341E-03_dp, 1.5009431826959475E-02_dp, -3.1333769326353706E-03_dp, &
         -9.0848760047894626E-03_dp, 6.8972370390152915E-03_dp, -1.6104367213749559E-03_dp, &
         4.3349564546588484E+00_dp, -1.6338639472670711E+00_dp, -6.2039446459744241E-01_dp, &
         3.7176531406914165E-01_dp, 2.0468278416324942E-01_dp, -9.7117001036191072E-02_dp, &
         -1.3346018024426126E-01_dp, 6.9248942972530952E-02_dp, 6.5294776930919118E-02_dp, &
         -5.9636007438036251E-02_dp, -2.7144286812367025E-03_dp, 1.8580653270522128E-02_dp, &
         -5.8846609276546023E-03_dp, 8.7127006245330918E+00_dp, -5.2814575268392803E+00_dp, &
         -1.2377871952109918E+00_dp, 1.7721827561018202E+00_dp, 4.1504199215701543E-01_dp, &
         -6.2649087632354694E-01_dp, -3.5476660021691170E-01_dp, 3.9492435289379613E-01_dp, &
         2.4669057876623324E-01_dp, -4.5374450788098053E-01_dp, 2.0534450489776798E-01_dp, &
         -1.4337629279395947E-02_dp, -9.5633384489265550E-03_dp, 1.9864553062133062E+01_dp, &
         -1.7268416697475640E+01_dp, -1.2605611184415808E+00_dp, 7.1361682987646899E+00_dp, &
         -2.8155468494418773E-01_dp, -2.8885437478337219E+00_dp, -3.2115843314446185E-01_dp, &
         1.8474192462862677E+00_dp, 5.0459240430275332E-01_dp, -2.4256645534485304E+00_dp, &
         1.8935550944143869E+00_dp, -6.5130637173581107E-01_dp, 8.4425479665035150E-02_dp, &
         4.9415943100591043E+01_dp, -5.8710018731217680E+01_dp, 7.6100350158161136E+00_dp, &
         2.6437526205424952E+01_dp, -1.0210049310126077E+01_dp, -1.0364129503643207E+01_dp, &
         4.2703723895971040E+00_dp, 7.2550799233044732E+00_dp, -2.8007257975971140E+00_dp, &
         -7.9321546402115857E+00_dp, 9.8731611373346890E+00_dp, -4.8000107616726417E+00_dp, &
         9.1625956317232737E-01_dp, 1.3509570103059647E+02_dp, -2.1615670162551694E+02_dp, &
         8.7042223042420432E+01_dp, 8.5074953085441251E+01_dp, -8.7680431005856249E+01_dp, &
         -1.4573782688795355E+01_dp, 4.1553388972824948E+01_dp, 1.5857868513717005E+01_dp, &
         -4.3825384512915626E+01_dp, 1.1242194914554007E+01_dp, 1.9564845495143668E+01_dp, &
         -1.6811781491598083E+01_dp, 4.2614301447710945E+00_dp, 4.2182283329260298E+02_dp, &
         -9.0232374730691106E+02_dp, 7.1913307941971050E+02_dp, 6.5788926678577425E+01_dp, &
         -5.6847305624230580E+02_dp, 3.2522974861279704E+02_dp, 1.3143394779042333E+02_dp, &
         -1.4623977249362667E+02_dp, -1.9485496537035539E+02_dp, 3.9309282560397878E+02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/-2.8863592203480596E+02_dp, &
         1.0528430973221714E+02_dp, -1.5857007223981109E+01_dp, 1.8276946029107571E+03_dp, &
         -4.8112032981737193E+03_dp, 6.1258592600600687E+03_dp, -4.0381542724034271E+03_dp, &
         3.2440486268768652E+01_dp, 2.9756759864086989E+03_dp, -3.5826087401832801E+03_dp, &
         2.5220758904019117E+03_dp, -1.1725016022892619E+03_dp, 2.9714869827611420E+02_dp, &
         3.6408367716617519E+01_dp, -6.1718990215158165E+01_dp, 1.6819583425152295E+01_dp, &
         3.6008296661068129E-01_dp, -2.4934495356088673E-02_dp, -2.6462026390146498E-02_dp, &
         9.5876084101441991E-03_dp, -1.1331107464939878E-02_dp, 1.6586262055557337E-02_dp, &
         -1.6106632470944200E-02_dp, 1.2751304928977305E-02_dp, -8.6732664212007485E-03_dp, &
         4.0240760309193475E-03_dp, -4.4825706457146837E-04_dp, -6.1748144054248283E-04_dp, &
         2.5357640280289075E-04_dp, 1.3098679540438654E+00_dp, -1.2749258837080399E-01_dp, &
         -1.3073323485317337E-01_dp, 5.8516929367271715E-02_dp, -5.6951328739513037E-02_dp, &
         8.4819766678372446E-02_dp, -8.3438172999370994E-02_dp, 6.3534150436398582E-02_dp, &
         -4.0789452590645821E-02_dp, 1.7207516063066464E-02_dp, 2.7854405286715241E-05_dp, &
         -4.2694378230801753E-03_dp, 1.5551774834561508E-03_dp, 3.1817466966750025E+00_dp, &
         -4.8685729139374834E-01_dp, -4.7508372620822747E-01_dp, 2.7226827027853312E-01_dp, &
         -2.1644926109007404E-01_dp, 3.2162422597469387E-01_dp, -3.2373129824840324E-01_dp, &
         2.3402591064647971E-01_dp, -1.3552991687865093E-01_dp, 4.6569436455071500E-02_dp, &
         1.2573171658256766E-02_dp, -2.1996218996840747E-02_dp, 7.2391100075863378E-03_dp, &
         7.6667308457077841E+00_dp, -1.8064423526176003E+00_dp, -1.6558509837359154E+00_dp, &
         1.2167933969012499E+00_dp, -8.2685269049351195E-01_dp, 1.1702629066210879E+00_dp, &
         -1.2160740673457533E+00_dp, 8.3129428882777301E-01_dp, -4.0346334453079419E-01_dp, &
         7.6181979654980650E-02_dp, 1.0538509979804817E-01_dp, -1.0594678252053578E-01_dp, &
         3.2100908588300212E-02_dp, 1.9789605527050799E+01_dp, -6.9317551385828775E+00_dp, &
         -5.8179930349634565E+00_dp, 5.6410261132198896E+00_dp, -3.4720604618155404E+00_dp, &
         4.3222477406821946E+00_dp, -4.6593280205438399E+00_dp, 2.9973049622920191E+00_dp, &
         -9.8458606194260101E-01_dp, -3.0033580201486476E-01_dp, 7.5698139555564226E-01_dp, &
         -5.3686026095251194E-01_dp, 1.4785400367332710E-01_dp, 5.7071028866399622E+01_dp, &
         -2.9401869129872736E+01_dp, -2.1376956390611706E+01_dp, 2.9364241928690973E+01_dp, &
         -1.8013055669371493E+01_dp, 1.7254148374056300E+01_dp, -1.8753111107981081E+01_dp, &
         1.1159554233110121E+01_dp, -1.2752966385306053E-01_dp, -6.1325355952629250E+00_dp, &
         6.3822854246393721E+00_dp, -3.4517625645107035E+00_dp, 8.3020380724742204E-01_dp, &
         1.9521201133501947E+02_dp, -1.5220438343077086E+02_dp, -8.1912358679862876E+01_dp, &
         1.9161075517584456E+02_dp, -1.3980358819312374E+02_dp, 8.7370038717781611E+01_dp, &
         -7.2988008432725636E+01_dp, 3.0136598201372607E+01_dp, 4.4492815957746984E+01_dp, &
         -9.0138493143008020E+01_dp, 7.8015315638660624E+01_dp, -3.7190554397321939E+01_dp, &
         7.9761307531966610E+00_dp, 9.1278852097032075E+02_dp, -1.1881974595454724E+03_dp, &
         -1.2420992251855267E+02_dp, 1.8801289344962393E+03_dp, -2.4556997768600449E+03_dp, &
         1.5365426797607518E+03_dp, 1.0676304658797669E+02_dp, -1.6868122887497107E+03_dp, &
         2.7261083335976618E+03_dp, -2.8690190040817129E+03_dp, 2.0730030060640806E+03_dp, &
         -9.2805090617485291E+02_dp, 1.9285336281750696E+02_dp, 3.0202120791487510E+00_dp, &
         -2.2929050737844739E-01_dp, -2.4117235349996635E-01_dp, 9.2704544171716446E-02_dp, &
         -1.0360124397461544E-01_dp, 1.5260718160236281E-01_dp, -1.4863649841980880E-01_dp, &
         1.1646276459958013E-01_dp, -7.8115523264564363E-02_dp, 3.5445051921955989E-02_dp, &
         -3.0265424876823456E-03_dp, -6.2106108531044283E-03_dp, 2.4560375548071142E-03_dp, &
         4.9977115377020143E+00_dp, -7.0713497336708797E-01_dp, -7.0201845826538045E-01_dp, &
         3.7205228555849740E-01_dp, -3.1239914375103917E-01_dp, 4.6962581162570477E-01_dp, &
         -4.6809922294208084E-01_dp, 3.4405231198789693E-01_dp, -2.0774765511908164E-01_dp, &
         7.7981264100472250E-02_dp, 1.1926822036550332E-02_dp, -2.9234106504804597E-02_dp, &
         9.9111598962051163E-03_dp, 1.0982352744981107E+01_dp, -2.5586757519844334E+00_dp, &
         -2.3924834863157098E+00_dp, 1.6377235454994263E+00_dp, -1.1452343882258973E+00_dp, &
         1.6742775852013179E+00_dp, -1.7208177689510733E+00_dp, 1.1933688801441626E+00_dp, &
         -6.2068748093326198E-01_dp, 1.5757464118201009E-01_dp, 1.2092332567847532E-01_dp, &
         -1.3968280008702197E-01_dp, 4.3442636832466526E-02_dp, 2.7798834753362176E+01_dp, &
         -9.5235310673252087E+00_dp, -8.2506504296193413E+00_dp, 7.3158471930055118E+00_dp, &
         -4.5365544170813825E+00_dp, 6.0400318657414305E+00_dp, -6.4510290250044848E+00_dp, &
         4.2170504980565306E+00_dp, -1.6565364294458416E+00_dp, -6.7984032425921506E-02_dp, &
         8.3432875689469077E-01_dp, -6.6751055992496788E-01_dp, 1.9155267749846455E-01_dp, &
         7.8875815934450927E+01_dp, -3.8326646554760494E+01_dp, -2.9647322447318992E+01_dp, &
         3.5592030615395316E+01_dp, -2.1068887318345901E+01_dp, 2.3036574856677053E+01_dp, &
         -2.5457687546678919E+01_dp, 1.5603470262769573E+01_dp, -2.6161630333998538E+00_dp, &
         -4.7974113923994901E+00_dp, 6.0888371971161597E+00_dp, -3.6509041918220748E+00_dp, &
         9.3827458104319261E-01_dp, 2.5820747118786790E+02_dp, -1.7912518679501466E+02_dp, &
         -1.1382161244744694E+02_dp, 2.0430275857235807E+02_dp, -1.3010643642425052E+02_dp, &
         1.0051327222541408E+02_dp, -1.0601266064614593E+02_dp, 5.8288692345255235E+01_dp, &
         2.1127565037978961E+01_dp, -6.5273552596841654E+01_dp, 5.8416007114044838E+01_dp, &
         -2.8627550070207121E+01_dp, 6.3938023439562945E+00_dp, 1.0530804258036487E+03_dp, &
         -1.0937614495281734E+03_dp, -4.3098080435867519E+02_dp, 1.5667138220310976E+03_dp, &
         -1.3165673699354647E+03_dp, 6.6347634454225249E+02_dp, -3.4185303502890736E+02_dp, &
         -1.1466724937319730E+01_dp, 6.1723480049892203E+02_dp, -1.0180911237826692E+03_dp, &
         8.6238446090989885E+02_dp, -4.0590511772790057E+02_dp, 8.5233625649272710E+01_dp, &
         6.6217329112409152E+03_dp, -1.1425351525376182E+04_dp, 2.2492305374254370E+03_dp, &
         1.8881455299767505E+04_dp, -3.4404996456886765E+04_dp, 2.7559397466881259E+04_dp, &
         2.4019070309659278E+03_dp, -3.9817790294105704E+04_dp, 6.4303769223224357E+04_dp, &
         -6.3967990080210388E+04_dp, 4.2899287320689975E+04_dp, -1.7914528413855718E+04_dp, &
         3.5182149345345074E+03_dp, 3.6486926164490469E-01_dp, 2.0204681876392647E-08_dp, &
         -2.9063033115982715E-02_dp, 2.5307089232711903E-02_dp, -1.7387851050024779E-02_dp, &
         -1.1298488481693316E-02_dp, 6.1784825098588868E-02_dp, -1.1326788530175186E-01_dp, &
         1.2736917040891757E-01_dp, -8.2068207035220206E-02_dp, 1.1361290138449257E-02_dp, &
         2.2817399150212796E-02_dp, -1.2307392917240636E-02_dp, 1.3345106038416172E+00_dp, &
         1.0473100168474207E-07_dp, -1.5067373422508695E-01_dp, 1.3120139232313632E-01_dp, &
         -8.3771660636503986E-02_dp, -6.9708760654556362E-02_dp, 3.3279408293031759E-01_dp, &
         -5.9029754060722750E-01_dp, 6.3961521945451127E-01_dp, -3.7704261134121214E-01_dp, &
         1.4222371809919012E-03_dp, 1.5638278661184743E-01_dp, -7.4976139536898731E-02_dp, &
         3.2767572317506888E+00_dp, 4.0757791931997376E-07_dp, -5.8650748464803737E-01_dp, &
         5.1070951349780291E-01_dp, -2.9175849747269827E-01_dp, -3.3130653070655985E-01_dp, &
         1.3602786476515449E+00_dp, -2.3079473886452635E+00_dp, 2.3670793977651505E+00_dp, &
         -1.1932637133271387E+00_dp, -3.1535910527251165E-01_dp, 8.1970673219695112E-01_dp, &
         -3.5341210513182930E-01_dp, 8.0233464032319297E+00_dp, 1.5469932072109146E-06_dp, &
         -2.2267438908414117E+00_dp, 1.9389654319071237E+00_dp, -9.5066660364792799E-01_dp, &
         -1.5320952612124052E+00_dp, 5.4440801207413720E+00_dp, -8.7625182894010312E+00_dp, &
         8.3459802828849448E+00_dp, -3.1789409559564277E+00_dp, -2.7443202471077637E+00_dp, &
         4.1173052329751831E+00_dp, -1.6328471027246423E+00_dp, 2.1178994737488193E+01_dp, &
         6.1163444588653944E-06_dp, -8.8069616938961222E+00_dp, 7.6687600580722854E+00_dp, &
         -2.9352571881725296E+00_dp, -7.4996883609505600E+00_dp, 2.2879137403427048E+01_dp, &
         -3.4324990064612756E+01_dp, 2.9082103747106494E+01_dp, -4.8140937848072012E+00_dp, &
         -1.9508713405196168E+01_dp, 2.1824003337870533E+01_dp, -8.0461535482045914E+00_dp, &
         6.3097771483862608E+01_dp, 2.7109366339854733E-05_dp, -3.9053754979821946E+01_dp, &
         3.4006411809445268E+01_dp, -7.5765986227444024E+00_dp, -4.2752431653712904E+01_dp, &
         1.0918545502277176E+02_dp, -1.4682601957283225E+02_dp, 9.7806062121677598E+01_dp, &
         3.5785765692400041E+01_dp, -1.4814419919148372E+02_dp, 1.3542279919125832E+02_dp, &
         -4.6600546953699642E+01_dp, 2.2763840675790064E+02_dp, 1.5137505265291369E-04_dp, &
         -2.1821644080557203E+02_dp, 1.9001333141048849E+02_dp, 1.1018780585175488E+01_dp, &
         -3.3195327790129170E+02_dp, 6.6654678166863403E+02_dp, -7.1329548357082911E+02_dp, &
         1.5569450837479587E+02_dp, 8.4955015710139378E+02_dp, -1.4931577695169062E+03_dp, &
         1.1595826219001069E+03_dp, -3.7132444207641549E+02_dp, 1.1910914431216361E+03_dp, &
         1.4263721165942622E-03_dp, -2.0551073023908939E+03_dp, 1.7895023674272838E+03_dp, &
         1.4236436615084440E+03_dp, -5.4213598658487108E+03_dp, 6.4193939031260470E+03_dp, &
         -4.3652668119034223E+02_dp, -1.3307350777633279E+04_dp, 2.8004934478322662E+04_dp, &
         -3.1609945290997650E+04_dp, 2.0246981620964787E+04_dp, -5.7903787141059511E+03_dp, &
         3.0643056444925736E+00_dp, 1.8646661584793930E-07_dp, -2.6823124660733616E-01_dp, &
         2.3356648801342425E-01_dp, -1.5749179181334688E-01_dp, -1.0949264322249078E-01_dp, &
         5.7614538690757122E-01_dp, -1.0470107757684963E+00_dp, 1.1661521479359405E+00_dp, &
         -7.3514781124312867E-01_dp, 7.8269126791499333E-02_dp, 2.2825560761526997E-01_dp, &
         -1.1877863457204990E-01_dp, 5.1352531685543763E+00_dp, 5.8811945999022206E-07_dp, &
         -8.4624001750433542E-01_dp, 7.3687550809193658E-01_dp, -4.3833952146723459E-01_dp, &
         -4.4767431890582660E-01_dp, 1.9312272697015882E+00_dp, -3.3286199469950279E+00_dp, &
         3.4839435414394124E+00_dp, -1.8685419680693067E+00_dp, -2.8596457697133809E-01_dp, &
         1.0725167156433020E+00_dp, -4.7794242799465891E-01_dp, 1.1485629169810354E+01_dp, &
         2.1753987536946991E-06_dp, -3.1309987505895300E+00_dp, 2.7263579761484240E+00_dp, &
         -1.4087375048075199E+00_dp, -2.0285000971994687E+00_dp, 7.5358003620778735E+00_dp, &
         -1.2345857753739226E+01_dp, 1.2071996259175435E+01_dp, -5.1411109695351618E+00_dp, &
         -3.1071844239563102E+00_dp, 5.3071010378293604E+00_dp, -2.1574669369328108E+00_dp, &
         2.9697399912139918E+01_dp, 8.3135628316181069E-06_dp, -1.1969267103599444E+01_dp, &
         1.0422379107290249E+01_dp, -4.4016512622802200E+00_dp, -9.4725197213611363E+00_dp, &
         3.0480241673650344E+01_dp, -4.6979770727514214E+01_dp, 4.1763601254545264E+01_dp, &
         -1.0744664685407409E+01_dp, -2.1930155353081439E+01_dp, 2.6765835808358442E+01_dp, &
         -1.0113169118847527E+01_dp, 8.6657911400504887E+01_dp, 3.4682726546868686E-05_dp, &
         -4.9954155938197879E+01_dp, 4.3498076605632711E+01_dp, -1.2732165518503718E+01_dp, &
         -4.9378251472346385E+01_dp, 1.3584067834177924E+02_dp, -1.9220755204893482E+02_dp, &
         1.4478597281785360E+02_dp, 1.1301880394535623E+01_dp, -1.5313609799645960E+02_dp, &
         1.5075591149117258E+02_dp, -5.3325987923237058E+01_dp, 2.9560124699329253E+02_dp, &
         1.7113342586626282E-04_dp, -2.4662390250801528E+02_dp, 2.1474962274139023E+02_dp, &
         -2.0392071197782663E+01_dp, -3.1789596824276714E+02_dp, 7.2496736168975849E+02_dp/)
      REAL(KIND=dp), DIMENSION(13, 16, 4), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03/), (/13, 16, 4/))

      INTEGER                                            :: irange

      IF (Rc <= 1.0000000000000000E+01_dp) Rc = 1.0000000000000000E+01_dp
      IF (Rc >= 6.3447246551513672E+03_dp) Rc = 6.3447246551513672E+03_dp
      IF (Rc >= 1.0000000000000000E+01_dp .AND. Rc <= 4.8301780076010026E+01_dp) THEN
         irange = 1
         L_b = 1.0000000000000000E+01_dp
         U_b = 4.8301780076010026E+01_dp
      END IF
      IF (Rc >= 4.8301780076010026E+01_dp .AND. Rc <= 4.3944134912624889E+02_dp) THEN
         irange = 2
         L_b = 4.8301780076010026E+01_dp
         U_b = 4.3944134912624889E+02_dp
      END IF
      IF (Rc >= 4.3944134912624889E+02_dp .AND. Rc <= 2.4651524781467556E+03_dp) THEN
         irange = 3
         L_b = 4.3944134912624889E+02_dp
         U_b = 2.4651524781467556E+03_dp
      END IF
      IF (Rc >= 2.4651524781467556E+03_dp .AND. Rc <= 6.3447246551513672E+03_dp) THEN
         irange = 4
         L_b = 2.4651524781467556E+03_dp
         U_b = 6.3447246551513672E+03_dp
      END IF

      fit_coef(1:13, 1:16) = coefdata(1:13, 1:16, irange)

   END SUBROUTINE get_fit_coef_k08

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k09(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 18)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(136), PARAMETER :: c03 = (/-8.4609670938814907E+02_dp, &
         5.2521889455815938E+02_dp, -1.8158803681882837E+01_dp, -1.9195843056246903E+02_dp, &
         4.7038182235191250E+01_dp, 4.4962618821808661E+01_dp, 2.3952550076762327E+03_dp, &
         7.3919658606953450E-04_dp, -3.3678854403725500E+03_dp, 2.7630103823335253E+03_dp, &
         1.5760188487094081E+03_dp, -6.2507126178305934E+03_dp, 7.5762009915563758E+03_dp, &
         -3.4997817421370810E+03_dp, -3.6562358282784421E+03_dp, 6.5885187381838932E+03_dp, &
         -1.0669477891854467E+03_dp, -5.1923485018358997E+03_dp, 3.6296377875939629E+03_dp, &
         2.8852081079160179E+00_dp, 4.0102719961273801E-08_dp, -1.7659002023480172E-01_dp, &
         1.4488106202271947E-01_dp, -1.0039965813412362E-01_dp, -2.5605452610356936E-02_dp, &
         2.2190893738774328E-01_dp, -3.9868714638912034E-01_dp, 4.1666545240789737E-01_dp, &
         -2.0838248751367475E-01_dp, -7.8702085558092494E-02_dp, 1.8499348107948621E-01_dp, &
         -8.5003584077023686E-02_dp, 4.5954934072634712E+00_dp, 1.1523486520096986E-07_dp, &
         -5.0806402051936073E-01_dp, 4.1683398596614762E-01_dp, -2.6579259191602511E-01_dp, &
         -1.1173061386143904E-01_dp, 6.8157899224885998E-01_dp, -1.1725048395640862E+00_dp, &
         1.1888726238653182E+00_dp, -5.7031447220490772E-01_dp, -2.3373683427653494E-01_dp, &
         5.0947643811010013E-01_dp, -2.2753724549285903E-01_dp, 9.5744538319119368E+00_dp, &
         3.8935451948906365E-07_dp, -1.7187840838704733E+00_dp, 1.4101497165394266E+00_dp, &
         -8.2424578220474254E-01_dp, -5.0165771967854866E-01_dp, 2.4411624004207271E+00_dp, &
         -4.0381205898500108E+00_dp, 3.9746616048513781E+00_dp, -1.8240116710035414E+00_dp, &
         -8.1274767364034362E-01_dp, 1.6405984125323814E+00_dp, -7.0946369768700668E-01_dp, &
         2.2756152830064220E+01_dp, 1.3380012629757024E-06_dp, -5.9153103075446083E+00_dp, &
         4.8531141769756569E+00_dp, -2.5426282738934458E+00_dp, -2.2119329262709027E+00_dp, &
         8.9099642025165444E+00_dp, -1.4124325804320934E+01_dp, 1.3421000524921119E+01_dp, &
         -5.8143931761506558E+00_dp, -2.8750831570457684E+00_dp, 5.2778875665252425E+00_dp, &
         -2.1815020322245369E+00_dp, 5.9577031872248192E+01_dp, 4.8466710971685504E-06_dp, &
         -2.1470062487220464E+01_dp, 1.7614691738721081E+01_dp, -7.8543518759744195E+00_dp, &
         -1.0297543487661295E+01_dp, 3.4589595106069012E+01_dp, -5.2023617051867674E+01_dp, &
         4.7104796520358647E+01_dp, -1.8668699591613251E+01_dp, -1.0751139468718122E+01_dp, &
         1.7201278360974712E+01_dp, -6.5744756865732210E+00_dp, 1.7421142487352890E+02_dp, &
         1.9480399941922716E-05_dp, -8.6547598834693005E+01_dp, 7.1005987098306903E+01_dp, &
         -2.3983059370901209E+01_dp, -5.4191420793490103E+01_dp, 1.5104757590596600E+02_dp, &
         -2.1163527408094700E+02_dp, 1.7786418457885577E+02_dp, -5.9642025337733756E+01_dp, &
         -4.4718701051976034E+01_dp, 5.6683066711160087E+01_dp, -1.8062035214880027E+01_dp, &
         5.9442761921776025E+02_dp, 9.3312420949589195E-05_dp, -4.1653329002850757E+02_dp, &
         3.4173289748064514E+02_dp, -5.8910745159325486E+01_dp, -3.5416677215144733E+02_dp, &
         8.0179526534346644E+02_dp, -1.0066258052744280E+03_dp, 7.3393741106438824E+02_dp, &
         -1.5060845806689406E+02_dp, -2.2080860417911666E+02_dp, 1.6107768368211609E+02_dp, &
         -1.4997422612735614E+01_dp, 2.6003372356118971E+03_dp, 6.1017236225054279E-04_dp, &
         -2.7485887342448536E+03_dp, 2.2549733577908974E+03_dp, 2.9333015363387631E+02_dp, &
         -3.4638286965522780E+03_dp, 5.9673723572603731E+03_dp, -5.9287627550561756E+03_dp, &
         2.6547680091869743E+03_dp, 1.0828729490854735E+03_dp, -1.4147281790223578E+03_dp, &
         -6.5594324803846439E+02_dp, 9.5647367522013610E+02_dp, 1.9075646404982639E+04_dp, &
         7.9756724721310873E-03_dp, -3.6633955682749525E+04_dp, 3.0054168139170702E+04_dp, &
         3.0244619250697619E+04_dp, -8.9553753522321902E+04_dp, 7.9875735584514681E+04_dp, &
         1.7524458945017868E+04_dp, -1.3188240705569120E+05_dp, 1.3592402872600380E+05_dp, &
         -1.4644005644830598E+01_dp, -1.0996321547256157E+05_dp, 6.7843413283968956E+04_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/2.7491556688515650E-01_dp, &
         -1.0138397977284885E-01_dp, -3.9049726894672010E-02_dp, 1.2625689412991524E-02_dp, &
         1.2491380414983764E-02_dp, -3.7765616302317911E-03_dp, -5.1175283467075521E-03_dp, &
         1.4632298825565039E-03_dp, 3.6642955364173022E-03_dp, -2.0156097072982762E-03_dp, &
         -1.1372055647295432E-03_dp, 1.3123912090724280E-03_dp, -3.4398637877680914E-04_dp, &
         9.2040539096736784E-01_dp, -4.1664024852120668E-01_dp, -1.1285853825987181E-01_dp, &
         7.9706962779770921E-02_dp, 3.7420429120305111E-02_dp, -3.0919600141887049E-02_dp, &
         -1.5284758454925678E-02_dp, 1.2652127413759585E-02_dp, 1.7202917438349393E-02_dp, &
         -2.2424746152411607E-02_dp, 7.9553605887680342E-03_dp, 5.1972329893920990E-04_dp, &
         -6.8699382298445435E-04_dp, 1.9006321915664388E+00_dp, -1.1695348091685061E+00_dp, &
         -1.2189417451274155E-01_dp, 3.0841663498130323E-01_dp, 2.8149207201224088E-02_dp, &
         -1.3459045672526931E-01_dp, -4.2172084983072775E-03_dp, 5.7470448789934472E-02_dp, &
         3.9892748401996053E-02_dp, -1.0792452192241736E-01_dp, 7.7031387642835639E-02_dp, &
         -2.4799303141627441E-02_dp, 2.9641145262798221E-03_dp, 3.6226601304189745E+00_dp, &
         -3.0653465294668503E+00_dp, 2.3997996878917224E-01_dp, 9.2020850811311550E-01_dp, &
         -2.1053920858265254E-01_dp, -4.0407144020499658E-01_dp, 1.4478946512027779E-01_dp, &
         1.7808216466386911E-01_dp, -6.7382162449458046E-05_dp, -2.9692981490093184E-01_dp, &
         3.0359529933075358E-01_dp, -1.3377430156320294E-01_dp, 2.3545839240890390E-02_dp, &
         6.8829600678148175E+00_dp, -7.8109151543133537E+00_dp, 2.1524611550975314E+00_dp, &
         2.1963549588737723E+00_dp, -1.4083698809615777E+00_dp, -8.1873516409798242E-01_dp, &
         8.3000971643672017E-01_dp, 3.4924147841610120E-01_dp, -4.6131744612525610E-01_dp, &
         -3.2768291484595113E-01_dp, 6.7712900531194742E-01_dp, -3.8698623980190255E-01_dp, &
         8.1552378085722771E-02_dp, 1.3361445315337988E+01_dp, -1.9720693805354671E+01_dp, &
         9.8170523591388061E+00_dp, 3.6112843572095294E+00_dp, -5.7557505888315017E+00_dp, &
         -2.6911215231012703E-01_dp, 2.9734538015789798E+00_dp, -1.1080337220770373E-01_dp, &
         -2.2704004653913077E+00_dp, 1.3354167923098297E+00_dp, 2.7827788457258407E-01_dp, &
         -5.3396440146110924E-01_dp, 1.5485305521629095E-01_dp, 2.6868227237507195E+01_dp, &
         -5.0102045492014518E+01_dp, 3.7811908718509507E+01_dp, -1.8294630056592489E+00_dp, &
         -1.7506470965502057E+01_dp, 7.8480672911596105E+00_dp, 6.3178773982403840E+00_dp, &
         -4.8161399112143526E+00_dp, -5.2120084586700539E+00_dp, 9.0278934674271571E+00_dp, &
         -5.4400779115770668E+00_dp, 1.4969760611556289E+00_dp, -1.3708613238816467E-01_dp, &
         5.7398225707624746E+01_dp, -1.3068240758584577E+02_dp, 1.3611896593086911E+02_dp, &
         -5.6244554349383378E+01_dp, -2.8041309641109571E+01_dp, 4.3346676847580149E+01_dp, &
         -6.8148749615071447E+00_dp, -1.7910192391133652E+01_dp, 7.4742481485171650E+00_dp, &
         1.1402642814244142E+01_dp, -1.4702415316240296E+01_dp, 7.0796474962768450E+00_dp, &
         -1.3388574138054052E+00_dp, 1.4718735946959035E+02_dp, -3.8655381690361162E+02_dp, &
         5.0264423442263518E+02_dp, -3.6114141158090524E+02_dp, 9.0095327950470534E+01_dp, &
         8.0537937520274326E+01_dp, -7.6164154126460289E+01_dp, 3.8131872647713339E+00_dp, &
         2.5322962717490988E+01_dp, -7.4949103011308527E+00_dp, -9.1658479772166626E+00_dp, &
         7.7135520272691753E+00_dp, -1.8523169694290555E+00_dp, 2.2601006812612785E+00_dp, &
         -8.7997390790539198E-01_dp, -3.1237293051750925E-01_dp, 1.2648961191291591E-01_dp, &
         1.0173399066050458E-01_dp, -4.1985217142443479E-02_dp, -4.2043767126443553E-02_dp, &
         1.6589539959258010E-02_dp, 3.3723886270116563E-02_dp, -2.5599276642292213E-02_dp, &
         -3.3723992759883159E-03_dp, 9.1543230425048736E-03_dp, -2.7035579659107495E-03_dp, &
         3.0558414722769203E+00_dp, -1.8423525687001594E+00_dp, -2.4806358565768091E-01_dp, &
         4.8522067556668447E-01_dp, 8.1780516282716875E-02_dp, -2.1134587830838658E-01_dp, &
         -2.9333397736968799E-02_dp, 8.8869329486145510E-02_dp, 7.9743471514237316E-02_dp, &
         -1.7011594446843087E-01_dp, 1.0584292297514415E-01_dp, -2.7622708244182885E-02_dp, &
         1.9266826647592555E-03_dp, 5.0439918903115455E+00_dp, -4.6147378247056725E+00_dp, &
         3.4606299922222433E-01_dp, 1.4907521779331261E+00_dp, -2.6458521972713844E-01_dp, &
         -6.8256595629361283E-01_dp, 1.8400335770241980E-01_dp, 3.0044010867457915E-01_dp, &
         6.7687204379828558E-02_dp, -5.5503131110925941E-01_dp, 5.1852137763862582E-01_dp, &
         -2.1508674337333056E-01_dp, 3.5696005935841650E-02_dp, 9.2338356558127259E+00_dp, &
         -1.1596541857331431E+01_dp, 3.2196780214282321E+00_dp, 3.6242503908319130E+00_dp, &
         -2.0715462391169384E+00_dp, -1.5043039760553865E+00_dp, 1.2499534478269310E+00_dp, &
         6.7566189217826012E-01_dp, -6.0067926803925731E-01_dp, -8.3969188762166769E-01_dp, &
         1.3289290929674367E+00_dp, -7.1228918020345477E-01_dp, 1.4490578224151213E-01_dp, &
         1.7897430920456454E+01_dp, -2.8993974640936074E+01_dp, 1.4484254763706325E+01_dp, &
         6.4853429598526233E+00_dp, -8.8659405206488398E+00_dp, -1.3072959846374910E+00_dp, &
         4.8716122165178666E+00_dp, 3.7210099832799376E-01_dp, -3.7219944942361369E+00_dp, &
         1.4592913559025713E+00_dp, 1.2136291248627868E+00_dp, -1.2214837337140676E+00_dp, &
         3.1786347913744367E-01_dp, 3.6268902880949476E+01_dp, -7.3109988017877853E+01_dp, &
         5.5198663558559915E+01_dp, 1.5723454940839439E+00_dp, -2.9119283252532057E+01_dp, &
         9.2853577555148661E+00_dp, 1.2848162547779781E+01_dp, -6.1114123559080635E+00_dp, &
         -1.1475135076695258E+01_dp, 1.5069041357505670E+01_dp, -7.1864984784676569E+00_dp, &
         1.1624317623595268E+00_dp, 8.5663539809219230E-02_dp, 7.7586632783688742E+01_dp, &
         -1.8852347266621780E+02_dp, 1.9745424592717342E+02_dp, -6.6990539473087694E+01_dp, &
         -6.4607063534478783E+01_dp, 7.0129963577739574E+01_dp, 5.6788736642728228E+00_dp, &
         -3.8031820871312668E+01_dp, -2.1740311275891591E+00_dp, 4.4736349997022785E+01_dp, &
         -4.2964456895307912E+01_dp, 1.8563951012438558E+01_dp, -3.2717753104214178E+00_dp, &
         1.8607799112651591E+02_dp, -5.1602169316042023E+02_dp, 6.8887347424898587E+02_dp, &
         -4.7458042855350305E+02_dp, 4.4622418554771478E+01_dp, 2.1616014044968463E+02_dp, &
         -1.5545535106352176E+02_dp, -3.2603087830560838E+01_dp, 1.1189461104665806E+02_dp, &
         -6.2453572059122330E+01_dp, 1.4004003528879079E+00_dp, 1.2024016844350291E+01_dp, &
         -3.7743438683099675E+00_dp, 6.9298673553197830E+02_dp, -1.9786469723055493E+03_dp, &
         2.8890520642257911E+03_dp, -2.5529058035051989E+03_dp, 1.2587845207337734E+03_dp, &
         -7.7992454428378821E+01_dp, -3.2321033215482868E+02_dp, 1.6424491980317885E+02_dp, &
         2.4365533617277901E+01_dp, -3.9605274406289432E+01_dp, -6.8896488944108363E+00_dp, &
         1.6649305464354185E+01_dp, -4.9462414102750252E+00_dp, 3.2469980045956409E-01_dp, &
         -5.0884891508484555E-02_dp, -3.0604197193935530E-02_dp, 2.7427469620202637E-03_dp, &
         6.5288788489622235E-03_dp, 1.4751518742325842E-03_dp, -4.4569860753834612E-03_dp, &
         7.3984873623845994E-04_dp, 1.3725182324794782E-03_dp, -1.3330062499350304E-04_dp, &
         -1.0097670883910479E-03_dp, 7.2572714897585022E-04_dp, -1.6788642792246261E-04_dp, &
         1.1375940158990296E+00_dp, -2.3608340938492453E-01_dp, -1.2626283245351805E-01_dp, &
         3.0203907418656358E-02_dp, 3.1785577520873423E-02_dp, 6.6484062921600383E-04_dp, &
         -2.3186399042989306E-02_dp, 6.3183607198872387E-03_dp, 8.6381015194322997E-03_dp, &
         -3.8842596338162639E-03_dp, -3.7122714446579746E-03_dp, 3.6216395853979934E-03_dp, &
         -9.2922534715789357E-04_dp, 2.5662576301481543E+00_dp, -7.8767800972008517E-01_dp, &
         -3.4787162045624048E-01_dp, 1.7519834794011546E-01_dp, 1.0184464478563594E-01_dp, &
         -2.9402771689865572E-02_dp, -8.3714434652696970E-02_dp, 3.7221173571793567E-02_dp, &
         3.9228710671209921E-02_dp, -3.5132688186151159E-02_dp, -7.2078001512086579E-04_dp, &
         1.0273709728764327E-02_dp, -3.3154780668181900E-03_dp, 5.5500130403737691E+00_dp, &
         -2.4995868120429363E+00_dp, -8.4108231466303029E-01_dp, 7.8254760866878970E-01_dp, &
         2.5130274709663863E-01_dp, -2.1094801260314211E-01_dp, -2.5468188988232654E-01_dp, &
         1.8562195286267533E-01_dp, 1.5278152349671439E-01_dp, -2.2434059692085659E-01_dp, &
         8.4198910115564699E-02_dp, 3.7327357965615668E-03_dp, -7.0014770989066926E-03_dp, &
         1.2388533146386802E+01_dp, -7.9005727746800547E+00_dp, -1.6768805892952201E+00_dp, &
         3.1303583041097780E+00_dp, 2.9757366586635242E-01_dp, -1.0543635092197530E+00_dp, &
         -5.9806717900324746E-01_dp, 8.4749241097756289E-01_dp, 4.7566246950522223E-01_dp, &
         -1.1850653348692619E+00_dp, 7.6550203921832072E-01_dp, -2.1047358296363022E-01_dp, &
         1.7042274233523055E-02_dp, 2.9329613292850858E+01_dp, -2.5787172597477593E+01_dp, &
         -1.4204821244535877E+00_dp, 1.1964763351996828E+01_dp, -1.9597065992887659E+00_dp, &
         -4.3627229820786857E+00_dp, -3.9222745547684623E-01_dp, 3.6450051138022639E+00_dp, &
         4.9711724176984340E-01_dp, -4.9839121455852364E+00_dp, 4.5843906987281251E+00_dp, &
         -1.8840025476130395E+00_dp, 3.1143090085006336E-01_dp, 7.5288964164161186E+01_dp, &
         -9.0335134300893770E+01_dp, 1.4213295300102013E+01_dp, 4.4161535993359188E+01_dp, &
         -2.3939960756625741E+01_dp, -1.3093263270166732E+01_dp, 7.9641392095377705E+00_dp, &
         1.3766695386166811E+01_dp, -9.8502878978357327E+00_dp, -1.0558330710285986E+01_dp, &
         1.7789872343339511E+01_dp, -9.7871793631893524E+00_dp, 2.0435013450669803E+00_dp, &
         2.1650557215451059E+02_dp, -3.5651641429679739E+02_dp, 1.6757116494673738E+02_dp, &
         1.3519758005338315E+02_dp, -1.9366619517796931E+02_dp, 2.9213274253505041E+01_dp, &
         6.3389691171209662E+01_dp, 1.8187088316420692E+01_dp, -1.0836900917299248E+02_dp, &
         8.7474533360265895E+01_dp, -2.0470329785844417E+01_dp, -7.7295411530191132E+00_dp, &
         3.9109991716175272E+00_dp, 7.6235912249499665E+02_dp, -1.7181530353164492E+03_dp, &
         1.6377791571835323E+03_dp, -3.3946784295902592E+02_dp, -9.7252903864036193E+02_dp, &
         1.2354270580997018E+03_dp, -6.6463719849297411E+02_dp, 1.4060533750444611E+02_dp, &
         -5.3565858987836350E+01_dp, 1.4823802021219080E+02_dp, -1.4694213881725742E+02_dp, &
         6.5887306625564349E+01_dp, -1.1862205010260631E+01_dp, 2.6990399563726419E+00_dp, &
         -4.5606541750914703E-01_dp, -2.6625736807926959E-01_dp, 3.3786194855342760E-02_dp, &
         5.9720956580049278E-02_dp, 1.0206299110287578E-02_dp, -4.1412493078630895E-02_dp, &
         7.9885005336768396E-03_dp, 1.3447541252729038E-02_dp, -2.6017691043546186E-03_dp, &
         -8.7910286454442373E-03_dp, 6.7236958305186539E-03_dp, -1.5940141945756799E-03_dp, &
         4.0846562583048689E+00_dp, -1.1927022865417531E+00_dp, -5.5724233644012455E-01_dp, &
         2.3929021555236676E-01_dp, 1.6353741898513949E-01_dp, -2.9735107611800447E-02_dp, &
         -1.2793124562894490E-01_dp, 4.7930701290913279E-02_dp, 5.5361774728157669E-02_dp, &
         -3.9226885480863859E-02_dp, -1.1058674730255031E-02_dp, 1.8542247519973694E-02_dp, &
         -5.3076927066474976E-03_dp, 7.9215231943523268E+00_dp, -3.6845503235373944E+00_dp, &
         -1.3154610882028632E+00_dp, 1.1148266588452116E+00_dp, 4.2745695912670312E-01_dp, &
         -2.7683155181369429E-01_dp, -4.0167185145554668E-01_dp, 2.4904589781084449E-01_dp, &
         2.2516861354249776E-01_dp, -2.7804985084751915E-01_dp, 6.9038251045187016E-02_dp, &
         2.9781052441439151E-02_dp, -1.4527696704126849E-02_dp, 1.7339641636383043E+01_dp, &
         -1.1465395481325789E+01_dp, -2.7302697069655961E+00_dp, 4.4629359221611402E+00_dp, &
         7.1614646223286882E-01_dp, -1.4605259953560206E+00_dp, -1.0335490616197762E+00_dp, &
         1.1536772096440751E+00_dp, 7.7992282360725562E-01_dp, -1.5718730701413242E+00_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/8.6574572823904627E-01_dp, &
         -1.6889080712338411E-01_dp, -3.8729713566779244E-03_dp, 4.1076303571384891E+01_dp, &
         -3.6684994925539350E+01_dp, -3.4268355391161074E+00_dp, 1.6994010089726881E+01_dp, &
         -1.2988970233372894E+00_dp, -6.3669393876591824E+00_dp, -1.4789915209996456E+00_dp, &
         5.0459641372112225E+00_dp, 1.7604359362697857E+00_dp, -7.4519441974173777E+00_dp, &
         6.0042280072110019E+00_dp, -2.1903299806572583E+00_dp, 3.1342632696136241E-01_dp, &
         1.0537644349674036E+02_dp, -1.2495001078702305E+02_dp, 1.1671977638986354E+01_dp, &
         6.3509049743298789E+01_dp, -2.4624518831017006E+01_dp, -2.2991251058742133E+01_dp, &
         6.5440301808037606E+00_dp, 2.0726271091951617E+01_dp, -5.4834521038296886E+00_dp, &
         -2.4964517024507781E+01_dp, 2.9635321224675756E+01_dp, -1.4308796101738487E+01_dp, &
         2.7331662205056446E+00_dp, 2.9884951418781924E+02_dp, -4.7112538859165682E+02_dp, &
         1.6517376144736465E+02_dp, 2.2431316178052239E+02_dp, -2.1004912340787996E+02_dp, &
         -3.6841233998403162E+01_dp, 8.5018701750455719E+01_dp, 6.4270674161414831E+01_dp, &
         -1.1762295966994319E+02_dp, 1.4128076172212385E+01_dp, 6.8638839101450117E+01_dp, &
         -5.2639766289471062E+01_dp, 1.2875679284434117E+01_dp, 9.7963199739588015E+02_dp, &
         -2.0726596540764804E+03_dp, 1.5278705496087439E+03_dp, 4.2594838759546570E+02_dp, &
         -1.5473674085543355E+03_dp, 7.8543126899292076E+02_dp, 3.4775861155014280E+02_dp, &
         -2.1213591402524492E+02_dp, -7.6234790005893728E+02_dp, 1.1942226627680336E+03_dp, &
         -7.9341217678609769E+02_dp, 2.6356711429793290E+02_dp, -3.5098730768030379E+01_dp, &
         4.4532080783536394E+03_dp, -1.1976038810115935E+04_dp, 1.5216348304854013E+04_dp, &
         -9.1497778902350692E+03_dp, -2.5770601345297018E+03_dp, 1.1628742339594626E+04_dp, &
         -1.3276634853192494E+04_dp, 9.1770221622615209E+03_dp, -3.7823343104978308E+03_dp, &
         2.4407451616112442E+02_dp, 8.2977752842562359E+02_dp, -5.3212154229074201E+02_dp, &
         1.1938018424864173E+02_dp, 3.4262178292778683E-01_dp, -1.7261992206626605E-02_dp, &
         -2.2336602339871350E-02_dp, 1.1372975909513491E-02_dp, -1.4637411764816409E-02_dp, &
         1.9682898826632621E-02_dp, -1.9508686827707740E-02_dp, 1.5180239022615439E-02_dp, &
         -8.0494684598632436E-03_dp, 4.3996140717912250E-04_dp, 3.5809746378920688E-03_dp, &
         -2.8752058938169818E-03_dp, 7.9428830938936665E-04_dp, 1.2227594053594877E+00_dp, &
         -8.4033698327114534E-02_dp, -1.0674419107588769E-01_dp, 6.0462282288982330E-02_dp, &
         -7.0785676831264299E-02_dp, 9.5635549300517764E-02_dp, -9.4351933521006390E-02_dp, &
         7.0755447131140681E-02_dp, -3.4019702880768367E-02_dp, -3.3728225702757239E-03_dp, &
         2.1611313953339161E-02_dp, -1.5999188759917224E-02_dp, 4.3118753534726753E-03_dp, &
         2.8601521163296666E+00_dp, -2.9979848800794423E-01_dp, -3.7094815914666179E-01_dp, &
         2.4059291327186513E-01_dp, -2.5155247937875885E-01_dp, 3.3934844711305995E-01_dp, &
         -3.3320333083079079E-01_dp, 2.3651738850263895E-01_dp, -9.4854817960181387E-02_dp, &
         -4.0172372650363469E-02_dp, 9.8175845918375776E-02_dp, -6.7106282688909516E-02_dp, &
         1.7607790831527539E-02_dp, 6.5195485295712299E+00_dp, -1.0269942852581080E+00_dp, &
         -1.2316140242437659E+00_dp, 9.2057237167479200E-01_dp, -8.6609478324363198E-01_dp, &
         1.1499978387095577E+00_dp, -1.1250457879495033E+00_dp, 7.4504215559187936E-01_dp, &
         -2.1409102804328831E-01_dp, -2.5465251320434851E-01_dp, 4.2237116881553893E-01_dp, &
         -2.7025866808821075E-01_dp, 6.9222424341596048E-02_dp, 1.5601098721282909E+01_dp, &
         -3.5605032547397730E+00_dp, -4.1021042659113336E+00_dp, 3.5928399226527388E+00_dp, &
         -3.0663280855742157E+00_dp, 3.9096825239762416E+00_dp, -3.8079613332081270E+00_dp, &
         2.2900409541690148E+00_dp, -2.3404263306851089E-01_dp, -1.4199262179831273E+00_dp, &
         1.8507969187271571E+00_dp, -1.1142464342469358E+00_dp, 2.7855655064891799E-01_dp, &
         4.0485584978500526E+01_dp, -1.3109279621523161E+01_dp, -1.4273206648108861E+01_dp, &
         1.5134358148420791E+01_dp, -1.1885791275039221E+01_dp, 1.3904458082083121E+01_dp, &
         -1.3369820062766205E+01_dp, 6.8691260200838169E+00_dp, 1.8693426127542194E+00_dp, &
         -8.1523519343559254E+00_dp, 8.8821179988688765E+00_dp, -5.0221150779769559E+00_dp, &
         1.2204509968663211E+00_dp, 1.1797058394435172E+02_dp, -5.4419195200734706E+01_dp, &
         -5.4178470328491755E+01_dp, 7.3774886159171885E+01_dp, -5.5331810110639871E+01_dp, &
         5.4316621798868489E+01_dp, -4.9351105617752985E+01_dp, 1.7430778986280817E+01_dp, &
         2.6471231206118816E+01_dp, -5.4492738860827032E+01_dp, 5.1545112335746424E+01_dp, &
         -2.7216839193839707E+01_dp, 6.3684016980887135E+00_dp, 4.1049490910859993E+02_dp, &
         -2.8101620718403206E+02_dp, -2.3507330230842464E+02_dp, 4.6729944581388298E+02_dp, &
         -3.7007836528600853E+02_dp, 2.5620942949620382E+02_dp, -1.6276307350822648E+02_dp, &
         -4.7413775559957635E+01_dp, 3.3696413558725703E+02_dp, -5.0564162778438396E+02_dp, &
         4.3214949005292641E+02_dp, -2.1394266212194319E+02_dp, 4.7786865055818609E+01_dp, &
         1.9870073185576307E+03_dp, -2.2734345390187736E+03_dp, -1.0366644301592682E+03_dp, &
         4.9782231070513590E+03_dp, -5.8165480519264056E+03_dp, 2.9032478426631178E+03_dp, &
         2.0625281232662687E+03_dp, -7.2292182775277870E+03_dp, 1.0969088485231096E+04_dp, &
         -1.1554611433571949E+04_dp, 8.4527956060934266E+03_dp, -3.8415737929474558E+03_dp, &
         8.1110303330682132E+02_dp, 2.8606866433551854E+00_dp, -1.5669366038116742E-01_dp, &
         -2.0177594319939973E-01_dp, 1.0575962777565361E-01_dp, -1.3259082085957813E-01_dp, &
         1.7861218892252503E-01_dp, -1.7679239908469763E-01_dp, 1.3625980189079434E-01_dp, &
         -7.0554930091096479E-02_dp, 1.3046383793103491E-03_dp, 3.4551474369077309E-02_dp, &
         -2.7081602811949396E-02_dp, 7.4285468910542333E-03_dp, 4.5253856130077423E+00_dp, &
         -4.4517797470657283E-01_dp, -5.5534985895292899E-01_dp, 3.4609637177398361E-01_dp, &
         -3.7311782888019995E-01_dp, 5.0533840314273726E-01_dp, -4.9657643901723980E-01_dp, &
         3.5871272457032366E-01_dp, -1.5359061407378005E-01_dp, -4.6227951687665789E-02_dp, &
         1.3588639509874001E-01_dp, -9.4986322229029627E-02_dp, 2.5115468494290173E-02_dp, &
         9.3387030569139586E+00_dp, -1.4879589096614800E+00_dp, -1.7988658334718912E+00_dp, &
         1.2995109815487618E+00_dp, -1.2482340118753112E+00_dp, 1.6737963248029610E+00_dp, &
         -1.6379906508005568E+00_dp, 1.1048426169812811E+00_dp, -3.5465330257263172E-01_dp, &
         -3.2226064337605276E-01_dp, 5.7855934320968894E-01_dp, -3.7626009267967558E-01_dp, &
         9.6955612305782160E-02_dp, 2.1950328829710198E+01_dp, -5.0511260135942582E+00_dp, &
         -5.8862988319603264E+00_dp, 4.9441392015736385E+00_dp, -4.2994380387089803E+00_dp, &
         5.5897309543939446E+00_dp, -5.4527815856197339E+00_dp, 3.3734109423832814E+00_dp, &
         -5.5303193700046960E-01_dp, -1.7798059879978836E+00_dp, 2.4612001888686756E+00_dp, &
         -1.5076561686602952E+00_dp, 3.7961828447184426E-01_dp, 5.6677731876899855E+01_dp, &
         -1.8013663018122998E+01_dp, -1.9988781444904490E+01_dp, 1.9975665481283382E+01_dp, &
         -1.5915517724387783E+01_dp, 1.9377941803388104E+01_dp, -1.8776127280497441E+01_dp, &
         1.0211721332179954E+01_dp, 1.2166506526464049E+00_dp, -9.7300470047856056E+00_dp, &
         1.1178017920928982E+01_dp, -6.4583697384013634E+00_dp, 1.5861476220437252E+00_dp, &
         1.6266409498267876E+02_dp, -7.0874377152702721E+01_dp, -7.3232748732520108E+01_dp, &
         9.0638778034246016E+01_dp, -6.7833024061559215E+01_dp, 7.2645518074708306E+01_dp, &
         -6.8582720788466375E+01_dp, 2.9344572977161508E+01_dp, 2.3658760653171736E+01_dp, &
         -5.8749244736075021E+01_dp, 5.8359148143262004E+01_dp, -3.1682065751004121E+01_dp, &
         7.5424009759039885E+00_dp, 5.3985696776971338E+02_dp, -3.2878377292720995E+02_dp, &
         -3.0196845089030052E+02_dp, 4.9836773218871861E+02_dp, -3.7106297234816765E+02_dp, &
         3.1114524824685083E+02_dp, -2.6001905721418905E+02_dp, 4.3657614880237190E+01_dp, &
         2.5997836938226095E+02_dp, -4.4142314871954403E+02_dp, 3.9224023622037282E+02_dp, &
         -1.9948747668116454E+02_dp, 4.5558727242996810E+01_dp, 2.2518322147207505E+03_dp, &
         -2.0304752941146858E+03_dp, -1.4511666087126071E+03_dp, 3.8178668066270552E+03_dp, &
         -3.2459924653373946E+03_dp, 1.7680131097711978E+03_dp, -5.2399989243414564E+02_dp, &
         -1.3019719452938477E+03_dp, 3.7028205187866570E+03_dp, -5.0416739250637856E+03_dp, &
         4.1663670072848045E+03_dp, -2.0159642364314470E+03_dp, 4.4166941378618947E+02_dp, &
         1.4828081608364595E+04_dp, -2.2574049136099864E+04_dp, -4.2565500168210820E+03_dp, &
         5.5084825389384707E+04_dp, -8.2670207348266471E+04_dp, 4.6919568554688740E+04_dp, &
         4.8435800070201876E+04_dp, -1.5814297608851999E+05_dp, 2.2592917315340103E+05_dp, &
         -2.1798583714533239E+05_dp, 1.4577424945232089E+05_dp, -6.1348593359510611E+04_dp, &
         1.2191827498138327E+04_dp, 3.4532051271897191E-01_dp, 4.4100447073726646E-09_dp, &
         -1.9415646007041806E-02_dp, 1.5929327108429592E-02_dp, -1.1176571510952480E-02_dp, &
         -2.5877640413336351E-03_dp, 2.4137268336731563E-02_dp, -4.3674857388440270E-02_dp, &
         4.5860088560280322E-02_dp, -2.3078520914289882E-02_dp, -8.6082658270545242E-03_dp, &
         2.0469378895770350E-02_dp, -9.4442160284978153E-03_dp, 1.2359461442291915E+00_dp, &
         2.1613211945043984E-08_dp, -9.5224148443146792E-02_dp, 7.8125394169053466E-02_dp, &
         -5.2254864411876351E-02_dp, -1.6917118000855980E-02_dp, 1.2319600286256702E-01_dp, &
         -2.1709123959554127E-01_dp, 2.2390587928808975E-01_dp, -1.1000370261743218E-01_dp, &
         -4.3039738152667616E-02_dp, 9.7915585548516884E-02_dp, -4.4457276303720650E-02_dp, &
         2.9074409526161196E+00_dp, 7.7826482682482066E-08_dp, -3.4324867297490447E-01_dp, &
         2.8161341307550997E-01_dp, -1.7555608121572958E-01_dp, -8.2110170119991194E-02_dp, &
         4.6764741429324741E-01_dp, -7.9579019571154763E-01_dp, 8.0041604656281196E-01_dp, &
         -3.7948529579609275E-01_dp, -1.5906346472215843E-01_dp, 3.3960719652526311E-01_dp, &
         -1.5041051454561974E-01_dp, 6.6825165744494983E+00_dp, 2.6947500124974605E-07_dp, &
         -1.1899989328294034E+00_dp, 9.7631566011870097E-01_dp, -5.5706896221715130E-01_dp, &
         -3.6977101750415975E-01_dp, 1.7130054180455190E+00_dp, -2.8047552355031407E+00_dp, &
         2.7378646139272180E+00_dp, -1.2400284060281412E+00_dp, -5.6607288056274385E-01_dp, &
         1.1176163709452969E+00_dp, -4.7849425689590552E-01_dp, 1.6170325428899595E+01_dp, &
         9.4669250986364349E-07_dp, -4.1874399518362004E+00_dp, 3.4355103239851768E+00_dp, &
         -1.7339742570487151E+00_dp, -1.6747275450356629E+00_dp, 6.4121308809247504E+00_dp, &
         -1.0027369248263646E+01_dp, 9.4133727686313424E+00_dp, -3.9922245386706305E+00_dp, &
         -2.0492157416285433E+00_dp, 3.6364108946934444E+00_dp, -1.4765386616651448E+00_dp, &
         4.2602647072076110E+01_dp, 3.5480189961398575E-06_dp, -1.5730419782630479E+01_dp, &
         1.2905699113131829E+01_dp, -5.3595105861997663E+00_dp, -8.1972811001257586E+00_dp, &
         2.5924257535046760E+01_dp, -3.8176552987451025E+01_dp, 3.3842711727305677E+01_dp, &
         -1.2840341384820070E+01_dp, -7.9417840034024891E+00_dp, 1.1922026639891879E+01_dp, &
         -4.3665403821207924E+00_dp, 1.2689163866512382E+02_dp, 1.5118075434269657E-05_dp, &
         -6.7275555506731095E+01_dp, 5.5194557090186521E+01_dp, -1.5537886798726163E+01_dp, &
         -4.7253294284612615E+01_dp, 1.2145285812656974E+02_dp, -1.6368251320712150E+02_dp, &
         1.3132457381744328E+02_dp, -3.8713782602971918E+01_dp, -3.5040157951924371E+01_dp, &
         3.7792017841024283E+01_dp, -1.0016991740016113E+01_dp, 4.5777593919577163E+02_dp, &
         8.1637641969554641E-05_dp, -3.6580205008674420E+02_dp, 3.0011032726156765E+02_dp, &
         -1.3953035592512641E+01_dp, -3.7345160055057738E+02_dp, 7.4221682637239917E+02_dp/)
      REAL(KIND=dp), DIMENSION(13, 18, 4), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03/), (/13, 18, 4/))

      INTEGER                                            :: irange

      IF (Rc <= 1.0000000000000000E+01_dp) Rc = 1.0000000000000000E+01_dp
      IF (Rc >= 1.2759075164794922E+04_dp) Rc = 1.2759075164794922E+04_dp
      IF (Rc >= 1.0000000000000000E+01_dp .AND. Rc <= 1.1566345417137983E+02_dp) THEN
         irange = 1
         L_b = 1.0000000000000000E+01_dp
         U_b = 1.1566345417137983E+02_dp
      END IF
      IF (Rc >= 1.1566345417137983E+02_dp .AND. Rc <= 1.1355688414729136E+03_dp) THEN
         irange = 2
         L_b = 1.1566345417137983E+02_dp
         U_b = 1.1355688414729136E+03_dp
      END IF
      IF (Rc >= 1.1355688414729136E+03_dp .AND. Rc <= 5.9775803550807823E+03_dp) THEN
         irange = 3
         L_b = 1.1355688414729136E+03_dp
         U_b = 5.9775803550807823E+03_dp
      END IF
      IF (Rc >= 5.9775803550807823E+03_dp .AND. Rc <= 1.2759075164794922E+04_dp) THEN
         irange = 4
         L_b = 5.9775803550807823E+03_dp
         U_b = 1.2759075164794922E+04_dp
      END IF

      fit_coef(1:13, 1:18) = coefdata(1:13, 1:18, irange)

   END SUBROUTINE get_fit_coef_k09

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k10(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 20)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(240), PARAMETER :: c03 = (/-6.3360994610628929E-01_dp, &
         9.7649518951989356E-01_dp, -1.0559988395317237E+00_dp, 7.8078414899977844E-01_dp, &
         -3.5500159005989329E-01_dp, 7.4673135592582374E-02_dp, 2.6285615412587124E+00_dp, &
         1.5860173767897436E-07_dp, -3.6913389708324101E-01_dp, 3.7465972064134995E-01_dp, &
         -3.1561014124123743E-01_dp, -5.7643482677198092E-02_dp, 9.2410177205176591E-01_dp, &
         -2.1835473996583254E+00_dp, 3.2970352241440182E+00_dp, -3.4960505581040451E+00_dp, &
         2.5278721705611331E+00_dp, -1.1186977080493512E+00_dp, 2.2749024280873054E-01_dp, &
         5.7493721709618049E+00_dp, 5.1131034673141817E-07_dp, -1.1909353988603624E+00_dp, &
         1.2087611583778604E+00_dp, -9.6018131262641071E-01_dp, -3.0452567745790421E-01_dp, &
         3.1471609396333315E+00_dp, -7.1728858417880028E+00_dp, 1.0580386353620003E+01_dp, &
         -1.0957261904298493E+01_dp, 7.7029457637581844E+00_dp, -3.2879502327488281E+00_dp, &
         6.3689329465605082E-01_dp, 1.3024889596738207E+01_dp, 1.6466878135911835E-06_dp, &
         -3.8389522217497509E+00_dp, 3.8964046167903090E+00_dp, -2.8718997405626689E+00_dp, &
         -1.4373626984851318E+00_dp, 1.0761889985401876E+01_dp, -2.3552965075047659E+01_dp, &
         3.3769385385787992E+01_dp, -3.3923608671499672E+01_dp, 2.2944182267670296E+01_dp, &
         -9.2816114732207033E+00_dp, 1.6586565047476383E+00_dp, 3.1432665091848261E+01_dp, &
         5.5045818221119110E-06_dp, -1.2848255060030274E+01_dp, 1.3040500498346995E+01_dp, &
         -8.6561473620662710E+00_dp, -6.7615560396841685E+00_dp, 3.8551980935564238E+01_dp, &
         -8.0342597911471927E+01_dp, 1.1094756895369210E+02_dp, -1.0670977647200603E+02_dp, &
         6.7945774392356341E+01_dp, -2.4996446349757960E+01_dp, 3.7523703565476070E+00_dp, &
         8.2776639766289406E+01_dp, 1.9927227577400854E-05_dp, -4.6589600081344805E+01_dp, &
         4.7286523741856939E+01_dp, -2.6646560758118390E+01_dp, -3.4199978931791321E+01_dp, &
         1.5164754664098274E+02_dp, -2.9661527406436210E+02_dp, 3.8780081459610972E+02_dp, &
         -3.4748842481061382E+02_dp, 1.9747486810222125E+02_dp, -5.7764787565672627E+01_dp, &
         3.9741149863670704E+00_dp, 2.4653488710994122E+02_dp, 8.2811812634509398E-05_dp, &
         -1.9410155849192833E+02_dp, 1.9700392058258890E+02_dp, -8.1422141386870905E+01_dp, &
         -2.0290030574898262E+02_dp, 6.9905487313819572E+02_dp, -1.2476934937878245E+03_dp, &
         1.4859985026297729E+03_dp, -1.1509015539974444E+03_dp, 4.7294812978393372E+02_dp, &
         -1.1374554419169920E+01_dp, -4.9602668093775350E+01_dp, 8.8939704304992472E+02_dp, &
         4.3913317689329946E-04_dp, -1.0337351134189328E+03_dp, 1.0491819395923051E+03_dp, &
         -1.5674657748993846E+02_dp, -1.6456617734696722E+03_dp, 4.2436228771404622E+03_dp, &
         -6.4096922420241744E+03_dp, 6.0719638526939207E+03_dp, -2.5708936520526772E+03_dp, &
         -1.4140287964490178E+03_dp, 2.4275532997866176E+03_dp, -9.6578744659039239E+02_dp, &
         4.6536562243205735E+03_dp, 3.9438897095978847E-03_dp, -9.3476143934673928E+03_dp, &
         9.4871562777538929E+03_dp, 5.1466047555988271E+03_dp, -2.8237091397710159E+04_dp, &
         4.3842210795647632E+04_dp, -2.9527423734560154E+04_dp, -2.8125488385754445E+04_dp, &
         1.0440567050324763E+05_dp, -1.3678916485878875E+05_dp, 9.3760628448597679E+04_dp, &
         -2.7703671833937438E+04_dp, 2.7347133607954048E+00_dp, 8.9893013343578908E-08_dp, &
         -2.0901102881611386E-01_dp, 2.1214036968796418E-01_dp, -1.9242931099074201E-01_dp, &
         -4.6222152733960808E-03_dp, 4.8292883261738540E-01_dp, -1.2026007686844928E+00_dp, &
         1.8735094311067497E+00_dp, -2.0464478608991055E+00_dp, 1.5298400325094081E+00_dp, &
         -7.0456884498313244E-01_dp, 1.5049712771824222E-01_dp, 4.1788056884786338E+00_dp, &
         2.3922482564732447E-07_dp, -5.5667371811520094E-01_dp, 5.6500721075193983E-01_dp, &
         -4.8270517088007758E-01_dp, -7.3153070762063260E-02_dp, 1.3743457496872575E+00_dp, &
         -3.2780462608547563E+00_dp, 4.9787719700231952E+00_dp, -5.3097509004990568E+00_dp, &
         3.8648752483084237E+00_dp, -1.7244548775870321E+00_dp, 3.5436350814717738E-01_dp, &
         8.2149049298385233E+00_dp, 7.5131065115636197E-07_dp, -1.7496558689886887E+00_dp, &
         1.7758451683120040E+00_dp, -1.4284432030977785E+00_dp, -4.1105297713265365E-01_dp, &
         4.5754862006213592E+00_dp, -1.0506814920577511E+01_dp, 1.5577169057113929E+01_dp, &
         -1.6217703754793195E+01_dp, 1.1475030104148107E+01_dp, -4.9399708582597350E+00_dp, &
         9.6830600717441639E-01_dp, 1.8261194412040087E+01_dp, 2.3800351273101264E-06_dp, &
         -5.5474414243405885E+00_dp, 5.6304652949007590E+00_dp, -4.2224080442212388E+00_dp, &
         -1.9292280668308683E+00_dp, 1.5363529652698020E+01_dp, -3.3932770509938848E+01_dp, &
         4.8979621527255397E+01_dp, -4.9571574115119340E+01_dp, 3.3856280052461059E+01_dp, &
         -1.3889532396024331E+01_dp, 2.5377006524511256E+00_dp, 4.4058973569111785E+01_dp, &
         7.7835134661955194E-06_dp, -1.8161678578197755E+01_dp, 1.8433441512633923E+01_dp, &
         -1.2590211927409817E+01_dp, -8.8344038021778299E+00_dp, 5.3618887193845289E+01_dp, &
         -1.1320132345144582E+02_dp, 1.5796945077014172E+02_dp, -1.5386547456867649E+02_dp, &
         9.9772878898220497E+01_dp, -3.7833032575922452E+01_dp, 6.0353219200773358E+00_dp, &
         1.1568580381278532E+02_dp, 2.7277035891925943E-05_dp, -6.3738620396398332E+01_dp, &
         6.4692155157605299E+01_dp, -3.8542281982295471E+01_dp, -4.2526459742796327E+01_dp, &
         2.0263872121445911E+02_dp, -4.0468067321359194E+02_dp, 5.3921183956405559E+02_dp, &
         -4.9575997036083578E+02_dp, 2.9437494939739389E+02_dp, -9.4973641633375749E+01_dp, &
         1.0052032668988744E+01_dp, 3.3844441481139603E+02_dp, 1.0708615847444910E-04_dp, &
         -2.5073874169991208E+02_dp, 2.5448859830947686E+02_dp, -1.2076885762164136E+02_dp, &
         -2.3028236778445304E+02_dp, 8.7071015385976250E+02_dp, -1.6152471653168971E+03_dp, &
         2.0059864954392992E+03_dp, -1.6663062126313935E+03_dp, 8.1553595526745528E+02_dp, &
         -1.4640889696758140E+02_dp, -2.6500762414317144E+01_dp, 1.1548838669601862E+03_dp, &
         5.0406126126581669E-04_dp, -1.1839335596846738E+03_dp, 1.2016308566170446E+03_dp, &
         -3.4715885980007397E+02_dp, -1.5427529687412427E+03_dp, 4.5828692243129517E+03_dp, &
         -7.6048801944765419E+03_dp, 8.2881674827925654E+03_dp, -5.3672515338847079E+03_dp, &
         9.8514173264072872E+02_dp, 1.1594528623077892E+03_dp, -6.5416602461076945E+02_dp, &
         5.0521007201290549E+03_dp, 3.2560476887267091E-03_dp, -7.6890070328778293E+03_dp, &
         7.8038462290402367E+03_dp, 3.9687198253925811E+02_dp, -1.5428524978799329E+04_dp, &
         3.4098140303215056E+04_dp, -4.5038353268521678E+04_dp, 3.2436968649400857E+04_dp, &
         3.7529823967807060E+03_dp, -3.4972750744217890E+04_dp, 3.3219164891303109E+04_dp, &
         -1.1400731607075682E+04_dp, 3.7061397677567882E+04_dp, 4.2647739208431207E-02_dp, &
         -1.0102666755386532E+05_dp, 1.0253508247597031E+05_dp, 1.0602636418265585E+05_dp, &
         -4.0728952867028292E+05_dp, 4.6976444940836332E+05_dp, 5.8500781101532266E+04_dp, &
         -1.2055500949944141E+06_dp, 2.3598896259778733E+06_dp, -2.5585975849316372E+06_dp, &
         1.5841592037308544E+06_dp, -4.3843418788217887E+05_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/2.5561005851553148E-01_dp, &
         -9.8727639373792656E-02_dp, -3.5824853081321391E-02_dp, 1.3639809429163310E-02_dp, &
         1.2140312218587666E-02_dp, -4.6051578600266093E-03_dp, -5.2440292404026837E-03_dp, &
         1.6141788048728977E-03_dp, 4.9068438183985687E-03_dp, -3.8643109440260576E-03_dp, &
         -1.3246866398568792E-05_dp, 9.9919760932901634E-04_dp, -3.1484202326100442E-04_dp, &
         8.4272796404557193E-01_dp, -3.8901613573165461E-01_dp, -1.0067616997785502E-01_dp, &
         7.5588609850681651E-02_dp, 3.4648448246217420E-02_dp, -3.0893154344547202E-02_dp, &
         -1.4586932713617575E-02_dp, 1.2131777659737109E-02_dp, 2.0819225767407759E-02_dp, &
         -2.8910339615850700E-02_dp, 1.3280887656117454E-02_dp, -1.7063570583648409E-03_dp, &
         -2.9519647286739830E-04_dp, 1.6891663892562130E+00_dp, -1.0264543983393131E+00_dp, &
         -1.0644178450031999E-01_dp, 2.6441701384654470E-01_dp, 2.6556013010702633E-02_dp, &
         -1.1958261699563323E-01_dp, -4.5401290095952633E-03_dp, 5.1008274807051057E-02_dp, &
         4.5353450819585334E-02_dp, -1.1487744057269965E-01_dp, 8.4790329561634001E-02_dp, &
         -2.9250363584235169E-02_dp, 3.9794003608289051E-03_dp, 3.0862814349867831E+00_dp, &
         -2.5146525088601708E+00_dp, 1.6647176896057306E-01_dp, 7.3389153069620650E-01_dp, &
         -1.4890295143927060E-01_dp, -3.3532945050846130E-01_dp, 1.0826191575795263E-01_dp, &
         1.5317085471740924E-01_dp, 2.1245529597518711E-02_dp, -2.8990367580166487E-01_dp, &
         2.8971901471196243E-01_dp, -1.2768396771603524E-01_dp, 2.2620465138046251E-02_dp, &
         5.5612937791608781E+00_dp, -5.9652373960236220E+00_dp, 1.4638363568522026E+00_dp, &
         1.6757996697253348E+00_dp, -9.4543407221718756E-01_dp, -6.8535446452406956E-01_dp, &
         5.7877541374209429E-01_dp, 3.2873788086829692E-01_dp, -3.1394297986526876E-01_dp, &
         -3.7134857531460552E-01_dp, 6.2436702901871433E-01_dp, -3.4134184222136776E-01_dp, &
         7.0290493767657503E-02_dp, 1.0134025998635400E+01_dp, -1.3929286667259428E+01_dp, &
         6.1321763420892008E+00_dp, 2.9330166883940492E+00_dp, -3.6140733174101785E+00_dp, &
         -6.7590309312420871E-01_dp, 1.9913635070734654E+00_dp, 2.9670856040397131E-01_dp, &
         -1.6412798075937387E+00_dp, 5.9540353669355439E-01_dp, 5.5718678791010678E-01_dp, &
         -5.3739746596150106E-01_dp, 1.3768553709528983E-01_dp, 1.8880739332952572E+01_dp, &
         -3.2415164403909472E+01_dp, 2.1344280000859261E+01_dp, 1.8599622720347415E+00_dp, &
         -1.0841361053976007E+01_dp, 2.4783487242921880E+00_dp, 4.9433547282134356E+00_dp, &
         -1.5295881914775833E+00_dp, -4.7733675746929487E+00_dp, 5.4211714199806815E+00_dp, &
         -2.2746955524517021E+00_dp, 2.2802023048670481E-01_dp, 6.9582676550312542E-02_dp, &
         3.6335716382001010E+01_dp, -7.6018306813772128E+01_dp, 6.8732879845752208E+01_dp, &
         -1.6316998250306515E+01_dp, -2.4243518494944528E+01_dp, 1.8917610164460754E+01_dp, &
         5.5832946785483388E+00_dp, -1.0154671149564965E+01_dp, -5.3612772885590845E+00_dp, &
         1.6258259712293338E+01_dp, -1.2814243372405180E+01_dp, 4.8354454594839140E+00_dp, &
         -7.5082718737346388E-01_dp, 7.4215448848755400E+01_dp, -1.8335462266162904E+02_dp, &
         2.1358039078323498E+02_dp, -1.1557194919902095E+02_dp, -1.6853548075444962E+01_dp, &
         6.4520986380399108E+01_dp, -2.2746737191584021E+01_dp, -2.2062101091379237E+01_dp, &
         1.7789460655504136E+01_dp, 9.0031524935346070E+00_dp, -1.8542030042521016E+01_dp, &
         1.0164879826355794E+01_dp, -2.0735114143734910E+00_dp, 1.8341446865250575E+02_dp, &
         -5.0877391178005558E+02_dp, 7.0449341000362926E+02_dp, -5.5517124915990166E+02_dp, &
         1.8704002208932476E+02_dp, 8.2551818013538082E+01_dp, -1.0679714348268803E+02_dp, &
         9.0991109491737330E+00_dp, 4.1126744560482862E+01_dp, -1.8803518402205185E+01_dp, &
         -8.8575702273859580E+00_dp, 1.0182151829754702E+01_dp, -2.6408507910563315E+00_dp, &
         2.0934716095248742E+00_dp, -8.4746055605151593E-01_dp, -2.8427750491443193E-01_dp, &
         1.3068018915578153E-01_dp, 9.7384350056846619E-02_dp, -4.7397825964349709E-02_dp, &
         -4.2114616553585213E-02_dp, 1.7316564939632989E-02_dp, 4.3715952641648077E-02_dp, &
         -4.1554266092337727E-02_dp, 7.4054798396290897E-03_dp, 5.5977770808678454E-03_dp, &
         -2.2373612160795555E-03_dp, 2.7210036656294760E+00_dp, -1.6335020259554927E+00_dp, &
         -2.0680846939368511E-01_dp, 4.2349947782997177E-01_dp, 6.8122910566226311E-02_dp, &
         -1.9210163151722040E-01_dp, -2.3321895976892021E-02_dp, 8.0146433140384166E-02_dp, &
         8.7876154568316808E-02_dp, -1.8834452288276832E-01_dp, 1.2728988197515606E-01_dp, &
         -3.9391516952941304E-02_dp, 4.4839495381117809E-03_dp, 4.2368736629375707E+00_dp, &
         -3.7802973669397950E+00_dp, 2.8694988069048666E-01_dp, 1.1892835304153320E+00_dp, &
         -2.1162718986798351E-01_dp, -5.6362634575947124E-01_dp, 1.5250211850697887E-01_dp, &
         2.5537886963697748E-01_dp, 7.9649433013108686E-02_dp, -5.2911720038955190E-01_dp, &
         5.0012923442134083E-01_dp, -2.1252561189029046E-01_dp, 3.6393804438124989E-02_dp, &
         7.2758481879346171E+00_dp, -8.8167800124863280E+00_dp, 2.3330958495202210E+00_dp, &
         2.7108042687959344E+00_dp, -1.4637274671449114E+00_dp, -1.1890746322064221E+00_dp, &
         9.0715716740563646E-01_dp, 5.8170114998851308E-01_dp, -4.2989157564081804E-01_dp, &
         -7.8489539565179522E-01_dp, 1.1632291726097599E+00_dp, -6.1430639748629101E-01_dp, &
         1.2402938836877603E-01_dp, 1.3170364441482098E+01_dp, -2.0369157267246411E+01_dp, &
         9.4288085210953678E+00_dp, 4.8217116031330969E+00_dp, -5.6861997313549448E+00_dp, &
         -1.3780670845149023E+00_dp, 3.2453891736082223E+00_dp, 6.9371123634570475E-01_dp, &
         -2.7043289246339368E+00_dp, 6.9595595974366786E-01_dp, 1.2333657622496881E+00_dp, &
         -1.0391669343871834E+00_dp, 2.5704620438014530E-01_dp, 2.4716167001767921E+01_dp, &
         -4.7051480106065966E+01_dp, 3.2155480577188463E+01_dp, 3.9571391906830309E+00_dp, &
         -1.7481873346539064E+01_dp, 2.9769207206580353E+00_dp, 8.6315702725343542E+00_dp, &
         -1.8390311278544129E+00_dp, -8.6328002259793184E+00_dp, 8.8292799810814202E+00_dp, &
         -3.1386229669783341E+00_dp, -2.3715217346491377E-02_dp, 2.0535568808319654E-01_dp, &
         4.8051665196315099E+01_dp, -1.0976439154416751E+02_dp, 1.0235279436257439E+02_dp, &
         -2.0383771987790944E+01_dp, -4.2573982577312300E+01_dp, 2.8489936514971767E+01_dp, &
         1.3632688101723078E+01_dp, -1.6113634285018911E+01_dp, -1.5125986568044409E+01_dp, &
         3.3059561366308991E+01_dp, -2.3793635327665104E+01_dp, 8.3081513867812440E+00_dp, &
         -1.1774786371362731E+00_dp, 9.8355243119061484E+01_dp, -2.6174079961725596E+02_dp, &
         3.1507317101618884E+02_dp, -1.6085984367937596E+02_dp, -5.3691559005483263E+01_dp, &
         1.1770786081977101E+02_dp, -2.4626973913536151E+01_dp, -5.6303483179808985E+01_dp, &
         2.8847539985623889E+01_dp, 3.7907983907152932E+01_dp, -5.4916726607563149E+01_dp, &
         2.8300062799138427E+01_dp, -5.6436900258842000E+00_dp, 2.2780023889155092E+02_dp, &
         -6.6928883833527561E+02_dp, 9.6616701087926413E+02_dp, -7.6933915858921614E+02_dp, &
         2.0967951970184987E+02_dp, 2.2235553142415941E+02_dp, -2.3712254219885492E+02_dp, &
         8.8535187310337022E+00_dp, 1.3723076658510743E+02_dp, -1.0892853389708965E+02_dp, &
         3.0016643813667478E+01_dp, 3.2568204150430908E+00_dp, -2.6599172689595973E+00_dp, &
         8.4295965555020791E+02_dp, -2.5052951870900652E+03_dp, 3.8112871716634299E+03_dp, &
         -3.5307900192682346E+03_dp, 1.8732914678507132E+03_dp, -2.3164482385830567E+02_dp, &
         -3.8311716432588844E+02_dp, 1.7973313419763957E+02_dp, 9.3149821542615896E+01_dp, &
         -1.0608074184561853E+02_dp, 1.5210241656314690E+01_dp, 1.7089336905208167E+01_dp, &
         -6.3589478647152200E+00_dp, 3.0592957565289286E-01_dp, -5.1110929330852831E-02_dp, &
         -2.8987289080467823E-02_dp, 3.3561895390932014E-03_dp, 7.5514448011207201E-03_dp, &
         1.5139979831753692E-04_dp, -4.1708804842884705E-03_dp, 6.3660879882102028E-04_dp, &
         1.8200902121666236E-03_dp, -4.4081413042472650E-04_dp, -1.0534265079120433E-03_dp, &
         8.4103592321804256E-04_dp, -2.0186508759774654E-04_dp, 1.0525777831189882E+00_dp, &
         -2.2605017958924992E-01_dp, -1.1386949796815091E-01_dp, 2.9774643144139291E-02_dp, &
         3.3954500666697272E-02_dp, -5.4949904510822227E-03_dp, -1.9959719518336631E-02_dp, &
         5.5330636693116369E-03_dp, 1.0456790729486607E-02_dp, -6.1821591900185022E-03_dp, &
         -2.5994663457028877E-03_dp, 3.4568072502680707E-03_dp, -9.4576165761213102E-04_dp, &
         2.2919336198838280E+00_dp, -7.0509806233015249E-01_dp, -2.9188032941530057E-01_dp, &
         1.5316780817827591E-01_dp, 1.0010940584584704E-01_dp, -4.5873718932214910E-02_dp, &
         -6.4374651563716148E-02_dp, 3.0933040465657206E-02_dp, 4.2615946369704352E-02_dp, &
         -4.3329755675445232E-02_dp, 7.3263430836030315E-03_dp, 6.6279030394463092E-03_dp, &
         -2.6348013649783435E-03_dp, 4.7111266696595502E+00_dp, -2.0753802897217151E+00_dp, &
         -6.5267974687686825E-01_dp, 6.1871821696202411E-01_dp, 2.3731527891030152E-01_dp, &
         -2.2970544582560379E-01_dp, -1.7157602804264291E-01_dp, 1.4120918754911052E-01_dp, &
         1.4835360673290199E-01_dp, -2.2594398043496730E-01_dp, 1.0338085286845002E-01_dp, &
         -1.1064326042678516E-02_dp, -3.2650162585572047E-03_dp, 9.8473743889612759E+00_dp, &
         -6.0131980554589237E+00_dp, -1.2129491990772383E+00_dp, 2.2254376519254429E+00_dp, &
         3.6302894568783933E-01_dp, -9.4270878563778415E-01_dp, -3.3552531531919016E-01_dp, &
         5.7912063800107416E-01_dp, 4.2064135755829263E-01_dp, -9.8733374612298730E-01_dp, &
         6.6042114139286523E-01_dp, -1.9752444965211660E-01_dp, 2.0598380207893886E-02_dp, &
         2.1453757316653739E+01_dp, -1.7636195844827874E+01_dp, -1.1722260578775618E+00_dp, &
         7.5540641100286328E+00_dp, -5.1916302328323216E-01_dp, -3.4197651131825317E+00_dp, &
         -4.8633514010784806E-02_dp, 2.2205434162034567E+00_dp, 6.2862136700524063E-01_dp, &
         -3.5713176709083991E+00_dp, 3.1667300178926499E+00_dp, -1.2773737284357061E+00_dp, &
         2.0804275445421447E-01_dp, 4.9455179001078307E+01_dp, -5.3788204979410104E+01_dp, &
         5.5826058527127156E+00_dp, 2.4646678349298128E+01_dp, -9.0061724463921937E+00_dp, &
         -1.0659622095516202E+01_dp, 4.4620017759335129E+00_dp, 7.7884822664291757E+00_dp, &
         -3.4529539803486271E+00_dp, -8.5214099311418341E+00_dp, 1.1263431473903086E+01_dp, &
         -5.6921726426224888E+00_dp, 1.1214589809695603E+00_dp, 1.2269759721177081E+02_dp, &
         -1.7567244385618403E+02_dp, 5.7907863244030423E+01_dp, 7.4194919093471611E+01_dp, &
         -6.3873800026211477E+01_dp, -1.9928072277598243E+01_dp, 3.3660133146703842E+01_dp, &
         1.9714768039645122E+01_dp, -4.2363027488591541E+01_dp, 1.1603745291817669E+01_dp, &
         1.6417376687682371E+01_dp, -1.4366009866157320E+01_dp, 3.6184944957788190E+00_dp, &
         3.3684139089809150E+02_dp, -6.3751305400246281E+02_dp, 4.1416908794179335E+02_dp, &
         1.3688234571140524E+02_dp, -3.7063608979057182E+02_dp, 1.2143722822915174E+02_dp, &
         1.3731882710289389E+02_dp, -4.6346221563435897E+01_dp, -2.0268131602498934E+02_dp, &
         2.7850969054636107E+02_dp, -1.6672316424649100E+02_dp, 4.9139777093543977E+01_dp, &
         -5.4771033390485266E+00_dp, 1.1166365406065504E+03_dp, -2.7547693853452047E+03_dp, &
         3.0256333468051180E+03_dp, -1.2247747589089884E+03_dp, -1.0361905831178201E+03_dp, &
         1.8075416359519638E+03_dp, -1.0299807587967950E+03_dp, 1.1242398674681689E+02_dp, &
         5.2546835084300625E+01_dp, 2.0209424570528384E+02_dp, -2.8565373044965975E+02_dp, &
         1.4996446737543528E+02_dp, -3.0374236420884017E+01_dp, 2.5318787052609042E+00_dp, &
         -4.5232522380132079E-01_dp, -2.4894203765998704E-01_dp, 3.7844517569960621E-02_dp, &
         6.7481113092261444E-02_dp, -1.8350883302710324E-03_dp, -3.7893183999051401E-02_dp, &
         6.9921199011088479E-03_dp, 1.7385860790483609E-02_dp, -5.9031381997973701E-03_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/-8.4867222607997398E-03_dp, &
         7.4160230219779905E-03_dp, -1.8338341118192985E-03_dp, 3.6656246502838239E+00_dp, &
         -1.0862821007061922E+00_dp, -4.7299838594788851E-01_dp, 2.1791529395255307E-01_dp, &
         1.6235258776392880E-01_dp, -5.9164629917823931E-02_dp, -1.0182204867540855E-01_dp, &
         4.1466211524979202E-02_dp, 6.2596474192031126E-02_dp, -5.4084127674425816E-02_dp, &
         8.8525531803891741E-04_dp, 1.4025724547496657E-02_dp, -4.6032712047529446E-03_dp, &
         6.6769922009854454E+00_dp, -3.0970371708527678E+00_dp, -1.0140989359103294E+00_dp, &
         9.1150511260383493E-01_dp, 3.9216519842874853E-01_dp, -3.2543013169962537E-01_dp, &
         -2.7778597097915597E-01_dp, 1.9772597022734270E-01_dp, 2.2579283394692193E-01_dp, &
         -3.0593985439637572E-01_dp, 1.1622044051596585E-01_dp, 2.0410252870715667E-03_dp, &
         -8.3966296975702334E-03_dp, 1.3623320699171776E+01_dp, -8.8558893274426165E+00_dp, &
         -1.9149805228610246E+00_dp, 3.2774967810751803E+00_dp, 6.8478362995256370E-01_dp, &
         -1.3680993330511229E+00_dp, -5.9476667203260292E-01_dp, 8.2139559031258302E-01_dp, &
         6.9090776235235218E-01_dp, -1.4058701920349088E+00_dp, 8.5413836920236308E-01_dp, &
         -2.1810741182595644E-01_dp, 1.4255082384362025E-02_dp, 2.9748852628521725E+01_dp, &
         -2.5626593581102714E+01_dp, -2.2281795814990457E+00_dp, 1.1083089209567575E+01_dp, &
         -1.3736928750062674E-01_dp, -5.0763678761476774E+00_dp, -5.0249418564834214E-01_dp, &
         3.1849945099789339E+00_dp, 1.4726022355354449E+00_dp, -5.4992302941830431E+00_dp, &
         4.4719918101122813E+00_dp, -1.6727141406904336E+00_dp, 2.4977009750404219E-01_dp, &
         6.9037893211784962E+01_dp, -7.6964135274020265E+01_dp, 5.4824348905454112E+00_dp, &
         3.6332914820361268E+01_dp, -1.0145340748522697E+01_dp, -1.6794616142996990E+01_dp, &
         4.6404572235591202E+00_dp, 1.1643820898102305E+01_dp, -1.5727289341303272E+00_dp, &
         -1.6439340401781347E+01_dp, 1.8245888309513088E+01_dp, -8.5306553392815161E+00_dp, &
         1.5879795273877024E+00_dp, 1.7182022390659438E+02_dp, -2.4613776736324084E+02_dp, &
         6.7186754104591998E+01_dp, 1.1416051467226143E+02_dp, -7.7001090430108874E+01_dp, &
         -4.3041023883777321E+01_dp, 4.2330169680693920E+01_dp, 3.6483814289573353E+01_dp, &
         -4.6424717513479678E+01_dp, -1.3561530898217379E+01_dp, 4.6964917734628436E+01_dp, &
         -2.9562882553835443E+01_dp, 6.6027180009057211E+00_dp, 4.6773081093747555E+02_dp, &
         -8.6370782747484213E+02_dp, 4.7050209287821843E+02_dp, 2.9908873234382622E+02_dp, &
         -4.7104852448475293E+02_dp, 1.6889908350425834E+01_dp, 2.3791881917263390E+02_dp, &
         3.6767254629124110E+01_dp, -3.4434736297874019E+02_dp, 2.9580663084422144E+02_dp, &
         -8.4122378160736815E+01_dp, -1.4744687754938745E+01_dp, 1.0220647890227653E+01_dp, &
         1.4510103014196029E+03_dp, -3.4619055478794439E+03_dp, 3.2324145408972931E+03_dp, &
         -2.0171830139352292E+02_dp, -2.4226448619655516E+03_dp, 1.8793554815989780E+03_dp, &
         3.7645024254807635E+02_dp, -1.0119465757689406E+03_dp, -5.4242419196334629E+02_dp, &
         1.9807634322220583E+03_dp, -1.7934401838025353E+03_dp, 7.8212645881853916E+02_dp, &
         -1.4202019050520397E+02_dp, 6.2496239305148747E+03_dp, -1.7931776674939312E+04_dp, &
         2.4916823442349287E+04_dp, -1.8499646284135801E+04_dp, 2.2582520021478135E+03_dp, &
         1.1773422282371406E+04_dp, -1.5812069270655764E+04_dp, 1.1855294782461493E+04_dp, &
         -6.1493986813322972E+03_dp, 2.3758121921599381E+03_dp, -6.9588428484459973E+02_dp, &
         1.3762196426902483E+02_dp, -1.1913138218513865E+01_dp, 3.2470276062374315E-01_dp, &
         -1.9003588934512133E-02_dp, -1.8355145860536166E-02_dp, 6.0475341198086311E-03_dp, &
         -6.7274274564422471E-03_dp, 1.0743771705419822E-02_dp, -1.1306871751006364E-02_dp, &
         1.0146638916666086E-02_dp, -8.7364678660853939E-03_dp, 6.4093949316994117E-03_dp, &
         -3.3864501126037571E-03_dp, 1.0980797299645153E-03_dp, -1.6237200712845642E-04_dp, &
         1.1376077378517864E+00_dp, -8.8168978454148964E-02_dp, -8.2966960947689752E-02_dp, &
         3.2220461688134687E-02_dp, -3.0778141035688236E-02_dp, 4.9924652120375348E-02_dp, &
         -5.3251003530151837E-02_dp, 4.6974990767227179E-02_dp, -3.9775722053066218E-02_dp, &
         2.9079117771650204E-02_dp, -1.5334031453898066E-02_dp, 4.9235548695483276E-03_dp, &
         -7.1155987460411857E-04_dp, 2.5663032612728065E+00_dp, -2.9417413930700842E-01_dp, &
         -2.6658115314553854E-01_dp, 1.2655515440162460E-01_dp, -1.0170953149105190E-01_dp, &
         1.6637618853203828E-01_dp, -1.8127899261011421E-01_dp, 1.5639283645717517E-01_dp, &
         -1.2892869855711223E-01_dp, 9.3691422159345950E-02_dp, -4.9367488223686778E-02_dp, &
         1.5676761588867189E-02_dp, -2.1960515393188405E-03_dp, 5.5501559534518865E+00_dp, &
         -9.3353792174523964E-01_dp, -8.0931985604485024E-01_dp, 4.6774970561508383E-01_dp, &
         -3.2499339716017250E-01_dp, 5.2455331020275098E-01_dp, -5.8731514467566936E-01_dp, &
         4.9583685161075564E-01_dp, -3.9426634261390675E-01_dp, 2.8367707651694279E-01_dp, &
         -1.4984448875803760E-01_dp, 4.7282829011273904E-02_dp, -6.4299987442053754E-03_dp, &
         1.2388957662684190E+01_dp, -2.9507992375405814E+00_dp, -2.4213308891809810E+00_dp, &
         1.7155106808462175E+00_dp, -1.0620015587182443E+00_dp, 1.6352185410468572E+00_dp, &
         -1.8946239202850306E+00_dp, 1.5686621116997250E+00_dp, -1.1845997162664987E+00_dp, &
         8.3552549739590276E-01_dp, -4.4419948824934929E-01_dp, 1.4059544874631708E-01_dp, &
         -1.8723783798815802E-02_dp, 2.9330516223788635E+01_dp, -9.6326512564302238E+00_dp, &
         -7.3400314978953354E+00_dp, 6.5269889977849020E+00_dp, -3.7394128859062916E+00_dp, &
         5.2122480243556142E+00_dp, -6.2925500085850317E+00_dp, 5.1355169298916543E+00_dp, &
         -3.5833520026413055E+00_dp, 2.4087579835578392E+00_dp, -1.2892152657735232E+00_dp, &
         4.1552495501679831E-01_dp, -5.5252748929712804E-02_dp, 7.5280704759745987E+01_dp, &
         -3.3767849596353095E+01_dp, -2.3070194711950077E+01_dp, 2.6937913499898457E+01_dp, &
         -1.5152121320132757E+01_dp, 1.7635732437963757E+01_dp, -2.2164824763993362E+01_dp, &
         1.8037707168064706E+01_dp, -1.1050554303680990E+01_dp, 6.4306484729918472E+00_dp, &
         -3.3403778828301771E+00_dp, 1.1062691086411864E+00_dp, -1.4900634156067252E-01_dp, &
         2.1606787442647337E+02_dp, -1.3382440328877573E+02_dp, -7.6047551699211525E+01_dp, &
         1.2760861971346685E+02_dp, -7.7609951567506357E+01_dp, 6.7526383179937113E+01_dp, &
         -8.4589255870543653E+01_dp, 7.0008722016727575E+01_dp, -3.3736145353795237E+01_dp, &
         9.4396713074454439E+00_dp, -1.5637007492391390E+00_dp, 1.1171056998266990E-02_dp, &
         1.1212771079965393E-01_dp, 7.3384991880302323E+02_dp, -6.5417335621128620E+02_dp, &
         -2.4586250567902340E+02_dp, 7.5968666626958600E+02_dp, -5.8873803425324638E+02_dp, &
         3.5879791502938718E+02_dp, -3.3980659001725331E+02_dp, 2.7598165136284933E+02_dp, &
         -5.9140159002267268E+01_dp, -1.2294529972421002E+02_dp, 1.4244705471792136E+02_dp, &
         -7.0098207002701088E+01_dp, 1.4534293555696086E+01_dp, 3.3844616772454051E+03_dp, &
         -4.8034622373676111E+03_dp, 2.7407731795956414E+02_dp, 6.4933384274191558E+03_dp, &
         -9.3671786693748436E+03_dp, 6.8293651088856996E+03_dp, -1.6751505221392342E+03_dp, &
         -3.1810158919697988E+03_dp, 6.3337086240843646E+03_dp, -7.0202386816691314E+03_dp, &
         5.1859554354248066E+03_dp, -2.3452002831301538E+03_dp, 4.8871710263553763E+02_dp, &
         2.6990664833406162E+00_dp, -1.7032360609079394E-01_dp, -1.6339064623599808E-01_dp, &
         5.6343825413842273E-02_dp, -6.0034058384596536E-02_dp, 9.6355578617489959E-02_dp, &
         -1.0174675727771637E-01_dp, 9.0884582873824907E-02_dp, -7.7924356474501258E-02_dp, &
         5.7118288518374742E-02_dp, -3.0159305381508530E-02_dp, 9.7524462861410255E-03_dp, &
         -1.4333289835950528E-03_dp, 4.0847254939908550E+00_dp, -4.4543680259376395E-01_dp, &
         -4.0790852964917546E-01_dp, 1.8397050605060919E-01_dp, -1.5369689027094091E-01_dp, &
         2.5238408456140987E-01_dp, -2.7306185583974296E-01_dp, 2.3685643812599680E-01_dp, &
         -1.9696262819285476E-01_dp, 1.4343603886285661E-01_dp, -7.5523339024722219E-02_dp, &
         2.4017052583640244E-02_dp, -3.3873525180420443E-03_dp, 7.9217355528550035E+00_dp, &
         -1.3760861532373267E+00_dp, -1.2035284285567731E+00_dp, 6.7184227370314420E-01_dp, &
         -4.7463289875497050E-01_dp, 7.7591654058981052E-01_dp, -8.6294791141385430E-01_dp, &
         7.3053252189346307E-01_dp, -5.8632946501607508E-01_dp, 4.2329064886995371E-01_dp, &
         -2.2316303933581608E-01_dp, 7.0275129511077339E-02_dp, -9.5676479706661693E-03_dp, &
         1.7340282840591684E+01_dp, -4.2821512957303280E+00_dp, -3.5551190120613683E+00_dp, &
         2.4241473256398329E+00_dp, -1.5142511846518363E+00_dp, 2.3855617310533646E+00_dp, &
         -2.7426998606577166E+00_dp, 2.2741890274553316E+00_dp, -1.7414520515624128E+00_dp, &
         1.2378423595437182E+00_dp, -6.5653655779588638E-01_dp, 2.0668914871276739E-01_dp, &
         -2.7419850897391889E-02_dp, 4.1078055491562758E+01_dp, -1.3702317382050255E+01_dp, &
         -1.0636229224223397E+01_dp, 8.9943416364770226E+00_dp, -5.1483412011517853E+00_dp, &
         7.4730665242250467E+00_dp, -8.9464889682750020E+00_dp, 7.2963810538971217E+00_dp, &
         -5.2128753174443299E+00_dp, 3.5795300074268930E+00_dp, -1.9173641280639240E+00_dp, &
         6.1233155609547130E-01_dp, -8.0560572449398848E-02_dp, 1.0537664003823548E+02_dp, &
         -4.6684798343138667E+01_dp, -3.2994097251570423E+01_dp, 3.5741451453175586E+01_dp, &
         -1.9689433576200049E+01_dp, 2.4641696738638490E+01_dp, -3.0837544464503011E+01_dp, &
         2.4961159129982445E+01_dp, -1.5996745583167973E+01_dp, 1.0020617009760857E+01_dp, &
         -5.3706889305591723E+00_dp, 1.7753999846542772E+00_dp, -2.3852515204726438E-01_dp, &
         2.9869868484411023E+02_dp, -1.7630985569199339E+02_dp, -1.0788199695181154E+02_dp, &
         1.5896261410654549E+02_dp, -9.0745960005483639E+01_dp, 8.9261769188135247E+01_dp, &
         -1.1499671293603262E+02_dp, 9.4016858776158358E+01_dp, -5.0104327036931210E+01_dp, &
         2.2553445772362544E+01_dp, -1.0392621676839312E+01_dp, 3.4909481019108575E+00_dp, &
         -4.6411039824149081E-01_dp, 9.7281571855615607E+02_dp, -7.8276566598170348E+02_dp, &
         -3.6841750220545742E+02_dp, 8.4221015840709913E+02_dp, -5.5710576843649176E+02_dp, &
         3.9075007223954356E+02_dp, -4.6691622115267910E+02_dp, 3.9364528127783149E+02_dp, &
         -1.4353811191269111E+02_dp, -3.5775860066911733E+01_dp, 6.3797892295263324E+01_dp, &
         -3.0108403820523300E+01_dp, 5.9947780639426460E+00_dp, 3.9315484795337375E+03_dp, &
         -4.5545648190814945E+03_dp, -1.0245924129063264E+03_dp, 5.9086864031696177E+03_dp, &
         -5.4423397520665330E+03_dp, 2.8987063024985546E+03_dp, -1.8396758304894188E+03_dp, &
         1.2823546603176267E+03_dp, 3.1149245284221831E+02_dp, -1.8673961299648126E+03_dp, &
         1.9363718166711049E+03_dp, -9.6551061755211992E+02_dp, 2.0354900388393716E+02_dp, &
         2.4280135349637152E+04_dp, -4.4803296384207482E+04_dp, 1.6182899528768005E+04_dp, &
         5.9523036280870147E+04_dp, -1.2394614179673967E+05_dp, 1.1722133038315580E+05_dp, &
         -3.1352348768700398E+04_dp, -8.6202695459030496E+04_dp, 1.6982689346809132E+05_dp, &
         -1.7890858277514763E+05_dp, 1.2286186888783661E+05_dp, -5.1808342561852995E+04_dp, &
         1.0198816861077163E+04_dp, 3.2867451616851961E-01_dp, 1.0003187174814238E-08_dp, &
         -2.3255655571169480E-02_dp, 2.3603849289304341E-02_dp, -2.1600208550223220E-02_dp, &
         -1.2742608714917504E-04_dp, 5.3167626511118796E-02_dp, -1.3331453857752948E-01_dp, &
         2.0849621635227758E-01_dp, -2.2855126470711301E-01_dp, 1.7151231132777350E-01_dp, &
         -7.9338970506037887E-02_dp, 1.7034794798258115E-02_dp, 1.1561315825459209E+00_dp, &
         4.6876842564306299E-08_dp, -1.0903099661435091E-01_dp, 1.1066333209885304E-01_dp, &
         -9.7910572631274409E-02_dp, -7.4543419120169617E-03_dp, 2.5924139627124398E-01_dp/)
      REAL(KIND=dp), DIMENSION(13, 20, 4), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03/), (/13, 20, 4/))

      INTEGER                                            :: irange

      IF (Rc <= 1.0000000000000000E+01_dp) Rc = 1.0000000000000000E+01_dp
      IF (Rc >= 2.4789161682128906E+04_dp) Rc = 2.4789161682128906E+04_dp
      IF (Rc >= 1.0000000000000000E+01_dp .AND. Rc <= 1.3035677773075426E+02_dp) THEN
         irange = 1
         L_b = 1.0000000000000000E+01_dp
         U_b = 1.3035677773075426E+02_dp
      END IF
      IF (Rc >= 1.3035677773075426E+02_dp .AND. Rc <= 1.4908189080699865E+03_dp) THEN
         irange = 2
         L_b = 1.3035677773075426E+02_dp
         U_b = 1.4908189080699865E+03_dp
      END IF
      IF (Rc >= 1.4908189080699865E+03_dp .AND. Rc <= 8.7983353697963448E+03_dp) THEN
         irange = 3
         L_b = 1.4908189080699865E+03_dp
         U_b = 8.7983353697963448E+03_dp
      END IF
      IF (Rc >= 8.7983353697963448E+03_dp .AND. Rc <= 2.4789161682128906E+04_dp) THEN
         irange = 4
         L_b = 8.7983353697963448E+03_dp
         U_b = 2.4789161682128906E+04_dp
      END IF

      fit_coef(1:13, 1:20) = coefdata(1:13, 1:20, irange)

   END SUBROUTINE get_fit_coef_k10

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k11(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 22)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(230), PARAMETER :: c04 = (/-2.5319035989384675E+00_dp, &
         -3.1868909327109335E-01_dp, 7.3902114944100665E+00_dp, -1.8431274144078589E+01_dp, &
         2.9378426424405447E+01_dp, -3.3161724592145767E+01_dp, 2.5816413812770147E+01_dp, &
         -1.2469330304705172E+01_dp, 2.8128079338690761E+00_dp, 2.4584353206019742E+01_dp, &
         -7.0898120270105360E-07_dp, -8.4777162976221128E+00_dp, 9.0616788124866332E+00_dp, &
         -7.2743279493553201E+00_dp, -2.0496520134147098E+00_dp, 2.4245583092874156E+01_dp, &
         -5.7939288071844402E+01_dp, 9.0054831214542489E+01_dp, -9.9384665876997374E+01_dp, &
         7.5600795805376052E+01_dp, -3.5621843185101483E+01_dp, 7.8236158425344327E+00_dp, &
         5.9289248842465618E+01_dp, -2.3090660386475871E-06_dp, -2.7406069249001057E+01_dp, &
         2.9293762062917892E+01_dp, -2.1436160745311739E+01_dp, -1.1100871619137575E+01_dp, &
         8.4659605462739805E+01_dp, -1.9223819873089482E+02_dp, 2.8912014235586781E+02_dp, &
         -3.0919939021055404E+02_dp, 2.2729338871058800E+02_dp, -1.0300649725896395E+02_dp, &
         2.1628399735758872E+01_dp, 1.5611877204908581E+02_dp, -8.2470664905436727E-06_dp, &
         -9.6867258565390344E+01_dp, 1.0353887823856932E+02_dp, -6.5682095402519849E+01_dp, &
         -6.0937613047746630E+01_dp, 3.2814251019497499E+02_dp, -6.9844405525445029E+02_dp, &
         1.0025237366016256E+03_dp, -1.0211592975562098E+03_dp, 7.0836288602503657E+02_dp, &
         -2.9837334150052988E+02_dp, 5.6921862071198518E+01_dp, 4.6496398441680941E+02_dp, &
         -3.4223577383871753E-05_dp, -3.9563037071188347E+02_dp, 4.2287603418762706E+02_dp, &
         -2.0653649047208501E+02_dp, -3.8172667391864337E+02_dp, 1.5030594921729898E+03_dp, &
         -2.9227874116449611E+03_dp, 3.8875752535201673E+03_dp, -3.6138894664576283E+03_dp, &
         2.2048207474371115E+03_dp, -7.5725030323513738E+02_dp, 9.7823912240919910E+01_dp, &
         1.6773973436726878E+03_dp, -1.8434170056840727E-04_dp, -2.0728088439834223E+03_dp, &
         2.2155290245149013E+03_dp, -5.1394027894319083E+02_dp, -3.2225108046723080E+03_dp, &
         9.1480623850220691E+03_dp, -1.5210320266230407E+04_dp, 1.7053021607531919E+04_dp, &
         -1.1990799071970599E+04_dp, 3.6225067334732939E+03_dp, 1.1364153809614309E+03_dp, &
         -9.4031209009192617E+02_dp, 8.7767660769754748E+03_dp, -1.7301255885247059E-03_dp, &
         -1.8469732055515993E+04_dp, 1.9741000016554910E+04_dp, 8.6834780160185710E+03_dp, &
         -5.7182355776311117E+04_dp, 9.7112250915990357E+04_dp, -8.4229126418227272E+04_dp, &
         -1.3935593484706684E+04_dp, 1.5898931395778846E+05_dp, -2.3301918179538654E+05_dp, &
         1.6673068363491533E+05_dp, -5.0314774080942996E+04_dp, 2.6059126819246052E+00_dp, &
         -1.5184941808505334E-08_dp, -1.8508505661036193E-01_dp, 1.9783570907686668E-01_dp, &
         -1.9674587968365256E-01_dp, 3.6851207151313237E-02_dp, 4.0611699310341248E-01_dp, &
         -1.1473620212627975E+00_dp, 1.9481749518031135E+00_dp, -2.3151160483320039E+00_dp, &
         1.8923593661054907E+00_dp, -9.5933207303263768E-01_dp, 2.2714210609760052E-01_dp, &
         3.8468115363487008E+00_dp, -3.8029440074718104E-08_dp, -4.6155963237640873E-01_dp, &
         4.9335593943885447E-01_dp, -4.6884491464814559E-01_dp, 4.5025271220622491E-02_dp, &
         1.0863412220747446E+00_dp, -2.9376032021086864E+00_dp, 4.8861385280514016E+00_dp, &
         -5.7142841824577904E+00_dp, 4.6031605955599106E+00_dp, -2.3011975883717044E+00_dp, &
         5.3755533579904435E-01_dp, 7.2028738158550514E+00_dp, -1.1279238254521241E-07_dp, &
         -1.3633777485153236E+00_dp, 1.4572967288789407E+00_dp, -1.3243244997028747E+00_dp, &
         2.7095192289746449E-03_dp, 3.4096983315886717E+00_dp, -8.8778329414439394E+00_dp, &
         1.4487046010265646E+01_dp, -1.6682435020852303E+01_dp, 1.3244072546969868E+01_dp, &
         -6.5263496906408722E+00_dp, 1.5028959640528783E+00_dp, 1.5155338303550476E+01_dp, &
         -3.3609978102604631E-07_dp, -4.0448729535600725E+00_dp, 4.3235039861438427E+00_dp, &
         -3.7401938738456062E+00_dp, -3.9813778724854831E-01_dp, 1.0727618136824205E+01_dp, &
         -2.6919155675271707E+01_dp, 4.3059272745573644E+01_dp, -4.8755479915707021E+01_dp, &
         3.8074633815792446E+01_dp, -1.8449947088829020E+01_dp, 4.1761221031219460E+00_dp, &
         3.4280442119778854E+01_dp, -1.0209663151194477E-06_dp, -1.2222273717908696E+01_dp, &
         1.3064175401809598E+01_dp, -1.0625423963861708E+01_dp, -2.6578485949092507E+00_dp, &
         3.4550403061250776E+01_dp, -8.3242948341915039E+01_dp, 1.3004394299969496E+02_dp, &
         -1.4419493228330839E+02_dp, 1.1023246137753506E+02_dp, -5.2221835584331593E+01_dp, &
         1.1538123809656815E+01_dp, 8.3023233971686949E+01_dp, -3.2566579599055860E-06_dp, &
         -3.8724653854538957E+01_dp, 4.1391988205209223E+01_dp, -3.0989658913000362E+01_dp, &
         -1.4177989050298210E+01_dp, 1.1765010876316009E+02_dp, -2.7042253291528965E+02_dp, &
         4.1008960465882154E+02_dp, -4.4218461562488704E+02_dp, 3.2805416607523796E+02_dp, &
         -1.5027909722707574E+02_dp, 3.1961374875560626E+01_dp, 2.1815261062484299E+02_dp, &
         -1.1263123220565498E-05_dp, -1.3272987707048557E+02_dp, 1.4187169395564845E+02_dp, &
         -9.4224265155115006E+01_dp, -7.4404769563706964E+01_dp, 4.3837099398368440E+02_dp, &
         -9.5183455799001024E+02_dp, 1.3871602035765698E+03_dp, -1.4365079812470631E+03_dp, &
         1.0170715004310093E+03_dp, -4.4009662202309175E+02_dp, 8.7140147118769207E+01_dp, &
         6.3829241773452395E+02_dp, -4.4064460415062701E-05_dp, -5.1269111346857051E+02_dp, &
         5.4799986683226075E+02_dp, -2.9947328222405213E+02_dp, -4.2617722685256570E+02_dp, &
         1.8705077766963821E+03_dp, -3.7732358405587056E+03_dp, 5.1863433327248795E+03_dp, &
         -5.0256458027042509E+03_dp, 3.2619991942285269E+03_dp, -1.2467431574611608E+03_dp, &
         2.0321857334508204E+02_dp, 2.1780996488924352E+03_dp, -2.0907405555531795E-04_dp, &
         -2.3849923698991861E+03_dp, 2.5492235035473764E+03_dp, -9.3326450903590683E+02_dp, &
         -2.9722230635951501E+03_dp, 9.8396661700376753E+03_dp, -1.7829859393390787E+04_dp, &
         2.2115783909639369E+04_dp, -1.8612912915580167E+04_dp, 9.4926673873045820E+03_dp, &
         -2.0581853251291723E+03_dp, -1.3506399200968644E+02_dp, 9.5282286909788218E+03_dp, &
         -1.3879365845395903E-03_dp, -1.5291429799079764E+04_dp, 1.6344150714985619E+04_dp, &
         -5.8506000394263890E+02_dp, -3.0670459034251217E+04_dp, 7.3829472150507368E+04_dp, &
         -1.0879378872595044E+05_dp, 1.0183991390866101E+05_dp, -4.2693488672623760E+04_dp, &
         -2.3602726867402875E+04_dp, 4.0091252301356195E+04_dp, -1.5836929022957091E+04_dp, &
         6.9897560151235826E+04_dp, -1.8816291773517638E-02_dp, -1.9854704384469349E+05_dp, &
         2.1221227425035721E+05_dp, 1.9513181165124831E+05_dp, -8.3228931356148690E+05_dp, &
         1.0686042433396007E+06_dp, -1.6890882361242565E+05_dp, -2.0275393275062682E+06_dp, &
         4.3886071943292571E+06_dp, -4.9464312502375301E+06_dp, 3.1275647455199584E+06_dp, &
         -8.7618952854779735E+05_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.9140383852290629E-01_dp, &
         -6.8714312208041611E-02_dp, -9.1979524758077930E-03_dp, 4.0542016308222793E-03_dp, &
         1.3043896673939879E-03_dp, -3.3077626826676678E-04_dp, -1.1713733973944338E-04_dp, &
         7.6096932007647678E-05_dp, 1.6862094749663900E-05_dp, -1.8859482112709049E-05_dp, &
         -1.0541642966027691E-05_dp, 8.6799165647851962E-06_dp, -1.1431845819938354E-06_dp, &
         6.0489131859947465E-01_dp, -2.4024677183256804E-01_dp, -1.8798045436205172E-02_dp, &
         1.5837928504487560E-02_dp, 2.6102782284167180E-03_dp, -1.5652126400000133E-03_dp, &
         -1.3827151075313952E-04_dp, 3.1557964494168748E-04_dp, 2.4958120214882143E-06_dp, &
         -7.5449303381703686E-05_dp, -1.6643512580341317E-05_dp, 2.8571234433609825E-05_dp, &
         -6.0597558303770187E-06_dp, 1.1168795367902835E+00_dp, -5.2641558523934429E-01_dp, &
         6.6463431140833772E-03_dp, 3.7711310583160011E-02_dp, -1.4443909125474224E-03_dp, &
         -4.2130528794393661E-03_dp, 7.2183333855785543E-04_dp, 7.3960445605806635E-04_dp, &
         -1.9348822131823822E-04_dp, -1.6874282736821310E-04_dp, 3.6412043398419911E-05_dp, &
         4.1084858267757457E-05_dp, -1.5134273688639719E-05_dp, 1.8158447362882504E+00_dp, &
         -1.0478545011498810E+00_dp, 1.2341009021397938E-01_dp, 7.1058959620305007E-02_dp, &
         -1.9347584416030746E-02_dp, -7.7203335451403827E-03_dp, 3.8020800688892561E-03_dp, &
         1.1445156153540811E-03_dp, -7.9235221490085630E-04_dp, -2.4612837134322182E-04_dp, &
         2.1358675908042475E-04_dp, 1.7784602018876602E-06_dp, -1.9703398840932572E-05_dp, &
         2.8299287437087801E+00_dp, -2.0047036100998326E+00_dp, 4.5453821241428582E-01_dp, &
         9.8865732993666661E-02_dp, -6.6582584274414200E-02_dp, -7.1570894058818569E-03_dp, &
         1.1198660032441313E-02_dp, 5.0949933891323033E-04_dp, -2.0685167422699120E-03_dp, &
         -9.6150256589892975E-05_dp, 5.5979142435495538E-04_dp, -1.6195041286557900E-04_dp, &
         3.3998670478786237E-07_dp, 4.3574186531348724E+00_dp, -3.7415946807175851E+00_dp, &
         1.2610537308451397E+00_dp, 4.7862913363269956E-02_dp, -1.6352351758172967E-01_dp, &
         1.3772664688980325E-02_dp, 2.4218199648024418E-02_dp, -4.1260474719994672E-03_dp, &
         -4.0082156154816646E-03_dp, 8.1922589521136123E-04_dp, 1.0199455777660676E-03_dp, &
         -5.4787322426774396E-04_dp, 7.8504581614977811E-05_dp, 6.7237595753310817E+00_dp, &
         -6.8634354719280006E+00_dp, 3.0796744406753831E+00_dp, -3.0614743416373752E-01_dp, &
         -3.1107420468808400E-01_dp, 9.3817776800566161E-02_dp, 3.5729378037653900E-02_dp, &
         -1.8761603678412703E-02_dp, -4.7366169578511481E-03_dp, 3.5316646495869753E-03_dp, &
         1.0851842798406590E-03_dp, -1.2373787163121499E-03_dp, 2.8130651342912039E-04_dp, &
         1.0510156788199021E+01_dp, -1.2466153777035323E+01_dp, 6.9869766995910823E+00_dp, &
         -1.5498053185363254E+00_dp, -3.9434499629744990E-01_dp, 2.9189754861993961E-01_dp, &
         9.5684893046540861E-03_dp, -4.6899766144214212E-02_dp, 2.9892393239476815E-03_dp, &
         8.1317647341407617E-03_dp, -1.1496803534589862E-03_dp, -1.5588269285254301E-03_dp, &
         5.6436279367429989E-04_dp, 1.6929026309404673E+01_dp, -2.2732969289472557E+01_dp, &
         1.5167967467215167E+01_dp, -5.0599918378658693E+00_dp, 7.1901703409584045E-02_dp, &
         6.0237191357615583E-01_dp, -1.3337725393148273E-01_dp, -6.4361089066435215E-02_dp, &
         2.7783835281892384E-02_dp, 8.5230341553865878E-03_dp, -6.6163986118154151E-03_dp, &
         2.1386671992928557E-04_dp, 4.7004720455020335E-04_dp, 2.9361444810427511E+01_dp, &
         -4.3304500814209732E+01_dp, 3.2853995310162027E+01_dp, -1.4060189959048198E+01_dp, &
         2.4294238784534623E+00_dp, 7.2536802429446512E-01_dp, -4.3389039409549884E-01_dp, &
         -1.3185076785459927E-02_dp, 5.6402610165301202E-02_dp, -1.1720630499261352E-03_dp, &
         -1.0821082259544815E-02_dp, 3.2422174075987468E-03_dp, -5.9058577933889726E-05_dp, &
         6.4371986201665507E+01_dp, -1.0067505697246480E+02_dp, 8.2674796182294259E+01_dp, &
         -4.0680037489638053E+01_dp, 1.0796803179141461E+01_dp, -9.8211559590887343E-02_dp, &
         -8.6947315752611443E-01_dp, 1.1519815681783646E-01_dp, 8.8122286786261939E-02_dp, &
         -1.4859222473260573E-02_dp, -1.7476608605390593E-02_dp, 7.7549011480211169E-03_dp, &
         -7.7298631180373564E-04_dp, 1.5512634652715025E+00_dp, -5.7165382658867425E-01_dp, &
         -6.8264917568339606E-02_dp, 3.4929866387098558E-02_dp, 9.6619672882347681E-03_dp, &
         -3.0449530315484925E-03_dp, -8.1140934294873649E-04_dp, 6.7166809665575506E-04_dp, &
         1.0553881582976249E-04_dp, -1.6470352066950413E-04_dp, -7.5508519829998230E-05_dp, &
         7.2215264404265218E-05_dp, -1.1149724276607350E-05_dp, 1.8009352545445239E+00_dp, &
         -8.5251974129151409E-01_dp, 7.6410708533761022E-03_dp, 6.3828560675533136E-02_dp, &
         -1.7202481171730995E-03_dp, -7.4505177561230681E-03_dp, 1.0572044356321228E-03_dp, &
         1.3218647507928171E-03_dp, -3.0561817587206915E-04_dp, -3.0299445474565189E-04_dp, &
         5.5344610050092803E-05_dp, 8.1834190969111756E-05_dp, -2.9680329778920826E-05_dp, &
         2.3527449274293573E+00_dp, -1.5147537958221933E+00_dp, 2.3115619180855354E-01_dp, &
         1.1284255760843230E-01_dp, -3.6177847817668474E-02_dp, -1.3539664544268073E-02_dp, &
         6.7019226962421333E-03_dp, 1.9653435924636472E-03_dp, -1.4109355286997759E-03_dp, &
         -4.1385824743472909E-04_dp, 4.1549350181307716E-04_dp, -1.1959801065863547E-05_dp, &
         -3.7255770501802784E-05_dp, 3.3224921863045691E+00_dp, -2.7860632861596644E+00_dp, &
         7.8052485998277177E-01_dp, 1.4483194966481014E-01_dp, -1.1737504026440210E-01_dp, &
         -1.1679042020471016E-02_dp, 1.9403002086844354E-02_dp, 6.6735285522260949E-04_dp, &
         -3.5780226530484115E-03_dp, -7.4687021790351436E-05_dp, 1.0364093806503746E-03_dp, &
         -3.5212253142996770E-04_dp, 1.4459347180379926E-05_dp, 4.9188989071416884E+00_dp, &
         -5.0912381643341389E+00_dp, 2.0338384923721380E+00_dp, 3.4520315816372826E-02_dp, &
         -2.7527398518237012E-01_dp, 2.6213290332193811E-02_dp, 4.1056443632656141E-02_dp, &
         -7.6033292189327589E-03_dp, -6.7068508750127874E-03_dp, 1.6134398217065211E-03_dp, &
         1.7099199099071764E-03_dp, -1.0188154531831771E-03_dp, 1.6585315316805667E-04_dp, &
         7.5105710318956387E+00_dp, -9.1933741951079870E+00_dp, 4.7596295380909268E+00_dp, &
         -5.7877693805389485E-01_dp, -5.0633729223025081E-01_dp, 1.6657960364469732E-01_dp, &
         5.9754102084310180E-02_dp, -3.3655424854068201E-02_dp, -7.8470267984706195E-03_dp, &
         6.4378442077806190E-03_dp, 1.7214765195746398E-03_dp, -2.1640300991494017E-03_dp, &
         5.0773164374421677E-04_dp, 1.1787767168757027E+01_dp, -1.6479896456900001E+01_dp, &
         1.0474433181368957E+01_dp, -2.6176038877413412E+00_dp, -6.1374050521456192E-01_dp, &
         5.1744487961845398E-01_dp, 1.0209694593808110E-02_dp, -8.7555357959509111E-02_dp, &
         6.7878804533424266E-03_dp, 1.5788464928805319E-02_dp, -2.5231574145546760E-03_dp, &
         -3.0178852806152551E-03_dp, 1.1280928755338241E-03_dp, 1.9240897295596305E+01_dp, &
         -2.9689511846963928E+01_dp, 2.2151809035042952E+01_dp, -8.1959918765769775E+00_dp, &
         2.3949103147685374E-01_dp, 1.0808974332068280E+00_dp, -2.7708880902203492E-01_dp, &
         -1.2364385254000407E-01_dp, 6.3036168861553526E-02_dp, 1.6885878085071752E-02_dp, &
         -1.6735028081352580E-02_dp, 2.0042184897993667E-03_dp, 7.5921274060652174E-04_dp, &
         3.3999306723020787E+01_dp, -5.5636509752237529E+01_dp, 4.6326714382390342E+01_dp, &
         -2.1820258573048438E+01_dp, 4.3276152184902941E+00_dp, 1.2172727004629018E+00_dp, &
         -9.0694960221451648E-01_dp, 3.1554492336115077E-02_dp, 1.2559453953809244E-01_dp, &
         -2.0064318592328701E-02_dp, -2.2560731494297201E-02_dp, 1.1385938701128969E-02_dp, &
         -1.5858929811645066E-03_dp, 7.2814831077229968E+01_dp, -1.2072622608954870E+02_dp, &
         1.0532732283136377E+02_dp, -5.5787958513676813E+01_dp, 1.6603414490986122E+01_dp, &
         -7.3755690320892586E-01_dp, -1.4210730327967394E+00_dp, 3.7793417576420085E-01_dp, &
         9.1264545598547908E-02_dp, -4.8606229226741834E-02_dp, -1.3188114627426912E-02_dp, &
         1.2197365335229388E-02_dp, -2.2862081919550218E-03_dp, 2.7358153743749347E+02_dp, &
         -4.4385936209458333E+02_dp, 3.8212861114580585E+02_dp, -2.0280907585143663E+02_dp, &
         6.3861686467341308E+01_dp, -7.0005683460188894E+00_dp, -2.6528795806599486E+00_dp, &
         7.0810042010323637E-01_dp, 2.6455253738091272E-01_dp, -7.9687526859254745E-02_dp, &
         -5.7953598000465630E-02_dp, 3.2570226713321167E-02_dp, -4.4496456437617135E-03_dp, &
         2.5710198422843333E-01_dp, -7.6161818320978847E-02_dp, -2.7674187351960450E-02_dp, &
         8.5898642065098691E-03_dp, 8.0536613015059072E-03_dp, -2.6174486149182770E-03_dp, &
         -3.2661823845164801E-03_dp, 9.2754900950433358E-04_dp, 2.1088398186065527E-03_dp, &
         -1.1473435220555634E-03_dp, -6.3993457458976226E-04_dp, 7.5761255017756373E-04_dp, &
         -2.0390887377095055E-04_dp, 8.4861453798468323E-01_dp, -3.0105534605114320E-01_dp, &
         -8.5111841975141406E-02_dp, 4.8094032800166463E-02_dp, 2.6040827909755683E-02_dp, &
         -1.7293930943123634E-02_dp, -1.0786968802803141E-02_dp, 6.6125719406229183E-03_dp, &
         9.1882171776696397E-03_dp, -9.6727488215151876E-03_dp, 1.7446919122676161E-03_dp, &
         1.4418793705765515E-03_dp, -5.8840280185385851E-04_dp, 1.7047248048976706E+00_dp, &
         -7.9815206934993621E-01_dp, -1.2993709521910998E-01_dp, 1.7342171454714553E-01_dp, &
         3.9041789120375842E-02_dp, -6.9360101166953014E-02_dp, -1.5275816598696328E-02_dp, &
         2.7921918147956529E-02_dp, 2.3894544071414819E-02_dp, -4.4643419061525462E-02_dp, &
         2.3898972121543908E-02_dp, -4.4247812420037886E-03_dp, -1.6586226979854507E-04_dp, &
         3.1244261122013506E+00_dp, -1.9657261428423933E+00_dp, -6.2195521540174485E-02_dp, &
         5.0922738991989702E-01_dp, -9.3639444916459660E-03_dp, -2.1322782553698055E-01_dp, &
         1.5504548523492871E-02_dp, 8.8634058985128744E-02_dp, 3.9464172801052973E-02_dp, &
         -1.4553362414306942E-01_dp, 1.1134286669408361E-01_dp, -3.8113881848227303E-02_dp, &
         4.9718947417133237E-03_dp, 5.6515577756942257E+00_dp, -4.6901233718069122E+00_dp, &
         4.9998880825294556E-01_dp, 1.2889768367120584E+00_dp, -3.4014775985320350E-01_dp, &
         -5.3198012841393982E-01_dp, 2.0501714262866522E-01_dp, 2.2366241606561824E-01_dp, &
         -1.2094839938125915E-02_dp, -3.4853283957718817E-01_dp, 3.5500068302721494E-01_dp, &
         -1.5402651260478087E-01_dp, 2.6586199081616392E-02_dp, 1.0342072806880291E+01_dp, &
         -1.1025837566627148E+01_dp, 2.8320859276363368E+00_dp, 2.8353118375827213E+00_dp, &
         -1.6643808637692934E+00_dp, -1.0312784859484334E+00_dp, 9.2317469166909838E-01_dp, &
         4.1524786772077965E-01_dp, -4.1115535857203517E-01_dp, -5.0528937194956192E-01_dp, &
         8.1638468056711944E-01_dp, -4.3490956585830648E-01_dp, 8.7714392284929860E-02_dp, &
         1.9340679951428509E+01_dp, -2.5875245987552002E+01_dp, 1.1011927062253607E+01_dp, &
         5.0172148595146115E+00_dp, -5.9922364423090562E+00_dp, -1.0100072524430268E+00_dp, &
         3.0463673833777176E+00_dp, 2.3135881603355785E-01_dp, -1.9255994055152754E+00_dp, &
         3.7336390355963817E-01_dp, 1.0535281004009800E+00_dp, -8.2899219614509112E-01_dp, &
         2.0043276697370435E-01_dp, 3.7199991651764925E+01_dp, -6.1293516863993275E+01_dp, &
         3.7969498172512104E+01_dp, 3.9173683745057399E+00_dp, -1.8033846800996791E+01_dp, &
         3.6542142546232963E+00_dp, 7.6853116612266872E+00_dp, -2.5741014212521396E+00_dp, &
         -5.6157937298118208E+00_dp, 5.6662223853193741E+00_dp, -1.5475935532793732E+00_dp, &
         -3.8649136037865522E-01_dp, 2.2567746326676860E-01_dp, 7.4213954026122252E+01_dp, &
         -1.4804950267684572E+02_dp, 1.2495507784125873E+02_dp, -2.4776264817547396E+01_dp, &
         -4.2930419560902017E+01_dp, 2.9472395507465627E+01_dp, 1.0097402469676231E+01_dp, &
         -1.5262809248340325E+01_dp, -7.2300215031127406E+00_dp, 2.0357934514863608E+01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/-1.4561539999459816E+01_dp, &
         4.8544787554718685E+00_dp, -6.3387746031848102E-01_dp, 1.5731981998839461E+02_dp, &
         -3.7106120150889751E+02_dp, 4.0538316285381757E+02_dp, -1.9862435563464234E+02_dp, &
         -4.2159595434504858E+01_dp, 1.1123116084380190E+02_dp, -3.2616679588589356E+01_dp, &
         -3.5941786883652419E+01_dp, 2.2789754163995845E+01_dp, 1.8371151060137738E+01_dp, &
         -2.9170310760936879E+01_dp, 1.4831275429341607E+01_dp, -2.8863895844841174E+00_dp, &
         4.0083846543407321E+02_dp, -1.0677816135164219E+03_dp, 1.4026777969921070E+03_dp, &
         -1.0380548329826443E+03_dp, 3.1150334049658153E+02_dp, 1.6486545559165077E+02_dp, &
         -1.8516523777840163E+02_dp, 1.6518152941373941E+01_dp, 5.6979659267802994E+01_dp, &
         -1.7647184173214058E+01_dp, -2.1103736761399592E+01_dp, 1.7917739029353150E+01_dp, &
         -4.3165125341576296E+00_dp, 2.1062827676264901E+00_dp, -6.5430642611969270E-01_dp, &
         -2.2384079169103308E-01_dp, 8.2390744712270661E-02_dp, 6.6262172019689369E-02_dp, &
         -2.6633980874697070E-02_dp, -2.7111579335168608E-02_dp, 9.7002448728929856E-03_dp, &
         1.8782298233897637E-02_dp, -1.2801673272427807E-02_dp, -3.2239388584569518E-03_dp, &
         5.7460187984439293E-03_dp, -1.6569702053578139E-03_dp, 2.7457621605201994E+00_dp, &
         -1.2692591549164844E+00_dp, -2.2947324376136316E-01_dp, 2.7323021428913846E-01_dp, &
         7.5164358453219907E-02_dp, -1.0818724938592214E-01_dp, -3.1583296898807685E-02_dp, &
         4.3084465728622491E-02_dp, 4.1482878645338658E-02_dp, -6.7933308032627218E-02_dp, &
         3.1679176585648089E-02_dp, -3.3812888555787906E-03_dp, -1.0360838277370485E-03_dp, &
         4.2942636924137831E+00_dp, -2.9557754507532943E+00_dp, -7.3406311761844992E-02_dp, &
         8.0895545247383172E-01_dp, -2.2458835272879123E-03_dp, -3.4387294218152459E-01_dp, &
         1.3473260271944267E-02_dp, 1.4295454214800754E-01_dp, 7.4353360154718565E-02_dp, &
         -2.3866533192162417E-01_dp, 1.7424538034935477E-01_dp, -5.6402793269764662E-02_dp, &
         6.6940377272607246E-03_dp, 7.4095497593660964E+00_dp, -6.9353228276198537E+00_dp, &
         8.3586764619106058E-01_dp, 2.0316025967186220E+00_dp, -5.1322687373588960E-01_dp, &
         -8.6661941980502355E-01_dp, 3.0532915478203720E-01_dp, 3.6856373700427469E-01_dp, &
         1.4157957888213758E-02_dp, -6.0168493995655148E-01_dp, 5.8832895145574027E-01_dp, &
         -2.4863584549454967E-01_dp, 4.1841159809315402E-02_dp, 1.3476526156516739E+01_dp, &
         -1.6130585523051849E+01_dp, 4.4061937443889301E+00_dp, 4.4430534169571274E+00_dp, &
         -2.5570162915545356E+00_dp, -1.7175195154266936E+00_dp, 1.4376957431381163E+00_dp, &
         7.1766464786951778E-01_dp, -5.9257675525589137E-01_dp, -9.6205557835963396E-01_dp, &
         1.4268339903708875E+00_dp, -7.4132394072868890E-01_dp, 1.4725124240356049E-01_dp, &
         2.5398852071685731E+01_dp, -3.7572387919172385E+01_dp, 1.6636186660780080E+01_dp, &
         7.9929423757903564E+00_dp, -9.2705027979453689E+00_dp, -1.9503834582624382E+00_dp, &
         4.8813044930751017E+00_dp, 6.0504538533406726E-01_dp, -3.0998271168328229E+00_dp, &
         2.6411993677728995E-01_dp, 2.0646553446087554E+00_dp, -1.5110641191317531E+00_dp, &
         3.5664182271533801E-01_dp, 4.9435136714811982E+01_dp, -8.8597920159101832E+01_dp, &
         5.6353728135153659E+01_dp, 7.6467691457190172E+00_dp, -2.8439752545361475E+01_dp, &
         4.3311736788952206E+00_dp, 1.3061842905815471E+01_dp, -3.3304225848462199E+00_dp, &
         -9.9097932669193582E+00_dp, 8.8616730772757997E+00_dp, -1.6316122579259393E+00_dp, &
         -1.1613800601835296E+00_dp, 4.8402556303837513E-01_dp, 9.9762188382444535E+01_dp, &
         -2.1339565978974167E+02_dp, 1.8334922150404742E+02_dp, -2.8939147749315442E+01_dp, &
         -7.2890617526936850E+01_dp, 4.2713837113284967E+01_dp, 2.2721643681065917E+01_dp, &
         -2.3847907629562187E+01_dp, -1.8549681716076698E+01_dp, 3.8300080834612302E+01_dp, &
         -2.4638773637155111E+01_dp, 7.2597188948475342E+00_dp, -7.5984073743085179E-01_dp, &
         2.1197486994109249E+02_dp, -5.3038671301074385E+02_dp, 5.8971315115622019E+02_dp, &
         -2.6418140422078056E+02_dp, -1.1373718823220391E+02_dp, 1.9211692177267807E+02_dp, &
         -2.7169765446157413E+01_dp, -8.4379798332666354E+01_dp, 2.7260492365389506E+01_dp, &
         6.7376634921239955E+01_dp, -8.0434216086053510E+01_dp, 3.8099371156341618E+01_dp, &
         -7.1604086942794369E+00_dp, 5.0533178791676357E+02_dp, -1.4141062067950954E+03_dp, &
         1.9146864532541163E+03_dp, -1.3990795604874068E+03_dp, 2.9162932525768673E+02_dp, &
         4.3994914666230397E+02_dp, -3.9085522986369466E+02_dp, -8.2879522036785254E+00_dp, &
         2.0611979984798336E+02_dp, -1.2695223934311078E+02_dp, 7.8185610329907984E+00_dp, &
         2.1623126780307519E+01_dp, -7.0837826672811230E+00_dp, 1.8793380650637498E+03_dp, &
         -5.3818022034402502E+03_dp, 7.8117470886846950E+03_dp, -6.8754098839868311E+03_dp, &
         3.4302358931127628E+03_dp, -3.3780948387252249E+02_dp, -7.1180847792872282E+02_dp, &
         3.4000649096567986E+02_dp, 9.5803695633720451E+01_dp, -1.0706529225064254E+02_dp, &
         -1.7431236331464156E+01_dp, 4.1838749445416617E+01_dp, -1.2304788741999435E+01_dp, &
         2.9693722288060909E-01_dp, -4.0503765415150689E-02_dp, -2.3374347340216213E-02_dp, &
         2.4649712953909672E-03_dp, 5.4970824224843205E-03_dp, 4.8287397465528485E-04_dp, &
         -3.3174719260196080E-03_dp, 5.7632661521211167E-04_dp, 1.2446956470640266E-03_dp, &
         -2.6247595657409464E-04_dp, -7.8065068838181328E-04_dp, 6.0684870308787948E-04_dp, &
         -1.4446753144174501E-04_dp, 1.0132434862618578E+00_dp, -1.7523694112647539E-01_dp, &
         -9.2556727590122526E-02_dp, 1.9924243846417989E-02_dp, 2.4530815588521580E-02_dp, &
         -1.2987424585305914E-03_dp, -1.5482592236816674E-02_dp, 4.0053098423139359E-03_dp, &
         6.6615628610347507E-03_dp, -3.1644246597228834E-03_dp, -2.5541557205489267E-03_dp, &
         2.6188403853415528E-03_dp, -6.7853784962391733E-04_dp, 2.1711398209555157E+00_dp, &
         -5.2976846220389984E-01_dp, -2.4286083825272872E-01_dp, 9.7659023189940866E-02_dp, &
         7.3533850048130775E-02_dp, -1.9551255900919134E-02_dp, -4.9877998593511656E-02_dp, &
         1.9506507612301518E-02_dp, 2.5578617398895529E-02_dp, -2.0627500236759959E-02_dp, &
         -1.7452808686614640E-03_dp, 6.8372280060992642E-03_dp, -2.1043878950266792E-03_dp, &
         4.3616246933218727E+00_dp, -1.5063751348282000E+00_dp, -5.7486437306662230E-01_dp, &
         3.8265267045798651E-01_dp, 1.8943507598881318E-01_dp, -1.0722794884839189E-01_dp, &
         -1.4175159806446239E-01_dp, 8.1999399523214028E-02_dp, 8.7843503751238977E-02_dp, &
         -1.0363055386269830E-01_dp, 2.6555214303521067E-02_dp, 9.8867863529191486E-03_dp, &
         -4.9996570844378710E-03_dp, 8.8539699844047153E+00_dp, -4.1984202894252292E+00_dp, &
         -1.2428634026751828E+00_dp, 1.3448235920919132E+00_dp, 4.0461929150339243E-01_dp, &
         -4.5733788385795898E-01_dp, -3.5409248304565399E-01_dp, 3.1850290939545917E-01_dp, &
         2.7325234985046681E-01_dp, -4.5652048431894104E-01_dp, 2.2079297546908300E-01_dp, &
         -2.9478620765750235E-02_dp, -5.1895829058320950E-03_dp, 1.8605315387814702E+01_dp, &
         -1.1760213994880345E+01_dp, -2.3008734839719169E+00_dp, 4.4982007461290801E+00_dp, &
         5.2311515349431637E-01_dp, -1.7368601503569794E+00_dp, -7.0107041944374471E-01_dp, &
         1.1883211570158956E+00_dp, 7.2483817123112193E-01_dp, -1.8304322778786442E+00_dp, &
         1.2325954726922457E+00_dp, -3.6723777123307916E-01_dp, 3.7564073010966953E-02_dp, &
         4.1004452121855905E+01_dp, -3.3883822198342756E+01_dp, -2.4660978577221768E+00_dp, &
         1.4765625376284200E+01_dp, -1.3297140620311558E+00_dp, -6.0822406363831671E+00_dp, &
         -4.2952269081750616E-01_dp, 4.3254569166267594E+00_dp, 1.0855910889639027E+00_dp, &
         -6.4825677999145492E+00_dp, 5.6839301455511970E+00_dp, -2.2574263920344908E+00_dp, &
         3.6057185979037681E-01_dp, 9.6016519297543255E+01_dp, -1.0286313276941833E+02_dp, &
         8.6326255820698279E+00_dp, 4.8051414088075639E+01_dp, -1.7261755163232731E+01_dp, &
         -1.8906364729866077E+01_dp, 6.5295315087461869E+00_dp, 1.5019872639436668E+01_dp, &
         -5.3377881739836885E+00_dp, -1.6841583305043358E+01_dp, 2.1034770339353262E+01_dp, &
         -1.0362819237186125E+01_dp, 2.0054323897259563E+00_dp, 2.4319276556516138E+02_dp, &
         -3.3866420245722674E+02_dp, 9.9763909609854409E+01_dp, 1.5027385014986709E+02_dp, &
         -1.2031406869863480E+02_dp, -3.7241043779740231E+01_dp, 5.5370114102753035E+01_dp, &
         4.1642802529325209E+01_dp, -7.1093015443508278E+01_dp, 6.7827155853688756E+00_dp, &
         4.1302786677291920E+01_dp, -3.0651138576665996E+01_dp, 7.3344210412222370E+00_dp, &
         6.8664225651899642E+02_dp, -1.2587153573562946E+03_dp, 7.5483998218097088E+02_dp, &
         3.3826607275853559E+02_dp, -7.3529640146317138E+02_dp, 2.1476437158088527E+02_dp, &
         2.4921315487310400E+02_dp, -3.9413815726828012E+01_dp, -4.0163723829203337E+02_dp, &
         4.8255471931724361E+02_dp, -2.5345967916897050E+02_dp, 5.9457400566630319E+01_dp, &
         -3.2565309975817023E+00_dp, 2.3594394496881578E+03_dp, -5.6892144384380426E+03_dp, &
         5.9833743128811575E+03_dp, -2.0231849696975166E+03_dp, -2.6090219824590527E+03_dp, &
         3.9842874719195775E+03_dp, -2.2508330708123085E+03_dp, 3.7033823002291069E+02_dp, &
         -3.1902484361391252E+01_dp, 4.8115948604471549E+02_dp, -5.8171212108552209E+02_dp, &
         2.8977583679044648E+02_dp, -5.6846166243147294E+01_dp, 2.4525299653129071E+00_dp, &
         -3.5637069115264741E-01_dp, -2.0104566416523126E-01_dp, 2.6779238503295028E-02_dp, &
         4.8939905628270081E-02_dp, 2.4792825698676576E-03_dp, -2.9869803743560470E-02_dp, &
         5.8425383143963141E-03_dp, 1.1635030317581013E-02_dp, -3.2996081706425048E-03_dp, &
         -6.5241888531386676E-03_dp, 5.3759987970630431E-03_dp, -1.3062621653998619E-03_dp, &
         3.4788507276835041E+00_dp, -8.2208076410363606E-01_dp, -3.8961998160802858E-01_dp, &
         1.4052212399279815E-01_dp, 1.1721421643645048E-01_dp, -2.4240480071210848E-02_dp, &
         -7.7735694825084065E-02_dp, 2.7075038378894713E-02_dp, 3.7905398760191382E-02_dp, &
         -2.6448267280219422E-02_dp, -6.6537554846141100E-03_dp, 1.1815563026552645E-02_dp, &
         -3.3909248684933454E-03_dp, 6.1543722219710162E+00_dp, -2.2568137161530410E+00_dp, &
         -8.7825522652867727E-01_dp, 5.6706955570340978E-01_dp, 2.9972495377368685E-01_dp, &
         -1.5163443438723695E-01_dp, -2.1990818300811518E-01_dp, 1.1667113151985914E-01_dp, &
         1.3131118147888812E-01_dp, -1.4121395152578772E-01_dp, 2.5319103809441905E-02_dp, &
         2.1127868156859679E-02_dp, -8.5766267142760727E-03_dp, 1.2157059618437597E+01_dp, &
         -6.2094066481490042E+00_dp, -1.8855484638051927E+00_dp, 1.9867289226165976E+00_dp, &
         6.5796325888957663E-01_dp, -6.6000282590724135E-01_dp, -5.5900550981956598E-01_dp, &
         4.5520937544335560E-01_dp, 4.1683712730881195E-01_dp, -6.3698977586433025E-01_dp, &
         2.7310677955566764E-01_dp, -1.5884701923367262E-02_dp, -1.3177785753804044E-02_dp, &
         2.5597273547085955E+01_dp, -1.7189605326643218E+01_dp, -3.5459040501619303E+00_dp, &
         6.5908292406530711E+00_dp, 9.9536824328795226E-01_dp, -2.5268919270014778E+00_dp, &
         -1.1778857876024893E+00_dp, 1.6973853342765182E+00_dp, 1.1755279384183819E+00_dp, &
         -2.6199621348432438E+00_dp, 1.6332594658854560E+00_dp, -4.2930601501144439E-01_dp, &
         3.0668797181816410E-02_dp, 5.6890294980110632E+01_dp, -4.8935675805853023E+01_dp, &
         -4.3928513391337605E+00_dp, 2.1493900023941237E+01_dp, -9.2561396202764701E-01_dp, &
         -8.9832110266967078E+00_dp, -1.2945023961231210E+00_dp, 6.1971050020908542E+00_dp, &
         2.4122252521897924E+00_dp, -9.7999682646599489E+00_dp, 7.9639626818274714E+00_dp, &
         -2.9529703827741800E+00_dp, 4.3437020501980872E-01_dp, 1.3418598639315178E+02_dp, &
         -1.4639064170062176E+02_dp, 7.9688822975923612E+00_dp, 6.9963146941480773E+01_dp, &
         -1.9798038249044243E+01_dp, -2.9411357486525350E+01_dp, 6.3698447544865715E+00_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/2.2105999150585166E+01_dp, &
         -2.2551854693657698E+00_dp, -3.0252257829864192E+01_dp, 3.2690274416605568E+01_dp, &
         -1.5006475584487145E+01_dp, 2.7494470603903425E+00_dp, 3.4045498207568971E+02_dp, &
         -4.7133760448357106E+02_dp, 1.1327780050673924E+02_dp, 2.2523014020490135E+02_dp, &
         -1.4330073759632975E+02_dp, -7.7468008856218788E+01_dp, 6.7539805363605566E+01_dp, &
         7.1533214810405553E+01_dp, -7.4426510844032748E+01_dp, -4.0867003585558841E+01_dp, &
         9.5519686500466420E+01_dp, -5.6748771136081224E+01_dp, 1.2334798380537688E+01_dp, &
         9.5030306864288514E+02_dp, -1.6854513316308066E+03_dp, 8.3319880141164163E+02_dp, &
         6.4915029955371108E+02_dp, -8.9512232378940689E+02_dp, 1.1151023474977780E+01_dp, &
         4.0937510009100765E+02_dp, 1.1637171492092472E+02_dp, -6.0214982550573529E+02_dp, &
         4.3122386361734880E+02_dp, -4.9217676673429537E+01_dp, -7.4489120696598462E+01_dp, &
         2.7655792789203122E+01_dp, 3.0475256987427633E+03_dp, -7.0066488989471227E+03_dp, &
         6.0846071431580522E+03_dp, 2.4654355076157873E+02_dp, -5.0986530061647118E+03_dp, &
         3.5130484940953365E+03_dp, 8.2990840282523322E+02_dp, -1.5418369657761907E+03_dp, &
         -1.6637950125880895E+03_dp, 4.0893665083154874E+03_dp, -3.3614370283246321E+03_dp, &
         1.3720459965327138E+03_dp, -2.3508756179939306E+02_dp, 1.3543997817893476E+04_dp, &
         -3.8219712441683943E+04_dp, 5.1571871314927179E+04_dp, -3.5616662338932052E+04_dp, &
         -5.5991015438729119E+02_dp, 3.0560634575586610E+04_dp, -3.8034115203375237E+04_dp, &
         2.7684234144350565E+04_dp, -1.3389558668848904E+04_dp, 3.9243885697378510E+03_dp, &
         -2.1067618872891239E+02_dp, -3.4140069782619736E+02_dp, 1.0514659620128288E+02_dp, &
         3.1070151999119766E-01_dp, -1.5418900622590423E-02_dp, -1.3385402496793672E-02_dp, &
         4.0104520214988143E-03_dp, -4.0137665511710309E-03_dp, 6.0634593372117873E-03_dp, &
         -6.0335242023330767E-03_dp, 5.1416827988300775E-03_dp, -4.2479700168415501E-03_dp, &
         3.0278142624862105E-03_dp, -1.5771186418613689E-03_dp, 5.1319284677445436E-04_dp, &
         -7.7760892249170883E-05_dp, 1.0738088654114670E+00_dp, -6.9011733174083323E-02_dp, &
         -5.8571528350665605E-02_dp, 2.0236912375192588E-02_dp, -1.7748843399198674E-02_dp, &
         2.7179396127280601E-02_dp, -2.7370169804265022E-02_dp, 2.3012442691885241E-02_dp, &
         -1.8752163526337080E-02_dp, 1.3306638853040436E-02_dp, -6.8968599129474242E-03_dp, &
         2.2148106641816261E-03_dp, -3.2713412872578977E-04_dp, 2.3587191042167146E+00_dp, &
         -2.1892062702350010E-01_dp, -1.7981032764373062E-01_dp, 7.4266715447842788E-02_dp, &
         -5.5755162509579577E-02_dp, 8.6197238895336856E-02_dp, -8.8434819428318698E-02_dp, &
         7.3044554544590024E-02_dp, -5.8253678591248331E-02_dp, 4.1045428330803928E-02_dp, &
         -2.1141989971175491E-02_dp, 6.6663194149084379E-03_dp, -9.4698866181811470E-04_dp, &
         4.9095569364132761E+00_dp, -6.5653821881663610E-01_dp, -5.1926150788332637E-01_dp, &
         2.5545610262410418E-01_dp, -1.6752475780931578E-01_dp, 2.5748215779005090E-01_dp, &
         -2.7035809285075363E-01_dp, 2.1945387900012747E-01_dp, -1.7026830150641939E-01_dp, &
         1.1880809008201974E-01_dp, -6.0862416134477300E-02_dp, 1.8834269810612231E-02_dp, &
         -2.5543692261934688E-03_dp, 1.0428495867052428E+01_dp, -1.9432357381147800E+00_dp, &
         -1.4692140211872782E+00_dp, 8.6331929124290829E-01_dp, -5.0585478836643105E-01_dp, &
         7.5530509804086554E-01_dp, -8.1607591368868604E-01_dp, 6.5190588923750659E-01_dp, &
         -4.8767325783852300E-01_dp, 3.3517775887994816E-01_dp, -1.7093793530670998E-01_dp, &
         5.1961983765136888E-02_dp, -6.6742743223630954E-03_dp, 2.3180306769496532E+01_dp, &
         -5.8479292153610993E+00_dp, -4.1741626920582231E+00_dp, 2.9735016542960855E+00_dp, &
         -1.5967301459661869E+00_dp, 2.2376916273966088E+00_dp, -2.5039185219249935E+00_dp, &
         1.9742019032983462E+00_dp, -1.4031603691615286E+00_dp, 9.3839465982933501E-01_dp, &
         -4.7641240927423684E-01_dp, 1.4259287153642736E-01_dp, -1.7178530431716743E-02_dp, &
         5.4807444543914080E+01_dp, -1.8428736049538433E+01_dp, -1.2159840401125734E+01_dp, &
         1.0799384839018646E+01_dp, -5.5132864599838030E+00_dp, 6.8824744137130160E+00_dp, &
         -8.0181186424414683E+00_dp, 6.2811263339245347E+00_dp, -4.1414130588660143E+00_dp, &
         2.6139308618654691E+00_dp, -1.3115897057255532E+00_dp, 3.8743455254506703E-01_dp, &
         -4.3118917292752940E-02_dp, 1.4054749225956735E+02_dp, -6.2916617869030020E+01_dp, &
         -3.6937557851795980E+01_dp, 4.2950396779726667E+01_dp, -2.2089900157704871E+01_dp, &
         2.2746517401589692E+01_dp, -2.7468240068761069E+01_dp, 2.1674175015936296E+01_dp, &
         -1.2736556725406421E+01_dp, 6.9130613132454206E+00_dp, -3.2447386195397336E+00_dp, &
         9.1820903948266930E-01_dp, -8.4183705014911528E-02_dp, 4.0295699079346474E+02_dp, &
         -2.4412729159645014E+02_dp, -1.1742697356266694E+02_dp, 1.9701773109595578E+02_dp, &
         -1.1199516755711144E+02_dp, 8.6331114706981623E+01_dp, -1.0268037391817907E+02_dp, &
         8.3451889210861594E+01_dp, -4.0566080642770373E+01_dp, 1.1849996075031711E+01_dp, &
         -1.5608524774197274E+00_dp, -6.5060728335316376E-01_dp, 3.8305022351595713E-01_dp, &
         1.3662736150558032E+03_dp, -1.1715066138716820E+03_dp, -3.5576387006757238E+02_dp, &
         1.1360365394454777E+03_dp, -8.3466379882382341E+02_dp, 4.6643712739732109E+02_dp, &
         -4.1192955562637843E+02_dp, 3.3476739896275893E+02_dp, -9.8910890189152127E+01_dp, &
         -9.7849413214247178E+01_dp, 1.3013359839770587E+02_dp, -6.7589528042984526E+01_dp, &
         1.4836210780641425E+01_dp, 6.2804767862049530E+03_dp, -8.4385527636566749E+03_dp, &
         7.8564045247073238E+02_dp, 9.2835293778113901E+03_dp, -1.2752206647198502E+04_dp, &
         8.9081514173809228E+03_dp, -2.5221109786186362E+03_dp, -2.7419974694190250E+03_dp, &
         5.8446477853057622E+03_dp, -6.5356503656889390E+03_dp, 4.9036628540405673E+03_dp, &
         -2.2739363291736840E+03_dp, 4.8962251827596492E+02_dp, 2.5741763632522208E+00_dp, &
         -1.3688666727294529E-01_dp, -1.1813071559062696E-01_dp, 3.6811630925133884E-02_dp, &
         -3.5503146549161348E-02_dp, 5.3860501688968561E-02_dp, -5.3757184624549099E-02_dp, &
         4.5644883856042782E-02_dp, -3.7578566225404821E-02_dp, 2.6754376121288694E-02_dp, &
         -1.3916750691682382E-02_dp, 4.5126084600130203E-03_dp, -6.7926878267826000E-04_dp, &
         3.7683048696758537E+00_dp, -3.3588966026720962E-01_dp, -2.7816323679430133E-01_dp, &
         1.1023403695203432E-01_dp, -8.5451463047366089E-02_dp, 1.3240533806228441E-01_dp, &
         -1.3509314769377362E-01_dp, 1.1202128312859926E-01_dp, -8.9894231488644935E-02_dp, &
         6.3470572830902017E-02_dp, -3.2727234081501758E-02_dp, 1.0359096690123611E-02_dp, &
         -1.4853875358901360E-03_dp, 6.9727268104466820E+00_dp, -9.7719657754485867E-01_dp, &
         -7.7724485344015970E-01_dp, 3.7367281005024411E-01_dp, -2.4776335609741895E-01_dp, &
         3.8410532223734450E-01_dp, -4.0146502572297815E-01_dp, 3.2637402854653347E-01_dp, &
         -2.5462309800085997E-01_dp, 1.7806400215143009E-01_dp, -9.1212601775455024E-02_dp, &
         2.8257494209588008E-02_dp, -3.8517458799869888E-03_dp, 1.4477920229338737E+01_dp, &
         -2.8534043059134824E+00_dp, -2.1721992056082700E+00_dp, 1.2466640681189123E+00_dp, &
         -7.3460581190645624E-01_dp, 1.1126441446066606E+00_dp, -1.1960225615602900E+00_dp, &
         9.5605186786242724E-01_dp, -7.2049164256777010E-01_dp, 4.9716022632596263E-01_dp, &
         -2.5348072114831677E-01_dp, 7.7049718869009298E-02_dp, -9.9385241865572148E-03_dp, &
         3.2252510027816605E+01_dp, -8.4617202290005658E+00_dp, -6.1023655425242858E+00_dp, &
         4.2181327536141584E+00_dp, -2.2670906799979624E+00_dp, 3.2527495661814285E+00_dp, &
         -3.6180228600037601E+00_dp, 2.8512301914963061E+00_dp, -2.0497082727121163E+00_dp, &
         1.3825637028872737E+00_dp, -7.0227883206852093E-01_dp, 2.0995045275702587E-01_dp, &
         -2.5391882706450129E-02_dp, 7.6671906271617928E+01_dp, -2.6191644580623201E+01_dp, &
         -1.7589785027199280E+01_dp, 1.4956788717489948E+01_dp, -7.5745756904485431E+00_dp, &
         9.8459676211007903E+00_dp, -1.1400125550680388E+01_dp, 8.9043865524691590E+00_dp, &
         -5.9853757336841316E+00_dp, 3.8609879714924378E+00_dp, -1.9491156032947727E+00_dp, &
         5.7551390258433999E-01_dp, -6.4555191580898369E-02_dp, 1.9670731225175837E+02_dp, &
         -8.7092067535475124E+01_dp, -5.2912261467103285E+01_dp, 5.7371747173768014E+01_dp, &
         -2.8714960392858028E+01_dp, 3.1727692420638270E+01_dp, -3.8282028198777425E+01_dp, &
         2.9960818291058100E+01_dp, -1.8238144214054639E+01_dp, 1.0608034141638724E+01_dp, &
         -5.2023410679539275E+00_dp, 1.5132791560087548E+00_dp, -1.5181291183013737E-01_dp, &
         5.5714705001563368E+02_dp, -3.2263481480447831E+02_dp, -1.6778611836804163E+02_dp, &
         2.4762291567849428E+02_dp, -1.3146191026312459E+02_dp, 1.1356350617240891E+02_dp, &
         -1.3983007561219821E+02_dp, 1.1183958095997933E+02_dp, -5.8395910760663028E+01_dp, &
         2.4697691761515443E+01_dp, -9.5671291643381515E+00_dp, 2.2817076345157226E+00_dp, &
         -5.4636457220685601E-02_dp, 1.8121516134579620E+03_dp, -1.4090173856477702E+03_dp, &
         -5.4917522369887411E+02_dp, 1.2757799353011990E+03_dp, -7.9838106640274214E+02_dp, &
         5.0012033797473867E+02_dp, -5.6022159293750929E+02_dp, 4.6860203326723456E+02_dp, &
         -1.8680839076443570E+02_dp, -1.5501517335518674E+01_dp, 5.8637375937772674E+01_dp, &
         -3.1899570998327739E+01_dp, 7.2944841057857248E+00_dp, 7.3075215016852726E+03_dp, &
         -8.0716600625892179E+03_dp, -1.3185554249594411E+03_dp, 8.6801383834222961E+03_dp, &
         -7.6321624820157049E+03_dp, 3.8266089588782174E+03_dp, -2.2802301436819621E+03_dp, &
         1.6446395857376338E+03_dp, 6.5040243939953628E+00_dp, -1.6194859543158686E+03_dp, &
         1.7774545762751463E+03_dp, -9.1126046751063029E+02_dp, 1.9853994304916299E+02_dp, &
         4.4936883720346559E+04_dp, -7.7985520384416654E+04_dp, 2.9091936453752249E+04_dp, &
         8.2600120859143732E+04_dp, -1.6542933631024099E+05_dp, 1.5026229391384314E+05_dp, &
         -4.6996941258423743E+04_dp, -7.8034667778134157E+04_dp, 1.5828259652379624E+05_dp, &
         -1.6452541439182087E+05_dp, 1.1259380546309771E+05_dp, -4.8049115944930432E+04_dp, &
         9.6994935473796631E+03_dp, 3.1427203235400236E-01_dp, -1.7052748574378306E-09_dp, &
         -2.0798188655925094E-02_dp, 2.2230997852694379E-02_dp, -2.2254212098492477E-02_dp, &
         4.4543234851814424E-03_dp, 4.5140639377175884E-02_dp, -1.2840956067326359E-01_dp, &
         2.1871157485341305E-01_dp, -2.6051786656559134E-01_dp, 2.1339394919157553E-01_dp, &
         -1.0839462535404525E-01_dp, 2.5713078268260792E-02_dp, 1.0898619734312922E+00_dp, &
         -7.7200614526554230E-09_dp, -9.3932730256275687E-02_dp, 1.0040375386032285E-01_dp, &
         -9.8022840144582693E-02_dp, 1.4771149788048780E-02_dp, 2.1228972641591992E-01_dp, &
         -5.8873273993490682E-01_dp, 9.9111659249563344E-01_dp, -1.1700776653422673E+00_dp, &
         9.5074998676885270E-01_dp, -4.7927322262063310E-01_dp, 1.1286696036611708E-01_dp, &
         2.4099686080266722E+00_dp, -2.4887589540378409E-08_dp, -3.0179173326782494E-01_dp, &
         3.2258169443715523E-01_dp, -3.0369141944218259E-01_dp, 2.3279662310325418E-02_dp, &
         7.1970589930972140E-01_dp, -1.9299446920382815E+00_dp, 3.1967747715415227E+00_dp, &
         -3.7261628430281388E+00_dp, 2.9923091861400359E+00_dp, -1.4913736558069244E+00_dp, &
         3.4734441592637294E-01_dp, 5.0643496588107419E+00_dp, -7.5994415175293662E-08_dp, &
         -9.1798921280659829E-01_dp, 9.8122645165553024E-01_dp, -8.8561110596829973E-01_dp, &
         -1.1263108188627524E-02_dp, 2.3148160729056904E+00_dp, -5.9940875579716151E+00_dp, &
         9.7525037514957766E+00_dp, -1.1202621964616382E+01_dp, 8.8723971586232349E+00_dp, &
         -4.3615551235670349E+00_dp, 1.0019327026030862E+00_dp, 1.0890417035686898E+01_dp, &
         -2.2965810094845916E-07_dp, -2.7617022482760696E+00_dp, 2.9519410935497650E+00_dp/)
      REAL(KIND=dp), DIMENSION(13, 22, 5), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04/), (/13, 22, 5/))

      INTEGER                                            :: irange

      IF (Rc <= 1.0000000000000000E+01_dp) Rc = 1.0000000000000000E+01_dp
      IF (Rc >= 4.6752176284790039E+04_dp) Rc = 4.6752176284790039E+04_dp
      IF (Rc >= 1.0000000000000000E+01_dp .AND. Rc <= 2.8424619208291968E+01_dp) THEN
         irange = 1
         L_b = 1.0000000000000000E+01_dp
         U_b = 2.8424619208291968E+01_dp
      END IF
      IF (Rc >= 2.8424619208291968E+01_dp .AND. Rc <= 3.0981600906727294E+02_dp) THEN
         irange = 2
         L_b = 2.8424619208291968E+01_dp
         U_b = 3.0981600906727294E+02_dp
      END IF
      IF (Rc >= 3.0981600906727294E+02_dp .AND. Rc <= 3.3622492755925045E+03_dp) THEN
         irange = 3
         L_b = 3.0981600906727294E+02_dp
         U_b = 3.3622492755925045E+03_dp
      END IF
      IF (Rc >= 3.3622492755925045E+03_dp .AND. Rc <= 1.6183902110703035E+04_dp) THEN
         irange = 4
         L_b = 3.3622492755925045E+03_dp
         U_b = 1.6183902110703035E+04_dp
      END IF
      IF (Rc >= 1.6183902110703035E+04_dp .AND. Rc <= 4.6752176284790039E+04_dp) THEN
         irange = 5
         L_b = 1.6183902110703035E+04_dp
         U_b = 4.6752176284790039E+04_dp
      END IF

      fit_coef(1:13, 1:22) = coefdata(1:13, 1:22, irange)

   END SUBROUTINE get_fit_coef_k11

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k12(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 24)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(360), PARAMETER :: c04 = (/-3.4185637511153556E+02_dp, &
         3.0668182445881496E+02_dp, -4.1209664021967029E+02_dp, 3.7330789502174974E+02_dp, &
         -2.3245333897278655E+02_dp, 1.3019687281058577E+02_dp, -7.4166370900253412E+01_dp, &
         3.1434861912054572E+01_dp, -6.1992398402862987E+00_dp, 3.3062345634073872E+03_dp, &
         -2.8380291130889623E+03_dp, -1.1095051926787444E+03_dp, 2.9595972359353800E+03_dp, &
         -2.0418648669157474E+03_dp, 1.3579924383957912E+03_dp, -1.6383680580108412E+03_dp, &
         1.5632799559570371E+03_dp, -8.2188370092199273E+02_dp, 1.7721434488923131E+02_dp, &
         2.5281668531481952E+01_dp, -1.8196384478159434E+01_dp, 1.9842571350266878E+00_dp, &
         1.3328564943766472E+04_dp, -1.6084133970863943E+04_dp, -2.4194016592280573E+03_dp, &
         1.9708051080372439E+04_dp, -1.9066259070011958E+04_dp, 1.0516314706873116E+04_dp, &
         -6.9116247247954307E+03_dp, 5.9899122414862086E+03_dp, -1.8790330931175195E+03_dp, &
         -3.1628969452433375E+03_dp, 4.4070324751659182E+03_dp, -2.3872165372019858E+03_dp, &
         5.1970220864424766E+02_dp, 8.1916821529150999E+04_dp, -1.5385697042232822E+05_dp, &
         6.4357021299851127E+04_dp, 1.8222531336464582E+05_dp, -3.9926912182453059E+05_dp, &
         3.9765324348418182E+05_dp, -1.5259165941629582E+05_dp, -1.9235890817676744E+05_dp, &
         4.4346076434029645E+05_dp, -4.8289048921236163E+05_dp, 3.3598895851814025E+05_dp, &
         -1.4269220555320667E+05_dp, 2.8226493599413534E+04_dp, 3.0164685170479322E-01_dp, &
         -5.5172203451047762E-10_dp, -1.8506044942376530E-02_dp, 2.0535362130493114E-02_dp, &
         -2.1955161428588353E-02_dp, 7.9539755938173202E-03_dp, 3.5963658879027555E-02_dp, &
         -1.1551812301088075E-01_dp, 2.0972249709379689E-01_dp, -2.6358369943822746E-01_dp, &
         2.2691979390776368E-01_dp, -1.2076707889951548E-01_dp, 2.9914909381961639E-02_dp, &
         1.0337362727860777E+00_dp, -2.4278618133657428E-09_dp, -8.0987705054095693E-02_dp, &
         8.9868498016994175E-02_dp, -9.4244322343980405E-02_dp, 3.0706481031845208E-02_dp, &
         1.6421381213801789E-01_dp, -5.1354447393972757E-01_dp, 9.2295913308558519E-01_dp, &
         -1.1521564902907171E+00_dp, 9.8650626274080910E-01_dp, -5.2258730390222685E-01_dp, &
         1.2892814675256531E-01_dp, 2.2336006209022599E+00_dp, -7.5196557381381133E-09_dp, &
         -2.4886328089901064E-01_dp, 2.7615231697783232E-01_dp, -2.8157691375482552E-01_dp, &
         7.6447427732108442E-02_dp, 5.3416311823157037E-01_dp, -1.6122104338954852E+00_dp, &
         2.8570238241898216E+00_dp, -3.5321996045285347E+00_dp, 3.0005760718496060E+00_dp, &
         -1.5787021349008945E+00_dp, 3.8715834373013674E-01_dp, 4.5408396268854370E+00_dp, &
         -2.1963117911792124E-08_dp, -7.2047148070424816E-01_dp, 7.9947350516824989E-01_dp, &
         -7.8952927462140421E-01_dp, 1.6405035377058730E-01_dp, 1.6397303078861616E+00_dp, &
         -4.7727711329957341E+00_dp, 8.3301112377347231E+00_dp, -1.0188327765377178E+01_dp, &
         8.5774320423583141E+00_dp, -4.4773330972176026E+00_dp, 1.0903113937324100E+00_dp, &
         9.3585396225448729E+00_dp, -6.3091718142079549E-08_dp, -2.0488465862004168E+00_dp, &
         2.2735056908505160E+00_dp, -2.1632353036088312E+00_dp, 2.8342599157808113E-01_dp, &
         4.9563849142841772E+00_dp, -1.3893621567301645E+01_dp, 2.3845002853830717E+01_dp, &
         -2.8807589805822740E+01_dp, 2.3998952097393417E+01_dp, -1.2409549730494412E+01_dp, &
         2.9962231581101673E+00_dp, 2.0035661685047266E+01_dp, -1.8318861950190785E-07_dp, &
         -5.8768822305456201E+00_dp, 6.5212788393059284E+00_dp, -5.9260748466884259E+00_dp, &
         1.9007154556845168E-01_dp, 1.5195230767450031E+01_dp, -4.0880688284867844E+01_dp, &
         6.8798962920532261E+01_dp, -8.1885938580434171E+01_dp, 6.7326941283367191E+01_dp, &
         -3.4396112195934521E+01_dp, 8.2124819537730627E+00_dp, 4.5186837524564041E+01_dp, &
         -5.5189048315790405E-07_dp, -1.7435293518428235E+01_dp, 1.9347016240234929E+01_dp, &
         -1.6555028673455720E+01_dp, -1.7282546301330646E+00_dp, 4.8592375625498889E+01_dp, &
         -1.2478022163971495E+02_dp, 2.0501118398605493E+02_dp, -2.3937993744058983E+02_dp, &
         1.9340954070002567E+02_dp, -9.7185875629100849E+01_dp, 2.2841035703260431E+01_dp, &
         1.0895632742952016E+02_dp, -1.7753617695147041E-06_dp, -5.4961807655449583E+01_dp, &
         6.0987992289219527E+01_dp, -4.7993145360250452E+01_dp, -1.4816878684426364E+01_dp, &
         1.6706696991861321E+02_dp, -4.0611982341150724E+02_dp, 6.4689668450834165E+02_dp, &
         -7.3581605952141376E+02_dp, 5.7978919474213126E+02_dp, -2.8419268389089797E+02_dp, &
         6.5166070829987390E+01_dp, 2.8689243638403678E+02_dp, -6.3312119940337709E-06_dp, &
         -1.9053537366222281E+02_dp, 2.1142529518359913E+02_dp, -1.4641605175592150E+02_dp, &
         -9.5966800148703925E+01_dp, 6.4223751431253402E+02_dp, -1.4586588921357418E+03_dp, &
         2.2256725177045328E+03_dp, -2.4338616595046051E+03_dp, 1.8418263857593236E+03_dp, &
         -8.6495958991931241E+02_dp, 1.8950502445083771E+02_dp, 8.5444004181695163E+02_dp, &
         -2.6571362038851006E-05_dp, -7.6613502361500741E+02_dp, 8.5012645497370215E+02_dp, &
         -4.6842775108287498E+02_dp, -6.5473848351985168E+02_dp, 2.9351311783685087E+03_dp, &
         -6.0796556523797653E+03_dp, 8.6674356970216486E+03_dp, -8.8364985676925335E+03_dp, &
         6.1665429803353672E+03_dp, -2.6257276231829951E+03_dp, 5.0946079410937239E+02_dp, &
         3.0824642958474647E+03_dp, -1.4788465867979594E-04_dp, -3.9609773254235906E+03_dp, &
         4.3951626916942769E+03_dp, -1.3291224125020340E+03_dp, -5.8270602373543688E+03_dp, &
         1.7945918393741063E+04_dp, -3.1907297760403359E+04_dp, 3.9407886847883412E+04_dp, &
         -3.3300333492099067E+04_dp, 1.7255106888719325E+04_dp, -3.9825476124982529E+03_dp, &
         -1.2999954606432709E+02_dp, 1.6128597938285850E+04_dp, -1.5030638475001011E-03_dp, &
         -3.4863202377279144E+04_dp, 3.8683719996946420E+04_dp, 1.3482962106710645E+04_dp, &
         -1.0745578386256201E+05_dp, 1.9505941245695992E+05_dp, -1.9666018180780555E+05_dp, &
         4.4741584065591967E+04_dp, 2.0843454148318552E+05_dp, -3.5985775021980889E+05_dp, &
         2.7195812467571569E+05_dp, -8.4386235731227309E+04_dp, 2.4940301661524327E+00_dp, &
         -4.8761432479024613E-09_dp, -1.6331906973896218E-01_dp, 1.8122810750947294E-01_dp, &
         -1.9277728419454013E-01_dp, 6.8005697039744004E-02_dp, 3.2103938239821989E-01_dp, &
         -1.0237726178221629E+00_dp, 1.8536557765917583E+00_dp, -2.3255419407411013E+00_dp, &
         1.9991982697808994E+00_dp, -1.0626828301829472E+00_dp, 2.6295769459071416E-01_dp, &
         3.5755973818818960E+00_dp, -1.1603952754322735E-08_dp, -3.8475394639504729E-01_dp, &
         4.2694415860455948E-01_dp, -4.3819791361753802E-01_dp, 1.2459309034144038E-01_dp, &
         8.1545646945528116E-01_dp, -2.4809209517018500E+00_dp, 4.4108062606870044E+00_dp, &
         -5.4655694478252199E+00_dp, 4.6516682189795713E+00_dp, -2.4513933721396328E+00_dp, &
         6.0204233745356073E-01_dp, 6.4226158559830946E+00_dp, -3.2805352108466541E-08_dp, &
         -1.0773255604203609E+00_dp, 1.1954579885591146E+00_dp, -1.1849990846141789E+00_dp, &
         2.5515816259290564E-01_dp, 2.4367480635924981E+00_dp, -7.1215213584239168E+00_dp, &
         1.2451828043400571E+01_dp, -1.5249497001548516E+01_dp, 1.2852721953337637E+01_dp, &
         -6.7156694314339243E+00_dp, 1.6368497429599620E+00_dp, 1.2902584607163559E+01_dp, &
         -9.3010917588220549E-08_dp, -3.0241468331042811E+00_dp, 3.3557497057646462E+00_dp, &
         -3.2062969899464329E+00_dp, 4.4807915038362744E-01_dp, 7.2715631913010199E+00_dp, &
         -2.0466348197975517E+01_dp, 3.5193113499959807E+01_dp, -4.2579681063128852E+01_dp, &
         3.5517565733358850E+01_dp, -1.8387068020037940E+01_dp, 4.4442003578290663E+00_dp, &
         2.7685125754473791E+01_dp, -2.6655333069596634E-07_dp, -8.5660748337068267E+00_dp, &
         9.5053423418833010E+00_dp, -8.6904919031127577E+00_dp, 3.9482814908508695E-01_dp, &
         2.1976974588949293E+01_dp, -5.9436677744795340E+01_dp, 1.0029335513740615E+02_dp, &
         -1.1962215910808953E+02_dp, 9.8539996723451168E+01_dp, -5.0431243284742301E+01_dp, &
         1.2060994098735554E+01_dp, 6.2917458626872921E+01_dp, -7.9134239123148668E-07_dp, &
         -2.5068529212304448E+01_dp, 2.7817223140209801E+01_dp, -2.4046915225943547E+01_dp, &
         -1.9395939337241239E+00_dp, 6.9089786072269447E+01_dp, -1.7876906933349807E+02_dp, &
         2.9493693173592038E+02_dp, -3.4556771722578929E+02_dp, 2.8010388997491992E+02_dp, &
         -1.4118580950500959E+02_dp, 3.3281675996370566E+01_dp, 1.5253253791878910E+02_dp, &
         -2.4954055601737901E-06_dp, -7.7617749203342299E+01_dp, 8.6128067009403992E+01_dp, &
         -6.9075429768151551E+01_dp, -1.8021885618043012E+01_dp, 2.3192299426396963E+02_dp, &
         -5.7052514794751085E+02_dp, 9.1526774152540247E+02_dp, -1.0476060088374984E+03_dp, &
         8.3055190254818183E+02_dp, -4.0964513685395752E+02_dp, 9.4525150515697661E+01_dp, &
         4.0087287412388724E+02_dp, -8.6103900253895508E-06_dp, -2.6137783643603069E+02_dp, &
         2.9003519532875379E+02_dp, -2.0888205957244327E+02_dp, -1.1370705124392683E+02_dp, &
         8.5737866354046969E+02_dp, -1.9862642690863304E+03_dp, 3.0712706338441053E+03_dp, &
         -3.4013694733239226E+03_dp, 2.6087459336779320E+03_dp, -1.2431758759458942E+03_dp, &
         2.7676845196812326E+02_dp, 1.1729505165460264E+03_dp, -3.3934203984688906E-05_dp, &
         -9.9532307429726222E+02_dp, 1.1044435943440794E+03_dp, -6.6950542960751693E+02_dp, &
         -7.1437477357816522E+02_dp, 3.6476398644962173E+03_dp, -7.8338686873921197E+03_dp, &
         1.1489560023181602E+04_dp, -1.2078261929568855E+04_dp, 8.7467165628783914E+03_dp, &
         -3.9033893165920181E+03_dp, 8.0543786542223984E+02_dp, 4.0025763692524602E+03_dp, &
         -1.6441200621815803E-04_dp, -4.5748681371839593E+03_dp, 5.0763799034785297E+03_dp, &
         -2.1894243070054299E+03_dp, -5.2679217031519338E+03_dp, 1.9223347825793080E+04_dp, &
         -3.7099987002346737E+04_dp, 4.9772930678433557E+04_dp, -4.7215872180717706E+04_dp, &
         2.9876971372619882E+04_dp, -1.0991976018488713E+04_dp, 1.6682216582907936E+03_dp, &
         1.7509520393482337E+04_dp, -1.1430899598941567E-03_dp, -2.9021900105895245E+04_dp, &
         3.2202837223613624E+04_dp, -3.5725050225571181E+03_dp, -5.6472687752002603E+04_dp, &
         1.4551089741279720E+05_dp, -2.3088420890704289E+05_dp, 2.4762084854022125E+05_dp, &
         -1.6041754781689952E+05_dp, 3.1958918982707906E+04_dp, 3.1336956460431076E+04_dp, &
         -1.8132152598903405E+04_dp, 1.2844704220123403E+05_dp, -1.6871224076075653E-02_dp, &
         -3.7306629896332178E+05_dp, 4.1394566251683736E+05_dp, 3.3744969046701316E+05_dp, &
         -1.5791948214027057E+06_dp, 2.1917044569001873E+06_dp, -7.9115467427151953E+05_dp, &
         -3.0601023356753197E+06_dp, 7.4427363873937679E+06_dp, -8.7248223821029570E+06_dp, &
         5.6367831322922539E+06_dp, -1.6014423862358064E+06_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.8031277973538481E-01_dp, &
         -6.7333985536224536E-02_dp, -8.9222590966419636E-03_dp, 4.3313443199331000E-03_dp, &
         1.3612994231258369E-03_dp, -4.2134528860866431E-04_dp, -1.3958926498276367E-04_dp, &
         1.0459985085575114E-04_dp, 2.4321788489287324E-05_dp, -2.5628407616984446E-05_dp, &
         -1.6458382252127667E-05_dp, 1.3859125845187524E-05_dp, -2.2063994735235424E-06_dp, &
         5.6642829897676783E-01_dp, -2.3140933582208004E-01_dp, -1.8778663252719714E-02_dp, &
         1.6290358696371920E-02_dp, 2.8000138622044312E-03_dp, -1.8197685206671524E-03_dp, &
         -1.8239631460980203E-04_dp, 4.0942989349671970E-04_dp, 1.5801658807221979E-05_dp, &
         -1.0194929622221692E-04_dp, -2.8507608430196145E-05_dp, 4.3492764958275464E-05_dp, &
         -9.6234726899503233E-06_dp, 1.0336794690647568E+00_dp, -4.9295449395081187E-01_dp, &
         2.0533773060304626E-03_dp, 3.7338083754647818E-02_dp, -7.9252017933650757E-04_dp, &
         -4.6073387291294529E-03_dp, 6.9737642509595181E-04_dp, 9.2284371103869235E-04_dp, &
         -2.0181932998978786E-04_dp, -2.3287683121935485E-04_dp, 3.9390431126406291E-05_dp, &
         6.1279897499314255E-05_dp, -2.1659512655332614E-05_dp, 1.6524775107115341E+00_dp, &
         -9.4977382091404372E-01_dp, 9.9098600636927089E-02_dp, 6.9382247845005254E-02_dp, &
         -1.6738833641680006E-02_dp, -8.4516567685186013E-03_dp, 3.7770914842681614E-03_dp, &
         1.4714946234771721E-03_dp, -8.7049922044455698E-04_dp, -3.7422653125675199E-04_dp, &
         2.7591748697712180E-04_dp, 1.0743072823795072E-05_dp, -2.7724542973182781E-05_dp, &
         2.5209403407090294E+00_dp, -1.7572535732534307E+00_dp, 3.6675784179170207E-01_dp, &
         1.0132030522062821E-01_dp, -5.8095720513577059E-02_dp, -9.6754073273882992E-03_dp, &
         1.1051073079906482E-02_dp, 1.1782991243745834E-03_dp, -2.2976290173779054E-03_dp, &
         -3.1775371778666181E-04_dp, 7.5498578709957390E-04_dp, -2.0358509226504188E-04_dp, &
         -2.1487703220423853E-06_dp, 3.7842663535150787E+00_dp, -3.1719815711214094E+00_dp, &
         9.9628607256145418E-01_dp, 8.4642747014531908E-02_dp, -1.4338485390045752E-01_dp, &
         3.6775926065065463E-03_dp, 2.4387343709671966E-02_dp, -2.3922521075792551E-03_dp, &
         -4.6280864130883170E-03_dp, 4.5453420746634649E-04_dp, 1.4427825127764225E-03_dp, &
         -6.9362473654104538E-04_dp, 9.3510079815132331E-05_dp, 5.6675748977110523E+00_dp, &
         -5.6248348603552669E+00_dp, 2.3650242434167574E+00_dp, -1.2518277136594844E-01_dp, &
         -2.8499324383843427E-01_dp, 6.0582406722422502E-02_dp, 4.1263478851215951E-02_dp, &
         -1.4469635429703774E-02_dp, -6.9085971227368830E-03_dp, 2.9757772888861674E-03_dp, &
         1.9815826737071439E-03_dp, -1.5730264350839068E-03_dp, 3.2218540254269558E-04_dp, &
         8.5395933853630233E+00_dp, -9.8513995851521390E+00_dp, 5.1996720804059038E+00_dp, &
         -8.9780495348559786E-01_dp, -4.3997137995311281E-01_dp, 2.1423762405988919E-01_dp, &
         4.1578026070946399E-02_dp, -4.2129670371580948E-02_dp, -4.3453273897259689E-03_dp, &
         8.5341896614408736E-03_dp, 8.3854210658043302E-04_dp, -2.5753940654037861E-03_dp, &
         7.3005171953852588E-04_dp, 1.3061522995672423E+01_dp, -1.7160442399057352E+01_dp, &
         1.0881660469438872E+01_dp, -3.0868825863712459E+00_dp, -3.6654990016924677E-01_dp, &
         5.1609758519564186E-01_dp, -3.5298257683150376E-02_dp, -8.1154201018169170E-02_dp, &
         1.4393228245430536E-02_dp, 1.5144340450492988E-02_dp, -5.1252673274943747E-03_dp, &
         -1.8479081064741922E-03_dp, 9.9116641653779008E-04_dp, 2.0632157021397099E+01_dp, &
         -3.0169058816343398E+01_dp, 2.2140092264250899E+01_dp, -8.5303719478715294E+00_dp, &
         6.8478977718455336E-01_dp, 8.8233432059397909E-01_dp, -2.8211262686916566E-01_dp, &
         -8.6566980543862401E-02_dp, 5.5120484707143334E-02_dp, 1.2338345616801711E-02_dp, &
         -1.4829218126644110E-02_dp, 2.4716302212440498E-03_dp, 4.4101212104042314E-04_dp, &
         3.5217338955590499E+01_dp, -5.5810497183858885E+01_dp, 4.5632291320845674E+01_dp, &
         -2.1446416086860207E+01_dp, 4.5351223518918955E+00_dp, 8.5236412809032913E-01_dp, &
         -7.1131676788516363E-01_dp, 7.8756575443881364E-03_dp, 9.5952279114652053E-02_dp, &
         -4.2251471621967060E-03_dp, -2.2019449280769526E-02_dp, 8.2657727156656199E-03_dp, &
         -6.8583155310803173E-04_dp, 7.6395962185656373E+01_dp, -1.2728686310186325E+02_dp, &
         1.1124646527486497E+02_dp, -5.8598736421586601E+01_dp, 1.7014576293478193E+01_dp, &
         -5.7631736469304418E-01_dp, -1.3501560044152556E+00_dp, 2.1084835464631169E-01_dp, &
         1.5298196818925494E-01_dp, -2.8319608074091947E-02_dp, -3.6419431099310340E-02_dp, &
         1.8351239837091964E-02_dp, -2.4603666608287310E-03_dp, 1.4591908177422581E+00_dp, &
         -5.5767651166491705E-01_dp, -6.6479842568471656E-02_dp, 3.6890552285677343E-02_dp, &
         1.0112353440941297E-02_dp, -3.7573056532886231E-03_dp, -9.7365738212055173E-04_dp, &
         9.0541872635710951E-04_dp, 1.5916301956753239E-04_dp, -2.2301214373999525E-04_dp, &
         -1.1911948208237914E-04_dp, 1.1340817590337231E-04_dp, -1.9950792471753723E-05_dp, &
         1.6661756196836854E+00_dp, -7.9898374218448487E-01_dp, 1.7550890038460737E-03_dp, &
         6.2809001477724419E-02_dp, -9.4026377706547348E-04_dp, -8.0360148249423204E-03_dp, &
         1.0619161940127371E-03_dp, 1.6209032795216753E-03_dp, -3.2460691552359526E-04_dp, &
         -4.1050813878500998E-04_dp, 6.0867379772286893E-05_dp, 1.1638427836478115E-04_dp, &
         -4.1022245181328590E-05_dp, 2.1179439068493049E+00_dp, -1.3568029141401121E+00_dp, &
         1.9362507650379276E-01_dp, 1.0737274042258362E-01_dp, -3.2523712445134516E-02_dp, &
         -1.4173642482044387E-02_dp, 6.8145250252902613E-03_dp, 2.3962722609825390E-03_dp, &
         -1.5785664936154885E-03_dp, -6.1334891145104148E-04_dp, 5.4453171956963262E-04_dp, &
         -1.1974429657690983E-05_dp, -4.8455232546300578E-05_dp, 2.8966195759216218E+00_dp, &
         -2.3985545624839126E+00_dp, 6.4264735748769419E-01_dp, 1.4303859261575749E-01_dp, &
         -1.0370087260206109E-01_dp, -1.4442128737280619E-02_dp, 1.9189264767116777E-02_dp, &
         1.4633618315939528E-03_dp, -3.9872332174864444E-03_dp, -3.8808657146881335E-04_dp, &
         1.4034332418797093E-03_dp, -4.5827591441622000E-04_dp, 1.7452248924342918E-05_dp, &
         4.1458283694621647E+00_dp, -4.2276944439107638E+00_dp, 1.6204734519661337E+00_dp, &
         8.3699114369403166E-02_dp, -2.4005358926470990E-01_dp, 1.2019379195118291E-02_dp, &
         4.0476206444562082E-02_dp, -5.3365562632643688E-03_dp, -7.5759092434434594E-03_dp, &
         1.1703318240633225E-03_dp, 2.4154624739185747E-03_dp, -1.3300232599431343E-03_dp, &
         2.1206653667717895E-04_dp, 6.1070095083601350E+00_dp, -7.3692662730790639E+00_dp, &
         3.6569615163999236E+00_dp, -3.0543234096798377E-01_dp, -4.5325437989138290E-01_dp, &
         1.1340063624980586E-01_dp, 6.5345011752561535E-02_dp, -2.6683850539911569E-02_dp, &
         -1.0609290569814111E-02_dp, 5.6881166467098749E-03_dp, 2.8996179473254218E-03_dp, &
         -2.6780441022469402E-03_dp, 5.9375790281117737E-04_dp, 9.1967607356270840E+00_dp, &
         -1.2728161122633493E+01_dp, 7.7585514743513535E+00_dp, -1.6031523579554121E+00_dp, &
         -6.5971663204666997E-01_dp, 3.7822167096008902E-01_dp, 5.8075256175536355E-02_dp, &
         -7.5356462157257068E-02_dp, -4.7610280346296668E-03_dp, 1.5528173541741448E-02_dp, &
         5.0291646970573997E-04_dp, -4.2910396246577829E-03_dp, 1.2926070332399792E-03_dp, &
         1.4196696181971468E+01_dp, -2.1917883458855773E+01_dp, 1.5801136777519396E+01_dp, &
         -5.1163101205901276E+00_dp, -4.3929530502807101E-01_dp, 8.9353294761283697E-01_dp, &
         -9.5307670162679783E-02_dp, -1.4563133541061388E-01_dp, 3.3385614060572416E-02_dp, &
         2.8263201419445248E-02_dp, -1.2151984828830927E-02_dp, -2.4648614361214064E-03_dp, &
         1.7803770851374343E-03_dp, 2.2826825664059452E+01_dp, -3.8162174617271774E+01_dp, &
         3.1357657732432617E+01_dp, -1.3555584080685593E+01_dp, 1.4927242940990013E+00_dp, &
         1.4849933729996307E+00_dp, -5.9022811136976339E-01_dp, -1.3676085124973847E-01_dp, &
         1.2318046476580022E-01_dp, 1.6651678250228036E-02_dp, -3.5419068294572351E-02_dp, &
         1.0103555469163566E-02_dp, -2.8242465666916748E-04_dp, 3.9949146857561630E+01_dp, &
         -6.9806819497162493E+01_dp, 6.2526091490839079E+01_dp, -3.2491521274710166E+01_dp, &
         8.0495785945780884E+00_dp, 1.1449352957658303E+00_dp, -1.4001449115453595E+00_dp, &
         1.4849591069311463E-01_dp, 1.8413922498441526E-01_dp, -4.3848567993047871E-02_dp, &
         -3.5796486617126269E-02_dp, 2.2231295011594484E-02_dp, -3.8635690919857713E-03_dp, &
         8.5422617602734633E+01_dp, -1.5007516551248531E+02_dp, 1.3876195898375653E+02_dp, &
         -7.8464142326741523E+01_dp, 2.5573232411746961E+01_dp, -2.0124727833037532E+00_dp, &
         -1.9580157250354535E+00_dp, 5.9506449783845983E-01_dp, 1.4004086634287427E-01_dp, &
         -7.8164255350495107E-02_dp, -2.7687831019897382E-02_dp, 2.4943146447903092E-02_dp, &
         -4.9581832391613840E-03_dp, 3.2243620561770621E+02_dp, -5.5446459400958940E+02_dp, &
         5.0455670468847723E+02_dp, -2.8330528858502907E+02_dp, 9.4651307478009102E+01_dp, &
         -1.1293415812091316E+01_dp, -4.1837250114735332E+00_dp, 1.1914362296732655E+00_dp, &
         4.8166287147916081E-01_dp, -1.4958729633043041E-01_dp, -1.2337653487937358E-01_dp, &
         7.5229238812777835E-02_dp, -1.2091269567906382E-02_dp, 2.4402285772918880E-01_dp, &
         -7.3529180505381958E-02_dp, -2.5723746423618070E-02_dp, 8.9632749393919135E-03_dp, &
         7.8378134914843661E-03_dp, -2.9623515386430414E-03_dp, -3.3201464839523391E-03_dp, &
         1.0078877137589960E-03_dp, 2.5955945748679025E-03_dp, -1.8637178008176481E-03_dp, &
         -2.3022221614225430E-04_dp, 6.5838716850614683E-04_dp, -1.9830435925879039E-04_dp, &
         7.9761273417323364E-01_dp, -2.8289567084615880E-01_dp, -7.7983815660254863E-02_dp, &
         4.6178699662303369E-02_dp, 2.4703159881149541E-02_dp, -1.7384731956342849E-02_dp, &
         -1.0573811255673717E-02_dp, 6.4828830942470356E-03_dp, 1.0745058972820175E-02_dp, &
         -1.2304204780420294E-02_dp, 3.6994644715572381E-03_dp, 7.1771353633863510E-04_dp, &
         -4.7696615443758658E-04_dp, 1.5722119399201502E+00_dp, -7.2022637684883928E-01_dp, &
         -1.1778618034469690E-01_dp, 1.5513182133862150E-01_dp, 3.6741006041689533E-02_dp, &
         -6.4119193681736464E-02_dp, -1.4800413983640810E-02_dp, 2.5585125964366535E-02_dp, &
         2.6791486817367951E-02_dp, -4.9141584177700877E-02_dp, 2.8220269783745606E-02_dp, &
         -6.5502100323003949E-03_dp, 2.6599045470973711E-04_dp, 2.8051373012201362E+00_dp, &
         -1.6968205795059628E+00_dp, -6.8520667990219081E-02_dp, 4.3238094864773730E-01_dp, &
         1.3604877002541815E-03_dp, -1.8671384960156767E-01_dp, 8.8883053763241949E-03_dp, &
         7.8093230403048317E-02_dp, 4.6271896457208229E-02_dp, -1.4709229627835357E-01_dp, &
         1.1306355579348493E-01_dp, -3.9830635885835595E-02_dp, 5.4972776007275979E-03_dp, &
         4.9068374696158550E+00_dp, -3.8649342494189476E+00_dp, 3.3961436603861528E-01_dp, &
         1.0516701804754258E+00_dp, -2.3599107021791954E-01_dp, -4.5209301669418722E-01_dp, &
         1.4887292136073554E-01_dp, 1.9549898360847917E-01_dp, 1.9083808478825574E-02_dp, &
         -3.4003659083377596E-01_dp, 3.3304060399936797E-01_dp, -1.4290330281954144E-01_dp, &
         2.4575713534518952E-02_dp, 8.6327262342711766E+00_dp, -8.6449632701709422E+00_dp, &
         1.9398206133689215E+00_dp, 2.2633523514966289E+00_dp, -1.1404630111455665E+00_dp, &
         -8.9656935075118105E-01_dp, 6.5655059844600916E-01_dp, 3.9178915736248293E-01_dp, &
         -2.5745765584870500E-01_dp, -5.4365659786191212E-01_dp, 7.5377793521530634E-01_dp, &
         -3.8414425007098330E-01_dp, 7.5502828561301874E-02_dp, 1.5429572442140103E+01_dp, &
         -1.9205543991334537E+01_dp, 7.1837956925844102E+00_dp, 4.1527008022568523E+00_dp, &
         -3.9604373888031548E+00_dp, -1.2184620720634423E+00_dp, 2.1268242758491254E+00_dp, &
         4.8977374234146870E-01_dp, -1.3567457043687021E+00_dp, -1.4342868044936791E-01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/1.1553671463453488E+00_dp, &
         -7.6780781804979914E-01_dp, 1.7390307652239254E-01_dp, 2.8165264234356147E+01_dp, &
         -4.2781018965163859E+01_dp, 2.3179056548048951E+01_dp, 5.2439089413556950E+00_dp, &
         -1.1643001503346785E+01_dp, 4.5770373921564939E-01_dp, 5.5851597940226441E+00_dp, &
         -5.5713676796962119E-01_dp, -4.3682925715598220E+00_dp, 2.9977088433118744E+00_dp, &
         1.1551526130399045E-01_dp, -8.4503890035033269E-01_dp, 2.6553468866554275E-01_dp, &
         5.2755173634217869E+01_dp, -9.6327032783315346E+01_dp, 7.0566200135216633E+01_dp, &
         -3.8183788481120411E+00_dp, -2.9391413038382652E+01_dp, 1.1731796742283295E+01_dp, &
         1.0979398073072346E+01_dp, -6.6194752458984993E+00_dp, -9.5048580745993743E+00_dp, &
         1.3808973411209031E+01_dp, -7.1703659500595673E+00_dp, 1.4743357570136750E+00_dp, &
         -1.8794771258963321E-02_dp, 1.0222132432274876E+02_dp, -2.2102822149401919E+02_dp, &
         2.0974623410290044E+02_dp, -6.6317175075640407E+01_dp, -5.5810649104847528E+01_dp, &
         5.5721941193426112E+01_dp, 6.1874002936069470E+00_dp, -2.6416717623490523E+01_dp, &
         -5.5906480645235490E+00_dp, 3.2874933451592149E+01_dp, -2.8094658694055632E+01_dp, &
         1.1072416668088719E+01_dp, -1.7823600622119833E+00_dp, 2.1028466549514980E+02_dp, &
         -5.2642609969482317E+02_dp, 6.2125118053900360E+02_dp, -3.5670022652999478E+02_dp, &
         -1.1979421659544816E+01_dp, 1.6071306995897604E+02_dp, -6.8887503091643865E+01_dp, &
         -4.4648249302777593E+01_dp, 4.3014857970719397E+01_dp, 1.6335752686095919E+01_dp, &
         -3.9503484670996066E+01_dp, 2.2138801724899430E+01_dp, -4.5544474926067293E+00_dp, &
         5.2271868060407382E+02_dp, -1.4526730604347720E+03_dp, 2.0000601499106572E+03_dp, &
         -1.5790696077296959E+03_dp, 5.6252542900631101E+02_dp, 1.8000459328370113E+02_dp, &
         -2.6699351388563537E+02_dp, 3.0766701448493226E+01_dp, 9.2468215263256795E+01_dp, &
         -3.9890430775649456E+01_dp, -2.4228475522268266E+01_dp, 2.5320986601337847E+01_dp, &
         -6.4467381198271339E+00_dp, 1.9943221690264703E+00_dp, -6.2720822070394933E-01_dp, &
         -2.0718877314833953E-01_dp, 8.3665992635281738E-02_dp, 6.3970965828668153E-02_dp, &
         -2.8909340921094925E-02_dp, -2.7244049755787644E-02_dp, 1.0154578919302151E-02_dp, &
         2.2755693177166175E-02_dp, -1.8948581830615388E-02_dp, 6.0818547316571962E-04_dp, &
         4.6474201410700100E-03_dp, -1.5474535300080714E-03_dp, 2.5344309071612208E+00_dp, &
         -1.1517789229917388E+00_dp, -2.0475816431506066E-01_dp, 2.4704626449142972E-01_dp, &
         6.8510892925257152E-02_dp, -1.0144220474534189E-01_dp, -2.9276180034024692E-02_dp, &
         3.9983114315973144E-02_dp, 4.6113046413331867E-02_dp, -7.6975277049864677E-02_dp, &
         4.0552717012763861E-02_dp, -7.6506455032915865E-03_dp, -1.9384086829425090E-04_dp, &
         3.8149008989961728E+00_dp, -2.5425820634363938E+00_dp, -6.9974141904847256E-02_dp, &
         6.8655295447602072E-01_dp, 3.7667761955125154E-03_dp, -3.0089857114650986E-01_dp, &
         9.5441871944521653E-03_dp, 1.2570484744488805E-01_dp, 8.1291754180774958E-02_dp, &
         -2.4131410671430067E-01_dp, 1.8097245043205007E-01_dp, -6.2022990804356130E-02_dp, &
         8.2193545283439501E-03_dp, 6.3114819275149774E+00_dp, -5.6792022806645264E+00_dp, &
         6.2279323003148646E-01_dp, 1.6459331144460243E+00_dp, -3.7800717585122690E-01_dp, &
         -7.2663629659756079E-01_dp, 2.3264111653024058E-01_dp, 3.1646630721128349E-01_dp, &
         4.7657307643253675E-02_dp, -5.7090704941545778E-01_dp, 5.4910822545064875E-01_dp, &
         -2.3284314940812098E-01_dp, 3.9588194044005813E-02_dp, 1.0984388893348632E+01_dp, &
         -1.2553778627354628E+01_dp, 3.1375915699757435E+00_dp, 3.4851328166564612E+00_dp, &
         -1.8017015924479218E+00_dp, -1.4359380873215706E+00_dp, 1.0422397069450537E+00_dp, &
         6.4072298568256425E-01_dp, -3.8510916515210597E-01_dp, -9.4069946857544329E-01_dp, &
         1.2691484307865526E+00_dp, -6.4142347298575852E-01_dp, 1.2542319982761238E-01_dp, &
         1.9742724664081642E+01_dp, -2.7660695606300102E+01_dp, 1.1126190485661816E+01_dp, &
         6.3320842667644364E+00_dp, -6.2040316849472417E+00_dp, -1.9845491148459973E+00_dp, &
         3.4016074403877203E+00_dp, 8.4883284732790321E-01_dp, -2.1910518713819105E+00_dp, &
         -3.4310026768968166E-01_dp, 2.0043227872868097E+00_dp, -1.3129608772761314E+00_dp, &
         2.9596309142603439E-01_dp, 3.6446675442084000E+01_dp, -6.1312690384351988E+01_dp, &
         3.5062440706052982E+01_dp, 8.1141251417433384E+00_dp, -1.8238397593931307E+01_dp, &
         4.2444498167686012E-01_dp, 9.0917913358856648E+00_dp, -6.4145530970932474E-01_dp, &
         -7.3727065993355358E+00_dp, 4.8554897125262979E+00_dp, 4.0288046417382256E-01_dp, &
         -1.5272111598694897E+00_dp, 4.6864448241919043E-01_dp, 6.9151902111365331E+01_dp, &
         -1.3775697029732524E+02_dp, 1.0517986253443161E+02_dp, -4.2845246310611023E+00_dp, &
         -4.6935409816300506E+01_dp, 1.7360744079299344E+01_dp, 1.9042670125521109E+01_dp, &
         -1.0054681312184911E+01_dp, -1.7340855643857253E+01_dp, 2.3641147423075047E+01_dp, &
         -1.1697618310899770E+01_dp, 2.1262428960669113E+00_dp, 6.0872426969359870E-02_dp, &
         1.3558768380220121E+02_dp, -3.1578118874680166E+02_dp, 3.0979445611347273E+02_dp, &
         -9.1822777560522482E+01_dp, -9.7277945816883886E+01_dp, 8.7519582106791688E+01_dp, &
         1.8616391208601357E+01_dp, -4.4903204542560921E+01_dp, -1.8690573605516008E+01_dp, &
         6.6433565837523219E+01_dp, -5.3403019488106459E+01_dp, 2.0231742919655424E+01_dp, &
         -3.1305074294263489E+00_dp, 2.7963732579939841E+02_dp, -7.4606125632766100E+02_dp, &
         9.0940141987222216E+02_dp, -5.0646326675575170E+02_dp, -7.8609616213890064E+01_dp, &
         2.9311367852905386E+02_dp, -9.4979204093817344E+01_dp, -1.1347623458453775E+02_dp, &
         8.2031344824942224E+01_dp, 5.6241860637037483E+01_dp, -1.0224730452002980E+02_dp, &
         5.5340684759985947E+01_dp, -1.1297342601063139E+01_dp, 6.5120722704318109E+02_dp, &
         -1.9033734753449851E+03_dp, 2.7232474210995215E+03_dp, -2.1823518641635665E+03_dp, &
         6.8086154256333919E+02_dp, 4.7729134285430774E+02_dp, -5.7292260599128042E+02_dp, &
         6.1733473299802547E+01_dp, 2.6831002853289289E+02_dp, -2.0906396401698387E+02_dp, &
         4.5742535109750555E+01_dp, 1.4910878567831897E+01_dp, -7.1528632268068026E+00_dp, &
         2.4115194635293756E+03_dp, -7.1360888524468037E+03_dp, 1.0703481900930154E+04_dp, &
         -9.7696053852248097E+03_dp, 5.1324020712244110E+03_dp, -6.9246367124994219E+02_dp, &
         -9.3295400850484577E+02_dp, 4.2973937177886751E+02_dp, 2.2651805907394606E+02_dp, &
         -2.3513604873721363E+02_dp, 1.2713777537393858E+01_dp, 5.2470351852753829E+01_dp, &
         -1.7444400068472273E+01_dp, 2.8440161633359956E-01_dp, -3.9823006517888213E-02_dp, &
         -2.3076160901873654E-02_dp, 2.8993315015824581E-03_dp, 6.3510245125292433E-03_dp, &
         -1.8708028901690619E-05_dp, -3.7414722349339334E-03_dp, 5.3858115549491718E-04_dp, &
         2.0941650894001022E-03_dp, -9.3028620337434742E-04_dp, -7.2074637484255898E-04_dp, &
         7.4135593147601206E-04_dp, -1.9162678311097998E-04_dp, 9.5981147138166634E-01_dp, &
         -1.6725087562945692E-01_dp, -8.9182688444622890E-02_dp, 2.0572731901361569E-02_dp, &
         2.7187174656436549E-02_dp, -3.6082603824213978E-03_dp, -1.6627641029942530E-02_dp, &
         3.7750956656711155E-03_dp, 1.0628480501116633E-02_dp, -7.1153643941278205E-03_dp, &
         -1.2143377855200034E-03_dp, 2.7012120466684050E-03_dp, -7.8946988266133150E-04_dp, &
         2.0129964635290527E+00_dp, -4.8454573123042377E-01_dp, -2.2604994573980364E-01_dp, &
         9.2677928858034012E-02_dp, 7.7785406032969739E-02_dp, -2.5979505885131083E-02_dp, &
         -5.0328987859669862E-02_dp, 1.7997890535406402E-02_dp, 3.8319048248298730E-02_dp, &
         -3.6325104322509204E-02_dp, 6.7213705329095992E-03_dp, 4.8397181344369666E-03_dp, &
         -1.9887682453394975E-03_dp, 3.9222318106291012E+00_dp, -1.3148639911853994E+00_dp, &
         -5.1737762994905812E-01_dp, 3.4039821500135192E-01_dp, 1.9442866605932332E-01_dp, &
         -1.1862440428797107E-01_dp, -1.3471081518767758E-01_dp, 7.2722253898947764E-02_dp, &
         1.2394645141103253E-01_dp, -1.5424078283263418E-01_dp, 6.0957620821992728E-02_dp, &
         -2.0727719708758920E-03_dp, -3.2589949122940306E-03_dp, 7.6600225072517931E+00_dp, &
         -3.4806043928458892E+00_dp, -1.0922250581494755E+00_dp, 1.1238446877190527E+00_dp, &
         4.2100547038706426E-01_dp, -4.5205119802000077E-01_dp, -3.1961008012678949E-01_dp, &
         2.6922377501542977E-01_dp, 3.6631461573210744E-01_dp, -5.8790288209541941E-01_dp, &
         3.2417909023377850E-01_dp, -7.2737310780591674E-02_dp, 2.5717607245155992E-03_dp, &
         1.5357195767319828E+01_dp, -9.1855216845760417E+00_dp, -2.0474502120861211E+00_dp, &
         3.5186908819973559E+00_dp, 6.9067364763413841E-01_dp, -1.5707666205541846E+00_dp, &
         -6.2320076264913815E-01_dp, 9.5548605981562995E-01_dp, 9.6164257386927365E-01_dp, &
         -2.0694392628725198E+00_dp, 1.4288052626518093E+00_dp, -4.6175690844427097E-01_dp, &
         5.7161598844966313E-02_dp, 3.1967087548643718E+01_dp, -2.4634023685889019E+01_dp, &
         -2.8379221339115919E+00_dp, 1.0758070470373076E+01_dp, 9.5880676096911616E-02_dp, &
         -5.1371158552123646E+00_dp, -6.1308866074368273E-01_dp, 3.3299562667716351E+00_dp, &
         1.9037543102815111E+00_dp, -6.6312995485896993E+00_dp, 5.5769281199888221E+00_dp, &
         -2.1992036324059452E+00_dp, 3.5304429234184187E-01_dp, 6.9739204456194557E+01_dp, &
         -6.8387481704787120E+01_dp, 1.1321639890564044E+00_dp, 3.2573484471282640E+01_dp, &
         -6.9804580122418587E+00_dp, -1.5736074279056968E+01_dp, 2.9654243601725896E+00_dp, &
         1.1422619493838827E+01_dp, -2.6727212260300343E-01_dp, -1.7588114164311783E+01_dp, &
         1.8884242504372157E+01_dp, -8.7648445734311817E+00_dp, 1.6290708530849365E+00_dp, &
         1.6128088895603187E+02_dp, -2.0051884945874767E+02_dp, 3.6821045909182011E+01_dp, &
         9.7251659706901165E+01_dp, -5.0845173213144840E+01_dp, -4.1755777685783144E+01_dp, &
         2.8048895067605439E+01_dp, 3.6657246361996478E+01_dp, -3.4007473970245094E+01_dp, &
         -2.2004679796240055E+01_dp, 4.6900402309890616E+01_dp, -2.7316097181854502E+01_dp, &
         5.8577092013741536E+00_dp, 4.0194017966122044E+02_dp, -6.3674119443127381E+02_dp, &
         2.5850957390809663E+02_dp, 2.6952251874150846E+02_dp, -2.9073790305138749E+02_dp, &
         -5.2677944928443630E+01_dp, 1.5741217884382442E+02_dp, 8.1820204921580611E+01_dp, &
         -2.5405433230660509E+02_dp, 1.5943577712260654E+02_dp, -4.6996586786307697E+00_dp, &
         -3.5301765944409738E+01_dp, 1.1655632510420780E+01_dp, 1.1102009036442944E+03_dp, &
         -2.2680830885994883E+03_dp, 1.6376815639623894E+03_dp, 4.1340967237756740E+02_dp, &
         -1.5047177979548774E+03_dp, 6.2874431977984602E+02_dp, 5.3667599466253898E+02_dp, &
         -2.6715850493363729E+02_dp, -9.0644414621246210E+02_dp, 1.4354167004689343E+03_dp, &
         -9.8255347076517205E+02_dp, 3.4531439529808881E+02_dp, -5.0862246712184223E+01_dp, &
         3.7086333759961367E+03_dp, -9.7206127257815970E+03_dp, 1.1363579106801713E+04_dp, &
         -5.0930367254613293E+03_dp, -3.8587254648260623E+03_dp, 7.5131622983303614E+03_dp, &
         -4.5750330568023464E+03_dp, 4.8521172705364671E+02_dp, 4.7555952848897863E+02_dp, &
         6.7920802819038545E+02_dp, -1.2191938699136147E+03_dp, 6.8845273583955679E+02_dp, &
         -1.4522061491176112E+02_dp, 2.3426708866702475E+00_dp, -3.4763950540505012E-01_dp, &
         -1.9719523876834619E-01_dp, 3.0016600517614064E-02_dp, 5.5870741584836549E-02_dp, &
         -2.0603641124710137E-03_dp, -3.3236739085782500E-02_dp, 5.4972021666434001E-03_dp, &
         1.9287416985295423E-02_dp, -9.7735471490287135E-03_dp, -5.4153420961059818E-03_dp, &
         6.2996178645704911E-03_dp, -1.6741808988248179E-03_dp, 3.2323307090443092E+00_dp, &
         -7.5868899537917556E-01_dp, -3.6403112314198777E-01_dp, 1.3638032105863709E-01_dp, &
         1.2454451915538212E-01_dp, -3.5272622120219609E-02_dp, -7.9465036314149431E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/2.5411208403973456E-02_dp, &
         5.7808629859203861E-02_dp, -5.0196351778645114E-02_dp, 5.2149034144752344E-03_dp, &
         9.5271530622531974E-03_dp, -3.3982153227889275E-03_dp, 5.4948218832066447E+00_dp, &
         -1.9768544283573861E+00_dp, -7.8540931963005123E-01_dp, 5.1224166160961682E-01_dp, &
         3.0432975281369379E-01_dp, -1.7343930124124304E-01_dp, -2.0939874165255690E-01_dp, &
         1.0529951288679236E-01_dp, 1.8723710201929586E-01_dp, -2.2094924684309869E-01_dp, &
         7.8605553221171381E-02_dp, 3.2239680553835400E-03_dp, -6.1056020491761899E-03_dp, &
         1.0388070420621753E+01_dp, -5.1639778369451559E+00_dp, -1.6336599184903002E+00_dp, &
         1.6828507996508695E+00_dp, 6.6282593487173314E-01_dp, -6.6705890421444214E-01_dp, &
         -5.0042252719289448E-01_dp, 3.9062209984428897E-01_dp, 5.5969261337300114E-01_dp, &
         -8.5374958521504130E-01_dp, 4.4514864411394023E-01_dp, -8.7078324077810784E-02_dp, &
         -6.3901198934322614E-04_dp, 2.0837835829959378E+01_dp, -1.3485610187672910E+01_dp, &
         -3.0562817152450337E+00_dp, 5.2210972603116508E+00_dp, 1.1380806549217413E+00_dp, &
         -2.3202212441266843E+00_dp, -1.0087684674158810E+00_dp, 1.3808656481997015E+00_dp, &
         1.5160097727120327E+00_dp, -3.0497799109640908E+00_dp, 2.0185152497063346E+00_dp, &
         -6.1672935366387194E-01_dp, 6.9010721045408896E-02_dp, 4.3786529696255222E+01_dp, &
         -3.5837705535464202E+01_dp, -4.3770867029967091E+00_dp, 1.5847699276648193E+01_dp, &
         5.9230278789709845E-01_dp, -7.6207423069692428E+00_dp, -1.2403077215715186E+00_dp, &
         4.7968232553191292E+00_dp, 3.3649722547708287E+00_dp, -1.0082904709615054E+01_dp, &
         8.1126465226757194E+00_dp, -3.0773616343906505E+00_dp, 4.7309562815703760E-01_dp, &
         9.6534365122504383E+01_dp, -9.8600884120947967E+01_dp, 2.9621706296987543E-01_dp, &
         4.7831849282858727E+01_dp, -8.1795187324838583E+00_dp, -2.3764904478653722E+01_dp, &
         2.9262740978603858E+00_dp, 1.6552408974929406E+01_dp, 2.8091737862874124E+00_dp, &
         -2.9119085060489422E+01_dp, 2.9028575924228921E+01_dp, -1.2928669504768751E+01_dp, &
         2.3245151888545674E+00_dp, 2.2512257878709937E+02_dp, -2.8579384581407447E+02_dp, &
         4.4953255218681591E+01_dp, 1.4382256723080877E+02_dp, -6.3886295794702583E+01_dp, &
         -6.7487344867895843E+01_dp, 3.4987698625952298E+01_dp, 5.5556684237578004E+01_dp, &
         -3.3788647650616660E+01_dp, -5.6043553929673735E+01_dp, 8.4310762498784698E+01_dp, &
         -4.5100448710609861E+01_dp, 9.2349413542319656E+00_dp, 5.6281441264089858E+02_dp, &
         -8.9092661108477409E+02_dp, 3.1621440639103065E+02_dp, 4.1867392990338300E+02_dp, &
         -3.6830256822695566E+02_dp, -1.3786095760420562E+02_dp, 2.1164276757772936E+02_dp, &
         1.5857658366809042E+02_dp, -3.2038368617142976E+02_dp, 1.0004479022816162E+02_dp, &
         1.0938350249740985E+02_dp, -1.0143008716163932E+02_dp, 2.5908560637399010E+01_dp, &
         1.5406066303216392E+03_dp, -3.0712567653270539E+03_dp, 1.9107443589577488E+03_dp, &
         9.8795055555435533E+02_dp, -1.9737417466756906E+03_dp, 2.8775990611293059E+02_dp, &
         9.9019328005226532E+02_dp, 7.8824363513483021E+01_dp, -1.7670587145406548E+03_dp, &
         1.9470064745111699E+03_dp, -9.6841150263156567E+02_dp, 2.0775921744212744E+02_dp, &
         -6.8055324053062165E+00_dp, 4.8137396326482822E+03_dp, -1.2191613808338789E+04_dp, &
         1.2266285257455142E+04_dp, -1.4891625095205279E+03_dp, -9.2959576828369536E+03_dp, &
         8.2521109655672644E+03_dp, 9.7025008633460880E+02_dp, -4.8188657990924448E+03_dp, &
         -1.0477423269137444E+03_dp, 7.7758077154163093E+03_dp, -7.7950611646956877E+03_dp, &
         3.6117091083907817E+03_dp, -6.8647813181077140E+02_dp, 2.0868376196912817E+04_dp, &
         -6.3171096044027225E+04_dp, 9.2357701118304758E+04_dp, -7.2478877239780646E+04_dp, &
         1.0873910093902665E+04_dp, 4.7193165269176498E+04_dp, -6.6614746747051075E+04_dp, &
         5.1476389226075742E+04_dp, -2.7387825042622444E+04_dp, 1.1182959389903041E+04_dp, &
         -3.8548989260871235E+03_dp, 1.0658643031361141E+03_dp, -1.6150497978759398E+02_dp, &
         2.9848190654960571E-01_dp, -1.4531646368209182E-02_dp, -1.3349182894683412E-02_dp, &
         4.3341324342453481E-03_dp, -4.5914461419488779E-03_dp, 7.5114003286678209E-03_dp, &
         -8.1777684041123705E-03_dp, 7.6589093791208356E-03_dp, -6.9685324431716066E-03_dp, &
         5.4975257982100012E-03_dp, -3.1992401573521976E-03_dp, 1.1765944960906449E-03_dp, &
         -2.0399784249967458E-04_dp, 1.0199389612090075E+00_dp, -6.3105771054488224E-02_dp, &
         -5.6858180138521432E-02_dp, 2.0836254442772213E-02_dp, -1.9744550724843178E-02_dp, &
         3.2668133118974374E-02_dp, -3.5931234657235918E-02_dp, 3.3317168158417933E-02_dp, &
         -3.0032438804812003E-02_dp, 2.3676679695673334E-02_dp, -1.3795044313291295E-02_dp, &
         5.0675158738836823E-03_dp, -8.7502220183366809E-04_dp, 2.1914353100451938E+00_dp, &
         -1.9179798443558940E-01_dp, -1.6799355534241783E-01_dp, 7.1931114531331136E-02_dp, &
         -5.9484431545979891E-02_dp, 9.9336123697354881E-02_dp, -1.1099688936746152E-01_dp, &
         1.0155478022832697E-01_dp, -9.0225752064209283E-02_dp, 7.1039085577344563E-02_dp, &
         -4.1492188330594100E-02_dp, 1.5231043613680471E-02_dp, -2.6165816928386950E-03_dp, &
         4.4195039100565916E+00_dp, -5.4858126359256787E-01_dp, -4.6531366469875657E-01_dp, &
         2.3227879552470015E-01_dp, -1.7002004191404002E-01_dp, 2.8353543550268284E-01_dp, &
         -3.2295870972875435E-01_dp, 2.9162100501061888E-01_dp, -2.5446398227605715E-01_dp, &
         1.9990573122381708E-01_dp, -1.1722136047578775E-01_dp, 4.3074512262631613E-02_dp, &
         -7.3698908494786572E-03_dp, 9.0158149241474579E+00_dp, -1.5390032846184512E+00_dp, &
         -1.2578078408173032E+00_dp, 7.3256099290466847E-01_dp, -4.8255480252021304E-01_dp, &
         7.9110935203918809E-01_dp, -9.2223368631764924E-01_dp, 8.2261801505822885E-01_dp, &
         -7.0152375312903192E-01_dp, 5.4882264799559055E-01_dp, -3.2374652428183265E-01_dp, &
         1.1943745886308012E-01_dp, -2.0399323236584468E-02_dp, 1.9060405525354774E+01_dp, &
         -4.3442410134014242E+00_dp, -3.3923750244595929E+00_dp, 2.3277758289827348E+00_dp, &
         -1.4032292783490181E+00_dp, 2.2111149370970846E+00_dp, -2.6514360156752161E+00_dp, &
         2.3400879330259072E+00_dp, -1.9352609112533496E+00_dp, 1.5013805536357132E+00_dp, &
         -8.9331348604117622E-01_dp, 3.3245002695248244E-01_dp, -5.6923036199617680E-02_dp, &
         4.2321796419078780E+01_dp, -1.2635737527625851E+01_dp, -9.3019892409007774E+00_dp, &
         7.6613108253972051E+00_dp, -4.3221837921659887E+00_dp, 6.3307699962540935E+00_dp, &
         -7.8520270548486826E+00_dp, 6.8807442180215030E+00_dp, -5.4525567047496342E+00_dp, &
         4.1550041539563418E+00_dp, -2.5001930251736764E+00_dp, 9.4579959622261767E-01_dp, &
         -1.6363574334611850E-01_dp, 1.0003827134700703E+02_dp, -3.8829042199894573E+01_dp, &
         -2.6360290750789250E+01_dp, 2.6857452022470046E+01_dp, -1.4676880411075732E+01_dp, &
         1.9015127170909437E+01_dp, -2.4479606083190603E+01_dp, 2.1453660248266239E+01_dp, &
         -1.5990524893902380E+01_dp, 1.1692590213258955E+01_dp, -7.1089925888310566E+00_dp, &
         2.7650175937063479E+00_dp, -4.9013357258124818E-01_dp, 2.5650279105853537E+02_dp, &
         -1.3001613271097199E+02_dp, -7.8215890320689127E+01_dp, 1.0374859376545058E+02_dp, &
         -5.7993318532586997E+01_dp, 6.1925129027835808E+01_dp, -8.2129941440824297E+01_dp, &
         7.2996479221982142E+01_dp, -4.9711740845864256E+01_dp, 3.2670927648676113E+01_dp, &
         -1.9601482847566309E+01_dp, 7.9400385997630769E+00_dp, -1.4745807885995803E+00_dp, &
         7.3530003145728017E+02_dp, -4.9664201976979172E+02_dp, -2.4279101607234665E+02_dp, &
         4.6398305934827602E+02_dp, -2.8940008684901966E+02_dp, 2.3403868192771887E+02_dp, &
         -3.0250596676982525E+02_dp, 2.7884938757156840E+02_dp, -1.6597480330952183E+02_dp, &
         7.7022134764390358E+01_dp, -3.6333234939720029E+01_dp, 1.4834832375776335E+01_dp, &
         -3.0059213601888550E+00_dp, 2.4925639619878293E+03_dp, -2.3516399736727049E+03_dp, &
         -7.0669849118397849E+02_dp, 2.6123262478824795E+03_dp, -2.1096774881906767E+03_dp, &
         1.2752227522425612E+03_dp, -1.2190189717372978E+03_dp, 1.1433571541782767E+03_dp, &
         -5.3704043024832720E+02_dp, -8.5050052992266330E+01_dp, 2.6432025336564107E+02_dp, &
         -1.4590397426454322E+02_dp, 3.0667907896264261E+01_dp, 1.1452856000529706E+04_dp, &
         -1.6733730631591166E+04_dp, 2.0124842686447407E+03_dp, 2.0782792338251493E+04_dp, &
         -3.1172495661780686E+04_dp, 2.3937895402534603E+04_dp, -8.4050963154502351E+03_dp, &
         -5.8783772978304578E+03_dp, 1.4993404742007511E+04_dp, -1.7419881693507228E+04_dp, &
         1.3154166369648410E+04_dp, -6.0320093070800103E+03_dp, 1.2691258478430473E+03_dp, &
         2.4661275214127114E+00_dp, -1.2798267129549132E-01_dp, -1.1697137480153171E-01_dp, &
         3.9257803663157738E-02_dp, -4.0319593020954211E-02_dp, 6.6187689447338618E-02_dp, &
         -7.2247928359382055E-02_dp, 6.7482097544372566E-02_dp, -6.1252477865463867E-02_dp, &
         4.8314626699507028E-02_dp, -2.8123924424198074E-02_dp, 1.0339419858641128E-02_dp, &
         -1.7906361701448847E-03_dp, 3.5103259814502201E+00_dp, -2.9727217547633278E-01_dp, &
         -2.6205887594376975E-01_dp, 1.0857238160676247E-01_dp, -9.2141383796487078E-02_dp, &
         1.5406834578313486E-01_dp, -1.7142772300076523E-01_dp, 1.5726488808531711E-01_dp, &
         -1.4025477042228404E-01_dp, 1.1047969889471382E-01_dp, -6.4470419314857599E-02_dp, &
         2.3658250747070033E-02_dp, -4.0673029740227054E-03_dp, 6.2410599788925678E+00_dp, &
         -8.2138392898720347E-01_dp, -6.9906495654527301E-01_dp, 3.4410291328684828E-01_dp, &
         -2.5343941141731630E-01_dp, 4.2515725467164173E-01_dp, -4.8284871607865038E-01_dp, &
         4.3629017443259160E-01_dp, -3.8177479405146436E-01_dp, 3.0011747927293064E-01_dp, &
         -1.7581115021036545E-01_dp, 6.4536453913093872E-02_dp, -1.1036630823958494E-02_dp, &
         1.2396354706697132E+01_dp, -2.2747834976246768E+00_dp, -1.8659037974661266E+00_dp, &
         1.0729486834844844E+00_dp, -7.0823291387398257E-01_dp, 1.1716255191494653E+00_dp, &
         -1.3616528360542641E+00_dp, 1.2146420854278106E+00_dp, -1.0394227304439734E+00_dp, &
         8.1420256203015140E-01_dp, -4.7970758970705091E-01_dp, 1.7666567904333222E-01_dp, &
         -3.0134269374265207E-02_dp, 2.6262194331493351E+01_dp, -6.3444832902403547E+00_dp, &
         -4.9805965392317857E+00_dp, 3.3626392280343960E+00_dp, -2.0268609276880918E+00_dp, &
         3.2376830951333413E+00_dp, -3.8685734302070061E+00_dp, 3.4125011524571320E+00_dp, &
         -2.8364792541794244E+00_dp, 2.2064813341891907E+00_dp, -1.3109104909016629E+00_dp, &
         4.8647143855153607E-01_dp, -8.3081066217279925E-02_dp, 5.8791642664628654E+01_dp, &
         -1.8223978038331911E+01_dp, -1.3535672130713879E+01_dp, 1.0886746267759019E+01_dp, &
         -6.1196900002518406E+00_dp, 9.1643774491776391E+00_dp, -1.1318457824166819E+01_dp, &
         9.9022716234880352E+00_dp, -7.9117201762867282E+00_dp, 6.0668367158067333E+00_dp, &
         -3.6456805775684451E+00_dp, 1.3726588060126073E+00_dp, -2.3632132863188687E-01_dp, &
         1.3990455162403524E+02_dp, -5.5127425979272914E+01_dp, -3.8049785729902737E+01_dp, &
         3.7322312094101129E+01_dp, -2.0162618467243483E+01_dp, 2.7134468713776865E+01_dp, &
         -3.4792138438282116E+01_dp, 3.0368635438402379E+01_dp, -2.2961340267827463E+01_dp, &
         1.7064301226195642E+01_dp, -1.0388397417931811E+01_dp, 4.0108403964288808E+00_dp, &
         -7.0448220676099105E-01_dp, 3.5898010619414435E+02_dp, -1.8010251190337905E+02_dp, &
         -1.1207515297883917E+02_dp, 1.3926745663661529E+02_dp, -7.5591370594985008E+01_dp, &
         8.6197068058446064E+01_dp, -1.1457176948523994E+02_dp, 1.0089516510941472E+02_dp, &
         -7.0493679332727382E+01_dp, 4.8625058388200841E+01_dp, -2.9732088584351676E+01_dp, &
         1.1947228503152537E+01_dp, -2.1840920197937304E+00_dp, 1.0166774272757106E+03_dp, &
         -6.5774062040784861E+02_dp, -3.4819173908900984E+02_dp, 5.8698390708558020E+02_dp/)
      REAL(KIND=dp), DIMENSION(13, 24, 5), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04/), (/13, 24, 5/))

      INTEGER                                            :: irange

      IF (Rc <= 1.0000000000000000E+01_dp) Rc = 1.0000000000000000E+01_dp
      IF (Rc >= 8.5914001464843750E+04_dp) Rc = 8.5914001464843750E+04_dp
      IF (Rc >= 1.0000000000000000E+01_dp .AND. Rc <= 3.1589054611062615E+01_dp) THEN
         irange = 1
         L_b = 1.0000000000000000E+01_dp
         U_b = 3.1589054611062615E+01_dp
      END IF
      IF (Rc >= 3.1589054611062615E+01_dp .AND. Rc <= 3.7775485225370909E+02_dp) THEN
         irange = 2
         L_b = 3.1589054611062615E+01_dp
         U_b = 3.7775485225370909E+02_dp
      END IF
      IF (Rc >= 3.7775485225370909E+02_dp .AND. Rc <= 5.0208914580369265E+03_dp) THEN
         irange = 3
         L_b = 3.7775485225370909E+02_dp
         U_b = 5.0208914580369265E+03_dp
      END IF
      IF (Rc >= 5.0208914580369265E+03_dp .AND. Rc <= 2.9478561984966764E+04_dp) THEN
         irange = 4
         L_b = 5.0208914580369265E+03_dp
         U_b = 2.9478561984966764E+04_dp
      END IF
      IF (Rc >= 2.9478561984966764E+04_dp .AND. Rc <= 8.5914001464843750E+04_dp) THEN
         irange = 5
         L_b = 2.9478561984966764E+04_dp
         U_b = 8.5914001464843750E+04_dp
      END IF

      fit_coef(1:13, 1:24) = coefdata(1:13, 1:24, irange)

   END SUBROUTINE get_fit_coef_k12

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k13(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 26)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.8838771004380783E-01_dp, &
         -7.8098504554216852E-02_dp, -1.5170263296839995E-02_dp, 8.2399169935392508E-03_dp, &
         3.5286892847422469E-03_dp, -1.8215194066877041E-03_dp, -9.0884305814463765E-04_dp, &
         6.8756727909990219E-04_dp, 3.7413977081317160E-04_dp, -2.6705465221031334E-04_dp, &
         -2.0802563944277461E-04_dp, 2.1290418502755695E-04_dp, -5.2523675707973026E-05_dp, &
         5.9434702831532238E-01_dp, -2.7182658534478105E-01_dp, -3.5999729360206109E-02_dp, &
         3.2340253606252901E-02_dp, 7.9358310080228263E-03_dp, -7.7844468594381001E-03_dp, &
         -1.8014703887347175E-03_dp, 2.8423929944335724E-03_dp, 9.1015987102848297E-04_dp, &
         -1.4560488645141763E-03_dp, -7.3222872823034983E-05_dp, 4.8860545790434685E-04_dp, &
         -1.5085140620238292E-04_dp, 1.0937368125504157E+00_dp, -5.9132092667529690E-01_dp, &
         -1.8412335702747615E-02_dp, 7.9331829469009338E-02_dp, 1.0973664242275066E-03_dp, &
         -2.0316082349166990E-02_dp, 1.2853862027112464E-03_dp, 7.0775762212150883E-03_dp, &
         3.4751426157424701E-04_dp, -4.5696828146617506E-03_dp, 2.1223196342842286E-03_dp, &
         1.0726013520497198E-05_dp, -1.5736650409045057E-04_dp, 1.7694613840435505E+00_dp, &
         -1.1676651668520612E+00_dp, 1.0071479256453719E-01_dp, 1.6228303909602290E-01_dp, &
         -3.6185203797936424E-02_dp, -4.1378623218189241E-02_dp, 1.5022662586176605E-02_dp, &
         1.3484748664965476E-02_dp, -3.6199651811571775E-03_dp, -1.0288246654012838E-02_dp, &
         8.7176610157029440E-03_dp, -2.6010672222318890E-03_dp, 2.1442841577039209E-04_dp, &
         2.7395947604311148E+00_dp, -2.2168158237937692E+00_dp, 4.6058197011790003E-01_dp, &
         2.8018864797075105E-01_dp, -1.4313062865562740E-01_dp, -6.4381187850727661E-02_dp, &
         5.1866744842633442E-02_dp, 1.8402911677378096E-02_dp, -1.5630053920933622E-02_dp, &
         -1.6666410769786210E-02_dp, 2.1958245525177594E-02_dp, -9.4695529478873997E-03_dp, &
         1.5123425133968700E-03_dp, 4.1819739103477458E+00_dp, -4.1085299248060032E+00_dp, &
         1.3587434487195018E+00_dp, 3.8010926364059955E-01_dp, -3.8974887265449265E-01_dp, &
         -5.7767530618699811E-02_dp, 1.3019397541063171E-01_dp, 8.6445563808137237E-03_dp, &
         -4.2595670770739558E-02_dp, -1.5008742214007073E-02_dp, 4.0771284378325862E-02_dp, &
         -2.2393274717496293E-02_dp, 4.4193958726736074E-03_dp, 6.3740799341119825E+00_dp, &
         -7.4856748747981650E+00_dp, 3.4153135374383137E+00_dp, 2.5465839179415151E-01_dp, &
         -8.7770919045013052E-01_dp, 7.1377608346000626E-02_dp, 2.6510553418483845E-01_dp, &
         -4.9379331756714537E-02_dp, -8.9569832352192805E-02_dp, 1.6019869840805893E-02_dp, &
         5.4400827506418850E-02_dp, -4.0239804495804533E-02_dp, 9.3312689304133906E-03_dp, &
         9.7621486867205380E+00_dp, -1.3475250019084454E+01_dp, 7.8972763434980600E+00_dp, &
         -7.1565769086151843E-01_dp, -1.6844459659464686E+00_dp, 5.4494023769013600E-01_dp, &
         4.2698529938918661E-01_dp, -2.2418651826182923E-01_dp, -1.4388654754391872E-01_dp, &
         1.1755325675552541E-01_dp, 2.8878081758365690E-02_dp, -5.2435038016586624E-02_dp, &
         1.5197313495000905E-02_dp, 1.5092987100272357E+01_dp, -2.4074749737500255E+01_dp, &
         1.7349987294818295E+01_dp, -4.1717529748993130E+00_dp, -2.5981745668440559E+00_dp, &
         1.7898646177351853E+00_dp, 4.1191953332704456E-01_dp, -6.1075785276949601E-01_dp, &
         -1.2882271191956388E-01_dp, 3.4442105482404745E-01_dp, -1.1990741103975205E-01_dp, &
         -1.9713219340545846E-02_dp, 1.5129094650791500E-02_dp, 2.3710574140122770E+01_dp, &
         -4.2934018545625399E+01_dp, 3.6833617889420324E+01_dp, -1.4114831970803618E+01_dp, &
         -2.2472542537243165E+00_dp, 4.3088109269149406E+00_dp, -4.2518679827460348E-01_dp, &
         -1.1804673921484345E+00_dp, 1.7962982669589683E-01_dp, 6.4586932424038102E-01_dp, &
         -4.9112812550688612E-01_dp, 1.3546324224979389E-01_dp, -9.3472046563761128E-03_dp, &
         3.8416250559670239E+01_dp, -7.7380918682867062E+01_dp, 7.6669856574547836E+01_dp, &
         -3.9692668605817900E+01_dp, 4.0800907639980108E+00_dp, 7.6431547349679390E+00_dp, &
         -3.2426728324429890E+00_dp, -1.3282905867506054E+00_dp, 1.0159546829067250E+00_dp, &
         6.3516967916479361E-01_dp, -9.3377190156086576E-01_dp, 4.1042024028249102E-01_dp, &
         -6.7363978656170859E-02_dp, 6.6985406661457603E+01_dp, -1.4642414370063676E+02_dp, &
         1.6193310131526860E+02_dp, -1.0257065719645773E+02_dp, 2.8871318786796056E+01_dp, &
         7.8634102783568860E+00_dp, -8.5200231797122488E+00_dp, 1.2264537347714767E-01_dp, &
         1.9987657976971298E+00_dp, 7.7158069079422978E-02_dp, -1.1542897495755511E+00_dp, &
         6.7737646619243352E-01_dp, -1.3657792494227855E-01_dp, 1.4740538370956807E+02_dp, &
         -3.3932989241907069E+02_dp, 4.0219547600147121E+02_dp, -2.8677094173034959E+02_dp, &
         1.1237202082027906E+02_dp, -4.8057836168082861E+00_dp, -1.6514621066683070E+01_dp, &
         3.5357507348208697E+00_dp, 3.3035381871180327E+00_dp, -7.4033248045744904E-01_dp, &
         -1.6598699253649027E+00_dp, 1.2042118842716378E+00_dp, -2.6920467176461421E-01_dp, &
         1.5261740947232876E+00_dp, -6.4895738266931657E-01_dp, -1.1564452965859112E-01_dp, &
         7.0980947338764838E-02_dp, 2.6676015641548347E-02_dp, -1.6136455032911386E-02_dp, &
         -6.7340073746126466E-03_dp, 6.0342465353809367E-03_dp, 2.8819686877824442E-03_dp, &
         -2.5718142851708921E-03_dp, -1.3273365637482889E-03_dp, 1.6243550200292600E-03_dp, &
         -4.2055786525236277E-04_dp, 1.7635041108560561E+00_dp, -9.5779430337441518E-01_dp, &
         -3.3226684026626346E-02_dp, 1.3208807728323771E-01_dp, 3.3225866313692422E-03_dp, &
         -3.4496297667900466E-02_dp, 1.4674666962247453E-03_dp, 1.2123713242266174E-02_dp, &
         8.7892456369925113E-04_dp, -7.8123662744860019E-03_dp, 3.3746734942086509E-03_dp, &
         2.1407084187177066E-04_dp, -3.1845714516548978E-04_dp, 2.2858230764507561E+00_dp, &
         -1.6828828513757415E+00_dp, 2.1342349744092762E-01_dp, 2.5321127952670880E-01_dp, &
         -7.0205604577718933E-02_dp, -6.7430673522987303E-02_dp, 2.7518708712689594E-02_dp, &
         2.1972840982545115E-02_dp, -6.8503456890888544E-03_dp, -1.7678791877657449E-02_dp, &
         1.5800632995275406E-02_dp, -4.9541923552455531E-03_dp, 4.5213141208361152E-04_dp, &
         3.1971135931052292E+00_dp, -3.0674576959479736E+00_dp, 8.2727565667300873E-01_dp, &
         4.1192795171230862E-01_dp, -2.5340716065788732E-01_dp, -9.7677767113723460E-02_dp, &
         9.0618959728272830E-02_dp, 2.7208794713592679E-02_dp, -2.7958096268066484E-02_dp, &
         -2.7052906430807815E-02_dp, 3.8668773332420213E-02_dp, -1.7422012026548538E-02_dp, &
         2.8945916612542781E-03_dp, 4.6799701618791243E+00_dp, -5.5646620604346912E+00_dp, &
         2.2397817355646192E+00_dp, 5.0208145461724241E-01_dp, -6.4842250541548618E-01_dp, &
         -6.7592147376930631E-02_dp, 2.1681380729701397E-01_dp, 4.0599420621051435E-03_dp, &
         -7.2943200049801968E-02_dp, -1.8572586454638652E-02_dp, 6.6945743501341068E-02_dp, &
         -3.8891162759268932E-02_dp, 7.9717226708572522E-03_dp, 7.0397079895684032E+00_dp, &
         -9.9842598075051558E+00_dp, 5.3386101825655814E+00_dp, 1.7446186675610861E-01_dp, &
         -1.3948157912004475E+00_dp, 1.7791991429819645E-01_dp, 4.2211621944934152E-01_dp, &
         -1.0340330103130466E-01_dp, -1.4705013349847670E-01_dp, 4.5565811980008253E-02_dp, &
         7.7147756940248571E-02_dp, -6.4077566137299757E-02_dp, 1.5603547582033700E-02_dp, &
         1.0786587572187873E+01_dp, -1.7760112751903939E+01_dp, 1.1918980063394129E+01_dp, &
         -1.5539236970758330E+00_dp, -2.5746531195043820E+00_dp, 9.9669167169678763E-01_dp, &
         6.4516326868847929E-01_dp, -4.0607915441255105E-01_dp, -2.2302549663448712E-01_dp, &
         2.2969575799370925E-01_dp, 1.1320877228687479E-02_dp, -7.2862905150103019E-02_dp, &
         2.3237485312351126E-02_dp, 1.6798695470837391E+01_dp, -3.1441464870566055E+01_dp, &
         2.5557113319742395E+01_dp, -7.2192306979436189E+00_dp, -3.7887425571978977E+00_dp, &
         3.0784789752716946E+00_dp, 5.3318993850287921E-01_dp, -1.0629617159049316E+00_dp, &
         -1.6729312653851172E-01_dp, 6.2482228533775386E-01_dp, -2.7075437119322843E-01_dp, &
         -7.6526272998893416E-04_dp, 1.9795696185616975E-02_dp, 2.6698448272325432E+01_dp, &
         -5.5678518659905293E+01_dp, 5.3281403655988036E+01_dp, -2.2925621416228282E+01_dp, &
         -2.7448097677013457E+00_dp, 7.2733797930171429E+00_dp, -1.0652244860941229E+00_dp, &
         -2.0439448593750167E+00_dp, 4.3742024469413332E-01_dp, 1.1815987016225704E+00_dp, &
         -1.0071997092922647E+00_dp, 3.2214298124885665E-01_dp, -3.3769710235736237E-02_dp, &
         4.3948016443764217E+01_dp, -9.9715998087285030E+01_dp, 1.0903180020753425E+02_dp, &
         -6.2349038976881253E+01_dp, 8.4191726553756379E+00_dp, 1.2761370267104230E+01_dp, &
         -6.4816307926839150E+00_dp, -2.1407881050035464E+00_dp, 2.2505811330603973E+00_dp, &
         9.4541418370340546E-01_dp, -1.8728571043550890E+00_dp, 9.3165721994730266E-01_dp, &
         -1.7212775126811192E-01_dp, 7.8045875406746134E+01_dp, -1.8661598260984715E+02_dp, &
         2.2414431510122250E+02_dp, -1.5518159917081857E+02_dp, 4.9918019842175177E+01_dp, &
         1.1453185884785420E+01_dp, -1.6531485621219627E+01_dp, 1.9110397390256724E+00_dp, &
         4.1466549895305009E+00_dp, -1.2773383144404289E+00_dp, -1.3711697775898219E+00_dp, &
         1.1359083217574477E+00_dp, -2.6834986232570840E-01_dp, 1.6737659404428402E+02_dp, &
         -4.0546870390462601E+02_dp, 5.0715726314925251E+02_dp, -3.8648362061791448E+02_dp, &
         1.6825446300611861E+02_dp, -1.5904401111568793E+01_dp, -2.4753480319349315E+01_dp, &
         9.7300048676002913E+00_dp, 3.4706756298406813E+00_dp, -2.7684975236319453E+00_dp, &
         -7.2826041594920199E-01_dp, 1.1315837672625995E+00_dp, -3.0636127893899234E-01_dp, &
         6.2801459623898256E+02_dp, -1.4934850582877016E+03_dp, 1.8458135305428518E+03_dp, &
         -1.4058647887554894E+03_dp, 6.3659031322995588E+02_dp, -1.0314180734592853E+02_dp, &
         -5.1253805381204558E+01_dp, 2.0084334666874064E+01_dp, 1.0632240773130309E+01_dp, &
         -4.7733655815247635E+00_dp, -4.8730760134796665E+00_dp, 4.2825744412767870E+00_dp, &
         -1.0293980933009121E+00_dp, 2.4470051605057291E-01_dp, -6.2837867140790746E-02_dp, &
         -2.3825186082600196E-02_dp, 7.5784912887244772E-03_dp, 7.5082468515253169E-03_dp, &
         -2.6455605788751222E-03_dp, -3.3362505117537893E-03_dp, 7.8091376511942739E-04_dp, &
         3.0325200376611385E-03_dp, -2.3347738852380210E-03_dp, 3.9936843186023246E-05_dp, &
         5.8244826441623864E-04_dp, -1.9032202744327542E-04_dp, 8.0022162325527124E-01_dp, &
         -2.4209391447252487E-01_dp, -7.6405614359062499E-02_dp, 3.8989771396722744E-02_dp, &
         2.5419789016015160E-02_dp, -1.5256813711493509E-02_dp, -1.1473513632992018E-02_dp, &
         5.2085413881328025E-03_dp, 1.2850293721253818E-02_dp, -1.3920130297941820E-02_dp, &
         4.3891393295565915E-03_dp, 5.8004137460729440E-04_dp, -4.7036883093655793E-04_dp, &
         1.5788599826464651E+00_dp, -6.1762979410631058E-01_dp, -1.3565786277770472E-01_dp, &
         1.3250818605318840E-01_dp, 4.7423141009731530E-02_dp, -5.6828580488514448E-02_dp, &
         -2.1224822901882736E-02_dp, 2.1560191062029284E-02_dp, 3.4712659716572912E-02_dp, &
         -5.3570121520492134E-02_dp, 2.8798870932988822E-02_dp, -6.1083685002502114E-03_dp, &
         1.1251907951229265E-04_dp, 2.8208126201441122E+00_dp, -1.4584430358075993E+00_dp, &
         -1.6668814068976517E-01_dp, 3.7939412477145662E-01_dp, 5.2727438203323901E-02_dp, &
         -1.7212640736974685E-01_dp, -1.9204293272473808E-02_dp, 7.0402478790583201E-02_dp, &
         7.4962963356748721E-02_dp, -1.6470750799943457E-01_dp, 1.1464911243407885E-01_dp, &
         -3.7209893008122726E-02_dp, 4.6211560443511707E-03_dp, 4.9425542029764467E+00_dp, &
         -3.3299793049425701E+00_dp, -1.5240471603939173E-02_dp, 9.6617116551502813E-01_dp, &
         -5.4487412666999822E-02_dp, -4.4981543663568441E-01_dp, 5.4235895382646740E-02_dp, &
         1.9616328040136954E-01_dp, 1.1396958212193117E-01_dp, -4.2101103801084033E-01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/3.5632823078643500E-01_dp, &
         -1.4001776162509430E-01_dp, 2.2286666712875530E-02_dp, 8.7125285885676789E+00_dp, &
         -7.4679426200738366E+00_dp, 8.3805209926335500E-01_dp, 2.2477163058559704E+00_dp, &
         -6.0060735276290700E-01_dp, -1.0331040612400506E+00_dp, 3.9698448728282515E-01_dp, &
         4.7797940178876780E-01_dp, 1.1436011046665012E-02_dp, -8.7501751935782923E-01_dp, &
         9.1713404922379360E-01_dp, -4.1604826319244603E-01_dp, 7.5505830008108218E-02_dp, &
         1.5605853014796613E+01_dp, -1.6640738573993023E+01_dp, 3.9973532644106071E+00_dp, &
         4.7869251169441966E+00_dp, -2.5684017026470856E+00_dp, -2.0236077415630396E+00_dp, &
         1.5736463565912060E+00_dp, 9.8572630187217014E-01_dp, -7.7868319784195095E-01_dp, &
         -1.2477373616278120E+00_dp, 1.9014513769547758E+00_dp, -1.0102166550823963E+00_dp, &
         2.0453369882137556E-01_dp, 2.8549616253145054E+01_dp, -3.7205075770883383E+01_dp, &
         1.4271765352571258E+01_dp, 8.9932074739321770E+00_dp, -8.6798219050700265E+00_dp, &
         -2.8631184094001214E+00_dp, 4.9827139884730887E+00_dp, 1.4257317765234938E+00_dp, &
         -3.8503419132165200E+00_dp, 1.8060419072857684E-01_dp, 2.6148900288649957E+00_dp, &
         -1.8695443351115448E+00_dp, 4.3631072153907585E-01_dp, 5.3552984772777414E+01_dp, &
         -8.4175532531260529E+01_dp, 4.6043669292644864E+01_dp, 1.2581369109054569E+01_dp, &
         -2.5785157300270445E+01_dp, 3.6426327688394516E-01_dp, 1.3334789657578380E+01_dp, &
         -4.7226254964308417E-01_dp, -1.2526270169907576E+01_dp, 9.4142206495359755E+00_dp, &
         -9.7670970926161171E-01_dp, -1.6726174022636047E+00_dp, 5.9874706221520879E-01_dp, &
         1.0346475698874463E+02_dp, -1.9421400275760911E+02_dp, 1.4292733600886447E+02_dp, &
         -2.1682521535311881E+00_dp, -6.7899121747504481E+01_dp, 2.4906170745096762E+01_dp, &
         2.8087316716428784E+01_dp, -1.4449093427136949E+01_dp, -2.8736513116670000E+01_dp, &
         4.0767438543005156E+01_dp, -2.2398062557021266E+01_dp, 5.4405001864859486E+00_dp, &
         -3.4631065659938148E-01_dp, 2.0750616910138754E+02_dp, -4.6065142702781014E+02_dp, &
         4.4009500583974625E+02_dp, -1.2652217098767277E+02_dp, -1.4255338662322163E+02_dp, &
         1.3009558694582910E+02_dp, 2.4337692154912165E+01_dp, -6.8052930184984419E+01_dp, &
         -2.3709608232610183E+01_dp, 9.9442268821700665E+01_dp, -8.4274944171226693E+01_dp, &
         3.3636040311091861E+01_dp, -5.5275330736437418E+00_dp, 4.4235457824221572E+02_dp, &
         -1.1417219968036247E+03_dp, 1.3690824125389506E+03_dp, -7.7065096290590293E+02_dp, &
         -8.8442075039759473E+01_dp, 4.2234029954766072E+02_dp, -1.6089957770769391E+02_dp, &
         -1.4258882763239106E+02_dp, 1.2603829491926012E+02_dp, 5.0804366296990459E+01_dp, &
         -1.1980481260857118E+02_dp, 6.7459853564141966E+01_dp, -1.3960462430030432E+01_dp, &
         1.1324481119281395E+03_dp, -3.2684217630798435E+03_dp, 4.6328527365543323E+03_dp, &
         -3.7266964190295712E+03_dp, 1.2856665138480816E+03_dp, 5.6297970920071646E+02_dp, &
         -7.5128039018902086E+02_dp, 7.0908728648373369E+01_dp, 2.9917130128753172E+02_dp, &
         -1.5232184462135913E+02_dp, -4.8594625636135675E+01_dp, 6.7185755709058370E+01_dp, &
         -1.8010386679280117E+01_dp, 2.0001036223672011E+00_dp, -5.3620407054782981E-01_dp, &
         -1.9434054411717244E-01_dp, 7.0618726869747614E-02_dp, 6.2242609610535821E-02_dp, &
         -2.5597292416828135E-02_dp, -2.7813756603712155E-02_dp, 7.9491914776396028E-03_dp, &
         2.6680695709919679E-02_dp, -2.2810343453090697E-02_dp, 2.7161809486165871E-03_dp, &
         4.0718478547287487E-03_dp, -1.4880205641258518E-03_dp, 2.5450615398879743E+00_dp, &
         -9.8744397592581534E-01_dp, -2.2902266959137468E-01_dp, 2.0922986206151875E-01_dp, &
         8.2779560412023001E-02_dp, -8.8686545550320794E-02_dp, -3.7958590200944345E-02_dp, &
         3.2998767331186209E-02_dp, 5.7441855270214746E-02_dp, -8.2952472588137552E-02_dp, &
         4.1461489832071360E-02_dp, -7.2309981798897591E-03_dp, -3.4486800387515759E-04_dp, &
         3.8383945496778296E+00_dp, -2.1858886322144140E+00_dp, -2.2638494850866908E-01_dp, &
         5.9654548742227309E-01_dp, 8.2375367991094961E-02_dp, -2.7228188920384877E-01_dp, &
         -3.2991216309614123E-02_dp, 1.1069144407749401E-01_dp, 1.2288856763142823E-01_dp, &
         -2.6231090706733151E-01_dp, 1.7948962970980861E-01_dp, -5.6968770565566255E-02_dp, &
         6.8079040584616184E-03_dp, 6.3639884662472506E+00_dp, -4.8950253392955796E+00_dp, &
         6.7035562751566802E-02_dp, 1.4977453112460624E+00_dp, -9.4857290242106540E-02_dp, &
         -7.0786182174407719E-01_dp, 8.3499909648241832E-02_dp, 3.0861847626582534E-01_dp, &
         1.9124789670380135E-01_dp, -6.7776371800663660E-01_dp, 5.6824943668256445E-01_dp, &
         -2.2147300860757518E-01_dp, 3.4925516438103342E-02_dp, 1.1100384232886654E+01_dp, &
         -1.0849365072746094E+01_dp, 1.4500169429794771E+00_dp, 3.4281034563810699E+00_dp, &
         -9.5548658310972046E-01_dp, -1.6125902800476346E+00_dp, 6.2195077294496282E-01_dp, &
         7.5138806761421051E-01_dp, 4.4232035492933294E-02_dp, -1.4236095789326435E+00_dp, &
         1.4787960249515941E+00_dp, -6.6792149947902735E-01_dp, 1.2078043058881406E-01_dp, &
         1.9997254830061060E+01_dp, -2.3977706500886121E+01_dp, 6.3262970427801459E+00_dp, &
         7.2118993888159375E+00_dp, -4.0034174842598853E+00_dp, -3.1512209966526510E+00_dp, &
         2.4653796429483354E+00_dp, 1.5660770527154297E+00_dp, -1.1858015495148018E+00_dp, &
         -2.0863186414852812E+00_dp, 3.1158573924681550E+00_dp, -1.6479978556059860E+00_dp, &
         3.3293613764598917E-01_dp, 3.7002091361283142E+01_dp, -5.3343789674459856E+01_dp, &
         2.1798301205375779E+01_dp, 1.3490742966183797E+01_dp, -1.3399645593923985E+01_dp, &
         -4.5399329672371129E+00_dp, 7.8331708134372793E+00_dp, 2.3842813657243171E+00_dp, &
         -6.1312609096237622E+00_dp, 6.9373920027707620E-02_dp, 4.4439580524589513E+00_dp, &
         -3.1287274765904054E+00_dp, 7.2687886436644078E-01_dp, 7.0330646344558374E+01_dp, &
         -1.2042385478592142E+02_dp, 6.8896945123466992E+01_dp, 1.9248160282366094E+01_dp, &
         -3.9785945322970619E+01_dp, -1.0757313037469426E-01_dp, 2.1300262950471573E+01_dp, &
         -5.9398954626295727E-02_dp, -2.0726405213802341E+01_dp, 1.5013319224419423E+01_dp, &
         -1.0206915169707462E+00_dp, -3.0611082438918387E+00_dp, 1.0504012964699687E+00_dp, &
         1.3771889621960588E+02_dp, -2.7778114229196626E+02_dp, 2.1092710837289934E+02_dp, &
         4.9544962436490181E-01_dp, -1.0656701841455923E+02_dp, 3.5610982335235697E+01_dp, &
         4.7522077277759607E+01_dp, -2.0742921049299113E+01_dp, -5.1214696900496918E+01_dp, &
         6.8525574756543946E+01_dp, -3.6081947353505832E+01_dp, 8.0689565180881910E+00_dp, &
         -3.3115643672049860E-01_dp, 2.7951222706689759E+02_dp, -6.5906712429888637E+02_dp, &
         6.4323098194907232E+02_dp, -1.6677447485585066E+02_dp, -2.4162966307021762E+02_dp, &
         1.9677031498186125E+02_dp, 6.0574620887850372E+01_dp, -1.0971429429810229E+02_dp, &
         -6.6686880639177701E+01_dp, 1.9421790532342177E+02_dp, -1.5486964911422615E+02_dp, &
         5.9425108459636675E+01_dp, -9.4080145408358931E+00_dp, 5.9731652416151724E+02_dp, &
         -1.6234750522173472E+03_dp, 1.9836471844267119E+03_dp, -1.0519495956941482E+03_dp, &
         -2.9350943708598510E+02_dp, 7.3559653074966434E+02_dp, -1.8444178048301910E+02_dp, &
         -3.3747019072196520E+02_dp, 2.0501261919104977E+02_dp, 1.9472432317933209E+02_dp, &
         -3.1351374896534389E+02_dp, 1.6602020342171713E+02_dp, -3.3574624921311873E+01_dp, &
         1.4305952826928835E+03_dp, -4.3148429996300356E+03_dp, 6.2934312136573408E+03_dp, &
         -5.0322189338736098E+03_dp, 1.3515555677503437E+03_dp, 1.4914175783605847E+03_dp, &
         -1.5590391042662025E+03_dp, 2.5540483265629923E+01_dp, 9.2917473249960813E+02_dp, &
         -7.1183738950361192E+02_dp, 1.7736950337469372E+02_dp, 3.3996199940030365E+01_dp, &
         -2.0265055920456604E+01_dp, 5.3260400513964350E+03_dp, -1.6436105385546998E+04_dp, &
         2.5548454078470353E+04_dp, -2.4072536527098953E+04_dp, 1.2929492785729244E+04_dp, &
         -1.5539864332605835E+03_dp, -2.8010923463387767E+03_dp, 1.3421428470689034E+03_dp, &
         6.7886952374328268E+02_dp, -8.2031294944055298E+02_dp, 1.5680250796613109E+02_dp, &
         1.0539258593554351E+02_dp, -4.3152598585647880E+01_dp, 2.7522136583648199E-01_dp, &
         -3.2217110577238542E-02_dp, -1.7033147481772627E-02_dp, 2.0210122815219334E-03_dp, &
         3.9289578029764833E-03_dp, -3.9930449290441994E-05_dp, -1.9887361467012953E-03_dp, &
         3.6898435626645432E-04_dp, 7.3280730329796837E-04_dp, -1.1210128908484710E-04_dp, &
         -5.1629849116574338E-04_dp, 3.8585425500095789E-04_dp, -9.0580414399608951E-05_dp, &
         9.2165986573140568E-01_dp, -1.3249464016020585E-01_dp, -6.5239324099907584E-02_dp, &
         1.3110109010924940E-02_dp, 1.6453340473596419E-02_dp, -1.8265742727726945E-03_dp, &
         -8.6103449023340675E-03_dp, 2.2159504354805157E-03_dp, 3.4794700520068469E-03_dp, &
         -1.2533943941271351E-03_dp, -1.8168584053897300E-03_dp, 1.5997969588019021E-03_dp, &
         -3.9784620927895906E-04_dp, 1.9041320449174821E+00_dp, -3.7234333772363021E-01_dp, &
         -1.6365326189221707E-01_dp, 5.5534422369305905E-02_dp, 4.6089398503204683E-02_dp, &
         -1.2265328668024897E-02_dp, -2.5382790394936194E-02_dp, 9.3490699632645686E-03_dp, &
         1.1622195051795290E-02_dp, -7.3570743970728667E-03_dp, -3.2261443745580896E-03_dp, &
         4.2824229569487044E-03_dp, -1.1778304868165397E-03_dp, 3.6316569669487175E+00_dp, &
         -9.7730941204074340E-01_dp, -3.7309973697493126E-01_dp, 1.9511449684300722E-01_dp, &
         1.1472304455885286E-01_dp, -5.3617761770916240E-02_dp, -6.7215522012602877E-02_dp, &
         3.4468238642844179E-02_dp, 3.5251414090503870E-02_dp, -3.2843194103284207E-02_dp, &
         -6.1769442529466899E-05_dp, 9.2371576567999401E-03_dp, -3.0152739805040869E-03_dp, &
         6.9044315846491511E+00_dp, -2.4954078084649751E+00_dp, -7.9752233983078080E-01_dp, &
         6.1910169828591066E-01_dp, 2.5700006957784816E-01_dp, -1.9664588521910206E-01_dp, &
         -1.6279316206764638E-01_dp, 1.1756639080199344E-01_dp, 9.9690988498802063E-02_dp, &
         -1.2855067522072000E-01_dp, 3.3851769665174079E-02_dp, 1.2487678317695030E-02_dp, &
         -6.4025099233952784E-03_dp, 1.3402802913338570E+01_dp, -6.3224482797045658E+00_dp, &
         -1.5803882488352667E+00_dp, 1.8639565830791769E+00_dp, 4.9440377028852905E-01_dp, &
         -6.6058719060697912E-01_dp, -3.5203000086883229E-01_dp, 3.8416444626006885E-01_dp, &
         2.6219933074993468E-01_dp, -4.6750014207051527E-01_dp, 2.1565685107100283E-01_dp, &
         -1.9590633219356376E-02_dp, -8.3188142246124162E-03_dp, 2.6847189168859529E+01_dp, &
         -1.6164600954038654E+01_dp, -2.7348299274314170E+00_dp, 5.4804038804993898E+00_dp, &
         6.3908178640514468E-01_dp, -2.1090707289755835E+00_dp, -6.0397632454309369E-01_dp, &
         1.2258676026166442E+00_dp, 6.0900085656743030E-01_dp, -1.6095450133803784E+00_dp, &
         1.0185781424420182E+00_dp, -2.6410536241534338E-01_dp, 1.7190485044331690E-02_dp, &
         5.5920030994857392E+01_dp, -4.2359987862833194E+01_dp, -3.0851699293052208E+00_dp, &
         1.6003551733561032E+01_dp, -7.5845896807267987E-01_dp, -6.4758768912300226E+00_dp, &
         -2.7997762367110551E-01_dp, 3.8438366690735388E+00_dp, 9.6178031191591695E-01_dp, &
         -5.1789423674060124E+00_dp, 4.1897184836450609E+00_dp, -1.5066589469240397E+00_dp, &
         2.1023744160454813E-01_dp, 1.2214165365194877E+02_dp, -1.1560316535122182E+02_dp, &
         5.2431654538003061E+00_dp, 4.6756646171042270E+01_dp, -1.1638136061220804E+01_dp, &
         -1.8825294369662465E+01_dp, 4.6901099778948048E+00_dp, 1.1684043831662752E+01_dp, &
         -1.5690448976060742E+00_dp, -1.4469520844727743E+01_dp, 1.5229016181347497E+01_dp, &
         -6.7413191274813640E+00_dp, 1.1871712862612187E+00_dp, 2.8292219949744720E+02_dp, &
         -3.3466585644002333E+02_dp, 6.5399374381441106E+01_dp, 1.3547954808764075E+02_dp, &
         -7.2400177578816127E+01_dp, -4.7247337505445195E+01_dp, 3.5254490079699167E+01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/3.1951406154569082E+01_dp, &
         -2.7863833003973856E+01_dp, -2.4615790099798474E+01_dp, 4.4154903132889864E+01_dp, &
         -2.4321043128719936E+01_dp, 5.0336537486842516E+00_dp, 7.0657858026784504E+02_dp, &
         -1.0528573743078830E+03_dp, 4.1625512279061735E+02_dp, 3.6427444425345516E+02_dp, &
         -3.8770748215573474E+02_dp, -4.8519737539136528E+01_dp, 1.8255864495792844E+02_dp, &
         4.9003198502854644E+01_dp, -1.8896312119490256E+02_dp, 8.1337993659320531E+01_dp, &
         4.5525635099598908E+01_dp, -5.2286864309302139E+01_dp, 1.4107728865636403E+01_dp, &
         1.9572003502958112E+03_dp, -3.7275124443070590E+03_dp, 2.5346776221187474E+03_dp, &
         5.1376728855467570E+02_dp, -1.9308763300790281E+03_dp, 7.9050799076938517E+02_dp, &
         5.7143225950275246E+02_dp, -3.4470294413325115E+02_dp, -6.8440314851472851E+02_dp, &
         1.0471215545682312E+03_dp, -6.3503878270573148E+02_dp, 1.8429203829230039E+02_dp, &
         -1.9468221079821902E+01_dp, 6.5615697945100492E+03_dp, -1.5941114910537892E+04_dp, &
         1.7283771807502479E+04_dp, -7.2779267165695510E+03_dp, -4.8084545703952344E+03_dp, &
         8.8537836032320683E+03_dp, -5.0434849558105634E+03_dp, 5.9362692294717317E+02_dp, &
         2.7660294111299510E+02_dp, 8.0161530341845844E+02_dp, -1.1650257023506490E+03_dp, &
         6.0897733306934356E+02_dp, -1.2237560126278971E+02_dp, 2.2627540419257364E+00_dp, &
         -2.7968575736450180E-01_dp, -1.4518668455604195E-01_dp, 2.0265924722043079E-02_dp, &
         3.4338618611475344E-02_dp, -1.2556065673962331E-03_dp, -1.7531784982831655E-02_dp, &
         3.5793660343997852E-03_dp, 6.6231604787169770E-03_dp, -1.3867866855376788E-03_dp, &
         -4.3395264768239942E-03_dp, 3.3684735669525651E-03_dp, -8.0229330712122920E-04_dp, &
         3.0613725230519471E+00_dp, -5.8640457056345652E-01_dp, -2.6339818110169105E-01_dp, &
         8.2855001233820474E-02_dp, 7.3722316481618991E-02_dp, -1.7015690354151666E-02_dp, &
         -4.0113596090714465E-02_dp, 1.3619066177511220E-02_dp, 1.7832370619744201E-02_dp, &
         -1.0005418487325966E-02_dp, -6.1379340236641386E-03_dp, 7.0291927722965558E-03_dp, &
         -1.8751685593083232E-03_dp, 5.0576570915087578E+00_dp, -1.4710755111222258E+00_dp, &
         -5.6312085551876734E-01_dp, 2.9587574582447002E-01_dp, 1.7758263331967478E-01_dp, &
         -7.9480406519325253E-02_dp, -1.0353547727129607E-01_dp, 5.0871576222639547E-02_dp, &
         5.3464895969474464E-02_dp, -4.7025455703949920E-02_dp, -2.8649889086355728E-03_dp, &
         1.5163927370369992E-02_dp, -4.7605351840397164E-03_dp, 9.2666854098509734E+00_dp, &
         -3.7037123145769315E+00_dp, -1.1810845674556205E+00_dp, 9.3266985234712430E-01_dp, &
         3.9541518329595232E-01_dp, -2.9288519060055263E-01_dp, -2.4999841635163031E-01_dp, &
         1.7350630646750761E-01_dp, 1.5126192191822729E-01_dp, -1.8602668410133402E-01_dp, &
         4.2058952933339604E-02_dp, 2.3004266980576980E-02_dp, -1.0410473830431016E-02_dp, &
         1.7967075304979097E+01_dp, -9.2886003844612315E+00_dp, -2.3166503256513042E+00_dp, &
         2.7793145895593954E+00_dp, 7.6579508186925827E-01_dp, -9.8065539968708293E-01_dp, &
         -5.4434577208491419E-01_dp, 5.6487894573947284E-01_dp, 4.0128002438923693E-01_dp, &
         -6.7985365844418810E-01_dp, 2.9335374977600032E-01_dp, -1.3760417124864824E-02_dp, &
         -1.5579665225528172E-02_dp, 3.6330014596264910E+01_dp, -2.3558077039701011E+01_dp, &
         -4.0032080208899892E+00_dp, 8.1023328627947695E+00_dp, 1.0572474369212368E+00_dp, &
         -3.1246715021136469E+00_dp, -9.7288399455208796E-01_dp, 1.7980101554684516E+00_dp, &
         9.6548627911437968E-01_dp, -2.3618248937200388E+00_dp, 1.4256316289637909E+00_dp, &
         -3.3674236693570941E-01_dp, 1.3035310486665157E-02_dp, 7.6564297545359281E+01_dp, &
         -6.1327459250609188E+01_dp, -4.7038085252667656E+00_dp, 2.3522134846498993E+01_dp, &
         -6.2074232510114502E-01_dp, -9.6274919508419270E+00_dp, -7.4907562113744564E-01_dp, &
         5.6473524210207273E+00_dp, 1.7877391201538855E+00_dp, -7.7760616092732224E+00_dp, &
         6.0028051651270777E+00_dp, -2.0515988031519519E+00_dp, 2.6498768518488430E-01_dp, &
         1.6908758380027180E+02_dp, -1.6620256620995710E+02_dp, 5.8556451666173768E+00_dp, &
         6.8594902354899986E+01_dp, -1.4646639889433427E+01_dp, -2.8503986567123022E+01_dp, &
         5.4011736632665412E+00_dp, 1.7407266610755649E+01_dp, -2.1695266703893210E-01_dp, &
         -2.3183117951883965E+01_dp, 2.2747354795410811E+01_dp, -9.6248972377227879E+00_dp, &
         1.6232770520249808E+00_dp, 3.9496620712272983E+02_dp, -4.7630587177234480E+02_dp, &
         8.2634984633242880E+01_dp, 2.0043881475177201E+02_dp, -9.3256133931078750E+01_dp, &
         -7.6938636554802741E+01_dp, 4.5382945051577323E+01_dp, 5.0541291906442254E+01_dp, &
         -3.0180631932035514E+01_dp, -5.1804380534411720E+01_dp, 7.3703321624210730E+01_dp, &
         -3.7832717690973290E+01_dp, 7.4886753735631828E+00_dp, 9.8936735122169955E+02_dp, &
         -1.4723135861653050E+03_dp, 5.1661296919801975E+02_dp, 5.6775300121587384E+02_dp, &
         -4.9823328967242446E+02_dp, -1.4291981765414513E+02_dp, 2.5077300933486598E+02_dp, &
         1.1280997635389373E+02_dp, -2.4043309466984022E+02_dp, 1.8533960613943030E+01_dp, &
         1.5008777209583707E+02_dp, -1.0988215815683799E+02_dp, 2.6126498062203950E+01_dp, &
         2.7151255050497193E+03_dp, -5.0457127958801984E+03_dp, 2.9824323650975698E+03_dp, &
         1.2838985854569405E+03_dp, -2.5571835909206930E+03_dp, 4.2544336126025911E+02_dp, &
         1.0994628841103713E+03_dp, -1.2735938497924820E+02_dp, -1.2861687784670435E+03_dp, &
         1.2636200369734484E+03_dp, -4.3735779232907760E+02_dp, -1.3792064567093755E+01_dp, &
         3.3103909845024816E+01_dp, 8.5120394945424250E+03_dp, -1.9972380653537413E+04_dp, &
         1.8712883353138033E+04_dp, -2.4296838299905257E+03_dp, -1.1695432627920640E+04_dp, &
         9.8982067857411967E+03_dp, 6.8744963878254964E+02_dp, -4.5712455095728219E+03_dp, &
         -1.0807863326674956E+03_dp, 6.7875580160934505E+03_dp, -6.3715584765173253E+03_dp, &
         2.7904520792497096E+03_dp, -5.0512451195331687E+02_dp, 3.7015110168653046E+04_dp, &
         -1.0359356753096696E+05_dp, 1.3988074174530094E+05_dp, -1.0150724524384555E+05_dp, &
         1.4669819151632983E+04_dp, 5.4938546232760098E+04_dp, -7.1495818225217707E+04_dp, &
         5.0437331500296590E+04_dp, -2.4139642994171907E+04_dp, 8.5816358667112163E+03_dp, &
         -2.4305935666411028E+03_dp, 5.1492651788062108E+02_dp, -5.4029802229116505E+01_dp, &
         2.8766437095860359E-01_dp, -1.2841648915210033E-02_dp, -1.1749544511224883E-02_dp, &
         3.9124751913915002E-03_dp, -4.1340943883577163E-03_dp, 6.7857802334998418E-03_dp, &
         -7.5294031286692407E-03_dp, 7.2175306126584524E-03_dp, -6.7009368542467326E-03_dp, &
         5.3904146360225216E-03_dp, -3.2096830445151438E-03_dp, 1.2124425357336903E-03_dp, &
         -2.1637406162705776E-04_dp, 9.7356679780566036E-01_dp, -5.4345431749096215E-02_dp, &
         -4.8904874253147546E-02_dp, 1.8036518570911850E-02_dp, -1.7361622406643535E-02_dp, &
         2.8759362863345999E-02_dp, -3.2180437961153392E-02_dp, 3.0604917054349139E-02_dp, &
         -2.8207332208422350E-02_dp, 2.2678944358017036E-02_dp, -1.3518527121475700E-02_dp, &
         5.1039031798528591E-03_dp, -9.0864035037277143E-04_dp, 2.0530666914935956E+00_dp, &
         -1.5918196985192454E-01_dp, -1.3979713465522753E-01_dp, 5.8990016258980232E-02_dp, &
         -5.0479493514827861E-02_dp, 8.4318744789823125E-02_dp, -9.5580479758503561E-02_dp, &
         8.9926670184525562E-02_dp, -8.1951748270375741E-02_dp, 6.5826636399679089E-02_dp, &
         -3.9316553314932813E-02_dp, 1.4841315675822706E-02_dp, -2.6340794368276848E-03_dp, &
         4.0316235573187873E+00_dp, -4.3717905236020804E-01_dp, -3.7355720461015557E-01_dp, &
         1.8022229937573989E-01_dp, -1.3839328046856253E-01_dp, 2.3139075155391939E-01_dp, &
         -2.6640028065715393E-01_dp, 2.4795661325252652E-01_dp, -2.2289816786634609E-01_dp, &
         1.7876458006240628E-01_dp, -1.0708594501709379E-01_dp, 4.0458337362132495E-02_dp, &
         -7.1627969268156346E-03_dp, 7.9515127037254940E+00_dp, -1.1725795416101994E+00_dp, &
         -9.7139505078684429E-01_dp, 5.3572393763784321E-01_dp, -3.7368202650800936E-01_dp, &
         6.1869498945049961E-01_dp, -7.2552440352875003E-01_dp, 6.6842200013424780E-01_dp, &
         -5.9088129861833538E-01_dp, 4.7264433898376779E-01_dp, -2.8431635449514581E-01_dp, &
         1.0769203775409864E-01_dp, -1.9041687452794855E-02_dp, 1.6132294510459047E+01_dp, &
         -3.1414300269804429E+00_dp, -2.5084546797336125E+00_dp, 1.5919506211215777E+00_dp, &
         -1.0199649239248365E+00_dp, 1.6474808867343209E+00_dp, -1.9752177536142645E+00_dp, &
         1.8027216083947462E+00_dp, -1.5596608718908076E+00_dp, 1.2416080767881754E+00_dp, &
         -7.5126157559420226E-01_dp, 2.8603278415470540E-01_dp, -5.0617454119410023E-02_dp, &
         3.4064569212641622E+01_dp, -8.5769738548301273E+00_dp, -6.5423787884993834E+00_dp, &
         4.8406599765190608E+00_dp, -2.8877282478001232E+00_dp, 4.4534784008756505E+00_dp, &
         -5.4834935225951478E+00_dp, 4.9661825471664365E+00_dp, -4.1752937122300180E+00_dp, &
         3.2934090517963024E+00_dp, -2.0084125823659398E+00_dp, 7.7180910515991674E-01_dp, &
         -1.3721280155604926E-01_dp, 7.5626882675942724E+01_dp, -2.4341047240189681E+01_dp, &
         -1.7493213476710846E+01_dp, 1.5398174306816493E+01_dp, -8.7275747587617456E+00_dp, &
         1.2451306283479628E+01_dp, -1.5814730127661882E+01_dp, 1.4263812050450625E+01_dp, &
         -1.1531252583607163E+01_dp, 8.9274705109811467E+00_dp, -5.4952489571487435E+00_dp, &
         2.1450145097837918E+00_dp, -3.8566671695744043E-01_dp, 1.7878769023909365E+02_dp, &
         -7.3409482660028232E+01_dp, -4.8603840481754148E+01_dp, 5.2510737919075773E+01_dp, &
         -2.9190434914792892E+01_dp, 3.6789271998095046E+01_dp, -4.8290125513308809E+01_dp, &
         4.3685642487586676E+01_dp, -3.3425817156494972E+01_dp, 2.4846648059952049E+01_dp, &
         -1.5396980073755357E+01_dp, 6.1596692561473594E+00_dp, -1.1327889201084458E+00_dp, &
         4.5853266997398288E+02_dp, -2.4219834828292343E+02_dp, -1.4185912981973058E+02_dp, &
         1.9818507281291306E+02_dp, -1.1363958650629472E+02_dp, 1.1864500714177012E+02_dp, &
         -1.5936064769189525E+02_dp, 1.4653462807503183E+02_dp, -1.0363950474324729E+02_dp, &
         6.9803513230716248E+01_dp, -4.2493410132349823E+01_dp, 1.7571722994223919E+01_dp, &
         -3.3652356136199639E+00_dp, 1.3149209017368075E+03_dp, -9.1410171353466365E+02_dp, &
         -4.3355981239428490E+02_dp, 8.6854889246174923E+02_dp, -5.5717720241772929E+02_dp, &
         4.4729901591974419E+02_dp, -5.8023403991716850E+02_dp, 5.5363750980013492E+02_dp, &
         -3.5036288911768412E+02_dp, 1.7673630965596453E+02_dp, -8.7764967631492297E+01_dp, &
         3.6165599936797761E+01_dp, -7.4275840619886511E+00_dp, 4.4599678882045982E+03_dp, &
         -4.2852510926403547E+03_dp, -1.2345242410977692E+03_dp, 4.8020310925072372E+03_dp, &
         -3.9712192527651519E+03_dp, 2.4421962410592055E+03_dp, -2.3444652516106917E+03_dp, &
         2.2748220642734714E+03_dp, -1.2153075874252681E+03_dp, 4.2618549893802005E+01_dp, &
         3.6957056541266746E+02_dp, -2.2467629199390655E+02_dp, 4.8301366431393674E+01_dp, &
         2.0515552105599014E+04_dp, -3.0251434673870317E+04_dp, 3.8760590293516493E+03_dp, &
         3.7614063826257821E+04_dp, -5.7087356760206945E+04_dp, 4.4626244651310022E+04_dp, &
         -1.6930953342940349E+04_dp, -8.6342938793903559E+03_dp, 2.5039453054521509E+04_dp, &
         -2.9727441111072243E+04_dp, 2.2663651551108436E+04_dp, -1.0448354492148857E+04_dp, &
         2.2049629835911287E+03_dp, 2.3711821326754712E+00_dp, -1.1232868750598661E-01_dp, &
         -1.0232817697756741E-01_dp, 3.5035029792710649E-02_dp, -3.6080149837860856E-02_dp, &
         5.9384272779209310E-02_dp, -6.6034549944533560E-02_dp, 6.3165197301667780E-02_dp, &
         -5.8533667022985264E-02_dp, 4.7080124186021241E-02_dp, -2.8040678094242519E-02_dp, &
         1.0590404330460124E-02_dp, -1.8887168866975497E-03_dp, 3.2950022547352660E+00_dp, &
         -2.4869033985187267E-01_dp, -2.1951170275277102E-01_dp, 9.0239524034429530E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c04 = (/-7.8836509708512373E-02_dp, &
         1.3178061660818258E-01_dp, -1.4891101086305505E-01_dp, 1.4038052557472530E-01_dp, &
         -1.2827505792189706E-01_dp, 1.0306674552536969E-01_dp, -6.1520533848826123E-02_dp, &
         2.3216931723502306E-02_dp, -4.1222161991565217E-03_dp, 5.6592328962619902E+00_dp, &
         -6.5682359828028736E-01_dp, -5.6224091737511694E-01_dp, 2.6926868745701449E-01_dp, &
         -2.0725768669908209E-01_dp, 3.4798639207117904E-01_dp, -3.9993937306967442E-01_dp, &
         3.7233564292827848E-01_dp, -3.3521912095457151E-01_dp, 2.6894454176131877E-01_dp, &
         -1.6101715682600454E-01_dp, 6.0796203847490786E-02_dp, -1.0759696759551582E-02_dp, &
         1.0820432159449139E+01_dp, -1.7387845708330643E+00_dp, -1.4425878741646769E+00_dp, &
         7.9167895347904427E-01_dp, -5.5153935438914514E-01_dp, 9.1858113260673568E-01_dp, &
         -1.0754845180690036E+00_dp, 9.9058914214781446E-01_dp, -8.7711044783346115E-01_dp, &
         7.0204070734569368E-01_dp, -4.2204323840163149E-01_dp, 1.5970999475494435E-01_dp, &
         -2.8218316930005444E-02_dp, 2.1975384938130706E+01_dp, -4.6086047370349190E+00_dp, &
         -3.6880188824890627E+00_dp, 2.3253201370596082E+00_dp, -1.4865249175920201E+00_dp, &
         2.4206052762959338E+00_dp, -2.8969618325004678E+00_dp, 2.6424885821728354E+00_dp, &
         -2.2914719419367309E+00_dp, 1.8263977484192504E+00_dp, -1.1042793052824569E+00_dp, &
         4.1984999040571597E-01_dp, -7.4204462154064413E-02_dp, 4.6836016437591510E+01_dp, &
         -1.2461263924295558E+01_dp, -9.5424753844301833E+00_dp, 6.9849024224173073E+00_dp, &
         -4.1532087460680200E+00_dp, 6.4834239581427680E+00_dp, -7.9652050621281747E+00_dp, &
         7.2063417067380673E+00_dp, -6.0811310415530810E+00_dp, 4.8089182798816328E+00_dp, &
         -2.9302138012528847E+00_dp, 1.1235538973792896E+00_dp, -1.9930560719274457E-01_dp, &
         1.0501296069740016E+02_dp, -3.5006611463522972E+01_dp, -2.5350753062858892E+01_dp, &
         2.1897185983038352E+01_dp, -1.2338035196247318E+01_dp, 1.7958095428676263E+01_dp, &
         -2.2746802023037219E+01_dp, 2.0474886895879344E+01_dp, -1.6659351440344015E+01_dp, &
         1.2973198485786074E+01_dp, -7.9821240373924969E+00_dp, 3.1045561581164107E+00_dp, &
         -5.5595266514842701E-01_dp, 2.5001854850064910E+02_dp, -1.0412293597281749E+02_dp, &
         -7.0012689271087496E+01_dp, 7.3141042075103357E+01_dp, -4.0139216815255786E+01_dp, &
         5.2378449759973961E+01_dp, -6.8602643632320721E+01_dp, 6.1798794377474557E+01_dp, &
         -4.7836268566179115E+01_dp, 3.6088120915352725E+01_dp, -2.2418738636667779E+01_dp, &
         8.9199275986530679E+00_dp, -1.6284440849442408E+00_dp, 6.4170847130306493E+02_dp, &
         -3.3564734161105105E+02_dp, -2.0319074233870467E+02_dp, 2.6700999172427709E+02_dp, &
         -1.4865370577376467E+02_dp, 1.6483905993906203E+02_dp, -2.2243071554945487E+02_dp, &
         2.0268427674625798E+02_dp, -1.4634920891505303E+02_dp, 1.0281958316987021E+02_dp, &
         -6.3797063574792020E+01_dp, 2.6253291146070694E+01_dp, -4.9653658173425370E+00_dp, &
         1.8179870531301624E+03_dp, -1.2124262579162116E+03_dp, -6.2269753019103860E+02_dp, &
         1.1041853779458711E+03_dp, -6.6269936764476824E+02_dp, 5.8460724515644620E+02_dp, &
         -7.9040872092006543E+02_dp, 7.4186477271070237E+02_dp, -4.8589122354266311E+02_dp, &
         2.8409351239811036E+02_dp, -1.6399352031256032E+02_dp, 7.0149332037368637E+01_dp, &
         -1.4138078938532878E+01_dp, 5.9147915732141646E+03_dp, -5.1836184148231387E+03_dp, &
         -1.9529616708993781E+03_dp, 5.4738511859880100E+03_dp, -3.8869620939089637E+03_dp, &
         2.5977148376502064E+03_dp, -3.1249029145093127E+03_dp, 3.0852665470111756E+03_dp, &
         -1.7708800172300193E+03_dp, 5.3074618739108712E+02_dp, -6.6377545473118147E+01_dp, &
         9.3948507249713860E+00_dp, -4.8647165579413887E+00_dp, 2.3862435794753223E+04_dp, &
         -2.9152813027206714E+04_dp, -4.0489848903694369E+03_dp, 3.5895063967658702E+04_dp, &
         -3.5486880126624863E+04_dp, 2.0119320410757315E+04_dp, -1.3494757960194409E+04_dp, &
         1.2166038486582942E+04_dp, -5.1302990256297890E+03_dp, -4.1716588565092670E+03_dp, &
         7.0751423497119849E+03_dp, -4.0035381834602572E+03_dp, 8.8748758779155071E+02_dp, &
         1.4686983146318793E+05_dp, -2.7739659047737991E+05_dp, 1.1760797338476461E+05_dp, &
         3.2807041818098148E+05_dp, -7.2517285349087697E+05_dp, 7.3150274091493548E+05_dp, &
         -2.9882997403572657E+05_dp, -3.1816573683119891E+05_dp, 7.7265175039547053E+05_dp, &
         -8.5110852845767897E+05_dp, 5.9525059737962019E+05_dp, -2.5347666927566967E+05_dp, &
         5.0206766624547767E+04_dp, 2.9045854252487524E-01_dp, -9.3872029923743098E-09_dp, &
         -1.6474589315474152E-02_dp, 1.8803148185921940E-02_dp, -2.1136377875972792E-02_dp, &
         1.0402051234732080E-02_dp, 2.7413508770381487E-02_dp, -1.0020420108756088E-01_dp, &
         1.9200020290840114E-01_dp, -2.5139939316453958E-01_dp, 2.2429428108518476E-01_dp, &
         -1.2324671659253950E-01_dp, 3.1416116213041342E-02_dp, 9.8543060415902617E-01_dp, &
         -3.9997789833057460E-08_dp, -7.0180574115183969E-02_dp, 8.0100018518159663E-02_dp, &
         -8.8669643824093369E-02_dp, 4.1168277654631887E-02_dp, 1.2222716004100626E-01_dp, &
         -4.3376557308075792E-01_dp, 8.2348524196548867E-01_dp, -1.0722306945113518E+00_dp, &
         9.5266694074435665E-01_dp, -5.2175051061240585E-01_dp, 1.3263698690739856E-01_dp, &
         2.0879815633551813E+00_dp, -1.1830858233257060E-07_dp, -2.0751780374971132E-01_dp, &
         2.3684851551994904E-01_dp, -2.5637862072869982E-01_dp, 1.0839683313493143E-01_dp, &
         3.8437385693103709E-01_dp, -1.3114065729390199E+00_dp, 2.4576389073052467E+00_dp, &
         -3.1745152556387226E+00_dp, 2.8036316578514442E+00_dp, -1.5280660715176184E+00_dp, &
         3.8691002445303968E-01_dp, 4.1280139396328233E+00_dp, -3.2843232523142131E-07_dp, &
         -5.7587200156643770E-01_dp, 6.5726548338828250E-01_dp, -6.9376957208833223E-01_dp, &
         2.6019334435131042E-01_dp, 1.1358828382244106E+00_dp, -3.7246497078122731E+00_dp, &
         6.8843611147032977E+00_dp, -8.8148483202359351E+00_dp, 7.7329874627101871E+00_dp, &
         -4.1917406592895956E+00_dp, 1.0565313621508341E+00_dp, 8.2115292599550127E+00_dp, &
         -8.9137561697151123E-07_dp, -1.5622875225930539E+00_dp, 1.7830984621994226E+00_dp, &
         -1.8292461199409722E+00_dp, 5.8447218617253227E-01_dp, 3.2858522987931740E+00_dp, &
         -1.0351403024506384E+01_dp, 1.8851002197235399E+01_dp, -2.3904379330116988E+01_dp, &
         2.0812769462943820E+01_dp, -1.1211467536930467E+01_dp, 2.8109944480167490E+00_dp, &
         1.6833509752553130E+01_dp, -2.4208347916624588E-06_dp, -4.2408362534061510E+00_dp, &
         4.8402218801150756E+00_dp, -4.7995870275879025E+00_dp, 1.2056343958145297E+00_dp, &
         9.5508495555428166E+00_dp, -2.8841634205337105E+01_dp, 5.1653143559226265E+01_dp, &
         -6.4765365903098953E+01_dp, 5.5884702982899860E+01_dp, -2.9877273510186630E+01_dp, &
         7.4426681746137389E+00_dp, 3.5994294559688392E+01_dp, -6.7185205986334922E-06_dp, &
         -1.1762365188147919E+01_dp, 1.3424795806058500E+01_dp, -1.2758760271073882E+01_dp, &
         2.0732889158938947E+00_dp, 2.8558432889956102E+01_dp, -8.2347703324680765E+01_dp, &
         1.4461633713546519E+02_dp, -1.7885516484206732E+02_dp, 1.5260695584062722E+02_dp, &
         -8.0803840777326485E+01_dp, 1.9960600023375445E+01_dp, 8.1157457670966565E+01_dp, &
         -1.9458094008644022E-05_dp, -3.4039161550681698E+01_dp, 3.8849992038065054E+01_dp, &
         -3.4927696505148425E+01_dp, 1.4181938934757687E+00_dp, 8.9934070855182398E+01_dp, &
         -2.4623095414911904E+02_dp, 4.2223594118137123E+02_dp, -5.1315822696394446E+02_dp, &
         4.3143053246951132E+02_dp, -2.2548223291042308E+02_dp, 5.5058148831168019E+01_dp, &
         1.9568047986849353E+02_dp, -6.0244262192986751E-05_dp, -1.0527645970752303E+02_dp, &
         1.2015506150003682E+02_dp, -1.0001639314265297E+02_dp, -1.4009458871747229E+01_dp, &
         3.0653332588013888E+02_dp, -7.9043502502542060E+02_dp, 1.3149170528475586E+03_dp, &
         -1.5608831673196967E+03_dp, 1.2853217606932562E+03_dp, -6.5913017450105701E+02_dp, &
         1.5816639621647650E+02_dp, 5.1524126870993928E+02_dp, -2.0601342380317144E-04_dp, &
         -3.5945477251100289E+02_dp, 4.1025436478750186E+02_dp, -3.0394364960641099E+02_dp, &
         -1.3411341942097474E+02_dp, 1.1740847511426989E+03_dp, -2.8153287548247690E+03_dp, &
         4.4935466649789141E+03_dp, -5.1526402294564268E+03_dp, 4.1075902980563333E+03_dp, &
         -2.0415866612899454E+03_dp, 4.7535023510802057E+02_dp, 1.5345203816734511E+03_dp, &
         -8.1996534428424410E-04_dp, -1.4272114721930720E+03_dp, 1.6288998293799721E+03_dp, &
         -9.8331472929207700E+02_dp, -1.0461299234634332E+03_dp, 5.3699628756611428E+03_dp, &
         -1.1700010870023980E+04_dp, 1.7522011300616890E+04_dp, -1.8931976535897018E+04_dp, &
         1.4192216319515732E+04_dp, -6.6082119103015066E+03_dp, 1.4354040244532143E+03_dp, &
         5.5359106840411023E+03_dp, -4.2116907744044847E-03_dp, -7.2978223910327642E+03_dp, &
         8.3290217141779613E+03_dp, -3.0205311888974306E+03_dp, -9.9636784453742785E+03_dp, &
         3.3034312432889732E+04_dp, -6.1784943330522874E+04_dp, 8.1361262384593938E+04_dp, &
         -7.5874525508600782E+04_dp, 4.6962603202435283E+04_dp, -1.6649687551097129E+04_dp, &
         2.3288381807198862E+03_dp, 2.8965940366189952E+04_dp, -3.7068444267143071E-02_dp, &
         -6.3565705168485889E+04_dp, 7.2545702838451150E+04_dp, 1.9451051149951527E+04_dp, &
         -1.9181670386420426E+05_dp, 3.6675255147557601E+05_dp, -4.0868713263820746E+05_dp, &
         1.8884705302897276E+05_dp, 2.3077739263055730E+05_dp, -5.1954761094616010E+05_dp, &
         4.2094053714259033E+05_dp, -1.3519426853223852E+05_dp, 2.3956446261391355E+00_dp, &
         -8.2259922726024793E-08_dp, -1.4435805810148564E-01_dp, 1.6476195304653121E-01_dp, &
         -1.8446144579171345E-01_dp, 8.9437025922576366E-02_dp, 2.4318075444129192E-01_dp, &
         -8.8181344063145117E-01_dp, 1.6854750516438730E+00_dp, -2.2036434429093879E+00_dp, &
         1.9639043208955855E+00_dp, -1.0782019877436793E+00_dp, 2.7464330062693504E-01_dp, &
         3.3494963759865057E+00_dp, -1.8445757660486408E-07_dp, -3.2356867403047279E-01_dp, &
         3.6930217014151012E-01_dp, -4.0166065484929164E-01_dp, 1.7339244143990060E-01_dp, &
         5.9190709926720098E-01_dp, -2.0357074336995966E+00_dp, 3.8253641076662022E+00_dp, &
         -4.9496028636172511E+00_dp, 4.3769722375906133E+00_dp, -2.3880787149857583E+00_dp, &
         6.0519139225210616E-01_dp, 5.8040000291750982E+00_dp, -4.9307010104776223E-07_dp, &
         -8.6457441827499182E-01_dp, 9.8677305172000451E-01_dp, -1.0435592710758324E+00_dp, &
         3.9518781491624627E-01_dp, 1.6981221840783756E+00_dp, -5.5841325524658263E+00_dp, &
         1.0332178192642571E+01_dp, -1.3238703735787881E+01_dp, 1.1620225704849606E+01_dp, &
         -6.3016961957941975E+00_dp, 1.5889503305951129E+00_dp, 1.1205889768777084E+01_dp, &
         -1.3209488146781724E-06_dp, -2.3152576012558059E+00_dp, 2.6424922127194401E+00_dp, &
         -2.7155773855174825E+00_dp, 8.7695933535320614E-01_dp, 4.8533303358467910E+00_dp, &
         -1.5324929900939482E+01_dp, 2.7934350894109556E+01_dp, -3.5445461425090642E+01_dp, &
         3.0877155066264496E+01_dp, -1.6640212268612309E+01_dp, 4.1736667265992375E+00_dp, &
         2.3003668057025042E+01_dp, -3.5482604603519923E-06_dp, -6.2161341452072962E+00_dp, &
         7.0947026113909137E+00_dp, -7.0528063622407950E+00_dp, 1.8077921798260204E+00_dp, &
         1.3939173281458793E+01_dp, -4.2219371371025446E+01_dp, 7.5708235249965000E+01_dp, &
         -9.5012975060705713E+01_dp, 8.2045747816253453E+01_dp, -4.3891539618467377E+01_dp, &
         1.0939766218489742E+01_dp, 4.9637684295142599E+01_dp, -9.7463695738026281E-06_dp, &
         -1.7064535222921169E+01_dp, 1.9476350353952395E+01_dp, -1.8590436540559089E+01_dp, &
         3.1925275522724674E+00_dp, 4.1158324783342913E+01_dp, -1.1921515605115852E+02_dp, &
         2.0979218330367425E+02_dp, -2.5985281813994447E+02_dp, 2.2199832068654112E+02_dp, &
         -1.1767584219618503E+02_dp, 2.9097217542476354E+01_dp, 1.1295662396434658E+02_dp/)
      REAL(KIND=dp), DIMENSION(90), PARAMETER :: c05 = (/-2.7907357414271074E-05_dp, &
         -4.8826106427338516E+01_dp, 5.5726828156877396E+01_dp, -5.0511477178292488E+01_dp, &
         2.9783692652581260E+00_dp, 1.2761594545085276E+02_dp, -3.5194583503612256E+02_dp, &
         6.0567238480576850E+02_dp, -7.3810346857458831E+02_dp, 6.2201716742048916E+02_dp, &
         -3.2577937709337687E+02_dp, 7.9700609240616870E+01_dp, 2.7392126417294304E+02_dp, &
         -8.5014454364132483E-05_dp, -1.4859738660861041E+02_dp, 1.6959859615136114E+02_dp, &
         -1.4348055742340679E+02_dp, -1.4470906151360252E+01_dp, 4.2504884354925576E+02_dp, &
         -1.1092388852844888E+03_dp, 1.8571994555057452E+03_dp, -2.2161045374396167E+03_dp, &
         1.8334805028895823E+03_dp, -9.4436397774699310E+02_dp, 2.2753963301137293E+02_dp, &
         7.1993494851256742E+02_dp, -2.8273561718275677E-04_dp, -4.9354947315699241E+02_dp, &
         5.6330058572048620E+02_dp, -4.3192291483542755E+02_dp, -1.5060398376844597E+02_dp, &
         1.5657840098177583E+03_dp, -3.8312638445092366E+03_dp, 6.1904109019097459E+03_dp, &
         -7.1744215164009311E+03_dp, 5.7783733007334085E+03_dp, -2.9012807842372895E+03_dp, &
         6.8229724960201429E+02_dp, 2.1065420056611683E+03_dp, -1.0664058503153191E-03_dp, &
         -1.8579764758900715E+03_dp, 2.1205446683502560E+03_dp, -1.3916730341031403E+03_dp, &
         -1.1054018979421496E+03_dp, 6.6616451407058066E+03_dp, -1.5054577477071649E+04_dp, &
         2.3136946283622921E+04_dp, -2.5637844083293843E+04_dp, 1.9746988549792022E+04_dp, &
         -9.4727339041338582E+03_dp, 2.1263470523015208E+03_dp, 7.1883726760512764E+03_dp, &
         -4.8683316646283012E-03_dp, -8.4557043884527229E+03_dp, 9.6505775179423381E+03_dp, &
         -4.6968102107445948E+03_dp, -8.7928604739436014E+03_dp, 3.5236438954359772E+04_dp, &
         -7.1416261140412083E+04_dp, 1.0113068401234987E+05_dp, -1.0299681880727473E+05_dp, &
         7.2039640653129478E+04_dp, -3.0819607047464066E+04_dp, 6.0214483913251506E+03_dp, &
         3.1445989346200007E+04_dp, -3.0789770328952418E-02_dp, -5.3162976062730428E+04_dp, &
         6.0674445128441919E+04_dp, -1.0676731288363026E+04_dp, -9.8618787001921461E+04_dp, &
         2.6917328710149962E+05_dp, -4.5080663875526667E+05_dp, 5.2615990446274763E+05_dp, &
         -4.0824771802677517E+05_dp, 1.7414668875913633E+05_dp, -1.1753804603528921E+04_dp, &
         -1.4548822319029938E+04_dp, 2.3068275228418666E+05_dp, -3.9689873696814088E-01_dp, &
         -6.7753046300068102E+05_dp, 7.7323771401862660E+05_dp, 5.5823523395335092E+05_dp, &
         -2.8474678682125942E+06_dp, 4.1915016169727501E+06_dp, -2.1435582467267131E+06_dp, &
         -4.2583193954617819E+06_dp, 1.1939190162504816E+07_dp, -1.4588866431316955E+07_dp, &
         9.6445890214436613E+06_dp, -2.7844693650067835E+06_dp/)
      REAL(KIND=dp), DIMENSION(13, 26, 5), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04, c05/), (/13, 26, 5/))

      INTEGER                                            :: irange

      IF (Rc <= 1.0000000000000000E+01_dp) Rc = 1.0000000000000000E+01_dp
      IF (Rc >= 1.5429601669311523E+05_dp) Rc = 1.5429601669311523E+05_dp
      IF (Rc >= 1.0000000000000000E+01_dp .AND. Rc <= 6.6546340495335400E+01_dp) THEN
         irange = 1
         L_b = 1.0000000000000000E+01_dp
         U_b = 6.6546340495335400E+01_dp
      END IF
      IF (Rc >= 6.6546340495335400E+01_dp .AND. Rc <= 8.8617575097672625E+02_dp) THEN
         irange = 2
         L_b = 6.6546340495335400E+01_dp
         U_b = 8.8617575097672625E+02_dp
      END IF
      IF (Rc >= 8.8617575097672625E+02_dp .AND. Rc <= 8.9415021812443374E+03_dp) THEN
         irange = 3
         L_b = 8.8617575097672625E+02_dp
         U_b = 8.9415021812443374E+03_dp
      END IF
      IF (Rc >= 8.9415021812443374E+03_dp .AND. Rc <= 5.2961915553991363E+04_dp) THEN
         irange = 4
         L_b = 8.9415021812443374E+03_dp
         U_b = 5.2961915553991363E+04_dp
      END IF
      IF (Rc >= 5.2961915553991363E+04_dp .AND. Rc <= 1.5429601669311523E+05_dp) THEN
         irange = 5
         L_b = 5.2961915553991363E+04_dp
         U_b = 1.5429601669311523E+05_dp
      END IF

      fit_coef(1:13, 1:26) = coefdata(1:13, 1:26, irange)

   END SUBROUTINE get_fit_coef_k13

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k14(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 28)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(220), PARAMETER :: c05 = (/-4.9474578762585689E-04_dp, &
         -2.5939403957639643E+03_dp, 3.0352874416549189E+03_dp, -1.9742442806612846E+03_dp, &
         -1.5969761709828460E+03_dp, 9.5497831811419619E+03_dp, -2.1791219784053799E+04_dp, &
         3.4007167063977518E+04_dp, -3.8426789298450683E+04_dp, 3.0290020612758195E+04_dp, &
         -1.4913740382161952E+04_dp, 3.4437263625752439E+03_dp, 9.7404664944032957E+03_dp, &
         -2.5387359568679966E-03_dp, -1.3140569581716762E+04_dp, 1.5376194259263802E+04_dp, &
         -6.3931081330686866E+03_dp, -1.6591959902315488E+04_dp, 5.9198447699858742E+04_dp, &
         -1.1572075729785630E+05_dp, 1.6031285635678476E+05_dp, -1.6009686057185219E+05_dp, &
         1.0949859530426569E+05_dp, -4.5488875426595805E+04_dp, 8.5191784067115168E+03_dp, &
         5.0965737544466407E+04_dp, -2.2599730435201849E-02_dp, -1.1343132088409952E+05_dp, &
         1.3272584229156727E+05_dp, 2.6288738068185427E+04_dp, -3.3499444278650463E+05_dp, &
         6.7051826395931852E+05_dp, -8.0700177432248381E+05_dp, 5.0680536128272885E+05_dp, &
         1.7276049875705931E+05_dp, -7.0698870636937302E+05_dp, 6.3021843911543733E+05_dp, &
         -2.1121368435571375E+05_dp, 2.3082323868282559E+00_dp, -2.4017040804702034E-08_dp, &
         -1.2891053118339380E-01_dp, 1.5084735253153100E-01_dp, -1.7633642214925768E-01_dp, &
         1.0478107279403694E-01_dp, 1.8056901435863362E-01_dp, -7.6047719259316937E-01_dp, &
         1.5304155276743796E+00_dp, -2.0748704774691396E+00_dp, 1.9060146879751447E+00_dp, &
         -1.0744161491583117E+00_dp, 2.8012092652976173E-01_dp, 3.1578187462227962E+00_dp, &
         -5.1567365353821572E-08_dp, -2.7650905240083334E-01_dp, 3.2356254508419818E-01_dp, &
         -3.6928537030522152E-01_dp, 2.0370200953976045E-01_dp, 4.2500849576389860E-01_dp, &
         -1.6818132491347872E+00_dp, 3.3294488150428045E+00_dp, -4.4727192170593986E+00_dp, &
         4.0824995398514528E+00_dp, -2.2901310317535963E+00_dp, 5.9479992770208523E-01_dp, &
         5.3022457526794620E+00_dp, -1.3231464400206018E-07_dp, -7.0879601306783835E-01_dp, &
         8.2941094524127557E-01_dp, -9.2491385814921556E-01_dp, 4.7111974770571152E-01_dp, &
         1.1800589827897907E+00_dp, -4.4312033118085346E+00_dp, 8.6424235805572369E+00_dp, &
         -1.1510883031319265E+01_dp, 1.0442741903695572E+01_dp, -5.8305757103901508E+00_dp, &
         1.5087006641831338E+00_dp, 9.8889975023413292E+00_dp, -3.4049023321205348E-07_dp, &
         -1.8221988606703043E+00_dp, 2.1322782325784657E+00_dp, -2.3235177347953093E+00_dp, &
         1.0834769924777288E+00_dp, 3.2578049089354004E+00_dp, -1.1683733051537128E+01_dp, &
         2.2469887916547904E+01_dp, -2.9680347503960153E+01_dp, 2.6764943785092932E+01_dp, &
         -1.4874114918252536E+01_dp, 3.8343556288753748E+00_dp, 1.9545871935955041E+01_dp, &
         -8.7460941189600850E-07_dp, -4.6754910887806371E+00_dp, 5.4711030020406684E+00_dp, &
         -5.8086891294068961E+00_dp, 2.4198279950235375E+00_dp, 8.9835555363766275E+00_dp, &
         -3.0776859857929882E+01_dp, 5.8311567939739909E+01_dp, -7.6326069298093813E+01_dp, &
         6.8369012228472059E+01_dp, -3.7794300439853124E+01_dp, 9.7011963380606723E+00_dp, &
         4.0387743977509039E+01_dp, -2.2802119015403580E-06_dp, -1.2173600984053655E+01_dp, &
         1.4245122575347299E+01_dp, -1.4663497964021953E+01_dp, 5.2167187058988729E+00_dp, &
         2.5244728902273341E+01_dp, -8.2452555727692769E+01_dp, 1.5363036456312540E+02_dp, &
         -1.9899236224931769E+02_dp, 1.7684535970055489E+02_dp, -9.7143346163942994E+01_dp, &
         2.4806140772967233E+01_dp, 8.7294816240247798E+01_dp, -6.1325232975282236E-06_dp, &
         -3.2687761451378606E+01_dp, 3.8250019320805777E+01_dp, -3.7906149312736396E+01_dp, &
         1.0554564587180893E+01_dp, 7.3600534264141842E+01_dp, -2.2847415147597096E+02_dp, &
         4.1762130504843361E+02_dp, -5.3421933339350812E+02_dp, 4.7021823009778637E+02_dp, &
         -2.5627821741621284E+02_dp, 6.5016683550652573E+01_dp, 1.9872757775908869E+02_dp, &
         -1.7288920250713497E-05_dp, -9.1967062673265403E+01_dp, 1.0761628002271890E+02_dp, &
         -1.0162265582966705E+02_dp, 1.7864790639628893E+01_dp, 2.2661260001503362E+02_dp, &
         -6.6576659613427580E+02_dp, 1.1897709602645350E+03_dp, -1.4987894458995706E+03_dp, &
         1.3032857841974862E+03_dp, -7.0315224410172698E+02_dp, 1.7686291682282348E+02_dp, &
         4.8195697213806216E+02_dp, -5.2053770241544080E-05_dp, -2.7616012508311786E+02_dp, &
         3.2315103970603087E+02_dp, -2.8617592824287709E+02_dp, 8.9668308727996777E+00_dp, &
         7.5242665407679601E+02_dp, -2.0796726290417987E+03_dp, 3.6157344965745160E+03_dp, &
         -4.4661943306924250E+03_dp, 3.8213658267393344E+03_dp, -2.0333395036853592E+03_dp, &
         5.0534042637321267E+02_dp, 1.2667258569739490E+03_dp, -1.7160973677295130E-04_dp, &
         -9.0710910652439338E+02_dp, 1.0614575664009355E+03_dp, -8.5818358894893083E+02_dp, &
         -1.6322151149271565E+02_dp, 2.7714382485293463E+03_dp, -7.1430478120354646E+03_dp, &
         1.1989987328766261E+04_dp, -1.4418456057909183E+04_dp, 1.2054815500043296E+04_dp, &
         -6.2834888773905532E+03_dp, 1.5330588576635123E+03_dp, 3.7064712285605369E+03_dp, &
         -6.4340614075563080E-04_dp, -3.3826536493755129E+03_dp, 3.9582066825917277E+03_dp, &
         -2.7744445291648535E+03_dp, -1.6114983670703609E+03_dp, 1.1820868123783597E+04_dp, &
         -2.8002618049413450E+04_dp, 4.4775862324928894E+04_dp, -5.1714676070415371E+04_dp, &
         4.1654323461293316E+04_dp, -2.0958910595013618E+04_dp, 4.9457226648390688E+03_dp, &
         1.2647982284255992E+04_dp, -2.9295350801209737E-03_dp, -1.5266725138724667E+04_dp, &
         1.7864187777650670E+04_dp, -9.5712259110815667E+03_dp, -1.4224742236571192E+04_dp, &
         6.2861882562494437E+04_dp, -1.3310141181211534E+05_dp, 1.9698892332270893E+05_dp, &
         -2.1146680880798530E+05_dp, 1.5786348899689523E+05_dp, -7.3269193628805137E+04_dp, &
         1.5867217147397359E+04_dp, 5.5329397744117472E+04_dp, -1.8590496792721553E-02_dp, &
         -9.5252723032791575E+04_dp, 1.1145715883398660E+05_dp, -2.5984748457593378E+04_dp, &
         -1.6824051744760206E+05_dp, 4.8493312244212057E+05_dp, -8.5016802056288661E+05_dp, &
         1.0575469301659723E+06_dp, -9.1592531311393599E+05_dp, 5.0087306406000553E+05_dp, &
         -1.3516194494539662E+05_dp, 4.9738319216673990E+03_dp, 4.0588762038136885E+05_dp, &
         -2.4363178639415667E-01_dp, -1.2048752475226286E+06_dp, 1.4098050382019044E+06_dp, &
         9.0383108002498816E+05_dp, -5.0245780711403918E+06_dp, 7.7751430360333202E+06_dp, &
         -4.9183263330898508E+06_dp, -5.5122178990547452E+06_dp, 1.8712359903900910E+07_dp, &
         -2.3939217503688838E+07_dp, 1.6218284186391968E+07_dp, -4.7629949218999417E+06_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.8855631887642382E-01_dp, &
         -8.1829231046632769E-02_dp, -1.8950903102790816E-02_dp, 1.1349898287774197E-02_dp, &
         5.7338935886024562E-03_dp, -3.6475885523936957E-03_dp, -2.1458998142384776E-03_dp, &
         1.4787405505064602E-03_dp, 1.8498537181843471E-03_dp, -1.9614757255708645E-03_dp, &
         2.8699603397392510E-04_dp, 3.2492719703003728E-04_dp, -1.2121707854025808E-04_dp, &
         5.9493156461017915E-01_dp, -2.8489994642562694E-01_dp, -4.7541679614047191E-02_dp, &
         4.4726439188938094E-02_dp, 1.3790272162181918E-02_dp, -1.5382737388282837E-02_dp, &
         -4.7841715747880403E-03_dp, 6.2784351364303555E-03_dp, 5.7333760277949548E-03_dp, &
         -9.0331645457294250E-03_dp, 3.7807714485503436E-03_dp, -1.8396222276796279E-04_dp, &
         -1.8364694814447738E-04_dp, 1.0949959297761693E+00_dp, -6.2009062084330724E-01_dp, &
         -3.7748088011084188E-02_dp, 1.1122644744095318E-01_dp, 6.8260044178406754E-03_dp, &
         -4.0370249812718174E-02_dp, -5.5457356283545060E-05_dp, 1.6531192945984253E-02_dp, &
         8.5778848944558107E-03_dp, -2.5250508617166625E-02_dp, 1.7117124110630336E-02_dp, &
         -5.0352617710504076E-03_dp, 5.0598347337982492E-04_dp, 1.7718966086100012E+00_dp, &
         -1.2253060640836630E+00_dp, 7.5765729412832300E-02_dp, 2.3434768509302889E-01_dp, &
         -4.2525503260917592E-02_dp, -8.5582364672439362E-02_dp, 2.5281751460349305E-02_dp, &
         3.4833638640193537E-02_dp, 3.6751693692180167E-03_dp, -5.3823028208479444E-02_dp, &
         4.8983441632312753E-02_dp, -1.9467338617706484E-02_dp, 3.0785669504690746E-03_dp, &
         2.7440213704191687E+00_dp, -2.3281411168383572E+00_dp, 4.3930618759633422E-01_dp, &
         4.2911680926061485E-01_dp, -1.9302208134528706E-01_dp, -1.4788624688089952E-01_dp, &
         9.7831155740447415E-02_dp, 5.8724172544457748E-02_dp, -2.7041547750709143E-02_dp, &
         -8.6324492364363792E-02_dp, 1.0519866830647166E-01_dp, -4.9884486746561728E-02_dp, &
         9.2457653338246029E-03_dp, 4.1894298957234364E+00_dp, -4.3191941088244246E+00_dp, &
         1.3686987339076067E+00_dp, 6.6657149222134049E-01_dp, -5.5741457317177090E-01_dp, &
         -1.9026437846512242E-01_dp, 2.6250861035098944E-01_dp, 6.9486254930157559E-02_dp, &
         -1.1978905880115075E-01_dp, -8.3501865329870154E-02_dp, 1.7398289631638658E-01_dp, &
         -9.8507600097967690E-02_dp, 2.0521730335798096E-02_dp, 6.3847973817046322E+00_dp, &
         -7.8795893221752813E+00_dp, 3.5278548969550236E+00_dp, 7.7209394256709085E-01_dp, &
         -1.3260774785983964E+00_dp, -8.5419091308552686E-02_dp, 5.7755377384646511E-01_dp, &
         7.2297671316667676E-03_dp, -3.2877429737301961E-01_dp, 5.4329500825053369E-02_dp, &
         1.9667270764503703E-01_dp, -1.5053976717237291E-01_dp, 3.5936788548594702E-02_dp, &
         9.7702622697290806E+00_dp, -1.4206713723167212E+01_dp, 8.2873928530423839E+00_dp, &
         1.5985839391104073E-01_dp, -2.7552157435202256E+00_dp, 5.0282122616039127E-01_dp, &
         1.0695678398638087E+00_dp, -2.7464467867173409E-01_dp, -6.9906236411372280E-01_dp, &
         5.2101436020119296E-01_dp, 1.9094129828846389E-02_dp, -1.4958848604568709E-01_dp, &
         4.7222211859139149E-02_dp, 1.5062823541404761E+01_dp, -2.5419370366400564E+01_dp, &
         1.8439365337386121E+01_dp, -2.8259558244586622E+00_dp, -4.9818385838230759E+00_dp, &
         2.3299538569069234E+00_dp, 1.5633478274790813E+00_dp, -1.0657994340246808E+00_dp, &
         -1.1646925905079819E+00_dp, 1.6169430708139660E+00_dp, -6.8586900071703705E-01_dp, &
         5.1235251652361868E-02_dp, 2.8103584030951066E-02_dp, 2.3472242707636227E+01_dp, &
         -4.5307745602761770E+01_dp, 3.9608173260058393E+01_dp, -1.2469876908780039E+01_dp, &
         -7.2257373755414482E+00_dp, 6.7879665547641341E+00_dp, 1.1935112761475244E+00_dp, &
         -2.7643113712900522E+00_dp, -1.2334897284577950E+00_dp, 3.5447845919979937E+00_dp, &
         -2.4336600487586937E+00_dp, 7.5137492324226218E-01_dp, -8.4754014224136878E-02_dp, &
         3.7181849843463610E+01_dp, -8.0854637759410053E+01_dp, 8.3057755018582043E+01_dp, &
         -3.9158151519184301E+01_dp, -5.2153320438563107E+00_dp, 1.5431055894370806E+01_dp, &
         -2.5633281395845020E+00_dp, -5.2800516403359623E+00_dp, 6.0861233401307058E-01_dp, &
         5.3386911924762455E+00_dp, -5.2478418325908383E+00_dp, 2.1981670777697651E+00_dp, &
         -3.6896267338134209E-01_dp, 6.0727490472197132E+01_dp, -1.4613546055640202E+02_dp, &
         1.7197354948450220E+02_dp, -1.0678616520983826E+02_dp, 1.5368848940430027E+01_dp, &
         2.6491023605600720E+01_dp, -1.4185436270935341E+01_dp, -6.0093236121070364E+00_dp, &
         6.0468948387099335E+00_dp, 3.8095072307193587E+00_dp, -7.1410151818281697E+00_dp, &
         3.7963272372520303E+00_dp, -7.5598613223750699E-01_dp, 1.0661517386729834E+02_dp, &
         -2.7755039079968356E+02_dp, 3.6291502644599831E+02_dp, -2.7257937328142549E+02_dp, &
         9.2825581734245375E+01_dp, 2.6166156567735605E+01_dp, -3.5739847667562742E+01_dp, &
         5.4601725492393982E-01_dp, 1.3032015168131986E+01_dp, -2.3134927090253932E+00_dp, &
         -6.3108826059815888E+00_dp, 4.6857674694757945E+00_dp, -1.0832084080117372E+00_dp, &
         2.3570689411128097E+02_dp, -6.4519456443598233E+02_dp, 9.0223673164973013E+02_dp, &
         -7.5922137375758791E+02_dp, 3.5264115885243655E+02_dp, -2.0989321935745537E+01_dp, &
         -6.8957744635106934E+01_dp, 1.6380082435687662E+01_dp, 2.3421535674813764E+01_dp, &
         -1.2505473268777655E+01_dp, -5.5686200874137679E+00_dp, 6.7565951949370930E+00_dp, &
         -1.7824763136779194E+00_dp, 1.5275739094339187E+00_dp, -6.8001171975001351E-01_dp, &
         -1.4605823697436165E-01_dp, 9.7828030861734694E-02_dp, 4.3904288873512894E-02_dp, &
         -3.2132952455475414E-02_dp, -1.6222343050214385E-02_dp, 1.3057021240371581E-02_dp, &
         1.5021168482517486E-02_dp, -1.7795844885780759E-02_dp, 4.1587460671975805E-03_dp, &
         1.8886977368616257E-03_dp, -8.6439907575529956E-04_dp, 1.7655488232124410E+00_dp, &
         -1.0043589160047419E+00_dp, -6.4881261437890497E-02_dp, 1.8408970278911196E-01_dp, &
         1.3505968144741386E-02_dp, -6.7697647980935388E-02_dp, -1.4104212623691880E-03_dp, &
         2.7789428772147087E-02_dp, 1.5476046749690131E-02_dp, -4.2895663223273178E-02_dp, &
         2.8323690685459327E-02_dp, -7.9885591764726257E-03_dp, 7.2098905617051475E-04_dp, &
         2.2893354930889771E+00_dp, -1.7662816033235220E+00_dp, 1.8403075477202577E-01_dp, &
         3.6407316931374950E-01_dp, -8.6454388639239418E-02_dp, -1.3732188470605050E-01_dp, &
         4.7671855898007404E-02_dp, 5.6196647324577562E-02_dp, 3.2014114180754462E-03_dp, &
         -8.9528747999384858E-02_dp, 8.4420352577280325E-02_dp, -3.4350841194717596E-02_dp, &
         5.5441745297109164E-03_dp, 3.2032258575320007E+00_dp, -3.2225125249560986E+00_dp, &
         8.1611931560454820E-01_dp, 6.3294842865066814E-01_dp, -3.4599308009457763E-01_dp, &
         -2.2371900126276956E-01_dp, 1.7236277609166772E-01_dp, 8.9265078243513069E-02_dp, &
         -5.3947604792708909E-02_dp, -1.3470741896044980E-01_dp, 1.7532582289902740E-01_dp, &
         -8.5701035571192671E-02_dp, 1.6222152583645552E-02_dp, 4.6899825303755200E+00_dp, &
         -5.8524582698866139E+00_dp, 2.2916446118015137E+00_dp, 9.1184997824365766E-01_dp, &
         -9.3131462349314253E-01_dp, -2.5350817564214317E-01_dp, 4.3807172987756571E-01_dp, &
         9.0958165596269375E-02_dp, -2.1576258659606037E-01_dp, -1.0117137327241729E-01_dp, &
         2.6846770077261650E-01_dp, -1.5994395563622057E-01_dp, 3.4270164544904627E-02_dp, &
         7.0536044987990474E+00_dp, -1.0515715134514387E+01_dp, 5.5637593381103434E+00_dp, &
         8.8711884475899694E-01_dp, -2.1104915247624749E+00_dp, -8.1670293239061807E-03_dp, &
         9.2046944679099196E-01_dp, -4.1741761795501105E-02_dp, -5.6379831367556721E-01_dp, &
         1.8783938079412943E-01_dp, 2.4581866594692456E-01_dp, -2.2071302879217192E-01_dp, &
         5.5588648839480607E-02_dp, 1.0795630733012548E+01_dp, -1.8740084666512331E+01_dp, &
         1.2578600627233692E+01_dp, -3.9401650322504667E-01_dp, -4.2184539233060248E+00_dp, &
         1.0568277055345616E+00_dp, 1.6271512205476346E+00_dp, -5.4967161451363589E-01_dp, &
         -1.1438065064071192E+00_dp, 1.0422345804455202E+00_dp, -1.6188730221930633E-01_dp, &
         -1.6095431410441777E-01_dp, 6.2846743147671746E-02_dp, 1.6752881329272565E+01_dp, &
         -3.3242655345938644E+01_dp, 2.7270848317691222E+01_dp, -5.5130334892731048E+00_dp, &
         -7.3490722258440666E+00_dp, 4.1336683392056157E+00_dp, 2.2217585152496331E+00_dp, &
         -1.8973233340123217E+00_dp, -1.7925847714925607E+00_dp, 2.8980257671054690E+00_dp, &
         -1.4704996700404205E+00_dp, 2.6049665793720700E-01_dp, 1.1444579602089748E-02_dp, &
         2.6366237165338223E+01_dp, -5.8894321733298867E+01_dp, 5.7521256562790946E+01_dp, &
         -2.1015040091428240E+01_dp, -1.0093110698669442E+01_dp, 1.1421770115513882E+01_dp, &
         1.2209454420475951E+00_dp, -4.7246394893237085E+00_dp, -1.6540481482401830E+00_dp, &
         6.0698584679024128E+00_dp, -4.5445980010104039E+00_dp, 1.5511339489005815E+00_dp, &
         -2.0629950073069225E-01_dp, 4.2291510559584431E+01_dp, -1.0465649315579324E+02_dp, &
         1.1897915051644942E+02_dp, -6.2607403829996528E+01_dp, -5.4209821188017733E+00_dp, &
         2.5469691021577223E+01_dp, -5.7934227183011000E+00_dp, -8.9450737399818969E+00_dp, &
         2.1130287274913053E+00_dp, 8.8579986644448212E+00_dp, -9.6059615158346361E+00_dp, &
         4.2847833469932413E+00_dp, -7.6136171546806874E-01_dp, 7.0150982322458802E+01_dp, &
         -1.8840659621861244E+02_dp, 2.4308030630616531E+02_dp, -1.6583105497089588E+02_dp, &
         3.0198027338055482E+01_dp, 4.3137539621256487E+01_dp, -2.7521349718582130E+01_dp, &
         -9.6205094694534150E+00_dp, 1.4155430281644003E+01_dp, 3.0610045263527681E+00_dp, &
         -1.1486453193732402E+01_dp, 6.8958779080000516E+00_dp, -1.4735848497004862E+00_dp, &
         1.2521662423711092E+02_dp, -3.5446949046616129E+02_dp, 5.0095178511895006E+02_dp, &
         -4.0924920293285766E+02_dp, 1.5804385724333696E+02_dp, 3.6770269869154461E+01_dp, &
         -6.7438382515530122E+01_dp, 8.2106438781909876E+00_dp, 2.7715716832749418E+01_dp, &
         -1.6544637350100938E+01_dp, -1.8960171584647745E+00_dp, 5.0230826620958835E+00_dp, &
         -1.4542149169175325E+00_dp, 2.6891905074142079E+02_dp, -7.7262572953372262E+02_dp, &
         1.1373217134902056E+03_dp, -1.0196332565900419E+03_dp, 5.2341144727679796E+02_dp, &
         -6.1641665846344743E+01_dp, -1.0165109376939833E+02_dp, 4.4685439920384923E+01_dp, &
         2.7038021992864703E+01_dp, -3.0106447966381346E+01_dp, 6.1843452503431493E+00_dp, &
         3.2384773408937977E+00_dp, -1.3899494804286554E+00_dp, 1.0072868276053762E+03_dp, &
         -2.8456064913973623E+03_dp, 4.1444213836381341E+03_dp, -3.7170886189465332E+03_dp, &
         1.9829818649289750E+03_dp, -3.8268355348089460E+02_dp, -2.1461243783889245E+02_dp, &
         9.3530786144130829E+01_dp, 8.1716992654821155E+01_dp, -6.6241416941144067E+01_dp, &
         -4.3199384438482218E+00_dp, 1.9601634716394887E+01_dp, -5.9258959456424520E+00_dp, &
         2.3570619090301786E-01_dp, -5.5162312535829566E-02_dp, -1.8870279804443474E-02_dp, &
         5.9631585077347333E-03_dp, 5.2228509160063231E-03_dp, -1.8582311804501090E-03_dp, &
         -2.0959249000252859E-03_dp, 6.5294058434205886E-04_dp, 1.3620688255760083E-03_dp, &
         -8.3422748679588279E-04_dp, -3.2153474478749297E-04_dp, 4.5243977001023549E-04_dp, &
         -1.2591995144422757E-04_dp, 7.6587911600589598E-01_dp, -2.0875636096653921E-01_dp, &
         -6.0166290603338807E-02_dp, 2.9038121823338276E-02_dp, 1.7462634992676026E-02_dp, &
         -9.9487645929256206E-03_dp, -7.1366951595613149E-03_dp, 3.6713354494958815E-03_dp, &
         5.4941565312529340E-03_dp, -5.1227538205351423E-03_dp, 3.8905702168758006E-04_dp, &
         1.1382083100254657E-03_dp, -3.9779179490065093E-04_dp, 1.4924251913695632E+00_dp, &
         -5.1831019334185813E-01_dp, -1.0671352469554986E-01_dp, 9.3839903685068102E-02_dp, &
         3.2490232771881414E-02_dp, -3.4916436688308940E-02_dp, -1.3379509563995902E-02_dp, &
         1.3447654845333435E-02_dp, 1.4076996424835847E-02_dp, -2.0087908870094155E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/7.7725039723149673E-03_dp, &
         1.1145706792479703E-04_dp, -5.6097890673260616E-04_dp, 2.6197138414537928E+00_dp, &
         -1.1878855591113402E+00_dp, -1.3669541461047707E-01_dp, 2.5880924857422194E-01_dp, &
         3.9410663226141771E-02_dp, -1.0161291074659663E-01_dp, -1.4845376770825033E-02_dp, &
         4.0384191854304169E-02_dp, 2.9596626868951859E-02_dp, -6.2931154480019291E-02_dp, &
         3.6745945037802222E-02_dp, -8.3488837850756864E-03_dp, 2.5001869374288916E-04_dp, &
         4.4903769459069487E+00_dp, -2.6294155124375926E+00_dp, -5.5431338024852818E-02_dp, &
         6.4004292000876550E-01_dp, -1.1281945374730250E-02_dp, -2.5862693537498738E-01_dp, &
         1.5347628611049238E-02_dp, 1.0516496345674861E-01_dp, 4.7284909960751409E-02_dp, &
         -1.6707094690313704E-01_dp, 1.2446700525343438E-01_dp, -4.1183388221565025E-02_dp, &
         5.0811070975616355E-03_dp, 7.7146703444230118E+00_dp, -5.7066349635112275E+00_dp, &
         4.4359998413126639E-01_dp, 1.4564973028002151E+00_dp, -2.8264653691673164E-01_dp, &
         -5.8828533359250146E-01_dp, 1.6490349909564089E-01_dp, 2.4225156854407701E-01_dp, &
         2.9313258361025520E-02_dp, -3.8043499940822334E-01_dp, 3.4922268468144424E-01_dp, &
         -1.4038340165679025E-01_dp, 2.2435164708763913E-02_dp, 1.3420634891703846E+01_dp, &
         -1.2271053658859397E+01_dp, 2.2633203019324784E+00_dp, 3.0726848709689221E+00_dp, &
         -1.2487174695539813E+00_dp, -1.1832394997237425E+00_dp, 6.8072898930880155E-01_dp, &
         4.8388636378359617E-01_dp, -1.7386926345701451E-01_dp, -7.0971295084118979E-01_dp, &
         8.3444731110830883E-01_dp, -3.9085840738204530E-01_dp, 7.1877924239686128E-02_dp, &
         2.3755596487151276E+01_dp, -2.6373483711130035E+01_dp, 7.9674708316710250E+00_dp, &
         5.9098701221696377E+00_dp, -4.1849668364839099E+00_dp, -1.9539106200179810E+00_dp, &
         2.1803405618804597E+00_dp, 7.4953893987111575E-01_dp, -1.0194767520566530E+00_dp, &
         -8.6460471090210544E-01_dp, 1.6344868078214525E+00_dp, -9.0779850081241142E-01_dp, &
         1.8748507335135670E-01_dp, 4.2925038468224415E+01_dp, -5.7082416348496004E+01_dp, &
         2.4765245687853245E+01_dp, 9.6276167314465653E+00_dp, -1.2283278522629377E+01_dp, &
         -1.7758222239526547E+00_dp, 5.9861494789457375E+00_dp, 3.8161879101773316E-01_dp, &
         -3.6179634751249212E+00_dp, 5.5154525020794043E-01_dp, 2.1657632444483794E+00_dp, &
         -1.6489718555108390E+00_dp, 3.9384533483919071E-01_dp, 7.9426971888974009E+01_dp, &
         -1.2525063212940651E+02_dp, 7.3064335204686586E+01_dp, 9.0748492057192287E+00_dp, &
         -3.2688968221322831E+01_dp, 4.6389115697221399E+00_dp, 1.4061626099831328E+01_dp, &
         -3.4152596807585240E+00_dp, -9.7690378370180984E+00_dp, 7.9275061732557406E+00_dp, &
         -6.3075146750735867E-01_dp, -1.6211192224607696E+00_dp, 5.6714810234651192E-01_dp, &
         1.5111743347765383E+02_dp, -2.8035613867011256E+02_dp, 2.1163825913793113E+02_dp, &
         -2.3170413390173934E+01_dp, -7.7181490178355602E+01_dp, 3.6759610462492383E+01_dp, &
         2.4870389441934346E+01_dp, -1.9605319761356419E+01_dp, -1.8579389164179108E+01_dp, &
         3.0778110404354127E+01_dp, -1.6525494041464249E+01_dp, 3.4913991145457492E+00_dp, &
         -6.0347075936734364E-02_dp, 2.9794125234183332E+02_dp, -6.4468281291214532E+02_dp, &
         6.1350680831270904E+02_dp, -2.0829770735425490E+02_dp, -1.3833043913329655E+02_dp, &
         1.5148578230935769E+02_dp, 6.9315274799328979E+00_dp, -6.5837068538278658E+01_dp, &
         -5.7004307370424572E+00_dp, 7.0003047590575193E+01_dp, -6.1017865054790505E+01_dp, &
         2.3916329575898800E+01_dp, -3.7939335527980216E+00_dp, 6.2392037586903575E+02_dp, &
         -1.5490091248195286E+03_dp, 1.8055806125944553E+03_dp, -1.0360644128662241E+03_dp, &
         -2.3754789025977079E+00_dp, 4.1933356341119969E+02_dp, -1.8722583764205237E+02_dp, &
         -1.0263434903918866E+02_dp, 1.0213234231038781E+02_dp, 3.9318968261747074E+01_dp, &
         -9.2918679848214254E+01_dp, 5.1419134926361373E+01_dp, -1.0473717025421935E+01_dp, &
         1.5736554804419109E+03_dp, -4.3215073024710609E+03_dp, 5.8432235337189013E+03_dp, &
         -4.5330421089344291E+03_dp, 1.6066632187764733E+03_dp, 4.6048780056295902E+02_dp, &
         -7.0252498313679098E+02_dp, 9.1567126340641437E+01_dp, 2.1597646777874064E+02_dp, &
         -8.2131833864215508E+01_dp, -7.0306756274872612E+01_dp, 6.5598722538748618E+01_dp, &
         -1.6248587274560606E+01_dp, 1.9235349518444129E+00_dp, -4.6850080403507538E-01_dp, &
         -1.5364029163795673E-01_dp, 5.4610434870493545E-02_dp, 4.3121661743123693E-02_dp, &
         -1.7541190516424592E-02_dp, -1.7406011867825987E-02_dp, 6.2616285228102664E-03_dp, &
         1.1811592459976310E-02_dp, -8.2397883673727759E-03_dp, -1.8295153305730790E-03_dp, &
         3.5354495356089955E-03_dp, -1.0301987802026438E-03_dp, 2.4066428993785296E+00_dp, &
         -8.3136637617833942E-01_dp, -1.7886500853362544E-01_dp, 1.4914120143861584E-01_dp, &
         5.5966341016553810E-02_dp, -5.4968206983759794E-02_dp, -2.3425654868914157E-02_dp, &
         2.1010142579782284E-02_dp, 2.3289369427945999E-02_dp, -3.1028342173622683E-02_dp, &
         1.0636269323796589E-02_dp, 1.1399336207735340E-03_dp, -1.0908676737398028E-03_dp, &
         3.5374997704069169E+00_dp, -1.7739971394119649E+00_dp, -1.8178559222630922E-01_dp, &
         4.0632103157798233E-01_dp, 5.8516869919659278E-02_dp, -1.6056878832594537E-01_dp, &
         -2.3367297155693821E-02_dp, 6.3796767939619234E-02_dp, 4.7371253056304186E-02_dp, &
         -9.9633558625783894E-02_dp, 5.7560660830321013E-02_dp, -1.2736642321518233E-02_dp, &
         2.7575782625087675E-04_dp, 5.7011521784732837E+00_dp, -3.8423532249372361E+00_dp, &
         -2.9642882719094453E-03_dp, 9.8774471298748978E-01_dp, -3.1786519630271917E-02_dp, &
         -4.0451564192974432E-01_dp, 2.7198223261336543E-02_dp, 1.6513101391234630E-01_dp, &
         7.5034065435769917E-02_dp, -2.6516920000527039E-01_dp, 1.9761798053561705E-01_dp, &
         -6.5322652133406461E-02_dp, 8.0342411952895529E-03_dp, 9.6554344109573194E+00_dp, &
         -8.2331879085041422E+00_dp, 8.3524862618218354E-01_dp, 2.2044272451218019E+00_dp, &
         -4.7078752828290560E-01_dp, -9.0706474881686883E-01_dp, 2.6681861233539966E-01_dp, &
         3.7661042003703288E-01_dp, 4.5852032020570602E-02_dp, -6.0242502481174587E-01_dp, &
         5.5531680066652089E-01_dp, -2.2393716090138521E-01_dp, 3.5887703311978866E-02_dp, &
         1.6859940902920791E+01_dp, -1.7545862869698755E+01_dp, 3.6977782998502717E+00_dp, &
         4.5720615862129801E+00_dp, -1.9866126947412956E+00_dp, -1.8011363549668156E+00_dp, &
         1.0806874902511809E+00_dp, 7.4694472223237718E-01_dp, -2.7704282161362154E-01_dp, &
         -1.1227694450891366E+00_dp, 1.3298672608779971E+00_dp, -6.2600213805058025E-01_dp, &
         1.1557688164935315E-01_dp, 3.0155576587510065E+01_dp, -3.7501345657849221E+01_dp, &
         1.2398647864361692E+01_dp, 8.6755274737656904E+00_dp, -6.5226597957606725E+00_dp, &
         -2.9431221713683908E+00_dp, 3.4329970746320631E+00_dp, 1.1571252075338336E+00_dp, &
         -1.6281640072535604E+00_dp, -1.3658547679386668E+00_dp, 2.6187047264334584E+00_dp, &
         -1.4634970806763015E+00_dp, 3.0358383342450590E-01_dp, 5.5192028159048192E+01_dp, &
         -8.0962165488777188E+01_dp, 3.7555961619090525E+01_dp, 1.3994299732088162E+01_dp, &
         -1.8953701082797508E+01_dp, -2.6423170124495257E+00_dp, 9.4281011864102400E+00_dp, &
         6.1853673917080609E-01_dp, -5.8552996587823642E+00_dp, 9.3429973997614690E-01_dp, &
         3.5009224511745121E+00_dp, -2.6876808075590444E+00_dp, 6.4489189999489860E-01_dp, &
         1.0352661090439724E+02_dp, -1.7762218682439524E+02_dp, 1.0901655066345789E+02_dp, &
         1.3191430681842892E+01_dp, -5.0434357437755253E+01_dp, 7.0189212422942697E+00_dp, &
         2.2479507564050024E+01_dp, -5.1663403209676666E+00_dp, -1.6257258538476439E+01_dp, &
         1.3112384987334350E+01_dp, -1.0057705657907436E+00_dp, -2.7227979411177912E+00_dp, &
         9.5157750952935138E-01_dp, 1.9966810298056458E+02_dp, -3.9817155755323489E+02_dp, &
         3.1231785694487940E+02_dp, -3.2246955189265840E+01_dp, -1.2137675854903618E+02_dp, &
         5.5648852060851773E+01_dp, 4.2338485599882645E+01_dp, -3.0713631015622173E+01_dp, &
         -3.3372363691858347E+01_dp, 5.2473159250251847E+01_dp, -2.7472517355576315E+01_dp, &
         5.4871815877910279E+00_dp, 5.4181088396383879E-03_dp, 3.9842112960611036E+02_dp, &
         -9.1713248145130524E+02_dp, 8.9864978724752018E+02_dp, -2.9197223519853361E+02_dp, &
         -2.3777047005535999E+02_dp, 2.3747379514120729E+02_dp, 2.9189361759954174E+01_dp, &
         -1.1235024332118923E+02_dp, -2.5228006012287803E+01_dp, 1.3666058782503137E+02_dp, &
         -1.1319713359079486E+02_dp, 4.3059250023675311E+01_dp, -6.6306785236143213E+00_dp, &
         8.3643909862800683E+02_dp, -2.1911581438517564E+03_dp, 2.6233201738911630E+03_dp, &
         -1.4629002578603033E+03_dp, -1.5531076779586360E+02_dp, 7.5096190656559827E+02_dp, &
         -2.6513199393656015E+02_dp, -2.5306291751220505E+02_dp, 1.9176224755216879E+02_dp, &
         1.2171911094119822E+02_dp, -2.2483087333153333E+02_dp, 1.2085756124968174E+02_dp, &
         -2.4483946936633131E+01_dp, 1.9747033956733776E+03_dp, -5.6690082004591668E+03_dp, &
         7.9229880703795752E+03_dp, -6.2070030014723598E+03_dp, 1.9283153836933900E+03_dp, &
         1.2213601402606237E+03_dp, -1.4629461716889127E+03_dp, 1.7493772651497704E+02_dp, &
         6.0147021238394541E+02_dp, -4.2899362538784328E+02_dp, 5.2240406552156081E+01_dp, &
         5.8761089300646205E+01_dp, -2.1062064230578645E+01_dp, 7.3293182665232616E+03_dp, &
         -2.1390250361943301E+04_dp, 3.1430512423742246E+04_dp, -2.8054361932601565E+04_dp, &
         1.4411560239732573E+04_dp, -1.9245318108120027E+03_dp, -2.4770186693515961E+03_dp, &
         1.1408945117448648E+03_dp, 5.1709534814360416E+02_dp, -5.0521280755054079E+02_dp, &
         -3.5875569643025585E+01_dp, 1.5800852296117296E+02_dp, -4.8036967247801527E+01_dp, &
         2.6675516114666215E-01_dp, -3.0945355721284026E-02_dp, -1.7441440218854541E-02_dp, &
         2.3047365097519259E-03_dp, 4.6492436549931005E-03_dp, -1.2182731057025141E-04_dp, &
         -2.6753899877316541E-03_dp, 4.3133453711237618E-04_dp, 1.4760115578547502E-03_dp, &
         -6.7275877348429937E-04_dp, -5.0478172177037875E-04_dp, 5.2672478258151327E-04_dp, &
         -1.3681815219244913E-04_dp, 8.8716914786510592E-01_dp, -1.2489263551188355E-01_dp, &
         -6.6159449712017482E-02_dp, 1.3827016752023739E-02_dp, 1.9071901611964450E-02_dp, &
         -2.2900215117290321E-03_dp, -1.1259614209975406E-02_dp, 2.5040307358945432E-03_dp, &
         6.8228796479948492E-03_dp, -4.2395551594276134E-03_dp, -1.1853357048091078E-03_dp, &
         1.9375771856216684E-03_dp, -5.4683602988019086E-04_dp, 1.8085261758917643E+00_dp, &
         -3.4145203794204038E-01_dp, -1.6386900603394911E-01_dp, 5.5237575273534616E-02_dp, &
         5.2221001537078073E-02_dp, -1.3801174494884877E-02_dp, -3.2078301452457908E-02_dp, &
         1.0176963366296529E-02_dp, 2.2098874328449269E-02_dp, -1.8474914130739314E-02_dp, &
         1.0331691102100885E-03_dp, 4.1466751154130713E-03_dp, -1.4017603905798956E-03_dp, &
         3.3844323807837364E+00_dp, -8.6969086014254704E-01_dp, -3.7007448537416970E-01_dp, &
         1.8591067907939515E-01_dp, 1.2836223540354583E-01_dp, -5.7108370411460219E-02_dp, &
         -8.2699128673198688E-02_dp, 3.6214620670927643E-02_dp, 6.5374140319992780E-02_dp, &
         -6.9244074088779542E-02_dp, 1.8299747317206676E-02_dp, 5.5068417777118968E-03_dp, &
         -2.9577800214382753E-03_dp, 6.2831509132867023E+00_dp, -2.1503252769443533E+00_dp, &
         -7.9051209248628784E-01_dp, 5.6778852490396248E-01_dp, 2.8990581299559376E-01_dp, &
         -2.0042137877984503E-01_dp, -1.9784689657716242E-01_dp, 1.1942781843739504E-01_dp, &
         1.8202124035828973E-01_dp, -2.3777194345333863E-01_dp, 9.8820369139971026E-02_dp, &
         -6.2744671316908003E-03_dp, -4.3699523495120468E-03_dp, 1.1856221271082283E+01_dp, &
         -5.2568207235502564E+00_dp, -1.5969164750173954E+00_dp, 1.6461069205315015E+00_dp, &
         5.9172061051969216E-01_dp, -6.4697064642905655E-01_dp, -4.3622701630183242E-01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/3.7852818456494469E-01_dp, &
         4.8067555347683755E-01_dp, -7.7452581263261322E-01_dp, 4.1934407249887440E-01_dp, &
         -8.9234614648328783E-02_dp, 1.6775465007990663E-03_dp, 2.2972132228999541E+01_dp, &
         -1.2898901071815617E+01_dp, -2.9797860863789967E+00_dp, 4.6560435611069622E+00_dp, &
         1.0045232845484455E+00_dp, -1.9936051302786315E+00_dp, -8.4531105176808208E-01_dp, &
         1.1798428768823785E+00_dp, 1.1850316641391561E+00_dp, -2.4272082281943557E+00_dp, &
         1.5965055566291400E+00_dp, -4.8094513367183356E-01_dp, 5.2209596934856856E-02_dp, &
         4.6003576082453456E+01_dp, -3.2201217855327890E+01_dp, -4.7001617664051283E+00_dp, &
         1.3071605527953821E+01_dp, 8.5579097973198270E-01_dp, -5.9644456694242445E+00_dp, &
         -1.1661056085226225E+00_dp, 3.6676542174044551E+00_dp, 2.5328675562596046E+00_dp, &
         -7.2813422948153237E+00_dp, 5.6732651577216711E+00_dp, -2.0778493148878012E+00_dp, &
         3.0597521984172754E-01_dp, 9.5841344366560733E+01_dp, -8.2867209315180261E+01_dp, &
         -3.6667980705527965E+00_dp, 3.6810280324243195E+01_dp, -3.9027065352258030E+00_dp, &
         -1.7352727135169747E+01_dp, 7.6980699945374509E-01_dp, 1.1432996249675226E+01_dp, &
         3.2078234224178548E+00_dp, -2.0167035286623779E+01_dp, 1.8815816418699637E+01_dp, &
         -7.9856869761396307E+00_dp, 1.3743706676833676E+00_dp, 2.0942078257013398E+02_dp, &
         -2.2297406228411018E+02_dp, 1.7976991795066482E+01_dp, 1.0433146940440160E+02_dp, &
         -3.3593272459503659E+01_dp, -4.7934874758223437E+01_dp, 1.6761961834599632E+01_dp, &
         3.5307462464268248E+01_dp, -1.0676959688175113E+01_dp, -4.5186467607051981E+01_dp, &
         5.5390800171275316E+01_dp, -2.7302072934110832E+01_dp, 5.2930427116892105E+00_dp, &
         4.8534437062529378E+02_dp, -6.3827294390772897E+02_dp, 1.5269862076475343E+02_dp, &
         2.9397876898906549E+02_dp, -1.8666196685194382E+02_dp, -1.1267861965482709E+02_dp, &
         1.0302221723226668E+02_dp, 1.0064780441992042E+02_dp, -1.2667959907993337E+02_dp, &
         -2.7955679739330272E+01_dp, 1.1636709243491977E+02_dp, -7.4278132336508548E+01_dp, &
         1.6627724231379677E+01_dp, 1.2129457544590350E+03_dp, -1.9898851218084496E+03_dp, &
         8.9667454269579969E+02_dp, 7.6499660811561989E+02_dp, -9.4460273295149761E+02_dp, &
         -7.9982084258437382E+01_dp, 4.9143681651282088E+02_dp, 1.7577870976013608E+02_dp, &
         -7.5560526889570701E+02_dp, 5.5420839419383049E+02_dp, -9.7827251223882456E+01_dp, &
         -6.5994619153302679E+01_dp, 2.7322580954253830E+01_dp, 3.3629380940608125E+03_dp, &
         -6.9947471832857618E+03_dp, 5.2569129046974813E+03_dp, 9.5662372221786302E+02_dp, &
         -4.5223771656660383E+03_dp, 2.1388578525409239E+03_dp, 1.4458258137230293E+03_dp, &
         -9.7995910573338404E+02_dp, -2.3056338385778722E+03_dp, 3.9539197752574801E+03_dp, &
         -2.7936521089069415E+03_dp, 1.0067700692768599E+03_dp, -1.5230449843591444E+02_dp, &
         1.1287607903620121E+04_dp, -2.9770904965426005E+04_dp, 3.5105294651488788E+04_dp, &
         -1.6371982820338246E+04_dp, -1.0740533254974671E+04_dp, 2.2130378005569939E+04_dp, &
         -1.3681292248100397E+04_dp, 1.5045907807150209E+03_dp, 1.5506170412673375E+03_dp, &
         1.7390235047408335E+03_dp, -3.3865390974169732E+03_dp, 1.9425096877525270E+03_dp, &
         -4.1245955612424916E+02_dp, 2.1894397151835716E+00_dp, -2.6731395928049057E-01_dp, &
         -1.4827254994752054E-01_dp, 2.2501949118577419E-02_dp, 4.0392878417795181E-02_dp, &
         -2.0509313301437277E-03_dp, -2.3398171933515566E-02_dp, 4.1436012276036211E-03_dp, &
         1.3239233793160993E-02_dp, -6.6345121073634033E-03_dp, -3.9199986704907238E-03_dp, &
         4.4620529455242711E-03_dp, -1.1821222579064985E-03_dp, 2.9104358928220684E+00_dp, &
         -5.4034884374736947E-01_dp, -2.6382498632517509E-01_dp, 8.3423398368213414E-02_dp, &
         8.3577127933356615E-02_dp, -1.9596957492619492E-02_dp, -5.0894988316946159E-02_dp, &
         1.5007727408384405E-02_dp, 3.4099429011554168E-02_dp, -2.6790339407196072E-02_dp, &
         -2.0861762641897246E-04_dp, 7.2003288329720860E-03_dp, -2.3042530203007294E-03_dp, &
         4.6854870375971425E+00_dp, -1.3091559374612594E+00_dp, -5.5604618168351427E-01_dp, &
         2.8363733949175418E-01_dp, 1.9722299942319932E-01_dp, -8.5724412786704784E-02_dp, &
         -1.2678366277947672E-01_dp, 5.3898967563011380E-02_dp, 9.9143592343204467E-02_dp, &
         -1.0221615198869149E-01_dp, 2.4645321523710888E-02_dp, 9.8027723202895409E-03_dp, &
         -4.7528229423811326E-03_dp, 8.3449364444476224E+00_dp, -3.1882784788266245E+00_dp, &
         -1.1623777798125512E+00_dp, 8.5932749444592937E-01_dp, 4.4006378157100107E-01_dp, &
         -3.0113783137236694E-01_dp, -3.0064333658838277E-01_dp, 1.7727771396624017E-01_dp, &
         2.7500235701742781E-01_dp, -3.5164491355954486E-01_dp, 1.4063352553582015E-01_dp, &
         -5.3061541860887768E-03_dp, -7.3941722723575680E-03_dp, 1.5695844879515739E+01_dp, &
         -7.7144784516164409E+00_dp, -2.3165992414383219E+00_dp, 2.4635839948403095E+00_dp, &
         8.9286928883939432E-01_dp, -9.6611614969126347E-01_dp, -6.6064108196517513E-01_dp, &
         5.5828529367439539E-01_dp, 7.2643541444293003E-01_dp, -1.1445679759723104E+00_dp, &
         6.0431143104611718E-01_dp, -1.2059693868497956E-01_dp, -2.0295090300914784E-04_dp, &
         3.0683428112515067E+01_dp, -1.8785984761642442E+01_dp, -4.2812807144192888E+00_dp, &
         6.9041136192452699E+00_dp, 1.5281967131175347E+00_dp, -2.9618491125087343E+00_dp, &
         -1.2909060055596731E+00_dp, 1.7294637699195108E+00_dp, 1.8115339805586135E+00_dp, &
         -3.5961083965134120E+00_dp, 2.3152629313913131E+00_dp, -6.7559378765190137E-01_dp, &
         6.8467347739389728E-02_dp, 6.2200394666586682E+01_dp, -4.6645513761643670E+01_dp, &
         -6.7361062081138403E+00_dp, 1.9252946239137508E+01_dp, 1.4565082813699259E+00_dp, &
         -8.8439939458906807E+00_dp, -1.8841850547213637E+00_dp, 5.3507933397910890E+00_dp, &
         4.0391961741194118E+00_dp, -1.0903704455343615E+01_dp, 8.2986239071848669E+00_dp, &
         -2.9674077434627351E+00_dp, 4.2355807369087978E-01_dp, 1.3120724805702122E+02_dp, &
         -1.1950869933893051E+02_dp, -5.5268211617968701E+00_dp, 5.3988318262052609E+01_dp, &
         -4.7298665550070194E+00_dp, -2.5840083994726847E+01_dp, 3.9210953422711342E-01_dp, &
         1.6659556581442590E+01_dp, 6.3342926153832790E+00_dp, -3.1128246651007331E+01_dp, &
         2.8060315969863534E+01_dp, -1.1623356406032910E+01_dp, 1.9545704467217686E+00_dp, &
         2.8992281668366485E+02_dp, -3.1986021667598641E+02_dp, 2.2957602830758823E+01_dp, &
         1.5295309506645341E+02_dp, -4.3977099689006337E+01_dp, -7.2842752827744860E+01_dp, &
         2.1240405440259895E+01_dp, 5.1988276730246653E+01_dp, -7.3415208227066628E+00_dp, &
         -7.6698168738036316E+01_dp, 8.6864530178114521E+01_dp, -4.1342020475172049E+01_dp, &
         7.8179704154115273E+00_dp, 6.7758098148149588E+02_dp, -9.0749625734977383E+02_dp, &
         1.9757172893678359E+02_dp, 4.3536816979902625E+02_dp, -2.4496931051749999E+02_dp, &
         -1.8568872790586860E+02_dp, 1.3676710116513863E+02_dp, 1.5660775089407684E+02_dp, &
         -1.4802736727848816E+02_dp, -1.0503595627665368E+02_dp, 2.1660252926971174E+02_dp, &
         -1.2550574533876951E+02_dp, 2.6842986324742590E+01_dp, 1.6983835710324086E+03_dp, &
         -2.7822602350002267E+03_dp, 1.1271088712112896E+03_dp, 1.1986681939191794E+03_dp, &
         -1.2298857533356895E+03_dp, -3.0022794237993980E+02_dp, 6.8655519186623917E+02_dp, &
         3.7926097327365630E+02_dp, -1.0131416464518528E+03_dp, 4.8887586578955757E+02_dp, &
         1.5852175267429436E+02_dp, -2.3155394891253087E+02_dp, 6.4867684171672508E+01_dp, &
         4.6647494493474896E+03_dp, -9.4718275026936517E+03_dp, 6.2415276573638348E+03_dp, &
         2.5681659744319936E+03_dp, -6.0603657311991774E+03_dp, 1.3404500115938827E+03_dp, &
         2.8367905176183167E+03_dp, -2.1312330539639728E+02_dp, -4.7786812976805413E+03_dp, &
         5.7342345176005756E+03_dp, -3.1010309026004620E+03_dp, 7.8357948046534830E+02_dp, &
         -5.9923253166677775E+01_dp, 1.4640183187865061E+04_dp, -3.7306916780068022E+04_dp, &
         3.8220031520362521E+04_dp, -6.2809044505597049E+03_dp, -2.6739415575057752E+04_dp, &
         2.5342729378908178E+04_dp, 1.2027700096180724E+03_dp, -1.4004701367972091E+04_dp, &
         -8.0202934875566336E+02_dp, 1.9450058535585955E+04_dp, -2.0396420577318539E+04_dp, &
         9.6146216278380598E+03_dp, -1.8456278447511768E+03_dp, 6.3728309588229160E+04_dp, &
         -1.9308392701878195E+05_dp, 2.8218302673358534E+05_dp, -2.2238214531146569E+05_dp, &
         3.7703671533692228E+04_dp, 1.3438185722279068E+05_dp, -1.8992995882626902E+05_dp, &
         1.4401513945138350E+05_dp, -7.4358417823577242E+04_dp, 2.9628900862353326E+04_dp, &
         -1.0445794075970865E+04_dp, 3.1174329654997350E+03_dp, -5.1375865413784766E+02_dp, &
         2.7795875304093992E-01_dp, -1.1508375559190450E-02_dp, -1.0528648097849034E-02_dp, &
         3.5980305121916505E-03_dp, -3.8046195921370984E-03_dp, 6.2928606186302520E-03_dp, &
         -7.1251215695943898E-03_dp, 6.9881088552248760E-03_dp, -6.6162502135251478E-03_dp, &
         5.4184133074171737E-03_dp, -3.2882049375220451E-03_dp, 1.2671473115865579E-03_dp, &
         -2.3060149967202290E-04_dp, 9.3295242665284350E-01_dp, -4.7623070133923896E-02_dp, &
         -4.2945598361340688E-02_dp, 1.6016174188036112E-02_dp, -1.5648429729476797E-02_dp, &
         2.6076911362094328E-02_dp, -2.9735563415651905E-02_dp, 2.8978847027043658E-02_dp, &
         -2.7276350421693667E-02_dp, 2.2330036824631784E-02_dp, -1.3565168762016968E-02_dp, &
         5.2270999974457248E-03_dp, -9.4990442008638116E-04_dp, 1.9360104180064521E+00_dp, &
         -1.3504439794671419E-01_dp, -1.1921435653077275E-01_dp, 5.0013006045341514E-02_dp, &
         -4.4100911881113043E-02_dp, 7.4033972248669144E-02_dp, -8.5347616016655026E-02_dp, &
         8.2444773869143534E-02_dp, -7.6903614035383622E-02_dp, 6.2915522354710013E-02_dp, &
         -3.8288814369473227E-02_dp, 1.4757779755536996E-02_dp, -2.6771107475512521E-03_dp, &
         3.7157437793497445E+00_dp, -3.5796615175869162E-01_dp, -3.0857012350133511E-01_dp, &
         1.4568419777186900E-01_dp, -1.1664877750278947E-01_dp, 1.9623630214417229E-01_dp, &
         -2.2916236385065755E-01_dp, 2.1940133178072269E-01_dp, -2.0246833223513266E-01_dp, &
         1.6546652950802851E-01_dp, -1.0094501676761625E-01_dp, 3.8944876117642600E-02_dp, &
         -7.0548301311543953E-03_dp, 7.1202810265841023E+00_dp, -9.2376909827681186E-01_dp, &
         -7.7556752166862253E-01_dp, 4.1180402375069397E-01_dp, -3.0219866001661555E-01_dp, &
         5.0555399088003872E-01_dp, -5.9930306647530585E-01_dp, 5.6887825139019754E-01_dp, &
         -5.1830619426463687E-01_dp, 4.2285671923122187E-01_dp, -2.5880981419593402E-01_dp, &
         1.0005085991364741E-01_dp, -1.8112309200688195E-02_dp, 1.3952971079099326E+01_dp, &
         -2.3685732416421441E+00_dp, -1.9287282183697398E+00_dp, 1.1570319194902128E+00_dp, &
         -7.8432675091254189E-01_dp, 1.2914143383187207E+00_dp, -1.5583873119306928E+00_dp, &
         1.4673017772412877E+00_dp, -1.3158217342584138E+00_dp, 1.0703445295466383E+00_dp, &
         -6.5795894538224242E-01_dp, 2.5527153608136027E-01_dp, -4.6237736535300604E-02_dp, &
         2.8263649428018645E+01_dp, -6.1403574477253944E+00_dp, -4.8190110843037894E+00_dp, &
         3.2973992582013865E+00_dp, -2.0826346006536562E+00_dp, 3.3265966189184617E+00_dp, &
         -4.1006380665334703E+00_dp, 3.8330107038269401E+00_dp, -3.3679881939021974E+00_dp, &
         2.7252672290525357E+00_dp, -1.6848847973698691E+00_dp, 6.5763593614104732E-01_dp, &
         -1.1942745604194921E-01_dp, 5.9663070196049929E+01_dp, -1.6360639408342024E+01_dp, &
         -1.2260457072232811E+01_dp, 9.7081093752801308E+00_dp, -5.7831301102621087E+00_dp, &
         8.7770738779350062E+00_dp, -1.1095172356986756E+01_dp, 1.0314862293028543E+01_dp, &
         -8.8233143705921826E+00_dp, 7.0699089843034164E+00_dp, -4.4024286308950149E+00_dp, &
         1.7350313146222107E+00_dp, -3.1699935067362639E-01_dp, 1.3246229877703036E+02_dp, &
         -4.5573195710967674E+01_dp, -3.2154929851246941E+01_dp, 3.0089567929013004E+01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c04 = (/-1.7228109793468004E+01_dp, &
         2.4108285826088757E+01_dp, -3.1362122687167119E+01_dp, 2.9105748579791722E+01_dp, &
         -2.4011018345543842E+01_dp, 1.8874442598138845E+01_dp, -1.1844908362658735E+01_dp, &
         4.7389221341553531E+00_dp, -8.7620232886140437E-01_dp, 3.1319177678658934E+02_dp, &
         -1.3543802765633262E+02_dp, -8.7944249327211026E+01_dp, 1.0040343220265778E+02_dp, &
         -5.6926025700847731E+01_dp, 7.0347720403350522E+01_dp, -9.4219186285633356E+01_dp, &
         8.7905354648438234E+01_dp, -6.8986573372461578E+01_dp, 5.2100412701990848E+01_dp, &
         -3.2821779741733202E+01_dp, 1.3425508412863300E+01_dp, -2.5361926958927188E+00_dp, &
         8.0339006237151955E+02_dp, -4.4154448996371457E+02_dp, -2.5318298497317585E+02_dp, &
         3.7186998732251453E+02_dp, -2.1890233755863807E+02_dp, 2.2518951062497459E+02_dp, &
         -3.0685526869647856E+02_dp, 2.9150513350211389E+02_dp, -2.1341406154345805E+02_dp, &
         1.4697365335701707E+02_dp, -9.0644866260771636E+01_dp, 3.8104543183927014E+01_dp, &
         -7.4637665568833587E+00_dp, 2.3044953848651026E+03_dp, -1.6499343359986115E+03_dp, &
         -7.6329960501783944E+02_dp, 1.6025382549709414E+03_dp, -1.0575686504725684E+03_dp, &
         8.4751945418489061E+02_dp, -1.1070023776663299E+03_dp, 1.0916471478962596E+03_dp, &
         -7.2797923049358553E+02_dp, 3.9110381271759127E+02_dp, -2.0138082123320197E+02_dp, &
         8.3652333810026974E+01_dp, -1.7347611535672375E+01_dp, 7.8198644325855139E+03_dp, &
         -7.6691862797499716E+03_dp, -2.1277219734434225E+03_dp, 8.7238270472921049E+03_dp, &
         -7.3961781231503164E+03_dp, 4.6349606671658858E+03_dp, -4.4852525866010510E+03_dp, &
         4.4961185730836178E+03_dp, -2.6534205822788208E+03_dp, 4.4195703445871567E+02_dp, &
         4.6808102343591264E+02_dp, -3.3365812265666904E+02_dp, 7.3857858542740814E+01_dp, &
         3.6001520527708155E+04_dp, -5.3761146837784952E+04_dp, 7.3437178260156779E+03_dp, &
         6.7383560520903891E+04_dp, -1.0382851510328006E+05_dp, 8.2809008442700942E+04_dp, &
         -3.3648889547776744E+04_dp, -1.2152726558027885E+04_dp, 4.1767740945577811E+04_dp, &
         -5.0768326126419706E+04_dp, 3.9062230702036293E+04_dp, -1.8084327244239081E+04_dp, &
         3.8220474286796107E+03_dp, 2.2865375061568898E+00_dp, -1.0007118233392230E-01_dp, &
         -9.1205849541503861E-02_dp, 3.1914700695052964E-02_dp, -3.3024390409020138E-02_dp, &
         5.4742737190489880E-02_dp, -6.2096136424361001E-02_dp, 6.0798325126475448E-02_dp, &
         -5.7475761170505700E-02_dp, 4.7065964441797066E-02_dp, -2.8569616417685219E-02_dp, &
         1.1009159982093808E-02_dp, -2.0027154539687037E-03_dp, 3.1115361446896435E+00_dp, &
         -2.1232738914527213E-01_dp, -1.8819487078015418E-01_dp, 7.7315071595482399E-02_dp, &
         -6.9322784492525144E-02_dp, 1.1642512083091060E-01_dp, -1.3389325658449983E-01_dp, &
         1.2953231476076033E-01_dp, -1.2106215221167207E-01_dp, 9.9062132162017258E-02_dp, &
         -6.0257874244903968E-02_dp, 2.3219784186475492E-02_dp, -4.2129071313934735E-03_dp, &
         5.1842137469273535E+00_dp, -5.3869903618846926E-01_dp, -4.6467926474567495E-01_dp, &
         2.1884730673172473E-01_dp, -1.7511335735521882E-01_dp, 2.9552329416544632E-01_dp, &
         -3.4478150622705367E-01_dp, 3.3007122270082284E-01_dp, -3.0482944942907819E-01_dp, &
         2.4917344203144623E-01_dp, -1.5196363061763693E-01_dp, 5.8604719193897523E-02_dp, &
         -1.0613235696316487E-02_dp, 9.5870634342206369E+00_dp, -1.3711377799260795E+00_dp, &
         -1.1512181070647907E+00_dp, 6.1170380517292722E-01_dp, -4.4707416132200822E-01_dp, &
         7.5102145460616387E-01_dp, -8.8972944046919122E-01_dp, 8.4414679164287010E-01_dp, &
         -7.6958368953859069E-01_dp, 6.2806586528884911E-01_dp, -3.8429278159433666E-01_dp, &
         1.4848195960403690E-01_dp, -2.6866781887765284E-02_dp, 1.8775367525208154E+01_dp, &
         -3.4798037363209451E+00_dp, -2.8341983855308737E+00_dp, 1.7006763707506694E+00_dp, &
         -1.1475201905709496E+00_dp, 1.8990746846216189E+00_dp, -2.2902563189929959E+00_dp, &
         2.1549148240268319E+00_dp, -1.9340794515934272E+00_dp, 1.5741729298799292E+00_dp, &
         -9.6735481018495850E-01_dp, 3.7504025715761041E-01_dp, -6.7883194547012568E-02_dp, &
         3.8394133622760471E+01_dp, -8.9466981562030412E+00_dp, -7.0282409736060609E+00_dp, &
         4.7994825716760108E+00_dp, -3.0169750276153215E+00_dp, 4.8524390956684931E+00_dp, &
         -5.9767609648935736E+00_dp, 5.5816268619870462E+00_dp, -4.9114774016645590E+00_dp, &
         3.9786679592871335E+00_dp, -2.4588990184794475E+00_dp, 9.5872721501601932E-01_dp, &
         -1.7391335067678992E-01_dp, 8.1981262849015678E+01_dp, -2.3666893439637072E+01_dp, &
         -1.7784927738480093E+01_dp, 1.3988139300590353E+01_dp, -8.2896531742092350E+00_dp, &
         1.2715968834351342E+01_dp, -1.6055214717081412E+01_dp, 1.4906635835795530E+01_dp, &
         -1.2785323476847376E+01_dp, 1.0268923034784310E+01_dp, -6.3923625597427707E+00_dp, &
         2.5150132422925426E+00_dp, -4.5864970193217036E-01_dp, 1.8391160612598173E+02_dp, &
         -6.5390879598833081E+01_dp, -4.6443209608932030E+01_dp, 4.2801184197871855E+01_dp, &
         -2.4328927514598988E+01_dp, 3.4666363971514954E+01_dp, -4.5025337204534765E+01_dp, &
         4.1693953701520407E+01_dp, -3.4572251788912048E+01_dp, 2.7325062230277453E+01_dp, &
         -1.7152184655455500E+01_dp, 6.8433372267470105E+00_dp, -1.2610819268429596E+00_dp, &
         4.3796390510262017E+02_dp, -1.9195269795546488E+02_dp, -1.2647230203731048E+02_dp, &
         1.4008459142409632E+02_dp, -7.8348031326383392E+01_dp, 9.9970385018688177E+01_dp, &
         -1.3379095649086614E+02_dp, 1.2428806571516949E+02_dp, -9.8477370673399108E+01_dp, &
         7.5392430125667914E+01_dp, -4.7653420100430495E+01_dp, 1.9413552472266861E+01_dp, &
         -3.6453602426561473E+00_dp, 1.1243161739960412E+03_dp, -6.1212969783621759E+02_dp, &
         -3.6255926320818696E+02_dp, 5.0245861498975745E+02_dp, -2.8720395504014408E+02_dp, &
         3.1238265825952902E+02_dp, -4.2845042492346460E+02_dp, 4.0342871100091708E+02_dp, &
         -3.0042323465953200E+02_dp, 2.1481280954412966E+02_dp, -1.3503140509729658E+02_dp, &
         5.6622775693708434E+01_dp, -1.0979355574831448E+01_dp, 3.1860045248229740E+03_dp, &
         -2.1911750278244835E+03_dp, -1.0979433579892768E+03_dp, 2.0455378472242323E+03_dp, &
         -1.2650045628271755E+03_dp, 1.1053428711376532E+03_dp, -1.5077317283874904E+03_dp, &
         1.4637188657276913E+03_dp, -1.0022416048330521E+03_dp, 6.0744063403024404E+02_dp, &
         -3.5491809474315596E+02_dp, 1.5302472467201832E+02_dp, -3.1334099788707611E+01_dp, &
         1.0369219521864628E+04_dp, -9.2957237022933496E+03_dp, -3.3928819922045304E+03_dp, &
         9.9967321591592136E+03_dp, -7.3074927639255448E+03_dp, 4.9261330266433706E+03_dp, &
         -5.9355709483907294E+03_dp, 6.0546709760503109E+03_dp, -3.7342004116968578E+03_dp, &
         1.3546695864229043E+03_dp, -3.3203558186954626E+02_dp, 9.8087982450693005E+01_dp, &
         -2.5243614322976278E+01_dp, 4.1857193674808586E+04_dp, -5.1932163797323534E+04_dp, &
         -6.6662468997263541E+03_dp, 6.4686288551278376E+04_dp, -6.5456790256758810E+04_dp, &
         3.8148473534729375E+04_dp, -2.6150038718665659E+04_dp, 2.4469246590260118E+04_dp, &
         -1.2503913145751001E+04_dp, -4.7022652434212587E+03_dp, 1.1183043091421301E+04_dp, &
         -6.6810238092235504E+03_dp, 1.5108300867808391E+03_dp, 2.5791029976014077E+05_dp, &
         -4.9175046576096409E+05_dp, 2.1233667264455085E+05_dp, 5.8463884123176278E+05_dp, &
         -1.3091052712645272E+06_dp, 1.3418160613037723E+06_dp, -5.8138326599118719E+05_dp, &
         -5.2549948487002426E+05_dp, 1.3542841741758550E+06_dp, -1.5110583922361752E+06_dp, &
         1.0622068635754019E+06_dp, -4.5309312113443937E+05_dp, 8.9720576460898301E+04_dp, &
         2.8045183132902857E-01_dp, -2.7574727763525926E-09_dp, -1.4802663783945194E-02_dp, &
         1.7321648653547309E-02_dp, -2.0314993734315081E-02_dp, 1.2188204632943614E-02_dp, &
         2.0453784311654996E-02_dp, -8.6946069733443987E-02_dp, 1.7538269591846428E-01_dp, &
         -2.3808144401169168E-01_dp, 2.1890014297902782E-01_dp, -1.2347609817248510E-01_dp, &
         3.2209458319119928E-02_dp, 9.4329847413827783E-01_dp, -1.1481913919029099E-08_dp, &
         -6.1604975697409345E-02_dp, 7.2088324040036539E-02_dp, -8.3499507949812257E-02_dp, &
         4.8263387714568257E-02_dp, 8.9536268205010608E-02_dp, -3.6778510358608535E-01_dp, &
         7.3540775557696592E-01_dp, -9.9349209869932542E-01_dp, 9.1038678244132876E-01_dp, &
         -5.1222364328553327E-01_dp, 1.3335054733354942E-01_dp, 1.9654700761868769E+00_dp, &
         -3.2854337036704686E-08_dp, -1.7614103276566062E-01_dp, 2.0611488927720040E-01_dp, &
         -2.3440872269300381E-01_dp, 1.2780362813152599E-01_dp, 2.7418366288979301E-01_dp, &
         -1.0758527122495292E+00_dp, 2.1248452756492062E+00_dp, -2.8506473734734499E+00_dp, &
         2.5994713960584761E+00_dp, -1.4571442401661396E+00_dp, 3.7823535850863244E-01_dp, &
         3.7941873168540208E+00_dp, -8.7953097131811444E-08_dp, -4.7113570657516318E-01_dp, &
         5.5130825970707775E-01_dp, -6.1431256183219685E-01_dp, 3.1203070731979304E-01_dp, &
         7.8613980333408151E-01_dp, -2.9472754643680608E+00_dp, 5.7453376661729898E+00_dp, &
         -7.6498729270533463E+00_dp, 6.9384231145105142E+00_dp, -3.8732848550438543E+00_dp, &
         1.0020928498724406E+00_dp, 7.3237070985773096E+00_dp, -2.2941985723774246E-07_dp, &
         -1.2277519929151568E+00_dp, 1.4366756798913567E+00_dp, -1.5650422648304791E+00_dp, &
         7.2887187219335803E-01_dp, 2.1962398474654630E+00_dp, -7.8721698163374203E+00_dp, &
         1.5136322777799633E+01_dp, -1.9990503168612264E+01_dp, 1.8024781758740165E+01_dp, &
         -1.0015971255468424E+01_dp, 2.5817837059480153E+00_dp, 1.4477457547611932E+01_dp, &
         -5.9542599972157469E-07_dp, -3.1829060629238062E+00_dp, 3.7245298868254464E+00_dp, &
         -3.9522312305350300E+00_dp, 1.6423393285449270E+00_dp, 6.1216563001120274E+00_dp, &
         -2.0953918860770138E+01_dp, 3.9686837113822115E+01_dp, -5.1935176897343290E+01_dp, &
         4.6512173586647570E+01_dp, -2.5707915801780647E+01_dp, 6.5979718743648332E+00_dp, &
         2.9632190699838180E+01_dp, -1.5657622900881051E-06_dp, -8.3586532272678387E+00_dp, &
         9.7810033020084397E+00_dp, -1.0055079501227743E+01_dp, 3.5508201891380611E+00_dp, &
         1.7377218506217933E+01_dp, -5.6648334572026592E+01_dp, 1.0547111094388774E+02_dp, &
         -1.3654326425592572E+02_dp, 1.2129760971899201E+02_dp, -6.6608157929477457E+01_dp, &
         1.7004088480309473E+01_dp, 6.3338122170095403E+01_dp, -4.2444484549751587E-06_dp, &
         -2.2620321314472132E+01_dp, 2.6469466355873809E+01_dp, -2.6151213910751117E+01_dp, &
         7.1147272422339958E+00_dp, 5.1213821549050515E+01_dp, -1.5836988990793796E+02_dp, &
         2.8901910480803514E+02_dp, -3.6930520446037576E+02_dp, 3.2477605951571485E+02_dp, &
         -1.7687999522963116E+02_dp, 4.4846043091496973E+01_dp, 1.4279947172069063E+02_dp, &
         -1.2080229456795278E-05_dp, -6.4239399665069129E+01_dp, 7.5170425103726117E+01_dp, &
         -7.0510474822968519E+01_dp, 1.1363573169598904E+01_dp, 1.5996866863552859E+02_dp, &
         -4.6665591365326327E+02_dp, 8.3131873091270222E+02_dp, -1.0448759235890909E+03_dp, &
         9.0690542685405194E+02_dp, -4.8852833718934801E+02_dp, 1.2271331760718991E+02_dp, &
         3.4430186167302065E+02_dp, -3.6917560063034570E-05_dp, -1.9573425600940379E+02_dp, &
         2.2903991435173910E+02_dp, -1.9998078868584736E+02_dp, -3.6337191265627644E-01_dp, &
         5.4328299769275566E+02_dp, -1.4832817481049449E+03_dp, 2.5633679410315954E+03_dp, &
         -3.1519867754826223E+03_dp, 2.6865378429486905E+03_dp, -1.4246759516690913E+03_dp, &
         3.5301663828659417E+02_dp, 9.0657027822430757E+02_dp, -1.2505227876874551E-04_dp, &
         -6.6016663782193780E+02_dp, 7.7249587755502171E+02_dp, -6.0564642881603584E+02_dp, &
         -1.6334533603788654E+02_dp, 2.0809169575290675E+03_dp, -5.2520819705711947E+03_dp, &
         8.7142803953445527E+03_dp, -1.0381449011333283E+04_dp, 8.6066748064158692E+03_dp, &
         -4.4514243053148093E+03_dp, 1.0783194357224340E+03_dp, 2.6999974727364965E+03_dp/)
      REAL(KIND=dp), DIMENSION(13, 28, 5), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04, c05/), (/13, 28, 5/))

      INTEGER                                            :: irange

      IF (Rc <= 1.0000000000000000E+01_dp) Rc = 1.0000000000000000E+01_dp
      IF (Rc >= 2.7148483276367188E+05_dp) Rc = 2.7148483276367188E+05_dp
      IF (Rc >= 1.0000000000000000E+01_dp .AND. Rc <= 1.0920229970956126E+02_dp) THEN
         irange = 1
         L_b = 1.0000000000000000E+01_dp
         U_b = 1.0920229970956126E+02_dp
      END IF
      IF (Rc >= 1.0920229970956126E+02_dp .AND. Rc <= 1.1839524369543471E+03_dp) THEN
         irange = 2
         L_b = 1.0920229970956126E+02_dp
         U_b = 1.1839524369543471E+03_dp
      END IF
      IF (Rc >= 1.1839524369543471E+03_dp .AND. Rc <= 1.5414599739111178E+04_dp) THEN
         irange = 3
         L_b = 1.1839524369543471E+03_dp
         U_b = 1.5414599739111178E+04_dp
      END IF
      IF (Rc >= 1.5414599739111178E+04_dp .AND. Rc <= 9.3151526902721278E+04_dp) THEN
         irange = 4
         L_b = 1.5414599739111178E+04_dp
         U_b = 9.3151526902721278E+04_dp
      END IF
      IF (Rc >= 9.3151526902721278E+04_dp .AND. Rc <= 2.7148483276367188E+05_dp) THEN
         irange = 5
         L_b = 9.3151526902721278E+04_dp
         U_b = 2.7148483276367188E+05_dp
      END IF

      fit_coef(1:13, 1:28) = coefdata(1:13, 1:28, irange)

   END SUBROUTINE get_fit_coef_k14

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k15(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 30)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(350), PARAMETER :: c05 = (/-3.0692468631051978E-08_dp, &
         -3.9609383383921665E-01_dp, 4.7587580740675417E-01_dp, -5.5706375318965173E-01_dp, &
         3.4546263714113901E-01_dp, 5.4776062905207357E-01_dp, -2.4141552540426781E+00_dp, &
         4.9577847041515470E+00_dp, -6.8295343742172383E+00_dp, 6.3573367557181761E+00_dp, &
         -3.6226122444655293E+00_dp, 9.5268597153976176E-01_dp, 6.6179227443553872E+00_dp, &
         -7.7387082815093436E-08_dp, -9.9678078330036968E-01_dp, 1.1975533847961941E+00_dp, &
         -1.3762136834418899E+00_dp, 8.0743054822484472E-01_dp, 1.4912431514031330E+00_dp, &
         -6.2314927241780405E+00_dp, 1.2629990199235216E+01_dp, -1.7275639875772516E+01_dp, &
         1.6004550790906951E+01_dp, -9.0878072804552286E+00_dp, 2.3834753802899566E+00_dp, &
         1.2682718868072019E+01_dp, -1.9334328719304757E-07_dp, -2.4848694599881327E+00_dp, &
         2.9853718860501335E+00_dp, -3.3596378831067573E+00_dp, 1.8411087716077625E+00_dp, &
         4.0272695241999301E+00_dp, -1.5957538805213288E+01_dp, 3.1889777329262149E+01_dp, &
         -4.3281478360785940E+01_dp, 3.9883474913142209E+01_dp, -2.2556907487546944E+01_dp, &
         5.8978299479488649E+00_dp, 2.5023174807569418E+01_dp, -4.8642081259672986E-07_dp, &
         -6.2351984490271315E+00_dp, 7.4910846767260342E+00_dp, -8.2248711412084816E+00_dp, &
         4.1238878257772091E+00_dp, 1.0990309451485148E+01_dp, -4.1230623601049615E+01_dp, &
         8.1118417047199273E+01_dp, -1.0912326990788030E+02_dp, 9.9935525762372606E+01_dp, &
         -5.6257102890489350E+01_dp, 1.4655642668172396E+01_dp, 5.1192402268189625E+01_dp, &
         -1.2506631640235665E-06_dp, -1.5980438605202909E+01_dp, 1.9199175975593889E+01_dp, &
         -2.0457978476944227E+01_dp, 9.0671788455391216E+00_dp, 3.0813008462505493E+01_dp, &
         -1.0915584619445536E+02_dp, 2.1098187077206578E+02_dp, -2.8088454943931174E+02_dp, &
         2.5533840000208855E+02_dp, -1.4292670658071711E+02_dp, 3.7067966084892596E+01_dp, &
         1.0941038451297312E+02_dp, -3.3354017068023612E-06_dp, -4.2446927792656162E+01_dp, &
         5.0996396254497533E+01_dp, -5.2338301546774389E+01_dp, 1.9247992459076787E+01_dp, &
         9.0228533336751497E+01_dp, -3.0069997073722988E+02_dp, 5.6936309163352882E+02_dp, &
         -7.4856520686405372E+02_dp, 6.7429591378758153E+02_dp, -3.7476540313467729E+02_dp, &
         9.6644172995788892E+01_dp, 2.4666624220966949E+02_dp, -9.3826543851981905E-06_dp, &
         -1.1877946256653507E+02_dp, 1.4270319798697713E+02_dp, -1.3946465595615874E+02_dp, &
         3.6961694322931955E+01_dp, 2.8120044470115204E+02_dp, -8.7706262371593266E+02_dp, &
         1.6203584176977990E+03_dp, -2.0972847074266356E+03_dp, 1.8671502744648556E+03_dp, &
         -1.0280983406275520E+03_dp, 2.6312195463073755E+02_dp, 5.9473096717460919E+02_dp, &
         -2.8449948791129570E-05_dp, -3.5758635317581138E+02_dp, 4.2960773213012766E+02_dp, &
         -3.9251600529353743E+02_dp, 4.5183929392732445E+01_dp, 9.5587078733574140E+02_dp, &
         -2.7694370693632336E+03_dp, 4.9627218611957405E+03_dp, -6.2929514552940627E+03_dp, &
         5.5135886103861476E+03_dp, -2.9965165662533723E+03_dp, 7.5862722603535246E+02_dp, &
         1.5659660472206674E+03_dp, -9.5986179588260765E-05_dp, -1.1939116417183877E+03_dp, &
         1.4343713151150068E+03_dp, -1.1851399647021360E+03_dp, -1.5231964375206397E+02_dp, &
         3.6741780805293706E+03_dp, -9.7734189198990425E+03_dp, 1.6822530274757435E+04_dp, &
         -2.0720426077965767E+04_dp, 1.7726884301747647E+04_dp, -9.4410952942756994E+03_dp, &
         2.3490775569298803E+03_dp, 4.6638457658101452E+03_dp, -3.8015552197131998E-04_dp, &
         -4.6500281550825130E+03_dp, 5.5865303975222787E+03_dp, -3.8837673064102523E+03_dp, &
         -2.3640043516048258E+03_dp, 1.6965406299396385E+04_dp, -4.0565840455321268E+04_dp, &
         6.5811540329868585E+04_dp, -7.7320579116722583E+04_dp, 6.3441897875273360E+04_dp, &
         -3.2532774480567186E+04_dp, 7.8216440760836831E+03_dp, 1.6825213139968786E+04_dp, &
         -1.9715775615195739E-03_dp, -2.3368672017163331E+04_dp, 2.8074707632439560E+04_dp, &
         -1.3055332264394696E+04_dp, -2.7517846784628873E+04_dp, 1.0619501478660316E+05_dp, &
         -2.1693608692558491E+05_dp, 3.1502086695178028E+05_dp, -3.3334547694157349E+05_dp, &
         2.4553599204317702E+05_dp, -1.1238955180628877E+05_dp, 2.3972326899658266E+04_dp, &
         8.8035762654044956E+04_dp, -1.8243112038117435E-02_dp, -2.0014249841264417E+05_dp, &
         2.4044019203317922E+05_dp, 3.2918527777915639E+04_dp, -5.8564151349304337E+05_dp, &
         1.2278214162309726E+06_dp, -1.5862044513149876E+06_dp, 1.2269856342422077E+06_dp, &
         -1.6000752081890625E+05_dp, -8.0881279434379749E+05_dp, 8.5894001577282173E+05_dp, &
         -3.0531479437217204E+05_dp, 2.2298855158487991E+00_dp, -9.0336389369874778E-09_dp, &
         -1.1705905877090302E-01_dp, 1.4063753695885436E-01_dp, -1.7131704892021460E-01_dp, &
         1.1823577935867260E-01_dp, 1.3211160731762517E-01_dp, -6.7148667896732095E-01_dp, &
         1.4224586147700080E+00_dp, -1.9911772021989478E+00_dp, 1.8732253894330964E+00_dp, &
         -1.0756592628497046E+00_dp, 2.8453654479619789E-01_dp, 2.9930213382361659E+00_dp, &
         -1.8670378462002057E-08_dp, -2.4143656806336988E-01_dp, 2.9006740649273083E-01_dp, &
         -3.4629089408668917E-01_dp, 2.2683680893051139E-01_dp, 3.0404690414859736E-01_dp, &
         -1.4297696355269953E+00_dp, 2.9800638894488669E+00_dp, -4.1371932160927800E+00_dp, &
         3.8711217608603641E+00_dp, -2.2142406528000071E+00_dp, 5.8398891025144894E-01_dp, &
         4.8875160233343715E+00_dp, -4.6204570149850092E-08_dp, -5.9629866259653674E-01_dp, &
         7.1640628120411509E-01_dp, -8.3868041124278481E-01_dp, 5.2019826135940084E-01_dp, &
         8.2463242915711699E-01_dp, -3.6348492405505075E+00_dp, 7.4650304783387114E+00_dp, &
         -1.0283768881675757E+01_dp, 9.5730498855986266E+00_dp, -5.4551689976883058E+00_dp, &
         1.4346476675560549E+00_dp, 8.8414991600481354E+00_dp, -1.1482774693269792E-07_dp, &
         -1.4789938872294259E+00_dp, 1.7768943240759401E+00_dp, -2.0408571717597774E+00_dp, &
         1.1953236620504990E+00_dp, 2.2183761876611556E+00_dp, -9.2555318513362081E+00_dp, &
         1.8752221113980344E+01_dp, -2.5645072882046485E+01_dp, 2.3755380780125787E+01_dp, &
         -1.3487795474893785E+01_dp, 3.5372434275309308E+00_dp, 1.6909293215817328E+01_dp, &
         -2.8397263995252731E-07_dp, -3.6495531995060122E+00_dp, 4.3846461857239438E+00_dp, &
         -4.9313380943000595E+00_dp, 2.6968368321956553E+00_dp, 5.9302864628254115E+00_dp, &
         -2.3462796400928930E+01_dp, 4.6870864451652814E+01_dp, -6.3602080860714842E+01_dp, &
         5.8601551184019506E+01_dp, -3.3140382879672700E+01_dp, 8.6644568567671278E+00_dp, &
         3.3671685943037645E+01_dp, -7.0886445358394887E-07_dp, -9.0866956422509908E+00_dp, &
         1.0916927061893476E+01_dp, -1.1982122131221457E+01_dp, 5.9998216294135123E+00_dp, &
         1.6041977958350834E+01_dp, -6.0136339183264646E+01_dp, 1.1829300280981532E+02_dp, &
         -1.5911935257728072E+02_dp, 1.4571566750741661E+02_dp, -8.2025918687439358E+01_dp, &
         2.1368285429166992E+01_dp, 6.9717468070339038E+01_dp, -1.8115637114197503E-06_dp, &
         -2.3149864597531785E+01_dp, 2.7812649747426072E+01_dp, -2.9648862569096533E+01_dp, &
         1.3165828947892159E+01_dp, 4.4609325588012467E+01_dp, -1.5814634948605772E+02_dp, &
         3.0576431202144215E+02_dp, -4.0715554275312178E+02_dp, 3.7018587829115569E+02_dp, &
         -2.0724045709870501E+02_dp, 5.3753492193265338E+01_dp, 1.5076605862295096E+02_dp, &
         -4.8029447707686627E-06_dp, -6.1142104656738780E+01_dp, 7.3457079767394291E+01_dp, &
         -7.5551348219208521E+01_dp, 2.8115955488796168E+01_dp, 1.2939155360574125E+02_dp, &
         -4.3261107002125891E+02_dp, 8.2012582176643787E+02_dp, -1.0791090432146957E+03_dp, &
         9.7263069656922221E+02_dp, -5.4083603585758897E+02_dp, 1.3952472833308838E+02_dp, &
         3.4326171137332619E+02_dp, -1.3401787740328214E-05_dp, -1.6977952594693747E+02_dp, &
         2.0397539508271359E+02_dp, -2.0046385534576143E+02_dp, 5.5534906744180304E+01_dp, &
         3.9775987919436119E+02_dp, -1.2493663191105957E+03_dp, 2.3147249863985589E+03_dp, &
         -3.0017062754705989E+03_dp, 2.6762457872586815E+03_dp, -1.4753536354877817E+03_dp, &
         3.7795655560340720E+02_dp, 8.3250447697650543E+02_dp, -4.0066178299457473E-05_dp, &
         -5.0433618822359546E+02_dp, 6.0591475756705029E+02_dp, -5.6061550959797250E+02_dp, &
         8.0693121371922274E+01_dp, 1.3220373413009017E+03_dp, -3.8795859841778079E+03_dp, &
         6.9916807710050844E+03_dp, -8.9011916360067389E+03_dp, 7.8237206451432530E+03_dp, &
         -4.2633113757326564E+03_dp, 1.0817488919723778E+03_dp, 2.1880792776836724E+03_dp, &
         -1.3156759027407118E-04_dp, -1.6415118517756680E+03_dp, 1.9721227309225790E+03_dp, &
         -1.6756252916105827E+03_dp, -9.7731609047837892E+01_dp, 4.8848208178157192E+03_dp, &
         -1.3281764469769068E+04_dp, 2.3115603902558436E+04_dp, -2.8709560935485304E+04_dp, &
         2.4734396997303353E+04_dp, -1.3253308640711919E+04_dp, 3.3150401953712826E+03_dp, &
         6.4023792745151923E+03_dp, -4.9316125127201718E-04_dp, -6.0727629604647600E+03_dp, &
         7.2958187858552392E+03_dp, -5.4291483890577892E+03_dp, -2.2232679468127371E+03_dp, &
         2.0941874763722215E+04_dp, -5.2044169864559364E+04_dp, 8.6393786455101945E+04_dp, &
         -1.0345994694934737E+05_dp, 8.6382110934542667E+04_dp, -4.5019043803292414E+04_dp, &
         1.0986821504031903E+04_dp, 2.1847515732361062E+04_dp, -2.2579708925754478E-03_dp, &
         -2.7213243572988889E+04_dp, 3.2693722118754918E+04_dp, -1.9030680015810074E+04_dp, &
         -2.2781523256788889E+04_dp, 1.1219253462313319E+05_dp, -2.4822897274719740E+05_dp, &
         3.8292601430606667E+05_dp, -4.3022911687807390E+05_dp, 3.3804652466094872E+05_dp, &
         -1.6611379681486270E+05_dp, 3.8304153833710385E+04_dp, 9.5573339232628467E+04_dp, &
         -1.4588462469079619E-02_dp, -1.6866513691120152E+05_dp, 2.0262927744864495E+05_dp, &
         -5.7768403952849316E+04_dp, -2.8684582736809686E+05_dp, 8.7522813270953298E+05_dp, &
         -1.6053947767154609E+06_dp, 2.1164271865422688E+06_dp, -2.0010625502698994E+06_dp, &
         1.2728583679706273E+06_dp, -4.7452866201810562E+05_dp, 7.3939317332975275E+04_dp, &
         7.0111074483186123E+05_dp, -2.0104415008030188E-01_dp, -2.1194663306621597E+06_dp, &
         2.5461696440140237E+06_dp, 1.4577959403430095E+06_dp, -8.8776723555507679E+06_dp, &
         1.4429270679797074E+07_dp, -1.0764718275979346E+07_dp, -6.0081024069717359E+06_dp, &
         2.8423455083762139E+07_dp, -3.8351286309245549E+07_dp, 2.6572649225598279E+07_dp, &
         -7.9000536567775914E+06_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.8728308171438685E-01_dp, &
         -7.4958225804522882E-02_dp, -1.7654418697317781E-02_dp, 1.0337296188123402E-02_dp, &
         5.3074367941912289E-03_dp, -3.4737735619981940E-03_dp, -2.0980081422520588E-03_dp, &
         1.3332352270175691E-03_dp, 1.8928638178553657E-03_dp, -2.0371064972525598E-03_dp, &
         4.3322700993716177E-04_dp, 2.4269923006729229E-04_dp, -1.0645391970489572E-04_dp, &
         5.9050213462098911E-01_dp, -2.6045916825257870E-01_dp, -4.6013842770087673E-02_dp, &
         4.0578808574448226E-02_dp, 1.3431038359853705E-02_dp, -1.4448196212394048E-02_dp, &
         -5.0801158293652263E-03_dp, 5.6830395388698567E-03_dp, 6.0399026225143836E-03_dp, &
         -9.0362581333669131E-03_dp, 3.8977272533399529E-03_dp, -2.8631051093057830E-04_dp, &
         -1.6255826015267508E-04_dp, 1.0853661382142048E+00_dp, -5.6505629769884813E-01_dp, &
         -4.5259537292808079E-02_dp, 1.0107558488280066E-01_dp, 1.0246730819933846E-02_dp, &
         -3.7841969991286292E-02_dp, -2.4466922115694272E-03_dp, 1.5238296869265590E-02_dp, &
         1.0051703287389212E-02_dp, -2.4895630648353886E-02_dp, 1.6271959664872358E-02_dp, &
         -4.6274361677158590E-03_dp, 4.2879818988516113E-04_dp, 1.7528845246909897E+00_dp, &
         -1.1123944492275302E+00_dp, 3.5424018217895502E-02_dp, 2.1579693702045874E-01_dp, &
         -2.5857027453275973E-02_dp, -8.1949081571572172E-02_dp, 1.6563668405901510E-02_dp, &
         3.3437407219596017E-02_dp, 9.0789227225405479E-03_dp, -5.4452189093168082E-02_dp, &
         4.6324965691288536E-02_dp, -1.7622865195560243E-02_dp, 2.6635992084654233E-03_dp, &
         2.7078938439530917E+00_dp, -2.1055687575566426E+00_dp, 3.1071744605798479E-01_dp, &
         4.0776524304565614E-01_dp, -1.4276623438669639E-01_dp, -1.5004288372475155E-01_dp, &
         7.4789568797600153E-02_dp, 6.1204081755821510E-02_dp, -1.1450125801862587E-02_dp, &
         -9.5333496012500740E-02_dp, 1.0315290568119541E-01_dp, -4.6301294653322549E-02_dp, &
         8.2327674140336218E-03_dp, 4.1222585977458319E+00_dp, -3.8916375985052856E+00_dp, &
         1.0287256818377424E+00_dp, 6.7824953475735794E-01_dp, -4.3616962492030631E-01_dp, &
         -2.2442617116461031E-01_dp, 2.1410714787692231E-01_dp, 8.9232609313346464E-02_dp, &
         -8.4070976342451670E-02_dp, -1.2224787823475017E-01_dp, 1.8617969237834020E-01_dp, &
         -9.6992308351370335E-02_dp, 1.9240741102009955E-02_dp, 6.2614437064378148E+00_dp, &
         -7.0734996616320798E+00_dp, 2.7109877698494551E+00_dp, 9.4315201130905646E-01_dp, &
         -1.0794537771539827E+00_dp, -2.2631895031272395E-01_dp, 4.9923803036845726E-01_dp, &
         7.9249948225847855E-02_dp, -2.6692610962273050E-01_dp, -5.7380279775207862E-02_dp, &
         2.6060510417597316E-01_dp, -1.6580437343809790E-01_dp, 3.6703525275681770E-02_dp, &
         9.5442495162132879E+00_dp, -1.2707406352368116E+01_dp, 6.4338288503783216E+00_dp, &
         8.6207574579597068E-01_dp, -2.3483916970462482E+00_dp, 7.3668001058521435E-02_dp, &
         1.0019608211825608E+00_dp, -7.4722380119410070E-02_dp, -6.3758685173908902E-01_dp, &
         2.7271888085324325E-01_dp, 2.1084047383313775E-01_dp, -2.1733444967013762E-01_dp, &
         5.6641676594043365E-02_dp, 1.4643304263035175E+01_dp, -2.2655257244471311E+01_dp, &
         1.4392217701024219E+01_dp, -5.9435672642207360E-01_dp, -4.5777070581320389E+00_dp, &
         1.2351960033513358E+00_dp, 1.7170546319191513E+00_dp, -6.1224168013341784E-01_dp, &
         -1.2399540344239814E+00_dp, 1.1799686690987607E+00_dp, -2.3435024824910564E-01_dp, &
         -1.4444306875541876E-01_dp, 6.1909340705238292E-02_dp, 2.2661556393863005E+01_dp, &
         -4.0221154146253184E+01_dp, 3.1024585967572101E+01_dp, -6.1984371213472800E+00_dp, &
         -7.8268184197828807E+00_dp, 4.4496347113412762E+00_dp, 2.2809196015538231E+00_dp, &
         -1.9772352681903975E+00_dp, -1.9037690088995247E+00_dp, 3.0942613517855539E+00_dp, &
         -1.5987186090989016E+00_dp, 3.0213023980747233E-01_dp, 5.8046939919012327E-03_dp, &
         3.5468093717051559E+01_dp, -7.1336923975099211E+01_dp, 6.5243630185390288E+01_dp, &
         -2.2937295080112825E+01_dp, -1.0565705083043978E+01_dp, 1.1806144892783585E+01_dp, &
         1.2178868152228470E+00_dp, -4.7251305539334565E+00_dp, -1.7396453277841573E+00_dp, &
         6.1545408517662787E+00_dp, -4.5896192312693147E+00_dp, 1.5674169440757093E+00_dp, &
         -2.0955667192577354E-01_dp, 5.6442005976612165E+01_dp, -1.2697432446862335E+02_dp, &
         1.3491229212895021E+02_dp, -6.7527741358861761E+01_dp, -5.5728868452042661E+00_dp, &
         2.5434552707557021E+01_dp, -5.4812962163209651E+00_dp, -8.5501736428865840E+00_dp, &
         1.6761122225991096E+00_dp, 8.5806543180294934E+00_dp, -8.9923594919231888E+00_dp, &
         3.9271082076494404E+00_dp, -6.8512753005700844E-01_dp, 9.2592250966591095E+01_dp, &
         -2.2931021529690631E+02_dp, 2.7694141294096011E+02_dp, -1.7833095335890167E+02_dp, &
         3.0475551450022131E+01_dp, 4.2007083036843376E+01_dp, -2.4692838382505418E+01_dp, &
         -9.2334955726867634E+00_dp, 1.0827009524998724E+01_dp, 5.3897134051317694E+00_dp, &
         -1.1613244577817873E+01_dp, 6.3881771027172425E+00_dp, -1.2985721875418899E+00_dp, &
         1.6317349757925933E+02_dp, -4.3567621344686893E+02_dp, 5.8173974531331419E+02_dp, &
         -4.4774963325360090E+02_dp, 1.5972672378438833E+02_dp, 3.9090311776792390E+01_dp, &
         -5.9596575929838380E+01_dp, 1.7921854401952180E+00_dp, 2.2619550601715890E+01_dp, &
         -5.3934015665785919E+00_dp, -9.7558137801544511E+00_dp, 7.6935514963494267E+00_dp, &
         -1.8168321790144826E+00_dp, 3.6167981936470210E+02_dp, -1.0134714814220670E+03_dp, &
         1.4434041436150949E+03_dp, -1.2377699820298194E+03_dp, 5.8888285175980138E+02_dp, &
         -4.1630388543112794E+01_dp, -1.1428599149682097E+02_dp, 2.8274776131364867E+01_dp, &
         4.1140189506559487E+01_dp, -2.4185930418501993E+01_dp, -7.4527591947965899E+00_dp, &
         1.0860057531580511E+01_dp, -2.9572025136611222E+00_dp, 1.5169954011304849E+00_dp, &
         -6.2259099720483668E-01_dp, -1.3712328672328261E-01_dp, 8.8962828652861925E-02_dp, &
         4.1043676954109272E-02_dp, -3.0443932125223352E-02_dp, -1.6102573780172595E-02_dp, &
         1.1773644941326244E-02_dp, 1.5459317774057306E-02_dp, -1.8230644358955633E-02_dp, &
         5.1169302495753487E-03_dp, 1.3183518368033997E-03_dp, -7.5903468555029124E-04_dp, &
         1.7499525908457525E+00_dp, -9.1531468832248197E-01_dp, -7.6236631412840317E-02_dp, &
         1.6660751635474580E-01_dp, 1.8633663668780786E-02_dp, -6.2945543301738627E-02_dp, &
         -5.1025796217109790E-03_dp, 2.5374090389752808E-02_dp, 1.7578523782323396E-02_dp, &
         -4.1763776533811942E-02_dp, 2.6693483469204270E-02_dp, -7.3228880092403711E-03_dp, &
         6.1452305084940938E-04_dp, 2.2619483691280533E+00_dp, -1.6013976329479644E+00_dp, &
         1.1397591566142432E-01_dp, 3.3354035175947067E-01_dp, -5.7655973122046261E-02_dp, &
         -1.3000506221511882E-01_dp, 3.2967403963862302E-02_dp, 5.3494998718352296E-02_dp, &
         1.1629723916519787E-02_dp, -8.8697135876886060E-02_dp, 7.8248427005448054E-02_dp, &
         -3.0608523079921474E-02_dp, 4.7649824200322270E-03_dp, 3.1532640805277889E+00_dp, &
         -2.9086028536368387E+00_dp, 6.0362531752264503E-01_dp, 6.0074812985031767E-01_dp, &
         -2.6167841326003777E-01_dp, -2.2598188022853924E-01_dp, 1.3384536380554243E-01_dp, &
         9.3175482448094510E-02_dp, -2.8552228684050263E-02_dp, -1.4717381630923509E-01_dp, &
         1.6884009102232561E-01_dp, -7.8198141800493393E-02_dp, 1.4252294977810340E-02_dp, &
         4.5990362784635748E+00_dp, -5.2610480580945511E+00_dp, 1.7554142769832759E+00_dp, &
         9.3954313197653416E-01_dp, -7.3530977939486497E-01_dp, -3.0941341340797157E-01_dp, &
         3.5949602821775650E-01_dp, 1.2391323817308118E-01_dp, -1.5674994650740204E-01_dp, &
         -1.6497906356785436E-01_dp, 2.8695923379282917E-01_dp, -1.5603347654001742E-01_dp, &
         3.1792664959924305E-02_dp, 6.8890308910671356E+00_dp, -9.4168852534476279E+00_dp, &
         4.3188327286694310E+00_dp, 1.1783565195132464E+00_dp, -1.7254059330117610E+00_dp, &
         -2.3869963236274130E-01_dp, 7.9885232236465131E-01_dp, 7.7283423048285574E-02_dp, &
         -4.6310391349780267E-01_dp, -1.4416187514689229E-03_dp, 3.5710470924999077E-01_dp, &
         -2.4826140681908590E-01_dp, 5.7194043278611120E-02_dp, 1.0497250232124662E+01_dp, &
         -1.6719832965151884E+01_dp, 9.8216896813060970E+00_dp, 7.3970611364224259E-01_dp, &
         -3.6059125166637358E+00_dp, 3.6498570473675707E-01_dp, 1.5345652988820813E+00_dp, &
         -2.2332312617279992E-01_dp, -1.0543504968226272E+00_dp, 6.3278478456932230E-01_dp, &
         1.6617016052927272E-01_dp, -2.8118812153705103E-01_dp, 8.0364317707803778E-02_dp, &
         1.6202514503988390E+01_dp, -2.9552478542397893E+01_dp, 2.1354172370954117E+01_dp, &
         -2.0112687983411681E+00_dp, -6.7855159638167732E+00_dp, 2.3873633162830670E+00_dp, &
         2.4998632165545791E+00_dp, -1.1623801758158443E+00_dp, -1.9504681292485571E+00_dp, &
         2.2152489068540460E+00_dp, -7.3792333984291780E-01_dp, -6.4276694004470475E-02_dp, &
         6.8740941716736517E-02_dp, 2.5304655431445173E+01_dp, -5.2159151005688130E+01_dp, &
         4.5132654192248751E+01_dp, -1.1338185998416552E+01_dp, -1.1160559135910367E+01_dp, &
         7.6739991523479407E+00_dp, 3.0247629570482868E+00_dp, -3.4345182182881113E+00_dp, &
         -2.7853008863066906E+00_dp, 5.3467839342257841E+00_dp, -3.1694123271238630E+00_dp, &
         8.0694157345556727E-01_dp, -5.5813093165602519E-02_dp, 4.0044203390039350E+01_dp, &
         -9.2178078631682354E+01_dp, 9.3592414174368287E+01_dp, -3.7812448013221193E+01_dp, &
         -1.4050944707063607E+01_dp, 1.9459317299329200E+01_dp, 6.0508974835963314E-01_dp, &
         -7.9026281849074591E+00_dp, -2.0036084707040196E+00_dp, 1.0164747665646736E+01_dp, &
         -8.2753537306586118E+00_dp, 3.0766573481774491E+00_dp, -4.5919018524870359E-01_dp, &
         6.4548216002732048E+01_dp, -1.6372767342710813E+02_dp, 1.9149111780202836E+02_dp, &
         -1.0651796580505838E+02_dp, -3.6942339641588529E+00_dp, 4.1116703654714840E+01_dp, &
         -1.1688259826259763E+01_dp, -1.4092944078806669E+01_dp, 4.9534063773839927E+00_dp, &
         1.3327217036380015E+01_dp, -1.5720082956234478E+01_dp, 7.3285690650141238E+00_dp, &
         -1.3498478899609199E+00_dp, 1.0751974782069394E+02_dp, -2.9506101073517016E+02_dp, &
         3.8886438902426795E+02_dp, -2.7414973759844793E+02_dp, 5.7813607870059755E+01_dp, &
         6.6686184229942938E+01_dp, -4.6907689236635093E+01_dp, -1.3942873569870745E+01_dp, &
         2.4627269489905036E+01_dp, 2.2908131761383355E+00_dp, -1.7205170554994950E+01_dp, &
         1.0876408321603687E+01_dp, -2.3799479399999801E+00_dp, 1.9247091864637630E+02_dp, &
         -5.5613733430051036E+02_dp, 7.9963641497991784E+02_dp, -6.6716934500841262E+02_dp, &
         2.6869114067386937E+02_dp, 5.1976162118599937E+01_dp, -1.0977479209219821E+02_dp, &
         1.5320418459417469E+01_dp, 4.6056433428995703E+01_dp, -2.9674185020400326E+01_dp, &
         -1.1567576408411366E+00_dp, 7.5442392510615788E+00_dp, -2.2833978807688808E+00_dp, &
         4.1371492246319275E+02_dp, -1.2139560601068151E+03_dp, 1.8159618918225744E+03_dp, &
         -1.6547693041000880E+03_dp, 8.6670487570533533E+02_dp, -1.1144406945065859E+02_dp, &
         -1.6484069878511906E+02_dp, 7.3842511245161845E+01_dp, 4.7114364272953480E+01_dp, &
         -5.3558177553213220E+01_dp, 1.2533328664185559E+01_dp, 4.7084386636147091E+00_dp, &
         -2.2343883375336238E+00_dp, 1.5482456811918644E+03_dp, -4.4722809407060004E+03_dp, &
         6.6239888100139933E+03_dp, -6.0381436371333639E+03_dp, 3.2764369520591445E+03_dp, &
         -6.5019366566941835E+02_dp, -3.5765412817763291E+02_dp, 1.5908326595301082E+02_dp, &
         1.4549695520260664E+02_dp, -1.2481558144644566E+02_dp, 1.0775656086159311E+00_dp, &
         3.0587066325522777E+01_dp, -9.7246696807779784E+00_dp, 2.3046596783109666E-01_dp, &
         -5.0266863054748409E-02_dp, -1.7179737588739360E-02_dp, 5.4654515693096711E-03_dp, &
         4.7852361622912525E-03_dp, -1.7421101769864551E-03_dp, -1.9440310548800761E-03_dp, &
         6.0661250198343168E-04_dp, 1.3341594386686755E-03_dp, -8.7661120974839130E-04_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/-2.3785652584517211E-04_dp, &
         4.0415260781371566E-04_dp, -1.1566827381466792E-04_dp, 7.4614843698202982E-01_dp, &
         -1.8831297390879426E-01_dp, -5.5307781300718430E-02_dp, 2.5755321163507486E-02_dp, &
         1.6130804272204945E-02_dp, -8.8949071784985307E-03_dp, -6.6594057854666763E-03_dp, &
         3.2468867718668313E-03_dp, 5.3068142089810330E-03_dp, -4.9387141407180018E-03_dp, &
         4.6635469014696219E-04_dp, 1.0225379001983694E-03_dp, -3.6432448373067977E-04_dp, &
         1.4438155770093133E+00_dp, -4.6034617711746817E-01_dp, -1.0114569577453580E-01_dp, &
         8.0847736706817566E-02_dp, 3.1132826193078431E-02_dp, -3.0098300767058127E-02_dp, &
         -1.2990120467489992E-02_dp, 1.1474677064022006E-02_dp, 1.3524876701271211E-02_dp, &
         -1.8252222127657793E-02_dp, 6.7665303400730129E-03_dp, 2.6588974984116941E-04_dp, &
         -5.3997537121426489E-04_dp, 2.5092558836526182E+00_dp, -1.0370895440920307E+00_dp, &
         -1.4293476604484445E-01_dp, 2.1892543997539182E-01_dp, 4.3916720461941325E-02_dp, &
         -8.5906691296838311E-02_dp, -1.7581973163101349E-02_dp, 3.3856445186467174E-02_dp, &
         2.9188840795829587E-02_dp, -5.5487833762756317E-02_dp, 3.0644984612438222E-02_dp, &
         -6.2518601968366934E-03_dp, -1.3619319925894898E-05_dp, 4.2480175628654671E+00_dp, &
         -2.2552731443674090E+00_dp, -1.2020190507153349E-01_dp, 5.3601909417709703E-01_dp, &
         2.3123136649126307E-02_dp, -2.1742428493087468E-01_dp, -3.1971501625727987E-03_dp, &
         8.7966369542184522E-02_dp, 5.2098492049143776E-02_dp, -1.4627449223179681E-01_dp, &
         1.0200338814556090E-01_dp, -3.1449802576644731E-02_dp, 3.4350745589740235E-03_dp, &
         7.1934846151788383E+00_dp, -4.8045602148293982E+00_dp, 1.6800723431165815E-01_dp, &
         1.2178133966689220E+00_dp, -1.3794075755207405E-01_dp, -4.9998735924019683E-01_dp, &
         8.8402497336550506E-02_dp, 2.0630420687207099E-01_dp, 6.3021419984179414E-02_dp, &
         -3.4072681253649273E-01_dp, 2.8686006501533551E-01_dp, -1.0813824760173801E-01_dp, &
         1.6124430353177774E-02_dp, 1.2310857286275628E+01_dp, -1.0127157366490705E+01_dp, &
         1.3116248102932406E+00_dp, 2.5981695903237028E+00_dp, -7.5308905349267552E-01_dp, &
         -1.0473666363263530E+00_dp, 4.2542510815825957E-01_dp, 4.3632942640926314E-01_dp, &
         -2.7104988102069912E-02_dp, -6.9029606232100948E-01_dp, 7.0742413257606307E-01_dp, &
         -3.0835516320907685E-01_dp, 5.3474117194919846E-02_dp, 2.1395537244919190E+01_dp, &
         -2.1294952582127532E+01_dp, 4.9586759348241420E+00_dp, 5.1888176836291438E+00_dp, &
         -2.6758759362963236E+00_dp, -1.9431216985141384E+00_dp, 1.4461702367488496E+00_dp, &
         7.9948224935758594E-01_dp, -5.1733385566911505E-01_dp, -1.1192375715667364E+00_dp, &
         1.5165197860827433E+00_dp, -7.5951327716264216E-01_dp, 1.4699498210546361E-01_dp, &
         3.7874378642397900E+01_dp, -4.4985821130051526E+01_dp, 1.5631115834221772E+01_dp, &
         9.4321652649337349E+00_dp, -8.0952721883237579E+00_dp, -2.8567401038783609E+00_dp, &
         4.1805740387217281E+00_dp, 1.0764667230663318E+00_dp, -2.2360065629356378E+00_dp, &
         -9.4199681401382451E-01_dp, 2.6381666350602235E+00_dp, -1.5885150294156811E+00_dp, &
         3.4308353829740224E-01_dp, 6.8468294337154987E+01_dp, -9.6066459756805102E+01_dp, &
         4.5738898453688122E+01_dp, 1.4111679011798346E+01_dp, -2.2267791999698446E+01_dp, &
         -1.5136616324510050E+00_dp, 1.0688893528377216E+01_dp, -1.3084462945928754E-02_dp, &
         -7.0303230583103691E+00_dp, 2.5138353770466910E+00_dp, 2.7198542807232915E+00_dp, &
         -2.5443022123523957E+00_dp, 6.4822382263330280E-01_dp, 1.2677231414292079E+02_dp, &
         -2.0856628210561396E+02_dp, 1.2977414621627011E+02_dp, 9.1386695755711109E+00_dp, &
         -5.6431954234367460E+01_dp, 1.1601893649962131E+01_dp, 2.3601434993236484E+01_dp, &
         -7.4033248447262663E+00_dp, -1.7502732203666131E+01_dp, 1.6822593726600431E+01_dp, &
         -3.9321362151268588E+00_dp, -1.6129758489863975E+00_dp, 7.7708342864698754E-01_dp, &
         2.4137930903364659E+02_dp, -4.6294363698594930E+02_dp, 3.6567746448701314E+02_dp, &
         -5.4264739702960071E+01_dp, -1.2743717778047937E+02_dp, 6.9426575766974068E+01_dp, &
         3.8441422516024915E+01_dp, -3.5937690433228575E+01_dp, -3.0556184645453925E+01_dp, &
         5.6709057679208890E+01_dp, -3.3613907949953663E+01_dp, 8.7069360708662646E+00_dp, &
         -6.5071547786799400E-01_dp, 4.7630568634676882E+02_dp, -1.0576120213436998E+03_dp, &
         1.0387559056977509E+03_dp, -3.8252814075560741E+02_dp, -2.1458148531810613E+02_dp, &
         2.6341292507433673E+02_dp, -4.9256154892799719E-01_dp, -1.1226068848553658E+02_dp, &
         -2.8240527873382217E+00_dp, 1.1790110895706198E+02_dp, -1.0801679538988816E+02_dp, &
         4.3992074063654670E+01_dp, -7.2605761932187214E+00_dp, 9.9831523554887349E+02_dp, &
         -2.5291408915064985E+03_dp, 3.0129078291719397E+03_dp, -1.7907805874724606E+03_dp, &
         5.2655917157171132E+01_dp, 6.9446212730128070E+02_dp, -3.3406235416814116E+02_dp, &
         -1.6432224599188541E+02_dp, 1.8182149839084897E+02_dp, 5.5649378552951724E+01_dp, &
         -1.5404182921741005E+02_dp, 8.7668748811935586E+01_dp, -1.8136282582373241E+01_dp, &
         2.5197956634819393E+03_dp, -7.0358606845934346E+03_dp, 9.6643808587164713E+03_dp, &
         -7.6367755567523545E+03_dp, 2.8044634842378887E+03_dp, 7.1741892008980801E+02_dp, &
         -1.1890761393357277E+03_dp, 1.6354332870871900E+02_dp, 3.8067937600501506E+02_dp, &
         -1.6046444938930748E+02_dp, -1.0923720079478032E+02_dp, 1.0988436092388882E+02_dp, &
         -2.7749973961794815E+01_dp, 1.8790883577505788E+00_dp, -4.2578976954083553E-01_dp, &
         -1.4015384201338235E-01_dp, 4.9535242205172317E-02_dp, 3.9556319497385098E-02_dp, &
         -1.6188345328963884E-02_dp, -1.6150186727266109E-02_dp, 5.7212758934123784E-03_dp, &
         1.1515686373747173E-02_dp, -8.4039177187524849E-03_dp, -1.2421200032158236E-03_dp, &
         3.1542038254185724E-03_dp, -9.4397800109121256E-04_dp, 2.3286072431387423E+00_dp, &
         -7.3963153685831751E-01_dp, -1.6817064802096532E-01_dp, 1.2874106425262241E-01_dp, &
         5.2793112327834887E-02_dp, -4.7503985299264515E-02_dp, -2.2274548170645277E-02_dp, &
         1.7979044108164671E-02_dp, 2.2209117750315765E-02_dp, -2.8376437312440522E-02_dp, &
         9.4867238052117418E-03_dp, 1.1424442652292026E-03_dp, -1.0083455255258623E-03_dp, &
         3.3727466736396918E+00_dp, -1.5448224549513210E+00_dp, -1.9258486598122063E-01_dp, &
         3.4252847181451407E-01_dp, 6.4318683841940644E-02_dp, -1.3514789671089780E-01_dp, &
         -2.6692286257403969E-02_dp, 5.3252686100318150E-02_dp, 4.5847091590627692E-02_dp, &
         -8.7415775234976054E-02_dp, 4.8250758426842198E-02_dp, -9.7951517189096889E-03_dp, &
         -4.0080346635974667E-05_dp, 5.3476974926650058E+00_dp, -3.2821481672447801E+00_dp, &
         -1.0673693072908484E-01_dp, 8.2336777253315929E-01_dp, 1.9953202180531222E-02_dp, &
         -3.3773322054580296E-01_dp, -1.4581831738714061E-04_dp, 1.3707953425659358E-01_dp, &
         8.0179761639936148E-02_dp, -2.2969150949183195E-01_dp, 1.6148118046840720E-01_dp, &
         -5.0234143578115772E-02_dp, 5.5756585628377389E-03_dp, 8.9052398487289590E+00_dp, &
         -6.8986923960309667E+00_dp, 4.0813160978388663E-01_dp, 1.8330154014443996E+00_dp, &
         -2.4965908117099087E-01_dp, -7.6390746125752773E-01_dp, 1.4990729029415834E-01_dp, &
         3.1721789359789093E-01_dp, 9.3176773742444513E-02_dp, -5.3063124869930844E-01_dp, &
         4.5174324220642409E-01_dp, -1.7188470254374158E-01_dp, 2.5897541226569322E-02_dp, &
         1.5277197043057445E+01_dp, -1.4403808006208923E+01_dp, 2.2520306793354425E+00_dp, &
         3.8402789683619147E+00_dp, -1.2282086708774107E+00_dp, -1.5752333655294370E+00_dp, &
         6.8512948891006886E-01_dp, 6.6286709270779132E-01_dp, -5.6777181683095333E-02_dp, &
         -1.0658001277397429E+00_dp, 1.1088446843283359E+00_dp, -4.8809164974479147E-01_dp, &
         8.5368915942929544E-02_dp, 2.6809203330113430E+01_dp, -3.0105648181028421E+01_dp, &
         7.8907784967325929E+00_dp, 7.5527940324183627E+00_dp, -4.2136321310777394E+00_dp, &
         -2.8787838332303517E+00_dp, 2.2822315773743074E+00_dp, 1.2025374446858241E+00_dp, &
         -8.4310251040346118E-01_dp, -1.7073511450150149E+00_dp, 2.3701711370977949E+00_dp, &
         -1.2003354391249612E+00_dp, 2.3417510086396071E-01_dp, 4.8050803348005907E+01_dp, &
         -6.3414162259620326E+01_dp, 2.4012810181586435E+01_dp, 1.3543281743456987E+01_dp, &
         -1.2518399110121845E+01_dp, -4.1483228966192884E+00_dp, 6.5359361416230444E+00_dp, &
         1.6010875650124006E+00_dp, -3.5935349636567349E+00_dp, -1.3505908589460529E+00_dp, &
         4.1102852487568118E+00_dp, -2.5128140904440519E+00_dp, 5.4740471934507018E-01_dp, &
         8.8043083290018245E+01_dp, -1.3537232179695650E+02_dp, 6.8833207015861163E+01_dp, &
         1.9954488880701138E+01_dp, -3.4131409980955162E+01_dp, -1.9360725759409532E+00_dp, &
         1.6704761193414530E+01_dp, -8.8596046107263363E-02_dp, -1.1369840543223333E+01_dp, &
         4.3533647172741814E+00_dp, 4.1462824454929388E+00_dp, -4.0304785375198096E+00_dp, &
         1.0392352576178361E+00_dp, 1.6528695869768484E+02_dp, -2.9435766562155197E+02_dp, &
         1.9269303927992763E+02_dp, 1.2272006016669405E+01_dp, -8.6558626633572814E+01_dp, &
         1.8084227124270082E+01_dp, 3.7422286676621773E+01_dp, -1.1504422400508385E+01_dp, &
         -2.8981028446875509E+01_dp, 2.8061685767204931E+01_dp, -6.8454282426917770E+00_dp, &
         -2.5075498308103796E+00_dp, 1.2590902098959449E+00_dp, 3.1904415962372269E+02_dp, &
         -6.5517407839010923E+02_dp, 5.3795044575370639E+02_dp, -7.8460961642246900E+01_dp, &
         -1.9941242952975134E+02_dp, 1.0626216155817104E+02_dp, 6.5024974663538401E+01_dp, &
         -5.6853636598935005E+01_dp, -5.4516670856876033E+01_dp, 9.6886573999137951E+01_dp, &
         -5.6786140068674428E+01_dp, 1.4494723569712843E+01_dp, -1.0273091986637437E+00_dp, &
         6.3717770820107091E+02_dp, -1.5006743273417519E+03_dp, 1.5183508248241242E+03_dp, &
         -5.4249059143678107E+02_dp, -3.6871786881056681E+02_dp, 4.1511942398217616E+02_dp, &
         2.6225823385079689E+01_dp, -1.9254671249051862E+02_dp, -2.7660102820898491E+01_dp, &
         2.2884670976955013E+02_dp, -2.0078072107272666E+02_dp, 8.0102960484258318E+01_dp, &
         -1.2994587476880236E+01_dp, 1.3388641449458219E+03_dp, -3.5707279333293045E+03_dp, &
         4.3696883768952457E+03_dp, -2.5408264232775923E+03_dp, -1.5720657008223867E+02_dp, &
         1.2433964069705148E+03_dp, -4.9064532646415466E+02_dp, -4.0438252913184124E+02_dp, &
         3.5221285081678133E+02_dp, 1.6421609423683356E+02_dp, -3.5828174338405978E+02_dp, &
         1.9958566836052518E+02_dp, -4.1225182836963377E+01_dp, 3.1630578325136571E+03_dp, &
         -9.2174296603112798E+03_dp, 1.3078767927116834E+04_dp, -1.0455636907823062E+04_dp, &
         3.4380325323762208E+03_dp, 1.9056619449914899E+03_dp, -2.4483725222720182E+03_dp, &
         3.3598576622327710E+02_dp, 1.0045407494218127E+03_dp, -7.4953512509145685E+02_dp, &
         1.1389182240118748E+02_dp, 8.8631597307204402E+01_dp, -3.3784175186635018E+01_dp, &
         1.1741583647618587E+04_dp, -3.4772802927998528E+04_dp, 5.1757382963840319E+04_dp, &
         -4.6796698845286672E+04_dp, 2.4404793826910518E+04_dp, -3.4240409533000625E+03_dp, &
         -4.1525449932535876E+03_dp, 1.9174273170875924E+03_dp, 9.6404934848061453E+02_dp, &
         -9.6135633177130569E+02_dp, -6.2667702498202678E+00_dp, 2.5568713212589310E+02_dp, &
         -8.0871228602797459E+01_dp, 2.5910399582486437E-01_dp, -2.8215937471986335E-02_dp, &
         -1.6057948736797313E-02_dp, 2.2232618509844711E-03_dp, 4.4103499063233055E-03_dp, &
         -1.6453858095913664E-04_dp, -2.6110557878466982E-03_dp, 3.9824644955139122E-04_dp, &
         1.6273475095842586E-03_dp, -8.9025885910407965E-04_dp, -3.6671922956720823E-04_dp, &
         4.8507440460371750E-04_dp, -1.3211072877791544E-04_dp, 8.5654480285554591E-01_dp, &
         -1.1200920153309295E-01_dp, -6.0375651330303620E-02_dp, 1.2480943363126322E-02_dp, &
         1.7766988818178821E-02_dp, -2.1391630741560913E-03_dp, -1.0741465894792950E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/2.2061351968797813E-03_dp, &
         7.2651800654771788E-03_dp, -4.9666091131882828E-03_dp, -6.0509045722459322E-04_dp, &
         1.7074706793122701E-03_dp, -5.0886032182887841E-04_dp, 1.7258143200373688E+00_dp, &
         -2.9884323352649256E-01_dp, -1.4775030321223576E-01_dp, 4.7248271588260116E-02_dp, &
         4.7619516819167255E-02_dp, -1.1816286083326438E-02_dp, -2.9752710649639229E-02_dp, &
         8.5535969429659311E-03_dp, 2.2544415185358772E-02_dp, -1.9478392039671260E-02_dp, &
         2.3867184506599064E-03_dp, 3.3938096349746282E-03_dp, -1.2408011528001253E-03_dp, &
         3.1765354659041769E+00_dp, -7.4113425888104711E-01_dp, -3.3028597049657266E-01_dp, &
         1.5279087435777911E-01_dp, 1.1534912548347100E-01_dp, -4.6610003150119948E-02_dp, &
         -7.4975530724277845E-02_dp, 2.9204637910627697E-02_dp, 6.4144154361418626E-02_dp, &
         -6.7383151583557066E-02_dp, 1.9495127476223385E-02_dp, 3.9216434341502104E-03_dp, &
         -2.5026214876007966E-03_dp, 5.7762677335774812E+00_dp, -1.7814057289603056E+00_dp, &
         -7.0223258614858830E-01_dp, 4.5053468321128493E-01_dp, 2.6008563697492804E-01_dp, &
         -1.5727265430895401E-01_dp, -1.7715192567226051E-01_dp, 9.2605163231566937E-02_dp, &
         1.7280977666896552E-01_dp, -2.1622833321996968E-01_dp, 8.9907102665867045E-02_dp, &
         -6.5342786988899322E-03_dp, -3.6405862618677188E-03_dp, 1.0635866709342723E+01_dp, &
         -4.2221910963666494E+00_dp, -1.4280030736515645E+00_dp, 1.2620175013991994E+00_dp, &
         5.4430896273848306E-01_dp, -4.8950416510612965E-01_dp, -3.9333543661332560E-01_dp, &
         2.8233335910502383E-01_dp, 4.4606392158013780E-01_dp, -6.6340801329142474E-01_dp, &
         3.4698053836989273E-01_dp, -7.0117046739817404E-02_dp, 3.8524475772982433E-04_dp, &
         2.0028892471519779E+01_dp, -1.0003628886391914E+01_dp, -2.7543527547087376E+00_dp, &
         3.4459409293308343E+00_dp, 1.0227949830416387E+00_dp, -1.4575186303587055E+00_dp, &
         -8.0548916230565581E-01_dp, 8.4662036227712156E-01_dp, 1.1006704135061391E+00_dp, &
         -1.9762595229695696E+00_dp, 1.2293924485257903E+00_dp, -3.4776573573238989E-01_dp, &
         3.3411960379456256E-02_dp, 3.8800864934511303E+01_dp, -2.3979969508026311E+01_dp, &
         -4.8691713652289721E+00_dp, 9.3268439917937549E+00_dp, 1.5091857344004103E+00_dp, &
         -4.2309476977858385E+00_dp, -1.4145743330439313E+00_dp, 2.5351120694533700E+00_dp, &
         2.5243975989462042E+00_dp, -5.7315045646482972E+00_dp, 4.1375154729827681E+00_dp, &
         -1.4175776019905513E+00_dp, 1.9259934383182015E-01_dp, 7.7735814299357841E+01_dp, &
         -5.8814947029987437E+01_dp, -6.9572678017406089E+00_dp, 2.5308616104245754E+01_dp, &
         3.9853220364063036E-01_dp, -1.2070957719421918E+01_dp, -1.4815249980025791E+00_dp, &
         7.6544448352112955E+00_dp, 4.8490644610502533E+00_dp, -1.5967127077365880E+01_dp, &
         1.3341122150477743E+01_dp, -5.2520565274679489E+00_dp, 8.4320063077592255E-01_dp, &
         1.6206206464295840E+02_dp, -1.4930321627961305E+02_dp, -2.5067856011273091E+00_dp, &
         6.9341075313618006E+01_dp, -1.0743440903075324E+01_dp, -3.3716158577793486E+01_dp, &
         3.7254112287325314E+00_dp, 2.3363694407054414E+01_dp, 3.9659586871550068E+00_dp, &
         -4.0721927683672071E+01_dp, 4.0561860559316933E+01_dp, -1.8073377097484137E+01_dp, &
         3.2530077490973341E+00_dp, 3.5442984306993384E+02_dp, -3.9746210109659887E+02_dp, &
         4.2890577383048353E+01_dp, 1.9202721291279789E+02_dp, -7.1796204519006139E+01_dp, &
         -8.9522197012670105E+01_dp, 3.7812207284886419E+01_dp, 7.1003674952178457E+01_dp, &
         -3.3126726641687128E+01_dp, -7.9733890323907957E+01_dp, 1.0865212885707727E+02_dp, &
         -5.6048850673878945E+01_dp, 1.1208474159757820E+01_dp, 8.2232427943744574E+02_dp, &
         -1.1282193717418054E+03_dp, 3.0056386653438210E+02_dp, 5.2970835948779984E+02_dp, &
         -3.6912742804955758E+02_dp, -1.9945931211622650E+02_dp, 2.0796916315615783E+02_dp, &
         1.9918225114845831E+02_dp, -2.9472016065289944E+02_dp, 7.2671923695521441E+00_dp, &
         1.9096380738566870E+02_dp, -1.3490590492082302E+02_dp, 3.1463294283517353E+01_dp, &
         2.0580837910555242E+03_dp, -3.4946061703309515E+03_dp, 1.6696299583911045E+03_dp, &
         1.3450505077893019E+03_dp, -1.7913069927563238E+03_dp, -8.6372308442249675E+01_dp, &
         9.4017698393274759E+02_dp, 3.3494319705863421E+02_dp, -1.5979707059674208E+03_dp, &
         1.3250952509876529E+03_dp, -3.9493447453139942E+02_dp, -4.2074272991575519E+01_dp, &
         3.7749026605926751E+01_dp, 5.7172884504696012E+03_dp, -1.2228637615334226E+04_dp, &
         9.5260345584609622E+03_dp, 1.5243163670468698E+03_dp, -8.3375278186301748E+03_dp, &
         4.2132976943189569E+03_dp, 2.6329863532489203E+03_dp, -1.9930161570578837E+03_dp, &
         -4.4234144073523730E+03_dp, 7.9728267713326386E+03_dp, -5.8508031847927323E+03_dp, &
         2.1939029662054095E+03_dp, -3.4737249846121421E+02_dp, 1.9237695027274051E+04_dp, &
         -5.1947738927857921E+04_dp, 6.2765011200887377E+04_dp, -3.0309589558643987E+04_dp, &
         -1.9264261229194974E+04_dp, 4.1446219371737265E+04_dp, -2.6281354883180597E+04_dp, &
         2.8159263824857394E+03_dp, 3.6485501407647080E+03_dp, 2.6770249060636979E+03_dp, &
         -6.2669352427291251E+03_dp, 3.7308014050905122E+03_dp, -8.0644597819162743E+02_dp, &
         2.1234916041378580E+00_dp, -2.4267551942842563E-01_dp, -1.3618390064078997E-01_dp, &
         2.1235253488344166E-02_dp, 3.8121053083591247E-02_dp, -2.2512374350561323E-03_dp, &
         -2.2690956898994009E-02_dp, 3.7722692795086480E-03_dp, 1.4456191546859473E-02_dp, &
         -8.4462833259858951E-03_dp, -2.6996025350312465E-03_dp, 4.0618392843868737E-03_dp, &
         -1.1299679877086958E-03_dp, 2.7792848344652787E+00_dp, -4.7477598542627597E-01_dp, &
         -2.3798512429820987E-01_dp, 7.2098824833812886E-02_dp, 7.6285142890366947E-02_dp, &
         -1.7095747870246605E-02_dp, -4.7341370096482475E-02_dp, 1.2775854073981439E-02_dp, &
         3.5074556718704991E-02_dp, -2.8956193783707895E-02_dp, 2.2732905932494967E-03_dp, &
         5.9324463384063473E-03_dp, -2.0476473783333669E-03_dp, 4.3726378328909492E+00_dp, &
         -1.1147668218629847E+00_dp, -4.9438793138131359E-01_dp, 2.3411800744875491E-01_dp, &
         1.7621745035600736E-01_dp, -7.0675331650434867E-02_dp, -1.1444595532273523E-01_dp, &
         4.3810434696352277E-02_dp, 9.7399112758829939E-02_dp, -1.0075502118647087E-01_dp, &
         2.7820211439353416E-02_dp, 6.8061814563366834E-03_dp, -3.9600522823777032E-03_dp, &
         7.5940664502338420E+00_dp, -2.6359395650789450E+00_dp, -1.0267420292656975E+00_dp, &
         6.8391874036213196E-01_dp, 3.9112093438927609E-01_dp, -2.3797924906131696E-01_dp, &
         -2.6702577764788687E-01_dp, 1.3830923397715175E-01_dp, 2.6053229138318296E-01_dp, &
         -3.2264392621733523E-01_dp, 1.3155017102629712E-01_dp, -7.8585587277630424E-03_dp, &
         -5.8776989356745333E-03_dp, 1.3906829345279375E+01_dp, -6.1815615927462932E+00_dp, &
         -2.0567479439921703E+00_dp, 1.8929310304780897E+00_dp, 8.0939883147882108E-01_dp, &
         -7.3450777930442601E-01_dp, -5.8793748940266832E-01_dp, 4.1849011345367765E-01_dp, &
         6.6959615955201890E-01_dp, -9.8583966364769504E-01_dp, 5.0895589523531948E-01_dp, &
         -9.9254205792793884E-02_dp, -6.1164209449963590E-04_dp, 2.6401188276808139E+01_dp, &
         -1.4535395682302440E+01_dp, -3.9190359193691893E+00_dp, 5.1181176445344025E+00_dp, &
         1.5115068531544618E+00_dp, -2.1714821317110071E+00_dp, -1.2002962513354241E+00_dp, &
         1.2461281841515912E+00_dp, 1.6551300034059584E+00_dp, -2.9317809023065591E+00_dp, &
         1.8031912714428560E+00_dp, -5.0050761339182093E-01_dp, 4.5835377439546375E-02_dp, &
         5.1774174778275956E+01_dp, -3.4673590093429169E+01_dp, -6.8605657769909634E+00_dp, &
         1.3753843857799250E+01_dp, 2.2488088677398350E+00_dp, -6.2744810231042134E+00_dp, &
         -2.1279575450853696E+00_dp, 3.7095435731655004E+00_dp, 3.8509567321319773E+00_dp, &
         -8.5280597524752579E+00_dp, 6.0785754666089655E+00_dp, -2.0517020855153323E+00_dp, &
         2.7264394937468583E-01_dp, 1.0509588382613930E+02_dp, -8.4780704698196232E+01_dp, &
         -9.7392349655477037E+00_dp, 3.7137208312733918E+01_dp, 8.3554407985177170E-01_dp, &
         -1.7880263854377880E+01_dp, -2.4083329597635625E+00_dp, 1.1151723834747575E+01_dp, &
         7.7871979605114570E+00_dp, -2.4043267182405760E+01_dp, 1.9735664203306662E+01_dp, &
         -7.6526126546828053E+00_dp, 1.2084415666148431E+00_dp, 2.2189324774713293E+02_dp, &
         -2.1464604053209558E+02_dp, -3.6807906910198707E+00_dp, 1.0146351820682573E+02_dp, &
         -1.4155528812748599E+01_dp, -5.0203644978574609E+01_dp, 4.3001718188217284E+00_dp, &
         3.4014323314955611E+01_dp, 9.1644378901327990E+00_dp, -6.3442475370119638E+01_dp, &
         6.1193423228111932E+01_dp, -2.6745055759988510E+01_dp, 4.7357733742760555E+00_dp, &
         4.9075000758815480E+02_dp, -5.6911173591276133E+02_dp, 5.7194456067601791E+01_dp, &
         2.8119762788146772E+02_dp, -9.6106533145203329E+01_dp, -1.3626933155906542E+02_dp, &
         4.9851263918608808E+01_dp, 1.0456184482930480E+02_dp, -3.2146041163311445E+01_dp, &
         -1.3945447237077744E+02_dp, 1.7320187517099365E+02_dp, -8.6336243126137077E+01_dp, &
         1.6891573599822905E+01_dp, 1.1481520205431834E+03_dp, -1.6025998192547047E+03_dp, &
         3.9445867091993847E+02_dp, 7.8459322025471704E+02_dp, -4.9045650179813879E+02_dp, &
         -3.3175436609396507E+02_dp, 2.8078846062724392E+02_dp, 3.1070835070821977E+02_dp, &
         -3.6220306330311195E+02_dp, -1.1681704840662950E+02_dp, 3.7714350711930985E+02_dp, &
         -2.3526844262559749E+02_dp, 5.2160651991091960E+01_dp, 2.8817218555073487E+03_dp, &
         -4.8844832039242247E+03_dp, 2.1166301273677968E+03_dp, 2.1145167459343247E+03_dp, &
         -2.3531376839776899E+03_dp, -4.6965613633104721E+02_dp, 1.3278822822389677E+03_dp, &
         7.3531833182374783E+02_dp, -2.2060454899663455E+03_dp, 1.3552254192307953E+03_dp, &
         -1.8597548287900345E+00_dp, -3.2827436822361659E+02_dp, 1.0578152183175671E+02_dp, &
         7.9287162776111199E+03_dp, -1.6556637711428240E+04_dp, 1.1377040542438039E+04_dp, &
         4.3472707467271775E+03_dp, -1.1253948176882712E+04_dp, 2.8844083146489602E+03_dp, &
         5.2635488814341406E+03_dp, -5.3349077678909532E+02_dp, -9.5980797679285060E+03_dp, &
         1.2237468602717436E+04_dp, -7.1752996430490848E+03_dp, 2.0904595579719871E+03_dp, &
         -2.3076473049581341E+02_dp, 2.4941513553111414E+04_dp, -6.5057154343987058E+04_dp, &
         6.8510522134202620E+04_dp, -1.2564664624188634E+04_dp, -4.8494990702839394E+04_dp, &
         4.8066227692409921E+04_dp, 1.2754284959000174E+03_dp, -2.7700527143446598E+04_dp, &
         1.3436695087102896E+03_dp, 3.5023043646265716E+04_dp, -3.8543590945586038E+04_dp, &
         1.8589546507271829E+04_dp, -3.6217344190998742E+03_dp, 1.0880372505911512E+05_dp, &
         -3.3682866668262816E+05_dp, 5.0268295205038588E+05_dp, -4.0507797875013901E+05_dp, &
         7.2631515561032633E+04_dp, 2.4905603733649486E+05_dp, -3.5947552860654687E+05_dp, &
         2.7628042300900549E+05_dp, -1.4381451338213816E+05_dp, 5.7733437538747756E+04_dp, &
         -2.0832981857862389E+04_dp, 6.4928622544881764E+03_dp, -1.1195098273879023E+03_dp, &
         2.6915793781734648E-01_dp, -1.0381942599208310E-02_dp, -9.3650521014913869E-03_dp, &
         3.2090193795957971E-03_dp, -3.3386986359229646E-03_dp, 5.5539179173563335E-03_dp, &
         -6.3480573143521249E-03_dp, 6.2871036161375651E-03_dp, -6.0131949795456498E-03_dp, &
         4.9810441408132712E-03_dp, -3.0620311237843996E-03_dp, 1.1962570988126766E-03_dp, &
         -2.2066239674921657E-04_dp, 8.9689216873395949E-01_dp, -4.2122586464803081E-02_dp, &
         -3.7513882433882192E-02_dp, 1.3881724731562524E-02_dp, -1.3477012568179458E-02_dp, &
         2.2562073529423785E-02_dp, -2.5951410486921640E-02_dp, 2.5567379426983138E-02_dp, &
         -2.4333781051323347E-02_dp, 2.0151995146635144E-02_dp, -1.2400482709093779E-02_dp, &
         4.8452373650168124E-03_dp, -8.9294844328912499E-04_dp, 1.8352119311378547E+00_dp, &
         -1.1605842392533414E-01_dp, -1.0141001120766774E-01_dp, 4.1699067057004879E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c04 = (/-3.6931298185134437E-02_dp, &
         6.2239013775737956E-02_dp, -7.2288174936489244E-02_dp, 7.0693666064322772E-02_dp, &
         -6.6776211112181863E-02_dp, 5.5274684972707100E-02_dp, -3.4069532004586962E-02_dp, &
         1.3318163740503416E-02_dp, -2.4516852040202333E-03_dp, 3.4526910112953169E+00_dp, &
         -2.9811951866509651E-01_dp, -2.5502175885566530E-01_dp, 1.1666479573438553E-01_dp, &
         -9.4640066895441091E-02_dp, 1.5992007690589682E-01_dp, -1.8786362795493361E-01_dp, &
         1.8234170407841388E-01_dp, -1.7071583364783530E-01_dp, 1.4120711271809522E-01_dp, &
         -8.7222157926976809E-02_dp, 3.4129137457941487E-02_dp, -6.2772574750481792E-03_dp, &
         6.4526874369355705E+00_dp, -7.4379137647143534E-01_dp, -6.2169351557404495E-01_dp, &
         3.1608406508912079E-01_dp, -2.3664910820793422E-01_dp, 3.9861181596578510E-01_dp, &
         -4.7437842206341091E-01_dp, 4.5710809927671725E-01_dp, -4.2357791972371833E-01_dp, &
         3.4994703099419200E-01_dp, -2.1674549913705246E-01_dp, 8.4953665492070438E-02_dp, &
         -1.5619616785382430E-02_dp, 1.2272761702136265E+01_dp, -1.8365536182683235E+00_dp, &
         -1.4952269059085537E+00_dp, 8.4751536974247987E-01_dp, -5.8923289605863272E-01_dp, &
         9.8181746694991578E-01_dp, -1.1862663094962580E+00_dp, 1.1351701430563665E+00_dp, &
         -1.0389137932494279E+00_dp, 8.5664247790845927E-01_dp, -5.3242501279025167E-01_dp, &
         2.0925723568583102E-01_dp, -3.8489862677363318E-02_dp, 2.4000086828210332E+01_dp, &
         -4.5581060135710860E+00_dp, -3.5977120742775472E+00_dp, 2.2897011921494417E+00_dp, &
         -1.4872248187048727E+00_dp, 2.4261115525821300E+00_dp, -2.9842698163078438E+00_dp, &
         2.8374294543564580E+00_dp, -2.5569428859304861E+00_dp, 2.1013503214810827E+00_dp, &
         -1.3118947003867396E+00_dp, 5.1782664621522245E-01_dp, -9.5402662711553898E-02_dp, &
         4.8587093531959233E+01_dp, -1.1532062155028678E+01_dp, -8.7655022066070973E+00_dp, &
         6.3319337945364706E+00_dp, -3.8731156981586130E+00_dp, 6.0963207775734327E+00_dp, &
         -7.6591948339186171E+00_dp, 7.2433544682793292E+00_dp, -6.3996585265106960E+00_dp, &
         5.2289327669288426E+00_dp, -3.2829199929814634E+00_dp, 1.3043816113094733E+00_dp, &
         -2.4119202153591454E-01_dp, 1.0254367589701396E+02_dp, -3.0156545743278251E+01_dp, &
         -2.1861824137429267E+01_dp, 1.8198871770079609E+01_dp, -1.0610092557902053E+01_dp, &
         1.5784369861157300E+01_dp, -2.0325106722771388E+01_dp, 1.9159263068907812E+01_dp, &
         -1.6497575716105022E+01_dp, 1.3338797204809440E+01_dp, -8.4308703027030507E+00_dp, &
         3.3832494619455296E+00_dp, -6.2994976928592783E-01_dp, 2.2763518387273865E+02_dp, &
         -8.2764595584286539E+01_dp, -5.6398311753951880E+01_dp, 5.5286001032641892E+01_dp, &
         -3.1300545607087930E+01_dp, 4.2735085101753064E+01_dp, -5.6557182696520613E+01_dp, &
         5.3353437804874766E+01_dp, -4.4397860550024987E+01_dp, 3.5190517825708255E+01_dp, &
         -2.2390621695313161E+01_dp, 9.1181332326342606E+00_dp, -1.7187754447680910E+00_dp, &
         5.3814419453876985E+02_dp, -2.4299588074893020E+02_dp, -1.5201445631333939E+02_dp, &
         1.8130374185702280E+02_dp, -1.0262892549571430E+02_dp, 1.2338122205446149E+02_dp, &
         -1.6769634862214531E+02_dp, 1.5947931529626564E+02_dp, -1.2673253607941503E+02_dp, &
         9.6503470517689962E+01_dp, -6.1507970117755306E+01_dp, 2.5568992544284665E+01_dp, &
         -4.9210909009850994E+00_dp, 1.3801918068313626E+03_dp, -7.8409403684044321E+02_dp, &
         -4.3139798224082045E+02_dp, 6.6108802464127928E+02_dp, -3.9172133571873263E+02_dp, &
         3.9236375791502667E+02_dp, -5.4013676730762052E+02_dp, 5.2452481311060501E+02_dp, &
         -3.9173273726869706E+02_dp, 2.7308075960502873E+02_dp, -1.6996111729874119E+02_dp, &
         7.2338992388678662E+01_dp, -1.4396994651161968E+01_dp, 3.9580600472914825E+03_dp, &
         -2.9036032599844793E+03_dp, -1.2781768816295728E+03_dp, 2.8070929268418577E+03_dp, &
         -1.8752951429725697E+03_dp, 1.4753386767800280E+03_dp, -1.9323503518765785E+03_dp, &
         1.9527579205070861E+03_dp, -1.3486577520305257E+03_dp, 7.5381327111334554E+02_dp, &
         -3.9743710790630257E+02_dp, 1.6635147838756436E+02_dp, -3.4761194955563198E+01_dp, &
         1.3425684708567962E+04_dp, -1.3383900660812744E+04_dp, -3.4320824963960499E+03_dp, &
         1.5050487619482205E+04_dp, -1.2942745398196372E+04_dp, 8.1171714061498142E+03_dp, &
         -7.8406387741109365E+03_dp, 8.0705849662959281E+03_dp, -5.1239781721026657E+03_dp, &
         1.3200473018363366E+03_dp, 4.4071137447532215E+02_dp, -4.2582815928175688E+02_dp, &
         9.8145031152881089E+01_dp, 6.1763293249195012E+04_dp, -9.3023683746086099E+04_dp, &
         1.4302377887690851E+04_dp, 1.1395909986937273E+05_dp, -1.7772764092445478E+05_dp, &
         1.4368085671526083E+05_dp, -6.1917915753482834E+04_dp, -1.3748361328957566E+04_dp, &
         6.2361092152953963E+04_dp, -7.7904069086697375E+04_dp, 6.0598777613532206E+04_dp, &
         -2.8232839487469380E+04_dp, 5.9912034166305712E+03_dp, 2.2102099829502819E+00_dp, &
         -8.9807091694595470E-02_dp, -8.0738852355447699E-02_dp, 2.8245551420289328E-02_dp, &
         -2.8838152951585126E-02_dp, 4.8060961814259832E-02_dp, -5.5022894223366300E-02_dp, &
         5.4417947301434594E-02_dp, -5.1980922605685237E-02_dp, 4.3055991607175798E-02_dp, &
         -2.6474712012505622E-02_dp, 1.0343216742807535E-02_dp, -1.9074395117850485E-03_dp, &
         2.9526385249227900E+00_dp, -1.8342158808651834E-01_dp, -1.6079974432000388E-01_dp, &
         6.4990457572363936E-02_dp, -5.8357582370356993E-02_dp, 9.8375835274615853E-02_dp, &
         -1.1403666430432409E-01_dp, 1.1164874104323870E-01_dp, -1.0561858908323142E-01_dp, &
         8.7439013178214162E-02_dp, -5.3873844636419935E-02_dp, 2.1055369777301082E-02_dp, &
         -3.8763341662694521E-03_dp, 4.7882425300284677E+00_dp, -4.4880027377068837E-01_dp, &
         -3.8387620646428122E-01_dp, 1.7581567932182518E-01_dp, -1.4218651556165815E-01_dp, &
         2.4088533741840665E-01_dp, -2.8285896497383756E-01_dp, 2.7446561521278501E-01_dp, &
         -2.5704300468522123E-01_dp, 2.1264001116035619E-01_dp, -1.3132372928749306E-01_dp, &
         5.1372473735998153E-02_dp, -9.4466605494810223E-03_dp, 8.5963629598619455E+00_dp, &
         -1.1032812111617041E+00_dp, -9.2128991958738193E-01_dp, 4.7072149697983523E-01_dp, &
         -3.5013599087368830E-01_dp, 5.9164664653692711E-01_dp, -7.0411617076286537E-01_dp, &
         6.7802505793072532E-01_dp, -6.2831000597729314E-01_dp, 5.1918399266101378E-01_dp, &
         -3.2153146818208928E-01_dp, 1.2598734950885529E-01_dp, -2.3156182641453846E-02_dp, &
         1.6307279051808887E+01_dp, -2.6964639955096659E+00_dp, -2.1929083193674388E+00_dp, &
         1.2494850984921542E+00_dp, -8.6264580098333044E-01_dp, 1.4424439050266553E+00_dp, &
         -1.7431000698797154E+00_dp, 1.6666423180736005E+00_dp, -1.5253713037802321E+00_dp, &
         1.2581191094820894E+00_dp, -7.8188837190149219E-01_dp, 3.0719187328985836E-01_dp, &
         -5.6478457126488940E-02_dp, 3.2180867728506904E+01_dp, -6.6411072684078274E+00_dp, &
         -5.2372172602238214E+00_dp, 3.3470888636894665E+00_dp, -2.1592514567307126E+00_dp, &
         3.5371868143609047E+00_dp, -4.3517183020335670E+00_dp, 4.1336520605961979E+00_dp, &
         -3.7258847437211742E+00_dp, 3.0636078261990956E+00_dp, -1.9124981724704819E+00_dp, &
         7.5452097435395959E-01_dp, -1.3892551807807468E-01_dp, 6.5943104829375315E+01_dp, &
         -1.6707127044375667E+01_dp, -1.2698264773736545E+01_dp, 9.1863412650630529E+00_dp, &
         -5.5829483431324674E+00_dp, 8.8392123953951618E+00_dp, -1.1105112337604499E+01_dp, &
         1.0490324644517454E+01_dp, -9.2755649942267286E+00_dp, 7.5866003373259350E+00_dp, &
         -4.7628121299140389E+00_dp, 1.8909156066204140E+00_dp, -3.4932019459840230E-01_dp, &
         1.4087116058808186E+02_dp, -4.3470586509878942E+01_dp, -3.1572239382059106E+01_dp, &
         2.6185590050497311E+01_dp, -1.5166880594846853E+01_dp, 2.2778700310398051E+01_dp, &
         -2.9319687356015418E+01_dp, 2.7596362755558843E+01_dp, -2.3808410787503707E+01_dp, &
         1.9293100477047371E+01_dp, -1.2194695218999417E+01_dp, 4.8872630176721907E+00_dp, &
         -9.0854459291729428E-01_dp, 3.1603181254524878E+02_dp, -1.1853513661034587E+02_dp, &
         -8.1246680173306473E+01_dp, 7.8648347652941197E+01_dp, -4.4153278086217064E+01_dp, &
         6.1303806049424246E+01_dp, -8.1073505393952701E+01_dp, 7.6296406201925507E+01_dp, &
         -6.3752564021982288E+01_dp, 5.0794561685790477E+01_dp, -3.2341634414896959E+01_dp, &
         1.3141737689621349E+01_dp, -2.4701157665658586E+00_dp, 7.5252253575264456E+02_dp, &
         -3.4418682220066631E+02_dp, -2.1837415788962826E+02_dp, 2.5329125041502920E+02_dp, &
         -1.4128547056745032E+02_dp, 1.7507672278319296E+02_dp, -2.3804626639739695E+02_dp, &
         2.2535235168628145E+02_dp, -1.8051776327743372E+02_dp, 1.3923468265213427E+02_dp, &
         -8.9091837917185714E+01_dp, 3.6920629587167475E+01_dp, -7.0690524517695410E+00_dp, &
         1.9315421410301064E+03_dp, -1.0874625239245886E+03_dp, -6.1808932166184934E+02_dp, &
         8.9545464671544005E+02_dp, -5.1479696749349444E+02_dp, 5.4356197242025212E+02_dp, &
         -7.5448078467711980E+02_dp, 7.2603369830764473E+02_dp, -5.4986172074879164E+02_dp, &
         3.9669672465398293E+02_dp, -2.5164163702286163E+02_dp, 1.0700313706124334E+02_dp, &
         -2.1113910665526991E+01_dp, 5.4723123751409612E+03_dp, -3.8610955995775034E+03_dp, &
         -1.8448756713449409E+03_dp, 3.5966262696431140E+03_dp, -2.2512699759682537E+03_dp, &
         1.9194105802260040E+03_dp, -2.6322931011174383E+03_dp, 2.6185867118885690E+03_dp, &
         -1.8443677996699814E+03_dp, 1.1419121865305731E+03_dp, -6.7280482181537468E+02_dp, &
         2.9210777012637755E+02_dp, -6.0511721609154549E+01_dp, 1.7804801411705495E+04_dp, &
         -1.6259024258501489E+04_dp, -5.5718594192341498E+03_dp, 1.7345758837342113E+04_dp, &
         -1.2877366253533812E+04_dp, 8.5915553852783196E+03_dp, -1.0315077162695961E+04_dp, &
         1.0800104463257729E+04_dp, -7.0138651831887073E+03_dp, 2.8554212903400171E+03_dp, &
         -8.8722261141725619E+02_dp, 2.9569669149664014E+02_dp, -6.9226682688697807E+01_dp, &
         7.1835823339602095E+04_dp, -9.0186165593538084E+04_dp, -9.7248478944525286E+03_dp, &
         1.1062096220530488E+05_dp, -1.1369726450560044E+05_dp, 6.7036834170998452E+04_dp, &
         -4.6207804389370787E+04_dp, 4.4594452729411365E+04_dp, -2.6072626782514293E+04_dp, &
         -2.9846178239841133E+03_dp, 1.5681375871473874E+04_dp, -1.0022357789314863E+04_dp, &
         2.3179238419310946E+03_dp, 4.4219529370203713E+05_dp, -8.4755186000937573E+05_dp, &
         3.7879188093690283E+05_dp, 9.7631641875634436E+05_dp, -2.2196339514296870E+06_dp, &
         2.3054796315979110E+06_dp, -1.0628907039914103E+06_dp, -7.6347796690832789E+05_dp, &
         2.1400073212989480E+06_dp, -2.4192321832669177E+06_dp, 1.7082231092388814E+06_dp, &
         -7.3012565960855584E+05_dp, 1.4473968969337529E+05_dp, 2.7143097167899877E-01_dp, &
         -1.0426354655888225E-09_dp, -1.3514349234655127E-02_dp, 1.6236462271473035E-02_dp, &
         -1.9832094707145951E-02_dp, 1.3779856524964107E-02_dp, 1.5011176255910088E-02_dp, &
         -7.7179118599429580E-02_dp, 1.6386504807733129E-01_dp, -2.2964169829422010E-01_dp, &
         2.1619787855597516E-01_dp, -1.2421278261439007E-01_dp, 3.2870250330679252E-02_dp, &
         9.0613777954751162E-01_dp, -4.2561353091248025E-09_dp, -5.5108681647653930E-02_dp, &
         6.6208860978777484E-02_dp, -8.0041705305285135E-02_dp, 5.4189321111835380E-02_dp, &
         6.4928039936621940E-02_dp, -3.2000393170497327E-01_dp, 6.7367170849599256E-01_dp, &
         -9.4004438108949862E-01_dp, 8.8254299622584398E-01_dp, -5.0603178442964947E-01_dp, &
         1.3370755800909928E-01_dp, 1.8607801318806969E+00_dp, -1.1832063710451179E-08_dp, &
         -1.5296272228579366E-01_dp, 1.8377288757765206E-01_dp, -2.1880198683428151E-01_dp, &
         1.4228504563245828E-01_dp, 1.9523650997654252E-01_dp, -9.0945729598550951E-01_dp, &
         1.8916140218456958E+00_dp, -2.6232469600311217E+00_dp, 2.4527671312473118E+00_dp, &
         -1.4022183672866397E+00_dp, 3.6967613733809346E-01_dp, 3.5186336824147015E+00_dp/)
      REAL(KIND=dp), DIMENSION(13, 30, 5), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04, c05/), (/13, 30, 5/))

      INTEGER                                            :: irange

      IF (Rc <= 1.5000000000000000E+01_dp) Rc = 1.5000000000000000E+01_dp
      IF (Rc >= 4.6894989013671875E+05_dp) Rc = 4.6894989013671875E+05_dp
      IF (Rc >= 1.5000000000000000E+01_dp .AND. Rc <= 1.6990973774601900E+02_dp) THEN
         irange = 1
         L_b = 1.5000000000000000E+01_dp
         U_b = 1.6990973774601900E+02_dp
      END IF
      IF (Rc >= 1.6990973774601900E+02_dp .AND. Rc <= 1.8995280520207393E+03_dp) THEN
         irange = 2
         L_b = 1.6990973774601900E+02_dp
         U_b = 1.8995280520207393E+03_dp
      END IF
      IF (Rc >= 1.8995280520207393E+03_dp .AND. Rc <= 2.6390317079362849E+04_dp) THEN
         irange = 3
         L_b = 1.8995280520207393E+03_dp
         U_b = 2.6390317079362849E+04_dp
      END IF
      IF (Rc >= 2.6390317079362849E+04_dp .AND. Rc <= 1.5948665335135348E+05_dp) THEN
         irange = 4
         L_b = 2.6390317079362849E+04_dp
         U_b = 1.5948665335135348E+05_dp
      END IF
      IF (Rc >= 1.5948665335135348E+05_dp .AND. Rc <= 4.6894989013671875E+05_dp) THEN
         irange = 5
         L_b = 1.5948665335135348E+05_dp
         U_b = 4.6894989013671875E+05_dp
      END IF

      fit_coef(1:13, 1:30) = coefdata(1:13, 1:30, irange)

   END SUBROUTINE get_fit_coef_k15

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k16(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 32)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.8440983439459679E-01_dp, &
         -7.0479249003300562E-02_dp, -1.6832381157590469E-02_dp, 9.9837323526780937E-03_dp, &
         5.1989041899428937E-03_dp, -3.5423416042951205E-03_dp, -2.1543663571483209E-03_dp, &
         1.3096125508292179E-03_dp, 2.2049401312140603E-03_dp, -2.5023058617921153E-03_dp, &
         7.5375163633056852E-04_dp, 1.3637466463565460E-04_dp, -9.2994202830172699E-05_dp, &
         5.8054036572020107E-01_dp, -2.4380379683949494E-01_dp, -4.4938696140059450E-02_dp, &
         3.8768465512213003E-02_dp, 1.3593778767841709E-02_dp, -1.4461499926526256E-02_dp, &
         -5.4490788053341418E-03_dp, 5.5553275407733936E-03_dp, 7.1422948686751375E-03_dp, &
         -1.0583913052061498E-02_dp, 4.9741685262334190E-03_dp, -6.6432140641299532E-04_dp, &
         -1.0892699705666875E-04_dp, 1.0638318043371691E+00_dp, -5.2505303151143268E-01_dp, &
         -4.9624782924737383E-02_dp, 9.5687743884690382E-02_dp, 1.2767947916143167E-02_dp, &
         -3.7404011533309391E-02_dp, -3.9949046346714071E-03_dp, 1.4966928016166042E-02_dp, &
         1.2577310120692639E-02_dp, -2.8100049351042718E-02_dp, 1.8433425451462267E-02_dp, &
         -5.4088352179882731E-03_dp, 5.4854194494455884E-04_dp, 1.7106633703662533E+00_dp, &
         -1.0248769125439021E+00_dp, 1.0241427657345478E-02_dp, 2.0420458035779018E-01_dp, &
         -1.5264072128014391E-02_dp, -8.1320404223678117E-02_dp, 1.1517312423593138E-02_dp, &
         3.3558605291626020E-02_dp, 1.4544544223415962E-02_dp, -6.0901876447184505E-02_dp, &
         5.0292199523917204E-02_dp, -1.8965089106949350E-02_dp, 2.8601543886679452E-03_dp, &
         2.6282972200588932E+00_dp, -1.9229483513273367E+00_dp, 2.2722345074070921E-01_dp, &
         3.9046743014432933E-01_dp, -1.1085109620769071E-01_dp, -1.5292292332946034E-01_dp, &
         6.1324562893261297E-02_dp, 6.4440760127347133E-02_dp, -3.1375758896852156E-05_dp, &
         -1.0943572242374154E-01_dp, 1.1119871011947056E-01_dp, -4.8686829057736594E-02_dp, &
         8.5187843378692653E-03_dp, 3.9756816802545427E+00_dp, -3.5229306793001349E+00_dp, &
         8.0023523017838849E-01_dp, 6.7084504352855900E-01_dp, -3.5600097515911666E-01_dp, &
         -2.4559425187376815E-01_dp, 1.8405067150112531E-01_dp, 1.0450118269262665E-01_dp, &
         -6.2242399538416641E-02_dp, -1.5509553332024303E-01_dp, 2.0567739453273481E-01_dp, &
         -1.0262895944488672E-01_dp, 1.9845825902533719E-02_dp, 5.9958030217368909E+00_dp, &
         -6.3467161270531429E+00_dp, 2.1437674998301484E+00_dp, 1.0109234785696470E+00_dp, &
         -9.0400327994454444E-01_dp, -3.1019661013539263E-01_dp, 4.4434854634858123E-01_dp, &
         1.3034314501804636E-01_dp, -2.3283606473742849E-01_dp, -1.3205232328255015E-01_dp, &
         3.1131037411240431E-01_dp, -1.8221061043989620E-01_dp, 3.8754681380380226E-02_dp, &
         9.0676760085489132E+00_dp, -1.1299033181149410E+01_dp, 5.1048832785744436E+00_dp, &
         1.2130368920639967E+00_dp, -2.0156334343903222E+00_dp, -1.8774490141514191E-01_dp, &
         9.3144160276729693E-01_dp, 6.6775735953655421E-02_dp, -6.0860647626862330E-01_dp, &
         1.1967109798777947E-01_dp, 3.3525976868624457E-01_dp, -2.6382031281912727E-01_dp, &
         6.3577702221923230E-02_dp, 1.3791772574427243E+01_dp, -1.9957844758745054E+01_dp, &
         1.1393907501217026E+01_dp, 6.3027745141587310E-01_dp, -4.0736217099431213E+00_dp, &
         5.3209770609899276E-01_dp, 1.7196075448940431E+00_dp, -2.7410147020725067E-01_dp, &
         -1.2993291005460068E+00_dp, 9.1405021206328108E-01_dp, 3.9446714460368994E-02_dp, &
         -2.6218217849092257E-01_dp, 8.2004458930670890E-02_dp, 2.1134043943396996E+01_dp, &
         -3.5091359042775800E+01_dp, 2.4438380342850234E+01_dp, -2.5312397126433797E+00_dp, &
         -7.4649061114523123E+00_dp, 2.7785338021255463E+00_dp, 2.7113234706643428E+00_dp, &
         -1.2881207709853775E+00_dp, -2.3173947170071818E+00_dp, 2.7571681079983406E+00_dp, &
         -1.0669271742866862E+00_dp, 3.3286659884836270E-02_dp, 5.7463683876706147E-02_dp, &
         3.2679093868926657E+01_dp, -6.1590985372944964E+01_dp, 5.1071575868628280E+01_dp, &
         -1.2865796070437154E+01_dp, -1.1968112302623199E+01_dp, 8.4107024757715152E+00_dp, &
         3.1677115274179708E+00_dp, -3.6601777916288216E+00_dp, -3.2433702877664414E+00_dp, &
         6.2231339219062374E+00_dp, -3.8192853459283764E+00_dp, 1.0504543812661566E+00_dp, &
         -9.4464855653482102E-02_dp, 5.1112233923296436E+01_dp, -1.0820984617012243E+02_dp, &
         1.0486168192581441E+02_dp, -4.1564959865575254E+01_dp, -1.4578956048017877E+01_dp, &
         2.0463299891738025E+01_dp, 5.4796957878465291E-01_dp, -8.1431617361849096E+00_dp, &
         -2.3864407534639644E+00_dp, 1.1108185317660142E+01_dp, -9.0914165395682556E+00_dp, &
         3.4289059180469894E+00_dp, -5.2330290922697853E-01_dp, 8.1288207408222235E+01_dp, &
         -1.9111334897720442E+02_dp, 2.1279120241913324E+02_dp, -1.1477244836095247E+02_dp, &
         -3.0743808414701879E+00_dp, 4.1497540322186111E+01_dp, -1.1492293983396848E+01_dp, &
         -1.3925871166929845E+01_dp, 4.2907496508986602E+00_dp, 1.3859380628329522E+01_dp, &
         -1.5876854948348997E+01_dp, 7.3152892639960951E+00_dp, -1.3362731143808417E+00_dp, &
         1.3327942970575734E+02_dp, -3.4300787025442804E+02_dp, 4.3047110358093215E+02_dp, &
         -2.9164886994241795E+02_dp, 5.9859717276872587E+01_dp, 6.4920747245308959E+01_dp, &
         -4.3129286137280374E+01_dp, -1.4066222386165592E+01_dp, 2.0329366289960053E+01_dp, &
         6.4662789320134308E+00_dp, -1.8713821361854901E+01_dp, 1.0879757257521486E+01_dp, &
         -2.2809868996647227E+00_dp, 2.3476896298756850E+02_dp, -6.4871239939014094E+02_dp, &
         8.9470567514658705E+02_dp, -7.1469544614179904E+02_dp, 2.7168514298197266E+02_dp, &
         5.4663915078208277E+01_dp, -9.8551569029646402E+01_dp, 4.4291571928112470E+00_dp, &
         4.0875449189810638E+01_dp, -1.3810716456952706E+01_dp, -1.3765103917995090E+01_dp, &
         1.2346735490952792E+01_dp, -3.0319784854436125E+00_dp, 5.2021357980075481E+02_dp, &
         -1.5047681469099909E+03_dp, 2.2056583940220830E+03_dp, -1.9489466301285634E+03_dp, &
         9.6172634089814062E+02_dp, -8.2178385250715877E+01_dp, -1.8689166188128539E+02_dp, &
         4.7992813706474685E+01_dp, 7.5507434893328252E+01_dp, -5.0888531694752380E+01_dp, &
         -6.1648337992718236E+00_dp, 1.6427444516897769E+01_dp, -4.7804380737887202E+00_dp, &
         1.4931443468447578E+00_dp, -5.8471071386162488E-01_dp, -1.3137927495216153E-01_dp, &
         8.5629508847014146E-02_dp, 4.0460885624517218E-02_dp, -3.0854519142842370E-02_dp, &
         -1.6668874154937584E-02_dp, 1.1539311755993547E-02_dp, 1.8057976799045881E-02_dp, &
         -2.2047699644401675E-02_dp, 7.7604920500705474E-03_dp, 4.2470957702766204E-04_dp, &
         -6.4150955581103778E-04_dp, 1.7150666636291758E+00_dp, -8.5070446961324930E-01_dp, &
         -8.2587877714910274E-02_dp, 1.5733665025528792E-01_dp, 2.2362433751069832E-02_dp, &
         -6.1920203471438347E-02_dp, -7.4462817802991854E-03_dp, 2.4762989163368353E-02_dp, &
         2.1594517266609357E-02_dp, -4.6861787787818826E-02_dp, 3.0232817720782037E-02_dp, &
         -8.6536690927158984E-03_dp, 8.2816489219117028E-04_dp, 2.2012593013113242E+00_dp, &
         -1.4708994286649284E+00_dp, 7.0779599888676187E-02_dp, 3.1369855773415140E-01_dp, &
         -4.0224055770866012E-02_dp, -1.2774489263868016E-01_dp, 2.4938170206917257E-02_dp, &
         5.3277305829681154E-02_dp, 1.9570090404201004E-02_dp, -9.7451375771407875E-02_dp, &
         8.3591537572802813E-02_dp, -3.2502927863455926E-02_dp, 5.0679445972248672E-03_dp, &
         3.0435610577413037E+00_dp, -2.6439734331166180E+00_dp, 4.6643978960348892E-01_dp, &
         5.7255374553028604E-01_dp, -2.0967596109489331E-01_dp, -2.2871315539930384E-01_dp, &
         1.1214157555126254E-01_dp, 9.7841452353856731E-02_dp, -1.1952067880384304E-02_dp, &
         -1.6611264833371459E-01_dp, 1.7882104458009906E-01_dp, -8.0938254356085149E-02_dp, &
         1.4553336693307755E-02_dp, 4.4013915563626602E+00_dp, -4.7370367251945105E+00_dp, &
         1.3952141799394053E+00_dp, 9.3083326856233284E-01_dp, -6.0770432754104109E-01_dp, &
         -3.4146109652219514E-01_dp, 3.1148445412961517E-01_dp, 1.4799614852691056E-01_dp, &
         -1.2454577962361596E-01_dp, -2.1157820423498469E-01_dp, 3.1273896919979172E-01_dp, &
         -1.6264093750898964E-01_dp, 3.2321056217098403E-02_dp, 6.5363325586350669E+00_dp, &
         -8.4002734007267286E+00_dp, 3.4516744838336901E+00_dp, 1.2958797798191262E+00_dp, &
         -1.4530782665536190E+00_dp, -3.7130371833416770E-01_dp, 7.1329068164973419E-01_dp, &
         1.5886534628143975E-01_dp, -4.1300823765828609E-01_dp, -1.1297420382509935E-01_dp, &
         4.3095219716435623E-01_dp, -2.7099924501127154E-01_dp, 5.9765906484496450E-02_dp, &
         9.8718409209594444E+00_dp, -1.4774703156334951E+01_dp, 7.8344307203439723E+00_dp, &
         1.3100128552371193E+00_dp, -3.1029964454694561E+00_dp, -4.8425439664709267E-02_dp, &
         1.4305547840984860E+00_dp, 1.7043059862931740E-03_dp, -1.0173113806737391E+00_dp, &
         3.9971961481904772E-01_dp, 3.5717334078499841E-01_dp, -3.5227483495704071E-01_dp, &
         9.0840950487899280E-02_dp, 1.5094285692399961E+01_dp, -2.5863006570028976E+01_dp, &
         1.6941424618914212E+01_dp, -8.2529211252986553E-02_dp, -6.0502265317186374E+00_dp, &
         1.2842402390008709E+00_dp, 2.5244804302560042E+00_dp, -6.2891082730515735E-01_dp, &
         -2.0682717362404395E+00_dp, 1.8199209168598900E+00_dp, -3.1696402449624006E-01_dp, &
         -2.4788872827770345E-01_dp, 1.0040053068940764E-01_dp, 2.3326532076830389E+01_dp, &
         -4.5194032269559408E+01_dp, 3.5550320300834244E+01_dp, -5.6756451356893560E+00_dp, &
         -1.0703822653505913E+01_dp, 5.0650560255240329E+00_dp, 3.7433826066761373E+00_dp, &
         -2.3546914919859425E+00_dp, -3.4852289677497725E+00_dp, 4.8766148395113742E+00_dp, &
         -2.3683936704709061E+00_dp, 3.9477013548749840E-01_dp, 2.4217597532977045E-02_dp, &
         3.6438523276203654E+01_dp, -7.9026710872235796E+01_dp, 7.3153148523645797E+01_dp, &
         -2.2447806103578166E+01_dp, -1.6400252521887804E+01_dp, 1.4117189879671018E+01_dp, &
         3.7704063367345246E+00_dp, -6.2200433997860021E+00_dp, -4.4324532589543466E+00_dp, &
         1.0293341416331923E+01_dp, -7.0471929333047125E+00_dp, 2.2491704032299409E+00_dp, &
         -2.7472784846345960E-01_dp, 5.7650008601931596E+01_dp, -1.3859335346138556E+02_dp, &
         1.4854728177346706E+02_dp, -6.7174062977350260E+01_dp, -1.8000776577693308E+01_dp, &
         3.3025212703350597E+01_dp, -1.6648355635515739E+00_dp, -1.3341370484527044E+01_dp, &
         -1.9519662111886593E+00_dp, 1.7476000515060889E+01_dp, -1.5668489541211233E+01_dp, &
         6.3504551706926540E+00_dp, -1.0447824464894551E+00_dp, 9.2889712584894980E+01_dp, &
         -2.4463928647750311E+02_dp, 2.9889577966035040E+02_dp, -1.7875521526141426E+02_dp, &
         4.2730769050089581E+00_dp, 6.5596464491298477E+01_dp, -2.3416502194370231E+01_dp, &
         -2.2316628174048979E+01_dp, 1.1587124874044365E+01_dp, 1.9473620072500285E+01_dp, &
         -2.6128292622546354E+01_dp, 1.2919004312807344E+01_dp, -2.4858470852747883E+00_dp, &
         1.5466343631749228E+02_dp, -4.3876623267221055E+02_dp, 5.9917637433678522E+02_dp, &
         -4.4383926863394601E+02_dp, 1.1056185052763135E+02_dp, 9.9764791389111096E+01_dp, &
         -8.0352055369394023E+01_dp, -1.9622721204759706E+01_dp, 4.5089137417242156E+01_dp, &
         -3.6046484399499579E+00_dp, -2.4148729613563599E+01_dp, 1.6877244514385687E+01_dp, &
         -3.8474659319176436E+00_dp, 2.7677631153916957E+02_dp, -8.2445521989176632E+02_dp, &
         1.2212144990999409E+03_dp, -1.0555839720133740E+03_dp, 4.5253808804371664E+02_dp, &
         6.5091562231667240E+01_dp, -1.7662774539141580E+02_dp, 2.8862592004300790E+01_dp, &
         7.9025353206514183E+01_dp, -5.7883045427627493E+01_dp, 5.0699996458841383E+00_dp, &
         9.8497860011119158E+00_dp, -3.3639190612129344E+00_dp, 5.9487274112930379E+02_dp, &
         -1.7978611801015538E+03_dp, 2.7622843520701335E+03_dp, -2.5880461605405826E+03_dp, &
         1.4022414211295804E+03_dp, -2.0359258628022113E+02_dp, -2.6091417556034258E+02_dp, &
         1.2051226872116960E+02_dp, 8.5298374145724267E+01_dp, -1.0115805307780569E+02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/2.9539145031653501E+01_dp, &
         4.7715049407002184E+00_dp, -3.2712144543127324E+00_dp, 2.2264348078485641E+03_dp, &
         -6.6288008445597088E+03_dp, 1.0084012629385183E+04_dp, -9.4383385665132246E+03_dp, &
         5.2647790264708256E+03_dp, -1.0875754774918837E+03_dp, -5.8837187049175338E+02_dp, &
         2.6613872996811381E+02_dp, 2.7202180766861471E+02_dp, -2.5302823108588976E+02_dp, &
         3.0050455194523508E+01_dp, 4.2500745288190885E+01_dp, -1.5231116966455813E+01_dp, &
         2.2506815302769312E-01_dp, -4.6864684645862111E-02_dp, -1.6076066373471819E-02_dp, &
         5.2617151773664064E-03_dp, 4.6066685262613890E-03_dp, -1.7521312997179688E-03_dp, &
         -1.9251133062711246E-03_dp, 5.9345994494066007E-04_dp, 1.4834834984639620E-03_dp, &
         -1.0955734366602742E-03_dp, -1.0027326064082555E-04_dp, 3.6348878322400217E-04_dp, &
         -1.1127160112187439E-04_dp, 7.2603005105721208E-01_dp, -1.7378410728319532E-01_dp, &
         -5.1996715280760691E-02_dp, 2.4016913773170116E-02_dp, 1.5560321159120889E-02_dp, &
         -8.5587123146005032E-03_dp, -6.5878194833641881E-03_dp, 3.0580493812781056E-03_dp, &
         5.8185879511744283E-03_dp, -5.6506429717535292E-03_dp, 9.6145797073991931E-04_dp, &
         8.4659623403035588E-04_dp, -3.3812623825664234E-04_dp, 1.3950206706580586E+00_dp, &
         -4.1818083799113709E-01_dp, -9.6697550419933886E-02_dp, 7.3114311638722271E-02_dp, &
         3.0573982779230305E-02_dp, -2.7876982279685374E-02_dp, -1.3064679941770251E-02_dp, &
         1.0479459233157986E-02_dp, 1.4673025417197233E-02_dp, -1.9452677781320156E-02_dp, &
         7.6812744538724814E-03_dp, -1.3354340974799377E-04_dp, -4.6374652770920057E-04_dp, &
         2.4002817735423130E+00_dp, -9.2576009390463998E-01_dp, -1.4413374515017688E-01_dp, &
         1.9374327535805080E-01_dp, 4.6522602576221726E-02_dp, -7.7663309125627397E-02_dp, &
         -1.9365836194771904E-02_dp, 3.0366121689986602E-02_dp, 3.1934293985556542E-02_dp, &
         -5.6570946869567498E-02_dp, 3.1245323124887436E-02_dp, -6.6302180297778618E-03_dp, &
         9.0727048872160795E-05_dp, 4.0131529792607168E+00_dp, -1.9771947673658135E+00_dp, &
         -1.5401727009837554E-01_dp, 4.6736033429357465E-01_dp, 4.2297016009197330E-02_dp, &
         -1.9390246830191787E-01_dp, -1.3451643644637289E-02_dp, 7.8314725324450052E-02_dp, &
         6.0048855340030217E-02_dp, -1.4534657554682370E-01_dp, 9.8739620200257255E-02_dp, &
         -3.0047651240179940E-02_dp, 3.2482467714997627E-03_dp, 6.6977661937400432E+00_dp, &
         -4.1338315742806850E+00_dp, 1.1682547993851978E-02_dp, 1.0522365350263854E+00_dp, &
         -5.6267873028670708E-02_dp, -4.4464162072743280E-01_dp, 4.6411192993530540E-02_dp, &
         1.8465630884986559E-01_dp, 8.9370042068134745E-02_dp, -3.3628926971532436E-01_dp, &
         2.6985734900416353E-01_dp, -9.8821388710520416E-02_dp, 1.4332394513831495E-02_dp, &
         1.1276246206952271E+01_dp, -8.5406849172019346E+00_dp, 7.5682719586455571E-01_dp, &
         2.2414673914939178E+00_dp, -4.6842658114719654E-01_dp, -9.4428627475309324E-01_dp, &
         2.8194134886271960E-01_dp, 4.0121325571233513E-01_dp, 6.1942620093743794E-02_dp, &
         -6.9757764732780769E-01_dp, 6.6046001816422062E-01_dp, -2.7625327157326457E-01_dp, &
         4.6339073107553620E-02_dp, 1.9243131938746824E+01_dp, -1.7572180196348960E+01_dp, &
         3.1911493525192705E+00_dp, 4.5301719814922450E+00_dp, -1.7901890264940798E+00_dp, &
         -1.8378320927758121E+00_dp, 1.0148404640536322E+00_dp, 7.9197328474954454E-01_dp, &
         -2.3336384070350974E-01_dp, -1.2451080784453024E+00_dp, 1.4494114850308317E+00_dp, &
         -6.8378647123057212E-01_dp, 1.2708688064825907E-01_dp, 3.3380032277175935E+01_dp, &
         -3.6241091255762910E+01_dp, 1.0285722844271358E+01_dp, 8.5886708142726427E+00_dp, &
         -5.5551083227948252E+00_dp, -3.1307221701362038E+00_dp, 3.0284928629647809E+00_dp, &
         1.3394612443539415E+00_dp, -1.4247249996215281E+00_dp, -1.6410029693292962E+00_dp, &
         2.7598440670059219E+00_dp, -1.4960510679975429E+00_dp, 3.0535228441060097E-01_dp, &
         5.8990537925489349E+01_dp, -7.5352687006393992E+01_dp, 2.9945186044685350E+01_dp, &
         1.4666282746247029E+01_dp, -1.5519933203166778E+01_dp, -3.9261174027125954E+00_dp, &
         8.0505809846827407E+00_dp, 1.5262322012801828E+00_dp, -5.1007207081578763E+00_dp, &
         -2.3367163638289626E-01_dp, 4.0544900581412744E+00_dp, -2.7636968296083531E+00_dp, &
         6.3238101915035738E-01_dp, 1.0645784500162654E+02_dp, -1.5877150896583035E+02_dp, &
         8.3344160108688257E+01_dp, 1.9521330509514783E+01_dp, -4.0311527122326183E+01_dp, &
         3.5169704418224251E-01_dp, 1.9265087220786128E+01_dp, -1.1490927641348434E+00_dp, &
         -1.4483864243908888E+01_dp, 8.3828197319727593E+00_dp, 2.1044231861397189E+00_dp, &
         -3.6152559380555696E+00_dp, 1.0333455730248924E+00_dp, 1.9673957337934047E+02_dp, &
         -3.4071102187005641E+02_dp, 2.2816253678927330E+02_dp, 3.5432086518209376E+00_dp, &
         -9.7344029351124050E+01_dp, 2.7012016585447444E+01_dp, 3.9849386666885657E+01_dp, &
         -1.5594715004093725E+01_dp, -3.3275570936735846E+01_dp, 3.7857227723125447E+01_dp, &
         -1.4266884105267037E+01_dp, 2.5646566664270998E-01_dp, 8.2958028001983131E-01_dp, &
         3.7380897364001447E+02_dp, -7.4843365384486981E+02_dp, 6.2534346604636403E+02_dp, &
         -1.2011555987584157E+02_dp, -2.0913981636077924E+02_dp, 1.3223097128254875E+02_dp, &
         5.8482694002845740E+01_dp, -6.7001391152085858E+01_dp, -5.2620142808837230E+01_dp, &
         1.1084495171472038E+02_dp, -7.3002889231094073E+01_dp, 2.2414543977585499E+01_dp, &
         -2.5763128135272004E+00_dp, 7.3590228308277256E+02_dp, -1.6937719308092217E+03_dp, &
         1.7365973342231957E+03_dp, -7.0191484123187968E+02_dp, -3.2398843711181541E+02_dp, &
         4.6233640099881148E+02_dp, -2.5325301041616040E+01_dp, -1.9579955506646129E+02_dp, &
         9.2621990919570578E+00_dp, 2.0559952760345703E+02_dp, -2.0101605399136827E+02_dp, &
         8.5917578479674788E+01_dp, -1.4858279542741453E+01_dp, 1.5386988002229782E+03_dp, &
         -4.0169102660006606E+03_dp, 4.9443246955768172E+03_dp, -3.0833947921566423E+03_dp, &
         2.0533986354945955E+02_dp, 1.1496536148736168E+03_dp, -6.0713368333607593E+02_dp, &
         -2.6680381334708136E+02_dp, 3.4114146837188082E+02_dp, 6.5931431069094316E+01_dp, &
         -2.5767149565086402E+02_dp, 1.5357507474959058E+02_dp, -3.2545640706779942E+01_dp, &
         3.8760158209907208E+03_dp, -1.1106072353977881E+04_dp, 1.5652437946556933E+04_dp, &
         -1.2734756310437266E+04_dp, 4.9118485439787391E+03_dp, 1.0865528975480108E+03_dp, &
         -2.0311041115258663E+03_dp, 2.9284227544796153E+02_dp, 7.0832752282331637E+02_dp, &
         -3.4845714151078687E+02_dp, -1.5188466316875682E+02_dp, 1.8252641460497227E+02_dp, &
         -4.7889305821155830E+01_dp, 1.8334270761363285E+00_dp, -3.9591024095020638E-01_dp, &
         -1.3126349325857345E-01_dp, 4.7222262371844306E-02_dp, 3.8076484409822732E-02_dp, &
         -1.6052628355116935E-02_dp, -1.5975826849070771E-02_dp, 5.5282813708112865E-03_dp, &
         1.2748899132651596E-02_dp, -1.0203081612794597E-02_dp, -7.6057568131617672E-05_dp, &
         2.7891576352124907E-03_dp, -8.9966226534393592E-04_dp, 2.2501471998076918E+00_dp, &
         -6.7293363442777321E-01_dp, -1.5991691179414610E-01_dp, 1.1671686002652161E-01_dp, &
         5.1331330187183055E-02_dp, -4.4152248495071197E-02_dp, -2.2121001215930056E-02_dp, &
         1.6471795598964997E-02_dp, 2.3995999348520584E-02_dp, -3.0506564606459299E-02_dp, &
         1.1201649086166804E-02_dp, 3.8654761046905383E-04_dp, -8.6532927608573725E-04_dp, &
         3.2106589696873193E+00_dp, -1.3748442557689724E+00_dp, -1.9499364840046871E-01_dp, &
         3.0218577867948093E-01_dp, 6.7411675514350539E-02_dp, -1.2176826893310169E-01_dp, &
         -2.8771344876437632E-02_dp, 4.7632499769699695E-02_dp, 4.9529748656412943E-02_dp, &
         -8.8817206722143702E-02_dp, 4.9444418187113308E-02_dp, -1.0655147393424186E-02_dp, &
         1.9439228710540853E-04_dp, 5.0066760706889388E+00_dp, -2.8639206962938242E+00_dp, &
         -1.6134501992474284E-01_dp, 7.1478932034735954E-01_dp, 4.8155111521357881E-02_dp, &
         -2.9955711646643696E-01_dp, -1.4766362863594271E-02_dp, 1.2140643553439820E-01_dp, &
         9.0761344780329142E-02_dp, -2.2640716820860976E-01_dp, 1.5602634912722926E-01_dp, &
         -4.8310196750658556E-02_dp, 5.3965878202100323E-03_dp, 8.1953446443739981E+00_dp, &
         -5.9031688012409029E+00_dp, 1.6445721911292407E-01_dp, 1.5757599912732250E+00_dp, &
         -1.2596820581475607E-01_dp, -6.7456338496403423E-01_dp, 8.6896320749570238E-02_dp, &
         2.8184553616038971E-01_dp, 1.2976825380286353E-01_dp, -5.1736963246432022E-01_dp, &
         4.2200984843870865E-01_dp, -1.5680233470759714E-01_dp, 2.3138933575817758E-02_dp, &
         1.3809992873162685E+01_dp, -1.2073777985207707E+01_dp, 1.4046415603947400E+00_dp, &
         3.2935120871963131E+00_dp, -7.9491410275893204E-01_dp, -1.4077655697256062E+00_dp, &
         4.6583962717584887E-01_dp, 6.0356328563664763E-01_dp, 7.2315826275806253E-02_dp, &
         -1.0589440789395266E+00_dp, 1.0229181937437832E+00_dp, -4.3382313133258776E-01_dp, &
         7.3703880601393990E-02_dp, 2.3775983411163235E+01_dp, -2.4678311578916002E+01_dp, &
         5.2289598442606424E+00_dp, 6.5489084608638590E+00_dp, -2.8658851271641992E+00_dp, &
         -2.6940912874542859E+00_dp, 1.6164260715043517E+00_dp, 1.1759602756276122E+00_dp, &
         -4.1631339246830373E-01_dp, -1.8583417620069387E+00_dp, 2.2261661929860863E+00_dp, &
         -1.0658414136189160E+00_dp, 2.0032087127389880E-01_dp, 4.1737161851743600E+01_dp, &
         -5.0724899954717543E+01_dp, 1.6048134070688963E+01_dp, 1.2233841848958997E+01_dp, &
         -8.6515035875208675E+00_dp, -4.4977564335736266E+00_dp, 4.7366008010596605E+00_dp, &
         1.9604112600157124E+00_dp, -2.3289929066425041E+00_dp, -2.3525147200130276E+00_dp, &
         4.1996293132940137E+00_dp, -2.3189605445392569E+00_dp, 4.7873798450047977E-01_dp, &
         7.4736777736519500E+01_dp, -1.0539402492295558E+02_dp, 4.5498177785332551E+01_dp, &
         2.0552859353942374E+01_dp, -2.3801828295887351E+01_dp, -5.3984340277393645E+00_dp, &
         1.2478979390566309E+01_dp, 2.1568019965817968E+00_dp, -8.2046138856525843E+00_dp, &
         1.3417335317994308E-01_dp, 6.0431791851769976E+00_dp, -4.2563589579645420E+00_dp, &
         9.8809406039481740E-01_dp, 1.3673546697572985E+02_dp, -2.2238025265186792E+02_dp, &
         1.2456773646258677E+02_dp, 2.6565789009467544E+01_dp, -6.1369218065318094E+01_dp, &
         1.5783873793130154E+00_dp, 2.9847304045089373E+01_dp, -2.0693059266805047E+00_dp, &
         -2.3361145261027204E+01_dp, 1.4410244118512024E+01_dp, 2.4894867024678495E+00_dp, &
         -5.4517863752830902E+00_dp, 1.6052091637823835E+00_dp, 2.5623223867233128E+02_dp, &
         -4.7860134044984517E+02_dp, 3.3730097004867315E+02_dp, 1.8218678998248234E+00_dp, &
         -1.4844387126254162E+02_dp, 4.2634582495319499E+01_dp, 6.2609487759576318E+01_dp, &
         -2.4599352208490920E+01_dp, -5.4743329555941450E+01_dp, 6.3294007287866073E+01_dp, &
         -2.4856415765702479E+01_dp, 1.1220010346374052E+00_dp, 1.2319782966587371E+00_dp, &
         4.9355841129397038E+02_dp, -1.0553521276769141E+03_dp, 9.1745816655949625E+02_dp, &
         -1.7734721772935907E+02_dp, -3.2570604320747339E+02_dp, 2.0420589402118949E+02_dp, &
         9.8255664197738639E+01_dp, -1.0685912026305078E+02_dp, -9.2991815283395937E+01_dp, &
         1.8938941289256240E+02_dp, -1.2451863070929338E+02_dp, 3.8321023072916731E+01_dp, &
         -4.4243955972078108E+00_dp, 9.8341942206132705E+02_dp, -2.3964944718903166E+03_dp, &
         2.5343558839934922E+03_dp, -1.0068167115408598E+03_dp, -5.5803706876145225E+02_dp, &
         7.3313067449087373E+02_dp, -6.6657041466680189E-01_dp, -3.3788668166577065E+02_dp, &
         -1.7628434198318345E+01_dp, 3.9655688909902926E+02_dp, -3.7408946589251713E+02_dp, &
         1.5787812785698458E+02_dp, -2.7089362134634474E+01_dp, 2.0614418378112468E+03_dp, &
         -5.6578757894510300E+03_dp, 7.1620850676558530E+03_dp, -4.4041308675995369E+03_dp, &
         -5.1684129400340517E+01_dp, 2.0625349133742861E+03_dp, -9.2575644960768261E+02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/-6.5703600920047631E+02_dp, &
         6.8515562434659739E+02_dp, 1.8173876061092901E+02_dp, -5.6990417291447488E+02_dp, &
         3.3635689037567329E+02_dp, -7.1542971828322536E+01_dp, 4.8609029109261082E+03_dp, &
         -1.4519102949367327E+04_dp, 2.1140422991455973E+04_dp, -1.7456530345889521E+04_dp, &
         6.1840046223145300E+03_dp, 2.8967239834451047E+03_dp, -4.1351692035042397E+03_dp, &
         6.5605620689472300E+02_dp, 1.7465620750833602E+03_dp, -1.4086350854964323E+03_dp, &
         2.9723222124582003E+02_dp, 1.1297881533760143E+02_dp, -5.1706127610453933E+01_dp, &
         1.8037684068839269E+04_dp, -5.4688608792211548E+04_dp, 8.3235695251795070E+04_dp, &
         -7.6979367072450463E+04_dp, 4.1185108453331864E+04_dp, -6.1794736925636562E+03_dp, &
         -6.9941711390058144E+03_dp, 3.2391970281109980E+03_dp, 1.9275654950388387E+03_dp, &
         -2.0014061535060721E+03_dp, 1.9033610113525435E+02_dp, 3.8478416754321637E+02_dp, &
         -1.3375337573784293E+02_dp, 2.5233269038969647E-01_dp, -2.6125055124024122E-02_dp, &
         -1.5445007406931326E-02_dp, 2.2557733632006770E-03_dp, 4.5182814756085403E-03_dp, &
         -1.7015400805549943E-04_dp, -2.8956549151850828E-03_dp, 3.4556093573692947E-04_dp, &
         2.2860774424317346E-03_dp, -1.6142483329644956E-03_dp, -2.0080309005498026E-05_dp, &
         4.1938671314358758E-04_dp, -1.3211727515346973E-04_dp, 8.2985625449170086E-01_dp, &
         -1.0224002952564909E-01_dp, -5.7664608710827639E-02_dp, 1.1974872702710729E-02_dp, &
         1.7932222830025043E-02_dp, -2.0242221071433827E-03_dp, -1.1683621416807838E-02_dp, &
         1.9133988199667928E-03_dp, 9.8961258386357637E-03_dp, -8.0196005442698740E-03_dp, &
         1.0252959799708511E-03_dp, 1.3039819801815705E-03_dp, -4.7833675979909446E-04_dp, &
         1.6553598506640155E+00_dp, -2.6704534598816121E-01_dp, -1.3976189982584905E-01_dp, &
         4.3195924781359557E-02_dp, 4.7199394458458561E-02_dp, -1.0781926573923930E-02_dp, &
         -3.1578308961048213E-02_dp, 7.3609046611193265E-03_dp, 2.9529666637054093E-02_dp, &
         -2.8085086647952978E-02_dp, 7.5639805082039629E-03_dp, 1.8106942675531527E-03_dp, &
         -1.0454801887820322E-03_dp, 3.0038146571211781E+00_dp, -6.4706656618816383E-01_dp, &
         -3.0986077711226090E-01_dp, 1.3471966868340191E-01_dp, 1.1282630726578560E-01_dp, &
         -4.1339044226723645E-02_dp, -7.7965682549816187E-02_dp, 2.4795390028214060E-02_dp, &
         8.1004935022273516E-02_dp, -8.8852693289714887E-02_dp, 3.3669485741906306E-02_dp, &
         -1.0427673934340815E-03_dp, -1.7583837649226093E-03_dp, 5.3661110332212125E+00_dp, &
         -1.5177683390219008E+00_dp, -6.5603757824139586E-01_dp, 3.8500786992205382E-01_dp, &
         2.5336092688609296E-01_dp, -1.3567978709745959E-01_dp, -1.8192450564210702E-01_dp, &
         7.7309535643977162E-02_dp, 2.1128094777967329E-01_dp, -2.6436364174291122E-01_dp, &
         1.2387404838719954E-01_dp, -1.9643916427159049E-02_dp, -1.4373481356760645E-03_dp, &
         9.6763664742460129E+00_dp, -3.5034680326316079E+00_dp, -1.3384736160499082E+00_dp, &
         1.0464400018507101E+00_dp, 5.3683599645419833E-01_dp, -4.1063751344235755E-01_dp, &
         -4.0422596672493061E-01_dp, 2.3132613484214665E-01_dp, 5.3106145549846739E-01_dp, &
         -7.5794244173349756E-01_dp, 4.1567726800030025E-01_dp, -9.8798754939939740E-02_dp, &
         5.6616607065354827E-03_dp, 1.7788408183604410E+01_dp, -8.0593752690254270E+00_dp, &
         -2.6302227053318825E+00_dp, 2.7712044719380975E+00_dp, 1.0604974174861457E+00_dp, &
         -1.1889029862167542E+00_dp, -8.5146872301255649E-01_dp, 6.8030468396528943E-01_dp, &
         1.2912221791895706E+00_dp, -2.1240738522644662E+00_dp, 1.3266071942599185E+00_dp, &
         -3.9109485214509854E-01_dp, 4.2305366773465025E-02_dp, 3.3518413862389274E+01_dp, &
         -1.8679186314635402E+01_dp, -4.9162127064446723E+00_dp, 7.2666289813213654E+00_dp, &
         1.8606253476611085E+00_dp, -3.3597721710782356E+00_dp, -1.6564498861825205E+00_dp, &
         1.9990348858286913E+00_dp, 3.0053373009755533E+00_dp, -5.8494608250533311E+00_dp, &
         4.1063242486654916E+00_dp, -1.3990589490178600E+00_dp, 1.9147802002843936E-01_dp, &
         6.5026247054793515E+01_dp, -4.4052475259993862E+01_dp, -8.3815680918103901E+00_dp, &
         1.9083151743714176E+01_dp, 2.3780161434136025E+00_dp, -9.3704244036259556E+00_dp, &
         -2.7194194265982845E+00_dp, 5.9402535323345900E+00_dp, 6.4281786309828632E+00_dp, &
         -1.5756775907653372E+01_dp, 1.2412062950466199E+01_dp, -4.7334254279650443E+00_dp, &
         7.4159825374866106E-01_dp, 1.3052226392818551E+02_dp, -1.0674760116218118E+02_dp, &
         -1.1161921902585187E+01_dp, 5.0609207068618112E+01_dp, -1.0533924589029899E+00_dp, &
         -2.5890748254175101E+01_dp, -2.2676107030737405E+00_dp, 1.7997385078036434E+01_dp, &
         1.0898308868992313E+01_dp, -4.0625202454028170E+01_dp, 3.6416772710287425E+01_dp, &
         -1.5358584568725442E+01_dp, 2.6525693124594736E+00_dp, 2.7273422484799687E+02_dp, &
         -2.6858858523224052E+02_dp, -4.7885740801452348E-01_dp, 1.3627417330611453E+02_dp, &
         -2.6358992902829311E+01_dp, -7.0393695696082460E+01_dp, 9.9494808404702209E+00_dp, &
         5.5679648303730211E+01_dp, 2.3191677764627250E+00_dp, -9.3785981251887762E+01_dp, &
         1.0074347646251888E+02_dp, -4.7292967213979878E+01_dp, 8.8974333107109391E+00_dp, &
         5.9814256543256761E+02_dp, -7.1057805327339554E+02_dp, 8.8601488863604075E+01_dp, &
         3.7264206640783578E+02_dp, -1.5551496835779633E+02_dp, -1.8230942230852452E+02_dp, &
         8.3174041344709863E+01_dp, 1.7313461443852231E+02_dp, -1.1433889332027171E+02_dp, &
         -1.4732698175357447E+02_dp, 2.3885411412899521E+02_dp, -1.3081884815432261E+02_dp, &
         2.7131160357794776E+01_dp, 1.3926535234572304E+03_dp, -2.0093649803688027E+03_dp, &
         5.7417596311169530E+02_dp, 1.0197340958636103E+03_dp, -7.6503574679237079E+02_dp, &
         -3.9341161249586122E+02_dp, 4.3789084981339414E+02_dp, 5.0339035381666918E+02_dp, &
         -8.3869418930766665E+02_dp, 2.3396482300537855E+02_dp, 2.9455554478541080E+02_dp, &
         -2.6058293586631680E+02_dp, 6.5188966276420842E+01_dp, 3.5014744519247643E+03_dp, &
         -6.2163991386173120E+03_dp, 3.1151468986944483E+03_dp, 2.5849457928935071E+03_dp, &
         -3.6379187148580500E+03_dp, -1.0371922354864618E+02_dp, 1.9541413742804682E+03_dp, &
         9.4382928584982380E+02_dp, -4.1940808589968328E+03_dp, 3.9117942893214040E+03_dp, &
         -1.6320017495377126E+03_dp, 2.1892754217913259E+02_dp, 2.8366828833340946E+01_dp, &
         9.7875448079300859E+03_dp, -2.1801596144554653E+04_dp, 1.7647392751717292E+04_dp, &
         3.0182497292630583E+03_dp, -1.6860787344740485E+04_dp, 8.9519787903568704E+03_dp, &
         5.5338276194636665E+03_dp, -4.2835976536269254E+03_dp, -1.0643234050261457E+04_dp, &
         1.9747629350798474E+04_dp, -1.5091225398714651E+04_dp, 5.9161713721962315E+03_dp, &
         -9.8289027954559174E+02_dp, 3.3196247063191586E+04_dp, -9.3274647246994995E+04_dp, &
         1.1696367892260675E+05_dp, -5.7983849396112921E+04_dp, -4.0430038391972135E+04_dp, &
         8.8668357693324142E+04_dp, -5.8371126061495066E+04_dp, 6.0751284512953389E+03_dp, &
         1.0622268134494479E+04_dp, 3.2655837179587106E+03_dp, -1.2844528875114578E+04_dp, &
         8.2097522311486537E+03_dp, -1.8311415720327325E+03_dp, 2.0653637422040756E+00_dp, &
         -2.2384945891375232E-01_dp, -1.3073607242437763E-01_dp, 2.1162136800823701E-02_dp, &
         3.8890411553423945E-02_dp, -2.2306624268842003E-03_dp, -2.5029344308511840E-02_dp, &
         3.2754819697756528E-03_dp, 2.0135897811441424E-02_dp, -1.4787158586580834E-02_dp, &
         4.3510596001332895E-04_dp, 3.4129326945082478E-03_dp, -1.1123282044359531E-03_dp, &
         2.6671755379167270E+00_dp, -4.2558997726367698E-01_dp, -2.2518643634186017E-01_dp, &
         6.6476925859854485E-02_dp, 7.5670871490107108E-02_dp, -1.5785369201729928E-02_dp, &
         -5.0359789910624432E-02_dp, 1.1050039154106801E-02_dp, 4.6262019696085968E-02_dp, &
         -4.2730599514122225E-02_dp, 1.0454611624572414E-02_dp, 3.4881697628749296E-03_dp, &
         -1.7584705205137370E-03_dp, 4.1130012975486210E+00_dp, -9.7197400656395661E-01_dp, &
         -4.6237399676090191E-01_dp, 2.0698158854957260E-01_dp, 1.7154740439363861E-01_dp, &
         -6.3194136479580720E-02_dp, -1.1852187781079761E-01_dp, 3.7373697623651093E-02_dp, &
         1.2303837057702460E-01_dp, -1.3406730347811235E-01_dp, 5.0073347721850299E-02_dp, &
         -1.0526296887983370E-03_dp, -2.7741605641166435E-03_dp, 6.9879106304886687E+00_dp, &
         -2.2398959339089957E+00_dp, -9.5500233539439394E-01_dp, 5.8532810178543637E-01_dp, &
         3.7832389585154519E-01_dp, -2.0649380273593768E-01_dp, -2.7235995790339301E-01_dp, &
         1.1589245735429003E-01_dp, 3.1782510260520691E-01_dp, -3.9666849007623201E-01_dp, &
         1.8494292684281363E-01_dp, -2.8775671938967632E-02_dp, -2.3253247744458075E-03_dp, &
         1.2504091216408819E+01_dp, -5.1133715337985244E+00_dp, -1.9181898554326646E+00_dp, &
         1.5709030401577699E+00_dp, 7.9081643367877807E-01_dp, -6.1856497603188698E-01_dp, &
         -5.9852994860204733E-01_dp, 3.4380179574626718E-01_dp, 7.9330856385958726E-01_dp, &
         -1.1300425774704193E+00_dp, 6.1780772258320593E-01_dp, -1.4576277626532783E-01_dp, &
         8.0559421794824965E-03_dp, 2.3150699771274905E+01_dp, -1.1671570206269971E+01_dp, &
         -3.7216455982735082E+00_dp, 4.1174047756097680E+00_dp, 1.5454379483839249E+00_dp, &
         -1.7754265965038893E+00_dp, -1.2509441410240485E+00_dp, 1.0031593326500263E+00_dp, &
         1.9232600826293296E+00_dp, -3.1538645783335792E+00_dp, 1.9636054804026819E+00_dp, &
         -5.7580937729700632E-01_dp, 6.1579345420484835E-02_dp, 4.4147095393639411E+01_dp, &
         -2.6922086577205693E+01_dp, -6.8815493972125763E+00_dp, 1.0716119799573967E+01_dp, &
         2.6933147969167028E+00_dp, -4.9862251392598829E+00_dp, -2.4265145944659903E+00_dp, &
         2.9279766729955137E+00_dp, 4.4935316525485280E+00_dp, -8.6797598171501242E+00_dp, &
         6.0658299107057587E+00_dp, -2.0551305190725611E+00_dp, 2.7895387328935622E-01_dp, &
         8.6795469321420356E+01_dp, -6.3334977146765148E+01_dp, -1.1616855764666131E+01_dp, &
         2.7995473727190131E+01_dp, 3.4578165687230404E+00_dp, -1.3855839284346782E+01_dp, &
         -4.0148856187294424E+00_dp, 8.6500178158021424E+00_dp, 9.7804892449733156E+00_dp, &
         -2.3485823007059793E+01_dp, 1.8360707367435733E+01_dp, -6.9529236408812043E+00_dp, &
         1.0806322167710474E+00_dp, 1.7656622530201321E+02_dp, -1.5328123200897573E+02_dp, &
         -1.5291064937327004E+01_dp, 7.3985834147322677E+01_dp, -1.2152988380722913E+00_dp, &
         -3.8270202581466002E+01_dp, -3.6377050794632200E+00_dp, 2.6084885989267164E+01_dp, &
         1.7676198736610498E+01_dp, -6.1439500469950268E+01_dp, 5.4294556266848126E+01_dp, &
         -2.2673187037308736E+01_dp, 3.8805401802065420E+00_dp, 3.7365419495324414E+02_dp, &
         -3.8515598503602843E+02_dp, -4.0361864625047167E-01_dp, 1.9885938187600482E+02_dp, &
         -3.5770781805075885E+01_dp, -1.0466970552404520E+02_dp, 1.2589120977813629E+01_dp, &
         8.0595793051374187E+01_dp, 1.1782181052160290E+01_dp, -1.4784344384024041E+02_dp, &
         1.5365187663766355E+02_dp, -7.0959472033717248E+01_dp, 1.3190525000857415E+01_dp, &
         8.2863322469379375E+02_dp, -1.0157657443516359E+03_dp, 1.1986988187949704E+02_dp, &
         5.4449676449347032E+02_dp, -2.1036288129400012E+02_dp, -2.7738557943311429E+02_dp, &
         1.1136872225739332E+02_dp, 2.5324789577013900E+02_dp, -1.2776987419497914E+02_dp, &
         -2.7397753121367208E+02_dp, 3.8961772786829607E+02_dp, -2.0551276375773824E+02_dp, &
         4.1739997168171449E+01_dp, 1.9450907147920134E+03_dp, -2.8508892351002860E+03_dp, &
         7.5747523129342949E+02_dp, 1.5070567428821669E+03_dp, -1.0217458181146775E+03_dp, &
         -6.5694865280062390E+02_dp, 5.9368339749825054E+02_dp, 7.7692645932202890E+02_dp, &
         -1.0697937873179803E+03_dp, 3.8548984201540087E+01_dp, 6.7608544747530505E+02_dp, &
         -4.8167622236142665E+02_dp, 1.1263022376557032E+02_dp, 4.9024755081192952E+03_dp, &
         -8.6785968466350550E+03_dp, 3.9572621953771327E+03_dp, 4.0485213527526844E+03_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c04 = (/-4.7874743699806668E+03_dp, &
         -8.5329386064550408E+02_dp, 2.7526945075798449E+03_dp, 1.9553980517436300E+03_dp, &
         -5.9289564226732664E+03_dp, 4.4771348003328458E+03_dp, -1.0901273950898408E+03_dp, &
         -3.0130386512273043E+02_dp, 1.6238769196873562E+02_dp, 1.3563319365162426E+04_dp, &
         -2.9459071556804978E+04_dp, 2.1056290292594003E+04_dp, 8.3902840919604496E+03_dp, &
         -2.2673291511349471E+04_dp, 6.3284247944587814E+03_dp, 1.0923165055552865E+04_dp, &
         -4.1277480924433718E+02_dp, -2.4024037869185162E+04_dp, 3.2150151957564813E+04_dp, &
         -2.0428697339350008E+04_dp, 6.7276912202797048E+03_dp, -9.1758726122739427E+02_dp, &
         4.2980689084572172E+04_dp, -1.1637801167987248E+05_dp, 1.2686014747328563E+05_dp, &
         -2.3138864168402757E+04_dp, -9.8492147420019101E+04_dp, 1.0071830242916004E+05_dp, &
         3.2911879605145018E+03_dp, -6.4873615309764507E+04_dp, 9.6403278074733826E+03_dp, &
         7.4571802297396542E+04_dp, -8.6979131802506905E+04_dp, 4.3117461760700382E+04_dp, &
         -8.5450772808429192E+03_dp, 1.8880572840095649E+05_dp, -6.0818896836644155E+05_dp, &
         9.4238894917301007E+05_dp, -7.8416009400037490E+05_dp, 1.3237171408866072E+05_dp, &
         5.4779421601664904E+05_dp, -8.1171169360100222E+05_dp, 6.4638916088477266E+05_dp, &
         -3.4673780482035381E+05_dp, 1.4009010473194942E+05_dp, -4.9102063709230773E+04_dp, &
         1.4876957124327950E+04_dp, -2.5652831322436327E+03_dp, 2.6121103454662853E-01_dp, &
         -9.2614016920939818E-03_dp, -8.3000893742049778E-03_dp, 2.9074778484552504E-03_dp, &
         -3.0144050099379580E-03_dp, 4.9532380283510919E-03_dp, -5.6871174776962855E-03_dp, &
         5.6833144062626998E-03_dp, -5.4644435872497057E-03_dp, 4.5424442498369442E-03_dp, &
         -2.8087470813445226E-03_dp, 1.1070902641492602E-03_dp, -2.0646299963951532E-04_dp, &
         8.6492801066328462E-01_dp, -3.6931303166526513E-02_dp, -3.2730494983751128E-02_dp, &
         1.2246168926080620E-02_dp, -1.1973281160824956E-02_dp, 1.9775557228632324E-02_dp, &
         -2.2824439482019589E-02_dp, 2.2709199342063304E-02_dp, -2.1745218503398112E-02_dp, &
         1.8069693810495297E-02_dp, -1.1180161722380632E-02_dp, 4.4065971039600003E-03_dp, &
         -8.2110935783146887E-04_dp, 1.7482557856955379E+00_dp, -9.9194016925639519E-02_dp, &
         -8.6451169814672935E-02_dp, 3.5464507882586600E-02_dp, -3.2021249054978267E-02_dp, &
         5.3185104302619933E-02_dp, -6.1882498472844098E-02_dp, 6.1184989135095555E-02_dp, &
         -5.8218328328857059E-02_dp, 4.8348235667150061E-02_dp, -2.9946177814150546E-02_dp, &
         1.1804489751035784E-02_dp, -2.1971644622055773E-03_dp, 3.2323943569806324E+00_dp, &
         -2.4780428150533954E-01_dp, -2.1198615319295999E-01_dp, 9.5537150881642635E-02_dp, &
         -7.9829124976059013E-02_dp, 1.3295600640545427E-01_dp, -1.5615621853217190E-01_dp, &
         1.5340084396316397E-01_dp, -1.4489460259257397E-01_dp, 1.2023015115955925E-01_dp, &
         -7.4572268679463607E-02_dp, 2.9408259186661307E-02_dp, -5.4683663834212944E-03_dp, &
         5.9110512696058004E+00_dp, -6.0023777314923710E-01_dp, -5.0327168989784998E-01_dp, &
         2.4893207402559139E-01_dp, -1.9361685019651018E-01_dp, 3.2197727902061518E-01_dp, &
         -3.8217882779932483E-01_dp, 3.7308039130993181E-01_dp, -3.4946309083981375E-01_dp, &
         2.8964679258249543E-01_dp, -1.7996777762962227E-01_dp, 7.1034404853642361E-02_dp, &
         -1.3199465528400126E-02_dp, 1.0956673224198525E+01_dp, -1.4346595057416782E+00_dp, &
         -1.1762540800120738E+00_dp, 6.3994506883957525E-01_dp, -4.6541568109437514E-01_dp, &
         7.6857524192355064E-01_dp, -9.2344838294545883E-01_dp, 8.9595278289335112E-01_dp, &
         -8.3103399203641393E-01_dp, 6.8765431279190914E-01_dp, -4.2822296013721550E-01_dp, &
         1.6927920220984846E-01_dp, -3.1446728843768625E-02_dp, 2.0792596884134536E+01_dp, &
         -3.4314401576726481E+00_dp, -2.7415334111503054E+00_dp, 1.6495658618315001E+00_dp, &
         -1.1261669765784517E+00_dp, 1.8334764395621579E+00_dp, -2.2345254604368661E+00_dp, &
         2.1552250426936013E+00_dp, -1.9753669693004268E+00_dp, 1.6304657176257347E+00_dp, &
         -1.0182640362604070E+00_dp, 4.0351035593520562E-01_dp, -7.4988906494427660E-02_dp, &
         4.0641614403913103E+01_dp, -8.3157277744128777E+00_dp, -6.4427715366084595E+00_dp, &
         4.3229852088919545E+00_dp, -2.7837340360875835E+00_dp, 4.4241729410937412E+00_dp, &
         -5.4829090349509961E+00_dp, 5.2597965380569276E+00_dp, -4.7499578296245764E+00_dp, &
         3.9051365477525373E+00_dp, -2.4477687477727605E+00_dp, 9.7366758168919942E-01_dp, &
         -1.8121195058941905E-01_dp, 8.2279753709829350E+01_dp, -2.0658726903050777E+01_dp, &
         -1.5417087397668558E+01_dp, 1.1667289558374994E+01_dp, -7.1366807342421215E+00_dp, &
         1.0920441396316519E+01_dp, -1.3801231712035245E+01_dp, 1.3182065256871045E+01_dp, &
         -1.1682971218060507E+01_dp, 9.5429009419005446E+00_dp, -6.0084889711222393E+00_dp, &
         2.4038519801319982E+00_dp, -4.4882474577803322E-01_dp, 1.7368668276644541E+02_dp, &
         -5.3256212659671924E+01_dp, -3.7915111660268572E+01_dp, 3.2864991327369893E+01_dp, &
         -1.9307015368211516E+01_dp, 2.7897233436619405E+01_dp, -3.6056588333988557E+01_dp, &
         3.4359603001317005E+01_dp, -2.9723444391223833E+01_dp, 2.4009950934101433E+01_dp, &
         -1.5194650895871973E+01_dp, 6.1315866703523074E+00_dp, -1.1517999700175321E+00_dp, &
         3.8567284052451515E+02_dp, -1.4449356721851993E+02_dp, -9.6726585251915836E+01_dp, &
         9.8143836449527910E+01_dp, -5.6333493633577575E+01_dp, 7.4782940357757127E+01_dp, &
         -9.9046173142738638E+01_dp, 9.4524979091637363E+01_dp, -7.9209350305708099E+01_dp, &
         6.2708234355782480E+01_dp, -3.9861447795304208E+01_dp, 1.6288701950845873E+01_dp, &
         -3.0927002635391019E+00_dp, 9.1208802703023264E+02_dp, -4.2025964940054922E+02_dp, &
         -2.5838953620053741E+02_dp, 3.1708958164778818E+02_dp, -1.8269319770114419E+02_dp, &
         2.1443864190015620E+02_dp, -2.9053026229725748E+02_dp, 2.7958028873552331E+02_dp, &
         -2.2455334922664062E+02_dp, 1.7097859906654173E+02_dp, -1.0856602341332952E+02_dp, &
         4.5132022374191045E+01_dp, -8.7247003764599960E+00_dp, 2.3403906153903949E+03_dp, &
         -1.3455186670211949E+03_dp, -7.2807901502605125E+02_dp, 1.1411003502760752E+03_dp, &
         -6.8879925865566565E+02_dp, 6.7936705271442918E+02_dp, -9.2766071574488012E+02_dp, &
         9.1080489850452716E+02_dp, -6.9147508025247521E+02_dp, 4.8478972066941219E+02_dp, &
         -2.9990358024065625E+02_dp, 1.2694128276053435E+02_dp, -2.5248397806722892E+01_dp, &
         6.7163110511762943E+03_dp, -4.9502119609359797E+03_dp, -2.1460672665525426E+03_dp, &
         4.7894312825043899E+03_dp, -3.2493183252210265E+03_dp, 2.5513263672070620E+03_dp, &
         -3.3004059244053074E+03_dp, 3.3620235704706211E+03_dp, -2.3803057066797624E+03_dp, &
         1.3701475041710678E+03_dp, -7.2715726697134289E+02_dp, 3.0023731366800985E+02_dp, &
         -6.1929348822443536E+01_dp, 2.2806657666318617E+04_dp, -2.2697692191739989E+04_dp, &
         -5.7569649161039997E+03_dp, 2.5430927220737100E+04_dp, -2.2037671719779755E+04_dp, &
         1.3975577910469839E+04_dp, -1.3428763146233510E+04_dp, 1.3846589383992001E+04_dp, &
         -9.1032979373949202E+03_dp, 2.8138129998584268E+03_dp, 3.1040518294611104E+02_dp, &
         -5.4082375860751756E+02_dp, 1.3353967691905902E+02_dp, 1.0514267073816214E+05_dp, &
         -1.5728134173123082E+05_dp, 2.3393458473274026E+04_dp, 1.9191748448950576E+05_dp, &
         -2.9755239107713493E+05_dp, 2.4052918017103928E+05_dp, -1.0595433575621128E+05_dp, &
         -1.7465087519767349E+04_dp, 9.6696437844672575E+04_dp, -1.2300599526839080E+05_dp, &
         9.6573441236828832E+04_dp, -4.5315525505061276E+04_dp, 9.6777412850281053E+03_dp, &
         2.1416242820167533E+00_dp, -7.9748902326817347E-02_dp, -7.1261894234958337E-02_dp, &
         2.5408305090802094E-02_dp, -2.5927397116631480E-02_dp, 4.2666546644005751E-02_dp, &
         -4.9054480275301224E-02_dp, 4.8964284685653690E-02_dp, -4.7028426616405591E-02_dp, &
         3.9089916076761162E-02_dp, -2.4174446086355380E-02_dp, 9.5283680688875664E-03_dp, &
         -1.7765574154452902E-03_dp, 2.8149194258528634E+00_dp, -1.5742724721103563E-01_dp, &
         -1.3756798805731643E-01_dp, 5.5658044121983691E-02_dp, -5.0812419090658374E-02_dp, &
         8.4411296062409852E-02_dp, -9.8069441497070670E-02_dp, 9.7050981626127075E-02_dp, &
         -9.2449750838156014E-02_dp, 7.6786841095660932E-02_dp, -4.7549390881222578E-02_dp, &
         1.8741467289636684E-02_dp, -3.4887110096130883E-03_dp, 4.4566699723520529E+00_dp, &
         -3.7284061867202412E-01_dp, -3.1873352555549722E-01_dp, 1.4418917656863958E-01_dp, &
         -1.1991661862767018E-01_dp, 2.0013354404185627E-01_dp, -2.3504848379348631E-01_dp, &
         2.3080624953499043E-01_dp, -2.1800504811681376E-01_dp, 1.8090710775770399E-01_dp, &
         -1.1219870258910787E-01_dp, 4.4239450395016547E-02_dp, -8.2247162512252506E-03_dp, &
         7.7935726447770772E+00_dp, -8.8875668865490143E-01_dp, -7.4396707707618104E-01_dp, &
         3.7089303027758863E-01_dp, -2.8613997455148843E-01_dp, 4.7698560347380187E-01_dp, &
         -5.6641616408552975E-01_dp, 5.5250474842806918E-01_dp, -5.1735291166967712E-01_dp, &
         4.2882674642307389E-01_dp, -2.6644201552357072E-01_dp, 1.0514781473582199E-01_dp, &
         -1.9533141443740406E-02_dp, 1.4376697262131565E+01_dp, -2.1020595111518539E+00_dp, &
         -1.7200884547239541E+00_dp, 9.4395322212651767E-01_dp, -6.8061166413961394E-01_dp, &
         1.1266400272816803E+00_dp, -1.3545508846472831E+00_dp, 1.3130363890239167E+00_dp, &
         -1.2173504455983082E+00_dp, 1.0074175588607479E+00_dp, -6.2736593117696249E-01_dp, &
         2.4795521216282801E-01_dp, -4.6047734204784939E-02_dp, 2.7511316014784537E+01_dp, &
         -4.9898964170307982E+00_dp, -3.9785547201608060E+00_dp, 2.4140746181922275E+00_dp, &
         -1.6342762543947815E+00_dp, 2.6673110732257239E+00_dp, -3.2532834256468250E+00_dp, &
         3.1347513281824200E+00_dp, -2.8718563783342388E+00_dp, 2.3708595750580499E+00_dp, &
         -1.4807456214225743E+00_dp, 5.8665018395297608E-01_dp, -1.0898260449107275E-01_dp, &
         5.4436762909174121E+01_dp, -1.2032475553577383E+01_dp, -9.3058204956370005E+00_dp, &
         6.2880148178096746E+00_dp, -4.0175609932092797E+00_dp, 6.4042779737722988E+00_dp, &
         -7.9429695413364838E+00_dp, 7.6117413807947472E+00_dp, -6.8720650528250387E+00_dp, &
         5.6519758743547346E+00_dp, -3.5429970722615485E+00_dp, 1.4088855536325260E+00_dp, &
         -2.6208092423393642E-01_dp, 1.1164736368007399E+02_dp, -2.9791384407380139E+01_dp, &
         -2.2212377539439530E+01_dp, 1.6877037072718817E+01_dp, -1.0250572296995630E+01_dp, &
         1.5756255142999594E+01_dp, -1.9924053783286329E+01_dp, 1.9007981150929719E+01_dp, &
         -1.6850401908340178E+01_dp, 1.3775589300147089E+01_dp, -8.6745222956516148E+00_dp, &
         3.4686490862588455E+00_dp, -6.4714752047628832E-01_dp, 2.3860418477390155E+02_dp, &
         -7.6548665940035832E+01_dp, -5.4555576502769235E+01_dp, 4.7223564660740450E+01_dp, &
         -2.7552539514865845E+01_dp, 4.0129145306232964E+01_dp, -5.1876753274658711E+01_dp, &
         4.9362356501653707E+01_dp, -4.2758141683395408E+01_dp, 3.4607643058962353E+01_dp, &
         -2.1907296486378485E+01_dp, 8.8320535955436519E+00_dp, -1.6569481566503010E+00_dp, &
         5.3545799381479731E+02_dp, -2.0661914588910187E+02_dp, -1.3901714342570125E+02_dp, &
         1.3960064752137760E+02_dp, -7.9455954526732498E+01_dp, 1.0705506450207533E+02_dp, &
         -1.4178741954776285E+02_dp, 1.3500627788698125E+02_dp, -1.1350858118917807E+02_dp, &
         9.0286381762194210E+01_dp, -5.7449961163831816E+01_dp, 2.3438033636020098E+01_dp, &
         -4.4397143591330375E+00_dp, 1.2754605500542539E+03_dp, -5.9491668930678827E+02_dp, &
         -3.7066724475406403E+02_dp, 4.4339421427221095E+02_dp, -2.5185768559746347E+02_dp, &
         3.0387188886435422E+02_dp, -4.1221355071839258E+02_dp, 3.9500349025307065E+02_dp, &
         -3.1942180687176204E+02_dp, 2.4606495659517455E+02_dp, -1.5690263401286234E+02_dp, &
         6.5084151112193439E+01_dp, -1.2528409779731549E+01_dp, 3.2752436692089077E+03_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c05 = (/-1.8663851312283325E+03_dp, &
         -1.0423380612707174E+03_dp, 1.5483806206375675E+03_dp, -9.0806872083263704E+02_dp, &
         9.4017350130709883E+02_dp, -1.2957349086372242E+03_dp, 1.2616460034642841E+03_dp, &
         -9.6943667338306091E+02_dp, 7.0087846283955878E+02_dp, -4.4162263728305550E+02_dp, &
         1.8703907422642797E+02_dp, -3.6951030936082461E+01_dp, 9.2847122827431940E+03_dp, &
         -6.5872616019199959E+03_dp, -3.0955142203029591E+03_dp, 6.1516693509907564E+03_dp, &
         -3.9218013076513857E+03_dp, 3.3172151564545989E+03_dp, -4.4931295716623717E+03_dp, &
         4.5137348900129173E+03_dp, -3.2498919726863614E+03_dp, 2.0440411940202880E+03_dp, &
         -1.1971496951864885E+03_dp, 5.1322722240655889E+02_dp, -1.0554920856973189E+02_dp, &
         3.0235039579578988E+04_dp, -2.7601902396650934E+04_dp, -9.3215803184808592E+03_dp, &
         2.9388322193901040E+04_dp, -2.2102684178004485E+04_dp, 1.4848401915745148E+04_dp, &
         -1.7579341720367716E+04_dp, 1.8492406134381723E+04_dp, -1.2397481070128109E+04_dp, &
         5.4498033753094114E+03_dp, -1.8981227282010670E+03_dp, 6.4138733136082431E+02_dp, &
         -1.3956433688914467E+02_dp, 1.2216150610526593E+05_dp, -1.5253694884673052E+05_dp, &
         -1.6429729416520167E+04_dp, 1.8607958595858968E+05_dp, -1.9196043087545509E+05_dp, &
         1.1475537459925326E+05_dp, -7.9761919488329018E+04_dp, 7.6847623039708662E+04_dp, &
         -4.7002818213517261E+04_dp, -4.9407381312193525E+02_dp, 2.2584965140133190E+04_dp, &
         -1.5141091047763175E+04_dp, 3.5667736604964957E+03_dp, 7.5406264750572247E+05_dp, &
         -1.4326299671747899E+06_dp, 6.2617574112644594E+05_dp, 1.6499953681879421E+06_dp, &
         -3.7046203068710547E+06_dp, 3.8236409013257986E+06_dp, -1.7734757970713889E+06_dp, &
         -1.1979318870654979E+06_dp, 3.4208098012167024E+06_dp, -3.8751620860478538E+06_dp, &
         2.7423423745086947E+06_dp, -1.1761698637398595E+06_dp, 2.3420810572297117E+05_dp, &
         2.6324296748844861E-01_dp, 2.7730709344764889E-09_dp, -1.2207869322742759E-02_dp, &
         1.4895462150693508E-02_dp, -1.8697336568206407E-02_dp, 1.4407704487284264E-02_dp, &
         9.8927256082207531E-03_dp, -6.4832137218925986E-02_dp, 1.4426877307723651E-01_dp, &
         -2.0783702334434193E-01_dp, 1.9989272270072017E-01_dp, -1.1694483602928235E-01_dp, &
         3.1444827999764245E-02_dp, 8.7304841863481009E-01_dp, 1.1103688365004590E-08_dp, &
         -4.8894552708827893E-02_dp, 5.9658791963436070E-02_dp, -7.4239686035906957E-02_dp, &
         5.6121910416534672E-02_dp, 4.2622494290017153E-02_dp, -2.6407945174133429E-01_dp, &
         5.8267974366545527E-01_dp, -8.3609588478556940E-01_dp, 8.0215794742252378E-01_dp, &
         -4.6848952780524716E-01_dp, 1.2581178648852193E-01_dp, 1.7701371812793694E+00_dp, &
         3.0005295459047330E-08_dp, -1.3217904746214579E-01_dp, 1.6127847054189673E-01_dp, &
         -1.9811745819523222E-01_dp, 1.4539987789906042E-01_dp, 1.2715376214627466E-01_dp, &
         -7.3137864438796607E-01_dp, 1.5942745621335248E+00_dp, -2.2744691502925121E+00_dp, &
         2.1742633873086681E+00_dp, -1.2666391454859651E+00_dp, 3.3951831663787274E-01_dp, &
         3.2872518940714639E+00_dp, 7.5457088814963605E-08_dp, -3.3254988508622568E-01_dp, &
         4.0576105221431225E-01_dp, -4.9135955091474298E-01_dp, 3.4845147454343889E-01_dp, &
         3.5251584271462333E-01_dp, -1.8875150259828974E+00_dp, 4.0622047380355637E+00_dp, &
         -5.7594642621988239E+00_dp, 5.4840978290474425E+00_dp, -3.1859681965527034E+00_dp, &
         8.5223243994700304E-01_dp, 6.0444281533037323E+00_dp, 1.8405860589428224E-07_dp, &
         -8.1156394914969088E-01_dp, 9.9022995646443068E-01_dp, -1.1807985506013581E+00_dp, &
         8.0545436367079470E-01_dp, 9.4407597393184040E-01_dp, -4.7270998413371172E+00_dp, &
         1.0041641519313384E+01_dp, -1.4145249627502430E+01_dp, 1.3413036931983747E+01_dp, &
         -7.7692521878901104E+00_dp, 2.0736589380956820E+00_dp, 1.1276776635033665E+01_dp, &
         4.4330328247789032E-07_dp, -1.9557137822199457E+00_dp, 2.3862631597335069E+00_dp, &
         -2.7970925554909138E+00_dp, 1.8223553162021626E+00_dp, 2.4945994656705284E+00_dp, &
         -1.1704415746788525E+01_dp, 2.4524089594280941E+01_dp, -3.4305229650253693E+01_dp, &
         3.2381675186685143E+01_dp, -1.8695316093989742E+01_dp, 4.9776584290105079E+00_dp, &
         2.1561789642603674E+01_dp, 1.0694691683032124E-06_dp, -4.7211887146110669E+00_dp, &
         5.7605517287947574E+00_dp, -6.6200439159302640E+00_dp, 4.0750308632033789E+00_dp, &
         6.6167943654334040E+00_dp, -2.9092299880744793E+01_dp, 6.0053394751935848E+01_dp, &
         -8.3351989681794308E+01_dp, 7.8273643558272752E+01_dp, -4.5022143326781794E+01_dp, &
         1.1953365515829327E+01_dp, 4.2515726163437314E+01_dp, 2.6176175888860676E-06_dp, &
         -1.1564453268637152E+01_dp, 1.4110338828481353E+01_dp, -1.5839917900098961E+01_dp, &
         9.0605780190336862E+00_dp, 1.7879972721762254E+01_dp, -7.3580492861370900E+01_dp, &
         1.4939077315607656E+02_dp, -2.0550794788511661E+02_dp, 1.9183345232179832E+02_dp, &
         -1.0985654847974544E+02_dp, 2.9069197364239841E+01_dp, 8.6965326344200449E+01_dp, &
         6.5794423109474356E-06_dp, -2.9095311161257293E+01_dp, 3.5500532806834372E+01_dp, &
         -3.8732079625741207E+01_dp, 2.0049687270975561E+01_dp, 4.9909299757334580E+01_dp, &
         -1.9182953635638975E+02_dp, 3.8224197279200575E+02_dp, -5.2037560304457224E+02_dp, &
         4.8229079686506594E+02_dp, -2.7472873500261977E+02_dp, 7.2399140112324929E+01_dp, &
         1.8585904474506876E+02_dp, 1.7201658225003901E-05_dp, -7.6160919574509762E+01_dp, &
         9.2927327352209844E+01_dp, -9.7830738138831876E+01_dp, 4.3762988977931663E+01_dp, &
         1.4604784584109120E+02_dp, -5.2262953630197603E+02_dp, 1.0191335036472396E+03_dp, &
         -1.3702049146643706E+03_dp, 1.2588374043796580E+03_dp, -7.1234187972917664E+02_dp, &
         1.8675582601848978E+02_dp, 4.1901695827903075E+02_dp, 4.7491082333765805E-05_dp, &
         -2.1060582176594269E+02_dp, 2.5696907210161868E+02_dp, -2.5824967221804940E+02_dp, &
         9.0897718031470191E+01_dp, 4.5605578804061344E+02_dp, -1.5124879782856940E+03_dp, &
         2.8748868483526799E+03_dp, -3.8058921632187958E+03_dp, 3.4576924488814939E+03_dp, &
         -1.9398311729774050E+03_dp, 5.0511116864969131E+02_dp, 1.0102801407739152E+03_dp, &
         1.4124779360070820E-04_dp, -6.2777550617602583E+02_dp, 7.6597355093601072E+02_dp, &
         -7.2225725676902391E+02_dp, 1.5432026922197824E+02_dp, 1.5564202303382963E+03_dp, &
         -4.7511882745640760E+03_dp, 8.7516133649064413E+03_dp, -1.1355288142397600E+04_dp, &
         1.0162253283370977E+04_dp, -5.6337171128600294E+03_dp, 1.4529129986222329E+03_dp, &
         2.6601339384885455E+03_dp, 4.6606639202797416E-04_dp, -2.0782638055180396E+03_dp, &
         2.5357620486883484E+03_dp, -2.1748960778962787E+03_dp, -1.9622446905636050E+01_dp, &
         6.0167417094472839E+03_dp, -1.6721095387390953E+04_dp, 2.9566028222436140E+04_dp, &
         -3.7299442151022238E+04_dp, 3.2650804589094801E+04_dp, -1.7775490007433858E+04_dp, &
         4.5156255093209511E+03_dp, 7.9225561991071781E+03_dp, 1.7918715932261867E-03_dp, &
         -8.0336669286121869E+03_dp, 9.8020987286611507E+03_dp, -7.1547112917715904E+03_dp, &
         -3.1511198074212939E+03_dp, 2.7990647407511457E+04_dp, -6.9404808605822764E+04_dp, &
         1.1568723250772001E+05_dp, -1.3955704611726975E+05_dp, 1.1762161553204221E+05_dp, &
         -6.1959192901078532E+04_dp, 1.5294013166136992E+04_dp, 2.8581283113673122E+04_dp, &
         8.8478202618300621E-03_dp, -4.0093340340633775E+04_dp, 4.8918432619865300E+04_dp, &
         -2.4727656895868771E+04_dp, -4.2695894630299779E+04_dp, 1.7698207404610471E+05_dp, &
         -3.7294359948259842E+05_dp, 5.5833895825864724E+05_dp, -6.1205082918743347E+05_dp, &
         4.6999752306235145E+05_dp, -2.2582000828196839E+05_dp, 5.0916331737571825E+04_dp, &
         1.4954788596490613E+05_dp, 7.3099845234523828E-02_dp, -3.4101003337474016E+05_dp, &
         4.1605872519837652E+05_dp, 3.3781671415115379E+04_dp, -9.6249467433489673E+05_dp, &
         2.0843532691271894E+06_dp, -2.8167171929273452E+06_dp, 2.4303395386896660E+06_dp, &
         -8.2647776569912373E+05_dp, -8.1568634633921518E+05_dp, 1.1303807273279666E+06_dp, &
         -4.3211741616069851E+05_dp, 2.1591311325597387E+00_dp, 2.3904566691105513E-08_dp, &
         -1.0524216061902127E-01_dp, 1.2841147400320774E-01_dp, -1.6081987966006114E-01_dp, &
         1.2330752066629211E-01_dp, 8.6989463657353730E-02_dp, -5.6142224268033980E-01_dp, &
         1.2464929077820950E+00_dp, -1.7938392888269015E+00_dp, 1.7241491969739160E+00_dp, &
         -1.0082371381189261E+00_dp, 2.7101096999794005E-01_dp, 2.8496287705980246E+00_dp, &
         4.7574479365230764E-08_dp, -2.0956063301401973E-01_dp, 2.5569575239259368E-01_dp, &
         -3.1475928439865386E-01_dp, 2.3213358106314164E-01_dp, 1.9858313580613890E-01_dp, &
         -1.1552055679733595E+00_dp, 2.5229914218962830E+00_dp, -3.6027621950380175E+00_dp, &
         3.4460540887413100E+00_dp, -2.0083606425817289E+00_dp, 5.3849845894628512E-01_dp, &
         4.5392171215693038E+00_dp, 1.1355602415595152E-07_dp, -5.0046032814876074E-01_dp, &
         6.1063713098642736E-01_dp, -7.3915636311573130E-01_dp, 5.2365825674300659E-01_dp, &
         5.3203716081063890E-01_dp, -2.8430782160250718E+00_dp, 6.1165660979115399E+00_dp, &
         -8.6707645013669357E+00_dp, 8.2553871441875692E+00_dp, -4.7956128548962340E+00_dp, &
         1.2827415834620399E+00_dp, 7.9911174357421331E+00_dp, 2.7267735672504561E-07_dp, &
         -1.2023389067054850E+00_dp, 1.4670340742178860E+00_dp, -1.7474593675113077E+00_dp, &
         1.1886260371952238E+00_dp, 1.4078415190225411E+00_dp, -7.0174340690018511E+00_dp, &
         1.4893651164758403E+01_dp, -2.0970955209319627E+01_dp, 1.9879948538498052E+01_dp, &
         -1.1512860614973324E+01_dp, 3.0724084595660521E+00_dp, 1.4845870303287574E+01_dp, &
         6.4993344169887501E-07_dp, -2.8673891829941871E+00_dp, 3.4986433007173297E+00_dp, &
         -4.0957052042130364E+00_dp, 2.6589353154739181E+00_dp, 3.6828533437700273E+00_dp, &
         -1.7199574435158372E+01_dp, 3.6002283622712199E+01_dp, -5.0336430093192462E+01_dp, &
         4.7498896436592311E+01_dp, -2.7416944125665005E+01_dp, 7.2985761158201008E+00_dp, &
         2.8630233686196082E+01_dp, 1.5561676463892027E-06_dp, -6.8699281572824669E+00_dp, &
         8.3823328274713553E+00_dp, -9.6205476983389229E+00_dp, 5.8991888169639166E+00_dp, &
         9.6885900913409895E+00_dp, -4.2426673846413735E+01_dp, 8.7497348366870099E+01_dp, &
         -1.2138586087506182E+02_dp, 1.1395537051718760E+02_dp, -6.5531527032459493E+01_dp, &
         1.7395737984478455E+01_dp, 5.7149278573182144E+01_dp, 3.7895280519993520E-06_dp, &
         -1.6742161277774240E+01_dp, 2.0427905916032042E+01_dp, -2.2908625009614020E+01_dp, &
         1.3060369342628176E+01_dp, 2.6002569703236279E+01_dp, -1.0671561128972567E+02_dp, &
         2.1652100588293317E+02_dp, -2.9775421501190749E+02_dp, 2.7788090034510896E+02_dp, &
         -1.5910812572414045E+02_dp, 4.2096728824521215E+01_dp, 1.1840509257693286E+02_dp, &
         9.4900573035783242E-06_dp, -4.1965931868056849E+01_dp, 5.1204572200420344E+01_dp, &
         -5.5852949973774983E+01_dp, 2.8888133500973176E+01_dp, 7.2090364616085481E+01_dp, &
         -2.7692734999520661E+02_dp, 5.5175378925874941E+02_dp, -7.5112695001410100E+02_dp, &
         6.9615211100672354E+02_dp, -3.9655272221580424E+02_dp, 1.0450396601481178E+02_dp, &
         2.5609655068148811E+02_dp, 2.4715305054824862E-05_dp, -1.0941937632957631E+02_dp, &
         1.3350746185992543E+02_dp, -1.4075541179078257E+02_dp, 6.3373447021950305E+01_dp, &
         2.0911980757912349E+02_dp, -7.5029290516763729E+02_dp, 1.4644128290940855E+03_dp, &
         -1.9700135465819471E+03_dp, 1.8106713348643566E+03_dp, -1.0249522673245622E+03_dp, &
         2.6878483032699012E+02_dp, 5.8309882443897629E+02_dp, 6.7807520171925752E-05_dp, &
         -3.0064117576976992E+02_dp, 3.6682509804999006E+02_dp, -3.7040717157728307E+02_dp, &
         1.3406424149107912E+02_dp, 6.4428398410363116E+02_dp, -2.1519786183493156E+03_dp, &
         4.1011295900086225E+03_dp, -5.4383331157833045E+03_dp, 4.9469571799236483E+03_dp/)
      REAL(KIND=dp), DIMENSION(80), PARAMETER :: c06 = (/-2.7780629763055545E+03_dp, &
         7.2394748688135314E+02_dp, 1.4141872703259592E+03_dp, 1.9924393057961005E-04_dp, &
         -8.8514298674834572E+02_dp, 1.0799982553752052E+03_dp, -1.0299886964497700E+03_dp, &
         2.4613962885626509E+02_dp, 2.1496502911088132E+03_dp, -6.6513723885247055E+03_dp, &
         1.2319717518748630E+04_dp, -1.6044225972732615E+04_dp, 1.4399549815435170E+04_dp, &
         -8.0011168005279378E+03_dp, 2.0673302306451947E+03_dp, 3.7169273499830429E+03_dp, &
         6.4174536681622018E-04_dp, -2.8588922904614119E+03_dp, 3.4882377284724071E+03_dp, &
         -3.0697095638713813E+03_dp, 1.6426883687990320E+02_dp, 7.9840717598249521E+03_dp, &
         -2.2710590528982131E+04_dp, 4.0596610232320745E+04_dp, -5.1616663987160551E+04_dp, &
         4.5469685184816553E+04_dp, -2.4885092322884550E+04_dp, 6.3498617053543667E+03_dp, &
         1.0875832953021774E+04_dp, 2.3481599355844865E-03_dp, -1.0504850853833614E+04_dp, &
         1.2817288640026694E+04_dp, -9.9607576780561449E+03_dp, -2.6337897159272597E+03_dp, &
         3.4451262132612253E+04_dp, -8.8941818154280394E+04_dp, 1.5161895422610914E+05_dp, &
         -1.8618139413710040E+05_dp, 1.5936755599614693E+05_dp, -8.5114171194832204E+04_dp, &
         2.1268239024390470E+04_dp, 3.7112755999414359E+04_dp, 1.0383653918374779E-02_dp, &
         -4.6784795849598871E+04_dp, 5.7083114992325580E+04_dp, -3.5339178754296212E+04_dp, &
         -3.3888742049394437E+04_dp, 1.8605716164815886E+05_dp, -4.2519544494572078E+05_dp, &
         6.7441504677537933E+05_dp, -7.8031535165577661E+05_dp, 6.3288379982644250E+05_dp, &
         -3.2161368971558253E+05_dp, 7.6774261966426260E+04_dp, 1.6235209876331326E+05_dp, &
         6.3058311437979311E-02_dp, -2.8828244204753474E+05_dp, 3.5173407478189125E+05_dp, &
         -1.1586186796904098E+05_dp, -4.5916736844352842E+05_dp, 1.4672512106301824E+06_dp, &
         -2.7737577271967172E+06_dp, 3.7900648353620353E+06_dp, -3.7631762994996398E+06_dp, &
         2.5690955686708130E+06_dp, -1.0676056448933431E+06_dp, 2.0022423762086249E+05_dp, &
         1.1909890543833333E+06_dp, 7.5897543064246109E-01_dp, -3.6014432096696794E+06_dp, &
         4.3939694879051335E+06_dp, 2.2270474259830317E+06_dp, -1.4748142024036435E+07_dp, &
         2.4772021603756029E+07_dp, -2.0427694973111704E+07_dp, -5.1425641553283399E+06_dp, &
         4.1035528362137891E+07_dp, -5.8536317653180525E+07_dp, 4.1604269769282125E+07_dp, &
         -1.2580153353393231E+07_dp/)
      REAL(KIND=dp), DIMENSION(13, 32, 5), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04, c05, c06/), (/13, 32, 5/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+01_dp) Rc = 2.0000000000000000E+01_dp
      IF (Rc >= 7.9661361694335938E+05_dp) Rc = 7.9661361694335938E+05_dp
      IF (Rc >= 2.0000000000000000E+01_dp .AND. Rc <= 2.4398919835450684E+02_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+01_dp
         U_b = 2.4398919835450684E+02_dp
      END IF
      IF (Rc >= 2.4398919835450684E+02_dp .AND. Rc <= 2.9062022203852043E+03_dp) THEN
         irange = 2
         L_b = 2.4398919835450684E+02_dp
         U_b = 2.9062022203852043E+03_dp
      END IF
      IF (Rc >= 2.9062022203852043E+03_dp .AND. Rc <= 4.6200345281597729E+04_dp) THEN
         irange = 3
         L_b = 2.9062022203852043E+03_dp
         U_b = 4.6200345281597729E+04_dp
      END IF
      IF (Rc >= 4.6200345281597729E+04_dp .AND. Rc <= 2.7305247284985101E+05_dp) THEN
         irange = 4
         L_b = 4.6200345281597729E+04_dp
         U_b = 2.7305247284985101E+05_dp
      END IF
      IF (Rc >= 2.7305247284985101E+05_dp .AND. Rc <= 7.9661361694335938E+05_dp) THEN
         irange = 5
         L_b = 2.7305247284985101E+05_dp
         U_b = 7.9661361694335938E+05_dp
      END IF

      fit_coef(1:13, 1:32) = coefdata(1:13, 1:32, irange)

   END SUBROUTINE get_fit_coef_k16

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k17(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 34)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(252), PARAMETER :: c07 = (/1.8370815219029825E+05_dp, &
         -2.2571407833669311E+05_dp, 1.9425009817913763E+05_dp, -1.0472915274294355E+05_dp, &
         2.6504347685948280E+04_dp, 4.7817002324082285E+04_dp, 2.5450661362359097E-02_dp, &
         -6.6225824542394374E+04_dp, 8.1048394804074065E+04_dp, -4.3435260565217926E+04_dp, &
         -6.1476285747099682E+04_dp, 2.7171168546155246E+05_dp, -5.8303836555695033E+05_dp, &
         8.8719630466380890E+05_dp, -9.9159773972621036E+05_dp, 7.7951289244560874E+05_dp, &
         -3.8487843289358675E+05_dp, 8.9463560970755076E+04_dp, 2.5019631140224831E+05_dp, &
         2.1264619593581804E-01_dp, -5.5981397809465544E+05_dp, 6.8509542672483542E+05_dp, &
         2.1284725467874094E+04_dp, -1.4753034153249892E+06_dp, 3.2502583758630129E+06_dp, &
         -4.4832573914568834E+06_dp, 4.0445406981888828E+06_dp, -1.6710630248330070E+06_dp, &
         -9.4341987280607189E+05_dp, 1.6221639053145065E+06_dp, -6.5652039505004522E+05_dp, &
         2.0948115699131753E+00_dp, 3.6258494519234841E-08_dp, -9.3232132132602369E-02_dp, &
         1.1410199411295747E-01_dp, -1.4480521511281780E-01_dp, 1.1917207481518551E-01_dp, &
         4.9241583812778933E-02_dp, -4.4186107155067733E-01_dp, 1.0258906798051275E+00_dp, &
         -1.5172966271163439E+00_dp, 1.4945236952314433E+00_dp, -8.9549977851517348E-01_dp, &
         2.4671541335739663E-01_dp, 2.7235696297903660E+00_dp, 6.9927670959677010E-08_dp, &
         -1.7984023158039161E-01_dp, 2.2009708683611603E-01_dp, -2.7519253143108352E-01_dp, &
         2.1973689951913447E-01_dp, 1.1430694800789225E-01_dp, -8.8115350487381972E-01_dp, &
         2.0114407245057633E+00_dp, -2.9526094256716577E+00_dp, 2.8950743978986537E+00_dp, &
         -1.7292996895200985E+00_dp, 4.7535789164424885E-01_dp, 4.2427066521500594E+00_dp, &
         1.6193093735503707E-07_dp, -4.1653266916325560E-01_dp, 5.0977241977686150E-01_dp, &
         -6.2812103646832995E-01_dp, 4.8620027382563846E-01_dp, 3.0785581451135241E-01_dp, &
         -2.1047443629858482E+00_dp, 4.7300914686869833E+00_dp, -6.8940188870393388E+00_dp, &
         6.7301912758278446E+00_dp, -4.0080099252645738E+00_dp, 1.0993200811710275E+00_dp, &
         7.2887750366020878E+00_dp, 3.7798556209097436E-07_dp, -9.7246230665416322E-01_dp, &
         1.1901449948154916E+00_dp, -1.4463642323703936E+00_dp, 1.0857865684868881E+00_dp, &
         8.1159807200523648E-01_dp, -5.0502163134081739E+00_dp, 1.1193077721641433E+01_dp, &
         -1.6208047074016658E+01_dp, 1.5759089654312840E+01_dp, -9.3586250334120251E+00_dp, &
         2.5616011278036979E+00_dp, 1.3195622118541367E+01_dp, 8.7449260204547286E-07_dp, &
         -2.2502894227249288E+00_dp, 2.7540086220032598E+00_dp, -3.2983231545819116E+00_dp, &
         2.3932042656303105E+00_dp, 2.1010585628596576E+00_dp, -1.2010575120676899E+01_dp, &
         2.6251583881954033E+01_dp, -3.7760474141713388E+01_dp, 3.6560400169648275E+01_dp, &
         -2.1647605107930151E+01_dp, 5.9123811733713181E+00_dp, 2.4741076971735783E+01_dp, &
         2.0256543020120554E-06_dp, -5.2136786706623308E+00_dp, 6.3807388758373085E+00_dp, &
         -7.5175090758348935E+00_dp, 5.2393521484233307E+00_dp, 5.4345512047312257E+00_dp, &
         -2.8642817974261821E+01_dp, 6.1688125010087411E+01_dp, -8.8091580505469821E+01_dp, &
         8.4897604321769748E+01_dp, -5.0103611036029179E+01_dp, 1.3650920716373861E+01_dp, &
         4.7846812164549789E+01_dp, 4.7505238242518463E-06_dp, -1.2230201238224922E+01_dp, &
         1.4967872450044803E+01_dp, -1.7304693765421220E+01_dp, 1.1480222650031589E+01_dp, &
         1.4238881963448762E+01_dp, -6.9310801973768122E+01_dp, 1.4691230332041425E+02_dp, &
         -2.0810906978525890E+02_dp, 1.9951727876211754E+02_dp, -1.1730830010948584E+02_dp, &
         3.1871643718694187E+01_dp, 9.5584382658373627E+01_dp, 1.1389673415432144E-05_dp, &
         -2.9331774923459527E+01_dp, 3.5897526849215701E+01_dp, -4.0593292635133999E+01_dp, &
         2.5300639865679354E+01_dp, 3.8217190441467515E+01_dp, -1.7193734266608891E+02_dp, &
         3.5812529431878522E+02_dp, -5.0270766323901529E+02_dp, 4.7906560065789068E+02_dp, &
         -2.8044896737160479E+02_dp, 7.5945642333907642E+01_dp, 1.9807989409896766E+02_dp, &
         2.8179303689685462E-05_dp, -7.2597490992070050E+01_dp, 8.8847958009412466E+01_dp, &
         -9.7843279644225888E+01_dp, 5.6165358634741970E+01_dp, 1.0621909187312551E+02_dp, &
         -4.4160242824475296E+02_dp, 9.0211884406401828E+02_dp, -1.2531534970604021E+03_dp, &
         1.1858384963678866E+03_dp, -6.9062251088863718E+02_dp, 1.8628549560366449E+02_dp, &
         4.2844728563604173E+02_dp, 7.2680619722386190E-05_dp, -1.8733325328642434E+02_dp, &
         2.2926633331658493E+02_dp, -2.4440219730170580E+02_dp, 1.2508379111590767E+02_dp, &
         3.0936543510420876E+02_dp, -1.1871960018194495E+03_dp, 2.3726584254496997E+03_dp, &
         -3.2557421748838055E+03_dp, 3.0549203135871107E+03_dp, -1.7679853070233212E+03_dp, &
         4.7457670212220614E+02_dp, 9.7553162024702920E+02_dp, 1.9784461526304437E-04_dp, &
         -5.1025103218277997E+02_dp, 6.2446593588851465E+02_dp, -6.3883562219794624E+02_dp, &
         2.7468433715161183E+02_dp, 9.5798846328704099E+02_dp, -3.3853330597180307E+03_dp, &
         6.5968917599317774E+03_dp, -8.9193738742063597E+03_dp, 8.2819962181539267E+03_dp, &
         -4.7550782366077383E+03_dp, 1.2684749997005017E+03_dp, 2.3659593721200058E+03_dp, &
         5.7767824463795267E-04_dp, -1.4910686829640895E+03_dp, 1.8248273467436229E+03_dp, &
         -1.7679369070925552E+03_dp, 5.5958786970336223E+02_dp, 3.2150931773552397E+03_dp, &
         -1.0419368658525378E+04_dp, 1.9704230885199639E+04_dp, -2.6153107469601153E+04_dp, &
         2.3957244240913482E+04_dp, -1.3610636015655184E+04_dp, 3.6004270680947861E+03_dp, &
         6.2184859119400480E+03_dp, 1.8515069283753457E-03_dp, -4.7844525772745565E+03_dp, &
         5.8553838083439105E+03_dp, -5.2557820977780702E+03_dp, 7.6998584833558868E+02_dp, &
         1.2019149643632090E+04_dp, -3.5476725058178665E+04_dp, 6.4648163585761089E+04_dp, &
         -8.3736358762410033E+04_dp, 7.5284898688267742E+04_dp, -4.2133475096368573E+04_dp, &
         1.1009804491050614E+04_dp, 1.8195464521428708E+04_dp, 6.7517479161416302E-03_dp, &
         -1.7477227521677822E+04_dp, 2.1389179519798992E+04_dp, -1.7071068989116044E+04_dp, &
         -2.4214225086274514E+03_dp, 5.2234983048949362E+04_dp, -1.3874773753806672E+05_dp, &
         2.4073094092437049E+05_dp, -3.0103520143692737E+05_dp, 2.6302584995883796E+05_dp, &
         -1.4370762781981824E+05_dp, 3.6795383161880185E+04_dp, 6.2090310305363819E+04_dp, &
         2.9820864621100219E-02_dp, -7.7417968024254675E+04_dp, 9.4745982916105684E+04_dp, &
         -6.1136770748532195E+04_dp, -4.6361387820341610E+04_dp, 2.8437287677000294E+05_dp, &
         -6.6357048964616924E+05_dp, 1.0697148236834274E+06_dp, -1.2605114585738920E+06_dp, &
         1.0444647787696669E+06_dp, -5.4377630601913563E+05_dp, 1.3327154055909728E+05_dp, &
         2.7161799040869653E+05_dp, 1.8172876846851641E-01_dp, -4.7458463411040365E+05_dp, &
         5.8080137302666670E+05_dp, -2.1206348195483550E+05_dp, -6.8472287525603513E+05_dp, &
         2.2646124682806111E+06_dp, -4.3463120374242561E+06_dp, 6.0384830132842269E+06_dp, &
         -6.1288439298421443E+06_dp, 4.3074969014170449E+06_dp, -1.8594197839029063E+06_dp, &
         3.6681152123045758E+05_dp, 1.9925461761031416E+06_dp, 2.2253842343468082E+00_dp, &
         -5.8978754757524757E+06_dp, 7.2176798918735152E+06_dp, 3.1998615207433747E+06_dp, &
         -2.2851538821395889E+07_dp, 3.8978847055747673E+07_dp, -3.3766691905692071E+07_dp, &
         -3.6817494449235639E+06_dp, 5.8417659409558006E+07_dp, -8.7460434017794341E+07_dp, &
         6.4007727007001713E+07_dp, -1.9852653644652657E+07_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.3076079314075767E-01_dp, &
         -3.1204465976027999E-02_dp, -1.7122144125202067E-03_dp, 7.1244719472004776E-04_dp, &
         9.7914036643084913E-05_dp, -1.6905608998459976E-05_dp, -7.4913349592830019E-05_dp, &
         3.9524844540922514E-04_dp, -1.3207548101866621E-03_dp, 2.4893343556511901E-03_dp, &
         -1.7630401832362664E-03_dp, -1.5303430582117692E-03_dp, 2.4954212764800787E-03_dp, &
         4.0176042278607560E-01_dp, -1.0052269981409805E-01_dp, -3.7752999908446687E-03_dp, &
         2.3355451575862705E-03_dp, 2.2164896522723778E-04_dp, -5.9872225699415800E-05_dp, &
         -2.3417557017388007E-04_dp, 1.2627576376189641E-03_dp, -4.2195733035849375E-03_dp, &
         7.9523531821684996E-03_dp, -5.6322897078267794E-03_dp, -4.8887888709380819E-03_dp, &
         7.9720806192536044E-03_dp, 7.0203193009585230E-01_dp, -1.9160546782655125E-01_dp, &
         -1.2596497264163648E-03_dp, 4.4901354476166479E-03_dp, 1.0663595310664382E-04_dp, &
         -1.2452948785005740E-04_dp, -4.2231891451488128E-04_dp, 2.3703226650564047E-03_dp, &
         -7.9222542238009102E-03_dp, 1.4928487774690918E-02_dp, -1.0573462306857859E-02_dp, &
         -9.1774203746157711E-03_dp, 1.4966211175066604E-02_dp, 1.0539449282453617E+00_dp, &
         -3.2251390565619303E-01_dp, 1.0615457224673873E-02_dp, 7.2905132110783599E-03_dp, &
         -4.7936555312612936E-04_dp, -2.0448213231640043E-04_dp, -6.6082699844020952E-04_dp, &
         3.9081706133278242E-03_dp, -1.3071632605714272E-02_dp, 2.4627501045524507E-02_dp, &
         -1.7442957122418997E-02_dp, -1.5139885809291837E-02_dp, 2.4690040632825656E-02_dp, &
         1.4843975248485286E+00_dp, -5.1701819383514402E-01_dp, 3.9543265998243871E-02_dp, &
         1.0365663424557522E-02_dp, -1.8681910570608855E-03_dp, -2.5533490666031401E-04_dp, &
         -9.7612040167318833E-04_dp, 6.1129491849905623E-03_dp, -2.0475177295793314E-02_dp, &
         3.8568511251337811E-02_dp, -2.7314928846016738E-02_dp, -2.3709957461900095E-02_dp, &
         3.8663702667761222E-02_dp, 2.0272052023813556E+00_dp, -8.0777194692517917E-01_dp, &
         9.8050801990368502E-02_dp, 1.2353931861675717E-02_dp, -4.4992921000651621E-03_dp, &
         -1.6348775215867598E-04_dp, -1.4060441185015371E-03_dp, 9.2907069709225621E-03_dp, &
         -3.1187261758263946E-02_dp, 5.8734750953777816E-02_dp, -4.1589834778631818E-02_dp, &
         -3.6106670683926118E-02_dp, 5.8868554322763382E-02_dp, 2.7262530833535337E+00_dp, &
         -1.2406060229341269E+00_dp, 2.0638055327688459E-01_dp, 9.9642442503349842E-03_dp, &
         -8.8630281922776159E-03_dp, 3.0333232987031180E-04_dp, -2.0137745161970229E-03_dp, &
         1.3848156069161642E-02_dp, -4.6621551236882400E-02_dp, 8.7785867514915933E-02_dp, &
         -6.2143870434135301E-02_dp, -5.3964809521018703E-02_dp, 8.7958357674071022E-02_dp, &
         3.6399268594035759E+00_dp, -1.8805467044522048E+00_dp, 3.9684620223953421E-01_dp, &
         -3.6979224560712339E-03_dp, -1.5290732711395427E-02_dp, 1.5650362097787600E-03_dp, &
         -2.9164870249643996E-03_dp, 2.0338155085769950E-02_dp, -6.8712559932418329E-02_dp, &
         1.2936460946821179E-01_dp, -9.1545670622710518E-02_dp, -7.9523365645068900E-02_dp, &
         1.2956570819945470E-01_dp, 4.8477396000212751E+00_dp, -2.8204823695808767E+00_dp, &
         7.2037192283678975E-01_dp, -4.2006232839993926E-02_dp, -2.3408946961019392E-02_dp, &
         4.3011253069062654E-03_dp, -4.3418036026321299E-03_dp, 2.9528398299157512E-02_dp, &
         -1.0014893969272984E-01_dp, 1.8854172213152740E-01_dp, -1.3337119045002854E-01_dp, &
         -1.1589897868070902E-01_dp, 1.8874729813192132E-01_dp, 6.4610001308210583E+00_dp, &
         -4.1941537938416324E+00_dp, 1.2562579246714793E+00_dp, -1.2956693816744225E-01_dp, &
         -3.0923800259146223E-02_dp, 9.4582394952995325E-03_dp, -6.7291628461492317E-03_dp, &
         4.2513667531429149E-02_dp, -1.4472119835853223E-01_dp, 2.7248024980715652E-01_dp, &
         -1.9267411822422414E-01_dp, -1.6749461934299184E-01_dp, 2.7264883596712303E-01_dp, &
         8.6416370834085008E+00_dp, -6.1970932928169926E+00_dp, 2.1269516693256731E+00_dp, &
         -3.0975515701583661E-01_dp, -3.1212680848427189E-02_dp, 1.8014526702952737E-02_dp, &
         -1.0880146572029471E-02_dp, 6.0921073275536620E-02_dp, -2.0789469712183004E-01_dp, &
         3.9152299938109048E-01_dp, -2.7675631174570009E-01_dp, -2.4066959573985081E-01_dp, &
         3.9159410702887432E-01_dp, 1.1639225306345036E+01_dp, -9.1246532821522290E+00_dp, &
         3.5218372776759885E+00_dp, -6.5629264554887390E-01_dp, -9.0889294350904488E-03_dp, &
         3.0161449462820341E-02_dp, -1.8099973853997281E-02_dp, 8.7313637818274087E-02_dp, &
         -2.9790804111642638E-01_dp, 5.6124720781308091E-01_dp, -3.9662215564129433E-01_dp, &
         -3.4500609303341645E-01_dp, 5.6114294060164027E-01_dp, 1.5873255748499172E+01_dp, &
         -1.3453102441637940E+01_dp, 5.7419478039571841E+00_dp, -1.2916346325142669E+00_dp, &
         6.5759417771617396E-02_dp, 4.3530054327603068E-02_dp, -3.0166292171774273E-02_dp, &
         1.2600955452173773E-01_dp, -4.2822600401379651E-01_dp, 8.0701621705779869E-01_dp, &
         -5.7019201205676040E-01_dp, -4.9611094674881584E-01_dp, 8.0664222793806906E-01_dp, &
         2.2143206347171624E+01_dp, -2.0046617832223532E+01_dp, 9.3071997135744269E+00_dp, &
         -2.4230046225771531E+00_dp, 2.4836527841931291E-01_dp, 5.0611968862569114E-02_dp, &
         -4.9048621468263262E-02_dp, 1.8493192924859195E-01_dp, -6.2385283643435641E-01_dp, &
         1.1757719724259648E+00_dp, -8.3061767393793995E-01_dp, -7.2286268514831931E-01_dp, &
         1.1749965024749465E+00_dp, 3.2270336161918614E+01_dp, -3.0801175619216782E+01_dp, &
         1.5292207073605903E+01_dp, -4.4474171161113727E+00_dp, 6.3789574241791069E-01_dp, &
         3.6622362193611202E-02_dp, -7.7335776114862156E-02_dp, 2.8112010508678736E-01_dp, &
         -9.4070495460370629E-01_dp, 1.7725454560696385E+00_dp, -1.2520539024975985E+00_dp, &
         -1.0898461557958572E+00_dp, 1.7711168752874866E+00_dp, 5.1688635070150774E+01_dp, &
         -5.1282387193841267E+01_dp, 2.6737329772837441E+01_dp, -8.4076215182313732E+00_dp, &
         1.4596804851316547E+00_dp, -2.2785940581509477E-02_dp, -1.2416451365357957E-01_dp, &
         4.6408964352239751E-01_dp, -1.5442693251973152E+00_dp, 2.9088148034596015E+00_dp, &
         -2.0544486079149733E+00_dp, -1.7885651298283725E+00_dp, 2.9061229744851547E+00_dp, &
         1.0762496921374742E+02_dp, -1.0933096205262035E+02_dp, 5.8725372851529393E+01_dp, &
         -1.9345285433471407E+01_dp, 3.7179008532913183E+00_dp, -1.9659850716063998E-01_dp, &
         -2.4853838702353581E-01_dp, 9.8299055919831080E-01_dp, -3.2634289123959395E+00_dp, &
         6.1454812603833870E+00_dp, -4.3401377360820046E+00_dp, -3.7787705767862518E+00_dp, &
         6.1392696513115714E+00_dp, 1.0523497373659305E+00_dp, -2.5417364415230703E-01_dp, &
         -1.2822466959058445E-02_dp, 5.8356833156815681E-03_dp, 7.3675755090544322E-04_dp, &
         -1.4172040208393219E-04_dp, -6.0556073785800386E-04_dp, 3.2127278163324629E-03_dp, &
         -1.0735448720595803E-02_dp, 2.0233569467854696E-02_dp, -1.4330270606267244E-02_dp, &
         -1.2438789926505517E-02_dp, 2.0283242462557815E-02_dp, 1.1287287038620462E+00_dp, &
         -3.1013327621056686E-01_dp, -1.6066847757949096E-03_dp, 7.3809503661043836E-03_dp, &
         1.4495723902261054E-04_dp, -2.1120727007587788E-04_dp, -6.8141332001349158E-04_dp, &
         3.8348376691482336E-03_dp, -1.2815205295062257E-02_dp, 2.4148412584623379E-02_dp, &
         -1.7103912448526319E-02_dp, -1.4845470277022208E-02_dp, 2.4209810789034943E-02_dp, &
         1.2882283178090226E+00_dp, -4.3040597485938342E-01_dp, 2.4831881244751369E-02_dp, &
         9.9485126813913909E-03_dp, -1.2174764509962526E-03_dp, -3.0299547417090487E-04_dp, &
         -8.3788259132067935E-04_dp, 5.1512432549364667E-03_dp, -1.7230336261564266E-02_dp, &
         3.2458796239020947E-02_dp, -2.2990522835058941E-02_dp, -1.9954198064667916E-02_dp, &
         3.2543079449201776E-02_dp, 1.5448514878021482E+00_dp, -6.3250090306882112E-01_dp, &
         7.5173375330629649E-02_dp, 1.2239074374363137E-02_dp, -3.6909008476000097E-03_dp, &
         -3.0904460311629461E-04_dp, -1.0873965370004825E-03_dp, 7.3078182273191615E-03_dp, &
         -2.4497680697525372E-02_dp, 4.6136049459575182E-02_dp, -3.2674527368890323E-02_dp, &
         -2.8361888660189474E-02_dp, 4.6251077222106737E-02_dp, 1.9211073988689866E+00_dp, &
         -9.4502175925954501E-01_dp, 1.6441311125312935E-01_dp, 1.1514168833412042E-02_dp, &
         -7.7156781634275936E-03_dp, -2.5314063749536728E-05_dp, -1.4593192552218414E-03_dp, &
         1.0532414112812576E-02_dp, -3.5426284263123525E-02_dp, 6.6700389834680351E-02_dp, &
         -4.7225177778801364E-02_dp, -4.1002894262538177E-02_dp, 6.6845407444861935E-02_dp, &
         2.4503233903027750E+00_dp, -1.4111691869136846E+00_dp, 3.1685836597950173E-01_dp, &
         2.3051678588932364E-03_dp, -1.3672390712846821E-02_dp, 9.1023388741843267E-04_dp, &
         -2.0204001124313029E-03_dp, 1.5157830646097843E-02_dp, -5.1202256225836684E-02_dp, &
         9.6382554516795824E-02_dp, -6.8210683093346311E-02_dp, -5.9248334038998998E-02_dp, &
         9.6542522830884450E-02_dp, 3.1805574548812579E+00_dp, -2.0942909063464801E+00_dp, &
         5.7089561010061707E-01_dp, -2.5833725640180539E-02_dp, -2.1491678096792352E-02_dp, &
         3.0989304847712494E-03_dp, -2.9159806953795845E-03_dp, 2.1665920578763782E-02_dp, &
         -7.3546522325358535E-02_dp, 1.3842587042657298E-01_dp, -9.7913493025603968E-02_dp, &
         -8.5091612722190096E-02_dp, 1.3856859671520547E-01_dp, 4.1811646792046675E+00_dp, &
         -3.0863241265112706E+00_dp, 9.8614321254810988E-01_dp, -9.2071074916824508E-02_dp, &
         -2.9756968197499723E-02_dp, 7.4456170391292899E-03_dp, -4.4508265047520533E-03_dp, &
         3.0758812754125595E-02_dp, -1.0495169546575114E-01_dp, 1.9754193374336126E-01_dp, &
         -1.3965469631124944E-01_dp, -1.2142837818243153E-01_dp, 1.9761867354288118E-01_dp, &
         5.5543701913701007E+00_dp, -4.5209989365988195E+00_dp, 1.6539762241301386E+00_dp, &
         -2.3030816535912507E-01_dp, -3.3813532281998536E-02_dp, 1.5086566272402728E-02_dp, &
         -7.2342195006746086E-03_dp, 4.3480249136807009E-02_dp, -1.4903057718459220E-01_dp, &
         2.8058221599374905E-01_dp, -1.9827032648170015E-01_dp, -1.7246987861993432E-01_dp, &
         2.8052964857063045E-01_dp, 7.4574112588355606E+00_dp, -6.5963933721010299E+00_dp, &
         2.7132460533356846E+00_dp, -4.9834201048625321E-01_dp, -2.2167462115990084E-02_dp, &
         2.6872562330083130E-02_dp, -1.2387908764726825E-02_dp, 6.1461561859636082E-02_dp, &
         -2.1112442139248186E-01_dp, 3.9769622119135700E-01_dp, -2.8093246365215413E-01_dp, &
         -2.4445770977980782E-01_dp, 3.9743993120230359E-01_dp, 1.0149199555623431E+01_dp, &
         -9.6192305108604046E+00_dp, 4.3758772166955557E+00_dp, -9.9115581475053982E-01_dp, &
         2.9545326239106471E-02_dp, 4.1823403732506717E-02_dp, -2.1690138678753676E-02_dp, &
         8.7456012212983200E-02_dp, -2.9961125255873033E-01_dp, 5.6476258723175166E-01_dp, &
         -3.9887202332975563E-01_dp, -3.4716515228541883E-01_dp, 5.6422730177346647E-01_dp, &
         1.4099231208017288E+01_dp, -1.4107075328742125E+01_dp, 6.9778612606041159E+00_dp, &
         -1.8607045468744079E+00_dp, 1.6704246179217699E-01_dp, 5.3958212328329369E-02_dp, &
         -3.7277902595866740E-02_dp, 1.2645660436554337E-01_dp, -4.2911347940038297E-01_dp, &
         8.0927294663161464E-01_dp, -5.7153148233841800E-01_dp, -4.9752142302887042E-01_dp, &
         8.0839552282034810E-01_dp, 2.0271891440726208E+01_dp, -2.1062627665694670E+01_dp, &
         1.1112095708204244E+01_dp, -3.3547769325343713E+00_dp, 4.6741458495922611E-01_dp, &
         4.8143364079041427E-02_dp, -6.0571498843629903E-02_dp, 1.8810682220477090E-01_dp, &
         -6.2907199074458620E-01_dp, 1.1862382813167087E+00_dp, -8.3776299569573820E-01_dp, &
         -7.2938082345320598E-01_dp, 1.1849370069427336E+00_dp, 3.0981138444088945E+01_dp, &
         -3.2826924199015650E+01_dp, 1.8061271201881183E+01_dp, -5.9402359654032049E+00_dp, &
         1.0560336370860732E+00_dp, -1.3076665170325774E-03_dp, -9.1952845171647243E-02_dp, &
         2.9356117158718775E-01_dp, -9.6956307740968917E-01_dp, 1.8269216521737346E+00_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/-1.2901998665259486E+00_dp, &
         -1.1234438324718303E+00_dp, 1.8249388295785653E+00_dp, 5.3052663877251987E+01_dp, &
         -5.6300287553126552E+01_dp, 3.1510803678504963E+01_dp, -1.0839542135686512E+01_dp, &
         2.1894315183922908E+00_dp, -1.2602122478496375E-01_dp, -1.3889457312206924E-01_dp, &
         5.0337725032938441E-01_dp, -1.6571300126908901E+00_dp, 3.1201665791015056E+00_dp, &
         -2.2033768080640694E+00_dp, -1.9187234757026690E+00_dp, 3.1167184219995239E+00_dp, &
         1.1438382828329563E+02_dp, -1.1990346708888428E+02_dp, 6.6738432612919837E+01_dp, &
         -2.3077325719032441E+01_dp, 4.8314221537097728E+00_dp, -3.8128451926731016E-01_dp, &
         -2.6117870547203120E-01_dp, 1.0707535522056886E+00_dp, -3.5392609127171060E+00_dp, &
         6.6628013349508590E+00_dp, -4.7050599511008340E+00_dp, -4.0970099516280252E+00_dp, &
         6.6554064206249119E+00_dp, 4.4215573031384639E+02_dp, -4.5609383814257563E+02_dp, &
         2.4967139254642808E+02_dp, -8.4637858247865637E+01_dp, 1.7233338058664309E+01_dp, &
         -1.2625615465109370E+00_dp, -9.8044928324825564E-01_dp, 4.0814152411288465E+00_dp, &
         -1.3536108639012429E+01_dp, 2.5486079632183817E+01_dp, -1.7998229964061135E+01_dp, &
         -1.5671083055262866E+01_dp, 2.5458811210562828E+01_dp, 1.8010738006583224E-01_dp, &
         -6.2745911390591533E-02_dp, -1.3976556185891661E-02_dp, 7.9986178737707029E-03_dp, &
         3.8749477893130122E-03_dp, -2.5821008682339651E-03_dp, -1.4958653576095052E-03_dp, &
         9.6456448930898895E-04_dp, 1.1417500384510227E-03_dp, -1.1797959682526289E-03_dp, &
         1.5700881689688416E-04_dp, 2.1623209723117793E-04_dp, -8.1228210530532070E-05_dp, &
         5.6570602726698116E-01_dp, -2.1563012466646606E-01_dp, -3.7806458016327996E-02_dp, &
         3.0569342974261178E-02_dp, 1.0311060275225992E-02_dp, -1.0309865308339399E-02_dp, &
         -3.9036553568344260E-03_dp, 3.9203404473260196E-03_dp, 3.6495598931388052E-03_dp, &
         -5.0687843416395749E-03_dp, 1.7273963399649264E-03_dp, 1.9928519421558662E-04_dp, &
         -1.8203789728090053E-04_dp, 1.0320570826693001E+00_dp, -4.5936001731287074E-01_dp, &
         -4.4411061460762145E-02_dp, 7.4144797958186112E-02_dp, 1.0753445855713722E-02_dp, &
         -2.6121803125198153E-02_dp, -3.5691764025816315E-03_dp, 1.0112104989405936E-02_dp, &
         6.3659041969645553E-03_dp, -1.3809991703339446E-02_dp, 7.7140379343898745E-03_dp, &
         -1.5474246435163541E-03_dp, -1.6652411530409433E-05_dp, 1.6490282386969555E+00_dp, &
         -8.8535365615839856E-01_dp, -4.0715433614121159E-03_dp, 1.5650434575864441E-01_dp, &
         -6.3381624762084273E-03_dp, -5.6308039349758643E-02_dp, 4.9139660534489723E-03_dp, &
         2.2014370541855809E-02_dp, 7.6064754785856734E-03_dp, -3.1129921875843455E-02_dp, &
         2.3050532267550796E-02_dp, -7.4383126577924879E-03_dp, 8.7131148771299547E-04_dp, &
         2.5133929798909564E+00_dp, -1.6395076438389586E+00_dp, 1.4912090565157865E-01_dp, &
         2.9864401700894339E-01_dp, -6.6701724542314406E-02_dp, -1.0663176597759881E-01_dp, &
         3.3393538746138207E-02_dp, 4.1661750655118041E-02_dp, 1.8488948835995406E-03_dp, &
         -5.9792963151484434E-02_dp, 5.4834911193647876E-02_dp, -2.1671245162956085E-02_dp, &
         3.3770605802150699E-03_dp, 3.7665004427790230E+00_dp, -2.9643013129955884E+00_dp, &
         5.5460479257789530E-01_dp, 5.1853399079636375E-01_dp, -2.2256772225024538E-01_dp, &
         -1.7685996741929338E-01_dp, 1.0486353197058355E-01_dp, 6.7964719385209413E-02_dp, &
         -2.4139367958340034E-02_dp, -9.6856881358269989E-02_dp, 1.1068443291141189E-01_dp, &
         -5.0365313485448475E-02_dp, 8.9913262883314117E-03_dp, 5.6212689576437533E+00_dp, &
         -5.2696679183943491E+00_dp, 1.4981304567183242E+00_dp, 8.0921229093198388E-01_dp, &
         -5.7220665206718013E-01_dp, -2.4595464439568815E-01_dp, 2.5924803186583678E-01_dp, &
         8.9778626060307662E-02_dp, -9.7021696248226069E-02_dp, -1.2255486037880553E-01_dp, &
         1.9217594222861736E-01_dp, -9.9958611073620751E-02_dp, 1.9715217021775512E-02_dp, &
         8.4047702141954037E+00_dp, -9.2546240707189753E+00_dp, 3.5534752138358008E+00_dp, &
         1.0782156596209165E+00_dp, -1.2868181607551561E+00_dp, -2.3746783847288316E-01_dp, &
         5.5681719910211935E-01_dp, 7.0579962978307528E-02_dp, -2.6253042844784452E-01_dp, &
         -7.8903735544427714E-02_dp, 2.7820728538275824E-01_dp, -1.7127967005997424E-01_dp, &
         3.7227062473712304E-02_dp, 1.2626716828773741E+01_dp, -1.6118082462763581E+01_dp, &
         7.8581631083902526E+00_dp, 9.9481739788304591E-01_dp, -2.6354708004856406E+00_dp, &
         5.4288386772138826E-02_dp, 1.0677957643381861E+00_dp, -8.1556028930718116E-02_dp, &
         -5.8349972470961986E-01_dp, 1.6434754008913957E-01_dp, 2.9118537433730779E-01_dp, &
         -2.4503791491071500E-01_dp, 6.0378888487433051E-02_dp, 1.9091662225464063E+01_dp, &
         -2.7927209435631788E+01_dp, 1.6649978899989840E+01_dp, -3.7733057875206194E-01_dp, &
         -4.9668781274772158E+00_dp, 1.1090269089810683E+00_dp, 1.8166148798971244E+00_dp, &
         -5.6474461880155935E-01_dp, -1.1104410579548303E+00_dp, 8.5108302171346073E-01_dp, &
         3.8997884915739786E-02_dp, -2.5276025183160106E-01_dp, 7.9555438462666675E-02_dp, &
         2.9087515158702328E+01_dp, -4.8266848623079582E+01_dp, 3.4310610939149257E+01_dp, &
         -5.4385586613019150E+00_dp, -8.5091233691378658E+00_dp, 3.9330354951833977E+00_dp, &
         2.6016553035322700E+00_dp, -1.7480235778329487E+00_dp, -1.7799379145508136E+00_dp, &
         2.3589742723119596E+00_dp, -8.6871581926008334E-01_dp, -2.5405644527526649E-02_dp, &
         6.5524802080553057E-02_dp, 4.4716407424310830E+01_dp, -8.3401654526387858E+01_dp, &
         6.9392101713584879E+01_dp, -2.0013526749110621E+01_dp, -1.2568573670815580E+01_dp, &
         1.0394404054784644E+01_dp, 2.5120938544689957E+00_dp, -4.1812179718438625E+00_dp, &
         -2.1440743380530605E+00_dp, 5.0859869680028833E+00_dp, -3.0893954222950697E+00_dp, &
         7.8092848188977992E-01_dp, -5.0011270818283222E-02_dp, 6.9526312537902172E+01_dp, &
         -1.4443332206413996E+02_dp, 1.3855136404356625E+02_dp, -5.7705992232865768E+01_dp, &
         -1.2976025146161891E+01_dp, 2.3238517986306515E+01_dp, -1.1539823334646924E+00_dp, &
         -8.2494908804502867E+00_dp, -7.8117315080707395E-01_dp, 8.8027481030006598E+00_dp, &
         -7.3145712081593359E+00_dp, 2.6860366240642017E+00_dp, -3.8959830551301688E-01_dp, &
         1.0992188203872426E+02_dp, -2.5173108252917748E+02_dp, 2.7439346526745004E+02_dp, &
         -1.4922339600333521E+02_dp, 4.9511964753249949E+00_dp, 4.3977905438400001E+01_dp, &
         -1.4483825851053977E+01_dp, -1.2571750398213693E+01_dp, 5.2268271154719423E+00_dp, &
         1.0879079479809656E+01_dp, -1.2909711450908476E+01_dp, 5.8882269567429590E+00_dp, &
         -1.0555338673912253E+00_dp, 1.7922144960276142E+02_dp, -4.4649909942956606E+02_dp, &
         5.4358551771857742E+02_dp, -3.6239825682407479E+02_dp, 8.2915134399015301E+01_dp, &
         6.3991587425066804E+01_dp, -4.6200832193128967E+01_dp, -1.0539208929328979E+01_dp, &
         1.7867396014859761E+01_dp, 5.7571582062942861E+00_dp, -1.6184489993979646E+01_dp, &
         9.2303743866032342E+00_dp, -1.9050792144269630E+00_dp, 3.1419936221955288E+02_dp, &
         -8.3634297532370749E+02_dp, 1.1110728196113650E+03_dp, -8.6002680865678747E+02_dp, &
         3.2634452019348345E+02_dp, 4.5766391820609520E+01_dp, -9.8967485930911280E+01_dp, &
         8.3582056990308278E+00_dp, 3.3401494394153403E+01_dp, -8.4113932333341950E+00_dp, &
         -1.4692641867662479E+01_dp, 1.1653736548115734E+01_dp, -2.7552478960951516E+00_dp, &
         6.9397291379150988E+02_dp, -1.9276242869869666E+03_dp, 2.7092202860876819E+03_dp, &
         -2.2995420208229034E+03_dp, 1.0985958796049201E+03_dp, -1.0518020171631299E+02_dp, &
         -1.8516772374256206E+02_dp, 5.1320419469952256E+01_dp, 6.0714424144403772E+01_dp, &
         -3.4538746632826253E+01_dp, -1.3916031203934184E+01_dp, 1.7774507932922663E+01_dp, &
         -4.7391960892775407E+00_dp, 1.4574807068110414E+00_dp, -5.1966690003049509E-01_dp, &
         -1.0937121555649043E-01_dp, 6.8282130920804027E-02_dp, 3.0254852511833141E-02_dp, &
         -2.2336606510276111E-02_dp, -1.1641347572123872E-02_dp, 8.3890438873829229E-03_dp, &
         9.3122892529834168E-03_dp, -1.0437466316246218E-02_dp, 2.0493938245759241E-03_dp, &
         1.4556641442013175E-03_dp, -6.1406129186868910E-04_dp, 1.6635774394994376E+00_dp, &
         -7.4447036116082299E-01_dp, -7.3265463407581985E-02_dp, 1.2175244735677020E-01_dp, &
         1.8417461157197813E-02_dp, -4.3129640640968338E-02_dp, -6.3261088523512048E-03_dp, &
         1.6717304979147734E-02_dp, 1.0813189391350427E-02_dp, -2.2870001477521811E-02_dp, &
         1.2540227392196780E-02_dp, -2.3939994363031116E-03_dp, -6.7292329081708585E-05_dp, &
         2.1130083057428242E+00_dp, -1.2646245625758314E+00_dp, 3.9026678312747663E-02_dp, &
         2.3868043891980367E-01_dp, -2.2107148945619371E-02_dp, -8.7624034141812510E-02_dp, &
         1.2400332443529075E-02_dp, 3.4503268350899216E-02_dp, 9.9312975904139698E-03_dp, &
         -4.9617985830045279E-02_dp, 3.8605161734445871E-02_dp, -1.3143058004433409E-02_dp, &
         1.6819268858947601E-03_dp, 2.8861358995389326E+00_dp, -2.2379012503086972E+00_dp, &
         3.2529535767432088E-01_dp, 4.3452974907619385E-01_dp, -1.3098878900070884E-01_dp, &
         -1.5802763888684246E-01_dp, 6.3048141625604320E-02_dp, 6.2230049682745238E-02_dp, &
         -3.4334538372675287E-03_dp, -9.1072403941971233E-02_dp, 8.8902484985361505E-02_dp, &
         -3.6762642551420852E-02_dp, 6.0063306694479717E-03_dp, 4.1212721248896012E+00_dp, &
         -3.9524567137450393E+00_dp, 9.8864687807522755E-01_dp, 7.1583307090425163E-01_dp, &
         -3.8544667410777939E-01_dp, -2.4530041815719975E-01_dp, 1.7970618099537952E-01_dp, &
         9.4646106681016307E-02_dp, -5.1054392753752079E-02_dp, -1.3697099489584089E-01_dp, &
         1.7083844262209280E-01_dp, -8.1262942364707796E-02_dp, 1.5034543179272131E-02_dp, &
         6.0427874405411135E+00_dp, -6.9114188446690799E+00_dp, 2.4368523205155839E+00_dp, &
         1.0422276743751018E+00_dp, -9.2514306563957271E-01_dp, -3.0462890230255507E-01_dp, &
         4.1835014026158401E-01_dp, 1.0930727144152716E-01_dp, -1.7419009015490974E-01_dp, &
         -1.4775669359208615E-01_dp, 2.7804753282859701E-01_dp, -1.5293629530276798E-01_dp, &
         3.1260563924463657E-02_dp, 9.0090022431937165E+00_dp, -1.1984085393120617E+01_dp, &
         5.4766120418255344E+00_dp, 1.2306964512849667E+00_dp, -1.9841323771701029E+00_dp, &
         -2.0178112750263197E-01_dp, 8.5733779402876675E-01_dp, 4.2784199415162032E-02_dp, &
         -4.3931636150909686E-01_dp, -2.1473925344957952E-02_dp, 3.6199744674856460E-01_dp, &
         -2.4527309632224395E-01_dp, 5.5802785639214363E-02_dp, 1.3591622850222656E+01_dp, &
         -2.0671385649481678E+01_dp, 1.1690911823198414E+01_dp, 7.2296817333192265E-01_dp, &
         -3.9122753813128814E+00_dp, 3.9756373580542609E-01_dp, 1.5730792402830771E+00_dp, &
         -2.5426387228982428E-01_dp, -9.2821684388540859E-01_dp, 4.5454888767816393E-01_dp, &
         2.7835397546226770E-01_dp, -3.1429802698364062E-01_dp, 8.3881632785542004E-02_dp, &
         2.0708310558399841E+01_dp, -3.5573130360984386E+01_dp, 2.4173095046726132E+01_dp, &
         -1.9624509389026361E+00_dp, -7.1194347943715250E+00_dp, 2.2395796302706463E+00_dp, &
         2.5414676107335459E+00_dp, -1.0901424546362217E+00_dp, -1.6791478429090076E+00_dp, &
         1.6519357564458712E+00_dp, -2.9686578874346814E-01_dp, -2.3345958133610148E-01_dp, &
         9.5629002374604352E-02_dp, 3.1849060054820505E+01_dp, -6.1216326263024023E+01_dp, &
         4.8955578605990361E+01_dp, -1.0495665728696784E+01_dp, -1.1727745979077595E+01_dp, &
         6.8444384400614950E+00_dp, 3.3437289894171416E+00_dp, -3.0185515513605332E+00_dp, &
         -2.5103650595979259E+00_dp, 4.0981205243496870E+00_dp, -1.9744886540859123E+00_dp, &
         2.7900551212540414E-01_dp, 3.9609099066905094E-02_dp, 4.9480572244963817E+01_dp, &
         -1.0554295952164991E+02_dp, 9.7784802913758284E+01_dp, -3.3609900633466118E+01_dp, &
         -1.6293722769181993E+01_dp, 1.6978324437699740E+01_dp, 2.3989997683487170E+00_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/-6.8292295875978279E+00_dp, &
         -2.5859581635928466E+00_dp, 8.2643295812249669E+00_dp, -5.7320634779870945E+00_dp, &
         1.7650956249120218E+00_dp, -1.9567508222089966E-01_dp, 7.7834138139689216E+01_dp, &
         -1.8267266020330240E+02_dp, 1.9349073152056008E+02_dp, -9.1481588588459843E+01_dp, &
         -1.3822783835670613E+01_dp, 3.6637703347062448E+01_dp, -4.7917952106854855E+00_dp, &
         -1.2983441330102631E+01_dp, 5.6026423699990802E-01_dp, 1.3524777870542334E+01_dp, &
         -1.2577242006230376E+01_dp, 5.0578704679411794E+00_dp, -8.1335121215823181E-01_dp, &
         1.2469972524515586E+02_dp, -3.1853576272145068E+02_dp, 3.8049814102352741E+02_dp, &
         -2.2932721454772252E+02_dp, 1.9316804878393324E+01_dp, 6.7651019536813379E+01_dp, &
         -2.8556483274297605E+01_dp, -1.8915160882598556E+01_dp, 1.2394440997926818E+01_dp, &
         1.4599065487288630E+01_dp, -2.0997926004529241E+01_dp, 1.0415399006452484E+01_dp, &
         -1.9915861476472365E+00_dp, 2.0658312239862570E+02_dp, -5.6523320243902458E+02_dp, &
         7.4772242646605821E+02_dp, -5.4520823589717588E+02_dp, 1.4978849258832480E+02_dp, &
         9.3899472952695064E+01_dp, -8.4549171246107946E+01_dp, -1.1373667866556911E+01_dp, &
         3.7590687286378447E+01_dp, -2.2520800129803975E+00_dp, -2.1045686666495470E+01_dp, &
         1.4379152932515209E+01_dp, -3.2398302906841523E+00_dp, 3.6839922338686353E+02_dp, &
         -1.0537707025969034E+03_dp, 1.5006891025872155E+03_dp, -1.2562081672322104E+03_dp, &
         5.3825286433010865E+02_dp, 4.2498035313542289E+01_dp, -1.7156738426552047E+02_dp, &
         3.5936393290874996E+01_dp, 6.0177434607070502E+01_dp, -4.0248403369709855E+01_dp, &
         -2.7568019702721092E+00_dp, 1.1202607958032809E+01_dp, -3.3375095387496079E+00_dp, &
         7.9098125896269880E+02_dp, -2.2904299594238068E+03_dp, 3.3675571212100285E+03_dp, &
         -3.0249252348258237E+03_dp, 1.5837024358091369E+03_dp, -2.4417286558439139E+02_dp, &
         -2.4707398238911915E+02_dp, 1.1661864596469552E+02_dp, 6.5275940183027629E+01_dp, &
         -7.1150510177242779E+01_dp, 1.0564083514283231E+01_dp, 1.0686772402576475E+01_dp, &
         -4.0141425428278463E+00_dp, 2.9638192147333571E+03_dp, -8.4573529628979268E+03_dp, &
         1.2304466526150441E+04_dp, -1.1014363127026381E+04_dp, 5.8847019588549192E+03_dp, &
         -1.1791468203519848E+03_dp, -5.8911861079676396E+02_dp, 2.7195710510299079E+02_dp, &
         2.1685800896742452E+02_dp, -1.7746143301747480E+02_dp, -1.4574518198891317E+01_dp, &
         5.4805869945752612E+01_dp, -1.6433623764473392E+01_dp, 2.2036130805159376E-01_dp, &
         -4.5076804566774459E-02_dp, -1.6076370477688957E-02_dp, 5.5534315752250891E-03_dp, &
         5.0461774412725036E-03_dp, -2.0604744482374273E-03_dp, -2.2689084798617320E-03_dp, &
         5.9099274703236049E-04_dp, 2.3874131512182165E-03_dp, -2.1618256480298356E-03_dp, &
         4.3684618125611787E-04_dp, 2.5191107660970112E-04_dp, -1.0861832710107730E-04_dp, &
         7.0865218236726824E-01_dp, -1.6570352438040939E-01_dp, -5.2249512547067715E-02_dp, &
         2.4669238373278793E-02_dp, 1.7087628445085885E-02_dp, -9.6906052918900448E-03_dp, &
         -7.7468377444847780E-03_dp, 3.0933546649085024E-03_dp, 9.2517865757837520E-03_dp, &
         -1.0056570203112576E-02_dp, 3.4720383247476838E-03_dp, 1.6952014914339465E-04_dp, &
         -2.7523734696923005E-04_dp, 1.3534864020651076E+00_dp, -3.9336467243616868E-01_dp, &
         -9.8656745599151094E-02_dp, 7.3087777433389595E-02_dp, 3.4093865889400041E-02_dp, &
         -3.0490786739633617E-02_dp, -1.5507854335106205E-02_dp, 1.0740733786632572E-02_dp, &
         2.3081322466090199E-02_dp, -3.1460955605477889E-02_dp, 1.5456737064545917E-02_dp, &
         -2.6763084092928935E-03_dp, -1.2381650776694105E-04_dp, 2.3090234088808299E+00_dp, &
         -8.5776891469241912E-01_dp, -1.5338027586293296E-01_dp, 1.8991620137293749E-01_dp, &
         5.4828916209970144E-02_dp, -8.3025281965828188E-02_dp, -2.4296812195096610E-02_dp, &
         3.1469074113093229E-02_dp, 5.0330137251257137E-02_dp, -8.5515406097693397E-02_dp, &
         5.1866160160934108E-02_dp, -1.4179706064901558E-02_dp, 1.2560395955065946E-03_dp, &
         3.8197475422582392E+00_dp, -1.8037294425434220E+00_dp, -1.8899434601612383E-01_dp, &
         4.5191973881055397E-01_dp, 6.3713987378258216E-02_dp, -2.0444755855365859E-01_dp, &
         -2.4088171877681655E-02_dp, 8.2291692084297763E-02_dp, 9.7362650813698964E-02_dp, &
         -2.0934041478609497E-01_dp, 1.4766839822240871E-01_dp, -4.9288591476610653E-02_dp, &
         6.4499622528296279E-03_dp, 6.2966160150403754E+00_dp, -3.7109866191524525E+00_dp, &
         -1.0522184486152265E-01_dp, 1.0085149071581117E+00_dp, 2.5323534878071412E-03_dp, &
         -4.6630138967976670E-01_dp, 2.1076711993510598E-02_dp, 1.9828760927368860E-01_dp, &
         1.5998299484831749E-01_dp, -4.6825207004727926E-01_dp, 3.7645280036684625E-01_dp, &
         -1.4285165419296325E-01_dp, 2.2001423131429464E-02_dp, 1.0454409969612419E+01_dp, &
         -7.5375644876744872E+00_dp, 3.9449043473913742E-01_dp, 2.1412348960549363E+00_dp, &
         -3.0355677655830265E-01_dp, -9.9615781067345155E-01_dp, 2.1711098138944293E-01_dp, &
         4.4692476224298294E-01_dp, 1.8591520945768222E-01_dp, -9.5485093882726246E-01_dp, &
         8.7779557412188514E-01_dp, -3.6914472091627620E-01_dp, 6.2977146118752103E-02_dp, &
         1.7567505895441929E+01_dp, -1.5225622537868910E+01_dp, 2.1193499070085169E+00_dp, &
         4.3510998858589849E+00_dp, -1.3318083138416941E+00_dp, -1.9898827710887639E+00_dp, &
         8.4640106489365297E-01_dp, 9.4249431258822869E-01_dp, -3.7448636838545483E-02_dp, &
         -1.7262268981146505E+00_dp, 1.8723774805378806E+00_dp, -8.6854820742115435E-01_dp, &
         1.6062602638927947E-01_dp, 2.9958154145281728E+01_dp, -3.0775664710589034E+01_dp, &
         7.2009528080278491E+00_dp, 8.4351100129247225E+00_dp, -4.3138466133479865E+00_dp, &
         -3.6417064107767114E+00_dp, 2.6101763292430844E+00_dp, 1.8247368979816314E+00_dp, &
         -1.1799262417500072E+00_dp, -2.5186933824400533E+00_dp, 3.5685918319046439E+00_dp, &
         -1.8540412040242573E+00_dp, 3.7039085515950843E-01_dp, 5.1951978548581039E+01_dp, &
         -6.2581108665881274E+01_dp, 2.1182571267411070E+01_dp, 1.5325960406620894E+01_dp, &
         -1.2296597851170151E+01_dp, -5.7213940765686004E+00_dp, 7.1318501498457465E+00_dp, &
         3.0496412877399690E+00_dp, -5.0553615396661806E+00_dp, -1.8070209122459882E+00_dp, &
         5.6345736738491370E+00_dp, -3.4763373861946438E+00_dp, 7.6167100749572603E-01_dp, &
         9.1798276836285979E+01_dp, -1.2863475356597820E+02_dp, 5.8496101669050397E+01_dp, &
         2.4744077629967428E+01_dp, -3.2497120098098861E+01_dp, -5.8945347267566541E+00_dp, &
         1.7820196379746736E+01_dp, 3.4558058060427674E+00_dp, -1.5893088806805018E+01_dp, &
         5.7568757129573056E+00_dp, 5.3329494587249027E+00_dp, -5.1537631328962918E+00_dp, &
         1.3188423737811730E+00_dp, 1.6564728585867815E+02_dp, -2.6848193454407402E+02_dp, &
         1.5714914105375516E+02_dp, 2.9015793106076185E+01_dp, -8.0936336478366655E+01_dp, &
         6.0337388562561021E+00_dp, 4.0541279455914804E+01_dp, -2.5358711523755098E+00_dp, &
         -4.1489451794467044E+01_dp, 3.4994248985293758E+01_dp, -7.6772498820742348E+00_dp, &
         -3.3389802476741770E+00_dp, 1.5472908363224536E+00_dp, 3.0609437432716891E+02_dp, &
         -5.7157248282740568E+02_dp, 4.1925052593925301E+02_dp, -1.1600417484219872E+01_dp, &
         -1.8849689345974764E+02_dp, 6.6110045063594768E+01_dp, 7.9669860475699736E+01_dp, &
         -3.4687343159162722E+01_dp, -8.8625160627562337E+01_dp, 1.1831559439724820E+02_dp, &
         -6.2643850352276246E+01_dp, 1.4313449464928395E+01_dp, -6.9462104331569086E-01_dp, &
         5.8152372234088853E+02_dp, -1.2471092561355499E+03_dp, 1.1264512659705742E+03_dp, &
         -2.6350048877660970E+02_dp, -3.8978177911431425E+02_dp, 2.8606115420791843E+02_dp, &
         1.0671832025528735E+02_dp, -1.4850079552469927E+02_dp, -1.2702723207166953E+02_dp, &
         2.9628275973114745E+02_dp, -2.1971106597050107E+02_dp, 7.9255756055034652E+01_dp, &
         -1.1738659870534550E+01_dp, 1.1446896056961609E+03_dp, -2.8062684434259190E+03_dp, &
         3.0792684302402877E+03_dp, -1.3773330948044859E+03_dp, -5.6315721236341506E+02_dp, &
         9.4384383423094675E+02_dp, -9.1298241779040950E+01_dp, -4.3161000322227324E+02_dp, &
         5.6869256603055426E+01_dp, 4.6537109542580185E+02_dp, -4.9778545245192629E+02_dp, &
         2.2723211423906713E+02_dp, -4.1682116814902678E+01_dp, 2.3931486245110582E+03_dp, &
         -6.6248847444164976E+03_dp, 8.6607172498776090E+03_dp, -5.7963496001557987E+03_dp, &
         5.6074777366093997E+02_dp, 2.2520445132559971E+03_dp, -1.3103161337837334E+03_dp, &
         -5.8712209332349289E+02_dp, 8.9204271290751376E+02_dp, -1.1639397991700832E+01_dp, &
         -5.0558768097343659E+02_dp, 3.3540358542052877E+02_dp, -7.4587683513078233E+01_dp, &
         6.0278043205707454E+03_dp, -1.8260441253553501E+04_dp, 2.7198231026846435E+04_dp, &
         -2.3440178398044740E+04_dp, 9.7072987485955946E+03_dp, 1.9816258432428326E+03_dp, &
         -4.2204293880177993E+03_dp, 5.7832304639916708E+02_dp, 1.8692380816063537E+03_dp, &
         -1.2109634326915912E+03_dp, -5.4026291769502137E+01_dp, 3.1528828654208075E+02_dp, &
         -9.4786678043392129E+01_dp, 1.7937094180089550E+00_dp, -3.7993808012512725E-01_dp, &
         -1.3139337881541585E-01_dp, 4.9430879424174734E-02_dp, 4.1715801584312942E-02_dp, &
         -1.8655007611371892E-02_dp, -1.8806529335258913E-02_dp, 5.5337190143686437E-03_dp, &
         2.0442542773902748E-02_dp, -1.9495932303552625E-02_dp, 4.7817782309535614E-03_dp, &
         1.6859794985952033E-03_dp, -8.4516904829219403E-04_dp, 2.1832690683719891E+00_dp, &
         -6.3377544246855000E-01_dp, -1.6248271996603472E-01_dp, 1.1690489216647633E-01_dp, &
         5.6807251626275680E-02_dp, -4.8432873661308128E-02_dp, -2.6022849299146757E-02_dp, &
         1.6853696869085871E-02_dp, 3.7648457313444655E-02_dp, -4.9878575440930099E-02_dp, &
         2.3642088947003155E-02_dp, -3.6336297159725440E-03_dp, -3.3755205819095051E-04_dp, &
         3.0753249432146119E+00_dp, -1.2701520583639132E+00_dp, -2.0849770294749378E-01_dp, &
         2.9526772321314387E-01_dp, 7.8995542205080829E-02_dp, -1.2972869147593624E-01_dp, &
         -3.5582090248904033E-02_dp, 4.9305013374453963E-02_dp, 7.7353411801215979E-02_dp, &
         -1.3353482456470839E-01_dp, 8.1872538186610708E-02_dp, -2.2734499256472550E-02_dp, &
         2.0934547665217729E-03_dp, 4.7271509364185604E+00_dp, -2.6008689221752657E+00_dp, &
         -2.1378302448991618E-01_dp, 6.8829890166551655E-01_dp, 7.8155780076278006E-02_dp, &
         -3.1427392410807997E-01_dp, -2.9055243870744554E-02_dp, 1.2728338790291072E-01_dp, &
         1.4542960480058301E-01_dp, -3.2288152522122560E-01_dp, 2.3145832634114052E-01_dp, &
         -7.8570015248536254E-02_dp, 1.0527188402678012E-02_dp, 7.6239589870130429E+00_dp, &
         -5.2716113961903401E+00_dp, -1.1543925130729318E-02_dp, 1.5034494543003840E+00_dp, &
         -4.1512319049193738E-02_dp, -7.0326534010639863E-01_dp, 5.2072123155519456E-02_dp, &
         3.0170745289904521E-01_dp, 2.3013280605622619E-01_dp, -7.1093920590192394E-01_dp, &
         5.8207150943308372E-01_dp, -2.2424295809781689E-01_dp, 3.5093648865483100E-02_dp, &
         1.2651454564011702E+01_dp, -1.0594016000348530E+01_dp, 8.6079526700003395E-01_dp, &
         3.1309762712299110E+00_dp, -5.5343393928245810E-01_dp, -1.4752776243845915E+00_dp, &
         3.7341588237894630E-01_dp, 6.7003976934246967E-01_dp, 2.4249290952305499E-01_dp, &
         -1.4249882020825018E+00_dp, 1.3395429928449589E+00_dp, -5.7182949222827995E-01_dp, &
         9.8841365839720668E-02_dp, 2.1429965966478342E+01_dp, -2.1247628542515503E+01_dp, &
         3.6287863654310013E+00_dp, 6.2581819127944680E+00_dp, -2.1863113529437315E+00_dp, &
         -2.8967923413102912E+00_dp, 1.3681886836307373E+00_dp, 1.3955776856915574E+00_dp, &
         -1.5897239290899864E-01_dp, -2.5212216600023445E+00_dp, 2.8219668706971945E+00_dp, &
         -1.3307700093912902E+00_dp, 2.4913514170866757E-01_dp, 3.6963550418790938E+01_dp, &
         -4.2781735704421529E+01_dp, 1.1468857291085394E+01_dp, 1.1956021064279158E+01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c04 = (/-6.7963494820353745E+00_dp, &
         -5.2037548606327730E+00_dp, 4.1015955423288375E+00_dp, 2.6719939603737344E+00_dp, &
         -2.0314435192987417E+00_dp, -3.5279015440942261E+00_dp, 5.3011832033957269E+00_dp, &
         -2.8122613024625087E+00_dp, 5.6915484145040629E-01_dp, 6.4926664043803811E+01_dp, &
         -8.6899342041810002E+01_dp, 3.2560767520934299E+01_dp, 2.1401514791066290E+01_dp, &
         -1.8945123297568252E+01_dp, -7.9472372366379886E+00_dp, 1.1031467887006638E+01_dp, &
         4.4055773998180028E+00_dp, -8.2307667741735777E+00_dp, -1.9496849372131122E+00_dp, &
         8.1532056320068413E+00_dp, -5.2079110157651822E+00_dp, 1.1601051300795284E+00_dp, &
         1.1627932827440712E+02_dp, -1.7882190134095538E+02_dp, 8.8065313863488811E+01_dp, &
         3.3852987626345111E+01_dp, -4.9425190938476227E+01_dp, -7.4515204120012459E+00_dp, &
         2.7343411149027549E+01_dp, 4.8027447886962538E+00_dp, -2.5480067206341694E+01_dp, &
         1.0841781549031017E+01_dp, 6.8639785168072986E+00_dp, -7.5117713591262296E+00_dp, &
         1.9816437521619110E+00_dp, 2.1273862203414356E+02_dp, -3.7428369024165295E+02_dp, &
         2.3346238297745839E+02_dp, 3.7654110424279338E+01_dp, -1.2236868407636074E+02_dp, &
         1.1842889589859949E+01_dp, 6.2188087891764269E+01_dp, -4.6949238597179814E+00_dp, &
         -6.6545313811818971E+01_dp, 5.8913671022159285E+01_dp, -1.5425914307453965E+01_dp, &
         -3.9040075343478216E+00_dp, 2.2156462706329005E+00_dp, 3.9862923384302263E+02_dp, &
         -8.0001729806986441E+02_dp, 6.1729015727077262E+02_dp, -2.5059892452212821E+01_dp, &
         -2.8573151086745020E+02_dp, 1.0461889229880697E+02_dp, 1.2389826235882019E+02_dp, &
         -5.4939060293644125E+01_dp, -1.4442389426530306E+02_dp, 1.9642382267159812E+02_dp, &
         -1.0695650576544364E+02_dp, 2.5875876655613105E+01_dp, -1.6417786429176835E+00_dp, &
         7.6777159359322945E+02_dp, -1.7537133335705055E+03_dp, 1.6482226768525409E+03_dp, &
         -3.9211708766427637E+02_dp, -6.0392305987307441E+02_dp, 4.4333559405067615E+02_dp, &
         1.7787450181484434E+02_dp, -2.3734798751752581E+02_dp, -2.2095920452471174E+02_dp, &
         5.0350326325990295E+02_dp, -3.7436592829883779E+02_dp, 1.3589060804004382E+02_dp, &
         -2.0292510387655746E+01_dp, 1.5296195207071578E+03_dp, -3.9622350559145198E+03_dp, &
         4.4857786928390806E+03_dp, -1.9870420059955213E+03_dp, -9.7134936492482530E+02_dp, &
         1.4998589385742246E+03_dp, -7.2892049877523178E+01_dp, -7.4631312910496581E+02_dp, &
         3.4005793563591304E+01_dp, 8.9042700432665424E+02_dp, -9.2292552288497313E+02_dp, &
         4.1790581952714399E+02_dp, -7.6433006503481721E+01_dp, 3.2060087628137712E+03_dp, &
         -9.3159464375811276E+03_dp, 1.2527209661291899E+04_dp, -8.3053185228384245E+03_dp, &
         2.0211803182403949E+02_dp, 4.0355629264948511E+03_dp, -2.0304106473113000E+03_dp, &
         -1.4364497108817213E+03_dp, 1.8398623005265499E+03_dp, -9.3541599607360126E+00_dp, &
         -1.0484935445809742E+03_dp, 6.9858483574260197E+02_dp, -1.5661215432484158E+02_dp, &
         7.5591031206278803E+03_dp, -2.3841921827888185E+04_dp, 3.6673405643627026E+04_dp, &
         -3.2125806874645870E+04_dp, 1.2397751752493754E+04_dp, 5.3198673575983330E+03_dp, &
         -8.5053962353377374E+03_dp, 1.3907593831382503E+03_dp, 4.2903518293564521E+03_dp, &
         -3.9548845908239964E+03_dp, 1.3227466776399026E+03_dp, -1.3766911144115962E+01_dp, &
         -7.0037204280174294E+01_dp, 2.8049596976249963E+04_dp, -8.9761481908365226E+04_dp, &
         1.4403023637372538E+05_dp, -1.4044772416853864E+05_dp, 7.9387433338889328E+04_dp, &
         -1.2949273069561426E+04_dp, -1.4404242828698498E+04_dp, 6.6705867587399225E+03_dp, &
         5.5214430686198393E+03_dp, -6.3725355581415643E+03_dp, 1.8293089995376904E+03_dp, &
         3.3442777562420838E+02_dp, -2.1590463856083068E+02_dp, 2.4597046908671585E-01_dp, &
         -2.4131515313491576E-02_dp, -1.4593663353954249E-02_dp, 2.2311800236706780E-03_dp, &
         4.4466949112589912E-03_dp, -1.8404928039251618E-04_dp, -2.9927966501996396E-03_dp, &
         2.6735816123596879E-04_dp, 2.7889490611109772E-03_dp, -2.2452342695937332E-03_dp, &
         3.6231203373800745E-04_dp, 3.0171331853223634E-04_dp, -1.1748533887842431E-04_dp, &
         8.0512062220512370E-01_dp, -9.3210652061862007E-02_dp, -5.4083956239739717E-02_dp, &
         1.1275869730545569E-02_dp, 1.7399565786205721E-02_dp, -1.9029409691100781E-03_dp, &
         -1.1869164318760249E-02_dp, 1.5167816775997040E-03_dp, 1.1742529707275380E-02_dp, &
         -1.0427559162066466E-02_dp, 2.5937524919136981E-03_dp, 7.7071430704681043E-04_dp, &
         -4.0179192653180969E-04_dp, 1.5913824833072012E+00_dp, -2.3872493245015622E-01_dp, &
         -1.2969157416835858E-01_dp, 3.8869865775083785E-02_dp, 4.4967623155562533E-02_dp, &
         -9.6244127552489699E-03_dp, -3.1355874505723864E-02_dp, 5.8983064535336914E-03_dp, &
         3.3784953332417815E-02_dp, -3.3869634002932601E-02_dp, 1.1722170608515301E-02_dp, &
         2.1904742245513310E-04_dp, -7.8368264635060559E-04_dp, 2.8504450774327998E+00_dp, &
         -5.6612512571131524E-01_dp, -2.8459684321944051E-01_dp, 1.1706520518696334E-01_dp, &
         1.0585372455591220E-01_dp, -3.5695445721860848E-02_dp, -7.5839023807993372E-02_dp, &
         1.9812098408427733E-02_dp, 8.9483211008635308E-02_dp, -1.0040209061808522E-01_dp, &
         4.2905149773221109E-02_dp, -5.0125013089657635E-03_dp, -1.0293435042649617E-03_dp, &
         5.0103582234362856E+00_dp, -1.2983685620824574E+00_dp, -5.9781702803632175E-01_dp, &
         3.2458841568543156E-01_dp, 2.3547324163176134E-01_dp, -1.1375382180367667E-01_dp, &
         -1.7422262057421442E-01_dp, 6.1111170753431865E-02_dp, 2.2598196223392233E-01_dp, &
         -2.8189868001107560E-01_dp, 1.4003280395960366E-01_dp, -2.7702266029124158E-02_dp, &
         2.3088321236861272E-04_dp, 8.8649193387428031E+00_dp, -2.9256588227359037E+00_dp, &
         -1.2152264078580886E+00_dp, 8.5696828796979208E-01_dp, 4.9899734214418356E-01_dp, &
         -3.3446674272664295E-01_dp, -3.8401558036591399E-01_dp, 1.7999675719079797E-01_dp, &
         5.5182828585602572E-01_dp, -7.6573695808554709E-01_dp, 4.2908723819998862E-01_dp, &
         -1.0932416618525242E-01_dp, 8.4551173907571678E-03_dp, 1.5946104667560101E+01_dp, &
         -6.5535743239849209E+00_dp, -2.3980297467455691E+00_dp, 2.2035136645166129E+00_dp, &
         1.0045777679055572E+00_dp, -9.4065923178633748E-01_dp, -8.1356526611789204E-01_dp, &
         5.1914828261288715E-01_dp, 1.3116012772845194E+00_dp, -2.0399962005155539E+00_dp, &
         1.2690404031225635E+00_dp, -3.7979574497355501E-01_dp, 4.2950086794989274E-02_dp, &
         2.9311091290898574E+01_dp, -1.4740506211940973E+01_dp, -4.5764593211909546E+00_dp, &
         5.6031566250743383E+00_dp, 1.8826906137057735E+00_dp, -2.5823838485985520E+00_dp, &
         -1.6411360147721636E+00_dp, 1.4915672640388642E+00_dp, 3.0268734350794677E+00_dp, &
         -5.3670886083417155E+00_dp, 3.6770733976275096E+00_dp, -1.2384352707729609E+00_dp, &
         1.6844098822094386E-01_dp, 5.5269891112628450E+01_dp, -3.3588337863064204E+01_dp, &
         -8.3060983403295587E+00_dp, 1.4245858269068547E+01_dp, 3.0679193808833980E+00_dp, &
         -7.0049866947254937E+00_dp, -3.0485764419964387E+00_dp, 4.3234572865357226E+00_dp, &
         6.6611087254448593E+00_dp, -1.3946268257843466E+01_dp, 1.0509462585713232E+01_dp, &
         -3.8962221995669863E+00_dp, 5.9561183401098095E-01_dp, 1.0733973915516557E+02_dp, &
         -7.8192038336418292E+01_dp, -1.3621080023967423E+01_dp, 3.6519885331470277E+01_dp, &
         3.1798382551855648E+00_dp, -1.8903100012045829E+01_dp, -4.6635840130611124E+00_dp, &
         1.2767153185864498E+01_dp, 1.3205696710402899E+01_dp, -3.5455541355133271E+01_dp, &
         2.9604035476104617E+01_dp, -1.1959941279367301E+01_dp, 1.9943808572689838E+00_dp, &
         2.1574110229087182E+02_dp, -1.8759091378514898E+02_dp, -1.6548194406177625E+01_dp, &
         9.5013977871018469E+01_dp, -5.4491937574046752E+00_dp, -5.0801726334298664E+01_dp, &
         -2.6993423034851483E+00_dp, 3.8656939223630474E+01_dp, 1.8853421819247718E+01_dp, &
         -8.5648508663829901E+01_dp, 8.1175335514105896E+01_dp, -3.5754750625577735E+01_dp, &
         6.4231041798051232E+00_dp, 4.5152529264002453E+02_dp, -4.6839884015400759E+02_dp, &
         7.0294758026416391E+00_dp, 2.5195876228718237E+02_dp, -5.8200098802286071E+01_dp, &
         -1.3476486579792476E+02_dp, 2.3084889406857080E+01_dp, 1.1999084006699820E+02_dp, &
         -1.4733025794216744E+01_dp, -1.7989156147155072E+02_dp, 2.0847512854987923E+02_dp, &
         -1.0186724973687066E+02_dp, 1.9730960727082149E+01_dp, 9.9218910784116417E+02_dp, &
         -1.2321348321492649E+03_dp, 1.7515742609839145E+02_dp, 6.8070921799959092E+02_dp, &
         -3.1209102904765376E+02_dp, -3.4057274892355639E+02_dp, 1.6687726273187675E+02_dp, &
         3.7518950074597132E+02_dp, -3.1475910486719397E+02_dp, -2.1096647744491858E+02_dp, &
         4.4206246642368058E+02_dp, -2.5724080612716710E+02_dp, 5.5041848263909372E+01_dp, &
         2.3157777551412210E+03_dp, -3.4705465948827878E+03_dp, 1.0560938859361700E+03_dp, &
         1.8457238342746532E+03_dp, -1.4757403839796784E+03_dp, -7.0935408588748635E+02_dp, &
         8.4504740826609702E+02_dp, 1.0995279604639743E+03_dp, -1.9951407194066446E+03_dp, &
         8.8275263167062758E+02_dp, 3.0984268568964131E+02_dp, -4.2142428256752106E+02_dp, &
         1.1510935171638715E+02_dp, 5.8410998191536091E+03_dp, -1.0715057679305270E+04_dp, &
         5.5881348394814086E+03_dp, 4.6493931663950134E+03_dp, -6.8728966621051050E+03_dp, &
         -4.1924139258334478E+01_dp, 3.7206102777900010E+03_dp, 2.1202646040436580E+03_dp, &
         -9.2875013030481878E+03_dp, 9.2985656387185409E+03_dp, -4.4824184492676550E+03_dp, &
         9.6577982916343615E+02_dp, -4.0485333749672158E+01_dp, 1.6398745921063954E+04_dp, &
         -3.7596389625372154E+04_dp, 3.1336122401268269E+04_dp, 5.4048022371223969E+03_dp, &
         -3.1578107170577830E+04_dp, 1.7491375187499234E+04_dp, 1.0558791507837142E+04_dp, &
         -8.6578643507329816E+03_dp, -2.1657103563023124E+04_dp, 4.1477022148521210E+04_dp, &
         -3.2576961817922802E+04_dp, 1.3109718009568427E+04_dp, -2.2341346231179591E+03_dp, &
         5.5933770441125977E+04_dp, -1.6151312582203400E+05_dp, 2.0784369795231955E+05_dp, &
         -1.0524097468765127E+05_dp, -7.7042400583628056E+04_dp, 1.7217692234932515E+05_dp, &
         -1.1629573180003346E+05_dp, 1.1490194503876031E+04_dp, 2.5783214909699680E+04_dp, &
         1.0864388170919431E+03_dp, -2.2701386651906592E+04_dp, 1.5634205854882433E+04_dp, &
         -3.5877583180293955E+03_dp, 2.0109439642887108E+00_dp, -2.0605705561447749E-01_dp, &
         -1.2328473266793621E-01_dp, 2.0611707730536010E-02_dp, 3.8124292134014297E-02_dp, &
         -2.2565406289175886E-03_dp, -2.5747130192002296E-02_dp, 2.5568925787927566E-03_dp, &
         2.4379620109946373E-02_dp, -2.0171664468884437E-02_dp, 3.7637966228315930E-03_dp, &
         2.3586802910798275E-03_dp, -9.7522517747690776E-04_dp, 2.5650771305620559E+00_dp, &
         -3.8147293599496823E-01_dp, -2.0907944912054602E-01_dp, 6.0270406124987005E-02_dp, &
         7.2175105576574167E-02_dp, -1.4263061186970644E-02_dp, -5.0116196827266875E-02_dp, &
         8.8786858753379574E-03_dp, 5.3249935798219614E-02_dp, -5.2294415587463297E-02_dp, &
         1.7261637124157323E-02_dp, 9.2009373639893323E-04_dp, -1.3430266626453476E-03_dp, &
         3.8828170221494824E+00_dp, -8.4885642696270247E-01_dp, -4.2339961499366990E-01_dp, &
         1.8016120959522738E-01_dp, 1.6028857054099307E-01_dp, -5.4950371377315503E-02_dp, &
         -1.1487630174987833E-01_dp, 3.0013157516531445E-02_dp, 1.3585245394835330E-01_dp, &
         -1.5217460713811157E-01_dp, 6.4788928372283699E-02_dp, -7.4203095016135562E-03_dp, &
         -1.6019171078434956E-03_dp, 6.4636980648849729E+00_dp, -1.9099631989520829E+00_dp, &
         -8.6651521700026635E-01_dp, 4.9374292598778025E-01_dp, 3.4952603809371302E-01_dp, &
         -1.7396151407323862E-01_dp, -2.5934913691469813E-01_dp, 9.2007238929733004E-02_dp, &
         3.3903391371774966E-01_dp, -4.2381412446029582E-01_dp, 2.1088118199107522E-01_dp, &
         -4.1855791726591811E-02_dp, 3.8884987212805058E-04_dp, 1.1321884964418704E+01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c05 = (/-4.2539604263036663E+00_dp, &
         -1.7332279212896191E+00_dp, 1.2863899425408518E+00_dp, 7.2978346101421798E-01_dp, &
         -5.0542068060802536E-01_dp, -5.6453129220842824E-01_dp, 2.6844891959416117E-01_dp, &
         8.2065620557106000E-01_dp, -1.1421990003730462E+00_dp, 6.4148861901481036E-01_dp, &
         -1.6396257011823606E-01_dp, 1.2796070085478475E-02_dp, 2.0487853209932744E+01_dp, &
         -9.4517537769870792E+00_dp, -3.3759733637452087E+00_dp, 3.2722131653337412E+00_dp, &
         1.4507872941308217E+00_dp, -1.4073698425446566E+00_dp, -1.1842661757288730E+00_dp, &
         7.6778103478918880E-01_dp, 1.9398286469535142E+00_dp, -3.0258530426324368E+00_dp, &
         1.8858507556008846E+00_dp, -5.6545804339658101E-01_dp, 6.4128521008321809E-02_dp, &
         3.8095174335565567E+01_dp, -2.1154495441708484E+01_dp, -6.3728233078180931E+00_dp, &
         8.2563638968778950E+00_dp, 2.6905728467335490E+00_dp, -3.8356557488916927E+00_dp, &
         -2.3727163256891193E+00_dp, 2.1906341809472898E+00_dp, 4.4719906295754370E+00_dp, &
         -7.9395530803829812E+00_dp, 5.4435322355460913E+00_dp, -1.8339025028763316E+00_dp, &
         2.4941000766842886E-01_dp, 7.2794607863908539E+01_dp, -4.8093277265190174E+01_dp, &
         -1.1453054881238076E+01_dp, 2.0880285157258903E+01_dp, 4.3451760510124844E+00_dp, &
         -1.0353879135082220E+01_dp, -4.3940844971702182E+00_dp, 6.3120734481245231E+00_dp, &
         9.8973417965680408E+00_dp, -2.0647117445046018E+01_dp, 1.5538811111774422E+01_dp, &
         -5.7518348230309915E+00_dp, 8.7742951648656331E-01_dp, 1.4332617622926281E+02_dp, &
         -1.1190083363446227E+02_dp, -1.8573930061634336E+01_dp, 5.3338145508390340E+01_dp, &
         4.4631414700445591E+00_dp, -2.7868449098144676E+01_dp, -6.7549296888413908E+00_dp, &
         1.8541177811553482E+01_dp, 2.0031367391249091E+01_dp, -5.2810064519241095E+01_dp, &
         4.3877834330533581E+01_dp, -1.7658169541203652E+01_dp, 2.9331138916030355E+00_dp, &
         2.9197759925863352E+02_dp, -2.6852040701030040E+02_dp, -2.2082783044921865E+01_dp, &
         1.3846051549368181E+02_dp, -7.6189827475657532E+00_dp, -7.4936271383495338E+01_dp, &
         -4.2799690014967950E+00_dp, 5.5901437494488015E+01_dp, 3.1119328948155172E+01_dp, &
         -1.2977676359921708E+02_dp, 1.2145063411930138E+02_dp, -5.3095774315419469E+01_dp, &
         9.4800311502936196E+00_dp, 6.1887006052817526E+02_dp, -6.7028990777509409E+02_dp, &
         1.1201345750961481E+01_dp, 3.6684552949179283E+02_dp, -8.0419700058574776E+01_dp, &
         -2.0013407508924359E+02_dp, 3.0658023480604339E+01_dp, 1.7336606540154875E+02_dp, &
         -4.2234575269977981E+00_dp, -2.8652377100140529E+02_dp, 3.2016960404528265E+02_dp, &
         -1.5405177965842552E+02_dp, 2.9536421581524074E+01_dp, 1.3750163796936240E+03_dp, &
         -1.7589799109221406E+03_dp, 2.3961277028254179E+02_dp, 9.9301157485970555E+02_dp, &
         -4.2575252374052224E+02_dp, -5.1817603461276519E+02_dp, 2.2602380506510636E+02_dp, &
         5.4799637021657588E+02_dp, -3.8063255724265480E+02_dp, -4.3209967552717637E+02_dp, &
         7.3643375490972539E+02_dp, -4.1008578171950961E+02_dp, 8.5854890766784024E+01_dp, &
         3.2350942724992897E+03_dp, -4.9196762475376900E+03_dp, 1.4005845015490424E+03_dp, &
         2.7240803683332629E+03_dp, -1.9809371044851687E+03_dp, -1.1912780067707383E+03_dp, &
         1.1497315070977018E+03_dp, 1.6931690976040784E+03_dp, -2.6191658279561266E+03_dp, &
         6.8165720237798473E+02_dp, 9.6374997813179459E+02_dp, -8.3276968242760586E+02_dp, &
         2.0676706863037907E+02_dp, 8.1777664155924858E+03_dp, -1.4947471572678814E+04_dp, &
         7.1193687656003467E+03_dp, 7.2704699942482303E+03_dp, -9.0701497685694721E+03_dp, &
         -1.3829479450049473E+03_dp, 5.2343291775954631E+03_dp, 4.3188817279079030E+03_dp, &
         -1.3395512259314450E+04_dp, 1.1338333668387049E+04_dp, -4.0650141614603253E+03_dp, &
         1.8500545674922793E+02_dp, 1.8404119432961565E+02_dp, 2.2712516525332092E+04_dp, &
         -5.0737374807702967E+04_dp, 3.7414818324359119E+04_dp, 1.5002435412785790E+04_dp, &
         -4.2430746564604517E+04_dp, 1.2806412701410625E+04_dp, 2.0720578130718106E+04_dp, &
         -2.6832946122836898E+02_dp, -5.0834266091037163E+04_dp, 7.0480937773566868E+04_dp, &
         -4.6917539041131182E+04_dp, 1.6400456910054389E+04_dp, -2.4239470406997216E+03_dp, &
         7.2346780322228064E+04_dp, -2.0101001503791616E+05_dp, 2.2465714313474763E+05_dp, &
         -4.1692320974158160E+04_dp, -1.8428799273171904E+05_dp, 1.9351303403408095E+05_dp, &
         6.5115811053828320E+03_dp, -1.3528658618327085E+05_dp, 3.3663152446055996E+04_dp, &
         1.3655504570022557E+05_dp, -1.6923190168772679E+05_dp, 8.5874640094980365E+04_dp, &
         -1.7241422711442276E+04_dp, 3.1939792849337222E+05_dp, -1.0568561779921057E+06_dp, &
         1.6796432468799651E+06_dp, -1.4288114404238255E+06_dp, 2.3272192404290100E+05_dp, &
         1.0799323551139913E+06_dp, -1.6297117460196065E+06_dp, 1.3266412491764352E+06_dp, &
         -7.2330143312487204E+05_dp, 2.9143884697513317E+05_dp, -9.9073557291649820E+04_dp, &
         2.9147576845580217E+04_dp, -5.0077315833356834E+03_dp, 2.5399771849532937E-01_dp, &
         -8.3033887225757580E-03_dp, -7.6240573657738115E-03_dp, 2.8522290409030179E-03_dp, &
         -3.0540795652697220E-03_dp, 4.9663641512514419E-03_dp, -5.8165865884663133E-03_dp, &
         5.9728067889719752E-03_dp, -5.8369679736555158E-03_dp, 4.8909245070782462E-03_dp, &
         -3.0462113393202300E-03_dp, 1.2108511933957031E-03_dp, -2.2780712761746828E-04_dp, &
         8.3638364305768520E-01_dp, -3.2608326200435286E-02_dp, -2.9657073560502185E-02_dp, &
         1.1717032907011333E-02_dp, -1.1961865091520774E-02_dp, 1.9527087573284384E-02_dp, &
         -2.2960112994842900E-02_dp, 2.3490832187062813E-02_dp, -2.2880044526118807E-02_dp, &
         1.9162416281042115E-02_dp, -1.1938349942081816E-02_dp, 4.7445678978824480E-03_dp, &
         -8.9197277469357805E-04_dp, 1.6724533608911527E+00_dp, -8.5614488407303832E-02_dp, &
         -7.6755331134380211E-02_dp, 3.2773035396635355E-02_dp, -3.1314839841457949E-02_dp, &
         5.1345159212853306E-02_dp, -6.0742187908847688E-02_dp, 6.1817450219547841E-02_dp, &
         -5.9900400959394209E-02_dp, 5.0128559376177953E-02_dp, -3.1246531279548323E-02_dp, &
         1.2416023551359569E-02_dp, -2.3317856969346015E-03_dp, 3.0453512199679000E+00_dp, &
         -2.0863090315180247E-01_dp, -1.8408153868430499E-01_dp, 8.5169636305938420E-02_dp, &
         -7.6204738289630397E-02_dp, 1.2525692165315666E-01_dp, -1.4923544481462697E-01_dp, &
         1.5103590459034816E-01_dp, -1.4548513388214768E-01_dp, 1.2163468207365730E-01_dp, &
         -7.5871468061049080E-02_dp, 3.0148268152665792E-02_dp, -5.6563574453512922E-03_dp, &
         5.4638310929832095E+00_dp, -4.9229245214837142E-01_dp, -4.2703130683166335E-01_dp, &
         2.1395029978012914E-01_dp, -1.7998803117127329E-01_dp, 2.9569827835085399E-01_dp, &
         -3.5510720815802771E-01_dp, 3.5742646671849382E-01_dp, -3.4201443477799331E-01_dp, &
         2.8560882144847355E-01_dp, -1.7831621218447033E-01_dp, 7.0874969102503266E-02_dp, &
         -1.3286288180972624E-02_dp, 9.9026613863259954E+00_dp, -1.1436820473123184E+00_dp, &
         -9.7368133633730680E-01_dp, 5.2913145605271328E-01_dp, -4.1979778183307109E-01_dp, &
         6.8678629087967658E-01_dp, -8.3225011865141296E-01_dp, 8.3313008194203841E-01_dp, &
         -7.9113948287700431E-01_dp, 6.5965573759522067E-01_dp, -4.1234076860417629E-01_dp, &
         1.6399646715620800E-01_dp, -3.0723929620273597E-02_dp, 1.8310843645113124E+01_dp, &
         -2.6497670237209929E+00_dp, -2.2084953089663300E+00_dp, 1.3073414834056065E+00_dp, &
         -9.8030423668319455E-01_dp, 1.5894351154630064E+00_dp, -1.9463691136255397E+00_dp, &
         1.9378432641209999E+00_dp, -1.8235505263636096E+00_dp, 1.5173210666977606E+00_dp, &
         -9.4993178976001713E-01_dp, 3.7824140655577221E-01_dp, -7.0840946919823303E-02_dp, &
         3.4734198615946212E+01_dp, -6.1913475701008718E+00_dp, -5.0342585633204626E+00_dp, &
         3.2672480067080283E+00_dp, -2.3210081478694335E+00_dp, 3.7057412387846540E+00_dp, &
         -4.5938530888721134E+00_dp, 4.5493414807353512E+00_dp, -4.2340607625465676E+00_dp, &
         3.5125800180349342E+00_dp, -2.2034834637331495E+00_dp, 8.7903170473612890E-01_dp, &
         -1.6467198958207585E-01_dp, 6.7907060498973522E+01_dp, -1.4739076394337873E+01_dp, &
         -1.1636844538005446E+01_dp, 8.3520518383431011E+00_dp, -5.6406452024251346E+00_dp, &
         8.7907314207810714E+00_dp, -1.1054622145984020E+01_dp, 1.0892905149402702E+01_dp, &
         -1.0000415655277743E+01_dp, 8.2597416232575309E+00_dp, -5.1944403144978049E+00_dp, &
         2.0783350249549093E+00_dp, -3.8976488880608956E-01_dp, 1.3754189884920987E+02_dp, &
         -3.6112261566861719E+01_dp, -2.7506089296448486E+01_dp, 2.2081510409742627E+01_dp, &
         -1.4257561271282713E+01_dp, 2.1425371349192375E+01_dp, -2.7393711620526801E+01_dp, &
         2.6880120365914127E+01_dp, -2.4256007023950794E+01_dp, 1.9896595191742847E+01_dp, &
         -1.2550072865137675E+01_dp, 5.0440291354648172E+00_dp, -9.4830475587057206E-01_dp, &
         2.9051478597825161E+02_dp, -9.2077480881216914E+01_dp, -6.7042751616157403E+01_dp, &
         6.1122066501813805E+01_dp, -3.8074088991537131E+01_dp, 5.4229813974137763E+01_dp, &
         -7.0650491998597872E+01_dp, 6.9148772922131428E+01_dp, -6.1030422251934006E+01_dp, &
         4.9501109185729597E+01_dp, -3.1320759235557357E+01_dp, 1.2672891600799508E+01_dp, &
         -2.3940255865041160E+00_dp, 6.4557191457606052E+02_dp, -2.4763182690496433E+02_dp, &
         -1.7000441338399474E+02_dp, 1.7979797134338500E+02_dp, -1.0963259713335488E+02_dp, &
         1.4445301204007444E+02_dp, -1.9199563810236253E+02_dp, 1.8801003542743408E+02_dp, &
         -1.6118263890219302E+02_dp, 1.2820143562173234E+02_dp, -8.1268684993159169E+01_dp, &
         3.3201648029135029E+01_dp, -6.3257064933095632E+00_dp, 1.5281770790560115E+03_dp, &
         -7.1515428321482318E+02_dp, -4.5276182324024262E+02_dp, 5.7338834147225236E+02_dp, &
         -3.5031930447070647E+02_dp, 4.1265596366351411E+02_dp, -5.5821259522651553E+02_dp, &
         5.4994852484809564E+02_dp, -4.5353406400816010E+02_dp, 3.4777962951959489E+02_dp, &
         -2.1955613483579080E+02_dp, 9.0846938995140135E+01_dp, -1.7557396032513779E+01_dp, &
         3.9261873180851890E+03_dp, -2.2769021329796587E+03_dp, -1.2770843901758556E+03_dp, &
         2.0407998981381138E+03_dp, -1.2978099365872522E+03_dp, 1.3049821930118717E+03_dp, &
         -1.7706647534018759E+03_dp, 1.7719691434495785E+03_dp, -1.3872679059180307E+03_dp, &
         9.8676308339813556E+02_dp, -6.0572190633134835E+02_dp, 2.5343498559405921E+02_dp, &
         -5.0067017168943025E+01_dp, 1.1287275974519674E+04_dp, -8.3427028800217486E+03_dp, &
         -3.7990573642673103E+03_dp, 8.4927259072795641E+03_dp, -5.9962833139676068E+03_dp, &
         4.8878274555726575E+03_dp, -6.2818683353242614E+03_dp, 6.4697978795118643E+03_dp, &
         -4.7394997944678944E+03_dp, 2.8301056804072027E+03_dp, -1.5087936765466591E+03_dp, &
         6.0963560380239039E+02_dp, -1.2304467591412092E+02_dp, 3.8437627273177568E+04_dp, &
         -3.8176312227787312E+04_dp, -1.0634568511977122E+04_dp, 4.4919022886395134E+04_dp, &
         -3.9765187408990765E+04_dp, 2.6349805711576602E+04_dp, -2.5672293572137780E+04_dp, &
         2.6451194932949369E+04_dp, -1.7901001830063731E+04_dp, 6.2663747959682069E+03_dp, &
         -7.9046224464493122E+01_dp, -7.7755945708786203E+02_dp, 2.1450377554813323E+02_dp, &
         1.7818869100699757E+05_dp, -2.6537447971430060E+05_dp, 3.1712593310074910E+04_dp, &
         3.4422146844366990E+05_dp, -5.3053517524802696E+05_dp, 4.3145350047972775E+05_dp, &
         -1.9228471756741623E+05_dp, -3.1796392782362986E+04_dp, 1.7950338120613602E+05_dp, &
         -2.3038705409991340E+05_dp, 1.8179053082015848E+05_dp, -8.5539393856822993E+04_dp, &
         1.8297112165331673E+04_dp, 2.0796368651156318E+00_dp, -7.1211773323306998E-02_dp, &
         -6.5222594833814318E-02_dp, 2.4758880150403027E-02_dp, -2.6172471480240402E-02_dp, &
         4.2607134478414486E-02_dp, -4.9952200375915141E-02_dp, 5.1244059328384543E-02_dp, &
         -5.0035111663177034E-02_dp, 4.1920311429239167E-02_dp, -2.6111046961287777E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c06 = (/1.0378429926308540E-02_dp, &
         -1.9521878957809889E-03_dp, 2.6944073713370376E+00_dp, -1.3634250253884062E-01_dp, &
         -1.2248962009310110E-01_dp, 5.1739894797576326E-02_dp, -4.9861795300424228E-02_dp, &
         8.1764747263490686E-02_dp, -9.6628896510658951E-02_dp, 9.8409054684497427E-02_dp, &
         -9.5437204809848047E-02_dp, 7.9879375951563511E-02_dp, -4.9785180785078605E-02_dp, &
         1.9781814649150051E-02_dp, -3.7155106897152532E-03_dp, 4.1754065857664679E+00_dp, &
         -3.1351013649643816E-01_dp, -2.7633289430167707E-01_dp, 1.2854478131208438E-01_dp, &
         -1.1437860765449445E-01_dp, 1.8829614057148475E-01_dp, -2.2438400638247086E-01_dp, &
         2.2698373182776574E-01_dp, -2.1859649315655161E-01_dp, 1.8276068899158232E-01_dp, &
         -1.1399571200820550E-01_dp, 4.5292156705966799E-02_dp, -8.4963179977451710E-03_dp, &
         7.1321772449454741E+00_dp, -7.2704365835075646E-01_dp, -6.2939140508259084E-01_dp, &
         3.1836503791119075E-01_dp, -2.6546680650849463E-01_dp, 4.3689678639928536E-01_dp, &
         -5.2499672256707541E-01_dp, 5.2799497456180577E-01_dp, -5.0495196946337140E-01_dp, &
         4.2165902676363498E-01_dp, -2.6325905625353568E-01_dp, 1.0462371284509191E-01_dp, &
         -1.9608344465334434E-02_dp, 1.2834515530424907E+01_dp, -1.6706472138204553E+00_dp, &
         -1.4188832578032204E+00_dp, 7.7932291344618254E-01_dp, -6.1248631370805595E-01_dp, &
         1.0036315751279516E+00_dp, -1.2172307293345568E+00_dp, 1.2173794784926901E+00_dp, &
         -1.1552141520797918E+00_dp, 9.6318085874165005E-01_dp, -6.0209450103939621E-01_dp, &
         2.3943637699340395E-01_dp, -4.4845635855033761E-02_dp, 2.3907644544268969E+01_dp, &
         -3.8409238287608631E+00_dp, -3.1927866471055242E+00_dp, 1.9107508410085583E+00_dp, &
         -1.4194444132695128E+00_dp, 2.3047039849449251E+00_dp, -2.8251319104705459E+00_dp, &
         2.8099411726428549E+00_dp, -2.6421180276849054E+00_dp, 2.1983609044304298E+00_dp, &
         -1.3764093323588191E+00_dp, 5.4799175204302053E-01_dp, -1.0260464712972778E-01_dp, &
         4.5900219693354039E+01_dp, -8.9317693484926668E+00_dp, -7.2428742526117622E+00_dp, &
         4.7498551429452860E+00_dp, -3.3446060263422406E+00_dp, 5.3474540805623176E+00_dp, &
         -6.6363730788695472E+00_dp, 6.5652622867554413E+00_dp, -6.1054269491608801E+00_dp, &
         5.0652251409978781E+00_dp, -3.1778332100528144E+00_dp, 1.2675663829725419E+00_dp, &
         -2.3738207936101491E-01_dp, 9.0939009801418251E+01_dp, -2.1207951305576458E+01_dp, &
         -1.6703836717599909E+01_dp, 1.2094601423920158E+01_dp, -8.1037239824558487E+00_dp, &
         1.2652193762144261E+01_dp, -1.5927629386120495E+01_dp, 1.5678136007593519E+01_dp, &
         -1.4384696049040816E+01_dp, 1.1883309669508408E+01_dp, -7.4743470332662731E+00_dp, &
         2.9899963557394957E+00_dp, -5.6051343972032064E-01_dp, 1.8664874611056891E+02_dp, &
         -5.1882278871112675E+01_dp, -3.9458004229447603E+01_dp, 3.1854721012338533E+01_dp, &
         -2.0425886295138511E+01_dp, 3.0791578686657544E+01_dp, -3.9403681473957676E+01_dp, &
         3.8622643042566317E+01_dp, -3.4847520475636244E+01_dp, 2.8602543521658834E+01_dp, &
         -1.8045198909363659E+01_dp, 7.2502140242936921E+00_dp, -1.3622957283347905E+00_dp, &
         3.9915188049530610E+02_dp, -1.3203707855054867E+02_dp, -9.6178119362902649E+01_dp, &
         8.7705958866288142E+01_dp, -5.4284439736638681E+01_dp, 7.7802647990103949E+01_dp, &
         -1.0141935559157001E+02_dp, 9.9136132884069752E+01_dp, -8.7572824663943678E+01_dp, &
         7.1146305460401948E+01_dp, -4.5034774737857425E+01_dp, 1.8210581017182808E+01_dp, &
         -3.4368084704922492E+00_dp, 8.9640126039779705E+02_dp, -3.5362542578945090E+02_dp, &
         -2.4381286270301160E+02_dp, 2.5567043117350039E+02_dp, -1.5473035694623934E+02_dp, &
         2.0643065091772985E+02_dp, -2.7449876024024695E+02_dp, 2.6829509790813256E+02_dp, &
         -2.3063853177772037E+02_dp, 1.8419775073300784E+02_dp, -1.1690662622678340E+02_dp, &
         4.7711080741790596E+01_dp, -9.0740124097056771E+00_dp, 2.1371273099507980E+03_dp, &
         -1.0117484690800054E+03_dp, -6.4833923675694632E+02_dp, 8.0221902155722398E+02_dp, &
         -4.8409799486432814E+02_dp, 5.8407650925627354E+02_dp, -7.9151735905004318E+02_dp, &
         7.7715699922847898E+02_dp, -6.4473901677909225E+02_dp, 4.9950084061013206E+02_dp, &
         -3.1670516768758597E+02_dp, 1.3088948582201553E+02_dp, -2.5216009702956097E+01_dp, &
         5.4941634656312544E+03_dp, -3.1579566223942602E+03_dp, -1.8240594402316519E+03_dp, &
         2.7722798032040118E+03_dp, -1.7188010021120501E+03_dp, 1.8049036287011472E+03_dp, &
         -2.4720525322843669E+03_dp, 2.4575593777441154E+03_dp, -1.9454017485166187E+03_dp, &
         1.4216353652861419E+03_dp, -8.8794262472737205E+02_dp, 3.7236649124083812E+02_dp, &
         -7.3224838489909345E+01_dp, 1.5598907311734893E+04_dp, -1.1102735305658800E+04_dp, &
         -5.4496895758397304E+03_dp, 1.0920783488778621E+04_dp, -7.2896930341331999E+03_dp, &
         6.3648779067823270E+03_dp, -8.5400815400530537E+03_dp, 8.7029682626190970E+03_dp, &
         -6.4829758274370470E+03_dp, 4.1818738901113929E+03_dp, -2.4318300865338133E+03_dp, &
         1.0217539466953230E+03_dp, -2.0696183440979297E+02_dp, 5.0910784596133592E+04_dp, &
         -4.6410694622357332E+04_dp, -1.6759575861048310E+04_dp, 5.1878639177724821E+04_dp, &
         -4.0256109414878149E+04_dp, 2.8334810650844160E+04_dp, -3.3461954000327562E+04_dp, &
         3.5317418178144384E+04_dp, -2.4510772031663375E+04_dp, 1.1577080835902063E+04_dp, &
         -4.3532742678888289E+03_dp, 1.4492552333627673E+03_dp, -2.9351788438751623E+02_dp, &
         2.0646310805017679E+05_dp, -2.5649451003031401E+05_dp, -3.4599088204528663E+04_dp, &
         3.2876093443253404E+05_dp, -3.4295275058681634E+05_dp, 2.1298567642959324E+05_dp, &
         -1.5333540067239694E+05_dp, 1.4673765999310571E+05_dp, -9.1327695628929680E+04_dp, &
         4.0267201625102462E+03_dp, 3.8179083987894133E+04_dp, -2.6616546576670629E+04_dp, &
         6.3775166535376393E+03_dp, 1.2836671172308552E+06_dp, -2.4267891820420851E+06_dp, &
         9.8426137787763157E+05_dp, 3.0244705295962645E+06_dp, -6.6314140027439278E+06_dp, &
         6.7915841815625075E+06_dp, -3.0584748882667138E+06_dp, -2.3579722151722158E+06_dp, &
         6.4360152608277015E+06_dp, -7.2732554452784210E+06_dp, 5.1574036293220147E+06_dp, &
         -2.2174478505159747E+06_dp, 4.4243860364206659E+05_dp, 2.5576624471415554E-01_dp, &
         4.2237570088002974E-09_dp, -1.0860357249158544E-02_dp, 1.3291431359068883E-02_dp, &
         -1.6900731238607920E-02_dp, 1.3962484092300874E-02_dp, 5.5824330603980879E-03_dp, &
         -5.1241539388943998E-02_dp, 1.1924290064807870E-01_dp, -1.7653776055574968E-01_dp, &
         1.7399289115806521E-01_dp, -1.0429706396163735E-01_dp, 2.8742956296286825E-02_dp, &
         8.4334174286158137E-01_dp, 1.6647331740500862E-08_dp, -4.2808548044690892E-02_dp, &
         5.2391164460175478E-02_dp, -6.6127942116811278E-02_dp, 5.3832866343907902E-02_dp, &
         2.4299402435477839E-02_dp, -2.0540730598820642E-01_dp, 4.7389926254432846E-01_dp, &
         -6.9894838310754248E-01_dp, 6.8730283012545457E-01_dp, -4.1135207214867037E-01_dp, &
         1.1323604650004115E-01_dp, 1.6907728029821778E+00_dp, 4.3944714380629980E-08_dp, &
         -1.1301966435617677E-01_dp, 1.3831887050265168E-01_dp, -1.7265894786567948E-01_dp, &
         1.3739412392853392E-01_dp, 7.3149394658010375E-02_dp, -5.5568362275266503E-01_dp, &
         1.2661948729096475E+00_dp, -1.8571387585078984E+00_dp, 1.8200395995789005E+00_dp, &
         -1.0867803636088169E+00_dp, 2.9866422519059560E-01_dp, 3.0901288057755520E+00_dp, &
         1.0772079547945405E-07_dp, -2.7708675910432151E-01_dp, 3.3911191201867580E-01_dp, &
         -4.1813678737718835E-01_dp, 3.2415947918917831E-01_dp, 2.0334817882080883E-01_dp, &
         -1.3978592144512130E+00_dp, 3.1438040492554027E+00_dp, -4.5835634903417697E+00_dp, &
         4.4755448980546646E+00_dp, -2.6656753625759242E+00_dp, 7.3121673386587283E-01_dp, &
         5.5698272110483344E+00_dp, 2.5576027456667277E-07_dp, -6.5799686641669253E-01_dp, &
         8.0528746329636880E-01_dp, -9.8004848950174905E-01_dp, 7.3810118982122419E-01_dp, &
         5.4249771450196238E-01_dp, -3.4069578044619200E+00_dp, 7.5616388404833437E+00_dp, &
         -1.0956710664782413E+01_dp, 1.0657500411874638E+01_dp, -6.3307799811259393E+00_dp, &
         1.7331860358171081E+00_dp, 1.0149759614132440E+01_dp, 5.9816188704636412E-07_dp, &
         -1.5391916143113304E+00_dp, 1.8837342032506716E+00_dp, -2.2598790657906993E+00_dp, &
         1.6463608426933884E+00_dp, 1.4189272312199954E+00_dp, -8.1875793783310566E+00_dp, &
         1.7923938935972270E+01_dp, -2.5801297646966532E+01_dp, 2.4993035113196413E+01_dp, &
         -1.4803360412194611E+01_dp, 4.0440613015781999E+00_dp, 1.8885544797422838E+01_dp, &
         1.3962193082829181E-06_dp, -3.5935492300793954E+00_dp, 4.3979504208452127E+00_dp, &
         -5.1910280549710484E+00_dp, 3.6347018516401288E+00_dp, 3.7005632026129085E+00_dp, &
         -1.9673683832970518E+01_dp, 4.2439595400855616E+01_dp, -6.0651898779119989E+01_dp, &
         5.8481834431875996E+01_dp, -3.4526009414347030E+01_dp, 9.4091800758801245E+00_dp, &
         3.6082927628028706E+01_dp, 3.2901193014272859E-06_dp, -8.4702295022993539E+00_dp, &
         1.0366249728117493E+01_dp, -1.2006521585567233E+01_dp, 8.0045054262700059E+00_dp, &
         9.7574715974163553E+00_dp, -4.7844176520379150E+01_dp, 1.0156232558264223E+02_dp, &
         -1.4397490593201323E+02_dp, 1.3809637723282293E+02_dp, -8.1222582326131445E+01_dp, &
         2.2073010705828167E+01_dp, 7.1134245645863814E+01_dp, 7.9097869358485311E-06_dp, &
         -2.0369747780417786E+01_dp, 2.4929401403798764E+01_dp, -2.8233117758625095E+01_dp, &
         1.7675076772578425E+01_dp, 2.6333062643496763E+01_dp, -1.1907948950314673E+02_dp, &
         2.4831111955679535E+02_dp, -3.4875954493874565E+02_dp, 3.3248082048627452E+02_dp, &
         -1.9468856176014137E+02_dp, 5.2732162646628915E+01_dp, 1.4549672148303156E+02_dp, &
         1.9606098644651494E-05_dp, -5.0510545608226444E+01_dp, 6.1816996598134779E+01_dp, &
         -6.8124325068817015E+01_dp, 3.9197095706813819E+01_dp, 7.3632728580914488E+01_dp, &
         -3.0676194479178452E+02_dp, 6.2695742728703999E+02_dp, -8.7111493883908997E+02_dp, &
         8.2443456649865914E+02_dp, -4.8018933005153883E+02_dp, 1.2953318523991635E+02_dp, &
         3.1094648618254405E+02_dp, 5.0696414767541625E-05_dp, -1.3067261929794532E+02_dp, &
         1.5992265246911447E+02_dp, -1.7032412326959059E+02_dp, 8.6865456810714306E+01_dp, &
         2.1627551271719494E+02_dp, -8.2833324696946715E+02_dp, 1.6543740153928604E+03_dp, &
         -2.2691680798433426E+03_dp, 2.1285371190344981E+03_dp, -1.2315575770352384E+03_dp, &
         3.3052109092285252E+02_dp, 7.0102332426653004E+02_dp, 1.3873488339462509E-04_dp, &
         -3.5783165787516657E+02_dp, 4.3792884674714333E+02_dp, -4.4617993522197833E+02_dp, &
         1.8813865534569888E+02_dp, 6.7882008059907184E+02_dp, -2.3814302002445670E+03_dp, &
         4.6290899430085301E+03_dp, -6.2490819879456340E+03_dp, 5.7958888003535612E+03_dp, &
         -3.3247298296427966E+03_dp, 8.8628594048774914E+02_dp, 1.6902170532998441E+03_dp, &
         4.0974181820914717E-04_dp, -1.0577892962329620E+03_dp, 1.2945628573020617E+03_dp, &
         -1.2412789684761844E+03_dp, 3.6518353070382403E+02_dp, 2.3309880593651797E+03_dp, &
         -7.4457710895261789E+03_dp, 1.4002929494342166E+04_dp, -1.8518696747336475E+04_dp, &
         1.6917288236393451E+04_dp, -9.5900986554442443E+03_dp, 2.5323928944806203E+03_dp, &
         4.4504521254867541E+03_dp, 1.3448249990557044E-03_dp, -3.4764938517534570E+03_dp, &
         4.2546538860367436E+03_dp, -3.7288910130907971E+03_dp, 3.3785610612894226E+02_dp, &
         9.0749096864554558E+03_dp, -2.6126306878210624E+04_dp, 4.7085181240615682E+04_dp, &
         -6.0517025375512349E+04_dp, 5.4074356432769353E+04_dp, -3.0109150792826324E+04_dp, &
         7.8344472467430223E+03_dp, 1.3254579477277694E+04_dp, 5.1532479271565908E-03_dp, &
         -1.3351228791461199E+04_dp, 1.6339623735533780E+04_dp, -1.2301930953556473E+04_dp, &
         -3.6684681819111252E+03_dp, 4.2559142880279360E+04_dp, -1.0831862123749210E+05_dp/)
      REAL(KIND=dp), DIMENSION(13, 34, 6), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04, c05, c06, c07/), (/13, 34, 6/))

      INTEGER                                            :: irange

      IF (Rc <= 2.0000000000000000E+01_dp) Rc = 2.0000000000000000E+01_dp
      IF (Rc >= 1.3327491760253906E+06_dp) Rc = 1.3327491760253906E+06_dp
      IF (Rc >= 2.0000000000000000E+01_dp .AND. Rc <= 3.0823724363668457E+01_dp) THEN
         irange = 1
         L_b = 2.0000000000000000E+01_dp
         U_b = 3.0823724363668457E+01_dp
      END IF
      IF (Rc >= 3.0823724363668457E+01_dp .AND. Rc <= 3.1985602006226213E+02_dp) THEN
         irange = 2
         L_b = 3.0823724363668457E+01_dp
         U_b = 3.1985602006226213E+02_dp
      END IF
      IF (Rc >= 3.1985602006226213E+02_dp .AND. Rc <= 4.5183964268225072E+03_dp) THEN
         irange = 3
         L_b = 3.1985602006226213E+02_dp
         U_b = 4.5183964268225072E+03_dp
      END IF
      IF (Rc >= 4.5183964268225072E+03_dp .AND. Rc <= 7.8644668751788398E+04_dp) THEN
         irange = 4
         L_b = 4.5183964268225072E+03_dp
         U_b = 7.8644668751788398E+04_dp
      END IF
      IF (Rc >= 7.8644668751788398E+04_dp .AND. Rc <= 4.6977420089882286E+05_dp) THEN
         irange = 5
         L_b = 7.8644668751788398E+04_dp
         U_b = 4.6977420089882286E+05_dp
      END IF
      IF (Rc >= 4.6977420089882286E+05_dp .AND. Rc <= 1.3327491760253906E+06_dp) THEN
         irange = 6
         L_b = 4.6977420089882286E+05_dp
         U_b = 1.3327491760253906E+06_dp
      END IF

      fit_coef(1:13, 1:34) = coefdata(1:13, 1:34, irange)

   END SUBROUTINE get_fit_coef_k17

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k18(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 36)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(340), PARAMETER :: c06 = (/-4.2209902072425757E+02_dp, &
         1.1367036683435998E+02_dp, 2.4004100563047302E+02_dp, 2.6442809338541683E-05_dp, &
         -9.2920139462935452E+01_dp, 1.1857226657184384E+02_dp, -1.3712334751086243E+02_dp, &
         8.9614212066663356E+01_dp, 1.2954531399327851E+02_dp, -6.1204117014894439E+02_dp, &
         1.3038237052684744E+03_dp, -1.8532736683780777E+03_dp, 1.7733881888322414E+03_dp, &
         -1.0349286327234870E+03_dp, 2.7775933648182263E+02_dp, 5.1299854071485697E+02_dp, &
         6.7687074050697422E-05_dp, -2.3807455011160289E+02_dp, 3.0379853851555936E+02_dp, &
         -3.3999113773695723E+02_dp, 2.0053502227186451E+02_dp, 3.8602011516394481E+02_dp, &
         -1.6462886216839220E+03_dp, 3.4237001138452265E+03_dp, -4.8064857217217213E+03_dp, &
         4.5625744054824872E+03_dp, -2.6475930973206077E+03_dp, 7.0759015679511151E+02_dp, &
         1.1565450996747870E+03_dp, 1.8361085330211829E-04_dp, -6.4662037045130342E+02_dp, &
         8.2512803376189629E+02_dp, -8.8496728903978203E+02_dp, 4.4602324539501097E+02_dp, &
         1.2288632621141262E+03_dp, -4.7252668279211357E+03_dp, 9.5564883176047988E+03_dp, &
         -1.3215309577113516E+04_dp, 1.2419557109971771E+04_dp, -7.1549681548517901E+03_dp, &
         1.9018665194857740E+03_dp, 2.7885119751535453E+03_dp, 5.3798578681814367E-04_dp, &
         -1.8979726842190619E+03_dp, 2.4219260114581057E+03_dp, -2.4509451315420783E+03_dp, &
         9.3302092536482598E+02_dp, 4.2792080086216884E+03_dp, -1.4781335864606486E+04_dp, &
         2.8910881818664639E+04_dp, -3.9221458601856801E+04_dp, 3.6377452574481533E+04_dp, &
         -2.0754440568654329E+04_dp, 5.4759314189430270E+03_dp, 7.3423343437531503E+03_dp, &
         1.7523478403478510E-03_dp, -6.1987624682842961E+03_dp, 7.9099609072407902E+03_dp, &
         -7.3469720853921908E+03_dp, 1.3590388991922907E+03_dp, 1.6895282676913615E+04_dp, &
         -5.2015173713595323E+04_dp, 9.7530203120827064E+04_dp, -1.2891927678944364E+05_dp, &
         1.1735562908487483E+05_dp, -6.6007676931770868E+04_dp, 1.7222837048140107E+04_dp, &
         2.1867340894031433E+04_dp, 6.6610965479193367E-03_dp, -2.3670256838196394E+04_dp, &
         3.0204363587247230E+04_dp, -2.4289431365742868E+04_dp, -4.4793087397477648E+03_dp, &
         8.0399556072960419E+04_dp, -2.1698457697832960E+05_dp, 3.8372266226749582E+05_dp, &
         -4.8749358651122235E+05_dp, 4.3044669176928164E+05_dp, -2.3628550514088655E+05_dp, &
         6.0447814633299982E+04_dp, 7.8888258338212341E+04_dp, 3.2559081234041241E-02_dp, &
         -1.1677721904936734E+05_dp, 1.4901143678641456E+05_dp, -8.7205284381827863E+04_dp, &
         -1.0593928818725457E+05_dp, 5.2180063081218430E+05_dp, -1.1828235545327321E+06_dp, &
         1.8923972467739715E+06_dp, -2.2220643842051183E+06_dp, 1.8334827911676331E+06_dp, &
         -9.4845074792786303E+05_dp, 2.3035540725488932E+05_dp, 4.1277266013211373E+05_dp, &
         2.6625034408035703E-01_dp, -9.8168542137299629E+05_dp, 1.2526189136341743E+06_dp, &
         -1.6524351056354993E+04_dp, -2.7322480720883626E+06_dp, 6.3977913782139570E+06_dp, &
         -9.5302632217170522E+06_dp, 9.9340828649581820E+06_dp, -6.5415577947778786E+06_dp, &
         1.6425385275771332E+06_dp, 9.1574731282429548E+05_dp, -6.1255945160736761E+05_dp, &
         2.0360025664444423E+00_dp, 2.5713335383125262E-08_dp, -9.0033864102026753E-02_dp, &
         1.1488977052985240E-01_dp, -1.5339668231823797E-01_dp, 1.3943382095570961E-01_dp, &
         2.3403104409030068E-02_dp, -4.3679119831297458E-01_dp, 1.0806320687760871E+00_dp, &
         -1.6418760026934989E+00_dp, 1.6349608211334608E+00_dp, -9.8011211272180954E-01_dp, &
         2.6815864103304071E-01_dp, 2.6117528386499291E+00_dp, 4.8175407514969741E-08_dp, &
         -1.6873111063388158E-01_dp, 2.1531310300017245E-01_dp, -2.8376201294612502E-01_dp, &
         2.5179501088112777E-01_dp, 6.2880810206506932E-02_dp, -8.4876332472530991E-01_dp, &
         2.0624430496348749E+00_dp, -3.1112002808161887E+00_dp, 3.0855832382580046E+00_dp, &
         -1.8448645916831607E+00_dp, 5.0382953298570421E-01_dp, 3.9873063048304820E+00_dp, &
         1.0846158670211902E-07_dp, -3.7998757133131345E-01_dp, 4.8489142926013917E-01_dp, &
         -6.3086126110902885E-01_dp, 5.4610493299372398E-01_dp, 1.8329858289416012E-01_dp, &
         -1.9772478991069387E+00_dp, 4.7253027814453725E+00_dp, -7.0796876549675778E+00_dp, &
         6.9941004208554265E+00_dp, -4.1711166061111760E+00_dp, 1.1370891467073994E+00_dp, &
         6.7001043314976272E+00_dp, 2.4667185517548435E-07_dp, -8.6443414945752739E-01_dp, &
         1.1030796184932681E+00_dp, -1.4180463633457838E+00_dp, 1.1985397640306839E+00_dp, &
         5.0364134607866440E-01_dp, -4.6338532066208620E+00_dp, 1.0914232611086101E+01_dp, &
         -1.6252611187619642E+01_dp, 1.5999475073408924E+01_dp, -9.5194949221022522E+00_dp, &
         2.5908465591882055E+00_dp, 1.1851418633552694E+01_dp, 5.5547702724000489E-07_dp, &
         -1.9471816828933401E+00_dp, 2.4847416347076288E+00_dp, -3.1545013399132991E+00_dp, &
         2.5980360723724814E+00_dp, 1.3345505038407754E+00_dp, -1.0749199950188043E+01_dp, &
         2.4957927423377040E+01_dp, -3.6937125272225181E+01_dp, 3.6230950319225336E+01_dp, &
         -2.1505397766830995E+01_dp, 5.8430179763106009E+00_dp, 2.1671718972763752E+01_dp, &
         1.2490628723124012E-06_dp, -4.3799641771907298E+00_dp, 5.5891415983185286E+00_dp, &
         -6.9981143458634438E+00_dp, 5.5940275426254855E+00_dp, 3.4904237077772149E+00_dp, &
         -2.4933029773709567E+01_dp, 5.7033160660992166E+01_dp, -8.3854458025895269E+01_dp, &
         8.1931375876771739E+01_dp, -4.8504788239908393E+01_dp, 1.3154247835903428E+01_dp, &
         4.0763243915536528E+01_dp, 2.8331268185654416E-06_dp, -9.9385399289344427E+00_dp, &
         1.2682267211216036E+01_dp, -1.5631827819043766E+01_dp, 1.2059187258319630E+01_dp, &
         9.1510546139662399E+00_dp, -5.8458649242606107E+01_dp, 1.3161495724281883E+02_dp, &
         -1.9212585630556700E+02_dp, 1.8691186329645851E+02_dp, -1.1033319945556754E+02_dp, &
         2.9859331950975829E+01_dp, 7.8908010778509308E+01_dp, 6.5386761281513677E-06_dp, &
         -2.2948202615132029E+01_dp, 2.9283481126861815E+01_dp, -3.5445800826485623E+01_dp, &
         2.6183727841564750E+01_dp, 2.4329279868365493E+01_dp, -1.3982752010927283E+02_dp, &
         3.0947642362503592E+02_dp, -4.4818522192425121E+02_dp, 4.3391508597732701E+02_dp, &
         -2.5529342544325627E+02_dp, 6.8925136413432071E+01_dp, 1.5767917857157383E+02_dp, &
         1.5478496535958464E-05_dp, -5.4354086424776057E+01_dp, 6.9359488995831569E+01_dp, &
         -8.2189060026724775E+01_dp, 5.7491106121444517E+01_dp, 6.6265828876160398E+01_dp, &
         -3.4412046626243608E+02_dp, 7.4760950475972925E+02_dp, -1.0730888794956911E+03_dp, &
         1.0331995461163826E+03_dp, -6.0557217116128845E+02_dp, 1.6304262960925644E+02_dp, &
         3.2678336205766686E+02_dp, 3.7894877275203884E-05_dp, -1.3316371456228885E+02_dp, &
         1.6992573781710709E+02_dp, -1.9629816380606513E+02_dp, 1.2788120327692410E+02_dp, &
         1.8683976508725286E+02_dp, -8.7919648882012120E+02_dp, 1.8714058585605610E+03_dp, &
         -2.6590278379766187E+03_dp, 2.5438261029985429E+03_dp, -1.4843168619653752E+03_dp, &
         3.9832397479258418E+02_dp, 7.0684637540150163E+02_dp, 9.6867269795911109E-05_dp, &
         -3.4069427410525071E+02_dp, 4.3474797507118456E+02_dp, -4.8678867720092836E+02_dp, &
         2.8761368885618197E+02_dp, 5.5179921872736327E+02_dp, -2.3562414577310824E+03_dp, &
         4.9020823980878231E+03_dp, -6.8836792573968105E+03_dp, 6.5355383068211950E+03_dp, &
         -3.7929919281300995E+03_dp, 1.0138137129943455E+03_dp, 1.6094255057648668E+03_dp, &
         2.6161705713683781E-04_dp, -9.2119908813007453E+02_dp, 1.1755078805571836E+03_dp, &
         -1.2652473969070488E+03_dp, 6.4695642698297149E+02_dp, 1.7319053641298342E+03_dp, &
         -6.7103673916556281E+03_dp, 1.3602270272451808E+04_dp, -1.8834904642552094E+04_dp, &
         1.7716918035172537E+04_dp, -1.0213676976914645E+04_dp, 2.7162972544969493E+03_dp, &
         3.9033474711962631E+03_dp, 7.5844834317149079E-04_dp, -2.6748123810493139E+03_dp, &
         3.4132212597088978E+03_dp, -3.4872403633089002E+03_dp, 1.3999647101781925E+03_dp, &
         5.8893036531862190E+03_dp, -2.0663289718638232E+04_dp, 4.0631504446060877E+04_dp, &
         -5.5298982600424795E+04_dp, 5.1405733453152017E+04_dp, -2.9378611655370314E+04_dp, &
         7.7615915923671664E+03_dp, 1.0259227531224709E+04_dp, 2.4144621632299105E-03_dp, &
         -8.5342291700014030E+03_dp, 1.0890155897749342E+04_dp, -1.0343731878628403E+04_dp, &
         2.4584152986258155E+03_dp, 2.2303664961658811E+04_dp, -7.0523457767913642E+04_dp, &
         1.3364816601587992E+05_dp, -1.7787359000140551E+05_dp, 1.6273980884683644E+05_dp, &
         -9.1893860960891921E+04_dp, 2.4051527420140064E+04_dp, 3.0018788916766829E+04_dp, &
         8.7440843910547086E-03_dp, -3.1015062922638364E+04_dp, 3.9576778234732534E+04_dp, &
         -3.3617796506662242E+04_dp, -1.2664803729988839E+03_dp, 9.8208694230122404E+04_dp, &
         -2.7712338104205776E+05_dp, 5.0070250342457759E+05_dp, -6.4583598645926779E+05_dp, &
         5.7713900782644108E+05_dp, -3.1991818541374581E+05_dp, 8.2502449933280441E+04_dp, &
         1.0243629251738689E+05_dp, 3.8315196281821000E-02_dp, -1.3673145392947458E+05_dp, &
         1.7447480748440127E+05_dp, -1.2128767346456964E+05_dp, -7.4731473660592732E+04_dp, &
         5.4209346541303862E+05_dp, -1.3366186725741250E+06_dp, 2.2534054543823483E+06_dp, &
         -2.7624422501378348E+06_dp, 2.3684409505399591E+06_dp, -1.2681911269942287E+06_dp, &
         3.1767018404991477E+05_dp, 4.4811404216243100E+05_dp, 2.3079117700841423E-01_dp, &
         -8.3433208879112056E+05_dp, 1.0646234386996420E+06_dp, -4.3900763949985575E+05_dp, &
         -1.2299926310603216E+06_dp, 4.3873424472257951E+06_dp, -8.9070735680865049E+06_dp, &
         1.3145595299621349E+07_dp, -1.4320366639666997E+07_dp, 1.0965291282036081E+07_dp, &
         -5.2620189925113497E+06_dp, 1.1859559777996268E+06_dp, 3.2872930098168775E+06_dp, &
         2.7482124839381648E+00_dp, -1.0319720350104814E+07_dp, 1.3167559540232308E+07_dp, &
         5.3141307787938649E+06_dp, -4.2806172437992170E+07_dp, 7.7700737994820118E+07_dp, &
         -7.6880123309893116E+07_dp, 1.7343391064140473E+07_dp, 7.7860350009209916E+07_dp, &
         -1.3364460712376091E+08_dp, 1.0093529309506236E+08_dp, -3.1433866759847842E+07_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.8104907156518354E-01_dp, &
         -6.1817005353024562E-02_dp, -1.5365071642354823E-02_dp, 9.1489217818269322E-03_dp, &
         4.9967787273344161E-03_dp, -3.5367914652488903E-03_dp, -2.2075007488833458E-03_dp, &
         1.2029276142200300E-03_dp, 2.8328851353903797E-03_dp, -3.4438772536686739E-03_dp, &
         1.4299272644572190E-03_dp, -1.0931267316481758E-04_dp, -5.5894388646958038E-05_dp, &
         5.6894444769190411E-01_dp, -2.1274164519782207E-01_dp, -4.2890815190342713E-02_dp, &
         3.5035480705044961E-02_dp, 1.3874240699660001E-02_dp, -1.4090755641772217E-02_dp, &
         -6.0016188181493280E-03_dp, 5.1060704779823706E-03_dp, 9.3683628268561563E-03_dp, &
         -1.3681594752909918E-02_dp, 7.1674311571180219E-03_dp, -1.4730236769969446E-03_dp, &
         1.7117880701502458E-05_dp, 1.0389637522125577E+00_dp, -4.5428176175934221E-01_dp, &
         -5.6266159555088222E-02_dp, 8.5555486145475090E-02_dp, 1.7149370136037647E-02_dp, &
         -3.5874852805917547E-02_dp, -6.6722317760714526E-03_dp, 1.3972200727155170E-02_dp, &
         1.7715212621408963E-02_dp, -3.4443652948503337E-02_dp, 2.2673023691549964E-02_dp, &
         -6.9481802096620863E-03_dp, 7.9021807437504202E-04_dp, 1.6623569960343112E+00_dp, &
         -8.7799922392501561E-01_dp, -2.8938611876543164E-02_dp, 1.8305597014563521E-01_dp, &
         2.6823924227795206E-03_dp, -7.8638677468377882E-02_dp, 2.5052055477927018E-03_dp, &
         3.2539080971677445E-02_dp, 2.5797503204750667E-02_dp, -7.3577309739427957E-02_dp, &
         5.7896954096048170E-02_dp, -2.1479855679151314E-02_dp, 3.2218345849532076E-03_dp, &
         2.5381059435843234E+00_dp, -1.6305859324151268E+00_dp, 9.9940175243888832E-02_dp, &
         3.5719858476252214E-01_dp, -5.7992644536333599E-02_dp, -1.5378842508458723E-01_dp, &
         3.6992487976869874E-02_dp, 6.6983136752400169E-02_dp, 2.4029970044661862E-02_dp, &
         -1.3686421359551432E-01_dp, 1.2628872883212300E-01_dp, -5.3004210025772776E-02_dp, &
         9.0124008618107931E-03_dp, 3.8112298597695000E+00_dp, -2.9567738386099500E+00_dp, &
         4.6159700647669799E-01_dp, 6.4257759684283600E-01_dp, -2.2510024931676006E-01_dp, &
         -2.6991449769951259E-01_dp, 1.2839677612061992E-01_dp, 1.2324005868213329E-01_dp, &
         -1.3633470926998539E-02_dp, -2.1853429296999061E-01_dp, 2.4123664332249928E-01_dp, &
         -1.1237127932737824E-01_dp, 2.0806220268101396E-02_dp, 5.7008390024972444E+00_dp, &
         -5.2719032477064589E+00_dp, 1.3286136063792067E+00_dp, 1.0639787628820851E+00_dp, &
         -6.1724692798358538E-01_dp, -4.1836007817403786E-01_dp, 3.3558355186522965E-01_dp, &
         2.0049791059094066E-01_dp, -1.4522348760099926E-01_dp, -2.7848113288997439E-01_dp, &
         4.0294851159988648E-01_dp, -2.0991973963891600E-01_dp, 4.1918373279036081E-02_dp, &
         8.5444952115836337E+00_dp, -9.2870882366649212E+00_dp, 3.2542494456110536E+00_dp, &
         1.5891401650173187E+00_dp, -1.4538307368581864E+00_dp, -5.3611578876987864E-01_dp, &
         7.5799867206976324E-01_dp, 2.7324277500873972E-01_dp, -4.8297349611741452E-01_dp, &
         -1.9621208011655733E-01_dp, 5.6521268371561728E-01_dp, -3.4379015311921068E-01_dp, &
         7.4666191775106056E-02_dp, 1.2869608889008939E+01_dp, -1.6227381735998069E+01_dp, &
         7.3457273329354766E+00_dp, 2.0007984970256341E+00_dp, -3.1135650279596425E+00_dp, &
         -4.2035573694616968E-01_dp, 1.5423312586717872E+00_dp, 2.4895859945054985E-01_dp, &
         -1.2143416136828173E+00_dp, 3.0888829641492249E-01_dp, 5.6592145994268328E-01_dp, &
         -4.7133862783050268E-01_dp, 1.1553402263192626E-01_dp, 1.9510047195341826E+01_dp, &
         -2.8215377333929258E+01_dp, 1.5800988109124468E+01_dp, 1.5753381956046189E+00_dp, &
         -6.1888970404653927E+00_dp, 4.5340289216053564E-01_dp, 2.8529274310442476E+00_dp, &
         -1.1985631183237680E-01_dp, -2.5919936322336810E+00_dp, 1.7904459124895560E+00_dp, &
         3.3185572647898222E-03_dp, -4.5120198605812339E-01_dp, 1.4430781237161322E-01_dp, &
         2.9799686613683338E+01_dp, -4.8951476312595261E+01_dp, 3.2965439950410698E+01_dp, &
         -1.6925499201780276E+00_dp, -1.1444787017566560E+01_dp, 3.2893773175079954E+00_dp, &
         4.7232798353674390E+00_dp, -1.3946455243162874E+00_dp, -4.8138784821254603E+00_dp, &
         5.1806730186148213E+00_dp, -1.9484855421973464E+00_dp, 5.7565415788330868E-02_dp, &
         1.0328584787941356E-01_dp, 4.5900987586785305E+01_dp, -8.4931873768978448E+01_dp, &
         6.7417416166242447E+01_dp, -1.2917213048507392E+01_dp, -1.9369123838067868E+01_dp, &
         1.0603449957874252E+01_dp, 6.5793074414015802E+00_dp, -4.6781732888897354E+00_dp, &
         -7.6130671417336284E+00_dp, 1.1733674325796130E+01_dp, -6.7380063796773628E+00_dp, &
         1.7431827284203307E+00_dp, -1.3742674245542177E-01_dp, 7.1384655908206980E+01_dp, &
         -1.4766110646876876E+02_dp, 1.3608901365331167E+02_dp, -4.4554212981906190E+01_dp, &
         -2.8355600002395651E+01_dp, 2.7091425069539174E+01_dp, 5.9567085220792704E+00_dp, &
         -1.1750053684710588E+01_dp, -9.1674840428949729E+00_dp, 2.2293423124322853E+01_dp, &
         -1.6371253102977199E+01_dp, 5.7669097930761808E+00_dp, -8.2353886063174842E-01_dp, &
         1.1233183308389430E+02_dp, -2.5780369487207054E+02_dp, 2.7242671117115714E+02_dp, &
         -1.2599753109481574E+02_dp, -2.8803649715248678E+01_dp, 5.9808941846251201E+01_dp, &
         -4.4333363973394899E+00_dp, -2.4294323397425206E+01_dp, -3.7425366988330224E+00_dp, &
         3.4394335271359864E+01_dp, -3.1907344026729994E+01_dp, 1.3411678831415276E+01_dp, &
         -2.2957107908605567E+00_dp, 1.7976027905757553E+02_dp, -4.5369809560786518E+02_dp, &
         5.4292666408037110E+02_dp, -3.2469460257856400E+02_dp, 1.3642493310471760E+01_dp, &
         1.1322144991645266E+02_dp, -4.1415851597078642E+01_dp, -3.9344925288346268E+01_dp, &
         2.0505996019668014E+01_dp, 3.5640047716368947E+01_dp, -4.8066784685431486E+01_dp, &
         2.3931298245668948E+01_dp, -4.6302204728291798E+00_dp, 2.9648400099998616E+02_dp, &
         -8.1318193308728723E+02_dp, 1.0852146967185377E+03_dp, -7.9253616514248370E+02_dp, &
         1.9916533921620839E+02_dp, 1.6646952551534545E+02_dp, -1.3078063936770988E+02_dp, &
         -3.7561085470724848E+01_dp, 7.2748490941620716E+01_dp, 3.1770585593328282E+00_dp, &
         -4.8379526363826820E+01_dp, 3.1452688247501925E+01_dp, -6.9488664317550564E+00_dp, &
         5.2490296655177804E+02_dp, -1.5381034281001564E+03_dp, 2.2397394630871509E+03_dp, &
         -1.8980625603236201E+03_dp, 7.8793326721990013E+02_dp, 1.2072768890951619E+02_dp, &
         -2.8333018924346351E+02_dp, 1.5827025468773952E+01_dp, 1.4107764823365741E+02_dp, &
         -7.1508335180567997E+01_dp, -2.2202983634603370E+01_dp, 3.1298832279951391E+01_dp, &
         -8.4329929352037851E+00_dp, 1.1671482535247758E+03_dp, -3.5701355790217854E+03_dp, &
         5.5029528129983019E+03_dp, -5.1174808457482495E+03_dp, 2.6785762412568697E+03_dp, &
         -2.8427948185688939E+02_dp, -5.3279071516229101E+02_dp, 1.4157049520139412E+02_dp, &
         2.6643897464789472E+02_dp, -2.1826868704262932E+02_dp, 2.6638738596435580E+01_dp, &
         3.3936069509129055E+01_dp, -1.2188642274117770E+01_dp, 1.4652812785816598E+00_dp, &
         -5.1216317098725084E-01_dp, -1.2107111132320940E-01_dp, 7.8124136614970738E-02_dp, &
         3.9373299417261073E-02_dp, -3.0559387326754693E-02_dp, -1.7326558360561344E-02_dp, &
         1.0589249771069641E-02_dp, 2.3291147261995698E-02_dp, -2.9747108410639177E-02_dp, &
         1.3279817871973330E-02_dp, -1.5925954187249202E-03_dp, -3.3358537008160058E-04_dp, &
         1.6747706949071501E+00_dp, -7.3619993787034099E-01_dp, -9.2524849392815814E-02_dp, &
         1.4009317461191056E-01_dp, 2.8993455579017200E-02_dp, -5.8947753721227604E-02_dp, &
         -1.1544746505501786E-02_dp, 2.2888806638087296E-02_dp, 2.9752756422560003E-02_dp, &
         -5.6904268755167908E-02_dp, 3.7075509291121668E-02_dp, -1.1205940796047410E-02_dp, &
         1.2414097457956487E-03_dp, 2.1320569710835793E+00_dp, -1.2554495472557166E+00_dp, &
         2.0663593584744656E-03_dp, 2.7880698885400201E-01_dp, -1.0341136500847944E-02_dp, &
         -1.2182591843005953E-01_dp, 1.0375885305318683E-02_dp, 5.1177746698586865E-02_dp, &
         3.5915727679690139E-02_dp, -1.1457196488446013E-01_dp, 9.3552441080291246E-02_dp, &
         -3.5818521010418942E-02_dp, 5.5618136531600925E-03_dp, 2.9198954720291050E+00_dp, &
         -2.2293473658819969E+00_dp, 2.5486825622381931E-01_dp, 5.2106454908534805E-01_dp, &
         -1.2338568730458786E-01_dp, -2.2813706694695601E-01_dp, 7.2867460607848183E-02_dp, &
         1.0156141288573106E-01_dp, 2.2927036394505256E-02_dp, -2.0292387455972360E-01_dp, &
         1.9739334607294629E-01_dp, -8.5711263409840824E-02_dp, 1.5008093390084035E-02_dp, &
         4.1809650084217482E+00_dp, -3.9498692026171232E+00_dp, 8.5920971095646503E-01_dp, &
         8.9595233084619708E-01_dp, -4.0043826925842790E-01_dp, -3.7963806599838434E-01_dp, &
         2.2335390257611457E-01_dp, 1.7832326489249364E-01_dp, -5.3406140763388878E-02_dp, &
         -3.0094516220647582E-01_dp, 3.5944195478704649E-01_dp, -1.7390038059962756E-01_dp, &
         3.3089261580483392E-02_dp, 6.1472397966948646E+00_dp, -6.9285464394150278E+00_dp, &
         2.2070567132326251E+00_dp, 1.4063446953296506E+00_dp, -1.0126899665970956E+00_dp, &
         -5.4584355985905086E-01_dp, 5.4616991585738983E-01_dp, 2.7226973838426027E-01_dp, &
         -2.8505286023998866E-01_dp, -3.2863288737826130E-01_dp, 5.6233816955748084E-01_dp, &
         -3.0862618834553968E-01_dp, 6.3566187976697816E-02_dp, 9.1900790489272914E+00_dp, &
         -1.2052670995254264E+01_dp, 5.0808380779589832E+00_dp, 1.9509309699776836E+00_dp, &
         -2.2667200167260235E+00_dp, -6.0497438189785557E-01_dp, 1.1769408070141436E+00_dp, &
         3.3139818277041716E-01_dp, -8.4084448230590603E-01_dp, -7.5893210001079320E-02_dp, &
         7.0499033302583869E-01_dp, -4.7230545858400730E-01_dp, 1.0717094028717249E-01_dp, &
         1.3903288184276960E+01_dp, -2.0861113079586804E+01_dp, 1.1028873449856791E+01_dp, &
         2.1247273818735861E+00_dp, -4.6741212415573994E+00_dp, -2.1321596221060898E-01_dp, &
         2.3014636907442734E+00_dp, 1.9447225381512184E-01_dp, -1.9861181112351922E+00_dp, &
         9.1330548244123566E-01_dp, 4.9319113247404922E-01_dp, -5.7394999803708169E-01_dp, &
         1.5315431429978804E-01_dp, 2.1241141277279649E+01_dp, -3.6033094867817468E+01_dp, &
         2.3102278552533658E+01_dp, 7.3272373739498387E-01_dp, -8.9990510035402540E+00_dp, &
         1.4661536441210716E+00_dp, 4.0908947410985483E+00_dp, -5.4327633403126974E-01_dp, &
         -4.0397081958320689E+00_dp, 3.4883517046190322E+00_dp, -7.4279189342756524E-01_dp, &
         -3.5883086711361650E-01_dp, 1.6165467197230585E-01_dp, 3.2750700659675516E+01_dp, &
         -6.2260617506502861E+01_dp, 4.7306599319039115E+01_dp, -5.3630667640010694E+00_dp, &
         -1.6135965596523331E+01_dp, 6.2850804710731145E+00_dp, 6.4437698302620037E+00_dp, &
         -2.7654299843452557E+00_dp, -7.1374063007838009E+00_dp, 8.9855168947369108E+00_dp, &
         -4.2936389770830674E+00_dp, 7.3708343499722073E-01_dp, 3.2368466859189433E-02_dp, &
         5.0971587949712941E+01_dp, -1.0781940838147709E+02_dp, 9.5470719828705825E+01_dp, &
         -2.3933268645557146E+01_dp, -2.6335385696434315E+01_dp, 1.8016780861843714E+01_dp, &
         8.2132156983356950E+00_dp, -8.1534450671776550E+00_dp, -1.0525394951164396E+01_dp, &
         1.9008340533796282E+01_dp, -1.2282071103729981E+01_dp, 3.7805441921538856E+00_dp, &
         -4.4383200779161114E-01_dp, 8.0151692226233365E+01_dp, -1.8743506352274528E+02_dp, &
         1.9092478332081694E+02_dp, -7.3582255682872955E+01_dp, -3.6322792299277531E+01_dp, &
         4.3581813355095242E+01_dp, 5.1785945753389013E+00_dp, -1.9329067501593595E+01_dp, &
         -1.0809469621978822E+01_dp, 3.4142519766130000E+01_dp, -2.7441723274242250E+01_dp, &
         1.0488104966880156E+01_dp, -1.6444555466331974E+00_dp, 1.2763045722861104E+02_dp, &
         -3.2766564765272250E+02_dp, 3.7973217630740623E+02_dp, -1.9781100978811261E+02_dp, &
         -3.0157621574998899E+01_dp, 9.3237621383503154E+01_dp, -1.4664555123007290E+01_dp, &
         -3.8610014643549675E+01_dp, 1.9561581150219733E+00_dp, 4.9178714877907694E+01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/-5.0683586321199940E+01_dp, &
         2.2685614985361322E+01_dp, -4.0911306259362732E+00_dp, 2.0692715444769155E+02_dp, &
         -5.7784708851700361E+02_dp, 7.5313473849594516E+02_dp, -4.9591356082585395E+02_dp, &
         4.6934746915158954E+01_dp, 1.7278593891248715E+02_dp, -7.9461883488471031E+01_dp, &
         -6.0640581635323905E+01_dp, 4.9709605032605261E+01_dp, 3.9785417005473363E+01_dp, &
         -7.0707070950857386E+01_dp, 3.8301537914504102E+01_dp, -7.8036753533627845E+00_dp, &
         3.4643294122869480E+02_dp, -1.0373289229669426E+03_dp, 1.4964894908260794E+03_dp, &
         -1.1883576064664164E+03_dp, 3.5395000506872071E+02_dp, 2.4438341252020561E+02_dp, &
         -2.3525196154593584E+02_dp, -4.7569852552505999E+01_dp, 1.5264487252042383E+02_dp, &
         -4.7249655989039340E+01_dp, -4.6426435042560392E+01_dp, 4.1614986793760508E+01_dp, &
         -1.0251258037382264E+01_dp, 6.2235104125450675E+02_dp, -1.9530000742095463E+03_dp, &
         3.0381848576161515E+03_dp, -2.7721007214385299E+03_dp, 1.2886472933015646E+03_dp, &
         1.1997495108070251E+02_dp, -4.8923220063164257E+02_dp, 8.9320317305675530E+01_dp, &
         2.5417223540252016E+02_dp, -2.1855818229530024E+02_dp, 5.3225449020928423E+01_dp, &
         1.4035211359660455E+01_dp, -7.4385089727473837E+00_dp, 1.3392562526989720E+03_dp, &
         -4.2661156874133503E+03_dp, 6.8704505045145133E+03_dp, -6.7486562942964165E+03_dp, &
         3.8557980607995291E+03_dp, -6.4041197681422841E+02_dp, -7.2028812629191964E+02_dp, &
         3.4296722416927707E+02_dp, 2.9920668043243813E+02_dp, -3.7866211734835207E+02_dp, &
         1.4679850489782487E+02_dp, -8.2562501236068311E+00_dp, -6.1258993537203876E+00_dp, &
         5.0065694781245784E+03_dp, -1.5735707066352348E+04_dp, 2.5115152335838251E+04_dp, &
         -2.4641268459826806E+04_dp, 1.4432198334563878E+04_dp, -3.1899549141849984E+03_dp, &
         -1.6897674776327710E+03_dp, 7.8118394579369010E+02_dp, 9.8213799629630739E+02_dp, &
         -1.0349003003258824E+03_dp, 2.8781829158913928E+02_dp, 5.6520242044190624E+01_dp, &
         -3.5018492673502557E+01_dp, 2.1672646531042281E-01_dp, -4.0280327089000083E-02_dp, &
         -1.4141841987473180E-02_dp, 4.7576924044397707E-03_dp, 4.2680527641714202E-03_dp, &
         -1.7121426116300659E-03_dp, -1.8727973098150161E-03_dp, 5.2727512023623743E-04_dp, &
         1.7921888843793570E-03_dp, -1.5480544238380691E-03_dp, 2.0854567230273012E-04_dp, &
         2.5776529992692170E-04_dp, -9.6383744843509111E-05_dp, 6.9533459997442570E-01_dp, &
         -1.4709564320905877E-01_dp, -4.6296062125088105E-02_dp, 2.0685418524643968E-02_dp, &
         1.4532396018071178E-02_dp, -7.8206235021459156E-03_dp, -6.4330680110342292E-03_dp, &
         2.5911543805967269E-03_dp, 6.8719471260085367E-03_dp, -7.0817523244719764E-03_dp, &
         1.9939578385323588E-03_dp, 4.5553991899127580E-04_dp, -2.7486672885741699E-04_dp, &
         1.3220348620246702E+00_dp, -3.4559558249226729E-01_dp, -8.9054370747977230E-02_dp, &
         5.9965845713972753E-02_dp, 2.9540216889318294E-02_dp, -2.3945083153648589E-02_dp, &
         -1.3172332990522255E-02_dp, 8.5295161742936637E-03_dp, 1.6992003498685241E-02_dp, &
         -2.1830534111603173E-02_dp, 9.4603169609586910E-03_dp, -9.0624437091804298E-04_dp, &
         -3.1639345757305219E-04_dp, 2.2408348572564285E+00_dp, -7.4493621182222391E-01_dp, &
         -1.4437935708506164E-01_dp, 1.5345216731463060E-01_dp, 4.9981931944418634E-02_dp, &
         -6.4068353634840702E-02_dp, -2.2064095582909744E-02_dp, 2.4136625260774338E-02_dp, &
         3.7163716242542298E-02_dp, -5.8961636814152951E-02_dp, 3.2540613635375810E-02_dp, &
         -7.3389754359831682E-03_dp, 2.7235853348761487E-04_dp, 3.6772058442244018E+00_dp, &
         -1.5479451960257449E+00_dp, -1.9841574174749130E-01_dp, 3.6154130846694421E-01_dp, &
         6.8141128228278253E-02_dp, -1.5626923127364656E-01_dp, -2.8071652682116390E-02_dp, &
         6.1659614693382665E-02_dp, 7.3720675498160068E-02_dp, -1.4480983948301773E-01_dp, &
         9.4136312025964453E-02_dp, -2.8018590746203921E-02_dp, 2.9795911304120759E-03_dp, &
         6.0051316400388179E+00_dp, -3.1459525542112772E+00_dp, -1.9087960860761674E-01_dp, &
         8.0233225074828274E-01_dp, 5.0436194677495369E-02_dp, -3.5562444186561837E-01_dp, &
         -1.0808342335055487E-02_dp, 1.4631431396534988E-01_dp, 1.3021643058978571E-01_dp, &
         -3.2912594390655925E-01_dp, 2.4429333403753731E-01_dp, -8.5093523653718778E-02_dp, &
         1.1740285611601968E-02_dp, 9.8661407780096138E+00_dp, -6.3079683609443906E+00_dp, &
         6.4668417524656907E-02_dp, 1.7019207709152095E+00_dp, -1.0864690051990379E-01_dp, &
         -7.6490539642302580E-01_dp, 9.2463045587798220E-02_dp, 3.2754329325028214E-01_dp, &
         1.9036075703407310E-01_dp, -6.9593654311936159E-01_dp, 5.8500589456574437E-01_dp, &
         -2.2833809379472947E-01_dp, 3.6085090534356240E-02_dp, 1.6387345232640328E+01_dp, &
         -1.2566581816514574E+01_dp, 1.0833488063785175E+00_dp, 3.4780856044371822E+00_dp, &
         -7.0298676205420429E-01_dp, -1.5603437365117936E+00_dp, 4.5358830961481994E-01_dp, &
         6.9487899892425453E-01_dp, 1.6306574054540915E-01_dp, -1.3565280066889112E+00_dp, &
         1.3044257170639009E+00_dp, -5.6107520030086444E-01_dp, 9.7339437137894266E-02_dp, &
         2.7590362528083165E+01_dp, -2.5020953292207636E+01_dp, 4.2199543975733604E+00_dp, &
         6.8574653755300341E+00_dp, -2.5008013151358859E+00_dp, -2.9959805517111318E+00_dp, &
         1.5109398081007834E+00_dp, 1.3866679059684508E+00_dp, -2.7320241923625971E-01_dp, &
         -2.3552778799554877E+00_dp, 2.6919927683133067E+00_dp, -1.2750517732227757E+00_dp, &
         2.3894973963551455E-01_dp, 4.7176080794132197E+01_dp, -5.0043699620567423E+01_dp, &
         1.2963098635058795E+01_dp, 1.2956187817919837E+01_dp, -7.4293132600153990E+00_dp, &
         -5.2715667610509955E+00_dp, 4.3201347820454323E+00_dp, 2.5318631031919860E+00_dp, &
         -2.0681119464958941E+00_dp, -3.2740269844269108E+00_dp, 5.0056055519591984E+00_dp, &
         -2.6579602424397017E+00_dp, 5.3739683583410036E-01_dp, 8.2067870773483293E+01_dp, &
         -1.0099618717254077E+02_dp, 3.6251613177931354E+01_dp, 2.2955720667473543E+01_dp, &
         -2.0162813094031463E+01_dp, -7.8515331287445758E+00_dp, 1.1267353610368847E+01_dp, &
         3.8948541826633067E+00_dp, -7.6972479076468243E+00_dp, -1.9606037339217177E+00_dp, &
         7.7570215809451124E+00_dp, -4.9132929567084096E+00_dp, 1.0893358659958350E+00_dp, &
         1.4552872970382410E+02_dp, -2.0653403368911415E+02_dp, 9.7133525385800795E+01_dp, &
         3.5870243090002134E+01_dp, -5.1612920970104334E+01_dp, -6.8814440565183297E+00_dp, &
         2.7253921030056905E+01_dp, 3.4513432377643651E+00_dp, -2.2786048871961697E+01_dp, &
         8.9085416689751842E+00_dp, 7.2036115050483476E+00_dp, -7.2654992116157242E+00_dp, &
         1.8811183601144852E+00_dp, 2.6365252354702568E+02_dp, -4.2974049946773454E+02_dp, &
         2.5594897338735430E+02_dp, 3.8712804569966629E+01_dp, -1.2572180308400272E+02_dp, &
         1.3481924026494765E+01_dp, 6.0461082538064041E+01_dp, -7.3930191264914962E+00_dp, &
         -5.7463183771426053E+01_dp, 4.9457275105138656E+01_dp, -1.0700147395230934E+01_dp, &
         -4.9418175760713634E+00_dp, 2.2591130172972433E+00_dp, 4.8938924956022072E+02_dp, &
         -9.1372186196753569E+02_dp, 6.7434387995902102E+02_dp, -3.2155447353298307E+01_dp, &
         -2.8807904077131656E+02_dp, 1.0787587156933184E+02_dp, 1.1604641801031272E+02_dp, &
         -5.7495945231578801E+01_dp, -1.1982979280830179E+02_dp, 1.6386009398044138E+02_dp, &
         -8.5958117255417733E+01_dp, 1.8875664993395461E+01_dp, -6.7677172652916551E-01_dp, &
         9.3447004994210636E+02_dp, -1.9946219423819334E+03_dp, 1.7982346755590097E+03_dp, &
         -4.3838272051099790E+02_dp, -5.8789690476235876E+02_dp, 4.4360653113106764E+02_dp, &
         1.4921155606077679E+02_dp, -2.2511093876438125E+02_dp, -1.6695730504683976E+02_dp, &
         4.0917420561017377E+02_dp, -3.0271931512428114E+02_dp, 1.0796446015104219E+02_dp, &
         -1.5703656639435989E+01_dp, 1.8498047274145415E+03_dp, -4.4984616584856849E+03_dp, &
         4.8992366073257654E+03_dp, -2.1977425810406944E+03_dp, -8.3345216169575804E+02_dp, &
         1.4355964803086322E+03_dp, -1.5812943023377832E+02_dp, -6.2781654249010364E+02_dp, &
         9.2380286057259525E+01_dp, 6.5527187346706739E+02_dp, -6.9983355536916895E+02_dp, &
         3.1751326162224495E+02_dp, -5.7861253194568050E+01_dp, 3.8898521960899338E+03_dp, &
         -1.0660258370556912E+04_dp, 1.3786469354695366E+04_dp, -9.1465237275417621E+03_dp, &
         9.3013441723421602E+02_dp, 3.4076602842055540E+03_dp, -1.9886757703666833E+03_dp, &
         -8.2892218311343163E+02_dp, 1.2569071836213366E+03_dp, 3.2509800881158377E+01_dp, &
         -7.6144423093941646E+02_dp, 4.9335766114991003E+02_dp, -1.0865110989329460E+02_dp, &
         9.8449165454890644E+03_dp, -2.9503930183220888E+04_dp, 4.3421888480353453E+04_dp, &
         -3.6978110090308204E+04_dp, 1.5168356908680847E+04_dp, 2.9788779106407983E+03_dp, &
         -6.3846589497297919E+03_dp, 9.3105734344115615E+02_dp, 2.6308310447041681E+03_dp, &
         -1.6146136611487138E+03_dp, -1.9731348014190954E+02_dp, 5.0643537932149997E+02_dp, &
         -1.4599209745780860E+02_dp, 1.7630990396200976E+00_dp, -3.3892327099449571E-01_dp, &
         -1.1576929041053773E-01_dp, 4.2074501755300822E-02_dp, 3.5318613391807509E-02_dp, &
         -1.5361815626986417E-02_dp, -1.5538770819816210E-02_dp, 4.8377403154006675E-03_dp, &
         1.5296287290468806E-02_dp, -1.3888302637667885E-02_dp, 2.4825409810328397E-03_dp, &
         1.8872230081976539E-03_dp, -7.6714693939084367E-04_dp, 2.1325733506956639E+00_dp, &
         -5.5729186875233505E-01_dp, -1.4615706136468917E-01_dp, 9.6026856885913170E-02_dp, &
         4.8904598893406155E-02_dp, -3.8100135476357294E-02_dp, -2.1911777538135460E-02_dp, &
         1.3454311097079341E-02_dp, 2.7637903087209887E-02_dp, -3.4616620258338109E-02_dp, &
         1.4428685212043010E-02_dp, -1.0336299096049299E-03_dp, -5.9505869733514696E-04_dp, &
         2.9744753512390276E+00_dp, -1.1003780471193960E+00_dp, -1.9785159455719278E-01_dp, &
         2.3780637253194109E-01_dp, 7.1963817509901273E-02_dp, -9.9748145826735912E-02_dp, &
         -3.2190554128659045E-02_dp, 3.7695550143958513E-02_dp, 5.6651162671219071E-02_dp, &
         -9.1715891141988914E-02_dp, 5.1468581747088488E-02_dp, -1.1995276658616574E-02_dp, &
         5.5439593134832155E-04_dp, 4.5219935606964325E+00_dp, -2.2236721397856853E+00_dp, &
         -2.3730761168654535E-01_dp, 5.4856278124972280E-01_dp, 8.7445847277008756E-02_dp, &
         -2.3912431815003291E-01_dp, -3.6191282630300727E-02_dp, 9.4910314828061157E-02_dp, &
         1.0914636445746341E-01_dp, -2.2219329966642776E-01_dp, 1.4753788499623779E-01_dp, &
         -4.5129424300298711E-02_dp, 5.0589183084530339E-03_dp, 7.2107757897887188E+00_dp, &
         -4.4495581221107656E+00_dp, -1.5949892284174227E-01_dp, 1.1916501833218205E+00_dp, &
         3.8159808588412052E-02_dp, -5.3381719311232834E-01_dp, 1.1987078987062662E-04_dp, &
         2.2134348100162349E-01_dp, 1.8661057319208169E-01_dp, -4.9690591490122710E-01_dp, &
         3.7695345860164003E-01_dp, -1.3414080875310089E-01_dp, 1.9025830154460060E-02_dp, &
         1.1826586706724369E+01_dp, -8.8233400141049501E+00_dp, 3.3211087191020516E-01_dp, &
         2.4798971011020616E+00_dp, -2.4321974910475175E-01_dp, -1.1276491372635893E+00_dp, &
         1.7580577072785533E-01_dp, 4.8773057223976912E-01_dp, 2.5762633845103161E-01_dp, &
         -1.0335790498546602E+00_dp, 8.8984162729180138E-01_dp, -3.5392500454239423E-01_dp, &
         5.7015366649628613E-02_dp, 1.9787246843226512E+01_dp, -1.7445132201772360E+01_dp, &
         2.0255129628444672E+00_dp, 4.9871197966879164E+00_dp, -1.2082606849268420E+00_dp, &
         -2.2629643535078277E+00_dp, 7.5536562784156702E-01_dp, 1.0207918987429019E+00_dp, &
         1.7192575597177209E-01_dp, -1.9796532433429683E+00_dp, 1.9593380309152029E+00_dp, &
         -8.5836023264597161E-01_dp, 1.5125817632192159E-01_dp, 3.3681131542986975E+01_dp, &
         -3.4585438137019239E+01_dp, 6.9509291872553263E+00_dp, 9.6987345107099117E+00_dp, &
         -4.0173741038954836E+00_dp, -4.2756064482503922E+00_dp, 2.4041920018559479E+00_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/2.0123022827718544E+00_dp, &
         -5.8064769221536761E-01_dp, -3.3561931133425400E+00_dp, 3.9944119179735149E+00_dp, &
         -1.9296809249289084E+00_dp, 3.6684710023030970E-01_dp, 5.8314392583364203E+01_dp, &
         -6.9071915064420764E+01_dp, 2.0273316780786242E+01_dp, 1.8091276263426014E+01_dp, &
         -1.1553296198966734E+01_dp, -7.3776447143848456E+00_dp, 6.7127519171282239E+00_dp, &
         3.6256900346252308E+00_dp, -3.4861066477531804E+00_dp, -4.4280915350121814E+00_dp, &
         7.3155197483498444E+00_dp, -3.9817222244987160E+00_dp, 8.1716033590013382E-01_dp, &
         1.0279442872714777E+02_dp, -1.3951572323665187E+02_dp, 5.5126922655550800E+01_dp, &
         3.1593961331540505E+01_dp, -3.0773981811737219E+01_dp, -1.0609400500428917E+01_dp, &
         1.7269150027597533E+01_dp, 5.4550950252540860E+00_dp, -1.2412007314711813E+01_dp, &
         -1.6944363104250255E+00_dp, 1.1010026315963469E+01_dp, -7.2647707908744747E+00_dp, &
         1.6412253503340308E+00_dp, 1.8478392611717035E+02_dp, -2.8606063621708472E+02_dp, &
         1.4520791993720482E+02_dp, 4.8248284045813477E+01_dp, -7.7922693756595990E+01_dp, &
         -7.9408226077719561E+00_dp, 4.1485197987820960E+01_dp, 4.3678621514692084E+00_dp, &
         -3.6233004314355206E+01_dp, 1.6591239288982003E+01_dp, 8.9301404752582219E+00_dp, &
         -1.0434258526682950E+01_dp, 2.7944372114336824E+00_dp, 3.3944758158812471E+02_dp, &
         -5.9758838258126650E+02_dp, 3.7841395922782914E+02_dp, 4.8417305572292761E+01_dp, &
         -1.8893041807173518E+02_dp, 2.4741002322332704E+01_dp, 9.2060942372558330E+01_dp, &
         -1.2828891284744461E+01_dp, -9.1422655242511411E+01_dp, 8.2905747183313039E+01_dp, &
         -2.1697703734039433E+01_dp, -5.6962675534492826E+00_dp, 3.1991345493662133E+00_dp, &
         6.3893928008600767E+02_dp, -1.2768305025665254E+03_dp, 9.8951899589348307E+02_dp, &
         -6.0443695658991189E+01_dp, -4.3440590534894858E+02_dp, 1.7014936207019227E+02_dp, &
         1.7915396372204998E+02_dp, -9.1142430248866432E+01_dp, -1.9369374933733687E+02_dp, &
         2.7062993560444278E+02_dp, -1.4648606822230460E+02_dp, 3.4401173794203238E+01_dp, &
         -1.8643660577461350E+00_dp, 1.2368631912166727E+03_dp, -2.8021015677999771E+03_dp, &
         2.6245851326980301E+03_dp, -6.5184739072448338E+02_dp, -9.0662683327423667E+02_dp, &
         6.8595985546862971E+02_dp, 2.4708292671840775E+02_dp, -3.5944943741598553E+02_dp, &
         -2.8802980788436065E+02_dp, 6.9079812496750924E+02_dp, -5.1324954557812202E+02_dp, &
         1.8452547451235932E+02_dp, -2.7127058028799510E+01_dp, 2.4780267312523883E+03_dp, &
         -6.3481274679156131E+03_dp, 7.1223078050965232E+03_dp, -3.1670682000868255E+03_dp, &
         -1.4343302633933163E+03_dp, 2.2735595623482527E+03_dp, -1.4893036405924593E+02_dp, &
         -1.0812056752629833E+03_dp, 7.5011532640876709E+01_dp, 1.2375394471103587E+03_dp, &
         -1.2835762825052959E+03_dp, 5.7825279537734082E+02_dp, -1.0513964454975101E+02_dp, &
         5.2240008736035516E+03_dp, -1.4989583733809592E+04_dp, 1.9904817992583929E+04_dp, &
         -1.3071959006342418E+04_dp, 4.1890770123089112E+02_dp, 6.0736481096366579E+03_dp, &
         -3.0780323603297416E+03_dp, -2.0133593320735015E+03_dp, 2.5522083062383117E+03_dp, &
         1.0175091050765498E+02_dp, -1.5715434454560022E+03_dp, 1.0200855278610385E+03_dp, &
         -2.2627481009306229E+02_dp, 1.2373847610812090E+04_dp, -3.8548285445828238E+04_dp, &
         5.8497379094576674E+04_dp, -5.0546288117095348E+04_dp, 1.9279080238692830E+04_dp, &
         8.0456785161874832E+03_dp, -1.2762585183952415E+04_dp, 2.1391906252619819E+03_dp, &
         6.0395885681160671E+03_dp, -5.3701230424802343E+03_dp, 1.5974805021564043E+03_dp, &
         1.1957935329310017E+02_dp, -1.2701661891975539E+02_dp, 4.5956408995458441E+04_dp, &
         -1.4543165723892135E+05_dp, 2.3046137165679206E+05_dp, -2.2184270858179423E+05_dp, &
         1.2378962647413493E+05_dp, -1.9979906545203474E+04_dp, -2.1878488592470068E+04_dp, &
         1.0209913111179272E+04_dp, 7.6876868768320492E+03_dp, -8.6795300997867835E+03_dp, &
         2.1149514187852687E+03_dp, 7.3353058897293010E+02_dp, -3.6026759783745257E+02_dp, &
         2.3973743190763880E-01_dp, -2.1972173943175737E-02_dp, -1.2933433013553911E-02_dp, &
         2.0263255241757100E-03_dp, 3.8558136859502359E-03_dp, -2.3496377608067925E-04_dp, &
         -2.4948639401902385E-03_dp, 2.7317726766600036E-04_dp, 2.1977795539286837E-03_dp, &
         -1.7085734424693412E-03_dp, 1.8807577698138860E-04_dp, 2.9444043588833577E-04_dp, &
         -1.0495679299232483E-04_dp, 7.8119548885908119E-01_dp, -8.3814148577581293E-02_dp, &
         -4.7508135810116980E-02_dp, 9.8019626915272322E-03_dp, 1.4868280045836485E-02_dp, &
         -1.8006665222561487E-03_dp, -9.7377554307787259E-03_dp, 1.4125595007739784E-03_dp, &
         9.0509059488131458E-03_dp, -7.7385054530521970E-03_dp, 1.5388691452156726E-03_dp, &
         8.5747629581810734E-04_dp, -3.6661944358525332E-04_dp, 1.5306841908146380E+00_dp, &
         -2.1063022274943391E-01_dp, -1.1241196620289133E-01_dp, 3.2360998322751140E-02_dp, &
         3.7665015340093901E-02_dp, -8.0772948523086009E-03_dp, -2.5159527309634066E-02_dp, &
         5.0554039417750034E-03_dp, 2.5273349130359068E-02_dp, -2.4353394331749184E-02_dp, &
         7.2955857496937882E-03_dp, 9.9758288171452984E-04_dp, -7.6511170564210666E-04_dp, &
         2.7079852668931923E+00_dp, -4.8920830337568294E-01_dp, -2.4324920879045647E-01_dp, &
         9.4157225722568336E-02_dp, 8.7017036286837601E-02_dp, -2.8177328136367519E-02_dp, &
         -5.9539436788281913E-02_dp, 1.5947059865743421E-02_dp, 6.4998406125438488E-02_dp, &
         -7.0010003351382771E-02_dp, 2.6870981588887691E-02_dp, -1.2287304809945091E-03_dp, &
         -1.2626866804900177E-03_dp, 4.6871272746842072E+00_dp, -1.0979373448364005E+00_dp, &
         -5.0439064698554803E-01_dp, 2.5329109255504134E-01_dp, 1.9065417291741427E-01_dp, &
         -8.5926530886646244E-02_dp, -1.3414673664499402E-01_dp, 4.6615086648998535E-02_dp, &
         1.5969458749606510E-01_dp, -1.9077663190094568E-01_dp, 8.6799804720832860E-02_dp, &
         -1.2802583451222354E-02_dp, -1.2752933836137432E-03_dp, 8.1448100654677873E+00_dp, &
         -2.4178207414003650E+00_dp, -1.0140790036261034E+00_dp, 6.4948091684454845E-01_dp, &
         4.0002662216023965E-01_dp, -2.4319032508794547E-01_dp, -2.9094633344678594E-01_dp, &
         1.3045357990505776E-01_dp, 3.8023332842632163E-01_dp, -5.0306877230005320E-01_dp, &
         2.6113769839458034E-01_dp, -5.6298924203416423E-02_dp, 1.8123949457853315E-03_dp, &
         1.4352898973567878E+01_dp, -5.2818184817482203E+00_dp, -1.9859061847519863E+00_dp, &
         1.6209923485665547E+00_dp, 8.0480146245262718E-01_dp, -6.5980134024898218E-01_dp, &
         -6.0983822335215754E-01_dp, 3.5728072539684819E-01_dp, 8.8464717874409349E-01_dp, &
         -1.3018445886435062E+00_dp, 7.5507980073580228E-01_dp, -2.0136154790359589E-01_dp, &
         1.7515223022048523E-02_dp, 2.5776640136003774E+01_dp, -1.1552325240946562E+01_dp, &
         -3.7871289474355541E+00_dp, 3.9950291812053065E+00_dp, 1.5381543034175054E+00_dp, &
         -1.7488517416194553E+00_dp, -1.2307547949552229E+00_dp, 9.7227430479047705E-01_dp, &
         2.0152195969554660E+00_dp, -3.3340448599009007E+00_dp, 2.1353902718112705E+00_dp, &
         -6.5792855059375088E-01_dp, 7.7757862605825775E-02_dp, 4.7340489049606639E+01_dp, &
         -2.5502622340548470E+01_dp, -6.9771561423823139E+00_dp, 9.8241770037760148E+00_dp, &
         2.7041503874198094E+00_dp, -4.5827496067732367E+00_dp, -2.3499942979626960E+00_dp, &
         2.6589465299988495E+00_dp, 4.4638845876459596E+00_dp, -8.4790434572470250E+00_dp, &
         5.9637608044446919E+00_dp, -2.0511810835588626E+00_dp, 2.8556951006602344E-01_dp, &
         8.9206434643405373E+01_dp, -5.7246392101459577E+01_dp, -1.2118388129984879E+01_dp, &
         2.4298560398108567E+01_dp, 3.9314821788160956E+00_dp, -1.1961915849435366E+01_dp, &
         -4.0288255301413383E+00_dp, 7.3737596068968578E+00_dp, 9.3863903576693257E+00_dp, &
         -2.1365482846607684E+01_dp, 1.6512440427129999E+01_dp, -6.2254495487803139E+00_dp, &
         9.6619197544078494E-01_dp, 1.7312884123481155E+02_dp, -1.3162884253179351E+02_dp, &
         -1.8435423873064877E+01_dp, 6.0837883751267668E+01_dp, 2.3531394283228084E+00_dp, &
         -3.1220104773374494E+01_dp, -5.0963636597563147E+00_dp, 2.0884887034584892E+01_dp, &
         1.7434242907928009E+01_dp, -5.2737432022531095E+01_dp, 4.5207653351126943E+01_dp, &
         -1.8537109806799002E+01_dp, 3.1268716041690032E+00_dp, 3.4769947397156164E+02_dp, &
         -3.1249298768889707E+02_dp, -1.7437461781643293E+01_dp, 1.5498716759905759E+02_dp, &
         -1.5841772368812352E+01_dp, -8.1377538168402268E+01_dp, 1.8724669219067327E+00_dp, &
         6.0675444135269096E+01_dp, 2.0816898730371967E+01_dp, -1.2351369514321331E+02_dp, &
         1.2071107009758312E+02_dp, -5.3951445844698021E+01_dp, 9.7892956954004831E+00_dp, &
         7.2703354843362490E+02_dp, -7.7313108411402254E+02_dp, 3.5023735034956978E+01_dp, &
         4.0297017473577085E+02_dp, -1.1042256365889763E+02_dp, -2.0941813845985146E+02_dp, &
         5.2087715690791505E+01_dp, 1.8039722228227356E+02_dp, -4.1352516140890167E+01_dp, &
         -2.4923882189703463E+02_dp, 3.0196465579696888E+02_dp, -1.5000260743356392E+02_dp, &
         2.9343587506438965E+01_dp, 1.5958129013599184E+03_dp, -2.0168957177610125E+03_dp, &
         3.4240220658253907E+02_dp, 1.0672373281250943E+03_dp, -5.3613234405799597E+02_dp, &
         -5.1084300344864096E+02_dp, 2.9925364586857182E+02_dp, 5.3763268220233249E+02_dp, &
         -5.0331545545250009E+02_dp, -2.5847420072564825E+02_dp, 6.2069474368006570E+02_dp, &
         -3.7002167042097642E+02_dp, 8.0101106126116946E+01_dp, 3.7194250317395058E+03_dp, &
         -5.6366551640950183E+03_dp, 1.8508257063062345E+03_dp, 2.8281524590337158E+03_dp, &
         -2.4073828332691851E+03_dp, -9.9894429778883796E+02_dp, 1.3896422496078037E+03_dp, &
         1.4795882623602579E+03_dp, -2.9286651253849855E+03_dp, 1.3827022620856706E+03_dp, &
         3.7841886473089392E+02_dp, -5.8826257543824522E+02_dp, 1.6416552679796825E+02_dp, &
         9.3643659386802919E+03_dp, -1.7268153319694040E+04_dp, 9.3340833529031097E+03_dp, &
         6.8721045784191338E+03_dp, -1.0801964919124486E+04_dp, 3.3862132734645331E+02_dp, &
         5.7555001034296838E+03_dp, 2.4166845618580087E+03_dp, -1.2988122107330906E+04_dp, &
         1.3288937622953321E+04_dp, -6.4646546543504810E+03_dp, 1.4032801485599530E+03_dp, &
         -6.0214798988559934E+01_dp, 2.6224691656659063E+04_dp, -6.0092717495861041E+04_dp, &
         5.0748946885645579E+04_dp, 6.4296047875928125E+03_dp, -4.7887862033455087E+04_dp, &
         2.7735324023937988E+04_dp, 1.4938668072319664E+04_dp, -1.4316737379758091E+04_dp, &
         -2.8451024275428434E+04_dp, 5.7244647432622391E+04_dp, -4.5386091585022587E+04_dp, &
         1.8311548691603668E+04_dp, -3.1224051036545457E+03_dp, 8.9159916001930978E+04_dp, &
         -2.5586332776355915E+05_dp, 3.2843439673832123E+05_dp, -1.7020742611488042E+05_dp, &
         -1.0846427488360634E+05_dp, 2.5316054359400918E+05_dp, -1.7089677062997792E+05_dp, &
         1.6823614758843752E+04_dp, 3.7175896672429459E+04_dp, 2.2014007110624079E+03_dp, &
         -3.3289475790468583E+04_dp, 2.2732719096251702E+04_dp, -5.1983853159329319E+03_dp, &
         1.9578084287802890E+00_dp, -1.8700206095269101E-01_dp, -1.0900328443936919E-01_dp, &
         1.8470854213375257E-02_dp, 3.2926833694674286E-02_dp, -2.5244548830637811E-03_dp, &
         -2.1371119257102094E-02_dp, 2.5373572035113992E-03_dp, 1.9096936556220606E-02_dp, &
         -1.5244081816584576E-02_dp, 2.0668203447746073E-03_dp, 2.3576566574104695E-03_dp, &
         -8.7491071381053489E-04_dp, 2.4679674747272893E+00_dp, -3.3737601133348555E-01_dp, &
         -1.8139040419117963E-01_dp, 5.0523868752123523E-02_dp, 6.0550371224777057E-02_dp, &
         -1.2172815962913154E-02_dp, -4.0308427219634048E-02_dp, 7.7287143244938123E-03_dp, &
         4.0019771312533585E-02_dp, -3.7854821497454338E-02_dp, 1.0769335283667015E-02_dp, &
         1.9476127553671263E-03_dp, -1.2801087884499606E-03_dp, 3.6694426304025538E+00_dp, &
         -7.3185799640057625E-01_dp, -3.6064251381359935E-01_dp, 1.4504547183962732E-01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c04 = (/1.3124621600522313E-01_dp, &
         -4.3574928379419176E-02_dp, -8.9899230702256708E-02_dp, 2.4290637640954200E-02_dp, &
         9.8569915866746516E-02_dp, -1.0636170473106374E-01_dp, 4.0918764154109298E-02_dp, &
         -1.9272580657675622E-03_dp, -1.9052149763963443E-03_dp, 5.9890800326505680E+00_dp, &
         -1.6089545544303487E+00_dp, -7.2735892893642817E-01_dp, 3.8519972111009043E-01_dp, &
         2.8134695163873363E-01_dp, -1.3173769525033738E-01_dp, -1.9866704055549164E-01_dp, &
         7.0436060618907739E-02_dp, 2.3886114608119013E-01_dp, -2.8708303282688091E-01_dp, &
         1.3162016111650496E-01_dp, -1.9942386857965708E-02_dp, -1.7671512955273788E-03_dp, &
         1.0277071420779189E+01_dp, -3.4998322063470488E+00_dp, -1.4378747336383482E+00_dp, &
         9.7407812913651848E-01_dp, 5.8100968550036769E-01_dp, -3.6798347669688403E-01_dp, &
         -4.2502364982377133E-01_dp, 1.9512289769610702E-01_dp, 5.6300271102111910E-01_dp, &
         -7.5027827357326204E-01_dp, 3.9235942713735511E-01_dp, -8.5921250073101446E-02_dp, &
         3.1413061827254385E-03_dp, 1.8195484597436025E+01_dp, -7.5799465416797727E+00_dp, &
         -2.7779407333989208E+00_dp, 2.4038088840560472E+00_dp, 1.1529366402453889E+00_dp, &
         -9.8759392678972879E-01_dp, -8.8098967842317599E-01_dp, 5.2980884220561242E-01_dp, &
         1.3004888148740070E+00_dp, -1.9285239537624044E+00_dp, 1.1258532118551621E+00_dp, &
         -3.0321009022168460E-01_dp, 2.7043746144518259E-02_dp, 3.3035404761027252E+01_dp, &
         -1.6491595898542446E+01_dp, -5.2375830412033784E+00_dp, 5.8765409929385362E+00_dp, &
         2.1768388716426843E+00_dp, -2.5969780543056746E+00_dp, -1.7626117368164964E+00_dp, &
         1.4322289011391323E+00_dp, 2.9516064099497408E+00_dp, -4.9190419937796754E+00_dp, &
         3.1669116909973813E+00_dp, -9.8168653418504981E-01_dp, 1.1715471750362567E-01_dp, &
         6.1469293824180966E+01_dp, -3.6319806298406590E+01_dp, -9.5514353824431790E+00_dp, &
         1.4372704901961265E+01_dp, 3.7807422307232739E+00_dp, -6.7679914605420413E+00_dp, &
         -3.3434548753675188E+00_dp, 3.8965635434521699E+00_dp, 6.5408575540338090E+00_dp, &
         -1.2492746836072200E+01_dp, 8.8166589002534259E+00_dp, -3.0419114467219512E+00_dp, &
         4.2503541134711348E-01_dp, 1.1743554310779896E+02_dp, -8.1509386397962075E+01_dp, &
         -1.6406506874595095E+01_dp, 3.5425877135044423E+01_dp, 5.4008009960033583E+00_dp, &
         -1.7605685937044811E+01_dp, -5.7037271668928087E+00_dp, 1.0759910584842750E+01_dp, &
         1.3847290180791390E+01_dp, -3.1533337618401688E+01_dp, 2.4393105499495121E+01_dp, &
         -9.2014396622544474E+00_dp, 1.4283532789355629E+00_dp, 2.3109385131857380E+02_dp, &
         -1.8761966730640114E+02_dp, -2.4520412603580802E+01_dp, 8.8516278234992583E+01_dp, &
         2.9328252339946874E+00_dp, -4.5889710850574829E+01_dp, -7.1950183014523059E+00_dp, &
         3.0366027067610979E+01_dp, 2.6306482065369778E+01_dp, -7.8357372730095506E+01_dp, &
         6.6967780172219548E+01_dp, -2.7398393744283439E+01_dp, 4.6109478575715883E+00_dp, &
         4.7043109624338825E+02_dp, -4.4606475685262001E+02_dp, -2.1735952206843098E+01_dp, &
         2.2527950930573971E+02_dp, -2.3012422758538580E+01_dp, -1.1981796817643516E+02_dp, &
         2.5291315036908233E+00_dp, 8.8003829705726233E+01_dp, 3.4937617704092141E+01_dp, &
         -1.8678178196032340E+02_dp, 1.8052624320492154E+02_dp, -8.0190538925395487E+01_dp, &
         1.4478678662148862E+01_dp, 9.9623504829206684E+02_dp, -1.1043779912072609E+03_dp, &
         5.3154531274160959E+01_dp, 5.8589592555521381E+02_dp, -1.5504952430522471E+02_dp, &
         -3.1084069476317563E+02_dp, 7.2155851070132712E+01_dp, 2.6196369615057944E+02_dp, &
         -3.7901934092893228E+01_dp, -3.9677065726243325E+02_dp, 4.6351722925927106E+02_dp, &
         -2.2696455653345569E+02_dp, 4.3990267926862309E+01_dp, 2.2110877219599788E+03_dp, &
         -2.8764768741892444E+03_dp, 4.7492320131449105E+02_dp, 1.5567264701241727E+03_dp, &
         -7.3938540087289493E+02_dp, -7.7880598566919457E+02_dp, 4.1299391414202665E+02_dp, &
         7.9154965016367521E+02_dp, -6.3312008083422745E+02_dp, -5.4669713260373055E+02_dp, &
         1.0339049127853618E+03_dp, -5.8970336298410371E+02_dp, 1.2498590069609844E+02_dp, &
         5.1953268832294525E+03_dp, -7.9884476380606466E+03_dp, 2.4793733613308304E+03_dp, &
         4.1824523715100868E+03_dp, -3.2616952824520654E+03_dp, -1.7015636062823082E+03_dp, &
         1.9191005411391400E+03_dp, 2.3134019608791955E+03_dp, -3.9142318595195161E+03_dp, &
         1.2048151664984891E+03_dp, 1.2558303204971614E+03_dp, -1.1624542002305350E+03_dp, &
         2.9431974485633702E+02_dp, 1.3110918396829140E+04_dp, -2.4102687398992093E+04_dp, &
         1.1995982410614766E+04_dp, 1.0841615122259704E+04_dp, -1.4393293492492468E+04_dp, &
         -1.5198734249430386E+03_dp, 8.2366314984664277E+03_dp, 5.3369330859444490E+03_dp, &
         -1.8989222784588852E+04_dp, 1.6516142437368548E+04_dp, -6.0985673599833599E+03_dp, &
         3.7054895141125257E+02_dp, 2.4857740696458779E+02_dp, 3.6333087513022270E+04_dp, &
         -8.1221783017847876E+04_dp, 6.1132709382630717E+04_dp, 2.0704857862231747E+04_dp, &
         -6.5197332745595239E+04_dp, 2.1778211045947377E+04_dp, 3.0647530299908398E+04_dp, &
         -4.3442782810474673E+03_dp, -6.8761517136658484E+04_dp, 9.8370328388803697E+04_dp, &
         -6.6087566887000561E+04_dp, 2.3191079659114679E+04_dp, -3.4339593811428126E+03_dp, &
         1.1539025686929584E+05_dp, -3.1920426799906261E+05_dp, 3.5835520091848128E+05_dp, &
         -7.7332737818564172E+04_dp, -2.7179397309597122E+05_dp, 2.9504892028661666E+05_dp, &
         -1.5850909591480361E+03_dp, -1.9513271077423554E+05_dp, 5.8739957732688774E+04_dp, &
         1.8218684831437774E+05_dp, -2.3253340265484373E+05_dp, 1.1891512847722569E+05_dp, &
         -2.3965471466063856E+04_dp, 5.0796510014340468E+05_dp, -1.6654594573956954E+06_dp, &
         2.6259920107028279E+06_dp, -2.2320118568638815E+06_dp, 4.1945039078323619E+05_dp, &
         1.5259082255208537E+06_dp, -2.3070026158079393E+06_dp, 1.8438958389471469E+06_dp, &
         -9.8021455412542529E+05_dp, 3.8852172824498301E+05_dp, -1.3583505516405738E+05_dp, &
         4.2733204548142872E+04_dp, -7.7320302758653297E+03_dp, 2.4714592967992341E-01_dp, &
         -7.8085959775806109E-03_dp, -6.7886161955341851E-03_dp, 2.3575304344995184E-03_dp, &
         -2.3517034282114465E-03_dp, 3.9158802999507076E-03_dp, -4.5461153332725527E-03_dp, &
         4.5805194409910457E-03_dp, -4.4524512104898387E-03_dp, 3.7555627758705215E-03_dp, &
         -2.3599409622613479E-03_dp, 9.4509400621173940E-04_dp, -1.7884563401109535E-04_dp, &
         8.0966640464626982E-01_dp, -3.0233867484354637E-02_dp, -2.6043673616899996E-02_dp, &
         9.5433966009009930E-03_dp, -9.0783414688469580E-03_dp, 1.5175856706915760E-02_dp, &
         -1.7697460473983996E-02_dp, 1.7774158182275648E-02_dp, -1.7223209104464284E-02_dp, &
         1.4524811645378713E-02_dp, -9.1332381344281059E-03_dp, 3.6583813732935489E-03_dp, &
         -6.9204130195118878E-04_dp, 1.6030458569493182E+00_dp, -7.7712739846721682E-02_dp, &
         -6.6010044423433833E-02_dp, 2.6127869801634368E-02_dp, -2.3253711186275654E-02_dp, &
         3.9050839842992117E-02_dp, -4.5857572874862529E-02_dp, 4.5840961396754240E-02_dp, &
         -4.4204734538243924E-02_dp, 3.7266920335001827E-02_dp, -2.3458701939590761E-02_dp, &
         9.4004046670134924E-03_dp, -1.7773797446858689E-03_dp, 2.8781590421606613E+00_dp, &
         -1.8502266072689696E-01_dp, -1.5472615939338874E-01_dp, 6.6339708849260826E-02_dp, &
         -5.5249687600741691E-02_dp, 9.3041185647943017E-02_dp, -1.1013517526810855E-01_dp, &
         1.0955921820654839E-01_dp, -1.0506048446608686E-01_dp, 8.8531090619630326E-02_dp, &
         -5.5801054481577961E-02_dp, 2.2374634069922589E-02_dp, -4.2288141306295558E-03_dp, &
         5.0740496965044271E+00_dp, -4.2609069373861314E-01_dp, -3.5047238596729913E-01_dp, &
         1.6258723345136517E-01_dp, -1.2725913237484188E-01_dp, 2.1429433697276487E-01_dp, &
         -2.5589943506352225E-01_dp, 2.5335777197170445E-01_dp, -2.4146172104155522E-01_dp, &
         2.0334399963063937E-01_dp, -1.2836270975711614E-01_dp, 5.1515644012971963E-02_dp, &
         -9.7343687583029415E-03_dp, 9.0087543486586252E+00_dp, -9.6436054026559159E-01_dp, &
         -7.7907974479593267E-01_dp, 3.9137647622410721E-01_dp, -2.8888988562914647E-01_dp, &
         4.8474201985199905E-01_dp, -5.8455659582923736E-01_dp, 5.7609137874519123E-01_dp, &
         -5.4522036473334778E-01_dp, 4.5874071108557374E-01_dp, -2.9010852464938575E-01_dp, &
         1.1657483507620653E-01_dp, -2.2028508267688093E-02_dp, 1.6269081610823477E+01_dp, &
         -2.1706703701271688E+00_dp, -1.7186386235096978E+00_dp, 9.3797983745884228E-01_dp, &
         -6.5457017839938392E-01_dp, 1.0896280805830605E+00_dp, -1.3287688787666494E+00_dp, &
         1.3036771339867574E+00_dp, -1.2238033857287740E+00_dp, 1.0283519414815741E+00_dp, &
         -6.5175578160501146E-01_dp, 2.6235561770753085E-01_dp, -4.9594225091627114E-02_dp, &
         3.0039520360403436E+01_dp, -4.9089049295402107E+00_dp, -3.7977674798271108E+00_dp, &
         2.2636667351052835E+00_dp, -1.4972289723671712E+00_dp, 2.4584044262615730E+00_dp, &
         -3.0368285262935673E+00_dp, 2.9668326974499704E+00_dp, -2.7581511636752905E+00_dp, &
         2.3131661843896207E+00_dp, -1.4699607563494359E+00_dp, 5.9316114483746940E-01_dp, &
         -1.1222595041176153E-01_dp, 5.6934882930425353E+01_dp, -1.1255095485895385E+01_dp, &
         -8.4747322577007207E+00_dp, 5.5559279273831361E+00_dp, -3.4946282450544621E+00_dp, &
         5.6164991647309872E+00_dp, -7.0421238497195029E+00_dp, 6.8536265224353468E+00_dp, &
         -6.2970979175561093E+00_dp, 5.2656213569535302E+00_dp, -3.3569884616944177E+00_dp, &
         1.3592370978249693E+00_dp, -2.5758842009927652E-01_dp, 1.1122973850857802E+02_dp, &
         -2.6389905796723561E+01_dp, -1.9237652263540834E+01_dp, 1.3999317461544457E+01_dp, &
         -8.4157473265769305E+00_dp, 1.3101786091490547E+01_dp, -1.6712238563942623E+01_dp, &
         1.6216483953281678E+01_dp, -1.4686105782654973E+01_dp, 1.2224250268447630E+01_dp, &
         -7.8233760607390384E+00_dp, 3.1827221380884012E+00_dp, -6.0484510597143726E-01_dp, &
         2.2511331110403108E+02_dp, -6.3849510824888611E+01_dp, -4.4736822972493066E+01_dp, &
         3.6570941293529650E+01_dp, -2.1172368843978809E+01_dp, 3.1477570716288106E+01_dp, &
         -4.0948178433469003E+01_dp, 3.9674042582809108E+01_dp, -3.5289493665272310E+01_dp, &
         2.9158345876804262E+01_dp, -1.8743398238805774E+01_dp, 7.6757245316723974E+00_dp, &
         -1.4653860927918039E+00_dp, 4.7504666597161093E+02_dp, -1.6105638525337199E+02_dp, &
         -1.0734835568336860E+02_dp, 1.0017665095009950E+02_dp, -5.6502935380445450E+01_dp, &
         7.8671378853738148E+01_dp, -1.0460354615827933E+02_dp, 1.0145728089695098E+02_dp, &
         -8.8217580108401904E+01_dp, 7.2003186291951550E+01_dp, -4.6494111524940820E+01_dp, &
         1.9214944281734663E+01_dp, -3.6952935742615893E+00_dp, 1.0544524739814613E+03_dp, &
         -4.2902221397650425E+02_dp, -2.6782200683344763E+02_dp, 2.9185470703014238E+02_dp, &
         -1.6321979346586420E+02_dp, 2.0720494992345783E+02_dp, -2.8184954114105011E+02_dp, &
         2.7486750102605271E+02_dp, -2.3215822832786790E+02_dp, 1.8548517561875852E+02_dp, &
         -1.2013476439901228E+02_dp, 5.0268802195171816E+01_dp, -9.7793374602533572E+00_dp, &
         2.4925376021791531E+03_dp, -1.2280929973471482E+03_dp, -7.0003357512153480E+02_dp, &
         9.2189747375624108E+02_dp, -5.2511408044203222E+02_dp, 5.8621847685149055E+02_dp, &
         -8.1277799265203862E+02_dp, 8.0292827818442333E+02_dp, -6.5338665830524042E+02_dp, &
         5.0183079090547267E+02_dp, -3.2380554523618252E+02_dp, 1.3763419521849823E+02_dp, &
         -2.7255890559133967E+01_dp, 6.3918219717306083E+03_dp, -3.8763187937858406E+03_dp, &
         -1.9257015318780568E+03_dp, 3.2472036360073544E+03_dp, -1.9642544031850207E+03_dp, &
         1.8422466462758632E+03_dp, -2.5561629910519391E+03_dp, 2.5895289445906710E+03_dp, &
         -2.0116564542023402E+03_dp, 1.4283611462124322E+03_dp, -8.9489172870925711E+02_dp, &
         3.8577334167655448E+02_dp, -7.8417479541745152E+01_dp, 1.8326761523831301E+04_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c05 = (/-1.4073380941871266E+04_dp, &
         -5.4884212128021491E+03_dp, 1.3338114839397269E+04_dp, -9.1704389994968824E+03_dp, &
         6.9234192947715492E+03_dp, -8.9848495511906458E+03_dp, 9.4904838722939803E+03_dp, &
         -7.0131996973248588E+03_dp, 4.2053129160131894E+03_dp, -2.2892321189981762E+03_dp, &
         9.5853233917332932E+02_dp, -2.0059595633199862E+02_dp, 6.2145484992675003E+04_dp, &
         -6.3683464018306629E+04_dp, -1.3478172242723740E+04_dp, 6.9122642519095578E+04_dp, &
         -6.1149663704367988E+04_dp, 3.8450079402543517E+04_dp, -3.6623205602609836E+04_dp, &
         3.9290615378684073E+04_dp, -2.8219652444004834E+04_dp, 1.1412807988657049E+04_dp, &
         -1.7224522927570106E+03_dp, -3.9040817007118409E+02_dp, 1.4646631734514426E+02_dp, &
         2.8572736221717932E+05_dp, -4.3463112005624449E+05_dp, 8.0601652340817658E+04_dp, &
         5.0202883058267244E+05_dp, -7.9735137443422247E+05_dp, 6.5856450151112222E+05_dp, &
         -3.1469388927102822E+05_dp, 4.5380852232874913E+03_dp, 1.9064981445481826E+05_dp, &
         -2.6045344183859305E+05_dp, 2.0964678182384724E+05_dp, -9.9723937310119960E+04_dp, &
         2.1478645199727333E+04_dp, 2.0209842869144579E+00_dp, -6.6718986697254859E-02_dp, &
         -5.7864043748775733E-02_dp, 2.0385124539857375E-02_dp, -2.0076648172407022E-02_dp, &
         3.3467236798597394E-02_dp, -3.8898880867761311E-02_dp, 3.9159779450315686E-02_dp, &
         -3.8033899789202824E-02_dp, 3.2079462322166769E-02_dp, -2.0161695424411413E-02_dp, &
         8.0746211326208697E-03_dp, -1.5278524818652483E-03_dp, 2.5837103332261946E+00_dp, &
         -1.2412305755869253E-01_dp, -1.0562856977936072E-01_dp, 4.1401929220466462E-02_dp, &
         -3.7137069265504780E-02_dp, 6.2373047952612670E-02_dp, -7.3167139494141137E-02_dp, &
         7.3181444229383899E-02_dp, -7.0620695217909493E-02_dp, 5.9541226705652497E-02_dp, &
         -3.7473058713811576E-02_dp, 1.5014631941726080E-02_dp, -2.8389529827758256E-03_dp, &
         3.9243795810678530E+00_dp, -2.7752925964967601E-01_dp, -2.3174959372548332E-01_dp, &
         1.0011305293692697E-01_dp, -8.2767548665314883E-02_dp, 1.3960986500361180E-01_dp, &
         -1.6533811213529379E-01_dp, 1.6437972410323695E-01_dp, -1.5757476076598648E-01_dp, &
         1.3278585216426170E-01_dp, -8.3698567928797013E-02_dp, 3.3559074454691487E-02_dp, &
         -6.3419102382915853E-03_dp, 6.5574218511326112E+00_dp, -6.2720528041932788E-01_dp, &
         -5.1457373180315114E-01_dp, 2.4165147820461136E-01_dp, -1.8706272866170806E-01_dp, &
         3.1556412614357066E-01_dp, -3.7722243153075635E-01_dp, 3.7313797192010750E-01_dp, &
         -3.5535802476053696E-01_dp, 2.9926050182479369E-01_dp, -1.8893716237645972E-01_dp, &
         7.5825057518831993E-02_dp, -1.4325630615601024E-02_dp, 1.1531103735581528E+01_dp, &
         -1.4032618163838810E+00_dp, -1.1301837639020289E+00_dp, 5.7553236097795246E-01_dp, &
         -4.1986809658151530E-01_dp, 7.0558155272988798E-01_dp, -8.5201620306659231E-01_dp, &
         8.3882578576766720E-01_dp, -7.9314185990164077E-01_dp, 6.6733090862986277E-01_dp, &
         -4.2210637535148637E-01_dp, 1.6962027057736975E-01_dp, -3.2047172429002457E-02_dp, &
         2.0953856325861402E+01_dp, -3.1332175704610203E+00_dp, -2.4722201318657344E+00_dp, &
         1.3685689515151018E+00_dp, -9.4397783930672019E-01_dp, 1.5731408557496862E+00_dp, &
         -1.9214695312017522E+00_dp, 1.8831652333939095E+00_dp, -1.7658826005426529E+00_dp, &
         1.4838298143011921E+00_dp, -9.4067732071431376E-01_dp, 3.7868226521725007E-01_dp, &
         -7.1573121884394186E-02_dp, 3.9141027382473368E+01_dp, -7.0510935911389439E+00_dp, &
         -5.4349441219378498E+00_dp, 3.2859057843378880E+00_dp, -2.1492051697234911E+00_dp, &
         3.5316064256642075E+00_dp, -4.3703582897803956E+00_dp, 4.2649336499090476E+00_dp, &
         -3.9602705121144273E+00_dp, 3.3213171066858487E+00_dp, -2.1112848586815800E+00_dp, &
         8.5203244277433265E-01_dp, -1.6118013209548254E-01_dp, 7.5179666844985803E+01_dp, &
         -1.6129525112712493E+01_dp, -1.2099880210005864E+01_dp, 8.0393543887536687E+00_dp, &
         -5.0050419349925148E+00_dp, 8.0491356578510747E+00_dp, -1.0111118000567620E+01_dp, &
         9.8296164480187027E+00_dp, -9.0208256786749921E+00_dp, 7.5435906846169685E+00_dp, &
         -4.8109961052478889E+00_dp, 1.9481547611953693E+00_dp, -3.6913036892062134E-01_dp, &
         1.4889534986689205E+02_dp, -3.7795695700544215E+01_dp, -2.7459767911560213E+01_dp, &
         2.0211149927742266E+01_dp, -1.2040077089570843E+01_dp, 1.8764352859283580E+01_dp, &
         -2.3978110155581604E+01_dp, 2.3240884657405484E+01_dp, -2.1026891338026122E+01_dp, &
         1.7506149076342183E+01_dp, -1.1208140964030401E+01_dp, 4.5598936248731166E+00_dp, &
         -8.6634263812127699E-01_dp, 3.0539686947459830E+02_dp, -9.1444333087771867E+01_dp, &
         -6.3921240045689750E+01_dp, 5.2663651852895761E+01_dp, -3.0245493927589340E+01_dp, &
         4.5084324511456899E+01_dp, -5.8738721586867534E+01_dp, 5.6842137146892654E+01_dp, &
         -5.0535367379859217E+01_dp, 4.1783876072414230E+01_dp, -2.6871281987162011E+01_dp, &
         1.1002744908733408E+01_dp, -2.0995744007601536E+00_dp, 6.5253194924819445E+02_dp, &
         -2.3049888975569243E+02_dp, -1.5360350190567564E+02_dp, 1.4363449886237967E+02_dp, &
         -8.0388980829080381E+01_dp, 1.1261707068019670E+02_dp, -1.4991417030352827E+02_dp, &
         1.4517931236170054E+02_dp, -1.2629435322089957E+02_dp, 1.0326348811780197E+02_dp, &
         -6.6723417286080917E+01_dp, 2.7562125203340528E+01_dp, -5.2956364086681749E+00_dp, &
         1.4638759455072827E+03_dp, -6.1204832725053041E+02_dp, -3.8364710618768743E+02_dp, &
         4.1505445999875218E+02_dp, -2.2991842627563597E+02_dp, 2.9570852999172786E+02_dp, &
         -4.0267828686987724E+02_dp, 3.9173441007461383E+02_dp, -3.3158823108198732E+02_dp, &
         2.6608267952167319E+02_dp, -1.7259790833879237E+02_dp, 7.2150618982427929E+01_dp, &
         -1.4010375050526267E+01_dp, 3.4854439155843625E+03_dp, -1.7372008876781651E+03_dp, &
         -1.0030747630399879E+03_dp, 1.2911405329593324E+03_dp, -7.2402985904807974E+02_dp, &
         8.2901556502561914E+02_dp, -1.1526573479976369E+03_dp, 1.1334817444713410E+03_dp, &
         -9.2688170377254664E+02_dp, 7.1957041790704716E+02_dp, -4.6656972587699590E+02_dp, &
         1.9808433106431835E+02_dp, -3.9093212258748324E+01_dp, 8.9452333593947151E+03_dp, &
         -5.3804654916003874E+03_dp, -2.7615678829835888E+03_dp, 4.4211939369662241E+03_dp, &
         -2.5941258974348689E+03_dp, 2.5441305071386205E+03_dp, -3.5725792116391317E+03_dp, &
         3.5873032039055993E+03_dp, -2.8097510721061076E+03_dp, 2.0498805737947519E+03_dp, &
         -1.3081644460504003E+03_dp, 5.6519369679006809E+02_dp, -1.1429316751443122E+02_dp, &
         2.5338916917680017E+04_dp, -1.8766091152234465E+04_dp, -7.9810318917773557E+03_dp, &
         1.7229548905712414E+04_dp, -1.1116122770842045E+04_dp, 8.9620856529374523E+03_dp, &
         -1.2236743640136612E+04_dp, 1.2740974975183684E+04_dp, -9.4927044237527389E+03_dp, &
         6.1054857729098449E+03_dp, -3.6163779068096483E+03_dp, 1.5701523669074877E+03_dp, &
         -3.2867702174385209E+02_dp, 8.2423613014789240E+04_dp, -7.7738621638431970E+04_dp, &
         -2.2856223996885383E+04_dp, 8.0654743982346874E+04_dp, -6.1910908781176040E+04_dp, &
         4.0486392582576533E+04_dp, -4.7535009010394941E+04_dp, 5.2027077505429726E+04_dp, &
         -3.7183064561553911E+04_dp, 1.8156573347610247E+04_dp, -7.2562244437422933E+03_dp, &
         2.6129376328153694E+03_dp, -5.6069917781247318E+02_dp, 3.3241938028525532E+05_dp, &
         -4.2449719865456008E+05_dp, -2.7991543060281649E+04_dp, 4.9815416292546922E+05_dp, &
         -5.2713017647717579E+05_dp, 3.1848201456787082E+05_dp, -2.2093992761096239E+05_dp, &
         2.2247276428230319E+05_dp, -1.5785444477482370E+05_dp, 3.4768551468795849E+04_dp, &
         3.5161092722404210E+04_dp, -2.9940407867245645E+04_dp, 7.4882756263552019E+03_dp, &
         2.0447139661067696E+06_dp, -3.9283854089172743E+06_dp, 1.8502764473921761E+06_dp, &
         4.1951868191134324E+06_dp, -9.7529644361083508E+06_dp, 1.0318888468459859E+07_dp, &
         -5.2802722255622474E+06_dp, -2.1595666128842793E+06_dp, 7.7832591679149270E+06_dp, &
         -9.0678311244019158E+06_dp, 6.4686899047139660E+06_dp, -2.7819922020553104E+06_dp, &
         5.5456672891937534E+05_dp, 2.4890279420896633E-01_dp, 3.0066734866592875E-09_dp, &
         -1.0527327234770156E-02_dp, 1.3433637283418252E-02_dp, -1.7966097650019941E-02_dp, &
         1.6380268724846176E-02_dp, 2.5827143469974796E-03_dp, -5.0828042512973688E-02_dp, &
         1.2605197871020793E-01_dp, -1.9169993841817892E-01_dp, 1.9099346055836872E-01_dp, &
         -1.1453423015330329E-01_dp, 3.1344018718015468E-02_dp, 8.1648099483442071E-01_dp, &
         1.1681804630596778E-08_dp, -4.0907368999085904E-02_dp, 5.2200775677796407E-02_dp, &
         -6.9370421329701656E-02_dp, 6.2517319547751971E-02_dp, 1.2303423022201899E-02_dp, &
         -2.0110939199019731E-01_dp, 4.9426482807255301E-01_dp, -7.4900279581754514E-01_dp, &
         7.4475006092198848E-01_dp, -4.4603037474417395E-01_dp, 1.2195270515337567E-01_dp, &
         1.6206094508793791E+00_dp, 3.0183277900631536E-08_dp, -1.0571814695626443E-01_dp, &
         1.3490399708668319E-01_dp, -1.7755777128909234E-01_dp, 1.5716554526027085E-01_dp, &
         4.0575662816921231E-02_dp, -5.3363368225929542E-01_dp, 1.2944475547431196E+00_dp, &
         -1.9512977152300366E+00_dp, 1.9344516123948587E+00_dp, -1.1563005205015051E+00_dp, &
         3.1572545247324441E-01_dp, 2.9200997436150145E+00_dp, 7.2272211306585797E-08_dp, &
         -2.5319659711553028E-01_dp, 3.2309704851649657E-01_dp, -4.2074358141218299E-01_dp, &
         3.6486472342550103E-01_dp, 1.2013651265592906E-01_dp, -1.3142416201487426E+00_dp, &
         3.1444431701789064E+00_dp, -4.7133821392535511E+00_dp, 4.6576470048302268E+00_dp, &
         -2.7781948677092143E+00_dp, 7.5745674162454923E-01_dp, 5.1709355253819069E+00_dp, &
         1.6742699148040768E-07_dp, -5.8671080264769337E-01_dp, 7.4868485784133132E-01_dp, &
         -9.6403362067884502E-01_dp, 8.1750470049367896E-01_dp, 3.3369738745733207E-01_dp, &
         -3.1320371788601724E+00_dp, 7.3913387513511770E+00_dp, -1.1015588989961810E+01_dp, &
         1.0849123684781043E+01_dp, -6.4570948721058814E+00_dp, 1.7577598775352674E+00_dp, &
         9.2287628542971909E+00_dp, 3.8134106077365812E-07_dp, -1.3367097614968169E+00_dp, &
         1.7057363546008353E+00_dp, -2.1697290395664353E+00_dp, 1.7942979800129226E+00_dp, &
         8.9447410621817325E-01_dp, -7.3445546549349077E+00_dp, 1.7090148060084527E+01_dp, &
         -2.5316713230788089E+01_dp, 2.4846267033206001E+01_dp, -1.4753191637652103E+01_dp, &
         4.0094693816808542E+00_dp, 1.6766111681930639E+01_dp, 8.6435256783724784E-07_dp, &
         -3.0308060554179863E+00_dp, 3.8675214211329658E+00_dp, -4.8529498283031085E+00_dp, &
         3.8976870882182326E+00_dp, 2.3616553134743259E+00_dp, -1.7167702511583300E+01_dp, &
         3.9359851775548030E+01_dp, -5.7927748597033606E+01_dp, 5.6632670017808159E+01_dp, &
         -3.3540723665316413E+01_dp, 9.0986198218588719E+00_dp, 3.1168178536681012E+01_dp, &
         1.9700422170976357E-06_dp, -6.9105394933144373E+00_dp, 8.8183289442136772E+00_dp, &
         -1.0894028309812372E+01_dp, 8.4485589750840440E+00_dp, 6.2361739723108975E+00_dp, &
         -4.0446807666182103E+01_dp, 9.1267594151432164E+01_dp, -1.3336330401014530E+02_dp, &
         1.2982241038880991E+02_dp, -7.6664694454963950E+01_dp, 2.0753697286099698E+01_dp, &
         5.9534897429557844E+01_dp, 4.5573631978173512E-06_dp, -1.5993863131481948E+01_dp, &
         2.0409268076898947E+01_dp, -2.4759174458477368E+01_dp, 1.8389829887940540E+01_dp, &
         1.6674262282079525E+01_dp, -9.7004865265531805E+01_dp, 2.1513624544240966E+02_dp, &
         -3.1185418300656380E+02_dp, 3.0209632244299593E+02_dp, -1.7780645081741756E+02_dp, &
         4.8018190809835360E+01_dp, 1.1735957000719010E+02_dp, 1.0796734300526574E-05_dp, &
         -3.7912420090540877E+01_dp, 4.8378813528148328E+01_dp, -5.7435863659353501E+01_dp, &
         4.0377808851989506E+01_dp, 4.5658311276489918E+01_dp, -2.3911739347016501E+02_dp, &
         5.2031388028459696E+02_dp, -7.4739624739763394E+02_dp, 7.1994201689992678E+02_dp/)
      REAL(KIND=dp), DIMENSION(13, 36, 5), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04, c05, c06/), (/13, 36, 5/))

      INTEGER                                            :: irange

      IF (Rc <= 4.0000000000000000E+01_dp) Rc = 4.0000000000000000E+01_dp
      IF (Rc >= 2.1987628936767578E+06_dp) Rc = 2.1987628936767578E+06_dp
      IF (Rc >= 4.0000000000000000E+01_dp .AND. Rc <= 5.5746054090099528E+02_dp) THEN
         irange = 1
         L_b = 4.0000000000000000E+01_dp
         U_b = 5.5746054090099528E+02_dp
      END IF
      IF (Rc >= 5.5746054090099528E+02_dp .AND. Rc <= 7.4758824984654611E+03_dp) THEN
         irange = 2
         L_b = 5.5746054090099528E+02_dp
         U_b = 7.4758824984654611E+03_dp
      END IF
      IF (Rc >= 7.4758824984654611E+03_dp .AND. Rc <= 1.2462901325470702E+05_dp) THEN
         irange = 3
         L_b = 7.4758824984654611E+03_dp
         U_b = 1.2462901325470702E+05_dp
      END IF
      IF (Rc >= 1.2462901325470702E+05_dp .AND. Rc <= 7.3667319606433529E+05_dp) THEN
         irange = 4
         L_b = 1.2462901325470702E+05_dp
         U_b = 7.3667319606433529E+05_dp
      END IF
      IF (Rc >= 7.3667319606433529E+05_dp .AND. Rc <= 2.1987628936767578E+06_dp) THEN
         irange = 5
         L_b = 7.3667319606433529E+05_dp
         U_b = 2.1987628936767578E+06_dp
      END IF

      fit_coef(1:13, 1:36) = coefdata(1:13, 1:36, irange)

   END SUBROUTINE get_fit_coef_k18

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k19(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 38)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(164), PARAMETER :: c08 = (/1.0499223216920694E+01_dp, &
         4.4951728164305678E+00_dp, -4.0844623976889828E+01_dp, 9.6469089733765301E+01_dp, &
         -1.4457308237433509E+02_dp, 1.4358591897299843E+02_dp, -8.6337055270254567E+01_dp, &
         2.3772804326078404E+01_dp, 6.6359118115924133E+01_dp, -2.8591296818375736E-06_dp, &
         -1.7227311155696889E+01_dp, 2.2124688216575461E+01_dp, -2.7524752734790738E+01_dp, &
         2.2372786649592609E+01_dp, 1.2155184152551417E+01_dp, -9.4518163584077797E+01_dp, &
         2.1941577176458151E+02_dp, -3.2637649842310663E+02_dp, 3.2272807635497901E+02_dp, &
         -1.9348838342099501E+02_dp, 5.3166737579708254E+01_dp, 1.2849926566575485E+02_dp, &
         -6.5250582816014808E-06_dp, -3.9291091128744178E+01_dp, 5.0460729666257137E+01_dp, &
         -6.1689440682005809E+01_dp, 4.8223722297260849E+01_dp, 3.3157059718603954E+01_dp, &
         -2.2390910129266675E+02_dp, 5.1024271893998286E+02_dp, -7.5272887345661684E+02_dp, &
         7.4065729983418430E+02_dp, -4.4258825906333180E+02_dp, 1.2132797157588969E+02_dp, &
         2.5680058641412631E+02_dp, -1.5312146095985592E-05_dp, -9.2132612013296807E+01_dp, &
         1.1832391986938538E+02_dp, -1.4171780637542841E+02_dp, 1.0550987736400363E+02_dp, &
         9.2295690191958684E+01_dp, -5.4710074862993338E+02_dp, 1.2219574601299419E+03_dp, &
         -1.7860867470382068E+03_dp, 1.7476196856227098E+03_dp, -1.0403464625313584E+03_dp, &
         2.8441474799760618E+02_dp, 5.3222213783565780E+02_dp, -3.7234716544473444E-05_dp, &
         -2.2382764453118818E+02_dp, 2.8745677810311867E+02_dp, -3.3594463439207505E+02_dp, &
         2.3480947417458006E+02_dp, 2.6515448809283430E+02_dp, -1.3903458609430929E+03_dp, &
         3.0378333769430460E+03_dp, -4.3939128957463727E+03_dp, 4.2714720839997299E+03_dp, &
         -2.5314697435154421E+03_dp, 6.8983232947349563E+02_dp, 1.1512267982072294E+03_dp, &
         -9.4695299828165009E-05_dp, -5.6855359361921421E+02_dp, 7.3017979817497314E+02_dp, &
         -8.2805783277340424E+02_dp, 5.3123615799294919E+02_dp, 7.9604494681759957E+02_dp, &
         -3.7117311299625480E+03_dp, 7.9142883179915798E+03_dp, -1.1309056810379891E+04_dp, &
         1.0909850675069327E+04_dp, -6.4311777917450936E+03_dp, 1.7456613494233954E+03_dp, &
         2.6212437093759077E+03_dp, -2.5485209034150442E-04_dp, -1.5277300945628297E+03_dp, &
         1.9620247585698864E+03_dp, -2.1419889374981735E+03_dp, 1.2132646910388833E+03_dp, &
         2.5352427487043710E+03_dp, -1.0542926029232349E+04_dp, 2.1867497472080187E+04_dp, &
         -3.0801081479878208E+04_dp, 2.9437600466280288E+04_dp, -1.7238435953682736E+04_dp, &
         4.6562418705978034E+03_dp, 6.3573171344602770E+03_dp, -7.3750251193115297E-04_dp, &
         -4.4115279609245035E+03_dp, 5.6656029053998782E+03_dp, -5.8829581406359675E+03_dp, &
         2.7234456726279182E+03_dp, 8.7354783549759413E+03_dp, -3.2415576280063124E+04_dp, &
         6.5114222442079190E+04_dp, -9.0113853476169999E+04_dp, 8.5113563712506788E+04_dp, &
         -4.9417241881668429E+04_dp, 1.3262343922822860E+04_dp, 1.6709034360467489E+04_dp, &
         -2.3486683702591328E-03_dp, -1.4005755867923166E+04_dp, 1.7987158768947684E+04_dp, &
         -1.7415019499264930E+04_dp, 5.3886108848086124E+03_dp, 3.3486192491280985E+04_dp, &
         -1.1058426237334734E+05_dp, 2.1373071702116015E+05_dp, -2.8916673365143739E+05_dp, &
         2.6883819423513982E+05_dp, -1.5426221397195625E+05_dp, 4.1027854475280248E+04_dp, &
         4.8891105989388227E+04_dp, -8.5375882415901910E-03_dp, -5.0669203017368134E+04_dp, &
         6.5072639338247180E+04_dp, -5.6622801394688104E+04_dp, 3.0202589453576870E+03_dp, &
         1.4914012489017315E+05_dp, -4.3485803392159345E+05_dp, 8.0005087015935709E+05_dp, &
         -1.0489157105679847E+06_dp, 9.5284382085401053E+05_dp, -5.3704348597552557E+05_dp, &
         1.4082751009772695E+05_dp, 1.6683629894264581E+05_dp, -3.7790306832555944E-02_dp, &
         -2.2242916225572405E+05_dp, 2.8565589535384666E+05_dp, -2.0552031775314044E+05_dp, &
         -9.7947481898987331E+04_dp, 8.3244217354800832E+05_dp, -2.1022382927533309E+06_dp, &
         3.6045398713238072E+06_dp, -4.4935831417357307E+06_dp, 3.9211888536126167E+06_dp, &
         -2.1383095567867076E+06_dp, 5.4562133604672970E+05_dp, 7.2983594458675524E+05_dp, &
         -2.3366549317086566E-01_dp, -1.3516445238670940E+06_dp, 1.7358328797716717E+06_dp, &
         -7.6941638889201207E+05_dp, -1.8347737779277351E+06_dp, 6.8172234608326862E+06_dp, &
         -1.4087562537947292E+07_dp, 2.1161535706049107E+07_dp, -2.3534962968572598E+07_dp, &
         1.8463007196192324E+07_dp, -9.1061534880313724E+06_dp, 2.1147690424670577E+06_dp, &
         5.3539598700785628E+06_dp, -3.0211150365300039E+00_dp, -1.6646735851000162E+07_dp, &
         2.1377543596431356E+07_dp, 7.5558560062402077E+06_dp, -6.6595246625963897E+07_dp, &
         1.2300183044218574E+08_dp, -1.2663664724774376E+08_dp, 4.1031313200280927E+07_dp, &
         1.0283971430652012E+08_dp, -1.9310895904024291E+08_dp, 1.5058167673706385E+08_dp, &
         -4.7897275226966202E+07_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.4159553890012480E-01_dp, &
         -3.2212259532342614E-02_dp, -2.5211569112270739E-03_dp, 1.0259263161798598E-03_dp, &
         1.7009451604343799E-04_dp, -7.4233588848268952E-05_dp, -1.4091849072407304E-05_dp, &
         8.9792205912272354E-06_dp, -6.1068721080120308E-07_dp, 2.3673956901460579E-06_dp, &
         -1.4083255944590247E-06_dp, -2.7195755501531476E-06_dp, 2.8146140048823581E-06_dp, &
         4.3687738412731247E-01_dp, -1.0507959602935127E-01_dp, -6.2038156589891806E-03_dp, &
         3.4867260532288986E-03_dp, 4.0214827530637130E-04_dp, -2.5825399367695534E-04_dp, &
         -3.1252892528233610E-05_dp, 3.0947573806607806E-05_dp, -3.6098038277409914E-06_dp, &
         7.3259195158793407E-06_dp, -4.3056106890235391E-06_dp, -8.7173329450119566E-06_dp, &
         9.0229802552072765E-06_dp, 7.6969894835349451E-01_dp, -2.0480550688388621E-01_dp, &
         -5.1614525784059537E-03_dp, 7.1434595006984259E-03_dp, 2.5699998302359232E-04_dp, &
         -5.4217957031654817E-04_dp, -9.9497238896661655E-06_dp, 6.3976878971682059E-05_dp, &
         -1.2652209576345682E-05_dp, 1.3188831511557239E-05_dp, -7.3852575250832500E-06_dp, &
         -1.6514325039434995E-05_dp, 1.7056531547933975E-05_dp, 1.1694161663883196E+00_dp, &
         -3.5457725458566364E-01_dp, 6.0405099172997122E-03_dp, 1.2652283618377483E-02_dp, &
         -6.9573612869031239E-04_dp, -9.6570008981377035E-04_dp, 9.3185669048836430E-05_dp, &
         1.1170291259476769E-04_dp, -3.4059568676163194E-05_dp, 2.1411436640986164E-05_dp, &
         -1.0587098957813508E-05_dp, -2.7739876771009426E-05_dp, 2.8436369456750092E-05_dp, &
         1.6722923778351919E+00_dp, -5.8623250288584683E-01_dp, 3.6803481489419941E-02_dp, &
         2.0292829632816360E-02_dp, -3.1641052891675505E-03_dp, -1.5095879003062503E-03_dp, &
         3.4669256907157714E-04_dp, 1.7005900081597947E-04_dp, -7.7249793311931302E-05_dp, &
         3.5028575965622117E-05_dp, -1.3652892102255319E-05_dp, -4.4739744474324642E-05_dp, &
         4.5186940571270951E-05_dp, 2.3252215053670944E+00_dp, -9.4543684120515115E-01_dp, &
         1.0338931425768211E-01_dp, 2.9396745206708937E-02_dp, -8.2573760376551154E-03_dp, &
         -2.0274975592751354E-03_dp, 8.4994675565299837E-04_dp, 2.1942491207429306E-04_dp, &
         -1.5506894551206067E-04_dp, 5.9943198545642097E-05_dp, -1.6238288004320363E-05_dp, &
         -7.1013836471124102E-05_dp, 7.0143453024049232E-05_dp, 3.1906616395439409E+00_dp, &
         -1.4990569225058550E+00_dp, 2.3356101549045380E-01_dp, 3.7152742084143882E-02_dp, &
         -1.7592764802963623E-02_dp, -2.1213171272414589E-03_dp, 1.7296120802156852E-03_dp, &
         2.1165538052484334E-04_dp, -2.8264733608776007E-04_dp, 1.0684070365301925E-04_dp, &
         -1.8316063933529629E-05_dp, -1.1178382801896222E-04_dp, 1.0737911151933133E-04_dp, &
         4.3534382015127369E+00_dp, -2.3459506946039794E+00_dp, 4.7408783987239850E-01_dp, &
         3.6262330606119872E-02_dp, -3.3299942574138243E-02_dp, -9.3140059419598656E-04_dp, &
         3.1137054827187964E-03_dp, 5.1022668227572255E-05_dp, -4.7327734643169223E-04_dp, &
         1.9371495091259679E-04_dp, -2.1116509746927200E-05_dp, -1.7465725270621350E-04_dp, &
         1.6279100510407628E-04_dp, 5.9303714588445304E+00_dp, -3.6329281082192590E+00_dp, &
         9.0285285883188005E-01_dp, 1.0499413585049452E-02_dp, -5.7706685341736653E-02_dp, &
         3.1914219035582910E-03_dp, 5.0564950350439003E-03_dp, -4.3044903889887784E-04_dp, &
         -7.2872478490099949E-04_dp, 3.4876195691509751E-04_dp, -2.8964892295533430E-05_dp, &
         -2.7029244144607034E-04_dp, 2.4488000058885411E-04_dp, 8.0842352986683856E+00_dp, &
         -5.5785127765670346E+00_dp, 1.6482906476540169E+00_dp, -7.3501953025711345E-02_dp, &
         -9.2213493322770707E-02_dp, 1.3150366867854405E-02_dp, 7.3635168477600224E-03_dp, &
         -1.4982003516811509E-03_dp, -1.0196873457808851E-03_dp, 6.1292284623671504E-04_dp, &
         -5.2523833942698411E-05_dp, -4.1290691731029868E-04_dp, 3.6576881274268374E-04_dp, &
         1.1044653770852007E+01_dp, -8.5086264786359092E+00_dp, 2.9204021062989258E+00_dp, &
         -2.8111787188864740E-01_dp, -1.3429818303779237E-01_dp, 3.3638242999846295E-02_dp, &
         9.2262993367846335E-03_dp, -3.5217929571347388E-03_dp, -1.2493756075931993E-03_dp, &
         1.0399803214705392E-03_dp, -1.1423406576364230E-04_dp, -6.2036982403473006E-04_dp, &
         5.4255571555972932E-04_dp, 1.5140946973553808E+01_dp, -1.2911167463604777E+01_dp, &
         5.0599558722610176E+00_dp, -7.3604696739389575E-01_dp, -1.7049477007591834E-01_dp, &
         7.1395167777044383E-02_dp, 8.5304346040664409E-03_dp, -6.9043372440724740E-03_dp, &
         -1.1917442674394089E-03_dp, 1.6877417649437795E-03_dp, -2.5692597772479920E-04_dp, &
         -9.1320979479733819E-04_dp, 7.9917704737221692E-04_dp, 2.0857767729149380E+01_dp, &
         -1.9522972409838573E+01_dp, 8.6166217738442601E+00_dp, -1.6655977945022149E+00_dp, &
         -1.6130946336101359E-01_dp, 1.3407695333391192E-01_dp, 7.3462104420434426E-04_dp, &
         -1.1786497079900466E-02_dp, -4.1197943246183656E-04_dp, 2.5837162514269450E-03_dp, &
         -5.5402838065590862E-04_dp, -1.3110273196903820E-03_dp, 1.1688682117688078E-03_dp, &
         2.8939950737247685E+01_dp, -2.9480526610440588E+01_dp, 1.4476451649973361E+01_dp, &
         -3.4764436718654159E+00_dp, -1.1492003298758746E-02_dp, 2.2502856445645558E-01_dp, &
         -2.2365897846246961E-02_dp, -1.7268108576283563E-02_dp, 1.7427539085000711E-03_dp, &
         3.6430042980382660E-03_dp, -1.1065725030894821E-03_dp, -1.8275133932960331E-03_dp, &
         1.6976814016934469E-03_dp, 4.0618956502082611E+01_dp, -4.4622059577162375E+01_dp, &
         2.4090578510499956E+01_dp, -6.8810902494300450E+00_dp, 4.8297850963216071E-01_dp, &
         3.2924474886243160E-01_dp, -7.2369776779368955E-02_dp, -2.0078222605003318E-02_dp, &
         5.8293068973593737E-03_dp, 4.5736167580590990E-03_dp, -1.9885526061121971E-03_dp, &
         -2.4869733482698786E-03_dp, 2.4561874555545063E-03_dp, 5.8186666034640020E+01_dp, &
         -6.8227495131103126E+01_dp, 3.9981099201342317E+01_dp, -1.3134066085088833E+01_dp, &
         1.7219507581901394E+00_dp, 3.8681398325887517E-01_dp, -1.5929705043037320E-01_dp, &
         -1.3820689038096395E-02_dp, 1.1500886478842712E-02_dp, 4.9890084140602840E-03_dp, &
         -3.1430775711513877E-03_dp, -3.4228763531890033E-03_dp, 3.5888038606415973E-03_dp, &
         8.6758629460574809E+01_dp, -1.0730358374322792E+02_dp, 6.7254446024704066E+01_dp, &
         -2.4650559229093147E+01_dp, 4.4657471444515995E+00_dp, 2.5973390816722725E-01_dp, &
         -2.8281531420486583E-01_dp, 8.6395325668061733E-03_dp, 1.7190940344631079E-02_dp, &
         4.9683014527230921E-03_dp, -4.4797399994558479E-03_dp, -5.0527887500770325E-03_dp, &
         5.4485658862832098E-03_dp, 1.4144570865336470E+02_dp, -1.8203037827531273E+02_dp, &
         1.1996729501753330E+02_dp, -4.7597068467016200E+01_dp, 1.0433251046275085E+01_dp, &
         -3.0715161860465295E-01_dp, -4.4333987528632623E-01_dp, 5.2266426831911433E-02_dp, &
         2.2622601468742540E-02_dp, 5.6448052547070247E-03_dp, -6.6314050067125879E-03_dp, &
         -8.2807273014170093E-03_dp, 8.9883473255153619E-03_dp, 2.9783846082084170E+02_dp, &
         -3.9281745649836722E+02_dp, 2.6705245994554446E+02_dp, -1.1116159567569434E+02_dp, &
         2.7022970008284052E+01_dp, -2.0393625814260337E+00_dp, -7.8644885792012886E-01_dp, &
         1.4664747765129174E-01_dp, 3.6698696189955908E-02_dp, 1.0255959134649445E-02_dp, &
         -1.3496519985266115E-02_dp, -1.7057146077249077E-02_dp, 1.8839418118163746E-02_dp, &
         1.1407412355136231E+00_dp, -2.6322614975892222E-01_dp, -1.9304807485994203E-02_dp, &
         8.4818056715191947E-03_dp, 1.2921553214948484E-03_dp, -6.1798626362584322E-04_dp, &
         -1.0574734927292684E-04_dp, 7.4567513561016663E-05_dp, -6.0228382320527032E-06_dp, &
         1.9073396980006737E-05_dp, -1.1324448161278565E-05_dp, -2.2120940011110670E-05_dp, &
         2.2897251630575986E-05_dp, 1.2383058680166454E+00_dp, -3.3177779766973436E-01_dp, &
         -8.0046278756808614E-03_dp, 1.1734934753880745E-02_dp, 3.9208403634974115E-04_dp, &
         -8.9816173297452740E-04_dp, -1.3958158920641618E-05_dp, 1.0603060144415561E-04_dp, &
         -2.0721841907929155E-05_dp, 2.1014251840363297E-05_dp, -1.1928943188164544E-05_dp, &
         -2.6670935778900919E-05_dp, 2.7588164189909763E-05_dp, 1.4436406826714989E+00_dp, &
         -4.8123537796251287E-01_dp, 1.9845018196607334E-02_dp, 1.7971084322191626E-02_dp, &
         -1.8324606541230578E-03_dp, -1.4018145539285401E-03_dp, 2.1266040253605433E-04_dp, &
         1.6089571777858162E-04_dp, -5.5758950174576915E-05_dp, 2.6807942924299252E-05_dp, &
         -1.2671269591892968E-05_dp, -3.6769890090760815E-05_dp, 3.7679474165084254E-05_dp, &
         1.7780289941084260E+00_dp, -7.3782695706333956E-01_dp, 7.5852316169277137E-02_dp, &
         2.6321369018930191E-02_dp, -6.2530501547849882E-03_dp, -1.9778677729210049E-03_dp, &
         6.5769145784193284E-04_dp, 2.1870962072471271E-04_dp, -1.2174982968187595E-04_dp, &
         4.0911670934797237E-05_dp, -1.2765102734804405E-05_dp, -5.4569409911678445E-05_dp, &
         5.4688757320726091E-05_dp, 2.2759204975194458E+00_dp, -1.1453785133959720E+00_dp, &
         1.8079354115664725E-01_dp, 3.4432382914558510E-02_dp, -1.4225585470058071E-02_dp, &
         -2.2837089659632549E-03_dp, 1.4367339690524530E-03_dp, 2.3591365549879067E-04_dp, &
         -2.3239837931551176E-04_dp, 7.1833055128279899E-05_dp, -1.1507142004923511E-05_dp, &
         -8.3724655681505198E-05_dp, 8.1177843733305105E-05_dp, 2.9887064285877973E+00_dp, &
         -1.7721307937649995E+00_dp, 3.6991743513447239E-01_dp, 3.6672484709771344E-02_dp, &
         -2.7596569461930998E-02_dp, -1.6078064447974708E-03_dp, 2.6752315429307439E-03_dp, &
         1.2818773343005159E-04_dp, -4.0049270094921513E-04_dp, 1.3443710602048975E-04_dp, &
         -9.0985942281952549E-06_dp, -1.2990401189688037E-04_dp, 1.2115099317095869E-04_dp, &
         3.9907366836701370E+00_dp, -2.7216952914669283E+00_dp, 7.0185367490954009E-01_dp, &
         2.0790650087634404E-02_dp, -4.8523023243476252E-02_dp, 1.4087441403410760E-03_dp, &
         4.4519866778520399E-03_dp, -2.5248221642837002E-04_dp, -6.2960311984174694E-04_dp, &
         2.5254010380090493E-04_dp, -8.2259066803737854E-06_dp, -2.0128177724252732E-04_dp, &
         1.8061401811982056E-04_dp, 5.3884484703298714E+00_dp, -4.1497815192118628E+00_dp, &
         1.2730263450206805E+00_dp, -3.8210239808827151E-02_dp, -7.8752008993259112E-02_dp, &
         9.1787131489447105E-03_dp, 6.6677701366350964E-03_dp, -1.1427359524673077E-03_dp, &
         -8.9785110256968275E-04_dp, 4.6106575711624083E-04_dp, -1.6852240096092439E-05_dp, &
         -3.0877518141384864E-04_dp, 2.6828132709743447E-04_dp, 7.3341836390666773E+00_dp, &
         -6.2894029533210078E+00_dp, 2.2407098704133954E+00_dp, -1.8918988940658871E-01_dp, &
         -1.1761608329525802E-01_dp, 2.5701025480113431E-02_dp, 8.7678899455993949E-03_dp, &
         -2.8850805066351374E-03_dp, -1.1285234289110727E-03_dp, 8.0654726904078294E-04_dp, &
         -5.2692171698939168E-05_dp, -4.6577350369995772E-04_dp, 3.9642409742979582E-04_dp, &
         1.0047710850872544E+01_dp, -9.4890315198232518E+00_dp, 3.8596089366441526E+00_dp, &
         -5.2472117726711742E-01_dp, -1.5718795036158192E-01_dp, 5.7072740328761015E-02_dp, &
         9.1939643128644485E-03_dp, -5.9066984136172977E-03_dp, -1.1387477711487443E-03_dp, &
         1.3430321326247108E-03_dp, -1.5016650487631499E-04_dp, -6.8710530153629065E-04_dp, &
         5.8196249923196924E-04_dp, 1.3851688904176260E+01_dp, -1.4271598955417318E+01_dp, &
         6.5394232734786746E+00_dp, -1.2157582321122917E+00_dp, -1.7145472518196578E-01_dp, &
         1.1137485470268504E-01_dp, 4.3748739759179806E-03_dp, -1.0567793142941557E-02_dp, &
         -5.5398723777692134E-04_dp, 2.1139656566692543E-03_dp, -3.7128217874996786E-04_dp, &
         -9.8650715039269197E-04_dp, 8.4812423349403076E-04_dp, 1.9235334619586016E+01_dp, &
         -2.1430578544580889E+01_dp, 1.0935348023770239E+01_dp, -2.5703290931021043E+00_dp, &
         -9.3684439339661404E-02_dp, 1.9608483863631485E-01_dp, -1.2873315513389102E-02_dp, &
         -1.6635857332202259E-02_dp, 1.3017736724714114E-03_dp, 3.0919884927816301E-03_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/-8.1899012855183456E-04_dp, &
         -1.3697632531141069E-03_dp, 1.2268039489488590E-03_dp, 2.6982367118180953E+01_dp, &
         -3.2203467268313922E+01_dp, 1.8094671738936693E+01_dp, -5.1291575498824864E+00_dp, &
         2.2721945369117036E-01_dp, 3.0848319451415068E-01_dp, -5.4944705448018556E-02_dp, &
         -2.1918290310463735E-02_dp, 5.3889292561536137E-03_dp, 4.0275986677558995E-03_dp, &
         -1.6284311851828951E-03_dp, -1.8215037039926844E-03_dp, 1.7609141110641021E-03_dp, &
         3.8463536907902750E+01_dp, -4.8640580195082968E+01_dp, 2.9722457220673043E+01_dp, &
         -9.8185297472519668E+00_dp, 1.1017236982453140E+00_dp, 4.1094138704569305E-01_dp, &
         -1.3790144274071689E-01_dp, -1.9836806896771739E-02_dp, 1.2281130435731527E-02_dp, &
         4.2485407574071928E-03_dp, -2.8502932676645093E-03_dp, -2.3204621866155586E-03_dp, &
         2.5139038804018183E-03_dp, 5.6403019544607950E+01_dp, -7.4558074137380629E+01_dp, &
         4.8792909062352564E+01_dp, -1.8222521073735450E+01_dp, 3.1113259029642006E+00_dp, &
         3.8525550418227494E-01_dp, -2.6956980951719761E-01_dp, 1.7730572492532272E-03_dp, &
         2.0231218118677733E-02_dp, 2.8971793876300693E-03_dp, -4.1670706906750512E-03_dp, &
         -3.0299779823530571E-03_dp, 3.6417755428282965E-03_dp, 8.7215443808268901E+01_dp, &
         -1.1851941798879672E+02_dp, 8.1443910354759353E+01_dp, -3.3294997428270143E+01_dp, &
         7.2813751191796143E+00_dp, -1.4949444862504643E-02_dp, -4.2555970602879750E-01_dp, &
         5.3440375091183941E-02_dp, 2.3936805411273999E-02_dp, 4.2943352588307808E-04_dp, &
         -4.9592740606350010E-03_dp, -4.7297595630543931E-03_dp, 5.6368552675931643E-03_dp, &
         1.4975251635328343E+02_dp, -2.0531240455284888E+02_dp, 1.4459250832690373E+02_dp, &
         -6.2316167155448824E+01_dp, 1.5665274878658106E+01_dp, -1.1636674097249506E+00_dp, &
         -5.5714790450689045E-01_dp, 1.2874387847404120E-01_dp, 2.0981878079326013E-02_dp, &
         1.6097121510941226E-04_dp, -6.0483121208980773E-03_dp, -8.8514602159330009E-03_dp, &
         9.7828665118702071E-03_dp, 3.2129985348974549E+02_dp, -4.3705037011365653E+02_dp, &
         3.0768674866967689E+02_dp, -1.3421668175563934E+02_dp, 3.5382166323770129E+01_dp, &
         -3.7150156401778789E+00_dp, -7.8066308454448430E-01_dp, 2.3019139497858526E-01_dp, &
         2.6698155626115754E-02_dp, 6.9031498421899722E-03_dp, -1.3061502264041481E-02_dp, &
         -1.8618105437648663E-02_dp, 2.0549011501193169E-02_dp, 1.2327094475164540E+03_dp, &
         -1.6517971287587309E+03_dp, 1.1453145166666175E+03_dp, -4.9100832192840704E+02_dp, &
         1.2656644760117726E+02_dp, -1.2735766360571926E+01_dp, -2.6884467290906815E+00_dp, &
         6.7843516874228083E-01_dp, 1.2057077781690270E-01_dp, 4.0211801204278225E-02_dp, &
         -5.5885010022905059E-02_dp, -6.8430520456983487E-02_dp, 7.7231660087382012E-02_dp, &
         1.8496027008239549E-01_dp, -5.2635957315967263E-02_dp, -1.3363208633747327E-02_dp, &
         7.0753080700979715E-03_dp, 4.0250101520871492E-03_dp, -2.5572951671070160E-03_dp, &
         -1.7101514222507485E-03_dp, 9.0274083105137438E-04_dp, 1.7429230871490228E-03_dp, &
         -1.9224938737565124E-03_dp, 5.7596057684833031E-04_dp, 1.0659543229708461E-04_dp, &
         -7.1924049427382215E-05_dp, 5.8244506483690739E-01_dp, -1.8223607569173905E-01_dp, &
         -3.8829020775716601E-02_dp, 2.7386770439359200E-02_dp, 1.1779696794311099E-02_dp, &
         -1.0280482019437921E-02_dp, -4.9672977515278658E-03_dp, 3.7715941587838780E-03_dp, &
         5.9254776501040194E-03_dp, -7.8754309889097502E-03_dp, 3.3370240188511616E-03_dp, &
         -2.4573980952193470E-04_dp, -1.3766621817983240E-04_dp, 1.0679356036687564E+00_dp, &
         -3.9301786697228080E-01_dp, -5.7337229831602214E-02_dp, 6.8138158271713281E-02_dp, &
         1.7189370118560782E-02_dp, -2.6680191557440031E-02_dp, -6.9610330874151343E-03_dp, &
         1.0234093194331958E-02_dp, 1.1928832504988295E-02_dp, -2.0825809621203371E-02_dp, &
         1.1845502328236113E-02_dp, -2.7625162573559322E-03_dp, 1.2266527208606984E-04_dp, &
         1.7186762769732131E+00_dp, -7.6842283829929103E-01_dp, -5.2094219929163819E-02_dp, &
         1.4952064221850928E-01_dp, 1.3105518526079931E-02_dp, -6.0277036362458583E-02_dp, &
         -3.8993134189913609E-03_dp, 2.3998100221899753E-02_dp, 1.9722360015163132E-02_dp, &
         -4.7449318260552320E-02_dp, 3.3022861705973044E-02_dp, -1.0525079839838058E-02_dp, &
         1.2565018126329791E-03_dp, 2.6433292578569403E+00_dp, -1.4442699325289519E+00_dp, &
         1.8749612571812301E-02_dp, 3.0152805079239342E-01_dp, -1.9072651296996059E-02_dp, &
         -1.2313703162394884E-01_dp, 1.3982139796035626E-02_dp, 5.0529807099352154E-02_dp, &
         2.5915913834361832E-02_dp, -9.6301299125299111E-02_dp, 7.8387553232621393E-02_dp, &
         -2.9347212002760113E-02_dp, 4.3984494251308407E-03_dp, 4.0031890214864516E+00_dp, &
         -2.6506777836645301E+00_dp, 2.5002709371286003E-01_dp, 5.6678418963927524E-01_dp, &
         -1.2166589054182736E-01_dp, -2.3025504915310566E-01_dp, 6.8512851967640287E-02_dp, &
         9.6897025962713448E-02_dp, 1.8655348484778574E-02_dp, -1.7509091764540302E-01_dp, &
         1.6512362797291477E-01_dp, -6.9241563143609311E-02_dp, 1.1677077362884182E-02_dp, &
         6.0451957328475263E+00_dp, -4.7842190085775247E+00_dp, 8.4607600090330592E-01_dp, &
         9.9838014493270644E-01_dp, -3.8386031304018225E-01_dp, -3.9385951905586181E-01_dp, &
         2.0432223719801115E-01_dp, 1.6924599740654714E-01_dp, -3.2820891989357739E-02_dp, &
         -2.8096972471170906E-01_dp, 3.1382191127514508E-01_dp, -1.4536535595852404E-01_dp, &
         2.6668413841726423E-02_dp, 9.1549135569876245E+00_dp, -8.5345664784207340E+00_dp, &
         2.2346205651622792E+00_dp, 1.6411516718069554E+00_dp, -9.8335661502361238E-01_dp, &
         -6.0422667740241987E-01_dp, 5.0601416326899085E-01_dp, 2.6377746002222652E-01_dp, &
         -1.9726347500816235E-01_dp, -3.7523198569027921E-01_dp, 5.3625497856328741E-01_dp, &
         -2.7599511569361151E-01_dp, 5.4528659506659791E-02_dp, 1.3943035992310987E+01_dp, &
         -1.5109323197529408E+01_dp, 5.2959149276944357E+00_dp, 2.4661972731333157E+00_dp, &
         -2.2567333201120254E+00_dp, -7.7787392178326420E-01_dp, 1.1214179222198557E+00_dp, &
         3.4190618100067971E-01_dp, -6.1152850879396203E-01_dp, -3.2726172380330121E-01_dp, &
         7.9909066319447652E-01_dp, -4.7152537410281026E-01_dp, 1.0076309501925237E-01_dp, &
         2.1388611942642935E+01_dp, -2.6637185765365800E+01_dp, 1.1824571495957908E+01_dp, &
         3.1834095908537434E+00_dp, -4.8063853776563112E+00_dp, -6.3222648705256546E-01_dp, &
         2.2812373088061939E+00_dp, 2.7053699574642753E-01_dp, -1.5164506840508636E+00_dp, &
         1.9390080320187825E-01_dp, 9.3774412092831516E-01_dp, -7.0197405735123353E-01_dp, &
         1.6622614771280861E-01_dp, 3.3079428604239290E+01_dp, -4.6898194866003884E+01_dp, &
         2.5464018502546349E+01_dp, 2.7484480884475584E+00_dp, -9.6279273060536603E+00_dp, &
         5.9552677798513798E-01_dp, 4.2759743597723183E+00_dp, -3.0699576070505968E-01_dp, &
         -3.2669097676275891E+00_dp, 1.8987807572689366E+00_dp, 4.8132579922105273E-01_dp, &
         -8.2065702179074362E-01_dp, 2.3362433782091457E-01_dp, 5.1621626508739872E+01_dp, &
         -8.2660353947017612E+01_dp, 5.3607937311050435E+01_dp, -1.8797644714387658E+00_dp, &
         -1.8143027819974449E+01_dp, 4.7369680178784570E+00_dp, 7.2692132436006371E+00_dp, &
         -2.2184482777750349E+00_dp, -6.2154142523989133E+00_dp, 6.1059357543242241E+00_dp, &
         -1.6522745114655870E+00_dp, -4.1115789631675914E-01_dp, 2.3756928000969607E-01_dp, &
         8.1351582550594543E+01_dp, -1.4615422101218340E+02_dp, 1.1129430956108159E+02_dp, &
         -1.8800774059904494E+01_dp, -3.1684177587873151E+01_dp, 1.5820010337807913E+01_dp, &
         1.0616280037682005E+01_dp, -7.1547281508105547E+00_dp, -1.0210813394710581E+01_dp, &
         1.4851459426239472E+01_dp, -7.5823159284462474E+00_dp, 1.4755007442866093E+00_dp, &
         7.1704854962532146E-03_dp, 1.2961216352300849E+02_dp, -2.5972094896967849E+02_dp, &
         2.2922833292335898E+02_dp, -6.8618818175144867E+01_dp, -4.8903263939216217E+01_dp, &
         4.1885218071518977E+01_dp, 1.0862853016310472E+01_dp, -1.8022963902540312E+01_dp, &
         -1.3119830070208033E+01_dp, 3.0238247725930801E+01_dp, -2.0742505936307470E+01_dp, &
         6.6612181005611406E+00_dp, -8.2499200913490378E-01_dp, 2.0918002109462682E+02_dp, &
         -4.6477061154654319E+02_dp, 4.7040392223308885E+02_dp, -2.0231664523073340E+02_dp, &
         -5.6353025630719443E+01_dp, 9.6488871901532022E+01_dp, -3.3153170467196271E+00_dp, &
         -3.8112379830274591E+01_dp, -7.2773894428920212E+00_dp, 5.0927199428565473E+01_dp, &
         -4.4589076648414782E+01_dp, 1.7786229263712446E+01_dp, -2.8840133398763834E+00_dp, &
         3.4348422405496660E+02_dp, -8.4023075962910389E+02_dp, 9.6509249025561269E+02_dp, &
         -5.4322732013402708E+02_dp, 1.1614085203846807E+00_dp, 1.9306947976062503E+02_dp, &
         -6.0953011969071319E+01_dp, -6.4473641869529004E+01_dp, 2.5599475521189774E+01_dp, &
         6.1661793378629682E+01_dp, -7.5299742803616596E+01_dp, 3.5784746893679127E+01_dp, &
         -6.6920352740423770E+00_dp, 5.8082030750243996E+02_dp, -1.5482290606357024E+03_dp, &
         1.9913534204232064E+03_dp, -1.3842027389980281E+03_dp, 3.0176915978714999E+02_dp, &
         3.0701230954874183E+02_dp, -2.1433469154899214E+02_dp, -6.6515928662840182E+01_dp, &
         1.0587060354314622E+02_dp, 2.4158616076495992E+01_dp, -8.8819422206168397E+01_dp, &
         5.3284774844267872E+01_dp, -1.1352258425217249E+01_dp, 1.0507824282448173E+03_dp, &
         -3.0034485650910306E+03_dp, 4.2378849656165157E+03_dp, -3.4559382913401400E+03_dp, &
         1.3449288846420820E+03_dp, 2.6444208061663255E+02_dp, -4.9877900967018985E+02_dp, &
         2.5127991468850134E+01_dp, 2.1767335090020083E+02_dp, -8.6072216194219422E+01_dp, &
         -6.0325796108173265E+01_dp, 6.0083116119914528E+01_dp, -1.5156462785073193E+01_dp, &
         2.3717539683229165E+03_dp, -7.0993078298523542E+03_dp, 1.0652966496878691E+04_dp, &
         -9.6140748259170232E+03_dp, 4.8504224379382586E+03_dp, -4.4395885474091079E+02_dp, &
         -9.6288919502292026E+02_dp, 2.5917249174048044E+02_dp, 4.0854133722363883E+02_dp, &
         -3.0143744269409973E+02_dp, -4.5418714874973256E+00_dp, 7.5623186578469742E+01_dp, &
         -2.3436275595093466E+01_dp, 1.4977112766698923E+00_dp, -4.3677613112273472E-01_dp, &
         -1.0626472632002160E-01_dp, 6.0579861136460085E-02_dp, 3.2094769702270376E-02_dp, &
         -2.2140138450254837E-02_dp, -1.3621095148298569E-02_dp, 7.9037816732341545E-03_dp, &
         1.4423141126330896E-02_dp, -1.6738731357106910E-02_dp, 5.6195012107057513E-03_dp, &
         5.0623354758645299E-04_dp, -5.2923485963463817E-04_dp, 1.7217160143503936E+00_dp, &
         -6.3675005873485835E-01_dp, -9.4151747966956217E-02_dp, 1.1116729873281835E-01_dp, &
         2.8800708837475177E-02_dp, -4.3580539400664406E-02_dp, -1.1821195607030708E-02_dp, &
         1.6681071628927815E-02_dp, 1.9836760036767896E-02_dp, -3.4063478523860956E-02_dp, &
         1.9103471265017376E-02_dp, -4.3249027191962740E-03_dp, 1.5532200025133099E-04_dp, &
         2.2127628736646106E+00_dp, -1.1034935942153976E+00_dp, -4.3577466272544045E-02_dp, &
         2.2803871592728966E-01_dp, 1.0104708726025797E-02_dp, -9.3215357617339387E-02_dp, &
         -1.8053387009684143E-03_dp, 3.7448006485105621E-02_dp, 2.8302008106949928E-02_dp, &
         -7.3968859165537221E-02_dp, 5.3453226207458779E-02_dp, -1.7780936779581476E-02_dp, &
         2.2736357196637734E-03_dp, 3.0642372558529560E+00_dp, -1.9876242013174268E+00_dp, &
         1.1168757098085445E-01_dp, 4.4279774497587876E-01_dp, -5.4403602466477574E-02_dp, &
         -1.8361663846171278E-01_dp, 3.2420048824757658E-02_dp, 7.6305200794048844E-02_dp, &
         3.2103644959620231E-02_dp, -1.4486824002301485E-01_dp, 1.2355203927413878E-01_dp, &
         -4.8092052541477776E-02_dp, 7.5259001655546970E-03_dp, 4.4383894095260246E+00_dp, &
         -3.5679718788094252E+00_dp, 5.1476908985660208E-01_dp, 8.0172648142794001E-01_dp, &
         -2.2976447620038551E-01_dp, -3.2956771002295171E-01_dp, 1.2448866609398915E-01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/1.4079316176694362E-01_dp, &
         1.0204817722398823E-02_dp, -2.5165636141352937E-01_dp, 2.5091425850523480E-01_dp, &
         -1.0907652615992168E-01_dp, 1.8995561214566335E-02_dp, 6.6017064494743369E+00_dp, &
         -6.3396206574849705E+00_dp, 1.4645915581906761E+00_dp, 1.3567138102239846E+00_dp, &
         -6.4776462885481734E-01_dp, -5.3664210086520570E-01_dp, 3.4037092684481507E-01_dp, &
         2.3476223625427250E-01_dp, -8.7556632295100814E-02_dp, -3.7830556673362864E-01_dp, &
         4.5738992125084094E-01_dp, -2.2026577794379970E-01_dp, 4.1599788241118318E-02_dp, &
         9.9858326732371161E+00_dp, -1.1174218555192047E+01_dp, 3.5747006721239503E+00_dp, &
         2.1296809807914152E+00_dp, -1.5636677908078949E+00_dp, -7.6880577851663401E-01_dp, &
         8.0089159003032162E-01_dp, 3.4298150442274511E-01_dp, -3.6805510673665137E-01_dp, &
         -4.4644114872399454E-01_dp, 7.4222316810942002E-01_dp, -4.0152913150399111E-01_dp, &
         8.1829151439933356E-02_dp, 1.5289854701740817E+01_dp, -1.9608078427701756E+01_dp, &
         8.0917990845543581E+00_dp, 3.0041832124620726E+00_dp, -3.4494990735255384E+00_dp, &
         -8.6609625198945162E-01_dp, 1.7095858524603544E+00_dp, 3.9120562018981797E-01_dp, &
         -1.0360932314069153E+00_dp, -2.2758467416281059E-01_dp, 1.0203793428499848E+00_dp, &
         -6.5239517137454839E-01_dp, 1.4501295283759075E-01_dp, 2.3652443598918335E+01_dp, &
         -3.4362906201199991E+01_dp, 1.7536976813856338E+01_dp, 3.4319229466110013E+00_dp, &
         -7.1298251478464278E+00_dp, -3.5204255806543472E-01_dp, 3.3680128935599196E+00_dp, &
         1.5319402163742263E-01_dp, -2.4370830597624149E+00_dp, 8.1997373400602092E-01_dp, &
         9.8380055990672011E-01_dp, -8.9662392912364008E-01_dp, 2.2625667920011164E-01_dp, &
         3.6944557452608692E+01_dp, -6.0295621236170625E+01_dp, 3.7003398215145729E+01_dp, &
         1.6834081679861252E+00_dp, -1.3925568490337929E+01_dp, 1.9888475074288903E+00_dp, &
         6.1139267076108847E+00_dp, -9.4588148977566311E-01_dp, -5.0444398561656572E+00_dp, &
         3.8065719195626566E+00_dp, -1.6831147750489744E-01_dp, -8.5900446249491913E-01_dp, &
         2.8780543981072360E-01_dp, 5.8274202677422771E+01_dp, -1.0614856307111749E+02_dp, &
         7.6800301489740932E+01_dp, -7.0079594145933406E+00_dp, -2.5617180005330020E+01_dp, &
         8.9815644644679367E+00_dp, 9.9914118876838458E+00_dp, -4.1991995337660075E+00_dp, &
         -9.2225161080668876E+00_dp, 1.0699436467194234E+01_dp, -4.1744895657292922E+00_dp, &
         1.6968874745604984E-01_dp, 2.1150225597430636E-01_dp, 9.2874008697647312E+01_dp, &
         -1.8781074240763044E+02_dp, 1.5788030861616468E+02_dp, -3.4972432415418027E+01_dp, &
         -4.3527351698049202E+01_dp, 2.6710127846672595E+01_dp, 1.3641876922340176E+01_dp, &
         -1.2188594976459228E+01_dp, -1.4352260173596875E+01_dp, 2.4342038573890058E+01_dp, &
         -1.4279132767898336E+01_dp, 3.7137073675422725E+00_dp, -2.8757236845263162E-01_dp, &
         1.4971128490394287E+02_dp, -3.3447762227306481E+02_dp, 3.2305460093335603E+02_dp, &
         -1.1305851039826165E+02_dp, -6.4420176838297635E+01_dp, 6.7169772276610786E+01_dp, &
         1.1142459865038420E+01_dp, -2.9260375212110183E+01_dp, -1.6413218989655853E+01_dp, &
         4.7209519021715842E+01_dp, -3.5494808693208000E+01_dp, 1.2580735551183873E+01_dp, &
         -1.7935944649833950E+00_dp, 2.4458399494694996E+02_dp, -6.0055061871622843E+02_dp, &
         6.6027498087928825E+02_dp, -3.1693333009014430E+02_dp, -6.6059668231172040E+01_dp, &
         1.5055616810459574E+02_dp, -1.6035040303862690E+01_dp, -6.0205525165442332E+01_dp, &
         -2.1832622908315549E+00_dp, 7.5674463602179827E+01_dp, -7.2506243056554936E+01_dp, &
         3.0788078901881761E+01_dp, -5.2942708588931886E+00_dp, 4.0681467176552570E+02_dp, &
         -1.0899373555775655E+03_dp, 1.3513251866930286E+03_dp, -8.2903784544199414E+02_dp, &
         3.8380965916357390E+01_dp, 2.9721645411143447E+02_dp, -1.1594319332969438E+02_dp, &
         -9.9972427583413506E+01_dp, 6.1208430774721975E+01_dp, 8.0856974870358002E+01_dp, &
         -1.1693883280777918E+02_dp, 5.9389649528398166E+01_dp, -1.1628989395474973E+01_dp, &
         6.9722251479747388E+02_dp, -2.0140066190150276E+03_dp, 2.7790283348654434E+03_dp, &
         -2.0796098972388722E+03_dp, 5.2869394312949373E+02_dp, 4.6652987556767943E+02_dp, &
         -3.8221920854190296E+02_dp, -9.2752399207456079E+01_dp, 2.1910131922818687E+02_dp, &
         -2.2923675949907071E+01_dp, -1.1207458200684393E+02_dp, 7.9761888545321156E+01_dp, &
         -1.8298506671237661E+01_dp, 1.2744662066142864E+03_dp, -3.8866078972090563E+03_dp, &
         5.8310452947172871E+03_dp, -5.0818244125139472E+03_dp, 2.1788811500495763E+03_dp, &
         3.4081840675497295E+02_dp, -8.7556575958971519E+02_dp, 1.2965937398586175E+02_dp, &
         4.1054055060278523E+02_dp, -2.9796372200111415E+02_dp, 2.6343327262803022E+01_dp, &
         5.0424032533223901E+01_dp, -1.7235143944934833E+01_dp, 2.7604064480154507E+03_dp, &
         -8.6019432779733779E+03_dp, 1.3474832660141481E+04_dp, -1.2816761276941676E+04_dp, &
         7.0149847866566979E+03_dp, -9.9103658229985535E+02_dp, -1.3749290812180745E+03_dp, &
         6.2574740788577901E+02_dp, 4.7802482876747689E+02_dp, -5.7305069900934677E+02_dp, &
         1.8314783781950712E+02_dp, 1.5464936084215722E+01_dp, -1.5810901025213084E+01_dp, &
         1.0274044897416868E+04_dp, -3.1659430536269701E+04_dp, 4.9322572161259443E+04_dp, &
         -4.7163561469793902E+04_dp, 2.6891304291869375E+04_dp, -5.7571317404894771E+03_dp, &
         -3.0444779122812329E+03_dp, 1.4450586928263419E+03_dp, 1.4829738315237773E+03_dp, &
         -1.4849083905506739E+03_dp, 2.8382264810325569E+02_dp, 1.7516211991683696E+02_dp, &
         -7.2468749687854157E+01_dp, 2.1318222830584657E-01_dp, -3.3684034054268112E-02_dp, &
         -1.0832627512700675E-02_dp, 3.3145187729886286E-03_dp, 2.7275281503664904E-03_dp, &
         -9.9837098618575626E-04_dp, -1.0357840754560805E-03_dp, 3.5875531553344507E-04_dp, &
         6.0259734691266913E-04_dp, -3.5423446604539031E-04_dp, -1.7556949097338197E-04_dp, &
         2.2034591354691384E-04_dp, -6.0116378539179255E-05_dp, 6.8243285676064791E-01_dp, &
         -1.2222900203139513E-01_dp, -3.5650354344508521E-02_dp, 1.4124232653500343E-02_dp, &
         9.3169904587289860E-03_dp, -4.4428123313210782E-03_dp, -3.5817403682234456E-03_dp, &
         1.6213661243337648E-03_dp, 2.2617072347383576E-03_dp, -1.7251617279990248E-03_dp, &
         -3.0117239462936477E-04_dp, 6.9071851016345640E-04_dp, -2.0557421403978617E-04_dp, &
         1.2918736815826171E+00_dp, -2.8431849514217356E-01_dp, -6.9503682880585616E-02_dp, &
         4.0079778234935019E-02_dp, 1.9169899730441080E-02_dp, -1.3254501345108947E-02_dp, &
         -7.4879325459152992E-03_dp, 4.9273958248561584E-03_dp, 5.4431528418503244E-03_dp, &
         -5.6527461842056205E-03_dp, 6.4400303030163011E-04_dp, 1.1358340960384338E-03_dp, &
         -4.1629994324220604E-04_dp, 2.1761841600359864E+00_dp, -6.0602677049954834E-01_dp, &
         -1.1600614333211007E-01_dp, 1.0094374685942331E-01_dp, 3.3524435971905575E-02_dp, &
         -3.4830286515735560E-02_dp, -1.3217120228077264E-02_dp, 1.3167612202744752E-02_dp, &
         1.1658344857093769E-02_dp, -1.6002826381700426E-02_dp, 4.9726471567697389E-03_dp, &
         1.0322574937840596E-03_dp, -6.7937956240013083E-04_dp, 3.5436318724159870E+00_dp, &
         -1.2448897785654798E+00_dp, -1.7047825329078076E-01_dp, 2.3516821229813373E-01_dp, &
         4.9997768154929345E-02_dp, -8.3991468599014865E-02_dp, -1.9461645231788477E-02_dp, &
         3.2227076132211548E-02_dp, 2.2927327872461805E-02_dp, -4.0981673652954999E-02_dp, &
         1.8645143724385239E-02_dp, -1.4710909163262332E-03_dp, -7.9628203284441701E-04_dp, &
         5.7352535342850013E+00_dp, -2.5003246419581067E+00_dp, -2.0277450208455863E-01_dp, &
         5.1788148421079194E-01_dp, 5.4494531427097280E-02_dp, -1.9004943455969087E-01_dp, &
         -1.9106874087512889E-02_dp, 7.3855738319935568E-02_dp, 4.1005914394026836E-02_dp, &
         -9.7462681711003807E-02_dp, 5.6025894780165515E-02_dp, -1.1671638371992546E-02_dp, &
         -1.2317531086698035E-05_dp, 9.3283234643961990E+00_dp, -4.9517090926784189E+00_dp, &
         -1.1237619085161876E-01_dp, 1.0938292009097146E+00_dp, -3.6578939270913327E-04_dp, &
         -4.0903466934184857E-01_dp, 1.1571092173393985E-02_dp, 1.6054688245457299E-01_dp, &
         6.3796913867728455E-02_dp, -2.1826491336856391E-01_dp, 1.4979809357033552E-01_dp, &
         -4.3231775436116136E-02_dp, 3.9227329925919244E-03_dp, 1.5322921886299406E+01_dp, &
         -9.7350236118409814E+00_dp, 3.8501066580685789E-01_dp, 2.2354405913551587E+00_dp, &
         -2.5000970304965098E-01_dp, -8.4251562237350353E-01_dp, 1.4017068692381385E-01_dp, &
         3.3257684318281500E-01_dp, 7.3319727667607221E-02_dp, -4.6190571472728487E-01_dp, &
         3.7010532372901361E-01_dp, -1.2889503527555926E-01_dp, 1.6973488867294845E-02_dp, &
         2.5485983352834001E+01_dp, -1.9107468063951504E+01_dp, 2.0402311845689449E+00_dp, &
         4.4371579701624126E+00_dp, -1.0529422595840883E+00_dp, -1.6585347909210049E+00_dp, &
         5.4410632305326356E-01_dp, 6.5376504315884909E-01_dp, 4.1661500351891439E-04_dp, &
         -9.1692979997264157E-01_dp, 8.5690261662255818E-01_dp, -3.4197738694486834E-01_dp, &
         5.3687965659661094E-02_dp, 4.2999540010684285E+01_dp, -3.7622665990809367E+01_dp, &
         6.7752865496365642E+00_dp, 8.5406611669389410E+00_dp, -3.3166201661466737E+00_dp, &
         -3.0842874201654951E+00_dp, 1.6635414194652112E+00_dp, 1.1980859949014355E+00_dp, &
         -3.7016765925282619E-01_dp, -1.6659214344665423E+00_dp, 1.8593601546633085E+00_dp, &
         -8.3381793576913044E-01_dp, 1.4684476667934723E-01_dp, 7.3709904847928598E+01_dp, &
         -7.4630438320318333E+01_dp, 1.9452643501026422E+01_dp, 1.5793696701925368E+01_dp, &
         -9.2675373118737916E+00_dp, -5.2467842188148071E+00_dp, 4.5372916469923439E+00_dp, &
         1.9496492078416041E+00_dp, -1.6349099319243841E+00_dp, -2.5906847912536093E+00_dp, &
         3.7265500877505593E+00_dp, -1.8799923390423963E+00_dp, 3.6327549283233479E-01_dp, &
         1.2859438624321660E+02_dp, -1.4972520267351325E+02_dp, 5.2383826750269051E+01_dp, &
         2.7367488598061911E+01_dp, -2.4219671211959398E+01_dp, -7.4290230400283290E+00_dp, &
         1.1475384287478459E+01_dp, 2.3922423306082199E+00_dp, -5.2921831798037999E+00_dp, &
         -2.6730364381610836E+00_dp, 6.6169572702353063E+00_dp, -3.8645306885918953E+00_dp, &
         8.1818488592343996E-01_dp, 2.2876608607133824E+02_dp, -3.0495185259837291E+02_dp, &
         1.3706237077883176E+02_dp, 4.1439813867536394E+01_dp, -6.0428830848804978E+01_dp, &
         -5.3399496761646414E+00_dp, 2.7145380315073801E+01_dp, 1.0963663429032429E-01_dp, &
         -1.4667285308366607E+01_dp, 2.0716234645629714E+00_dp, 9.1721609250763034E+00_dp, &
         -6.9025941825134423E+00_dp, 1.6398955739095382E+00_dp, 4.1594691714391070E+02_dp, &
         -6.3301093600335355E+02_dp, 3.5553694173797044E+02_dp, 4.0753399602504416E+01_dp, &
         -1.4452338056841705E+02_dp, 1.7908795314096562E+01_dp, 5.9142120003589028E+01_dp, &
         -1.3457064524764537E+01_dp, -3.5600526788255479E+01_dp, 2.3692561370282714E+01_dp, &
         4.1369265076592301E+00_dp, -9.1363336250763520E+00_dp, 2.7024750645260252E+00_dp, &
         7.7521438841071472E+02_dp, -1.3447541987663474E+03_dp, 9.2710345217552867E+02_dp, &
         -5.5662691197356949E+01_dp, -3.2666994854804346E+02_dp, 1.1949426674679283E+02_dp, &
         1.1148985958051878E+02_dp, -6.4189392858052045E+01_dp, -7.2080786102618873E+01_dp, &
         9.1605160930916881E+01_dp, -3.3190298346214078E+01_dp, -1.1968423393425880E+00_dp, &
         2.5799094822024733E+00_dp, 1.4870387159310956E+03_dp, -2.9373961910368489E+03_dp, &
         2.4568455127217608E+03_dp, -5.7572036132261246E+02_dp, -6.5915764219002381E+02_dp, &
         4.7301818737714882E+02_dp, 1.3863635256023798E+02_dp, -2.1631236718932931E+02_dp, &
         -9.4803921196274686E+01_dp, 2.5426811844807858E+02_dp, -1.6576110763951490E+02_dp, &
         4.7407189814529652E+01_dp, -4.5099294020295337E+00_dp, 2.9586882462973167E+03_dp, &
         -6.6390589153338169E+03_dp, 6.6755402509804217E+03_dp, -2.7894432451376142E+03_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c04 = (/-9.1855299569510419E+02_dp, &
         1.5086598025400613E+03_dp, -1.6982251398122094E+02_dp, -5.5006675710728985E+02_dp, &
         8.6766882152744344E+01_dp, 4.7256010495026828E+02_dp, -4.6889366551390941E+02_dp, &
         1.9581559031410015E+02_dp, -3.2716661828879559E+01_dp, 6.2548032409521911E+03_dp, &
         -1.5788950870448605E+04_dp, 1.8797183431156966E+04_dp, -1.1500723521848280E+04_dp, &
         1.1311068823604562E+03_dp, 3.5696653498968658E+03_dp, -1.9548307992473563E+03_dp, &
         -6.5409377594636578E+02_dp, 8.8415097294428301E+02_dp, 2.0828579656796524E+02_dp, &
         -6.8559158021726762E+02_dp, 3.9487798673853291E+02_dp, -8.1813494801799735E+01_dp, &
         1.5900659171374931E+04_dp, -4.3865264886054203E+04_dp, 5.9368981473015963E+04_dp, &
         -4.6496642597520760E+04_dp, 1.7576030674673242E+04_dp, 3.1090094533818497E+03_dp, &
         -6.2551252886095626E+03_dp, 1.0254795734419797E+03_dp, 1.7700143703154242E+03_dp, &
         -6.9300411007170703E+02_dp, -5.9600061313823699E+02_dp, 5.5645136364522648E+02_dp, &
         -1.3789134413973542E+02_dp, 1.7333022186632598E+00_dp, -2.8295145016009599E-01_dp, &
         -8.8783345881969303E-02_dp, 2.9137501520617400E-02_dp, 2.2582343344786884E-02_dp, &
         -8.8870496740718532E-03_dp, -8.6050069441200982E-03_dp, 3.2075742114612876E-03_dp, &
         5.1124562607267323E-03_dp, -3.2396586093912825E-03_dp, -1.2782830098815570E-03_dp, &
         1.7892266110778137E-03_dp, -4.9828059483079956E-04_dp, 2.0839177842079417E+00_dp, &
         -4.5883342975975994E-01_dp, -1.1377653794136394E-01_dp, 6.4262118792829045E-02_dp, &
         3.1589922005670561E-02_dp, -2.1128047919275923E-02_dp, -1.2370808453459683E-02_dp, &
         7.8307499965700952E-03_dp, 8.8525819512241723E-03_dp, -8.9036430575215944E-03_dp, &
         7.7278653207137722E-04_dp, 1.9574113869564288E-03_dp, -6.9432538273257499E-04_dp, &
         2.8790965126363099E+00_dp, -8.9291494000038507E-01_dp, -1.5980217852948253E-01_dp, &
         1.5593871145279911E-01_dp, 4.8212590796327996E-02_dp, -5.4049164283410382E-02_dp, &
         -1.9262868395269681E-02_dp, 2.0444212857397715E-02_dp, 1.7597567867215835E-02_dp, &
         -2.4946097804552481E-02_dp, 8.1983007050586126E-03_dp, 1.3052693431584528E-03_dp, &
         -9.9046233632061080E-04_dp, 4.3304768482591722E+00_dp, -1.7813853808128957E+00_dp, &
         -2.1010494972641455E-01_dp, 3.5547178388532347E-01_dp, 6.5506905553492881E-02_dp, &
         -1.2798724844398165E-01_dp, -2.5933808745545982E-02_dp, 4.9217194247932812E-02_dp, &
         3.3506489973277181E-02_dp, -6.3047933152743540E-02_dp, 3.0110932740404765E-02_dp, &
         -3.2402195006371740E-03_dp, -1.0036785562597503E-03_dp, 6.8299081160337032E+00_dp, &
         -3.5205397689128302E+00_dp, -2.0822929848963964E-01_dp, 7.6631150098260470E-01_dp, &
         5.8155387220312556E-02_dp, -2.8404749249030603E-01_dp, -1.9917085295170896E-02_dp, &
         1.1077158299302442E-01_dp, 5.8009365155884812E-02_dp, -1.4754924485395102E-01_dp, &
         8.8224148397498189E-02_dp, -2.0017588821696984E-02_dp, 5.1524515103809545E-04_dp, &
         1.1076132539365005E+01_dp, -6.8920385138882194E+00_dp, 1.0149059418049949E-02_dp, &
         1.5881453532787797E+00_dp, -5.0514917369298917E-02_dp, -6.0049701223304608E-01_dp, &
         3.6944486690867563E-02_dp, 2.3676450668964813E-01_dp, 8.5900161567404182E-02_dp, &
         -3.2547794450289452E-01_dp, 2.3138356568902774E-01_dp, -7.0029009427246630E-02_dp, &
         7.1213088371686710E-03_dp, 1.8313497271878418E+01_dp, -1.3441485775194154E+01_dp, &
         8.8580011501243916E-01_dp, 3.1948031252200022E+00_dp, -4.7058826034590084E-01_dp, &
         -1.2174961672232065E+00_dp, 2.4957240968045155E-01_dp, 4.8310590222314870E-01_dp, &
         8.6483956963456024E-02_dp, -6.7940646376194236E-01_dp, 5.6338924998429407E-01_dp, &
         -2.0291123162723659E-01_dp, 2.8032368883630622E-02_dp, 3.0780617013122821E+01_dp, &
         -2.6257569697681561E+01_dp, 3.5521591798072638E+00_dp, 6.2587792976548826E+00_dp, &
         -1.7460143870796279E+00_dp, -2.3619614892636438E+00_dp, 8.8818001701109950E-01_dp, &
         9.3606848206238136E-01_dp, -5.0794860926712801E-02_dp, -1.3302240326510510E+00_dp, &
         1.2893794094134079E+00_dp, -5.2876901741570692E-01_dp, 8.5451865781826092E-02_dp, &
         5.2567287396157639E+01_dp, -5.1604816743560406E+01_dp, 1.0866611400666184E+01_dp, &
         1.1908868775950642E+01_dp, -5.2359341711279104E+00_dp, -4.3250063052196532E+00_dp, &
         2.6159067633220250E+00_dp, 1.6876932949921251E+00_dp, -6.6168203286073823E-01_dp, &
         -2.3751835992262973E+00_dp, 2.7700021611620960E+00_dp, -1.2732445592684205E+00_dp, &
         2.2902755062897792E-01_dp, 9.1285593514466271E+01_dp, -1.0241866626620062E+02_dp, &
         2.9991030801919234E+01_dp, 2.1769589255916912E+01_dp, -1.4249979282246676E+01_dp, &
         -7.2044673782332227E+00_dp, 6.9889285697343118E+00_dp, 2.6782377021603785E+00_dp, &
         -2.6677863139577163E+00_dp, -3.5770528758936435E+00_dp, 5.4957103661007007E+00_dp, &
         -2.8440183689455196E+00_dp, 5.5949615778311945E-01_dp, 1.6140675350929016E+02_dp, &
         -2.0596934838864317E+02_dp, 7.8952124827283598E+01_dp, 3.7178752267348052E+01_dp, &
         -3.6659944916518448E+01_dp, -9.7656735399369428E+00_dp, 1.7464710569932944E+01_dp, &
         3.0679469902932719E+00_dp, -8.3851627852850044E+00_dp, -3.2636862826796440E+00_dp, &
         9.6185791182519313E+00_dp, -5.7987749350325295E+00_dp, 1.2497007872223715E+00_dp, &
         2.9110313668489431E+02_dp, -4.2112906631689452E+02_dp, 2.0368167006668779E+02_dp, &
         5.4837106594561327E+01_dp, -9.0638707836013396E+01_dp, -5.3325509510817621E+00_dp, &
         4.1073827539931052E+01_dp, -8.9364940761646350E-01_dp, -2.3013798898787297E+01_dp, &
         4.9430590435737045E+00_dp, 1.2851186457902051E+01_dp, -1.0246075994503601E+01_dp, &
         2.4884401830064262E+00_dp, 5.3670606902267980E+02_dp, -8.7847920003528930E+02_dp, &
         5.2349235503272166E+02_dp, 4.8749806265367148E+01_dp, -2.1602127708232794E+02_dp, &
         3.1729321388441790E+01_dp, 8.9567868142842315E+01_dp, -2.2390677578649012E+01_dp, &
         -5.6017285780368070E+01_dp, 4.0387845593738454E+01_dp, 3.4667266516341293E+00_dp, &
         -1.3141255438243856E+01_dp, 4.0592288247989385E+00_dp, 1.0143591938132444E+03_dp, &
         -1.8766993856001438E+03_dp, 1.3564571841587124E+03_dp, -9.8696563163335753E+01_dp, &
         -4.9033312510266632E+02_dp, 1.8780383859095875E+02_dp, 1.7118665610355563E+02_dp, &
         -1.0185617481240935E+02_dp, -1.1558713439161205E+02_dp, 1.5151598909811219E+02_dp, &
         -5.8664397974976858E+01_dp, 6.3267833198642698E-01_dp, 3.6766915824367761E+00_dp, &
         1.9726076865000123E+03_dp, -4.1232818836429924E+03_dp, 3.5781197436471589E+03_dp, &
         -8.5509074775503893E+02_dp, -1.0123738678655812E+03_dp, 7.2969707150714839E+02_dp, &
         2.2828683121334075E+02_dp, -3.4563856543415034E+02_dp, -1.6400405779415590E+02_dp, &
         4.2769450363111764E+02_dp, -2.8094753019966493E+02_dp, 8.1695817906598137E+01_dp, &
         -8.0867822790318300E+00_dp, 3.9722127523258755E+03_dp, -9.3649615551361039E+03_dp, &
         9.6872702091042847E+03_dp, -4.0152449268719452E+03_dp, -1.5781648063622556E+03_dp, &
         2.3817881583693224E+03_dp, -1.7512622509953385E+02_dp, -9.4466208919082828E+02_dp, &
         8.8832120896167638E+01_dp, 8.8054897643088452E+02_dp, -8.4980315712561492E+02_dp, &
         3.5239533036204887E+02_dp, -5.8700991234149221E+01_dp, 8.4183972987919678E+03_dp, &
         -2.2200192543210582E+04_dp, 2.7095512964836042E+04_dp, -1.6398496436394067E+04_dp, &
         5.8588432027345868E+02_dp, 6.3339571905285420E+03_dp, -3.0333212281281785E+03_dp, &
         -1.5835275583468535E+03_dp, 1.7348972889932504E+03_dp, 5.0425398982402010E+02_dp, &
         -1.4466434445749767E+03_dp, 8.3093046218870779E+02_dp, -1.7343507171721367E+02_dp, &
         2.0025459522044373E+04_dp, -5.7346919835612738E+04_dp, 7.9917525300310852E+04_dp, &
         -6.3405260767018226E+04_dp, 2.2238895649329628E+04_dp, 8.4397777042427315E+03_dp, &
         -1.2412923298499947E+04_dp, 2.2353310629429734E+03_dp, 4.1643117321199643E+03_dp, &
         -2.9557407553812504E+03_dp, 1.6207931060663054E+02_dp, 5.4977314747273965E+02_dp, &
         -1.7928307380491745E+02_dp, 7.4440121917214812E+04_dp, -2.1677815388585156E+05_dp, &
         3.1577001931789052E+05_dp, -2.7930028207177343E+05_dp, 1.4320805176019639E+05_dp, &
         -2.1245008317147029E+04_dp, -2.1554991554281445E+04_dp, 9.9376537143063269E+03_dp, &
         4.6744532080188510E+03_dp, -4.2905179066964693E+03_dp, -5.3149376874680104E+02_dp, &
         1.4881154495908686E+03_dp, -4.3984467711458336E+02_dp, 2.3404090378356895E-01_dp, &
         -2.0097322397469076E-02_dp, -1.1609180582537834E-02_dp, 1.8601938852069603E-03_dp, &
         3.4063516337523563E-03_dp, -2.5328555239247899E-04_dp, -2.1526388476503741E-03_dp, &
         2.6541196853022917E-04_dp, 1.8337856370389378E-03_dp, -1.4007221924521144E-03_dp, &
         1.1370777613700699E-04_dp, 2.7147984801578659E-04_dp, -9.3281631667276296E-05_dp, &
         7.5958720353542519E-01_dp, -7.5810202725780582E-02_dp, -4.2313502138352527E-02_dp, &
         8.6657522638846456E-03_dp, 1.2968885710289797E-02_dp, -1.6608526759463512E-03_dp, &
         -8.2851212755105860E-03_dp, 1.2914133195331631E-03_dp, 7.4052964778980055E-03_dp, &
         -6.1849568325369995E-03_dp, 1.0315421017034565E-03_dp, 8.3158039993379697E-04_dp, &
         -3.2797209696527126E-04_dp, 1.4768419304919898E+00_dp, -1.8729413362838598E-01_dp, &
         -9.8945335892159453E-02_dp, 2.7515284560837457E-02_dp, 3.2275791153780403E-02_dp, &
         -6.8395663023995482E-03_dp, -2.0986896600458896E-02_dp, 4.3486105991930634E-03_dp, &
         2.0127441003510580E-02_dp, -1.8849674488696535E-02_dp, 5.0191511478448469E-03_dp, &
         1.2331926287606307E-03_dp, -7.0207519810811295E-04_dp, 2.5840964671288975E+00_dp, &
         -4.2690659349976090E-01_dp, -2.1145308910161936E-01_dp, 7.7555603458720743E-02_dp, &
         7.3310657354651587E-02_dp, -2.2709496661448086E-02_dp, -4.8700826546977563E-02_dp, &
         1.3059501608799757E-02_dp, 5.0389115213805714E-02_dp, -5.2544419268517009E-02_dp, &
         1.8412947257736761E-02_dp, 3.6589716526811415E-04_dp, -1.2543984993995988E-03_dp, &
         4.4117567910870852E+00_dp, -9.3963835482824520E-01_dp, -4.3338088445410117E-01_dp, &
         2.0292403047377275E-01_dp, 1.5835719985334940E-01_dp, -6.6721718153669785E-02_dp, &
         -1.0781515206665805E-01_dp, 3.6564985747604667E-02_dp, 1.2068898656418255E-01_dp, &
         -1.3900826230358432E-01_dp, 5.8565521670857705E-02_dp, -5.8908104279455217E-03_dp, &
         -1.7476722084679196E-03_dp, 7.5445309724776202E+00_dp, -2.0272400898868317E+00_dp, &
         -8.6251312311759643E-01_dp, 5.0671870987293466E-01_dp, 3.2885263809384618E-01_dp, &
         -1.8271408651872414E-01_dp, -2.3038824323149212E-01_dp, 9.8194271811996164E-02_dp, &
         2.8055474423540522E-01_dp, -3.5598861752761291E-01_dp, 1.7250528507673124E-01_dp, &
         -3.0809943938599049E-02_dp, -8.6747384125205782E-04_dp, 1.3055861531649315E+01_dp, &
         -4.3314881266793979E+00_dp, -1.6761083589215517E+00_dp, 1.2312731236949950E+00_dp, &
         6.5894439998792775E-01_dp, -4.8036396219622884E-01_dp, -4.7776938854359963E-01_dp, &
         2.5796491007239708E-01_dp, 6.3870537313146425E-01_dp, -8.9484473757249905E-01_dp, &
         4.8676186523295428E-01_dp, -1.1467909217180164E-01_dp, 6.3856877573706490E-03_dp, &
         2.2973813103234260E+01_dp, -9.2447576119387715E+00_dp, -3.1862366340950352E+00_dp, &
         2.9511614190933089E+00_dp, 1.2698033917738343E+00_dp, -1.2342653176010880E+00_dp, &
         -9.6141838186275863E-01_dp, 6.7220233082876679E-01_dp, 1.4302140919443622E+00_dp, &
         -2.2279767372545760E+00_dp, 1.3410180260314670E+00_dp, -3.7636626450613619E-01_dp, &
         3.6869243870961514E-02_dp, 4.1235933833121898E+01_dp, -1.9856535292982361E+01_dp, &
         -5.9070868285783193E+00_dp, 7.0462234223485760E+00_dp, 2.3163307108458544E+00_dp, &
         -3.1353286429629565E+00_dp, -1.8613396596490337E+00_dp, 1.7553943424387961E+00_dp, &
         3.1455571650317942E+00_dp, -5.5222826187397756E+00_dp, 3.6478129213769028E+00_dp, &
         -1.1619081276403027E+00_dp, 1.4446834803157857E-01_dp, 7.5710886366993023E+01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c05 = (/-4.3207068963149325E+01_dp, &
         -1.0554104925448858E+01_dp, 1.6886407464491334E+01_dp, 3.8180534827516577E+00_dp, &
         -7.9351781128404237E+00_dp, -3.3784899123822600E+00_dp, 4.6319521704833901E+00_dp, &
         6.7239240450100590E+00_dp, -1.3642372104076109E+01_dp, 9.8589344031369794E+00_dp, &
         -3.4720965735924705E+00_dp, 4.9693345251046339E-01_dp, 1.4263572197170834E+02_dp, &
         -9.5857301698247568E+01_dp, -1.7588925878417403E+01_dp, 4.0872193751092674E+01_dp, &
         4.8184722668567055E+00_dp, -2.0108942400948376E+01_dp, -5.3282163675227920E+00_dp, &
         1.2435599975018286E+01_dp, 1.3556226217519358E+01_dp, -3.3461490386803618E+01_dp, &
         2.6523447839347572E+01_dp, -1.0184357103374717E+01_dp, 1.6085526053328969E+00_dp, &
         2.7676317157922375E+02_dp, -2.1826938584162323E+02_dp, -2.4695849258246266E+01_dp, &
         1.0043798763217026E+02_dp, -1.3947867483132029E-01_dp, -5.1127927754529111E+01_dp, &
         -5.1560469151207808E+00_dp, 3.4156690568543311E+01_dp, 2.3554036850825067E+01_dp, &
         -8.0434522638194011E+01_dp, 7.0744011691971110E+01_dp, -2.9459728084185446E+01_dp, &
         5.0323051191546622E+00_dp, 5.5569710557487417E+02_dp, -5.1390329950924320E+02_dp, &
         -1.5624537990859787E+01_dp, 2.5163004909295555E+02_dp, -3.5139741515483045E+01_dp, &
         -1.3006473245475482E+02_dp, 1.0550946938596450E+01_dp, 9.6255888915734857E+01_dp, &
         2.2051927486829232E+01_dp, -1.8305163287186298E+02_dp, 1.8429047593569263E+02_dp, &
         -8.3575794890910672E+01_dp, 1.5320836610304321E+01_dp, 1.1616226247615350E+03_dp, &
         -1.2623095596855940E+03_dp, 8.6933886681923610E+01_dp, 6.4414090263090793E+02_dp, &
         -1.9978982244793221E+02_dp, -3.2660522885185867E+02_dp, 1.0098963434615148E+02_dp, &
         2.7721445435633876E+02_dp, -8.9111509314974313E+01_dp, -3.5498277231331673E+02_dp, &
         4.4957236371547003E+02_dp, -2.2695337686830618E+02_dp, 4.4827771222570036E+01_dp, &
         2.5488397812481312E+03_dp, -3.2720196575275063E+03_dp, 6.2514938447284021E+02_dp, &
         1.6797717015255266E+03_dp, -9.0609391756137052E+02_dp, -7.7376478784489177E+02_dp, &
         5.1365112699708175E+02_dp, 7.9705181518637903E+02_dp, -8.1808800039951734E+02_dp, &
         -3.1602106846125673E+02_dp, 8.9474824450035931E+02_dp, -5.4669103099058145E+02_dp, &
         1.1970060288606828E+02_dp, 5.9381087032667047E+03_dp, -9.0911589870974240E+03_dp, &
         3.1519440108030135E+03_dp, 4.3742938194787512E+03_dp, -3.9151712191612141E+03_dp, &
         -1.4306765495221396E+03_dp, 2.2531425467709741E+03_dp, 2.0894645666101583E+03_dp, &
         -4.4544178967310481E+03_dp, 2.2383591529298183E+03_dp, 4.4379030282863710E+02_dp, &
         -8.3686062012151353E+02_dp, 2.3941035244133383E+02_dp, 1.4941870061839112E+04_dp, &
         -2.7698725637069860E+04_dp, 1.5368590597048256E+04_dp, 1.0343886022762168E+04_dp, &
         -1.7085228275250280E+04_dp, 1.0290154787655192E+03_dp, 8.9420986845818061E+03_dp, &
         2.9316545272252147E+03_dp, -1.8997312206692230E+04_dp, 1.9823842525145723E+04_dp, &
         -9.7693564885756969E+03_dp, 2.1647867710169985E+03_dp, -1.0378839433368972E+02_dp, &
         4.1812142330903938E+04_dp, -9.5876465655889318E+04_dp, 8.1756920297039134E+04_dp, &
         7.9039564977741147E+03_dp, -7.3842379280105539E+04_dp, 4.4204923551467022E+04_dp, &
         2.1730397459397547E+04_dp, -2.3137741549322684E+04_dp, -3.9615691029120862E+04_dp, &
         8.2843892021041815E+04_dp, -6.6187145198822764E+04_dp, 2.6773721003469775E+04_dp, &
         -4.5704387359470866E+03_dp, 1.4201349500169407E+05_dp, -4.0614829221312096E+05_dp, &
         5.2070095300974668E+05_dp, -2.7373429328130849E+05_dp, -1.5922589262630625E+05_dp, &
         3.8367646454422048E+05_dp, -2.5932110490762870E+05_dp, 2.5826860555147938E+04_dp, &
         5.5653603356457446E+04_dp, 3.5258106303212062E+03_dp, -4.9963042898806314E+04_dp, &
         3.4015110532782375E+04_dp, -7.7664796273009124E+03_dp, 1.9093974668075968E+00_dp, &
         -1.7054400904209266E-01_dp, -9.7641580671686004E-02_dp, 1.6766255796770199E-02_dp, &
         2.8988444707523806E-02_dp, -2.5620769652305881E-03_dp, -1.8370205087387178E-02_dp, &
         2.4203819467354571E-03_dp, 1.5850000696904308E-02_dp, -1.2409755539837933E-02_dp, &
         1.3104701010835478E-03_dp, 2.1950558162222062E-03_dp, -7.7840696413045292E-04_dp, &
         2.3816405893090655E+00_dp, -3.0058399408021230E-01_dp, -1.5977869438048026E-01_dp, &
         4.3205664883079593E-02_dp, 5.1954260872828326E-02_dp, -1.0435767802951716E-02_dp, &
         -3.3687805923760396E-02_dp, 6.7210749032354392E-03_dp, 3.1994491176302167E-02_dp, &
         -2.9480232130482186E-02_dp, 7.4474480162902649E-03_dp, 2.2084644917421768E-03_dp, &
         -1.1620522957490494E-03_dp, 3.4843292108184070E+00_dp, -6.3704889535412579E-01_dp, &
         -3.1247856056891354E-01_dp, 1.1946809479052665E-01_dp, 1.1015354223955937E-01_dp, &
         -3.5224341675936091E-02_dp, -7.3289581156592246E-02_dp, 1.9950208126348418E-02_dp, &
         7.6285287328668447E-02_dp, -7.9932304695661163E-02_dp, 2.8268292772225596E-02_dp, &
         3.8090423507330300E-04_dp, -1.8684578732096576E-03_dp, 5.5863369323075549E+00_dp, &
         -1.3713721433368866E+00_dp, -6.2196190605453838E-01_dp, 3.0824039077130033E-01_dp, &
         2.3240679161688174E-01_dp, -1.0242926268918769E-01_dp, -1.5885664856881426E-01_dp, &
         5.5322292014197520E-02_dp, 1.7989426377403095E-01_dp, -2.0915679370975637E-01_dp, &
         8.9364744914267866E-02_dp, -9.7174197412381834E-03_dp, -2.4358156829036412E-03_dp, &
         9.4101499417614978E+00_dp, -2.9205066473747960E+00_dp, -1.2163674262839712E+00_dp, &
         7.5863769149748450E-01_dp, 4.7467102033224490E-01_dp, -2.7655989190772234E-01_dp, &
         -3.3454490509347323E-01_dp, 1.4696126355412018E-01_dp, 4.1354189925528473E-01_dp, &
         -5.3037498985122311E-01_dp, 2.6031262682300332E-01_dp, -4.8168408431760540E-02_dp, &
         -7.7871669409946837E-04_dp, 1.6338815605587858E+01_dp, -6.1833514746412481E+00_dp, &
         -2.3311621003398191E+00_dp, 1.8217844419225089E+00_dp, 9.3758726271025805E-01_dp, &
         -7.1865224719975085E-01_dp, -6.8550844653228826E-01_dp, 3.8262747601717012E-01_dp, &
         9.3366571673349752E-01_dp, -1.3231064225590690E+00_dp, 7.2780109182395636E-01_dp, &
         -1.7510717849303473E-01_dp, 1.0697404777480228E-02_dp, 2.9045107955015176E+01_dp, &
         -1.3122115286638619E+01_dp, -4.3806637776150632E+00_dp, 4.3296889561695737E+00_dp, &
         1.7839258348899563E+00_dp, -1.8308198205695498E+00_dp, -1.3662114443942674E+00_dp, &
         9.9043960784398033E-01_dp, 2.0798949625159446E+00_dp, -3.2780668652253730E+00_dp, &
         1.9918330621946101E+00_dp, -5.6663166075695159E-01_dp, 5.7158452667943484E-02_dp, &
         5.2800048392041845E+01_dp, -2.8109720042733336E+01_dp, -8.0405851706326441E+00_dp, &
         1.0279577923775973E+01_dp, 3.2138452521312746E+00_dp, -4.6231746304589763E+00_dp, &
         -2.6241582510606816E+00_dp, 2.5740075419555373E+00_dp, 4.5650980871610862E+00_dp, &
         -8.1046297603881445E+00_dp, 5.3943552024246850E+00_dp, -1.7331884894412846E+00_dp, &
         2.1832820004259532E-01_dp, 9.8274043380510179E+01_dp, -6.1150381159704402E+01_dp, &
         -1.4222299310017810E+01_dp, 2.4549816525007763E+01_dp, 5.2132243676002723E+00_dp, &
         -1.1655504278147887E+01_dp, -4.7278213600031389E+00_dp, 6.7676550797734309E+00_dp, &
         9.7732086339422466E+00_dp, -2.0017061889583658E+01_dp, 1.4544370076050541E+01_dp, &
         -5.1483846630015107E+00_dp, 7.4121547100618823E-01_dp, 1.8774247624315791E+02_dp, &
         -1.3585965292035331E+02_dp, -2.3384417210458281E+01_dp, 5.9307322699677300E+01_dp, &
         6.3425555998420622E+00_dp, -2.9474617795317286E+01_dp, -7.3842765889098940E+00_dp, &
         1.8118043634237178E+01_dp, 1.9851466755132208E+01_dp, -4.9221535131083812E+01_dp, &
         3.9121944733752876E+01_dp, -1.5053605862917172E+01_dp, 2.3820642460959474E+00_dp, &
         3.6938617681120809E+02_dp, -3.1008390641542189E+02_dp, -3.1908899493735571E+01_dp, &
         1.4562271786149682E+02_dp, -1.2559633861124211E+00_dp, -7.4923876238492468E+01_dp, &
         -6.9384384355100055E+00_dp, 4.9653981256260799E+01_dp, 3.5331302879921083E+01_dp, &
         -1.1918108482817887E+02_dp, 1.0467883559485558E+02_dp, -4.3551657522660385E+01_dp, &
         7.4318254787716498E+00_dp, 7.5178053722265770E+02_dp, -7.3185923065526492E+02_dp, &
         -1.6464637400435404E+01_dp, 3.6485923276725026E+02_dp, -5.1587363386350894E+01_dp, &
         -1.9111170221147160E+02_dp, 1.5490715798392426E+01_dp, 1.3978292003195943E+02_dp, &
         3.8110337744866719E+01_dp, -2.7620328681571948E+02_dp, 2.7537298325765300E+02_dp, &
         -1.2425623890169337E+02_dp, 2.2689284891309541E+01_dp, 1.5916154447623735E+03_dp, &
         -1.8003650102566801E+03_dp, 1.3061885225064071E+02_dp, 9.3513854431018569E+02_dp, &
         -2.8305798212942278E+02_dp, -4.8432743152230262E+02_dp, 1.4241629292606842E+02_dp, &
         4.0370510754094880E+02_dp, -9.9685940038875515E+01_dp, -5.6492023569870310E+02_dp, &
         6.8962159878288605E+02_dp, -3.4344985934760621E+02_dp, 6.7268170706916536E+01_dp, &
         3.5313329066420033E+03_dp, -4.6623582136066398E+03_dp, 8.7352686043406538E+02_dp, &
         2.4489787138449915E+03_dp, -1.2584736353041660E+03_dp, -1.1810855855446016E+03_dp, &
         7.1672145189107039E+02_dp, 1.1795141993550299E+03_dp, -1.0569593213292389E+03_dp, &
         -7.0128668015997494E+02_dp, 1.4914938794172158E+03_dp, -8.7124256006302119E+02_dp, &
         1.8684842614746950E+02_dp, 8.2940863893933129E+03_dp, -1.2880093011243731E+04_dp, &
         4.2495122306124758E+03_dp, 6.4758081698510341E+03_dp, -5.3383669126949280E+03_dp, &
         -2.4685864914850517E+03_dp, 3.1413524828014874E+03_dp, 3.3033312228152809E+03_dp, &
         -6.0335109923486816E+03_dp, 2.1284718544868892E+03_dp, 1.6537769702474902E+03_dp, &
         -1.6613883929668332E+03_dp, 4.2939992860686181E+02_dp, 2.0920103579440787E+04_dp, &
         -3.8669758427570443E+04_dp, 1.9866909734865130E+04_dp, 1.6420139899251604E+04_dp, &
         -2.2915184173335136E+04_dp, -1.5997192354548479E+03_dp, 1.2941589682976841E+04_dp, &
         7.0023617449621897E+03_dp, -2.8057100680185933E+04_dp, 2.5033445993726258E+04_dp, &
         -9.5663441270494550E+03_dp, 7.6757608917678635E+02_dp, 3.3218276502429876E+02_dp, &
         5.7934335499806577E+04_dp, -1.2968557309336646E+05_dp, 9.9056954372674198E+04_dp, &
         2.9476068192368246E+04_dp, -1.0138652618742791E+05_dp, 3.6368515099655910E+04_dp, &
         4.6073637136932237E+04_dp, -1.0492843328428793E+04_dp, -9.7848110148587089E+04_dp, &
         1.4355967431747538E+05_dp, -9.7281253297314208E+04_dp, 3.4312801274826343E+04_dp, &
         -5.1027940556969679E+03_dp, 1.8382562965656386E+05_dp, -5.0724133133864752E+05_dp, &
         5.7124379095245234E+05_dp, -1.3430281607756464E+05_dp, -4.1104787472683209E+05_dp, &
         4.5698302097076108E+05_dp, -1.4772171067732785E+04_dp, -2.9013984608709166E+05_dp, &
         9.7841936518310904E+04_dp, 2.5630153474370242E+05_dp, -3.3461736758467171E+05_dp, &
         1.7210001134766926E+05_dp, -3.4775431355638953E+04_dp, 8.0851544458545872E+05_dp, &
         -2.6363345449983794E+06_dp, 4.1361998354174164E+06_dp, -3.5121162276064917E+06_dp, &
         7.0893150895957707E+05_dp, 2.2570973383430163E+06_dp, -3.4154205506155374E+06_dp, &
         2.6954736607133211E+06_dp, -1.4066197350891286E+06_dp, 5.4951201324373030E+05_dp, &
         -1.9512919132011509E+05_dp, 6.3990065228808700E+04_dp, -1.1942735238243258E+04_dp, &
         2.4099562266117897E-01_dp, -7.2049876247263943E-03_dp, -6.4116087584877362E-03_dp, &
         2.3477247701907569E-03_dp, -2.4131694227638371E-03_dp, 4.0371201990794785E-03_dp, &
         -4.8026052577681737E-03_dp, 4.9757874425969597E-03_dp, -4.9284265776128131E-03_dp, &
         4.2028583295808742E-03_dp, -2.6617139437737393E-03_dp, 1.0721512574701931E-03_dp, &
         -2.0360724244160395E-04_dp, 7.8600093766184087E-01_dp, -2.7552722482330382E-02_dp, &
         -2.4320818015224920E-02_dp, 9.3274629486534248E-03_dp, -9.2079352354389778E-03_dp, &
         1.5452692636155842E-02_dp, -1.8450575101340298E-02_dp, 1.9063757888167280E-02_dp, &
         -1.8832830863784338E-02_dp, 1.6057085510439863E-02_dp, -1.0174585816282621E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c06 = (/4.0992188628857576E-03_dp, &
         -7.7827564716362676E-04_dp, 1.5427835617899610E+00_dp, -6.9505394993586406E-02_dp, &
         -6.0594885888198051E-02_dp, 2.4860739271249380E-02_dp, -2.3168221093078572E-02_dp, &
         3.9027511015698728E-02_dp, -4.6867833756562442E-02_dp, 4.8230445510777237E-02_dp, &
         -4.7451819606639183E-02_dp, 4.0444324071817472E-02_dp, -2.5649820333915784E-02_dp, &
         1.0337965959194499E-02_dp, -1.9621474197091537E-03_dp, 2.7361460521635825E+00_dp, &
         -1.6210915733697953E-01_dp, -1.3938372870854468E-01_dp, 6.1370009081324846E-02_dp, &
         -5.3953321352611480E-02_dp, 9.1109307724040633E-02_dp, -1.1013676364888772E-01_dp, &
         1.1285815532268084E-01_dp, -1.1051732784079558E-01_dp, 9.4154657670159986E-02_dp, &
         -5.9775221985383778E-02_dp, 2.4105091177407285E-02_dp, -4.5740232709998250E-03_dp, &
         4.7504664477072316E+00_dp, -3.6540640732652457E-01_dp, -3.0963304480188580E-01_dp, &
         1.4617823006458483E-01_dp, -1.2163472991350244E-01_dp, 2.0546897879778972E-01_dp, &
         -2.5017072872400081E-01_dp, 2.5528811465811574E-01_dp, -2.4871800987649034E-01_dp, &
         2.1177224606033893E-01_dp, -1.3460697362966942E-01_dp, 5.4321881683681905E-02_dp, &
         -1.0306557157615706E-02_dp, 8.2846134626133292E+00_dp, -8.0838138618414457E-01_dp, &
         -6.7434278259781932E-01_dp, 3.4154350399965178E-01_dp, -2.6969855520786901E-01_dp, &
         4.5455382056357474E-01_dp, -5.5786575129632443E-01_dp, 5.6695864745421198E-01_dp, &
         -5.4922218190823124E-01_dp, 4.6727981986625922E-01_dp, -2.9742716810172054E-01_dp, &
         1.2014838410171294E-01_dp, -2.2797188357298853E-02_dp, 1.4659069700977295E+01_dp, &
         -1.7747757205152632E+00_dp, -1.4550424144280802E+00_dp, 7.9267096661767655E-01_dp, &
         -5.9501413092507949E-01_dp, 9.9738596204312624E-01_dp, -1.2351218666597548E+00_dp, &
         1.2502002535211376E+00_dp, -1.2031718276241572E+00_dp, 1.0225830615341298E+00_dp, &
         -6.5195914949791578E-01_dp, 2.6371662619257846E-01_dp, -5.0052760660977261E-02_dp, &
         2.6448156436321245E+01_dp, -3.9033379057570827E+00_dp, -3.1379124640819920E+00_dp, &
         1.8465185787105332E+00_dp, -1.3195424388936023E+00_dp, 2.1908275051266819E+00_dp, &
         -2.7410453141653983E+00_dp, 2.7635561809966092E+00_dp, -2.6392193814105300E+00_dp, &
         2.2397534106652977E+00_dp, -1.4308096005721775E+00_dp, 5.7980482655024335E-01_dp, &
         -1.1011473830562242E-01_dp, 4.8827566212215686E+01_dp, -8.6709145648362504E+00_dp, &
         -6.8145806531242146E+00_dp, 4.3563149107877930E+00_dp, -2.9691106736363246E+00_dp, &
         4.8564608241711680E+00_dp, -6.1483109774475455E+00_dp, 6.1755130247165688E+00_dp, &
         -5.8438669005291510E+00_dp, 4.9487151155432789E+00_dp, -3.1688647001286903E+00_dp, &
         1.2872479155148069E+00_dp, -2.4474367721377402E-01_dp, 9.2556502815334994E+01_dp, &
         -1.9603282099708711E+01_dp, -1.5002883382810225E+01_dp, 1.0494569374376898E+01_dp, &
         -6.8410742350968299E+00_dp, 1.0944746242954064E+01_dp, -1.4045586508307052E+01_dp, &
         1.4060381579443410E+01_dp, -1.3158840677577247E+01_dp, 1.1107899993760929E+01_dp, &
         -7.1328849625134936E+00_dp, 2.9070868330198718E+00_dp, -5.5374097833048719E-01_dp, &
         1.8086387815603692E+02_dp, -4.5449136805946708E+01_dp, -3.3697127799416755E+01_dp, &
         2.6030360696858182E+01_dp, -1.6300016117567488E+01_dp, 2.5261110581327873E+01_dp, &
         -3.2926409290629678E+01_dp, 3.2875420065927933E+01_dp, -3.0353354761618480E+01_dp, &
         2.5499432007902808E+01_dp, -1.6427830450896202E+01_dp, 6.7254398764218735E+00_dp, &
         -1.2848018257579334E+00_dp, 3.6615307455742288E+02_dp, -1.0895580944560962E+02_dp, &
         -7.7707894068596588E+01_dp, 6.7083059154405120E+01_dp, -4.0625930911416084E+01_dp, &
         6.0191592289485307E+01_dp, -7.9850740811585609E+01_dp, 7.9626930467390395E+01_dp, &
         -7.2291475212265667E+01_dp, 6.0275016793375315E+01_dp, -3.8970639832754756E+01_dp, &
         1.6051798367917286E+01_dp, -3.0804040922174489E+00_dp, 7.7295926933662383E+02_dp, &
         -2.7274630136859054E+02_dp, -1.8524424755998632E+02_dp, 1.8158011767619479E+02_dp, &
         -1.0746016288132108E+02_dp, 1.4949540595499082E+02_dp, -2.0220074697334576E+02_dp, &
         2.0184753824607512E+02_dp, -1.7940902459359080E+02_dp, 1.4776377024332152E+02_dp, &
         -9.5855229939181896E+01_dp, 3.9808239583015897E+01_dp, -7.6927802615479832E+00_dp, &
         1.7164950125258283E+03_dp, -7.2191533002872904E+02_dp, -4.5988985220715472E+02_dp, &
         5.2344948407025970E+02_dp, -3.0757874260869397E+02_dp, 3.9198001571233146E+02_dp, &
         -5.4073857326588131E+02_dp, 5.4249811741022972E+02_dp, -4.6932848633916359E+02_dp, &
         3.7853944596752183E+02_dp, -2.4590148935655793E+02_dp, 1.0323206520266561E+02_dp, &
         -2.0162118950386379E+01_dp, 4.0597984203925230E+03_dp, -2.0554408966225374E+03_dp, &
         -1.1981956551284402E+03_dp, 1.6379438988785680E+03_dp, -9.7944271492137625E+02_dp, &
         1.1058705957255131E+03_dp, -1.5495244988136308E+03_dp, 1.5727925045092773E+03_dp, &
         -1.3143543851594472E+03_dp, 1.0206697674392115E+03_dp, -6.5915665050050154E+02_dp, &
         2.8033098204930724E+02_dp, -5.5627786985458201E+01_dp, 1.0418737555296313E+04_dp, &
         -6.4587705877741246E+03_dp, -3.2930962522435352E+03_dp, 5.7217121047356122E+03_dp, &
         -3.6200039792272896E+03_dp, 3.4703091404762904E+03_dp, -4.8499754392515288E+03_dp, &
         5.0352943316006931E+03_dp, -4.0300238041629141E+03_dp, 2.9069303818723929E+03_dp, &
         -1.8196316384987094E+03_dp, 7.8120202146243457E+02_dp, -1.5851065777637234E+02_dp, &
         2.9904840276131377E+04_dp, -2.3365682875309405E+04_dp, -9.4213557234108248E+03_dp, &
         2.3341037180846441E+04_dp, -1.6664545324000497E+04_dp, 1.3018178357102739E+04_dp, &
         -1.7013313819623374E+04_dp, 1.8330955289497884E+04_dp, -1.4004140465670787E+04_dp, &
         8.6573316238168354E+03_dp, -4.7417789173697429E+03_dp, 1.9635793580241952E+03_dp, &
         -4.0643170846761330E+02_dp, 1.0157948762160608E+05_dp, -1.0548383390007413E+05_dp, &
         -2.3722704012505274E+04_dp, 1.2044737338200401E+05_dp, -1.0942907243642975E+05_dp, &
         7.1576985837329717E+04_dp, -6.9634226489907858E+04_dp, 7.5769005808596703E+04_dp, &
         -5.6409114203720761E+04_dp, 2.4902335273688092E+04_dp, -5.3550087892323263E+03_dp, &
         2.8390167115008719E+01_dp, 1.4643329837669990E+02_dp, 4.6857889664810052E+05_dp, &
         -7.2038307415608899E+05_dp, 1.2509311324741252E+05_dp, 8.8091622462875850E+05_dp, &
         -1.4127842818682543E+06_dp, 1.1880475416018902E+06_dp, -5.8502506049625343E+05_dp, &
         2.4743808750507604E+04_dp, 3.3699929860959132E+05_dp, -4.7117288930957159E+05_dp, &
         3.8186109459051176E+05_dp, -1.8185745287369000E+05_dp, 3.9096271590969533E+04_dp, &
         1.9685202318288957E+00_dp, -6.1360979308742736E-02_dp, -5.4488336108174731E-02_dp, &
         2.0199240738539220E-02_dp, -2.0538828062587423E-02_dp, 3.4390780667720153E-02_dp, &
         -4.0950900871404537E-02_dp, 4.2396877890840033E-02_dp, -4.1964658244183287E-02_dp, &
         3.5784807205071087E-02_dp, -2.2666021560142281E-02_dp, 9.1304409682086869E-03_dp, &
         -1.7338039826641876E-03_dp, 2.4873431607352861E+00_dp, -1.1127896744503744E-01_dp, &
         -9.7160647648538795E-02_dp, 3.9549196995538598E-02_dp, -3.7088461256170678E-02_dp, &
         6.2482162326251998E-02_dp, -7.4973797450662821E-02_dp, 7.7187912043400322E-02_dp, &
         -7.5984338726898373E-02_dp, 6.4767341178840990E-02_dp, -4.1070002288843978E-02_dp, &
         1.6551536662038395E-02_dp, -3.1415083350615496E-03_dp, 3.7115819235040099E+00_dp, &
         -2.4263735069592671E-01_dp, -2.0828276881828411E-01_dp, 9.2474300956586888E-02_dp, &
         -8.0673281317597170E-02_dp, 1.3641415406027710E-01_dp, -1.6499696868142674E-01_dp, &
         1.6897498706163439E-01_dp, -1.6540000447304345E-01_dp, 1.4091102128378122E-01_dp, &
         -8.9465194191119327E-02_dp, 3.6077418740267916E-02_dp, -6.8451878020381164E-03_dp, &
         6.0819380402782102E+00_dp, -5.3592127980283966E-01_dp, -4.5287536801814943E-01_dp, &
         2.1661663264454648E-01_dp, -1.7820374811642009E-01_dp, 3.0146509927127713E-01_dp, &
         -3.6745474931004735E-01_dp, 3.7463188233201217E-01_dp, -3.6470166760065986E-01_dp, &
         3.1051645875911826E-01_dp, -1.9740158338751423E-01_dp, 7.9665597214015602E-02_dp, &
         -1.5113366050427108E-02_dp, 1.0479537652915743E+01_dp, -1.1712760954555530E+00_dp, &
         -9.7387580140040209E-01_dp, 5.0050202014332057E-01_dp, -3.9044301463179548E-01_dp, &
         6.5882177079888460E-01_dp, -8.0968653097802212E-01_dp, 8.2205031893424119E-01_dp, &
         -7.9554850265578514E-01_dp, 6.7681690340488077E-01_dp, -4.3089093066787693E-01_dp, &
         1.7407475175520742E-01_dp, -3.3025834490739599E-02_dp, 1.8635108288599536E+01_dp, &
         -2.5497121657871862E+00_dp, -2.0826834920345156E+00_dp, 1.1522254517634021E+00_dp, &
         -8.5439272544641709E-01_dp, 1.4332103583869653E+00_dp, -1.7777542118465444E+00_dp, &
         1.7975307886658498E+00_dp, -1.7279183793446284E+00_dp, 1.4684523252875816E+00_dp, &
         -9.3648068065724421E-01_dp, 3.7885018835420770E-01_dp, -7.1898509086654641E-02_dp, &
         3.3994590239686019E+01_dp, -5.5786535866163760E+00_dp, -4.4666506461735587E+00_dp, &
         2.6703461173069210E+00_dp, -1.8856235312128826E+00_dp, 3.1315158571185848E+00_dp, &
         -3.9253072081456857E+00_dp, 3.9531937219561639E+00_dp, -3.7704367525106743E+00_dp, &
         3.1994378130940819E+00_dp, -2.0445460353306122E+00_dp, 8.2864620464328365E-01_dp, &
         -1.5736316384893229E-01_dp, 6.3588405510579335E+01_dp, -1.2363461132070416E+01_dp, &
         -9.6753997605292525E+00_dp, 6.2824458697670735E+00_dp, -4.2341320945537095E+00_dp, &
         6.9247378274318594E+00_dp, -8.7844588705872138E+00_dp, 8.8136469170395682E+00_dp, &
         -8.3287909317576716E+00_dp, 7.0522992389582715E+00_dp, -4.5175728840847738E+00_dp, &
         1.8354907072519964E+00_dp, -3.4896177050152971E-01_dp, 1.2220829809176205E+02_dp, &
         -2.7944214858134188E+01_dp, -2.1296201547453109E+01_dp, 1.5115164278434872E+01_dp, &
         -9.7538219914620825E+00_dp, 1.5600604221804334E+01_dp, -2.0061710148318788E+01_dp, &
         2.0061333896341090E+01_dp, -1.8749429601270812E+01_dp, 1.5826299380047390E+01_dp, &
         -1.0167002373267730E+01_dp, 4.1445597664016303E+00_dp, -7.8940481877510260E-01_dp, &
         2.4212494294883325E+02_dp, -6.4848567095130747E+01_dp, -4.7896931247182728E+01_dp, &
         3.7461397357650824E+01_dp, -2.3251790808727218E+01_dp, 3.6039980420712112E+01_dp, &
         -4.7067767611689661E+01_dp, 4.6945415655437806E+01_dp, -4.3293284411662249E+01_dp, &
         3.6374015775739899E+01_dp, -2.3444250992847859E+01_dp, 9.5995302583925408E+00_dp, &
         -1.8336252417158323E+00_dp, 4.9678277300087296E+02_dp, -1.5564477617682064E+02_dp, &
         -1.1069764568693903E+02_dp, 9.6407239682845059E+01_dp, -5.7938205328089253E+01_dp, &
         8.5979370733647627E+01_dp, -1.1425483537959983E+02_dp, 1.1380794672668260E+02_dp, &
         -1.0325159247360516E+02_dp, 8.6134415628011325E+01_dp, -5.5718825689552844E+01_dp, &
         2.2951002054178900E+01_dp, -4.4030493121892027E+00_dp, 1.0618466971760956E+03_dp, &
         -3.8969710352881111E+02_dp, -2.6450171371001460E+02_dp, 2.6007018186143932E+02_dp, &
         -1.5279188655132614E+02_dp, 2.1360891249135918E+02_dp, -2.8931621709083004E+02_dp, &
         2.8840853173138242E+02_dp, -2.5640283554040690E+02_dp, 2.1150347347655622E+02_dp, &
         -1.3730820428083371E+02_dp, 5.7010139200952608E+01_dp, -1.1009287497802793E+01_dp, &
         2.3831525849788013E+03_dp, -1.0288984124938795E+03_dp, -6.5778574392811515E+02_dp, &
         7.4418907067655823E+02_dp, -4.3344091595675718E+02_dp, 5.5871797886676870E+02_dp, &
         -7.7182037553705447E+02_dp, 7.7263905710645179E+02_dp, -6.6963947277791476E+02_dp, &
         5.4224742270029606E+02_dp, -3.5283793388597093E+02_dp, 1.4803941533425319E+02_dp, &
         -2.8871200173717554E+01_dp, 5.6772385530114016E+03_dp, -2.9062895360890998E+03_dp, &
         -1.7148517487548181E+03_dp, 2.2947989851698503E+03_dp, -1.3525758366401831E+03_dp, &
         1.5626038948903565E+03_dp, -2.1964155478564353E+03_dp, 2.2203872701695277E+03_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c07 = (/-1.8635835707916260E+03_dp, &
         1.4615094004676805E+03_dp, -9.4860776301801513E+02_dp, 4.0324630910185732E+02_dp, &
         -7.9802696400978931E+01_dp, 1.4580340019039370E+04_dp, -8.9648221759975659E+03_dp, &
         -4.7160531127075310E+03_dp, 7.7970756952852253E+03_dp, -4.7953125668139064E+03_dp, &
         4.7904213388018679E+03_dp, -6.7759341874801794E+03_dp, 6.9800960085149118E+03_dp, &
         -5.6283983143506020E+03_dp, 4.1614628365758181E+03_dp, -2.6520659708846315E+03_dp, &
         1.1425936294797418E+03_dp, -2.3097558189199052E+02_dp, 4.1339464053298638E+04_dp, &
         -3.1163959542726632E+04_dp, -1.3657950683331510E+04_dp, 3.0183606937467408E+04_dp, &
         -2.0297598820224881E+04_dp, 1.6867253952668740E+04_dp, -2.3145329533718246E+04_dp, &
         2.4629319790813181E+04_dp, -1.8956801723961318E+04_dp, 1.2473061205766879E+04_dp, &
         -7.3797593844846533E+03_dp, 3.1721822281891500E+03_dp, -6.5940325840740331E+02_dp, &
         1.3465183476229242E+05_dp, -1.2878154349397100E+05_dp, -3.9551202896785959E+04_dp, &
         1.4058730137702916E+05_dp, -1.1150513454150937E+05_dp, 7.5934850042442980E+04_dp, &
         -9.0027245525794060E+04_dp, 1.0007380677391199E+05_dp, -7.4106008002716495E+04_dp, &
         3.8168389198024262E+04_dp, -1.5926680674690388E+04_dp, 5.6819996226797621E+03_dp, &
         -1.1787084557648736E+03_dp, 5.4426445660263638E+05_dp, -7.0252376902961778E+05_dp, &
         -5.5472023615715923E+04_dp, 8.6719603863591724E+05_dp, -9.3672058996953757E+05_dp, &
         5.8718722970563383E+05_dp, -4.2246478920076904E+05_dp, 4.3113805307864543E+05_dp, &
         -3.1790388144140469E+05_dp, 8.9294940174578951E+04_dp, 4.9835062579173107E+04_dp, &
         -4.9965645162383262E+04_dp, 1.2952090357486064E+04_dp, 3.3621760396656832E+06_dp, &
         -6.5229123935961705E+06_dp, 3.0062160391263529E+06_dp, 7.4531686564585082E+06_dp, &
         -1.7301829485568821E+07_dp, 1.8484963345212858E+07_dp, -9.5516135467152409E+06_dp, &
         -4.0021370551100709E+06_dp, 1.4467922484597271E+07_dp, -1.6960571811769929E+07_dp, &
         1.2138850561498789E+07_dp, -5.2211586557498844E+06_dp, 1.0381744850822145E+06_dp, &
         2.4257259348117230E-01_dp, -1.5851429704634029E-09_dp, -9.5774069407972163E-03_dp, &
         1.2300098639931586E-02_dp, -1.6677557521014314E-02_dp, 1.5981292332277006E-02_dp, &
         -2.8119788057016178E-04_dp, -4.1411633666622608E-02_dp, 1.0826576167260124E-01_dp, &
         -1.6888673263610837E-01_dp, 1.7154817499141597E-01_dp, -1.0466126191680286E-01_dp, &
         2.9111302266020433E-02_dp, 7.9204097116662908E-01_dp, -6.0818994537216391E-09_dp, &
         -3.6740838833795079E-02_dp, 4.7185619635050413E-02_dp, -6.3625273088436185E-02_dp, &
         6.0397868728266732E-02_dp, 7.5718647217901276E-04_dp, -1.6182361088743230E-01_dp, &
         4.1907586547824699E-01_dp, -6.5145319082734277E-01_dp, 6.6046883780165377E-01_dp, &
         -4.0246846324937580E-01_dp, 1.1185374791833239E-01_dp, 1.5580568229128680E+00_dp, &
         -1.5419744030696444E-08_dp, -9.3127809560014999E-02_dp, 1.1960239960607019E-01_dp, &
         -1.5991695786356333E-01_dp, 1.4960080232162254E-01_dp, 8.9510077773447185E-03_dp, &
         -4.2148836269473983E-01_dp, 1.0765070072706033E+00_dp, -1.6647789307570040E+00_dp, &
         1.6830424466049698E+00_dp, -1.0237444784742269E+00_dp, 2.8416557439803364E-01_dp, &
         2.7718621061429261E+00_dp, -3.6163513981536213E-08_dp, -2.1834886480726926E-01_dp, &
         2.8042152161200956E-01_dp, -3.7145151138366161E-01_dp, 3.4176124001058317E-01_dp, &
         3.9044747975845936E-02_dp, -1.0171603426069817E+00_dp, 2.5602863098988093E+00_dp, &
         -3.9374011756654790E+00_dp, 3.9683779293500430E+00_dp, -2.4091012786988979E+00_dp, &
         6.6779737409364437E-01_dp, 4.8311936966109696E+00_dp, -8.1986505697328631E-08_dp, &
         -4.9486997262600380E-01_dp, 6.3555244877383477E-01_dp, -8.3363638388937433E-01_dp, &
         7.5337822797117326E-01_dp, 1.3086191484801463E-01_dp, -2.3728506879221536E+00_dp, &
         5.8867883672658534E+00_dp, -9.0020738404552869E+00_dp, 9.0443088573663957E+00_dp, &
         -5.4794215122691758E+00_dp, 1.5167430446672299E+00_dp, 8.4637235711253460E+00_dp, &
         -1.8249091149929743E-07_dp, -1.1011450973952845E+00_dp, 1.4141801024460730E+00_dp, &
         -1.8355159334262476E+00_dp, 1.6263138799610943E+00_dp, 3.9073092232731099E-01_dp, &
         -5.4376260018676819E+00_dp, 1.3293600447123115E+01_dp, -2.0209696632944365E+01_dp, &
         2.0237473247415195E+01_dp, -1.2234446749358080E+01_dp, 3.3815255705843770E+00_dp, &
         1.5053546138154489E+01_dp, -4.0334165130598051E-07_dp, -2.4328249179107155E+00_dp, &
         3.1244306626860210E+00_dp, -4.0084799014945292E+00_dp, 3.4724453638380903E+00_dp, &
         1.1020158511749709E+00_dp, -1.2389534810286600E+01_dp, 2.9830068384681105E+01_dp, &
         -4.5066858817381963E+01_dp, 4.4968188905675269E+01_dp, -2.7122037857043569E+01_dp, &
         7.4841528688448022E+00_dp, 2.7318806501465819E+01_dp, -8.9373103139548194E-07_dp, &
         -5.3883015584663889E+00_dp, 6.9200904347821890E+00_dp, -8.7625163748632833E+00_dp, &
         7.3930201736202728E+00_dp, 3.0266739418528634E+00_dp, -2.8356377784336452E+01_dp, &
         6.7176527817712639E+01_dp, -1.0080221398959092E+02_dp, 1.0018754691787824E+02_dp, &
         -6.0271077807630199E+01_dp, 1.6601208530464657E+01_dp, 5.0769359622143128E+01_dp, &
         -2.0022110266338276E-06_dp, -1.2064973865474135E+01_dp, 1.5494803546470303E+01_dp, &
         -1.9327179050582775E+01_dp, 1.5798630009622276E+01_dp, 8.2524723662168071E+00_dp, &
         -6.5779344719277219E+01_dp, 1.5314732669567439E+02_dp, -2.2809101477129849E+02_dp, &
         2.2570758430280716E+02_dp, -1.3538681440746080E+02_dp, 3.7214435740581607E+01_dp, &
         9.6966640947802489E+01_dp, -4.5708702824465076E-06_dp, -2.7525782848920322E+01_dp, &
         3.5350791499997797E+01_dp, -4.3326681349790739E+01_dp, 3.4065698277940633E+01_dp, &
         2.2663529783272740E+01_dp, -1.5595875787184744E+02_dp, 3.5632753228509131E+02_dp, &
         -5.2627723474899005E+02_dp, 5.1819280410896465E+02_dp, -3.0979403001144362E+02_dp, &
         8.4952402372718240E+01_dp, 1.9114310962046079E+02_dp, -1.0718381310198472E-05_dp, &
         -6.4495689652753413E+01_dp, 8.2830422426384970E+01_dp, -9.9422424907189324E+01_dp, &
         7.4415368027743341E+01_dp, 6.3487036604073083E+01_dp, -3.8117146994730530E+02_dp, &
         8.5310336400357869E+02_dp, -1.2481132204175772E+03_dp, 1.2219161469975245E+03_dp, &
         -7.2767253907573388E+02_dp, 1.9898804514956225E+02_dp, 3.9095152547016971E+02_dp, &
         -2.6044693436334060E-05_dp, -1.5656455100212540E+02_dp, 2.0107231298299899E+02_dp, &
         -2.3531086163800461E+02_dp, 1.6507461742953816E+02_dp, 1.8371250260285103E+02_dp, &
         -9.6951867862351389E+02_dp, 2.1208296289805753E+03_dp, -3.0692048830008334E+03_dp, &
         2.9846327006006663E+03_dp, -1.7692097125381856E+03_dp, 4.8218836520790217E+02_dp, &
         8.3551256508341919E+02_dp, -6.6276418699394831E-05_dp, -3.9790489467747528E+02_dp, &
         5.1101973017952952E+02_dp, -5.7941288348416379E+02_dp, 3.7150601573538091E+02_dp, &
         5.5698245764046999E+02_dp, -2.5960834505545481E+03_dp, 5.5345357608728345E+03_dp, &
         -7.9075149042235771E+03_dp, 7.6276140487209941E+03_dp, -4.4959994932460531E+03_dp, &
         1.2203067212487733E+03_dp, 1.8836460432374965E+03_dp, -1.7906558419193089E-04_dp, &
         -1.0732156319450071E+03_dp, 1.3783032272117819E+03_dp, -1.5000340075540309E+03_dp, &
         8.4018168997033501E+02_dp, 1.8007044978323806E+03_dp, -7.4288132647801831E+03_dp, &
         1.5374600294490150E+04_dp, -2.1628986973636649E+04_dp, 2.0654257790969808E+04_dp, &
         -1.2087613800492580E+04_dp, 3.2634643687020621E+03_dp, 4.5416034797674138E+03_dp, &
         -5.2365813191155002E-04_dp, -3.1308942252975776E+03_dp, 4.0209190324650995E+03_dp, &
         -4.1383392506453802E+03_dp, 1.8377028026049759E+03_dp, 6.3590629378903286E+03_dp, &
         -2.3198652550819283E+04_dp, 4.6349523517001166E+04_dp, -6.3943547760604444E+04_dp, &
         6.0263738399936170E+04_dp, -3.4932837998398638E+04_dp, 9.3635357641943356E+03_dp, &
         1.1958338781224160E+04_dp, -1.7071768156226256E-03_dp, -1.0169488624467269E+04_dp, &
         1.3060348340403734E+04_dp, -1.2381213086186415E+04_dp, 3.2314378644640055E+03_dp, &
         2.5435849905159415E+04_dp, -8.1602901155135565E+04_dp, 1.5601679506781770E+05_dp, &
         -2.0965853398512429E+05_dp, 1.9396469663853475E+05_dp, -1.1088271281624625E+05_dp, &
         2.9404441290840648E+04_dp, 3.5614977210684003E+04_dp, -6.5265362044827011E-03_dp, &
         -3.8636947008429313E+04_dp, 4.9619945449366554E+04_dp, -4.0999782846154798E+04_dp, &
         -3.3222504922255262E+03_dp, 1.2256002817686876E+05_dp, -3.4080284163623268E+05_dp, &
         6.1352219860694092E+05_dp, -7.9241303014367342E+05_dp, 7.1153559384913277E+05_dp, &
         -3.9732963315665961E+05_dp, 1.0341010372176219E+05_dp, 1.2848400434943594E+05_dp, &
         -3.2448901589420744E-02_dp, -1.8969566877398198E+05_dp, 2.4361640221589044E+05_dp, &
         -1.4921343554708638E+05_dp, -1.5091238856037860E+05_dp, 8.0563254617528804E+05_dp, &
         -1.8647433709473095E+06_dp, 3.0346692605618830E+06_dp, -3.6281356744333007E+06_dp, &
         3.0528225777004999E+06_dp, -1.6122300376973741E+06_dp, 3.9997604947801877E+05_dp, &
         6.7227601896939008E+05_dp, -2.8143215138966765E-01_dp, -1.5867302113995703E+06_dp, &
         2.0376973639408911E+06_dp, -1.1025901634165533E+05_dp, -4.2036295031807078E+06_dp, &
         1.0044743271165637E+07_dp, -1.5282870940461934E+07_dp, 1.6518963178353507E+07_dp, &
         -1.1769749475844450E+07_dp, 4.0871559754661410E+06_dp, 5.6596099490722292E+05_dp, &
         -7.4109536787231569E+05_dp, 1.9819560025056868E+00_dp, -1.3511569339814371E-08_dp, &
         -8.1633253443258433E-02_dp, 1.0484017653035560E-01_dp, -1.4194524209112128E-01_dp, &
         1.3568557553371516E-01_dp, -1.3234983778788343E-03_dp, -3.5470515598290253E-01_dp, &
         9.2500068772695143E-01_dp, -1.4416049296037063E+00_dp, 1.4635924511712637E+00_dp, &
         -8.9265360989689835E-01_dp, 2.4823595537612544E-01_dp, 2.5117886684929931E+00_dp, &
         -2.4671854792314193E-08_dp, -1.4901109047023370E-01_dp, 1.9137231477123434E-01_dp, &
         -2.5614818172008030E-01_dp, 2.4006670513485362E-01_dp, 1.2935867778898439E-02_dp, &
         -6.7220614776726906E-01_dp, 1.7197502043268926E+00_dp, -2.6612289097071167E+00_dp, &
         2.6913697392321345E+00_dp, -1.6374477528376692E+00_dp, 4.5458459181050526E-01_dp, &
         3.7650561907510078E+00_dp, -5.4161605376903836E-08_dp, -3.2700972896614366E-01_dp, &
         4.1997270677737458E-01_dp, -5.5572559086377915E-01_dp, 5.1034976841239854E-01_dp, &
         6.1519345356843586E-02_dp, -1.5283321353925809E+00_dp, 3.8408489596559470E+00_dp, &
         -5.9031588254427607E+00_dp, 5.9476108822547875E+00_dp, -3.6098724362817247E+00_dp, &
         1.0005011402155195E+00_dp, 6.2003958313268743E+00_dp, -1.2037059248503575E-07_dp, &
         -7.2652194441561535E-01_dp, 9.3305879954457793E-01_dp, -1.2217147867466474E+00_dp, &
         1.1004967117720030E+00_dp, 2.0336664708568603E-01_dp, -3.5018460450989166E+00_dp, &
         8.6657256927685911E+00_dp, -1.3238463235718530E+01_dp, 1.3293210387322198E+01_dp, &
         -8.0507159019795775E+00_dp, 2.2279456762186776E+00_dp, 1.0739206251612034E+01_dp, &
         -2.6473797876852818E-07_dp, -1.5973297281490579E+00_dp, 2.0514206663572310E+00_dp, &
         -2.6570672559948862E+00_dp, 2.3448587301741366E+00_dp, 5.9565428178472757E-01_dp, &
         -7.9344395190937238E+00_dp, 1.9342826383683839E+01_dp, -2.9372682545317804E+01_dp, &
         2.9394222808923349E+01_dp, -1.7762737595349108E+01_dp, 4.9080935520946882E+00_dp, &
         1.9202199285653183E+01_dp, -5.8024376773466123E-07_dp, -3.4996092621466643E+00_dp, &
         4.4944812099998321E+00_dp, -5.7527459450885452E+00_dp, 4.9604902525102608E+00_dp, &
         1.6548760224475922E+00_dp, -1.7934154017551453E+01_dp, 4.3051248399276048E+01_dp, &
         -6.4961623842017133E+01_dp, 6.4774151275958900E+01_dp, -3.9049982184732009E+01_dp, &
         1.0772159604979572E+01_dp, 3.5239797671962883E+01_dp, -1.2791831843911305E-06_dp, &
         -7.7116293148741368E+00_dp, 9.9038942846037248E+00_dp, -1.2509090324776968E+01_dp/)
      REAL(KIND=dp), DIMENSION(13, 38, 6), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04, c05, c06, c07, c08/), (/13, 38, 6/))

      INTEGER                                            :: irange

      IF (Rc <= 5.0000000000000000E+01_dp) Rc = 5.0000000000000000E+01_dp
      IF (Rc >= 3.5810871124267578E+06_dp) Rc = 3.5810871124267578E+06_dp
      IF (Rc >= 5.0000000000000000E+01_dp .AND. Rc <= 9.6916963080248536E+01_dp) THEN
         irange = 1
         L_b = 5.0000000000000000E+01_dp
         U_b = 9.6916963080248536E+01_dp
      END IF
      IF (Rc >= 9.6916963080248536E+01_dp .AND. Rc <= 1.2185306936104239E+03_dp) THEN
         irange = 2
         L_b = 9.6916963080248536E+01_dp
         U_b = 1.2185306936104239E+03_dp
      END IF
      IF (Rc >= 1.2185306936104239E+03_dp .AND. Rc <= 1.2217104533148158E+04_dp) THEN
         irange = 3
         L_b = 1.2185306936104239E+03_dp
         U_b = 1.2217104533148158E+04_dp
      END IF
      IF (Rc >= 1.2217104533148158E+04_dp .AND. Rc <= 1.9815072618369377E+05_dp) THEN
         irange = 4
         L_b = 1.2217104533148158E+04_dp
         U_b = 1.9815072618369377E+05_dp
      END IF
      IF (Rc >= 1.9815072618369377E+05_dp .AND. Rc <= 1.2188439903425989E+06_dp) THEN
         irange = 5
         L_b = 1.9815072618369377E+05_dp
         U_b = 1.2188439903425989E+06_dp
      END IF
      IF (Rc >= 1.2188439903425989E+06_dp .AND. Rc <= 3.5810871124267578E+06_dp) THEN
         irange = 6
         L_b = 1.2188439903425989E+06_dp
         U_b = 3.5810871124267578E+06_dp
      END IF

      fit_coef(1:13, 1:38) = coefdata(1:13, 1:38, irange)

   END SUBROUTINE get_fit_coef_k19

! **************************************************************************************************
!> \brief ...
!> \param Rc ...
!> \param L_b ...
!> \param U_b ...
!> \param fit_coef ...
! **************************************************************************************************
   SUBROUTINE get_fit_coef_k20(Rc, L_b, U_b, fit_coef)
      REAL(KIND=dp)                                      :: Rc, L_b, U_b
      REAL(KIND=dp), DIMENSION(13, 40)                   :: fit_coef

      REAL(KIND=dp), DIMENSION(320), PARAMETER :: c08 = (/3.4482542694423209E+03_dp, &
         9.3380638606834109E+03_dp, -3.6141948437832449E+04_dp, 7.3811576243227406E+04_dp, &
         -1.0360614380245237E+05_dp, 9.9273882530972987E+04_dp, -5.8503038150581720E+04_dp, &
         1.5941507970166829E+04_dp, 1.9245609411100908E+04_dp, 1.5758311389454111E-03_dp, &
         -1.6493149055229915E+04_dp, 2.1319315251857988E+04_dp, -2.0594905686666327E+04_dp, &
         6.7450368572538291E+03_dp, 3.7943888089438202E+04_dp, -1.2716219726681517E+05_dp, &
         2.4802524381258758E+05_dp, -3.3897801223706041E+05_dp, 3.1885864066745085E+05_dp, &
         -1.8535737515320248E+05_dp, 4.9984902189495566E+04_dp, 5.7318324293103229E+04_dp, &
         5.9071432526945937E-03_dp, -6.2380841746085105E+04_dp, 8.0634158042088748E+04_dp, &
         -6.8284731483765951E+04_dp, 5.4681930309783763E+02_dp, 1.8545057497723409E+05_dp, &
         -5.3191922255590232E+05_dp, 9.7505194465025037E+05_dp, -1.2805757822087191E+06_dp, &
         1.1696153509061884E+06_dp, -6.6458955725748790E+05_dp, 1.7602465115753218E+05_dp, &
         2.0678064129765515E+05_dp, 2.8344038829274502E-02_dp, -3.0494093965445616E+05_dp, &
         3.9416660234818154E+05_dp, -2.5130146580176344E+05_dp, -2.1177938413183455E+05_dp, &
         1.2359116116152727E+06_dp, -2.9217324689694936E+06_dp, 4.8357925748930285E+06_dp, &
         -5.8846345861200411E+06_dp, 5.0469092229344351E+06_dp, -2.7194600229616752E+06_dp, &
         6.8869342689229129E+05_dp, 1.0819530962975370E+06_dp, 2.2283944744663364E-01_dp, &
         -2.5391699671295690E+06_dp, 3.2820491767507303E+06_dp, -2.9879597797710146E+05_dp, &
         -6.4273657969231484E+06_dp, 1.5669734066822972E+07_dp, -2.4314998224766523E+07_dp, &
         2.7124297775591992E+07_dp, -2.0552349360272344E+07_dp, 8.5713366254611593E+06_dp, &
         -3.7935610363443429E+05_dp, -8.3928846510169783E+05_dp, 1.9320589271290354E+00_dp, &
         7.2621425609846251E-09_dp, -7.4528731965208886E-02_dp, 9.6338004250833206E-02_dp, &
         -1.3214657018893700E-01_dp, 1.3186456446235820E-01_dp, -2.0166775874314861E-02_dp, &
         -2.8991359670853600E-01_dp, 8.0125762227548336E-01_dp, -1.2825636229353268E+00_dp, &
         1.3287373087728476E+00_dp, -8.2519003619160525E-01_dp, 2.3343121150717866E-01_dp, &
         2.4217999169457665E+00_dp, 1.2934929155472219E-08_dp, -1.3279748613131781E-01_dp, &
         1.7165785375157250E-01_dp, -2.3307585350609009E-01_dp, 2.2877594664050899E-01_dp, &
         -2.3346318222297352E-02_dp, -5.3715564779172731E-01_dp, 1.4543022329299131E+00_dp, &
         -2.3114184688427013E+00_dp, 2.3856880646661343E+00_dp, -1.4781477166371391E+00_dp, &
         4.1748286135655011E-01_dp, 3.5699082894477758E+00_dp, 2.7703387762280234E-08_dp, &
         -2.8453205801269615E-01_dp, 3.6779425246710412E-01_dp, -4.9428962102081125E-01_dp, &
         4.7696326068383171E-01_dp, -2.3216522248151596E-02_dp, -1.1945691723631664E+00_dp, &
         3.1721266716571264E+00_dp, -5.0071605495889964E+00_dp, 5.1491452709951524E+00_dp, &
         -3.1830440507795661E+00_dp, 8.9760588696477739E-01_dp, 5.7714745756709691E+00_dp, &
         6.0203678580378609E-08_dp, -6.1856466138001853E-01_dp, 7.9957416304825757E-01_dp, &
         -1.0645544695648401E+00_dp, 1.0109461180627741E+00_dp, 1.9498075311132036E-03_dp, &
         -2.6818798561564088E+00_dp, 7.0044906403410883E+00_dp, -1.0989948493463517E+01_dp, &
         1.1264809560038588E+01_dp, -6.9492617232208316E+00_dp, 1.9569143430774056E+00_dp, &
         9.8063851627247747E+00_dp, 1.2943227446405601E-07_dp, -1.3303860307558404E+00_dp, &
         1.7196942244173261E+00_dp, -2.2679112076886052E+00_dp, 2.1180849214493009E+00_dp, &
         1.1718272413659171E-01_dp, -5.9500812130151042E+00_dp, 1.5295418487216114E+01_dp, &
         -2.3856483405279011E+01_dp, 2.4374146939476983E+01_dp, -1.5005553650474871E+01_dp, &
         4.2196153451488474E+00_dp, 1.7182196524894046E+01_dp, 2.7681980610640121E-07_dp, &
         -2.8465978348393266E+00_dp, 3.6795911913729618E+00_dp, -4.8029054852258346E+00_dp, &
         4.4032195652656236E+00_dp, 5.0834247955654877E-01_dp, -1.3143792363416638E+01_dp, &
         3.3245320187678388E+01_dp, -5.1533317361899485E+01_dp, 5.2471959863900395E+01_dp, &
         -3.2233041672826495E+01_dp, 9.0504077773870328E+00_dp, 3.0845352294831098E+01_dp, &
         5.9398312369643793E-07_dp, -6.1111889591236563E+00_dp, 7.8994899897307205E+00_dp, &
         -1.0193982404223158E+01_dp, 9.1495116492406492E+00_dp, 1.6952456171928449E+00_dp, &
         -2.9178719666384300E+01_dp, 7.2568642728771039E+01_dp, -1.1174619343091653E+02_dp, &
         1.1336163692278953E+02_dp, -6.9471395810743132E+01_dp, 1.9474058838425890E+01_dp, &
         5.6681482844036843E+01_dp, 1.2880602601267124E-06_dp, -1.3260114690148326E+01_dp, &
         1.7140381644716403E+01_dp, -2.1836576138010962E+01_dp, 1.9120563971615155E+01_dp, &
         5.1270149635423383E+00_dp, -6.5601700286248231E+01_dp, 1.6028123406894895E+02_dp, &
         -2.4505157526179127E+02_dp, 2.4758883306634189E+02_dp, -1.5133127650620855E+02_dp, &
         4.2343189674877024E+01_dp, 1.0677930992732443E+02_dp, 2.8406374165609642E-06_dp, &
         -2.9263985515033859E+01_dp, 3.7827402760261627E+01_dp, -4.7491127745716220E+01_dp, &
         4.0381636711719807E+01_dp, 1.4885135448607762E+01_dp, -1.5037491992215962E+02_dp, &
         3.6054446889308713E+02_dp, -5.4693897575387132E+02_dp, 5.5012746515130448E+02_dp, &
         -3.3526230511591672E+02_dp, 9.3615003920193601E+01_dp, 2.0679533410485035E+02_dp, &
         6.4094168244319851E-06_dp, -6.6085152661798745E+01_dp, 8.5423381867854914E+01_dp, &
         -1.0544823391402035E+02_dp, 8.6528860573033953E+01_dp, 4.2714834499989976E+01_dp, &
         -3.5371775303988545E+02_dp, 8.3117344347077687E+02_dp, -1.2500496388367815E+03_dp, &
         1.2510302711576599E+03_dp, -7.5988036054840700E+02_dp, 2.1168378030342967E+02_dp, &
         4.1328664733661651E+02_dp, 1.4889806008040278E-05_dp, -1.5368210432639367E+02_dp, &
         1.9865336911012668E+02_dp, -2.4040314537499890E+02_dp, 1.8872866671662570E+02_dp, &
         1.2351690805087296E+02_dp, -8.5972088745757685E+02_dp, 1.9767740484365552E+03_dp, &
         -2.9445141512794567E+03_dp, 2.9300403857222186E+03_dp, -1.7729344307768993E+03_dp, &
         4.9255725456854503E+02_dp, 8.5654938775191738E+02_dp, 3.5872303016542066E-05_dp, &
         -3.7072601294454631E+02_dp, 4.7920949648375620E+02_dp, -5.6631511928892780E+02_dp, &
         4.1997532374726688E+02_dp, 3.6557905857252490E+02_dp, -2.1763659954041159E+03_dp, &
         4.8870520139000928E+03_dp, -7.2006664507013447E+03_dp, 7.1181616235454258E+03_dp, &
         -4.2879275292267075E+03_dp, 1.1874668147376524E+03_dp, 1.8527695319692295E+03_dp, &
         9.0414525740239963E-05_dp, -9.3593883032510507E+02_dp, 1.2098163913708674E+03_dp, &
         -1.3887279432851076E+03_dp, 9.5392042649204359E+02_dp, 1.1242551714786039E+03_dp, &
         -5.7944326329618743E+03_dp, 1.2675523603724201E+04_dp, -1.8443335807050775E+04_dp, &
         1.8090734173380570E+04_dp, -1.0839637669115842E+04_dp, 2.9902567983122644E+03_dp, &
         4.2185980993912917E+03_dp, 2.4111494993262110E-04_dp, -2.5013688864927140E+03_dp, &
         3.2333243651908087E+03_dp, -3.5775439041628538E+03_dp, 2.2018907529049170E+03_dp, &
         3.6526563185232840E+03_dp, -1.6430818396518105E+04_dp, 3.4901684323665315E+04_dp, &
         -5.0036642596126949E+04_dp, 4.8619455802354423E+04_dp, -2.8940491721336319E+04_dp, &
         7.9451453468075997E+03_dp, 1.0231390945985855E+04_dp, 6.9083287944081352E-04_dp, &
         -7.1882842790158111E+03_dp, 9.2917205293700135E+03_dp, -9.7955483893616456E+03_dp, &
         5.0678948479689298E+03_dp, 1.2800733529560235E+04_dp, -5.0479717751341195E+04_dp, &
         1.0365862765556478E+05_dp, -1.4595280309876031E+05_dp, 1.4014598216542669E+05_dp, &
         -8.2717100120144009E+04_dp, 2.2565914851566984E+04_dp, 2.6891322723431167E+04_dp, &
         2.1742181990913696E-03_dp, -2.2721782303665794E+04_dp, 2.9370569957731357E+04_dp, &
         -2.8945217516976965E+04_dp, 1.0780164053029053E+04_dp, 4.9798530718129528E+04_dp, &
         -1.7223393818390579E+05_dp, 3.3968227753935917E+05_dp, -4.6738665105989081E+05_dp, &
         4.4176278346948413E+05_dp, -2.5773351553389218E+05_dp, 6.9696495544733130E+04_dp, &
         7.8684769327617381E+04_dp, 7.7807434887690492E-03_dp, -8.1870048706807342E+04_dp, &
         1.0582629970733885E+05_dp, -9.4133820403246747E+04_dp, 1.2453504120756526E+04_dp, &
         2.2473369976015511E+05_dp, -6.7807284436999192E+05_dp, 1.2707645472222809E+06_dp, &
         -1.6939265958734713E+06_dp, 1.5647390178375142E+06_dp, -8.9707114806702349E+05_dp, &
         2.3930069996788309E+05_dp, 2.6850437188439124E+05_dp, 3.3617187911264466E-02_dp, &
         -3.5802500077804463E+05_dp, 4.6278519403414475E+05_dp, -3.4336642578337889E+05_dp, &
         -1.2299914820280354E+05_dp, 1.2698214439994770E+06_dp, -3.2863510817135861E+06_dp, &
         5.7313706772967186E+06_dp, -7.2659207395268613E+06_dp, 6.4530123779051183E+06_dp, &
         -3.5837813575776960E+06_dp, 9.3152895714921399E+05_dp, 1.1745893617395842E+06_dp, &
         1.9823345837793316E-01_dp, -2.1674806769945836E+06_dp, 2.8016646898503760E+06_dp, &
         -1.3208943556501460E+06_dp, -2.7152441097245929E+06_dp, 1.0527862934716102E+07_dp, &
         -2.2141959028414574E+07_dp, 3.3833815704067335E+07_dp, -3.8375451355696611E+07_dp, &
         3.0793030373920619E+07_dp, -1.5573075367459131E+07_dp, 3.7151288981291451E+06_dp, &
         8.6165998717888258E+06_dp, 2.2428828403966108E+00_dp, -2.6590263782914072E+07_dp, &
         3.4369174208174527E+07_dp, 1.0603207649981629E+07_dp, -1.0297545596222048E+08_dp, &
         1.9337680822524059E+08_dp, -2.0606965563558266E+08_dp, 8.3351497204227954E+07_dp, &
         1.3400532965008634E+08_dp, -2.7949474181137478E+08_dp, 2.2551932538988832E+08_dp, &
         -7.3358489823807165E+07_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c01 = (/1.4818427186393188E-01_dp, &
         -3.9684125490741781E-02_dp, -4.5544691512272280E-03_dp, 2.2868531356116861E-03_dp, &
         5.4988095348367525E-04_dp, -3.2090841260734175E-04_dp, -9.2010045248866118E-05_dp, &
         6.1214638366653809E-05_dp, 2.2665438652455234E-05_dp, -1.8900573386763588E-05_dp, &
         1.9737341940349761E-06_dp, -4.0098588379752778E-07_dp, 7.9423337565922097E-07_dp, &
         4.5845629764106710E-01_dp, -1.3050008497162252E-01_dp, -1.1746015834640975E-02_dp, &
         7.9238498027192957E-03_dp, 1.3726359502348783E-03_dp, -1.1375480080520228E-03_dp, &
         -2.2424819941956490E-04_dp, 2.1867699685858426E-04_dp, 5.8049724664724386E-05_dp, &
         -6.6759583496292301E-05_dp, 1.1220242809507542E-05_dp, -1.3505737470233449E-06_dp, &
         2.1298269803731967E-06_dp, 8.1205262169756842E-01_dp, -2.5796327574778904E-01_dp, &
         -1.2091050279258431E-02_dp, 1.6773647500350573E-02_dp, 1.1878661490072853E-03_dp, &
         -2.4697243937931864E-03_dp, -1.6605322411568808E-04_dp, 4.7805723848042757E-04_dp, &
         5.6067445750992246E-05_dp, -1.4470898201670074E-04_dp, 3.9438442939321917E-05_dp, &
         -4.1919645541969900E-06_dp, 2.9194474822921137E-06_dp, 1.2433837102757057E+00_dp, &
         -4.5449390505874582E-01_dp, 2.8772391163010618E-03_dp, 3.0979924684101916E-02_dp, &
         -1.2867784281123959E-03_dp, -4.6126614533263735E-03_dp, 3.2856131571824614E-04_dp, &
         8.9223222108096990E-04_dp, -3.3421577643852616E-05_dp, -2.7142959004058064E-04_dp, &
         1.0953992802467019E-04_dp, -1.5370427764270796E-05_dp, 3.6238473019299857E-06_dp, &
         1.7957394158840545E+00_dp, -7.6574403746434661E-01_dp, 4.8537842200104014E-02_dp, &
         5.2360822807135658E-02_dp, -8.3350755683544234E-03_dp, -7.7088356658967239E-03_dp, &
         1.6907326728181175E-03_dp, 1.4727004194089355E-03_dp, -2.9848376058600709E-04_dp, &
         -4.5890070262808070E-04_dp, 2.5968046486014553E-04_dp, -5.0492382323740779E-05_dp, &
         6.4133621780081155E-06_dp, 2.5262136301711431E+00_dp, -1.2588607298899273E+00_dp, &
         1.5253497398909188E-01_dp, 8.1421469717775821E-02_dp, -2.3821436496503780E-02_dp, &
         -1.1489040350698325E-02_dp, 4.6191262169961445E-03_dp, 2.1284229023209771E-03_dp, &
         -8.7863297220430183E-04_dp, -6.9885410634119751E-04_dp, 5.4823184687904317E-04_dp, &
         -1.4084932910802381E-04_dp, 1.6899366991871437E-05_dp, 3.5123659329872625E+00_dp, &
         -2.0346082761913720E+00_dp, 3.6366813808794807E-01_dp, 1.1503446620296379E-01_dp, &
         -5.3909505121735528E-02_dp, -1.4743484990619836E-02_dp, 1.0146826347631005E-02_dp, &
         2.5497745431026582E-03_dp, -1.9698501663441875E-03_dp, -9.3634077876968234E-04_dp, &
         1.0547263092030720E-03_dp, -3.4197789086921887E-04_dp, 4.6897245033077326E-05_dp, &
         4.8615658543578890E+00_dp, -3.2456179761655868E+00_dp, 7.6674797762756552E-01_dp, &
         1.4154002387869766E-01_dp, -1.0779042624053684E-01_dp, -1.4342190696884314E-02_dp, &
         1.9628242867656083E-02_dp, 2.0123316603197143E-03_dp, -3.8036526812309686E-03_dp, &
         -1.0255285079106839E-03_dp, 1.8698313419945043E-03_dp, -7.4237378807489164E-04_dp, &
         1.1820859561659198E-04_dp, 6.7243640488374226E+00_dp, -5.1241379505231510E+00_dp, &
         1.5073836478498175E+00_dp, 1.3090312459656980E-01_dp, -1.9797012483116577E-01_dp, &
         -3.5368923851968148E-03_dp, 3.4498661615170442E-02_dp, -9.2463996990214416E-04_dp, &
         -6.5658463176524523E-03_dp, -6.6485773777846281E-04_dp, 3.0624077548521331E-03_dp, &
         -1.4675840447194452E-03_dp, 2.6687950020770981E-04_dp, 9.3139083034785379E+00_dp, &
         -8.0244497096949736E+00_dp, 2.8332477118834278E+00_dp, 1.5522414229484492E-02_dp, &
         -3.3893468747140731E-01_dp, 3.0800758429289306E-02_dp, 5.5506097333631294E-02_dp, &
         -8.8311029352697995E-03_dp, -1.0194631369398816E-02_dp, 6.8342519451028135E-04_dp, &
         4.6062825095427749E-03_dp, -2.6705893223239008E-03_dp, 5.4643420350779194E-04_dp, &
         1.2934585921249521E+01_dp, -1.2487983193387047E+01_dp, 5.1625830662902139E+00_dp, &
         -3.4641867453590114E-01_dp, -5.4137725689650251E-01_dp, 1.1230929110201276E-01_dp, &
         8.0834531173369778E-02_dp, -2.5867633057602857E-02_dp, -1.3957958298784179E-02_dp, &
         3.8652333980646594E-03_dp, 6.2435692573387415E-03_dp, -4.4931052938362573E-03_dp, &
         1.0271118708257464E-03_dp, 1.8025229894947888E+01_dp, -1.9343860235887981E+01_dp, &
         9.1975266340592636E+00_dp, -1.2397685603690685E+00_dp, -7.9561029845592168E-01_dp, &
         2.8043792713585208E-01_dp, 1.0205012717274836E-01_dp, -5.8031621983963111E-02_dp, &
         -1.5652637896035132E-02_dp, 1.0030483755240904E-02_dp, 7.2529698214355435E-03_dp, &
         -6.9701747701425868E-03_dp, 1.7852314445813721E-03_dp, 2.5226686181538120E+01_dp, &
         -2.9865961999886299E+01_dp, 1.6110802529841504E+01_dp, -3.2194475911241494E+00_dp, &
         -1.0303513955388184E+00_dp, 5.9474554354538078E-01_dp, 9.6101297590719212E-02_dp, &
         -1.1237368378646433E-01_dp, -1.0220072288871951E-02_dp, 2.0372906608218531E-02_dp, &
         6.0853027356999037E-03_dp, -9.8194206242380334E-03_dp, 2.8679427689706485E-03_dp, &
         3.5494031265018705E+01_dp, -4.6025018877931025E+01_dp, 2.7852880135496353E+01_dp, &
         -7.3401774215925366E+00_dp, -1.0186566668613926E+00_dp, 1.1300029991052669E+00_dp, &
         1.1295976109331713E-02_dp, -1.9301508698030923E-01_dp, 1.2241843214785244E-02_dp, &
         3.5142635468207804E-02_dp, -8.2517129158308055E-05_dp, -1.2012456289332549E-02_dp, &
         4.1969205275104700E-03_dp, 5.0301713674828171E+01_dp, -7.0918826017007646E+01_dp, &
         4.7664290809984450E+01_dp, -1.5561816383882222E+01_dp, -1.7868631517805517E-01_dp, &
         1.9389162015454402E+00_dp, -2.5170989059236282E-01_dp, -2.8945054924244118E-01_dp, &
         6.7800129944182816E-02_dp, 5.1001095146721069E-02_dp, -1.5261130863393309E-02_dp, &
         -1.1165054631480001E-02_dp, 5.3713423174483309E-03_dp, 7.2073804790616393E+01_dp, &
         -1.0960262316704710E+02_dp, 8.0985249028988193E+01_dp, -3.1454847999518755E+01_dp, &
         2.8131045955063794E+00_dp, 2.9325467224256498E+00_dp, -8.4832494632348165E-01_dp, &
         -3.5274801045453463E-01_dp, 1.7425095910591704E-01_dp, 5.6535696476405536E-02_dp, &
         -4.2150084007274102E-02_dp, -3.6504731904105233E-03_dp, 5.5123410746420135E-03_dp, &
         1.0525713658114117E+02_dp, -1.7104517278891666E+02_dp, 1.3737253723605420E+02_dp, &
         -6.1507403720685481E+01_dp, 1.0708265591623199E+01_dp, 3.6132806701224829E+00_dp, &
         -1.9428164726167740E+00_dp, -2.7177296350258656E-01_dp, 3.3358187899034614E-01_dp, &
         3.1602468053086578E-02_dp, -7.6640928780817658E-02_dp, 1.2767086056539591E-02_dp, &
         3.7015373482892546E-03_dp, 1.5961638373875255E+02_dp, -2.7407208180371384E+02_dp, &
         2.3597941902719086E+02_dp, -1.1831815835540151E+02_dp, 2.8965862566108601E+01_dp, &
         2.6695849787510491E+00_dp, -3.5799821757574688E+00_dp, 1.0012437250743024E-01_dp, &
         5.1714245220836208E-01_dp, -3.8887168121094023E-02_dp, -1.0903222378111407E-01_dp, &
         3.6184421001653103E-02_dp, 7.4184612058038470E-06_dp, 2.6370713876922872E+02_dp, &
         -4.7201989715682282E+02_dp, 4.2842096437204810E+02_dp, -2.3341869455214848E+02_dp, &
         6.9905383809584009E+01_dp, -2.5086723240417781E+00_dp, -5.7498093848956318E+00_dp, &
         8.7370241169426954E-01_dp, 7.2690318270483367E-01_dp, -1.5862760095796691E-01_dp, &
         -1.4433107481771534E-01_dp, 6.7767335244992954E-02_dp, -5.0600724217742732E-03_dp, &
         5.6003009143569716E+02_dp, -1.0287297849968459E+03_dp, 9.6498150903106421E+02_dp, &
         -5.5317904766641254E+02_dp, 1.8492063710688871E+02_dp, -1.9101040737585297E+01_dp, &
         -1.0231770142026468E+01_dp, 2.5465845134577312E+00_dp, 1.2703101988094840E+00_dp, &
         -4.1883844922802427E-01_dp, -2.4642046076164414E-01_dp, 1.4229171344177202E-01_dp, &
         -1.4875308764204570E-02_dp, 1.1946368463753128E+00_dp, -3.2495379783849715E-01_dp, &
         -3.5228366362451968E-02_dp, 1.9001587589605372E-02_dp, 4.2259462352662765E-03_dp, &
         -2.6842657875635143E-03_dp, -7.0387680924777728E-04_dp, 5.1324943831075869E-04_dp, &
         1.7519190242729584E-04_dp, -1.5790677102250255E-04_dp, 1.9118266393283089E-05_dp, &
         -3.2127263681394069E-06_dp, 6.1720484771177976E-06_dp, 1.3069315340948260E+00_dp, &
         -4.1805297500996297E-01_dp, -1.9174651660622710E-02_dp, 2.7527354291961389E-02_dp, &
         1.8936163303805468E-03_dp, -4.0779517864187736E-03_dp, -2.6637654979470501E-04_dp, &
         7.9164825616038234E-04_dp, 9.0616104454619150E-05_dp, -2.3837680235491997E-04_dp, &
         6.3556401416228639E-05_dp, -5.3380803695105839E-06_dp, 4.2840530293578485E-06_dp, &
         1.5445250149914536E+00_dp, -6.2287628679630180E-01_dp, 2.1843824089781916E-02_dp, &
         4.4718022389577831E-02_dp, -4.1973146173572883E-03_dp, -6.7826557553283739E-03_dp, &
         8.9075466785271920E-04_dp, 1.3203001424771435E-03_dp, -1.3887733251644589E-04_dp, &
         -3.9798804282131669E-04_dp, 1.7705748242976088E-04_dp, -2.1402551771088614E-05_dp, &
         2.2801890600343344E-06_dp, 1.9346618851392008E+00_dp, -9.7926756464428466E-01_dp, &
         1.0781255156226323E-01_dp, 7.0204267153656671E-02_dp, -1.7083979272462264E-02_dp, &
         -1.0472010647795030E-02_dp, 3.3479446665571036E-03_dp, 2.0046454310945101E-03_dp, &
         -6.3478895071837540E-04_dp, -6.2462188170554333E-04_dp, 4.0883505029748853E-04_dp, &
         -7.7995690482165342E-05_dp, 4.6890795156480489E-06_dp, 2.5218110054658807E+00_dp, &
         -1.5547109430512907E+00_dp, 2.7533116442749239E-01_dp, 1.0132320908146478E-01_dp, &
         -4.1829776586315127E-02_dp, -1.4151497493277881E-02_dp, 8.0101836422415695E-03_dp, &
         2.5858991445510790E-03_dp, -1.5782258678436166E-03_dp, -8.7134756801638584E-04_dp, &
         8.3107610724402121E-04_dp, -2.2325382352150771E-04_dp, 2.1184031900921113E-05_dp, &
         3.3730785337688096E+00_dp, -2.4562132009879658E+00_dp, 5.8822468241512738E-01_dp, &
         1.2938256628900288E-01_dp, -8.6117944085599613E-02_dp, -1.5346369501680956E-02_dp, &
         1.6105261132836165E-02_dp, 2.4662787067982008E-03_dp, -3.1947037676917276E-03_dp, &
         -1.0145572939541253E-03_dp, 1.5309117164874644E-03_dp, -5.3600219225621295E-04_dp, &
         7.0324364856116483E-05_dp, 4.5868466733976998E+00_dp, -3.8497189820072570E+00_dp, &
         1.1557884102835358E+00_dp, 1.3236494031740839E-01_dp, -1.6068493389320054E-01_dp, &
         -8.7123214871505979E-03_dp, 2.9032643032267314E-02_dp, 4.4156030667138075E-04_dp, &
         -5.6942421453560228E-03_dp, -7.9246900794011370E-04_dp, 2.5835195275365408E-03_dp, &
         -1.1316665352679273E-03_dp, 1.8389558466057045E-04_dp, 6.3058543134695348E+00_dp, &
         -5.9904367486606711E+00_dp, 2.1632706388868730E+00_dp, 6.0862754524593270E-02_dp, &
         -2.7876465363888409E-01_dp, 1.6240870724100132E-02_dp, 4.7821555177730193E-02_dp, &
         -5.6736395491679926E-03_dp, -9.1125860875598368E-03_dp, 2.7444438602033391E-04_dp, &
         3.9878923020532997E-03_dp, -2.1556997302423510E-03_dp, 4.1038336500161027E-04_dp, &
         8.7367679966561802E+00_dp, -9.2696858881557223E+00_dp, 3.9225237552290722E+00_dp, &
         -1.8871498473285156E-01_dp, -4.5266022613648998E-01_dp, 7.8678436144091410E-02_dp, &
         7.1755861676809696E-02_dp, -1.9502646657472384E-02_dp, -1.2970460818387342E-02_dp, &
         2.9640688478987454E-03_dp, 5.5440927395923583E-03_dp, -3.7503139256693141E-03_dp, &
         8.1466890564677789E-04_dp, 1.2179715928272094E+01_dp, -1.4287247815361980E+01_dp, &
         6.9559238252243123E+00_dp, -8.2452423618630410E-01_dp, -6.8317025405101606E-01_dp, &
         2.1141333388888769E-01_dp, 9.5346291787819212E-02_dp, -4.6499824144226799E-02_dp, &
         -1.5623028966346148E-02_dp, 8.3849673127276043E-03_dp, 6.6500033547118049E-03_dp, &
         -5.9740661993351847E-03_dp, 1.4691884865455198E-03_dp, 1.7073926981017319E+01_dp, &
         -2.1963630226832553E+01_dp, 1.2134829552891802E+01_dp, -2.2528424944623944E+00_dp, &
         -9.3240764730384440E-01_dp, 4.6660358949448766E-01_dp, 1.0216844176487623E-01_dp, &
         -9.3824494827588173E-02_dp, -1.3105510915164605E-02_dp, 1.7849498315239853E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c02 = (/5.9915911031709962E-03_dp, &
         -8.6423694074923528E-03_dp, 2.4291300518560965E-03_dp, 2.4071366084770816E+01_dp, &
         -3.3716515170389293E+01_dp, 2.0906788438349480E+01_dp, -5.2492585742829574E+00_dp, &
         -1.0597221453010373E+00_dp, 9.1826115756590143E-01_dp, 5.3116568030487622E-02_dp, &
         -1.6840980447490020E-01_dp, 2.7804548003092637E-03_dp, 3.2324581009718446E-02_dp, &
         1.0845459715807160E-03_dp, -1.1013651706859648E-02_dp, 3.6730611525692776E-03_dp, &
         3.4164354226130186E+01_dp, -5.1747145255461746E+01_dp, 3.5667082252749204E+01_dp, &
         -1.1265379545809436E+01_dp, -6.7939439898041476E-01_dp, 1.6480605577661318E+00_dp, &
         -1.3378965635839782E-01_dp, -2.6979010160400613E-01_dp, 4.7149105264278728E-02_dp, &
         5.0747602507635034E-02_dp, -1.2294628821884937E-02_dp, -1.1142091315825215E-02_dp, &
         4.9501418663369838E-03_dp, 4.8931770063798176E+01_dp, -7.9540390842429701E+01_dp, &
         6.0373542926316304E+01_dp, -2.2962155467014412E+01_dp, 1.1360885415171147E+00_dp, &
         2.6763499892965861E+00_dp, -6.0987396392114657E-01_dp, -3.6997629107051999E-01_dp, &
         1.4345000248927353E-01_dp, 6.5524578572070433E-02_dp, -3.9862967761504962E-02_dp, &
         -4.8299270547189094E-03_dp, 5.4310464886229816E-03_dp, 7.1084878218521894E+01_dp, &
         -1.2286891263137933E+02_dp, 1.0163325401682647E+02_dp, -4.5111853024962613E+01_dp, &
         6.4218229411147210E+00_dp, 3.7532636523243239E+00_dp, -1.6002854623816907E+00_dp, &
         -3.7101035717591935E-01_dp, 3.1409505327231513E-01_dp, 5.3941115043966030E-02_dp, &
         -8.3818143472302709E-02_dp, 1.4207263645641415E-02_dp, 3.3546858581714827E-03_dp, &
         1.0588222696611157E+02_dp, -1.9226833008452053E+02_dp, 1.7101798076439422E+02_dp, &
         -8.6186553476897572E+01_dp, 1.9271662529722853E+01_dp, 3.8930409361722291E+00_dp, &
         -3.2922928695756104E+00_dp, -6.0113623291277216E-02_dp, 5.4062487480709565E-01_dp, &
         -2.4107047161548600E-02_dp, -1.2869189718624949E-01_dp, 4.7717461953990542E-02_dp, &
         -2.9342027944556193E-03_dp, 1.6546877079051265E+02_dp, -3.1068544904103140E+02_dp, &
         2.9180626201261043E+02_dp, -1.6205766726181861E+02_dp, 4.7344515303703460E+01_dp, &
         7.0564709632446621E-01_dp, -5.4990736658749366E+00_dp, 8.1374078900166369E-01_dp, &
         7.1175171802245729E-01_dp, -1.8680907638713953E-01_dp, -1.3919953678136415E-01_dp, &
         8.1913606331528166E-02_dp, -1.1490899293904754E-02_dp, 2.8527479104831519E+02_dp, &
         -5.4324338868812106E+02_dp, 5.2597081274478194E+02_dp, -3.1032555466251750E+02_dp, &
         1.0592128649569207E+02_dp, -9.9230478811039369E+00_dp, -7.5217801155503921E+00_dp, &
         2.2260389199008581E+00_dp, 7.4389817085116938E-01_dp, -3.7487661915364073E-01_dp, &
         -1.2135071368815714E-01_dp, 1.0472900042735607E-01_dp, -1.7467242484406273E-02_dp, &
         6.1069908841010852E+02_dp, -1.1574002440882412E+03_dp, 1.1246410459943154E+03_dp, &
         -6.7559373392291536E+02_dp, 2.4469690351149762E+02_dp, -3.4981011467122990E+01_dp, &
         -1.0447270088817676E+01_dp, 4.1199824733302801E+00_dp, 1.0753683950263284E+00_dp, &
         -6.1667707322635523E-01_dp, -1.9231863825085113E-01_dp, 1.6351050609684276E-01_dp, &
         -2.4438171008690827E-02_dp, 2.3309249136020708E+03_dp, -4.3538180199326071E+03_dp, &
         4.1700553851275899E+03_dp, -2.4659990924985104E+03_dp, 8.7681545314646689E+02_dp, &
         -1.2252110351183714E+02_dp, -3.4744515550283161E+01_dp, 1.1919976509846366E+01_dp, &
         4.5071725742760815E+00_dp, -1.9152099076735201E+00_dp, -8.5859628246475261E-01_dp, &
         5.7717184691703094E-01_dp, -7.0410149927279939E-02_dp, 1.8682839453664454E-01_dp, &
         -4.6619677822158430E-02_dp, -1.2244969271828127E-02_dp, 6.0589256164018451E-03_dp, &
         3.6233383923308837E-03_dp, -2.1697547440741212E-03_dp, -1.5324832573249004E-03_dp, &
         7.6307014457020446E-04_dp, 1.5009945716157327E-03_dp, -1.5948622920885888E-03_dp, &
         4.2721023924721354E-04_dp, 1.2452965510300625E-04_dp, -6.8101344859438604E-05_dp, &
         5.8892228459182272E-01_dp, -1.6187708429798231E-01_dp, -3.6609865098592521E-02_dp, &
         2.3522804102595687E-02_dp, 1.0989418080689319E-02_dp, -8.7225886606387588E-03_dp, &
         -4.6358129182236165E-03_dp, 3.1750312025791852E-03_dp, 5.1844812097367743E-03_dp, &
         -6.5275901023297469E-03_dp, 2.5192599245746250E-03_dp, -2.7790386311357066E-05_dp, &
         -1.5519525312227047E-04_dp, 1.0819380779186476E+00_dp, -3.5078545793786275E-01_dp, &
         -5.8317591891871842E-02_dp, 5.8961787807382718E-02_dp, 1.7698109526329605E-02_dp, &
         -2.2770460521331585E-02_dp, -7.3205917206429702E-03_dp, 8.6317273038198361E-03_dp, &
         1.0821379193248763E-02_dp, -1.7385782357341700E-02_dp, 9.1208897862528268E-03_dp, &
         -1.7582601800537039E-03_dp, -2.5708035752802444E-05_dp, 1.7461298655116124E+00_dp, &
         -6.8967176721527312E-01_dp, -6.6751042045674530E-02_dp, 1.3103208070134215E-01_dp, &
         1.9338401160822482E-02_dp, -5.2191210709296623E-02_dp, -7.2199928465625264E-03_dp, &
         2.0478760930027633E-02_dp, 1.9174065370320473E-02_dp, -4.0384174885238051E-02_dp, &
         2.6020899384681384E-02_dp, -7.4705034438871248E-03_dp, 7.2143802459460730E-04_dp, &
         2.6950773766538831E+00_dp, -1.3037225406473845E+00_dp, -3.2614496759563683E-02_dp, &
         2.6916418726799823E-01_dp, 3.0053439263977664E-03_dp, -1.0921441669530184E-01_dp, &
         2.4831672286212188E-03_dp, 4.4091697655829862E-02_dp, 2.9150030585017250E-02_dp, &
         -8.4774815208886814E-02_dp, 6.3626601084289261E-02_dp, -2.2005245386043593E-02_dp, &
         2.9767021718693147E-03_dp, 4.0984375451052717E+00_dp, -2.4066178073051301E+00_dp, &
         1.1383731173835311E-01_dp, 5.1909474912068099E-01_dp, -6.2603108434964416E-02_dp, &
         -2.1175654622555612E-01_dp, 3.8133384598550361E-02_dp, 8.7587360901696096E-02_dp, &
         3.4278741909182162E-02_dp, -1.6260127734933916E-01_dp, 1.3944397087807067E-01_dp, &
         -5.4447820679856637E-02_dp, 8.5453383708596811E-03_dp, 6.2176136529295158E+00_dp, &
         -4.3693211280877762E+00_dp, 5.2774370702314355E-01_dp, 9.4773880309914649E-01_dp, &
         -2.4690257552432496E-01_dp, -3.8251497933289635E-01_dp, 1.3533123197212421E-01_dp, &
         1.6155301558102667E-01_dp, 1.4718947044978946E-02_dp, -2.8450075717598156E-01_dp, &
         2.7982684165522875E-01_dp, -1.2045734281774570E-01_dp, 2.0790000977242391E-02_dp, &
         9.4634111053439103E+00_dp, -7.8418332358592870E+00_dp, 1.5417436172364343E+00_dp, &
         1.6414265351909745E+00_dp, -6.9529406365081492E-01_dp, -6.4050441706217742E-01_dp, &
         3.6647261823447669E-01_dp, 2.7522127699205079E-01_dp, -7.9228003624491242E-02_dp, &
         -4.4498374478349312E-01_dp, 5.1708841341964662E-01_dp, -2.4407600743956154E-01_dp, &
         4.5405289677376504E-02_dp, 1.4490868628033503E+01_dp, -1.3971233939072121E+01_dp, &
         3.8542271282507108E+00_dp, 2.6797445232398029E+00_dp, -1.6991502505639382E+00_dp, &
         -9.7162873904940783E-01_dp, 8.6994103205685502E-01_dp, 4.2251973253482306E-01_dp, &
         -3.5811394605060659E-01_dp, -5.8327449682364263E-01_dp, 8.7181325262147957E-01_dp, &
         -4.5547158946665350E-01_dp, 9.0828400398500778E-02_dp, 2.2357425070187674E+01_dp, &
         -2.4796353116963004E+01_dp, 8.9144869553283570E+00_dp, 4.0371985902019292E+00_dp, &
         -3.8186248869629074E+00_dp, -1.2467737038604281E+00_dp, 1.8925278595991524E+00_dp, &
         5.4265414205624996E-01_dp, -1.0448438269132569E+00_dp, -4.9352647807166389E-01_dp, &
         1.2974717296843923E+00_dp, -7.7527513958093197E-01_dp, 1.6672848898454873E-01_dp, &
         3.4790250541342665E+01_dp, -4.3969980740698468E+01_dp, 1.9713761658579021E+01_dp, &
         5.2896392037293767E+00_dp, -8.0806472982724422E+00_dp, -1.0170835399174816E+00_dp, &
         3.8335251168155606E+00_dp, 4.2219424033780178E-01_dp, -2.5430562807078081E+00_dp, &
         3.5748238853390918E-01_dp, 1.5419198749245833E+00_dp, -1.1644678241284623E+00_dp, &
         2.7666437341984224E-01_dp, 5.4644357897813208E+01_dp, -7.8095758505940708E+01_dp, &
         4.2420187894560037E+01_dp, 4.8151968808568224E+00_dp, -1.6245875469915777E+01_dp, &
         9.6243602774289128E-01_dp, 7.2326780768599983E+00_dp, -5.2544186970366702E-01_dp, &
         -5.4789383956688651E+00_dp, 3.1531099403716132E+00_dp, 8.4886474149751467E-01_dp, &
         -1.3986356059167204E+00_dp, 3.9650600036767097E-01_dp, 8.6699804225947645E+01_dp, &
         -1.3923361555998270E+02_dp, 8.9786173246194181E+01_dp, -2.2858823629011713E+00_dp, &
         -3.0992469263910380E+01_dp, 7.7741399253657599E+00_dp, 1.2505917232015936E+01_dp, &
         -3.7068949594152851E+00_dp, -1.0555346369807513E+01_dp, 1.0188356709429819E+01_dp, &
         -2.6001240535680141E+00_dp, -7.9853252277010678E-01_dp, 4.2303845984206989E-01_dp, &
         1.3907554680969875E+02_dp, -2.4965842431321846E+02_dp, 1.8831846945865320E+02_dp, &
         -2.9489704359881337E+01_dp, -5.5316316864452965E+01_dp, 2.6447389036265232E+01_dp, &
         1.8854798651484575E+01_dp, -1.2120609962757287E+01_dp, -1.7763715122527280E+01_dp, &
         2.5231831140577533E+01_dp, -1.2544891115185020E+01_dp, 2.2668677068936747E+00_dp, &
         6.8708716177258419E-02_dp, 2.2580720743716594E+02_dp, -4.5104097133606439E+02_dp, &
         3.9351587358059925E+02_dp, -1.1220803805233957E+02_dp, -8.8584998126696050E+01_dp, &
         7.1690786222667725E+01_dp, 2.0820379520370651E+01_dp, -3.1251600385405137E+01_dp, &
         -2.3842474793325504E+01_dp, 5.2733349756860250E+01_dp, -3.5442716338070284E+01_dp, &
         1.1104138822595329E+01_dp, -1.3193636495537726E+00_dp, 3.7178643267228017E+02_dp, &
         -8.2258367035709261E+02_dp, 8.2255509467989646E+02_dp, -3.4122567397482152E+02_dp, &
         -1.1069056818136373E+02_dp, 1.7023549623079217E+02_dp, -1.0915097896851977E+00_dp, &
         -6.8273494641212551E+01_dp, -1.5708419656731760E+01_dp, 9.2205658507516148E+01_dp, &
         -7.8945038472936645E+01_dp, 3.0974630589084953E+01_dp, -4.9401496535210629E+00_dp, &
         6.2326881452254656E+02_dp, -1.5188270356173509E+03_dp, 1.7254522561739329E+03_dp, &
         -9.4519869568904846E+02_dp, -2.7999063550782221E+01_dp, 3.5481857290613175E+02_dp, &
         -1.0099523826322665E+02_dp, -1.2123045796995562E+02_dp, 4.0970797552648818E+01_dp, &
         1.1891202216538902E+02_dp, -1.3972808947295965E+02_dp, 6.5322452115099665E+01_dp, &
         -1.2075994070187011E+01_dp, 1.0756923535067231E+03_dp, -2.8616598027309929E+03_dp, &
         3.6507795405336346E+03_dp, -2.4909049195091211E+03_dp, 4.8798405235788942E+02_dp, &
         5.9835790021628566E+02_dp, -3.8831159871198366E+02_dp, -1.3665662485667869E+02_dp, &
         1.9320859447294200E+02_dp, 5.7805579962611311E+01_dp, -1.7452945217439253E+02_dp, &
         1.0228317193023724E+02_dp, -2.1552120617229455E+01_dp, 1.9817023447702200E+03_dp, &
         -5.6689826598340678E+03_dp, 7.9668997146417360E+03_dp, -6.4333249134176367E+03_dp, &
         2.4219830207338719E+03_dp, 5.7967220853445599E+02_dp, -9.5984621210825696E+02_dp, &
         3.2978769612793030E+01_dp, 4.1917841670563894E+02_dp, -1.5534544314105773E+02_dp, &
         -1.2421868756049768E+02_dp, 1.1843742614679046E+02_dp, -2.9556703864728348E+01_dp, &
         4.5306956901767944E+03_dp, -1.3608434324291762E+04_dp, 2.0416922578512735E+04_dp, &
         -1.8375713453892236E+04_dp, 9.1917215327749964E+03_dp, -7.4781042033255528E+02_dp, &
         -1.8977895694601118E+03_dp, 5.0289039209206737E+02_dp, 7.8898968358442164E+02_dp, &
         -5.7658145419403400E+02_dp, -1.4194226893666439E+01_dp, 1.4827723853173043E+02_dp, &
         -4.5635215304885165E+01_dp, 1.5132189223523540E+00_dp, -3.8714502724407229E-01_dp, &
         -9.8018889420702443E-02_dp, 5.1908250545652360E-02_dp, 2.9129428048283938E-02_dp, &
         -1.8776985333903432E-02_dp, -1.2318593850205199E-02_dp, 6.6695589588984475E-03_dp, &
         1.2466570600707088E-02_dp, -1.3874197741980808E-02_dp, 4.1907856070478042E-03_dp, &
         7.5271368574731106E-04_dp, -5.1647877466666774E-04_dp, 1.7444005611942730E+00_dp, &
         -5.6824316669549102E-01_dp, -9.5567270387908632E-02_dp, 9.5953975937291464E-02_dp, &
         2.9440932840795256E-02_dp, -3.7050946327260699E-02_dp, -1.2287347330040912E-02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c03 = (/1.4012464467116794E-02_dp, &
         1.7870937744543961E-02_dp, -2.8284778305657239E-02_dp, 1.4614940139686226E-02_dp, &
         -2.7014027690474133E-03_dp, -7.8952988099918739E-05_dp, 2.2522270551584125E+00_dp, &
         -9.9237513069804251E-01_dp, -7.1665511461607434E-02_dp, 1.9968203248776958E-01_dp, &
         2.1586264982050893E-02_dp, -8.0464217361784704E-02_dp, -7.7695595383537727E-03_dp, &
         3.1814347737576659E-02_dp, 2.7811012184441854E-02_dp, -6.2649965060119633E-02_dp, &
         4.1918544103320216E-02_dp, -1.2666876155687441E-02_dp, 1.3660025331692276E-03_dp, &
         3.1355625943697762E+00_dp, -1.7996536940652996E+00_dp, 2.2381691245553063E-02_dp, &
         3.9612212867481145E-01_dp, -1.6449727468430150E-02_dp, -1.6295147202484259E-01_dp, &
         1.2800103481059741E-02_dp, 6.6491589187543318E-02_dp, 3.8496472917503681E-02_dp, &
         -1.2749901018687695E-01_dp, 1.0004618939892632E-01_dp, -3.6137844427782950E-02_dp, &
         5.1770195187856929E-03_dp, 4.5668263189960809E+00_dp, -3.2508325919683241E+00_dp, &
         2.9442091272203441E-01_dp, 7.3905390144916405E-01_dp, -1.3384430189810520E-01_dp, &
         -3.0519473426292121E-01_dp, 7.5103144196923433E-02_dp, 1.2781992372799392E-01_dp, &
         3.7492879106017292E-02_dp, -2.3587781768150057E-01_dp, 2.1232133279926324E-01_dp, &
         -8.6033979418852449E-02_dp, 1.4025330806284918E-02_dp, 6.8306176002396661E+00_dp, &
         -5.8119086384988803E+00_dp, 9.7490754567894877E-01_dp, 1.3049740304463888E+00_dp, &
         -4.3490563945641791E-01_dp, -5.3072489778072540E-01_dp, 2.3270303823013319E-01_dp, &
         2.2740302928499284E-01_dp, -8.6507435588130746E-03_dp, -3.9468254647236639E-01_dp, &
         4.1189269031114867E-01_dp, -1.8379654833529319E-01_dp, 3.2710121302165010E-02_dp, &
         1.0390575731969639E+01_dp, -1.0309231722090752E+01_dp, 2.5464315850199029E+00_dp, &
         2.1830827795271524E+00_dp, -1.1319594654536267E+00_dp, -8.4971117936736751E-01_dp, &
         5.9122626817509949E-01_dp, 3.7127494548695916E-01_dp, -1.7896260036582676E-01_dp, &
         -5.7922303833023392E-01_dp, 7.3368980026983677E-01_dp, -3.6037859396806582E-01_dp, &
         6.9003335230075069E-02_dp, 1.6002364677374977E+01_dp, -1.8210615137406258E+01_dp, &
         6.0073025151814532E+00_dp, 3.4250138514203066E+00_dp, -2.6414806725325550E+00_dp, &
         -1.2096574762879975E+00_dp, 1.3472659206289432E+00_dp, 5.3635873100939280E-01_dp, &
         -6.4288697341708967E-01_dp, -6.6970318883021285E-01_dp, 1.1814039303496457E+00_dp, &
         -6.4954054414694495E-01_dp, 1.3361067762784279E-01_dp, 2.4905165004773185E+01_dp, &
         -3.2138208303483374E+01_dp, 1.3409824399537735E+01_dp, 4.8815305907327238E+00_dp, &
         -5.7496874849540909E+00_dp, -1.3655975920609458E+00_dp, 2.8416424417161545E+00_dp, &
         6.0744001443999063E-01_dp, -1.7319413268413426E+00_dp, -3.1102100539977884E-01_dp, &
         1.6338737133746211E+00_dp, -1.0583411311910322E+00_dp, 2.3663443320973138E-01_dp, &
         3.9149516015116362E+01_dp, -5.6813706285333311E+01_dp, 2.8965448354605932E+01_dp, &
         5.7421686338785047E+00_dp, -1.1871980500371624E+01_dp, -5.6543446935291142E-01_dp, &
         5.6045107680728776E+00_dp, 2.2503023424962715E-01_dp, -4.0269224337355114E+00_dp, &
         1.3631950103938133E+00_dp, 1.6221634173962662E+00_dp, -1.4818038878243160E+00_dp, &
         3.7423258125912723E-01_dp, 6.2165735856370389E+01_dp, -1.0082025924844557E+02_dp, &
         6.1330857086424636E+01_dp, 3.3187227725079316E+00_dp, -2.3380689687069282E+01_dp, &
         3.1683638576480098E+00_dp, 1.0296270093353964E+01_dp, -1.5512204348401999E+00_dp, &
         -8.3821659603468390E+00_dp, 6.2014658906591276E+00_dp, -1.3435796755846094E-01_dp, &
         -1.4991994694273796E+00_dp, 4.9235900829999746E-01_dp, 9.9764309320929868E+01_dp, &
         -1.7992761720012572E+02_dp, 1.2838124909734995E+02_dp, -1.0067710243253384E+01_dp, &
         -4.3747416798516454E+01_dp, 1.4588791418335694E+01_dp, 1.7234209303970911E+01_dp, &
         -6.9157222786756467E+00_dp, -1.5616463216416106E+01_dp, 1.7661695341240645E+01_dp, &
         -6.5632441414580995E+00_dp, 4.6107861550632338E-02_dp, 4.0429578336215916E-01_dp, &
         1.6192996213509178E+02_dp, -3.2345183259818674E+02_dp, 2.6728260309780984E+02_dp, &
         -5.4934721948077964E+01_dp, -7.6420235817495239E+01_dp, 4.4341723962570370E+01_dp, &
         2.4617318327247116E+01_dp, -2.0464011397274241E+01_dp, -2.5136972926424946E+01_dp, &
         4.1157382314364590E+01_dp, -2.3454249403793014E+01_dp, 5.7905373012948962E+00_dp, &
         -3.7151924625073762E-01_dp, 2.6612804981637356E+02_dp, -5.8660374248587323E+02_dp, &
         5.5597284125538818E+02_dp, -1.8432038838199645E+02_dp, -1.1859595383908474E+02_dp, &
         1.1449618036240184E+02_dp, 2.3096079841726318E+01_dp, -5.0496124411414627E+01_dp, &
         -3.0760587855042580E+01_dp, 8.2441989484414663E+01_dp, -6.0440740405265736E+01_dp, &
         2.0892064007476247E+01_dp, -2.8821707975965407E+00_dp, 4.4369669215955679E+02_dp, &
         -1.0749724494514057E+03_dp, 1.1593969816222766E+03_dp, -5.3302317687626135E+02_dp, &
         -1.3715111027236267E+02_dp, 2.6529860970775599E+02_dp, -1.9004859388517797E+01_dp, &
         -1.0782865542385048E+02_dp, -1.0024949426649870E+01_dp, 1.3842513763413518E+02_dp, &
         -1.2847895610869600E+02_dp, 5.3470240396583165E+01_dp, -9.0328668602524544E+00_dp, &
         7.5346534185182429E+02_dp, -1.9952392505299072E+03_dp, 2.4299447215024820E+03_dp, &
         -1.4396109992918582E+03_dp, 1.3298234472702070E+01_dp, 5.4802453254041689E+02_dp, &
         -1.9023333858956633E+02_dp, -1.8964738445274472E+02_dp, 9.7587567615069020E+01_dp, &
         1.6354143452048581E+02_dp, -2.1971374673088104E+02_dp, 1.0879392628260086E+02_dp, &
         -2.0963583502054494E+01_dp, 1.3170419393603040E+03_dp, -3.7739560335424503E+03_dp, &
         5.1335488235209250E+03_dp, -3.7408311327837364E+03_dp, 8.4211641182365474E+02_dp, &
         9.2551814404559434E+02_dp, -6.8464014949108071E+02_dp, -2.0411488709587312E+02_dp, &
         3.9392227255553178E+02_dp, -2.7256439657068738E+00_dp, -2.3813864878020763E+02_dp, &
         1.5893698689400233E+02_dp, -3.5584326778431219E+01_dp, 2.4452352059405862E+03_dp, &
         -7.4377745780056084E+03_dp, 1.1067488576469721E+04_dp, -9.4896153343244987E+03_dp, &
         3.8782719555563267E+03_dp, 8.3846279328892012E+02_dp, -1.6963477989104424E+03_dp, &
         1.9083731740985843E+02_dp, 8.1093174916219630E+02_dp, -5.4226175427579835E+02_dp, &
         1.2518886616972225E+01_dp, 1.1490785535678236E+02_dp, -3.6491764381788030E+01_dp, &
         5.3330085765150261E+03_dp, -1.6666877738151037E+04_dp, 2.6076020406118962E+04_dp, &
         -2.4677010132749703E+04_dp, 1.3307416777568624E+04_dp, -1.6319434420381626E+03_dp, &
         -2.8258156451600034E+03_dp, 1.2347677417808713E+03_dp, 9.6224967002132507E+02_dp, &
         -1.1349195336895932E+03_dp, 3.5716055241575810E+02_dp, 3.3514646317404001E+01_dp, &
         -3.1835363222344533E+01_dp, 1.9791355268743882E+04_dp, -6.1293055686968582E+04_dp, &
         9.5680928609287061E+04_dp, -9.1568459060692185E+04_dp, 5.2201017929389076E+04_dp, &
         -1.1121707130041303E+04_dp, -5.9896108180577212E+03_dp, 2.8711813866255552E+03_dp, &
         2.8442192738211143E+03_dp, -2.8748668019934485E+03_dp, 5.5449597276475947E+02_dp, &
         3.3697546690239642E+02_dp, -1.4001963907445077E+02_dp, 2.1435457880133899E-01_dp, &
         -3.0445050547489378E-02_dp, -1.1502220181206861E-02_dp, 3.3927624439129855E-03_dp, &
         3.4189557542295174E-03_dp, -1.2218199554707497E-03_dp, -1.5066279978736984E-03_dp, &
         3.5867793931900605E-04_dp, 1.4303870829534720E-03_dp, -1.1751047917745799E-03_dp, &
         1.1448519522410556E-04_dp, 2.2484723844012693E-04_dp, -7.9625602391486381E-05_dp, &
         6.8669139798068002E-01_dp, -1.1070670760471460E-01_dp, -3.8815146888622191E-02_dp, &
         1.4421740546154914E-02_dp, 1.1981770590339011E-02_dp, -5.3869891046450912E-03_dp, &
         -5.3259041256285061E-03_dp, 1.6932851375693670E-03_dp, 5.4587177464697767E-03_dp, &
         -5.1318888796206240E-03_dp, 1.1113119925963892E-03_dp, 5.5707042656183924E-04_dp, &
         -2.5031832857251932E-04_dp, 1.3017958498163928E+00_dp, -2.5836549390329550E-01_dp, &
         -7.9522740248279142E-02_dp, 4.0966970905658096E-02_dp, 2.5961846371615496E-02_dp, &
         -1.5996708083253323E-02_dp, -1.1668689158977025E-02_dp, 5.3987213548460644E-03_dp, &
         1.3553939593849051E-02_dp, -1.5161416199164019E-02_dp, 5.2920214407258222E-03_dp, &
         2.4789363134447497E-04_dp, -4.1875707155075752E-04_dp, 2.1973724796967060E+00_dp, &
         -5.5274701142067917E-01_dp, -1.4357245126583923E-01_dp, 1.0384774390949460E-01_dp, &
         4.9473349286753690E-02_dp, -4.2206845894195689E-02_dp, -2.2382617295611081E-02_dp, &
         1.5059572073191307E-02_dp, 3.0398772810786283E-02_dp, -4.0085034231927559E-02_dp, &
         1.8281218393724739E-02_dp, -2.3368359760019852E-03_dp, -4.1946031605447358E-04_dp, &
         3.5872393736741892E+00_dp, -1.1397626678673476E+00_dp, -2.3978729107103397E-01_dp, &
         2.4474181628521541E-01_dp, 8.5876014179872409E-02_dp, -1.0302284902670406E-01_dp, &
         -3.8644600170405448E-02_dp, 3.8463608204633855E-02_dp, 6.3915061285770847E-02_dp, &
         -9.8422842593201207E-02_dp, 5.3458921537296901E-02_dp, -1.1737485346159822E-02_dp, &
         3.5223316761622068E-04_dp, 5.8230089184481608E+00_dp, -2.2981115043123683E+00_dp, &
         -3.6708206680875766E-01_dp, 5.4793630683732597E-01_dp, 1.3260167763700370E-01_dp, &
         -2.3785074235442757E-01_dp, -5.7719416353132290E-02_dp, 9.2450227914151389E-02_dp, &
         1.2669173511185230E-01_dp, -2.2861142653418301E-01_dp, 1.4176511400260272E-01_dp, &
         -3.9625027891297522E-02_dp, 3.6743327870223182E-03_dp, 9.5024717953790585E+00_dp, &
         -4.5698286444549199E+00_dp, -4.8749963554224052E-01_dp, 1.1830931580696589E+00_dp, &
         1.6497571972527433E-01_dp, -5.2711345036646928E-01_dp, -6.2998011226482770E-02_dp, &
         2.1284167669799156E-01_dp, 2.3538540644563008E-01_dp, -5.0792488101412725E-01_dp, &
         3.5245039897125219E-01_dp, -1.1446896115953678E-01_dp, 1.4289249240853091E-02_dp, &
         1.5666042754500245E+01_dp, -9.0233915532177953E+00_dp, -4.5159576117067335E-01_dp, &
         2.4887534553074904E+00_dp, 9.0670470640627859E-02_dp, -1.1312846755872035E+00_dp, &
         3.8111343009444320E-03_dp, 4.7424674847483844E-01_dp, 4.0054202834589053E-01_dp, &
         -1.0842444122478243E+00_dp, 8.3507352445861216E-01_dp, -3.0284416515089485E-01_dp, &
         4.4096366950424772E-02_dp, 2.6161032398783654E+01_dp, -1.7794137065441063E+01_dp, &
         1.9957922278720683E-01_dp, 5.1325134511234731E+00_dp, -3.7238280322014222E-01_dp, &
         -2.3605015078492113E+00_dp, 3.1646959944127984E-01_dp, 1.0276213558669789E+00_dp, &
         5.8276774951493282E-01_dp, -2.2183031836814724E+00_dp, 1.8984835446720232E+00_dp, &
         -7.5410521451439838E-01_dp, 1.2153634803586906E-01_dp, 4.4332125750992894E+01_dp, &
         -3.5216911762790389E+01_dp, 2.7515312485019741E+00_dp, 1.0407651938533359E+01_dp, &
         -2.0146926217048411E+00_dp, -4.7839201365532205E+00_dp, 1.3535664467773134E+00_dp, &
         2.1644804230399943E+00_dp, 5.4450764825895215E-01_dp, -4.2990170260660685E+00_dp, &
         4.1392834856622303E+00_dp, -1.7868794706520530E+00_dp, 3.1134654318639488E-01_dp, &
         7.6360718814403441E+01_dp, -7.0253763630464462E+01_dp, 1.0664558219640512E+01_dp, &
         2.0743007038958023E+01_dp, -6.9544924163722843E+00_dp, -9.3467664173592819E+00_dp, &
         4.3712655485208352E+00_dp, 4.3994628592890503E+00_dp, -5.9790313958115127E-01_dp, &
         -7.6609316918298695E+00_dp, 8.5784839688301471E+00_dp, -4.0319809725090563E+00_dp, &
         7.5205840776261002E-01_dp, 1.3392877071998257E+02_dp, -1.4183042758875911E+02_dp, &
         3.3121435990401395E+01_dp, 4.0426295721862452E+01_dp, -2.0721850212906318E+01_dp, &
         -1.7263628579210469E+01_dp, 1.2547672408466081E+01_dp, 8.4591865266308197E+00_dp, &
         -5.6249151706448863E+00_dp, -1.1592672327907211E+01_dp, 1.6508808139297699E+01_dp, &
         -8.5771693846067460E+00_dp, 1.7125060439946929E+00_dp, 2.3966601282404417E+02_dp, &
         -2.9090772769614887E+02_dp, 9.4528375241536281E+01_dp, 7.5979855023555004E+01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c04 = (/-5.7499561390798192E+01_dp, &
         -2.8667295698939792E+01_dp, 3.3623712254987076E+01_dp, 1.4613232172192909E+01_dp, &
         -2.2425748408516260E+01_dp, -1.0625183610714041E+01_dp, 2.7811282349984850E+01_dp, &
         -1.6721270864181005E+01_dp, 3.6190172602275243E+00_dp, 4.3864077011236401E+02_dp, &
         -6.0869407672507202E+02_dp, 2.6061415383600962E+02_dp, 1.3295484378240570E+02_dp, &
         -1.5302514848145449E+02_dp, -3.6108018483517967E+01_dp, 8.5434219515898235E+01_dp, &
         1.9050431653711076E+01_dp, -7.1236044045457149E+01_dp, 1.7168053422465423E+01_dp, &
         3.3523780444104382E+01_dp, -2.7691407702644518E+01_dp, 6.7871470608408817E+00_dp, &
         8.2347485996135970E+02_dp, -1.3050923635822317E+03_dp, 7.1254991634726696E+02_dp, &
         1.9443051128113493E+02_dp, -3.9477087573253618E+02_dp, 1.1049102653040705E+00_dp, &
         2.0471622937020155E+02_dp, -1.0305127225062771E+00_dp, -1.9658691426799172E+02_dp, &
         1.4177328838717565E+02_dp, -9.4534452677625325E+00_dp, -2.8803213147124975E+01_dp, &
         9.8371070114488415E+00_dp, 1.5915710687424632E+03_dp, -2.8820355770147166E+03_dp, &
         1.9663633983704333E+03_dp, 1.1799337397837300E+02_dp, -9.8270228323130664E+02_dp, &
         2.4974930541290666E+02_dp, 4.4552578036203141E+02_dp, -1.4025777789091055E+02_dp, &
         -4.6557102028422628E+02_dp, 5.4786103049593976E+02_dp, -2.4864894143143812E+02_dp, &
         3.6996297729905450E+01_dp, 3.6444195240281498E+00_dp, 3.1826214065789259E+03_dp, &
         -6.5947551051531327E+03_dp, 5.5553678841919664E+03_dp, -8.2212981706636367E+02_dp, &
         -2.2837692981734585E+03_dp, 1.3151448072342987E+03_dp, 7.6228584365461666E+02_dp, &
         -7.2065304049405756E+02_dp, -8.2516821940278430E+02_dp, 1.5841227049850456E+03_dp, &
         -1.0731502300722150E+03_dp, 3.5302215207929873E+02_dp, -4.6365877199563009E+01_dp, &
         6.6383831975153153E+03_dp, -1.5758014247231216E+04_dp, 1.6270823641939674E+04_dp, &
         -5.9638112817090259E+03_dp, -4.3121794406394965E+03_dp, 5.1144301182847603E+03_dp, &
         1.8462464529373887E+02_dp, -2.5218095815008292E+03_dp, -2.3054431334728457E+02_dp, &
         3.1514734149831415E+03_dp, -2.9921625738998023E+03_dp, 1.2837985283984451E+03_dp, &
         -2.2471322113598779E+02_dp, 1.4761997601240251E+04_dp, -3.9939749414678743E+04_dp, &
         5.0012213870608291E+04_dp, -3.0242466715096940E+04_dp, -1.1894217165033751E+03_dp, &
         1.5195501545453046E+04_dp, -6.8048688127790801E+03_dp, -4.7801527082331741E+03_dp, &
         5.1277990028047161E+03_dp, 1.1223380413403229E+03_dp, -4.0040636605906966E+03_dp, &
         2.3860787033124807E+03_dp, -5.0750360903198020E+02_dp, 3.9204413377125406E+04_dp, &
         -1.1769916409401350E+05_dp, 1.7194986056957813E+05_dp, -1.4277204840178991E+05_dp, &
         5.2345034138577081E+04_dp, 1.9880789878096351E+04_dp, -2.9425779677239327E+04_dp, &
         3.4033464159393416E+03_dp, 1.2155031462654326E+04_dp, -7.1247822519758038E+03_dp, &
         -1.0526634908337862E+03_dp, 2.3358638291647071E+03_dp, -6.6372768936816408E+02_dp, &
         1.7431528419766495E+00_dp, -2.5588307542896549E-01_dp, -9.4854900124262329E-02_dp, &
         2.9794665014952194E-02_dp, 2.8483247338140741E-02_dp, -1.0841558810136384E-02_dp, &
         -1.2582020468356882E-02_dp, 3.2484447101857397E-03_dp, 1.2184683287913497E-02_dp, &
         -1.0393043078939668E-02_dp, 1.3734683830726878E-03_dp, 1.7378837582184665E-03_dp, &
         -6.4675618806164781E-04_dp, 2.0999282096752392E+00_dp, -4.1684702634050547E-01_dp, &
         -1.2966447121561134E-01_dp, 6.5531871892109725E-02_dp, 4.2460205357563126E-02_dp, &
         -2.5433368014272466E-02_dp, -1.9110254921504966E-02_dp, 8.5170411027574670E-03_dp, &
         2.1895399268815847E-02_dp, -2.4046927930312594E-02_dp, 8.0713367078222315E-03_dp, &
         6.1894677827445367E-04_dp, -7.1620971319254038E-04_dp, 2.9103286794502852E+00_dp, &
         -8.1511239504650834E-01_dp, -2.0272420121372503E-01_dp, 1.6000857299250551E-01_dp, &
         7.2180276628084461E-02_dp, -6.5213326048744386E-02_dp, -3.2904145344168502E-02_dp, &
         2.3370721007330255E-02_dp, 4.5890422026197968E-02_dp, -6.1738612510888358E-02_dp, &
         2.8848028203159046E-02_dp, -4.0746891743398553E-03_dp, -5.3779869066474179E-04_dp, &
         4.3929184784296806E+00_dp, -1.6330912689205386E+00_dp, -3.1604791770547475E-01_dp, &
         3.6946636122180865E-01_dp, 1.1843930520825167E-01_dp, -1.5647910975653245E-01_dp, &
         -5.3772992985622166E-02_dp, 5.8807839980255634E-02_dp, 9.4040122713274651E-02_dp, &
         -1.4935257927475140E-01_dp, 8.3382587423705509E-02_dp, -1.9374603363277509E-02_dp, &
         8.9154656100587327E-04_dp, 6.9535647507228511E+00_dp, -3.2407207810334469E+00_dp, &
         -4.5450216314505965E-01_dp, 8.1120311488282126E-01_dp, 1.7199729289635310E-01_dp, &
         -3.5502288370822022E-01_dp, -7.5071517391571155E-02_dp, 1.3903228386390781E-01_dp, &
         1.8246528140825818E-01_dp, -3.4175970207501360E-01_dp, 2.1737747326196696E-01_dp, &
         -6.3021723009680411E-02_dp, 6.3596358141532021E-03_dp, 1.1318724501442688E+01_dp, &
         -6.3711340630689817E+00_dp, -5.4290267282805349E-01_dp, 1.7222212774009904E+00_dp, &
         1.8800546629110781E-01_dp, -7.7473483399344423E-01_dp, -6.8074344065584824E-02_dp, &
         3.1549770235302227E-01_dp, 3.3134230367045064E-01_dp, -7.4915846108706863E-01_dp, &
         5.3259600474652258E-01_dp, -1.7774557823373618E-01_dp, 2.3124115670092340E-02_dp, &
         1.8787690401309099E+01_dp, -1.2481591332976498E+01_dp, -3.3240738329162378E-01_dp, &
         3.5746152138023852E+00_dp, 1.6582305790403686E-02_dp, -1.6418374158804947E+00_dp, &
         6.0369727099797245E-02_dp, 6.9497178729859110E-01_dp, 5.4759326848085665E-01_dp, &
         -1.5808873877659888E+00_dp, 1.2475010854139670E+00_dp, -4.6259424760260659E-01_dp, &
         6.9150781216906557E-02_dp, 3.1709193194004314E+01_dp, -2.4500936384364149E+01_dp, &
         8.9530424805676101E-01_dp, 7.2967750917633403E+00_dp, -7.8033873418639665E-01_dp, &
         -3.3908784872316935E+00_dp, 5.7873892445767738E-01_dp, 1.4926423753313225E+00_dp, &
         7.5568258108913688E-01_dp, -3.2012379368822348E+00_dp, 2.8106680717115542E+00_dp, &
         -1.1384202117706752E+00_dp, 1.8702710303308939E-01_dp, 5.4397061089613771E+01_dp, &
         -4.8406686443798705E+01_dp, 5.0922674916099249E+00_dp, 1.4682021374582865E+01_dp, &
         -3.4028187575347335E+00_dp, -6.8129880929714179E+00_dp, 2.2134139481608028E+00_dp, &
         3.1225310594685567E+00_dp, 5.6939022073493750E-01_dp, -6.1380258185628369E+00_dp, &
         6.0832724690905824E+00_dp, -2.6743583995551066E+00_dp, 4.7321157108309642E-01_dp, &
         9.4927520965517601E+01_dp, -9.6626388107481489E+01_dp, 1.7423709956011194E+01_dp, &
         2.9087093576015224E+01_dp, -1.1024713061645821E+01_dp, -1.3204431435260902E+01_dp, &
         6.8519114491064999E+00_dp, 6.3123437593929230E+00_dp, -1.3914514684572890E+00_dp, &
         -1.0781645944177450E+01_dp, 1.2524886415280703E+01_dp, -5.9961542854634038E+00_dp, &
         1.1336014805592638E+00_dp, 1.6875369326167805E+02_dp, -1.9555844833588057E+02_dp, &
         5.1452978374839745E+01_dp, 5.6408183065757186E+01_dp, -3.1869524561997707E+01_dp, &
         -2.4169756910093785E+01_dp, 1.9246480032110174E+01_dp, 1.2080073322081908E+01_dp, &
         -9.4081964481552394E+00_dp, -1.5842846845631168E+01_dp, 2.3929916792690261E+01_dp, &
         -1.2694153793065942E+01_dp, 2.5676754426641781E+00_dp, 3.0617464355184120E+02_dp, &
         -4.0267737873989802E+02_dp, 1.4297706304699409E+02_dp, 1.0552060360349593E+02_dp, &
         -8.6951847266308548E+01_dp, -3.9583498741809187E+01_dp, 5.0971829663593091E+01_dp, &
         2.0756769842847500E+01_dp, -3.5608237890535712E+01_dp, -1.2635454822619382E+01_dp, &
         3.9827896269018439E+01_dp, -2.4641903351275278E+01_dp, 5.4102958418032170E+00_dp, &
         5.6824683386744607E+02_dp, -8.4670140808710323E+02_dp, 3.8789802089036772E+02_dp, &
         1.8361314535314401E+02_dp, -2.2924223071701195E+02_dp, -4.8028389813997478E+01_dp, &
         1.2888310185706845E+02_dp, 2.6716683644805315E+01_dp, -1.1141263626421915E+02_dp, &
         3.3081570439399378E+01_dp, 4.6048288731971617E+01_dp, -4.0522118414987965E+01_dp, &
         1.0132031238850658E+01_dp, 1.0818700689018240E+03_dp, -1.8254063981441693E+03_dp, &
         1.0493135940933003E+03_dp, 2.6594497727926023E+02_dp, -5.8924539721155190E+02_dp, &
         1.0644611347412503E+01_dp, 3.0960803248045971E+02_dp, -3.5086367758831214E+00_dp, &
         -3.0809815313509745E+02_dp, 2.3143304305545996E+02_dp, -2.4814271428416337E+01_dp, &
         -4.0649889869082884E+01_dp, 1.4627459402254191E+01_dp, 2.1203027222125434E+03_dp, &
         -4.0538804318582847E+03_dp, 2.8727141030336588E+03_dp, 1.5342090598361241E+02_dp, &
         -1.4719332795253706E+03_dp, 3.8093938541663044E+02_dp, 6.8480755414744362E+02_dp, &
         -2.1085331534342177E+02_dp, -7.4477742346268985E+02_dp, 8.8210214793824537E+02_dp, &
         -4.0743642895084110E+02_dp, 6.4513827918347346E+01_dp, 4.7653632712549898E+00_dp, &
         4.2971394915069723E+03_dp, -9.3250590824070587E+03_dp, 8.0591228301709589E+03_dp, &
         -1.1497631929121603E+03_dp, -3.4872799439592382E+03_dp, 1.9571861069180222E+03_dp, &
         1.2474018707737353E+03_dp, -1.0953820294698519E+03_dp, -1.4138330704595035E+03_dp, &
         2.6149313818024643E+03_dp, -1.7568818980540248E+03_dp, 5.7462043816387722E+02_dp, &
         -7.4876139865788033E+01_dp, 9.0670446226651657E+03_dp, -2.2366825283029859E+04_dp, &
         2.3435539043591565E+04_dp, -8.1542691185909571E+03_dp, -7.1386507741248670E+03_dp, &
         7.7340914601856848E+03_dp, 8.4604251076613593E+02_dp, -4.0914375105218251E+03_dp, &
         -9.5634197023942659E+02_dp, 5.8324252454260395E+03_dp, -5.3006063534238547E+03_dp, &
         2.2296251828676996E+03_dp, -3.8479712431583698E+02_dp, 2.0208180308777472E+04_dp, &
         -5.6497936633236641E+04_dp, 7.1357135395308476E+04_dp, -4.1006680278242755E+04_dp, &
         -6.7704662049699600E+03_dp, 2.5578407436596837E+04_dp, -8.5907976304829099E+03_dp, &
         -1.0696042935228554E+04_dp, 8.5873059932908618E+03_dp, 4.3094498446251810E+03_dp, &
         -9.2350938585649274E+03_dp, 5.1924066610847412E+03_dp, -1.0815797042462589E+03_dp, &
         5.0205247084381823E+04_dp, -1.5546532286454976E+05_dp, 2.3094943138771786E+05_dp, &
         -1.8852660057995931E+05_dp, 5.4423083001147803E+04_dp, 5.2087747303112206E+04_dp, &
         -5.7021055651190843E+04_dp, 1.6354776707089343E+03_dp, 3.3477410681770663E+04_dp, &
         -2.5610268529159377E+04_dp, 6.0980258016920243E+03_dp, 1.4484722062152532E+03_dp, &
         -7.8447165419517830E+02_dp, 1.8887929346820957E+05_dp, -6.0391416820596519E+05_dp, &
         9.6236000817517866E+05_dp, -9.2625010870439373E+05_dp, 5.0796682925588451E+05_dp, &
         -6.3746047988236256E+04_dp, -1.1255310501935899E+05_dp, 5.5203234544571715E+04_dp, &
         2.9119699905471261E+04_dp, -3.7170133622062072E+04_dp, 9.5002199317095183E+03_dp, &
         3.0563342511347387E+03_dp, -1.5546850581139452E+03_dp, 2.2978154445120763E-01_dp, &
         -1.5786385884921532E-02_dp, -8.6068171899530838E-03_dp, 1.2219346454486037E-03_dp, &
         1.9728968192239591E-03_dp, 4.8787622431389017E-06_dp, -1.2138585236970429E-03_dp, &
         2.8402149565212066E-04_dp, 5.0364218620146115E-04_dp, -2.0104688190957889E-04_dp, &
         -2.3690088629292757E-04_dp, 2.2016840122659345E-04_dp, -5.5777402901406545E-05_dp, &
         7.4358607798286502E-01_dp, -5.9062732873971499E-02_dp, -3.1312168545837522E-02_dp, &
         5.5131899950392558E-03_dp, 7.4627215090000507E-03_dp, -2.9791391683538964E-04_dp, &
         -4.6367875641200578E-03_dp, 1.2011301801606153E-03_dp, 2.0057385312576845E-03_dp, &
         -9.6341439171540230E-04_dp, -7.8097734459690806E-04_dp, 8.0779389096505641E-04_dp, &
         -2.1048482828717736E-04_dp, 1.4375579606459543E+00_dp, -1.4409331357953084E-01_dp, &
         -7.3049415971812165E-02_dp, 1.6928103093679515E-02_dp, 1.8417725387954376E-02_dp, &
         -1.9453592218161570E-03_dp, -1.1634328023640639E-02_dp, 3.4714019819061617E-03_dp, &
         5.3490931351907831E-03_dp, -3.1907950343243647E-03_dp, -1.4616713023815479E-03_dp, &
         1.8903207684751867E-03_dp, -5.1679973398017858E-04_dp, 2.4951739894697100E+00_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c05 = (/-3.2389785098025670E-01_dp, &
         -1.5588421416935677E-01_dp, 4.6469218783312088E-02_dp, 4.1557942308222309E-02_dp, &
         -7.5142092188104671E-03_dp, -2.6806031107589168E-02_dp, 9.2275566818059377E-03_dp, &
         1.3149892067901462E-02_dp, -9.4462212865469193E-03_dp, -1.9880182010796601E-03_dp, &
         3.9637132356526448E-03_dp, -1.1550825018293159E-03_dp, 4.2174292864101881E+00_dp, &
         -7.0273735438846152E-01_dp, -3.1958858729380285E-01_dp, 1.1897201461891156E-01_dp, &
         8.9479712944561557E-02_dp, -2.3599626911379866E-02_dp, -5.9172184220296763E-02_dp, &
         2.3418908433583512E-02_dp, 3.0998875121130358E-02_dp, -2.6100693280547826E-02_dp, &
         -8.2991968088997443E-04_dp, 7.6967034888079256E-03_dp, -2.4440953982349913E-03_dp, &
         7.1283946095005790E+00_dp, -1.4935062082301014E+00_dp, -6.3792079328025064E-01_dp, &
         2.9134212583550634E-01_dp, 1.8609503658057067E-01_dp, -6.6849972955691214E-02_dp, &
         -1.2679209365908473E-01_dp, 5.7788861383825232E-02_dp, 7.1125670220970105E-02_dp, &
         -6.9155756511572325E-02_dp, 7.4772007009126021E-03_dp, 1.3632029714336589E-02_dp, &
         -4.9228934078391299E-03_dp, 1.2173813507842725E+01_dp, -3.1400013912848532E+00_dp, &
         -1.2483074964973158E+00_dp, 6.9485765409820177E-01_dp, 3.7611257997832859E-01_dp, &
         -1.7890856047138773E-01_dp, -2.6578443159351284E-01_dp, 1.4054639374393529E-01_dp, &
         1.6047068051743860E-01_dp, -1.7899907540313789E-01_dp, 4.0285268988641564E-02_dp, &
         2.0634397349215801E-02_dp, -9.3741744134483111E-03_dp, 2.1107642864276819E+01_dp, &
         -6.5843431773724026E+00_dp, -2.4050262723680573E+00_dp, 1.6350096374661267E+00_dp, &
         7.3996123655732526E-01_dp, -4.6373337696649547E-01_dp, -5.4747966879238541E-01_dp, &
         3.4034950017735888E-01_dp, 3.5845721258746249E-01_dp, -4.5829008373764241E-01_dp, &
         1.5001000994476824E-01_dp, 1.9940954618215254E-02_dp, -1.6426829440793874E-02_dp, &
         3.7266334989026802E+01_dp, -1.3867340387230591E+01_dp, -4.5696148656214710E+00_dp, &
         3.8316970122330818E+00_dp, 1.4108582397700304E+00_dp, -1.1818903323968282E+00_dp, &
         -1.1088915296439605E+00_dp, 8.2704736824941671E-01_dp, 7.9552904564683069E-01_dp, &
         -1.1705005968235462E+00_dp, 4.9161298485989080E-01_dp, -2.5476053684134296E-02_dp, &
         -2.4313596900675315E-02_dp, 6.7166643382165645E+01_dp, -2.9516187500335366E+01_dp, &
         -8.5451854440662061E+00_dp, 9.0099953442390479E+00_dp, 2.5665490024398552E+00_dp, &
         -2.9920833686411239E+00_dp, -2.1979231493373184E+00_dp, 2.0291778674850649E+00_dp, &
         1.7530383607299391E+00_dp, -2.9992079584428746E+00_dp, 1.5201611647974216E+00_dp, &
         -2.4963049256105524E-01_dp, -1.9576317627699253E-02_dp, 1.2391224737051206E+02_dp, &
         -6.3860904547818556E+01_dp, -1.5596243868028813E+01_dp, 2.1390445119414000E+01_dp, &
         4.2665327922213292E+00_dp, -7.5789334284051302E+00_dp, -4.2009119095859688E+00_dp, &
         5.0530481463321744E+00_dp, 3.8091681592822395E+00_dp, -7.7351382657783132E+00_dp, &
         4.5739670847276521E+00_dp, -1.1088234533884338E+00_dp, 5.7788156102331803E-02_dp, &
         2.3472890639646755E+02_dp, -1.4127311447474941E+02_dp, -2.7175909935628077E+01_dp, &
         5.1557286383939626E+01_dp, 5.6020406213249556E+00_dp, -1.9305968542309145E+01_dp, &
         -7.4481250666061314E+00_dp, 1.2828955385228877E+01_dp, 7.9993419405969064E+00_dp, &
         -2.0087543738369188E+01_dp, 1.3607203316826732E+01_dp, -4.1171412084629671E+00_dp, &
         4.3982127296754747E-01_dp, 4.5834886708487102E+02_dp, -3.2158236377162405E+02_dp, &
         -4.2500822021650286E+01_dp, 1.2681749970623152E+02_dp, 1.0331207737661403E+00_dp, &
         -4.9592638546541615E+01_dp, -1.0824736389255063E+01_dp, 3.3331519274645522E+01_dp, &
         1.5343959985063165E+01_dp, -5.2295332834515250E+01_dp, 4.0323068230649206E+01_dp, &
         -1.4240819192072191E+01_dp, 1.9740981942447056E+00_dp, 9.2713486252741257E+02_dp, &
         -7.5881923406724286E+02_dp, -4.6362583755929869E+01_dp, 3.1985382205090053E+02_dp, &
         -3.4656066891479590E+01_dp, -1.2831744385936224E+02_dp, -4.9055399255105359E+00_dp, &
         8.8808607339626235E+01_dp, 2.1618208770971705E+01_dp, -1.3449947060170553E+02_dp, &
         1.1892412424943839E+02_dp, -4.7580164322585055E+01_dp, 7.6650405494235141E+00_dp, &
         1.9552554903261487E+03_dp, -1.8730960487320954E+03_dp, 4.5978532346317706E+01_dp, &
         8.3029128979817312E+02_dp, -2.0862888435830430E+02_dp, -3.3108436117182572E+02_dp, &
         5.8912046909874640E+01_dp, 2.4205987690357750E+02_dp, -1.5528194709333421E+01_dp, &
         -3.2781660557594995E+02_dp, 3.4387197703428257E+02_dp, -1.5465208352189103E+02_dp, &
         2.7818133740736446E+01_dp, 4.3367077238709990E+03_dp, -4.8949990956050187E+03_dp, &
         6.3916888657632228E+02_dp, 2.2193372246794902E+03_dp, -9.7830000114740096E+02_dp, &
         -8.2164251683291820E+02_dp, 4.0343060366377574E+02_dp, 6.6365758988347898E+02_dp, &
         -3.5763798810336135E+02_dp, -6.5703543744238846E+02_dp, 9.2616781594453050E+02_dp, &
         -4.7750440193060797E+02_dp, 9.5070584811891436E+01_dp, 1.0241393178730643E+04_dp, &
         -1.3770619099685142E+04_dp, 3.6724807315668318E+03_dp, 6.0414514428452330E+03_dp, &
         -4.3826177232463688E+03_dp, -1.6912184393750047E+03_dp, 2.0017186887168280E+03_dp, &
         1.7046426163395738E+03_dp, -2.4055027646909889E+03_dp, -2.6196488354718417E+02_dp, &
         1.9094004144765859E+03_dp, -1.2635639019999310E+03_dp, 2.8729634415947623E+02_dp, &
         2.6233288034113684E+04_dp, -4.2749978247629486E+04_dp, 1.9270216857049596E+04_dp, &
         1.5749822825859681E+04_dp, -2.0118002433411439E+04_dp, 4.6708746343642822E+01_dp, &
         8.6670798148225003E+03_dp, 2.7018890777277093E+03_dp, -1.2539246090174649E+04_dp, &
         8.7032187871660572E+03_dp, -9.0580632943659941E+02_dp, -1.5024640379499738E+03_dp, &
         5.4315272376617247E+02_dp, 7.5229570283632551E+04_dp, -1.5238226013483954E+05_dp, &
         1.0963939977708698E+05_dp, 2.3699789741127177E+04_dp, -9.5127308763506648E+04_dp, &
         4.5638161082392988E+04_dp, 2.3668280641175519E+04_dp, -1.5344112755035001E+04_dp, &
         -4.0975894536356587E+04_dp, 6.5391139391724340E+04_dp, -4.3370233536422988E+04_dp, &
         1.4481627041640608E+04_dp, -1.9779208943771914E+03_dp, 2.6391498082474648E+05_dp, &
         -6.7578257857516059E+05_dp, 7.6143463474363345E+05_dp, -3.1336546372684737E+05_dp, &
         -2.7916162801665330E+05_dp, 5.0127439350212057E+05_dp, -3.0745998913516488E+05_dp, &
         5.4150294767096661E+04_dp, 9.3778431953998916E+03_dp, 4.7079371166969926E+04_dp, &
         -6.9604664413008839E+04_dp, 3.7411996446493475E+04_dp, -7.6881721943026469E+03_dp, &
         1.8732918832849885E+00_dp, -1.3367410054138293E-01_dp, -7.2352034213483338E-02_dp, &
         1.0909107202525322E-02_dp, 1.6759377488280897E-02_dp, -1.4463247410901762E-04_dp, &
         -1.0337062028004253E-02_dp, 2.4860738197468890E-03_dp, 4.3367529883595983E-03_dp, &
         -1.8252291072459487E-03_dp, -1.9460343870739156E-03_dp, 1.8559495947100581E-03_dp, &
         -4.7355776784216511E-04_dp, 2.3185521937218758E+00_dp, -2.3156055207094098E-01_dp, &
         -1.1794102738288226E-01_dp, 2.6687995574725933E-02_dp, 2.9649726476853933E-02_dp, &
         -2.8985963429409112E-03_dp, -1.8676999303934575E-02_dp, 5.4709623980301811E-03_dp, &
         8.5219863451357503E-03_dp, -4.9468255033507554E-03_dp, -2.4659023720406152E-03_dp, &
         3.0700450686352012E-03_dp, -8.3320990292621406E-04_dp, 3.3517744635877706E+00_dp, &
         -4.8229570440427993E-01_dp, -2.3009121135863886E-01_dp, 7.1478533606108507E-02_dp, &
         6.2277049326382221E-02_dp, -1.1852516550801632E-02_dp, -4.0220971604340308E-02_dp, &
         1.4012449854287748E-02_dp, 1.9865467670200296E-02_dp, -1.4442483293443300E-02_dp, &
         -2.8348238293451422E-03_dp, 5.9174029177587939E-03_dp, -1.7331773541936861E-03_dp, &
         5.3031945680151331E+00_dp, -1.0221259715172619E+00_dp, -4.5807387470277838E-01_dp, &
         1.8030806698751617E-01_dp, 1.3089607103055093E-01_dp, -3.6736851851184617E-02_dp, &
         -8.6877938204299371E-02_dp, 3.5099225799060947E-02_dp, 4.6044781064172931E-02_dp, &
         -3.9504731274502158E-02_dp, -5.0833388917190258E-04_dp, 1.1128529395928691E-02_dp, &
         -3.5796831935575950E-03_dp, 8.8118115386268983E+00_dp, -2.1430470190320698E+00_dp, &
         -8.9884915867193405E-01_dp, 4.3510992099274409E-01_dp, 2.6780537007744093E-01_dp, &
         -1.0216018744057363E-01_dp, -1.8352430459115163E-01_dp, 8.5686435047266335E-02_dp, &
         1.0440791497056170E-01_dp, -1.0353084282964660E-01_dp, 1.2964730094686697E-02_dp, &
         1.9176877419226003E-02_dp, -7.0877110142053322E-03_dp, 1.5082358533124399E+01_dp, &
         -4.4627150879444883E+00_dp, -1.7360210374491152E+00_dp, 1.0253983925629231E+00_dp, &
         5.3391870561552945E-01_dp, -2.6940028777485425E-01_dp, -3.8042443552829347E-01_dp, &
         2.0666296997581385E-01_dp, 2.3347998166148740E-01_dp, -2.6561413847670828E-01_dp, &
         6.3743999442031934E-02_dp, 2.7867301306668354E-02_dp, -1.3280682087156985E-02_dp, &
         2.6402399708460496E+01_dp, -9.3010729536877754E+00_dp, -3.3106893280192309E+00_dp, &
         2.3920970871922971E+00_dp, 1.0385304890158198E+00_dp, -6.9082350726466202E-01_dp, &
         -7.7708115592966898E-01_dp, 4.9760389999262650E-01_dp, 5.1852022401451869E-01_dp, &
         -6.7581533521215909E-01_dp, 2.2983801183532374E-01_dp, 2.3431816929989063E-02_dp, &
         -2.2860171257873246E-02_dp, 4.7194342446443912E+01_dp, -1.9531042894379219E+01_dp, &
         -6.2415251534557656E+00_dp, 5.5741252723118899E+00_dp, 1.9607368455071121E+00_dp, &
         -1.7473248563445485E+00_dp, -1.5646513259774506E+00_dp, 1.2048716833535995E+00_dp, &
         1.1473081165267458E+00_dp, -1.7190501090563362E+00_dp, 7.4009478335832446E-01_dp, &
         -4.9906011575939872E-02_dp, -3.2866315872047365E-02_dp, 8.6211948052474540E+01_dp, &
         -4.1554142631369068E+01_dp, -1.1600185792684515E+01_dp, 1.3061955480753966E+01_dp, &
         3.5320613718765403E+00_dp, -4.4004006892425798E+00_dp, -3.0887855024650599E+00_dp, &
         2.9501535623742106E+00_dp, 2.5271027814898610E+00_dp, -4.3943167718995682E+00_dp, &
         2.2632129132314391E+00_dp, -3.9043244143139921E-01_dp, -2.3099495422322626E-02_dp, &
         1.6127098938108156E+02_dp, -9.0039525643435795E+01_dp, -2.1053710089218743E+01_dp, &
         3.0952950172066647E+01_dp, 5.7949336008223513E+00_dp, -1.1107924269366810E+01_dp, &
         -5.8875195813116301E+00_dp, 7.3386204896080276E+00_dp, 5.5043800120782960E+00_dp, &
         -1.1322988619866953E+01_dp, 6.7597230603340916E+00_dp, -1.6667162278158267E+00_dp, &
         9.4315470904943952E-02_dp, 3.0980553856001376E+02_dp, -1.9972543431441864E+02_dp, &
         -3.6433629844988346E+01_dp, 7.4549189020146230E+01_dp, 7.3841116630708612E+00_dp, &
         -2.8240512722579421E+01_dp, -1.0420298428282244E+01_dp, 1.8622263811434856E+01_dp, &
         1.1642216142881098E+01_dp, -2.9427949723634473E+01_dp, 2.0016125436219472E+01_dp, &
         -6.0848036620178672E+00_dp, 6.5557403628564404E-01_dp, 6.1339991529169913E+02_dp, &
         -4.5611357812319022E+02_dp, -5.6273315070163605E+01_dp, 1.8335187300879113E+02_dp, &
         3.6753316352523047E-01_dp, -7.2514546876872799E+01_dp, -1.5147050715967145E+01_dp, &
         4.8376562193982153E+01_dp, 2.2769214359032059E+01_dp, -7.6885524107612937E+01_dp, &
         5.9199566480206009E+01_dp, -2.0866616660745976E+01_dp, 2.8826441383546944E+00_dp, &
         1.2575364403629890E+03_dp, -1.0794714024934246E+03_dp, -5.8881576395916468E+01_dp, &
         4.6255041181187573E+02_dp, -5.0617974041393261E+01_dp, -1.8799904813345231E+02_dp, &
         -7.1004252575506328E+00_dp, 1.2896162702553036E+02_dp, 3.4617522397077572E+01_dp, &
         -1.9966691145147081E+02_dp, 1.7499140865274785E+02_dp, -6.9518624039918890E+01_dp, &
         1.1110268324191310E+01_dp, 2.6853251507125306E+03_dp, -2.6690269182691513E+03_dp, &
         7.2463711925762794E+01_dp, 1.2014230806845137E+03_dp, -2.9322356189139873E+02_dp, &
         -4.8871486180307096E+02_dp, 8.0189444854946785E+01_dp, 3.5253998447998373E+02_dp, &
         -4.4344999858279763E+00_dp, -4.9961491943629647E+02_dp, 5.1183627581614923E+02_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c06 = (/-2.2712994716763473E+02_dp, &
         4.0385181038065113E+01_dp, 6.0197819306524152E+03_dp, -6.9664015206915119E+03_dp, &
         8.8000828312450119E+02_dp, 3.2181044508507798E+03_dp, -1.3451011011345620E+03_dp, &
         -1.2427951210535916E+03_dp, 5.4994829125785282E+02_dp, 9.7855377199584927E+02_dp, &
         -4.2361871901332773E+02_dp, -1.0938340303256634E+03_dp, 1.4307426913415468E+03_dp, &
         -7.1802169806496249E+02_dp, 1.4043890455439342E+02_dp, 1.4319961817475929E+04_dp, &
         -1.9465824088680063E+04_dp, 4.8714430805874626E+03_dp, 8.8417682243147774E+03_dp, &
         -5.8918891377734217E+03_dp, -2.8286954277399191E+03_dp, 2.7351780669919335E+03_dp, &
         2.6465333742965945E+03_dp, -3.0382947594497841E+03_dp, -1.2233930010289332E+03_dp, &
         3.3943700985724181E+03_dp, -2.0626667671078540E+03_dp, 4.5139108215934982E+02_dp, &
         3.6710729399870332E+04_dp, -5.9406878052428168E+04_dp, 2.4439529663225516E+04_dp, &
         2.4144745080947698E+04_dp, -2.6373382875748357E+04_dp, -3.2022704306036744E+03_dp, &
         1.2266276717430037E+04_dp, 5.7681700935135077E+03_dp, -1.6920132638314262E+04_dp, &
         8.0315452344738505E+03_dp, 2.9576472354183184E+03_dp, -4.0703120694405825E+03_dp, &
         1.1327630799673143E+03_dp, 1.0388253490761654E+05_dp, -2.0409751637993401E+05_dp, &
         1.2925285017231829E+05_dp, 5.4968049843609188E+04_dp, -1.2447761109417022E+05_dp, &
         3.2483521887329320E+04_dp, 4.7616175662189016E+04_dp, -4.6528610805370972E+03_dp, &
         -7.6633080388216375E+04_dp, 8.7935387930962621E+04_dp, -4.3615780121465235E+04_dp, &
         8.9761933428718621E+03_dp, -1.5583757123341093E+02_dp, 3.3941315862276591E+05_dp, &
         -8.2872883215502277E+05_dp, 7.9822211386825750E+05_dp, -8.4074217002807636E+04_dp, &
         -5.7957777085555380E+05_dp, 5.1317015480880765E+05_dp, 5.0080177118401307E+03_dp, &
         -2.1554743060333602E+05_dp, -6.8710719687172372E+04_dp, 3.7597911959563685E+05_dp, &
         -3.5898885567466140E+05_dp, 1.6051728031093234E+05_dp, -2.9622503438194970E+04_dp, &
         1.5375178283066794E+06_dp, -4.5263227540075248E+06_dp, 6.3451073165447786E+06_dp, &
         -4.6240857945008520E+06_dp, 2.4500245785436130E+05_dp, 3.5323273923941231E+06_dp, &
         -4.5233461773433862E+06_dp, 3.2676821863851110E+06_dp, -1.5689703219501805E+06_dp, &
         5.1341607784453162E+05_dp, -1.1143590110389760E+05_dp, 1.3969999513488921E+04_dp, &
         -4.2927092431353054E+02_dp, 2.3528218884138855E-01_dp, -6.1622809221375703E-03_dp, &
         -5.1621212301375691E-03_dp, 1.8256971283183122E-03_dp, -1.7757001927316855E-03_dp, &
         2.7587868056723900E-03_dp, -3.1055628572827037E-03_dp, 3.0788733937437561E-03_dp, &
         -2.9393177388430942E-03_dp, 2.4422551665436885E-03_dp, -1.5309508783932135E-03_dp, &
         6.2070307138328850E-04_dp, -1.2037481424199890E-04_dp, 7.6427517492838903E-01_dp, &
         -2.3301267122622669E-02_dp, -1.9379395197215672E-02_dp, 7.1363033962990527E-03_dp, &
         -6.7034010276187488E-03_dp, 1.0440998275629443E-02_dp, -1.1789250952983365E-02_dp, &
         1.1659409239080437E-02_dp, -1.1104556236899659E-02_dp, 9.2224304313982004E-03_dp, &
         -5.7812574625441581E-03_dp, 2.3431890007339553E-03_dp, -4.5412494246756713E-04_dp, &
         1.4884452945163158E+00_dp, -5.7780418368486818E-02_dp, -4.7524525258133422E-02_dp, &
         1.8574891035076632E-02_dp, -1.6590444169109660E-02_dp, 2.5921469668594304E-02_dp, &
         -2.9408901004723566E-02_dp, 2.8978898590561707E-02_dp, -2.7497730106109036E-02_dp, &
         2.2820362291516471E-02_dp, -1.4306212773108158E-02_dp, 5.7958440272391429E-03_dp, &
         -1.1221744170445015E-03_dp, 2.6105984893027929E+00_dp, -1.3223929038081408E-01_dp, &
         -1.0742680517105528E-01_dp, 4.4715325992489605E-02_dp, -3.7927286855369426E-02_dp, &
         5.9386145153437431E-02_dp, -6.7746283118929551E-02_dp, 6.6496265207841132E-02_dp, &
         -6.2831201303096690E-02_dp, 5.2098699546617035E-02_dp, -3.2665012485418560E-02_dp, &
         1.3227984141057456E-02_dp, -2.5585133273128336E-03_dp, 4.4702121379027533E+00_dp, &
         -2.9229427267092511E-01_dp, -2.3438486623605742E-01_dp, 1.0383070807863753E-01_dp, &
         -8.3847740273577026E-02_dp, 1.3135552153343474E-01_dp, -1.5073934324268146E-01_dp, &
         1.4739044102412518E-01_dp, -1.3863284292466410E-01_dp, 1.1484099496527858E-01_dp, &
         -7.2018264561787296E-02_dp, 2.9154836504191747E-02_dp, -5.6333601427675465E-03_dp, &
         7.6709266037709876E+00_dp, -6.3341405923166583E-01_dp, -5.0094369630952285E-01_dp, &
         2.3625700783289558E-01_dp, -1.8202632717122910E-01_dp, 2.8474545391588596E-01_dp, &
         -3.2890009162789791E-01_dp, 3.2037234643104090E-01_dp, -2.9982983600410501E-01_dp, &
         2.4809503450350234E-01_dp, -1.5563160367252610E-01_dp, 6.2990071438569953E-02_dp, &
         -1.2159286810559670E-02_dp, 1.3326570559389127E+01_dp, -1.3598533563961679E+00_dp, &
         -1.0592963255701000E+00_dp, 5.3297178906709941E-01_dp, -3.9225796386832035E-01_dp, &
         6.1125294976131017E-01_dp, -7.1120845081937767E-01_dp, 6.9013373426681579E-01_dp, &
         -6.4223671050719877E-01_dp, 5.3070213765084473E-01_dp, -3.3305766295054873E-01_dp, &
         1.3479152751847270E-01_dp, -2.5995336263354377E-02_dp, 2.3553113968366869E+01_dp, &
         -2.9176276962014072E+00_dp, -2.2345934256000408E+00_dp, 1.2036034488905603E+00_dp, &
         -8.4690962758329502E-01_dp, 1.3107060305534961E+00_dp, -1.5377199381631459E+00_dp, &
         1.4864614399378571E+00_dp, -1.3743037289490871E+00_dp, 1.1337221588786441E+00_dp, &
         -7.1192149066697252E-01_dp, 2.8815685634273214E-01_dp, -5.5524774908207841E-02_dp, &
         4.2483858960737933E+01_dp, -6.3034996022484222E+00_dp, -4.7356656620562889E+00_dp, &
         2.7433815095122966E+00_dp, -1.8473371186027441E+00_dp, 2.8284164031479886E+00_dp, &
         -3.3498320968923769E+00_dp, 3.2259356440170177E+00_dp, -2.9598546271753077E+00_dp, &
         2.4364391557222511E+00_dp, -1.5311624837953171E+00_dp, 6.1999991359852169E-01_dp, &
         -1.1937851432712819E-01_dp, 7.8435361033567304E+01_dp, -1.3807243700243477E+01_dp, &
         -1.0144752030644700E+01_dp, 6.3577896872198973E+00_dp, -4.1032215968515038E+00_dp, &
         6.1834653423448023E+00_dp, -7.4031369163085019E+00_dp, 7.1034525244070581E+00_dp, &
         -6.4588685672011907E+00_dp, 5.3015382264941371E+00_dp, -3.3350883832865144E+00_dp, &
         1.3515222561641431E+00_dp, -2.6009258886737091E-01_dp, 1.4870314684342992E+02_dp, &
         -3.0864834118508210E+01_dp, -2.2093258811557757E+01_dp, 1.5088268587806473E+01_dp, &
         -9.3558068710238516E+00_dp, 1.3783237313514368E+01_dp, -1.6707764004413356E+01_dp, &
         1.5978749580565010E+01_dp, -1.4371991456513806E+01_dp, 1.1750956854689353E+01_dp, &
         -7.4016805238109020E+00_dp, 3.0035875232982088E+00_dp, -5.7793452739336704E-01_dp, &
         2.9063992566159970E+02_dp, -7.0899303612086015E+01_dp, -4.9191236361965750E+01_dp, &
         3.6944732038741407E+01_dp, -2.2096224350928072E+01_dp, 3.1535478662551061E+01_dp, &
         -3.8769926784154620E+01_dp, 3.6981135439762390E+01_dp, -3.2826571132059875E+01_dp, &
         2.6693078989105775E+01_dp, -1.6838972567744609E+01_dp, 6.8482490585591780E+00_dp, &
         -1.3183464225945833E+00_dp, 5.8853959957153813E+02_dp, -1.6866367548444850E+02_dp, &
         -1.1264217046922096E+02_dp, 9.4139630285255294E+01_dp, -5.4641451607748081E+01_dp, &
         7.4621772797404674E+01_dp, -9.3205667243426078E+01_dp, 8.8778688875908884E+01_dp, &
         -7.7534194668658188E+01_dp, 6.2541347341539755E+01_dp, -3.9518334321274587E+01_dp, &
         1.6126307270781510E+01_dp, -3.1093032562817622E+00_dp, 1.2428008833245567E+03_dp, &
         -4.1947785521914795E+02_dp, -2.6701285426652703E+02_dp, 2.5226385512723408E+02_dp, &
         -1.4345141862564316E+02_dp, 1.8433585054421314E+02_dp, -2.3423907181798836E+02_dp, &
         2.2325867412280667E+02_dp, -1.9105874160160232E+02_dp, 1.5221467577400151E+02_dp, &
         -9.6307754715699645E+01_dp, 3.9500236966449748E+01_dp, -7.6415388902361059E+00_dp, &
         2.7608915592989856E+03_dp, -1.1041691712029990E+03_dp, -6.6001010192912725E+02_dp, &
         7.2065717805192821E+02_dp, -4.0742001889818755E+02_dp, 4.8140472788779994E+02_dp, &
         -6.2229993000320292E+02_dp, 5.9562070307754016E+02_dp, -4.9676841427948847E+02_dp, &
         3.8786350248878023E+02_dp, -2.4530498580103026E+02_dp, 1.0134423863055014E+02_dp, &
         -1.9728796542023890E+01_dp, 6.5330539597289735E+03_dp, -3.1289896695941670E+03_dp, &
         -1.7144513670222454E+03_dp, 2.2366769584555950E+03_dp, -1.2862337799451041E+03_dp, &
         1.3545526987521687E+03_dp, -1.7732225521374326E+03_dp, 1.7145546967607834E+03_dp, &
         -1.3838754531006168E+03_dp, 1.0431149756407765E+03_dp, -6.5542908071128045E+02_dp, &
         2.7332077381383078E+02_dp, -5.3786830655175223E+01_dp, 1.6776395337771821E+04_dp, &
         -9.7928996284343357E+03_dp, -4.7065652179644330E+03_dp, 7.7563596005434492E+03_dp, &
         -4.7062831481779740E+03_dp, 4.2440444961091507E+03_dp, -5.5256490153285331E+03_dp, &
         5.4499874871954971E+03_dp, -4.2216100644922999E+03_dp, 2.9760556872278821E+03_dp, &
         -1.8186260564964466E+03_dp, 7.6369653195481897E+02_dp, -1.5282174308169036E+02_dp, &
         4.8196026704304764E+04_dp, -3.5311796147047455E+04_dp, -1.3499860063436423E+04_dp, &
         3.1444987874104710E+04_dp, -2.1409313414419648E+04_dp, 1.5890405287246669E+04_dp, &
         -1.9335940318034241E+04_dp, 1.9685996849087067E+04_dp, -1.4546076258636267E+04_dp, &
         8.9035324606561226E+03_dp, -4.8468802353721112E+03_dp, 1.9729740359294485E+03_dp, &
         -4.0033508034806044E+02_dp, 1.6394238215154631E+05_dp, -1.5905119636599772E+05_dp, &
         -3.4636012557649119E+04_dp, 1.6159468155671351E+05_dp, -1.3873213467843784E+05_dp, &
         8.6609937728528355E+04_dp, -7.9189862325619106E+04_dp, 8.0498770341005089E+04_dp, &
         -5.6800098872495153E+04_dp, 2.4318200214846263E+04_dp, -5.3486374635222237E+03_dp, &
         1.4978633693132076E+02_dp, 1.3126224230101201E+02_dp, 7.5834075124608912E+05_dp, &
         -1.0864193654712576E+06_dp, 1.6418235171236837E+05_dp, 1.1876482137492767E+06_dp, &
         -1.7741895129516888E+06_dp, 1.3985011610223595E+06_dp, -6.4697226138982608E+05_dp, &
         1.5340670486444154E+04_dp, 3.7060552578940440E+05_dp, -5.1246119196262851E+05_dp, &
         4.1905388782645378E+05_dp, -2.0351458872316248E+05_dp, 4.4969428447952065E+04_dp, &
         1.9199342325621056E+00_dp, -5.2325699613614447E-02_dp, -4.3750513777635607E-02_dp, &
         1.5640015963402101E-02_dp, -1.5070969781007941E-02_dp, 2.3431356008326149E-02_dp, &
         -2.6397531067032395E-02_dp, 2.6153766119785420E-02_dp, -2.4952874087669753E-02_dp, &
         2.0730661937501563E-02_dp, -1.2995238603525089E-02_dp, 5.2682907391888040E-03_dp, &
         -1.0215189419958690E-03_dp, 2.4002587623301448E+00_dp, -9.2692544030490695E-02_dp, &
         -7.6334957658037619E-02_dp, 2.9644232823144580E-02_dp, -2.6611713402345558E-02_dp, &
         4.1582159829453708E-02_dp, -4.7147451242685809E-02_dp, 4.6475506392680971E-02_dp, &
         -4.4120374420174135E-02_dp, 3.6619045036844551E-02_dp, -2.2956089487110191E-02_dp, &
         9.3003902294016821E-03_dp, -1.8008841146529171E-03_dp, 3.5238883296645747E+00_dp, &
         -1.9744854381747892E-01_dp, -1.6011934282987575E-01_dp, 6.7242589745269768E-02_dp, &
         -5.6584064388948051E-02_dp, 8.8699180368368491E-02_dp, -1.0124759733146146E-01_dp, &
         9.9318595794023787E-02_dp, -9.3797388513328375E-02_dp, 7.7768501343562266E-02_dp, &
         -4.8758543054448447E-02_dp, 1.9743071622430480E-02_dp, -3.8179581523259494E-03_dp, &
         5.6716908861136890E+00_dp, -4.2699382481264164E-01_dp, -3.4141921609801612E-01_dp, &
         1.5331722645320270E-01_dp, -1.2238630321990675E-01_dp, 1.9196242245137163E-01_dp, &
         -2.2053214264272086E-01_dp, 2.1543129874600328E-01_dp, -2.0245400581809397E-01_dp, &
         1.6768202331832444E-01_dp, -1.0515504828494937E-01_dp, 4.2563447502959498E-02_dp, &
         -8.2219895159869709E-03_dp, 9.5923398264069970E+00_dp, -9.1351432511338171E-01_dp, &
         -7.2002060302993121E-01_dp, 3.4478262562537942E-01_dp, -2.6236445928893765E-01_dp, &
         4.1080033835374796E-01_dp, -4.7515478577796189E-01_dp, 4.6235013115468138E-01_dp/)
      REAL(KIND=dp), DIMENSION(400), PARAMETER :: c07 = (/-4.3224032563729953E-01_dp, &
         3.5758483946236974E-01_dp, -2.2431923396808751E-01_dp, 9.0777959192790053E-02_dp, &
         -1.7518061809443625E-02_dp, 1.6725498665042689E+01_dp, -1.9435939058523524E+00_dp, &
         -1.5082483802542699E+00_dp, 7.7123067496772879E-01_dp, -5.6050187357065773E-01_dp, &
         8.7386773329950440E-01_dp, -1.0184147160321777E+00_dp, 9.8713775000143567E-01_dp, &
         -9.1748347756413584E-01_dp, 7.5796037251628001E-01_dp, -4.7570197410379467E-01_dp, &
         1.9249696312506864E-01_dp, -3.7112410202830466E-02_dp, 2.9867921112026917E+01_dp, &
         -4.1467710457417182E+00_dp, -3.1626313649547511E+00_dp, 1.7323863084109623E+00_dp, &
         -1.2038693567355865E+00_dp, 1.8631804826928859E+00_dp, -2.1899213857816511E+00_dp, &
         2.1144847654909018E+00_dp, -1.9521830909580684E+00_dp, 1.6099731981821668E+00_dp, &
         -1.0110621403038136E+00_dp, 4.0919444918433101E-01_dp, -7.8821746413100843E-02_dp, &
         5.4567905169406032E+01_dp, -8.9359125078559316E+00_dp, -6.6829857757469942E+00_dp, &
         3.9380155360006719E+00_dp, -2.6201981403151646E+00_dp, 4.0098204947674123E+00_dp, &
         -4.7586752532202130E+00_dp, 4.5773565964791283E+00_dp, -4.1933229746486287E+00_dp, &
         3.4506609624332976E+00_dp, -2.1687844901543256E+00_dp, 8.7811842578345611E-01_dp, &
         -1.6902289559128800E-01_dp, 1.0213272364646357E+02_dp, -1.9568894490980483E+01_dp, &
         -1.4310663259886253E+01_dp, 9.1189354843768466E+00_dp, -5.8202843298748856E+00_dp, &
         8.7633901669854701E+00_dp, -1.0514412487755664E+01_dp, 1.0077333974037595E+01_dp, &
         -9.1480969346309582E+00_dp, 7.5063522206702977E+00_dp, -4.7227755759865069E+00_dp, &
         1.9137734403969708E+00_dp, -3.6817562821028288E-01_dp, 1.9634789478808003E+02_dp, &
         -4.3805003142660127E+01_dp, -3.1210983480765918E+01_dp, 2.1646060102378659E+01_dp, &
         -1.3290007437531367E+01_dp, 1.9559470489534739E+01_dp, -2.3760530475729208E+01_dp, &
         2.2699107507081777E+01_dp, -2.0384285092804305E+01_dp, 1.6661734343002301E+01_dp, &
         -1.0496738624690892E+01_dp, 4.2593912604244579E+00_dp, -8.1930517662403424E-01_dp, &
         3.8911243929907391E+02_dp, -1.0084567929859232E+02_dp, -6.9675590237727150E+01_dp, &
         5.3023392551490424E+01_dp, -3.1443103504058708E+01_dp, 4.4848086974289387E+01_dp, &
         -5.5248419076016638E+01_dp, 5.2644425241400143E+01_dp, -4.6665529615040874E+01_dp, &
         3.7940330501057126E+01_dp, -2.3939320376490915E+01_dp, 9.7353325974390632E+00_dp, &
         -1.8734799945633591E+00_dp, 7.9857425247754088E+02_dp, -2.4041614411913994E+02_dp, &
         -1.6005711637448368E+02_dp, 1.3505138321984933E+02_dp, -7.7813606689547342E+01_dp, &
         1.0635214708532304E+02_dp, -1.3307542579001566E+02_dp, 1.2661963840999799E+02_dp, &
         -1.1047991689937190E+02_dp, 8.9136639272186287E+01_dp, -5.6339722379846961E+01_dp, &
         2.2987396443453235E+01_dp, -4.4301925174203562E+00_dp, 1.7074190243146495E+03_dp, &
         -5.9849403848656630E+02_dp, -3.8056898159220981E+02_dp, 3.6095620533881140E+02_dp, &
         -2.0384611935344842E+02_dp, 2.6298454814944631E+02_dp, -3.3468311722540773E+02_dp, &
         3.1858350725758493E+02_dp, -2.7261989113275348E+02_dp, 2.1744286850398822E+02_dp, &
         -1.3765043100313622E+02_dp, 5.6438287133666776E+01_dp, -1.0910502440622219E+01_dp, &
         3.8334094340274683E+03_dp, -1.5723908956163691E+03_dp, -9.4281020491913853E+02_dp, &
         1.0242611119269077E+03_dp, -5.7428751500833221E+02_dp, 6.8547311891561162E+02_dp, &
         -8.8749541005121341E+02_dp, 8.4777812676065867E+02_dp, -7.0810279313566184E+02_dp, &
         5.5469907294877623E+02_dp, -3.5128294755457512E+02_dp, 1.4504011186174179E+02_dp, &
         -2.8198572213336920E+01_dp, 9.1361304192165644E+03_dp, -4.4226274015114341E+03_dp, &
         -2.4512781079839870E+03_dp, 3.1345626862048653E+03_dp, -1.7784413126977358E+03_dp, &
         1.9127074406180236E+03_dp, -2.5124489551522956E+03_dp, 2.4206332112005493E+03_dp, &
         -1.9613261391666301E+03_dp, 1.4910365526546971E+03_dp, -9.4077115625989279E+02_dp, &
         3.9209237872283859E+02_dp, -7.6978949238131889E+01_dp, 2.3476804836368778E+04_dp, &
         -1.3592633254062974E+04_dp, -6.7328872117040401E+03_dp, 1.0577880817867974E+04_dp, &
         -6.2497295482203444E+03_dp, 5.8566563032174154E+03_dp, -7.7179494949154041E+03_dp, &
         7.5607028303424113E+03_dp, -5.8984755977119921E+03_dp, 4.2500229634280931E+03_dp, &
         -2.6366823005050137E+03_dp, 1.1101454768091173E+03_dp, -2.2141619218110822E+02_dp, &
         6.6614484110205842E+04_dp, -4.7108388287965288E+04_dp, -1.9523320398634307E+04_dp, &
         4.0705057477236631E+04_dp, -2.6182350826008271E+04_dp, 2.0606177082035400E+04_dp, &
         -2.6289727061071782E+04_dp, 2.6497961110492906E+04_dp, -1.9762650337886316E+04_dp, &
         1.2821176945806279E+04_dp, -7.4626721995415519E+03_dp, 3.1351727273388092E+03_dp, &
         -6.3906389016555295E+02_dp, 2.1722047847281824E+05_dp, -1.9421988001893414E+05_dp, &
         -5.7005765806413925E+04_dp, 1.8870977976178579E+05_dp, -1.4215955667570615E+05_dp, &
         9.2469073856111092E+04_dp, -1.0223677196613803E+05_dp, 1.0679192105089303E+05_dp, &
         -7.6172821818506985E+04_dp, 3.9251745674129270E+04_dp, -1.6935502287943222E+04_dp, &
         6.1072726038873288E+03_dp, -1.2312049579639786E+03_dp, 8.7962812129087350E+05_dp, &
         -1.0583094001579208E+06_dp, -8.7726089363737672E+04_dp, 1.1620532119345395E+06_dp, &
         -1.1801685233582184E+06_dp, 7.0461861958712188E+05_dp, -4.8091128165205766E+05_dp, &
         4.5337411670429324E+05_dp, -3.0573011401678680E+05_dp, 6.8451145563412196E+04_dp, &
         6.0365054336862682E+04_dp, -5.2641113555194272E+04_dp, 1.3429156007824822E+04_dp, &
         5.4534124986841530E+06_dp, -9.8505910169624574E+06_dp, 4.1125439885084690E+06_dp, &
         1.0145049015186040E+07_dp, -2.1737139793601342E+07_dp, 2.1610941564617176E+07_dp, &
         -1.0288896330046380E+07_dp, -4.5951658076810353E+06_dp, 1.5022207247350259E+07_dp, &
         -1.7099728394099496E+07_dp, 1.2171622980394198E+07_dp, -5.3020153298938526E+06_dp, &
         1.0818379069385820E+06_dp, 2.3670956436095472E-01_dp, 8.5466042782728561E-10_dp, &
         -8.7706516429572629E-03_dp, 1.1337199418484344E-02_dp, -1.5571021209192253E-02_dp, &
         1.5569362340987348E-02_dp, -2.4778844274855861E-03_dp, -3.3946179190323028E-02_dp, &
         9.4071406905596136E-02_dp, -1.5071574728552392E-01_dp, 1.5621592696596148E-01_dp, &
         -9.7043862738537434E-02_dp, 2.7457399409936200E-02_dp, 7.6968000393308511E-01_dp, &
         3.2401450648662975E-09_dp, -3.3256925970225171E-02_dp, 4.2988865647330857E-02_dp, &
         -5.8756741104987492E-02_dp, 5.8295248976357440E-02_dp, -7.8860907725801094E-03_dp, &
         -1.3118813251172537E-01_dp, 3.5989835704601603E-01_dp, -5.7462950608239283E-01_dp, &
         5.9452625957445671E-01_dp, -3.6891620144798964E-01_dp, 1.0430150010072287E-01_dp, &
         1.5018763179292813E+00_dp, 8.0666558227795613E-09_dp, -8.2819934409362694E-02_dp, &
         1.0705542955448639E-01_dp, -1.4523496967030083E-01_dp, 1.4235570567914205E-01_dp, &
         -1.3911957597914196E-02_dp, -3.3604701889620353E-01_dp, 9.0831435978670716E-01_dp, &
         -1.4428049024652592E+00_dp, 1.4887036137719374E+00_dp, -9.2220669302792391E-01_dp, &
         2.6043051480110496E-01_dp, 2.6414098653587632E+00_dp, 1.8542770560249732E-08_dp, &
         -1.9043939699928683E-01_dp, 2.4616740044234692E-01_dp, -3.3120274807858702E-01_dp, &
         3.2019629965984886E-01_dp, -1.7514299065979757E-02_dp, -7.9626913862920790E-01_dp, &
         2.1188458241158257E+00_dp, -3.3470422670498117E+00_dp, 3.4433083070863675E+00_dp, &
         -2.1290723817341952E+00_dp, 6.0049052563642336E-01_dp, 4.5384820955717666E+00_dp, &
         4.1171481244663121E-08_dp, -4.2299027025076608E-01_dp, 5.4676919672569546E-01_dp, &
         -7.2930405536067833E-01_dp, 6.9476844166586404E-01_dp, -5.7289627094700328E-03_dp, &
         -1.8223453736507966E+00_dp, 4.7747854172686690E+00_dp, -7.5003110665766188E+00_dp, &
         7.6927313673657078E+00_dp, -4.7475264948136440E+00_dp, 1.3372664279873743E+00_dp, &
         7.8192504156388125E+00_dp, 8.9643714677238562E-08_dp, -9.2134105247858755E-01_dp, &
         1.1909512854185862E+00_dp, -1.5740081463467599E+00_dp, 1.4756520617606019E+00_dp, &
         6.3250121089543448E-02_dp, -4.0913969501850129E+00_dp, 1.0554863467153654E+01_dp, &
         -1.6484455241456860E+01_dp, 1.6854353882166951E+01_dp, -1.0380878084152998E+01_dp, &
         2.9200587692297555E+00_dp, 1.3645885165943300E+01_dp, 1.9343550021859286E-07_dp, &
         -1.9889558975248320E+00_dp, 2.5709795975545706E+00_dp, -3.3639874881302920E+00_dp, &
         3.0976636513324340E+00_dp, 3.1247903773887803E-01_dp, -9.1142796291772967E+00_dp, &
         2.3139806188864029E+01_dp, -3.5920374960036447E+01_dp, 3.6603598484231796E+01_dp, &
         -2.2496614716254765E+01_dp, 6.3188068542226405E+00_dp, 2.4240315177351366E+01_dp, &
         4.1735160476847653E-07_dp, -4.2934724986403312E+00_dp, 5.5498602800946042E+00_dp, &
         -7.1808885190927310E+00_dp, 6.4773412225047196E+00_dp, 1.0915405525625834E+00_dp, &
         -2.0338678533491709E+01_dp, 5.0778253166199057E+01_dp, -7.8310233286780999E+01_dp, &
         7.9509375168412404E+01_dp, -4.8752154654454642E+01_dp, 1.3671218890716160E+01_dp, &
         4.3973636879745392E+01_dp, 9.0731441379826837E-07_dp, -9.3394244687918899E+00_dp, &
         1.2072392500013727E+01_dp, -1.5423438922958889E+01_dp, 1.3579561953094331E+01_dp, &
         3.3846975946055915E+00_dp, -4.5839585399730254E+01_dp, 1.1242639699367048E+02_dp, &
         -1.7215228837760810E+02_dp, 1.7408628097597264E+02_dp, -1.0646507180414103E+02_dp, &
         2.9801105395511023E+01_dp, 8.1711262931542294E+01_dp, 2.0013821500952097E-06_dp, &
         -2.0615673482163860E+01_dp, 2.6648367171543072E+01_dp, -3.3552503990308622E+01_dp, &
         2.8697356550258888E+01_dp, 9.9848598970918747E+00_dp, -1.0512807173143871E+02_dp, &
         2.5297323747845883E+02_dp, -3.8433225027457678E+02_dp, 3.8690514140568553E+02_dp, &
         -2.3592263278951384E+02_dp, 6.5902326877061824E+01_dp, 1.5605879855050961E+02_dp, &
         4.5098443290657727E-06_dp, -4.6494277087601013E+01_dp, 6.0099711541289679E+01_dp, &
         -7.4393330371570187E+01_dp, 6.1409065971332915E+01_dp, 2.8984093656381923E+01_dp, &
         -2.4713866438037331E+02_dp, 5.8259718635376578E+02_dp, -8.7740192761657318E+02_dp, &
         8.7878792290311787E+02_dp, -5.3405825659980758E+02_dp, 1.4883009303091876E+02_dp, &
         3.0762435042193624E+02_dp, 1.0455746769612150E-05_dp, -1.0790746717317136E+02_dp, &
         1.3948392350653936E+02_dp, -1.6920163788653028E+02_dp, 1.3356024260004625E+02_dp, &
         8.4612560681060771E+01_dp, -6.0021504339910359E+02_dp, 1.3835888832168876E+03_dp, &
         -2.0632260865604612E+03_dp, 2.0544267400630752E+03_dp, -1.2436499579901790E+03_dp, &
         3.4561729478378913E+02_dp, 6.2919318362091553E+02_dp, 2.5144905278024844E-05_dp, &
         -2.5985287556874499E+02_dp, 3.3589217889390994E+02_dp, -3.9757888190288548E+02_dp, &
         2.9600943596616952E+02_dp, 2.5281006909384925E+02_dp, -1.5196535530174046E+03_dp, &
         3.4176113708246221E+03_dp, -5.0389952265142629E+03_dp, 4.9832566201452173E+03_dp, &
         -3.0026770436385773E+03_dp, 8.3169610915648207E+02_dp, 1.3446642485369698E+03_dp, &
         6.3355693696567805E-05_dp, -6.5587608695957738E+02_dp, 8.4780072445181315E+02_dp, &
         -9.7324930571061304E+02_dp, 6.6865816098059202E+02_dp, 7.8642740762815720E+02_dp, &
         -4.0561879015463091E+03_dp, 8.8735395997333635E+03_dp, -1.2911085013308702E+04_dp, &
         1.2663863439439963E+04_dp, -7.5877330082628787E+03_dp, 2.0931234638857791E+03_dp, &
         3.0315178012421479E+03_dp, 1.6945412072655205E-04_dp, -1.7583989533457666E+03_dp, &
         2.2729448301394605E+03_dp, -2.5079766439284267E+03_dp, 1.5298178702512521E+03_dp, &
         2.5970516090366850E+03_dp, -1.1583854023864891E+04_dp, 2.4553873694398699E+04_dp, &
         -3.5160886963347439E+04_dp, 3.4138635891960170E+04_dp, -2.0309554368209294E+04_dp, &
         5.5733463618645619E+03_dp, 7.3092031064017010E+03_dp, 4.9005183646288716E-04_dp, &
         -5.1024325120903641E+03_dp, 6.5955046072210243E+03_dp, -6.8957569912961944E+03_dp/)
      REAL(KIND=dp), DIMENSION(13, 40, 6), PARAMETER :: &
         coefdata = RESHAPE((/c01, c02, c03, c04, c05, c06, c07, c08/), (/13, 40, 6/))

      INTEGER                                            :: irange

      IF (Rc <= 6.0000000000000000E+01_dp) Rc = 6.0000000000000000E+01_dp
      IF (Rc >= 5.7633609771728516E+06_dp) Rc = 5.7633609771728516E+06_dp
      IF (Rc >= 6.0000000000000000E+01_dp .AND. Rc <= 1.9739889204664721E+02_dp) THEN
         irange = 1
         L_b = 6.0000000000000000E+01_dp
         U_b = 1.9739889204664721E+02_dp
      END IF
      IF (Rc >= 1.9739889204664721E+02_dp .AND. Rc <= 2.4626386425352434E+03_dp) THEN
         irange = 2
         L_b = 1.9739889204664721E+02_dp
         U_b = 2.4626386425352434E+03_dp
      END IF
      IF (Rc >= 2.4626386425352434E+03_dp .AND. Rc <= 3.3579125411394947E+04_dp) THEN
         irange = 3
         L_b = 2.4626386425352434E+03_dp
         U_b = 3.3579125411394947E+04_dp
      END IF
      IF (Rc >= 3.3579125411394947E+04_dp .AND. Rc <= 3.9047249374088546E+05_dp) THEN
         irange = 4
         L_b = 3.3579125411394947E+04_dp
         U_b = 3.9047249374088546E+05_dp
      END IF
      IF (Rc >= 3.9047249374088546E+05_dp .AND. Rc <= 1.9873406038870332E+06_dp) THEN
         irange = 5
         L_b = 3.9047249374088546E+05_dp
         U_b = 1.9873406038870332E+06_dp
      END IF
      IF (Rc >= 1.9873406038870332E+06_dp .AND. Rc <= 5.7633609771728516E+06_dp) THEN
         irange = 6
         L_b = 1.9873406038870332E+06_dp
         U_b = 5.7633609771728516E+06_dp
      END IF

      fit_coef(1:13, 1:40) = coefdata(1:13, 1:40, irange)

   END SUBROUTINE get_fit_coef_k20

END MODULE minimax_rpa
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate MP2 energy with laplace approach
!> \par History
!>      11.2012 created [Mauro Del Ben]
! **************************************************************************************************
MODULE mp2_laplace
!
   USE cp_fm_types,                     ONLY: cp_fm_get_info,&
                                              cp_fm_set_all,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_sum
   USE minimax,                         ONLY: get_minimax_coeff
   USE mp2_types,                       ONLY: integ_mat_buffer_type
   USE rpa_communication,               ONLY: fm_redistribute,&
                                              initialize_buffer,&
                                              release_buffer
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_laplace'

   PUBLIC :: laplace_minimax_approx

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param Emp2 ...
!> \param para_env ...
!> \param para_env_RPA ...
!> \param unit_nr ...
!> \param homo ...
!> \param virtual ...
!> \param dimen_RI ...
!> \param dimen_ia ...
!> \param Eigenval ...
!> \param num_integ_points ...
!> \param num_integ_group ...
!> \param color_rpa_group ...
!> \param fm_mat_S ...
!> \param fm_mat_Q_gemm ...
!> \param fm_mat_Q ...
!> \param homo_beta ...
!> \param virtual_beta ...
!> \param dimen_ia_beta ...
!> \param Eigenval_beta ...
!> \param fm_mat_S_beta ...
!> \param fm_mat_Q_gemm_beta ...
!> \param fm_mat_Q_beta ...
! **************************************************************************************************
   SUBROUTINE laplace_minimax_approx(Emp2, para_env, para_env_RPA, unit_nr, homo, virtual, dimen_RI, dimen_ia, Eigenval, &
                                     num_integ_points, num_integ_group, color_rpa_group, &
                                     fm_mat_S, fm_mat_Q_gemm, fm_mat_Q, &
                                     homo_beta, virtual_beta, dimen_ia_beta, Eigenval_beta, fm_mat_S_beta, &
                                     fm_mat_Q_gemm_beta, fm_mat_Q_beta)
      REAL(KIND=dp)                                      :: Emp2
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_RPA
      INTEGER                                            :: unit_nr, homo, virtual, dimen_RI, &
                                                            dimen_ia
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: num_integ_points, num_integ_group, &
                                                            color_rpa_group
      TYPE(cp_fm_type), POINTER                          :: fm_mat_S, fm_mat_Q_gemm, fm_mat_Q
      INTEGER, OPTIONAL                                  :: homo_beta, virtual_beta, dimen_ia_beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_mat_S_beta, fm_mat_Q_gemm_beta, &
                                                            fm_mat_Q_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'laplace_minimax_approx', &
         routineP = moduleN//':'//routineN

      INTEGER :: avirt, handle, i_global, iiB, iocc, j_global, jjB, jquad, my_num_dgemm_call, &
         ncol_local, nrow_local, number_of_rec, number_of_rec_beta, number_of_send, &
         number_of_send_beta
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: map_rec_size, map_rec_size_beta, &
                                                            map_send_size, map_send_size_beta, &
                                                            RPA_proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_size_source, local_size_source_beta
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: first_cycle, my_open_shell
      REAL(KIND=dp) :: actual_flop_rate, alpha_old, E_Range, Emax, Emax_beta, Emin, Emin_beta, &
         laplace_transf, my_flop_rate, t_end, t_start, trace_XX, w_old
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: aj, awj, wj
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_rec_beta, &
                                                            buffer_send, buffer_send_beta

      CALL timeset(routineN, handle)

      my_open_shell = .FALSE.
      IF (PRESENT(homo_beta) .AND. &
          PRESENT(virtual_beta) .AND. &
          PRESENT(dimen_ia_beta) .AND. &
          PRESENT(Eigenval_beta) .AND. &
          PRESENT(fm_mat_S_beta) .AND. &
          PRESENT(fm_mat_Q_gemm_beta) .AND. &
          PRESENT(fm_mat_Q_beta)) my_open_shell = .TRUE.

      ! Calculate weights and exponents with minimax approximation
      ALLOCATE (awj(2*num_integ_points))
      awj = 0.0_dp

      Emin = 2.0_dp*(Eigenval(homo+1)-Eigenval(homo))
      Emax = 2.0_dp*(MAXVAL(Eigenval)-MINVAL(Eigenval))
      IF (my_open_shell) THEN
         IF (homo_beta > 0) THEN
            Emin_beta = 2.0_dp*(Eigenval_beta(homo_beta+1)-Eigenval_beta(homo_beta))
            Emax_beta = 2.0_dp*(MAXVAL(Eigenval_beta)-MINVAL(Eigenval_beta))
            Emin = MIN(Emin, Emin_beta)
            Emax = MAX(Emax, Emax_beta)
         END IF
      END IF
      E_Range = Emax/Emin
      IF (E_Range < 2.0_dp) E_Range = 2.0_dp

      CALL get_minimax_coeff(num_integ_points, E_Range, awj)

      ALLOCATE (aj(num_integ_points))
      aj = 0.0_dp
      ALLOCATE (wj(num_integ_points))
      wj = 0.0_dp

      DO jquad = 1, num_integ_points
         aj(jquad) = awj(jquad)
         wj(jquad) = awj(jquad+num_integ_points)
      END DO

      DEALLOCATE (awj)

      IF (unit_nr > 0) THEN
         WRITE (UNIT=unit_nr, FMT="(T3,A,T66,F15.4)") &
            "MINIMAX| Range of the energy denominator:", E_Range
         WRITE (UNIT=unit_nr, FMT="(T3,A,T54,A,T72,A)") "MINIMAX| Parameters:", "Weights", "Exponents"
         DO jquad = 1, num_integ_points
            WRITE (UNIT=unit_nr, FMT="(T41,F20.10,F20.10)") wj(jquad), aj(jquad)
         END DO
         CALL m_flush(unit_nr)
      END IF

      ! scale the minimax parameters
      aj(:) = aj(:)/Emin
      wj(:) = wj(:)/Emin

      ! initialize buffer for matrix redistribution
      CALL initialize_buffer(fm_mat_Q_gemm, fm_mat_Q, RPA_proc_map, buffer_rec, buffer_send, &
                             number_of_rec, number_of_send, &
                             map_send_size, map_rec_size, local_size_source, para_env_RPA)
      IF (my_open_shell) THEN
         CALL initialize_buffer(fm_mat_Q_gemm_beta, fm_mat_Q_beta, RPA_proc_map, buffer_rec_beta, buffer_send_beta, &
                                number_of_rec_beta, number_of_send_beta, &
                                map_send_size_beta, map_rec_size_beta, local_size_source_beta, para_env_RPA)
      END IF

      Emp2 = 0.0_dp
      first_cycle = .TRUE.
      alpha_old = 0.0_dp
      w_old = 0.0_dp
      my_num_dgemm_call = 0
      my_flop_rate = 0.0_dp
      DO jquad = 1, num_integ_points
         IF (MODULO(jquad, num_integ_group) /= color_rpa_group) CYCLE

         !XXX ! copy fm_mat_S into fm_mat_G
         !XXX CALL cp_fm_to_fm(source=fm_mat_S,destination=fm_mat_G)

         !XXX ! get info of fm_mat_G
         !XXX CALL cp_fm_get_info(matrix=fm_mat_G,&
         !XXX                     nrow_local=nrow_local,&
         !XXX                     ncol_local=ncol_local,&
         !XXX                     row_indices=row_indices,&
         !XXX                     col_indices=col_indices)
         !XXX

         ! get info of fm_mat_S
         CALL cp_fm_get_info(matrix=fm_mat_S, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)

         ! update G matrix with the new value of w and a
         IF (first_cycle) THEN
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)

                  iocc = MAX(1, i_global-1)/virtual+1
                  avirt = i_global-(iocc-1)*virtual

                  laplace_transf = SQRT(wj(jquad))*EXP(Eigenval(iocc)*aj(jquad))*EXP(-Eigenval(avirt+homo)*aj(jquad))
                  laplace_transf = SQRT(laplace_transf)

                  fm_mat_S%local_data(iiB, jjB) = fm_mat_S%local_data(iiB, jjB)*laplace_transf

               END DO
            END DO
         ELSE
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)

                  iocc = MAX(1, i_global-1)/virtual+1
                  avirt = i_global-(iocc-1)*virtual

                  laplace_transf = SQRT(wj(jquad)/w_old)* &
                                   EXP(Eigenval(iocc)*(aj(jquad)-alpha_old))*EXP(-Eigenval(avirt+homo)*(aj(jquad)-alpha_old))
                  laplace_transf = SQRT(laplace_transf)

                  fm_mat_S%local_data(iiB, jjB) = fm_mat_S%local_data(iiB, jjB)*laplace_transf

               END DO
            END DO
         END IF

         t_start = m_walltime()
         CALL cp_gemm(transa="T", transb="N", m=dimen_RI, n=dimen_RI, k=dimen_ia, alpha=1.0_dp, &
                      matrix_a=fm_mat_S, matrix_b=fm_mat_S, beta=0.0_dp, &
                      matrix_c=fm_mat_Q_gemm)
         t_end = m_walltime()
         actual_flop_rate = 2.0_dp*REAL(dimen_ia, KIND=dp)*dimen_RI*REAL(dimen_RI, KIND=dp)/(MAX(0.01_dp, t_end-t_start))
         IF (para_env_RPA%mepos == 0) my_flop_rate = my_flop_rate+actual_flop_rate
         my_num_dgemm_call = my_num_dgemm_call+1

         ! copy/redistribute fm_mat_Q_gemm to fm_mat_Q
         CALL cp_fm_set_all(matrix=fm_mat_Q, alpha=0.0_dp)
         CALL fm_redistribute(fm_mat_Q_gemm, fm_mat_Q, RPA_proc_map, buffer_rec, buffer_send, &
                              number_of_send, &
                              map_send_size, map_rec_size, local_size_source, para_env_RPA)

         IF (my_open_shell) THEN
            !XXX  the same for the beta spin
            !XXX ! copy fm_mat_S into fm_mat_G
            !XXX CALL cp_fm_to_fm(source=fm_mat_S_beta,destination=fm_mat_G_beta)
            !XXX ! get info of fm_mat_G_beta
            !XXX CALL cp_fm_get_info(matrix=fm_mat_G_beta,&
            !XXX                     nrow_local=nrow_local,&
            !XXX                     ncol_local=ncol_local,&
            !XXX                     row_indices=row_indices,&
            !XXX                     col_indices=col_indices)
            !XXX

            ! the same for the beta spin
            CALL cp_fm_get_info(matrix=fm_mat_S_beta, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)
            ! update G matrix with the new value of w and a
            IF (first_cycle) THEN
               DO jjB = 1, ncol_local
                  j_global = col_indices(jjB)
                  DO iiB = 1, nrow_local
                     i_global = row_indices(iiB)

                     iocc = MAX(1, i_global-1)/virtual_beta+1
                     avirt = i_global-(iocc-1)*virtual_beta

                     laplace_transf = SQRT(wj(jquad))*EXP(Eigenval_beta(iocc)*aj(jquad))* &
                                      EXP(-Eigenval_beta(avirt+homo_beta)*aj(jquad))
                     laplace_transf = SQRT(laplace_transf)

                     fm_mat_S_beta%local_data(iiB, jjB) = fm_mat_S_beta%local_data(iiB, jjB)*laplace_transf

                  END DO
               END DO
            ELSE
               DO jjB = 1, ncol_local
                  j_global = col_indices(jjB)
                  DO iiB = 1, nrow_local
                     i_global = row_indices(iiB)

                     iocc = MAX(1, i_global-1)/virtual_beta+1
                     avirt = i_global-(iocc-1)*virtual_beta

                     laplace_transf = SQRT(wj(jquad)/w_old)* &
                                      EXP(Eigenval_beta(iocc)*(aj(jquad)-alpha_old))* &
                                      EXP(-Eigenval_beta(avirt+homo_beta)*(aj(jquad)-alpha_old))
                     laplace_transf = SQRT(laplace_transf)

                     fm_mat_S_beta%local_data(iiB, jjB) = fm_mat_S_beta%local_data(iiB, jjB)*laplace_transf

                  END DO
               END DO
            END IF

            t_start = m_walltime()
            CALL cp_gemm(transa="T", transb="N", m=dimen_RI, n=dimen_RI, k=dimen_ia_beta, alpha=1.0_dp, &
                         matrix_a=fm_mat_S_beta, matrix_b=fm_mat_S_beta, beta=0.0_dp, &
                         matrix_c=fm_mat_Q_gemm_beta)
            t_end = m_walltime()
            actual_flop_rate = 2.0_dp*REAL(dimen_ia_beta, KIND=dp)*dimen_RI*REAL(dimen_RI, KIND=dp)/(MAX(0.01_dp, t_end-t_start))
            IF (para_env_RPA%mepos == 0) my_flop_rate = my_flop_rate+actual_flop_rate
            my_num_dgemm_call = my_num_dgemm_call+1

            ! copy/redistribute fm_mat_Q_gemm to fm_mat_Q
            CALL cp_fm_set_all(matrix=fm_mat_Q_beta, alpha=0.0_dp)
            CALL fm_redistribute(fm_mat_Q_gemm_beta, fm_mat_Q_beta, RPA_proc_map, buffer_rec_beta, buffer_send_beta, &
                                 number_of_send_beta, &
                                 map_send_size_beta, map_rec_size_beta, local_size_source_beta, para_env_RPA)

         END IF

         ! get info of fm_mat_Q
         CALL cp_fm_get_info(matrix=fm_mat_Q, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)

         ! calcualte the trace of the product Q*Q
         trace_XX = 0.0_dp
         DO jjB = 1, ncol_local
            j_global = col_indices(jjB)
            DO iiB = 1, nrow_local
               i_global = row_indices(iiB)
               IF (my_open_shell) THEN
                  trace_XX = trace_XX+fm_mat_Q%local_data(iiB, jjB)*fm_mat_Q_beta%local_data(iiB, jjB)
               ELSE
                  trace_XX = trace_XX+fm_mat_Q%local_data(iiB, jjB)*fm_mat_Q%local_data(iiB, jjB)
               END IF
            END DO
         END DO

         Emp2 = Emp2-trace_XX

         ! save alpha, w and reset the first_cycle flag
         first_cycle = .FALSE.
         alpha_old = aj(jquad)
         w_old = wj(jquad)
      END DO

      CALL mp_sum(Emp2, para_env%group)

      IF (para_env_RPA%mepos == 0) my_flop_rate = my_flop_rate/REAL(MAX(my_num_dgemm_call, 1), KIND=dp)/1.0E9_dp
      CALL mp_sum(my_flop_rate, para_env%group)
      my_flop_rate = my_flop_rate/para_env%num_pe
      IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="(T3,A,T66,F15.2)") &
         "PERFORMANCE| PDGEMM flop rate (Gflops / MPI rank):", my_flop_rate

      ! release buffer
      CALL release_buffer(RPA_proc_map, buffer_rec, buffer_send, &
                          number_of_rec, number_of_send, &
                          map_send_size, map_rec_size, local_size_source)

      IF (my_open_shell) THEN
         CALL release_buffer(RPA_proc_map, buffer_rec_beta, buffer_send_beta, &
                             number_of_rec_beta, number_of_send_beta, &
                             map_send_size_beta, map_rec_size_beta, local_size_source_beta)
      END IF

      DEALLOCATE (aj)
      DEALLOCATE (wj)

      CALL timestop(handle)

   END SUBROUTINE laplace_minimax_approx

END MODULE mp2_laplace
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate RI-RPA energy
!> \par History
!>      06.2012 created [Mauro Del Ben]
!>      04.2015 GW routines added [Jan Wilhelm]
! **************************************************************************************************
MODULE rpa_ri_gpw
   USE bibliography,                    ONLY: DelBen2013,&
                                              DelBen2015,&
                                              cite_reference
   USE cp_blacs_env,                    ONLY: cp_blacs_env_create,&
                                              cp_blacs_env_release,&
                                              cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_deallocate_matrix_set, &
        cp_dbcsr_desymmetrize, cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_init_p, cp_dbcsr_p_type, &
        cp_dbcsr_release_p, cp_dbcsr_set, cp_dbcsr_type
   USE cp_fm_basic_linalg,              ONLY: cp_fm_invert,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_syrk,&
                                              cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_invert
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: &
        cp_fm_create, cp_fm_get_info, cp_fm_indxg2l, cp_fm_indxg2p, cp_fm_p_type, cp_fm_release, &
        cp_fm_set_all, cp_fm_set_element, cp_fm_to_fm, cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: ri_rpa_g0w0_crossing_newton,&
                                              ri_rpa_g0w0_crossing_none,&
                                              ri_rpa_g0w0_crossing_z_shot,&
                                              wfc_mm_style_gemm,&
                                              wfc_mm_style_syrk
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE machine,                         ONLY: m_flush,&
                                              m_memory,&
                                              m_walltime
   USE mathconstants,                   ONLY: pi
   USE message_passing,                 ONLY: &
        mp_bcast, mp_comm_split_direct, mp_irecv, mp_isend, mp_min, mp_sendrecv, mp_sum, mp_sync, &
        mp_wait, mp_waitall
   USE minimax,                         ONLY: check_range,&
                                              get_minimax_coeff
   USE minimax_rpa,                     ONLY: get_rpa_minimax_coeff
   USE mp2_laplace,                     ONLY: laplace_minimax_approx
   USE mp2_types,                       ONLY: integ_mat_buffer_type,&
                                              mp2_type
   USE rpa_communication,               ONLY: fm_redistribute,&
                                              initialize_buffer,&
                                              release_buffer
   USE rpa_im_time,                     ONLY: compute_fm_mat_Q_omega,&
                                              fill_fm_with_2d_array
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rpa_ri_gpw'

   PUBLIC :: rpa_ri_compute_en

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param Erpa ...
!> \param mp2_env ...
!> \param BIb_C ...
!> \param BIb_C_gw ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param color_sub ...
!> \param ends_array ...
!> \param ends_B_virtual ...
!> \param ends_B_all ...
!> \param sizes_array ...
!> \param sizes_B_virtual ...
!> \param sizes_B_all ...
!> \param starts_array ...
!> \param starts_B_virtual ...
!> \param starts_B_all ...
!> \param mo_coeff ...
!> \param Eigenval ...
!> \param nmo ...
!> \param homo ...
!> \param dimen_RI ...
!> \param gw_corr_lev_occ ...
!> \param gw_corr_lev_virt ...
!> \param unit_nr ...
!> \param do_ri_sos_laplace_mp2 ...
!> \param my_do_gw ...
!> \param BIb_C_beta ...
!> \param homo_beta ...
!> \param Eigenval_beta ...
!> \param ends_B_virtual_beta ...
!> \param sizes_B_virtual_beta ...
!> \param starts_B_virtual_beta ...
!> \param mo_coeff_beta ...
!> \param BIb_C_gw_beta ...
!> \param gw_corr_lev_occ_beta ...
!> \param gw_corr_lev_virt_beta ...
!> \param do_im_time ...
!> \param mat_munu ...
!> \param mat_B_munu ...
!> \param my_Lrows ...
!> \param eps_filter ...
! **************************************************************************************************
   SUBROUTINE rpa_ri_compute_en(Erpa, mp2_env, BIb_C, BIb_C_gw, para_env, para_env_sub, color_sub, &
                                ends_array, ends_B_virtual, ends_B_all, sizes_array, sizes_B_virtual, sizes_B_all, &
                                starts_array, starts_B_virtual, starts_B_all, mo_coeff, &
                                Eigenval, nmo, homo, dimen_RI, gw_corr_lev_occ, gw_corr_lev_virt, &
                                unit_nr, do_ri_sos_laplace_mp2, my_do_gw, &
                                BIb_C_beta, homo_beta, Eigenval_beta, &
                                ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta, &
                                mo_coeff_beta, BIb_C_gw_beta, gw_corr_lev_occ_beta, gw_corr_lev_virt_beta, &
                                do_im_time, mat_munu, mat_B_munu, my_Lrows, eps_filter)
      REAL(KIND=dp)                                      :: Erpa
      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C, BIb_C_gw
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER                                            :: color_sub
      INTEGER, ALLOCATABLE, DIMENSION(:) :: ends_array, ends_B_virtual, ends_B_all, sizes_array, &
         sizes_B_virtual, sizes_B_all, starts_array, starts_B_virtual, starts_B_all
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: nmo, homo, dimen_RI, gw_corr_lev_occ, &
                                                            gw_corr_lev_virt, unit_nr
      LOGICAL                                            :: do_ri_sos_laplace_mp2, my_do_gw
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :), &
         OPTIONAL                                        :: BIb_C_beta
      INTEGER, OPTIONAL                                  :: homo_beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      INTEGER, ALLOCATABLE, DIMENSION(:), OPTIONAL       :: ends_B_virtual_beta, &
                                                            sizes_B_virtual_beta, &
                                                            starts_B_virtual_beta
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: mo_coeff_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :), &
         OPTIONAL                                        :: BIb_C_gw_beta
      INTEGER, OPTIONAL                                  :: gw_corr_lev_occ_beta, &
                                                            gw_corr_lev_virt_beta
      LOGICAL, OPTIONAL                                  :: do_im_time
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mat_B_munu
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :), &
         OPTIONAL                                        :: my_Lrows
      REAL(KIND=dp), OPTIONAL                            :: eps_filter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rpa_ri_compute_en', &
         routineP = moduleN//':'//routineN

      INTEGER :: best_integ_group_size, best_num_integ_point, color_rpa_group, crossing_search, &
         dimen_ia, dimen_ia_beta, dimen_nm_gw, handle, handle2, i, ierr, iiB, &
         input_integ_group_size, integ_group_size, jjB, max_iter_fit, min_integ_group_size, &
         my_B_all_end, my_B_all_size, my_B_all_start, my_B_size, my_B_size_beta, my_B_virtual_end, &
         my_B_virtual_end_beta, my_B_virtual_start, my_B_virtual_start_beta, my_group_L_end, &
         my_group_L_size, my_group_L_start, my_homo_beta, my_ia_end, my_ia_end_beta, my_ia_size, &
         my_ia_size_beta, my_ia_start, my_ia_start_beta, my_nm_gw_end, my_nm_gw_size, &
         my_nm_gw_start
      INTEGER :: ncol_block_mat, ngroup, nrow_block_mat, num_integ_group, num_integ_points, &
         num_poles, pos_integ_group, virtual, virtual_beta
      INTEGER(KIND=int_8)                                :: mem
      INTEGER, ALLOCATABLE, DIMENSION(:) :: ends_ia, ends_ia_beta, ends_nm_gw, sizes_ia, &
         sizes_ia_beta, sizes_nm_gw, starts_ia, starts_ia_beta, starts_nm_gw, sub_proc_map
      LOGICAL                                            :: check_fit, do_minimax_quad, &
                                                            my_do_im_time, my_open_shell, &
                                                            skip_integ_group_opt
      REAL(KIND=dp) :: allowed_memory, avail_mem, E_Range, Emax, Emax_beta, Emin, Emin_beta, &
         ext_scaling, fermi_level_offset, mem_for_iaK, mem_for_QK, mem_min, mem_per_group, &
         mem_real, needed_mem, omega_max_fit, stop_crit
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIb_C_2D, BIb_C_2D_beta, BIb_C_2D_gw, &
                                                            BIb_C_2D_gw_beta
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fm_mat_Q_omega
      TYPE(cp_fm_type), POINTER :: fm_mat_Q, fm_mat_Q_beta, fm_mat_Q_gemm, fm_mat_Q_gemm_beta, &
         fm_mat_R_gw, fm_mat_R_gw_beta, fm_mat_R_gw_gemm, fm_mat_R_gw_gemm_beta, fm_mat_S, &
         fm_mat_S_beta, fm_mat_S_gw, fm_mat_S_gw_beta, fm_mo_coeff_occ, fm_mo_coeff_occ_beta, &
         fm_mo_coeff_virt, fm_mo_coeff_virt_beta, fm_scaled_dm_occ, fm_scaled_dm_virt
      TYPE(cp_para_env_type), POINTER                    :: para_env_RPA

      CALL timeset(routineN, handle)

      CALL cite_reference(DelBen2013)
      CALL cite_reference(DelBen2015)

      my_open_shell = .FALSE.
      IF (PRESENT(BIb_C_beta) .AND. &
          PRESENT(ends_B_virtual_beta) .AND. &
          PRESENT(sizes_B_virtual_beta) .AND. &
          PRESENT(starts_B_virtual_beta) .AND. &
          PRESENT(homo_beta) .AND. &
          PRESENT(Eigenval_beta)) my_open_shell = .TRUE.

      my_do_im_time = .FALSE.
      IF (PRESENT(do_im_time)) my_do_im_time = do_im_time

      virtual = nmo-homo

      IF (do_ri_sos_laplace_mp2) THEN
         num_integ_points = mp2_env%ri_laplace%n_quadrature
         input_integ_group_size = mp2_env%ri_laplace%integ_group_size

         ! check the range for the minimax approximation
         Emin = 2.0_dp*(Eigenval(homo+1)-Eigenval(homo))
         Emax = 2.0_dp*(MAXVAL(Eigenval)-MINVAL(Eigenval))
         IF (my_open_shell) THEN
            IF (homo_beta > 0) THEN
               Emin_beta = 2.0_dp*(Eigenval_beta(homo_beta+1)-Eigenval_beta(homo_beta))
               Emax_beta = 2.0_dp*(MAXVAL(Eigenval_beta)-MINVAL(Eigenval_beta))
               Emin = MIN(Emin, Emin_beta)
               Emax = MAX(Emax, Emax_beta)
            END IF
         END IF
         E_Range = Emax/Emin
         IF (E_Range < 2.0_dp) E_Range = 2.0_dp
         ierr = 0
         CALL check_range(num_integ_points, E_Range, ierr)
         IF (ierr /= 0) THEN
            jjB = num_integ_points-1
            DO iiB = 1, jjB
               num_integ_points = num_integ_points-1
               ierr = 0
               CALL check_range(num_integ_points, E_Range, ierr)
               IF (ierr == 0) EXIT
            END DO
         END IF
         IF(.NOT.(num_integ_points >= 1))CALL cp__a("rpa_ri_gpw.F",240)
      ELSE
         num_integ_points = mp2_env%ri_rpa%rpa_num_quad_points
         input_integ_group_size = mp2_env%ri_rpa%rpa_integ_group_size
         do_minimax_quad = mp2_env%ri_rpa%minimax_quad
         IF (do_minimax_quad .AND. num_integ_points > 20) THEN
            CALL cp_warn(cp__l("rpa_ri_gpw.F",246), &
                         "The required number of quadrature point exceeds the maximum possible in the "// &
                         "Minimax quadrature scheme. The number of quadrature point has been reset to 20.")
            num_integ_points = 20
         END IF
      END IF
      allowed_memory = mp2_env%mp2_memory

      ngroup = para_env%num_pe/para_env_sub%num_pe

      ! for imaginary time, we have other memory
      IF (my_do_im_time) THEN

         integ_group_size = ngroup
         best_num_integ_point = num_integ_points

      ELSE

         ! Calculate available memory and create integral group according to that
         ! mem_for_iaK is the memory needed for storing the 3 centre integrals
         mem_for_iaK = REAL(homo, KIND=dp)*virtual*dimen_RI*8.0_dp/(1024_dp**2)
         mem_for_QK = REAL(dimen_RI, KIND=dp)*dimen_RI*8.0_dp/(1024_dp**2)

         IF (my_open_shell) THEN
            virtual_beta = nmo-homo_beta
            mem_for_iaK = mem_for_iaK+REAL(homo_beta, KIND=dp)*virtual_beta*dimen_RI*8.0_dp/(1024_dp**2)
            mem_for_QK = mem_for_QK*2.0_dp
         END IF

         CALL m_memory(mem)
         mem_real = (mem+1024*1024-1)/(1024*1024)
         ! mp_min .... a hack.. it should be mp_max, but as it turns out, on some processes the previously freed memory (hfx)
         ! has not been given back to the OS yet.
         CALL mp_min(mem_real, para_env%group)

         mem_min = 2.0_dp*REAL(homo, KIND=dp)*MAXVAL(sizes_B_virtual)*MAXVAL(sizes_array)*8.0_dp/(1024**2)

         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T68,F9.2,A4)') 'RI_INFO| Minimum required memory per MPI process:', &
            mem_min, ' MiB'

         mem_real = allowed_memory-mem_real
         mem_real = MAX(mem_real, mem_min)

         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T68,F9.2,A4)') 'RI_INFO| Available memory per MPI process:', &
            mem_real, ' MiB'

         mem_per_group = mem_real*para_env_sub%num_pe

         needed_mem = mem_for_iaK*2.0_dp+mem_for_QK*3.0_dp

         ! here we try to find the best rpa/lapleca group size
         skip_integ_group_opt = .FALSE.

         IF (input_integ_group_size > 0) THEN
            IF (MOD(input_integ_group_size, para_env_sub%num_pe) == 0) THEN
               best_integ_group_size = input_integ_group_size/para_env_sub%num_pe
               IF (MOD(ngroup, best_integ_group_size) == 0) THEN
                  num_integ_group = ngroup/best_integ_group_size
                  IF ((num_integ_points > num_integ_group) .AND. MOD(num_integ_points, num_integ_group) == 0) THEN
                     best_num_integ_point = num_integ_points/num_integ_group
                     skip_integ_group_opt = .TRUE.
                  ELSE
                     IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'NUM_QUAD_POINTS not multiple of the number of INTEG_GROUP'
                  END IF
               ELSE
                  IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'Total number of groups not multiple of SIZE_INTEG_GROUP'
               END IF
            ELSE
               IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'SIZE_INTEG_GROUP not multiple of GROUP_SIZE'
            END IF
         END IF

         IF (.NOT. skip_integ_group_opt) THEN
            best_integ_group_size = ngroup
            best_num_integ_point = num_integ_points

            min_integ_group_size = MAX(1, ngroup/num_integ_points)

            integ_group_size = min_integ_group_size-1
            DO iiB = min_integ_group_size+1, ngroup
               integ_group_size = integ_group_size+1

               ! check that the ngroup is a multiple of  integ_group_size
               IF (MOD(ngroup, integ_group_size) /= 0) CYCLE

               ! check for memory
               avail_mem = integ_group_size*mem_per_group
               IF (avail_mem < needed_mem) CYCLE

               ! check the number of integration points is a multiple of the  number of integ_group
               num_integ_group = ngroup/integ_group_size
               IF (num_integ_points < num_integ_group) CYCLE
               IF (MOD(num_integ_points, num_integ_group) /= 0) CYCLE

               ! if all the test passed then decide
               IF ((num_integ_points/num_integ_group) < best_num_integ_point) THEN
                  best_num_integ_point = num_integ_points/num_integ_group
                  best_integ_group_size = integ_group_size
               END IF

            END DO
         END IF

         integ_group_size = best_integ_group_size

      END IF

      IF (unit_nr > 0) THEN
         IF (do_ri_sos_laplace_mp2) THEN
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "RI_INFO| Group size for laplace numerical integration:", integ_group_size*para_env_sub%num_pe
            WRITE (UNIT=unit_nr, FMT="(T3,A)") &
               "INTEG_INFO| MINIMAX approximation"
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "INTEG_INFO| Number of integration points:", num_integ_points
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "INTEG_INFO| Number of integration points per Laplace group:", best_num_integ_point
         ELSE
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "RI_INFO| Group size for frequency integration:", integ_group_size*para_env_sub%num_pe
            IF (do_minimax_quad) THEN
               WRITE (UNIT=unit_nr, FMT="(T3,A)") &
                  "INTEG_INFO| MINIMAX quadrature"
            ELSE
               WRITE (UNIT=unit_nr, FMT="(T3,A)") &
                  "INTEG_INFO| Clenshaw-Curtius quadrature"
            END IF
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "INTEG_INFO| Number of integration points:", num_integ_points
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "INTEG_INFO| Number of integration points per RPA group:", best_num_integ_point
         END IF
         CALL m_flush(unit_nr)
      END IF

      num_integ_group = ngroup/integ_group_size

      pos_integ_group = MOD(color_sub, integ_group_size)
      color_rpa_group = color_sub/integ_group_size

      my_group_L_size = sizes_array(color_sub)
      my_group_L_start = starts_array(color_sub)
      my_group_L_end = ends_array(color_sub)

      ! reordering is not necessary for imaginary time
      IF (.NOT. my_do_im_time) THEN

         my_B_virtual_start = starts_B_virtual(para_env_sub%mepos)
         my_B_virtual_end = ends_B_virtual(para_env_sub%mepos)
         my_B_size = sizes_B_virtual(para_env_sub%mepos)

         IF (my_open_shell) THEN
            my_B_virtual_start_beta = starts_B_virtual_beta(para_env_sub%mepos)
            my_B_virtual_end_beta = ends_B_virtual_beta(para_env_sub%mepos)
            my_B_size_beta = sizes_B_virtual_beta(para_env_sub%mepos)
            my_homo_beta = homo_beta
         ELSE
            my_B_virtual_start_beta = my_B_virtual_start
            my_B_virtual_end_beta = my_B_virtual_end
            my_B_size_beta = my_B_size
            my_homo_beta = homo
         END IF

      END IF ! not imaginary time

      IF (my_do_gw) THEN
         my_B_all_start = starts_B_all(para_env_sub%mepos)
         my_B_all_end = ends_B_all(para_env_sub%mepos)
         my_B_all_size = sizes_B_all(para_env_sub%mepos)

         ext_scaling = mp2_env%ri_g0w0%scaling
         num_poles = mp2_env%ri_g0w0%num_poles
         omega_max_fit = mp2_env%ri_g0w0%omega_max_fit
         stop_crit = mp2_env%ri_g0w0%stop_crit
         max_iter_fit = mp2_env%ri_g0w0%max_iter_fit
         check_fit = mp2_env%ri_g0w0%check_fit
         crossing_search = mp2_env%ri_g0w0%crossing_search
         fermi_level_offset = mp2_env%ri_g0w0%fermi_level_offset
      END IF

      CALL timeset(routineN//"_reorder", handle2)

      ! create the sub_proc_map
      ALLOCATE (sub_proc_map(-para_env_sub%num_pe:2*para_env_sub%num_pe-1))
      DO i = 0, para_env_sub%num_pe-1
         sub_proc_map(i) = i
         sub_proc_map(-i-1) = para_env_sub%num_pe-i-1
         sub_proc_map(para_env_sub%num_pe+i) = i
      END DO

      ! not necessary for imaginary time

      IF (my_do_im_time) THEN

         dimen_ia = homo*virtual

      ELSE

         ! reorder the local data in such a way to help the next stage of matrix creation
         ! now the data inside the group are divided into a ia x K matrix
         CALL calculate_BIb_C_2D(BIb_C_2D, BIb_C, para_env_sub, dimen_ia, homo, virtual, &
                                 sizes_ia, starts_ia, ends_ia, &
                                 sizes_B_virtual, starts_B_virtual, ends_B_virtual, &
                                 sub_proc_map, my_ia_size, my_ia_start, my_ia_end, my_group_L_size, &
                                 my_B_size, my_B_virtual_start)
         DEALLOCATE (BIb_C)
         DEALLOCATE (starts_B_virtual)
         DEALLOCATE (ends_B_virtual)
         DEALLOCATE (sizes_B_virtual)

         ! The same for open shell
         IF (my_open_shell) THEN
            CALL calculate_BIb_C_2D(BIb_C_2D_beta, BIb_C_beta, para_env_sub, &
                                    dimen_ia_beta, homo_beta, virtual_beta, &
                                    sizes_ia_beta, starts_ia_beta, ends_ia_beta, &
                                    sizes_B_virtual_beta, starts_B_virtual_beta, ends_B_virtual_beta, &
                                    sub_proc_map, my_ia_size_beta, my_ia_start_beta, my_ia_end_beta, my_group_L_size, &
                                    my_B_size_beta, my_B_virtual_start_beta)

            DEALLOCATE (BIb_C_beta)
            DEALLOCATE (starts_B_virtual_beta)
            DEALLOCATE (ends_B_virtual_beta)
            DEALLOCATE (sizes_B_virtual_beta)

         END IF

      END IF

      ! in the GW case, BIb_C_2D_gw is an nm x K matrix, with n: number of corr GW levels, m=nmo
      IF (my_do_gw) THEN
         CALL calculate_BIb_C_2D(BIb_C_2D_gw, BIb_C_gw, para_env_sub, dimen_nm_gw, &
                                 gw_corr_lev_occ+gw_corr_lev_virt, nmo, &
                                 sizes_nm_gw, starts_nm_gw, ends_nm_gw, &
                                 sizes_B_all, starts_B_all, ends_B_all, &
                                 sub_proc_map, my_nm_gw_size, my_nm_gw_start, my_nm_gw_end, my_group_L_size, &
                                 my_B_all_size, my_B_all_start)

         ! The same for open shell
         IF (my_open_shell) THEN
            DEALLOCATE (sizes_nm_gw, starts_nm_gw, ends_nm_gw)
            CALL calculate_BIb_C_2D(BIb_C_2D_gw_beta, BIb_C_gw_beta, para_env_sub, dimen_nm_gw, &
                                    gw_corr_lev_occ+gw_corr_lev_virt, nmo, &
                                    sizes_nm_gw, starts_nm_gw, ends_nm_gw, &
                                    sizes_B_all, starts_B_all, ends_B_all, &
                                    sub_proc_map, my_nm_gw_size, my_nm_gw_start, my_nm_gw_end, my_group_L_size, &
                                    my_B_all_size, my_B_all_start)
            DEALLOCATE (BIb_C_gw_beta)
         END IF

         DEALLOCATE (BIb_C_gw)
         DEALLOCATE (starts_B_all)
         DEALLOCATE (ends_B_all)
         DEALLOCATE (sizes_B_all)

      END IF

      CALL timestop(handle2)

      ! now create the matrices needed for the calculation, Q, S and G
      ! Q and G will have omega dependence
      CALL create_integ_mat(BIb_C_2D, para_env, para_env_sub, color_sub, ngroup, integ_group_size, &
                            dimen_RI, dimen_ia, dimen_ia, color_rpa_group, &
                            mp2_env%block_size_row, mp2_env%block_size_col, unit_nr, &
                            my_ia_size, my_ia_start, my_ia_end, &
                            my_group_L_size, my_group_L_start, my_group_L_end, &
                            para_env_RPA, fm_mat_S, fm_mat_Q_gemm, fm_mat_Q, &
                            nrow_block_mat, ncol_block_mat, &
                            do_im_time=my_do_im_time, fm_scaled_dm_occ=fm_scaled_dm_occ, &
                            fm_scaled_dm_virt=fm_scaled_dm_virt, mo_coeff=mo_coeff, &
                            fm_mo_coeff_occ=fm_mo_coeff_occ, &
                            fm_mo_coeff_virt=fm_mo_coeff_virt, &
                            nmo=nmo, homo=homo, num_integ_points=num_integ_points, &
                            fm_mat_Q_omega=fm_mat_Q_omega, &
                            do_gw=my_do_gw, mat_B_munu=mat_B_munu)

      IF (my_open_shell) THEN

         ! for imaginary time, we only have to build beta mo coefficients
         IF (my_do_im_time) THEN

            CALL create_occ_virt_mo_coeffs(fm_mo_coeff_occ_beta, fm_mo_coeff_virt_beta, mo_coeff_beta, &
                                           nmo, homo)

            ! for RPA with imaginary frequency, we have to build the same matrices for beta as for alpha
         ELSE

            CALL create_integ_mat(BIb_C_2D_beta, para_env, para_env_sub, color_sub, ngroup, integ_group_size, &
                                  dimen_RI, dimen_ia_beta, dimen_ia_beta, color_rpa_group, &
                                  mp2_env%block_size_row, mp2_env%block_size_col, unit_nr, &
                                  my_ia_size_beta, my_ia_start_beta, my_ia_end_beta, &
                                  my_group_L_size, my_group_L_start, my_group_L_end, &
                                  para_env_RPA, fm_mat_S_beta, fm_mat_Q_gemm_beta, fm_mat_Q_beta, &
                                  nrow_block_mat, ncol_block_mat, &
                                  .TRUE., fm_mat_Q%matrix_struct%context)

         END IF

      END IF

      ! for GW, we need other matrix fm_mat_S
      IF (my_do_gw) THEN
         CALL create_integ_mat(BIb_C_2D_gw, para_env, para_env_sub, color_sub, ngroup, integ_group_size, &
                               dimen_RI, dimen_nm_gw, dimen_ia, color_rpa_group, &
                               mp2_env%block_size_row, mp2_env%block_size_col, unit_nr, &
                               my_nm_gw_size, my_nm_gw_start, my_nm_gw_end, &
                               my_group_L_size, my_group_L_start, my_group_L_end, &
                               para_env_RPA, fm_mat_S_gw, fm_mat_R_gw_gemm, fm_mat_R_gw, &
                               nrow_block_mat, ncol_block_mat, &
                               .TRUE., fm_mat_Q%matrix_struct%context, fm_mat_Q%matrix_struct%context)

         ! for GW, we don't need fm_mat_R_gw_gemm (in contrast to RPA)
         CALL cp_fm_release(fm_mat_R_gw_gemm)

         IF (my_open_shell) THEN
            CALL create_integ_mat(BIb_C_2D_gw_beta, para_env, para_env_sub, color_sub, ngroup, integ_group_size, &
                                  dimen_RI, dimen_nm_gw, dimen_ia, color_rpa_group, &
                                  mp2_env%block_size_row, mp2_env%block_size_col, unit_nr, &
                                  my_nm_gw_size, my_nm_gw_start, my_nm_gw_end, &
                                  my_group_L_size, my_group_L_start, my_group_L_end, &
                                  para_env_RPA, fm_mat_S_gw_beta, fm_mat_R_gw_gemm_beta, fm_mat_R_gw_beta, &
                                  nrow_block_mat, ncol_block_mat, &
                                  .TRUE., fm_mat_Q%matrix_struct%context, fm_mat_Q%matrix_struct%context, .TRUE.)

         END IF

      END IF

      IF (my_open_shell) THEN

         IF (do_ri_sos_laplace_mp2) THEN
            ! go with laplace MINIMAX MP2
            CALL laplace_minimax_approx(Erpa, para_env, para_env_RPA, unit_nr, homo, virtual, dimen_RI, dimen_ia, Eigenval, &
                                        num_integ_points, num_integ_group, color_rpa_group, &
                                        fm_mat_S, fm_mat_Q_gemm, fm_mat_Q, &
                                        homo_beta, virtual_beta, dimen_ia_beta, Eigenval_beta, fm_mat_S_beta, &
                                        fm_mat_Q_gemm_beta, fm_mat_Q_beta)
         ELSE
            ! go with clenshaw-curtius/minimax quadrature
            CALL rpa_num_int(Erpa, mp2_env, para_env, para_env_RPA, para_env_sub, unit_nr, &
                             homo, virtual, dimen_RI, dimen_ia, dimen_nm_gw, &
                             Eigenval, num_integ_points, num_integ_group, color_rpa_group, &
                             fm_mat_S, fm_mat_Q_gemm, fm_mat_Q, fm_mat_S_gw, fm_mat_R_gw, &
                             my_do_gw, gw_corr_lev_occ, gw_corr_lev_virt, num_poles, ext_scaling, omega_max_fit, &
                             stop_crit, check_fit, fermi_level_offset, crossing_search, &
                             max_iter_fit, &
                             mp2_env%ri_rpa%mm_style, do_minimax_quad, &
                             homo_beta, virtual_beta, dimen_ia_beta, Eigenval_beta, fm_mat_S_beta, &
                             fm_mat_Q_gemm_beta, fm_mat_Q_beta, fm_mat_S_gw_beta, &
                             gw_corr_lev_occ_beta, gw_corr_lev_virt_beta, &
                             do_im_time=my_do_im_time, fm_scaled_dm_occ=fm_scaled_dm_occ, &
                             fm_scaled_dm_virt=fm_scaled_dm_virt, &
                             fm_mo_coeff_occ=fm_mo_coeff_occ, &
                             fm_mo_coeff_virt=fm_mo_coeff_virt, &
                             fm_mat_Q_omega=fm_mat_Q_omega, &
                             mat_munu=mat_munu, mat_B_munu=mat_B_munu, my_Lrows=my_Lrows, &
                             my_group_L_size=my_group_L_size, my_group_L_start=my_group_L_start, &
                             ngroup=ngroup, eps_filter=eps_filter, sizes_array=sizes_array, &
                             fm_mo_coeff_occ_beta=fm_mo_coeff_occ_beta, &
                             fm_mo_coeff_virt_beta=fm_mo_coeff_virt_beta)
         END IF
      ELSE
         IF (do_ri_sos_laplace_mp2) THEN
            ! go with laplace MINIMAX MP2
            CALL laplace_minimax_approx(Erpa, para_env, para_env_RPA, unit_nr, homo, virtual, dimen_RI, dimen_ia, Eigenval, &
                                        num_integ_points, num_integ_group, color_rpa_group, &
                                        fm_mat_S, fm_mat_Q_gemm, fm_mat_Q)
         ELSE
            ! go with clenshaw-curtius/minimax quadrature
            ! here, we also do the quasi-particle-energy correction for G0W0
            CALL rpa_num_int(Erpa, mp2_env, para_env, para_env_RPA, para_env_sub, unit_nr, &
                             homo, virtual, dimen_RI, dimen_ia, dimen_nm_gw, &
                             Eigenval, num_integ_points, num_integ_group, color_rpa_group, &
                             fm_mat_S, fm_mat_Q_gemm, fm_mat_Q, fm_mat_S_gw, fm_mat_R_gw, &
                             my_do_gw, gw_corr_lev_occ, gw_corr_lev_virt, num_poles, ext_scaling, omega_max_fit, &
                             stop_crit, check_fit, fermi_level_offset, crossing_search, &
                             max_iter_fit, &
                             mp2_env%ri_rpa%mm_style, do_minimax_quad, &
                             do_im_time=my_do_im_time, fm_scaled_dm_occ=fm_scaled_dm_occ, &
                             fm_scaled_dm_virt=fm_scaled_dm_virt, &
                             fm_mo_coeff_occ=fm_mo_coeff_occ, &
                             fm_mo_coeff_virt=fm_mo_coeff_virt, &
                             fm_mat_Q_omega=fm_mat_Q_omega, &
                             mat_munu=mat_munu, mat_B_munu=mat_B_munu, my_Lrows=my_Lrows, &
                             my_group_L_size=my_group_L_size, my_group_L_start=my_group_L_start, &
                             ngroup=ngroup, eps_filter=eps_filter, sizes_array=sizes_array)
         END IF
      END IF

      IF (.NOT. my_do_im_time) THEN

         DEALLOCATE (sizes_ia)
         DEALLOCATE (starts_ia)
         DEALLOCATE (ends_ia)

      END IF

      DEALLOCATE (sub_proc_map)

      DEALLOCATE (sizes_array)
      DEALLOCATE (starts_array)
      DEALLOCATE (ends_array)

      CALL cp_para_env_release(para_env_RPA)

      IF (.NOT. my_do_im_time) THEN
         CALL cp_fm_release(fm_mat_S)
      END IF
      !XXX CALL cp_fm_release(fm_mat_G)
      CALL cp_fm_release(fm_mat_Q_gemm)
      CALL cp_fm_release(fm_mat_Q)
      IF (my_open_shell .AND. (.NOT. my_do_im_time)) THEN
         DEALLOCATE (sizes_ia_beta)
         DEALLOCATE (starts_ia_beta)
         DEALLOCATE (ends_ia_beta)
         CALL cp_fm_release(fm_mat_S_beta)
         !XXX CALL cp_fm_release(fm_mat_G_beta)
         CALL cp_fm_release(fm_mat_Q_gemm_beta)
         CALL cp_fm_release(fm_mat_Q_beta)
      END IF

      IF (my_do_gw) THEN
         DEALLOCATE (sizes_nm_gw)
         DEALLOCATE (starts_nm_gw)
         DEALLOCATE (ends_nm_gw)
         CALL cp_fm_release(fm_mat_S_gw)
         CALL cp_fm_release(fm_mat_R_gw)
         IF (my_open_shell) THEN
            CALL cp_fm_release(fm_mat_S_gw_beta)
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE rpa_ri_compute_en

! **************************************************************************************************
!> \brief reorder the local data in such a way to help the next stage of matrix creation;
!>        now the data inside the group are divided into a ia x K matrix (BIb_C_2D);
!>        Subroutine created to avoid massive double coding for GW
!> \param BIb_C_2D ...
!> \param BIb_C ...
!> \param para_env_sub ...
!> \param dimen_ia ...
!> \param homo ...
!> \param virtual ...
!> \param sizes_ia ...
!> \param starts_ia ...
!> \param ends_ia ...
!> \param sizes_B_virtual ...
!> \param starts_B_virtual ...
!> \param ends_B_virtual ...
!> \param sub_proc_map ...
!> \param my_ia_size ...
!> \param my_ia_start ...
!> \param my_ia_end ...
!> \param my_group_L_size ...
!> \param my_B_size ...
!> \param my_B_virtual_start ...
!> \author Jan Wilhelm, 03/2015
! **************************************************************************************************
   SUBROUTINE calculate_BIb_C_2D(BIb_C_2D, BIb_C, para_env_sub, dimen_ia, homo, virtual, &
                                 sizes_ia, starts_ia, ends_ia, &
                                 sizes_B_virtual, starts_B_virtual, ends_B_virtual, &
                                 sub_proc_map, my_ia_size, my_ia_start, my_ia_end, my_group_L_size, &
                                 my_B_size, my_B_virtual_start)

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIb_C_2D
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      INTEGER                                            :: dimen_ia, homo, virtual
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: sizes_ia, starts_ia, ends_ia, &
                                                            sizes_B_virtual, starts_B_virtual, &
                                                            ends_B_virtual, sub_proc_map
      INTEGER                                            :: my_ia_size, my_ia_start, my_ia_end, &
                                                            my_group_L_size, my_B_size, &
                                                            my_B_virtual_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_BIb_C_2D', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: occ_chunk = 128

      INTEGER :: ia_global, iiB, iproc, itmp(2), jjB, occ_high, occ_low, proc_receive, proc_send, &
         proc_shift, rec_B_size, rec_B_virtual_end, rec_B_virtual_start
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C_rec

      dimen_ia = homo*virtual

      ALLOCATE (sizes_ia(0:para_env_sub%num_pe-1))
      sizes_ia = 0
      ALLOCATE (starts_ia(0:para_env_sub%num_pe-1))
      starts_ia = 0
      ALLOCATE (ends_ia(0:para_env_sub%num_pe-1))
      ends_ia = 0

      DO iproc = 0, para_env_sub%num_pe-1
         itmp = get_limit(dimen_ia, para_env_sub%num_pe, iproc)
         starts_ia(iproc) = itmp(1)
         ends_ia(iproc) = itmp(2)
         sizes_ia(iproc) = itmp(2)-itmp(1)+1
      END DO

      my_ia_start = starts_ia(para_env_sub%mepos)
      my_ia_end = ends_ia(para_env_sub%mepos)
      my_ia_size = sizes_ia(para_env_sub%mepos)

      ! reorder data
      ALLOCATE (BIb_C_2D(my_ia_size, my_group_L_size))

!$OMP     PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,ia_global) &
!$OMP              SHARED(homo,my_B_size,virtual,my_B_virtual_start,my_ia_start,my_ia_end,BIb_C,BIb_C_2D,&
!$OMP              my_group_L_size)
      DO iiB = 1, homo
         DO jjB = 1, my_B_size
            ia_global = (iiB-1)*virtual+my_B_virtual_start+jjB-1
            IF (ia_global >= my_ia_start .AND. ia_global <= my_ia_end) THEN
               BIb_C_2D(ia_global-my_ia_start+1, 1:my_group_L_size) = BIb_C(1:my_group_L_size, jjB, iiB)
            END IF
         END DO
      END DO

      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = sub_proc_map(para_env_sub%mepos+proc_shift)
         proc_receive = sub_proc_map(para_env_sub%mepos-proc_shift)

         rec_B_size = sizes_B_virtual(proc_receive)
         rec_B_virtual_end = ends_B_virtual(proc_receive)
         rec_B_virtual_start = starts_B_virtual(proc_receive)

         ! do this in chunks to avoid high memory overhead  for both BIb_C_rec and buffers in mp_sendrecv
         ! TODO: fix this more cleanly with a rewrite sending only needed data etc.
         ! TODO: occ_chunk should presumably be precomputed so that messages are limited to e.g. 100MiB.
         ALLOCATE (BIb_C_rec(my_group_L_size, rec_B_size, MIN(homo, occ_chunk)))

         DO occ_low = 1, homo, occ_chunk
            occ_high = MIN(homo, occ_low+occ_chunk-1)
            CALL mp_sendrecv(BIb_C(:, :, occ_low:occ_high), proc_send, &
                             BIb_C_rec(:, :, 1:occ_high-occ_low+1), proc_receive, &
                             para_env_sub%group)
!$OMP          PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,ia_global) &
!$OMP                   SHARED(occ_low,occ_high,rec_B_size,virtual,rec_B_virtual_start,my_ia_start,my_ia_end,BIb_C_rec,BIb_C_2D,&
!$OMP                          my_group_L_size)
            DO iiB = occ_low, occ_high
               DO jjB = 1, rec_B_size
                  ia_global = (iiB-1)*virtual+rec_B_virtual_start+jjB-1
                  IF (ia_global >= my_ia_start .AND. ia_global <= my_ia_end) THEN
                     BIb_C_2D(ia_global-my_ia_start+1, 1:my_group_L_size) = BIb_C_rec(1:my_group_L_size, jjB, iiB-occ_low+1)
                  END IF
               END DO
            END DO
         ENDDO

         DEALLOCATE (BIb_C_rec)
      END DO

   END SUBROUTINE calculate_BIb_C_2D

! **************************************************************************************************
!> \brief ...
!> \param BIb_C_2D ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param color_sub ...
!> \param ngroup ...
!> \param integ_group_size ...
!> \param dimen_RI ...
!> \param dimen_ia ...
!> \param dimen_ia_for_block_size ...
!> \param color_rpa_group ...
!> \param ext_row_block_size ...
!> \param ext_col_block_size ...
!> \param unit_nr ...
!> \param my_ia_size ...
!> \param my_ia_start ...
!> \param my_ia_end ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
!> \param my_group_L_end ...
!> \param para_env_RPA ...
!> \param fm_mat_S ...
!> \param fm_mat_Q_gemm ...
!> \param fm_mat_Q ...
!> \param nrow_block_mat ...
!> \param ncol_block_mat ...
!> \param beta_case ...
!> \param blacs_env_ext ...
!> \param blacs_env_ext_S ...
!> \param do_gw_open_shell ...
!> \param do_im_time ...
!> \param fm_scaled_dm_occ ...
!> \param fm_scaled_dm_virt ...
!> \param mo_coeff ...
!> \param fm_mo_coeff_occ ...
!> \param fm_mo_coeff_virt ...
!> \param nmo ...
!> \param homo ...
!> \param num_integ_points ...
!> \param fm_mat_Q_omega ...
!> \param do_gw ...
!> \param mat_B_munu ...
! **************************************************************************************************
   SUBROUTINE create_integ_mat(BIb_C_2D, para_env, para_env_sub, color_sub, ngroup, integ_group_size, &
                               dimen_RI, dimen_ia, dimen_ia_for_block_size, color_rpa_group, &
                               ext_row_block_size, ext_col_block_size, unit_nr, &
                               my_ia_size, my_ia_start, my_ia_end, &
                               my_group_L_size, my_group_L_start, my_group_L_end, &
                               para_env_RPA, fm_mat_S, fm_mat_Q_gemm, fm_mat_Q, &
                               nrow_block_mat, ncol_block_mat, beta_case, &
                               blacs_env_ext, blacs_env_ext_S, do_gw_open_shell, &
                               do_im_time, fm_scaled_dm_occ, fm_scaled_dm_virt, mo_coeff, &
                               fm_mo_coeff_occ, fm_mo_coeff_virt, &
                               nmo, homo, &
                               num_integ_points, fm_mat_Q_omega, do_gw, &
                               mat_B_munu)

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIb_C_2D
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER :: color_sub, ngroup, integ_group_size, dimen_RI, dimen_ia, dimen_ia_for_block_size, &
         color_rpa_group, ext_row_block_size, ext_col_block_size, unit_nr, my_ia_size, &
         my_ia_start, my_ia_end, my_group_L_size, my_group_L_start, my_group_L_end
      TYPE(cp_para_env_type), POINTER                    :: para_env_RPA
      TYPE(cp_fm_type), POINTER                          :: fm_mat_S, fm_mat_Q_gemm, fm_mat_Q
      INTEGER                                            :: nrow_block_mat, ncol_block_mat
      LOGICAL, OPTIONAL                                  :: beta_case
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env_ext, blacs_env_ext_S
      LOGICAL, OPTIONAL                                  :: do_gw_open_shell, do_im_time
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_scaled_dm_occ, fm_scaled_dm_virt, &
                                                            mo_coeff, fm_mo_coeff_occ, &
                                                            fm_mo_coeff_virt
      INTEGER, OPTIONAL                                  :: nmo, homo, num_integ_points
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: fm_mat_Q_omega
      LOGICAL, OPTIONAL                                  :: do_gw
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mat_B_munu

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_integ_mat', &
         routineP = moduleN//':'//routineN

      INTEGER :: comm_exchange, comm_rpa, dummy_proc, end_col_block, end_row_block, grid_2D(2), &
         handle, handle2, handle3, i, i_global, i_local, i_sub, iiB, iii, iproc, iproc_col, &
         iproc_row, itmp(2), j_global, j_local, j_sub, jjB, mepos_in_RPA_group, my_num_col_blocks, &
         my_num_row_blocks, my_pcol, my_prow, mypcol, myprow, nblkcols_local, nblkcols_total, &
         nblkrows_local, nblkrows_total, ncol_block, ncol_local, nfullcols_local, nfullcols_total, &
         nfullrows_local, nfullrows_total, npcol, nprow, nrow_block, nrow_local, num_rec_cols, &
         number_of_rec, number_of_send, proc_receive, proc_receive_static
      INTEGER :: proc_send, proc_send_static, proc_shift, rec_counter, rec_ia_end, rec_ia_size, &
         rec_ia_start, rec_L_end, rec_L_size, rec_L_start, rec_pcol, rec_prow, ref_send_pcol, &
         ref_send_prow, row_col_proc_ratio, send_counter, send_pcol, send_prow, size_rec_buffer, &
         size_send_buffer, start_col_block, start_row_block, sub_sub_color
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iii_vet, index_col_rec, map_rec_size, &
                                                            map_send_size, req_send, RPA_proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :) :: blocks_ranges_col, blocks_ranges_row, grid_2_mepos, &
         grid_ref_2_send_pos, group_grid_2_mepos, mepos_2_grid, RPA_info_end, RPA_info_size, &
         RPA_info_start
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: block_distribution_M
      INTEGER, DIMENSION(:), POINTER :: col_blk_offset, col_blk_size, col_indices, local_cols, &
         local_rows, proc_col_dist, proc_row_dist, row_blk_offset, row_blk_size, row_indices
      LOGICAL                                            :: my_beta_case, my_blacs_ext, &
                                                            my_blacs_S_ext, my_do_gw, &
                                                            my_do_im_time, my_gw_open_shell
      REAL(KIND=dp)                                      :: part_ia, part_RI
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: BIb_C_rec
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env, blacs_env_Q
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_para_env_type), POINTER                    :: para_env_exchange
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_send

      CALL timeset(routineN, handle)

      my_beta_case = .FALSE.
      IF (PRESENT(beta_case)) my_beta_case = beta_case

      my_blacs_ext = .FALSE.
      IF (PRESENT(blacs_env_ext)) my_blacs_ext = .TRUE.

      my_blacs_S_ext = .FALSE.
      IF (PRESENT(blacs_env_ext_S)) my_blacs_S_ext = .TRUE.

      my_gw_open_shell = .FALSE.
      IF (PRESENT(do_gw_open_shell)) my_gw_open_shell = do_gw_open_shell

      my_do_im_time = .FALSE.
      IF (PRESENT(do_im_time)) my_do_im_time = do_im_time

      my_do_gw = .FALSE.
      IF (PRESENT(do_gw)) my_do_gw = do_gw

      ! create the RPA para_env
      IF (.NOT. my_beta_case) THEN
         CALL mp_comm_split_direct(para_env%group, comm_rpa, color_rpa_group)
         NULLIFY (para_env_RPA)
         CALL cp_para_env_create(para_env_RPA, comm_rpa)
      END IF

      ! create the RPA blacs env
      IF (my_blacs_S_ext) THEN
         NULLIFY (blacs_env)
         blacs_env => blacs_env_ext_S
         NULLIFY (fm_struct)
         CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=dimen_ia, &
                                  ncol_global=dimen_RI, para_env=para_env_RPA)
      ELSE
         NULLIFY (blacs_env)
         IF (para_env_RPA%num_pe > 1) THEN
            row_col_proc_ratio = dimen_ia_for_block_size/dimen_RI
            row_col_proc_ratio = MAX(1, row_col_proc_ratio)
            IF (.FALSE.) THEN
               iproc_row = MAX(INT(SQRT(REAL(para_env_RPA%num_pe*row_col_proc_ratio, KIND=dp))), 1)-1
               DO iproc = 1, para_env_RPA%num_pe
                  iproc_row = iproc_row+1
                  IF (MOD(para_env_RPA%num_pe, iproc_row) == 0) EXIT
               END DO
            ELSE
               iproc_row = MIN(MAX(INT(SQRT(REAL(para_env_RPA%num_pe*row_col_proc_ratio, KIND=dp))), 1), para_env_RPA%num_pe)+1
               DO iproc = 1, para_env_RPA%num_pe
                  iproc_row = iproc_row-1
                  IF (MOD(para_env_RPA%num_pe, iproc_row) == 0) EXIT
               END DO
            END IF

            iproc_col = para_env_RPA%num_pe/iproc_row
            grid_2D(1) = iproc_row
            grid_2D(2) = iproc_col
         ELSE
            grid_2D = 1
         END IF
         CALL cp_blacs_env_create(blacs_env=blacs_env, para_env=para_env_RPA, grid_2d=grid_2d)

         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "MATRIX_INFO| Number row processes:", grid_2D(1)
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "MATRIX_INFO| Number column processes:", grid_2D(2)
         END IF
         ! CALL cp_blacs_env_create(blacs_env=blacs_env, para_env=para_env_RPA)

         ! define the block_size for the row
         IF (ext_row_block_size > 0) THEN
            nrow_block_mat = ext_row_block_size
         ELSE
            nrow_block_mat = dimen_ia_for_block_size/grid_2D(1)/2
            nrow_block_mat = MAX(nrow_block_mat, 1)
         END IF

         ! define the block_size for the column
         IF (ext_col_block_size > 0) THEN
            ncol_block_mat = ext_col_block_size
         ELSE
            ncol_block_mat = dimen_RI/grid_2D(2)/2
            ncol_block_mat = MAX(ncol_block_mat, 1)
         END IF

         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "MATRIX_INFO| Row block size:", nrow_block_mat
            WRITE (UNIT=unit_nr, FMT="(T3,A,T75,i6)") &
               "MATRIX_INFO| Column block size:", ncol_block_mat
         END IF

         NULLIFY (fm_struct)
         ! CALL cp_fm_struct_create(fm_struct,context=blacs_env,nrow_global=dimen_ia,&
         !                          ncol_global=dimen_RI,para_env=para_env_RPA)
         CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=dimen_ia, &
                                  ncol_global=dimen_RI, para_env=para_env_RPA, &
                                  nrow_block=nrow_block_mat, ncol_block=ncol_block_mat, force_block=.TRUE.)

      END IF ! external blacs_env

      ! for imaginary time we do not need the fm_mat_S
      IF (.NOT. my_do_im_time) THEN

         ! create the S full matrix, that is the (ia|K) matrix with K colomns
         ! and homo*virtual rows
         NULLIFY (fm_mat_S)
         CALL cp_fm_create(fm_mat_S, fm_struct, name="fm_mat_S")

         CALL cp_fm_set_all(matrix=fm_mat_S, alpha=0.0_dp)

         ! fill the matrix
         CALL cp_fm_get_info(matrix=fm_mat_S, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices, &
                             nrow_block=nrow_block, &
                             ncol_block=ncol_block)
         myprow = fm_mat_S%matrix_struct%context%mepos(1)
         mypcol = fm_mat_S%matrix_struct%context%mepos(2)
         nprow = fm_mat_S%matrix_struct%context%num_pe(1)
         npcol = fm_mat_S%matrix_struct%context%num_pe(2)

         ! create the RPA proc_map
         ALLOCATE (RPA_proc_map(-para_env_RPA%num_pe:2*para_env_RPA%num_pe-1))
         RPA_proc_map = 0
         DO i = 0, para_env_RPA%num_pe-1
            RPA_proc_map(i) = i
            RPA_proc_map(-i-1) = para_env_RPA%num_pe-i-1
            RPA_proc_map(para_env_RPA%num_pe+i) = i
         END DO

         ! create the info array, first index: 1-> L info, 2-> ia info
         ALLOCATE (RPA_info_start(2, 0:para_env_RPA%num_pe-1))
         RPA_info_start = 0
         ALLOCATE (RPA_info_end(2, 0:para_env_RPA%num_pe-1))
         RPA_info_end = 0
         ALLOCATE (RPA_info_size(2, 0:para_env_RPA%num_pe-1))
         RPA_info_size = 0

         RPA_info_start(1, para_env_RPA%mepos) = my_group_L_start
         RPA_info_start(2, para_env_RPA%mepos) = my_ia_start

         RPA_info_end(1, para_env_RPA%mepos) = my_group_L_end
         RPA_info_end(2, para_env_RPA%mepos) = my_ia_end

         RPA_info_size(1, para_env_RPA%mepos) = my_group_L_size
         RPA_info_size(2, para_env_RPA%mepos) = my_ia_size

         CALL mp_sum(RPA_info_start, para_env_RPA%group)
         CALL mp_sum(RPA_info_end, para_env_RPA%group)
         CALL mp_sum(RPA_info_size, para_env_RPA%group)

         IF (.TRUE.) THEN
            CALL timeset(routineN//"_comm", handle2)
            ! new replication scheme
            ! 0) create array with processes positions
            CALL timeset(routineN//"_info", handle3)
            ALLOCATE (grid_2_mepos(0:nprow-1, 0:npcol-1))
            grid_2_mepos = 0
            ALLOCATE (mepos_2_grid(0:para_env_RPA%num_pe-1, 2))
            mepos_2_grid = 0

            grid_2_mepos(myprow, mypcol) = para_env_RPA%mepos
            mepos_2_grid(para_env_RPA%mepos, 1) = myprow
            mepos_2_grid(para_env_RPA%mepos, 2) = mypcol

            CALL mp_sum(grid_2_mepos, para_env_RPA%group)
            CALL mp_sum(mepos_2_grid, para_env_RPA%group)

            ! 1) loop over my local data and define a map for the proc to send data
            ALLOCATE (map_send_size(0:para_env_RPA%num_pe-1))
            map_send_size = 0
            DO jjB = my_group_L_start, my_group_L_end
               send_pcol = cp_fm_indxg2p(jjB, ncol_block, dummy_proc, &
                                         fm_mat_S%matrix_struct%first_p_pos(2), npcol)
               DO iiB = my_ia_start, my_ia_end
                  send_prow = cp_fm_indxg2p(iiB, nrow_block, dummy_proc, &
                                            fm_mat_S%matrix_struct%first_p_pos(1), nprow)
                  proc_send = grid_2_mepos(send_prow, send_pcol)
                  map_send_size(proc_send) = map_send_size(proc_send)+1
               END DO
            END DO

            ! 2) loop over my local data of fm_mat_S and define a map for the proc from which rec data
            ALLOCATE (map_rec_size(0:para_env_RPA%num_pe-1))
            map_rec_size = 0
            mepos_in_RPA_group = MOD(color_sub, integ_group_size)
            ALLOCATE (group_grid_2_mepos(0:para_env_sub%num_pe-1, 0:integ_group_size-1))
            group_grid_2_mepos = 0
            group_grid_2_mepos(para_env_sub%mepos, mepos_in_RPA_group) = para_env_RPA%mepos
            CALL mp_sum(group_grid_2_mepos, para_env_RPA%group)
            part_ia = FLOAT(dimen_ia)/FLOAT(para_env_sub%num_pe)
            part_RI = FLOAT(dimen_RI)/FLOAT(ngroup)

            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               ! check the group holding this element
               ! dirty way, if someone has a better idea ...
               rec_pcol = INT(FLOAT(j_global-1)/part_RI)
               rec_pcol = MAX(0, rec_pcol)
               rec_pcol = MIN(rec_pcol, ngroup-1)
               DO
                  itmp = get_limit(dimen_RI, ngroup, rec_pcol)
                  IF (j_global >= itmp(1) .AND. j_global <= itmp(2)) EXIT
                  IF (j_global < itmp(1)) rec_pcol = rec_pcol-1
                  IF (j_global > itmp(2)) rec_pcol = rec_pcol+1
               END DO
               ! if the group is not in the same RPA group cycle
               IF ((rec_pcol/integ_group_size) .NE. color_rpa_group) CYCLE
               ! convert global position to position into the RPA group
               rec_pcol = MOD(rec_pcol, integ_group_size)

               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  ! check the process in the group holding this element
                  rec_prow = INT(FLOAT(i_global-1)/part_ia)
                  rec_prow = MAX(0, rec_prow)
                  rec_prow = MIN(rec_prow, para_env_sub%num_pe-1)
                  DO
                     itmp = get_limit(dimen_ia, para_env_sub%num_pe, rec_prow)
                     IF (i_global >= itmp(1) .AND. i_global <= itmp(2)) EXIT
                     IF (i_global < itmp(1)) rec_prow = rec_prow-1
                     IF (i_global > itmp(2)) rec_prow = rec_prow+1
                  END DO

                  proc_receive = group_grid_2_mepos(rec_prow, rec_pcol)

                  map_rec_size(proc_receive) = map_rec_size(proc_receive)+1

               END DO ! i_global
            END DO ! j_global

            ! 3) check if the local data has to be stored in the new fm_mat_S
            IF (map_rec_size(para_env_RPA%mepos) > 0) THEN
               DO jjB = 1, ncol_local
                  j_global = col_indices(jjB)
                  IF (j_global >= my_group_L_start .AND. j_global <= my_group_L_end) THEN
                     DO iiB = 1, nrow_local
                        i_global = row_indices(iiB)
                        IF (i_global >= my_ia_start .AND. i_global <= my_ia_end) THEN
                           fm_mat_S%local_data(iiB, jjB) = BIb_C_2D(i_global-my_ia_start+1, j_global-my_group_L_start+1)
                        END IF
                     END DO
                  END IF
               END DO
            END IF
            CALL timestop(handle3)

            ! 4) reorder data in the send_buffer
            CALL timeset(routineN//"_buffer_s", handle3)
            number_of_send = 0
            DO proc_shift = 1, para_env_RPA%num_pe-1
               proc_send = RPA_proc_map(para_env_RPA%mepos+proc_shift)
               IF (map_send_size(proc_send) > 0) THEN
                  number_of_send = number_of_send+1
               END IF
            END DO

            ALLOCATE (buffer_send(number_of_send))

            ! this array given a pair (ref_send_prow,ref_send_pcol) returns
            ! the position in the buffer_send associated to that process
            ALLOCATE (grid_ref_2_send_pos(0:nprow-1, 0:npcol-1))
            grid_ref_2_send_pos = 0

            send_counter = 0
            DO proc_shift = 1, para_env_RPA%num_pe-1
               proc_send = RPA_proc_map(para_env_RPA%mepos+proc_shift)
               size_send_buffer = map_send_size(proc_send)

               IF (map_send_size(proc_send) > 0) THEN
                  send_counter = send_counter+1
                  ! prepare the sending buffer
                  ALLOCATE (buffer_send(send_counter)%msg(size_send_buffer))
                  buffer_send(send_counter)%msg = 0.0_dp
                  buffer_send(send_counter)%proc = proc_send

                  ref_send_prow = mepos_2_grid(proc_send, 1)
                  ref_send_pcol = mepos_2_grid(proc_send, 2)

                  grid_ref_2_send_pos(ref_send_prow, ref_send_pcol) = send_counter

                  ! iii=0
                  ! DO iiB=my_ia_start, my_ia_end
                  !   send_prow=cp_fm_indxg2p(iiB,nrow_block,dummy_proc,&
                  !                           fm_mat_S%matrix_struct%first_p_pos(1),nprow)
                  !   IF(send_prow.NE.ref_send_prow) CYCLE
                  !   DO jjB=my_group_L_start, my_group_L_end
                  !     send_pcol=cp_fm_indxg2p(jjB,ncol_block,dummy_proc,&
                  !                             fm_mat_S%matrix_struct%first_p_pos(2),npcol)
                  !     IF(send_pcol.NE.ref_send_pcol) CYCLE
                  !     iii=iii+1
                  !     buffer_send(send_counter)%msg(iii)=BIb_C_2D(iiB-my_ia_start+1,jjB-my_group_L_start+1)
                  !   END DO
                  ! END DO

               END IF

            END DO

            ! loop over the locally held data and fill the buffer_send
            ! for doing that we need an array index
            ALLOCATE (iii_vet(number_of_send))

            iii_vet = 0
            DO iiB = my_ia_start, my_ia_end
               send_prow = cp_fm_indxg2p(iiB, nrow_block, dummy_proc, &
                                         fm_mat_S%matrix_struct%first_p_pos(1), nprow)
               DO jjB = my_group_L_start, my_group_L_end
                  send_pcol = cp_fm_indxg2p(jjB, ncol_block, dummy_proc, &
                                            fm_mat_S%matrix_struct%first_p_pos(2), npcol)
                  ! we don't need to send to ourselves
                  IF (grid_2_mepos(send_prow, send_pcol) == para_env_RPA%mepos) CYCLE

                  send_counter = grid_ref_2_send_pos(send_prow, send_pcol)
                  iii_vet(send_counter) = iii_vet(send_counter)+1
                  iii = iii_vet(send_counter)
                  buffer_send(send_counter)%msg(iii) = BIb_C_2D(iiB-my_ia_start+1, jjB-my_group_L_start+1)
               END DO
            END DO

            DEALLOCATE (iii_vet)
            DEALLOCATE (grid_ref_2_send_pos)
            DEALLOCATE (BIb_C_2D)
            CALL timestop(handle3)

            ! 5) create the buffer for receive and post the irecv
            CALL timeset(routineN//"_isendrecv", handle3)
            number_of_rec = 0
            DO proc_shift = 1, para_env_RPA%num_pe-1
               proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)
               IF (map_rec_size(proc_receive) > 0) THEN
                  number_of_rec = number_of_rec+1
               END IF
            END DO

            ALLOCATE (buffer_rec(number_of_rec))

            rec_counter = 0
            DO proc_shift = 1, para_env_RPA%num_pe-1
               proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)
               size_rec_buffer = map_rec_size(proc_receive)

               IF (map_rec_size(proc_receive) > 0) THEN
                  rec_counter = rec_counter+1
                  ! prepare the buffer for receive
                  ALLOCATE (buffer_rec(rec_counter)%msg(size_rec_buffer))
                  buffer_rec(rec_counter)%msg = 0.0_dp
                  buffer_rec(rec_counter)%proc = proc_receive

                  ! post the receiving message
                  CALL mp_irecv(buffer_rec(rec_counter)%msg, proc_receive, para_env_RPA%group, buffer_rec(rec_counter)%msg_req)
               END IF
            END DO

            ! 6) send data
            ALLOCATE (req_send(number_of_send))
            send_counter = 0
            DO proc_shift = 1, para_env_RPA%num_pe-1
               proc_send = RPA_proc_map(para_env_RPA%mepos+proc_shift)
               IF (map_send_size(proc_send) > 0) THEN
                  send_counter = send_counter+1
                  CALL mp_isend(buffer_send(send_counter)%msg, proc_send, para_env_RPA%group, buffer_send(send_counter)%msg_req)
                  req_send(send_counter) = buffer_send(send_counter)%msg_req
               END IF
            END DO
            CALL timestop(handle3)

            ! 8) fill the fm_mat_S matrix
            CALL timeset(routineN//"_fill", handle3)
            ! In order to perform this step efficiently first we have to know the
            ! ranges of the blocks over which a given process hold its local data.
            ! Start with the rows ...
            my_num_row_blocks = 1
            DO iiB = 1, nrow_local-1
               IF (ABS(row_indices(iiB+1)-row_indices(iiB)) == 1) CYCLE
               my_num_row_blocks = my_num_row_blocks+1
            END DO
            ALLOCATE (blocks_ranges_row(2, my_num_row_blocks))
            blocks_ranges_row = 0
            blocks_ranges_row(1, 1) = row_indices(1)
            iii = 1
            DO iiB = 1, nrow_local-1
               IF (ABS(row_indices(iiB+1)-row_indices(iiB)) == 1) CYCLE
               iii = iii+1
               blocks_ranges_row(2, iii-1) = row_indices(iiB)
               blocks_ranges_row(1, iii) = row_indices(iiB+1)
            END DO
            blocks_ranges_row(2, my_num_row_blocks) = row_indices(MAX(nrow_local, 1))

            ! and columns
            my_num_col_blocks = 1
            DO jjB = 1, ncol_local-1
               IF (ABS(col_indices(jjB+1)-col_indices(jjB)) == 1) CYCLE
               my_num_col_blocks = my_num_col_blocks+1
            END DO
            ALLOCATE (blocks_ranges_col(2, my_num_col_blocks))
            blocks_ranges_col = 0
            blocks_ranges_col(1, 1) = col_indices(1)
            iii = 1
            DO jjB = 1, ncol_local-1
               IF (ABS(col_indices(jjB+1)-col_indices(jjB)) == 1) CYCLE
               iii = iii+1
               blocks_ranges_col(2, iii-1) = col_indices(jjB)
               blocks_ranges_col(1, iii) = col_indices(jjB+1)
            END DO
            blocks_ranges_col(2, my_num_col_blocks) = col_indices(MAX(ncol_local, 1))

            rec_counter = 0
            DO proc_shift = 1, para_env_RPA%num_pe-1
               proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)
               size_rec_buffer = map_rec_size(proc_receive)

               IF (map_rec_size(proc_receive) > 0) THEN
                  rec_counter = rec_counter+1

                  rec_L_size = RPA_info_size(1, proc_receive)
                  rec_L_start = RPA_info_start(1, proc_receive)
                  rec_L_end = RPA_info_end(1, proc_receive)
                  ! precompute the number of received columns and relative index
                  num_rec_cols = 0
                  DO jjB = 1, my_num_col_blocks
                     start_col_block = MAX(blocks_ranges_col(1, jjB), rec_L_start)
                     end_col_block = MIN(blocks_ranges_col(2, jjB), rec_L_end)
                     DO j_sub = start_col_block, end_col_block
                        num_rec_cols = num_rec_cols+1
                     END DO
                  END DO
                  ALLOCATE (index_col_rec(num_rec_cols))
                  index_col_rec = 0
                  iii = 0
                  DO jjB = 1, my_num_col_blocks
                     start_col_block = MAX(blocks_ranges_col(1, jjB), rec_L_start)
                     end_col_block = MIN(blocks_ranges_col(2, jjB), rec_L_end)
                     DO j_sub = start_col_block, end_col_block
                        iii = iii+1
                        j_local = cp_fm_indxg2l(j_sub, ncol_block, dummy_proc, &
                                                fm_mat_S%matrix_struct%first_p_pos(2), npcol)
                        index_col_rec(iii) = j_local
                     END DO
                  END DO

                  rec_ia_size = RPA_info_size(2, proc_receive)
                  rec_ia_start = RPA_info_start(2, proc_receive)
                  rec_ia_end = RPA_info_end(2, proc_receive)

                  ! wait for the message
                  CALL mp_wait(buffer_rec(rec_counter)%msg_req)

                  iii = 0
                  DO iiB = 1, my_num_row_blocks
                     start_row_block = MAX(blocks_ranges_row(1, iiB), rec_ia_start)
                     end_row_block = MIN(blocks_ranges_row(2, iiB), rec_ia_end)
                     DO i_sub = start_row_block, end_row_block
                        i_local = cp_fm_indxg2l(i_sub, nrow_block, dummy_proc, &
                                                fm_mat_S%matrix_struct%first_p_pos(1), nprow)
                        ! DO jjB=1, my_num_col_blocks
                        !   start_col_block=MAX(blocks_ranges_col(1,jjB),rec_L_start)
                        !   end_col_block=MIN(blocks_ranges_col(2,jjB),rec_L_end)
                        !   DO j_sub=start_col_block, end_col_block
                        !     j_local=cp_fm_indxg2l(j_sub,ncol_block,dummy_proc,&
                        !                           fm_mat_S%matrix_struct%first_p_pos(2),npcol)
                        !     iii=iii+1
                        !     fm_mat_S%local_data(i_local,j_local)=buffer_rec(rec_counter)%msg(iii)
                        !   END DO
                        ! END DO
                        DO jjB = 1, num_rec_cols
                           iii = iii+1
                           j_local = index_col_rec(jjB)
                           fm_mat_S%local_data(i_local, j_local) = buffer_rec(rec_counter)%msg(iii)
                        END DO
                     END DO
                  END DO

                  ! iii=0
                  ! DO iiB=rec_ia_start, rec_ia_end
                  !   rec_prow=cp_fm_indxg2p(iiB,nrow_block,dummy_proc,&
                  !                    fm_mat_S%matrix_struct%first_p_pos(1),nprow)
                  !   IF(rec_prow.NE.myprow) CYCLE
                  !   i_local=cp_fm_indxg2l(iiB,nrow_block,dummy_proc,&
                  !                         fm_mat_S%matrix_struct%first_p_pos(1),nprow)
                  !   DO jjB=rec_L_start, rec_L_end
                  !     rec_pcol=cp_fm_indxg2p(jjB,ncol_block,dummy_proc,&
                  !                      fm_mat_S%matrix_struct%first_p_pos(2),npcol)
                  !     IF(rec_pcol.NE.mypcol) CYCLE
                  !     j_local=cp_fm_indxg2l(jjB,ncol_block,dummy_proc,&
                  !                     fm_mat_S%matrix_struct%first_p_pos(2),npcol)
                  !     iii=iii+1
                  !     fm_mat_S%local_data(i_local,j_local)=buffer_rec(rec_counter)%msg(iii)
                  !   END DO
                  ! END DO

                  DEALLOCATE (buffer_rec(rec_counter)%msg)
                  DEALLOCATE (index_col_rec)

               END IF
            END DO
            DEALLOCATE (buffer_rec)

            DEALLOCATE (blocks_ranges_row)
            DEALLOCATE (blocks_ranges_col)

            CALL timestop(handle3)

            ! wait for all messeges to be sent
            CALL timeset(routineN//"_waitall", handle3)
            CALL mp_waitall(req_send(:))
            DO send_counter = 1, number_of_send
               DEALLOCATE (buffer_send(send_counter)%msg)
            END DO
            DEALLOCATE (buffer_send)
            CALL timestop(handle3)

            DEALLOCATE (group_grid_2_mepos)
            DEALLOCATE (map_send_size)
            DEALLOCATE (map_rec_size)
            DEALLOCATE (grid_2_mepos)
            DEALLOCATE (mepos_2_grid)

            CALL timestop(handle2)
         ELSE
            ! old inefficient replication scheme
            ! local data
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               IF (j_global >= my_group_L_start .AND. j_global <= my_group_L_end) THEN
                  DO iiB = 1, nrow_local
                     i_global = row_indices(iiB)
                     IF (i_global >= my_ia_start .AND. i_global <= my_ia_end) THEN
                        fm_mat_S%local_data(iiB, jjB) = BIb_C_2D(i_global-my_ia_start+1, j_global-my_group_L_start+1)
                     END IF
                  END DO
               END IF
            END DO

            proc_send_static = RPA_proc_map(para_env_RPA%mepos+1)
            proc_receive_static = RPA_proc_map(para_env_RPA%mepos-1)

            ! start collect data from other proc in the RPA group
            DO proc_shift = 1, para_env_RPA%num_pe-1
               proc_send = RPA_proc_map(para_env_RPA%mepos+proc_shift)
               proc_receive = RPA_proc_map(para_env_RPA%mepos-proc_shift)

               rec_L_size = RPA_info_size(1, proc_receive)
               rec_L_start = RPA_info_start(1, proc_receive)
               rec_L_end = RPA_info_end(1, proc_receive)

               rec_ia_size = RPA_info_size(2, proc_receive)
               rec_ia_start = RPA_info_start(2, proc_receive)
               rec_ia_end = RPA_info_end(2, proc_receive)

               ALLOCATE (BIb_C_rec(rec_ia_size, rec_L_size))
               BIb_C_rec = 0.0_dp

               CALL mp_sendrecv(BIb_C_2D, proc_send_static, &
                                BIb_C_rec, proc_receive_static, &
                                para_env_RPA%group)

               DO jjB = 1, ncol_local
                  j_global = col_indices(jjB)
                  IF (j_global >= rec_L_start .AND. j_global <= rec_L_end) THEN
                     DO iiB = 1, nrow_local
                        i_global = row_indices(iiB)
                        IF (i_global >= rec_ia_start .AND. i_global <= rec_ia_end) THEN
                           fm_mat_S%local_data(iiB, jjB) = BIb_C_rec(i_global-rec_ia_start+1, j_global-rec_L_start+1)
                        END IF
                     END DO
                  END IF
               END DO

               DEALLOCATE (BIb_C_2D)
               ALLOCATE (BIb_C_2D(rec_ia_size, rec_L_size))
               BIb_C_2D(:, :) = BIb_C_rec

               DEALLOCATE (BIb_C_rec)
            END DO

            DEALLOCATE (BIb_C_2D)
         END IF

         ! deallocaete the info array
         DEALLOCATE (RPA_info_start)
         DEALLOCATE (RPA_info_end)
         DEALLOCATE (RPA_info_size)

         ! mp_sum the local data across processes belonging to different RPA group.
         ! first create the para_env then mp_sum
         sub_sub_color = para_env_RPA%mepos
         CALL mp_comm_split_direct(para_env%group, comm_exchange, sub_sub_color)
         NULLIFY (para_env_exchange)
         CALL cp_para_env_create(para_env_exchange, comm_exchange)

         CALL timeset(routineN//"_sum", handle2)
         CALL mp_sum(fm_mat_S%local_data, para_env_exchange%group)
         CALL timestop(handle2)

         CALL cp_para_env_release(para_env_exchange)

      END IF ! not imag. time

      ! create the twin matrix for the mat-mat-mul (mat_Q)
      !XXX CALL cp_fm_create(fm_mat_G,fm_struct,name="fm_mat_G")
      !XXX CALL cp_fm_set_all(matrix=fm_mat_G,alpha=0.0_dp)

      CALL cp_fm_struct_release(fm_struct)

      IF (.NOT. my_gw_open_shell) THEN
         ! create the Q matrix dimen_RIxdimen_RI where the result of the mat-mat-mult will be stored
         NULLIFY (fm_mat_Q_gemm)
         NULLIFY (fm_struct)
         !CALL cp_fm_struct_create(fm_struct,context=blacs_env,nrow_global=dimen_RI,&
         !                         ncol_global=dimen_RI,para_env=para_env_RPA)
         CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=dimen_RI, &
                                  ncol_global=dimen_RI, para_env=para_env_RPA, &
                                  nrow_block=nrow_block_mat, ncol_block=ncol_block_mat, force_block=.TRUE.)
         CALL cp_fm_create(fm_mat_Q_gemm, fm_struct, name="fm_mat_Q_gemm")
         CALL cp_fm_struct_release(fm_struct)

         CALL cp_fm_set_all(matrix=fm_mat_Q_gemm, alpha=0.0_dp)

         ! create the Q matrix with a different blacs env
         NULLIFY (blacs_env_Q)
         IF (my_blacs_ext) THEN
            blacs_env_Q => blacs_env_ext
         ELSE
            CALL cp_blacs_env_create(blacs_env=blacs_env_Q, para_env=para_env_RPA)
         END IF

         NULLIFY (fm_mat_Q)
         NULLIFY (fm_struct)
         CALL cp_fm_struct_create(fm_struct, context=blacs_env_Q, nrow_global=dimen_RI, &
                                  ncol_global=dimen_RI, para_env=para_env_RPA)
         CALL cp_fm_create(fm_mat_Q, fm_struct, name="fm_mat_Q")

         CALL cp_fm_struct_release(fm_struct)

         CALL cp_fm_set_all(matrix=fm_mat_Q, alpha=0.0_dp)
      END IF

      ! in case we do imaginary time, we allocate fm_scaled_dm_occ and fm_scaled_dm_virt
      IF (my_do_im_time) THEN

         CALL cp_fm_create(fm_scaled_dm_occ, mo_coeff%matrix_struct)
         CALL cp_fm_set_all(fm_scaled_dm_occ, 0.0_dp)

         CALL cp_fm_create(fm_scaled_dm_virt, mo_coeff%matrix_struct)
         CALL cp_fm_set_all(fm_scaled_dm_virt, 0.0_dp)

         CALL create_occ_virt_mo_coeffs(fm_mo_coeff_occ, fm_mo_coeff_virt, mo_coeff, &
                                        nmo, homo)

         ! in case we do GW, we directly work with Q(it) (or Pi(it)) and do not need Q(iw)
         IF (.NOT. my_do_gw) THEN
            ALLOCATE (fm_mat_Q_omega(num_integ_points))
            DO iii = 1, num_integ_points
               CALL cp_fm_create(fm_mat_Q_omega(iii)%matrix, fm_mat_Q%matrix_struct)
               CALL cp_fm_set_all(fm_mat_Q_omega(iii)%matrix, 0.0_dp)
            END DO
         END IF

         CALL cp_dbcsr_get_info(mat_B_munu(1)%matrix, nblkrows_total, nblkcols_total, &
                                nfullrows_total, nfullcols_total, &
                                nblkrows_local, nblkcols_local, &
                                nfullrows_local, nfullcols_local, &
                                my_prow, my_pcol, &
                                local_rows, local_cols, proc_row_dist, proc_col_dist, &
                                row_blk_size, col_blk_size, row_blk_offset, col_blk_offset)

         ALLOCATE (block_distribution_M(ngroup, nblkrows_total, nblkcols_total))
         block_distribution_M = 0

         ! take care: color_sub starts with 0
         block_distribution_M(color_sub+1, local_rows, local_cols) = para_env%mepos

         CALL mp_sum(block_distribution_M, para_env%group)

         DEALLOCATE (block_distribution_M)

      END IF ! imag. time

      ! release blacs_env
      IF (.NOT. my_blacs_ext) CALL cp_blacs_env_release(blacs_env_Q)
      IF (.NOT. my_blacs_S_ext) CALL cp_blacs_env_release(blacs_env)

      CALL timestop(handle)

   END SUBROUTINE create_integ_mat

! **************************************************************************************************
!> \brief ...
!> \param fm_mo_coeff_occ ...
!> \param fm_mo_coeff_virt ...
!> \param mo_coeff ...
!> \param nmo ...
!> \param homo ...
! **************************************************************************************************
   SUBROUTINE create_occ_virt_mo_coeffs(fm_mo_coeff_occ, fm_mo_coeff_virt, mo_coeff, &
                                        nmo, homo)
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_mo_coeff_occ, fm_mo_coeff_virt, &
                                                            mo_coeff
      INTEGER                                            :: nmo, homo

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_occ_virt_mo_coeffs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icol_global, irow_global

      CALL timeset(routineN, handle)

      CALL cp_fm_create(fm_mo_coeff_occ, mo_coeff%matrix_struct)
      CALL cp_fm_set_all(fm_mo_coeff_occ, 0.0_dp)
      CALL cp_fm_to_fm(mo_coeff, fm_mo_coeff_occ)

      ! set all virtual MO coeffs to zero
      DO irow_global = 1, nmo
         DO icol_global = homo+1, nmo
            CALL cp_fm_set_element(fm_mo_coeff_occ, irow_global, icol_global, 0.0_dp)
         END DO
      END DO

      CALL cp_fm_create(fm_mo_coeff_virt, mo_coeff%matrix_struct)
      CALL cp_fm_set_all(fm_mo_coeff_virt, 0.0_dp)
      CALL cp_fm_to_fm(mo_coeff, fm_mo_coeff_virt)

      ! set all occupied MO coeffs to zero
      DO irow_global = 1, nmo
         DO icol_global = 1, homo
            CALL cp_fm_set_element(fm_mo_coeff_virt, irow_global, icol_global, 0.0_dp)
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE create_occ_virt_mo_coeffs

! **************************************************************************************************
!> \brief ...
!> \param Erpa ...
!> \param mp2_env ...
!> \param para_env ...
!> \param para_env_RPA ...
!> \param para_env_sub ...
!> \param unit_nr ...
!> \param homo ...
!> \param virtual ...
!> \param dimen_RI ...
!> \param dimen_ia ...
!> \param dimen_nm_gw ...
!> \param Eigenval ...
!> \param num_integ_points ...
!> \param num_integ_group ...
!> \param color_rpa_group ...
!> \param fm_mat_S ...
!> \param fm_mat_Q_gemm ...
!> \param fm_mat_Q ...
!> \param fm_mat_S_gw ...
!> \param fm_mat_R_gw ...
!> \param my_do_gw ...
!> \param gw_corr_lev_occ ...
!> \param gw_corr_lev_virt ...
!> \param num_poles ...
!> \param ext_scaling ...
!> \param omega_max_fit ...
!> \param stop_crit ...
!> \param check_fit ...
!> \param fermi_level_offset ...
!> \param crossing_search ...
!> \param max_iter_fit ...
!> \param mm_style ...
!> \param do_minimax_quad ...
!> \param homo_beta ...
!> \param virtual_beta ...
!> \param dimen_ia_beta ...
!> \param Eigenval_beta ...
!> \param fm_mat_S_beta ...
!> \param fm_mat_Q_gemm_beta ...
!> \param fm_mat_Q_beta ...
!> \param fm_mat_S_gw_beta ...
!> \param gw_corr_lev_occ_beta ...
!> \param gw_corr_lev_virt_beta ...
!> \param do_im_time ...
!> \param fm_scaled_dm_occ ...
!> \param fm_scaled_dm_virt ...
!> \param fm_mo_coeff_occ ...
!> \param fm_mo_coeff_virt ...
!> \param fm_mat_Q_omega ...
!> \param mat_munu ...
!> \param mat_B_munu ...
!> \param my_Lrows ...
!> \param my_group_L_size ...
!> \param my_group_L_start ...
!> \param ngroup ...
!> \param eps_filter ...
!> \param sizes_array ...
!> \param fm_mo_coeff_occ_beta ...
!> \param fm_mo_coeff_virt_beta ...
! **************************************************************************************************
   SUBROUTINE rpa_num_int(Erpa, mp2_env, para_env, para_env_RPA, para_env_sub, unit_nr, &
                          homo, virtual, dimen_RI, dimen_ia, dimen_nm_gw, &
                          Eigenval, num_integ_points, num_integ_group, color_rpa_group, &
                          fm_mat_S, fm_mat_Q_gemm, fm_mat_Q, fm_mat_S_gw, fm_mat_R_gw, &
                          my_do_gw, gw_corr_lev_occ, gw_corr_lev_virt, num_poles, ext_scaling, omega_max_fit, &
                          stop_crit, check_fit, fermi_level_offset, crossing_search, &
                          max_iter_fit, mm_style, do_minimax_quad, &
                          homo_beta, virtual_beta, dimen_ia_beta, Eigenval_beta, fm_mat_S_beta, &
                          fm_mat_Q_gemm_beta, fm_mat_Q_beta, fm_mat_S_gw_beta, &
                          gw_corr_lev_occ_beta, gw_corr_lev_virt_beta, &
                          do_im_time, fm_scaled_dm_occ, fm_scaled_dm_virt, &
                          fm_mo_coeff_occ, fm_mo_coeff_virt, fm_mat_Q_omega, mat_munu, mat_B_munu, &
                          my_Lrows, my_group_L_size, my_group_L_start, &
                          ngroup, eps_filter, &
                          sizes_array, fm_mo_coeff_occ_beta, fm_mo_coeff_virt_beta)
      REAL(KIND=dp)                                      :: Erpa
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_RPA, para_env_sub
      INTEGER                                            :: unit_nr, homo, virtual, dimen_RI, &
                                                            dimen_ia, dimen_nm_gw
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: num_integ_points, num_integ_group, &
                                                            color_rpa_group
      TYPE(cp_fm_type), POINTER                          :: fm_mat_S, fm_mat_Q_gemm, fm_mat_Q, &
                                                            fm_mat_S_gw, fm_mat_R_gw
      LOGICAL                                            :: my_do_gw
      INTEGER                                            :: gw_corr_lev_occ, gw_corr_lev_virt, &
                                                            num_poles
      REAL(KIND=dp)                                      :: ext_scaling, omega_max_fit, stop_crit
      LOGICAL                                            :: check_fit
      REAL(KIND=dp)                                      :: fermi_level_offset
      INTEGER                                            :: crossing_search, max_iter_fit, mm_style
      LOGICAL                                            :: do_minimax_quad
      INTEGER, OPTIONAL                                  :: homo_beta, virtual_beta, dimen_ia_beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_mat_S_beta, fm_mat_Q_gemm_beta, &
                                                            fm_mat_Q_beta, fm_mat_S_gw_beta
      INTEGER, OPTIONAL                                  :: gw_corr_lev_occ_beta, &
                                                            gw_corr_lev_virt_beta
      LOGICAL, OPTIONAL                                  :: do_im_time
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_scaled_dm_occ, fm_scaled_dm_virt, &
                                                            fm_mo_coeff_occ, fm_mo_coeff_virt
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: fm_mat_Q_omega
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mat_B_munu
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :), &
         OPTIONAL                                        :: my_Lrows
      INTEGER, OPTIONAL                                  :: my_group_L_size, my_group_L_start, ngroup
      REAL(KIND=dp), OPTIONAL                            :: eps_filter
      INTEGER, ALLOCATABLE, DIMENSION(:), OPTIONAL       :: sizes_array
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_mo_coeff_occ_beta, &
                                                            fm_mo_coeff_virt_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rpa_num_int', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: im_unit, re_unit
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: Lambda, Lambda_without_offset
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: vec_Sigma_c_gw, vec_Sigma_c_gw_beta
      INTEGER :: avirt, contour_def_end, contour_def_level_tot, contour_def_start, count_ev_sc_GW, &
         gw_corr_lev_tot, handle, handle2, handle3, handle4, i_ener, i_global, i_mo, ierr, iiB, &
         iii, index_contour_def, info_chol, iocc, iquad, iter_ev_sc, j_global, j_mo, jjB, jquad, &
         m_global, m_global_beta, max_size_RI_group, my_num_dgemm_call, n_global, n_global_beta, &
         n_level_gw, n_level_gw_ref, ncol_local, nm_global, nmo, nrow_local, num_fit_points, &
         num_real_freq_cont_def, num_residues, num_var, num_x_nodes_max, num_x_nodes_min, &
         num_x_nodes_min_temp, num_x_nodes_step, number_of_rec, number_of_rec_beta
      INTEGER :: number_of_send, number_of_send_beta
      INTEGER, ALLOCATABLE, DIMENSION(:) :: alpha_beta_from_jquad, map_rec_size, &
         map_rec_size_beta, map_send_size, map_send_size_beta, mo_from_jquad, residue_from_jquad, &
         RPA_proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_size_source, local_size_source_beta
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL :: do_contour_def, do_mix_exchange, do_ri_Sigma_x, first_cycle, first_cycle_im_time, &
         first_cycle_im_time_beta, my_do_im_time, my_open_shell, remove_neg_virt_energies
      REAL(KIND=dp) :: a_scaling, actual_flop_rate, alpha, contour_def_offset, e_fermi, E_Range, &
         eigen_diff, Emax, Emax_beta, Emin, Emin_beta, FComega, frac_rix, ls_error, max_error_min, &
         my_flop_rate, omega, omega_i, omega_old, sign_occ_virt, t_end, t_start, tau, tau_old
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:) :: Eigenval_last, Eigenval_last_beta, Eigenval_scf, &
         Eigenval_scf_beta, Lambda_Im, Lambda_Re, m_value, m_value_beta, Q_log, real_freq, tau_tj, &
         tau_wj, tj, trace_Qomega, vec_gw_energ, vec_gw_energ_beta, vec_gw_energ_error_fit, &
         vec_gw_energ_error_fit_beta, vec_omega_fit_gw, vec_omega_fit_gw_sign, vec_omega_gw, &
         vec_Sigma_c_gw_real_freq, vec_Sigma_c_gw_real_freq_beta, vec_Sigma_x_gw, &
         vec_Sigma_x_gw_beta, vec_tau_fit_gw_sign, vec_W_gw, vec_W_gw_beta, wj, x_tw, z_value, &
         z_value_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: my_Lrows_transposed, tau_wj_2d, &
                                                            trace_2d, vec_W_gw_store
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_M_munu_P_occ, mat_M_munu_P_virt
      TYPE(cp_dbcsr_type), POINTER                       :: mat_M_munu_occ, mat_M_munu_virt
      TYPE(cp_fm_type), POINTER :: fm_mat_L, fm_mat_S_contour_def, fm_mat_S_contour_def_beta, &
         fm_mat_S_gw_work, fm_mat_S_gw_work_beta, fm_mat_work
      TYPE(integ_mat_buffer_type), ALLOCATABLE, &
         DIMENSION(:)                                    :: buffer_rec, buffer_rec_beta, &
                                                            buffer_send, buffer_send_beta

      CALL timeset(routineN, handle)

      my_open_shell = .FALSE.
      IF (PRESENT(homo_beta) .AND. &
          PRESENT(virtual_beta) .AND. &
          PRESENT(dimen_ia_beta) .AND. &
          PRESENT(Eigenval_beta) .AND. &
          PRESENT(fm_mat_S_beta) .AND. &
          PRESENT(fm_mat_Q_gemm_beta) .AND. &
          PRESENT(fm_mat_Q_beta)) my_open_shell = .TRUE.

      my_do_im_time = .FALSE.
      IF (PRESENT(do_im_time)) my_do_im_time = do_im_time

      do_contour_def = .FALSE.

      ALLOCATE (tj(num_integ_points))
      tj = 0.0_dp

      ALLOCATE (wj(num_integ_points))
      wj = 0.0_dp

      ALLOCATE (Q_log(dimen_RI))

      ALLOCATE (trace_Qomega(dimen_RI))

      ! set up the least-square time grid and other matrices specifically for imag time
      IF (my_do_im_time) THEN

         ALLOCATE (tau_tj(num_integ_points))
         tau_tj = 0.0_dp

         ALLOCATE (tau_wj_2d(num_integ_points, num_integ_points))
         tau_wj_2d = 0.0_dp

         ALLOCATE (tau_wj(num_integ_points))
         tau_wj = 0.0_dp

         ! determine maximum size of aux. bf group
         max_size_RI_group = MAXVAL(sizes_array)

         NULLIFY (mat_M_munu_occ)
         CALL cp_dbcsr_init_p(mat_M_munu_occ)
!JW 2del      CALL cp_dbcsr_create(matrix=mat_M_munu_occ,template=mat_munu%matrix)
         CALL cp_dbcsr_copy(mat_M_munu_occ, mat_munu%matrix)
         CALL cp_dbcsr_set(mat_M_munu_occ, 0.0_dp)

         NULLIFY (mat_M_munu_virt)
         CALL cp_dbcsr_init_p(mat_M_munu_virt)
!JW 2del      CALL cp_dbcsr_create(matrix=mat_M_munu_virt,template=mat_munu%matrix)
         CALL cp_dbcsr_copy(mat_M_munu_virt, mat_munu%matrix)
         CALL cp_dbcsr_set(mat_M_munu_virt, 0.0_dp)

         CALL timeset(routineN//"_im_time_alloc_trace_2d", handle4)

         ALLOCATE (trace_2d(my_group_L_size, dimen_RI))
         trace_2d(:, :) = 0.0_dp

         CALL timestop(handle4)

         CALL timeset(routineN//"_im_time_alloc_mat_M_munu_P_occ", handle4)

         NULLIFY (mat_M_munu_P_occ)
         CALL cp_dbcsr_allocate_matrix_set(mat_M_munu_P_occ, my_group_L_size)
         DO iii = 1, my_group_L_size
            ALLOCATE (mat_M_munu_P_occ(iii)%matrix)

            CALL cp_dbcsr_init(mat_M_munu_P_occ(iii)%matrix)
            CALL cp_dbcsr_desymmetrize(mat_B_munu(1)%matrix, mat_M_munu_P_occ(iii)%matrix)
            CALL cp_dbcsr_set(mat_M_munu_P_occ(iii)%matrix, 0.0_dp)
         END DO

         CALL timestop(handle4)

         CALL timeset(routineN//"_im_time_alloc_mat_M_munu_P_virt", handle4)

         NULLIFY (mat_M_munu_P_virt)
         CALL cp_dbcsr_allocate_matrix_set(mat_M_munu_P_virt, max_size_RI_group)
         DO iii = 1, max_size_RI_group
            ALLOCATE (mat_M_munu_P_virt(iii)%matrix)

            CALL cp_dbcsr_init(mat_M_munu_P_virt(iii)%matrix)
            CALL cp_dbcsr_desymmetrize(mat_B_munu(1)%matrix, mat_M_munu_P_virt(iii)%matrix)
            CALL cp_dbcsr_set(mat_M_munu_P_virt(iii)%matrix, 0.0_dp)
         END DO

         CALL timestop(handle4)

         NULLIFY (fm_mat_work)
         CALL cp_fm_create(fm_mat_work, fm_mat_Q%matrix_struct)
         CALL cp_fm_to_fm(fm_mat_Q, fm_mat_work)
         CALL cp_fm_set_all(matrix=fm_mat_work, alpha=0.0_dp)

         NULLIFY (fm_mat_L)
         CALL cp_fm_create(fm_mat_L, fm_mat_Q%matrix_struct)
         CALL cp_fm_to_fm(fm_mat_Q, fm_mat_L)
         CALL cp_fm_set_all(matrix=fm_mat_L, alpha=0.0_dp)

      END IF

      IF (do_minimax_quad) THEN
         ! MINIMAX quadrature
         ALLOCATE (x_tw(2*num_integ_points))
         x_tw = 0.0_dp

         Emin = Eigenval(homo+1)-Eigenval(homo)
         Emax = MAXVAL(Eigenval)-MINVAL(Eigenval)
         IF (my_open_shell) THEN
            IF (homo_beta > 0) THEN
               Emin_beta = Eigenval_beta(homo_beta+1)-Eigenval_beta(homo_beta)
               Emax_beta = MAXVAL(Eigenval_beta)-MINVAL(Eigenval_beta)
               Emin = MIN(Emin, Emin_beta)
               Emax = MAX(Emax, Emax_beta)
            END IF
         END IF
         E_Range = Emax/Emin

         ierr = 0
         CALL get_rpa_minimax_coeff(num_integ_points, E_Range, x_tw, ierr)

         DO jquad = 1, num_integ_points
            tj(jquad) = x_tw(jquad)
            wj(jquad) = x_tw(jquad+num_integ_points)
         END DO

         DEALLOCATE (x_tw)

         IF (unit_nr > 0) THEN
            WRITE (UNIT=unit_nr, FMT="(T3,A,T66,F15.4)") &
               "INTEG_INFO| Range for the minimax approximation:", E_Range
            WRITE (UNIT=unit_nr, FMT="(T3,A,T54,A,T72,A)") "INTEG_INFO| Minimax parameters:", "Weights", "Abscissas"
            DO jquad = 1, num_integ_points
               WRITE (UNIT=unit_nr, FMT="(T41,F20.10,F20.10)") wj(jquad), tj(jquad)
            END DO
            CALL m_flush(unit_nr)
         END IF

         ! scale the minimax parameters
         tj(:) = tj(:)*Emin
         wj(:) = wj(:)*Emin

         ! set up the minimax time grid
         IF (my_do_im_time) THEN

            ALLOCATE (x_tw(2*num_integ_points))
            x_tw = 0.0_dp

            CALL get_minimax_coeff(num_integ_points, E_Range, x_tw)

            DO jquad = 1, num_integ_points
               tau_tj(jquad) = x_tw(jquad)/2.0_dp
               tau_wj(jquad) = x_tw(jquad+num_integ_points)/2.0_dp
            END DO

            DEALLOCATE (x_tw)

            IF (unit_nr > 0) THEN
               WRITE (UNIT=unit_nr, FMT="(T3,A,T66,F15.4)") &
                  "INTEG_INFO| Range for the minimax approximation:", E_Range
               WRITE (UNIT=unit_nr, FMT="(T3,A,T54,A,T72,A)") &
                  "INTEG_INFO| Minimax parameters of the time grid:", "Weights", "Abscissas"
               DO jquad = 1, num_integ_points
                  WRITE (UNIT=unit_nr, FMT="(T41,F20.10,F20.10)") tau_wj(jquad), tau_tj(jquad)
               END DO
               CALL m_flush(unit_nr)
            END IF

            ! scale grid from [1,R] to [Emin,Emax]
            tau_tj(:) = tau_tj(:)/Emin
            tau_wj(:) = tau_wj(:)/Emin

            !JW hack that should be defined in the input
            num_x_nodes_min = 5*num_integ_points

            num_x_nodes_max = 5*num_integ_points

            IF (num_x_nodes_min > num_x_nodes_max) THEN

               num_x_nodes_min_temp = num_x_nodes_min

               num_x_nodes_min = num_x_nodes_max

               num_x_nodes_max = num_x_nodes_min_temp

            END IF

            num_x_nodes_step = num_integ_points

            ! get the integration weights of the tau-grid specifically for every omega-point (because we are doing a
            ! Fourier transform from imag. time to imag. frequency)
            CALL get_rpa_least_sq_weights_tau(num_integ_points, tau_tj, tau_wj_2d, tj, &
                                              num_x_nodes_min, num_x_nodes_max, num_x_nodes_step, &
                                              Emin, Emax, ls_error, max_error_min)
            IF (unit_nr > 0) THEN
               WRITE (UNIT=unit_nr, FMT="(T3,A,T66,ES15.2)") &
                  "INTEG_INFO| Maximum deviation of the imag. time fit:", max_error_min
            END IF

         END IF

      ELSE
         ! Clenshaw-Curtius quadrature
         DO jquad = 1, num_integ_points-1
            tj(jquad) = jquad*pi/(2.0_dp*num_integ_points)
            wj(jquad) = pi/(num_integ_points*SIN(tj(jquad))**2)
         END DO
         tj(num_integ_points) = pi/2.0_dp
         wj(num_integ_points) = pi/(2.0_dp*num_integ_points*SIN(tj(num_integ_points))**2)

         a_scaling = 1.0_dp
         IF (my_open_shell) THEN
            CALL calc_scaling_factor(a_scaling, para_env, para_env_RPA, homo, virtual, Eigenval, &
                                     num_integ_points, num_integ_group, color_rpa_group, &
                                     tj, wj, fm_mat_S, &
                                     homo_beta, virtual_beta, dimen_ia_beta, Eigenval_beta, fm_mat_S_beta)
         ELSE
            CALL calc_scaling_factor(a_scaling, para_env, para_env_RPA, homo, virtual, Eigenval, &
                                     num_integ_points, num_integ_group, color_rpa_group, &
                                     tj, wj, fm_mat_S)
         END IF

         ! for G0W0, we may set the scaling factor by hand
         IF (my_do_gw .AND. ext_scaling > 0.0_dp) THEN
            a_scaling = ext_scaling
         END IF

         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.5)') 'INTEG_INFO| Scaling parameter:', a_scaling

         wj(:) = wj(:)*a_scaling

      END IF

      ! initialize buffer for matrix redistribution
      CALL initialize_buffer(fm_mat_Q_gemm, fm_mat_Q, RPA_proc_map, buffer_rec, buffer_send, &
                             number_of_rec, number_of_send, &
                             map_send_size, map_rec_size, local_size_source, para_env_RPA)
      IF (my_open_shell .AND. (.NOT. my_do_im_time)) THEN
         CALL initialize_buffer(fm_mat_Q_gemm_beta, fm_mat_Q_beta, RPA_proc_map, buffer_rec_beta, buffer_send_beta, &
                                number_of_rec_beta, number_of_send_beta, &
                                map_send_size_beta, map_rec_size_beta, local_size_source_beta, para_env_RPA)
      END IF

      IF (my_open_shell) THEN
         alpha = 2.0_dp
      ELSE
         alpha = 4.0_dp
      END IF

      nmo = homo+virtual

      IF (my_do_gw) THEN
         gw_corr_lev_tot = gw_corr_lev_occ+gw_corr_lev_virt

         ALLOCATE (Eigenval_scf(nmo))
         Eigenval_scf(:) = Eigenval(:)

         ALLOCATE (Eigenval_last(nmo))
         Eigenval_last(:) = Eigenval(:)

         ! in the case of HF_diag approach of X. Blase (PRB 83, 115103 (2011), Sec. IV), we subtract the
         ! XC potential and add exact exchange
         IF (mp2_env%ri_g0w0%hf_like_ev_start) THEN
            DO n_level_gw = 1, gw_corr_lev_tot
               n_level_gw_ref = n_level_gw+homo-gw_corr_lev_occ
               Eigenval(n_level_gw_ref) = Eigenval(n_level_gw_ref)+ &
                                          mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(n_level_gw_ref, 1)
            END DO
         END IF

         ! Eigenval for beta
         IF (my_open_shell) THEN
            ALLOCATE (Eigenval_scf_beta(nmo))
            Eigenval_scf_beta(:) = Eigenval_beta(:)

            ALLOCATE (Eigenval_last_beta(nmo))
            Eigenval_last_beta(:) = Eigenval_beta(:)

            ! in the case of HF_diag approach of X. Blase (PRB 83, 115103 (2011), Sec. IV), we subtract the
            ! XC potential and add exact exchange
            IF (mp2_env%ri_g0w0%hf_like_ev_start) THEN
               DO n_level_gw = 1, gw_corr_lev_tot
                  n_level_gw_ref = n_level_gw+homo-gw_corr_lev_occ
                  Eigenval_beta(n_level_gw_ref) = Eigenval_beta(n_level_gw_ref)+ &
                                                  mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(n_level_gw_ref, 2)
               END DO
            END IF
         END IF

         ! create temporary matrix to store B*([1+Q(iw')]^-1-1), has the same size as B
         NULLIFY (fm_mat_S_gw_work)
         CALL cp_fm_create(fm_mat_S_gw_work, fm_mat_S_gw%matrix_struct)
         CALL cp_fm_set_all(matrix=fm_mat_S_gw_work, alpha=0.0_dp)

         IF (my_open_shell) THEN
            NULLIFY (fm_mat_S_gw_work_beta)
            CALL cp_fm_create(fm_mat_S_gw_work_beta, fm_mat_S_gw%matrix_struct)
            CALL cp_fm_set_all(matrix=fm_mat_S_gw_work_beta, alpha=0.0_dp)
         END IF

         ALLOCATE (vec_W_gw(dimen_nm_gw))

         vec_W_gw = 0.0_dp

         IF (my_do_im_time) THEN
            ! hack JW for testing; vec_W_gw_store has to be removed afterwards
            ALLOCATE (vec_W_gw_store(dimen_nm_gw, num_integ_points))
            vec_W_gw_store = 0.0_dp
         END IF

         IF (my_open_shell) THEN
            ALLOCATE (vec_W_gw_beta(dimen_nm_gw))
            vec_W_gw_beta = 0.0_dp
         END IF

         im_unit = (0.0_dp, 1.0_dp)
         re_unit = (1.0_dp, 0.0_dp)

         ! fill the omega_frequency vector
         ALLOCATE (vec_omega_gw(num_integ_points))
         vec_omega_gw = 0.0_dp

         DO jquad = 1, num_integ_points
            IF (do_minimax_quad) THEN
               omega = tj(jquad)
            ELSE
               omega = a_scaling/TAN(tj(jquad))
            END IF
            vec_omega_gw(jquad) = omega
         END DO

         ! determine number of fit points in the interval [0,w_max] for virt, or [-w_max,0] for occ
         num_fit_points = 0

         DO jquad = 1, num_integ_points
            IF (vec_omega_gw(jquad) < omega_max_fit) THEN
               num_fit_points = num_fit_points+1
            END IF
         END DO

         ! allocate imaginary time matrices
         IF (my_do_im_time) THEN
            num_fit_points = 2*num_integ_points
            ALLOCATE (vec_tau_fit_gw_sign(2*num_fit_points))
            vec_tau_fit_gw_sign = 0.0_dp
         END IF

         ! create new arrays containing omega values at which we calculate vec_Sigma_c_gw
         ALLOCATE (vec_omega_fit_gw(num_fit_points))

         ALLOCATE (vec_omega_fit_gw_sign(num_fit_points))

         ALLOCATE (vec_Sigma_c_gw(gw_corr_lev_tot, num_fit_points))

         vec_Sigma_c_gw = (0.0_dp, 0.0_dp)

         IF (my_open_shell) THEN
            ALLOCATE (vec_Sigma_c_gw_beta(gw_corr_lev_tot, num_fit_points))
            vec_Sigma_c_gw_beta = (0.0_dp, 0.0_dp)
         END IF

         do_ri_Sigma_x = mp2_env%ri_g0w0%do_ri_Sigma_x

         do_mix_exchange = mp2_env%ri_g0w0%mix_exchange

         ! in case we do RI for Sigma_x, we calculate Sigma_x right here
         IF (do_ri_Sigma_x .OR. do_mix_exchange) THEN

            IF (do_mix_exchange) THEN
               frac_rix = 1.0_dp-mp2_env%ri_g0w0%frac_exx
            ELSE
               frac_rix = 1.0_dp
            END IF

            ALLOCATE (vec_Sigma_x_gw(nmo))
            vec_Sigma_x_gw = 0.0_dp

            IF (my_open_shell) THEN
               ALLOCATE (vec_Sigma_x_gw_beta(nmo))
               vec_Sigma_x_gw_beta = 0.0_dp
            END IF

            CALL cp_fm_get_info(matrix=fm_mat_S_gw, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)

            CALL mp_sync(para_env%group)

            ! loop over (nm) index
            DO iiB = 1, nrow_local

               ! this is needed for correct values within parallelization, see analogue
               ! statement below
               IF (MODULO(1, num_integ_group) /= color_rpa_group) CYCLE

               nm_global = row_indices(iiB)

               ! transform the index nm to n and m, formulae copied from Mauro's code
               n_global = MAX(1, nm_global-1)/nmo+1
               m_global = nm_global-(n_global-1)*nmo
               n_global = n_global+homo-gw_corr_lev_occ

               IF (my_open_shell) THEN
                  n_global_beta = MAX(1, nm_global-1)/nmo+1
                  m_global_beta = nm_global-(n_global_beta-1)*nmo
                  n_global_beta = n_global_beta+homo_beta-gw_corr_lev_occ_beta
               END IF

               IF (m_global <= homo) THEN

                  ! loop over auxiliary basis functions
                  DO jjB = 1, ncol_local

                     ! Sigma_x_n = -sum_m^occ sum_P (B_(nm)^P)^2
                     vec_Sigma_x_gw(n_global) = &
                        vec_Sigma_x_gw(n_global)- &
                        fm_mat_S_gw%local_data(iiB, jjB)**2

                  END DO

               END IF

               ! The same for beta
               IF (my_open_shell) THEN

                  IF (m_global_beta <= homo_beta) THEN

                     ! loop over auxiliary basis functions
                     DO jjB = 1, ncol_local

                        ! Sigma_x_n = -sum_m^occ sum_P (B_(nm)^P)^2
                        vec_Sigma_x_gw_beta(n_global_beta) = &
                           vec_Sigma_x_gw_beta(n_global_beta)- &
                           fm_mat_S_gw_beta%local_data(iiB, jjB)**2

                     END DO

                  END IF

               END IF

            END DO

            CALL mp_sum(vec_Sigma_x_gw, para_env%group)

            mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 1) = &
               mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 1)+ &
               frac_rix*vec_Sigma_x_gw(:)

            IF (my_open_shell) THEN

               CALL mp_sum(vec_Sigma_x_gw_beta, para_env%group)

               mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 2) = &
                  mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 2)+ &
                  frac_rix*vec_Sigma_x_gw_beta(:)

            END IF

         END IF

         ! fill the omega vector with frequencies, where we calculate the self-energy
         iquad = 0
         DO jquad = 1, num_integ_points
            IF (vec_omega_gw(jquad) < omega_max_fit) THEN
               iquad = iquad+1
               vec_omega_fit_gw(iquad) = vec_omega_gw(jquad)
            END IF
         END DO

         DEALLOCATE (vec_omega_gw)

         ! check if we want to do the contour deformation technique instead of analytical continuation
         contour_def_start = mp2_env%ri_g0w0%contour_def_start
         contour_def_end = mp2_env%ri_g0w0%contour_def_end

         IF (contour_def_start > 0 .AND. contour_def_end >= contour_def_start) THEN
            do_contour_def = .TRUE.
         END IF
         contour_def_offset = mp2_env%ri_g0w0%contour_def_offset

         IF (do_contour_def) THEN

            ! check range
            IF (contour_def_start < homo-gw_corr_lev_occ+1) THEN
               contour_def_start = homo-gw_corr_lev_occ+1
            END IF

            IF (contour_def_end > homo+gw_corr_lev_virt) THEN
               contour_def_end = homo+gw_corr_lev_virt
            END IF

            contour_def_level_tot = contour_def_end-contour_def_start+1

            num_real_freq_cont_def = 2*contour_def_level_tot

            ! set up array for the self-energy as function of real energy. one energy is slightly
            ! above and one below the DFT MO energy => the space is 2* number of levels which are
            ! treated by contour deformation technique
            ALLOCATE (vec_Sigma_c_gw_real_freq(num_real_freq_cont_def))

            IF (my_open_shell) THEN
               ALLOCATE (vec_Sigma_c_gw_real_freq_beta(num_real_freq_cont_def))
            END IF

            NULLIFY (fm_mat_S_contour_def)
            CALL cp_fm_create(fm_mat_S_contour_def, fm_mat_S%matrix_struct)
            CALL cp_fm_to_fm(source=fm_mat_S, destination=fm_mat_S_contour_def)

            IF (my_open_shell) THEN
               NULLIFY (fm_mat_S_contour_def_beta)
               CALL cp_fm_create(fm_mat_S_contour_def_beta, fm_mat_S_beta%matrix_struct)
               CALL cp_fm_to_fm(source=fm_mat_S_beta, destination=fm_mat_S_contour_def_beta)
            END IF

         END IF

         ! arrays storing the complex fit parameters a_0, a_1, b_1, a_2, b_2, ...
         num_var = 2*num_poles+1
         ALLOCATE (Lambda(num_var))
         Lambda = (0.0_dp, 0.0_dp)
         ALLOCATE (Lambda_without_offset(num_var))
         Lambda_without_offset = (0.0_dp, 0.0_dp)
         ALLOCATE (Lambda_Re(num_var))
         Lambda_Re = 0.0_dp
         ALLOCATE (Lambda_Im(num_var))
         Lambda_Im = 0.0_dp

         ! arrays storing the correlation self-energy, stat. error and z-shot value
         ALLOCATE (vec_gw_energ(gw_corr_lev_tot))
         vec_gw_energ = 0.0_dp
         ALLOCATE (vec_gw_energ_error_fit(gw_corr_lev_tot))
         vec_gw_energ_error_fit = 0.0_dp
         ALLOCATE (z_value(gw_corr_lev_tot))
         z_value = 0.0_dp
         ALLOCATE (m_value(gw_corr_lev_tot))
         m_value = 0.0_dp

         ! the same for beta
         IF (my_open_shell) THEN
            ALLOCATE (vec_gw_energ_beta(gw_corr_lev_tot))
            vec_gw_energ_beta = 0.0_dp
            ALLOCATE (vec_gw_energ_error_fit_beta(gw_corr_lev_tot))
            vec_gw_energ_error_fit_beta = 0.0_dp
            ALLOCATE (z_value_beta(gw_corr_lev_tot))
            z_value_beta = 0.0_dp
            ALLOCATE (m_value_beta(gw_corr_lev_tot))
            m_value_beta = 0.0_dp
         END IF

      END IF

      Erpa = 0.0_dp
      first_cycle = .TRUE.
      omega_old = 0.0_dp
      my_num_dgemm_call = 0
      my_flop_rate = 0.0_dp

      IF (my_do_gw) THEN
         iter_ev_sc = mp2_env%ri_g0w0%iter_ev_sc
      ELSE
         iter_ev_sc = 1
      END IF

      DO count_ev_sc_GW = 1, iter_ev_sc

         ! reset some values, important when doing eigenvalue self-consistent GW
         IF (my_do_gw) THEN
            Erpa = 0.0_dp
            vec_Sigma_c_gw = (0.0_dp, 0.0_dp)
            vec_gw_energ = 0.0_dp
            vec_gw_energ_error_fit = 0.0_dp
            z_value = 0.0_dp
            m_value = 0.0_dp
            first_cycle = .TRUE.
            IF (my_open_shell) THEN
               vec_Sigma_c_gw_beta = (0.0_dp, 0.0_dp)
               vec_gw_energ_beta = 0.0_dp
               vec_gw_energ_error_fit_beta = 0.0_dp
               z_value_beta = 0.0_dp
               m_value_beta = 0.0_dp
            END IF
         END IF

         ! calculate Q_PQ(it)
         IF (my_do_im_time) THEN

            first_cycle_im_time = .TRUE.
            first_cycle_im_time_beta = .TRUE.
            e_fermi = (Eigenval(homo)+Eigenval(homo+1))*0.5_dp
            tau = 0.0_dp
            tau_old = 0.0_dp

            IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="(T3,A,T66,ES15.2)") &
               "DBCSR_INFO| Eps filter for imaginary time:", eps_filter

            ! iteration over time points
            DO jquad = 1, num_integ_points

               CALL timeset(routineN//"_im_time_mat_op", handle3)

               tau_old = tau

               IF (do_minimax_quad) THEN
                  tau = tau_tj(jquad)
               ELSE
                  tau = a_scaling/TAN(tj(jquad))
               END IF

               ! compute the matrix Q(it) and Fourier transform it directly to fm_mat_Q_omega(iw)
! take care for beta case we have to set first_cycle_im_time_beta to true !!!!!!!!!!!
               CALL compute_fm_mat_Q_omega(fm_mat_Q_omega, fm_scaled_dm_occ, &
                                           fm_scaled_dm_virt, fm_mo_coeff_occ, fm_mo_coeff_virt, &
                                           fm_mat_Q, mat_M_munu_occ, mat_M_munu_virt, mat_M_munu_P_occ, &
                                           mat_M_munu_P_virt, mat_B_munu, mat_munu, trace_2d, tau_wj_2d, tj, wj, &
                                           first_cycle_im_time, do_minimax_quad, &
                                           tau, tau_old, e_fermi, eps_filter, a_scaling, alpha, &
                                           Eigenval, nmo, my_group_L_size, &
                                           my_group_L_start, dimen_RI, max_size_RI_group, ngroup, &
                                           num_integ_points, jquad, mp2_env, para_env, para_env_sub)

               first_cycle_im_time = .FALSE.

               ! the same for open shell, use fm_mo_coeff_occ_beta and fm_mo_coeff_virt_beta
               IF (my_open_shell) THEN

                  CALL compute_fm_mat_Q_omega(fm_mat_Q_omega, fm_scaled_dm_occ, &
                                              fm_scaled_dm_virt, fm_mo_coeff_occ_beta, fm_mo_coeff_virt_beta, &
                                              fm_mat_Q, mat_M_munu_occ, mat_M_munu_virt, mat_M_munu_P_occ, &
                                              mat_M_munu_P_virt, mat_B_munu, mat_munu, trace_2d, tau_wj_2d, tj, wj, &
                                              first_cycle_im_time_beta, do_minimax_quad, &
                                              tau, tau_old, e_fermi, eps_filter, a_scaling, alpha, &
                                              Eigenval, nmo, my_group_L_size, &
                                              my_group_L_start, dimen_RI, max_size_RI_group, ngroup, &
                                              num_integ_points, jquad, mp2_env, para_env, para_env_sub)

                  first_cycle_im_time_beta = .FALSE.

               END IF

               CALL timestop(handle3)

            END DO !jquad time points loop

            ALLOCATE (my_Lrows_transposed(my_group_L_size, dimen_RI))

            ! JW hack OMP statement missing
            DO iiB = 1, my_group_L_size
               DO jjB = 1, dimen_RI
                  my_Lrows_transposed(iiB, jjB) = my_Lrows(jjB, iiB)
               END DO
            END DO

            DEALLOCATE (my_Lrows)

            CALL fill_fm_with_2d_array(fm_mat_L, my_Lrows_transposed, para_env, my_group_L_start, my_group_L_size, &
                                       dimen_RI, ngroup)

            DEALLOCATE (my_Lrows_transposed)

         END IF ! do im time

         num_residues = 0

         ! get the energies for evaluating the residues
         IF (do_contour_def) THEN
            vec_Sigma_c_gw_real_freq = 0.0_dp
            IF (my_open_shell) THEN
               vec_Sigma_c_gw_real_freq_beta = 0.0_dp
            END IF

            ! determine how many residues we need for the contour deformation technique
            DO i_mo = contour_def_start, contour_def_end
               IF (i_mo <= homo) THEN
                  DO j_mo = 1, homo
                     IF (Eigenval(i_mo)-contour_def_offset < Eigenval(j_mo)) THEN
                        num_residues = num_residues+1
                     END IF
                  END DO
               ELSE
                  DO j_mo = homo+1, nmo
                     IF (Eigenval(i_mo)+contour_def_offset > Eigenval(j_mo)) THEN
                        num_residues = num_residues+1
                     END IF
                  END DO
               END IF
            END DO

            ! add residues for beta
            IF (my_open_shell) THEN
               DO i_mo = contour_def_start, contour_def_end
                  IF (i_mo <= homo_beta) THEN
                     DO j_mo = 1, homo_beta
                        IF (Eigenval_beta(i_mo)-contour_def_offset < Eigenval_beta(j_mo)) THEN
                           num_residues = num_residues+1
                        END IF
                     END DO
                  ELSE
                     DO j_mo = homo_beta+1, nmo
                        IF (Eigenval_beta(i_mo)+contour_def_offset > Eigenval_beta(j_mo)) THEN
                           num_residues = num_residues+1
                        END IF
                     END DO
                  END IF
               END DO
            END IF

            ! allocate real frequencies of the residues
            ALLOCATE (real_freq(num_residues))

            ! for getting the MO to which the residue is beloning
            ALLOCATE (mo_from_jquad(num_residues))

            ! for getting the residue (connected with the MO m)
            ALLOCATE (residue_from_jquad(num_residues))

            ! for spin information
            ALLOCATE (alpha_beta_from_jquad(num_residues))

            i_ener = 0
            ! determine which residue number belongs to which MO, residue and energy
            DO i_mo = contour_def_start, contour_def_end
               IF (i_mo <= homo) THEN
                  DO j_mo = 1, homo
                     IF (Eigenval(i_mo)-contour_def_offset < Eigenval(j_mo)) THEN
                        i_ener = i_ener+1
                        mo_from_jquad(i_ener) = i_mo
                        residue_from_jquad(i_ener) = j_mo
                        real_freq(i_ener) = Eigenval(j_mo)-Eigenval(i_mo)
                        alpha_beta_from_jquad(i_ener) = 1
                     END IF
                  END DO
               ELSE
                  DO j_mo = homo+1, nmo
                     IF (Eigenval(i_mo)+contour_def_offset > Eigenval(j_mo)) THEN
                        i_ener = i_ener+1
                        mo_from_jquad(i_ener) = i_mo
                        residue_from_jquad(i_ener) = j_mo
                        real_freq(i_ener) = Eigenval(j_mo)-Eigenval(i_mo)
                        alpha_beta_from_jquad(i_ener) = 1
                     END IF
                  END DO
               END IF
            END DO

            ! residues for beta
            IF (my_open_shell) THEN

               DO i_mo = contour_def_start, contour_def_end
                  IF (i_mo <= homo_beta) THEN
                     DO j_mo = 1, homo_beta
                        IF (Eigenval_beta(i_mo)-contour_def_offset < Eigenval_beta(j_mo)) THEN
                           i_ener = i_ener+1
                           mo_from_jquad(i_ener) = i_mo
                           residue_from_jquad(i_ener) = j_mo
                           real_freq(i_ener) = Eigenval_beta(j_mo)-Eigenval_beta(i_mo)
                           alpha_beta_from_jquad(i_ener) = 2
                        END IF
                     END DO
                  ELSE
                     DO j_mo = homo_beta+1, nmo
                        IF (Eigenval_beta(i_mo)+contour_def_offset > Eigenval_beta(j_mo)) THEN
                           i_ener = i_ener+1
                           mo_from_jquad(i_ener) = i_mo
                           residue_from_jquad(i_ener) = j_mo
                           real_freq(i_ener) = Eigenval_beta(j_mo)-Eigenval_beta(i_mo)
                           alpha_beta_from_jquad(i_ener) = 2
                        END IF
                     END DO
                  END IF
               END DO

            END IF

         END IF

         DO jquad = 1, num_integ_points+num_residues

            IF (MODULO(jquad, num_integ_group) /= color_rpa_group) CYCLE

            CALL timeset(routineN//"_RPA_matrix_operations", handle3)

            IF (jquad <= num_integ_points) THEN

               IF (do_minimax_quad) THEN
                  omega = tj(jquad)
               ELSE
                  omega = a_scaling/TAN(tj(jquad))
               END IF

            ELSE
               ! we do contour deformation for GW, where we have a real frequency
               omega = real_freq(jquad-num_integ_points)

            END IF

            IF (my_do_im_time) THEN

               ! hack; is necessary that buffer_release does not complain
               CALL fm_redistribute(fm_mat_Q_gemm, fm_mat_Q, RPA_proc_map, buffer_rec, buffer_send, &
                                    number_of_send, &
                                    map_send_size, map_rec_size, local_size_source, para_env_RPA)

               ! in case we do imag time, we already calculated fm_mat_Q by a Fourier transform from im. time
               CALL cp_fm_to_fm(fm_mat_Q_omega(jquad)%matrix, fm_mat_Q)

               ! multiplication with RI metric/Coulomb operator
               CALL cp_gemm('N', 'T', dimen_RI, dimen_RI, dimen_RI, 1.0_dp, fm_mat_Q, fm_mat_L, &
                            0.0_dp, fm_mat_work)

               CALL cp_gemm('N', 'N', dimen_RI, dimen_RI, dimen_RI, 1.0_dp, fm_mat_L, fm_mat_work, &
                            0.0_dp, fm_mat_Q)

            ELSE

               ! get info of fm_mat_S
               CALL cp_fm_get_info(matrix=fm_mat_S, &
                                   nrow_local=nrow_local, &
                                   ncol_local=ncol_local, &
                                   row_indices=row_indices, &
                                   col_indices=col_indices)

               ! remove eigenvalue part of S matrix from the last eigenvalue self-c. cycle
               IF (first_cycle .AND. count_ev_sc_GW > 1) THEN
!$OMP             PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,iocc,avirt,eigen_diff,i_global,j_global) &
!$OMP                         SHARED(ncol_local,nrow_local,col_indices,row_indices,Eigenval_last,fm_mat_S,virtual,homo,omega_old,&
!$OMP                                do_contour_def,fm_mat_S_contour_def)
                  DO jjB = 1, ncol_local
                     j_global = col_indices(jjB)
                     DO iiB = 1, nrow_local
                        i_global = row_indices(iiB)

                        iocc = MAX(1, i_global-1)/virtual+1
                        avirt = i_global-(iocc-1)*virtual
                        eigen_diff = Eigenval_last(avirt+homo)-Eigenval_last(iocc)

                        ! for the contour deformation, we have the original B-matrix
                        IF (do_contour_def) THEN
                           fm_mat_S%local_data(iiB, jjB) = fm_mat_S_contour_def%local_data(iiB, jjB)
                        ELSE
                           fm_mat_S%local_data(iiB, jjB) = fm_mat_S%local_data(iiB, jjB)/ &
                                                           SQRT(eigen_diff/(eigen_diff**2+omega_old**2))
                        END IF

                     END DO
                  END DO

               END IF

               ! update G matrix with the new value of omega
               IF (first_cycle) THEN
                  ! In this case just update the matrix (symmetric form) witrh
                  ! SQRT((epsi_a-epsi_i)/((epsi_a-epsi_i)**2+omega**2))
!$OMP             PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,iocc,avirt,eigen_diff,i_global,j_global) &
!$OMP                         SHARED(ncol_local,nrow_local,col_indices,row_indices,Eigenval,fm_mat_S,virtual,homo,omega)
                  DO jjB = 1, ncol_local
                     j_global = col_indices(jjB)
                     DO iiB = 1, nrow_local
                        i_global = row_indices(iiB)

                        iocc = MAX(1, i_global-1)/virtual+1
                        avirt = i_global-(iocc-1)*virtual
                        eigen_diff = Eigenval(avirt+homo)-Eigenval(iocc)

                        fm_mat_S%local_data(iiB, jjB) = fm_mat_S%local_data(iiB, jjB)* &
                                                        SQRT(eigen_diff/(eigen_diff**2+omega**2))

                     END DO
                  END DO
               ELSE
                  ! In this case the update has to remove the old omega component thus
                  ! SQRT(((epsi_a-epsi_i)**2+omega_old**2)/((epsi_a-epsi_i)**2+omega**2))
!$OMP             PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,iocc,avirt,eigen_diff,i_global,j_global) &
!$OMP                         SHARED(ncol_local,nrow_local,col_indices,row_indices,Eigenval,&
!$OMP                                fm_mat_S,virtual,homo,omega,omega_old,real_freq,jquad,&
!$OMP                                num_integ_points,fm_mat_S_contour_def)
                  DO jjB = 1, ncol_local
                     j_global = col_indices(jjB)
                     DO iiB = 1, nrow_local
                        i_global = row_indices(iiB)

                        iocc = MAX(1, i_global-1)/virtual+1
                        avirt = i_global-(iocc-1)*virtual
                        eigen_diff = Eigenval(avirt+homo)-Eigenval(iocc)

                        IF (jquad <= num_integ_points) THEN

                           fm_mat_S%local_data(iiB, jjB) = fm_mat_S%local_data(iiB, jjB)* &
                                                           SQRT((eigen_diff**2+omega_old**2)/(eigen_diff**2+omega**2))

                        ELSE

                           ! reset
                           fm_mat_S%local_data(iiB, jjB) = fm_mat_S_contour_def%local_data(iiB, jjB)

                           ! update
                           fm_mat_S%local_data(iiB, jjB) = fm_mat_S%local_data(iiB, jjB)*eigen_diff/(eigen_diff**2-omega**2)

                        END IF
                     END DO
                  END DO
               END IF

               ! alpha=4.0 is valid only for closed shell systems (alpha is 2.0 in the spin-orbital basis)
               t_start = m_walltime()
               SELECT CASE (mm_style)
               CASE (wfc_mm_style_gemm)
                  ! waste-fully computes the full symmetrix matrix, but maybe faster than cp_fm_syrk for optimized cp_fm_gemm
                  IF (jquad <= num_integ_points) THEN
                     ! RPA for imaginary frequencies
                     CALL cp_gemm(transa="T", transb="N", m=dimen_RI, n=dimen_RI, k=dimen_ia, alpha=alpha, &
                                  matrix_a=fm_mat_S, matrix_b=fm_mat_S, beta=0.0_dp, &
                                  matrix_c=fm_mat_Q_gemm)
                  ELSE
                     ! GW contour deformation for real frequencies with fm_mat_S_contour_def
                     CALL cp_gemm(transa="T", transb="N", m=dimen_RI, n=dimen_RI, k=dimen_ia, alpha=alpha, &
                                  matrix_a=fm_mat_S, matrix_b=fm_mat_S_contour_def, beta=0.0_dp, &
                                  matrix_c=fm_mat_Q_gemm)
                  END IF
               CASE (wfc_mm_style_syrk)
                  ! will only compute the upper half of the matrix, which is fine, since we only use it for cholesky later
                  CALL cp_fm_syrk(uplo='U', trans='T', k=dimen_ia, alpha=alpha, matrix_a=fm_mat_S, &
                                  ia=1, ja=1, beta=0.0_dp, matrix_c=fm_mat_Q_gemm)
                  IF (jquad > num_integ_points) THEN
                     CALL cp__b("rpa_ri_gpw.F",2733,"GW contour deformation does not work with wfc_mm_style_syrk.")
                  END IF
               CASE DEFAULT
                  CALL cp__b("rpa_ri_gpw.F",2736,"")
               END SELECT
               t_end = m_walltime()

               actual_flop_rate = 2.0_dp*REAL(dimen_ia, KIND=dp)*dimen_RI*REAL(dimen_RI, KIND=dp)
               actual_flop_rate = actual_flop_rate/(MAX(0.01_dp, t_end-t_start))
               IF (para_env_RPA%mepos == 0) my_flop_rate = my_flop_rate+actual_flop_rate
               my_num_dgemm_call = my_num_dgemm_call+1

               ! copy/redistribute fm_mat_Q_gemm to fm_mat_Q
               CALL cp_fm_set_all(matrix=fm_mat_Q, alpha=0.0_dp)
               CALL fm_redistribute(fm_mat_Q_gemm, fm_mat_Q, RPA_proc_map, buffer_rec, buffer_send, &
                                    number_of_send, &
                                    map_send_size, map_rec_size, local_size_source, para_env_RPA)

               IF (my_open_shell) THEN
                  ! the same for the beta spin
                  !XXX ! copy fm_mat_S into fm_mat_G
                  !XXX CALL cp_fm_to_fm(source=fm_mat_S_beta,destination=fm_mat_G_beta)
                  !XXX ! get info of fm_mat_G_beta
                  !XXX CALL cp_fm_get_info(matrix=fm_mat_G_beta,&
                  !XXX                     nrow_local=nrow_local,&
                  !XXX                     ncol_local=ncol_local,&
                  !XXX                     row_indices=row_indices,&
                  !XXX                     col_indices=col_indices)
                  !XXX

                  ! get info of fm_mat_S_beta
                  CALL cp_fm_get_info(matrix=fm_mat_S_beta, &
                                      nrow_local=nrow_local, &
                                      ncol_local=ncol_local, &
                                      row_indices=row_indices, &
                                      col_indices=col_indices)

                  ! remove eigenvalue part of S matrix from the last eigenvalue self-c. cycle
                  IF (first_cycle .AND. count_ev_sc_GW > 1) THEN
!$OMP               PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,iocc,avirt,eigen_diff,i_global,j_global) &
!$OMP                           SHARED(ncol_local,nrow_local,col_indices,row_indices,Eigenval_last_beta,&
!$OMP                                  fm_mat_S_beta,virtual_beta,homo_beta,omega_old,do_contour_def,&
!$OMP                                  fm_mat_S_contour_def_beta)
                     DO jjB = 1, ncol_local
                        j_global = col_indices(jjB)
                        DO iiB = 1, nrow_local
                           i_global = row_indices(iiB)

                           iocc = MAX(1, i_global-1)/virtual_beta+1
                           avirt = i_global-(iocc-1)*virtual_beta
                           eigen_diff = Eigenval_last_beta(avirt+homo_beta)-Eigenval_last_beta(iocc)

                           ! for the contour deformation, the last omega was a residue, so negative sign
                           IF (do_contour_def) THEN
                              fm_mat_S_beta%local_data(iiB, jjB) = fm_mat_S_contour_def_beta%local_data(iiB, jjB)
                           ELSE
                              fm_mat_S_beta%local_data(iiB, jjB) = fm_mat_S_beta%local_data(iiB, jjB)/ &
                                                                   SQRT(eigen_diff/(eigen_diff**2+omega_old**2))
                           END IF

                        END DO

                     END DO

                  END IF

                  ! update G matrix with the new value of omega
                  IF (first_cycle) THEN
!$OMP               PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,iocc,avirt,eigen_diff,i_global,j_global) &
!$OMP                           SHARED(ncol_local,nrow_local,col_indices,row_indices,&
!$OMP                                  Eigenval_beta,fm_mat_S_beta,virtual_beta,homo_beta,omega)
                     DO jjB = 1, ncol_local
                        j_global = col_indices(jjB)
                        DO iiB = 1, nrow_local
                           i_global = row_indices(iiB)

                           iocc = MAX(1, i_global-1)/virtual_beta+1
                           avirt = i_global-(iocc-1)*virtual_beta
                           eigen_diff = Eigenval_beta(avirt+homo_beta)-Eigenval_beta(iocc)

                           fm_mat_S_beta%local_data(iiB, jjB) = fm_mat_S_beta%local_data(iiB, jjB)* &
                                                                SQRT(eigen_diff/(eigen_diff**2+omega**2))

                        END DO
                     END DO
                  ELSE
!$OMP               PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,iocc,avirt,eigen_diff,i_global,j_global) &
!$OMP                           SHARED(ncol_local,nrow_local,col_indices,row_indices,jquad,num_integ_points,&
!$OMP                                  Eigenval_beta,fm_mat_S_beta,virtual_beta,homo_beta,omega,omega_old, &
!$OMP                                  fm_mat_S_contour_def_beta)
                     DO jjB = 1, ncol_local
                        j_global = col_indices(jjB)
                        DO iiB = 1, nrow_local
                           i_global = row_indices(iiB)

                           iocc = MAX(1, i_global-1)/virtual_beta+1
                           avirt = i_global-(iocc-1)*virtual_beta
                           eigen_diff = Eigenval_beta(avirt+homo_beta)-Eigenval_beta(iocc)

                           IF (jquad <= num_integ_points) THEN

                              fm_mat_S_beta%local_data(iiB, jjB) = fm_mat_S_beta%local_data(iiB, jjB)* &
                                                                   SQRT((eigen_diff**2+omega_old**2)/(eigen_diff**2+omega**2))

                           ELSE

                              ! reset
                              fm_mat_S_beta%local_data(iiB, jjB) = fm_mat_S_contour_def_beta%local_data(iiB, jjB)

                              ! update
                              fm_mat_S_beta%local_data(iiB, jjB) = &
                                 fm_mat_S_beta%local_data(iiB, jjB)*eigen_diff/(eigen_diff**2-omega**2)

                           END IF

                        END DO
                     END DO
                  END IF

                  ! alpha=4.0 is valid only for closed shell systems (alpha is 2.0 in the spin-orbital basis)
                  t_start = m_walltime()
                  SELECT CASE (mm_style)
                  CASE (wfc_mm_style_gemm)
                     IF (jquad <= num_integ_points) THEN
                        CALL cp_gemm(transa="T", transb="N", m=dimen_RI, n=dimen_RI, k=dimen_ia_beta, alpha=alpha, &
                                     matrix_a=fm_mat_S_beta, matrix_b=fm_mat_S_beta, beta=0.0_dp, &
                                     matrix_c=fm_mat_Q_gemm_beta)
                     ELSE
                        CALL cp_gemm(transa="T", transb="N", m=dimen_RI, n=dimen_RI, k=dimen_ia_beta, alpha=alpha, &
                                     matrix_a=fm_mat_S_beta, matrix_b=fm_mat_S_contour_def_beta, beta=0.0_dp, &
                                     matrix_c=fm_mat_Q_gemm_beta)
                     END IF
                  CASE (wfc_mm_style_syrk)
                     ! will only compute the upper half of the matrix, which is fine, since we only use it for cholesky later
                     CALL cp_fm_syrk(uplo='U', trans='T', k=dimen_ia, alpha=alpha, matrix_a=fm_mat_S, &
                                     ia=1, ja=1, beta=0.0_dp, matrix_c=fm_mat_Q_gemm)
                  CASE DEFAULT
                     CALL cp__b("rpa_ri_gpw.F",2870,"")
                  END SELECT

                  t_end = m_walltime()
                  actual_flop_rate = 2.0_dp*REAL(dimen_ia_beta, KIND=dp)* &
                                     dimen_RI*REAL(dimen_RI, KIND=dp)/(MAX(0.01_dp, t_end-t_start))
                  IF (para_env_RPA%mepos == 0) my_flop_rate = my_flop_rate+actual_flop_rate
                  my_num_dgemm_call = my_num_dgemm_call+1

                  ! copy/redistribute fm_mat_Q_gemm to fm_mat_Q
                  CALL cp_fm_set_all(matrix=fm_mat_Q_beta, alpha=0.0_dp)
                  CALL fm_redistribute(fm_mat_Q_gemm_beta, fm_mat_Q_beta, RPA_proc_map, buffer_rec_beta, buffer_send_beta, &
                                       number_of_send_beta, &
                                       map_send_size_beta, map_rec_size_beta, local_size_source_beta, para_env_RPA)

                  CALL cp_fm_scale_and_add(alpha=1.0_dp, matrix_a=fm_mat_Q, beta=1.0_dp, matrix_b=fm_mat_Q_beta)
                  ! fm_mat_Q%local_data=fm_mat_Q%local_data+fm_mat_Q_beta%local_data

               END IF ! open shell

            END IF ! im time

            ! get info of fm_mat_Q
            CALL cp_fm_get_info(matrix=fm_mat_Q, &
                                nrow_local=nrow_local, &
                                ncol_local=ncol_local, &
                                row_indices=row_indices, &
                                col_indices=col_indices)

            ! calcualte the trace of Q and add 1 on the diagonal
            trace_Qomega = 0.0_dp
!$OMP         PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,i_global,j_global) &
!$OMP                     SHARED(ncol_local,nrow_local,col_indices,row_indices,trace_Qomega,fm_mat_Q,dimen_RI)
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               DO iiB = 1, nrow_local
                  i_global = row_indices(iiB)
                  IF (j_global == i_global .AND. i_global <= dimen_RI) THEN
                     trace_Qomega(i_global) = fm_mat_Q%local_data(iiB, jjB)
                     fm_mat_Q%local_data(iiB, jjB) = fm_mat_Q%local_data(iiB, jjB)+1.0_dp
                  END IF
               END DO
            END DO
            CALL mp_sum(trace_Qomega, para_env_RPA%group)

            IF (jquad <= num_integ_points) THEN

               ! calculate Trace(Log(Matrix)) as Log(DET(Matrix)) via cholesky decomposition
               CALL cp_fm_cholesky_decompose(matrix=fm_mat_Q, n=dimen_RI, info_out=info_chol)
               IF(.NOT.(info_chol == 0))CALL cp__a("rpa_ri_gpw.F",2919)

               ! get info of cholesky_decomposed(fm_mat_Q)
               CALL cp_fm_get_info(matrix=fm_mat_Q, &
                                   nrow_local=nrow_local, &
                                   ncol_local=ncol_local, &
                                   row_indices=row_indices, &
                                   col_indices=col_indices)

               Q_log = 0.0_dp
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,i_global,j_global) &
!$OMP                       SHARED(ncol_local,nrow_local,col_indices,row_indices,Q_log,fm_mat_Q,dimen_RI)
               DO jjB = 1, ncol_local
                  j_global = col_indices(jjB)
                  DO iiB = 1, nrow_local
                     i_global = row_indices(iiB)
                     IF (j_global == i_global .AND. i_global <= dimen_RI) THEN
                        Q_log(i_global) = 2.0_dp*LOG(fm_mat_Q%local_data(iiB, jjB))
                     END IF
                  END DO
               END DO
               CALL mp_sum(Q_log, para_env_RPA%group)

               FComega = 0.0_dp
               DO iiB = 1, dimen_RI
                  IF (MODULO(iiB, para_env_RPA%num_pe) /= para_env_RPA%mepos) CYCLE
                  ! FComega=FComega+(LOG(Q_log(iiB))-trace_Qomega(iiB))/2.0_dp
                  FComega = FComega+(Q_log(iiB)-trace_Qomega(iiB))/2.0_dp
               END DO

               Erpa = Erpa+FComega*wj(jquad)

            END IF

            ! save omega and reset the first_cycle flag
            first_cycle = .FALSE.
            omega_old = omega

            CALL timestop(handle3)

            ! the actual G0W0 calculation
            IF (my_do_gw) THEN
               CALL timeset(routineN//"_G0W0_matrix_operations", handle2)

               IF (jquad <= num_integ_points) THEN

                  ! calculate [1+Q(iw')]^-1
                  CALL cp_fm_cholesky_invert(fm_mat_Q)
                  ! symmetrize the result, fm_mat_R_gw is only temporary work matrix
                  CALL cp_fm_upper_to_full(fm_mat_Q, fm_mat_R_gw)

               ELSE

                  ! inverted matrix is written to fm_mat_R_gw
                  CALL cp_fm_invert(fm_mat_Q, fm_mat_R_gw)

                  CALL cp_fm_to_fm(source=fm_mat_R_gw, destination=fm_mat_Q)

               END IF

               ! subtract 1 from the diagonal to get rid of exchange self-energy
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(jjB,iiB,i_global,j_global) &
!$OMP                       SHARED(ncol_local,nrow_local,col_indices,row_indices,fm_mat_Q,dimen_RI)
               DO jjB = 1, ncol_local
                  j_global = col_indices(jjB)
                  DO iiB = 1, nrow_local
                     i_global = row_indices(iiB)
                     IF (j_global == i_global .AND. i_global <= dimen_RI) THEN
                        fm_mat_Q%local_data(iiB, jjB) = fm_mat_Q%local_data(iiB, jjB)-1.0_dp
                     END IF
                  END DO
               END DO

               ! copy/redistribute fm_mat_Q to fm_mat_Q_gemm
               !        CALL cp_fm_set_all(matrix=fm_mat_Q_gemm,alpha=0.0_dp)
               !        CALL fm_redistribute(fm_mat_Q,fm_mat_Q_gemm,RPA_proc_map,buffer_rec_gw,buffer_send_gw,&
               !                             number_of_send_gw,&
               !                             map_send_size_gw,map_rec_size_gw,local_size_source_gw,para_env_RPA)

               ! S_work_(nm)Q = B_(nm)P * ([1+Q]^-1-1)_PQ
               CALL timeset(routineN//"_mult_B_f(Pi)_gw", handle3)
               !        CALL cp_gemm(transa="N",transb="N",m=dimen_nm_gw,n=dimen_RI,k=dimen_RI,alpha=1.0_dp,&
               !                     matrix_a=fm_mat_S_gw,matrix_b=fm_mat_Q_gemm,beta=0.0_dp,&
               !                     matrix_c=fm_mat_S_gw_work)
               CALL cp_gemm(transa="N", transb="N", m=dimen_nm_gw, n=dimen_RI, k=dimen_RI, alpha=1.0_dp, &
                            matrix_a=fm_mat_S_gw, matrix_b=fm_mat_Q, beta=0.0_dp, &
                            matrix_c=fm_mat_S_gw_work)
               IF (my_open_shell) THEN
                  CALL cp_gemm(transa="N", transb="N", m=dimen_nm_gw, n=dimen_RI, k=dimen_RI, alpha=1.0_dp, &
                               matrix_a=fm_mat_S_gw_beta, matrix_b=fm_mat_Q, beta=0.0_dp, &
                               matrix_c=fm_mat_S_gw_work_beta)
               END IF

               CALL timestop(handle3)

               ! vector W_(nm) = S_work_(nm)Q * [B_(nm)Q]^T

               CALL cp_fm_get_info(matrix=fm_mat_S_gw, &
                                   nrow_local=nrow_local, &
                                   ncol_local=ncol_local, &
                                   row_indices=row_indices, &
                                   col_indices=col_indices)

               vec_W_gw = 0.0_dp
               IF (my_open_shell) THEN
                  vec_W_gw_beta = 0.0_dp
               END IF

               DO iiB = 1, nrow_local
                  nm_global = row_indices(iiB)
                  DO jjB = 1, ncol_local
                     vec_W_gw(nm_global) = vec_W_gw(nm_global)+ &
                                           fm_mat_S_gw_work%local_data(iiB, jjB)*fm_mat_S_gw%local_data(iiB, jjB)
                     IF (my_open_shell) THEN
                        vec_W_gw_beta(nm_global) = vec_W_gw_beta(nm_global)+ &
                                                   fm_mat_S_gw_work_beta%local_data(iiB, jjB)*fm_mat_S_gw_beta%local_data(iiB, jjB)
                     END IF
                  END DO

                  ! transform the index nm of vec_W_gw back to n and m, formulae copied from Mauro's code
                  n_global = MAX(1, nm_global-1)/nmo+1
                  m_global = nm_global-(n_global-1)*nmo
                  n_global = n_global+homo-gw_corr_lev_occ

                  IF (my_open_shell) THEN
                     n_global_beta = MAX(1, nm_global-1)/nmo+1
                     m_global_beta = nm_global-(n_global_beta-1)*nmo
                     n_global_beta = n_global_beta+homo_beta-gw_corr_lev_occ_beta
                  END IF

                  ! compute self-energy for imaginary frequencies
                  IF (jquad <= num_integ_points) THEN

                     DO iquad = 1, num_fit_points

                        ! for occ orbitals, we compute the self-energy for negative frequencies
                        IF (n_global <= homo) THEN
                           sign_occ_virt = -1.0_dp
                        ELSE
                           sign_occ_virt = 1.0_dp
                        END IF

                        omega_i = vec_omega_fit_gw(iquad)*sign_occ_virt

                        ! set the Fermi energy for occ orbitals slightly above the HOMO and
                        ! for virt orbitals slightly below the LUMO
                        IF (n_global <= homo) THEN
                           e_fermi = Eigenval(homo)+fermi_level_offset
                        ELSE
                           e_fermi = Eigenval(homo+1)-fermi_level_offset
                        END IF

                        ! update the self-energy (use that vec_W_gw(iw) is symmetric), divide the integration
                        ! weight by 2, because the integration is from -infty to +infty and not just 0 to +infty
                        ! as for RPA, also we need for virtual orbitals a complex conjugate
                        vec_Sigma_c_gw(n_global-homo+gw_corr_lev_occ, iquad) = &
                           vec_Sigma_c_gw(n_global-homo+gw_corr_lev_occ, iquad)- &
                           0.5_dp/pi*wj(jquad)/2.0_dp*vec_W_gw(nm_global)/(im_unit*(omega+omega_i)+e_fermi-Eigenval(m_global))- &
                           0.5_dp/pi*wj(jquad)/2.0_dp*vec_W_gw(nm_global)/(im_unit*(-omega+omega_i)+e_fermi-Eigenval(m_global))

                        ! the same for beta
                        IF (my_open_shell) THEN
                           ! for occ orbitals, we compute the self-energy for negative frequencies
                           IF (n_global_beta <= homo_beta) THEN
                              sign_occ_virt = -1.0_dp
                           ELSE
                              sign_occ_virt = 1.0_dp
                           END IF

                           omega_i = vec_omega_fit_gw(iquad)*sign_occ_virt

                           ! set the Fermi energy for occ orbitals slightly above the HOMO and
                           ! for virt orbitals slightly below the LUMO
                           IF (n_global_beta <= homo_beta) THEN
                              e_fermi = Eigenval_beta(homo_beta)+fermi_level_offset
                           ELSE
                              e_fermi = Eigenval_beta(homo_beta+1)-fermi_level_offset
                           END IF

                           ! update the self-energy (use that vec_W_gw(iw) is symmetric), divide the integration
                           ! weight by 2, because the integration is from -infty to +infty and not just 0 to +infty
                           ! as for RPA, also we need for virtual orbitals a complex conjugate
                           vec_Sigma_c_gw_beta(n_global_beta-homo_beta+gw_corr_lev_occ_beta, iquad) = &
                              vec_Sigma_c_gw_beta(n_global_beta-homo_beta+gw_corr_lev_occ_beta, iquad)- &
                              0.5_dp/pi*wj(jquad)/2.0_dp*vec_W_gw_beta(nm_global)/ &
                              (im_unit*(omega+omega_i)+e_fermi-Eigenval_beta(m_global))- &
                              0.5_dp/pi*wj(jquad)/2.0_dp*vec_W_gw_beta(nm_global)/ &
                              (im_unit*(-omega+omega_i)+e_fermi-Eigenval_beta(m_global))
                        END IF
                     END DO ! iquad
                  END IF ! check imaginary frequency

                  IF (do_contour_def) THEN

                     IF (my_open_shell) THEN

                        CALL contour_def_integrate_and_residues(vec_Sigma_c_gw_real_freq, wj, vec_W_gw, Eigenval, &
                                                                omega, contour_def_offset, e_fermi, &
                                                                mo_from_jquad, residue_from_jquad, alpha_beta_from_jquad, &
                                                                contour_def_start, contour_def_end, n_global, jquad, &
                                                                num_integ_points, m_global, nm_global, &
                                                                alpha_beta_case=.TRUE., alpha=.TRUE.)

                        CALL contour_def_integrate_and_residues(vec_Sigma_c_gw_real_freq_beta, wj, vec_W_gw_beta, Eigenval_beta, &
                                                                omega, contour_def_offset, e_fermi, &
                                                                mo_from_jquad, residue_from_jquad, alpha_beta_from_jquad, &
                                                                contour_def_start, contour_def_end, n_global_beta, jquad, &
                                                                num_integ_points, m_global_beta, nm_global, &
                                                                alpha_beta_case=.TRUE., beta=.TRUE.)

                     ELSE

                        CALL contour_def_integrate_and_residues(vec_Sigma_c_gw_real_freq, wj, vec_W_gw, Eigenval, &
                                                                omega, contour_def_offset, e_fermi, &
                                                                mo_from_jquad, residue_from_jquad, alpha_beta_from_jquad, &
                                                                contour_def_start, contour_def_end, n_global, jquad, &
                                                                num_integ_points, m_global, nm_global)
                     END IF

                  END IF ! contour deformation

               END DO ! iiB

               CALL timestop(handle2)

            END IF ! GW

         END DO ! jquad

         CALL mp_sum(Erpa, para_env%group)
         Erpa = Erpa/(pi*2.0_dp)
         IF (do_minimax_quad) Erpa = Erpa/2.0_dp

         IF (.NOT. my_do_im_time) THEN

            IF (para_env_RPA%mepos == 0) my_flop_rate = my_flop_rate/REAL(MAX(my_num_dgemm_call, 1), KIND=dp)/1.0E9_dp
            CALL mp_sum(my_flop_rate, para_env%group)
            my_flop_rate = my_flop_rate/para_env%num_pe
            IF (unit_nr > 0) WRITE (UNIT=unit_nr, FMT="(T3,A,T66,F15.2)") &
               "PERFORMANCE| PDGEMM flop rate (Gflops / MPI rank):", my_flop_rate
         END IF

         ! G0W0 postprocessing: Fitting + correction of MO energies
         IF (my_do_gw) THEN

            CALL timeset(routineN//"_G0W0_fit", handle3)

            CALL mp_sum(vec_Sigma_c_gw, para_env%group)

            IF (my_open_shell) THEN
               CALL mp_sum(vec_Sigma_c_gw_beta, para_env%group)
            END IF

            IF (do_contour_def) THEN

               CALL mp_sum(vec_Sigma_c_gw_real_freq, para_env%group)

               IF (my_open_shell) THEN
                  CALL mp_sum(vec_Sigma_c_gw_real_freq_beta, para_env%group)
               END IF

            END IF

            CALL mp_sync(para_env%group)

            ! fit the self-energy on imaginary frequency axis and evaluate the fit on the MO energy of the SCF
            DO n_level_gw = 1, gw_corr_lev_tot
               ! processes perform different fits
               IF (MODULO(n_level_gw, para_env%num_pe) /= para_env%mepos) CYCLE

               CALL fit_and_continuation(vec_gw_energ, vec_gw_energ_error_fit, vec_omega_fit_gw, &
                                         z_value, m_value, vec_Sigma_c_gw, mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 1), &
                                         Eigenval, Eigenval_scf, n_level_gw, gw_corr_lev_occ, num_poles, &
                                         num_fit_points, max_iter_fit, crossing_search, homo, check_fit, stop_crit, &
                                         fermi_level_offset)

               IF (my_open_shell) THEN
                  CALL fit_and_continuation( &
                     vec_gw_energ_beta, vec_gw_energ_error_fit_beta, vec_omega_fit_gw, &
                     z_value_beta, m_value_beta, vec_Sigma_c_gw_beta, mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 2), &
                     Eigenval_beta, Eigenval_scf_beta, n_level_gw, &
                     gw_corr_lev_occ_beta, num_poles, &
                     num_fit_points, max_iter_fit, crossing_search, homo_beta, check_fit, stop_crit, &
                     fermi_level_offset)

               END IF

            END DO ! n_level_gw

            CALL mp_sum(vec_gw_energ_error_fit, para_env%group)
            CALL mp_sum(vec_gw_energ, para_env%group)
            CALL mp_sum(z_value, para_env%group)
            CALL mp_sum(m_value, para_env%group)

            IF (my_open_shell) THEN
               CALL mp_sum(vec_gw_energ_error_fit_beta, para_env%group)
               CALL mp_sum(vec_gw_energ_beta, para_env%group)
               CALL mp_sum(z_value_beta, para_env%group)
               CALL mp_sum(m_value_beta, para_env%group)
            END IF

            remove_neg_virt_energies = mp2_env%ri_g0w0%remove_neg_virt_energies

            ! in case we do the contour deformation technique, we transfer vec_Sigma_c_gw_real_freq to
            ! vec_gw_energ and we compute the z-value, m-value
            ! the results from the analytic continuation are therefore overwritten
            IF (do_contour_def) THEN

               IF (unit_nr > 0) THEN
                  WRITE (unit_nr, *) ''
                  WRITE (unit_nr, '(T3,A,I4,A,I3)') 'The following MOs have been corrected by contour deformation: MO', &
                     contour_def_start, '  -  MO', contour_def_end
                  WRITE (unit_nr, *) ''
               END IF

               index_contour_def = 1

               ! only correct levels for which contour deformation is enabled
               DO n_level_gw = contour_def_start-(homo-gw_corr_lev_occ), contour_def_end-(homo-gw_corr_lev_occ)

                  ! reset the values from analytic continuation
                  vec_gw_energ(n_level_gw) = 0.0_dp
                  vec_gw_energ_error_fit(n_level_gw) = 0.0_dp
                  z_value(n_level_gw) = 0.0_dp
                  m_value(n_level_gw) = 0.0_dp

                  CALL compute_z_and_m_contour_def(vec_gw_energ, z_value, m_value, &
                                                   vec_Sigma_c_gw_real_freq, contour_def_offset, &
                                                   n_level_gw, index_contour_def)

                  index_contour_def = index_contour_def+1

               END DO

               ! the same for beta
               IF (my_open_shell) THEN

                  index_contour_def = 1

                  ! only correct levels for which contour deformation is enabled
                  DO n_level_gw = contour_def_start-(homo_beta-gw_corr_lev_occ_beta), &
                                  contour_def_end-(homo_beta-gw_corr_lev_occ_beta)

                     ! reset the values from analytic continuation
                     vec_gw_energ_beta(n_level_gw) = 0.0_dp
                     vec_gw_energ_error_fit_beta(n_level_gw) = 0.0_dp
                     z_value_beta(n_level_gw) = 0.0_dp
                     m_value_beta(n_level_gw) = 0.0_dp

                     CALL compute_z_and_m_contour_def(vec_gw_energ_beta, z_value_beta, m_value_beta, &
                                                      vec_Sigma_c_gw_real_freq_beta, contour_def_offset, &
                                                      n_level_gw, index_contour_def)

                     index_contour_def = index_contour_def+1

                  END DO

               END IF

               DEALLOCATE (real_freq)
               DEALLOCATE (mo_from_jquad)
               DEALLOCATE (residue_from_jquad)
               DEALLOCATE (alpha_beta_from_jquad)

            END IF

            ! print the quasiparticle energies and update Eigenval in case you do eigenvalue self-consistent GW
            IF (my_open_shell) THEN

               CALL print_and_update_for_ev_sc( &
                  vec_gw_energ, vec_gw_energ_error_fit, &
                  z_value, m_value, mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 1), Eigenval, &
                  Eigenval_last, Eigenval_scf, gw_corr_lev_occ, gw_corr_lev_virt, gw_corr_lev_tot, &
                  count_ev_sc_GW, crossing_search, homo, nmo, unit_nr, mp2_env%ri_g0w0%print_gw_details, &
                  do_alpha=.TRUE.)

               CALL print_and_update_for_ev_sc( &
                  vec_gw_energ_beta, vec_gw_energ_error_fit_beta, &
                  z_value_beta, m_value_beta, mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 2), Eigenval_beta, &
                  Eigenval_last_beta, Eigenval_scf_beta, gw_corr_lev_occ_beta, gw_corr_lev_virt_beta, gw_corr_lev_tot, &
                  count_ev_sc_GW, crossing_search, homo_beta, nmo, unit_nr, mp2_env%ri_g0w0%print_gw_details, &
                  do_beta=.TRUE.)

            ELSE

               CALL print_and_update_for_ev_sc( &
                  vec_gw_energ, vec_gw_energ_error_fit, &
                  z_value, m_value, mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, 1), Eigenval, &
                  Eigenval_last, Eigenval_scf, gw_corr_lev_occ, gw_corr_lev_virt, gw_corr_lev_tot, &
                  count_ev_sc_GW, crossing_search, homo, nmo, unit_nr, mp2_env%ri_g0w0%print_gw_details)

            END IF

            CALL timestop(handle3)

            ! if HOMO-LUMO gap differs by less then mp2_env%ri_g0w0%eps_ev_sc_iter, exit ev sc GW loop
            IF (ABS(Eigenval(homo)-Eigenval_last(homo)-Eigenval(homo+1)+Eigenval_last(homo+1)) &
                < mp2_env%ri_g0w0%eps_ev_sc_iter) THEN
               EXIT
            END IF

         END IF ! my_do_gw if

      END DO !ev_sc_gw_loop

      ! release buffer
      CALL release_buffer(RPA_proc_map, buffer_rec, buffer_send, &
                          number_of_rec, number_of_send, &
                          map_send_size, map_rec_size, local_size_source)

      IF (my_open_shell) THEN
         CALL release_buffer(RPA_proc_map, buffer_rec_beta, buffer_send_beta, &
                             number_of_rec_beta, number_of_send_beta, &
                             map_send_size_beta, map_rec_size_beta, local_size_source_beta)
      END IF

      IF (my_do_gw) THEN
         CALL cp_fm_release(fm_mat_S_gw_work)
         DEALLOCATE (vec_Sigma_c_gw)
         DEALLOCATE (vec_omega_fit_gw)
         DEALLOCATE (vec_omega_fit_gw_sign)
         DEALLOCATE (Lambda)
         DEALLOCATE (Lambda_without_offset)
         DEALLOCATE (Lambda_Re)
         DEALLOCATE (Lambda_Im)
         DEALLOCATE (z_value)
         DEALLOCATE (m_value)
         DEALLOCATE (vec_gw_energ)
         DEALLOCATE (vec_gw_energ_error_fit)
         DEALLOCATE (mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw)
         DEALLOCATE (Eigenval_last)
         DEALLOCATE (Eigenval_scf)
         IF (do_ri_Sigma_x) THEN
            DEALLOCATE (vec_Sigma_x_gw)
         END IF
         IF (my_open_shell) THEN
            CALL cp_fm_release(fm_mat_S_gw_work_beta)
            DEALLOCATE (vec_W_gw_beta)
            DEALLOCATE (vec_Sigma_c_gw_beta)
            DEALLOCATE (z_value_beta)
            DEALLOCATE (m_value_beta)
            DEALLOCATE (vec_gw_energ_beta)
            DEALLOCATE (vec_gw_energ_error_fit_beta)
            DEALLOCATE (Eigenval_last_beta)
            DEALLOCATE (Eigenval_scf_beta)
            IF (do_ri_Sigma_x) THEN
               DEALLOCATE (vec_Sigma_x_gw_beta)
            END IF
         END IF
         IF (do_contour_def) THEN
            DEALLOCATE (vec_Sigma_c_gw_real_freq)
            CALL cp_fm_release(fm_mat_S_contour_def)
            IF (my_open_shell) THEN
               DEALLOCATE (vec_Sigma_c_gw_real_freq_beta)
               CALL cp_fm_release(fm_mat_S_contour_def_beta)
            END IF
         END IF
      END IF

      IF (my_do_im_time) THEN

         CALL cp_fm_release(fm_scaled_dm_occ)
         CALL cp_fm_release(fm_scaled_dm_virt)
         CALL cp_fm_release(fm_mo_coeff_occ)
         CALL cp_fm_release(fm_mo_coeff_virt)

         IF (my_open_shell) THEN
            CALL cp_fm_release(fm_mo_coeff_occ_beta)
            CALL cp_fm_release(fm_mo_coeff_virt_beta)
         END IF

         CALL cp_fm_release(fm_mat_L)
         CALL cp_fm_release(fm_mat_work)

         CALL cp_dbcsr_release_p(mat_M_munu_occ)
         CALL cp_dbcsr_release_p(mat_M_munu_virt)

         CALL cp_dbcsr_deallocate_matrix_set(mat_M_munu_P_occ)
         CALL cp_dbcsr_deallocate_matrix_set(mat_M_munu_P_virt)

         DEALLOCATE (trace_2d)

         IF (my_do_gw) THEN
            DEALLOCATE (vec_tau_fit_gw_sign)
            DEALLOCATE (vec_W_gw_store)
         END IF

         IF (.NOT. my_do_gw) THEN
            DO iquad = 1, num_integ_points
               CALL cp_fm_release(fm_mat_Q_omega(iquad)%matrix)
            END DO
            DEALLOCATE (fm_mat_Q_omega)
         END IF

         CALL cp_dbcsr_deallocate_matrix_set(mat_B_munu)

!JW 2del      DO iii=1,my_group_L_size
!        CALL cp_dbcsr_release(mat_B_munu(iii)%matrix)
!        DEALLOCATE(mat_B_munu(iii)%matrix)
!      END DO
!      DEALLOCATE(mat_B_munu)

      END IF

      DEALLOCATE (tj)
      DEALLOCATE (wj)
      DEALLOCATE (Q_log)
      DEALLOCATE (trace_Qomega)

      CALL timestop(handle)

   END SUBROUTINE rpa_num_int

! **************************************************************************************************
!> \brief ...
!> \param a_scaling_ext ...
!> \param para_env ...
!> \param para_env_RPA ...
!> \param homo ...
!> \param virtual ...
!> \param Eigenval ...
!> \param num_integ_points ...
!> \param num_integ_group ...
!> \param color_rpa_group ...
!> \param tj_ext ...
!> \param wj_ext ...
!> \param fm_mat_S ...
!> \param homo_beta ...
!> \param virtual_beta ...
!> \param dimen_ia_beta ...
!> \param Eigenval_beta ...
!> \param fm_mat_S_beta ...
! **************************************************************************************************
   SUBROUTINE calc_scaling_factor(a_scaling_ext, para_env, para_env_RPA, homo, virtual, Eigenval, &
                                  num_integ_points, num_integ_group, color_rpa_group, &
                                  tj_ext, wj_ext, fm_mat_S, &
                                  homo_beta, virtual_beta, dimen_ia_beta, Eigenval_beta, fm_mat_S_beta)
      REAL(KIND=dp)                                      :: a_scaling_ext
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_RPA
      INTEGER                                            :: homo, virtual
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: num_integ_points, num_integ_group, &
                                                            color_rpa_group
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tj_ext, wj_ext
      TYPE(cp_fm_type), POINTER                          :: fm_mat_S
      INTEGER, OPTIONAL                                  :: homo_beta, virtual_beta, dimen_ia_beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: fm_mat_S_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_scaling_factor', &
         routineP = moduleN//':'//routineN

      INTEGER :: avirt, color_col, color_col_beta, color_row, color_row_beta, comm_col, &
         comm_col_beta, comm_row, comm_row_beta, handle, i_global, icycle, iiB, iocc, jjB, jquad, &
         ncol_local, ncol_local_beta, nrow_local, nrow_local_beta
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, col_indices_beta, &
                                                            row_indices, row_indices_beta
      LOGICAL                                            :: my_open_shell
      REAL(KIND=dp) :: a_high, a_low, a_scaling, conv_param, eigen_diff, eps, first_deriv, four, &
         left_term, one, pig, right_term, right_term_ref, right_term_ref_beta, step, two, zero
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cottj, D_ia, D_ia_beta, iaia_RI, &
                                                            iaia_RI_beta, iaia_RI_dp, &
                                                            iaia_RI_dp_beta, M_ia, M_ia_beta, tj, &
                                                            wj
      TYPE(cp_para_env_type), POINTER                    :: para_env_col, para_env_col_beta, &
                                                            para_env_row, para_env_row_beta

      CALL timeset(routineN, handle)

      my_open_shell = .FALSE.
      IF (PRESENT(homo_beta) .AND. &
          PRESENT(virtual_beta) .AND. &
          PRESENT(dimen_ia_beta) .AND. &
          PRESENT(Eigenval_beta) .AND. &
          PRESENT(fm_mat_S_beta)) my_open_shell = .TRUE.

      ZERO = 0.0_dp
      ONE = 1.0_dp
      TWO = 2.0_dp
      FOUR = 4.0_dp
      PIG = pi
      eps = 1.0E-10_dp

      ALLOCATE (cottj(num_integ_points))

      ALLOCATE (tj(num_integ_points))

      ALLOCATE (wj(num_integ_points))

      ! calculate the cotangent of the abscissa tj
      DO jquad = 1, num_integ_points
         tj(jquad) = tj_ext(jquad)
         wj(jquad) = wj_ext(jquad)
         cottj(jquad) = ONE/TAN(tj(jquad))
      END DO

      ! calculate the (ia|ia) RI integrals
      ! ----------------------------------
      ! 1) get info fm_mat_S
      !XXX CALL cp_fm_to_fm(source=fm_mat_S,destination=fm_mat_G)
      CALL cp_fm_get_info(matrix=fm_mat_S, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices)

      ! allocate the local buffer of iaia_RI integrals (dp kind)
      ALLOCATE (iaia_RI_dp(nrow_local))
      iaia_RI_dp = 0.0_dp

      ! 2) perform the local multiplication SUM_K (ia|K)*(ia|K)
      DO jjB = 1, ncol_local
         DO iiB = 1, nrow_local
            iaia_RI_dp(iiB) = iaia_RI_dp(iiB)+fm_mat_S%local_data(iiB, jjB)*fm_mat_S%local_data(iiB, jjB)
         END DO
      END DO

      ! 3) sum the result with the processes of the RPA_group having the same rows
      !          _______K_______               _
      !         |   |   |   |   |             | |
      !     --> | 1 | 5 | 9 | 13|   SUM -->   | |
      !         |___|__ |___|___|             |_|
      !         |   |   |   |   |             | |
      !     --> | 2 | 6 | 10| 14|   SUM -->   | |
      !      ia |___|___|___|___|             |_|   (ia|ia)_RI
      !         |   |   |   |   |             | |
      !     --> | 3 | 7 | 11| 15|   SUM -->   | |
      !         |___|___|___|___|             |_|
      !         |   |   |   |   |             | |
      !     --> | 4 | 8 | 12| 16|   SUM -->   | |
      !         |___|___|___|___|             |_|
      !

      color_row = fm_mat_S%matrix_struct%context%mepos(1)
      CALL mp_comm_split_direct(para_env_RPA%group, comm_row, color_row)
      NULLIFY (para_env_row)
      CALL cp_para_env_create(para_env_row, comm_row)

      CALL mp_sum(iaia_RI_dp, para_env_row%group)

      ! convert the iaia_RI_dp into double-double precision
      ALLOCATE (iaia_RI(nrow_local))
      DO iiB = 1, nrow_local
         iaia_RI(iiB) = iaia_RI_dp(iiB)
      END DO
      DEALLOCATE (iaia_RI_dp)

      ! 4) calculate the right hand term, D_ia is the matrix containing the
      ! orbital energy differences, M_ia is the diagonal of the full RPA 'excitation'
      ! matrix
      ALLOCATE (D_ia(nrow_local))

      ALLOCATE (M_ia(nrow_local))

      DO iiB = 1, nrow_local
         i_global = row_indices(iiB)

         iocc = MAX(1, i_global-1)/virtual+1
         avirt = i_global-(iocc-1)*virtual
         eigen_diff = Eigenval(avirt+homo)-Eigenval(iocc)

         D_ia(iiB) = eigen_diff
      END DO

      DO iiB = 1, nrow_local
         M_ia(iiB) = D_ia(iiB)*D_ia(iiB)+TWO*D_ia(iiB)*iaia_RI(iiB)
      END DO

      right_term_ref = ZERO
      DO iiB = 1, nrow_local
         right_term_ref = right_term_ref+(SQRT(M_ia(iiB))-D_ia(iiB)-iaia_RI(iiB))
      END DO
      right_term_ref = right_term_ref/TWO
      ! right_term_ref=accurate_sum((SQRT(M_ia)-D_ia-iaia_RI))/2.0_dp

      ! sum the result with the processes of the RPA_group having the same col
      color_col = fm_mat_S%matrix_struct%context%mepos(2)
      CALL mp_comm_split_direct(para_env_RPA%group, comm_col, color_col)
      NULLIFY (para_env_col)
      CALL cp_para_env_create(para_env_col, comm_col)

      ! allocate communication array for columns
      CALL mp_sum(right_term_ref, para_env_col%group)

      ! In the open shell case do point 1-2-3 for the beta spin
      IF (my_open_shell) THEN
         !XXX CALL cp_fm_to_fm(source=fm_mat_S_beta,destination=fm_mat_G_beta)
         CALL cp_fm_get_info(matrix=fm_mat_S_beta, &
                             nrow_local=nrow_local_beta, &
                             ncol_local=ncol_local_beta, &
                             row_indices=row_indices_beta, &
                             col_indices=col_indices_beta)

         ALLOCATE (iaia_RI_dp_beta(nrow_local_beta))
         iaia_RI_dp_beta = 0.0_dp

         DO jjB = 1, ncol_local_beta
            DO iiB = 1, nrow_local_beta
               iaia_RI_dp_beta(iiB) = iaia_RI_dp_beta(iiB)+fm_mat_S_beta%local_data(iiB, jjB)*fm_mat_S_beta%local_data(iiB, jjB)
            END DO
         END DO

         color_row_beta = fm_mat_S_beta%matrix_struct%context%mepos(1)
         CALL mp_comm_split_direct(para_env_RPA%group, comm_row_beta, color_row_beta)
         NULLIFY (para_env_row_beta)
         CALL cp_para_env_create(para_env_row_beta, comm_row_beta)

         CALL mp_sum(iaia_RI_dp_beta, para_env_row_beta%group)

         ALLOCATE (iaia_RI_beta(nrow_local_beta))
         DO iiB = 1, nrow_local_beta
            iaia_RI_beta(iiB) = iaia_RI_dp_beta(iiB)
         END DO
         DEALLOCATE (iaia_RI_dp_beta)

         ALLOCATE (D_ia_beta(nrow_local_beta))

         ALLOCATE (M_ia_beta(nrow_local_beta))

         DO iiB = 1, nrow_local_beta
            i_global = row_indices_beta(iiB)

            iocc = MAX(1, i_global-1)/virtual_beta+1
            avirt = i_global-(iocc-1)*virtual_beta
            eigen_diff = Eigenval_beta(avirt+homo_beta)-Eigenval_beta(iocc)

            D_ia_beta(iiB) = eigen_diff
         END DO

         DO iiB = 1, nrow_local_beta
            M_ia_beta(iiB) = D_ia_beta(iiB)*D_ia_beta(iiB)+TWO*D_ia_beta(iiB)*iaia_RI_beta(iiB)
         END DO

         right_term_ref_beta = ZERO
         DO iiB = 1, nrow_local_beta
            right_term_ref_beta = right_term_ref_beta+(SQRT(M_ia_beta(iiB))-D_ia_beta(iiB)-iaia_RI_beta(iiB))
         END DO
         right_term_ref_beta = right_term_ref_beta/TWO

         ! sum the result with the processes of the RPA_group having the same col
         color_col_beta = fm_mat_S_beta%matrix_struct%context%mepos(2)
         CALL mp_comm_split_direct(para_env_RPA%group, comm_col_beta, color_col_beta)
         NULLIFY (para_env_col_beta)
         CALL cp_para_env_create(para_env_col_beta, comm_col_beta)

         CALL mp_sum(right_term_ref_beta, para_env_col_beta%group)

         right_term_ref = right_term_ref+right_term_ref_beta
      END IF

      ! bcast the result
      IF (para_env%mepos == 0) THEN
         CALL mp_bcast(right_term_ref, 0, para_env%group)
      ELSE
         right_term_ref = 0.0_dp
         CALL mp_bcast(right_term_ref, 0, para_env%group)
      END IF

      ! 5) start iteration for solving the non-linear equation by bisection
      ! find limit, here step=0.5 seems a good compromise
      conv_param = 100.0_dp*EPSILON(right_term_ref)
      step = 0.5_dp
      a_low = 0.0_dp
      a_high = step
      right_term = -right_term_ref
      DO icycle = 1, num_integ_points*2
         a_scaling = a_high

         CALL calculate_objfunc(a_scaling, left_term, first_deriv, num_integ_points, my_open_shell, &
                                ZERO, ONE, M_ia, cottj, wj, D_ia, D_ia_beta, M_ia_beta, &
                                nrow_local, nrow_local_beta, num_integ_group, color_rpa_group, &
                                para_env, para_env_row, para_env_row_beta)
         left_term = left_term/FOUR/PIG*a_scaling

         IF (ABS(left_term) > ABS(right_term) .OR. ABS(left_term+right_term) <= conv_param) EXIT
         a_low = a_high
         a_high = a_high+step

      END DO

      IF (ABS(left_term+right_term) >= conv_param) THEN
         IF (a_scaling >= 2*num_integ_points*step) THEN
            a_scaling = 1.0_dp
         ELSE

            DO icycle = 1, num_integ_points*2
               a_scaling = (a_low+a_high)/2.0_dp

               CALL calculate_objfunc(a_scaling, left_term, first_deriv, num_integ_points, my_open_shell, &
                                      ZERO, ONE, M_ia, cottj, wj, D_ia, D_ia_beta, M_ia_beta, &
                                      nrow_local, nrow_local_beta, num_integ_group, color_rpa_group, &
                                      para_env, para_env_row, para_env_row_beta)
               left_term = left_term/FOUR/PIG*a_scaling

               IF (ABS(left_term) > ABS(right_term)) THEN
                  a_high = a_scaling
               ELSE
                  a_low = a_scaling
               END IF

               IF (ABS(a_high-a_low) < 1.0D-5) EXIT

            END DO

         END IF
      END IF

      a_scaling_ext = a_scaling
      CALL mp_bcast(a_scaling_ext, 0, para_env%group)

      DEALLOCATE (cottj)
      DEALLOCATE (tj)
      DEALLOCATE (wj)
      DEALLOCATE (iaia_RI)
      DEALLOCATE (D_ia)
      DEALLOCATE (M_ia)
      CALL cp_para_env_release(para_env_row)
      CALL cp_para_env_release(para_env_col)

      IF (my_open_shell) THEN
         DEALLOCATE (iaia_RI_beta)
         DEALLOCATE (D_ia_beta)
         DEALLOCATE (M_ia_beta)
         CALL cp_para_env_release(para_env_row_beta)
         CALL cp_para_env_release(para_env_col_beta)
      END IF

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param a_scaling ...
!> \param left_term ...
!> \param first_deriv ...
!> \param num_integ_points ...
!> \param my_open_shell ...
!> \param ZERO ...
!> \param ONE ...
!> \param M_ia ...
!> \param cottj ...
!> \param wj ...
!> \param D_ia ...
!> \param D_ia_beta ...
!> \param M_ia_beta ...
!> \param nrow_local ...
!> \param nrow_local_beta ...
!> \param num_integ_group ...
!> \param color_rpa_group ...
!> \param para_env ...
!> \param para_env_row ...
!> \param para_env_row_beta ...
! **************************************************************************************************
   SUBROUTINE calculate_objfunc(a_scaling, left_term, first_deriv, num_integ_points, my_open_shell, &
                                ZERO, ONE, M_ia, cottj, wj, D_ia, D_ia_beta, M_ia_beta, &
                                nrow_local, nrow_local_beta, num_integ_group, color_rpa_group, &
                                para_env, para_env_row, para_env_row_beta)
      REAL(KIND=dp)                                      :: a_scaling, left_term, first_deriv
      INTEGER                                            :: num_integ_points
      LOGICAL                                            :: my_open_shell
      REAL(KIND=dp)                                      :: zero, one
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: M_ia, cottj, wj, D_ia, D_ia_beta, &
                                                            M_ia_beta
      INTEGER                                            :: nrow_local, nrow_local_beta, &
                                                            num_integ_group, color_rpa_group
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_row, para_env_row_beta

      INTEGER                                            :: iiB, jquad
      REAL(KIND=dp)                                      :: first_deriv_beta, left_term_beta, omega

      left_term = ZERO
      first_deriv = ZERO
      left_term_beta = ZERO
      first_deriv_beta = ZERO
      DO jquad = 1, num_integ_points
         ! parallelize over integration points
         IF (MODULO(jquad, num_integ_group) /= color_rpa_group) CYCLE
         omega = a_scaling*cottj(jquad)

         DO iiB = 1, nrow_local
            ! parallelize over ia elements in the para_env_row group
            IF (MODULO(iiB, para_env_row%num_pe) /= para_env_row%mepos) CYCLE
            ! calculate left_term
            left_term = left_term+wj(jquad)* &
                        (LOG(ONE+(M_ia(iiB)-D_ia(iiB)**2)/(omega**2+D_ia(iiB)**2))- &
                         (M_ia(iiB)-D_ia(iiB)**2)/(omega**2+D_ia(iiB)**2))
            first_deriv = first_deriv+wj(jquad)*cottj(jquad)**2* &
                          ((-M_ia(iiB)+D_ia(iiB)**2)**2/((omega**2+D_ia(iiB)**2)**2*(omega**2+M_ia(iiB))))
         END DO

         IF (my_open_shell) THEN
            DO iiB = 1, nrow_local_beta
               ! parallelize over ia elements in the para_env_row group
               IF (MODULO(iiB, para_env_row_beta%num_pe) /= para_env_row_beta%mepos) CYCLE
               ! calculate left_term
               left_term_beta = left_term_beta+wj(jquad)* &
                                (LOG(ONE+(M_ia_beta(iiB)-D_ia_beta(iiB)**2)/(omega**2+D_ia_beta(iiB)**2))- &
                                 (M_ia_beta(iiB)-D_ia_beta(iiB)**2)/(omega**2+D_ia_beta(iiB)**2))
               first_deriv_beta = &
                  first_deriv_beta+wj(jquad)*cottj(jquad)**2* &
                  ((-M_ia_beta(iiB)+D_ia_beta(iiB)**2)**2/((omega**2+D_ia_beta(iiB)**2)**2*(omega**2+M_ia_beta(iiB))))
            END DO
         END IF

      END DO

      ! sum the contribution from all proc, starting form the row group
      CALL mp_sum(left_term, para_env%group)
      CALL mp_sum(first_deriv, para_env%group)

      IF (my_open_shell) THEN
         CALL mp_sum(left_term_beta, para_env%group)
         CALL mp_sum(first_deriv_beta, para_env%group)

         left_term = left_term+left_term_beta
         first_deriv = first_deriv+first_deriv_beta
      END IF

   END SUBROUTINE calculate_objfunc

! **************************************************************************************************
!> \brief Calculate integration weights for the tau grid (in dependency of the omega node)
!> \param num_integ_points ...
!> \param tau_tj ...
!> \param tau_wj_2d ...
!> \param omega_tj ...
!> \param num_x_nodes_min ...
!> \param num_x_nodes_max ...
!> \param num_x_nodes_step ...
!> \param E_min ...
!> \param E_max ...
!> \param ls_error ...
!> \param max_error_min ...
! **************************************************************************************************
   SUBROUTINE get_rpa_least_sq_weights_tau(num_integ_points, tau_tj, tau_wj_2d, omega_tj, &
                                           num_x_nodes_min, num_x_nodes_max, num_x_nodes_step, &
                                           E_min, E_max, ls_error, max_error_min)

      INTEGER                                            :: num_integ_points
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tau_tj
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: tau_wj_2d
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: omega_tj
      INTEGER                                            :: num_x_nodes_min, num_x_nodes_max, &
                                                            num_x_nodes_step
      REAL(KIND=dp)                                      :: E_min, E_max, ls_error, max_error_min

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_rpa_least_sq_weights_tau', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iii, info, jjj, jquad, lwork, &
                                                            max_iter_fit, max_iter_fit_abord, &
                                                            num_check_chi2, num_x_nodes, x_node_min
      REAL(KIND=dp)                                      :: chi2, chi2_min, chi2_min_jquad, &
                                                            max_error, max_error_min_jquad, omega
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tau_wj_work, work_array, x_values, &
                                                            y_values
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: mat_A

      CALL timeset(routineN, handle)

      max_error_min = 0.0_dp
      chi2_min = 0.0_dp

      num_check_chi2 = 300000

      max_iter_fit = 10000000
      max_iter_fit_abord = 2000

      ! loop over all omega frequency points
      DO jquad = 1, num_integ_points

         chi2_min_jquad = 100.0_dp

         ! since the fitting is strongly dependent on the initial guess and the number of evaluations points of 2x/(x^2+omega_k^2)
         ! we search for the optimal fit among different number of evaluation points of 1/x and among different initial guesses
         ! NINT is the function to round down

         DO num_x_nodes = num_x_nodes_min, num_x_nodes_max, num_x_nodes_step

            ALLOCATE (x_values(num_x_nodes))

            ! set the x-values uniformly in the interval [Emin,Emax]
            DO iii = 1, num_x_nodes
               x_values(iii) = E_min+(iii-1.0_dp)/(num_x_nodes-1.0_dp)*(E_max-E_min)
            END DO

            ALLOCATE (y_values(num_x_nodes))

            ALLOCATE (mat_A(num_x_nodes, num_integ_points))
            mat_A = 0.0_dp

            ALLOCATE (tau_wj_work(num_integ_points))
            tau_wj_work = 0.0_dp

            ALLOCATE (work_array(2*num_integ_points))
            work_array = 0.0_dp

            omega = omega_tj(jquad)

            ! y=2x/(x^2+omega_k^2)
            DO iii = 1, num_x_nodes
               y_values(iii) = 2.0_dp*x_values(iii)/((x_values(iii))**2+omega**2)
            END DO

            ! calculate mat_A
            DO jjj = 1, num_integ_points

               DO iii = 1, num_x_nodes

                  mat_A(iii, jjj) = COS(omega*tau_tj(jjj))*EXP(-x_values(iii)*tau_tj(jjj))

               END DO

            END DO

            ! Query the optimal workspace.
            CALL DGELS('N', num_x_nodes, num_integ_points, 1, mat_A, num_x_nodes, y_values, num_x_nodes, &
                       work_array, -1, info)

            lwork = INT(work_array(1))

            DEALLOCATE (work_array)

            ALLOCATE (work_array(lwork))
            work_array = 0.0_dp

            ! calculate least squares solution for the integration weights
            CALL DGELS('N', num_x_nodes, num_integ_points, 1, mat_A, num_x_nodes, y_values, num_x_nodes, &
                       work_array, lwork, info)

            tau_wj_work(:) = y_values(1:num_integ_points)

            ! Recalculate chi2

            ! first, deallocate x and y
            DEALLOCATE (x_values)
            DEALLOCATE (y_values)

            ALLOCATE (x_values(num_check_chi2))

            ! set the x-values uniformly in the interval [E_min,E_max]
            DO iii = 1, num_check_chi2
               x_values(iii) = E_min+(iii-1.0_dp)/(num_x_nodes-1.0_dp)*(E_max-E_min)
            END DO

            ALLOCATE (y_values(num_check_chi2))

            DO iii = 1, num_check_chi2
               y_values(iii) = 2.0_dp*x_values(iii)/((x_values(iii))**2+omega**2)
            END DO

            CALL calc_chi2_fit_tau_grid_with_cosine(chi2, omega, tau_tj, tau_wj_work, x_values, &
                                                    y_values, num_integ_points, num_x_nodes)

            CALL calc_max_error_fit_tau_grid_with_cosine(max_error, omega, tau_tj, tau_wj_work, x_values, &
                                                         y_values, num_integ_points, num_x_nodes)

            IF (chi2 < chi2_min_jquad) THEN
               chi2_min_jquad = chi2
               max_error_min_jquad = max_error

               x_node_min = num_x_nodes

               tau_wj_2d(jquad, :) = tau_wj_work(:)

            END IF

            DEALLOCATE (x_values)
            DEALLOCATE (y_values)
            DEALLOCATE (mat_A)
            DEALLOCATE (tau_wj_work)
            DEALLOCATE (work_array)

         END DO

         IF (chi2_min_jquad > chi2_min) THEN

            chi2_min = chi2_min_jquad

            max_error_min = max_error_min_jquad

         END IF

      END DO

      ls_error = SQRT(chi2_min/num_check_chi2)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Calculate the matrix mat_N_gw containing the second derivatives
!>        with respect to the fitting parameters. The second derivatives are
!>        calculated numerically by finite differences.
!> \param N_ij matrix element
!> \param Lambda fitting parameters
!> \param Sigma_c ...
!> \param vec_omega_fit_gw ...
!> \param i ...
!> \param j ...
!> \param num_poles ...
!> \param num_fit_points ...
!> \param n_level_gw ...
!> \param h  ...
! **************************************************************************************************
   SUBROUTINE calc_mat_N(N_ij, Lambda, Sigma_c, vec_omega_fit_gw, i, j, &
                         num_poles, num_fit_points, n_level_gw, h)
      REAL(KIND=dp)                                      :: N_ij
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: Lambda
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: Sigma_c
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: vec_omega_fit_gw
      INTEGER                                            :: i, j, num_poles, num_fit_points, &
                                                            n_level_gw
      REAL(KIND=dp)                                      :: h

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_mat_N', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: im_unit, re_unit
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: Lambda_tmp
      INTEGER                                            :: handle, num_var
      REAL(KIND=dp)                                      :: chi2, chi2_sum

      CALL timeset(routineN, handle)

      num_var = 2*num_poles+1
      ALLOCATE (Lambda_tmp(num_var))
      Lambda_tmp = (0.0_dp, 0.0_dp)
      chi2_sum = 0.0_dp
      re_unit = (1.0_dp, 0.0_dp)
      im_unit = (0.0_dp, 1.0_dp)

      !test
      Lambda_tmp(:) = Lambda(:)
      CALL calc_chi2(chi2, Lambda_tmp, Sigma_c, vec_omega_fit_gw, num_poles, &
                     num_fit_points, n_level_gw)

      ! Fitting parameters with offset h
      Lambda_tmp(:) = Lambda(:)
      IF (MODULO(i, 2) == 0) THEN
         Lambda_tmp(i/2) = Lambda_tmp(i/2)+h*re_unit
      ELSE
         Lambda_tmp((i+1)/2) = Lambda_tmp((i+1)/2)+h*im_unit
      END IF
      IF (MODULO(j, 2) == 0) THEN
         Lambda_tmp(j/2) = Lambda_tmp(j/2)+h*re_unit
      ELSE
         Lambda_tmp((j+1)/2) = Lambda_tmp((j+1)/2)+h*im_unit
      END IF
      CALL calc_chi2(chi2, Lambda_tmp, Sigma_c, vec_omega_fit_gw, num_poles, &
                     num_fit_points, n_level_gw)
      chi2_sum = chi2_sum+chi2

      IF (MODULO(i, 2) == 0) THEN
         Lambda_tmp(i/2) = Lambda_tmp(i/2)-2.0_dp*h*re_unit
      ELSE
         Lambda_tmp((i+1)/2) = Lambda_tmp((i+1)/2)-2.0_dp*h*im_unit
      END IF
      CALL calc_chi2(chi2, Lambda_tmp, Sigma_c, vec_omega_fit_gw, num_poles, &
                     num_fit_points, n_level_gw)
      chi2_sum = chi2_sum-chi2

      IF (MODULO(j, 2) == 0) THEN
         Lambda_tmp(j/2) = Lambda_tmp(j/2)-2.0_dp*h*re_unit
      ELSE
         Lambda_tmp((j+1)/2) = Lambda_tmp((j+1)/2)-2.0_dp*h*im_unit
      END IF
      CALL calc_chi2(chi2, Lambda_tmp, Sigma_c, vec_omega_fit_gw, num_poles, &
                     num_fit_points, n_level_gw)
      chi2_sum = chi2_sum+chi2

      IF (MODULO(i, 2) == 0) THEN
         Lambda_tmp(i/2) = Lambda_tmp(i/2)+2.0_dp*h*re_unit
      ELSE
         Lambda_tmp((i+1)/2) = Lambda_tmp((i+1)/2)+2.0_dp*h*im_unit
      END IF
      CALL calc_chi2(chi2, Lambda_tmp, Sigma_c, vec_omega_fit_gw, num_poles, &
                     num_fit_points, n_level_gw)
      chi2_sum = chi2_sum-chi2

      ! Second derivative with symmetric difference quotient
      N_ij = 1.0_dp/2.0_dp*chi2_sum/(4.0_dp*h*h)

      DEALLOCATE (Lambda_tmp)

      CALL timestop(handle)

   END SUBROUTINE calc_mat_N

! **************************************************************************************************
!> \brief Calculate chi2
!> \param chi2 ...
!> \param Lambda fitting parameters
!> \param Sigma_c ...
!> \param vec_omega_fit_gw ...
!> \param num_poles ...
!> \param num_fit_points ...
!> \param n_level_gw ...
! **************************************************************************************************
   SUBROUTINE calc_chi2(chi2, Lambda, Sigma_c, vec_omega_fit_gw, num_poles, &
                        num_fit_points, n_level_gw)
      REAL(KIND=dp)                                      :: chi2
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: Lambda
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: Sigma_c
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: vec_omega_fit_gw
      INTEGER                                            :: num_poles, num_fit_points, n_level_gw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_chi2', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: func_val, im_unit
      INTEGER                                            :: handle, iii, jjj, kkk

      CALL timeset(routineN, handle)

      im_unit = (0.0_dp, 1.0_dp)
      chi2 = 0.0_dp
      DO kkk = 1, num_fit_points
         func_val = Lambda(1)
         DO iii = 1, num_poles
            jjj = iii*2
            ! calculate value of the fit function
            func_val = func_val+Lambda(jjj)/(im_unit*vec_omega_fit_gw(kkk)-Lambda(jjj+1))
         END DO
         chi2 = chi2+(ABS(Sigma_c(n_level_gw, kkk)-func_val))**2
      END DO

      CALL timestop(handle)

   END SUBROUTINE calc_chi2

! **************************************************************************************************
!> \brief Calculate chi2
!> \param chi2 ...
!> \param beta fitting parameters
!> \param Sigma_c ...
!> \param vec_tau_fit_gw ...
!> \param num_poles ...
!> \param num_fit_points ...
! **************************************************************************************************
   SUBROUTINE calc_chi2_im_time(chi2, beta, Sigma_c, vec_tau_fit_gw, num_poles, &
                                num_fit_points)
      REAL(KIND=dp)                                      :: chi2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: beta, Sigma_c, vec_tau_fit_gw
      INTEGER                                            :: num_poles, num_fit_points

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_chi2_im_time', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: func_val
      INTEGER                                            :: handle, iii, jjj, kkk

      CALL timeset(routineN, handle)

      chi2 = 0.0_dp
      DO kkk = 1, num_fit_points
         func_val = 0.0_dp
         DO iii = 1, num_poles
            jjj = iii*3
            ! calculate value of the fit function
            IF (vec_tau_fit_gw(kkk) > 0) THEN
               func_val = func_val+beta(jjj-2)*EXP(-vec_tau_fit_gw(kkk)*beta(jjj))
            ELSE
               func_val = func_val+beta(jjj-1)*EXP(vec_tau_fit_gw(kkk)*beta(jjj))
            END IF
         END DO
         chi2 = chi2+(ABS(Sigma_c(kkk)-func_val))**2

      END DO

      CALL timestop(handle)

   END SUBROUTINE calc_chi2_im_time

! **************************************************************************************************
!> \brief Calculate chi2 when setting up the grid for tau integration with cosine
!> \param chi2 ...
!> \param omega ...
!> \param tau_tj ...
!> \param tau_wj_work ...
!> \param x_values ...
!> \param y_values ...
!> \param num_integ_points ...
!> \param num_x_nodes ...
! **************************************************************************************************
   SUBROUTINE calc_chi2_fit_tau_grid_with_cosine(chi2, omega, tau_tj, tau_wj_work, x_values, &
                                                 y_values, num_integ_points, num_x_nodes)

      REAL(KIND=dp)                                      :: chi2, omega
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tau_tj, tau_wj_work, x_values, y_values
      INTEGER                                            :: num_integ_points, num_x_nodes

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_chi2_fit_tau_grid_with_cosine', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, kkk
      REAL(KIND=dp)                                      :: func_val

      CALL timeset(routineN, handle)

      chi2 = 0.0_dp
      DO kkk = 1, num_x_nodes

         func_val = 0.0_dp

         CALL eval_fit_func_tau_grid_cosine(func_val, x_values(kkk), num_integ_points, tau_tj, tau_wj_work, omega)

         chi2 = chi2+(ABS(y_values(kkk)-func_val))**2

      END DO

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param max_error ...
!> \param omega ...
!> \param tau_tj ...
!> \param tau_wj_work ...
!> \param x_values ...
!> \param y_values ...
!> \param num_integ_points ...
!> \param num_x_nodes ...
! **************************************************************************************************
   SUBROUTINE calc_max_error_fit_tau_grid_with_cosine(max_error, omega, tau_tj, tau_wj_work, x_values, &
                                                      y_values, num_integ_points, num_x_nodes)

      REAL(KIND=dp)                                      :: max_error, omega
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tau_tj, tau_wj_work, x_values, y_values
      INTEGER                                            :: num_integ_points, num_x_nodes

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_max_error_fit_tau_grid_with_cosine', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, kkk
      REAL(KIND=dp)                                      :: func_val

      CALL timeset(routineN, handle)

      max_error = 0.0_dp

      DO kkk = 1, num_x_nodes

         func_val = 0.0_dp

         CALL eval_fit_func_tau_grid_cosine(func_val, x_values(kkk), num_integ_points, tau_tj, tau_wj_work, omega)

         IF (ABS(y_values(kkk)-func_val) > max_error) THEN
            max_error = ABS(y_values(kkk)-func_val)
         END IF

      END DO

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Calculate the maximum deviation between the fit and the computed self-ener
!> \param max_dev ...
!> \param Lambda fitting parameters
!> \param Sigma_c ...
!> \param vec_omega_fit_gw ...
!> \param num_poles ...
!> \param num_fit_points ...
!> \param n_level_gw ...
! **************************************************************************************************
   SUBROUTINE calc_max_dev(max_dev, Lambda, Sigma_c, vec_omega_fit_gw, num_poles, &
                           num_fit_points, n_level_gw)
      REAL(KIND=dp)                                      :: max_dev
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: Lambda
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: Sigma_c
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: vec_omega_fit_gw
      INTEGER                                            :: num_poles, num_fit_points, n_level_gw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_max_dev', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: func_val, im_unit
      INTEGER                                            :: handle, iii, jjj, kkk

      CALL timeset(routineN, handle)

      im_unit = (0.0_dp, 1.0_dp)
      max_dev = 0.0_dp
      DO kkk = 1, num_fit_points
         func_val = Lambda(1)
         DO iii = 1, num_poles
            jjj = iii*2
            ! calculate value of the fit function
            func_val = func_val+Lambda(jjj)/(im_unit*vec_omega_fit_gw(kkk)-Lambda(jjj+1))
         END DO
         IF (ABS(Sigma_c(n_level_gw, kkk)-func_val) > max_dev) THEN
            max_dev = ABS(Sigma_c(n_level_gw, kkk)-func_val)
         END IF
      END DO

      CALL timestop(handle)

   END SUBROUTINE calc_max_dev

! **************************************************************************************************
!> \brief Evaluate fit function
!> \param func_val result of fit evaluation
!> \param Lambda fitting parameters
!> \param omega real or complex energy
!> \param e_fermi the Fermi level
!> \param num_poles ...
! **************************************************************************************************
   SUBROUTINE evaluate_fit_function(func_val, Lambda, omega, e_fermi, num_poles)
      COMPLEX(KIND=dp)                                   :: func_val
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: Lambda
      COMPLEX(KIND=dp)                                   :: omega
      REAL(KIND=dp)                                      :: e_fermi
      INTEGER                                            :: num_poles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'evaluate_fit_function', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iii, jjj

      func_val = Lambda(1)
      DO iii = 1, num_poles
         jjj = iii*2
         ! calculate value of the fit function
         func_val = func_val+Lambda(jjj)/(omega-e_fermi-Lambda(jjj+1))
      END DO

   END SUBROUTINE evaluate_fit_function

! **************************************************************************************************
!> \brief Fits the complex self-energy of n_level_gw to a multi-pole model and evaluates the
!>        fit at the energy eigenvalue of the SCF. Real part of the correlation self-energy
!>        is written to vec_gw_energ. Also calculates the statistical error of the correlation
!>        self-energy due to the fit
!> \param vec_gw_energ ...
!> \param vec_gw_energ_error_fit ...
!> \param vec_omega_fit_gw ...
!> \param z_value ...
!> \param m_value ...
!> \param vec_Sigma_c_gw ...
!> \param vec_Sigma_x_minus_vxc_gw ...
!> \param Eigenval ...
!> \param Eigenval_scf ...
!> \param n_level_gw ...
!> \param gw_corr_lev_occ ...
!> \param num_poles ...
!> \param num_fit_points ...
!> \param max_iter_fit ...
!> \param crossing_search ...
!> \param homo ...
!> \param check_fit ...
!> \param stop_crit ...
!> \param fermi_level_offset ...
! **************************************************************************************************
   SUBROUTINE fit_and_continuation(vec_gw_energ, vec_gw_energ_error_fit, vec_omega_fit_gw, &
                                   z_value, m_value, vec_Sigma_c_gw, vec_Sigma_x_minus_vxc_gw, &
                                   Eigenval, Eigenval_scf, n_level_gw, gw_corr_lev_occ, num_poles, &
                                   num_fit_points, max_iter_fit, crossing_search, homo, check_fit, stop_crit, &
                                   fermi_level_offset)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: vec_gw_energ, vec_gw_energ_error_fit, &
                                                            vec_omega_fit_gw, z_value, m_value
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: vec_Sigma_c_gw
      REAL(KIND=dp), DIMENSION(:)                        :: vec_Sigma_x_minus_vxc_gw, Eigenval, &
                                                            Eigenval_scf
      INTEGER                                            :: n_level_gw, gw_corr_lev_occ, num_poles, &
                                                            num_fit_points, max_iter_fit, &
                                                            crossing_search, homo
      LOGICAL                                            :: check_fit
      REAL(KIND=dp)                                      :: stop_crit, fermi_level_offset

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fit_and_continuation', &
         routineP = moduleN//':'//routineN

      CHARACTER(5)                                       :: MO_number
      COMPLEX(KIND=dp)                                   :: func_val, im_unit, one, re_unit, rho1, &
                                                            zero
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: dLambda, dLambda_2, Lambda, &
                                                            Lambda_without_offset, vec_b_gw, &
                                                            vec_b_gw_copy
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: mat_A_gw, mat_B_gw
      INTEGER                                            :: handle4, ierr, iii, iiter, info, &
                                                            integ_range, jjj, jquad, kkk, &
                                                            n_level_gw_ref, num_var, output_unit, &
                                                            xpos
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipiv
      LOGICAL                                            :: could_exit
      REAL(KIND=dp) :: chi2, chi2_old, delta, deriv_val_real, e_fermi, gw_energ, Ldown, &
         level_energ_GW, Lup, range_step, ScalParam, sign_occ_virt, stat_error
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Lambda_Im, Lambda_Re, stat_errors, &
                                                            vec_N_gw, vec_omega_fit_gw_sign
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: mat_N_gw

      output_unit = cp_logger_get_default_io_unit()

      im_unit = (0.0_dp, 1.0_dp)
      re_unit = (1.0_dp, 0.0_dp)

      num_var = 2*num_poles+1
      ALLOCATE (Lambda(num_var))
      Lambda = (0.0_dp, 0.0_dp)
      ALLOCATE (Lambda_without_offset(num_var))
      Lambda_without_offset = (0.0_dp, 0.0_dp)
      ALLOCATE (Lambda_Re(num_var))
      Lambda_Re = 0.0_dp
      ALLOCATE (Lambda_Im(num_var))
      Lambda_Im = 0.0_dp

      ALLOCATE (vec_omega_fit_gw_sign(num_fit_points))

      IF (n_level_gw <= gw_corr_lev_occ) THEN
         sign_occ_virt = -1.0_dp
      ELSE
         sign_occ_virt = 1.0_dp
      END IF

      DO jquad = 1, num_fit_points
         vec_omega_fit_gw_sign(jquad) = ABS(vec_omega_fit_gw(jquad))*sign_occ_virt
      END DO

      ! initial guess
      range_step = (vec_omega_fit_gw_sign(num_fit_points)-vec_omega_fit_gw_sign(1))/(num_poles-1)
      DO iii = 1, num_poles
         Lambda_Im(2*iii+1) = vec_omega_fit_gw_sign(1)+(iii-1)*range_step
      END DO
      range_step = (vec_omega_fit_gw_sign(num_fit_points)-vec_omega_fit_gw_sign(1))/num_poles
      DO iii = 1, num_poles
         Lambda_Re(2*iii+1) = ABS(vec_omega_fit_gw_sign(1)+(iii-0.5_dp)*range_step)
      END DO

      DO iii = 1, num_var
         Lambda(iii) = Lambda_Re(iii)+im_unit*Lambda_Im(iii)
      END DO

      CALL calc_chi2(chi2_old, Lambda, vec_Sigma_c_gw, vec_omega_fit_gw_sign, num_poles, &
                     num_fit_points, n_level_gw)

      ALLOCATE (mat_A_gw(num_poles+1, num_poles+1))
      ALLOCATE (vec_b_gw(num_poles+1))
      ALLOCATE (ipiv(num_poles+1))
      mat_A_gw = (0.0_dp, 0.0_dp)
      vec_b_gw = 0.0_dp

      DO iii = 1, num_poles+1
         mat_A_gw(iii, 1) = (1.0_dp, 0.0_dp)
      END DO
      integ_range = num_fit_points/num_poles
      DO kkk = 1, num_poles+1
         xpos = (kkk-1)*integ_range+1
         xpos = MIN(xpos, num_fit_points)
         ! calculate coefficient at this point
         DO iii = 1, num_poles
            jjj = iii*2
            func_val = (1.0_dp, 0.0_dp)/(im_unit*vec_omega_fit_gw_sign(xpos)- &
                                         CMPLX(Lambda_Re(jjj+1), Lambda_Im(jjj+1), KIND=dp))
            mat_A_gw(kkk, iii+1) = func_val
         END DO
         vec_b_gw(kkk) = vec_Sigma_c_gw(n_level_gw, xpos)
      END DO

      ! Solve system of linear equations
      CALL ZGETRF(num_poles+1, num_poles+1, mat_A_gw, num_poles+1, ipiv, info)

      CALL ZGETRS('N', num_poles+1, 1, mat_A_gw, num_poles+1, ipiv, vec_b_gw, num_poles+1, info)

      Lambda_Re(1) = REAL(vec_b_gw(1))
      Lambda_Im(1) = AIMAG(vec_b_gw(1))
      DO iii = 1, num_poles
         jjj = iii*2
         Lambda_Re(jjj) = REAL(vec_b_gw(iii+1))
         Lambda_Im(jjj) = AIMAG(vec_b_gw(iii+1))
      END DO

      DEALLOCATE (mat_A_gw)
      DEALLOCATE (vec_b_gw)
      DEALLOCATE (ipiv)

      ALLOCATE (mat_A_gw(num_var*2, num_var*2))
      ALLOCATE (mat_B_gw(num_fit_points, num_var*2))
      ALLOCATE (dLambda(num_fit_points))
      ALLOCATE (dLambda_2(num_fit_points))
      ALLOCATE (vec_b_gw(num_var*2))
      ALLOCATE (vec_b_gw_copy(num_var*2))
      ALLOCATE (ipiv(num_var*2))

      ScalParam = 0.01_dp
      Ldown = 1.5_dp
      Lup = 10.0_dp
      could_exit = .FALSE.

      ! iteration loop for fitting
      DO iiter = 1, max_iter_fit

         CALL timeset(routineN//"_fit_loop_1", handle4)

         ! calc delta lambda
         DO iii = 1, num_var
            Lambda(iii) = Lambda_Re(iii)+im_unit*Lambda_Im(iii)
         END DO
         dLambda = (0.0_dp, 0.0_dp)

         DO kkk = 1, num_fit_points
            func_val = Lambda(1)
            DO iii = 1, num_poles
               jjj = iii*2
               func_val = func_val+Lambda(jjj)/(vec_omega_fit_gw_sign(kkk)*im_unit-Lambda(jjj+1))
            END DO
            dLambda(kkk) = vec_Sigma_c_gw(n_level_gw, kkk)-func_val
         END DO
         rho1 = SUM(dLambda*dLambda)

         ! fill matrix
         mat_B_gw = (0.0_dp, 0.0_dp)
         DO iii = 1, num_fit_points
            mat_B_gw(iii, 1) = 1.0_dp
            mat_B_gw(iii, num_var+1) = im_unit
         END DO
         DO iii = 1, num_poles
            jjj = iii*2
            DO kkk = 1, num_fit_points
               mat_B_gw(kkk, jjj) = 1.0_dp/(im_unit*vec_omega_fit_gw_sign(kkk)-Lambda(jjj+1))
               mat_B_gw(kkk, jjj+num_var) = im_unit/(im_unit*vec_omega_fit_gw_sign(kkk)-Lambda(jjj+1))
               mat_B_gw(kkk, jjj+1) = Lambda(jjj)/(im_unit*vec_omega_fit_gw_sign(kkk)-Lambda(jjj+1))**2
               mat_B_gw(kkk, jjj+1+num_var) = (-Lambda_Im(jjj)+im_unit*Lambda_Re(jjj))/ &
                                              (im_unit*vec_omega_fit_gw_sign(kkk)-Lambda(jjj+1))**2
            END DO
         END DO

         CALL timestop(handle4)

         CALL timeset(routineN//"_fit_matmul_1", handle4)

         one = (1.0_dp, 0.0_dp)
         zero = (0.0_dp, 0.0_dp)
         CALL zgemm('C', 'N', num_var*2, num_var*2, num_fit_points, one, mat_B_gw, num_fit_points, mat_B_gw, num_fit_points, &
                    zero, mat_A_gw, num_var*2)
         CALL timestop(handle4)

         CALL timeset(routineN//"_fit_zgemv_1", handle4)
         CALL zgemv('C', num_fit_points, num_var*2, one, mat_B_gw, num_fit_points, dLambda, 1, &
                    zero, vec_b_gw, 1)

         CALL timestop(handle4)

         ! scale diagonal elements of a_mat
         DO iii = 1, num_var*2
            mat_A_gw(iii, iii) = mat_A_gw(iii, iii)+ScalParam*mat_A_gw(iii, iii)
         END DO

         ! solve linear system
         ierr = 0
         ipiv = 0

         CALL timeset(routineN//"_fit_lin_eq_2", handle4)

         CALL ZGETRF(2*num_var, 2*num_var, mat_A_gw, 2*num_var, ipiv, info)

         CALL ZGETRS('N', 2*num_var, 1, mat_A_gw, 2*num_var, ipiv, vec_b_gw, 2*num_var, info)

         CALL timestop(handle4)

         DO iii = 1, num_var
            Lambda(iii) = Lambda_Re(iii)+im_unit*Lambda_Im(iii)+vec_b_gw(iii)+vec_b_gw(iii+num_var)
         END DO

         ! calculate chi2
         CALL calc_chi2(chi2, Lambda, vec_Sigma_c_gw, vec_omega_fit_gw_sign, num_poles, &
                        num_fit_points, n_level_gw)

         IF (chi2 < chi2_old) THEN
            ScalParam = MAX(ScalParam/Ldown, 1E-12_dp)
            DO iii = 1, num_var
               Lambda_Re(iii) = Lambda_Re(iii)+REAL(vec_b_gw(iii)+vec_b_gw(iii+num_var))
               Lambda_Im(iii) = Lambda_Im(iii)+AIMAG(vec_b_gw(iii)+vec_b_gw(iii+num_var))
            END DO
            IF (chi2_old/chi2-1.0_dp < stop_crit) could_exit = .TRUE.
            chi2_old = chi2
         ELSE
            ScalParam = ScalParam*Lup
         END IF
         IF (ScalParam > 100.0_dp .AND. could_exit) EXIT

         IF (ScalParam > 1E+10_dp) ScalParam = 1E-4_dp

         n_level_gw_ref = n_level_gw+homo-gw_corr_lev_occ
         IF (iiter == max_iter_fit) THEN
            WRITE (MO_number, "(I3)") n_level_gw_ref
            CALL cp_warn(cp__l("rpa_ri_gpw.F",4607), &
                         "The fit for corrected level n ="//MO_number//" did not converge. "// &
                         "For levels close to HOMO or LUMO, this is normally no issue. "// &
                         "To avoid this warning, you can (1) increase the "// &
                         "number of fit iterations MAX_ITER_FIT, or you can (2) increase the number "// &
                         "of RPA integration points (then, Sigma_c(i*omega) is more accurate) "// &
                         "or  you can (3) decrease "// &
                         "the fit range by setting the keyword OMEGA_MAX_FIT (in Hartree). ")

         END IF

      END DO

      ! change a_0 [Lambda(1)], so that Sigma(i0) = Fit(i0)
      func_val = Lambda(1)
      DO iii = 1, num_poles
         jjj = iii*2
         ! calculate value of the fit function
         func_val = func_val+Lambda(jjj)/(-Lambda(jjj+1))
      END DO

      Lambda_Re(1) = Lambda_Re(1)-REAL(func_val)+REAL(vec_Sigma_c_gw(n_level_gw, num_fit_points))
      Lambda_Im(1) = Lambda_Im(1)-AIMAG(func_val)+AIMAG(vec_Sigma_c_gw(n_level_gw, num_fit_points))

      Lambda_without_offset(:) = Lambda(:)

      DO iii = 1, num_var
         Lambda(iii) = CMPLX(Lambda_Re(iii), Lambda_Im(iii), KIND=dp)
      END DO

      ! print self_energy and fit on the imaginary frequency axis if required
      IF (check_fit) THEN

         IF (output_unit > 0) THEN

            WRITE (output_unit, *) ' '
            WRITE (output_unit, '(T3,A,I5)') 'Check the GW fit for molecular orbital', n_level_gw_ref
            WRITE (output_unit, '(T3,A)') '-------------------------------------------'
            WRITE (output_unit, *)
            WRITE (output_unit, '(T3,5A)') '  omega (i*eV)    ', 'Re(fit) (eV)    ', &
               'Im(fit) (eV)  ', 'Re(Sig_c) (eV)  ', &
               'Im(Sig_c) (eV)'

         END IF

         DO kkk = 1, num_fit_points
            func_val = Lambda(1)
            DO iii = 1, num_poles
               jjj = iii*2
               ! calculate value of the fit function
               func_val = func_val+Lambda(jjj)/(im_unit*vec_omega_fit_gw_sign(kkk)-Lambda(jjj+1))
            END DO
            WRITE (output_unit, '(1F16.3,4F16.5)') vec_omega_fit_gw_sign(kkk)*27.211_dp, REAL(func_val)*27.211_dp, &
               AIMAG(func_val)*27.211_dp, REAL(vec_Sigma_c_gw(n_level_gw, kkk))*27.211_dp, &
               AIMAG(vec_Sigma_c_gw(n_level_gw, kkk))*27.211_dp
         END DO

         WRITE (output_unit, *) ' '

      END IF

      ! correct the electronic levels
      IF (n_level_gw <= gw_corr_lev_occ) THEN
         e_fermi = Eigenval(homo)+fermi_level_offset
      ELSE
         e_fermi = Eigenval(homo+1)-fermi_level_offset
      END IF

      ! either Z-shot or no crossing search for evaluating Sigma_c
      IF (crossing_search == ri_rpa_g0w0_crossing_none) THEN

         ! calculate func val on the real axis
         ! gw_energ = only correlation part of the self energy
         func_val = Lambda(1)
         DO iii = 1, num_poles
            jjj = iii*2
            func_val = func_val+Lambda(jjj)/(Eigenval(n_level_gw_ref)-e_fermi-Lambda(jjj+1))
         END DO

         gw_energ = REAL(func_val)
         vec_gw_energ(n_level_gw) = gw_energ

      ELSE IF (crossing_search == ri_rpa_g0w0_crossing_z_shot .OR. &
               crossing_search == ri_rpa_g0w0_crossing_newton) THEN

         ! calculate Sigma_c_fit(e_n) and Z
         func_val = Lambda(1)
         z_value(n_level_gw) = 1.0_dp
         DO iii = 1, num_poles
            jjj = iii*2
            z_value(n_level_gw) = z_value(n_level_gw)+REAL(Lambda(jjj)/ &
                                                           (Eigenval(n_level_gw_ref)-e_fermi-Lambda(jjj+1))**2)
            func_val = func_val+Lambda(jjj)/(Eigenval(n_level_gw_ref)-e_fermi-Lambda(jjj+1))
         END DO
         ! m is the slope of the correl self-energy
         m_value(n_level_gw) = 1.0_dp-z_value(n_level_gw)
         z_value(n_level_gw) = 1.0_dp/z_value(n_level_gw)
         gw_energ = REAL(func_val)
         vec_gw_energ(n_level_gw) = gw_energ

         ! in case one wants to do Newton-Raphson on top of the Z-shot
         IF (crossing_search == ri_rpa_g0w0_crossing_newton) THEN

            level_energ_GW = (Eigenval_scf(n_level_gw_ref)- &
                              m_value(n_level_gw)*Eigenval(n_level_gw_ref)+ &
                              vec_gw_energ(n_level_gw)+ &
                              vec_Sigma_x_minus_vxc_gw(n_level_gw_ref))* &
                             z_value(n_level_gw)

            ! Newton-Raphson iteration
            DO kkk = 1, 1000

               ! calculate the value of the fit function for level_energ_GW
               func_val = Lambda(1)
               z_value(n_level_gw) = 1.0_dp
               DO iii = 1, num_poles
                  jjj = iii*2
                  func_val = func_val+Lambda(jjj)/(level_energ_GW-e_fermi-Lambda(jjj+1))
               END DO

               ! calculate the derivative of the fit function for level_energ_GW
               deriv_val_real = -1.0_dp
               DO iii = 1, num_poles
                  jjj = iii*2
                  deriv_val_real = deriv_val_real+REAL(Lambda(jjj))/((ABS(level_energ_GW-e_fermi-Lambda(jjj+1)))**2) &
                                   -(REAL(Lambda(jjj))*(level_energ_GW-e_fermi)-REAL(Lambda(jjj)*CONJG(Lambda(jjj+1))))* &
                                   2.0_dp*(level_energ_GW-e_fermi-REAL(Lambda(jjj+1)))/ &
                                   ((ABS(level_energ_GW-e_fermi-Lambda(jjj+1)))**2)

               END DO

               delta = (Eigenval_scf(n_level_gw_ref)+vec_Sigma_x_minus_vxc_gw(n_level_gw_ref)+REAL(func_val)-level_energ_GW)/ &
                       deriv_val_real

               level_energ_GW = level_energ_GW-delta

               IF (ABS(delta) < 1.0E-08) EXIT

            END DO

            ! update the GW-energy by Newton-Raphson and set the Z-value to 1

            vec_gw_energ(n_level_gw) = REAL(func_val)
            z_value(n_level_gw) = 1.0_dp
            m_value(n_level_gw) = 0.0_dp

         END IF ! Newton-Raphson on top of Z-shot

      END IF ! decision crossing search none, Z-shot

      !   --------------------------------------------
      !  | calculate statistical error due to fitting |
      !   --------------------------------------------

      ! estimate the statistical error of the calculated Sigma_c(i*omega)
      ! by sqrt(chi2/n), where n is the number of fit points

      CALL calc_chi2(chi2, Lambda_without_offset, vec_Sigma_c_gw, vec_omega_fit_gw_sign, num_poles, &
                     num_fit_points, n_level_gw)

      ! Estimate the statistical error of every fit point
      stat_error = SQRT(chi2/num_fit_points)

      ! allocate N array containing the second derivatives of chi^2
      ALLOCATE (vec_N_gw(num_var*2))
      vec_N_gw = 0.0_dp

      ALLOCATE (mat_N_gw(num_var*2, num_var*2))
      mat_N_gw = 0.0_dp

      DO iii = 1, num_var*2
         CALL calc_mat_N(vec_N_gw(iii), Lambda_without_offset, vec_Sigma_c_gw, vec_omega_fit_gw_sign, &
                         iii, iii, num_poles, num_fit_points, n_level_gw, 0.001_dp)
      END DO

      DO iii = 1, num_var*2
         DO jjj = 1, num_var*2
            CALL calc_mat_N(mat_N_gw(iii, jjj), Lambda_without_offset, vec_Sigma_c_gw, vec_omega_fit_gw_sign, &
                            iii, jjj, num_poles, num_fit_points, n_level_gw, 0.001_dp)
         END DO
      END DO

      CALL DGETRF(2*num_var, 2*num_var, mat_N_gw, 2*num_var, ipiv, info)

      ! vec_b_gw is only working array
      CALL DGETRI(2*num_var, mat_N_gw, 2*num_var, ipiv, vec_b_gw, 2*num_var, info)

      ALLOCATE (stat_errors(2*num_var))
      stat_errors = 0.0_dp

      DO iii = 1, 2*num_var
         stat_errors(iii) = SQRT(ABS(mat_N_gw(iii, iii)))*stat_error
      END DO

      ! Compute error of Sigma_c on real axis according to error propagation

      vec_gw_energ_error_fit(n_level_gw) = 0.0_dp

      DO kkk = 1, num_poles
         vec_gw_energ_error_fit(n_level_gw) = vec_gw_energ_error_fit(n_level_gw)+ &
                                              (stat_errors(4*kkk-1)+stat_errors(4*kkk))* &
                                              ABS(1.0_dp/(Eigenval(n_level_gw_ref)-e_fermi-Lambda(2*kkk+1))- &
                                                  1.0_dp/(-Lambda(2*kkk+1)))+ &
                                              (stat_errors(4*kkk+1)+stat_errors(4*kkk+2))*ABS(Lambda(2*kkk))* &
                                              ABS(1.0_dp/(Eigenval(n_level_gw_ref)-e_fermi-Lambda(2*kkk+1))**2- &
                                                  1.0_dp/(-Lambda(2*kkk+1))**2)
      END DO

      DEALLOCATE (mat_N_gw)
      DEALLOCATE (vec_N_gw)
      DEALLOCATE (mat_A_gw)
      DEALLOCATE (mat_B_gw)
      DEALLOCATE (stat_errors)
      DEALLOCATE (dLambda)
      DEALLOCATE (dLambda_2)
      DEALLOCATE (vec_b_gw)
      DEALLOCATE (vec_b_gw_copy)
      DEALLOCATE (ipiv)
      DEALLOCATE (vec_omega_fit_gw_sign)
      DEALLOCATE (Lambda)
      DEALLOCATE (Lambda_without_offset)
      DEALLOCATE (Lambda_Re)
      DEALLOCATE (Lambda_Im)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param vec_gw_energ ...
!> \param z_value ...
!> \param m_value ...
!> \param vec_Sigma_c_gw_real_freq ...
!> \param contour_def_offset ...
!> \param n_level_gw ...
!> \param index_contour_def ...
! **************************************************************************************************
   SUBROUTINE compute_z_and_m_contour_def(vec_gw_energ, z_value, m_value, &
                                          vec_Sigma_c_gw_real_freq, contour_def_offset, &
                                          n_level_gw, index_contour_def)

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: vec_gw_energ, z_value, m_value, &
                                                            vec_Sigma_c_gw_real_freq
      REAL(KIND=dp)                                      :: contour_def_offset
      INTEGER                                            :: n_level_gw, index_contour_def

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_z_and_m_contour_def', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: derivative

! average of Sigma_c(e_n+delta) and Sigma_c(e_n-delta)

      vec_gw_energ(n_level_gw) = 0.5_dp*(vec_Sigma_c_gw_real_freq(2*index_contour_def-1) &
                                         +vec_Sigma_c_gw_real_freq(2*index_contour_def))

      derivative = 0.5_dp*(vec_Sigma_c_gw_real_freq(2*index_contour_def) &
                           -vec_Sigma_c_gw_real_freq(2*index_contour_def-1)) &
                   /contour_def_offset

      z_value(n_level_gw) = 1.0_dp/(1.0_dp-derivative)

      m_value(n_level_gw) = derivative

   END SUBROUTINE

! **************************************************************************************************
!> \brief Evaluate the fit function at imaginary time
!> \param func_val ...
!> \param tau ...
!> \param num_poles ...
!> \param beta ...
! **************************************************************************************************
   SUBROUTINE eval_fit_func_im_time(func_val, tau, num_poles, beta)

      REAL(KIND=dp)                                      :: func_val, tau
      INTEGER                                            :: num_poles
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_fit_func_im_time', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iii, jjj

      func_val = 0.0_dp
      DO iii = 1, num_poles
         jjj = iii*3
         ! calculate value of the fit function
         IF (tau > 0) THEN
            func_val = func_val+beta(jjj-2)*EXP(-tau*beta(jjj))
         ELSE
            func_val = func_val+beta(jjj-1)*EXP(tau*beta(jjj))
         END IF
      END DO

   END SUBROUTINE

! **************************************************************************************************
!> \brief Evaluate fit function when calculating tau grid
!> \param func_val ...
!> \param x_value ...
!> \param num_integ_points ...
!> \param tj ...
!> \param wj ...
! **************************************************************************************************
   SUBROUTINE eval_fit_func_tau_grid(func_val, x_value, num_integ_points, &
                                     tj, wj)

      REAL(KIND=dp)                                      :: func_val, x_value
      INTEGER                                            :: num_integ_points
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tj, wj

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_fit_func_tau_grid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iii

      func_val = 0.0_dp

      DO iii = 1, num_integ_points

         ! calculate value of the fit function
         func_val = func_val+wj(iii)*EXP(-2.0_dp*x_value*tj(iii)**2)

      END DO

   END SUBROUTINE

! **************************************************************************************************
!> \brief Evaluate fit function when calculating tau grid for Fourier transform
!> \param func_val ...
!> \param x_value ...
!> \param num_integ_points ...
!> \param tau_tj ...
!> \param tau_wj_work ...
!> \param omega ...
! **************************************************************************************************
   SUBROUTINE eval_fit_func_tau_grid_cosine(func_val, x_value, num_integ_points, tau_tj, tau_wj_work, omega)

      REAL(KIND=dp)                                      :: func_val, x_value
      INTEGER                                            :: num_integ_points
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tau_tj, tau_wj_work
      REAL(KIND=dp)                                      :: omega

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_fit_func_tau_grid_cosine', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iii

      func_val = 0.0_dp

      DO iii = 1, num_integ_points

         ! calculate value of the fit function
         func_val = func_val+tau_wj_work(iii)*COS(omega*tau_tj(iii))*EXP(-x_value*tau_tj(iii))

      END DO

   END SUBROUTINE

! **************************************************************************************************
!> \brief Prints the GW stuff to the output and optinally to an external file.
!>        Also updates the eigenvalues for eigenvalue-self-consistent GW
!> \param vec_gw_energ ...
!> \param vec_gw_energ_error_fit ...
!> \param z_value ...
!> \param m_value ...
!> \param vec_Sigma_x_minus_vxc_gw ...
!> \param Eigenval ...
!> \param Eigenval_last ...
!> \param Eigenval_scf ...
!> \param gw_corr_lev_occ ...
!> \param gw_corr_lev_virt ...
!> \param gw_corr_lev_tot ...
!> \param count_ev_sc_GW ...
!> \param crossing_search ...
!> \param homo ...
!> \param nmo ...
!> \param unit_nr ...
!> \param print_gw_details ...
!> \param do_alpha ...
!> \param do_beta ...
! **************************************************************************************************
   SUBROUTINE print_and_update_for_ev_sc(vec_gw_energ, vec_gw_energ_error_fit, &
                                         z_value, m_value, vec_Sigma_x_minus_vxc_gw, Eigenval, &
                                         Eigenval_last, Eigenval_scf, gw_corr_lev_occ, gw_corr_lev_virt, gw_corr_lev_tot, &
                                         count_ev_sc_GW, crossing_search, homo, nmo, unit_nr, print_gw_details, &
                                         do_alpha, do_beta)

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: vec_gw_energ, vec_gw_energ_error_fit, &
                                                            z_value, m_value
      REAL(KIND=dp), DIMENSION(:)                        :: vec_Sigma_x_minus_vxc_gw, Eigenval
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Eigenval_last, Eigenval_scf
      INTEGER                                            :: gw_corr_lev_occ, gw_corr_lev_virt, &
                                                            gw_corr_lev_tot, count_ev_sc_GW, &
                                                            crossing_search, homo, nmo, unit_nr
      LOGICAL                                            :: print_gw_details
      LOGICAL, OPTIONAL                                  :: do_alpha, do_beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_and_update_for_ev_sc', &
         routineP = moduleN//':'//routineN

      CHARACTER(4)                                       :: occ_virt
      INTEGER                                            :: handle, n_level_gw, n_level_gw_ref
      LOGICAL                                            :: do_closed_shell, my_do_alpha, my_do_beta
      REAL(KIND=dp)                                      :: eigen_diff

      CALL timeset(routineN, handle)

      IF (PRESENT(do_alpha)) THEN
         my_do_alpha = do_alpha
      ELSE
         my_do_alpha = .FALSE.
      END IF

      IF (PRESENT(do_alpha)) THEN
         my_do_alpha = do_alpha
      ELSE
         my_do_alpha = .FALSE.
      END IF

      IF (PRESENT(do_beta)) THEN
         my_do_beta = do_beta
      ELSE
         my_do_beta = .FALSE.
      END IF

      do_closed_shell = .NOT. (my_do_alpha .OR. my_do_beta)

      Eigenval_last(:) = Eigenval(:)

      IF (unit_nr > 0) THEN

         WRITE (unit_nr, *) ' '

         IF (do_closed_shell) THEN
            WRITE (unit_nr, '(T3,A)') 'GW quasiparticle energies'
            WRITE (unit_nr, '(T3,A)') '-------------------------'
         ELSE IF (my_do_alpha) THEN
            WRITE (unit_nr, '(T3,A)') 'GW quasiparticle energies of alpha spins'
            WRITE (unit_nr, '(T3,A)') '----------------------------------------'
         ELSE IF (my_do_beta) THEN
            WRITE (unit_nr, '(T3,A)') 'GW quasiparticle energies of beta spins'
            WRITE (unit_nr, '(T3,A)') '---------------------------------------'
         END IF

      END IF

      IF (unit_nr > 0 .AND. (.NOT. print_gw_details)) THEN
         WRITE (unit_nr, *) ' '
         WRITE (unit_nr, '(T5,A)') 'Molecular orbital        MO energy after SCF (eV)        G0W0 QP energy (eV)'
      END IF

      IF (unit_nr > 0 .AND. print_gw_details) THEN
         WRITE (unit_nr, '(T3,A)') ' '
         WRITE (unit_nr, '(T3,A)') 'The GW quasiparticle energies are calculated according to: '
      END IF

      IF (crossing_search == ri_rpa_g0w0_crossing_none) THEN

         DO n_level_gw = 1, gw_corr_lev_tot
            n_level_gw_ref = n_level_gw+homo-gw_corr_lev_occ
            Eigenval(n_level_gw_ref) = Eigenval_scf(n_level_gw_ref)+vec_gw_energ(n_level_gw)+ &
                                       vec_Sigma_x_minus_vxc_gw(n_level_gw_ref)
         END DO

         IF (unit_nr > 0 .AND. print_gw_details) THEN
            WRITE (unit_nr, '(T3,A)') ' '
            WRITE (unit_nr, '(T3,A)') 'E_GW = E_SCF +  Sigc(E_SCF) + Sigx - vxc'
            WRITE (unit_nr, '(T3,A)') ' '
            WRITE (unit_nr, '(T3,A)') 'The energy unit of the following table is eV. Sigc_fit is a very conservative'
            WRITE (unit_nr, '(T3,A)') 'estimate of the statistical error of the correlation self-energy caused by the'
            WRITE (unit_nr, '(T3,A)') 'fitting.'
            WRITE (unit_nr, '(T3,A)') ' '
            WRITE (unit_nr, '(T14,2A)') 'MO        E_SCF         Sigc', &
               '     Sigc_fit     Sigx-vxc         E_GW'
         END IF

         DO n_level_gw = 1, gw_corr_lev_tot
            n_level_gw_ref = n_level_gw+homo-gw_corr_lev_occ
            IF (n_level_gw <= gw_corr_lev_occ) THEN
               occ_virt = 'occ'
            ELSE
               occ_virt = 'vir'
            END IF

            IF (unit_nr > 0 .AND. (.NOT. print_gw_details)) THEN
               WRITE (unit_nr, '(T5,I9,3A,2F27.3)') &
                  n_level_gw_ref, ' ( ', occ_virt, ')     ', &
                  Eigenval_last(n_level_gw_ref)*27.211_dp, &
                  Eigenval(n_level_gw_ref)*27.211_dp
            END IF

            IF (unit_nr > 0 .AND. print_gw_details) THEN
               WRITE (unit_nr, '(T4,I4,3A,5F13.3)') &
                  n_level_gw_ref, ' ( ', occ_virt, ')', &
                  Eigenval_last(n_level_gw_ref)*27.211_dp, &
                  vec_gw_energ(n_level_gw)*27.211_dp, &
                  vec_gw_energ_error_fit(n_level_gw)*27.211_dp, &
                  vec_Sigma_x_minus_vxc_gw(n_level_gw_ref)*27.211_dp, &
                  Eigenval(n_level_gw_ref)*27.211_dp
            END IF
         END DO

         ! z-shot
      ELSE

         DO n_level_gw = 1, gw_corr_lev_tot
            n_level_gw_ref = n_level_gw+homo-gw_corr_lev_occ
            Eigenval(n_level_gw_ref) = (Eigenval_scf(n_level_gw_ref)- &
                                        m_value(n_level_gw)*Eigenval(n_level_gw_ref)+ &
                                        vec_gw_energ(n_level_gw)+ &
                                        vec_Sigma_x_minus_vxc_gw(n_level_gw_ref))* &
                                       z_value(n_level_gw)
         END DO

         IF (unit_nr > 0 .AND. print_gw_details) THEN
            WRITE (unit_nr, '(T3,A)') 'E_GW = E_SCF + Z * ( Sigc(E_SCF) + Sigx - vxc )'
            WRITE (unit_nr, '(T3,A)') ' '
            WRITE (unit_nr, '(T3,A)') 'The energy unit of the following table is eV.  Sigc_fit is a very conservative'
            WRITE (unit_nr, '(T3,2A)') 'estimate of the statistical error of the fitting.'
            WRITE (unit_nr, '(T3,A)') ' '
            WRITE (unit_nr, '(T13,2A)') 'MO      E_SCF       Sigc', &
               '   Sigc_fit   Sigx-vxc          Z       E_GW'
         END IF

         DO n_level_gw = 1, gw_corr_lev_tot
            n_level_gw_ref = n_level_gw+homo-gw_corr_lev_occ
            IF (n_level_gw <= gw_corr_lev_occ) THEN
               occ_virt = 'occ'
            ELSE
               occ_virt = 'vir'
            END IF

            IF (unit_nr > 0 .AND. (.NOT. print_gw_details)) THEN
               WRITE (unit_nr, '(T5,I9,3A,2F27.3)') &
                  n_level_gw_ref, ' ( ', occ_virt, ')     ', &
                  Eigenval_last(n_level_gw_ref)*27.211_dp, &
                  Eigenval(n_level_gw_ref)*27.211_dp
            END IF

            IF (unit_nr > 0 .AND. print_gw_details) THEN
               WRITE (unit_nr, '(T3,I4,3A,6F11.3)') &
                  n_level_gw_ref, ' ( ', occ_virt, ')', &
                  Eigenval_last(n_level_gw_ref)*27.211_dp, &
                  vec_gw_energ(n_level_gw)*27.211_dp, &
                  vec_gw_energ_error_fit(n_level_gw)*27.211_dp, &
                  vec_Sigma_x_minus_vxc_gw(n_level_gw_ref)*27.211_dp, &
                  z_value(n_level_gw), &
                  Eigenval(n_level_gw_ref)*27.211_dp
            END IF
         END DO

      END IF ! z-shot vs. no crossing

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, *) ' '
      END IF

      ! for eigenvalue self-consistent GW, all eigenvalues have to be corrected
      ! 1) the occupied; check if there are occupied MOs not being corrected by GW
      IF (gw_corr_lev_occ < homo .AND. gw_corr_lev_occ > 0) THEN

         ! calculate average GW correction for occupied orbitals
         eigen_diff = 0.0_dp

         DO n_level_gw = 1, gw_corr_lev_occ
            n_level_gw_ref = n_level_gw+homo-gw_corr_lev_occ
            eigen_diff = eigen_diff+Eigenval(n_level_gw_ref)-Eigenval_last(n_level_gw_ref)
         END DO
         eigen_diff = eigen_diff/gw_corr_lev_occ

         ! correct the eigenvalues of the occupied orbitals which have not been corrected by GW
         DO n_level_gw = 1, homo-gw_corr_lev_occ
            Eigenval(n_level_gw) = Eigenval(n_level_gw)+eigen_diff
         END DO

      END IF

      ! 2) the virtual: check if there are virtual orbitals not being corrected by GW
      IF (gw_corr_lev_virt < nmo-homo .AND. gw_corr_lev_virt > 0) THEN

         ! calculate average GW correction for virtual orbitals
         eigen_diff = 0.0_dp
         DO n_level_gw = 1, gw_corr_lev_virt
            n_level_gw_ref = n_level_gw+homo
            eigen_diff = eigen_diff+Eigenval(n_level_gw_ref)-Eigenval_last(n_level_gw_ref)
         END DO
         eigen_diff = eigen_diff/gw_corr_lev_virt

         ! correct the eigenvalues of the virtual orbitals which have not been corrected by GW
         DO n_level_gw = homo+gw_corr_lev_virt+1, nmo
            Eigenval(n_level_gw) = Eigenval(n_level_gw)+eigen_diff
         END DO

      END IF

      IF ((gw_corr_lev_occ == 0 .OR. gw_corr_lev_virt == 0) .AND. count_ev_sc_GW > 1) THEN
         CALL cp_warn(cp__l("rpa_ri_gpw.F",5204), &
                      "Please increase for eigenvalue-self-consistent GW, the number of "// &
                      "corrected occupied and/or virtual orbitals above 0.")
      END IF

      CALL timestop(handle)

   END SUBROUTINE print_and_update_for_ev_sc

! **************************************************************************************************
!> \brief Do the numerical integration for the contour deformation technique:
!>        First, numerical integration on the imag. freq. axis, then residues
!> \param vec_Sigma_c_gw_real_freq ...
!> \param wj ...
!> \param vec_W_gw ...
!> \param Eigenval ...
!> \param omega ...
!> \param contour_def_offset ...
!> \param e_fermi ...
!> \param mo_from_jquad ...
!> \param residue_from_jquad ...
!> \param alpha_beta_from_jquad ...
!> \param contour_def_start ...
!> \param contour_def_end ...
!> \param n_global ...
!> \param jquad ...
!> \param num_integ_points ...
!> \param m_global ...
!> \param nm_global ...
!> \param alpha_beta_case ...
!> \param alpha ...
!> \param beta ...
! **************************************************************************************************
   SUBROUTINE contour_def_integrate_and_residues(vec_Sigma_c_gw_real_freq, wj, vec_W_gw, Eigenval, &
                                                 omega, contour_def_offset, e_fermi, &
                                                 mo_from_jquad, residue_from_jquad, alpha_beta_from_jquad, &
                                                 contour_def_start, contour_def_end, n_global, jquad, &
                                                 num_integ_points, m_global, nm_global, &
                                                 alpha_beta_case, alpha, beta)

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: vec_Sigma_c_gw_real_freq, wj, vec_W_gw
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      REAL(KIND=dp)                                      :: omega, contour_def_offset, e_fermi
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: mo_from_jquad, residue_from_jquad, &
                                                            alpha_beta_from_jquad
      INTEGER                                            :: contour_def_start, contour_def_end, &
                                                            n_global, jquad, num_integ_points, &
                                                            m_global, nm_global
      LOGICAL, OPTIONAL                                  :: alpha_beta_case, alpha, beta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'contour_def_integrate_and_residues', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, index_contour_def, &
                                                            m_level_of_real_energy, &
                                                            n_level_of_real_energy, spin, &
                                                            spin_of_residue
      LOGICAL                                            :: my_alpha, my_alpha_beta_case, my_beta
      REAL(KIND=dp)                                      :: real_energy

      CALL timeset(routineN, handle)

      my_alpha_beta_case = .FALSE.
      IF (PRESENT(alpha_beta_case)) my_alpha_beta_case = alpha_beta_case

      my_alpha = .FALSE.
      IF (PRESENT(alpha)) my_alpha = alpha

      my_beta = .FALSE.
      IF (PRESENT(beta)) my_beta = beta

      spin = 1
      IF (my_alpha_beta_case) THEN
         IF (my_alpha) spin = 1
         IF (my_beta) spin = 2
      END IF

      ! check whether we apply contour deformation to the level n_global
      IF (contour_def_start <= n_global .AND. n_global <= contour_def_end) THEN

         ! check whether we do integration over imag. frequencies or we compute the residues
         IF (jquad <= num_integ_points) THEN

            ! do the integration on the imaginary frequency axis, first for energy slightly below e_n
            real_energy = Eigenval(n_global)-contour_def_offset

            index_contour_def = 2*(n_global-contour_def_start+1)-1

            CALL numerical_integrate_contour_def(vec_Sigma_c_gw_real_freq(index_contour_def), vec_W_gw(nm_global), &
                                                 wj(jquad), omega, real_energy, Eigenval(m_global))

            ! now, for energy slightly above e_n
            real_energy = Eigenval(n_global)+contour_def_offset

            index_contour_def = 2*(n_global-contour_def_start+1)

            CALL numerical_integrate_contour_def(vec_Sigma_c_gw_real_freq(index_contour_def), vec_W_gw(nm_global), &
                                                 wj(jquad), omega, real_energy, Eigenval(m_global))

            ! residues
         ELSE

            ! for jquad > num_integ_points, we add the residue, but only for a single MO n and for a single residue m
            n_level_of_real_energy = mo_from_jquad(jquad-num_integ_points)
            m_level_of_real_energy = residue_from_jquad(jquad-num_integ_points)

            spin_of_residue = 1
            IF (my_alpha_beta_case) THEN
               spin_of_residue = alpha_beta_from_jquad(jquad-num_integ_points)
            END IF

            ! only add residue to a specific level
            IF (n_level_of_real_energy == n_global .AND. &
                m_level_of_real_energy == m_global .AND. &
                spin == spin_of_residue) THEN

               ! add residue, first for energy slightly below e_n
               index_contour_def = 2*(n_global-contour_def_start+1)-1

               IF (e_fermi-Eigenval(n_global) < omega .AND. omega < -contour_def_offset) THEN
                  ! update for residue
                  vec_Sigma_c_gw_real_freq(index_contour_def) = &
                     vec_Sigma_c_gw_real_freq(index_contour_def) &
                     +vec_W_gw(nm_global)
               ELSE IF (-contour_def_offset < omega .AND. omega < e_fermi-Eigenval(n_global)) THEN
                  vec_Sigma_c_gw_real_freq(index_contour_def) = &
                     vec_Sigma_c_gw_real_freq(index_contour_def) &
                     -vec_W_gw(nm_global)
               END IF

               ! add residue, first for energy slightly above e_n
               index_contour_def = 2*(n_global-contour_def_start+1)

               IF (e_fermi-Eigenval(n_global) < omega .AND. omega < contour_def_offset) THEN
                  ! update for residue
                  vec_Sigma_c_gw_real_freq(index_contour_def) = &
                     vec_Sigma_c_gw_real_freq(index_contour_def) &
                     +vec_W_gw(nm_global)
               ELSE IF (contour_def_offset < omega .AND. omega < e_fermi-Eigenval(n_global)) THEN
                  vec_Sigma_c_gw_real_freq(index_contour_def) = &
                     vec_Sigma_c_gw_real_freq(index_contour_def) &
                     -vec_W_gw(nm_global)
               END IF

            END IF ! check whether we have the right residue

         END IF ! imag. frequency / real frequency

      END IF ! check contour deformation for n_global

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param Sigma_c_to_update ...
!> \param W_gw_element ...
!> \param weight ...
!> \param omega ...
!> \param real_energy ...
!> \param eigenval ...
! **************************************************************************************************
   SUBROUTINE numerical_integrate_contour_def(Sigma_c_to_update, W_gw_element, &
                                              weight, omega, real_energy, eigenval)

      REAL(KIND=dp)                                      :: Sigma_c_to_update, W_gw_element, weight, &
                                                            omega, real_energy, eigenval

      CHARACTER(LEN=*), PARAMETER :: routineN = 'numerical_integrate_contour_def', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: im_unit
      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      im_unit = (0.0_dp, 1.0_dp)

      Sigma_c_to_update = Sigma_c_to_update- &
                          REAL( &
                          0.5_dp/pi*weight/2.0_dp*W_gw_element/(im_unit*omega+real_energy-eigenval)+ &
                          0.5_dp/pi*weight/2.0_dp*W_gw_element/(-im_unit*omega+real_energy-eigenval) &
                          )

      CALL timestop(handle)

   END SUBROUTINE

END MODULE rpa_ri_gpw
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate MP2 energy using pw
!> \par History
!>      10.2011 created [Joost VandeVondele and Mauro Del Ben]
! **************************************************************************************************
MODULE mp2_gpw
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: BLACS_GRID_SQUARE,&
                                              cp_blacs_env_create,&
                                              cp_blacs_env_release,&
                                              cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_clear_mempools, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_distribution_release, &
        cp_dbcsr_filter, cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_init_p, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, &
        cp_dbcsr_reserve_all_blocks, cp_dbcsr_set, cp_dbcsr_type, dbcsr_distribution_obj, &
        dbcsr_type_no_symmetry, dbcsr_type_real_default, dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_dist2d_to_dist,&
                                              cp_dbcsr_m_by_n_from_row_template,&
                                              cp_dbcsr_m_by_n_from_template
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_add_default_logger,&
                                              cp_get_default_logger,&
                                              cp_logger_create,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_release,&
                                              cp_logger_set,&
                                              cp_logger_type,&
                                              cp_rm_default_logger
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_release,&
                                              distribution_1d_type
   USE distribution_2d_types,           ONLY: distribution_2d_release,&
                                              distribution_2d_type
   USE distribution_methods,            ONLY: distribute_molecules_1d,&
                                              distribute_molecules_2d
   USE input_section_types,             ONLY: section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE machine,                         ONLY: default_output_unit,&
                                              m_flush,&
                                              m_memory
   USE message_passing,                 ONLY: mp_comm_split_direct,&
                                              mp_max,&
                                              mp_min,&
                                              mp_sendrecv,&
                                              mp_sum
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE mp2_cphf,                        ONLY: solve_z_vector_eq
   USE mp2_ri_gpw,                      ONLY: mp2_ri_gpw_compute_en,&
                                              mp2_ri_gpw_compute_in
   USE mp2_ri_grad,                     ONLY: calc_ri_mp2_nonsep
   USE mp2_types,                       ONLY: mp2_type
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE pw_env_methods,                  ONLY: pw_env_create,&
                                              pw_env_rebuild
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_release,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_scale,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_release
   USE qs_collocate_density,            ONLY: calculate_wavefunction
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_integrate_potential,          ONLY: integrate_v_rspace
   USE qs_interactions,                 ONLY: init_interaction_radii
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: deallocate_neighbor_list_set,&
                                              neighbor_list_set_p_type
   USE qs_neighbor_lists,               ONLY: atom2d_build,&
                                              atom2d_cleanup,&
                                              build_neighbor_lists,&
                                              local_atoms_type,&
                                              pair_radius_setup
   USE rpa_ri_gpw,                      ONLY: rpa_ri_compute_en
   USE task_list_methods,               ONLY: generate_qs_task_list
   USE task_list_types,                 ONLY: allocate_task_list,&
                                              deallocate_task_list,&
                                              task_list_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_gpw'

   PUBLIC :: mp2_gpw_main

CONTAINS

! **************************************************************************************************
!> \brief with a big bang to mp2
!> \param qs_env ...
!> \param mp2_env ...
!> \param Emp2 ...
!> \param Emp2_Cou ...
!> \param Emp2_EX ...
!> \param Emp2_S ...
!> \param Emp2_T ...
!> \param mos_mp2 ...
!> \param para_env ...
!> \param unit_nr ...
!> \param calc_forces ...
!> \param calc_ex ...
!> \param do_ri_mp2 ...
!> \param do_ri_rpa ...
!> \param do_ri_sos_laplace_mp2 ...
!> \author Mauro Del Ben and Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE mp2_gpw_main(qs_env, mp2_env, Emp2, Emp2_Cou, Emp2_EX, Emp2_S, Emp2_T, &
                           mos_mp2, para_env, unit_nr, calc_forces, calc_ex, do_ri_mp2, do_ri_rpa, &
                           do_ri_sos_laplace_mp2)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp)                                      :: Emp2, Emp2_Cou, Emp2_EX, Emp2_S, Emp2_T
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos_mp2
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: unit_nr
      LOGICAL, INTENT(IN)                                :: calc_forces
      LOGICAL                                            :: calc_ex
      LOGICAL, OPTIONAL                                  :: do_ri_mp2, do_ri_rpa, &
                                                            do_ri_sos_laplace_mp2

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_gpw_main', routineP = moduleN//':'//routineN

      INTEGER :: blacs_grid_layout, color_sub, comm_sub, dimen, dimen_RI, gw_corr_lev_occ, &
         gw_corr_lev_occ_beta, gw_corr_lev_virt, gw_corr_lev_virt_beta, handle, homo, homo_beta, &
         i, i_multigrid, local_unit_nr, my_group_L_end, my_group_L_size, my_group_L_start, &
         n_multigrid, nelectron, nelectron_beta, nmo, nspins, ri_metric, ri_metric_gw
      INTEGER, ALLOCATABLE, DIMENSION(:) :: ends_array, ends_B_all, ends_B_virtual, &
         ends_B_virtual_beta, sizes_array, sizes_B_all, sizes_B_virtual, sizes_B_virtual_beta, &
         starts_array, starts_B_all, starts_B_virtual, starts_B_virtual_beta
      LOGICAL :: blacs_repeatable, do_im_time, my_do_gw, my_do_ri_mp2, my_do_ri_rpa, &
         my_do_ri_sos_laplace_mp2, skip_load_balance_distributed
      REAL(KIND=dp) :: cutoff_old, Emp2_AB, Emp2_BB, Emp2_Cou_BB, Emp2_d2_AB, Emp2_d_AB, &
         Emp2_EX_BB, eps_gvg_rspace_old, eps_mp2_old, eps_pgf_orb_old, eps_rho_rspace_old, &
         progression_factor, relative_cutoff_old
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: e_cutoff_old, Eigenval, Eigenval_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: my_Lrows
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C, BIb_C_beta, BIb_C_gw, &
                                                            BIb_C_gw_beta
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues, mo_eigenvalues_beta
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env_sub
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mat_B_munu
      TYPE(cp_dbcsr_type), POINTER :: mo_coeff_all, mo_coeff_all_beta, mo_coeff_gw, &
         mo_coeff_gw_beta, mo_coeff_o, mo_coeff_o_beta, mo_coeff_v, mo_coeff_v_beta
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_beta
      TYPE(cp_logger_type), POINTER                      :: logger, logger_sub
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(local_atoms_type), ALLOCATABLE, DIMENSION(:)  :: atom2d
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb_sub
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env_sub
      TYPE(pw_p_type)                                    :: pot_g, rho_g, rho_r
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(task_list_type), POINTER                      :: task_list_sub

      CALL timeset(routineN, handle)

      ! check if we want to do ri-mp2
      my_do_ri_mp2 = .FALSE.
      IF (PRESENT(do_ri_mp2)) my_do_ri_mp2 = do_ri_mp2

      ! check if we want to do ri-rpa
      my_do_ri_rpa = .FALSE.
      IF (PRESENT(do_ri_rpa)) my_do_ri_rpa = do_ri_rpa

      ! check if we want to do ri-sos-laplace-mp2
      my_do_ri_sos_laplace_mp2 = .FALSE.
      IF (PRESENT(do_ri_sos_laplace_mp2)) my_do_ri_sos_laplace_mp2 = do_ri_sos_laplace_mp2

      ! check if we want to do imaginary time
      do_im_time = mp2_env%ri_rpa%do_im_time

      ! Get the number of spins
      nspins = SIZE(mos_mp2)

      ! ... setup needed to be able to qs_integrate in a subgroup.
      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, mos=mos)
      CALL get_mo_set(mo_set=mos_mp2(1)%mo_set, nelectron=nelectron, &
                      eigenvalues=mo_eigenvalues, nmo=nmo, homo=homo, &
                      mo_coeff=mo_coeff, nao=dimen)
      IF (nspins == 2) THEN
         CALL get_mo_set(mo_set=mos_mp2(2)%mo_set, nelectron=nelectron_beta, &
                         eigenvalues=mo_eigenvalues_beta, homo=homo_beta, &
                         mo_coeff=mo_coeff_beta)
      ENDIF

      CALL get_mo_set(mo_set=mos(1)%mo_set)

      ! a para_env
      color_sub = para_env%mepos/mp2_env%mp2_num_proc
      CALL mp_comm_split_direct(para_env%group, comm_sub, color_sub)
      NULLIFY (para_env_sub)
      CALL cp_para_env_create(para_env_sub, comm_sub)

      ! each of the sub groups might need to generate output
      logger => cp_get_default_logger()
      IF (para_env%mepos == para_env%source) THEN
         local_unit_nr = cp_logger_get_default_unit_nr(logger, local=.FALSE.)
      ELSE
         local_unit_nr = default_output_unit
      ENDIF

      IF (unit_nr > 0) THEN
         WRITE (UNIT=unit_nr, FMT="(T3,A,T71,F10.1)") &
            "GPW_INFO| Density cutoff [a.u.]:", mp2_env%mp2_gpw%cutoff*0.5_dp
         WRITE (UNIT=unit_nr, FMT="(T3,A,T71,F10.1)") &
            "GPW_INFO| Relative density cutoff [a.u.]:", mp2_env%mp2_gpw%relative_cutoff*0.5_dp
         CALL m_flush(unit_nr)
      ENDIF

      ! a logger
      NULLIFY (logger_sub)
      CALL cp_logger_create(logger_sub, para_env=para_env_sub, &
                            default_global_unit_nr=local_unit_nr, close_global_unit_on_dealloc=.FALSE.)
      CALL cp_logger_set(logger_sub, local_filename="MP2_localLog")
      ! set to a custom print level (we could also have a different print level for para_env%source)
      logger_sub%iter_info%print_level = mp2_env%mp2_gpw%print_level
      CALL cp_add_default_logger(logger_sub)

      ! a blacs_env (ignore the globenv stored defaults for now)
      blacs_grid_layout = BLACS_GRID_SQUARE
      blacs_repeatable = .TRUE.
      NULLIFY (blacs_env_sub)
      CALL cp_blacs_env_create(blacs_env_sub, para_env_sub, &
                               blacs_grid_layout, &
                               blacs_repeatable)

      ! get stuff
      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      molecule_set=molecule_set, &
                      molecule_kind_set=molecule_kind_set, &
                      dft_control=dft_control)

      CALL create_mat_munu(mat_munu, qs_env, mp2_env, dft_control, atomic_kind_set, qs_kind_set, &
                           atom2d, molecule_kind_set, &
                           molecule_set, sab_orb_sub, particle_set, cell, blacs_env_sub, &
                           eps_pgf_orb_old, eps_rho_rspace_old, eps_gvg_rspace_old, &
                           unit_nr)

      ! check which RI metric we want to have
      ri_metric = mp2_env%ri_metric

      ! check if we want to do ri-g0w0 on top of ri-rpa
      my_do_gw = mp2_env%ri_rpa%do_ri_g0w0
      gw_corr_lev_occ = mp2_env%ri_g0w0%corr_mos_occ
      gw_corr_lev_virt = mp2_env%ri_g0w0%corr_mos_virt

      ! check Coulomb metric for GW matrix elements B^nm_P
      ri_metric_gw = mp2_env%ri_g0w0%ri_metric

      ! and the array of mos
      ALLOCATE (Eigenval(dimen))
      Eigenval(:) = mo_eigenvalues(:)
      IF (nspins == 2) THEN
         ALLOCATE (Eigenval_beta(dimen))
         Eigenval_beta(:) = mo_eigenvalues_beta(:)
      ENDIF

      ! new routine: replicate a full matrix from one para_env to a smaller one
      ! keeping the memory usage as small as possible in this case the
      ! output the two part of the C matrix (virtual, occupied)
      CALL replicate_mat_to_subgroup(mp2_env, para_env, para_env_sub, mo_coeff, dimen, homo, mat_munu, &
                                     mo_coeff_o, mo_coeff_v, mo_coeff_all, mo_coeff_gw, my_do_gw, &
                                     gw_corr_lev_occ, gw_corr_lev_virt)

      ! if open shell case replicate also the coefficient matrix for the beta orbitals
      IF (nspins == 2) THEN

         gw_corr_lev_occ_beta = mp2_env%ri_g0w0%corr_mos_occ_beta
         gw_corr_lev_virt_beta = mp2_env%ri_g0w0%corr_mos_virt_beta

         CALL replicate_mat_to_subgroup(mp2_env, para_env, para_env_sub, mo_coeff_beta, dimen, homo_beta, mat_munu, &
                                        mo_coeff_o_beta, mo_coeff_v_beta, mo_coeff_all_beta, mo_coeff_gw_beta, &
                                        my_do_gw, gw_corr_lev_occ_beta, gw_corr_lev_virt_beta)
      END IF

      ! hack hack hack XXXXXXXXXXXXXXX rebuilds the pw_en with the new cutoffs
      progression_factor = dft_control%qs_control%progression_factor
      n_multigrid = SIZE(dft_control%qs_control%e_cutoff)
      ALLOCATE (e_cutoff_old(n_multigrid))
      e_cutoff_old(:) = dft_control%qs_control%e_cutoff
      cutoff_old = dft_control%qs_control%cutoff

      dft_control%qs_control%cutoff = mp2_env%mp2_gpw%cutoff*0.5_dp
      dft_control%qs_control%e_cutoff(1) = dft_control%qs_control%cutoff
      DO i_multigrid = 2, n_multigrid
         dft_control%qs_control%e_cutoff(i_multigrid) = dft_control%qs_control%e_cutoff(i_multigrid-1) &
                                                        /progression_factor
      END DO

      relative_cutoff_old = dft_control%qs_control%relative_cutoff
      dft_control%qs_control%relative_cutoff = mp2_env%mp2_gpw%relative_cutoff*0.5_dp

      ! a pw_env
      NULLIFY (pw_env_sub)
      CALL pw_env_create(pw_env_sub)
      CALL pw_env_rebuild(pw_env_sub, qs_env, para_env_sub)

      CALL pw_env_get(pw_env_sub, auxbas_pw_pool=auxbas_pw_pool, &
                      poisson_env=poisson_env)
      ! hack hack hack XXXXXXXXXXXXXXX

      ! now we need a task list, hard code skip_load_balance_distributed
      NULLIFY (task_list_sub)
      skip_load_balance_distributed = dft_control%qs_control%skip_load_balance_distributed
      CALL allocate_task_list(task_list_sub)
      CALL generate_qs_task_list(ks_env, task_list_sub, &
                                 reorder_rs_grid_ranks=.TRUE., soft_valid=.FALSE., &
                                 skip_load_balance_distributed=skip_load_balance_distributed, &
                                 pw_env_external=pw_env_sub, sab_orb_external=sab_orb_sub)

      ! get some of the grids ready
      NULLIFY (rho_r%pw, rho_g%pw, pot_g%pw)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_r%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, pot_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      ! run the FFT once, to set up buffers and to take into account the memory
      rho_r%pw%cr3d = 0.0D0
      CALL pw_transfer(rho_r%pw, rho_g%pw)

      ! now we're kind of ready to go....
      Emp2_S = 0.0_dp
      Emp2_T = 0.0_dp
      IF (my_do_ri_mp2 .OR. my_do_ri_rpa .OR. my_do_ri_sos_laplace_mp2) THEN
         ! RI-GPW integrals (same stuff for both RPA and MP2)
         IF (nspins == 2) THEN
            ! open shell case (RI) here the (ia|K) integrals are computed for both the alpha and beta components
            CALL mp2_ri_gpw_compute_in( &
               BIb_C, BIb_C_gw, ends_array, ends_B_virtual, sizes_array, &
               sizes_B_virtual, starts_array, starts_B_virtual, &
               dimen_RI, qs_env, para_env, para_env_sub, color_sub, dft_control, cell, particle_set, &
               atomic_kind_set, qs_kind_set, mo_coeff, nmo, homo, rho_r, rho_g, pot_g, &
               mat_munu, sab_orb_sub, pw_env_sub, poisson_env, auxbas_pw_pool, task_list_sub, &
               mo_coeff_o, mo_coeff_v, mo_coeff_all, mo_coeff_gw, &
               mp2_env%mp2_gpw%eps_filter, unit_nr, &
               mp2_env%mp2_memory, mp2_env%calc_PQ_cond_num, calc_forces, blacs_env_sub, my_do_gw, &
               starts_B_all, sizes_B_all, ends_B_all, gw_corr_lev_occ, gw_corr_lev_virt, &
               BIb_C_beta, BIb_C_gw_beta, ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta, &
               homo_beta, mo_coeff_o_beta, mo_coeff_v_beta, mo_coeff_all_beta, mo_coeff_gw_beta, &
               do_im_time=do_im_time, mat_B_munu=mat_B_munu, ri_metric=ri_metric, &
               ri_metric_gw=ri_metric_gw, my_Lrows=my_Lrows)
         ELSE
            ! closed shell case (RI)
            CALL mp2_ri_gpw_compute_in(BIb_C, BIb_C_gw, ends_array, ends_B_virtual, sizes_array, &
                                       sizes_B_virtual, starts_array, starts_B_virtual, &
                                       dimen_RI, qs_env, para_env, para_env_sub, color_sub, dft_control, cell, particle_set, &
                                       atomic_kind_set, qs_kind_set, mo_coeff, nmo, homo, rho_r, rho_g, pot_g, &
                                       mat_munu, sab_orb_sub, pw_env_sub, poisson_env, auxbas_pw_pool, task_list_sub, &
                                       mo_coeff_o, mo_coeff_v, mo_coeff_all, mo_coeff_gw, &
                                       mp2_env%mp2_gpw%eps_filter, unit_nr, &
                                       mp2_env%mp2_memory, mp2_env%calc_PQ_cond_num, calc_forces, blacs_env_sub, my_do_gw, &
                                       starts_B_all, sizes_B_all, ends_B_all, gw_corr_lev_occ, gw_corr_lev_virt, &
                                       do_im_time=do_im_time, mat_B_munu=mat_B_munu, ri_metric=ri_metric, &
                                       ri_metric_gw=ri_metric_gw, my_Lrows=my_Lrows)

         END IF
      ELSE
         ! Canonical MP2-GPW
         IF (nspins == 2) THEN
            ! alpha-alpha and alpha-beta components
            IF (unit_nr > 0) WRITE (unit_nr, *)
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'Alpha (ia|'
            CALL mp2_gpw_compute( &
               Emp2, Emp2_Cou, Emp2_EX, qs_env, para_env, para_env_sub, color_sub, dft_control, cell, particle_set, &
               atomic_kind_set, qs_kind_set, mo_coeff, Eigenval, nmo, homo, rho_r, rho_g, pot_g, &
               mat_munu, pw_env_sub, poisson_env, auxbas_pw_pool, task_list_sub, &
               mo_coeff_o, mo_coeff_v, mp2_env%mp2_gpw%eps_filter, unit_nr, &
               mp2_env%mp2_memory, calc_ex, blacs_env_sub, &
               homo_beta, mo_coeff_o_beta, mo_coeff_v_beta, Eigenval_beta, Emp2_AB)

            ! beta-beta component
            IF (unit_nr > 0) WRITE (unit_nr, *)
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'Beta (ia|'
            CALL mp2_gpw_compute( &
               Emp2_BB, Emp2_Cou_BB, Emp2_EX_BB, qs_env, para_env, para_env_sub, color_sub, dft_control, cell, particle_set, &
               atomic_kind_set, qs_kind_set, mo_coeff_beta, Eigenval_beta, nmo, homo_beta, rho_r, rho_g, pot_g, &
               mat_munu, pw_env_sub, poisson_env, auxbas_pw_pool, task_list_sub, &
               mo_coeff_o_beta, mo_coeff_v_beta, mp2_env%mp2_gpw%eps_filter, unit_nr, &
               mp2_env%mp2_memory, calc_ex, blacs_env_sub)

         ELSE
            ! closed shell case
            CALL mp2_gpw_compute( &
               Emp2, Emp2_Cou, Emp2_EX, qs_env, para_env, para_env_sub, color_sub, dft_control, cell, particle_set, &
               atomic_kind_set, qs_kind_set, mo_coeff, Eigenval, nmo, homo, rho_r, rho_g, pot_g, &
               mat_munu, pw_env_sub, poisson_env, auxbas_pw_pool, task_list_sub, &
               mo_coeff_o, mo_coeff_v, mp2_env%mp2_gpw%eps_filter, unit_nr, &
               mp2_env%mp2_memory, calc_ex, blacs_env_sub)
         END IF
      END IF

      ! and now free the whole lot
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_r%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_g%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, pot_g%pw)

      ! Free possibly large buffers allocated by dbcsr on the GPU,
      ! large hybrid dgemm/pdgemm's coming later will need the space.
      CALL cp_dbcsr_clear_mempools()

      ! moved down
      ! CALL deallocate_task_list(task_list_sub)
      ! CALL pw_env_release(pw_env_sub)

      IF (calc_forces) THEN
         ! make a copy of mo_coeff_o and mo_coeff_v
         NULLIFY (mp2_env%ri_grad%mo_coeff_o)
         CALL cp_dbcsr_init_p(mp2_env%ri_grad%mo_coeff_o)
         CALL cp_dbcsr_copy(mp2_env%ri_grad%mo_coeff_o, mo_coeff_o, name="mo_coeff_o")
         NULLIFY (mp2_env%ri_grad%mo_coeff_v)
         CALL cp_dbcsr_init_p(mp2_env%ri_grad%mo_coeff_v)
         CALL cp_dbcsr_copy(mp2_env%ri_grad%mo_coeff_v, mo_coeff_v, name="mo_coeff_v")
         IF (nspins == 2) THEN
            NULLIFY (mp2_env%ri_grad%mo_coeff_o_beta)
            CALL cp_dbcsr_init_p(mp2_env%ri_grad%mo_coeff_o_beta)
            CALL cp_dbcsr_copy(mp2_env%ri_grad%mo_coeff_o_beta, mo_coeff_o_beta, name="mo_coeff_o_b")
            NULLIFY (mp2_env%ri_grad%mo_coeff_v_beta)
            CALL cp_dbcsr_init_p(mp2_env%ri_grad%mo_coeff_v_beta)
            CALL cp_dbcsr_copy(mp2_env%ri_grad%mo_coeff_v_beta, mo_coeff_v_beta, name="mo_coeff_v_b")
         ENDIF
         my_group_L_size = sizes_array(color_sub)
         my_group_L_start = starts_array(color_sub)
         my_group_L_end = ends_array(color_sub)
      END IF

      CALL cp_dbcsr_release(mo_coeff_o)
      DEALLOCATE (mo_coeff_o)
      CALL cp_dbcsr_release(mo_coeff_v)
      DEALLOCATE (mo_coeff_v)
      IF (my_do_gw) THEN
         CALL cp_dbcsr_release(mo_coeff_all)
         DEALLOCATE (mo_coeff_all)
      END IF

      IF (nspins == 2) THEN
         CALL cp_dbcsr_release(mo_coeff_o_beta)
         DEALLOCATE (mo_coeff_o_beta)
         CALL cp_dbcsr_release(mo_coeff_v_beta)
         DEALLOCATE (mo_coeff_v_beta)
         IF (my_do_gw) THEN
            CALL cp_dbcsr_release(mo_coeff_all_beta)
            DEALLOCATE (mo_coeff_all_beta)
         END IF
      END IF

      IF (.NOT. calc_forces) THEN

         CALL cp_dbcsr_release(mat_munu%matrix)
         DEALLOCATE (mat_munu%matrix)

         DO i = 1, SIZE(sab_orb_sub)
            CALL deallocate_neighbor_list_set(sab_orb_sub(i)%neighbor_list_set)
         END DO
         DEALLOCATE (sab_orb_sub)

      END IF

      ! decide if to doing RI-RPA or RI-MP2
      IF (my_do_ri_rpa .OR. my_do_ri_sos_laplace_mp2) THEN

         IF (do_im_time) THEN

            ! for imaginary time, we need a non-sparse dbcsr matrix mat_munu

            eps_mp2_old = mp2_env%mp2_gpw%eps_grid
            mp2_env%mp2_gpw%eps_grid = 1.0E-290_dp

            CALL create_mat_munu(mat_munu, qs_env, mp2_env, dft_control, atomic_kind_set, qs_kind_set, &
                                 atom2d, molecule_kind_set, &
                                 molecule_set, sab_orb_sub, particle_set, cell, blacs_env_sub, &
                                 eps_pgf_orb_old, eps_rho_rspace_old, eps_gvg_rspace_old, &
                                 unit_nr)

            mp2_env%mp2_gpw%eps_grid = eps_mp2_old

         END IF

         ! RI-RPA
         IF (nspins == 2) THEN
            CALL rpa_ri_compute_en(Emp2, mp2_env, BIb_C, BIb_C_gw, para_env, para_env_sub, color_sub, &
                                   ends_array, ends_B_virtual, ends_B_all, sizes_array, sizes_B_virtual, sizes_B_all, &
                                   starts_array, starts_B_virtual, starts_B_all, mo_coeff, &
                                   Eigenval, nmo, homo, dimen_RI, gw_corr_lev_occ, gw_corr_lev_virt, &
                                   unit_nr, my_do_ri_sos_laplace_mp2, my_do_gw, &
                                   BIb_C_beta, homo_beta, Eigenval_beta, &
                                   ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta, &
                                   mo_coeff_beta, BIb_C_gw_beta, gw_corr_lev_occ_beta, gw_corr_lev_virt_beta, &
                                   do_im_time=do_im_time, mat_munu=mat_munu, mat_B_munu=mat_B_munu, my_Lrows=my_Lrows, &
                                   eps_filter=mp2_env%mp2_gpw%eps_filter)
         ELSE
            CALL rpa_ri_compute_en(Emp2, mp2_env, BIb_C, BIb_C_gw, para_env, para_env_sub, color_sub, &
                                   ends_array, ends_B_virtual, ends_B_all, sizes_array, sizes_B_virtual, sizes_B_all, &
                                   starts_array, starts_B_virtual, starts_B_all, mo_coeff, &
                                   Eigenval, nmo, homo, dimen_RI, gw_corr_lev_occ, gw_corr_lev_virt, &
                                   unit_nr, my_do_ri_sos_laplace_mp2, my_do_gw, do_im_time=do_im_time, &
                                   mat_munu=mat_munu, mat_B_munu=mat_B_munu, my_Lrows=my_Lrows, &
                                   eps_filter=mp2_env%mp2_gpw%eps_filter)
         END IF

         IF (do_im_time) THEN

            CALL cp_dbcsr_release(mat_munu%matrix)
            DEALLOCATE (mat_munu%matrix)

            DO i = 1, SIZE(sab_orb_sub)
               CALL deallocate_neighbor_list_set(sab_orb_sub(i)%neighbor_list_set)
            END DO
            DEALLOCATE (sab_orb_sub)

         END IF

      ELSE
         IF (my_do_ri_mp2) THEN
            ! RI-MP2-GPW compute energy
            IF (nspins == 2) THEN
               ! alpha-alpha component
               CALL mp2_ri_gpw_compute_en( &
                  Emp2, Emp2_Cou, Emp2_EX, BIb_C, mp2_env, para_env, para_env_sub, color_sub, &
                  ends_array, ends_B_virtual, sizes_array, sizes_B_virtual, starts_array, starts_B_virtual, &
                  Eigenval, nmo, homo, dimen_RI, unit_nr, calc_forces, calc_ex, &
                  open_shell_SS=.TRUE.)

               ! beta-beta component
               CALL mp2_ri_gpw_compute_en( &
                  Emp2_BB, Emp2_Cou_BB, Emp2_EX_BB, BIb_C_beta, mp2_env, para_env, para_env_sub, color_sub, &
                  ends_array, ends_B_virtual_beta, sizes_array, &
                  sizes_B_virtual_beta, starts_array, starts_B_virtual_beta, &
                  Eigenval_beta, nmo, homo_beta, dimen_RI, unit_nr, calc_forces, calc_ex, &
                  open_shell_SS=.TRUE.)

               ! alpha-beta case
               CALL mp2_ri_gpw_compute_en( &
                  Emp2_d_AB, Emp2_AB, Emp2_d2_AB, BIb_C, mp2_env, para_env, para_env_sub, color_sub, &
                  ends_array, ends_B_virtual, sizes_array, sizes_B_virtual, starts_array, starts_B_virtual, &
                  Eigenval, nmo, homo, dimen_RI, unit_nr, calc_forces, .FALSE., &
                  .FALSE., BIb_C_beta, homo_beta, Eigenval_beta, &
                  ends_B_virtual_beta, sizes_B_virtual_beta, starts_B_virtual_beta)

            ELSE
               ! closed shell case
               CALL mp2_ri_gpw_compute_en( &
                  Emp2, Emp2_Cou, Emp2_EX, BIb_C, mp2_env, para_env, para_env_sub, color_sub, &
                  ends_array, ends_B_virtual, sizes_array, sizes_B_virtual, starts_array, starts_B_virtual, &
                  Eigenval, nmo, homo, dimen_RI, unit_nr, calc_forces, calc_ex)
            END IF
            ! if we need forces time to calculate the MP2 non-separable contribution
            ! and start coputing the Largrangian
            IF (calc_forces) THEN
               ! since we have to compute again integrals reinitialize the stuff we need
               ! get some of the grids ready
               NULLIFY (rho_r%pw, rho_g%pw, pot_g%pw)
               CALL pw_pool_create_pw(auxbas_pw_pool, rho_r%pw, &
                                      use_data=REALDATA3D, &
                                      in_space=REALSPACE)
               CALL pw_pool_create_pw(auxbas_pw_pool, rho_g%pw, &
                                      use_data=COMPLEXDATA1D, &
                                      in_space=RECIPROCALSPACE)
               CALL pw_pool_create_pw(auxbas_pw_pool, pot_g%pw, &
                                      use_data=COMPLEXDATA1D, &
                                      in_space=RECIPROCALSPACE)

               ! the mu_nu matrix (again)
               ! XXXXXXXXXXXXXXXXXXXXXXXXX
               ! ! build a dbcsr matrix the hard way
               ! CALL get_particle_set(particle_set=particle_set,nsgf=rbs)
               ! CALL array_nullify (row_blk_sizes)
               ! CALL array_new (row_blk_sizes, rbs, gift=.TRUE.)
               ! ALLOCATE(mat_munu%matrix, STAT=stat)
               ! CPPostcondition(stat==0,cp_failure_level,routineP,failure)
               ! CALL cp_dbcsr_init(mat_munu%matrix)
               ! CALL cp_dbcsr_create(matrix=mat_munu%matrix,&
               !            name="(ai|munu)",&
               !            dist=dbcsr_dist_sub, &
               !            matrix_type=dbcsr_type_symmetric,&
               !            ! matrix_type=dbcsr_type_no_symmetry,&
               !            row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
               !            nze=0)
               ! CALL cp_dbcsr_alloc_block_from_nbl(mat_munu%matrix,sab_orb_sub)
               ! CALL array_release (row_blk_sizes)

               IF (nspins == 2) THEN ! Open shell
                  CALL calc_ri_mp2_nonsep(qs_env, mp2_env, para_env, para_env_sub, dft_control, cell, &
                                          particle_set, atomic_kind_set, qs_kind_set, mo_coeff, nmo, homo, dimen_RI, Eigenval, &
                                          my_group_L_start, my_group_L_end, my_group_L_size, rho_r, rho_g, pot_g, &
                                          mat_munu, pw_env_sub, poisson_env, auxbas_pw_pool, task_list_sub, &
                                          blacs_env_sub, Eigenval_beta, homo_beta, mo_coeff_beta)
               ELSE ! Closed shell
                  CALL calc_ri_mp2_nonsep(qs_env, mp2_env, para_env, para_env_sub, dft_control, cell, &
                                          particle_set, atomic_kind_set, qs_kind_set, mo_coeff, nmo, homo, dimen_RI, Eigenval, &
                                          my_group_L_start, my_group_L_end, my_group_L_size, rho_r, rho_g, pot_g, &
                                          mat_munu, pw_env_sub, poisson_env, auxbas_pw_pool, task_list_sub, &
                                          blacs_env_sub)
               ENDIF

               ! release
               CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_r%pw)
               CALL pw_pool_give_back_pw(auxbas_pw_pool, rho_g%pw)
               CALL pw_pool_give_back_pw(auxbas_pw_pool, pot_g%pw)

               CALL cp_dbcsr_release(mp2_env%ri_grad%mo_coeff_o)
               DEALLOCATE (mp2_env%ri_grad%mo_coeff_o)

               CALL cp_dbcsr_release(mp2_env%ri_grad%mo_coeff_v)
               DEALLOCATE (mp2_env%ri_grad%mo_coeff_v)

               IF (nspins == 2) THEN
                  CALL cp_dbcsr_release(mp2_env%ri_grad%mo_coeff_o_beta)
                  DEALLOCATE (mp2_env%ri_grad%mo_coeff_o_beta)

                  CALL cp_dbcsr_release(mp2_env%ri_grad%mo_coeff_v_beta)
                  DEALLOCATE (mp2_env%ri_grad%mo_coeff_v_beta)
               ENDIF

               CALL cp_dbcsr_release(mat_munu%matrix)
               DEALLOCATE (mat_munu%matrix)

               DO i = 1, SIZE(sab_orb_sub)
                  CALL deallocate_neighbor_list_set(sab_orb_sub(i)%neighbor_list_set)
               END DO
               DEALLOCATE (sab_orb_sub)

            END IF
         END IF

         IF (nspins == 2) THEN
            ! make order on the MP2 energy contributions
            Emp2_Cou = Emp2_Cou*0.25_dp
            Emp2_EX = Emp2_EX*0.5_dp

            Emp2_Cou_BB = Emp2_Cou_BB*0.25_dp
            Emp2_EX_BB = Emp2_EX_BB*0.5_dp

            Emp2_S = Emp2_AB
            Emp2_T = Emp2_Cou+Emp2_Cou_BB+Emp2_EX+Emp2_EX_BB

            Emp2_Cou = Emp2_Cou+Emp2_Cou_BB+Emp2_AB
            Emp2_EX = Emp2_EX+Emp2_EX_BB
            Emp2 = Emp2_EX+Emp2_Cou
         END IF

      END IF

      !XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXx
      ! moved from above
      IF (my_do_gw) THEN
         CALL cp_dbcsr_release(mo_coeff_gw)
         DEALLOCATE (mo_coeff_gw)
         IF (nspins == 2) THEN
            CALL cp_dbcsr_release(mo_coeff_gw_beta)
            DEALLOCATE (mo_coeff_gw_beta)
         END IF
      END IF

      CALL deallocate_task_list(task_list_sub)

      CALL pw_env_release(pw_env_sub)

      ! CALL cp_dbcsr_distribution_release(dbcsr_dist_sub)
      ! DEALLOCATE(dbcsr_dist_sub,STAT=stat)
      ! CPPostcondition(stat==0,cp_failure_level,routineP,failure)

      ! DO i=1,SIZE(sab_orb_sub)
      !    CALL deallocate_neighbor_list_set(sab_orb_sub(i)%neighbor_list_set)
      ! END DO
      ! DEALLOCATE(sab_orb_sub,stat=stat)
      ! CPPostconditionNoFail(stat==0,cp_warning_level,routineP)

      ! CALL distribution_2d_release(distribution_2d_sub)

      ! CALL distribution_1d_release(local_particles_sub)
      ! CALL distribution_1d_release(local_molecules_sub)
      !XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXx

      ! re-init the radii to be able to generate pair lists with MP2-appropriate screening
      dft_control%qs_control%eps_pgf_orb = eps_pgf_orb_old
      dft_control%qs_control%eps_rho_rspace = eps_rho_rspace_old
      dft_control%qs_control%eps_gvg_rspace = eps_gvg_rspace_old
      CALL init_interaction_radii(dft_control%qs_control, atomic_kind_set, qs_kind_set)

      ! restore the initial value of the cutoff
      dft_control%qs_control%e_cutoff = e_cutoff_old
      dft_control%qs_control%cutoff = cutoff_old
      dft_control%qs_control%relative_cutoff = relative_cutoff_old

      CALL cp_blacs_env_release(blacs_env_sub)

      CALL cp_rm_default_logger()
      CALL cp_logger_release(logger_sub)

      CALL cp_para_env_release(para_env_sub)

      ! finally solve the z-vector equation if forces are required
      IF (calc_forces) THEN
         IF (nspins == 2) THEN
            CALL solve_z_vector_eq(qs_env, mp2_env, para_env, dft_control, &
                                   atomic_kind_set, mo_coeff, nmo, homo, Eigenval, unit_nr, &
                                   Eigenval_beta, homo_beta, mo_coeff_beta)
         ELSE
            CALL solve_z_vector_eq(qs_env, mp2_env, para_env, dft_control, &
                                   atomic_kind_set, mo_coeff, nmo, homo, Eigenval, unit_nr)
         ENDIF
      END IF

      DEALLOCATE (Eigenval)
      IF (nspins == 2) THEN
         DEALLOCATE (Eigenval_beta)
      END IF

      CALL timestop(handle)

   END SUBROUTINE mp2_gpw_main

! **************************************************************************************************
!> \brief ...
!> \param Emp2 ...
!> \param Emp2_Cou ...
!> \param Emp2_EX ...
!> \param qs_env ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param color_sub ...
!> \param dft_control ...
!> \param cell ...
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param mo_coeff ...
!> \param Eigenval ...
!> \param nmo ...
!> \param homo ...
!> \param rho_r ...
!> \param rho_g ...
!> \param pot_g ...
!> \param mat_munu ...
!> \param pw_env_sub ...
!> \param poisson_env ...
!> \param auxbas_pw_pool ...
!> \param task_list_sub ...
!> \param mo_coeff_o ...
!> \param mo_coeff_v ...
!> \param eps_filter ...
!> \param unit_nr ...
!> \param mp2_memory ...
!> \param calc_ex ...
!> \param blacs_env_sub ...
!> \param homo_beta ...
!> \param mo_coeff_o_beta ...
!> \param mo_coeff_v_beta ...
!> \param Eigenval_beta ...
!> \param Emp2_AB ...
! **************************************************************************************************
   SUBROUTINE mp2_gpw_compute(Emp2, Emp2_Cou, Emp2_EX, qs_env, para_env, para_env_sub, color_sub, dft_control, &
                              cell, particle_set, atomic_kind_set, qs_kind_set, mo_coeff, Eigenval, nmo, homo, &
                              rho_r, rho_g, pot_g, mat_munu, pw_env_sub, &
                              poisson_env, auxbas_pw_pool, task_list_sub, mo_coeff_o, mo_coeff_v, eps_filter, unit_nr, &
                              mp2_memory, calc_ex, blacs_env_sub, homo_beta, mo_coeff_o_beta, &
                              mo_coeff_v_beta, Eigenval_beta, Emp2_AB)

      REAL(KIND=dp)                                      :: Emp2, Emp2_Cou, Emp2_EX
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      INTEGER                                            :: color_sub
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      REAL(KIND=dp), DIMENSION(:)                        :: Eigenval
      INTEGER                                            :: nmo, homo
      TYPE(pw_p_type)                                    :: rho_r, rho_g, pot_g
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(pw_env_type), POINTER                         :: pw_env_sub
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(task_list_type), POINTER                      :: task_list_sub
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_o, mo_coeff_v
      REAL(KIND=dp)                                      :: eps_filter
      INTEGER                                            :: unit_nr
      REAL(KIND=dp)                                      :: mp2_memory
      LOGICAL                                            :: calc_ex
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env_sub
      INTEGER, OPTIONAL                                  :: homo_beta
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: mo_coeff_o_beta, mo_coeff_v_beta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: Eigenval_beta
      REAL(KIND=dp), OPTIONAL                            :: Emp2_AB

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_gpw_compute', &
         routineP = moduleN//':'//routineN

      INTEGER :: a, a_group_counter, b, b_global, b_group_counter, blk, col, col_offset, col_size, &
         color_counter, comm_exchange, EX_end, EX_end_send, EX_start, EX_start_send, &
         group_counter, handle, handle2, handle3, i, i_counter, i_group_counter, index_proc_shift, &
         j, max_b_size, max_batch_size_A, max_batch_size_I, max_row_col_local, mepos_in_EX_group, &
         my_A_batch_size, my_A_virtual_end, my_A_virtual_start, my_B_size, my_B_virtual_end, &
         my_B_virtual_start, my_I_batch_size, my_I_occupied_end, my_I_occupied_start, &
         my_q_position, ncol_local, nfullcols_total, nfullrows_total, ngroup, nrow_local, one, p
      INTEGER :: p_best, proc_receive, proc_send, q, q_best, row, row_offset, row_size, size_EX, &
         size_EX_send, size_of_exchange_group, sub_sub_color, virtual, virtual_beta, wfn_calc, &
         wfn_calc_best
      INTEGER(KIND=int_8)                                :: mem
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: proc_map, sub_proc_map, vector_B_sizes, &
                                                            vector_batch_A_size_group, &
                                                            vector_batch_I_size_group
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: color_array, exchange_group_sizes, &
                                                            local_col_row_info
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: do_alpha_beta
      REAL(KIND=dp)                                      :: mem_min, mem_real, mem_try, pair_energy, &
                                                            wfn_size
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: my_Cocc, my_Cvirt
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: BIb_C, BIb_Ex, BIb_send
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_block
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type)                                :: matrix_ia_jb, matrix_ia_jb_beta, &
                                                            matrix_ia_jnu, matrix_ia_jnu_beta
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_BIb_jb
      TYPE(cp_para_env_type), POINTER                    :: para_env_exchange
      TYPE(pw_p_type)                                    :: psi_a
      TYPE(pw_p_type), ALLOCATABLE, DIMENSION(:)         :: psi_i

      CALL timeset(routineN, handle)

      do_alpha_beta = .FALSE.
      IF (PRESENT(homo_beta) .AND. &
          PRESENT(mo_coeff_o_beta) .AND. &
          PRESENT(mo_coeff_v_beta) .AND. &
          PRESENT(Eigenval_beta) .AND. &
          PRESENT(Emp2_AB)) do_alpha_beta = .TRUE.

      ! initialize and create the matrix (ia|jnu)
      CALL cp_dbcsr_init(matrix_ia_jnu)
      CALL cp_dbcsr_create(matrix_ia_jnu, template=mo_coeff_o)

      ! Allocate Sparse matrices: (ia|jb)
      CALL cp_dbcsr_init(matrix_ia_jb)
      CALL cp_dbcsr_m_by_n_from_template(matrix_ia_jb, template=mo_coeff_o, m=homo, n=nmo-homo, &
                                         sym=dbcsr_type_no_symmetry)

      ! set all to zero in such a way that the memory is actually allocated
      CALL cp_dbcsr_set(matrix_ia_jnu, 0.0_dp)
      CALL cp_dbcsr_set(matrix_ia_jb, 0.0_dp)
      CALL cp_dbcsr_set(mat_munu%matrix, 0.0_dp)

      IF (calc_ex) THEN
         ! create the analogous of matrix_ia_jb in fm type
         NULLIFY (fm_BIb_jb)
         NULLIFY (fm_struct)
         CALL cp_dbcsr_get_info(matrix_ia_jb, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)
         CALL cp_fm_struct_create(fm_struct, context=blacs_env_sub, nrow_global=nfullrows_total, &
                                  ncol_global=nfullcols_total, para_env=para_env_sub)
         CALL cp_fm_create(fm_BIb_jb, fm_struct, name="fm_BIb_jb")

         CALL copy_dbcsr_to_fm(matrix_ia_jb, fm_BIb_jb)
         CALL cp_fm_struct_release(fm_struct)

         CALL cp_fm_get_info(matrix=fm_BIb_jb, &
                             nrow_local=nrow_local, &
                             ncol_local=ncol_local, &
                             row_indices=row_indices, &
                             col_indices=col_indices)

         max_row_col_local = MAX(nrow_local, ncol_local)
         CALL mp_max(max_row_col_local, para_env_sub%group)

         ALLOCATE (local_col_row_info(0:max_row_col_local, 2))
         local_col_row_info = 0
         ! 0,1 nrows
         local_col_row_info(0, 1) = nrow_local
         local_col_row_info(1:nrow_local, 1) = row_indices(1:nrow_local)
         ! 0,2 ncols
         local_col_row_info(0, 2) = ncol_local
         local_col_row_info(1:ncol_local, 2) = col_indices(1:ncol_local)
      END IF

      IF (do_alpha_beta) THEN
         ! initialize and create the matrix (ia|jnu)
         CALL cp_dbcsr_init(matrix_ia_jnu_beta)
         CALL cp_dbcsr_create(matrix_ia_jnu_beta, template=mo_coeff_o_beta)

         ! Allocate Sparse matrices: (ia|jb)
         CALL cp_dbcsr_init(matrix_ia_jb_beta)
         CALL cp_dbcsr_m_by_n_from_template(matrix_ia_jb_beta, template=mo_coeff_o_beta, m=homo_beta, n=nmo-homo_beta, &
                                            sym=dbcsr_type_no_symmetry)

         virtual_beta = nmo-homo_beta

         CALL cp_dbcsr_set(matrix_ia_jnu_beta, 0.0_dp)
         CALL cp_dbcsr_set(matrix_ia_jb_beta, 0.0_dp)
      END IF

      CALL m_memory(mem)
      mem_real = (mem+1024*1024-1)/(1024*1024)
      ! mp_min .... a hack.. it should be mp_max, but as it turns out, on some processes the previously freed memory (hfx)
      ! has not been given back to the OS yet.
      CALL mp_min(mem_real, para_env%group)

      virtual = nmo-homo

      wfn_size = REAL(SIZE(rho_r%pw%cr3d), KIND=dp)
      CALL mp_max(wfn_size, para_env%group)

      ngroup = para_env%num_pe/para_env_sub%num_pe

      ! calculate the minimal memory required per MPI task (p=occupied division,q=virtual division)
      p_best = ngroup
      q_best = 1
      mem_min = HUGE(0)
      DO p = 1, ngroup
         q = ngroup/p
         IF (p*q .NE. ngroup) CYCLE

         CALL estimate_memory_usage(wfn_size, p, q, para_env_sub%num_pe, nmo, virtual, homo, calc_ex, mem_try)

         IF (mem_try <= mem_min) THEN
            mem_min = mem_try
            p_best = p
            q_best = q
         END IF
      END DO
      IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T69,F9.2,A3)') 'Minimum required memory per MPI process for MP2:', &
         mem_min, ' MB'

      mem_real = mp2_memory-mem_real
      mem_real = MAX(mem_real, mem_min)
      IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T69,F9.2,A3)') 'Available memory per MPI process for MP2:', &
         mem_real, ' MB'

      wfn_calc_best = HUGE(wfn_calc_best)
      DO p = 1, ngroup
         q = ngroup/p
         IF (p*q .NE. ngroup) CYCLE

         CALL estimate_memory_usage(wfn_size, p, q, para_env_sub%num_pe, nmo, virtual, homo, calc_ex, mem_try)

         IF (mem_try > mem_real) CYCLE
         wfn_calc = ((homo+p-1)/p)+((virtual+q-1)/q)
         IF (wfn_calc < wfn_calc_best) THEN
            wfn_calc_best = wfn_calc
            p_best = p
            q_best = q
         ENDIF
      ENDDO

      max_batch_size_I = (homo+p_best-1)/p_best
      max_batch_size_A = (virtual+q_best-1)/q_best

      IF (unit_nr > 0) THEN
         WRITE (UNIT=unit_nr, FMT="(T3,A,T77,i4)") &
            "MP2_GPW| max. batch size for the occupied states:", max_batch_size_I
         WRITE (UNIT=unit_nr, FMT="(T3,A,T77,i4)") &
            "MP2_GPW| max. batch size for the virtual states:", max_batch_size_A
      ENDIF

      ALLOCATE (vector_batch_I_size_group(0:p_best-1))
      ALLOCATE (vector_batch_A_size_group(0:q_best-1))

      vector_batch_I_size_group = max_batch_size_I
      IF (SUM(vector_batch_I_size_group) /= homo) THEN
         one = 1
         IF (SUM(vector_batch_I_size_group) > homo) one = -1
         i = -1
         DO
            i = i+1
            vector_batch_I_size_group(i) = vector_batch_I_size_group(i)+one
            IF (SUM(vector_batch_I_size_group) == homo) EXIT
            IF (i == p_best-1) i = -1
         END DO
      END IF

      vector_batch_A_size_group = max_batch_size_A
      IF (SUM(vector_batch_A_size_group) /= virtual) THEN
         one = 1
         IF (SUM(vector_batch_A_size_group) > virtual) one = -1
         i = -1
         DO
            i = i+1
            vector_batch_A_size_group(i) = vector_batch_A_size_group(i)+one
            IF (SUM(vector_batch_A_size_group) == virtual) EXIT
            IF (i == q_best-1) i = -1
         END DO
      END IF

      !XXXXXXXXXXXXX inverse group distribution
      group_counter = 0
      a_group_counter = 0
      my_A_virtual_start = 1
      DO j = 0, q_best-1
         my_I_occupied_start = 1
         i_group_counter = 0
         DO i = 0, p_best-1
            group_counter = group_counter+1
            IF (color_sub == group_counter-1) EXIT
            my_I_occupied_start = my_I_occupied_start+vector_batch_I_size_group(i)
            i_group_counter = i_group_counter+1
         END DO
         my_q_position = j
         IF (color_sub == group_counter-1) EXIT
         my_A_virtual_start = my_A_virtual_start+vector_batch_A_size_group(j)
         a_group_counter = a_group_counter+1
      END DO
      !XXXXXXXXXXXXX inverse group distribution

      my_I_occupied_end = my_I_occupied_start+vector_batch_I_size_group(i_group_counter)-1
      my_I_batch_size = vector_batch_I_size_group(i_group_counter)
      my_A_virtual_end = my_A_virtual_start+vector_batch_A_size_group(a_group_counter)-1
      my_A_batch_size = vector_batch_A_size_group(a_group_counter)

      DEALLOCATE (vector_batch_I_size_group)
      DEALLOCATE (vector_batch_A_size_group)

      ! replicate on a local array on proc 0 the occupied and virtual wavevectior
      ! needed for the calculation of the WF's by calculate_wavefunction
      ! (external vector)
      CALL grep_occ_virt_wavefunc(para_env_sub, nmo, &
                                  my_I_occupied_start, my_I_occupied_end, my_I_batch_size, &
                                  my_A_virtual_start, my_A_virtual_end, my_A_batch_size, &
                                  mo_coeff_o, mo_coeff_v, my_Cocc, my_Cvirt)

      ! divide the b states in the sub_group in such a way to create
      ! b_start and b_end for each proc inside the sub_group
      max_b_size = (virtual+para_env_sub%num_pe-1)/para_env_sub%num_pe
      ALLOCATE (vector_B_sizes(0:para_env_sub%num_pe-1))
      vector_B_sizes = max_b_size
      IF (SUM(vector_B_sizes) /= virtual) THEN
         one = 1
         IF (SUM(vector_B_sizes) > virtual) one = -1
         i = -1
         DO
            i = i+1
            vector_B_sizes(i) = vector_B_sizes(i)+one
            IF (SUM(vector_B_sizes) == virtual) EXIT
            IF (i == para_env_sub%num_pe-1) i = -1
         END DO
      END IF
      ! now give to each proc its b_start and b_end
      b_group_counter = 0
      my_B_virtual_start = 1
      DO j = 0, para_env_sub%num_pe-1
         b_group_counter = b_group_counter+1
         IF (b_group_counter-1 == para_env_sub%mepos) EXIT
         my_B_virtual_start = my_B_virtual_start+vector_B_sizes(j)
      END DO
      my_B_virtual_end = my_B_virtual_start+vector_B_sizes(para_env_sub%mepos)-1
      my_B_size = vector_B_sizes(para_env_sub%mepos)

      DEALLOCATE (vector_B_sizes)

      ! create an array containing a different "color" for each pair of
      ! A_start and B_start, communication will take place only among
      ! those proc that have the same A_start and B_start
      ALLOCATE (color_array(0:para_env_sub%num_pe-1, 0:q_best-1))
      color_array = 0
      color_counter = 0
      DO j = 0, q_best-1
         DO i = 0, para_env_sub%num_pe-1
            color_counter = color_counter+1
            color_array(i, j) = color_counter
         END DO
      END DO
      sub_sub_color = color_array(para_env_sub%mepos, my_q_position)

      DEALLOCATE (color_array)

      ! now create a group that contains all the proc that have the same 2 virtual starting points
      ! in this way it is possible to sum the common integrals needed for the full MP2 energy
      ! in mp_comm_split_direct the color is give by my_a_virtual_start and my_b_virtual_start
      CALL mp_comm_split_direct(para_env%group, comm_exchange, sub_sub_color)
      NULLIFY (para_env_exchange)
      CALL cp_para_env_create(para_env_exchange, comm_exchange)

      ! crate the proc maps
      ALLOCATE (proc_map(-para_env_exchange%num_pe:2*para_env_exchange%num_pe-1))
      DO i = 0, para_env_exchange%num_pe-1
         proc_map(i) = i
         proc_map(-i-1) = para_env_exchange%num_pe-i-1
         proc_map(para_env_exchange%num_pe+i) = i
      END DO

      ALLOCATE (sub_proc_map(-para_env_sub%num_pe:2*para_env_sub%num_pe-1))
      DO i = 0, para_env_sub%num_pe-1
         sub_proc_map(i) = i
         sub_proc_map(-i-1) = para_env_sub%num_pe-i-1
         sub_proc_map(para_env_sub%num_pe+i) = i
      END DO

      ! create an array containing the information for communication
      ALLOCATE (exchange_group_sizes(0:para_env_exchange%num_pe-1, 3))
      exchange_group_sizes = 0
      exchange_group_sizes(para_env_exchange%mepos, 1) = my_I_occupied_start
      exchange_group_sizes(para_env_exchange%mepos, 2) = my_I_occupied_end
      exchange_group_sizes(para_env_exchange%mepos, 3) = my_I_batch_size
      CALL mp_sum(exchange_group_sizes, para_env_exchange%group)
      mepos_in_EX_group = para_env_exchange%mepos
      size_of_exchange_group = para_env_exchange%num_pe

      NULLIFY (psi_a%pw)
      CALL pw_pool_create_pw(auxbas_pw_pool, psi_a%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      ALLOCATE (psi_i(my_I_occupied_start:my_I_occupied_end))
      DO i = my_I_occupied_start, my_I_occupied_end
         NULLIFY (psi_i(i)%pw)
         CALL pw_pool_create_pw(auxbas_pw_pool, psi_i(i)%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL calculate_wavefunction(mo_coeff, i, psi_i(i), rho_g, atomic_kind_set, &
                                     qs_kind_set, cell, dft_control, particle_set, &
                                     pw_env_sub, external_vector=my_Cocc(:, i-my_I_occupied_start+1))
      END DO

      Emp2 = 0.0_dp
      Emp2_Cou = 0.0_dp
      Emp2_EX = 0.0_dp
      IF (do_alpha_beta) Emp2_AB = 0.0_dp
      IF (calc_ex) THEN
         ALLOCATE (BIb_C(my_B_size, homo, my_I_batch_size))
      END IF

      CALL timeset(routineN//"_loop", handle2)
      DO a = homo+my_A_virtual_start, homo+my_A_virtual_end

         IF (calc_ex) BIb_C = 0.0_dp

         ! psi_a
         CALL calculate_wavefunction(mo_coeff, a, psi_a, rho_g, atomic_kind_set, &
                                     qs_kind_set, cell, dft_control, particle_set, &
                                     pw_env_sub, external_vector=my_Cvirt(:, a-(homo+my_A_virtual_start)+1))
         i_counter = 0
         DO i = my_I_occupied_start, my_I_occupied_end
            i_counter = i_counter+1

            ! potential
            CALL timeset(routineN//"_pot", handle3)
            rho_r%pw%cr3d = psi_i(i)%pw%cr3d*psi_a%pw%cr3d
            CALL pw_transfer(rho_r%pw, rho_g%pw)
            CALL pw_poisson_solve(poisson_env, rho_g%pw, pair_energy, pot_g%pw)
            CALL pw_transfer(pot_g%pw, rho_r%pw)
            CALL pw_scale(rho_r%pw, rho_r%pw%pw_grid%dvol)
            CALL timestop(handle3)

            ! and finally (ia|munu)
            CALL timeset(routineN//"_int", handle3)
            CALL cp_dbcsr_set(mat_munu%matrix, 0.0_dp)
            CALL integrate_v_rspace(rho_r, hmat=mat_munu, qs_env=qs_env, &
                                    calculate_forces=.FALSE., compute_tau=.FALSE., gapw=.FALSE., &
                                    pw_env_external=pw_env_sub, task_list_external=task_list_sub)
            CALL timestop(handle3)

            ! multiply and goooooooo ...
            CALL timeset(routineN//"_mult_o", handle3)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mat_munu%matrix, mo_coeff_o, &
                                   0.0_dp, matrix_ia_jnu, filter_eps=eps_filter)
            IF (do_alpha_beta) THEN
               ! transform orbitals using the beta coeff matrix
               CALL cp_dbcsr_multiply("N", "N", 1.0_dp, mat_munu%matrix, mo_coeff_o_beta, &
                                      0.0_dp, matrix_ia_jnu_beta, filter_eps=eps_filter)
            END IF
            CALL timestop(handle3)
            CALL timeset(routineN//"_mult_v", handle3)
            CALL cp_dbcsr_multiply("T", "N", 1.0_dp, matrix_ia_jnu, mo_coeff_v, &
                                   0.0_dp, matrix_ia_jb, filter_eps=eps_filter)
            IF (do_alpha_beta) THEN
               ! transform orbitals using the beta coeff matrix
               CALL cp_dbcsr_multiply("T", "N", 1.0_dp, matrix_ia_jnu_beta, mo_coeff_v_beta, &
                                      0.0_dp, matrix_ia_jb_beta, filter_eps=eps_filter)
            END IF
            CALL timestop(handle3)

            CALL timeset(routineN//"_E_Cou", handle3)
            CALL cp_dbcsr_iterator_start(iter, matrix_ia_jb)
            DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
               CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                                 row_size=row_size, col_size=col_size, &
                                                 row_offset=row_offset, col_offset=col_offset)
               DO b = 1, col_size
               DO j = 1, row_size
                  ! Compute the coulomb MP2 energy
                  Emp2_Cou = Emp2_Cou-2.0_dp*data_block(j, b)**2/ &
                             (Eigenval(a)+Eigenval(homo+col_offset+b-1)-Eigenval(i)-Eigenval(row_offset+j-1))
               ENDDO
               ENDDO
            ENDDO
            CALL cp_dbcsr_iterator_stop(iter)
            IF (do_alpha_beta) THEN
               ! Compute the coulomb only= SO = MP2 alpha-beta  MP2 energy component
               CALL cp_dbcsr_iterator_start(iter, matrix_ia_jb_beta)
               DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
                  CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                                    row_size=row_size, col_size=col_size, &
                                                    row_offset=row_offset, col_offset=col_offset)
                  DO b = 1, col_size
                  DO j = 1, row_size
                     ! Compute the coulomb MP2 energy alpha beta case
                     Emp2_AB = Emp2_AB-data_block(j, b)**2/ &
                               (Eigenval(a)+Eigenval_beta(homo_beta+col_offset+b-1)-Eigenval(i)-Eigenval_beta(row_offset+j-1))
                  ENDDO
                  ENDDO
               ENDDO
               CALL cp_dbcsr_iterator_stop(iter)
            END IF
            CALL timestop(handle3)

            ! now collect my local data from all the other members of the group
            ! b_start, b_end
            IF (calc_ex) THEN
               CALL timeset(routineN//"_E_Ex_1", handle3)
               CALL copy_dbcsr_to_fm(matrix_ia_jb, fm_BIb_jb)
               CALL grep_my_integrals(para_env_sub, fm_BIb_jb, BIb_C(1:my_B_size, 1:homo, i_counter), max_row_col_local, &
                                      sub_proc_map, local_col_row_info, &
                                      my_B_virtual_end, my_B_virtual_start)
               CALL timestop(handle3)
            END IF

         END DO

         IF (calc_ex) THEN
            CALL timeset(routineN//"_E_Ex_2", handle3)
            ! calculate the contribution to MP2 energy for my local data
            DO i = 1, my_I_batch_size
               DO j = my_I_occupied_start, my_I_occupied_end
                  DO b = 1, my_B_size
                     b_global = b-1+my_B_virtual_start
                     Emp2_EX = Emp2_EX+BIb_C(b, j, i)*BIb_C(b, i+my_I_occupied_start-1, j-my_I_occupied_start+1) &
                               /(Eigenval(a)+Eigenval(homo+b_global)-Eigenval(i+my_I_occupied_start-1)-Eigenval(j))
                  END DO
               END DO
            END DO

            ! start communicating and collecting exchange contributions from
            ! other processes in my exchange group
            DO index_proc_shift = 1, size_of_exchange_group-1
               proc_send = proc_map(mepos_in_EX_group+index_proc_shift)
               proc_receive = proc_map(mepos_in_EX_group-index_proc_shift)

               EX_start = exchange_group_sizes(proc_receive, 1)
               EX_end = exchange_group_sizes(proc_receive, 2)
               size_EX = exchange_group_sizes(proc_receive, 3)

               ALLOCATE (BIb_EX(my_B_size, my_I_batch_size, size_EX))
               BIb_EX = 0.0_dp

               EX_start_send = exchange_group_sizes(proc_send, 1)
               EX_end_send = exchange_group_sizes(proc_send, 2)
               size_EX_send = exchange_group_sizes(proc_send, 3)

               ALLOCATE (BIb_send(my_B_size, size_EX_send, my_I_batch_size))
               BIb_send(1:my_B_size, 1:size_EX_send, 1:my_I_batch_size) = &
                  BIb_C(1:my_B_size, EX_start_send:EX_end_send, 1:my_I_batch_size)

               ! send and receive the exchange array
               CALL mp_sendrecv(BIb_send, proc_send, BIb_EX, proc_receive, para_env_exchange%group)

               DO i = 1, my_I_batch_size
                  DO j = 1, size_EX
                     DO b = 1, my_B_size
                        b_global = b-1+my_B_virtual_start
                        Emp2_EX = Emp2_EX+BIb_C(b, j+EX_start-1, i)*BIb_EX(b, i, j) &
                                  /(Eigenval(a)+Eigenval(homo+b_global)-Eigenval(i+my_I_occupied_start-1)-Eigenval(j+EX_start-1))
                     END DO
                  END DO
               END DO

               DEALLOCATE (BIb_EX)
               DEALLOCATE (BIb_send)

            END DO
            CALL timestop(handle3)
         END IF

      ENDDO
      CALL timestop(handle2)

      CALL mp_sum(Emp2_Cou, para_env%group)
      CALL mp_sum(Emp2_EX, para_env%group)
      Emp2 = Emp2_Cou+Emp2_EX
      IF (do_alpha_beta) CALL mp_sum(Emp2_AB, para_env%group)

      DEALLOCATE (my_Cocc)
      DEALLOCATE (my_Cvirt)

      IF (calc_ex) THEN
         CALL cp_fm_release(fm_BIb_jb)
         DEALLOCATE (local_col_row_info)
         DEALLOCATE (BIb_C)
      END IF
      DEALLOCATE (proc_map)
      DEALLOCATE (sub_proc_map)
      DEALLOCATE (exchange_group_sizes)

      CALL cp_para_env_release(para_env_exchange)

      CALL cp_dbcsr_release(matrix_ia_jnu)
      CALL cp_dbcsr_release(matrix_ia_jb)
      IF (do_alpha_beta) THEN
         CALL cp_dbcsr_release(matrix_ia_jnu_beta)
         CALL cp_dbcsr_release(matrix_ia_jb_beta)
      END IF

      DO i = my_I_occupied_start, my_I_occupied_end
         CALL pw_release(psi_i(i)%pw)
      END DO
      DEALLOCATE (psi_i)

      CALL pw_pool_give_back_pw(auxbas_pw_pool, psi_a%pw)

      CALL timestop(handle)

   END SUBROUTINE mp2_gpw_compute

! **************************************************************************************************
!> \brief ...
!> \param wfn_size ...
!> \param p ...
!> \param q ...
!> \param num_w ...
!> \param nmo ...
!> \param virtual ...
!> \param homo ...
!> \param calc_ex ...
!> \param mem_try ...
! **************************************************************************************************
   SUBROUTINE estimate_memory_usage(wfn_size, p, q, num_w, nmo, virtual, homo, calc_ex, mem_try)
      REAL(KIND=dp)                                      :: wfn_size
      INTEGER                                            :: p, q, num_w, nmo, virtual, homo
      LOGICAL                                            :: calc_ex
      REAL(KIND=dp)                                      :: mem_try

      mem_try = 0.0_dp
      ! integrals
      mem_try = mem_try+virtual*REAL(homo, KIND=dp)**2/(p*num_w)
      ! array for the coefficient matrix and wave vectors
      mem_try = mem_try+REAL(homo, KIND=dp)*nmo/p+ &
                REAL(virtual, KIND=dp)*nmo/q+ &
                2.0_dp*MAX(REAL(homo, KIND=dp)*nmo/p, REAL(virtual, KIND=dp)*nmo/q)
      ! temporary array for MO integrals and MO integrals to be exchanged
      IF (calc_ex) THEN
         mem_try = mem_try+2.0_dp*MAX(virtual*REAL(homo, KIND=dp)*MIN(1, num_w-1)/num_w, &
                                      virtual*REAL(homo, KIND=dp)**2/(p*p*num_w))
      ELSE
         mem_try = mem_try+2.0_dp*virtual*REAL(homo, KIND=dp)
      END IF
      ! wfn
      mem_try = mem_try+((homo+p-1)/p)*wfn_size
      ! Mb
      mem_try = mem_try*8.0D+00/1024.0D+00**2

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param para_env_sub ...
!> \param fm_BIb_jb ...
!> \param BIb_jb ...
!> \param max_row_col_local ...
!> \param proc_map ...
!> \param local_col_row_info ...
!> \param my_B_virtual_end ...
!> \param my_B_virtual_start ...
! **************************************************************************************************
   SUBROUTINE grep_my_integrals(para_env_sub, fm_BIb_jb, BIb_jb, max_row_col_local, &
                                proc_map, local_col_row_info, &
                                my_B_virtual_end, my_B_virtual_start)
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      TYPE(cp_fm_type), POINTER                          :: fm_BIb_jb
      REAL(KIND=dp), DIMENSION(:, :)                     :: BIb_jb
      INTEGER                                            :: max_row_col_local
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: proc_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_col_row_info
      INTEGER                                            :: my_B_virtual_end, my_B_virtual_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'grep_my_integrals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_global, iiB, j_global, jjB, ncol_rec, &
                                                            nrow_rec, proc_receive, proc_send, &
                                                            proc_shift
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: rec_col_row_info
      INTEGER, DIMENSION(:), POINTER                     :: col_indices_rec, row_indices_rec
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_BI, rec_BI

      ALLOCATE (rec_col_row_info(0:max_row_col_local, 2))

      rec_col_row_info(:, :) = local_col_row_info

      nrow_rec = rec_col_row_info(0, 1)
      ncol_rec = rec_col_row_info(0, 2)

      ALLOCATE (row_indices_rec(nrow_rec))
      row_indices_rec = rec_col_row_info(1:nrow_rec, 1)

      ALLOCATE (col_indices_rec(ncol_rec))
      col_indices_rec = rec_col_row_info(1:ncol_rec, 2)

      ! accumulate data on BIb_jb buffer starting from myself
      DO jjB = 1, ncol_rec
         j_global = col_indices_rec(jjB)
         IF (j_global >= my_B_virtual_start .AND. j_global <= my_B_virtual_end) THEN
            DO iiB = 1, nrow_rec
               i_global = row_indices_rec(iiB)
               BIb_jb(j_global-my_B_virtual_start+1, i_global) = fm_BIb_jb%local_data(iiB, jjB)
            END DO
         END IF
      END DO

      DEALLOCATE (row_indices_rec)
      DEALLOCATE (col_indices_rec)

      IF (para_env_sub%num_pe > 1) THEN
         ALLOCATE (local_BI(nrow_rec, ncol_rec))
         local_BI(1:nrow_rec, 1:ncol_rec) = fm_BIb_jb%local_data(1:nrow_rec, 1:ncol_rec)

         DO proc_shift = 1, para_env_sub%num_pe-1
            proc_send = proc_map(para_env_sub%mepos+proc_shift)
            proc_receive = proc_map(para_env_sub%mepos-proc_shift)

            ! first exchange information on the local data
            rec_col_row_info = 0
            CALL mp_sendrecv(local_col_row_info, proc_send, rec_col_row_info, proc_receive, para_env_sub%group)
            nrow_rec = rec_col_row_info(0, 1)
            ncol_rec = rec_col_row_info(0, 2)

            ALLOCATE (row_indices_rec(nrow_rec))
            row_indices_rec = rec_col_row_info(1:nrow_rec, 1)

            ALLOCATE (col_indices_rec(ncol_rec))
            col_indices_rec = rec_col_row_info(1:ncol_rec, 2)

            ALLOCATE (rec_BI(nrow_rec, ncol_rec))
            rec_BI = 0.0_dp

            ! then send and receive the real data
            CALL mp_sendrecv(local_BI, proc_send, rec_BI, proc_receive, para_env_sub%group)

            ! accumulate the received data on BIb_jb buffer
            DO jjB = 1, ncol_rec
               j_global = col_indices_rec(jjB)
               IF (j_global >= my_B_virtual_start .AND. j_global <= my_B_virtual_end) THEN
                  DO iiB = 1, nrow_rec
                     i_global = row_indices_rec(iiB)
                     BIb_jb(j_global-my_B_virtual_start+1, i_global) = rec_BI(iiB, jjB)
                  END DO
               END IF
            END DO

            DEALLOCATE (col_indices_rec)
            DEALLOCATE (row_indices_rec)
            DEALLOCATE (rec_BI)
         END DO

         DEALLOCATE (local_BI)
      END IF

      DEALLOCATE (rec_col_row_info)

   END SUBROUTINE grep_my_integrals

! **************************************************************************************************
!> \brief ...
!> \param para_env_sub ...
!> \param dimen ...
!> \param my_I_occupied_start ...
!> \param my_I_occupied_end ...
!> \param my_I_batch_size ...
!> \param my_A_virtual_start ...
!> \param my_A_virtual_end ...
!> \param my_A_batch_size ...
!> \param mo_coeff_o ...
!> \param mo_coeff_v ...
!> \param my_Cocc ...
!> \param my_Cvirt ...
! **************************************************************************************************
   SUBROUTINE grep_occ_virt_wavefunc(para_env_sub, dimen, &
                                     my_I_occupied_start, my_I_occupied_end, my_I_batch_size, &
                                     my_A_virtual_start, my_A_virtual_end, my_A_batch_size, &
                                     mo_coeff_o, mo_coeff_v, my_Cocc, my_Cvirt)

      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      INTEGER :: dimen, my_I_occupied_start, my_I_occupied_end, my_I_batch_size, &
         my_A_virtual_start, my_A_virtual_end, my_A_batch_size
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_o, mo_coeff_v
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: my_Cocc, my_Cvirt

      CHARACTER(LEN=*), PARAMETER :: routineN = 'grep_occ_virt_wavefunc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, col_offset, col_size, handle, &
                                                            i, i_global, j, j_global, row, &
                                                            row_offset, row_size
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)

      ALLOCATE (my_Cocc(dimen, my_I_batch_size))
      my_Cocc = 0.0_dp

      ALLOCATE (my_Cvirt(dimen, my_A_batch_size))
      my_Cvirt = 0.0_dp

      ! accumulate data from mo_coeff_o into Cocc
      CALL cp_dbcsr_iterator_start(iter, mo_coeff_o)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                           row_size=row_size, col_size=col_size, &
                                           row_offset=row_offset, col_offset=col_offset)
         DO j = 1, col_size
            j_global = col_offset+j-1
            IF (j_global >= my_I_occupied_start .AND. j_global <= my_I_occupied_end) THEN
               DO i = 1, row_size
                  i_global = row_offset+i-1
                  my_Cocc(i_global, j_global-my_I_occupied_start+1) = data_block(i, j)
               END DO
            END IF
         END DO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL mp_sum(my_Cocc, para_env_sub%group)

      ! accumulate data from mo_coeff_o into Cocc
      CALL cp_dbcsr_iterator_start(iter, mo_coeff_v)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                           row_size=row_size, col_size=col_size, &
                                           row_offset=row_offset, col_offset=col_offset)
         DO j = 1, col_size
            j_global = col_offset+j-1
            IF (j_global >= my_A_virtual_start .AND. j_global <= my_A_virtual_end) THEN
               DO i = 1, row_size
                  i_global = row_offset+i-1
                  my_Cvirt(i_global, j_global-my_A_virtual_start+1) = data_block(i, j)
               END DO
            END IF
         END DO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL mp_sum(my_Cvirt, para_env_sub%group)

      CALL timestop(handle)

   END SUBROUTINE grep_occ_virt_wavefunc

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
!> \param para_env ...
!> \param para_env_sub ...
!> \param mo_coeff ...
!> \param dimen ...
!> \param homo ...
!> \param mat_munu ...
!> \param mo_coeff_o ...
!> \param mo_coeff_v ...
!> \param mo_coeff_all ...
!> \param mo_coeff_gw ...
!> \param my_do_gw ...
!> \param gw_corr_lev_occ ...
!> \param gw_corr_lev_virt ...
!> \param only_mo_coeff_all ...
! **************************************************************************************************
   SUBROUTINE replicate_mat_to_subgroup(mp2_env, para_env, para_env_sub, mo_coeff, dimen, homo, mat_munu, &
                                        mo_coeff_o, mo_coeff_v, mo_coeff_all, mo_coeff_gw, my_do_gw, &
                                        gw_corr_lev_occ, gw_corr_lev_virt, only_mo_coeff_all)
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env, para_env_sub
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      INTEGER                                            :: dimen, homo
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_o, mo_coeff_v, mo_coeff_all, &
                                                            mo_coeff_gw
      LOGICAL                                            :: my_do_gw
      INTEGER                                            :: gw_corr_lev_occ, gw_corr_lev_virt
      LOGICAL, OPTIONAL                                  :: only_mo_coeff_all

      CHARACTER(LEN=*), PARAMETER :: routineN = 'replicate_mat_to_subgroup', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk, col, col_offset, col_size, handle, i_global, iiB, iproc, itmp(2), j_global, &
         jjB, max_row_col_local, my_mu_end, my_mu_size, my_mu_start, ncol_local, ncol_rec, &
         nrow_local, nrow_rec, proc_receive, proc_receive_static, proc_send, proc_send_static, &
         proc_shift, row, row_offset, row_size, virtual
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_array, proc_map, sizes_array, &
                                                            starts_array
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: local_col_row_info, rec_col_row_info
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, col_indices_rec, &
                                                            row_indices, row_indices_rec
      LOGICAL                                            :: my_only_mo_coeff_all
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: C, Cgw, Cocc, Cvirt, rec_C
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_block, local_C, local_C_internal
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_coeff

      my_only_mo_coeff_all = .FALSE.
      IF (PRESENT(only_mo_coeff_all)) my_only_mo_coeff_all = only_mo_coeff_all

      CALL timeset(routineN, handle)

      ALLOCATE (sizes_array(0:para_env_sub%num_pe-1))
      ALLOCATE (starts_array(0:para_env_sub%num_pe-1))
      starts_array = 0
      ALLOCATE (ends_array(0:para_env_sub%num_pe-1))
      ends_array = 0

      DO iproc = 0, para_env_sub%num_pe-1
         itmp = get_limit(dimen, para_env_sub%num_pe, iproc)
         starts_array(iproc) = itmp(1)
         ends_array(iproc) = itmp(2)
         sizes_array(iproc) = itmp(2)-itmp(1)+1
      ENDDO

      my_mu_size = sizes_array(para_env_sub%mepos)
      my_mu_start = starts_array(para_env_sub%mepos)
      my_mu_end = ends_array(para_env_sub%mepos)

      ! local storage for the C matrix
      ALLOCATE (C(my_mu_size, dimen))
      C = 0.0_dp

      ! proc_map, vector that replicate the processor numbers also
      ! for negative and positive number > num_pe
      ! needed to know which is the processor, to respect to another one,
      ! for a given shift
      ALLOCATE (proc_map(-para_env%num_pe:2*para_env%num_pe-1))
      DO iiB = 0, para_env%num_pe-1
         proc_map(iiB) = iiB
         proc_map(-iiB-1) = para_env%num_pe-iiB-1
         proc_map(para_env%num_pe+iiB) = iiB
      END DO

      CALL cp_fm_get_info(matrix=mo_coeff, &
                          matrix_struct=fm_struct_coeff, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          local_data=local_C_internal)

      ALLOCATE (local_C(nrow_local, ncol_local))
      local_C = local_C_internal(1:nrow_local, 1:ncol_local)
      NULLIFY (local_C_internal)

      max_row_col_local = MAX(nrow_local, ncol_local)
      CALL mp_max(max_row_col_local, para_env%group)

      ALLOCATE (local_col_row_info(0:max_row_col_local, 2))
      local_col_row_info = 0
      ! 0,1 nrows
      local_col_row_info(0, 1) = nrow_local
      local_col_row_info(1:nrow_local, 1) = row_indices(1:nrow_local)
      ! 0,2 ncols
      local_col_row_info(0, 2) = ncol_local
      local_col_row_info(1:ncol_local, 2) = col_indices(1:ncol_local)

      ALLOCATE (rec_col_row_info(0:max_row_col_local, 2))

      ! accumulate data on C buffer starting from myself
      DO iiB = 1, nrow_local
         i_global = row_indices(iiB)
         IF (i_global >= my_mu_start .AND. i_global <= my_mu_end) THEN
            DO jjB = 1, ncol_local
               j_global = col_indices(jjB)
               C(i_global-my_mu_start+1, j_global) = local_C(iiB, jjB)
            END DO
         END IF
      END DO

      ! start ring communication for collecting the data from the other
      proc_send_static = proc_map(para_env%mepos+1)
      proc_receive_static = proc_map(para_env%mepos-1)
      DO proc_shift = 1, para_env%num_pe-1
         proc_send = proc_map(para_env%mepos+proc_shift)
         proc_receive = proc_map(para_env%mepos-proc_shift)

         ! first exchange information on the local data
         rec_col_row_info = 0
         CALL mp_sendrecv(local_col_row_info, proc_send_static, rec_col_row_info, proc_receive_static, para_env%group)
         nrow_rec = rec_col_row_info(0, 1)
         ncol_rec = rec_col_row_info(0, 2)

         ALLOCATE (row_indices_rec(nrow_rec))
         row_indices_rec = rec_col_row_info(1:nrow_rec, 1)

         ALLOCATE (col_indices_rec(ncol_rec))
         col_indices_rec = rec_col_row_info(1:ncol_rec, 2)

         ALLOCATE (rec_C(nrow_rec, ncol_rec))
         rec_C = 0.0_dp

         ! then send and receive the real data
         CALL mp_sendrecv(local_C, proc_send_static, rec_C, proc_receive_static, para_env%group)

         ! accumulate the received data on C buffer
         DO iiB = 1, nrow_rec
            i_global = row_indices_rec(iiB)
            IF (i_global >= my_mu_start .AND. i_global <= my_mu_end) THEN
               DO jjB = 1, ncol_rec
                  j_global = col_indices_rec(jjB)
                  C(i_global-my_mu_start+1, j_global) = rec_C(iiB, jjB)
               END DO
            END IF
         END DO

         local_col_row_info(:, :) = rec_col_row_info
         DEALLOCATE (local_C)
         ALLOCATE (local_C(nrow_rec, ncol_rec))
         local_C = rec_C

         DEALLOCATE (col_indices_rec)
         DEALLOCATE (row_indices_rec)
         DEALLOCATE (rec_C)
      END DO

      DEALLOCATE (local_col_row_info)
      DEALLOCATE (rec_col_row_info)
      DEALLOCATE (proc_map)

      ! proc_map, for the sub_group
      ALLOCATE (proc_map(-para_env_sub%num_pe:2*para_env_sub%num_pe-1))
      DO iiB = 0, para_env_sub%num_pe-1
         proc_map(iiB) = iiB
         proc_map(-iiB-1) = para_env_sub%num_pe-iiB-1
         proc_map(para_env_sub%num_pe+iiB) = iiB
      END DO

      ! split the C matrix into occupied and virtual
      ALLOCATE (Cocc(my_mu_size, homo))
      Cocc(1:my_mu_size, 1:homo) = C(1:my_mu_size, 1:homo)

      virtual = dimen-homo
      ALLOCATE (Cvirt(my_mu_size, virtual))
      Cvirt(1:my_mu_size, 1:virtual) = C(1:my_mu_size, homo+1:dimen)

      IF (.NOT. my_only_mo_coeff_all) THEN
         ! create and fill mo_coeff_o, mo_coeff_v and mo_coeff_all
         CALL build_mo_coeff_v_o_all(mp2_env, para_env_sub, mo_coeff_o, Cocc, &
                                     homo, blk, row, col, row_size, col_size, row_offset, &
                                     col_offset, i_global, j_global, my_mu_start, my_mu_end, &
                                     mat_munu, iter, &
                                     data_block, ends_array, proc_map, &
                                     sizes_array, starts_array)

         CALL build_mo_coeff_v_o_all(mp2_env, para_env_sub, mo_coeff_v, Cvirt, &
                                     virtual, blk, row, col, row_size, col_size, row_offset, &
                                     col_offset, i_global, j_global, my_mu_start, my_mu_end, &
                                     mat_munu, iter, &
                                     data_block, ends_array, proc_map, &
                                     sizes_array, starts_array)
      ELSE
         DEALLOCATE (Cocc, Cvirt)
      END IF

      IF (my_do_gw .OR. my_only_mo_coeff_all) THEN

         IF (my_do_gw) THEN
            ! also cut levels homo-gw_corr_lev_occ+1, ..., lumo+gw_corr_lev_virt-1 of C
            ALLOCATE (Cgw(my_mu_size, gw_corr_lev_occ+gw_corr_lev_virt))
            Cgw(1:my_mu_size, 1:(gw_corr_lev_occ+gw_corr_lev_virt)) = &
               C(1:my_mu_size, homo-gw_corr_lev_occ+1:homo+gw_corr_lev_virt)
            CALL build_mo_coeff_v_o_all(mp2_env, para_env_sub, mo_coeff_gw, Cgw, &
                                        gw_corr_lev_occ+gw_corr_lev_virt, &
                                        blk, row, col, row_size, col_size, row_offset, &
                                        col_offset, i_global, j_global, my_mu_start, my_mu_end, &
                                        mat_munu, iter, &
                                        data_block, ends_array, proc_map, &
                                        sizes_array, starts_array)
         END IF

         ! all levels
         CALL build_mo_coeff_v_o_all(mp2_env, para_env_sub, mo_coeff_all, C, &
                                     dimen, blk, row, col, row_size, col_size, row_offset, &
                                     col_offset, i_global, j_global, my_mu_start, my_mu_end, &
                                     mat_munu, iter, &
                                     data_block, ends_array, proc_map, &
                                     sizes_array, starts_array)

      ELSE
         DEALLOCATE (C)
      END IF

      DEALLOCATE (proc_map)
      DEALLOCATE (sizes_array)
      DEALLOCATE (starts_array)
      DEALLOCATE (ends_array)
      DEALLOCATE (local_C)

      CALL timestop(handle)

   END SUBROUTINE replicate_mat_to_subgroup

! **************************************************************************************************
!> \brief Encapsulate the building of dbcsr_matrices mo_coeff_(v,o,all)
!> \param mp2_env ...
!> \param para_env_sub ...
!> \param mo_coeff_to_build ...
!> \param Cread ...
!> \param number_of_level ...
!> \param blk ...
!> \param row ...
!> \param col ...
!> \param row_size ...
!> \param col_size ...
!> \param row_offset ...
!> \param col_offset ...
!> \param i_global ...
!> \param j_global ...
!> \param my_mu_start ...
!> \param my_mu_end ...
!> \param mat_munu ...
!> \param iter ...
!> \param data_block ...
!> \param ends_array ...
!> \param proc_map ...
!> \param sizes_array ...
!> \param starts_array ...
!> \author Jan Wilhelm, Code by Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE build_mo_coeff_v_o_all(mp2_env, para_env_sub, mo_coeff_to_build, Cread, &
                                     number_of_level, blk, row, col, row_size, col_size, row_offset, &
                                     col_offset, i_global, j_global, my_mu_start, my_mu_end, &
                                     mat_munu, iter, &
                                     data_block, ends_array, proc_map, &
                                     sizes_array, starts_array)
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_to_build
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Cread
      INTEGER                                            :: number_of_level, blk, row, col, &
                                                            row_size, col_size, row_offset, &
                                                            col_offset, i_global, j_global, &
                                                            my_mu_start, my_mu_end
      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(cp_dbcsr_iterator)                            :: iter
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_block
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ends_array, proc_map, sizes_array, &
                                                            starts_array

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_mo_coeff_v_o_all', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, proc_receive, proc_send, &
                                                            proc_shift, rec_mu_end, rec_mu_size, &
                                                            rec_mu_start
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rec_C

      CALL timeset(routineN, handle)

      NULLIFY (mo_coeff_to_build)
      CALL cp_dbcsr_init_p(mo_coeff_to_build)
      CALL cp_dbcsr_m_by_n_from_row_template(mo_coeff_to_build, template=mat_munu%matrix, n=number_of_level, &
                                             sym=dbcsr_type_no_symmetry, data_type=dbcsr_type_real_default)
      CALL cp_dbcsr_reserve_all_blocks(mo_coeff_to_build)

      ! accumulate data on mo_coeff_to_build starting from myself
      CALL cp_dbcsr_iterator_start(iter, mo_coeff_to_build)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                           row_size=row_size, col_size=col_size, &
                                           row_offset=row_offset, col_offset=col_offset)
         DO i = 1, row_size
            i_global = row_offset+i-1
            IF (i_global >= my_mu_start .AND. i_global <= my_mu_end) THEN
               DO j = 1, col_size
                  j_global = col_offset+j-1
                  data_block(i, j) = Cread(i_global-my_mu_start+1, col_offset+j-1)
               ENDDO
            END IF
         ENDDO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      ! start ring communication in the subgroup for collecting the data from the other
      ! proc (occupied)
      DO proc_shift = 1, para_env_sub%num_pe-1
         proc_send = proc_map(para_env_sub%mepos+proc_shift)
         proc_receive = proc_map(para_env_sub%mepos-proc_shift)

         rec_mu_start = starts_array(proc_receive)
         rec_mu_end = ends_array(proc_receive)
         rec_mu_size = sizes_array(proc_receive)

         ALLOCATE (rec_C(rec_mu_size, number_of_level))
         rec_C = 0.0_dp

         ! then send and receive the real data
         CALL mp_sendrecv(Cread, proc_send, rec_C, proc_receive, para_env_sub%group)

         ! accumulate data on mo_coeff_to_build the data received from proc_rec
         CALL cp_dbcsr_iterator_start(iter, mo_coeff_to_build)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_block, blk, &
                                              row_size=row_size, col_size=col_size, &
                                              row_offset=row_offset, col_offset=col_offset)
            DO i = 1, row_size
               i_global = row_offset+i-1
               IF (i_global >= rec_mu_start .AND. i_global <= rec_mu_end) THEN
                  DO j = 1, col_size
                     j_global = col_offset+j-1
                     data_block(i, j) = rec_C(i_global-rec_mu_start+1, col_offset+j-1)
                  ENDDO
               END IF
            ENDDO
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)

         DEALLOCATE (rec_C)

      END DO
      CALL cp_dbcsr_filter(mo_coeff_to_build, mp2_env%mp2_gpw%eps_filter)

      DEALLOCATE (Cread)

      CALL timestop(handle)

   END SUBROUTINE build_mo_coeff_v_o_all

! **************************************************************************************************
!> \brief Encapsulate the building of dbcsr_matrix mat_munu
!> \param mat_munu ...
!> \param qs_env ...
!> \param mp2_env ...
!> \param dft_control ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param atom2d ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param sab_orb_sub ...
!> \param particle_set ...
!> \param cell ...
!> \param blacs_env_sub ...
!> \param eps_pgf_orb_old ...
!> \param eps_rho_rspace_old ...
!> \param eps_gvg_rspace_old ...
!> \param unit_nr ...
!> \author Jan Wilhelm, code by Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE create_mat_munu(mat_munu, qs_env, mp2_env, dft_control, atomic_kind_set, qs_kind_set, &
                              atom2d, molecule_kind_set, &
                              molecule_set, sab_orb_sub, particle_set, cell, blacs_env_sub, &
                              eps_pgf_orb_old, eps_rho_rspace_old, eps_gvg_rspace_old, &
                              unit_nr)

      TYPE(cp_dbcsr_p_type)                              :: mat_munu
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(local_atoms_type), ALLOCATABLE, DIMENSION(:)  :: atom2d
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb_sub
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env_sub
      REAL(kind=dp)                                      :: eps_pgf_orb_old, eps_rho_rspace_old, &
                                                            eps_gvg_rspace_old
      INTEGER                                            :: unit_nr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_mat_munu', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, natom, nkind
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_sizes
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: orb_present
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: orb_radius
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: pair_radius
      REAL(kind=dp)                                      :: subcells
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist_sub
      TYPE(distribution_1d_type), POINTER                :: local_molecules_sub, local_particles_sub
      TYPE(distribution_2d_type), POINTER                :: distribution_2d_sub
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set

      CALL timeset(routineN, handle)

      ! re-init the radii to be able to generate pair lists with MP2-appropriate screening
      eps_pgf_orb_old = dft_control%qs_control%eps_pgf_orb
      eps_rho_rspace_old = dft_control%qs_control%eps_rho_rspace
      eps_gvg_rspace_old = dft_control%qs_control%eps_gvg_rspace

      ! hack hack hack XXXXXXXXXXXXXXX ... to be fixed
      dft_control%qs_control%eps_pgf_orb = mp2_env%mp2_gpw%eps_grid
      dft_control%qs_control%eps_rho_rspace = mp2_env%mp2_gpw%eps_grid
      dft_control%qs_control%eps_gvg_rspace = mp2_env%mp2_gpw%eps_grid
      CALL init_interaction_radii(dft_control%qs_control, atomic_kind_set, qs_kind_set)

      ! get a distribution_1d
      NULLIFY (local_particles_sub, local_molecules_sub)
      CALL distribute_molecules_1d(atomic_kind_set=atomic_kind_set, &
                                   particle_set=particle_set, &
                                   local_particles=local_particles_sub, &
                                   molecule_kind_set=molecule_kind_set, &
                                   molecule_set=molecule_set, &
                                   local_molecules=local_molecules_sub, &
                                   force_env_section=qs_env%input)

      ! get a distribution_2d
      NULLIFY (distribution_2d_sub)
      CALL distribute_molecules_2d(cell=cell, &
                                   atomic_kind_set=atomic_kind_set, &
                                   qs_kind_set=qs_kind_set, &
                                   particle_set=particle_set, &
                                   molecule_kind_set=molecule_kind_set, &
                                   molecule_set=molecule_set, &
                                   distribution_2d=distribution_2d_sub, &
                                   blacs_env=blacs_env_sub, &
                                   force_env_section=qs_env%input)

      ! Build the sub  orbital-orbital overlap neighbor lists
      NULLIFY (sab_orb_sub)
      CALL section_vals_val_get(qs_env%input, "DFT%SUBCELLS", r_val=subcells)
      nkind = SIZE(atomic_kind_set)
      ALLOCATE (atom2d(nkind))

      CALL atom2d_build(atom2d, local_particles_sub, distribution_2d_sub, atomic_kind_set, &
                        molecule_set, molecule_only=.FALSE., particle_set=particle_set)

      ALLOCATE (orb_present(nkind))
      ALLOCATE (orb_radius(nkind))
      ALLOCATE (pair_radius(nkind, nkind))

      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
         IF (ASSOCIATED(orb_basis_set)) THEN
            orb_present(ikind) = .TRUE.
            CALL get_gto_basis_set(gto_basis_set=orb_basis_set, kind_radius=orb_radius(ikind))
         ELSE
            orb_present(ikind) = .FALSE.
            orb_radius(ikind) = 0.0_dp
         ENDIF
      END DO

      CALL pair_radius_setup(orb_present, orb_present, orb_radius, orb_radius, pair_radius)

      CALL build_neighbor_lists(sab_orb_sub, particle_set, atom2d, cell, pair_radius, &
                                mic=.FALSE., subcells=subcells, molecular=.FALSE., name="sab_orb_sub")
      CALL atom2d_cleanup(atom2d)
      DEALLOCATE (atom2d)
      DEALLOCATE (orb_present, orb_radius, pair_radius)

      ! a dbcsr_dist
      ALLOCATE (dbcsr_dist_sub)
      CALL cp_dbcsr_dist2d_to_dist(distribution_2d_sub, dbcsr_dist_sub, unit_nr)

      ! build a dbcsr matrix the hard way
      natom = SIZE(particle_set)
      ALLOCATE (row_blk_sizes(natom))
      CALL get_particle_set(particle_set, qs_kind_set, nsgf=row_blk_sizes)
      NULLIFY (mat_munu%matrix)
      ALLOCATE (mat_munu%matrix)
      CALL cp_dbcsr_init(mat_munu%matrix)
      CALL cp_dbcsr_create(matrix=mat_munu%matrix, &
                           name="(ai|munu)", &
                           dist=dbcsr_dist_sub, matrix_type=dbcsr_type_symmetric, &
                           row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                           nze=0)
      CALL cp_dbcsr_alloc_block_from_nbl(mat_munu%matrix, sab_orb_sub)

      DEALLOCATE (row_blk_sizes)

      CALL cp_dbcsr_distribution_release(dbcsr_dist_sub)
      DEALLOCATE (dbcsr_dist_sub)

      CALL distribution_2d_release(distribution_2d_sub)

      CALL distribution_1d_release(local_particles_sub)
      CALL distribution_1d_release(local_molecules_sub)

      CALL timestop(handle)

   END SUBROUTINE create_mat_munu

END MODULE mp2_gpw
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Rountines to calculate MP2 energy
!> \par History
!>      05.2011 created [Mauro Del Ben]
!> \author Mauro Del Ben
! **************************************************************************************************
MODULE mp2
   USE admm_methods,                    ONLY: admm_mo_merge_ks_matrix
   USE admm_types,                      ONLY: admm_type
   USE admm_utils,                      ONLY: admm_correct_for_eigenvalues,&
                                              admm_uncorrect_for_eigenvalues
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE bibliography,                    ONLY: DelBen2012,&
                                              DelBen2015b,&
                                              cite_reference
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_desymmetrize, cp_dbcsr_get_diag, &
        cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_init_p, cp_dbcsr_multiply, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_release_p, cp_dbcsr_set, cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr
   USE cp_fm_basic_linalg,              ONLY: cp_fm_triangular_invert
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE hfx_energy_potential,            ONLY: integrate_four_center
   USE hfx_types,                       ONLY: &
        alloc_containers, dealloc_containers, hfx_basis_info_type, hfx_basis_type, &
        hfx_container_type, hfx_create_basis_types, hfx_init_container, hfx_release_basis_types, &
        hfx_type
   USE input_constants,                 ONLY: cholesky_inverse,&
                                              do_mp2_potential_TShPSC,&
                                              hfx_do_eval_energy,&
                                              xc_none
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE machine,                         ONLY: m_flush,&
                                              m_memory,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_comm_split_direct,&
                                              mp_max,&
                                              mp_sum,&
                                              mp_sync
   USE mp2_direct_method,               ONLY: mp2_canonical_direct_single_batch
   USE mp2_gpw,                         ONLY: mp2_gpw_main
   USE mp2_optimize_ri_basis,           ONLY: optimize_ri_basis_main
   USE mp2_types,                       ONLY: &
        mp2_biel_type, mp2_method_direct, mp2_method_gpw, mp2_method_laplace, &
        mp2_ri_optimize_basis, mp2_type, ri_mp2_laplace, ri_mp2_method_gpw, ri_rpa_method_gpw
   USE particle_types,                  ONLY: particle_type
   USE qs_energy_types,                 ONLY: allocate_qs_energy,&
                                              deallocate_qs_energy,&
                                              qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_build_kohn_sham_matrix
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_matrix_pools,                 ONLY: mpools_create,&
                                              mpools_rebuild_fm_pools,&
                                              mpools_release,&
                                              qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: allocate_mo_set,&
                                              deallocate_mo_set,&
                                              get_mo_set,&
                                              init_mo_set,&
                                              mo_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_methods,                  ONLY: eigensolver
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE virial_types,                    ONLY: virial_type

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2'

   PUBLIC :: mp2_main

CONTAINS

! **************************************************************************************************
!> \brief the main entry point for MP2 calculations
!> \param qs_env ...
!> \param calc_forces ...
!> \author Mauro Del Ben
! **************************************************************************************************
   SUBROUTINE mp2_main(qs_env, calc_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calc_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'mp2_main', routineP = moduleN//':'//routineN

      INTEGER :: bin, cholesky_method, dimen, handle, handle2, i, i_thread, iatom, ikind, irep, &
         ispin, max_nset, my_bin_size, n_rep_hf, n_threads, nao, natom, ncol_block, nelec_alpha, &
         nelec_beta, nelectron, nfullcols_total, nfullrows_total, nkind, nrow_block, nspins, &
         unit_nr
      INTEGER(KIND=int_8)                                :: mem
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of
      LOGICAL                                            :: calc_ex, do_admm, &
                                                            do_dynamic_load_balancing, do_exx, &
                                                            do_gw, free_hfx_buffer
      REAL(KIND=dp) :: E_ex_from_GW, Emp2, Emp2_AA, Emp2_AA_Cou, Emp2_AA_ex, Emp2_AB, Emp2_AB_Cou, &
         Emp2_AB_ex, Emp2_BB, Emp2_BB_Cou, Emp2_BB_ex, Emp2_Cou, Emp2_ex, Emp2_S, Emp2_T, maxocc, &
         mem_real, t1, t2, t3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: Auto, Auto_alpha, Auto_beta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: C, C_alpha, C_beta
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_ks_aux, matrix_s
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_fm_type), POINTER                          :: fm_matrix_ks, fm_matrix_s, &
                                                            fm_matrix_work, mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(hfx_basis_info_type)                          :: basis_info
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_parameter
      TYPE(hfx_container_type), DIMENSION(:), POINTER    :: integral_containers
      TYPE(hfx_container_type), POINTER                  :: maxval_container
      TYPE(hfx_type), POINTER                            :: actual_x_data
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_mp2
      TYPE(mp2_biel_type)                                :: mp2_biel
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_matrix_pools_type), POINTER                :: my_mpools
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(section_vals_type), POINTER                   :: hfx_sections, input
      TYPE(virial_type), POINTER                         :: virial

! If SCF has not converged we should abort MP2 calculation
!JW 2del    IF (qs_env%mp2_env%hf_fail) THEN
!       CALL cp_abort(cp__l("mp2.F",172),"SCF not converged: "//&
!                      "not possible to run MP2")
!    ENDIF

      NULLIFY (virial, dft_control, blacs_env)
      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      CALL cite_reference(DelBen2012)

      CALL get_qs_env(qs_env, &
                      input=input, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      para_env=para_env, &
                      blacs_env=blacs_env, &
                      energy=energy, &
                      rho=rho, &
                      mos=mos, &
                      scf_env=scf_env, virial=virial, &
                      matrix_ks=matrix_ks, &
                      matrix_s=matrix_s, &
                      matrix_ks_aux_fit=matrix_ks_aux, &
                      mp2_env=mp2_env, &
                      admm_env=admm_env)

      unit_nr = cp_print_key_unit_nr(logger, input, "DFT%XC%WF_CORRELATION%MP2_INFO", &
                                     extension=".mp2Log")

      IF (unit_nr > 0) THEN
         IF (mp2_env%method .NE. ri_rpa_method_gpw) THEN
            WRITE (unit_nr, *)
            WRITE (unit_nr, *)
            WRITE (unit_nr, '(T2,A)') 'MP2 section'
            WRITE (unit_nr, '(T2,A)') '-----------'
            WRITE (unit_nr, *)
         ELSE
            WRITE (unit_nr, *)
            WRITE (unit_nr, *)
            WRITE (unit_nr, '(T2,A)') 'RI-RPA section'
            WRITE (unit_nr, '(T2,A)') '--------------'
            WRITE (unit_nr, *)
         END IF
      ENDIF

      IF (calc_forces) THEN
         CALL cite_reference(DelBen2015b)
         IF (mp2_env%method /= ri_mp2_method_gpw) THEN
            CALL cp__b("mp2.F",222,"No forces/gradients for the selected method.")
         END IF
      ENDIF

      IF (mp2_env%mp2_num_proc <= 0 .OR. mp2_env%mp2_num_proc > para_env%num_pe .OR. &
          MOD(para_env%num_pe, mp2_env%mp2_num_proc) .NE. 0) THEN
         IF (unit_nr > 0 .AND. mp2_env%mp2_num_proc .NE. -1) &
            WRITE (unit_nr, '(T3,A,T76,I5)') 'Requested number of processes per group:', mp2_env%mp2_num_proc
         mp2_env%mp2_num_proc = para_env%num_pe
      ENDIF
      IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T76,I5)') 'Used number of processes per group:', mp2_env%mp2_num_proc
      IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T68,F9.2,A4)') 'Maximum allowed memory usage per MPI processes:', &
         mp2_env%mp2_memory, ' MiB'

      IF ((mp2_env%method .NE. mp2_method_gpw) .AND. &
          (mp2_env%method .NE. ri_mp2_method_gpw) .AND. &
          (mp2_env%method .NE. ri_rpa_method_gpw) .AND. &
          (mp2_env%method .NE. ri_mp2_laplace)) THEN
         CALL m_memory(mem)
         mem_real = (mem+1024*1024-1)/(1024*1024)
         CALL mp_max(mem_real, para_env%group)
         mp2_env%mp2_memory = mp2_env%mp2_memory-mem_real
         IF (mp2_env%mp2_memory < 0.0_dp) mp2_env%mp2_memory = 1.0_dp

         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T68,F9.2,A4)') 'Available memory per MPI processes for MP2:', &
            mp2_env%mp2_memory, ' MiB'
      END IF

      IF (unit_nr > 0) CALL m_flush(unit_nr)

      nspins = dft_control%nspins
      natom = SIZE(particle_set, 1)

      ALLOCATE (kind_of(natom))

      CALL get_atomic_kind_set(atomic_kind_set, kind_of=kind_of)
      nkind = SIZE(atomic_kind_set, 1)

      CALL hfx_create_basis_types(basis_parameter, basis_info, qs_kind_set, do_admm=.FALSE.)

      dimen = 0
      max_nset = 0
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         dimen = dimen+SUM(basis_parameter(ikind)%nsgf)
         max_nset = MAX(max_nset, basis_parameter(ikind)%nset)
      END DO

      CALL get_mo_set(mo_set=mos(1)%mo_set, nao=nao)
      IF(.NOT.(dimen == nao))CALL cp__a("mp2.F",271)

      ! diagonalize the KS matrix in order to have the full set of MO's
      ! get S and KS matrices in fm_type (create also a working array)
      NULLIFY (fm_matrix_s)
      NULLIFY (fm_matrix_ks)
      NULLIFY (fm_matrix_work)
      NULLIFY (fm_struct)
      CALL cp_dbcsr_get_info(matrix_s(1)%matrix, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullcols_total, para_env=para_env)
      CALL cp_fm_create(fm_matrix_s, fm_struct, name="fm_matrix_s")
      CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, fm_matrix_s)

      CALL cp_fm_create(fm_matrix_ks, fm_struct, name="fm_matrix_ks")

      CALL cp_fm_create(fm_matrix_work, fm_struct, name="fm_matrix_work")
      CALL cp_fm_set_all(matrix=fm_matrix_work, alpha=0.0_dp)

      CALL cp_fm_struct_release(fm_struct)

      CALL cp_fm_get_info(matrix=fm_matrix_ks, nrow_block=nrow_block, ncol_block=ncol_block)

      ! calculate S^(-1/2) (cholescky decomposition)
      CALL cp_fm_cholesky_decompose(fm_matrix_s)
      CALL cp_fm_triangular_invert(fm_matrix_s)

      NULLIFY (mos_mp2)
      ALLOCATE (mos_mp2(nspins))
      DO ispin = 1, nspins
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, maxocc=maxocc, nelectron=nelectron)

         NULLIFY (mos_mp2(ispin)%mo_set)
         CALL allocate_mo_set(mo_set=mos_mp2(ispin)%mo_set, &
                              nao=nao, &
                              nmo=nao, &
                              nelectron=nelectron, &
                              n_el_f=REAL(nelectron, dp), &
                              maxocc=maxocc, &
                              flexible_electron_count=dft_control%relax_multiplicity)
      END DO

      NULLIFY (my_mpools)
      CALL mpools_create(mpools=my_mpools)
      CALL mpools_rebuild_fm_pools(mpools=my_mpools, &
                                   mos=mos_mp2, &
                                   blacs_env=blacs_env, &
                                   para_env=para_env)

      DO ispin = 1, nspins

         ! If ADMM we should make the ks matrix up-to-date
         IF (dft_control%do_admm) THEN
            CALL admm_correct_for_eigenvalues(ispin, admm_env, matrix_ks(ispin)%matrix)
         END IF

         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, fm_matrix_ks)

         IF (dft_control%do_admm) THEN
            CALL admm_uncorrect_for_eigenvalues(ispin, admm_env, matrix_ks(ispin)%matrix)
         END IF

         CALL init_mo_set(mos_mp2(ispin)%mo_set, &
                          fm_pool=my_mpools%ao_mo_fm_pools(ispin)%pool, &
                          name="mp2_mos")

         ! diagonalize KS matrix
         cholesky_method = cholesky_inverse
         CALL eigensolver(matrix_ks_fm=fm_matrix_ks, &
                          mo_set=mos_mp2(ispin)%mo_set, &
                          ortho=fm_matrix_s, &
                          work=fm_matrix_work, &
                          cholesky_method=cholesky_method, &
                          do_level_shift=.FALSE., &
                          level_shift=0.0_dp, &
                          matrix_u_fm=null(), &
                          use_jacobi=.FALSE.)
      END DO

      CALL cp_fm_release(fm_matrix_s)
      CALL cp_fm_release(fm_matrix_ks)
      CALL cp_fm_release(fm_matrix_work)
      CALL mpools_release(mpools=my_mpools)

      hfx_sections => section_vals_get_subs_vals(input, "DFT%XC%HF")

      !   build the table of index
      t1 = m_walltime()
      ALLOCATE (mp2_biel%index_table(natom, max_nset))

      CALL build_index_table(natom, max_nset, mp2_biel%index_table, basis_parameter, kind_of)

      ! free the hfx_container (for now if forces are required we don't release the HFX stuff)
      free_hfx_buffer = .FALSE.
      IF (ASSOCIATED(qs_env%x_data)) THEN
         free_hfx_buffer = .TRUE.
         IF (calc_forces .AND. (.NOT. mp2_env%ri_mp2%free_hfx_buffer)) free_hfx_buffer = .FALSE.
      END IF

      IF (virial%pv_numer) THEN
         ! in the case of numerical stress we don't have to free the HFX integrals
         free_hfx_buffer = .FALSE.
         mp2_env%ri_mp2%free_hfx_buffer = .FALSE.
      END IF
      ! calculate the matrix sigma_x - vxc for G0W0
      t3 = 0
      IF (mp2_env%ri_rpa%do_ri_g0w0) THEN
         CALL compute_vec_Sigma_x_minus_vxc_gw(qs_env, mp2_env, mos_mp2, E_ex_from_GW, t3)
      END IF
      IF (free_hfx_buffer) THEN
         CALL timeset(routineN//"_free_hfx", handle2)
         CALL section_vals_get(hfx_sections, n_repetition=n_rep_hf)
         n_threads = 1
!$       n_threads = omp_get_max_threads()

         DO irep = 1, n_rep_hf
            DO i_thread = 0, n_threads-1
               actual_x_data => qs_env%x_data(irep, i_thread+1)

               do_dynamic_load_balancing = .TRUE.
               IF (n_threads == 1 .OR. actual_x_data%memory_parameter%do_disk_storage) do_dynamic_load_balancing = .FALSE.

               IF (do_dynamic_load_balancing) THEN
                  my_bin_size = SIZE(actual_x_data%distribution_energy)
               ELSE
                  my_bin_size = 1
               END IF

               IF (.NOT. actual_x_data%memory_parameter%do_all_on_the_fly) THEN
                  CALL dealloc_containers(actual_x_data, hfx_do_eval_energy)
                  !  CALL alloc_containers(actual_x_data, my_bin_size, hfx_do_eval_energy)
                  !
                  !  DO bin=1, my_bin_size
                  !    maxval_container => actual_x_data%maxval_container(bin)
                  !    integral_containers => actual_x_data%integral_containers(:,bin)
                  !    CALL hfx_init_container(maxval_container, actual_x_data%memory_parameter%actual_memory_usage, .FALSE.)
                  !    DO i=1,64
                  !      CALL hfx_init_container(integral_containers(i), &
                  !               actual_x_data%memory_parameter%actual_memory_usage, .FALSE.)
                  !    END DO
                  !  END DO
               END IF
            END DO
         END DO
         CALL timestop(handle2)
      END IF

      Emp2 = 0.D+00
      Emp2_Cou = 0.D+00
      Emp2_ex = 0.D+00
      calc_ex = .TRUE.

      t1 = m_walltime()
      SELECT CASE (mp2_env%method)
      CASE (mp2_method_laplace)
         CALL cp__b("mp2.F",426,"laplace not implemented")
      CASE (mp2_method_direct)
         IF (unit_nr > 0) WRITE (unit_nr, *)

         IF (nspins == 2) THEN
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'Unrestricted Canonical Direct Methods:'
            ! for now, require the mos to be always present

            ! get the alpha coeff and eigenvalues
            CALL get_mo_set(mo_set=mos_mp2(1)%mo_set, &
                            nelectron=nelec_alpha, &
                            eigenvalues=mo_eigenvalues, &
                            mo_coeff=mo_coeff)
            ALLOCATE (C_alpha(dimen, dimen))

            ALLOCATE (Auto_alpha(dimen))

            CALL cp_fm_get_submatrix(mo_coeff, C_alpha, 1, 1, dimen, dimen, .FALSE.)
            Auto_alpha(:) = mo_eigenvalues(:)

            ! get the beta coeff and eigenvalues
            CALL get_mo_set(mo_set=mos_mp2(2)%mo_set, &
                            nelectron=nelec_beta, &
                            eigenvalues=mo_eigenvalues, &
                            mo_coeff=mo_coeff)
            ALLOCATE (C_beta(dimen, dimen))

            ALLOCATE (Auto_beta(dimen))

            CALL cp_fm_get_submatrix(mo_coeff, C_beta, 1, 1, dimen, dimen, .FALSE.)
            Auto_beta(:) = mo_eigenvalues(:)

            ! calculate the alpha-alpha MP2
            Emp2_AA = 0.0_dp
            Emp2_AA_Cou = 0.0_dp
            Emp2_AA_ex = 0.0_dp
            CALL mp2_direct_energy(dimen, nelec_alpha, nelec_alpha, mp2_biel, &
                                   mp2_env, C_alpha, Auto_alpha, Emp2_AA, Emp2_AA_Cou, Emp2_AA_ex, &
                                   qs_env, rho, para_env, &
                                   unit_nr)
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'MP2 Energy Alpha-Alpha = ', Emp2_AA
            IF (unit_nr > 0) WRITE (unit_nr, *)

            Emp2_BB = 0.0_dp
            Emp2_BB_Cou = 0.0_dp
            Emp2_BB_ex = 0.0_dp
            CALL mp2_direct_energy(dimen, nelec_beta, nelec_beta, mp2_biel, mp2_env, &
                                   C_beta, Auto_beta, Emp2_BB, Emp2_BB_Cou, Emp2_BB_ex, &
                                   qs_env, rho, para_env, &
                                   unit_nr)
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'MP2 Energy Beta-Beta= ', Emp2_BB
            IF (unit_nr > 0) WRITE (unit_nr, *)

            Emp2_AB = 0.0_dp
            Emp2_AB_Cou = 0.0_dp
            Emp2_AB_ex = 0.0_dp
            CALL mp2_direct_energy(dimen, nelec_alpha, nelec_beta, mp2_biel, mp2_env, C_alpha, &
                                   Auto_alpha, Emp2_AB, Emp2_AB_Cou, Emp2_AB_ex, &
                                   qs_env, rho, para_env, &
                                   unit_nr, C_beta, Auto_beta)
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'MP2 Energy Alpha-Beta= ', Emp2_AB
            IF (unit_nr > 0) WRITE (unit_nr, *)

            Emp2 = Emp2_AA+Emp2_BB+Emp2_AB*2.0_dp !+Emp2_BA
            Emp2_Cou = Emp2_AA_Cou+Emp2_BB_Cou+Emp2_AB_Cou*2.0_dp !+Emp2_BA
            Emp2_ex = Emp2_AA_ex+Emp2_BB_ex+Emp2_AB_ex*2.0_dp !+Emp2_BA

            Emp2_S = Emp2_AB*2.0_dp
            Emp2_T = Emp2_AA+Emp2_BB

         ELSE

            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'Canonical Direct Methods:'
            CALL get_mo_set(mo_set=mos_mp2(1)%mo_set, &
                            nelectron=nelectron, &
                            eigenvalues=mo_eigenvalues, &
                            mo_coeff=mo_coeff)
            ALLOCATE (C(dimen, dimen))

            ALLOCATE (Auto(dimen))

            CALL cp_fm_get_submatrix(mo_coeff, C, 1, 1, dimen, dimen, .FALSE.)
            Auto(:) = mo_eigenvalues(:)

            CALL mp2_direct_energy(dimen, nelectron/2, nelectron/2, mp2_biel, mp2_env, C, Auto, Emp2, Emp2_Cou, Emp2_ex, &
                                   qs_env, rho, para_env, &
                                   unit_nr)

         END IF

      CASE (mp2_ri_optimize_basis)
         ! optimize ri basis set or tests for RI-MP2 gradients
         IF (unit_nr > 0) THEN
            WRITE (unit_nr, *)
            WRITE (unit_nr, '(T3,A)') 'Optimization of the auxiliary RI-MP2 basis'
            WRITE (unit_nr, *)
         END IF
         CALL get_mo_set(mo_set=mos_mp2(1)%mo_set, &
                         nelectron=nelectron, &
                         eigenvalues=mo_eigenvalues, &
                         mo_coeff=mo_coeff)
         ALLOCATE (C(dimen, dimen))

         ALLOCATE (Auto(dimen))

         CALL cp_fm_get_submatrix(mo_coeff, C, 1, 1, dimen, dimen, .FALSE.)
         Auto(:) = mo_eigenvalues(:)

         IF (nspins == 2) THEN
            ! get the beta coeff and eigenvalues open-shell case
            CALL get_mo_set(mo_set=mos_mp2(2)%mo_set, &
                            nelectron=nelec_beta, &
                            eigenvalues=mo_eigenvalues, &
                            mo_coeff=mo_coeff)
            ALLOCATE (C_beta(dimen, dimen))

            ALLOCATE (Auto_beta(dimen))

            CALL cp_fm_get_submatrix(mo_coeff, C_beta, 1, 1, dimen, dimen, .FALSE.)
            Auto_beta(:) = mo_eigenvalues(:)

            ! optimize basis
            CALL optimize_ri_basis_main(Emp2, Emp2_Cou, Emp2_ex, Emp2_S, Emp2_T, dimen, natom, nelectron, &
                                        mp2_biel, mp2_env, C, Auto, &
                                        kind_of, qs_env, rho, para_env, &
                                        unit_nr, &
                                        nelec_beta, C_beta, Auto_beta)

         ELSE
            ! optimize basis
            CALL optimize_ri_basis_main(Emp2, Emp2_Cou, Emp2_ex, Emp2_S, Emp2_T, dimen, natom, nelectron/2, &
                                        mp2_biel, mp2_env, C, Auto, &
                                        kind_of, qs_env, rho, para_env, &
                                        unit_nr)
         END IF

      CASE (mp2_method_gpw)
         ! check if calculate the exchange contribution
         IF (mp2_env%scale_T == 0.0_dp) calc_ex = .FALSE.

         ! go with mp2_gpw
         CALL mp2_gpw_main(qs_env, mp2_env, Emp2, Emp2_Cou, Emp2_EX, Emp2_S, Emp2_T, &
                           mos_mp2, para_env, unit_nr, calc_forces, calc_ex)

      CASE (ri_mp2_method_gpw)
         ! check if calculate the exchange contribution
         IF (mp2_env%scale_T == 0.0_dp) calc_ex = .FALSE.

         ! go with mp2_gpw
         CALL mp2_gpw_main(qs_env, mp2_env, Emp2, Emp2_Cou, Emp2_EX, Emp2_S, Emp2_T, &
                           mos_mp2, para_env, unit_nr, calc_forces, calc_ex, do_ri_mp2=.TRUE.)

      CASE (ri_rpa_method_gpw)
         ! perform RI-RPA energy calculation (since most part of the calculation
         ! is actually equal to the RI-MP2-GPW we decided to put RPA in the MP2
         ! section to avoid code replication)

         calc_ex = .FALSE.

         ! go with ri_rpa_gpw
         CALL mp2_gpw_main(qs_env, mp2_env, Emp2, Emp2_Cou, Emp2_EX, Emp2_S, Emp2_T, &
                           mos_mp2, para_env, unit_nr, calc_forces, calc_ex, do_ri_rpa=.TRUE.)

      CASE (ri_mp2_laplace)
         ! perform RI-SOS-Laplace-MP2 energy calculation, most part of the code in common
         ! with the RI-RPA part

         ! In SOS-MP2 only the coulomb-like contribution of the MP2 energy is computed
         calc_ex = .FALSE.

         ! go with sos_laplace_mp2_gpw
         CALL mp2_gpw_main(qs_env, mp2_env, Emp2, Emp2_Cou, Emp2_EX, Emp2_S, Emp2_T, &
                           mos_mp2, para_env, unit_nr, calc_forces, calc_ex, do_ri_sos_laplace_mp2=.TRUE.)

      CASE DEFAULT
         CALL cp__b("mp2.F",601,"")
      END SELECT
      t2 = m_walltime()
      IF (unit_nr > 0) WRITE (unit_nr, *)
      IF (mp2_env%method .NE. ri_rpa_method_gpw) THEN
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.6)') 'Total MP2 Time=', t2-t1
         IF (mp2_env%method == ri_mp2_laplace) THEN
            Emp2_S = Emp2
            Emp2_T = 0.0_dp
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'MP2 Energy SO component (singlet) = ', Emp2_S
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'Scaling factor SO                 = ', mp2_env%scale_S
         ELSE
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'MP2 Coulomb Energy = ', Emp2_Cou/2.0_dp
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'MP2 Exchange Energy = ', Emp2_ex
            IF (nspins == 1) THEN
               ! valid only in the closed shell case
               Emp2_S = Emp2_Cou/2.0_dp
               IF (calc_ex) THEN
                  Emp2_T = Emp2_ex+Emp2_Cou/2.0_dp
               ELSE
                  ! unknown if Emp2_ex is not computed
                  Emp2_T = 0.0_dp
               ENDIF
            END IF
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'MP2 Energy SO component (singlet) = ', Emp2_S
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'MP2 Energy SS component (triplet) = ', Emp2_T
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'Scaling factor SO                 = ', mp2_env%scale_S
            IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'Scaling factor SS                 = ', mp2_env%scale_T
         END IF
         Emp2_S = Emp2_S*mp2_env%scale_S
         Emp2_T = Emp2_T*mp2_env%scale_T
         Emp2 = Emp2_S+Emp2_T
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'Second order perturbation energy  =   ', Emp2
      ELSE
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.6)') 'Total RI-RPA Time=', t2-t1
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'RI-RPA energy  =   ', Emp2
      END IF
      IF (unit_nr > 0) WRITE (unit_nr, *)

      ! we have it !!!!
      energy%mp2 = Emp2
      energy%total = energy%total+Emp2

      DO ispin = 1, nspins
         CALL deallocate_mo_set(mo_set=mos_mp2(ispin)%mo_set)
      END DO
      DEALLOCATE (mos_mp2)

      ! if necessary reallocate hfx buffer
      IF (free_hfx_buffer .AND. (.NOT. calc_forces)) THEN
         CALL timeset(routineN//"_alloc_hfx", handle2)
         DO irep = 1, n_rep_hf
            DO i_thread = 0, n_threads-1
               actual_x_data => qs_env%x_data(irep, i_thread+1)

               do_dynamic_load_balancing = .TRUE.
               IF (n_threads == 1 .OR. actual_x_data%memory_parameter%do_disk_storage) do_dynamic_load_balancing = .FALSE.

               IF (do_dynamic_load_balancing) THEN
                  my_bin_size = SIZE(actual_x_data%distribution_energy)
               ELSE
                  my_bin_size = 1
               END IF

               IF (.NOT. actual_x_data%memory_parameter%do_all_on_the_fly) THEN
                  ! CALL dealloc_containers(actual_x_data, hfx_do_eval_energy)
                  CALL alloc_containers(actual_x_data, my_bin_size, hfx_do_eval_energy)

                  DO bin = 1, my_bin_size
                     maxval_container => actual_x_data%maxval_container(bin)
                     integral_containers => actual_x_data%integral_containers(:, bin)
                     CALL hfx_init_container(maxval_container, actual_x_data%memory_parameter%actual_memory_usage, .FALSE.)
                     DO i = 1, 64
                        CALL hfx_init_container(integral_containers(i), actual_x_data%memory_parameter%actual_memory_usage, .FALSE.)
                     END DO
                  END DO
               END IF
            END DO
         END DO
         CALL timestop(handle2)
      END IF

      CALL hfx_release_basis_types(basis_parameter)

      ! if required calculate the EXX contribution from the DFT density
      IF (mp2_env%method == ri_rpa_method_gpw) THEN
         do_exx = .FALSE.
         hfx_sections => section_vals_get_subs_vals(input, "DFT%XC%WF_CORRELATION%RI_RPA%HF")
         CALL section_vals_get(hfx_sections, explicit=do_exx)
         IF (do_exx) THEN
            do_gw = mp2_env%ri_rpa%do_ri_g0w0
            do_admm = mp2_env%ri_rpa%do_admm
            CALL calculate_exx(qs_env, unit_nr, do_gw, do_admm, E_ex_from_GW, t3)
         END IF
      END IF

      CALL cp_print_key_finished_output(unit_nr, logger, input, &
                                        "DFT%XC%WF_CORRELATION%MP2_INFO")

      CALL timestop(handle)

   END SUBROUTINE mp2_main

! **************************************************************************************************
!> \brief ...
!> \param natom ...
!> \param max_nset ...
!> \param index_table ...
!> \param basis_parameter ...
!> \param kind_of ...
! **************************************************************************************************
   SUBROUTINE build_index_table(natom, max_nset, index_table, basis_parameter, kind_of)
      INTEGER                                            :: natom, max_nset
      INTEGER, DIMENSION(natom, max_nset)                :: index_table
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_parameter
      INTEGER, DIMENSION(natom)                          :: kind_of

      INTEGER                                            :: counter, iatom, ikind, iset, nset

      index_table = -HUGE(0)
      counter = 0
      DO iatom = 1, natom
         ikind = kind_of(iatom)
         nset = basis_parameter(ikind)%nset
         DO iset = 1, nset
            index_table(iatom, iset) = counter+1
            counter = counter+basis_parameter(ikind)%nsgf(iset)
         END DO
      END DO

   END SUBROUTINE build_index_table

! **************************************************************************************************
!> \brief ...
!> \param dimen ...
!> \param occ_i ...
!> \param occ_j ...
!> \param mp2_biel ...
!> \param mp2_env ...
!> \param C_i ...
!> \param Auto_i ...
!> \param Emp2 ...
!> \param Emp2_Cou ...
!> \param Emp2_ex ...
!> \param qs_env ...
!> \param rho ...
!> \param para_env ...
!> \param unit_nr ...
!> \param C_j ...
!> \param Auto_j ...
! **************************************************************************************************
   SUBROUTINE mp2_direct_energy(dimen, occ_i, occ_j, mp2_biel, mp2_env, C_i, Auto_i, Emp2, Emp2_Cou, Emp2_ex, &
                                qs_env, rho, para_env, &
                                unit_nr, C_j, Auto_j)
      INTEGER                                            :: dimen, occ_i, occ_j
      TYPE(mp2_biel_type)                                :: mp2_biel
      TYPE(mp2_type), POINTER                            :: mp2_env
      REAL(KIND=dp), DIMENSION(dimen, dimen)             :: C_i
      REAL(KIND=dp), DIMENSION(dimen)                    :: Auto_i
      REAL(KIND=dp)                                      :: Emp2, Emp2_Cou, Emp2_ex
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: unit_nr
      REAL(KIND=dp), DIMENSION(dimen, dimen), OPTIONAL   :: C_j
      REAL(KIND=dp), DIMENSION(dimen), OPTIONAL          :: Auto_j

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_direct_energy', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: zero = 0.D+00

      INTEGER :: batch_number, color_sub, comm_sub, counter, elements_ij_proc, group_counter, &
         handle, i, i_batch, i_batch_start, i_group_counter, j, j_batch_start, j_group_counter, &
         last_batch, max_batch_number, max_batch_size, max_set, minimum_memory_needed, &
         my_batch_size, my_I_batch_size, my_I_occupied_end, my_I_occupied_start, my_J_batch_size, &
         my_J_occupied_end, my_J_occupied_start, natom, Ni_occupied, Nj_occupied, number_groups, &
         number_i_subset, number_j_subset, one, sqrt_number_groups, total_I_size_batch_group, &
         total_J_size_batch_group, virt_i, virt_j
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: batch_sizes, batch_sizes_tmp, &
                                                            vector_batch_I_size_group, &
                                                            vector_batch_J_size_group
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ij_list_proc, ij_list_proc_temp, &
                                                            ij_matrix
      LOGICAL                                            :: alpha_beta_case
      TYPE(cp_para_env_type), POINTER                    :: para_env_sub

      CALL timeset(routineN, handle)

      alpha_beta_case = .FALSE.
      IF (PRESENT(C_j) .AND. PRESENT(Auto_j)) alpha_beta_case = .TRUE.

      IF (unit_nr > 0 .AND. mp2_env%potential_parameter%potential_type == do_mp2_potential_TShPSC) THEN
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T64,F12.6,A5)') 'Truncated MP2 method, Rt=', &
            mp2_env%potential_parameter%truncation_radius, ' Bohr'
      END IF

      ! create the local para env
      ! each para_env_sub corresponds to a group that is going to compute
      ! all the integrals. To each group a batch I is assigned and the
      ! communication takes place only inside the group
      number_groups = para_env%num_pe/mp2_env%mp2_num_proc
      IF (number_groups*mp2_env%mp2_num_proc /= para_env%num_pe) THEN
         CALL cp__b("mp2.F",803," The number of processors needs to be a multiple of the processors per group. ")
      END IF
      IF (number_groups > occ_i*occ_j) THEN
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'Number of groups greater then the number of IJ pairs!'
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A)') 'Consider using more processors per group for improved efficiency'
      END IF

      color_sub = para_env%mepos/mp2_env%mp2_num_proc
      CALL mp_comm_split_direct(para_env%group, comm_sub, color_sub)
      NULLIFY (para_env_sub)
      CALL cp_para_env_create(para_env_sub, comm_sub)

      ! calculate the maximal size of the batch, according to the maximum RS size
      max_set = SIZE(mp2_biel%index_table, 2)
      minimum_memory_needed = (8*(max_set**4))/1024**2
      IF (minimum_memory_needed > mp2_env%mp2_memory) THEN
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T66,F12.6,A3)') 'Memory required below the minimum, new memory:', &
            minimum_memory_needed, 'MiB'
         mp2_env%mp2_memory = minimum_memory_needed
      END IF

      ! Distribute the batches over the groups in
      ! a rectangular fashion, bigger size for J index
      ! the sizes of the I batches should be as small as possible
      sqrt_number_groups = INT(SQRT(REAL(number_groups, KIND=dp)))
      DO i = 1, number_groups
         IF (MOD(number_groups, i) == 0) THEN
            IF (sqrt_number_groups/i <= 1) THEN
               number_j_subset = i
               EXIT
            END IF
         END IF
      END DO
      number_i_subset = number_groups/number_j_subset

      IF (number_i_subset < number_j_subset) THEN
         number_i_subset = number_j_subset
         number_j_subset = number_groups/number_i_subset
      END IF

      ! Distribute the I index and the J index over groups
      total_I_size_batch_group = occ_i/number_i_subset
      IF (total_I_size_batch_group < 1) total_I_size_batch_group = 1
      ALLOCATE (vector_batch_I_size_group(0:number_i_subset-1))

      vector_batch_I_size_group = 0
      DO i = 0, number_i_subset-1
         vector_batch_I_size_group(i) = total_I_size_batch_group
      END DO
      IF (SUM(vector_batch_I_size_group) /= occ_i) THEN
         one = 1
         IF (SUM(vector_batch_I_size_group) > occ_i) one = -1
         i = -1
         DO
            i = i+1
            vector_batch_I_size_group(i) = vector_batch_I_size_group(i)+one
            IF (SUM(vector_batch_I_size_group) == occ_i) EXIT
            IF (i == number_i_subset-1) i = -1
         END DO
      END IF

      total_J_size_batch_group = occ_j/number_j_subset
      IF (total_J_size_batch_group < 1) total_J_size_batch_group = 1
      ALLOCATE (vector_batch_J_size_group(0:number_j_subset-1))

      vector_batch_J_size_group = 0
      DO i = 0, number_J_subset-1
         vector_batch_J_size_group(i) = total_J_size_batch_group
      END DO
      IF (SUM(vector_batch_J_size_group) /= occ_j) THEN
         one = 1
         IF (SUM(vector_batch_J_size_group) > occ_j) one = -1
         i = -1
         DO
            i = i+1
            vector_batch_J_size_group(i) = vector_batch_J_size_group(i)+one
            IF (SUM(vector_batch_J_size_group) == occ_j) EXIT
            IF (i == number_J_subset-1) i = -1
         END DO
      END IF

      ! now the starting and ending I and J occupied orbitals are assigned to each group
      group_counter = 0
      i_group_counter = 0
      my_I_occupied_start = 1
      DO i = 0, number_i_subset-1
         my_J_occupied_start = 1
         j_group_counter = 0
         DO j = 0, number_j_subset-1
            group_counter = group_counter+1
            IF (color_sub == group_counter-1) EXIT
            my_J_occupied_start = my_J_occupied_start+vector_batch_J_size_group(j)
            j_group_counter = j_group_counter+1
         END DO
         IF (color_sub == group_counter-1) EXIT
         my_I_occupied_start = my_I_occupied_start+vector_batch_I_size_group(i)
         i_group_counter = i_group_counter+1
      END DO
      my_I_occupied_end = my_I_occupied_start+vector_batch_I_size_group(i_group_counter)-1
      my_I_batch_size = vector_batch_I_size_group(i_group_counter)
      my_J_occupied_end = my_J_occupied_start+vector_batch_J_size_group(j_group_counter)-1
      my_J_batch_size = vector_batch_J_size_group(j_group_counter)

      DEALLOCATE (vector_batch_I_size_group)
      DEALLOCATE (vector_batch_J_size_group)

      max_batch_size = MIN( &
                       MAX(1, &
                           INT(mp2_env%mp2_memory*INT(1024, KIND=int_8)**2/ &
                               (8*(2*dimen-occ_i)*INT(dimen, KIND=int_8)*my_J_batch_size/para_env_sub%num_pe))) &
                       , my_I_batch_size)
      IF (max_batch_size < 1) THEN
         max_batch_size = INT((8*(occ_i+1)*INT(dimen, KIND=int_8)**2/para_env%num_pe)/1024**2)
         IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T72,I6,A3)') 'More memory required, at least:', max_batch_size, 'MiB'
         max_batch_size = 1
      END IF

      ! create the size of the batches inside the group
      my_batch_size = my_I_batch_size
      ALLOCATE (batch_sizes(my_batch_size))

      batch_sizes = -HUGE(0)
      batch_number = 0
      DO i = 1, my_batch_size
         IF (i*max_batch_size > my_batch_size) EXIT
         batch_number = batch_number+1
         batch_sizes(i) = max_batch_size
      END DO
      last_batch = my_batch_size-max_batch_size*batch_number
      IF (last_batch > 0) THEN
         batch_number = batch_number+1
         batch_sizes(batch_number) = last_batch
      END IF

      ALLOCATE (batch_sizes_tmp(batch_number))
      batch_sizes_tmp(1:batch_number) = batch_sizes(1:batch_number)
      DEALLOCATE (batch_sizes)
      ALLOCATE (batch_sizes(batch_number))
      batch_sizes(:) = batch_sizes_tmp
      DEALLOCATE (batch_sizes_tmp)

      max_batch_size = MAXVAL(batch_sizes)
      CALL mp_max(max_batch_size, para_env%group)
      max_batch_number = batch_number
      CALL mp_max(max_batch_number, para_env%group)
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T3,A,T76,I5)') 'Maximum used batch size: ', max_batch_size
         WRITE (unit_nr, '(T3,A,T76,I5)') 'Number of integral recomputations: ', max_batch_number
         CALL m_flush(unit_nr)
      END IF

      ! Batches sizes exceed the occupied orbitals allocated for group
      IF(.NOT.(SUM(batch_sizes) <= my_batch_size))CALL cp__a("mp2.F",955)

      virt_i = dimen-occ_i
      virt_j = dimen-occ_j
      natom = SIZE(mp2_biel%index_table, 1)

      CALL mp_sync(para_env%group)
      Emp2 = zero
      Emp2_Cou = zero
      Emp2_ex = zero
      i_batch_start = my_I_occupied_start-1
      j_batch_start = my_J_occupied_start-1
      Nj_occupied = my_J_batch_size
      DO i_batch = 1, batch_number

         Ni_occupied = batch_sizes(i_batch)

         counter = -1
         ALLOCATE (ij_matrix(Ni_occupied, Nj_occupied))

         ij_matrix = 0
         DO i = 1, Ni_occupied
            DO j = 1, Nj_occupied
               counter = counter+1
               IF (MOD(counter, para_env_sub%num_pe) == para_env_sub%mepos) THEN
                  ij_matrix(i, j) = ij_matrix(i, j)+1
               END IF
            END DO
         END DO

         ALLOCATE (ij_list_proc_temp(Ni_occupied*occ_j, 2))

         elements_ij_proc = 0
         DO i = 1, Ni_occupied
            DO j = 1, Nj_occupied
               IF (ij_matrix(i, j) == 0) CYCLE
               elements_ij_proc = elements_ij_proc+1
               ij_list_proc_temp(elements_ij_proc, 1) = i
               ij_list_proc_temp(elements_ij_proc, 2) = j
            END DO
         END DO
         DEALLOCATE (ij_matrix)

         ALLOCATE (ij_list_proc(elements_ij_proc, 2))
         DO i = 1, elements_ij_proc
            ij_list_proc(i, 1) = ij_list_proc_temp(i, 1)
            ij_list_proc(i, 2) = ij_list_proc_temp(i, 2)
         END DO
         DEALLOCATE (ij_list_proc_temp)

         IF (.NOT. alpha_beta_case) THEN
            CALL mp2_canonical_direct_single_batch(Emp2, Emp2_Cou, Emp2_ex, mp2_env, qs_env, rho, para_env_sub, &
                                                   mp2_biel, dimen, C_i, Auto_i, i_batch_start, Ni_occupied, occ_i, &
                                                   elements_ij_proc, ij_list_proc, Nj_occupied, j_batch_start)
         ELSE
            CALL mp2_canonical_direct_single_batch(Emp2, Emp2_Cou, Emp2_ex, mp2_env, qs_env, rho, para_env_sub, &
                                                   mp2_biel, dimen, C_i, Auto_i, i_batch_start, Ni_occupied, occ_i, &
                                                   elements_ij_proc, ij_list_proc, Nj_occupied, j_batch_start, &
                                                   occ_j, C_j, Auto_j)
         END IF

         i_batch_start = i_batch_start+Ni_occupied

         DEALLOCATE (ij_list_proc)

      END DO

      CALL mp_sum(Emp2_Cou, para_env%group)
      CALL mp_sum(Emp2_Ex, para_env%group)
      CALL mp_sum(Emp2, para_env%group)

      CALL cp_para_env_release(para_env_sub)

      CALL timestop(handle)

   END SUBROUTINE mp2_direct_energy

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param unit_nr ...
!> \param do_gw ...
!> \param do_admm ...
!> \param E_ex_from_GW ...
!> \param t3 ...
! **************************************************************************************************
   SUBROUTINE calculate_exx(qs_env, unit_nr, do_gw, do_admm, E_ex_from_GW, t3)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER                                            :: unit_nr
      LOGICAL                                            :: do_gw, do_admm
      REAL(KIND=dp)                                      :: E_ex_from_GW, t3

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_exx', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, irep, n_rep_hf, ns
      REAL(KIND=dp)                                      :: t1, t2
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_ks_aux_fit, rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_ks_2d, rho_ao_2d
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: hfx_sections, input

      CALL timeset(routineN, handle)

      t1 = m_walltime()

      NULLIFY (hfx_sections, input, para_env, matrix_ks, matrix_ks_aux_fit, rho, rho_ao)

      ! if ADMM for exchange energy, get rho_aux_fit and matrix_ks_aux_fit
      IF (do_admm) THEN
         CALL get_qs_env(qs_env=qs_env, &
                         input=input, &
                         para_env=para_env, &
                         energy=energy, &
                         rho_aux_fit=rho, &
                         matrix_ks=matrix_ks, &
                         matrix_ks_aux_fit=matrix_ks_aux_fit)
      ELSE
         CALL get_qs_env(qs_env=qs_env, &
                         input=input, &
                         para_env=para_env, &
                         energy=energy, &
                         rho=rho, &
                         matrix_ks=matrix_ks, &
                         matrix_ks_aux_fit=matrix_ks_aux_fit)
      END IF

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      hfx_sections => section_vals_get_subs_vals(input, "DFT%XC%WF_CORRELATION%RI_RPA%HF")

      CALL section_vals_get(hfx_sections, n_repetition=n_rep_hf)

      ! put matrix_ks to zero
      DO i = 1, SIZE(matrix_ks)
         CALL cp_dbcsr_set(matrix_ks(i)%matrix, 0.0_dp)
         IF (do_admm) THEN
            CALL cp_dbcsr_set(matrix_ks_aux_fit(i)%matrix, 0.0_dp)
         END IF
      END DO

      ! Remove the Exchange-correlation energy contributions from the total energy
      energy%total = energy%total-(energy%exc+energy%exc1+energy%ex+energy%exc_aux_fit)

      energy%exc = 0.0_dp
      energy%exc1 = 0.0_dp
      energy%exc_aux_fit = 0.0_dp
      energy%ex = 0.0_dp

      ! take the exact exchange energy from GW or calculate it
      IF (do_gw) THEN
         energy%total = energy%total+E_ex_from_GW
         energy%ex = E_ex_from_GW
      ELSE

         DO irep = 1, n_rep_hf
            ns = SIZE(rho_ao)
            rho_ao_2d(1:ns, 1:1) => rho_ao(1:ns)
            ns = SIZE(matrix_ks)
            IF (do_admm) THEN
               matrix_ks_2d(1:ns, 1:1) => matrix_ks_aux_fit(1:ns)
            ELSE
               matrix_ks_2d(1:ns, 1:1) => matrix_ks(1:ns)
            END IF
            CALL integrate_four_center(qs_env, matrix_ks_2d, energy, rho_ao_2d, hfx_sections, &
                                       para_env, .TRUE., irep, .TRUE., &
                                       ispin=1, do_exx=.TRUE.)
         END DO

         ! include the EXX contribution to the total energy
         energy%total = energy%total+energy%ex
      END IF

      t2 = m_walltime()

      IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.6)') 'Total EXX Time=', t2-t1+t3
      IF (unit_nr > 0) WRITE (unit_nr, '(T3,A,T56,F25.14)') 'EXX energy  =   ', energy%ex

      ! reset to zero the Hartree-Fock energy
      energy%ex = 0.0_dp

      CALL timestop(handle)

   END SUBROUTINE calculate_exx

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param mp2_env ...
!> \param mos_mp2 ...
!> \param energy_ex ...
!> \param t3 ...
!> \par History
!>      04.2015 created
!> \author Jan Wilhelm
! **************************************************************************************************
   SUBROUTINE compute_vec_Sigma_x_minus_vxc_gw(qs_env, mp2_env, mos_mp2, energy_ex, t3)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mp2_type), POINTER                            :: mp2_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos_mp2
      REAL(KIND=dp)                                      :: energy_ex, t3

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_vec_Sigma_x_minus_vxc_gw', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dimen, gw_corr_lev_occ, &
                                                            gw_corr_lev_virt, homo, irep, ispin, &
                                                            myfun, myfun_aux, myfun_prim, &
                                                            n_rep_hf, nmo, ns
      LOGICAL                                            :: charge_constrain_tmp, do_admm_rpa, &
                                                            do_hfx, do_mix_exchange, do_ri_Sigma_x
      REAL(KIND=dp)                                      :: frac_exx, hfx_fraction, t1, t2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: vec_Sigma_x_minus_vxc_gw
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_ks_aux_fit, &
                                                            matrix_sigma_x_minus_vxc, rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_ks_2d, rho_ao_2d
      TYPE(cp_dbcsr_type)                                :: matrix_tmp, matrix_tmp_2, mo_coeff_b
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy_dummy
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: hfx_sections, input, xc_section, &
                                                            xc_section_admm_aux, &
                                                            xc_section_admm_prim

      NULLIFY (admm_env, matrix_ks, matrix_ks_aux_fit, rho_ao, matrix_sigma_x_minus_vxc, input, &
               xc_section, xc_section_admm_aux, xc_section_admm_prim, hfx_sections, rho, &
               energy_dummy, dft_control, para_env, ks_env, mo_coeff)

      t1 = m_walltime()

      do_admm_rpa = mp2_env%ri_rpa%do_admm
      do_ri_Sigma_x = mp2_env%ri_g0w0%do_ri_Sigma_x

      ! In case we do admm for RPA/GW, get the aux. density matrix rho_aux_fit, otherwise
      ! just the normal density matrix rho
      IF (do_admm_rpa) THEN
         CALL get_qs_env(qs_env, &
                         admm_env=admm_env, &
                         matrix_ks=matrix_ks, &
                         matrix_ks_aux_fit=matrix_ks_aux_fit, &
                         rho_aux_fit=rho, &
                         input=input, &
                         dft_control=dft_control, &
                         para_env=para_env, &
                         ks_env=ks_env)
      ELSE
         CALL get_qs_env(qs_env, &
                         admm_env=admm_env, &
                         matrix_ks=matrix_ks, &
                         matrix_ks_aux_fit=matrix_ks_aux_fit, &
                         rho=rho, &
                         input=input, &
                         dft_control=dft_control, &
                         para_env=para_env, &
                         ks_env=ks_env)
      END IF

      CALL qs_rho_get(rho, rho_ao=rho_ao)

      DO ispin = 1, SIZE(matrix_ks)
         ! If ADMM we should make the ks matrix up-to-date
         IF (dft_control%do_admm) THEN
            CALL admm_correct_for_eigenvalues(ispin, admm_env, matrix_ks(ispin)%matrix)
         END IF
      END DO

      ! initialize matrix_sigma_x_minus_vxc
      ALLOCATE (matrix_sigma_x_minus_vxc(SIZE(matrix_ks)))
      DO ispin = 1, SIZE(matrix_ks)
         NULLIFY (matrix_sigma_x_minus_vxc(ispin)%matrix)
         CALL cp_dbcsr_init_p(matrix_sigma_x_minus_vxc(ispin)%matrix)
         CALL cp_dbcsr_copy(matrix_sigma_x_minus_vxc(ispin)%matrix, matrix_ks(ispin)%matrix, &
                            name="Matrix VXC of spin "//cp_to_string(ispin))
         CALL cp_dbcsr_set(matrix_ks(ispin)%matrix, 0.0_dp)
      ENDDO

      ! set DFT functional to none and hfx_fraction to zero
      hfx_sections => section_vals_get_subs_vals(input, "DFT%XC%HF")
      CALL section_vals_get(hfx_sections, explicit=do_hfx)
      IF (do_hfx) THEN
         hfx_fraction = qs_env%x_data(1, 1)%general_parameter%fraction
         qs_env%x_data(:, :)%general_parameter%fraction = 0.0_dp
      END IF
      xc_section => section_vals_get_subs_vals(input, "DFT%XC")
      CALL section_vals_val_get(xc_section, "XC_FUNCTIONAL%_SECTION_PARAMETERS_", &
                                i_val=myfun)
      CALL section_vals_val_set(xc_section, "XC_FUNCTIONAL%_SECTION_PARAMETERS_", &
                                i_val=xc_none)

      ! in ADMM, also set the XC functional for ADMM correction to none
      ! do not do this if we do ADMM for Sigma_x
      IF (dft_control%do_admm) THEN
         xc_section_admm_aux => section_vals_get_subs_vals(admm_env%xc_section_aux, &
                                                           "XC_FUNCTIONAL")
         CALL section_vals_val_get(xc_section_admm_aux, "_SECTION_PARAMETERS_", &
                                   i_val=myfun_aux)
         CALL section_vals_val_set(xc_section_admm_aux, "_SECTION_PARAMETERS_", &
                                   i_val=xc_none)

         ! the same for the primary basis
         xc_section_admm_prim => section_vals_get_subs_vals(admm_env%xc_section_primary, &
                                                            "XC_FUNCTIONAL")
         CALL section_vals_val_get(xc_section_admm_aux, "_SECTION_PARAMETERS_", &
                                   i_val=myfun_prim)
         CALL section_vals_val_set(xc_section_admm_prim, "_SECTION_PARAMETERS_", &
                                   i_val=xc_none)

         ! for ADMMQ/S, set the charge_constrain to false (otherwise wrong results)
         charge_constrain_tmp = .FALSE.
         IF (admm_env%charge_constrain) THEN
            admm_env%charge_constrain = .FALSE.
            charge_constrain_tmp = .TRUE.
         END IF

      END IF

      ! if we do ADMM for Sigma_x, we write the ADMM correction into matrix_ks_aux_fit
      ! and therefore we should set it to zero
      IF (do_admm_rpa) THEN
         DO ispin = 1, dft_control%nspins
            CALL cp_dbcsr_set(matrix_ks_aux_fit(ispin)%matrix, 0.0_dp)
         END DO

!      ! set the hf fraction to minus one to get the ADMM correction negatively
!      ! this is correct because we need ADMM for the exchange self energy Sigma_x
!      ! but it is calculated in -vxc
!      qs_env%x_data(:,:)%general_parameter%fraction = -1.0_dp
!
!      ! do not search for already existing functionals because in Sigma_x we
!      ! have 100 % Fock exchange
!      CALL create_admm_xc_section(qs_env, xc_section, admm_env, error, search_func=.FALSE.)
!
!      qs_env%x_data(:,:)%general_parameter%fraction = 0.0_dp

      END IF

      ! calculate KS-matrix without XC and without HF
      CALL qs_ks_build_kohn_sham_matrix(qs_env=qs_env, calculate_forces=.FALSE., &
                                        just_energy=.FALSE.)

!JW    PRINT *, 'matrix_ks_aux_fit after qs_ks_build_kohn_sham_matrix: '
!JW    CALL cp_dbcsr_print(matrix_ks_aux_fit(1)%matrix,error=error)

!JW    PRINT *, 'matrix_ks after build ks matrix: '
!JW    CALL cp_dbcsr_print(matrix_ks(1)%matrix,error=error)

      ! set the DFT functional and HF fraction back
      CALL section_vals_val_set(xc_section, "XC_FUNCTIONAL%_SECTION_PARAMETERS_", &
                                i_val=myfun)
      IF (do_hfx) THEN
         qs_env%x_data(:, :)%general_parameter%fraction = hfx_fraction
      END IF

      IF (dft_control%do_admm) THEN
         xc_section_admm_aux => section_vals_get_subs_vals(admm_env%xc_section_aux, &
                                                           "XC_FUNCTIONAL")
         xc_section_admm_prim => section_vals_get_subs_vals(admm_env%xc_section_primary, &
                                                            "XC_FUNCTIONAL")

         CALL section_vals_val_set(xc_section_admm_aux, "_SECTION_PARAMETERS_", &
                                   i_val=myfun_aux)
         CALL section_vals_val_set(xc_section_admm_prim, "_SECTION_PARAMETERS_", &
                                   i_val=myfun_prim)
         IF (charge_constrain_tmp) THEN
            admm_env%charge_constrain = .TRUE.
         END IF
      END IF

      ! remove the single-particle part (kin. En + Hartree pot) and change the sign
      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_add(matrix_sigma_x_minus_vxc(ispin)%matrix, matrix_ks(ispin)%matrix, -1.0_dp, 1.0_dp)
      END DO

      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_set(matrix_ks(ispin)%matrix, 0.0_dp)
         IF (do_admm_rpa) THEN
            CALL cp_dbcsr_set(matrix_ks_aux_fit(ispin)%matrix, 0.0_dp)
         END IF
      END DO

      hfx_sections => section_vals_get_subs_vals(input, "DFT%XC%WF_CORRELATION%RI_RPA%HF")

      CALL section_vals_get(hfx_sections, n_repetition=n_rep_hf)

      CALL allocate_qs_energy(energy_dummy)

      energy_dummy%ex = 0.0_dp

!JW    PRINT *, 'matrix_ks before int 4 c:'
!JW    CALL cp_dbcsr_print(matrix_ks(1)%matrix,error=error)

      ! in most cases, we calculate the exchange self-energy here. But if we do only RI for
      ! the exchange self-energy, we do not calculate exchange here
      IF (.NOT. do_ri_Sigma_x .OR. do_mix_exchange) THEN

         ! add here HFX (=Sigma_exchange) to matrix_sigma_x_minus_vxc
         DO irep = 1, n_rep_hf
            ns = SIZE(rho_ao)
            rho_ao_2d(1:ns, 1:1) => rho_ao(1:ns)
            ns = SIZE(matrix_ks)
            IF (do_admm_rpa) THEN
               matrix_ks_2d(1:ns, 1:1) => matrix_ks_aux_fit(1:ns)
            ELSE
               matrix_ks_2d(1:ns, 1:1) => matrix_ks(1:ns)
            END IF

            CALL integrate_four_center(qs_env, matrix_ks_2d, energy_dummy, rho_ao_2d, hfx_sections, &
                                       para_env, .TRUE., irep, .TRUE., &
                                       ispin=1, do_exx=.TRUE.)
         END DO

      END IF

!JW    PRINT *, 'matrix_ks after int 4 c:'
!JW    CALL cp_dbcsr_print(matrix_ks(1)%matrix,error=error)

!JW    PRINT *, 'rho_ao for auxiliary basis hopefully'
!JW    CALL cp_dbcsr_print(rho_ao(1)%matrix,error=error)

!JW    PRINT *, 'matrix_ks_aux_fit after int 4c: '
!JW    CALL cp_dbcsr_print(matrix_ks_aux_fit(1)%matrix,error=error)

      ! transform Fock-Matrix (calculated in integrate_four_center, written in matrix_ks_aux_fit in case
      ! of ADMM) from ADMM basis to primary basis
      IF (do_admm_rpa) THEN
         CALL admm_mo_merge_ks_matrix(qs_env)
      END IF

!JW    PRINT *, 'matrix_ks after merging: '
!JW    CALL cp_dbcsr_print(matrix_ks(1)%matrix,error=error)

      do_mix_exchange = mp2_env%ri_g0w0%mix_exchange

      IF (do_mix_exchange) THEN
         frac_exx = mp2_env%ri_g0w0%frac_exx
      ELSE
         frac_exx = 1.0_dp
      END IF

      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_add(matrix_sigma_x_minus_vxc(ispin)%matrix, matrix_ks(ispin)%matrix, 1.0_dp, frac_exx)
      END DO

      energy_ex = energy_dummy%ex

      CALL deallocate_qs_energy(energy_dummy)

      CALL cp_dbcsr_init(mo_coeff_b)
      CALL cp_dbcsr_desymmetrize(matrix_ks(1)%matrix, mo_coeff_b)
      CALL cp_dbcsr_set(mo_coeff_b, 0.0_dp)

      ! Transform matrix_sigma_x_minus_vxc to MO basis
      DO ispin = 1, dft_control%nspins

         CALL get_mo_set(mo_set=mos_mp2(ispin)%mo_set, &
                         mo_coeff=mo_coeff, &
                         nmo=nmo, &
                         homo=homo, &
                         nao=dimen)

         IF (ispin == 1) THEN
            ALLOCATE (vec_Sigma_x_minus_vxc_gw(nmo, dft_control%nspins))
            vec_Sigma_x_minus_vxc_gw = 0.0_dp
         END IF

         CALL cp_dbcsr_set(mo_coeff_b, 0.0_dp)
         CALL copy_fm_to_dbcsr(mo_coeff, mo_coeff_b, keep_sparsity=.TRUE.)

         ! initialize matrix_tmp and matrix_tmp2
         IF (ispin == 1) THEN
            CALL cp_dbcsr_init(matrix_tmp)
            CALL cp_dbcsr_create(matrix_tmp, template=mo_coeff_b)
            CALL cp_dbcsr_copy(matrix_tmp, mo_coeff_b)
            CALL cp_dbcsr_set(matrix_tmp, 0.0_dp)

            CALL cp_dbcsr_init(matrix_tmp_2)
            CALL cp_dbcsr_create(matrix_tmp_2, template=mo_coeff_b)
            CALL cp_dbcsr_copy(matrix_tmp_2, mo_coeff_b)
            CALL cp_dbcsr_set(matrix_tmp_2, 0.0_dp)
         END IF

         gw_corr_lev_occ = mp2_env%ri_g0w0%corr_mos_occ
         gw_corr_lev_virt = mp2_env%ri_g0w0%corr_mos_virt
         ! if corrected occ/virt levels exceed the number of occ/virt levels,
         ! correct all occ/virt level energies
         IF (gw_corr_lev_occ > homo) gw_corr_lev_occ = homo
         IF (gw_corr_lev_virt > dimen-homo) gw_corr_lev_virt = dimen-homo
         IF (ispin == 1) THEN
            mp2_env%ri_g0w0%corr_mos_occ = gw_corr_lev_occ
            mp2_env%ri_g0w0%corr_mos_virt = gw_corr_lev_virt
         ELSE IF (ispin == 2) THEN
            ! ensure that the total number of corrected MOs is the same for alpha and beta, important
            ! for parallelization
            IF (mp2_env%ri_g0w0%corr_mos_occ+mp2_env%ri_g0w0%corr_mos_virt /= &
                gw_corr_lev_occ+gw_corr_lev_virt) THEN
               gw_corr_lev_virt = mp2_env%ri_g0w0%corr_mos_occ+mp2_env%ri_g0w0%corr_mos_virt-gw_corr_lev_occ
            END IF
            mp2_env%ri_g0w0%corr_mos_occ_beta = gw_corr_lev_occ
            mp2_env%ri_g0w0%corr_mos_virt_beta = gw_corr_lev_virt

         END IF

         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, matrix_sigma_x_minus_vxc(ispin)%matrix, &
                                mo_coeff_b, 0.0_dp, matrix_tmp, first_column=homo+1-gw_corr_lev_occ, &
                                last_column=homo+gw_corr_lev_virt)

         CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, mo_coeff_b, &
                                matrix_tmp, 0.0_dp, matrix_tmp_2, first_row=homo+1-gw_corr_lev_occ, &
                                last_row=homo+gw_corr_lev_virt)

         CALL cp_dbcsr_get_diag(matrix_tmp_2, vec_Sigma_x_minus_vxc_gw(:, ispin))

         CALL cp_dbcsr_set(matrix_tmp, 0.0_dp)
         CALL cp_dbcsr_set(matrix_tmp_2, 0.0_dp)

      END DO

      CALL mp_sum(vec_Sigma_x_minus_vxc_gw, para_env%group)

      CALL cp_dbcsr_release(mo_coeff_b)
      CALL cp_dbcsr_release(matrix_tmp)
      CALL cp_dbcsr_release(matrix_tmp_2)

      DO ispin = 1, SIZE(matrix_ks)
         CALL cp_dbcsr_release_p(matrix_sigma_x_minus_vxc(ispin)%matrix)
      END DO

      ALLOCATE (mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(nmo, dft_control%nspins))

      ! store vec_Sigma_x_minus_vxc_gw in the mp2_environment
      mp2_env%ri_g0w0%vec_Sigma_x_minus_vxc_gw(:, :) = vec_Sigma_x_minus_vxc_gw(:, :)

      ! clean up
      DEALLOCATE (matrix_sigma_x_minus_vxc, vec_Sigma_x_minus_vxc_gw)

      t2 = m_walltime()

      t3 = t2-t1

   END SUBROUTINE compute_vec_Sigma_x_minus_vxc_gw

END MODULE mp2

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Split and build its own idependent core_core SE interaction module
!> \author Teodoro Laino [tlaino] - 05.2009
!> \par History
!>      Teodoro Laino (05.2009) [tlaino] - create
! **************************************************************************************************
MODULE se_core_core
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_array_init,&
                                              atprop_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type,&
                                              semi_empirical_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_get,&
                                              ewald_pw_type
   USE input_constants,                 ONLY: do_method_am1,&
                                              do_method_mndo,&
                                              do_method_mndod,&
                                              do_method_pdg,&
                                              do_method_pm3,&
                                              do_method_pm6,&
                                              do_method_pnnl,&
                                              do_method_rm1
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE semi_empirical_int_arrays,       ONLY: rij_threshold
   USE semi_empirical_integrals,        ONLY: corecore,&
                                              dcorecore
   USE semi_empirical_types,            ONLY: get_se_param,&
                                              se_int_control_type,&
                                              se_taper_type,&
                                              semi_empirical_p_type,&
                                              semi_empirical_type,&
                                              setup_se_int_control_type
   USE semi_empirical_utils,            ONLY: finalize_se_taper,&
                                              get_se_type,&
                                              initialize_se_taper
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'se_core_core'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.

   PUBLIC :: se_core_core_interaction

CONTAINS

! **************************************************************************************************
!> \brief Evaluates the core-core interactions for NDDO methods
!> \param qs_env ...
!> \param para_env ...
!> \param calculate_forces ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE se_core_core_interaction(qs_env, para_env, calculate_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(in)                                :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'se_core_core_interaction', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, handle, iab, iatom, &
                                                            ikind, itype, jatom, jkind, natom, &
                                                            nkind
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: anag, atener, defined, use_virial
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: se_defined
      REAL(KIND=dp)                                      :: delta, dr1, dr3inv(3), enuc, enucij, &
                                                            enuclear, r2inv, r3inv, rinv
      REAL(KIND=dp), DIMENSION(3)                        :: force_ab, rij
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_se
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(se_int_control_type)                          :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_control_type), POINTER         :: se_control
      TYPE(semi_empirical_p_type), DIMENSION(:), POINTER :: se_kind_param
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a, se_kind_b
      TYPE(virial_type), POINTER                         :: virial

      enuclear = 0.0_dp
      NULLIFY (dft_control, cell, force, particle_set, se_control, se_taper, atomic_kind_set, &
               virial, atprop)

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("se_core_core.F",125)
      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, cell=cell, se_taper=se_taper, &
                      virial=virial, atprop=atprop, energy=energy)

      CALL initialize_se_taper(se_taper, coulomb=.TRUE.)
      ! Parameters
      se_control => dft_control%qs_control%se_control
      anag = se_control%analytical_gradients
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      CALL setup_se_int_control_type(se_int_control, do_ewald_r3=se_control%do_ewald_r3, &
                                     do_ewald_gks=se_control%do_ewald_gks, integral_screening=se_control%integral_screening, &
                                     shortrange=(se_control%do_ewald .OR. se_control%do_ewald_gks), &
                                     max_multipole=se_control%max_multipole, pc_coulomb_int=.FALSE.)

      ! atomic energy decomposition
      atener = atprop%energy
      IF (atener) THEN
         CALL get_qs_env(qs_env=qs_env, particle_set=particle_set)
         natom = SIZE(particle_set)
         CALL atprop_array_init(atprop%atecc, natom)
      END IF

      ! Retrieve some information if GKS ewald scheme is used
      IF (se_control%do_ewald_gks) THEN
         CALL get_qs_env(qs_env=qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw)
         CALL ewald_env_get(ewald_env, alpha=se_int_control%ewald_gks%alpha)
         CALL ewald_pw_get(ewald_pw, pw_big_pool=se_int_control%ewald_gks%pw_pool, &
                           dg=se_int_control%ewald_gks%dg)
         ! Virial not implemented
         IF(.NOT.(.NOT. use_virial))CALL cp__a("se_core_core.F",154)
      END IF

      CALL get_qs_env(qs_env=qs_env, sab_se=sab_se, atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set)

      nkind = SIZE(atomic_kind_set)
      ! Possibly compute forces
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, particle_set=particle_set, force=force)
         natom = SIZE(particle_set)
         ALLOCATE (atom_of_kind(natom))
         delta = se_control%delta
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)
      END IF

      itype = get_se_type(dft_control%qs_control%method_id)

      ALLOCATE (se_kind_param(nkind), se_defined(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
         se_kind_param(ikind)%se_param => se_kind_a
         CALL get_se_param(se_kind_a, defined=defined)
         se_defined(ikind) = defined
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_se)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, r=rij)
         IF (.NOT. se_defined(ikind)) CYCLE
         IF (.NOT. se_defined(jkind)) CYCLE
         se_kind_a => se_kind_param(ikind)%se_param
         se_kind_b => se_kind_param(jkind)%se_param
         iab = ikind+nkind*(jkind-1)
         dr1 = DOT_PRODUCT(rij, rij)
         enucij = 0._dp
         IF (dr1 > rij_threshold) THEN
            SELECT CASE (dft_control%qs_control%method_id)
            CASE (do_method_mndo, do_method_am1, do_method_pm3, do_method_pm6, do_method_pdg, &
                  do_method_rm1, do_method_mndod, do_method_pnnl)

               ! Core-Core energy term
               CALL corecore(se_kind_a, se_kind_b, rij, enuc=enuc, itype=itype, anag=anag, &
                             se_int_control=se_int_control, se_taper=se_taper)
               enucij = enucij+enuc
               ! Residual integral (1/R^3) correction
               IF (se_int_control%do_ewald_r3) THEN
                  r2inv = 1.0_dp/dr1
                  rinv = SQRT(r2inv)
                  r3inv = rinv**3
                  ! Core-Core term
                  enucij = enucij+se_kind_a%expns3_int(jkind)%expns3%core_core*r3inv
               END IF

               ! Core-Core Derivatives
               IF (calculate_forces) THEN
                  atom_a = atom_of_kind(iatom)
                  atom_b = atom_of_kind(jatom)

                  CALL dcorecore(se_kind_a, se_kind_b, rij, denuc=force_ab, itype=itype, delta=delta, &
                                 anag=anag, se_int_control=se_int_control, se_taper=se_taper)

                  ! Residual integral (1/R^3) correction
                  IF (se_int_control%do_ewald_r3) THEN
                     dr3inv = -3.0_dp*rij*r3inv*r2inv
                     ! Derivatives of core-core terms
                     force_ab = force_ab+se_kind_a%expns3_int(jkind)%expns3%core_core*dr3inv
                  END IF
                  IF (use_virial) THEN
                     CALL virial_pair_force(virial%pv_virial, -1.0_dp, force_ab, rij)
                  END IF

                  ! Sum up force components
                  force(ikind)%all_potential(1, atom_a) = force(ikind)%all_potential(1, atom_a)-force_ab(1)
                  force(jkind)%all_potential(1, atom_b) = force(jkind)%all_potential(1, atom_b)+force_ab(1)

                  force(ikind)%all_potential(2, atom_a) = force(ikind)%all_potential(2, atom_a)-force_ab(2)
                  force(jkind)%all_potential(2, atom_b) = force(jkind)%all_potential(2, atom_b)+force_ab(2)

                  force(ikind)%all_potential(3, atom_a) = force(ikind)%all_potential(3, atom_a)-force_ab(3)
                  force(jkind)%all_potential(3, atom_b) = force(jkind)%all_potential(3, atom_b)+force_ab(3)
               END IF
            CASE DEFAULT
               CALL cp__b("se_core_core.F",236,"")
            END SELECT
         ELSE
            IF (se_int_control%do_ewald_gks) THEN
               ! Core-Core energy term (self term in periodic systems)
               CALL corecore(se_kind_a, se_kind_b, rij, enuc=enuc, itype=itype, anag=anag, &
                             se_int_control=se_int_control, se_taper=se_taper)
               enucij = enucij+0.5_dp*enuc
            END IF
         END IF
         IF (atener) THEN
            atprop%atecc(iatom) = atprop%atecc(iatom)+0.5_dp*enucij
            atprop%atecc(jatom) = atprop%atecc(jatom)+0.5_dp*enucij
         END IF
         enuclear = enuclear+enucij
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (se_kind_param, se_defined)

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind)
      END IF

      CALL mp_sum(enuclear, para_env%group)
      energy%core_overlap = enuclear
      energy%core_overlap0 = enuclear

      CALL finalize_se_taper(se_taper)
      CALL timestop(handle)
   END SUBROUTINE se_core_core_interaction

END MODULE se_core_core

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the Hamiltonian integral matrix <a|H|b> for
!>      semi-empirical methods
!> \author JGH
! **************************************************************************************************
MODULE se_core_matrix
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_deallocate_matrix, &
        cp_dbcsr_distribute, cp_dbcsr_get_block_diag, cp_dbcsr_get_block_p, cp_dbcsr_init, &
        cp_dbcsr_p_type, cp_dbcsr_replicate_all, cp_dbcsr_set, cp_dbcsr_sum_replicated, &
        cp_dbcsr_type
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: do_method_am1,&
                                              do_method_mndo,&
                                              do_method_mndod,&
                                              do_method_pdg,&
                                              do_method_pm3,&
                                              do_method_pm6,&
                                              do_method_pnnl,&
                                              do_method_rm1
   USE input_section_types,             ONLY: section_vals_val_get
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: evolt
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_overlap,                      ONLY: build_overlap_matrix
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE semi_empirical_int_arrays,       ONLY: rij_threshold
   USE semi_empirical_types,            ONLY: get_se_param,&
                                              semi_empirical_type
   USE semi_empirical_utils,            ONLY: get_se_type
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'se_core_matrix'

   PUBLIC :: build_se_core_matrix

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param para_env ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE build_se_core_matrix(qs_env, para_env, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_se_core_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER :: after, atom_a, atom_b, handle, i, iatom, icol, icor, ikind, inode, irow, itype, &
         iw, j, jatom, jkind, natom, natorb_a, nkind, nr_a, nra, nrb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nrt
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: defined, found, omit_headers, use_virial
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: se_defined
      REAL(KIND=dp)                                      :: delta, dr, econst, eheat, eisol, kh, &
                                                            udd, uff, upp, uss, ZPa, ZPb, ZSa, ZSb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ZPt, ZSt
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: hmt, umt
      REAL(KIND=dp), DIMENSION(16)                       :: ha, hb, ua
      REAL(KIND=dp), DIMENSION(3)                        :: force_ab, rij
      REAL(KIND=dp), DIMENSION(:), POINTER               :: beta_a, sto_exponents_a
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dsmat, h_block, h_blocka, pabmat, pamat, &
                                                            s_block
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_h, matrix_p, matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: diagmat_h, diagmat_p
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a
      TYPE(virial_type), POINTER                         :: virial

!   REAL(KIND=dp), DIMENSION(3)              :: R

      CALL timeset(routineN, handle)

      NULLIFY (logger, energy)
      logger => cp_get_default_logger()

      NULLIFY (rho, force, atomic_kind_set, qs_kind_set, sab_orb, &
               diagmat_h, diagmat_p, particle_set, matrix_p, ks_env)

      CALL get_qs_env(qs_env, &
                      matrix_s=matrix_s, &
                      matrix_h=matrix_h, &
                      ks_env=ks_env, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control, &
                      energy=energy, &
                      force=force, &
                      virial=virial, &
                      rho=rho, &
                      sab_orb=sab_orb)

      ! calculate overlap matrix
      IF (calculate_forces) THEN
         CALL build_overlap_matrix(ks_env, nderivative=1, matrix_s=matrix_s, &
                                   matrix_name="OVERLAP", &
                                   basis_type_a="ORB", &
                                   basis_type_b="ORB", &
                                   sab_nl=sab_orb, &
                                   qs_input=qs_env%input)
         CALL set_ks_env(ks_env, matrix_s=matrix_s)
         use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      ELSE
         CALL build_overlap_matrix(ks_env, matrix_s=matrix_s, &
                                   matrix_name="OVERLAP", &
                                   basis_type_a="ORB", &
                                   basis_type_b="ORB", &
                                   sab_nl=sab_orb, &
                                   qs_input=qs_env%input)
         CALL set_ks_env(ks_env, matrix_s=matrix_s)
         use_virial = .FALSE.
      END IF

      IF (calculate_forces) THEN
         CALL qs_rho_get(rho, rho_ao=matrix_p)

         IF (SIZE(matrix_p) == 2) THEN
            CALL cp_dbcsr_add(matrix_p(1)%matrix, matrix_p(2)%matrix, alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
         END IF
         natom = SIZE(particle_set)
         ALLOCATE (atom_of_kind(natom))
         delta = dft_control%qs_control%se_control%delta
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  atom_of_kind=atom_of_kind)
         ALLOCATE (diagmat_p)
         CALL cp_dbcsr_init(diagmat_p)
         CALL cp_dbcsr_get_block_diag(matrix_p(1)%matrix, diagmat_p)
         CALL cp_dbcsr_replicate_all(diagmat_p)
      END IF

      ! Allocate the core Hamiltonian matrix
      CALL cp_dbcsr_allocate_matrix_set(matrix_h, 1)
      ALLOCATE (matrix_h(1)%matrix)
      CALL cp_dbcsr_init(matrix_h(1)%matrix)
      CALL cp_dbcsr_copy(matrix_h(1)%matrix, matrix_s(1)%matrix, "CORE HAMILTONIAN MATRIX")
      CALL cp_dbcsr_set(matrix_h(1)%matrix, 0.0_dp)

      ! Allocate a diagonal block matrix
      ALLOCATE (diagmat_h)
      CALL cp_dbcsr_init(diagmat_h)
      CALL cp_dbcsr_get_block_diag(matrix_s(1)%matrix, diagmat_h)
      CALL cp_dbcsr_set(diagmat_h, 0.0_dp)
      CALL cp_dbcsr_replicate_all(diagmat_h)

      ! kh might be set in qs_control
      itype = get_se_type(dft_control%qs_control%method_id)
      kh = 0.5_dp

      nkind = SIZE(atomic_kind_set)

      ALLOCATE (se_defined(nkind))
      ALLOCATE (hmt(16, nkind))
      ALLOCATE (umt(16, nkind))

      ALLOCATE (ZSt(nkind))
      ALLOCATE (ZPt(nkind))
      ALLOCATE (nrt(nkind))

      econst = 0.0_dp
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom)
         CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
         CALL get_se_param(se_kind_a, defined=defined, natorb=natorb_a, &
                           beta=beta_a, uss=uss, upp=upp, udd=udd, uff=uff, eisol=eisol, eheat=eheat, &
                           nr=nr_a, sto_exponents=sto_exponents_a)
         econst = econst-(eisol-eheat)*REAL(natom, dp)
         se_defined(ikind) = (defined .AND. natorb_a >= 1)
         hmt(1, ikind) = beta_a(0)
         hmt(2:4, ikind) = beta_a(1)
         hmt(5:9, ikind) = beta_a(2)
         hmt(10:16, ikind) = beta_a(3)
         umt(1, ikind) = uss
         umt(2:4, ikind) = upp
         umt(5:9, ikind) = udd
         umt(10:16, ikind) = uff

         ZSt(ikind) = sto_exponents_a(0)
         ZPt(ikind) = sto_exponents_a(1)
         nrt(ikind) = nr_a

      END DO
      energy%core_self = econst

      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, inode=inode, r=rij)
         IF (.NOT. se_defined(ikind)) CYCLE
         IF (.NOT. se_defined(jkind)) CYCLE
         ha(1:16) = hmt(1:16, ikind)
         ua(1:16) = umt(1:16, ikind)
         hb(1:16) = hmt(1:16, jkind)

         nra = nrt(ikind)
         nrb = nrt(jkind)
         ZSa = ZSt(ikind)
         ZSb = ZSt(jkind)
         ZPa = ZPt(ikind)
         ZPb = ZPt(jkind)

         IF (inode == 1) THEN
            SELECT CASE (dft_control%qs_control%method_id)
            CASE (do_method_am1, do_method_rm1, do_method_mndo, do_method_pdg, &
                  do_method_pm3, do_method_pm6, do_method_mndod, do_method_pnnl)
               NULLIFY (h_blocka)
               CALL cp_dbcsr_get_block_p(diagmat_h, iatom, iatom, h_blocka, found)
               IF(.NOT.(ASSOCIATED(h_blocka)))CALL cp__a("se_core_matrix.F",261)
               IF (calculate_forces) THEN
                  CALL cp_dbcsr_get_block_p(diagmat_p, iatom, iatom, pamat, found)
                  IF(.NOT.(ASSOCIATED(pamat)))CALL cp__a("se_core_matrix.F",264)
               END IF
            END SELECT
         END IF
         dr = SUM(rij(:)**2)
         IF (iatom == jatom .AND. dr < rij_threshold) THEN

            SELECT CASE (dft_control%qs_control%method_id)
            CASE DEFAULT
               CALL cp__b("se_core_matrix.F",273,"")
            CASE (do_method_am1, do_method_rm1, do_method_mndo, do_method_pdg, &
                  do_method_pm3, do_method_pm6, do_method_mndod, do_method_pnnl)
               DO i = 1, SIZE(h_blocka, 1)
                  h_blocka(i, i) = h_blocka(i, i)+ua(i)
               END DO
            END SELECT

         ELSE
            IF (iatom <= jatom) THEN
               irow = iatom
               icol = jatom
            ELSE
               irow = jatom
               icol = iatom
            END IF
            NULLIFY (h_block)
            CALL cp_dbcsr_get_block_p(matrix_h(1)%matrix, &
                                      irow, icol, h_block, found)
            IF(.NOT.(ASSOCIATED(h_block)))CALL cp__a("se_core_matrix.F",292)
            ! two-centre one-electron term
            NULLIFY (s_block)

!          CALL cp_dbcsr_get_block_p(matrix_s(1)%matrix,&
!               irow,icol,s_block,found)
!          CPPostcondition(ASSOCIATED(s_block),cp_failure_level,routineP,failure)
!
!          if( irow == iatom )then
!            R= -rij
!            call makeS(R,nra,nrb,ZSa,ZSb,ZPa,ZPb,S)
!          else
!            R= rij
!            call makeS(R,nrb,nra,ZSb,ZSa,ZPb,ZPa,S)
!          endif
!
!          do i=1,4
!            do j=1,4
!              s_block(i,j)=S(ix(i),ix(j))
!            enddo
!          enddo

            CALL cp_dbcsr_get_block_p(matrix_s(1)%matrix, &
                                      irow, icol, s_block, found)
            IF(.NOT.(ASSOCIATED(s_block)))CALL cp__a("se_core_matrix.F",316)
            IF (irow == iatom) THEN
               DO i = 1, SIZE(h_block, 1)
                  DO j = 1, SIZE(h_block, 2)
                     h_block(i, j) = h_block(i, j)+kh*(ha(i)+hb(j))*s_block(i, j)
                  END DO
               END DO
            ELSE
               DO i = 1, SIZE(h_block, 1)
                  DO j = 1, SIZE(h_block, 2)
                     h_block(i, j) = h_block(i, j)+kh*(ha(j)+hb(i))*s_block(i, j)
                  END DO
               END DO
            END IF
            IF (calculate_forces) THEN
               atom_a = atom_of_kind(iatom)
               atom_b = atom_of_kind(jatom)

!            if( irow == iatom )then
!              R= -rij
!              call makedS(R,nra,nrb,ZSa,ZSb,ZPa,ZPb,dS)
!            else
!              R= rij
!              call makedS(R,nrb,nra,ZSb,ZSa,ZPb,ZPa,dS)
!            endif

               CALL cp_dbcsr_get_block_p(matrix_p(1)%matrix, irow, icol, pabmat, found)
               IF(.NOT.(ASSOCIATED(pabmat)))CALL cp__a("se_core_matrix.F",343)
               DO icor = 1, 3
                  force_ab(icor) = 0._dp

!              CALL cp_dbcsr_get_block_p(matrix_s(icor+1)%matrix,irow,icol,dsmat,found)
!              CPPostcondition(ASSOCIATED(dsmat),cp_failure_level,routineP,failure)
!
!              do i=1,4
!                do j=1,4
!                  dsmat(i,j)=dS(ix(i),ix(j),icor)
!                enddo
!              enddo

                  CALL cp_dbcsr_get_block_p(matrix_s(icor+1)%matrix, irow, icol, dsmat, found)
                  IF(.NOT.(ASSOCIATED(dsmat)))CALL cp__a("se_core_matrix.F",357)
                  dsmat = 2._dp*kh*dsmat*pabmat
                  IF (irow == iatom) THEN
                     DO i = 1, SIZE(h_block, 1)
                        DO j = 1, SIZE(h_block, 2)
                           force_ab(icor) = force_ab(icor)+(ha(i)+hb(j))*dsmat(i, j)
                        END DO
                     END DO
                  ELSE
                     DO i = 1, SIZE(h_block, 1)
                        DO j = 1, SIZE(h_block, 2)
                           force_ab(icor) = force_ab(icor)+(ha(j)+hb(i))*dsmat(i, j)
                        END DO
                     END DO
                  END IF
               END DO
            END IF

         END IF

         IF (calculate_forces .AND. (iatom /= jatom .OR. dr > rij_threshold)) THEN
            IF (irow == iatom) force_ab = -force_ab
            force(ikind)%all_potential(:, atom_a) = &
               force(ikind)%all_potential(:, atom_a)-force_ab(:)
            force(jkind)%all_potential(:, atom_b) = &
               force(jkind)%all_potential(:, atom_b)+force_ab(:)
            IF (use_virial) THEN
               CALL virial_pair_force(virial%pv_virial, -1.0_dp, force_ab, rij)
            END IF
         END IF

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (se_defined, hmt, umt, ZSt, ZPt, nrt)

      CALL cp_dbcsr_sum_replicated(diagmat_h)
      CALL cp_dbcsr_distribute(diagmat_h)
      CALL cp_dbcsr_add(matrix_h(1)%matrix, diagmat_h, 1.0_dp, 1.0_dp)
      CALL set_ks_env(ks_env, matrix_h=matrix_h)

      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "DFT%PRINT%AO_MATRICES/CORE_HAMILTONIAN"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/CORE_HAMILTONIAN", &
                                   extension=".Log")
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
         CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
         after = MIN(MAX(after, 1), 16)
         CALL cp_dbcsr_write_sparse_matrix(matrix_h(1)%matrix, 4, after, qs_env, para_env, &
                                           scale=evolt, output_unit=iw, omit_headers=omit_headers)
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "DFT%PRINT%AO_MATRICES/CORE_HAMILTONIAN")
      END IF

      IF (calculate_forces) THEN
         IF (SIZE(matrix_p) == 2) THEN
            CALL cp_dbcsr_add(matrix_p(1)%matrix, matrix_p(2)%matrix, alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
         END IF
         DEALLOCATE (atom_of_kind)
         CALL cp_dbcsr_deallocate_matrix(diagmat_p)
      END IF

      CALL cp_dbcsr_deallocate_matrix(diagmat_h)

      CALL timestop(handle)

   END SUBROUTINE build_se_core_matrix

! **************************************************************************************************
!> \brief ...
!> \param R ...
!> \param nra ...
!> \param nrb ...
!> \param ZSA ...
!> \param ZSB ...
!> \param ZPA ...
!> \param ZPB ...
!> \param S ...
! **************************************************************************************************
   SUBROUTINE makeS(R, nra, nrb, ZSA, ZSB, ZPA, ZPB, S)

      REAL(kind=dp), DIMENSION(3)                        :: R
      INTEGER                                            :: nra, nrb
      REAL(kind=dp)                                      :: ZSA, ZSB, ZPA, ZPB
      REAL(kind=dp), DIMENSION(4, 4)                     :: S

      INTEGER, DIMENSION(4, 4), PARAMETER :: &
         nc1 = RESHAPE((/2, 4, 4, 6, 4, 3, 6, 7, 4, 6, 4, 8, 6, 7, 8, 5/), (/4, 4/)), &
         nc2 = RESHAPE((/4, 4, 8, 8, 6, 8, 6, 12, 8, 8, 12, 8, 10, 12, 14, 16/), (/4, 4/)), &
         nc3 = RESHAPE((/4, 6, 8, 10, 4, 8, 8, 12, 8, 6, 12, 14, 8, 12, 8, 16/), (/4, 4/)), &
         nc4 = RESHAPE((/4, 8, 11, 14, 8, 6, 12, 14, 11, 12, 10, 20, 14, 14, 20, 12/), (/4, 4/)), &
         nc5 = RESHAPE((/2, 4, 6, 8, 4, 4, 8, 8, 6, 8, 6, 12, 8, 8, 12, 8/), (/4, 4/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c1 = RESHAPE((/1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 &
         , 1, 1, 1, 1, -1, 1, 2, 3, -1, -2, 1, 2, -2, -1, -3, 1, -3, -2, -1, -4, 0, -1, -2, 2, -1, &
         1, -2, -1, 2, -2, 3, -3, 2, -1, -3, 6, 0, -1, -1, -2, 1, 0, -2, -4, -1, 2, -1, -3, 2, 4, 3&
         , -4, 0, 0, 0, -3, 0, 0, 1, -1, 0, 1, 0, 3, -3, -1, 3, 1, 0, 0, 0, -1, 0, 0, 1, 2, 0, -1, &
         0, 3, 1, -2, -3, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, -1, 0, 1, -1, 0, 0, 0, 0, 0, 0, 0, 0 &
         , 0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c2 = RESHAPE((/1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 &
         , 1, 1, 1, 1, -1, 1, 1, 2, -2, -1, 1, 1, -3, -2, -1, 1, -4, -3, -2, -1, 1, -1, 1, 1, 1, 1 &
         , -2, 1, 1, 1, 1, -3, 1, 1, 1, 1, -1, -1, -1, 2, 1, -1, -2, -2, 3, -2, -2, -3, 6, 2, -1, &
         -3, 0, 0, 1, -2, -2, -1, 1, 1, -3, 2, -1, 3, -4, -3, -2, -1, 0, 0, -1, -1, 1, 1, 1, -2, -1&
         , -1, 2, 3, -4, 2, 4, 3, 0, 0, -1, -2, 0, -1, 0, -2, 3, 2, -2, -1, 6, 2, -1, -3, 0, 0, -1,&
         -1, 0, 1, 0, 1, -1, -1, 1, -1, 1, -3, -1, 3, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, 2, 0, -4, 2, 4&
         , 3, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 1, 1, -2, -3, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0&
         , 0, -3, -1, 3, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 1, 1, -1, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, -2, -3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c3 = RESHAPE((/-1, -1, -1, -1, -1, -1, -1, -1, -1 &
         , -1, -1, -1, -1, -1, -1, -1, -1, -2, -3, -4, 1, -1, -2, -3, 1, 1, -1, -2, 2, 1, 1, -1, 1 &
         , 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 3, 1, 1, -1, -3, -6, -1, 1, 2, -2, 1, -2, 2, 1, &
         -2, 2, -3, 3, 0, 2, 3, 4, 0, 1, 2, 3, -1, -1, 1, 2, -2, -1, -3, 1, 0, 1, -1, -4, 0, 1, 1, &
         2, -1, 1, 2, 4, 1, -2, 3, 3, 0, 0, 3, 6, 0, -1, -2, 2, -1, 0, -2, -1, 2, -2, 1, -3, 0, 0, &
         1, -1, 0, -1, -1, 3, 1, 0, -1, 1, -1, -1, -1, -3, 0, 0, 0, 4, 0, 0, 0, -2, 0, 0, -2, -4, 0&
         , 2, 0, -3, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, -1, -2, 0, 1, 0, -3, 0, 0, 0, 0, 0, 0, 0, -3, 0 &
         , 0, 1, -1, 0, 1, 0, 3, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, -1, 0, -1, 0, 1, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c4 = RESHAPE((/-1, -1, -1, -1, -1, -1, -1, -1, -1 &
         , -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -2, &
         -3, 1, 1, -1, -2, 2, 1, 2, -1, 3, 2, 1, 3, -1, 1, 2, 3, -1, -1, 1, 2, -2, -1, -1, 1, -3, &
         -2, -1, -2, 0, 1, -1, -3, 1, -1, 1, 1, -1, 1, -1, 2, -3, 1, 2, -1, 0, -1, 2, 4, -1, 1, -1 &
         , -1, 2, -1, -1, -1, 4, -1, -1, -3, 0, 1, -1, -1, -1, 0, 1, 2, -1, -1, -1, -1, -1, -2, -1 &
         , 3, 0, -1, 2, -1, 1, 0, -1, -2, -2, 1, 2, 2, 1, 2, -2, 1, 0, 0, -2, 4, 0, 0, -1, 1, 2, -1&
         , 1, -1, -4, 1, 1, 2, 0, 0, 1, -3, 0, 0, 1, -1, 1, 1, -1, -1, 3, -1, 1, -3, 0, 0, -1, 3, 0&
         , 0, -1, -2, -1, 1, 0, -1, 3, 2, -1, -1, 0, 0, 0, -3, 0, 0, 1, 2, 0, -1, 0, -1, -3, -2, -1&
         , 1, 0, 0, 0, 1, 0, 0, 0, -1, 0, 0, 0, 2, -1, -1, 2, 0, 0, 0, 0, -1, 0, 0, 0, 1, 0, 0, 0, &
         -1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 2, 0, 0, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c5 = RESHAPE((/-1, -1, -1, -1, -1, -1, -1, -1, -1 &
         , -1, -1, -1, -1, -1, -1, -1, 1, -1, -2, -3, 1, 1, -1, -2, 2, 1, 2, -1, 3, 2, 1, 3, 0, 1, &
         -1, -3, 1, 1, 1, 1, -1, 1, 1, 2, -3, 1, 2, 1, 0, 1, 1, 1, -1, -1, 1, 2, 1, 1, -1, 1, 1, -2&
         , 1, -3, 0, 0, 2, -1, 0, 0, 1, 2, -2, -1, -2, 2, 1, -2, -2, -3, 0, 0, 1, 3, 0, 0, 1, 1, 1,&
         -1, 1, 1, -3, 1, -1, 1, 0, 0, 0, 3, 0, 0, -1, -2, 0, -1, 0, -1, 3, 2, -1, 3, 0, 0, 0, 1, 0&
         , 0, -1, -1, 0, 1, 0, -2, -1, -1, -2, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0,&
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0,&
         1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, &
         20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma1 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4, 4, 6, 4, 5, 6, 6, 0, 1, 1, 3, 1, 0, 3, 4, 1, 3&
         , 2, 5, 3, 4, 5, 4, 0, 0, 0, 2, 0, 0, 2, 3, 0, 2, 0, 4, 2, 3, 4, 2, 0, 0, 0, 1, 0, 0, 1, 2&
         , 0, 1, 0, 3, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 2, 0, 1, 2, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma2 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 1, 2, 3, 4, 2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7, 1, 1, 3, 4, 2, 3, 3, 5, 3, 4&
         , 5, 5, 4, 5, 6, 7, 0, 0, 2, 3, 1, 2, 2, 4, 2, 3, 4, 4, 3, 4, 5, 6, 0, 0, 2, 2, 1, 2, 1, 4&
         , 2, 2, 4, 3, 3, 4, 5, 6, 0, 0, 1, 1, 0, 1, 0, 3, 1, 1, 3, 2, 2, 3, 4, 5, 0, 0, 1, 1, 0, 1&
         , 0, 3, 1, 1, 3, 1, 2, 3, 4, 5, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 1, 2, 3, 4, 0, 0, 0, 0&
         , 0, 0, 0, 2, 0, 0, 2, 0, 1, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 2, 3, 0, 0&
         , 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma3 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 1, 2, 3, 4, 2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7, 1, 2, 3, 4, 1, 3, 4, 5, 3, 3&
         , 5, 6, 4, 5, 5, 7, 0, 1, 2, 3, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4, 4, 6, 0, 1, 2, 3, 0, 2, 2, 4&
         , 2, 1, 4, 5, 2, 4, 3, 6, 0, 0, 1, 2, 0, 1, 1, 3, 1, 0, 3, 4, 1, 3, 2, 5, 0, 0, 1, 2, 0, 1&
         , 1, 3, 1, 0, 3, 4, 1, 3, 1, 5, 0, 0, 0, 1, 0, 0, 0, 2, 0, 0, 2, 3, 0, 2, 0, 4, 0, 0, 0, 1&
         , 0, 0, 0, 2, 0, 0, 2, 3, 0, 2, 0, 4, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 2, 0, 1, 0, 3, 0, 0&
         , 0, 0, 0, 0, 0, 1, 0, 0, 1, 2, 0, 1, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 2&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma4 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7, 5, 6, 7, 8, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4&
         , 4, 6, 4, 5, 6, 6, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4, 4, 6, 4, 5, 6, 6, 0, 1, 2, 3, 1, 0, 3, 4&
         , 2, 3, 4, 5, 3, 4, 5, 6, 0, 1, 2, 3, 1, 0, 3, 4, 2, 3, 2, 5, 3, 4, 5, 4, 0, 0, 2, 3, 0, 0&
         , 2, 3, 2, 2, 2, 5, 3, 3, 5, 4, 0, 0, 1, 2, 0, 0, 2, 3, 1, 2, 2, 4, 2, 3, 4, 2, 0, 0, 1, 2&
         , 0, 0, 1, 2, 1, 1, 0, 4, 2, 2, 4, 2, 0, 0, 0, 2, 0, 0, 1, 2, 0, 1, 0, 4, 2, 2, 4, 2, 0, 0&
         , 0, 1, 0, 0, 0, 1, 0, 0, 0, 3, 1, 1, 3, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 3, 1, 1, 3, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0&
         , 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2&
         , 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma5 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4, 4, 6, 4, 5, 6, 6, 0, 1, 2, 3, 1, 2, 3, 4, 2, 3&
         , 4, 5, 3, 4, 5, 6, 0, 0, 2, 3, 0, 0, 3, 3, 2, 3, 2, 5, 3, 3, 5, 4, 0, 0, 1, 2, 0, 0, 2, 3&
         , 1, 2, 2, 4, 2, 3, 4, 4, 0, 0, 0, 2, 0, 0, 2, 2, 0, 2, 0, 4, 2, 2, 4, 2, 0, 0, 0, 1, 0, 0&
         , 1, 1, 0, 1, 0, 3, 1, 1, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 3, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb1 = RESHAPE((/0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 0, 2, 3, 2, 2, 4, 2, 2, 3, 2&
         , 4, 2, 2, 2, 2, 4, 0, 3, 4, 3, 3, 0, 3, 3, 4, 3, 6, 3, 3, 3, 3, 6, 0, 0, 0, 4, 0, 0, 4, 4&
         , 0, 4, 0, 4, 4, 4, 4, 8, 0, 0, 0, 5, 0, 0, 5, 5, 0, 5, 0, 5, 5, 5, 5, 0, 0, 0, 0, 0, 0, 0&
         , 0, 6, 0, 0, 0, 6, 0, 6, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 7, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb2 = RESHAPE((/1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1&
         , 1, 1, 1, 1, 2, 0, 2, 2, 2, 2, 0, 2, 2, 2, 2, 0, 2, 2, 2, 2, 0, 3, 0, 0, 0, 0, 3, 0, 0, 0&
         , 0, 3, 0, 0, 0, 0, 1, 2, 3, 1, 3, 3, 2, 3, 3, 1, 3, 2, 3, 3, 3, 3, 0, 0, 1, 4, 1, 1, 5, 1&
         , 1, 4, 1, 5, 1, 1, 1, 1, 0, 0, 4, 5, 2, 4, 4, 4, 4, 5, 4, 4, 4, 4, 4, 4, 0, 0, 2, 3, 0, 2&
         , 0, 2, 2, 3, 2, 7, 2, 2, 2, 2, 0, 0, 3, 4, 0, 3, 0, 5, 3, 4, 5, 6, 5, 5, 5, 5, 0, 0, 0, 0&
         , 0, 0, 0, 3, 0, 0, 3, 0, 3, 3, 3, 3, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 6, 0, 4, 6, 6, 6, 0, 0&
         , 0, 0, 0, 0, 0, 4, 0, 0, 4, 0, 0, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 0, 0, 5, 7, 7&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 6, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb3 = RESHAPE((/1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1&
         , 1, 1, 1, 1, 2, 2, 2, 2, 0, 2, 2, 2, 2, 0, 2, 2, 2, 2, 0, 2, 0, 0, 0, 0, 3, 0, 0, 0, 0, 3&
         , 0, 0, 0, 0, 3, 0, 1, 3, 3, 3, 2, 3, 1, 3, 3, 2, 3, 3, 1, 3, 2, 3, 0, 1, 1, 1, 0, 1, 4, 1&
         , 1, 5, 1, 1, 4, 1, 5, 1, 0, 2, 4, 4, 0, 4, 5, 4, 4, 4, 4, 4, 5, 4, 4, 4, 0, 0, 2, 2, 0, 2&
         , 3, 2, 2, 0, 2, 2, 3, 2, 7, 2, 0, 0, 3, 5, 0, 3, 4, 5, 3, 0, 5, 5, 4, 5, 6, 5, 0, 0, 0, 3&
         , 0, 0, 0, 3, 0, 0, 3, 3, 0, 3, 0, 3, 0, 0, 0, 4, 0, 0, 0, 6, 0, 0, 6, 6, 0, 6, 0, 6, 0, 0&
         , 0, 0, 0, 0, 0, 4, 0, 0, 4, 4, 0, 4, 0, 4, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 7, 0, 5, 0, 7&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0&
         , 0, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb4 = RESHAPE((/2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2&
         , 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 3, 3, 3, 4, 3, 3, 3, 3&
         , 4, 3, 3, 3, 3, 4, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 0, 2, 4, 4, 2, 4, 4, 2&
         , 4, 4, 0, 4, 4, 2, 4, 0, 0, 0, 2, 2, 0, 2, 0, 0, 2, 0, 6, 2, 2, 0, 2, 6, 0, 3, 0, 0, 3, 0&
         , 5, 5, 0, 5, 4, 0, 0, 5, 0, 2, 0, 1, 3, 5, 1, 0, 1, 1, 3, 1, 2, 5, 5, 1, 5, 8, 0, 0, 1, 3&
         , 0, 0, 4, 6, 1, 4, 6, 3, 3, 6, 3, 6, 0, 0, 4, 1, 0, 0, 2, 4, 4, 2, 4, 1, 1, 4, 1, 4, 0, 0&
         , 2, 4, 0, 0, 5, 5, 2, 5, 0, 6, 4, 5, 6, 8, 0, 0, 0, 2, 0, 0, 3, 3, 0, 3, 0, 4, 2, 3, 4, 6&
         , 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 2, 5, 6, 2, 0, 0, 0, 0, 3, 0, 0, 0, 4, 0, 0, 0, 7, 3, 4&
         , 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3&
         , 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 4, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 5, 0, 0, 5, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb5 = RESHAPE((/2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2&
         , 2, 2, 2, 2, 0, 3, 3, 3, 3, 4, 3, 3, 3, 3, 4, 3, 3, 3, 3, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4&
         , 0, 4, 4, 0, 4, 0, 0, 1, 0, 0, 1, 2, 0, 5, 0, 0, 6, 0, 0, 5, 0, 6, 0, 0, 1, 5, 0, 0, 5, 1&
         , 1, 5, 2, 5, 5, 1, 5, 2, 0, 0, 2, 1, 0, 0, 1, 6, 2, 1, 4, 1, 1, 6, 1, 8, 0, 0, 0, 2, 0, 0&
         , 2, 3, 0, 2, 0, 6, 2, 3, 6, 4, 0, 0, 0, 3, 0, 0, 3, 4, 0, 3, 0, 2, 3, 4, 2, 6, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 3, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))

      INTEGER                                            :: k, k1, k2, mu
      REAL(kind=dp)                                      :: cp, ct, fac1, fac2, J, Jc, Jcc, Jss, rr, &
                                                            sp, st, xx, yy, za, zb
      REAL(kind=dp), DIMENSION(3)                        :: v
      REAL(kind=dp), DIMENSION(3, 3)                     :: Arot

      S(:, :) = 0.0_dp

      v(:) = R(:)
      rr = SQRT(DOT_PRODUCT(v, v))

      IF (rr < 1.0e-20_dp) THEN

         DO mu = 1, 4
            S(mu, mu) = 1.0_dp
         ENDDO

      ELSE

         fac1 = 1.0_dp
         IF (nra == 1) THEN
            fac1 = fac1*2.0_dp
         ELSE
            IF (nra == 2) THEN
               fac1 = fac1*SQRT(4.0_dp/3.0_dp)
            ELSE
               IF (nra == 3) THEN
                  fac1 = fac1*SQRT(8.0_dp/45.0_dp)
               ELSE
                  IF (nra == 4) THEN
                     fac1 = fac1*SQRT(4.0_dp/315.0_dp)
                  ELSE
                     WRITE (*, *) 'nra= ', nra
                     RETURN
                  ENDIF
               ENDIF
            ENDIF
         ENDIF
         IF (nrb == 1) THEN
            fac1 = fac1*2.0_dp
         ELSE
            IF (nrb == 2) THEN
               fac1 = fac1*SQRT(4.0_dp/3.0_dp)
            ELSE
               IF (nrb == 3) THEN
                  fac1 = fac1*SQRT(8.0_dp/45.0_dp)
               ELSE
                  IF (nrb == 4) THEN
                     fac1 = fac1*SQRT(4.0_dp/315.0_dp)
                  ELSE
                     WRITE (*, *) 'nrb= ', nrb
                     RETURN
                  ENDIF
               ENDIF
            ENDIF
         ENDIF

         ct = -v(3)/rr
         IF (ABS(ct) < 1.0_dp) THEN
            st = SQRT(1.0_dp-ct**2)
            cp = -v(1)/(rr*st)
            sp = -v(2)/(rr*st)
            Arot(1, 1) = ct*cp
            Arot(1, 2) = -sp
            Arot(1, 3) = st*cp
            Arot(2, 1) = ct*sp
            Arot(2, 2) = cp
            Arot(2, 3) = st*sp
            Arot(3, 1) = -st
            Arot(3, 2) = 0.0_dp
            Arot(3, 3) = ct
         ELSE
            Arot(1, 1) = ct
            Arot(1, 2) = 0.0_dp
            Arot(1, 3) = 0.0_dp
            Arot(2, 1) = 0.0_dp
            Arot(2, 2) = 1.0_dp
            Arot(2, 3) = 0.0_dp
            Arot(3, 1) = 0.0_dp
            Arot(3, 2) = 0.0_dp
            Arot(3, 3) = ct
         ENDIF

         za = ZSA
         zb = ZSB
         fac2 = SQRT(za**(2*nra+1)*zb**(2*nrb+1))
         xx = 0.5_dp*rr*(za+zb)
         yy = 0.5_dp*rr*(za-zb)

         J = 0.0_dp
         DO k = 1, nc1(nra, nrb)
            J = J+REAL(c1(nra, nrb, k), dp)*AA(ma1(nra, nrb, k), xx)*BB(mb1(nra, nrb, k), yy)
         ENDDO
         J = J*rr**(nra+nrb+1)
         J = J/2.0_dp**(nra+nrb+2)

         S(1, 1) = S(1, 1)+fac1*fac2*J

         za = ZPA
         zb = ZSB
         fac2 = SQRT(za**(2*nra+1)*zb**(2*nrb+1))
         xx = 0.5_dp*rr*(za+zb)
         yy = 0.5_dp*rr*(za-zb)

         Jc = 0.0_dp
         DO k = 1, nc2(nra, nrb)
            Jc = Jc+REAL(c2(nra, nrb, k), dp)*AA(ma2(nra, nrb, k), xx)*BB(mb2(nra, nrb, k), yy)
         ENDDO
         Jc = Jc*rr**(nra+nrb+1)
         Jc = Jc/2.0_dp**(nra+nrb+2)

         DO k1 = 1, 3
            S(k1+1, 1) = S(k1+1, 1) &
         &          +SQRT(3.0_dp)*Arot(k1, 3)*fac1*fac2*Jc
         ENDDO

         za = ZSA
         zb = ZPB
         fac2 = SQRT(za**(2*nra+1)*zb**(2*nrb+1))
         xx = 0.5_dp*rr*(za+zb)
         yy = 0.5_dp*rr*(za-zb)

         Jc = 0.0_dp
         DO k = 1, nc3(nra, nrb)
            Jc = Jc+REAL(c3(nra, nrb, k), dp)*AA(ma3(nra, nrb, k), xx)*BB(mb3(nra, nrb, k), yy)
         ENDDO
         Jc = Jc*rr**(nra+nrb+1)
         Jc = Jc/2.0_dp**(nra+nrb+2)

         DO k1 = 1, 3
            S(1, k1+1) = S(1, k1+1) &
         &          -SQRT(3.0_dp)*Arot(k1, 3)*fac1*fac2*Jc
         ENDDO

         za = ZPA
         zb = ZPB
         fac2 = SQRT(za**(2*nra+1)*zb**(2*nrb+1))
         xx = 0.5_dp*rr*(za+zb)
         yy = 0.5_dp*rr*(za-zb)

         Jss = 0.0_dp
         DO k = 1, nc4(nra, nrb)
            Jss = Jss+REAL(c4(nra, nrb, k), dp)*AA(ma4(nra, nrb, k), xx)*BB(mb4(nra, nrb, k), yy)
         ENDDO
         Jss = Jss*rr**(nra+nrb+1)
         Jss = Jss/2.0_dp**(nra+nrb+2)

         Jcc = 0.0_dp
         DO k = 1, nc5(nra, nrb)
            Jcc = Jcc+REAL(c5(nra, nrb, k), dp)*AA(ma5(nra, nrb, k), xx)*BB(mb5(nra, nrb, k), yy)
         ENDDO
         Jcc = Jcc*rr**(nra+nrb+1)
         Jcc = Jcc/2.0_dp**(nra+nrb+2)

         DO k1 = 1, 3
            DO k2 = 1, 3
               S(k1+1, k2+1) = S(k1+1, k2+1) &
          &            +1.5_dp*Arot(k1, 1)*Arot(k2, 1)*fac1*fac2*Jss &
          &            +1.5_dp*Arot(k1, 2)*Arot(k2, 2)*fac1*fac2*Jss &
          &            -3.0_dp*Arot(k1, 3)*Arot(k2, 3)*fac1*fac2*Jcc
            ENDDO
         ENDDO

      ENDIF

   END SUBROUTINE makeS

! **************************************************************************************************
!> \brief ...
!> \param R ...
!> \param nra ...
!> \param nrb ...
!> \param ZSA ...
!> \param ZSB ...
!> \param ZPA ...
!> \param ZPB ...
!> \param dS ...
! **************************************************************************************************
   SUBROUTINE makedS(R, nra, nrb, ZSA, ZSB, ZPA, ZPB, dS)

      REAL(kind=dp), DIMENSION(3)                        :: R
      INTEGER                                            :: nra, nrb
      REAL(kind=dp)                                      :: ZSA, ZSB, ZPA, ZPB
      REAL(kind=dp), DIMENSION(4, 4, 3)                  :: dS

      INTEGER, DIMENSION(4, 4), PARAMETER :: &
         nc1 = RESHAPE((/2, 4, 4, 6, 4, 3, 6, 7, 4, 6, 4, 8, 6, 7, 8, 5/), (/4, 4/)), &
         nc2 = RESHAPE((/4, 4, 8, 8, 6, 8, 6, 12, 8, 8, 12, 8, 10, 12, 14, 16/), (/4, 4/)), &
         nc3 = RESHAPE((/4, 6, 8, 10, 4, 8, 8, 12, 8, 6, 12, 14, 8, 12, 8, 16/), (/4, 4/)), &
         nc4 = RESHAPE((/4, 8, 11, 14, 8, 6, 12, 14, 11, 12, 10, 20, 14, 14, 20, 12/), (/4, 4/)), &
         nc5 = RESHAPE((/2, 4, 6, 8, 4, 4, 8, 8, 6, 8, 6, 12, 8, 8, 12, 8/), (/4, 4/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c1 = RESHAPE((/1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 &
         , 1, 1, 1, 1, -1, 1, 2, 3, -1, -2, 1, 2, -2, -1, -3, 1, -3, -2, -1, -4, 0, -1, -2, 2, -1, &
         1, -2, -1, 2, -2, 3, -3, 2, -1, -3, 6, 0, -1, -1, -2, 1, 0, -2, -4, -1, 2, -1, -3, 2, 4, 3&
         , -4, 0, 0, 0, -3, 0, 0, 1, -1, 0, 1, 0, 3, -3, -1, 3, 1, 0, 0, 0, -1, 0, 0, 1, 2, 0, -1, &
         0, 3, 1, -2, -3, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, -1, 0, 1, -1, 0, 0, 0, 0, 0, 0, 0, 0 &
         , 0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c2 = RESHAPE((/1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 &
         , 1, 1, 1, 1, -1, 1, 1, 2, -2, -1, 1, 1, -3, -2, -1, 1, -4, -3, -2, -1, 1, -1, 1, 1, 1, 1 &
         , -2, 1, 1, 1, 1, -3, 1, 1, 1, 1, -1, -1, -1, 2, 1, -1, -2, -2, 3, -2, -2, -3, 6, 2, -1, &
         -3, 0, 0, 1, -2, -2, -1, 1, 1, -3, 2, -1, 3, -4, -3, -2, -1, 0, 0, -1, -1, 1, 1, 1, -2, -1&
         , -1, 2, 3, -4, 2, 4, 3, 0, 0, -1, -2, 0, -1, 0, -2, 3, 2, -2, -1, 6, 2, -1, -3, 0, 0, -1,&
         -1, 0, 1, 0, 1, -1, -1, 1, -1, 1, -3, -1, 3, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, 2, 0, -4, 2, 4&
         , 3, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 1, 1, -2, -3, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0&
         , 0, -3, -1, 3, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 1, 1, -1, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, -2, -3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &
         0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c3 = RESHAPE((/-1, -1, -1, -1, -1, -1, -1, -1, -1 &
         , -1, -1, -1, -1, -1, -1, -1, -1, -2, -3, -4, 1, -1, -2, -3, 1, 1, -1, -2, 2, 1, 1, -1, 1 &
         , 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 3, 1, 1, -1, -3, -6, -1, 1, 2, -2, 1, -2, 2, 1, &
         -2, 2, -3, 3, 0, 2, 3, 4, 0, 1, 2, 3, -1, -1, 1, 2, -2, -1, -3, 1, 0, 1, -1, -4, 0, 1, 1, &
         2, -1, 1, 2, 4, 1, -2, 3, 3, 0, 0, 3, 6, 0, -1, -2, 2, -1, 0, -2, -1, 2, -2, 1, -3, 0, 0, &
         1, -1, 0, -1, -1, 3, 1, 0, -1, 1, -1, -1, -1, -3, 0, 0, 0, 4, 0, 0, 0, -2, 0, 0, -2, -4, 0&
         , 2, 0, -3, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, -1, -2, 0, 1, 0, -3, 0, 0, 0, 0, 0, 0, 0, -3, 0 &
         , 0, 1, -1, 0, 1, 0, 3, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, -1, 0, -1, 0, 1, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c4 = RESHAPE((/-1, -1, -1, -1, -1, -1, -1, -1, -1 &
         , -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -2, &
         -3, 1, 1, -1, -2, 2, 1, 2, -1, 3, 2, 1, 3, -1, 1, 2, 3, -1, -1, 1, 2, -2, -1, -1, 1, -3, &
         -2, -1, -2, 0, 1, -1, -3, 1, -1, 1, 1, -1, 1, -1, 2, -3, 1, 2, -1, 0, -1, 2, 4, -1, 1, -1 &
         , -1, 2, -1, -1, -1, 4, -1, -1, -3, 0, 1, -1, -1, -1, 0, 1, 2, -1, -1, -1, -1, -1, -2, -1 &
         , 3, 0, -1, 2, -1, 1, 0, -1, -2, -2, 1, 2, 2, 1, 2, -2, 1, 0, 0, -2, 4, 0, 0, -1, 1, 2, -1&
         , 1, -1, -4, 1, 1, 2, 0, 0, 1, -3, 0, 0, 1, -1, 1, 1, -1, -1, 3, -1, 1, -3, 0, 0, -1, 3, 0&
         , 0, -1, -2, -1, 1, 0, -1, 3, 2, -1, -1, 0, 0, 0, -3, 0, 0, 1, 2, 0, -1, 0, -1, -3, -2, -1&
         , 1, 0, 0, 0, 1, 0, 0, 0, -1, 0, 0, 0, 2, -1, -1, 2, 0, 0, 0, 0, -1, 0, 0, 0, 1, 0, 0, 0, &
         -1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 2, 0, 0, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 0, 0, &
         0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: c5 = RESHAPE((/-1, -1, -1, -1, -1, -1, -1, -1, -1 &
         , -1, -1, -1, -1, -1, -1, -1, 1, -1, -2, -3, 1, 1, -1, -2, 2, 1, 2, -1, 3, 2, 1, 3, 0, 1, &
         -1, -3, 1, 1, 1, 1, -1, 1, 1, 2, -3, 1, 2, 1, 0, 1, 1, 1, -1, -1, 1, 2, 1, 1, -1, 1, 1, -2&
         , 1, -3, 0, 0, 2, -1, 0, 0, 1, 2, -2, -1, -2, 2, 1, -2, -2, -3, 0, 0, 1, 3, 0, 0, 1, 1, 1,&
         -1, 1, 1, -3, 1, -1, 1, 0, 0, 0, 3, 0, 0, -1, -2, 0, -1, 0, -1, 3, 2, -1, 3, 0, 0, 0, 1, 0&
         , 0, -1, -1, 0, 1, 0, -2, -1, -1, -2, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0,&
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0,&
         1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 &
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, &
         20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma1 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4, 4, 6, 4, 5, 6, 6, 0, 1, 1, 3, 1, 0, 3, 4, 1, 3&
         , 2, 5, 3, 4, 5, 4, 0, 0, 0, 2, 0, 0, 2, 3, 0, 2, 0, 4, 2, 3, 4, 2, 0, 0, 0, 1, 0, 0, 1, 2&
         , 0, 1, 0, 3, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 2, 0, 1, 2, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma2 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 1, 2, 3, 4, 2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7, 1, 1, 3, 4, 2, 3, 3, 5, 3, 4&
         , 5, 5, 4, 5, 6, 7, 0, 0, 2, 3, 1, 2, 2, 4, 2, 3, 4, 4, 3, 4, 5, 6, 0, 0, 2, 2, 1, 2, 1, 4&
         , 2, 2, 4, 3, 3, 4, 5, 6, 0, 0, 1, 1, 0, 1, 0, 3, 1, 1, 3, 2, 2, 3, 4, 5, 0, 0, 1, 1, 0, 1&
         , 0, 3, 1, 1, 3, 1, 2, 3, 4, 5, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 1, 2, 3, 4, 0, 0, 0, 0&
         , 0, 0, 0, 2, 0, 0, 2, 0, 1, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 2, 3, 0, 0&
         , 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma3 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 1, 2, 3, 4, 2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7, 1, 2, 3, 4, 1, 3, 4, 5, 3, 3&
         , 5, 6, 4, 5, 5, 7, 0, 1, 2, 3, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4, 4, 6, 0, 1, 2, 3, 0, 2, 2, 4&
         , 2, 1, 4, 5, 2, 4, 3, 6, 0, 0, 1, 2, 0, 1, 1, 3, 1, 0, 3, 4, 1, 3, 2, 5, 0, 0, 1, 2, 0, 1&
         , 1, 3, 1, 0, 3, 4, 1, 3, 1, 5, 0, 0, 0, 1, 0, 0, 0, 2, 0, 0, 2, 3, 0, 2, 0, 4, 0, 0, 0, 1&
         , 0, 0, 0, 2, 0, 0, 2, 3, 0, 2, 0, 4, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 2, 0, 1, 0, 3, 0, 0&
         , 0, 0, 0, 0, 0, 1, 0, 0, 1, 2, 0, 1, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 2&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma4 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7, 5, 6, 7, 8, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4&
         , 4, 6, 4, 5, 6, 6, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4, 4, 6, 4, 5, 6, 6, 0, 1, 2, 3, 1, 0, 3, 4&
         , 2, 3, 4, 5, 3, 4, 5, 6, 0, 1, 2, 3, 1, 0, 3, 4, 2, 3, 2, 5, 3, 4, 5, 4, 0, 0, 2, 3, 0, 0&
         , 2, 3, 2, 2, 2, 5, 3, 3, 5, 4, 0, 0, 1, 2, 0, 0, 2, 3, 1, 2, 2, 4, 2, 3, 4, 2, 0, 0, 1, 2&
         , 0, 0, 1, 2, 1, 1, 0, 4, 2, 2, 4, 2, 0, 0, 0, 2, 0, 0, 1, 2, 0, 1, 0, 4, 2, 2, 4, 2, 0, 0&
         , 0, 1, 0, 0, 0, 1, 0, 0, 0, 3, 1, 1, 3, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 3, 1, 1, 3, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0&
         , 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2&
         , 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: ma5 = RESHAPE((/2, 3, 4, 5, 3, 4, 5, 6, 4, 5, 6, 7&
         , 5, 6, 7, 8, 0, 2, 3, 4, 2, 2, 4, 5, 3, 4, 4, 6, 4, 5, 6, 6, 0, 1, 2, 3, 1, 2, 3, 4, 2, 3&
         , 4, 5, 3, 4, 5, 6, 0, 0, 2, 3, 0, 0, 3, 3, 2, 3, 2, 5, 3, 3, 5, 4, 0, 0, 1, 2, 0, 0, 2, 3&
         , 1, 2, 2, 4, 2, 3, 4, 4, 0, 0, 0, 2, 0, 0, 2, 2, 0, 2, 0, 4, 2, 2, 4, 2, 0, 0, 0, 1, 0, 0&
         , 1, 1, 0, 1, 0, 3, 1, 1, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 3, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb1 = RESHAPE((/0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 0, 2, 3, 2, 2, 4, 2, 2, 3, 2&
         , 4, 2, 2, 2, 2, 4, 0, 3, 4, 3, 3, 0, 3, 3, 4, 3, 6, 3, 3, 3, 3, 6, 0, 0, 0, 4, 0, 0, 4, 4&
         , 0, 4, 0, 4, 4, 4, 4, 8, 0, 0, 0, 5, 0, 0, 5, 5, 0, 5, 0, 5, 5, 5, 5, 0, 0, 0, 0, 0, 0, 0&
         , 0, 6, 0, 0, 0, 6, 0, 6, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 7, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb2 = RESHAPE((/1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1&
         , 1, 1, 1, 1, 2, 0, 2, 2, 2, 2, 0, 2, 2, 2, 2, 0, 2, 2, 2, 2, 0, 3, 0, 0, 0, 0, 3, 0, 0, 0&
         , 0, 3, 0, 0, 0, 0, 1, 2, 3, 1, 3, 3, 2, 3, 3, 1, 3, 2, 3, 3, 3, 3, 0, 0, 1, 4, 1, 1, 5, 1&
         , 1, 4, 1, 5, 1, 1, 1, 1, 0, 0, 4, 5, 2, 4, 4, 4, 4, 5, 4, 4, 4, 4, 4, 4, 0, 0, 2, 3, 0, 2&
         , 0, 2, 2, 3, 2, 7, 2, 2, 2, 2, 0, 0, 3, 4, 0, 3, 0, 5, 3, 4, 5, 6, 5, 5, 5, 5, 0, 0, 0, 0&
         , 0, 0, 0, 3, 0, 0, 3, 0, 3, 3, 3, 3, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 6, 0, 4, 6, 6, 6, 0, 0&
         , 0, 0, 0, 0, 0, 4, 0, 0, 4, 0, 0, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 0, 0, 5, 7, 7&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 6, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb3 = RESHAPE((/1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1&
         , 1, 1, 1, 1, 2, 2, 2, 2, 0, 2, 2, 2, 2, 0, 2, 2, 2, 2, 0, 2, 0, 0, 0, 0, 3, 0, 0, 0, 0, 3&
         , 0, 0, 0, 0, 3, 0, 1, 3, 3, 3, 2, 3, 1, 3, 3, 2, 3, 3, 1, 3, 2, 3, 0, 1, 1, 1, 0, 1, 4, 1&
         , 1, 5, 1, 1, 4, 1, 5, 1, 0, 2, 4, 4, 0, 4, 5, 4, 4, 4, 4, 4, 5, 4, 4, 4, 0, 0, 2, 2, 0, 2&
         , 3, 2, 2, 0, 2, 2, 3, 2, 7, 2, 0, 0, 3, 5, 0, 3, 4, 5, 3, 0, 5, 5, 4, 5, 6, 5, 0, 0, 0, 3&
         , 0, 0, 0, 3, 0, 0, 3, 3, 0, 3, 0, 3, 0, 0, 0, 4, 0, 0, 0, 6, 0, 0, 6, 6, 0, 6, 0, 6, 0, 0&
         , 0, 0, 0, 0, 0, 4, 0, 0, 4, 4, 0, 4, 0, 4, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 7, 0, 5, 0, 7&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0&
         , 0, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb4 = RESHAPE((/2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2&
         , 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 3, 3, 3, 4, 3, 3, 3, 3&
         , 4, 3, 3, 3, 3, 4, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 0, 2, 4, 4, 2, 4, 4, 2&
         , 4, 4, 0, 4, 4, 2, 4, 0, 0, 0, 2, 2, 0, 2, 0, 0, 2, 0, 6, 2, 2, 0, 2, 6, 0, 3, 0, 0, 3, 0&
         , 5, 5, 0, 5, 4, 0, 0, 5, 0, 2, 0, 1, 3, 5, 1, 0, 1, 1, 3, 1, 2, 5, 5, 1, 5, 8, 0, 0, 1, 3&
         , 0, 0, 4, 6, 1, 4, 6, 3, 3, 6, 3, 6, 0, 0, 4, 1, 0, 0, 2, 4, 4, 2, 4, 1, 1, 4, 1, 4, 0, 0&
         , 2, 4, 0, 0, 5, 5, 2, 5, 0, 6, 4, 5, 6, 8, 0, 0, 0, 2, 0, 0, 3, 3, 0, 3, 0, 4, 2, 3, 4, 6&
         , 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 2, 5, 6, 2, 0, 0, 0, 0, 3, 0, 0, 0, 4, 0, 0, 0, 7, 3, 4&
         , 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3&
         , 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 4, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 5, 0, 0, 5, 0/), (/4, 4, 20/))
      INTEGER, DIMENSION(4, 4, 20), PARAMETER :: mb5 = RESHAPE((/2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2&
         , 2, 2, 2, 2, 0, 3, 3, 3, 3, 4, 3, 3, 3, 3, 4, 3, 3, 3, 3, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4&
         , 0, 4, 4, 0, 4, 0, 0, 1, 0, 0, 1, 2, 0, 5, 0, 0, 6, 0, 0, 5, 0, 6, 0, 0, 1, 5, 0, 0, 5, 1&
         , 1, 5, 2, 5, 5, 1, 5, 2, 0, 0, 2, 1, 0, 0, 1, 6, 2, 1, 4, 1, 1, 6, 1, 8, 0, 0, 0, 2, 0, 0&
         , 2, 3, 0, 2, 0, 6, 2, 3, 6, 4, 0, 0, 0, 3, 0, 0, 3, 4, 0, 3, 0, 2, 3, 4, 2, 6, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 3, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0&
         , 0, 0, 0, 0, 0, 0, 0, 0/), (/4, 4, 20/))

      INTEGER                                            :: k, k1, k2, mu
      REAL(kind=dp)                                      :: cp, ct, dJ, dJc, dJcc, dJss, dxx, dyy, &
                                                            f, fac1, fac2, J, Jc, Jcc, Jss, rr, &
                                                            sp, st, w, w1, w2, xx, yy, za, zb
      REAL(kind=dp), DIMENSION(3)                        :: dcp, dct, dsp, dst, v
      REAL(kind=dp), DIMENSION(3, 3)                     :: Arot
      REAL(kind=dp), DIMENSION(3, 3, 3)                  :: dArot

      dS(:, :, :) = 0.0_dp

      v(:) = R(:)
      rr = SQRT(DOT_PRODUCT(v, v))

      IF (rr < 1.0e-20_dp) THEN

         DO mu = 1, 4
            dS(mu, mu, :) = 0.0_dp
         ENDDO

      ELSE

         fac1 = 1.0_dp
         IF (nra == 1) THEN
            fac1 = fac1*2.0_dp
         ELSE
            IF (nra == 2) THEN
               fac1 = fac1*SQRT(4.0_dp/3.0_dp)
            ELSE
               IF (nra == 3) THEN
                  fac1 = fac1*SQRT(8.0_dp/45.0_dp)
               ELSE
                  IF (nra == 4) THEN
                     fac1 = fac1*SQRT(4.0_dp/315.0_dp)
                  ELSE
                     WRITE (*, *) 'nra= ', nra
                     RETURN
                  ENDIF
               ENDIF
            ENDIF
         ENDIF
         IF (nrb == 1) THEN
            fac1 = fac1*2.0_dp
         ELSE
            IF (nrb == 2) THEN
               fac1 = fac1*SQRT(4.0_dp/3.0_dp)
            ELSE
               IF (nrb == 3) THEN
                  fac1 = fac1*SQRT(8.0_dp/45.0_dp)
               ELSE
                  IF (nrb == 4) THEN
                     fac1 = fac1*SQRT(4.0_dp/315.0_dp)
                  ELSE
                     WRITE (*, *) 'nrb= ', nrb
                     RETURN
                  ENDIF
               ENDIF
            ENDIF
         ENDIF

         ct = -v(3)/rr
         IF (ABS(ct) >= 1.0_dp) THEN

            dct(:) = v(:)*v(3)/rr**3
            dct(3) = dct(3)-1.0_dp/rr

            Arot(1, 1) = ct
            Arot(1, 2) = 0.0_dp
            Arot(1, 3) = 0.0_dp
            Arot(2, 1) = 0.0_dp
            Arot(2, 2) = 1.0_dp
            Arot(2, 3) = 0.0_dp
            Arot(3, 1) = 0.0_dp
            Arot(3, 2) = 0.0_dp
            Arot(3, 3) = ct

            dArot(1, 1, :) = dct(:)
            dArot(1, 2, :) = 0.0_dp
            dArot(1, 3, :) = 0.0_dp
            dArot(2, 1, :) = 0.0_dp
            dArot(2, 2, :) = 0.0_dp
            dArot(2, 3, :) = 0.0_dp
            dArot(3, 1, :) = 0.0_dp
            dArot(3, 2, :) = 0.0_dp
            dArot(3, 3, :) = dct(:)

         ELSE

            xx = SQRT(v(1)**2+v(2)**2)
            st = xx/rr
            cp = -v(1)/xx
            sp = -v(2)/xx

            dct(:) = v(:)*v(3)/rr**3
            dct(3) = dct(3)-1.0_dp/rr
            dst(:) = -ct*dct(:)/st
            dcp(:) = v(:)*v(1)/(rr**3*st)
            dcp(:) = dcp(:)+v(1)*dst(:)/(rr*st**2)
            dcp(1) = dcp(1)-1.0_dp/(rr*st)
            dsp(:) = v(:)*v(2)/(rr**3*st)
            dsp(:) = dsp(:)+v(2)*dst(:)/(rr*st**2)
            dsp(2) = dsp(2)-1.0_dp/(rr*st)

            Arot(1, 1) = ct*cp
            Arot(1, 2) = -sp
            Arot(1, 3) = st*cp
            Arot(2, 1) = ct*sp
            Arot(2, 2) = cp
            Arot(2, 3) = st*sp
            Arot(3, 1) = -st
            Arot(3, 2) = 0.0_dp
            Arot(3, 3) = ct

            dArot(1, 1, :) = dct(:)*cp+ct*dcp(:)
            dArot(1, 2, :) = -dsp(:)
            dArot(1, 3, :) = dst(:)*cp+st*dcp(:)
            dArot(2, 1, :) = dct(:)*sp+ct*dsp(:)
            dArot(2, 2, :) = dcp(:)
            dArot(2, 3, :) = dst(:)*sp+st*dsp(:)
            dArot(3, 1, :) = -dst(:)
            dArot(3, 2, :) = 0.0_dp
            dArot(3, 3, :) = dct(:)

         ENDIF

         za = ZSA
         zb = ZSB
         fac2 = SQRT(za**(2*nra+1)*zb**(2*nrb+1))
         xx = 0.5_dp*rr*(za+zb)
         yy = 0.5_dp*rr*(za-zb)
         dxx = 0.5_dp*(za+zb)
         dyy = 0.5_dp*(za-zb)

         w = 0.0_dp
         w1 = 0.0_dp
         w2 = 0.0_dp
         f = rr**(nra+nrb+1)/2.0_dp**(nra+nrb+2)
         DO k = 1, nc1(nra, nrb)
            w = w+REAL(c1(nra, nrb, k), dp)*AA(ma1(nra, nrb, k), xx)*BB(mb1(nra, nrb, k), yy)
            w1 = w1+REAL(c1(nra, nrb, k), dp)*AA(ma1(nra, nrb, k)+1, xx)*BB(mb1(nra, nrb, k), yy)
            w2 = w2+REAL(c1(nra, nrb, k), dp)*AA(ma1(nra, nrb, k), xx)*BB(mb1(nra, nrb, k)+1, yy)
         ENDDO
         J = f*w
         dJ = f*REAL(nra+nrb+1, dp)*w/rr
         dJ = dJ-dxx*f*w1
         dJ = dJ-dyy*f*w2

         dS(1, 1, :) = dS(1, 1, :)+fac1*fac2*dJ*v(:)/rr

         za = ZPA
         zb = ZSB
         fac2 = SQRT(za**(2*nra+1)*zb**(2*nrb+1))
         xx = 0.5_dp*rr*(za+zb)
         yy = 0.5_dp*rr*(za-zb)
         dxx = 0.5_dp*(za+zb)
         dyy = 0.5_dp*(za-zb)

         w = 0.0_dp
         w1 = 0.0_dp
         w2 = 0.0_dp
         f = rr**(nra+nrb+1)/2.0_dp**(nra+nrb+2)
         DO k = 1, nc2(nra, nrb)
            w = w+REAL(c2(nra, nrb, k), dp)*AA(ma2(nra, nrb, k), xx)*BB(mb2(nra, nrb, k), yy)
            w1 = w1+REAL(c2(nra, nrb, k), dp)*AA(ma2(nra, nrb, k)+1, xx)*BB(mb2(nra, nrb, k), yy)
            w2 = w2+REAL(c2(nra, nrb, k), dp)*AA(ma2(nra, nrb, k), xx)*BB(mb2(nra, nrb, k)+1, yy)
         ENDDO
         Jc = f*w
         dJc = f*REAL(nra+nrb+1, dp)*w/rr
         dJc = dJc-dxx*f*w1
         dJc = dJc-dyy*f*w2

         DO k1 = 1, 3
            dS(k1+1, 1, :) = dS(k1+1, 1, :) &
         &          +SQRT(3.0_dp)*Arot(k1, 3)*fac1*fac2*dJc*v(:)/rr &
         &          +SQRT(3.0_dp)*dArot(k1, 3, :)*fac1*fac2*Jc
         ENDDO

         za = ZSA
         zb = ZPB
         fac2 = SQRT(za**(2*nra+1)*zb**(2*nrb+1))
         xx = 0.5_dp*rr*(za+zb)
         yy = 0.5_dp*rr*(za-zb)
         dxx = 0.5_dp*(za+zb)
         dyy = 0.5_dp*(za-zb)

         w = 0.0_dp
         w1 = 0.0_dp
         w2 = 0.0_dp
         f = rr**(nra+nrb+1)/2.0_dp**(nra+nrb+2)
         DO k = 1, nc3(nra, nrb)
            w = w+REAL(c3(nra, nrb, k), dp)*AA(ma3(nra, nrb, k), xx)*BB(mb3(nra, nrb, k), yy)
            w1 = w1+REAL(c3(nra, nrb, k), dp)*AA(ma3(nra, nrb, k)+1, xx)*BB(mb3(nra, nrb, k), yy)
            w2 = w2+REAL(c3(nra, nrb, k), dp)*AA(ma3(nra, nrb, k), xx)*BB(mb3(nra, nrb, k)+1, yy)
         ENDDO
         Jc = f*w
         dJc = f*REAL(nra+nrb+1, dp)*w/rr
         dJc = dJc-dxx*f*w1
         dJc = dJc-dyy*f*w2

         DO k1 = 1, 3
            dS(1, k1+1, :) = dS(1, k1+1, :) &
         &          -SQRT(3.0_dp)*Arot(k1, 3)*fac1*fac2*dJc*v(:)/rr &
         &          -SQRT(3.0_dp)*dArot(k1, 3, :)*fac1*fac2*Jc
         ENDDO

         za = ZPA
         zb = ZPB
         fac2 = SQRT(za**(2*nra+1)*zb**(2*nrb+1))
         xx = 0.5_dp*rr*(za+zb)
         yy = 0.5_dp*rr*(za-zb)
         dxx = 0.5_dp*(za+zb)
         dyy = 0.5_dp*(za-zb)

         w = 0.0_dp
         w1 = 0.0_dp
         w2 = 0.0_dp
         f = rr**(nra+nrb+1)/2.0_dp**(nra+nrb+2)
         DO k = 1, nc4(nra, nrb)
            w = w+REAL(c4(nra, nrb, k), dp)*AA(ma4(nra, nrb, k), xx)*BB(mb4(nra, nrb, k), yy)
            w1 = w1+REAL(c4(nra, nrb, k), dp)*AA(ma4(nra, nrb, k)+1, xx)*BB(mb4(nra, nrb, k), yy)
            w2 = w2+REAL(c4(nra, nrb, k), dp)*AA(ma4(nra, nrb, k), xx)*BB(mb4(nra, nrb, k)+1, yy)
         ENDDO
         Jss = f*w
         dJss = f*REAL(nra+nrb+1, dp)*w/rr
         dJss = dJss-dxx*f*w1
         dJss = dJss-dyy*f*w2

         w = 0.0_dp
         w1 = 0.0_dp
         w2 = 0.0_dp
         f = rr**(nra+nrb+1)/2.0_dp**(nra+nrb+2)
         DO k = 1, nc5(nra, nrb)
            w = w+REAL(c5(nra, nrb, k), dp)*AA(ma5(nra, nrb, k), xx)*BB(mb5(nra, nrb, k), yy)
            w1 = w1+REAL(c5(nra, nrb, k), dp)*AA(ma5(nra, nrb, k)+1, xx)*BB(mb5(nra, nrb, k), yy)
            w2 = w2+REAL(c5(nra, nrb, k), dp)*AA(ma5(nra, nrb, k), xx)*BB(mb5(nra, nrb, k)+1, yy)
         ENDDO
         Jcc = f*w
         dJcc = f*REAL(nra+nrb+1, dp)*w/rr
         dJcc = dJcc-dxx*f*w1
         dJcc = dJcc-dyy*f*w2

         DO k1 = 1, 3
            DO k2 = 1, 3
               dS(k1+1, k2+1, :) = dS(k1+1, k2+1, :) &
          &            +1.5_dp*Arot(k1, 1)*Arot(k2, 1)*fac1*fac2*dJss*v(:)/rr &
          &            +1.5_dp*dArot(k1, 1, :)*Arot(k2, 1)*fac1*fac2*Jss &
          &            +1.5_dp*Arot(k1, 1)*dArot(k2, 1, :)*fac1*fac2*Jss &
          &            +1.5_dp*Arot(k1, 2)*Arot(k2, 2)*fac1*fac2*dJss*v(:)/rr &
          &            +1.5_dp*dArot(k1, 2, :)*Arot(k2, 2)*fac1*fac2*Jss &
          &            +1.5_dp*Arot(k1, 2)*dArot(k2, 2, :)*fac1*fac2*Jss &
          &            -3.0_dp*Arot(k1, 3)*Arot(k2, 3)*fac1*fac2*dJcc*v(:)/rr &
          &            -3.0_dp*dArot(k1, 3, :)*Arot(k2, 3)*fac1*fac2*Jcc &
          &            -3.0_dp*Arot(k1, 3)*dArot(k2, 3, :)*fac1*fac2*Jcc
            ENDDO
         ENDDO

      ENDIF

   END SUBROUTINE makedS

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param x ...
!> \retval AA ...
! **************************************************************************************************
   FUNCTION AA(n, x)

      INTEGER                                            :: n
      REAL(kind=dp)                                      :: x, AA

      REAL(kind=dp)                                      :: p

      IF (n == 0) THEN
         p = 1.0_dp
      ELSE
         IF (n == 1) THEN
            p = 1.0_dp+x
         ELSE
            IF (n == 2) THEN
               p = 2.0_dp+x*( &
                   2.0_dp+x)
            ELSE
               IF (n == 3) THEN
                  p = 6.0_dp+x*( &
                      6.0_dp+x*( &
                      3.0_dp+x))
               ELSE
                  IF (n == 4) THEN
                     p = 24.0_dp+x*( &
                         24.0_dp+x*( &
                         12.0_dp+x*( &
                         4.0_dp+x)))
                  ELSE
                     IF (n == 5) THEN
                        p = 120.0_dp+x*( &
                            120.0_dp+x*( &
                            60.0_dp+x*( &
                            20.0_dp+x*( &
                            5.0_dp+x))))
                     ELSE
                        IF (n == 6) THEN
                           p = 720.0_dp+x*( &
                               720.0_dp+x*( &
                               360.0_dp+x*( &
                               120.0_dp+x*( &
                               30.0_dp+x*( &
                               6.0_dp+x)))))
                        ELSE
                           IF (n == 7) THEN
                              p = 5040.0_dp+x*( &
                                  5040.0_dp+x*( &
                                  2520.0_dp+x*( &
                                  840.0_dp+x*( &
                                  210.0_dp+x*( &
                                  42.0_dp+x*( &
                                  7.0_dp+x))))))
                           ELSE
                              IF (n == 8) THEN
                                 p = 40320.0_dp+x*( &
                                     40320.0_dp+x*( &
                                     20160.0_dp+x*( &
                                     6720.0_dp+x*( &
                                     1680.0_dp+x*( &
                                     336.0_dp+x*( &
                                     56.0_dp+x*( &
                                     8.0_dp+x)))))))
                              ELSE
                                 IF (n == 9) THEN
                                    p = 362880.0_dp+x*( &
                                        362880.0_dp+x*( &
                                        181440.0_dp+x*( &
                                        60480.0_dp+x*( &
                                        15120.0_dp+x*( &
                                        3024.0_dp+x*( &
                                        504.0_dp+x*( &
                                        72.0_dp+x*( &
                                        9.0_dp+x))))))))
                                 ELSE
                                    IF (n == 10) THEN
                                       p = 3628800.0_dp+x*( &
                                           3628800.0_dp+x*( &
                                           1814400.0_dp+x*( &
                                           604800.0_dp+x*( &
                                           151200.0_dp+x*( &
                                           30240.0_dp+x*( &
                                           5040.0_dp+x*( &
                                           720.0_dp+x*( &
                                           90.0_dp+x*( &
                                           10.0_dp+x)))))))))
                                    ELSE
                                       p = 1.0_dp
                                       WRITE (*, *) ' n= ', n, ' in AA(n,x) '
                                    ENDIF
                                 ENDIF
                              ENDIF
                           ENDIF
                        ENDIF
                     ENDIF
                  ENDIF
               ENDIF
            ENDIF
         ENDIF
      ENDIF

      AA = EXP(-x)*p/x**(n+1)

   END FUNCTION AA

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param y ...
!> \retval BB ...
! **************************************************************************************************
   FUNCTION BB(n, y)

      INTEGER                                            :: n
      REAL(kind=dp)                                      :: y, BB

      IF (ABS(y) > 1.0e-20_dp) THEN
         BB = REAL((-1)**(n+1), dp)*AA(n, -y)-AA(n, y)
      ELSE
         IF (MOD(n, 2) == 0) THEN
            BB = 2.0_dp/REAL(n+1, dp)
         ELSE
            BB = -y*2.0_dp/REAL(n+2, dp)
         ENDIF
      ENDIF

   END FUNCTION BB

END MODULE se_core_matrix

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief xas_scf for the tp method
!>       It is repeaated for every atom that have to be excited
!> \par History
!>      created 05.2005
!> \author MI (05.2005)
! **************************************************************************************************
MODULE xas_tp_scf
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_array_utils_r,                ONLY: cp_2d_r_p_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_p_type
   USE cp_external_control,             ONLY: external_control
   USE cp_fm_types,                     ONLY: cp_fm_get_submatrix,&
                                              cp_fm_init_random,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: ot_precond_full_kinetic,&
                                              ot_precond_solver_default,&
                                              xas_dscf
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE message_passing,                 ONLY: mp_sync
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE preconditioner,                  ONLY: make_preconditioner
   USE preconditioner_types,            ONLY: destroy_preconditioner,&
                                              init_preconditioner,&
                                              preconditioner_type
   USE qs_charges_types,                ONLY: qs_charges_type
   USE qs_density_mixing_types,         ONLY: broyden_mixing_new_nr,&
                                              broyden_mixing_nr,&
                                              direct_mixing_nr,&
                                              gspace_mixing_nr,&
                                              multisecant_mixing_nr,&
                                              no_mixing_nr,&
                                              pulay_mixing_nr
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gspace_mixing,                ONLY: gspace_mixing,&
                                              self_consistency_check
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type
   USE qs_loc_methods,                  ONLY: qs_loc_driver
   USE qs_loc_types,                    ONLY: get_qs_loc_env,&
                                              localized_wfn_control_type,&
                                              qs_loc_env_new_type
   USE qs_mo_methods,                   ONLY: calculate_density_matrix,&
                                              calculate_subspace_eigenvalues
   USE qs_mo_occupation,                ONLY: set_mo_occupation
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_ot_eigensolver,               ONLY: ot_eigensolver
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf,                          ONLY: scf_env_cleanup,&
                                              scf_env_do_scf
   USE qs_scf_diagonalization,          ONLY: general_eigenproblem
   USE qs_scf_initialization,           ONLY: qs_scf_env_initialize
   USE qs_scf_methods,                  ONLY: scf_env_density_mixing
   USE qs_scf_output,                   ONLY: qs_scf_print_summary
   USE qs_scf_types,                    ONLY: general_diag_method_nr,&
                                              qs_scf_env_type,&
                                              scf_env_release
   USE scf_control_types,               ONLY: scf_control_type
   USE xas_control,                     ONLY: xas_control_type
   USE xas_env_types,                   ONLY: get_xas_env,&
                                              set_xas_env,&
                                              xas_environment_type
   USE xas_restart,                     ONLY: find_excited_core_orbital,&
                                              xas_write_restart
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xas_tp_scf'

   ! *** Public subroutines ***

   PUBLIC :: xas_do_tp_scf, xes_scf_once

CONTAINS

! **************************************************************************************************
!> \brief perform an scf loop to calculate the xas spectrum
!>      given by the excitation of a inner state of a selected atom
!>      by using the transition potential method
!> \param dft_control ...
!> \param xas_env the environment for XAS  calculations
!> \param iatom ...
!> \param scf_env the scf_env where to perform the scf procedure
!> \param qs_env the qs_env, the scf_env and xas_env live in
!> \param xas_section ...
!> \param scf_section ...
!> \param converged ...
!> \param should_stop ...
!> \par History
!>      05.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE xas_do_tp_scf(dft_control, xas_env, iatom, scf_env, qs_env, &
                            xas_section, scf_section, converged, should_stop)

      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(xas_environment_type), POINTER                :: xas_env
      INTEGER, INTENT(IN)                                :: iatom
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: xas_section, scf_section
      LOGICAL, INTENT(OUT)                               :: converged, should_stop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xas_do_tp_scf', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle2, ispin, iter_count, &
                                                            nspin, output_unit
      LOGICAL                                            :: diis_step, energy_only, exit_loop, gapw
      REAL(KIND=dp)                                      :: t1, t2
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao_kp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(xas_control_type), POINTER                    :: xas_control

      CALL timeset(routineN, handle)
      NULLIFY (xas_control, matrix_s, matrix_ks, para_env, rho_ao_kp)
      NULLIFY (rho, energy, scf_control, logger, ks_env, mos, atomic_kind_set)
      NULLIFY (qs_charges)

      logger => cp_get_default_logger()
      t1 = m_walltime()
      converged = .TRUE.

      IF(.NOT.(ASSOCIATED(xas_env)))CALL cp__a("xas_tp_scf.F",171)
      IF(.NOT.(xas_env%ref_count > 0))CALL cp__a("xas_tp_scf.F",172)
      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("xas_tp_scf.F",173)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("xas_tp_scf.F",174)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("xas_tp_scf.F",175)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("xas_tp_scf.F",176)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      matrix_s=matrix_s, energy=energy, &
                      qs_charges=qs_charges, &
                      ks_env=ks_env, para_env=para_env)

      CALL get_xas_env(xas_env, scf_control=scf_control)

      energy_only = .FALSE.
      output_unit = cp_print_key_unit_nr(logger, xas_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".xasLog")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") "XAS_TP_SCF WAVEFUNCTION OPTIMIZATION"
      END IF

      !   GAPW method must be used
      gapw = dft_control%qs_control%gapw
      IF(.NOT.(gapw))CALL cp__a("xas_tp_scf.F",195)
      xas_control => dft_control%xas_control

      CALL cp_add_iter_level(logger%iter_info, "XAS_SCF")

      CALL get_qs_env(qs_env, matrix_ks=matrix_ks, rho=rho, mos=mos)
      CALL qs_rho_get(rho, rho_ao_kp=rho_ao_kp)

      iter_count = 0
      diis_step = .FALSE.
      nspin = dft_control%nspins

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, &
                FMT="(/,T3,A,T12,A,T31,A,T40,A,T60,A,T75,A/,T3,A)") &
            "Step", "Update method", "Time", "Convergence", "Total energy", "Change", &
            REPEAT("-", 77)
      END IF

      !   *** SCF loop ***

      energy%tot_old = 0.0_dp
      scf_loop: DO
         CALL timeset(routineN//"_inner_loop", handle2)

         exit_loop = .FALSE.
         IF (output_unit > 0) CALL m_flush(output_unit)

         iter_count = iter_count+1
         CALL cp_iterate(logger%iter_info, last=.FALSE., iter_nr=iter_count)

         ! ** here qs_env%rho%rho_r and qs_env%rho%rho_g should be up to date

         CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., just_energy=energy_only)

         SELECT CASE (scf_env%method)
         CASE DEFAULT
            CALL cp_abort(cp__l("xas_tp_scf.F",232), &
                          "unknown scf method method for core level spectroscopy"// &
                          cp_to_string(scf_env%method))

         CASE (general_diag_method_nr) ! diagonalisation (default)
            scf_env%iter_count = iter_count
            CALL general_eigenproblem(scf_env, mos, matrix_ks, &
                                      matrix_s, scf_control, scf_section, &
                                      diis_step)
            CALL find_excited_core_orbital(xas_env, mos, matrix_s)
            CALL set_mo_occupation(mo_set=mos(1)%mo_set, &
                                   smear=scf_control%smear, &
                                   xas_env=xas_env)
            CALL set_mo_occupation(mo_set=mos(2)%mo_set, &
                                   smear=scf_control%smear)
            DO ispin = 1, nspin
               ! does not yet handle k-points
               CALL calculate_density_matrix(mos(ispin)%mo_set, &
                                             scf_env%p_mix_new(ispin, 1)%matrix)
            END DO
            energy%kTS = 0.0_dp
            energy%efermi = 0.0_dp
            DO ispin = 1, nspin
               energy%kTS = energy%kTS+mos(ispin)%mo_set%kTS
               energy%efermi = energy%efermi+mos(ispin)%mo_set%mu
            ENDDO
            energy%efermi = energy%efermi/REAL(nspin, KIND=dp)

         END SELECT

         SELECT CASE (scf_env%mixing_method)
         CASE (direct_mixing_nr)
            CALL scf_env_density_mixing(scf_env%p_mix_new, &
                                        scf_env%mixing_store, rho_ao_kp, para_env, scf_env%iter_delta, scf_env%iter_count, &
                                        diis=diis_step)
         CASE (gspace_mixing_nr, pulay_mixing_nr, broyden_mixing_nr, &
               broyden_mixing_new_nr, multisecant_mixing_nr)
            ! Compute the difference p_out-p_in
            CALL self_consistency_check(rho_ao_kp, scf_env%p_delta, para_env, scf_env%p_mix_new, &
                                        delta=scf_env%iter_delta)
         CASE (no_mixing_nr)
         CASE DEFAULT
            CALL cp_abort(cp__l("xas_tp_scf.F",274), &
                          "unknown scf mixing method: "// &
                          cp_to_string(scf_env%mixing_method))
         END SELECT

         t2 = m_walltime()

         IF (output_unit > 0 .AND. scf_env%print_iter_line) THEN
            WRITE (UNIT=output_unit, &
                   FMT="(T2,I5,1X,A,T20,E8.2,1X,F6.1,1X,F14.8,1X,F20.10,1X,ES9.2)") &
               iter_count, TRIM(scf_env%iter_method), &
               scf_env%iter_param, t2-t1, scf_env%iter_delta, energy%total, &
               energy%total-energy%tot_old
         END IF
         energy%tot_old = energy%total

         ! ** convergence check
         CALL external_control(should_stop, "XASSCF", target_time=qs_env%target_time, &
                               start_time=qs_env%start_time)
         IF (scf_env%iter_delta < scf_control%eps_scf) THEN
            IF (output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(/,T3,A,I5,A/)") &
                  "*** SCF run converged in ", iter_count, " steps ***"
            END IF
            exit_loop = .TRUE.
         ELSE IF (should_stop .OR. iter_count == scf_control%max_scf) THEN
            IF (output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(/,T3,A,/)") &
                  "*** SCF run NOT converged ***"
            END IF
            converged = .FALSE.
            exit_loop = .TRUE.
         END IF
         !   *** Exit if we have finished with the SCF inner loop ***
         IF (exit_loop) THEN
            ! now, print out energies and charges corresponding to the obtained wfn
            ! (this actually is not 100% consistent at this point)!
            CALL qs_scf_print_summary(output_unit, qs_env)
            CALL cp_iterate(logger%iter_info, last=.TRUE., iter_nr=iter_count)
         END IF

         ! ** Write restart file **
         CALL xas_write_restart(xas_env, xas_section, qs_env, xas_control%xas_method, &
                                iatom)

         IF (exit_loop) THEN
            CALL timestop(handle2)
            EXIT scf_loop
         END IF

         IF (.NOT. BTEST(cp_print_key_should_output(logger%iter_info, &
                                                    xas_section, "PRINT%ITERATION_INFO/TIME_CUMUL"), cp_p_file)) t1 = m_walltime()

         !   *** mixing methods have the new density matrix in p_mix_new
         IF (scf_env%mixing_method > 0) THEN
            DO ispin = 1, nspin
               ! does not yet handle k-points
               CALL cp_dbcsr_copy(rho_ao_kp(ispin, 1)%matrix, scf_env%p_mix_new(ispin, 1)%matrix)
            END DO
         ENDIF

         ! ** update qs_env%rho
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         IF (scf_env%mixing_method >= gspace_mixing_nr) THEN
            CALL gspace_mixing(qs_env, scf_env%mixing_method, scf_env%mixing_store, rho, para_env, &
                               scf_env%iter_count)
         END IF

         CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)
         CALL timestop(handle2)

      END DO scf_loop

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,(T3,A,T55,F25.14))") &
            "Ionization potential of the excited atom:      ", xas_env%IP_energy
         CALL m_flush(output_unit)
      END IF

      CALL mp_sync(para_env%group)
      CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)

      CALL cls_prepare_states(xas_control, xas_env, qs_env, iatom, xas_section, output_unit)

      CALL mp_sync(para_env%group)

      CALL cp_print_key_finished_output(output_unit, logger, xas_section, &
                                        "PRINT%PROGRAM_RUN_INFO")
      CALL cp_rm_iter_level(logger%iter_info, "XAS_SCF")

      CALL timestop(handle)

   END SUBROUTINE xas_do_tp_scf
! **************************************************************************************************
!> \brief  Post processing of the optimized wfn in XAS scheme, as preparation for
!>         the calculation of the spectrum
!> \param xas_control ...
!> \param xas_env the environment for XAS  calculations
!> \param qs_env the qs_env, the scf_env and xas_env live in
!> \param iatom index of the excited atom
!> \param xas_section ...
!> \param output_unit ...
!> \par History
!>      05.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE cls_prepare_states(xas_control, xas_env, qs_env, iatom, xas_section, output_unit)

      TYPE(xas_control_type)                             :: xas_control
      TYPE(xas_environment_type), POINTER                :: xas_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: iatom
      TYPE(section_vals_type), POINTER                   :: xas_section
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cls_prepare_states', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ikind, isgf, ispin, istate, &
                                                            j, my_kind, my_state, nao, natom, &
                                                            nexc_search, nmo, nvirtual2, uno_iter, &
                                                            xas_estate
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf
      INTEGER, DIMENSION(:), POINTER                     :: mykind_of_kind
      REAL(dp), DIMENSION(:, :), POINTER                 :: centers_wfn
      REAL(KIND=dp)                                      :: component, dist, max_overlap, ra(3), &
                                                            rac(3), rc(3), sto_state_overlap, &
                                                            uno_eps
      REAL(KIND=dp), DIMENSION(:), POINTER               :: all_evals, eigenvalues, uno_evals
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer, vecbuffer2
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_2d_r_p_type), DIMENSION(:), POINTER        :: stogto_overlap
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: kinetic, matrix_ks, matrix_s
      TYPE(cp_fm_type), POINTER                          :: all_vectors, excvec_coeff, &
                                                            excvec_overlap, mo_coeff, uno_orbs
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(preconditioner_type), POINTER                 :: local_preconditioner
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: loc_section, print_loc_section

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind, dft_control, matrix_s, matrix_ks, qs_kind_set, kinetic)
      NULLIFY (cell, particle_set, local_preconditioner, vecbuffer, vecbuffer2)
      NULLIFY (dft_control, loc_section, mos, mo_coeff, eigenvalues)
      NULLIFY (centers_wfn, mykind_of_kind, qs_loc_env, localized_wfn_control, stogto_overlap)
      NULLIFY (all_evals, all_vectors, excvec_coeff, excvec_overlap, uno_evals, para_env, blacs_env)

      IF(.NOT.(ASSOCIATED(xas_env)))CALL cp__a("xas_tp_scf.F",430)

      CALL get_qs_env(qs_env, &
                      cell=cell, &
                      dft_control=dft_control, &
                      matrix_ks=matrix_ks, &
                      matrix_s=matrix_s, &
                      kinetic=kinetic, &
                      mos=mos, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      para_env=para_env, &
                      blacs_env=blacs_env)

      ! Some elements from the xas_env
      CALL get_xas_env(xas_env=xas_env, &
                       all_vectors=all_vectors, all_evals=all_evals, &
                       excvec_coeff=excvec_coeff, &
                       nvirtual2=nvirtual2, xas_estate=xas_estate, &
                       excvec_overlap=excvec_overlap, nexc_search=nexc_search, scf_control=scf_control)
      IF(.NOT.(ASSOCIATED(excvec_overlap)))CALL cp__a("xas_tp_scf.F",450)

      CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff, nao=nao, &
                      eigenvalues=eigenvalues)

      ALLOCATE (vecbuffer(1, nao))
      vecbuffer = 0.0_dp
      ALLOCATE (vecbuffer2(1, nao))
      vecbuffer2 = 0.0_dp
      natom = SIZE(particle_set, 1)
      ALLOCATE (first_sgf(natom))
      CALL get_particle_set(particle_set, qs_kind_set, first_sgf=first_sgf)
      ALLOCATE (centers_wfn(3, nexc_search))
      centers_wfn = 0.0_dp

      ! Possible only for emission only
      IF (scf_control%use_ot) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") " Eigenstates are derived "// &
               "from the MOs optimized by OT. Follows localization of the core states"// &
               " to identify the excited orital. "
         END IF

         CALL get_xas_env(xas_env=xas_env, &
                          mykind_of_kind=mykind_of_kind, qs_loc_env=qs_loc_env, &
                          stogto_overlap=stogto_overlap)
         CALL get_qs_loc_env(qs_loc_env=qs_loc_env, &
                             localized_wfn_control=localized_wfn_control)
         loc_section => section_vals_get_subs_vals(xas_section, "LOCALIZE")
         print_loc_section => section_vals_get_subs_vals(loc_section, "PRINT")
         CALL qs_loc_driver(qs_env, qs_loc_env, print_loc_section, myspin=1)
         ra(1:3) = particle_set(iatom)%r(1:3)

         NULLIFY (atomic_kind)
         atomic_kind => particle_set(iatom)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              kind_number=ikind)
         my_kind = mykind_of_kind(ikind)

         CALL cp_fm_get_submatrix(mo_coeff, vecbuffer2, 1, my_state, &
                                  nao, 1, transpose=.TRUE.)

         ! Rotate the wfn to get the eigenstate of the KS hamiltonian
         ! Only ispin=1 should be needed
         DO ispin = 1, dft_control%nspins
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo, &
                            eigenvalues=eigenvalues)
            CALL calculate_subspace_eigenvalues(mo_coeff, &
                                                matrix_ks(ispin)%matrix, eigenvalues, &
                                                do_rotation=.TRUE.)
         END DO ! ispin

         !Search for the core state to be excited
         max_overlap = 0.0_dp
         DO istate = 1, nexc_search
            centers_wfn(1, istate) = localized_wfn_control%centers_set(1)%array(1, istate)
            centers_wfn(2, istate) = localized_wfn_control%centers_set(1)%array(2, istate)
            centers_wfn(3, istate) = localized_wfn_control%centers_set(1)%array(3, istate)

            rc(1:3) = centers_wfn(1:3, istate)
            rac = pbc(ra, rc, cell)
            dist = rac(1)*rac(1)+rac(2)*rac(2)+rac(3)*rac(3)

            IF (dist < 1.0_dp) THEN
               CALL cp_fm_get_submatrix(mo_coeff, vecbuffer, 1, istate, &
                                        nao, 1, transpose=.TRUE.)
               sto_state_overlap = 0.0_dp
               DO i = 1, SIZE(stogto_overlap(my_kind)%array, 1)
                  component = 0.0_dp
                  DO j = 1, SIZE(stogto_overlap(my_kind)%array, 2)
                     isgf = first_sgf(iatom)+j-1
                     component = component+stogto_overlap(my_kind)%array(i, j)*vecbuffer(1, isgf)
                  END DO ! j size
                  sto_state_overlap = sto_state_overlap+ &
                                      component*component
               END DO ! i size
               IF (sto_state_overlap .GT. max_overlap) THEN
                  max_overlap = sto_state_overlap
                  my_state = istate
               END IF
            END IF
            xas_estate = my_state
         END DO !  istate

         CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_get_submatrix(mo_coeff, vecbuffer, 1, xas_estate, &
                                  nao, 1, transpose=.TRUE.)
         CALL cp_fm_set_submatrix(excvec_coeff, vecbuffer2, 1, 1, &
                                  nao, 1, transpose=.TRUE.)
         !
      END IF

      CALL mp_sync(para_env%group)
      !Calculate the virtual states from the KS matrix matrix_ks(1)
      IF (nvirtual2 .GT. 0) THEN
         NULLIFY (mo_coeff)
         CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,/,T2,A,I5,A,I6,A)") " Calculation of ", nvirtual2, &
               " additional virtual states of the subspace complementary to the "// &
               " lowest ", nmo, " states"
         END IF

         NULLIFY (uno_orbs, uno_evals, local_preconditioner)
         ALLOCATE (local_preconditioner)
         CALL init_preconditioner(local_preconditioner, para_env=para_env, &
                                  blacs_env=blacs_env)

         CALL make_preconditioner(local_preconditioner, &
                                  precon_type=ot_precond_full_kinetic, &
                                  solver_type=ot_precond_solver_default, &
                                  matrix_h=matrix_ks(1)%matrix, &
                                  matrix_s=matrix_s(1)%matrix, &
                                  matrix_t=kinetic(1)%matrix, &
                                  convert_precond_to_dbcsr=.TRUE., &
                                  mo_set=mos(1)%mo_set, energy_gap=0.2_dp)

         CALL get_xas_env(xas_env=xas_env, unoccupied_orbs=uno_orbs, &
                          unoccupied_evals=uno_evals, unoccupied_eps=uno_eps, unoccupied_max_iter=uno_iter)
         CALL cp_fm_init_random(uno_orbs, nvirtual2)

         CALL ot_eigensolver(matrix_h=matrix_ks(1)%matrix, matrix_s=matrix_s(1)%matrix, &
                             matrix_c_fm=uno_orbs, matrix_orthogonal_space_fm=mo_coeff, &
                             preconditioner=local_preconditioner, eps_gradient=uno_eps, &
                             iter_max=uno_iter, size_ortho_space=nmo)

         CALL calculate_subspace_eigenvalues(uno_orbs, matrix_ks(1)%matrix, &
                                             uno_evals, do_rotation=.TRUE.)
         CALL destroy_preconditioner(local_preconditioner)

         DEALLOCATE (local_preconditioner)
      END IF

      CALL mp_sync(para_env%group)
      ! Prapare arrays for the calculation of the spectrum
      IF (.NOT. xas_control%xas_method == xas_dscf) THEN
         ! Copy the final vectors in the array
         NULLIFY (all_vectors, all_evals)
         CALL get_xas_env(xas_env=xas_env, all_vectors=all_vectors, &
                          all_evals=all_evals)
         CALL get_mo_set(mos(1)%mo_set, eigenvalues=eigenvalues, mo_coeff=mo_coeff, &
                         nmo=nmo)

         CALL cp_fm_to_fm(mo_coeff, all_vectors, ncol=nmo, &
                          source_start=1, target_start=1)
         DO istate = 1, nmo
            all_evals(istate) = eigenvalues(istate)
         ENDDO
         IF (nvirtual2 .GT. 0) THEN
            CALL cp_fm_to_fm(uno_orbs, all_vectors, ncol=nvirtual2, &
                             source_start=1, target_start=1+nmo)
            DO istate = 1, nvirtual2
               all_evals(istate+nmo) = uno_evals(istate)
            END DO
         END IF
      END IF

      DEALLOCATE (vecbuffer)
      DEALLOCATE (vecbuffer2)
      DEALLOCATE (centers_wfn, first_sgf)

      CALL timestop(handle)

   END SUBROUTINE cls_prepare_states

! **************************************************************************************************
!> \brief  SCF for emission spectra calculations: vacancy in valence
!> \param qs_env the qs_env, the scf_env and xas_env live in
!> \param xas_env the environment for XAS  calculations
!> \param converged ...
!> \param should_stop ...
!> \par History
!>      05.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE xes_scf_once(qs_env, xas_env, converged, should_stop)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(xas_environment_type), POINTER                :: xas_env
      LOGICAL, INTENT(OUT)                               :: converged, should_stop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xes_scf_once', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, istate, nmo, nvirtual, &
                                                            nvirtual2, output_unit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: all_evals, eigenvalues
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_fm_type), POINTER                          :: all_vectors, mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: dft_section, scf_section, xas_section
      TYPE(xas_control_type), POINTER                    :: xas_control

      NULLIFY (dft_control, scf_control, scf_env, matrix_ks, mos, para_env, xas_control)
      NULLIFY (dft_section, xas_section, scf_section, all_vectors, mo_coeff, all_evals, eigenvalues)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(xas_env)))CALL cp__a("xas_tp_scf.F",655)
      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      matrix_ks=matrix_ks, mos=mos, para_env=para_env)

      xas_control => dft_control%xas_control
      CALL get_xas_env(xas_env, scf_control=scf_control)

      dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")
      xas_section => section_vals_get_subs_vals(dft_section, "XAS")
      scf_section => section_vals_get_subs_vals(xas_section, "SCF")

      IF (xas_env%homo_occ == 0) THEN

         NULLIFY (scf_env)
         CALL get_xas_env(xas_env, scf_env=scf_env)
         IF (.NOT. ASSOCIATED(scf_env)) THEN
            CALL qs_scf_env_initialize(qs_env, scf_env, scf_control, scf_section)
            CALL set_xas_env(xas_env, scf_env=scf_env)
            CALL scf_env_release(scf_env)
            CALL get_xas_env(xas_env=xas_env, scf_env=scf_env)
         ELSE
            CALL qs_scf_env_initialize(qs_env, scf_env, scf_control, scf_section)
         ENDIF

         CALL scf_env_do_scf(scf_env=scf_env, scf_control=scf_control, qs_env=qs_env, &
                             converged=converged, should_stop=should_stop)
         CALL scf_env_cleanup(scf_env)

      END IF

      !   The eigenstate of the KS Hamiltonian are nedeed
      NULLIFY (mo_coeff, eigenvalues)
      IF (scf_control%use_ot) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T10,A,/)") &
               "Get eigenstates and eigenvalues from ground state MOs"
         END IF
         DO ispin = 1, SIZE(mos)
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, &
                            eigenvalues=eigenvalues)
            CALL calculate_subspace_eigenvalues(mo_coeff, &
                                                matrix_ks(ispin)%matrix, eigenvalues, &
                                                do_rotation=.TRUE.)
         END DO
      END IF
      CALL get_xas_env(xas_env=xas_env, all_vectors=all_vectors, &
                       all_evals=all_evals, nvirtual2=nvirtual2)
      CALL get_mo_set(mos(1)%mo_set, eigenvalues=eigenvalues, mo_coeff=mo_coeff, nmo=nmo)

      CALL cp_fm_to_fm(mo_coeff, all_vectors, ncol=nmo, &
                       source_start=1, target_start=1)
      DO istate = 1, nmo
         all_evals(istate) = eigenvalues(istate)
      ENDDO

      IF (nvirtual2 /= 0) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T10,A,/)") &
               "WARNING: for this XES calculation additional unoccupied MOs are not needed"
         END IF
         nvirtual2 = 0
         nvirtual = nmo
         CALL set_xas_env(xas_env=xas_env, nvirtual=nvirtual, nvirtual2=nvirtual2)
      END IF

      CALL timestop(handle)

   END SUBROUTINE xes_scf_once

END MODULE xas_tp_scf
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief driver for the xas calculation and xas_scf for the tp method
!> \par History
!>      created 05.2005
!>      replace overlap integral routine [07.2014,JGH]
!> \author MI (05.2005)
! **************************************************************************************************
MODULE xas_methods

   USE ai_contraction,                  ONLY: block_add,&
                                              contraction
   USE ai_overlap,                      ONLY: overlap_ab
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: &
        allocate_sto_basis_set, create_gto_from_sto_basis, deallocate_sto_basis_set, &
        get_gto_basis_set, gto_basis_set_type, init_orb_basis_set, set_sto_basis_set, srules, &
        sto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_array_utils_r,                ONLY: cp_2d_r_p_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        convert_offsets_to_sizes, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_init, cp_dbcsr_p_type, cp_dbcsr_set, cp_dbcsr_type, dbcsr_distribution_obj, &
        dbcsr_type_antisymmetric
   USE cp_dbcsr_operations,             ONLY: copy_fm_to_dbcsr,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_external_control,             ONLY: external_control
   USE cp_fm_pool_types,                ONLY: fm_pool_create_fm
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_element,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_p_type,&
                                              cp_fm_set_all,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: &
        do_loc_none, state_loc_list, state_loc_range, xas_1s_type, xas_2p_type, xas_2s_type, &
        xas_dip_len2, xas_dip_vel, xas_dscf, xas_tp_fh, xas_tp_flex, xas_tp_hh, xas_tp_xfh, &
        xas_tp_xhh, xes_tp_val
   USE input_section_types,             ONLY: section_get_lval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: ncoset
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE periodic_table,                  ONLY: ptable
   USE physcon,                         ONLY: evolt
   USE qs_diis,                         ONLY: qs_diis_b_clear,&
                                              qs_diis_b_create
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_loc_methods,                  ONLY: qs_loc_driver,&
                                              qs_print_cubes
   USE qs_loc_types,                    ONLY: localized_wfn_control_type,&
                                              qs_loc_env_create,&
                                              qs_loc_env_new_type,&
                                              qs_loc_env_release
   USE qs_loc_utils,                    ONLY: qs_loc_control_init,&
                                              qs_loc_env_init,&
                                              set_loc_centers,&
                                              set_loc_wfn_lists
   USE qs_matrix_pools,                 ONLY: mpools_get,&
                                              qs_matrix_pools_type
   USE qs_mo_io,                        ONLY: write_mo_set
   USE qs_mo_methods,                   ONLY: calculate_subspace_eigenvalues
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              set_mo_set
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_operators_ao,                 ONLY: p_xyz_ao,&
                                              rRc_xyz_ao
   USE qs_pdos,                         ONLY: calculate_projected_dos
   USE qs_scf,                          ONLY: scf_env_cleanup
   USE qs_scf_initialization,           ONLY: qs_scf_env_initialize
   USE qs_scf_types,                    ONLY: qs_scf_env_type,&
                                              scf_env_release
   USE scf_control_types,               ONLY: scf_c_create,&
                                              scf_c_read_parameters,&
                                              scf_c_release,&
                                              scf_control_type
   USE xas_control,                     ONLY: read_xas_control,&
                                              write_xas_control,&
                                              xas_control_create,&
                                              xas_control_type
   USE xas_env_types,                   ONLY: get_xas_env,&
                                              set_xas_env,&
                                              xas_env_create,&
                                              xas_env_release,&
                                              xas_environment_type
   USE xas_restart,                     ONLY: xas_read_restart
   USE xas_tp_scf,                      ONLY: xas_do_tp_scf,&
                                              xes_scf_once
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xas_methods'

! *** Public subroutines ***

   PUBLIC :: xas

CONTAINS

! **************************************************************************************************
!> \brief Driver for xas calculations
!>      The initial mos are prepared
!>      A loop on the atoms to be excited is started
!>      For each atom the state to be excited is identified
!>      An scf optimization using the TP scheme or TD-DFT is used
!>      to evaluate the spectral energies and oscillator strengths
!> \param qs_env the qs_env, the xas_env lives in
!> \param dft_control ...
!> \par History
!>      05.2005 created [MI]
!> \author MI
!> \note
!>      the iteration counter is not finilized yet
!>      only the transition potential approach is active
!>      the localization can be switched off, otherwise
!>      it uses by default the berry phase approach
!>      The number of states to be localized is xas_control%nexc_search
!>      In general only the core states are needed
! **************************************************************************************************
   SUBROUTINE xas(qs_env, dft_control)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(dft_control_type), POINTER                    :: dft_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xas', routineP = moduleN//':'//routineN

      INTEGER :: handle, homo, i, iat, iatom, ispin, my_homo(2), my_nelectron(2), nao, nexc_atoms, &
         nexc_search, nmo, nspins, output_unit, state_to_be_excited
      INTEGER, DIMENSION(2)                              :: added_mos
      INTEGER, DIMENSION(:), POINTER                     :: state_of_atom
      LOGICAL                                            :: ch_method_flags, converged, my_uocc(2), &
                                                            should_stop, skip_scf, &
                                                            transition_potential
      REAL(dp)                                           :: maxocc, occ_estate, tmp, xas_nelectron
      REAL(dp), DIMENSION(:), POINTER                    :: eigenvalues
      REAL(dp), DIMENSION(:, :), POINTER                 :: vecbuffer
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, op_sm, ostrength_sm
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_b
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: groundstate_coeff
      TYPE(cp_fm_type), POINTER                          :: all_vectors, excvec_coeff, mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: dft_section, loc_section, &
                                                            print_loc_section, scf_section, &
                                                            xas_section
      TYPE(xas_control_type), POINTER                    :: xas_control
      TYPE(xas_environment_type), POINTER                :: xas_env

      CALL timeset(routineN, handle)

      transition_potential = .FALSE.
      skip_scf = .FALSE.
      converged = .TRUE.
      should_stop = .FALSE.
      ch_method_flags = .FALSE.

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      NULLIFY (xas_env, groundstate_coeff, ostrength_sm, op_sm)
      NULLIFY (excvec_coeff, state_of_atom, qs_loc_env, cell, scf_env)
      NULLIFY (matrix_ks)
      NULLIFY (all_vectors, state_of_atom, xas_control)
      NULLIFY (vecbuffer, op_sm, mo_coeff_b)
      NULLIFY (dft_section, xas_section, scf_section, loc_section, print_loc_section)
      dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")
      xas_section => section_vals_get_subs_vals(dft_section, "XAS")
      scf_section => section_vals_get_subs_vals(xas_section, "SCF")
      loc_section => section_vals_get_subs_vals(xas_section, "LOCALIZE")
      print_loc_section => section_vals_get_subs_vals(loc_section, "PRINT")

      output_unit = cp_print_key_unit_nr(logger, xas_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".Log")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T3,A,/,T25,A,/,T3,A,/)") &
            REPEAT("=", 77), &
            "START CORE LEVEL SPECTROSCOPY CALCULATION", &
            REPEAT("=", 77)
      END IF

!   Create the xas environment
      CALL get_qs_env(qs_env, xas_env=xas_env)
      IF (.NOT. ASSOCIATED(xas_env)) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T5,A)") &
               "Create and initialize the xas environment"
         END IF
         CALL xas_env_create(xas_env)
         CALL xas_env_init(xas_env, qs_env, dft_section, logger)
         xas_control => dft_control%xas_control
         CALL set_qs_env(qs_env, xas_env=xas_env)
         CALL xas_env_release(xas_env)
         CALL get_qs_env(qs_env, xas_env=xas_env)
      END IF

!   Initialize the type of calculation
      NULLIFY (atomic_kind_set, qs_kind_set, scf_control, mos, para_env, particle_set)
      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, &
                      cell=cell, scf_control=scf_control, &
                      matrix_ks=matrix_ks, mos=mos, para_env=para_env, &
                      particle_set=particle_set)

!   The eigenstate of the KS Hamiltonian are nedeed
      NULLIFY (mo_coeff, eigenvalues)
      IF (scf_control%use_ot) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T10,A,/)") &
               "Get eigenstates and eigenvalues from ground state MOs"
         END IF
         DO ispin = 1, dft_control%nspins
            CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nao=nao, nmo=nmo, &
                            eigenvalues=eigenvalues, homo=homo)
            CALL calculate_subspace_eigenvalues(mo_coeff, &
                                                matrix_ks(ispin)%matrix, eigenvalues, &
                                                do_rotation=.TRUE.)
         END DO
      END IF
!   In xas SCF we need to use the same number of MOS as for GS
      added_mos = scf_control%added_mos
      NULLIFY (scf_control)
!   Consider to use get function for this
      CALL get_xas_env(xas_env, scf_control=scf_control)
      scf_control%added_mos = added_mos

!   Set initial occupation numbers, and store the original ones
      my_homo = 0
      my_nelectron = 0
      DO ispin = 1, dft_control%nspins
         CALL get_mo_set(mos(ispin)%mo_set, nelectron=my_nelectron(ispin), maxocc=maxocc, &
                         homo=my_homo(ispin), uniform_occupation=my_uocc(ispin))
      END DO

      nspins = dft_control%nspins
! at the moment the only implemented method for XAS and XES calculations
      transition_potential = .TRUE. !(xas_control%xas_method==xas_tp_hh).OR.&
      !                           (xas_control%xas_method==xas_tp_fh).OR.&
      !                          (xas_control%xas_method==xas_tp_xhh).OR.&
      !                         (xas_control%xas_method==xas_tp_xfh).OR.&
      !                        (xas_control%xas_method==xas_dscf)
      IF (nspins == 1 .AND. transition_potential) THEN
         CALL cp__b("xas_methods.F",285,"XAS with TP method requires LSD calculations")
      END IF

!   Set of states among which there is the state to be excited
      CALL get_mo_set(mos(1)%mo_set, nao=nao, homo=homo)
      IF (xas_control%nexc_search < 0) xas_control%nexc_search = homo
      nexc_search = xas_control%nexc_search

      CALL get_xas_env(xas_env=xas_env, &
                       state_of_atom=state_of_atom, all_vectors=all_vectors, &
                       groundstate_coeff=groundstate_coeff, excvec_coeff=excvec_coeff, &
                       nexc_atoms=nexc_atoms, occ_estate=occ_estate, xas_nelectron=xas_nelectron)

      CALL set_xas_env(xas_env=xas_env, nexc_search=nexc_search)

      !Define the qs_loc_env : to find centers, spread and possibly localize them
      CALL get_xas_env(xas_env=xas_env, qs_loc_env=qs_loc_env)
      IF (qs_loc_env%do_localize) THEN
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T10,A,/)") &
               "Localize a sub-set of MOs with spin alpha,"// &
               " to better identify the core states"
            IF ( &
               qs_loc_env%localized_wfn_control%set_of_states == state_loc_range) THEN
               WRITE (UNIT=output_unit, FMT="( A , I7, A, I7)") " The sub-set contains states from ", &
                  qs_loc_env%localized_wfn_control%lu_bound_states(1, 1), " to ", &
                  qs_loc_env%localized_wfn_control%lu_bound_states(2, 1)
            ELSEIF (qs_loc_env%localized_wfn_control%set_of_states == state_loc_list) THEN
               WRITE (UNIT=output_unit, FMT="( A )") " The sub-set contains states given in the input list"
            END IF

         END IF
         CALL qs_loc_driver(qs_env, qs_loc_env, print_loc_section, myspin=1)
      END IF

      IF(.NOT.(ASSOCIATED(groundstate_coeff)))CALL cp__a("xas_methods.F",320)
      DO ispin = 1, nspins
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, mo_coeff_b=mo_coeff_b, nmo=nmo)
         CALL cp_fm_to_fm(mo_coeff, groundstate_coeff(ispin)%matrix, nmo, 1, 1)
         IF (ASSOCIATED(mo_coeff_b)) THEN

         END IF
      END DO

!   SCF for only XES using occupied core and empty homo (only one SCF)
!   Probably better not to do the localization in this case, but only single out the
!   core orbital for the specific atom for which the spectrum is computed
      IF (xas_control%xas_method == xes_tp_val .AND. &
          xas_control%xes_core_occupation == 1.0_dp) THEN
         IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,/,T10,A)') &
            "START Core Level Spectroscopy Calculation for the Emission Spectrum"
         IF (xas_control%xes_homo_occupation == 1) THEN
            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(T10,A,/,A)') &
               "The core state is fully occupied and XES from ground state calculation.", &
               " No SCF is needed, MOS already available"
         ELSE IF (xas_control%xes_homo_occupation == 0) THEN
            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(T10,A,/,A)') &
               "The core state is fully occupied and the homo is empty", &
               " (final state of the core hole decay). Only one SCF is needed (not one per atom)"
         END IF
         skip_scf = .TRUE.

         CALL set_xas_env(xas_env=xas_env, xas_estate=-1, homo_occ=xas_control%xes_homo_occupation)
         CALL xes_scf_once(qs_env, xas_env, converged, should_stop)

         IF (converged .AND. .NOT. should_stop .AND. xas_control%xes_homo_occupation == 0) THEN
            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,T10,A,I6)') &
               "SCF with empty homo converged "
         ELSE IF (.NOT. converged .OR. should_stop) THEN
            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,T10,A,I6)') &
               "SCF with empty homo NOT converged"
            ! Release what has to be released
            IF (ASSOCIATED(vecbuffer)) THEN
               DEALLOCATE (vecbuffer)
               DEALLOCATE (op_sm)
            END IF

            DO ispin = 1, dft_control%nspins
               CALL set_mo_set(mos(ispin)%mo_set, homo=my_homo(ispin), &
                               uniform_occupation=my_uocc(ispin), nelectron=my_nelectron(ispin))
               CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
               CALL cp_fm_to_fm(groundstate_coeff(ispin)%matrix, mos(ispin)%mo_set%mo_coeff, nmo, 1, 1)
            END DO

            IF (output_unit > 0) THEN
               WRITE (UNIT=output_unit, FMT="(/,T3,A,/,T25,A,/,T3,A,/)") &
                  REPEAT("=", 77), &
                  "END CORE LEVEL SPECTROSCOPY CALCULATION", &
                  REPEAT("=", 77)
            END IF

            CALL xas_env_release(qs_env%xas_env)

            CALL cp_print_key_finished_output(output_unit, logger, xas_section, &
                                              "PRINT%PROGRAM_RUN_INFO")
            CALL timestop(handle)
            RETURN
         END IF
      END IF

      ! Assign the character of the selected core states
      ! through the overlap with atomic-like states
      CALL cls_assign_core_states(xas_control, xas_env, qs_loc_env%localized_wfn_control, &
                                  qs_env)

      IF (skip_scf) THEN
         CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff)
         CALL cp_fm_to_fm(mo_coeff, all_vectors, ncol=nexc_search, &
                          source_start=1, target_start=1)
      END IF

      ALLOCATE (vecbuffer(1, nao))
      ALLOCATE (op_sm(3))

      ! copy the coefficients of the mos in a temporary fm with the right structure
      IF (transition_potential) THEN

         ! Calculate the operator
         CALL get_xas_env(xas_env=xas_env, ostrength_sm=ostrength_sm)
         DO i = 1, 3
            NULLIFY (op_sm(i)%matrix)
            op_sm(i)%matrix => ostrength_sm(i)%matrix
         END DO
         IF (xas_control%dipole_form == xas_dip_vel) THEN
            CALL p_xyz_ao(op_sm, qs_env)
         END IF
      END IF

      ! DO SCF if required
      DO iat = 1, nexc_atoms
         iatom = xas_env%exc_atoms(iat)
         ! determine which state has to be excited in the global list
         state_to_be_excited = state_of_atom(iat)

         ! Take the state_to_be_excited vector from the full set and copy into excvec_coeff
         CALL get_mo_set(mos(1)%mo_set, nmo=nmo)
         CALL get_xas_env(xas_env, occ_estate=occ_estate, xas_nelectron=xas_nelectron)
         tmp = xas_nelectron+1.0_dp-occ_estate
         IF (nmo < tmp) &
            CALL cp__b("xas_methods.F",424,"CLS: the required method needs added_mos to the ground state")
         ! If the restart file for this atom exists, the mos and the
         ! occupation numbers are overwritten
         ! It is necessary that the restart is for the same xas method
         ! otherwise the number of electrons and the occupation numbers
         ! may not  be consistent
         IF (xas_control%xas_restart) THEN
            CALL xas_read_restart(xas_env, xas_section, qs_env, xas_control%xas_method, iatom, &
                                  state_to_be_excited)
         END IF

         CALL set_xas_env(xas_env=xas_env, xas_estate=state_to_be_excited)
         CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff)
         IF(.NOT.(ASSOCIATED(excvec_coeff)))CALL cp__a("xas_methods.F",437)
         CALL cp_fm_get_submatrix(mo_coeff, vecbuffer, 1, state_to_be_excited, &
                                  nao, 1, transpose=.TRUE.)
         CALL cp_fm_set_submatrix(excvec_coeff, vecbuffer, 1, 1, &
                                  nao, 1, transpose=.TRUE.)

         IF (transition_potential) THEN

            IF (.NOT. skip_scf) THEN
               IF (output_unit > 0) THEN
                  WRITE (UNIT=output_unit, FMT='(/,T5,A)') REPEAT("-", 75)
                  IF (xas_control%xas_method == xas_dscf) THEN
                     WRITE (UNIT=output_unit, FMT='(/,/,T10,A,I6)') &
                        "START DeltaSCF for the first excited state from the core state of ATOM ", iatom
                  ELSE
                     WRITE (UNIT=output_unit, FMT='(/,T10,A,I6)') &
                        "Start Core Level Spectroscopy Calculation with TP approach for ATOM ", iatom
                     WRITE (UNIT=output_unit, FMT='(T10,A,T50,f10.4)') "Occupation of the core orbital", &
                        occ_estate
                     WRITE (UNIT=output_unit, FMT='(T10,A,T50,f10.4)') "Number of electrons in Spin 1", &
                        xas_nelectron
                  END IF
               END IF

               CALL get_xas_env(xas_env=xas_env, scf_env=scf_env)
               IF (.NOT. ASSOCIATED(scf_env)) THEN
                  CALL qs_scf_env_initialize(qs_env, scf_env, scf_control, scf_section)
!     Moved here from qs_scf_env_initialize to be able to have more scf_env
                  CALL set_xas_env(xas_env, scf_env=scf_env)
                  CALL scf_env_release(scf_env)
                  CALL get_xas_env(xas_env=xas_env, scf_env=scf_env)
               ELSE
                  CALL qs_scf_env_initialize(qs_env, scf_env, scf_control, scf_section)
               ENDIF

               DO ispin = 1, SIZE(mos)
                  IF (ASSOCIATED(mos(ispin)%mo_set%mo_coeff_b)) THEN !fm->dbcsr
                     CALL copy_fm_to_dbcsr(mos(ispin)%mo_set%mo_coeff, &
                                           mos(ispin)%mo_set%mo_coeff_b) !fm->dbcsr
                  ENDIF !fm->dbcsr
               ENDDO !fm->dbcsr

               IF (.NOT. scf_env%skip_diis) THEN
                  IF (.NOT. ASSOCIATED(scf_env%scf_diis_buffer)) THEN
                     CALL qs_diis_b_create(scf_env%scf_diis_buffer, nbuffer=scf_control%max_diis)
                  END IF
                  CALL qs_diis_b_clear(scf_env%scf_diis_buffer)
               END IF

               CALL xas_do_tp_scf(dft_control, xas_env, iatom, scf_env, qs_env, &
                                  xas_section, scf_section, converged, should_stop)

               CALL external_control(should_stop, "CLS", target_time=qs_env%target_time, &
                                     start_time=qs_env%start_time)
               IF (should_stop) THEN
                  CALL scf_env_cleanup(scf_env)
                  EXIT
               END IF

            END IF
            ! SCF DONE

!   *** Write last wavefunction to screen ***
            DO ispin = 1, nspins
               CALL write_mo_set(mos(ispin)%mo_set, atomic_kind_set, qs_kind_set, particle_set, 4, &
                                 dft_section)
            ENDDO

         ELSE
            ! Core level spectroscopy by TDDFT is not yet implemented
            ! the states defined by the rotation are the ground state orbitals
            ! the initial state from which I excite should be localized
            ! I take the excitations from lumo to nmo
         END IF

         IF (converged) THEN
            CALL cls_calculate_spectrum(xas_control, xas_env, qs_env, xas_section, &
                                        iatom)
         ELSE
            IF (output_unit > 0) WRITE (UNIT=output_unit, FMT='(/,/,T10,A,I6)') &
               "SCF with core hole NOT converged for ATOM ", iatom
         END IF

         IF (.NOT. skip_scf) THEN
            ! Reset the initial core orbitals.
            ! The valence orbitals are taken from the last SCF,
            ! it should be a better initial guess
            CALL get_qs_env(qs_env, mos=mos)
            DO ispin = 1, nspins
               CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
               CALL cp_fm_to_fm(groundstate_coeff(ispin)%matrix, mos(ispin)%mo_set%mo_coeff, nmo, 1, 1)
            END DO
            IF (iat == nexc_atoms) THEN
               CALL scf_env_cleanup(scf_env)
               CALL scf_env_release(xas_env%scf_env)
            END IF
         END IF

      END DO ! iat = 1,nexc_atoms

      ! END of Calculation

      ! Release what has to be released
      IF (ASSOCIATED(vecbuffer)) THEN
         DEALLOCATE (vecbuffer)
         DEALLOCATE (op_sm)
      END IF

      DO ispin = 1, dft_control%nspins
         CALL set_mo_set(mos(ispin)%mo_set, homo=my_homo(ispin), &
                         uniform_occupation=my_uocc(ispin), nelectron=my_nelectron(ispin))
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
         CALL cp_fm_to_fm(groundstate_coeff(ispin)%matrix, mos(ispin)%mo_set%mo_coeff, nmo, 1, 1)
      END DO

      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T3,A,/,T25,A,/,T3,A,/)") &
            REPEAT("=", 77), &
            "END CORE LEVEL SPECTROSCOPY CALCULATION", &
            REPEAT("=", 77)
      END IF

      CALL xas_env_release(qs_env%xas_env)

      CALL cp_print_key_finished_output(output_unit, logger, xas_section, &
                                        "PRINT%PROGRAM_RUN_INFO")
      CALL timestop(handle)

   END SUBROUTINE xas

! **************************************************************************************************
!> \brief allocate and initialize the structure needed for the xas calculation
!> \param xas_env the environment for XAS  calculations
!> \param qs_env the qs_env, the xas_env lives in
!> \param dft_section ...
!> \param logger ...
!> \par History
!>      05.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE xas_env_init(xas_env, qs_env, dft_section, logger)

      TYPE(xas_environment_type), POINTER                :: xas_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: dft_section
      TYPE(cp_logger_type), POINTER                      :: logger

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_env_init', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_sto
      INTEGER :: homo, i, iat, iatom, ik, ikind, ispin, j, l, lfomo, n_mo(2), n_rep, nao, natom, &
         ncubes, nelectron, nexc_atoms, nexc_search, nj, nk, nkind, nmo, nmoloc(2), norb(0:3), &
         nsgf_gto, nsgf_sto, nspins, nvirtual, nvirtual2, nwork
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf, kind_type_tmp, kind_z_tmp, &
                                                            last_sgf
      INTEGER, DIMENSION(4, 7)                           :: ne
      INTEGER, DIMENSION(:), POINTER                     :: bounds, list, lq, nq, row_blk_sizes
      LOGICAL                                            :: ihavethis
      REAL(dp)                                           :: nele, occ_estate, occ_homo, &
                                                            occ_homo_plus, zatom
      REAL(dp), DIMENSION(:), POINTER                    :: sto_zet
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_struct_type), POINTER                   :: tmp_fm_struct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dbcsr_distribution_obj), POINTER              :: dbcsr_dist
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: loc_section, xas_section
      TYPE(sto_basis_set_type), POINTER                  :: sto_basis_set
      TYPE(xas_control_type), POINTER                    :: xas_control

      n_mo(1:2) = 0
      IF(.NOT.(ASSOCIATED(xas_env)))CALL cp__a("xas_methods.F",620)

      NULLIFY (atomic_kind_set, qs_kind_set, dft_control, scf_control, matrix_s, mos, mpools)
      NULLIFY (para_env, particle_set, xas_control)
      NULLIFY (qs_loc_env)
      NULLIFY (sab_orb)
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      dft_control=dft_control, &
                      mpools=mpools, &
                      matrix_s=matrix_s, mos=mos, &
                      para_env=para_env, particle_set=particle_set, &
                      sab_orb=sab_orb, &
                      dbcsr_dist=dbcsr_dist)

      xas_section => section_vals_get_subs_vals(dft_section, "XAS")
      CALL xas_control_create(dft_control%xas_control)
      CALL read_xas_control(dft_control%xas_control, xas_section)
      CALL write_xas_control(dft_control%xas_control, dft_section)
      xas_control => dft_control%xas_control
      CALL scf_c_create(scf_control)
      CALL scf_c_read_parameters(scf_control, xas_section)
      CALL set_xas_env(xas_env, scf_control=scf_control)
      CALL scf_c_release(scf_control)
      CALL get_xas_env(xas_env, scf_control=scf_control)

      nexc_search = xas_control%nexc_search
      IF (nexc_search < 0) THEN
         ! ground state occupation
         CALL get_mo_set(mos(1)%mo_set, nmo=nmo, lfomo=lfomo)
         nexc_search = lfomo-1
      END IF
      nexc_atoms = xas_control%nexc_atoms
      ALLOCATE (xas_env%exc_atoms(nexc_atoms))
      xas_env%exc_atoms = xas_control%exc_atoms
      CALL set_xas_env(xas_env=xas_env, nexc_search=nexc_search, &
                       nexc_atoms=nexc_atoms)

      CALL mpools_get(mpools, ao_mo_fm_pools=xas_env%ao_mo_fm_pools)

      NULLIFY (mo_coeff)
      CALL get_mo_set(mos(1)%mo_set, nao=nao, homo=homo, nmo=nmo, mo_coeff=mo_coeff, nelectron=nelectron)

      nvirtual2 = 0
      IF (xas_control%added_mos .GT. 0) THEN
         nvirtual2 = MIN(xas_control%added_mos, nao-nmo)
         xas_env%unoccupied_eps = xas_control%eps_added
         xas_env%unoccupied_max_iter = xas_control%max_iter_added
      END IF
      nvirtual = nmo+nvirtual2

      n_mo(1:2) = nmo

      ALLOCATE (xas_env%centers_wfn(3, nexc_search))
      ALLOCATE (xas_env%atom_of_state(nexc_search))
      ALLOCATE (xas_env%type_of_state(nexc_search))
      ALLOCATE (xas_env%state_of_atom(nexc_atoms))
      ALLOCATE (xas_env%mykind_of_atom(nexc_atoms))
      nkind = SIZE(atomic_kind_set, 1)
      ALLOCATE (xas_env%mykind_of_kind(nkind))
      xas_env%mykind_of_kind = 0

      ! create a new matrix structure nao x 1
      NULLIFY (tmp_fm_struct)
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                               ncol_global=1, para_env=para_env, context=mo_coeff%matrix_struct%context)
      CALL cp_fm_create(xas_env%excvec_coeff, tmp_fm_struct)
      CALL cp_fm_struct_release(tmp_fm_struct)

      NULLIFY (tmp_fm_struct)
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=1, &
                               ncol_global=nexc_search, para_env=para_env, &
                               context=mo_coeff%matrix_struct%context)
      CALL cp_fm_create(xas_env%excvec_overlap, tmp_fm_struct)
      CALL cp_fm_struct_release(tmp_fm_struct)

      nspins = SIZE(mos, 1)

      ! initialize operators for the calculation of the oscillator strengths
      IF (xas_control%xas_method == xas_tp_hh) THEN
         occ_estate = 0.5_dp
         nele = REAL(nelectron, dp)-0.5_dp
         occ_homo = 1.0_dp
         occ_homo_plus = 0._dp
      ELSEIF (xas_control%xas_method == xas_tp_xhh) THEN
         occ_estate = 0.5_dp
         nele = REAL(nelectron, dp)
         occ_homo = 1.0_dp
         occ_homo_plus = 0.5_dp
      ELSEIF (xas_control%xas_method == xas_tp_fh) THEN
         occ_estate = 0.0_dp
         nele = REAL(nelectron, dp)-1.0_dp
         occ_homo = 1.0_dp
         occ_homo_plus = 0._dp
      ELSEIF (xas_control%xas_method == xas_tp_xfh) THEN
         occ_estate = 0.0_dp
         nele = REAL(nelectron, dp)
         occ_homo = 1.0_dp
         occ_homo_plus = 1._dp
      ELSEIF (xas_control%xas_method == xes_tp_val) THEN
         occ_estate = xas_control%xes_core_occupation
         nele = REAL(nelectron, dp)-xas_control%xes_core_occupation
         occ_homo = xas_control%xes_homo_occupation
      ELSEIF (xas_control%xas_method == xas_dscf) THEN
         occ_estate = 0.0_dp
         nele = REAL(nelectron, dp)
         occ_homo = 1.0_dp
         occ_homo_plus = 1._dp
      ELSEIF (xas_control%xas_method == xas_tp_flex) THEN
         nele = REAL(xas_control%nel_tot, dp)
         occ_estate = REAL(xas_control%xas_core_occupation, dp)
         IF (nele < 0.0_dp) nele = REAL(nelectron, dp)-(1.0_dp-occ_estate)
         occ_homo = 1.0_dp
      ENDIF
      CALL set_xas_env(xas_env=xas_env, occ_estate=occ_estate, xas_nelectron=nele, &
                       nvirtual2=nvirtual2, nvirtual=nvirtual, homo_occ=occ_homo)

      ! Initialize the list of orbitals for cube files printing
      IF (BTEST(cp_print_key_should_output(logger%iter_info, xas_section, &
                                           "PRINT%CLS_FUNCTION_CUBES"), cp_p_file)) THEN
         NULLIFY (bounds, list)
         CALL section_vals_val_get(xas_section, &
                                   "PRINT%CLS_FUNCTION_CUBES%CUBES_LU_BOUNDS", &
                                   i_vals=bounds)
         ncubes = bounds(2)-bounds(1)+1
         IF (ncubes > 0) THEN
            ALLOCATE (xas_control%list_cubes(ncubes))

            DO ik = 1, ncubes
               xas_control%list_cubes(ik) = bounds(1)+(ik-1)
            END DO
         END IF

         IF (.NOT. ASSOCIATED(xas_control%list_cubes)) THEN
            CALL section_vals_val_get(xas_section, &
                                      "PRINT%CLS_FUNCTION_CUBES%CUBES_LIST", &
                                      n_rep_val=n_rep)
            ncubes = 0
            DO ik = 1, n_rep
               NULLIFY (list)
               CALL section_vals_val_get(xas_section, &
                                         "PRINT%CLS_FUNCTION_CUBES%CUBES_LIST", &
                                         i_rep_val=ik, i_vals=list)
               IF (ASSOCIATED(list)) THEN
                  CALL reallocate(xas_control%list_cubes, 1, ncubes+SIZE(list))
                  DO i = 1, SIZE(list)
                     xas_control%list_cubes(i+ncubes) = list(i)
                  END DO
                  ncubes = ncubes+SIZE(list)
               END IF
            END DO ! ik
         END IF

         IF (.NOT. ASSOCIATED(xas_control%list_cubes)) THEN
            ncubes = MAX(10, xas_control%added_mos/10)
            ncubes = MIN(ncubes, xas_control%added_mos)
            ALLOCATE (xas_control%list_cubes(ncubes))
            DO ik = 1, ncubes
               xas_control%list_cubes(ik) = homo+ik
            END DO
         END IF
      ELSE
         NULLIFY (xas_control%list_cubes)
      ENDIF

      NULLIFY (tmp_fm_struct)
      nwork = nvirtual
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, ncol_global=nwork, &
                               para_env=para_env, context=mo_coeff%matrix_struct%context)
      CALL cp_fm_create(xas_env%fm_work, tmp_fm_struct)
      CALL cp_fm_struct_release(tmp_fm_struct)

      ALLOCATE (xas_env%groundstate_coeff(nspins))
      DO ispin = 1, nspins
         NULLIFY (xas_env%groundstate_coeff(ispin)%matrix)
         CALL get_mo_set(mos(ispin)%mo_set, nao=nao, nmo=nmo)
         CALL fm_pool_create_fm(xas_env%ao_mo_fm_pools(ispin)%pool, &
                                xas_env%groundstate_coeff(ispin)%matrix, &
                                name="xas_env%mo0"//TRIM(ADJUSTL(cp_to_string(ispin))))
      END DO ! ispin

      NULLIFY (tmp_fm_struct)
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=1, &
                               ncol_global=nvirtual, para_env=para_env, &
                               context=mo_coeff%matrix_struct%context)
      ALLOCATE (xas_env%dip_fm_set(2, 3))
      DO i = 1, 3
         DO j = 1, 2
            NULLIFY (xas_env%dip_fm_set(j, i)%matrix)
            CALL cp_fm_create(xas_env%dip_fm_set(j, i)%matrix, tmp_fm_struct)
         END DO
      END DO
      CALL cp_fm_struct_release(tmp_fm_struct)

      !Array to store all the eigenstates: occupied and the required not occupied
      IF (nvirtual2 .GT. 0) THEN
         ALLOCATE (xas_env%unoccupied_evals(nvirtual2))
         NULLIFY (tmp_fm_struct)
         CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                                  ncol_global=nvirtual2, &
                                  para_env=para_env, context=mo_coeff%matrix_struct%context)
         NULLIFY (xas_env%unoccupied_orbs)
         CALL cp_fm_create(xas_env%unoccupied_orbs, tmp_fm_struct)
         CALL cp_fm_struct_release(tmp_fm_struct)
      END IF

      NULLIFY (tmp_fm_struct)
      CALL cp_fm_struct_create(tmp_fm_struct, nrow_global=nao, &
                               ncol_global=nvirtual, &
                               para_env=para_env, context=mo_coeff%matrix_struct%context)
      NULLIFY (xas_env%all_vectors)
      CALL cp_fm_create(xas_env%all_vectors, tmp_fm_struct)
      CALL cp_fm_struct_release(tmp_fm_struct)

      ! Array to store all the energies needed  for the spectrum
      ALLOCATE (xas_env%all_evals(nvirtual))

      IF (xas_control%dipole_form == xas_dip_len2) THEN
         CALL cp_dbcsr_allocate_matrix_set(xas_env%ostrength_sm, 3)
         DO i = 1, 3
            ALLOCATE (xas_env%ostrength_sm(i)%matrix)
            CALL cp_dbcsr_init(xas_env%ostrength_sm(i)%matrix)
            CALL cp_dbcsr_copy(xas_env%ostrength_sm(i)%matrix, matrix_s(1)%matrix, &
                               "xas_env%ostrength_sm"// &
                               "-"//TRIM(ADJUSTL(cp_to_string(i))))
            CALL cp_dbcsr_set(xas_env%ostrength_sm(i)%matrix, 0.0_dp)
         END DO
      ELSEIF (xas_control%dipole_form == xas_dip_vel) THEN
         !
         ! prepare for allocation
         natom = SIZE(particle_set, 1)
         ALLOCATE (first_sgf(natom))
         ALLOCATE (last_sgf(natom))
         CALL get_particle_set(particle_set, qs_kind_set, &
                               first_sgf=first_sgf, &
                               last_sgf=last_sgf)
         ALLOCATE (row_blk_sizes(natom))
         CALL convert_offsets_to_sizes(first_sgf, row_blk_sizes, last_sgf)
         DEALLOCATE (first_sgf)
         DEALLOCATE (last_sgf)
         !
         !
         CALL cp_dbcsr_allocate_matrix_set(xas_env%ostrength_sm, 3)
         ALLOCATE (xas_env%ostrength_sm(1)%matrix)
         CALL cp_dbcsr_init(xas_env%ostrength_sm(1)%matrix)
         CALL cp_dbcsr_create(matrix=xas_env%ostrength_sm(1)%matrix, &
                              name="xas_env%ostrength_sm", &
                              dist=dbcsr_dist, matrix_type=dbcsr_type_antisymmetric, &
                              row_blk_size=row_blk_sizes, col_blk_size=row_blk_sizes, &
                              nze=0, mutable_work=.TRUE.)
         CALL cp_dbcsr_alloc_block_from_nbl(xas_env%ostrength_sm(1)%matrix, sab_orb)
         CALL cp_dbcsr_set(xas_env%ostrength_sm(1)%matrix, 0.0_dp)
         DO i = 2, 3
            ALLOCATE (xas_env%ostrength_sm(i)%matrix)
            CALL cp_dbcsr_init(xas_env%ostrength_sm(i)%matrix)
            CALL cp_dbcsr_copy(xas_env%ostrength_sm(i)%matrix, xas_env%ostrength_sm(1)%matrix, &
                               "xas_env%ostrength_sm-"//TRIM(ADJUSTL(cp_to_string(i))))
            CALL cp_dbcsr_set(xas_env%ostrength_sm(i)%matrix, 0.0_dp)
         END DO

         DEALLOCATE (row_blk_sizes)
      END IF

      ! Define the qs_loc_env : to find centers, spread and possibly localize them
      IF (.NOT. (ASSOCIATED(xas_env%qs_loc_env))) THEN
         CALL qs_loc_env_create(qs_loc_env)
         CALL set_xas_env(xas_env=xas_env, qs_loc_env=qs_loc_env)
         CALL qs_loc_env_release(qs_loc_env)
         CALL get_xas_env(xas_env=xas_env, qs_loc_env=qs_loc_env)
         loc_section => section_vals_get_subs_vals(xas_section, "LOCALIZE")
         CALL qs_loc_control_init(qs_loc_env, loc_section, do_homo=.TRUE., &
                                  do_xas=.TRUE., nloc_xas=nexc_search)
         IF (.NOT. qs_loc_env%do_localize) THEN
            qs_loc_env%localized_wfn_control%localization_method = do_loc_none
         ELSE
            nmoloc = qs_loc_env%localized_wfn_control%nloc_states
            CALL set_loc_wfn_lists(qs_loc_env%localized_wfn_control, nmoloc, n_mo, nspins)
            CALL set_loc_centers(qs_loc_env%localized_wfn_control, nmoloc, nspins)
            CALL qs_loc_env_init(qs_loc_env, qs_loc_env%localized_wfn_control, &
                                 qs_env, myspin=1, do_localize=qs_loc_env%do_localize)
         END IF
      END IF

      !Type of state
      norb = 0
      ALLOCATE (nq(1), lq(1), sto_zet(1))
      IF (xas_control%state_type == xas_1s_type) THEN
         norb(0) = 1
         nq(1) = 1
         lq(1) = 0
      ELSEIF (xas_control%state_type == xas_2s_type) THEN
         norb(0) = 2
         nq(1) = 2
         lq(1) = 0
      ELSEIF (xas_control%state_type == xas_2p_type) THEN
         norb(0) = 2
         norb(1) = 1
         nq(1) = 2
         lq(1) = 1
      ELSE
         CALL cp__b("xas_methods.F",921,"XAS type of state not implemented")
      END IF

!   Find core orbitals of right angular momentum
      ALLOCATE (kind_type_tmp(nkind))
      ALLOCATE (kind_z_tmp(nkind))
      kind_type_tmp = 0
      kind_z_tmp = 0
      nk = 0
      DO iat = 1, nexc_atoms
         iatom = xas_env%exc_atoms(iat)
         NULLIFY (atomic_kind)
         atomic_kind => particle_set(iatom)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, kind_number=ikind)
         CALL get_qs_kind(qs_kind_set(ikind), zeff=zatom)
         ihavethis = .FALSE.
         DO ik = 1, nk
            IF (ikind == kind_type_tmp(ik)) THEN
               ihavethis = .TRUE.
               xas_env%mykind_of_atom(iat) = ik
               EXIT
            END IF
         END DO
         IF (.NOT. ihavethis) THEN
            nk = nk+1
            kind_type_tmp(nk) = ikind
            kind_z_tmp(nk) = INT(zatom)
            xas_env%mykind_of_atom(iat) = nk
            xas_env%mykind_of_kind(ikind) = nk
         END IF
      END DO ! iat

      ALLOCATE (xas_env%my_gto_basis(nk))
      ALLOCATE (xas_env%stogto_overlap(nk))
      DO ik = 1, nk
         NULLIFY (xas_env%my_gto_basis(ik)%gto_basis_set, sto_basis_set)
         ne = 0
         DO l = 1, 4 !lq(1)+1
            nj = 2*(l-1)+1
            DO i = l, 7 ! nq(1)
               ne(l, i) = ptable(kind_z_tmp(ik))%e_conv(l-1)-2*nj*(i-l)
               ne(l, i) = MAX(ne(l, i), 0)
               ne(l, i) = MIN(ne(l, i), 2*nj)
            END DO
         END DO

         sto_zet(1) = srules(kind_z_tmp(ik), ne, nq(1), lq(1))
         CALL allocate_sto_basis_set(sto_basis_set)
         name_sto = 'xas_tmp_sto'
         CALL set_sto_basis_set(sto_basis_set, nshell=1, nq=nq, &
                                lq=lq, zet=sto_zet, name=name_sto)
         CALL create_gto_from_sto_basis(sto_basis_set, &
                                        xas_env%my_gto_basis(ik)%gto_basis_set, xas_control%ngauss)
         CALL deallocate_sto_basis_set(sto_basis_set)
         xas_env%my_gto_basis(ik)%gto_basis_set%norm_type = 2
         CALL init_orb_basis_set(xas_env%my_gto_basis(ik)%gto_basis_set)

         ikind = kind_type_tmp(ik)
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, nsgf=nsgf_gto)
         CALL get_gto_basis_set(gto_basis_set=xas_env%my_gto_basis(ik)%gto_basis_set, nsgf=nsgf_sto)
         ALLOCATE (xas_env%stogto_overlap(ik)%array(nsgf_sto, nsgf_gto))

         CALL calc_stogto_overlap(xas_env%my_gto_basis(ik)%gto_basis_set, orb_basis_set, &
                                  xas_env%stogto_overlap(ik)%array)
      END DO

      DEALLOCATE (nq, lq, sto_zet)
      DEALLOCATE (kind_type_tmp, kind_z_tmp)

   END SUBROUTINE xas_env_init

! **************************************************************************************************
!> \brief Calculate and write the spectrum relative to the core level excitation
!>      of a specific atom. It works for TP approach, because of the definition
!>      of the oscillator strengths as  matrix elements of the dipole operator
!> \param xas_control ...
!> \param xas_env ...
!> \param qs_env ...
!> \param xas_section ...
!> \param iatom index of the excited atom
!> \par History
!>      03.2006 created [MI]
!> \author MI
!> \note
!>      for the tddft calculation should be re-thought
! **************************************************************************************************
   SUBROUTINE cls_calculate_spectrum(xas_control, xas_env, qs_env, xas_section, &
                                     iatom)

      TYPE(xas_control_type)                             :: xas_control
      TYPE(xas_environment_type), POINTER                :: xas_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: xas_section
      INTEGER, INTENT(IN)                                :: iatom

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cls_calculate_spectrum', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: homo, i, lfomo, nabs, nmo, nvirtual, &
                                                            output_unit, xas_estate
      LOGICAL                                            :: append_cube, length
      REAL(dp)                                           :: rc(3)
      REAL(dp), DIMENSION(:), POINTER                    :: all_evals
      REAL(dp), DIMENSION(:, :), POINTER                 :: sp_ab, sp_em
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op_sm, ostrength_sm
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: dip_fm_set
      TYPE(cp_fm_type), POINTER                          :: all_vectors, excvec_coeff, fm_work
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      NULLIFY (fm_work, ostrength_sm, op_sm, dip_fm_set)
      NULLIFY (all_evals, all_vectors, excvec_coeff)
      NULLIFY (mos, particle_set, sp_em, sp_ab)
      ALLOCATE (op_sm(3))

      CALL get_qs_env(qs_env=qs_env, &
                      mos=mos, particle_set=particle_set)

      CALL get_mo_set(mos(1)%mo_set, homo=homo, lfomo=lfomo, nmo=nmo)
      CALL get_xas_env(xas_env=xas_env, all_vectors=all_vectors, xas_estate=xas_estate, &
                       all_evals=all_evals, dip_fm_set=dip_fm_set, excvec_coeff=excvec_coeff, &
                       fm_work=fm_work, ostrength_sm=ostrength_sm, nvirtual=nvirtual)

      nabs = nvirtual-lfomo+1
      ALLOCATE (sp_em(6, homo))
      ALLOCATE (sp_ab(6, nabs))
      IF(.NOT.(ASSOCIATED(excvec_coeff)))CALL cp__a("xas_methods.F",1054)

      IF (.NOT. xas_control%xas_method == xas_dscf) THEN
         ! Calculate the spectrum
         IF (xas_control%dipole_form == xas_dip_len2) THEN
            rc(1:3) = particle_set(iatom)%r(1:3)
            DO i = 1, 3
               NULLIFY (op_sm(i)%matrix)
               op_sm(i)%matrix => ostrength_sm(i)%matrix
            END DO
            CALL rRc_xyz_ao(op_sm, qs_env, rc, order=1, minimum_image=.TRUE.)
            CALL spectrum_dip_vel(dip_fm_set, op_sm, mos, excvec_coeff, &
                                  all_vectors, all_evals, fm_work, &
                                  sp_em, sp_ab, xas_estate, nvirtual)
            DO i = 1, SIZE(ostrength_sm, 1)
               CALL cp_dbcsr_set(ostrength_sm(i)%matrix, 0.0_dp)
            END DO
         ELSE
            DO i = 1, 3
               NULLIFY (op_sm(i)%matrix)
               op_sm(i)%matrix => ostrength_sm(i)%matrix
            END DO
            CALL spectrum_dip_vel(dip_fm_set, op_sm, mos, excvec_coeff, &
                                  all_vectors, all_evals, fm_work, &
                                  sp_em, sp_ab, xas_estate, nvirtual)
         END IF
      END IF

      CALL get_mo_set(mos(1)%mo_set, lfomo=lfomo)
      ! writw thw spectrum, if the file exists it is appended
      IF (.NOT. xas_control%xas_method == xas_dscf) THEN
         length = (.NOT. xas_control%dipole_form == xas_dip_vel)
         CALL xas_write(sp_em, sp_ab, xas_estate, &
                        xas_section, iatom, lfomo, length=length)
      END IF

      DEALLOCATE (sp_em)
      DEALLOCATE (sp_ab)

      IF (BTEST(cp_print_key_should_output(logger%iter_info, xas_section, &
                                           "PRINT%CLS_FUNCTION_CUBES"), cp_p_file)) THEN
         append_cube = section_get_lval(xas_section, "PRINT%CLS_FUNCTION_CUBES%APPEND")
         CALL xas_print_cubes(xas_control, qs_env, xas_section, mos, all_vectors, &
                              iatom, append_cube)
      END IF

      IF (BTEST(cp_print_key_should_output(logger%iter_info, xas_section, &
                                           "PRINT%PDOS"), cp_p_file)) THEN
         CALL xas_pdos(qs_env, xas_section, mos, iatom)
      END IF

      DEALLOCATE (op_sm)

   END SUBROUTINE cls_calculate_spectrum

! **************************************************************************************************
!> \brief write the spectrum for each atom in a different output file
!> \param sp_em ...
!> \param sp_ab ...
!> \param estate ...
!> \param xas_section ...
!> \param iatom index of the excited atom
!> \param lfomo ...
!> \param length ...
!> \par History
!>      05.2005 created [MI]
!> \author MI
!> \note
!>      the iteration counter is not finilized yet
! **************************************************************************************************
   SUBROUTINE xas_write(sp_em, sp_ab, estate, xas_section, iatom, &
                        lfomo, length)

      REAL(dp), DIMENSION(:, :), POINTER                 :: sp_em, sp_ab
      INTEGER, INTENT(IN)                                :: estate
      TYPE(section_vals_type), POINTER                   :: xas_section
      INTEGER, INTENT(IN)                                :: iatom, lfomo
      LOGICAL, INTENT(IN)                                :: length

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_write', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: mittle_ab, mittle_em, my_act, my_pos
      INTEGER                                            :: i, istate, out_sp_ab, out_sp_em
      REAL(dp)                                           :: ene2
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      my_pos = "APPEND"
      my_act = "WRITE"

      mittle_em = "xes_at"//TRIM(ADJUSTL(cp_to_string(iatom)))
      out_sp_em = cp_print_key_unit_nr(logger, xas_section, "PRINT%XES_SPECTRUM", &
                                       extension=".spectrum", file_position=my_pos, file_action=my_act, &
                                       file_form="FORMATTED", middle_name=TRIM(mittle_em))

      IF (out_sp_em > 0) THEN
         WRITE (out_sp_em, '(A,I6,A,I6,A,I6)') " Emission spectrum for atom ", iatom, &
            ", index of excited core MO is", estate, ", # of lines ", SIZE(sp_em, 2)
         ene2 = 1.0_dp
         DO istate = estate, SIZE(sp_em, 2)
            IF (length) ene2 = sp_em(1, istate)*sp_em(1, istate)
            WRITE (out_sp_em, '(I6,5F16.8,F10.5)') istate, sp_em(1, istate)*evolt, &
               sp_em(2, istate)*ene2, sp_em(3, istate)*ene2, &
               sp_em(4, istate)*ene2, sp_em(5, istate)*ene2, sp_em(6, istate)
         END DO
      END IF
      CALL cp_print_key_finished_output(out_sp_em, logger, xas_section, &
                                        "PRINT%XES_SPECTRUM")

      mittle_ab = "xas_at"//TRIM(ADJUSTL(cp_to_string(iatom)))
      out_sp_ab = cp_print_key_unit_nr(logger, xas_section, "PRINT%XAS_SPECTRUM", &
                                       extension=".spectrum", file_position=my_pos, file_action=my_act, &
                                       file_form="FORMATTED", middle_name=TRIM(mittle_ab))

      IF (out_sp_ab > 0) THEN
         WRITE (out_sp_ab, '(A,I6,A,I6,A,I6)') " Absorption spectrum for atom ", iatom, &
            ", index of excited core MO is", estate, ", # of lines ", SIZE(sp_ab, 2)
         ene2 = 1.0_dp
         DO i = 1, SIZE(sp_ab, 2)
            istate = lfomo-1+i
            IF (length) ene2 = sp_ab(1, i)*sp_ab(1, i)
            WRITE (out_sp_ab, '(I6,5F16.8,F10.5)') istate, sp_ab(1, i)*evolt, &
               sp_ab(2, i)*ene2, sp_ab(3, i)*ene2, &
               sp_ab(4, i)*ene2, sp_ab(5, i)*ene2, sp_ab(6, i)
         END DO
      END IF

      CALL cp_print_key_finished_output(out_sp_ab, logger, xas_section, &
                                        "PRINT%XAS_SPECTRUM")

   END SUBROUTINE xas_write

! **************************************************************************************************
!> \brief write the cube files for a set of selected states
!> \param xas_control provide number ant indexes of the states to be printed
!> \param qs_env ...
!> \param xas_section ...
!> \param mos mos from which the states to be printed are extracted
!> \param all_vectors ...
!> \param iatom index of the atom that has been excited
!> \param append_cube ...
!> \par History
!>      08.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE xas_print_cubes(xas_control, qs_env, xas_section, &
                              mos, all_vectors, iatom, append_cube)

      TYPE(xas_control_type)                             :: xas_control
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: xas_section
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_fm_type), POINTER                          :: all_vectors
      INTEGER, INTENT(IN)                                :: iatom
      LOGICAL, INTENT(IN)                                :: append_cube

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_print_cubes', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_mittle, my_pos
      INTEGER                                            :: homo, istate0, nspins, nstates
      REAL(dp), DIMENSION(:, :), POINTER                 :: centers
      TYPE(section_vals_type), POINTER                   :: print_key

      nspins = SIZE(mos)

      print_key => section_vals_get_subs_vals(xas_section, "PRINT%CLS_FUNCTION_CUBES")
      my_mittle = 'at'//TRIM(ADJUSTL(cp_to_string(iatom)))
      nstates = SIZE(xas_control%list_cubes, 1)

      IF (xas_control%do_centers) THEN
         ! one might like to calculate the centers of the xas orbital (without localizing them)
      ELSE
         ALLOCATE (centers(6, nstates))
         centers = 0.0_dp
      END IF

      CALL get_mo_set(mos(1)%mo_set, homo=homo)
      istate0 = 0

      my_pos = "REWIND"
      IF (append_cube) THEN
         my_pos = "APPEND"
      END IF

      CALL qs_print_cubes(qs_env, all_vectors, nstates, xas_control%list_cubes, &
                          centers, print_key, my_mittle, state0=istate0, file_position=my_pos)

      DEALLOCATE (centers)

   END SUBROUTINE xas_print_cubes

! **************************************************************************************************
!> \brief write the PDOS after the XAS SCF, i.e., with one excited core
!> \param qs_env ...
!> \param xas_section ...
!> \param mos mos from which the eigenvalues and expansion coeffiecients are obtained
!> \param iatom index of the atom that has been excited
!> \par History
!>      03.2016 created [MI]
!> \author MI
! **************************************************************************************************

   SUBROUTINE xas_pdos(qs_env, xas_section, mos, iatom)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: xas_section
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      INTEGER, INTENT(IN)                                :: iatom

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_pdos', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: xas_mittle
      INTEGER                                            :: ispin
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      NULLIFY (atomic_kind_set, particle_set, qs_kind_set)
      xas_mittle = 'xasat'//TRIM(ADJUSTL(cp_to_string(iatom)))//'_'

      CALL get_qs_env(qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      particle_set=particle_set, &
                      qs_kind_set=qs_kind_set)

      DO ispin = 1, 2
         CALL calculate_projected_dos(mos(ispin)%mo_set, atomic_kind_set, qs_kind_set, particle_set, qs_env, &
                                      xas_section, ispin, xas_mittle)
      END DO

   END SUBROUTINE xas_pdos
! **************************************************************************************************
!> \brief Calculation of the spectrum when the dipole approximation
!>      in the velocity form is used.
!> \param fm_set components of the position operator in a full matrix form
!>                already multiplied by the coefficiets
!>                only the terms <C_i Op C_f> are calculated where
!>                C_i are the coefficients of the excited state
!> \param op_sm components of the position operator for the dipole
!>               in a sparse matrix form (cos and sin)
!>               calculated for the basis functions
!> \param mos wavefunctions coefficients
!> \param excvec coefficients of the excited orbital
!> \param all_vectors ...
!> \param all_evals ...
!> \param fm_work work space
!> \param sp_em ...
!> \param sp_ab ...
!> \param estate index of the excited state
!> \param nstate ...
!> \par History
!>      06.2005 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE spectrum_dip_vel(fm_set, op_sm, mos, excvec, &
                               all_vectors, all_evals, fm_work, sp_em, sp_ab, estate, nstate)

      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: fm_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: op_sm
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_fm_type), POINTER                          :: excvec, all_vectors
      REAL(dp), DIMENSION(:), POINTER                    :: all_evals
      TYPE(cp_fm_type), POINTER                          :: fm_work
      REAL(dp), DIMENSION(:, :), POINTER                 :: sp_em, sp_ab
      INTEGER, INTENT(IN)                                :: estate, nstate

      CHARACTER(LEN=*), PARAMETER :: routineN = 'spectrum_dip_vel', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: homo, i, i_abs, istate, lfomo, nao, nmo
      REAL(dp)                                           :: dip(3), ene_f, ene_i
      REAL(dp), DIMENSION(:), POINTER                    :: eigenvalues, occupation_numbers

      IF(.NOT.(ASSOCIATED(fm_set)))CALL cp__a("xas_methods.F",1330)
      IF(.NOT.(ASSOCIATED(mos)))CALL cp__a("xas_methods.F",1331)
      NULLIFY (eigenvalues, occupation_numbers)

      CALL get_mo_set(mos(1)%mo_set, eigenvalues=eigenvalues, occupation_numbers=occupation_numbers, &
                      nao=nao, nmo=nmo, homo=homo, lfomo=lfomo)

      DO i = 1, SIZE(fm_set, 2)
         IF(.NOT.(ASSOCIATED(fm_set(1, i)%matrix)))CALL cp__a("xas_methods.F",1338)
         CALL cp_fm_set_all(fm_set(1, i)%matrix, 0.0_dp)
         CALL cp_fm_set_all(fm_work, 0.0_dp)
         CALL cp_dbcsr_sm_fm_multiply(op_sm(i)%matrix, all_vectors, fm_work, ncol=nstate)
         CALL cp_gemm("T", "N", 1, nstate, nao, 1.0_dp, excvec, &
                      fm_work, 0.0_dp, fm_set(1, i)%matrix, b_first_col=1)
      END DO

      sp_em = 0.0_dp
      sp_ab = 0.0_dp
      ene_i = eigenvalues(estate)

      DO istate = 1, nstate
         ene_f = all_evals(istate)

         DO i = 1, 3
            CALL cp_fm_get_element(fm_set(1, i)%matrix, 1, istate, dip(i))
         END DO
         IF (istate <= homo) THEN
            sp_em(1, istate) = ene_f-ene_i
            sp_em(2, istate) = dip(1)
            sp_em(3, istate) = dip(2)
            sp_em(4, istate) = dip(3)
            sp_em(5, istate) = dip(1)*dip(1)+dip(2)*dip(2)+dip(3)*dip(3)
            sp_em(6, istate) = occupation_numbers(istate)
         END IF
         IF (istate >= lfomo) THEN
            i_abs = istate-lfomo+1
            sp_ab(1, i_abs) = ene_f-ene_i
            sp_ab(2, i_abs) = dip(1)
            sp_ab(3, i_abs) = dip(2)
            sp_ab(4, i_abs) = dip(3)
            sp_ab(5, i_abs) = dip(1)*dip(1)+dip(2)*dip(2)+dip(3)*dip(3)
            IF (istate <= nmo) sp_ab(6, i_abs) = occupation_numbers(istate)
         END IF

      END DO

   END SUBROUTINE spectrum_dip_vel

! **************************************************************************************************
!> \brief ...
!> \param base_a ...
!> \param base_b ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE calc_stogto_overlap(base_a, base_b, matrix)

      TYPE(gto_basis_set_type), POINTER                  :: base_a, base_b
      REAL(dp), DIMENSION(:, :), POINTER                 :: matrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_stogto_overlap', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iset, jset, ldsab, maxcoa, maxcob, maxl, &
                                                            maxla, maxlb, na, nb, nseta, nsetb, &
                                                            nsgfa, nsgfb, sgfa, sgfb
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa_set, nsgfb_set
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      REAL(dp)                                           :: rab(3)
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: sab, work
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rpgfa, rpgfb, scon_a, scon_b, sphi_a, &
                                                            sphi_b, zeta, zetb

      NULLIFY (la_max, la_min, lb_max, lb_min)
      NULLIFY (npgfa, npgfb, nsgfa_set, nsgfb_set)
      NULLIFY (first_sgfa, first_sgfb)
      NULLIFY (rpgfa, rpgfb, sphi_a, sphi_b, zeta, zetb)

      CALL get_gto_basis_set(gto_basis_set=base_a, nsgf=nsgfa, nsgf_set=nsgfa_set, lmax=la_max, &
                             lmin=la_min, npgf=npgfa, pgf_radius=rpgfa, &
                             sphi=sphi_a, scon=scon_a, zet=zeta, first_sgf=first_sgfa, &
                             maxco=maxcoa, nset=nseta, maxl=maxla)

      CALL get_gto_basis_set(gto_basis_set=base_b, nsgf=nsgfb, nsgf_set=nsgfb_set, lmax=lb_max, &
                             lmin=lb_min, npgf=npgfb, pgf_radius=rpgfb, &
                             sphi=sphi_b, scon=scon_b, zet=zetb, first_sgf=first_sgfb, &
                             maxco=maxcob, nset=nsetb, maxl=maxlb)
      ! Initialize and allocate
      rab = 0.0_dp
      matrix = 0.0_dp

      ldsab = MAX(maxcoa, maxcob, nsgfa, nsgfb)
      maxl = MAX(maxla, maxlb)

      ALLOCATE (sab(ldsab, ldsab))
      ALLOCATE (work(ldsab, ldsab))

      DO iset = 1, nseta

         na = npgfa(iset)*(ncoset(la_max(iset))-ncoset(la_min(iset)-1))
         sgfa = first_sgfa(1, iset)

         DO jset = 1, nsetb

            nb = npgfb(jset)*(ncoset(lb_max(jset))-ncoset(lb_min(iset)-1))
            sgfb = first_sgfb(1, jset)

            CALL overlap_ab(la_max(iset), la_min(iset), npgfa(iset), rpgfa(:, iset), zeta(:, iset), &
                            lb_max(jset), lb_min(jset), npgfb(jset), rpgfb(:, jset), zetb(:, jset), &
                            rab, sab)
            CALL contraction(sab, work, ca=scon_a(:, sgfa:), na=na, ma=nsgfa_set(iset), &
                             cb=scon_b(:, sgfb:), nb=nb, mb=nsgfb_set(jset))
            CALL block_add("IN", work, nsgfa_set(iset), nsgfb_set(jset), matrix, sgfa, sgfb)

         END DO ! jset
      END DO ! iset

      DEALLOCATE (sab, work)

   END SUBROUTINE calc_stogto_overlap

! **************************************************************************************************
!> \brief Starting from a set of mos, determine on which atom are centered
!>      and if they are of the right type (1s,2s ...)
!>      to be used in the specific core level spectrum calculation
!>      The set of states need to be from the core, otherwise the
!>      characterization of the type is not valid, since it assumes that
!>      the orbital is localizad on a specific atom
!>      It is probably reccomandable to run a localization cycle before
!>      proceeding to the assignment of the type
!>      The type is determined by computing the overalp with a
!>      type specific, minimal, STO bais set
!> \param xas_control ...
!> \param xas_env ...
!> \param localized_wfn_control ...
!> \param qs_env ...
!> \par History
!>      03.2006 created [MI]
!> \author MI
! **************************************************************************************************
   SUBROUTINE cls_assign_core_states(xas_control, xas_env, localized_wfn_control, qs_env)

      TYPE(xas_control_type)                             :: xas_control
      TYPE(xas_environment_type), POINTER                :: xas_env
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cls_assign_core_states', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: homo, i, iat, iatom, ikind, isgf, &
                                                            istate, j, my_kind, nao, natom, &
                                                            nexc_atoms, nexc_search, output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: first_sgf
      INTEGER, DIMENSION(3)                              :: perd0
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_state, mykind_of_kind, &
                                                            state_of_atom, state_of_mytype, &
                                                            type_of_state
      REAL(dp)                                           :: component, dist, distmin, maxocc, ra(3), &
                                                            rac(3), rc(3)
      REAL(dp), DIMENSION(:), POINTER                    :: max_overlap, sto_state_overlap
      REAL(dp), DIMENSION(:, :), POINTER                 :: centers_wfn
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vecbuffer
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_2d_r_p_type), DIMENSION(:), POINTER        :: stogto_overlap
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      NULLIFY (cell, mos, particle_set)
      NULLIFY (atom_of_state, centers_wfn, mykind_of_kind, state_of_atom)
      NULLIFY (stogto_overlap, type_of_state, max_overlap, qs_kind_set)
      NULLIFY (state_of_mytype, type_of_state, sto_state_overlap)

      NULLIFY (logger)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      CALL get_qs_env(qs_env=qs_env, cell=cell, mos=mos, particle_set=particle_set, &
                      qs_kind_set=qs_kind_set)

      ! The Berry operator can be used only for periodic systems
      ! If an isolated system is used the periodicity is overimposed
      perd0(1:3) = cell%perd(1:3)
      cell%perd(1:3) = 1

      CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff, maxocc=maxocc, nao=nao, homo=homo)

      CALL get_xas_env(xas_env=xas_env, &
                       centers_wfn=centers_wfn, atom_of_state=atom_of_state, &
                       mykind_of_kind=mykind_of_kind, &
                       type_of_state=type_of_state, state_of_atom=state_of_atom, &
                       stogto_overlap=stogto_overlap, nexc_atoms=nexc_atoms, nexc_search=nexc_search)

      ! scratch array for the state
      ALLOCATE (vecbuffer(1, nao))
      natom = SIZE(particle_set)

      ALLOCATE (first_sgf(natom))
      CALL get_particle_set(particle_set, qs_kind_set, first_sgf=first_sgf)
      ALLOCATE (sto_state_overlap(nexc_search))
      ALLOCATE (max_overlap(natom))
      max_overlap = 0.0_dp
      ALLOCATE (state_of_mytype(natom))
      state_of_mytype = 0
      atom_of_state = 0

      DO istate = 1, nexc_search
         centers_wfn(1, istate) = localized_wfn_control%centers_set(1)%array(1, istate)
         centers_wfn(2, istate) = localized_wfn_control%centers_set(1)%array(2, istate)
         centers_wfn(3, istate) = localized_wfn_control%centers_set(1)%array(3, istate)

         ! Assign the state to the closest atom
         distmin = 100.0_dp

         DO iat = 1, nexc_atoms
            iatom = xas_control%exc_atoms(iat)
            ra(1:3) = particle_set(iatom)%r(1:3)
            rc(1:3) = centers_wfn(1:3, istate)
            rac = pbc(ra, rc, cell)
            dist = rac(1)*rac(1)+rac(2)*rac(2)+rac(3)*rac(3)

            IF (dist < distmin) THEN
               atom_of_state(istate) = iatom
               distmin = dist
            END IF
         END DO
         IF (atom_of_state(istate) /= 0) THEN
            !Character of the state
            CALL cp_fm_get_submatrix(mo_coeff, vecbuffer, 1, istate, &
                                     nao, 1, transpose=.TRUE.)

            iatom = atom_of_state(istate)

            NULLIFY (atomic_kind)
            atomic_kind => particle_set(iatom)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 kind_number=ikind)

            my_kind = mykind_of_kind(ikind)

            sto_state_overlap(istate) = 0.0_dp
            DO i = 1, SIZE(stogto_overlap(my_kind)%array, 1)
               component = 0.0_dp
               DO j = 1, SIZE(stogto_overlap(my_kind)%array, 2)
                  isgf = first_sgf(iatom)+j-1
                  component = component+stogto_overlap(my_kind)%array(i, j)*vecbuffer(1, isgf)
               END DO
               sto_state_overlap(istate) = sto_state_overlap(istate)+ &
                                           component*component
            END DO
            IF (sto_state_overlap(istate) > max_overlap(iatom)) THEN
               state_of_mytype(iatom) = istate
               max_overlap(iatom) = sto_state_overlap(istate)
            END IF
         END IF
      END DO ! istate

      ! In the set of states, assign the index of the state to be excited for iatom
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T10,A,/)") &
            "List the atoms to be excited and the relative of MOs index "
      END IF
      DO iat = 1, nexc_atoms
         iatom = xas_env%exc_atoms(iat)
         state_of_atom(iat) = state_of_mytype(iatom)
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T10,A,I6,T32,A,I6)") 'Atom: ', iatom, &
               "MO index", state_of_atom(iat)
         END IF
         IF (state_of_atom(iat) == 0 .OR. state_of_atom(iat) .GT. homo) THEN
            CALL cp__b("xas_methods.F",1604,"A wrong state has been selected for excitation, check the Wannier centers")
         END IF
      END DO

      ! Set back the correct periodicity
      cell%perd(1:3) = perd0(1:3)

      DEALLOCATE (vecbuffer)
      DEALLOCATE (first_sgf)
      DEALLOCATE (sto_state_overlap)
      DEALLOCATE (max_overlap)
      DEALLOCATE (state_of_mytype)

   END SUBROUTINE cls_assign_core_states

END MODULE xas_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of dispersion in DFTB
!> \author JGH
! **************************************************************************************************
MODULE qs_dftb_dispersion
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_array_init,&
                                              atprop_type
   USE cp_control_types,                ONLY: dft_control_type,&
                                              dftb_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: dispersion_d3,&
                                              dispersion_uff
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type,&
                                              qs_dftb_pairpot_type
   USE qs_dftb_utils,                   ONLY: get_dftb_atom_param
   USE qs_dispersion_pairpot,           ONLY: calculate_dispersion_pairpot
   USE qs_dispersion_types,             ONLY: qs_dispersion_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dftb_dispersion'

   PUBLIC :: calculate_dftb_dispersion

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param para_env ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE calculate_dftb_dispersion(qs_env, para_env, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_dftb_dispersion', &
         routineP = moduleN//':'//routineN

      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(dftb_control_type), POINTER                   :: dftb_control
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      TYPE(qs_energy_type), POINTER                      :: energy

      CALL get_qs_env(qs_env=qs_env, &
                      energy=energy, &
                      dft_control=dft_control)

      energy%dispersion = 0._dp

      dftb_control => dft_control%qs_control%dftb_control
      IF (dftb_control%dispersion) THEN
         SELECT CASE (dftb_control%dispersion_type)
         CASE (dispersion_uff)
            CALL calculate_dispersion_uff(qs_env, para_env, calculate_forces)
         CASE (dispersion_d3)
            CALL get_qs_env(qs_env=qs_env, dispersion_env=dispersion_env)
            CALL calculate_dispersion_pairpot(qs_env, dispersion_env, &
                                              energy%dispersion, calculate_forces)
         CASE DEFAULT
            CALL cp__b("qs_dftb_dispersion.F",90,"")
         END SELECT
      END IF

   END SUBROUTINE calculate_dftb_dispersion

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param para_env ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE calculate_dispersion_uff(qs_env, para_env, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_dispersion_uff', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, handle, iatom, ikind, &
                                                            jatom, jkind, natom, nkind
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: use_virial
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: define_kind
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: rc_kind
      REAL(KIND=dp)                                      :: a, b, c, devdw, dij, dr, eij, evdw, fac, &
                                                            rc, x0ij, xij, xp
      REAL(KIND=dp), DIMENSION(3)                        :: fdij, rij
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(dftb_control_type), POINTER                   :: dftb_control
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_vdw
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_kind_a
      TYPE(qs_dftb_pairpot_type), DIMENSION(:, :), &
         POINTER                                         :: dftb_potential
      TYPE(qs_dftb_pairpot_type), POINTER                :: dftb_param_ij
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, sab_vdw, atprop)

      CALL get_qs_env(qs_env=qs_env, &
                      energy=energy, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      virial=virial, atprop=atprop, &
                      dft_control=dft_control)

      energy%dispersion = 0._dp

      dftb_control => dft_control%qs_control%dftb_control

      IF (dftb_control%dispersion) THEN

         NULLIFY (dftb_potential)
         CALL get_qs_env(qs_env=qs_env, dftb_potential=dftb_potential)
         IF (calculate_forces) THEN
            NULLIFY (force, particle_set)
            CALL get_qs_env(qs_env=qs_env, &
                            particle_set=particle_set, &
                            force=force)
            natom = SIZE(particle_set)
            ALLOCATE (atom_of_kind(natom))
            CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)
            use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
         ELSE
            use_virial = .FALSE.
         END IF
         nkind = SIZE(atomic_kind_set)
         ALLOCATE (define_kind(nkind), rc_kind(nkind))
         DO ikind = 1, nkind
            CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_kind_a)
            CALL get_dftb_atom_param(dftb_kind_a, defined=define_kind(ikind), rcdisp=rc_kind(ikind))
         END DO

         evdw = 0._dp

         IF (atprop%energy) THEN
            CALL get_qs_env(qs_env=qs_env, particle_set=particle_set)
            natom = SIZE(particle_set)
            CALL atprop_array_init(atprop%atevdw, natom)
         END IF

         CALL get_qs_env(qs_env=qs_env, sab_vdw=sab_vdw)
         CALL neighbor_list_iterator_create(nl_iterator, sab_vdw)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, r=rij)
            IF ((.NOT. define_kind(ikind)) .OR. (.NOT. define_kind(jkind))) CYCLE
            rc = rc_kind(ikind)+rc_kind(jkind)
            ! vdW potential
            dr = SQRT(SUM(rij(:)**2))
            IF (dr <= rc .AND. dr > 0.001_dp) THEN
               fac = 1._dp
               IF (iatom == jatom) fac = 0.5_dp
               ! retrieve information on potential
               dftb_param_ij => dftb_potential(ikind, jkind)
               ! vdW parameter
               xij = dftb_param_ij%xij
               dij = dftb_param_ij%dij
               x0ij = dftb_param_ij%x0ij
               a = dftb_param_ij%a
               b = dftb_param_ij%b
               c = dftb_param_ij%c
               IF (dr > x0ij) THEN
                  ! This is the standard London contribution.
                  ! UFF1 - Eq. 20 (long-range)
                  xp = xij/dr
                  eij = dij*(-2._dp*xp**6+xp**12)*fac
                  evdw = evdw+eij
                  IF (calculate_forces .AND. (dr > 0.001_dp)) THEN
                     devdw = dij*12._dp*(xp**6-xp**12)/dr*fac
                     atom_a = atom_of_kind(iatom)
                     atom_b = atom_of_kind(jatom)
                     fdij(:) = devdw*rij(:)/dr
                     force(ikind)%dispersion(:, atom_a) = &
                        force(ikind)%dispersion(:, atom_a)-fdij(:)
                     force(jkind)%dispersion(:, atom_b) = &
                        force(jkind)%dispersion(:, atom_b)+fdij(:)
                  END IF
               ELSE
                  ! Shorter distance.
                  ! London contribution should converge to a finite value.
                  ! Using a parabola of the form (y = A - Bx**5 -Cx**10).
                  ! Analytic parameters by forcing energy, first and second
                  ! derivatives to be continuous.
                  eij = (A-B*dr**5-C*dr**10)*fac
                  evdw = evdw+eij
                  IF (calculate_forces .AND. (dr > 0.001_dp)) THEN
                     atom_a = atom_of_kind(iatom)
                     atom_b = atom_of_kind(jatom)
                     devdw = (-5*B*dr**4-10*C*dr**9)*fac
                     fdij(:) = devdw*rij(:)/dr
                     force(ikind)%dispersion(:, atom_a) = &
                        force(ikind)%dispersion(:, atom_a)-fdij(:)
                     force(jkind)%dispersion(:, atom_b) = &
                        force(jkind)%dispersion(:, atom_b)+fdij(:)
                  END IF
               END IF
               IF (atprop%energy) THEN
                  atprop%atevdw(iatom) = atprop%atevdw(iatom)+0.5_dp*eij
                  atprop%atevdw(jatom) = atprop%atevdw(jatom)+0.5_dp*eij
               END IF
               IF (calculate_forces .AND. (dr > 0.001_dp) .AND. use_virial) THEN
                  CALL virial_pair_force(virial%pv_virial, -1._dp, fdij, rij)
                  IF (atprop%stress) THEN
                     CALL virial_pair_force(atprop%atstress(:, :, iatom), -0.5_dp, fdij, rij)
                     CALL virial_pair_force(atprop%atstress(:, :, jatom), -0.5_dp, fdij, rij)
                  END IF
               END IF
            END IF
         END DO
         CALL neighbor_list_iterator_release(nl_iterator)

         IF (calculate_forces) THEN
            DEALLOCATE (atom_of_kind)
         END IF

         ! set dispersion energy
         CALL mp_sum(evdw, para_env%group)
         energy%dispersion = evdw

      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_dispersion_uff

END MODULE qs_dftb_dispersion

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to handle an external electrostatic field
!>        The external field can be generic and is provided by user input
! **************************************************************************************************
MODULE qs_external_potential
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE force_fields_util,               ONLY: get_generic_info
   USE fparser,                         ONLY: evalf,&
                                              evalfd,&
                                              finalizef,&
                                              initf,&
                                              parsef
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_bcast
   USE particle_types,                  ONLY: particle_type
   USE pw_grid_types,                   ONLY: PW_MODE_LOCAL
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE realspace_grid_cube,             ONLY: cube_to_pw
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_external_potential'

! *** Public subroutines ***
   PUBLIC :: external_e_potential, &
             external_c_potential

CONTAINS

! **************************************************************************************************
!> \brief  Computes the external potential on the grid
!> \param qs_env ...
!> \date   12.2009
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE external_e_potential(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'external_e_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo_global, bo_local
      LOGICAL                                            :: read_from_cube, static_potential
      REAL(kind=dp)                                      :: dvol, efunc, scaling_factor
      REAL(kind=dp), DIMENSION(3)                        :: dr, grid_p
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type), POINTER                           :: v_ee
      TYPE(section_vals_type), POINTER                   :: ext_pot_section, input

      CALL timeset(routineN, handle)
      NULLIFY (v_ee, input, ext_pot_section, dft_control)
      CALL get_qs_env(qs_env, &
                      vee=v_ee, &
                      input=input, &
                      dft_control=dft_control)
      IF (dft_control%apply_external_potential) THEN
         ext_pot_section => section_vals_get_subs_vals(input, "DFT%EXTERNAL_POTENTIAL")
         CALL section_vals_val_get(ext_pot_section, "STATIC", l_val=static_potential)
         CALL section_vals_val_get(ext_pot_section, "READ_FROM_CUBE", l_val=read_from_cube)

         IF ((.NOT. static_potential) .OR. dft_control%eval_external_potential) THEN
            IF (read_from_cube) THEN
               CALL section_vals_val_get(ext_pot_section, "SCALING_FACTOR", r_val=scaling_factor)
               CALL cube_to_pw(v_ee%pw, 'pot.cube', scaling_factor)
               dft_control%eval_external_potential = .FALSE.
            ELSE
               dr = v_ee%pw%pw_grid%dr
               dvol = v_ee%pw%pw_grid%dvol
               v_ee%pw%cr3d = 0.0_dp

               bo_local = v_ee%pw%pw_grid%bounds_local
               bo_global = v_ee%pw%pw_grid%bounds

               DO k = bo_local(1, 3), bo_local(2, 3)
                  DO j = bo_local(1, 2), bo_local(2, 2)
                     DO i = bo_local(1, 1), bo_local(2, 1)
                        grid_p(1) = (i-bo_global(1, 1))*dr(1)
                        grid_p(2) = (j-bo_global(1, 2))*dr(2)
                        grid_p(3) = (k-bo_global(1, 3))*dr(3)
                        CALL get_external_potential(grid_p, ext_pot_section, func=efunc)
                        v_ee%pw%cr3d(i, j, k) = v_ee%pw%cr3d(i, j, k)+efunc
                     END DO
                  END DO
               END DO
               dft_control%eval_external_potential = .FALSE.
            END IF
         END IF
      END IF
      CALL timestop(handle)
   END SUBROUTINE external_e_potential

! **************************************************************************************************
!> \brief  Computes the force and the energy due to the external potential on the cores
!> \param qs_env ...
!> \param calculate_forces ...
!> \date   12.2009
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE external_c_potential(qs_env, calculate_forces)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, OPTIONAL                                  :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'external_c_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ikind, natom, &
                                                            nkind
      INTEGER, DIMENSION(:), POINTER                     :: list
      LOGICAL                                            :: my_force, pot_on_grid
      REAL(KIND=dp)                                      :: ee_core_ener, efunc, zeff
      REAL(KIND=dp), DIMENSION(3)                        :: dfunc, r
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_p_type), POINTER                           :: v_ee
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: ext_pot_section, input

      CALL timeset(routineN, handle)
      NULLIFY (dft_control)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      energy=energy, &
                      particle_set=particle_set, &
                      input=input, &
                      cell=cell, &
                      dft_control=dft_control)

      IF (dft_control%apply_external_potential) THEN
         IF (dft_control%eval_external_potential) THEN !ensure that external potential is loaded to grid
            CALL external_e_potential(qs_env)
         END IF
         my_force = .FALSE.
         IF (PRESENT(calculate_forces)) my_force = calculate_forces
         ext_pot_section => section_vals_get_subs_vals(input, "DFT%EXTERNAL_POTENTIAL")
         ee_core_ener = 0.0_dp
         nkind = SIZE(atomic_kind_set)

         !check if external potential on grid has been loaded from a file instead of giving a function
         CALL section_vals_val_get(ext_pot_section, "READ_FROM_CUBE", l_val=pot_on_grid)
         IF (pot_on_grid) CALL get_qs_env(qs_env, vee=v_ee, input=input)

         DO ikind = 1, SIZE(atomic_kind_set)
            CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=list, natom=natom)
            CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)

            natom = SIZE(list)
            DO iatom = 1, natom
               atom_a = list(iatom)
               !pbc returns r(i) in range [-cell%hmat(i,i)/2, cell%hmat(i,i)/2]
               !for periodic dimensions (assuming the cell is orthorombic).
               !This is not consistent with the potential on grid, where r(i) is
               !in range [0, cell%hmat(i,i)]
               !Use new pbc function with switch positive_range=.TRUE.
               r(:) = pbc(particle_set(atom_a)%r(:), cell, positive_range=.TRUE.)
               !if potential is on grid, interpolate the value at r,
               !otherwise evaluate the given function
               IF (pot_on_grid) THEN
                  CALL interpolate_external_potential(r, v_ee%pw, func=efunc, &
                                                      dfunc=dfunc, calc_derivatives=my_force)
               ELSE
                  CALL get_external_potential(r, ext_pot_section, func=efunc, &
                                              dfunc=dfunc, calc_derivatives=my_force)
               END IF
               ee_core_ener = ee_core_ener+zeff*efunc
               IF (my_force) THEN
                  CALL get_qs_env(qs_env=qs_env, force=force)
                  force(ikind)%eev(:, iatom) = dfunc*zeff
               END IF
            END DO
         END DO
         energy%ee_core = ee_core_ener
      END IF
      CALL timestop(handle)
   END SUBROUTINE external_c_potential

! **************************************************************************************************
!> \brief  Low level function for computing the potential and the derivatives
!> \param r                position in realspace
!> \param ext_pot_section ...
!> \param func             external potential at r
!> \param dfunc            derivative of the external potential at r
!> \param calc_derivatives Whether to calulate dfunc
!> \date   12.2009
!> \par History
!>      12.2009            created [tlaino]
!>      11.2014            reading external cube file added [Juha Ritala & Matt Watkins]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE get_external_potential(r, ext_pot_section, func, dfunc, calc_derivatives)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: r
      TYPE(section_vals_type), POINTER                   :: ext_pot_section
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: func, dfunc(3)
      LOGICAL, INTENT(IN), OPTIONAL                      :: calc_derivatives

      CHARACTER(len=*), PARAMETER :: routineN = 'get_external_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: coupling_function
      CHARACTER(LEN=default_string_length)               :: def_error, this_error
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: my_par
      INTEGER                                            :: handle, j
      LOGICAL                                            :: check, my_force
      REAL(KIND=dp)                                      :: dedf, dx, err, lerr
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_val

      CALL timeset(routineN, handle)
      NULLIFY (my_par, my_val)
      my_force = .FALSE.
      IF (PRESENT(calc_derivatives)) my_force = calc_derivatives
      check = PRESENT(dfunc) .EQV. PRESENT(calc_derivatives)
      IF(.NOT.(check))CALL cp__a("qs_external_potential.F",247)
      CALL section_vals_val_get(ext_pot_section, "DX", r_val=dx)
      CALL section_vals_val_get(ext_pot_section, "ERROR_LIMIT", r_val=lerr)
      CALL get_generic_info(ext_pot_section, "FUNCTION", coupling_function, my_par, my_val, &
                            input_variables=(/"X", "Y", "Z"/), i_rep_sec=1)
      CALL initf(1)
      CALL parsef(1, TRIM(coupling_function), my_par)

      my_val(1) = r(1)
      my_val(2) = r(2)
      my_val(3) = r(3)

      IF (PRESENT(func)) func = evalf(1, my_val)
      IF (my_force) THEN
         DO j = 1, 3
            dedf = evalfd(1, j, my_val, dx, err)
            IF (ABS(err) > lerr) THEN
               WRITE (this_error, "(A,G12.6,A)") "(", err, ")"
               WRITE (def_error, "(A,G12.6,A)") "(", lerr, ")"
               CALL compress(this_error, .TRUE.)
               CALL compress(def_error, .TRUE.)
               CALL cp_warn(cp__l("qs_external_potential.F",268), &
                            'ASSERTION (cond) failed at line '//cp_to_string(269)// &
                            ' Error '//TRIM(this_error)//' in computing numerical derivatives larger then'// &
                            TRIM(def_error)//' .')
            END IF
            dfunc(j) = dedf
         END DO
      END IF
      DEALLOCATE (my_par)
      DEALLOCATE (my_val)
      CALL finalizef()
      CALL timestop(handle)
   END SUBROUTINE get_external_potential

! **************************************************************************************************
!> \brief                  subroutine that interpolates the value of the external
!>                         potential at position r based on the values on the realspace grid
!> \param r                 ...
!> \param grid             external potential pw grid, vee
!> \param func             value of vee at r
!> \param dfunc            derivatives of vee at r
!> \param calc_derivatives calc dfunc
! **************************************************************************************************
   SUBROUTINE interpolate_external_potential(r, grid, func, dfunc, calc_derivatives)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: r
      TYPE(pw_type), POINTER                             :: grid
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: func, dfunc(3)
      LOGICAL, INTENT(IN), OPTIONAL                      :: calc_derivatives

      CHARACTER(len=*), PARAMETER :: routineN = 'interpolate_external_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: buffer_i, buffer_j, buffer_k, &
                                                            data_source, fd_extra_point, gid, &
                                                            handle, i, i_pbc, ip, j, j_pbc, k, &
                                                            k_pbc, my_rank, num_pe, tag
      INTEGER, DIMENSION(3)                              :: lbounds, lbounds_local, lower_inds, &
                                                            ubounds, ubounds_local, upper_inds
      LOGICAL                                            :: check, my_force
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: bcast_buffer
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: grid_buffer
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: dgrid
      REAL(KIND=dp), DIMENSION(3)                        :: dr, subgrid_origin

      CALL timeset(routineN, handle)
      my_force = .FALSE.
      IF (PRESENT(calc_derivatives)) my_force = calc_derivatives
      check = PRESENT(dfunc) .EQV. PRESENT(calc_derivatives)
      IF(.NOT.(check))CALL cp__a("qs_external_potential.F",316)

      IF (my_force) THEN
         ALLOCATE (grid_buffer(0:3, 0:3, 0:3))
         ALLOCATE (bcast_buffer(0:3))
         ALLOCATE (dgrid(1:2, 1:2, 1:2, 3))
         fd_extra_point = 1
      ELSE
         ALLOCATE (grid_buffer(1:2, 1:2, 1:2))
         ALLOCATE (bcast_buffer(1:2))
         fd_extra_point = 0
      END IF

      ! The values of external potential on grid are distributed among the
      ! processes, so first we have to gather them up
      gid = grid%pw_grid%para%group
      my_rank = grid%pw_grid%para%my_pos
      num_pe = grid%pw_grid%para%group_size
      tag = 1

      dr = grid%pw_grid%dr
      lbounds = grid%pw_grid%bounds(1, :)
      ubounds = grid%pw_grid%bounds(2, :)
      lbounds_local = grid%pw_grid%bounds_local(1, :)
      ubounds_local = grid%pw_grid%bounds_local(2, :)

      ! Determine the indices of grid points that are needed
      lower_inds = lbounds+FLOOR(r/dr)-fd_extra_point
      upper_inds = lower_inds+1+2*fd_extra_point

      DO i = lower_inds(1), upper_inds(1)
         ! If index is out of global bounds, assume periodic boundary conditions
         i_pbc = pbc_index(i, lbounds(1), ubounds(1))
         buffer_i = i-lower_inds(1)+1-fd_extra_point
         DO j = lower_inds(2), upper_inds(2)
            j_pbc = pbc_index(j, lbounds(2), ubounds(2))
            buffer_j = j-lower_inds(2)+1-fd_extra_point

            ! Find the process that has the data for indices i_pbc and j_pbc
            ! and store the data to bcast_buffer. Assuming that each process has full z data
            IF (grid%pw_grid%para%mode .NE. PW_MODE_LOCAL) THEN
               DO ip = 0, num_pe-1
                  IF (grid%pw_grid%para%bo(1, 1, ip, 1) <= i_pbc-lbounds(1)+1 .AND. &
                      grid%pw_grid%para%bo(2, 1, ip, 1) >= i_pbc-lbounds(1)+1 .AND. &
                      grid%pw_grid%para%bo(1, 2, ip, 1) <= j_pbc-lbounds(2)+1 .AND. &
                      grid%pw_grid%para%bo(2, 2, ip, 1) >= j_pbc-lbounds(2)+1) THEN
                     data_source = ip
                     EXIT
                  END IF
               END DO
               IF (my_rank == data_source) THEN
                  IF (lower_inds(3) >= lbounds(3) .AND. upper_inds(3) <= ubounds(3)) THEN
                     bcast_buffer(:) = &
                        grid%cr3d(i_pbc, j_pbc, lower_inds(3):upper_inds(3))
                  ELSE
                     DO k = lower_inds(3), upper_inds(3)
                        k_pbc = pbc_index(k, lbounds(3), ubounds(3))
                        buffer_k = k-lower_inds(3)+1-fd_extra_point
                        bcast_buffer(buffer_k) = &
                           grid%cr3d(i_pbc, j_pbc, k_pbc)
                     END DO
                  END IF
               END IF
               ! data_source sends data to everyone
               CALL mp_bcast(bcast_buffer, data_source, gid)
               grid_buffer(buffer_i, buffer_j, :) = bcast_buffer
            ELSE
               grid_buffer(buffer_i, buffer_j, :) = grid%cr3d(i_pbc, j_pbc, lower_inds(3):upper_inds(3))
            END IF
         END DO
      END DO

      ! Now that all the processes have local external potential data around r,
      ! interpolate the value at r
      subgrid_origin = (lower_inds-lbounds+fd_extra_point)*dr
      func = trilinear_interpolation(r, grid_buffer(1:2, 1:2, 1:2), subgrid_origin, dr)

      ! If the derivative of the potential is needed, approximate the derivative at grid
      ! points using finite differences, and then interpolate the value at r
      IF (my_force) THEN
         CALL d_finite_difference(grid_buffer, dr, dgrid)
         DO i = 1, 3
            dfunc(i) = trilinear_interpolation(r, dgrid(:, :, :, i), subgrid_origin, dr)
         END DO
         DEALLOCATE (dgrid)
      END IF

      DEALLOCATE (grid_buffer)
      CALL timestop(handle)
   END SUBROUTINE interpolate_external_potential

! **************************************************************************************************
!> \brief       subroutine that uses finite differences to approximate the partial
!>              derivatives of the potential based on the given values on grid
!> \param grid  tiny bit of external potential vee
!> \param dr    step size for finite difference
!> \param dgrid derivatives of grid
! **************************************************************************************************
   SUBROUTINE d_finite_difference(grid, dr, dgrid)
      REAL(KIND=dp), DIMENSION(0:, 0:, 0:), INTENT(IN)   :: grid
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dr
      REAL(KIND=dp), DIMENSION(1:, 1:, 1:, :), &
         INTENT(OUT)                                     :: dgrid

      INTEGER                                            :: i, j, k

      DO i = 1, SIZE(dgrid, 1)
         DO j = 1, SIZE(dgrid, 2)
            DO k = 1, SIZE(dgrid, 3)
               dgrid(i, j, k, 1) = 0.5*(grid(i+1, j, k)-grid(i-1, j, k))/dr(1)
               dgrid(i, j, k, 2) = 0.5*(grid(i, j+1, k)-grid(i, j-1, k))/dr(2)
               dgrid(i, j, k, 3) = 0.5*(grid(i, j, k+1)-grid(i, j, k-1))/dr(3)
            END DO
         END DO
      END DO
   END SUBROUTINE d_finite_difference

! **************************************************************************************************
!> \brief             trilinear interpolation function that interpolates value at r based
!>                    on 2x2x2 grid points around r in subgrid
!> \param r           where to interpolate to
!> \param subgrid     part of external potential on a grid
!> \param origin      center of grid
!> \param dr          step size
!> \retval value_at_r interpolated value of external potential
! **************************************************************************************************
   FUNCTION trilinear_interpolation(r, subgrid, origin, dr) RESULT(value_at_r)
      REAL(KIND=dp), DIMENSION(3)                        :: r
      REAL(KIND=dp), DIMENSION(:, :, :)                  :: subgrid
      REAL(KIND=dp), DIMENSION(3)                        :: origin, dr
      REAL(KIND=dp)                                      :: value_at_r

      REAL(KIND=dp), DIMENSION(3)                        :: norm_r, norm_r_rev

      norm_r = (r-origin)/dr
      norm_r_rev = 1-norm_r
      value_at_r = subgrid(1, 1, 1)*PRODUCT(norm_r_rev)+ &
                   subgrid(2, 1, 1)*norm_r(1)*norm_r_rev(2)*norm_r_rev(3)+ &
                   subgrid(1, 2, 1)*norm_r_rev(1)*norm_r(2)*norm_r_rev(3)+ &
                   subgrid(1, 1, 2)*norm_r_rev(1)*norm_r_rev(2)*norm_r(3)+ &
                   subgrid(1, 2, 2)*norm_r_rev(1)*norm_r(2)*norm_r(3)+ &
                   subgrid(2, 1, 2)*norm_r(1)*norm_r_rev(2)*norm_r(3)+ &
                   subgrid(2, 2, 1)*norm_r(1)*norm_r(2)*norm_r_rev(3)+ &
                   subgrid(2, 2, 2)*PRODUCT(norm_r)
   END FUNCTION trilinear_interpolation

! **************************************************************************************************
!> \brief          get a correct value for possible out of bounds index using periodic
!>                  boundary conditions
!> \param i ...
!> \param lowbound ...
!> \param upbound ...
!> \retval pbc_index ...
! **************************************************************************************************
   FUNCTION pbc_index(i, lowbound, upbound)
      INTEGER                                            :: i, lowbound, upbound, pbc_index

      IF (i < lowbound) THEN
         pbc_index = upbound+i-lowbound+1
      ELSE IF (i > upbound) THEN
         pbc_index = lowbound+i-upbound-1
      ELSE
         pbc_index = i
      END IF
   END FUNCTION pbc_index

END MODULE qs_external_potential
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief SCPTB short range repulsive core interaction
!> \author Juerg Hutter - 12.2011
!> \par History
!>      JHU (12.2011) - create
! **************************************************************************************************
MODULE scptb_core_interactions
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_array_init,&
                                              atprop_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE scptb_types,                     ONLY: get_scptb_parameter,&
                                              scptb_parameter_p_type,&
                                              scptb_parameter_type
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'scptb_core_interactions'

   PUBLIC :: scptb_core_interaction

CONTAINS

! **************************************************************************************************
!> \brief Evaluates the repulsive core interactions for SCPTB methods
!> \param qs_env ...
!> \param calculate_forces ...
!> \date 12.2011 [jhu]
!> \author Juerg Hutter [jhu] - University of Zurich
! **************************************************************************************************
   SUBROUTINE scptb_core_interaction(qs_env, calculate_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(in)                                :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'scptb_core_interaction', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, handle, iatom, ikind, &
                                                            jatom, jkind, natom, nkind
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: atener, defined, use_virial
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: scptb_defined
      REAL(KIND=dp)                                      :: dr1, enuc, enucij, enuclear
      REAL(KIND=dp), DIMENSION(3)                        :: force_ab, rij
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_core
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(scptb_parameter_p_type), DIMENSION(:), &
         POINTER                                         :: scptb_kind_param
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind_a, scptb_kind_b
      TYPE(virial_type), POINTER                         :: virial

      enuclear = 0.0_dp
      NULLIFY (atprop)
      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("scptb_core_interactions.F",94)
      CALL get_qs_env(qs_env=qs_env, para_env=para_env, dft_control=dft_control, &
                      cell=cell, virial=virial, energy=energy, atprop=atprop)

      ! Parameters
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      ! atomic energy decomposition
      atener = atprop%energy
      IF (atener) THEN
         CALL get_qs_env(qs_env=qs_env, particle_set=particle_set)
         natom = SIZE(particle_set)
         CALL atprop_array_init(atprop%atecc, natom)
      END IF

      CALL get_qs_env(qs_env=qs_env, sab_core=sab_core, atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set)

      nkind = SIZE(atomic_kind_set)
      ! Possibly compute forces
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, particle_set=particle_set, force=force)
         natom = SIZE(particle_set)
         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind)
      END IF

      ALLOCATE (scptb_kind_param(nkind), scptb_defined(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind_a)
         scptb_kind_param(ikind)%scptb_param => scptb_kind_a
         CALL get_scptb_parameter(scptb_kind_a, defined=defined)
         scptb_defined(ikind) = defined
      END DO
      CALL neighbor_list_iterator_create(nl_iterator, sab_core)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, r=rij)
         IF (.NOT. scptb_defined(ikind)) CYCLE
         IF (.NOT. scptb_defined(jkind)) CYCLE
         scptb_kind_a => scptb_kind_param(ikind)%scptb_param
         scptb_kind_b => scptb_kind_param(jkind)%scptb_param
         dr1 = DOT_PRODUCT(rij, rij)
         enucij = 0._dp
         IF (dr1 > 0.00001_dp) THEN
            ! Core-Core energy and derivatives
            IF (calculate_forces) THEN
               CALL corecore(scptb_kind_a, scptb_kind_b, rij, enuc=enuc, denuc=force_ab)

               atom_a = atom_of_kind(iatom)
               atom_b = atom_of_kind(jatom)

               ! Sum up force components
               force(ikind)%all_potential(1, atom_a) = force(ikind)%all_potential(1, atom_a)-force_ab(1)
               force(jkind)%all_potential(1, atom_b) = force(jkind)%all_potential(1, atom_b)+force_ab(1)

               force(ikind)%all_potential(2, atom_a) = force(ikind)%all_potential(2, atom_a)-force_ab(2)
               force(jkind)%all_potential(2, atom_b) = force(jkind)%all_potential(2, atom_b)+force_ab(2)

               force(ikind)%all_potential(3, atom_a) = force(ikind)%all_potential(3, atom_a)-force_ab(3)
               force(jkind)%all_potential(3, atom_b) = force(jkind)%all_potential(3, atom_b)+force_ab(3)

               IF (use_virial) THEN
                  CALL virial_pair_force(virial%pv_virial, -1._dp, force_ab, rij)
               END IF

            ELSE
               CALL corecore(scptb_kind_a, scptb_kind_b, rij, enuc=enuc)
            END IF
            enucij = enucij+enuc
         END IF
         IF (atener) THEN
            atprop%atecc(iatom) = atprop%atecc(iatom)+0.5_dp*enucij
            atprop%atecc(jatom) = atprop%atecc(jatom)+0.5_dp*enucij
         END IF
         enuclear = enuclear+enucij
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (scptb_kind_param, scptb_defined)

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind)
      END IF

      CALL mp_sum(enuclear, para_env%group)
      energy%repulsive = enuclear

      CALL timestop(handle)

   END SUBROUTINE scptb_core_interaction

! **************************************************************************************************
!> \brief ...
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param enuc ...
!> \param denuc ...
! **************************************************************************************************
   SUBROUTINE corecore(sepi, sepj, rij, enuc, denuc)
      TYPE(scptb_parameter_type), POINTER                :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), INTENT(OUT)                              :: enuc
      REAL(dp), DIMENSION(3), INTENT(OUT), OPTIONAL      :: denuc

      CHARACTER(len=*), PARAMETER :: routineN = 'corecore', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: a, c, d, de, expa, x

!
!  (c+d*x*x) * exp[-(a*x)]
!

      a = SQRT(sepi%crep(1)*sepj%crep(1))
      c = 0.5_dp*(sepi%crep(2)+sepj%crep(2))
      d = 0.5_dp*(sepi%crep(3)+sepj%crep(3))

      x = SQRT(rij(1)**2+rij(2)**2+rij(3)**2)
      expa = EXP(-(a*x))

      enuc = (c+d*x**2)*expa

      IF (PRESENT(denuc)) THEN
         !de = (d-(c+d*x)*(a+2._dp*b*x)) * expa
         de = (-a*c+(-a*x+2.0_dp)*(d*x))*expa
         denuc = de*rij/x
      END IF

   END SUBROUTINE corecore

END MODULE scptb_core_interactions

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility subroutine for qs energy calculation
!> \par History
!>      none
!> \author MK (29.10.2002)
! **************************************************************************************************
MODULE qs_energy_utils
   USE atprop_types,                    ONLY: atprop_array_add,&
                                              atprop_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_control_utils,                ONLY: read_becke_section,&
                                              read_ddapc_section
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_copy, cp_dbcsr_get_block_p, &
        cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_p_type, cp_dbcsr_set, cp_dbcsr_type
   USE cp_external_control,             ONLY: external_control
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE efield_utils,                    ONLY: calculate_ecore_efield
   USE et_coupling,                     ONLY: calc_et_coupling
   USE input_constants,                 ONLY: kg_tnadd_atomic,&
                                              kg_tnadd_embed
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kg_environment,                  ONLY: kg_build_neighborlist,&
                                              kg_build_subsets
   USE kinds,                           ONLY: dp
   USE kpoint_methods,                  ONLY: kpoint_density_matrices,&
                                              kpoint_density_transform,&
                                              kpoint_init_cell_index
   USE kpoint_types,                    ONLY: kpoint_type
   USE molecule_types_new,              ONLY: molecule_of_atom,&
                                              molecule_type
   USE mp2,                             ONLY: mp2_main
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_pool_types,                   ONLY: pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: pw_p_type
   USE qs_core_energies,                ONLY: calculate_ecore_overlap,&
                                              calculate_ecore_self
   USE qs_core_hamiltonian,             ONLY: build_core_hamiltonian_matrix
   USE qs_dftb_dispersion,              ONLY: calculate_dftb_dispersion
   USE qs_dftb_matrices,                ONLY: build_dftb_matrices
   USE qs_dispersion_pairpot,           ONLY: calculate_dispersion_pairpot
   USE qs_dispersion_types,             ONLY: qs_dispersion_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_methods,          ONLY: qs_env_update_s_mstruct
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_external_density,             ONLY: external_read_density
   USE qs_external_potential,           ONLY: external_c_potential,&
                                              external_e_potential
   USE qs_integrate_potential,          ONLY: integrate_v_core_rspace
   USE qs_ks_methods,                   ONLY: calculate_w_matrix,&
                                              calculate_w_matrix_ot,&
                                              qs_ks_allocate_basics,&
                                              qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                              set_ks_env
   USE qs_linres_module,                ONLY: linres_calculation_low
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_neighbor_lists,               ONLY: build_qs_neighbor_lists
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf,                          ONLY: scf
   USE scf_control_types,               ONLY: scf_control_type
   USE scptb_core_interactions,         ONLY: scptb_core_interaction
   USE scptb_core_matrix,               ONLY: build_scptb_core_matrix
   USE se_core_core,                    ONLY: se_core_core_interaction
   USE se_core_matrix,                  ONLY: build_se_core_matrix
   USE xas_methods,                     ONLY: xas
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_energy_utils'

   PUBLIC :: qs_energies_init, &
             qs_energies_compute_matrix_w, &
             qs_energies_properties, &
             qs_energies_mp2

CONTAINS

! **************************************************************************************************
!> \brief Refactoring of qs_energies_scf. Driver routine for the initial
!>        setup and calculations for a qs energy calculation
!> \param qs_env ...
!> \param calc_forces ...
!> \par History
!>      05.2013 created [Florian Schiffmann]
! **************************************************************************************************

   SUBROUTINE qs_energies_init(qs_env, calc_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calc_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_energies_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: img, ispin, nimg, nspin
      LOGICAL                                            :: has_unit_metric, molecule_only
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_w_mp2
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_s, matrix_w
      TYPE(cp_dbcsr_type), POINTER                       :: matrix
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      NULLIFY (ks_env, matrix_w, matrix_w_mp2, matrix_s, dft_control)

      CALL qs_energies_init_kg(qs_env, molecule_only)
      CALL qs_energies_init_hamiltonians(qs_env, calc_forces, molecule_only)
      CALL qs_ks_allocate_basics(qs_env)
      CALL get_qs_env(qs_env, has_unit_metric=has_unit_metric)

      ! if need forces allocate energy weighted density matrices
      IF (calc_forces .AND. .NOT. has_unit_metric) THEN
         CALL get_qs_env(qs_env, &
                         ks_env=ks_env, &
                         matrix_s_kp=matrix_s, &
                         dft_control=dft_control)
         nspin = dft_control%nspins
         nimg = dft_control%nimages
         matrix => matrix_s(1, 1)%matrix
         CALL cp_dbcsr_allocate_matrix_set(matrix_w, nspin, nimg)
         DO ispin = 1, nspin
            DO img = 1, nimg
               ALLOCATE (matrix_w(ispin, img)%matrix)
               CALL cp_dbcsr_init(matrix_w(ispin, img)%matrix)
               CALL cp_dbcsr_copy(matrix_w(ispin, img)%matrix, matrix, name="W MATRIX")
               CALL cp_dbcsr_set(matrix_w(ispin, img)%matrix, 0.0_dp)
            END DO
         END DO
         CALL set_ks_env(ks_env, matrix_w_kp=matrix_w)
         IF (ASSOCIATED(qs_env%mp2_env)) THEN
            ! create the MP2 energy weighted density matrix
            CALL cp_dbcsr_allocate_matrix_set(matrix_w_mp2, dft_control%nspins)
            DO ispin = 1, nspin
               ALLOCATE (matrix_w_mp2(ispin)%matrix)
               CALL cp_dbcsr_init(matrix_w_mp2(ispin)%matrix)
               CALL cp_dbcsr_copy(matrix_w_mp2(ispin)%matrix, matrix, &
                                  name="W MATRIX MP2")
               CALL cp_dbcsr_set(matrix_w_mp2(ispin)%matrix, 0.0_dp)
            END DO
            CALL set_ks_env(ks_env, matrix_w_mp2=matrix_w_mp2)
         END IF
      ENDIF

   END SUBROUTINE qs_energies_init

! **************************************************************************************************
!> \brief Refactoring of qs_energies_scf. Puts initialization of the Kim-Gordon
!>        settings into separate subroutine
!> \param qs_env ...
!> \param molecule_only ...
!> \par History
!>      05.2013 created [Florian Schiffmann]
! **************************************************************************************************

   SUBROUTINE qs_energies_init_kg(qs_env, molecule_only)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL                                            :: molecule_only

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_energies_init_kg', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, isubset, natom
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, para_env)
      molecule_only = .FALSE.

      CALL get_qs_env(qs_env, dft_control=dft_control, para_env=para_env)
      IF (dft_control%qs_control%do_kg) THEN

         ! create neighbor lists with molecular blocks
         molecule_only = .TRUE.

         ! get the set of molecules
         CALL get_qs_env(qs_env=qs_env, molecule_set=molecule_set, natom=natom)

         qs_env%kg_env%natom = natom

         ! store set of molecules in kg_env
         qs_env%kg_env%molecule_set => molecule_set

         ! build the (new) full neighborlist
         CALL kg_build_neighborlist(qs_env, sab_orb=qs_env%kg_env%sab_orb_full)

         IF (.NOT. ALLOCATED(qs_env%kg_env%atom_to_molecule)) THEN
            ALLOCATE (qs_env%kg_env%atom_to_molecule(natom))

            ! get the mapping from atoms to molecules
            CALL molecule_of_atom(molecule_set, atom_to_mol=qs_env%kg_env%atom_to_molecule)
         END IF

         IF (qs_env%kg_env%tnadd_method == kg_tnadd_embed) THEN
            ! allocate the subset list
            IF (.NOT. ASSOCIATED(qs_env%kg_env%subset_of_mol)) ALLOCATE (qs_env%kg_env%subset_of_mol(SIZE(molecule_set)))

            CALL kg_build_subsets(qs_env%kg_env, para_env)

            DO isubset = 1, qs_env%kg_env%nsubsets
               ! build the (new) molecular neighborlist of the current subset
               CALL kg_build_neighborlist(qs_env, sab_orb=qs_env%kg_env%subset(isubset)%sab_orb, molecular=.TRUE., &
                                          subset_of_mol=qs_env%kg_env%subset_of_mol, current_subset=isubset)
            END DO
         ELSE IF (qs_env%kg_env%tnadd_method == kg_tnadd_atomic) THEN
            ! build the A-C list for the nonadditive kinetic energy potential
            CALL kg_build_neighborlist(qs_env, sac_kin=qs_env%kg_env%sac_kin)
         END IF

      END IF
      CALL timestop(handle)

   END SUBROUTINE qs_energies_init_kg

! **************************************************************************************************
!> \brief Refactoring of qs_energies_scf. Moves computation of the different
!>        core hamiltonians into separate subroutine
!> \param qs_env        QS environment
!> \param calc_forces   Calculate forces
!> \param molecule_only restrict neighbor list to molecules
!> \par History
!>      05.2013 created [Florian Schiffmann]
!>      08.2014 Kpoints [JGH]
! **************************************************************************************************

   SUBROUTINE qs_energies_init_hamiltonians(qs_env, calc_forces, molecule_only)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calc_forces
      LOGICAL                                            :: molecule_only

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_energies_init_hamiltonians', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: do_kpoints
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_nl
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(section_vals_type), POINTER                   :: input

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, &
                      input=input, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      kpoints=kpoints, &
                      do_kpoints=do_kpoints)

      ! create neighbor lists for standard use in QS
      CALL build_qs_neighbor_lists(qs_env, para_env, molecular=molecule_only, &
                                   force_env_section=input)

      ! calculate cell index for k-point calculations
      IF (do_kpoints) THEN
         CALL get_qs_env(qs_env, sab_kp=sab_nl)
         CALL kpoint_init_cell_index(kpoints, sab_nl, para_env, dft_control)
      ENDIF

      dft_control%qs_control%becke_control%need_pot = .TRUE.

      ! Calculate the overlap and the core Hamiltonian integral matrix
      IF (dft_control%qs_control%semi_empirical) THEN
         CALL build_se_core_matrix(qs_env=qs_env, para_env=para_env, &
                                   calculate_forces=.FALSE.)
         CALL qs_env_update_s_mstruct(qs_env)
         CALL se_core_core_interaction(qs_env, para_env, calculate_forces=.FALSE.)
         CALL get_qs_env(qs_env=qs_env, dispersion_env=dispersion_env, energy=energy)
         CALL calculate_dispersion_pairpot(qs_env, dispersion_env, energy%dispersion, calc_forces)
      ELSEIF (dft_control%qs_control%dftb) THEN
         CALL build_dftb_matrices(qs_env=qs_env, para_env=para_env, &
                                  calculate_forces=.FALSE.)
         CALL calculate_dftb_dispersion(qs_env=qs_env, para_env=para_env, &
                                        calculate_forces=.FALSE.)
         CALL qs_env_update_s_mstruct(qs_env)
      ELSEIF (dft_control%qs_control%scptb) THEN
         CALL build_scptb_core_matrix(qs_env=qs_env, calculate_forces=.FALSE.)
         CALL qs_env_update_s_mstruct(qs_env)
         CALL scptb_core_interaction(qs_env, calculate_forces=.FALSE.)
         CALL get_qs_env(qs_env=qs_env, dispersion_env=dispersion_env, energy=energy)
         CALL calculate_dispersion_pairpot(qs_env, dispersion_env, energy%dispersion, calc_forces)
      ELSE
         CALL build_core_hamiltonian_matrix(qs_env=qs_env, calculate_forces=.FALSE.)
         CALL qs_env_update_s_mstruct(qs_env)
         CALL calculate_ecore_self(qs_env)
         CALL calculate_ecore_efield(qs_env, calculate_forces=.FALSE.)
         CALL calculate_ecore_overlap(qs_env, para_env, calculate_forces=.FALSE.)
         !swap external_e_potential before external_c_potential, to ensure
         !that external potential on grid is loaded before calculating energy of cores
         CALL external_e_potential(qs_env)
         IF (.NOT. dft_control%qs_control%gapw) THEN
            CALL external_c_potential(qs_env, calculate_forces=.FALSE.)
         END IF

         ! ZMP addition to read external density
         CALL external_read_density(qs_env)

         ! Add possible pair potential dispersion energy - Evaluate first so we can print
         ! energy info at the end of the SCF
         CALL get_qs_env(qs_env=qs_env, dispersion_env=dispersion_env, energy=energy)
         CALL calculate_dispersion_pairpot(qs_env, dispersion_env, energy%dispersion, calc_forces)

      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_energies_init_hamiltonians

! **************************************************************************************************
!> \brief Refactoring of qs_energies_scf. Moves computation of matrix_w
!>        into separate subroutine
!> \param qs_env ...
!> \param calc_forces ...
!> \par History
!>      05.2013 created [Florian Schiffmann]
! **************************************************************************************************

   SUBROUTINE qs_energies_compute_matrix_w(qs_env, calc_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calc_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_energies_compute_matrix_w', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, is, ispin, nao, nspin
      LOGICAL                                            :: do_kpoints, has_unit_metric
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s, matrix_w, &
                                                            matrix_w_mp2, mo_derivs, rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_s_kp, matrix_w_kp
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: fmwork
      TYPE(cp_fm_struct_type), POINTER                   :: ao_ao_fmstruct
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_nl
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(scf_control_type), POINTER                    :: scf_control

      CALL timeset(routineN, handle)

      ! if calculate forces, time to compute the w matrix
      CALL get_qs_env(qs_env, has_unit_metric=has_unit_metric)

      IF (calc_forces .AND. .NOT. has_unit_metric) THEN
         CALL get_qs_env(qs_env, do_kpoints=do_kpoints)

         IF (do_kpoints) THEN

            CALL get_qs_env(qs_env, &
                            matrix_w_kp=matrix_w_kp, &
                            matrix_s_kp=matrix_s_kp, &
                            sab_orb=sab_nl, &
                            mos=mos, &
                            kpoints=kpoints)

            CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff, nao=nao)
            CALL cp_fm_struct_create(fmstruct=ao_ao_fmstruct, nrow_global=nao, ncol_global=nao, &
                                     template_fmstruct=mo_coeff%matrix_struct)

            ALLOCATE (fmwork(2))
            DO is = 1, SIZE(fmwork)
               NULLIFY (fmwork(is)%matrix)
               CALL cp_fm_create(fmwork(is)%matrix, matrix_struct=ao_ao_fmstruct)
            END DO
            CALL cp_fm_struct_release(ao_ao_fmstruct)

            ! energy weighted density matrices in k-space
            CALL kpoint_density_matrices(kpoints, energy_weighted=.TRUE.)
            ! energy weighted density matrices in real space
            CALL kpoint_density_transform(kpoints, matrix_w_kp, .TRUE., &
                                          matrix_s_kp(1, 1)%matrix, sab_nl, fmwork)

            DO is = 1, SIZE(fmwork)
               CALL cp_fm_release(fmwork(is)%matrix)
            END DO
            DEALLOCATE (fmwork)

         ELSE

            NULLIFY (dft_control, rho_ao)
            CALL get_qs_env(qs_env, &
                            matrix_w=matrix_w, &
                            matrix_ks=matrix_ks, &
                            matrix_s=matrix_s, &
                            matrix_w_mp2=matrix_w_mp2, &
                            mo_derivs=mo_derivs, &
                            scf_control=scf_control, &
                            mos=mos, &
                            rho=rho, &
                            dft_control=dft_control)

            CALL qs_rho_get(rho, rho_ao=rho_ao)

            nspin = SIZE(mos)
            DO ispin = 1, nspin
               mo_set => mos(ispin)%mo_set
               IF (dft_control%roks) THEN
                  IF (scf_control%use_ot) THEN
                     IF (ispin > 1) THEN
                        ! not very elegant, indeed ...
                        CALL cp_dbcsr_set(matrix_w(ispin)%matrix, 0.0_dp)
                     ELSE
                        CALL calculate_w_matrix_ot(mo_set, mo_derivs(ispin)%matrix, &
                                                   matrix_w(ispin)%matrix, matrix_s(1)%matrix)
                     END IF
                  ELSE
                     CALL calculate_w_matrix(mo_set=mo_set, &
                                             matrix_ks=matrix_ks(ispin)%matrix, &
                                             matrix_p=rho_ao(ispin)%matrix, &
                                             matrix_w=matrix_w(ispin)%matrix)
                  END IF
               ELSE
                  IF (scf_control%use_ot) THEN
                     CALL calculate_w_matrix_ot(mo_set, mo_derivs(ispin)%matrix, &
                                                matrix_w(ispin)%matrix, matrix_s(1)%matrix)
                  ELSE
                     CALL calculate_w_matrix(mo_set, matrix_w(ispin)%matrix)
                  END IF
               END IF
               ! if MP2 time to update the W matrix with the MP2 contribution
               IF (ASSOCIATED(qs_env%mp2_env)) THEN
                  CALL cp_dbcsr_add(matrix_w(ispin)%matrix, matrix_w_mp2(ispin)%matrix, 1.0_dp, -1.0_dp)
               END IF
            END DO

         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_energies_compute_matrix_w

! **************************************************************************************************
!> \brief Refactoring of qs_energies_scf. Moves computation of properties
!>        into separate subroutine
!> \param qs_env ...
!> \par History
!>      05.2013 created [Florian Schiffmann]
! **************************************************************************************************

   SUBROUTINE qs_energies_properties(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_energies_properties', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: do_et
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: v_hartree_rspace
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(section_vals_type), POINTER                   :: input, rest_b_section

      NULLIFY (atprop, energy, pw_env)
      CALL timeset(routineN, handle)

      NULLIFY (v_hartree_rspace%pw)

      ! atomic energies using Mulliken partition
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      input=input, &
                      atprop=atprop, &
                      energy=energy, &
                      v_hartree_rspace=v_hartree_rspace%pw, &
                      pw_env=pw_env)
      IF (atprop%energy) THEN
         CALL qs_energies_mulliken(qs_env)
         IF (.NOT. dft_control%qs_control%semi_empirical .AND. &
             .NOT. dft_control%qs_control%dftb) THEN
            ! Nuclear charge correction
            CALL integrate_v_core_rspace(v_hartree_rspace, qs_env)
            ! Kohn-Sham Functional corrections
         END IF
         CALL atprop_array_add(atprop%atener, atprop%ateb)
         CALL atprop_array_add(atprop%atener, atprop%ateself)
         CALL atprop_array_add(atprop%atener, atprop%atexc)
         CALL atprop_array_add(atprop%atener, atprop%atecoul)
         CALL atprop_array_add(atprop%atener, atprop%atevdw)
         CALL atprop_array_add(atprop%atener, atprop%atecc)
         CALL atprop_array_add(atprop%atener, atprop%ate1c)
      END IF

      ! **********  Calculate the electron transfer coupling elements********
      do_et = .FALSE.
      do_et = dft_control%qs_control%et_coupling_calc
      IF (do_et) THEN
         qs_env%et_coupling%energy = energy%total
         qs_env%et_coupling%keep_matrix = .TRUE.
         qs_env%et_coupling%first_run = .TRUE.
         CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., just_energy=.TRUE.)
         qs_env%et_coupling%first_run = .FALSE.
         IF (dft_control%qs_control%ddapc_restraint) THEN
            rest_b_section => section_vals_get_subs_vals(input, "PROPERTIES%ET_COUPLING%DDAPC_RESTRAINT_B")
            CALL read_ddapc_section(qs_control=dft_control%qs_control, &
                                    ddapc_restraint_section=rest_b_section)
         END IF
         IF (dft_control%qs_control%becke_restraint) THEN
            rest_b_section => section_vals_get_subs_vals(input, "PROPERTIES%ET_COUPLING%BECKE_RESTRAINT_B")
            CALL read_becke_section(qs_control=dft_control%qs_control, &
                                    becke_section=rest_b_section)
         END IF
         CALL scf(qs_env=qs_env)
         qs_env%et_coupling%keep_matrix = .TRUE.

         CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., just_energy=.TRUE.)
         CALL calc_et_coupling(qs_env)
         IF (dft_control%qs_control%becke_restraint) THEN
            CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)
            CALL pw_pool_give_back_pw(auxbas_pw_pool, &
                                      dft_control%qs_control%becke_control%becke_pot%pw)
            dft_control%qs_control%becke_control%need_pot = .TRUE.
         END IF
      END IF

      !Properties
      IF (dft_control%do_xas_calculation) THEN
         CALL xas(qs_env, dft_control)
      END IF

      ! Compute Linear Response properties as post-scf
      IF (.NOT. qs_env%linres_run) THEN
         CALL linres_calculation_low(qs_env)
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_energies_properties

! **************************************************************************************************
!> \brief   Use a simple Mulliken-like energy decomposition
!> \param qs_env ...
!> \date    07.2011
!> \author  JHU
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE qs_energies_mulliken(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_energies_mulliken', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_h, matrix_ks, rho_ao
      TYPE(qs_rho_type), POINTER                         :: rho

      NULLIFY (atprop, matrix_h, matrix_ks, rho, rho_ao)
      CALL get_qs_env(qs_env=qs_env, matrix_ks=matrix_ks, matrix_h=matrix_h, &
                      rho=rho, atprop=atprop)
      CALL qs_rho_get(rho, rho_ao=rho_ao)

      IF (atprop%energy) THEN
         ! E = 0.5*Tr(H*P+F*P)
         atprop%atener = 0._dp
         DO ispin = 1, SIZE(rho_ao)
            CALL atom_trace(matrix_h(1)%matrix, rho_ao(ispin)%matrix, &
                            0.5_dp, atprop%atener)
            CALL atom_trace(matrix_ks(ispin)%matrix, rho_ao(ispin)%matrix, &
                            0.5_dp, atprop%atener)
         END DO

      END IF

   END SUBROUTINE qs_energies_mulliken

! **************************************************************************************************
!> \brief Compute partial trace of product of two matrices
!> \param amat ...
!> \param bmat ...
!> \param factor ...
!> \param atrace ...
!> \par History
!>      06.2004 created [Joost VandeVondele]
!> \note
!>      charges are computed per spin in the LSD case
! **************************************************************************************************
   SUBROUTINE atom_trace(amat, bmat, factor, atrace)
      TYPE(cp_dbcsr_type), POINTER                       :: amat, bmat
      REAL(kind=dp), INTENT(IN)                          :: factor
      REAL(KIND=dp), DIMENSION(:), POINTER               :: atrace

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_trace', routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, iblock_col, iblock_row, nblock
      LOGICAL                                            :: found
      REAL(kind=dp)                                      :: btr, mult
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a_block, b_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL cp_dbcsr_get_info(bmat, nblkrows_total=nblock)
      IF(.NOT.(nblock == SIZE(atrace)))CALL cp__a("qs_energy_utils.F",631)

      CALL cp_dbcsr_iterator_start(iter, bmat)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, b_block, blk)
         CALL cp_dbcsr_get_block_p(matrix=amat, &
                                   row=iblock_row, col=iblock_col, BLOCK=a_block, found=found)

         ! we can cycle if a block is not present
         IF (.NOT. (ASSOCIATED(b_block) .AND. ASSOCIATED(a_block))) CYCLE

         IF (iblock_row .EQ. iblock_col) THEN
            mult = 0.5_dp ! avoid double counting of diagonal blocks
         ELSE
            mult = 1.0_dp
         ENDIF
         btr = factor*mult*SUM(a_block*b_block)
         atrace(iblock_row) = atrace(iblock_row)+btr
         atrace(iblock_col) = atrace(iblock_col)+btr

      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

   END SUBROUTINE atom_trace

! **************************************************************************************************
!> \brief Enters the mp2 part of cp2k
!> \param qs_env ...
!> \param calc_forces ...
! **************************************************************************************************

   SUBROUTINE qs_energies_mp2(qs_env, calc_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calc_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_energies_mp2', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: should_stop

! Compute MP2 energy

      IF (ASSOCIATED(qs_env%mp2_env)) THEN

         CALL external_control(should_stop, "MP2", target_time=qs_env%target_time, &
                               start_time=qs_env%start_time)

         CALL mp2_main(qs_env=qs_env, calc_forces=calc_forces)
      ENDIF

   END SUBROUTINE qs_energies_mp2

END MODULE qs_energy_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Perform a QUICKSTEP wavefunction optimization (single point)
!> \par History
!>      none
!> \author MK (29.10.2002)
! **************************************************************************************************
MODULE qs_energy
   USE almo_scf,                        ONLY: almo_entry_scf
   USE cp_control_types,                ONLY: dft_control_type
   USE dm_ls_scf,                       ONLY: ls_scf
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_energy_utils,                 ONLY: qs_energies_compute_matrix_w,&
                                              qs_energies_init,&
                                              qs_energies_mp2,&
                                              qs_energies_properties
   USE qs_environment_methods,          ONLY: qs_env_rebuild_pw_env
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_scf,                          ONLY: scf
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_energy'

   PUBLIC :: qs_energies

CONTAINS

! **************************************************************************************************
!> \brief   Driver routine for QUICKSTEP single point wavefunction optimization.
!> \param qs_env ...
!> \param consistent_energies ...
!> \param calc_forces ...
!> \date    29.10.2002
!> \par History
!>          - consistent_energies option added (25.08.2005, TdK)
!>          - introduced driver for energy in order to properly decide between
!>            SCF or RTP (fschiff 02.09)
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE qs_energies(qs_env, consistent_energies, calc_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: consistent_energies, calc_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_energies', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: my_calc_forces, run_rtp
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy

      CALL timeset(routineN, handle)

      my_calc_forces = .FALSE.
      IF (PRESENT(calc_forces)) my_calc_forces = calc_forces

      CALL qs_env_rebuild_pw_env(qs_env)

      CALL get_qs_env(qs_env=qs_env, run_rtp=run_rtp)
      IF (.NOT. run_rtp) THEN

         NULLIFY (dft_control, energy)
         CALL qs_energies_init(qs_env, my_calc_forces)
         CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, energy=energy)

         ! *** Perform a SCF run ***
         IF (dft_control%qs_control%do_ls_scf) THEN
            CALL ls_scf(qs_env=qs_env)
         ELSE IF (dft_control%qs_control%do_almo_scf) THEN
            CALL almo_entry_scf(qs_env=qs_env, calc_forces=my_calc_forces)
         ELSE
            CALL scf(qs_env=qs_env)

            ! Compute MP2 energy
            CALL qs_energies_mp2(qs_env, my_calc_forces)

            ! if calculate forces, time to compute the w matrix
            CALL qs_energies_compute_matrix_w(qs_env, my_calc_forces)

         END IF

         IF (PRESENT(consistent_energies)) THEN
            IF (consistent_energies) THEN
               CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE., just_energy=.TRUE.)
               ! add MP2 energy if necessary
               IF (ASSOCIATED(qs_env%mp2_env)) THEN
                  energy%total = energy%total+energy%mp2
               END IF
            END IF
         END IF

         CALL qs_energies_properties(qs_env)

      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_energies

END MODULE qs_energy
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module to perform a counterpoise correction (BSSE)
!> \par History
!>      6.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE bsse
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE cp2k_info,                       ONLY: write_restart_header
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_methods,               ONLY: create_small_subsys
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_release,&
                                              cp_subsys_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: do_qs
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set,&
                                              section_vals_write
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE particle_list_types,             ONLY: particle_list_type
   USE qs_energy,                       ONLY: qs_energies
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment,                  ONLY: qs_init
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_env_create,&
                                              qs_env_release,&
                                              qs_environment_type
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'bsse'

   PUBLIC :: do_bsse_calculation

CONTAINS

! **************************************************************************************************
!> \brief Perform an COUNTERPOISE CORRECTION (BSSE)
!>      For a 2-body system the correction scheme can be represented as:
!>
!>      E_{AB}^{2}        = E_{AB}(AB) - E_A(AB) - E_B(AB)  [BSSE-corrected interaction energy]
!>      E_{AB}^{2,uncorr} = E_{AB}(AB) - E_A(A)  - E_B(B)
!>      E_{AB}^{CP}       = E_{AB}(AB) + [ E_A(A) - E_A(AB) ] + [ E_B(B) - E_B(AB) ]
!>                                                          [CP-corrected total energy of AB]
!> \param force_env ...
!> \param globenv ...
!> \par History
!>      06.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE do_bsse_calculation(force_env, globenv)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_bsse_calculation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, istart, k, num_of_conf, Num_of_Frag
      INTEGER, DIMENSION(:, :), POINTER                  :: conf
      LOGICAL                                            :: explicit, should_stop
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Em
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: bsse_section, fragment_energies_section, &
                                                            n_frags, root_section

      NULLIFY (bsse_section, n_frags, Em, conf)
      logger => cp_get_default_logger()
      root_section => force_env%root_section
      bsse_section => section_vals_get_subs_vals(force_env%force_env_section, "BSSE")
      n_frags => section_vals_get_subs_vals(bsse_section, "FRAGMENT")
      CALL section_vals_get(n_frags, n_repetition=Num_of_Frag)

      ! Number of configurations
      num_of_conf = 0
      DO k = 1, Num_of_frag
         num_of_conf = num_of_conf+FACT(Num_of_frag)/(FACT(k)*FACT(Num_of_frag-k))
      END DO
      ALLOCATE (conf(num_of_conf, Num_of_frag))
      ALLOCATE (Em(num_of_conf))
      CALL gen_Nbody_conf(Num_of_frag, conf)

      should_stop = .FALSE.
      istart = 0
      fragment_energies_section => section_vals_get_subs_vals(bsse_section, "FRAGMENT_ENERGIES")
      CALL section_vals_get(fragment_energies_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(fragment_energies_section, "_DEFAULT_KEYWORD_", n_rep_val=istart)
         DO i = 1, istart
            CALL section_vals_val_get(fragment_energies_section, "_DEFAULT_KEYWORD_", r_val=Em(i), &
                                      i_rep_val=i)
         END DO
      END IF
      ! Setup the iteration level for BSSE
      CALL cp_add_iter_level(logger%iter_info, "BSSE")
      CALL cp_iterate(logger%iter_info, last=.FALSE., iter_nr=istart)

      ! Evaluating the energy of the N-body cluster terms
      DO i = istart+1, num_of_conf
         CALL cp_iterate(logger%iter_info, last=(i == num_of_conf), iter_nr=i)
         CALL eval_bsse_energy(conf(i, :), Em(i), force_env, n_frags, &
                               root_section, globenv, should_stop)
         IF (should_stop) EXIT

         ! If no signal was received in the inner loop let's check also at this stage
         CALL external_control(should_stop, "BSSE", globenv=globenv)
         IF (should_stop) EXIT

         ! Dump Restart info only if the calculation of the energy of a configuration
         ! ended nicely..
         CALL section_vals_val_set(fragment_energies_section, "_DEFAULT_KEYWORD_", r_val=Em(i), &
                                   i_rep_val=i)
         CALL write_bsse_restart(bsse_section, root_section)
      END DO
      IF (.NOT. should_stop) CALL dump_bsse_results(conf, Em, num_of_frag, bsse_section)
      CALL cp_rm_iter_level(logger%iter_info, "BSSE")
      DEALLOCATE (Em)
      DEALLOCATE (conf)

   END SUBROUTINE do_bsse_calculation

! **************************************************************************************************
!> \brief Evaluate the N-body energy contribution to the BSSE evaluation
!> \param conf ...
!> \param Em ...
!> \param force_env ...
!> \param n_frags ...
!> \param root_section ...
!> \param globenv ...
!> \param should_stop ...
!> \par History
!>      07.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE eval_bsse_energy(conf, Em, force_env, n_frags, root_section, &
                               globenv, should_stop)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: conf
      REAL(KIND=dp), INTENT(OUT)                         :: Em
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: n_frags, root_section
      TYPE(global_environment_type), POINTER             :: globenv
      LOGICAL, INTENT(OUT)                               :: should_stop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_bsse_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, Num_of_sub_conf, Num_of_sub_frag
      INTEGER, DIMENSION(:, :), POINTER                  :: conf_loc
      REAL(KIND=dp)                                      :: my_energy
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Em_loc

      NULLIFY (conf_loc, Em_loc)
      should_stop = .FALSE.
      ! Count the number of subconfiguration to evaluate..
      Num_of_sub_frag = COUNT(conf == 1)
      Num_of_sub_conf = 0
      IF (Num_of_sub_frag == 1) THEN
         CALL eval_bsse_energy_low(force_env, conf, conf, n_frags, root_section, globenv, Em)
      ELSE
         my_energy = 0.0_dp
         DO k = 1, Num_of_sub_frag
            Num_of_sub_conf = Num_of_sub_conf+ &
                              FACT(Num_of_sub_frag)/(FACT(k)*FACT(Num_of_sub_frag-k))
         END DO
         ALLOCATE (conf_loc(Num_of_sub_conf, Num_of_sub_frag))
         ALLOCATE (Em_loc(Num_of_sub_conf))
         Em_loc = 0.0_dp
         CALL gen_Nbody_conf(Num_of_sub_frag, conf_loc)
         CALL make_plan_conf(conf, conf_loc)
         DO i = 1, Num_of_sub_conf
            CALL eval_bsse_energy_low(force_env, conf, conf_loc(i, :), n_frags, &
                                      root_section, globenv, Em_loc(i))
            CALL external_control(should_stop, "BSSE", globenv=globenv)
            IF (should_stop) EXIT
         END DO
         ! Energy
         k = COUNT(conf == 1)
         DO i = 1, Num_of_sub_conf
            j = COUNT(conf_loc(i, :) == 1)
            my_energy = my_energy+(-1.0_dp)**(k+j)*Em_loc(i)
         END DO
         Em = my_energy
         DEALLOCATE (Em_loc)
         DEALLOCATE (conf_loc)
      END IF

   END SUBROUTINE eval_bsse_energy

! **************************************************************************************************
!> \brief Evaluate the N-body energy contribution to the BSSE evaluation
!> \param force_env ...
!> \param conf ...
!> \param conf_loc ...
!> \param n_frags ...
!> \param root_section ...
!> \param globenv ...
!> \param energy ...
!> \par History
!>      07.2005 created [tlaino]
!>      2014/09/17 made atom list to be read from repeated occurance of LIST [LTong]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE eval_bsse_energy_low(force_env, conf, conf_loc, n_frags, &
                                   root_section, globenv, energy)
      TYPE(force_env_type), POINTER                      :: force_env
      INTEGER, DIMENSION(:), INTENT(IN)                  :: conf, conf_loc
      TYPE(section_vals_type), POINTER                   :: n_frags, root_section
      TYPE(global_environment_type), POINTER             :: globenv
      REAL(KIND=dp), INTENT(OUT)                         :: energy

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_bsse_energy_low', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name
      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: atom_type
      INTEGER                                            :: i, ir, isize, j, k, method_name_id, &
                                                            my_targ, n_rep, num_of_frag, old_size, &
                                                            present_charge, present_multpl
      INTEGER, DIMENSION(:), POINTER                     :: atom_index, atom_list, my_conf, tmplist
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys, subsys_loc
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(qs_energy_type), POINTER                      :: qs_energy
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: bsse_section, dft_section, &
                                                            force_env_section, subsys_section

      CALL section_vals_get(n_frags, n_repetition=num_of_frag)
      IF(.NOT.(SIZE(conf) == num_of_frag))CALL cp__a("bsse.F",256)
      NULLIFY (subsys_loc, subsys, particles, para_env, cell, atom_index, atom_type, tmplist, &
               force_env_section)
      CALL force_env_get(force_env, force_env_section=force_env_section)
      CALL section_vals_val_get(force_env_section, "METHOD", i_val=method_name_id)
      bsse_section => section_vals_get_subs_vals(force_env_section, "BSSE")
      subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      dft_section => section_vals_get_subs_vals(force_env_section, "DFT")

      ALLOCATE (my_conf(SIZE(conf)))
      my_conf = conf
      CALL force_env_get(force_env=force_env, subsys=subsys, para_env=para_env, &
                         cell=cell)
      CALL cp_subsys_get(subsys, particles=particles)
      isize = 0
      ALLOCATE (atom_index(isize))
      DO i = 1, num_of_frag
         IF (conf(i) == 1) THEN
            !
            ! Get the list of atoms creating the present fragment
            !
            old_size = isize
            CALL section_vals_val_get(n_frags, "LIST", i_rep_section=i, n_rep_val=n_rep)
            IF (n_rep /= 0) THEN
               DO ir = 1, n_rep
                  CALL section_vals_val_get(n_frags, "LIST", i_rep_section=i, i_rep_val=ir, i_vals=tmplist)
                  CALL reallocate(atom_index, 1, isize+SIZE(tmplist))
                  atom_index(isize+1:isize+SIZE(tmplist)) = tmplist
                  isize = SIZE(atom_index)
               END DO
            END IF
            my_conf(i) = isize-old_size
            IF(.NOT.(conf(i) /= 0))CALL cp__a("bsse.F",288)
         END IF
      END DO
      CALL conf_info_setup(present_charge, present_multpl, conf, conf_loc, bsse_section, &
                           dft_section)
      !
      ! Get names and modify the ghost ones
      !
      ALLOCATE (atom_type(isize))
      DO j = 1, isize
         my_targ = atom_index(j)
         DO k = 1, SIZE(particles%els)
            CALL get_atomic_kind(particles%els(k)%atomic_kind, atom_list=atom_list, name=name)
            IF (ANY(atom_list == my_targ)) EXIT
         END DO
         atom_type(j) = name
      END DO
      DO i = 1, SIZE(conf_loc)
         IF (my_conf(i) /= 0 .AND. conf_loc(i) == 0) THEN
            DO j = SUM(my_conf(1:i-1))+1, SUM(my_conf(1:i))
               atom_type(j) = TRIM(atom_type(j))//"_ghost"
            END DO
         END IF
      END DO
      CALL dump_bsse_info(atom_index, atom_type, conf, conf_loc, bsse_section, &
                          present_charge, present_multpl)
      !
      ! Let's start setting up environments and calculations
      !
      energy = 0.0_dp
      IF (method_name_id == do_qs) THEN
         CALL create_small_subsys(subsys_loc, big_subsys=subsys, &
                                  small_para_env=para_env, small_cell=cell, sub_atom_index=atom_index, &
                                  sub_atom_kind_name=atom_type, para_env=para_env, &
                                  force_env_section=force_env_section, subsys_section=subsys_section)

         CALL qs_env_create(qs_env, globenv)
         CALL qs_init(qs_env, para_env, globenv, root_section, cp_subsys=subsys_loc, &
                      force_env_section=force_env_section, subsys_section=subsys_section, &
                      use_motion_section=.FALSE.)
         CALL cp_subsys_release(subsys_loc)

         !
         ! Evaluate Energy
         !
         CALL qs_energies(qs_env)
         CALL get_qs_env(qs_env, energy=qs_energy)
         energy = qs_energy%total
         CALL qs_env_release(qs_env)
      ELSE
         CALL cp__b("bsse.F",338,"")
      END IF
      DEALLOCATE (atom_index)
      DEALLOCATE (atom_type)
      DEALLOCATE (my_conf)

   END SUBROUTINE eval_bsse_energy_low

! **************************************************************************************************
!> \brief Dumps bsse information (configuration fragment)
!> \param atom_index ...
!> \param atom_type ...
!> \param conf ...
!> \param conf_loc ...
!> \param bsse_section ...
!> \param present_charge ...
!> \param present_multpl ...
!> \par History
!>      07.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE dump_bsse_info(atom_index, atom_type, conf, conf_loc, bsse_section, &
                             present_charge, present_multpl)
      INTEGER, DIMENSION(:), POINTER                     :: atom_index
      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: atom_type
      INTEGER, DIMENSION(:), INTENT(IN)                  :: conf, conf_loc
      TYPE(section_vals_type), POINTER                   :: bsse_section
      INTEGER, INTENT(IN)                                :: present_charge, present_multpl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_bsse_info', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, istat, iw
      CHARACTER(len=20*SIZE(conf))                       :: conf_loc_s, conf_s
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, bsse_section, "PRINT%PROGRAM_RUN_INFO", &
                                extension=".log")
      IF (iw > 0) THEN
         WRITE (conf_s, fmt="(1000I0)", iostat=istat) conf; 
         IF (istat .NE. 0) conf_s = "exceeded"
         CALL compress(conf_s, full=.TRUE.)
         WRITE (conf_loc_s, fmt="(1000I0)", iostat=istat) conf_loc; 
         IF (istat .NE. 0) conf_loc_s = "exceeded"
         CALL compress(conf_loc_s, full=.TRUE.)

         WRITE (UNIT=iw, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T5,A,T30,A,T55,A,T80,A)") &
            "-", "BSSE CALCULATION", "FRAGMENT CONF: "//TRIM(conf_s), "FRAGMENT SUBCONF: "//TRIM(conf_loc_s), "-"
         WRITE (UNIT=iw, FMT="(T2,A,T30,A,I6,T55,A,I6,T80,A)") "-", "CHARGE =", present_charge, "MULTIPLICITY =", &
            present_multpl, "-"
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T20,A,T60,A,T80,A)") "-", "ATOM INDEX", "ATOM NAME", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T20,A,T60,A,T80,A)") "-", "----------", "---------", "-"
         DO i = 1, SIZE(atom_index)
            WRITE (UNIT=iw, FMT="(T2,A,T20,I6,T61,A,T80,A)") "-", atom_index(i), TRIM(atom_type(i)), "-"
         END DO
         WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)

         CALL cp_print_key_finished_output(iw, logger, bsse_section, &
                                           "PRINT%PROGRAM_RUN_INFO")

      END IF
   END SUBROUTINE dump_bsse_info

! **************************************************************************************************
!> \brief Read modified parameters for configurations
!> \param present_charge ...
!> \param present_multpl ...
!> \param conf ...
!> \param conf_loc ...
!> \param bsse_section ...
!> \param dft_section ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE conf_info_setup(present_charge, present_multpl, conf, conf_loc, &
                              bsse_section, dft_section)
      INTEGER, INTENT(OUT)                               :: present_charge, present_multpl
      INTEGER, DIMENSION(:), INTENT(IN)                  :: conf, conf_loc
      TYPE(section_vals_type), POINTER                   :: bsse_section, dft_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'conf_info_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nconf
      INTEGER, DIMENSION(:), POINTER                     :: glb_conf, sub_conf
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: configurations

      present_charge = 0
      present_multpl = 0
      NULLIFY (configurations, glb_conf, sub_conf)
      ! Loop over all configurations to pick up the right one
      configurations => section_vals_get_subs_vals(bsse_section, "CONFIGURATION")
      CALL section_vals_get(configurations, explicit=explicit, n_repetition=nconf)
      IF (explicit) THEN
         DO i = 1, nconf
            CALL section_vals_val_get(configurations, "GLB_CONF", i_rep_section=i, i_vals=glb_conf)
            IF (SIZE(glb_conf) /= SIZE(conf)) &
               CALL cp_abort(cp__l("bsse.F",442), &
                             "GLB_CONF requires a binary description of the configuration. Number of integer "// &
                             "different from the number of fragments defined!")
            CALL section_vals_val_get(configurations, "SUB_CONF", i_rep_section=i, i_vals=sub_conf)
            IF (SIZE(sub_conf) /= SIZE(conf)) &
               CALL cp_abort(cp__l("bsse.F",447), &
                             "SUB_CONF requires a binary description of the configuration. Number of integer "// &
                             "different from the number of fragments defined!")
            IF (ALL(conf == glb_conf) .AND. ALL(conf_loc == sub_conf)) THEN
               CALL section_vals_val_get(configurations, "CHARGE", i_rep_section=i, &
                                         i_val=present_charge)
               CALL section_vals_val_get(configurations, "MULTIPLICITY", i_rep_section=i, &
                                         i_val=present_multpl)
            END IF
         END DO
      END IF
      ! Setup parameter for this configuration
      CALL section_vals_val_set(dft_section, "CHARGE", i_val=present_charge)
      CALL section_vals_val_set(dft_section, "MULTIPLICITY", i_val=present_multpl)
   END SUBROUTINE conf_info_setup

! **************************************************************************************************
!> \brief Dumps results
!> \param conf ...
!> \param Em ...
!> \param num_of_frag ...
!> \param bsse_section ...
!> \par History
!>      09.2007 created [tlaino]
!> \author Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE dump_bsse_results(conf, Em, num_of_frag, bsse_section)
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: conf
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Em
      INTEGER, INTENT(IN)                                :: num_of_frag
      TYPE(section_vals_type), POINTER                   :: bsse_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_bsse_results', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iw
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, bsse_section, "PRINT%PROGRAM_RUN_INFO", &
                                extension=".log")

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T36,A,T80,A)") &
            "-", "BSSE RESULTS", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=iw, FMT="(T2,A,T20,A,F16.6,T80,A)") "-", "CP-corrected Total energy:", SUM(Em), "-"
         WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
         DO i = 1, SIZE(conf, 1)
            IF (i .GT. 1) THEN
               IF (SUM(conf(i-1, :)) == 1 .AND. SUM(conf(i, :)) /= 1) THEN
                  WRITE (UNIT=iw, FMT="(T2,A,T80,A)") "-", "-"
               END IF
            END IF
            WRITE (UNIT=iw, FMT="(T2,A,T24,I3,A,F16.6,T80,A)") "-", SUM(conf(i, :)), "-body contribution:", Em(i), "-"
         END DO
         WRITE (UNIT=iw, FMT="(T2,A,T20,A,F16.6,T80,A)") "-", "BSSE-free interaction energy:", SUM(Em(Num_of_frag+1:)), "-"
         WRITE (UNIT=iw, FMT="(T2,A)") REPEAT("-", 79)
      END IF

      CALL cp_print_key_finished_output(iw, logger, bsse_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE dump_bsse_results

! **************************************************************************************************
!> \brief generate the N-body configuration for the N-body BSSE evaluation
!> \param Num_of_frag ...
!> \param conf ...
!> \par History
!>      07.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE gen_Nbody_conf(Num_of_frag, conf)
      INTEGER, INTENT(IN)                                :: Num_of_frag
      INTEGER, DIMENSION(:, :), POINTER                  :: conf

      INTEGER                                            :: k, my_ind

      my_ind = 0
      !
      ! Set up the N-body configurations
      !
      conf = 0
      DO k = 1, Num_of_frag
         CALL build_Nbody_conf(1, Num_of_frag, conf, k, my_ind)
      END DO
   END SUBROUTINE gen_Nbody_conf

! **************************************************************************************************
!> \brief ...
!> \param ldown ...
!> \param lup ...
!> \param conf ...
!> \param k ...
!> \param my_ind ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE build_Nbody_conf(ldown, lup, conf, k, my_ind)
      INTEGER, INTENT(IN)                                :: ldown, lup
      INTEGER, DIMENSION(:, :), POINTER                  :: conf
      INTEGER, INTENT(IN)                                :: k
      INTEGER, INTENT(INOUT)                             :: my_ind

      INTEGER                                            :: i, kloc, my_ind0

      kloc = k-1
      my_ind0 = my_ind
      IF (kloc /= 0) THEN
         DO i = ldown, lup
            CALL build_Nbody_conf(i+1, lup, conf, kloc, my_ind)
            conf(my_ind0+1:my_ind, i) = 1
            my_ind0 = my_ind
         END DO
      ELSE
         DO i = ldown, lup
            my_ind = my_ind+1
            conf(my_ind, i) = 1
         END DO
      END IF
   END SUBROUTINE build_Nbody_conf

! **************************************************************************************************
!> \brief ...
!> \param num ...
!> \retval my_fact ...
! **************************************************************************************************
   RECURSIVE FUNCTION FACT(num) RESULT(my_fact)
      INTEGER, INTENT(IN)                                :: num
      INTEGER                                            :: my_fact

      IF (num <= 1) THEN
         my_fact = 1
      ELSE
         my_fact = num*FACT(num-1)
      END IF
   END FUNCTION FACT

! **************************************************************************************************
!> \brief ...
!> \param main_conf ...
!> \param conf ...
! **************************************************************************************************
   SUBROUTINE make_plan_conf(main_conf, conf)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: main_conf
      INTEGER, DIMENSION(:, :), POINTER                  :: conf

      INTEGER                                            :: i, ind
      INTEGER, DIMENSION(:, :), POINTER                  :: tmp_conf

      ALLOCATE (tmp_conf(SIZE(conf, 1), SIZE(main_conf)))
      tmp_conf = 0
      ind = 0
      DO i = 1, SIZE(main_conf)
         IF (main_conf(i) /= 0) THEN
            ind = ind+1
            tmp_conf(:, i) = conf(:, ind)
         END IF
      END DO
      DEALLOCATE (conf)
      ALLOCATE (conf(SIZE(tmp_conf, 1), SIZE(tmp_conf, 2)))
      conf = tmp_conf
      DEALLOCATE (tmp_conf)

   END SUBROUTINE make_plan_conf

! **************************************************************************************************
!> \brief Writes restart for BSSE calculations
!> \param bsse_section ...
!> \param root_section ...
!> \par History
!>      01.2008 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE write_bsse_restart(bsse_section, root_section)

      TYPE(section_vals_type), POINTER                   :: bsse_section, root_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_bsse_restart', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ires
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      ires = cp_print_key_unit_nr(logger, bsse_section, "PRINT%RESTART", &
                                  extension=".restart", do_backup=.FALSE., file_position="REWIND")

      IF (ires > 0) THEN
         CALL write_restart_header(ires)
         CALL section_vals_write(root_section, unit_nr=ires, hide_root=.TRUE.)
      ENDIF

      CALL cp_print_key_finished_output(ires, logger, bsse_section, &
                                        "PRINT%RESTART")

   END SUBROUTINE write_bsse_restart

END MODULE bsse
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods and functions on the EIP environment
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
MODULE eip_environment
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_methods,                    ONLY: read_cell,&
                                              write_cell
   USE cell_types,                      ONLY: cell_release,&
                                              cell_type,&
                                              get_cell
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_methods,               ONLY: cp_subsys_create
   USE cp_subsys_types,                 ONLY: cp_subsys_release,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_release,&
                                              distribution_1d_type
   USE distribution_methods,            ONLY: distribute_molecules_1d
   USE eip_environment_types,           ONLY: eip_env_set,&
                                              eip_environment_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE molecule_kind_types,             ONLY: molecule_kind_type,&
                                              write_molecule_kind_set
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_methods,                ONLY: write_fist_particle_coordinates,&
                                              write_particle_distances,&
                                              write_structure_data
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'eip_environment'

! *** Public subroutines ***

   PUBLIC :: eip_init

CONTAINS

! **************************************************************************************************
!> \brief Initialize the eip environment
!> \param eip_env The eip environment to retain
!> \param root_section ...
!> \param para_env ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE eip_init(eip_env, root_section, para_env, force_env_section, &
                       subsys_section)
      TYPE(eip_environment_type), POINTER                :: eip_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: use_ref_cell
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: cell_section, colvar_section, eip_section

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(eip_env)))CALL cp__a("eip_environment.F",87)

      ! nullifying  pointers
      NULLIFY (cell_section, colvar_section, eip_section, cell, cell_ref, &
               subsys)

      IF (.NOT. ASSOCIATED(subsys_section)) THEN
         subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      END IF
      cell_section => section_vals_get_subs_vals(subsys_section, "CELL")
      colvar_section => section_vals_get_subs_vals(subsys_section, "COLVAR")
      eip_section => section_vals_get_subs_vals(force_env_section, "EIP")

      CALL eip_env_set(eip_env=eip_env, eip_input=eip_section, &
                       force_env_input=force_env_section)

      CALL read_cell(cell=cell, cell_ref=cell_ref, use_ref_cell=use_ref_cell, cell_section=cell_section, &
                     para_env=para_env)
      CALL get_cell(cell=cell, abc=abc)
      CALL write_cell(cell=cell, subsys_section=subsys_section)

      CALL cp_subsys_create(subsys, para_env, root_section)

      CALL eip_init_subsys(eip_env=eip_env, subsys=subsys, cell=cell, &
                           cell_ref=cell_ref, use_ref_cell=use_ref_cell, &
                           subsys_section=subsys_section)

      CALL cell_release(cell)
      CALL cell_release(cell_ref)
      CALL cp_subsys_release(subsys)

      CALL timestop(handle)

   END SUBROUTINE eip_init

! **************************************************************************************************
!> \brief Initialize the eip environment
!> \param eip_env The eip environment of matter
!> \param subsys the subsys
!> \param cell Pointer to the actual simulation cell
!> \param cell_ref Pointer to the reference cell, used e.g. in NPT simulations
!> \param use_ref_cell Logical which indicates if cell_ref is in use
!> \param subsys_section ...
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE eip_init_subsys(eip_env, subsys, cell, cell_ref, use_ref_cell, subsys_section)
      TYPE(eip_environment_type), POINTER                :: eip_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      LOGICAL, INTENT(in)                                :: use_ref_cell
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_init_subsys', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, natom
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

!   ------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, molecule_kind_set, particle_set, molecule_set, &
               local_molecules, local_particles)

      particle_set => subsys%particles%els
      atomic_kind_set => subsys%atomic_kinds%els
      molecule_kind_set => subsys%molecule_kinds_new%els
      molecule_set => subsys%molecules_new%els

!   *** Print the molecule kind set ***
      CALL write_molecule_kind_set(molecule_kind_set, subsys_section)

!   *** Print the atomic coordinates
      CALL write_fist_particle_coordinates(particle_set, subsys_section, &
                                           charges=Null())
      CALL write_particle_distances(particle_set, cell=cell, &
                                    subsys_section=subsys_section)
      CALL write_structure_data(particle_set, cell=cell, &
                                input_section=subsys_section)

!   *** Distribute molecules and atoms using the new data structures ***
      CALL distribute_molecules_1d(atomic_kind_set=atomic_kind_set, &
                                   particle_set=particle_set, &
                                   local_particles=local_particles, &
                                   molecule_kind_set=molecule_kind_set, &
                                   molecule_set=molecule_set, &
                                   local_molecules=local_molecules, &
                                   force_env_section=eip_env%force_env_input)

      natom = SIZE(particle_set)

      ALLOCATE (eip_env%eip_forces(3, natom))

      eip_env%eip_forces(:, :) = 0.0_dp

      CALL cp_subsys_set(subsys, cell=cell)
      CALL eip_env_set(eip_env=eip_env, subsys=subsys, &
                       cell_ref=cell_ref, use_ref_cell=use_ref_cell, &
                       local_molecules=local_molecules, &
                       local_particles=local_particles)

      CALL distribution_1d_release(local_particles)
      CALL distribution_1d_release(local_molecules)

      CALL eip_init_model(eip_env=eip_env)

      CALL timestop(handle)

   END SUBROUTINE eip_init_subsys

! **************************************************************************************************
!> \brief Initialize the empirical interatomic potnetial (force field) model
!> \param eip_env The eip environment to retain
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
   SUBROUTINE eip_init_model(eip_env)
      TYPE(eip_environment_type), POINTER                :: eip_env

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_init_model', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: eip_atomic_kind_name
      INTEGER                                            :: handle, i
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_ptr
      TYPE(section_vals_type), POINTER                   :: eip_section

!   ------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, atomic_kind_ptr, eip_section)

      eip_section => section_vals_get_subs_vals(eip_env%force_env_input, &
                                                "EIP")

      atomic_kind_set => eip_env%subsys%atomic_kinds%els

      ! loop over all kinds
      DO i = 1, SIZE(atomic_kind_set)
         atomic_kind_ptr => eip_env%subsys%atomic_kinds%els(i)
         CALL get_atomic_kind(atomic_kind=atomic_kind_ptr, &
                              name=eip_atomic_kind_name)
         SELECT CASE (eip_atomic_kind_name)
         CASE ("SI", "Si")
            CALL section_vals_val_get(section_vals=eip_section, &
                                      keyword_name="EIP-Model", &
                                      i_val=eip_env%eip_model)
         CASE DEFAULT
            CALL cp_abort(cp__l("eip_environment.F",244), &
                          "EIP models for other elements"// &
                          "than Si isn't implemented yet.")
         END SELECT
      END DO

      CALL timestop(handle)

   END SUBROUTINE eip_init_model

END MODULE eip_environment
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      10.2005 split input_cp2k into smaller modules [fawzi]
!> \author teo & fawzi
! **************************************************************************************************
MODULE input_cp2k_free_energy
   USE bibliography,                    ONLY: BarducBus2008,&
                                              VandenCic2006
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              high_print_level,&
                                              low_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        do_fe_ac, do_fe_meta, do_fe_ui, do_wall_gaussian, do_wall_m, do_wall_none, do_wall_p, &
        do_wall_quadratic, do_wall_quartic, do_wall_reflective, gaussian
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_free_energy'

   PUBLIC :: create_metavar_section, &
             create_fe_section

!***
CONTAINS

! **************************************************************************************************
!> \brief creates the free energy section
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_fe_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_fe_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (subsection, keyword, print_key)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",65)
      CALL section_create(section, name="free_energy", &
                          description="Controls the calculation of free energy and free energy derivatives"// &
                          " with different possible methods", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="METHOD", &
                          description="Defines the method to use to compute free energy.", &
                          usage="METHOD (METADYN|UI|AC)", &
                          enum_c_vals=s2a("METADYN", "UI", "AC"), &
                          enum_i_vals=(/do_fe_meta, do_fe_ui, do_fe_ac/), &
                          enum_desc=s2a("Metadynamics", &
                                        "Umbrella Integration", &
                                        "Alchemical Change"), &
                          default_i_val=do_fe_meta, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_metadyn_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ui_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ac_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(print_key, "free_energy_info", &
                                       description="Controls the printing of basic and summary information during the"// &
                                       " Free Energy calculation", &
                                       print_level=low_print_level, each_iter_names=s2a("MD"), &
                                       each_iter_values=(/1/), add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_fe_section

! **************************************************************************************************
!> \brief creates the metadynamics section
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_metadyn_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_metadyn_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",119)
      CALL section_create(section, name="metadyn", &
                          description="This section sets parameters to set up a calculation of metadynamics.", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE., &
                          citations=(/VandenCic2006/))

      NULLIFY (subsection, keyword, print_key)

      CALL keyword_create(keyword, name="USE_PLUMED", &
                          description="Specify whether to use plumed as an external metadynamics driver.", &
                          usage="USE_PLUMED .FALSE./.TRUE.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PLUMED_INPUT_FILE", &
                          description="Specify the file name of the external plumed input file", &
                          usage="PLUMED_INPUT_FILE ./FILENAME", &
                          default_c_val="./plumed.dat")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIN_NT_HILLS", &
                          description="Specify the minimum MD step interval between spawning "// &
                          "two hills. If specified, it must be >= than NT_HILLS. In case MIN_DISP "// &
                          "is used, if MIN_DISP is satisfied before MIN_NT_HILLS MD steps have been "// &
                          "performed, the MD will continue without any spawning until MIN_NT_HILLS is "// &
                          "reached. The default value has the net effect of skipping this check.", &
                          usage="MIN_NT_HILLS {integer}", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NT_HILLS", &
                          description="Specify the maximum MD step interval between spawning "// &
                          "two hills. When negative, no new hills are spawned and only "// &
                          "the hills read from SPAWNED_HILLS_* are in effect. The latter"// &
                          "is useful when one wants to add a custom constant bias potential.", &
                          usage="NT_HILLS {integer}", default_i_val=30)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPERATURE", &
                          description="If a Lagrangian scheme is used the temperature for the collective "// &
                          "variables is specified. ", usage="TEMPERATURE <REAL>", &
                          default_r_val=0.0_dp, unit_str='K')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      !RG Adaptive hills
      CALL keyword_create(keyword, name="MIN_DISP", &
                          description="Minimum displacement between hills before placing a new hill.", &
                          usage="MIN_DISP <REAL>", &
                          default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OLD_HILL_NUMBER", &
                          description="Index of the last hill spawned for this walker.Needed to calculate MIN_DISP", &
                          usage="OLD_HILL_NUMBER <INT>", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="OLD_HILL_STEP", &
                          description="Timestep of the last hill spawned for this walker.Needed to calculate MIN_DISP", &
                          usage="OLD_HILL_STEP <INT>", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !RG Adaptive hills

      !Hills tail damping
      CALL keyword_create(keyword, name="HILL_TAIL_CUTOFF", &
                          description="By setting this variable larger than 0 the tail of the Gaussian hill "// &
                          " is damped to zero faster. The Gaussian function is multiplied by a cutoff function"// &
                          " that becomes active at |x-X0|>HILL_TAIL_CUTOFF*SCALE, where X0 is the location of "// &
                          "the Gaussian and SCALE is the width of the Gaussian. For more than one METAVAR"// &
                          " X0 and SCALE are METAVAR-dependent. "// &
                          " (1-(|x-X0|/HILL_TAIL_CUTOFF*SCALE)^P_EXP)/(1-(|x-X0|/HILL_TAIL_CUTOFF*SCALE)^Q_EXP)", &
                          usage="HILL_TAIL_CUTOFF <REAL>", &
                          default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="P_EXPONENT", &
                          description="Exponent at the numerator of the cutoff function to damp the tail of the Gaussian.", &
                          usage="P_EXPONENT <INT>", &
                          default_i_val=8)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="Q_EXPONENT", &
                          description="Exponent at the denominator of the cutoff function to damp the tail of the Gaussian.", &
                          usage="Q_EXPONENT <INT>", &
                          default_i_val=20)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SLOW_GROWTH", &
                          description="Let the last hill grow slowly over NT_HILLS. ", &
                          usage="SLOW_GROWTH {logical}", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMP_TOL", &
                          description="If a Lagrangian scheme is used the temperature tolerance for the collective "// &
                          "variables is specified.", usage="TEMP_TOL <REAL>", &
                          unit_str='K', default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LANGEVIN", &
                          description="If a Lagrangian scheme is used the eq. motion of the COLVARS are integrated "// &
                          "with a LANGEVIN scheme.", &
                          usage="LANGEVIN {logical}", &
                          citations=(/VandenCic2006/), &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WW", &
                          description="Specifies the height of the gaussian to spawn. Default 0.1 .", &
                          usage="WW <REAL>", unit_str='hartree', default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DO_HILLS", &
                          description="This keyword enables the spawning of the hills. Default .FALSE.", &
                          usage="DO_HILLS", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WELL_TEMPERED", &
                          description="This keyword enables Well-tempered metadynamics. Default .FALSE.", &
                          usage="WELL_TEMPERED", citations=(/BarducBus2008/), &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DELTA_T", &
                          description="If Well-tempered metaD is used, the temperature parameter "// &
                          "must be specified.", usage="DELTA_T <REAL>", &
                          unit_str='K', default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WTGAMMA", &
                          description="If Well-tempered metaD is used, the gamma parameter "// &
                          "must be specified if not DELTA_T.", usage="WTGAMMA <REAL>", &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LAGRANGE", &
                          description="Specifies whether an extended-lagrangian should be used. Default .FALSE.", &
                          usage="LAGRANGE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="step_start_val", &
                          description="The starting step value for metadynamics", &
                          usage="step_start_val <integer>", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="nhills_start_val", &
                          description="The starting value of previously spawned hills", &
                          usage="nhills_start_val <integer>", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="COLVAR_AVG_TEMPERATURE_RESTART", &
                          description="COLVAR average temperature. Only for restarting purposes.", &
                          usage="COLVAR_AVG_TEMPERATURE_RESTART 0.0", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="TAMCSteps", &
                          description="Number of sampling points for z", &
                          usage="TAMCSteps <integer>", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="timestep", &
                          description="The length of an integration step for colvars (TAMC only)", &
                          usage="timestep <real>", default_r_val=cp_unit_to_cp2k(value=0.5_dp, unit_str="fs"), &
                          unit_str="fs")

      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_metavar_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_multiple_walkers_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="print", &
                          description="Controls the printing properties during an metadynamics run", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)
      NULLIFY (print_key)

      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing of basic and summary information during"// &
                                       " metadynamics.", &
                                       print_level=low_print_level, each_iter_names=s2a("MD", "METADYNAMICS"), &
                                       each_iter_values=(/1, 1/), add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "temperature_colvar", &
                                       description="Controls the printing of the temperature of COLVARS in an "// &
                                       "extended lagrangian scheme.", &
                                       print_level=low_print_level, each_iter_names=s2a("MD", "METADYNAMICS"), &
                                       each_iter_values=(/1, 1/), add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "COLVAR", &
                                       description="Controls the printing of COLVAR summary information during"// &
                                       " metadynamics. When an extended Lagrangian use used, the files"// &
                                       " contain (in order): colvar value of the extended Lagrangian, "// &
                                       " instantaneous colvar value, force due to the harmonic term of the extended "// &
                                       " Lagrangian and the force due to the previously spawned hills,"// &
                                       " the force due to the walls, the velocities in the extended "// &
                                       " Lagrangian, the potential of the harmonic term of the"// &
                                       " Lagrangian, the potential energy of the hills, the potential"// &
                                       " energy of the walls and the temperature of the extended"// &
                                       " Lagrangian. When the extended Lagrangian is not used, all"// &
                                       " related fields are omitted.", &
                                       print_level=low_print_level, each_iter_names=s2a("MD", "METADYNAMICS"), &
                                       each_iter_values=(/1, 1/), add_last=add_last_numeric, filename="COLVAR")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "HILLS", &
                                       description="Controls the printing of HILLS summary information during"// &
                                       " metadynamics. The file contains: instantaneous colvar value, width of "// &
                                       " the spawned gaussian and height of the gaussian. According the value of "// &
                                       " the EACH keyword this file may not be synchronized with the COLVAR file.", &
                                       print_level=high_print_level, each_iter_names=s2a("MD", "METADYNAMICS"), &
                                       each_iter_values=(/1, 1/), add_last=add_last_numeric, filename="HILLS")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_metadyn_history(subsection, section)
   END SUBROUTINE create_metadyn_section

! **************************************************************************************************
!> \brief creates the multiple walker section
!> \param section the section to be created
!> \author teodoro laino [tlaino] 10.2008
! **************************************************************************************************
   SUBROUTINE create_multiple_walkers_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_multiple_walkers_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",384)
      CALL section_create(section, name="MULTIPLE_WALKERS", &
                          description="Enables and configures the metadynamics using multiple walkers.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection, keyword)
      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Controls the usage of the multiple walkers in a metadynamics run.", &
                          usage="&MULTIPLE_WALKERS T", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WALKER_ID", &
                          description="Sets the walker ID for the local metadynamics run.", &
                          usage="WALKER_ID <INTEGER>", type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUMBER_OF_WALKERS", &
                          description="Sets the total number of walkers in the metadynamic run.", &
                          usage="NUMBER_OF_WALKERS <INTEGER>", type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WALKER_COMM_FREQUENCY", &
                          description="Sets the frequency (in unit of spawned hills) for the "// &
                          "communication between the several walkers, in order to update the "// &
                          "local list of hills with the ones coming from the other walkers", &
                          usage="WALKER_COMM_FREQUENCY <INTEGER>", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WALKERS_STATUS", &
                          description="Stores the status of the several walkers in the local run.", &
                          usage="WALKERS_STATUS <INTEGER> .. <INTEGER>", type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="WALKERS_FILE_NAME", &
                          description="Specify the basename for the NUMBER_OF_WALKERS files used to "// &
                          "communicate between the walkers. Absolute path can be input as well "// &
                          "together with the filename. One file will be created for each spawned hill.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specified the communication filename for each walker.", repeats=.TRUE., &
                          usage="{String}", type_of_var=lchar_t, n_var=1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_multiple_walkers_section

! **************************************************************************************************
!> \brief creates the alchemical section for free energy evaluation
!> \param section the section to be created
!> \author teodoro laino [tlaino] 04.2007
! **************************************************************************************************
   SUBROUTINE create_ac_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ac_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",450)
      CALL section_create(section, name="ALCHEMICAL_CHANGE", &
                          description="Controls the calculation of delta free energies"// &
                          " with the alchemical change method.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="PARAMETER", &
                          description="Defines the perturbing parameter of the alchemical change tranformation", &
                          usage="PARAMETERS k", type_of_var=char_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHTING_FUNCTION", &
                          description="Specifies the weighting function (umbrella potential, part of the mixing function)", &
                          usage="WEIGHTING_FUNCTION (E1+E2-LOG(E1/E2))", type_of_var=lchar_t, &
                          n_var=1, default_lc_val="0")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_CONV", &
                          description="Set the relative tolerance for the convergence of the free energy derivative", &
                          usage="EPS_CONV <REAL>", &
                          default_r_val=1.0E-2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NEQUIL_STEPS", &
                          description="Set the number of equilibration steps, skipped to compute averages", &
                          usage="NEQUIL_STEPS <INTEGER>", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ac_section

! **************************************************************************************************
!> \brief creates the umbrella integration section
!> \param section the section to be created
!> \author teodoro laino [tlaino] 01.2007
! **************************************************************************************************
   SUBROUTINE create_ui_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ui_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",499)
      CALL section_create(section, name="umbrella_integration", &
                          description="Controls the calculation of free energy derivatives"// &
                          " with the umbrella integration method.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection)
      CALL create_uvar_conv_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_uvar_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_ui_section

! **************************************************************************************************
!> \brief Creates the velocity section
!> \param section the section to create
!> \param metadyn_section ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_metadyn_history(section, metadyn_section)
      TYPE(section_type), POINTER                        :: section, metadyn_section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_metadyn_history', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",530)
      CALL section_create(section, name="SPAWNED_HILLS_POS", &
                          description="The position of the spawned hills during metadynamics."// &
                          "Used for RESTART.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify the spawned hills", repeats=.TRUE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(metadyn_section, section)
      CALL section_release(section)

      CALL section_create(section, name="SPAWNED_HILLS_SCALE", &
                          description="The scales of the spawned hills during metadynamics."// &
                          "Used for RESTART. When a scale is zero in one or more "// &
                          "directions, the Gaussian hill is assumed to be infinitely wide "// &
                          "in those directions. The latter can be used to combine spawned "// &
                          "hills from multiple 1D metadynamics runs in one multidimensional "// &
                          "metadynamics run.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify the spawned hills", repeats=.TRUE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(metadyn_section, section)
      CALL section_release(section)

      CALL section_create(section, name="SPAWNED_HILLS_HEIGHT", &
                          description="The height of the spawned hills during metadynamics."// &
                          "Used for RESTART.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify the spawned hills", repeats=.TRUE., &
                          usage="{Real}", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(metadyn_section, section)
      CALL section_release(section)

      CALL section_create(section, name="SPAWNED_HILLS_INVDT", &
                          description="The inverse of the DELTA_T parameter used for Well-Tempered metadynamics."// &
                          "Used for RESTART.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify the spawned hills", repeats=.TRUE., &
                          usage="{Real}", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(metadyn_section, section)
      CALL section_release(section)
      !
      ! Extended Lagrangian
      !
      CALL section_create(section, name="EXT_LAGRANGE_SS0", &
                          description="Colvar position within an extended Lagrangian formalism."// &
                          "Used for RESTART.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specified the positions", repeats=.TRUE., &
                          usage="{Real}", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(metadyn_section, section)
      CALL section_release(section)

      CALL section_create(section, name="EXT_LAGRANGE_VVP", &
                          description="Colvar velocities within an extended Lagrangian formalism."// &
                          "Used for RESTART.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specified the velocities", repeats=.TRUE., &
                          usage="{Real}", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(metadyn_section, section)
      CALL section_release(section)

      CALL section_create(section, name="EXT_LAGRANGE_SS", &
                          description="Colvar Theta within an extended Lagrangian formalism."// &
                          "Used for RESTART.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specified the theta", repeats=.TRUE., &
                          usage="{Real}", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(metadyn_section, section)
      CALL section_release(section)

      CALL section_create(section, name="EXT_LAGRANGE_FS", &
                          description="Colvar force within an extended Lagrangian formalism."// &
                          "Used for RESTART.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specified the theta", repeats=.TRUE., &
                          usage="{Real}", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(metadyn_section, section)
      CALL section_release(section)

   END SUBROUTINE create_metadyn_history

! **************************************************************************************************
!> \brief creates the metavar section
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_metavar_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_metavar_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection, wall_section

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",657)
      CALL section_create(section, name="METAVAR", &
                          description="This section specify the nature of the collective variables.", &
                          n_keywords=1, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="LAMBDA", &
                          description="Specifies the lambda parameter of the collective variable in the"// &
                          " extended lagrangian scheme.", &
                          usage="LAMBDA <REAL>", unit_str='internal_cp2k', type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MASS", &
                          description="Specifies the mass parameter of the collective variable in the"// &
                          " extended lagrangian scheme.", usage="MASS <REAL>", unit_str='amu', type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GAMMA", &
                          description="Specifies the friction term in Langevin integration of the collective variable in the"// &
                          " extended lagrangian scheme.", &
                          citations=(/VandenCic2006/), &
                          usage="GAMMA {real}", type_of_var=real_t, unit_str="fs^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SCALE", &
                          variants=(/"WIDTH"/), &
                          description="Specifies the scale factor for the following collective variable. The history "// &
                          "dependent term has the expression: WW * Sum_{j=1}^{nhills} Prod_{k=1}^{ncolvar} "// &
                          "[EXP[-0.5*((ss-ss0(k,j))/SCALE(k))^2]], "// &
                          "where ncolvar is the number of defined METAVAR and nhills is the number of spawned hills. ", &
                          usage="SCALE <REAL>", type_of_var=real_t, unit_str='internal_cp2k')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COLVAR", &
                          description="Specifies the colvar on which to apply metadynamics.", &
                          usage="COLVAR {integer}", type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Wall section
      NULLIFY (wall_section, subsection)
      CALL section_create(wall_section, name="WALL", &
                          description="Controls the activation of walls on COLVAR during a metadynamic run.", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)

      CALL keyword_create( &
         keyword, name="TYPE", &
         description="Specify the type of wall", &
         usage=" TYPE (REFLECTIVE|QUADRATIC|QUARTIC|GAUSSIAN|NONE)", &
         enum_c_vals=s2a("REFLECTIVE", "QUADRATIC", "QUARTIC", "GAUSSIAN", "NONE"), &
         enum_desc=s2a("Reflective wall. Colvar velocity is inverted when the colvar is beyond the wall position.", &
                       "Applies a quadratic potential at the wall position.", &
                       "Applies a quartic potential at the wall position.", &
                       "Applies a gaussian potential at the wall position.", &
                       "No walls are applied."), &
         enum_i_vals=(/do_wall_reflective, do_wall_quadratic, do_wall_quartic, do_wall_gaussian, do_wall_none/), &
         default_i_val=do_wall_none)
      CALL section_add_keyword(wall_section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POSITION", &
                          description="Specify the value of the colvar for the wall position", &
                          usage="POSITION <REAL>", unit_str='internal_cp2k', &
                          type_of_var=real_t)
      CALL section_add_keyword(wall_section, keyword)
      CALL keyword_release(keyword)

      ! Reflective wall
      CALL section_create(subsection, name="REFLECTIVE", &
                          description="Parameters controlling the reflective wall", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="DIRECTION", &
                          description="Specify the direction of the wall.", &
                          usage=" TYPE (WALL_PLUS|WALL_MINUS)", &
                          enum_c_vals=s2a("WALL_PLUS", "WALL_MINUS"), &
                          enum_desc=s2a("Wall extends from the position towards larger values of COLVAR", &
                                        "Wall extends from the position towards smaller values of COLVAR"), &
                          enum_i_vals=(/do_wall_p, do_wall_m/), default_i_val=do_wall_p)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(wall_section, subsection)
      CALL section_release(subsection)

      ! Quadratic wall
      CALL section_create(subsection, name="QUADRATIC", &
                          description="Parameters controlling the quadratic wall", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="DIRECTION", &
                          description="Specify the direction of the wall.", &
                          usage=" TYPE (WALL_PLUS|WALL_MINUS)", &
                          enum_c_vals=s2a("WALL_PLUS", "WALL_MINUS"), &
                          enum_desc=s2a("Wall extends from the position towards larger values of COLVAR", &
                                        "Wall extends from the position towards smaller values of COLVAR"), &
                          enum_i_vals=(/do_wall_p, do_wall_m/), default_i_val=do_wall_p)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Specify the value of the quadratic potential constant: K*(CV-POS)^2", &
                          usage="K <REAL>", unit_str='hartree', &
                          type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(wall_section, subsection)
      CALL section_release(subsection)

      ! Quartic wall
      CALL section_create(subsection, name="QUARTIC", &
                          description="Parameters controlling the quartic wall", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="DIRECTION", &
                          description="Specify the direction of the wall.", &
                          usage=" TYPE (WALL_PLUS|WALL_MINUS)", &
                          enum_c_vals=s2a("WALL_PLUS", "WALL_MINUS"), &
                          enum_desc=s2a("Wall extends from the position towards larger values of COLVAR", &
                                        "Wall extends from the position towards smaller values of COLVAR"), &
                          enum_i_vals=(/do_wall_p, do_wall_m/), default_i_val=do_wall_p)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Specify the value of the quartic potential constant: K*(CV-(POS+/-(1/K^(1/4))))^4", &
                          usage="K <REAL>", unit_str='hartree', &
                          type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(wall_section, subsection)
      CALL section_release(subsection)

      ! Gaussian wall
      CALL section_create(subsection, name="GAUSSIAN", &
                          description="Parameters controlling the gaussian wall.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL keyword_create(keyword, name="WW", &
                          description="Specify the height of the gaussian: WW*e^(-((CV-POS)/sigma)^2)", &
                          usage="K <REAL>", unit_str='hartree', &
                          type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIGMA", &
                          description="Specify the width of the gaussian: WW*e^(-((CV-POS)/sigma)^2)", &
                          usage="SIGMA <REAL>", unit_str='internal_cp2k', &
                          type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(wall_section, subsection)
      CALL section_release(subsection)

      CALL section_add_subsection(section, wall_section)
      CALL section_release(wall_section)

   END SUBROUTINE create_metavar_section

! **************************************************************************************************
!> \brief creates the uvar section
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_uvar_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_uvar_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",835)
      CALL section_create(section, name="UVAR", &
                          description="This section specify the nature of the collective variables"// &
                          " used in computing the free energy.", &
                          n_keywords=1, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="COLVAR", &
                          description="Specifies the colvar used to compute free energy", &
                          usage="COLVAR {integer}", type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_uvar_section

! **************************************************************************************************
!> \brief  creates the section specifying parameters to control the convergence
!>         of the free energy
!> \param section the section to be created
!> \author teodoro laino [tlaino] 01.2007
! **************************************************************************************************
   SUBROUTINE create_uvar_conv_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_uvar_conv_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_free_energy.F",864)
      CALL section_create(section, name="CONVERGENCE_CONTROL", &
                          description="This section specify parameters controlling the convergence"// &
                          " of the free energy.", &
                          n_keywords=1, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="COARSE_GRAINED_WIDTH", &
                          variants=(/"CG_WIDTH"/), &
                          description="Width of segments in MD steps to generate the set of"// &
                          " coarse grained data, providing a correlation independent data set.", &
                          usage="COARSE_GRAINED_WIDTH <INTEGER>", default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_COARSE_GRAINED_WIDTH", &
                          variants=(/"MAX_CG_WIDTH"/), &
                          description="Max Width of segments in MD steps to generate the set of"// &
                          " coarse grained data.", &
                          usage="MAX_COARSE_GRAINED_WIDTH <INTEGER>", default_i_val=200)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COARSE_GRAINED_POINTS", &
                          variants=(/"CG_POINTS"/), &
                          description="Set the minimum amount of coarse grained points to collect"// &
                          " before starting the statistical analysis", &
                          usage="COARSE_GRAINED_POINTS <INTEGER>", default_i_val=30)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_CONV", &
                          description="Set the relative tolerance for the convergence of the collective"// &
                          " variable averages used to compute the free energy.", &
                          usage="EPS_CONV <REAL>", &
                          default_r_val=1.0E-2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K_CONFIDENCE_LIMIT", &
                          description="Set the confidence limit for the Mann-Kendall trend test.", &
                          usage="K_CONFIDENCE_LIMIT <REAL>", &
                          default_r_val=0.90_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SW_CONFIDENCE_LIMIT", &
                          description="Set the confidence limit for the Shapiro-Wilks normality test.", &
                          usage="SW_CONFIDENCE_LIMIT <REAL>", &
                          default_r_val=0.90_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VN_CONFIDENCE_LIMIT", &
                          description="Set the confidence limit for the Von Neumann serial correlation test.", &
                          usage="VN_CONFIDENCE_LIMIT <REAL>", &
                          default_r_val=0.90_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_uvar_conv_section

END MODULE input_cp2k_free_energy
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Performs the metadynamics calculation
!> \par History
!>      01.2005 created [fawzi and ale]
!>      11.2007 Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
MODULE metadynamics_utils
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_constants,                 ONLY: do_fe_meta,&
                                              do_wall_gaussian,&
                                              do_wall_m,&
                                              do_wall_none,&
                                              do_wall_p,&
                                              do_wall_quadratic,&
                                              do_wall_quartic,&
                                              do_wall_reflective
   USE input_cp2k_free_energy,          ONLY: create_metavar_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_get_subsection,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE machine,                         ONLY: m_mov
   USE message_passing,                 ONLY: mp_bcast
   USE metadynamics_types,              ONLY: hills_env_type,&
                                              meta_env_type,&
                                              metadyn_create,&
                                              metavar_type,&
                                              multiple_walkers_type
   USE physcon,                         ONLY: kelvin
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'metadynamics_utils'

   PUBLIC :: metadyn_read, &
             synchronize_multiple_walkers, &
             add_hill_single, &
             restart_hills, &
             get_meta_iter_level, &
             meta_walls

CONTAINS

! **************************************************************************************************
!> \brief reads metadynamics section
!> \param meta_env ...
!> \param force_env ...
!> \param root_section ...
!> \param para_env ...
!> \param fe_section ...
!> \par History
!>      04.2004 created
!> \author Teodoro Laino [tlaino] - University of Zurich. 11.2007
! **************************************************************************************************
   SUBROUTINE metadyn_read(meta_env, force_env, root_section, para_env, fe_section)
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), OPTIONAL, POINTER         :: fe_section

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_read', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: walkers_file_name
      INTEGER                                            :: handle, i, id_method, n_colvar, n_rep, &
                                                            number_allocated_colvars
      INTEGER, DIMENSION(:), POINTER                     :: walkers_status
      LOGICAL                                            :: check, explicit
      REAL(kind=dp)                                      :: dt
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: md_section, metadyn_section, &
                                                            metavar_section, walkers_section

      NULLIFY (subsys)
      CALL timeset(routineN, handle)

      CALL section_vals_get(fe_section, explicit=explicit)
      IF (explicit) THEN
         number_allocated_colvars = 0
         CALL force_env_get(force_env, subsys=subsys)
         IF (ASSOCIATED(subsys%colvar_p)) THEN
            number_allocated_colvars = SIZE(subsys%colvar_p)
         END IF
         CALL section_vals_val_get(fe_section, "METHOD", i_val=id_method)
         IF (id_method /= do_fe_meta) THEN
            CALL timestop(handle)
            RETURN
         ENDIF
         metadyn_section => section_vals_get_subs_vals(fe_section, "METADYN")
         IF(.NOT.(.NOT. ASSOCIATED(meta_env)))CALL cp__a("metadynamics_utils.F",117)

         md_section => section_vals_get_subs_vals(root_section, "MOTION%MD")
         CALL section_vals_val_get(md_section, "TIMESTEP", r_val=dt)

         metavar_section => section_vals_get_subs_vals(metadyn_section, "METAVAR")
         CALL section_vals_get(metavar_section, n_repetition=n_colvar)
         CALL metadyn_create(meta_env, n_colvar=n_colvar, &
                             dt=dt, para_env=para_env, metadyn_section=metadyn_section)

         !Check if using plumed. If so, only get the file name and read nothing else
         CALL section_vals_val_get(metadyn_section, "USE_PLUMED", l_val=meta_env%use_plumed)
         IF (meta_env%use_plumed .EQV. .TRUE.) THEN
            CALL section_vals_val_get(metadyn_section, "PLUMED_INPUT_FILE", c_val=meta_env%plumed_input_file)
            meta_env%plumed_input_file = TRIM(meta_env%plumed_input_file)//CHAR(0)
            meta_env%langevin = .FALSE.
            CALL timestop(handle)
            RETURN
         END IF

         CALL section_vals_val_get(metadyn_section, "DO_HILLS", l_val=meta_env%do_hills)
         CALL section_vals_val_get(metadyn_section, "LAGRANGE", l_val=meta_env%extended_lagrange)
         CALL section_vals_val_get(metadyn_section, "TAMCSteps", i_val=meta_env%TAMCSteps)
         IF (meta_env%TAMCSteps < 0) THEN
            CALL cp__b("metadynamics_utils.F",141,"TAMCSteps must be positive!")
         ENDIF
         CALL section_vals_val_get(metadyn_section, "Timestep", r_val=meta_env%zdt)
         IF (meta_env%zdt <= 0.0_dp) THEN
            CALL cp__b("metadynamics_utils.F",145,"Timestep must be positive!")
         ENDIF
         CALL section_vals_val_get(metadyn_section, "WW", r_val=meta_env%hills_env%ww)
         CALL section_vals_val_get(metadyn_section, "NT_HILLS", i_val=meta_env%hills_env%nt_hills)
         CALL section_vals_val_get(metadyn_section, "MIN_NT_HILLS", i_val=meta_env%hills_env%min_nt_hills)
         IF (meta_env%hills_env%nt_hills <= 0) THEN
            meta_env%hills_env%min_nt_hills = meta_env%hills_env%nt_hills
            CALL cp_warn(cp__l("metadynamics_utils.F",152), &
                         "NT_HILLS has a value <= 0; "// &
                         "Setting MIN_NT_HILLS to the same value! "// &
                         "Overriding input specification!")
         END IF
         check = meta_env%hills_env%nt_hills >= meta_env%hills_env%min_nt_hills
         IF (.NOT. check) &
            CALL cp_abort(cp__l("metadynamics_utils.F",159), "MIN_NT_HILLS must have a value smaller or equal to NT_HILLS! "// &
                          "Cross check with the input reference!")
         !RG Adaptive hills
         CALL section_vals_val_get(metadyn_section, "MIN_DISP", r_val=meta_env%hills_env%min_disp)
         CALL section_vals_val_get(metadyn_section, "OLD_HILL_NUMBER", i_val=meta_env%hills_env%old_hill_number)
         CALL section_vals_val_get(metadyn_section, "OLD_HILL_STEP", i_val=meta_env%hills_env%old_hill_step)

         !Hills tail damping
         CALL section_vals_val_get(metadyn_section, "HILL_TAIL_CUTOFF", r_val=meta_env%hills_env%tail_cutoff)
         CALL section_vals_val_get(metadyn_section, "P_EXPONENT", i_val=meta_env%hills_env%p_exp)
         CALL section_vals_val_get(metadyn_section, "Q_EXPONENT", i_val=meta_env%hills_env%q_exp)

         CALL section_vals_val_get(metadyn_section, "SLOW_GROWTH", l_val=meta_env%hills_env%slow_growth)

         !RG Adaptive hills
         CALL section_vals_val_get(metadyn_section, "STEP_START_VAL", i_val=meta_env%n_steps)
         IF(.NOT.(meta_env%n_steps >= 0))CALL cp__a("metadynamics_utils.F",175)
         CALL section_vals_val_get(metadyn_section, "NHILLS_START_VAL", &
                                   i_val=meta_env%hills_env%n_hills)
         CALL section_vals_val_get(metadyn_section, "TEMPERATURE", r_val=meta_env%temp_wanted)
         CALL section_vals_val_get(metadyn_section, "LANGEVIN", l_val=meta_env%langevin)
         CALL section_vals_val_get(metadyn_section, "TEMP_TOL", explicit=meta_env%tempcontrol, &
                                   r_val=meta_env%toll_temp)
         CALL section_vals_val_get(metadyn_section, "WELL_TEMPERED", l_val=meta_env%well_tempered)
         CALL section_vals_val_get(metadyn_section, "DELTA_T", explicit=meta_env%hills_env%wtcontrol, &
                                   r_val=meta_env%delta_t)
         CALL section_vals_val_get(metadyn_section, "WTGAMMA", explicit=check, &
                                   r_val=meta_env%wtgamma)
         IF (meta_env%well_tempered) THEN
            meta_env%hills_env%wtcontrol = meta_env%hills_env%wtcontrol .OR. check
            check = meta_env%hills_env%wtcontrol
            IF (.NOT. check) &
               CALL cp_abort(cp__l("metadynamics_utils.F",191), "When using Well-Tempered metadynamics, "// &
                             "DELTA_T (or WTGAMMA) should be explicitly specified.")
            IF (meta_env%extended_lagrange) &
               CALL cp_abort(cp__l("metadynamics_utils.F",194), &
                             "Well-Tempered metadynamics not possible with extended-lagrangian formulation.")
            IF (meta_env%hills_env%min_disp > 0.0_dp) &
               CALL cp_abort(cp__l("metadynamics_utils.F",197), &
                             "Well-Tempered metadynamics not possible with Adaptive hills.")
         ENDIF

         CALL section_vals_val_get(metadyn_section, "COLVAR_AVG_TEMPERATURE_RESTART", &
                                   r_val=meta_env%avg_temp)
         ! Parsing Metavar Section
         DO i = 1, n_colvar
            CALL metavar_read(meta_env%metavar(i), meta_env%extended_lagrange, &
                              meta_env%langevin, i, metavar_section)
            check = (meta_env%metavar(i)%icolvar <= number_allocated_colvars)
            IF (.NOT. check) &
               CALL cp_abort(cp__l("metadynamics_utils.F",209), &
                             "An error occured in the specification of COLVAR for METAVAR. "// &
                             "Specified COLVAR #("//TRIM(ADJUSTL(cp_to_string(meta_env%metavar(i)%icolvar)))//") "// &
                             "is larger than the maximum number of COLVARS defined in the SUBSYS ("// &
                             TRIM(ADJUSTL(cp_to_string(number_allocated_colvars)))//") !")
         END DO

         ! Parsing the Multiple Walkers Info
         IF (meta_env%do_multiple_walkers) THEN
            NULLIFY (walkers_status)
            walkers_section => section_vals_get_subs_vals(metadyn_section, "MULTIPLE_WALKERS")

            ! General setup for walkers
            CALL section_vals_val_get(walkers_section, "WALKER_ID", &
                                      i_val=meta_env%multiple_walkers%walker_id)
            CALL section_vals_val_get(walkers_section, "NUMBER_OF_WALKERS", &
                                      i_val=meta_env%multiple_walkers%walkers_tot_nr)
            CALL section_vals_val_get(walkers_section, "WALKER_COMM_FREQUENCY", &
                                      i_val=meta_env%multiple_walkers%walkers_freq_comm)

            ! Handle status and file names
            ALLOCATE (meta_env%multiple_walkers%walkers_status(meta_env%multiple_walkers%walkers_tot_nr))
            ALLOCATE (meta_env%multiple_walkers%walkers_file_name(meta_env%multiple_walkers%walkers_tot_nr))
            CALL section_vals_val_get(walkers_section, "WALKERS_STATUS", explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(walkers_section, "WALKERS_STATUS", i_vals=walkers_status)
               check = (SIZE(walkers_status) == meta_env%multiple_walkers%walkers_tot_nr)
               IF (.NOT. check) &
                  CALL cp_abort(cp__l("metadynamics_utils.F",237), &
                                "Number of Walkers specified in the input does not match with the "// &
                                "size of the WALKERS_STATUS. Please check your input and in case "// &
                                "this is a restart run consider the possibility to switch off the "// &
                                "RESTART_WALKERS in the EXT_RESTART section! ")
               meta_env%multiple_walkers%walkers_status = walkers_status
            ELSE
               meta_env%multiple_walkers%walkers_status = 0
            END IF
            meta_env%multiple_walkers%n_hills_local = &
               meta_env%multiple_walkers%walkers_status(meta_env%multiple_walkers%walker_id)

            CALL section_vals_val_get(walkers_section, "WALKERS_FILE_NAME%_DEFAULT_KEYWORD_", &
                                      n_rep_val=n_rep)
            check = (n_rep == meta_env%multiple_walkers%walkers_tot_nr)
            IF (.NOT. check) &
               CALL cp_abort(cp__l("metadynamics_utils.F",253), &
                             "Number of Walkers specified in the input does not match with the "// &
                             "number of Walkers File names provided. Please check your input and in case "// &
                             "this is a restart run consider the possibility to switch off the "// &
                             "RESTART_WALKERS in the EXT_RESTART section! ")
            DO i = 1, n_rep
               CALL section_vals_val_get(walkers_section, "WALKERS_FILE_NAME%_DEFAULT_KEYWORD_", &
                                         i_rep_val=i, c_val=walkers_file_name)
               meta_env%multiple_walkers%walkers_file_name(i) = walkers_file_name
            END DO
         END IF

         ! Print Metadynamics Info
         CALL print_metadyn_info(meta_env, n_colvar, metadyn_section)
      END IF

      CALL timestop(handle)

   END SUBROUTINE metadyn_read

! **************************************************************************************************
!> \brief prints information on the metadynamics run
!> \param meta_env ...
!> \param n_colvar ...
!> \param metadyn_section ...
!> \author Teodoro Laino [tlaino] - University of Zurich. 10.2008
! **************************************************************************************************
   SUBROUTINE print_metadyn_info(meta_env, n_colvar, metadyn_section)
      TYPE(meta_env_type), POINTER                       :: meta_env
      INTEGER, INTENT(IN)                                :: n_colvar
      TYPE(section_vals_type), POINTER                   :: metadyn_section

      CHARACTER(len=*), PARAMETER :: routineN = 'print_metadyn_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=10)                                  :: my_id, my_tag
      INTEGER                                            :: handle, i, iw, j
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section, wall_section, work_section

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, metadyn_section, &
                                "PRINT%PROGRAM_RUN_INFO", extension=".metadynLog")
      NULLIFY (section, enum, keyword)
      CALL create_metavar_section(section)
      wall_section => section_get_subsection(section, "WALL")
      IF (iw > 0) THEN
         WRITE (iw, '( /A )') ' METADYN| Meta Dynamics Protocol '
         WRITE (iw, '( A,T71,I10)') ' METADYN| Number of interval time steps to spawn hills', &
            meta_env%hills_env%nt_hills
         WRITE (iw, '( A,T71,I10)') ' METADYN| Number of previously spawned hills', &
            meta_env%hills_env%n_hills
         IF (meta_env%extended_lagrange) THEN
            WRITE (iw, '( A )') ' METADYN| Extended Lagrangian Scheme '
            IF (meta_env%tempcontrol) WRITE (iw, '( A,T71,F10.2)') &
               ' METADYN| Collective Variables Temperature control', meta_env%toll_temp
            IF (meta_env%langevin) THEN
               WRITE (iw, '(A,T71)') ' METADYN| Langevin Thermostat in use for COLVAR '
               WRITE (iw, '(A,T71,F10.4)') ' METADYN| Langevin Thermostat. Target Temperature = ', &
                  meta_env%temp_wanted*kelvin
            ENDIF
            WRITE (iw, '(A,T71,F10.4)') ' METADYN| COLVARS restarted average temperature ', &
               meta_env%avg_temp
         END IF
         IF (meta_env%do_hills) THEN
            WRITE (iw, '( A )') ' METADYN| Spawning the Hills '
            WRITE (iw, '( A,T71,F10.3)') ' METADYN| Height of the Spawned Gaussian', meta_env%hills_env%ww
            !RG Adaptive hills
            IF (meta_env%hills_env%min_disp .GT. 0.0_dp) THEN
               WRITE (iw, '(A)') ' METADYN| Adapative meta time step is activated'
               WRITE (iw, '(A,T71,F10.4)') ' METADYN| Minimum displacement for next hill', &
                  meta_env%hills_env%min_disp
            END IF
            !RG Adaptive hills
         END IF

         IF (meta_env%well_tempered) THEN
            WRITE (iw, '( A )') ' METADYN| Well-Tempered metadynamics '
            IF (meta_env%delta_t > EPSILON(1._dp)) THEN
               WRITE (iw, '( A,T71,F10.3)') ' METADYN| Temperature parameter (Delta T) [K]', meta_env%delta_t*kelvin
            ELSE
               WRITE (iw, '( A,T71,F10.3)') ' METADYN| Temperature parameter (gamma)', meta_env%wtgamma
            END IF
         END IF

         IF (meta_env%do_multiple_walkers) THEN
            WRITE (iw, '( A,T71,A10)') ' METADYN| Multiple Walkers', '   ENABLED'
            WRITE (iw, '( A,T71,I10)') ' METADYN| Number of Multiple Walkers', &
               meta_env%multiple_walkers%walkers_tot_nr
            WRITE (iw, '( A,T71,I10)') ' METADYN| Local Walker ID', &
               meta_env%multiple_walkers%walker_id
            WRITE (iw, '( A,T71,I10)') ' METADYN| Walker Communication Frequency', &
               meta_env%multiple_walkers%walkers_freq_comm
            DO i = 1, meta_env%multiple_walkers%walkers_tot_nr
               my_tag = ""
               IF (i == meta_env%multiple_walkers%walker_id) my_tag = " ( Local )"
               my_id = '( '//TRIM(ADJUSTL(cp_to_string(i)))//' )'
               WRITE (iw, '(/,A,T71,A10)') ' WALKERS| Walker ID'//TRIM(my_tag), ADJUSTR(my_id)
               WRITE (iw, '(  A,T71,I10)') ' WALKERS| Number of Hills communicated', &
                  meta_env%multiple_walkers%walkers_status(i)
               WRITE (iw, '(  A,T24,A57)') ' WALKERS| Base Filename', &
                  ADJUSTR(meta_env%multiple_walkers%walkers_file_name(i) (1:57))
            END DO
            WRITE (iw, '(/)')
         END IF

         WRITE (iw, '( A,T71,I10)') ' METADYN| Number of collective variables', meta_env%n_colvar
         DO i = 1, n_colvar
            WRITE (iw, '( A )') '          '//'----------------------------------------------------------------------'
            WRITE (iw, '( A,T71,I10)') ' METAVARS| Collective Variable Number', meta_env%metavar(i)%icolvar
            IF (meta_env%extended_lagrange) THEN
               WRITE (iw, '( A,T71,F10.6)') ' METAVARS| Lambda Parameter', meta_env%metavar(i)%lambda
               WRITE (iw, '( A,T66,F15.6)') ' METAVARS| Collective Variable Mass', meta_env%metavar(i)%mass
            END IF
            WRITE (iw, '( A,T71,F10.6)') ' METAVARS| Scaling factor', meta_env%metavar(i)%delta_s
            IF (meta_env%langevin) WRITE (iw, '( A,T71,F10.6)') ' METAVARS| Friction for Langevin Thermostat', &
               meta_env%metavar(i)%gamma
            IF (meta_env%metavar(i)%do_wall) THEN
               WRITE (iw, '( A,T71,I10)') ' METAVARS| Number of Walls present', SIZE(meta_env%metavar(i)%walls)
               DO j = 1, SIZE(meta_env%metavar(i)%walls)
                  keyword => section_get_keyword(wall_section, "TYPE")
                  CALL keyword_get(keyword, enum=enum)
                  WRITE (iw, '(/,A,5X,I10,T50,A,T70,A11)') ' METAVARS| Wall Number:', j, 'Type of Wall:', &
                     ADJUSTR(TRIM(enum_i2c(enum, meta_env%metavar(i)%walls(j)%id_type)))
                  ! Type of wall IO
                  SELECT CASE (meta_env%metavar (i)%walls (j)%id_type)
                  CASE (do_wall_none)
                     ! Do Nothing
                     CYCLE
                  CASE (do_wall_reflective)
                     work_section => section_get_subsection(wall_section, "REFLECTIVE")
                     keyword => section_get_keyword(work_section, "DIRECTION")
                     CALL keyword_get(keyword, enum=enum)
                     WRITE (iw, '(A,T70,A11)') ' METAVARS| Wall direction', &
                        ADJUSTR(TRIM(enum_i2c(enum, meta_env%metavar(i)%walls(j)%id_direction)))
                  CASE (do_wall_quadratic)
                     work_section => section_get_subsection(wall_section, "QUADRATIC")
                     keyword => section_get_keyword(work_section, "DIRECTION")
                     CALL keyword_get(keyword, enum=enum)
                     WRITE (iw, '(A,T70,A11)') ' METAVARS| Wall direction', &
                        ADJUSTR(TRIM(enum_i2c(enum, meta_env%metavar(i)%walls(j)%id_direction)))
                     WRITE (iw, '(A,T70,F11.6)') ' METAVARS| Constant K of the quadratic potential', &
                        meta_env%metavar(i)%walls(j)%k_quadratic
                  CASE (do_wall_gaussian)
                     WRITE (iw, '(A,T70,F11.6)') ' METAVARS| Height of the Wall Gaussian', &
                        meta_env%metavar(i)%walls(j)%ww_gauss
                     WRITE (iw, '(A,T70,F11.6)') ' METAVARS| Scale of the Wall Gaussian', &
                        meta_env%metavar(i)%walls(j)%sigma_gauss
                  END SELECT
                  WRITE (iw, '(A,T70,F11.6)') ' METAVARS| Wall location', &
                     meta_env%metavar(i)%walls(j)%pos
               END DO
            END IF
            WRITE (iw, '( A )') '          '//'----------------------------------------------------------------------'
         ENDDO
      END IF
      CALL section_release(section)
      CALL cp_print_key_finished_output(iw, logger, metadyn_section, "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE print_metadyn_info

! **************************************************************************************************
!> \brief reads metavar section
!> \param metavar ...
!> \param extended_lagrange ...
!> \param langevin ...
!> \param icol ...
!> \param metavar_section ...
!> \par History
!>      04.2004 created
!> \author alessandro laio and fawzi mohamed
!>      Teodoro Laino [tlaino] - University of Zurich. 11.2007
! **************************************************************************************************
   SUBROUTINE metavar_read(metavar, extended_lagrange, langevin, icol, metavar_section)
      TYPE(metavar_type), INTENT(INOUT)                  :: metavar
      LOGICAL, INTENT(IN)                                :: extended_lagrange, langevin
      INTEGER, INTENT(IN)                                :: icol
      TYPE(section_vals_type), OPTIONAL, POINTER         :: metavar_section

      CHARACTER(len=*), PARAMETER :: routineN = 'metavar_read', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, n_walls
      TYPE(section_vals_type), POINTER                   :: wall_section, work_section

      CALL timeset(routineN, handle)

      CALL section_vals_val_get(metavar_section, "COLVAR", i_rep_section=icol, i_val=metavar%icolvar)
      CALL section_vals_val_get(metavar_section, "SCALE", i_rep_section=icol, r_val=metavar%delta_s)
      ! Walls
      wall_section => section_vals_get_subs_vals(metavar_section, "WALL", i_rep_section=icol)
      CALL section_vals_get(wall_section, n_repetition=n_walls)
      IF (n_walls /= 0) THEN
         metavar%do_wall = .TRUE.
         ALLOCATE (metavar%walls(n_walls))
         DO i = 1, n_walls
            CALL section_vals_val_get(wall_section, "TYPE", i_rep_section=i, i_val=metavar%walls(i)%id_type)
            CALL section_vals_val_get(wall_section, "POSITION", i_rep_section=i, r_val=metavar%walls(i)%pos)
            SELECT CASE (metavar%walls (i)%id_type)
            CASE (do_wall_none)
               ! Just cycle..
               CYCLE
            CASE (do_wall_reflective)
               work_section => section_vals_get_subs_vals(wall_section, "REFLECTIVE", i_rep_section=i)
               CALL section_vals_val_get(work_section, "DIRECTION", i_val=metavar%walls(i)%id_direction)
            CASE (do_wall_quadratic)
               work_section => section_vals_get_subs_vals(wall_section, "QUADRATIC", i_rep_section=i)
               CALL section_vals_val_get(work_section, "DIRECTION", i_val=metavar%walls(i)%id_direction)
               CALL section_vals_val_get(work_section, "K", r_val=metavar%walls(i)%k_quadratic)
            CASE (do_wall_quartic)
               work_section => section_vals_get_subs_vals(wall_section, "QUARTIC", i_rep_section=i)
               CALL section_vals_val_get(work_section, "DIRECTION", i_val=metavar%walls(i)%id_direction)
               CALL section_vals_val_get(work_section, "K", r_val=metavar%walls(i)%k_quartic)
               SELECT CASE (metavar%walls (i)%id_direction)
               CASE (do_wall_m)
                  metavar%walls(i)%pos0 = metavar%walls(i)%pos+(0.05_dp/metavar%walls(i)%k_quartic**(1/4))
               CASE (do_wall_p)
                  metavar%walls(i)%pos0 = metavar%walls(i)%pos-(0.05_dp/metavar%walls(i)%k_quartic**(1/4))
               END SELECT
            CASE (do_wall_gaussian)
               work_section => section_vals_get_subs_vals(wall_section, "GAUSSIAN", i_rep_section=i)
               CALL section_vals_val_get(work_section, "WW", r_val=metavar%walls(i)%ww_gauss)
               CALL section_vals_val_get(work_section, "SIGMA", r_val=metavar%walls(i)%sigma_gauss)
            END SELECT
         END DO
      END IF
      ! Setup few more parameters for extended lagrangian
      IF (extended_lagrange) THEN
         CALL section_vals_val_get(metavar_section, "MASS", i_rep_section=icol, r_val=metavar%mass)
         CALL section_vals_val_get(metavar_section, "LAMBDA", i_rep_section=icol, r_val=metavar%lambda)
         IF (langevin) THEN
            CALL section_vals_val_get(metavar_section, "GAMMA", i_rep_section=icol, r_val=metavar%gamma)
         END IF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE metavar_read

! **************************************************************************************************
!> \brief  Synchronize with the rest of the walkers
!> \param multiple_walkers ...
!> \param hills_env ...
!> \param colvars ...
!> \param n_colvar ...
!> \param metadyn_section ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
! **************************************************************************************************
   SUBROUTINE synchronize_multiple_walkers(multiple_walkers, hills_env, colvars, &
                                           n_colvar, metadyn_section)
      TYPE(multiple_walkers_type), POINTER               :: multiple_walkers
      TYPE(hills_env_type), POINTER                      :: hills_env
      TYPE(metavar_type), DIMENSION(:), POINTER          :: colvars
      INTEGER, INTENT(IN)                                :: n_colvar
      TYPE(section_vals_type), POINTER                   :: metadyn_section

      CHARACTER(len=*), PARAMETER :: routineN = 'synchronize_multiple_walkers', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename, tmpname
      INTEGER                                            :: delta_hills, handle, i, i_hills, ih, iw, &
                                                            unit_nr
      LOGICAL                                            :: exist
      REAL(KIND=dp)                                      :: invdt, ww
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: delta_s_save, ss0_save
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: delta_s_ss0_buf
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      para_env => logger%para_env

      ! Locally dump information on file..
      IF (para_env%mepos == para_env%source) THEN
         ! Generate file name for the specific Hill
         i = multiple_walkers%walker_id
         filename = TRIM(multiple_walkers%walkers_file_name(i))//"_"// &
                    TRIM(ADJUSTL(cp_to_string(multiple_walkers%n_hills_local)))
         tmpname = TRIM(filename)//".tmp"
         CALL open_file(file_name=tmpname, file_status="UNKNOWN", &
                        file_form="FORMATTED", file_action="WRITE", &
                        file_position="APPEND", unit_number=unit_nr)
         WRITE (unit_nr, *) hills_env%ww_history(hills_env%n_hills)
         DO ih = 1, n_colvar
            WRITE (unit_nr, *) hills_env%ss_history(ih, hills_env%n_hills)
            WRITE (unit_nr, *) hills_env%delta_s_history(ih, hills_env%n_hills)
         END DO
         IF (hills_env%wtcontrol) WRITE (unit_nr, *) hills_env%invdt_history(hills_env%n_hills)
         CALL close_file(unit_nr)
         CALL m_mov(tmpname, filename)
      END IF

      IF (MODULO(multiple_walkers%n_hills_local, multiple_walkers%walkers_freq_comm) == 0) THEN
         ! Store colvars information
         ALLOCATE (ss0_save(n_colvar))
         ALLOCATE (delta_s_save(n_colvar))
         ALLOCATE (delta_s_ss0_buf(2, 0:n_colvar))
         delta_s_ss0_buf = 0
         DO i = 1, n_colvar
            ss0_save(i) = colvars(i)%ss0
            delta_s_save(i) = colvars(i)%delta_s
         END DO

         ! Watch for other walkers's file and update
         DO i = 1, multiple_walkers%walkers_tot_nr
            IF (i == multiple_walkers%walker_id) THEN
               ! Update local counter
               multiple_walkers%walkers_status(i) = multiple_walkers%n_hills_local
               CYCLE
            END IF

            i_hills = multiple_walkers%walkers_status(i)+1
            filename = TRIM(multiple_walkers%walkers_file_name(i))//"_"// &
                       TRIM(ADJUSTL(cp_to_string(i_hills)))

            IF (para_env%mepos == para_env%source) THEN
               INQUIRE (FILE=TRIM(filename), EXIST=exist)
            ENDIF
            CALL mp_bcast(exist, para_env%source, para_env%group)
            DO WHILE (exist)
               ! Read information from the walker's file
               ! We shouldn't care too much about the concurrency of these I/O instructions..
               ! In case, they can be fixed in the future..
               IF (para_env%mepos == para_env%source) THEN
                  CALL open_file(file_name=filename, file_status="OLD", &
                                 file_form="FORMATTED", file_action="READ", &
                                 file_position="REWIND", unit_number=unit_nr)
                  READ (unit_nr, *) delta_s_ss0_buf(1, 0)
                  DO ih = 1, n_colvar
                     READ (unit_nr, *) delta_s_ss0_buf(1, ih)
                     READ (unit_nr, *) delta_s_ss0_buf(2, ih)
                  END DO
                  IF (hills_env%wtcontrol) READ (unit_nr, *) delta_s_ss0_buf(2, 0)
                  CALL close_file(unit_nr)
               ENDIF
               CALL mp_bcast(delta_s_ss0_buf, para_env%source, para_env%group)
               ww = delta_s_ss0_buf(1, 0)
               IF (hills_env%wtcontrol) invdt = delta_s_ss0_buf(2, 0)
               DO ih = 1, n_colvar
                  colvars(ih)%ss0 = delta_s_ss0_buf(1, ih)
                  colvars(ih)%delta_s = delta_s_ss0_buf(2, ih)
               ENDDO

               ! Add this hill to the history dependent terms
               IF (hills_env%wtcontrol) THEN
                  CALL add_hill_single(hills_env, colvars, ww, hills_env%n_hills, n_colvar, invdt=invdt)
               ELSE
                  CALL add_hill_single(hills_env, colvars, ww, hills_env%n_hills, n_colvar)
               END IF

               i_hills = i_hills+1
               filename = TRIM(multiple_walkers%walkers_file_name(i))//"_"// &
                          TRIM(ADJUSTL(cp_to_string(i_hills)))
               IF (para_env%mepos == para_env%source) THEN
                  INQUIRE (FILE=TRIM(filename), EXIST=exist)
               ENDIF
               CALL mp_bcast(exist, para_env%source, para_env%group)
            END DO

            delta_hills = i_hills-1-multiple_walkers%walkers_status(i)
            multiple_walkers%walkers_status(i) = i_hills-1
            iw = cp_print_key_unit_nr(logger, metadyn_section, "PRINT%PROGRAM_RUN_INFO", &
                                      extension=".metadynLog")
            IF (iw > 0) THEN
               WRITE (iw, '(T2,A,I0,A,I0,A,I0,A)') 'WALKERS| Walker #', i, '. Reading [', delta_hills, &
                  '] Hills. Total number of Hills acquired [', multiple_walkers%walkers_status(i), ']'
            END IF
            CALL cp_print_key_finished_output(iw, logger, metadyn_section, &
                                              "PRINT%PROGRAM_RUN_INFO")
         END DO

         ! Restore colvars information
         DO i = 1, n_colvar
            colvars(i)%ss0 = ss0_save(i)
            colvars(i)%delta_s = delta_s_save(i)
         END DO
         DEALLOCATE (ss0_save)
         DEALLOCATE (delta_s_save)
      END IF

      CALL timestop(handle)

   END SUBROUTINE synchronize_multiple_walkers

! **************************************************************************************************
!> \brief Add a single Hill
!> \param hills_env ...
!> \param colvars ...
!> \param ww ...
!> \param n_hills ...
!> \param n_colvar ...
!> \param invdt ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
! **************************************************************************************************
   SUBROUTINE add_hill_single(hills_env, colvars, ww, n_hills, n_colvar, invdt)
      TYPE(hills_env_type), POINTER                      :: hills_env
      TYPE(metavar_type), DIMENSION(:), POINTER          :: colvars
      REAL(KIND=dp), INTENT(IN)                          :: ww
      INTEGER, INTENT(INOUT)                             :: n_hills
      INTEGER, INTENT(IN)                                :: n_colvar
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: invdt

      CHARACTER(len=*), PARAMETER :: routineN = 'add_hill_single', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      LOGICAL                                            :: wtcontrol
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tnp
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: tmp

      CALL timeset(routineN, handle)

      wtcontrol = PRESENT(invdt)
      NULLIFY (tmp, tnp)
      IF (SIZE(hills_env%ss_history, 2) < n_hills+1) THEN
         ALLOCATE (tmp(n_colvar, n_hills+100))
         tmp(:, :n_hills) = hills_env%ss_history
         tmp(:, n_hills+1:) = 0.0_dp
         DEALLOCATE (hills_env%ss_history)
         hills_env%ss_history => tmp
         NULLIFY (tmp)
      ENDIF
      IF (SIZE(hills_env%delta_s_history, 2) < n_hills+1) THEN
         ALLOCATE (tmp(n_colvar, n_hills+100))
         tmp(:, :n_hills) = hills_env%delta_s_history
         tmp(:, n_hills+1:) = 0.0_dp
         DEALLOCATE (hills_env%delta_s_history)
         hills_env%delta_s_history => tmp
         NULLIFY (tmp)
      ENDIF
      IF (SIZE(hills_env%ww_history) < n_hills+1) THEN
         ALLOCATE (tnp(n_hills+100))
         tnp(1:n_hills) = hills_env%ww_history
         tnp(n_hills+1:) = 0.0_dp
         DEALLOCATE (hills_env%ww_history)
         hills_env%ww_history => tnp
         NULLIFY (tnp)
      ENDIF
      IF (wtcontrol) THEN
         IF (SIZE(hills_env%invdt_history) < n_hills+1) THEN
            ALLOCATE (tnp(n_hills+100))
            tnp(1:n_hills) = hills_env%invdt_history
            tnp(n_hills+1:) = 0.0_dp
            DEALLOCATE (hills_env%invdt_history)
            hills_env%invdt_history => tnp
            NULLIFY (tnp)
         ENDIF
      ENDIF
      n_hills = n_hills+1
      ! Now add the hill
      DO i = 1, n_colvar
         hills_env%ss_history(i, n_hills) = colvars(i)%ss0
         hills_env%delta_s_history(i, n_hills) = colvars(i)%delta_s
      ENDDO
      hills_env%ww_history(n_hills) = ww
      IF (wtcontrol) hills_env%invdt_history(n_hills) = invdt

      CALL timestop(handle)

   END SUBROUTINE add_hill_single

! **************************************************************************************************
!> \brief Restart Hills Information
!> \param ss_history ...
!> \param delta_s_history ...
!> \param ww_history ...
!> \param ww ...
!> \param n_hills ...
!> \param n_colvar ...
!> \param colvars ...
!> \param metadyn_section ...
!> \param invdt_history ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
! **************************************************************************************************
   SUBROUTINE restart_hills(ss_history, delta_s_history, ww_history, ww, &
                            n_hills, n_colvar, colvars, metadyn_section, invdt_history)
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ss_history, delta_s_history
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ww_history
      REAL(KIND=dp)                                      :: ww
      INTEGER, INTENT(IN)                                :: n_hills, n_colvar
      TYPE(metavar_type), DIMENSION(:), POINTER          :: colvars
      TYPE(section_vals_type), POINTER                   :: metadyn_section
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: invdt_history

      CHARACTER(len=*), PARAMETER :: routineN = 'restart_hills', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, ndum
      LOGICAL                                            :: explicit, wtcontrol
      REAL(KIND=dp)                                      :: rval
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rvals
      TYPE(section_vals_type), POINTER                   :: hills_history

      CALL timeset(routineN, handle)

      wtcontrol = PRESENT(invdt_history)
      NULLIFY (rvals)
      hills_history => section_vals_get_subs_vals(metadyn_section, "SPAWNED_HILLS_POS")
      CALL section_vals_get(hills_history, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(hills_history, "_DEFAULT_KEYWORD_", n_rep_val=ndum)
         ! ss_history, delta_s_history, ww_history, invdt_history : deallocate and reallocate with the proper size
         DEALLOCATE (ss_history)
         DEALLOCATE (delta_s_history)
         DEALLOCATE (ww_history)
         IF (wtcontrol) THEN
            DEALLOCATE (invdt_history)
         ENDIF
         !
         IF(.NOT.(n_hills == ndum))CALL cp__a("metadynamics_utils.F",768)
         ALLOCATE (ss_history(n_colvar, n_hills))
         ALLOCATE (delta_s_history(n_colvar, n_hills))
         ALLOCATE (ww_history(n_hills))
         IF (wtcontrol) THEN
            ALLOCATE (invdt_history(n_hills))
         ENDIF
         !
         DO i = 1, n_hills
            CALL section_vals_val_get(hills_history, "_DEFAULT_KEYWORD_", &
                                      i_rep_val=i, r_vals=rvals)
            IF(.NOT.(SIZE(rvals) == n_colvar))CALL cp__a("metadynamics_utils.F",779)
            ss_history(1:n_colvar, i) = rvals
         END DO
         !
         hills_history => section_vals_get_subs_vals(metadyn_section, "SPAWNED_HILLS_SCALE")
         CALL section_vals_get(hills_history, explicit=explicit)
         IF (explicit) THEN
            ! delta_s_history
            CALL section_vals_val_get(hills_history, "_DEFAULT_KEYWORD_", n_rep_val=ndum)
            IF(.NOT.(n_hills == ndum))CALL cp__a("metadynamics_utils.F",788)
            DO i = 1, n_hills
               CALL section_vals_val_get(hills_history, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=i, r_vals=rvals)
               IF(.NOT.(SIZE(rvals) == n_colvar))CALL cp__a("metadynamics_utils.F",792)
               delta_s_history(1:n_colvar, i) = rvals
            END DO
         ELSE
            CALL cp_warn(cp__l("metadynamics_utils.F",796), &
                         "Section SPAWNED_HILLS_SCALE is not present! Setting the scales of the"// &
                         "restarted hills according the parameters specified in the input file. ")
            DO i = 1, n_hills
               DO j = 1, n_colvar
                  delta_s_history(j, i) = colvars(i)%delta_s
               END DO
            END DO
         END IF
         !
         hills_history => section_vals_get_subs_vals(metadyn_section, "SPAWNED_HILLS_HEIGHT")
         CALL section_vals_get(hills_history, explicit=explicit)
         IF (explicit) THEN
            ! ww_history
            CALL section_vals_val_get(hills_history, "_DEFAULT_KEYWORD_", &
                                      n_rep_val=ndum)
            IF(.NOT.(n_hills == ndum))CALL cp__a("metadynamics_utils.F",812)
            DO i = 1, n_hills
               CALL section_vals_val_get(hills_history, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=i, r_val=rval)
               IF(.NOT.(SIZE(rvals) == n_colvar))CALL cp__a("metadynamics_utils.F",816)
               ww_history(i) = rval
            END DO
         ELSE
            CALL cp_warn(cp__l("metadynamics_utils.F",820), &
                         "Section SPAWNED_HILLS_HEIGHT is not present! Setting the height of the"// &
                         " restarted hills according the parameters specified in the input file. ")
            ww_history = ww
         END IF
         !
         hills_history => section_vals_get_subs_vals(metadyn_section, "SPAWNED_HILLS_INVDT")
         CALL section_vals_get(hills_history, explicit=explicit)
         IF (wtcontrol) THEN
            IF (explicit) THEN
               ! invdt_history
               CALL section_vals_val_get(hills_history, "_DEFAULT_KEYWORD_", &
                                         n_rep_val=ndum)
               IF(.NOT.(n_hills == ndum))CALL cp__a("metadynamics_utils.F",833)
               DO i = 1, n_hills
                  CALL section_vals_val_get(hills_history, "_DEFAULT_KEYWORD_", &
                                            i_rep_val=i, r_val=rval)
                  IF(.NOT.(SIZE(rvals) == n_colvar))CALL cp__a("metadynamics_utils.F",837)
                  invdt_history(i) = rval
               END DO
            ELSE
               CALL cp_warn(cp__l("metadynamics_utils.F",841), &
                            "Section SPAWNED_HILLS_INVDT is not present! Restarting from standard"// &
                            " metadynamics run i.e. setting 1/(Delta T) equal to zero. ")
               invdt_history = 0._dp
            END IF
         ELSE
            IF (explicit) THEN
               CALL cp_abort(cp__l("metadynamics_utils.F",848), &
                             "Found section SPAWNED_HILLS_INVDT while restarting a standard metadynamics run..."// &
                             " Cannot restart metadynamics from well-tempered MetaD runs. ")
            END IF
         END IF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE restart_hills

! **************************************************************************************************
!> \brief Retrieves the iteration level for the metadynamics loop
!> \param meta_env ...
!> \param iter_nr ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
! **************************************************************************************************
   SUBROUTINE get_meta_iter_level(meta_env, iter_nr)
      TYPE(meta_env_type), POINTER                       :: meta_env
      INTEGER, INTENT(OUT)                               :: iter_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'get_meta_iter_level', &
         routineP = moduleN//':'//routineN

      IF (meta_env%do_multiple_walkers) THEN
         iter_nr = meta_env%multiple_walkers%n_hills_local
      ELSE
         iter_nr = meta_env%hills_env%n_hills
      END IF

   END SUBROUTINE get_meta_iter_level

! **************************************************************************************************
!> \brief ...
!> \param meta_env ...
!> \par History
!>      11.2007 [created] [tlaino]
!> \author Teodoro Laino - University of Zurich - 11.2007
! **************************************************************************************************
   SUBROUTINE meta_walls(meta_env)
      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(len=*), PARAMETER :: routineN = 'meta_walls', routineP = moduleN//':'//routineN

      INTEGER                                            :: ih, iwall
      REAL(dp)                                           :: ddp, delta_s, dfunc, diff_ss, dp2, &
                                                            efunc, ww
      TYPE(metavar_type), DIMENSION(:), POINTER          :: colvars

      colvars => meta_env%metavar
      ! Forces from the Walls
      DO ih = 1, SIZE(colvars)
         IF (colvars(ih)%do_wall) THEN
            colvars(ih)%epot_walls = 0.0_dp
            colvars(ih)%ff_walls = 0.0_dp
            DO iwall = 1, SIZE(colvars(ih)%walls)
               SELECT CASE (colvars (ih)%walls (iwall)%id_type)
               CASE (do_wall_reflective, do_wall_none)
                  ! Do Nothing.. treated in the main metadyn function
                  CYCLE
               CASE (do_wall_quadratic)
                  diff_ss = colvars(ih)%ss0-colvars(ih)%walls(iwall)%pos
                  IF (colvars(ih)%periodic) THEN
                     ! The difference of a periodic COLVAR is always within [-pi,pi]
                     diff_ss = SIGN(1.0_dp, ASIN(SIN(diff_ss)))*ACOS(COS(diff_ss))
                  END IF
                  efunc = colvars(ih)%walls(iwall)%k_quadratic*diff_ss**2
                  dfunc = 2.0_dp*colvars(ih)%walls(iwall)%k_quadratic*diff_ss
                  SELECT CASE (colvars (ih)%walls (iwall)%id_direction)
                  CASE (do_wall_p)
                     IF (diff_ss > 0.0_dp) THEN
                        colvars(ih)%ff_walls = colvars(ih)%ff_walls-dfunc
                        colvars(ih)%epot_walls = colvars(ih)%epot_walls+efunc
                     END IF
                  CASE (do_wall_m)
                     IF (diff_ss < 0.0_dp) THEN
                        colvars(ih)%ff_walls = colvars(ih)%ff_walls-dfunc
                        colvars(ih)%epot_walls = colvars(ih)%epot_walls+efunc
                     END IF
                  END SELECT
               CASE (do_wall_quartic)
                  diff_ss = colvars(ih)%ss0-colvars(ih)%walls(iwall)%pos0
                  IF (colvars(ih)%periodic) THEN
                     ! The difference of a periodic COLVAR is always within [-pi,pi]
                     diff_ss = SIGN(1.0_dp, ASIN(SIN(diff_ss)))*ACOS(COS(diff_ss))
                  END IF
                  efunc = colvars(ih)%walls(iwall)%k_quartic*diff_ss*diff_ss**4
                  dfunc = 4.0_dp*colvars(ih)%walls(iwall)%k_quartic*diff_ss**3
                  SELECT CASE (colvars (ih)%walls (iwall)%id_direction)
                  CASE (do_wall_p)
                     IF (diff_ss > 0.0_dp) THEN
                        colvars(ih)%ff_walls = colvars(ih)%ff_walls-dfunc
                        colvars(ih)%epot_walls = colvars(ih)%epot_walls+efunc
                     END IF
                  CASE (do_wall_m)
                     IF (diff_ss < 0.0_dp) THEN
                        colvars(ih)%ff_walls = colvars(ih)%ff_walls-dfunc
                        colvars(ih)%epot_walls = colvars(ih)%epot_walls+efunc
                     END IF
                  END SELECT
               CASE (do_wall_gaussian)
                  diff_ss = colvars(ih)%ss0-colvars(ih)%walls(iwall)%pos
                  IF (colvars(ih)%periodic) THEN
                     ! The difference of a periodic COLVAR is always within [-pi,pi]
                     diff_ss = SIGN(1.0_dp, ASIN(SIN(diff_ss)))*ACOS(COS(diff_ss))
                  END IF
                  ww = colvars(ih)%walls(iwall)%ww_gauss
                  delta_s = colvars(ih)%walls(iwall)%sigma_gauss
                  ddp = (diff_ss)/delta_s
                  dp2 = ddp**2
                  efunc = ww*EXP(-0.5_dp*dp2)
                  dfunc = -efunc*ddp/delta_s
                  colvars(ih)%ff_walls = colvars(ih)%ff_walls-dfunc
                  colvars(ih)%epot_walls = colvars(ih)%epot_walls+efunc
               END SELECT
            END DO
         END IF
      ENDDO
   END SUBROUTINE meta_walls

END MODULE metadynamics_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE beta_gamma_psi
 ! not tested in the case where dp would stand for single precision
  
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

 IMPLICIT NONE
 PRIVATE

 PUBLIC :: gamln,erf,psi

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param i ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION ipmpar (i) RESULT(fn_val)
!-----------------------------------------------------------------------

!     IPMPAR PROVIDES THE INTEGER MACHINE CONSTANTS FOR THE COMPUTER
!     THAT IS USED. IT IS ASSUMED THAT THE ARGUMENT I IS AN INTEGER
!     HAVING ONE OF THE VALUES 1-10. IPMPAR(I) HAS THE VALUE ...

!  INTEGERS.

!     ASSUME INTEGERS ARE REPRESENTED IN THE N-DIGIT, BASE-A FORM

!               SIGN ( X(N-1)*A**(N-1) + ... + X(1)*A + X(0) )

!               WHERE 0 .LE. X(I) .LT. A FOR I=0,...,N-1.

!     IPMPAR(1) = A, THE BASE (radix).

!     IPMPAR(2) = N, THE NUMBER OF BASE-A DIGITS (digits).

!     IPMPAR(3) = A**N - 1, THE LARGEST MAGNITUDE (huge).

!  FLOATING-POINT NUMBERS.

!     IT IS ASSUMED THAT THE SINGLE AND DOUBLE PRECISION FLOATING
!     POINT ARITHMETICS HAVE THE SAME BASE, SAY B, AND THAT THE
!     NONZERO NUMBERS ARE REPRESENTED IN THE FORM

!               SIGN (B**E) * (X(1)/B + ... + X(M)/B**M)

!               WHERE X(I) = 0,1,...,B-1 FOR I=1,...,M,
!               X(1) .GE. 1, AND EMIN .LE. E .LE. EMAX.

!     IPMPAR(4) = B, THE BASE.

!  SINGLE-PRECISION

!     IPMPAR(5) = M, THE NUMBER OF BASE-B DIGITS.

!     IPMPAR(6) = EMIN, THE SMALLEST EXPONENT E.

!     IPMPAR(7) = EMAX, THE LARGEST EXPONENT E.

!  DOUBLE-PRECISION

!     IPMPAR(8) = M, THE NUMBER OF BASE-B DIGITS.

!     IPMPAR(9) = EMIN, THE SMALLEST EXPONENT E.

!     IPMPAR(10) = EMAX, THE LARGEST EXPONENT E.

!-----------------------------------------------------------------------

      INTEGER, INTENT(IN)                                :: i
      INTEGER                                            :: fn_val

SELECT CASE(i)
  CASE( 1)
    fn_val = RADIX(i)
  CASE( 2)
    fn_val = DIGITS(i)
  CASE( 3)
    fn_val = HUGE(i)
  CASE( 4)
    fn_val = RADIX(1.0)
  CASE( 5)
    fn_val = DIGITS(1.0)
  CASE( 6)
    fn_val = MINEXPONENT(1.0)
  CASE( 7)
    fn_val = MAXEXPONENT(1.0)
  CASE( 8)
    fn_val = DIGITS(1.0e0_dp)
  CASE( 9)
    fn_val = MINEXPONENT(1.0e0_dp)
  CASE(10)
    fn_val = MAXEXPONENT(1.0e0_dp)
  CASE DEFAULT
    CALL cp__b("beta_gamma_psi.F",102,"unknown case")
END SELECT

END FUNCTION ipmpar

! **************************************************************************************************
!> \brief ...
!> \param i ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION dpmpar (i) RESULT(fn_val)
!-----------------------------------------------------------------------

!     DPMPAR PROVIDES THE DOUBLE PRECISION MACHINE CONSTANTS FOR
!     THE COMPUTER BEING USED. IT IS ASSUMED THAT THE ARGUMENT
!     I IS AN INTEGER HAVING ONE OF THE VALUES 1, 2, OR 3. IF THE
!     DOUBLE PRECISION ARITHMETIC BEING USED HAS M BASE B DIGITS AND
!     ITS SMALLEST AND LARGEST EXPONENTS ARE EMIN AND EMAX, THEN

!        DPMPAR(1) = B**(1 - M), THE MACHINE PRECISION,

!        DPMPAR(2) = B**(EMIN - 1), THE SMALLEST MAGNITUDE,

!        DPMPAR(3) = B**EMAX*(1 - B**(-M)), THE LARGEST MAGNITUDE.
!-----------------------------------------------------------------------

      INTEGER, INTENT(IN)                                :: i
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER                                :: one = 1._dp

! Local variable

SELECT CASE (i)
  CASE (1)
    fn_val = EPSILON(one)
  CASE (2)
    fn_val = TINY(one)
  CASE (3)
    fn_val = HUGE(one)
END SELECT

END FUNCTION dpmpar

! **************************************************************************************************
!> \brief ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION epsln () RESULT(fn_val)
!--------------------------------------------------------------------
!     THE EVALUATION OF LN(EPS) WHERE EPS IS THE SMALLEST NUMBER
!     SUCH THAT 1.0 + EPS .GT. 1.0 .  L IS A DUMMY ARGUMENT.
!--------------------------------------------------------------------
      REAL                                               :: fn_val

      REAL, PARAMETER                                    :: one = 1.0

! Local variable

fn_val = LOG( EPSILON(one) )

END FUNCTION epsln

! **************************************************************************************************
!> \brief ...
!> \param l ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION exparg (l) RESULT(fn_val)
!--------------------------------------------------------------------
!     IF L = 0 THEN  EXPARG(L) = THE LARGEST POSITIVE W FOR WHICH
!     EXP(W) CAN BE COMPUTED.
!
!     IF L IS NONZERO THEN  EXPARG(L) = THE LARGEST NEGATIVE W FOR
!     WHICH THE COMPUTED VALUE OF EXP(W) IS NONZERO.
!
!     NOTE... ONLY AN APPROXIMATE VALUE FOR EXPARG(L) IS NEEDED.
!--------------------------------------------------------------------
      INTEGER, INTENT(IN)                                :: l
      REAL                                               :: fn_val

      REAL, PARAMETER                                    :: one = 1.0

! Local variable

IF (l == 0) THEN
  fn_val = LOG( HUGE(one) )
ELSE
  fn_val = LOG( TINY(one) )
END IF

END FUNCTION exparg

! **************************************************************************************************
!> \brief ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION depsln () RESULT(fn_val)
!--------------------------------------------------------------------
!     THE EVALUATION OF LN(EPS) WHERE EPS IS THE SMALLEST NUMBER
!     SUCH THAT 1.e0_dp + EPS .GT. 1.e0_dp .  L IS A DUMMY ARGUMENT.
!--------------------------------------------------------------------
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER                                :: one = 1._dp

! Local variable

fn_val = LOG( EPSILON(one) )

END FUNCTION depsln

! **************************************************************************************************
!> \brief ...
!> \param l ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION dxparg (l) RESULT(fn_val)
!--------------------------------------------------------------------
!     IF L = 0 THEN  DXPARG(L) = THE LARGEST POSITIVE W FOR WHICH
!     DEXP(W) CAN BE COMPUTED.
!
!     IF L IS NONZERO THEN  DXPARG(L) = THE LARGEST NEGATIVE W FOR
!     WHICH THE COMPUTED VALUE OF DEXP(W) IS NONZERO.
!
!     NOTE... ONLY AN APPROXIMATE VALUE FOR DXPARG(L) IS NEEDED.
!--------------------------------------------------------------------
      INTEGER, INTENT(IN)                                :: l
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER                                :: one = 1._dp

! Local variable

IF (l == 0) THEN
  fn_val = LOG( HUGE(one) )
ELSE
  fn_val = LOG( TINY(one) )
END IF

END FUNCTION dxparg

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION alnrel(a) RESULT(fn_val)
!-----------------------------------------------------------------------
!            EVALUATION OF THE FUNCTION LN(1 + A)
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: half = 0.5e0_dp, one = 1.e0_dp, p1 = -.129418923021993e+01_dp, &
         p2 = .405303492862024e+00_dp, p3 = -.178874546012214e-01_dp, &
         q1 = -.162752256355323e+01_dp, q2 = .747811014037616e+00_dp, &
         q3 = -.845104217945565e-01_dp, two = 2.e0_dp, zero = 0.e0_dp

      REAL(dp)                                           :: t, t2, w, x

!--------------------------

IF (ABS(a) <= 0.375e0_dp) THEN
  t = a/(a + two)
  t2 = t*t
  w = (((p3*t2 + p2)*t2 + p1)*t2 + one)/ (((q3*t2 + q2)*t2 + q1)*t2 + one)
  fn_val = two*t*w
ELSE
  x = one + a
  IF (a < zero) x = (a + half) + half
  fn_val = LOG(x)
END IF

END FUNCTION alnrel

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION erf (x) RESULT(fn_val)
!-----------------------------------------------------------------------
!             EVALUATION OF THE REAL ERROR FUNCTION
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: x
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: a(5) = (/ .771058495001320e-04_dp, -.133733772997339e-02_dp,  &
         .323076579225834e-01_dp,  .479137145607681e-01_dp, .128379167095513e+00_dp /), &
         b(3) = (/ .301048631703895e-02_dp,  .538971687740286e-01_dp,.375795757275549e+00_dp /), &
         c = .564189583547756e0_dp, p(8) = (/-1.36864857382717e-07_dp,  5.64195517478974e-01_dp, &
         7.21175825088309e+00_dp,  4.31622272220567e+01_dp, 1.52989285046940e+02_dp,&
         3.39320816734344e+02_dp, 4.51918953711873e+02_dp,  3.00459261020162e+02_dp /)
      REAL(dp), PARAMETER :: q(8) = (/ 1.00000000000000e+00_dp,  1.27827273196294e+01_dp, &
         7.70001529352295e+01_dp,  2.77585444743988e+02_dp,6.38980264465631e+02_dp,  &
         9.31354094850610e+02_dp, 7.90950925327898e+02_dp,  3.00459260956983e+02_dp /), r(5) = (/ &
         2.10144126479064e+00_dp,  2.62370141675169e+01_dp, 2.13688200555087e+01_dp,  &
         4.65807828718470e+00_dp, 2.82094791773523e-01_dp /), s(4) = (/ 9.41537750555460e+01_dp,  &
         1.87114811799590e+02_dp, 9.90191814623914e+01_dp, 1.80124575948747e+01_dp /)

      REAL(dp)                                           :: ax, bot, t, top, x2

ax = ABS(x)
IF (ax < 0.5e0_dp) THEN
  t = x*x
  top = ((((a(1)*t + a(2))*t + a(3))*t + a(4))*t + a(5)) + 1.0e0_dp
  bot = ((b(1)*t + b(2))*t + b(3))*t + 1.0e0_dp
  fn_val = x*(top/bot)
  RETURN

ELSE IF (ax < 4.0e0_dp) THEN
  top = ((((((p(1)*ax + p(2))*ax + p(3))*ax + p(4))*ax + p(5))*ax  &
        + p(6))*ax + p(7))*ax + p(8)
  bot = ((((((q(1)*ax + q(2))*ax + q(3))*ax + q(4))*ax + q(5))*ax  &
        + q(6))*ax + q(7))*ax + q(8)
  fn_val = 0.5e0_dp + (0.5e0_dp - EXP(-x*x)*top/bot)
  IF (x < 0.0e0_dp) fn_val = -fn_val
  RETURN

ELSE IF (ax < 5.8e0_dp) THEN
  x2 = x*x
  t = 1.0e0_dp/x2
  top = (((r(1)*t + r(2))*t + r(3))*t + r(4))*t + r(5)
  bot = (((s(1)*t + s(2))*t + s(3))*t + s(4))*t + 1.0e0_dp
  fn_val = (c - top/(x2*bot)) / ax
  fn_val = 0.5e0_dp + (0.5e0_dp - EXP(-x2)*fn_val)
  IF (x < 0.0e0_dp) fn_val = -fn_val
  RETURN

ELSE
  fn_val = SIGN(1.0e0_dp,x)
END IF

END FUNCTION erf

! **************************************************************************************************
!> \brief ...
!> \param ind ...
!> \param x ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION erfc1 (ind, x) RESULT(fn_val)
!-----------------------------------------------------------------------
!         EVALUATION OF THE COMPLEMENTARY ERROR FUNCTION

!          ERFC1(IND,X) = ERFC(X)            IF IND = 0
!          ERFC1(IND,X) = EXP(X*X)*ERFC(X)   OTHERWISE
!-----------------------------------------------------------------------
      INTEGER, INTENT(IN)                                :: ind
      REAL(dp), INTENT(IN)                               :: x
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: a(5) = (/ .771058495001320e-04_dp, -.133733772997339e-02_dp,  &
         .323076579225834e-01_dp,  .479137145607681e-01_dp,  .128379167095513e+00_dp /), &
         b(3) = (/ .301048631703895e-02_dp,  .538971687740286e-01_dp,.375795757275549e+00_dp /), &
         c = .564189583547756e0_dp, p(8) = (/-1.36864857382717e-07_dp,  5.64195517478974e-01_dp,  &
         7.21175825088309e+00_dp,  4.31622272220567e+01_dp,  1.52989285046940e+02_dp,&
         3.39320816734344e+02_dp,  4.51918953711873e+02_dp,  3.00459261020162e+02_dp /)
      REAL(dp), PARAMETER :: q(8) = (/ 1.00000000000000e+00_dp,  1.27827273196294e+01_dp,  &
         7.70001529352295e+01_dp,  2.77585444743988e+02_dp,6.38980264465631e+02_dp,  &
         9.31354094850610e+02_dp,  7.90950925327898e+02_dp,  3.00459260956983e+02_dp /), r(5) = (/ &
         2.10144126479064e+00_dp,  2.62370141675169e+01_dp,  2.13688200555087e+01_dp,  &
         4.65807828718470e+00_dp,  2.82094791773523e-01_dp /), s(4) = (/ 9.41537750555460e+01_dp,  &
         1.87114811799590e+02_dp,  9.90191814623914e+01_dp, 1.80124575948747e+01_dp /)

      REAL(dp)                                           :: ax, bot, e, t, top, w

ax = ABS(x)
IF (ax <= 0.5e0_dp) THEN
  t = x*x
  top = ((((a(1)*t + a(2))*t + a(3))*t + a(4))*t + a(5)) + 1.0e0_dp
  bot = ((b(1)*t + b(2))*t + b(3))*t + 1.0e0_dp
  fn_val = 0.5e0_dp + (0.5e0_dp - x*(top/bot))
  IF (ind /= 0) fn_val = EXP(t) * fn_val
  RETURN
ENDIF

!                  0.5 < ABS(X) <= 4

IF (ax <= 4.0e0_dp) THEN
top = ((((((p(1)*ax + p(2))*ax + p(3))*ax + p(4))*ax + p(5))*ax  &
      + p(6))*ax + p(7))*ax + p(8)
bot = ((((((q(1)*ax + q(2))*ax + q(3))*ax + q(4))*ax + q(5))*ax  &
      + q(6))*ax + q(7))*ax + q(8)
fn_val = top/bot
ELSE

!                      ABS(X) > 4

IF (x <= -5.6e0_dp) THEN
!             LIMIT VALUE FOR LARGE NEGATIVE X
  fn_val = 2.0e0_dp
  IF (ind /= 0) fn_val = 2.0e0_dp*EXP(x*x)
  RETURN
ENDIF
IF (ind==0.AND.(x > 100.0e0_dp.OR.x*x > -dxparg(1))) THEN
  !             LIMIT VALUE FOR LARGE POSITIVE X
  !                       WHEN IND = 0
  fn_val = 0.0e0_dp
  RETURN
ENDIF

t = (1.0e0_dp/x)**2
top = (((r(1)*t + r(2))*t + r(3))*t + r(4))*t + r(5)
bot = (((s(1)*t + s(2))*t + s(3))*t + s(4))*t + 1.0e0_dp
fn_val = (c - t*top/bot)/ax
ENDIF

!                      FINAL ASSEMBLY

IF (ind /= 0) THEN
  IF (x < 0.0e0_dp) fn_val = 2.0e0_dp*EXP(x*x) - fn_val
  RETURN
END IF
w = x * x
t = w
e = w - t
fn_val = ((0.5e0_dp + (0.5e0_dp - e)) * EXP(-t)) * fn_val
IF (x < 0.0e0_dp) fn_val = 2.0e0_dp - fn_val

END FUNCTION erfc1

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION gam1(a) RESULT(fn_val)
!-----------------------------------------------------------------------
!     COMPUTATION OF 1/GAMMA(A+1) - 1  FOR -0.5 <= A <= 1.5
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: p(7) = (/  .577215664901533e+00_dp, -.409078193005776e+00_dp,&
         -.230975380857675e+00_dp,  .597275330452234e-01_dp, .766968181649490e-02_dp, &
         -.514889771323592e-02_dp, .589597428611429e-03_dp /), q(5) = (/  .100000000000000e+01_dp, &
         .427569613095214e+00_dp, .158451672430138e+00_dp,  .261132021441447e-01_dp, &
         .423244297896961e-02_dp /)
      REAL(dp), PARAMETER :: r(9) = (/ -.422784335098468e+00_dp, -.771330383816272e+00_dp, &
         -.244757765222226e+00_dp,  .118378989872749e+00_dp, .930357293360349e-03_dp,&
         -.118290993445146e-01_dp, .223047661158249e-02_dp,  .266505979058923e-03_dp, &
         -.132674909766242e-03_dp /), s1 = .273076135303957e+00_dp, s2 = .559398236957378e-01_dp

      REAL(dp)                                           :: bot, d, t, top, w

t = a
d = a - 0.5e0_dp
IF (d > 0.0e0_dp) t = d - 0.5e0_dp

IF (t > 0.e0_dp) THEN
  top = (((((p(7)*t + p(6))*t + p(5))*t + p(4))*t + p(3))*t + p(2))*t + p(1)
  bot = (((q(5)*t + q(4))*t + q(3))*t + q(2))*t + 1.0e0_dp
  w = top/bot
  IF (d > 0.0e0_dp) THEN
    fn_val = (t/a)*((w - 0.5e0_dp) - 0.5e0_dp)
  ELSE
    fn_val = a*w
  END IF
ELSE IF (t < 0.e0_dp) THEN
  top = (((((((r(9)*t + r(8))*t + r(7))*t + r(6))*t + r(5))*t  &
           + r(4))*t + r(3))*t + r(2))*t + r(1)
  bot = (s2*t + s1)*t + 1.0e0_dp
  w = top/bot
  IF (d > 0.0e0_dp) THEN
    fn_val = t*w/a
  ELSE
    fn_val = a*((w + 0.5e0_dp) + 0.5e0_dp)
  END IF
ELSE
  fn_val = 0.0e0_dp
END IF

END FUNCTION gam1

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION algdiv (a, b) RESULT(fn_val)
!-----------------------------------------------------------------------

!     COMPUTATION OF LN(GAMMA(B)/GAMMA(A+B)) WHEN B >= 8

!                         --------

!     IN THIS ALGORITHM, DEL(X) IS THE FUNCTION DEFINED BY
!     LN(GAMMA(X)) = (X - 0.5)*LN(X) - X + 0.5*LN(2*PI) + DEL(X).

!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: c0 = .833333333333333e-01_dp, c1 = -.277777777760991e-02_dp, &
         c2 = .793650666825390e-03_dp, c3 = -.595202931351870e-03_dp, &
         c4 = .837308034031215e-03_dp, c5 = -.165322962780713e-02_dp

      REAL(dp)                                           :: c, d, h, s11, s3, s5, s7, s9, t, u, v, &
                                                            w, x, x2

IF (a > b) THEN
  h = b/a
  c = 1.0e0_dp/(1.0e0_dp + h)
  x = h/(1.0e0_dp + h)
  d = a + (b - 0.5e0_dp)
ELSE
  h = a/b
  c = h/(1.0e0_dp + h)
  x = 1.0e0_dp/(1.0e0_dp + h)
  d = b + (a - 0.5e0_dp)
END IF

!                SET SN = (1 - X**N)/(1 - X)

x2 = x*x
s3 = 1.0e0_dp + (x + x2)
s5 = 1.0e0_dp + (x + x2*s3)
s7 = 1.0e0_dp + (x + x2*s5)
s9 = 1.0e0_dp + (x + x2*s7)
s11 = 1.0e0_dp + (x + x2*s9)

!                SET W = DEL(B) - DEL(A + B)

t = (1.0e0_dp/b)**2
w = ((((c5*s11*t + c4*s9)*t + c3*s7)*t + c2*s5)*t + c1*s3)*t + c0
w = w*(c/b)

!                    COMBINE THE RESULTS

u = d*alnrel(a/b)
v = a*(LOG(b) - 1.0e0_dp)
IF (u > v) THEN
  fn_val = (w - v) - u
ELSE
  fn_val = (w - u) - v
END IF

END FUNCTION algdiv

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION rexp (x) RESULT(fn_val)
!-----------------------------------------------------------------------
!            EVALUATION OF THE FUNCTION EXP(X) - 1
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: x
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: p1 = .914041914819518e-09_dp, p2 = .238082361044469e-01_dp, &
         q1 = -.499999999085958e+00_dp, q2 = .107141568980644e+00_dp, &
         q3 = -.119041179760821e-01_dp, q4 = .595130811860248e-03_dp

      REAL(dp)                                           :: e

IF (ABS(x) < 0.15e0_dp) THEN
  fn_val = x*(((p2*x + p1)*x + 1.0e0_dp)/((((q4*x + q3)*x + q2)*x + q1)*x + 1.0e0_dp))
  RETURN
END IF

IF (x < 0.0e0_dp) THEN
   IF (x > -37.0e0_dp) THEN
     fn_val = (EXP(x) - 0.5e0_dp) - 0.5e0_dp
   ELSE
     fn_val = -1.0e0_dp
   ENDIF
ELSE
   e = EXP(x)
   fn_val = e*(0.5e0_dp + (0.5e0_dp - 1.0e0_dp/e))
ENDIF

END FUNCTION rexp

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param y ...
!> \param w ...
!> \param eps ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE bgrat (a, b, x, y, w, eps, ierr)
!-----------------------------------------------------------------------
!     ASYMPTOTIC EXPANSION FOR IX(A,B) WHEN A IS LARGER THAN B.
!     THE RESULT OF THE EXPANSION IS ADDED TO W. IT IS ASSUMED
!     THAT A <= 15 AND B <= 1.  EPS IS THE TOLERANCE USED.
!     IERR IS A VARIABLE THAT REPORTS THE STATUS OF THE RESULTS.
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, x, y
      REAL(dp), INTENT(INOUT)                            :: w
      REAL(dp), INTENT(IN)                               :: eps
      INTEGER, INTENT(OUT)                               :: ierr

      REAL(dp), PARAMETER                                :: half = 0.5e0_dp, one = 1.e0_dp, &
                                                            quarter = 0.25e0_dp, zero = 0.e0_dp

      INTEGER                                            :: i, n
      REAL(dp)                                           :: bm1, bp2n, c(30), cn, coef, d(30), dj, &
                                                            j, l, lnx, n2, nu, q, r, s, sum, t, &
                                                            t2, tol, u, v, z

bm1 = (b - half) - half
nu = a + half*bm1
IF (y > 0.375e0_dp) THEN
  lnx = LOG(x)
ELSE
  lnx = alnrel(-y)
ENDIF
z = -nu*lnx
IF (b*z == zero) THEN
   !               THE EXPANSION CANNOT BE COMPUTED
   ierr = 1
   RETURN
ENDIF

!                 COMPUTATION OF THE EXPANSION
!                 SET R = EXP(-Z)*Z**B/GAMMA(B)

r = b*(one + gam1(b))*EXP(b*LOG(z))
r = r*EXP(a*lnx)*EXP(half*bm1*lnx)
u = algdiv(b,a) + b*LOG(nu)
u = r*EXP(-u)
IF (u == zero) THEN
   !               THE EXPANSION CANNOT BE COMPUTED
   ierr = 1
   RETURN
ENDIF
CALL grat1 (b, z, r, q=q, eps=eps)

tol = 15.0e0_dp*eps
v = quarter*(one/nu)**2
t2 = quarter*lnx*lnx
l = w/u
j = q/r
sum = j
t = one
cn = one
n2 = zero
DO n = 1,30
  bp2n = b + n2
  j = (bp2n*(bp2n + one)*j + (z + bp2n + one)*t)*v
  n2 = n2 + 2.0e0_dp
  t = t*t2
  cn = cn/(n2*(n2 + one))
  c(n) = cn
  s = zero
  IF (.NOT.(n == 1)) THEN
    coef = b - n
    DO i = 1, n-1
      s = s + coef*c(i)*d(n-i)
      coef = coef + b
    END DO
  ENDIF
  d(n) = bm1*cn + s/n
  dj = d(n)*j
  sum = sum + dj
  IF (sum <= zero) THEN
   !               THE EXPANSION CANNOT BE COMPUTED
   ierr = 1
   RETURN
  ENDIF
  IF (ABS(dj) <= tol*(sum + l)) EXIT
END DO

!                    ADD THE RESULTS TO W

ierr = 0
w = w + u*sum

END SUBROUTINE bgrat

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param x ...
!> \param r ...
!> \param p ...
!> \param q ...
!> \param eps ...
! **************************************************************************************************
SUBROUTINE grat1 (a, x, r, p, q, eps)
!-----------------------------------------------------------------------
!           EVALUATION OF P(A,X) AND Q(A,X) WHERE A <= 1 AND
!        THE INPUT ARGUMENT R HAS THE VALUE E**(-X)*X**A/GAMMA(A)
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, x, r
      REAL(dp), INTENT(OUT), OPTIONAL                    :: p, q
      REAL(dp), INTENT(IN)                               :: eps

      REAL(dp), PARAMETER                                :: half = 0.5e0_dp, one = 1.e0_dp, &
                                                            quarter = 0.25e0_dp, three = 3.e0_dp, &
                                                            two = 2.e0_dp, zero = 0.e0_dp

      REAL(dp)                                           :: a2n, a2nm1, an, b2n, b2nm1, c, g, h, j, &
                                                            l, pp, qq, sum, t, tol, w, z

IF (a*x == zero) THEN
  IF (x <= a) THEN
    pp = zero
    qq = one
  ELSE
    pp = one
    qq = zero
  ENDIF
  IF (PRESENT(p)) p = pp
  IF (PRESENT(q)) q = qq
  RETURN
ENDIF
IF (a == half) THEN
IF (x < quarter) THEN
    pp = erf(SQRT(x))
    qq = half + (half - pp)
  ELSE
    qq = erfc1(0,SQRT(x))
    pp = half + (half - qq)
  END IF
  IF (PRESENT(p)) p = pp
  IF (PRESENT(q)) q = qq
  RETURN
ENDIF
IF (x < 1.1e0_dp) THEN
  !             TAYLOR SERIES FOR P(A,X)/X**A

  an = three
  c = x
  sum = x/(a + three)
  tol = three*eps/(a + one)
  an = an + one
  c = -c*(x/an)
  t = c/(a + an)
  sum = sum + t
  DO WHILE (ABS(t) > tol)
    an = an + one
    c = -c*(x/an)
    t = c/(a + an)
    sum = sum + t
  END DO
  j = a*x*((sum/6.0e0_dp - half/(a + two))*x + one/(a + one))

  z = a*LOG(x)
  h = gam1(a)
  g = one + h
  IF ((x < quarter.AND.z > -.13394e0_dp).OR.a < x/2.59e0_dp) THEN
    l = rexp(z)
    qq = ((half + (half + l))*j - l)*g - h
    IF (qq <= zero) THEN
      pp = one
      qq = zero
    ELSE
      pp = half + (half - qq)
    ENDIF
  ELSE
    w = EXP(z)
    pp = w*g*(half + (half - j))
    qq = half + (half - pp)
  ENDIF
ELSE
  !              CONTINUED FRACTION EXPANSION

  tol = 8.0e0_dp*eps
  a2nm1 = one
  a2n = one
  b2nm1 = x
  b2n = x + (one - a)
  c = one
  DO
    a2nm1 = x*a2n + c*a2nm1
    b2nm1 = x*b2n + c*b2nm1
    c = c + one
    a2n = a2nm1 + (c - a)*a2n
    b2n = b2nm1 + (c - a)*b2n
    a2nm1 = a2nm1/b2n
    b2nm1 = b2nm1/b2n
    a2n = a2n/b2n
    b2n = one
    IF (ABS(a2n - a2nm1/b2nm1) < tol*a2n) EXIT
  END DO

  qq = r*a2n
  pp = half + (half - qq)
ENDIF

IF (PRESENT(p)) p = pp
IF (PRESENT(q)) q = qq

END SUBROUTINE grat1

! **************************************************************************************************
!> \brief ...
!> \param mu ...
!> \param x ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION esum (mu, x) RESULT(fn_val)
!-----------------------------------------------------------------------
!                    EVALUATION OF EXP(MU + X)
!-----------------------------------------------------------------------
      INTEGER, INTENT(IN)                                :: mu
      REAL(dp), INTENT(IN)                               :: x
      REAL(dp)                                           :: fn_val

      REAL(dp)                                           :: w

IF (x > 0.0e0_dp) THEN
  IF (mu > 0.OR.mu+x<0.0_dp) THEN
    w = mu
    fn_val = EXP(w)*EXP(x)
  ELSE
    w=mu+x
    fn_val = EXP(w)
  ENDIF
ELSE
  IF (mu < 0.OR.mu+x<0.0_dp) THEN
    w = mu
    fn_val = EXP(w)*EXP(x)
  ELSE
    w=mu+x
    fn_val = EXP(w)
  ENDIF
ENDIF

END FUNCTION esum

! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION rlog1(x) RESULT(fn_val)
!-----------------------------------------------------------------------
!             EVALUATION OF THE FUNCTION X - LN(1 + X)
!-----------------------------------------------------------------------
!     A = RLOG (0.7)
!     B = RLOG (4/3)
!------------------------
      REAL(dp), INTENT(IN)                               :: x
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: a = .566749439387324e-01_dp, b = .456512608815524e-01_dp, &
         p0 = .333333333333333e+00_dp, p1 = -.224696413112536e+00_dp, &
         p2 = .620886815375787e-02_dp, q1 = -.127408923933623e+01_dp, q2 = .354508718369557e+00_dp

      REAL(dp)                                           :: r, t, u, up2, w, w1

IF (x < -0.39e0_dp .OR. x > 0.57e0_dp) THEN
   w = (x + 0.5e0_dp) + 0.5e0_dp
   fn_val = x - LOG(w)
   RETURN
ENDIF

!                 ARGUMENT REDUCTION

IF (x < -0.18e0_dp) THEN
  u = (x + 0.3e0_dp)/0.7e0_dp
  up2 = u + 2.0e0_dp
  w1 = a - u*0.3e0_dp
ELSEIF (x >  0.18e0_dp) THEN
  t = 0.75e0_dp*x
  u = t - 0.25e0_dp
  up2 = t + 1.75e0_dp
  w1 = b + u/3.0e0_dp
ELSE
  u = x
  up2 = u + 2.0e0_dp
  w1 = 0.0e0_dp
ENDIF

!                  SERIES EXPANSION

r = u/up2
t = r*r
w = ((p2*t + p1)*t + p0)/((q2*t + q1)*t + 1.0e0_dp)
fn_val = r*(u - 2.0e0_dp*t*w) + w1

END FUNCTION rlog1

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION gamln (a) RESULT(fn_val)
!-----------------------------------------------------------------------
!            EVALUATION OF LN(GAMMA(A)) FOR POSITIVE A
!-----------------------------------------------------------------------
!     WRITTEN BY ALFRED H. MORRIS
!          NAVAL SURFACE WARFARE CENTER
!          DAHLGREN, VIRGINIA
!--------------------------
!     D = 0.5*(LN(2*PI) - 1)
!--------------------------
      REAL(dp), INTENT(IN)                               :: a
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: c0 = .833333333333333e-01_dp, c1 = -.277777777760991e-02_dp, &
         c2 = .793650666825390e-03_dp, c3 = -.595202931351870e-03_dp, &
         c4 = .837308034031215e-03_dp, c5 = -.165322962780713e-02_dp, d = .418938533204673e0_dp

      INTEGER                                            :: i, n
      REAL(dp)                                           :: t, w

IF (a <= 0.8e0_dp) THEN
  fn_val = gamln1(a) - LOG(a)
ELSEIF (a <= 2.25e0_dp) THEN
  t = (a - 0.5e0_dp) - 0.5e0_dp
  fn_val = gamln1(t)
ELSEIF (a < 10.0e0_dp) THEN
  n = INT(a - 1.25e0_dp)
  t = a
  w = 1.0e0_dp
  DO i = 1, n
    t = t - 1.0e0_dp
    w = t*w
  END DO
  fn_val = gamln1(t - 1.0e0_dp) + LOG(w)
ELSE
  t = (1.0e0_dp/a)**2
  w = (((((c5*t + c4)*t + c3)*t + c2)*t + c1)*t + c0)/a
  fn_val = (d + w) + (a - 0.5e0_dp)*(LOG(a) - 1.0e0_dp)
ENDIF

END FUNCTION gamln

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION gamln1 (a) RESULT(fn_val)
!-----------------------------------------------------------------------
!     EVALUATION OF LN(GAMMA(1 + A)) FOR -0.2 .LE. A .LE. 1.25
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: p0 = .577215664901533e+00_dp, p1 = .844203922187225e+00_dp, &
         p2 = -.168860593646662e+00_dp, p3 = -.780427615533591e+00_dp, &
         p4 = -.402055799310489e+00_dp, p5 = -.673562214325671e-01_dp, &
         p6 = -.271935708322958e-02_dp, q1 = .288743195473681e+01_dp, &
         q2 = .312755088914843e+01_dp, q3 = .156875193295039e+01_dp, q4 = .361951990101499e+00_dp, &
         q5 = .325038868253937e-01_dp, q6 = .667465618796164e-03_dp, r0 = .422784335098467e+00_dp, &
         r1 = .848044614534529e+00_dp, r2 = .565221050691933e+00_dp, r3 = .156513060486551e+00_dp, &
         r4 = .170502484022650e-01_dp, r5 = .497958207639485e-03_dp
      REAL(dp), PARAMETER :: s1 = .124313399877507e+01_dp, s2 = .548042109832463e+00_dp, &
         s3 = .101552187439830e+00_dp, s4 = .713309612391000e-02_dp, s5 = .116165475989616e-03_dp

      REAL(dp)                                           :: w, x

IF (a < 0.6e0_dp) THEN
  w = ((((((p6*a + p5)*a + p4)*a + p3)*a + p2)*a + p1)*a + p0)/  &
      ((((((q6*a + q5)*a + q4)*a + q3)*a + q2)*a + q1)*a + 1.0e0_dp)
  fn_val = -a*w
ELSE
  x = (a - 0.5e0_dp) - 0.5e0_dp
  w = (((((r5*x + r4)*x + r3)*x + r2)*x + r1)*x + r0)/  &
      (((((s5*x + s4)*x + s3)*x + s2)*x + s1)*x + 1.0e0_dp)
  fn_val = x*w
ENDIF

END FUNCTION gamln1

! **************************************************************************************************
!> \brief ...
!> \param xx ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION psi(xx) RESULT(fn_val)
!---------------------------------------------------------------------

!                 EVALUATION OF THE DIGAMMA FUNCTION

!                           -----------

!     PSI(XX) IS ASSIGNED THE VALUE 0 WHEN THE DIGAMMA FUNCTION CANNOT
!     BE COMPUTED.

!     THE MAIN COMPUTATION INVOLVES EVALUATION OF RATIONAL CHEBYSHEV
!     APPROXIMATIONS PUBLISHED IN MATH. COMP. 27, 123-127(1973) BY
!     CODY, STRECOK AND THACHER.

!---------------------------------------------------------------------
!     PSI WAS WRITTEN AT ARGONNE NATIONAL LABORATORY FOR THE FUNPACK
!     PACKAGE OF SPECIAL FUNCTION SUBROUTINES. PSI WAS MODIFIED BY
!     A.H. MORRIS (NSWC).
!---------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: xx
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: dx0 = 1.461632144968362341262659542325721325e0_dp, p1(7) = (/ &
         .895385022981970e-02_dp, .477762828042627e+01_dp,  .142441585084029e+03_dp, &
         .118645200713425e+04_dp,  .363351846806499e+04_dp,.413810161269013e+04_dp,  &
         .130560269827897e+04_dp /), p2(4) = (/ -.212940445131011e+01_dp, -.701677227766759e+01_dp,&
         -.448616543918019e+01_dp,-.648157123766197e+00_dp /), piov4 = .785398163397448e0_dp, q1(6)&
         = (/ .448452573429826e+02_dp, .520752771467162e+03_dp,  .221000799247830e+04_dp, &
         .364127349079381e+04_dp,  .190831076596300e+04_dp,.691091682714533e-05_dp /)
      REAL(dp), PARAMETER :: q2(4) = (/  .322703493791143e+02_dp, .892920700481861e+02_dp,  &
         .546117738103215e+02_dp,  .777788548522962e+01_dp /)

      INTEGER                                            :: i, m, n, nq
      REAL(dp)                                           :: aug, den, sgn, upper, w, x, xmax1, xmx0, &
                                                            xsmall, z

!---------------------------------------------------------------------
!     PIOV4 = PI/4
!     DX0 = ZERO OF PSI TO EXTENDED PRECISION
!---------------------------------------------------------------------
!---------------------------------------------------------------------
!     COEFFICIENTS FOR RATIONAL APPROXIMATION OF
!     PSI(X) / (X - X0),  0.5 <= X <= 3.0
!---------------------------------------------------------------------
!---------------------------------------------------------------------
!     COEFFICIENTS FOR RATIONAL APPROXIMATION OF
!     PSI(X) - LN(X) + 1 / (2*X),  X > 3.0
!---------------------------------------------------------------------
!---------------------------------------------------------------------
!     MACHINE DEPENDENT CONSTANTS ...
!        XMAX1  = THE SMALLEST POSITIVE FLOATING POINT CONSTANT
!                 WITH ENTIRELY INTEGER REPRESENTATION.  ALSO USED
!                 AS NEGATIVE OF LOWER BOUND ON ACCEPTABLE NEGATIVE
!                 ARGUMENTS AND AS THE POSITIVE ARGUMENT BEYOND WHICH
!                 PSI MAY BE REPRESENTED AS ALOG(X).
!        XSMALL = ABSOLUTE ARGUMENT BELOW WHICH PI*COTAN(PI*X)
!                 MAY BE REPRESENTED BY 1/X.
!---------------------------------------------------------------------

xmax1 = ipmpar(3)
xmax1 = MIN(xmax1, 1.0e0_dp/dpmpar(1))
xsmall = 1.e-9_dp
!---------------------------------------------------------------------
x = xx
aug = 0.0e0_dp
IF (x < 0.5e0_dp) THEN
!---------------------------------------------------------------------
!     X .LT. 0.5,  USE REFLECTION FORMULA
!     PSI(1-X) = PSI(X) + PI * COTAN(PI*X)
!---------------------------------------------------------------------
IF (ABS(x) <= xsmall) THEN
  IF (x == 0.0e0_dp) THEN
    !     ERROR RETURN
    fn_val = 0.0e0_dp
    RETURN
  ENDIF
!---------------------------------------------------------------------
!     0 .LT. ABS(X) .LE. XSMALL.  USE 1/X AS A SUBSTITUTE
!     FOR  PI*COTAN(PI*X)
!---------------------------------------------------------------------
  aug = -1.0e0_dp / x
  x = 1.0e0_dp - x
ELSE
!---------------------------------------------------------------------
!     REDUCTION OF ARGUMENT FOR COTAN
!---------------------------------------------------------------------
w = - x
sgn = piov4
IF (w <= 0.0e0_dp) THEN
  w = - w
  sgn = -sgn
ENDIF
!---------------------------------------------------------------------
!     MAKE AN ERROR EXIT IF X .LE. -XMAX1
!---------------------------------------------------------------------
IF (w >= xmax1) THEN
  !     ERROR RETURN
  fn_val = 0.0e0_dp
  RETURN
ENDIF
nq = INT(w)
w = w - nq
nq = INT(w*4.0e0_dp)
w = 4.0e0_dp * (w - nq * .25e0_dp)
!---------------------------------------------------------------------
!     W IS NOW RELATED TO THE FRACTIONAL PART OF  4.0 * X.
!     ADJUST ARGUMENT TO CORRESPOND TO VALUES IN FIRST
!     QUADRANT AND DETERMINE SIGN
!---------------------------------------------------------------------
n = nq / 2
IF ((n+n) /= nq) w = 1.0e0_dp - w
z = piov4 * w
m = n / 2
IF ((m+m) /= n) sgn = - sgn
!---------------------------------------------------------------------
!     DETERMINE FINAL VALUE FOR  -PI*COTAN(PI*X)
!---------------------------------------------------------------------
n = (nq + 1) / 2
m = n / 2
m = m + m
IF (m /= n) THEN
   aug = sgn * ((SIN(z) / COS(z)) * 4.0e0_dp)
ELSE
  !---------------------------------------------------------------------
  !     CHECK FOR SINGULARITY
  !---------------------------------------------------------------------
  IF (z == 0.0e0_dp) THEN
    !     ERROR RETURN
    fn_val = 0.0e0_dp
    RETURN
  ENDIF
!---------------------------------------------------------------------
!     USE COS/SIN AS A SUBSTITUTE FOR COTAN, AND
!     SIN/COS AS A SUBSTITUTE FOR TAN
!---------------------------------------------------------------------
  aug = sgn * ((COS(z) / SIN(z)) * 4.0e0_dp)
ENDIF
x = 1.0e0_dp - x
ENDIF
ENDIF
IF (x <= 3.0e0_dp) THEN
!---------------------------------------------------------------------
!     0.5 .LE. X .LE. 3.0
!---------------------------------------------------------------------
  den = x
  upper = p1(1) * x

  DO i = 1, 5
    den = (den + q1(i)) * x
    upper = (upper + p1(i+1)) * x
  END DO

  den = (upper + p1(7)) / (den + q1(6))
  xmx0 = x - dx0
  fn_val = den * xmx0 + aug
  RETURN
ENDIF
!---------------------------------------------------------------------
!     IF X .GE. XMAX1, PSI = LN(X)
!---------------------------------------------------------------------
IF (x < xmax1) THEN
!---------------------------------------------------------------------
!     3.0 .LT. X .LT. XMAX1
!---------------------------------------------------------------------
w = 1.0e0_dp / (x * x)
den = w
upper = p2(1) * w

DO i = 1, 3
  den = (den + q2(i)) * w
  upper = (upper + p2(i+1)) * w
END DO

aug = upper / (den + q2(4)) - 0.5e0_dp / x + aug
ENDIF
fn_val = aug + LOG(x)

END FUNCTION psi

! **************************************************************************************************
!> \brief ...
!> \param a0 ...
!> \param b0 ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION betaln (a0, b0) RESULT(fn_val)
!-----------------------------------------------------------------------
!     EVALUATION OF THE LOGARITHM OF THE BETA FUNCTION
!-----------------------------------------------------------------------
!     E = 0.5*LN(2*PI)
!--------------------------
      REAL(dp), INTENT(IN)                               :: a0, b0
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER                                :: e = .918938533204673e0_dp

      INTEGER                                            :: i, n
      REAL(dp)                                           :: a, b, c, h, u, v, w, z

!--------------------------

a = MIN(a0,b0)
b = MAX(a0,b0)
!-----------------------------------------------------------------------
!                   PROCEDURE WHEN A .GE. 8
!-----------------------------------------------------------------------
IF (a >= 8.0e0_dp) THEN
  w = bcorr(a,b)
  h = a/b
  c = h/(1.0e0_dp + h)
  u = -(a - 0.5e0_dp)*LOG(c)
  v = b*alnrel(h)
  IF (u > v) THEN
    fn_val = (((-0.5e0_dp*LOG(b) + e) + w) - v) - u
  ELSE
    fn_val = (((-0.5e0_dp*LOG(b) + e) + w) - u) - v
  ENDIF
!-----------------------------------------------------------------------
!                   PROCEDURE WHEN A .LT. 1
!-----------------------------------------------------------------------
ELSEIF (a < 1.0e0_dp) THEN
  IF (b < 8.0e0_dp) THEN
    fn_val = gamln(a) + (gamln(b) - gamln(a + b))
  ELSE
    fn_val = gamln(a) + algdiv(a,b)
  ENDIF
!-----------------------------------------------------------------------
!                PROCEDURE WHEN 1 .LE. A .LT. 8
!-----------------------------------------------------------------------
ELSEIF (a <= 2.0e0_dp) THEN
  IF (b <= 2.0e0_dp) THEN
    fn_val = gamln(a) + gamln(b) - gsumln(a,b)
    RETURN
  ENDIF
  w = 0.0e0_dp
  IF (b < 8.0e0_dp) THEN
    !                 REDUCTION OF B WHEN B .LT. 8

    n = INT(b - 1.0e0_dp)
    z = 1.0e0_dp
    DO i = 1,n
      b = b - 1.0e0_dp
      z = z * (b/(a + b))
    END DO
    fn_val = w + LOG(z) + (gamln(a) + (gamln(b) - gsumln(a,b)))
    RETURN
  ENDIF
  fn_val = gamln(a) + algdiv(a,b)
!                REDUCTION OF A WHEN B .LE. 1000
ELSEIF (b <= 1000.0e0_dp) THEN
  n = INT(a - 1.0e0_dp)
  w = 1.0e0_dp
  DO i = 1, n
    a = a - 1.0e0_dp
    h = a/b
    w = w * (h/(1.0e0_dp + h))
  END DO
  w = LOG(w)
  IF (b >= 8.0e0_dp) THEN
    fn_val = w + gamln(a) + algdiv(a,b)
    RETURN
  ENDIF

  !                 REDUCTION OF B WHEN B .LT. 8

  n = INT(b - 1.0e0_dp)
  z = 1.0e0_dp
  DO i = 1,n
    b = b - 1.0e0_dp
    z = z * (b/(a + b))
  END DO
  fn_val = w + LOG(z) + (gamln(a) + (gamln(b) - gsumln(a,b)))
ELSE
!                REDUCTION OF A WHEN B .GT. 1000

  n = INT(a - 1.0e0_dp)
  w = 1.0e0_dp
  DO i = 1,n
    a = a - 1.0e0_dp
    w = w * (a/(1.0e0_dp + a/b))
  END DO
  fn_val = (LOG(w) - n*LOG(b)) + (gamln(a) + algdiv(a,b))
ENDIF

END FUNCTION betaln

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION gsumln (a, b) RESULT(fn_val)
!-----------------------------------------------------------------------
!          EVALUATION OF THE FUNCTION LN(GAMMA(A + B))
!          FOR 1 .LE. A .LE. 2  AND  1 .LE. B .LE. 2
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b
      REAL(dp)                                           :: fn_val

      REAL(dp)                                           :: x

x = a + b - 2.e0_dp
IF (x <= 0.25e0_dp) THEN
  fn_val = gamln1(1.0e0_dp + x)
ELSEIF (x <= 1.25e0_dp) THEN
  fn_val = gamln1(x) + alnrel(x)
ELSE
  fn_val = gamln1(x - 1.0e0_dp) + LOG(x*(1.0e0_dp + x))
ENDIF
END FUNCTION gsumln

! **************************************************************************************************
!> \brief ...
!> \param a0 ...
!> \param b0 ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION bcorr (a0, b0) RESULT(fn_val)
!-----------------------------------------------------------------------

!     EVALUATION OF  DEL(A0) + DEL(B0) - DEL(A0 + B0)  WHERE
!     LN(GAMMA(A)) = (A - 0.5)*LN(A) - A + 0.5*LN(2*PI) + DEL(A).
!     IT IS ASSUMED THAT A0 .GE. 8 AND B0 .GE. 8.

!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a0, b0
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER :: c0 = .833333333333333e-01_dp, c1 = -.277777777760991e-02_dp, &
         c2 = .793650666825390e-03_dp, c3 = -.595202931351870e-03_dp, &
         c4 = .837308034031215e-03_dp, c5 = -.165322962780713e-02_dp

      REAL(dp)                                           :: a, b, c, h, s11, s3, s5, s7, s9, t, w, &
                                                            x, x2

a = MIN(a0, b0)
b = MAX(a0, b0)

h = a/b
c = h/(1.0e0_dp + h)
x = 1.0e0_dp/(1.0e0_dp + h)
x2 = x*x

!                SET SN = (1 - X**N)/(1 - X)

s3 = 1.0e0_dp + (x + x2)
s5 = 1.0e0_dp + (x + x2*s3)
s7 = 1.0e0_dp + (x + x2*s5)
s9 = 1.0e0_dp + (x + x2*s7)
s11 = 1.0e0_dp + (x + x2*s9)

!                SET W = DEL(B) - DEL(A + B)

t = (1.0e0_dp/b)**2
w = ((((c5*s11*t + c4*s9)*t + c3*s7)*t + c2*s5)*t + c1*s3)*t + c0
w = w*(c/b)

!                   COMPUTE  DEL(A) + W

t = (1.0e0_dp/a)**2
fn_val = (((((c5*t + c4)*t + c3)*t + c2)*t + c1)*t + c0)/a + w
RETURN
END FUNCTION bcorr

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param y ...
!> \param w ...
!> \param w1 ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE bratio (a, b, x, y, w, w1, ierr)
!-----------------------------------------------------------------------

!            EVALUATION OF THE INCOMPLETE BETA FUNCTION IX(A,B)

!                     --------------------

!     IT IS ASSUMED THAT A AND B ARE NONNEGATIVE, AND THAT X <= 1
!     AND Y = 1 - X.  BRATIO ASSIGNS W AND W1 THE VALUES

!                      W  = IX(A,B)
!                      W1 = 1 - IX(A,B)

!     IERR IS A VARIABLE THAT REPORTS THE STATUS OF THE RESULTS.
!     IF NO INPUT ERRORS ARE DETECTED THEN IERR IS SET TO 0 AND
!     W AND W1 ARE COMPUTED. OTHERWISE, IF AN ERROR IS DETECTED,
!     THEN W AND W1 ARE ASSIGNED THE VALUE 0 AND IERR IS SET TO
!     ONE OF THE FOLLOWING VALUES ...

!        IERR = 1  IF A OR B IS NEGATIVE
!        IERR = 2  IF A = B = 0
!        IERR = 3  IF X .LT. 0 OR X .GT. 1
!        IERR = 4  IF Y .LT. 0 OR Y .GT. 1
!        IERR = 5  IF X + Y .NE. 1
!        IERR = 6  IF X = A = 0
!        IERR = 7  IF Y = B = 0

!--------------------
!     WRITTEN BY ALFRED H. MORRIS, JR.
!        NAVAL SURFACE WARFARE CENTER
!        DAHLGREN, VIRGINIA
!     REVISED ... APRIL 1993
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, x, y
      REAL(dp), INTENT(OUT)                              :: w, w1
      INTEGER, INTENT(OUT)                               :: ierr

      INTEGER                                            :: ierr1, ind, n
      REAL(dp)                                           :: a0, b0, eps, lambda, t, x0, y0, z

!-----------------------------------------------------------------------
!     ****** EPS IS A MACHINE DEPENDENT CONSTANT. EPS IS THE SMALLEST
!            FLOATING POINT NUMBER FOR WHICH 1.0 + EPS .GT. 1.0

eps = dpmpar(1)

w = 0.0e0_dp
w1 = 0.0e0_dp
IF (a < 0.0e0_dp .OR. b < 0.0e0_dp) THEN
  ierr = 1
  RETURN
ENDIF
IF (a == 0.0e0_dp .AND. b == 0.0e0_dp) THEN
  ierr = 2
  RETURN
ENDIF
IF (x < 0.0e0_dp .OR. x > 1.0e0_dp) THEN
  ierr = 3
  RETURN
ENDIF
IF (y < 0.0e0_dp .OR. y > 1.0e0_dp) THEN
  ierr = 4
  RETURN
ENDIF
z = ((x + y) - 0.5e0_dp) - 0.5e0_dp
IF (ABS(z) > 3.0e0_dp*eps) THEN
  ierr = 5
  RETURN
ENDIF

ierr = 0

IF (x == 0.0e0_dp .OR. a == 0.0e0_dp) THEN
   IF (x == 0.0e0_dp .AND. a == 0.0e0_dp) THEN
      ierr = 6
   ELSE
      w = 0.0e0_dp
      w1 = 1.0e0_dp
   ENDIF
   RETURN
ENDIF

IF (y == 0.0e0_dp .OR. b == 0.0e0_dp) THEN
   IF (y == 0.0e0_dp .AND. b == 0.0e0_dp) THEN
      ierr = 7
   ELSE
      w = 1.0e0_dp
      w1 = 0.0e0_dp
   ENDIF
   RETURN
ENDIF

eps = MAX(eps, 1.e-15_dp)
IF (MAX(a,b) < 1.e-3_dp*eps) THEN
!           PROCEDURE FOR A AND B .LT. 1.E-3*EPS
   w = b/(a + b)
   w1 = a/(a + b)
   RETURN
ENDIF

ind = 0
a0 = a
b0 = b
x0 = x
y0 = y
IF (MIN(a0, b0) > 1.0e0_dp) GO TO 30

!             PROCEDURE FOR A0 .LE. 1 OR B0 .LE. 1

IF (x > 0.5e0_dp) THEN
  ind = 1
  a0 = b
  b0 = a
  x0 = y
  y0 = x
ENDIF

IF (b0 < MIN(eps,eps*a0)) GO TO 80
IF (a0 < MIN(eps,eps*b0) .AND. b0*x0 <= 1.0e0_dp) GO TO 90
IF (MAX(a0, b0) > 1.0e0_dp) GO TO 20
IF (a0 >= MIN(0.2e0_dp, b0)) GO TO 100
IF (x0**a0 <= 0.9e0_dp) GO TO 100
IF (x0 >= 0.3e0_dp) GO TO 110
n = 20
GO TO 130

20 IF (b0 <= 1.0e0_dp) GO TO 100
IF (x0 >= 0.3e0_dp) GO TO 110
IF (x0 < 0.1e0_dp.AND.(x0*b0)**a0 <= 0.7e0_dp) GO TO 100
IF (b0 > 15.0e0_dp) GO TO 131
n = 20
GO TO 130

!             PROCEDURE FOR A0 .GT. 1 AND B0 .GT. 1

30 IF (a > b) THEN
   lambda = (a + b)*y - b
ELSE
   lambda = a - (a + b)*x
ENDIF

IF (lambda < 0.0e0_dp) THEN
   ind = 1
   a0 = b
   b0 = a
   x0 = y
   y0 = x
   lambda = ABS(lambda)
ENDIF

IF (b0 < 40.0e0_dp .AND. b0*x0 <= 0.7e0_dp) GO TO 100
IF (b0 < 40.0e0_dp) GO TO 140
IF (a0 > b0) GO TO 50
IF (a0 <= 100.0e0_dp) GO TO 120
IF (lambda > 0.03e0_dp*a0) GO TO 120
GO TO 180
50 IF (b0 <= 100.0e0_dp) GO TO 120
IF (lambda > 0.03e0_dp*b0) GO TO 120
GO TO 180

!            EVALUATION OF THE APPROPRIATE ALGORITHM

80 w = fpser(a0, b0, x0, eps)
w1 = 0.5e0_dp + (0.5e0_dp - w)
GO TO 220

90 w1 = apser(a0, b0, x0, eps)
w = 0.5e0_dp + (0.5e0_dp - w1)
GO TO 220

100 w = bpser(a0, b0, x0, eps)
w1 = 0.5e0_dp + (0.5e0_dp - w)
GO TO 220

110 w1 = bpser(b0, a0, y0, eps)
w = 0.5e0_dp + (0.5e0_dp - w1)
GO TO 220

120 w = bfrac(a0, b0, x0, y0, lambda, 15.0e0_dp*eps)
w1 = 0.5e0_dp + (0.5e0_dp - w)
GO TO 220

130 w1 = bup(b0, a0, y0, x0, n, eps)
b0 = b0 + n
131 CALL bgrat (b0, a0, y0, x0, w1, eps, ierr1)
IF (ierr1 > 0) CALL cp__b("beta_gamma_psi.F",1535,"Error in BGRAT")
w = 0.5e0_dp + (0.5e0_dp - w1)
GO TO 220

140 n = INT(b0)
b0 = b0 - n
IF (b0 /= 0.0e0_dp) GO TO 141
n = n - 1
b0 = 1.0e0_dp
141 w = bup(b0, a0, y0, x0, n, eps)
IF (x0 > 0.7e0_dp) GO TO 150
w = w + bpser(a0, b0, x0, eps)
w1 = 0.5e0_dp + (0.5e0_dp - w)
GO TO 220

150 IF (a0 > 15.0e0_dp) GO TO 151
n = 20
w = w + bup(a0, b0, x0, y0, n, eps)
a0 = a0 + n
151 CALL bgrat (a0, b0, x0, y0, w, eps, ierr1)
w1 = 0.5e0_dp + (0.5e0_dp - w)
GO TO 220

180 w = basym(a0, b0, lambda, 100.0e0_dp*eps)
w1 = 0.5e0_dp + (0.5e0_dp - w)
GO TO 220

!               TERMINATION OF THE PROCEDURE

220 IF (ind == 0) RETURN
t = w
w = w1
w1 = t

END SUBROUTINE bratio

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param eps ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION fpser (a, b, x, eps) RESULT(fn_val)
!-----------------------------------------------------------------------

!                 EVALUATION OF I (A,B)
!                                X

!          FOR B .LT. MIN(EPS,EPS*A) AND X .LE. 0.5.

!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, x, eps
      REAL(dp)                                           :: fn_val

      REAL(dp)                                           :: an, c, s, t, tol

!                  SET  FPSER = X**A

fn_val = 1.0e0_dp
IF (a > 1.e-3_dp*eps) THEN
  fn_val = 0.0e0_dp
  t = a*LOG(x)
  IF (t < dxparg(1)) RETURN
  fn_val = EXP(t)
ENDIF

!                NOTE THAT 1/B(A,B) = B

fn_val = (b/a)*fn_val
tol = eps/a
an = a + 1.0e0_dp
t = x
s = t/an
an = an + 1.0e0_dp
t = x*t
c = t/an
s = s + c
DO WHILE (ABS(c) > tol)
  an = an + 1.0e0_dp
  t = x*t
  c = t/an
  s = s + c
ENDDO

fn_val = fn_val*(1.0e0_dp + a*s)

END FUNCTION fpser

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param eps ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION apser (a, b, x, eps) RESULT(fn_val)
!-----------------------------------------------------------------------
!     APSER YIELDS THE INCOMPLETE BETA RATIO I(SUB(1-X))(B,A) FOR
!     A .LE. MIN(EPS,EPS*B), B*X .LE. 1, AND X .LE. 0.5. USED WHEN
!     A IS VERY SMALL. USE ONLY IF ABOVE INEQUALITIES ARE SATISFIED.
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, x, eps
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER                                :: g = .577215664901533e0_dp

      REAL(dp)                                           :: aj, bx, c, j, s, t, tol

bx = b*x
t = x - bx
IF (b*eps > 2.e-2_dp) THEN
  c = LOG(bx) + g + t
ELSE
  c = LOG(x) + psi(b) + g + t
ENDIF

tol = 5.0e0_dp*eps*ABS(c)
j = 1.0e0_dp
s = 0.0e0_dp
j = j + 1.0e0_dp
t = t*(x - bx/j)
aj = t/j
s = s + aj
DO WHILE (ABS(aj) > tol)
  t = t*(x - bx/j)
  aj = t/j
  s = s + aj
ENDDO

fn_val = -a*(c + s)

END FUNCTION apser

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param eps ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION bpser (a, b, x, eps) RESULT(fn_val)
!-----------------------------------------------------------------------
!     POWER SERIES EXPANSION FOR EVALUATING IX(A,B) WHEN B .LE. 1
!     OR B*X .LE. 0.7.  EPS IS THE TOLERANCE USED.
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, x, eps
      REAL(dp)                                           :: fn_val

      INTEGER                                            :: i, m
      REAL(dp)                                           :: a0, apb, b0, c, n, sum, t, tol, u, w, z

fn_val = 0.0e0_dp
IF (x == 0.0e0_dp) RETURN
!-----------------------------------------------------------------------
!            COMPUTE THE FACTOR X**A/(A*BETA(A,B))
!-----------------------------------------------------------------------
a0 = MIN(a,b)
b0 = MAX(a,b)
IF (a0 >= 1.0e0_dp) THEN
  z = a*LOG(x) - betaln(a,b)
  fn_val = EXP(z)/a
ELSEIF (b0 >= 8.0e0_dp) THEN
  u = gamln1(a0) + algdiv(a0,b0)
  z = a*LOG(x) - u
  fn_val = (a0/a)*EXP(z)
ELSEIF (b0 > 1.0e0_dp) THEN
!         PROCEDURE FOR A0 .LT. 1 AND 1 .LT. B0 .LT. 8

  u = gamln1(a0)
  m = INT(b0 - 1.0e0_dp)
  IF (m >= 1) THEN
    c = 1.0e0_dp
    DO i = 1, m
      b0 = b0 - 1.0e0_dp
      c = c*(b0/(a0 + b0))
    END DO
    u = LOG(c) + u
  ENDIF

  z = a*LOG(x) - u
  b0 = b0 - 1.0e0_dp
  apb = a0 + b0
  IF (apb > 1.0e0_dp) THEN
    u = a0 + b0 - 1.e0_dp
    t = (1.0e0_dp + gam1(u))/apb
  ELSE
    t = 1.0e0_dp + gam1(apb)
  ENDIF
  fn_val = EXP(z)*(a0/a)*(1.0e0_dp + gam1(b0))/t
ELSE

  !            PROCEDURE FOR A0 .LT. 1 AND B0 .LE. 1

  fn_val = x**a
  IF (fn_val == 0.0e0_dp) RETURN

  apb = a + b
  IF (apb > 1.0e0_dp) THEN
    u = a + b - 1.e0_dp
    z = (1.0e0_dp + gam1(u))/apb
  ELSE
    z = 1.0e0_dp + gam1(apb)
  ENDIF

  c = (1.0e0_dp + gam1(a))*(1.0e0_dp + gam1(b))/z
  fn_val = fn_val*c*(b/apb)
ENDIF

!            PROCEDURE FOR A0 .LT. 1 AND B0 .GE. 8

IF (fn_val == 0.0e0_dp .OR. a <= 0.1e0_dp*eps) RETURN
!-----------------------------------------------------------------------
!                     COMPUTE THE SERIES
!-----------------------------------------------------------------------
sum = 0.0e0_dp
n = 0.0e0_dp
c = 1.0e0_dp
tol = eps/a
n = n + 1.0e0_dp
c = c*(0.5e0_dp + (0.5e0_dp - b/n))*x
w = c/(a + n)
sum = sum + w
DO WHILE (ABS(w) > tol)
  n = n + 1.0e0_dp
  c = c*(0.5e0_dp + (0.5e0_dp - b/n))*x
  w = c/(a + n)
  sum = sum + w
END DO
fn_val = fn_val*(1.0e0_dp + a*sum)

END FUNCTION bpser

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param y ...
!> \param n ...
!> \param eps ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION bup (a, b, x, y, n, eps) RESULT(fn_val)
!-----------------------------------------------------------------------
!     EVALUATION OF IX(A,B) - IX(A+N,B) WHERE N IS A POSITIVE INTEGER.
!     EPS IS THE TOLERANCE USED.
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, x, y
      INTEGER, INTENT(IN)                                :: n
      REAL(dp), INTENT(IN)                               :: eps
      REAL(dp)                                           :: fn_val

      INTEGER                                            :: i, k, kp1, mu, nm1
      REAL(dp)                                           :: ap1, apb, d, l, r, t, w

!          OBTAIN THE SCALING FACTOR EXP(-MU) AND
!             EXP(MU)*(X**A*Y**B/BETA(A,B))/A

apb = a + b
ap1 = a + 1.0e0_dp
mu = 0
d = 1.0e0_dp
IF (.NOT.(n == 1 .OR. a < 1.0e0_dp.OR.apb < 1.1e0_dp*ap1)) THEN
  mu = INT(ABS(dxparg(1)))
  k = INT(dxparg(0))
  IF (k < mu) mu = k
  t = mu
  d = EXP(-t)
ENDIF

fn_val = brcmp1(mu,a,b,x,y)/a
IF (n == 1 .OR. fn_val == 0.0e0_dp) RETURN
nm1 = n - 1
w = d

!          LET K BE THE INDEX OF THE MAXIMUM TERM

k = 0
IF (b > 1.0e0_dp) THEN
  IF (y <= 1.e-4_dp) THEN
    k = nm1
    DO i = 1,k
      l = i - 1
      d = ((apb + l)/(ap1 + l))*x*d
      w = w + d
    END DO
    IF (k == nm1) THEN
      fn_val = fn_val*w
      RETURN
    ENDIF
  ELSE
    r = (b - 1.0e0_dp)*x/y - a
    IF (r >= 1.0e0_dp) THEN
      k = nm1
      t = nm1
      IF (r < t) k = INT(r)

!          ADD THE INCREASING TERMS OF THE SERIES

      DO i = 1,k
        l = i - 1
        d = ((apb + l)/(ap1 + l))*x*d
        w = w + d
      END DO
      IF (k == nm1) THEN 
        fn_val = fn_val*w
        RETURN
      ENDIF
    ENDIF
  ENDIF
ENDIF

!          ADD THE REMAINING TERMS OF THE SERIES

kp1 = k + 1
DO i = kp1,nm1
  l = i - 1
  d = ((apb + l)/(ap1 + l))*x*d
  w = w + d
  IF (d <= eps*w) EXIT
END DO

!               TERMINATE THE PROCEDURE

fn_val = fn_val*w

END FUNCTION bup

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param y ...
!> \param lambda ...
!> \param eps ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION bfrac (a, b, x, y, lambda, eps) RESULT(fn_val)
!-----------------------------------------------------------------------
!     CONTINUED FRACTION EXPANSION FOR IX(A,B) WHEN A,B .GT. 1.
!     IT IS ASSUMED THAT  LAMBDA = (A + B)*Y - B.
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, x, y, lambda, eps
      REAL(dp)                                           :: fn_val

      REAL(dp)                                           :: alpha, an, anp1, beta, bn, bnp1, c, c0, &
                                                            c1, e, n, p, r, r0, s, t, w, yp1

fn_val = brcomp(a,b,x,y)
IF (fn_val == 0.0e0_dp) RETURN

c = 1.0e0_dp + lambda
c0 = b/a
c1 = 1.0e0_dp + 1.0e0_dp/a
yp1 = y + 1.0e0_dp

n = 0.0e0_dp
p = 1.0e0_dp
s = a + 1.0e0_dp
an = 0.0e0_dp
bn = 1.0e0_dp
anp1 = 1.0e0_dp
bnp1 = c/c1
r = c1/c

!        CONTINUED FRACTION CALCULATION

DO WHILE(.TRUE.)
n = n + 1.0e0_dp
t = n/a
w = n*(b - n)*x
e = a/s
alpha = (p*(p + c0)*e*e)*(w*x)
IF (alpha <= 0.0e0_dp) THEN
!                 TERMINATION
  fn_val = fn_val*r
  RETURN
ENDIF
e = (1.0e0_dp + t)/(c1 + t + t)
beta = n + w/s + e*(c + n*yp1)
p = 1.0e0_dp + t
s = s + 2.0e0_dp

!        UPDATE AN, BN, ANP1, AND BNP1

t = alpha*an + beta*anp1
an = anp1
anp1 = t
t = alpha*bn + beta*bnp1
bn = bnp1
bnp1 = t
r0 = r
r = anp1/bnp1
IF (ABS(r - r0) <= eps*r) THEN
!                 TERMINATION
  fn_val = fn_val*r
  RETURN
ENDIF

!        RESCALE AN, BN, ANP1, AND BNP1

an = an/bnp1
bn = bn/bnp1
anp1 = r
bnp1 = 1.0e0_dp
ENDDO

END FUNCTION bfrac

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param y ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION brcomp (a, b, x, y) RESULT(fn_val)
!-----------------------------------------------------------------------
!               EVALUATION OF X**A*Y**B/BETA(A,B)
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, x, y
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER                                :: const = 0.398942280401433e0_dp

      INTEGER                                            :: i, n
      REAL(dp)                                           :: a0, apb, b0, c, e, h, lambda, lnx, lny, &
                                                            t, u, v, x0, y0, z

!-----------------
!     CONST = 1/SQRT(2*PI)
!-----------------

fn_val = 0.0e0_dp
IF (x == 0.0e0_dp .OR. y == 0.0e0_dp) RETURN
a0 = MIN(a,b)
IF (a0 >= 8.0e0_dp) THEN
!-----------------------------------------------------------------------
!              PROCEDURE FOR A .GE. 8 AND B .GE. 8
!-----------------------------------------------------------------------
   IF (a > b) THEN
      h = b/a
      x0 = 1.0e0_dp/(1.0e0_dp + h)
      y0 = h/(1.0e0_dp + h)
      lambda = (a + b)*y - b
   ELSE
      h = a/b
      x0 = h/(1.0e0_dp + h)
      y0 = 1.0e0_dp/(1.0e0_dp + h)
      lambda = a - (a + b)*x
   ENDIF

   e = -lambda/a
   IF (ABS(e) > 0.6e0_dp) THEN
      u = e - LOG(x/x0)
   ELSE
      u = rlog1(e)
   ENDIF

   e = lambda/b
   IF (ABS(e) > 0.6e0_dp) THEN
      v = e - LOG(y/y0)
   ELSE
      v = rlog1(e)
   ENDIF

   z = EXP(-(a*u + b*v))
   fn_val = const*SQRT(b*x0)*z*EXP(-bcorr(a,b))
   RETURN
ENDIF

IF (x > 0.375e0_dp) THEN
  IF (y > 0.375e0_dp) THEN
    lnx = LOG(x)
    lny = LOG(y)
  ELSE
    lnx = alnrel(-y)
    lny = LOG(y)
  ENDIF
ELSE
  lnx = LOG(x)
  lny = alnrel(-x)
ENDIF

z = a*lnx + b*lny
IF (a0 >= 1.0e0_dp) THEN
  z = z - betaln(a,b)
  fn_val = EXP(z)
  RETURN
ENDIF
!-----------------------------------------------------------------------
!              PROCEDURE FOR A .LT. 1 OR B .LT. 1
!-----------------------------------------------------------------------
b0 = MAX(a,b)
IF (b0 >= 8.0e0_dp) THEN
!                   ALGORITHM FOR B0 .GE. 8
   u = gamln1(a0) + algdiv(a0,b0)
   fn_val = a0*EXP(z - u)
ENDIF
IF (b0 <= 1.0e0_dp) THEN

!                   ALGORITHM FOR B0 .LE. 1

  fn_val = EXP(z)
  IF (fn_val == 0.0e0_dp) RETURN

  apb = a + b
  IF (apb > 1.0e0_dp) THEN
    u = a + b - 1.e0_dp
    z = (1.0e0_dp + gam1(u))/apb
  ELSE
    z = 1.0e0_dp + gam1(apb)
  ENDIF

  c = (1.0e0_dp + gam1(a))*(1.0e0_dp + gam1(b))/z
  fn_val = fn_val*(a0*c)/(1.0e0_dp + a0/b0)
  RETURN
ENDIF

!                ALGORITHM FOR 1 .LT. B0 .LT. 8

u = gamln1(a0)
n = INT(b0 - 1.0e0_dp)
IF (n >= 1) THEN
   c = 1.0e0_dp
   DO i = 1, n
      b0 = b0 - 1.0e0_dp
      c = c*(b0/(a0 + b0))
   END DO
   u = LOG(c) + u
ENDIF

z = z - u
b0 = b0 - 1.0e0_dp
apb = a0 + b0
IF (apb > 1.0e0_dp) THEN
   u = a0 + b0 - 1.e0_dp
   t = (1.0e0_dp + gam1(u))/apb
ELSE
   t = 1.0e0_dp + gam1(apb)
ENDIF
fn_val = a0*EXP(z)*(1.0e0_dp + gam1(b0))/t



END FUNCTION brcomp

! **************************************************************************************************
!> \brief ...
!> \param mu ...
!> \param a ...
!> \param b ...
!> \param x ...
!> \param y ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION brcmp1 (mu, a, b, x, y) RESULT(fn_val)
!-----------------------------------------------------------------------
!          EVALUATION OF  EXP(MU) * (X**A*Y**B/BETA(A,B))
!-----------------------------------------------------------------------
      INTEGER, INTENT(IN)                                :: mu
      REAL(dp), INTENT(IN)                               :: a, b, x, y
      REAL(dp)                                           :: fn_val

      REAL(dp), PARAMETER                                :: const = 0.398942280401433e0_dp

      INTEGER                                            :: i, n
      REAL(dp)                                           :: a0, apb, b0, c, e, h, lambda, lnx, lny, &
                                                            t, u, v, x0, y0, z

!-----------------
!     CONST = 1/SQRT(2*PI)
!-----------------

a0 = MIN(a,b)
IF (a0 >= 8.0e0_dp) THEN
!-----------------------------------------------------------------------
!              PROCEDURE FOR A .GE. 8 AND B .GE. 8
!-----------------------------------------------------------------------
  IF (a > b) THEN
    h = b/a
    x0 = 1.0e0_dp/(1.0e0_dp + h)
    y0 = h/(1.0e0_dp + h)
    lambda = (a + b)*y - b
  ENDIF
  h = a/b
  x0 = h/(1.0e0_dp + h)
  y0 = 1.0e0_dp/(1.0e0_dp + h)
  lambda = a - (a + b)*x

  e = -lambda/a
  IF (ABS(e) > 0.6e0_dp) THEN
    u = e - LOG(x/x0)
  ELSE
    u = rlog1(e)
  ENDIF

  e = lambda/b
  IF (ABS(e) <= 0.6e0_dp) THEN
    v = rlog1(e)
  ELSE
    v = e - LOG(y/y0)
  ENDIF

  z = esum(mu,-(a*u + b*v))
  fn_val = const*SQRT(b*x0)*z*EXP(-bcorr(a,b))
ENDIF

IF (x > 0.375e0_dp) THEN
  IF (y > 0.375e0_dp) THEN
    lnx = LOG(x)
    lny = LOG(y)
  ELSE
    lnx = alnrel(-y)
    lny = LOG(y)
  ENDIF
ELSE
  lnx = LOG(x)
  lny = alnrel(-x)
ENDIF
z = a*lnx + b*lny
IF (a0 >= 1.0e0_dp) THEN
  z = z - betaln(a,b)
  fn_val = esum(mu,z)
  RETURN
ENDIF
!-----------------------------------------------------------------------
!              PROCEDURE FOR A .LT. 1 OR B .LT. 1
!-----------------------------------------------------------------------
b0 = MAX(a,b)
IF (b0 >= 8.0e0_dp) THEN
!                   ALGORITHM FOR B0 .GE. 8

  u = gamln1(a0) + algdiv(a0,b0)
  fn_val = a0*esum(mu,z - u)
  RETURN
ENDIF
IF (b0 <= 1.0e0_dp) THEN

!                   ALGORITHM FOR B0 .LE. 1

  fn_val = esum(mu,z)
  IF (fn_val == 0.0e0_dp) RETURN

  apb = a + b
  IF (apb > 1.0e0_dp) THEN
    u = a + b - 1.e0_dp
    z = (1.0e0_dp + gam1(u))/apb
  ELSE
    z = 1.0e0_dp + gam1(apb)
  ENDIF

  c = (1.0e0_dp + gam1(a))*(1.0e0_dp + gam1(b))/z
  fn_val = fn_val*(a0*c)/(1.0e0_dp + a0/b0)
  RETURN
ENDIF

!                ALGORITHM FOR 1 .LT. B0 .LT. 8

u = gamln1(a0)
n = INT(b0 - 1.0e0_dp)
IF (n >= 1) THEN
  c = 1.0e0_dp
  DO i = 1, n
    b0 = b0 - 1.0e0_dp
    c = c*(b0/(a0 + b0))
  END DO
  u = LOG(c) + u
ENDIF

z = z - u
b0 = b0 - 1.0e0_dp
apb = a0 + b0
IF (apb > 1.0e0_dp) THEN
  u = a0 + b0 - 1.e0_dp
  t = (1.0e0_dp + gam1(u))/apb
ELSE
  t = 1.0e0_dp + gam1(apb)
ENDIF
fn_val = a0*esum(mu,z)*(1.0e0_dp + gam1(b0))/t

END FUNCTION brcmp1

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param lambda ...
!> \param eps ...
!> \retval fn_val ...
! **************************************************************************************************
FUNCTION basym (a, b, lambda, eps) RESULT(fn_val)
!-----------------------------------------------------------------------
!     ASYMPTOTIC EXPANSION FOR IX(A,B) FOR LARGE A AND B.
!     LAMBDA = (A + B)*Y - B  AND EPS IS THE TOLERANCE USED.
!     IT IS ASSUMED THAT LAMBDA IS NONNEGATIVE AND THAT
!     A AND B ARE GREATER THAN OR EQUAL TO 15.
!-----------------------------------------------------------------------
      REAL(dp), INTENT(IN)                               :: a, b, lambda, eps
      REAL(dp)                                           :: fn_val

      INTEGER, PARAMETER                                 :: num = 20
      REAL(dp), PARAMETER                                :: e0 = 1.12837916709551e0_dp, &
                                                            e1 = .353553390593274e0_dp

      INTEGER                                            :: i, im1, imj, j, m, mm1, mmj, n, np1
      REAL(dp)                                           :: a0(21), b0(21), bsum, c(21), d(21), &
                                                            dsum, f, h, h2, hn, j0, j1, r, r0, r1, &
                                                            s, sum, t, t0, t1, u, w, w0, z, z0, &
                                                            z2, zn, znm1

!------------------------
!     ****** NUM IS THE MAXIMUM VALUE THAT N CAN TAKE IN THE DO LOOP
!            ENDING AT STATEMENT 50. IT IS REQUIRED THAT NUM BE EVEN.
!            THE ARRAYS A0, B0, C, D HAVE DIMENSION NUM + 1.
!------------------------
!     E0 = 2/SQRT(PI)
!     E1 = 2**(-3/2)
!------------------------

fn_val = 0.0e0_dp
IF (a >= b) THEN
   r0 = 1.0e0_dp/(1.0e0_dp + h)
   r1 = (b - a)/a
   w0 = 1.0e0_dp/SQRT(b*(1.0e0_dp + h))
ELSE
   h = a/b
   r0 = 1.0e0_dp/(1.0e0_dp + h)
   r1 = (b - a)/b
   w0 = 1.0e0_dp/SQRT(a*(1.0e0_dp + h))
ENDIF

20 f = a*rlog1(-lambda/a) + b*rlog1(lambda/b)
t = EXP(-f)
IF (t == 0.0e0_dp) RETURN
z0 = SQRT(f)
z = 0.5e0_dp*(z0/e1)
z2 = f + f

a0(1) = (2.0e0_dp/3.0e0_dp)*r1
c(1) = - 0.5e0_dp*a0(1)
d(1) = - c(1)
j0 = (0.5e0_dp/e0)*erfc1(1,z0)
j1 = e1
sum = j0 + d(1)*w0*j1

s = 1.0e0_dp
h2 = h*h
hn = 1.0e0_dp
w = w0
znm1 = z
zn = z2
DO n = 2, num, 2
  hn = h2*hn
  a0(n) = 2.0e0_dp*r0*(1.0e0_dp + h*hn)/(n + 2.0e0_dp)
  np1 = n + 1
  s = s + hn
  a0(np1) = 2.0e0_dp*r1*s/(n + 3.0e0_dp)

  DO i = n, np1
    r = -0.5e0_dp*(i + 1.0e0_dp)
    b0(1) = r*a0(1)
    DO m = 2, i
      bsum = 0.0e0_dp
      mm1 = m - 1
      DO j = 1, mm1
        mmj = m - j
        bsum = bsum + (j*r - mmj)*a0(j)*b0(mmj)
      END DO
      b0(m) = r*a0(m) + bsum/m
    END DO
    c(i) = b0(i)/(i + 1.0e0_dp)

    dsum = 0.0e0_dp
    im1 = i - 1
    DO j = 1, im1
      imj = i - j
      dsum = dsum + d(imj)*c(j)
    END DO
    d(i) = -(dsum + c(i))
  END DO

  j0 = e1*znm1 + (n - 1.0e0_dp)*j0
  j1 = e1*zn + n*j1
  znm1 = z2*znm1
  zn = z2*zn
  w = w0*w
  t0 = d(n)*w*j0
  w = w0*w
  t1 = d(np1)*w*j1
  sum = sum + (t0 + t1)
  IF ((ABS(t0) + ABS(t1)) <= eps*sum) EXIT
END DO

u = EXP(-bcorr(a,b))
fn_val = e0*t*u*sum

END FUNCTION basym

END MODULE beta_gamma_psi
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief methods used in the flexible partitioning scheme
!> \par History
!>      04.2006 [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE fp_methods

   USE beta_gamma_psi,                  ONLY: erf,&
                                              gamln,&
                                              psi
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_iter_string,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE fp_types,                        ONLY: fp_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fac,&
                                              maxfac,&
                                              oorootpi
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: fp_eval

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fp_methods'

CONTAINS

! **************************************************************************************************
!> \brief computest the forces and the energy due to the flexible potential & bias,
!>     and writes the weights file
!> \param fp_env ...
!> \param subsys ...
!> \param cell ...
!> \par History
!>      04.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE fp_eval(fp_env, subsys, cell)
      TYPE(fp_type), POINTER                             :: fp_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'fp_eval', routineP = moduleN//':'//routineN

      CHARACTER(LEN=15)                                  :: tmpstr
      INTEGER                                            :: handle, i, icenter, iparticle, &
                                                            output_unit
      LOGICAL                                            :: zero_weight
      REAL(KIND=dp)                                      :: c, dcdr, kT, r, rab(3), sf, strength
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(particle_list_type), POINTER                  :: particles_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(fp_env)))CALL cp__a("fp_methods.F",71)
      IF(.NOT.(fp_env%use_fp))CALL cp__a("fp_methods.F",72)
      IF(.NOT.(ASSOCIATED(subsys)))CALL cp__a("fp_methods.F",73)
      CALL cp_subsys_get(subsys, particles=particles_list)
      particles => particles_list%els

      ! compute the force due to the reflecting walls
      ! and count the distribution in discrete and contiguous ways
      zero_weight = .FALSE.
      fp_env%restraint_energy = 0.0_dp
      icenter = fp_env%central_atom
      strength = fp_env%strength
      fp_env%i1 = 0; fp_env%i2 = 0; fp_env%o1 = 0; fp_env%o2 = 0
      fp_env%ri1 = 0.0_dp; fp_env%ri2 = 0.0_dp; fp_env%ro1 = 0.0_dp; fp_env%ro2 = 0.0_dp
      fp_env%energy = 0.0_dp

      ! inner particles
      DO i = 1, SIZE(fp_env%inner_atoms)
         iparticle = fp_env%inner_atoms(i)
         rab = particles(iparticle)%r-particles(icenter)%r
         rab = pbc(rab, cell)
         r = SQRT(SUM(rab**2))
         ! constraint wall  (they feel to outer wall)
         IF (r > fp_env%outer_radius) THEN
            zero_weight = .TRUE.
            fp_env%restraint_energy = fp_env%restraint_energy+0.5_dp*strength*(r-fp_env%outer_radius)**2
            sf = strength*(r-fp_env%outer_radius)/r
            particles(iparticle)%f = particles(iparticle)%f-sf*rab
            particles(icenter)%f = particles(icenter)%f+sf*rab
         ENDIF
         ! count the distribution
         IF (r > fp_env%inner_radius) THEN
            fp_env%i2 = fp_env%i2+1
         ELSE
            fp_env%i1 = fp_env%i1+1
         ENDIF
         ! smooth count the distribution
         CALL smooth_count(r, fp_env%inner_radius, fp_env%smooth_width, c, dcdr)
         fp_env%ri1 = fp_env%ri1+c
         fp_env%ri2 = fp_env%ri2+(1.0_dp-c)
      ENDDO

      ! outer particles
      DO i = 1, SIZE(fp_env%outer_atoms)
         iparticle = fp_env%outer_atoms(i)
         rab = particles(iparticle)%r-particles(icenter)%r
         rab = pbc(rab, cell)
         r = SQRT(SUM(rab**2))
         ! constraint wall (they feel the inner wall)
         IF (r < fp_env%inner_radius) THEN
            zero_weight = .TRUE.
            fp_env%restraint_energy = fp_env%restraint_energy+ &
                                      0.5_dp*strength*(r-fp_env%inner_radius)**2
            sf = strength*(r-fp_env%inner_radius)/r
            particles(iparticle)%f = particles(iparticle)%f-sf*rab
            particles(icenter)%f = particles(icenter)%f+sf*rab
         ENDIF
         ! count the distribution
         IF (r > fp_env%outer_radius) THEN
            fp_env%o2 = fp_env%o2+1
         ELSE
            fp_env%o1 = fp_env%o1+1
         ENDIF
         ! smooth count the distribution
         CALL smooth_count(r, fp_env%outer_radius, fp_env%smooth_width, c, dcdr)
         fp_env%ro1 = fp_env%ro1+c
         fp_env%ro2 = fp_env%ro2+(1.0_dp-c)
      ENDDO
      fp_env%energy = fp_env%energy+fp_env%restraint_energy

      ! the combinatorial weight
      i = fp_env%i2+fp_env%o1
      IF(.NOT.(i <= maxfac))CALL cp__a("fp_methods.F",143)
      fp_env%comb_weight = (fac(fp_env%i2)*fac(fp_env%o1))/fac(i)

      ! we can add the bias potential now.
      ! this bias has the form
      ! kT * { ln[(o1+i2)!] - ln[o1!] - ln[i2!] }
      ! where the smooth counts are used for o1 and i2
      fp_env%bias_energy = 0.0_dp
      IF (fp_env%bias) THEN
         kT = fp_env%temperature
         fp_env%bias_energy = kT*(gamln(fp_env%ro1+fp_env%ri2+1)- &
                                  gamln(fp_env%ro1+1)-gamln(fp_env%ri2+1))

         ! and add the corresponding forces
         ! inner particles
         DO i = 1, SIZE(fp_env%inner_atoms)
            iparticle = fp_env%inner_atoms(i)
            rab = particles(iparticle)%r-particles(icenter)%r
            rab = pbc(rab, cell)
            r = SQRT(SUM(rab**2))
            CALL smooth_count(r, fp_env%inner_radius, fp_env%smooth_width, c, dcdr)
            sf = kT*(psi(fp_env%ro1+fp_env%ri2+1)-psi(fp_env%ri2+1))*(-dcdr)/r
            particles(iparticle)%f = particles(iparticle)%f-sf*rab
            particles(icenter)%f = particles(icenter)%f+sf*rab
         ENDDO
         ! outer particles
         DO i = 1, SIZE(fp_env%outer_atoms)
            iparticle = fp_env%outer_atoms(i)
            rab = particles(iparticle)%r-particles(icenter)%r
            rab = pbc(rab, cell)
            r = SQRT(SUM(rab**2))
            CALL smooth_count(r, fp_env%outer_radius, fp_env%smooth_width, c, dcdr)
            sf = kT*(psi(fp_env%ro1+fp_env%ri2+1)-psi(fp_env%ro1+1))*(dcdr)/r
            particles(iparticle)%f = particles(iparticle)%f-sf*rab
            particles(icenter)%f = particles(icenter)%f+sf*rab
         ENDDO
      ENDIF
      fp_env%energy = fp_env%energy+fp_env%bias_energy
      fp_env%bias_weight = EXP(fp_env%bias_energy/kT)

      ! if this configuration is a valid one, compute its weight
      IF (zero_weight) THEN
         fp_env%weight = 0.0_dp
      ELSE
         fp_env%weight = fp_env%comb_weight*fp_env%bias_weight
      ENDIF

      ! put weights and other info on file
      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, fp_env%print_section, "", &
                                         extension=".weights")
      IF (output_unit > 0) THEN
         tmpstr = cp_iter_string(logger%iter_info, fp_env%print_section)
         WRITE (output_unit, '(T2,A15,6(1X,F16.10),4(1X,I4),4(1X,F16.10))') &
            tmpstr, &
            fp_env%weight, fp_env%comb_weight, fp_env%bias_weight, &
            fp_env%energy, fp_env%restraint_energy, fp_env%bias_energy, &
            fp_env%i1, fp_env%i2, fp_env%o1, fp_env%o2, &
            fp_env%ri1, fp_env%ri2, fp_env%ro1, fp_env%ro2
      ENDIF

      CALL cp_print_key_finished_output(output_unit, logger, fp_env%print_section, &
                                        "")

      CALL timestop(handle)

   END SUBROUTINE fp_eval

! **************************************************************************************************
!> \brief counts in a smooth way (error function with width=width)
!>      if r is closer than r1. Returns 1.0 for the count=c if r<<r1
!>      and the derivative wrt r dcdr
!> \param r ...
!> \param r1 ...
!> \param width ...
!> \param c ...
!> \param dcdr ...
!> \par History
!>      04.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE smooth_count(r, r1, width, c, dcdr)
      REAL(KIND=dp), INTENT(IN)                          :: r, r1, width
      REAL(KIND=dp), INTENT(OUT)                         :: c, dcdr

      REAL(KIND=dp)                                      :: arg

      arg = (r1-r)/width

      c = (1.0_dp+erf(arg))/2.0_dp
      dcdr = (-oorootpi/width)*EXP(-arg**2)

   END SUBROUTINE

END MODULE fp_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines needed for EMD
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

MODULE rt_propagation_forces
   USE admm_types,                      ONLY: admm_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE cp_control_types,                ONLY: dft_control_type,&
                                              rtp_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_col_block_sizes, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_deallocate_matrix, &
        cp_dbcsr_distribution, cp_dbcsr_get_block_p, cp_dbcsr_get_data_size, &
        cp_dbcsr_get_data_type, cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_row_block_sizes, &
        cp_dbcsr_type, dbcsr_create_dist_r_unrot, dbcsr_distribution_obj, &
        dbcsr_distribution_release, dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: copy_fm_to_dbcsr,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_struct,                    ONLY: cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: one,&
                                              zero
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: add_qs_force,&
                                              qs_force_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_overlap,                      ONLY: build_overlap_force
   USE rt_propagation_types,            ONLY: get_rtp,&
                                              rt_prop_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: calc_c_mat_force, &
             rt_admm_force

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_propagation_forces'

CONTAINS

! **************************************************************************************************
!> \brief calculates the three additional force contributions needed in EMD
!>        P_imag*C , P_imag*B*S^-1*S_der , P*S^-1*H*S_der
!>        driver routine
!> \param qs_env ...
!> \par History
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE calc_c_mat_force(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_c_mat_force', &
         routineP = moduleN//':'//routineN

      IF (qs_env%rtp%linear_scaling) THEN
         CALL calc_c_mat_force_ls(qs_env)
      ELSE
         CALL calc_c_mat_force_fm(qs_env)
      END IF

   END SUBROUTINE calc_c_mat_force

! **************************************************************************************************
!> \brief standard treatment for fm MO based calculations
!>        P_imag*C , P_imag*B*S^-1*S_der , P*S^-1*H*S_der
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE calc_c_mat_force_fm(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_c_mat_force_fm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, im, ispin, nao, natom, nmo, re
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size
      REAL(KIND=dp)                                      :: alpha
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: C_mat, S_der, SinvB, SinvH
      TYPE(cp_dbcsr_type)                                :: db_mo_tmp1, db_mo_tmp2, db_mos_im, &
                                                            db_mos_re
      TYPE(cp_dbcsr_type), POINTER                       :: rho_im_sparse, tmp_dbcsr
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new
      TYPE(dbcsr_distribution_obj)                       :: dist
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(rt_prop_type), POINTER                        :: rtp

      CALL timeset(routineN, handle)

      NULLIFY (rtp, particle_set, atomic_kind_set, mos)
      NULLIFY (tmp_dbcsr, rho_im_sparse)
      CALL get_qs_env(qs_env=qs_env, rtp=rtp, particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, mos=mos, force=force)

      CALL get_rtp(rtp=rtp, C_mat=C_mat, S_der=S_der, &
                   SinvH=SinvH, SinvB=SinvB, mos_new=mos_new)

      ALLOCATE (tmp_dbcsr)
      ALLOCATE (rho_im_sparse)
      CALL cp_dbcsr_init(tmp_dbcsr)
      CALL cp_dbcsr_create(tmp_dbcsr, template=SinvH(1)%matrix)
      CALL cp_dbcsr_init(rho_im_sparse)
      CALL cp_dbcsr_create(rho_im_sparse, template=SinvH(1)%matrix)

      natom = SIZE(particle_set)
      ALLOCATE (atom_of_kind(natom))
      ALLOCATE (kind_of(natom))

      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind, kind_of=kind_of)

      DO ispin = 1, SIZE(SinvH)
         re = 2*ispin-1
         im = 2*ispin
         alpha = mos(ispin)%mo_set%maxocc

         CALL get_mo_set(mos(ispin)%mo_set, nao=nao, nmo=nmo)

         NULLIFY (col_blk_size)
         CALL dbcsr_create_dist_r_unrot(dist, cp_dbcsr_distribution(SinvB(ispin)%matrix), nmo, col_blk_size)

         CALL cp_dbcsr_init(db_mos_re)
         CALL cp_dbcsr_create(db_mos_re, "D", dist, dbcsr_type_no_symmetry, &
                              cp_dbcsr_row_block_sizes(SinvB(ispin)%matrix), col_blk_size, &
                              nze=0)
         CALL cp_dbcsr_init(db_mos_im)
         CALL cp_dbcsr_create(db_mos_im, template=db_mos_re)
         CALL cp_dbcsr_init(db_mo_tmp1)
         CALL cp_dbcsr_create(db_mo_tmp1, template=db_mos_re)
         CALL cp_dbcsr_init(db_mo_tmp2)
         CALL cp_dbcsr_create(db_mo_tmp2, template=db_mos_re)

         CALL copy_fm_to_dbcsr(mos_new(im)%matrix, db_mos_im)
         CALL copy_fm_to_dbcsr(mos_new(re)%matrix, db_mos_re)

         CALL cp_dbcsr_multiply("N", "N", alpha, SinvB(ispin)%matrix, db_mos_im, 0.0_dp, db_mo_tmp1)
         CALL cp_dbcsr_multiply("N", "N", alpha, SinvH(ispin)%matrix, db_mos_re, 1.0_dp, db_mo_tmp1)
         CALL cp_dbcsr_multiply("N", "N", -alpha, SinvB(ispin)%matrix, db_mos_re, 0.0_dp, db_mo_tmp2)
         CALL cp_dbcsr_multiply("N", "N", alpha, SinvH(ispin)%matrix, db_mos_im, 1.0_dp, db_mo_tmp2)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, db_mo_tmp1, db_mos_re, 0.0_dp, tmp_dbcsr)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, db_mo_tmp2, db_mos_im, 1.0_dp, tmp_dbcsr)

         CALL cp_dbcsr_multiply("N", "T", alpha, db_mos_re, db_mos_im, 0.0_dp, rho_im_sparse)
         CALL cp_dbcsr_multiply("N", "T", -alpha, db_mos_im, db_mos_re, 1.0_dp, rho_im_sparse)

         CALL compute_forces(force, tmp_dbcsr, S_der, rho_im_sparse, C_mat, kind_of, atom_of_kind)

         CALL cp_dbcsr_release(db_mos_re)
         CALL cp_dbcsr_release(db_mos_im)
         CALL cp_dbcsr_release(db_mo_tmp1)
         CALL cp_dbcsr_release(db_mo_tmp2)

         DEALLOCATE (col_blk_size)
         CALL dbcsr_distribution_release(dist)

      END DO

      DO i = 1, SIZE(force)
         force(i)%ehrenfest(:, :) = -force(i)%ehrenfest(:, :)
      END DO

      CALL cp_dbcsr_release(tmp_dbcsr)
      CALL cp_dbcsr_release(rho_im_sparse)
      DEALLOCATE (atom_of_kind)
      DEALLOCATE (kind_of)
      DEALLOCATE (tmp_dbcsr)
      DEALLOCATE (rho_im_sparse)

      CALL timestop(handle)

   END SUBROUTINE calc_c_mat_force_fm

! **************************************************************************************************
!> \brief special treatment ofr linear scaling
!>        P_imag*C , P_imag*B*S^-1*S_der , P*S^-1*H*S_der
!> \param qs_env ...
!> \par History
!>      02.2014 switched to dbcsr matrices [Samuel Andermatt]
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE calc_c_mat_force_ls(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_c_mat_force_ls', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, im, ispin, natom, re
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: C_mat, rho_new, S_der, SinvB, SinvH
      TYPE(cp_dbcsr_type), POINTER                       :: S_inv, S_minus_half, tmp, tmp2, tmp3
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CALL timeset(routineN, handle)
      NULLIFY (rtp, particle_set, atomic_kind_set, dft_control)

      CALL get_qs_env(qs_env, &
                      rtp=rtp, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      force=force, &
                      dft_control=dft_control)

      rtp_control => dft_control%rtp_control
      CALL get_rtp(rtp=rtp, C_mat=C_mat, S_der=S_der, S_inv=S_inv, &
                   SinvH=SinvH, SinvB=SinvB)

      natom = SIZE(particle_set)
      ALLOCATE (atom_of_kind(natom))
      ALLOCATE (kind_of(natom))

      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind, kind_of=kind_of)

      NULLIFY (tmp)
      ALLOCATE (tmp)
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=SinvB(1)%matrix)
      NULLIFY (tmp2)
      ALLOCATE (tmp2)
      CALL cp_dbcsr_init(tmp2)
      CALL cp_dbcsr_create(tmp2, template=SinvB(1)%matrix)
      NULLIFY (tmp3)
      ALLOCATE (tmp3)
      CALL cp_dbcsr_init(tmp3)
      CALL cp_dbcsr_create(tmp3, template=SinvB(1)%matrix)

      CALL get_rtp(rtp=rtp, rho_new=rho_new, S_minus_half=S_minus_half)

      DO ispin = 1, SIZE(SinvH)
         re = 2*ispin-1
         im = 2*ispin
         IF (rtp_control%orthonormal) THEN
            CALL cp_dbcsr_multiply("N", "N", one, S_minus_half, SinvH(ispin)%matrix, zero, tmp2, &
                                   filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_multiply("N", "N", one, tmp2, rho_new(re)%matrix, zero, tmp3, &
                                   filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_multiply("N", "N", one, tmp3, S_minus_half, zero, tmp, &
                                   filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_multiply("N", "N", one, S_minus_half, SinvB(ispin)%matrix, zero, tmp2, &
                                   filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_multiply("N", "N", one, tmp2, rho_new(im)%matrix, zero, tmp3, &
                                   filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_multiply("N", "N", one, tmp3, S_minus_half, one, tmp, &
                                   filter_eps=rtp%filter_eps)
         ELSE
            CALL cp_dbcsr_multiply("N", "N", one, SinvH(ispin)%matrix, rho_new(re)%matrix, zero, tmp, &
                                   filter_eps=rtp%filter_eps)
            CALL cp_dbcsr_multiply("N", "N", one, SinvB(ispin)%matrix, rho_new(im)%matrix, one, tmp, &
                                   filter_eps=rtp%filter_eps)
         ENDIF

         CALL compute_forces(force, tmp, S_der, rho_new(im)%matrix, C_mat, kind_of, atom_of_kind)

      END DO

      ! recall QS forces, at this point have the other sign.
      DO i = 1, SIZE(force)
         force(i)%ehrenfest(:, :) = -force(i)%ehrenfest(:, :)
      END DO

      CALL cp_dbcsr_deallocate_matrix(tmp)
      CALL cp_dbcsr_deallocate_matrix(tmp2)
      CALL cp_dbcsr_deallocate_matrix(tmp3)

      DEALLOCATE (atom_of_kind)
      DEALLOCATE (kind_of)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param force ...
!> \param tmp ...
!> \param S_der ...
!> \param rho_im ...
!> \param C_mat ...
!> \param kind_of ...
!> \param atom_of_kind ...
! **************************************************************************************************
   SUBROUTINE compute_forces(force, tmp, S_der, rho_im, C_mat, kind_of, atom_of_kind)
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(cp_dbcsr_type), POINTER                       :: tmp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: S_der
      TYPE(cp_dbcsr_type), POINTER                       :: rho_im
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: C_mat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of, atom_of_kind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_forces', routineP = moduleN//':'//routineN

      INTEGER                                            :: col_atom, i, ikind, kind_atom, row_atom
      LOGICAL                                            :: found
      REAL(dp), DIMENSION(:), POINTER                    :: block_values, block_values2
      TYPE(cp_dbcsr_iterator)                            :: iter

      DO i = 1, 3
         !Calculate the sum over the hadmard product
         !S_der part

         CALL cp_dbcsr_iterator_start(iter, tmp)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row_atom, col_atom, block_values)
            CALL cp_dbcsr_get_block_p(S_der(i)%matrix, row_atom, col_atom, block_values2, found=found)
            IF (found) THEN
               ikind = kind_of(col_atom)
               kind_atom = atom_of_kind(col_atom)
               !The block_values are in a vector format,
               ! so the dot_product is the sum over all elements of the hamand product, that I need
               force(ikind)%ehrenfest(i, kind_atom) = force(ikind)%ehrenfest(i, kind_atom)+ &
                                                      2.0_dp*DOT_PRODUCT(block_values, block_values2)
            ENDIF
         END DO
         CALL cp_dbcsr_iterator_stop(iter)

         !C_mat part

         CALL cp_dbcsr_iterator_start(iter, rho_im)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row_atom, col_atom, block_values)
            CALL cp_dbcsr_get_block_p(C_mat(i)%matrix, row_atom, col_atom, block_values2, found=found)
            IF (found) THEN
               ikind = kind_of(col_atom)
               kind_atom = atom_of_kind(col_atom)
               !The block_values are in a vector format, so the dot_product is
               ! the sum over all elements of the hamand product, that I need
               force(ikind)%ehrenfest(i, kind_atom) = force(ikind)%ehrenfest(i, kind_atom)+ &
                                                      2.0_dp*DOT_PRODUCT(block_values, block_values2)
            ENDIF
         END DO
         CALL cp_dbcsr_iterator_stop(iter)
      END DO

   END SUBROUTINE compute_forces

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE rt_admm_force(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rt_admm_force', routineP = moduleN//':'//routineN

      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: KS_aux_im, KS_aux_re, matrix_s_aux_fit, &
                                                            matrix_s_aux_fit_vs_orb
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos, mos_admm
      TYPE(rt_prop_type), POINTER                        :: rtp

      CALL get_qs_env(qs_env, &
                      admm_env=admm_env, &
                      rtp=rtp, &
                      matrix_ks_aux_fit=KS_aux_re, &
                      matrix_ks_aux_fit_im=KS_aux_im, &
                      matrix_s_aux_fit=matrix_s_aux_fit, &
                      matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb)

      CALL get_rtp(rtp=rtp, mos_new=mos, admm_mos=mos_admm)

      ! currently only none option
      CALL rt_admm_forces_none(qs_env, admm_env, KS_aux_re, KS_aux_im, &
                               matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, mos_admm, mos)

   END SUBROUTINE rt_admm_force

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param admm_env ...
!> \param KS_aux_re ...
!> \param KS_aux_im ...
!> \param matrix_s_aux_fit ...
!> \param matrix_s_aux_fit_vs_orb ...
!> \param mos_admm ...
!> \param mos ...
! **************************************************************************************************
   SUBROUTINE rt_admm_forces_none(qs_env, admm_env, KS_aux_re, KS_aux_im, matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, mos_admm, mos)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: KS_aux_re, KS_aux_im, matrix_s_aux_fit, &
                                                            matrix_s_aux_fit_vs_orb
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_admm, mos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rt_admm_forces_none', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: im, ispin, nao, natom, naux, nmo, re
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: admm_force
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_w_q, matrix_w_s
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: tmp_aux_aux, tmp_aux_mo, tmp_aux_mo1, &
                                                            tmp_aux_nao
      TYPE(cp_fm_struct_type), POINTER                   :: mstruct
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_aux_fit_asymm, sab_aux_fit_vs_orb
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      NULLIFY (sab_aux_fit_asymm, sab_aux_fit_vs_orb, ks_env)
!   CALL cp_fm_create(tmp_aux_aux,admm_env%fm_struct_tmp,name="fm matrix")

      CALL get_qs_env(qs_env, &
                      sab_aux_fit_asymm=sab_aux_fit_asymm, &
                      sab_aux_fit_vs_orb=sab_aux_fit_vs_orb, &
                      ks_env=ks_env)

      ALLOCATE (matrix_w_s)
      CALL cp_dbcsr_init(matrix_w_s)
      CALL cp_dbcsr_create(matrix_w_s, 'W MATRIX AUX S', &
                           cp_dbcsr_distribution(matrix_s_aux_fit(1)%matrix), dbcsr_type_no_symmetry, &
                           cp_dbcsr_row_block_sizes(matrix_s_aux_fit(1)%matrix), &
                           cp_dbcsr_col_block_sizes(matrix_s_aux_fit(1)%matrix), &
                           cp_dbcsr_get_data_size(matrix_s_aux_fit(1)%matrix), &
                           cp_dbcsr_get_data_type(matrix_s_aux_fit(1)%matrix))
      CALL cp_dbcsr_alloc_block_from_nbl(matrix_w_s, sab_aux_fit_asymm)

      ALLOCATE (matrix_w_q)
      CALL cp_dbcsr_init(matrix_w_q)
      CALL cp_dbcsr_copy(matrix_w_q, matrix_s_aux_fit_vs_orb(1)%matrix, &
                         "W MATRIX AUX Q")

      DO ispin = 1, SIZE(KS_aux_re)
         re = 2*ispin-1; im = 2*ispin
         naux = admm_env%nao_aux_fit; nmo = admm_env%nmo(ispin); nao = admm_env%nao_orb

         ALLOCATE (tmp_aux_aux(2), tmp_aux_nao(2), tmp_aux_mo(2), tmp_aux_mo1(2))
         CALL cp_fm_create(tmp_aux_aux(1)%matrix, admm_env%work_aux_aux%matrix_struct, name="taa")
         CALL cp_fm_create(tmp_aux_aux(2)%matrix, admm_env%work_aux_aux%matrix_struct, name="taa")
         CALL cp_fm_create(tmp_aux_nao(1)%matrix, admm_env%work_aux_orb%matrix_struct, name="tao")
         CALL cp_fm_create(tmp_aux_nao(2)%matrix, admm_env%work_aux_orb%matrix_struct, name="tao")
         mstruct => admm_env%work_aux_nmo(ispin)%matrix%matrix_struct
         CALL cp_fm_create(tmp_aux_mo(1)%matrix, mstruct, name="tam")
         CALL cp_fm_create(tmp_aux_mo(2)%matrix, mstruct, name="tam")
         CALL cp_fm_create(tmp_aux_mo1(1)%matrix, mstruct, name="tam")
         CALL cp_fm_create(tmp_aux_mo1(2)%matrix, mstruct, name="tam")

! First calculate H=KS_aux*C~, real part ends on work_aux_aux2, imaginary part ends at work_aux_aux3
         CALL cp_dbcsr_sm_fm_multiply(KS_aux_re(ispin)%matrix, mos_admm(re)%matrix, tmp_aux_mo(re)%matrix, nmo, 4.0_dp, 0.0_dp)
         CALL cp_dbcsr_sm_fm_multiply(KS_aux_re(ispin)%matrix, mos_admm(im)%matrix, tmp_aux_mo(im)%matrix, nmo, 4.0_dp, 0.0_dp)
         CALL cp_dbcsr_sm_fm_multiply(KS_aux_im(ispin)%matrix, mos_admm(im)%matrix, tmp_aux_mo(re)%matrix, nmo, -4.0_dp, 1.0_dp)
         CALL cp_dbcsr_sm_fm_multiply(KS_aux_im(ispin)%matrix, mos_admm(re)%matrix, tmp_aux_mo(im)%matrix, nmo, 4.0_dp, 1.0_dp)

! Next step compute S-1*H
         CALL cp_gemm('N', 'N', naux, nmo, naux, 1.0_dp, admm_env%S_inv, tmp_aux_mo(re)%matrix, 0.0_dp, tmp_aux_mo1(re)%matrix)
         CALL cp_gemm('N', 'N', naux, nmo, naux, 1.0_dp, admm_env%S_inv, tmp_aux_mo(im)%matrix, 0.0_dp, tmp_aux_mo1(im)%matrix)

! Here we go on with Ws=S-1*H * C^H (take care of sign of the imaginary part!!!)

         CALL cp_gemm("N", "T", naux, nao, nmo, -1.0_dp, tmp_aux_mo1(re)%matrix, mos(re)%matrix, 0.0_dp, &
                      tmp_aux_nao(re)%matrix)
         CALL cp_gemm("N", "T", naux, nao, nmo, -1.0_dp, tmp_aux_mo1(im)%matrix, mos(im)%matrix, 1.0_dp, &
                      tmp_aux_nao(re)%matrix)
         CALL cp_gemm("N", "T", naux, nao, nmo, 1.0_dp, tmp_aux_mo1(re)%matrix, mos(im)%matrix, 0.0_dp, &
                      tmp_aux_nao(im)%matrix)
         CALL cp_gemm("N", "T", naux, nao, nmo, -1.0_dp, tmp_aux_mo1(im)%matrix, mos(re)%matrix, 1.0_dp, &
                      tmp_aux_nao(im)%matrix)

! Let's do the final bit  Wq=S-1*H * C^H * A^T
         CALL cp_gemm('N', 'T', naux, naux, nao, -1.0_dp, tmp_aux_nao(re)%matrix, admm_env%A, 0.0_dp, tmp_aux_aux(re)%matrix)
         CALL cp_gemm('N', 'T', naux, naux, nao, -1.0_dp, tmp_aux_nao(im)%matrix, admm_env%A, 0.0_dp, tmp_aux_aux(im)%matrix)

         ! *** copy to sparse matrix
         CALL copy_fm_to_dbcsr(tmp_aux_nao(re)%matrix, matrix_w_q, keep_sparsity=.TRUE.)

         ! *** copy to sparse matrix
         CALL copy_fm_to_dbcsr(tmp_aux_aux(re)%matrix, matrix_w_s, keep_sparsity=.TRUE.)

! *** This can be done in one call w_total = w_alpha + w_beta
         ! allocate force vector
         CALL get_qs_env(qs_env=qs_env, natom=natom)
         ALLOCATE (admm_force(3, natom))
         admm_force = 0.0_dp
         CALL build_overlap_force(ks_env, admm_force, &
                                  basis_type_a="AUX_FIT", basis_type_b="AUX_FIT", &
                                  sab_nl=sab_aux_fit_asymm, matrix_p=matrix_w_s)
         CALL build_overlap_force(ks_env, admm_force, &
                                  basis_type_a="AUX_FIT", basis_type_b="ORB", &
                                  sab_nl=sab_aux_fit_vs_orb, matrix_p=matrix_w_q)
         ! add forces
         CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, &
                         force=force)
         CALL add_qs_force(admm_force, force, "overlap_admm", atomic_kind_set)
         DEALLOCATE (admm_force)

         ! *** Deallocated weighted density matrices
         CALL cp_dbcsr_deallocate_matrix(matrix_w_s)
         CALL cp_dbcsr_deallocate_matrix(matrix_w_q)
         CALL cp_fm_vect_dealloc(tmp_aux_aux)
         CALL cp_fm_vect_dealloc(tmp_aux_nao)
         CALL cp_fm_vect_dealloc(tmp_aux_mo)
         CALL cp_fm_vect_dealloc(tmp_aux_mo1)
      END DO

   END SUBROUTINE rt_admm_forces_none

END MODULE rt_propagation_forces
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Quickstep force driver routine
!> \author MK (12.06.2002)
! **************************************************************************************************
MODULE qs_force
   USE admm_methods,                    ONLY: calc_aux_mo_derivs_none,&
                                              calc_mixed_overlap_force
   USE admm_types,                      ONLY: admm_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_set
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_fm_types,                     ONLY: cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE dft_plus_u,                      ONLY: plus_u
   USE efield_utils,                    ONLY: calculate_ecore_efield
   USE input_constants,                 ONLY: do_admm_purify_none
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              int_size
   USE message_passing,                 ONLY: mp_sum
   USE mulliken,                        ONLY: mulliken_restraint
   USE particle_types,                  ONLY: particle_type
   USE qs_core_energies,                ONLY: calculate_ecore_overlap,&
                                              calculate_ecore_self
   USE qs_core_hamiltonian,             ONLY: build_core_hamiltonian_matrix
   USE qs_dftb_dispersion,              ONLY: calculate_dftb_dispersion
   USE qs_dftb_matrices,                ONLY: build_dftb_matrices
   USE qs_energy,                       ONLY: qs_energies
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_methods,          ONLY: qs_env_rebuild_pw_env
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_external_potential,           ONLY: external_c_potential,&
                                              external_e_potential
   USE qs_force_types,                  ONLY: allocate_qs_force,&
                                              qs_force_type,&
                                              zero_qs_force
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change,&
                                              qs_ks_env_type,&
                                              set_ks_env
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type
   USE qs_rho_methods,                  ONLY: qs_rho_update_rho
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_scf_initialization,           ONLY: qs_scf_compute_properties
   USE qs_subsys_types,                 ONLY: qs_subsys_set,&
                                              qs_subsys_type
   USE rt_propagation_forces,           ONLY: calc_c_mat_force,&
                                              rt_admm_force
   USE scptb_core_interactions,         ONLY: scptb_core_interaction
   USE scptb_core_matrix,               ONLY: build_scptb_core_matrix
   USE se_core_core,                    ONLY: se_core_core_interaction
   USE se_core_matrix,                  ONLY: build_se_core_matrix
   USE termination,                     ONLY: stop_memory
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_force'

! *** Public subroutines ***

   PUBLIC :: qs_calc_energy_force

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param calc_force ...
!> \param consistent_energies ...
!> \param linres ...
! **************************************************************************************************
   SUBROUTINE qs_calc_energy_force(qs_env, calc_force, consistent_energies, linres)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL                                            :: calc_force, consistent_energies, linres

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_calc_energy_force', &
         routineP = moduleN//':'//routineN

      qs_env%linres_run = linres
      CALL set_qs_env(qs_env)
      IF (calc_force) THEN
         CALL qs_forces(qs_env)
      ELSE
         CALL qs_energies(qs_env, calc_forces=.FALSE., &
                          consistent_energies=consistent_energies)
      END IF
      CALL get_qs_env(qs_env)
   END SUBROUTINE qs_calc_energy_force

! **************************************************************************************************
!> \brief   Calculate the Quickstep forces.
!> \param qs_env ...
!> \date    29.10.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE qs_forces(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_forces', routineP = moduleN//':'//routineN

      INTEGER                                            :: after, dir, handle, i, iatom, ic, ikind, &
                                                            ispin, iw, natom, nkind, nspin, &
                                                            output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of, natom_of_kind
      LOGICAL                                            :: has_unit_metric, omit_headers
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: matrix_ks_aux_fit, matrix_p_mp2, matrix_s, &
         matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, matrix_w, matrix_w_mp2, rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_w_kp
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_aux_fit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_aux_fit
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: dft_section, input, print_section
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! rebuild plane wave environment
      CALL qs_env_rebuild_pw_env(qs_env)

      ! zero out the forces in particle set
      CALL get_qs_env(qs_env, particle_set=particle_set)
      natom = SIZE(particle_set)
      DO iatom = 1, natom
         particle_set(iatom)%f = 0.0_dp
      END DO

      ! get atom mapping
      NULLIFY (atomic_kind_set)
      CALL get_qs_env(qs_env, atomic_kind_set=atomic_kind_set)
      ALLOCATE (atom_of_kind(natom), kind_of(natom))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               atom_of_kind=atom_of_kind, &
                               kind_of=kind_of)

      NULLIFY (force, subsys)
      CALL get_qs_env(qs_env, &
                      force=force, &
                      subsys=subsys)
      IF (.NOT. ASSOCIATED(force)) THEN
         !   *** Allocate the force data structure ***
         nkind = SIZE(atomic_kind_set)
         ALLOCATE (natom_of_kind(nkind))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  natom_of_kind=natom_of_kind)
         CALL allocate_qs_force(force, natom_of_kind)
         DEALLOCATE (natom_of_kind)
         CALL qs_subsys_set(subsys, force=force)
      END IF
      CALL zero_qs_force(force)

      ! Set parameter for P screening with MP2
      IF (ASSOCIATED(qs_env%mp2_env)) qs_env%mp2_env%not_last_hfx = .TRUE.

      ! recalculate energy with forces
      CALL qs_energies(qs_env, calc_forces=.TRUE.)

      NULLIFY (dft_control, para_env)
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      para_env=para_env)

      ! Now we handle some special cases
      ! Maybe some of these would be better dealt with in qs_energies?
      IF (qs_env%run_rtp) THEN
         NULLIFY (matrix_w, matrix_s, ks_env)
         CALL get_qs_env(qs_env, &
                         ks_env=ks_env, &
                         matrix_w=matrix_w, &
                         matrix_s=matrix_s)
         CALL cp_dbcsr_allocate_matrix_set(matrix_w, dft_control%nspins)
         DO ispin = 1, dft_control%nspins
            ALLOCATE (matrix_w(ispin)%matrix)
            CALL cp_dbcsr_init(matrix_w(ispin)%matrix)
            CALL cp_dbcsr_copy(matrix_w(ispin)%matrix, matrix_s(1)%matrix, &
                               name="W MATRIX")
            CALL cp_dbcsr_set(matrix_w(ispin)%matrix, 0.0_dp)
         END DO
         CALL set_ks_env(ks_env, matrix_w=matrix_w)

         CALL calc_c_mat_force(qs_env)
         IF (dft_control%do_admm) CALL rt_admm_force(qs_env)
      END IF
      ! from an eventual Mulliken restraint
      IF (dft_control%qs_control%mulliken_restraint) THEN
         NULLIFY (matrix_w, matrix_s, rho)
         CALL get_qs_env(qs_env, &
                         matrix_w=matrix_w, &
                         matrix_s=matrix_s, &
                         rho=rho)
         NULLIFY (rho_ao)
         CALL qs_rho_get(rho, rho_ao=rho_ao)
         CALL mulliken_restraint(dft_control%qs_control%mulliken_restraint_control, &
                                 para_env, matrix_s(1)%matrix, rho_ao, w_matrix=matrix_w)
      END IF
      ! Add non-Pulay contribution of DFT+U to W matrix, since it has also to be
      ! digested with overlap matrix derivatives
      IF (dft_control%dft_plus_u) THEN
         NULLIFY (matrix_w)
         CALL get_qs_env(qs_env, matrix_w=matrix_w)
         CALL plus_u(qs_env=qs_env, matrix_w=matrix_w)
      END IF

      ! Write W Matrix to output (if requested)
      CALL get_qs_env(qs_env, has_unit_metric=has_unit_metric)
      IF (.NOT. has_unit_metric) THEN
         NULLIFY (matrix_w_kp)
         CALL get_qs_env(qs_env, matrix_w_kp=matrix_w_kp)
         nspin = SIZE(matrix_w_kp, 1)
         DO ispin = 1, nspin
            IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                                 qs_env%input, "DFT%PRINT%AO_MATRICES/W_MATRIX"), cp_p_file)) THEN
               iw = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%AO_MATRICES/W_MATRIX", &
                                         extension=".Log")
               CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%NDIGITS", i_val=after)
               CALL section_vals_val_get(qs_env%input, "DFT%PRINT%AO_MATRICES%OMIT_HEADERS", l_val=omit_headers)
               after = MIN(MAX(after, 1), 16)
               DO ic = 1, SIZE(matrix_w_kp, 2)
                  CALL cp_dbcsr_write_sparse_matrix(matrix_w_kp(ispin, ic)%matrix, 4, after, qs_env, &
                                                    para_env, output_unit=iw, omit_headers=omit_headers)
               END DO
               CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                                 "DFT%PRINT%AO_MATRICES/W_MATRIX")
            END IF
         END DO
      ENDIF

      ! Compute core forces (also overwrites matrix_w)
      IF (dft_control%qs_control%semi_empirical) THEN
         CALL build_se_core_matrix(qs_env=qs_env, para_env=para_env, &
                                   calculate_forces=.TRUE.)
         CALL se_core_core_interaction(qs_env, para_env, calculate_forces=.TRUE.)
      ELSEIF (dft_control%qs_control%dftb) THEN
         CALL build_dftb_matrices(qs_env=qs_env, para_env=para_env, &
                                  calculate_forces=.TRUE.)
         CALL calculate_dftb_dispersion(qs_env=qs_env, para_env=para_env, &
                                        calculate_forces=.TRUE.)
      ELSEIF (dft_control%qs_control%scptb) THEN
         CALL build_scptb_core_matrix(qs_env=qs_env, calculate_forces=.TRUE.)
         CALL scptb_core_interaction(qs_env, calculate_forces=.TRUE.)
      ELSE
         CALL build_core_hamiltonian_matrix(qs_env=qs_env, calculate_forces=.TRUE.)
         CALL calculate_ecore_self(qs_env)
         CALL calculate_ecore_overlap(qs_env, para_env, calculate_forces=.TRUE.)
         CALL calculate_ecore_efield(qs_env, calculate_forces=.TRUE.)
         !swap external_e_potential before external_c_potential, to ensure
         !that external potential on grid is loaded before calculating energy of cores
         CALL external_e_potential(qs_env)
         IF (.NOT. dft_control%qs_control%gapw) THEN
            CALL external_c_potential(qs_env, calculate_forces=.TRUE.)
         END IF
      END IF

      ! Compute grid-based forces
      CALL qs_ks_update_qs_env(qs_env, calculate_forces=.TRUE.)

      ! MP2 Code
      IF (ASSOCIATED(qs_env%mp2_env)) THEN
         NULLIFY (matrix_p_mp2, matrix_w_mp2, rho, ks_env, energy)
         CALL get_qs_env(qs_env, &
                         matrix_p_mp2=matrix_p_mp2, &
                         matrix_w_mp2=matrix_w_mp2, &
                         ks_env=ks_env, &
                         rho=rho, &
                         energy=energy)
         NULLIFY (rho_ao)
         CALL qs_rho_get(rho, rho_ao=rho_ao)

         ! with MP2 we have to recalculate the SCF energy with the
         ! correct density
         DO ispin = 1, dft_control%nspins
            CALL cp_dbcsr_add(rho_ao(ispin)%matrix, matrix_p_mp2(ispin)%matrix, 1.0_dp, -1.0_dp)
         END DO
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)
         CALL qs_ks_update_qs_env(qs_env, just_energy=.TRUE.)
         energy%total = energy%total+energy%mp2

         ! Compute MP2 properties
         CALL get_qs_env(qs_env=qs_env, input=input)
         dft_section => section_vals_get_subs_vals(input, "DFT")
         ! Get the HF+MP2 density
         DO ispin = 1, dft_control%nspins
            CALL cp_dbcsr_add(rho_ao(ispin)%matrix, matrix_p_mp2(ispin)%matrix, 1.0_dp, 1.0_dp)
         END DO
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)
         CALL qs_scf_compute_properties(qs_env, dft_section, wf_type='MP2   ')
         ! Get everything back
         DO ispin = 1, dft_control%nspins
            CALL cp_dbcsr_add(rho_ao(ispin)%matrix, matrix_p_mp2(ispin)%matrix, 1.0_dp, -1.0_dp)
         END DO
         CALL qs_rho_update_rho(rho, qs_env=qs_env)
         CALL qs_ks_did_change(ks_env, rho_changed=.TRUE.)

         ! deallocate mp2_W
         CALL cp_dbcsr_deallocate_matrix_set(matrix_w_mp2)
         CALL set_ks_env(ks_env, matrix_w_mp2=Null())

      END IF

      ! Add forces resulting from wavefunction fitting
      IF (dft_control%do_admm_dm) THEN
         CALL cp__b("qs_force.F",350,"Forces with ADMM DM methods not implemented")
      END IF
      IF (dft_control%do_admm_mo .AND. .NOT. qs_env%run_rtp) THEN
         NULLIFY (matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, matrix_ks_aux_fit, &
                  mos_aux_fit, mos, admm_env)
         CALL get_qs_env(qs_env=qs_env, &
                         matrix_s_aux_fit=matrix_s_aux_fit, &
                         matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb, &
                         matrix_ks_aux_fit=matrix_ks_aux_fit, &
                         mos_aux_fit=mos_aux_fit, &
                         mos=mos, &
                         admm_env=admm_env)
         DO ispin = 1, dft_control%nspins
            mo_set => mos(ispin)%mo_set
            CALL get_mo_set(mo_set=mo_set, mo_coeff=mo_coeff)
            ! if no purification we need to calculate the H matrix for forces
            IF (admm_env%purification_method == do_admm_purify_none) THEN
               CALL get_mo_set(mo_set=mos_aux_fit(ispin)%mo_set, mo_coeff=mo_coeff_aux_fit)
               CALL calc_aux_mo_derivs_none(ispin, qs_env%admm_env, mo_set, &
                                            mo_coeff_aux_fit, matrix_ks_aux_fit)
            END IF
         END DO
         CALL calc_mixed_overlap_force(qs_env)
      END IF

      !  *** replicate forces ***
      DO ikind = 1, SIZE(force)
         CALL mp_sum(force(ikind)%overlap, para_env%group)
         CALL mp_sum(force(ikind)%overlap_admm, para_env%group)
         CALL mp_sum(force(ikind)%kinetic, para_env%group)
         CALL mp_sum(force(ikind)%gth_ppl, para_env%group)
         CALL mp_sum(force(ikind)%gth_nlcc, para_env%group)
         CALL mp_sum(force(ikind)%gth_ppnl, para_env%group)
         CALL mp_sum(force(ikind)%all_potential, para_env%group)
         CALL mp_sum(force(ikind)%core_overlap, para_env%group)
         CALL mp_sum(force(ikind)%rho_core, para_env%group)
         CALL mp_sum(force(ikind)%rho_elec, para_env%group)
         CALL mp_sum(force(ikind)%rho_lri_elec, para_env%group)
         CALL mp_sum(force(ikind)%vhxc_atom, para_env%group)
         CALL mp_sum(force(ikind)%g0s_Vh_elec, para_env%group)
         CALL mp_sum(force(ikind)%fock_4c, para_env%group)
         CALL mp_sum(force(ikind)%mp2_non_sep, para_env%group)
         CALL mp_sum(force(ikind)%mp2_sep, para_env%group)
         CALL mp_sum(force(ikind)%repulsive, para_env%group)
         CALL mp_sum(force(ikind)%dispersion, para_env%group)
         CALL mp_sum(force(ikind)%ehrenfest, para_env%group)

         force(ikind)%total(:, :) = force(ikind)%total(:, :)+ &
                                    force(ikind)%core_overlap(:, :)+ &
                                    force(ikind)%gth_ppl(:, :)+ &
                                    force(ikind)%gth_nlcc(:, :)+ &
                                    force(ikind)%gth_ppnl(:, :)+ &
                                    force(ikind)%all_potential(:, :)+ &
                                    force(ikind)%kinetic(:, :)+ &
                                    force(ikind)%overlap(:, :)+ &
                                    force(ikind)%overlap_admm(:, :)+ &
                                    force(ikind)%rho_core(:, :)+ &
                                    force(ikind)%rho_elec(:, :)+ &
                                    force(ikind)%rho_lri_elec(:, :)+ &
                                    force(ikind)%vhxc_atom(:, :)+ &
                                    force(ikind)%g0s_Vh_elec(:, :)+ &
                                    force(ikind)%fock_4c(:, :)+ &
                                    force(ikind)%mp2_non_sep(:, :)+ &
                                    force(ikind)%mp2_sep(:, :)+ &
                                    force(ikind)%repulsive(:, :)+ &
                                    force(ikind)%dispersion(:, :)+ &
                                    force(ikind)%ehrenfest(:, :)+ &
                                    force(ikind)%efield(:, :)+ &
                                    force(ikind)%eev(:, :)

      END DO

      DO iatom = 1, natom
         ikind = kind_of(iatom)
         i = atom_of_kind(iatom)
         ! XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
         ! the force is - dE/dR, what is called force is actually the gradient
         ! Things should have the right name
         ! The minus sign below is a hack
         ! XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
         force(ikind)%other(1:3, i) = -particle_set(iatom)%f(1:3)+force(ikind)%ch_pulay(1:3, i)
         force(ikind)%total(1:3, i) = force(ikind)%total(1:3, i)+force(ikind)%other(1:3, i)
         particle_set(iatom)%f = -force(ikind)%total(1:3, i)
      END DO

      NULLIFY (virial, energy)
      CALL get_qs_env(qs_env=qs_env, virial=virial, energy=energy)
      !   *** distribute virial ***
      IF (virial%pv_availability) THEN
         CALL mp_sum(virial%pv_virial, para_env%group)
         !  *** add the volume terms of the virial ***
         IF ((.NOT. virial%pv_numer) .AND. &
             (.NOT. (dft_control%qs_control%dftb .OR. &
                     dft_control%qs_control%scptb .OR. &
                     dft_control%qs_control%semi_empirical))) THEN
            DO dir = 1, 3
               virial%pv_virial(dir, dir) = virial%pv_virial(dir, dir)-energy%exc &
                                            -2.0_dp*energy%hartree
               IF (dft_control%do_admm) THEN
                  virial%pv_virial(dir, dir) = virial%pv_virial(dir, dir)-energy%exc_aux_fit
               END IF
               ! The factor 2 is a hack. It compensates the plus sign in h_stress/pw_poisson_solve.
               ! The sign in pw_poisson_solve is correct for FIST, but not for QS.
               ! There should be a more elegant solution to that...
            END DO
         END IF
      END IF

      output_unit = cp_print_key_unit_nr(logger, qs_env%input, "DFT%PRINT%DERIVATIVES", &
                                         extension=".Log")
      print_section => section_vals_get_subs_vals(qs_env%input, "DFT%PRINT%DERIVATIVES")
      IF (dft_control%qs_control%semi_empirical) THEN
         CALL write_forces(force, atomic_kind_set, 2, output_unit=output_unit, &
                           print_section=print_section)
      ELSE IF (dft_control%qs_control%dftb) THEN
         CALL write_forces(force, atomic_kind_set, 4, output_unit=output_unit, &
                           print_section=print_section)
      ELSE IF (dft_control%qs_control%gapw) THEN
         CALL write_forces(force, atomic_kind_set, 1, output_unit=output_unit, &
                           print_section=print_section)
      ELSE IF (dft_control%qs_control%scptb) THEN
         CALL write_forces(force, atomic_kind_set, 5, output_unit=output_unit, &
                           print_section=print_section)
      ELSE
         CALL write_forces(force, atomic_kind_set, 0, output_unit=output_unit, &
                           print_section=print_section)
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, qs_env%input, &
                                        "DFT%PRINT%DERIVATIVES")

      ! deallocate W Matrix:
      NULLIFY (ks_env, matrix_w_kp)
      CALL get_qs_env(qs_env=qs_env, &
                      matrix_w_kp=matrix_w_kp, &
                      ks_env=ks_env)
      CALL cp_dbcsr_deallocate_matrix_set(matrix_w_kp)
      CALL set_ks_env(ks_env, matrix_w=Null(), matrix_w_kp=Null())

      DEALLOCATE (atom_of_kind, kind_of)

      CALL timestop(handle)

   END SUBROUTINE qs_forces

! **************************************************************************************************
!> \brief   Write a Quickstep force data structure to output unit
!> \param qs_force ...
!> \param atomic_kind_set ...
!> \param ftype ...
!> \param output_unit ...
!> \param print_section ...
!> \date    05.06.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_forces(qs_force, atomic_kind_set, ftype, output_unit, &
                           print_section)

      TYPE(qs_force_type), DIMENSION(:), POINTER         :: qs_force
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      INTEGER, INTENT(IN)                                :: ftype, output_unit
      TYPE(section_vals_type), POINTER                   :: print_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_forces', routineP = moduleN//':'//routineN

      CHARACTER(LEN=13)                                  :: fmtstr5
      CHARACTER(LEN=15)                                  :: fmtstr4
      CHARACTER(LEN=20)                                  :: fmtstr3
      CHARACTER(LEN=35)                                  :: fmtstr2
      CHARACTER(LEN=48)                                  :: fmtstr1
      INTEGER                                            :: i, iatom, ikind, istat, my_ftype, natom, &
                                                            ndigits
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of
      REAL(KIND=dp), DIMENSION(3)                        :: grand_total

      IF (output_unit > 0) THEN

         IF (.NOT. ASSOCIATED(qs_force)) THEN
            CALL cp_abort(cp__l("qs_force.F",528), &
                          "The qs_force pointer is not associated "// &
                          "and cannot be printed")
         END IF

         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  natom=natom)
         ALLOCATE (atom_of_kind(natom), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 536, &
                                          "atom_of_kind", natom*int_size)
         ALLOCATE (kind_of(natom), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 539, &
                                          "kind_of", natom*int_size)
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  atom_of_kind=atom_of_kind, &
                                  kind_of=kind_of)

         ! Variable precision output of the forces
         CALL section_vals_val_get(print_section, "NDIGITS", &
                                   i_val=ndigits)

         fmtstr1 = "(/,/,T2,A,/,/,T3,A,T11,A,T23,A,T40,A1,2(  X,A1))"
         WRITE (UNIT=fmtstr1(41:42), FMT="(I2)") ndigits+5

         fmtstr2 = "(/,(T2,I5,4X,I4,T18,A,T34,3F  .  ))"
         WRITE (UNIT=fmtstr2(32:33), FMT="(I2)") ndigits
         WRITE (UNIT=fmtstr2(29:30), FMT="(I2)") ndigits+6

         fmtstr3 = "(/,T3,A,T34,3F  .  )"
         WRITE (UNIT=fmtstr3(18:19), FMT="(I2)") ndigits
         WRITE (UNIT=fmtstr3(15:16), FMT="(I2)") ndigits+6

         fmtstr4 = "((T34,3F  .  ))"
         WRITE (UNIT=fmtstr4(12:13), FMT="(I2)") ndigits
         WRITE (UNIT=fmtstr4(9:10), FMT="(I2)") ndigits+6

         fmtstr5 = "(/T2,A//T3,A)"

         WRITE (UNIT=output_unit, FMT=fmtstr1) &
            "FORCES [a.u.]", "Atom", "Kind", "Component", "X", "Y", "Z"

         grand_total(:) = 0.0_dp

         my_ftype = ftype

         SELECT CASE (my_ftype)
         CASE DEFAULT
            DO iatom = 1, natom
               ikind = kind_of(iatom)
               i = atom_of_kind(iatom)
               WRITE (UNIT=output_unit, FMT=fmtstr2) &
                  iatom, ikind, "         total", qs_force(ikind)%total(1:3, i)
               grand_total(1:3) = grand_total(1:3)+qs_force(ikind)%total(1:3, i)
            END DO
         CASE (0)
            DO iatom = 1, natom
               ikind = kind_of(iatom)
               i = atom_of_kind(iatom)
               WRITE (UNIT=output_unit, FMT=fmtstr2) &
                  iatom, ikind, "       overlap", qs_force(ikind)%overlap(1:3, i), &
                  iatom, ikind, "  overlap_admm", qs_force(ikind)%overlap_admm(1:3, i), &
                  iatom, ikind, "       kinetic", qs_force(ikind)%kinetic(1:3, i), &
                  iatom, ikind, "       gth_ppl", qs_force(ikind)%gth_ppl(1:3, i), &
                  iatom, ikind, "      gth_nlcc", qs_force(ikind)%gth_nlcc(1:3, i), &
                  iatom, ikind, "      gth_ppnl", qs_force(ikind)%gth_ppnl(1:3, i), &
                  iatom, ikind, "  core_overlap", qs_force(ikind)%core_overlap(1:3, i), &
                  iatom, ikind, "      rho_core", qs_force(ikind)%rho_core(1:3, i), &
                  iatom, ikind, "      rho_elec", qs_force(ikind)%rho_elec(1:3, i), &
                  iatom, ikind, "      rho_lri_elec", qs_force(ikind)%rho_lri_elec(1:3, i), &
                  iatom, ikind, "      ch_pulay", qs_force(ikind)%ch_pulay(1:3, i), &
                  iatom, ikind, "    dispersion", qs_force(ikind)%dispersion(1:3, i), &
                  iatom, ikind, "         other", qs_force(ikind)%other(1:3, i), &
                  iatom, ikind, "       fock_4c", qs_force(ikind)%fock_4c(1:3, i), &
                  iatom, ikind, "     ehrenfest", qs_force(ikind)%ehrenfest(1:3, i), &
                  iatom, ikind, "        efield", qs_force(ikind)%efield(1:3, i), &
                  iatom, ikind, "           eev", qs_force(ikind)%eev(1:3, i), &
                  iatom, ikind, "   mp2_non_sep", qs_force(ikind)%mp2_non_sep(1:3, i), &
                  iatom, ikind, "       mp2_sep", qs_force(ikind)%mp2_sep(1:3, i), &
                  iatom, ikind, "         total", qs_force(ikind)%total(1:3, i)
               grand_total(1:3) = grand_total(1:3)+qs_force(ikind)%total(1:3, i)
            END DO
         CASE (1)
            DO iatom = 1, natom
               ikind = kind_of(iatom)
               i = atom_of_kind(iatom)
               WRITE (UNIT=output_unit, FMT=fmtstr2) &
                  iatom, ikind, "       overlap", qs_force(ikind)%overlap(1:3, i), &
                  iatom, ikind, "  overlap_admm", qs_force(ikind)%overlap_admm(1:3, i), &
                  iatom, ikind, "       kinetic", qs_force(ikind)%kinetic(1:3, i), &
                  iatom, ikind, "       gth_ppl", qs_force(ikind)%gth_ppl(1:3, i), &
                  iatom, ikind, "      gth_nlcc", qs_force(ikind)%gth_nlcc(1:3, i), &
                  iatom, ikind, "      gth_ppnl", qs_force(ikind)%gth_ppnl(1:3, i), &
                  iatom, ikind, " all_potential", qs_force(ikind)%all_potential(1:3, i), &
                  iatom, ikind, "  core_overlap", qs_force(ikind)%core_overlap(1:3, i), &
                  iatom, ikind, "      rho_core", qs_force(ikind)%rho_core(1:3, i), &
                  iatom, ikind, "      rho_elec", qs_force(ikind)%rho_elec(1:3, i), &
                  iatom, ikind, "      rho_lri_elec", qs_force(ikind)%rho_lri_elec(1:3, i), &
                  iatom, ikind, "     vhxc_atom", qs_force(ikind)%vhxc_atom(1:3, i), &
                  iatom, ikind, "   g0s_Vh_elec", qs_force(ikind)%g0s_Vh_elec(1:3, i), &
                  iatom, ikind, "      ch_pulay", qs_force(ikind)%ch_pulay(1:3, i), &
                  iatom, ikind, "    dispersion", qs_force(ikind)%dispersion(1:3, i), &
                  iatom, ikind, "       fock_4c", qs_force(ikind)%fock_4c(1:3, i), &
                  iatom, ikind, "        efield", qs_force(ikind)%efield(1:3, i), &
                  iatom, ikind, "           eev", qs_force(ikind)%eev(1:3, i), &
                  iatom, ikind, "   mp2_non_sep", qs_force(ikind)%mp2_non_sep(1:3, i), &
                  iatom, ikind, "       mp2_sep", qs_force(ikind)%mp2_sep(1:3, i), &
                  iatom, ikind, "         total", qs_force(ikind)%total(1:3, i)
               grand_total(1:3) = grand_total(1:3)+qs_force(ikind)%total(1:3, i)
            END DO
         CASE (2)
            DO iatom = 1, natom
               ikind = kind_of(iatom)
               i = atom_of_kind(iatom)
               WRITE (UNIT=output_unit, FMT=fmtstr2) &
                  iatom, ikind, " all_potential", qs_force(ikind)%all_potential(1:3, i), &
                  iatom, ikind, "      rho_elec", qs_force(ikind)%rho_elec(1:3, i), &
                  iatom, ikind, "      rho_lri_elec", qs_force(ikind)%rho_lri_elec(1:3, i), &
                  iatom, ikind, "         total", qs_force(ikind)%total(1:3, i)
               grand_total(1:3) = grand_total(1:3)+qs_force(ikind)%total(1:3, i)
            END DO
         CASE (3)
            DO iatom = 1, natom
               ikind = kind_of(iatom)
               i = atom_of_kind(iatom)
               WRITE (UNIT=output_unit, FMT=fmtstr2) &
                  iatom, ikind, "        overlap", qs_force(ikind)%overlap(1:3, i), &
                  iatom, ikind, "overlap_admm", qs_force(ikind)%overlap_admm(1:3, i), &
                  iatom, ikind, "        kinetic", qs_force(ikind)%kinetic(1:3, i), &
                  iatom, ikind, "        gth_ppl", qs_force(ikind)%gth_ppl(1:3, i), &
                  iatom, ikind, "       gth_nlcc", qs_force(ikind)%gth_nlcc(1:3, i), &
                  iatom, ikind, "       gth_ppnl", qs_force(ikind)%gth_ppnl(1:3, i), &
                  iatom, ikind, "   core_overlap", qs_force(ikind)%core_overlap(1:3, i), &
                  iatom, ikind, "       rho_core", qs_force(ikind)%rho_core(1:3, i), &
                  iatom, ikind, "       rho_elec", qs_force(ikind)%rho_elec(1:3, i), &
                  iatom, ikind, "       rho_lri_elec", qs_force(ikind)%rho_lri_elec(1:3, i), &
                  iatom, ikind, "       ch_pulay", qs_force(ikind)%ch_pulay(1:3, i), &
                  iatom, ikind, "        fock_4c", qs_force(ikind)%fock_4c(1:3, i), &
                  iatom, ikind, "   mp2_non_sep", qs_force(ikind)%mp2_non_sep(1:3, i), &
                  iatom, ikind, "       mp2_sep", qs_force(ikind)%mp2_sep(1:3, i), &
                  iatom, ikind, "          total", qs_force(ikind)%total(1:3, i)
               grand_total(1:3) = grand_total(1:3)+qs_force(ikind)%total(1:3, i)
            END DO
         CASE (4)
            DO iatom = 1, natom
               ikind = kind_of(iatom)
               i = atom_of_kind(iatom)
               WRITE (UNIT=output_unit, FMT=fmtstr2) &
                  iatom, ikind, "  all_potential", qs_force(ikind)%all_potential(1:3, i), &
                  iatom, ikind, "        overlap", qs_force(ikind)%overlap(1:3, i), &
                  iatom, ikind, "   overlap_admm", qs_force(ikind)%overlap_admm(1:3, i), &
                  iatom, ikind, "       rho_elec", qs_force(ikind)%rho_elec(1:3, i), &
                  iatom, ikind, "       rho_lri_elec", qs_force(ikind)%rho_lri_elec(1:3, i), &
                  iatom, ikind, "      repulsive", qs_force(ikind)%repulsive(1:3, i), &
                  iatom, ikind, "     dispersion", qs_force(ikind)%dispersion(1:3, i), &
                  iatom, ikind, "     ehrenfest", qs_force(ikind)%ehrenfest(1:3, i), &
                  iatom, ikind, "          total", qs_force(ikind)%total(1:3, i)
               grand_total(1:3) = grand_total(1:3)+qs_force(ikind)%total(1:3, i)
            END DO
         CASE (5)
            DO iatom = 1, natom
               ikind = kind_of(iatom)
               i = atom_of_kind(iatom)
               WRITE (UNIT=output_unit, FMT=fmtstr2) &
                  iatom, ikind, "       overlap", qs_force(ikind)%overlap(1:3, i), &
                  iatom, ikind, "       kinetic", qs_force(ikind)%kinetic(1:3, i), &
                  iatom, ikind, "      rho_elec", qs_force(ikind)%rho_elec(1:3, i), &
                  iatom, ikind, "      rho_lri_elec", qs_force(ikind)%rho_lri_elec(1:3, i), &
                  iatom, ikind, "    dispersion", qs_force(ikind)%dispersion(1:3, i), &
                  iatom, ikind, " all potential", qs_force(ikind)%all_potential(1:3, i), &
                  iatom, ikind, "         other", qs_force(ikind)%other(1:3, i), &
                  iatom, ikind, "         total", qs_force(ikind)%total(1:3, i)
               grand_total(1:3) = grand_total(1:3)+qs_force(ikind)%total(1:3, i)
            END DO
         END SELECT

         WRITE (UNIT=output_unit, FMT=fmtstr3) "Sum of total", grand_total(1:3)

         DEALLOCATE (atom_of_kind)
         DEALLOCATE (kind_of)

      END IF

   END SUBROUTINE write_forces

END MODULE qs_force
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      EAM potential
!> \author CJM, I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
MODULE manybody_eam

   USE bibliography,                    ONLY: Foiles1986,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                              neighbor_kind_pairs_type
   USE fist_nonbond_env_types,          ONLY: eam_type,&
                                              fist_nonbond_env_get,&
                                              fist_nonbond_env_set,&
                                              fist_nonbond_env_type,&
                                              pos_type
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: matvec_3x3
   USE message_passing,                 ONLY: mp_sum
   USE pair_potential_types,            ONLY: ea_type,&
                                              eam_pot_type,&
                                              pair_potential_pp_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: get_force_eam, density_nonbond
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'manybody_eam'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param fist_nonbond_env ...
!> \param particle_set ...
!> \param cell ...
!> \param para_env ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE density_nonbond(fist_nonbond_env, particle_set, cell, para_env)
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(particle_type), DIMENSION(:), INTENT(INOUT)   :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'density_nonbond', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, handle, i, i_a, i_b, &
                                                            iend, igrp, ikind, ilist, ipair, &
                                                            iparticle, istart, jkind, kind_a, &
                                                            kind_b, nkinds, nparticle
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: eam_kinds_index
      LOGICAL                                            :: do_eam
      REAL(KIND=dp)                                      :: fac, rab2, rab2_max
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v, rab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rho
      TYPE(eam_pot_type), POINTER                        :: eam_a, eam_b
      TYPE(eam_type), DIMENSION(:), POINTER              :: eam_data
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update, r_last_update_pbc

      CALL timeset(routineN, handle)
      do_eam = .FALSE.
      CALL fist_nonbond_env_get(fist_nonbond_env, nonbonded=nonbonded, &
                                potparm=potparm, r_last_update=r_last_update, &
                                r_last_update_pbc=r_last_update_pbc, eam_data=eam_data)
      nkinds = SIZE(potparm%pot, 1)
      ALLOCATE (eam_kinds_index(nkinds, nkinds))
      eam_kinds_index = -1
      DO ikind = 1, nkinds
         DO jkind = ikind, nkinds
            DO i = 1, SIZE(potparm%pot(ikind, jkind)%pot%type)
               IF (potparm%pot(ikind, jkind)%pot%type(i) == ea_type) THEN
                  ! At the moment we allow only 1 EAM per each kinds pair..
                  IF(.NOT.(eam_kinds_index(ikind, jkind) == -1))CALL cp__a("manybody_eam.F",87)
                  IF(.NOT.(eam_kinds_index(jkind, ikind) == -1))CALL cp__a("manybody_eam.F",88)
                  eam_kinds_index(ikind, jkind) = i
                  eam_kinds_index(jkind, ikind) = i
                  do_eam = .TRUE.
               END IF
            END DO
         END DO
      END DO

      nparticle = SIZE(particle_set)

      IF (do_eam) THEN
         IF (.NOT. ASSOCIATED(eam_data)) THEN
            ALLOCATE (eam_data(nparticle))
            CALL fist_nonbond_env_set(fist_nonbond_env, eam_data=eam_data)
         ENDIF
         DO i = 1, nparticle
            eam_data(i)%rho = 0.0_dp
            eam_data(i)%f_embed = 0.0_dp
         ENDDO
      ENDIF

      ! Only if EAM potential are present
      IF (do_eam) THEN
         ! Add citation
         CALL cite_reference(Foiles1986)
         NULLIFY (eam_a, eam_b)
         ALLOCATE (rho(nparticle))
         rho = 0._dp
         ! Starting the force loop
         DO ilist = 1, nonbonded%nlists
            neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
            IF (neighbor_kind_pair%npairs == 0) CYCLE
            Kind_Group_Loop: DO igrp = 1, neighbor_kind_pair%ngrp_kind
               istart = neighbor_kind_pair%grp_kind_start(igrp)
               iend = neighbor_kind_pair%grp_kind_end(igrp)
               ikind = neighbor_kind_pair%ij_kind(1, igrp)
               jkind = neighbor_kind_pair%ij_kind(2, igrp)

               i = eam_kinds_index(ikind, jkind)
               IF (i == -1) CYCLE
               rab2_max = potparm%pot(ikind, jkind)%pot%rcutsq
               CALL matvec_3x3(cell_v, cell%hmat, REAL(neighbor_kind_pair%cell_vector, KIND=dp))
               DO ipair = istart, iend
                  atom_a = neighbor_kind_pair%list(1, ipair)
                  atom_b = neighbor_kind_pair%list(2, ipair)
                  fac = 1.0_dp
                  IF (atom_a == atom_b) fac = 0.5_dp
                  rab = r_last_update_pbc(atom_b)%r-r_last_update_pbc(atom_a)%r
                  rab = rab+cell_v
                  rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
                  IF (rab2 <= rab2_max) THEN
                     kind_a = particle_set(atom_a)%atomic_kind%kind_number
                     kind_b = particle_set(atom_b)%atomic_kind%kind_number
                     i_a = eam_kinds_index(kind_a, kind_a)
                     i_b = eam_kinds_index(kind_b, kind_b)
                     eam_a => potparm%pot(kind_a, kind_a)%pot%set(i_a)%eam
                     eam_b => potparm%pot(kind_b, kind_b)%pot%set(i_b)%eam
                     CALL get_rho_eam(eam_a, eam_b, rab2, atom_a, atom_b, rho, fac)
                  END IF
               END DO
            END DO Kind_Group_Loop
         END DO
         CALL mp_sum(rho, para_env%group)
         DO iparticle = 1, nparticle
            eam_data(iparticle)%rho = rho(iparticle)
         END DO

         DEALLOCATE (rho)
      END IF
      DEALLOCATE (eam_kinds_index)
      CALL timestop(handle)

   END SUBROUTINE density_nonbond

! **************************************************************************************************
!> \brief ...
!> \param eam_a ...
!> \param eam_b ...
!> \param rab2 ...
!> \param atom_a ...
!> \param atom_b ...
!> \param rho ...
!> \param fac ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE get_rho_eam(eam_a, eam_b, rab2, atom_a, atom_b, rho, fac)
      TYPE(eam_pot_type), POINTER                        :: eam_a, eam_b
      REAL(dp), INTENT(IN)                               :: rab2
      INTEGER, INTENT(IN)                                :: atom_a, atom_b
      REAL(dp), INTENT(INOUT)                            :: rho(:)
      REAL(dp), INTENT(IN)                               :: fac

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_rho_eam', routineP = moduleN//':'//routineN

      INTEGER                                            :: index
      REAL(dp)                                           :: qq, rab, rhoi, rhoj

      rab = SQRT(rab2)

      ! Particle A
      index = INT(rab/eam_b%drar)+1
      IF (index > eam_b%npoints) THEN
         index = eam_b%npoints
      ELSEIF (index < 1) THEN
         index = 1
      ENDIF
      qq = rab-eam_b%rval(index)
      rhoi = eam_b%rho(index)+qq*eam_b%rhop(index)

      ! Particle B
      index = INT(rab/eam_a%drar)+1
      IF (index > eam_a%npoints) THEN
         index = eam_a%npoints
      ELSEIF (index < 1) THEN
         index = 1
      ENDIF
      qq = rab-eam_a%rval(index)
      rhoj = eam_a%rho(index)+qq*eam_a%rhop(index)

      rho(atom_a) = rho(atom_a)+rhoi*fac
      rho(atom_b) = rho(atom_b)+rhoj*fac
   END SUBROUTINE get_rho_eam

! **************************************************************************************************
!> \brief ...
!> \param rab2 ...
!> \param eam_a ...
!> \param eam_b ...
!> \param eam_data ...
!> \param atom_a ...
!> \param atom_b ...
!> \param f_eam ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE get_force_eam(rab2, eam_a, eam_b, eam_data, atom_a, atom_b, f_eam)
      REAL(dp), INTENT(IN)                               :: rab2
      TYPE(eam_pot_type), POINTER                        :: eam_a, eam_b
      TYPE(eam_type), INTENT(IN)                         :: eam_data(:)
      INTEGER, INTENT(IN)                                :: atom_a, atom_b
      REAL(dp), INTENT(OUT)                              :: f_eam

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_force_eam', routineP = moduleN//':'//routineN

      INTEGER                                            :: index
      REAL(KIND=dp)                                      :: denspi, denspj, fcp, qq, rab

      rab = SQRT(rab2)

      ! Particle A
      index = INT(rab/eam_a%drar)+1
      IF (index > eam_a%npoints) THEN
         index = eam_a%npoints
      ELSEIF (index < 1) THEN
         index = 1
      ENDIF
      qq = rab-eam_a%rval(index)
      IF (index == eam_a%npoints) THEN
         denspi = eam_a%rhop(index)+qq*(eam_a%rhop(index)-eam_a%rhop(index-1))/eam_a%drar
      ELSE
         denspi = eam_a%rhop(index)+qq*(eam_a%rhop(index+1)-eam_a%rhop(index))/eam_a%drar
      END IF

      ! Particle B
      index = INT(rab/eam_b%drar)+1
      IF (index > eam_b%npoints) THEN
         index = eam_b%npoints
      ELSEIF (index < 1) THEN
         index = 1
      ENDIF
      qq = rab-eam_b%rval(index)
      IF (index == eam_b%npoints) THEN
         denspj = eam_b%rhop(index)+qq*(eam_b%rhop(index)-eam_b%rhop(index-1))/eam_b%drar
      ELSE
         denspj = eam_b%rhop(index)+qq*(eam_b%rhop(index+1)-eam_b%rhop(index))/eam_b%drar
      END IF

      fcp = denspj*eam_data(atom_a)%f_embed+denspi*eam_data(atom_b)%f_embed
      f_eam = fcp/rab
   END SUBROUTINE get_force_eam

END MODULE manybody_eam

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE manybody_quip

USE cp_log_handling, ONLY: cp_logger_get_default_io_unit
  USE atomic_kind_types,               ONLY: atomic_kind_type
  USE bibliography,                    ONLY: QUIP_ref,&
                                             cite_reference
  USE cell_types,                      ONLY: cell_type
  USE cp_para_types,                   ONLY: cp_para_env_type
  USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_get,&
                                             fist_nonbond_env_set,&
                                             fist_nonbond_env_type,&
                                             quip_data_type
  USE kinds,                           ONLY: dp
  USE pair_potential_types,            ONLY: pair_potential_pp_type,&
                                             pair_potential_single_type,&
                                             quip_pot_type,&
                                             quip_type
  USE particle_types,                  ONLY: particle_type
  USE physcon,                         ONLY: angstrom,&
                                             evolt

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC quip_energy_store_force_virial, quip_add_force_virial

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'manybody_quip'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param particle_set ...
!> \param cell ...
!> \param atomic_kind_set ...
!> \param potparm ...
!> \param fist_nonbond_env ...
!> \param pot_quip ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE quip_energy_store_force_virial(particle_set, cell, atomic_kind_set, potparm, fist_nonbond_env, &
                                             pot_quip, para_env)
      TYPE(particle_type), POINTER             :: particle_set(:)
      TYPE(cell_type), POINTER                 :: cell
      TYPE(atomic_kind_type), POINTER          :: atomic_kind_set(:)
      TYPE(pair_potential_pp_type), POINTER    :: potparm
      TYPE(fist_nonbond_env_type), POINTER     :: fist_nonbond_env
      REAL(kind=dp)                            :: pot_quip
      TYPE(cp_para_env_type), OPTIONAL, &
         POINTER                                :: para_env

      CHARACTER(LEN=*), PARAMETER :: &
         routineN = 'quip_energy_store_force_virial', &
         routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(particle_set))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(cell))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(atomic_kind_set))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(potparm))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(fist_nonbond_env))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pot_quip))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(para_env))==-1) EXIT ;  END DO ; ENDIF
      CALL cp_abort(cp__l("manybody_quip.F",90), "In order to use QUIP you need to download "// &
                    "and install the libAtoms/QUIP library (check CP2K manual for details)")
   END SUBROUTINE quip_energy_store_force_virial

! **************************************************************************************************
!> \brief ...
!> \param fist_nonbond_env ...
!> \param force ...
!> \param virial ...
! **************************************************************************************************
   SUBROUTINE quip_add_force_virial(fist_nonbond_env, force, virial)
      TYPE(fist_nonbond_env_type), POINTER     :: fist_nonbond_env
      REAL(KIND=dp)                            :: force(:, :), virial(3, 3)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'quip_add_force_virial', &
                                     routineP = moduleN//':'//routineN


      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(fist_nonbond_env))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(force))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(virial))==-1) EXIT ;  END DO ; ENDIF
      RETURN
   END SUBROUTINE quip_add_force_virial

END MODULE manybody_quip
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Efficient tersoff implementation
!> \author CJM, I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
MODULE manybody_tersoff

   USE cell_types,                      ONLY: cell_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                              neighbor_kind_pairs_type
   USE fist_nonbond_env_types,          ONLY: pos_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE mathlib,                         ONLY: matvec_3x3
   USE pair_potential_types,            ONLY: pair_potential_pp_type,&
                                              pair_potential_single_type,&
                                              tersoff_pot_type,&
                                              tersoff_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: setup_tersoff_arrays, destroy_tersoff_arrays, &
             tersoff_forces, tersoff_energy
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'manybody_tersoff'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param pot_loc ...
!> \param tersoff ...
!> \param r_last_update_pbc ...
!> \param atom_a ...
!> \param atom_b ...
!> \param nloc_size ...
!> \param full_loc_list ...
!> \param loc_cell_v ...
!> \param cell_v ...
!> \param drij ...
!> \author Teodoro Laino [tlaino] - 12.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE tersoff_energy(pot_loc, tersoff, r_last_update_pbc, atom_a, atom_b, nloc_size, &
                             full_loc_list, loc_cell_v, cell_v, drij)

      REAL(KIND=dp), INTENT(OUT)                         :: pot_loc
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: atom_a, atom_b, nloc_size
      INTEGER, DIMENSION(2, 1:nloc_size)                 :: full_loc_list
      REAL(KIND=dp), DIMENSION(3, 1:nloc_size)           :: loc_cell_v
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      REAL(KIND=dp)                                      :: drij

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tersoff_energy', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: b_ij, f_A, f_C, f_R

      b_ij = ter_b_ij(tersoff, r_last_update_pbc, atom_a, atom_b, nloc_size, &
                      full_loc_list, loc_cell_v, cell_v, tersoff%rcutsq)
      f_C = ter_f_C(tersoff, drij)
      f_A = ter_f_A(tersoff, drij)
      f_R = ter_f_R(tersoff, drij)
      pot_loc = f_C*(f_R+b_ij*f_A)

   END SUBROUTINE tersoff_energy

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r ...
!> \retval ter_f_C ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_f_C(tersoff, r)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: ter_f_C

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_f_C', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: bigD, bigR, RmD, RpD

      bigR = tersoff%bigR
      bigD = tersoff%bigD
      RmD = tersoff%bigR-tersoff%bigD
      RpD = tersoff%bigR+tersoff%bigD
      ter_f_C = 0.0_dp
      IF (r < RmD) ter_f_C = 1.0_dp
      IF (r > RpD) ter_f_C = 0.0_dp
      IF ((r < RpD) .AND. (r > RmD)) THEN
         ter_f_C = 0.5_dp*(1.0_dp-SIN(0.5_dp*PI*(r-bigR)/(bigD)))
      END IF
   END FUNCTION ter_f_C

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r ...
!> \retval ter_f_C_d ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_f_C_d(tersoff, r)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: ter_f_C_d

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_f_C_d', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: bigD, bigR, RmD, RpD

      bigR = tersoff%bigR
      bigD = tersoff%bigD
      RmD = tersoff%bigR-tersoff%bigD
      RpD = tersoff%bigR+tersoff%bigD
      ter_f_C_d = 0.0_dp
      IF (r < RmD) ter_f_C_d = 0.0_dp
      IF (r > RpD) ter_f_C_d = 0.0_dp
      IF ((r < RpD) .AND. (r > RmD)) THEN
         ter_f_C_d = (0.25_dp*PI/bigD)*COS(0.5_dp*PI*(r-bigR)/(bigD))/r
      END IF

   END FUNCTION ter_f_C_d

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r ...
!> \retval ter_f_R ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_f_R(tersoff, r)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: ter_f_R

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_f_R', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: A, lambda1

      A = tersoff%A
      lambda1 = tersoff%lambda1
      ter_f_R = 0.0_dp
      ter_f_R = A*EXP(-lambda1*r)

   END FUNCTION ter_f_R

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r ...
!> \retval ter_f_R_d ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_f_R_d(tersoff, r)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: ter_f_R_d

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_f_R_d', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: A, f_R, lambda1

      A = tersoff%A
      lambda1 = tersoff%lambda1
      f_R = A*EXP(-lambda1*r)
      ter_f_R_d = 0.0_dp
      ter_f_R_d = lambda1*f_R/r

   END FUNCTION ter_f_R_d

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r ...
!> \retval ter_f_A ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_f_A(tersoff, r)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: ter_f_A

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_f_A', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: B, lambda2

      B = tersoff%B
      lambda2 = tersoff%lambda2
      ter_f_A = 0.0_dp
      ter_f_A = -B*EXP(-lambda2*r)

   END FUNCTION ter_f_A

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r ...
!> \retval ter_f_A_d ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_f_A_d(tersoff, r)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: ter_f_A_d

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_f_A_d', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: B, lambda2

      B = tersoff%B
      lambda2 = tersoff%lambda2
      ter_f_A_d = 0.0_dp
      ter_f_A_d = -B*lambda2*EXP(-lambda2*r)/r

   END FUNCTION ter_f_A_d

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \retval ter_a_ij ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_a_ij(tersoff)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp)                                      :: ter_a_ij

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_a_ij', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: alpha, n

      n = tersoff%n
      alpha = tersoff%alpha
      ter_a_ij = 0.0_dp
      !Note alpha = 0.0_dp for the parameters in the paper so using simplified term
      !ter_a_ij = (1.0_dp+(alpha*ter_n_ij(tersoff,iparticle,jparticle,r))**n)**(-0.5_dp/n)
      ter_a_ij = 1.0_dp

   END FUNCTION ter_a_ij

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r_last_update_pbc ...
!> \param iparticle ...
!> \param jparticle ...
!> \param n_loc_size ...
!> \param full_loc_list ...
!> \param loc_cell_v ...
!> \param cell_v ...
!> \param rcutsq ...
!> \retval ter_b_ij ...
!> \author I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
   FUNCTION ter_b_ij(tersoff, r_last_update_pbc, iparticle, jparticle, n_loc_size, &
                     full_loc_list, loc_cell_v, cell_v, rcutsq)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: iparticle, jparticle, n_loc_size
      INTEGER, DIMENSION(2, 1:n_loc_size)                :: full_loc_list
      REAL(KIND=dp), DIMENSION(3, 1:n_loc_size)          :: loc_cell_v
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      REAL(KIND=dp)                                      :: ter_b_ij

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_b_ij', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: beta, n, zeta_ij

      n = tersoff%n
      beta = tersoff%beta
      ter_b_ij = 0.0_dp
      zeta_ij = ter_zeta_ij(tersoff, r_last_update_pbc, iparticle, jparticle, &
                            n_loc_size, full_loc_list, loc_cell_v, cell_v, rcutsq)
      ter_b_ij = (1.0_dp+(beta*zeta_ij)**n)**(-0.5_dp/n)

   END FUNCTION ter_b_ij

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r_last_update_pbc ...
!> \param iparticle ...
!> \param jparticle ...
!> \param n_loc_size ...
!> \param full_loc_list ...
!> \param loc_cell_v ...
!> \param cell_v ...
!> \param rcutsq ...
!> \retval ter_b_ij_d ...
!> \author I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
   FUNCTION ter_b_ij_d(tersoff, r_last_update_pbc, iparticle, jparticle, n_loc_size, &
                       full_loc_list, loc_cell_v, cell_v, rcutsq)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: iparticle, jparticle, n_loc_size
      INTEGER, DIMENSION(2, 1:n_loc_size)                :: full_loc_list
      REAL(KIND=dp), DIMENSION(3, 1:n_loc_size)          :: loc_cell_v
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      REAL(KIND=dp)                                      :: ter_b_ij_d

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_b_ij_d', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: beta, beta_n, n, zeta_ij, zeta_ij_n, &
                                                            zeta_ij_nm1

      n = tersoff%n
      beta = tersoff%beta
      beta_n = beta**n
      zeta_ij = ter_zeta_ij(tersoff, r_last_update_pbc, iparticle, jparticle, n_loc_size, &
                            full_loc_list, loc_cell_v, cell_v, rcutsq)
      zeta_ij_nm1 = 0.0_dp
      IF (zeta_ij > 0.0_dp) zeta_ij_nm1 = zeta_ij**(n-1.0_dp)
      zeta_ij_n = zeta_ij**(n)

      ter_b_ij_d = 0.0_dp
      ter_b_ij_d = -0.5_dp*beta_n*zeta_ij_nm1* &
                   ((1.0_dp+beta_n*zeta_ij_n)**((-0.5_dp/n)-1.0_dp))

   END FUNCTION ter_b_ij_d

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r_last_update_pbc ...
!> \param iparticle ...
!> \param jparticle ...
!> \param n_loc_size ...
!> \param full_loc_list ...
!> \param loc_cell_v ...
!> \param cell_v ...
!> \param rcutsq ...
!> \retval ter_zeta_ij ...
!> \par History
!>      Using a local list of neighbors - [tlaino] 2007
!> \author I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
   FUNCTION ter_zeta_ij(tersoff, r_last_update_pbc, iparticle, jparticle, n_loc_size, &
                        full_loc_list, loc_cell_v, cell_v, rcutsq)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: iparticle, jparticle, n_loc_size
      INTEGER, DIMENSION(2, 1:n_loc_size)                :: full_loc_list
      REAL(KIND=dp), DIMENSION(3, 1:n_loc_size)          :: loc_cell_v
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      REAL(KIND=dp)                                      :: ter_zeta_ij

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_zeta_ij', routineP = moduleN//':'//routineN

      INTEGER                                            :: ilist, kparticle
      REAL(KIND=dp)                                      :: cell_v_2(3), costheta, drij, drik, &
                                                            expterm, f_C, gterm, lambda3, n, &
                                                            rab2_max, rij(3), rik(3)

      ter_zeta_ij = 0.0_dp
      n = tersoff%n
      lambda3 = tersoff%lambda3
      rab2_max = rcutsq
      rij(:) = r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v
      drij = SQRT(DOT_PRODUCT(rij, rij))
      ter_zeta_ij = 0.0_dp
      DO ilist = 1, n_loc_size
         kparticle = full_loc_list(2, ilist)
         IF (kparticle == jparticle) CYCLE
         cell_v_2 = loc_cell_v(:, ilist)
         rik(:) = r_last_update_pbc(kparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v_2
         drik = DOT_PRODUCT(rik, rik)
         IF (drik > rab2_max) CYCLE
         drik = SQRT(drik)
         costheta = DOT_PRODUCT(rij, rik)/(drij*drik)
         IF (costheta < -1.0_dp) costheta = -1.0_dp
         IF (costheta > +1.0_dp) costheta = +1.0_dp
         f_C = ter_f_C(tersoff, drik)
         gterm = ter_g(tersoff, costheta)
         expterm = EXP((lambda3*(drij-drik))**3)
         ter_zeta_ij = ter_zeta_ij+f_C*gterm*expterm
      END DO

   END FUNCTION ter_zeta_ij

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r_last_update_pbc ...
!> \param iparticle ...
!> \param jparticle ...
!> \param f_nonbond ...
!> \param pv_nonbond ...
!> \param prefactor ...
!> \param n_loc_size ...
!> \param full_loc_list ...
!> \param loc_cell_v ...
!> \param cell_v ...
!> \param rcutsq ...
!> \param use_virial ...
!> \par History
!>       Using a local list of neighbors - [tlaino] 2007
!> \author I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
   SUBROUTINE ter_zeta_ij_d(tersoff, r_last_update_pbc, iparticle, jparticle, f_nonbond, pv_nonbond, prefactor, &
                            n_loc_size, full_loc_list, loc_cell_v, cell_v, rcutsq, use_virial)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: iparticle, jparticle
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: f_nonbond, pv_nonbond
      REAL(KIND=dp), INTENT(IN)                          :: prefactor
      INTEGER, INTENT(IN)                                :: n_loc_size
      INTEGER, DIMENSION(2, 1:n_loc_size)                :: full_loc_list
      REAL(KIND=dp), DIMENSION(3, 1:n_loc_size)          :: loc_cell_v
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      LOGICAL, INTENT(IN)                                :: use_virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_zeta_ij_d', routineP = moduleN//':'//routineN

      INTEGER                                            :: ilist, kparticle, nparticle
      REAL(KIND=dp)                                      :: costheta, drij, drik, expterm, &
                                                            expterm_d, f_C, f_C_d, gterm, gterm_d, &
                                                            lambda3, n, rab2_max
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v_2, dcosdri, dcosdrj, dcosdrk, &
                                                            dri, drj, drk, rij, rij_hat, rik, &
                                                            rik_hat

      n = tersoff%n
      lambda3 = tersoff%lambda3
      rab2_max = rcutsq

      rij(:) = r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v
      drij = SQRT(DOT_PRODUCT(rij, rij))
      rij_hat(:) = rij(:)/drij

      nparticle = SIZE(r_last_update_pbc)
      DO ilist = 1, n_loc_size
         kparticle = full_loc_list(2, ilist)
         IF (kparticle == jparticle) CYCLE
         cell_v_2 = loc_cell_v(:, ilist)
         rik(:) = r_last_update_pbc(kparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v_2
         drik = DOT_PRODUCT(rik, rik)

         IF (drik > rab2_max) CYCLE
         drik = SQRT(drik)
         rik_hat(:) = rik(:)/drik
         costheta = DOT_PRODUCT(rij, rik)/(drij*drik)
         IF (costheta < -1.0_dp) costheta = -1.0_dp
         IF (costheta > +1.0_dp) costheta = +1.0_dp

         dcosdrj(:) = (1.0_dp/(drij))*(rik_hat(:)-costheta*rij_hat(:))
         dcosdrk(:) = (1.0_dp/(drik))*(rij_hat(:)-costheta*rik_hat(:))
         dcosdri(:) = -(dcosdrj(:)+dcosdrk(:))

         f_C = ter_f_C(tersoff, drik)
         f_C_d = ter_f_C_d(tersoff, drik)
         gterm = ter_g(tersoff, costheta)
         gterm_d = ter_g_d(tersoff, costheta) !still need d(costheta)/dR term
         expterm = EXP((lambda3*(drij-drik))**3)
         expterm_d = (3.0_dp)*(lambda3**3)*((drij-drik)**2)*expterm

         dri = f_C_d*gterm*expterm*(rik) &
               +f_C*gterm_d*expterm*(dcosdri) &
               +f_C*gterm*expterm_d*(-rij_hat+rik_hat)

         !No f_C_d component for Rj
         drj = f_C*gterm_d*expterm*(dcosdrj) &
               +f_C*gterm*expterm_d*(rij_hat)

         drk = f_C_d*gterm*expterm*(-rik) &
               +f_C*gterm_d*expterm*(dcosdrk) &
               +f_C*gterm*expterm_d*(-rik_hat)

         f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+prefactor*dri(1)
         f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+prefactor*dri(2)
         f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+prefactor*dri(3)

         f_nonbond(1, jparticle) = f_nonbond(1, jparticle)+prefactor*drj(1)
         f_nonbond(2, jparticle) = f_nonbond(2, jparticle)+prefactor*drj(2)
         f_nonbond(3, jparticle) = f_nonbond(3, jparticle)+prefactor*drj(3)

         f_nonbond(1, kparticle) = f_nonbond(1, kparticle)+prefactor*drk(1)
         f_nonbond(2, kparticle) = f_nonbond(2, kparticle)+prefactor*drk(2)
         f_nonbond(3, kparticle) = f_nonbond(3, kparticle)+prefactor*drk(3)

         IF (use_virial) THEN
            pv_nonbond(1, 1) = pv_nonbond(1, 1)+prefactor*(rij(1)*drj(1)+rik(1)*drk(1))
            pv_nonbond(1, 2) = pv_nonbond(1, 2)+prefactor*(rij(1)*drj(2)+rik(1)*drk(2))
            pv_nonbond(1, 3) = pv_nonbond(1, 3)+prefactor*(rij(1)*drj(3)+rik(1)*drk(3))

            pv_nonbond(2, 1) = pv_nonbond(2, 1)+prefactor*(rij(2)*drj(1)+rik(2)*drk(1))
            pv_nonbond(2, 2) = pv_nonbond(2, 2)+prefactor*(rij(2)*drj(2)+rik(2)*drk(2))
            pv_nonbond(2, 3) = pv_nonbond(2, 3)+prefactor*(rij(2)*drj(3)+rik(2)*drk(3))

            pv_nonbond(3, 1) = pv_nonbond(3, 1)+prefactor*(rij(3)*drj(1)+rik(3)*drk(1))
            pv_nonbond(3, 2) = pv_nonbond(3, 2)+prefactor*(rij(3)*drj(2)+rik(3)*drk(2))
            pv_nonbond(3, 3) = pv_nonbond(3, 3)+prefactor*(rij(3)*drj(3)+rik(3)*drk(3))
         END IF
      END DO
   END SUBROUTINE ter_zeta_ij_d

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param costheta ...
!> \retval ter_g ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_g(tersoff, costheta)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp), INTENT(IN)                          :: costheta
      REAL(KIND=dp)                                      :: ter_g

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_g', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: c, c2, d, d2, h

      c = tersoff%c
      d = tersoff%d
      h = tersoff%h
      c2 = c*c
      d2 = d*d
      ter_g = 0.0_dp
      ter_g = 1.0_dp+(c2/d2)-(c2)/(d2+(h-costheta)**2)

   END FUNCTION ter_g

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param costheta ...
!> \retval ter_g_d ...
!> \author I-Feng W. Kuo
! **************************************************************************************************
   FUNCTION ter_g_d(tersoff, costheta)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      REAL(KIND=dp), INTENT(IN)                          :: costheta
      REAL(KIND=dp)                                      :: ter_g_d

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ter_g_d', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: c, c2, d, d2, h, hc, sintheta

      c = tersoff%c
      d = tersoff%d
      h = tersoff%h
      c2 = c*c
      d2 = d*d
      hc = h-costheta

      sintheta = SQRT(1.0-costheta**2)

      ter_g_d = 0.0_dp
      ! Still need d(costheta)/dR
      ter_g_d = (-2.0_dp*c2*hc)/(d2+hc**2)**2
   END FUNCTION ter_g_d

! **************************************************************************************************
!> \brief ...
!> \param tersoff ...
!> \param r_last_update_pbc ...
!> \param cell_v ...
!> \param n_loc_size ...
!> \param full_loc_list ...
!> \param loc_cell_v ...
!> \param iparticle ...
!> \param jparticle ...
!> \param f_nonbond ...
!> \param pv_nonbond ...
!> \param use_virial ...
!> \param rcutsq ...
!> \par History
!>       Using a local list of neighbors - [tlaino] 2007
!> \author I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
   SUBROUTINE tersoff_forces(tersoff, r_last_update_pbc, cell_v, n_loc_size, &
                             full_loc_list, loc_cell_v, iparticle, jparticle, f_nonbond, pv_nonbond, &
                             use_virial, rcutsq)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      INTEGER, INTENT(IN)                                :: n_loc_size
      INTEGER, DIMENSION(2, 1:n_loc_size)                :: full_loc_list
      REAL(KIND=dp), DIMENSION(3, 1:n_loc_size)          :: loc_cell_v
      INTEGER, INTENT(IN)                                :: iparticle, jparticle
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: f_nonbond, pv_nonbond
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tersoff_forces', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: b_ij, b_ij_d, drij, f_A, f_A1, f_A2, &
                                                            f_A_d, f_C, f_C_d, f_R, f_R1, f_R2, &
                                                            f_R_d, fac, prefactor, rij(3), &
                                                            rij_hat(3)

      CALL timeset(routineN, handle)
      rij(:) = r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v
      drij = SQRT(DOT_PRODUCT(rij, rij))
      rij_hat(:) = rij(:)/drij

      fac = -0.5_dp
      b_ij = ter_b_ij(tersoff, r_last_update_pbc, iparticle, jparticle, n_loc_size, full_loc_list, loc_cell_v, cell_v, rcutsq)
      b_ij_d = ter_b_ij_d(tersoff, r_last_update_pbc, iparticle, jparticle, n_loc_size, full_loc_list, loc_cell_v, cell_v, rcutsq)
      f_A = ter_f_A(tersoff, drij)
      f_A_d = ter_f_A_d(tersoff, drij)
      f_C = ter_f_C(tersoff, drij)
      f_C_d = ter_f_C_d(tersoff, drij)
      f_R = ter_f_R(tersoff, drij)
      f_R_d = ter_f_R_d(tersoff, drij)

      ! Lets do the easy one first, the repulsive term
      ! Note a_ij = 1.0_dp so just going to ignore it...
      f_R1 = f_C_d*f_R*fac
      f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+f_R1*rij(1)
      f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+f_R1*rij(2)
      f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+f_R1*rij(3)
      f_nonbond(1, jparticle) = f_nonbond(1, jparticle)-f_R1*rij(1)
      f_nonbond(2, jparticle) = f_nonbond(2, jparticle)-f_R1*rij(2)
      f_nonbond(3, jparticle) = f_nonbond(3, jparticle)-f_R1*rij(3)

      IF (use_virial) THEN
         pv_nonbond(1, 1) = pv_nonbond(1, 1)-f_R1*rij(1)*rij(1)
         pv_nonbond(1, 2) = pv_nonbond(1, 2)-f_R1*rij(1)*rij(2)
         pv_nonbond(1, 3) = pv_nonbond(1, 3)-f_R1*rij(1)*rij(3)
         pv_nonbond(2, 1) = pv_nonbond(2, 1)-f_R1*rij(2)*rij(1)
         pv_nonbond(2, 2) = pv_nonbond(2, 2)-f_R1*rij(2)*rij(2)
         pv_nonbond(2, 3) = pv_nonbond(2, 3)-f_R1*rij(2)*rij(3)
         pv_nonbond(3, 1) = pv_nonbond(3, 1)-f_R1*rij(3)*rij(1)
         pv_nonbond(3, 2) = pv_nonbond(3, 2)-f_R1*rij(3)*rij(2)
         pv_nonbond(3, 3) = pv_nonbond(3, 3)-f_R1*rij(3)*rij(3)
      END IF

      f_R2 = f_C*f_R_d*fac
      f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+f_R2*rij(1)
      f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+f_R2*rij(2)
      f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+f_R2*rij(3)
      f_nonbond(1, jparticle) = f_nonbond(1, jparticle)-f_R2*rij(1)
      f_nonbond(2, jparticle) = f_nonbond(2, jparticle)-f_R2*rij(2)
      f_nonbond(3, jparticle) = f_nonbond(3, jparticle)-f_R2*rij(3)

      IF (use_virial) THEN
         pv_nonbond(1, 1) = pv_nonbond(1, 1)-f_R2*rij(1)*rij(1)
         pv_nonbond(1, 2) = pv_nonbond(1, 2)-f_R2*rij(1)*rij(2)
         pv_nonbond(1, 3) = pv_nonbond(1, 3)-f_R2*rij(1)*rij(3)
         pv_nonbond(2, 1) = pv_nonbond(2, 1)-f_R2*rij(2)*rij(1)
         pv_nonbond(2, 2) = pv_nonbond(2, 2)-f_R2*rij(2)*rij(2)
         pv_nonbond(2, 3) = pv_nonbond(2, 3)-f_R2*rij(2)*rij(3)
         pv_nonbond(3, 1) = pv_nonbond(3, 1)-f_R2*rij(3)*rij(1)
         pv_nonbond(3, 2) = pv_nonbond(3, 2)-f_R2*rij(3)*rij(2)
         pv_nonbond(3, 3) = pv_nonbond(3, 3)-f_R2*rij(3)*rij(3)
      END IF

      ! Lets do the f_A1 piece derivative of F_C
      f_A1 = f_C_d*b_ij*f_A*fac
      f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+f_A1*rij(1)
      f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+f_A1*rij(2)
      f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+f_A1*rij(3)
      f_nonbond(1, jparticle) = f_nonbond(1, jparticle)-f_A1*rij(1)
      f_nonbond(2, jparticle) = f_nonbond(2, jparticle)-f_A1*rij(2)
      f_nonbond(3, jparticle) = f_nonbond(3, jparticle)-f_A1*rij(3)

      IF (use_virial) THEN
         pv_nonbond(1, 1) = pv_nonbond(1, 1)-f_A1*rij(1)*rij(1)
         pv_nonbond(1, 2) = pv_nonbond(1, 2)-f_A1*rij(1)*rij(2)
         pv_nonbond(1, 3) = pv_nonbond(1, 3)-f_A1*rij(1)*rij(3)
         pv_nonbond(2, 1) = pv_nonbond(2, 1)-f_A1*rij(2)*rij(1)
         pv_nonbond(2, 2) = pv_nonbond(2, 2)-f_A1*rij(2)*rij(2)
         pv_nonbond(2, 3) = pv_nonbond(2, 3)-f_A1*rij(2)*rij(3)
         pv_nonbond(3, 1) = pv_nonbond(3, 1)-f_A1*rij(3)*rij(1)
         pv_nonbond(3, 2) = pv_nonbond(3, 2)-f_A1*rij(3)*rij(2)
         pv_nonbond(3, 3) = pv_nonbond(3, 3)-f_A1*rij(3)*rij(3)
      END IF

      ! Lets do the f_A2 piece derivative of F_A
      f_A2 = f_C*b_ij*f_A_d*fac
      f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+f_A2*rij(1)
      f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+f_A2*rij(2)
      f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+f_A2*rij(3)
      f_nonbond(1, jparticle) = f_nonbond(1, jparticle)-f_A2*rij(1)
      f_nonbond(2, jparticle) = f_nonbond(2, jparticle)-f_A2*rij(2)
      f_nonbond(3, jparticle) = f_nonbond(3, jparticle)-f_A2*rij(3)

      IF (use_virial) THEN
         pv_nonbond(1, 1) = pv_nonbond(1, 1)-f_A2*rij(1)*rij(1)
         pv_nonbond(1, 2) = pv_nonbond(1, 2)-f_A2*rij(1)*rij(2)
         pv_nonbond(1, 3) = pv_nonbond(1, 3)-f_A2*rij(1)*rij(3)
         pv_nonbond(2, 1) = pv_nonbond(2, 1)-f_A2*rij(2)*rij(1)
         pv_nonbond(2, 2) = pv_nonbond(2, 2)-f_A2*rij(2)*rij(2)
         pv_nonbond(2, 3) = pv_nonbond(2, 3)-f_A2*rij(2)*rij(3)
         pv_nonbond(3, 1) = pv_nonbond(3, 1)-f_A2*rij(3)*rij(1)
         pv_nonbond(3, 2) = pv_nonbond(3, 2)-f_A2*rij(3)*rij(2)
         pv_nonbond(3, 3) = pv_nonbond(3, 3)-f_A2*rij(3)*rij(3)
      END IF

      ! Lets do the f_A3 piece derivative of b_ij
      prefactor = f_C*b_ij_d*f_A*fac ! Note need to do d(Zeta_ij)/dR
      CALL ter_zeta_ij_d(tersoff, r_last_update_pbc, iparticle, jparticle, f_nonbond, pv_nonbond, prefactor, &
                         n_loc_size, full_loc_list, loc_cell_v, cell_v, rcutsq, use_virial)
      CALL timestop(handle)
   END SUBROUTINE tersoff_forces

! **************************************************************************************************
!> \brief ...
!> \param nonbonded ...
!> \param potparm ...
!> \param glob_loc_list ...
!> \param glob_cell_v ...
!> \param glob_loc_list_a ...
!> \param cell ...
!> \par History
!>      Fast implementation of the tersoff potential - [tlaino] 2007
!> \author Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE setup_tersoff_arrays(nonbonded, potparm, glob_loc_list, glob_cell_v, glob_loc_list_a, cell)
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      INTEGER, DIMENSION(:, :), POINTER                  :: glob_loc_list
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: glob_cell_v
      INTEGER, DIMENSION(:), POINTER                     :: glob_loc_list_a
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_tersoff_arrays', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iend, igrp, ikind, ilist, &
                                                            ipair, istart, jkind, nkinds, npairs, &
                                                            npairs_tot
      INTEGER, DIMENSION(:), POINTER                     :: work_list, work_list2
      INTEGER, DIMENSION(:, :), POINTER                  :: list
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v, cvi
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rwork_list
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      TYPE(pair_potential_single_type), POINTER          :: pot

      IF(.NOT.(.NOT. ASSOCIATED(glob_loc_list)))CALL cp__a("manybody_tersoff.F",743)
      IF(.NOT.(.NOT. ASSOCIATED(glob_loc_list_a)))CALL cp__a("manybody_tersoff.F",744)
      IF(.NOT.(.NOT. ASSOCIATED(glob_cell_v)))CALL cp__a("manybody_tersoff.F",745)
      CALL timeset(routineN, handle)
      npairs_tot = 0
      nkinds = SIZE(potparm%pot, 1)
      DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         npairs = neighbor_kind_pair%npairs
         IF (npairs == 0) CYCLE
         Kind_Group_Loop1: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            iend = neighbor_kind_pair%grp_kind_end(igrp)
            ikind = neighbor_kind_pair%ij_kind(1, igrp)
            jkind = neighbor_kind_pair%ij_kind(2, igrp)
            pot => potparm%pot(ikind, jkind)%pot
            npairs = iend-istart+1
            IF (pot%no_mb) CYCLE
            DO i = 1, SIZE(pot%type)
               IF (pot%type(i) == tersoff_type) npairs_tot = npairs_tot+npairs
            END DO
         END DO Kind_Group_Loop1
      END DO
      ALLOCATE (work_list(npairs_tot))
      ALLOCATE (work_list2(npairs_tot))
      ALLOCATE (glob_loc_list(2, npairs_tot))
      ALLOCATE (glob_cell_v(3, npairs_tot))
      ! Fill arrays with data
      npairs_tot = 0
      DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         npairs = neighbor_kind_pair%npairs
         IF (npairs == 0) CYCLE
         Kind_Group_Loop2: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            iend = neighbor_kind_pair%grp_kind_end(igrp)
            ikind = neighbor_kind_pair%ij_kind(1, igrp)
            jkind = neighbor_kind_pair%ij_kind(2, igrp)
            list => neighbor_kind_pair%list
            cvi = neighbor_kind_pair%cell_vector
            pot => potparm%pot(ikind, jkind)%pot
            npairs = iend-istart+1
            IF (pot%no_mb) CYCLE
            CALL matvec_3x3(cell_v, cell%hmat, cvi)
            DO i = 1, SIZE(pot%type)
               ! TERSOFF
               IF (pot%type(i) == tersoff_type) THEN
                  DO ipair = 1, npairs
                     glob_loc_list(:, npairs_tot+ipair) = list(:, istart-1+ipair)
                     glob_cell_v(1:3, npairs_tot+ipair) = cell_v(1:3)
                  END DO
                  npairs_tot = npairs_tot+npairs
               END IF
            END DO
         END DO Kind_Group_Loop2
      END DO
      ! Order the arrays w.r.t. the first index of glob_loc_list
      CALL sort(glob_loc_list(1, :), npairs_tot, work_list)
      DO ipair = 1, npairs_tot
         work_list2(ipair) = glob_loc_list(2, work_list(ipair))
      END DO
      glob_loc_list(2, :) = work_list2
      DEALLOCATE (work_list2)
      ALLOCATE (rwork_list(3, npairs_tot))
      DO ipair = 1, npairs_tot
         rwork_list(:, ipair) = glob_cell_v(:, work_list(ipair))
      END DO
      glob_cell_v = rwork_list
      DEALLOCATE (rwork_list)
      DEALLOCATE (work_list)
      ALLOCATE (glob_loc_list_a(npairs_tot))
      glob_loc_list_a = glob_loc_list(1, :)
      CALL timestop(handle)
   END SUBROUTINE setup_tersoff_arrays

! **************************************************************************************************
!> \brief ...
!> \param glob_loc_list ...
!> \param glob_cell_v ...
!> \param glob_loc_list_a ...
!> \par History
!>      Fast implementation of the tersoff potential - [tlaino] 2007
!> \author Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE destroy_tersoff_arrays(glob_loc_list, glob_cell_v, glob_loc_list_a)
      INTEGER, DIMENSION(:, :), POINTER                  :: glob_loc_list
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: glob_cell_v
      INTEGER, DIMENSION(:), POINTER                     :: glob_loc_list_a

      CHARACTER(LEN=*), PARAMETER :: routineN = 'destroy_tersoff_arrays', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(glob_loc_list)) THEN
         DEALLOCATE (glob_loc_list)
      END IF
      IF (ASSOCIATED(glob_loc_list_a)) THEN
         DEALLOCATE (glob_loc_list_a)
      END IF
      IF (ASSOCIATED(glob_cell_v)) THEN
         DEALLOCATE (glob_cell_v)
      END IF

   END SUBROUTINE destroy_tersoff_arrays

END MODULE manybody_tersoff

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!  \brief implementation of dipole and three-body part of Siepmann-Sprik
!         potential
!>        dipole term: 3rd term in Eq. (1) in J. Chem. Phys., Vol. 102, p.511
!>        three-body term: Eq. (4) in J. Chem. Phys., Vol. 102, p. 511
!>        remaining terms of Siepmann-Sprik potential can be given via
!>        the GENPOT section
!> \par History
!>      12.2012 created
!> \author Dorothea Golze
! **************************************************************************************************
MODULE manybody_siepmann

   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                              neighbor_kind_pairs_type
   USE fist_nonbond_env_types,          ONLY: pos_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: matvec_3x3
   USE message_passing,                 ONLY: mp_sum
   USE pair_potential_types,            ONLY: pair_potential_pp_type,&
                                              pair_potential_single_type,&
                                              siepmann_pot_type,&
                                              siepmann_type
   USE particle_types,                  ONLY: particle_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: setup_siepmann_arrays, destroy_siepmann_arrays, &
             siepmann_energy, siepmann_forces_v2, siepmann_forces_v3, &
             print_nr_ions_siepmann
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'manybody_siepmann'

CONTAINS

! **************************************************************************************************
!> \brief  energy of two-body dipole term and three-body term
!> \param pot_loc ...
!> \param siepmann ...
!> \param r_last_update_pbc ...
!> \param atom_a ...
!> \param atom_b ...
!> \param nloc_size ...
!> \param full_loc_list ...
!> \param cell_v ...
!> \param cell ...
!> \param drij ...
!> \param particle_set ...
!> \param nr_oh number of OH- ions near surface
!> \param nr_h3o number of hydronium ions near surface
!> \author Dorothea Golze - 11.2012 - University of Zurich
! **************************************************************************************************
   SUBROUTINE siepmann_energy(pot_loc, siepmann, r_last_update_pbc, atom_a, atom_b, &
                              nloc_size, full_loc_list, cell_v, cell, drij, particle_set, &
                              nr_oh, nr_h3o)

      REAL(KIND=dp), INTENT(OUT)                         :: pot_loc
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: atom_a, atom_b, nloc_size
      INTEGER, DIMENSION(2, 1:nloc_size)                 :: full_loc_list
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp)                                      :: drij
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(INOUT)                             :: nr_oh, nr_h3o

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siepmann_energy', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: a_ij, D, E, f2, Phi_ij, pot_loc_v2, &
                                                            pot_loc_v3

      a_ij = siep_a_ij(siepmann, r_last_update_pbc, atom_a, atom_b, nloc_size, &
                       full_loc_list, cell_v, siepmann%rcutsq, particle_set, &
                       cell)
      Phi_ij = siep_Phi_ij(siepmann, r_last_update_pbc, atom_a, atom_b, &
                           cell_v, cell, siepmann%rcutsq, particle_set, nr_oh, nr_h3o)
      f2 = siep_f2(siepmann, drij)
      D = siepmann%D
      E = siepmann%E

      !two-body part --> dipole term
      pot_loc_v2 = -D*f2*drij**(-3)*Phi_ij

      !three-body part
      pot_loc_v3 = E*f2*drij**(-siepmann%beta)*a_ij

      pot_loc = pot_loc_v2+pot_loc_v3

   END SUBROUTINE siepmann_energy

! **************************************************************************************************
!> \brief f2(r) corresponds to Equation (2) in J. Chem. Phys., Vol. 102, p.511
!> \param siepmann ...
!> \param r distance between oxygen and metal atom
!> \retval siep_f2 ...
! **************************************************************************************************
   FUNCTION siep_f2(siepmann, r)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: siep_f2

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siep_f2', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: rcut

      rcut = SQRT(siepmann%rcutsq)
      siep_f2 = 0.0_dp
      IF (r < rcut) THEN
         siep_f2 = EXP(siepmann%B/(r-rcut))
      END IF
   END FUNCTION siep_f2

! **************************************************************************************************
!> \brief f2(r)_d derivative of f2
!> \param siepmann ...
!> \param r distance between oxygen and metal atom
!> \retval siep_f2_d ...
! **************************************************************************************************
   FUNCTION siep_f2_d(siepmann, r)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      REAL(KIND=dp), INTENT(IN)                          :: r
      REAL(KIND=dp)                                      :: siep_f2_d

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siep_f2_d', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: B, rcut

      rcut = SQRT(siepmann%rcutsq)
      B = siepmann%B
      siep_f2_d = 0.0_dp
      IF (r < rcut) THEN
         siep_f2_d = -B*EXP(B/(r-rcut))/(r-rcut)**2
      END IF

   END FUNCTION siep_f2_d

! **************************************************************************************************
!> \brief exponential part of three-body term, see Equation (4) in J. Chem.
!>        Phys., Vol. 102, p.511
!> \param siepmann ...
!> \param r_last_update_pbc ...
!> \param iparticle ...
!> \param jparticle ...
!> \param n_loc_size ...
!> \param full_loc_list ...
!> \param cell_v ...
!> \param rcutsq ...
!> \param particle_set ...
!> \param cell ...
!> \retval siep_a_ij ...
!> \par History
!>      Using a local list of neighbors
! **************************************************************************************************
   FUNCTION siep_a_ij(siepmann, r_last_update_pbc, iparticle, jparticle, n_loc_size, &
                      full_loc_list, cell_v, rcutsq, particle_set, cell)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: iparticle, jparticle, n_loc_size
      INTEGER, DIMENSION(2, 1:n_loc_size)                :: full_loc_list
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp)                                      :: siep_a_ij

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siep_a_ij', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: ilist, kparticle
      REAL(KIND=dp)                                      :: costheta, drji, drjk, F, rab2_max, &
                                                            rji(3), rjk(3), theta

      siep_a_ij = 0.0_dp
      rab2_max = rcutsq
      F = siepmann%F
      CALL get_atomic_kind(atomic_kind=particle_set(iparticle)%atomic_kind, &
                           element_symbol=element_symbol)
      IF (element_symbol /= "O") RETURN
      rji(:) = -1.0_dp*(r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v)
      drji = SQRT(DOT_PRODUCT(rji, rji))
      DO ilist = 1, n_loc_size
         kparticle = full_loc_list(2, ilist)
         IF (kparticle == jparticle) CYCLE
         rjk(:) = pbc(r_last_update_pbc(jparticle)%r(:), r_last_update_pbc(kparticle)%r(:), cell)
         drjk = DOT_PRODUCT(rjk, rjk)
         IF (drjk > rab2_max) CYCLE
         drjk = SQRT(drjk)
         costheta = DOT_PRODUCT(rji, rjk)/(drji*drjk)
         IF (costheta < -1.0_dp) costheta = -1.0_dp
         IF (costheta > +1.0_dp) costheta = +1.0_dp
         theta = ACOS(costheta)
         siep_a_ij = siep_a_ij+EXP(F*(COS(theta/2.0_dp))**2)
      END DO
   END FUNCTION siep_a_ij

! **************************************************************************************************
!> \brief derivative of a_ij
!> \param siepmann ...
!> \param r_last_update_pbc ...
!> \param iparticle ...
!> \param jparticle ...
!> \param f_nonbond ...
!> \param prefactor ...
!> \param n_loc_size ...
!> \param full_loc_list ...
!> \param cell_v ...
!> \param cell ...
!> \param rcutsq ...
!> \param use_virial ...
!> \par History
!>       Using a local list of neighbors
! **************************************************************************************************
   SUBROUTINE siep_a_ij_d(siepmann, r_last_update_pbc, iparticle, jparticle, f_nonbond, &
                          prefactor, n_loc_size, full_loc_list, &
                          cell_v, cell, rcutsq, use_virial)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: iparticle, jparticle
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: f_nonbond
      REAL(KIND=dp), INTENT(IN)                          :: prefactor
      INTEGER, INTENT(IN)                                :: n_loc_size
      INTEGER, DIMENSION(2, 1:n_loc_size)                :: full_loc_list
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      LOGICAL, INTENT(IN)                                :: use_virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siep_a_ij_d', routineP = moduleN//':'//routineN

      INTEGER                                            :: ilist, kparticle, nparticle
      REAL(KIND=dp)                                      :: costheta, d_expterm, dcos_thetahalf, &
                                                            drji, drjk, F, rab2_max, theta
      REAL(KIND=dp), DIMENSION(3)                        :: dcosdri, dcosdrj, dcosdrk, dri, drj, &
                                                            drk, rji, rji_hat, rjk, rjk_hat

      rab2_max = rcutsq
      F = siepmann%F

      rji(:) = -1.0_dp*(r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v)
      drji = SQRT(DOT_PRODUCT(rji, rji))
      rji_hat(:) = rji(:)/drji

      nparticle = SIZE(r_last_update_pbc)
      DO ilist = 1, n_loc_size
         kparticle = full_loc_list(2, ilist)
         IF (kparticle == jparticle) CYCLE
         rjk(:) = pbc(r_last_update_pbc(jparticle)%r(:), r_last_update_pbc(kparticle)%r(:), cell)
         drjk = DOT_PRODUCT(rjk, rjk)
         IF (drjk > rab2_max) CYCLE
         drjk = SQRT(drjk)
         rjk_hat(:) = rjk(:)/drjk
         costheta = DOT_PRODUCT(rji, rjk)/(drji*drjk)
         IF (costheta < -1.0_dp) costheta = -1.0_dp
         IF (costheta > +1.0_dp) costheta = +1.0_dp

         dcosdri(:) = (1.0_dp/(drji))*(rjk_hat(:)-costheta*rji_hat(:))
         dcosdrk(:) = (1.0_dp/(drjk))*(rji_hat(:)-costheta*rjk_hat(:))
         dcosdrj(:) = -(dcosdri(:)+dcosdrk(:))

         theta = ACOS(costheta)
         dcos_thetahalf = -1.0_dp/(2.0_dp*SQRT(1-costheta**2))
         d_expterm = -2.0_dp*F*COS(theta/2.0_dp)*SIN(theta/2.0_dp) &
                     *EXP(F*(COS(theta/2.0_dp))**2)

         dri = d_expterm*dcos_thetahalf*dcosdri

         drj = d_expterm*dcos_thetahalf*dcosdrj

         drk = d_expterm*dcos_thetahalf*dcosdrk

         f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+prefactor*dri(1)
         f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+prefactor*dri(2)
         f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+prefactor*dri(3)

         f_nonbond(1, jparticle) = f_nonbond(1, jparticle)+prefactor*drj(1)
         f_nonbond(2, jparticle) = f_nonbond(2, jparticle)+prefactor*drj(2)
         f_nonbond(3, jparticle) = f_nonbond(3, jparticle)+prefactor*drj(3)

         f_nonbond(1, kparticle) = f_nonbond(1, kparticle)+prefactor*drk(1)
         f_nonbond(2, kparticle) = f_nonbond(2, kparticle)+prefactor*drk(2)
         f_nonbond(3, kparticle) = f_nonbond(3, kparticle)+prefactor*drk(3)

         IF (use_virial) THEN
            CALL cp_abort(cp__l("manybody_siepmann.F",302), &
                          "using virial with Siepmann-Sprik"// &
                          " not implemented")
         END IF
      END DO
   END SUBROUTINE siep_a_ij_d
! **************************************************************************************************
!> \brief Phi_ij corresponds to Equation (3) in J. Chem. Phys., Vol. 102, p.511
!> \param siepmann ...
!> \param r_last_update_pbc ...
!> \param iparticle ...
!> \param jparticle ...
!> \param cell_v ...
!> \param cell ...
!> \param rcutsq ...
!> \param particle_set ...
!> \param nr_oh ...
!> \param nr_h3o ...
!> \retval siep_Phi_ij ...
!> \par History
!>      Using a local list of neighbors
! **************************************************************************************************
   FUNCTION siep_Phi_ij(siepmann, r_last_update_pbc, iparticle, jparticle, &
                        cell_v, cell, rcutsq, particle_set, nr_oh, nr_h3o)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: iparticle, jparticle
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(INOUT), OPTIONAL                   :: nr_oh, nr_h3o
      REAL(KIND=dp)                                      :: siep_Phi_ij

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siep_Phi_ij', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: count_h, iatom, index_h1, index_h2, natom
      REAL(KIND=dp)                                      :: cosphi, drih, drix, drji, h_max_dist, &
                                                            rab2_max, rih(3), rih1(3), rih2(3), &
                                                            rix(3), rji(3)

      siep_Phi_ij = 0.0_dp
      count_h = 0
      index_h1 = 0
      index_h2 = 0
      rab2_max = rcutsq
      h_max_dist = 2.27_dp ! 1.2 angstrom
      natom = SIZE(particle_set)
      CALL get_atomic_kind(atomic_kind=particle_set(iparticle)%atomic_kind, &
                           element_symbol=element_symbol)
      IF (element_symbol /= "O") RETURN
      rji(:) = -1.0_dp*(r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v)
      drji = SQRT(DOT_PRODUCT(rji, rji))

      DO iatom = 1, natom
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                              element_symbol=element_symbol)
         IF (element_symbol /= "H") CYCLE
         rih(:) = pbc(r_last_update_pbc(iparticle)%r(:), r_last_update_pbc(iatom)%r(:), cell)
         drih = SQRT(DOT_PRODUCT(rih, rih))
         IF (drih >= h_max_dist) CYCLE
         count_h = count_h+1
         IF (count_h == 1) THEN
            index_h1 = iatom
         ELSEIF (count_h == 2) THEN
            index_h2 = iatom
         ENDIF
      ENDDO

      IF (count_h == 0) THEN
         CALL cp__b("manybody_siepmann.F",373,"No H atoms for O found")
      ELSEIF (count_h == 1) THEN
         IF (siepmann%allow_oh_formation) THEN
            IF (PRESENT(nr_oh)) nr_oh = nr_oh+1
            siep_Phi_ij = 0.0_dp
         ELSE
            CALL cp__b("manybody_siepmann.F",379,"Only one H atom of O atom found")
         ENDIF
      ELSEIF (count_h == 3) THEN
         IF (siepmann%allow_h3o_formation) THEN
            IF (PRESENT(nr_h3o)) nr_h3o = nr_h3o+1
            siep_Phi_ij = 0.0_dp
         ELSE
            CALL cp__b("manybody_siepmann.F",386,"Three H atoms for O atom found")
         ENDIF
      ELSEIF (count_h > 3) THEN
         CALL cp__b("manybody_siepmann.F",389,"Error in Siepmann-Sprik part: too many H atoms for O")
      ENDIF

      IF (count_h == 2) THEN
         !dipole vector rix of the H2O molecule
         rih1(:) = pbc(r_last_update_pbc(iparticle)%r(:), r_last_update_pbc(index_h1)%r(:), cell)
         rih2(:) = pbc(r_last_update_pbc(iparticle)%r(:), r_last_update_pbc(index_h2)%r(:), cell)
         rix(:) = rih1(:)+rih2(:)
         drix = SQRT(DOT_PRODUCT(rix, rix))
         cosphi = DOT_PRODUCT(rji, rix)/(drji*drix)
         IF (cosphi < -1.0_dp) cosphi = -1.0_dp
         IF (cosphi > +1.0_dp) cosphi = +1.0_dp
         siep_Phi_ij = EXP(-8.0_dp*((cosphi-1)/4.0_dp)**4)
      ENDIF
   END FUNCTION siep_Phi_ij

! **************************************************************************************************
!> \brief derivative of Phi_ij
!> \param siepmann ...
!> \param r_last_update_pbc ...
!> \param iparticle ...
!> \param jparticle ...
!> \param f_nonbond ...
!> \param prefactor ...
!> \param cell_v ...
!> \param cell ...
!> \param rcutsq ...
!> \param use_virial ...
!> \param particle_set ...
!> \par History
!>       Using a local list of neighbors
! **************************************************************************************************
   SUBROUTINE siep_Phi_ij_d(siepmann, r_last_update_pbc, iparticle, jparticle, f_nonbond, &
                            prefactor, cell_v, cell, rcutsq, use_virial, particle_set)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      INTEGER, INTENT(IN)                                :: iparticle, jparticle
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: f_nonbond
      REAL(KIND=dp), INTENT(IN)                          :: prefactor
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      LOGICAL, INTENT(IN)                                :: use_virial
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siep_Phi_ij_d', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      INTEGER                                            :: count_h, iatom, index_h1, index_h2, natom
      REAL(KIND=dp)                                      :: cosphi, dphi, drih, drix, drji, &
                                                            h_max_dist, Phi_ij, rab2_max
      REAL(KIND=dp), DIMENSION(3)                        :: dcosdrh, dcosdri, dcosdrj, drh, dri, &
                                                            drj, rih, rih1, rih2, rix, rix_hat, &
                                                            rji, rji_hat

      count_h = 0
      index_h1 = 0
      index_h2 = 0
      rab2_max = rcutsq
      h_max_dist = 2.27_dp ! 1.2 angstrom
      natom = SIZE(particle_set)
      Phi_ij = siep_Phi_ij(siepmann, r_last_update_pbc, iparticle, jparticle, &
                           cell_v, cell, rcutsq, &
                           particle_set)
      rji(:) = -1.0_dp*(r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v)
      drji = SQRT(DOT_PRODUCT(rji, rji))
      rji_hat(:) = rji(:)/drji

      DO iatom = 1, natom
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                              element_symbol=element_symbol)
         IF (element_symbol /= "H") CYCLE
         rih(:) = pbc(r_last_update_pbc(iparticle)%r(:), r_last_update_pbc(iatom)%r(:), cell)
         drih = SQRT(DOT_PRODUCT(rih, rih))
         IF (drih >= h_max_dist) CYCLE
         count_h = count_h+1
         IF (count_h == 1) THEN
            index_h1 = iatom
         ELSEIF (count_h == 2) THEN
            index_h2 = iatom
         ENDIF
      ENDDO

      IF (count_h == 0) THEN
         CALL cp__b("manybody_siepmann.F",473,"No H atoms for O found")
      ELSEIF (count_h == 1 .AND. .NOT. siepmann%allow_oh_formation) THEN
         CALL cp__b("manybody_siepmann.F",475,"Only one H atom for O atom found")
      ELSEIF (count_h == 3 .AND. .NOT. siepmann%allow_h3o_formation) THEN
         CALL cp__b("manybody_siepmann.F",477,"Three H atoms for O atom found")
      ELSEIF (count_h > 3) THEN
         CALL cp__b("manybody_siepmann.F",479,"Error in Siepmann-Sprik part: too many H atoms for O")
      ENDIF

      IF (count_h == 2) THEN
         !dipole vector rix of the H2O molecule
         rih1(:) = pbc(r_last_update_pbc(iparticle)%r(:), r_last_update_pbc(index_h1)%r(:), cell)
         rih2(:) = pbc(r_last_update_pbc(iparticle)%r(:), r_last_update_pbc(index_h2)%r(:), cell)
         rix(:) = rih1(:)+rih2(:)
         drix = SQRT(DOT_PRODUCT(rix, rix))
         rix_hat(:) = rix(:)/drix
         cosphi = DOT_PRODUCT(rji, rix)/(drji*drix)
         IF (cosphi < -1.0_dp) cosphi = -1.0_dp
         IF (cosphi > +1.0_dp) cosphi = +1.0_dp

         dcosdrj(:) = (1.0_dp/(drji))*(-rix_hat(:)+cosphi*rji_hat(:))
         ! for H atoms:
         dcosdrh(:) = (1.0_dp/(drix))*(rji_hat(:)-cosphi*rix_hat(:))
         dcosdri(:) = -dcosdrj-2.0_dp*dcosdrh

         dphi = Phi_ij*(-8.0_dp)*((cosphi-1)/4.0_dp)**3

         dri = dphi*dcosdri
         drj = dphi*dcosdrj
         drh = dphi*dcosdrh

         f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+prefactor*dri(1)
         f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+prefactor*dri(2)
         f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+prefactor*dri(3)

         f_nonbond(1, jparticle) = f_nonbond(1, jparticle)+prefactor*drj(1)
         f_nonbond(2, jparticle) = f_nonbond(2, jparticle)+prefactor*drj(2)
         f_nonbond(3, jparticle) = f_nonbond(3, jparticle)+prefactor*drj(3)

         f_nonbond(1, index_h1) = f_nonbond(1, index_h1)+prefactor*drh(1)
         f_nonbond(2, index_h1) = f_nonbond(2, index_h1)+prefactor*drh(2)
         f_nonbond(3, index_h1) = f_nonbond(3, index_h1)+prefactor*drh(3)

         f_nonbond(1, index_h2) = f_nonbond(1, index_h2)+prefactor*drh(1)
         f_nonbond(2, index_h2) = f_nonbond(2, index_h2)+prefactor*drh(2)
         f_nonbond(3, index_h2) = f_nonbond(3, index_h2)+prefactor*drh(3)

         IF (use_virial) THEN
            CALL cp_abort(cp__l("manybody_siepmann.F",521), &
                          "using virial with Siepmann-Sprik"// &
                          " not implemented")
         END IF

      ENDIF
   END SUBROUTINE siep_Phi_ij_d

! **************************************************************************************************
!> \brief forces generated by the three-body term
!> \param siepmann ...
!> \param r_last_update_pbc ...
!> \param cell_v ...
!> \param n_loc_size ...
!> \param full_loc_list ...
!> \param iparticle ...
!> \param jparticle ...
!> \param f_nonbond ...
!> \param use_virial ...
!> \param rcutsq ...
!> \param cell ...
!> \param particle_set ...
!> \par History
!>       Using a local list of neighbors
! **************************************************************************************************
   SUBROUTINE siepmann_forces_v3(siepmann, r_last_update_pbc, cell_v, n_loc_size, &
                                 full_loc_list, iparticle, jparticle, f_nonbond, &
                                 use_virial, rcutsq, cell, particle_set)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      INTEGER, INTENT(IN)                                :: n_loc_size
      INTEGER, DIMENSION(2, 1:n_loc_size)                :: full_loc_list
      INTEGER, INTENT(IN)                                :: iparticle, jparticle
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: f_nonbond
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siepmann_forces_v3', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      REAL(KIND=dp)                                      :: a_ij, beta, drji, E, f2, f2_d, f_A1, &
                                                            f_A2, fac, prefactor, rji(3), &
                                                            rji_hat(3)

      beta = siepmann%beta
      E = siepmann%E

      CALL get_atomic_kind(atomic_kind=particle_set(iparticle)%atomic_kind, &
                           element_symbol=element_symbol)
      IF (element_symbol /= "O") RETURN

      rji(:) = -1.0_dp*(r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v)
      drji = SQRT(DOT_PRODUCT(rji, rji))
      rji_hat(:) = rji(:)/drji

      fac = -1.0_dp !gradient to force
      a_ij = siep_a_ij(siepmann, r_last_update_pbc, iparticle, jparticle, n_loc_size, &
                       full_loc_list, cell_v, rcutsq, particle_set, cell)
      f2 = siep_f2(siepmann, drji)
      f2_d = siep_f2_d(siepmann, drji)

      ! Lets do the f_A1 piece derivative of  f2
      f_A1 = E*f2_d*drji**(-beta)*a_ij*fac*(1.0_dp/drji)
      f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+f_A1*rji(1)
      f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+f_A1*rji(2)
      f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+f_A1*rji(3)
      f_nonbond(1, jparticle) = f_nonbond(1, jparticle)-f_A1*rji(1)
      f_nonbond(2, jparticle) = f_nonbond(2, jparticle)-f_A1*rji(2)
      f_nonbond(3, jparticle) = f_nonbond(3, jparticle)-f_A1*rji(3)

      IF (use_virial) THEN
         CALL cp_abort(cp__l("manybody_siepmann.F",596), &
                       "using virial with Siepmann-Sprik"// &
                       " not implemented")
      END IF

      ! Lets do the f_A2 piece derivative of rji**(-beta)
      f_A2 = E*f2*(-beta)*drji**(-beta-1)*a_ij*fac*(1.0_dp/drji)
      f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+f_A2*rji(1)
      f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+f_A2*rji(2)
      f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+f_A2*rji(3)
      f_nonbond(1, jparticle) = f_nonbond(1, jparticle)-f_A2*rji(1)
      f_nonbond(2, jparticle) = f_nonbond(2, jparticle)-f_A2*rji(2)
      f_nonbond(3, jparticle) = f_nonbond(3, jparticle)-f_A2*rji(3)

      IF (use_virial) THEN
         CALL cp_abort(cp__l("manybody_siepmann.F",611), &
                       "using virial with Siepmann-Sprik"// &
                       " not implemented")
      END IF

      ! Lets do the f_A3 piece derivative: of a_ij
      prefactor = E*f2*drji**(-beta)*fac
      CALL siep_a_ij_d(siepmann, r_last_update_pbc, iparticle, jparticle, f_nonbond, &
                       prefactor, n_loc_size, full_loc_list, cell_v, &
                       cell, rcutsq, use_virial)

   END SUBROUTINE siepmann_forces_v3

! **************************************************************************************************
!> \brief forces generated by the dipole term
!> \param siepmann ...
!> \param r_last_update_pbc ...
!> \param cell_v ...
!> \param cell ...
!> \param iparticle ...
!> \param jparticle ...
!> \param f_nonbond ...
!> \param use_virial ...
!> \param rcutsq ...
!> \param particle_set ...
!> \par History
!>       Using a local list of neighbors
! **************************************************************************************************
   SUBROUTINE siepmann_forces_v2(siepmann, r_last_update_pbc, cell_v, cell, &
                                 iparticle, jparticle, f_nonbond, use_virial, rcutsq, particle_set)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: iparticle, jparticle
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: f_nonbond
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), INTENT(IN)                          :: rcutsq
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'siepmann_forces_v2', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      REAL(KIND=dp)                                      :: D, drji, f2, f2_d, f_A1, f_A2, fac, &
                                                            Phi_ij, prefactor, rji(3)

      D = siepmann%D

      CALL get_atomic_kind(atomic_kind=particle_set(iparticle)%atomic_kind, &
                           element_symbol=element_symbol)
      IF (element_symbol /= "O") RETURN

      rji(:) = -1.0_dp*(r_last_update_pbc(jparticle)%r(:)-r_last_update_pbc(iparticle)%r(:)+cell_v)
      drji = SQRT(DOT_PRODUCT(rji, rji))

      fac = -1.0_dp
      Phi_ij = siep_Phi_ij(siepmann, r_last_update_pbc, iparticle, jparticle, &
                           cell_v, cell, rcutsq, particle_set)
      f2 = siep_f2(siepmann, drji)
      f2_d = siep_f2_d(siepmann, drji)

      ! Lets do the f_A1 piece derivative of  f2
      f_A1 = -D*f2_d*drji**(-3)*Phi_ij*fac*(1.0_dp/drji)
      f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+f_A1*rji(1)
      f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+f_A1*rji(2)
      f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+f_A1*rji(3)
      f_nonbond(1, jparticle) = f_nonbond(1, jparticle)-f_A1*rji(1)
      f_nonbond(2, jparticle) = f_nonbond(2, jparticle)-f_A1*rji(2)
      f_nonbond(3, jparticle) = f_nonbond(3, jparticle)-f_A1*rji(3)

      IF (use_virial) THEN
         CALL cp_abort(cp__l("manybody_siepmann.F",683), &
                       "using virial with Siepmann-Sprik"// &
                       " not implemented")
      END IF

!   ! Lets do the f_A2 piece derivative of rji**(-3)
      f_A2 = -D*f2*(-3.0_dp)*drji**(-4)*Phi_ij*fac*(1.0_dp/drji)
      f_nonbond(1, iparticle) = f_nonbond(1, iparticle)+f_A2*rji(1)
      f_nonbond(2, iparticle) = f_nonbond(2, iparticle)+f_A2*rji(2)
      f_nonbond(3, iparticle) = f_nonbond(3, iparticle)+f_A2*rji(3)
      f_nonbond(1, jparticle) = f_nonbond(1, jparticle)-f_A2*rji(1)
      f_nonbond(2, jparticle) = f_nonbond(2, jparticle)-f_A2*rji(2)
      f_nonbond(3, jparticle) = f_nonbond(3, jparticle)-f_A2*rji(3)

      IF (use_virial) THEN
         CALL cp_abort(cp__l("manybody_siepmann.F",698), &
                       "using virial with Siepmann-Sprik"// &
                       " not implemented")
      END IF

      ! Lets do the f_A3 piece derivative: of Phi_ij
      prefactor = -D*f2*drji**(-3)*fac
      CALL siep_Phi_ij_d(siepmann, r_last_update_pbc, iparticle, jparticle, f_nonbond, &
                         prefactor, cell_v, cell, &
                         rcutsq, use_virial, particle_set)

   END SUBROUTINE siepmann_forces_v2

! **************************************************************************************************
!> \brief ...
!> \param nonbonded ...
!> \param potparm ...
!> \param glob_loc_list ...
!> \param glob_cell_v ...
!> \param glob_loc_list_a ...
!> \param cell ...
!> \par History
! **************************************************************************************************
   SUBROUTINE setup_siepmann_arrays(nonbonded, potparm, glob_loc_list, glob_cell_v, &
                                    glob_loc_list_a, cell)
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      INTEGER, DIMENSION(:, :), POINTER                  :: glob_loc_list
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: glob_cell_v
      INTEGER, DIMENSION(:), POINTER                     :: glob_loc_list_a
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_siepmann_arrays', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iend, igrp, ikind, ilist, &
                                                            ipair, istart, jkind, nkinds, npairs, &
                                                            npairs_tot
      INTEGER, DIMENSION(:), POINTER                     :: work_list, work_list2
      INTEGER, DIMENSION(:, :), POINTER                  :: list
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v, cvi
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rwork_list
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      TYPE(pair_potential_single_type), POINTER          :: pot

      IF(.NOT.(.NOT. ASSOCIATED(glob_loc_list)))CALL cp__a("manybody_siepmann.F",743)
      IF(.NOT.(.NOT. ASSOCIATED(glob_loc_list_a)))CALL cp__a("manybody_siepmann.F",744)
      IF(.NOT.(.NOT. ASSOCIATED(glob_cell_v)))CALL cp__a("manybody_siepmann.F",745)
      CALL timeset(routineN, handle)
      npairs_tot = 0
      nkinds = SIZE(potparm%pot, 1)
      DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         npairs = neighbor_kind_pair%npairs
         IF (npairs == 0) CYCLE
         Kind_Group_Loop1: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            iend = neighbor_kind_pair%grp_kind_end(igrp)
            ikind = neighbor_kind_pair%ij_kind(1, igrp)
            jkind = neighbor_kind_pair%ij_kind(2, igrp)
            pot => potparm%pot(ikind, jkind)%pot
            npairs = iend-istart+1
            IF (pot%no_mb) CYCLE
            DO i = 1, SIZE(pot%type)
               IF (pot%type(i) == siepmann_type) npairs_tot = npairs_tot+npairs
            END DO
         END DO Kind_Group_Loop1
      END DO
      ALLOCATE (work_list(npairs_tot))
      ALLOCATE (work_list2(npairs_tot))
      ALLOCATE (glob_loc_list(2, npairs_tot))
      ALLOCATE (glob_cell_v(3, npairs_tot))
      ! Fill arrays with data
      npairs_tot = 0
      DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         npairs = neighbor_kind_pair%npairs
         IF (npairs == 0) CYCLE
         Kind_Group_Loop2: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            iend = neighbor_kind_pair%grp_kind_end(igrp)
            ikind = neighbor_kind_pair%ij_kind(1, igrp)
            jkind = neighbor_kind_pair%ij_kind(2, igrp)
            list => neighbor_kind_pair%list
            cvi = neighbor_kind_pair%cell_vector
            pot => potparm%pot(ikind, jkind)%pot
            npairs = iend-istart+1
            IF (pot%no_mb) CYCLE
            CALL matvec_3x3(cell_v, cell%hmat, cvi)
            DO i = 1, SIZE(pot%type)
               ! SIEPMANN
               IF (pot%type(i) == siepmann_type) THEN
                  DO ipair = 1, npairs
                     glob_loc_list(:, npairs_tot+ipair) = list(:, istart-1+ipair)
                     glob_cell_v(1:3, npairs_tot+ipair) = cell_v(1:3)
                  END DO
                  npairs_tot = npairs_tot+npairs
               END IF
            END DO
         END DO Kind_Group_Loop2
      END DO
      ! Order the arrays w.r.t. the first index of glob_loc_list
      CALL sort(glob_loc_list(1, :), npairs_tot, work_list)
      DO ipair = 1, npairs_tot
         work_list2(ipair) = glob_loc_list(2, work_list(ipair))
      END DO
      glob_loc_list(2, :) = work_list2
      DEALLOCATE (work_list2)
      ALLOCATE (rwork_list(3, npairs_tot))
      DO ipair = 1, npairs_tot
         rwork_list(:, ipair) = glob_cell_v(:, work_list(ipair))
      END DO
      glob_cell_v = rwork_list
      DEALLOCATE (rwork_list)
      DEALLOCATE (work_list)
      ALLOCATE (glob_loc_list_a(npairs_tot))
      glob_loc_list_a = glob_loc_list(1, :)
      CALL timestop(handle)
   END SUBROUTINE setup_siepmann_arrays

! **************************************************************************************************
!> \brief ...
!> \param glob_loc_list ...
!> \param glob_cell_v ...
!> \param glob_loc_list_a ...
! **************************************************************************************************
   SUBROUTINE destroy_siepmann_arrays(glob_loc_list, glob_cell_v, glob_loc_list_a)
      INTEGER, DIMENSION(:, :), POINTER                  :: glob_loc_list
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: glob_cell_v
      INTEGER, DIMENSION(:), POINTER                     :: glob_loc_list_a

      CHARACTER(LEN=*), PARAMETER :: routineN = 'destroy_siepmann_arrays', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(glob_loc_list)) THEN
         DEALLOCATE (glob_loc_list)
      END IF
      IF (ASSOCIATED(glob_loc_list_a)) THEN
         DEALLOCATE (glob_loc_list_a)
      END IF
      IF (ASSOCIATED(glob_cell_v)) THEN
         DEALLOCATE (glob_cell_v)
      END IF

   END SUBROUTINE destroy_siepmann_arrays

! **************************************************************************************************
!> \brief prints the number of OH- ions or H3O+ ions near surface
!> \param nr_ions ... number of ions
!> \param mm_section ...
!> \param para_env ...
!> \param count_oh flag indicating if number OH- or H3O+ is printed
! **************************************************************************************************
   SUBROUTINE print_nr_ions_siepmann(nr_ions, mm_section, para_env, count_oh)
      INTEGER, INTENT(INOUT)                             :: nr_ions
      TYPE(section_vals_type), POINTER                   :: mm_section
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      LOGICAL, INTENT(IN)                                :: count_oh

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_nr_ions_siepmann', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iw
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)

      CALL mp_sum(nr_ions, para_env%group)
      logger => cp_get_default_logger()

      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%PROGRAM_RUN_INFO", &
                                extension=".mmLog")

      IF (iw > 0 .AND. nr_ions > 0 .AND. count_oh) THEN
         WRITE (iw, '(/,A,T71,I10,/)') " SIEPMANN: number of OH- ions at surface", nr_ions
      ENDIF
      IF (iw > 0 .AND. nr_ions > 0 .AND. .NOT. count_oh) THEN
         WRITE (iw, '(/,A,T71,I10,/)') " SIEPMANN: number of H3O+ ions at surface", nr_ions
      ENDIF

      CALL cp_print_key_finished_output(iw, logger, mm_section, "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE print_nr_ions_siepmann

END MODULE manybody_siepmann

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Efficient tersoff implementation and general "lifting" of manybody_potential module
!>      12.2007 [tlaino] - Splitting manybody module : In this module we should only
!>                         keep the main routines for computing energy and forces of
!>                         manybody potentials. Each potential should have his own module!
!> \author CJM, I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
MODULE manybody_potential

   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                              neighbor_kind_pairs_type
   USE fist_nonbond_env_types,          ONLY: eam_type,&
                                              fist_nonbond_env_get,&
                                              fist_nonbond_env_type,&
                                              pos_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE manybody_eam,                    ONLY: get_force_eam
   USE manybody_quip,                   ONLY: quip_add_force_virial,&
                                              quip_energy_store_force_virial
   USE manybody_siepmann,               ONLY: destroy_siepmann_arrays,&
                                              print_nr_ions_siepmann,&
                                              setup_siepmann_arrays,&
                                              siepmann_energy,&
                                              siepmann_forces_v2,&
                                              siepmann_forces_v3
   USE manybody_tersoff,                ONLY: destroy_tersoff_arrays,&
                                              setup_tersoff_arrays,&
                                              tersoff_energy,&
                                              tersoff_forces
   USE mathlib,                         ONLY: matvec_3x3
   USE message_passing,                 ONLY: mp_sum
   USE pair_potential_types,            ONLY: &
        ea_type, eam_pot_type, pair_potential_pp_type, pair_potential_single_type, quip_type, &
        siepmann_pot_type, siepmann_type, tersoff_pot_type, tersoff_type
   USE particle_types,                  ONLY: particle_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: energy_manybody
   PUBLIC :: force_nonbond_manybody
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'manybody_potential'

CONTAINS

! **************************************************************************************************
!> \brief computes the embedding contribution to the energy
!> \param fist_nonbond_env ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param particle_set ...
!> \param cell ...
!> \param pot_manybody ...
!> \param para_env ...
!> \param mm_section ...
!> \par History
!>      tlaino [2007] - New algorithm for tersoff potential
!> \author CJM, I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
   SUBROUTINE energy_manybody(fist_nonbond_env, atomic_kind_set, local_particles, &
                              particle_set, cell, pot_manybody, para_env, mm_section)

      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(cell_type), POINTER                           :: cell
      REAL(dp), INTENT(INOUT)                            :: pot_manybody
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'energy_manybody', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, handle, i, iend, ifirst, igrp, ikind, ilast, ilist, indexa, &
         ipair, iparticle, iparticle_local, istart, iunique, jkind, junique, mpair, nkinds, &
         nloc_size, npairs, nparticle, nparticle_local, nr_h3O, nr_oh, nunique
      INTEGER, DIMENSION(:), POINTER                     :: glob_loc_list_a, work_list
      INTEGER, DIMENSION(:, :), POINTER                  :: glob_loc_list, list, sort_list
      LOGICAL                                            :: allow_h3o_formation, allow_oh_formation, &
                                                            any_quip, any_siepmann, any_tersoff
      REAL(KIND=dp)                                      :: drij, embed, pot_loc, pot_quip, qr, &
                                                            rab2_max, rij(3)
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v, cvi
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: glob_cell_v
      REAL(KIND=dp), POINTER                             :: fembed(:)
      TYPE(eam_pot_type), POINTER                        :: eam
      TYPE(eam_type), DIMENSION(:), POINTER              :: eam_data
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      TYPE(pair_potential_single_type), POINTER          :: pot
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(tersoff_pot_type), POINTER                    :: tersoff

      NULLIFY (eam, siepmann, tersoff)
      any_tersoff = .FALSE.
      any_siepmann = .FALSE.
      any_quip = .FALSE.
      allow_oh_formation = .FALSE.
      allow_h3O_formation = .FALSE.
      CALL timeset(routineN, handle)
      CALL fist_nonbond_env_get(fist_nonbond_env, r_last_update_pbc=r_last_update_pbc, &
                                potparm=potparm, eam_data=eam_data)
      ! EAM requires a single loop
      DO ikind = 1, SIZE(atomic_kind_set)
         pot => potparm%pot(ikind, ikind)%pot
         DO i = 1, SIZE(pot%type)
            IF (pot%type(i) /= ea_type) CYCLE
            eam => pot%set(i)%eam
            nparticle = SIZE(particle_set)
            ALLOCATE (fembed(nparticle))
            fembed(:) = 0._dp
            IF(.NOT.(ASSOCIATED(eam_data)))CALL cp__a("manybody_potential.F",128)
            ! computation of embedding function and energy
            nparticle_local = local_particles%n_el(ikind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(ikind)%array(iparticle_local)
               indexa = INT(eam_data(iparticle)%rho/eam%drhoar)+1
               IF (indexa > eam%npoints-1) indexa = eam%npoints-1
               qr = eam_data(iparticle)%rho-eam%rhoval(indexa)

               embed = eam%frho(indexa)+qr*eam%frhop(indexa)
               fembed(iparticle) = eam%frhop(indexa)+qr*(eam%frhop(indexa+1)-eam%frhop(indexa))/eam%drhoar

               pot_manybody = pot_manybody+embed
            END DO
            ! communicate data
            CALL mp_sum(fembed, para_env%group)
            DO iparticle = 1, nparticle
               IF (particle_set(iparticle)%atomic_kind%kind_number == ikind) THEN
                  eam_data(iparticle)%f_embed = fembed(iparticle)
               END IF
            END DO

            DEALLOCATE (fembed)
         END DO
      END DO
      ! Other manybody potential
      DO ikind = 1, SIZE(atomic_kind_set)
         DO jkind = ikind, SIZE(atomic_kind_set)
            pot => potparm%pot(ikind, jkind)%pot
            any_tersoff = any_tersoff .OR. ANY(pot%type == tersoff_type)
            any_quip = any_quip .OR. ANY(pot%type == quip_type)
            any_siepmann = any_siepmann .OR. ANY(pot%type == siepmann_type)
         END DO
      END DO
      CALL fist_nonbond_env_get(fist_nonbond_env, nonbonded=nonbonded, natom_types=nkinds)
      ! QUIP
      IF (any_quip) THEN
         CALL quip_energy_store_force_virial(particle_set, cell, atomic_kind_set, potparm, &
                                             fist_nonbond_env, pot_quip, para_env)
         pot_manybody = pot_manybody+pot_quip
      ENDIF
      ! TERSOFF
      IF (any_tersoff) THEN
         NULLIFY (glob_loc_list, glob_cell_v, glob_loc_list_a)
         CALL setup_tersoff_arrays(nonbonded, potparm, glob_loc_list, glob_cell_v, glob_loc_list_a, cell)
         DO ilist = 1, nonbonded%nlists
            neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
            npairs = neighbor_kind_pair%npairs
            IF (npairs == 0) CYCLE
            Kind_Group_Loop: DO igrp = 1, neighbor_kind_pair%ngrp_kind
               istart = neighbor_kind_pair%grp_kind_start(igrp)
               iend = neighbor_kind_pair%grp_kind_end(igrp)
               ikind = neighbor_kind_pair%ij_kind(1, igrp)
               jkind = neighbor_kind_pair%ij_kind(2, igrp)
               list => neighbor_kind_pair%list
               cvi = neighbor_kind_pair%cell_vector
               pot => potparm%pot(ikind, jkind)%pot
               DO i = 1, SIZE(pot%type)
                  IF (pot%type(i) /= tersoff_type) CYCLE
                  rab2_max = pot%set(i)%tersoff%rcutsq
                  CALL matvec_3x3(cell_v, cell%hmat, cvi)
                  pot => potparm%pot(ikind, jkind)%pot
                  tersoff => pot%set(i)%tersoff
                  npairs = iend-istart+1
                  IF (npairs /= 0) THEN
                     ALLOCATE (sort_list(2, npairs), work_list(npairs))
                     sort_list = list(:, istart:iend)
                     ! Sort the list of neighbors, this increases the efficiency for single
                     ! potential contributions
                     CALL sort(sort_list(1, :), npairs, work_list)
                     DO ipair = 1, npairs
                        work_list(ipair) = sort_list(2, work_list(ipair))
                     END DO
                     sort_list(2, :) = work_list
                     ! find number of unique elements of array index 1
                     nunique = 1
                     DO ipair = 1, npairs-1
                        IF (sort_list(1, ipair+1) /= sort_list(1, ipair)) nunique = nunique+1
                     END DO
                     ipair = 1
                     junique = sort_list(1, ipair)
                     ifirst = 1
                     DO iunique = 1, nunique
                        atom_a = junique
                        IF (glob_loc_list_a(ifirst) > atom_a) CYCLE
                        DO mpair = ifirst, SIZE(glob_loc_list_a)
                           IF (glob_loc_list_a(mpair) == atom_a) EXIT
                        END DO
                        ifirst = mpair
                        DO mpair = ifirst, SIZE(glob_loc_list_a)
                           IF (glob_loc_list_a(mpair) /= atom_a) EXIT
                        END DO
                        ilast = mpair-1
                        nloc_size = 0
                        IF (ifirst /= 0) nloc_size = ilast-ifirst+1
                        DO WHILE (ipair <= npairs)
                           IF (sort_list(1, ipair) /= junique) EXIT
                           atom_b = sort_list(2, ipair)
                           ! Energy terms
                           pot_loc = 0.0_dp
                           rij(:) = r_last_update_pbc(atom_b)%r(:)-r_last_update_pbc(atom_a)%r(:)+cell_v
                           drij = DOT_PRODUCT(rij, rij)
                           ipair = ipair+1
                           IF (drij > rab2_max) CYCLE
                           drij = SQRT(drij)
                           CALL tersoff_energy(pot_loc, tersoff, r_last_update_pbc, atom_a, atom_b, nloc_size, &
                                               glob_loc_list(:, ifirst:ilast), glob_cell_v(:, ifirst:ilast), cell_v, drij)
                           pot_manybody = pot_manybody+0.5_dp*pot_loc
                        END DO
                        ifirst = ilast+1
                        IF (ipair <= npairs) junique = sort_list(1, ipair)
                     END DO
                     DEALLOCATE (sort_list, work_list)
                  END IF
               END DO
            END DO Kind_Group_Loop
         END DO
         CALL destroy_tersoff_arrays(glob_loc_list, glob_cell_v, glob_loc_list_a)
      END IF

      !SIEPMANN POTENTIAL
      IF (any_siepmann) THEN
         NULLIFY (glob_loc_list, glob_cell_v, glob_loc_list_a)
         CALL setup_siepmann_arrays(nonbonded, potparm, glob_loc_list, glob_cell_v, glob_loc_list_a, cell)
         nr_oh = 0
         nr_h3O = 0
         DO ilist = 1, nonbonded%nlists
            neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
            npairs = neighbor_kind_pair%npairs
            IF (npairs == 0) CYCLE
            Kind_Group_Loop_2: DO igrp = 1, neighbor_kind_pair%ngrp_kind
               istart = neighbor_kind_pair%grp_kind_start(igrp)
               iend = neighbor_kind_pair%grp_kind_end(igrp)
               ikind = neighbor_kind_pair%ij_kind(1, igrp)
               jkind = neighbor_kind_pair%ij_kind(2, igrp)
               list => neighbor_kind_pair%list
               cvi = neighbor_kind_pair%cell_vector
               pot => potparm%pot(ikind, jkind)%pot
               DO i = 1, SIZE(pot%type)
                  IF (pot%type(i) /= siepmann_type) CYCLE
                  rab2_max = pot%set(i)%siepmann%rcutsq
                  CALL matvec_3x3(cell_v, cell%hmat, cvi)
                  pot => potparm%pot(ikind, jkind)%pot
                  siepmann => pot%set(i)%siepmann
                  IF (siepmann%allow_oh_formation) allow_oh_formation = .TRUE.
                  IF (siepmann%allow_h3o_formation) allow_h3o_formation = .TRUE.
                  npairs = iend-istart+1
                  IF (npairs /= 0) THEN
                     ALLOCATE (sort_list(2, npairs), work_list(npairs))
                     sort_list = list(:, istart:iend)
                     ! Sort the list of neighbors, this increases the efficiency for single
                     ! potential contributions
                     CALL sort(sort_list(1, :), npairs, work_list)
                     DO ipair = 1, npairs
                        work_list(ipair) = sort_list(2, work_list(ipair))
                     END DO
                     sort_list(2, :) = work_list
                     ! find number of unique elements of array index 1
                     nunique = 1
                     DO ipair = 1, npairs-1
                        IF (sort_list(1, ipair+1) /= sort_list(1, ipair)) nunique = nunique+1
                     END DO
                     ipair = 1
                     junique = sort_list(1, ipair)
                     ifirst = 1
                     DO iunique = 1, nunique
                        atom_a = junique
                        IF (glob_loc_list_a(ifirst) > atom_a) CYCLE
                        DO mpair = ifirst, SIZE(glob_loc_list_a)
                           IF (glob_loc_list_a(mpair) == atom_a) EXIT
                        END DO
                        ifirst = mpair
                        DO mpair = ifirst, SIZE(glob_loc_list_a)
                           IF (glob_loc_list_a(mpair) /= atom_a) EXIT
                        END DO
                        ilast = mpair-1
                        nloc_size = 0
                        IF (ifirst /= 0) nloc_size = ilast-ifirst+1
                        DO WHILE (ipair <= npairs)
                           IF (sort_list(1, ipair) /= junique) EXIT
                           atom_b = sort_list(2, ipair)
                           ! Energy terms
                           pot_loc = 0.0_dp
                           rij(:) = r_last_update_pbc(atom_b)%r(:)-r_last_update_pbc(atom_a)%r(:)+cell_v
                           drij = DOT_PRODUCT(rij, rij)
                           ipair = ipair+1
                           IF (drij > rab2_max) CYCLE
                           drij = SQRT(drij)
                           CALL siepmann_energy(pot_loc, siepmann, r_last_update_pbc, atom_a, atom_b, nloc_size, &
                                                glob_loc_list(:, ifirst:ilast), cell_v, cell, drij, &
                                                particle_set, nr_oh, nr_h3O)
                           pot_manybody = pot_manybody+pot_loc
                        END DO
                        ifirst = ilast+1
                        IF (ipair <= npairs) junique = sort_list(1, ipair)
                     END DO
                     DEALLOCATE (sort_list, work_list)
                  END IF
               END DO
            END DO Kind_Group_Loop_2
         END DO
         CALL destroy_siepmann_arrays(glob_loc_list, glob_cell_v, glob_loc_list_a)
         IF (allow_oh_formation) THEN
            CALL print_nr_ions_siepmann(nr_oh, mm_section, para_env, count_oh=.TRUE.)
         ENDIF
         IF (allow_h3o_formation) THEN
            CALL print_nr_ions_siepmann(nr_h3o, mm_section, para_env, count_oh=.FALSE.)
         ENDIF
      END IF

      CALL timestop(handle)
   END SUBROUTINE energy_manybody

! **************************************************************************************************
!> \brief ...
!> \param fist_nonbond_env ...
!> \param particle_set ...
!> \param cell ...
!> \param f_nonbond ...
!> \param pv_nonbond ...
!> \param use_virial ...
!> \par History
!>      Fast implementation of the tersoff potential - [tlaino] 2007
!> \author I-Feng W. Kuo, Teodoro Laino
! **************************************************************************************************
   SUBROUTINE force_nonbond_manybody(fist_nonbond_env, particle_set, cell, &
                                     f_nonbond, pv_nonbond, use_virial)

      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: f_nonbond, pv_nonbond
      LOGICAL, INTENT(IN)                                :: use_virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'force_nonbond_manybody', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, handle, i, i_a, i_b, iend, ifirst, igrp, ikind, ilast, ilist, &
         ipair, istart, iunique, jkind, junique, kind_a, kind_b, mpair, nkinds, nloc_size, npairs, &
         nunique
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: eam_kinds_index
      INTEGER, DIMENSION(:), POINTER                     :: glob_loc_list_a, work_list
      INTEGER, DIMENSION(:, :), POINTER                  :: glob_loc_list, list, sort_list
      LOGICAL                                            :: any_quip, any_siepmann, any_tersoff
      REAL(KIND=dp) :: f_eam, fac, fr(3), ptens11, ptens12, ptens13, ptens21, ptens22, ptens23, &
         ptens31, ptens32, ptens33, rab(3), rab2, rab2_max, rtmp(3)
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v, cvi
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: glob_cell_v
      TYPE(eam_pot_type), POINTER                        :: eam_a, eam_b
      TYPE(eam_type), DIMENSION(:), POINTER              :: eam_data
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      TYPE(pair_potential_single_type), POINTER          :: pot
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update_pbc
      TYPE(siepmann_pot_type), POINTER                   :: siepmann
      TYPE(tersoff_pot_type), POINTER                    :: tersoff

      any_tersoff = .FALSE.
      any_quip = .FALSE.
      any_siepmann = .FALSE.
      CALL timeset(routineN, handle)
      NULLIFY (eam_a, eam_b, tersoff, siepmann)

      CALL fist_nonbond_env_get(fist_nonbond_env, nonbonded=nonbonded, potparm=potparm, &
                                natom_types=nkinds, eam_data=eam_data, r_last_update_pbc=r_last_update_pbc)

      ! Initializing the potential energy, pressure tensor and force
      IF (use_virial) THEN
         ptens11 = 0.0_dp; ptens12 = 0.0_dp; ptens13 = 0.0_dp
         ptens21 = 0.0_dp; ptens22 = 0.0_dp; ptens23 = 0.0_dp
         ptens31 = 0.0_dp; ptens32 = 0.0_dp; ptens33 = 0.0_dp
      END IF

      nkinds = SIZE(potparm%pot, 1)
      ALLOCATE (eam_kinds_index(nkinds, nkinds))
      eam_kinds_index = -1
      DO ikind = 1, nkinds
         DO jkind = ikind, nkinds
            DO i = 1, SIZE(potparm%pot(ikind, jkind)%pot%type)
               IF (potparm%pot(ikind, jkind)%pot%type(i) == ea_type) THEN
                  ! At the moment we allow only 1 EAM per each kinds pair..
                  IF(.NOT.(eam_kinds_index(ikind, jkind) == -1))CALL cp__a("manybody_potential.F",410)
                  IF(.NOT.(eam_kinds_index(jkind, ikind) == -1))CALL cp__a("manybody_potential.F",411)
                  eam_kinds_index(ikind, jkind) = i
                  eam_kinds_index(jkind, ikind) = i
               END IF
            END DO
         END DO
      END DO

      ! QUIP
      IF (use_virial) &
         CALL quip_add_force_virial(fist_nonbond_env, f_nonbond, pv_nonbond)

      ! starting the force loop
      DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         npairs = neighbor_kind_pair%npairs
         IF (npairs == 0) CYCLE
         Kind_Group_Loop1: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            iend = neighbor_kind_pair%grp_kind_end(igrp)
            ikind = neighbor_kind_pair%ij_kind(1, igrp)
            jkind = neighbor_kind_pair%ij_kind(2, igrp)
            list => neighbor_kind_pair%list
            cvi = neighbor_kind_pair%cell_vector
            pot => potparm%pot(ikind, jkind)%pot
            IF (pot%no_mb) CYCLE
            rab2_max = pot%rcutsq
            CALL matvec_3x3(cell_v, cell%hmat, cvi)
            any_tersoff = any_tersoff .OR. ANY(pot%type == tersoff_type)
            any_siepmann = any_siepmann .OR. ANY(pot%type == siepmann_type)
            i = eam_kinds_index(ikind, jkind)
            IF (i == -1) CYCLE
            ! EAM
            IF(.NOT.(ASSOCIATED(eam_data)))CALL cp__a("manybody_potential.F",444)
            DO ipair = istart, iend
               atom_a = list(1, ipair)
               atom_b = list(2, ipair)
               fac = 1.0_dp
               IF (atom_a == atom_b) fac = 0.5_dp
               kind_a = particle_set(atom_a)%atomic_kind%kind_number
               kind_b = particle_set(atom_b)%atomic_kind%kind_number
               i_a = eam_kinds_index(kind_a, kind_a)
               i_b = eam_kinds_index(kind_b, kind_b)
               eam_a => potparm%pot(kind_a, kind_a)%pot%set(i_a)%eam
               eam_b => potparm%pot(kind_b, kind_b)%pot%set(i_b)%eam

               !set this outside the potential type in case need multiple potentials
               !Do everything necessary for EAM here
               rab = r_last_update_pbc(atom_b)%r-r_last_update_pbc(atom_a)%r
               rab = rab+cell_v
               rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
               IF (rab2 <= rab2_max) THEN
                  CALL get_force_eam(rab2, eam_a, eam_b, eam_data, atom_a, atom_b, f_eam)
                  f_eam = f_eam*fac

                  fr(1) = -f_eam*rab(1)
                  fr(2) = -f_eam*rab(2)
                  fr(3) = -f_eam*rab(3)
                  f_nonbond(1, atom_a) = f_nonbond(1, atom_a)-fr(1)
                  f_nonbond(2, atom_a) = f_nonbond(2, atom_a)-fr(2)
                  f_nonbond(3, atom_a) = f_nonbond(3, atom_a)-fr(3)

                  f_nonbond(1, atom_b) = f_nonbond(1, atom_b)+fr(1)
                  f_nonbond(2, atom_b) = f_nonbond(2, atom_b)+fr(2)
                  f_nonbond(3, atom_b) = f_nonbond(3, atom_b)+fr(3)
                  IF (use_virial) THEN
                     ptens11 = ptens11+rab(1)*fr(1)
                     ptens21 = ptens21+rab(2)*fr(1)
                     ptens31 = ptens31+rab(3)*fr(1)
                     ptens12 = ptens12+rab(1)*fr(2)
                     ptens22 = ptens22+rab(2)*fr(2)
                     ptens32 = ptens32+rab(3)*fr(2)
                     ptens13 = ptens13+rab(1)*fr(3)
                     ptens23 = ptens23+rab(2)*fr(3)
                     ptens33 = ptens33+rab(3)*fr(3)
                  END IF
               ENDIF
            END DO
         END DO Kind_Group_Loop1
      END DO
      DEALLOCATE (eam_kinds_index)

      ! Special way of handling the tersoff potential..
      IF (any_tersoff) THEN
         NULLIFY (glob_loc_list, glob_cell_v, glob_loc_list_a)
         CALL setup_tersoff_arrays(nonbonded, potparm, glob_loc_list, glob_cell_v, glob_loc_list_a, cell)
         DO ilist = 1, nonbonded%nlists
            neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
            npairs = neighbor_kind_pair%npairs
            IF (npairs == 0) CYCLE
            Kind_Group_Loop2: DO igrp = 1, neighbor_kind_pair%ngrp_kind
               istart = neighbor_kind_pair%grp_kind_start(igrp)
               iend = neighbor_kind_pair%grp_kind_end(igrp)
               ikind = neighbor_kind_pair%ij_kind(1, igrp)
               jkind = neighbor_kind_pair%ij_kind(2, igrp)
               list => neighbor_kind_pair%list
               cvi = neighbor_kind_pair%cell_vector
               pot => potparm%pot(ikind, jkind)%pot

               IF (pot%no_mb) CYCLE
               rab2_max = pot%rcutsq
               CALL matvec_3x3(cell_v, cell%hmat, cvi)
               DO i = 1, SIZE(pot%type)
                  ! TERSOFF
                  IF (pot%type(i) == tersoff_type) THEN
                     npairs = iend-istart+1
                     tersoff => pot%set(i)%tersoff
                     ALLOCATE (sort_list(2, npairs), work_list(npairs))
                     sort_list = list(:, istart:iend)
                     ! Sort the list of neighbors, this increases the efficiency for single
                     ! potential contributions
                     CALL sort(sort_list(1, :), npairs, work_list)
                     DO ipair = 1, npairs
                        work_list(ipair) = sort_list(2, work_list(ipair))
                     END DO
                     sort_list(2, :) = work_list
                     ! find number of unique elements of array index 1
                     nunique = 1
                     DO ipair = 1, npairs-1
                        IF (sort_list(1, ipair+1) /= sort_list(1, ipair)) nunique = nunique+1
                     END DO
                     ipair = 1
                     junique = sort_list(1, ipair)
                     ifirst = 1
                     DO iunique = 1, nunique
                        atom_a = junique
                        IF (glob_loc_list_a(ifirst) > atom_a) CYCLE
                        DO mpair = ifirst, SIZE(glob_loc_list_a)
                           IF (glob_loc_list_a(mpair) == atom_a) EXIT
                        END DO
                        ifirst = mpair
                        DO mpair = ifirst, SIZE(glob_loc_list_a)
                           IF (glob_loc_list_a(mpair) /= atom_a) EXIT
                        END DO
                        ilast = mpair-1
                        nloc_size = 0
                        IF (ifirst /= 0) nloc_size = ilast-ifirst+1
                        DO WHILE (ipair <= npairs)
                           IF (sort_list(1, ipair) /= junique) EXIT
                           atom_b = sort_list(2, ipair)
                           ! Derivative terms
                           rtmp = r_last_update_pbc(atom_b)%r(:)-r_last_update_pbc(atom_a)%r(:)+cell_v
                           ipair = ipair+1
                           IF (DOT_PRODUCT(rtmp, rtmp) <= tersoff%rcutsq) THEN
                              CALL tersoff_forces(tersoff, r_last_update_pbc, cell_v, &
                                                  nloc_size, glob_loc_list(:, ifirst:ilast), glob_cell_v(:, ifirst:ilast), &
                                                  atom_a, atom_b, f_nonbond, pv_nonbond, use_virial, tersoff%rcutsq)
                           END IF
                        END DO
                        ifirst = ilast+1
                        IF (ipair <= npairs) junique = sort_list(1, ipair)
                     END DO
                     DEALLOCATE (sort_list, work_list)
                  END IF
               END DO
            END DO Kind_Group_Loop2
         END DO
         CALL destroy_tersoff_arrays(glob_loc_list, glob_cell_v, glob_loc_list_a)
      END IF
      ! Special way of handling the siepmann potential..
      IF (any_siepmann) THEN
         NULLIFY (glob_loc_list, glob_cell_v, glob_loc_list_a)
         CALL setup_siepmann_arrays(nonbonded, potparm, glob_loc_list, glob_cell_v, glob_loc_list_a, cell)
         DO ilist = 1, nonbonded%nlists
            neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
            npairs = neighbor_kind_pair%npairs
            IF (npairs == 0) CYCLE
            Kind_Group_Loop3: DO igrp = 1, neighbor_kind_pair%ngrp_kind
               istart = neighbor_kind_pair%grp_kind_start(igrp)
               iend = neighbor_kind_pair%grp_kind_end(igrp)
               ikind = neighbor_kind_pair%ij_kind(1, igrp)
               jkind = neighbor_kind_pair%ij_kind(2, igrp)
               list => neighbor_kind_pair%list
               cvi = neighbor_kind_pair%cell_vector
               pot => potparm%pot(ikind, jkind)%pot

               IF (pot%no_mb) CYCLE
               rab2_max = pot%rcutsq
               CALL matvec_3x3(cell_v, cell%hmat, cvi)
               DO i = 1, SIZE(pot%type)
                  ! SIEPMANN
                  IF (pot%type(i) == siepmann_type) THEN
                     npairs = iend-istart+1
                     siepmann => pot%set(i)%siepmann
                     ALLOCATE (sort_list(2, npairs), work_list(npairs))
                     sort_list = list(:, istart:iend)
                     ! Sort the list of neighbors, this increases the efficiency for single
                     ! potential contributions
                     CALL sort(sort_list(1, :), npairs, work_list)
                     DO ipair = 1, npairs
                        work_list(ipair) = sort_list(2, work_list(ipair))
                     END DO
                     sort_list(2, :) = work_list
                     ! find number of unique elements of array index 1
                     nunique = 1
                     DO ipair = 1, npairs-1
                        IF (sort_list(1, ipair+1) /= sort_list(1, ipair)) nunique = nunique+1
                     END DO
                     ipair = 1
                     junique = sort_list(1, ipair)
                     ifirst = 1
                     DO iunique = 1, nunique
                        atom_a = junique
                        IF (glob_loc_list_a(ifirst) > atom_a) CYCLE
                        DO mpair = ifirst, SIZE(glob_loc_list_a)
                           IF (glob_loc_list_a(mpair) == atom_a) EXIT
                        END DO
                        ifirst = mpair
                        DO mpair = ifirst, SIZE(glob_loc_list_a)
                           IF (glob_loc_list_a(mpair) /= atom_a) EXIT
                        END DO
                        ilast = mpair-1
                        nloc_size = 0
                        IF (ifirst /= 0) nloc_size = ilast-ifirst+1
                        DO WHILE (ipair <= npairs)
                           IF (sort_list(1, ipair) /= junique) EXIT
                           atom_b = sort_list(2, ipair)
                           ! Derivative terms
                           rtmp = r_last_update_pbc(atom_b)%r(:)-r_last_update_pbc(atom_a)%r(:)+cell_v
                           ipair = ipair+1
                           IF (DOT_PRODUCT(rtmp, rtmp) <= siepmann%rcutsq) THEN
                              CALL siepmann_forces_v2(siepmann, r_last_update_pbc, cell_v, cell, &
                                                      atom_a, atom_b, f_nonbond, use_virial, siepmann%rcutsq, &
                                                      particle_set)
                              CALL siepmann_forces_v3(siepmann, r_last_update_pbc, cell_v, &
                                                      nloc_size, glob_loc_list(:, ifirst:ilast), &
                                                      atom_a, atom_b, f_nonbond, use_virial, siepmann%rcutsq, &
                                                      cell, particle_set)
                           END IF
                        END DO
                        ifirst = ilast+1
                        IF (ipair <= npairs) junique = sort_list(1, ipair)
                     END DO
                     DEALLOCATE (sort_list, work_list)
                  END IF
               END DO
            END DO Kind_Group_Loop3
         END DO
         CALL destroy_siepmann_arrays(glob_loc_list, glob_cell_v, glob_loc_list_a)
      END IF
      IF (use_virial) THEN
         pv_nonbond(1, 1) = pv_nonbond(1, 1)+ptens11
         pv_nonbond(1, 2) = pv_nonbond(1, 2)+ptens12
         pv_nonbond(1, 3) = pv_nonbond(1, 3)+ptens13
         pv_nonbond(2, 1) = pv_nonbond(2, 1)+ptens21
         pv_nonbond(2, 2) = pv_nonbond(2, 2)+ptens22
         pv_nonbond(2, 3) = pv_nonbond(2, 3)+ptens23
         pv_nonbond(3, 1) = pv_nonbond(3, 1)+ptens31
         pv_nonbond(3, 2) = pv_nonbond(3, 2)+ptens32
         pv_nonbond(3, 3) = pv_nonbond(3, 3)+ptens33
      END IF
      CALL timestop(handle)
   END SUBROUTINE force_nonbond_manybody

END MODULE manybody_potential

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (11 May 2001) : cleaning up of support structures
!>      CJM & HAF (27 July 2001): fixed bug with handling of cutoff larger than
!>                                half the boxsize.
!>      07.02.2005: getting rid of scaled_to_real calls in force loop (MK)
!>      22.06.2013: OpenMP parallelisation of pair interaction loop (MK)
!> \author CJM
! **************************************************************************************************
MODULE fist_nonbond_force
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                              neighbor_kind_pairs_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_get,&
                                              fist_nonbond_env_type,&
                                              pos_type
   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size
   USE machine,                         ONLY: m_memory
   USE mathconstants,                   ONLY: oorootpi,&
                                              sqrthalf
   USE mathlib,                         ONLY: matvec_3x3
   USE message_passing,                 ONLY: mp_sum
   USE pair_potential_coulomb,          ONLY: potential_coulomb
   USE pair_potential_types,            ONLY: nosh_nosh,&
                                              nosh_sh,&
                                              pair_potential_pp_type,&
                                              pair_potential_single_type,&
                                              sh_sh,&
                                              siepmann_type,&
                                              tersoff_type
   USE particle_types,                  ONLY: particle_type
   USE shell_potential_types,           ONLY: get_shell,&
                                              shell_kind_type
   USE splines_methods,                 ONLY: potential_s
   USE splines_types,                   ONLY: spline_data_p_type,&
                                              spline_factor_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_nonbond_force'
   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .FALSE.

   PUBLIC :: force_nonbond, &
             bonded_correct_gaussian

CONTAINS

! **************************************************************************************************
!> \brief Calculates the force and the potential of the minimum image, and
!>      the pressure tensor
!> \param fist_nonbond_env ...
!> \param ewald_env ...
!> \param particle_set ...
!> \param cell ...
!> \param pot_nonbond ...
!> \param f_nonbond ...
!> \param pv_nonbond ...
!> \param fshell_nonbond ...
!> \param fcore_nonbond ...
!> \param atprop_env ...
!> \param atomic_kind_set ...
!> \param use_virial ...
! **************************************************************************************************
   SUBROUTINE force_nonbond(fist_nonbond_env, ewald_env, particle_set, cell, &
                            pot_nonbond, f_nonbond, pv_nonbond, fshell_nonbond, fcore_nonbond, &
                            atprop_env, atomic_kind_set, use_virial)

      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(OUT)                         :: pot_nonbond
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: f_nonbond, pv_nonbond
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT), &
         OPTIONAL                                        :: fshell_nonbond, fcore_nonbond
      TYPE(atprop_type), POINTER                         :: atprop_env
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      LOGICAL, INTENT(IN)                                :: use_virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'force_nonbond', routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, ewald_type, handle, i, iend, igrp, ikind, ilist, ipair, istart, &
         istat, j, kind_a, kind_b, nkind, npairs, shell_a, shell_b, shell_type
      INTEGER, DIMENSION(:, :), POINTER                  :: list
      LOGICAL                                            :: all_terms, do_multipoles, full_nl, &
                                                            shell_present
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: is_shell_kind
      REAL(KIND=dp) :: alpha, beta, beta_a, beta_b, energy, etot, fac_ei, fac_kind, fac_vdw, &
         fscalar, mm_radius_a, mm_radius_b, qcore_a, qcore_b, qeff_a, qeff_b, qshell_a, qshell_b, &
         rab2, rab2_com, rab2_max
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: mm_radius, qcore, qeff, qshell
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v, cvi, fatom_a, fatom_b, fcore_a, &
                                                            fcore_b, fshell_a, fshell_b, rab, &
                                                            rab_cc, rab_com, rab_cs, rab_sc, rab_ss
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv, pv_thread
      REAL(KIND=dp), DIMENSION(3, 4)                     :: rab_list
      REAL(KIND=dp), DIMENSION(4)                        :: rab2_list
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ij_kind_full_fac
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: ei_interaction_cutoffs
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      TYPE(pair_potential_pp_type), POINTER              :: potparm, potparm14
      TYPE(pair_potential_single_type), POINTER          :: pot
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update, r_last_update_pbc, &
                                                            rcore_last_update_pbc, &
                                                            rshell_last_update_pbc
      TYPE(shell_kind_type), POINTER                     :: shell_kind
      TYPE(spline_data_p_type), DIMENSION(:), POINTER    :: spline_data
      TYPE(spline_factor_type), POINTER                  :: spl_f

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      NULLIFY (pot, rshell_last_update_pbc, spl_f, ij_kind_full_fac)
      CALL fist_nonbond_env_get(fist_nonbond_env, nonbonded=nonbonded, &
                                potparm14=potparm14, potparm=potparm, r_last_update=r_last_update, &
                                r_last_update_pbc=r_last_update_pbc, natom_types=nkind, &
                                rshell_last_update_pbc=rshell_last_update_pbc, &
                                rcore_last_update_pbc=rcore_last_update_pbc, &
                                ij_kind_full_fac=ij_kind_full_fac)
      CALL ewald_env_get(ewald_env, alpha=alpha, ewald_type=ewald_type, &
                         do_multipoles=do_multipoles, &
                         interaction_cutoffs=ei_interaction_cutoffs)

      ! Initializing the potential energy, pressure tensor and force
      pot_nonbond = 0.0_dp
      f_nonbond(:, :) = 0.0_dp

      IF (use_virial) THEN
         pv_nonbond(:, :) = 0.0_dp
      END IF
      shell_present = .FALSE.
      IF (PRESENT(fshell_nonbond)) THEN
         IF(.NOT.(PRESENT(fcore_nonbond)))CALL cp__a("fist_nonbond_force.F",157)
         fshell_nonbond = 0.0_dp
         fcore_nonbond = 0.0_dp
         shell_present = .TRUE.
      END IF
      ! Load atomic kind information
      ALLOCATE (mm_radius(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 164, "mm_radius", dp_size*nkind)
      ALLOCATE (qeff(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 166, "qeff", dp_size*nkind)
      ALLOCATE (qcore(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 168, "qcore", dp_size*nkind)
      ALLOCATE (qshell(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 170, "qshell", dp_size*nkind)
      ALLOCATE (is_shell_kind(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 172, "is_shell_kind", int_size*nkind)
      IF(.NOT.(istat == 0))CALL cp__a("fist_nonbond_force.F",173)
      DO ikind = 1, nkind
         atomic_kind => atomic_kind_set(ikind)
         CALL get_atomic_kind(atomic_kind, &
                              qeff=qeff(ikind), &
                              mm_radius=mm_radius(ikind), &
                              shell=shell_kind)
         is_shell_kind(ikind) = ASSOCIATED(shell_kind)
         IF (ASSOCIATED(shell_kind)) THEN
            CALL get_shell(shell=shell_kind, &
                           charge_core=qcore(ikind), &
                           charge_shell=qshell(ikind))
         ELSE
            qcore(ikind) = 0.0_dp
            qshell(ikind) = 0.0_dp
         END IF
      END DO
      ! Starting the force loop
      Lists: DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         npairs = neighbor_kind_pair%npairs
         IF (npairs == 0) CYCLE
         list => neighbor_kind_pair%list
         cvi = neighbor_kind_pair%cell_vector
         CALL matvec_3x3(cell_v, cell%hmat, cvi)
         Kind_Group_Loop: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            iend = neighbor_kind_pair%grp_kind_end(igrp)
!$OMP           PARALLEL DEFAULT(NONE) &
!$OMP                    PRIVATE(ipair,atom_a,atom_b,kind_a,kind_b,fac_kind,pot) &
!$OMP                    PRIVATE(fac_ei,fac_vdw,atomic_kind,full_nl,qcore_a,qshell_a) &
!$OMP                    PRIVATE(qeff_a,qcore_b,qshell_b,qeff_b,mm_radius_a,mm_radius_b) &
!$OMP                    PRIVATE(shell_kind,beta,beta_a,beta_b,spl_f,spline_data) &
!$OMP                    PRIVATE(shell_type,all_terms,rab_cc,rab_cs,rab_sc,rab_ss) &
!$OMP                    PRIVATE(rab_list,rab2_list,rab_com,rab2_com,pv,pv_thread) &
!$OMP                    PRIVATE(rab,rab2,rab2_max,fscalar,energy) &
!$OMP                    PRIVATE(shell_a,shell_b,etot,fatom_a,fatom_b) &
!$OMP                    PRIVATE(fcore_a,fcore_b,fshell_a,fshell_b,i,j) &
!$OMP                    SHARED(shell_present) &
!$OMP                    SHARED(istart,iend,list,particle_set,ij_kind_full_fac) &
!$OMP                    SHARED(neighbor_kind_pair,atomic_kind_set,fist_nonbond_env) &
!$OMP                    SHARED(potparm,potparm14,do_multipoles,r_last_update_pbc) &
!$OMP                    SHARED(use_virial,ei_interaction_cutoffs,alpha,cell_v) &
!$OMP                    SHARED(rcore_last_update_pbc,rshell_last_update_pbc) &
!$OMP                    SHARED(f_nonbond,fcore_nonbond,fshell_nonbond,logger) &
!$OMP                    SHARED(ewald_type,pot_nonbond,pv_nonbond,atprop_env) &
!$OMP                    SHARED(is_shell_kind,mm_radius,qcore,qeff,qshell)
            IF (use_virial) pv_thread(:, :) = 0.0_dp
!$OMP           DO
            Pairs: DO ipair = istart, iend
               atom_a = list(1, ipair)
               atom_b = list(2, ipair)
               ! Get actual atomic kinds, since atom_a is not always of
               ! kind_a and atom_b of kind_b, ie. they might be swapped.
               kind_a = particle_set(atom_a)%atomic_kind%kind_number
               kind_b = particle_set(atom_b)%atomic_kind%kind_number

               fac_kind = ij_kind_full_fac(kind_a, kind_b)
               ! take the proper potential
               pot => potparm%pot(kind_a, kind_b)%pot
               IF (ipair <= neighbor_kind_pair%nscale) THEN
                  IF (neighbor_kind_pair%is_onfo(ipair)) THEN
                     pot => potparm14%pot(kind_a, kind_b)%pot
                  END IF
               END IF

               ! Determine the scaling factors
               fac_ei = fac_kind
               fac_vdw = fac_kind
               full_nl = ANY(pot%type == tersoff_type) .OR. ANY(pot%type == siepmann_type)
               IF ((.NOT. full_nl) .AND. (atom_a == atom_b)) THEN
                  fac_ei = 0.5_dp*fac_ei
                  fac_vdw = 0.5_dp*fac_vdw
               END IF
               ! decide which interactions to compute
               IF (do_multipoles) THEN
                  fac_ei = 0.0_dp
               END IF
               IF (ipair <= neighbor_kind_pair%nscale) THEN
                  fac_ei = fac_ei*neighbor_kind_pair%ei_scale(ipair)
                  fac_vdw = fac_vdw*neighbor_kind_pair%vdw_scale(ipair)
               END IF

               IF (fac_ei > 0.0_dp) THEN
                  ! Get the electrostatic parameters for the atoms a and b
                  mm_radius_a = mm_radius(kind_a)
                  mm_radius_b = mm_radius(kind_b)
                  IF (ASSOCIATED(fist_nonbond_env%charges)) THEN
                     qeff_a = fist_nonbond_env%charges(atom_a)
                     qeff_b = fist_nonbond_env%charges(atom_b)
                  ELSE
                     qeff_a = qeff(kind_a)
                     qeff_b = qeff(kind_b)
                  END IF
                  IF (is_shell_kind(kind_a)) THEN
                     qcore_a = qcore(kind_a)
                     qshell_a = qshell(kind_a)
                     IF ((qcore_a == 0.0_dp) .AND. (qshell_a == 0.0_dp)) fac_ei = 0.0_dp
                  ELSE
                     qcore_a = qeff_a
                     qshell_a = HUGE(0.0_dp)
                     IF (qeff_a == 0.0_dp) fac_ei = 0.0_dp
                  END IF
                  IF (is_shell_kind(kind_b)) THEN
                     qcore_b = qcore(kind_b)
                     qshell_b = qshell(kind_b)
                     IF ((qcore_b == 0.0_dp) .AND. (qshell_b == 0.0_dp)) fac_ei = 0.0_dp
                  ELSE
                     qcore_b = qeff_b
                     qshell_b = HUGE(0.0_dp)
                     IF (qeff_b == 0.0_dp) fac_ei = 0.0_dp
                  END IF
                  ! Derive beta parameters
                  beta = 0.0_dp
                  beta_a = 0.0_dp
                  beta_b = 0.0_dp
                  IF (mm_radius_a > 0) THEN
                     beta_a = sqrthalf/mm_radius_a
                  END IF
                  IF (mm_radius_b > 0) THEN
                     beta_b = sqrthalf/mm_radius_b
                  END IF
                  IF ((mm_radius_a > 0) .OR. (mm_radius_b > 0)) THEN
                     beta = sqrthalf/SQRT(mm_radius_a*mm_radius_a+mm_radius_b*mm_radius_b)
                  END IF
               END IF

               ! In case we have only manybody potentials and no charges, this
               ! pair of atom types can be ignored here.
               IF (pot%no_pp .AND. (fac_ei == 0.0)) CYCLE

               ! Setup spline_data set
               spl_f => pot%spl_f
               spline_data => pot%pair_spline_data
               shell_type = pot%shell_type
               IF (shell_type /= nosh_nosh) THEN
                  IF(.NOT.(.NOT. do_multipoles))CALL cp__a("fist_nonbond_force.F",309)
                  IF(.NOT.(shell_present))CALL cp__a("fist_nonbond_force.F",310)
               END IF
               rab2_max = pot%rcutsq

               ! compute the relative vector(s) for this pair
               IF (shell_type /= nosh_nosh) THEN
                  ! do shell
                  all_terms = .TRUE.
                  IF (shell_type == sh_sh) THEN
                     shell_a = particle_set(atom_a)%shell_index
                     shell_b = particle_set(atom_b)%shell_index
                     rab_cc = rcore_last_update_pbc(shell_b)%r-rcore_last_update_pbc(shell_a)%r
                     rab_cs = rshell_last_update_pbc(shell_b)%r-rcore_last_update_pbc(shell_a)%r
                     rab_sc = rcore_last_update_pbc(shell_b)%r-rshell_last_update_pbc(shell_a)%r
                     rab_ss = rshell_last_update_pbc(shell_b)%r-rshell_last_update_pbc(shell_a)%r
                     rab_list(1:3, 1) = rab_cc(1:3)+cell_v(1:3)
                     rab_list(1:3, 2) = rab_cs(1:3)+cell_v(1:3)
                     rab_list(1:3, 3) = rab_sc(1:3)+cell_v(1:3)
                     rab_list(1:3, 4) = rab_ss(1:3)+cell_v(1:3)
                  ELSE IF ((shell_type == nosh_sh) .AND. (particle_set(atom_a)%shell_index /= 0)) THEN
                     shell_a = particle_set(atom_a)%shell_index
                     shell_b = 0
                     rab_cc = r_last_update_pbc(atom_b)%r-rcore_last_update_pbc(shell_a)%r
                     rab_sc = 0.0_dp
                     rab_cs = 0.0_dp
                     rab_ss = r_last_update_pbc(atom_b)%r-rshell_last_update_pbc(shell_a)%r
                     rab_list(1:3, 1) = rab_cc(1:3)+cell_v(1:3)
                     rab_list(1:3, 2) = 0.0_dp
                     rab_list(1:3, 3) = 0.0_dp
                     rab_list(1:3, 4) = rab_ss(1:3)+cell_v(1:3)
                  ELSE IF ((shell_type == nosh_sh) .AND. (particle_set(atom_b)%shell_index /= 0)) THEN
                     shell_b = particle_set(atom_b)%shell_index
                     shell_a = 0
                     rab_cc = rcore_last_update_pbc(shell_b)%r-r_last_update_pbc(atom_a)%r
                     rab_sc = 0.0_dp
                     rab_cs = 0.0_dp
                     rab_ss = rshell_last_update_pbc(shell_b)%r-r_last_update_pbc(atom_a)%r
                     rab_list(1:3, 1) = rab_cc(1:3)+cell_v(1:3)
                     rab_list(1:3, 2) = 0.0_dp
                     rab_list(1:3, 3) = 0.0_dp
                     rab_list(1:3, 4) = rab_ss(1:3)+cell_v(1:3)
                  END IF
                  ! Compute the term only if all the pairs (cc,cs,sc,ss) are within the cut-off
                  Check_terms: DO i = 1, 4
                     rab2_list(i) = rab_list(1, i)**2+rab_list(2, i)**2+rab_list(3, i)**2
                     IF (rab2_list(i) >= rab2_max) THEN
                        all_terms = .FALSE.
                        EXIT Check_terms
                     END IF
                  END DO Check_terms
                  rab_com = r_last_update_pbc(atom_b)%r-r_last_update_pbc(atom_a)%r
               ELSE
                  ! not do shell
                  rab_cc = r_last_update_pbc(atom_b)%r-r_last_update_pbc(atom_a)%r
                  rab_com = rab_cc
                  shell_a = 0
                  shell_b = 0
               END IF
               rab_com = rab_com+cell_v
               rab2_com = rab_com(1)**2+rab_com(2)**2+rab_com(3)**2

               ! compute the interactions for the current pair
               etot = 0.0_dp
               fatom_a(:) = 0.0_dp
               fatom_b(:) = 0.0_dp
               fcore_a(:) = 0.0_dp
               fcore_b(:) = 0.0_dp
               fshell_a(:) = 0.0_dp
               fshell_b(:) = 0.0_dp
               IF (use_virial) pv(:, :) = 0.0_dp
               IF (shell_type /= nosh_nosh) THEN
                  ! do shell
                  IF ((rab2_com <= rab2_max) .AND. all_terms) THEN
                     IF (fac_ei > 0) THEN
                        ! core-core or core-ion/ion-core: Coulomb only
                        rab = rab_list(:, 1)
                        rab2 = rab2_list(1)
                        fscalar = 0.0_dp
                        IF (shell_a == 0) THEN
                           ! atom a is a plain ion and can have beta_a > 0
                           energy = potential_coulomb(rab2, fscalar, fac_ei*qeff_a*qcore_b, &
                                                      ewald_type, alpha, beta_a, &
                                                      ei_interaction_cutoffs(2, kind_a, kind_b))
                           CALL add_force_nonbond(fatom_a, fcore_b, pv, fscalar, rab, use_virial)
                        ELSE IF (shell_b == 0) THEN
                           ! atom b is a plain ion and can have beta_b > 0
                           energy = potential_coulomb(rab2, fscalar, fac_ei*qcore_a*qeff_b, &
                                                      ewald_type, alpha, beta_b, &
                                                      ei_interaction_cutoffs(2, kind_b, kind_a))
                           CALL add_force_nonbond(fcore_a, fatom_b, pv, fscalar, rab, use_virial)
                        ELSE
                           ! core-core interaction is always pure point charge
                           energy = potential_coulomb(rab2, fscalar, fac_ei*qcore_a*qcore_b, &
                                                      ewald_type, alpha, 0.0_dp, &
                                                      ei_interaction_cutoffs(1, kind_a, kind_b))
                           CALL add_force_nonbond(fcore_a, fcore_b, pv, fscalar, rab, use_virial)
                        END IF
                        etot = etot+energy
                     END IF

                     IF (shell_type == sh_sh) THEN
                        ! shell-shell: VDW + Coulomb
                        rab = rab_list(:, 4)
                        rab2 = rab2_list(4)
                        fscalar = 0.0_dp
                        IF (fac_vdw > 0) THEN
                           energy = potential_s(spline_data, rab2, fscalar, spl_f, logger)
                           etot = etot+energy*fac_vdw
                           fscalar = fscalar*fac_vdw
                        END IF
                        IF (fac_ei > 0) THEN
                           ! note that potential_coulomb increments fscalar
                           energy = potential_coulomb(rab2, fscalar, fac_ei*qshell_a*qshell_b, &
                                                      ewald_type, alpha, beta, &
                                                      ei_interaction_cutoffs(3, kind_a, kind_b))
                           etot = etot+energy
                        END IF
                        CALL add_force_nonbond(fshell_a, fshell_b, pv, fscalar, rab, use_virial)

                        IF (fac_ei > 0) THEN
                           ! core-shell: Coulomb only
                           rab = rab_list(:, 2)
                           rab2 = rab2_list(2)
                           fscalar = 0.0_dp
                           ! swap kind_a and kind_b to get the right cutoff
                           energy = potential_coulomb(rab2, fscalar, fac_ei*qcore_a*qshell_b, &
                                                      ewald_type, alpha, beta_b, &
                                                      ei_interaction_cutoffs(2, kind_b, kind_a))
                           etot = etot+energy
                           CALL add_force_nonbond(fcore_a, fshell_b, pv, fscalar, rab, use_virial)

                           ! shell-core: Coulomb only
                           rab = rab_list(:, 3)
                           rab2 = rab2_list(3)
                           fscalar = 0.0_dp
                           energy = potential_coulomb(rab2, fscalar, fac_ei*qshell_a*qcore_b, &
                                                      ewald_type, alpha, beta_a, &
                                                      ei_interaction_cutoffs(2, kind_a, kind_b))
                           etot = etot+energy
                           CALL add_force_nonbond(fshell_a, fcore_b, pv, fscalar, rab, use_virial)
                        END IF
                     ELSE IF ((shell_type == nosh_sh) .AND. (shell_a == 0)) THEN
                        ! ion-shell: VDW + Coulomb
                        rab = rab_list(:, 4)
                        rab2 = rab2_list(4)
                        fscalar = 0.0_dp
                        IF (fac_vdw > 0) THEN
                           energy = potential_s(spline_data, rab2, fscalar, spl_f, logger)
                           etot = etot+energy*fac_vdw
                           fscalar = fscalar*fac_vdw
                        END IF
                        IF (fac_ei > 0) THEN
                           ! note that potential_coulomb increments fscalar
                           energy = potential_coulomb(rab2, fscalar, fac_ei*qeff_a*qshell_b, &
                                                      ewald_type, alpha, beta, &
                                                      ei_interaction_cutoffs(3, kind_a, kind_b))
                           etot = etot+energy
                        END IF
                        CALL add_force_nonbond(fatom_a, fshell_b, pv, fscalar, rab, use_virial)
                     ELSE IF ((shell_type == nosh_sh) .AND. (shell_b == 0)) THEN
                        ! shell-ion : VDW + Coulomb
                        rab = rab_list(:, 4)
                        rab2 = rab2_list(4)
                        fscalar = 0.0_dp
                        IF (fac_vdw > 0) THEN
                           energy = potential_s(spline_data, rab2, fscalar, spl_f, logger)
                           etot = etot+energy*fac_vdw
                           fscalar = fscalar*fac_vdw
                        END IF
                        IF (fac_ei > 0) THEN
                           ! note that potential_coulomb increments fscalar
                           energy = potential_coulomb(rab2, fscalar, fac_ei*qshell_a*qeff_b, &
                                                      ewald_type, alpha, beta, &
                                                      ei_interaction_cutoffs(3, kind_a, kind_b))
                           etot = etot+energy
                        END IF
                        CALL add_force_nonbond(fshell_a, fatom_b, pv, fscalar, rab, use_virial)
                     END IF
                  END IF
               ELSE
                  IF (rab2_com <= rab2_max) THEN
                     ! NO SHELL MODEL...
                     ! Ion-Ion: no shell model, VDW + coulomb
                     rab = rab_com
                     rab2 = rab2_com
                     fscalar = 0.0_dp
                     IF (fac_vdw > 0) THEN
                        energy = potential_s(spline_data, rab2, fscalar, spl_f, logger)
                        etot = etot+energy*fac_vdw
                        fscalar = fscalar*fac_vdw
                     END IF
                     IF (fac_ei > 0) THEN
                        ! note that potential_coulomb increments fscalar
                        energy = potential_coulomb(rab2, fscalar, fac_ei*qeff_a*qeff_b, &
                                                   ewald_type, alpha, beta, &
                                                   ei_interaction_cutoffs(3, kind_a, kind_b))
                        etot = etot+energy
                     END IF
                     CALL add_force_nonbond(fatom_a, fatom_b, pv, fscalar, rab, use_virial)
                  END IF
               END IF
               ! Nonbonded energy
!$OMP              ATOMIC
               pot_nonbond = pot_nonbond+etot
               IF (atprop_env%energy) THEN
                  ! Update atomic energies
!$OMP                 ATOMIC
                  atprop_env%atener(atom_a) = atprop_env%atener(atom_a)+0.5_dp*etot
!$OMP                 ATOMIC
                  atprop_env%atener(atom_b) = atprop_env%atener(atom_b)+0.5_dp*etot
               END IF
               ! Nonbonded forces
               DO i = 1, 3
!$OMP                 ATOMIC
                  f_nonbond(i, atom_a) = f_nonbond(i, atom_a)+fatom_a(i)
!$OMP                 ATOMIC
                  f_nonbond(i, atom_b) = f_nonbond(i, atom_b)+fatom_b(i)
               END DO
               IF (shell_a > 0) THEN
                  DO i = 1, 3
!$OMP                    ATOMIC
                     fcore_nonbond(i, shell_a) = fcore_nonbond(i, shell_a)+fcore_a(i)
!$OMP                    ATOMIC
                     fshell_nonbond(i, shell_a) = fshell_nonbond(i, shell_a)+fshell_a(i)
                  END DO
               END IF
               IF (shell_b > 0) THEN
                  DO i = 1, 3
!$OMP                    ATOMIC
                     fcore_nonbond(i, shell_b) = fcore_nonbond(i, shell_b)+fcore_b(i)
!$OMP                    ATOMIC
                     fshell_nonbond(i, shell_b) = fshell_nonbond(i, shell_b)+fshell_b(i)
                  END DO
               END IF
               ! Add the contribution of the current pair to the total pressure tensor
               IF (use_virial) THEN
                  DO i = 1, 3
                     DO j = 1, 3
                        pv_thread(j, i) = pv_thread(j, i)+pv(j, i)
                        ! Update atomic stress tensors
                        IF (atprop_env%stress) THEN
!$OMP                          ATOMIC
                           atprop_env%atstress(j, i, atom_a) = atprop_env%atstress(j, i, atom_a)+ &
                                                               0.5_dp*pv(j, i)
!$OMP                          ATOMIC
                           atprop_env%atstress(j, i, atom_b) = atprop_env%atstress(j, i, atom_b)+ &
                                                               0.5_dp*pv(j, i)
                        END IF
                     END DO
                  END DO
               END IF
            END DO Pairs
!$OMP           END DO
            IF (use_virial) THEN
               DO i = 1, 3
                  DO j = 1, 3
!$OMP                    ATOMIC
                     pv_nonbond(j, i) = pv_nonbond(j, i)+pv_thread(j, i)
                  END DO
               END DO
            END IF
!$OMP           END PARALLEL
         END DO Kind_Group_Loop
      END DO Lists

      !sample peak memory
      CALL m_memory()

      DEALLOCATE (mm_radius)
      DEALLOCATE (qeff)
      DEALLOCATE (qcore)
      DEALLOCATE (qshell)
      DEALLOCATE (is_shell_kind)

      CALL timestop(handle)

   END SUBROUTINE force_nonbond

   ! **************************************************************************************************
   !> \brief Adds a non-bonding contribution to the total force and optionally to
   !>        the virial.
   ! **************************************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param f_nonbond_a ...
!> \param f_nonbond_b ...
!> \param pv ...
!> \param fscalar ...
!> \param rab ...
!> \param use_virial ...
! **************************************************************************************************
   SUBROUTINE add_force_nonbond(f_nonbond_a, f_nonbond_b, pv, fscalar, rab, use_virial)

      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: f_nonbond_a, f_nonbond_b
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(INOUT)      :: pv
      REAL(KIND=dp), INTENT(IN)                          :: fscalar
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      LOGICAL, INTENT(IN)                                :: use_virial

      REAL(KIND=dp), DIMENSION(3)                        :: fr

      fr(1) = fscalar*rab(1)
      fr(2) = fscalar*rab(2)
      fr(3) = fscalar*rab(3)
      f_nonbond_a(1) = f_nonbond_a(1)-fr(1)
      f_nonbond_a(2) = f_nonbond_a(2)-fr(2)
      f_nonbond_a(3) = f_nonbond_a(3)-fr(3)
      f_nonbond_b(1) = f_nonbond_b(1)+fr(1)
      f_nonbond_b(2) = f_nonbond_b(2)+fr(2)
      f_nonbond_b(3) = f_nonbond_b(3)+fr(3)
      IF (use_virial) THEN
         pv(1, 1) = pv(1, 1)+rab(1)*fr(1)
         pv(1, 2) = pv(1, 2)+rab(1)*fr(2)
         pv(1, 3) = pv(1, 3)+rab(1)*fr(3)
         pv(2, 1) = pv(2, 1)+rab(2)*fr(1)
         pv(2, 2) = pv(2, 2)+rab(2)*fr(2)
         pv(2, 3) = pv(2, 3)+rab(2)*fr(3)
         pv(3, 1) = pv(3, 1)+rab(3)*fr(1)
         pv(3, 2) = pv(3, 2)+rab(3)*fr(2)
         pv(3, 3) = pv(3, 3)+rab(3)*fr(3)
      END IF

   END SUBROUTINE

! **************************************************************************************************
!> \brief corrects electrostatics for bonded terms
!> \param fist_nonbond_env ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param particle_set ...
!> \param ewald_env ...
!> \param v_bonded_corr ...
!> \param pv_bc ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param atprop_env ...
!> \param cell ...
!> \param use_virial ...
!> \par History
!>      Splitted routines to clean and to fix a bug with the tensor whose
!>      original definition was not correct for PBC.. [Teodoro Laino -06/2007]
! **************************************************************************************************
   SUBROUTINE bonded_correct_gaussian(fist_nonbond_env, atomic_kind_set, &
                                      local_particles, particle_set, ewald_env, v_bonded_corr, pv_bc, &
                                      shell_particle_set, core_particle_set, atprop_env, cell, use_virial)

      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      REAL(KIND=dp), INTENT(OUT)                         :: v_bonded_corr
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: pv_bc
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      TYPE(atprop_type), POINTER                         :: atprop_env
      TYPE(cell_type), POINTER                           :: cell
      LOGICAL, INTENT(IN)                                :: use_virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'bonded_correct_gaussian', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, group, handle, iatom, iend, igrp, ilist, ipair, istart, kind_a, &
         kind_b, natoms_per_kind, nkind, npairs, shell_a, shell_b
      INTEGER, DIMENSION(:, :), POINTER                  :: list
      LOGICAL                                            :: a_is_shell, b_is_shell, do_multipoles, &
                                                            full_nl, shell_adiabatic
      REAL(KIND=dp)                                      :: alpha, const, fac_cor, fac_ei, qcore_a, &
                                                            qcore_b, qeff_a, qeff_b, qshell_a, &
                                                            qshell_b
      REAL(KIND=dp), DIMENSION(3)                        :: rca, rcb, rsa, rsb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ij_kind_full_fac
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      TYPE(pair_potential_pp_type), POINTER              :: potparm, potparm14
      TYPE(pair_potential_single_type), POINTER          :: pot
      TYPE(shell_kind_type), POINTER                     :: shell_kind

      CALL timeset(routineN, handle)

      ! Initializing values
      IF (use_virial) pv_bc = 0.0_dp
      v_bonded_corr = 0.0_dp

      CALL fist_nonbond_env_get(fist_nonbond_env, nonbonded=nonbonded, &
                                potparm14=potparm14, potparm=potparm, &
                                ij_kind_full_fac=ij_kind_full_fac)
      CALL ewald_env_get(ewald_env, alpha=alpha, do_multipoles=do_multipoles, &
                         group=group)
      ! Defining the constants
      const = 2.0_dp*alpha*oorootpi

      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               shell_adiabatic=shell_adiabatic)

      Lists: DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         npairs = neighbor_kind_pair%nscale
         IF (npairs == 0) CYCLE
         list => neighbor_kind_pair%list
         Kind_Group_Loop: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            IF (istart > npairs) THEN
               EXIT
            END IF
            iend = MIN(npairs, neighbor_kind_pair%grp_kind_end(igrp))

            Pairs: DO ipair = istart, iend
               atom_a = list(1, ipair)
               atom_b = list(2, ipair)
               ! Get actual atomic kinds, since atom_a is not always of
               ! kind_a and atom_b of kind_b, ie. they might be swapped.
               kind_a = particle_set(atom_a)%atomic_kind%kind_number
               kind_b = particle_set(atom_b)%atomic_kind%kind_number

               ! take the proper potential, only for full_nl test
               pot => potparm%pot(kind_a, kind_b)%pot
               IF (ipair <= neighbor_kind_pair%nscale) THEN
                  IF (neighbor_kind_pair%is_onfo(ipair)) THEN
                     pot => potparm14%pot(kind_a, kind_b)%pot
                  END IF
               END IF

               ! Determine the scaling factors
               fac_ei = ij_kind_full_fac(kind_a, kind_b)
               full_nl = ANY(pot%type == tersoff_type) .OR. ANY(pot%type == siepmann_type)
               IF ((.NOT. full_nl) .AND. (atom_a == atom_b)) THEN
                  fac_ei = fac_ei*0.5_dp
               END IF
               IF (ipair <= neighbor_kind_pair%nscale) THEN
                  fac_ei = fac_ei*neighbor_kind_pair%ei_scale(ipair)
               END IF
               ! The amount of correction is related to the
               ! amount of scaling as follows:
               fac_cor = 1.0_dp-fac_ei
               IF (fac_cor <= 0.0_dp) CYCLE

               ! Parameters for kind a
               atomic_kind => atomic_kind_set(kind_a)
               CALL get_atomic_kind(atomic_kind, qeff=qeff_a, shell=shell_kind)
               IF (ASSOCIATED(fist_nonbond_env%charges)) qeff_a = fist_nonbond_env%charges(atom_a)
               a_is_shell = ASSOCIATED(shell_kind)
               IF (a_is_shell) THEN
                  CALL get_shell(shell=shell_kind, charge_core=qcore_a, &
                                 charge_shell=qshell_a)
                  shell_a = particle_set(atom_a)%shell_index
                  rca = core_particle_set(shell_a)%r
                  rsa = shell_particle_set(shell_a)%r
               ELSE
                  qcore_a = qeff_a
                  qshell_a = HUGE(0.0_dp)
                  shell_a = 0
                  rca = particle_set(atom_a)%r
                  rsa = 0.0_dp
               END IF

               ! Parameters for kind b
               atomic_kind => atomic_kind_set(kind_b)
               CALL get_atomic_kind(atomic_kind, qeff=qeff_b, shell=shell_kind)
               IF (ASSOCIATED(fist_nonbond_env%charges)) qeff_b = fist_nonbond_env%charges(atom_b)
               b_is_shell = ASSOCIATED(shell_kind)
               IF (b_is_shell) THEN
                  CALL get_shell(shell=shell_kind, charge_core=qcore_b, &
                                 charge_shell=qshell_b)
                  shell_b = particle_set(atom_b)%shell_index
                  rcb = core_particle_set(shell_b)%r
                  rsb = shell_particle_set(shell_b)%r
               ELSE
                  qcore_b = qeff_b
                  qshell_b = HUGE(0.0_dp)
                  shell_b = 0
                  rcb = particle_set(atom_b)%r
                  rsb = 0.0_dp
               END IF

               ! First part: take care of core/ion-core/ion correction
               IF (a_is_shell .AND. b_is_shell) THEN
                  ! correct for core-core interaction
                  CALL bonded_correct_gaussian_low(rca, rcb, cell, &
                                                   v_bonded_corr, core_particle_set, core_particle_set, &
                                                   shell_a, shell_b, .TRUE., alpha, qcore_a, qcore_b, &
                                                   const, fac_cor, pv_bc, atprop_env, use_virial)
               ELSE IF (a_is_shell) THEN
                  ! correct for core-ion interaction
                  CALL bonded_correct_gaussian_low(rca, rcb, cell, &
                                                   v_bonded_corr, core_particle_set, particle_set, &
                                                   shell_a, atom_b, .TRUE., alpha, qcore_a, qcore_b, &
                                                   const, fac_cor, pv_bc, atprop_env, use_virial)
               ELSE IF (b_is_shell) THEN
                  ! correct for ion-core interaction
                  CALL bonded_correct_gaussian_low(rca, rcb, cell, &
                                                   v_bonded_corr, particle_set, core_particle_set, &
                                                   atom_a, shell_b, .TRUE., alpha, qcore_a, qcore_b, &
                                                   const, fac_cor, pv_bc, atprop_env, use_virial)
               ELSE
                  ! correct for ion-ion interaction
                  CALL bonded_correct_gaussian_low(rca, rcb, cell, &
                                                   v_bonded_corr, particle_set, particle_set, &
                                                   atom_a, atom_b, .TRUE., alpha, qcore_a, qcore_b, &
                                                   const, fac_cor, pv_bc, atprop_env, use_virial)
               END IF

               ! Second part: take care of shell-shell, shell-core/ion and
               ! core/ion-shell corrections
               IF (a_is_shell .AND. b_is_shell) THEN
                  ! correct for shell-shell interaction
                  CALL bonded_correct_gaussian_low(rsa, rsa, cell, &
                                                   v_bonded_corr, shell_particle_set, shell_particle_set, &
                                                   shell_a, shell_b, shell_adiabatic, alpha, qshell_a, &
                                                   qshell_b, const, fac_cor, pv_bc, atprop_env, use_virial)
               END IF
               IF (a_is_shell) THEN
                  IF (b_is_shell) THEN
                     ! correct for shell-core interaction
                     CALL bonded_correct_gaussian_low(rsa, rcb, cell, &
                                                      v_bonded_corr, shell_particle_set, core_particle_set, &
                                                      shell_a, shell_b, shell_adiabatic, alpha, qshell_a, qcore_b, &
                                                      const, fac_cor, pv_bc, atprop_env, use_virial)
                  ELSE
                     ! correct for shell-ion interaction
                     CALL bonded_correct_gaussian_low(rsa, rcb, cell, &
                                                      v_bonded_corr, shell_particle_set, particle_set, &
                                                      shell_a, atom_b, shell_adiabatic, alpha, qshell_a, qcore_b, &
                                                      const, fac_cor, pv_bc, atprop_env, use_virial)
                  END IF
               END IF
               IF (b_is_shell) THEN
                  IF (a_is_shell) THEN
                     ! correct for core-shell interaction
                     CALL bonded_correct_gaussian_low(rca, rsb, cell, &
                                                      v_bonded_corr, core_particle_set, shell_particle_set, &
                                                      shell_a, shell_b, shell_adiabatic, alpha, qcore_a, qshell_b, &
                                                      const, fac_cor, pv_bc, atprop_env, use_virial)
                  ELSE
                     ! correct for ion-shell interaction
                     CALL bonded_correct_gaussian_low(rca, rsb, cell, &
                                                      v_bonded_corr, particle_set, shell_particle_set, &
                                                      atom_a, shell_b, shell_adiabatic, alpha, qcore_a, qshell_b, &
                                                      const, fac_cor, pv_bc, atprop_env, use_virial)
                  END IF
               END IF
            END DO Pairs
         END DO Kind_Group_Loop
      END DO Lists

      ! Always correct core-shell interaction within one atom.
      nkind = SIZE(atomic_kind_set)
      DO kind_a = 1, nkind
         ! parameters for kind a
         atomic_kind => atomic_kind_set(kind_a)
         CALL get_atomic_kind(atomic_kind, shell=shell_kind)
         IF (ASSOCIATED(shell_kind)) THEN
            CALL get_shell(shell=shell_kind, charge_core=qcore_a, &
                           charge_shell=qshell_a)

            natoms_per_kind = local_particles%n_el(kind_a)
            DO iatom = 1, natoms_per_kind

               ! Data for atom a
               atom_a = local_particles%list(kind_a)%array(iatom)
               shell_a = particle_set(atom_a)%shell_index
               rca = core_particle_set(shell_a)%r
               rsa = shell_particle_set(shell_a)%r

               CALL bonded_correct_gaussian_low_sh(rca, rsa, cell, &
                                                   v_bonded_corr, core_particle_set, shell_particle_set, &
                                                   shell_a, shell_adiabatic, alpha, qcore_a, qshell_a, &
                                                   const, pv_bc, atprop_env, use_virial)

            END DO
         END IF
      END DO

      CALL mp_sum(v_bonded_corr, group)

      CALL timestop(handle)

   END SUBROUTINE bonded_correct_gaussian

! **************************************************************************************************
!> \brief ...
!> \param r1 ...
!> \param r2 ...
!> \param cell ...
!> \param v_bonded_corr ...
!> \param particle_set1 ...
!> \param particle_set2 ...
!> \param i ...
!> \param j ...
!> \param shell_adiabatic ...
!> \param alpha ...
!> \param q1 ...
!> \param q2 ...
!> \param const ...
!> \param fac_cor ...
!> \param pv_bc ...
!> \param atprop_env ...
!> \param use_virial ...
!> \par History
!>      Splitted routines to clean and to fix a bug with the tensor whose
!>      original definition was not correct for PBC..
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE bonded_correct_gaussian_low(r1, r2, cell, v_bonded_corr, &
                                          particle_set1, particle_set2, i, j, shell_adiabatic, alpha, q1, q2, &
                                          const, fac_cor, pv_bc, atprop_env, use_virial)
      REAL(KIND=dp), DIMENSION(3)                        :: r1, r2
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(INOUT)                       :: v_bonded_corr
      TYPE(particle_type), POINTER                       :: particle_set1(:), particle_set2(:)
      INTEGER, INTENT(IN)                                :: i, j
      LOGICAL, INTENT(IN)                                :: shell_adiabatic
      REAL(KIND=dp), INTENT(IN)                          :: alpha, q1, q2, const, fac_cor
      REAL(KIND=dp), INTENT(INOUT)                       :: pv_bc(3, 3)
      TYPE(atprop_type), POINTER                         :: atprop_env
      LOGICAL, INTENT(IN)                                :: use_virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'bonded_correct_gaussian_low', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: ac1 = 0.254829592_dp, ac2 = -0.284496736_dp, &
         ac3 = 1.421413741_dp, ac4 = -1.453152027_dp, ac5 = 1.061405429_dp, pc = 0.3275911_dp

      INTEGER                                            :: iatom, jatom
      REAL(KIND=dp)                                      :: arg, dij, e_arg_arg, errf, fscalar, &
                                                            idij, rijsq, tc, vbc
      REAL(KIND=dp), DIMENSION(3)                        :: fij_com, rij
      REAL(KIND=dp), DIMENSION(3, 3)                     :: fbc

      rij = r1-r2
      rij = pbc(rij, cell)
      rijsq = rij(1)*rij(1)+rij(2)*rij(2)+rij(3)*rij(3)
      idij = 1.0_dp/SQRT(rijsq)
      dij = rijsq*idij
      arg = alpha*dij
      e_arg_arg = EXP(-arg**2)
      tc = 1.0_dp/(1.0_dp+pc*arg)

      ! Defining errf=1-erfc
      errf = 1.0_dp-((((ac5*tc+ac4)*tc+ac3)*tc+ac2)*tc+ac1)*tc*e_arg_arg

      ! Getting the potential
      vbc = -q1*q2*idij*errf*fac_cor
      v_bonded_corr = v_bonded_corr+vbc
      IF (atprop_env%energy) THEN
         iatom = particle_set1(i)%atom_index
         atprop_env%atener(iatom) = atprop_env%atener(iatom)+0.5_dp*vbc
         jatom = particle_set2(j)%atom_index
         atprop_env%atener(jatom) = atprop_env%atener(jatom)+0.5_dp*vbc
      END IF

      ! Subtracting the force from the total force
      fscalar = q1*q2*idij**2*(idij*errf-const*e_arg_arg)*fac_cor

      particle_set1(i)%f(1) = particle_set1(i)%f(1)-fscalar*rij(1)
      particle_set1(i)%f(2) = particle_set1(i)%f(2)-fscalar*rij(2)
      particle_set1(i)%f(3) = particle_set1(i)%f(3)-fscalar*rij(3)

      particle_set2(j)%f(1) = particle_set2(j)%f(1)+fscalar*rij(1)
      particle_set2(j)%f(2) = particle_set2(j)%f(2)+fscalar*rij(2)
      particle_set2(j)%f(3) = particle_set2(j)%f(3)+fscalar*rij(3)

      IF (use_virial .AND. shell_adiabatic) THEN
         fij_com = fscalar*rij
         fbc(1, 1) = -fij_com(1)*rij(1)
         fbc(1, 2) = -fij_com(1)*rij(2)
         fbc(1, 3) = -fij_com(1)*rij(3)
         fbc(2, 1) = -fij_com(2)*rij(1)
         fbc(2, 2) = -fij_com(2)*rij(2)
         fbc(2, 3) = -fij_com(2)*rij(3)
         fbc(3, 1) = -fij_com(3)*rij(1)
         fbc(3, 2) = -fij_com(3)*rij(2)
         fbc(3, 3) = -fij_com(3)*rij(3)
         pv_bc(:, :) = pv_bc(:, :)+fbc(:, :)
         IF (atprop_env%stress) THEN
            ! Atomic stress tensors
            iatom = particle_set1(i)%atom_index
            atprop_env%atstress(:, :, iatom) = atprop_env%atstress(:, :, iatom)+0.5_dp*fbc(:, :)
            jatom = particle_set2(j)%atom_index
            atprop_env%atstress(:, :, jatom) = atprop_env%atstress(:, :, jatom)+0.5_dp*fbc(:, :)
         END IF
      END IF

   END SUBROUTINE bonded_correct_gaussian_low

! **************************************************************************************************
!> \brief specific for shell models cleans the interaction core-shell on the same
!>      atom
!> \param r1 ...
!> \param r2 ...
!> \param cell ...
!> \param v_bonded_corr ...
!> \param core_particle_set ...
!> \param shell_particle_set ...
!> \param i ...
!> \param shell_adiabatic ...
!> \param alpha ...
!> \param q1 ...
!> \param q2 ...
!> \param const ...
!> \param pv_bc ...
!> \param atprop_env ...
!> \param use_virial ...
!> \par History
!>      Splitted routines to clean and to fix a bug with the tensor whose
!>      original definition was not correct for PBC..
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE bonded_correct_gaussian_low_sh(r1, r2, cell, v_bonded_corr, &
                                             core_particle_set, shell_particle_set, i, shell_adiabatic, alpha, q1, q2, &
                                             const, pv_bc, atprop_env, use_virial)
      REAL(KIND=dp), DIMENSION(3)                        :: r1, r2
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(INOUT)                       :: v_bonded_corr
      TYPE(particle_type), POINTER                       :: core_particle_set(:), &
                                                            shell_particle_set(:)
      INTEGER, INTENT(IN)                                :: i
      LOGICAL, INTENT(IN)                                :: shell_adiabatic
      REAL(KIND=dp), INTENT(IN)                          :: alpha, q1, q2, const
      REAL(KIND=dp), INTENT(INOUT)                       :: pv_bc(3, 3)
      TYPE(atprop_type), POINTER                         :: atprop_env
      LOGICAL, INTENT(IN)                                :: use_virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'bonded_correct_gaussian_low_sh', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: ac1 = 0.254829592_dp, ac2 = -0.284496736_dp, &
         ac3 = 1.421413741_dp, ac4 = -1.453152027_dp, ac5 = 1.061405429_dp, pc = 0.3275911_dp

      INTEGER                                            :: iatom
      REAL(KIND=dp)                                      :: arg, dij, e_arg_arg, efac, errf, ffac, &
                                                            fscalar, idij, rijsq, tc, tc2, tc4, vbc
      REAL(KIND=dp), DIMENSION(3)                        :: fr, rij
      REAL(KIND=dp), DIMENSION(3, 3)                     :: fbc

      rij = r1-r2
      rij = pbc(rij, cell)
      rijsq = rij(1)*rij(1)+rij(2)*rij(2)+rij(3)*rij(3)
      dij = SQRT(rijsq)
      ! Two possible limiting cases according the value of dij
      arg = alpha*dij
      ! and this is a magic number.. it is related to the order expansion
      ! and to the value of the polynomial coefficients
      IF (arg > 0.355_dp) THEN
         idij = 1.0_dp/dij
         e_arg_arg = EXP(-arg*arg)
         tc = 1.0_dp/(1.0_dp+pc*arg)
         ! defining errf = 1 - erfc
         errf = 1.0_dp-((((ac5*tc+ac4)*tc+ac3)*tc+ac2)*tc+ac1)*tc*e_arg_arg
         efac = idij*errf
         ffac = idij**2*(efac-const*e_arg_arg)
      ELSE
         tc = arg*arg
         tc2 = tc*tc
         tc4 = tc2*tc2
         efac = const*(1.0_dp-tc/3.0_dp+tc2/10.0_dp-tc*tc2/42.0_dp+tc4/216.0_dp- &
                       tc*tc4/1320.0_dp+tc2*tc4/9360.0_dp)
         ffac = const*alpha**2*(2.0_dp/3.0_dp-2.0_dp*tc/5.0_dp+tc2/7.0_dp-tc*tc2/27.0_dp+ &
                                tc4/132.0_dp-tc*tc4/780.0_dp)
      END IF

      ! getting the potential
      vbc = -q1*q2*efac
      v_bonded_corr = v_bonded_corr+vbc
      IF (atprop_env%energy) THEN
         iatom = shell_particle_set(i)%atom_index
         atprop_env%atener(iatom) = atprop_env%atener(iatom)+vbc
      END IF

      ! subtracting the force from the total force
      fscalar = q1*q2*ffac
      fr(:) = fscalar*rij(:)

      core_particle_set(i)%f(1) = core_particle_set(i)%f(1)-fr(1)
      core_particle_set(i)%f(2) = core_particle_set(i)%f(2)-fr(2)
      core_particle_set(i)%f(3) = core_particle_set(i)%f(3)-fr(3)

      shell_particle_set(i)%f(1) = shell_particle_set(i)%f(1)+fr(1)
      shell_particle_set(i)%f(2) = shell_particle_set(i)%f(2)+fr(2)
      shell_particle_set(i)%f(3) = shell_particle_set(i)%f(3)+fr(3)

      IF (use_virial .AND. shell_adiabatic) THEN
         fbc(1, 1) = -fr(1)*rij(1)
         fbc(1, 2) = -fr(1)*rij(2)
         fbc(1, 3) = -fr(1)*rij(3)
         fbc(2, 1) = -fr(2)*rij(1)
         fbc(2, 2) = -fr(2)*rij(2)
         fbc(2, 3) = -fr(2)*rij(3)
         fbc(3, 1) = -fr(3)*rij(1)
         fbc(3, 2) = -fr(3)*rij(2)
         fbc(3, 3) = -fr(3)*rij(3)
         pv_bc(:, :) = pv_bc(:, :)+fbc(:, :)
         IF (atprop_env%stress) THEN
            ! Atomic stress tensors
            iatom = shell_particle_set(i)%atom_index
            atprop_env%atstress(:, :, iatom) = atprop_env%atstress(:, :, iatom)+fbc(:, :)
         END IF
      END IF

   END SUBROUTINE bonded_correct_gaussian_low_sh

END MODULE fist_nonbond_force
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      CJM APRIL-30-2009: only uses fist_env
!>      Teodoro Laino [tlaino] - 05.2009 : Generalization to different Ewald
!>                                         methods (initial framework)
!> \author CJM
! **************************************************************************************************

MODULE fist_pol_scf
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_type
   USE ewalds_multipole,                ONLY: ewald_multipole_evaluate
   USE fist_energy_types,               ONLY: fist_energy_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_type
   USE input_constants,                 ONLY: do_fist_pol_cg,&
                                              do_fist_pol_sc
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE multipole_types,                 ONLY: multipole_type
   USE particle_types,                  ONLY: particle_type
   USE pw_poisson_types,                ONLY: do_ewald_ewald,&
                                              do_ewald_pme,&
                                              do_ewald_spme
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   LOGICAL, PRIVATE                     :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_pol_scf'

   PUBLIC :: fist_pol_evaluate

CONTAINS

! **************************************************************************************************
!> \brief Main driver for evaluating energy/forces in a polarizable forcefield
!> \param atomic_kind_set ...
!> \param multipoles ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param fist_nonbond_env ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param thermo ...
!> \param vg_coulomb ...
!> \param pot_nonbond ...
!> \param f_nonbond ...
!> \param fg_coulomb ...
!> \param use_virial ...
!> \param pv_g ...
!> \param pv_nonbond ...
!> \param mm_section ...
!> \param do_ipol ...
!> \author Toon.Verstraelen@gmail.com (2010-03-01)
! **************************************************************************************************
   SUBROUTINE fist_pol_evaluate(atomic_kind_set, multipoles, ewald_env, &
                                ewald_pw, fist_nonbond_env, cell, particle_set, local_particles, &
                                thermo, vg_coulomb, pot_nonbond, f_nonbond, fg_coulomb, use_virial, &
                                pv_g, pv_nonbond, mm_section, do_ipol)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(multipole_type), POINTER                      :: multipoles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(fist_energy_type), POINTER                    :: thermo
      REAL(KIND=dp)                                      :: vg_coulomb, pot_nonbond
      REAL(KIND=dp), DIMENSION(:, :)                     :: f_nonbond, fg_coulomb
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_g, pv_nonbond
      TYPE(section_vals_type), POINTER                   :: mm_section
      INTEGER                                            :: do_ipol

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_pol_evaluate', &
         routineP = moduleN//':'//routineN

      SELECT CASE (do_ipol)
      CASE (do_fist_pol_sc)
         CALL fist_pol_evaluate_sc(atomic_kind_set, multipoles, ewald_env, &
                                   ewald_pw, fist_nonbond_env, cell, particle_set, local_particles, &
                                   thermo, vg_coulomb, pot_nonbond, f_nonbond, fg_coulomb, use_virial, &
                                   pv_g, pv_nonbond, mm_section)
      CASE (do_fist_pol_cg)
         CALL fist_pol_evaluate_cg(atomic_kind_set, multipoles, ewald_env, &
                                   ewald_pw, fist_nonbond_env, cell, particle_set, local_particles, &
                                   thermo, vg_coulomb, pot_nonbond, f_nonbond, fg_coulomb, use_virial, &
                                   pv_g, pv_nonbond, mm_section)
      END SELECT

   END SUBROUTINE fist_pol_evaluate

! **************************************************************************************************
!> \brief Self-consistent solver for a polarizable force-field
!> \param atomic_kind_set ...
!> \param multipoles ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param fist_nonbond_env ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param thermo ...
!> \param vg_coulomb ...
!> \param pot_nonbond ...
!> \param f_nonbond ...
!> \param fg_coulomb ...
!> \param use_virial ...
!> \param pv_g ...
!> \param pv_nonbond ...
!> \param mm_section ...
!> \author Toon.Verstraelen@gmail.com (2010-03-01)
!> \note
!>    Method: Given an initial guess of the induced dipoles, the electrostatic
!>    field is computed at each dipole. Then new induced dipoles are computed
!>    following p = alpha x E. This is repeated until a convergence criterion is
!>    met. The convergence is measured as the RSMD of the derivatives of the
!>    electrostatic energy (including dipole self-energy) towards the components
!>    of the dipoles.
! **************************************************************************************************
   SUBROUTINE fist_pol_evaluate_sc(atomic_kind_set, multipoles, ewald_env, ewald_pw, &
                                   fist_nonbond_env, cell, particle_set, local_particles, thermo, vg_coulomb, &
                                   pot_nonbond, f_nonbond, fg_coulomb, use_virial, pv_g, pv_nonbond, mm_section)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(multipole_type), POINTER                      :: multipoles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(fist_energy_type), POINTER                    :: thermo
      REAL(KIND=dp)                                      :: vg_coulomb, pot_nonbond
      REAL(KIND=dp), DIMENSION(:, :)                     :: f_nonbond, fg_coulomb
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_g, pv_nonbond
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_pol_evaluate_sc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ewald_type, handle, i, iatom, ii, ikind, &
                                                            iter, iw, iw2, j, max_ipol_iter, &
                                                            natom_of_kind, natoms, nkind, ntot
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: iwarn
      REAL(KIND=dp)                                      :: apol, cpol, eps_pol, pot_nonbond_local, &
                                                            rmsd, tmp_trace
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: efield1, efield2
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      NULLIFY (logger, atomic_kind)
      logger => cp_get_default_logger()

      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%ITER_INFO", &
                                extension=".mmLog")
      iw2 = cp_print_key_unit_nr(logger, mm_section, "PRINT%EWALD_INFO", &
                                 extension=".mmLog")

      CALL ewald_env_get(ewald_env, max_ipol_iter=max_ipol_iter, eps_pol=eps_pol, &
                         ewald_type=ewald_type)

      natoms = SIZE(particle_set)
      ALLOCATE (efield1(3, natoms))
      ALLOCATE (efield2(9, natoms))

      nkind = SIZE(atomic_kind_set)
      IF (iw > 0) WRITE (iw, FMT='(/,T5,"POL_SCF|","Method: self-consistent")')
      IF (iw > 0) WRITE (iw, FMT='(T5,"POL_SCF|","  Iteration",7X,"Conv.",T49,"Electrostatic & Induction Energy")')
      pol_scf: DO iter = 1, max_ipol_iter
         ! Evaluate the electrostatic with Ewald schemes
         CALL eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, cell, &
                             particle_set, local_particles, vg_coulomb, pot_nonbond_local, thermo, &
                             multipoles, do_correction_bonded=.TRUE., do_forces=.FALSE., &
                             do_stress=.FALSE., do_efield=.TRUE., iw2=iw2, do_debug=.FALSE., &
                             atomic_kind_set=atomic_kind_set, mm_section=mm_section, &
                             efield1=efield1, efield2=efield2)
         CALL mp_sum(pot_nonbond_local, logger%para_env%group)

         ! compute the new dipoles, qudrupoles, and check for convergence
         ntot = 0
         rmsd = 0.0_dp
         thermo%e_induction = 0.0_dp
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, apol=apol, cpol=cpol, natom=natom_of_kind, atom_list=atom_list)
            ! ignore atoms with dipole and quadrupole polarizability zero
            IF (apol == 0 .AND. cpol == 0) CYCLE
            ! increment counter correctly
            IF (apol /= 0) ntot = ntot+natom_of_kind
            IF (cpol /= 0) ntot = ntot+natom_of_kind

            DO iatom = 1, natom_of_kind
               ii = atom_list(iatom)
               IF (apol /= 0) THEN
                  DO i = 1, 3
                     ! the rmsd of the derivatives of the energy towards the
                     ! components of the atomic dipole moments
                     rmsd = rmsd+(multipoles%dipoles(i, ii)/apol-efield1(i, ii))**2
                  END DO
               END IF
               IF (cpol /= 0) THEN
                  rmsd = rmsd+(multipoles%quadrupoles(1, 1, ii)/cpol-efield2(1, ii))**2
                  rmsd = rmsd+(multipoles%quadrupoles(2, 1, ii)/cpol-efield2(2, ii))**2
                  rmsd = rmsd+(multipoles%quadrupoles(3, 1, ii)/cpol-efield2(3, ii))**2
                  rmsd = rmsd+(multipoles%quadrupoles(1, 2, ii)/cpol-efield2(4, ii))**2
                  rmsd = rmsd+(multipoles%quadrupoles(2, 2, ii)/cpol-efield2(5, ii))**2
                  rmsd = rmsd+(multipoles%quadrupoles(3, 2, ii)/cpol-efield2(6, ii))**2
                  rmsd = rmsd+(multipoles%quadrupoles(1, 3, ii)/cpol-efield2(7, ii))**2
                  rmsd = rmsd+(multipoles%quadrupoles(2, 3, ii)/cpol-efield2(8, ii))**2
                  rmsd = rmsd+(multipoles%quadrupoles(3, 3, ii)/cpol-efield2(9, ii))**2
               END IF
! compute dipole
               multipoles%dipoles(:, ii) = apol*efield1(:, ii)
! compute quadrupole
               IF (cpol /= 0) THEN
                  multipoles%quadrupoles(1, 1, ii) = cpol*efield2(1, ii)
                  multipoles%quadrupoles(2, 1, ii) = cpol*efield2(2, ii)
                  multipoles%quadrupoles(3, 1, ii) = cpol*efield2(3, ii)
                  multipoles%quadrupoles(1, 2, ii) = cpol*efield2(4, ii)
                  multipoles%quadrupoles(2, 2, ii) = cpol*efield2(5, ii)
                  multipoles%quadrupoles(3, 2, ii) = cpol*efield2(6, ii)
                  multipoles%quadrupoles(1, 3, ii) = cpol*efield2(7, ii)
                  multipoles%quadrupoles(2, 3, ii) = cpol*efield2(8, ii)
                  multipoles%quadrupoles(3, 3, ii) = cpol*efield2(9, ii)
               END IF
               ! Compute the new induction term while we are here
               IF (apol /= 0) THEN
                  thermo%e_induction = thermo%e_induction+ &
                                       DOT_PRODUCT(multipoles%dipoles(:, ii), &
                                                   multipoles%dipoles(:, ii))/apol/2.0_dp
               END IF
               IF (cpol /= 0) THEN
                  tmp_trace = 0._dp
                  DO i = 1, 3
                     DO j = 1, 3
                        tmp_trace = tmp_trace+ &
                                    multipoles%quadrupoles(i, j, ii)*multipoles%quadrupoles(i, j, ii)
                     END DO
                  END DO
                  thermo%e_induction = thermo%e_induction+tmp_trace/cpol/6.0_dp
               END IF
            END DO
         END DO
         rmsd = SQRT(rmsd/REAL(ntot, KIND=dp))
         IF (iw > 0) THEN
            ! print the energy that is minimized (this is electrostatic + induction)
            WRITE (iw, FMT='(T5,"POL_SCF|",5X,I5,5X,E12.6,T61,F20.10)') iter, &
               rmsd, vg_coulomb+pot_nonbond_local+thermo%e_induction
         END IF
         IF (rmsd <= eps_pol) THEN
            IF (iw > 0) WRITE (iw, FMT='(T5,"POL_SCF|",1X,"Self-consistent Polarization achieved.")')
            EXIT pol_scf
         END IF

         iwarn = ((rmsd > eps_pol) .AND. (iter == max_ipol_iter))
         IF (iwarn .AND. iw > 0) WRITE (iw, FMT='(T5,"POL_SCF|",1X,"Self-consistent Polarization not converged!")')
         IF (iwarn) &
            CALL cp__w("fist_pol_scf.F",280,"Self-consistent Polarization not converged! ")
      END DO pol_scf

      ! Now evaluate after convergence to obtain forces and converged energies
      CALL eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, cell, &
                          particle_set, local_particles, vg_coulomb, pot_nonbond_local, thermo, &
                          multipoles, do_correction_bonded=.TRUE., do_forces=.TRUE., &
                          do_stress=use_virial, do_efield=.FALSE., iw2=iw2, do_debug=.FALSE., &
                          atomic_kind_set=atomic_kind_set, mm_section=mm_section, &
                          forces_local=fg_coulomb, forces_glob=f_nonbond, &
                          pv_local=pv_g, pv_glob=pv_nonbond)
      pot_nonbond = pot_nonbond+pot_nonbond_local
      CALL mp_sum(pot_nonbond_local, logger%para_env%group)

      IF (iw > 0) THEN
         ! print the energy that is minimized (this is electrostatic + induction)
         WRITE (iw, FMT='(T5,"POL_SCF|",5X,"Final",T61,F20.10,/)') &
            vg_coulomb+pot_nonbond_local+thermo%e_induction
      END IF

      ! Deallocate working arrays
      DEALLOCATE (efield1)
      DEALLOCATE (efield2)
      CALL cp_print_key_finished_output(iw2, logger, mm_section, &
                                        "PRINT%EWALD_INFO")
      CALL cp_print_key_finished_output(iw, logger, mm_section, &
                                        "PRINT%ITER_INFO")

      CALL timestop(handle)
   END SUBROUTINE fist_pol_evaluate_sc

! **************************************************************************************************
!> \brief Conjugate-gradient solver for a polarizable force-field
!> \param atomic_kind_set ...
!> \param multipoles ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param fist_nonbond_env ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param thermo ...
!> \param vg_coulomb ...
!> \param pot_nonbond ...
!> \param f_nonbond ...
!> \param fg_coulomb ...
!> \param use_virial ...
!> \param pv_g ...
!> \param pv_nonbond ...
!> \param mm_section ...
!> \author Toon.Verstraelen@gmail.com (2010-03-01)
!> \note
!>     Method: The dipoles are found by minimizing the sum of the electrostatic
!>     and the induction energy directly using a conjugate gradient method. This
!>     routine assumes that the energy is a quadratic function of the dipoles.
!>     Finding the minimum is then done by solving a linear system. This will
!>     not work for polarizable force fields that include hyperpolarizability.
!>
!>     The implementation of the conjugate gradient solver for linear systems
!>     is described in chapter 2.7 Sparse Linear Systems, under the section
!>     "Conjugate Gradient Method for a Sparse System". Although the inducible
!>     dipoles are the solution of a dense linear system, the same algorithm is
!>     still recommended for this situation. One does not have access to the
!>     entire hardness kernel to compute the solution with conventional linear
!>     algebra routines, but one only has a function that computes the dot
!>     product of the hardness kernel and a vector. (This is the routine that
!>     computes the electrostatic field at the atoms for a given vector of
!>     inducible dipoles.) Given such function, the conjugate gradient method
!>     is an efficient way to compute the solution of a linear system, and it
!>     scales well towards many degrees of freedom in terms of efficiency and
!>     memory usage.
! **************************************************************************************************
   SUBROUTINE fist_pol_evaluate_cg(atomic_kind_set, multipoles, ewald_env, ewald_pw, &
                                   fist_nonbond_env, cell, particle_set, local_particles, thermo, vg_coulomb, &
                                   pot_nonbond, f_nonbond, fg_coulomb, use_virial, pv_g, pv_nonbond, mm_section)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(multipole_type), POINTER                      :: multipoles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(fist_energy_type), POINTER                    :: thermo
      REAL(KIND=dp)                                      :: vg_coulomb, pot_nonbond
      REAL(KIND=dp), DIMENSION(:, :)                     :: f_nonbond, fg_coulomb
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_g, pv_nonbond
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_pol_evaluate_cg', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ewald_type, handle, i, iatom, ii, ikind, &
                                                            iter, iw, iw2, max_ipol_iter, &
                                                            natom_of_kind, natoms, nkind, ntot
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: iwarn
      REAL(KIND=dp)                                      :: alpha, apol, beta, denom, eps_pol, &
                                                            pot_nonbond_local, rmsd
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: conjugate, conjugate_applied, efield1, &
                                                            efield1_ext, residual, tmp_dipoles
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      NULLIFY (logger, atomic_kind)
      logger => cp_get_default_logger()

      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%ITER_INFO", &
                                extension=".mmLog")
      iw2 = cp_print_key_unit_nr(logger, mm_section, "PRINT%EWALD_INFO", &
                                 extension=".mmLog")

      CALL ewald_env_get(ewald_env, max_ipol_iter=max_ipol_iter, eps_pol=eps_pol, &
                         ewald_type=ewald_type)

      ! allocate work arrays
      natoms = SIZE(particle_set)
      ALLOCATE (efield1(3, natoms))
      ALLOCATE (tmp_dipoles(3, natoms))
      ALLOCATE (residual(3, natoms))
      ALLOCATE (conjugate(3, natoms))
      ALLOCATE (conjugate_applied(3, natoms))
      ALLOCATE (efield1_ext(3, natoms))

      ! Compute the 'external' electrostatic field (all inducible dipoles
      ! equal to zero). this is required for the conjugate gradient solver.
      ! We assume that all dipoles are inducible dipoles.
      tmp_dipoles(:, :) = multipoles%dipoles ! backup of dipoles
      multipoles%dipoles = 0.0_dp
      CALL eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, cell, &
                          particle_set, local_particles, vg_coulomb, pot_nonbond_local, thermo, &
                          multipoles, do_correction_bonded=.TRUE., do_forces=.FALSE., &
                          do_stress=.FALSE., do_efield=.TRUE., iw2=iw2, do_debug=.FALSE., &
                          atomic_kind_set=atomic_kind_set, mm_section=mm_section, &
                          efield1=efield1_ext)
      multipoles%dipoles = tmp_dipoles ! restore backup

      ! Compute the electric field with the initial guess of the dipoles.
      CALL eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, cell, &
                          particle_set, local_particles, vg_coulomb, pot_nonbond_local, thermo, &
                          multipoles, do_correction_bonded=.TRUE., do_forces=.FALSE., &
                          do_stress=.FALSE., do_efield=.TRUE., iw2=iw2, do_debug=.FALSE., &
                          atomic_kind_set=atomic_kind_set, mm_section=mm_section, &
                          efield1=efield1)

      ! Compute the first residual explicitly.
      nkind = SIZE(atomic_kind_set)
      ntot = 0
      residual = 0.0_dp
      DO ikind = 1, nkind
         atomic_kind => atomic_kind_set(ikind)
         CALL get_atomic_kind(atomic_kind, apol=apol, natom=natom_of_kind, atom_list=atom_list)
         ! ignore atoms with polarizability zero
         IF (apol == 0) CYCLE
         ntot = ntot+natom_of_kind
         DO iatom = 1, natom_of_kind
            ii = atom_list(iatom)
            DO i = 1, 3
               ! residual = b - A x
               residual(i, ii) = efield1(i, ii)-multipoles%dipoles(i, ii)/apol
            END DO
         END DO
      END DO
      ! The first conjugate residual is equal to the residual.
      conjugate(:, :) = residual

      IF (iw > 0) WRITE (iw, FMT='(/,T5,"POL_SCF|","Method: conjugate-gradient")')
      IF (iw > 0) WRITE (iw, FMT='(T5,"POL_SCF|","  Iteration",7X,"Conv.",T49,"Electrostatic & Induction Energy")')
      pol_scf: DO iter = 1, max_ipol_iter
         IF (debug_this_module) THEN
            ! In principle the residual must not be computed explicitly any more. It
            ! is obtained in an indirect way below. When the DEBUG flag is set, the
            ! explicit residual is computed and compared with the implicitly derived
            ! residual as a double check.
            CALL eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, cell, &
                                particle_set, local_particles, vg_coulomb, pot_nonbond_local, thermo, &
                                multipoles, do_correction_bonded=.TRUE., do_forces=.FALSE., &
                                do_stress=.FALSE., do_efield=.TRUE., iw2=iw2, do_debug=.FALSE., &
                                atomic_kind_set=atomic_kind_set, mm_section=mm_section, &
                                efield1=efield1)
            ! inapropriate use of denom to check the error on the residual
            denom = 0.0_dp
         END IF
         rmsd = 0.0_dp
         ! Compute the rmsd of the residual.
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, apol=apol, natom=natom_of_kind, atom_list=atom_list)
            ! ignore atoms with polarizability zero
            IF (apol == 0) CYCLE
            DO iatom = 1, natom_of_kind
               ii = atom_list(iatom)
               DO i = 1, 3
                  ! residual = b - A x
                  rmsd = rmsd+residual(i, ii)**2
                  IF (debug_this_module) THEN
                     denom = denom+(residual(i, ii)-(efield1(i, ii)- &
                                                     multipoles%dipoles(i, ii)/apol))**2
                  END IF
               END DO
            END DO
         END DO
         rmsd = SQRT(rmsd/ntot)
         IF (iw > 0) THEN
            WRITE (iw, FMT='(T5,"POL_SCF|",5X,I5,5X,E12.6,T67,"(not computed)")') iter, rmsd
            IF (debug_this_module) THEN
               denom = SQRT(denom/ntot)
               WRITE (iw, FMT='(T5,"POL_SCF|",5X,"Error on implicit residual:",5X,E12.6)') denom
            END IF
         END IF

         ! Apply the hardness kernel to the conjugate residual.
         ! We assume that all non-zero dipoles are inducible dipoles.
         tmp_dipoles(:, :) = multipoles%dipoles ! backup of dipoles
         multipoles%dipoles = conjugate
         CALL eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, cell, &
                             particle_set, local_particles, vg_coulomb, pot_nonbond_local, thermo, &
                             multipoles, do_correction_bonded=.TRUE., do_forces=.FALSE., &
                             do_stress=.FALSE., do_efield=.TRUE., iw2=iw2, do_debug=.FALSE., &
                             atomic_kind_set=atomic_kind_set, mm_section=mm_section, &
                             efield1=conjugate_applied)
         multipoles%dipoles = tmp_dipoles ! restore backup
         conjugate_applied(:, :) = efield1_ext-conjugate_applied

         ! Finish conjugate_applied and compute alpha from the conjugate gradient algorithm.
         alpha = 0.0_dp
         denom = 0.0_dp
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, apol=apol, natom=natom_of_kind, atom_list=atom_list)
            ! ignore atoms with polarizability zero
            IF (apol == 0) CYCLE
            DO iatom = 1, natom_of_kind
               ii = atom_list(iatom)
               DO i = 1, 3
                  conjugate_applied(i, ii) = conjugate_applied(i, ii)+conjugate(i, ii)/apol
               END DO
               alpha = alpha+DOT_PRODUCT(residual(:, ii), residual(:, ii))
               denom = denom+DOT_PRODUCT(conjugate(:, ii), conjugate_applied(:, ii))
            END DO
         END DO
         alpha = alpha/denom

         ! Compute the new residual and beta from the conjugate gradient method.
         beta = 0.0_dp
         denom = 0.0_dp
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, apol=apol, natom=natom_of_kind, atom_list=atom_list)
            IF (apol == 0) CYCLE
            DO iatom = 1, natom_of_kind
               ii = atom_list(iatom)
               denom = denom+DOT_PRODUCT(residual(:, ii), residual(:, ii))
               DO i = 1, 3
                  residual(i, ii) = residual(i, ii)-alpha*conjugate_applied(i, ii)
               END DO
               beta = beta+DOT_PRODUCT(residual(:, ii), residual(:, ii))
            END DO
         END DO
         beta = beta/denom

         ! Compute the new dipoles, the new conjugate residual, and the induction
         ! energy.
         thermo%e_induction = 0.0_dp
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, apol=apol, natom=natom_of_kind, atom_list=atom_list)
            ! ignore atoms with polarizability zero
            IF (apol == 0) CYCLE
            DO iatom = 1, natom_of_kind
               ii = atom_list(iatom)
               DO i = 1, 3
                  multipoles%dipoles(i, ii) = multipoles%dipoles(i, ii)+alpha*conjugate(i, ii)
                  conjugate(i, ii) = residual(i, ii)+beta*conjugate(i, ii)
                  thermo%e_induction = thermo%e_induction+multipoles%dipoles(i, ii)**2/apol/2.0_dp
               END DO
            END DO
         END DO

         ! Quit if rmsd is low enough.
         IF (rmsd <= eps_pol) THEN
            IF (iw > 0) WRITE (iw, FMT='(T5,"POL_SCF|",1X,"Self-consistent Polarization achieved.")')
            EXIT pol_scf
         END IF

         ! Print warning when not converged.
         iwarn = ((rmsd > eps_pol) .AND. (iter == max_ipol_iter))
         IF (iwarn .AND. iw > 0) WRITE (iw, FMT='(T5,"POL_SCF|",1X,"Self-consistent Polarization not converged!")')
         IF (iwarn) &
            CALL cp__w("fist_pol_scf.F",572,"Self-consistent Polarization not converged! ")
      END DO pol_scf

      IF (debug_this_module) THEN
         ! Now evaluate after convergence to obtain forces and converged energies
         CALL eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, cell, &
                             particle_set, local_particles, vg_coulomb, pot_nonbond_local, thermo, &
                             multipoles, do_correction_bonded=.TRUE., do_forces=.TRUE., &
                             do_stress=use_virial, do_efield=.TRUE., iw2=iw2, do_debug=.FALSE., &
                             atomic_kind_set=atomic_kind_set, mm_section=mm_section, &
                             forces_local=fg_coulomb, forces_glob=f_nonbond, &
                             pv_local=pv_g, pv_glob=pv_nonbond, efield1=efield1)

         ! Do a final check on the convergence: compute the residual explicitely
         rmsd = 0.0_dp
         DO ikind = 1, nkind
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind, apol=apol, natom=natom_of_kind, atom_list=atom_list)
            ! ignore atoms with polarizability zero
            IF (apol == 0) CYCLE
            DO iatom = 1, natom_of_kind
               ii = atom_list(iatom)
               DO i = 1, 3
                  ! residual = b - A x
                  rmsd = rmsd+(efield1(i, ii)-multipoles%dipoles(i, ii)/apol)**2
               END DO
            END DO
         END DO
         rmsd = SQRT(rmsd/ntot)
         IF (iw > 0) WRITE (iw, FMT='(T5,"POL_SCF|",1X,"Final RMSD of residual:",5X,E12.6)') rmsd
         ! Stop program when congergence is not reached after all
         IF (rmsd > eps_pol) THEN
            CALL cp__w("fist_pol_scf.F",604,"Error in the conjugate gradient method for self-consistent polarization! ")
         END IF
      ELSE
         ! Now evaluate after convergence to obtain forces and converged energies
         CALL eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, cell, &
                             particle_set, local_particles, vg_coulomb, pot_nonbond_local, thermo, &
                             multipoles, do_correction_bonded=.TRUE., do_forces=.TRUE., &
                             do_stress=use_virial, do_efield=.FALSE., iw2=iw2, do_debug=.FALSE., &
                             atomic_kind_set=atomic_kind_set, mm_section=mm_section, &
                             forces_local=fg_coulomb, forces_glob=f_nonbond, &
                             pv_local=pv_g, pv_glob=pv_nonbond)
      ENDIF
      pot_nonbond = pot_nonbond+pot_nonbond_local
      CALL mp_sum(pot_nonbond_local, logger%para_env%group)

      IF (iw > 0) THEN
         WRITE (iw, FMT='(T5,"POL_SCF|",5X,"Final",T61,F20.10,/)') &
            vg_coulomb+pot_nonbond_local+thermo%e_induction
      END IF

      ! Deallocate working arrays
      DEALLOCATE (efield1)
      DEALLOCATE (tmp_dipoles)
      DEALLOCATE (residual)
      DEALLOCATE (conjugate)
      DEALLOCATE (conjugate_applied)
      DEALLOCATE (efield1_ext)
      CALL cp_print_key_finished_output(iw2, logger, mm_section, &
                                        "PRINT%EWALD_INFO")
      CALL cp_print_key_finished_output(iw, logger, mm_section, &
                                        "PRINT%ITER_INFO")

      CALL timestop(handle)
   END SUBROUTINE fist_pol_evaluate_cg

! **************************************************************************************************
!> \brief Main driver for evaluating electrostatic in polarible forcefields
!>        All the dependence on the Ewald method should go here!
!> \param ewald_type ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param fist_nonbond_env ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param vg_coulomb ...
!> \param pot_nonbond ...
!> \param thermo ...
!> \param multipoles ...
!> \param do_correction_bonded ...
!> \param do_forces ...
!> \param do_stress ...
!> \param do_efield ...
!> \param iw2 ...
!> \param do_debug ...
!> \param atomic_kind_set ...
!> \param mm_section ...
!> \param efield0 ...
!> \param efield1 ...
!> \param efield2 ...
!> \param forces_local ...
!> \param forces_glob ...
!> \param pv_local ...
!> \param pv_glob ...
!> \author Teodoro Laino [tlaino] 05.2009
! **************************************************************************************************
   SUBROUTINE eval_pol_ewald(ewald_type, ewald_env, ewald_pw, fist_nonbond_env, &
                             cell, particle_set, local_particles, vg_coulomb, pot_nonbond, thermo, &
                             multipoles, do_correction_bonded, do_forces, do_stress, do_efield, iw2, &
                             do_debug, atomic_kind_set, mm_section, efield0, efield1, efield2, forces_local, &
                             forces_glob, pv_local, pv_glob)

      INTEGER, INTENT(IN)                                :: ewald_type
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), INTENT(OUT)                         :: vg_coulomb, pot_nonbond
      TYPE(fist_energy_type), POINTER                    :: thermo
      TYPE(multipole_type), POINTER                      :: multipoles
      LOGICAL, INTENT(IN)                                :: do_correction_bonded, do_forces, &
                                                            do_stress, do_efield
      INTEGER, INTENT(IN)                                :: iw2
      LOGICAL, INTENT(IN)                                :: do_debug
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(section_vals_type), POINTER                   :: mm_section
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: efield0
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL           :: efield1, efield2
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: forces_local, forces_glob, pv_local, &
                                                            pv_glob

      CHARACTER(len=*), PARAMETER :: routineN = 'eval_pol_ewald', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      pot_nonbond = 0.0_dp ! Initialization..
      vg_coulomb = 0.0_dp ! Initialization..
      SELECT CASE (ewald_type)
      CASE (do_ewald_ewald)
         CALL ewald_multipole_evaluate(ewald_env, ewald_pw, fist_nonbond_env, cell, &
                                       particle_set, local_particles, vg_coulomb, pot_nonbond, thermo%e_neut, &
                                       thermo%e_self, multipoles%task, do_correction_bonded=do_correction_bonded, &
                                       do_forces=do_forces, do_stress=do_stress, do_efield=do_efield, &
                                       radii=multipoles%radii, charges=multipoles%charges, &
                                       dipoles=multipoles%dipoles, quadrupoles=multipoles%quadrupoles, &
                                       forces_local=forces_local, forces_glob=forces_glob, pv_local=pv_local, &
                                       pv_glob=pv_glob, iw=iw2, do_debug=do_debug, atomic_kind_set=atomic_kind_set, &
                                       mm_section=mm_section, efield0=efield0, efield1=efield1, efield2=efield2)
      CASE (do_ewald_pme)
         CALL cp__b("fist_pol_scf.F",718,"Multipole Ewald not yet implemented within a PME scheme!")
      CASE (do_ewald_spme)
         CALL cp__b("fist_pol_scf.F",720,"Multipole Ewald not yet implemented within a SPME scheme!")
      END SELECT
      CALL timestop(handle)
   END SUBROUTINE eval_pol_ewald

END MODULE fist_pol_scf
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Torsions added (DG) 05-Dec-2000
!> \author CJM
! **************************************************************************************************
MODULE mol_force

   USE force_field_kind_types,          ONLY: &
        do_ff_amber, do_ff_charmm, do_ff_cubic, do_ff_fues, do_ff_g87, do_ff_g96, do_ff_harmonic, &
        do_ff_legendre, do_ff_mixed_bend_stretch, do_ff_mm2, do_ff_mm3, do_ff_mm4, do_ff_morse, &
        do_ff_opls, do_ff_quartic, legendre_data_type
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mol_force'
   PUBLIC :: force_bonds, force_bends, force_torsions, force_imp_torsions, force_opbends
   PUBLIC :: get_pv_bond, get_pv_bend, get_pv_torsion

CONTAINS

! **************************************************************************************************
!> \brief Computes the forces from the bonds
!> \param id_type ...
!> \param rij ...
!> \param r0 ...
!> \param k ...
!> \param cs ...
!> \param energy ...
!> \param fscalar ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE force_bonds(id_type, rij, r0, k, cs, energy, fscalar)
      INTEGER, INTENT(IN)                                :: id_type
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rij
      REAL(KIND=dp), INTENT(IN)                          :: r0, k(3), cs
      REAL(KIND=dp), INTENT(OUT)                         :: energy, fscalar

      CHARACTER(len=*), PARAMETER :: routineN = 'force_bonds', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: f12 = 1.0_dp/2.0_dp, &
                                                            f13 = 1.0_dp/3.0_dp, &
                                                            f14 = 1.0_dp/4.0_dp

      REAL(KIND=dp)                                      :: dij, disp

      SELECT CASE (id_type)
      CASE (do_ff_quartic)
         dij = SQRT(DOT_PRODUCT(rij, rij))
         disp = dij-r0
         energy = (f12*k(1)+(f13*k(2)+f14*k(3)*disp)*disp)*disp*disp
         fscalar = ((k(1)+(k(2)+k(3)*disp)*disp)*disp)/dij
      CASE (do_ff_morse)
         dij = SQRT(DOT_PRODUCT(rij, rij))
         disp = dij-r0
         energy = k(1)*((1-EXP(-k(2)*disp))**2-1)
         fscalar = 2*k(1)*k(2)*EXP(-k(2)*disp)*(1-EXP(-k(2)*disp))/dij
      CASE (do_ff_cubic)
         dij = SQRT(DOT_PRODUCT(rij, rij))
         disp = dij-r0
         energy = k(1)*disp**2*(1+cs*disp+7.0_dp/12.0_dp*cs**2*disp**2)
         fscalar = (2.0_dp*k(1)*disp*(1+cs*disp+7.0_dp/12.0_dp*cs**2*disp**2)+ &
                    k(1)*disp**2*(cs+2.0_dp*7.0_dp/12.0_dp*cs**2*disp))/dij
      CASE (do_ff_g96)
         ! From GROMOS...
         ! V = (1/4)*Kb*(rij**2 - bij**2)**2
         dij = DOT_PRODUCT(rij, rij)
         disp = dij-r0*r0
         energy = f14*k(1)*disp*disp
         fscalar = k(1)*disp
      CASE (do_ff_charmm, do_ff_amber)
         dij = SQRT(DOT_PRODUCT(rij, rij))
         disp = dij-r0
         IF (ABS(disp) < EPSILON(1.0_dp)) THEN
            energy = 0.0_dp
            fscalar = 0.0_dp
         ELSE
            energy = k(1)*disp*disp
            fscalar = 2.0_dp*k(1)*disp/dij
         END IF
      CASE (do_ff_harmonic, do_ff_g87)
         dij = SQRT(DOT_PRODUCT(rij, rij))
         disp = dij-r0
         IF (ABS(disp) < EPSILON(1.0_dp)) THEN
            energy = 0.0_dp
            fscalar = 0.0_dp
         ELSE
            energy = f12*k(1)*disp*disp
            fscalar = k(1)*disp/dij
         END IF
      CASE (do_ff_fues)
         dij = SQRT(DOT_PRODUCT(rij, rij))
         disp = r0/dij
         energy = f12*k(1)*r0*r0*(1.0_dp+disp*(disp-2.0_dp))
         fscalar = k(1)*r0*disp*disp*(1.0_dp-disp)/dij
      CASE DEFAULT
         CALL cp__b("mol_force.F",103,"Unmatched bond kind")
      END SELECT

   END SUBROUTINE force_bonds

! **************************************************************************************************
!> \brief Computes the forces from the bends
!> \param id_type ...
!> \param b12 ...
!> \param b32 ...
!> \param d12 ...
!> \param d32 ...
!> \param id12 ...
!> \param id32 ...
!> \param dist ...
!> \param theta ...
!> \param theta0 ...
!> \param k ...
!> \param cb ...
!> \param r012 ...
!> \param r032 ...
!> \param kbs12 ...
!> \param kbs32 ...
!> \param kss ...
!> \param legendre ...
!> \param g1 ...
!> \param g2 ...
!> \param g3 ...
!> \param energy ...
!> \param fscalar ...
!> \par History
!>      Legendre Bonding Potential added 2015-11-02 [Felix Uhl]
!> \author CJM
! **************************************************************************************************
   SUBROUTINE force_bends(id_type, b12, b32, d12, d32, id12, id32, dist, &
                          theta, theta0, k, cb, r012, r032, kbs12, kbs32, kss, legendre, g1, g2, g3, energy, fscalar)
      INTEGER, INTENT(IN)                                :: id_type
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: b12, b32
      REAL(KIND=dp), INTENT(IN)                          :: d12, d32, id12, id32, dist, theta, &
                                                            theta0, k, cb, r012, r032, kbs12, &
                                                            kbs32, kss
      TYPE(legendre_data_type), INTENT(IN)               :: legendre
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: g1, g2, g3
      REAL(KIND=dp), INTENT(OUT)                         :: energy, fscalar

      CHARACTER(len=*), PARAMETER :: routineN = 'force_bends', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: f12 = 1.0_dp/2.0_dp

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: ctheta, denom, disp12, disp32, y0, y1, &
                                                            y2, yd0, yd1, yd2

      SELECT CASE (id_type)
      CASE (do_ff_g96)
         energy = f12*k*(COS(theta)-theta0)**2
         fscalar = -k*(COS(theta)-theta0)
         g1 = (b32*id32-b12*id12*COS(theta))*id12
         g3 = (b12*id12-b32*id32*COS(theta))*id32
         g2 = -g1-g3
      CASE (do_ff_charmm, do_ff_amber)
         denom = id12*id12*id32*id32
         energy = k*(theta-theta0)**2
         fscalar = 2.0_dp*k*(theta-theta0)/SIN(theta)
         g1 = (b32*d12*d32-dist*d32*id12*b12)*denom
         g2 = (-(b12+b32)*d12*d32+dist*(d32*id12*b12+id32*d12*b32))*denom
         g3 = (b12*d12*d32-dist*id32*d12*b32)*denom
      CASE (do_ff_cubic)
         denom = id12*id12*id32*id32
         energy = k*(theta-theta0)**2*(1.0_dp+cb*(theta-theta0))
         fscalar = (2.0_dp*k*(theta-theta0)*(1.0_dp+cb*(theta-theta0))+k*(theta-theta0)**2*cb)/SIN(theta)
         g1 = (b32*d12*d32-dist*d32*id12*b12)*denom
         g2 = (-(b12+b32)*d12*d32+dist*(d32*id12*b12+id32*d12*b32))*denom
         g3 = (b12*d12*d32-dist*id32*d12*b32)*denom
      CASE (do_ff_mixed_bend_stretch)

         ! 1) cubic term in theta (do_ff_cubic)
         energy = k*(theta-theta0)**2*(1.0_dp+cb*(theta-theta0))
         fscalar = (2.0_dp*k*(theta-theta0)*(1.0_dp+cb*(theta-theta0))+k*(theta-theta0)**2*cb)/SIN(theta)
         denom = id12*id12*id32*id32
         g1 = (b32*d12*d32-dist*d32*id12*b12)*denom*fscalar
         g2 = (-(b12+b32)*d12*d32+dist*(d32*id12*b12+id32*d12*b32))*denom*fscalar
         g3 = (b12*d12*d32-dist*id32*d12*b32)*denom*fscalar

         ! 2) stretch-stretch term
         disp12 = d12-r012
         disp32 = d32-r032
         energy = energy+kss*disp12*disp32
         g1 = g1-kss*disp32*id12*b12
         g2 = g2+kss*disp32*id12*b12
         g3 = g3-kss*disp12*id32*b32
         g2 = g2+kss*disp12*id32*b32

         ! 3) bend-stretch term
         energy = energy+kbs12*disp12*(theta-theta0)+kbs32*disp32*(theta-theta0)
         fscalar = (kbs12*disp12+kbs32*disp32)/SIN(theta)
         denom = id12*id12*id32*id32

         ! 3a) bend part
         g1 = g1+(b32*d12*d32-dist*d32*id12*b12)*denom*fscalar
         g2 = g2+(-(b12+b32)*d12*d32+dist*(d32*id12*b12+id32*d12*b32))*denom*fscalar
         g3 = g3+(b12*d12*d32-dist*id32*d12*b32)*denom*fscalar

         ! 3b) stretch part
         g1 = g1-kbs12*(theta-theta0)*id12*b12
         g2 = g2+kbs12*(theta-theta0)*id12*b12
         g3 = g3-kbs32*(theta-theta0)*id32*b32
         g2 = g2+kbs32*(theta-theta0)*id32*b32

         ! fscalar is already included in g1, g2 and g3
         fscalar = 1.0_dp

      CASE (do_ff_harmonic, do_ff_g87)
         denom = id12*id12*id32*id32
         energy = f12*k*(theta-theta0)**2
         fscalar = k*(theta-theta0)/SIN(theta)
         g1 = (b32*d12*d32-dist*d32*id12*b12)*denom
         g2 = (-(b12+b32)*d12*d32+dist*(d32*id12*b12+id32*d12*b32))*denom
         g3 = (b12*d12*d32-dist*id32*d12*b32)*denom

      CASE (do_ff_mm3)

         ! 1) up to sixth order in theta
         energy = k*(theta-theta0)**2*(0.5_dp+(theta-theta0)* &
                                       (-0.007_dp+(theta-theta0)*(2.8E-5_dp+(theta-theta0)* &
                                                                  (-3.5E-7_dp+(theta-theta0)*4.5E-10_dp))))

         fscalar = k*(theta-theta0)*(1.0_dp+(theta-theta0)* &
                                     (-0.021_dp+(theta-theta0)*(1.12E-4_dp+ &
                                                                (theta-theta0)*(-1.75E-6_dp+(theta-theta0)*2.7E-9_dp))))/ &
                   SIN(theta)

         denom = id12*id12*id32*id32
         g1 = (b32*d12*d32-dist*d32*id12*b12)*denom*fscalar
         g2 = (-(b12+b32)*d12*d32+dist*(d32*id12*b12+id32*d12*b32))*denom*fscalar
         g3 = (b12*d12*d32-dist*id32*d12*b32)*denom*fscalar
         ! 2) bend-stretch term
         disp12 = d12-r012
         disp32 = d32-r032
         energy = energy+kbs12*disp12*(theta-theta0)+kbs32*disp32*(theta-theta0)
         fscalar = (kbs12*disp12+kbs32*disp32)/SIN(theta)
         denom = id12*id12*id32*id32

         ! 2a) bend part
         g1 = g1+(b32*d12*d32-dist*d32*id12*b12)*denom*fscalar
         g2 = g2+(-(b12+b32)*d12*d32+dist*(d32*id12*b12+id32*d12*b32))*denom*fscalar
         g3 = g3+(b12*d12*d32-dist*id32*d12*b32)*denom*fscalar

         ! 2b) stretch part
         g1 = g1-kbs12*(theta-theta0)*id12*b12
         g2 = g2+kbs12*(theta-theta0)*id12*b12
         g3 = g3-kbs32*(theta-theta0)*id32*b32
         g2 = g2+kbs32*(theta-theta0)*id32*b32

         ! fscalar is already included in g1, g2 and g3
         fscalar = 1.0_dp
      CASE (do_ff_legendre)
         ! Calculates f(x) = sum_{n=0}^{nmax} c(n) * P(n,x)
         !
         ! Legendre Polynomials recursion formula:
         ! P(n+1,x) = x*(2n+1)/(n+1) * P(n,x) - n/(n+1) * P(n-1,x)     n = 1, 2,... ; P(0,x) = 1, P(1,x) = x, ...
         ! P(n+1,x) = alpha(n,x) * P(n,x) + beta(n,x) * P(n-1,x)
         ! with alpha(n,x) = x*(2n+1)/(n+1)
         ! and  beta(n,x) = -n/(n+1)
         !
         ! Therefore
         ! f(x) = sum_{n=0}^{nmax} c(n) * P(n,x)
         ! can be calculated using a Clenshaw recursion
         ! (c(n) are the coefficients for the Legendre polynomial expansion)
         !
         ! f(x) = beta(1,x)*P(0,x)*y(2) + P(1,x)*y(1) + P(0,x)*c(0)
         ! beta(1,x) = -0.5
         ! y(k) is calculated in the following way:
         ! y(nmax+1) = 0
         ! y(nmax+2) = 0
         ! y(k) = alpha(k,x)*y(k+1) + beta(k+1,x)*y(k+2) + c(k)

         ! Calculates f'(x) = sum_{n=0}^{nmax} c(n) * P'(n,x)
         !
         ! Recursion formula for the m-th derivatives of Legendre Polynomials
         ! P^{m}(n+1,x) = x*(2n+1)/(n+1-m) * P^{m}(n,x) -(n+m)/(n+1-m) * P^{m}(n-1,x)   n = 1, 2, ... ; m = 1, ..., n-1
         ! For first derivative:
         ! P'(n+1,x) = x*(2n+1)/n * P'(n,x) - (n+1)/n * P'(n-1,x) with P(0,x) = 0; P(1,x) = 1
         ! P'(n+1,x) = alpha(n,x) * P'(n,x) + beta(n,x) * P'(n-1,x)
         ! with alpha(n,x) = x*(2n+1)/n
         ! and  beta(n,x) = -1*(n+1)/n
         !
         ! Therefore Clenshaw recursion can be used
         ! f'(x) = beta(1,x)*P'(0,x)*y(2) + P'(1,x)*y(1) + P'(0,x)*c(0)
         !       = beta(1,x)*0*y(2)      +        y(1) + 0
         !       = y(1)
         ! y(k) is calculated in the following way:
         ! y(nmax+1) = 0
         ! y(nmax+2) = 0
         ! y(k) = alpha(k,x)*y(k+1) + beta(k+1,x)*y(k+2) + c(k)
         !
         ctheta = COS(theta)
         y1 = 0.0d0
         y2 = 0.0d0
         yd1 = 0.0d0
         yd2 = 0.0d0
         DO i = legendre%order, 2, -1
            y0 = (2*i-1)*ctheta*y1/i-i*y2/(i+1)+legendre%coeffs(i)
            y2 = y1
            y1 = y0
            yd0 = (2*i-1)*ctheta*yd1/(i-1)-(i+1)*yd2/i+legendre%coeffs(i)
            yd2 = yd1
            yd1 = yd0
         END DO

         energy = -f12*y2+ctheta*y1+legendre%coeffs(1)
         fscalar = -yd1
         g1 = (b32*id32-b12*id12*ctheta)*id12
         g3 = (b12*id12-b32*id32*ctheta)*id32
         g2 = -g1-g3

      CASE DEFAULT
         CALL cp__b("mol_force.F",319,"Unmatched bend kind")
      END SELECT

   END SUBROUTINE force_bends

! **************************************************************************************************
!> \brief Computes the forces from the torsions
!> \param id_type ...
!> \param s32 ...
!> \param is32 ...
!> \param ism ...
!> \param isn ...
!> \param dist1 ...
!> \param dist2 ...
!> \param tm ...
!> \param tn ...
!> \param t12 ...
!> \param k ...
!> \param phi0 ...
!> \param m ...
!> \param gt1 ...
!> \param gt2 ...
!> \param gt3 ...
!> \param gt4 ...
!> \param energy ...
!> \param fscalar ...
!> \par History
!>      none
!> \author DG
! **************************************************************************************************
   SUBROUTINE force_torsions(id_type, s32, is32, ism, isn, dist1, dist2, tm, &
                             tn, t12, k, phi0, m, gt1, gt2, gt3, gt4, energy, fscalar)
      INTEGER, INTENT(IN)                                :: id_type
      REAL(KIND=dp), INTENT(IN)                          :: s32, is32, ism, isn, dist1, dist2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: tm, tn, t12
      REAL(KIND=dp), INTENT(IN)                          :: k, phi0
      INTEGER, INTENT(IN)                                :: m
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: gt1, gt2, gt3, gt4
      REAL(KIND=dp), INTENT(OUT)                         :: energy, fscalar

      CHARACTER(len=*), PARAMETER :: routineN = 'force_torsions', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: cosphi, phi

      cosphi = DOT_PRODUCT(tm, tn)*ism*isn
      IF (cosphi > 1.0_dp) cosphi = 1.0_dp
      IF (cosphi < -1.0_dp) cosphi = -1.0_dp
      phi = SIGN(ACOS(cosphi), DOT_PRODUCT(t12, tn))

      ! Select force field
      SELECT CASE (id_type)
      CASE (do_ff_charmm, do_ff_g87, do_ff_g96, do_ff_amber, do_ff_opls)
         ! compute energy
         energy = k*(1.0_dp+COS(m*phi-phi0))

         ! compute fscalar
         fscalar = k*m*SIN(m*phi-phi0)
      CASE DEFAULT
         CALL cp__b("mol_force.F",377,"Unmatched torsion kind")
      END SELECT

      ! compute the gradients
      gt1 = (s32*ism*ism)*tm
      gt4 = -(s32*isn*isn)*tn
      gt2 = (dist1*is32**2-1.0_dp)*gt1-dist2*is32**2*gt4
      gt3 = (dist2*is32**2-1.0_dp)*gt4-dist1*is32**2*gt1
   END SUBROUTINE force_torsions

! **************************************************************************************************
!> \brief Computes the forces from the improper torsions
!> \param id_type ...
!> \param s32 ...
!> \param is32 ...
!> \param ism ...
!> \param isn ...
!> \param dist1 ...
!> \param dist2 ...
!> \param tm ...
!> \param tn ...
!> \param t12 ...
!> \param k ...
!> \param phi0 ...
!> \param gt1 ...
!> \param gt2 ...
!> \param gt3 ...
!> \param gt4 ...
!> \param energy ...
!> \param fscalar ...
!> \par History
!>      none
!> \author DG
! **************************************************************************************************
   SUBROUTINE force_imp_torsions(id_type, s32, is32, ism, isn, dist1, dist2, tm, &
                                 tn, t12, k, phi0, gt1, gt2, gt3, gt4, energy, fscalar)
      INTEGER, INTENT(IN)                                :: id_type
      REAL(KIND=dp), INTENT(IN)                          :: s32, is32, ism, isn, dist1, dist2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: tm, tn, t12
      REAL(KIND=dp), INTENT(IN)                          :: k, phi0
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: gt1, gt2, gt3, gt4
      REAL(KIND=dp), INTENT(OUT)                         :: energy, fscalar

      CHARACTER(len=*), PARAMETER :: routineN = 'force_imp_torsions', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: f12 = 1.0_dp/2.0_dp

      REAL(KIND=dp)                                      :: cosphi, phi

! define cosphi

      cosphi = DOT_PRODUCT(tm, tn)*ism*isn
      IF (cosphi > 1.0_dp) cosphi = 1.0_dp
      IF (cosphi < -1.0_dp) cosphi = -1.0_dp
      phi = SIGN(ACOS(cosphi), DOT_PRODUCT(t12, tn))

      SELECT CASE (id_type)
      CASE (do_ff_charmm)
         ! compute energy
         energy = k*(phi-phi0)**2

         ! compute fscalar
         fscalar = -2.0_dp*k*(phi-phi0)

      CASE (do_ff_harmonic, do_ff_g87, do_ff_g96)
         ! compute energy
         energy = f12*k*(phi-phi0)**2

         ! compute fscalar
         fscalar = -k*(phi-phi0)

      CASE DEFAULT
         CALL cp__b("mol_force.F",449,"Unmatched improper kind")
      END SELECT

      ! compute the gradients
      gt1 = (s32*ism*ism)*tm
      gt4 = -(s32*isn*isn)*tn
      gt2 = (dist1*is32**2-1.0_dp)*gt1-dist2*is32**2*gt4
      gt3 = (dist2*is32**2-1.0_dp)*gt4-dist1*is32**2*gt1
   END SUBROUTINE force_imp_torsions

   ! **************************************************************************************************
!> \brief Computes the forces from the out of plane bends
!> \param id_type ...
!> \param s32 ...
!> \param tm ...
!> \param t41 ...
!> \param t42 ...
!> \param t43 ...
!> \param k ...
!> \param phi0 ...
!> \param gt1 ...
!> \param gt2 ...
!> \param gt3 ...
!> \param gt4 ...
!> \param energy ...
!> \param fscalar ...
!> \par History
!>      none
!> \author Louis Vanduyfhuys
! **************************************************************************************************
   SUBROUTINE force_opbends(id_type, s32, tm, &
                            t41, t42, t43, k, phi0, gt1, gt2, gt3, gt4, energy, fscalar)

      INTEGER, INTENT(IN)                                :: id_type
      REAL(KIND=dp), INTENT(IN)                          :: s32
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: tm, t41, t42, t43
      REAL(KIND=dp), INTENT(IN)                          :: k, phi0
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: gt1, gt2, gt3, gt4
      REAL(KIND=dp), INTENT(OUT)                         :: energy, fscalar

      CHARACTER(len=*), PARAMETER :: routineN = 'force_opbends', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: f12 = 1.0_dp/2.0_dp

      REAL(KIND=dp)                                      :: b, C, cosphi, D, E, is41, phi
      REAL(KIND=dp), DIMENSION(3)                        :: db_dq1, db_dq2, db_dq3, dC_dq1, dC_dq2, &
                                                            dC_dq3, dD_dq1, dD_dq2, dD_dq3, &
                                                            dE_dq1, dE_dq2, dE_dq3

!compute the energy and the gradients of cos(phi), see
!   "Efficient treatment of out-of-plane bend and improper torsion interactions in
!   MM2, MM3 and MM4 Molecular mechanicsd calculations.", Robert E. Tuzun, Donald W.Noid,
!   Bobby G Sumpter, Chemical and Analytical Sciences Division, Oak Ridge National
!   Laboratory, P.O. Box 2008, Oak Ridge, Tennesse 37831-6497
!CAUTION: in the paper r_ij = rj - ri
!in fist_intra_force.F t_ij = ri - rj
!hence a minus sign needs to be added to convert r_ij vectors in t_ij vectors
!tm is the normal of the plane 123, tm = t32 x t12 (= w from paper)
!tn = - t41 x tm (= a from paper, for minus sign see CAUTION above)
!Computing some necessary variables (see paper)

      E = DOT_PRODUCT(-t41, tm)
      C = DOT_PRODUCT(tm, tm)
      D = E**2/C
      b = DOT_PRODUCT(t41, t41)-D

      !inverse norm of t41
      is41 = 1.0_dp/SQRT(DOT_PRODUCT(t41, t41))

      cosphi = SQRT(b)*is41
      IF (cosphi > 1.0_dp) cosphi = 1.0_dp
      IF (cosphi < -1.0_dp) cosphi = -1.0_dp
      phi = SIGN(ACOS(cosphi), DOT_PRODUCT(tm, -t41))

      SELECT CASE (id_type)
      CASE (do_ff_mm2, do_ff_mm3, do_ff_mm4)
         ! compute energy
         energy = k*(phi-phi0)**2

         ! compute fscalar
         fscalar = 2.0_dp*k*(phi-phi0)*is41

      CASE (do_ff_harmonic)
         ! compute energy
         energy = f12*k*(phi-phi0)**2

         ! compute fscalar
         fscalar = k*(phi-phi0)*is41

      CASE DEFAULT
         CALL cp__b("mol_force.F",538,"Unmatched opbend kind")
      END SELECT

      !Computing the necessary intermediate variables. dX_dqi is the gradient
      !of X with respect to the coordinates of particle i.

      dE_dq1(1) = (t42(2)*t43(3)-t43(2)*t42(3))
      dE_dq1(2) = (-t42(1)*t43(3)+t43(1)*t42(3))
      dE_dq1(3) = (t42(1)*t43(2)-t43(1)*t42(2))

      dE_dq2(1) = (t43(2)*t41(3)-t41(2)*t43(3))
      dE_dq2(2) = (-t43(1)*t41(3)+t41(1)*t43(3))
      dE_dq2(3) = (t43(1)*t41(2)-t41(1)*t43(2))

      dE_dq3(1) = (t41(2)*t42(3)-t42(2)*t41(3))
      dE_dq3(2) = (-t41(1)*t42(3)+t42(1)*t41(3))
      dE_dq3(3) = (t41(1)*t42(2)-t42(1)*t41(2))

      dC_dq1 = 2.0_dp*((t42-t41)*s32**2-(t42-t43)*DOT_PRODUCT(t42-t41, t42-t43))
      dC_dq3 = 2.0_dp*((t42-t43)*DOT_PRODUCT(t41-t42, t41-t42) &
                       -(t42-t41)*DOT_PRODUCT(t42-t41, t42-t43))
      !C only dependent of atom 1 2 and 3, using translational invariance we find
      dC_dq2 = -(dC_dq1+dC_dq3)

      dD_dq1 = (2.0_dp*E*dE_dq1-D*dC_dq1)/C
      dD_dq2 = (2.0_dp*E*dE_dq2-D*dC_dq2)/C
      dD_dq3 = (2.0_dp*E*dE_dq3-D*dC_dq3)/C

      db_dq1 = -2.0_dp*t41-dD_dq1
      db_dq2 = -dD_dq2
      db_dq3 = -dD_dq3

      !gradients of cos(phi), gt4 is calculated using translational invariance.
      !The 1/r41 factor from the paper is absorbed in fscalar.
      !If phi = 0 then sin(phi) = 0 and the regular formula for calculating gt
      !won't work because of the sine function in the denominator. A further
      !analytic simplification is needed.
      IF (phi == 0) THEN
         gt1 = -SIGN(1.0_dp, phi)/SQRT(C)*dE_dq1
         gt2 = -SIGN(1.0_dp, phi)/SQRT(C)*dE_dq2
         gt3 = -SIGN(1.0_dp, phi)/SQRT(C)*dE_dq3
         gt4 = -(gt1+gt2+gt3)

      ELSE
         gt1 = (1.0_dp/(2.0_dp*SQRT(b))*db_dq1+cosphi*t41*is41)/SIN(phi)
         gt2 = (1.0_dp/(2.0_dp*SQRT(b))*db_dq2)/SIN(phi)
         gt3 = (1.0_dp/(2.0_dp*SQRT(b))*db_dq3)/SIN(phi)
         gt4 = -(gt1+gt2+gt3)
      END IF
   END SUBROUTINE force_opbends

! **************************************************************************************************
!> \brief Computes the pressure tensor from the bonds
!> \param f12 ...
!> \param r12 ...
!> \param pv_bond ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE get_pv_bond(f12, r12, pv_bond)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: f12, r12
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: pv_bond

      CHARACTER(len=*), PARAMETER :: routineN = 'get_pv_bond', routineP = moduleN//':'//routineN

      pv_bond(1, 1) = pv_bond(1, 1)+f12(1)*r12(1)
      pv_bond(1, 2) = pv_bond(1, 2)+f12(1)*r12(2)
      pv_bond(1, 3) = pv_bond(1, 3)+f12(1)*r12(3)
      pv_bond(2, 1) = pv_bond(2, 1)+f12(2)*r12(1)
      pv_bond(2, 2) = pv_bond(2, 2)+f12(2)*r12(2)
      pv_bond(2, 3) = pv_bond(2, 3)+f12(2)*r12(3)
      pv_bond(3, 1) = pv_bond(3, 1)+f12(3)*r12(1)
      pv_bond(3, 2) = pv_bond(3, 2)+f12(3)*r12(2)
      pv_bond(3, 3) = pv_bond(3, 3)+f12(3)*r12(3)

   END SUBROUTINE get_pv_bond

! **************************************************************************************************
!> \brief Computes the pressure tensor from the bends
!> \param f1 ...
!> \param f3 ...
!> \param r12 ...
!> \param r32 ...
!> \param pv_bend ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE get_pv_bend(f1, f3, r12, r32, pv_bend)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: f1, f3, r12, r32
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: pv_bend

      CHARACTER(len=*), PARAMETER :: routineN = 'get_pv_bend', routineP = moduleN//':'//routineN

      pv_bend(1, 1) = pv_bend(1, 1)+f1(1)*r12(1)
      pv_bend(1, 1) = pv_bend(1, 1)+f3(1)*r32(1)
      pv_bend(1, 2) = pv_bend(1, 2)+f1(1)*r12(2)
      pv_bend(1, 2) = pv_bend(1, 2)+f3(1)*r32(2)
      pv_bend(1, 3) = pv_bend(1, 3)+f1(1)*r12(3)
      pv_bend(1, 3) = pv_bend(1, 3)+f3(1)*r32(3)
      pv_bend(2, 1) = pv_bend(2, 1)+f1(2)*r12(1)
      pv_bend(2, 1) = pv_bend(2, 1)+f3(2)*r32(1)
      pv_bend(2, 2) = pv_bend(2, 2)+f1(2)*r12(2)
      pv_bend(2, 2) = pv_bend(2, 2)+f3(2)*r32(2)
      pv_bend(2, 3) = pv_bend(2, 3)+f1(2)*r12(3)
      pv_bend(2, 3) = pv_bend(2, 3)+f3(2)*r32(3)
      pv_bend(3, 1) = pv_bend(3, 1)+f1(3)*r12(1)
      pv_bend(3, 1) = pv_bend(3, 1)+f3(3)*r32(1)
      pv_bend(3, 2) = pv_bend(3, 2)+f1(3)*r12(2)
      pv_bend(3, 2) = pv_bend(3, 2)+f3(3)*r32(2)
      pv_bend(3, 3) = pv_bend(3, 3)+f1(3)*r12(3)
      pv_bend(3, 3) = pv_bend(3, 3)+f3(3)*r32(3)

   END SUBROUTINE get_pv_bend

! **************************************************************************************************
!> \brief Computes the pressure tensor from the torsions (also used for impr
!>        and opbend)
!> \param f1 ...
!> \param f3 ...
!> \param f4 ...
!> \param r12 ...
!> \param r32 ...
!> \param r43 ...
!> \param pv_torsion ...
!> \par History
!>      none
!> \author DG
! **************************************************************************************************
   SUBROUTINE get_pv_torsion(f1, f3, f4, r12, r32, r43, pv_torsion)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: f1, f3, f4, r12, r32, r43
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: pv_torsion

      CHARACTER(len=*), PARAMETER :: routineN = 'get_pv_torsion', routineP = moduleN//':'//routineN

      pv_torsion(1, 1) = pv_torsion(1, 1)+f1(1)*r12(1)
      pv_torsion(1, 1) = pv_torsion(1, 1)+(f3(1)+f4(1))*r32(1)
      pv_torsion(1, 1) = pv_torsion(1, 1)+f4(1)*r43(1)
      pv_torsion(1, 2) = pv_torsion(1, 2)+f1(1)*r12(2)
      pv_torsion(1, 2) = pv_torsion(1, 2)+(f3(1)+f4(1))*r32(2)
      pv_torsion(1, 2) = pv_torsion(1, 2)+f4(1)*r43(2)
      pv_torsion(1, 3) = pv_torsion(1, 3)+f1(1)*r12(3)
      pv_torsion(1, 3) = pv_torsion(1, 3)+(f3(1)+f4(1))*r32(3)
      pv_torsion(1, 3) = pv_torsion(1, 3)+f4(1)*r43(3)
      pv_torsion(2, 1) = pv_torsion(2, 1)+f1(2)*r12(1)
      pv_torsion(2, 1) = pv_torsion(2, 1)+(f3(2)+f4(2))*r32(1)
      pv_torsion(2, 1) = pv_torsion(2, 1)+f4(2)*r43(1)
      pv_torsion(2, 2) = pv_torsion(2, 2)+f1(2)*r12(2)
      pv_torsion(2, 2) = pv_torsion(2, 2)+(f3(2)+f4(2))*r32(2)
      pv_torsion(2, 2) = pv_torsion(2, 2)+f4(2)*r43(2)
      pv_torsion(2, 3) = pv_torsion(2, 3)+f1(2)*r12(3)
      pv_torsion(2, 3) = pv_torsion(2, 3)+(f3(2)+f4(2))*r32(3)
      pv_torsion(2, 3) = pv_torsion(2, 3)+f4(2)*r43(3)
      pv_torsion(3, 1) = pv_torsion(3, 1)+f1(3)*r12(1)
      pv_torsion(3, 1) = pv_torsion(3, 1)+(f3(3)+f4(3))*r32(1)
      pv_torsion(3, 1) = pv_torsion(3, 1)+f4(3)*r43(1)
      pv_torsion(3, 2) = pv_torsion(3, 2)+f1(3)*r12(2)
      pv_torsion(3, 2) = pv_torsion(3, 2)+(f3(3)+f4(3))*r32(2)
      pv_torsion(3, 2) = pv_torsion(3, 2)+f4(3)*r43(2)
      pv_torsion(3, 3) = pv_torsion(3, 3)+f1(3)*r12(3)
      pv_torsion(3, 3) = pv_torsion(3, 3)+(f3(3)+f4(3))*r32(3)
      pv_torsion(3, 3) = pv_torsion(3, 3)+f4(3)*r43(3)

   END SUBROUTINE get_pv_torsion

END MODULE mol_force

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Torsions added (DG) 05-Dec-2000
!>      Variable names changed (DG) 05-Dec-2000
!> \author CJM
! **************************************************************************************************
MODULE fist_intra_force

   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE kinds,                           ONLY: dp
   USE mol_force,                       ONLY: force_bends,&
                                              force_bonds,&
                                              force_imp_torsions,&
                                              force_opbends,&
                                              force_torsions,&
                                              get_pv_bend,&
                                              get_pv_bond,&
                                              get_pv_torsion
   USE molecule_kind_types,             ONLY: &
        bend_type, bond_type, get_molecule_kind, impr_type, molecule_kind_type, opbend_type, &
        shell_type, torsion_type, ub_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              molecule_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_intra_force'
   PUBLIC :: force_intra_control

CONTAINS

! **************************************************************************************************
!> \brief Computes the the intramolecular energies, forces, and pressure tensors
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param particle_set ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param pot_bond ...
!> \param pot_bend ...
!> \param pot_urey_bradley ...
!> \param pot_torsion ...
!> \param pot_imp_torsion ...
!> \param pot_opbend ...
!> \param pot_shell ...
!> \param pv_bond ...
!> \param pv_bend ...
!> \param pv_urey_bradley ...
!> \param pv_torsion ...
!> \param pv_imp_torsion ...
!> \param pv_opbend ...
!> \param f_bond ...
!> \param f_bend ...
!> \param f_torsion ...
!> \param f_ub ...
!> \param f_imptor ...
!> \param f_opbend ...
!> \param cell ...
!> \param use_virial ...
!> \param atprop_env ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE force_intra_control(molecule_set, molecule_kind_set, &
                                  local_molecules, particle_set, shell_particle_set, core_particle_set, &
                                  pot_bond, pot_bend, pot_urey_bradley, pot_torsion, pot_imp_torsion, &
                                  pot_opbend, pot_shell, pv_bond, pv_bend, pv_urey_bradley, pv_torsion, &
                                  pv_imp_torsion, pv_opbend, f_bond, f_bend, f_torsion, f_ub, &
                                  f_imptor, f_opbend, cell, use_virial, atprop_env)

      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(particle_type), POINTER                       :: particle_set(:), shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pot_bond, pot_bend, pot_urey_bradley, &
                                                            pot_torsion, pot_imp_torsion, &
                                                            pot_opbend, pot_shell
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: pv_bond, pv_bend, pv_urey_bradley, &
                                                            pv_torsion, pv_imp_torsion, pv_opbend
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: f_bond, f_bend, f_torsion, f_ub, &
                                                            f_imptor, f_opbend
      TYPE(cell_type), POINTER                           :: cell
      LOGICAL, INTENT(IN)                                :: use_virial
      TYPE(atprop_type), POINTER                         :: atprop_env

      CHARACTER(len=*), PARAMETER :: routineN = 'force_intra_control', &
         routineP = moduleN//':'//routineN

      INTEGER :: first_atom, handle, i, ibend, ibond, ikind, imol, imul, index_a, index_b, &
         index_c, index_d, iopbend, ishell, itorsion, nbends, nbonds, nimptors, nkind, &
         nmol_per_kind, nopbends, nshell, ntorsions, nub
      LOGICAL                                            :: atener, atstress
      REAL(KIND=dp)                                      :: d12, d32, dist, dist1, dist2, energy, &
                                                            fscalar, id12, id32, is32, ism, isn, &
                                                            k2_spring, k4_spring, r2, s32, sm, sn, &
                                                            theta
      REAL(KIND=dp), DIMENSION(3)                        :: b12, b32, g1, g2, g3, gt1, gt2, gt3, &
                                                            gt4, k1, k2, k3, k4, rij, t12, t32, &
                                                            t34, t41, t42, t43, tm, tn
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ener_a
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pv_a
      TYPE(bend_type), POINTER                           :: bend_list(:)
      TYPE(bond_type), POINTER                           :: bond_list(:)
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(impr_type), POINTER                           :: impr_list(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(opbend_type), POINTER                         :: opbend_list(:)
      TYPE(shell_type), POINTER                          :: shell_list(:)
      TYPE(torsion_type), POINTER                        :: torsion_list(:)
      TYPE(ub_type), POINTER                             :: ub_list(:)

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF (PRESENT(f_bond)) f_bond = 0.0_dp
      IF (PRESENT(f_bend)) f_bend = 0.0_dp
      IF (PRESENT(f_torsion)) f_torsion = 0.0_dp
      IF (PRESENT(f_imptor)) f_imptor = 0.0_dp
      IF (PRESENT(f_ub)) f_ub = 0.0_dp

      pot_bond = 0.0_dp
      pot_bend = 0.0_dp
      pot_urey_bradley = 0.0_dp
      pot_torsion = 0.0_dp
      pot_imp_torsion = 0.0_dp
      pot_opbend = 0.0_dp
      pot_shell = 0.0_dp

      atener = atprop_env%energy
      IF (atener) ener_a => atprop_env%atener
      atstress = atprop_env%stress
      IF (atstress) pv_a => atprop_env%atstress

      nkind = SIZE(molecule_kind_set)
      MOL: DO ikind = 1, nkind
         nmol_per_kind = local_molecules%n_el(ikind)

         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            molecule_kind => molecule%molecule_kind
            CALL get_molecule_kind(molecule_kind, nbend=nbends, nbond=nbonds, &
                                   nimpr=nimptors, nub=nub, ntorsion=ntorsions, &
                                   nopbend=nopbends, nshell=nshell, &
                                   bond_list=bond_list, ub_list=ub_list, &
                                   bend_list=bend_list, torsion_list=torsion_list, &
                                   impr_list=impr_list, opbend_list=opbend_list, &
                                   shell_list=shell_list)

            CALL get_molecule(molecule, first_atom=first_atom)

            BOND: DO ibond = 1, nbonds
               index_a = bond_list(ibond)%a+first_atom-1
               index_b = bond_list(ibond)%b+first_atom-1
               rij = particle_set(index_a)%r-particle_set(index_b)%r
               rij = pbc(rij, cell)
               CALL force_bonds(bond_list(ibond)%bond_kind%id_type, rij, &
                                bond_list(ibond)%bond_kind%r0, &
                                bond_list(ibond)%bond_kind%k, &
                                bond_list(ibond)%bond_kind%cs, &
                                energy, fscalar)
               pot_bond = pot_bond+energy
               IF (atener) THEN
                  ener_a(index_a) = ener_a(index_a)+0.5_dp*energy
                  ener_a(index_b) = ener_a(index_b)+0.5_dp*energy
               END IF

               particle_set(index_a)%f(1) = particle_set(index_a)%f(1)-rij(1)*fscalar
               particle_set(index_a)%f(2) = particle_set(index_a)%f(2)-rij(2)*fscalar
               particle_set(index_a)%f(3) = particle_set(index_a)%f(3)-rij(3)*fscalar
               particle_set(index_b)%f(1) = particle_set(index_b)%f(1)+rij(1)*fscalar
               particle_set(index_b)%f(2) = particle_set(index_b)%f(2)+rij(2)*fscalar
               particle_set(index_b)%f(3) = particle_set(index_b)%f(3)+rij(3)*fscalar

               ! computing the pressure tensor
               k2 = -rij*fscalar
               IF (use_virial) CALL get_pv_bond(k2, rij, pv_bond)
               IF (atstress) THEN
                  CALL get_pv_bond(0.5_dp*k2, rij, pv_a(:, :, index_a))
                  CALL get_pv_bond(0.5_dp*k2, rij, pv_a(:, :, index_b))
               END IF

               ! the contribution from the bonds. ONLY FOR DEBUG
               IF (PRESENT(f_bond)) THEN
                  f_bond(1, index_a) = f_bond(1, index_a)-rij(1)*fscalar
                  f_bond(2, index_a) = f_bond(2, index_a)-rij(2)*fscalar
                  f_bond(3, index_a) = f_bond(3, index_a)-rij(3)*fscalar
                  f_bond(1, index_b) = f_bond(1, index_b)+rij(1)*fscalar
                  f_bond(2, index_b) = f_bond(2, index_b)+rij(2)*fscalar
                  f_bond(3, index_b) = f_bond(3, index_b)+rij(3)*fscalar
               END IF

            END DO BOND

            SHELL: DO ishell = 1, nshell
               index_a = shell_list(ishell)%a+first_atom-1
               index_b = particle_set(index_a)%shell_index
               rij = core_particle_set(index_b)%r-shell_particle_set(index_b)%r
               rij = pbc(rij, cell)
               k2_spring = shell_list(ishell)%shell_kind%k2_spring
               k4_spring = shell_list(ishell)%shell_kind%k4_spring
               r2 = DOT_PRODUCT(rij, rij)
               energy = 0.5_dp*(k2_spring+k4_spring*r2/12.0_dp)*r2
               fscalar = k2_spring+k4_spring*r2/6.0_dp
               pot_shell = pot_shell+energy
               IF (atener) THEN
                  ener_a(index_a) = ener_a(index_a)+energy
               END IF
               core_particle_set(index_b)%f(1) = core_particle_set(index_b)%f(1)-rij(1)*fscalar
               core_particle_set(index_b)%f(2) = core_particle_set(index_b)%f(2)-rij(2)*fscalar
               core_particle_set(index_b)%f(3) = core_particle_set(index_b)%f(3)-rij(3)*fscalar
               shell_particle_set(index_b)%f(1) = shell_particle_set(index_b)%f(1)+rij(1)*fscalar
               shell_particle_set(index_b)%f(2) = shell_particle_set(index_b)%f(2)+rij(2)*fscalar
               shell_particle_set(index_b)%f(3) = shell_particle_set(index_b)%f(3)+rij(3)*fscalar
               ! Compute the pressure tensor, if requested
               IF (use_virial) THEN
                  k1 = -rij*fscalar
                  CALL get_pv_bond(k1, rij, pv_bond)
               END IF
               IF (atstress) THEN
                  CALL get_pv_bond(k1, rij, pv_a(:, :, index_a))
               END IF
            END DO SHELL

            UREY_BRADLEY: DO ibend = 1, nub
               index_a = ub_list(ibend)%a+first_atom-1
               index_b = ub_list(ibend)%c+first_atom-1
               rij = particle_set(index_a)%r-particle_set(index_b)%r
               rij = pbc(rij, cell)
               CALL force_bonds(ub_list(ibend)%ub_kind%id_type, rij, &
                                ub_list(ibend)%ub_kind%r0, &
                                ub_list(ibend)%ub_kind%k, 0.0_dp, energy, fscalar)
               pot_urey_bradley = pot_urey_bradley+energy
               IF (atener) THEN
                  ener_a(index_a) = ener_a(index_a)+0.5_dp*energy
                  ener_a(index_b) = ener_a(index_b)+0.5_dp*energy
               END IF
               particle_set(index_a)%f(1) = particle_set(index_a)%f(1)-rij(1)*fscalar
               particle_set(index_a)%f(2) = particle_set(index_a)%f(2)-rij(2)*fscalar
               particle_set(index_a)%f(3) = particle_set(index_a)%f(3)-rij(3)*fscalar
               particle_set(index_b)%f(1) = particle_set(index_b)%f(1)+rij(1)*fscalar
               particle_set(index_b)%f(2) = particle_set(index_b)%f(2)+rij(2)*fscalar
               particle_set(index_b)%f(3) = particle_set(index_b)%f(3)+rij(3)*fscalar

               ! computing the pressure tensor
               k2 = -rij*fscalar
               IF (use_virial) CALL get_pv_bond(k2, rij, pv_urey_bradley)
               IF (atstress) THEN
                  CALL get_pv_bond(0.5_dp*k2, rij, pv_a(:, :, index_a))
                  CALL get_pv_bond(0.5_dp*k2, rij, pv_a(:, :, index_b))
               END IF

               ! the contribution from the ub. ONLY FOR DEBUG
               IF (PRESENT(f_ub)) THEN
                  f_ub(:, index_a) = f_ub(:, index_a)-rij*fscalar
                  f_ub(:, index_b) = f_ub(:, index_b)+rij*fscalar
               END IF

            END DO UREY_BRADLEY

            BEND: DO ibend = 1, nbends
               index_a = bend_list(ibend)%a+first_atom-1
               index_b = bend_list(ibend)%b+first_atom-1
               index_c = bend_list(ibend)%c+first_atom-1
               b12 = particle_set(index_a)%r-particle_set(index_b)%r
               b32 = particle_set(index_c)%r-particle_set(index_b)%r
               b12 = pbc(b12, cell)
               b32 = pbc(b32, cell)
               d12 = SQRT(DOT_PRODUCT(b12, b12))
               id12 = 1.0_dp/d12
               d32 = SQRT(DOT_PRODUCT(b32, b32))
               id32 = 1.0_dp/d32
               dist = DOT_PRODUCT(b12, b32)
               theta = (dist*id12*id32)
               IF (theta < -1.0_dp) theta = -1.0_dp
               IF (theta > +1.0_dp) theta = +1.0_dp
               theta = ACOS(theta)
               CALL force_bends(bend_list(ibend)%bend_kind%id_type, &
                                b12, b32, d12, d32, id12, id32, dist, theta, &
                                bend_list(ibend)%bend_kind%theta0, &
                                bend_list(ibend)%bend_kind%k, &
                                bend_list(ibend)%bend_kind%cb, &
                                bend_list(ibend)%bend_kind%r012, &
                                bend_list(ibend)%bend_kind%r032, &
                                bend_list(ibend)%bend_kind%kbs12, &
                                bend_list(ibend)%bend_kind%kbs32, &
                                bend_list(ibend)%bend_kind%kss, &
                                bend_list(ibend)%bend_kind%legendre, &
                                g1, g2, g3, energy, fscalar)
               pot_bend = pot_bend+energy
               IF (atener) THEN
                  ener_a(index_a) = ener_a(index_a)+energy/3._dp
                  ener_a(index_b) = ener_a(index_b)+energy/3._dp
                  ener_a(index_c) = ener_a(index_c)+energy/3._dp
               END IF
               particle_set(index_a)%f(1) = particle_set(index_a)%f(1)+g1(1)*fscalar
               particle_set(index_a)%f(2) = particle_set(index_a)%f(2)+g1(2)*fscalar
               particle_set(index_a)%f(3) = particle_set(index_a)%f(3)+g1(3)*fscalar
               particle_set(index_b)%f(1) = particle_set(index_b)%f(1)+g2(1)*fscalar
               particle_set(index_b)%f(2) = particle_set(index_b)%f(2)+g2(2)*fscalar
               particle_set(index_b)%f(3) = particle_set(index_b)%f(3)+g2(3)*fscalar
               particle_set(index_c)%f(1) = particle_set(index_c)%f(1)+g3(1)*fscalar
               particle_set(index_c)%f(2) = particle_set(index_c)%f(2)+g3(2)*fscalar
               particle_set(index_c)%f(3) = particle_set(index_c)%f(3)+g3(3)*fscalar

               ! computing the pressure tensor
               k1 = fscalar*g1
               k3 = fscalar*g3
               IF (use_virial) CALL get_pv_bend(k1, k3, b12, b32, pv_bend)
               IF (atstress) THEN
                  k1 = fscalar*g1/3._dp
                  k3 = fscalar*g3/3._dp
                  CALL get_pv_bend(k1, k3, b12, b32, pv_a(:, :, index_a))
                  CALL get_pv_bend(k1, k3, b12, b32, pv_a(:, :, index_b))
                  CALL get_pv_bend(k1, k3, b12, b32, pv_a(:, :, index_c))
               END IF

               ! the contribution from the bends. ONLY FOR DEBUG
               IF (PRESENT(f_bend)) THEN
                  f_bend(:, index_a) = f_bend(:, index_a)+fscalar*g1
                  f_bend(:, index_b) = f_bend(:, index_b)+fscalar*g2
                  f_bend(:, index_c) = f_bend(:, index_c)+fscalar*g3
               END IF
            END DO BEND

            TORSION: DO itorsion = 1, ntorsions
               index_a = torsion_list(itorsion)%a+first_atom-1
               index_b = torsion_list(itorsion)%b+first_atom-1
               index_c = torsion_list(itorsion)%c+first_atom-1
               index_d = torsion_list(itorsion)%d+first_atom-1
               t12 = particle_set(index_a)%r-particle_set(index_b)%r
               t32 = particle_set(index_c)%r-particle_set(index_b)%r
               t34 = particle_set(index_c)%r-particle_set(index_d)%r
               t43 = particle_set(index_d)%r-particle_set(index_c)%r
               t12 = pbc(t12, cell)
               t32 = pbc(t32, cell)
               t34 = pbc(t34, cell)
               t43 = pbc(t43, cell)
               ! t12 x t32
               tm(1) = t12(2)*t32(3)-t32(2)*t12(3)
               tm(2) = -t12(1)*t32(3)+t32(1)*t12(3)
               tm(3) = t12(1)*t32(2)-t32(1)*t12(2)
               ! t32 x t34
               tn(1) = t32(2)*t34(3)-t34(2)*t32(3)
               tn(2) = -t32(1)*t34(3)+t34(1)*t32(3)
               tn(3) = t32(1)*t34(2)-t34(1)*t32(2)
               sm = SQRT(DOT_PRODUCT(tm, tm))
               ism = 1.0_dp/sm
               sn = SQRT(DOT_PRODUCT(tn, tn))
               isn = 1.0_dp/sn
               s32 = SQRT(DOT_PRODUCT(t32, t32))
               is32 = 1.0_dp/s32
               dist1 = DOT_PRODUCT(t12, t32)
               dist2 = DOT_PRODUCT(t34, t32)
               DO imul = 1, torsion_list(itorsion)%torsion_kind%nmul
                  CALL force_torsions(torsion_list(itorsion)%torsion_kind%id_type, &
                                      s32, is32, ism, isn, dist1, dist2, tm, tn, t12, &
                                      torsion_list(itorsion)%torsion_kind%k(imul), &
                                      torsion_list(itorsion)%torsion_kind%phi0(imul), &
                                      torsion_list(itorsion)%torsion_kind%m(imul), &
                                      gt1, gt2, gt3, gt4, energy, fscalar)
                  pot_torsion = pot_torsion+energy
                  IF (atener) THEN
                     ener_a(index_a) = ener_a(index_a)+energy*0.25_dp
                     ener_a(index_b) = ener_a(index_b)+energy*0.25_dp
                     ener_a(index_c) = ener_a(index_c)+energy*0.25_dp
                     ener_a(index_d) = ener_a(index_d)+energy*0.25_dp
                  END IF
                  particle_set(index_a)%f(1) = particle_set(index_a)%f(1)+gt1(1)*fscalar
                  particle_set(index_a)%f(2) = particle_set(index_a)%f(2)+gt1(2)*fscalar
                  particle_set(index_a)%f(3) = particle_set(index_a)%f(3)+gt1(3)*fscalar
                  particle_set(index_b)%f(1) = particle_set(index_b)%f(1)+gt2(1)*fscalar
                  particle_set(index_b)%f(2) = particle_set(index_b)%f(2)+gt2(2)*fscalar
                  particle_set(index_b)%f(3) = particle_set(index_b)%f(3)+gt2(3)*fscalar
                  particle_set(index_c)%f(1) = particle_set(index_c)%f(1)+gt3(1)*fscalar
                  particle_set(index_c)%f(2) = particle_set(index_c)%f(2)+gt3(2)*fscalar
                  particle_set(index_c)%f(3) = particle_set(index_c)%f(3)+gt3(3)*fscalar
                  particle_set(index_d)%f(1) = particle_set(index_d)%f(1)+gt4(1)*fscalar
                  particle_set(index_d)%f(2) = particle_set(index_d)%f(2)+gt4(2)*fscalar
                  particle_set(index_d)%f(3) = particle_set(index_d)%f(3)+gt4(3)*fscalar

                  ! computing the pressure tensor
                  k1 = fscalar*gt1
                  k3 = fscalar*gt3
                  k4 = fscalar*gt4
                  IF (use_virial) CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_torsion)
                  IF (atstress) THEN
                     k1 = 0.25_dp*fscalar*gt1
                     k3 = 0.25_dp*fscalar*gt3
                     k4 = 0.25_dp*fscalar*gt4
                     CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_a))
                     CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_b))
                     CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_c))
                     CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_d))
                  END IF

                  ! the contribution from the torsions. ONLY FOR DEBUG
                  IF (PRESENT(f_torsion)) THEN
                     f_torsion(:, index_a) = f_torsion(:, index_a)+fscalar*gt1
                     f_torsion(:, index_b) = f_torsion(:, index_b)+fscalar*gt2
                     f_torsion(:, index_c) = f_torsion(:, index_c)+fscalar*gt3
                     f_torsion(:, index_d) = f_torsion(:, index_d)+fscalar*gt4
                  END IF
               END DO
            END DO TORSION

            IMP_TORSION: DO itorsion = 1, nimptors
               index_a = impr_list(itorsion)%a+first_atom-1
               index_b = impr_list(itorsion)%b+first_atom-1
               index_c = impr_list(itorsion)%c+first_atom-1
               index_d = impr_list(itorsion)%d+first_atom-1
               t12 = particle_set(index_a)%r-particle_set(index_b)%r
               t32 = particle_set(index_c)%r-particle_set(index_b)%r
               t34 = particle_set(index_c)%r-particle_set(index_d)%r
               t43 = particle_set(index_d)%r-particle_set(index_c)%r
               t12 = pbc(t12, cell)
               t32 = pbc(t32, cell)
               t34 = pbc(t34, cell)
               t43 = pbc(t43, cell)
               ! t12 x t32
               tm(1) = t12(2)*t32(3)-t32(2)*t12(3)
               tm(2) = -t12(1)*t32(3)+t32(1)*t12(3)
               tm(3) = t12(1)*t32(2)-t32(1)*t12(2)
               ! t32 x t34
               tn(1) = t32(2)*t34(3)-t34(2)*t32(3)
               tn(2) = -t32(1)*t34(3)+t34(1)*t32(3)
               tn(3) = t32(1)*t34(2)-t34(1)*t32(2)
               sm = SQRT(DOT_PRODUCT(tm, tm))
               ism = 1.0_dp/sm
               sn = SQRT(DOT_PRODUCT(tn, tn))
               isn = 1.0_dp/sn
               s32 = SQRT(DOT_PRODUCT(t32, t32))
               is32 = 1.0_dp/s32
               dist1 = DOT_PRODUCT(t12, t32)
               dist2 = DOT_PRODUCT(t34, t32)
               CALL force_imp_torsions(impr_list(itorsion)%impr_kind%id_type, &
                                       s32, is32, ism, isn, dist1, dist2, tm, tn, t12, &
                                       impr_list(itorsion)%impr_kind%k, &
                                       impr_list(itorsion)%impr_kind%phi0, &
                                       gt1, gt2, gt3, gt4, energy, fscalar)
               pot_imp_torsion = pot_imp_torsion+energy
               IF (atener) THEN
                  ener_a(index_a) = ener_a(index_a)+energy*0.25_dp
                  ener_a(index_b) = ener_a(index_b)+energy*0.25_dp
                  ener_a(index_c) = ener_a(index_c)+energy*0.25_dp
                  ener_a(index_d) = ener_a(index_d)+energy*0.25_dp
               END IF
               particle_set(index_a)%f(1) = particle_set(index_a)%f(1)+gt1(1)*fscalar
               particle_set(index_a)%f(2) = particle_set(index_a)%f(2)+gt1(2)*fscalar
               particle_set(index_a)%f(3) = particle_set(index_a)%f(3)+gt1(3)*fscalar
               particle_set(index_b)%f(1) = particle_set(index_b)%f(1)+gt2(1)*fscalar
               particle_set(index_b)%f(2) = particle_set(index_b)%f(2)+gt2(2)*fscalar
               particle_set(index_b)%f(3) = particle_set(index_b)%f(3)+gt2(3)*fscalar
               particle_set(index_c)%f(1) = particle_set(index_c)%f(1)+gt3(1)*fscalar
               particle_set(index_c)%f(2) = particle_set(index_c)%f(2)+gt3(2)*fscalar
               particle_set(index_c)%f(3) = particle_set(index_c)%f(3)+gt3(3)*fscalar
               particle_set(index_d)%f(1) = particle_set(index_d)%f(1)+gt4(1)*fscalar
               particle_set(index_d)%f(2) = particle_set(index_d)%f(2)+gt4(2)*fscalar
               particle_set(index_d)%f(3) = particle_set(index_d)%f(3)+gt4(3)*fscalar

               ! computing the pressure tensor
               k1 = fscalar*gt1
               k3 = fscalar*gt3
               k4 = fscalar*gt4
               IF (use_virial) CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_imp_torsion)
               IF (atstress) THEN
                  k1 = 0.25_dp*fscalar*gt1
                  k3 = 0.25_dp*fscalar*gt3
                  k4 = 0.25_dp*fscalar*gt4
                  CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_a))
                  CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_b))
                  CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_c))
                  CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_d))
               END IF

               ! the contribution from the torsions. ONLY FOR DEBUG
               IF (PRESENT(f_imptor)) THEN
                  f_imptor(:, index_a) = f_imptor(:, index_a)+fscalar*gt1
                  f_imptor(:, index_b) = f_imptor(:, index_b)+fscalar*gt2
                  f_imptor(:, index_c) = f_imptor(:, index_c)+fscalar*gt3
                  f_imptor(:, index_d) = f_imptor(:, index_d)+fscalar*gt4
               END IF
            END DO IMP_TORSION

            OPBEND: DO iopbend = 1, nopbends
               index_a = opbend_list(iopbend)%a+first_atom-1
               index_b = opbend_list(iopbend)%b+first_atom-1
               index_c = opbend_list(iopbend)%c+first_atom-1
               index_d = opbend_list(iopbend)%d+first_atom-1

               t12 = particle_set(index_a)%r-particle_set(index_b)%r
               t32 = particle_set(index_c)%r-particle_set(index_b)%r
               t34 = particle_set(index_c)%r-particle_set(index_d)%r
               t43 = particle_set(index_d)%r-particle_set(index_c)%r
               t41 = particle_set(index_d)%r-particle_set(index_a)%r
               t42 = pbc(t41+t12, cell)
               t12 = pbc(t12, cell)
               t32 = pbc(t32, cell)
               t41 = pbc(t41, cell)
               t43 = pbc(t43, cell)
               ! tm = t32 x t12
               tm(1) = t32(2)*t12(3)-t12(2)*t32(3)
               tm(2) = -t32(1)*t12(3)+t12(1)*t32(3)
               tm(3) = t32(1)*t12(2)-t12(1)*t32(2)
               sm = SQRT(DOT_PRODUCT(tm, tm))
               s32 = SQRT(DOT_PRODUCT(t32, t32))
               CALL force_opbends(opbend_list(iopbend)%opbend_kind%id_type, &
                                  s32, tm, t41, t42, t43, &
                                  opbend_list(iopbend)%opbend_kind%k, &
                                  opbend_list(iopbend)%opbend_kind%phi0, &
                                  gt1, gt2, gt3, gt4, energy, fscalar)
               pot_opbend = pot_opbend+energy
               IF (atener) THEN
                  ener_a(index_a) = ener_a(index_a)+energy*0.25_dp
                  ener_a(index_b) = ener_a(index_b)+energy*0.25_dp
                  ener_a(index_c) = ener_a(index_c)+energy*0.25_dp
                  ener_a(index_d) = ener_a(index_d)+energy*0.25_dp
               END IF
               particle_set(index_a)%f(1) = particle_set(index_a)%f(1)+gt1(1)*fscalar
               particle_set(index_a)%f(2) = particle_set(index_a)%f(2)+gt1(2)*fscalar
               particle_set(index_a)%f(3) = particle_set(index_a)%f(3)+gt1(3)*fscalar
               particle_set(index_b)%f(1) = particle_set(index_b)%f(1)+gt2(1)*fscalar
               particle_set(index_b)%f(2) = particle_set(index_b)%f(2)+gt2(2)*fscalar
               particle_set(index_b)%f(3) = particle_set(index_b)%f(3)+gt2(3)*fscalar
               particle_set(index_c)%f(1) = particle_set(index_c)%f(1)+gt3(1)*fscalar
               particle_set(index_c)%f(2) = particle_set(index_c)%f(2)+gt3(2)*fscalar
               particle_set(index_c)%f(3) = particle_set(index_c)%f(3)+gt3(3)*fscalar
               particle_set(index_d)%f(1) = particle_set(index_d)%f(1)+gt4(1)*fscalar
               particle_set(index_d)%f(2) = particle_set(index_d)%f(2)+gt4(2)*fscalar
               particle_set(index_d)%f(3) = particle_set(index_d)%f(3)+gt4(3)*fscalar

               ! computing the pressure tensor
               k1 = fscalar*gt1
               k3 = fscalar*gt3
               k4 = fscalar*gt4

               IF (use_virial) CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_opbend)
               IF (atstress) THEN
                  k1 = 0.25_dp*fscalar*gt1
                  k3 = 0.25_dp*fscalar*gt3
                  k4 = 0.25_dp*fscalar*gt4
                  CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_a))
                  CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_b))
                  CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_c))
                  CALL get_pv_torsion(k1, k3, k4, t12, t32, t43, pv_a(:, :, index_d))
               END IF

               ! the contribution from the opbends. ONLY FOR DEBUG
               IF (PRESENT(f_opbend)) THEN
                  f_opbend(:, index_a) = f_opbend(:, index_a)+fscalar*gt1
                  f_opbend(:, index_b) = f_opbend(:, index_b)+fscalar*gt2
                  f_opbend(:, index_c) = f_opbend(:, index_c)+fscalar*gt3
                  f_opbend(:, index_d) = f_opbend(:, index_d)+fscalar*gt4
               END IF
            END DO OPBEND
         END DO
      END DO MOL

      CALL timestop(handle)

   END SUBROUTINE force_intra_control

END MODULE fist_intra_force

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (15-Mar-2001) : New routine ewald_setup (former pme_setup)
!>      JGH (23-Mar-2001) : Get rid of global variable ewald_grp
! **************************************************************************************************
MODULE ewalds

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE bibliography,                    ONLY: Ewald1921,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type
   USE dg_rho0_types,                   ONLY: dg_rho0_type
   USE dg_types,                        ONLY: dg_get,&
                                              dg_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_get,&
                                              ewald_pw_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              oorootpi,&
                                              pi
   USE mathlib,                         ONLY: matvec_3x3
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_poisson_types,                ONLY: do_ewald_none
   USE pw_pool_types,                   ONLY: pw_pool_type
   USE shell_potential_types,           ONLY: get_shell,&
                                              shell_kind_type
   USE structure_factor_types,          ONLY: structure_factor_type
   USE structure_factors,               ONLY: structure_factor_allocate,&
                                              structure_factor_deallocate,&
                                              structure_factor_evaluate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ewalds'

   PRIVATE
   PUBLIC :: ewald_evaluate, ewald_self, ewald_self_atom, ewald_print

CONTAINS

! **************************************************************************************************
!> \brief computes the potential and the force from the g-space part of
!>      the 1/r potential
!>      Ref.: J.-P. Hansen, Enrico Fermi School, 1985
!>      Note: Only the positive G-vectors are used in the sum.
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param cell ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param fg_coulomb ...
!> \param vg_coulomb ...
!> \param pv_g ...
!> \param use_virial ...
!> \param charges ...
!> \param e_coulomb ...
!> \param pv_coulomb ...
!> \par History
!>      JGH (21-Feb-2001) : changed name
!> \author CJM
! **************************************************************************************************
   SUBROUTINE ewald_evaluate(ewald_env, ewald_pw, cell, atomic_kind_set, particle_set, &
                             local_particles, fg_coulomb, vg_coulomb, pv_g, use_virial, charges, e_coulomb, &
                             pv_coulomb)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(cell_type), POINTER                           :: cell
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fg_coulomb
      REAL(KIND=dp), INTENT(OUT)                         :: vg_coulomb
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: pv_g
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges, e_coulomb
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: pv_coulomb

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ewald_evaluate', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: snode
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: summe
      INTEGER :: gpt, group, handle, iparticle, iparticle_kind, iparticle_local, lp, mp, nnodes, &
         node, np, nparticle_kind, nparticle_local
      INTEGER, DIMENSION(:, :), POINTER                  :: bds
      LOGICAL                                            :: atenergy, atstress, use_charge_array
      REAL(KIND=dp)                                      :: alpha, denom, e_igdotr, factor, &
                                                            four_alpha_sq, gauss, pref, q
      REAL(KIND=dp), DIMENSION(3)                        :: vec
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charge
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: rho0
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(dg_type), POINTER                             :: dg
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(structure_factor_type)                        :: exp_igr

      CALL timeset(routineN, handle)
      CALL cite_reference(Ewald1921)
      use_charge_array = PRESENT(charges)
      IF (use_charge_array) use_charge_array = ASSOCIATED(charges)
      atenergy = PRESENT(e_coulomb)
      IF (atenergy) atenergy = ASSOCIATED(e_coulomb)
      IF (atenergy) e_coulomb = 0._dp

      atstress = PRESENT(pv_coulomb)
      IF (atstress) atstress = ASSOCIATED(pv_coulomb)
      IF (atstress) pv_coulomb = 0._dp

      ! pointing
      CALL ewald_env_get(ewald_env, alpha=alpha, group=group)
      CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_pool, dg=dg)
      CALL dg_get(dg, dg_rho0=dg_rho0)
      rho0 => dg_rho0%density%pw%cr3d
      pw_grid => pw_pool%pw_grid
      bds => pw_grid%bounds

      ! allocating
      nparticle_kind = SIZE(atomic_kind_set)
      nnodes = 0
      DO iparticle_kind = 1, nparticle_kind
         nnodes = nnodes+local_particles%n_el(iparticle_kind)
      ENDDO

      CALL structure_factor_allocate(pw_grid%bounds, nnodes, exp_igr)

      ALLOCATE (summe(1:pw_grid%ngpts_cut))
      ALLOCATE (charge(1:nnodes))

      ! Initializing vg_coulomb and fg_coulomb
      vg_coulomb = 0.0_dp
      fg_coulomb = 0.0_dp
      IF (use_virial) pv_g = 0.0_dp
      ! defining four_alpha_sq
      four_alpha_sq = 4.0_dp*alpha**2
      ! zero node count
      node = 0
      DO iparticle_kind = 1, nparticle_kind
         nparticle_local = local_particles%n_el(iparticle_kind)
         IF (use_charge_array) THEN
            DO iparticle_local = 1, nparticle_local
               node = node+1
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               charge(node) = charges(iparticle)
               CALL matvec_3x3(vec, cell%h_inv, particle_set(iparticle)%r)
               CALL structure_factor_evaluate(vec, exp_igr%lb, &
                                              exp_igr%ex(:, node), exp_igr%ey(:, node), exp_igr%ez(:, node))
            END DO
         ELSE
            atomic_kind => atomic_kind_set(iparticle_kind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q)
            DO iparticle_local = 1, nparticle_local
               node = node+1
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               charge(node) = q
               CALL matvec_3x3(vec, cell%h_inv, particle_set(iparticle)%r)
               CALL structure_factor_evaluate(vec, exp_igr%lb, &
                                              exp_igr%ex(:, node), exp_igr%ey(:, node), exp_igr%ez(:, node))
            END DO
         END IF
      END DO

      summe(:) = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      ! looping over the positive g-vectors
      DO gpt = 1, pw_grid%ngpts_cut_local

         lp = pw_grid%mapl%pos(pw_grid%g_hat(1, gpt))
         mp = pw_grid%mapm%pos(pw_grid%g_hat(2, gpt))
         np = pw_grid%mapn%pos(pw_grid%g_hat(3, gpt))

         lp = lp+bds(1, 1)
         mp = mp+bds(1, 2)
         np = np+bds(1, 3)

         ! initializing sum to be used in the energy and force
         DO node = 1, nnodes
            summe(gpt) = summe(gpt)+charge(node)* &
                         (exp_igr%ex(lp, node) &
                          *exp_igr%ey(mp, node) &
                          *exp_igr%ez(np, node))
         END DO
      END DO
      CALL mp_sum(summe, group)

      pref = fourpi/pw_grid%vol

      ! looping over the positive g-vectors
      DO gpt = 1, pw_grid%ngpts_cut_local
         ! computing the potential energy
         lp = pw_grid%mapl%pos(pw_grid%g_hat(1, gpt))
         mp = pw_grid%mapm%pos(pw_grid%g_hat(2, gpt))
         np = pw_grid%mapn%pos(pw_grid%g_hat(3, gpt))

         lp = lp+bds(1, 1)
         mp = mp+bds(1, 2)
         np = np+bds(1, 3)

         IF (pw_grid%gsq(gpt) <= 1.0E-10_dp) CYCLE

         gauss = (rho0(lp, mp, np)*pw_grid%vol)**2/pw_grid%gsq(gpt)
         factor = gauss*REAL(summe(gpt)*CONJG(summe(gpt)), KIND=dp)
         vg_coulomb = vg_coulomb+factor

         ! atomic energies
         IF (atenergy) THEN
            DO node = 1, nnodes
               snode = CONJG(exp_igr%ex(lp, node) &
                             *exp_igr%ey(mp, node) &
                             *exp_igr%ez(np, node))
               e_coulomb(node) = e_coulomb(node)+gauss*charge(node)*REAL(summe(gpt)*snode, KIND=dp)
            END DO
         END IF

         ! computing the force
         node = 0
         DO node = 1, nnodes
            e_igdotr = AIMAG(summe(gpt)*CONJG &
                             (exp_igr%ex(lp, node) &
                              *exp_igr%ey(mp, node) &
                              *exp_igr%ez(np, node)))
            fg_coulomb(:, node) = fg_coulomb(:, node) &
                                  +charge(node)*gauss*e_igdotr*pw_grid%g(:, gpt)
         END DO

         ! compute the virial P*V
         denom = 1.0_dp/four_alpha_sq+1.0_dp/pw_grid%gsq(gpt)
         IF (use_virial) THEN
            pv_g(1, 1) = pv_g(1, 1)+factor*(1.0_dp-2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(1, gpt)*denom)
            pv_g(1, 2) = pv_g(1, 2)-factor*(2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(2, gpt)*denom)
            pv_g(1, 3) = pv_g(1, 3)-factor*(2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(3, gpt)*denom)
            pv_g(2, 1) = pv_g(2, 1)-factor*(2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(1, gpt)*denom)
            pv_g(2, 2) = pv_g(2, 2)+factor*(1.0_dp-2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(2, gpt)*denom)
            pv_g(2, 3) = pv_g(2, 3)-factor*(2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(3, gpt)*denom)
            pv_g(3, 1) = pv_g(3, 1)-factor*(2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(1, gpt)*denom)
            pv_g(3, 2) = pv_g(3, 2)-factor*(2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(2, gpt)*denom)
            pv_g(3, 3) = pv_g(3, 3)+factor*(1.0_dp-2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(3, gpt)*denom)
         END IF
         IF (atstress) THEN
            DO node = 1, nnodes
               snode = CONJG(exp_igr%ex(lp, node) &
                             *exp_igr%ey(mp, node) &
                             *exp_igr%ez(np, node))
               factor = gauss*charge(node)*REAL(summe(gpt)*snode, KIND=dp)
               pv_coulomb(1, 1, node) = pv_coulomb(1, 1, node)+factor*(1.0_dp-2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(1, gpt)*denom)
               pv_coulomb(1, 2, node) = pv_coulomb(1, 2, node)-factor*(2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(2, gpt)*denom)
               pv_coulomb(1, 3, node) = pv_coulomb(1, 3, node)-factor*(2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(3, gpt)*denom)
               pv_coulomb(2, 1, node) = pv_coulomb(2, 1, node)-factor*(2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(1, gpt)*denom)
               pv_coulomb(2, 2, node) = pv_coulomb(2, 2, node)+factor*(1.0_dp-2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(2, gpt)*denom)
               pv_coulomb(2, 3, node) = pv_coulomb(2, 3, node)-factor*(2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(3, gpt)*denom)
               pv_coulomb(3, 1, node) = pv_coulomb(3, 1, node)-factor*(2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(1, gpt)*denom)
               pv_coulomb(3, 2, node) = pv_coulomb(3, 2, node)-factor*(2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(2, gpt)*denom)
               pv_coulomb(3, 3, node) = pv_coulomb(3, 3, node)+factor*(1.0_dp-2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(3, gpt)*denom)
            END DO
         END IF
      END DO

      vg_coulomb = vg_coulomb*pref
      IF (use_virial) pv_g = pv_g*pref
      IF (atenergy) e_coulomb = e_coulomb*pref
      IF (atstress) pv_coulomb = pv_coulomb*pref

      fg_coulomb = fg_coulomb*(2.0_dp*pref)

      CALL structure_factor_deallocate(exp_igr)

      DEALLOCATE (charge, summe)

      CALL timestop(handle)

   END SUBROUTINE ewald_evaluate

! **************************************************************************************************
!> \brief Computes the self interaction from g-space
!>      and the neutralizing background
!> \param ewald_env ...
!> \param cell ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param e_self ...
!> \param e_neut ...
!> \param charges ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE ewald_self(ewald_env, cell, atomic_kind_set, local_particles, e_self, &
                         e_neut, charges)

      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(cell_type), POINTER                           :: cell
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), INTENT(OUT)                         :: e_self, e_neut
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ewald_self', routineP = moduleN//':'//routineN

      INTEGER                                            :: ewald_type, group, iparticle_kind, &
                                                            nparticle_kind, nparticle_local
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: alpha, mm_radius, q, q_neutg, q_self, &
                                                            q_sum, qcore, qshell
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(shell_kind_type), POINTER                     :: shell

      CALL ewald_env_get(ewald_env, ewald_type=ewald_type, &
                         alpha=alpha, group=group)
      q_neutg = 0.0_dp
      q_self = 0.0_dp
      q_sum = 0.0_dp
      nparticle_kind = SIZE(atomic_kind_set)
      IF (ASSOCIATED(charges)) THEN
         q_self = DOT_PRODUCT(charges, charges)
         q_sum = SUM(charges)
         ! check and abort..
         DO iparticle_kind = 1, nparticle_kind
            atomic_kind => atomic_kind_set(iparticle_kind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, mm_radius=mm_radius)
            IF (mm_radius > 0.0_dp) THEN
               CALL cp__b("ewalds.F",334,"Array of charges not implemented for mm_radius>0.0 !!")
            END IF
         END DO
      ELSE
         DO iparticle_kind = 1, nparticle_kind
            atomic_kind => atomic_kind_set(iparticle_kind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, mm_radius=mm_radius, &
                                 qeff=q, shell_active=is_shell, shell=shell)
            nparticle_local = local_particles%n_el(iparticle_kind)
            IF (is_shell) THEN
               CALL get_shell(shell=shell, charge_core=qcore, charge_shell=qshell)
               ! MI: the core-shell ES interaction, when core and shell belong to the same ion, is excluded
               !     in the nonbond correction term. Therefore, here the self interaction is computed entirely
               q_self = q_self+qcore*qcore*nparticle_local+qshell*qshell*nparticle_local
               q_sum = q_sum+qcore*nparticle_local+qshell*nparticle_local
               IF (mm_radius > 0) THEN
                  ! the core is always a point charge
                  q_neutg = q_neutg+2.0_dp*qshell*mm_radius**2
               END IF
            ELSE
               q_self = q_self+q*q*nparticle_local
               q_sum = q_sum+q*nparticle_local
               IF (mm_radius > 0) THEN
                  q_neutg = q_neutg+2.0_dp*q*mm_radius**2
               END IF
            END IF
         END DO

         CALL mp_sum(q_self, group)
         CALL mp_sum(q_sum, group)
      END IF

      e_neut = 0.0_dp
      e_self = 0.0_dp
      IF (ewald_type /= do_ewald_none) THEN
         e_self = -q_self*alpha*oorootpi
         e_neut = -q_sum*pi/(2.0_dp*cell%deth)*(q_sum/alpha**2-q_neutg)
      END IF

   END SUBROUTINE ewald_self

! **************************************************************************************************
!> \brief Computes the self interaction per atom
!> \param ewald_env ...
!> \param atomic_kind_set ...
!> \param local_particles ...
!> \param e_self ...
!> \param charges ...
!> \par History
!>      none
!> \author JHU from ewald_self
! **************************************************************************************************
   SUBROUTINE ewald_self_atom(ewald_env, atomic_kind_set, local_particles, e_self, &
                              charges)

      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: e_self(:)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ewald_self_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ewald_type, ii, iparticle_kind, &
                                                            iparticle_local, nparticle_kind, &
                                                            nparticle_local
      LOGICAL                                            :: is_shell
      REAL(KIND=dp)                                      :: alpha, fself, q, qcore, qshell
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(shell_kind_type), POINTER                     :: shell

      CALL ewald_env_get(ewald_env, ewald_type=ewald_type, alpha=alpha)

      fself = alpha*oorootpi

      IF (ewald_type /= do_ewald_none) THEN
         nparticle_kind = SIZE(atomic_kind_set)
         IF (ASSOCIATED(charges)) THEN
            CALL cp__b("ewalds.F",413,"Atomic energy not implemented for charges")
         ELSE
            DO iparticle_kind = 1, nparticle_kind
               atomic_kind => atomic_kind_set(iparticle_kind)
               nparticle_local = local_particles%n_el(iparticle_kind)
               CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q, &
                                    shell_active=is_shell, shell=shell)
               IF (is_shell) THEN
                  CALL get_shell(shell=shell, charge_core=qcore, charge_shell=qshell)
                  DO iparticle_local = 1, nparticle_local
                     ii = local_particles%list(iparticle_kind)%array(iparticle_local)
                     e_self(ii) = e_self(ii)-(qcore*qcore+qshell*qshell)*fself
                  END DO
               ELSE
                  DO iparticle_local = 1, nparticle_local
                     ii = local_particles%list(iparticle_kind)%array(iparticle_local)
                     e_self(ii) = e_self(ii)-q*q*fself
                  END DO
               END IF
            END DO
         END IF
      END IF

   END SUBROUTINE ewald_self_atom

! **************************************************************************************************
!> \brief ...
!> \param iw ...
!> \param pot_nonbond ...
!> \param e_gspace ...
!> \param e_self ...
!> \param e_neut ...
!> \param e_bonded ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE ewald_print(iw, pot_nonbond, e_gspace, e_self, e_neut, e_bonded)

      INTEGER, INTENT(IN)                                :: iw
      REAL(KIND=dp), INTENT(IN)                          :: pot_nonbond, e_gspace, e_self, e_neut, &
                                                            e_bonded

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ewald_print', routineP = moduleN//':'//routineN

      IF (iw > 0) THEN
         WRITE (iw, '( A, A )') ' *********************************', &
            '**********************************************'
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' INITIAL GSPACE ENERGY', &
            '[hartree]', '= ', e_gspace
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' INITIAL NONBONDED ENERGY', &
            '[hartree]', '= ', pot_nonbond
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' SELF ENERGY CORRECTION', &
            '[hartree]', '= ', e_self
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' NEUT. BACKGROUND', &
            '[hartree]', '= ', e_neut
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' BONDED CORRECTION', &
            '[hartree]', '= ', e_bonded
         WRITE (iw, '( A, A )') ' *********************************', &
            '**********************************************'
      END IF
   END SUBROUTINE ewald_print

END MODULE ewalds
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (21-Mar-2001) : Complete rewrite
!> \author CJM and APSI
! **************************************************************************************************
MODULE pme

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE atprop_types,                    ONLY: atprop_type
   USE bibliography,                    ONLY: cite_reference,&
                                              darden1993
   USE cell_types,                      ONLY: cell_type
   USE dg_rho0_types,                   ONLY: dg_rho0_type
   USE dg_types,                        ONLY: dg_get,&
                                              dg_type
   USE dgs,                             ONLY: dg_get_patch,&
                                              dg_get_strucfac,&
                                              dg_sum_patch,&
                                              dg_sum_patch_force_1d,&
                                              dg_sum_patch_force_3d
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_get,&
                                              ewald_pw_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi
   USE particle_types,                  ONLY: particle_type
   USE pme_tools,                       ONLY: get_center,&
                                              set_list
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_copy,&
                                              pw_derive,&
                                              pw_integral_a2b,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE realspace_grid_types,            ONLY: &
        pw2rs, realspace_grid_desc_type, realspace_grid_p_type, realspace_grid_type, rs2pw, &
        rs_grid_create, rs_grid_release, rs_grid_set_box, rs_grid_zero, rs_pw_transfer
   USE shell_potential_types,           ONLY: shell_kind_type
   USE structure_factor_types,          ONLY: structure_factor_type
   USE structure_factors,               ONLY: structure_factor_allocate,&
                                              structure_factor_deallocate,&
                                              structure_factor_init
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: pme_evaluate
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pme'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param box ...
!> \param particle_set ...
!> \param vg_coulomb ...
!> \param fg_coulomb ...
!> \param pv_g ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param fgshell_coulomb ...
!> \param fgcore_coulomb ...
!> \param use_virial ...
!> \param charges ...
!> \param atprop ...
!> \par History
!>      JGH (15-Mar-2001) : New electrostatic calculation and pressure tensor
!>      JGH (21-Mar-2001) : Complete rewrite
!>      JGH (21-Mar-2001) : Introduced real space density type for future
!>                          parallelisation
!> \author CJM and APSI
! **************************************************************************************************
   SUBROUTINE pme_evaluate(ewald_env, ewald_pw, box, particle_set, vg_coulomb, &
                           fg_coulomb, pv_g, shell_particle_set, core_particle_set, &
                           fgshell_coulomb, fgcore_coulomb, use_virial, charges, atprop)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(cell_type), POINTER                           :: box
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), INTENT(OUT)                         :: vg_coulomb
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT), &
         OPTIONAL                                        :: fg_coulomb, pv_g
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: shell_particle_set, core_particle_set
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT), &
         OPTIONAL                                        :: fgshell_coulomb, fgcore_coulomb
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges
      TYPE(atprop_type), POINTER                         :: atprop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pme_evaluate', routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, i, ig, ipart, j, nd(3), &
                                                            npart, nshell, p1, p2
      LOGICAL                                            :: is1_core, is2_core
      REAL(KIND=dp)                                      :: alpha, dvols, fat1, ffa, ffb
      REAL(KIND=dp), DIMENSION(3)                        :: fat
      REAL(KIND=dp), DIMENSION(3, 3)                     :: f_stress, h_stress
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(dg_type), POINTER                             :: dg
      TYPE(pw_grid_type), POINTER                        :: grid_b, grid_s
      TYPE(pw_p_type)                                    :: rhos1, rhos2
      TYPE(pw_p_type), DIMENSION(3)                      :: dphi_g
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: pw_big_pool, pw_small_pool
      TYPE(pw_type), POINTER                             :: phi_g, phi_r, rhob_g, rhob_r
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: drpot
      TYPE(realspace_grid_type), POINTER                 :: rden, rpot
      TYPE(structure_factor_type)                        :: exp_igr

      CALL timeset(routineN, handle)
      NULLIFY (poisson_env, rden, drpot)
      CALL cite_reference(Darden1993)
      CALL ewald_env_get(ewald_env, alpha=alpha, group=group)
      CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_big_pool, &
                        pw_small_pool=pw_small_pool, rs_desc=rs_desc, &
                        poisson_env=poisson_env, dg=dg)

      grid_b => pw_big_pool%pw_grid
      grid_s => pw_small_pool%pw_grid

      CALL dg_get(dg, dg_rho0=dg_rho0)

      npart = SIZE(particle_set)

      CALL structure_factor_init(exp_igr)

      IF (PRESENT(shell_particle_set)) THEN
         IF(.NOT.(ASSOCIATED(shell_particle_set)))CALL cp__a("pme.F",150)
         IF(.NOT.(ASSOCIATED(core_particle_set)))CALL cp__a("pme.F",151)
         nshell = SIZE(shell_particle_set)
         CALL structure_factor_allocate(grid_s%bounds, npart, exp_igr, &
                                        allocate_centre=.TRUE., allocate_shell_e=.TRUE., &
                                        allocate_shell_centre=.TRUE., nshell=nshell)

      ELSE
         CALL structure_factor_allocate(grid_s%bounds, npart, exp_igr, &
                                        allocate_centre=.TRUE.)
      END IF

      CALL pw_pool_create_pw(pw_small_pool, rhos1%pw, &
                             use_data=REALDATA3D)
      CALL pw_pool_create_pw(pw_small_pool, rhos2%pw, &
                             use_data=REALDATA3D)

      CALL rs_grid_create(rden, rs_desc)
      CALL rs_grid_set_box(grid_b, rs=rden)
      CALL rs_grid_zero(rden)

      IF (rden%desc%parallel .AND. rden%desc%distributed) THEN
         CALL get_center(particle_set, box, exp_igr%centre, grid_b%npts, 1)
      END IF
      IF (PRESENT(shell_particle_set) .AND. rden%desc%parallel .AND. rden%desc%distributed) THEN
         CALL get_center(shell_particle_set, box, exp_igr%shell_centre, grid_b%npts, 1)
         CALL get_center(core_particle_set, box, exp_igr%core_centre, grid_b%npts, 1)
      END IF

      !-------------- DENSITY CALCULATION ----------------

      ipart = 0
      DO

         CALL set_list(particle_set, npart, exp_igr%centre, p1, rden, ipart, exp_igr%core_centre)
         CALL set_list(particle_set, npart, exp_igr%centre, p2, rden, ipart, exp_igr%core_centre)
         IF (p1 == 0 .AND. p2 == 0) EXIT

         is1_core = (particle_set(p1)%shell_index /= 0)
         IF (p2 /= 0) THEN
            is2_core = (particle_set(p2)%shell_index /= 0)
         ELSE
            is2_core = .FALSE.
         END IF

         ! calculate function on small boxes (we use double packing in FFT)
         IF (is1_core .OR. is2_core) THEN
            CALL get_patch(dg, particle_set, exp_igr, box, p1, p2, grid_b, grid_s, &
                           rhos1, rhos2, is1_core=is1_core, is2_core=is2_core, &
                           core_particle_set=core_particle_set, charges=charges)

            ! add boxes to real space grid (big box)
            IF (is1_core) THEN
               CALL dg_sum_patch(rden, rhos1, exp_igr%core_centre(:, particle_set(p1)%shell_index))
            ELSE
               CALL dg_sum_patch(rden, rhos1, exp_igr%centre(:, p1))
            END IF
            IF (p2 /= 0 .AND. is2_core) THEN
               CALL dg_sum_patch(rden, rhos2, exp_igr%core_centre(:, particle_set(p2)%shell_index))
            ELSE IF (p2 /= 0) THEN
               CALL dg_sum_patch(rden, rhos2, exp_igr%centre(:, p2))
            END IF
         ELSE
            CALL get_patch(dg, particle_set, exp_igr, box, p1, p2, grid_b, grid_s, &
                           rhos1, rhos2, charges=charges)
            ! add boxes to real space grid (big box)
            CALL dg_sum_patch(rden, rhos1, exp_igr%centre(:, p1))
            IF (p2 /= 0) CALL dg_sum_patch(rden, rhos2, exp_igr%centre(:, p2))
         END IF

      END DO
      IF (PRESENT(shell_particle_set)) THEN
         ipart = 0
         DO
            CALL set_list(shell_particle_set, nshell, exp_igr%shell_centre, p1, rpot, ipart)
            CALL set_list(shell_particle_set, nshell, exp_igr%shell_centre, p2, rpot, ipart)
            IF (p1 == 0 .AND. p2 == 0) EXIT
            ! calculate function on small boxes (we use double packing in FFT)
            CALL get_patch(dg, shell_particle_set, exp_igr, box, p1, p2, grid_b, grid_s, &
                           rhos1, rhos2, is1_shell=.TRUE., is2_shell=.TRUE., charges=charges)
            ! add boxes to real space grid (big box)
            CALL dg_sum_patch(rpot, rhos1, exp_igr%shell_centre(:, p1))
            IF (p2 /= 0) CALL dg_sum_patch(rpot, rhos2, exp_igr%shell_centre(:, p2))
         END DO
      END IF

      CALL pw_pool_create_pw(pw_big_pool, rhob_r, use_data=REALDATA3D, in_space=REALSPACE)
      CALL rs_pw_transfer(rden, rhob_r, rs2pw)

      !-------------- ELECTROSTATIC CALCULATION -----------

      ! allocate intermediate arrays
      DO i = 1, 3
         NULLIFY (dphi_g(i)%pw)
         CALL pw_pool_create_pw(pw_big_pool, dphi_g(i)%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
      END DO
      CALL pw_pool_create_pw(pw_big_pool, phi_r, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)

      CALL pw_poisson_solve(poisson_env, rhob_r, vg_coulomb, phi_r, dphi_g, h_stress)

      ! atomic energies
      IF (atprop%energy .OR. atprop%stress) THEN
         dvols = rhos1%pw%pw_grid%dvol
         CALL rs_grid_create(rpot, rs_desc)
         CALL rs_pw_transfer(rpot, phi_r, pw2rs)
         ipart = 0
         DO
            CALL set_list(particle_set, npart, exp_igr%centre, p1, rden, ipart, exp_igr%core_centre)
            CALL set_list(particle_set, npart, exp_igr%centre, p2, rden, ipart, exp_igr%core_centre)
            IF (p1 == 0 .AND. p2 == 0) EXIT
            ! integrate box and potential
            CALL get_patch(dg, particle_set, exp_igr, box, p1, p2, grid_b, grid_s, &
                           rhos1, rhos2, charges=charges)
            ! add boxes to real space grid (big box)
            CALL dg_sum_patch_force_1d(rpot, rhos1, exp_igr%centre(:, p1), fat1)
            IF (atprop%energy) THEN
               atprop%atener(p1) = atprop%atener(p1)+0.5_dp*fat1*dvols
            END IF
            IF (atprop%stress) THEN
               atprop%atstress(1, 1, p1) = atprop%atstress(1, 1, p1)+0.5_dp*fat1*dvols
               atprop%atstress(2, 2, p1) = atprop%atstress(2, 2, p1)+0.5_dp*fat1*dvols
               atprop%atstress(3, 3, p1) = atprop%atstress(3, 3, p1)+0.5_dp*fat1*dvols
            END IF
            IF (p2 /= 0) THEN
               CALL dg_sum_patch_force_1d(rpot, rhos2, exp_igr%centre(:, p2), fat1)
               IF (atprop%energy) THEN
                  atprop%atener(p2) = atprop%atener(p2)+0.5_dp*fat1*dvols
               END IF
               IF (atprop%stress) THEN
                  atprop%atstress(1, 1, p2) = atprop%atstress(1, 1, p2)+0.5_dp*fat1*dvols
                  atprop%atstress(2, 2, p2) = atprop%atstress(2, 2, p2)+0.5_dp*fat1*dvols
                  atprop%atstress(3, 3, p2) = atprop%atstress(3, 3, p2)+0.5_dp*fat1*dvols
               END IF
            END IF
         END DO
         IF (atprop%stress) THEN
            CALL pw_pool_create_pw(pw_big_pool, phi_g, &
                                   use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
            CALL pw_pool_create_pw(pw_big_pool, rhob_g, &
                                   use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
            ffa = (0.5_dp/dg_rho0%zet(1))**2
            ffb = 1.0_dp/fourpi
            DO i = 1, 3
               DO ig = grid_b%first_gne0, grid_b%ngpts_cut_local
                  phi_g%cc(ig) = ffb*dphi_g(i)%pw%cc(ig)*(ffa*grid_b%gsq(ig)+1.0_dp)
                  phi_g%cc(ig) = phi_g%cc(ig)*poisson_env%green_fft%influence_fn%cc(ig)
               END DO
               IF (grid_b%have_g0) phi_g%cc(1) = 0.0_dp
               DO j = 1, i
                  CALL pw_copy(phi_g, rhob_g)
                  nd = 0
                  nd(j) = 1
                  CALL pw_derive(rhob_g, nd)
                  CALL pw_transfer(rhob_g, rhob_r)
                  CALL rs_pw_transfer(rpot, rhob_r, pw2rs)

                  ipart = 0
                  DO
                     CALL set_list(particle_set, npart, exp_igr%centre, p1, rden, ipart, exp_igr%core_centre)
                     CALL set_list(particle_set, npart, exp_igr%centre, p2, rden, ipart, exp_igr%core_centre)
                     IF (p1 == 0 .AND. p2 == 0) EXIT
                     ! integrate box and potential
                     CALL get_patch(dg, particle_set, exp_igr, box, p1, p2, grid_b, grid_s, &
                                    rhos1, rhos2, charges=charges)
                     ! add boxes to real space grid (big box)
                     CALL dg_sum_patch_force_1d(rpot, rhos1, exp_igr%centre(:, p1), fat1)
                     atprop%atstress(i, j, p1) = atprop%atstress(i, j, p1)+fat1*dvols
                     IF (i /= j) atprop%atstress(j, i, p1) = atprop%atstress(j, i, p1)+fat1*dvols
                     IF (p2 /= 0) THEN
                        CALL dg_sum_patch_force_1d(rpot, rhos2, exp_igr%centre(:, p2), fat1)
                        atprop%atstress(i, j, p2) = atprop%atstress(i, j, p2)+fat1*dvols
                        IF (i /= j) atprop%atstress(j, i, p2) = atprop%atstress(j, i, p2)+fat1*dvols
                     END IF
                  END DO
               END DO
            END DO
            CALL pw_pool_give_back_pw(pw_big_pool, phi_g)
            CALL pw_pool_give_back_pw(pw_big_pool, rhob_g)
         END IF
         CALL rs_grid_release(rpot)
      END IF

      CALL pw_pool_give_back_pw(pw_big_pool, phi_r)

      !---------- END OF ELECTROSTATIC CALCULATION --------

      !------------- STRESS TENSOR CALCULATION ------------

      IF ((use_virial) .AND. (PRESENT(pv_g))) THEN
         DO i = 1, 3
            DO j = i, 3
               f_stress(i, j) = pw_integral_a2b(dphi_g(i)%pw, dphi_g(j)%pw)
               f_stress(j, i) = f_stress(i, j)
            END DO
         END DO
         ffa = (1.0_dp/fourpi)*(0.5_dp/dg_rho0%zet(1))**2
         f_stress = -ffa*f_stress
         pv_g = h_stress+f_stress
      END IF

      !--------END OF STRESS TENSOR CALCULATION -----------

      ALLOCATE (drpot(1:3))
      DO i = 1, 3
         CALL rs_grid_create(drpot(i)%rs_grid, rs_desc)
         CALL rs_grid_set_box(grid_b, rs=drpot(i)%rs_grid)
         CALL pw_transfer(dphi_g(i)%pw, rhob_r)
         CALL pw_pool_give_back_pw(pw_big_pool, dphi_g(i)%pw)
         CALL rs_pw_transfer(drpot(i)%rs_grid, rhob_r, pw2rs)
      END DO

      CALL pw_pool_give_back_pw(pw_big_pool, rhob_r)

      !----------------- FORCE CALCULATION ----------------

      ! initialize the forces
      IF (PRESENT(fg_coulomb)) THEN
         fg_coulomb = 0.0_dp
         dvols = rhos1%pw%pw_grid%dvol

         ipart = 0
         DO

            CALL set_list(particle_set, npart, exp_igr%centre, p1, rden, ipart, exp_igr%core_centre)
            CALL set_list(particle_set, npart, exp_igr%centre, p2, rden, ipart, exp_igr%core_centre)
            IF (p1 == 0 .AND. p2 == 0) EXIT

            is1_core = (particle_set(p1)%shell_index /= 0)
            IF (p2 /= 0) THEN
               is2_core = (particle_set(p2)%shell_index /= 0)
            ELSE
               is2_core = .FALSE.
            END IF

            ! calculate function on small boxes (we use double packing in FFT)

            CALL get_patch_again(dg, particle_set, exp_igr, p1, p2, rhos1, rhos2, &
                                 is1_core=is1_core, is2_core=is2_core, charges=charges)

            ! sum boxes on real space grids (big box)
            IF (is1_core) THEN
               CALL dg_sum_patch_force_3d(drpot, rhos1, &
                                          exp_igr%core_centre(:, particle_set(p1)%shell_index), fat)
               fgcore_coulomb(1, particle_set(p1)%shell_index) = &
                  fgcore_coulomb(1, particle_set(p1)%shell_index)-fat(1)*dvols
               fgcore_coulomb(2, particle_set(p1)%shell_index) = &
                  fgcore_coulomb(2, particle_set(p1)%shell_index)-fat(2)*dvols
               fgcore_coulomb(3, particle_set(p1)%shell_index) = &
                  fgcore_coulomb(3, particle_set(p1)%shell_index)-fat(3)*dvols
            ELSE
               CALL dg_sum_patch_force_3d(drpot, rhos1, exp_igr%centre(:, p1), fat)
               fg_coulomb(1, p1) = fg_coulomb(1, p1)-fat(1)*dvols
               fg_coulomb(2, p1) = fg_coulomb(2, p1)-fat(2)*dvols
               fg_coulomb(3, p1) = fg_coulomb(3, p1)-fat(3)*dvols
            END IF
            IF (p2 /= 0 .AND. is2_core) THEN
               CALL dg_sum_patch_force_3d(drpot, rhos1, &
                                          exp_igr%core_centre(:, particle_set(p2)%shell_index), fat)
               fgcore_coulomb(1, particle_set(p2)%shell_index) = &
                  fgcore_coulomb(1, particle_set(p2)%shell_index)-fat(1)*dvols
               fgcore_coulomb(2, particle_set(p2)%shell_index) = &
                  fgcore_coulomb(2, particle_set(p2)%shell_index)-fat(2)*dvols
               fgcore_coulomb(3, particle_set(p2)%shell_index) = &
                  fgcore_coulomb(3, particle_set(p2)%shell_index)-fat(3)*dvols
            ELSEIF (p2 /= 0) THEN
               CALL dg_sum_patch_force_3d(drpot, rhos2, exp_igr%centre(:, p2), fat)
               fg_coulomb(1, p2) = fg_coulomb(1, p2)-fat(1)*dvols
               fg_coulomb(2, p2) = fg_coulomb(2, p2)-fat(2)*dvols
               fg_coulomb(3, p2) = fg_coulomb(3, p2)-fat(3)*dvols
            END IF

         END DO
      ENDIF
      IF (PRESENT(fgshell_coulomb)) THEN
         fgshell_coulomb = 0.0_dp
         dvols = rhos1%pw%pw_grid%dvol

         ipart = 0
         DO
            CALL set_list(shell_particle_set, nshell, exp_igr%shell_centre, p1, rden, ipart)
            CALL set_list(shell_particle_set, nshell, exp_igr%shell_centre, p2, rden, ipart)
            IF (p1 == 0 .AND. p2 == 0) EXIT

            ! calculate function on small boxes (we use double packing in FFT)
            CALL get_patch_again(dg, shell_particle_set, exp_igr, p1, p2, rhos1, rhos2, &
                                 is1_shell=.TRUE., is2_shell=.TRUE., charges=charges)

            ! sum boxes on real space grids (big box)
            CALL dg_sum_patch_force_3d(drpot, rhos1, exp_igr%shell_centre(:, p1), fat)
            fgshell_coulomb(1, p1) = fgshell_coulomb(1, p1)-fat(1)*dvols
            fgshell_coulomb(2, p1) = fgshell_coulomb(2, p1)-fat(2)*dvols
            fgshell_coulomb(3, p1) = fgshell_coulomb(3, p1)-fat(3)*dvols
            IF (p2 /= 0) THEN
               CALL dg_sum_patch_force_3d(drpot, rhos2, exp_igr%shell_centre(:, p2), fat)
               fgshell_coulomb(1, p2) = fgshell_coulomb(1, p2)-fat(1)*dvols
               fgshell_coulomb(2, p2) = fgshell_coulomb(2, p2)-fat(2)*dvols
               fgshell_coulomb(3, p2) = fgshell_coulomb(3, p2)-fat(3)*dvols
            END IF
         END DO

      END IF
      !--------------END OF FORCE CALCULATION -------------

      !------------------CLEANING UP ----------------------

      CALL rs_grid_release(rden)
      IF (ASSOCIATED(drpot)) THEN
         DO i = 1, 3
            CALL rs_grid_release(drpot(i)%rs_grid)
         END DO
         DEALLOCATE (drpot)
      END IF

      CALL pw_pool_give_back_pw(pw_small_pool, rhos1%pw)
      CALL pw_pool_give_back_pw(pw_small_pool, rhos2%pw)
      CALL structure_factor_deallocate(exp_igr)

      CALL timestop(handle)

   END SUBROUTINE pme_evaluate

! **************************************************************************************************
!> \brief Calculates local density in a small box
!> \param dg ...
!> \param particle_set ...
!> \param exp_igr ...
!> \param box ...
!> \param p1 ...
!> \param p2 ...
!> \param grid_b ...
!> \param grid_s ...
!> \param rhos1 ...
!> \param rhos2 ...
!> \param is1_core ...
!> \param is2_core ...
!> \param is1_shell ...
!> \param is2_shell ...
!> \param core_particle_set ...
!> \param charges ...
!> \par History
!>      JGH (23-Mar-2001) : Switch to integer from particle list pointers
!> \author JGH (21-Mar-2001)
! **************************************************************************************************
   SUBROUTINE get_patch(dg, particle_set, exp_igr, box, p1, p2, &
                        grid_b, grid_s, rhos1, rhos2, is1_core, is2_core, is1_shell, &
                        is2_shell, core_particle_set, charges)

      TYPE(dg_type), POINTER                             :: dg
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(structure_factor_type)                        :: exp_igr
      TYPE(cell_type), INTENT(IN)                        :: box
      INTEGER, INTENT(IN)                                :: p1, p2
      TYPE(pw_grid_type), INTENT(IN)                     :: grid_b, grid_s
      TYPE(pw_p_type)                                    :: rhos1, rhos2
      LOGICAL, OPTIONAL                                  :: is1_core, is2_core, is1_shell, is2_shell
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: core_particle_set
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_patch', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: ex1, ex2, ey1, ey2, ez1, ez2
      INTEGER, DIMENSION(:), POINTER                     :: center1, center2
      LOGICAL                                            :: my_is1_core, my_is1_shell, my_is2_core, &
                                                            my_is2_shell, use_charge_array
      REAL(KIND=dp)                                      :: q1, q2
      REAL(KIND=dp), DIMENSION(3)                        :: r1, r2
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(pw_p_type), POINTER                           :: rho0
      TYPE(shell_kind_type), POINTER                     :: shell

      NULLIFY (shell)
      use_charge_array = PRESENT(charges)
      IF (use_charge_array) use_charge_array = ASSOCIATED(charges)
      my_is1_core = .FALSE.
      my_is2_core = .FALSE.
      IF (PRESENT(is1_core)) my_is1_core = is1_core
      IF (PRESENT(is2_core)) my_is2_core = is2_core
      IF (my_is1_core .OR. my_is2_core) THEN
         IF(.NOT.(PRESENT(core_particle_set)))CALL cp__a("pme.F",536)
      END IF
      my_is1_shell = .FALSE.
      my_is2_shell = .FALSE.
      IF (PRESENT(is1_shell)) my_is1_shell = is1_shell
      IF (PRESENT(is2_shell)) my_is2_shell = is2_shell
      IF (my_is1_core .AND. my_is1_shell) THEN
         CALL cp__b("pme.F",543,"Shell-model: cannot be core and shell simultaneously")
      END IF

      CALL dg_get(dg, dg_rho0=dg_rho0)
      rho0 => dg_rho0%density

      IF (my_is1_core) THEN
         r1 = core_particle_set(particle_set(p1)%shell_index)%r
      ELSE
         r1 = particle_set(p1)%r
      END IF
      atomic_kind => particle_set(p1)%atomic_kind
      IF (my_is1_core) THEN
         CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
         q1 = shell%charge_core
      ELSE IF (my_is1_shell) THEN
         CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
         q1 = shell%charge_shell
      ELSE
         CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q1)
      END IF
      IF (use_charge_array) q1 = charges(p1)

      IF (my_is1_shell) THEN
         center1 => exp_igr%shell_centre(:, p1)
         ex1 => exp_igr%shell_ex(:, p1)
         ey1 => exp_igr%shell_ey(:, p1)
         ez1 => exp_igr%shell_ez(:, p1)
      ELSEIF (my_is1_core) THEN
         center1 => exp_igr%core_centre(:, particle_set(p1)%shell_index)
         ex1 => exp_igr%core_ex(:, particle_set(p1)%shell_index)
         ey1 => exp_igr%core_ey(:, particle_set(p1)%shell_index)
         ez1 => exp_igr%core_ez(:, particle_set(p1)%shell_index)
      ELSE
         center1 => exp_igr%centre(:, p1)
         ex1 => exp_igr%ex(:, p1)
         ey1 => exp_igr%ey(:, p1)
         ez1 => exp_igr%ez(:, p1)
      END IF

      CALL dg_get_strucfac(box%hmat, r1, grid_s%npts, grid_b%npts, center1, &
                           exp_igr%lb, ex1, ey1, ez1)

      IF (p2 /= 0) THEN
         IF (my_is2_core) THEN
            r2 = core_particle_set(particle_set(p2)%shell_index)%r
         ELSE
            r2 = particle_set(p2)%r
         END IF
         atomic_kind => particle_set(p2)%atomic_kind
         IF (my_is2_core) THEN
            CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
            q2 = shell%charge_core
         ELSE IF (my_is2_shell) THEN
            CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
            q2 = shell%charge_shell
         ELSE
            CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q2)
         END IF
         IF (use_charge_array) q2 = charges(p2)

         IF (my_is2_shell) THEN
            center2 => exp_igr%shell_centre(:, p2)
            ex2 => exp_igr%shell_ex(:, p2)
            ey2 => exp_igr%shell_ey(:, p2)
            ez2 => exp_igr%shell_ez(:, p2)
         ELSEIF (my_is2_core) THEN
            center2 => exp_igr%core_centre(:, particle_set(p2)%shell_index)
            ex2 => exp_igr%core_ex(:, particle_set(p2)%shell_index)
            ey2 => exp_igr%core_ey(:, particle_set(p2)%shell_index)
            ez2 => exp_igr%core_ez(:, particle_set(p2)%shell_index)
         ELSE
            center2 => exp_igr%centre(:, p2)
            ex2 => exp_igr%ex(:, p2)
            ey2 => exp_igr%ey(:, p2)
            ez2 => exp_igr%ez(:, p2)
         END IF
         CALL dg_get_strucfac(box%hmat, r2, grid_s%npts, grid_b%npts, center2, &
                              exp_igr%lb, ex2, ey2, ez2)
      END IF

      IF (p2 == 0) THEN
         CALL dg_get_patch(rho0, rhos1, q1, ex1, ey1, ez1)
      ELSE
         CALL dg_get_patch(rho0, rhos1, rhos2, q1, q2, ex1, ey1, ez1, ex2, ey2, ez2)
      END IF

   END SUBROUTINE get_patch

! **************************************************************************************************
!> \brief ...
!> \param dg ...
!> \param particle_set ...
!> \param exp_igr ...
!> \param p1 ...
!> \param p2 ...
!> \param rhos1 ...
!> \param rhos2 ...
!> \param is1_core ...
!> \param is2_core ...
!> \param is1_shell ...
!> \param is2_shell ...
!> \param charges ...
! **************************************************************************************************
   SUBROUTINE get_patch_again(dg, particle_set, exp_igr, p1, p2, rhos1, rhos2, is1_core, &
                              is2_core, is1_shell, is2_shell, charges)

      TYPE(dg_type), POINTER                             :: dg
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(structure_factor_type)                        :: exp_igr
      INTEGER, INTENT(IN)                                :: p1, p2
      TYPE(pw_p_type)                                    :: rhos1, rhos2
      LOGICAL, OPTIONAL                                  :: is1_core, is2_core, is1_shell, is2_shell
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_patch_again', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: ex1, ex2, ey1, ey2, ez1, ez2
      LOGICAL                                            :: my_is1_core, my_is1_shell, my_is2_core, &
                                                            my_is2_shell, use_charge_array
      REAL(KIND=dp)                                      :: q1, q2
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(pw_p_type), POINTER                           :: rho0
      TYPE(shell_kind_type), POINTER                     :: shell

      NULLIFY (shell)
      use_charge_array = PRESENT(charges)
      IF (use_charge_array) use_charge_array = ASSOCIATED(charges)
      my_is1_core = .FALSE.
      my_is2_core = .FALSE.
      IF (PRESENT(is1_core)) my_is1_core = is1_core
      IF (PRESENT(is2_core)) my_is2_core = is2_core
      my_is1_shell = .FALSE.
      my_is2_shell = .FALSE.
      IF (PRESENT(is1_shell)) my_is1_shell = is1_shell
      IF (PRESENT(is2_shell)) my_is2_shell = is2_shell

      CALL dg_get(dg, dg_rho0=dg_rho0)
      rho0 => dg_rho0%density

      atomic_kind => particle_set(p1)%atomic_kind
      IF (my_is1_core) THEN
         CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
         q1 = shell%charge_core
      ELSE IF (my_is1_shell) THEN
         CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
         q1 = shell%charge_shell
      ELSE
         CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q1)
      END IF
      IF (use_charge_array) q1 = charges(p1)
      IF (my_is1_core) THEN
         ex1 => exp_igr%core_ex(:, particle_set(p1)%shell_index)
         ey1 => exp_igr%core_ey(:, particle_set(p1)%shell_index)
         ez1 => exp_igr%core_ez(:, particle_set(p1)%shell_index)
      ELSEIF (my_is1_shell) THEN
         ex1 => exp_igr%shell_ex(:, p1)
         ey1 => exp_igr%shell_ey(:, p1)
         ez1 => exp_igr%shell_ez(:, p1)
      ELSE
         ex1 => exp_igr%ex(:, p1)
         ey1 => exp_igr%ey(:, p1)
         ez1 => exp_igr%ez(:, p1)
      END IF

      IF (p2 /= 0) THEN
         atomic_kind => particle_set(p2)%atomic_kind
         IF (my_is2_core) THEN
            CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
            q2 = shell%charge_core
         ELSE IF (my_is2_shell) THEN
            CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell)
            q2 = shell%charge_shell
         ELSE
            CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q2)
         END IF
         IF (use_charge_array) q2 = charges(p2)
         IF (my_is2_core) THEN
            ex2 => exp_igr%core_ex(:, particle_set(p2)%shell_index)
            ey2 => exp_igr%core_ey(:, particle_set(p2)%shell_index)
            ez2 => exp_igr%core_ez(:, particle_set(p2)%shell_index)
         ELSEIF (my_is2_shell) THEN
            ex2 => exp_igr%shell_ex(:, p2)
            ey2 => exp_igr%shell_ey(:, p2)
            ez2 => exp_igr%shell_ez(:, p2)
         ELSE
            ex2 => exp_igr%ex(:, p2)
            ey2 => exp_igr%ey(:, p2)
            ez2 => exp_igr%ez(:, p2)
         END IF
      END IF

      IF (p2 == 0) THEN
         CALL dg_get_patch(rho0, rhos1, q1, ex1, ey1, ez1)
      ELSE
         CALL dg_get_patch(rho0, rhos1, rhos2, q1, q2, &
                           ex1, ey1, ez1, ex2, ey2, ez2)
      END IF

   END SUBROUTINE get_patch_again

END MODULE pme

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Torsions added(DG)05-Dec-2000
!>      Variable names changed(DG)05-Dec-2000
!>      CJM SEPT-12-2002: int_env is now passed
!>      CJM NOV-30-2003: only uses fist_env
!> \author CJM & JGH
! **************************************************************************************************
MODULE fist_force
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_type,&
                                              init_cell,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_methods,               ONLY: cp_results_erase,&
                                              put_results
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_methods,                ONLY: ewald_pw_grid_update
   USE ewald_pw_types,                  ONLY: ewald_pw_type
   USE ewalds,                          ONLY: ewald_evaluate,&
                                              ewald_print,&
                                              ewald_self,&
                                              ewald_self_atom
   USE ewalds_multipole,                ONLY: ewald_multipole_evaluate
   USE exclusion_types,                 ONLY: exclusion_type
   USE fist_energy_types,               ONLY: fist_energy_type
   USE fist_environment_types,          ONLY: fist_env_get,&
                                              fist_environment_type
   USE fist_intra_force,                ONLY: force_intra_control
   USE fist_neighbor_list_control,      ONLY: list_control
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_type
   USE fist_nonbond_force,              ONLY: bonded_correct_gaussian,&
                                              force_nonbond
   USE fist_pol_scf,                    ONLY: fist_pol_evaluate
   USE input_constants,                 ONLY: do_fist_pol_none
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE manybody_eam,                    ONLY: density_nonbond
   USE manybody_potential,              ONLY: energy_manybody,&
                                              force_nonbond_manybody
   USE mathconstants,                   ONLY: twopi
   USE message_passing,                 ONLY: mp_sum
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE moments_utils,                   ONLY: get_reference_point
   USE multipole_types,                 ONLY: multipole_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: debye
   USE pme,                             ONLY: pme_evaluate
   USE pw_poisson_types,                ONLY: do_ewald_ewald,&
                                              do_ewald_none,&
                                              do_ewald_pme,&
                                              do_ewald_spme
   USE shell_potential_types,           ONLY: shell_kind_type
   USE spme,                            ONLY: spme_evaluate
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: fist_calc_energy_force
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_force'

! **************************************************************************************************
   TYPE debug_variables_type
      REAL(KIND=dp)                          :: pot_manybody, pot_bend, pot_torsion
      REAL(KIND=dp)                          :: pot_nonbond, pot_g, pot_bond
      REAL(KIND=dp)                          :: pot_imptors, pot_urey_bradley
      REAL(KIND=dp)                          :: pot_opbend
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: f_nonbond, f_g, f_bond, f_bend, &
                                                 f_torsion, f_imptors, f_ub, &
                                                 f_opbend
      REAL(KIND=dp), DIMENSION(3, 3)         :: pv_nonbond, pv_g, pv_bond, pv_ub, &
                                                pv_bend, pv_torsion, pv_imptors, &
                                                pv_opbend
   END TYPE debug_variables_type

CONTAINS

! **************************************************************************************************
!> \brief Calculates the total potential energy, total force, and the
!>      total pressure tensor from the potentials
!> \param fist_env ...
!> \param debug ...
!> \par History
!>      Harald Forbert(Dec-2000): Changes for multiple linked lists
!>      cjm, 20-Feb-2001: box_ref used to initialize ewald.  Now
!>                        have consistent restarts with npt and ewald
!>      JGH(15-Mar-2001): box_change replaces ensemble parameter
!>                          Call ewald_setup if box has changed
!>                          Consistent setup for PME and SPME
!>      cjm, 28-Feb-2006: box_change is gone
!> \author CJM & JGH
! **************************************************************************************************
   SUBROUTINE fist_calc_energy_force(fist_env, debug)
      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(debug_variables_type), INTENT(INOUT), &
         OPTIONAL                                        :: debug

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_calc_energy_force', &
         routineP = moduleN//':'//routineN

      INTEGER :: do_ipol, ewald_type, fg_coulomb_size, handle, i, ii, ikind, iparticle_kind, &
         iparticle_local, iw, iw2, j, natoms, nlocal_particles, node, nparticle_kind, &
         nparticle_local, nshell, shell_index
      LOGICAL                                            :: do_multipoles, shell_model_ad, &
                                                            shell_present, use_virial
      REAL(KIND=dp) :: mass, massc, masss, pot_bend, pot_bond, pot_imptors, pot_manybody, &
         pot_nonbond, pot_opbend, pot_shell, pot_torsion, pot_urey_bradley, umass, vg_coulomb, &
         xdum1, xdum2, xdum3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :) :: f_nonbond, f_total, fcore_nonbond, &
         fcore_total, fg_coulomb, fgcore_coulomb, fgshell_coulomb, fshell_nonbond, fshell_total
      REAL(KIND=dp), DIMENSION(3, 3)                     :: ident, pv_bc, pv_bend, pv_bond, pv_g, &
                                                            pv_imptors, pv_nonbond, pv_opbend, &
                                                            pv_torsion, pv_urey_bradley
      REAL(KIND=dp), DIMENSION(:), POINTER               :: e_coulomb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pv_coulomb
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(atprop_type), POINTER                         :: atprop_env
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(exclusion_type), DIMENSION(:), POINTER        :: exclusions
      TYPE(fist_energy_type), POINTER                    :: thermo
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(multipole_type), POINTER                      :: multipoles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(section_vals_type), POINTER                   :: force_env_section, mm_section, &
                                                            print_section
      TYPE(shell_kind_type), POINTER                     :: shell
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      NULLIFY (subsys, virial, atprop_env, para_env, force_env_section)
      logger => cp_get_default_logger()

      CALL fist_env_get(fist_env, &
                        subsys=subsys, &
                        para_env=para_env, &
                        input=force_env_section)

      CALL cp_subsys_get(subsys, &
                         virial=virial, &
                         atprop=atprop_env)

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      NULLIFY (atomic_kind, atomic_kind_set, cell, ewald_pw, ewald_env, &
               fist_nonbond_env, mm_section, local_molecules, local_particles, &
               molecule_kind_set, molecule_set, particle_set, print_section, &
               shell, shell_particle_set, core_particle_set, thermo, multipoles, &
               e_coulomb, pv_coulomb)

      mm_section => section_vals_get_subs_vals(force_env_section, "MM")
      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%DERIVATIVES", &
                                extension=".mmLog")
      iw2 = cp_print_key_unit_nr(logger, mm_section, "PRINT%EWALD_INFO", &
                                 extension=".mmLog")

      CALL fist_env_get(fist_env, ewald_pw=ewald_pw, ewald_env=ewald_env, &
                        local_particles=local_particles, particle_set=particle_set, &
                        atomic_kind_set=atomic_kind_set, molecule_set=molecule_set, &
                        local_molecules=local_molecules, thermo=thermo, &
                        molecule_kind_set=molecule_kind_set, fist_nonbond_env=fist_nonbond_env, &
                        cell=cell, shell_model=shell_present, shell_model_ad=shell_model_ad, &
                        multipoles=multipoles, exclusions=exclusions)

      CALL ewald_env_get(ewald_env, ewald_type=ewald_type, do_multipoles=do_multipoles, &
                         do_ipol=do_ipol)

      ! Initialize ewald grids
      CALL init_cell(cell)
      CALL ewald_pw_grid_update(ewald_pw, ewald_env, cell%hmat)

      natoms = SIZE(particle_set)
      nlocal_particles = 0
      nparticle_kind = SIZE(atomic_kind_set)
      DO ikind = 1, nparticle_kind
         nlocal_particles = nlocal_particles+local_particles%n_el(ikind)
      ENDDO

      ALLOCATE (f_nonbond(3, natoms))

      nshell = 0
      IF (shell_present) THEN
         CALL fist_env_get(fist_env, shell_particle_set=shell_particle_set, &
                           core_particle_set=core_particle_set)
         IF(.NOT.(ASSOCIATED(shell_particle_set)))CALL cp__a("fist_force.F",219)
         IF(.NOT.(ASSOCIATED(core_particle_set)))CALL cp__a("fist_force.F",220)
         nshell = SIZE(shell_particle_set)
         ALLOCATE (fshell_nonbond(3, nshell))
         ALLOCATE (fcore_nonbond(3, nshell))
      ELSE
         NULLIFY (shell_particle_set, core_particle_set)
      END IF

      IF (fist_nonbond_env%do_nonbonded) THEN
         ! First check with list_control to update neighbor lists
         IF (ASSOCIATED(exclusions)) THEN
            CALL list_control(atomic_kind_set, particle_set, local_particles, &
                              cell, fist_nonbond_env, para_env, mm_section, shell_particle_set, &
                              core_particle_set, exclusions=exclusions)
         ELSE
            CALL list_control(atomic_kind_set, particle_set, local_particles, &
                              cell, fist_nonbond_env, para_env, mm_section, shell_particle_set, &
                              core_particle_set)
         END IF
      END IF

      ! Initialize force, energy and pressure tensor arrays
      DO i = 1, natoms
         particle_set(i)%f(1) = 0.0_dp
         particle_set(i)%f(2) = 0.0_dp
         particle_set(i)%f(3) = 0.0_dp
      ENDDO
      IF (nshell > 0) THEN
         DO i = 1, nshell
            shell_particle_set(i)%f(1) = 0.0_dp
            shell_particle_set(i)%f(2) = 0.0_dp
            shell_particle_set(i)%f(3) = 0.0_dp
            core_particle_set(i)%f(1) = 0.0_dp
            core_particle_set(i)%f(2) = 0.0_dp
            core_particle_set(i)%f(3) = 0.0_dp
         END DO
      ENDIF

      IF (use_virial) THEN
         pv_bc = 0.0_dp
         pv_bond = 0.0_dp
         pv_bend = 0.0_dp
         pv_torsion = 0.0_dp
         pv_imptors = 0.0_dp
         pv_opbend = 0.0_dp
         pv_urey_bradley = 0.0_dp
         pv_nonbond = 0.0_dp
         pv_g = 0.0_dp
         virial%pv_virial = 0.0_dp
      END IF

      pot_nonbond = 0.0_dp
      pot_manybody = 0.0_dp
      pot_bond = 0.0_dp
      pot_bend = 0.0_dp
      pot_torsion = 0.0_dp
      pot_opbend = 0.0_dp
      pot_imptors = 0.0_dp
      pot_urey_bradley = 0.0_dp
      pot_shell = 0.0_dp
      vg_coulomb = 0.0_dp
      thermo%pot = 0.0_dp
      thermo%harm_shell = 0.0_dp

      ! Get real-space non-bonded forces:
      IF (iw > 0) THEN
         WRITE (iw, '(A)') " FIST:: FORCES IN INPUT..."
         WRITE (iw, '(3f15.9)') ((particle_set(i)%f(j), j=1, 3), i=1, SIZE(particle_set))
      END IF

      IF (fist_nonbond_env%do_nonbonded) THEN
         ! Compute density for EAM
         CALL density_nonbond(fist_nonbond_env, particle_set, cell, para_env)

         ! Compute embedding function and manybody energy
         CALL energy_manybody(fist_nonbond_env, atomic_kind_set, local_particles, particle_set, &
                              cell, pot_manybody, para_env, mm_section)

         ! Nonbond contribution + Manybody Forces
         IF (shell_present) THEN
            CALL force_nonbond(fist_nonbond_env, ewald_env, particle_set, cell, &
                               pot_nonbond, f_nonbond, pv_nonbond, &
                               fshell_nonbond=fshell_nonbond, fcore_nonbond=fcore_nonbond, &
                               atprop_env=atprop_env, &
                               atomic_kind_set=atomic_kind_set, use_virial=use_virial)
         ELSE
            CALL force_nonbond(fist_nonbond_env, ewald_env, particle_set, cell, &
                               pot_nonbond, f_nonbond, pv_nonbond, atprop_env=atprop_env, &
                               atomic_kind_set=atomic_kind_set, use_virial=use_virial)
            CALL force_nonbond_manybody(fist_nonbond_env, particle_set, cell, f_nonbond, pv_nonbond, &
                                        use_virial=use_virial)
         END IF
      ELSE
         f_nonbond = 0.0_dp
         IF (shell_present) THEN
            fshell_nonbond = 0.0_dp
            fcore_nonbond = 0.0_dp
         END IF
         IF (use_virial) THEN
            pv_nonbond = 0.0_dp
         END IF
      END IF

      IF (iw > 0) THEN
         WRITE (iw, '(A)') " FIST:: NONBOND + R-SPACE ELECTROSTATIC FORCES ..."
         WRITE (iw, '(3f15.9)') f_nonbond
         IF (shell_present .AND. shell_model_ad) THEN
            WRITE (iw, '(A)') " FIST:: SHELL NONBOND + R-SPACE ELECTROSTATIC FORCES ..."
            WRITE (iw, '(3f15.9)') fshell_nonbond
            WRITE (iw, '(A)') " FIST:: CORE NONBOND + R-SPACE ELECTROSTATIC FORCES ..."
            WRITE (iw, '(3f15.9)') fcore_nonbond
         END IF
      END IF

      ! Get g-space non-bonded forces:
      IF (ewald_type /= do_ewald_none) THEN
         ! Determine size of the forces array
         SELECT CASE (ewald_type)
         CASE (do_ewald_ewald)
            fg_coulomb_size = nlocal_particles
         CASE DEFAULT
            fg_coulomb_size = natoms
         END SELECT
         ! Allocate and zeroing arrays
         ALLOCATE (fg_coulomb(3, fg_coulomb_size))
         fg_coulomb = 0.0_dp
         IF (shell_present) THEN
            ALLOCATE (fgshell_coulomb(3, nshell))
            ALLOCATE (fgcore_coulomb(3, nshell))
            fgshell_coulomb = 0.0_dp
            fgcore_coulomb = 0.0_dp
         END IF
         IF (shell_present .AND. do_multipoles) THEN
            CALL cp__b("fist_force.F",353,"Multipoles and Core-Shell model not implemented.")
         END IF
         ! If not multipole: Compute self-interaction and neutralizing background
         ! for multipoles is handled separately..
         IF (.NOT. do_multipoles) THEN
            CALL ewald_self(ewald_env, cell, atomic_kind_set, local_particles, &
                            thermo%e_self, thermo%e_neut, fist_nonbond_env%charges)
            IF (atprop_env%energy) THEN
               CALL ewald_self_atom(ewald_env, atomic_kind_set, local_particles, &
                                    atprop_env%atener, fist_nonbond_env%charges)
               atprop_env%atener = atprop_env%atener+thermo%e_neut/SIZE(atprop_env%atener)
            END IF
         END IF

         ! Polarizable force-field
         IF (do_ipol /= do_fist_pol_none) THEN
            ! Check if an array of chagres was provided and in case abort due to lack of implementation
            IF (ASSOCIATED(fist_nonbond_env%charges)) THEN
               CALL cp__b("fist_force.F",371,"Polarizable force field and array charges not implemented!")
            END IF
            ! Converge the dipoles self-consistently
            CALL fist_pol_evaluate(atomic_kind_set, multipoles, ewald_env, &
                                   ewald_pw, fist_nonbond_env, cell, particle_set, &
                                   local_particles, thermo, vg_coulomb, pot_nonbond, f_nonbond, &
                                   fg_coulomb, use_virial, pv_g, pv_nonbond, mm_section, do_ipol)
         ELSE
            ! Non-Polarizable force-field
            SELECT CASE (ewald_type)
            CASE (do_ewald_ewald)
               ! Parallelisation over atoms --> allocate local atoms
               IF (shell_present) THEN
                  ! Check if an array of chagres was provided and in case abort due to lack of implementation
                  IF (ASSOCIATED(fist_nonbond_env%charges)) THEN
                     CALL cp__b("fist_force.F",386,"Core-Shell and array charges not implemented!")
                  END IF
                  IF (do_multipoles) THEN
                     CALL cp__b("fist_force.F",389,"Multipole Ewald and CORE-SHELL not yet implemented within Ewald sum!")
                  ELSE
                     CALL cp__b("fist_force.F",391,"Core-Shell model not yet implemented within Ewald sums.")
                  END IF
               ELSE
                  IF (do_multipoles) THEN
                     ! Check if an array of chagres was provided and in case abort due to lack of implementation
                     IF (ASSOCIATED(fist_nonbond_env%charges)) THEN
                        CALL cp__b("fist_force.F",397,"Multipole Ewald and array charges not implemented!")
                     END IF
                     CALL ewald_multipole_evaluate(ewald_env, ewald_pw, fist_nonbond_env, cell, &
                                                   particle_set, local_particles, vg_coulomb, pot_nonbond, thermo%e_neut, &
                                                   thermo%e_self, multipoles%task, do_correction_bonded=.TRUE., do_forces=.TRUE., &
                                                   do_stress=use_virial, do_efield=.FALSE., radii=multipoles%radii, &
                                                   charges=multipoles%charges, dipoles=multipoles%dipoles, &
                                                   quadrupoles=multipoles%quadrupoles, forces_local=fg_coulomb, &
                                                   forces_glob=f_nonbond, pv_local=pv_g, pv_glob=pv_nonbond, iw=iw2, &
                                                   do_debug=.TRUE., atomic_kind_set=atomic_kind_set, mm_section=mm_section)
                  ELSE
                     IF (atprop_env%energy) THEN
                        ALLOCATE (e_coulomb(fg_coulomb_size))
                     END IF
                     IF (atprop_env%stress) THEN
                        ALLOCATE (pv_coulomb(3, 3, fg_coulomb_size))
                     END IF
                     CALL ewald_evaluate(ewald_env, ewald_pw, cell, atomic_kind_set, particle_set, &
                                         local_particles, fg_coulomb, vg_coulomb, pv_g, use_virial=use_virial, &
                                         charges=fist_nonbond_env%charges, e_coulomb=e_coulomb, &
                                         pv_coulomb=pv_coulomb)
                  END IF
               END IF
            CASE (do_ewald_pme)
               ! Parallelisation over grids --> allocate all atoms
               IF (shell_present) THEN
                  ! Check if an array of chagres was provided and in case abort due to lack of implementation
                  IF (ASSOCIATED(fist_nonbond_env%charges)) THEN
                     CALL cp__b("fist_force.F",425,"Core-Shell and array charges not implemented!")
                  END IF
                  IF (do_multipoles) THEN
                     CALL cp__b("fist_force.F",428,"Multipole Ewald and CORE-SHELL not yet implemented within a PME scheme!")
                  ELSE
                     CALL pme_evaluate(ewald_env, ewald_pw, cell, particle_set, vg_coulomb, fg_coulomb, &
                                       pv_g, shell_particle_set=shell_particle_set, core_particle_set=core_particle_set, &
                                       fgshell_coulomb=fgshell_coulomb, fgcore_coulomb=fgcore_coulomb, use_virial=use_virial, &
                                       atprop=atprop_env)
                     CALL mp_sum(fgshell_coulomb, para_env%group)
                     CALL mp_sum(fgcore_coulomb, para_env%group)
                  END IF
               ELSE
                  IF (do_multipoles) THEN
                     CALL cp__b("fist_force.F",439,"Multipole Ewald not yet implemented within a PME scheme!")
                  ELSE
                     CALL pme_evaluate(ewald_env, ewald_pw, cell, particle_set, vg_coulomb, fg_coulomb, &
                                       pv_g, use_virial=use_virial, charges=fist_nonbond_env%charges, &
                                       atprop=atprop_env)
                  END IF
               END IF
               CALL mp_sum(fg_coulomb, para_env%group)
            CASE (do_ewald_spme)
               ! Parallelisation over grids --> allocate all atoms
               IF (shell_present) THEN
                  ! Check if an array of chagres was provided and in case abort due to lack of implementation
                  IF (ASSOCIATED(fist_nonbond_env%charges)) THEN
                     CALL cp__b("fist_force.F",452,"Core-Shell and array charges not implemented!")
                  END IF
                  IF (do_multipoles) THEN
                     CALL cp__b("fist_force.F",455,"Multipole Ewald and CORE-SHELL not yet implemented within a SPME scheme!")
                  ELSE
                     CALL spme_evaluate(ewald_env, ewald_pw, cell, particle_set, fg_coulomb, vg_coulomb, &
                                        pv_g, shell_particle_set=shell_particle_set, core_particle_set=core_particle_set, &
                                        fgshell_coulomb=fgshell_coulomb, fgcore_coulomb=fgcore_coulomb, use_virial=use_virial, &
                                        atprop=atprop_env)
                     CALL mp_sum(fgshell_coulomb, para_env%group)
                     CALL mp_sum(fgcore_coulomb, para_env%group)
                  END IF
               ELSE
                  IF (do_multipoles) THEN
                     CALL cp__b("fist_force.F",466,"Multipole Ewald not yet implemented within a SPME scheme!")
                  ELSE
                     CALL spme_evaluate(ewald_env, ewald_pw, cell, particle_set, fg_coulomb, vg_coulomb, &
                                        pv_g, use_virial=use_virial, charges=fist_nonbond_env%charges, &
                                        atprop=atprop_env)
                  END IF
               END IF
               CALL mp_sum(fg_coulomb, para_env%group)
            END SELECT
         END IF

         ! Subtract the interaction between screening charges. This is a
         ! correction in real-space and depends on the neighborlists. Therefore
         ! it is only executed if fist_nonbond_env%do_nonbonded is set.
         IF (fist_nonbond_env%do_nonbonded) THEN
            ! Correction for core-shell model
            IF (shell_present) THEN
               CALL bonded_correct_gaussian(fist_nonbond_env, atomic_kind_set, &
                                            local_particles, particle_set, ewald_env, thermo%e_bonded, &
                                            pv_bc, shell_particle_set=shell_particle_set, &
                                            core_particle_set=core_particle_set, atprop_env=atprop_env, cell=cell, &
                                            use_virial=use_virial)
            ELSE
               IF (.NOT. do_multipoles) THEN
                  CALL bonded_correct_gaussian(fist_nonbond_env, &
                                               atomic_kind_set, local_particles, particle_set, &
                                               ewald_env, thermo%e_bonded, pv_bc=pv_bc, atprop_env=atprop_env, cell=cell, &
                                               use_virial=use_virial)
               END IF
            END IF
         END IF

         IF (.NOT. do_multipoles) THEN
            ! Multipole code has its own printing routine.
            CALL ewald_print(iw2, pot_nonbond, vg_coulomb, thermo%e_self, thermo%e_neut, &
                             thermo%e_bonded)
         END IF
      ELSE
         IF (use_virial) THEN
            pv_g = 0.0_dp
            pv_bc = 0.0_dp
         END IF
         thermo%e_neut = 0.0_dp
      END IF

      IF (iw > 0) THEN
         IF (ALLOCATED(fg_coulomb)) THEN
            WRITE (iw, '(A)') " FIST:: NONBONDED ELECTROSTATIC FORCES IN G-SPACE..."
            WRITE (iw, '(3f15.9)') ((fg_coulomb(j, i), j=1, 3), i=1, SIZE(fg_coulomb, 2))
         END IF
         IF (shell_present .AND. shell_model_ad .AND. ALLOCATED(fgshell_coulomb)) THEN
            WRITE (iw, '(A)') " FIST:: SHELL NONBONDED ELECTROSTATIC FORCES IN G-SPACE..."
            WRITE (iw, '(3f15.9)') ((fgshell_coulomb(j, i), j=1, 3), i=1, SIZE(fgshell_coulomb, 2))
            WRITE (iw, '(A)') " FIST:: CORE NONBONDED ELECTROSTATIC FORCES IN G-SPACE..."
            WRITE (iw, '(3f15.9)') ((fgcore_coulomb(j, i), j=1, 3), i=1, SIZE(fgcore_coulomb, 2))
         END IF
      END IF

      ! Get intramolecular forces
      IF (PRESENT(debug)) THEN
         CALL force_intra_control(molecule_set, molecule_kind_set, &
                                  local_molecules, particle_set, shell_particle_set, &
                                  core_particle_set, pot_bond, pot_bend, pot_urey_bradley, &
                                  pot_torsion, pot_imptors, pot_opbend, pot_shell, pv_bond, pv_bend, &
                                  pv_urey_bradley, pv_torsion, pv_imptors, pv_opbend, &
                                  debug%f_bond, debug%f_bend, debug%f_torsion, debug%f_ub, &
                                  debug%f_imptors, debug%f_opbend, cell, use_virial, atprop_env)

      ELSE
         CALL force_intra_control(molecule_set, molecule_kind_set, &
                                  local_molecules, particle_set, shell_particle_set, &
                                  core_particle_set, pot_bond, pot_bend, pot_urey_bradley, &
                                  pot_torsion, pot_imptors, pot_opbend, pot_shell, pv_bond, pv_bend, &
                                  pv_urey_bradley, pv_torsion, pv_imptors, pv_opbend, &
                                  cell=cell, use_virial=use_virial, atprop_env=atprop_env)
      ENDIF

      IF (iw > 0) THEN
         xdum1 = cp_unit_from_cp2k(pot_bond, "kcalmol")
         xdum2 = cp_unit_from_cp2k(pot_bend, "kcalmol")
         xdum3 = cp_unit_from_cp2k(pot_urey_bradley, "kcalmol")
         WRITE (iw, '(A)') " FIST energy contributions in kcal/mol:"
         WRITE (iw, '(1x,"BOND    = ",f13.4,'// &
                '2x,"ANGLE   = ",f13.4,'// &
                '2x,"UBRAD   = ",f13.4)') xdum1, xdum2, xdum3
         xdum1 = cp_unit_from_cp2k(pot_torsion, "kcalmol")
         xdum2 = cp_unit_from_cp2k(pot_imptors, "kcalmol")
         xdum3 = cp_unit_from_cp2k(pot_opbend, "kcalmol")
         WRITE (iw, '(1x,"TORSION = ",f13.4,'// &
                '2x,"IMPTORS = ",f13.4,'// &
                '2x,"OPBEND  = ",f13.4)') xdum1, xdum2, xdum3

         WRITE (iw, '(A)') " FIST:: CORRECTED BONDED ELECTROSTATIC FORCES + INTERNAL FORCES..."
         WRITE (iw, '(3f15.9)') ((particle_set(i)%f(j), j=1, 3), i=1, SIZE(particle_set))
         IF (shell_present .AND. shell_model_ad) THEN
            WRITE (iw, '(A)') " FIST:: SHELL CORRECTED BONDED ELECTROSTATIC FORCES + INTERNAL FORCES..."
            WRITE (iw, '(3f15.9)') ((shell_particle_set(i)%f(j), j=1, 3), i=1, SIZE(shell_particle_set))
            WRITE (iw, '(A)') " FIST:: CORE CORRECTED BONDED ELECTROSTATIC FORCES + INTERNAL FORCES..."
            WRITE (iw, '(3f15.9)') ((core_particle_set(i)%f(j), j=1, 3), i=1, SIZE(core_particle_set))
         END IF
      END IF

      ! add up all the potential energies
      thermo%pot = pot_nonbond+pot_bond+pot_bend+pot_torsion+pot_opbend+ &
                   pot_imptors+pot_urey_bradley+pot_manybody+pot_shell

      CALL mp_sum(thermo%pot, para_env%group)
      IF (shell_present) THEN
         thermo%harm_shell = pot_shell
         CALL mp_sum(thermo%harm_shell, para_env%group)
      END IF
      ! add g-space contributions if needed
      IF (ewald_type /= do_ewald_none) THEN
         ! e_self, e_neut, and ebonded are already summed over all processors
         ! vg_coulomb is not calculated in parallel
         thermo%e_gspace = vg_coulomb
         thermo%pot = thermo%pot+thermo%e_self+thermo%e_neut
         thermo%pot = thermo%pot+vg_coulomb+thermo%e_bonded
         ! add the induction energy of the dipoles for polarizable models
         IF (do_ipol /= do_fist_pol_none) thermo%pot = thermo%pot+thermo%e_induction
      END IF

      ! add up all the forces
      !
      ! nonbonded forces might be calculated for atoms not on this node
      ! ewald forces are strictly local -> sum only over pnode
      ! We first sum the forces in f_nonbond, this allows for a more efficient
      ! global sum in the parallel code and in the end copy them back to part
      ALLOCATE (f_total(3, natoms))
      f_total = 0.0_dp
      DO i = 1, natoms
         f_total(1, i) = particle_set(i)%f(1)+f_nonbond(1, i)
         f_total(2, i) = particle_set(i)%f(2)+f_nonbond(2, i)
         f_total(3, i) = particle_set(i)%f(3)+f_nonbond(3, i)
      END DO
      IF (shell_present) THEN
         ALLOCATE (fshell_total(3, nshell))
         ALLOCATE (fcore_total(3, nshell))
         DO i = 1, nshell
            fshell_total(1, i) = shell_particle_set(i)%f(1)+fshell_nonbond(1, i)
            fshell_total(2, i) = shell_particle_set(i)%f(2)+fshell_nonbond(2, i)
            fshell_total(3, i) = shell_particle_set(i)%f(3)+fshell_nonbond(3, i)
            fcore_total(1, i) = core_particle_set(i)%f(1)+fcore_nonbond(1, i)
            fcore_total(2, i) = core_particle_set(i)%f(2)+fcore_nonbond(2, i)
            fcore_total(3, i) = core_particle_set(i)%f(3)+fcore_nonbond(3, i)
         END DO
      END IF

      IF (iw > 0) THEN
         WRITE (iw, '(A)') " FIST::(1)INTERNAL + ELECTROSTATIC BONDED + NONBONDED"
         WRITE (iw, '(3f15.9)') ((f_total(j, i), j=1, 3), i=1, natoms)
         IF (shell_present .AND. shell_model_ad) THEN
            WRITE (iw, '(A)') " FIST::(1)SHELL INTERNAL + ELECTROSTATIC BONDED + NONBONDED"
            WRITE (iw, '(3f15.9)') ((fshell_total(j, i), j=1, 3), i=1, nshell)
            WRITE (iw, '(A)') " FIST::(1)CORE INTERNAL + ELECTROSTATIC BONDED + NONBONDED"
            WRITE (iw, '(3f15.9)') ((fcore_total(j, i), j=1, 3), i=1, nshell)
         END IF
      END IF

      ! Adding in the reciprocal forces: EWALD is a special case because of distrubted data
      IF (ewald_type == do_ewald_ewald) THEN
         node = 0
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               ii = local_particles%list(iparticle_kind)%array(iparticle_local)
               node = node+1

               f_total(1, ii) = f_total(1, ii)+fg_coulomb(1, node)
               f_total(2, ii) = f_total(2, ii)+fg_coulomb(2, node)
               f_total(3, ii) = f_total(3, ii)+fg_coulomb(3, node)
               IF (PRESENT(debug)) THEN
                  debug%f_g(1, ii) = debug%f_g(1, ii)+fg_coulomb(1, node)
                  debug%f_g(2, ii) = debug%f_g(2, ii)+fg_coulomb(2, node)
                  debug%f_g(3, ii) = debug%f_g(3, ii)+fg_coulomb(3, node)
               ENDIF
               IF (atprop_env%energy) THEN
                  atprop_env%atener(ii) = atprop_env%atener(ii)+e_coulomb(node)
               END IF
               IF (atprop_env%stress) THEN
                  atprop_env%atstress(:, :, ii) = atprop_env%atstress(:, :, ii)+pv_coulomb(:, :, node)
               END IF
            END DO
         END DO
         IF (atprop_env%energy) THEN
            DEALLOCATE (e_coulomb)
         END IF
         IF (atprop_env%stress) THEN
            DEALLOCATE (pv_coulomb)
         END IF
      END IF

      IF (iw > 0) THEN
         WRITE (iw, '(A)') " FIST::(2)TOTAL FORCES(1)+ ELECTROSTATIC FORCES"
         WRITE (iw, '(3f15.9)') ((f_total(j, i), j=1, 3), i=1, natoms)
         IF (shell_present .AND. shell_model_ad) THEN
            WRITE (iw, '(A)') " FIST::(2)SHELL TOTAL FORCES(1)+ ELECTROSTATIC FORCES "
            WRITE (iw, '(3f15.9)') ((fshell_total(j, i), j=1, 3), i=1, nshell)
            WRITE (iw, '(A)') " FIST::(2)CORE TOTAL FORCES(1)+ ELECTROSTATIC FORCES"
            WRITE (iw, '(3f15.9)') ((fcore_total(j, i), j=1, 3), i=1, nshell)
         END IF
      END IF

      IF (use_virial) THEN
         ! Add up all the pressure tensors
         IF (ewald_type == do_ewald_none) THEN
            virial%pv_virial = pv_nonbond+pv_bond+pv_bend+pv_torsion+pv_imptors+pv_urey_bradley
            CALL mp_sum(virial%pv_virial, para_env%group)
         ELSE
            ident = 0.0_dp
            DO i = 1, 3
               ident(i, i) = 1.0_dp
            END DO

            virial%pv_virial = pv_nonbond+pv_bond+pv_bend+pv_torsion+pv_imptors+pv_urey_bradley+pv_bc
            CALL mp_sum(virial%pv_virial, para_env%group)

            virial%pv_virial = virial%pv_virial+ident*thermo%e_neut
            virial%pv_virial = virial%pv_virial+pv_g
         END IF
      END IF

      ! Sum total forces
      CALL mp_sum(f_total, para_env%group)
      IF (shell_present .AND. shell_model_ad) THEN
         CALL mp_sum(fshell_total, para_env%group)
         CALL mp_sum(fcore_total, para_env%group)
      END IF

      ! Assign to particle_set
      SELECT CASE (ewald_type)
      CASE (do_ewald_spme, do_ewald_pme)
         IF (shell_present .AND. shell_model_ad) THEN
            DO i = 1, natoms
               IF (particle_set(i)%shell_index /= 0) THEN
                  atomic_kind => particle_set(i)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell, mass=mass)
                  umass = 1.0_dp/mass
                  massc = shell%mass_core*umass
                  masss = shell%mass_shell*umass
                  shell_index = particle_set(i)%shell_index
                  fshell_total(1, shell_index) = fshell_total(1, shell_index)+particle_set(i)%f(1)*masss
                  fshell_total(2, shell_index) = fshell_total(2, shell_index)+particle_set(i)%f(2)*masss
                  fshell_total(3, shell_index) = fshell_total(3, shell_index)+particle_set(i)%f(3)*masss
                  fcore_total(1, shell_index) = fcore_total(1, shell_index)+particle_set(i)%f(1)*massc
                  fcore_total(2, shell_index) = fcore_total(2, shell_index)+particle_set(i)%f(2)*massc
                  fcore_total(3, shell_index) = fcore_total(3, shell_index)+particle_set(i)%f(3)*massc
               ELSE
                  particle_set(i)%f(1) = f_total(1, i)+fg_coulomb(1, i)
                  particle_set(i)%f(2) = f_total(2, i)+fg_coulomb(2, i)
                  particle_set(i)%f(3) = f_total(3, i)+fg_coulomb(3, i)
               END IF
            END DO

            DO i = 1, nshell
               shell_particle_set(i)%f(1) = fshell_total(1, i)+fgshell_coulomb(1, i)
               shell_particle_set(i)%f(2) = fshell_total(2, i)+fgshell_coulomb(2, i)
               shell_particle_set(i)%f(3) = fshell_total(3, i)+fgshell_coulomb(3, i)
               core_particle_set(i)%f(1) = fcore_total(1, i)+fgcore_coulomb(1, i)
               core_particle_set(i)%f(2) = fcore_total(2, i)+fgcore_coulomb(2, i)
               core_particle_set(i)%f(3) = fcore_total(3, i)+fgcore_coulomb(3, i)
            END DO

         ELSEIF (shell_present .AND. .NOT. shell_model_ad) THEN
            CALL cp__b("fist_force.F",730,"Non adiabatic shell-model not implemented.")
         ELSE
            DO i = 1, natoms
               particle_set(i)%f(1) = f_total(1, i)+fg_coulomb(1, i)
               particle_set(i)%f(2) = f_total(2, i)+fg_coulomb(2, i)
               particle_set(i)%f(3) = f_total(3, i)+fg_coulomb(3, i)
            END DO
         END IF
      CASE (do_ewald_ewald, do_ewald_none)
         IF (shell_present .AND. shell_model_ad) THEN
            DO i = 1, natoms
               IF (particle_set(i)%shell_index /= 0) THEN
                  atomic_kind => particle_set(i)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, shell=shell, mass=mass)
                  umass = 1.0_dp/mass
                  massc = shell%mass_core*umass
                  masss = shell%mass_shell*umass
                  shell_index = particle_set(i)%shell_index
                  fshell_total(1, shell_index) = fshell_total(1, shell_index)+particle_set(i)%f(1)*masss
                  fshell_total(2, shell_index) = fshell_total(2, shell_index)+particle_set(i)%f(2)*masss
                  fshell_total(3, shell_index) = fshell_total(3, shell_index)+particle_set(i)%f(3)*masss
                  fcore_total(1, shell_index) = fcore_total(1, shell_index)+particle_set(i)%f(1)*massc
                  fcore_total(2, shell_index) = fcore_total(2, shell_index)+particle_set(i)%f(2)*massc
                  fcore_total(3, shell_index) = fcore_total(3, shell_index)+particle_set(i)%f(3)*massc
               ELSE
                  particle_set(i)%f(1) = f_total(1, i)
                  particle_set(i)%f(2) = f_total(2, i)
                  particle_set(i)%f(3) = f_total(3, i)
               END IF
            END DO
            DO i = 1, nshell
               shell_particle_set(i)%f(1) = fshell_total(1, i)
               shell_particle_set(i)%f(2) = fshell_total(2, i)
               shell_particle_set(i)%f(3) = fshell_total(3, i)
               core_particle_set(i)%f(1) = fcore_total(1, i)
               core_particle_set(i)%f(2) = fcore_total(2, i)
               core_particle_set(i)%f(3) = fcore_total(3, i)
            END DO
         ELSEIF (shell_present .AND. .NOT. shell_model_ad) THEN
            CALL cp__b("fist_force.F",769,"Non adiabatic shell-model not implemented.")
         ELSE
            DO i = 1, natoms
               particle_set(i)%f(1) = f_total(1, i)
               particle_set(i)%f(2) = f_total(2, i)
               particle_set(i)%f(3) = f_total(3, i)
            END DO
         END IF
      END SELECT

      IF (iw > 0) THEN
         WRITE (iw, '(A)') " FIST::(3)TOTAL FORCES - THE END..."
         WRITE (iw, '(3f15.9)') ((particle_set(i)%f(j), j=1, 3), i=1, natoms)
         IF (shell_present .AND. shell_model_ad) THEN
            WRITE (iw, '(A)') " FIST::(3)SHELL TOTAL FORCES - THE END..."
            WRITE (iw, '(3f15.9)') ((shell_particle_set(i)%f(j), j=1, 3), i=1, nshell)
            WRITE (iw, '(A)') " FIST::(3)CORE TOTAL FORCES - THE END..."
            WRITE (iw, '(3f15.9)') ((core_particle_set(i)%f(j), j=1, 3), i=1, nshell)
         END IF
         WRITE (iw, '(A,f15.9)') "Energy after FIST calculation.. exiting now ::", thermo%pot
      END IF
      !
      ! If we are doing debugging, check if variables are present and assign
      !
      IF (PRESENT(debug)) THEN
         CALL mp_sum(pot_manybody, para_env%group)
         debug%pot_manybody = pot_manybody
         CALL mp_sum(pot_nonbond, para_env%group)
         debug%pot_nonbond = pot_nonbond
         CALL mp_sum(pot_bond, para_env%group)
         debug%pot_bond = pot_bond
         CALL mp_sum(pot_bend, para_env%group)
         debug%pot_bend = pot_bend
         CALL mp_sum(pot_torsion, para_env%group)
         debug%pot_torsion = pot_torsion
         CALL mp_sum(pot_imptors, para_env%group)
         debug%pot_imptors = pot_imptors
         CALL mp_sum(pot_opbend, para_env%group)
         debug%pot_opbend = pot_opbend
         CALL mp_sum(pot_urey_bradley, para_env%group)
         debug%pot_urey_bradley = pot_urey_bradley
         CALL mp_sum(f_nonbond, para_env%group)
         debug%f_nonbond = f_nonbond
         IF (use_virial) THEN
            CALL mp_sum(pv_nonbond, para_env%group)
            debug%pv_nonbond = pv_nonbond
            CALL mp_sum(pv_bond, para_env%group)
            debug%pv_bond = pv_bond
            CALL mp_sum(pv_bend, para_env%group)
            debug%pv_bend = pv_bend
            CALL mp_sum(pv_torsion, para_env%group)
            debug%pv_torsion = pv_torsion
            CALL mp_sum(pv_imptors, para_env%group)
            debug%pv_imptors = pv_imptors
            CALL mp_sum(pv_urey_bradley, para_env%group)
            debug%pv_ub = pv_urey_bradley
         END IF
         SELECT CASE (ewald_type)
         CASE (do_ewald_spme, do_ewald_pme)
            debug%pot_g = vg_coulomb
            debug%pv_g = pv_g
            debug%f_g = fg_coulomb
         CASE (do_ewald_ewald)
            debug%pot_g = vg_coulomb
            IF (use_virial) debug%pv_g = pv_g
         CASE default
            debug%pot_g = 0.0_dp
            debug%f_g = 0.0_dp
            IF (use_virial) debug%pv_g = 0.0_dp
         END SELECT
      END IF

      ! print properties if requested
      print_section => section_vals_get_subs_vals(mm_section, "PRINT")
      CALL print_fist(fist_env, print_section, atomic_kind_set, particle_set, cell)

      ! deallocating all local variables
      IF (ALLOCATED(fg_coulomb)) THEN
         DEALLOCATE (fg_coulomb)
      END IF
      IF (ALLOCATED(f_total)) THEN
         DEALLOCATE (f_total)
      END IF
      DEALLOCATE (f_nonbond)
      IF (shell_present) THEN
         DEALLOCATE (fshell_total)
         IF (ewald_type /= do_ewald_none) THEN
            DEALLOCATE (fgshell_coulomb)
         END IF
         DEALLOCATE (fshell_nonbond)
      END IF
      CALL cp_print_key_finished_output(iw, logger, mm_section, "PRINT%DERIVATIVES")
      CALL cp_print_key_finished_output(iw2, logger, mm_section, "PRINT%EWALD_INFO")
      CALL timestop(handle)

   END SUBROUTINE fist_calc_energy_force

! **************************************************************************************************
!> \brief Print properties number according the requests in input file
!> \param fist_env ...
!> \param print_section ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param cell ...
!> \par History
!>      [01.2006] created
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE print_fist(fist_env, print_section, atomic_kind_set, particle_set, cell)
      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(section_vals_type), POINTER                   :: print_section
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell

      INTEGER                                            :: unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (logger, print_key, fist_nonbond_env)
      logger => cp_get_default_logger()
      print_key => section_vals_get_subs_vals(print_section, "dipole")
      CALL fist_env_get(fist_env, fist_nonbond_env=fist_nonbond_env)
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), &
                cp_p_file)) THEN
         unit_nr = cp_print_key_unit_nr(logger, print_section, "dipole", &
                                        extension=".data", middle_name="MM_DIPOLE", log_filename=.FALSE.)
         CALL fist_dipole(fist_env, print_section, atomic_kind_set, particle_set, &
                          cell, unit_nr, fist_nonbond_env%charges)
         CALL cp_print_key_finished_output(unit_nr, logger, print_key)
      END IF

   END SUBROUTINE print_fist

! **************************************************************************************************
!> \brief Evaluates the Dipole of a classical charge distribution(point-like)
!>      possibly using the berry phase formalism
!> \param fist_env ...
!> \param print_section ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param cell ...
!> \param unit_nr ...
!> \param charges ...
!> \par History
!>      [01.2006] created
!>      [12.2007] tlaino - University of Zurich - debug and extended
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE fist_dipole(fist_env, print_section, atomic_kind_set, particle_set, &
                          cell, unit_nr, charges)
      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(section_vals_type), POINTER                   :: print_section
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: unit_nr
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges

      CHARACTER(LEN=default_string_length)               :: description, dipole_type
      COMPLEX(KIND=dp)                                   :: dzeta, dzphase(3), zeta, zphase(3)
      COMPLEX(KIND=dp), DIMENSION(3)                     :: dggamma, ggamma
      INTEGER                                            :: i, iparticle_kind, j, reference
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: do_berry
      REAL(KIND=dp) :: charge_tot, ci(3), dci(3), dipole(3), dipole_deriv(3), drcc(3), dria(3), &
         dtheta, gvec(3), q, rcc(3), ria(3), theta, tmp(3), via(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ref_point
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_result_type), POINTER                      :: results

      NULLIFY (atomic_kind)
      ! Reference point
      reference = section_get_ival(print_section, keyword_name="DIPOLE%REFERENCE")
      NULLIFY (ref_point)
      description = '[DIPOLE]'
      CALL section_vals_val_get(print_section, "DIPOLE%REF_POINT", r_vals=ref_point)
      CALL section_vals_val_get(print_section, "DIPOLE%PERIODIC", l_val=do_berry)

      CALL get_reference_point(rcc, drcc, fist_env=fist_env, reference=reference, ref_point=ref_point)

      ! Dipole deriv will be the derivative of the Dipole(dM/dt=\sum e_j v_j)
      dipole_deriv = 0.0_dp
      dipole = 0.0_dp
      IF (do_berry) THEN
         dipole_type = "[BERRY PHASE]"
         rcc = pbc(rcc, cell)
         charge_tot = 0._dp
         IF (ASSOCIATED(charges)) THEN
            charge_tot = SUM(charges)
         ELSE
            DO i = 1, SIZE(particle_set)
               atomic_kind => particle_set(i)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q)
               charge_tot = charge_tot+q
            END DO
         END IF
         ria = twopi*MATMUL(cell%h_inv, rcc)
         zphase = CMPLX(COS(ria), SIN(ria), dp)**charge_tot

         dria = twopi*MATMUL(cell%h_inv, drcc)
         dzphase = charge_tot*CMPLX(-SIN(ria), COS(ria), dp)**(charge_tot-1.0_dp)*dria

         ggamma = CMPLX(1.0_dp, 0.0_dp, KIND=dp)
         dggamma = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
         DO iparticle_kind = 1, SIZE(atomic_kind_set)
            atomic_kind => atomic_kind_set(iparticle_kind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q, atom_list=atom_list)

            DO i = 1, SIZE(atom_list)
               ria = particle_set(atom_list(i))%r(:)
               ria = pbc(ria, cell)
               via = particle_set(atom_list(i))%v(:)
               IF (ASSOCIATED(charges)) q = charges(atom_list(i))
               DO j = 1, 3
                  gvec = twopi*cell%h_inv(j, :)
                  theta = SUM(ria(:)*gvec(:))
                  dtheta = SUM(via(:)*gvec(:))
                  zeta = CMPLX(COS(theta), SIN(theta), KIND=dp)**(-q)
                  dzeta = -q*CMPLX(-SIN(theta), COS(theta), KIND=dp)**(-q-1.0_dp)*dtheta
                  dggamma(j) = dggamma(j)*zeta+ggamma(j)*dzeta
                  ggamma(j) = ggamma(j)*zeta
               END DO
            ENDDO
         END DO
         dggamma = dggamma*zphase+ggamma*dzphase
         ggamma = ggamma*zphase
         IF (ALL(REAL(ggamma, KIND=dp) /= 0.0_dp)) THEN
            tmp = AIMAG(ggamma)/REAL(ggamma, KIND=dp)
            ci = ATAN(tmp)
            dci = (1.0_dp/(1.0_dp+tmp**2))* &
                  (AIMAG(dggamma)*REAL(ggamma, KIND=dp)-AIMAG(ggamma)*REAL(dggamma, KIND=dp))/(REAL(ggamma, KIND=dp))**2

            dipole = MATMUL(cell%hmat, ci)/twopi
            dipole_deriv = MATMUL(cell%hmat, dci)/twopi
         END IF
         CALL fist_env_get(fist_env=fist_env, results=results)
         CALL cp_results_erase(results, description)
         CALL put_results(results, description, dipole)
      ELSE
         dipole_type = "[Non Periodic]"
         DO i = 1, SIZE(particle_set)
            atomic_kind => particle_set(i)%atomic_kind
            ria = particle_set(i)%r(:) ! no pbc(particle_set(i)%r(:),cell) so that the total dipole is the sum of the molecular dipoles
            CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=q)
            IF (ASSOCIATED(charges)) q = charges(i)
            dipole = dipole-q*(ria-rcc)
            dipole_deriv(:) = dipole_deriv(:)-q*(particle_set(i)%v(:)-drcc)
         END DO
         CALL fist_env_get(fist_env=fist_env, results=results)
         CALL cp_results_erase(results, description)
         CALL put_results(results, description, dipole)
      END IF
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(1X,A,T48,3F11.6)') "MM DIPOLE "//TRIM(dipole_type)//"(A.U.)|", dipole
         WRITE (unit_nr, '(1X,A,T48,3F11.6)') "MM DIPOLE "//TRIM(dipole_type)//"(Debye)|", dipole*debye
         WRITE (unit_nr, '(1X,A,T48,3F11.6)') "MM DIPOLE "//TRIM(dipole_type)//" DERIVATIVE(A.U.)|", dipole_deriv
      END IF

   END SUBROUTINE fist_dipole

END MODULE fist_force
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines used for force-mixing QM/MM calculations
!> \par History
!>      2.2012 created [noam]
!> \author Noam Bernstein
! **************************************************************************************************
MODULE qmmmx_util
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              set_atomic_kind
   USE cell_types,                      ONLY: cell_copy,&
                                              cell_type,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit,&
                                              cp_logger_get_default_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_deallocate,&
                                              fist_neighbor_type
   USE fist_neighbor_lists,             ONLY: build_fist_neighbor_lists
   USE input_section_types,             ONLY: &
        section_vals_add_values, section_vals_duplicate, section_vals_get, &
        section_vals_get_subs_vals, section_vals_get_subs_vals3, section_vals_release, &
        section_vals_remove_values, section_vals_set_subs_vals, section_vals_type, &
        section_vals_val_get, section_vals_val_set, section_vals_write
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE qmmm_types,                      ONLY: qmmm_env_get
   USE qmmm_types_low,                  ONLY: force_mixing_label_QM_core,&
                                              force_mixing_label_QM_core_list,&
                                              force_mixing_label_QM_dynamics,&
                                              force_mixing_label_QM_dynamics_list,&
                                              force_mixing_label_buffer,&
                                              force_mixing_label_buffer_list,&
                                              force_mixing_label_none,&
                                              force_mixing_label_termination
   USE qmmm_util,                       ONLY: apply_qmmm_translate
   USE qmmmx_types,                     ONLY: qmmmx_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmmx_util'

   PUBLIC :: setup_force_mixing_qmmm_sections, update_force_mixing_labels, &
             apply_qmmmx_translate

CONTAINS

! **************************************************************************************************
!> \brief Apply translation to the full system in order to center the QM
!>      system into the QM box
!> \param qmmmx_env ...
!> \par History
!>      08.2007 created [tlaino] - Zurich University
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE apply_qmmmx_translate(qmmmx_env)
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_qmmmx_translate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      TYPE(cell_type), POINTER                           :: cell_core, cell_extended
      TYPE(cp_subsys_type), POINTER                      :: subsys_core, subsys_extended
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_core, particles_extended

      NULLIFY (cell_core, cell_extended)
      NULLIFY (subsys_core, subsys_extended)
      NULLIFY (particles_core, particles_extended)

      ! want to center extended, and make core consistent with that
      CALL apply_qmmm_translate(qmmmx_env%ext)

      ! translate core fist particles
      CALL qmmm_env_get(qmmmx_env%ext, subsys=subsys_extended)
      CALL cp_subsys_get(subsys_extended, cell=cell_extended)
      CALL qmmm_env_get(qmmmx_env%core, subsys=subsys_core)
      CALL cp_subsys_get(subsys_core, cell=cell_core)
      particles_extended => subsys_extended%particles%els
      particles_core => subsys_core%particles%els
      DO ip = 1, SIZE(particles_extended)
         particles_core(ip)%r = particles_extended(ip)%r
      END DO
      CALL cell_copy(cell_extended, cell_core)

      ! The core QM particles will be updated the regular call
      ! to apply_qmmm_translate() from within qmmm_calc_energy_force()

   END SUBROUTINE apply_qmmmx_translate

! **************************************************************************************************
!> \brief ...
!> \param subsys ...
!> \param qmmm_section ...
!> \param labels_changed ...
!> \par History
!>      02.2012 created [noam]
!> \author Noam Bernstein
! **************************************************************************************************
   SUBROUTINE update_force_mixing_labels(subsys, qmmm_section, labels_changed)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: qmmm_section
      LOGICAL, OPTIONAL                                  :: labels_changed

      CHARACTER(len=*), PARAMETER :: routineN = 'update_force_mixing_labels', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), POINTER      :: adaptive_exclude_molecules(:)
      INTEGER :: i_rep_section, i_rep_val, ip, max_n_qm, n_new, n_rep_exclude, n_rep_section, &
         n_rep_val, natoms, output_unit, QM_extended_seed_min_label_val
      INTEGER, ALLOCATABLE                               :: new_full_labels(:), orig_full_labels(:)
      INTEGER, POINTER                                   :: broken_bonds(:), cur_indices(:), &
                                                            cur_labels(:), mm_index_entry(:), &
                                                            new_indices(:), new_labels(:)
      LOGICAL                                            :: explicit, QM_extended_seed_is_core_list
      REAL(dp), ALLOCATABLE                              :: nearest_dist(:)
      REAL(dp), POINTER                                  :: r_buf(:), r_core(:), r_qm(:)
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(fist_neighbor_type), POINTER                  :: nlist
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: force_mixing_section, &
                                                            non_adaptive_section, qm_kind_section, &
                                                            restart_section

      output_unit = cp_logger_get_default_io_unit()

      IF (debug_this_module .AND. output_unit > 0) WRITE (output_unit, *) "BOB starting update_force_mixing_labels"
      ! get cur indices, labels
      force_mixing_section => section_vals_get_subs_vals3(qmmm_section, "FORCE_MIXING")
      CALL get_force_mixing_indices(force_mixing_section, cur_indices, cur_labels)
      IF (debug_this_module .AND. output_unit > 0) WRITE (output_unit, *) "BOB got cur_indices ", SIZE(cur_indices)
      IF (debug_this_module .AND. output_unit > 0) WRITE (output_unit, *) "BOB got cur_labels ", SIZE(cur_labels)

      ! read from input
      ![NB] breakable bonds will come from here, too
      NULLIFY (r_core, r_qm, r_buf, adaptive_exclude_molecules, broken_bonds)
      CALL section_vals_val_get(force_mixing_section, "R_CORE", r_vals=r_core)
      CALL section_vals_val_get(force_mixing_section, "R_QM", r_vals=r_qm)
      CALL section_vals_val_get(force_mixing_section, "QM_EXTENDED_SEED_IS_ONLY_CORE_LIST", &
                                l_val=QM_extended_seed_is_core_list)
      CALL section_vals_val_get(force_mixing_section, "R_BUF", r_vals=r_buf)
      CALL section_vals_val_get(force_mixing_section, "MAX_N_QM", i_val=max_n_qm)

      CALL section_vals_val_get(force_mixing_section, "ADAPTIVE_EXCLUDE_MOLECULES", n_rep_val=n_rep_exclude)
      IF (n_rep_exclude > 0) THEN
         CALL section_vals_val_get(force_mixing_section, "ADAPTIVE_EXCLUDE_MOLECULES", c_vals=adaptive_exclude_molecules)
      ENDIF
      ![NB] need to read real list from input
      ! should be 2xN_bb integer arrays, with (1,:) indices of inside atoms, and (2,:) indices of outside atoms
      ! maybe also breakable_bond_types, with _atomic numbers_ of inside/outside atoms?
      ! separate lists for core/buffer?

      ! get particles, molecules
      NULLIFY (particles, molecules)
      CALL cp_subsys_get(subsys=subsys, particles=particles, molecules_new=molecules)
      particle_set => particles%els
      molecule_set => molecules%els

      natoms = SIZE(particle_set)

      ! initialize new indices, labels, and new_full_labels
      NULLIFY (new_indices, new_labels)
      CALL reallocate(new_indices, 1, SIZE(cur_indices))
      CALL reallocate(new_labels, 1, SIZE(cur_labels))
      new_indices = 0
      new_labels = force_mixing_label_none
      ALLOCATE (new_full_labels(natoms))
      new_full_labels = force_mixing_label_none

      ! neighbor list for various hysteretic distance calls
      NULLIFY (cell)
      CALL cp_subsys_get(subsys, cell=cell)
      NULLIFY (nlist)
      CALL make_neighbor_list(force_mixing_section, subsys, cell, MAX(r_core(2), r_qm(2), r_buf(2)), nlist)

      ! create labels for core_list from QM_KIND
      NULLIFY (mm_index_entry)
      qm_kind_section => section_vals_get_subs_vals3(qmmm_section, "QM_KIND")
      CALL section_vals_get(qm_kind_section, n_repetition=n_rep_section)
      n_new = 0
      DO i_rep_section = 1, n_rep_section
         CALL section_vals_val_get(qm_kind_section, "MM_INDEX", i_rep_section=i_rep_section, n_rep_val=n_rep_val)
         DO i_rep_val = 1, n_rep_val
            CALL section_vals_val_get(qm_kind_section, "MM_INDEX", i_rep_section=i_rep_section, i_rep_val=i_rep_val, &
                                      i_vals=mm_index_entry)
            DO ip = 1, SIZE(mm_index_entry)
               CALL add_new_label(mm_index_entry(ip), force_mixing_label_QM_core_list, n_new, new_indices, new_labels, &
                                  new_full_labels, max_n_qm)
            END DO ! ip
         END DO ! i_rep_val
      END DO ! i_rep_section

      IF (debug_this_module .AND. output_unit > 0) THEN
         WRITE (output_unit, *) "BOB core_list new_indices ", new_indices(1:n_new)
         WRITE (output_unit, *) "BOB core_list new_labels ", new_labels(1:n_new)
      ENDIF

      ! create labels for non adaptive QM and buffer regions from *_NON_ADAPTIVE&QM_KIND sections
      non_adaptive_section => section_vals_get_subs_vals(qmmm_section, "FORCE_MIXING%QM_NON_ADAPTIVE", &
                                                         can_return_null=.TRUE.)
      IF (ASSOCIATED(non_adaptive_section)) THEN
         qm_kind_section => section_vals_get_subs_vals3(non_adaptive_section, "QM_KIND")
         CALL section_vals_get(qm_kind_section, n_repetition=n_rep_section)
         DO i_rep_section = 1, n_rep_section
            CALL section_vals_val_get(qm_kind_section, "MM_INDEX", i_rep_section=i_rep_section, n_rep_val=n_rep_val)
            DO i_rep_val = 1, n_rep_val
               CALL section_vals_val_get(qm_kind_section, "MM_INDEX", i_rep_section=i_rep_section, i_rep_val=i_rep_val, &
                                         i_vals=mm_index_entry)
               DO ip = 1, SIZE(mm_index_entry)
                  CALL add_new_label(mm_index_entry(ip), force_mixing_label_QM_dynamics_list, n_new, new_indices, new_labels, &
                                     new_full_labels, max_n_qm)
               END DO ! ip
            END DO ! i_rep_val
         END DO ! i_rep_section
      ENDIF
      IF (debug_this_module .AND. output_unit > 0) THEN
         WRITE (output_unit, *) "BOB core_list + non adaptive QM new_indices ", new_indices(1:n_new)
         WRITE (output_unit, *) "BOB core_list + non adaptive QM new_labels ", new_labels(1:n_new)
      ENDIF
      non_adaptive_section => section_vals_get_subs_vals(qmmm_section, "FORCE_MIXING%BUFFER_NON_ADAPTIVE", &
                                                         can_return_null=.TRUE.)
      IF (ASSOCIATED(non_adaptive_section)) THEN
         qm_kind_section => section_vals_get_subs_vals3(non_adaptive_section, "QM_KIND")
         CALL section_vals_get(qm_kind_section, n_repetition=n_rep_section)
         DO i_rep_section = 1, n_rep_section
            CALL section_vals_val_get(qm_kind_section, "MM_INDEX", i_rep_section=i_rep_section, n_rep_val=n_rep_val)
            DO i_rep_val = 1, n_rep_val
               CALL section_vals_val_get(qm_kind_section, "MM_INDEX", i_rep_section=i_rep_section, i_rep_val=i_rep_val, &
                                         i_vals=mm_index_entry)
               DO ip = 1, SIZE(mm_index_entry)
                  CALL add_new_label(mm_index_entry(ip), force_mixing_label_buffer_list, n_new, new_indices, new_labels, &
                                     new_full_labels, max_n_qm)
               END DO ! ip
            END DO ! i_rep_val
         END DO ! i_rep_section
      ENDIF

      IF (debug_this_module .AND. output_unit > 0) THEN
         WRITE (output_unit, *) "BOB core_list + non adaptive QM+buffer new_indices ", new_indices(1:n_new)
         WRITE (output_unit, *) "BOB core_list + non adaptive QM+buffer new_labels ", new_labels(1:n_new)
      ENDIF

      ! allocate and initialize full atom set labels for hysteretic loops
      ALLOCATE (nearest_dist(natoms))

      ! orig_full_labels is full array (natoms) with orig labels
      ALLOCATE (orig_full_labels(natoms))
      orig_full_labels = force_mixing_label_none
      orig_full_labels(cur_indices(:)) = cur_labels(:)

      ! hysteretically set QM core from QM_core_list and radii, whole molecule
      ![NB] need to replace all the whole molecule stuff with pad to breakable bonds. not quite done
      ! (need intra molecule bond info, which isn't available for QM molecules yet)

      ! add core using hysteretic selection(core_list, r_core) + unbreakable bonds
      CALL add_layer_hysteretically( &
         nlist, particle_set, cell, nearest_dist, &
         orig_full_labels, new_full_labels, n_new, new_indices, new_labels, &
         force_mixing_label_QM_core_list, force_mixing_label_QM_core_list, force_mixing_label_QM_core, r_core, &
         max_n_qm, adaptive_exclude_molecules, molecule_set, broken_bonds)
      ![NB] should actually pass this back for making link sections?
      DEALLOCATE (broken_bonds)

      IF (debug_this_module .AND. output_unit > 0) THEN
         WRITE (output_unit, *) "BOB core new_indices ", new_indices(1:n_new)
         WRITE (output_unit, *) "BOB core new_labels ", new_labels(1:n_new)
      ENDIF

      ![NB] need more sophisticated QM extended, buffer rules

      ! add QM using hysteretic selection (core_list, r_qm) + unbreakable bonds
      IF (debug_this_module .AND. output_unit > 0) &
         WRITE (output_unit, *) "BOB QM_extended_seed_is_core_list ", QM_extended_seed_is_core_list
      IF (QM_extended_seed_is_core_list) THEN
         QM_extended_seed_min_label_val = force_mixing_label_QM_core_list
      ELSE ! QM region seed is all of core, not just core list + unbreakable bonds
         QM_extended_seed_min_label_val = force_mixing_label_QM_core
      ENDIF
      CALL add_layer_hysteretically(nlist, particle_set, cell, nearest_dist, &
                                    orig_full_labels, new_full_labels, n_new, new_indices, new_labels, &
                                    QM_extended_seed_min_label_val, force_mixing_label_QM_core_list, &
                                    force_mixing_label_QM_dynamics, r_qm, &
                                    max_n_qm, adaptive_exclude_molecules, molecule_set)

      IF (debug_this_module .AND. output_unit > 0) THEN
         WRITE (output_unit, *) "BOB extended new_indices ", new_indices(1:n_new)
         WRITE (output_unit, *) "BOB extended new_labels ", new_labels(1:n_new)
      ENDIF

      ! add buffer using hysteretic selection (>= QM extended, r_buf) + unbreakable bonds
      CALL add_layer_hysteretically( &
         nlist, particle_set, cell, nearest_dist, &
         orig_full_labels, new_full_labels, n_new, new_indices, new_labels, &
         force_mixing_label_QM_dynamics, force_mixing_label_QM_core_list, force_mixing_label_buffer, r_buf, &
         max_n_qm, adaptive_exclude_molecules, molecule_set, broken_bonds)
      ![NB] should actually pass this back for making link sections?
      DEALLOCATE (broken_bonds)

      IF (debug_this_module .AND. output_unit > 0) THEN
         WRITE (output_unit, *) "BOB buffer new_indices ", new_indices(1:n_new)
         WRITE (output_unit, *) "BOB buffer new_labels ", new_labels(1:n_new)
      ENDIF

      DEALLOCATE (nearest_dist)

      IF (PRESENT(labels_changed)) labels_changed = ANY(new_full_labels /= orig_full_labels)

      DEALLOCATE (orig_full_labels)
      DEALLOCATE (new_full_labels)

      ! reduce new indices, labels to actually used size
      CALL reallocate(new_indices, 1, n_new)
      CALL reallocate(new_labels, 1, n_new)

      ! save info in input structure
      restart_section => section_vals_get_subs_vals(qmmm_section, "FORCE_MIXING%RESTART_INFO")
      CALL section_vals_get(restart_section, explicit=explicit)
      IF (explicit) CALL section_vals_remove_values(restart_section)
      CALL section_vals_val_set(restart_section, "INDICES", i_vals_ptr=new_indices)
      CALL section_vals_val_set(restart_section, "LABELS", i_vals_ptr=new_labels)

      DEALLOCATE (cur_indices, cur_labels)
      CALL fist_neighbor_deallocate(nlist)

      ![NB] perhap be controlled by some &PRINT section?
      CALL cp_subsys_get(subsys, para_env=para_env)
      IF (para_env%ionode .AND. output_unit > 0) THEN
         WRITE (unit=output_unit, fmt='(A,A,I6,A,I5,A,I5,A,I5)') &
            "QMMM FORCE MIXING final count (not including links): ", &
            " N_QM core_list ", COUNT(new_labels == force_mixing_label_QM_core_list), &
            " N_QM core ", COUNT(new_labels == force_mixing_label_QM_core), &
            " N_QM extended ", COUNT(new_labels == force_mixing_label_QM_dynamics .OR. &
                                     new_labels == force_mixing_label_QM_dynamics_list), &
            " N_QM buffered ", COUNT(new_labels == force_mixing_label_buffer .OR. &
                                     new_labels == force_mixing_label_buffer_list)
      ENDIF

   END SUBROUTINE update_force_mixing_labels

! **************************************************************************************************
!> \brief ...
!> \param ip ...
!> \param label ...
!> \param n_new ...
!> \param new_indices ...
!> \param new_labels ...
!> \param new_full_labels ...
!> \param max_n_qm ...
! **************************************************************************************************
   SUBROUTINE add_new_label(ip, label, n_new, new_indices, new_labels, new_full_labels, max_n_qm)
      INTEGER                                            :: ip, label, n_new
      INTEGER, POINTER                                   :: new_indices(:), new_labels(:)
      INTEGER                                            :: new_full_labels(:), max_n_qm

      CHARACTER(len=*), PARAMETER :: routineN = 'add_new_label', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, old_index

      IF (new_full_labels(ip) > force_mixing_label_none) THEN ! already marked, just change mark
         old_index = -1
         DO i = 1, n_new
            IF (new_indices(i) == ip) THEN
               old_index = i
               EXIT
            ENDIF
         END DO
         IF (old_index <= 0) &
            CALL cp_abort(cp__l("qmmmx_util.F",388), &
                          "add_new_label found atom with a label "// &
                          "already set, but not in new_indices array")
         new_labels(old_index) = label
      ELSE
         n_new = n_new+1
         IF (n_new > max_n_qm) &
            CALL cp_abort(cp__l("qmmmx_util.F",395), &
                          "add_new_label tried to add more atoms "// &
                          "than allowed by &FORCE_MIXING&MAX_N_QM!")
         IF (n_new > SIZE(new_indices)) CALL reallocate(new_indices, 1, n_new+9)
         IF (n_new > SIZE(new_labels)) CALL reallocate(new_labels, 1, n_new+9)
         new_indices(n_new) = ip
         new_labels(n_new) = label
      ENDIF
      new_full_labels(ip) = label
   END SUBROUTINE add_new_label

! **************************************************************************************************
!> \brief ...
!> \param nlist ...
!> \param particle_set ...
!> \param cell ...
!> \param nearest_dist ...
!> \param orig_full_labels ...
!> \param new_full_labels ...
!> \param n_new ...
!> \param new_indices ...
!> \param new_labels ...
!> \param seed_min_label_val ...
!> \param seed_max_label_val ...
!> \param set_label_val ...
!> \param r_inout ...
!> \param max_n_qm ...
!> \param adaptive_exclude_molecules ...
!> \param molecule_set ...
!> \param broken_bonds ...
! **************************************************************************************************
   SUBROUTINE add_layer_hysteretically(nlist, particle_set, cell, nearest_dist, &
                                       orig_full_labels, new_full_labels, n_new, new_indices, new_labels, &
                                       seed_min_label_val, seed_max_label_val, set_label_val, r_inout, max_n_qm, &
                                       adaptive_exclude_molecules, molecule_set, broken_bonds)
      TYPE(fist_neighbor_type), POINTER                  :: nlist
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      REAL(dp)                                           :: nearest_dist(:)
      INTEGER                                            :: orig_full_labels(:), new_full_labels(:), &
                                                            n_new
      INTEGER, POINTER                                   :: new_indices(:), new_labels(:)
      INTEGER                                            :: seed_min_label_val, seed_max_label_val, &
                                                            set_label_val
      REAL(dp)                                           :: r_inout(2)
      INTEGER                                            :: max_n_qm
      CHARACTER(len=*), POINTER                          :: adaptive_exclude_molecules(:)
      TYPE(molecule_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_set
      INTEGER, OPTIONAL, POINTER                         :: broken_bonds(:)

      CHARACTER(len=*), PARAMETER :: routineN = 'add_layer_hysteretically', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_ind, im, im_exclude, ip, ipair, &
                                                            ipairkind, j_ind, output_unit
      LOGICAL :: adaptive_exclude, i_in_new_seed, i_outside_new_seed, j_in_new_seed, &
         j_outside_new_seed, molec_in_inner, molec_in_outer
      REAL(dp)                                           :: r_ij(3), r_ij_mag

      output_unit = cp_logger_get_default_unit_nr()

      IF (debug_this_module .AND. output_unit > 0) WRITE (output_unit, *) "BOB adding hysteretically seed ", &
         seed_min_label_val, seed_max_label_val, " set ", set_label_val, " r ", r_inout
      ! calculate nearest dist from each atom outside of new seed to nearest atom inside of new seed
      nearest_dist = HUGE(1.0_dp)
      ! loop over pairs of all kinds in random order
      DO ipairkind = 1, SIZE(nlist%neighbor_kind_pairs)
         DO ipair = 1, nlist%neighbor_kind_pairs(ipairkind)%npairs

            i_ind = nlist%neighbor_kind_pairs(ipairkind)%list(1, ipair)
            j_ind = nlist%neighbor_kind_pairs(ipairkind)%list(2, ipair)

            i_in_new_seed = (new_full_labels(i_ind) >= seed_min_label_val .AND. new_full_labels(i_ind) <= seed_max_label_val)
            i_outside_new_seed = (new_full_labels(i_ind) < seed_min_label_val)
            j_in_new_seed = (new_full_labels(j_ind) >= seed_min_label_val .AND. new_full_labels(j_ind) <= seed_max_label_val)
            j_outside_new_seed = (new_full_labels(j_ind) < seed_min_label_val)

            IF ((i_in_new_seed .AND. j_outside_new_seed) .OR. (j_in_new_seed .AND. i_outside_new_seed)) THEN
               r_ij = pbc(particle_set(i_ind)%r-particle_set(j_ind)%r, cell)
               r_ij_mag = SQRT(SUM(r_ij**2))
               IF (i_in_new_seed .AND. j_outside_new_seed .AND. (r_ij_mag < nearest_dist(j_ind))) THEN
                  nearest_dist(j_ind) = r_ij_mag
               ENDIF
               IF (j_in_new_seed .AND. i_outside_new_seed .AND. (r_ij_mag < nearest_dist(i_ind))) THEN
                  nearest_dist(i_ind) = r_ij_mag
               ENDIF
            ENDIF

         END DO
      END DO

      ![NB] this is whole molecule.  Should be replaced with labeling of individual atoms +
      ! pad_to_breakable_bonds (below), but QM molecule bond information isn't available yet
      DO im = 1, SIZE(molecule_set)
         ! molecule_set(im)%first_atom,molecule_set(im)%last_atom
         IF (ASSOCIATED(adaptive_exclude_molecules)) THEN
            adaptive_exclude = .FALSE.
            DO im_exclude = 1, SIZE(adaptive_exclude_molecules)
               IF (TRIM(molecule_set(im)%molecule_kind%name) == TRIM(adaptive_exclude_molecules(im_exclude)) .OR. &
                   TRIM(molecule_set(im)%molecule_kind%name) == '_QM_'//TRIM(adaptive_exclude_molecules(im_exclude))) &
                  adaptive_exclude = .TRUE.
            END DO
            IF (adaptive_exclude) CYCLE
         ENDIF
         molec_in_inner = ANY(nearest_dist(molecule_set(im)%first_atom:molecule_set(im)%last_atom) <= r_inout(1))
         molec_in_outer = ANY(nearest_dist(molecule_set(im)%first_atom:molecule_set(im)%last_atom) <= r_inout(2))
         IF (molec_in_inner) THEN
            DO ip = molecule_set(im)%first_atom, molecule_set(im)%last_atom
               ! labels are being rebuild from scratch, so never overwrite new label that's higher level
               IF (new_full_labels(ip) < set_label_val) &
                  CALL add_new_label(ip, set_label_val, n_new, new_indices, new_labels, new_full_labels, max_n_qm)
            END DO
         ELSE IF (molec_in_outer) THEN
            IF (ANY(orig_full_labels(molecule_set(im)%first_atom:molecule_set(im)%last_atom) >= set_label_val)) THEN
               DO ip = molecule_set(im)%first_atom, molecule_set(im)%last_atom
                  ! labels are being rebuild from scratch, so never overwrite new label that's higher level
                  IF (new_full_labels(ip) < set_label_val) &
                     CALL add_new_label(ip, set_label_val, n_new, new_indices, new_labels, new_full_labels, max_n_qm)
               END DO
            ENDIF
         ENDIF
      END DO
      IF (PRESENT(broken_bonds)) CALL reallocate(broken_bonds, 1, 0)

   END SUBROUTINE add_layer_hysteretically

! **************************************************************************************************
!> \brief ...
!> \param force_mixing_section ...
!> \param subsys ...
!> \param cell ...
!> \param r_max ...
!> \param nlist ...
! **************************************************************************************************
   SUBROUTINE make_neighbor_list(force_mixing_section, subsys, cell, r_max, nlist)
      TYPE(section_vals_type), POINTER                   :: force_mixing_section
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(cell_type), POINTER                           :: cell
      REAL(dp)                                           :: r_max
      TYPE(fist_neighbor_type), POINTER                  :: nlist

      CHARACTER(len=*), PARAMETER :: routineN = 'make_neighbor_list', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: kind_name
      CHARACTER(LEN=default_string_length), POINTER      :: kind_name_a(:)
      INTEGER                                            :: ik
      LOGICAL                                            :: skip_kind
      REAL(dp), ALLOCATABLE                              :: r_max_a(:, :), r_minsq_a(:, :)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      ALLOCATE (r_max_a(SIZE(subsys%atomic_kinds%els), SIZE(subsys%atomic_kinds%els)))
      ALLOCATE (r_minsq_a(SIZE(subsys%atomic_kinds%els), SIZE(subsys%atomic_kinds%els)))
      r_max_a = r_max
      r_minsq_a = EPSILON(1.0_dp)

      ! save kind names
      ALLOCATE (kind_name_a(SIZE(subsys%atomic_kinds%els)))
      DO ik = 1, SIZE(subsys%atomic_kinds%els)
         atomic_kind => subsys%atomic_kinds%els(ik)
         CALL get_atomic_kind(atomic_kind=atomic_kind, name=kind_name)
         kind_name_a(ik) = kind_name
      END DO

      ! overwrite kind names so that none are QM, and so excluding QM-QM interactions
      ! (which is not what we want) will not happen
      DO ik = 1, SIZE(subsys%atomic_kinds%els)
         atomic_kind => subsys%atomic_kinds%els(ik)
         CALL get_atomic_kind(atomic_kind=atomic_kind, name=kind_name)
         ! when atom is QM atom, kind_name is replaced with original
         ! mm kind name, and return status is logical .TRUE.
         skip_kind = qmmm_ff_precond_only_qm(kind_name)
         CALL set_atomic_kind(atomic_kind=atomic_kind, name=kind_name)
      END DO

      NULLIFY (nlist)
      CALL build_fist_neighbor_lists(subsys%atomic_kinds%els, subsys%particles%els, &
                                     cell=cell, r_max=r_max_a, r_minsq=r_minsq_a, &
                                     ei_scale14=1.0_dp, vdw_scale14=1.0_dp, nonbonded=nlist, &
                                     para_env=subsys%para_env, build_from_scratch=.TRUE., geo_check=.FALSE., &
                                     mm_section=force_mixing_section)

      DEALLOCATE (r_max_a, r_minsq_a)

      ! restore kind names
      DO ik = 1, SIZE(subsys%atomic_kinds%els)
         CALL set_atomic_kind(atomic_kind=atomic_kind, name=kind_name_a(ik))
      END DO
      DEALLOCATE (kind_name_a)

   END SUBROUTINE make_neighbor_list

! **************************************************************************************************
!> \brief ...
!> \param subsys ...
!> \param qmmm_section ...
!> \param qmmm_core_section ...
!> \param qmmm_extended_section ...
!> \par History
!>      02.2012 created [noam]
!> \author Noam Bernstein
! **************************************************************************************************
   SUBROUTINE setup_force_mixing_qmmm_sections(subsys, qmmm_section, qmmm_core_section, qmmm_extended_section)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: qmmm_section, qmmm_core_section, &
                                                            qmmm_extended_section

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_force_mixing_qmmm_sections', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length), POINTER      :: elem_mapping(:, :), elem_mapping_entry(:)
      INTEGER :: delta_charge, i_rep_section_core, i_rep_section_extended, i_rep_val_core, &
         i_rep_val_extended, ielem, ip, n_elements, output_unit
      INTEGER, POINTER                                   :: cur_indices(:), cur_labels(:)
      LOGICAL                                            :: mapped, new_element_core, &
                                                            new_element_extended
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      TYPE(section_vals_type), POINTER                   :: buffer_non_adaptive_section, &
                                                            dup_link_section, &
                                                            force_mixing_section, link_section, &
                                                            qm_kind_section

      NULLIFY (qmmm_core_section, qmmm_extended_section)
      output_unit = cp_logger_get_default_unit_nr()

      ! create new qmmm sections for core and extended
      CALL section_vals_duplicate(qmmm_section, qmmm_core_section)
      CALL section_vals_duplicate(qmmm_section, qmmm_extended_section)

      ! remove LINKs (specified by user for core) from extended
      link_section => section_vals_get_subs_vals(qmmm_extended_section, "LINK", can_return_null=.TRUE.)
      IF (ASSOCIATED(link_section)) THEN
         CALL section_vals_remove_values(link_section)
      END IF
      ! for LINKs to be added to extended
      buffer_non_adaptive_section => section_vals_get_subs_vals(qmmm_extended_section, "FORCE_MIXING%BUFFER_NON_ADAPTIVE", &
                                                                can_return_null=.TRUE.)
      link_section => section_vals_get_subs_vals(buffer_non_adaptive_section, "LINK", can_return_null=.TRUE.)
      IF (ASSOCIATED(link_section)) THEN
         NULLIFY (dup_link_section)
         CALL section_vals_duplicate(link_section, dup_link_section)
         CALL section_vals_set_subs_vals(qmmm_extended_section, "LINK", dup_link_section)
         CALL section_vals_release(dup_link_section)
      ENDIF

      IF (debug_this_module .AND. output_unit > 0) THEN
         link_section => section_vals_get_subs_vals(qmmm_core_section, "LINK", can_return_null=.TRUE.)
         WRITE (output_unit, *) "core section has LINKs ", ASSOCIATED(link_section)
         CALL section_vals_write(link_section, unit_nr=6)
         link_section => section_vals_get_subs_vals(qmmm_extended_section, "LINK", can_return_null=.TRUE.)
         WRITE (output_unit, *) "extended section has LINKs ", ASSOCIATED(link_section)
         CALL section_vals_write(link_section, unit_nr=6)
      ENDIF

      force_mixing_section => section_vals_get_subs_vals(qmmm_section, "FORCE_MIXING")

      ! get QM_KIND_ELEMENT_MAPPING
      CALL section_vals_val_get(force_mixing_section, "QM_KIND_ELEMENT_MAPPING", n_rep_val=n_elements)
      ALLOCATE (elem_mapping(2, n_elements))
      DO ielem = 1, n_elements
         CALL section_vals_val_get(force_mixing_section, "QM_KIND_ELEMENT_MAPPING", i_rep_val=ielem, c_vals=elem_mapping_entry)
         elem_mapping(1:2, ielem) = elem_mapping_entry(1:2)
      END DO

      ! get CUR_INDICES, CUR_LABELS
      CALL get_force_mixing_indices(force_mixing_section, cur_indices, cur_labels)
      IF (SIZE(cur_indices) <= 0) &
         CALL cp__b("qmmmx_util.F",663,"cur_indices is empty, found no QM atoms")

      IF (debug_this_module .AND. output_unit > 0) THEN
         WRITE (output_unit, *) "cur_indices ", cur_indices
         WRITE (output_unit, *) "cur_labels ", cur_labels
      ENDIF

      ! loop through elements and atoms, and set up new QM_KIND sections
      particles => subsys%particles%els

      DO ip = 1, SIZE(cur_indices)
         IF (cur_labels(ip) > force_mixing_label_none .AND. cur_labels(ip) < force_mixing_label_QM_core_list .AND. &
             cur_labels(ip) /= force_mixing_label_termination) THEN
            mapped = .FALSE.
            DO ielem = 1, n_elements
               IF (TRIM(particles(cur_indices(ip))%atomic_kind%element_symbol) == TRIM(elem_mapping(1, ielem))) THEN
                  mapped = .TRUE.
                  EXIT
               ENDIF
            END DO
            IF (.NOT. mapped) &
               CALL cp_abort(cp__l("qmmmx_util.F",684), &
                             "Force-mixing failed to find QM_KIND mapping for atom of type "// &
                             TRIM(particles(cur_indices(ip))%atomic_kind%element_symbol)// &
                             "! ")
         ENDIF
      END DO

      ! pre-existing QM_KIND section specifies list of core atom
      qm_kind_section => section_vals_get_subs_vals3(qmmm_section, "QM_KIND")
      CALL section_vals_get(qm_kind_section, n_repetition=i_rep_section_core)
      IF (i_rep_section_core <= 0) &
         CALL cp_abort(cp__l("qmmmx_util.F",695), &
                       "Force-mixing QM didn't find any QM_KIND sections, "// &
                       "so no core specified!")
      i_rep_section_extended = i_rep_section_core
      DO ielem = 1, n_elements
         new_element_core = .TRUE.
         new_element_extended = .TRUE.
         DO ip = 1, SIZE(cur_indices) ! particles with label
            IF (TRIM(particles(cur_indices(ip))%atomic_kind%element_symbol) /= TRIM(elem_mapping(1, ielem))) CYCLE
            ! extended
            ! if current particle is some sort of QM atom, and not in core list
            ! (those the user gave explicit QM_KIND sections for), and not a
            ! termination atom, need to make a QM_KIND section for it
            IF (cur_labels(ip) > force_mixing_label_none .AND. &
                cur_labels(ip) /= force_mixing_label_QM_core_list .AND. &
                cur_labels(ip) /= force_mixing_label_termination) THEN
               qm_kind_section => section_vals_get_subs_vals3(qmmm_extended_section, "QM_KIND")
               IF (new_element_extended) THEN ! add new QM_KIND section for this element
                  i_rep_section_extended = i_rep_section_extended+1
                  CALL section_vals_add_values(qm_kind_section)
                  CALL section_vals_val_set(qm_kind_section, "_SECTION_PARAMETERS_", i_rep_section=i_rep_section_extended, &
                                            c_val=elem_mapping(2, ielem))
                  i_rep_val_extended = 0
                  new_element_extended = .FALSE.
               ENDIF
               i_rep_val_extended = i_rep_val_extended+1
               CALL section_vals_val_set(qm_kind_section, "MM_INDEX", i_rep_section=i_rep_section_extended, &
                                         i_rep_val=i_rep_val_extended, i_val=cur_indices(ip))
            ENDIF ! is a non-termination QM atom

            ! core
            ! if current particle is a core QM atom, and not in core list (those the user
            ! gave explicit QM_KIND sections for, need to make a QM_KIND section for it
            IF (cur_labels(ip) == force_mixing_label_QM_core) THEN
               qm_kind_section => section_vals_get_subs_vals3(qmmm_core_section, "QM_KIND")
               IF (new_element_core) THEN ! add new QM_KIND section for this element
                  i_rep_section_core = i_rep_section_core+1
                  CALL section_vals_add_values(qm_kind_section)
                  CALL section_vals_val_set(qm_kind_section, "_SECTION_PARAMETERS_", i_rep_section=i_rep_section_core, &
                                            c_val=elem_mapping(2, ielem))
                  i_rep_val_core = 0
                  new_element_core = .FALSE.
               ENDIF
               i_rep_val_core = i_rep_val_core+1
               CALL section_vals_val_set(qm_kind_section, "MM_INDEX", i_rep_section=i_rep_section_core, &
                                         i_rep_val=i_rep_val_core, i_val=cur_indices(ip))
            ENDIF ! is a non-termination QM atom

         END DO ! atom index ip
      END DO ! element index ielem

      CALL section_vals_val_get(force_mixing_section, "EXTENDED_DELTA_CHARGE", i_val=delta_charge)
      CALL section_vals_val_set(qmmm_extended_section, "DELTA_CHARGE", i_val=delta_charge)

      ![NB] check
      DEALLOCATE (elem_mapping, cur_indices, cur_labels)

      IF (debug_this_module .AND. output_unit > 0) THEN
         WRITE (output_unit, *) "qmmm_core_section"
         CALL section_vals_write(qmmm_core_section, unit_nr=6)
         WRITE (output_unit, *) "qmmm_extended_section"
         CALL section_vals_write(qmmm_extended_section, unit_nr=6)
      ENDIF

   END SUBROUTINE setup_force_mixing_qmmm_sections

! **************************************************************************************************
!> \brief ...
!> \param force_mixing_section ...
!> \param indices ...
!> \param labels ...
! **************************************************************************************************
   SUBROUTINE get_force_mixing_indices(force_mixing_section, indices, labels)
      TYPE(section_vals_type), POINTER                   :: force_mixing_section
      INTEGER, POINTER                                   :: indices(:), labels(:)

      CHARACTER(len=*), PARAMETER :: routineN = 'get_force_mixing_indices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_rep_val, n_indices, n_labels, n_reps
      INTEGER, POINTER                                   :: indices_entry(:), labels_entry(:)
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: restart_section

      NULLIFY (indices, labels)
      restart_section => section_vals_get_subs_vals(force_mixing_section, "RESTART_INFO")
      CALL section_vals_get(restart_section, explicit=explicit)
      IF (.NOT. explicit) THEN ! no old indices, labels, return empty arrays
         ALLOCATE (indices(0))
         ALLOCATE (labels(0))
         RETURN
      ENDIF

      ![NB] maybe switch to reallocatable array
      CALL section_vals_val_get(restart_section, "INDICES", n_rep_val=n_reps)
      n_indices = 0
      DO i_rep_val = 1, n_reps
         CALL section_vals_val_get(restart_section, "INDICES", &
                                   i_rep_val=i_rep_val, i_vals=indices_entry)
         n_indices = n_indices+SIZE(indices_entry)
      END DO
      ALLOCATE (indices(n_indices))
      n_indices = 0
      DO i_rep_val = 1, n_reps
         CALL section_vals_val_get(restart_section, "INDICES", &
                                   i_rep_val=i_rep_val, i_vals=indices_entry)
         indices(n_indices+1:n_indices+SIZE(indices_entry)) = indices_entry
         n_indices = n_indices+SIZE(indices_entry)
      END DO

      CALL section_vals_val_get(restart_section, "LABELS", n_rep_val=n_reps)
      n_labels = 0
      DO i_rep_val = 1, n_reps
         CALL section_vals_val_get(restart_section, "LABELS", &
                                   i_rep_val=i_rep_val, i_vals=labels_entry)
         n_labels = n_labels+SIZE(labels_entry)
      END DO
      ALLOCATE (labels(n_labels))
      n_labels = 0
      DO i_rep_val = 1, n_reps
         CALL section_vals_val_get(restart_section, "LABELS", &
                                   i_rep_val=i_rep_val, i_vals=labels_entry)
         labels(n_labels+1:n_labels+SIZE(labels_entry)) = labels_entry
         n_labels = n_labels+SIZE(labels_entry)
      END DO

      IF (n_indices /= n_labels) &
         CALL cp__b("qmmmx_util.F",822,"got unequal numbers of force_mixing indices and labels!")
   END SUBROUTINE get_force_mixing_indices

END MODULE qmmmx_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A collection of methods to treat the QM/MM links
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_links_methods

   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: add_set_type,&
                                              qmmm_env_qm_type,&
                                              qmmm_imomm_link_type,&
                                              qmmm_links_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_links_methods'
   PUBLIC ::  qmmm_link_Imomm_coord, &
             qmmm_link_Imomm_forces, &
             qmmm_added_chrg_coord, &
             qmmm_added_chrg_forces

CONTAINS

! **************************************************************************************************
!> \brief correct the position for qm/mm IMOMM link type
!> \param qmmm_links ...
!> \param particles ...
!> \param qm_atom_index ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_link_Imomm_coord(qmmm_links, particles, qm_atom_index)
      TYPE(qmmm_links_type), POINTER                     :: qmmm_links
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_link_Imomm_coord', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ilink, ip, ip_mm, ip_qm, mm_index, &
                                                            n_imomm, qm_index
      REAL(KIND=dp)                                      :: alpha
      TYPE(qmmm_imomm_link_type), POINTER                :: my_link

      n_imomm = SIZE(qmmm_links%imomm)
      IF(.NOT.(n_imomm /= 0))CALL cp__a("qmmm_links_methods.F",58)
      DO ilink = 1, n_imomm
         my_link => qmmm_links%imomm(ilink)%link
         qm_index = my_link%qm_index
         mm_index = my_link%mm_index
         alpha = 1.0_dp/my_link%alpha
         DO ip = 1, SIZE(qm_atom_index)
            IF (qm_atom_index(ip) == qm_index) EXIT
         END DO
         IF (ip == SIZE(qm_atom_index)+1) &
            CALL cp_abort(cp__l("qmmm_links_methods.F",68), &
                          "QM atom index ("//cp_to_string(qm_index)//") specified in the LINK section nr.("// &
                          cp_to_string(ilink)//") is not defined as a QM atom! Please inspect your QM_KIND sections. ")
         ip_qm = ip
         DO ip = 1, SIZE(qm_atom_index)
            IF (qm_atom_index(ip) == mm_index) EXIT
         END DO
         IF (ip == SIZE(qm_atom_index)+1) &
            CALL cp_abort(cp__l("qmmm_links_methods.F",76), &
                          "Error in setting up the MM atom index ("//cp_to_string(mm_index)// &
                          ") specified in the LINK section nr.("//cp_to_string(ilink)//"). Please report this bug! ")
         ip_mm = ip
         particles(ip_mm)%r = alpha*particles(ip_mm)%r+(1.0_dp-alpha)*particles(ip_qm)%r
      END DO

   END SUBROUTINE qmmm_link_Imomm_coord

! **************************************************************************************************
!> \brief correct the forces for qm/mm IMOMM link type
!> \param qmmm_links ...
!> \param particles_qm ...
!> \param qm_atom_index ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_link_Imomm_forces(qmmm_links, particles_qm, qm_atom_index)
      TYPE(qmmm_links_type), POINTER                     :: qmmm_links
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_qm
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_link_Imomm_forces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ilink, ip, ip_mm, ip_qm, mm_index, &
                                                            n_imomm, qm_index
      REAL(KIND=dp)                                      :: alpha
      TYPE(qmmm_imomm_link_type), POINTER                :: my_link

      n_imomm = SIZE(qmmm_links%imomm)
      IF(.NOT.(n_imomm /= 0))CALL cp__a("qmmm_links_methods.F",108)
      DO ilink = 1, n_imomm
         my_link => qmmm_links%imomm(ilink)%link
         qm_index = my_link%qm_index
         mm_index = my_link%mm_index
         alpha = 1.0_dp/my_link%alpha
         DO ip = 1, SIZE(qm_atom_index)
            IF (qm_atom_index(ip) == qm_index) EXIT
         END DO
         IF (ip == SIZE(qm_atom_index)+1) &
            CALL cp_abort(cp__l("qmmm_links_methods.F",118), &
                          "QM atom index ("//cp_to_string(qm_index)//") specified in the LINK section nr.("// &
                          cp_to_string(ilink)//") is not defined as a QM atom! Please inspect your QM_KIND sections. ")
         ip_qm = ip
         DO ip = 1, SIZE(qm_atom_index)
            IF (qm_atom_index(ip) == mm_index) EXIT
         END DO
         IF (ip == SIZE(qm_atom_index)+1) &
            CALL cp_abort(cp__l("qmmm_links_methods.F",126), &
                          "Error in setting up the MM atom index ("//cp_to_string(mm_index)// &
                          ") specified in the LINK section nr.("//cp_to_string(ilink)//"). Please report this bug! ")
         ip_mm = ip
         particles_qm(ip_qm)%f = particles_qm(ip_qm)%f+particles_qm(ip_mm)%f*(1.0_dp-alpha)
         particles_qm(ip_mm)%f = particles_qm(ip_mm)%f*alpha
      END DO

   END SUBROUTINE qmmm_link_Imomm_forces

! **************************************************************************************************
!> \brief correct the position for added charges in qm/mm link scheme
!> \param qmmm_env ...
!> \param particles ...
!> \par History
!>      01.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_added_chrg_coord(qmmm_env, particles)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_added_chrg_coord', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, Index1, Index2
      REAL(KIND=dp)                                      :: alpha
      TYPE(add_set_type), POINTER                        :: added_charges

      added_charges => qmmm_env%added_charges

      DO i = 1, added_charges%num_mm_atoms
         Index1 = added_charges%add_env(i)%Index1
         Index2 = added_charges%add_env(i)%Index1
         alpha = added_charges%add_env(i)%alpha
         added_charges%added_particles(i)%r = alpha*particles(Index1)%r+(1.0_dp-alpha)*particles(Index2)%r
      END DO

   END SUBROUTINE qmmm_added_chrg_coord

! **************************************************************************************************
!> \brief correct the forces due to the  added charges in qm/mm link scheme
!> \param qmmm_env ...
!> \param particles ...
!> \par History
!>      01.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_added_chrg_forces(qmmm_env, particles)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_added_chrg_forces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, Index1, Index2
      REAL(KIND=dp)                                      :: alpha
      TYPE(add_set_type), POINTER                        :: added_charges

      added_charges => qmmm_env%added_charges

      DO i = 1, added_charges%num_mm_atoms
         Index1 = added_charges%add_env(i)%Index1
         Index2 = added_charges%add_env(i)%Index1
         alpha = added_charges%add_env(i)%alpha
         particles(Index1)%f = particles(Index1)%f+alpha*added_charges%added_particles(i)%f
         particles(Index2)%f = particles(Index2)%f+(1.0_dp-alpha)*added_charges%added_particles(i)%f
      END DO

   END SUBROUTINE qmmm_added_chrg_forces

END MODULE qmmm_links_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the MM potential by collocating the primitive Gaussian
!>      functions (pgf)
!> \par History
!>      7.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE mm_collocate_potential
   USE ao_util,                         ONLY: exp_radius
   USE cell_types,                      ONLY: cell_type
   USE cube_utils,                      ONLY: cube_info_type,&
                                              return_cube
   USE kinds,                           ONLY: dp
   USE pw_types,                        ONLY: pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mm_collocate_potential'

   PUBLIC :: collocate_gf_rspace_NoPBC, &
             integrate_gf_rspace_NoPBC
!***
CONTAINS

! **************************************************************************************************
!> \brief Main driver to collocate gaussian functions on grid
!>      without using periodic boundary conditions (NoPBC)
!> \param zetp ...
!> \param rp ...
!> \param scale ...
!> \param W ...
!> \param pwgrid ...
!> \param cube_info ...
!> \param eps_mm_rspace ...
!> \param xdat ...
!> \param ydat ...
!> \param zdat ...
!> \param bo2 ...
!> \param n_rep_real ...
!> \param mm_cell ...
!> \par History
!>      07.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE collocate_gf_rspace_NoPBC(zetp, rp, scale, W, pwgrid, cube_info, &
                                        eps_mm_rspace, xdat, ydat, zdat, bo2, n_rep_real, mm_cell)
      REAL(KIND=dp), INTENT(IN)                          :: zetp
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rp
      REAL(KIND=dp), INTENT(IN)                          :: scale, W
      TYPE(pw_type), POINTER                             :: pwgrid
      TYPE(cube_info_type), INTENT(IN)                   :: cube_info
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xdat, ydat, zdat
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bo2
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n_rep_real
      TYPE(cell_type), POINTER                           :: mm_cell

      INTEGER                                            :: ig, ix, iy, iz, xlb, xub, ylb, yub, zlb, &
                                                            zub
      INTEGER, DIMENSION(2, 3)                           :: bo, gbo
      INTEGER, DIMENSION(3)                              :: cubecenter, lb_cube, ub_cube
      INTEGER, DIMENSION(:), POINTER                     :: sphere_bounds
      REAL(KIND=dp)                                      :: radius, rpg, xap, yap, zap
      REAL(KIND=dp), DIMENSION(3)                        :: dr, my_shift, rpl
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid

      radius = exp_radius(0, zetp, eps_mm_rspace, scale*W)
      IF (radius .EQ. 0.0_dp) THEN
         RETURN
      END IF

!   *** properties of the grid ***
      rpl = rp
      dr(:) = pwgrid%pw_grid%dr(:)
      grid => pwgrid%cr3d
      bo = pwgrid%pw_grid%bounds_local
      gbo = pwgrid%pw_grid%bounds

!   *** get the sub grid properties for the given radius ***
      CALL return_cube(cube_info, radius, lb_cube, ub_cube, sphere_bounds)

      IF (ALL(n_rep_real == 0)) THEN
         cubecenter(:) = FLOOR(rpl(:)/dr(:))+gbo(1, :)
         zub = MIN(bo(2, 3), cubecenter(3)+ub_cube(3))
         zlb = MAX(bo(1, 3), cubecenter(3)+lb_cube(3))
         yub = MIN(bo(2, 2), cubecenter(2)+ub_cube(2))
         ylb = MAX(bo(1, 2), cubecenter(2)+lb_cube(2))
         xub = MIN(bo(2, 1), cubecenter(1)+ub_cube(1))
         xlb = MAX(bo(1, 1), cubecenter(1)+lb_cube(1))
         IF (zlb .GT. zub .OR. ylb .GT. yub .OR. xlb .GT. xub) RETURN
         DO ig = zlb, zub
            rpg = REAL(ig-gbo(1, 3), dp)*dr(3)-rpl(3)
            zap = EXP(-zetp*rpg**2)
            zdat(ig) = scale*W*zap
         ENDDO
         DO ig = ylb, yub
            rpg = REAL(ig-gbo(1, 2), dp)*dr(2)-rpl(2)
            yap = EXP(-zetp*rpg**2)
            ydat(ig) = yap
         ENDDO
         DO ig = xlb, xub
            rpg = REAL(ig-gbo(1, 1), dp)*dr(1)-rpl(1)
            xap = EXP(-zetp*rpg**2)
            xdat(ig) = xap
         ENDDO
         CALL collocate_gf_npbc(grid, xdat, ydat, zdat, bo, bo2, zlb, zub, ylb, yub, xlb, xub)
      ELSE
         DO iz = -n_rep_real(3), n_rep_real(3)
            my_shift(3) = mm_cell%hmat(3, 3)*REAL(iz, KIND=dp)
            DO iy = -n_rep_real(2), n_rep_real(2)
               my_shift(2) = mm_cell%hmat(2, 2)*REAL(iy, KIND=dp)
               DO ix = -n_rep_real(1), n_rep_real(1)
                  my_shift(1) = mm_cell%hmat(1, 1)*REAL(ix, KIND=dp)
                  rpl = rp+my_shift(:)
                  cubecenter(:) = FLOOR(rpl(:)/dr(:))+gbo(1, :)
                  zub = MIN(bo(2, 3), cubecenter(3)+ub_cube(3))
                  zlb = MAX(bo(1, 3), cubecenter(3)+lb_cube(3))
                  yub = MIN(bo(2, 2), cubecenter(2)+ub_cube(2))
                  ylb = MAX(bo(1, 2), cubecenter(2)+lb_cube(2))
                  xub = MIN(bo(2, 1), cubecenter(1)+ub_cube(1))
                  xlb = MAX(bo(1, 1), cubecenter(1)+lb_cube(1))
                  IF (zlb .GT. zub .OR. ylb .GT. yub .OR. xlb .GT. xub) CYCLE
                  DO ig = zlb, zub
                     rpg = REAL(ig-gbo(1, 3), dp)*dr(3)-rpl(3)
                     zap = EXP(-zetp*rpg**2)
                     zdat(ig) = scale*W*zap
                  ENDDO
                  DO ig = ylb, yub
                     rpg = REAL(ig-gbo(1, 2), dp)*dr(2)-rpl(2)
                     yap = EXP(-zetp*rpg**2)
                     ydat(ig) = yap
                  ENDDO
                  DO ig = xlb, xub
                     rpg = REAL(ig-gbo(1, 1), dp)*dr(1)-rpl(1)
                     xap = EXP(-zetp*rpg**2)
                     xdat(ig) = xap
                  ENDDO
                  CALL collocate_gf_npbc(grid, xdat, ydat, zdat, bo, bo2, zlb, zub, ylb, yub, xlb, xub)
               END DO
            END DO
         END DO
      END IF

   END SUBROUTINE collocate_gf_rspace_NoPBC

! **************************************************************************************************
!> \brief Main driver to integrate gaussian functions on a grid function
!>      without using periodic boundary conditions (NoPBC)
!>      Computes Forces.
!> \param zetp ...
!> \param rp ...
!> \param scale ...
!> \param W ...
!> \param pwgrid ...
!> \param cube_info ...
!> \param eps_mm_rspace ...
!> \param xdat ...
!> \param ydat ...
!> \param zdat ...
!> \param bo ...
!> \param force ...
!> \param n_rep_real ...
!> \param mm_cell ...
!> \par History
!>      07.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE integrate_gf_rspace_NoPBC(zetp, rp, scale, W, pwgrid, cube_info, &
                                        eps_mm_rspace, xdat, ydat, zdat, bo, force, n_rep_real, mm_cell)
      REAL(KIND=dp), INTENT(IN)                          :: zetp
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rp
      REAL(KIND=dp), INTENT(IN)                          :: scale, W
      TYPE(pw_type), POINTER                             :: pwgrid
      TYPE(cube_info_type), INTENT(IN)                   :: cube_info
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bo
      REAL(KIND=dp), DIMENSION(2, bo(1, 3):bo(2, 3))     :: zdat
      REAL(KIND=dp), DIMENSION(2, bo(1, 2):bo(2, 2))     :: ydat
      REAL(KIND=dp), DIMENSION(2, bo(1, 1):bo(2, 1))     :: xdat
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: force
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n_rep_real
      TYPE(cell_type), POINTER                           :: mm_cell

      INTEGER                                            :: ig, ix, iy, iz, xlb, xub, ylb, yub, zlb, &
                                                            zub
      INTEGER, DIMENSION(2, 3)                           :: gbo
      INTEGER, DIMENSION(3)                              :: cubecenter, lb_cube, ub_cube
      INTEGER, DIMENSION(:), POINTER                     :: sphere_bounds
      REAL(KIND=dp)                                      :: radius, rpg, xap, yap, zap
      REAL(KIND=dp), DIMENSION(3)                        :: dr, my_shift, rpl
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid

      force = 0.0_dp
      radius = exp_radius(0, zetp, eps_mm_rspace, scale*W)
      IF (radius .EQ. 0.0_dp) RETURN

!   *** properties of the grid ***
      rpl = rp
      dr(:) = pwgrid%pw_grid%dr(:)
      grid => pwgrid%cr3d(:, :, :)
      gbo = pwgrid%pw_grid%bounds

!   *** get the sub grid properties for the given radius ***
      CALL return_cube(cube_info, radius, lb_cube, ub_cube, sphere_bounds)

      IF (ALL(n_rep_real == 0)) THEN
         cubecenter(:) = FLOOR(rpl(:)/dr(:))+gbo(1, :)
         zub = MIN(bo(2, 3), cubecenter(3)+ub_cube(3))
         zlb = MAX(bo(1, 3), cubecenter(3)+lb_cube(3))
         yub = MIN(bo(2, 2), cubecenter(2)+ub_cube(2))
         ylb = MAX(bo(1, 2), cubecenter(2)+lb_cube(2))
         xub = MIN(bo(2, 1), cubecenter(1)+ub_cube(1))
         xlb = MAX(bo(1, 1), cubecenter(1)+lb_cube(1))
         IF (zlb .GT. zub .OR. ylb .GT. yub .OR. xlb .GT. xub) RETURN
         DO ig = zlb, zub
            rpg = REAL(ig-gbo(1, 3), dp)*dr(3)-rpl(3)
            zap = EXP(-zetp*rpg**2)
            zdat(1, ig) = scale*W*zap
            zdat(2, ig) = rpg*zdat(1, ig)*zetp*2.0_dp
         ENDDO
         DO ig = ylb, yub
            rpg = REAL(ig-gbo(1, 2), dp)*dr(2)-rpl(2)
            yap = EXP(-zetp*rpg**2)
            ydat(1, ig) = yap
            ydat(2, ig) = rpg*ydat(1, ig)*zetp*2.0_dp
         ENDDO
         DO ig = xlb, xub
            rpg = REAL(ig-gbo(1, 1), dp)*dr(1)-rpl(1)
            xap = EXP(-zetp*rpg**2)
            xdat(1, ig) = xap
            xdat(2, ig) = rpg*xdat(1, ig)*zetp*2.0_dp
         ENDDO
         CALL integrate_gf_npbc(grid, xdat, ydat, zdat, bo, zlb, zub, ylb, yub, xlb, xub, force)
      ELSE
         DO iz = -n_rep_real(3), n_rep_real(3)
            my_shift(3) = mm_cell%hmat(3, 3)*REAL(iz, KIND=dp)
            DO iy = -n_rep_real(2), n_rep_real(2)
               my_shift(2) = mm_cell%hmat(2, 2)*REAL(iy, KIND=dp)
               DO ix = -n_rep_real(1), n_rep_real(1)
                  my_shift(1) = mm_cell%hmat(1, 1)*REAL(ix, KIND=dp)
                  rpl = rp+my_shift(:)
                  cubecenter(:) = FLOOR(rpl(:)/dr(:))+gbo(1, :)
                  zub = MIN(bo(2, 3), cubecenter(3)+ub_cube(3))
                  zlb = MAX(bo(1, 3), cubecenter(3)+lb_cube(3))
                  yub = MIN(bo(2, 2), cubecenter(2)+ub_cube(2))
                  ylb = MAX(bo(1, 2), cubecenter(2)+lb_cube(2))
                  xub = MIN(bo(2, 1), cubecenter(1)+ub_cube(1))
                  xlb = MAX(bo(1, 1), cubecenter(1)+lb_cube(1))
                  IF (zlb .GT. zub .OR. ylb .GT. yub .OR. xlb .GT. xub) CYCLE
                  DO ig = zlb, zub
                     rpg = REAL(ig-gbo(1, 3), dp)*dr(3)-rpl(3)
                     zap = EXP(-zetp*rpg**2)
                     zdat(1, ig) = scale*W*zap
                     zdat(2, ig) = rpg*zdat(1, ig)*zetp*2.0_dp
                  ENDDO
                  DO ig = ylb, yub
                     rpg = REAL(ig-gbo(1, 2), dp)*dr(2)-rpl(2)
                     yap = EXP(-zetp*rpg**2)
                     ydat(1, ig) = yap
                     ydat(2, ig) = rpg*ydat(1, ig)*zetp*2.0_dp
                  ENDDO
                  DO ig = xlb, xub
                     rpg = REAL(ig-gbo(1, 1), dp)*dr(1)-rpl(1)
                     xap = EXP(-zetp*rpg**2)
                     xdat(1, ig) = xap
                     xdat(2, ig) = rpg*xdat(1, ig)*zetp*2.0_dp
                  ENDDO
                  CALL integrate_gf_npbc(grid, xdat, ydat, zdat, bo, &
                                         zlb, zub, ylb, yub, xlb, xub, force)
               END DO
            END DO
         END DO
      END IF
   END SUBROUTINE integrate_gf_rspace_NoPBC

END MODULE mm_collocate_potential
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the QMMM Hamiltonian integral matrix <a|\sum_i q_i|b> for
!>      semi-empirical methods
!> \author Teodoro Laino 04.2007 [created]
! **************************************************************************************************
MODULE qmmm_se_energy
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_set
   USE cp_dbcsr_output,                 ONLY: cp_dbcsr_write_sparse_matrix
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: &
        do_method_am1, do_method_mndo, do_method_mndod, do_method_pchg, do_method_pdg, &
        do_method_pm3, do_method_pm6, do_method_pnnl, do_method_rm1, do_qmmm_coulomb, &
        do_qmmm_gauss, do_qmmm_none, do_qmmm_pcharge, do_qmmm_swave
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE multipole_types,                 ONLY: do_multipole_none
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_type,&
                                              qmmm_pot_p_type,&
                                              qmmm_pot_type
   USE qmmm_util,                       ONLY: spherical_cutoff_factor
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_qmmm_types,                ONLY: qs_ks_qmmm_env_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_neighbor_lists,               ONLY: build_qs_neighbor_lists
   USE qs_overlap,                      ONLY: build_overlap_matrix
   USE semi_empirical_int_arrays,       ONLY: se_orbital_pointer
   USE semi_empirical_integrals,        ONLY: corecore,&
                                              rotnuc
   USE semi_empirical_types,            ONLY: get_se_param,&
                                              se_int_control_type,&
                                              se_taper_type,&
                                              semi_empirical_create,&
                                              semi_empirical_release,&
                                              semi_empirical_type,&
                                              setup_se_int_control_type
   USE semi_empirical_utils,            ONLY: get_se_type,&
                                              se_param_set_default
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_se_energy'

   PUBLIC :: build_se_qmmm_matrix

CONTAINS

! **************************************************************************************************
!> \brief Constructs the 1-el semi-empirical hamiltonian
!> \param qs_env ...
!> \param qmmm_env ...
!> \param particles_mm ...
!> \param mm_cell ...
!> \param para_env ...
!> \author Teodoro Laino 04.2007 [created]
! **************************************************************************************************
   SUBROUTINE build_se_qmmm_matrix(qs_env, qmmm_env, particles_mm, mm_cell, para_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_mm
      TYPE(cell_type), POINTER                           :: mm_cell
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'build_se_qmmm_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, ikind, itype, iw, &
                                                            natom, natorb_a, nkind
      INTEGER, DIMENSION(:), POINTER                     :: list
      LOGICAL                                            :: anag, defined, found
      REAL(KIND=dp)                                      :: enuclear
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: h_block_a
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_qm
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_ks_qmmm_env_type), POINTER                 :: ks_qmmm_env_loc
      TYPE(se_int_control_type)                          :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a, se_kind_mm

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()

      NULLIFY (matrix_s, atomic_kind_set, qs_kind_set, energy)
      NULLIFY (se_kind_a, se_kind_mm, se_taper, particles_qm, ks_env, sab_orb)
      CALL build_qs_neighbor_lists(qs_env, para_env, force_env_section=qs_env%input)
      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      matrix_s=matrix_s, &
                      energy=energy, &
                      sab_orb=sab_orb)

      CALL build_overlap_matrix(ks_env, matrix_s=matrix_s, &
                                matrix_name="OVERLAP", &
                                basis_type_a="ORB", &
                                basis_type_b="ORB", &
                                sab_nl=sab_orb, &
                                qs_input=qs_env%input)

      CALL set_ks_env(ks_env, matrix_s=matrix_s)
      CALL get_qs_env(qs_env=qs_env, &
                      se_taper=se_taper, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      ks_qmmm_env=ks_qmmm_env_loc, &
                      dft_control=dft_control, &
                      particle_set=particles_qm)

      SELECT CASE (dft_control%qs_control%method_id)
      CASE (do_method_am1, do_method_rm1, do_method_mndo, do_method_pdg, &
            do_method_pm3, do_method_pm6, do_method_mndod, do_method_pnnl)
         ! Go on with the calculation..
      CASE DEFAULT
         ! Otherwise stop..
         CALL cp__b("qmmm_se_energy.F",154,"Method not available")
      END SELECT
      anag = dft_control%qs_control%se_control%analytical_gradients
      ! Setup type for SE integral control
      CALL setup_se_int_control_type( &
         se_int_control, shortrange=.FALSE., do_ewald_r3=.FALSE., &
         do_ewald_gks=.FALSE., integral_screening=dft_control%qs_control%se_control%integral_screening, &
         max_multipole=do_multipole_none, pc_coulomb_int=.FALSE.)

      ! Allocate the core Hamiltonian matrix
      CALL cp_dbcsr_allocate_matrix_set(ks_qmmm_env_loc%matrix_h, 1)
      ALLOCATE (ks_qmmm_env_loc%matrix_h(1)%matrix)

      CALL cp_dbcsr_init(ks_qmmm_env_loc%matrix_h(1)%matrix)

      CALL cp_dbcsr_copy(ks_qmmm_env_loc%matrix_h(1)%matrix, matrix_s(1)%matrix, &
                         name="QMMM HAMILTONIAN MATRIX")
      CALL cp_dbcsr_set(ks_qmmm_env_loc%matrix_h(1)%matrix, 0.0_dp)

      SELECT CASE (qmmm_env%qmmm_coupl_type)
      CASE (do_qmmm_coulomb, do_qmmm_gauss, do_qmmm_swave, do_qmmm_pcharge)
         ! Create a fake semi-empirical type to handle the classical atom
         CALL semi_empirical_create(se_kind_mm)
         CALL se_param_set_default(se_kind_mm, 0, do_method_pchg)
         itype = get_se_type(se_kind_mm%typ)
         nkind = SIZE(atomic_kind_set)
         enuclear = 0.0_dp
         Kinds: DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=list)
            CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
            CALL get_se_param(se_kind_a, &
                              defined=defined, &
                              natorb=natorb_a)
            IF (.NOT. defined .OR. natorb_a < 1) CYCLE
            Atoms: DO i = 1, SIZE(list)
               iatom = list(i)
               ! Give back block
               NULLIFY (h_block_a)
               CALL cp_dbcsr_get_block_p(matrix=ks_qmmm_env_loc%matrix_h(1)%matrix, &
                                         row=iatom, col=iatom, BLOCK=h_block_a, found=found)

               IF (ASSOCIATED(h_block_a)) THEN
                  h_block_a = 0.0_dp
                  ! Real QM/MM computation
                  CALL build_se_qmmm_matrix_low(h_block_a, &
                                                se_kind_a, &
                                                se_kind_mm, &
                                                qmmm_env%Potentials, &
                                                particles_mm, &
                                                qmmm_env%mm_atom_chrg, &
                                                qmmm_env%mm_atom_index, &
                                                mm_cell, &
                                                iatom, &
                                                enuclear, &
                                                itype, &
                                                se_taper, &
                                                se_int_control, &
                                                anag, &
                                                qmmm_env%spherical_cutoff, &
                                                particles_qm)
                  ! Possibly added charges
                  IF (qmmm_env%move_mm_charges .OR. qmmm_env%add_mm_charges) THEN
                     CALL build_se_qmmm_matrix_low(h_block_a, &
                                                   se_kind_a, &
                                                   se_kind_mm, &
                                                   qmmm_env%added_charges%potentials, &
                                                   qmmm_env%added_charges%added_particles, &
                                                   qmmm_env%added_charges%mm_atom_chrg, &
                                                   qmmm_env%added_charges%mm_atom_index, &
                                                   mm_cell, &
                                                   iatom, &
                                                   enuclear, &
                                                   itype, &
                                                   se_taper, &
                                                   se_int_control, &
                                                   anag, &
                                                   qmmm_env%spherical_cutoff, &
                                                   particles_qm)
                  END IF
               END IF
            END DO Atoms
         END DO Kinds
         CALL mp_sum(enuclear, para_env%group)
         energy%qmmm_nu = enuclear
         CALL semi_empirical_release(se_kind_mm)
      CASE (do_qmmm_none)
         ! Zero Matrix
         CALL cp_dbcsr_set(ks_qmmm_env_loc%matrix_h(1)%matrix, 0.0_dp)
      END SELECT
      IF (BTEST(cp_print_key_should_output(logger%iter_info, &
                                           qs_env%input, "QMMM%PRINT%QMMM_MATRIX"), cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger, qs_env%input, "QMMM%PRINT%QMMM_MATRIX", &
                                   extension=".Log")
         CALL cp_dbcsr_write_sparse_matrix(ks_qmmm_env_loc%matrix_h(1)%matrix, 4, 6, qs_env, para_env, &
                                           scale=1.0_dp, output_unit=iw)
         CALL cp_print_key_finished_output(iw, logger, qs_env%input, &
                                           "QMMM%PRINT%QMMM_MATRIX")
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_se_qmmm_matrix

! **************************************************************************************************
!> \brief Low Level : Constructs the 1-el semi-empirical hamiltonian block
!> \param h_block_a ...
!> \param se_kind_a ...
!> \param se_kind_mm ...
!> \param potentials ...
!> \param particles_mm ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param mm_cell ...
!> \param IndQM ...
!> \param enuclear ...
!> \param itype ...
!> \param se_taper ...
!> \param se_int_control ...
!> \param anag ...
!> \param qmmm_spherical_cutoff ...
!> \param particles_qm ...
!> \author Teodoro Laino 04.2007 [created]
! **************************************************************************************************
   SUBROUTINE build_se_qmmm_matrix_low(h_block_a, se_kind_a, se_kind_mm, potentials, &
                                       particles_mm, mm_charges, mm_atom_index, &
                                       mm_cell, IndQM, enuclear, itype, se_taper, se_int_control, anag, &
                                       qmmm_spherical_cutoff, particles_qm)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: h_block_a
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a, se_kind_mm
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: potentials
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_mm
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(cell_type), POINTER                           :: mm_cell
      INTEGER, INTENT(IN)                                :: IndQM
      REAL(KIND=dp), INTENT(INOUT)                       :: enuclear
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      LOGICAL, INTENT(IN)                                :: anag
      REAL(KIND=dp), INTENT(IN)                          :: qmmm_spherical_cutoff(2)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_qm

      CHARACTER(len=*), PARAMETER :: routineN = 'build_se_qmmm_matrix_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i1, i1L, i2, Imm, Imp, IndMM, &
                                                            Ipot, j1, j1L
      REAL(KIND=dp)                                      :: enuc, rt1, rt2, rt3, sph_chrg_factor
      REAL(KIND=dp), DIMENSION(3)                        :: r_pbc, rij
      REAL(KIND=dp), DIMENSION(45)                       :: e1b
      TYPE(qmmm_pot_type), POINTER                       :: Pot

      CALL timeset(routineN, handle)
      ! Loop Over MM atoms
      ! Loop over Pot stores atoms with the same charge
      MainLoopPot: DO Ipot = 1, SIZE(Potentials)
         Pot => Potentials(Ipot)%Pot
         ! Loop over atoms belonging to this type
         LoopMM: DO Imp = 1, SIZE(Pot%mm_atom_index)
            Imm = Pot%mm_atom_index(Imp)
            IndMM = mm_atom_index(Imm)
            r_pbc = pbc(particles_mm(IndMM)%r-particles_qm(IndQM)%r, mm_cell)
            rt1 = r_pbc(1)
            rt2 = r_pbc(2)
            rt3 = r_pbc(3)
            rij = (/rt1, rt2, rt3/)
            se_kind_mm%zeff = mm_charges(Imm)
            ! Computes the screening factor for the spherical cutoff (if defined)
            IF (qmmm_spherical_cutoff(1) > 0.0_dp) THEN
               CALL spherical_cutoff_factor(qmmm_spherical_cutoff, rij, sph_chrg_factor)
               se_kind_mm%zeff = se_kind_mm%zeff*sph_chrg_factor
            END IF
            IF (ABS(se_kind_mm%zeff) <= EPSILON(0.0_dp)) CYCLE
            CALL rotnuc(se_kind_a, se_kind_mm, rij, itype=itype, e1b=e1b, anag=anag, &
                        se_int_control=se_int_control, se_taper=se_taper)
            CALL corecore(se_kind_a, se_kind_mm, rij, itype=itype, enuc=enuc, anag=anag, &
                          se_int_control=se_int_control, se_taper=se_taper)
            enuclear = enuclear+enuc
            ! Contribution to the iatom block
            ! Computation of the QMMM core matrix
            i2 = 0
            DO i1L = 1, se_kind_a%natorb
               i1 = se_orbital_pointer(i1L)
               DO j1L = 1, i1L-1
                  j1 = se_orbital_pointer(j1L)
                  i2 = i2+1
                  h_block_a(i1, j1) = h_block_a(i1, j1)+e1b(i2)
                  h_block_a(j1, i1) = h_block_a(i1, j1)
               END DO
               j1 = se_orbital_pointer(j1L)
               i2 = i2+1
               h_block_a(i1, j1) = h_block_a(i1, j1)+e1b(i2)
            END DO
         END DO LoopMM
      END DO MainLoopPot
      CALL timestop(handle)
   END SUBROUTINE build_se_qmmm_matrix_low

END MODULE qmmm_se_energy
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A collection of methods to treat the QM/MM electrostatic coupling
!> \par History
!>      5.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_gpw_energy
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_cube,          ONLY: cp_pw_to_cube
   USE cp_spline_utils,                 ONLY: pw_prolongate_s3
   USE cube_utils,                      ONLY: cube_info_type
   USE input_constants,                 ONLY: do_par_atom,&
                                              do_qmmm_coulomb,&
                                              do_qmmm_gauss,&
                                              do_qmmm_none,&
                                              do_qmmm_pcharge,&
                                              do_qmmm_swave,&
                                              spline3_nopbc_interp,&
                                              spline3_pbc_interp
   USE input_section_types,             ONLY: section_get_ivals,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum,&
                                              mp_sync
   USE mm_collocate_potential,          ONLY: collocate_gf_rspace_NoPBC
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_p_type,&
                                              pw_pools_create_pws,&
                                              pw_pools_give_back_pws
   USE pw_types,                        ONLY: REALDATA3D,&
                                              REALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qmmm_gaussian_types,             ONLY: qmmm_gaussian_p_type,&
                                              qmmm_gaussian_type
   USE qmmm_se_energy,                  ONLY: build_se_qmmm_matrix
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_type,&
                                              qmmm_per_pot_p_type,&
                                              qmmm_per_pot_type,&
                                              qmmm_pot_p_type,&
                                              qmmm_pot_type
   USE qmmm_util,                       ONLY: spherical_cutoff_factor
   USE qs_dftb_matrices,                ONLY: build_dftb_qmmm_matrix,&
                                              build_dftb_qmmm_matrix_pc,&
                                              build_dftb_qmmm_matrix_zero
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_qmmm_types,                ONLY: qs_ks_qmmm_env_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_gpw_energy'

   PUBLIC :: qmmm_el_coupling
   PUBLIC :: qmmm_elec_with_gaussian, &
             qmmm_elec_with_gaussian_LR, qmmm_elec_with_gaussian_LG
!***
CONTAINS

! **************************************************************************************************
!> \brief Main Driver to compute the QM/MM Electrostatic Coupling
!> \param qs_env ...
!> \param qmmm_env ...
!> \param mm_particles ...
!> \param mm_cell ...
!> \par History
!>      05.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_el_coupling(qs_env, qmmm_env, mm_particles, mm_cell)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      TYPE(cell_type), POINTER                           :: mm_cell

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_el_coupling', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw, iw2
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(qs_ks_qmmm_env_type), POINTER                 :: ks_qmmm_env_loc
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: input_section, interp_section, &
                                                            print_section

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      NULLIFY (ks_qmmm_env_loc, pw_pools, pw_env, input_section, dft_control)
      CALL get_qs_env(qs_env=qs_env, &
                      pw_env=pw_env, &
                      para_env=para_env, &
                      input=input_section, &
                      ks_qmmm_env=ks_qmmm_env_loc, &
                      subsys=subsys, &
                      dft_control=dft_control)
      CALL qs_subsys_get(subsys, particles=particles)

      CALL pw_env_get(pw_env=pw_env, pw_pools=pw_pools)
      print_section => section_vals_get_subs_vals(input_section, "QMMM%PRINT")
      iw = cp_print_key_unit_nr(logger, print_section, "PROGRAM_RUN_INFO", &
                                extension=".qmmmLog")
      IF (iw > 0) &
         WRITE (iw, '(T2,"QMMM|",1X,A)') "Information on the QM/MM Electrostatic Potential:"
      !
      ! Initializing vectors:
      !        Zeroing v_qmmm_rspace
      CALL pw_zero(ks_qmmm_env_loc%v_qmmm_rspace%pw)
      IF (dft_control%qs_control%semi_empirical) THEN
         ! SEMIEMPIRICAL
         SELECT CASE (qmmm_env%qmmm_coupl_type)
         CASE (do_qmmm_coulomb, do_qmmm_none)
            CALL build_se_qmmm_matrix(qs_env, qmmm_env, mm_particles, mm_cell, para_env)
            IF (qmmm_env%qmmm_coupl_type == do_qmmm_none) THEN
               IF (iw > 0) WRITE (iw, '(T2,"QMMM|",1X,A)') &
                  "No QM/MM Electrostatic coupling. Just Mechanical Coupling!"
            END IF
         CASE (do_qmmm_pcharge)
            CALL cp__b("qmmm_gpw_energy.F",148,"Point charge QM/MM electrostatic coupling not yet implemented for SE.")
         CASE (do_qmmm_gauss, do_qmmm_swave)
            CALL cp__b("qmmm_gpw_energy.F",150,"GAUSS or SWAVE QM/MM electrostatic coupling not yet implemented for SE.")
         CASE DEFAULT
            IF (iw > 0) WRITE (iw, '(T2,"QMMM|",1X,A)') "Unknown Coupling..."
            CALL cp__b("qmmm_gpw_energy.F",153,"")
         END SELECT
      ELSEIF (dft_control%qs_control%dftb) THEN
         ! DFTB
         SELECT CASE (qmmm_env%qmmm_coupl_type)
         CASE (do_qmmm_none)
            IF (iw > 0) WRITE (iw, '(T2,"QMMM|",1X,A)') &
               "No QM/MM Electrostatic coupling. Just Mechanical Coupling!"
            CALL build_dftb_qmmm_matrix_zero(qs_env, para_env)
         CASE (do_qmmm_coulomb)
            CALL build_dftb_qmmm_matrix(qs_env, qmmm_env, mm_particles, mm_cell, para_env)
         CASE (do_qmmm_pcharge)
            CALL build_dftb_qmmm_matrix_pc(qs_env, qmmm_env, mm_particles, mm_cell, para_env)
         CASE (do_qmmm_gauss, do_qmmm_swave)
            CALL cp__b("qmmm_gpw_energy.F",167,"GAUSS or SWAVE QM/MM electrostatic coupling not implemented for DFTB.")
         CASE DEFAULT
            IF (iw > 0) WRITE (iw, '(T2,"QMMM|",1X,A)') "Unknown Coupling..."
            CALL cp__b("qmmm_gpw_energy.F",170,"")
         END SELECT
      ELSE
         ! QS
         SELECT CASE (qmmm_env%qmmm_coupl_type)
         CASE (do_qmmm_coulomb)
            CALL cp__b("qmmm_gpw_energy.F",176,"Coulomb QM/MM electrostatic coupling not implemented for GPW/GAPW.")
         CASE (do_qmmm_pcharge)
            CALL cp__b("qmmm_gpw_energy.F",178,"Point Charge QM/MM electrostatic coupling not implemented for GPW/GAPW.")
         CASE (do_qmmm_gauss, do_qmmm_swave)
            IF (iw > 0) &
               WRITE (iw, '(T2,"QMMM|",1X,A)') &
               "QM/MM Coupling computed collocating the Gaussian Potential Functions."
            interp_section => section_vals_get_subs_vals(input_section, &
                                                         "QMMM%INTERPOLATOR")
            CALL qmmm_elec_with_gaussian(qmmm_env=qmmm_env, &
                                         v_qmmm=ks_qmmm_env_loc%v_qmmm_rspace, &
                                         mm_particles=mm_particles, &
                                         aug_pools=qmmm_env%aug_pools, &
                                         para_env=para_env, &
                                         eps_mm_rspace=qmmm_env%eps_mm_rspace, &
                                         cube_info=ks_qmmm_env_loc%cube_info, &
                                         pw_pools=pw_pools, &
                                         auxbas_grid=qmmm_env%gridlevel_info%auxbas_grid, &
                                         coarser_grid=qmmm_env%gridlevel_info%coarser_grid, &
                                         interp_section=interp_section, &
                                         mm_cell=mm_cell)
         CASE (do_qmmm_none)
            IF (iw > 0) WRITE (iw, '(T2,"QMMM|",1X,A)') &
               "No QM/MM Electrostatic coupling. Just Mechanical Coupling!"
         CASE DEFAULT
            IF (iw > 0) WRITE (iw, '(T2,"QMMM|",1X,A)') "Unknown Coupling..."
            CALL cp__b("qmmm_gpw_energy.F",202,"")
         END SELECT
         ! Dump info on the electrostatic potential if requested
         IF (BTEST(cp_print_key_should_output(logger%iter_info, print_section, &
                                              "POTENTIAL"), cp_p_file)) THEN
            iw2 = cp_print_key_unit_nr(logger, print_section, "POTENTIAL", &
                                       extension=".qmmmLog")
            CALL cp_pw_to_cube(ks_qmmm_env_loc%v_qmmm_rspace%pw, iw2, &
                               particles=particles, &
                               stride=section_get_ivals(print_section, "POTENTIAL%STRIDE"), &
                               title="QM/MM: MM ELECTROSTATIC POTENTIAL ")
            CALL cp_print_key_finished_output(iw2, logger, print_section, &
                                              "POTENTIAL")
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, print_section, &
                                        "PROGRAM_RUN_INFO")
      CALL timestop(handle)
   END SUBROUTINE qmmm_el_coupling

! **************************************************************************************************
!> \brief Compute the QM/MM electrostatic Interaction collocating the gaussian
!>      Electrostatic Potential
!> \param qmmm_env ...
!> \param v_qmmm ...
!> \param mm_particles ...
!> \param aug_pools ...
!> \param cube_info ...
!> \param para_env ...
!> \param eps_mm_rspace ...
!> \param pw_pools ...
!> \param auxbas_grid ...
!> \param coarser_grid ...
!> \param interp_section ...
!> \param mm_cell ...
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_elec_with_gaussian(qmmm_env, v_qmmm, mm_particles, &
                                      aug_pools, cube_info, para_env, eps_mm_rspace, pw_pools, &
                                      auxbas_grid, coarser_grid, interp_section, mm_cell)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(pw_p_type), INTENT(INOUT)                     :: v_qmmm
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: aug_pools
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      INTEGER, INTENT(IN)                                :: auxbas_grid, coarser_grid
      TYPE(section_vals_type), POINTER                   :: interp_section
      TYPE(cell_type), POINTER                           :: mm_cell

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_elec_with_gaussian', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle2, igrid, ilevel, &
                                                            kind_interp, lb(3), ngrids, ub(3)
      LOGICAL                                            :: shells
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: grids

      IF(.NOT.(ASSOCIATED(mm_particles)))CALL cp__a("qmmm_gpw_energy.F",264)
      IF(.NOT.(ASSOCIATED(qmmm_env%mm_atom_chrg)))CALL cp__a("qmmm_gpw_energy.F",265)
      IF(.NOT.(ASSOCIATED(qmmm_env%mm_atom_index)))CALL cp__a("qmmm_gpw_energy.F",266)
      IF(.NOT.(ASSOCIATED(aug_pools)))CALL cp__a("qmmm_gpw_energy.F",267)
      IF(.NOT.(ASSOCIATED(pw_pools)))CALL cp__a("qmmm_gpw_energy.F",268)
      !Statements
      CALL timeset(routineN, handle)
      ngrids = SIZE(pw_pools)
      CALL pw_pools_create_pws(aug_pools, grids, use_data=REALDATA3D, in_space=REALSPACE)
      DO igrid = 1, ngrids
         CALL pw_zero(grids(igrid)%pw)
      END DO

      shells = .FALSE.

      CALL qmmm_elec_with_gaussian_low(grids, mm_particles, &
                                       qmmm_env%mm_atom_chrg, qmmm_env%mm_atom_index, &
                                       cube_info, para_env, eps_mm_rspace, qmmm_env%pgfs, &
                                       auxbas_grid, coarser_grid, qmmm_env%potentials, &
                                       mm_cell=mm_cell, dOmmOqm=qmmm_env%dOmmOqm, periodic=qmmm_env%periodic, &
                                       per_potentials=qmmm_env%per_potentials, par_scheme=qmmm_env%par_scheme, &
                                       qmmm_spherical_cutoff=qmmm_env%spherical_cutoff, shells=shells)

      IF (qmmm_env%move_mm_charges .OR. qmmm_env%add_mm_charges) THEN
         CALL qmmm_elec_with_gaussian_low(grids, qmmm_env%added_charges%added_particles, &
                                          qmmm_env%added_charges%mm_atom_chrg, &
                                          qmmm_env%added_charges%mm_atom_index, &
                                          cube_info, para_env, eps_mm_rspace, qmmm_env%added_charges%pgfs, auxbas_grid, &
                                          coarser_grid, qmmm_env%added_charges%potentials, &
                                          mm_cell=mm_cell, dOmmOqm=qmmm_env%dOmmOqm, periodic=qmmm_env%periodic, &
                                          per_potentials=qmmm_env%per_potentials, par_scheme=qmmm_env%par_scheme, &
                                          qmmm_spherical_cutoff=qmmm_env%spherical_cutoff, shells=shells)
      END IF
      IF (qmmm_env%added_shells%num_mm_atoms .GT. 0) THEN
         shells = .TRUE.
         CALL qmmm_elec_with_gaussian_low(grids, qmmm_env%added_shells%added_particles, &
                                          qmmm_env%added_shells%mm_core_chrg, &
                                          qmmm_env%added_shells%mm_core_index, &
                                          cube_info, para_env, eps_mm_rspace, qmmm_env%added_shells%pgfs, auxbas_grid, &
                                          coarser_grid, qmmm_env%added_shells%potentials, &
                                          mm_cell=mm_cell, dOmmOqm=qmmm_env%dOmmOqm, periodic=qmmm_env%periodic, &
                                          per_potentials=qmmm_env%added_shells%per_potentials, &
                                          par_scheme=qmmm_env%par_scheme, qmmm_spherical_cutoff=qmmm_env%spherical_cutoff, &
                                          shells=shells)
      END IF
      ! Sumup all contributions according the parallelization scheme
      IF (qmmm_env%par_scheme == do_par_atom) THEN
         DO ilevel = 1, SIZE(grids)
            CALL mp_sum(grids(ilevel)%pw%cr3d, para_env%group)
         END DO
      END IF
      ! RealSpace Interpolation
      CALL section_vals_val_get(interp_section, "kind", i_val=kind_interp)
      SELECT CASE (kind_interp)
      CASE (spline3_nopbc_interp, spline3_pbc_interp)
         ! Spline Iterpolator
         CALL mp_sync(para_env%group)
         CALL timeset(TRIM(routineN)//":spline3Int", handle2)
         DO Ilevel = coarser_grid, auxbas_grid+1, -1
            CALL pw_prolongate_s3(grids(Ilevel)%pw, &
                                  grids(Ilevel-1)%pw, &
                                  aug_pools(Ilevel)%pool, &
                                  param_section=interp_section)
         END DO
         CALL timestop(handle2)
      CASE DEFAULT
         CALL cp__b("qmmm_gpw_energy.F",330,"")
      END SELECT
      lb = v_qmmm%pw%pw_grid%bounds_local(1, :)
      ub = v_qmmm%pw%pw_grid%bounds_local(2, :)

      v_qmmm%pw%cr3d = grids(auxbas_grid)%pw%cr3d(lb(1):ub(1), &
                                                  lb(2):ub(2), &
                                                  lb(3):ub(3))

      CALL pw_pools_give_back_pws(aug_pools, grids)

      CALL timestop(handle)
   END SUBROUTINE qmmm_elec_with_gaussian

! **************************************************************************************************
!> \brief Compute the QM/MM electrostatic Interaction collocating the gaussian
!>      Electrostatic Potential - Low Level
!> \param tmp_grid ...
!> \param mm_particles ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param cube_info ...
!> \param para_env ...
!> \param eps_mm_rspace ...
!> \param pgfs ...
!> \param auxbas_grid ...
!> \param coarser_grid ...
!> \param potentials ...
!> \param mm_cell ...
!> \param dOmmOqm ...
!> \param periodic ...
!> \param per_potentials ...
!> \param par_scheme ...
!> \param qmmm_spherical_cutoff ...
!> \param shells ...
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_elec_with_gaussian_low(tmp_grid, mm_particles, mm_charges, &
                                          mm_atom_index, cube_info, para_env, &
                                          eps_mm_rspace, pgfs, auxbas_grid, coarser_grid, &
                                          potentials, mm_cell, dOmmOqm, periodic, per_potentials, par_scheme, &
                                          qmmm_spherical_cutoff, shells)
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: tmp_grid
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      INTEGER, INTENT(IN)                                :: auxbas_grid, coarser_grid
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: potentials
      TYPE(cell_type), POINTER                           :: mm_cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dOmmOqm
      LOGICAL, INTENT(IN)                                :: periodic
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER   :: per_potentials
      INTEGER, INTENT(IN)                                :: par_scheme
      REAL(KIND=dp), INTENT(IN)                          :: qmmm_spherical_cutoff(2)
      LOGICAL, INTENT(IN)                                :: shells

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_elec_with_gaussian_low', &
         routineNb = 'qmmm_elec_gaussian_low', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle2, IGauss, ilevel, Imm, &
                                                            IndMM, IRadTyp, LIndMM, myind, &
                                                            n_rep_real(3)
      INTEGER, DIMENSION(2, 3)                           :: bo2
      REAL(KIND=dp)                                      :: alpha, height, sph_chrg_factor, W
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xdat, ydat, zdat
      TYPE(qmmm_gaussian_type), POINTER                  :: pgf
      TYPE(qmmm_per_pot_type), POINTER                   :: per_pot
      TYPE(qmmm_pot_type), POINTER                       :: pot

      NULLIFY (pgf, pot, per_pot, xdat, ydat, zdat)
      CALL timeset(routineN, handle)
      CALL timeset(routineNb//"_G", handle2)
      bo2 = tmp_grid(auxbas_grid)%pw%pw_grid%bounds
      ALLOCATE (xdat(bo2(1, 1):bo2(2, 1)))
      ALLOCATE (ydat(bo2(1, 2):bo2(2, 2)))
      ALLOCATE (zdat(bo2(1, 3):bo2(2, 3)))
      IF (par_scheme == do_par_atom) myind = 0
      Radius: DO IRadTyp = 1, SIZE(pgfs)
         pgf => pgfs(IRadTyp)%pgf
         pot => potentials(IRadTyp)%pot
         n_rep_real = 0
         IF (periodic) THEN
            per_pot => per_potentials(IRadTyp)%pot
            n_rep_real = per_pot%n_rep_real
         END IF
         Gaussian: DO IGauss = 1, pgf%Number_of_Gaussians
            alpha = 1.0_dp/pgf%Gk(IGauss)
            alpha = alpha*alpha
            height = pgf%Ak(IGauss)
            ilevel = pgf%grid_level(IGauss)
            Atoms: DO Imm = 1, SIZE(pot%mm_atom_index)
               IF (par_scheme == do_par_atom) THEN
                  myind = myind+1
                  IF (MOD(myind, para_env%num_pe) /= para_env%mepos) CYCLE Atoms
               END IF
               LIndMM = pot%mm_atom_index(Imm)
               IndMM = mm_atom_index(LIndMM)
               IF (shells) THEN
                  ra(:) = pbc(mm_particles(LIndMM)%r-dOmmOqm, mm_cell)+dOmmOqm
               ELSE
                  ra(:) = pbc(mm_particles(IndMM)%r-dOmmOqm, mm_cell)+dOmmOqm
               END IF
               W = mm_charges(LIndMM)*height
               ! Possible Spherical Cutoff
               IF (qmmm_spherical_cutoff(1) > 0.0_dp) THEN
                  CALL spherical_cutoff_factor(qmmm_spherical_cutoff, ra, sph_chrg_factor)
                  W = W*sph_chrg_factor
               END IF
               IF (ABS(W) <= EPSILON(0.0_dp)) CYCLE Atoms
               CALL collocate_gf_rspace_NoPBC(zetp=alpha, &
                                              rp=ra, &
                                              scale=-1.0_dp, &
                                              W=W, &
                                              pwgrid=tmp_grid(ilevel)%pw, &
                                              cube_info=cube_info(ilevel), &
                                              eps_mm_rspace=eps_mm_rspace, &
                                              xdat=xdat, &
                                              ydat=ydat, &
                                              zdat=zdat, &
                                              bo2=bo2, &
                                              n_rep_real=n_rep_real, &
                                              mm_cell=mm_cell)
            END DO Atoms
         END DO Gaussian
      END DO Radius
      IF (ASSOCIATED(xdat)) THEN
         DEALLOCATE (xdat)
      ENDIF
      IF (ASSOCIATED(ydat)) THEN
         DEALLOCATE (ydat)
      ENDIF
      IF (ASSOCIATED(zdat)) THEN
         DEALLOCATE (zdat)
      ENDIF
      CALL timestop(handle2)
      CALL timeset(routineNb//"_R", handle2)
      IF (periodic) THEN
         ! Long Range Part of the QM/MM Potential with Gaussians With Periodic Boundary Conditions
         CALL qmmm_elec_with_gaussian_LG(pgfs=pgfs, &
                                         cgrid=tmp_grid(coarser_grid)%pw, &
                                         mm_charges=mm_charges, &
                                         mm_atom_index=mm_atom_index, &
                                         mm_particles=mm_particles, &
                                         para_env=para_env, &
                                         per_potentials=per_potentials, &
                                         mm_cell=mm_cell, &
                                         dOmmOqm=dOmmOqm, &
                                         par_scheme=par_scheme, &
                                         qmmm_spherical_cutoff=qmmm_spherical_cutoff, &
                                         shells=shells)
      ELSE
         ! Long Range Part of the QM/MM Potential with Gaussians
         CALL qmmm_elec_with_gaussian_LR(pgfs=pgfs, &
                                         grid=tmp_grid(coarser_grid)%pw, &
                                         mm_charges=mm_charges, &
                                         mm_atom_index=mm_atom_index, &
                                         mm_particles=mm_particles, &
                                         para_env=para_env, &
                                         potentials=potentials, &
                                         mm_cell=mm_cell, &
                                         dOmmOqm=dOmmOqm, &
                                         par_scheme=par_scheme, &
                                         qmmm_spherical_cutoff=qmmm_spherical_cutoff, &
                                         shells=shells)
      END IF
      CALL timestop(handle2)
      CALL timestop(handle)

   END SUBROUTINE qmmm_elec_with_gaussian_low

! **************************************************************************************************
!> \brief Compute the QM/MM electrostatic Interaction collocating
!>      (1/R - Sum_NG Gaussians) on the coarser grid level in G-SPACE
!>      Long Range QM/MM Electrostatic Potential with Gaussian - Low Level
!>      PERIODIC BOUNDARY CONDITION VERSION
!> \param pgfs ...
!> \param cgrid ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param mm_particles ...
!> \param para_env ...
!> \param per_potentials ...
!> \param mm_cell ...
!> \param dOmmOqm ...
!> \param par_scheme ...
!> \param qmmm_spherical_cutoff ...
!> \param shells ...
!> \par History
!>      07.2004 created [tlaino]
!> \author Teodoro Laino
!> \note
!>      This version includes the explicit code of Eval_Interp_Spl3_pbc
!>      in order to achieve better performance
! **************************************************************************************************
   SUBROUTINE qmmm_elec_with_gaussian_LG(pgfs, cgrid, mm_charges, mm_atom_index, &
                                         mm_particles, para_env, per_potentials, &
                                         mm_cell, dOmmOqm, par_scheme, qmmm_spherical_cutoff, shells)
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(pw_type), POINTER                             :: cgrid
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER   :: per_potentials
      TYPE(cell_type), POINTER                           :: mm_cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dOmmOqm
      INTEGER, INTENT(IN)                                :: par_scheme
      REAL(KIND=dp), DIMENSION(2), INTENT(IN)            :: qmmm_spherical_cutoff
      LOGICAL                                            :: shells

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_elec_with_gaussian_LG', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii1, ii2, ii3, ii4, ij1, ij2, &
                                                            ij3, ij4, ik1, ik2, ik3, ik4, Imm, &
                                                            IndMM, IRadTyp, ivec(3), j, k, LIndMM, &
                                                            my_j, my_k, myind, npts(3)
      INTEGER, DIMENSION(2, 3)                           :: bo, gbo
      REAL(KIND=dp) :: a1, a2, a3, abc_X(4, 4), abc_X_Y(4), b1, b2, b3, c1, c2, c3, d1, d2, d3, &
         dr1, dr1c, dr2, dr2c, dr3, dr3c, e1, e2, e3, f1, f2, f3, g1, g2, g3, h1, h2, h3, p1, p2, &
         p3, q1, q2, q3, qt, r1, r2, r3, rt1, rt2, rt3, rv1, rv2, rv3, s1, s2, s3, s4, &
         sph_chrg_factor, t1, t2, t3, t4, u1, u2, u3, v1, v2, v3, v4, val, xd1, xd2, xd3, xs1, &
         xs2, xs3
      REAL(KIND=dp), DIMENSION(3)                        :: ra, vec
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid, grid2
      TYPE(pw_type), POINTER                             :: pw
      TYPE(qmmm_per_pot_type), POINTER                   :: per_pot

      CALL timeset(routineN, handle)
      NULLIFY (grid, pw)
      dr1c = cgrid%pw_grid%dr(1)
      dr2c = cgrid%pw_grid%dr(2)
      dr3c = cgrid%pw_grid%dr(3)
      gbo = cgrid%pw_grid%bounds
      bo = cgrid%pw_grid%bounds_local
      grid2 => cgrid%cr3d
      IF (par_scheme == do_par_atom) myind = 0
      Radius: DO IRadTyp = 1, SIZE(pgfs)
         per_pot => per_potentials(IRadTyp)%pot
         pw => per_pot%TabLR
         npts = pw%pw_grid%npts
         dr1 = pw%pw_grid%dr(1)
         dr2 = pw%pw_grid%dr(2)
         dr3 = pw%pw_grid%dr(3)
         grid => pw%cr3d(:, :, :)
         Atoms: DO Imm = 1, SIZE(per_pot%mm_atom_index)
            IF (par_scheme == do_par_atom) THEN
               myind = myind+1
               IF (MOD(myind, para_env%num_pe) /= para_env%mepos) CYCLE
            END IF
            LIndMM = per_pot%mm_atom_index(Imm)
            IndMM = mm_atom_index(LIndMM)
            qt = mm_charges(LIndMM)
            IF (shells) THEN
               ra(:) = pbc(mm_particles(LIndMM)%r-dOmmOqm, mm_cell)+dOmmOqm
            ELSE
               ra(:) = pbc(mm_particles(IndMM)%r-dOmmOqm, mm_cell)+dOmmOqm
            END IF
            ! Possible Spherical Cutoff
            IF (qmmm_spherical_cutoff(1) > 0.0_dp) THEN
               CALL spherical_cutoff_factor(qmmm_spherical_cutoff, ra, sph_chrg_factor)
               qt = qt*sph_chrg_factor
            END IF
            IF (ABS(qt) <= EPSILON(0.0_dp)) CYCLE Atoms
            rt1 = ra(1)
            rt2 = ra(2)
            rt3 = ra(3)
            LoopOnGrid: DO k = bo(1, 3), bo(2, 3)
               my_k = k-gbo(1, 3)
               xs3 = REAL(my_k, dp)*dr3c
               my_j = bo(1, 2)-gbo(1, 2)
               xs2 = REAL(my_j, dp)*dr2c
               rv3 = rt3-xs3
               vec(3) = rv3
               ivec(3) = FLOOR(vec(3)/pw%pw_grid%dr(3))
               xd3 = (vec(3)/dr3)-REAL(ivec(3), kind=dp)
               ik1 = MODULO(ivec(3)-1, npts(3))+1
               ik2 = MODULO(ivec(3), npts(3))+1
               ik3 = MODULO(ivec(3)+1, npts(3))+1
               ik4 = MODULO(ivec(3)+2, npts(3))+1
               p1 = 3.0_dp+xd3
               p2 = p1*p1
               p3 = p2*p1
               q1 = 2.0_dp+xd3
               q2 = q1*q1
               q3 = q2*q1
               r1 = 1.0_dp+xd3
               r2 = r1*r1
               r3 = r2*r1
               u1 = xd3
               u2 = u1*u1
               u3 = u2*u1
               v1 = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*p1+12.0_dp*p2-p3)
               v2 = -22.0_dp/3.0_dp+10.0_dp*q1-4.0_dp*q2+0.5_dp*q3
               v3 = 2.0_dp/3.0_dp-2.0_dp*r1+2.0_dp*r2-0.5_dp*r3
               v4 = 1.0_dp/6.0_dp*u3
               DO j = bo(1, 2), bo(2, 2)
                  xs1 = (bo(1, 1)-gbo(1, 1))*dr1c
                  rv2 = rt2-xs2
                  vec(2) = rv2
                  ivec(2) = FLOOR(vec(2)/pw%pw_grid%dr(2))
                  xd2 = (vec(2)/dr2)-REAL(ivec(2), kind=dp)
                  ij1 = MODULO(ivec(2)-1, npts(2))+1
                  ij2 = MODULO(ivec(2), npts(2))+1
                  ij3 = MODULO(ivec(2)+1, npts(2))+1
                  ij4 = MODULO(ivec(2)+2, npts(2))+1
                  e1 = 3.0_dp+xd2
                  e2 = e1*e1
                  e3 = e2*e1
                  f1 = 2.0_dp+xd2
                  f2 = f1*f1
                  f3 = f2*f1
                  g1 = 1.0_dp+xd2
                  g2 = g1*g1
                  g3 = g2*g1
                  h1 = xd2
                  h2 = h1*h1
                  h3 = h2*h1
                  s1 = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*e1+12.0_dp*e2-e3)
                  s2 = -22.0_dp/3.0_dp+10.0_dp*f1-4.0_dp*f2+0.5_dp*f3
                  s3 = 2.0_dp/3.0_dp-2.0_dp*g1+2.0_dp*g2-0.5_dp*g3
                  s4 = 1.0_dp/6.0_dp*h3
                  DO i = bo(1, 1), bo(2, 1)
                     rv1 = rt1-xs1
                     vec(1) = rv1
                     ivec(1) = FLOOR(vec(1)/pw%pw_grid%dr(1))
                     xd1 = (vec(1)/dr1)-REAL(ivec(1), kind=dp)
                     ii1 = MODULO(ivec(1)-1, npts(1))+1
                     ii2 = MODULO(ivec(1), npts(1))+1
                     ii3 = MODULO(ivec(1)+1, npts(1))+1
                     ii4 = MODULO(ivec(1)+2, npts(1))+1
                     !
                     ! Spline Interpolation
                     !

                     a1 = 3.0_dp+xd1
                     a2 = a1*a1
                     a3 = a2*a1
                     b1 = 2.0_dp+xd1
                     b2 = b1*b1
                     b3 = b2*b1
                     c1 = 1.0_dp+xd1
                     c2 = c1*c1
                     c3 = c2*c1
                     d1 = xd1
                     d2 = d1*d1
                     d3 = d2*d1
                     t1 = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*a1+12.0_dp*a2-a3)
                     t2 = -22.0_dp/3.0_dp+10.0_dp*b1-4.0_dp*b2+0.5_dp*b3
                     t3 = 2.0_dp/3.0_dp-2.0_dp*c1+2.0_dp*c2-0.5_dp*c3
                     t4 = 1.0_dp/6.0_dp*d3

                     abc_X(1, 1) = grid(ii1, ij1, ik1)*v1+grid(ii1, ij1, ik2)*v2+grid(ii1, ij1, ik3)*v3+grid(ii1, ij1, ik4)*v4
                     abc_X(1, 2) = grid(ii1, ij2, ik1)*v1+grid(ii1, ij2, ik2)*v2+grid(ii1, ij2, ik3)*v3+grid(ii1, ij2, ik4)*v4
                     abc_X(1, 3) = grid(ii1, ij3, ik1)*v1+grid(ii1, ij3, ik2)*v2+grid(ii1, ij3, ik3)*v3+grid(ii1, ij3, ik4)*v4
                     abc_X(1, 4) = grid(ii1, ij4, ik1)*v1+grid(ii1, ij4, ik2)*v2+grid(ii1, ij4, ik3)*v3+grid(ii1, ij4, ik4)*v4
                     abc_X(2, 1) = grid(ii2, ij1, ik1)*v1+grid(ii2, ij1, ik2)*v2+grid(ii2, ij1, ik3)*v3+grid(ii2, ij1, ik4)*v4
                     abc_X(2, 2) = grid(ii2, ij2, ik1)*v1+grid(ii2, ij2, ik2)*v2+grid(ii2, ij2, ik3)*v3+grid(ii2, ij2, ik4)*v4
                     abc_X(2, 3) = grid(ii2, ij3, ik1)*v1+grid(ii2, ij3, ik2)*v2+grid(ii2, ij3, ik3)*v3+grid(ii2, ij3, ik4)*v4
                     abc_X(2, 4) = grid(ii2, ij4, ik1)*v1+grid(ii2, ij4, ik2)*v2+grid(ii2, ij4, ik3)*v3+grid(ii2, ij4, ik4)*v4
                     abc_X(3, 1) = grid(ii3, ij1, ik1)*v1+grid(ii3, ij1, ik2)*v2+grid(ii3, ij1, ik3)*v3+grid(ii3, ij1, ik4)*v4
                     abc_X(3, 2) = grid(ii3, ij2, ik1)*v1+grid(ii3, ij2, ik2)*v2+grid(ii3, ij2, ik3)*v3+grid(ii3, ij2, ik4)*v4
                     abc_X(3, 3) = grid(ii3, ij3, ik1)*v1+grid(ii3, ij3, ik2)*v2+grid(ii3, ij3, ik3)*v3+grid(ii3, ij3, ik4)*v4
                     abc_X(3, 4) = grid(ii3, ij4, ik1)*v1+grid(ii3, ij4, ik2)*v2+grid(ii3, ij4, ik3)*v3+grid(ii3, ij4, ik4)*v4
                     abc_X(4, 1) = grid(ii4, ij1, ik1)*v1+grid(ii4, ij1, ik2)*v2+grid(ii4, ij1, ik3)*v3+grid(ii4, ij1, ik4)*v4
                     abc_X(4, 2) = grid(ii4, ij2, ik1)*v1+grid(ii4, ij2, ik2)*v2+grid(ii4, ij2, ik3)*v3+grid(ii4, ij2, ik4)*v4
                     abc_X(4, 3) = grid(ii4, ij3, ik1)*v1+grid(ii4, ij3, ik2)*v2+grid(ii4, ij3, ik3)*v3+grid(ii4, ij3, ik4)*v4
                     abc_X(4, 4) = grid(ii4, ij4, ik1)*v1+grid(ii4, ij4, ik2)*v2+grid(ii4, ij4, ik3)*v3+grid(ii4, ij4, ik4)*v4

                     abc_X_Y(1) = abc_X(1, 1)*t1+abc_X(2, 1)*t2+abc_X(3, 1)*t3+abc_X(4, 1)*t4
                     abc_X_Y(2) = abc_X(1, 2)*t1+abc_X(2, 2)*t2+abc_X(3, 2)*t3+abc_X(4, 2)*t4
                     abc_X_Y(3) = abc_X(1, 3)*t1+abc_X(2, 3)*t2+abc_X(3, 3)*t3+abc_X(4, 3)*t4
                     abc_X_Y(4) = abc_X(1, 4)*t1+abc_X(2, 4)*t2+abc_X(3, 4)*t3+abc_X(4, 4)*t4

                     val = abc_X_Y(1)*s1+abc_X_Y(2)*s2+abc_X_Y(3)*s3+abc_X_Y(4)*s4

                     grid2(i, j, k) = grid2(i, j, k)-val*qt
                     xs1 = xs1+dr1c
                  END DO
                  xs2 = xs2+dr2c
               END DO
            END DO LoopOnGrid
         END DO Atoms
      END DO Radius
      CALL timestop(handle)
   END SUBROUTINE qmmm_elec_with_gaussian_LG

! **************************************************************************************************
!> \brief Compute the QM/MM electrostatic Interaction collocating
!>      (1/R - Sum_NG Gaussians) on the coarser grid level.
!>      Long Range QM/MM Electrostatic Potential with Gaussian - Low Level
!> \param pgfs ...
!> \param grid ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param mm_particles ...
!> \param para_env ...
!> \param potentials ...
!> \param mm_cell ...
!> \param dOmmOqm ...
!> \param par_scheme ...
!> \param qmmm_spherical_cutoff ...
!> \param shells ...
!> \par History
!>      07.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_elec_with_gaussian_LR(pgfs, grid, mm_charges, mm_atom_index, &
                                         mm_particles, para_env, potentials, &
                                         mm_cell, dOmmOqm, par_scheme, qmmm_spherical_cutoff, shells)
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(pw_type), POINTER                             :: grid
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: potentials
      TYPE(cell_type), POINTER                           :: mm_cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dOmmOqm
      INTEGER, INTENT(IN)                                :: par_scheme
      REAL(KIND=dp), DIMENSION(2), INTENT(IN)            :: qmmm_spherical_cutoff
      LOGICAL                                            :: shells

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_elec_with_gaussian_LR', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, Imm, IndMM, IRadTyp, ix, j, &
                                                            k, LIndMM, my_j, my_k, myind, n1, n2, &
                                                            n3
      INTEGER, DIMENSION(2, 3)                           :: bo, gbo
      REAL(KIND=dp)                                      :: dr1, dr2, dr3, dx, qt, r, r2, rt1, rt2, &
                                                            rt3, rv1, rv2, rv3, rx, rx2, rx3, &
                                                            sph_chrg_factor, Term, xs1, xs2, xs3
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pot0_2
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid2
      TYPE(qmmm_pot_type), POINTER                       :: pot

      CALL timeset(routineN, handle)
      n1 = grid%pw_grid%npts(1)
      n2 = grid%pw_grid%npts(2)
      n3 = grid%pw_grid%npts(3)
      dr1 = grid%pw_grid%dr(1)
      dr2 = grid%pw_grid%dr(2)
      dr3 = grid%pw_grid%dr(3)
      gbo = grid%pw_grid%bounds
      bo = grid%pw_grid%bounds_local
      grid2 => grid%cr3d
      IF (par_scheme == do_par_atom) myind = 0
      Radius: DO IRadTyp = 1, SIZE(pgfs)
         pot => potentials(IRadTyp)%pot
         dx = Pot%dx
         pot0_2 => Pot%pot0_2
         Atoms: DO Imm = 1, SIZE(pot%mm_atom_index)
            IF (par_scheme == do_par_atom) THEN
               myind = myind+1
               IF (MOD(myind, para_env%num_pe) /= para_env%mepos) CYCLE
            END IF
            LIndMM = pot%mm_atom_index(Imm)
            IndMM = mm_atom_index(LIndMM)
            ra(:) = pbc(mm_particles(IndMM)%r-dOmmOqm, mm_cell)+dOmmOqm
            qt = mm_charges(LIndMM)
            IF (shells) &
               ra(:) = pbc(mm_particles(LIndMM)%r-dOmmOqm, mm_cell)+dOmmOqm
            ! Possible Spherical Cutoff
            IF (qmmm_spherical_cutoff(1) > 0.0_dp) THEN
               CALL spherical_cutoff_factor(qmmm_spherical_cutoff, ra, sph_chrg_factor)
               qt = qt*sph_chrg_factor
            END IF
            IF (ABS(qt) <= EPSILON(0.0_dp)) CYCLE Atoms
            rt1 = ra(1)
            rt2 = ra(2)
            rt3 = ra(3)
            LoopOnGrid: DO k = bo(1, 3), bo(2, 3)
               my_k = k-gbo(1, 3)
               xs3 = REAL(my_k, dp)*dr3
               my_j = bo(1, 2)-gbo(1, 2)
               xs2 = REAL(my_j, dp)*dr2
               rv3 = rt3-xs3
               DO j = bo(1, 2), bo(2, 2)
                  xs1 = (bo(1, 1)-gbo(1, 1))*dr1
                  rv2 = rt2-xs2
                  DO i = bo(1, 1), bo(2, 1)
                     rv1 = rt1-xs1
                     r2 = rv1*rv1+rv2*rv2+rv3*rv3
                     r = SQRT(r2)
                     ix = FLOOR(r/dx)+1
                     rx = (r-REAL(ix-1, dp)*dx)/dx
                     rx2 = rx*rx
                     rx3 = rx2*rx
                     Term = pot0_2(1, ix)*(1.0_dp-3.0_dp*rx2+2.0_dp*rx3) &
                            +pot0_2(2, ix)*(rx-2.0_dp*rx2+rx3) &
                            +pot0_2(1, ix+1)*(3.0_dp*rx2-2.0_dp*rx3) &
                            +pot0_2(2, ix+1)*(-rx2+rx3)
                     grid2(i, j, k) = grid2(i, j, k)-Term*qt
                     xs1 = xs1+dr1
                  END DO
                  xs2 = xs2+dr2
               END DO
            END DO LoopOnGrid
         END DO Atoms
      END DO Radius
      CALL timestop(handle)
   END SUBROUTINE qmmm_elec_with_gaussian_LR

END MODULE qmmm_gpw_energy
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of the derivative of the QMMM Hamiltonian integral
!>      matrix <a|\sum_i q_i|b> for semi-empirical methods
!> \author Teodoro Laino - 04.2007 [tlaino]
! **************************************************************************************************
MODULE qmmm_se_forces
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: &
        do_method_am1, do_method_mndo, do_method_mndod, do_method_pchg, do_method_pdg, &
        do_method_pm3, do_method_pm6, do_method_pnnl, do_method_rm1
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE multipole_types,                 ONLY: do_multipole_none
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_type,&
                                              qmmm_pot_p_type,&
                                              qmmm_pot_type
   USE qmmm_util,                       ONLY: spherical_cutoff_factor
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_qmmm_types,                ONLY: qs_ks_qmmm_env_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE semi_empirical_int_arrays,       ONLY: se_orbital_pointer
   USE semi_empirical_integrals,        ONLY: dcorecore,&
                                              drotnuc
   USE semi_empirical_types,            ONLY: get_se_param,&
                                              se_int_control_type,&
                                              se_taper_type,&
                                              semi_empirical_create,&
                                              semi_empirical_release,&
                                              semi_empirical_type,&
                                              setup_se_int_control_type
   USE semi_empirical_utils,            ONLY: get_se_type,&
                                              se_param_set_default
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_se_forces'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   PUBLIC :: deriv_se_qmmm_matrix

CONTAINS

! **************************************************************************************************
!> \brief Constructs the derivative w.r.t. 1-el semi-empirical hamiltonian
!>      QMMM terms
!> \param qs_env ...
!> \param qmmm_env ...
!> \param particles_mm ...
!> \param mm_cell ...
!> \param para_env ...
!> \param calc_force ...
!> \param Forces ...
!> \param Forces_added_charges ...
!> \author Teodoro Laino 04.2007 [created]
! **************************************************************************************************
   SUBROUTINE deriv_se_qmmm_matrix(qs_env, qmmm_env, particles_mm, mm_cell, para_env, &
                                   calc_force, Forces, Forces_added_charges)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_mm
      TYPE(cell_type), POINTER                           :: mm_cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(in), OPTIONAL                      :: calc_force
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Forces, Forces_added_charges

      CHARACTER(len=*), PARAMETER :: routineN = 'deriv_se_qmmm_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, ikind, iqm, ispin, &
                                                            itype, natom, natorb_a, nkind, &
                                                            number_qm_atoms
      INTEGER, DIMENSION(:), POINTER                     :: list
      LOGICAL                                            :: anag, defined, found
      REAL(KIND=dp)                                      :: delta, enuclear
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Forces_QM, p_block_a
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_p
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_qm
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_qmmm_env_type), POINTER                 :: ks_qmmm_env_loc
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(se_int_control_type)                          :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a, se_kind_mm

      CALL timeset(routineN, handle)
      IF (calc_force) THEN
         NULLIFY (rho, atomic_kind_set, qs_kind_set, se_taper)
         NULLIFY (se_kind_a, se_kind_mm, particles_qm)
         CALL get_qs_env(qs_env=qs_env, &
                         rho=rho, &
                         se_taper=se_taper, &
                         atomic_kind_set=atomic_kind_set, &
                         qs_kind_set=qs_kind_set, &
                         ks_qmmm_env=ks_qmmm_env_loc, &
                         dft_control=dft_control, &
                         particle_set=particles_qm, &
                         natom=number_qm_atoms)
         SELECT CASE (dft_control%qs_control%method_id)
         CASE (do_method_rm1, do_method_am1, do_method_mndo, do_method_pdg, &
               do_method_pm3, do_method_pm6, do_method_mndod, do_method_pnnl)
            ! Go on with the calculation..
         CASE DEFAULT
            ! Otherwise stop..
            CALL cp__b("qmmm_se_forces.F",126,"Method not available")
         END SELECT
         anag = dft_control%qs_control%se_control%analytical_gradients
         delta = dft_control%qs_control%se_control%delta
         ! Setup SE integral control type
         CALL setup_se_int_control_type( &
            se_int_control, shortrange=.FALSE., do_ewald_r3=.FALSE., &
            do_ewald_gks=.FALSE., integral_screening=dft_control%qs_control%se_control%integral_screening, &
            max_multipole=do_multipole_none, pc_coulomb_int=.FALSE.)

         ! Create a fake semi-empirical type to handle the classical atom
         ALLOCATE (Forces_QM(3, number_qm_atoms))
         CALL semi_empirical_create(se_kind_mm)
         CALL se_param_set_default(se_kind_mm, 0, do_method_pchg)
         itype = get_se_type(se_kind_mm%typ)
         nkind = SIZE(atomic_kind_set)
         enuclear = 0.0_dp
         Forces_QM = 0.0_dp
         CALL qs_rho_get(rho, rho_ao=matrix_p)

         DO ispin = 1, dft_control%nspins
            iqm = 0
            Kinds: DO ikind = 1, nkind
               CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=list)
               CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
               CALL get_se_param(se_kind_a, &
                                 defined=defined, &
                                 natorb=natorb_a)
               IF (.NOT. defined .OR. natorb_a < 1) CYCLE
               Atoms: DO i = 1, SIZE(list)
                  iqm = iqm+1
                  iatom = list(i)
                  ! Give back block
                  NULLIFY (p_block_a)
                  CALL cp_dbcsr_get_block_p(matrix=matrix_p(ispin)%matrix, &
                                            row=iatom, col=iatom, BLOCK=p_block_a, found=found)

                  IF (ASSOCIATED(p_block_a)) THEN
                     ! Expand derivative of geometrical factors
                     CALL deriv_se_qmmm_matrix_low(p_block_a, &
                                                   se_kind_a, &
                                                   se_kind_mm, &
                                                   qmmm_env%Potentials, &
                                                   particles_mm, &
                                                   qmmm_env%mm_atom_chrg, &
                                                   qmmm_env%mm_atom_index, &
                                                   mm_cell, &
                                                   iatom, &
                                                   itype, &
                                                   Forces, &
                                                   Forces_QM(:, iqm), &
                                                   se_taper, &
                                                   se_int_control, &
                                                   anag, &
                                                   delta, &
                                                   qmmm_env%spherical_cutoff, &
                                                   particles_qm)
                     ! Possibly added charges
                     IF (qmmm_env%move_mm_charges .OR. qmmm_env%add_mm_charges) THEN
                        CALL deriv_se_qmmm_matrix_low(p_block_a, &
                                                      se_kind_a, &
                                                      se_kind_mm, &
                                                      qmmm_env%added_charges%potentials, &
                                                      qmmm_env%added_charges%added_particles, &
                                                      qmmm_env%added_charges%mm_atom_chrg, &
                                                      qmmm_env%added_charges%mm_atom_index, &
                                                      mm_cell, &
                                                      iatom, &
                                                      itype, &
                                                      Forces_added_charges, &
                                                      Forces_QM(:, iqm), &
                                                      se_taper, &
                                                      se_int_control, &
                                                      anag, &
                                                      delta, &
                                                      qmmm_env%spherical_cutoff, &
                                                      particles_qm)
                     END IF
                  END IF
               END DO Atoms
            END DO Kinds
         END DO
         IF(.NOT.(iqm == number_qm_atoms))CALL cp__a("qmmm_se_forces.F",208)
         ! Transfer QM gradients to the QM particles..
         CALL mp_sum(Forces_QM, para_env%group)
         iqm = 0
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=list)
            CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
            CALL get_se_param(se_kind_a, &
                              defined=defined, &
                              natorb=natorb_a)
            IF (.NOT. defined .OR. natorb_a < 1) CYCLE
            DO i = 1, SIZE(list)
               iqm = iqm+1
               iatom = qmmm_env%qm_atom_index(list(i))
               particles_mm(iatom)%f(:) = particles_mm(iatom)%f(:)+Forces_QM(:, iqm)
            END DO
         END DO
         ! MM forces will be handled directly from the QMMM module in the same way
         ! as for GPW/GAPW methods
         DEALLOCATE (Forces_QM)
         CALL semi_empirical_release(se_kind_mm)

      END IF
      CALL timestop(handle)
   END SUBROUTINE deriv_se_qmmm_matrix

! **************************************************************************************************
!> \brief Low Level : Computes derivatives of the 1-el semi-empirical QMMM
!>                  hamiltonian block w.r.t. MM and QM coordinates
!> \param p_block_a ...
!> \param se_kind_a ...
!> \param se_kind_mm ...
!> \param potentials ...
!> \param particles_mm ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param mm_cell ...
!> \param IndQM ...
!> \param itype ...
!> \param forces ...
!> \param forces_qm ...
!> \param se_taper ...
!> \param se_int_control ...
!> \param anag ...
!> \param delta ...
!> \param qmmm_spherical_cutoff ...
!> \param particles_qm ...
!> \author Teodoro Laino 04.2007 [created]
! **************************************************************************************************
   SUBROUTINE deriv_se_qmmm_matrix_low(p_block_a, se_kind_a, se_kind_mm, &
                                       potentials, particles_mm, mm_charges, mm_atom_index, &
                                       mm_cell, IndQM, itype, forces, forces_qm, se_taper, &
                                       se_int_control, anag, delta, qmmm_spherical_cutoff, particles_qm)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p_block_a
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a, se_kind_mm
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: potentials
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_mm
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(cell_type), POINTER                           :: mm_cell
      INTEGER, INTENT(IN)                                :: IndQM, itype
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: forces
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: forces_qm
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      LOGICAL, INTENT(IN)                                :: anag
      REAL(KIND=dp), INTENT(IN)                          :: delta, qmmm_spherical_cutoff(2)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_qm

      CHARACTER(len=*), PARAMETER :: routineN = 'deriv_se_qmmm_matrix_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i1, i1L, i2, Imm, Imp, IndMM, &
                                                            Ipot, j1, j1L
      REAL(KIND=dp)                                      :: rt1, rt2, rt3, sph_chrg_factor
      REAL(KIND=dp), DIMENSION(3)                        :: denuc, force_ab, r_pbc, rij
      REAL(KIND=dp), DIMENSION(3, 45)                    :: de1b
      TYPE(qmmm_pot_type), POINTER                       :: Pot

      CALL timeset(routineN, handle)
      ! Loop Over MM atoms - parallelization over MM atoms...
      ! Loop over Pot stores atoms with the same charge
      MainLoopPot: DO Ipot = 1, SIZE(Potentials)
         Pot => Potentials(Ipot)%Pot
         ! Loop over atoms belonging to this type
         LoopMM: DO Imp = 1, SIZE(Pot%mm_atom_index)
            Imm = Pot%mm_atom_index(Imp)
            IndMM = mm_atom_index(Imm)
            r_pbc = pbc(particles_mm(IndMM)%r-particles_qm(IndQM)%r, mm_cell)
            rt1 = r_pbc(1)
            rt2 = r_pbc(2)
            rt3 = r_pbc(3)
            rij = (/rt1, rt2, rt3/)
            se_kind_mm%zeff = mm_charges(Imm)
            ! Computes the screening factor for the spherical cutoff
            IF (qmmm_spherical_cutoff(1) > 0.0_dp) THEN
               CALL spherical_cutoff_factor(qmmm_spherical_cutoff, rij, sph_chrg_factor)
               se_kind_mm%zeff = se_kind_mm%zeff*sph_chrg_factor
            END IF
            IF (ABS(se_kind_mm%zeff) <= EPSILON(0.0_dp)) CYCLE
            ! Integrals derivatives involving QM - MM atoms
            CALL drotnuc(se_kind_a, se_kind_mm, rij, itype=itype, de1b=de1b, &
                         se_int_control=se_int_control, anag=anag, delta=delta, &
                         se_taper=se_taper)
            CALL dcorecore(se_kind_a, se_kind_mm, rij, itype=itype, denuc=denuc, &
                           se_int_control=se_int_control, anag=anag, delta=delta, &
                           se_taper=se_taper)
            ! Nucler - Nuclear term
            force_ab(1:3) = -denuc(1:3)
            ! Force contribution from the QMMM Hamiltonian
            i2 = 0
            DO i1L = 1, se_kind_a%natorb
               i1 = se_orbital_pointer(i1L)
               DO j1L = 1, i1L-1
                  j1 = se_orbital_pointer(j1L)
                  i2 = i2+1
                  force_ab = force_ab-2.0_dp*de1b(:, i2)*p_block_a(i1, j1)
               END DO
               j1 = se_orbital_pointer(j1L)
               i2 = i2+1
               force_ab = force_ab-de1b(:, i2)*p_block_a(i1, j1)
            END DO
            ! The array of QM forces are really the forces
            forces_qm(:) = forces_qm(:)-force_ab
            ! The one of MM atoms are instead gradients
            forces(:, Imm) = forces(:, Imm)-force_ab
         END DO LoopMM
      END DO MainLoopPot
      CALL timestop(handle)
   END SUBROUTINE deriv_se_qmmm_matrix_low

END MODULE qmmm_se_forces
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to compute energy and forces in a QM/MM calculation
!> \par History
!>      05.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_gpw_forces
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_spline_utils,                 ONLY: pw_restrict_s3
   USE cube_utils,                      ONLY: cube_info_type
   USE input_constants,                 ONLY: do_par_atom,&
                                              do_qmmm_coulomb,&
                                              do_qmmm_gauss,&
                                              do_qmmm_none,&
                                              do_qmmm_pcharge,&
                                              do_qmmm_swave,&
                                              spline3_nopbc_interp,&
                                              spline3_pbc_interp
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_irecv,&
                                              mp_isend,&
                                              mp_sum,&
                                              mp_wait
   USE mm_collocate_potential,          ONLY: collocate_gf_rspace_NoPBC,&
                                              integrate_gf_rspace_NoPBC
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_integral_ab,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type,&
                                              pw_pools_create_pws,&
                                              pw_pools_give_back_pws
   USE pw_types,                        ONLY: REALDATA3D,&
                                              REALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qmmm_gaussian_types,             ONLY: qmmm_gaussian_p_type,&
                                              qmmm_gaussian_type
   USE qmmm_gpw_energy,                 ONLY: qmmm_elec_with_gaussian,&
                                              qmmm_elec_with_gaussian_LG,&
                                              qmmm_elec_with_gaussian_LR
   USE qmmm_se_forces,                  ONLY: deriv_se_qmmm_matrix
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_type,&
                                              qmmm_per_pot_p_type,&
                                              qmmm_per_pot_type,&
                                              qmmm_pot_p_type,&
                                              qmmm_pot_type
   USE qmmm_util,                       ONLY: spherical_cutoff_factor
   USE qs_dftb_matrices,                ONLY: deriv_dftb_qmmm_matrix,&
                                              deriv_dftb_qmmm_matrix_pc
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_qmmm_types,                ONLY: qs_ks_qmmm_env_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

  PRIVATE
  LOGICAL,       PARAMETER, PRIVATE    :: debug_this_module=.FALSE.
  REAL(KIND=dp), PARAMETER, PRIVATE    :: Dx = 0.01_dp     ! Debug Variables
  REAL(KIND=dp), PARAMETER, PRIVATE    :: MaxErr = 10.0_dp ! Debug Variables
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_gpw_forces'
  PUBLIC :: qmmm_forces

CONTAINS

! **************************************************************************************************
!> \brief General driver to Compute the contribution
!>      to the forces due to the QM/MM potential
!> \param qs_env ...
!> \param qmmm_env ...
!> \param mm_particles ...
!> \param calc_force ...
!> \param mm_cell ...
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE qmmm_forces(qs_env,qmmm_env,mm_particles,calc_force,mm_cell)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      LOGICAL, INTENT(in), OPTIONAL                      :: calc_force
      TYPE(cell_type), POINTER                           :: mm_cell

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_forces', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, image_IndMM, Imm, IndMM, &
                                                            ispin, iw
      LOGICAL                                            :: gapw, need_f, periodic
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Forces, Forces_added_charges, &
                                                            Forces_added_shells
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(pw_p_type), POINTER                           :: rho0_s_gs, rho_core
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pool
      TYPE(pw_type), POINTER                             :: rho_tot_r, rho_tot_r2
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_ks_qmmm_env_type), POINTER                 :: ks_qmmm_env_loc
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(section_vals_type), POINTER                   :: input_section, interp_section, &
                                                            print_section

    CALL timeset(routineN,handle)
    need_f = .TRUE.
    periodic = qmmm_env%periodic
    IF (PRESENT(calc_force)) need_f = calc_force
    NULLIFY(dft_control, ks_qmmm_env_loc, rho, pw_env, rho_tot_r, energy, Forces,&
            Forces_added_charges,input_section,rho0_s_gs, rho_r)
    CALL get_qs_env(qs_env=qs_env,&
         rho=rho,&
         rho_core=rho_core,&
         pw_env=pw_env,&
         energy=energy,&
         para_env=para_env,&
         input=input_section,&
         rho0_s_gs=rho0_s_gs,&
         dft_control=dft_control)

    CALL qs_rho_get(rho, rho_r=rho_r)

    logger => cp_get_default_logger()
    ks_qmmm_env_loc => qs_env%ks_qmmm_env
    interp_section => section_vals_get_subs_vals(input_section,"QMMM%INTERPOLATOR")
    print_section => section_vals_get_subs_vals(input_section,"QMMM%PRINT")
    iw=cp_print_key_unit_nr(logger,print_section,"PROGRAM_RUN_INFO",&
         extension=".qmmmLog")
    gapw = dft_control%qs_control%gapw
    ! If forces are required allocate these temporary arrays
    IF (need_f) THEN
       ALLOCATE(Forces(3,qmmm_env%num_mm_atoms))
       ALLOCATE(Forces_added_charges(3,qmmm_env%added_charges%num_mm_atoms))
       ALLOCATE(Forces_added_shells(3,qmmm_env%added_shells%num_mm_atoms))
       Forces(:,:)               = 0.0_dp
       Forces_added_charges(:,:) = 0.0_dp
       Forces_added_shells(:,:)  = 0.0_dp
    END IF
    IF (dft_control%qs_control%semi_empirical) THEN
       ! SEMIEMPIRICAL
       SELECT CASE(qmmm_env%qmmm_coupl_type)
       CASE(do_qmmm_coulomb)
          CALL deriv_se_qmmm_matrix(qs_env,qmmm_env,mm_particles,mm_cell,para_env,&
               need_f,Forces,Forces_added_charges)
       CASE(do_qmmm_pcharge)
          CALL cp__b("qmmm_gpw_forces.F",173,"Point Charge QM/MM electrostatic coupling not yet implemented for SE.")
       CASE(do_qmmm_gauss,do_qmmm_swave)
          CALL cp__b("qmmm_gpw_forces.F",175,"GAUSS or SWAVE QM/MM electrostatic coupling not yet implemented for SE.")
       CASE(do_qmmm_none)
          IF (iw>0) WRITE(iw,'(T2,"QMMM|",1X,A)')&
               "- No QM/MM Electrostatic coupling. Just Mechanical Coupling!"
       CASE DEFAULT
          IF (iw>0) WRITE(iw,'(T2,"QMMM|",1X,A)')"Unknown Coupling..."
          CALL cp__b("qmmm_gpw_forces.F",181,"")
       END SELECT
    ELSEIF (dft_control%qs_control%dftb) THEN
       ! DFTB
       SELECT CASE(qmmm_env%qmmm_coupl_type)
       CASE(do_qmmm_none)
          IF (iw>0) WRITE(iw,'(T2,"QMMM|",1X,A)')&
               "- No QM/MM Electrostatic coupling. Just Mechanical Coupling!"
       CASE(do_qmmm_coulomb)
          CALL deriv_dftb_qmmm_matrix(qs_env,qmmm_env,mm_particles,mm_cell,para_env,&
               need_f,Forces,Forces_added_charges)
       CASE(do_qmmm_pcharge)
          CALL deriv_dftb_qmmm_matrix_pc(qs_env,qmmm_env,mm_particles,mm_cell,para_env,&
               need_f,Forces,Forces_added_charges)
       CASE(do_qmmm_gauss,do_qmmm_swave)
          CALL cp__b("qmmm_gpw_forces.F",196,"GAUSS or SWAVE QM/MM electrostatic coupling not yet implemented for DFTB.")
       CASE DEFAULT
          IF (iw>0) WRITE(iw,'(T2,"QMMM|",1X,A)')"Unknown Coupling..."
          CALL cp__b("qmmm_gpw_forces.F",199,"")
       END SELECT
       IF (need_f) THEN
          Forces(:,:) = Forces(:,:) / REAL(para_env%num_pe,KIND=dp)
          Forces_added_charges(:,:) = Forces_added_charges(:,:) / REAL(para_env%num_pe,KIND=dp)
       ENDIF
    ELSE
       ! GPW/GAPW
       CALL pw_env_get(pw_env=pw_env,&
                       pw_pools=pw_pools,&
                       auxbas_pw_pool=auxbas_pool)
       CALL pw_pool_create_pw(auxbas_pool,rho_tot_r,&
                              in_space=REALSPACE,&
                              use_data=REALDATA3D)
       ! IF GAPW the core charge is replaced by the compensation charge
       IF(gapw) THEN
          IF( dft_control%qs_control%gapw_control%nopaw_as_gpw) THEN
            CALL pw_transfer(rho_core%pw,rho_tot_r)
            energy%qmmm_nu = pw_integral_ab ( rho_tot_r, ks_qmmm_env_loc%v_qmmm_rspace%pw)
            CALL pw_pool_create_pw(auxbas_pool,rho_tot_r2,&
                              in_space=REALSPACE,&
                              use_data=REALDATA3D)
            CALL pw_transfer(rho0_s_gs%pw,rho_tot_r2)
            CALL pw_axpy(rho_tot_r2,rho_tot_r)
            CALL pw_pool_give_back_pw(auxbas_pool,rho_tot_r2,accept_non_compatible=.TRUE.)
          ELSE
            CALL pw_transfer(rho0_s_gs%pw,rho_tot_r)
          !
          ! QM/MM Nuclear Electrostatic Potential already included through rho0
          !
            energy%qmmm_nu = 0.0_dp
          END IF
       ELSE
          CALL pw_transfer(rho_core%pw,rho_tot_r)
          !
          ! Computes the QM/MM Nuclear Electrostatic Potential
          !
          energy%qmmm_nu = pw_integral_ab ( rho_tot_r, ks_qmmm_env_loc%v_qmmm_rspace%pw)
       END IF
       IF (need_f) THEN
          !
          DO ispin=1,SIZE(rho_r)
             CALL pw_axpy(rho_r(ispin)%pw,rho_tot_r)
          END DO
          IF (iw>0) WRITE(iw,'(T2,"QMMM|",1X,A)')"Evaluating forces on MM atoms due to the:"
          ! Electrostatic Interaction type...
          SELECT CASE(qmmm_env%qmmm_coupl_type)
          CASE(do_qmmm_coulomb)
             CALL cp__b("qmmm_gpw_forces.F",247,"Coulomb QM/MM electrostatic coupling not implemented for GPW/GAPW.")
          CASE(do_qmmm_pcharge)
             CALL cp__b("qmmm_gpw_forces.F",249,"Point Charge QM/MM electrostatic coupling not yet implemented for GPW/GAPW.")
          CASE(do_qmmm_gauss,do_qmmm_swave)
             IF (iw>0) WRITE(iw,'(T2,"QMMM|",1X,A)')&
                  "- QM/MM Coupling computed collocating the Gaussian Potential Functions."
             CALL qmmm_forces_with_gaussian(rho=rho_tot_r,&
                                            qmmm_env=qmmm_env,&
                                            mm_particles=mm_particles,&
                                            aug_pools=qmmm_env%aug_pools,&
                                            auxbas_grid=qmmm_env%gridlevel_info%auxbas_grid,&
                                            coarser_grid=qmmm_env%gridlevel_info%coarser_grid,&
                                            para_env=para_env,&
                                            pw_pools=pw_pools,&
                                            eps_mm_rspace=qmmm_env%eps_mm_rspace,&
                                            cube_info=ks_qmmm_env_loc%cube_info,&
                                            Forces=Forces,&
                                            Forces_added_charges=Forces_added_charges,&
                                            Forces_added_shells=Forces_added_shells,&
                                            interp_section=interp_section,&
                                            iw=iw,&
                                            mm_cell=mm_cell)
          CASE(do_qmmm_none)
             IF (iw>0) WRITE(iw,'(T2,"QMMM|",1X,A)')&
                  "- No QM/MM Electrostatic coupling. Just Mechanical Coupling!"
          CASE DEFAULT
             IF (iw>0) WRITE(iw,'(T2,"QMMM|",1X,A)')"- Unknown Coupling..."
 CALL cp__b("qmmm_gpw_forces.F",274,"")
          END SELECT
       END IF
    END IF
    ! Correct Total Energy adding the contribution of the QM/MM nuclear interaction
    energy%total = energy%total +  energy%qmmm_nu
    ! Proceed if gradients are requested..
    IF (need_f) THEN
       !ikuo Temporary change to alleviate compiler problems on Intel with
       !array dimension of 0
       IF(qmmm_env%num_mm_atoms/=0)               CALL mp_sum(Forces,para_env%group)
       IF(qmmm_env%added_charges%num_mm_atoms/=0) CALL mp_sum(Forces_added_charges,para_env%group)
       IF(qmmm_env%added_shells%num_mm_atoms/=0) CALL mp_sum(Forces_added_shells,para_env%group)
       ! Debug Forces
       IF (debug_this_module) THEN
          IF ( (dft_control%qs_control%semi_empirical).OR.&
               (dft_control%qs_control%dftb))THEN
             WRITE(iw,*)"NO DEBUG AVAILABLE in module"//TRIM(routineN)
          ELSE
             ! Print Out Forces
             IF (iw>0) THEN
                DO Imm = 1, SIZE(qmmm_env%mm_atom_index)
                   WRITE(iw,*)"ANALYTICAL FORCES:"
                   IndMM = qmmm_env%mm_atom_index(Imm)
                   WRITE(iw,'(I6,3F15.9)') IndMM, Forces(:,Imm)
                END DO
             END IF
             CALL qmmm_debug_forces(rho=rho_tot_r,&
                                    qs_env=qs_env,&
                                    qmmm_env=qmmm_env,&
                                    Analytical_Forces=Forces,&
                                    mm_particles=mm_particles,&
                                    mm_atom_index=qmmm_env%mm_atom_index,&
                                    num_mm_atoms=qmmm_env%num_mm_atoms,&
                                    interp_section=interp_section,&
                                    mm_cell=mm_cell)
          END IF
       ENDIF
    END IF
    ! Give back rho_tot_t to auxbas_pool only for GPW/GAPW
    IF  ((.NOT.dft_control%qs_control%semi_empirical).AND.&
         (.NOT.dft_control%qs_control%dftb)) THEN
       CALL pw_pool_give_back_pw(auxbas_pool,rho_tot_r,&
            accept_non_compatible=.TRUE.)
    END IF
    IF (iw>0) THEN
       IF(.NOT. gapw) WRITE(iw,'(T2,"QMMM|",1X,A,T66,F15.9)')&
            "QM/MM Nuclear Electrostatic Potential :",energy%qmmm_nu
       WRITE(iw,'(T2,"QMMM|",1X,A,T66,F15.9)')&
            "QMMM Total Energy (QM + QMMM electronic + QMMM nuclear):",energy%total
       WRITE(iw,'(T2,"QMMM|",1X,A)')"MM energy NOT included in the above term!"//&
            " Check for:  FORCE_EVAL ( QMMM )"
       WRITE(iw,'(T2,"QMMM|",1X,A)')"that includes both QM, QMMM and MM energy terms!"
    END IF
    IF (need_f) THEN
       ! Transfer Forces
       DO Imm = 1, qmmm_env%num_mm_atoms
          IndMM = qmmm_env%mm_atom_index(Imm)

          !add image forces to Forces
          IF(qmmm_env%image_charge) THEN
             DO iatom=1,qmmm_env%num_image_mm_atoms
               image_IndMM=qmmm_env%image_charge_pot%image_mm_list(iatom)
               IF(image_IndMM.eq.IndMM) THEN
               Forces(:,Imm)=Forces(:,Imm)&
                             +qmmm_env%image_charge_pot%image_forcesMM(:,iatom)
               ENDIF
             ENDDO
          ENDIF

          ! Hack: In Forces there the gradients indeed...
          ! Minux sign to take care of this misunderstanding...
          mm_particles(IndMM)%f(:) = - Forces(:,Imm) + mm_particles(IndMM)%f(:)
       END DO
       DEALLOCATE(Forces)
       IF (qmmm_env%move_mm_charges.OR.qmmm_env%add_mm_charges) THEN
          DO Imm = 1, qmmm_env%added_charges%num_mm_atoms
             IndMM = qmmm_env%added_charges%mm_atom_index(Imm)
             ! Hack: In Forces there the gradients indeed...
             ! Minux sign to take care of this misunderstanding...
             qmmm_env%added_charges%added_particles(IndMM)%f(:) = - Forces_added_charges(:,Imm)
          END DO
       END IF
       DEALLOCATE(Forces_added_charges)
       IF (qmmm_env%added_shells%num_mm_atoms .GT. 0) THEN
          DO Imm = 1, qmmm_env%added_shells%num_mm_atoms
             IndMM = qmmm_env%added_shells%mm_core_index(Imm)
             ! Hack: In Forces there the gradients indeed...
             ! Minux sign to take care of this misunderstanding...
             qmmm_env%added_shells%added_particles(Imm)%f(:) = qmmm_env%added_shells%added_particles(Imm)%f(:) - &
                  Forces_added_shells(:,Imm)
                  
          END DO
       END IF
       DEALLOCATE(Forces_added_shells)
    END IF
    CALL cp_print_key_finished_output(iw,logger,print_section,"PROGRAM_RUN_INFO")
    CALL timestop(handle)

  END SUBROUTINE qmmm_forces

! **************************************************************************************************
!> \brief Evaluates the contribution to the forces due to the
!>      QM/MM potential computed collocating the Electrostatic
!>      Gaussian Potential.
!> \param rho ...
!> \param qmmm_env ...
!> \param mm_particles ...
!> \param aug_pools ...
!> \param auxbas_grid ...
!> \param coarser_grid ...
!> \param cube_info ...
!> \param para_env ...
!> \param eps_mm_rspace ...
!> \param pw_pools ...
!> \param Forces ...
!> \param Forces_added_charges ...
!> \param Forces_added_shells ...
!> \param interp_section ...
!> \param iw ...
!> \param mm_cell ...
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE qmmm_forces_with_gaussian(rho, qmmm_env, mm_particles,  &
       aug_pools, auxbas_grid, coarser_grid, cube_info, para_env,    &
       eps_mm_rspace, pw_pools, Forces, Forces_added_charges, Forces_added_shells,&
       interp_section, iw, mm_cell)
      TYPE(pw_type), POINTER                             :: rho
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: aug_pools
      INTEGER, INTENT(IN)                                :: auxbas_grid, coarser_grid
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Forces, Forces_added_charges, &
                                                            Forces_added_shells
      TYPE(section_vals_type), POINTER                   :: interp_section
      INTEGER, INTENT(IN)                                :: iw
      TYPE(cell_type), POINTER                           :: mm_cell

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_forces_with_gaussian', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, i, igrid, j, k, &
                                                            kind_interp, me, ngrids, request, stat
      INTEGER, DIMENSION(3)                              :: glb, gub, lb, ub
      INTEGER, DIMENSION(:), POINTER                     :: pos_of_x
      LOGICAL                                            :: shells
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: tmp
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: grids

! Statements

    CALL timeset(routineN,handle)
    NULLIFY(grids,tmp)
    IF(.NOT.(ASSOCIATED(mm_particles)))CALL cp__a("qmmm_gpw_forces.F",433)
    IF(.NOT.(ASSOCIATED(qmmm_env%mm_atom_chrg)))CALL cp__a("qmmm_gpw_forces.F",434)
    IF(.NOT.(ASSOCIATED(qmmm_env%mm_atom_index)))CALL cp__a("qmmm_gpw_forces.F",435)
    IF(.NOT.(ASSOCIATED(Forces)))CALL cp__a("qmmm_gpw_forces.F",436)
    !Statements
    ngrids=SIZE(pw_pools)
    CALL pw_pools_create_pws(aug_pools,grids,use_data=REALDATA3D,in_space=REALSPACE)
    DO igrid=1,ngrids
       CALL pw_zero(grids(igrid)%pw)
    END DO
    ! Collocate Density on multigrids
    lb = rho%pw_grid%bounds_local(1,:)
    ub = rho%pw_grid%bounds_local(2,:)
    grids(auxbas_grid) % pw % cr3d(lb(1):ub(1),&
                                   lb(2):ub(2),&
                                   lb(3):ub(3) )= rho % cr3d
    ! copy the boundaries
    DO i=lb(1),ub(1)
       grids(auxbas_grid) % pw %cr3d(i,ub(2)+1,ub(3)+1)=rho%cr3d(i,lb(2),lb(3))
    END DO
    DO k=lb(3),ub(3)
       DO i=lb(1),ub(1)
          grids(auxbas_grid) % pw %cr3d(i,ub(2)+1,k)=rho%cr3d(i,lb(2),k)
       END DO
    END DO
    DO j=lb(2),ub(2)
       DO i=lb(1),ub(1)
          grids(auxbas_grid) % pw %cr3d(i,j,ub(3)+1)=rho%cr3d(i,j,lb(3))
       END DO
    END DO
    pos_of_x => grids(auxbas_grid)%pw%pw_grid%para%pos_of_x
    group=grids(auxbas_grid)%pw%pw_grid%para%group
    me=grids(auxbas_grid)%pw%pw_grid%para%my_pos
    glb=rho%pw_grid%bounds(1,:)
    gub=rho%pw_grid%bounds(2,:)
    IF ((pos_of_x(glb(1)) .EQ. me).AND.(pos_of_x(gub(1)) .EQ. me)) THEN
       DO k=lb(3),ub(3)
          DO j=lb(2),ub(2)
             grids(auxbas_grid) % pw %cr3d(ub(1)+1,j,k)=rho%cr3d(lb(1),j,k)
          END DO
          grids(auxbas_grid) % pw %cr3d(ub(1)+1,ub(2)+1,k)=rho%cr3d(lb(1),lb(2),k)
       END DO
       DO j=lb(2),ub(2)
          grids(auxbas_grid) % pw %cr3d(ub(1)+1,j,ub(3)+1)=rho%cr3d(lb(1),j,lb(3))
       END DO
       grids(auxbas_grid) % pw %cr3d(ub(1)+1,ub(2)+1,ub(3)+1)=rho%cr3d(lb(1),lb(2),lb(3))
    ELSE IF (pos_of_x(glb(1)) .EQ. me) THEN
       ALLOCATE(tmp(rho%pw_grid%bounds_local(1,2):rho%pw_grid%bounds_local(2,2),&
            rho%pw_grid%bounds_local(1,3):rho%pw_grid%bounds_local(2,3)),&
            stat=stat)
       IF(.NOT.(stat==0))CALL cp__a("qmmm_gpw_forces.F",483)
       tmp=rho%cr3d(lb(1),:,:)
       CALL mp_isend(msgin=tmp,dest=pos_of_x(rho%pw_grid%bounds(2,1)),comm=group,&
            request=request,tag=112)
       CALL mp_wait(request)
    ELSE IF (pos_of_x(gub(1)) .EQ. me) THEN
       ALLOCATE(tmp(rho%pw_grid%bounds_local(1,2):rho%pw_grid%bounds_local(2,2),&
            rho%pw_grid%bounds_local(1,3):rho%pw_grid%bounds_local(2,3)),&
            stat=stat)
       IF(.NOT.(stat==0))CALL cp__a("qmmm_gpw_forces.F",492)
       CALL mp_irecv(msgout=tmp,source=pos_of_x(rho%pw_grid%bounds(1,1)),&
            comm=group,request=request,tag=112)
       CALL mp_wait(request)

       DO k=lb(3),ub(3)
          DO j=lb(2),ub(2)
             grids(auxbas_grid) % pw %cr3d(ub(1)+1,j,k)=tmp(j,k)
          END DO
          grids(auxbas_grid) % pw %cr3d(ub(1)+1,ub(2)+1,k)=tmp(lb(2),k)
       END DO
       DO j=lb(2),ub(2)
          grids(auxbas_grid) % pw %cr3d(ub(1)+1,j,ub(3)+1)=tmp(j,lb(3))
       END DO
       grids(auxbas_grid) % pw %cr3d(ub(1)+1,ub(2)+1,ub(3)+1)=tmp(lb(2),lb(3))
    END IF
    IF (ASSOCIATED(tmp)) THEN
       DEALLOCATE(tmp)
    END IF
    ! Further setup of parallelization scheme
    IF (qmmm_env%par_scheme==do_par_atom) THEN
       CALL mp_sum(grids(auxbas_grid)%pw%cr3d,para_env%group)
    END IF
    ! RealSpace Interpolation
    CALL section_vals_val_get(interp_section,"kind", i_val=kind_interp)
    SELECT CASE(kind_interp)
    CASE(spline3_nopbc_interp, spline3_pbc_interp)
       ! Spline Interpolator
       DO Igrid =  auxbas_grid, SIZE(grids)-1
          CALL pw_restrict_s3(grids(Igrid  )%pw,&
                              grids(Igrid+1)%pw,&
                              aug_pools(Igrid+1)%pool,&
                              param_section=interp_section)
       END DO
    CASE DEFAULT
       CALL cp__b("qmmm_gpw_forces.F",527,"")
    END SELECT

    shells = .FALSE.
    CALL qmmm_force_with_gaussian_low(grids, mm_particles,&
         qmmm_env%mm_atom_chrg, qmmm_env%mm_atom_index,&
         qmmm_env%num_mm_atoms, cube_info, para_env, eps_mm_rspace, auxbas_grid, &
         coarser_grid, qmmm_env%pgfs, qmmm_env%potentials, Forces, aug_pools,    &
         mm_cell, qmmm_env%dOmmOqm, qmmm_env%periodic, qmmm_env%per_potentials,  &
         iw, qmmm_env%par_scheme, qmmm_env%spherical_cutoff, shells)

    IF (qmmm_env%move_mm_charges.OR.qmmm_env%add_mm_charges) THEN
       CALL qmmm_force_with_gaussian_low(grids, qmmm_env%added_charges%added_particles,&
            qmmm_env%added_charges%mm_atom_chrg,&
            qmmm_env%added_charges%mm_atom_index, qmmm_env%added_charges%num_mm_atoms,&
            cube_info, para_env, eps_mm_rspace, auxbas_grid, coarser_grid, qmmm_env%added_charges%pgfs,&
            qmmm_env%added_charges%potentials, Forces_added_charges, aug_pools, mm_cell, &
            qmmm_env%dOmmOqm, qmmm_env%periodic, qmmm_env%per_potentials, iw, qmmm_env%par_scheme,&
            qmmm_env%spherical_cutoff, shells)
    END IF

    IF (qmmm_env%added_shells%num_mm_atoms .GT. 0) THEN
       shells = .TRUE.
       CALL qmmm_force_with_gaussian_low(grids, qmmm_env%added_shells%added_particles,&
            qmmm_env%added_shells%mm_core_chrg, &
            qmmm_env%added_shells%mm_core_index, qmmm_env%added_shells%num_mm_atoms,&
            cube_info, para_env, eps_mm_rspace, auxbas_grid, coarser_grid, qmmm_env%added_shells%pgfs,&
            qmmm_env%added_shells%potentials, Forces_added_shells, aug_pools, mm_cell, &
            qmmm_env%dOmmOqm, qmmm_env%periodic, qmmm_env%added_shells%per_potentials, iw, qmmm_env%par_scheme,&
            qmmm_env%spherical_cutoff, shells)
    END IF

    CALL pw_pools_give_back_pws(aug_pools,grids)
    CALL timestop(handle)

  END SUBROUTINE qmmm_forces_with_gaussian

! **************************************************************************************************
!> \brief Evaluates the contribution to the forces due to the
!>      QM/MM potential computed collocating the Electrostatic
!>      Gaussian Potential. Low Level
!> \param grids ...
!> \param mm_particles ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param num_mm_atoms ...
!> \param cube_info ...
!> \param para_env ...
!> \param eps_mm_rspace ...
!> \param auxbas_grid ...
!> \param coarser_grid ...
!> \param pgfs ...
!> \param potentials ...
!> \param Forces ...
!> \param aug_pools ...
!> \param mm_cell ...
!> \param dOmmOqm ...
!> \param periodic ...
!> \param per_potentials ...
!> \param iw ...
!> \param par_scheme ...
!> \param qmmm_spherical_cutoff ...
!> \param shells ...
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE qmmm_force_with_gaussian_low(grids, mm_particles, mm_charges, &
       mm_atom_index, num_mm_atoms, cube_info, para_env, &
       eps_mm_rspace, auxbas_grid, coarser_grid, pgfs, potentials, Forces, &
       aug_pools, mm_cell, dOmmOqm, periodic, per_potentials, iw, par_scheme,&
       qmmm_spherical_cutoff, shells)
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: grids
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      INTEGER, INTENT(IN)                                :: num_mm_atoms
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace
      INTEGER, INTENT(IN)                                :: auxbas_grid, coarser_grid
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: Potentials
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Forces
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: aug_pools
      TYPE(cell_type), POINTER                           :: mm_cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dOmmOqm
      LOGICAL, INTENT(in)                                :: periodic
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER   :: per_potentials
      INTEGER, INTENT(IN)                                :: iw, par_scheme
      REAL(KIND=dp), INTENT(IN)                          :: qmmm_spherical_cutoff(2)
      LOGICAL, INTENT(in)                                :: shells

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_force_with_gaussian_low', &
         routineNb = 'qmmm_forces_gaussian_low', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle2, IGauss, ilevel, Imm, &
                                                            IndMM, IRadTyp, LIndMM, myind, &
                                                            n_rep_real(3)
      INTEGER, DIMENSION(2, 3)                           :: bo
      REAL(KIND=dp)                                      :: alpha, dvol, height, sph_chrg_factor, W
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: xdat, ydat, zdat
      REAL(KIND=dp), DIMENSION(3)                        :: force, ra
      TYPE(qmmm_gaussian_type), POINTER                  :: pgf
      TYPE(qmmm_per_pot_type), POINTER                   :: per_pot
      TYPE(qmmm_pot_type), POINTER                       :: pot

    CALL timeset(routineN,handle)
    CALL timeset(routineNb//"_G",handle2)
    NULLIFY(pgf, pot, per_pot)
    IF (par_scheme==do_par_atom) myind=0
    Radius: DO IRadTyp = 1, SIZE(pgfs)
       pgf => pgfs(IRadTyp)%pgf
       pot => potentials(IRadTyp)%pot
       n_rep_real = 0
       IF (periodic) THEN
          per_pot => per_potentials(IRadTyp)%pot
          n_rep_real = per_pot%n_rep_real
       END IF
       Gaussian: DO IGauss = 1, pgf%Number_of_Gaussians
          alpha     = 1.0_dp / pgf%Gk(IGauss)
          alpha     = alpha * alpha
          height    = pgf%Ak(IGauss)
          ilevel    = pgf%grid_level(IGauss)
          dvol      = grids(ilevel)%pw%pw_grid%dvol
          bo        = grids(ilevel) % pw % pw_grid % bounds_local
          ALLOCATE (xdat(2,bo(1,1):bo(2,1)))
          ALLOCATE (ydat(2,bo(1,2):bo(2,2)))
          ALLOCATE (zdat(2,bo(1,3):bo(2,3)))
          Atoms: DO Imm = 1, SIZE(pot%mm_atom_index)
             IF (par_scheme==do_par_atom) THEN
                myind = myind + 1
                IF (MOD(myind,para_env%num_pe)/=para_env%mepos) CYCLE Atoms
             END IF
             LIndMM    =   pot%mm_atom_index(Imm)
             IndMM     =   mm_atom_index(LIndMM)
             IF (shells) THEN
                ra(:)     =   pbc(mm_particles(Imm)%r-dOmmOqm, mm_cell)+dOmmOqm
             ELSE
                ra(:)     =   pbc(mm_particles(IndMM)%r-dOmmOqm,mm_cell)+dOmmOqm
             END IF
             W         =   mm_charges(LIndMM) * height
             force     =   0.0_dp
             ! Possible Spherical Cutoff
             IF (qmmm_spherical_cutoff(1)>0.0_dp) THEN
                CALL spherical_cutoff_factor(qmmm_spherical_cutoff, ra, sph_chrg_factor)
                W = W * sph_chrg_factor
             END IF
             IF (ABS(W)<= EPSILON(0.0_dp)) CYCLE Atoms
             CALL integrate_gf_rspace_NoPBC(zetp=alpha,&
                                            rp=ra,&
                                            scale=-1.0_dp,&
                                            W=W,&
                                            pwgrid=grids(ilevel)%pw,&
                                            cube_info=cube_info(ilevel),&
                                            eps_mm_rspace=eps_mm_rspace,&
                                            xdat=xdat,&
                                            ydat=ydat,&
                                            zdat=zdat,&
                                            bo=bo,&
                                            force=force,&
                                            n_rep_real=n_rep_real,&
                                            mm_cell=mm_cell)
             force = force * dvol
             Forces(:,LIndMM) = Forces(:,LIndMM) + force(:)
             !
             ! Debug Statement
             !
             IF (debug_this_module) THEN
                CALL debug_integrate_gf_rspace_NoPBC(ilevel=ilevel,&
                                                     zetp=alpha,&
                                                     rp=ra,&
                                                     W=W,&
                                                     pwgrid=grids(ilevel)%pw,&
                                                     cube_info=cube_info(ilevel),&
                                                     eps_mm_rspace=eps_mm_rspace,&
                                                     aug_pools=aug_pools,&
                                                     debug_force=force,&
                                                     mm_cell=mm_cell,&
                                                     auxbas_grid=auxbas_grid,&
                                                     n_rep_real=n_rep_real,&
                                                     iw=iw)
             END IF
          END DO Atoms
          DEALLOCATE (xdat)
          DEALLOCATE (ydat)
          DEALLOCATE (zdat)
       END DO Gaussian
    END DO Radius
    CALL timestop(handle2)
    CALL timeset(routineNb//"_R",handle2)
    IF (periodic) THEN
       CALL qmmm_forces_with_gaussian_LG  (pgfs=pgfs,&
                                           cgrid=grids(coarser_grid)%pw,&
                                           num_mm_atoms=num_mm_atoms,&
                                           mm_charges=mm_charges,&
                                           mm_atom_index=mm_atom_index,&
                                           mm_particles=mm_particles,&
                                           para_env=para_env,&
                                           coarser_grid_level=coarser_grid,&
                                           Forces=Forces,&
                                           per_potentials=per_potentials,&
                                           aug_pools=aug_pools,&
                                           mm_cell=mm_cell,&
                                           dOmmOqm=dOmmOqm,&
                                           iw=iw,&
                                           par_scheme=par_scheme,&
                                           qmmm_spherical_cutoff=qmmm_spherical_cutoff,&
                                           shells=shells)
    ELSE
       CALL qmmm_forces_with_gaussian_LR  (pgfs=pgfs,&
                                           cgrid=grids(coarser_grid)%pw,&
                                           num_mm_atoms=num_mm_atoms,&
                                           mm_charges=mm_charges,&
                                           mm_atom_index=mm_atom_index,&
                                           mm_particles=mm_particles,&
                                           para_env=para_env,&
                                           coarser_grid_level=coarser_grid,&
                                           Forces=Forces,&
                                           potentials=potentials,&
                                           aug_pools=aug_pools,&
                                           mm_cell=mm_cell,&
                                           dOmmOqm=dOmmOqm,&
                                           iw=iw,&
                                           par_scheme=par_scheme,&
                                           qmmm_spherical_cutoff=qmmm_spherical_cutoff,&
                                           shells=shells)
    END IF
    CALL timestop(handle2)
    CALL timestop(handle)
  END SUBROUTINE qmmm_force_with_gaussian_low

! **************************************************************************************************
!> \brief Evaluates the contribution to the forces due to the Long Range
!>      part of the QM/MM potential computed collocating the Electrostatic
!>      Gaussian Potential.
!> \param pgfs ...
!> \param cgrid ...
!> \param num_mm_atoms ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param mm_particles ...
!> \param para_env ...
!> \param coarser_grid_level ...
!> \param Forces ...
!> \param per_potentials ...
!> \param aug_pools ...
!> \param mm_cell ...
!> \param dOmmOqm ...
!> \param iw ...
!> \param par_scheme ...
!> \param qmmm_spherical_cutoff ...
!> \param shells ...
!> \par History
!>      08.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE qmmm_forces_with_gaussian_LG  (pgfs,cgrid,num_mm_atoms,mm_charges,mm_atom_index,&
       mm_particles,para_env, coarser_grid_level,Forces, per_potentials,&
       aug_pools, mm_cell, dOmmOqm, iw, par_scheme, qmmm_spherical_cutoff, shells)
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(pw_type), POINTER                             :: cgrid
      INTEGER, INTENT(IN)                                :: num_mm_atoms
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: coarser_grid_level
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Forces
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER   :: per_potentials
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: aug_pools
      TYPE(cell_type), POINTER                           :: mm_cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dOmmOqm
      INTEGER, INTENT(IN)                                :: iw, par_scheme
      REAL(KIND=dp), DIMENSION(2), INTENT(IN)            :: qmmm_spherical_cutoff
      LOGICAL                                            :: shells

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_forces_with_gaussian_LG', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, ii1, ii2, ii3, ii4, ij1, ij2, ij3, ij4, ik1, ik2, ik3, ik4, Imm, &
         IndMM, IRadTyp, ivec(3), j, k, LIndMM, my_i, my_j, my_k, myind, npts(3)
      INTEGER, DIMENSION(2, 3)                           :: bo, gbo
      REAL(KIND=dp) :: a1, a2, a3, abc_X(4,4), abc_X_Y(4), b1, b2, b3, c1, c2, c3, d1, d2, d3, &
         dr1, dr1c, dr1i, dr2, dr2c, dr2i, dr3, dr3c, dr3i, dvol, e1, e2, e3, f1, f2, f3, fac, &
         ft1, ft2, ft3, g1, g2, g3, h1, h2, h3, p1, p2, p3, q1, q2, q3, qt, r1, r2, r3, rt1, rt2, &
         rt3, rv1, rv2, rv3, s1, s1d, s1o, s2, s2d, s2o, s3, s3d, s3o, s4, s4d, s4o, &
         sph_chrg_factor, t1, t1d, t1o, t2, t2d, t2o, t3, t3d, t3o, t4, t4d, t4o, u1, u2, u3, v1, &
         v1d, v1o, v2, v2d, v2o, v3, v3d, v3o, v4, v4d, v4o, xd1, xd2, xd3, xs1, xs2, xs3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: LForces
      REAL(KIND=dp), DIMENSION(3)                        :: ra, val, vec
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid, grid2
      TYPE(pw_type), POINTER                             :: pw
      TYPE(qmmm_per_pot_type), POINTER                   :: per_pot

    CALL timeset(routineN,handle)
    NULLIFY(grid)
    ALLOCATE(LForces(3,num_mm_atoms))
    LForces = 0.0_dp
    dr1c  = cgrid%pw_grid%dr(1)
    dr2c  = cgrid%pw_grid%dr(2)
    dr3c  = cgrid%pw_grid%dr(3)
    dvol  = cgrid%pw_grid%dvol
    gbo   = cgrid%pw_grid%bounds
    bo    =  cgrid%pw_grid%bounds_local
    grid  => cgrid%cr3d
    IF (par_scheme==do_par_atom) myind = 0
    Radius: DO IRadTyp = 1, SIZE(pgfs)
       per_pot => per_potentials(IRadTyp)%pot
       pw    => per_pot%TabLR
       grid2 => pw%cr3d(:,:,:)
       npts = pw%pw_grid%npts
       dr1  = pw%pw_grid%dr(1)
       dr2  = pw%pw_grid%dr(2)
       dr3  = pw%pw_grid%dr(3)
       dr1i = 1.0_dp / dr1
       dr2i = 1.0_dp / dr2
       dr3i = 1.0_dp / dr3
       Atoms: DO Imm = 1, SIZE(per_pot%mm_atom_index)
          IF (par_scheme==do_par_atom) THEN
             myind = myind + 1
             IF (MOD(myind,para_env%num_pe)/=para_env%mepos) CYCLE
          END IF
          LIndMM    =   per_pot%mm_atom_index(Imm)
          IndMM     =   mm_atom_index(LIndMM)
          IF (shells) THEN
             ra(:)     =   pbc(mm_particles(LIndMM)%r-dOmmOqm, mm_cell)+dOmmOqm
          ELSE
             ra(:)     =   pbc(mm_particles(IndMM)%r-dOmmOqm,mm_cell)+dOmmOqm
          END IF
          qt        =   mm_charges(LIndMM)
          ! Possible Spherical Cutoff
          IF (qmmm_spherical_cutoff(1)>0.0_dp) THEN
             CALL spherical_cutoff_factor(qmmm_spherical_cutoff, ra, sph_chrg_factor)
             qt = qt * sph_chrg_factor
          END IF
          IF (ABS(qt)<= EPSILON(0.0_dp)) CYCLE Atoms
          rt1 = ra(1)
          rt2 = ra(2)
          rt3 = ra(3)
          ft1 = 0.0_dp
          ft2 = 0.0_dp
          ft3 = 0.0_dp
          LoopOnGrid: DO k = bo(1,3), bo(2,3)
             my_k = k-gbo(1,3)
             xs3  = REAL(my_k,dp)*dr3c
             my_j = bo(1,2)-gbo(1,2)
             xs2 = REAL(my_j,dp)*dr2c
             rv3 = rt3 - xs3
             vec(3) = rv3
             ivec(3) = FLOOR(vec(3)/pw%pw_grid%dr(3))
                   ik1 = MODULO(ivec(3)-1,npts(3))+1
                   ik2 = MODULO(ivec(3)  ,npts(3))+1
                   ik3 = MODULO(ivec(3)+1,npts(3))+1
                   ik4 = MODULO(ivec(3)+2,npts(3))+1
             xd3  = (vec(3)/dr3)-REAL(ivec(3),kind=dp)
             p1  = 3.0_dp + xd3
             p2  = p1*p1
             p3  = p2*p1
             q1  = 2.0_dp + xd3
             q2  = q1*q1
             q3  = q2*q1
             r1  = 1.0_dp + xd3
             r2  = r1*r1
             r3  = r2*r1
             u1  = xd3
             u2  = u1*u1
             u3  = u2*u1
             v1o =   1.0_dp/6.0_dp * (64.0_dp - 48.0_dp*p1 + 12.0_dp*p2 - p3)
             v2o = -22.0_dp/3.0_dp + 10.0_dp*q1 - 4.0_dp*q2 + 0.5_dp*q3
             v3o =   2.0_dp/3.0_dp -  2.0_dp*r1 + 2.0_dp*r2 - 0.5_dp*r3
             v4o =   1.0_dp/6.0_dp*u3
             v1d =  -8.0_dp + 4.0_dp*p1 - 0.5_dp*p2
             v2d =  10.0_dp - 8.0_dp*q1 + 1.5_dp*q2
             v3d =  -2.0_dp + 4.0_dp*r1 - 1.5_dp*r2
             v4d =   0.5_dp*u2
             DO j =  bo(1,2), bo(2,2)
                my_i= bo(1,1)-gbo(1,1)
                xs1 = REAL(my_i,dp)*dr1c
                rv2 = rt2 - xs2
                vec(2) = rv2
                ivec(2) = FLOOR(vec(2)/pw%pw_grid%dr(2))
                   ij1 = MODULO(ivec(2)-1,npts(2))+1
                   ij2 = MODULO(ivec(2)  ,npts(2))+1
                   ij3 = MODULO(ivec(2)+1,npts(2))+1
                   ij4 = MODULO(ivec(2)+2,npts(2))+1
                xd2  = (vec(2)/dr2)-REAL(ivec(2),kind=dp)
                e1  = 3.0_dp + xd2
                e2  = e1*e1
                e3  = e2*e1
                f1  = 2.0_dp + xd2
                f2  = f1*f1
                f3  = f2*f1
                g1  = 1.0_dp + xd2
                g2  = g1*g1
                g3  = g2*g1
                h1  = xd2
                h2  = h1*h1
                h3  = h2*h1
                s1o =   1.0_dp/6.0_dp * (64.0_dp - 48.0_dp*e1 + 12.0_dp*e2 - e3)
                s2o = -22.0_dp/3.0_dp + 10.0_dp*f1 - 4.0_dp*f2 + 0.5_dp*f3
                s3o =   2.0_dp/3.0_dp -  2.0_dp*g1 + 2.0_dp*g2 - 0.5_dp*g3
                s4o =   1.0_dp/6.0_dp*h3
                s1d =  -8.0_dp + 4.0_dp*e1 - 0.5_dp*e2
                s2d =  10.0_dp - 8.0_dp*f1 + 1.5_dp*f2
                s3d =  -2.0_dp + 4.0_dp*g1 - 1.5_dp*g2
                s4d =   0.5_dp*h2
                DO i =  bo(1,1), bo(2,1)
                   rv1 = rt1 - xs1
                   vec(1) = rv1
                   ivec(1) = FLOOR(vec(1)/pw%pw_grid%dr(1))
                   ii1 = MODULO(ivec(1)-1,npts(1))+1
                   ii2 = MODULO(ivec(1)  ,npts(1))+1
                   ii3 = MODULO(ivec(1)+1,npts(1))+1
                   ii4 = MODULO(ivec(1)+2,npts(1))+1
                   xd1  = (vec(1)/dr1)-REAL(ivec(1),kind=dp)
                   a1  = 3.0_dp + xd1
                   a2  = a1*a1
                   a3  = a2*a1
                   b1  = 2.0_dp + xd1
                   b2  = b1*b1
                   b3  = b2*b1
                   c1  = 1.0_dp + xd1
                   c2  = c1*c1
                   c3  = c2*c1
                   d1  = xd1
                   d2  = d1*d1
                   d3  = d2*d1
                   t1o =   1.0_dp/6.0_dp * (64.0_dp - 48.0_dp*a1 + 12.0_dp*a2 - a3)
                   t2o = -22.0_dp/3.0_dp + 10.0_dp*b1 - 4.0_dp*b2 + 0.5_dp*b3
                   t3o =   2.0_dp/3.0_dp -  2.0_dp*c1 + 2.0_dp*c2 - 0.5_dp*c3
                   t4o =   1.0_dp/6.0_dp*d3
                   t1d =  -8.0_dp + 4.0_dp*a1 - 0.5_dp*a2
                   t2d =  10.0_dp - 8.0_dp*b1 + 1.5_dp*b2
                   t3d =  -2.0_dp + 4.0_dp*c1 - 1.5_dp*c2
                   t4d =   0.5_dp*d2


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! # v then t then s
!
! for ii in 1 2 3; do
! if [[ $ii == 1 ]]; then ld=t; fi
! if [[ $ii == 2 ]]; then ld=s; fi
! if [[ $ii == 3 ]]; then ld=v; fi
! #
! for l in t s v; do
! for i in 1 2 3 4; do
! if [[ $ld == $l ]]; then
!  echo "$l$i = $l${i}d*dr${ii}i"
! else
!  echo "$l$i = $l${i}o"
! fi
! done
! done
! #
! for i in 1 2 3 4; do
! for j in 1 2 3 4; do
! echo -n "abc_X($i,$j) = "
! for k in 1 2 3 4; do
! if [ $k == 4 ]; then
! echo "grid2(ii$i,ij$j,ik$k)*v$k"
! else
! echo -n "grid2(ii$i,ij$j,ik$k)*v$k + "
! fi
! done
! done
! done
! echo ""
! for j in 1 2 3 4; do
! echo -n "abc_X_Y($j) = "
! for k in 1 2 3 4; do
! if [ $k == 4 ]; then
! echo "abc_X($k,$j)*t$k"
! else
! echo -n "abc_X($k,$j)*t$k + "
! fi
! done
! done
! echo ""
! echo -n "val($ii) = "
! for k in 1 2 3 4; do
! if [ $k == 4 ]; then
! echo "abc_X_Y($k)*s$k"
! else
! echo -n "abc_X_Y($k)*s$k + "
! fi
! done
! echo ""
! done

                   t1     = t1d*dr1i
                   t2     = t2d*dr1i
                   t3     = t3d*dr1i
                   t4     = t4d*dr1i
                   s1     = s1o
                   s2     = s2o
                   s3     = s3o
                   s4     = s4o
                   v1     = v1o
                   v2     = v2o
                   v3     = v3o
                   v4     = v4o
                   abc_X(1,1) = grid2(ii1,ij1,ik1)*v1 + grid2(ii1,ij1,ik2)*v2 + grid2(ii1,ij1,ik3)*v3 + grid2(ii1,ij1,ik4)*v4
                   abc_X(1,2) = grid2(ii1,ij2,ik1)*v1 + grid2(ii1,ij2,ik2)*v2 + grid2(ii1,ij2,ik3)*v3 + grid2(ii1,ij2,ik4)*v4
                   abc_X(1,3) = grid2(ii1,ij3,ik1)*v1 + grid2(ii1,ij3,ik2)*v2 + grid2(ii1,ij3,ik3)*v3 + grid2(ii1,ij3,ik4)*v4
                   abc_X(1,4) = grid2(ii1,ij4,ik1)*v1 + grid2(ii1,ij4,ik2)*v2 + grid2(ii1,ij4,ik3)*v3 + grid2(ii1,ij4,ik4)*v4
                   abc_X(2,1) = grid2(ii2,ij1,ik1)*v1 + grid2(ii2,ij1,ik2)*v2 + grid2(ii2,ij1,ik3)*v3 + grid2(ii2,ij1,ik4)*v4
                   abc_X(2,2) = grid2(ii2,ij2,ik1)*v1 + grid2(ii2,ij2,ik2)*v2 + grid2(ii2,ij2,ik3)*v3 + grid2(ii2,ij2,ik4)*v4
                   abc_X(2,3) = grid2(ii2,ij3,ik1)*v1 + grid2(ii2,ij3,ik2)*v2 + grid2(ii2,ij3,ik3)*v3 + grid2(ii2,ij3,ik4)*v4
                   abc_X(2,4) = grid2(ii2,ij4,ik1)*v1 + grid2(ii2,ij4,ik2)*v2 + grid2(ii2,ij4,ik3)*v3 + grid2(ii2,ij4,ik4)*v4
                   abc_X(3,1) = grid2(ii3,ij1,ik1)*v1 + grid2(ii3,ij1,ik2)*v2 + grid2(ii3,ij1,ik3)*v3 + grid2(ii3,ij1,ik4)*v4
                   abc_X(3,2) = grid2(ii3,ij2,ik1)*v1 + grid2(ii3,ij2,ik2)*v2 + grid2(ii3,ij2,ik3)*v3 + grid2(ii3,ij2,ik4)*v4
                   abc_X(3,3) = grid2(ii3,ij3,ik1)*v1 + grid2(ii3,ij3,ik2)*v2 + grid2(ii3,ij3,ik3)*v3 + grid2(ii3,ij3,ik4)*v4
                   abc_X(3,4) = grid2(ii3,ij4,ik1)*v1 + grid2(ii3,ij4,ik2)*v2 + grid2(ii3,ij4,ik3)*v3 + grid2(ii3,ij4,ik4)*v4
                   abc_X(4,1) = grid2(ii4,ij1,ik1)*v1 + grid2(ii4,ij1,ik2)*v2 + grid2(ii4,ij1,ik3)*v3 + grid2(ii4,ij1,ik4)*v4
                   abc_X(4,2) = grid2(ii4,ij2,ik1)*v1 + grid2(ii4,ij2,ik2)*v2 + grid2(ii4,ij2,ik3)*v3 + grid2(ii4,ij2,ik4)*v4
                   abc_X(4,3) = grid2(ii4,ij3,ik1)*v1 + grid2(ii4,ij3,ik2)*v2 + grid2(ii4,ij3,ik3)*v3 + grid2(ii4,ij3,ik4)*v4
                   abc_X(4,4) = grid2(ii4,ij4,ik1)*v1 + grid2(ii4,ij4,ik2)*v2 + grid2(ii4,ij4,ik3)*v3 + grid2(ii4,ij4,ik4)*v4

                   abc_X_Y(1) = abc_X(1,1)*t1 + abc_X(2,1)*t2 + abc_X(3,1)*t3 + abc_X(4,1)*t4
                   abc_X_Y(2) = abc_X(1,2)*t1 + abc_X(2,2)*t2 + abc_X(3,2)*t3 + abc_X(4,2)*t4
                   abc_X_Y(3) = abc_X(1,3)*t1 + abc_X(2,3)*t2 + abc_X(3,3)*t3 + abc_X(4,3)*t4
                   abc_X_Y(4) = abc_X(1,4)*t1 + abc_X(2,4)*t2 + abc_X(3,4)*t3 + abc_X(4,4)*t4

                   val(1) = abc_X_Y(1)*s1 + abc_X_Y(2)*s2 + abc_X_Y(3)*s3 + abc_X_Y(4)*s4

                   t1     = t1o
                   t2     = t2o
                   t3     = t3o
                   t4     = t4o
                   s1     = s1d*dr2i
                   s2     = s2d*dr2i
                   s3     = s3d*dr2i
                   s4     = s4d*dr2i
                   !! v1 = v1o
                   !! v2 = v2o
                   !! v3 = v3o
                   !! v4 = v4o
                   !! abc_X(1,1) = grid2(ii1,ij1,ik1)*v1 + grid2(ii1,ij1,ik2)*v2 + grid2(ii1,ij1,ik3)*v3 + grid2(ii1,ij1,ik4)*v4
                   !! abc_X(1,2) = grid2(ii1,ij2,ik1)*v1 + grid2(ii1,ij2,ik2)*v2 + grid2(ii1,ij2,ik3)*v3 + grid2(ii1,ij2,ik4)*v4
                   !! abc_X(1,3) = grid2(ii1,ij3,ik1)*v1 + grid2(ii1,ij3,ik2)*v2 + grid2(ii1,ij3,ik3)*v3 + grid2(ii1,ij3,ik4)*v4
                   !! abc_X(1,4) = grid2(ii1,ij4,ik1)*v1 + grid2(ii1,ij4,ik2)*v2 + grid2(ii1,ij4,ik3)*v3 + grid2(ii1,ij4,ik4)*v4
                   !! abc_X(2,1) = grid2(ii2,ij1,ik1)*v1 + grid2(ii2,ij1,ik2)*v2 + grid2(ii2,ij1,ik3)*v3 + grid2(ii2,ij1,ik4)*v4
                   !! abc_X(2,2) = grid2(ii2,ij2,ik1)*v1 + grid2(ii2,ij2,ik2)*v2 + grid2(ii2,ij2,ik3)*v3 + grid2(ii2,ij2,ik4)*v4
                   !! abc_X(2,3) = grid2(ii2,ij3,ik1)*v1 + grid2(ii2,ij3,ik2)*v2 + grid2(ii2,ij3,ik3)*v3 + grid2(ii2,ij3,ik4)*v4
                   !! abc_X(2,4) = grid2(ii2,ij4,ik1)*v1 + grid2(ii2,ij4,ik2)*v2 + grid2(ii2,ij4,ik3)*v3 + grid2(ii2,ij4,ik4)*v4
                   !! abc_X(3,1) = grid2(ii3,ij1,ik1)*v1 + grid2(ii3,ij1,ik2)*v2 + grid2(ii3,ij1,ik3)*v3 + grid2(ii3,ij1,ik4)*v4
                   !! abc_X(3,2) = grid2(ii3,ij2,ik1)*v1 + grid2(ii3,ij2,ik2)*v2 + grid2(ii3,ij2,ik3)*v3 + grid2(ii3,ij2,ik4)*v4
                   !! abc_X(3,3) = grid2(ii3,ij3,ik1)*v1 + grid2(ii3,ij3,ik2)*v2 + grid2(ii3,ij3,ik3)*v3 + grid2(ii3,ij3,ik4)*v4
                   !! abc_X(3,4) = grid2(ii3,ij4,ik1)*v1 + grid2(ii3,ij4,ik2)*v2 + grid2(ii3,ij4,ik3)*v3 + grid2(ii3,ij4,ik4)*v4
                   !! abc_X(4,1) = grid2(ii4,ij1,ik1)*v1 + grid2(ii4,ij1,ik2)*v2 + grid2(ii4,ij1,ik3)*v3 + grid2(ii4,ij1,ik4)*v4
                   !! abc_X(4,2) = grid2(ii4,ij2,ik1)*v1 + grid2(ii4,ij2,ik2)*v2 + grid2(ii4,ij2,ik3)*v3 + grid2(ii4,ij2,ik4)*v4
                   !! abc_X(4,3) = grid2(ii4,ij3,ik1)*v1 + grid2(ii4,ij3,ik2)*v2 + grid2(ii4,ij3,ik3)*v3 + grid2(ii4,ij3,ik4)*v4
                   !! abc_X(4,4) = grid2(ii4,ij4,ik1)*v1 + grid2(ii4,ij4,ik2)*v2 + grid2(ii4,ij4,ik3)*v3 + grid2(ii4,ij4,ik4)*v4

                   abc_X_Y(1) = abc_X(1,1)*t1 + abc_X(2,1)*t2 + abc_X(3,1)*t3 + abc_X(4,1)*t4
                   abc_X_Y(2) = abc_X(1,2)*t1 + abc_X(2,2)*t2 + abc_X(3,2)*t3 + abc_X(4,2)*t4
                   abc_X_Y(3) = abc_X(1,3)*t1 + abc_X(2,3)*t2 + abc_X(3,3)*t3 + abc_X(4,3)*t4
                   abc_X_Y(4) = abc_X(1,4)*t1 + abc_X(2,4)*t2 + abc_X(3,4)*t3 + abc_X(4,4)*t4

                   val(2) = abc_X_Y(1)*s1 + abc_X_Y(2)*s2 + abc_X_Y(3)*s3 + abc_X_Y(4)*s4

                   t1     = t1o
                   t2     = t2o
                   t3     = t3o
                   t4     = t4o
                   s1     = s1o
                   s2     = s2o
                   s3     = s3o
                   s4     = s4o
                   v1     = v1d*dr3i
                   v2     = v2d*dr3i
                   v3     = v3d*dr3i
                   v4     = v4d*dr3i
                   abc_X(1,1) = grid2(ii1,ij1,ik1)*v1 + grid2(ii1,ij1,ik2)*v2 + grid2(ii1,ij1,ik3)*v3 + grid2(ii1,ij1,ik4)*v4
                   abc_X(1,2) = grid2(ii1,ij2,ik1)*v1 + grid2(ii1,ij2,ik2)*v2 + grid2(ii1,ij2,ik3)*v3 + grid2(ii1,ij2,ik4)*v4
                   abc_X(1,3) = grid2(ii1,ij3,ik1)*v1 + grid2(ii1,ij3,ik2)*v2 + grid2(ii1,ij3,ik3)*v3 + grid2(ii1,ij3,ik4)*v4
                   abc_X(1,4) = grid2(ii1,ij4,ik1)*v1 + grid2(ii1,ij4,ik2)*v2 + grid2(ii1,ij4,ik3)*v3 + grid2(ii1,ij4,ik4)*v4
                   abc_X(2,1) = grid2(ii2,ij1,ik1)*v1 + grid2(ii2,ij1,ik2)*v2 + grid2(ii2,ij1,ik3)*v3 + grid2(ii2,ij1,ik4)*v4
                   abc_X(2,2) = grid2(ii2,ij2,ik1)*v1 + grid2(ii2,ij2,ik2)*v2 + grid2(ii2,ij2,ik3)*v3 + grid2(ii2,ij2,ik4)*v4
                   abc_X(2,3) = grid2(ii2,ij3,ik1)*v1 + grid2(ii2,ij3,ik2)*v2 + grid2(ii2,ij3,ik3)*v3 + grid2(ii2,ij3,ik4)*v4
                   abc_X(2,4) = grid2(ii2,ij4,ik1)*v1 + grid2(ii2,ij4,ik2)*v2 + grid2(ii2,ij4,ik3)*v3 + grid2(ii2,ij4,ik4)*v4
                   abc_X(3,1) = grid2(ii3,ij1,ik1)*v1 + grid2(ii3,ij1,ik2)*v2 + grid2(ii3,ij1,ik3)*v3 + grid2(ii3,ij1,ik4)*v4
                   abc_X(3,2) = grid2(ii3,ij2,ik1)*v1 + grid2(ii3,ij2,ik2)*v2 + grid2(ii3,ij2,ik3)*v3 + grid2(ii3,ij2,ik4)*v4
                   abc_X(3,3) = grid2(ii3,ij3,ik1)*v1 + grid2(ii3,ij3,ik2)*v2 + grid2(ii3,ij3,ik3)*v3 + grid2(ii3,ij3,ik4)*v4
                   abc_X(3,4) = grid2(ii3,ij4,ik1)*v1 + grid2(ii3,ij4,ik2)*v2 + grid2(ii3,ij4,ik3)*v3 + grid2(ii3,ij4,ik4)*v4
                   abc_X(4,1) = grid2(ii4,ij1,ik1)*v1 + grid2(ii4,ij1,ik2)*v2 + grid2(ii4,ij1,ik3)*v3 + grid2(ii4,ij1,ik4)*v4
                   abc_X(4,2) = grid2(ii4,ij2,ik1)*v1 + grid2(ii4,ij2,ik2)*v2 + grid2(ii4,ij2,ik3)*v3 + grid2(ii4,ij2,ik4)*v4
                   abc_X(4,3) = grid2(ii4,ij3,ik1)*v1 + grid2(ii4,ij3,ik2)*v2 + grid2(ii4,ij3,ik3)*v3 + grid2(ii4,ij3,ik4)*v4
                   abc_X(4,4) = grid2(ii4,ij4,ik1)*v1 + grid2(ii4,ij4,ik2)*v2 + grid2(ii4,ij4,ik3)*v3 + grid2(ii4,ij4,ik4)*v4

                   abc_X_Y(1) = abc_X(1,1)*t1 + abc_X(2,1)*t2 + abc_X(3,1)*t3 + abc_X(4,1)*t4
                   abc_X_Y(2) = abc_X(1,2)*t1 + abc_X(2,2)*t2 + abc_X(3,2)*t3 + abc_X(4,2)*t4
                   abc_X_Y(3) = abc_X(1,3)*t1 + abc_X(2,3)*t2 + abc_X(3,3)*t3 + abc_X(4,3)*t4
                   abc_X_Y(4) = abc_X(1,4)*t1 + abc_X(2,4)*t2 + abc_X(3,4)*t3 + abc_X(4,4)*t4

                   val(3) = abc_X_Y(1)*s1 + abc_X_Y(2)*s2 + abc_X_Y(3)*s3 + abc_X_Y(4)*s4
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                   fac = grid(i,j,k)
                   ft1 = ft1 + val(1) * fac
                   ft2 = ft2 + val(2) * fac
                   ft3 = ft3 + val(3) * fac
                   xs1 = xs1 + dr1c
                END DO
                xs2 = xs2 + dr2c
             END DO
          END DO LoopOnGrid
          qt = - qt * dvol
          LForces(1,LindMM) = ft1 * qt
          LForces(2,LindMM) = ft2 * qt
          LForces(3,LindMM) = ft3 * qt

          Forces(1,LIndMM) = Forces(1,LIndMM) + LForces(1,LindMM)
          Forces(2,LIndMM) = Forces(2,LIndMM) + LForces(2,LindMM)
          Forces(3,LIndMM) = Forces(3,LIndMM) + LForces(3,LindMM)
       END DO Atoms
    END DO Radius
    !
    ! Debug Statement
    !
    IF (debug_this_module) THEN
       CALL debug_qmmm_forces_with_gauss_LG(pgfs=pgfs,&
                                            aug_pools=aug_pools,&
                                            rho=cgrid,&
                                            num_mm_atoms=num_mm_atoms,&
                                            mm_charges=mm_charges,&
                                            mm_atom_index=mm_atom_index,&
                                            mm_particles=mm_particles,&
                                            coarser_grid_level=coarser_grid_level,&
                                            debug_force=LForces,&
                                            per_potentials=per_potentials,&
                                            para_env=para_env,&
                                            mm_cell=mm_cell,&
                                            dOmmOqm=dOmmOqm,&
                                            iw=iw,&
                                            par_scheme=par_scheme,&
                                            qmmm_spherical_cutoff=qmmm_spherical_cutoff,&
                                            shells=shells)
    END IF
    DEALLOCATE(LForces)
    CALL timestop(handle)
  END SUBROUTINE qmmm_forces_with_gaussian_LG

! **************************************************************************************************
!> \brief Evaluates the contribution to the forces due to the Long Range
!>      part of the QM/MM potential computed collocating the Electrostatic
!>      Gaussian Potential.
!> \param pgfs ...
!> \param cgrid ...
!> \param num_mm_atoms ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param mm_particles ...
!> \param para_env ...
!> \param coarser_grid_level ...
!> \param Forces ...
!> \param potentials ...
!> \param aug_pools ...
!> \param mm_cell ...
!> \param dOmmOqm ...
!> \param iw ...
!> \param par_scheme ...
!> \param qmmm_spherical_cutoff ...
!> \param shells ...
!> \par History
!>      08.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE qmmm_forces_with_gaussian_LR  (pgfs,cgrid,num_mm_atoms,mm_charges,mm_atom_index,&
       mm_particles,para_env, coarser_grid_level,Forces, potentials,&
       aug_pools, mm_cell, dOmmOqm, iw, par_scheme, qmmm_spherical_cutoff, shells)
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(pw_type), POINTER                             :: cgrid
      INTEGER, INTENT(IN)                                :: num_mm_atoms
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: coarser_grid_level
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Forces
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: Potentials
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: aug_pools
      TYPE(cell_type), POINTER                           :: mm_cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dOmmOqm
      INTEGER, INTENT(IN)                                :: iw, par_scheme
      REAL(KIND=dp), DIMENSION(2), INTENT(IN)            :: qmmm_spherical_cutoff
      LOGICAL                                            :: shells

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_forces_with_gaussian_LR', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, Imm, IndMM, IRadTyp, ix, j, &
                                                            k, LIndMM, my_i, my_j, my_k, myind, &
                                                            n1, n2, n3
      INTEGER, DIMENSION(2, 3)                           :: bo, gbo
      REAL(KIND=dp)                                      :: dr1, dr2, dr3, dvol, dx, fac, ft1, ft2, &
                                                            ft3, qt, r, r2, rd1, rd2, rd3, rt1, &
                                                            rt2, rt3, rv1, rv2, rv3, rx, rx2, &
                                                            sph_chrg_factor, Term, xs1, xs2, xs3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: LForces
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pot0_2
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid
      TYPE(qmmm_pot_type), POINTER                       :: pot

    CALL timeset(routineN,handle)
    ALLOCATE(LForces(3,num_mm_atoms))
    LForces = 0.0_dp
    n1   = cgrid%pw_grid%npts(1)
    n2   = cgrid%pw_grid%npts(2)
    n3   = cgrid%pw_grid%npts(3)
    dr1  = cgrid%pw_grid%dr(1)
    dr2  = cgrid%pw_grid%dr(2)
    dr3  = cgrid%pw_grid%dr(3)
    dvol = cgrid%pw_grid%dvol
    gbo  = cgrid%pw_grid%bounds
    bo    =  cgrid%pw_grid%bounds_local
    grid  => cgrid%cr3d
    IF (par_scheme==do_par_atom) myind = 0
    Radius: DO IRadTyp = 1, SIZE(pgfs)
       pot => potentials(IRadTyp)%pot
       dx     =  Pot%dx
       pot0_2 => Pot%pot0_2
       Atoms: DO Imm = 1, SIZE(pot%mm_atom_index)
          IF (par_scheme==do_par_atom) THEN
             myind = myind + 1
             IF (MOD(myind,para_env%num_pe)/=para_env%mepos) CYCLE
          END IF
          LIndMM    =   pot%mm_atom_index(Imm)
          IndMM     =   mm_atom_index(LIndMM)
          ra(:)     =   pbc(mm_particles(IndMM)%r-dOmmOqm,mm_cell)+dOmmOqm
          IF (shells) &
               ra(:)     =   pbc(mm_particles(LIndMM)%r-dOmmOqm, mm_cell)+dOmmOqm
          qt        =   mm_charges(LIndMM)
          ! Possible Spherical Cutoff
          IF (qmmm_spherical_cutoff(1)>0.0_dp) THEN
             CALL spherical_cutoff_factor(qmmm_spherical_cutoff, ra, sph_chrg_factor)
             qt = qt * sph_chrg_factor
          END IF
          IF (ABS(qt)<= EPSILON(0.0_dp)) CYCLE Atoms
          rt1 = ra(1)
          rt2 = ra(2)
          rt3 = ra(3)
          ft1 = 0.0_dp
          ft2 = 0.0_dp
          ft3 = 0.0_dp
          LoopOnGrid: DO k = bo(1,3), bo(2,3)
             my_k = k-gbo(1,3)
             xs3  = REAL(my_k,dp)*dr3
             my_j = bo(1,2)-gbo(1,2)
             xs2 = REAL(my_j,dp)*dr2
             rv3 = rt3 - xs3
             DO j =  bo(1,2), bo(2,2)
                my_i= bo(1,1)-gbo(1,1)
                xs1 = REAL(my_i,dp)*dr1
                rv2 = rt2 - xs2
                DO i =  bo(1,1), bo(2,1)
                   rv1 = rt1 - xs1
                   r2  = rv1*rv1 + rv2*rv2 + rv3*rv3
                   r   = SQRT(r2)
                   ix  = FLOOR(r/dx)+1
                   rx  = (r-REAL(ix-1,dp)*dx)/dx
                   rx2 = rx*rx
                   Term = pot0_2(1,ix  )*(-6.0_dp*(rx-rx2))                &
                         +pot0_2(2,ix  )*(1.0_dp-4.0_dp*rx+3.0_dp*rx2)     &
                         +pot0_2(1,ix+1)*(6.0_dp*(rx-rx2))                 &
                         +pot0_2(2,ix+1)*(-2.0_dp*rx+3.0_dp*rx2)
                   fac = grid(i,j,k) * Term
                   IF ( r == 0.0_dp ) THEN
                      rd1 = 1.0_dp
                      rd2 = 1.0_dp
                      rd3 = 1.0_dp
                   ELSE
                      rd1 = rv1 / r
                      rd2 = rv2 / r
                      rd3 = rv3 / r
                   END IF
                   ft1 = ft1 + fac * rd1
                   ft2 = ft2 + fac * rd2
                   ft3 = ft3 + fac * rd3
                   xs1 = xs1 + dr1
                END DO
                xs2 = xs2 + dr2
             END DO
          END DO LoopOnGrid
          qt = - qt * dvol / dx
          LForces(1,LindMM) = ft1 * qt
          LForces(2,LindMM) = ft2 * qt
          LForces(3,LindMM) = ft3 * qt

          Forces(1,LIndMM) = Forces(1,LIndMM) + LForces(1,LindMM)
          Forces(2,LIndMM) = Forces(2,LIndMM) + LForces(2,LindMM)
          Forces(3,LIndMM) = Forces(3,LIndMM) + LForces(3,LindMM)
       END DO Atoms
    END DO Radius
    !
    ! Debug Statement
    !
    IF (debug_this_module) THEN
       CALL debug_qmmm_forces_with_gauss_LR(pgfs=pgfs,&
                                            aug_pools=aug_pools,&
                                            rho=cgrid,&
                                            num_mm_atoms=num_mm_atoms,&
                                            mm_charges=mm_charges,&
                                            mm_atom_index=mm_atom_index,&
                                            mm_particles=mm_particles,&
                                            coarser_grid_level=coarser_grid_level,&
                                            debug_force=LForces,&
                                            potentials=potentials,&
                                            para_env=para_env,&
                                            mm_cell=mm_cell,&
                                            dOmmOqm=dOmmOqm,&
                                            iw=iw,&
                                            par_scheme=par_scheme,&
                                            qmmm_spherical_cutoff=qmmm_spherical_cutoff,&
                                            shells=shells)
    END IF

    DEALLOCATE(LForces)
    CALL timestop(handle)
  END SUBROUTINE qmmm_forces_with_gaussian_LR

! **************************************************************************************************
!> \brief Evaluates numerically QM/MM forces and compares them with
!>      the analytically computed ones.
!>      It is evaluated only when debug_this_module is set to .TRUE.
!> \param rho ...
!> \param qs_env ...
!> \param qmmm_env ...
!> \param Analytical_Forces ...
!> \param mm_particles ...
!> \param mm_atom_index ...
!> \param num_mm_atoms ...
!> \param interp_section ...
!> \param mm_cell ...
!> \par History
!>      08.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE qmmm_debug_forces(rho,qs_env,qmmm_env,Analytical_Forces,&
                               mm_particles,mm_atom_index,num_mm_atoms,&
                               interp_section,mm_cell)
      TYPE(pw_type), POINTER                             :: rho
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Analytical_Forces
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      INTEGER, INTENT(IN)                                :: num_mm_atoms
      TYPE(section_vals_type), POINTER                   :: interp_section
      TYPE(cell_type), POINTER                           :: mm_cell

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_debug_forces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, I, IndMM, iw, J, K
      REAL(KIND=dp)                                      :: Coord_save
      REAL(KIND=dp), DIMENSION(2)                        :: energy
      REAL(KIND=dp), DIMENSION(3)                        :: Err
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Num_Forces
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: v_qmmm_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(qs_ks_qmmm_env_type), POINTER                 :: ks_qmmm_env_loc
      TYPE(section_vals_type), POINTER                   :: input_section, print_section

    CALL timeset(routineN,handle)
    NULLIFY( Num_Forces )
    CALL get_qs_env(qs_env=qs_env,&
                    pw_env=pw_env,&
                    input=input_section,&
                    para_env=para_env)

    print_section => section_vals_get_subs_vals(input_section,"QMMM%PRINT")
    logger => cp_get_default_logger()
    iw=cp_print_key_unit_nr(logger,print_section,"PROGRAM_RUN_INFO",extension=".qmmmLog")
    CALL pw_env_get(pw_env=pw_env, pw_pools=pw_pools)
    CALL pw_pool_create_pw(pw_pools(1)%pool, v_qmmm_rspace%pw,&
         use_data=REALDATA3D, in_space=REALSPACE)
    ALLOCATE(Num_Forces(3,num_mm_atoms))
    ks_qmmm_env_loc =>  qs_env%ks_qmmm_env
    IF (iw>0) WRITE(iw,'(/A)')"DEBUG SECTION:"
    Atoms: DO I = 1, num_mm_atoms
       IndMM = mm_atom_index(I)
       Coords: DO J = 1, 3
          Coord_save = mm_particles(IndMM)%r(J)
          energy = 0.0_dp
          Diff: DO K = 1, 2
             mm_particles(IndMM)%r(J) = Coord_save + (-1)**K * Dx
             CALL pw_zero(v_qmmm_rspace%pw)
             SELECT CASE(qmmm_env%qmmm_coupl_type)
             CASE(do_qmmm_coulomb)
                CALL cp__b("qmmm_gpw_forces.F",1424,"Coulomb QM/MM electrostatic coupling not implemented for GPW/GAPW.")
             CASE(do_qmmm_pcharge)
                CALL cp__b("qmmm_gpw_forces.F",1426,"Point Charge  QM/MM electrostatic coupling not implemented for GPW/GAPW.")
             CASE(do_qmmm_gauss,do_qmmm_swave)
                CALL    qmmm_elec_with_gaussian(qmmm_env=qmmm_env,&
                                                v_qmmm=v_qmmm_rspace,&
                                                mm_particles=mm_particles,&
                                                aug_pools=qmmm_env%aug_pools,&
                                                para_env=para_env,&
                                                eps_mm_rspace=qmmm_env%eps_mm_rspace,&
                                                cube_info=ks_qmmm_env_loc%cube_info,&
                                                pw_pools=pw_pools,&
                                                auxbas_grid=qmmm_env%gridlevel_info%auxbas_grid,&
                                                coarser_grid=qmmm_env%gridlevel_info%coarser_grid,&
                                                interp_section=interp_section,&
                                                mm_cell=mm_cell)
             CASE(do_qmmm_none)
                CYCLE Diff
             CASE DEFAULT
                IF (iw>0) WRITE(iw,'(T3,A)')"Unknown Coupling..."
                CALL cp__b("qmmm_gpw_forces.F",1444,"")
             END SELECT
             energy(K) =  pw_integral_ab ( rho, v_qmmm_rspace%pw)
          END DO Diff
          IF (iw>0) THEN
             WRITE(iw,'(A,I6,A,I3,A,2F15.9)')&
                  "DEBUG :: MM Atom = ",IndMM," Coord = ",J," Energies (+/-) :: ",energy(2), energy(1)
          END IF
          Num_Forces(J,I)  = ( energy(2) - energy(1) ) / (2.0_dp * Dx)
          mm_particles(IndMM)%r(J) = Coord_save
       END DO Coords
    END DO Atoms

    SELECT CASE(qmmm_env%qmmm_coupl_type)
    CASE(do_qmmm_coulomb)
       CALL cp__b("qmmm_gpw_forces.F",1459,"Coulomb QM/MM electrostatic coupling not implemented for GPW/GAPW.")
    CASE(do_qmmm_pcharge)
       CALL cp__b("qmmm_gpw_forces.F",1461,"Point Charge QM/MM electrostatic coupling not implemented for GPW/GAPW.")
    CASE(do_qmmm_gauss,do_qmmm_swave)
       IF (iw>0) WRITE(iw,'(/A/)')"CHECKING NUMERICAL Vs ANALYTICAL FORCES (Err%):"
       DO I =  1, num_mm_atoms
          IndMM = mm_atom_index(I)
          Err = 0.0_dp
          DO K = 1, 3
             IF (ABS(Num_Forces(K,I)) >= 5.0E-5_dp) THEN
                Err(K) = (Analytical_Forces(K,I)-Num_Forces(K,I))/Num_Forces(K,I)*100.0_dp
             END IF
          END DO
          IF (iw>0)&
               WRITE(iw,100)IndMM,Analytical_Forces(1,I),Num_Forces(1,I),Err(1),&
                                  Analytical_Forces(2,I),Num_Forces(2,I),Err(2),&
                                  Analytical_Forces(3,I),Num_Forces(3,I),Err(3)
          IF(.NOT.(ABS(Err(1))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1476)
          IF(.NOT.(ABS(Err(2))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1477)
          IF(.NOT.(ABS(Err(3))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1478)
       END DO
    CASE(do_qmmm_none)
       IF (iw>0) WRITE(iw,'(T3,A)')"No QM/MM Derivatives to debug. Just Mechanical Coupling!"
    CASE DEFAULT
       IF (iw>0) WRITE(iw,'(T3,A)')"Unknown Coupling..."
       CALL cp__b("qmmm_gpw_forces.F",1484,"")
    END SELECT
    CALL cp_print_key_finished_output(iw,logger,print_section,"PROGRAM_RUN_INFO")

    CALL pw_pool_give_back_pw ( pw_pools(1)%pool, v_qmmm_rspace%pw)
    DEALLOCATE(Num_Forces)
    CALL timestop(handle)
100 FORMAT(I5,2F15.9," ( ",F7.2," ) ",2F15.9," ( ",F7.2," ) ",2F15.9," ( ",F7.2," ) ")
  END SUBROUTINE qmmm_debug_forces

! **************************************************************************************************
!> \brief Debugs the integrate_gf_rspace_NoPBC.. It may helps ;-P
!> \param ilevel ...
!> \param zetp ...
!> \param rp ...
!> \param W ...
!> \param pwgrid ...
!> \param cube_info ...
!> \param eps_mm_rspace ...
!> \param aug_pools ...
!> \param debug_force ...
!> \param mm_cell ...
!> \param auxbas_grid ...
!> \param n_rep_real ...
!> \param iw ...
!> \par History
!>      08.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE debug_integrate_gf_rspace_NoPBC(ilevel, zetp, rp, W, pwgrid, cube_info,&
                                             eps_mm_rspace, aug_pools, debug_force,&
                                             mm_cell,auxbas_grid, n_rep_real, iw)
      INTEGER, INTENT(IN)                                :: ilevel
      REAL(KIND=dp), INTENT(IN)                          :: zetp
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rp
      REAL(KIND=dp), INTENT(IN)                          :: W
      TYPE(pw_type), POINTER                             :: pwgrid
      TYPE(cube_info_type), INTENT(IN)                   :: cube_info
      REAL(KIND=dp), INTENT(IN)                          :: eps_mm_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: aug_pools
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: debug_force
      TYPE(cell_type), POINTER                           :: mm_cell
      INTEGER, INTENT(IN)                                :: auxbas_grid
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n_rep_real
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'debug_integrate_gf_rspace_NoPBC', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, igrid, k, ngrids
      INTEGER, DIMENSION(2, 3)                           :: bo2
      INTEGER, SAVE                                      :: Icount
      REAL(KIND=dp), DIMENSION(2)                        :: energy
      REAL(KIND=dp), DIMENSION(3)                        :: Err, force, myrp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: xdat, ydat, zdat
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: grids

    DATA Icount /0/
    ! Statements
    CALL timeset(routineN,handle)
    NULLIFY(grids)
    !Statements
    ngrids = SIZE(aug_pools)
    CALL pw_pools_create_pws(aug_pools,grids,use_data=REALDATA3D,in_space=REALSPACE)
    DO igrid=1,ngrids
       CALL pw_zero(grids(igrid)%pw)
    END DO
    bo2 = grids(auxbas_grid)%pw%pw_grid%bounds
    ALLOCATE (xdat(bo2(1,1):bo2(2,1)))
    ALLOCATE (ydat(bo2(1,2):bo2(2,2)))
    ALLOCATE (zdat(bo2(1,3):bo2(2,3)))

    Icount = Icount + 1
    DO i = 1, 3
       DO k = 1, 2
          myrp = rp
          myrp(i) = myrp(i) + (-1.0_dp)**k * Dx
          CALL pw_zero(grids(ilevel)%pw)
          CALL collocate_gf_rspace_NoPBC(zetp=zetp,&
                                         rp=myrp,&
                                         scale=-1.0_dp,&
                                         W=W,&
                                         pwgrid=grids(ilevel)%pw,&
                                         cube_info=cube_info,&
                                         eps_mm_rspace=eps_mm_rspace,&
                                         xdat=xdat,&
                                         ydat=ydat,&
                                         zdat=zdat,&
                                         bo2=bo2,&
                                         n_rep_real=n_rep_real,&
                                         mm_cell=mm_cell)

          energy(k) = pw_integral_ab(pwgrid, grids(ilevel)%pw)
       END DO
       force(i) = ( energy(2) - energy(1) ) / (2.0_dp * Dx)
    END DO
    Err = 0.0_dp
    IF (ALL(force /= 0.0_dp)) THEN
       Err(1) = (debug_force(1)-force(1))/force(1)*100.0_dp
       Err(2) = (debug_force(2)-force(2))/force(2)*100.0_dp
       Err(3) = (debug_force(3)-force(3))/force(3)*100.0_dp
    END IF
    IF (iw>0) &
         WRITE(iw,100)Icount, debug_force(1), force(1), Err(1),&
                              debug_force(2), force(2), Err(2),&
                              debug_force(3), force(3), Err(3)
    IF(.NOT.(ABS(Err(1))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1590)
    IF(.NOT.(ABS(Err(2))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1591)
    IF(.NOT.(ABS(Err(3))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1592)

    IF (ASSOCIATED(xdat)) THEN
       DEALLOCATE (xdat)
    ENDIF
    IF (ASSOCIATED(ydat)) THEN
       DEALLOCATE (ydat)
    ENDIF
    IF (ASSOCIATED(zdat)) THEN
       DEALLOCATE (zdat)
    ENDIF

    CALL pw_pools_give_back_pws(aug_pools,grids)
    CALL timestop(handle)
100 FORMAT("Collocation   : ",I5,2F15.9," ( ",F7.2," ) ",2F15.9," ( ",F7.2," ) ",2F15.9," ( ",F7.2," ) ")
  END SUBROUTINE debug_integrate_gf_rspace_NoPBC

! **************************************************************************************************
!> \brief Debugs qmmm_forces_with_gaussian_LG.. It may helps too ... ;-]
!> \param pgfs ...
!> \param aug_pools ...
!> \param rho ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param mm_particles ...
!> \param num_mm_atoms ...
!> \param coarser_grid_level ...
!> \param per_potentials ...
!> \param debug_force ...
!> \param para_env ...
!> \param mm_cell ...
!> \param dOmmOqm ...
!> \param iw ...
!> \param par_scheme ...
!> \param qmmm_spherical_cutoff ...
!> \param shells ...
!> \par History
!>      08.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE debug_qmmm_forces_with_gauss_LG(pgfs,aug_pools, rho, mm_charges, mm_atom_index,&
       mm_particles, num_mm_atoms, coarser_grid_level, per_potentials,&
       debug_force, para_env, mm_cell,dOmmOqm,iw,par_scheme,qmmm_spherical_cutoff, shells)

      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: aug_pools
      TYPE(pw_type), POINTER                             :: rho
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      INTEGER, INTENT(IN)                                :: num_mm_atoms, coarser_grid_level
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER   :: per_potentials
      REAL(KIND=dp), DIMENSION(:, :)                     :: debug_force
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cell_type), POINTER                           :: mm_cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dOmmOqm
      INTEGER, INTENT(IN)                                :: iw, par_scheme
      REAL(KIND=dp), DIMENSION(2), INTENT(IN)            :: qmmm_spherical_cutoff
      LOGICAL                                            :: shells

      CHARACTER(len=*), PARAMETER :: routineN = 'debug_qmmm_forces_with_gauss_LG', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, I, igrid, IndMM, J, K, ngrids
      REAL(KIND=dp)                                      :: Coord_save
      REAL(KIND=dp), DIMENSION(2)                        :: energy
      REAL(KIND=dp), DIMENSION(3)                        :: Err
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Num_Forces
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: grids

    ALLOCATE(Num_Forces(3,num_mm_atoms))
    NULLIFY(grids)
    CALL timeset(routineN,handle)
    ngrids = SIZE(aug_pools)
    CALL pw_pools_create_pws(aug_pools,grids,use_data=REALDATA3D,in_space=REALSPACE)
    DO igrid=1,ngrids
       CALL pw_zero(grids(igrid)%pw)
    END DO
    Atoms: DO I = 1, num_mm_atoms
       IndMM = mm_atom_index(I)
       Coords: DO J = 1, 3
          Coord_save = mm_particles(IndMM)%r(J)
          energy = 0.0_dp
          Diff: DO K = 1, 2
             mm_particles(IndMM)%r(J) = Coord_save + (-1)**K * Dx
             CALL pw_zero(grids(coarser_grid_level)%pw)

             CALL qmmm_elec_with_gaussian_LG    (pgfs=pgfs,&
                                                 cgrid=grids(coarser_grid_level)%pw,&
                                                 mm_charges=mm_charges,&
                                                 mm_atom_index=mm_atom_index,&
                                                 mm_particles=mm_particles,&
                                                 para_env=para_env,&
                                                 per_potentials=per_potentials,&
                                                 mm_cell=mm_cell,&
                                                 dOmmOqm=dOmmOqm,&
                                                 par_scheme=par_scheme,&
                                                 qmmm_spherical_cutoff=qmmm_spherical_cutoff,&
                                                 shells=shells)

             energy(K) =  pw_integral_ab ( rho, grids(coarser_grid_level)%pw)
          END DO Diff
          IF (iw>0)&
               WRITE(iw,'(A,I6,A,I3,A,2F15.9)')&
               "DEBUG LR:: MM Atom = ",IndMM," Coord = ",J," Energies (+/-) :: ",energy(2), energy(1)
          Num_Forces(J,I)  = ( energy(2) - energy(1) ) / (2.0_dp * Dx)
          mm_particles(IndMM)%r(J) = Coord_save
       END DO Coords
    END DO Atoms

    DO I =  1, num_mm_atoms
       IndMM = mm_atom_index(I)
       Err = 0.0_dp
       IF (ALL(Num_Forces /= 0.0_dp)) THEN
          Err(1) = (debug_force(1,I)-Num_Forces(1,I))/Num_Forces(1,I)*100.0_dp
          Err(2) = (debug_force(2,I)-Num_Forces(2,I))/Num_Forces(2,I)*100.0_dp
          Err(3) = (debug_force(3,I)-Num_Forces(3,I))/Num_Forces(3,I)*100.0_dp
       END IF
       IF (iw>0) &
            WRITE(iw,100)IndMM,debug_force(1,I),Num_Forces(1,I),Err(1),&
                               debug_force(2,I),Num_Forces(2,I),Err(2),&
                               debug_force(3,I),Num_Forces(3,I),Err(3)
       IF(.NOT.(ABS(Err(1))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1714)
       IF(.NOT.(ABS(Err(2))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1715)
       IF(.NOT.(ABS(Err(3))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1716)
    END DO

    DEALLOCATE(Num_Forces)
    CALL pw_pools_give_back_pws(aug_pools,grids)
    CALL timestop(handle)
100 FORMAT("MM Atom LR    : ",I5,2F15.9," ( ",F7.2," ) ",2F15.9," ( ",F7.2," ) ",2F15.9," ( ",F7.2," ) ")
  END SUBROUTINE debug_qmmm_forces_with_gauss_LG

! **************************************************************************************************
!> \brief Debugs qmmm_forces_with_gaussian_LR.. It may helps too ... ;-]
!> \param pgfs ...
!> \param aug_pools ...
!> \param rho ...
!> \param mm_charges ...
!> \param mm_atom_index ...
!> \param mm_particles ...
!> \param num_mm_atoms ...
!> \param coarser_grid_level ...
!> \param potentials ...
!> \param debug_force ...
!> \param para_env ...
!> \param mm_cell ...
!> \param dOmmOqm ...
!> \param iw ...
!> \param par_scheme ...
!> \param qmmm_spherical_cutoff ...
!> \param shells ...
!> \par History
!>      08.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
  SUBROUTINE debug_qmmm_forces_with_gauss_LR(pgfs,aug_pools, rho, mm_charges, mm_atom_index,&
       mm_particles, num_mm_atoms, coarser_grid_level, potentials,&
       debug_force, para_env, mm_cell,dOmmOqm,iw, par_scheme, qmmm_spherical_cutoff, shells)

      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: aug_pools
      TYPE(pw_type), POINTER                             :: rho
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mm_charges
      INTEGER, DIMENSION(:), POINTER                     :: mm_atom_index
      TYPE(particle_type), DIMENSION(:), POINTER         :: mm_particles
      INTEGER, INTENT(IN)                                :: num_mm_atoms, coarser_grid_level
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: Potentials
      REAL(KIND=dp), DIMENSION(:, :)                     :: debug_force
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cell_type), POINTER                           :: mm_cell
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: dOmmOqm
      INTEGER, INTENT(IN)                                :: iw, par_scheme
      REAL(KIND=dp), DIMENSION(2), INTENT(IN)            :: qmmm_spherical_cutoff
      LOGICAL                                            :: shells

      CHARACTER(len=*), PARAMETER :: routineN = 'debug_qmmm_forces_with_gauss_LR', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, I, igrid, IndMM, J, K, ngrids
      REAL(KIND=dp)                                      :: Coord_save
      REAL(KIND=dp), DIMENSION(2)                        :: energy
      REAL(KIND=dp), DIMENSION(3)                        :: Err
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Num_Forces
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: grids

    ALLOCATE(Num_Forces(3,num_mm_atoms))
    NULLIFY(grids)
    CALL timeset(routineN,handle)
    ngrids = SIZE(aug_pools)
    CALL pw_pools_create_pws(aug_pools,grids,use_data=REALDATA3D,in_space=REALSPACE)
    DO igrid=1,ngrids
       CALL pw_zero(grids(igrid)%pw)
    END DO
    Atoms: DO I = 1, num_mm_atoms
       IndMM = mm_atom_index(I)
       Coords: DO J = 1, 3
          Coord_save = mm_particles(IndMM)%r(J)
          energy = 0.0_dp
          Diff: DO K = 1, 2
             mm_particles(IndMM)%r(J) = Coord_save + (-1)**K * Dx
             CALL pw_zero(grids(coarser_grid_level)%pw)

             CALL qmmm_elec_with_gaussian_LR    (pgfs=pgfs,&
                                                 grid=grids(coarser_grid_level)%pw,&
                                                 mm_charges=mm_charges,&
                                                 mm_atom_index=mm_atom_index,&
                                                 mm_particles=mm_particles,&
                                                 para_env=para_env,&
                                                 potentials=potentials,&
                                                 mm_cell=mm_cell,&
                                                 dOmmOqm=dOmmOqm,&
                                                 par_scheme=par_scheme,&
                                                 qmmm_spherical_cutoff=qmmm_spherical_cutoff,&
                                                 shells=shells)

             energy(K) =  pw_integral_ab ( rho, grids(coarser_grid_level)%pw)
          END DO Diff
          IF (iw>0)&
               WRITE(iw,'(A,I6,A,I3,A,2F15.9)')&
               "DEBUG LR:: MM Atom = ",IndMM," Coord = ",J," Energies (+/-) :: ",energy(2), energy(1)
          Num_Forces(J,I)  = ( energy(2) - energy(1) ) / (2.0_dp * Dx)
          mm_particles(IndMM)%r(J) = Coord_save
       END DO Coords
    END DO Atoms

    DO I =  1, num_mm_atoms
       IndMM = mm_atom_index(I)
       Err = 0.0_dp
       IF (ALL(Num_Forces(:,I) /= 0.0_dp)) THEN
          Err(1) = (debug_force(1,I)-Num_Forces(1,I))/Num_Forces(1,I)*100.0_dp
          Err(2) = (debug_force(2,I)-Num_Forces(2,I))/Num_Forces(2,I)*100.0_dp
          Err(3) = (debug_force(3,I)-Num_Forces(3,I))/Num_Forces(3,I)*100.0_dp
       END IF
       IF (iw>0)&
            WRITE(iw,100)IndMM,debug_force(1,I),Num_Forces(1,I),Err(1),&
                               debug_force(2,I),Num_Forces(2,I),Err(2),&
                               debug_force(3,I),Num_Forces(3,I),Err(3)
       IF(.NOT.(ABS(Err(1))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1830)
       IF(.NOT.(ABS(Err(2))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1831)
       IF(.NOT.(ABS(Err(3))<=MaxErr))CALL cp__a("qmmm_gpw_forces.F",1832)
    END DO

    DEALLOCATE(Num_Forces)
    CALL pw_pools_give_back_pws(aug_pools,grids)
    CALL timestop(handle)
100 FORMAT("MM Atom LR    : ",I5,2F15.9," ( ",F7.2," ) ",2F15.9," ( ",F7.2," ) ",2F15.9," ( ",F7.2," ) ")
  END SUBROUTINE debug_qmmm_forces_with_gauss_LR

END MODULE qmmm_gpw_forces
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates QM/MM energy and forces
!> \par History
!>      2015 Factored out of force_env_methods.F
!> \author Ole Schuett
! **************************************************************************************************
MODULE qmmm_force
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_iter_string,&
                                              cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_result_methods,               ONLY: cp_results_erase,&
                                              get_results,&
                                              put_results
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE fist_energy_types,               ONLY: fist_energy_type
   USE fist_environment_types,          ONLY: fist_env_get
   USE fist_force,                      ONLY: fist_calc_energy_force
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: debye
   USE qmmm_gpw_energy,                 ONLY: qmmm_el_coupling
   USE qmmm_gpw_forces,                 ONLY: qmmm_forces
   USE qmmm_links_methods,              ONLY: qmmm_added_chrg_coord,&
                                              qmmm_added_chrg_forces,&
                                              qmmm_link_Imomm_coord,&
                                              qmmm_link_Imomm_forces
   USE qmmm_types,                      ONLY: qmmm_env_type
   USE qmmm_types_low,                  ONLY: qmmm_links_type
   USE qmmm_util,                       ONLY: apply_qmmm_translate,&
                                              apply_qmmm_walls
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env
   USE qs_force,                        ONLY: qs_calc_energy_force
   USE qs_ks_qmmm_methods,              ONLY: ks_qmmm_env_rebuild
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_force'

   PUBLIC :: qmmm_calc_energy_force

CONTAINS

! **************************************************************************************************
!> \brief calculates the qm/mm energy and forces
!> \param qmmm_env ...
!> \param calc_force if also the forces should be calculated
!> \param consistent_energies ...
!> \param linres ...
!> \par History
!>      05.2004 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qmmm_calc_energy_force(qmmm_env, calc_force, consistent_energies, linres)
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      LOGICAL, INTENT(IN)                                :: calc_force, consistent_energies, linres

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_calc_energy_force', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description, iter
      INTEGER                                            :: ip, j, nres, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index
      LOGICAL                                            :: check, qmmm_added_chrg, qmmm_link, &
                                                            qmmm_link_imomm
      REAL(KIND=dp)                                      :: energy_mm, energy_qm
      REAL(KIND=dp), DIMENSION(3)                        :: dip_mm, dip_qm, dip_qmmm, max_coord, &
                                                            min_coord
      TYPE(cell_type), POINTER                           :: mm_cell, qm_cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_result_type), POINTER                      :: results_mm, results_qm, results_qmmm
      TYPE(cp_subsys_type), POINTER                      :: subsys, subsys_mm, subsys_qm
      TYPE(fist_energy_type), POINTER                    :: fist_energy
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_mm, particles_qm
      TYPE(qmmm_links_type), POINTER                     :: qmmm_links
      TYPE(qs_energy_type), POINTER                      :: qs_energy
      TYPE(section_vals_type), POINTER                   :: force_env_section, print_key

      min_coord = HUGE(0.0_dp)
      max_coord = -HUGE(0.0_dp)
      qmmm_link = .FALSE.
      qmmm_link_imomm = .FALSE.
      qmmm_added_chrg = .FALSE.
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      NULLIFY (subsys_mm, subsys_qm, subsys, qm_atom_index, particles_mm, particles_qm, qm_cell, mm_cell)
      NULLIFY (force_env_section, print_key, results_qmmm, results_qm, results_mm)

      CALL get_qs_env(qmmm_env%qs_env, input=force_env_section)
      print_key => section_vals_get_subs_vals(force_env_section, "QMMM%PRINT%DIPOLE")

      IF(.NOT.(ASSOCIATED(qmmm_env)))CALL cp__a("qmmm_force.F",111)
      IF(.NOT.(qmmm_env%ref_count > 0))CALL cp__a("qmmm_force.F",112)

      CALL apply_qmmm_translate(qmmm_env)

      CALL fist_env_get(qmmm_env%fist_env, cell=mm_cell, subsys=subsys_mm)
      CALL get_qs_env(qmmm_env%qs_env, cell=qm_cell, cp_subsys=subsys_qm)
      qm_atom_index => qmmm_env%qm%qm_atom_index
      qmmm_link = qmmm_env%qm%qmmm_link
      qmmm_links => qmmm_env%qm%qmmm_links
      qmmm_added_chrg = (qmmm_env%qm%move_mm_charges .OR. qmmm_env%qm%add_mm_charges)
      IF (qmmm_link) THEN
         IF(.NOT.(ASSOCIATED(qmmm_links)))CALL cp__a("qmmm_force.F",123)
         IF (ASSOCIATED(qmmm_links%imomm)) qmmm_link_imomm = (SIZE(qmmm_links%imomm) /= 0)
      END IF
      IF(.NOT.(ASSOCIATED(qm_atom_index)))CALL cp__a("qmmm_force.F",126)

      particles_mm => subsys_mm%particles%els
      particles_qm => subsys_qm%particles%els

      DO j = 1, 3
         IF (qm_cell%perd(j) == 1) CYCLE
         DO ip = 1, SIZE(particles_qm)
            check = (DOT_PRODUCT(qm_cell%h_inv(j, :), particles_qm(ip)%r) >= 0.0) .AND. &
                    (DOT_PRODUCT(qm_cell%h_inv(j, :), particles_qm(ip)%r) <= 1.0)
            IF (output_unit > 0 .AND. .NOT. check) THEN
               WRITE (unit=output_unit, fmt='("ip, j, pos, lat_pos ",2I6,6F12.5)') ip, j, &
                  particles_qm(ip)%r, DOT_PRODUCT(qm_cell%h_inv(j, :), particles_qm(ip)%r)
            ENDIF
            IF (.NOT. check) &
               CALL cp_abort(cp__l("qmmm_force.F",141), &
                             "QM/MM QM atoms must be fully contained in the same image of the QM box "// &
                             "- No wrapping of coordinates is allowed! ")
         END DO
      END DO

      ! If present QM/MM links (just IMOMM) correct the position of the qm-link atom
      IF (qmmm_link_imomm) CALL qmmm_link_Imomm_coord(qmmm_links, particles_qm, qm_atom_index)

      ! If add charges get their position NOW!
      IF (qmmm_added_chrg) CALL qmmm_added_chrg_coord(qmmm_env%qm, particles_mm)

      ! Initialize ks_qmmm_env
      CALL ks_qmmm_env_rebuild(qs_env=qmmm_env%qs_env, qmmm_env=qmmm_env%qm)

      ! Compute the short range QM/MM Electrostatic Potential
      CALL qmmm_el_coupling(qs_env=qmmm_env%qs_env, &
                            qmmm_env=qmmm_env%qm, &
                            mm_particles=particles_mm, &
                            mm_cell=mm_cell)

      ! Fist
      CALL fist_calc_energy_force(qmmm_env%fist_env)

      ! Print Out information on fist energy calculation...
      CALL fist_env_get(qmmm_env%fist_env, thermo=fist_energy)
      energy_mm = fist_energy%pot
      CALL cp_subsys_get(subsys_mm, results=results_mm)

      ! QS
      CALL qs_calc_energy_force(qmmm_env%qs_env, calc_force, consistent_energies, linres)

      ! QM/MM Interaction Potential forces
      CALL qmmm_forces(qmmm_env%qs_env, &
                       qmmm_env%qm, particles_mm, &
                       mm_cell=mm_cell, &
                       calc_force=calc_force)

      ! Forces of quadratic wall on QM atoms
      CALL apply_qmmm_walls(qmmm_env)

      ! Print Out information on QS energy calculation...
      CALL get_qs_env(qmmm_env%qs_env, energy=qs_energy)
      energy_qm = qs_energy%total
      CALL cp_subsys_get(subsys_qm, results=results_qm)

      !TODO: is really results_qm == results_qmmm ???
      CALL cp_subsys_get(subsys_qm, results=results_qmmm)

      IF (calc_force) THEN
         ! If present QM/MM links (just IMOMM) correct the position of the qm-link atom
         IF (qmmm_link_imomm) CALL qmmm_link_Imomm_forces(qmmm_links, particles_qm, qm_atom_index)
         particles_mm => subsys_mm%particles%els
         DO ip = 1, SIZE(qm_atom_index)
            particles_mm(qm_atom_index(ip))%f = particles_mm(qm_atom_index(ip))%f+particles_qm(ip)%f
         END DO
         ! If add charges get rid of their derivatives right NOW!
         IF (qmmm_added_chrg) CALL qmmm_added_chrg_forces(qmmm_env%qm, particles_mm)
      END IF

      ! Handle some output
      output_unit = cp_print_key_unit_nr(logger, force_env_section, "QMMM%PRINT%DERIVATIVES", &
                                         extension=".Log")
      IF (output_unit > 0) THEN
         WRITE (unit=output_unit, fmt='(/1X,A,F15.9)') "Energy after QMMM calculation: ", energy_qm
         IF (calc_force) THEN
            WRITE (unit=output_unit, fmt='(/1X,A)') "Derivatives on all atoms after QMMM calculation: "
            DO ip = 1, SIZE(particles_mm)
               WRITE (unit=output_unit, fmt='(1X,3F15.9)') particles_mm(ip)%f
            END DO
         END IF
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, force_env_section, &
                                        "QMMM%PRINT%DERIVATIVES")

      ! Dipole
      print_key => section_vals_get_subs_vals(force_env_section, "QMMM%PRINT%DIPOLE")
      IF (BTEST(cp_print_key_should_output(logger%iter_info, print_key), &
                cp_p_file)) THEN
         description = '[DIPOLE]'
         CALL get_results(results=results_qm, description=description, n_rep=nres)
         IF(.NOT.(nres <= 1))CALL cp__a("qmmm_force.F",222)
         CALL get_results(results=results_mm, description=description, n_rep=nres)
         IF(.NOT.(nres <= 1))CALL cp__a("qmmm_force.F",224)
         CALL get_results(results=results_qm, description=description, values=dip_qm)
         CALL get_results(results=results_mm, description=description, values=dip_mm)
         dip_qmmm = dip_qm+dip_mm
         CALL cp_results_erase(results=results_qmmm, description=description)
         CALL put_results(results=results_qmmm, description=description, values=dip_qmmm)

         output_unit = cp_print_key_unit_nr(logger, force_env_section, "QMMM%PRINT%DIPOLE", &
                                            extension=".Dipole")
         IF (output_unit > 0) THEN
            WRITE (unit=output_unit, fmt="(A)") "QMMM TOTAL DIPOLE"
            WRITE (unit=output_unit, fmt="(A,T31,A,T88,A)") &
               "# iter_level", "dipole(x,y,z)[atomic units]", &
               "dipole(x,y,z)[debye]"
            iter = cp_iter_string(logger%iter_info)
            WRITE (unit=output_unit, fmt="(a,6(es18.8))") &
               iter(1:15), dip_qmmm, dip_qmmm*debye
         END IF
      END IF

   END SUBROUTINE qmmm_calc_energy_force

END MODULE qmmm_force
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates QM/MM energy and forces with Force-Mixing
!> \par History
!>      2015 Factored out of force_env_methods.F
!> \author Ole Schuett
! **************************************************************************************************
MODULE qmmmx_force
   USE cell_types,                      ONLY: cell_type
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE fist_environment_types,          ONLY: fist_env_get
   USE input_constants,                 ONLY: do_fm_mom_conserv_QM,&
                                              do_fm_mom_conserv_buffer,&
                                              do_fm_mom_conserv_core,&
                                              do_fm_mom_conserv_equal_a,&
                                              do_fm_mom_conserv_equal_f,&
                                              do_fm_mom_conserv_none
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE particle_types,                  ONLY: particle_type
   USE qmmm_force,                      ONLY: qmmm_calc_energy_force
   USE qmmm_types,                      ONLY: qmmm_env_get,&
                                              qmmm_env_type
   USE qmmm_types_low,                  ONLY: force_mixing_label_QM_core,&
                                              force_mixing_label_QM_dynamics,&
                                              force_mixing_label_buffer
   USE qmmm_util,                       ONLY: apply_qmmm_unwrap,&
                                              apply_qmmm_wrap
   USE qmmmx_types,                     ONLY: qmmmx_env_type
   USE qmmmx_util,                      ONLY: apply_qmmmx_translate
   USE qs_environment_types,            ONLY: get_qs_env
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmmx_force'

   PUBLIC :: qmmmx_calc_energy_force

CONTAINS

! **************************************************************************************************
!> \brief calculates the qm/mm energy and forces
!> \param qmmmx_env ...
!> \param calc_force if also the forces should be calculated
!> \param consistent_energies ...
!> \param linres ...
!> \param require_consistent_energy_force ...
!> \par History
!>      05.2004 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qmmmx_calc_energy_force(qmmmx_env, calc_force, consistent_energies, linres, &
                                      require_consistent_energy_force)
      TYPE(qmmmx_env_type), POINTER                      :: qmmmx_env
      LOGICAL, INTENT(IN)                                :: calc_force, consistent_energies, linres
      LOGICAL, INTENT(IN), OPTIONAL :: require_consistent_energy_force

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmmx_calc_energy_force', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, mom_conserv_min_label, &
                                                            mom_conserv_n, mom_conserv_region, &
                                                            mom_conserv_type
      INTEGER, POINTER                                   :: cur_indices(:), cur_labels(:)
      REAL(dp)                                           :: delta_a(3), delta_f(3), &
                                                            mom_conserv_mass, total_f(3)
      TYPE(cp_subsys_type), POINTER                      :: subsys_primary, subsys_qmmm_core, &
                                                            subsys_qmmm_extended
      TYPE(particle_type), DIMENSION(:), POINTER         :: particles_primary, particles_qmmm_core, &
                                                            particles_qmmm_extended
      TYPE(section_vals_type), POINTER                   :: force_env_section

      IF (PRESENT(require_consistent_energy_force)) THEN
         IF (require_consistent_energy_force) &
            CALL cp_abort(cp__l("qmmmx_force.F",85), &
                          "qmmmx_energy_and_forces got require_consistent_energy_force but force mixing is active. ")
      ENDIF

      ! Possibly translate the system
      CALL apply_qmmmx_translate(qmmmx_env)

      ! actual energy force calculation
      CALL qmmmx_calc_energy_force_low(qmmmx_env%ext, calc_force, consistent_energies, linres, "ext")
      CALL qmmmx_calc_energy_force_low(qmmmx_env%core, calc_force, consistent_energies, linres, "core")

      ! get forces from subsys of each sub force env
      CALL qmmm_env_get(qmmmx_env%core, subsys=subsys_qmmm_core)
      CALL qmmm_env_get(qmmmx_env%ext, subsys=subsys_qmmm_extended)

      CALL get_qs_env(qmmmx_env%ext%qs_env, input=force_env_section)
      CALL section_vals_val_get(force_env_section, "QMMM%FORCE_MIXING%RESTART_INFO%INDICES", i_vals=cur_indices)
      CALL section_vals_val_get(force_env_section, "QMMM%FORCE_MIXING%RESTART_INFO%LABELS", i_vals=cur_labels)

      particles_qmmm_extended => subsys_qmmm_extended%particles%els
      particles_qmmm_core => subsys_qmmm_core%particles%els
      DO ip = 1, SIZE(cur_indices)
         IF (cur_labels(ip) >= force_mixing_label_QM_dynamics) THEN ! this is a QM atom
            ! copy (QM) force from extended calculation
            particles_qmmm_core(cur_indices(ip))%f = particles_qmmm_extended(cur_indices(ip))%f
         ENDIF
      END DO

      ! zero momentum
      CALL section_vals_val_get(force_env_section, "QMMM%FORCE_MIXING%MOMENTUM_CONSERVATION_TYPE", &
                                i_val=mom_conserv_type)
      IF (mom_conserv_type /= do_fm_mom_conserv_none) THEN
         CALL section_vals_val_get(force_env_section, "QMMM%FORCE_MIXING%MOMENTUM_CONSERVATION_REGION", &
                                   i_val=mom_conserv_region)

         IF (mom_conserv_region == do_fm_mom_conserv_core) THEN
            mom_conserv_min_label = force_mixing_label_QM_core
         ELSEIF (mom_conserv_region == do_fm_mom_conserv_QM) THEN
            mom_conserv_min_label = force_mixing_label_QM_dynamics
         ELSEIF (mom_conserv_region == do_fm_mom_conserv_buffer) THEN
            mom_conserv_min_label = force_mixing_label_buffer
         ELSE
            CALL cp__b("qmmmx_force.F",127,"Got unknown MOMENTUM_CONSERVATION_REGION (not CORE, QM, or BUFFER) !")
         ENDIF

         total_f = 0.0_dp
         DO ip = 1, SIZE(particles_qmmm_core)
            total_f(1:3) = total_f(1:3)+particles_qmmm_core(ip)%f(1:3)
         END DO
         IF (mom_conserv_type == do_fm_mom_conserv_equal_f) THEN
            mom_conserv_n = COUNT(cur_labels >= mom_conserv_min_label)
            delta_f = total_f/mom_conserv_n
            DO ip = 1, SIZE(cur_indices)
               IF (cur_labels(ip) >= mom_conserv_min_label) THEN
                  particles_qmmm_core(cur_indices(ip))%f = particles_qmmm_core(cur_indices(ip))%f-delta_f
               ENDIF
            END DO
         ELSE IF (mom_conserv_type == do_fm_mom_conserv_equal_a) THEN
            mom_conserv_mass = 0.0_dp
            DO ip = 1, SIZE(cur_indices)
               IF (cur_labels(ip) >= mom_conserv_min_label) &
                  mom_conserv_mass = mom_conserv_mass+particles_qmmm_core(cur_indices(ip))%atomic_kind%mass
            END DO
            delta_a = total_f/mom_conserv_mass
            DO ip = 1, SIZE(cur_indices)
               IF (cur_labels(ip) >= mom_conserv_min_label) THEN
                  particles_qmmm_core(cur_indices(ip))%f = particles_qmmm_core(cur_indices(ip))%f- &
                                                           particles_qmmm_core(cur_indices(ip))%atomic_kind%mass*delta_a
               ENDIF
            END DO
         ENDIF
      ENDIF

      CALL qmmm_env_get(qmmmx_env%ext, subsys=subsys_primary)
      particles_primary => subsys_primary%particles%els
      DO ip = 1, SIZE(particles_qmmm_core)
         particles_primary(ip)%f = particles_qmmm_core(ip)%f
      END DO

   END SUBROUTINE qmmmx_calc_energy_force

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env ...
!> \param calc_force ...
!> \param consistent_energies ...
!> \param linres ...
!> \param label ...
! **************************************************************************************************
   SUBROUTINE qmmmx_calc_energy_force_low(qmmm_env, calc_force, consistent_energies, linres, label)
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      LOGICAL, INTENT(IN)                                :: calc_force, consistent_energies, linres
      CHARACTER(*)                                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmmx_calc_energy_force_low', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_string_length)                   :: new_restart_fn, new_restart_hist_fn, &
                                                            old_restart_fn, old_restart_hist_fn
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index
      LOGICAL                                            :: saved_do_translate
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: saved_pos
      TYPE(cell_type), POINTER                           :: mm_cell
      TYPE(cp_subsys_type), POINTER                      :: subsys_mm, subsys_qm
      TYPE(section_vals_type), POINTER                   :: force_env_section

      NULLIFY (mm_cell, subsys_qm, subsys_mm, qm_atom_index)

      CALL get_qs_env(qmmm_env%qs_env, input=force_env_section)

      ! rewrite RESTART%FILENAME
      CALL section_vals_val_get(force_env_section, "DFT%SCF%PRINT%RESTART%FILENAME", &
                                c_val=old_restart_fn)
      new_restart_fn = TRIM(old_restart_fn)//"-"//TRIM(label)
      CALL section_vals_val_set(force_env_section, "DFT%SCF%PRINT%RESTART%FILENAME", &
                                c_val=new_restart_fn)

      ! rewrite RESTART_HISTORY%FILENAME
      CALL section_vals_val_get(force_env_section, "DFT%SCF%PRINT%RESTART_HISTORY%FILENAME", &
                                c_val=old_restart_hist_fn)
      new_restart_hist_fn = TRIM(old_restart_hist_fn)//"-"//TRIM(label)
      CALL section_vals_val_set(force_env_section, "DFT%SCF%PRINT%RESTART_HISTORY%FILENAME", &
                                c_val=new_restart_hist_fn)

      ! wrap positions before QM/MM calculation.
      ! Required if diffusion causes atoms outside of periodic box get added to QM
      CALL fist_env_get(qmmm_env%fist_env, cell=mm_cell, subsys=subsys_mm)
      CALL get_qs_env(qmmm_env%qs_env, cp_subsys=subsys_qm)
      qm_atom_index => qmmm_env%qm%qm_atom_index
      CALL apply_qmmm_wrap(subsys_mm, mm_cell, subsys_qm, qm_atom_index, saved_pos)

      ! Turn off box translation, it was already performed by apply_qmmmx_translate(),
      ! the particles coordinates will still be copied from MM to QM.
      saved_do_translate = qmmm_env%qm%do_translate
      qmmm_env%qm%do_translate = .FALSE.

      ! actual energy force calculation
      CALL qmmm_calc_energy_force(qmmm_env, calc_force, consistent_energies, linres)

      ! restore do_translate
      qmmm_env%qm%do_translate = saved_do_translate

      ! restore unwrapped positions
      CALL apply_qmmm_unwrap(subsys_mm, subsys_qm, qm_atom_index, saved_pos)

      ! restore RESTART filenames
      CALL section_vals_val_set(force_env_section, "DFT%SCF%PRINT%RESTART%FILENAME", &
                                c_val=old_restart_fn)
      CALL section_vals_val_set(force_env_section, "DFT%SCF%PRINT%RESTART_HISTORY%FILENAME", &
                                c_val=old_restart_hist_fn)

   END SUBROUTINE qmmmx_calc_energy_force_low

END MODULE qmmmx_force
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Handles all possible kinds of restraints in CP2K
!> \author Teodoro Laino 08.2006 [tlaino] - University of Zurich
!> \par    History
!>         Teodoro Laino [tlaino] - 11.2008 : Improved the fixd_list restraints
! **************************************************************************************************
MODULE restraint
   USE cell_types,                      ONLY: cell_type,&
                                              use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE colvar_methods,                  ONLY: colvar_eval_mol_f
   USE colvar_types,                    ONLY: colvar_counters,&
                                              diff_colvar
   USE constraint_fxd,                  ONLY: create_local_fixd_list,&
                                              release_local_fixd_list
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_set,&
                                              force_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: colvar_constraint_type,&
                                              fixd_constraint_type,&
                                              g3x3_constraint_type,&
                                              g4x6_constraint_type,&
                                              get_molecule_kind,&
                                              local_fixd_constraint_type,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              local_colvar_constraint_type,&
                                              molecule_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type,&
                                              update_particle_set
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: restraint_control
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'restraint'

CONTAINS

! **************************************************************************************************
!> \brief Computes restraints
!> \param force_env ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_control(force_env)

      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ifixd, ii, ikind, imol, &
                                                            iparticle, n3x3con_restraint, &
                                                            n4x6con_restraint, n_restraint, nkind, &
                                                            nmol_per_kind
      REAL(KIND=dp)                                      :: energy_3x3, energy_4x6, energy_colv, &
                                                            energy_fixd, extended_energies, k0, &
                                                            rab(3), rab2, targ(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: force
      TYPE(cell_type), POINTER                           :: cell
      TYPE(colvar_counters)                              :: ncolv
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(local_fixd_constraint_type), POINTER          :: lfixd_list(:)
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind, molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule, molecule_set(:)
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), POINTER                       :: particle_set(:)

      NULLIFY (cell, subsys, local_molecules, local_particles, fixd_list, molecule_kinds, &
               molecules, molecule_kind, molecule_kind_set, molecule, molecule_set, particles, &
               particle_set, gci, lfixd_list)
      CALL timeset(routineN, handle)
      CALL force_env_get(force_env=force_env, subsys=subsys, cell=cell)
      energy_4x6 = 0.0_dp
      energy_3x3 = 0.0_dp
      energy_colv = 0.0_dp
      energy_fixd = 0.0_dp
      n_restraint = 0
      CALL cp_subsys_get(subsys=subsys, particles=particles, molecules_new=molecules, &
                         local_particles=local_particles, local_molecules_new=local_molecules, &
                         gci=gci, molecule_kinds_new=molecule_kinds)

      nkind = molecule_kinds%n_els
      particle_set => particles%els
      molecule_set => molecules%els
      molecule_kind_set => molecule_kinds%els

      ! Intramolecular Restraints
      ALLOCATE (force(3, SIZE(particle_set)))
      force = 0.0_dp

      ! Create the list of locally fixed atoms
      CALL create_local_fixd_list(lfixd_list, nkind, molecule_kind_set, local_particles)

      DO ifixd = 1, SIZE(lfixd_list)
         ikind = lfixd_list(ifixd)%ikind
         ii = lfixd_list(ifixd)%ifixd_index
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind(molecule_kind, fixd_list=fixd_list)
         IF (fixd_list(ii)%restraint%active) THEN
            n_restraint = n_restraint+1
            iparticle = fixd_list(ii)%fixd
            k0 = fixd_list(ii)%restraint%k0
            targ = fixd_list(ii)%coord
            rab = 0.0_dp
            SELECT CASE (fixd_list (ii)%itype)
            CASE (use_perd_x)
               rab(1) = particle_set(iparticle)%r(1)-targ(1)
            CASE (use_perd_y)
               rab(2) = particle_set(iparticle)%r(2)-targ(2)
            CASE (use_perd_z)
               rab(3) = particle_set(iparticle)%r(3)-targ(3)
            CASE (use_perd_xy)
               rab(1) = particle_set(iparticle)%r(1)-targ(1)
               rab(2) = particle_set(iparticle)%r(2)-targ(2)
            CASE (use_perd_xz)
               rab(1) = particle_set(iparticle)%r(1)-targ(1)
               rab(3) = particle_set(iparticle)%r(3)-targ(3)
            CASE (use_perd_yz)
               rab(2) = particle_set(iparticle)%r(2)-targ(2)
               rab(3) = particle_set(iparticle)%r(3)-targ(3)
            CASE (use_perd_xyz)
               rab = particle_set(iparticle)%r-targ
            END SELECT
            rab2 = DOT_PRODUCT(rab, rab)
            ! Energy
            energy_fixd = energy_fixd+k0*rab2
            ! Forces
            force(:, iparticle) = force(:, iparticle)-2.0_dp*k0*rab
         END IF
      END DO
      CALL release_local_fixd_list(lfixd_list)

      ! Loop over other kind of Restraints
      MOL: DO ikind = 1, nkind
         molecule_kind => molecule_kind_set(ikind)
         nmol_per_kind = local_molecules%n_el(ikind)
         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            molecule_kind => molecule%molecule_kind

            CALL get_molecule_kind(molecule_kind, &
                                   ncolv=ncolv, &
                                   ng3x3_restraint=n3x3con_restraint, &
                                   ng4x6_restraint=n4x6con_restraint)
            ! 3x3
            IF (n3x3con_restraint /= 0) THEN
               n_restraint = n_restraint+n3x3con_restraint
               CALL restraint_3x3_int(molecule, particle_set, energy_3x3, force)
            ENDIF
            ! 4x6
            IF (n4x6con_restraint /= 0) THEN
               n_restraint = n_restraint+n4x6con_restraint
               CALL restraint_4x6_int(molecule, particle_set, energy_4x6, force)
            ENDIF
            ! collective variables
            IF (ncolv%nrestraint /= 0) THEN
               n_restraint = n_restraint+ncolv%nrestraint
               CALL restraint_colv_int(molecule, particle_set, cell, energy_colv, force)
            ENDIF
         END DO
      END DO MOL
      CALL mp_sum(n_restraint, force_env%para_env%group)
      IF (n_restraint > 0) THEN
         CALL mp_sum(energy_fixd, force_env%para_env%group)
         CALL mp_sum(energy_3x3, force_env%para_env%group)
         CALL mp_sum(energy_4x6, force_env%para_env%group)
         CALL mp_sum(energy_colv, force_env%para_env%group)
         CALL update_particle_set(particle_set, force_env%para_env%group, for=force, add=.TRUE.)
         force = 0.0_dp
         n_restraint = 0
      ENDIF
      ! Intermolecular Restraints
      IF (ASSOCIATED(gci)) THEN
         IF (gci%nrestraint > 0) THEN
            ! 3x3
            IF (gci%ng3x3_restraint /= 0) THEN
               n_restraint = n_restraint+gci%ng3x3_restraint
               CALL restraint_3x3_ext(gci, particle_set, energy_3x3, force)
            ENDIF
            ! 4x6
            IF (gci%ng4x6_restraint /= 0) THEN
               n_restraint = n_restraint+gci%ng4x6_restraint
               CALL restraint_4x6_ext(gci, particle_set, energy_4x6, force)
            ENDIF
            ! collective variables
            IF (gci%ncolv%nrestraint /= 0) THEN
               n_restraint = n_restraint+gci%ncolv%nrestraint
               CALL restraint_colv_ext(gci, particle_set, cell, energy_colv, force)
            ENDIF
            DO iparticle = 1, SIZE(particle_set)
               particle_set(iparticle)%f = particle_set(iparticle)%f+force(:, iparticle)
            END DO
         END IF
      END IF
      DEALLOCATE (force)

      ! Store restraint energies
      CALL force_env_get(force_env=force_env, additional_potential=extended_energies)
      extended_energies = extended_energies+energy_3x3+ &
                          energy_fixd+ &
                          energy_4x6+ &
                          energy_colv
      CALL force_env_set(force_env=force_env, additional_potential=extended_energies)
      CALL timestop(handle)

   END SUBROUTINE restraint_control

! **************************************************************************************************
!> \brief Computes restraints 3x3 - Intramolecular
!> \param molecule ...
!> \param particle_set ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_3x3_int(molecule, particle_set, energy, force)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_3x3_int', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, ng3x3
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, ng3x3=ng3x3, g3x3_list=g3x3_list, &
                             fixd_list=fixd_list)
      CALL get_molecule(molecule, first_atom=first_atom)
      CALL restraint_3x3_low(ng3x3, g3x3_list, fixd_list, first_atom, particle_set, &
                             energy, force)

   END SUBROUTINE restraint_3x3_int

! **************************************************************************************************
!> \brief Computes restraints 4x6 - Intramolecular
!> \param molecule ...
!> \param particle_set ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_4x6_int(molecule, particle_set, energy, force)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_4x6_int', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, ng4x6=ng4x6, g4x6_list=g4x6_list, &
                             fixd_list=fixd_list)
      CALL get_molecule(molecule, first_atom=first_atom)
      CALL restraint_4x6_low(ng4x6, g4x6_list, fixd_list, first_atom, particle_set, &
                             energy, force)

   END SUBROUTINE restraint_4x6_int

! **************************************************************************************************
!> \brief Computes restraints colv - Intramolecular
!> \param molecule ...
!> \param particle_set ...
!> \param cell ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_colv_int(molecule, particle_set, cell, energy, force)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_colv_int', &
         routineP = moduleN//':'//routineN

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)

      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, colv_list=colv_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, lcolv=lcolv)
      CALL restraint_colv_low(colv_list, fixd_list, lcolv, particle_set, &
                              cell, energy, force)

   END SUBROUTINE restraint_colv_int

! **************************************************************************************************
!> \brief Computes restraints 3x3 - Intermolecular
!> \param gci ...
!> \param particle_set ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_3x3_ext(gci, particle_set, energy, force)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_3x3_ext', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, ng3x3
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)

      first_atom = 1
      ng3x3 = gci%ng3x3
      g3x3_list => gci%g3x3_list
      fixd_list => gci%fixd_list
      CALL restraint_3x3_low(ng3x3, g3x3_list, fixd_list, first_atom, particle_set, &
                             energy, force)

   END SUBROUTINE restraint_3x3_ext

! **************************************************************************************************
!> \brief Computes restraints 4x6 - Intermolecular
!> \param gci ...
!> \param particle_set ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_4x6_ext(gci, particle_set, energy, force)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_4x6_ext', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)

      first_atom = 1
      ng4x6 = gci%ng4x6
      g4x6_list => gci%g4x6_list
      fixd_list => gci%fixd_list
      CALL restraint_4x6_low(ng4x6, g4x6_list, fixd_list, first_atom, particle_set, &
                             energy, force)

   END SUBROUTINE restraint_4x6_ext

! **************************************************************************************************
!> \brief Computes restraints colv - Intermolecular
!> \param gci ...
!> \param particle_set ...
!> \param cell ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_colv_ext(gci, particle_set, cell, energy, force)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_colv_ext', &
         routineP = moduleN//':'//routineN

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)

      colv_list => gci%colv_list
      fixd_list => gci%fixd_list
      lcolv => gci%lcolv
      CALL restraint_colv_low(colv_list, fixd_list, lcolv, particle_set, &
                              cell, energy, force)

   END SUBROUTINE restraint_colv_ext

! **************************************************************************************************
!> \brief Computes restraints 3x3 - Real 3x3 restraints
!> \param ng3x3 ...
!> \param g3x3_list ...
!> \param fixd_list ...
!> \param first_atom ...
!> \param particle_set ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_3x3_low(ng3x3, g3x3_list, fixd_list, first_atom, &
                                particle_set, energy, force)

      INTEGER                                            :: ng3x3
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      INTEGER, INTENT(IN)                                :: first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_3x3_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iconst, index_a, index_b, index_c
      REAL(KIND=dp)                                      :: k, rab, rac, rbc, tab, tac, tbc
      REAL(KIND=dp), DIMENSION(3)                        :: r0_12, r0_13, r0_23

      DO iconst = 1, ng3x3
         IF (.NOT. g3x3_list(iconst)%restraint%active) CYCLE
         index_a = g3x3_list(iconst)%a+first_atom-1
         index_b = g3x3_list(iconst)%b+first_atom-1
         index_c = g3x3_list(iconst)%c+first_atom-1
         r0_12(:) = particle_set(index_a)%r-particle_set(index_b)%r
         r0_13(:) = particle_set(index_a)%r-particle_set(index_c)%r
         r0_23(:) = particle_set(index_b)%r-particle_set(index_c)%r

         rab = SQRT(DOT_PRODUCT(r0_12, r0_12))
         rac = SQRT(DOT_PRODUCT(r0_13, r0_13))
         rbc = SQRT(DOT_PRODUCT(r0_23, r0_23))
         tab = rab-g3x3_list(ng3x3)%dab
         tac = rac-g3x3_list(ng3x3)%dac
         tbc = rbc-g3x3_list(ng3x3)%dbc
         k = g3x3_list(iconst)%restraint%k0
         ! Update Energy
         energy = energy+k*(tab**2+tac**2+tbc**2)
         ! Update Forces
         force(:, index_a) = force(:, index_a)-2.0_dp*k*(r0_12/rab*tab+r0_13/rac*tac)
         force(:, index_b) = force(:, index_b)-2.0_dp*k*(-r0_12/rab*tab+r0_23/rbc*tbc)
         force(:, index_c) = force(:, index_c)-2.0_dp*k*(-r0_13/rac*tac-r0_23/rbc*tbc)
         ! Fixed atoms
         IF (ASSOCIATED(fixd_list)) THEN
            IF (SIZE(fixd_list) > 0) THEN
               IF (ANY(fixd_list(:)%fixd == index_a)) force(:, index_a) = 0.0_dp
               IF (ANY(fixd_list(:)%fixd == index_b)) force(:, index_b) = 0.0_dp
               IF (ANY(fixd_list(:)%fixd == index_c)) force(:, index_c) = 0.0_dp
            END IF
         END IF
      END DO

   END SUBROUTINE restraint_3x3_low

! **************************************************************************************************
!> \brief Computes restraints 4x6 - Real 4x6 restraints
!> \param ng4x6 ...
!> \param g4x6_list ...
!> \param fixd_list ...
!> \param first_atom ...
!> \param particle_set ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_4x6_low(ng4x6, g4x6_list, fixd_list, first_atom, &
                                particle_set, energy, force)

      INTEGER                                            :: ng4x6
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      INTEGER, INTENT(IN)                                :: first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_4x6_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iconst, index_a, index_b, index_c, &
                                                            index_d
      REAL(KIND=dp)                                      :: k, rab, rac, rad, rbc, rbd, rcd, tab, &
                                                            tac, tad, tbc, tbd, tcd
      REAL(KIND=dp), DIMENSION(3)                        :: r0_12, r0_13, r0_14, r0_23, r0_24, r0_34

      DO iconst = 1, ng4x6
         IF (.NOT. g4x6_list(iconst)%restraint%active) CYCLE
         index_a = g4x6_list(iconst)%a+first_atom-1
         index_b = g4x6_list(iconst)%b+first_atom-1
         index_c = g4x6_list(iconst)%c+first_atom-1
         index_d = g4x6_list(iconst)%d+first_atom-1
         r0_12(:) = particle_set(index_a)%r-particle_set(index_b)%r
         r0_13(:) = particle_set(index_a)%r-particle_set(index_c)%r
         r0_14(:) = particle_set(index_a)%r-particle_set(index_d)%r
         r0_23(:) = particle_set(index_b)%r-particle_set(index_c)%r
         r0_24(:) = particle_set(index_b)%r-particle_set(index_d)%r
         r0_34(:) = particle_set(index_c)%r-particle_set(index_d)%r

         rab = SQRT(DOT_PRODUCT(r0_12, r0_12))
         rac = SQRT(DOT_PRODUCT(r0_13, r0_13))
         rad = SQRT(DOT_PRODUCT(r0_14, r0_14))
         rbc = SQRT(DOT_PRODUCT(r0_23, r0_23))
         rbd = SQRT(DOT_PRODUCT(r0_24, r0_24))
         rcd = SQRT(DOT_PRODUCT(r0_34, r0_34))

         tab = rab-g4x6_list(ng4x6)%dab
         tac = rac-g4x6_list(ng4x6)%dac
         tad = rad-g4x6_list(ng4x6)%dad
         tbc = rbc-g4x6_list(ng4x6)%dbc
         tbd = rbd-g4x6_list(ng4x6)%dbd
         tcd = rcd-g4x6_list(ng4x6)%dcd

         k = g4x6_list(iconst)%restraint%k0
         ! Update Energy
         energy = energy+k*(tab**2+tac**2+tad**2+tbc**2+tbd**2+tcd**2)
         ! Update Forces
         force(:, index_a) = force(:, index_a)-2.0_dp*k*(r0_12/rab*tab+r0_13/rac*tac+r0_14/rad*tad)
         force(:, index_b) = force(:, index_b)-2.0_dp*k*(-r0_12/rab*tab+r0_23/rbc*tbc+r0_24/rbd*tbd)
         force(:, index_c) = force(:, index_c)-2.0_dp*k*(-r0_13/rac*tac-r0_23/rbc*tbc+r0_34/rcd*tcd)
         force(:, index_d) = force(:, index_d)-2.0_dp*k*(-r0_14/rad*tad-r0_24/rbd*tbd-r0_34/rcd*tcd)
         ! Fixed atoms
         IF (ASSOCIATED(fixd_list)) THEN
            IF (SIZE(fixd_list) > 0) THEN
               IF (ANY(fixd_list(:)%fixd == index_a)) force(:, index_a) = 0.0_dp
               IF (ANY(fixd_list(:)%fixd == index_b)) force(:, index_b) = 0.0_dp
               IF (ANY(fixd_list(:)%fixd == index_c)) force(:, index_c) = 0.0_dp
               IF (ANY(fixd_list(:)%fixd == index_d)) force(:, index_d) = 0.0_dp
            END IF
         END IF
      END DO

   END SUBROUTINE restraint_4x6_low

! **************************************************************************************************
!> \brief Computes restraints colv - Real COLVAR restraints
!> \param colv_list ...
!> \param fixd_list ...
!> \param lcolv ...
!> \param particle_set ...
!> \param cell ...
!> \param energy ...
!> \param force ...
!> \author Teodoro Laino 08.2006 [tlaino]
! **************************************************************************************************
   SUBROUTINE restraint_colv_low(colv_list, fixd_list, lcolv, &
                                 particle_set, cell, energy, force)

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restraint_colv_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatm, iconst, ind
      REAL(KIND=dp)                                      :: k, tab, targ

      DO iconst = 1, SIZE(colv_list)
         IF (.NOT. colv_list(iconst)%restraint%active) CYCLE
         ! Update colvar
         CALL colvar_eval_mol_f(lcolv(iconst)%colvar, cell, &
                                particles=particle_set, fixd_list=fixd_list)

         k = colv_list(iconst)%restraint%k0
         targ = colv_list(iconst)%expected_value
         tab = diff_colvar(lcolv(iconst)%colvar, targ)
         ! Update Energy
         energy = energy+k*tab**2
         ! Update Forces
         DO iatm = 1, SIZE(lcolv(iconst)%colvar%i_atom)
            ind = lcolv(iconst)%colvar%i_atom(iatm)
            force(:, ind) = force(:, ind)-2.0_dp*k*tab*lcolv(iconst)%colvar%dsdr(:, iatm)
         END DO
      END DO

   END SUBROUTINE restraint_colv_low

END MODULE restraint
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Empirical interatomic potentials for Silicon
!> \note
!>      Stefan Goedecker's OpenMP implementation of Bazant's EDIP & Lenosky's
!>      empirical interatomic potentials for Silicon.
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
MODULE eip_silicon
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE eip_environment_types,           ONLY: eip_env_get,&
                                              eip_environment_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: angstrom,&
                                              evolt

  !$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE
  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'eip_silicon'

  ! *** Public subroutines ***
  PUBLIC :: eip_bazant, eip_lenosky

!***

CONTAINS

! **************************************************************************************************
!> \brief Interface routine of Goedecker's Bazant EDIP to CP2K
!> \param eip_env ...
!> \par Literature
!>      http://www-math.mit.edu/~bazant/EDIP
!>      M.Z. Bazant & E. Kaxiras: Modeling of Covalent Bonding in Solids by
!>                                Inversion of Cohesive Energy Curves;
!>                                Phys. Rev. Lett. 77, 4370 (1996)
!>      M.Z. Bazant, E. Kaxiras and J.F. Justo: Environment-dependent interatomic
!>                                              potential for bulk silicon;
!>                                              Phys. Rev. B 56, 8542-8552 (1997)
!>      S. Goedecker: Optimization and parallelization of a force field for silicon
!>                    using OpenMP; CPC 148, 1 (2002)
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
  SUBROUTINE eip_bazant(eip_env)
      TYPE(eip_environment_type), POINTER                :: eip_env

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_bazant', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iparticle, iparticle_kind, &
                                                            iparticle_local, iw, natom, &
                                                            nparticle_kind, nparticle_local
      REAL(KIND=dp)                                      :: ekin, ener, ener_var, mass
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rxyz
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: eip_section

!   ------------------------------------------------------------------------

    CALL timeset(routineN,handle)

    NULLIFY(cell, particle_set, eip_section, logger, atomic_kinds, &
            atomic_kind, local_particles, subsys, atomic_kind_set, para_env)

    ekin = 0.0_dp

    logger => cp_get_default_logger()

    IF(.NOT.(ASSOCIATED(eip_env)))CALL cp__a("eip_silicon.F",106)


    CALL eip_env_get(eip_env=eip_env, cell=cell, particle_set=particle_set, &
                     subsys=subsys, local_particles=local_particles, &
                     atomic_kind_set=atomic_kind_set)
    CALL get_cell(cell=cell, abc=abc)

    eip_section => section_vals_get_subs_vals(eip_env%force_env_input, "EIP")
    natom = SIZE(particle_set)
    !natom = local_particles%n_el(1)

    ALLOCATE(rxyz(3,natom))

    DO i = 1,natom
      !iparticle = local_particles%list(1)%array(i)
      rxyz(:,i) = particle_set(i)%r(:)*angstrom
    END DO

    CALL eip_bazant_silicon(nat=natom, alat=abc*angstrom, rxyz0=rxyz, &
                            fxyz=eip_env%eip_forces, ener=ener, &
                            coord=eip_env%coord_avg, ener_var=ener_var, &
                            coord_var=eip_env%coord_var, count=eip_env%count)

    !CALL get_part_ke(md_env, tbmd_energy%E_kinetic, int_grp=globalenv%para_env%group)
    CALL cp_subsys_get(subsys=subsys, atomic_kinds=atomic_kinds)

    nparticle_kind = atomic_kinds%n_els

    DO iparticle_kind = 1,nparticle_kind
      atomic_kind => atomic_kind_set(iparticle_kind)
      CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
      nparticle_local = local_particles%n_el(iparticle_kind)
      DO iparticle_local = 1,nparticle_local
        iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
        ekin = ekin +  0.5_dp * mass * &
                   (particle_set(iparticle)%v(1)* particle_set(iparticle)%v(1) &
                  + particle_set(iparticle)%v(2)* particle_set(iparticle)%v(2) &
                  + particle_set(iparticle)%v(3)* particle_set(iparticle)%v(3))
      END DO
    END DO

    ! sum all contributions to energy over calculated parts on all processors
    CALL cp_subsys_get(subsys=subsys, para_env=para_env)
    CALL mp_sum(ekin, para_env%group)
    eip_env%eip_kinetic_energy = ekin

    eip_env%eip_potential_energy = ener/evolt
    eip_env%eip_energy = eip_env%eip_kinetic_energy + eip_env%eip_potential_energy
    eip_env%eip_energy_var = ener_var/evolt

    DO i = 1,natom
      particle_set(i)%f(:) = eip_env%eip_forces(:,i)/evolt*angstrom
    END DO

    DEALLOCATE(rxyz)

    ! Print
    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%ENERGIES"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%ENERGIES", &
                                extension=".mmLog")

      CALL eip_print_energies(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%ENERGIES")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%ENERGIES_VAR"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%ENERGIES_VAR", &
                                extension=".mmLog")

      CALL eip_print_energy_var(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%ENERGIES_VAR")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%FORCES"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%FORCES", &
                                extension=".mmLog")

      CALL eip_print_forces(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%FORCES")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%COORD_AVG"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%COORD_AVG", &
                                extension=".mmLog")

      CALL eip_print_coord_avg(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%COORD_AVG")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%COORD_VAR"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%COORD_VAR", &
                                extension=".mmLog")

      CALL eip_print_coord_var(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%COORD_VAR")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%COUNT"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%COUNT", &
                                extension=".mmLog")

      CALL eip_print_count(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%COUNT")
    END IF


    CALL timestop(handle)

  END SUBROUTINE eip_bazant

! **************************************************************************************************
!> \brief Interface routine of Goedecker's Lenosky force field to CP2K
!> \param eip_env ...
!> \par Literature
!>      T. Lenosky, et. al.: Highly optimized empirical potential model of silicon;
!>                           Modelling Simul. Sci. Eng., 8 (2000)
!>      S. Goedecker: Optimization and parallelization of a force field for silicon
!>                    using OpenMP; CPC 148, 1 (2002)
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
  SUBROUTINE eip_lenosky(eip_env)
      TYPE(eip_environment_type), POINTER                :: eip_env

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_lenosky', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iparticle, iparticle_kind, &
                                                            iparticle_local, iw, natom, &
                                                            nparticle_kind, nparticle_local
      REAL(KIND=dp)                                      :: ekin, ener, ener_var, mass
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rxyz
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: eip_section

!   ------------------------------------------------------------------------

    CALL timeset(routineN,handle)

    NULLIFY(cell, particle_set, eip_section, logger, atomic_kinds, &
            atomic_kind, local_particles, subsys, atomic_kind_set, para_env)

    ekin = 0.0_dp

    logger => cp_get_default_logger()

    IF(.NOT.(ASSOCIATED(eip_env)))CALL cp__a("eip_silicon.F",274)


    CALL eip_env_get(eip_env=eip_env, cell=cell, particle_set=particle_set, &
                     subsys=subsys, local_particles=local_particles, &
                     atomic_kind_set=atomic_kind_set)
    CALL get_cell(cell=cell, abc=abc)

    eip_section => section_vals_get_subs_vals(eip_env%force_env_input, "EIP")
    natom = SIZE(particle_set)
    !natom = local_particles%n_el(1)

    ALLOCATE(rxyz(3,natom))

    DO i = 1,natom
      !iparticle = local_particles%list(1)%array(i)
      rxyz(:,i) = particle_set(i)%r(:)*angstrom
    END DO

    CALL eip_lenosky_silicon(nat=natom, alat=abc*angstrom, rxyz0=rxyz, &
                             fxyz=eip_env%eip_forces, ener=ener, &
                             coord=eip_env%coord_avg, ener_var=ener_var, &
                             coord_var=eip_env%coord_var, count=eip_env%count)

    !CALL get_part_ke(md_env, tbmd_energy%E_kinetic, int_grp=globalenv%para_env%group)
    CALL cp_subsys_get(subsys=subsys, atomic_kinds=atomic_kinds)

    nparticle_kind = atomic_kinds%n_els

    DO iparticle_kind = 1,nparticle_kind
      atomic_kind => atomic_kind_set(iparticle_kind)
      CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
      nparticle_local = local_particles%n_el(iparticle_kind)
      DO iparticle_local = 1,nparticle_local
        iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
        ekin = ekin +  0.5_dp * mass * &
                   (particle_set(iparticle)%v(1)* particle_set(iparticle)%v(1) &
                  + particle_set(iparticle)%v(2)* particle_set(iparticle)%v(2) &
                  + particle_set(iparticle)%v(3)* particle_set(iparticle)%v(3))
      END DO
    END DO

    ! sum all contributions to energy over calculated parts on all processors
    CALL cp_subsys_get(subsys=subsys, para_env=para_env)
    CALL mp_sum(ekin, para_env%group)
    eip_env%eip_kinetic_energy = ekin

    eip_env%eip_potential_energy = ener/evolt
    eip_env%eip_energy = eip_env%eip_kinetic_energy + eip_env%eip_potential_energy
    eip_env%eip_energy_var = ener_var/evolt

    DO i = 1,natom
      particle_set(i)%f(:) = eip_env%eip_forces(:,i)/evolt*angstrom
    END DO

    DEALLOCATE(rxyz)

    ! Print
    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%ENERGIES"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%ENERGIES", &
                                extension=".mmLog")

      CALL eip_print_energies(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%ENERGIES")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%ENERGIES_VAR"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%ENERGIES_VAR", &
                                extension=".mmLog")

      CALL eip_print_energy_var(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%ENERGIES_VAR")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%FORCES"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%FORCES", &
                                extension=".mmLog")

      CALL eip_print_forces(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%FORCES")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%COORD_AVG"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%COORD_AVG", &
                                extension=".mmLog")

      CALL eip_print_coord_avg(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%COORD_AVG")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%COORD_VAR"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%COORD_VAR", &
                                extension=".mmLog")

      CALL eip_print_coord_var(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%COORD_VAR")
    END IF

    IF (BTEST(cp_print_key_should_output(logger%iter_info, &
              eip_section, "PRINT%COUNT"), cp_p_file)) THEN
      iw = cp_print_key_unit_nr(logger, eip_section, "PRINT%COUNT", &
                                extension=".mmLog")

      CALL eip_print_count(eip_env=eip_env, output_unit=iw)
      CALL cp_print_key_finished_output(iw, logger, eip_section, &
                                        "PRINT%COUNT")
    END IF


    CALL timestop(handle)

  END SUBROUTINE eip_lenosky

! **************************************************************************************************
!> \brief Print routine for the EIP energies
!> \param eip_env The eip environment of matter
!> \param output_unit The output unit
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
!> \note
!>      As usual the EIP energies differ from the DFT energies!
!>      Only the relative energy differneces are correctly reproduced.
! **************************************************************************************************
  SUBROUTINE eip_print_energies(eip_env, output_unit)
      TYPE(eip_environment_type), POINTER                :: eip_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_print_energies', &
         routineP = moduleN//':'//routineN

!   ------------------------------------------------------------------------

    IF (output_unit > 0) THEN
      WRITE (UNIT=output_unit,FMT="(/,(T3,A,T55,F25.14))")&
             "Kinetic energy [Hartree]:        ",eip_env%eip_kinetic_energy, &
             "Potential energy [Hartree]:      ",eip_env%eip_potential_energy, &
             "Total EIP energy [Hartree]:      ",eip_env%eip_energy
    END IF

  END SUBROUTINE eip_print_energies

! **************************************************************************************************
!> \brief Print routine for the variance of the energy/atom
!> \param eip_env The eip environment of matter
!> \param output_unit The output unit
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
  SUBROUTINE eip_print_energy_var(eip_env, output_unit)
      TYPE(eip_environment_type), POINTER                :: eip_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_print_energy_var', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr

!   ------------------------------------------------------------------------

    unit_nr = output_unit

    IF (unit_nr > 0) THEN

      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) "The variance of the EIP energy/atom!"
      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) eip_env%eip_energy_var

    END IF

  END SUBROUTINE eip_print_energy_var

! **************************************************************************************************
!> \brief Print routine for the forces
!> \param eip_env The eip environment of matter
!> \param output_unit The output unit
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
  SUBROUTINE eip_print_forces(eip_env, output_unit)
      TYPE(eip_environment_type), POINTER                :: eip_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_print_forces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, natom, unit_nr
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

!   ------------------------------------------------------------------------

    NULLIFY(particle_set)

    unit_nr = output_unit

    IF (unit_nr > 0) THEN

      CALL eip_env_get(eip_env=eip_env, particle_set=particle_set)

      natom = SIZE(particle_set)

      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) "The EIP forces!"
      WRITE (unit_nr,*) ""
      WRITE(unit_nr, *) "Total EIP forces [Hartree/Bohr]"
      DO iatom = 1, natom
        WRITE (unit_nr,*) eip_env%eip_forces(1:3,iatom)
      END DO

    END IF

  END SUBROUTINE eip_print_forces

! **************************************************************************************************
!> \brief Print routine for the average coordination number
!> \param eip_env The eip environment of matter
!> \param output_unit The output unit
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
  SUBROUTINE eip_print_coord_avg(eip_env, output_unit)
      TYPE(eip_environment_type), POINTER                :: eip_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_print_coord_avg', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr

!   ------------------------------------------------------------------------

    unit_nr = output_unit

    IF (unit_nr > 0) THEN

      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) "The average coordination number!"
      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) eip_env%coord_avg

    END IF

  END SUBROUTINE eip_print_coord_avg

! **************************************************************************************************
!> \brief Print routine for the variance of the coordination number
!> \param eip_env The eip environment of matter
!> \param output_unit The output unit
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
  SUBROUTINE eip_print_coord_var(eip_env, output_unit)
      TYPE(eip_environment_type), POINTER                :: eip_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_print_coord_var', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr

!   ------------------------------------------------------------------------

    unit_nr = output_unit

    IF (unit_nr > 0) THEN

      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) "The variance of the coordination number!"
      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) eip_env%coord_var

    END IF

  END SUBROUTINE eip_print_coord_var

! **************************************************************************************************
!> \brief Print routine for the function call counter
!> \param eip_env The eip environment of matter
!> \param output_unit The output unit
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
  SUBROUTINE eip_print_count(eip_env, output_unit)
      TYPE(eip_environment_type), POINTER                :: eip_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'eip_print_count', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr

!   ------------------------------------------------------------------------

    unit_nr = output_unit

    IF (unit_nr > 0) THEN

      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) "The function call counter!"
      WRITE (unit_nr,*) ""
      WRITE (unit_nr,*) eip_env%count

    END IF

  END SUBROUTINE eip_print_count

! **************************************************************************************************
!> \brief Bazant's EDIP (environment-dependent interatomic potential) for Silicon
!>      by Stefan Goedecker
!> \param nat number of atoms
!> \param alat lattice constants of the orthorombic box containing the particles
!> \param rxyz0 atomic positions in Angstrom, may be modified on output.
!>               If an atom is outside the box the program will bring it back
!>               into the box by translations through alat
!> \param fxyz forces in eV/A
!> \param ener total energy in eV
!> \param coord average coordination number
!> \param ener_var variance of the energy/atom
!> \param coord_var variance of the coordination number
!> \param count count is increased by one per call, has to be initialized
!>                to 0.e0_dp before first call of eip_bazant
!> \par Literature
!>      http://www-math.mit.edu/~bazant/EDIP
!>      M.Z. Bazant & E. Kaxiras: Modeling of Covalent Bonding in Solids by
!>                                Inversion of Cohesive Energy Curves;
!>                                Phys. Rev. Lett. 77, 4370 (1996)
!>      M.Z. Bazant, E. Kaxiras and J.F. Justo: Environment-dependent interatomic
!>                                              potential for bulk silicon;
!>                                              Phys. Rev. B 56, 8542-8552 (1997)
!>      S. Goedecker: Optimization and parallelization of a force field for silicon
!>                    using OpenMP; CPC 148, 1 (2002)
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
    SUBROUTINE eip_bazant_silicon(nat,alat,rxyz0,fxyz,ener,coord,ener_var,&
                                  coord_var,count)

      INTEGER                                            :: nat
      REAL(KIND=dp)                                      :: alat, rxyz0, fxyz, ener, coord, &
                                                            ener_var, coord_var, count

        DIMENSION rxyz0(3,nat),fxyz(3,nat),alat(3)
        REAL(KIND=dp), ALLOCATABLE, DIMENSION(:,:) :: rxyz
        INTEGER, ALLOCATABLE, DIMENSION(:,:)       :: lsta
        INTEGER, ALLOCATABLE, DIMENSION(:)         :: lstb
        INTEGER, ALLOCATABLE, DIMENSION(:)         :: lay
        INTEGER, ALLOCATABLE, DIMENSION(:,:,:,:)   :: icell
        REAL(KIND=dp), ALLOCATABLE, DIMENSION(:,:) :: rel
        REAL(KIND=dp), ALLOCATABLE, DIMENSION(:,:) :: txyz
        REAL(KIND=dp), ALLOCATABLE, DIMENSION(:,:) :: s2,s3,sz
        INTEGER, ALLOCATABLE, DIMENSION(:)         :: num2,num3,numz

        REAL(KIND=dp) :: coord2, cut, cut2, ener2, rlc1i, rlc2i, rlc3i, tcoord, &
                         tcoord2, tener, tener2
        INTEGER       :: iam, iat, iat1, iat2, ii, i, il, in, indlst, indlstx, istop, &
                         istopg, l2, l3, laymx, ll1, ll2, ll3, lot, max_nbrs, myspace, &
                         l1, myspaceout, ncx, nn, nnbrx, npr

!        cut=par_a
        cut= 3.1213820e0_dp + 1.e-14_dp

        IF (count.eq.0)  OPEN(unit=10,file='bazant.mon',status='unknown')
        count=count+1.e0_dp

! linear scaling calculation of verlet list
        ll1=INT(alat(1)/cut)
        IF (ll1.lt.1) CALL cp__b("eip_silicon.F",657,"alat(1) too small")
        ll2=INT(alat(2)/cut)
        IF (ll2.lt.1) CALL cp__b("eip_silicon.F",659,"alat(2) too small")
        ll3=INT(alat(3)/cut)
        IF (ll3.lt.1) CALL cp__b("eip_silicon.F",661,"alat(3) too small")

! determine number of threads
        npr=1
!$OMP PARALLEL PRIVATE(iam)  SHARED (npr) DEFAULT(NONE)
!$       iam=omp_get_thread_num()
!$       if (iam.eq.0) npr=omp_get_num_threads()
!$OMP END PARALLEL

! linear scaling calculation of verlet list

     IF (npr.le.1) THEN !serial if too few processors to gain by parallelizing

! set ncx for serial case, ncx for parallel case set below
        ncx=16
        loop_ncx_s: DO
           ALLOCATE(icell(0:ncx,-1:ll1,-1:ll2,-1:ll3))
           icell(0,-1:ll1,-1:ll2,-1:ll3)=0
           rlc1i=ll1/alat(1)
           rlc2i=ll2/alat(2)
           rlc3i=ll3/alat(3)

           loop_iat_s: DO iat=1,nat
              rxyz0(1,iat)=MODULO(MODULO(rxyz0(1,iat),alat(1)),alat(1))
              rxyz0(2,iat)=MODULO(MODULO(rxyz0(2,iat),alat(2)),alat(2))
              rxyz0(3,iat)=MODULO(MODULO(rxyz0(3,iat),alat(3)),alat(3))
              l1=INT(rxyz0(1,iat)*rlc1i)
              l2=INT(rxyz0(2,iat)*rlc2i)
              l3=INT(rxyz0(3,iat)*rlc3i)

              ii=icell(0,l1,l2,l3)
              ii=ii+1
              icell(0,l1,l2,l3)=ii
              IF (ii.gt.ncx) THEN
                 WRITE(10,*) count,'NCX too small',ncx
                 DEALLOCATE(icell)
                 ncx=ncx*2
                 CYCLE loop_ncx_s
              ENDIF
              icell(ii,l1,l2,l3)=iat
           END DO loop_iat_s
           EXIT loop_ncx_s
        END DO loop_ncx_s

     ELSE  ! parallel case

! periodization of particles can be done in parallel
!$OMP PARALLEL DO SHARED (alat,nat,rxyz0) PRIVATE(iat) DEFAULT(NONE)
        DO iat=1,nat
           rxyz0(1,iat)=MODULO(MODULO(rxyz0(1,iat),alat(1)),alat(1))
           rxyz0(2,iat)=MODULO(MODULO(rxyz0(2,iat),alat(2)),alat(2))
           rxyz0(3,iat)=MODULO(MODULO(rxyz0(3,iat),alat(3)),alat(3))
        END DO
!$OMP END PARALLEL DO

! assignment to cell is done serially
! set ncx for parallel case, ncx for serial case set above
        ncx=16
        loop_ncx_p: DO
           ALLOCATE(icell(0:ncx,-1:ll1,-1:ll2,-1:ll3))
           icell(0,-1:ll1,-1:ll2,-1:ll3)=0

           rlc1i=ll1/alat(1)
           rlc2i=ll2/alat(2)
           rlc3i=ll3/alat(3)

           loop_iat_p: DO iat=1,nat
              l1=INT(rxyz0(1,iat)*rlc1i)
              l2=INT(rxyz0(2,iat)*rlc2i)
              l3=INT(rxyz0(3,iat)*rlc3i)
              ii=icell(0,l1,l2,l3)
              ii=ii+1
              icell(0,l1,l2,l3)=ii
              IF (ii.gt.ncx) THEN
                 WRITE(10,*) count,'NCX too small',ncx
                 DEALLOCATE(icell)
                 ncx=ncx*2
                 CYCLE loop_ncx_p
              ENDIF
              icell(ii,l1,l2,l3)=iat
           END DO loop_iat_p
           EXIT loop_ncx_p
        END DO loop_ncx_p

     ENDIF

! duplicate all atoms within boundary layer
        laymx=ncx*(2*ll1*ll2+2*ll1*ll3+2*ll2*ll3+4*ll1+4*ll2+4*ll3+8)
        nn=nat+laymx
        ALLOCATE(rxyz(3,nn),lay(nn))
        DO  iat=1,nat
        lay(iat)=iat
        rxyz(1,iat)=rxyz0(1,iat)
        rxyz(2,iat)=rxyz0(2,iat)
        rxyz(3,iat)=rxyz0(3,iat)
        ENDDO
        il=nat
! xy plane
        DO l2=0,ll2-1
        DO l1=0,ll1-1

        in=icell(0,l1,l2,0)
        icell(0,l1,l2,ll3)=in
        DO ii=1,in
        i=icell(ii,l1,l2,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",767,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,l2,ll3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,l1,l2,ll3-1)
        icell(0,l1,l2,-1)=in
        DO ii=1,in
        i=icell(ii,l1,l2,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",780,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,l2,-1)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        ENDDO
        ENDDO

! yz plane
        DO l3=0,ll3-1
        DO l2=0,ll2-1

        in=icell(0,0,l2,l3)
        icell(0,ll1,l2,l3)=in
        DO ii=1,in
        i=icell(ii,0,l2,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",800,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,l2,l3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,ll1-1,l2,l3)
        icell(0,-1,l2,l3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,l2,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",813,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,l2,l3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        ENDDO
        ENDDO

! xz plane
        DO l3=0,ll3-1
        DO l1=0,ll1-1

        in=icell(0,l1,0,l3)
        icell(0,l1,ll2,l3)=in
        DO ii=1,in
        i=icell(ii,l1,0,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",833,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,ll2,l3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,l1,ll2-1,l3)
        icell(0,l1,-1,l3)=in
        DO ii=1,in
        i=icell(ii,l1,ll2-1,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",846,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,-1,l3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        ENDDO
        ENDDO

! x axis
        DO l1=0,ll1-1

        in=icell(0,l1,0,0)
        icell(0,l1,ll2,ll3)=in
        DO ii=1,in
        i=icell(ii,l1,0,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",865,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,ll2,ll3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,l1,0,ll3-1)
        icell(0,l1,ll2,-1)=in
        DO ii=1,in
        i=icell(ii,l1,0,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",878,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,ll2,-1)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,l1,ll2-1,0)
        icell(0,l1,-1,ll3)=in
        DO ii=1,in
        i=icell(ii,l1,ll2-1,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",891,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,-1,ll3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,l1,ll2-1,ll3-1)
        icell(0,l1,-1,-1)=in
        DO ii=1,in
        i=icell(ii,l1,ll2-1,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",904,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,-1,-1)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        ENDDO

! y axis
        DO l2=0,ll2-1

        in=icell(0,0,l2,0)
        icell(0,ll1,l2,ll3)=in
        DO ii=1,in
        i=icell(ii,0,l2,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",922,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,l2,ll3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,0,l2,ll3-1)
        icell(0,ll1,l2,-1)=in
        DO ii=1,in
        i=icell(ii,0,l2,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",935,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,l2,-1)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,ll1-1,l2,0)
        icell(0,-1,l2,ll3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,l2,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",948,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,l2,ll3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,ll1-1,l2,ll3-1)
        icell(0,-1,l2,-1)=in
        DO ii=1,in
        i=icell(ii,ll1-1,l2,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",961,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,l2,-1)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        ENDDO

! z axis
        DO l3=0,ll3-1

        in=icell(0,0,0,l3)
        icell(0,ll1,ll2,l3)=in
        DO ii=1,in
        i=icell(ii,0,0,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",979,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,ll2,l3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,ll1-1,0,l3)
        icell(0,-1,ll2,l3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,0,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",992,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,ll2,l3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,0,ll2-1,l3)
        icell(0,ll1,-1,l3)=in
        DO ii=1,in
        i=icell(ii,0,ll2-1,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1005,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,-1,l3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,ll1-1,ll2-1,l3)
        icell(0,-1,-1,l3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,ll2-1,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1018,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,-1,l3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        ENDDO

! corners
        in=icell(0,0,0,0)
        icell(0,ll1,ll2,ll3)=in
        DO ii=1,in
        i=icell(ii,0,0,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1034,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,ll2,ll3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,ll1-1,0,0)
        icell(0,-1,ll2,ll3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,0,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1047,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,ll2,ll3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,0,ll2-1,0)
        icell(0,ll1,-1,ll3)=in
        DO ii=1,in
        i=icell(ii,0,ll2-1,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1060,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,-1,ll3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,ll1-1,ll2-1,0)
        icell(0,-1,-1,ll3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,ll2-1,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1073,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,-1,ll3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,0,0,ll3-1)
        icell(0,ll1,ll2,-1)=in
        DO ii=1,in
        i=icell(ii,0,0,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1086,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,ll2,-1)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,ll1-1,0,ll3-1)
        icell(0,-1,ll2,-1)=in
        DO ii=1,in
        i=icell(ii,ll1-1,0,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1099,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,ll2,-1)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,0,ll2-1,ll3-1)
        icell(0,ll1,-1,-1)=in
        DO ii=1,in
        i=icell(ii,0,ll2-1,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1112,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,-1,-1)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,ll1-1,ll2-1,ll3-1)
        icell(0,-1,-1,-1)=in
        DO ii=1,in
        i=icell(ii,ll1-1,ll2-1,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1125,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,-1,-1)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        ALLOCATE(lsta(2,nat))
        nnbrx=12
        loop_nnbrx: DO
           ALLOCATE(lstb(nnbrx*nat),rel(5,nnbrx*nat))

           indlstx=0

!$OMP PARALLEL DEFAULT(NONE) &
!$OMP PRIVATE(iat,cut2,iam,ii,indlst,l1,l2,l3,myspace,npr) &
!$OMP SHARED (indlstx,nat,nn,nnbrx,ncx,ll1,ll2,ll3,icell,lsta,lstb,lay, &
!$OMP rel,rxyz,cut,myspaceout)

           npr=1
!$         npr=omp_get_num_threads()
           iam=0
!$         iam=omp_get_thread_num()

           cut2=cut**2
! assign contiguous portions of the arrays lstb and rel to the threads
           myspace=(nat*nnbrx)/npr
           IF (iam.eq.0) myspaceout=myspace
! Verlet list, relative positions
           indlst=0
           loop_l3: DO l3=0,ll3-1
              loop_l2: DO l2=0,ll2-1
                 loop_l1: DO l1=0,ll1-1
                    loop_ii: DO ii=1,icell(0,l1,l2,l3)
                       iat=icell(ii,l1,l2,l3)
                       IF ( ((iat-1)*npr)/nat .EQ. iam) THEN
!                          write(*,*) 'sublstiat:iam,iat',iam,iat
                          lsta(1,iat)=iam*myspace+indlst+1
                          CALL sublstiat_b(iat,nn,ncx,ll1,ll2,ll3,l1,l2,l3,myspace, &
                                           rxyz,icell,lstb(iam*myspace+1),lay, &
                                           rel(1,iam*myspace+1),cut2,indlst)
                          lsta(2,iat)=iam*myspace+indlst
!                          write(*,'(a,4(x,i3),100(x,i2))') &
!                           'iam,iat,lsta',iam,iat,lsta(1,iat),lsta(2,iat), &
!                           (lstb(j),j=lsta(1,iat),lsta(2,iat))
                       END IF
                    END DO loop_ii
                 END DO loop_l1
              END DO loop_l2
           END DO loop_l3
!$OMP CRITICAL
           indlstx=MAX(indlstx,indlst)
!$OMP END CRITICAL
!$OMP END PARALLEL

           IF (indlstx.ge.myspaceout) THEN
              WRITE(10,*) count,'NNBRX too  small', nnbrx
              DEALLOCATE(lstb,rel)
              nnbrx=3*nnbrx/2
              CYCLE loop_nnbrx
           END IF
           EXIT loop_nnbrx
        END DO loop_nnbrx

        istopg=0

!$OMP PARALLEL DEFAULT(NONE)  &
!$OMP PRIVATE(iam,npr,iat,iat1,iat2,lot,istop,tcoord,tcoord2, &
!$OMP tener,tener2,txyz,s2,s3,sz,num2,num3,numz,max_nbrs) &
!$OMP SHARED (nat,nnbrx,lsta,lstb,rel,ener,ener2,fxyz,coord,coord2,istopg)

        npr=1
!$       npr=omp_get_num_threads()
        iam=0
!$       iam=omp_get_thread_num()

         max_nbrs=30

        IF (npr.ne.1) THEN
! PARALLEL CASE
! create temporary private scalars for reduction sum on energies and
!        temporary private array for reduction sum on forces
!$OMP CRITICAL
        ALLOCATE(txyz(3,nat),s2(max_nbrs,8),s3(max_nbrs,7),sz(max_nbrs,6),  &
                 num2(max_nbrs),num3(max_nbrs),numz(max_nbrs))
!$OMP END CRITICAL
        IF (iam.eq.0) THEN
        ener=0.e0_dp
        ener2=0.e0_dp
        coord=0.e0_dp
        coord2=0.e0_dp
        ENDIF
!$OMP DO
        DO iat=1,nat
        fxyz(1,iat)=0.e0_dp
        fxyz(2,iat)=0.e0_dp
        fxyz(3,iat)=0.e0_dp
        ENDDO
!$OMP BARRIER

! Each thread treats at most lot atoms
        lot=INT(float(nat)/float(npr)+.999999999999e0_dp)
        iat1=iam*lot+1
        iat2=MIN((iam+1)*lot,nat)
!       write(*,*) 'subfeniat:iat1,iat2,iam',iat1,iat2,iam
        CALL subfeniat_b(iat1,iat2,nat,lsta,lstb,rel,tener,tener2,  &
          tcoord,tcoord2,nnbrx,txyz,max_nbrs,istop,  &
          s2(1,1),s2(1,2),s2(1,3),s2(1,4),s2(1,5),s2(1,6),s2(1,7),s2(1,8),  &
          num2,s3(1,1),s3(1,2),s3(1,3),s3(1,4),s3(1,5),s3(1,6),s3(1,7),  &
          num3,sz(1,1),sz(1,2),sz(1,3),sz(1,4),sz(1,5),sz(1,6),numz)

!$OMP CRITICAL
        ener=ener+tener
        ener2=ener2+tener2
        coord=coord+tcoord
        coord2=coord2+tcoord2
        istopg=istopg+istop
        DO 8093,iat=1,nat
        fxyz(1,iat)=fxyz(1,iat)+txyz(1,iat)
        fxyz(2,iat)=fxyz(2,iat)+txyz(2,iat)
        fxyz(3,iat)=fxyz(3,iat)+txyz(3,iat)
8093    CONTINUE
        DEALLOCATE(txyz,s2,s3,sz,num2,num3,numz)
!$OMP END CRITICAL

        ELSE
! SERIAL CASE
        iat1=1
        iat2=nat
        ALLOCATE(s2(max_nbrs,8),s3(max_nbrs,7),sz(max_nbrs,6),  &
                 num2(max_nbrs),num3(max_nbrs),numz(max_nbrs))
        CALL subfeniat_b(iat1,iat2,nat,lsta,lstb,rel,ener,ener2,  &
          coord,coord2,nnbrx,fxyz,max_nbrs,istopg,  &
          s2(1,1),s2(1,2),s2(1,3),s2(1,4),s2(1,5),s2(1,6),s2(1,7),s2(1,8),  &
          num2,s3(1,1),s3(1,2),s3(1,3),s3(1,4),s3(1,5),s3(1,6),s3(1,7),  &
          num3,sz(1,1),sz(1,2),sz(1,3),sz(1,4),sz(1,5),sz(1,6),numz)
        DEALLOCATE(s2,s3,sz,num2,num3,numz)

        ENDIF
!$OMP END PARALLEL

!         write(*,*) 'ener,norm force', &
!                    ener,DNRM2(3*nat,fxyz,1)
        IF (istopg.gt.0) CALL cp__b("eip_silicon.F",1269,"DIMENSION ERROR (see WARNING above)")
        ener_var=ener2/nat-(ener/nat)**2
        coord=coord/nat
        coord_var=coord2/nat-coord**2

        DEALLOCATE(rxyz,icell,lay,lsta,lstb,rel)

        END SUBROUTINE eip_bazant_silicon

! **************************************************************************************************
!> \brief ...
!> \param iat1 ...
!> \param iat2 ...
!> \param nat ...
!> \param lsta ...
!> \param lstb ...
!> \param rel ...
!> \param ener ...
!> \param ener2 ...
!> \param coord ...
!> \param coord2 ...
!> \param nnbrx ...
!> \param ff ...
!> \param max_nbrs ...
!> \param istop ...
!> \param s2_t0 ...
!> \param s2_t1 ...
!> \param s2_t2 ...
!> \param s2_t3 ...
!> \param s2_dx ...
!> \param s2_dy ...
!> \param s2_dz ...
!> \param s2_r ...
!> \param num2 ...
!> \param s3_g ...
!> \param s3_dg ...
!> \param s3_rinv ...
!> \param s3_dx ...
!> \param s3_dy ...
!> \param s3_dz ...
!> \param s3_r ...
!> \param num3 ...
!> \param sz_df ...
!> \param sz_sum ...
!> \param sz_dx ...
!> \param sz_dy ...
!> \param sz_dz ...
!> \param sz_r ...
!> \param numz ...
! **************************************************************************************************
        SUBROUTINE subfeniat_b(iat1,iat2,nat,lsta,lstb,rel,ener,ener2,  &
          coord,coord2,nnbrx,ff,max_nbrs,istop,  &
          s2_t0,s2_t1,s2_t2,s2_t3,s2_dx,s2_dy,s2_dz,s2_r,  &
          num2,s3_g,s3_dg,s3_rinv,s3_dx,s3_dy,s3_dz,s3_r,  &
          num3,sz_df,sz_sum,sz_dx,sz_dy,sz_dz,sz_r,numz)
! This subroutine is a modification of a subroutine that is available at
! http://www-math.mit.edu/~bazant/EDIP/ and for which Martin Z. Bazant
! and Harvard University have a 1997 copyright.
! The modifications were done by S. Goedecker on April 10, 2002.
! The routines are included with the permission of M. Bazant into this package.

!  ------------------------- VARIABLE DECLARATIONS -------------------------
      INTEGER                                            :: iat1, iat2, nat, lsta(2,nat)
      REAL(KIND=dp)                                      :: ener, ener2, coord, coord2
      INTEGER                                            :: nnbrx
      REAL(KIND=dp)                                      :: rel(5,nnbrx*nat)
      INTEGER                                            :: lstb(nnbrx*nat)
      REAL(KIND=dp)                                      :: ff(3,nat)
      INTEGER                                            :: max_nbrs, istop
      REAL(KIND=dp) :: s2_t0(max_nbrs), s2_t1(max_nbrs), s2_t2(max_nbrs), s2_t3(max_nbrs), &
         s2_dx(max_nbrs), s2_dy(max_nbrs), s2_dz(max_nbrs), s2_r(max_nbrs)
      INTEGER                                            :: num2(max_nbrs)
      REAL(KIND=dp) :: s3_g(max_nbrs), s3_dg(max_nbrs), s3_rinv(max_nbrs), s3_dx(max_nbrs), &
         s3_dy(max_nbrs), s3_dz(max_nbrs), s3_r(max_nbrs)
      INTEGER                                            :: num3(max_nbrs)
      REAL(KIND=dp)                                      :: sz_df(max_nbrs), sz_sum(max_nbrs), &
                                                            sz_dx(max_nbrs), sz_dy(max_nbrs), &
                                                            sz_dz(max_nbrs), sz_r(max_nbrs)
      INTEGER                                            :: numz(max_nbrs)

      INTEGER                                            :: i, j, k, l, n, n2, n3, nj, nk, nl, nz
      REAL(KIND=dp) :: bmc, cmbinv, coord_iat, dEdrl, dEdrlx, dEdrly, dEdrlz, den, dhdl, dHdx, &
         dp1, dtau, dV2dZ, dV2ijx, dV2ijy, dV2ijz, dV2j, dV3dZ, dV3l, dV3ljx, dV3ljy, dV3ljz, &
         dV3lkx, dV3lky, dV3lkz, dV3rij, dV3rijx, dV3rijy, dV3rijz, dV3rik, dV3rikx, dV3riky, &
         dV3rikz, dwinv, dx, dxdZ, dy, dz, ener_iat, fjx, fjy, fjz, fkx, fky, fkz, fZ, H, lcos, &
         muhalf, par_a, par_alp, par_b, par_bet, par_bg, par_c, par_cap_A, par_cap_B, par_delta, &
         par_eta, par_gam, par_lam, par_mu, par_palp, par_Qo, par_rh, par_sig, pZ, Qort, r, rinv, &
         rmainv, rmbinv, tau, temp0, temp1, u1, u2, u3, u4, u5, winv, x, xarg
      REAL(KIND=dp) :: xinv, xinv3, Z

!   size of s2[]
!   atom ID numbers for s2[]
!   size of s3[]
!   atom ID numbers for s3[]
!   size of sz[]
!   atom ID numbers for sz[]
!   indices for the store arrays
!   EDIP parameters

          par_cap_A = 5.6714030e0_dp
          par_cap_B = 2.0002804e0_dp
          par_rh = 1.2085196e0_dp
          par_a = 3.1213820e0_dp
          par_sig = 0.5774108e0_dp
          par_lam = 1.4533108e0_dp
          par_gam = 1.1247945e0_dp
          par_b = 3.1213820e0_dp
          par_c = 2.5609104e0_dp
          par_delta = 78.7590539e0_dp
          par_mu = 0.6966326e0_dp
          par_Qo = 312.1341346e0_dp
          par_palp = 1.4074424e0_dp
          par_bet = 0.0070975e0_dp
          par_alp = 3.1083847e0_dp

          u1 = -0.165799e0_dp
          u2 = 32.557e0_dp
          u3 = 0.286198e0_dp
          u4 = 0.66e0_dp

          par_bg=par_a
          par_eta = par_delta/par_Qo

          DO i=1, nat
            ff(1,i) = 0.0e0_dp
            ff(2,i) = 0.0e0_dp
            ff(3,i) = 0.0e0_dp
          END DO

          coord=0.e0_dp
          coord2=0.e0_dp
          ener=0.e0_dp
          ener2=0.e0_dp
          istop=0

!   COMBINE COEFFICIENTS

          Qort = SQRT(par_Qo)
          muhalf = par_mu*0.5e0_dp
          u5 = u2*u4
          bmc = par_b-par_c
          cmbinv = 1.0e0_dp/(par_c-par_b)

!  --- LEVEL 1: OUTER LOOP OVER ATOMS ---

          DO 1000, i= iat1, iat2

!   RESET COORDINATION AND NEIGHBOR NUMBERS

            coord_iat=0.e0_dp
            ener_iat=0.e0_dp
            Z = 0.0e0_dp
            n2 = 1
            n3 = 1
            nz = 1

!  --- LEVEL 2: LOOP PREPASS OVER PAIRS ---

            DO n=lsta(1,i),lsta(2,i)
              j=lstb(n)

!   PARTS OF TWO-BODY INTERACTION r<par_a

                num2(n2) = j
                dx = -rel(1,n)
                dy = -rel(2,n)
                dz = -rel(3,n)
                r=rel(4,n)
                rinv=rel(5,n)
                rmainv = 1.e0_dp/(r-par_a)
                s2_t0(n2) = par_cap_A*EXP(par_sig*rmainv)
                s2_t1(n2) = (par_cap_B*rinv)**par_rh
                s2_t2(n2) = par_rh*rinv
                s2_t3(n2) = par_sig*rmainv*rmainv
                s2_dx(n2) = dx
                s2_dy(n2) = dy
                s2_dz(n2) = dz
                s2_r(n2)  = r
                n2 = n2 + 1
                IF (n2.gt.max_nbrs) THEN
                WRITE(*,*) 'WARNING enlarge max_nbrs'
                istop=1
                RETURN
                ENDIF

! coordination number calculated with soft cutoff between first
! nearest neighbor and midpoint of first and second nearest neighbor
        IF (r.le.2.36e0_dp) THEN
        coord_iat=coord_iat+1.e0_dp
        ELSE IF (r.ge.3.12e0_dp) THEN
        ELSE
        xarg=(r-2.36e0_dp)*(1.e0_dp/(3.12e0_dp-2.36e0_dp))
        coord_iat=coord_iat+(2*xarg+1.e0_dp)*(xarg-1.e0_dp)**2
        ENDIF

!   RADIAL PARTS OF THREE-BODY INTERACTION r<par_b

                IF(r .LT. par_bg)  THEN

                  num3(n3) = j
                  rmbinv = 1.e0_dp/(r-par_bg)
                  temp1 = par_gam*rmbinv
                  temp0 = EXP(temp1)
                  s3_g(n3) = temp0
                  s3_dg(n3) = -rmbinv*temp1*temp0
                  s3_dx(n3) = dx
                  s3_dy(n3) = dy
                  s3_dz(n3) = dz
                  s3_rinv(n3) = rinv
                  s3_r(n3) = r
                  n3 = n3 + 1
                  IF (n3.gt.max_nbrs) THEN
                  WRITE(*,*) 'WARNING enlarge max_nbrs'
                  istop=1
                  RETURN
                  ENDIF

!   COORDINATION AND NEIGHBOR FUNCTION par_c<r<par_b

                  IF(r .LT. par_b) THEN
                    IF(r .LT. par_c) THEN
                    Z = Z + 1.e0_dp
                   ELSE
                    xinv = bmc/(r-par_c)
                    xinv3 = xinv*xinv*xinv
                    den = 1.e0_dp/(1 - xinv3)
                    temp1 = par_alp*den
                    fZ = EXP(temp1)
                    Z = Z + fZ
                    numz(nz) = j
                    sz_df(nz) = fZ*temp1*den*3.e0_dp*xinv3*xinv*cmbinv
!   df/dr
                    sz_dx(nz) = dx
                    sz_dy(nz) = dy
                    sz_dz(nz) = dz
                    sz_r(nz) = r
                    nz = nz + 1
                    IF (nz.gt.max_nbrs) THEN
                    WRITE(*,*) 'WARNING enlarge max_nbrs'
                    istop=1
                    RETURN
                    ENDIF
                   END IF
!  r < par_C
                  END IF
!  r < par_b
                  END IF
!  r < par_bg
              END DO

!   ZERO ACCUMULATION ARRAY FOR ENVIRONMENT FORCES

              DO nl=1, nz-1
                sz_sum(nl)=0.e0_dp
              END DO

!   ENVIRONMENT-DEPENDENCE OF PAIR INTERACTION

              temp0 = par_bet*Z
              pZ = par_palp*EXP(-temp0*Z)
!   bond order
              dp1 = -2.e0_dp*temp0*pZ
!   derivative of bond order

!  --- LEVEL 2: LOOP FOR PAIR INTERACTIONS ---

            DO nj=1, n2-1

              temp0 = s2_t1(nj) - pZ

!   two-body energy V2(rij,Z)

              ener_iat = ener_iat + temp0*s2_t0(nj)

!   two-body forces

              dV2j = - s2_t0(nj) * (s2_t1(nj)*s2_t2(nj) + temp0 * s2_t3(nj))
!   dV2/dr
              dV2ijx = dV2j * s2_dx(nj)
              dV2ijy = dV2j * s2_dy(nj)
              dV2ijz = dV2j * s2_dz(nj)
              ff(1,i) = ff(1,i) + dV2ijx
              ff(2,i) = ff(2,i) + dV2ijy
              ff(3,i) = ff(3,i) + dV2ijz
              j = num2(nj)
              ff(1,j) = ff(1,j) - dV2ijx
              ff(2,j) = ff(2,j) - dV2ijy
              ff(3,j) = ff(3,j) - dV2ijz

!  --- LEVEL 3: LOOP FOR PAIR COORDINATION FORCES ---

              dV2dZ = - dp1 * s2_t0(nj)
              DO nl=1, nz-1
                 sz_sum(nl) =  sz_sum(nl) + dV2dZ
              END DO

            END DO

!   COORDINATION-DEPENDENCE OF THREE-BODY INTERACTION

              winv = Qort*EXP(-muhalf*Z)
!   inverse width of angular function
              dwinv = -muhalf*winv
!   its derivative
              temp0 = EXP(-u4*Z)
              tau = u1+u2*temp0*(u3-temp0)
!   -cosine of angular minimum
              dtau = u5*temp0*(2*temp0-u3)
!   its derivative

!  --- LEVEL 2: FIRST LOOP FOR THREE-BODY INTERACTIONS ---

            DO nj=1, n3-2

              j=num3(nj)

!  --- LEVEL 3: SECOND LOOP FOR THREE-BODY INTERACTIONS ---

              DO nk=nj+1, n3-1

                k=num3(nk)

!   angular function h(l,Z)

                lcos=s3_dx(nj)*s3_dx(nk)+s3_dy(nj)*s3_dy(nk)+s3_dz(nj)*s3_dz(nk)
                x = (lcos + tau)*winv
                temp0 = EXP(-x*x)

                H = par_lam*(1 - temp0 + par_eta*x*x)
                dHdx = 2*par_lam*x*(temp0 + par_eta)

                dhdl = dHdx*winv

!   three-body energy

                temp1 = s3_g(nj) * s3_g(nk)
                ener_iat = ener_iat + temp1*H

!   (-) radial force on atom j

                dV3rij = s3_dg(nj) * s3_g(nk) * H
                dV3rijx = dV3rij * s3_dx(nj)
                dV3rijy = dV3rij * s3_dy(nj)
                dV3rijz = dV3rij * s3_dz(nj)
                fjx = dV3rijx
                fjy = dV3rijy
                fjz = dV3rijz

!   (-) radial force on atom k

                dV3rik = s3_g(nj) * s3_dg(nk) * H
                dV3rikx = dV3rik * s3_dx(nk)
                dV3riky = dV3rik * s3_dy(nk)
                dV3rikz = dV3rik * s3_dz(nk)
                fkx = dV3rikx
                fky = dV3riky
                fkz = dV3rikz

!   (-) angular force on j

                dV3l = temp1*dhdl
                dV3ljx = dV3l * (s3_dx(nk) - lcos * s3_dx(nj)) * s3_rinv(nj)
                dV3ljy = dV3l * (s3_dy(nk) - lcos * s3_dy(nj)) * s3_rinv(nj)
                dV3ljz = dV3l * (s3_dz(nk) - lcos * s3_dz(nj)) * s3_rinv(nj)
                fjx = fjx + dV3ljx
                fjy = fjy + dV3ljy
                fjz = fjz + dV3ljz

!   (-) angular force on k

                dV3lkx = dV3l * (s3_dx(nj) - lcos * s3_dx(nk)) * s3_rinv(nk)
                dV3lky = dV3l * (s3_dy(nj) - lcos * s3_dy(nk)) * s3_rinv(nk)
                dV3lkz = dV3l * (s3_dz(nj) - lcos * s3_dz(nk)) * s3_rinv(nk)
                fkx = fkx + dV3lkx
                fky = fky + dV3lky
                fkz = fkz + dV3lkz

!   apply radial + angular forces to i, j, k

                ff(1,j) = ff(1,j) - fjx
                ff(2,j) = ff(2,j) - fjy
                ff(3,j) = ff(3,j) - fjz
                ff(1,k) = ff(1,k) - fkx
                ff(2,k) = ff(2,k) - fky
                ff(3,k) = ff(3,k) - fkz
                ff(1,i) = ff(1,i) + fjx + fkx
                ff(2,i) = ff(2,i) + fjy + fky
                ff(3,i) = ff(3,i) + fjz + fkz

!   prefactor for 4-body forces from coordination
                  dxdZ = dwinv*(lcos + tau) + winv*dtau
                  dV3dZ = temp1*dHdx*dxdZ

!  --- LEVEL 4: LOOP FOR THREE-BODY COORDINATION FORCES ---

                  DO nl=1, nz-1
                    sz_sum(nl) = sz_sum(nl) + dV3dZ
                  END DO
              END DO
            END DO

!  --- LEVEL 2: LOOP TO APPLY COORDINATION FORCES ---

            DO nl=1, nz-1

                dEdrl = sz_sum(nl) * sz_df(nl)
                dEdrlx = dEdrl * sz_dx(nl)
                dEdrly = dEdrl * sz_dy(nl)
                dEdrlz = dEdrl * sz_dz(nl)
                ff(1,i) = ff(1,i) + dEdrlx
                ff(2,i) = ff(2,i) + dEdrly
                ff(3,i) = ff(3,i) + dEdrlz
                l = numz(nl)
                ff(1,l) = ff(1,l) - dEdrlx
                ff(2,l) = ff(2,l) - dEdrly
                ff(3,l) = ff(3,l) - dEdrlz

            END DO

        coord=coord+coord_iat
        coord2=coord2+coord_iat**2
        ener = ener + ener_iat
        ener2 = ener2 + ener_iat**2

1000      CONTINUE

        RETURN
        END SUBROUTINE subfeniat_b

! **************************************************************************************************
!> \brief ...
!> \param iat ...
!> \param nn ...
!> \param ncx ...
!> \param ll1 ...
!> \param ll2 ...
!> \param ll3 ...
!> \param l1 ...
!> \param l2 ...
!> \param l3 ...
!> \param myspace ...
!> \param rxyz ...
!> \param icell ...
!> \param lstb ...
!> \param lay ...
!> \param rel ...
!> \param cut2 ...
!> \param indlst ...
! **************************************************************************************************
        SUBROUTINE sublstiat_b(iat,nn,ncx,ll1,ll2,ll3,l1,l2,l3,myspace, &
                   rxyz,icell,lstb,lay,rel,cut2,indlst)
! finds the neighbours of atom iat (specified by lsta and lstb) and and
! the relative position rel of iat with respect to these neighbours
      INTEGER                                            :: iat, nn, ncx, ll1, ll2, ll3, l1, l2, l3, &
                                                            myspace
      REAL(KIND=dp)                                      :: rxyz
      INTEGER                                            :: icell, lstb, lay
      REAL(KIND=dp)                                      :: rel, cut2
      INTEGER                                            :: indlst

        DIMENSION rxyz(3,nn),lay(nn),icell(0:ncx,-1:ll1,-1:ll2,-1:ll3), &
                  lstb(0:myspace-1),rel(5,0:myspace-1)

        INTEGER       :: jat, k1, k2, k3, jj
        REAL(KIND=dp) :: rr2, tt, tti, xrel, yrel, zrel

        DO 6003,k3=l3-1,l3+1
        DO 6002,k2=l2-1,l2+1
        DO 6001,k1=l1-1,l1+1
        DO 6000,jj=1,icell(0,k1,k2,k3)
          jat=icell(jj,k1,k2,k3)
          IF (jat.eq.iat) CYCLE
          xrel= rxyz(1,iat)-rxyz(1,jat)
          yrel= rxyz(2,iat)-rxyz(2,jat)
          zrel= rxyz(3,iat)-rxyz(3,jat)
          rr2=xrel**2 + yrel**2 + zrel**2
          IF ( rr2 .LE. cut2 ) THEN
           indlst=MIN(indlst,myspace-1)
           lstb(indlst)=lay(jat)
!        write(*,*) 'iat,indlst,lay(jat)',iat,indlst,lay(jat)
           tt=SQRT(rr2)
           tti=1.e0_dp/tt
           rel(1,indlst)=xrel*tti
           rel(2,indlst)=yrel*tti
           rel(3,indlst)=zrel*tti
           rel(4,indlst)=tt
           rel(5,indlst)=tti
           indlst= indlst+1
          ENDIF
6000        CONTINUE
6001        CONTINUE
6002        CONTINUE
6003        CONTINUE

        RETURN
        END SUBROUTINE sublstiat_b

! **************************************************************************************************
!> \brief Lenosky's "highly optimized empirical potential model of silicon"
!>      by Stefan Goedecker
!> \param nat number of atoms
!> \param alat lattice constants of the orthorombic box containing the particles
!> \param rxyz0 atomic positions in Angstrom, may be modified on output.
!>               If an atom is outside the box the program will bring it back
!>               into the box by translations through alat
!> \param fxyz forces in eV/A
!> \param ener total energy in eV
!> \param coord average coordination number
!> \param ener_var variance of the energy/atom
!> \param coord_var variance of the coordination number
!> \param count count is increased by one per call, has to be initialized
!>                to 0.e0_dp before first call of eip_bazant
!> \par Literature
!>      T. Lenosky, et. al.: Highly optimized empirical potential model of silicon;
!>                           Modelling Simul. Sci. Eng., 8 (2000)
!>      S. Goedecker: Optimization and parallelization of a force field for silicon
!>                    using OpenMP; CPC 148, 1 (2002)
!> \par History
!>      03.2006 initial create [tdk]
!> \author Thomas D. Kuehne (tkuehne@phys.chem.ethz.ch)
! **************************************************************************************************
    SUBROUTINE eip_lenosky_silicon(nat,alat,rxyz0,fxyz,ener,coord,ener_var,&
                                   coord_var,count)

      INTEGER                                            :: nat
      REAL(KIND=dp)                                      :: alat, rxyz0, fxyz, ener, coord, &
                                                            ener_var, coord_var, count

        DIMENSION rxyz0(3,nat),fxyz(3,nat),alat(3)
        REAL(KIND=dp), ALLOCATABLE, DIMENSION(:,:) :: rxyz
        INTEGER, ALLOCATABLE, DIMENSION(:,:)       :: lsta
        INTEGER, ALLOCATABLE, DIMENSION(:)         :: lstb
        INTEGER, ALLOCATABLE, DIMENSION(:)         :: lay
        INTEGER, ALLOCATABLE, DIMENSION(:,:,:,:)   :: icell
        REAL(KIND=dp), ALLOCATABLE, DIMENSION(:,:) :: rel
        REAL(KIND=dp), ALLOCATABLE, DIMENSION(:,:) :: txyz
        REAL(KIND=dp), ALLOCATABLE, DIMENSION(:,:) :: f2ij,f3ij,f3ik

        REAL(KIND=dp) :: coord2, cut, cut2, ener2, tcoord, &
                         tcoord2, tener, tener2
        INTEGER       :: i, iam, iat, iat1, iat2, ii, il, in, indlst, indlstx,&
                         istop, istopg, l1, l2, l3, ll1, ll2, ll3, lot, ncx, nn, &
                         nnbrx, npjkx, npjx, laymx, npr, rlc1i, rlc2i, rlc3i, &
                         myspace, myspaceout

!        tmax_phi= 0.4500000e+01_dp
!        cut=tmax_phi
        cut= 0.4500000e+01_dp

        IF (count.eq.0)  OPEN(unit=10,file='lenosky.mon',status='unknown')
        count=count+1.e0_dp

! linear scaling calculation of verlet list
        ll1=INT(alat(1)/cut)
        IF (ll1.lt.1) CALL cp__b("eip_silicon.F",1823,"alat(1) too small")
        ll2=INT(alat(2)/cut)
        IF (ll2.lt.1) CALL cp__b("eip_silicon.F",1825,"alat(2) too small")
        ll3=INT(alat(3)/cut)
        IF (ll3.lt.1) CALL cp__b("eip_silicon.F",1827,"alat(3) too small")

! determine number of threads
        npr=1
!$OMP PARALLEL PRIVATE(iam)  SHARED (npr) DEFAULT(NONE)
!$       iam=omp_get_thread_num()
!$       if (iam.eq.0) npr=omp_get_num_threads()
!$OMP END PARALLEL

! linear scaling calculation of verlet list

     IF (npr.le.1) THEN !serial if too few processors to gain by parallelizing

! set ncx for serial case, ncx for parallel case set below
        ncx=16
        loop_ncx_s: DO
           ALLOCATE(icell(0:ncx,-1:ll1,-1:ll2,-1:ll3))
           icell(0,-1:ll1,-1:ll2,-1:ll3)=0
           rlc1i=INT(ll1/alat(1))
           rlc2i=INT(ll2/alat(2))
           rlc3i=INT(ll3/alat(3))

           loop_iat_s: DO iat=1,nat
              rxyz0(1,iat)=MODULO(MODULO(rxyz0(1,iat),alat(1)),alat(1))
              rxyz0(2,iat)=MODULO(MODULO(rxyz0(2,iat),alat(2)),alat(2))
              rxyz0(3,iat)=MODULO(MODULO(rxyz0(3,iat),alat(3)),alat(3))
              l1=INT(rxyz0(1,iat)*rlc1i)
              l2=INT(rxyz0(2,iat)*rlc2i)
              l3=INT(rxyz0(3,iat)*rlc3i)

              ii=icell(0,l1,l2,l3)
              ii=ii+1
              icell(0,l1,l2,l3)=ii
              IF (ii.gt.ncx) THEN
                 WRITE(10,*) count,'NCX too small',ncx
                 DEALLOCATE(icell)
                 ncx=ncx*2
                 CYCLE loop_ncx_s
              END IF
              icell(ii,l1,l2,l3)=iat
           END DO loop_iat_s
           EXIT loop_ncx_s
        END DO loop_ncx_s

     ELSE  ! parallel case

! periodization of particles can be done in parallel
!$OMP PARALLEL DO SHARED (alat,nat,rxyz0) PRIVATE(iat) DEFAULT(NONE)
        DO iat=1,nat
           rxyz0(1,iat)=MODULO(MODULO(rxyz0(1,iat),alat(1)),alat(1))
           rxyz0(2,iat)=MODULO(MODULO(rxyz0(2,iat),alat(2)),alat(2))
           rxyz0(3,iat)=MODULO(MODULO(rxyz0(3,iat),alat(3)),alat(3))
        END DO
!$OMP END PARALLEL DO

! assignment to cell is done serially
! set ncx for parallel case, ncx for serial case set above
        ncx=16
        loop_ncx_p: DO
           ALLOCATE(icell(0:ncx,-1:ll1,-1:ll2,-1:ll3))
           icell(0,-1:ll1,-1:ll2,-1:ll3)=0
           rlc1i=INT(ll1/alat(1))
           rlc2i=INT(ll2/alat(2))
           rlc3i=INT(ll3/alat(3))

           loop_iat_p: DO iat=1,nat
              l1=INT(rxyz0(1,iat)*rlc1i)
              l2=INT(rxyz0(2,iat)*rlc2i)
              l3=INT(rxyz0(3,iat)*rlc3i)
              ii=icell(0,l1,l2,l3)
              ii=ii+1
              icell(0,l1,l2,l3)=ii
              IF (ii.gt.ncx) THEN
                 WRITE(10,*) count,'NCX too small',ncx
                 DEALLOCATE(icell)
                 ncx=ncx*2
                 CYCLE loop_ncx_p
              END IF
              icell(ii,l1,l2,l3)=iat
           END DO loop_iat_p
           EXIT loop_ncx_p
        END DO loop_ncx_p

    ENDIF

! duplicate all atoms within boundary layer
        laymx=ncx*(2*ll1*ll2+2*ll1*ll3+2*ll2*ll3+4*ll1+4*ll2+4*ll3+8)
        nn=nat+laymx
        ALLOCATE(rxyz(3,nn),lay(nn))
        DO  iat=1,nat
        lay(iat)=iat
        rxyz(1,iat)=rxyz0(1,iat)
        rxyz(2,iat)=rxyz0(2,iat)
        rxyz(3,iat)=rxyz0(3,iat)
        ENDDO
        il=nat
! xy plane
        DO l2=0,ll2-1
        DO l1=0,ll1-1

        in=icell(0,l1,l2,0)
        icell(0,l1,l2,ll3)=in
        DO ii=1,in
        i=icell(ii,l1,l2,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1932,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,l2,ll3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,l1,l2,ll3-1)
        icell(0,l1,l2,-1)=in
        DO ii=1,in
        i=icell(ii,l1,l2,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1945,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,l2,-1)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        ENDDO
        ENDDO

! yz plane
        DO l3=0,ll3-1
        DO l2=0,ll2-1

        in=icell(0,0,l2,l3)
        icell(0,ll1,l2,l3)=in
        DO ii=1,in
        i=icell(ii,0,l2,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1965,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,l2,l3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,ll1-1,l2,l3)
        icell(0,-1,l2,l3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,l2,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1978,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,l2,l3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        ENDDO
        ENDDO

! xz plane
        DO l3=0,ll3-1
        DO l1=0,ll1-1

        in=icell(0,l1,0,l3)
        icell(0,l1,ll2,l3)=in
        DO ii=1,in
        i=icell(ii,l1,0,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",1998,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,ll2,l3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,l1,ll2-1,l3)
        icell(0,l1,-1,l3)=in
        DO ii=1,in
        i=icell(ii,l1,ll2-1,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2011,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,-1,l3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        ENDDO
        ENDDO

! x axis
        DO l1=0,ll1-1

        in=icell(0,l1,0,0)
        icell(0,l1,ll2,ll3)=in
        DO ii=1,in
        i=icell(ii,l1,0,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2030,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,ll2,ll3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,l1,0,ll3-1)
        icell(0,l1,ll2,-1)=in
        DO ii=1,in
        i=icell(ii,l1,0,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2043,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,ll2,-1)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,l1,ll2-1,0)
        icell(0,l1,-1,ll3)=in
        DO ii=1,in
        i=icell(ii,l1,ll2-1,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2056,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,-1,ll3)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,l1,ll2-1,ll3-1)
        icell(0,l1,-1,-1)=in
        DO ii=1,in
        i=icell(ii,l1,ll2-1,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2069,"enlarge laymx")
        lay(il)=i
        icell(ii,l1,-1,-1)=il
        rxyz(1,il)=rxyz(1,i)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        ENDDO

! y axis
        DO l2=0,ll2-1

        in=icell(0,0,l2,0)
        icell(0,ll1,l2,ll3)=in
        DO ii=1,in
        i=icell(ii,0,l2,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2087,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,l2,ll3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,0,l2,ll3-1)
        icell(0,ll1,l2,-1)=in
        DO ii=1,in
        i=icell(ii,0,l2,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2100,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,l2,-1)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,ll1-1,l2,0)
        icell(0,-1,l2,ll3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,l2,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2113,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,l2,ll3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,ll1-1,l2,ll3-1)
        icell(0,-1,l2,-1)=in
        DO ii=1,in
        i=icell(ii,ll1-1,l2,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2126,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,l2,-1)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        ENDDO

! z axis
        DO l3=0,ll3-1

        in=icell(0,0,0,l3)
        icell(0,ll1,ll2,l3)=in
        DO ii=1,in
        i=icell(ii,0,0,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2144,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,ll2,l3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,ll1-1,0,l3)
        icell(0,-1,ll2,l3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,0,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2157,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,ll2,l3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,0,ll2-1,l3)
        icell(0,ll1,-1,l3)=in
        DO ii=1,in
        i=icell(ii,0,ll2-1,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2170,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,-1,l3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        in=icell(0,ll1-1,ll2-1,l3)
        icell(0,-1,-1,l3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,ll2-1,l3)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2183,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,-1,l3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)
        ENDDO

        ENDDO

! corners
        in=icell(0,0,0,0)
        icell(0,ll1,ll2,ll3)=in
        DO ii=1,in
        i=icell(ii,0,0,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2199,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,ll2,ll3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,ll1-1,0,0)
        icell(0,-1,ll2,ll3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,0,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2212,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,ll2,ll3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,0,ll2-1,0)
        icell(0,ll1,-1,ll3)=in
        DO ii=1,in
        i=icell(ii,0,ll2-1,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2225,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,-1,ll3)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,ll1-1,ll2-1,0)
        icell(0,-1,-1,ll3)=in
        DO ii=1,in
        i=icell(ii,ll1-1,ll2-1,0)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2238,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,-1,ll3)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)+alat(3)
        ENDDO

        in=icell(0,0,0,ll3-1)
        icell(0,ll1,ll2,-1)=in
        DO ii=1,in
        i=icell(ii,0,0,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2251,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,ll2,-1)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,ll1-1,0,ll3-1)
        icell(0,-1,ll2,-1)=in
        DO ii=1,in
        i=icell(ii,ll1-1,0,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2264,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,ll2,-1)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)+alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,0,ll2-1,ll3-1)
        icell(0,ll1,-1,-1)=in
        DO ii=1,in
        i=icell(ii,0,ll2-1,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2277,"enlarge laymx")
        lay(il)=i
        icell(ii,ll1,-1,-1)=il
        rxyz(1,il)=rxyz(1,i)+alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        in=icell(0,ll1-1,ll2-1,ll3-1)
        icell(0,-1,-1,-1)=in
        DO ii=1,in
        i=icell(ii,ll1-1,ll2-1,ll3-1)
        il=il+1
        IF (il.gt.nn) CALL cp__b("eip_silicon.F",2290,"enlarge laymx")
        lay(il)=i
        icell(ii,-1,-1,-1)=il
        rxyz(1,il)=rxyz(1,i)-alat(1)
        rxyz(2,il)=rxyz(2,i)-alat(2)
        rxyz(3,il)=rxyz(3,i)-alat(3)
        ENDDO

        ALLOCATE(lsta(2,nat))
        nnbrx=36
        loop_nnbrx: DO
           ALLOCATE(lstb(nnbrx*nat),rel(5,nnbrx*nat))

           indlstx=0

!$OMP PARALLEL DEFAULT(NONE)  &
!$OMP PRIVATE(iat,cut2,iam,ii,indlst,l1,l2,l3,myspace,npr) &
!$OMP SHARED (indlstx,nat,nn,nnbrx,ncx,ll1,ll2,ll3,icell,lsta,lstb,lay, &
!$OMP rel,rxyz,cut,myspaceout)

           npr=1
!$         npr=omp_get_num_threads()
           iam=0
!$         iam=omp_get_thread_num()

           cut2=cut**2
! assign contiguous portions of the arrays lstb and rel to the threads
           myspace=(nat*nnbrx)/npr
           IF (iam.eq.0) myspaceout=myspace
! Verlet list, relative positions
           indlst=0
           loop_l3: DO l3=0,ll3-1
              loop_l2: DO l2=0,ll2-1
                 loop_l1: DO l1=0,ll1-1
                    loop_ii: DO ii=1,icell(0,l1,l2,l3)
                       iat=icell(ii,l1,l2,l3)
                       IF ( ((iat-1)*npr)/nat .EQ. iam) THEN
!                          write(*,*) 'sublstiat:iam,iat',iam,iat
                          lsta(1,iat)=iam*myspace+indlst+1
                          CALL sublstiat_l(iat,nn,ncx,ll1,ll2,ll3,l1,l2,l3,myspace, &
                                           rxyz,icell,lstb(iam*myspace+1),lay, &
                                           rel(1,iam*myspace+1),cut2,indlst)
                                           lsta(2,iat)=iam*myspace+indlst
!                          write(*,'(a,4(x,i3),100(x,i2))') &
!                                'iam,iat,lsta',iam,iat,lsta(1,iat),lsta(2,iat), &
!                                (lstb(j),j=lsta(1,iat),lsta(2,iat))
                       END IF
                    END DO loop_ii
                 END DO loop_l1
              END DO loop_l2
           END DO loop_l3

!$OMP CRITICAL
           indlstx=MAX(indlstx,indlst)
!$OMP END CRITICAL
!$OMP END PARALLEL

           IF (indlstx.ge.myspaceout) THEN
               WRITE(10,*) count,'NNBRX too  small', nnbrx
               DEALLOCATE(lstb,rel)
               nnbrx=3*nnbrx/2
               CYCLE loop_nnbrx
           END IF
           EXIT loop_nnbrx
        END DO loop_nnbrx

        istopg=0
!$OMP PARALLEL DEFAULT(NONE)  &
!$OMP PRIVATE(iam,npr,iat,iat1,iat2,lot,istop,tcoord,tcoord2, &
!$OMP tener,tener2,txyz,f2ij,f3ij,f3ik,npjx,npjkx) &
!$OMP SHARED (nat,nnbrx,lsta,lstb,rel,ener,ener2,fxyz,coord,coord2,istopg)

        npr=1
!$       npr=omp_get_num_threads()
        iam=0
!$       iam=omp_get_thread_num()

        npjx=300 ; npjkx=6000

        IF (npr.ne.1) THEN
! PARALLEL CASE
! create temporary private scalars for reduction sum on energies and
!        temporary private array for reduction sum on forces
!$OMP CRITICAL
        ALLOCATE(txyz(3,nat),f2ij(3,npjx),f3ij(3,npjkx),f3ik(3,npjkx))
!$OMP END CRITICAL
        IF (iam.eq.0) THEN
        ener=0.e0_dp
        ener2=0.e0_dp
        coord=0.e0_dp
        coord2=0.e0_dp
        ENDIF
!$OMP DO
        DO 121,iat=1,nat
        fxyz(1,iat)=0.e0_dp
        fxyz(2,iat)=0.e0_dp
        fxyz(3,iat)=0.e0_dp
121     CONTINUE
!$OMP BARRIER

! Each thread treats at most lot atoms
        lot=INT(float(nat)/float(npr)+.999999999999e0_dp)
        iat1=iam*lot+1
        iat2=MIN((iam+1)*lot,nat)
!       write(*,*) 'subfeniat:iat1,iat2,iam',iat1,iat2,iam
        CALL subfeniat_l(iat1,iat2,nat,lsta,lstb,rel,tener,tener2, &
               tcoord,tcoord2,nnbrx,txyz,f2ij,npjx,f3ij,npjkx,f3ik,istop)
!$OMP CRITICAL
        ener=ener+tener
        ener2=ener2+tener2
        coord=coord+tcoord
        coord2=coord2+tcoord2
        istopg=istopg+istop
        DO 8093,iat=1,nat
        fxyz(1,iat)=fxyz(1,iat)+txyz(1,iat)
        fxyz(2,iat)=fxyz(2,iat)+txyz(2,iat)
        fxyz(3,iat)=fxyz(3,iat)+txyz(3,iat)
8093    CONTINUE
        DEALLOCATE(txyz,f2ij,f3ij,f3ik)
!$OMP END CRITICAL

        ELSE
! SERIAL CASE
        iat1=1
        iat2=nat
        ALLOCATE(f2ij(3,npjx),f3ij(3,npjkx),f3ik(3,npjkx))
        CALL subfeniat_l(iat1,iat2,nat,lsta,lstb,rel,ener,ener2, &
               coord,coord2,nnbrx,fxyz,f2ij,npjx,f3ij,npjkx,f3ik,istopg)
        DEALLOCATE(f2ij,f3ij,f3ik)

        ENDIF
!$OMP END PARALLEL

!         write(*,*) 'ener,norm force', &
!                    ener,DNRM2(3*nat,fxyz,1)
        IF (istopg.gt.0) CALL cp__b("eip_silicon.F",2425,"DIMENSION ERROR (see WARNING above)")
        ener_var=ener2/nat-(ener/nat)**2
        coord=coord/nat
        coord_var=coord2/nat-coord**2

        DEALLOCATE(rxyz,icell,lay,lsta,lstb,rel)

        END SUBROUTINE eip_lenosky_silicon

! **************************************************************************************************
!> \brief ...
!> \param iat1 ...
!> \param iat2 ...
!> \param nat ...
!> \param lsta ...
!> \param lstb ...
!> \param rel ...
!> \param tener ...
!> \param tener2 ...
!> \param tcoord ...
!> \param tcoord2 ...
!> \param nnbrx ...
!> \param txyz ...
!> \param f2ij ...
!> \param npjx ...
!> \param f3ij ...
!> \param npjkx ...
!> \param f3ik ...
!> \param istop ...
! **************************************************************************************************
        SUBROUTINE subfeniat_l(iat1,iat2,nat,lsta,lstb,rel,tener,tener2, &
               tcoord,tcoord2,nnbrx,txyz,f2ij,npjx,f3ij,npjkx,f3ik,istop)
! for a subset of atoms iat1 to iat2 the routine calculates the (partial) forces
! txyz acting on these atoms as well as on the atoms (jat, kat) interacting
! with them and their contribution to the energy (tener).
! In addition the coordination number tcoord and the second moment of the
! local energy tener2 and coordination number tcoord2 are returned
      INTEGER                                            :: iat1, iat2, nat, lsta, lstb
      REAL(KIND=dp)                                      :: rel, tener, tener2, tcoord, tcoord2
      INTEGER                                            :: nnbrx
      REAL(KIND=dp)                                      :: txyz, f2ij
      INTEGER                                            :: npjx
      REAL(KIND=dp)                                      :: f3ij
      INTEGER                                            :: npjkx
      REAL(KIND=dp)                                      :: f3ik
      INTEGER                                            :: istop

        DIMENSION lsta(2,nat),lstb(nnbrx*nat),rel(5,nnbrx*nat),txyz(3,nat)
        DIMENSION f2ij(3,npjx),f3ij(3,npjkx),f3ik(3,npjkx)
        REAL(KIND=dp), PARAMETER :: tmin_phi= 0.1500000e+01_dp
        REAL(KIND=dp), PARAMETER :: tmax_phi= 0.4500000e+01_dp
        REAL(KIND=dp), PARAMETER :: hi_phi= 3.00000000000e0_dp
        REAL(KIND=dp), PARAMETER :: hsixth_phi=5.55555555555556e-002_dp
        REAL(KIND=dp), PARAMETER :: h2sixth_phi=1.85185185185185e-002_dp
        REAL(KIND=dp), PARAMETER, DIMENSION(0:9) :: cof_phi  = &
                     (/ 0.69299400000000e+01_dp, -0.43995000000000e+00_dp, &
                       -0.17012300000000e+01_dp, -0.16247300000000e+01_dp, &
                       -0.99696000000000e+00_dp, -0.27391000000000e+00_dp, &
                       -0.24990000000000e-01_dp, -0.17840000000000e-01_dp, &
                       -0.96100000000000e-02_dp,  0.00000000000000e+00_dp /)
        REAL(KIND=dp), PARAMETER, DIMENSION(0:9) :: dof_phi  = &
                     (/ 0.16533229480429e+03_dp,  0.39415410391417e+02_dp, &
                        0.68710036300407e+01_dp,  0.53406950884203e+01_dp, &
                        0.15347960162782e+01_dp, -0.63347591535331e+01_dp, &
                       -0.17987794021458e+01_dp,  0.47429676211617e+00_dp, &
                       -0.40087646318907e-01_dp, -0.23942617684055e+00_dp /)
        REAL(KIND=dp), PARAMETER :: tmin_rho= 0.1500000e+01_dp
        REAL(KIND=dp), PARAMETER :: tmax_rho= 0.3500000e+01_dp
        REAL(KIND=dp), PARAMETER :: hi_rho= 5.00000000000e0_dp
        REAL(KIND=dp), PARAMETER :: hsixth_rho=3.33333333333333e-002_dp
        REAL(KIND=dp), PARAMETER :: h2sixth_rho=6.66666666666667e-003_dp
        REAL(KIND=dp), PARAMETER, DIMENSION(0:10) :: cof_rho =  &
                     (/ 0.13747000000000e+00_dp, -0.14831000000000e+00_dp, &
                       -0.55972000000000e+00_dp, -0.73110000000000e+00_dp, &
                       -0.76283000000000e+00_dp, -0.72918000000000e+00_dp, &
                       -0.66620000000000e+00_dp, -0.57328000000000e+00_dp, &
                       -0.40690000000000e+00_dp, -0.16662000000000e+00_dp, &
                        0.00000000000000e+00_dp /)
        REAL(KIND=dp), PARAMETER, DIMENSION(0:10) :: dof_rho =  &
                    (/ -0.32275496741918e+01_dp, -0.64119006516165e+01_dp, &
                        0.10030652280658e+02_dp,  0.22937915289857e+01_dp, &
                        0.17416816033995e+01_dp,  0.54648205741626e+00_dp, &
                        0.47189016693543e+00_dp,  0.20569572748420e+01_dp, &
                        0.23192807336964e+01_dp, -0.24908020962757e+00_dp, &
                       -0.12371959895186e+02_dp /)
        REAL(KIND=dp), PARAMETER :: tmin_fff= 0.1500000e+01_dp
        REAL(KIND=dp), PARAMETER :: tmax_fff= 0.3500000e+01_dp
        REAL(KIND=dp), PARAMETER :: hi_fff= 4.50000000000e0_dp
        REAL(KIND=dp), PARAMETER :: hsixth_fff=3.70370370370370e-002_dp
        REAL(KIND=dp), PARAMETER :: h2sixth_fff=8.23045267489712e-003_dp
        REAL(KIND=dp), PARAMETER, DIMENSION(0:9) :: cof_fff  = &
                     (/ 0.12503100000000e+01_dp,  0.86821000000000e+00_dp, &
                        0.60846000000000e+00_dp,  0.48756000000000e+00_dp, &
                        0.44163000000000e+00_dp,  0.37610000000000e+00_dp, &
                        0.27145000000000e+00_dp,  0.14814000000000e+00_dp, &
                        0.48550000000000e-01_dp,  0.00000000000000e+00_dp /)
        REAL(KIND=dp), PARAMETER, DIMENSION(0:9) :: dof_fff  = &
                     (/ 0.27904652711432e+02_dp, -0.45230754228635e+01_dp, &
                        0.50531739800222e+01_dp,  0.11806545027747e+01_dp, &
                       -0.66693699112098e+00_dp, -0.89430653829079e+00_dp, &
                       -0.50891685571587e+00_dp,  0.66278396115427e+00_dp, &
                        0.73976101109878e+00_dp,  0.25795319944506e+01_dp /)
        REAL(KIND=dp), PARAMETER :: tmin_uuu= -0.1770930e+01_dp
        REAL(KIND=dp), PARAMETER :: tmax_uuu= 0.7908520e+01_dp
        REAL(KIND=dp), PARAMETER :: hi_uuu= 0.723181585730594e0_dp
        REAL(KIND=dp), PARAMETER :: hsixth_uuu=0.230463095238095e0_dp
        REAL(KIND=dp), PARAMETER :: h2sixth_uuu=0.318679429600340e0_dp
        REAL(KIND=dp), PARAMETER, DIMENSION(0:7) :: cof_uuu  = &
                    (/ -0.10749300000000e+01_dp, -0.20045000000000e+00_dp, &
                        0.41422000000000e+00_dp,  0.87939000000000e+00_dp, &
                        0.12668900000000e+01_dp,  0.16299800000000e+01_dp, &
                        0.19773800000000e+01_dp,  0.23961800000000e+01_dp /)
        REAL(KIND=dp), PARAMETER, DIMENSION(0:7) :: dof_uuu  = &
                    (/ -0.14827125747284e+00_dp, -0.14922155328475e+00_dp, &
                       -0.70113224223509e-01_dp, -0.39449020349230e-01_dp, &
                       -0.15815242579643e-01_dp,  0.26112640061855e-01_dp, &
                       -0.13786974745095e+00_dp,  0.74941595372657e+00_dp /)
        REAL(KIND=dp), PARAMETER :: tmin_ggg= -0.1000000e+01_dp
        REAL(KIND=dp), PARAMETER :: tmax_ggg= 0.8001400e+00_dp
        REAL(KIND=dp), PARAMETER :: hi_ggg= 3.88858644327663e0_dp
        REAL(KIND=dp), PARAMETER :: hsixth_ggg=4.28604761904762e-002_dp
        REAL(KIND=dp), PARAMETER :: h2sixth_ggg=1.10221225156463e-002_dp
        REAL(KIND=dp), PARAMETER, DIMENSION(0:7) :: cof_ggg  = &
                     (/ 0.52541600000000e+01_dp,  0.23591500000000e+01_dp, &
                        0.11959500000000e+01_dp,  0.12299500000000e+01_dp, &
                        0.20356500000000e+01_dp,  0.34247400000000e+01_dp, &
                        0.49485900000000e+01_dp,  0.56179900000000e+01_dp /)
        REAL(KIND=dp), PARAMETER, DIMENSION(0:7) :: dof_ggg  = &
                     (/ 0.15826876132396e+02_dp,  0.31176239377907e+02_dp, &
                        0.16589446539683e+02_dp,  0.11083892500520e+02_dp, &
                        0.90887216383860e+01_dp,  0.54902279653967e+01_dp, &
                       -0.18823313223755e+02_dp, -0.77183416481005e+01_dp /)

       REAL(KIND=dp) :: a2_fff, a2_ggg, a_fff, a_ggg, b2_fff, b2_ggg, b_fff,&
                        b_ggg, cof1_fff, cof1_ggg, cof2_fff, cof2_ggg,cof3_fff,&
                        cof3_ggg, cof4_fff, cof4_ggg, cof_fff_khi, cof_fff_klo, &
                        cof_ggg_khi, cof_ggg_klo, coord_iat, costheta, dens, &
                        dens2, dens3, dof_fff_khi, dof_fff_klo, dof_ggg_khi, &
                        dof_ggg_klo, e_phi, e_uuu, ener_iat, ep_phi, ep_uuu, &
                        fij, fijp, fik, fikp, fxij, fxik, fyij, fyik, fzij, fzik,&
                        gjik, gjikp, rho, rhop, rij, rik, sij, sik, t1, t2, t3, t4, &
                        tt, tt_fff, tt_ggg, xarg, ypt1_fff, ypt1_ggg, ypt2_fff,&
                        ypt2_ggg, yt1_fff, yt1_ggg, yt2_fff, yt2_ggg

       INTEGER       :: iat, jat, jbr, jcnt, jkcnt, kat, kbr, khi_fff, khi_ggg,&
                        klo_fff, klo_ggg


! initialize temporary private scalars for reduction sum on energies and
! private workarray txyz for forces forces
        tener=0.e0_dp
        tener2=0.e0_dp
        tcoord=0.e0_dp
        tcoord2=0.e0_dp
        istop=0
        DO 121,iat=1,nat
        txyz(1,iat)=0.e0_dp
        txyz(2,iat)=0.e0_dp
        txyz(3,iat)=0.e0_dp
121     CONTINUE

! calculation of forces, energy

        DO 1000,iat=iat1,iat2

        dens2=0.e0_dp
        dens3=0.e0_dp
        jcnt=0
        jkcnt=0
        coord_iat=0.e0_dp
        ener_iat=0.e0_dp
        DO 2000,jbr=lsta(1,iat),lsta(2,iat)
        jat=lstb(jbr)
        jcnt=jcnt+1
        IF (jcnt.gt.npjx) THEN
            WRITE(*,*) 'WARNING: enlarge npjx'
            istop=1
            RETURN
        ENDIF

        fxij=rel(1,jbr)
        fyij=rel(2,jbr)
        fzij=rel(3,jbr)
        rij=rel(4,jbr)
        sij=rel(5,jbr)

! coordination number calculated with soft cutoff between first
! nearest neighbor and midpoint of first and second nearest neighbor
        IF (rij.le.2.36e0_dp) THEN
        coord_iat=coord_iat+1.e0_dp
        ELSE IF (rij.ge.3.12e0_dp) THEN
        ELSE
        xarg=(rij-2.36e0_dp)*(1.e0_dp/(3.12e0_dp-2.36e0_dp))
        coord_iat=coord_iat+(2*xarg+1.e0_dp)*(xarg-1.e0_dp)**2
        ENDIF

! pairpotential term
        CALL splint(cof_phi,dof_phi,tmin_phi,tmax_phi, &
           hsixth_phi,h2sixth_phi,hi_phi,10,rij,e_phi,ep_phi)
        ener_iat=ener_iat+(e_phi*.5e0_dp)
        txyz(1,iat)=txyz(1,iat)-fxij*(ep_phi*.5e0_dp)
        txyz(2,iat)=txyz(2,iat)-fyij*(ep_phi*.5e0_dp)
        txyz(3,iat)=txyz(3,iat)-fzij*(ep_phi*.5e0_dp)
        txyz(1,jat)=txyz(1,jat)+fxij*(ep_phi*.5e0_dp)
        txyz(2,jat)=txyz(2,jat)+fyij*(ep_phi*.5e0_dp)
        txyz(3,jat)=txyz(3,jat)+fzij*(ep_phi*.5e0_dp)

! 2 body embedding term
        CALL splint(cof_rho,dof_rho,tmin_rho,tmax_rho, &
             hsixth_rho,h2sixth_rho,hi_rho,11,rij,rho,rhop)
        dens2=dens2+rho
        f2ij(1,jcnt)=fxij*rhop
        f2ij(2,jcnt)=fyij*rhop
        f2ij(3,jcnt)=fzij*rhop

! 3 body embedding term
        CALL splint(cof_fff,dof_fff,tmin_fff,tmax_fff, &
             hsixth_fff,h2sixth_fff,hi_fff,10,rij,fij,fijp)

        DO 3000,kbr=lsta(1,iat),lsta(2,iat)
        kat=lstb(kbr)
        IF (kat.lt.jat) THEN
        jkcnt=jkcnt+1
        IF (jkcnt.gt.npjkx) THEN
            WRITE(*,*) 'WARNING: enlarge npjkx',npjkx
            istop=1
            RETURN
        ENDIF

! begin unoptimized original version:
!        fxik=rel(1,kbr)
!        fyik=rel(2,kbr)
!        fzik=rel(3,kbr)
!        rik=rel(4,kbr)
!        sik=rel(5,kbr)
!
!        call splint(cof_fff,dof_fff,tmin_fff,tmax_fff, &
!             hsixth_fff,h2sixth_fff,hi_fff,10,rik,fik,fikp)
!        costheta=fxij*fxik+fyij*fyik+fzij*fzik
!        call splint(cof_ggg,dof_ggg,tmin_ggg,tmax_ggg, &
!             hsixth_ggg,h2sixth_ggg,hi_ggg,8,costheta,gjik,gjikp)
! end unoptimized original version:

! begin optimized version
        rik=rel(4,kbr)
      IF (rik.gt.tmax_fff) THEN
        fikp=0.e0_dp ; fik=0.e0_dp
        gjik=0.e0_dp ;  gjikp=0.e0_dp ; sik=0.e0_dp
        costheta=0.e0_dp ; fxik=0.e0_dp ; fyik=0.e0_dp ; fzik=0.e0_dp
      ELSE IF (rik.lt.tmin_fff) THEN
        fxik=rel(1,kbr)
        fyik=rel(2,kbr)
        fzik=rel(3,kbr)
        costheta=fxij*fxik+fyij*fyik+fzij*fzik
        sik=rel(5,kbr)
        fikp=hi_fff*(cof_fff(1)-cof_fff(0)) -  &
             ( dof_fff(1)+2.e0_dp*dof_fff(0) )*hsixth_fff
        fik=cof_fff(0) + (rik-tmin_fff)*fikp
        tt_ggg=(costheta-tmin_ggg)*hi_ggg
        IF (costheta.gt.tmax_ggg) THEN
           gjikp=hi_ggg*(cof_ggg(8-1)-cof_ggg(8-2)) + &
                ( 2.e0_dp*dof_ggg(8-1)+dof_ggg(8-2) )*hsixth_ggg
                 gjik=cof_ggg(8-1) + (costheta-tmax_ggg)*gjikp
        ELSE
           klo_ggg=INT(tt_ggg)
           khi_ggg=klo_ggg+1
           cof_ggg_klo=cof_ggg(klo_ggg)
           dof_ggg_klo=dof_ggg(klo_ggg)
           b_ggg=tt_ggg-klo_ggg
           a_ggg=1.e0_dp-b_ggg
           cof_ggg_khi=cof_ggg(khi_ggg)
           dof_ggg_khi=dof_ggg(khi_ggg)
           b2_ggg=b_ggg*b_ggg
           gjik=a_ggg*cof_ggg_klo
           gjikp=cof_ggg_khi-cof_ggg_klo
           a2_ggg=a_ggg*a_ggg
           cof1_ggg=a2_ggg-1.e0_dp
           cof2_ggg=b2_ggg-1.e0_dp
           gjik=gjik+b_ggg*cof_ggg_khi
           gjikp=hi_ggg*gjikp
           cof3_ggg=3.e0_dp*b2_ggg
           cof4_ggg=3.e0_dp*a2_ggg
           cof1_ggg=a_ggg*cof1_ggg
           cof2_ggg=b_ggg*cof2_ggg
           cof3_ggg=cof3_ggg-1.e0_dp
           cof4_ggg=cof4_ggg-1.e0_dp
           yt1_ggg=cof1_ggg*dof_ggg_klo
           yt2_ggg=cof2_ggg*dof_ggg_khi
           ypt1_ggg=cof3_ggg*dof_ggg_khi
           ypt2_ggg=cof4_ggg*dof_ggg_klo
           gjik=gjik + (yt1_ggg+yt2_ggg)*h2sixth_ggg
           gjikp=gjikp + ( ypt1_ggg - ypt2_ggg )*hsixth_ggg
        ENDIF
      ELSE
        fxik=rel(1,kbr)
        tt_fff=rik-tmin_fff
        costheta=fxij*fxik
        fyik=rel(2,kbr)
        tt_fff=tt_fff*hi_fff
        costheta=costheta+fyij*fyik
        fzik=rel(3,kbr)
        klo_fff=INT(tt_fff)
        costheta=costheta+fzij*fzik
        sik=rel(5,kbr)
        tt_ggg=(costheta-tmin_ggg)*hi_ggg
        IF (costheta.gt.tmax_ggg) THEN
          gjikp=hi_ggg*(cof_ggg(8-1)-cof_ggg(8-2)) + &
                ( 2.e0_dp*dof_ggg(8-1)+dof_ggg(8-2) )*hsixth_ggg
                gjik=cof_ggg(8-1) + (costheta-tmax_ggg)*gjikp
          khi_fff=klo_fff+1
          cof_fff_klo=cof_fff(klo_fff)
          dof_fff_klo=dof_fff(klo_fff)
          b_fff=tt_fff-klo_fff
          a_fff=1.e0_dp-b_fff
          cof_fff_khi=cof_fff(khi_fff)
          dof_fff_khi=dof_fff(khi_fff)
          b2_fff=b_fff*b_fff
          fik=a_fff*cof_fff_klo
          fikp=cof_fff_khi-cof_fff_klo
          a2_fff=a_fff*a_fff
          cof1_fff=a2_fff-1.e0_dp
          cof2_fff=b2_fff-1.e0_dp
          fik=fik+b_fff*cof_fff_khi
          fikp=hi_fff*fikp
          cof3_fff=3.e0_dp*b2_fff
          cof4_fff=3.e0_dp*a2_fff
          cof1_fff=a_fff*cof1_fff
          cof2_fff=b_fff*cof2_fff
          cof3_fff=cof3_fff-1.e0_dp
          cof4_fff=cof4_fff-1.e0_dp
          yt1_fff=cof1_fff*dof_fff_klo
          yt2_fff=cof2_fff*dof_fff_khi
          ypt1_fff=cof3_fff*dof_fff_khi
          ypt2_fff=cof4_fff*dof_fff_klo
          fik=fik + (yt1_fff+yt2_fff)*h2sixth_fff
          fikp=fikp + ( ypt1_fff - ypt2_fff )*hsixth_fff
         ELSE
              klo_ggg=INT(tt_ggg)
              khi_ggg=klo_ggg+1
           khi_fff=klo_fff+1
              cof_ggg_klo=cof_ggg(klo_ggg)
           cof_fff_klo=cof_fff(klo_fff)
              dof_ggg_klo=dof_ggg(klo_ggg)
           dof_fff_klo=dof_fff(klo_fff)
              b_ggg=tt_ggg-klo_ggg
           b_fff=tt_fff-klo_fff
              a_ggg=1.e0_dp-b_ggg
           a_fff=1.e0_dp-b_fff
              cof_ggg_khi=cof_ggg(khi_ggg)
           cof_fff_khi=cof_fff(khi_fff)
              dof_ggg_khi=dof_ggg(khi_ggg)
           dof_fff_khi=dof_fff(khi_fff)
              b2_ggg=b_ggg*b_ggg
           b2_fff=b_fff*b_fff
              gjik=a_ggg*cof_ggg_klo
           fik=a_fff*cof_fff_klo
              gjikp=cof_ggg_khi-cof_ggg_klo
           fikp=cof_fff_khi-cof_fff_klo
              a2_ggg=a_ggg*a_ggg
           a2_fff=a_fff*a_fff
              cof1_ggg=a2_ggg-1.e0_dp
           cof1_fff=a2_fff-1.e0_dp
              cof2_ggg=b2_ggg-1.e0_dp
           cof2_fff=b2_fff-1.e0_dp
              gjik=gjik+b_ggg*cof_ggg_khi
           fik=fik+b_fff*cof_fff_khi
              gjikp=hi_ggg*gjikp
           fikp=hi_fff*fikp
              cof3_ggg=3.e0_dp*b2_ggg
           cof3_fff=3.e0_dp*b2_fff
              cof4_ggg=3.e0_dp*a2_ggg
           cof4_fff=3.e0_dp*a2_fff
              cof1_ggg=a_ggg*cof1_ggg
           cof1_fff=a_fff*cof1_fff
              cof2_ggg=b_ggg*cof2_ggg
           cof2_fff=b_fff*cof2_fff
              cof3_ggg=cof3_ggg-1.e0_dp
           cof3_fff=cof3_fff-1.e0_dp
              cof4_ggg=cof4_ggg-1.e0_dp
           cof4_fff=cof4_fff-1.e0_dp
              yt1_ggg=cof1_ggg*dof_ggg_klo
           yt1_fff=cof1_fff*dof_fff_klo
              yt2_ggg=cof2_ggg*dof_ggg_khi
           yt2_fff=cof2_fff*dof_fff_khi
              ypt1_ggg=cof3_ggg*dof_ggg_khi
           ypt1_fff=cof3_fff*dof_fff_khi
              ypt2_ggg=cof4_ggg*dof_ggg_klo
           ypt2_fff=cof4_fff*dof_fff_klo
              gjik=gjik + (yt1_ggg+yt2_ggg)*h2sixth_ggg
           fik=fik + (yt1_fff+yt2_fff)*h2sixth_fff
              gjikp=gjikp + ( ypt1_ggg - ypt2_ggg )*hsixth_ggg
           fikp=fikp + ( ypt1_fff - ypt2_fff )*hsixth_fff
         ENDIF
      ENDIF
! end optimized version

        tt=fij*fik
        dens3=dens3+tt*gjik

        t1=fijp*fik*gjik
        t2=sij*(tt*gjikp)
        f3ij(1,jkcnt)=fxij*t1 + (fxik-fxij*costheta)*t2
        f3ij(2,jkcnt)=fyij*t1 + (fyik-fyij*costheta)*t2
        f3ij(3,jkcnt)=fzij*t1 + (fzik-fzij*costheta)*t2

        t3=fikp*fij*gjik
        t4=sik*(tt*gjikp)
        f3ik(1,jkcnt)=fxik*t3 + (fxij-fxik*costheta)*t4
        f3ik(2,jkcnt)=fyik*t3 + (fyij-fyik*costheta)*t4
        f3ik(3,jkcnt)=fzik*t3 + (fzij-fzik*costheta)*t4
        ENDIF
3000        CONTINUE
2000        CONTINUE
        dens=dens2+dens3
        CALL splint(cof_uuu,dof_uuu,tmin_uuu,tmax_uuu, &
             hsixth_uuu,h2sixth_uuu,hi_uuu,8,dens,e_uuu,ep_uuu)
        ener_iat=ener_iat+e_uuu

! Only now ep_uu is known and the forces can be calculated, lets loop again
        jcnt=0
        jkcnt=0
        DO 2200,jbr=lsta(1,iat),lsta(2,iat)
        jat=lstb(jbr)
        jcnt=jcnt+1
        txyz(1,iat)=txyz(1,iat)-ep_uuu*f2ij(1,jcnt)
        txyz(2,iat)=txyz(2,iat)-ep_uuu*f2ij(2,jcnt)
        txyz(3,iat)=txyz(3,iat)-ep_uuu*f2ij(3,jcnt)
        txyz(1,jat)=txyz(1,jat)+ep_uuu*f2ij(1,jcnt)
        txyz(2,jat)=txyz(2,jat)+ep_uuu*f2ij(2,jcnt)
        txyz(3,jat)=txyz(3,jat)+ep_uuu*f2ij(3,jcnt)

! 3 body embedding term
        DO 3300,kbr=lsta(1,iat),lsta(2,iat)
        kat=lstb(kbr)
        IF (kat.lt.jat) THEN
        jkcnt=jkcnt+1

        txyz(1,iat)=txyz(1,iat)-ep_uuu*(f3ij(1,jkcnt)+f3ik(1,jkcnt))
        txyz(2,iat)=txyz(2,iat)-ep_uuu*(f3ij(2,jkcnt)+f3ik(2,jkcnt))
        txyz(3,iat)=txyz(3,iat)-ep_uuu*(f3ij(3,jkcnt)+f3ik(3,jkcnt))
        txyz(1,jat)=txyz(1,jat)+ep_uuu*f3ij(1,jkcnt)
        txyz(2,jat)=txyz(2,jat)+ep_uuu*f3ij(2,jkcnt)
        txyz(3,jat)=txyz(3,jat)+ep_uuu*f3ij(3,jkcnt)
        txyz(1,kat)=txyz(1,kat)+ep_uuu*f3ik(1,jkcnt)
        txyz(2,kat)=txyz(2,kat)+ep_uuu*f3ik(2,jkcnt)
        txyz(3,kat)=txyz(3,kat)+ep_uuu*f3ik(3,jkcnt)

        ENDIF
3300        CONTINUE
2200        CONTINUE

!        write(*,'(a,i4,x,e19.12,x,e10.3)') 'iat,ener_iat,coord_iat', &
!                                       iat,ener_iat,coord_iat
        tener=tener+ener_iat
        tener2=tener2+ener_iat**2
        tcoord=tcoord+coord_iat
        tcoord2=tcoord2+coord_iat**2

1000        CONTINUE

        END SUBROUTINE subfeniat_l

! **************************************************************************************************
!> \brief ...
!> \param iat ...
!> \param nn ...
!> \param ncx ...
!> \param ll1 ...
!> \param ll2 ...
!> \param ll3 ...
!> \param l1 ...
!> \param l2 ...
!> \param l3 ...
!> \param myspace ...
!> \param rxyz ...
!> \param icell ...
!> \param lstb ...
!> \param lay ...
!> \param rel ...
!> \param cut2 ...
!> \param indlst ...
! **************************************************************************************************
        SUBROUTINE sublstiat_l(iat,nn,ncx,ll1,ll2,ll3,l1,l2,l3,myspace, &
                   rxyz,icell,lstb,lay,rel,cut2,indlst)
! finds the neighbours of atom iat (specified by lsta and lstb) and and
! the relative position rel of iat with respect to these neighbours
      INTEGER                                            :: iat, nn, ncx, ll1, ll2, ll3, l1, l2, l3, &
                                                            myspace
      REAL(KIND=dp)                                      :: rxyz
      INTEGER                                            :: icell, lstb, lay
      REAL(KIND=dp)                                      :: rel, cut2
      INTEGER                                            :: indlst

        DIMENSION rxyz(3,nn),lay(nn),icell(0:ncx,-1:ll1,-1:ll2,-1:ll3), &
                  lstb(0:myspace-1),rel(5,0:myspace-1)

        INTEGER       :: jat,jj,k1,k2,k3
        REAL(KIND=dp) :: rr2,tt,xrel,yrel,zrel,tti

        loop_k3: DO k3=l3-1,l3+1
           loop_k2: DO k2=l2-1,l2+1
              loop_k1: DO k1=l1-1,l1+1
                 loop_jj: DO jj=1,icell(0,k1,k2,k3)
                    jat=icell(jj,k1,k2,k3)
                    IF (jat.eq.iat) CYCLE loop_k3
                    xrel= rxyz(1,iat)-rxyz(1,jat)
                    yrel= rxyz(2,iat)-rxyz(2,jat)
                    zrel= rxyz(3,iat)-rxyz(3,jat)
                    rr2=xrel**2 + yrel**2 + zrel**2
                    IF ( rr2 .LE. cut2 ) THEN
                       indlst=MIN(indlst,myspace-1)
                       lstb(indlst)=lay(jat)
!                       write(*,*) 'iat,indlst,lay(jat)',iat,indlst,lay(jat)
                       tt=SQRT(rr2)
                       tti=1.e0_dp/tt
                       rel(1,indlst)=xrel*tti
                       rel(2,indlst)=yrel*tti
                       rel(3,indlst)=zrel*tti
                       rel(4,indlst)=tt
                       rel(5,indlst)=tti
                       indlst= indlst+1
                    END IF
                 END DO loop_jj
              END DO loop_k1
           END DO loop_k2
        END DO loop_k3

        RETURN
        END SUBROUTINE sublstiat_l

! **************************************************************************************************
!> \brief ...
!> \param ya ...
!> \param y2a ...
!> \param tmin ...
!> \param tmax ...
!> \param hsixth ...
!> \param h2sixth ...
!> \param hi ...
!> \param n ...
!> \param x ...
!> \param y ...
!> \param yp ...
! **************************************************************************************************
        SUBROUTINE splint(ya,y2a,tmin,tmax,hsixth,h2sixth,hi,n,x,y,yp)
      REAL(KIND=dp)                                      :: ya, y2a, tmin, tmax, hsixth, h2sixth, hi
      INTEGER                                            :: n
      REAL(KIND=dp)                                      :: x, y, yp

        DIMENSION y2a(0:n-1),ya(0:n-1)
        REAL(KIND=dp) :: a,a2,b,b2,cof1,cof2,cof3,cof4,tt,&
          y2a_khi,ya_klo,y2a_klo,ya_khi,ypt1,ypt2,yt1,yt2
        INTEGER :: klo,khi

! interpolate if the argument is outside the cubic spline interval [tmin,tmax]
        tt=(x-tmin)*hi
        IF (x.lt.tmin) THEN
          yp=hi*(ya(1)-ya(0)) -  &
          ( y2a(1)+2.e0_dp*y2a(0) )*hsixth
          y=ya(0) + (x-tmin)*yp
        ELSE IF (x.gt.tmax) THEN
          yp=hi*(ya(n-1)-ya(n-2)) +  &
          ( 2.e0_dp*y2a(n-1)+y2a(n-2) )*hsixth
          y=ya(n-1) + (x-tmax)*yp
! otherwise evaluate cubic spline
        ELSE
          klo=INT(tt)
          khi=klo+1
          ya_klo=ya(klo)
          y2a_klo=y2a(klo)
          b=tt-klo
          a=1.e0_dp-b
          ya_khi=ya(khi)
          y2a_khi=y2a(khi)
          b2=b*b
          y=a*ya_klo
          yp=ya_khi-ya_klo
          a2=a*a
          cof1=a2-1.e0_dp
          cof2=b2-1.e0_dp
          y=y+b*ya_khi
          yp=hi*yp
          cof3=3.e0_dp*b2
          cof4=3.e0_dp*a2
          cof1=a*cof1
          cof2=b*cof2
          cof3=cof3-1.e0_dp
          cof4=cof4-1.e0_dp
          yt1=cof1*y2a_klo
          yt2=cof2*y2a_khi
          ypt1=cof3*y2a_khi
          ypt2=cof4*y2a_klo
          y=y + (yt1+yt2)*h2sixth
          yp=yp + ( ypt1 - ypt2 )*hsixth
        ENDIF
      RETURN
      END SUBROUTINE splint

END MODULE eip_silicon
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains routines useful for the application of constraints during MD
!> \par History
!>      none
! **************************************************************************************************
MODULE constraint_util
   USE cell_types,                      ONLY: cell_type
   USE colvar_methods,                  ONLY: colvar_eval_mol_f
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE kinds,                           ONLY: dp,&
                                              int_size
   USE mathlib,                         ONLY: matmul_3x3,&
                                              transpose_3d
   USE message_passing,                 ONLY: mp_sum
   USE molecule_kind_types,             ONLY: colvar_constraint_type,&
                                              fixd_constraint_type,&
                                              g3x3_constraint_type,&
                                              g4x6_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              local_colvar_constraint_type,&
                                              local_constraint_type,&
                                              local_g3x3_constraint_type,&
                                              local_g4x6_constraint_type,&
                                              molecule_type
   USE particle_types,                  ONLY: particle_type
   USE termination,                     ONLY: stop_memory
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: getold, &
             pv_constraint, &
             check_tol, &
             get_roll_matrix, &
             restore_temporary_set, &
             update_temporary_set

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'constraint_util'

CONTAINS

! **************************************************************************************************
!> \brief saves all of the old variables
!> \param gci ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param particle_set ...
!> \param cell ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE getold(gci, local_molecules, molecule_set, molecule_kind_set, &
                     particle_set, cell)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'getold', routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, i, ikind, imol, n3x3con, &
                                                            n4x6con, nkind, nmol_per_kind
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(local_constraint_type), POINTER               :: lci
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      NULLIFY (fixd_list)
      nkind = SIZE(molecule_kind_set)
      ! Intramolecular constraints
      MOL: DO ikind = 1, nkind
         nmol_per_kind = local_molecules%n_el(ikind)
         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            molecule_kind => molecule%molecule_kind
            CALL get_molecule_kind(molecule_kind, ng3x3=n3x3con, ng4x6=n4x6con, &
                                   colv_list=colv_list, g3x3_list=g3x3_list, g4x6_list=g4x6_list, &
                                   fixd_list=fixd_list)
            CALL get_molecule(molecule, lci=lci)
            IF (.NOT. ASSOCIATED(lci)) CYCLE
            CALL get_molecule(molecule, first_atom=first_atom, &
                              lcolv=lcolv, lg3x3=lg3x3, lg4x6=lg4x6)
            CALL getold_low(n3x3con, n4x6con, colv_list, g3x3_list, g4x6_list, fixd_list, &
                            lcolv, lg3x3, lg4x6, first_atom, particle_set, cell)
         END DO
      END DO MOL
      ! Intermolecular constraints
      IF (gci%ntot > 0) THEN
         n3x3con = gci%ng3x3
         n4x6con = gci%ng4x6
         colv_list => gci%colv_list
         g3x3_list => gci%g3x3_list
         g4x6_list => gci%g4x6_list
         fixd_list => gci%fixd_list
         lcolv => gci%lcolv
         lg3x3 => gci%lg3x3
         lg4x6 => gci%lg4x6
         CALL getold_low(n3x3con, n4x6con, colv_list, g3x3_list, g4x6_list, fixd_list, &
                         lcolv, lg3x3, lg4x6, 1, particle_set, cell)
      END IF
   END SUBROUTINE getold

! **************************************************************************************************
!> \brief saves all of the old variables - Low Level
!> \param n3x3con ...
!> \param n4x6con ...
!> \param colv_list ...
!> \param g3x3_list ...
!> \param g4x6_list ...
!> \param fixd_list ...
!> \param lcolv ...
!> \param lg3x3 ...
!> \param lg4x6 ...
!> \param first_atom ...
!> \param particle_set ...
!> \param cell ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE getold_low(n3x3con, n4x6con, colv_list, g3x3_list, g4x6_list, fixd_list, &
                         lcolv, lg3x3, lg4x6, first_atom, particle_set, cell)

      INTEGER, INTENT(IN)                                :: n3x3con, n4x6con
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      INTEGER, INTENT(IN)                                :: first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'getold_low', routineP = moduleN//':'//routineN

      INTEGER                                            :: iconst, index

      IF (ASSOCIATED(colv_list)) THEN
         ! Collective constraints
         DO iconst = 1, SIZE(colv_list)
            CALL colvar_eval_mol_f(lcolv(iconst)%colvar_old, cell, &
                                   particles=particle_set, fixd_list=fixd_list)
         ENDDO
      END IF
      ! 3x3 constraints
      DO iconst = 1, n3x3con
         index = g3x3_list(iconst)%a+first_atom-1
         lg3x3(iconst)%ra_old = particle_set(index)%r
         index = g3x3_list(iconst)%b+first_atom-1
         lg3x3(iconst)%rb_old = particle_set(index)%r
         index = g3x3_list(iconst)%c+first_atom-1
         lg3x3(iconst)%rc_old = particle_set(index)%r
      ENDDO
      ! 4x6 constraints
      DO iconst = 1, n4x6con
         index = g4x6_list(iconst)%a+first_atom-1
         lg4x6(iconst)%ra_old = particle_set(index)%r
         index = g4x6_list(iconst)%b+first_atom-1
         lg4x6(iconst)%rb_old = particle_set(index)%r
         index = g4x6_list(iconst)%c+first_atom-1
         lg4x6(iconst)%rc_old = particle_set(index)%r
         index = g4x6_list(iconst)%d+first_atom-1
         lg4x6(iconst)%rd_old = particle_set(index)%r
      ENDDO

   END SUBROUTINE getold_low

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param particle_set ...
!> \param virial ...
!> \param group ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE pv_constraint(gci, local_molecules, molecule_set, molecule_kind_set, &
                            particle_set, virial, group)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(virial_type), INTENT(INOUT)                   :: virial
      INTEGER, INTENT(IN)                                :: group

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pv_constraint', routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, i, ikind, imol, ng3x3, &
                                                            ng4x6, nkind, nmol_per_kind
      REAL(KIND=dp)                                      :: pv(3, 3)
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(local_constraint_type), POINTER               :: lci
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      pv = 0.0_dp
      nkind = SIZE(molecule_kind_set)
      ! Intramolecular Constraints
      MOL: DO ikind = 1, nkind
         nmol_per_kind = local_molecules%n_el(ikind)
         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            molecule_kind => molecule%molecule_kind
            CALL get_molecule_kind(molecule_kind, ng3x3=ng3x3, &
                                   ng4x6=ng4x6, g3x3_list=g3x3_list, g4x6_list=g4x6_list, &
                                   colv_list=colv_list)
            CALL get_molecule(molecule, lci=lci)
            IF (.NOT. ASSOCIATED(lci)) CYCLE
            CALL get_molecule(molecule, first_atom=first_atom, lg3x3=lg3x3, &
                              lg4x6=lg4x6, lcolv=lcolv)
            CALL pv_constraint_low(ng3x3, ng4x6, g3x3_list, g4x6_list, colv_list, &
                                   first_atom, lg3x3, lg4x6, lcolv, particle_set, pv)
         END DO
      END DO MOL
      ! Intermolecular constraints
      IF (gci%ntot > 0) THEN
         ng3x3 = gci%ng3x3
         ng4x6 = gci%ng4x6
         colv_list => gci%colv_list
         g3x3_list => gci%g3x3_list
         g4x6_list => gci%g4x6_list
         lcolv => gci%lcolv
         lg3x3 => gci%lg3x3
         lg4x6 => gci%lg4x6
         CALL pv_constraint_low(ng3x3, ng4x6, g3x3_list, g4x6_list, colv_list, &
                                1, lg3x3, lg4x6, lcolv, particle_set, pv)
      END IF
      CALL mp_sum(pv, group)
      ! Symmetrize PV
      pv(1, 2) = 0.5_dp*(pv(1, 2)+pv(2, 1))
      pv(2, 1) = pv(1, 2)
      pv(1, 3) = 0.5_dp*(pv(1, 3)+pv(3, 1))
      pv(3, 1) = pv(1, 3)
      pv(3, 2) = 0.5_dp*(pv(3, 2)+pv(2, 3))
      pv(2, 3) = pv(3, 2)
      ! Store in virial type
      virial%pv_constraint = pv

   END SUBROUTINE pv_constraint

! **************************************************************************************************
!> \brief ...
!> \param ng3x3 ...
!> \param ng4x6 ...
!> \param g3x3_list ...
!> \param g4x6_list ...
!> \param colv_list ...
!> \param first_atom ...
!> \param lg3x3 ...
!> \param lg4x6 ...
!> \param lcolv ...
!> \param particle_set ...
!> \param pv ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE pv_constraint_low(ng3x3, ng4x6, g3x3_list, g4x6_list, colv_list, &
                                first_atom, lg3x3, lg4x6, lcolv, particle_set, pv)

      INTEGER, INTENT(IN)                                :: ng3x3, ng4x6
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      INTEGER, INTENT(IN)                                :: first_atom
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(INOUT)      :: pv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pv_constraint_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iconst, index_a, index_b, index_c, &
                                                            index_d
      REAL(KIND=dp)                                      :: fc1(3), fc2(3), fc3(3), fc4(3), &
                                                            lambda_3x3(3), lambda_4x6(6)

      IF (ASSOCIATED(colv_list)) THEN
         ! Colvar Constraints
         DO iconst = 1, SIZE(colv_list)
            CALL pv_colv_eval(pv, lcolv(iconst), particle_set)
         END DO
      END IF
      ! 3x3
      DO iconst = 1, ng3x3
         !  pv gets updated with FULL multiplier
         lambda_3x3 = lg3x3(iconst)%lambda

         fc1 = lambda_3x3(1)*lg3x3(iconst)%fa+ &
               lambda_3x3(2)*lg3x3(iconst)%fb
         fc2 = -lambda_3x3(1)*lg3x3(iconst)%fa+ &
               lambda_3x3(3)*lg3x3(iconst)%fc
         fc3 = -lambda_3x3(2)*lg3x3(iconst)%fb- &
               lambda_3x3(3)*lg3x3(iconst)%fc
         index_a = g3x3_list(iconst)%a+first_atom-1
         index_b = g3x3_list(iconst)%b+first_atom-1
         index_c = g3x3_list(iconst)%c+first_atom-1

         !pv(1,1)
         pv(1, 1) = pv(1, 1)+fc1(1)*particle_set(index_a)%r(1)
         pv(1, 1) = pv(1, 1)+fc2(1)*particle_set(index_b)%r(1)
         pv(1, 1) = pv(1, 1)+fc3(1)*particle_set(index_c)%r(1)
         !pv(1,2)
         pv(1, 2) = pv(1, 2)+fc1(1)*particle_set(index_a)%r(2)
         pv(1, 2) = pv(1, 2)+fc2(1)*particle_set(index_b)%r(2)
         pv(1, 2) = pv(1, 2)+fc3(1)*particle_set(index_c)%r(2)
         !pv(1,3)
         pv(1, 3) = pv(1, 3)+fc1(1)*particle_set(index_a)%r(3)
         pv(1, 3) = pv(1, 3)+fc2(1)*particle_set(index_b)%r(3)
         pv(1, 3) = pv(1, 3)+fc3(1)*particle_set(index_c)%r(3)
         !pv(2,1)
         pv(2, 1) = pv(2, 1)+fc1(2)*particle_set(index_a)%r(1)
         pv(2, 1) = pv(2, 1)+fc2(2)*particle_set(index_b)%r(1)
         pv(2, 1) = pv(2, 1)+fc3(2)*particle_set(index_c)%r(1)
         !pv(2,2)
         pv(2, 2) = pv(2, 2)+fc1(2)*particle_set(index_a)%r(2)
         pv(2, 2) = pv(2, 2)+fc2(2)*particle_set(index_b)%r(2)
         pv(2, 2) = pv(2, 2)+fc3(2)*particle_set(index_c)%r(2)
         !pv(2,3)
         pv(2, 3) = pv(2, 3)+fc1(2)*particle_set(index_a)%r(3)
         pv(2, 3) = pv(2, 3)+fc2(2)*particle_set(index_b)%r(3)
         pv(2, 3) = pv(2, 3)+fc3(2)*particle_set(index_c)%r(3)
         !pv(3,1)
         pv(3, 1) = pv(3, 1)+fc1(3)*particle_set(index_a)%r(1)
         pv(3, 1) = pv(3, 1)+fc2(3)*particle_set(index_b)%r(1)
         pv(3, 1) = pv(3, 1)+fc3(3)*particle_set(index_c)%r(1)
         !pv(3,2)
         pv(3, 2) = pv(3, 2)+fc1(3)*particle_set(index_a)%r(2)
         pv(3, 2) = pv(3, 2)+fc2(3)*particle_set(index_b)%r(2)
         pv(3, 2) = pv(3, 2)+fc3(3)*particle_set(index_c)%r(2)
         !pv(3,3)
         pv(3, 3) = pv(3, 3)+fc1(3)*particle_set(index_a)%r(3)
         pv(3, 3) = pv(3, 3)+fc2(3)*particle_set(index_b)%r(3)
         pv(3, 3) = pv(3, 3)+fc3(3)*particle_set(index_c)%r(3)
      END DO

      ! 4x6
      DO iconst = 1, ng4x6
         !  pv gets updated with FULL multiplier
         lambda_4x6 = lg4x6(iconst)%lambda

         fc1 = lambda_4x6(1)*lg4x6(iconst)%fa+ &
               lambda_4x6(2)*lg4x6(iconst)%fb+ &
               lambda_4x6(3)*lg4x6(iconst)%fc
         fc2 = -lambda_4x6(1)*lg4x6(iconst)%fa+ &
               lambda_4x6(4)*lg4x6(iconst)%fd+ &
               lambda_4x6(5)*lg4x6(iconst)%fe
         fc3 = -lambda_4x6(2)*lg4x6(iconst)%fb- &
               lambda_4x6(4)*lg4x6(iconst)%fd+ &
               lambda_4x6(6)*lg4x6(iconst)%ff
         fc4 = -lambda_4x6(3)*lg4x6(iconst)%fc- &
               lambda_4x6(5)*lg4x6(iconst)%fe- &
               lambda_4x6(6)*lg4x6(iconst)%ff
         index_a = g4x6_list(iconst)%a+first_atom-1
         index_b = g4x6_list(iconst)%b+first_atom-1
         index_c = g4x6_list(iconst)%c+first_atom-1
         index_d = g4x6_list(iconst)%d+first_atom-1

         !pv(1,1)
         pv(1, 1) = pv(1, 1)+fc1(1)*particle_set(index_a)%r(1)
         pv(1, 1) = pv(1, 1)+fc2(1)*particle_set(index_b)%r(1)
         pv(1, 1) = pv(1, 1)+fc3(1)*particle_set(index_c)%r(1)
         pv(1, 1) = pv(1, 1)+fc4(1)*particle_set(index_d)%r(1)
         !pv(1,2)
         pv(1, 2) = pv(1, 2)+fc1(1)*particle_set(index_a)%r(2)
         pv(1, 2) = pv(1, 2)+fc2(1)*particle_set(index_b)%r(2)
         pv(1, 2) = pv(1, 2)+fc3(1)*particle_set(index_c)%r(2)
         pv(1, 2) = pv(1, 2)+fc4(1)*particle_set(index_d)%r(2)
         !pv(1,3)
         pv(1, 3) = pv(1, 3)+fc1(1)*particle_set(index_a)%r(3)
         pv(1, 3) = pv(1, 3)+fc2(1)*particle_set(index_b)%r(3)
         pv(1, 3) = pv(1, 3)+fc3(1)*particle_set(index_c)%r(3)
         pv(1, 3) = pv(1, 3)+fc4(1)*particle_set(index_d)%r(3)
         !pv(2,1)
         pv(2, 1) = pv(2, 1)+fc1(2)*particle_set(index_a)%r(1)
         pv(2, 1) = pv(2, 1)+fc2(2)*particle_set(index_b)%r(1)
         pv(2, 1) = pv(2, 1)+fc3(2)*particle_set(index_c)%r(1)
         pv(2, 1) = pv(2, 1)+fc4(2)*particle_set(index_d)%r(1)
         !pv(2,2)
         pv(2, 2) = pv(2, 2)+fc1(2)*particle_set(index_a)%r(2)
         pv(2, 2) = pv(2, 2)+fc2(2)*particle_set(index_b)%r(2)
         pv(2, 2) = pv(2, 2)+fc3(2)*particle_set(index_c)%r(2)
         pv(2, 2) = pv(2, 2)+fc4(2)*particle_set(index_d)%r(2)
         !pv(2,3)
         pv(2, 3) = pv(2, 3)+fc1(2)*particle_set(index_a)%r(3)
         pv(2, 3) = pv(2, 3)+fc2(2)*particle_set(index_b)%r(3)
         pv(2, 3) = pv(2, 3)+fc3(2)*particle_set(index_c)%r(3)
         pv(2, 3) = pv(2, 3)+fc4(2)*particle_set(index_d)%r(3)
         !pv(3,1)
         pv(3, 1) = pv(3, 1)+fc1(3)*particle_set(index_a)%r(1)
         pv(3, 1) = pv(3, 1)+fc2(3)*particle_set(index_b)%r(1)
         pv(3, 1) = pv(3, 1)+fc3(3)*particle_set(index_c)%r(1)
         pv(3, 1) = pv(3, 1)+fc4(3)*particle_set(index_d)%r(1)
         !pv(3,2)
         pv(3, 2) = pv(3, 2)+fc1(3)*particle_set(index_a)%r(2)
         pv(3, 2) = pv(3, 2)+fc2(3)*particle_set(index_b)%r(2)
         pv(3, 2) = pv(3, 2)+fc3(3)*particle_set(index_c)%r(2)
         pv(3, 2) = pv(3, 2)+fc4(3)*particle_set(index_d)%r(2)
         !pv(3,3)
         pv(3, 3) = pv(3, 3)+fc1(3)*particle_set(index_a)%r(3)
         pv(3, 3) = pv(3, 3)+fc2(3)*particle_set(index_b)%r(3)
         pv(3, 3) = pv(3, 3)+fc3(3)*particle_set(index_c)%r(3)
         pv(3, 3) = pv(3, 3)+fc4(3)*particle_set(index_d)%r(3)
      END DO

   END SUBROUTINE pv_constraint_low

! **************************************************************************************************
!> \brief ...
!> \param pv ...
!> \param lcolv ...
!> \param particle_set ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE pv_colv_eval(pv, lcolv, particle_set)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(INOUT)      :: pv
      TYPE(local_colvar_constraint_type), INTENT(IN)     :: lcolv
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      INTEGER                                            :: i, iatm, ind, j
      REAL(KIND=dp)                                      :: lambda, tmp
      REAL(KIND=dp), DIMENSION(3)                        :: f

      DO iatm = 1, SIZE(lcolv%colvar_old%i_atom)
         ind = lcolv%colvar_old%i_atom(iatm)
         f = -lcolv%colvar_old%dsdr(:, iatm)
         !  pv gets updated with FULL multiplier
         lambda = lcolv%lambda
         DO i = 1, 3
            tmp = lambda*particle_set(ind)%r(i)
            DO j = 1, 3
               pv(j, i) = pv(j, i)+f(j)*tmp
            END DO
         END DO
      END DO

   END SUBROUTINE pv_colv_eval

! **************************************************************************************************
!> \brief ...
!> \param roll_tol ...
!> \param iroll ...
!> \param char ...
!> \param matrix ...
!> \param veps ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE check_tol(roll_tol, iroll, char, matrix, veps)

      REAL(KIND=dp), INTENT(OUT)                         :: roll_tol
      INTEGER, INTENT(INOUT)                             :: iroll
      CHARACTER(LEN=*), INTENT(IN)                       :: char
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: matrix, veps

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_tol', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: local_tol
      REAL(KIND=dp), DIMENSION(3, 3)                     :: diff_rattle, diff_shake
      REAL(KIND=dp), DIMENSION(3, 3), SAVE               :: matrix_old, veps_old

      SELECT CASE (char)
      CASE ('SHAKE')
         IF (iroll == 1) THEN
            matrix_old = matrix
            roll_tol = -1.E10_dp
         ELSE
            roll_tol = 0.0_dp
            diff_shake = ABS(matrix_old-matrix)
            local_tol = MAXVAL(diff_shake)
            roll_tol = MAX(roll_tol, local_tol)
            matrix_old = matrix
         END IF
         iroll = iroll+1
      CASE ('RATTLE')
         IF (iroll == 1) THEN
            veps_old = veps
            roll_tol = -1.E+10_dp
         ELSE
            roll_tol = 0.0_dp
            ! compute tolerance on veps
            diff_rattle = ABS(veps-veps_old)
            local_tol = MAXVAL(diff_rattle)
            roll_tol = MAX(roll_tol, local_tol)
            veps_old = veps
         END IF
         iroll = iroll+1
      END SELECT

   END SUBROUTINE check_tol

! **************************************************************************************************
!> \brief ...
!> \param char ...
!> \param r_shake ...
!> \param v_shake ...
!> \param vector_r ...
!> \param vector_v ...
!> \param u ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE get_roll_matrix(char, r_shake, v_shake, vector_r, vector_v, u)

      CHARACTER(len=*), INTENT(IN)                       :: char
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT), &
         OPTIONAL                                        :: r_shake, v_shake
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: vector_r, vector_v
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: u

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_roll_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp), DIMENSION(3, 3)                     :: diag

      IF (PRESENT(r_shake)) r_shake = 0.0_dp
      IF (PRESENT(v_shake)) v_shake = 0.0_dp
      diag = 0.0_dp

      SELECT CASE (char)
      CASE ('SHAKE')
         IF (PRESENT(u) .AND. PRESENT(vector_v) .AND. &
             PRESENT(vector_r)) THEN
            diag(1, 1) = vector_r(1)
            diag(2, 2) = vector_r(2)
            diag(3, 3) = vector_r(3)
            r_shake = MATMUL_3X3(MATMUL_3X3(u, diag), TRANSPOSE_3D(u))
            diag(1, 1) = vector_v(1)
            diag(2, 2) = vector_v(2)
            diag(3, 3) = vector_v(3)
            v_shake = MATMUL_3X3(MATMUL_3X3(u, diag), TRANSPOSE_3D(u))
            diag = MATMUL_3x3(r_shake, v_shake)
            r_shake = diag
         ELSEIF (.NOT. PRESENT(u) .AND. PRESENT(vector_v) .AND. &
                 PRESENT(vector_r)) THEN
            DO i = 1, 3
               r_shake(i, i) = vector_r(i)*vector_v(i)
               v_shake(i, i) = vector_v(i)
            END DO
         ELSE
            CALL cp__b("constraint_util.F",579,"Not sufficient parameters")
         END IF
      CASE ('RATTLE')
         IF (PRESENT(u) .AND. PRESENT(vector_v)) THEN
            diag(1, 1) = vector_v(1)
            diag(2, 2) = vector_v(2)
            diag(3, 3) = vector_v(3)
            v_shake = MATMUL_3x3(MATMUL_3X3(u, diag), TRANSPOSE_3D(u))
         ELSEIF (.NOT. PRESENT(u) .AND. PRESENT(vector_v)) THEN
            DO i = 1, 3
               v_shake(i, i) = vector_v(i)
            END DO
         ELSE
            CALL cp__b("constraint_util.F",592,"Not sufficient parameters")
         END IF
      END SELECT

   END SUBROUTINE get_roll_matrix

! **************************************************************************************************
!> \brief ...
!> \param particle_set ...
!> \param local_particles ...
!> \param pos ...
!> \param vel ...
!> \par History
!>      Teodoro Laino [tlaino] 2007
! **************************************************************************************************
   SUBROUTINE restore_temporary_set(particle_set, local_particles, pos, vel)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: pos, vel

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restore_temporary_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iparticle, iparticle_kind, &
                                                            iparticle_local, nparticle_local, stat
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: wrk

      ALLOCATE (wrk(SIZE(particle_set)), stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 621, &
                                      "wrk", int_size*SIZE(particle_set))
      wrk = .TRUE.
      DO iparticle_kind = 1, SIZE(local_particles%n_el)
         nparticle_local = local_particles%n_el(iparticle_kind)
         DO iparticle_local = 1, nparticle_local
            iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
            wrk(iparticle) = .FALSE.
         END DO
      END DO
      IF (PRESENT(vel)) THEN
         DO iparticle = 1, SIZE(particle_set)
            IF (wrk(iparticle)) THEN
               vel(:, iparticle) = 0.0_dp
            END IF
         END DO
      END IF
      IF (PRESENT(pos)) THEN
         DO iparticle = 1, SIZE(particle_set)
            IF (wrk(iparticle)) THEN
               pos(:, iparticle) = 0.0_dp
            END IF
         END DO
      END IF
      DEALLOCATE (wrk)
   END SUBROUTINE restore_temporary_set

! **************************************************************************************************
!> \brief ...
!> \param group ...
!> \param pos ...
!> \param vel ...
!> \par History
!>      Teodoro Laino [tlaino] 2007
! **************************************************************************************************
   SUBROUTINE update_temporary_set(group, pos, vel)
      INTEGER, INTENT(IN)                                :: group
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: pos, vel

      IF (PRESENT(pos)) THEN
         CALL mp_sum(pos, group)
      END IF
      IF (PRESENT(vel)) THEN
         CALL mp_sum(vel, group)
      END IF
   END SUBROUTINE update_temporary_set

END MODULE constraint_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Type for storing MD parameters
!> \author CJM
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
!>         reorganization of the original routines/modules
! **************************************************************************************************
MODULE simpar_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
!> \brief Simulation parameter type for molecular dynamics
!> \par History
!>         created [CJM]
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2008
!>         reorganization of the original routines/modules
! **************************************************************************************************
   TYPE simpar_type
      INTEGER        :: nsteps
      REAL(KIND=dp) :: dt
      REAL(KIND=dp) :: dt_fact
      REAL(KIND=dp) :: dr_tol
      REAL(KIND=dp) :: dsc_tol
      REAL(KIND=dp) :: temp_ext
      REAL(KIND=dp) :: temp_baro_ext
      REAL(KIND=dp) :: temp_baro
      REAL(KIND=dp) :: temp_tol
      REAL(KIND=dp) :: temp_baro_tol
      REAL(KIND=dp) :: p_ext
      REAL(KIND=dp) :: cmass
      REAL(KIND=dp) :: cmass_nph
      REAL(KIND=dp) :: v0
      REAL(KIND=dp) :: e0
      REAL(KIND=dp) :: v_shock
      REAL(KIND=dp) :: p0
      REAL(KIND=dp) :: f_annealing
      REAL(KIND=dp) :: f_annealing_cell
      REAL(KIND=dp) :: gamma_nph
      INTEGER        :: ensemble
      LOGICAL        :: constraint
      LOGICAL        :: annealing
      LOGICAL        :: annealing_cell
      LOGICAL        :: dump_lm
      LOGICAL        :: angvel_zero
      LOGICAL        :: variable_dt
      INTEGER        :: nfree, nfree_rot_transl
      INTEGER        :: info_constraint
      INTEGER        :: lagrange_multipliers
      REAL(KIND=dp) :: tau_cell
      ! Constraints Parameters
      REAL(KIND=dp) :: shake_tol, roll_tol
      ! Langevin Parameters
      REAL(KIND=dp) :: gamma
      REAL(KIND=dp) :: noisy_gamma
      REAL(KIND=dp) :: shadow_gamma
      REAL(KIND=dp) :: var_w
      ! RESPA Parameters
      LOGICAL        :: multi_time_switch, do_respa
      INTEGER        :: n_time_steps
      ! SHELL parameters
      REAL(KIND=dp) :: temp_sh_ext
      REAL(KIND=dp) :: temp_sh_tol
      LOGICAL        :: temperature_per_kind
      LOGICAL        :: scale_temperature_per_kind
      LOGICAL        :: do_thermal_region
      ! ADIABATIC parameters
      REAL(KIND=dp) :: temp_slow
      REAL(KIND=dp) :: temp_fast
      REAL(KIND=dp) :: temp_tol_fast, temp_tol_slow
      INTEGER :: n_resp_fast
      ! Velocity softening Parameters
      INTEGER        :: soften_nsteps
      REAL(KIND=dp) :: soften_alpha
      REAL(KIND=dp) :: soften_delta
   END TYPE simpar_type

   PUBLIC :: simpar_type, &
             create_simpar_type, &
             release_simpar_type
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'simpar_types'

CONTAINS
! **************************************************************************************************
!> \brief Creates the simulation parameters type
!> \param simpar ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE create_simpar_type(simpar)
      TYPE(simpar_type), POINTER                         :: simpar

      CHARACTER(len=*), PARAMETER :: routineN = 'create_simpar_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(simpar)))CALL cp__a("simpar_types.F",104)
      ALLOCATE (simpar)
   END SUBROUTINE create_simpar_type

! **************************************************************************************************
!> \brief Releases the simulation parameters type
!> \param simpar ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE release_simpar_type(simpar)
      TYPE(simpar_type), POINTER                         :: simpar

      CHARACTER(len=*), PARAMETER :: routineN = 'release_simpar_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(simpar)))CALL cp__a("simpar_types.F",119)
      DEALLOCATE (simpar)
   END SUBROUTINE release_simpar_type

END MODULE simpar_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
! **************************************************************************************************
MODULE constraint_3x3

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE constraint_fxd,                  ONLY: check_fixed_atom_cns_g3x3
   USE kinds,                           ONLY: dp
   USE linear_systems,                  ONLY: solve_system
   USE mathlib,                         ONLY: dotprod_3d,&
                                              matvec_3x3
   USE molecule_kind_types,             ONLY: fixd_constraint_type,&
                                              g3x3_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              local_g3x3_constraint_type,&
                                              molecule_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: shake_3x3_int, &
             rattle_3x3_int, &
             shake_roll_3x3_int, &
             rattle_roll_3x3_int, &
             shake_3x3_ext, &
             rattle_3x3_ext, &
             shake_roll_3x3_ext, &
             rattle_roll_3x3_ext

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'constraint_3x3'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param molecule ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_3x3_int(molecule, particle_set, pos, vel, dt, ishake, &
                            max_sigma)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: first_atom, ng3x3
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, ng3x3=ng3x3, &
                             g3x3_list=g3x3_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, first_atom=first_atom, lg3x3=lg3x3)
      ! Real Shake
      CALL shake_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, ng3x3, &
                         particle_set, pos, vel, dt, ishake, max_sigma)

   END SUBROUTINE shake_3x3_int

! **************************************************************************************************
!> \brief ...
!> \param molecule ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param v_shake ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_roll_3x3_int(molecule, particle_set, pos, vel, r_shake, &
                                 v_shake, dt, ishake, max_sigma)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake, v_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: first_atom, ng3x3
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, ng3x3=ng3x3, &
                             g3x3_list=g3x3_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, first_atom=first_atom, lg3x3=lg3x3)
      ! Real Shake
      CALL shake_roll_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, ng3x3, &
                              particle_set, pos, vel, r_shake, v_shake, dt, ishake, max_sigma)

   END SUBROUTINE shake_roll_3x3_int

! **************************************************************************************************
!> \brief ...
!> \param molecule ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param veps ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_roll_3x3_int(molecule, particle_set, vel, r_rattle, dt, veps)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_rattle
      REAL(kind=dp), INTENT(in)                          :: dt
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: veps

      INTEGER                                            :: first_atom
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, &
                             g3x3_list=g3x3_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, first_atom=first_atom, lg3x3=lg3x3)
      ! Real Rattle
      CALL rattle_roll_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, &
                               particle_set, vel, r_rattle, dt, veps)

   END SUBROUTINE rattle_roll_3x3_int

! **************************************************************************************************
!> \brief ...
!> \param molecule ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_3x3_int(molecule, particle_set, vel, dt)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt

      INTEGER                                            :: first_atom
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, &
                             g3x3_list=g3x3_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, first_atom=first_atom, lg3x3=lg3x3)
      ! Real Rattle
      CALL rattle_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, &
                          particle_set, vel, dt)

   END SUBROUTINE rattle_3x3_int

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_3x3_ext(gci, particle_set, pos, vel, dt, ishake, &
                            max_sigma)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: first_atom, ng3x3
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)

      first_atom = 1
      ng3x3 = gci%ng3x3
      g3x3_list => gci%g3x3_list
      fixd_list => gci%fixd_list
      lg3x3 => gci%lg3x3
      ! Real Shake
      CALL shake_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, ng3x3, &
                         particle_set, pos, vel, dt, ishake, max_sigma)

   END SUBROUTINE shake_3x3_ext

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param v_shake ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_roll_3x3_ext(gci, particle_set, pos, vel, r_shake, &
                                 v_shake, dt, ishake, max_sigma)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake, v_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: first_atom, ng3x3
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)

      first_atom = 1
      ng3x3 = gci%ng3x3
      g3x3_list => gci%g3x3_list
      fixd_list => gci%fixd_list
      lg3x3 => gci%lg3x3
      ! Real Shake
      CALL shake_roll_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, ng3x3, &
                              particle_set, pos, vel, r_shake, v_shake, dt, ishake, max_sigma)

   END SUBROUTINE shake_roll_3x3_ext

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param veps ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_roll_3x3_ext(gci, particle_set, vel, r_rattle, dt, veps)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_rattle
      REAL(kind=dp), INTENT(in)                          :: dt
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: veps

      INTEGER                                            :: first_atom
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)

      first_atom = 1
      g3x3_list => gci%g3x3_list
      fixd_list => gci%fixd_list
      lg3x3 => gci%lg3x3
      ! Real Rattle
      CALL rattle_roll_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, &
                               particle_set, vel, r_rattle, dt, veps)

   END SUBROUTINE rattle_roll_3x3_ext

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_3x3_ext(gci, particle_set, vel, dt)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt

      INTEGER                                            :: first_atom
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)

      first_atom = 1
      g3x3_list => gci%g3x3_list
      fixd_list => gci%fixd_list
      lg3x3 => gci%lg3x3
      ! Real Rattle
      CALL rattle_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, &
                          particle_set, vel, dt)

   END SUBROUTINE rattle_3x3_ext

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param g3x3_list ...
!> \param lg3x3 ...
!> \param first_atom ...
!> \param ng3x3 ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, ng3x3, &
                            particle_set, pos, vel, dt, ishake, max_sigma)

      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      INTEGER, INTENT(IN)                                :: first_atom, ng3x3
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: iconst, index_a, index_b, index_c
      REAL(KIND=dp)                                      :: dtby2, dtsqby2, idtsq, imass1, imass2, &
                                                            imass3, sigma
      REAL(KIND=dp), DIMENSION(3)                        :: fc1, fc2, fc3, r0_12, r0_13, r0_23, r1, &
                                                            r12, r13, r2, r23, r3, v1, v2, v3, vec
      REAL(KIND=dp), DIMENSION(3, 1)                     :: bvec
      REAL(KIND=dp), DIMENSION(3, 3)                     :: amat, atemp
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      dtsqby2 = dt*dt*.5_dp
      idtsq = 1.0_dp/dt/dt
      dtby2 = dt*.5_dp
      DO iconst = 1, ng3x3
         IF (g3x3_list(iconst)%restraint%active) CYCLE
         index_a = g3x3_list(iconst)%a+first_atom-1
         index_b = g3x3_list(iconst)%b+first_atom-1
         index_c = g3x3_list(iconst)%c+first_atom-1
         IF (ishake == 1) THEN
            r0_12(:) = pos(:, index_a)-pos(:, index_b)
            r0_13(:) = pos(:, index_a)-pos(:, index_c)
            r0_23(:) = pos(:, index_b)-pos(:, index_c)
            atomic_kind => particle_set(index_a)%atomic_kind
            imass1 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_b)%atomic_kind
            imass2 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_c)%atomic_kind
            imass3 = 1.0_dp/atomic_kind%mass
            lg3x3(iconst)%fa = -2.0_dp*(lg3x3(iconst)%ra_old- &
                                        lg3x3(iconst)%rb_old)
            lg3x3(iconst)%fb = -2.0_dp*(lg3x3(iconst)%ra_old- &
                                        lg3x3(iconst)%rc_old)
            lg3x3(iconst)%fc = -2.0_dp*(lg3x3(iconst)%rb_old- &
                                        lg3x3(iconst)%rc_old)
            ! Check for fixed atom constraints
            CALL check_fixed_atom_cns_g3x3(imass1, imass2, imass3, &
                                           index_a, index_b, index_c, fixd_list, lg3x3(iconst))
            ! construct matrix
            amat(1, 1) = (imass1+imass2)*DOTPROD_3D(r0_12, lg3x3(iconst)%fa)
            amat(1, 2) = imass1*DOTPROD_3D(r0_12, lg3x3(iconst)%fb)
            amat(1, 3) = -imass2*DOTPROD_3D(r0_12, lg3x3(iconst)%fc)
            amat(2, 1) = imass1*DOTPROD_3D(r0_13, lg3x3(iconst)%fa)
            amat(2, 2) = (imass1+imass3)*DOTPROD_3D(r0_13, lg3x3(iconst)%fb)
            amat(2, 3) = imass3*DOTPROD_3D(r0_13, lg3x3(iconst)%fc)
            amat(3, 1) = -imass2*DOTPROD_3D(r0_23, lg3x3(iconst)%fa)
            amat(3, 2) = imass3*DOTPROD_3D(r0_23, lg3x3(iconst)%fb)
            amat(3, 3) = (imass3+imass2)*DOTPROD_3D(r0_23, lg3x3(iconst)%fc)
            ! Store values
            lg3x3(iconst)%r0_12 = r0_12
            lg3x3(iconst)%r0_13 = r0_13
            lg3x3(iconst)%r0_23 = r0_23
            lg3x3(iconst)%amat = amat
            lg3x3(iconst)%lambda_old = 0.0_dp
            lg3x3(iconst)%del_lambda = 0.0_dp
         ELSE
            ! Retrieve values
            r0_12 = lg3x3(iconst)%r0_12
            r0_13 = lg3x3(iconst)%r0_13
            r0_23 = lg3x3(iconst)%r0_23
            amat = lg3x3(iconst)%amat
            imass1 = lg3x3(iconst)%imass1
            imass2 = lg3x3(iconst)%imass2
            imass3 = lg3x3(iconst)%imass3
         END IF
         ! Iterate until convergence
         vec = lg3x3(iconst)%lambda(1)*lg3x3(iconst)%fa*(imass1+imass2)+ &
               lg3x3(iconst)%lambda(2)*imass1*lg3x3(iconst)%fb- &
               lg3x3(iconst)%lambda(3)*imass2*lg3x3(iconst)%fc
         bvec(1, 1) = g3x3_list(iconst)%dab*g3x3_list(iconst)%dab &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_12, r0_12)
         vec = lg3x3(iconst)%lambda(1)*lg3x3(iconst)%fa*imass1+ &
               lg3x3(iconst)%lambda(2)*(imass1+imass3)*lg3x3(iconst)%fb+ &
               lg3x3(iconst)%lambda(3)*imass3*lg3x3(iconst)%fc
         bvec(2, 1) = g3x3_list(iconst)%dac*g3x3_list(iconst)%dac &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_13, r0_13)
         vec = -lg3x3(iconst)%lambda(1)*lg3x3(iconst)%fa*imass2+ &
               lg3x3(iconst)%lambda(2)*imass3*lg3x3(iconst)%fb+ &
               lg3x3(iconst)%lambda(3)*(imass2+imass3)*lg3x3(iconst)%fc
         bvec(3, 1) = g3x3_list(iconst)%dbc*g3x3_list(iconst)%dbc &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_23, r0_23)
         bvec = bvec*idtsq
         ! get lambda
         atemp = amat
         CALL solve_system(atemp, 3, bvec)
         lg3x3(iconst)%lambda(:) = bvec(:, 1)
         lg3x3(iconst)%del_lambda(:) = lg3x3(iconst)%lambda(:)- &
                                       lg3x3(iconst)%lambda_old(:)
         lg3x3(iconst)%lambda_old(:) = lg3x3(iconst)%lambda(:)
         fc1 = lg3x3(iconst)%del_lambda(1)*lg3x3(iconst)%fa+ &
               lg3x3(iconst)%del_lambda(2)*lg3x3(iconst)%fb
         fc2 = -lg3x3(iconst)%del_lambda(1)*lg3x3(iconst)%fa+ &
               lg3x3(iconst)%del_lambda(3)*lg3x3(iconst)%fc
         fc3 = -lg3x3(iconst)%del_lambda(2)*lg3x3(iconst)%fb- &
               lg3x3(iconst)%del_lambda(3)*lg3x3(iconst)%fc
         r1(:) = pos(:, index_a)+imass1*dtsqby2*fc1(:)
         r2(:) = pos(:, index_b)+imass2*dtsqby2*fc2(:)
         r3(:) = pos(:, index_c)+imass3*dtsqby2*fc3(:)
         v1(:) = vel(:, index_a)+imass1*dtby2*fc1(:)
         v2(:) = vel(:, index_b)+imass2*dtby2*fc2(:)
         v3(:) = vel(:, index_c)+imass3*dtby2*fc3(:)
         r12 = r1-r2
         r13 = r1-r3
         r23 = r2-r3

         ! compute the tolerance:
         sigma = DOT_PRODUCT(r12, r12)-g3x3_list(iconst)%dab* &
                 g3x3_list(iconst)%dab
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r13, r13)-g3x3_list(iconst)%dac* &
                 g3x3_list(iconst)%dac
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r23, r23)-g3x3_list(iconst)%dbc* &
                 g3x3_list(iconst)%dbc
         max_sigma = MAX(max_sigma, ABS(sigma))
         ! update positions with full multiplier
         pos(:, index_a) = r1(:)
         pos(:, index_b) = r2(:)
         pos(:, index_c) = r3(:)

         ! update velocites with full multiplier
         vel(:, index_a) = v1(:)
         vel(:, index_b) = v2(:)
         vel(:, index_c) = v3(:)
      END DO

   END SUBROUTINE shake_3x3_low

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param g3x3_list ...
!> \param lg3x3 ...
!> \param first_atom ...
!> \param ng3x3 ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param v_shake ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_roll_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, ng3x3, &
                                 particle_set, pos, vel, r_shake, v_shake, dt, ishake, max_sigma)

      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      INTEGER, INTENT(IN)                                :: first_atom, ng3x3
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake, v_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: iconst, index_a, index_b, index_c
      REAL(KIND=dp)                                      :: dtby2, dtsqby2, idtsq, imass1, imass2, &
                                                            imass3, sigma
      REAL(KIND=dp), DIMENSION(3)                        :: f_roll1, f_roll2, f_roll3, fc1, fc2, &
                                                            fc3, r0_12, r0_13, r0_23, r1, r12, &
                                                            r13, r2, r23, r3, v1, v2, v3, vec
      REAL(KIND=dp), DIMENSION(3, 1)                     :: bvec
      REAL(KIND=dp), DIMENSION(3, 3)                     :: amat, atemp
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      dtsqby2 = dt*dt*.5_dp
      idtsq = 1.0_dp/dt/dt
      dtby2 = dt*.5_dp
      DO iconst = 1, ng3x3
         IF (g3x3_list(iconst)%restraint%active) CYCLE
         index_a = g3x3_list(iconst)%a+first_atom-1
         index_b = g3x3_list(iconst)%b+first_atom-1
         index_c = g3x3_list(iconst)%c+first_atom-1
         IF (ishake == 1) THEN
            r0_12(:) = pos(:, index_a)-pos(:, index_b)
            r0_13(:) = pos(:, index_a)-pos(:, index_c)
            r0_23(:) = pos(:, index_b)-pos(:, index_c)
            atomic_kind => particle_set(index_a)%atomic_kind
            imass1 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_b)%atomic_kind
            imass2 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_c)%atomic_kind
            imass3 = 1.0_dp/atomic_kind%mass
            lg3x3(iconst)%fa = -2.0_dp*(lg3x3(iconst)%ra_old- &
                                        lg3x3(iconst)%rb_old)
            lg3x3(iconst)%fb = -2.0_dp*(lg3x3(iconst)%ra_old- &
                                        lg3x3(iconst)%rc_old)
            lg3x3(iconst)%fc = -2.0_dp*(lg3x3(iconst)%rb_old- &
                                        lg3x3(iconst)%rc_old)
            ! Check for fixed atom constraints
            CALL check_fixed_atom_cns_g3x3(imass1, imass2, imass3, &
                                           index_a, index_b, index_c, fixd_list, lg3x3(iconst))
            ! rotate fconst:
            CALL matvec_3x3(f_roll1, r_shake, lg3x3(iconst)%fa)
            CALL matvec_3x3(f_roll2, r_shake, lg3x3(iconst)%fb)
            CALL matvec_3x3(f_roll3, r_shake, lg3x3(iconst)%fc)
            ! construct matrix
            amat(1, 1) = (imass1+imass2)*DOTPROD_3D(r0_12, f_roll1)
            amat(1, 2) = imass1*DOTPROD_3D(r0_12, f_roll2)
            amat(1, 3) = -imass2*DOTPROD_3D(r0_12, f_roll3)
            amat(2, 1) = imass1*DOTPROD_3D(r0_13, f_roll1)
            amat(2, 2) = (imass1+imass3)*DOTPROD_3D(r0_13, f_roll2)
            amat(2, 3) = imass3*DOTPROD_3D(r0_13, f_roll3)
            amat(3, 1) = -imass2*DOTPROD_3D(r0_23, f_roll1)
            amat(3, 2) = imass3*DOTPROD_3D(r0_23, f_roll2)
            amat(3, 3) = (imass3+imass2)*DOTPROD_3D(r0_23, f_roll3)
            ! Store values
            lg3x3(iconst)%r0_12 = r0_12
            lg3x3(iconst)%r0_13 = r0_13
            lg3x3(iconst)%r0_23 = r0_23
            lg3x3(iconst)%f_roll1 = f_roll1
            lg3x3(iconst)%f_roll2 = f_roll2
            lg3x3(iconst)%f_roll3 = f_roll3
            lg3x3(iconst)%amat = amat
            lg3x3(iconst)%lambda_old = 0.0_dp
            lg3x3(iconst)%del_lambda = 0.0_dp
         ELSE
            ! Retrieve values
            r0_12 = lg3x3(iconst)%r0_12
            r0_13 = lg3x3(iconst)%r0_13
            r0_23 = lg3x3(iconst)%r0_23
            f_roll1 = lg3x3(iconst)%f_roll1
            f_roll2 = lg3x3(iconst)%f_roll2
            f_roll3 = lg3x3(iconst)%f_roll3
            amat = lg3x3(iconst)%amat
            imass1 = lg3x3(iconst)%imass1
            imass2 = lg3x3(iconst)%imass2
            imass3 = lg3x3(iconst)%imass3
         END IF
         ! Iterate until convergence
         vec = lg3x3(iconst)%lambda(1)*f_roll1*(imass1+imass2)+ &
               lg3x3(iconst)%lambda(2)*imass1*f_roll2- &
               lg3x3(iconst)%lambda(3)*imass2*f_roll3
         bvec(1, 1) = g3x3_list(iconst)%dab*g3x3_list(iconst)%dab &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_12, r0_12)
         vec = lg3x3(iconst)%lambda(1)*f_roll1*imass1+ &
               lg3x3(iconst)%lambda(2)*(imass1+imass3)*f_roll2+ &
               lg3x3(iconst)%lambda(3)*imass3*f_roll3
         bvec(2, 1) = g3x3_list(iconst)%dac*g3x3_list(iconst)%dac &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_13, r0_13)
         vec = -lg3x3(iconst)%lambda(1)*f_roll1*imass2+ &
               lg3x3(iconst)%lambda(2)*imass3*f_roll2+ &
               lg3x3(iconst)%lambda(3)*(imass2+imass3)*f_roll3
         bvec(3, 1) = g3x3_list(iconst)%dbc*g3x3_list(iconst)%dbc &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_23, r0_23)
         bvec = bvec*idtsq

         ! get lambda
         atemp = amat
         CALL solve_system(atemp, 3, bvec)
         lg3x3(iconst)%lambda(:) = bvec(:, 1)
         lg3x3(iconst)%del_lambda(:) = lg3x3(iconst)%lambda(:)- &
                                       lg3x3(iconst)%lambda_old(:)
         lg3x3(iconst)%lambda_old(:) = lg3x3(iconst)%lambda(:)

         fc1 = lg3x3(iconst)%del_lambda(1)*lg3x3(iconst)%fa+ &
               lg3x3(iconst)%del_lambda(2)*lg3x3(iconst)%fb
         fc2 = -lg3x3(iconst)%del_lambda(1)*lg3x3(iconst)%fa+ &
               lg3x3(iconst)%del_lambda(3)*lg3x3(iconst)%fc
         fc3 = -lg3x3(iconst)%del_lambda(2)*lg3x3(iconst)%fb- &
               lg3x3(iconst)%del_lambda(3)*lg3x3(iconst)%fc
         CALL MATVEC_3x3(vec, r_shake, fc1)
         r1(:) = pos(:, index_a)+imass1*dtsqby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc2)
         r2(:) = pos(:, index_b)+imass2*dtsqby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc3)
         r3(:) = pos(:, index_c)+imass3*dtsqby2*vec
         CALL MATVEC_3x3(vec, v_shake, fc1)
         v1(:) = vel(:, index_a)+imass1*dtby2*vec
         CALL MATVEC_3x3(vec, v_shake, fc2)
         v2(:) = vel(:, index_b)+imass2*dtby2*vec
         CALL MATVEC_3x3(vec, v_shake, fc3)
         v3(:) = vel(:, index_c)+imass3*dtby2*vec
         r12 = r1-r2
         r13 = r1-r3
         r23 = r2-r3

         ! compute the tolerance:
         sigma = DOT_PRODUCT(r12, r12)-g3x3_list(iconst)%dab* &
                 g3x3_list(iconst)%dab
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r13, r13)-g3x3_list(iconst)%dac* &
                 g3x3_list(iconst)%dac
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r23, r23)-g3x3_list(iconst)%dbc* &
                 g3x3_list(iconst)%dbc
         max_sigma = MAX(max_sigma, ABS(sigma))

         ! update positions with full multiplier
         pos(:, index_a) = r1(:)
         pos(:, index_b) = r2(:)
         pos(:, index_c) = r3(:)

         ! update velocites with full multiplier
         vel(:, index_a) = v1(:)
         vel(:, index_b) = v2(:)
         vel(:, index_c) = v3(:)
      END DO
   END SUBROUTINE shake_roll_3x3_low

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param g3x3_list ...
!> \param lg3x3 ...
!> \param first_atom ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param veps ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_roll_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, &
                                  particle_set, vel, r_rattle, dt, veps)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      INTEGER, INTENT(IN)                                :: first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_rattle
      REAL(kind=dp), INTENT(in)                          :: dt
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: veps

      INTEGER                                            :: iconst, index_a, index_b, index_c
      REAL(KIND=dp)                                      :: dtby2, idt, imass1, imass2, imass3, mass
      REAL(KIND=dp), DIMENSION(3)                        :: f_roll1, f_roll2, f_roll3, fc1, fc2, &
                                                            fc3, lambda, r12, r13, r23, v12, v13, &
                                                            v23, vec
      REAL(KIND=dp), DIMENSION(3, 1)                     :: bvec
      REAL(KIND=dp), DIMENSION(3, 3)                     :: amat
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      idt = 1.0_dp/dt
      dtby2 = dt*.5_dp
      DO iconst = 1, SIZE(g3x3_list)
         IF (g3x3_list(iconst)%restraint%active) CYCLE
         index_a = g3x3_list(iconst)%a+first_atom-1
         index_b = g3x3_list(iconst)%b+first_atom-1
         index_c = g3x3_list(iconst)%c+first_atom-1
         v12(:) = vel(:, index_a)-vel(:, index_b)
         v13(:) = vel(:, index_a)-vel(:, index_c)
         v23(:) = vel(:, index_b)-vel(:, index_c)
         r12(:) = particle_set(index_a)%r(:)-particle_set(index_b)%r(:)
         r13(:) = particle_set(index_a)%r(:)-particle_set(index_c)%r(:)
         r23(:) = particle_set(index_b)%r(:)-particle_set(index_c)%r(:)
         atomic_kind => particle_set(index_a)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass1 = 1.0_dp/mass
         atomic_kind => particle_set(index_b)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass2 = 1.0_dp/mass
         atomic_kind => particle_set(index_c)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass3 = 1.0_dp/mass
         lg3x3(iconst)%fa = -2.0_dp*r12
         lg3x3(iconst)%fb = -2.0_dp*r13
         lg3x3(iconst)%fc = -2.0_dp*r23
         ! Check for fixed atom constraints
         CALL check_fixed_atom_cns_g3x3(imass1, imass2, imass3, &
                                        index_a, index_b, index_c, fixd_list, lg3x3(iconst))
         ! roll the fc
         CALL MATVEC_3x3(f_roll1, r_rattle, lg3x3(iconst)%fa)
         CALL MATVEC_3x3(f_roll2, r_rattle, lg3x3(iconst)%fb)
         CALL MATVEC_3x3(f_roll3, r_rattle, lg3x3(iconst)%fc)

         ! construct matrix
         amat(1, 1) = (imass1+imass2)*DOTPROD_3D(r12, f_roll1)
         amat(1, 2) = imass1*DOTPROD_3D(r12, f_roll2)
         amat(1, 3) = -imass2*DOTPROD_3D(r12, f_roll3)
         amat(2, 1) = imass1*DOTPROD_3D(r13, f_roll1)
         amat(2, 2) = (imass1+imass3)*DOTPROD_3D(r13, f_roll2)
         amat(2, 3) = imass3*DOTPROD_3D(r13, f_roll3)
         amat(3, 1) = -imass2*DOTPROD_3D(r23, f_roll1)
         amat(3, 2) = imass3*DOTPROD_3D(r23, f_roll2)
         amat(3, 3) = (imass2+imass3)*DOTPROD_3D(r23, f_roll3)

         ! construct solution vector
         CALL matvec_3x3(vec, veps, r12)
         bvec(1, 1) = DOTPROD_3D(r12, v12+vec)
         CALL matvec_3x3(vec, veps, r13)
         bvec(2, 1) = DOTPROD_3D(r13, v13+vec)
         CALL matvec_3x3(vec, veps, r23)
         bvec(3, 1) = DOTPROD_3D(r23, v23+vec)
         bvec = -bvec*2.0_dp*idt

         ! get lambda
         CALL solve_system(amat, 3, bvec)
         lambda(:) = bvec(:, 1)
         lg3x3(iconst)%lambda(:) = lambda

         fc1 = lambda(1)*f_roll1+ &
               lambda(2)*f_roll2
         fc2 = -lambda(1)*f_roll1+ &
               lambda(3)*f_roll3
         fc3 = -lambda(2)*f_roll2- &
               lambda(3)*f_roll3
         vel(:, index_a) = vel(:, index_a)+imass1*dtby2*fc1(:)
         vel(:, index_b) = vel(:, index_b)+imass2*dtby2*fc2(:)
         vel(:, index_c) = vel(:, index_c)+imass3*dtby2*fc3(:)
      END DO
   END SUBROUTINE rattle_roll_3x3_low

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param g3x3_list ...
!> \param lg3x3 ...
!> \param first_atom ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_3x3_low(fixd_list, g3x3_list, lg3x3, first_atom, &
                             particle_set, vel, dt)

      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g3x3_constraint_type), POINTER                :: g3x3_list(:)
      TYPE(local_g3x3_constraint_type), POINTER          :: lg3x3(:)
      INTEGER, INTENT(IN)                                :: first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt

      INTEGER                                            :: iconst, index_a, index_b, index_c
      REAL(KIND=dp)                                      :: dtby2, idt, imass1, imass2, imass3, mass
      REAL(KIND=dp), DIMENSION(3)                        :: fc1, fc2, fc3, r12, r13, r23, v12, v13, &
                                                            v23
      REAL(KIND=dp), DIMENSION(3, 1)                     :: bvec
      REAL(KIND=dp), DIMENSION(3, 3)                     :: amat
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      idt = 1.0_dp/dt
      dtby2 = dt*.5_dp
      DO iconst = 1, SIZE(g3x3_list)
         IF (g3x3_list(iconst)%restraint%active) CYCLE
         index_a = g3x3_list(iconst)%a+first_atom-1
         index_b = g3x3_list(iconst)%b+first_atom-1
         index_c = g3x3_list(iconst)%c+first_atom-1
         v12(:) = vel(:, index_a)-vel(:, index_b)
         v13(:) = vel(:, index_a)-vel(:, index_c)
         v23(:) = vel(:, index_b)-vel(:, index_c)
         r12(:) = particle_set(index_a)%r(:)-particle_set(index_b)%r(:)
         r13(:) = particle_set(index_a)%r(:)-particle_set(index_c)%r(:)
         r23(:) = particle_set(index_b)%r(:)-particle_set(index_c)%r(:)
         atomic_kind => particle_set(index_a)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass1 = 1.0_dp/mass
         atomic_kind => particle_set(index_b)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass2 = 1.0_dp/mass
         atomic_kind => particle_set(index_c)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass3 = 1.0_dp/mass
         lg3x3(iconst)%fa = -2.0_dp*r12
         lg3x3(iconst)%fb = -2.0_dp*r13
         lg3x3(iconst)%fc = -2.0_dp*r23
         ! Check for fixed atom constraints
         CALL check_fixed_atom_cns_g3x3(imass1, imass2, imass3, &
                                        index_a, index_b, index_c, fixd_list, lg3x3(iconst))
         ! construct matrix
         amat(1, 1) = (imass1+imass2)*DOTPROD_3D(r12, lg3x3(iconst)%fa)
         amat(1, 2) = imass1*DOTPROD_3D(r12, lg3x3(iconst)%fb)
         amat(1, 3) = -imass2*DOTPROD_3D(r12, lg3x3(iconst)%fc)
         amat(2, 1) = imass1*DOTPROD_3D(r13, lg3x3(iconst)%fa)
         amat(2, 2) = (imass1+imass3)*DOTPROD_3D(r13, lg3x3(iconst)%fb)
         amat(2, 3) = imass3*DOTPROD_3D(r13, lg3x3(iconst)%fc)
         amat(3, 1) = -imass2*DOTPROD_3D(r23, lg3x3(iconst)%fa)
         amat(3, 2) = imass3*DOTPROD_3D(r23, lg3x3(iconst)%fb)
         amat(3, 3) = (imass2+imass3)*DOTPROD_3D(r23, lg3x3(iconst)%fc)

         ! construct solution vector
         bvec(1, 1) = DOTPROD_3D(r12, v12)
         bvec(2, 1) = DOTPROD_3D(r13, v13)
         bvec(3, 1) = DOTPROD_3D(r23, v23)
         bvec = -bvec*2.0_dp*idt

         ! get lambda
         CALL solve_system(amat, 3, bvec)
         lg3x3(iconst)%lambda(:) = bvec(:, 1)

         fc1 = lg3x3(iconst)%lambda(1)*lg3x3(iconst)%fa+ &
               lg3x3(iconst)%lambda(2)*lg3x3(iconst)%fb
         fc2 = -lg3x3(iconst)%lambda(1)*lg3x3(iconst)%fa+ &
               lg3x3(iconst)%lambda(3)*lg3x3(iconst)%fc
         fc3 = -lg3x3(iconst)%lambda(2)*lg3x3(iconst)%fb- &
               lg3x3(iconst)%lambda(3)*lg3x3(iconst)%fc
         vel(:, index_a) = vel(:, index_a)+imass1*dtby2*fc1(:)
         vel(:, index_b) = vel(:, index_b)+imass2*dtby2*fc2(:)
         vel(:, index_c) = vel(:, index_c)+imass3*dtby2*fc3(:)
      END DO
   END SUBROUTINE rattle_3x3_low

END MODULE constraint_3x3
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to handle the virtual site constraint/restraint
!> \par History
!>      Teodoro Laino [tlaino] 12.2008 - Preparing for VIRTUAL SITE constraints
!>                                       (patch by Marcel Baer)
! **************************************************************************************************
MODULE constraint_vsite
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE kinds,                           ONLY: dp
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: get_molecule_kind,&
                                              molecule_kind_type,&
                                              vsite_constraint_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              molecule_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: shake_vsite_int, &
             shake_vsite_ext, &
             vsite_force_control

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'constraint_vsite'

CONTAINS

! **************************************************************************************************
!> \brief control force distribution for virtual sites
!> \param force_env ...
!> \date 12.2008
!> \par History
!>      - none
!> \author Marcel Baer
! **************************************************************************************************
   SUBROUTINE vsite_force_control(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      INTEGER                                            :: i, ikind, imol, nconstraint, nkind, &
                                                            nmol_per_kind, nvsitecon
      LOGICAL                                            :: do_ext_constraint
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      NULLIFY (gci, subsys, local_molecules, local_particles, &
               molecule_kinds)

      CALL force_env_get(force_env=force_env, subsys=subsys)

      CALL cp_subsys_get(subsys=subsys, local_particles=local_particles, &
                         particles=particles, local_molecules_new=local_molecules, &
                         molecule_kinds_new=molecule_kinds, gci=gci, molecules_new=molecules)

      molecule_kind_set => molecule_kinds%els
      molecule_set => molecules%els
      particle_set => particles%els
      nkind = SIZE(molecule_kind_set)
      ! Intermolecular Virtual Site Constraints
      do_ext_constraint = .FALSE.
      IF (ASSOCIATED(gci)) THEN
         do_ext_constraint = (gci%ntot /= 0)
      END IF
      ! Intramolecular Virtual Site Constraints
      MOL: DO ikind = 1, nkind
         nmol_per_kind = local_molecules%n_el(ikind)
         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            molecule_kind => molecule%molecule_kind
            CALL get_molecule_kind(molecule_kind, nconstraint=nconstraint, nvsite=nvsitecon)
            IF (nconstraint == 0) CYCLE
            IF (nvsitecon /= 0) &
               CALL force_vsite_int(molecule, particle_set)
         END DO
      END DO MOL
      ! Intermolecular Virtual Site Constraints
      IF (do_ext_constraint) THEN
         IF (gci%nvsite /= 0) &
            CALL force_vsite_ext(gci, particle_set)
      END IF

   END SUBROUTINE vsite_force_control

! **************************************************************************************************
!> \brief Intramolecular virtual site
!> \param molecule ...
!> \param pos ...
!> \par History
!>      12.2008 Marcel Baer
! **************************************************************************************************
   SUBROUTINE shake_vsite_int(molecule, pos)
      TYPE(molecule_type), POINTER                       :: molecule
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :)

      INTEGER                                            :: first_atom, nvsite
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(vsite_constraint_type), POINTER               :: vsite_list(:)

      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, nvsite=nvsite, vsite_list=vsite_list)
      CALL get_molecule(molecule, first_atom=first_atom)
      ! Real Shake
      CALL shake_vsite_low(vsite_list, nvsite, first_atom, pos)

   END SUBROUTINE shake_vsite_int

! **************************************************************************************************
!> \brief Intramolecular virtual site
!> \param gci ...
!> \param pos ...
!> \par History
!>      12.2008 Marcel Baer
! **************************************************************************************************
   SUBROUTINE shake_vsite_ext(gci, pos)

      TYPE(global_constraint_type), POINTER              :: gci
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :)

      INTEGER                                            :: first_atom, nvsite
      TYPE(vsite_constraint_type), POINTER               :: vsite_list(:)

      first_atom = 1
      nvsite = gci%nvsite
      vsite_list => gci%vsite_list
      ! Real Shake
      CALL shake_vsite_low(vsite_list, nvsite, first_atom, pos)

   END SUBROUTINE shake_vsite_ext

! **************************************************************************************************
!> \brief ...
!> \param vsite_list ...
!> \param nvsite ...
!> \param first_atom ...
!> \param pos ...
!> \par History
!>      12.2008 Marcel Bear
! **************************************************************************************************
   SUBROUTINE shake_vsite_low(vsite_list, nvsite, first_atom, pos)
      TYPE(vsite_constraint_type)                        :: vsite_list(:)
      INTEGER, INTENT(IN)                                :: nvsite, first_atom
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :)

      INTEGER                                            :: iconst, index_a, index_b, index_c, &
                                                            index_d
      REAL(KIND=dp), DIMENSION(3)                        :: r1, r2

      DO iconst = 1, nvsite
         IF (vsite_list(iconst)%restraint%active) CYCLE
         index_a = vsite_list(iconst)%a+first_atom-1
         index_b = vsite_list(iconst)%b+first_atom-1
         index_c = vsite_list(iconst)%c+first_atom-1
         index_d = vsite_list(iconst)%d+first_atom-1

         r1(:) = pos(:, index_b)-pos(:, index_c)
         r2(:) = pos(:, index_d)-pos(:, index_c)
         pos(:, index_a) = pos(:, index_c)+vsite_list(iconst)%wbc*r1(:)+ &
                           vsite_list(iconst)%wdc*r2(:)
      END DO
   END SUBROUTINE shake_vsite_low

! **************************************************************************************************
!> \brief Intramolecular virtual site
!> \param molecule ...
!> \param particle_set ...
!> \par History
!>      12.2008 Marcel Bear
! **************************************************************************************************
   SUBROUTINE force_vsite_int(molecule, particle_set)
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)

      INTEGER                                            :: first_atom, iconst, index_a, index_b, &
                                                            index_c, index_d, nvsite
      REAL(KIND=dp)                                      :: wb, wc, wd
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(vsite_constraint_type), POINTER               :: vsite_list(:)

      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, nvsite=nvsite, vsite_list=vsite_list)
      CALL get_molecule(molecule, first_atom=first_atom)

      DO iconst = 1, nvsite
         IF (vsite_list(iconst)%restraint%active) CYCLE
         index_a = vsite_list(iconst)%a+first_atom-1
         index_b = vsite_list(iconst)%b+first_atom-1
         index_c = vsite_list(iconst)%c+first_atom-1
         index_d = vsite_list(iconst)%d+first_atom-1

         wb = vsite_list(iconst)%wbc
         wd = vsite_list(iconst)%wdc
         wc = 1.0_dp-vsite_list(iconst)%wbc-vsite_list(iconst)%wdc

         particle_set(index_b)%f(:) = particle_set(index_b)%f(:)+wb*particle_set(index_a)%f(:)
         particle_set(index_c)%f(:) = particle_set(index_c)%f(:)+wc*particle_set(index_a)%f(:)
         particle_set(index_d)%f(:) = particle_set(index_d)%f(:)+wd*particle_set(index_a)%f(:)
         particle_set(index_a)%f(:) = 0.0_dp
      END DO

   END SUBROUTINE force_vsite_int

! **************************************************************************************************
!> \brief Intramolecular virtual site
!> \param gci ...
!> \param particle_set ...
!> \par History
!>      12.2008 Marcel Bear
! **************************************************************************************************
   SUBROUTINE force_vsite_ext(gci, particle_set)
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)

      INTEGER                                            :: first_atom, iconst, index_a, index_b, &
                                                            index_c, index_d, nvsite
      REAL(KIND=dp)                                      :: wb, wc, wd
      TYPE(vsite_constraint_type), POINTER               :: vsite_list(:)

      first_atom = 1
      nvsite = gci%nvsite
      vsite_list => gci%vsite_list
      ! Real Shake

      DO iconst = 1, nvsite
         IF (vsite_list(iconst)%restraint%active) CYCLE
         index_a = vsite_list(iconst)%a+first_atom-1
         index_b = vsite_list(iconst)%b+first_atom-1
         index_c = vsite_list(iconst)%c+first_atom-1
         index_d = vsite_list(iconst)%d+first_atom-1

         wb = vsite_list(iconst)%wbc
         wd = vsite_list(iconst)%wdc
         wc = 1.0_dp-vsite_list(iconst)%wbc-vsite_list(iconst)%wdc

         particle_set(index_b)%f(:) = particle_set(index_b)%f(:)+wb*particle_set(index_a)%f(:)
         particle_set(index_c)%f(:) = particle_set(index_c)%f(:)+wc*particle_set(index_a)%f(:)
         particle_set(index_d)%f(:) = particle_set(index_d)%f(:)+wd*particle_set(index_a)%f(:)
         particle_set(index_a)%f(:) = 0.0_dp
      END DO
   END SUBROUTINE force_vsite_ext

END MODULE constraint_vsite
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
! **************************************************************************************************
MODULE constraint_4x6

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE constraint_fxd,                  ONLY: check_fixed_atom_cns_g4x6
   USE kinds,                           ONLY: dp
   USE linear_systems,                  ONLY: solve_system
   USE mathlib,                         ONLY: dotprod_3d,&
                                              matvec_3x3
   USE molecule_kind_types,             ONLY: fixd_constraint_type,&
                                              g4x6_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              local_g4x6_constraint_type,&
                                              molecule_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: shake_4x6_int, &
             rattle_4x6_int, &
             shake_roll_4x6_int, &
             rattle_roll_4x6_int, &
             shake_4x6_ext, &
             rattle_4x6_ext, &
             shake_roll_4x6_ext, &
             rattle_roll_4x6_ext

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'constraint_4x6'

CONTAINS

! **************************************************************************************************
!> \brief Intramolecular shake_4x6
!> \param molecule ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_4x6_int(molecule, particle_set, pos, vel, dt, ishake, &
                            max_sigma)
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, ng4x6=ng4x6, &
                             fixd_list=fixd_list, g4x6_list=g4x6_list)
      CALL get_molecule(molecule, first_atom=first_atom, lg4x6=lg4x6)
      ! Real Shake
      CALL shake_4x6_low(fixd_list, g4x6_list, lg4x6, ng4x6, first_atom, &
                         particle_set, pos, vel, dt, ishake, max_sigma)

   END SUBROUTINE shake_4x6_int

! **************************************************************************************************
!> \brief Intramolecular shake_4x6_roll
!> \param molecule ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_roll_4x6_int(molecule, particle_set, pos, vel, r_shake, &
                                 dt, ishake, max_sigma)
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, ng4x6=ng4x6, &
                             fixd_list=fixd_list, g4x6_list=g4x6_list)
      CALL get_molecule(molecule, first_atom=first_atom, lg4x6=lg4x6)
      ! Real Shake
      CALL shake_roll_4x6_low(fixd_list, g4x6_list, lg4x6, ng4x6, first_atom, &
                              particle_set, pos, vel, r_shake, dt, ishake, max_sigma)

   END SUBROUTINE shake_roll_4x6_int

! **************************************************************************************************
!> \brief Intramolecular rattle_4x6
!> \param molecule ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_4x6_int(molecule, particle_set, vel, dt)
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, ng4x6=ng4x6, &
                             fixd_list=fixd_list, g4x6_list=g4x6_list)
      CALL get_molecule(molecule, first_atom=first_atom, lg4x6=lg4x6)
      ! Real Rattle
      CALL rattle_4x6_low(fixd_list, g4x6_list, lg4x6, first_atom, &
                          particle_set, vel, dt)

   END SUBROUTINE rattle_4x6_int

! **************************************************************************************************
!> \brief Intramolecular rattle_4x6_roll
!> \param molecule ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param veps ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_roll_4x6_int(molecule, particle_set, vel, r_rattle, dt, veps)
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_rattle
      REAL(kind=dp), INTENT(in)                          :: dt
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: veps

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, ng4x6=ng4x6, &
                             fixd_list=fixd_list, g4x6_list=g4x6_list)
      CALL get_molecule(molecule, first_atom=first_atom, lg4x6=lg4x6)
      ! Real Rattle
      CALL rattle_roll_4x6_low(fixd_list, g4x6_list, lg4x6, first_atom, &
                               particle_set, vel, r_rattle, dt, veps)

   END SUBROUTINE rattle_roll_4x6_int

! **************************************************************************************************
!> \brief Intramolecular shake_4x6
!> \param gci ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_4x6_ext(gci, particle_set, pos, vel, dt, ishake, &
                            max_sigma)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)

      first_atom = 1
      ng4x6 = gci%ng4x6
      fixd_list => gci%fixd_list
      g4x6_list => gci%g4x6_list
      lg4x6 => gci%lg4x6
      ! Real Shake
      CALL shake_4x6_low(fixd_list, g4x6_list, lg4x6, ng4x6, first_atom, &
                         particle_set, pos, vel, dt, ishake, max_sigma)

   END SUBROUTINE shake_4x6_ext

! **************************************************************************************************
!> \brief Intramolecular shake_4x6_roll
!> \param gci ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_roll_4x6_ext(gci, particle_set, pos, vel, r_shake, &
                                 dt, ishake, max_sigma)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)

      first_atom = 1
      ng4x6 = gci%ng4x6
      fixd_list => gci%fixd_list
      g4x6_list => gci%g4x6_list
      lg4x6 => gci%lg4x6
      ! Real Shake
      CALL shake_roll_4x6_low(fixd_list, g4x6_list, lg4x6, ng4x6, first_atom, &
                              particle_set, pos, vel, r_shake, dt, ishake, max_sigma)

   END SUBROUTINE shake_roll_4x6_ext

! **************************************************************************************************
!> \brief Intramolecular rattle_4x6
!> \param gci ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_4x6_ext(gci, particle_set, vel, dt)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)

      first_atom = 1
      ng4x6 = gci%ng4x6
      fixd_list => gci%fixd_list
      g4x6_list => gci%g4x6_list
      lg4x6 => gci%lg4x6
      ! Real Rattle
      CALL rattle_4x6_low(fixd_list, g4x6_list, lg4x6, first_atom, &
                          particle_set, vel, dt)

   END SUBROUTINE rattle_4x6_ext

! **************************************************************************************************
!> \brief Intramolecular rattle_4x6_roll
!> \param gci ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param veps ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_roll_4x6_ext(gci, particle_set, vel, r_rattle, dt, veps)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_rattle
      REAL(kind=dp), INTENT(in)                          :: dt
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: veps

      INTEGER                                            :: first_atom, ng4x6
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)

      first_atom = 1
      ng4x6 = gci%ng4x6
      fixd_list => gci%fixd_list
      g4x6_list => gci%g4x6_list
      lg4x6 => gci%lg4x6
      ! Real Rattle
      CALL rattle_roll_4x6_low(fixd_list, g4x6_list, lg4x6, first_atom, &
                               particle_set, vel, r_rattle, dt, veps)

   END SUBROUTINE rattle_roll_4x6_ext

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param g4x6_list ...
!> \param lg4x6 ...
!> \param ng4x6 ...
!> \param first_atom ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_4x6_low(fixd_list, g4x6_list, lg4x6, ng4x6, first_atom, &
                            particle_set, pos, vel, dt, ishake, max_sigma)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      INTEGER, INTENT(IN)                                :: ng4x6, first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: iconst, index_a, index_b, index_c, &
                                                            index_d
      REAL(KIND=dp)                                      :: dtby2, dtsqby2, idtsq, imass1, imass2, &
                                                            imass3, imass4, sigma
      REAL(KIND=dp), DIMENSION(3)                        :: fc1, fc2, fc3, fc4, r0_12, r0_13, r0_14, &
                                                            r0_23, r0_24, r0_34, r1, r12, r13, &
                                                            r14, r2, r23, r24, r3, r34, r4, v1, &
                                                            v2, v3, v4, vec
      REAL(KIND=dp), DIMENSION(6, 1)                     :: bvec
      REAL(KIND=dp), DIMENSION(6, 6)                     :: amat, atemp
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      dtsqby2 = dt*dt*.5_dp
      idtsq = 1.0_dp/dt/dt
      dtby2 = dt*.5_dp
      DO iconst = 1, ng4x6
         IF (g4x6_list(iconst)%restraint%active) CYCLE
         index_a = g4x6_list(iconst)%a+first_atom-1
         index_b = g4x6_list(iconst)%b+first_atom-1
         index_c = g4x6_list(iconst)%c+first_atom-1
         index_d = g4x6_list(iconst)%d+first_atom-1
         IF (ishake == 1) THEN
            r0_12(:) = pos(:, index_a)-pos(:, index_b)
            r0_13(:) = pos(:, index_a)-pos(:, index_c)
            r0_14(:) = pos(:, index_a)-pos(:, index_d)
            r0_23(:) = pos(:, index_b)-pos(:, index_c)
            r0_24(:) = pos(:, index_b)-pos(:, index_d)
            r0_34(:) = pos(:, index_c)-pos(:, index_d)
            atomic_kind => particle_set(index_a)%atomic_kind
            imass1 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_b)%atomic_kind
            imass2 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_c)%atomic_kind
            imass3 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_d)%atomic_kind
            imass4 = 1.0_dp/atomic_kind%mass
            lg4x6(iconst)%fa = -2.0_dp*(lg4x6(iconst)%ra_old- &
                                        lg4x6(iconst)%rb_old)
            lg4x6(iconst)%fb = -2.0_dp*(lg4x6(iconst)%ra_old- &
                                        lg4x6(iconst)%rc_old)
            lg4x6(iconst)%fc = -2.0_dp*(lg4x6(iconst)%ra_old- &
                                        lg4x6(iconst)%rd_old)
            lg4x6(iconst)%fd = -2.0_dp*(lg4x6(iconst)%rb_old- &
                                        lg4x6(iconst)%rc_old)
            lg4x6(iconst)%fe = -2.0_dp*(lg4x6(iconst)%rb_old- &
                                        lg4x6(iconst)%rd_old)
            lg4x6(iconst)%ff = -2.0_dp*(lg4x6(iconst)%rc_old- &
                                        lg4x6(iconst)%rd_old)
            ! Check for fixed atom constraints
            CALL check_fixed_atom_cns_g4x6(imass1, imass2, imass3, imass4, &
                                           index_a, index_b, index_c, index_d, fixd_list, lg4x6(iconst))
            ! construct matrix
            amat(1, 1) = (imass1+imass2)*DOTPROD_3D(r0_12, lg4x6(iconst)%fa)
            amat(1, 2) = imass1*DOTPROD_3D(r0_12, lg4x6(iconst)%fb)
            amat(1, 3) = imass1*DOTPROD_3D(r0_12, lg4x6(iconst)%fc)
            amat(1, 4) = -imass2*DOTPROD_3D(r0_12, lg4x6(iconst)%fd)
            amat(1, 5) = -imass2*DOTPROD_3D(r0_12, lg4x6(iconst)%fe)
            amat(1, 6) = 0.0_dp

            amat(2, 1) = imass1*DOTPROD_3D(r0_13, lg4x6(iconst)%fa)
            amat(2, 2) = (imass1+imass3)*DOTPROD_3D(r0_13, lg4x6(iconst)%fb)
            amat(2, 3) = imass1*DOTPROD_3D(r0_13, lg4x6(iconst)%fc)
            amat(2, 4) = imass3*DOTPROD_3D(r0_13, lg4x6(iconst)%fd)
            amat(2, 5) = 0.0_dp
            amat(2, 6) = -imass3*DOTPROD_3D(r0_13, lg4x6(iconst)%ff)

            amat(3, 1) = imass1*DOTPROD_3D(r0_14, lg4x6(iconst)%fa)
            amat(3, 2) = imass1*DOTPROD_3D(r0_14, lg4x6(iconst)%fb)
            amat(3, 3) = (imass1+imass4)*DOTPROD_3D(r0_14, lg4x6(iconst)%fc)
            amat(3, 4) = 0.0_dp
            amat(3, 5) = imass4*DOTPROD_3D(r0_14, lg4x6(iconst)%fe)
            amat(3, 6) = imass4*DOTPROD_3D(r0_14, lg4x6(iconst)%ff)

            amat(4, 1) = -imass2*DOTPROD_3D(r0_23, lg4x6(iconst)%fa)
            amat(4, 2) = imass3*DOTPROD_3D(r0_23, lg4x6(iconst)%fb)
            amat(4, 3) = 0.0_dp
            amat(4, 4) = (imass3+imass2)*DOTPROD_3D(r0_23, lg4x6(iconst)%fd)
            amat(4, 5) = imass2*DOTPROD_3D(r0_23, lg4x6(iconst)%fe)
            amat(4, 6) = -imass3*DOTPROD_3D(r0_23, lg4x6(iconst)%ff)

            amat(5, 1) = -imass2*DOTPROD_3D(r0_24, lg4x6(iconst)%fa)
            amat(5, 2) = 0.0_dp
            amat(5, 3) = imass4*DOTPROD_3D(r0_24, lg4x6(iconst)%fc)
            amat(5, 4) = imass2*DOTPROD_3D(r0_24, lg4x6(iconst)%fd)
            amat(5, 5) = (imass4+imass2)*DOTPROD_3D(r0_24, lg4x6(iconst)%fe)
            amat(5, 6) = imass4*DOTPROD_3D(r0_24, lg4x6(iconst)%ff)

            amat(6, 1) = 0.0_dp
            amat(6, 2) = -imass3*DOTPROD_3D(r0_34, lg4x6(iconst)%fb)
            amat(6, 3) = imass4*DOTPROD_3D(r0_34, lg4x6(iconst)%fc)
            amat(6, 4) = -imass3*DOTPROD_3D(r0_34, lg4x6(iconst)%fd)
            amat(6, 5) = imass4*DOTPROD_3D(r0_34, lg4x6(iconst)%fe)
            amat(6, 6) = (imass3+imass4)*DOTPROD_3D(r0_34, lg4x6(iconst)%ff)
            ! Store values
            lg4x6(iconst)%r0_12 = r0_12
            lg4x6(iconst)%r0_13 = r0_13
            lg4x6(iconst)%r0_14 = r0_14
            lg4x6(iconst)%r0_23 = r0_23
            lg4x6(iconst)%r0_24 = r0_24
            lg4x6(iconst)%r0_34 = r0_34
            lg4x6(iconst)%amat = amat
            lg4x6(iconst)%imass1 = imass1
            lg4x6(iconst)%imass2 = imass2
            lg4x6(iconst)%imass3 = imass3
            lg4x6(iconst)%imass4 = imass4
            lg4x6(iconst)%lambda_old = 0.0_dp
            lg4x6(iconst)%del_lambda = 0.0_dp
         ELSE
            ! Retrieve values
            r0_12 = lg4x6(iconst)%r0_12
            r0_13 = lg4x6(iconst)%r0_13
            r0_14 = lg4x6(iconst)%r0_14
            r0_23 = lg4x6(iconst)%r0_23
            r0_24 = lg4x6(iconst)%r0_24
            r0_34 = lg4x6(iconst)%r0_34
            amat = lg4x6(iconst)%amat
            imass1 = lg4x6(iconst)%imass1
            imass2 = lg4x6(iconst)%imass2
            imass3 = lg4x6(iconst)%imass3
            imass4 = lg4x6(iconst)%imass4
         END IF

         ! Iterate until convergence:
         vec = lg4x6(iconst)%lambda(1)*lg4x6(iconst)%fa*(imass1+imass2)+ &
               lg4x6(iconst)%lambda(2)*imass1*lg4x6(iconst)%fb+ &
               lg4x6(iconst)%lambda(3)*imass1*lg4x6(iconst)%fc- &
               lg4x6(iconst)%lambda(4)*imass2*lg4x6(iconst)%fd- &
               lg4x6(iconst)%lambda(5)*imass2*lg4x6(iconst)%fe
         bvec(1, 1) = g4x6_list(iconst)%dab*g4x6_list(iconst)%dab &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_12, r0_12)
         vec = lg4x6(iconst)%lambda(2)*lg4x6(iconst)%fb*(imass1+imass3)+ &
               lg4x6(iconst)%lambda(1)*imass1*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%lambda(3)*imass1*lg4x6(iconst)%fc+ &
               lg4x6(iconst)%lambda(4)*imass3*lg4x6(iconst)%fd- &
               lg4x6(iconst)%lambda(6)*imass3*lg4x6(iconst)%ff
         bvec(2, 1) = g4x6_list(iconst)%dac*g4x6_list(iconst)%dac &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_13, r0_13)
         vec = lg4x6(iconst)%lambda(3)*lg4x6(iconst)%fc*(imass1+imass4)+ &
               lg4x6(iconst)%lambda(1)*imass1*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%lambda(2)*imass1*lg4x6(iconst)%fb+ &
               lg4x6(iconst)%lambda(5)*imass4*lg4x6(iconst)%fe+ &
               lg4x6(iconst)%lambda(6)*imass4*lg4x6(iconst)%ff
         bvec(3, 1) = g4x6_list(iconst)%dad*g4x6_list(iconst)%dad &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_14, r0_14)
         vec = lg4x6(iconst)%lambda(4)*lg4x6(iconst)%fd*(imass2+imass3)- &
               lg4x6(iconst)%lambda(1)*imass2*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%lambda(2)*imass3*lg4x6(iconst)%fb+ &
               lg4x6(iconst)%lambda(5)*imass2*lg4x6(iconst)%fe- &
               lg4x6(iconst)%lambda(6)*imass3*lg4x6(iconst)%ff
         bvec(4, 1) = g4x6_list(iconst)%dbc*g4x6_list(iconst)%dbc &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_23, r0_23)
         vec = lg4x6(iconst)%lambda(5)*lg4x6(iconst)%fe*(imass2+imass4)- &
               lg4x6(iconst)%lambda(1)*imass2*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%lambda(3)*imass4*lg4x6(iconst)%fc+ &
               lg4x6(iconst)%lambda(4)*imass2*lg4x6(iconst)%fd+ &
               lg4x6(iconst)%lambda(6)*imass4*lg4x6(iconst)%ff
         bvec(5, 1) = g4x6_list(iconst)%dbd*g4x6_list(iconst)%dbd &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_24, r0_24)
         vec = lg4x6(iconst)%lambda(6)*lg4x6(iconst)%ff*(imass3+imass4)- &
               lg4x6(iconst)%lambda(2)*imass3*lg4x6(iconst)%fb+ &
               lg4x6(iconst)%lambda(3)*imass4*lg4x6(iconst)%fc- &
               lg4x6(iconst)%lambda(4)*imass3*lg4x6(iconst)%fd+ &
               lg4x6(iconst)%lambda(5)*imass4*lg4x6(iconst)%fe
         bvec(6, 1) = g4x6_list(iconst)%dcd*g4x6_list(iconst)%dcd &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_34, r0_34)
         bvec = bvec*idtsq

         ! get lambda
         atemp = amat
         CALL solve_system(atemp, 6, bvec)
         lg4x6(iconst)%lambda(:) = bvec(:, 1)
         lg4x6(iconst)%del_lambda(:) = lg4x6(iconst)%lambda(:)- &
                                       lg4x6(iconst)%lambda_old(:)
         lg4x6(iconst)%lambda_old(:) = lg4x6(iconst)%lambda(:)

         fc1 = lg4x6(iconst)%del_lambda(1)*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%del_lambda(2)*lg4x6(iconst)%fb+ &
               lg4x6(iconst)%del_lambda(3)*lg4x6(iconst)%fc
         fc2 = -lg4x6(iconst)%del_lambda(1)*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%del_lambda(4)*lg4x6(iconst)%fd+ &
               lg4x6(iconst)%del_lambda(5)*lg4x6(iconst)%fe
         fc3 = -lg4x6(iconst)%del_lambda(2)*lg4x6(iconst)%fb- &
               lg4x6(iconst)%del_lambda(4)*lg4x6(iconst)%fd+ &
               lg4x6(iconst)%del_lambda(6)*lg4x6(iconst)%ff
         fc4 = -lg4x6(iconst)%del_lambda(3)*lg4x6(iconst)%fc- &
               lg4x6(iconst)%del_lambda(5)*lg4x6(iconst)%fe- &
               lg4x6(iconst)%del_lambda(6)*lg4x6(iconst)%ff
         r1(:) = pos(:, index_a)+imass1*dtsqby2*fc1(:)
         r2(:) = pos(:, index_b)+imass2*dtsqby2*fc2(:)
         r3(:) = pos(:, index_c)+imass3*dtsqby2*fc3(:)
         r4(:) = pos(:, index_d)+imass4*dtsqby2*fc4(:)
         v1(:) = vel(:, index_a)+imass1*dtby2*fc1(:)
         v2(:) = vel(:, index_b)+imass2*dtby2*fc2(:)
         v3(:) = vel(:, index_c)+imass3*dtby2*fc3(:)
         v4(:) = vel(:, index_d)+imass4*dtby2*fc4(:)
         r12 = r1-r2
         r13 = r1-r3
         r14 = r1-r4
         r23 = r2-r3
         r24 = r2-r4
         r34 = r3-r4
         ! compute the tolerance:
         sigma = DOT_PRODUCT(r12, r12)-g4x6_list(iconst)%dab* &
                 g4x6_list(iconst)%dab
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r13, r13)-g4x6_list(iconst)%dac* &
                 g4x6_list(iconst)%dac
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r14, r14)-g4x6_list(iconst)%dad* &
                 g4x6_list(iconst)%dad
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r23, r23)-g4x6_list(iconst)%dbc* &
                 g4x6_list(iconst)%dbc
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r24, r24)-g4x6_list(iconst)%dbd* &
                 g4x6_list(iconst)%dbd
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r34, r34)-g4x6_list(iconst)%dcd* &
                 g4x6_list(iconst)%dcd
         max_sigma = MAX(max_sigma, ABS(sigma))

         ! update positions with full multiplier
         pos(:, index_a) = r1(:)
         pos(:, index_b) = r2(:)
         pos(:, index_c) = r3(:)
         pos(:, index_d) = r4(:)

         ! update velocites with full multiplier
         vel(:, index_a) = v1(:)
         vel(:, index_b) = v2(:)
         vel(:, index_c) = v3(:)
         vel(:, index_d) = v4(:)
      END DO
   END SUBROUTINE shake_4x6_low

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param g4x6_list ...
!> \param lg4x6 ...
!> \param ng4x6 ...
!> \param first_atom ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param dt ...
!> \param ishake ...
!> \param max_sigma ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE shake_roll_4x6_low(fixd_list, g4x6_list, lg4x6, ng4x6, first_atom, &
                                 particle_set, pos, vel, r_shake, dt, ishake, max_sigma)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      INTEGER, INTENT(IN)                                :: ng4x6, first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: iconst, index_a, index_b, index_c, &
                                                            index_d
      REAL(KIND=dp)                                      :: dtby2, dtsqby2, idtsq, imass1, imass2, &
                                                            imass3, imass4, sigma
      REAL(KIND=dp), DIMENSION(3) :: f_roll1, f_roll2, f_roll3, f_roll4, f_roll5, f_roll6, fc1, &
         fc2, fc3, fc4, r0_12, r0_13, r0_14, r0_23, r0_24, r0_34, r1, r12, r13, r14, r2, r23, r24, &
         r3, r34, r4, v1, v2, v3, v4, vec
      REAL(KIND=dp), DIMENSION(6, 1)                     :: bvec
      REAL(KIND=dp), DIMENSION(6, 6)                     :: amat, atemp
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      dtsqby2 = dt*dt*.5_dp
      idtsq = 1.0_dp/dt/dt
      dtby2 = dt*.5_dp
      DO iconst = 1, ng4x6
         IF (g4x6_list(iconst)%restraint%active) CYCLE
         index_a = g4x6_list(iconst)%a+first_atom-1
         index_b = g4x6_list(iconst)%b+first_atom-1
         index_c = g4x6_list(iconst)%c+first_atom-1
         index_d = g4x6_list(iconst)%d+first_atom-1
         IF (ishake == 1) THEN
            r0_12(:) = pos(:, index_a)-pos(:, index_b)
            r0_13(:) = pos(:, index_a)-pos(:, index_c)
            r0_23(:) = pos(:, index_b)-pos(:, index_c)
            r0_14(:) = pos(:, index_a)-pos(:, index_d)
            r0_24(:) = pos(:, index_b)-pos(:, index_d)
            r0_34(:) = pos(:, index_c)-pos(:, index_d)
            atomic_kind => particle_set(index_a)%atomic_kind
            imass1 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_b)%atomic_kind
            imass2 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_c)%atomic_kind
            imass3 = 1.0_dp/atomic_kind%mass
            atomic_kind => particle_set(index_d)%atomic_kind
            imass4 = 1.0_dp/atomic_kind%mass
            lg4x6(iconst)%fa = -2.0_dp*(lg4x6(iconst)%ra_old- &
                                        lg4x6(iconst)%rb_old)
            lg4x6(iconst)%fb = -2.0_dp*(lg4x6(iconst)%ra_old- &
                                        lg4x6(iconst)%rc_old)
            lg4x6(iconst)%fc = -2.0_dp*(lg4x6(iconst)%ra_old- &
                                        lg4x6(iconst)%rd_old)
            lg4x6(iconst)%fd = -2.0_dp*(lg4x6(iconst)%rb_old- &
                                        lg4x6(iconst)%rc_old)
            lg4x6(iconst)%fe = -2.0_dp*(lg4x6(iconst)%rb_old- &
                                        lg4x6(iconst)%rd_old)
            lg4x6(iconst)%ff = -2.0_dp*(lg4x6(iconst)%rc_old- &
                                        lg4x6(iconst)%rd_old)
            ! Check for fixed atom constraints
            CALL check_fixed_atom_cns_g4x6(imass1, imass2, imass3, imass4, &
                                           index_a, index_b, index_c, index_d, fixd_list, lg4x6(iconst))
            ! rotate fconst:
            CALL matvec_3x3(f_roll1, r_shake, lg4x6(iconst)%fa)
            CALL matvec_3x3(f_roll2, r_shake, lg4x6(iconst)%fb)
            CALL matvec_3x3(f_roll3, r_shake, lg4x6(iconst)%fc)
            CALL matvec_3x3(f_roll4, r_shake, lg4x6(iconst)%fd)
            CALL matvec_3x3(f_roll5, r_shake, lg4x6(iconst)%fe)
            CALL matvec_3x3(f_roll6, r_shake, lg4x6(iconst)%ff)

            ! construct matrix
            amat(1, 1) = (imass1+imass2)*DOTPROD_3D(r0_12, f_roll1)
            amat(1, 2) = imass1*DOTPROD_3D(r0_12, f_roll2)
            amat(1, 3) = imass1*DOTPROD_3D(r0_12, f_roll3)
            amat(1, 4) = -imass2*DOTPROD_3D(r0_12, f_roll4)
            amat(1, 5) = -imass2*DOTPROD_3D(r0_12, f_roll5)
            amat(1, 6) = 0.0_dp

            amat(2, 1) = imass1*DOTPROD_3D(r0_13, f_roll1)
            amat(2, 2) = (imass1+imass3)*DOTPROD_3D(r0_13, f_roll2)
            amat(2, 3) = imass1*DOTPROD_3D(r0_13, f_roll3)
            amat(2, 4) = imass3*DOTPROD_3D(r0_13, f_roll4)
            amat(2, 5) = 0.0_dp
            amat(2, 6) = -imass3*DOTPROD_3D(r0_13, f_roll6)

            amat(3, 1) = imass1*DOTPROD_3D(r0_14, f_roll1)
            amat(3, 2) = imass1*DOTPROD_3D(r0_14, f_roll2)
            amat(3, 3) = (imass1+imass4)*DOTPROD_3D(r0_14, f_roll3)
            amat(3, 4) = 0.0_dp
            amat(3, 5) = imass4*DOTPROD_3D(r0_14, f_roll5)
            amat(3, 6) = imass4*DOTPROD_3D(r0_14, f_roll6)

            amat(4, 1) = -imass2*DOTPROD_3D(r0_23, f_roll1)
            amat(4, 2) = imass3*DOTPROD_3D(r0_23, f_roll2)
            amat(4, 3) = 0.0_dp
            amat(4, 4) = (imass3+imass2)*DOTPROD_3D(r0_23, f_roll4)
            amat(4, 5) = imass2*DOTPROD_3D(r0_23, f_roll5)
            amat(4, 6) = -imass3*DOTPROD_3D(r0_23, f_roll6)

            amat(5, 1) = -imass2*DOTPROD_3D(r0_24, f_roll1)
            amat(5, 2) = 0.0_dp
            amat(5, 3) = imass4*DOTPROD_3D(r0_24, f_roll3)
            amat(5, 4) = imass2*DOTPROD_3D(r0_24, f_roll4)
            amat(5, 5) = (imass4+imass2)*DOTPROD_3D(r0_24, f_roll5)
            amat(5, 6) = imass4*DOTPROD_3D(r0_24, f_roll6)

            amat(6, 1) = 0.0_dp
            amat(6, 2) = -imass3*DOTPROD_3D(r0_34, f_roll2)
            amat(6, 3) = imass4*DOTPROD_3D(r0_34, f_roll3)
            amat(6, 4) = -imass3*DOTPROD_3D(r0_34, f_roll4)
            amat(6, 5) = imass4*DOTPROD_3D(r0_34, f_roll5)
            amat(6, 6) = (imass3+imass4)*DOTPROD_3D(r0_34, f_roll6)
            ! Store values
            lg4x6(iconst)%r0_12 = r0_12
            lg4x6(iconst)%r0_13 = r0_13
            lg4x6(iconst)%r0_14 = r0_14
            lg4x6(iconst)%r0_23 = r0_23
            lg4x6(iconst)%r0_24 = r0_24
            lg4x6(iconst)%r0_34 = r0_34
            lg4x6(iconst)%f_roll1 = f_roll1
            lg4x6(iconst)%f_roll2 = f_roll2
            lg4x6(iconst)%f_roll3 = f_roll3
            lg4x6(iconst)%f_roll4 = f_roll4
            lg4x6(iconst)%f_roll5 = f_roll5
            lg4x6(iconst)%f_roll6 = f_roll6
            lg4x6(iconst)%amat = amat
            lg4x6(iconst)%imass1 = imass1
            lg4x6(iconst)%imass2 = imass2
            lg4x6(iconst)%imass3 = imass3
            lg4x6(iconst)%imass4 = imass4
            lg4x6(iconst)%lambda_old = 0.0_dp
            lg4x6(iconst)%del_lambda = 0.0_dp
         ELSE
            ! Retrieve values
            r0_12 = lg4x6(iconst)%r0_12
            r0_13 = lg4x6(iconst)%r0_13
            r0_14 = lg4x6(iconst)%r0_14
            r0_23 = lg4x6(iconst)%r0_23
            r0_24 = lg4x6(iconst)%r0_24
            r0_34 = lg4x6(iconst)%r0_34
            f_roll1 = lg4x6(iconst)%f_roll1
            f_roll2 = lg4x6(iconst)%f_roll2
            f_roll3 = lg4x6(iconst)%f_roll3
            f_roll4 = lg4x6(iconst)%f_roll4
            f_roll5 = lg4x6(iconst)%f_roll5
            f_roll6 = lg4x6(iconst)%f_roll6
            amat = lg4x6(iconst)%amat
            imass1 = lg4x6(iconst)%imass1
            imass2 = lg4x6(iconst)%imass2
            imass3 = lg4x6(iconst)%imass3
            imass4 = lg4x6(iconst)%imass4
         END IF

         ! Iterate until convergence:
         vec = lg4x6(iconst)%lambda(1)*f_roll1*(imass1+imass2)+ &
               lg4x6(iconst)%lambda(2)*imass1*f_roll2+ &
               lg4x6(iconst)%lambda(3)*imass1*f_roll3- &
               lg4x6(iconst)%lambda(4)*imass2*f_roll4- &
               lg4x6(iconst)%lambda(5)*imass2*f_roll5
         bvec(1, 1) = g4x6_list(iconst)%dab*g4x6_list(iconst)%dab &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_12, r0_12)
         vec = lg4x6(iconst)%lambda(2)*f_roll2*(imass1+imass3)+ &
               lg4x6(iconst)%lambda(1)*imass1*f_roll1+ &
               lg4x6(iconst)%lambda(3)*imass1*f_roll3+ &
               lg4x6(iconst)%lambda(4)*imass3*f_roll4- &
               lg4x6(iconst)%lambda(6)*imass3*f_roll6
         bvec(2, 1) = g4x6_list(iconst)%dac*g4x6_list(iconst)%dac &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_13, r0_13)
         vec = lg4x6(iconst)%lambda(3)*f_roll3*(imass1+imass4)+ &
               lg4x6(iconst)%lambda(1)*imass1*f_roll1+ &
               lg4x6(iconst)%lambda(2)*imass1*f_roll2+ &
               lg4x6(iconst)%lambda(5)*imass4*f_roll5+ &
               lg4x6(iconst)%lambda(6)*imass4*f_roll6
         bvec(3, 1) = g4x6_list(iconst)%dad*g4x6_list(iconst)%dad &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_14, r0_14)
         vec = lg4x6(iconst)%lambda(4)*f_roll4*(imass2+imass3)- &
               lg4x6(iconst)%lambda(1)*imass2*f_roll1+ &
               lg4x6(iconst)%lambda(2)*imass3*f_roll2+ &
               lg4x6(iconst)%lambda(5)*imass2*f_roll5- &
               lg4x6(iconst)%lambda(6)*imass3*f_roll6
         bvec(4, 1) = g4x6_list(iconst)%dbc*g4x6_list(iconst)%dbc &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_23, r0_23)
         vec = lg4x6(iconst)%lambda(5)*f_roll5*(imass2+imass4)- &
               lg4x6(iconst)%lambda(1)*imass2*f_roll1+ &
               lg4x6(iconst)%lambda(3)*imass4*f_roll3+ &
               lg4x6(iconst)%lambda(4)*imass2*f_roll4+ &
               lg4x6(iconst)%lambda(6)*imass4*f_roll6
         bvec(5, 1) = g4x6_list(iconst)%dbd*g4x6_list(iconst)%dbd &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_24, r0_24)
         vec = lg4x6(iconst)%lambda(6)*f_roll6*(imass3+imass4)- &
               lg4x6(iconst)%lambda(2)*imass3*f_roll2+ &
               lg4x6(iconst)%lambda(3)*imass4*f_roll3- &
               lg4x6(iconst)%lambda(4)*imass3*f_roll4+ &
               lg4x6(iconst)%lambda(5)*imass4*f_roll5
         bvec(6, 1) = g4x6_list(iconst)%dcd*g4x6_list(iconst)%dcd &
                      -dtsqby2*dtsqby2*DOTPROD_3D(vec, vec)-DOTPROD_3D(r0_34, r0_34)
         bvec = bvec*idtsq

         ! get lambda
         atemp = amat
         CALL solve_system(atemp, 6, bvec)
         lg4x6(iconst)%lambda(:) = bvec(:, 1)
         lg4x6(iconst)%del_lambda(:) = lg4x6(iconst)%lambda(:)- &
                                       lg4x6(iconst)%lambda_old(:)
         lg4x6(iconst)%lambda_old(:) = lg4x6(iconst)%lambda(:)

         fc1 = lg4x6(iconst)%del_lambda(1)*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%del_lambda(2)*lg4x6(iconst)%fb+ &
               lg4x6(iconst)%del_lambda(3)*lg4x6(iconst)%fc
         fc2 = -lg4x6(iconst)%del_lambda(1)*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%del_lambda(4)*lg4x6(iconst)%fd+ &
               lg4x6(iconst)%del_lambda(5)*lg4x6(iconst)%fe
         fc3 = -lg4x6(iconst)%del_lambda(2)*lg4x6(iconst)%fb- &
               lg4x6(iconst)%del_lambda(4)*lg4x6(iconst)%fd+ &
               lg4x6(iconst)%del_lambda(6)*lg4x6(iconst)%ff
         fc4 = -lg4x6(iconst)%del_lambda(3)*lg4x6(iconst)%fc- &
               lg4x6(iconst)%del_lambda(5)*lg4x6(iconst)%fe- &
               lg4x6(iconst)%del_lambda(6)*lg4x6(iconst)%ff
         CALL MATVEC_3x3(vec, r_shake, fc1)
         r1(:) = pos(:, index_a)+imass1*dtsqby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc2)
         r2(:) = pos(:, index_b)+imass2*dtsqby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc3)
         r3(:) = pos(:, index_c)+imass3*dtsqby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc4)
         r4(:) = pos(:, index_d)+imass4*dtsqby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc1)
         v1(:) = vel(:, index_a)+imass1*dtby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc2)
         v2(:) = vel(:, index_b)+imass2*dtby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc3)
         v3(:) = vel(:, index_c)+imass3*dtby2*vec
         CALL MATVEC_3x3(vec, r_shake, fc4)
         v4(:) = vel(:, index_d)+imass4*dtby2*vec

         r12 = r1-r2
         r13 = r1-r3
         r23 = r2-r3
         r14 = r1-r4
         r24 = r2-r4
         r34 = r3-r4
         ! compute the tolerance:
         sigma = DOT_PRODUCT(r12, r12)-g4x6_list(iconst)%dab* &
                 g4x6_list(iconst)%dab
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r13, r13)-g4x6_list(iconst)%dac* &
                 g4x6_list(iconst)%dac
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r14, r14)-g4x6_list(iconst)%dad* &
                 g4x6_list(iconst)%dad
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r23, r23)-g4x6_list(iconst)%dbc* &
                 g4x6_list(iconst)%dbc
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r24, r24)-g4x6_list(iconst)%dbd* &
                 g4x6_list(iconst)%dbd
         max_sigma = MAX(max_sigma, ABS(sigma))
         sigma = DOT_PRODUCT(r34, r34)-g4x6_list(iconst)%dcd* &
                 g4x6_list(iconst)%dcd
         max_sigma = MAX(max_sigma, ABS(sigma))

         ! update positions with full multiplier
         pos(:, index_a) = r1(:)
         pos(:, index_b) = r2(:)
         pos(:, index_c) = r3(:)
         pos(:, index_d) = r4(:)

         ! update velocites with full multiplier
         vel(:, index_a) = v1(:)
         vel(:, index_b) = v2(:)
         vel(:, index_c) = v3(:)
         vel(:, index_d) = v4(:)
      END DO
   END SUBROUTINE shake_roll_4x6_low

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param g4x6_list ...
!> \param lg4x6 ...
!> \param first_atom ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_4x6_low(fixd_list, g4x6_list, lg4x6, first_atom, &
                             particle_set, vel, dt)

      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      INTEGER, INTENT(IN)                                :: first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt

      INTEGER                                            :: iconst, index_a, index_b, index_c, &
                                                            index_d
      REAL(KIND=dp)                                      :: dtby2, idt, imass1, imass2, imass3, &
                                                            imass4, mass
      REAL(KIND=dp), DIMENSION(3)                        :: fc1, fc2, fc3, fc4, r12, r13, r14, r23, &
                                                            r24, r34, v12, v13, v14, v23, v24, v34
      REAL(KIND=dp), DIMENSION(6, 1)                     :: bvec
      REAL(KIND=dp), DIMENSION(6, 6)                     :: amat
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      idt = 1.0_dp/dt
      dtby2 = dt*.5_dp
      DO iconst = 1, SIZE(g4x6_list)
         IF (g4x6_list(iconst)%restraint%active) CYCLE
         index_a = g4x6_list(iconst)%a+first_atom-1
         index_b = g4x6_list(iconst)%b+first_atom-1
         index_c = g4x6_list(iconst)%c+first_atom-1
         index_d = g4x6_list(iconst)%d+first_atom-1
         v12(:) = vel(:, index_a)-vel(:, index_b)
         v13(:) = vel(:, index_a)-vel(:, index_c)
         v14(:) = vel(:, index_a)-vel(:, index_d)
         v23(:) = vel(:, index_b)-vel(:, index_c)
         v24(:) = vel(:, index_b)-vel(:, index_d)
         v34(:) = vel(:, index_c)-vel(:, index_d)

         r12(:) = particle_set(index_a)%r(:)-particle_set(index_b)%r(:)
         r13(:) = particle_set(index_a)%r(:)-particle_set(index_c)%r(:)
         r14(:) = particle_set(index_a)%r(:)-particle_set(index_d)%r(:)
         r23(:) = particle_set(index_b)%r(:)-particle_set(index_c)%r(:)
         r24(:) = particle_set(index_b)%r(:)-particle_set(index_d)%r(:)
         r34(:) = particle_set(index_c)%r(:)-particle_set(index_d)%r(:)
         atomic_kind => particle_set(index_a)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass1 = 1.0_dp/mass
         atomic_kind => particle_set(index_b)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass2 = 1.0_dp/mass
         atomic_kind => particle_set(index_c)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass3 = 1.0_dp/mass
         atomic_kind => particle_set(index_d)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass4 = 1.0_dp/mass
         lg4x6(iconst)%fa = -2.0_dp*r12
         lg4x6(iconst)%fb = -2.0_dp*r13
         lg4x6(iconst)%fc = -2.0_dp*r14
         lg4x6(iconst)%fd = -2.0_dp*r23
         lg4x6(iconst)%fe = -2.0_dp*r24
         lg4x6(iconst)%ff = -2.0_dp*r34
         ! Check for fixed atom constraints
         CALL check_fixed_atom_cns_g4x6(imass1, imass2, imass3, imass4, &
                                        index_a, index_b, index_c, index_d, fixd_list, lg4x6(iconst))
         ! construct matrix
         amat(1, 1) = (imass1+imass2)*DOTPROD_3D(r12, lg4x6(iconst)%fa)
         amat(1, 2) = imass1*DOTPROD_3D(r12, lg4x6(iconst)%fb)
         amat(1, 3) = imass1*DOTPROD_3D(r12, lg4x6(iconst)%fc)
         amat(1, 4) = -imass2*DOTPROD_3D(r12, lg4x6(iconst)%fd)
         amat(1, 5) = -imass2*DOTPROD_3D(r12, lg4x6(iconst)%fe)
         amat(1, 6) = 0.0_dp

         amat(2, 1) = imass1*DOTPROD_3D(r13, lg4x6(iconst)%fa)
         amat(2, 2) = (imass1+imass3)*DOTPROD_3D(r13, lg4x6(iconst)%fb)
         amat(2, 3) = imass1*DOTPROD_3D(r13, lg4x6(iconst)%fc)
         amat(2, 4) = imass3*DOTPROD_3D(r13, lg4x6(iconst)%fd)
         amat(2, 5) = 0.0_dp
         amat(2, 6) = -imass3*DOTPROD_3D(r13, lg4x6(iconst)%ff)

         amat(3, 1) = imass1*DOTPROD_3D(r14, lg4x6(iconst)%fa)
         amat(3, 2) = imass1*DOTPROD_3D(r14, lg4x6(iconst)%fb)
         amat(3, 3) = (imass1+imass4)*DOTPROD_3D(r14, lg4x6(iconst)%fc)
         amat(3, 4) = 0.0_dp
         amat(3, 5) = imass4*DOTPROD_3D(r14, lg4x6(iconst)%fe)
         amat(3, 6) = imass4*DOTPROD_3D(r14, lg4x6(iconst)%ff)

         amat(4, 1) = -imass2*DOTPROD_3D(r23, lg4x6(iconst)%fa)
         amat(4, 2) = imass3*DOTPROD_3D(r23, lg4x6(iconst)%fb)
         amat(4, 3) = 0.0_dp
         amat(4, 4) = (imass3+imass2)*DOTPROD_3D(r23, lg4x6(iconst)%fd)
         amat(4, 5) = imass2*DOTPROD_3D(r23, lg4x6(iconst)%fe)
         amat(4, 6) = -imass3*DOTPROD_3D(r23, lg4x6(iconst)%ff)

         amat(5, 1) = -imass2*DOTPROD_3D(r24, lg4x6(iconst)%fa)
         amat(5, 2) = 0.0_dp
         amat(5, 3) = imass4*DOTPROD_3D(r24, lg4x6(iconst)%fc)
         amat(5, 4) = imass2*DOTPROD_3D(r24, lg4x6(iconst)%fd)
         amat(5, 5) = (imass4+imass2)*DOTPROD_3D(r24, lg4x6(iconst)%fe)
         amat(5, 6) = imass4*DOTPROD_3D(r24, lg4x6(iconst)%ff)

         amat(6, 1) = 0.0_dp
         amat(6, 2) = -imass3*DOTPROD_3D(r34, lg4x6(iconst)%fb)
         amat(6, 3) = imass4*DOTPROD_3D(r34, lg4x6(iconst)%fc)
         amat(6, 4) = -imass3*DOTPROD_3D(r34, lg4x6(iconst)%fd)
         amat(6, 5) = imass4*DOTPROD_3D(r34, lg4x6(iconst)%fe)
         amat(6, 6) = (imass3+imass4)*DOTPROD_3D(r34, lg4x6(iconst)%ff)

         ! construct solution vector
         bvec(1, 1) = DOTPROD_3D(r12, v12)
         bvec(2, 1) = DOTPROD_3D(r13, v13)
         bvec(3, 1) = DOTPROD_3D(r14, v14)
         bvec(4, 1) = DOTPROD_3D(r23, v23)
         bvec(5, 1) = DOTPROD_3D(r24, v24)
         bvec(6, 1) = DOTPROD_3D(r34, v34)
         bvec = -bvec*2.0_dp*idt

         ! get lambda
         CALL solve_system(amat, 6, bvec)
         lg4x6(iconst)%lambda(:) = bvec(:, 1)

         fc1 = lg4x6(iconst)%lambda(1)*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%lambda(2)*lg4x6(iconst)%fb+ &
               lg4x6(iconst)%lambda(3)*lg4x6(iconst)%fc
         fc2 = -lg4x6(iconst)%lambda(1)*lg4x6(iconst)%fa+ &
               lg4x6(iconst)%lambda(4)*lg4x6(iconst)%fd+ &
               lg4x6(iconst)%lambda(5)*lg4x6(iconst)%fe
         fc3 = -lg4x6(iconst)%lambda(2)*lg4x6(iconst)%fb- &
               lg4x6(iconst)%lambda(4)*lg4x6(iconst)%fd+ &
               lg4x6(iconst)%lambda(6)*lg4x6(iconst)%ff
         fc4 = -lg4x6(iconst)%lambda(3)*lg4x6(iconst)%fc- &
               lg4x6(iconst)%lambda(5)*lg4x6(iconst)%fe- &
               lg4x6(iconst)%lambda(6)*lg4x6(iconst)%ff
         vel(:, index_a) = vel(:, index_a)+imass1*dtby2*fc1(:)
         vel(:, index_b) = vel(:, index_b)+imass2*dtby2*fc2(:)
         vel(:, index_c) = vel(:, index_c)+imass3*dtby2*fc3(:)
         vel(:, index_d) = vel(:, index_d)+imass4*dtby2*fc4(:)
      END DO
   END SUBROUTINE rattle_4x6_low

! **************************************************************************************************
!> \brief ...
!> \param fixd_list ...
!> \param g4x6_list ...
!> \param lg4x6 ...
!> \param first_atom ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param veps ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE rattle_roll_4x6_low(fixd_list, g4x6_list, lg4x6, first_atom, &
                                  particle_set, vel, r_rattle, dt, veps)

      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(g4x6_constraint_type), POINTER                :: g4x6_list(:)
      TYPE(local_g4x6_constraint_type), POINTER          :: lg4x6(:)
      INTEGER, INTENT(IN)                                :: first_atom
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_rattle
      REAL(kind=dp), INTENT(in)                          :: dt
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: veps

      INTEGER                                            :: iconst, index_a, index_b, index_c, &
                                                            index_d
      REAL(KIND=dp)                                      :: dtby2, idt, imass1, imass2, imass3, &
                                                            imass4, mass
      REAL(KIND=dp), DIMENSION(3)                        :: f_roll1, f_roll2, f_roll3, f_roll4, &
                                                            f_roll5, f_roll6, fc1, fc2, fc3, fc4, &
                                                            r12, r13, r14, r23, r24, r34, v12, &
                                                            v13, v14, v23, v24, v34, vec
      REAL(KIND=dp), DIMENSION(6)                        :: lambda
      REAL(KIND=dp), DIMENSION(6, 1)                     :: bvec
      REAL(KIND=dp), DIMENSION(6, 6)                     :: amat
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      idt = 1.0_dp/dt
      dtby2 = dt*.5_dp
      DO iconst = 1, SIZE(g4x6_list)
         IF (g4x6_list(iconst)%restraint%active) CYCLE
         index_a = g4x6_list(iconst)%a+first_atom-1
         index_b = g4x6_list(iconst)%b+first_atom-1
         index_c = g4x6_list(iconst)%c+first_atom-1
         index_d = g4x6_list(iconst)%d+first_atom-1
         v12(:) = vel(:, index_a)-vel(:, index_b)
         v13(:) = vel(:, index_a)-vel(:, index_c)
         v14(:) = vel(:, index_a)-vel(:, index_d)
         v23(:) = vel(:, index_b)-vel(:, index_c)
         v24(:) = vel(:, index_b)-vel(:, index_d)
         v34(:) = vel(:, index_c)-vel(:, index_d)

         r12(:) = particle_set(index_a)%r(:)-particle_set(index_b)%r(:)
         r13(:) = particle_set(index_a)%r(:)-particle_set(index_c)%r(:)
         r14(:) = particle_set(index_a)%r(:)-particle_set(index_d)%r(:)
         r23(:) = particle_set(index_b)%r(:)-particle_set(index_c)%r(:)
         r24(:) = particle_set(index_b)%r(:)-particle_set(index_d)%r(:)
         r34(:) = particle_set(index_c)%r(:)-particle_set(index_d)%r(:)
         atomic_kind => particle_set(index_a)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass1 = 1.0_dp/mass
         atomic_kind => particle_set(index_b)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass2 = 1.0_dp/mass
         atomic_kind => particle_set(index_c)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass3 = 1.0_dp/mass
         atomic_kind => particle_set(index_d)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass4 = 1.0_dp/mass
         lg4x6(iconst)%fa = -2.0_dp*r12
         lg4x6(iconst)%fb = -2.0_dp*r13
         lg4x6(iconst)%fc = -2.0_dp*r14
         lg4x6(iconst)%fd = -2.0_dp*r23
         lg4x6(iconst)%fe = -2.0_dp*r24
         lg4x6(iconst)%ff = -2.0_dp*r34
         ! Check for fixed atom constraints
         CALL check_fixed_atom_cns_g4x6(imass1, imass2, imass3, imass4, &
                                        index_a, index_b, index_c, index_d, fixd_list, lg4x6(iconst))
         ! roll the fc
         CALL MATVEC_3x3(f_roll1, r_rattle, lg4x6(iconst)%fa)
         CALL MATVEC_3x3(f_roll2, r_rattle, lg4x6(iconst)%fb)
         CALL MATVEC_3x3(f_roll3, r_rattle, lg4x6(iconst)%fc)
         CALL MATVEC_3x3(f_roll4, r_rattle, lg4x6(iconst)%fd)
         CALL MATVEC_3x3(f_roll5, r_rattle, lg4x6(iconst)%fe)
         CALL MATVEC_3x3(f_roll6, r_rattle, lg4x6(iconst)%ff)
         ! construct matrix
         amat(1, 1) = (imass1+imass2)*DOTPROD_3D(r12, f_roll1)
         amat(1, 2) = imass1*DOTPROD_3D(r12, f_roll2)
         amat(1, 3) = imass1*DOTPROD_3D(r12, f_roll3)
         amat(1, 4) = -imass2*DOTPROD_3D(r12, f_roll4)
         amat(1, 5) = -imass2*DOTPROD_3D(r12, f_roll5)
         amat(1, 6) = 0.0_dp

         amat(2, 1) = imass1*DOTPROD_3D(r13, f_roll1)
         amat(2, 2) = (imass1+imass3)*DOTPROD_3D(r13, f_roll2)
         amat(2, 3) = imass1*DOTPROD_3D(r13, f_roll3)
         amat(2, 4) = imass3*DOTPROD_3D(r13, f_roll4)
         amat(2, 5) = 0.0_dp
         amat(2, 6) = -imass3*DOTPROD_3D(r13, f_roll6)

         amat(3, 1) = imass1*DOTPROD_3D(r14, f_roll1)
         amat(3, 2) = imass1*DOTPROD_3D(r14, f_roll2)
         amat(3, 3) = (imass1+imass4)*DOTPROD_3D(r14, f_roll3)
         amat(3, 4) = 0.0_dp
         amat(3, 5) = imass4*DOTPROD_3D(r14, f_roll5)
         amat(3, 6) = imass4*DOTPROD_3D(r14, f_roll6)

         amat(4, 1) = -imass2*DOTPROD_3D(r23, f_roll1)
         amat(4, 2) = imass3*DOTPROD_3D(r23, f_roll2)
         amat(4, 3) = 0.0_dp
         amat(4, 4) = (imass3+imass2)*DOTPROD_3D(r23, f_roll4)
         amat(4, 5) = imass2*DOTPROD_3D(r23, f_roll5)
         amat(4, 6) = -imass3*DOTPROD_3D(r23, f_roll6)

         amat(5, 1) = -imass2*DOTPROD_3D(r24, f_roll1)
         amat(5, 2) = 0.0_dp
         amat(5, 3) = imass4*DOTPROD_3D(r24, f_roll3)
         amat(5, 4) = imass2*DOTPROD_3D(r24, f_roll4)
         amat(5, 5) = (imass4+imass2)*DOTPROD_3D(r24, f_roll5)
         amat(5, 6) = imass4*DOTPROD_3D(r24, f_roll6)

         amat(6, 1) = 0.0_dp
         amat(6, 2) = -imass3*DOTPROD_3D(r34, f_roll2)
         amat(6, 3) = imass4*DOTPROD_3D(r34, f_roll3)
         amat(6, 4) = -imass3*DOTPROD_3D(r34, f_roll4)
         amat(6, 5) = imass4*DOTPROD_3D(r34, f_roll5)
         amat(6, 6) = (imass3+imass4)*DOTPROD_3D(r34, f_roll6)

         ! construct solution vector
         CALL matvec_3x3(vec, veps, r12)
         bvec(1, 1) = DOTPROD_3D(r12, v12+vec)
         CALL matvec_3x3(vec, veps, r13)
         bvec(2, 1) = DOTPROD_3D(r13, v13+vec)
         CALL matvec_3x3(vec, veps, r14)
         bvec(3, 1) = DOTPROD_3D(r14, v14+vec)
         CALL matvec_3x3(vec, veps, r23)
         bvec(4, 1) = DOTPROD_3D(r23, v23+vec)
         CALL matvec_3x3(vec, veps, r24)
         bvec(5, 1) = DOTPROD_3D(r24, v24+vec)
         CALL matvec_3x3(vec, veps, r34)
         bvec(6, 1) = DOTPROD_3D(r34, v34+vec)
         bvec = -bvec*2.0_dp*idt

         ! get lambda
         CALL solve_system(amat, 6, bvec)
         lambda(:) = bvec(:, 1)
         lg4x6(iconst)%lambda(:) = lambda

         fc1 = lambda(1)*f_roll1+ &
               lambda(2)*f_roll2+ &
               lambda(3)*f_roll3
         fc2 = -lambda(1)*f_roll1+ &
               lambda(4)*f_roll4+ &
               lambda(5)*f_roll5
         fc3 = -lambda(2)*f_roll2- &
               lambda(4)*f_roll4+ &
               lambda(6)*f_roll6
         fc4 = -lambda(3)*f_roll3- &
               lambda(5)*f_roll5- &
               lambda(6)*f_roll6
         vel(:, index_a) = vel(:, index_a)+imass1*dtby2*fc1(:)
         vel(:, index_b) = vel(:, index_b)+imass2*dtby2*fc2(:)
         vel(:, index_c) = vel(:, index_c)+imass3*dtby2*fc3(:)
         vel(:, index_d) = vel(:, index_d)+imass4*dtby2*fc4(:)
      END DO
   END SUBROUTINE rattle_roll_4x6_low

END MODULE constraint_4x6
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module that handles the COLLECTIVE constraints
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
MODULE constraint_clv
   USE cell_types,                      ONLY: cell_type
   USE colvar_methods,                  ONLY: colvar_eval_mol_f
   USE colvar_types,                    ONLY: colvar_type,&
                                              diff_colvar
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: matvec_3x3
   USE molecule_kind_types,             ONLY: colvar_constraint_type,&
                                              fixd_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              local_colvar_constraint_type,&
                                              molecule_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: shake_roll_colv_int, &
             rattle_roll_colv_int, &
             shake_colv_int, &
             rattle_colv_int, &
             shake_roll_colv_ext, &
             rattle_roll_colv_ext, &
             shake_colv_ext, &
             rattle_colv_ext, &
             shake_update_colv_int, &
             shake_update_colv_ext

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'constraint_clv'

CONTAINS

! **************************************************************************************************
!> \brief Intramolecular subroutine
!>      shake_colv algorithm for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param molecule ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE shake_colv_int(molecule, particle_set, pos, vel, dt, ishake, &
                             cell, imass, max_sigma)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      CHARACTER(len=*), PARAMETER :: routineN = 'shake_colv_int', routineP = moduleN//':'//routineN

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, colv_list=colv_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, lcolv=lcolv)
      ! Real Shake
      CALL shake_colv_low(fixd_list, colv_list, lcolv, &
                          particle_set, pos, vel, dt, ishake, cell, imass, max_sigma)

   END SUBROUTINE shake_colv_int

! **************************************************************************************************
!> \brief Intramolecular subroutine for updating the TARGET value of collective
!>        constraints
!> \param molecule ...
!> \param dt ...
!> \param motion_section ...
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE shake_update_colv_int(molecule, dt, motion_section)

      TYPE(molecule_type), POINTER                       :: molecule
      REAL(kind=dp), INTENT(in)                          :: dt
      TYPE(section_vals_type), POINTER                   :: motion_section

      CHARACTER(len=*), PARAMETER :: routineN = 'shake_update_colv_int', &
         routineP = moduleN//':'//routineN

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, colv_list=colv_list)
      ! Real update of the Shake target
      CALL shake_update_colv_low(colv_list, dt, motion_section)

   END SUBROUTINE shake_update_colv_int

! **************************************************************************************************
!> \brief Intramolecular subroutine
!>      rattle algorithm for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param molecule ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \param irattle ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE rattle_colv_int(molecule, particle_set, vel, dt, irattle, &
                              cell, imass, max_sigma)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: irattle
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, colv_list=colv_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, lcolv=lcolv)
      ! Real Rattle
      CALL rattle_colv_low(fixd_list, colv_list, lcolv, &
                           particle_set, vel, dt, irattle, cell, imass, max_sigma)

   END SUBROUTINE rattle_colv_int

! **************************************************************************************************
!> \brief Intramolecular subroutine
!>      shake algorithm (box allowed to change) for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param molecule ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param v_shake ...
!> \param dt ...
!> \param ishake ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE shake_roll_colv_int(molecule, particle_set, pos, vel, r_shake, v_shake, &
                                  dt, ishake, cell, imass, max_sigma)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake, v_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(in)                                :: ishake
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, colv_list=colv_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, lcolv=lcolv)
      ! Real Shake
      CALL shake_roll_colv_low(fixd_list, colv_list, lcolv, &
                               particle_set, pos, vel, r_shake, v_shake, dt, ishake, cell, &
                               imass, max_sigma)

   END SUBROUTINE shake_roll_colv_int

! **************************************************************************************************
!> \brief Intramolecular subroutine
!>      rattle algorithm (box allowed to change) for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param molecule ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param irattle ...
!> \param veps ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE rattle_roll_colv_int(molecule, particle_set, vel, r_rattle, &
                                   dt, irattle, veps, cell, imass, max_sigma)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), INTENT(IN)                          :: r_rattle(:, :), dt
      INTEGER, INTENT(in)                                :: irattle
      REAL(KIND=dp), INTENT(IN)                          :: veps(:, :)
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)
      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, colv_list=colv_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, lcolv=lcolv)
      ! Real Rattle
      CALL rattle_roll_colv_low(fixd_list, colv_list, lcolv, &
                                particle_set, vel, r_rattle, dt, irattle, veps, cell, &
                                imass, max_sigma)

   END SUBROUTINE rattle_roll_colv_int

! **************************************************************************************************
!> \brief Intermolecular subroutine
!>      shake_colv algorithm for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param gci ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE shake_colv_ext(gci, particle_set, pos, vel, dt, ishake, &
                             cell, imass, max_sigma)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      CHARACTER(len=*), PARAMETER :: routineN = 'shake_colv_ext', routineP = moduleN//':'//routineN

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)

      colv_list => gci%colv_list
      fixd_list => gci%fixd_list
      lcolv => gci%lcolv
      ! Real Shake
      CALL shake_colv_low(fixd_list, colv_list, lcolv, &
                          particle_set, pos, vel, dt, ishake, cell, imass, max_sigma)

   END SUBROUTINE shake_colv_ext

! **************************************************************************************************
!> \brief Intermolecular subroutine for updating the TARGET value for collective
!>        constraints
!> \param gci ...
!> \param dt ...
!> \param motion_section ...
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE shake_update_colv_ext(gci, dt, motion_section)

      TYPE(global_constraint_type), POINTER              :: gci
      REAL(kind=dp), INTENT(in)                          :: dt
      TYPE(section_vals_type), POINTER                   :: motion_section

      CHARACTER(len=*), PARAMETER :: routineN = 'shake_update_colv_ext', &
         routineP = moduleN//':'//routineN

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)

      colv_list => gci%colv_list
      ! Real update of the Shake target
      CALL shake_update_colv_low(colv_list, dt, motion_section)

   END SUBROUTINE shake_update_colv_ext

! **************************************************************************************************
!> \brief Intermolecular subroutine
!>      rattle algorithm for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param gci ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \param irattle ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE rattle_colv_ext(gci, particle_set, vel, dt, irattle, &
                              cell, imass, max_sigma)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: irattle
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)

      colv_list => gci%colv_list
      fixd_list => gci%fixd_list
      lcolv => gci%lcolv
      ! Real Rattle
      CALL rattle_colv_low(fixd_list, colv_list, lcolv, &
                           particle_set, vel, dt, irattle, cell, imass, max_sigma)

   END SUBROUTINE rattle_colv_ext

! **************************************************************************************************
!> \brief Intermolecular subroutine
!>      shake algorithm (box allowed to change) for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param gci ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param v_shake ...
!> \param dt ...
!> \param ishake ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE shake_roll_colv_ext(gci, particle_set, pos, vel, r_shake, v_shake, &
                                  dt, ishake, cell, imass, max_sigma)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake, v_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(in)                                :: ishake
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)

      colv_list => gci%colv_list
      fixd_list => gci%fixd_list
      lcolv => gci%lcolv
      ! Real Shake
      CALL shake_roll_colv_low(fixd_list, colv_list, lcolv, &
                               particle_set, pos, vel, r_shake, v_shake, dt, ishake, cell, &
                               imass, max_sigma)

   END SUBROUTINE shake_roll_colv_ext

! **************************************************************************************************
!> \brief Intermolecular subroutine
!>      rattle algorithm (box allowed to change) for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param gci ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param irattle ...
!> \param veps ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE rattle_roll_colv_ext(gci, particle_set, vel, r_rattle, &
                                   dt, irattle, veps, cell, imass, max_sigma)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), INTENT(IN)                          :: r_rattle(:, :), dt
      INTEGER, INTENT(in)                                :: irattle
      REAL(KIND=dp), INTENT(IN)                          :: veps(:, :)
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)

      colv_list => gci%colv_list
      fixd_list => gci%fixd_list
      lcolv => gci%lcolv
      ! Real Rattle
      CALL rattle_roll_colv_low(fixd_list, colv_list, lcolv, &
                                particle_set, vel, r_rattle, dt, irattle, veps, cell, &
                                imass, max_sigma)

   END SUBROUTINE rattle_roll_colv_ext

! **************************************************************************************************
!> \brief Real Shake subroutine - Low Level
!>      shake_colv algorithm for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param fixd_list ...
!> \param colv_list ...
!> \param lcolv ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param ishake ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE shake_colv_low(fixd_list, colv_list, lcolv, &
                             particle_set, pos, vel, dt, ishake, cell, imass, max_sigma)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: ishake
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      CHARACTER(len=*), PARAMETER :: routineN = 'shake_colv_low', routineP = moduleN//':'//routineN

      INTEGER                                            :: iconst
      REAL(KIND=dp)                                      :: del_lam, dtby2, dtsqby2, fdotf_sum

      dtsqby2 = dt*dt*.5_dp
      dtby2 = dt*.5_dp
      IF (ishake == 1) THEN
         DO iconst = 1, SIZE(colv_list)
            IF (colv_list(iconst)%restraint%active) CYCLE
            ! Update positions
            CALL update_con_colv(pos, dtsqby2, lcolv(iconst), &
                                 lambda=lcolv(iconst)%lambda, &
                                 imass=imass)
            ! Update velocities
            CALL update_con_colv(vel, dtby2, lcolv(iconst), &
                                 lambda=lcolv(iconst)%lambda, &
                                 imass=imass)
         END DO
      ELSE
         DO iconst = 1, SIZE(colv_list)
            IF (colv_list(iconst)%restraint%active) CYCLE
            ! Update colvar
            CALL colvar_eval_mol_f(lcolv(iconst)%colvar, cell, particles=particle_set, &
                                   pos=pos, fixd_list=fixd_list)
            lcolv(iconst)%sigma = diff_colvar(lcolv(iconst)%colvar, &
                                              colv_list(iconst)%expected_value)
            fdotf_sum = eval_Jac_colvar(lcolv(iconst)%colvar, &
                                        lcolv(iconst)%colvar_old, imass=imass)
            del_lam = 2.0_dp*lcolv(iconst)%sigma/(dt*dt*fdotf_sum)
            lcolv(iconst)%lambda = lcolv(iconst)%lambda+del_lam

            ! Update positions
            CALL update_con_colv(pos, dtsqby2, lcolv(iconst), &
                                 lambda=del_lam, &
                                 imass=imass)
            ! Update velocities
            CALL update_con_colv(vel, dtby2, lcolv(iconst), &
                                 lambda=del_lam, &
                                 imass=imass)
         END DO
      END IF
      ! computing the constraint and value of tolerance
      DO iconst = 1, SIZE(colv_list)
         IF (colv_list(iconst)%restraint%active) CYCLE
         CALL colvar_eval_mol_f(lcolv(iconst)%colvar, cell, particles=particle_set, &
                                pos=pos, fixd_list=fixd_list)
         lcolv(iconst)%sigma = diff_colvar(lcolv(iconst)%colvar, &
                                           colv_list(iconst)%expected_value)
         max_sigma = MAX(ABS(lcolv(iconst)%sigma), max_sigma)
      END DO
   END SUBROUTINE shake_colv_low

! **************************************************************************************************
!> \brief Real Shake subroutine - Low Level - for updating the TARGET value
!> \param colv_list ...
!> \param dt ...
!> \param motion_section ...
!> \date 02.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE shake_update_colv_low(colv_list, dt, motion_section)
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      REAL(kind=dp), INTENT(in)                          :: dt
      TYPE(section_vals_type), POINTER                   :: motion_section

      CHARACTER(len=*), PARAMETER :: routineN = 'shake_update_colv_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iconst, irep, n_rep
      LOGICAL                                            :: do_update_colvar, explicit
      REAL(KIND=dp)                                      :: clv_target, limit, new_clv_target, value
      TYPE(section_vals_type), POINTER                   :: collective_sections

! Update globally for restart

      collective_sections => section_vals_get_subs_vals(motion_section, "CONSTRAINT%COLLECTIVE")
      CALL section_vals_get(collective_sections, n_repetition=n_rep)
      IF (n_rep /= 0) THEN
         DO irep = 1, n_rep
            CALL section_vals_val_get(collective_sections, "TARGET_GROWTH", r_val=value, &
                                      i_rep_section=irep)
            IF (value /= 0.0_dp) THEN
               CALL section_vals_val_get(collective_sections, "TARGET", r_val=clv_target, &
                                         i_rep_section=irep)
               new_clv_target = clv_target+value*dt
               ! Check limits..
               CALL section_vals_val_get(collective_sections, "TARGET_LIMIT", explicit=explicit, &
                                         i_rep_section=irep)
               do_update_colvar = .TRUE.
               IF (explicit) THEN
                  CALL section_vals_val_get(collective_sections, "TARGET_LIMIT", r_val=limit, &
                                            i_rep_section=irep)
                  IF (value > 0.0_dp) THEN
                     IF (clv_target == limit) THEN
                        do_update_colvar = .FALSE.
                     ELSE IF (new_clv_target >= limit) THEN
                        new_clv_target = limit
                     END IF
                  ELSE
                     IF (clv_target == limit) THEN
                        do_update_colvar = .FALSE.
                     ELSE IF (new_clv_target <= limit) THEN
                        new_clv_target = limit
                     END IF
                  ENDIF
               END IF
               IF (do_update_colvar) THEN
                  CALL section_vals_val_set(collective_sections, "TARGET", r_val=new_clv_target, &
                                            i_rep_section=irep)
               END IF
            END IF
         END DO
      END IF

      ! Update locally the value to each processor
      DO iconst = 1, SIZE(colv_list)
         ! Update local to each processor
         IF (colv_list(iconst)%expected_value_growth_speed == 0.0_dp) CYCLE
         CALL section_vals_val_get(collective_sections, "TARGET", &
                                   r_val=colv_list(iconst)%expected_value, &
                                   i_rep_section=colv_list(iconst)%inp_seq_num)
      END DO

   END SUBROUTINE shake_update_colv_low

! **************************************************************************************************
!> \brief Real Rattle - Low Level
!>      rattle algorithm for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param fixd_list ...
!> \param colv_list ...
!> \param lcolv ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \param irattle ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE rattle_colv_low(fixd_list, colv_list, lcolv, &
                              particle_set, vel, dt, irattle, cell, imass, max_sigma)

      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(IN)                                :: irattle
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: iconst
      REAL(KIND=dp)                                      :: del_lam, dtby2, fdotf_sum

      dtby2 = dt*.5_dp
      IF (irattle == 1) THEN
         DO iconst = 1, SIZE(colv_list)
            IF (colv_list(iconst)%restraint%active) CYCLE
            ! Update colvar_old
            CALL colvar_eval_mol_f(lcolv(iconst)%colvar_old, cell, &
                                   particles=particle_set, fixd_list=fixd_list)
            ! Update velocities
            CALL update_con_colv(vel, dtby2, lcolv(iconst), &
                                 lambda=lcolv(iconst)%lambda, &
                                 imass=imass)
         END DO
      ELSE
         DO iconst = 1, SIZE(colv_list)
            IF (colv_list(iconst)%restraint%active) CYCLE
            lcolv(iconst)%sigma = rattle_con_eval(lcolv(iconst)%colvar_old, vel)
            fdotf_sum = eval_Jac_colvar(lcolv(iconst)%colvar_old, &
                                        lcolv(iconst)%colvar_old, imass=imass)
            del_lam = 2.0_dp*lcolv(iconst)%sigma/(dt*fdotf_sum)
            lcolv(iconst)%lambda = lcolv(iconst)%lambda+del_lam

            ! Update velocities
            CALL update_con_colv(vel, dtby2, lcolv(iconst), &
                                 lambda=del_lam, &
                                 imass=imass)
         END DO
      END IF

      DO iconst = 1, SIZE(colv_list)
         IF (colv_list(iconst)%restraint%active) CYCLE
         lcolv(iconst)%sigma = rattle_con_eval(lcolv(iconst)%colvar_old, vel)
         max_sigma = MAX(ABS(lcolv(iconst)%sigma), max_sigma)
      END DO

   END SUBROUTINE rattle_colv_low

! **************************************************************************************************
!> \brief Real shake_roll - Low Level
!>      shake algorithm (box allowed to change) for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param fixd_list ...
!> \param colv_list ...
!> \param lcolv ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param r_shake ...
!> \param v_shake ...
!> \param dt ...
!> \param ishake ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE shake_roll_colv_low(fixd_list, colv_list, lcolv, &
                                  particle_set, pos, vel, r_shake, v_shake, dt, ishake, cell, &
                                  imass, max_sigma)

      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_shake, v_shake
      REAL(kind=dp), INTENT(in)                          :: dt
      INTEGER, INTENT(in)                                :: ishake
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: iconst
      REAL(KIND=dp)                                      :: del_lam, dtby2, dtsqby2, fdotf_sum

      dtsqby2 = dt*dt*.5_dp
      dtby2 = dt*.5_dp
      IF (ishake == 1) THEN
         DO iconst = 1, SIZE(colv_list)
            IF (colv_list(iconst)%restraint%active) CYCLE
            ! Update positions
            CALL update_con_colv(pos, dtsqby2, lcolv(iconst), &
                                 lambda=lcolv(iconst)%lambda, &
                                 roll=.TRUE., rmat=r_shake, imass=imass)
            ! Update velocities
            CALL update_con_colv(vel, dtby2, lcolv(iconst), &
                                 lambda=lcolv(iconst)%lambda, &
                                 roll=.TRUE., rmat=v_shake, imass=imass)
         END DO
      ELSE
         DO iconst = 1, SIZE(colv_list)
            IF (colv_list(iconst)%restraint%active) CYCLE
            ! Update colvar
            CALL colvar_eval_mol_f(lcolv(iconst)%colvar, cell, particles=particle_set, &
                                   pos=pos, fixd_list=fixd_list)
            lcolv(iconst)%sigma = diff_colvar(lcolv(iconst)%colvar, &
                                              colv_list(iconst)%expected_value)
            fdotf_sum = eval_Jac_colvar(lcolv(iconst)%colvar, &
                                        lcolv(iconst)%colvar_old, roll=.TRUE., rmat=r_shake, &
                                        imass=imass)
            del_lam = 2.0_dp*lcolv(iconst)%sigma/(dt*dt*fdotf_sum)
            lcolv(iconst)%lambda = lcolv(iconst)%lambda+del_lam

            ! Update positions
            CALL update_con_colv(pos, dtsqby2, lcolv(iconst), &
                                 lambda=del_lam, &
                                 roll=.TRUE., rmat=r_shake, imass=imass)
            ! Update velocities
            CALL update_con_colv(vel, dtby2, lcolv(iconst), &
                                 lambda=del_lam, &
                                 roll=.TRUE., rmat=v_shake, imass=imass)
         END DO
      END IF
      ! computing the constraint and value of tolerance
      DO iconst = 1, SIZE(colv_list)
         IF (colv_list(iconst)%restraint%active) CYCLE
         CALL colvar_eval_mol_f(lcolv(iconst)%colvar, cell, particles=particle_set, &
                                pos=pos, fixd_list=fixd_list)
         lcolv(iconst)%sigma = diff_colvar(lcolv(iconst)%colvar, &
                                           colv_list(iconst)%expected_value)
         max_sigma = MAX(ABS(lcolv(iconst)%sigma), max_sigma)
      END DO

   END SUBROUTINE shake_roll_colv_low

! **************************************************************************************************
!> \brief Real Rattle_roll - Low Level
!>      rattle algorithm (box allowed to change) for collective variables constraints
!>      updates the multiplier one molecule type at a time
!> \param fixd_list ...
!> \param colv_list ...
!> \param lcolv ...
!> \param particle_set ...
!> \param vel ...
!> \param r_rattle ...
!> \param dt ...
!> \param irattle ...
!> \param veps ...
!> \param cell ...
!> \param imass ...
!> \param max_sigma ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE rattle_roll_colv_low(fixd_list, colv_list, lcolv, &
                                   particle_set, vel, r_rattle, dt, irattle, veps, cell, &
                                   imass, max_sigma)

      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), INTENT(IN)                          :: r_rattle(:, :), dt
      INTEGER, INTENT(in)                                :: irattle
      REAL(KIND=dp), INTENT(IN)                          :: veps(:, :)
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp), INTENT(INOUT)                       :: max_sigma

      INTEGER                                            :: iconst
      REAL(KIND=dp)                                      :: del_lam, dtby2, fdotf_sum

      dtby2 = dt*.5_dp
      IF (irattle == 1) THEN
         DO iconst = 1, SIZE(colv_list)
            IF (colv_list(iconst)%restraint%active) CYCLE
            ! Update colvar_old
            CALL colvar_eval_mol_f(lcolv(iconst)%colvar_old, cell, &
                                   particles=particle_set, fixd_list=fixd_list)
            ! Update velocities
            CALL update_con_colv(vel, dtby2, lcolv(iconst), &
                                 lambda=lcolv(iconst)%lambda, &
                                 imass=imass)
         END DO
      ELSE
         DO iconst = 1, SIZE(colv_list)
            IF (colv_list(iconst)%restraint%active) CYCLE
            lcolv(iconst)%sigma = rattle_con_eval(lcolv(iconst)%colvar_old, vel, &
                                                  roll=.TRUE., veps=veps, rmat=r_rattle, particles=particle_set)
            fdotf_sum = eval_Jac_colvar(lcolv(iconst)%colvar_old, &
                                        lcolv(iconst)%colvar_old, roll=.TRUE., &
                                        rmat=r_rattle, imass=imass)
            del_lam = 2.0_dp*lcolv(iconst)%sigma/(dt*fdotf_sum)
            lcolv(iconst)%lambda = lcolv(iconst)%lambda+del_lam
            ! Update velocities
            CALL update_con_colv(vel, dtby2, lcolv(iconst), &
                                 lambda=del_lam, &
                                 roll=.TRUE., rmat=r_rattle, imass=imass)
         END DO
      END IF
      ! computing the constraint and value of the tolerance
      DO iconst = 1, SIZE(colv_list)
         IF (colv_list(iconst)%restraint%active) CYCLE
         lcolv(iconst)%sigma = rattle_con_eval(lcolv(iconst)%colvar_old, vel, &
                                               roll=.TRUE., veps=veps, rmat=r_rattle, particles=particle_set)
         max_sigma = MAX(ABS(lcolv(iconst)%sigma), max_sigma)
      END DO

   END SUBROUTINE rattle_roll_colv_low

! **************************************************************************************************
!> \brief Update position/velocities
!> \param wrk ...
!> \param fac ...
!> \param lcolv ...
!> \param lambda ...
!> \param roll ...
!> \param rmat ...
!> \param imass ...
!> \par History
!>      Teodoro Laino [teo] created 04.2006
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE update_con_colv(wrk, fac, lcolv, lambda, roll, rmat, imass)
      REAL(KIND=dp), INTENT(INOUT)                       :: wrk(:, :)
      REAL(KIND=dp), INTENT(IN)                          :: fac
      TYPE(local_colvar_constraint_type), INTENT(IN)     :: lcolv
      REAL(KIND=dp), INTENT(IN)                          :: lambda
      LOGICAL, INTENT(in), OPTIONAL                      :: roll
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: rmat
      REAL(KIND=dp), DIMENSION(:)                        :: imass

      CHARACTER(len=*), PARAMETER :: routineN = 'update_con_colv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatm, ind
      LOGICAL                                            :: my_roll
      REAL(KIND=dp), DIMENSION(3)                        :: f_roll

      my_roll = .FALSE.
      IF (PRESENT(roll)) THEN
         my_roll = roll
         IF (my_roll) THEN
            IF(.NOT.(PRESENT(rmat)))CALL cp__a("constraint_clv.F",895)
         END IF
      END IF
      DO iatm = 1, SIZE(lcolv%colvar_old%i_atom)
         ind = lcolv%colvar_old%i_atom(iatm)
         !
         IF (my_roll) THEN
            ! If ROLL rotate forces
            CALL matvec_3x3(f_roll, rmat, lcolv%colvar_old%dsdr(:, iatm))
         ELSE
            f_roll = lcolv%colvar_old%dsdr(:, iatm)
         END IF
         wrk(:, ind) = wrk(:, ind)-imass(ind)*fac*lambda*f_roll
      END DO
   END SUBROUTINE update_con_colv

! **************************************************************************************************
!> \brief Evaluates the Jacobian of the collective variables constraints
!> \param colvar ...
!> \param colvar_old ...
!> \param roll ...
!> \param rmat ...
!> \param imass ...
!> \retval res ...
!> \par History
!>      Teodoro Laino [teo] created 04.2006
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   FUNCTION eval_Jac_colvar(colvar, colvar_old, roll, rmat, imass) RESULT(res)
      TYPE(colvar_type), POINTER                         :: colvar, colvar_old
      LOGICAL, INTENT(IN), OPTIONAL                      :: roll
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: rmat
      REAL(KIND=dp), DIMENSION(:)                        :: imass
      REAL(KIND=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'eval_Jac_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom
      LOGICAL                                            :: my_roll
      REAL(KIND=dp), DIMENSION(3)                        :: tmp1, tmp2, tmp3

      my_roll = .FALSE.
      IF (PRESENT(roll)) THEN
         my_roll = roll
         IF (my_roll) THEN
            IF(.NOT.(PRESENT(rmat)))CALL cp__a("constraint_clv.F",942)
         END IF
      END IF

      res = 0.0_dp
      DO i = 1, SIZE(colvar%i_atom)
         iatom = colvar%i_atom(i)
         tmp1 = colvar%dsdr(1:3, i)
         IF (my_roll) THEN
            tmp3 = colvar_old%dsdr(1:3, i)
            CALL matvec_3x3(tmp2, rmat, tmp3)
         ELSE
            tmp2 = colvar_old%dsdr(1:3, i)
         END IF
         res = res+DOT_PRODUCT(tmp1, tmp2)*imass(iatom)
      END DO

   END FUNCTION eval_Jac_colvar

! **************************************************************************************************
!> \brief Evaluates the constraint for the rattle scheme
!> \param colvar ...
!> \param vel ...
!> \param roll ...
!> \param veps ...
!> \param rmat ...
!> \param particles ...
!> \retval res ...
!> \par History
!>      Teodoro Laino [teo] created 04.2006
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   FUNCTION rattle_con_eval(colvar, vel, roll, veps, rmat, particles) RESULT(res)
      TYPE(colvar_type), POINTER                         :: colvar
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      LOGICAL, INTENT(IN), OPTIONAL                      :: roll
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: veps, rmat
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles
      REAL(KIND=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'rattle_con_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatm, ind
      LOGICAL                                            :: my_roll
      REAL(KIND=dp)                                      :: tmp(3)
      REAL(KIND=dp), DIMENSION(3)                        :: f_roll, pos, v_roll

      my_roll = .FALSE.
      IF (PRESENT(roll)) THEN
         my_roll = roll
         IF (my_roll) THEN
            IF(.NOT.(PRESENT(rmat)))CALL cp__a("constraint_clv.F",996)
            IF(.NOT.(PRESENT(veps)))CALL cp__a("constraint_clv.F",997)
            IF(.NOT.(PRESENT(particles)))CALL cp__a("constraint_clv.F",998)
         END IF
      END IF
      res = 0.0_dp
      DO iatm = 1, SIZE(colvar%i_atom)
         ind = colvar%i_atom(iatm)
         IF (my_roll) THEN
            pos = particles(ind)%r
            CALL matvec_3x3(f_roll, rmat, colvar%dsdr(:, iatm))
            CALL matvec_3x3(tmp, veps, pos)
            v_roll = vel(:, ind)+tmp
         ELSE
            f_roll = colvar%dsdr(:, iatm)
            v_roll = vel(:, ind)
         END IF
         res = res+DOT_PRODUCT(f_roll, v_roll)
      END DO

   END FUNCTION rattle_con_eval

END MODULE constraint_clv
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Teodoro Laino [tlaino] 2007 - Extension to Intermolecular constraints
! **************************************************************************************************
MODULE constraint
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE colvar_types,                    ONLY: colvar_counters
   USE constraint_3x3,                  ONLY: rattle_3x3_ext,&
                                              rattle_3x3_int,&
                                              rattle_roll_3x3_ext,&
                                              rattle_roll_3x3_int,&
                                              shake_3x3_ext,&
                                              shake_3x3_int,&
                                              shake_roll_3x3_ext,&
                                              shake_roll_3x3_int
   USE constraint_4x6,                  ONLY: rattle_4x6_ext,&
                                              rattle_4x6_int,&
                                              rattle_roll_4x6_ext,&
                                              rattle_roll_4x6_int,&
                                              shake_4x6_ext,&
                                              shake_4x6_int,&
                                              shake_roll_4x6_ext,&
                                              shake_roll_4x6_int
   USE constraint_clv,                  ONLY: &
        rattle_colv_ext, rattle_colv_int, rattle_roll_colv_ext, rattle_roll_colv_int, &
        shake_colv_ext, shake_colv_int, shake_roll_colv_ext, shake_roll_colv_int, &
        shake_update_colv_ext, shake_update_colv_int
   USE constraint_util,                 ONLY: check_tol,&
                                              get_roll_matrix,&
                                              restore_temporary_set,&
                                              update_temporary_set
   USE constraint_vsite,                ONLY: shake_vsite_ext,&
                                              shake_vsite_int
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE input_constants,                 ONLY: npt_f_ensemble,&
                                              npt_i_ensemble
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum
   USE molecule_kind_types,             ONLY: get_molecule_kind,&
                                              get_molecule_kind_set,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE particle_types,                  ONLY: particle_type
   USE simpar_types,                    ONLY: simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: shake_control, &
             rattle_control, &
             shake_roll_control, &
             rattle_roll_control, &
             shake_update_targets

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'constraint'
   INTEGER, PARAMETER, PRIVATE :: Max_Shake_Iter = 1000

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param shake_tol ...
!> \param log_unit ...
!> \param lagrange_mult ...
!> \param dump_lm ...
!> \param cell ...
!> \param group ...
!> \param local_particles ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - Extension to Intermolecular constraints
! **************************************************************************************************
   SUBROUTINE shake_control(gci, local_molecules, molecule_set, molecule_kind_set, &
                            particle_set, pos, vel, dt, shake_tol, log_unit, lagrange_mult, dump_lm, &
                            cell, group, local_particles)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt, shake_tol
      INTEGER, INTENT(in)                                :: log_unit, lagrange_mult
      LOGICAL, INTENT(IN)                                :: dump_lm
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(in)                                :: group
      TYPE(distribution_1d_type), POINTER                :: local_particles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'shake_control', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ikind, imol, ishake_ext, &
                                                            ishake_int, k, n3x3con, n4x6con, &
                                                            nconstraint, nkind, nmol_per_kind, &
                                                            nvsitecon
      LOGICAL                                            :: do_ext_constraint
      REAL(KIND=dp)                                      :: int_max_sigma, mass, max_sigma
      REAL(KIND=dp), DIMENSION(SIZE(pos, 2))             :: imass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(colvar_counters)                              :: ncolv
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)
      nkind = SIZE(molecule_kind_set)
      DO k = 1, SIZE(pos, 2)
         atomic_kind => particle_set(k)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass(k) = 1.0_dp/MAX(mass, EPSILON(0.0_dp))
      END DO
      do_ext_constraint = (gci%ntot /= 0)
      ishake_ext = 0
      max_sigma = -1.0E+10_dp
      Shake_Inter_Loop: DO WHILE ((ABS(max_sigma) >= shake_tol) .AND. (ishake_ext <= Max_Shake_Iter))
         max_sigma = 0.0_dp
         ishake_ext = ishake_ext+1
         ! Intramolecular Constraints
         MOL: DO ikind = 1, nkind
            nmol_per_kind = local_molecules%n_el(ikind)
            DO imol = 1, nmol_per_kind
               i = local_molecules%list(ikind)%array(imol)
               molecule => molecule_set(i)
               molecule_kind => molecule%molecule_kind
               CALL get_molecule_kind(molecule_kind, ncolv=ncolv, &
                                      ng3x3=n3x3con, ng4x6=n4x6con, &
                                      nconstraint=nconstraint, nvsite=nvsitecon)
               IF (nconstraint == 0) CYCLE
               ishake_int = 0
               int_max_sigma = -1.0E+10_dp
               Shake_Intra_Loop: DO WHILE ((ABS(int_max_sigma) >= shake_tol) .AND. (ishake_int <= Max_Shake_Iter))
                  int_max_sigma = 0.0_dp
                  ishake_int = ishake_int+1
                  ! 3x3
                  IF (n3x3con /= 0) &
                     CALL shake_3x3_int(molecule, particle_set, pos, vel, dt, ishake_int, &
                                        int_max_sigma)
                  ! 4x6
                  IF (n4x6con /= 0) &
                     CALL shake_4x6_int(molecule, particle_set, pos, vel, dt, ishake_int, &
                                        int_max_sigma)
                  ! Collective Variables
                  IF (ncolv%ntot /= 0) &
                     CALL shake_colv_int(molecule, particle_set, pos, vel, dt, ishake_int, &
                                         cell, imass, int_max_sigma)
               END DO Shake_Intra_Loop
               max_sigma = MAX(max_sigma, int_max_sigma)
               CALL shake_int_info(log_unit, i, ishake_int, max_sigma)
               ! Virtual Site
               IF (nvsitecon /= 0) &
                  CALL shake_vsite_int(molecule, pos)
            END DO
         END DO MOL
         ! Intermolecular constraints
         IF (do_ext_constraint) THEN
            CALL update_temporary_set(group, pos=pos, vel=vel)
            ! 3x3
            IF (gci%ng3x3 /= 0) &
               CALL shake_3x3_ext(gci, particle_set, pos, vel, dt, ishake_ext, &
                                  max_sigma)
            ! 4x6
            IF (gci%ng4x6 /= 0) &
               CALL shake_4x6_ext(gci, particle_set, pos, vel, dt, ishake_ext, &
                                  max_sigma)
            ! Collective Variables
            IF (gci%ncolv%ntot /= 0) &
               CALL shake_colv_ext(gci, particle_set, pos, vel, dt, ishake_ext, &
                                   cell, imass, max_sigma)
            ! Virtual Site
            IF (gci%nvsite /= 0) &
               CALL shake_vsite_ext(gci, pos)
            CALL restore_temporary_set(particle_set, local_particles, pos=pos, vel=vel)
         END IF
         CALL shake_ext_info(log_unit, ishake_ext, max_sigma)
      END DO Shake_Inter_Loop
      CALL dump_lagrange_mult(dump_lm, lagrange_mult, local_molecules, molecule_set, gci, &
                              molecule_kind_set, group, "S")

      CALL timestop(handle)
   END SUBROUTINE shake_control

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \param rattle_tol ...
!> \param log_unit ...
!> \param lagrange_mult ...
!> \param dump_lm ...
!> \param cell ...
!> \param group ...
!> \param local_particles ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - Extension to Intermolecular constraints
! **************************************************************************************************
   SUBROUTINE rattle_control(gci, local_molecules, molecule_set, molecule_kind_set, &
                             particle_set, vel, dt, rattle_tol, log_unit, lagrange_mult, dump_lm, cell, group, &
                             local_particles)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(kind=dp), INTENT(in)                          :: dt, rattle_tol
      INTEGER, INTENT(in)                                :: log_unit, lagrange_mult
      LOGICAL, INTENT(IN)                                :: dump_lm
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(in)                                :: group
      TYPE(distribution_1d_type), POINTER                :: local_particles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rattle_control', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ikind, imol, irattle_ext, &
                                                            irattle_int, k, n3x3con, n4x6con, &
                                                            nconstraint, nkind, nmol_per_kind
      LOGICAL                                            :: do_ext_constraint
      REAL(KIND=dp)                                      :: int_max_sigma, mass, max_sigma
      REAL(KIND=dp), DIMENSION(SIZE(vel, 2))             :: imass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(colvar_counters)                              :: ncolv
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)
      nkind = SIZE(molecule_kind_set)
      DO k = 1, SIZE(vel, 2)
         atomic_kind => particle_set(k)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass(k) = 1.0_dp/MAX(mass, EPSILON(0.0_dp))
      END DO
      do_ext_constraint = (gci%ntot /= 0)
      irattle_ext = 0
      max_sigma = -1.0E+10_dp
      Rattle_Inter_Loop: DO WHILE (ABS(max_sigma) >= rattle_tol)
         max_sigma = 0.0_dp
         irattle_ext = irattle_ext+1
         ! Intramolecular Constraints
         MOL: DO ikind = 1, nkind
            nmol_per_kind = local_molecules%n_el(ikind)
            DO imol = 1, nmol_per_kind
               i = local_molecules%list(ikind)%array(imol)
               molecule => molecule_set(i)
               molecule_kind => molecule%molecule_kind
               CALL get_molecule_kind(molecule_kind, ncolv=ncolv, ng3x3=n3x3con, &
                                      ng4x6=n4x6con, nconstraint=nconstraint)
               IF (nconstraint == 0) CYCLE
               irattle_int = 0
               int_max_sigma = -1.0E+10_dp
               Rattle_Intra_Loop: DO WHILE (ABS(int_max_sigma) >= rattle_tol)
                  int_max_sigma = 0.0_dp
                  irattle_int = irattle_int+1
                  ! 3x3
                  IF (n3x3con /= 0) &
                     CALL rattle_3x3_int(molecule, particle_set, vel, dt)
                  ! 4x6
                  IF (n4x6con /= 0) &
                     CALL rattle_4x6_int(molecule, particle_set, vel, dt)
                  ! Collective Variables
                  IF (ncolv%ntot /= 0) &
                     CALL rattle_colv_int(molecule, particle_set, vel, dt, &
                                          irattle_int, cell, imass, int_max_sigma)
               END DO Rattle_Intra_Loop
               max_sigma = MAX(max_sigma, int_max_sigma)
               CALL rattle_int_info(log_unit, i, irattle_int, max_sigma)
            END DO
         END DO MOL
         ! Intermolecular Constraints
         IF (do_ext_constraint) THEN
            CALL update_temporary_set(group, vel=vel)
            ! 3x3
            IF (gci%ng3x3 /= 0) &
               CALL rattle_3x3_ext(gci, particle_set, vel, dt)
            ! 4x6
            IF (gci%ng4x6 /= 0) &
               CALL rattle_4x6_ext(gci, particle_set, vel, dt)
            ! Collective Variables
            IF (gci%ncolv%ntot /= 0) &
               CALL rattle_colv_ext(gci, particle_set, vel, dt, &
                                    irattle_ext, cell, imass, max_sigma)
            CALL restore_temporary_set(particle_set, local_particles, vel=vel)
         END IF
         CALL rattle_ext_info(log_unit, irattle_ext, max_sigma)
      END DO Rattle_Inter_Loop
      CALL dump_lagrange_mult(dump_lm, lagrange_mult, local_molecules, molecule_set, gci, &
                              molecule_kind_set, group, "R")
      CALL timestop(handle)

   END SUBROUTINE rattle_control

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param particle_set ...
!> \param pos ...
!> \param vel ...
!> \param dt ...
!> \param simpar ...
!> \param roll_tol ...
!> \param iroll ...
!> \param vector_r ...
!> \param vector_v ...
!> \param group ...
!> \param u ...
!> \param cell ...
!> \param local_particles ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - Extension to Intermolecular constraints
! **************************************************************************************************
   SUBROUTINE shake_roll_control(gci, local_molecules, molecule_set, &
                                 molecule_kind_set, particle_set, pos, vel, dt, simpar, roll_tol, iroll, &
                                 vector_r, vector_v, group, u, cell, local_particles)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: pos(:, :), vel(:, :)
      REAL(KIND=dp), INTENT(IN)                          :: dt
      TYPE(simpar_type), INTENT(IN)                      :: simpar
      REAL(KIND=dp), INTENT(OUT)                         :: roll_tol
      INTEGER, INTENT(INOUT)                             :: iroll
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: vector_r, vector_v
      INTEGER, INTENT(IN)                                :: group
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: u
      TYPE(cell_type), POINTER                           :: cell
      TYPE(distribution_1d_type), POINTER                :: local_particles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'shake_roll_control', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, ikind, imol, ishake_ext, ishake_int, k, lagrange_mult, log_unit, &
         n3x3con, n4x6con, nconstraint, nkind, nmol_per_kind, nvsitecon
      LOGICAL                                            :: do_ext_constraint, dump_lm
      REAL(KIND=dp)                                      :: int_max_sigma, mass, max_sigma, shake_tol
      REAL(KIND=dp), DIMENSION(3, 3)                     :: r_shake, v_shake
      REAL(KIND=dp), DIMENSION(SIZE(pos, 2))             :: imass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(colvar_counters)                              :: ncolv
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)
      nkind = SIZE(molecule_kind_set)
      shake_tol = simpar%shake_tol
      log_unit = simpar%info_constraint
      lagrange_mult = simpar%lagrange_multipliers
      dump_lm = simpar%dump_lm
      ! setting up for roll
      IF (simpar%ensemble == npt_i_ensemble) THEN
         CALL get_roll_matrix('SHAKE', r_shake, v_shake, vector_r, vector_v)
      ELSE IF (simpar%ensemble == npt_f_ensemble) THEN
         CALL get_roll_matrix('SHAKE', r_shake, v_shake, vector_r, vector_v, u)
      END IF
      DO k = 1, SIZE(pos, 2)
         atomic_kind => particle_set(k)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass(k) = 1.0_dp/MAX(mass, EPSILON(0.0_dp))
      END DO
      do_ext_constraint = (gci%ntot /= 0)
      ishake_ext = 0
      max_sigma = -1.0E+10_dp
      Shake_Inter_Loop: DO WHILE (ABS(max_sigma) >= shake_tol)
         max_sigma = 0.0_dp
         ishake_ext = ishake_ext+1
         ! Intramolecular Constraints
         MOL: DO ikind = 1, nkind
            nmol_per_kind = local_molecules%n_el(ikind)
            DO imol = 1, nmol_per_kind
               i = local_molecules%list(ikind)%array(imol)
               molecule => molecule_set(i)
               molecule_kind => molecule%molecule_kind
               CALL get_molecule_kind(molecule_kind, ncolv=ncolv, &
                                      ng3x3=n3x3con, ng4x6=n4x6con, &
                                      nconstraint=nconstraint, nvsite=nvsitecon)
               IF (nconstraint == 0) CYCLE
               ishake_int = 0
               int_max_sigma = -1.0E+10_dp
               Shake_Roll_Intra_Loop: DO WHILE (ABS(int_max_sigma) >= shake_tol)
                  int_max_sigma = 0.0_dp
                  ishake_int = ishake_int+1
                  ! 3x3
                  IF (n3x3con /= 0) &
                     CALL shake_roll_3x3_int(molecule, particle_set, pos, vel, r_shake, &
                                             v_shake, dt, ishake_int, int_max_sigma)
                  ! 4x6
                  IF (n4x6con /= 0) &
                     CALL shake_roll_4x6_int(molecule, particle_set, pos, vel, r_shake, &
                                             dt, ishake_int, int_max_sigma)
                  ! Collective Variables
                  IF (ncolv%ntot /= 0) &
                     CALL shake_roll_colv_int(molecule, particle_set, pos, vel, r_shake, &
                                              v_shake, dt, ishake_int, cell, imass, int_max_sigma)
               END DO Shake_Roll_Intra_Loop
               max_sigma = MAX(max_sigma, int_max_sigma)
               CALL shake_int_info(log_unit, i, ishake_int, max_sigma)
               ! Virtual Site
               IF (nvsitecon /= 0) THEN
                  CALL cp__b("constraint.F",431,"Virtual Site Constraint/Restraint not implemented for SHAKE_ROLL!")
               END IF
            END DO
         END DO MOL
         ! Intermolecular constraints
         IF (do_ext_constraint) THEN
            CALL update_temporary_set(group, pos=pos, vel=vel)
            ! 3x3
            IF (gci%ng3x3 /= 0) &
               CALL shake_roll_3x3_ext(gci, particle_set, pos, vel, r_shake, &
                                       v_shake, dt, ishake_ext, max_sigma)
            ! 4x6
            IF (gci%ng4x6 /= 0) &
               CALL shake_roll_4x6_ext(gci, particle_set, pos, vel, r_shake, &
                                       dt, ishake_ext, max_sigma)
            ! Collective Variables
            IF (gci%ncolv%ntot /= 0) &
               CALL shake_roll_colv_ext(gci, particle_set, pos, vel, r_shake, &
                                        v_shake, dt, ishake_ext, cell, imass, max_sigma)
            ! Virtual Site
            IF (gci%nvsite /= 0) &
               CALL cp__b("constraint.F",452,"Virtual Site Constraint/Restraint not implemented for SHAKE_ROLL!")
            CALL restore_temporary_set(particle_set, local_particles, pos=pos, vel=vel)
         END IF
         CALL shake_ext_info(log_unit, ishake_ext, max_sigma)
      END DO Shake_Inter_Loop
      CALL dump_lagrange_mult(dump_lm, lagrange_mult, local_molecules, molecule_set, gci, &
                              molecule_kind_set, group, "S")
      CALL check_tol(roll_tol, iroll, 'SHAKE', r_shake)
      CALL timestop(handle)

   END SUBROUTINE shake_roll_control

! **************************************************************************************************
!> \brief ...
!> \param gci ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param particle_set ...
!> \param vel ...
!> \param dt ...
!> \param simpar ...
!> \param vector ...
!> \param veps ...
!> \param roll_tol ...
!> \param iroll ...
!> \param para_env ...
!> \param u ...
!> \param cell ...
!> \param local_particles ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - Extension to Intermolecular constraints
! **************************************************************************************************
   SUBROUTINE rattle_roll_control(gci, local_molecules, molecule_set, &
                                  molecule_kind_set, particle_set, vel, dt, simpar, vector, &
                                  veps, roll_tol, iroll, para_env, u, cell, local_particles)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel(:, :)
      REAL(KIND=dp), INTENT(IN)                          :: dt
      TYPE(simpar_type), INTENT(IN)                      :: simpar
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: vector
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: veps
      REAL(KIND=dp), INTENT(OUT)                         :: roll_tol
      INTEGER, INTENT(INOUT)                             :: iroll
      TYPE(cp_para_env_type), INTENT(IN)                 :: para_env
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: u
      TYPE(cell_type), POINTER                           :: cell
      TYPE(distribution_1d_type), POINTER                :: local_particles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rattle_roll_control', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, ikind, imol, irattle_ext, irattle_int, k, lagrange_mult, log_unit, &
         n3x3con, n4x6con, nconstraint, nkind, nmol_per_kind
      LOGICAL                                            :: do_ext_constraint, dump_lm
      REAL(KIND=dp)                                      :: int_max_sigma, mass, max_sigma, &
                                                            rattle_tol
      REAL(KIND=dp), DIMENSION(3, 3)                     :: r_rattle
      REAL(KIND=dp), DIMENSION(SIZE(vel, 2))             :: imass
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(colvar_counters)                              :: ncolv
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)
      ! initialize locals
      nkind = SIZE(molecule_kind_set)
      rattle_tol = simpar%shake_tol
      log_unit = simpar%info_constraint
      lagrange_mult = simpar%lagrange_multipliers
      dump_lm = simpar%dump_lm
      ! setting up for roll
      IF (simpar%ensemble == npt_i_ensemble) THEN
         CALL get_roll_matrix('RATTLE', v_shake=r_rattle, vector_v=vector)
      ELSE IF (simpar%ensemble == npt_f_ensemble) THEN
         CALL get_roll_matrix('RATTLE', v_shake=r_rattle, vector_v=vector, u=u)
      END IF
      DO k = 1, SIZE(vel, 2)
         atomic_kind => particle_set(k)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
         imass(k) = 1.0_dp/MAX(mass, EPSILON(0.0_dp))
      END DO
      do_ext_constraint = (gci%ntot /= 0)
      irattle_ext = 0
      max_sigma = -1.0E+10_dp
      Rattle_Inter_Loop: DO WHILE (ABS(max_sigma) >= rattle_tol)
         max_sigma = 0.0_dp
         irattle_ext = irattle_ext+1
         ! Intramolecular Constraints
         MOL: DO ikind = 1, nkind
            nmol_per_kind = local_molecules%n_el(ikind)
            DO imol = 1, nmol_per_kind
               i = local_molecules%list(ikind)%array(imol)
               molecule => molecule_set(i)
               molecule_kind => molecule%molecule_kind
               CALL get_molecule_kind(molecule_kind, ncolv=ncolv, &
                                      ng3x3=n3x3con, ng4x6=n4x6con, &
                                      nconstraint=nconstraint)
               IF (nconstraint == 0) CYCLE
               int_max_sigma = -1.0E+10_dp
               irattle_int = 0
               Rattle_Roll_Intramolecular: DO WHILE (ABS(int_max_sigma) >= rattle_tol)
                  int_max_sigma = 0.0_dp
                  irattle_int = irattle_int+1
                  ! 3x3
                  IF (n3x3con /= 0) &
                     CALL rattle_roll_3x3_int(molecule, particle_set, vel, r_rattle, dt, &
                                              veps)
                  ! 4x6
                  IF (n4x6con /= 0) &
                     CALL rattle_roll_4x6_int(molecule, particle_set, vel, r_rattle, dt, &
                                              veps)
                  ! Collective Variables
                  IF (ncolv%ntot /= 0) &
                     CALL rattle_roll_colv_int(molecule, particle_set, vel, r_rattle, dt, &
                                               irattle_int, veps, cell, imass, int_max_sigma)
               END DO Rattle_Roll_Intramolecular
               max_sigma = MAX(max_sigma, int_max_sigma)
               CALL rattle_int_info(log_unit, i, irattle_int, max_sigma)
            END DO
         END DO MOL
         ! Intermolecular Constraints
         IF (do_ext_constraint) THEN
            CALL update_temporary_set(para_env%group, vel=vel)
            ! 3x3
            IF (gci%ng3x3 /= 0) &
               CALL rattle_roll_3x3_ext(gci, particle_set, vel, r_rattle, dt, &
                                        veps)
            ! 4x6
            IF (gci%ng4x6 /= 0) &
               CALL rattle_roll_4x6_ext(gci, particle_set, vel, r_rattle, dt, &
                                        veps)
            ! Collective Variables
            IF (gci%ncolv%ntot /= 0) &
               CALL rattle_roll_colv_ext(gci, particle_set, vel, r_rattle, dt, &
                                         irattle_ext, veps, cell, imass, max_sigma)
            CALL restore_temporary_set(particle_set, local_particles, vel=vel)
         END IF
         CALL rattle_ext_info(log_unit, irattle_ext, max_sigma)
      END DO Rattle_Inter_Loop
      CALL dump_lagrange_mult(dump_lm, lagrange_mult, local_molecules, molecule_set, gci, &
                              molecule_kind_set, para_env%group, "R")
      CALL check_tol(roll_tol, iroll, 'RATTLE', veps=veps)
      CALL timestop(handle)
   END SUBROUTINE rattle_roll_control

! **************************************************************************************************
!> \brief ...
!> \param dump_lm ...
!> \param log_unit ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param gci ...
!> \param molecule_kind_set ...
!> \param group ...
!> \param id_type ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - Dumps lagrange multipliers
! **************************************************************************************************
   SUBROUTINE dump_lagrange_mult(dump_lm, log_unit, local_molecules, molecule_set, gci, &
                                 molecule_kind_set, group, id_type)
      LOGICAL, INTENT(IN)                                :: dump_lm
      INTEGER, INTENT(IN)                                :: log_unit
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      INTEGER, INTENT(IN)                                :: group
      CHARACTER(LEN=1), INTENT(IN)                       :: id_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_lagrange_mult', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: label
      INTEGER                                            :: handle, i, ikind, imol, j, my_index, &
                                                            n3x3con, n4x6con, nconstraint, nkind
      LOGICAL                                            :: do_ext_constraint, do_int_constraint
      REAL(KIND=dp), DIMENSION(:), POINTER               :: lagr
      TYPE(colvar_counters)                              :: ncolv
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      CALL timeset(routineN, handle)
      ! Total number of intramolecular constraints (distributed)
      CALL get_molecule_kind_set(molecule_kind_set=molecule_kind_set, &
                                 nconstraint=nconstraint)
      do_int_constraint = (nconstraint > 0)
      do_ext_constraint = (gci%ntot > 0)
      IF (dump_lm .AND. (do_int_constraint .OR. do_ext_constraint)) THEN
         nkind = SIZE(molecule_kind_set)
         ALLOCATE (lagr(nconstraint))
         lagr = 0.0_dp
         ! Dump lagrange multipliers for Intramolecular Constraints
         my_index = 0
         IF (do_int_constraint) THEN
            MOL: DO ikind = 1, nkind
               molecule_kind => molecule_kind_set(ikind)
               CALL get_molecule_kind(molecule_kind, &
                                      ncolv=ncolv, &
                                      ng3x3=n3x3con, &
                                      ng4x6=n4x6con)
               DO imol = 1, molecule_kind%nmolecule
                  i = molecule_kind%molecule_list(imol)
                  IF (ANY(local_molecules%list(ikind)%array == i)) THEN
                     molecule => molecule_set(i)
                     ! Collective Variables
                     DO j = 1, ncolv%ntot
                        lagr(my_index+1) = molecule%lci%lcolv(j)%lambda
                        my_index = my_index+1
                     END DO
                     ! 3x3
                     DO j = 1, n3x3con
                        lagr(my_index+1:my_index+3) = molecule%lci%lg3x3(j)%lambda(:)
                        my_index = my_index+3
                     END DO
                     ! 4x6
                     DO j = 1, n4x6con
                        lagr(my_index+1:my_index+6) = molecule%lci%lg4x6(j)%lambda(:)
                        my_index = my_index+6
                     END DO
                  ELSE
                     my_index = my_index+ncolv%ntot+3*n3x3con+6*n4x6con
                  END IF
               END DO
            END DO MOL
            CALL mp_sum(lagr, group)
         END IF
         ! Intermolecular constraints
         IF (do_ext_constraint) THEN
            CALL reallocate(lagr, 1, SIZE(lagr)+gci%ntot)
            ! Collective Variables
            DO j = 1, gci%ncolv%ntot
               lagr(my_index+1) = gci%lcolv(j)%lambda
               my_index = my_index+1
            END DO
            ! 3x3
            DO j = 1, gci%ng3x3
               lagr(my_index+1:my_index+3) = gci%lg3x3(j)%lambda(:)
               my_index = my_index+3
            END DO
            ! 4x6
            DO j = 1, gci%ng4x6
               lagr(my_index+1:my_index+6) = gci%lg4x6(j)%lambda(:)
               my_index = my_index+6
            END DO
         END IF
         IF (log_unit > 0) THEN
            IF (id_type == "S") THEN
               label = "Shake  Lagrangian Multipliers:"
            ELSEIF (id_type == "R") THEN
               label = "Rattle Lagrangian Multipliers:"
            ELSE
               CALL cp__b("constraint.F",710,"")
            END IF
            WRITE (log_unit, FMT='(A,T40,4F15.9)') TRIM(label), lagr(1:MIN(4, SIZE(lagr)))
            DO j = 5, SIZE(lagr), 4
               WRITE (log_unit, FMT='(T40,4F15.9)') lagr(j:MIN(j+3, SIZE(lagr)))
            END DO
         END IF
         DEALLOCATE (lagr)
      END IF
      CALL timestop(handle)

   END SUBROUTINE dump_lagrange_mult

! **************************************************************************************************
!> \brief Dumps convergence info about shake - intramolecular constraint loop
!> \param log_unit ...
!> \param i ...
!> \param ishake_int ...
!> \param max_sigma ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE shake_int_info(log_unit, i, ishake_int, max_sigma)
      INTEGER, INTENT(IN)                                :: log_unit, i, ishake_int
      REAL(KIND=dp), INTENT(IN)                          :: max_sigma

      CHARACTER(LEN=*), PARAMETER :: routineN = 'shake_int_info', routineP = moduleN//':'//routineN

      IF (log_unit > 0) THEN
         ! Dump info if requested
         WRITE (log_unit, '("SHAKE_INFO|",2X,2(A,I6),A,F15.9)') &
            "Molecule Nr.:", i, " Nr. Iterations:", ishake_int, " Max. Err.:", max_sigma
      END IF
      ! Notify a not converged SHAKE
      IF (ishake_int > Max_Shake_Iter) &
         CALL cp_warn(cp__l("constraint.F",745), &
                      "Shake NOT converged in "//cp_to_string(Max_Shake_Iter)//" iterations in the "// &
                      " intramolecular constraint loop for Molecule nr. "//cp_to_string(i)// &
                      ". CP2K continues but results could be meaningless. ")
   END SUBROUTINE shake_int_info

! **************************************************************************************************
!> \brief Dumps convergence info about shake - intermolecular constraint loop
!> \param log_unit ...
!> \param ishake_ext ...
!> \param max_sigma ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE shake_ext_info(log_unit, ishake_ext, max_sigma)
      INTEGER, INTENT(IN)                                :: log_unit, ishake_ext
      REAL(KIND=dp), INTENT(IN)                          :: max_sigma

      CHARACTER(LEN=*), PARAMETER :: routineN = 'shake_ext_info', routineP = moduleN//':'//routineN

      IF (log_unit > 0) THEN
         ! Dump info if requested
         WRITE (log_unit, '("SHAKE_INFO|",2X,A,I6,A,F15.9)') &
            "External Shake      Nr. Iterations:", ishake_ext, &
            " Max. Err.:", max_sigma
      END IF
      ! Notify a not converged SHAKE
      IF (ishake_ext > Max_Shake_Iter) &
         CALL cp_warn(cp__l("constraint.F",773), &
                      "Shake NOT converged in "//cp_to_string(Max_Shake_Iter)//" iterations in the "// &
                      " intermolecular constraint. CP2K continues but results could be meaningless. ")
   END SUBROUTINE shake_ext_info

! **************************************************************************************************
!> \brief Dumps convergence info about rattle - intramolecular constraint loop
!> \param log_unit ...
!> \param i ...
!> \param irattle_int ...
!> \param max_sigma ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE rattle_int_info(log_unit, i, irattle_int, max_sigma)
      INTEGER, INTENT(IN)                                :: log_unit, i, irattle_int
      REAL(KIND=dp), INTENT(IN)                          :: max_sigma

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rattle_int_info', &
         routineP = moduleN//':'//routineN

      IF (log_unit > 0) THEN
         ! Dump info if requested
         WRITE (log_unit, '("RATTLE_INFO|",1X,2(A,I6),A,F15.9)') &
            "Molecule Nr.:", i, " Nr. Iterations:", irattle_int, " Max. Err.:", max_sigma
      END IF
      ! Notify a not converged RATTLE
      IF (irattle_int > Max_shake_Iter) &
         CALL cp_warn(cp__l("constraint.F",801), &
                      "Rattle NOT converged in "//cp_to_string(Max_Shake_Iter)//" iterations in the "// &
                      " intramolecular constraint loop for Molecule nr. "//cp_to_string(i)// &
                      ". CP2K continues but results could be meaningless. ")
   END SUBROUTINE rattle_int_info

! **************************************************************************************************
!> \brief Dumps convergence info about rattle - intermolecular constraint loop
!> \param log_unit ...
!> \param irattle_ext ...
!> \param max_sigma ...
!> \par History
!>      Teodoro Laino [tlaino] 2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE rattle_ext_info(log_unit, irattle_ext, max_sigma)
      INTEGER, INTENT(IN)                                :: log_unit, irattle_ext
      REAL(KIND=dp), INTENT(IN)                          :: max_sigma

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rattle_ext_info', &
         routineP = moduleN//':'//routineN

      IF (log_unit > 0) THEN
         ! Dump info if requested
         WRITE (log_unit, '("RATTLE_INFO|",1X,A,I6,A,F15.9)') &
            "External Rattle     Nr. Iterations:", irattle_ext, &
            " Max. Err.:", max_sigma
      END IF
      ! Notify a not converged RATTLE
      IF (irattle_ext > Max_shake_Iter) &
         CALL cp_warn(cp__l("constraint.F",830), &
                      "Rattle NOT converged in "//cp_to_string(Max_Shake_Iter)//" iterations in the "// &
                      " intermolecular constraint. CP2K continues but results could be meaningless. ")
   END SUBROUTINE rattle_ext_info

! **************************************************************************************************
!> \brief Updates the TARGET of the COLLECTIVE constraints if the growth speed
!>        is different from zero.
!> \param gci ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param dt ...
!> \param root_section ...
!> \date    02.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE shake_update_targets(gci, local_molecules, molecule_set, &
                                   molecule_kind_set, dt, root_section)

      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      REAL(kind=dp), INTENT(in)                          :: dt
      TYPE(section_vals_type), POINTER                   :: root_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'shake_update_targets', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ikind, imol, nkind, &
                                                            nmol_per_kind
      LOGICAL                                            :: do_ext_constraint
      TYPE(colvar_counters)                              :: ncolv
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(section_vals_type), POINTER                   :: motion_section

      CALL timeset(routineN, handle)
      motion_section => section_vals_get_subs_vals(root_section, "MOTION")
      nkind = SIZE(molecule_kind_set)
      do_ext_constraint = (gci%ntot /= 0)
      ! Intramolecular Constraints
      MOL: DO ikind = 1, nkind
         nmol_per_kind = local_molecules%n_el(ikind)
         DO imol = 1, nmol_per_kind
            i = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set(i)
            molecule_kind => molecule%molecule_kind
            CALL get_molecule_kind(molecule_kind, ncolv=ncolv)

            ! Updating TARGETS for Collective Variables only
            IF (ncolv%ntot /= 0) CALL shake_update_colv_int(molecule, dt, motion_section)
         END DO
      END DO MOL
      ! Intermolecular constraints
      IF (do_ext_constraint) THEN
         ! Collective Variables
         IF (gci%ncolv%ntot /= 0) CALL shake_update_colv_ext(gci, dt, motion_section)
      END IF
      CALL timestop(handle)
   END SUBROUTINE shake_update_targets

END MODULE constraint
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Util force_env module
!> \author Teodoro Laino [tlaino] - 02.2011
! **************************************************************************************************
MODULE force_env_utils

   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE cell_types,                      ONLY: cell_type
   USE constraint,                      ONLY: rattle_control,&
                                              shake_control
   USE constraint_util,                 ONLY: getold
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: det_3x3,&
                                              jacobi
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_types_new,              ONLY: global_constraint_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: update_particle_set
   USE physcon,                         ONLY: pascal
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_env_utils'

   PUBLIC :: force_env_shake, &
             force_env_rattle, &
             rescale_forces, &
             write_stress_tensor, &
             write_forces

CONTAINS

! **************************************************************************************************
!> \brief perform shake (enforcing of constraints)
!> \param force_env the force env to shake
!> \param dt the dt for shake (if you are not interested in the velocities
!>        it can be any positive number)
!> \param shake_tol the tolerance for shake
!> \param log_unit if >0 then some information on the shake is printed,
!>        defaults to -1
!> \param lagrange_mult ...
!> \param dump_lm ...
!> \param pos ...
!> \param vel ...
!> \param compold ...
!> \param reset ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE force_env_shake(force_env, dt, shake_tol, log_unit, lagrange_mult, dump_lm, &
                              pos, vel, compold, reset)

      TYPE(force_env_type), POINTER                      :: force_env
      REAL(kind=dp), INTENT(IN), OPTIONAL                :: dt
      REAL(kind=dp), INTENT(IN)                          :: shake_tol
      INTEGER, INTENT(in), OPTIONAL                      :: log_unit, lagrange_mult
      LOGICAL, INTENT(IN), OPTIONAL                      :: dump_lm
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL, TARGET                                :: pos, vel
      LOGICAL, INTENT(IN), OPTIONAL                      :: compold, reset

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_shake', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, iparticle, iparticle_kind, iparticle_local, j, my_lagrange_mult, &
         my_log_unit, nparticle_kind, nparticle_local
      LOGICAL                                            :: has_pos, has_vel, my_dump_lm
      REAL(KIND=dp)                                      :: mydt
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: my_pos, my_vel
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(particle_list_type), POINTER                  :: particles

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_utils.F",97)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_utils.F",98)
      my_log_unit = -1
      IF (PRESENT(log_unit)) my_log_unit = log_unit
      my_lagrange_mult = -1
      IF (PRESENT(lagrange_mult)) my_lagrange_mult = lagrange_mult
      my_dump_lm = .FALSE.
      IF (PRESENT(dump_lm)) my_dump_lm = dump_lm
      NULLIFY (subsys, cell, molecules, molecule_kinds, local_molecules, particles, &
               my_pos, my_vel, gci)
      IF (PRESENT(pos)) my_pos => pos
      IF (PRESENT(vel)) my_vel => vel
      mydt = 0.1_dp
      IF (PRESENT(dt)) mydt = dt
      CALL force_env_get(force_env, subsys=subsys, cell=cell)
      CALL cp_subsys_get(subsys, &
                         atomic_kinds=atomic_kinds, &
                         local_molecules_new=local_molecules, &
                         local_particles=local_particles, &
                         molecules_new=molecules, &
                         molecule_kinds_new=molecule_kinds, &
                         particles=particles, &
                         gci=gci)
      nparticle_kind = atomic_kinds%n_els
      IF (PRESENT(compold)) THEN
         IF (compold) THEN
            CALL getold(gci, local_molecules, molecules%els, molecule_kinds%els, &
                        particles%els, cell)
         END IF
      END IF
      has_pos = .FALSE.
      IF (.NOT. ASSOCIATED(my_pos)) THEN
         has_pos = .TRUE.
         ALLOCATE (my_pos(3, particles%n_els))
         my_pos = 0.0_dp
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               my_pos(:, iparticle) = particles%els(iparticle)%r(:)
            END DO
         END DO
      END IF
      has_vel = .FALSE.
      IF (.NOT. ASSOCIATED(my_vel)) THEN
         has_vel = .TRUE.
         ALLOCATE (my_vel(3, particles%n_els))
         my_vel = 0.0_dp
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               my_vel(:, iparticle) = particles%els(iparticle)%v(:)
            END DO
         END DO
      END IF

      CALL shake_control(gci=gci, local_molecules=local_molecules, &
                         molecule_set=molecules%els, molecule_kind_set=molecule_kinds%els, &
                         particle_set=particles%els, pos=my_pos, vel=my_vel, dt=mydt, &
                         shake_tol=shake_tol, log_unit=my_log_unit, lagrange_mult=my_lagrange_mult, &
                         dump_lm=my_dump_lm, cell=cell, group=force_env%para_env%group, &
                         local_particles=local_particles)

      ! Possibly reset the lagrange multipliers
      IF (PRESENT(reset)) THEN
         IF (reset) THEN
            ! Reset Intramolecular constraints
            DO i = 1, SIZE(molecules%els)
               IF (ASSOCIATED(molecules%els(i)%lci%lcolv)) THEN
                  DO j = 1, SIZE(molecules%els(i)%lci%lcolv)
                     ! Reset langrange multiplier
                     molecules%els(i)%lci%lcolv(j)%lambda = 0.0_dp
                  END DO
               END IF
               IF (ASSOCIATED(molecules%els(i)%lci%lg3x3)) THEN
                  DO j = 1, SIZE(molecules%els(i)%lci%lg3x3)
                     ! Reset langrange multiplier
                     molecules%els(i)%lci%lg3x3(j)%lambda = 0.0_dp
                  END DO
               END IF
               IF (ASSOCIATED(molecules%els(i)%lci%lg4x6)) THEN
                  DO j = 1, SIZE(molecules%els(i)%lci%lg4x6)
                     ! Reset langrange multiplier
                     molecules%els(i)%lci%lg4x6(j)%lambda = 0.0_dp
                  END DO
               END IF
            END DO
            ! Reset Intermolecular constraints
            IF (ASSOCIATED(gci)) THEN
               IF (ASSOCIATED(gci%lcolv)) THEN
                  DO j = 1, SIZE(gci%lcolv)
                     ! Reset langrange multiplier
                     gci%lcolv(j)%lambda = 0.0_dp
                  END DO
               END IF
               IF (ASSOCIATED(gci%lg3x3)) THEN
                  DO j = 1, SIZE(gci%lg3x3)
                     ! Reset langrange multiplier
                     gci%lg3x3(j)%lambda = 0.0_dp
                  END DO
               END IF
               IF (ASSOCIATED(gci%lg4x6)) THEN
                  DO j = 1, SIZE(gci%lg4x6)
                     ! Reset langrange multiplier
                     gci%lg4x6(j)%lambda = 0.0_dp
                  END DO
               END IF
            END IF
         END IF
      END IF

      IF (has_pos) THEN
         CALL update_particle_set(particles%els, force_env%para_env%group, pos=my_pos)
         DEALLOCATE (my_pos)
      END IF
      IF (has_vel) THEN
         CALL update_particle_set(particles%els, force_env%para_env%group, vel=my_vel)
         DEALLOCATE (my_vel)
      END IF
      CALL timestop(handle)
   END SUBROUTINE force_env_shake

! **************************************************************************************************
!> \brief perform rattle (enforcing of constraints on velocities)
!>      This routine can be easily adapted to performe rattle on whatever
!>      other vector different from forces..
!> \param force_env the force env to shake
!> \param dt the dt for shake (if you are not interested in the velocities
!>        it can be any positive number)
!> \param shake_tol the tolerance for shake
!> \param log_unit if >0 then some information on the shake is printed,
!>        defaults to -1
!> \param lagrange_mult ...
!> \param dump_lm ...
!> \param vel ...
!> \param reset ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE force_env_rattle(force_env, dt, shake_tol, log_unit, lagrange_mult, dump_lm, &
                               vel, reset)

      TYPE(force_env_type), POINTER                      :: force_env
      REAL(kind=dp), INTENT(in), OPTIONAL                :: dt
      REAL(kind=dp), INTENT(in)                          :: shake_tol
      INTEGER, INTENT(in), OPTIONAL                      :: log_unit, lagrange_mult
      LOGICAL, INTENT(IN), OPTIONAL                      :: dump_lm
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL, TARGET                                :: vel
      LOGICAL, INTENT(IN), OPTIONAL                      :: reset

      CHARACTER(len=*), PARAMETER :: routineN = 'force_env_rattle', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, i, iparticle, iparticle_kind, iparticle_local, j, my_lagrange_mult, &
         my_log_unit, nparticle_kind, nparticle_local
      LOGICAL                                            :: has_vel, my_dump_lm
      REAL(KIND=dp)                                      :: mydt
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: my_vel
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(particle_list_type), POINTER                  :: particles

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_utils.F",266)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_utils.F",267)
      my_log_unit = -1
      IF (PRESENT(log_unit)) my_log_unit = log_unit
      my_lagrange_mult = -1
      IF (PRESENT(lagrange_mult)) my_lagrange_mult = lagrange_mult
      my_dump_lm = .FALSE.
      IF (PRESENT(dump_lm)) my_dump_lm = dump_lm
      NULLIFY (subsys, cell, molecules, molecule_kinds, local_molecules, particles, &
               my_vel)
      IF (PRESENT(vel)) my_vel => vel
      mydt = 0.1_dp
      IF (PRESENT(dt)) mydt = dt
      CALL force_env_get(force_env, subsys=subsys, cell=cell)
      CALL cp_subsys_get(subsys, &
                         atomic_kinds=atomic_kinds, &
                         local_molecules_new=local_molecules, &
                         local_particles=local_particles, &
                         molecules_new=molecules, &
                         molecule_kinds_new=molecule_kinds, &
                         particles=particles, &
                         gci=gci)
      nparticle_kind = atomic_kinds%n_els
      has_vel = .FALSE.
      IF (.NOT. ASSOCIATED(my_vel)) THEN
         has_vel = .TRUE.
         ALLOCATE (my_vel(3, particles%n_els))
         my_vel = 0.0_dp
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               my_vel(:, iparticle) = particles%els(iparticle)%v(:)
            END DO
         END DO
      END IF

      CALL rattle_control(gci=gci, local_molecules=local_molecules, &
                          molecule_set=molecules%els, molecule_kind_set=molecule_kinds%els, &
                          particle_set=particles%els, vel=my_vel, dt=mydt, &
                          rattle_tol=shake_tol, log_unit=my_log_unit, lagrange_mult=my_lagrange_mult, &
                          dump_lm=my_dump_lm, cell=cell, group=force_env%para_env%group, &
                          local_particles=local_particles)

      ! Possibly reset the lagrange multipliers
      IF (PRESENT(reset)) THEN
         IF (reset) THEN
            ! Reset Intramolecular constraints
            DO i = 1, SIZE(molecules%els)
               IF (ASSOCIATED(molecules%els(i)%lci%lcolv)) THEN
                  DO j = 1, SIZE(molecules%els(i)%lci%lcolv)
                     ! Reset langrange multiplier
                     molecules%els(i)%lci%lcolv(j)%lambda = 0.0_dp
                  END DO
               END IF
               IF (ASSOCIATED(molecules%els(i)%lci%lg3x3)) THEN
                  DO j = 1, SIZE(molecules%els(i)%lci%lg3x3)
                     ! Reset langrange multiplier
                     molecules%els(i)%lci%lg3x3(j)%lambda = 0.0_dp
                  END DO
               END IF
               IF (ASSOCIATED(molecules%els(i)%lci%lg4x6)) THEN
                  DO j = 1, SIZE(molecules%els(i)%lci%lg4x6)
                     ! Reset langrange multiplier
                     molecules%els(i)%lci%lg4x6(j)%lambda = 0.0_dp
                  END DO
               END IF
            END DO
            ! Reset Intermolecular constraints
            IF (ASSOCIATED(gci)) THEN
               IF (ASSOCIATED(gci%lcolv)) THEN
                  DO j = 1, SIZE(gci%lcolv)
                     ! Reset langrange multiplier
                     gci%lcolv(j)%lambda = 0.0_dp
                  END DO
               END IF
               IF (ASSOCIATED(gci%lg3x3)) THEN
                  DO j = 1, SIZE(gci%lg3x3)
                     ! Reset langrange multiplier
                     gci%lg3x3(j)%lambda = 0.0_dp
                  END DO
               END IF
               IF (ASSOCIATED(gci%lg4x6)) THEN
                  DO j = 1, SIZE(gci%lg4x6)
                     ! Reset langrange multiplier
                     gci%lg4x6(j)%lambda = 0.0_dp
                  END DO
               END IF
            END IF
         END IF
      END IF

      IF (has_vel) THEN
         CALL update_particle_set(particles%els, force_env%para_env%group, vel=my_vel)
      END IF
      DEALLOCATE (my_vel)
      CALL timestop(handle)
   END SUBROUTINE force_env_rattle

! **************************************************************************************************
!> \brief Rescale forces if requested
!> \param force_env the force env to shake
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE rescale_forces(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rescale_forces', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iparticle
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: force(3), max_value, mod_force
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: rescale_force_section

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("force_env_utils.F",383)
      IF(.NOT.(force_env%ref_count > 0))CALL cp__a("force_env_utils.F",384)
      rescale_force_section => section_vals_get_subs_vals(force_env%force_env_section, "RESCALE_FORCES")
      CALL section_vals_get(rescale_force_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(rescale_force_section, "MAX_FORCE", r_val=max_value)
         CALL force_env_get(force_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)
         DO iparticle = 1, SIZE(particles%els)
            force = particles%els(iparticle)%f(:)
            mod_force = SQRT(DOT_PRODUCT(force, force))
            IF ((mod_force > max_value) .AND. (mod_force /= 0.0_dp)) THEN
               force = force/mod_force*max_value
               particles%els(iparticle)%f(:) = force
            END IF
         END DO
      END IF
      CALL timestop(handle)
   END SUBROUTINE rescale_forces

! **************************************************************************************************
!> \brief Variable precision output of the stress tensor
!>
!> \param pv_virial ...
!> \param output_unit ...
!> \param cell ...
!> \param ndigits ...
!> \param numerical ...
!> \author MK (26.08.2010)
! **************************************************************************************************
   SUBROUTINE write_stress_tensor(pv_virial, output_unit, cell, ndigits, numerical)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: pv_virial
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: ndigits
      LOGICAL, INTENT(IN)                                :: numerical

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_stress_tensor', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=15)                                  :: fmtstr3
      CHARACTER(LEN=16)                                  :: fmtstr4
      CHARACTER(LEN=22)                                  :: fmtstr2
      CHARACTER(LEN=27)                                  :: fmtstr5
      CHARACTER(LEN=31)                                  :: fmtstr1
      INTEGER                                            :: n
      REAL(KIND=dp), DIMENSION(3)                        :: eigval
      REAL(KIND=dp), DIMENSION(3, 3)                     :: eigvec, stress_tensor

      IF (output_unit > 0) THEN
         IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("force_env_utils.F",434)
         stress_tensor(:, :) = pv_virial(:, :)/cell%deth*pascal*1.0E-9_dp
         n = MIN(MAX(1, ndigits), 20)
         fmtstr1 = "(/,T2,A,/,/,T13,A1,2(  X,A1))"
         WRITE (UNIT=fmtstr1(22:23), FMT="(I2)") n+7
         fmtstr2 = "(T3,A,T5,3(1X,F  .  ))"
         WRITE (UNIT=fmtstr2(16:17), FMT="(I2)") n+7
         WRITE (UNIT=fmtstr2(19:20), FMT="(I2)") n
         fmtstr3 = "(/,T3,A,F  .  )"
         WRITE (UNIT=fmtstr3(10:11), FMT="(I2)") n+8
         WRITE (UNIT=fmtstr3(13:14), FMT="(I2)") n
         IF (numerical) THEN
            WRITE (UNIT=output_unit, FMT=fmtstr1) &
               "NUMERICAL STRESS TENSOR [GPa]", "X", "Y", "Z"
         ELSE
            WRITE (UNIT=output_unit, FMT=fmtstr1) &
               "STRESS TENSOR [GPa]", "X", "Y", "Z"
         END IF
         WRITE (UNIT=output_unit, FMT=fmtstr2) "X", stress_tensor(1, 1:3)
         WRITE (UNIT=output_unit, FMT=fmtstr2) "Y", stress_tensor(2, 1:3)
         WRITE (UNIT=output_unit, FMT=fmtstr2) "Z", stress_tensor(3, 1:3)
         fmtstr4 = "(/,T3,A,ES  .  )"
         WRITE (UNIT=fmtstr4(11:12), FMT="(I2)") n+8
         WRITE (UNIT=fmtstr4(14:15), FMT="(I2)") n
         WRITE (UNIT=output_unit, FMT=fmtstr4) &
            "1/3 Trace(stress tensor): ", (stress_tensor(1, 1)+ &
                                           stress_tensor(2, 2)+ &
                                           stress_tensor(3, 3))/3.0_dp, &
            "Det(stress tensor)      : ", det_3x3(stress_tensor(:, 1), &
                                                  stress_tensor(:, 2), &
                                                  stress_tensor(:, 3))
         eigval(:) = 0.0_dp
         eigvec(:, :) = 0.0_dp
         CALL jacobi(stress_tensor, eigval, eigvec)
         fmtstr5 = "(/,/,T2,A,/,/,T5,3F  .  ,/)"
         WRITE (UNIT=fmtstr5(20:21), FMT="(I2)") n+8
         WRITE (UNIT=fmtstr5(23:24), FMT="(I2)") n
         WRITE (UNIT=output_unit, FMT=fmtstr5) &
            "EIGENVECTORS AND EIGENVALUES OF THE STRESS TENSOR", &
            eigval(1:3)
         WRITE (UNIT=output_unit, FMT=fmtstr2) " ", eigvec(1, 1:3)
         WRITE (UNIT=output_unit, FMT=fmtstr2) " ", eigvec(2, 1:3)
         WRITE (UNIT=output_unit, FMT=fmtstr2) " ", eigvec(3, 1:3)
      END IF

   END SUBROUTINE write_stress_tensor

! **************************************************************************************************
!> \brief Write forces
!>
!> \param particles ...
!> \param output_unit ...
!> \param label ...
!> \param ndigits ...
!> \param total_force ...
!> \param grand_total_force ...
!> \author MK (06.09.2010)
! **************************************************************************************************
   SUBROUTINE write_forces(particles, output_unit, label, ndigits, total_force, &
                           grand_total_force)

      TYPE(particle_list_type), POINTER                  :: particles
      INTEGER, INTENT(IN)                                :: output_unit
      CHARACTER(LEN=*), INTENT(IN)                       :: label
      INTEGER, INTENT(IN)                                :: ndigits
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: total_force
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT), &
         OPTIONAL                                        :: grand_total_force

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_forces', routineP = moduleN//':'//routineN

      CHARACTER(LEN=23)                                  :: fmtstr3
      CHARACTER(LEN=36)                                  :: fmtstr2
      CHARACTER(LEN=46)                                  :: fmtstr1
      INTEGER                                            :: i, ikind, iparticle, n

      IF (output_unit > 0) THEN
         IF(.NOT.(ASSOCIATED(particles)))CALL cp__a("force_env_utils.F",511)
         n = MIN(MAX(1, ndigits), 20)
         fmtstr1 = "(/,T2,A,/,/,T2,A,T11,A,T18,A,T35,A1,2(  X,A1))"
         WRITE (UNIT=fmtstr1(39:40), FMT="(I2)") n+6
         fmtstr2 = "(T2,I6,1X,I6,T21,A,T28,3(1X,F  .  ))"
         WRITE (UNIT=fmtstr2(33:34), FMT="(I2)") n
         WRITE (UNIT=fmtstr2(30:31), FMT="(I2)") n+6
         fmtstr3 = "(T2,A,T28,4(1X,F  .  ))"
         WRITE (UNIT=fmtstr3(20:21), FMT="(I2)") n
         WRITE (UNIT=fmtstr3(17:18), FMT="(I2)") n+6
         WRITE (UNIT=output_unit, FMT=fmtstr1) &
            label//" FORCES in [a.u.]", "# Atom", "Kind", "Element", "X", "Y", "Z"
         total_force(1:3) = 0.0_dp
         DO iparticle = 1, particles%n_els
            ikind = particles%els(iparticle)%atomic_kind%kind_number
            IF (particles%els(iparticle)%atom_index /= 0) THEN
               i = particles%els(iparticle)%atom_index
            ELSE
               i = iparticle
            END IF
            WRITE (UNIT=output_unit, FMT=fmtstr2) &
               i, ikind, particles%els(iparticle)%atomic_kind%element_symbol, &
               particles%els(iparticle)%f(1:3)
            total_force(1:3) = total_force(1:3)+particles%els(iparticle)%f(1:3)
         END DO
         WRITE (UNIT=output_unit, FMT=fmtstr3) &
            "SUM OF "//label//" FORCES", total_force(1:3), SQRT(SUM(total_force(:)**2))
      END IF

      IF (PRESENT(grand_total_force)) THEN
         grand_total_force(1:3) = grand_total_force(1:3)+total_force(1:3)
         WRITE (UNIT=output_unit, FMT="(A)") ""
         WRITE (UNIT=output_unit, FMT=fmtstr3) &
            "GRAND TOTAL FORCE", grand_total_force(1:3), SQRT(SUM(grand_total_force(:)**2))
      END IF

   END SUBROUTINE write_forces

END MODULE force_env_utils
