!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Update a QM/MM calculations with force mixing
!> \par History
!>      5.2004 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qmmmx_update
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_restart_force_eval,        ONLY: update_force_eval
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_release,&
                                              section_vals_type
   USE qmmm_create,                     ONLY: qmmm_env_create
   USE qmmm_types,                      ONLY: qmmm_env_get
   USE qmmmx_types,                     ONLY: qmmmx_env_release,&
                                              qmmmx_env_type
   USE qmmmx_util,                      ONLY: setup_force_mixing_qmmm_sections,&
                                              update_force_mixing_labels
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmmx_update'

   PUBLIC :: qmmmx_update_force_env

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
!> \param root_section ...
! **************************************************************************************************
   SUBROUTINE qmmmx_update_force_env(force_env, root_section)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_vals_type), POINTER                   :: root_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmmx_update_force_env', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: force_mixing_active, labels_changed
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds, new_atomic_kinds
      TYPE(cp_subsys_type), POINTER                      :: subsys, subsys_new
      TYPE(distribution_1d_type), POINTER                :: local_particles, new_local_particles
      TYPE(qmmmx_env_type), POINTER                      :: new_qmmmx_env
      TYPE(section_vals_type), POINTER                   :: qmmm_core_section, &
                                                            qmmm_extended_Section, &
                                                            qmmm_force_mixing, qmmm_section, &
                                                            subsys_section

! check everything for not null, because sometimes (e.g. metadynamics in parallel) it happens

      IF (.NOT. ASSOCIATED(force_env)) RETURN
      IF (.NOT. ASSOCIATED(force_env%force_env_section)) RETURN
      ! these two should never happen, because the sections exist, but just in case...
      qmmm_section => section_vals_get_subs_vals(force_env%force_env_section, "QMMM", can_return_null=.TRUE.)
      IF (.NOT. ASSOCIATED(qmmm_section)) RETURN
      qmmm_force_mixing => section_vals_get_subs_vals(qmmm_section, "FORCE_MIXING", can_return_null=.TRUE.)
      IF (.NOT. ASSOCIATED(qmmm_force_mixing)) RETURN
      CALL section_vals_get(qmmm_force_mixing, explicit=force_mixing_active)
      IF (.NOT. force_mixing_active) RETURN
      IF (.NOT. ASSOCIATED(force_env%qmmmx_env)) CALL cp__b("qmmmx_update.F",75,"force_env%qmmmx not associated")

      CALL force_env_get(force_env, subsys=subsys)
      CALL update_force_mixing_labels(subsys, qmmm_section, labels_changed=labels_changed)
      IF (.NOT. labels_changed) RETURN
      CALL cp__w("qmmmx_update.F",80,"Adaptive force-mixing labels changed, rebuilding QM/MM calculations! ")

      CALL update_force_eval(force_env, root_section, .FALSE.)

      ALLOCATE (new_qmmmx_env)

      ! using CUR_INDICES and CUR_LABELS, create appropriate QM_KIND sections for two QM/MM calculations
      CALL setup_force_mixing_qmmm_sections(subsys, qmmm_section, qmmm_core_section, qmmm_extended_section)

      subsys_section => section_vals_get_subs_vals(force_env%force_env_section, "SUBSYS")
      ![ADAPT] no sure about use_motion_section
      CALL qmmm_env_create(new_qmmmx_env%core, &
                           force_env%root_section, force_env%para_env, force_env%globenv, &
                           force_env%force_env_section, qmmm_core_section, subsys_section, use_motion_section=.TRUE., &
                           prev_subsys=subsys, ignore_outside_box=.TRUE.)
      CALL qmmm_env_create(new_qmmmx_env%ext, &
                           force_env%root_section, force_env%para_env, force_env%globenv, &
                           force_env%force_env_section, qmmm_extended_section, subsys_section, use_motion_section=.TRUE., &
                           prev_subsys=subsys, ignore_outside_box=.TRUE.)

      ! [NB] need to copy wiener process data, since it's not recreated when
      ! fist subsys is recreated by qmmm_env_create
      CALL qmmm_env_get(force_env%qmmmx_env%core, subsys=subsys)
      CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds, local_particles=local_particles)
      CALL qmmm_env_get(new_qmmmx_env%core, subsys=subsys_new)
      CALL cp_subsys_get(subsys_new, atomic_kinds=new_atomic_kinds, local_particles=new_local_particles)
      IF (ASSOCIATED(local_particles%local_particle_set)) THEN
         CALL copy_wiener_process(atomic_kinds, local_particles, new_atomic_kinds, new_local_particles)
      ENDIF

      CALL qmmm_env_get(force_env%qmmmx_env%ext, subsys=subsys)
      CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds, local_particles=local_particles)
      CALL qmmm_env_get(new_qmmmx_env%ext, subsys=subsys_new)
      CALL cp_subsys_get(subsys_new, atomic_kinds=new_atomic_kinds, local_particles=new_local_particles)
      IF (ASSOCIATED(local_particles%local_particle_set)) THEN
         CALL copy_wiener_process(atomic_kinds, local_particles, new_atomic_kinds, new_local_particles)
      ENDIF

      CALL section_vals_release(qmmm_core_section)
      CALL section_vals_release(qmmm_extended_section)

      ! release old qmmmx_env and point to new one
      CALL qmmmx_env_release(force_env%qmmmx_env)
      force_env%qmmmx_env => new_qmmmx_env

   END SUBROUTINE qmmmx_update_force_env

! **************************************************************************************************
!> \brief ...
!> \param from_local_particle_kinds ...
!> \param from_local_particles ...
!> \param to_local_particle_kinds ...
!> \param to_local_particles ...
! **************************************************************************************************
   SUBROUTINE copy_wiener_process(from_local_particle_kinds, from_local_particles, &
                                  to_local_particle_kinds, to_local_particles)
      TYPE(atomic_kind_list_type), POINTER               :: from_local_particle_kinds
      TYPE(distribution_1d_type), POINTER                :: from_local_particles
      TYPE(atomic_kind_list_type), POINTER               :: to_local_particle_kinds
      TYPE(distribution_1d_type), POINTER                :: to_local_particles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'copy_wiener_process', &
         routineP = moduleN//':'//routineN

      INTEGER :: from_iparticle_kind, from_iparticle_local(1), from_nparticle_kind, &
         from_nparticle_local, handle, to_iparticle_global, to_iparticle_kind, to_iparticle_local, &
         to_nparticle_kind, to_nparticle_local, tot_from_nparticle_local, tot_to_nparticle_local
      LOGICAL                                            :: found_it

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(from_local_particles)))CALL cp__a("qmmmx_update.F",150)
      IF(.NOT.(ASSOCIATED(to_local_particles)))CALL cp__a("qmmmx_update.F",151)

      IF (.NOT. ASSOCIATED(from_local_particles%local_particle_set)) RETURN
      IF(.NOT.(.NOT. ASSOCIATED(to_local_particles%local_particle_set)))CALL cp__a("qmmmx_update.F",154)

      from_nparticle_kind = from_local_particle_kinds%n_els
      to_nparticle_kind = to_local_particle_kinds%n_els

      ! make sure total number of particles hasn't changed, even if particle kinds have
      tot_from_nparticle_local = 0
      DO from_iparticle_kind = 1, from_nparticle_kind
         tot_from_nparticle_local = tot_from_nparticle_local+from_local_particles%n_el(from_iparticle_kind)
      END DO
      tot_to_nparticle_local = 0
      DO to_iparticle_kind = 1, to_nparticle_kind
         tot_to_nparticle_local = tot_to_nparticle_local+to_local_particles%n_el(to_iparticle_kind)
      END DO
      IF(.NOT.(tot_from_nparticle_local == tot_to_nparticle_local))CALL cp__a("qmmmx_update.F",168)

      ALLOCATE (to_local_particles%local_particle_set(to_nparticle_kind))
      DO to_iparticle_kind = 1, to_nparticle_kind

         to_nparticle_local = to_local_particles%n_el(to_iparticle_kind)
         ALLOCATE (to_local_particles%local_particle_set(to_iparticle_kind)%rng(to_nparticle_local))

         DO to_iparticle_local = 1, to_nparticle_local
            to_iparticle_global = to_local_particles%list(to_iparticle_kind)%array(to_iparticle_local)
            ALLOCATE (to_local_particles%local_particle_set(to_iparticle_kind)%rng(to_iparticle_local)%stream)

            found_it = .FALSE.
            ! find the matching kind/index where this particle was before
            DO from_iparticle_kind = 1, from_nparticle_kind
               from_nparticle_local = from_local_particles%n_el(from_iparticle_kind)
               IF (MINVAL(ABS(from_local_particles%list(from_iparticle_kind)%array(1:from_nparticle_local)- &
                              to_iparticle_global)) == 0) THEN
                  from_iparticle_local = &
                     MINLOC(ABS(from_local_particles%list(from_iparticle_kind)%array(1:from_nparticle_local)- &
                                to_iparticle_global))
                  to_local_particles%local_particle_set(to_iparticle_kind)%rng(to_iparticle_local)%stream = &
                     from_local_particles%local_particle_set(from_iparticle_kind)%rng(from_iparticle_local(1))%stream
                  found_it = .TRUE.
                  EXIT
               ENDIF
            END DO
            IF(.NOT.(found_it))CALL cp__a("qmmmx_update.F",195)

         END DO ! to_iparticle_local

      END DO ! to_iparticle_kind
      CALL timestop(handle)

   END SUBROUTINE copy_wiener_process

END MODULE qmmmx_update
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Performs the metadynamics calculation
!> \par History
!>      01.2005 created [fawzi and ale]
!>      11.2007 Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
MODULE metadynamics
USE cp_log_handling, ONLY: cp_logger_type, cp_get_default_logger
  USE bibliography,                    ONLY: VandenCic2006
  USE cell_types,                      ONLY: cell_type
  USE colvar_methods,                  ONLY: colvar_eval_glob_f
  USE colvar_types,                    ONLY: colvar_p_type,&
                                             torsion_colvar_id
  USE constraint_fxd,                  ONLY: fix_atom_control
  USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                             cp_iterate,&
                                             cp_print_key_finished_output,&
                                             cp_print_key_unit_nr,&
                                             cp_rm_iter_level
  USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                             cp_subsys_type
  USE force_env_types,                 ONLY: force_env_get,&
                                             force_env_type
  USE input_constants,                 ONLY: do_wall_m,&
                                             do_wall_p,&
                                             do_wall_reflective
  USE input_section_types,             ONLY: section_vals_get,&
                                             section_vals_get_subs_vals,&
                                             section_vals_type,&
                                             section_vals_val_get
  USE kinds,                           ONLY: dp
  USE message_passing,                 ONLY: mp_bcast, cp2k_is_parallel
  USE metadynamics_types,              ONLY: hills_env_type,&
                                             meta_env_type,&
                                             metavar_type,&
                                             multiple_walkers_type
  USE metadynamics_utils,              ONLY: add_hill_single,&
                                             get_meta_iter_level,&
                                             meta_walls,&
                                             restart_hills,&
                                             synchronize_multiple_walkers
  USE parallel_rng_types,              ONLY: next_random_number
  USE particle_list_types,             ONLY: particle_list_type
  USE physcon,                         ONLY: boltzmann,&
                                             femtoseconds,&
                                             joule,&
                                             kelvin
  USE reference_manager,               ONLY: cite_reference
  USE simpar_types,                    ONLY: simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'metadynamics'

   PUBLIC :: metadyn_forces, metadyn_integrator
   PUBLIC :: metadyn_velocities_colvar, metadyn_write_colvar
   PUBLIC :: metadyn_initialise_plumed, metadyn_finalise_plumed

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
!> \param simpar ...
!> \param itimes ...
! **************************************************************************************************
   SUBROUTINE metadyn_initialise_plumed(force_env, simpar, itimes)

      TYPE(force_env_type), POINTER            :: force_env
      TYPE(simpar_type), POINTER               :: simpar
      INTEGER, INTENT(IN)                      :: itimes

      CHARACTER(LEN=*), PARAMETER :: routineN = 'metadyn_initialise_plumed', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle



      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(simpar)))CALL cp__a("metadynamics.F",111)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("metadynamics.F",112)


      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(itimes))==-1) EXIT ;  END DO ; ENDIF

      CALL cp_abort(cp__l("metadynamics.F",145), &
                    "Requested to use plumed for metadynamics, but cp2k was"// &
                    " not compiled with plumed support.")

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief makes sure PLUMED is shut down cleanly
! **************************************************************************************************
   SUBROUTINE metadyn_finalise_plumed()

      CHARACTER(LEN=*), PARAMETER :: routineN = 'metadyn_finalise_plumed', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)


      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief  General driver for applying metadynamics
!> \param force_env ...
!> \param itimes ...
!> \param vel ...
!> \param rand ...
!> \date   01.2009
!> \par History
!>      01.2009 created
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE metadyn_integrator(force_env, itimes, vel, rand)
      TYPE(force_env_type), POINTER            :: force_env
      INTEGER, INTENT(IN)                      :: itimes
      REAL(KIND=dp), DIMENSION(:, :), &
         INTENT(INOUT), OPTIONAL                :: vel
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, &
         POINTER                                :: rand

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_integrator', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, plumed_itimes

      CALL timeset(routineN, handle)

      ! Apply Metadynamics
      IF (ASSOCIATED(force_env%meta_env)) THEN
         IF (force_env%meta_env%use_plumed .EQV. .TRUE.) THEN
            plumed_itimes = itimes

            CALL cp_abort(cp__l("metadynamics.F",292), &
                          "Requested to use plumed for metadynamics, but cp2k was"// &
                          " not compiled with plumed support.")

         ELSE
            IF (force_env%meta_env%langevin) THEN
               IF (.NOT. PRESENT(rand)) THEN
                  CALL cp__b("metadynamics.F",300,"Langevin on COLVAR not implemented for this MD ensemble!")
               END IF
               !    *** Velocity Verlet for Langevin S(t)->S(t+1)
               CALL metadyn_position_colvar(force_env)
               !    *** Forces from Vs and  S(X(t+1))
               CALL metadyn_forces(force_env)
               !    *** Velocity Verlet for Langeving *** v(t+1/2)--> v(t)
               CALL metadyn_velocities_colvar(force_env, rand)
            ELSE
               CALL metadyn_forces(force_env, vel)
            ENDIF
            !    *** Write down COVAR informations
            CALL metadyn_write_colvar(force_env)
         END IF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE metadyn_integrator

! **************************************************************************************************
!> \brief add forces to the subsys due to the metadynamics run
!>      possibly modifies the velocites (if reflective walls are applied)
!> \param force_env ...
!> \param vel ...
!> \par History
!>      04.2004 created
! **************************************************************************************************
   SUBROUTINE metadyn_forces(force_env, vel)
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: vel

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_forces', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, i_c, icolvar, ii, iwall
      LOGICAL                                            :: explicit
      REAL(kind=dp)                                      :: check_val, diff_ss, dt, ekin_w, fac_t, &
                                                            fft, norm, rval, scal, scalf, &
                                                            ss0_test, tol_ekin
      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: colvar_p
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: ss0_section, vvp_section

      NULLIFY (logger, meta_env)
      meta_env => force_env%meta_env
      IF (.NOT. ASSOCIATED(meta_env)) RETURN

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      NULLIFY (colvar_p, subsys, cv, ss0_section, vvp_section)
      CALL force_env_get(force_env, subsys=subsys)

      dt = meta_env%dt
      IF (.NOT. meta_env%restart) meta_env%n_steps = meta_env%n_steps+1

      ! Initialize velocity
      IF (meta_env%restart .AND. meta_env%extended_lagrange) THEN
         meta_env%ekin_s = 0.0_dp
         DO i_c = 1, meta_env%n_colvar
            cv => meta_env%metavar(i_c)
            cv%vvp = next_random_number(force_env%globenv%gaussian_rng_stream)
            meta_env%ekin_s = meta_env%ekin_s+0.5_dp*cv%mass*cv%vvp**2
         END DO
         ekin_w = 0.5_dp*meta_env%temp_wanted*REAL(meta_env%n_colvar, KIND=dp)
         fac_t = SQRT(ekin_w/MAX(meta_env%ekin_s, 1.0E-8_dp))
         DO i_c = 1, meta_env%n_colvar
            cv => meta_env%metavar(i_c)
            cv%vvp = cv%vvp*fac_t
         ENDDO
         meta_env%ekin_s = 0.0_dp
      END IF

      !    *** Velocity Verlet for Langevin S(t)->S(t+1)
      ! compute ss and the derivative of ss with respect to the atomic positions
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         icolvar = cv%icolvar
         CALL colvar_eval_glob_f(icolvar, force_env)
         cv%ss = subsys%colvar_p(icolvar)%colvar%ss

         ! Setup the periodic flag if the COLVAR is (-pi,pi] periodic
         cv%periodic = (subsys%colvar_p(icolvar)%colvar%type_id == torsion_colvar_id)

         ! Restart for Extended Lagrangian Metadynamics
         IF (meta_env%restart) THEN
            ! Initialize the position of the collective variable in the extended lagrange
            ss0_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_SS0")
            CALL section_vals_get(ss0_section, explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(ss0_section, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=i_c, r_val=rval)
               cv%ss0 = rval
            ELSE
               cv%ss0 = cv%ss
            END IF
            vvp_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_VVP")
            CALL section_vals_get(vvp_section, explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(vvp_section, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=i_c, r_val=rval)
               cv%vvp = rval
            END IF
         END IF
         !
         IF (.NOT. meta_env%extended_lagrange) THEN
            cv%ss0 = cv%ss
            cv%vvp = 0.0_dp
         END IF
      ENDDO
      ! History dependent forces (evaluated at s0)
      IF (meta_env%do_hills) CALL hills(meta_env)

      ! Apply walls to the colvars
      CALL meta_walls(meta_env)

      meta_env%restart = .FALSE.
      IF (.NOT. meta_env%extended_lagrange) THEN
         meta_env%ekin_s = 0.0_dp
         meta_env%epot_s = 0.0_dp
         meta_env%epot_walls = 0.0_dp
         DO i_c = 1, meta_env%n_colvar
            cv => meta_env%metavar(i_c)
            cv%epot_s = 0.0_dp
            cv%ff_s = 0.0_dp
            meta_env%epot_walls = meta_env%epot_walls+cv%epot_walls
            icolvar = cv%icolvar
            NULLIFY (particles)
            CALL cp_subsys_get(subsys, colvar_p=colvar_p, &
                               particles=particles)
            DO ii = 1, colvar_p(icolvar)%colvar%n_atom_s
               i = colvar_p(icolvar)%colvar%i_atom(ii)
               fft = cv%ff_hills+cv%ff_walls
               particles%els(i)%f = particles%els(i)%f+fft*colvar_p(icolvar)%colvar%dsdr(:, ii)
            ENDDO
         ENDDO
      ELSE
         meta_env%ekin_s = 0.0_dp
         meta_env%epot_s = 0.0_dp
         meta_env%epot_walls = 0.0_dp
         DO i_c = 1, meta_env%n_colvar
            cv => meta_env%metavar(i_c)
            diff_ss = cv%ss-cv%ss0
            IF (cv%periodic) THEN
               ! The difference of a periodic COLVAR is always within [-pi,pi]
               diff_ss = SIGN(1.0_dp, ASIN(SIN(diff_ss)))*ACOS(COS(diff_ss))
            END IF
            cv%epot_s = 0.5_dp*cv%lambda*(diff_ss)**2.0_dp
            cv%ff_s = cv%lambda*(diff_ss)
            icolvar = cv%icolvar
            ! forces on the atoms
            NULLIFY (particles)
            CALL cp_subsys_get(subsys, colvar_p=colvar_p, &
                               particles=particles)
            DO ii = 1, colvar_p(icolvar)%colvar%n_atom_s
               i = colvar_p(icolvar)%colvar%i_atom(ii)
               particles%els(i)%f = particles%els(i)%f-cv%ff_s*colvar_p(icolvar)%colvar%dsdr(:, ii)
            ENDDO
            !  velocity verlet on the s0 if NOT langevin
            IF (.NOT. meta_env%langevin) THEN
               fft = cv%ff_s+cv%ff_hills+cv%ff_walls
               cv%vvp = cv%vvp+dt*fft/cv%mass
               meta_env%ekin_s = meta_env%ekin_s+0.5_dp*cv%mass*cv%vvp**2
               meta_env%epot_s = meta_env%epot_s+cv%epot_s
               meta_env%epot_walls = meta_env%epot_walls+cv%epot_walls
            END IF
         ENDDO
         !  velocity rescaling on the s0
         IF (meta_env%tempcontrol .AND. (.NOT. meta_env%langevin)) THEN
            ekin_w = 0.5_dp*meta_env%temp_wanted*REAL(meta_env%n_colvar, KIND=dp)
            tol_ekin = 0.5_dp*meta_env%toll_temp*REAL(meta_env%n_colvar, KIND=dp)
            IF (ABS(ekin_w-meta_env%ekin_s) > tol_ekin) THEN
               fac_t = SQRT(ekin_w/MAX(meta_env%ekin_s, 1.0E-8_dp))
               DO i_c = 1, meta_env%n_colvar
                  cv => meta_env%metavar(i_c)
                  cv%vvp = cv%vvp*fac_t
               ENDDO
               meta_env%ekin_s = ekin_w
            ENDIF
         ENDIF
         ! Reflective Wall only for s0
         DO i_c = 1, meta_env%n_colvar
            cv => meta_env%metavar(i_c)
            IF (cv%do_wall) THEN
               DO iwall = 1, SIZE(cv%walls)
                  SELECT CASE (cv%walls (iwall)%id_type)
                  CASE (do_wall_reflective)
                     ss0_test = cv%ss0+dt*cv%vvp
                     IF (cv%periodic) THEN
                        ! A periodic COLVAR is always within [-pi,pi]
                        ss0_test = SIGN(1.0_dp, ASIN(SIN(ss0_test)))*ACOS(COS(ss0_test))
                     END IF
                     SELECT CASE (cv%walls (iwall)%id_direction)
                     CASE (do_wall_p)
                        IF ((ss0_test > cv%walls(iwall)%pos) .AND. (cv%vvp > 0)) cv%vvp = -cv%vvp
                     CASE (do_wall_m)
                        IF ((ss0_test < cv%walls(iwall)%pos) .AND. (cv%vvp < 0)) cv%vvp = -cv%vvp
                     END SELECT
                  END SELECT
               END DO
            ENDIF
         ENDDO
         ! Update of ss0 if NOT langevin
         IF (.NOT. meta_env%langevin) THEN
            DO i_c = 1, meta_env%n_colvar
               cv => meta_env%metavar(i_c)
               cv%ss0 = cv%ss0+dt*cv%vvp
               IF (cv%periodic) THEN
                  ! A periodic COLVAR is always within [-pi,pi]
                  cv%ss0 = SIGN(1.0_dp, ASIN(SIN(cv%ss0)))*ACOS(COS(cv%ss0))
               END IF
            ENDDO
         END IF
      ENDIF
      ! Constraints ONLY of Fixed Atom type
      CALL fix_atom_control(force_env)

      ! Reflective Wall only for ss
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         IF (cv%do_wall) THEN
            DO iwall = 1, SIZE(cv%walls)
               SELECT CASE (cv%walls (iwall)%id_type)
               CASE (do_wall_reflective)
                  SELECT CASE (cv%walls (iwall)%id_direction)
                  CASE (do_wall_p)
                     IF (cv%ss < cv%walls(iwall)%pos) CYCLE
                     check_val = -1.0_dp
                  CASE (do_wall_m)
                     IF (cv%ss > cv%walls(iwall)%pos) CYCLE
                     check_val = 1.0_dp
                  END SELECT
                  NULLIFY (particles)
                  icolvar = cv%icolvar
                  CALL cp_subsys_get(subsys, colvar_p=colvar_p, particles=particles)
                  scal = 0.0_dp
                  scalf = 0.0_dp
                  norm = 0.0_dp
                  DO ii = 1, colvar_p(icolvar)%colvar%n_atom_s
                     i = colvar_p(icolvar)%colvar%i_atom(ii)
                     IF (PRESENT(vel)) THEN
                        scal = scal+vel(1, i)*colvar_p(icolvar)%colvar%dsdr(1, ii)
                        scal = scal+vel(2, i)*colvar_p(icolvar)%colvar%dsdr(2, ii)
                        scal = scal+vel(3, i)*colvar_p(icolvar)%colvar%dsdr(3, ii)
                     ELSE
                        scal = scal+particles%els(i)%v(1)*colvar_p(icolvar)%colvar%dsdr(1, ii)
                        scal = scal+particles%els(i)%v(2)*colvar_p(icolvar)%colvar%dsdr(2, ii)
                        scal = scal+particles%els(i)%v(3)*colvar_p(icolvar)%colvar%dsdr(3, ii)
                     END IF
                     scalf = scalf+particles%els(i)%f(1)*colvar_p(icolvar)%colvar%dsdr(1, ii)
                     scalf = scalf+particles%els(i)%f(2)*colvar_p(icolvar)%colvar%dsdr(2, ii)
                     scalf = scalf+particles%els(i)%f(3)*colvar_p(icolvar)%colvar%dsdr(3, ii)
                     norm = norm+colvar_p(icolvar)%colvar%dsdr(1, ii)**2
                     norm = norm+colvar_p(icolvar)%colvar%dsdr(2, ii)**2
                     norm = norm+colvar_p(icolvar)%colvar%dsdr(3, ii)**2
                  ENDDO
                  IF (norm /= 0.0_dp) scal = scal/norm
                  IF (norm /= 0.0_dp) scalf = scalf/norm

                  IF (scal*check_val > 0.0_dp) CYCLE
                  DO ii = 1, colvar_p(icolvar)%colvar%n_atom_s
                     i = colvar_p(icolvar)%colvar%i_atom(ii)
                     IF (PRESENT(vel)) THEN
                        vel(:, i) = vel(:, i)-2.0_dp*colvar_p(icolvar)%colvar%dsdr(:, ii)*scal
                     ELSE
                        particles%els(i)%v(:) = particles%els(i)%v(:)-2.0_dp*colvar_p(icolvar)%colvar%dsdr(:, ii)*scal
                     END IF
                     ! Nullify forces along the colvar (this avoids the weird behaviors of the reflective wall)
                     particles%els(i)%f(:) = particles%els(i)%f(:)-colvar_p(icolvar)%colvar%dsdr(:, ii)*scalf
                  ENDDO
               END SELECT
            END DO
         END IF
      ENDDO

      CALL timestop(handle)
   END SUBROUTINE metadyn_forces

! **************************************************************************************************
!> \brief Evolves velocities COLVAR according to
!>        Vanden-Eijnden Ciccotti C.Phys.Letter 429 (2006) 310-316
!> \param force_env ...
!> \param rand ...
!> \date  01.2009
!> \author Fabio Sterpone and Teodoro Laino
! **************************************************************************************************
   SUBROUTINE metadyn_velocities_colvar(force_env, rand)
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT), &
         OPTIONAL                                        :: rand

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_velocities_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_c
      REAL(kind=dp)                                      :: diff_ss, dt, fft, sigma
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv

      NULLIFY (logger, meta_env, cv)
      meta_env => force_env%meta_env
      IF (.NOT. ASSOCIATED(meta_env)) RETURN

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      ! Add citation
      IF (meta_env%langevin) CALL cite_reference(VandenCic2006)

      dt = meta_env%dt
      ! History dependent forces (evaluated at s0)
      IF (meta_env%do_hills) CALL hills(meta_env)

      ! Evolve Velocities
      meta_env%ekin_s = 0.0_dp
      meta_env%epot_walls = 0.0_dp
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         diff_ss = cv%ss-cv%ss0
         IF (cv%periodic) THEN
            ! The difference of a periodic COLVAR is always within [-pi,pi]
            diff_ss = SIGN(1.0_dp, ASIN(SIN(diff_ss)))*ACOS(COS(diff_ss))
         END IF
         cv%epot_s = 0.5_dp*cv%lambda*(diff_ss)**2.0_dp
         cv%ff_s = cv%lambda*(diff_ss)

         fft = cv%ff_s+cv%ff_hills
         sigma = SQRT((meta_env%temp_wanted*kelvin)*2.0_dp*(boltzmann/joule)*cv%gamma/cv%mass)
         cv%vvp = cv%vvp+0.5_dp*dt*fft/cv%mass-0.5_dp*dt*cv%gamma*cv%vvp+ &
                  0.5_dp*SQRT(dt)*sigma*rand(i_c)
         meta_env%ekin_s = meta_env%ekin_s+0.5_dp*cv%mass*cv%vvp**2
         meta_env%epot_walls = meta_env%epot_walls+cv%epot_walls
      ENDDO
      CALL timestop(handle)

   END SUBROUTINE metadyn_velocities_colvar

! **************************************************************************************************
!> \brief Evolves COLVAR position
!>        Vanden-Eijnden Ciccotti C.Phys.Letter 429 (2006) 310-316
!> \param force_env ...
!> \date  01.2009
!> \author Fabio Sterpone and Teodoro Laino
! **************************************************************************************************
   SUBROUTINE metadyn_position_colvar(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_position_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_c
      REAL(kind=dp)                                      :: dt
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv

      NULLIFY (logger, meta_env, cv)
      meta_env => force_env%meta_env
      IF (.NOT. ASSOCIATED(meta_env)) RETURN

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      ! Add citation
      IF (meta_env%langevin) CALL cite_reference(VandenCic2006)
      dt = meta_env%dt

      ! Update of ss0
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         cv%ss0 = cv%ss0+dt*cv%vvp
         IF (cv%periodic) THEN
            ! A periodic COLVAR is always within [-pi,pi]
            cv%ss0 = SIGN(1.0_dp, ASIN(SIN(cv%ss0)))*ACOS(COS(cv%ss0))
         END IF
      ENDDO
      CALL timestop(handle)

   END SUBROUTINE metadyn_position_colvar

! **************************************************************************************************
!> \brief Write down COLVAR information evolved  according to
!>        Vanden-Eijnden Ciccotti C.Phys.Letter 429 (2006) 310-316
!> \param force_env ...
!> \date  01.2009
!> \author Fabio Sterpone and Teodoro Laino
! **************************************************************************************************
   SUBROUTINE metadyn_write_colvar(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_write_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, i_c, iw
      REAL(KIND=dp)                                      :: diff_ss, temp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv

      NULLIFY (logger, meta_env, cv)
      meta_env => force_env%meta_env
      IF (.NOT. ASSOCIATED(meta_env)) RETURN

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      ! If Langevin we need to recompute few quantities
      ! This does not apply to the standard lagrangian scheme since it is
      ! implemented with a plain Newton integration scheme.. while Langevin
      ! follows the correct Verlet integration.. This will have to be made
      ! uniform in the future (Teodoro Laino - 01.2009)
      IF (meta_env%langevin) THEN
         meta_env%ekin_s = 0.0_dp
         meta_env%epot_s = 0.0_dp
         DO i_c = 1, meta_env%n_colvar
            cv => meta_env%metavar(i_c)
            diff_ss = cv%ss-cv%ss0
            IF (cv%periodic) THEN
               ! The difference of a periodic COLVAR is always within [-pi,pi]
               diff_ss = SIGN(1.0_dp, ASIN(SIN(diff_ss)))*ACOS(COS(diff_ss))
            END IF
            cv%epot_s = 0.5_dp*cv%lambda*(diff_ss)**2.0_dp
            cv%ff_s = cv%lambda*(diff_ss)

            meta_env%epot_s = meta_env%epot_s+cv%epot_s
            meta_env%ekin_s = meta_env%ekin_s+0.5_dp*cv%mass*cv%vvp**2
         ENDDO
      END IF

      ! write COLVAR file
      iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                "PRINT%COLVAR", extension=".metadynLog")
      IF (iw > 0) THEN
         IF (meta_env%extended_lagrange) THEN
            WRITE (iw, '(f16.8,70f15.8)') meta_env%time*femtoseconds, &
               (meta_env%metavar(i)%ss0, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ss, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_s, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_hills, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_walls, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%vvp, i=1, meta_env%n_colvar), &
               meta_env%epot_s, &
               meta_env%hills_env%energy, &
               meta_env%epot_walls, &
               (meta_env%ekin_s)*2.0_dp/(REAL(meta_env%n_colvar, KIND=dp))*kelvin
         ELSE
            WRITE (iw, '(f16.8,40f13.5)') meta_env%time*femtoseconds, &
               (meta_env%metavar(i)%ss0, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_hills, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_walls, i=1, meta_env%n_colvar), &
               meta_env%hills_env%energy, &
               meta_env%epot_walls
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                        "PRINT%COLVAR")

      ! Temperature for COLVAR
      IF (meta_env%extended_lagrange) THEN
         temp = meta_env%ekin_s*2.0_dp/(REAL(meta_env%n_colvar, KIND=dp))*kelvin
         meta_env%avg_temp = (meta_env%avg_temp*REAL(meta_env%n_steps, KIND=dp)+ &
                              temp)/REAL(meta_env%n_steps+1, KIND=dp)
         iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                   "PRINT%TEMPERATURE_COLVAR", extension=".metadynLog")
         IF (iw > 0) THEN
            WRITE (iw, '(T2,79("-"))')
            WRITE (iw, '( A,T51,f10.2,T71,f10.2)') ' COLVARS INSTANTANEOUS/AVERAGE TEMPERATURE ', &
               temp, meta_env%avg_temp
            WRITE (iw, '(T2,79("-"))')
         ENDIF
         CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                           "PRINT%TEMPERATURE_COLVAR")
      END IF
      CALL timestop(handle)

   END SUBROUTINE metadyn_write_colvar

! **************************************************************************************************
!> \brief Major driver for adding hills and computing forces due to the history
!>        dependent term
!> \param meta_env ...
!> \par History
!>      04.2004 created
!>      10.2008 Teodoro Laino [tlaino] - University of Zurich
!>              Major rewriting and addition of multiple walkers
! **************************************************************************************************
   SUBROUTINE hills(meta_env)
      TYPE(meta_env_type), POINTER                       :: meta_env

      CHARACTER(len=*), PARAMETER :: routineN = 'hills', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, i_hills, ih, intermeta_steps, &
                                                            iter_nr, iw, n_colvar, n_hills_start, &
                                                            n_step
      LOGICAL                                            :: force_gauss
      REAL(KIND=dp)                                      :: cut_func, dfunc, diff, dp2, frac, &
                                                            slow_growth, V_now_here, V_to_fes, &
                                                            wtww, ww
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ddp, denf, diff_ss, local_last_hills, &
                                                            numf
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(hills_env_type), POINTER                      :: hills_env
      TYPE(metavar_type), DIMENSION(:), POINTER          :: colvars
      TYPE(multiple_walkers_type), POINTER               :: multiple_walkers

      CALL timeset(routineN, handle)

      NULLIFY (hills_env, multiple_walkers, logger, colvars, ddp, local_last_hills)
      hills_env => meta_env%hills_env
      logger => cp_get_default_logger()
      colvars => meta_env%metavar
      n_colvar = meta_env%n_colvar
      n_step = meta_env%n_steps

      ! Create a temporary logger level specific for metadynamics
      CALL cp_add_iter_level(logger%iter_info, "METADYNAMICS")
      CALL get_meta_iter_level(meta_env, iter_nr)
      CALL cp_iterate(logger%iter_info, last=.FALSE., iter_nr=iter_nr)

      ! Set-up restart if any
      IF (meta_env%hills_env%restart) THEN
         meta_env%hills_env%restart = .FALSE.
         IF (meta_env%well_tempered) THEN
            CALL restart_hills(hills_env%ss_history, hills_env%delta_s_history, hills_env%ww_history, &
                               hills_env%ww, hills_env%n_hills, n_colvar, colvars, meta_env%metadyn_section, &
                               invdt_history=hills_env%invdt_history)
         ELSE
            CALL restart_hills(hills_env%ss_history, hills_env%delta_s_history, hills_env%ww_history, &
                               hills_env%ww, hills_env%n_hills, n_colvar, colvars, meta_env%metadyn_section)
         END IF
      END IF

      ! Proceed with normal calculation
      intermeta_steps = n_step-hills_env%old_hill_step
      force_gauss = .FALSE.
      IF ((hills_env%min_disp > 0.0_dp) .AND. (hills_env%old_hill_number > 0) .AND. &
          (intermeta_steps >= hills_env%min_nt_hills)) THEN
         ALLOCATE (ddp(meta_env%n_colvar))
         ALLOCATE (local_last_hills(meta_env%n_colvar))

         local_last_hills(1:n_colvar) = hills_env%ss_history(1:n_colvar, hills_env%old_hill_number)

         !RG Calculate the displacement
         dp2 = 0.0_dp
         DO i = 1, n_colvar
            ddp(i) = colvars(i)%ss0-local_last_hills(i)
            IF (colvars(i)%periodic) THEN
               ! The difference of a periodic COLVAR is always within [-pi,pi]
               ddp(i) = SIGN(1.0_dp, ASIN(SIN(ddp(i))))*ACOS(COS(ddp(i)))
            END IF
            dp2 = dp2+ddp(i)**2
         ENDDO
         dp2 = SQRT(dp2)
         IF (dp2 > hills_env%min_disp) THEN
            force_gauss = .TRUE.
            iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                      "PRINT%PROGRAM_RUN_INFO", extension=".metadynLog")
            IF (iw > 0) THEN
               WRITE (UNIT=iw, FMT="(A,F0.6,A,F0.6)") &
                  " METADYN| Threshold value for COLVAR displacement exceeded: ", &
                  dp2, " > ", hills_env%min_disp
            END IF
            CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                              "PRINT%PROGRAM_RUN_INFO")
         END IF
         DEALLOCATE (ddp)
         DEALLOCATE (local_last_hills)
      END IF

      !RG keep into account adaptive hills
      IF (((MODULO(intermeta_steps, hills_env%nt_hills) == 0) .OR. force_gauss) &
          .AND. (.NOT. meta_env%restart) .AND. (hills_env%nt_hills > 0)) THEN
         IF (meta_env%do_multiple_walkers) multiple_walkers => meta_env%multiple_walkers

         n_hills_start = hills_env%n_hills
         ! Add the hill corresponding to this location
         IF (meta_env%well_tempered) THEN
            ! Well-Tempered scaling of hills height
            V_now_here = 0._dp
            DO ih = 1, hills_env%n_hills
               dp2 = 0._dp
               DO i = 1, n_colvar
                  diff = colvars(i)%ss0-hills_env%ss_history(i, ih)
                  IF (colvars(i)%periodic) THEN
                     ! The difference of a periodic COLVAR is always within [-pi,pi]
                     diff = SIGN(1.0_dp, ASIN(SIN(diff)))*ACOS(COS(diff))
                  END IF
                  diff = (diff)/hills_env%delta_s_history(i, ih)
                  dp2 = dp2+diff**2
               ENDDO
               V_to_fes = 1.0_dp+meta_env%wttemperature*hills_env%invdt_history(ih)
               V_now_here = V_now_here+hills_env%ww_history(ih)/V_to_fes*EXP(-0.5_dp*dp2)
            ENDDO
            wtww = hills_env%ww*EXP(-V_now_here*meta_env%invdt)
            ww = wtww*(1.0_dp+meta_env%wttemperature*meta_env%invdt)
            CALL add_hill_single(hills_env, colvars, ww, hills_env%n_hills, n_colvar, meta_env%invdt)
         ELSE
            CALL add_hill_single(hills_env, colvars, hills_env%ww, hills_env%n_hills, n_colvar)
         END IF
         ! Update local n_hills counter
         IF (meta_env%do_multiple_walkers) multiple_walkers%n_hills_local = multiple_walkers%n_hills_local+1

         hills_env%old_hill_number = hills_env%n_hills
         hills_env%old_hill_step = n_step

         ! Update iteration level for printing
         CALL get_meta_iter_level(meta_env, iter_nr)
         CALL cp_iterate(logger%iter_info, last=.FALSE., iter_nr=iter_nr)

         ! Print just program_run_info
         iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                   "PRINT%PROGRAM_RUN_INFO", extension=".metadynLog")
         IF (iw > 0) THEN
            IF (meta_env%do_multiple_walkers) THEN
               WRITE (iw, '(/,1X,"METADYN|",A,I0,A,I0,A,/)') &
                  ' Global/Local Hills number (', hills_env%n_hills, '/', multiple_walkers%n_hills_local, &
                  ') added.'
            ELSE
               WRITE (iw, '(/,1X,"METADYN|",A,I0,A,/)') ' Hills number (', hills_env%n_hills, ') added.'
            END IF
         END IF
         CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                           "PRINT%PROGRAM_RUN_INFO")

         ! Handle Multiple Walkers
         IF (meta_env%do_multiple_walkers) THEN
            ! Print Local Hills file if requested
            iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                      "PRINT%HILLS", middle_name="LOCAL", extension=".metadynLog")
            IF (iw > 0) THEN
               WRITE (iw, '(f12.1,30f13.5)') meta_env%time*femtoseconds, &
                  (hills_env%ss_history(ih, hills_env%n_hills), ih=1, n_colvar), &
                  (hills_env%delta_s_history(ih, hills_env%n_hills), ih=1, n_colvar), &
                  hills_env%ww_history(hills_env%n_hills)
            END IF
            CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                              "PRINT%HILLS")

            ! Check the communication buffer of the other walkers
            CALL synchronize_multiple_walkers(multiple_walkers, hills_env, colvars, &
                                              n_colvar, meta_env%metadyn_section)
         END IF

         ! Print Hills file if requested (for multiple walkers this file includes
         ! the Hills coming from all the walkers).
         iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                   "PRINT%HILLS", extension=".metadynLog")
         IF (iw > 0) THEN
            DO i_hills = n_hills_start+1, hills_env%n_hills
               WRITE (iw, '(f12.1,30f13.5)') meta_env%time*femtoseconds, &
                  (hills_env%ss_history(ih, i_hills), ih=1, n_colvar), &
                  (hills_env%delta_s_history(ih, i_hills), ih=1, n_colvar), &
                  hills_env%ww_history(i_hills)
            END DO
         END IF
         CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                           "PRINT%HILLS")
      END IF

      ! Computes forces due to the hills: history dependent term
      ALLOCATE (ddp(meta_env%n_colvar))
      ALLOCATE (diff_ss(meta_env%n_colvar))
      ALLOCATE (numf(meta_env%n_colvar))
      ALLOCATE (denf(meta_env%n_colvar))

      hills_env%energy = 0.0_dp
      DO ih = 1, n_colvar
         colvars(ih)%ff_hills = 0.0_dp
      ENDDO
      DO ih = 1, hills_env%n_hills
         slow_growth = 1.0_dp
         IF (hills_env%slow_growth .AND. (ih == hills_env%n_hills)) THEN
            slow_growth = REAL(n_step-hills_env%old_hill_step, dp)/REAL(hills_env%nt_hills, dp)
         END IF
         dp2 = 0._dp
         cut_func = 1.0_dp
         DO i = 1, n_colvar
            diff_ss(i) = colvars(i)%ss0-hills_env%ss_history(i, ih)
            IF (colvars(i)%periodic) THEN
               ! The difference of a periodic COLVAR is always within [-pi,pi]
               diff_ss(i) = SIGN(1.0_dp, ASIN(SIN(diff_ss(i))))*ACOS(COS(diff_ss(i)))
            END IF
            IF (hills_env%delta_s_history(i, ih) == 0.0_dp) THEN
               ! trick: scale = 0 is interpreted as infinitely wide Gaussian hill
               ! instead of infinitely narrow. This way one can combine several
               ! one-dimensional bias potentials in a multi-dimensional metadyn
               ! simulation.
               ddp(i) = 0.0_dp
            ELSE
               ddp(i) = (diff_ss(i))/hills_env%delta_s_history(i, ih)
            END IF
            dp2 = dp2+ddp(i)**2
            IF (hills_env%tail_cutoff > 0.0_dp) THEN
               frac = ABS(ddp(i))/hills_env%tail_cutoff
               numf(i) = frac**hills_env%p_exp
               denf(i) = frac**hills_env%q_exp
               cut_func = cut_func*(1.0_dp-numf(i))/(1.0_dp-denf(i))
            ENDIF
         ENDDO
         ! ff_hills contains the "force" due to the hills
         dfunc = hills_env%ww_history(ih)*EXP(-0.5_dp*dp2)*slow_growth
         IF (meta_env%well_tempered) dfunc = dfunc/(1.0_dp+meta_env%wttemperature*hills_env%invdt_history(ih))
         hills_env%energy = hills_env%energy+dfunc*cut_func
         DO i = 1, n_colvar
            IF (hills_env%delta_s_history(i, ih) /= 0.0_dp) THEN
               ! only apply a force when the Gaussian hill has a finite width in
               ! this direction
               colvars(i)%ff_hills = colvars(i)%ff_hills+ &
                                     ddp(i)/hills_env%delta_s_history(i, ih)*dfunc*cut_func
               IF (hills_env%tail_cutoff > 0.0_dp .AND. ABS(diff_ss(i)) > 10.E-5_dp) THEN
                  colvars(i)%ff_hills = colvars(i)%ff_hills+ &
                                        (hills_env%p_exp*numf(i)/(1.0_dp-numf(i))-hills_env%q_exp*denf(i)/(1.0_dp-denf(i)))* &
                                        dfunc*cut_func/ABS(diff_ss(i))
               END IF
            END IF
         ENDDO
      ENDDO
      DEALLOCATE (ddp)
      DEALLOCATE (diff_ss)
      DEALLOCATE (numf)
      DEALLOCATE (denf)

      CALL cp_rm_iter_level(logger%iter_info, "METADYNAMICS")

      CALL timestop(handle)

   END SUBROUTINE hills

END MODULE metadynamics
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
MODULE al_system_mapping

   USE al_system_types,                 ONLY: al_system_type,&
                                              al_thermo_create
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: map_info_type
   USE input_constants,                 ONLY: &
        do_thermo_communication, do_thermo_no_communication, isokin_ensemble, langevin_ensemble, &
        npe_f_ensemble, npe_i_ensemble, nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, &
        npt_f_ensemble, npt_i_ensemble, nve_ensemble, nvt_ensemble, reftraj_ensemble
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_mapping,              ONLY: thermostat_mapping_region
   USE thermostat_types,                ONLY: thermostat_info_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'al_system_mapping'

   PUBLIC :: al_to_particle_mapping

CONTAINS

! **************************************************************************************************
!> \brief Creates the thermostatting maps
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param al ...
!> \param para_env ...
!> \param gci ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE al_to_particle_mapping(thermostat_info, simpar, local_molecules, &
                                     molecule_set, molecule_kind_set, al, para_env, gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(al_system_type), POINTER                      :: al
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(LEN=*), PARAMETER :: routineN = 'al_to_particle_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imap, j, natoms_local, &
                                                            sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      REAL(KIND=dp)                                      :: fac
      TYPE(map_info_type), POINTER                       :: map_info

      NULLIFY (massive_atom_list, deg_of_freedom)
      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/al_system_mapping.F",78,'Unknown ensemble!')
      CASE (nve_ensemble, isokin_ensemble, npe_f_ensemble, npe_i_ensemble, nph_uniaxial_ensemble, &
            nph_uniaxial_damped_ensemble, reftraj_ensemble, langevin_ensemble)
         CALL cp__b("motion/thermostat/al_system_mapping.F",81,'Never reach this point!')
      CASE (nvt_ensemble, npt_i_ensemble, npt_f_ensemble)

         CALL setup_al_thermostat(al, thermostat_info, deg_of_freedom, &
                                  massive_atom_list, molecule_kind_set, local_molecules, molecule_set, &
                                  para_env, natoms_local, simpar, sum_of_thermostats, gci)

         ! Sum up the number of degrees of freedom on each thermostat.
         ! first: initialize the target
         map_info => al%map_info
         map_info%s_kin = 0.0_dp
         DO i = 1, 3
            DO j = 1, natoms_local
               map_info%p_kin(i, j)%point = map_info%p_kin(i, j)%point+1
            END DO
         END DO

         ! If thermostats are replicated but molecules distributed, we have to
         ! sum s_kin over all processors
         IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, para_env%group)

         ! We know the total number of system thermostats.
         IF ((sum_of_thermostats == 1) .AND. (map_info%dis_type /= do_thermo_no_communication)) THEN
            fac = map_info%s_kin(1)-deg_of_freedom(1)-simpar%nfree_rot_transl
            IF (fac == 0.0_dp) THEN
               CALL cp__b("motion/thermostat/al_system_mapping.F",106,'Zero degrees of freedom. Nothing to thermalize!')
            END IF
            al%nvt(1)%nkt = simpar%temp_ext*fac
            al%nvt(1)%degrees_of_freedom = FLOOR(fac)
         ELSE
            DO i = 1, al%loc_num_al
               imap = map_info%map_index(i)
               fac = (map_info%s_kin(imap)-deg_of_freedom(i))
               al%nvt(i)%nkt = simpar%temp_ext*fac
               al%nvt(i)%degrees_of_freedom = FLOOR(fac)
            END DO
         END IF

         DEALLOCATE (deg_of_freedom)
         DEALLOCATE (massive_atom_list)
      END SELECT

   END SUBROUTINE al_to_particle_mapping

! **************************************************************************************************
!> \brief Main general setup for AD_LANGEVIN thermostats
!> \param al ...
!> \param thermostat_info ...
!> \param deg_of_freedom ...
!> \param massive_atom_list ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param para_env ...
!> \param natoms_local ...
!> \param simpar ...
!> \param sum_of_thermostats ...
!> \param gci ...
!> \param shell ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
   SUBROUTINE setup_al_thermostat(al, thermostat_info, deg_of_freedom, &
                                  massive_atom_list, molecule_kind_set, local_molecules, molecule_set, &
                                  para_env, natoms_local, simpar, sum_of_thermostats, gci, shell)

      TYPE(al_system_type), POINTER                      :: al
      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(OUT)                               :: natoms_local
      TYPE(simpar_type), POINTER                         :: simpar
      INTEGER, INTENT(OUT)                               :: sum_of_thermostats
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_al_thermostat', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nkind, number, region
      LOGICAL                                            :: do_shell
      TYPE(map_info_type), POINTER                       :: map_info

      do_shell = .FALSE.
      IF (PRESENT(shell)) do_shell = shell
      map_info => al%map_info

      nkind = SIZE(molecule_kind_set)
      sum_of_thermostats = thermostat_info%sum_of_thermostats
      map_info%dis_type = thermostat_info%dis_type
      number = thermostat_info%number_of_thermostats
      region = al%region

      CALL thermostat_mapping_region(map_info, deg_of_freedom, massive_atom_list, &
                                     molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, &
                                     simpar, number, region, gci, do_shell, thermostat_info%map_loc_thermo_gen, &
                                     sum_of_thermostats)

      ! This is the local number of available thermostats
      al%loc_num_al = number
      al%glob_num_al = sum_of_thermostats
      CALL al_thermo_create(al)

   END SUBROUTINE setup_al_thermostat

END MODULE al_system_mapping
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
MODULE al_system_init

   USE al_system_mapping,               ONLY: al_to_particle_mapping
   USE al_system_types,                 ONLY: al_system_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_types,                ONLY: thermostat_info_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: initialize_al_part

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'al_system_init'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule ...
!> \param molecule_kind_set ...
!> \param para_env ...
!> \param al ...
!> \param al_section ...
!> \param gci ...
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
   SUBROUTINE initialize_al_part(thermostat_info, simpar, local_molecules, &
                                 molecule, molecule_kind_set, para_env, al, al_section, &
                                 gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(al_system_type), POINTER                      :: al
      TYPE(section_vals_type), POINTER                   :: al_section
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_al_part', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: restart

      restart = .FALSE.
      CALL al_to_particle_mapping(thermostat_info, simpar, local_molecules, &
                                  molecule, molecule_kind_set, al, para_env, gci)

      CALL restart_al(al, al_section, restart)

      IF (.NOT. restart) THEN
         CALL init_al_variables(al)
      ENDIF

   END SUBROUTINE initialize_al_part

! **************************************************************************************************
!> \brief ...
!> \param al ...
! **************************************************************************************************
   SUBROUTINE init_al_variables(al)
      TYPE(al_system_type), POINTER                      :: al

      CHARACTER(len=*), PARAMETER :: routineN = 'init_al_variables', &
         routineP = moduleN//':'//routineN

      al%nvt(:)%mass = al%nvt(:)%nkt*al%tau_nh**2
      al%nvt(:)%chi = 0.0_dp
   END SUBROUTINE init_al_variables

! **************************************************************************************************
!> \brief ...
!> \param al ...
!> \param al_section ...
!> \param restart ...
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
   SUBROUTINE restart_al(al, al_section, restart)
      TYPE(al_system_type), POINTER                      :: al
      TYPE(section_vals_type), POINTER                   :: al_section
      LOGICAL, INTENT(inout)                             :: restart

      CHARACTER(len=*), PARAMETER :: routineN = 'restart_al', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, my_index, n_rep
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: work_section

      restart = .FALSE.

      ! Possibly restart the initial thermostat DOF value
      work_section => section_vals_get_subs_vals(section_vals=al_section, &
                                                 subsection_name="CHI")
      CALL section_vals_get(work_section, explicit=explicit)
      restart = explicit
      IF (explicit) THEN
         CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                   n_rep_val=n_rep)
         IF (n_rep == al%glob_num_al) THEN
            DO i = 1, al%loc_num_al
               my_index = al%map_info%index(i)
               CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                         i_rep_val=my_index, r_val=al%nvt(i)%chi)
            END DO
         ELSE
            CALL cp_abort(cp__l("motion/thermostat/al_system_init.F",129), &
                          'Number pf restartable stream not equal to the number of'// &
                          ' total thermostats!')
         END IF
      END IF

      ! Possibly restart the initial thermostat mass
      work_section => section_vals_get_subs_vals(section_vals=al_section, &
                                                 subsection_name="MASS")
      CALL section_vals_get(work_section, explicit=explicit)
      IF (restart .NEQV. explicit) &
         CALL cp_abort(cp__l("motion/thermostat/al_system_init.F",140), &
                       "You need to define both CHI and MASS sections (or none) in the AD_LANGEVIN section")
      restart = restart .AND. explicit
      IF (explicit) THEN
         CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                   n_rep_val=n_rep)
         IF (n_rep == al%glob_num_al) THEN
            DO i = 1, al%loc_num_al
               my_index = al%map_info%index(i)
               CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                         i_rep_val=my_index, r_val=al%nvt(i)%mass)
            END DO
         ELSE
            CALL cp_abort(cp__l("motion/thermostat/al_system_init.F",153), &
                          'Number pf restartable stream not equal to the number of'// &
                          ' total thermostats!')
         END IF
      END IF

   END SUBROUTINE restart_al

END MODULE al_system_init
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
MODULE al_system_dynamics

   USE al_system_types,                 ONLY: al_system_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE constraint_fxd,                  ONLY: fix_atom_control
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: map_info_type
   USE force_env_types,                 ONLY: force_env_type
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              molecule_type
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE particle_types,                  ONLY: particle_type
   USE thermostat_utils,                ONLY: ke_region_particles,&
                                              vel_rescale_particles
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   LOGICAL, PARAMETER :: debug_this_module = .FALSE.
   PUBLIC :: al_particles

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'al_system_dynamics'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param al ...
!> \param force_env ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param local_particles ...
!> \param group ...
!> \param vel ...
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
   SUBROUTINE al_particles(al, force_env, molecule_kind_set, molecule_set, &
                           particle_set, local_molecules, local_particles, group, vel)

      TYPE(al_system_type), POINTER                      :: al
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      INTEGER, INTENT(IN)                                :: group
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'al_particles', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: my_shell_adiabatic
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)
      my_shell_adiabatic = .FALSE.
      map_info => al%map_info

      IF (debug_this_module) &
         CALL dump_vel(molecule_kind_set, molecule_set, local_molecules, particle_set, vel, "INIT")

      IF (al%tau_nh <= 0.0_dp) THEN
         CALL al_OU_step(0.5_dp, al, force_env, map_info, molecule_kind_set, molecule_set, &
                         particle_set, local_molecules, local_particles, vel)
         IF (debug_this_module) &
            CALL dump_vel(molecule_kind_set, molecule_set, local_molecules, particle_set, vel, "post OU")
      ELSE
         ! quarter step of Langevin using Ornstein-Uhlenbeck
         CALL al_OU_step(0.25_dp, al, force_env, map_info, molecule_kind_set, molecule_set, &
                         particle_set, local_molecules, local_particles, vel)
         IF (debug_this_module) &
            CALL dump_vel(molecule_kind_set, molecule_set, local_molecules, particle_set, vel, "post 1st OU")

         ! Compute the kinetic energy for the region to thermostat for the (T dependent chi step)
         CALL ke_region_particles(map_info, particle_set, molecule_kind_set, &
                                  local_molecules, molecule_set, group, vel=vel)
         ! quarter step of chi, and set vel drag factors for a half step
         CALL al_NH_quarter_step(al, map_info, set_half_step_vel_factors=.TRUE.)

         ! Now scale the particle velocities for a NH half step
         CALL vel_rescale_particles(map_info, molecule_kind_set, molecule_set, particle_set, &
                                    local_molecules, my_shell_adiabatic, vel=vel)
         ! Recompute the kinetic energy for the region to thermostat (for the T dependent chi step)
         CALL ke_region_particles(map_info, particle_set, molecule_kind_set, &
                                  local_molecules, molecule_set, group, vel=vel)
         IF (debug_this_module) &
            CALL dump_vel(molecule_kind_set, molecule_set, local_molecules, particle_set, vel, "post rescale_vel")

         ! quarter step of chi
         CALL al_NH_quarter_step(al, map_info, set_half_step_vel_factors=.FALSE.)

         ! quarter step of Langevin using Ornstein-Uhlenbeck
         CALL al_OU_step(0.25_dp, al, force_env, map_info, molecule_kind_set, molecule_set, &
                         particle_set, local_molecules, local_particles, vel)
         IF (debug_this_module) &
            CALL dump_vel(molecule_kind_set, molecule_set, local_molecules, particle_set, vel, "post 2nd OU")
      ENDIF

      ! Recompute the final kinetic energy for the region to thermostat
      CALL ke_region_particles(map_info, particle_set, molecule_kind_set, &
                               local_molecules, molecule_set, group, vel=vel)

      CALL timestop(handle)
   END SUBROUTINE al_particles

! **************************************************************************************************
!> \brief ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param local_molecules ...
!> \param particle_set ...
!> \param vel ...
!> \param label ...
! **************************************************************************************************
   SUBROUTINE dump_vel(molecule_kind_set, molecule_set, local_molecules, particle_set, vel, label)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(dp), OPTIONAL                                 :: vel(:, :)
      CHARACTER(len=*)                                   :: label

      INTEGER                                            :: first_atom, ikind, imol, imol_local, &
                                                            ipart, last_atom, nmol_local
      TYPE(molecule_type), POINTER                       :: molecule

      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_local = local_molecules%n_el(ikind)
         DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            DO ipart = first_atom, last_atom
               IF (PRESENT(vel)) THEN
                  WRITE (unit=*, fmt='("VEL ",A20," IPART ",I6," V ",3F20.10)') TRIM(label), ipart, vel(:, ipart)
               ELSE
                  WRITE (unit=*, fmt='("VEL ",A20," IPART ",I6," V ",3F20.10)') TRIM(label), ipart, particle_set(ipart)%v(:)
               ENDIF
            END DO
         END DO
      END DO
   END SUBROUTINE dump_vel

! **************************************************************************************************
!> \brief ...
!> \param step ...
!> \param al ...
!> \param force_env ...
!> \param map_info ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param local_particles ...
!> \param vel ...
! **************************************************************************************************
   SUBROUTINE al_OU_step(step, al, force_env, map_info, molecule_kind_set, molecule_set, &
                         particle_set, local_molecules, local_particles, vel)
      REAL(dp), INTENT(in)                               :: step
      TYPE(al_system_type), POINTER                      :: al
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(map_info_type), POINTER                       :: map_info
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'al_OU_step', routineP = moduleN//':'//routineN

      INTEGER :: first_atom, i, ii, ikind, imap, imol, imol_local, ipart, iparticle_kind, &
         iparticle_local, last_atom, nmol_local, nparticle, nparticle_kind, nparticle_local
      LOGICAL                                            :: check, present_vel
      REAL(KIND=dp)                                      :: mass
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: w(:, :)
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      present_vel = PRESENT(vel)

      ![NB] not a big deal, but could this be done once at init time?
      DO i = 1, al%loc_num_al
         imap = map_info%map_index(i)
         ! drag on velocities
         IF (al%tau_langevin > 0.0_dp) THEN
            map_info%v_scale(imap) = EXP(-step*al%dt/al%tau_langevin)
            map_info%s_kin(imap) = SQRT((al%nvt(i)%nkt/al%nvt(i)%degrees_of_freedom)*(1.0_dp-map_info%v_scale(imap)**2))
         ELSE
            map_info%v_scale(imap) = 1.0_dp
            map_info%s_kin(imap) = 0.0_dp
         ENDIF
         ! magnitude of random force, not including 1/sqrt(mass) part
      END DO

      nparticle = SIZE(particle_set)
      nparticle_kind = SIZE(local_particles%n_el)
      ALLOCATE (w(3, nparticle))
      w(:, :) = 0.0_dp
      check = (nparticle_kind <= SIZE(local_particles%n_el) .AND. nparticle_kind <= SIZE(local_particles%list))
      IF(.NOT.(check))CALL cp__a("motion/thermostat/al_system_dynamics.F",216)
      check = ASSOCIATED(local_particles%local_particle_set)
      IF(.NOT.(check))CALL cp__a("motion/thermostat/al_system_dynamics.F",218)
      DO iparticle_kind = 1, nparticle_kind
         nparticle_local = local_particles%n_el(iparticle_kind)
         check = (nparticle_local <= SIZE(local_particles%list(iparticle_kind)%array))
         IF(.NOT.(check))CALL cp__a("motion/thermostat/al_system_dynamics.F",222)
         DO iparticle_local = 1, nparticle_local
            ipart = local_particles%list(iparticle_kind)%array(iparticle_local)
            rng_stream => local_particles%local_particle_set(iparticle_kind)%rng(iparticle_local)%stream
            w(1, ipart) = next_random_number(rng_stream, variance=1.0_dp)
            w(2, ipart) = next_random_number(rng_stream, variance=1.0_dp)
            w(3, ipart) = next_random_number(rng_stream, variance=1.0_dp)
         END DO
      END DO

      CALL fix_atom_control(force_env, w)

      ii = 0
      DO ikind = 1, SIZE(molecule_kind_set)
         nmol_local = local_molecules%n_el(ikind)
         DO imol_local = 1, nmol_local
            imol = local_molecules%list(ikind)%array(imol_local)
            molecule => molecule_set(imol)
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            DO ipart = first_atom, last_atom
               ii = ii+1
               atomic_kind => particle_set(ipart)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, mass=mass)
               IF (present_vel) THEN
                  vel(:, ipart) = vel(:, ipart)*map_info%v_scale(ii)+ &
                                  map_info%s_kin(ii)/SQRT(mass)*w(:, ipart)
               ELSE
                  particle_set(ipart)%v(:) = particle_set(ipart)%v(:)*map_info%v_scale(ii)+ &
                                             map_info%s_kin(ii)/SQRT(mass)*w(:, ipart)
               ENDIF
            END DO
         END DO
      END DO

      DEALLOCATE (w)

   END SUBROUTINE al_OU_step

! **************************************************************************************************
!> \brief ...
!> \param al ...
!> \param map_info ...
!> \param set_half_step_vel_factors ...
!> \author Noam Bernstein [noamb] 02.2012
! **************************************************************************************************
   SUBROUTINE al_NH_quarter_step(al, map_info, set_half_step_vel_factors)
      TYPE(al_system_type), POINTER                      :: al
      TYPE(map_info_type), POINTER                       :: map_info
      LOGICAL, INTENT(in)                                :: set_half_step_vel_factors

      CHARACTER(len=*), PARAMETER :: routineN = 'al_NH_quarter_step', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imap
      REAL(KIND=dp)                                      :: decay, delta_K

![NB] how to deal with dt_fact?

      DO i = 1, al%loc_num_al
         IF (al%nvt(i)%mass > 0.0_dp) THEN
            imap = map_info%map_index(i)
            delta_K = 0.5_dp*(map_info%s_kin(imap)-al%nvt(i)%nkt)
            al%nvt(i)%chi = al%nvt(i)%chi+0.5_dp*al%dt*delta_K/al%nvt(i)%mass
            IF (set_half_step_vel_factors) THEN
               decay = EXP(-0.5_dp*al%dt*al%nvt(i)%chi)
               map_info%v_scale(imap) = decay
            ENDIF
         ELSE
            al%nvt(i)%chi = 0.0_dp
            IF (set_half_step_vel_factors) THEN
               map_info%v_scale(imap) = 1.0_dp
            ENDIF
         ENDIF
      END DO

   END SUBROUTINE al_NH_quarter_step

END MODULE al_system_dynamics
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      15.10.2007 Giovanni Bussi - Implementation validated.
!> \author Teodoro Laino - 09.2007 - University of Zurich [tlaino]
! **************************************************************************************************
MODULE csvr_system_utils

   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PARAMETER                   :: debug_this_module = .FALSE.
   PUBLIC                               :: rescaling_factor
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'csvr_system_utils'

CONTAINS

! **************************************************************************************************
!> \brief Stochastic velocity rescale, as described in
!>      Bussi, Donadio and Parrinello, J. Chem. Phys. 126, 014101 (2007)
!>
!>      This subroutine implements Eq.(A7) and returns the new value for the kinetic energy,
!>      which can be used to rescale the velocities.
!>      The procedure can be applied to all atoms or to smaller groups.
!>      If it is applied to intersecting groups in sequence, the kinetic energy
!>      that is given as an input (kk) has to be up-to-date with respect to the previous
!>      rescalings.
!>
!>      When applied to the entire system, and when performing standard molecular dynamics
!>      (fixed c.o.m. (center of mass))
!>      the degrees of freedom of the c.o.m. have to be discarded in the calculation of ndeg,
!>      and the c.o.m. momentum HAS TO BE SET TO ZERO.
!>      When applied to subgroups, one can chose to:
!>      (a) calculate the subgroup kinetic energy in the usual reference frame, and count
!>          the c.o.m. in ndeg
!>      (b) calculate the subgroup kinetic energy with respect to its c.o.m. motion, discard
!>          the c.o.m. in ndeg and apply the rescale factor with respect to the subgroup c.o.m.
!>          velocity.
!>      They should be almost equivalent.
!>      If the subgroups are expected to move one respect to the other, the choice (b)
!>      should be better.
!>
!>      If a null relaxation time is required (taut=0.0), the procedure reduces to an istantaneous
!>      randomization of the kinetic energy, as described in paragraph IIA.
!>
!>      HOW TO CALCULATE THE EFFECTIVE-ENERGY DRIFT
!>      The effective-energy (htilde) drift can be used to check the integrator against
!>      discretization errors.
!>      The easiest recipe is:
!>      htilde = h + conint
!>      where h is the total energy (kinetic + potential)
!>      and conint is a quantity accumulated along the trajectory as minus the sum of all
!>      the increments of kinetic energy due to the thermostat.
!>
!>      Variables:
!>       kk    ! present value of the kinetic energy of the atoms to be thermalized (in arbitrary units)
!>       sigma ! target average value of the kinetic energy (ndeg k_b T/2)  (in the same units as kk)
!>       ndeg  ! number of degrees of freedom of the atoms to be thermalized
!>       taut  ! relaxation time of the thermostat, in units of 'how often this routine is called'
!> \param kk ...
!> \param sigma ...
!> \param ndeg ...
!> \param taut ...
!> \param rng_stream ...
!> \retval my_res ...
!> \date 09.2007
!> \author Giovanni Bussi - ETH Zurich, Lugano 10.2007
! **************************************************************************************************
   FUNCTION rescaling_factor(kk, sigma, ndeg, taut, rng_stream) RESULT(my_res)
      REAL(KIND=dp), INTENT(IN)                          :: kk, sigma
      INTEGER, INTENT(IN)                                :: ndeg
      REAL(KIND=dp), INTENT(IN)                          :: taut
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      REAL(KIND=dp)                                      :: my_res

      CHARACTER(len=*), PARAMETER :: routineN = 'rescaling_factor', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: factor, resample, reverse, rr

      my_res = 0.0_dp
      IF (kk > 0.0_dp) THEN
         IF (taut > 0.1_dp) THEN
            factor = EXP(-1.0_dp/taut)
         ELSE
            factor = 0.0_dp
         END IF
         rr = next_random_number(rng_stream)
         reverse = 1.0_dp
         ! reverse of momentum is implemented to have the correct limit to Langevin dynamics for ndeg=1
         IF (rr < -SQRT(ndeg*kk*factor/(sigma*(1.0_dp-factor)))) reverse = -1.0_dp
         ! for ndeg/=1, the reverse of momentum is not necessary. in principles, it should be there.
         ! in practice, it is better to skip it to avoid unnecessary slowing down of the dynamics in the small taut regime
         ! anyway, this should not affect the final ensemble
         IF (ndeg /= 1) reverse = 1.0_dp
         resample = kk+(1.0_dp-factor)*(sigma*(sumnoises(ndeg-1, rng_stream)+rr**2)/REAL(ndeg, KIND=dp)-kk) &
                    +2.0_dp*rr*SQRT(kk*sigma/ndeg*(1.0_dp-factor)*factor)

         resample = MAX(0.0_dp, resample)
         my_res = reverse*SQRT(resample/kk)
      END IF
   END FUNCTION rescaling_factor

! **************************************************************************************************
!> \brief returns the sum of n independent gaussian noises squared
!>      (i.e. equivalent to summing the square of the return values of nn calls to gasdev)
!> \param nn ...
!> \param rng_stream ...
!> \retval sum_gauss ...
!> \date 09.2007
!> \author Teo - University of Zurich
! **************************************************************************************************
   FUNCTION sumnoises(nn, rng_stream) RESULT(sum_gauss)
      INTEGER, INTENT(IN)                                :: nn
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      REAL(KIND=dp)                                      :: sum_gauss

      CHARACTER(len=*), PARAMETER :: routineN = 'sumnoises', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      sum_gauss = 0.0_dp
      DO i = 1, nn
         sum_gauss = sum_gauss+next_random_number(rng_stream)**2
      END DO

   END FUNCTION sumnoises

END MODULE csvr_system_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
MODULE csvr_system_dynamics

   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE csvr_system_types,               ONLY: csvr_system_type
   USE csvr_system_utils,               ONLY: rescaling_factor
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: map_info_type,&
                                              npt_info_type
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE parallel_rng_types,              ONLY: rng_stream_type
   USE particle_types,                  ONLY: particle_type
   USE thermostat_utils,                ONLY: ke_region_baro,&
                                              ke_region_particles,&
                                              ke_region_shells,&
                                              vel_rescale_baro,&
                                              vel_rescale_particles,&
                                              vel_rescale_shells
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   LOGICAL, PARAMETER :: debug_this_module = .FALSE.
   PUBLIC :: csvr_particles, &
             csvr_barostat, &
             csvr_shells

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'csvr_system_dynamics'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param csvr ...
!> \param npt ...
!> \param group ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_barostat(csvr, npt, group)

      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(npt_info_type), DIMENSION(:, :), &
         INTENT(INOUT)                                   :: npt
      INTEGER, INTENT(IN)                                :: group

      CHARACTER(len=*), PARAMETER :: routineN = 'csvr_barostat', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)
      map_info => csvr%map_info

      ! Compute the kinetic energy of the barostat
      CALL ke_region_baro(map_info, npt, group)

      ! Apply the Canonical Sampling through Velocity Rescaling
      CALL do_csvr(csvr, map_info)

      ! Now scale the particle velocities
      CALL vel_rescale_baro(map_info, npt)

      ! Re-Compute the kinetic energy of the barostat
      CALL ke_region_baro(map_info, npt, group)

      ! Compute thermostat energy
      CALL do_csvr_eval_energy(csvr, map_info)

      CALL timestop(handle)
   END SUBROUTINE csvr_barostat

! **************************************************************************************************
!> \brief ...
!> \param csvr ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param group ...
!> \param shell_adiabatic ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_particles(csvr, molecule_kind_set, molecule_set, &
                             particle_set, local_molecules, group, shell_adiabatic, &
                             shell_particle_set, core_particle_set, vel, shell_vel, core_vel)

      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      INTEGER, INTENT(IN)                                :: group
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell_adiabatic
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :), shell_vel(:, :), &
                                                            core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'csvr_particles', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: my_shell_adiabatic
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)
      my_shell_adiabatic = .FALSE.
      IF (PRESENT(shell_adiabatic)) my_shell_adiabatic = shell_adiabatic
      map_info => csvr%map_info

      ! Compute the kinetic energy for the region to thermostat
      CALL ke_region_particles(map_info, particle_set, molecule_kind_set, &
                               local_molecules, molecule_set, group, vel)

      ! Apply the Canonical Sampling through Velocity Rescaling
      CALL do_csvr(csvr, map_info)

      ! Now scale the particle velocities
      CALL vel_rescale_particles(map_info, molecule_kind_set, molecule_set, particle_set, &
                                 local_molecules, my_shell_adiabatic, shell_particle_set, core_particle_set, &
                                 vel, shell_vel, core_vel)

      ! Re-Compute the kinetic energy for the region to thermostat
      CALL ke_region_particles(map_info, particle_set, molecule_kind_set, &
                               local_molecules, molecule_set, group, vel)

      ! Compute thermostat energy
      CALL do_csvr_eval_energy(csvr, map_info)

      CALL timestop(handle)
   END SUBROUTINE csvr_particles

! **************************************************************************************************
!> \brief ...
!> \param csvr ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param group ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_shells(csvr, atomic_kind_set, particle_set, local_particles, &
                          group, shell_particle_set, core_particle_set, vel, shell_vel, core_vel)

      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      INTEGER, INTENT(IN)                                :: group
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :), shell_vel(:, :), &
                                                            core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'csvr_shells', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(map_info_type), POINTER                       :: map_info

      CALL timeset(routineN, handle)
      map_info => csvr%map_info

      ! Compute the kinetic energy of the region to thermostat
      CALL ke_region_shells(map_info, particle_set, atomic_kind_set, local_particles, &
                            group, core_particle_set, shell_particle_set, core_vel, shell_vel)

      ! Apply the Canonical Sampling through Velocity Rescaling
      CALL do_csvr(csvr, map_info)

      ! Now scale the particle velocities
      CALL vel_rescale_shells(map_info, atomic_kind_set, particle_set, local_particles, &
                              shell_particle_set, core_particle_set, shell_vel, core_vel, vel)

      ! Re-Compute the kinetic energy of the region to thermostat
      CALL ke_region_shells(map_info, particle_set, atomic_kind_set, local_particles, &
                            group, core_particle_set, shell_particle_set, core_vel, shell_vel)

      ! Compute thermostat energy
      CALL do_csvr_eval_energy(csvr, map_info)

      CALL timestop(handle)
   END SUBROUTINE csvr_shells

! **************************************************************************************************
!> \brief ...
!> \param csvr ...
!> \param map_info ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE do_csvr(csvr, map_info)
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(map_info_type), POINTER                       :: map_info

      CHARACTER(len=*), PARAMETER :: routineN = 'do_csvr', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imap, ndeg
      REAL(KIND=dp)                                      :: kin_energy, kin_target, taut
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      DO i = 1, csvr%loc_num_csvr
         imap = map_info%map_index(i)
         kin_energy = map_info%s_kin(imap)
         csvr%nvt(i)%region_kin_energy = kin_energy
         kin_energy = kin_energy*0.5_dp
         kin_target = csvr%nvt(i)%nkt*0.5_dp
         ndeg = csvr%nvt(i)%degrees_of_freedom
         taut = csvr%tau_csvr/csvr%dt_fact
         rng_stream => csvr%nvt(i)%gaussian_rng_stream
         map_info%v_scale(imap) = rescaling_factor(kin_energy, kin_target, ndeg, taut, &
                                                   rng_stream)
      END DO

   END SUBROUTINE do_csvr

! **************************************************************************************************
!> \brief ...
!> \param csvr ...
!> \param map_info ...
!> \author 10.2007 [tlaino] - Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE do_csvr_eval_energy(csvr, map_info)
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(map_info_type), POINTER                       :: map_info

      CHARACTER(len=*), PARAMETER :: routineN = 'do_csvr_eval_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imap
      REAL(KIND=dp)                                      :: kin_energy_ar, kin_energy_br

      DO i = 1, csvr%loc_num_csvr
         imap = map_info%map_index(i)
         kin_energy_br = csvr%nvt(i)%region_kin_energy
         kin_energy_ar = map_info%s_kin(imap)
         csvr%nvt(i)%thermostat_energy = csvr%nvt(i)%thermostat_energy+ &
                                         0.5_dp*(kin_energy_br-kin_energy_ar)
      END DO

   END SUBROUTINE do_csvr_eval_energy

END MODULE csvr_system_dynamics
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
MODULE csvr_system_mapping

   USE cp_para_types,                   ONLY: cp_para_env_type
   USE csvr_system_types,               ONLY: csvr_system_type,&
                                              csvr_thermo_create
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_types,           ONLY: debug_isotropic_limit,&
                                              map_info_type
   USE input_constants,                 ONLY: &
        do_thermo_communication, do_thermo_no_communication, do_thermo_only_master, &
        isokin_ensemble, langevin_ensemble, npe_f_ensemble, npe_i_ensemble, &
        nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, npt_f_ensemble, npt_i_ensemble, &
        nve_ensemble, nvt_ensemble, reftraj_ensemble
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_mapping,              ONLY: init_baro_map_info,&
                                              thermostat_mapping_region
   USE thermostat_types,                ONLY: thermostat_info_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'csvr_system_mapping'

   PUBLIC :: csvr_to_particle_mapping, csvr_to_barostat_mapping, &
             csvr_to_shell_mapping

CONTAINS

! **************************************************************************************************
!> \brief Creates the thermostatting for the barostat
!> \param simpar ...
!> \param csvr ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_to_barostat_mapping(simpar, csvr)
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(csvr_system_type), POINTER                    :: csvr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csvr_to_barostat_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ndeg
      TYPE(map_info_type), POINTER                       :: map_info

      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/csvr_system_mapping.F",64,'Never reach this point!')
      CASE (npt_i_ensemble, npt_f_ensemble)
         map_info => csvr%map_info
         map_info%dis_type = do_thermo_only_master

         ! Counting the total number of thermostats ( 1 for both NPT_I and NPT_F )
         csvr%loc_num_csvr = 1
         csvr%glob_num_csvr = 1
         IF (simpar%ensemble == npt_f_ensemble) THEN
            ndeg = 9
         ELSE
            ndeg = 1
         ENDIF

         CALL init_baro_map_info(map_info, ndeg, csvr%loc_num_csvr)
         CALL csvr_thermo_create(csvr)

         ! Now that we know how many there are stick this into csvr%nkt
         ! (number of degrees of freedom times k_B T )
         DO i = 1, csvr%loc_num_csvr
            csvr%nvt(i)%nkt = simpar%temp_baro_ext*ndeg
            csvr%nvt(i)%degrees_of_freedom = ndeg
            IF (debug_isotropic_limit) THEN
               csvr%nvt(i)%nkt = simpar%temp_baro_ext
               csvr%nvt(i)%degrees_of_freedom = 1
            END IF
         END DO
      END SELECT

   END SUBROUTINE csvr_to_barostat_mapping

! **************************************************************************************************
!> \brief Creates the thermostatting maps
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param csvr ...
!> \param para_env ...
!> \param gci ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE csvr_to_particle_mapping(thermostat_info, simpar, local_molecules, &
                                       molecule_set, molecule_kind_set, csvr, para_env, gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csvr_to_particle_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imap, j, natoms_local, &
                                                            sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      REAL(KIND=dp)                                      :: fac
      TYPE(map_info_type), POINTER                       :: map_info

      NULLIFY (massive_atom_list, deg_of_freedom)
      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/csvr_system_mapping.F",131,'Unknown ensemble!')
      CASE (nve_ensemble, isokin_ensemble, npe_f_ensemble, npe_i_ensemble, nph_uniaxial_ensemble, &
            nph_uniaxial_damped_ensemble, reftraj_ensemble, langevin_ensemble)
         CALL cp__b("motion/thermostat/csvr_system_mapping.F",134,'Never reach this point!')
      CASE (nvt_ensemble, npt_i_ensemble, npt_f_ensemble)

         CALL setup_csvr_thermostat(csvr, thermostat_info, deg_of_freedom, &
                                    massive_atom_list, molecule_kind_set, local_molecules, molecule_set, &
                                    para_env, natoms_local, simpar, sum_of_thermostats, gci)

         ! Sum up the number of degrees of freedom on each thermostat.
         ! first: initialize the target
         map_info => csvr%map_info
         map_info%s_kin = 0.0_dp
         DO i = 1, 3
            DO j = 1, natoms_local
               map_info%p_kin(i, j)%point = map_info%p_kin(i, j)%point+1
            END DO
         END DO

         ! If thermostats are replicated but molecules distributed, we have to
         ! sum s_kin over all processors
         IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, para_env%group)

         ! We know the total number of system thermostats.
         IF ((sum_of_thermostats == 1) .AND. (map_info%dis_type /= do_thermo_no_communication)) THEN
            fac = map_info%s_kin(1)-deg_of_freedom(1)-simpar%nfree_rot_transl
            IF (fac == 0.0_dp) THEN
               CALL cp__b("motion/thermostat/csvr_system_mapping.F",159,'Zero degrees of freedom. Nothing to thermalize!')
            END IF
            csvr%nvt(1)%nkt = simpar%temp_ext*fac
            csvr%nvt(1)%degrees_of_freedom = FLOOR(fac)
         ELSE
            DO i = 1, csvr%loc_num_csvr
               imap = map_info%map_index(i)
               fac = (map_info%s_kin(imap)-deg_of_freedom(i))
               csvr%nvt(i)%nkt = simpar%temp_ext*fac
               csvr%nvt(i)%degrees_of_freedom = FLOOR(fac)
            END DO
         END IF

         DEALLOCATE (deg_of_freedom)
         DEALLOCATE (massive_atom_list)
      END SELECT

   END SUBROUTINE csvr_to_particle_mapping

! **************************************************************************************************
!> \brief Main general setup for CSVR thermostats
!> \param csvr ...
!> \param thermostat_info ...
!> \param deg_of_freedom ...
!> \param massive_atom_list ...
!> \param molecule_kind_set ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param para_env ...
!> \param natoms_local ...
!> \param simpar ...
!> \param sum_of_thermostats ...
!> \param gci ...
!> \param shell ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
   SUBROUTINE setup_csvr_thermostat(csvr, thermostat_info, deg_of_freedom, &
                                    massive_atom_list, molecule_kind_set, local_molecules, molecule_set, &
                                    para_env, natoms_local, simpar, sum_of_thermostats, gci, shell)

      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_atom_list
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(OUT)                               :: natoms_local
      TYPE(simpar_type), POINTER                         :: simpar
      INTEGER, INTENT(OUT)                               :: sum_of_thermostats
      TYPE(global_constraint_type), POINTER              :: gci
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_csvr_thermostat', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nkind, number, region
      LOGICAL                                            :: do_shell
      TYPE(map_info_type), POINTER                       :: map_info

      do_shell = .FALSE.
      IF (PRESENT(shell)) do_shell = shell
      map_info => csvr%map_info

      nkind = SIZE(molecule_kind_set)
      sum_of_thermostats = thermostat_info%sum_of_thermostats
      map_info%dis_type = thermostat_info%dis_type
      number = thermostat_info%number_of_thermostats
      region = csvr%region

      CALL thermostat_mapping_region(map_info, deg_of_freedom, massive_atom_list, &
                                     molecule_kind_set, local_molecules, molecule_set, para_env, natoms_local, &
                                     simpar, number, region, gci, do_shell, thermostat_info%map_loc_thermo_gen, &
                                     sum_of_thermostats)

      ! This is the local number of available thermostats
      csvr%loc_num_csvr = number
      csvr%glob_num_csvr = sum_of_thermostats
      CALL csvr_thermo_create(csvr)

   END SUBROUTINE setup_csvr_thermostat

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule_set ...
!> \param molecule_kind_set ...
!> \param csvr ...
!> \param para_env ...
!> \param gci ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
   SUBROUTINE csvr_to_shell_mapping(thermostat_info, simpar, local_molecules, &
                                    molecule_set, molecule_kind_set, csvr, para_env, gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csvr_to_shell_mapping', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imap, j, nshell_local, &
                                                            sum_of_thermostats
      INTEGER, DIMENSION(:), POINTER                     :: deg_of_freedom, massive_shell_list
      TYPE(map_info_type), POINTER                       :: map_info

      NULLIFY (massive_shell_list, deg_of_freedom)

      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/thermostat/csvr_system_mapping.F",277,'Unknown ensemble!')
      CASE (isokin_ensemble, nph_uniaxial_ensemble, &
            nph_uniaxial_damped_ensemble, reftraj_ensemble, langevin_ensemble)
         CALL cp__b("motion/thermostat/csvr_system_mapping.F",280,'Never reach this point!')
      CASE (nve_ensemble, npe_f_ensemble, npe_i_ensemble, nvt_ensemble, npt_i_ensemble, npt_f_ensemble)

         CALL setup_csvr_thermostat(csvr, thermostat_info, deg_of_freedom, massive_shell_list, &
                                    molecule_kind_set, local_molecules, molecule_set, para_env, nshell_local, &
                                    simpar, sum_of_thermostats, gci, shell=.TRUE.)

         map_info => csvr%map_info
         ! Sum up the number of degrees of freedom on each thermostat.
         ! first: initialize the target
         map_info%s_kin = 0.0_dp
         DO j = 1, nshell_local
            DO i = 1, 3
               map_info%p_kin(i, j)%point = map_info%p_kin(i, j)%point+1
            END DO
         END DO

         ! If thermostats are replicated but molecules distributed, we have to
         ! sum s_kin over all processors
         IF (map_info%dis_type == do_thermo_communication) CALL mp_sum(map_info%s_kin, para_env%group)

         ! Now that we know how many there are stick this into csvr%nkt
         ! (number of degrees of freedom times k_B T )
         DO i = 1, csvr%loc_num_csvr
            imap = map_info%map_index(i)
            csvr%nvt(i)%nkt = simpar%temp_sh_ext*map_info%s_kin(imap)
            csvr%nvt(i)%degrees_of_freedom = FLOOR(map_info%s_kin(imap))
         END DO

         DEALLOCATE (deg_of_freedom)
         DEALLOCATE (massive_shell_list)
      END SELECT

   END SUBROUTINE csvr_to_shell_mapping

END MODULE csvr_system_mapping
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
MODULE csvr_system_init

   USE cp_para_types,                   ONLY: cp_para_env_type
   USE csvr_system_mapping,             ONLY: csvr_to_barostat_mapping,&
                                              csvr_to_particle_mapping,&
                                              csvr_to_shell_mapping
   USE csvr_system_types,               ONLY: csvr_system_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE parallel_rng_types,              ONLY: read_rng_stream,&
                                              rng_record_length
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_types,                ONLY: thermostat_info_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: initialize_csvr_part, initialize_csvr_baro, &
             initialize_csvr_shell

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'csvr_system_init'

CONTAINS

! **************************************************************************************************
!> \brief fire up the thermostats, if NPT
!> \param simpar ...
!> \param csvr ...
!> \param csvr_section ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE initialize_csvr_baro(simpar, csvr, csvr_section)

      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(section_vals_type), POINTER                   :: csvr_section

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_csvr_baro', &
         routineP = moduleN//':'//routineN

      CALL csvr_to_barostat_mapping(simpar, csvr)
      CALL restart_csvr(csvr, csvr_section)

   END SUBROUTINE initialize_csvr_baro

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule ...
!> \param molecule_kind_set ...
!> \param para_env ...
!> \param csvr ...
!> \param csvr_section ...
!> \param gci ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE initialize_csvr_part(thermostat_info, simpar, local_molecules, &
                                   molecule, molecule_kind_set, para_env, csvr, csvr_section, &
                                   gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(section_vals_type), POINTER                   :: csvr_section
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_csvr_part', &
         routineP = moduleN//':'//routineN

      CALL csvr_to_particle_mapping(thermostat_info, simpar, local_molecules, &
                                    molecule, molecule_kind_set, csvr, para_env, gci)
      CALL restart_csvr(csvr, csvr_section)

   END SUBROUTINE initialize_csvr_part

! **************************************************************************************************
!> \brief ...
!> \param thermostat_info ...
!> \param simpar ...
!> \param local_molecules ...
!> \param molecule ...
!> \param molecule_kind_set ...
!> \param para_env ...
!> \param csvr ...
!> \param csvr_section ...
!> \param gci ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE initialize_csvr_shell(thermostat_info, simpar, local_molecules, &
                                    molecule, molecule_kind_set, para_env, csvr, csvr_section, &
                                    gci)

      TYPE(thermostat_info_type), POINTER                :: thermostat_info
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(molecule_type), POINTER                       :: molecule(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(section_vals_type), POINTER                   :: csvr_section
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_csvr_shell', &
         routineP = moduleN//':'//routineN

      CALL csvr_to_shell_mapping(thermostat_info, simpar, local_molecules, &
                                 molecule, molecule_kind_set, csvr, para_env, gci)
      CALL restart_csvr(csvr, csvr_section)

   END SUBROUTINE initialize_csvr_shell

! **************************************************************************************************
!> \brief ...
!> \param csvr ...
!> \param csvr_section ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE restart_csvr(csvr, csvr_section)
      TYPE(csvr_system_type), POINTER                    :: csvr
      TYPE(section_vals_type), POINTER                   :: csvr_section

      CHARACTER(len=*), PARAMETER :: routineN = 'restart_csvr', routineP = moduleN//':'//routineN

      CHARACTER(LEN=rng_record_length)                   :: rng_record
      INTEGER                                            :: i, my_index, n_rep
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: work_section

! Possibly restart the initial thermostat energy

      work_section => section_vals_get_subs_vals(section_vals=csvr_section, &
                                                 subsection_name="THERMOSTAT_ENERGY")
      CALL section_vals_get(work_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                   n_rep_val=n_rep)
         IF (n_rep == csvr%glob_num_csvr) THEN
            DO i = 1, csvr%loc_num_csvr
               my_index = csvr%map_info%index(i)
               CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                         i_rep_val=my_index, r_val=csvr%nvt(i)%thermostat_energy)
            END DO
         ELSE
            CALL cp_abort(cp__l("motion/thermostat/csvr_system_init.F",166), &
                          'Number pf restartable stream not equal to the number of'// &
                          ' total thermostats!')
         END IF
      END IF

      ! Possibly restart the random number generators for the different thermostats
      work_section => section_vals_get_subs_vals(section_vals=csvr_section, &
                                                 subsection_name="RNG_INIT")

      CALL section_vals_get(work_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                   n_rep_val=n_rep)
         IF (n_rep == csvr%glob_num_csvr) THEN
            DO i = 1, csvr%loc_num_csvr
               my_index = csvr%map_info%index(i)
               CALL section_vals_val_get(section_vals=work_section, keyword_name="_DEFAULT_KEYWORD_", &
                                         i_rep_val=my_index, c_val=rng_record)
               CALL read_rng_stream(rng_stream=csvr%nvt(i)%gaussian_rng_stream, &
                                    rng_record=rng_record)
            END DO
         ELSE
            CALL cp_abort(cp__l("motion/thermostat/csvr_system_init.F",189), &
                          'Number pf restartable stream not equal to the number of'// &
                          ' total thermostats!')
         END IF
      END IF
   END SUBROUTINE restart_csvr

END MODULE csvr_system_init
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Methods for Thermostats
!> \author teo [tlaino] - University of Zurich - 10.2007
! **************************************************************************************************
MODULE thermostat_methods
   USE al_system_dynamics,              ONLY: al_particles
   USE al_system_init,                  ONLY: initialize_al_part
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE bibliography,                    ONLY: VandeVondele2002,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE csvr_system_dynamics,            ONLY: csvr_barostat,&
                                              csvr_particles,&
                                              csvr_shells
   USE csvr_system_init,                ONLY: initialize_csvr_baro,&
                                              initialize_csvr_part,&
                                              initialize_csvr_shell
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE extended_system_dynamics,        ONLY: lnhc_barostat,&
                                              lnhc_particles,&
                                              lnhc_shells
   USE extended_system_init,            ONLY: initialize_nhc_baro,&
                                              initialize_nhc_fast,&
                                              initialize_nhc_part,&
                                              initialize_nhc_shell,&
                                              initialize_nhc_slow
   USE extended_system_types,           ONLY: npt_info_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE gle_system_dynamics,             ONLY: gle_particles,&
                                              initialize_gle_part
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: &
        do_region_global, do_thermo_al, do_thermo_csvr, do_thermo_gle, do_thermo_nose, &
        do_thermo_same_as_part, npe_f_ensemble, npe_i_ensemble, npt_f_ensemble, npt_i_ensemble, &
        nve_ensemble, nvt_adiabatic_ensemble, nvt_ensemble
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types,                      ONLY: qmmm_env_type
   USE simpar_types,                    ONLY: simpar_type
   USE thermostat_types,                ONLY: allocate_thermostats,&
                                              create_thermostat_type,&
                                              release_thermostat_info,&
                                              release_thermostat_type,&
                                              release_thermostats,&
                                              thermostat_type,&
                                              thermostats_type
   USE thermostat_utils,                ONLY: compute_degrees_of_freedom,&
                                              compute_nfree,&
                                              get_thermostat_energies,&
                                              setup_adiabatic_thermostat_info,&
                                              setup_thermostat_info
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: create_thermostats, &
             apply_thermostat_baro, &
             apply_thermostat_particles, &
             apply_thermostat_shells

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'thermostat_methods'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param thermostats ...
!> \param md_section ...
!> \param force_env ...
!> \param simpar ...
!> \param para_env ...
!> \param globenv ...
!> \param global_section ...
!> \par History
!>      10.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE create_thermostats(thermostats, md_section, force_env, simpar, &
                                 para_env, globenv, global_section)
      TYPE(thermostats_type), POINTER                    :: thermostats
      TYPE(section_vals_type), POINTER                   :: md_section
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: global_section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_thermostats', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: binary_restart_file_name
      INTEGER                                            :: n_rep, region, thermostat_type
      LOGICAL :: apply_general_thermo, apply_thermo_adiabatic, apply_thermo_baro, &
         apply_thermo_shell, explicit_adiabatic_fast, explicit_adiabatic_slow, explicit_baro, &
         explicit_barostat_section, explicit_part, explicit_shell, save_mem, shell_adiabatic, &
         shell_present
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(mol_new_list_type), POINTER                   :: molecules_new
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(qmmm_env_type), POINTER                       :: qmmm_env
      TYPE(section_vals_type), POINTER :: adiabatic_fast_section, adiabatic_slow_section, &
         barostat_section, print_section, region_section_fast, region_section_slow, &
         region_sections, thermo_baro_section, thermo_part_section, thermo_shell_section, &
         work_section

      NULLIFY (qmmm_env, cell)
      CALL allocate_thermostats(thermostats)
      adiabatic_fast_section => section_vals_get_subs_vals(md_section, "ADIABATIC_DYNAMICS%THERMOSTAT_FAST")
      adiabatic_slow_section => section_vals_get_subs_vals(md_section, "ADIABATIC_DYNAMICS%THERMOSTAT_SLOW")
      thermo_part_section => section_vals_get_subs_vals(md_section, "THERMOSTAT")
      thermo_shell_section => section_vals_get_subs_vals(md_section, "SHELL%THERMOSTAT")
      thermo_baro_section => section_vals_get_subs_vals(md_section, "BAROSTAT%THERMOSTAT")
      barostat_section => section_vals_get_subs_vals(md_section, "BAROSTAT")
      print_section => section_vals_get_subs_vals(md_section, "PRINT")

      CALL force_env_get(force_env, qmmm_env=qmmm_env, subsys=subsys, cell=cell)
      CALL section_vals_get(barostat_section, explicit=explicit_barostat_section)
      CALL section_vals_val_get(global_section, "SAVE_MEM", l_val=save_mem)
      CALL section_vals_get(thermo_part_section, explicit=explicit_part)
      CALL section_vals_get(thermo_shell_section, explicit=explicit_shell)
      CALL section_vals_get(thermo_baro_section, explicit=explicit_baro)
      CALL section_vals_get(adiabatic_fast_section, explicit=explicit_adiabatic_fast)
      CALL section_vals_get(adiabatic_slow_section, explicit=explicit_adiabatic_slow)

      apply_thermo_adiabatic = (simpar%ensemble == nvt_adiabatic_ensemble)

      apply_thermo_baro = (simpar%ensemble == npt_f_ensemble) .OR. &
                          (simpar%ensemble == npt_i_ensemble) .AND. &
                          (.NOT. apply_thermo_adiabatic)

      apply_general_thermo = apply_thermo_baro .OR. (simpar%ensemble == nvt_ensemble) .AND. &
                             (.NOT. apply_thermo_adiabatic)

      apply_thermo_shell = (simpar%ensemble == nve_ensemble) .OR. &
                           (simpar%ensemble == nvt_ensemble) .OR. &
                           (simpar%ensemble == npt_f_ensemble) .OR. &
                           (simpar%ensemble == npt_i_ensemble) .OR. &
                           (simpar%ensemble == npe_i_ensemble) .OR. &
                           (simpar%ensemble == npe_f_ensemble) .AND. &
                           (.NOT. apply_thermo_adiabatic)

      binary_restart_file_name = ""
      CALL section_vals_val_get(force_env%root_section, "EXT_RESTART%BINARY_RESTART_FILE_NAME", &
                                c_val=binary_restart_file_name)

      ! Compute Degrees of Freedom
      IF (simpar%ensemble == nvt_adiabatic_ensemble) THEN
         CALL cite_reference(VandeVondele2002)
         region = do_region_global
         region_section_fast => section_vals_get_subs_vals(adiabatic_fast_section, "DEFINE_REGION")
         region_section_slow => section_vals_get_subs_vals(adiabatic_slow_section, "DEFINE_REGION")
         IF (explicit_adiabatic_fast) CALL section_vals_val_get(adiabatic_fast_section, "REGION", i_val=region)
         IF (explicit_adiabatic_slow) CALL section_vals_val_get(adiabatic_slow_section, "REGION", i_val=region)
         CALL cp_subsys_get(subsys, molecule_kinds_new=molecule_kinds_new, local_molecules_new=local_molecules, &
                            molecules_new=molecules_new, gci=gci, particles=particles)
         CALL compute_nfree(cell, simpar, molecule_kinds_new%els, &
                            print_section, particles, gci)
         IF (explicit_adiabatic_fast .AND. explicit_adiabatic_slow) THEN
            IF (apply_thermo_adiabatic) THEN
               CALL create_thermostat_type(thermostats%thermostat_fast, simpar, adiabatic_fast_section, &
                                           label="FAST")
               CALL create_thermostat_type(thermostats%thermostat_slow, simpar, adiabatic_slow_section, &
                                           label="SLOW")
               CALL setup_adiabatic_thermostat_info(thermostats%thermostat_info_fast, &
                                                    molecule_kinds_new%els, local_molecules, molecules_new, particles, &
                                                    region, simpar%ensemble, region_sections=region_section_fast, &
                                                    qmmm_env=qmmm_env)

               CALL setup_adiabatic_thermostat_info(thermostats%thermostat_info_slow, &
                                                    molecule_kinds_new%els, local_molecules, molecules_new, particles, &
                                                    region, simpar%ensemble, region_sections=region_section_slow, &
                                                    qmmm_env=qmmm_env)

               ! Initialize or possibly restart Nose on Particles
               work_section => section_vals_get_subs_vals(adiabatic_fast_section, "NOSE")
               CALL initialize_nhc_fast(thermostats%thermostat_info_fast, simpar, local_molecules, &
                                        molecules_new%els, molecule_kinds_new%els, para_env, globenv, &
                                        thermostats%thermostat_fast%nhc, nose_section=work_section, gci=gci, &
                                        save_mem=save_mem)
               work_section => section_vals_get_subs_vals(adiabatic_slow_section, "NOSE")
               CALL initialize_nhc_slow(thermostats%thermostat_info_slow, simpar, local_molecules, &
                                        molecules_new%els, molecule_kinds_new%els, para_env, globenv, &
                                        thermostats%thermostat_slow%nhc, nose_section=work_section, gci=gci, &
                                        save_mem=save_mem)
            ENDIF
         ELSE
            CALL cp_warn(cp__l("motion/thermostat/thermostat_methods.F",222), &
                         "Adiabatic Thermostat has been defined but the ensemble provided "// &
                         "does not support thermostat for Particles! Ignoring thermostat input.")
         END IF
         CALL release_thermostat_info(thermostats%thermostat_info_fast)
         CALL release_thermostat_info(thermostats%thermostat_info_slow)
      ELSE
         region = do_region_global
         region_sections => section_vals_get_subs_vals(thermo_part_section, "DEFINE_REGION")
         IF (explicit_part) CALL section_vals_val_get(thermo_part_section, "REGION", i_val=region)
         CALL cp_subsys_get(subsys, molecule_kinds_new=molecule_kinds_new, local_molecules_new=local_molecules, &
                            molecules_new=molecules_new, gci=gci, particles=particles)
         CALL compute_degrees_of_freedom(thermostats, cell, simpar, molecule_kinds_new%els, &
                                         local_molecules, molecules_new, particles, print_section, region_sections, gci, &
                                         region, qmmm_env)

         ! Particles
         ! For constant temperature ensembles the thermostat is activated by default
         IF (explicit_part) THEN
            IF (apply_general_thermo) THEN
               CALL create_thermostat_type(thermostats%thermostat_part, simpar, thermo_part_section, &
                                           label="PARTICLES")
               ! Initialize thermostat
               IF (thermostats%thermostat_part%type_of_thermostat == do_thermo_nose) THEN
                  ! Initialize or possibly restart Nose on Particles
                  work_section => section_vals_get_subs_vals(thermo_part_section, "NOSE")
                  CALL initialize_nhc_part(thermostats%thermostat_info_part, simpar, local_molecules, &
                                           molecules_new%els, molecule_kinds_new%els, para_env, globenv, &
                                           thermostats%thermostat_part%nhc, nose_section=work_section, gci=gci, &
                                           save_mem=save_mem, binary_restart_file_name=binary_restart_file_name)
               ELSE IF (thermostats%thermostat_part%type_of_thermostat == do_thermo_csvr) THEN
                  ! Initialize or possibly restart CSVR thermostat on Particles
                  work_section => section_vals_get_subs_vals(thermo_part_section, "CSVR")
                  CALL initialize_csvr_part(thermostats%thermostat_info_part, simpar, local_molecules, &
                                            molecules_new%els, molecule_kinds_new%els, para_env, &
                                            thermostats%thermostat_part%csvr, csvr_section=work_section, &
                                            gci=gci)
               ELSE IF (thermostats%thermostat_part%type_of_thermostat == do_thermo_al) THEN
                  ! Initialize or possibly restart Ad-Langevin thermostat on Particles
                  work_section => section_vals_get_subs_vals(thermo_part_section, "AD_LANGEVIN")
                  CALL initialize_al_part(thermostats%thermostat_info_part, simpar, local_molecules, &
                                          molecules_new%els, molecule_kinds_new%els, para_env, &
                                          thermostats%thermostat_part%al, al_section=work_section, &
                                          gci=gci)
               ELSE IF (thermostats%thermostat_part%type_of_thermostat == do_thermo_gle) THEN
                  ! Initialize or possibly restart GLE thermostat on Particles
                  work_section => section_vals_get_subs_vals(thermo_part_section, "GLE")
                  CALL initialize_gle_part(thermostats%thermostat_info_part, simpar, local_molecules, &
                                           molecules_new%els, molecule_kinds_new%els, para_env, &
                                           thermostats%thermostat_part%gle, gle_section=work_section, &
                                           gci=gci, save_mem=save_mem)
               END IF
               CALL thermostat_info(thermostats%thermostat_part, "PARTICLES", thermo_part_section, &
                                    simpar, para_env)
            ELSE
               CALL cp_warn(cp__l("motion/thermostat/thermostat_methods.F",277), &
                            "Thermostat for Particles has been defined but the ensemble provided "// &
                            "does not support thermostat for Particles! Ignoring thermostat input.")
            END IF
         ELSE IF (apply_general_thermo) THEN
            CALL cp_abort(cp__l("motion/thermostat/thermostat_methods.F",282), &
                          "One constant temperature ensemble has been required, but no thermostat for the "// &
                          "particles has been defined. You may want to change your input and add a "// &
                          "THERMOSTAT section in the MD section.")
         END IF

         ! Core-Shell Model
         CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds)
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kinds%els, shell_present=shell_present)
         IF (shell_present) THEN
            IF (explicit_shell) THEN
               ! The thermostat is activated only if explicitely required
               ! It can be used to thermalize the shell-core motion when the temperature is not constant (nve, npe)
               IF (apply_thermo_shell) THEN
                  CALL create_thermostat_type(thermostats%thermostat_shell, simpar, thermo_shell_section, &
                                              label="SHELL")
                  CALL get_atomic_kind_set(atomic_kind_set=atomic_kinds%els, shell_adiabatic=shell_adiabatic)
                  region_sections => section_vals_get_subs_vals(thermo_shell_section, "DEFINE_REGION")
                  CALL section_vals_val_get(thermo_shell_section, "REGION", i_val=region)
                  CALL setup_thermostat_info( &
                     thermostats%thermostat_info_shell, molecule_kinds_new%els, &
                     local_molecules, molecules_new, particles, region, simpar%ensemble, shell=shell_adiabatic, &
                     region_sections=region_sections, qmmm_env=qmmm_env)
                  IF (shell_adiabatic) THEN
                     ! Initialize thermostat
                     IF (thermostats%thermostat_shell%type_of_thermostat == do_thermo_nose) THEN
                        ! Initialize or possibly restart Nose on Shells
                        work_section => section_vals_get_subs_vals(thermo_shell_section, "NOSE")
                        CALL initialize_nhc_shell(thermostats%thermostat_info_shell, simpar, local_molecules, &
                                                  molecules_new%els, molecule_kinds_new%els, para_env, globenv, &
                                                  thermostats%thermostat_shell%nhc, nose_section=work_section, gci=gci, &
                                                  save_mem=save_mem, binary_restart_file_name=binary_restart_file_name)
                     ELSE IF (thermostats%thermostat_shell%type_of_thermostat == do_thermo_csvr) THEN
                        ! Initialize or possibly restart CSVR thermostat on Shells
                        work_section => section_vals_get_subs_vals(thermo_shell_section, "CSVR")
                        CALL initialize_csvr_shell(thermostats%thermostat_info_shell, simpar, local_molecules, &
                                                   molecules_new%els, molecule_kinds_new%els, para_env, &
                                                   thermostats%thermostat_shell%csvr, csvr_section=work_section, gci=gci)
                     END IF
                     CALL thermostat_info(thermostats%thermostat_shell, "CORE-SHELL", thermo_shell_section, &
                                          simpar, para_env)
                  ELSE
                     CALL cp_warn(cp__l("motion/thermostat/thermostat_methods.F",324), &
                                  "Thermostat for Core-Shell motion only with adiabatic shell-model. "// &
                                  "Continuing calculation ignoring the thermostat info! No Thermostat "// &
                                  "applied to Shells!")
                     CALL release_thermostat_type(thermostats%thermostat_shell)
                     CALL release_thermostat_info(thermostats%thermostat_info_shell)
                  END IF
               ELSE
                  CALL cp_warn(cp__l("motion/thermostat/thermostat_methods.F",332), &
                               "Thermostat for Shells has been defined but for the selected ensemble the adiabatic  "// &
                               " shell model has not been implemented! Ignoring thermostat input.")
               END IF
            END IF
         ELSE IF (explicit_shell) THEN
            CALL cp_warn(cp__l("motion/thermostat/thermostat_methods.F",338), &
                         "Thermostat for Shells has been defined but the system provided "// &
                         "does not contain any Shells! Ignoring thermostat input.")
         END IF

         ! Barostat Temperature (not necessarily to be controlled by a thermostat)
         IF (explicit_barostat_section) THEN
            simpar%temp_baro_ext = simpar%temp_ext
            CALL section_vals_val_get(md_section, "BAROSTAT%TEMPERATURE", n_rep_val=n_rep)
            IF (n_rep /= 0) THEN
               CALL section_vals_val_get(md_section, "BAROSTAT%TEMPERATURE", r_val=simpar%temp_baro_ext)
               IF(.NOT.(simpar%temp_baro_ext >= 0.0_dp))CALL cp__a("motion/thermostat/thermostat_methods.F",349)
            END IF

            ! Setup Barostat Thermostat
            IF (apply_thermo_baro) THEN
               ! Check if we use the same thermostat as particles
               CALL section_vals_val_get(thermo_baro_section, "TYPE", i_val=thermostat_type)
               work_section => thermo_baro_section
               IF (thermostat_type == do_thermo_same_as_part) work_section => thermo_part_section

               CALL create_thermostat_type(thermostats%thermostat_baro, simpar, work_section, skip_region=.TRUE., &
                                           label="BAROSTAT")
               ! Initialize thermostat
               IF (thermostats%thermostat_baro%type_of_thermostat == do_thermo_nose) THEN
                  ! Initialize or possibly restart Nose on Barostat
                  work_section => section_vals_get_subs_vals(thermo_baro_section, "NOSE")
                  CALL initialize_nhc_baro(simpar, para_env, globenv, thermostats%thermostat_baro%nhc, &
                                           nose_section=work_section, save_mem=save_mem)
               ELSE IF (thermostats%thermostat_baro%type_of_thermostat == do_thermo_csvr) THEN
                  ! Initialize or possibly restart CSVR thermostat on Barostat
                  work_section => section_vals_get_subs_vals(thermo_baro_section, "CSVR")
                  CALL initialize_csvr_baro(simpar, thermostats%thermostat_baro%csvr, &
                                            csvr_section=work_section)
               END IF
               CALL thermostat_info(thermostats%thermostat_baro, "BAROSTAT", thermo_baro_section, &
                                    simpar, para_env)
               ! If thermostat for barostat uses a diffent kind than the one of the particles
               ! let's update infos in the input structure..
               IF (thermostat_type == do_thermo_same_as_part) THEN
                  CALL update_thermo_baro_section(thermostats%thermostat_baro, thermo_baro_section)
               END IF
            ELSE
               IF (explicit_baro) THEN
                  CALL cp_warn(cp__l("motion/thermostat/thermostat_methods.F",382), &
                               "Thermostat for Barostat has been defined but the ensemble provided "// &
                               "does not support thermostat for Barostat! Ignoring thermostat input.")
               END IF
               ! Let's remove the section
               CALL section_vals_remove_values(thermo_baro_section)
            END IF
         END IF

         ! Release the thermostats info..
         CALL release_thermostat_info(thermostats%thermostat_info_part)
         CALL release_thermostat_info(thermostats%thermostat_info_shell)

      ENDIF ! Adiabitic_NVT screening
      ! If no thermostats have been allocated deallocate the full structure
      IF ((.NOT. ASSOCIATED(thermostats%thermostat_part)) .AND. &
          (.NOT. ASSOCIATED(thermostats%thermostat_shell)) .AND. &
          (.NOT. ASSOCIATED(thermostats%thermostat_baro)) .AND. &
          (.NOT. ASSOCIATED(thermostats%thermostat_fast)) .AND. &
          (.NOT. ASSOCIATED(thermostats%thermostat_slow))) THEN
         CALL release_thermostats(thermostats)
      END IF

   END SUBROUTINE create_thermostats

! **************************************************************************************************
!> \brief ...
!> \param thermostat ...
!> \param section ...
!> \par History
!>      10.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE update_thermo_baro_section(thermostat, section)
      TYPE(thermostat_type), POINTER                     :: thermostat
      TYPE(section_vals_type), POINTER                   :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'update_thermo_baro_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_vals_type), POINTER                   :: work_section

      CALL section_vals_val_set(section, "TYPE", i_val=thermostat%type_of_thermostat)
      SELECT CASE (thermostat%type_of_thermostat)
      CASE (do_thermo_nose)
         work_section => section_vals_get_subs_vals(section, "NOSE")
         CALL section_vals_val_set(work_section, "LENGTH", i_val=thermostat%nhc%nhc_len)
         CALL section_vals_val_set(work_section, "YOSHIDA", i_val=thermostat%nhc%nyosh)
         CALL section_vals_val_set(work_section, "TIMECON", r_val=thermostat%nhc%tau_nhc)
         CALL section_vals_val_set(work_section, "MTS", i_val=thermostat%nhc%nc)
      CASE (do_thermo_csvr)
         work_section => section_vals_get_subs_vals(section, "CSVR")
         CALL section_vals_val_set(work_section, "TIMECON", r_val=thermostat%csvr%tau_csvr)
      CASE (do_thermo_al)
         work_section => section_vals_get_subs_vals(section, "AD_LANGEVIN")
         CALL section_vals_val_set(work_section, "TIMECON_NH", r_val=thermostat%al%tau_nh)
         CALL section_vals_val_set(work_section, "TIMECON_LANGEVIN", r_val=thermostat%al%tau_langevin)
      END SELECT
   END SUBROUTINE update_thermo_baro_section

! **************************************************************************************************
!> \brief ...
!> \param thermostat ...
!> \param label ...
!> \param section ...
!> \param simpar ...
!> \param para_env ...
!> \par History
!>      10.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE thermostat_info(thermostat, label, section, simpar, para_env)
      TYPE(thermostat_type), POINTER                     :: thermostat
      CHARACTER(LEN=*), INTENT(IN)                       :: label
      TYPE(section_vals_type), POINTER                   :: section
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'thermostat_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iw
      REAL(KIND=dp)                                      :: kin_energy, pot_energy, tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, section, "PRINT%THERMOSTAT_INFO", extension=".log")
      ! Total Tehrmostat Energy
      CALL get_thermostat_energies(thermostat, pot_energy, kin_energy, para_env)
      IF (iw > 0) THEN
         WRITE (iw, '( /,A )') ' THERMOSTAT| Thermostat Info for '//TRIM(label)
         SELECT CASE (thermostat%type_of_thermostat)
         CASE (do_thermo_nose)
            WRITE (iw, '( A,T63,A )') ' THERMOSTAT| Type of thermostat', 'Nose-Hoover-Chains'
            WRITE (iw, '( A,T77,I4 )') ' THERMOSTAT| Nose-Hoover-Chain length ', thermostat%nhc%nhc_len
            tmp = cp_unit_from_cp2k(thermostat%nhc%tau_nhc, "fs")
            WRITE (iw, '( A,A4,A,T71,F10.2 )') &
               ' THERMOSTAT| Nose-Hoover-Chain time constant [', 'fs', '] ', tmp
            WRITE (iw, '( A,T77,I4 )') ' THERMOSTAT| Order of Yoshida integrator ', &
               thermostat%nhc%nyosh
            WRITE (iw, '( A,T77,I4 )') ' THERMOSTAT| Number of multiple time steps', &
               thermostat%nhc%nc
            WRITE (iw, '( A,T69,F12.6 )') ' THERMOSTAT| Initial Potential Energy', &
               pot_energy
            WRITE (iw, '( A,T69,F12.6 )') ' THERMOSTAT| Initial Kinetic Energy', &
               kin_energy
         CASE (do_thermo_csvr)
            WRITE (iw, '( A,T44,A )') ' THERMOSTAT| Type of thermostat', 'Canonical Sampling/Velocity Rescaling'
            tmp = cp_unit_from_cp2k(thermostat%csvr%tau_csvr, "fs")*0.5_dp*simpar%dt
            WRITE (iw, '( A,A4,A,T71,F10.2 )') &
               ' THERMOSTAT| CSVR time constant [', 'fs', '] ', tmp
            WRITE (iw, '( A,T69,F12.6 )') &
               ' THERMOSTAT| Initial Kinetic Energy ', kin_energy
         CASE (do_thermo_al)
            WRITE (iw, '( A,T44,A )') ' THERMOSTAT| Type of thermostat', 'Adaptive Langevin'
            tmp = cp_unit_from_cp2k(thermostat%al%tau_nh, "fs")
            WRITE (iw, '( A,A4,A,T71,F10.2 )') &
               ' THERMOSTAT| AD_LANGEVIN NH time constant [', 'fs', '] ', tmp
            tmp = cp_unit_from_cp2k(thermostat%al%tau_langevin, "fs")
            WRITE (iw, '( A,A4,A,T71,F10.2 )') &
               ' THERMOSTAT| AD_LANGEVIN Langevin time constant [', 'fs', '] ', tmp
         END SELECT
         WRITE (iw, '( A,/ )') ' THERMOSTAT| End of Thermostat Info for '//TRIM(label)
      END IF
      CALL cp_print_key_finished_output(iw, logger, section, "PRINT%THERMOSTAT_INFO")

   END SUBROUTINE thermostat_info

! **************************************************************************************************
!> \brief ...
!> \param thermostat ...
!> \param npt ...
!> \param group ...
!> \par History
!>      10.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE apply_thermostat_baro(thermostat, npt, group)
      TYPE(thermostat_type), POINTER                     :: thermostat
      TYPE(npt_info_type), DIMENSION(:, :), POINTER      :: npt
      INTEGER, INTENT(IN)                                :: group

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_thermostat_baro', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(thermostat)) THEN
         IF (thermostat%type_of_thermostat == do_thermo_nose) THEN
            ! Apply Nose-Hoover Thermostat
            IF(.NOT.(ASSOCIATED(thermostat%nhc)))CALL cp__a("motion/thermostat/thermostat_methods.F",531)
            CALL lnhc_barostat(thermostat%nhc, npt, group)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_csvr) THEN
            ! Apply CSVR Thermostat
            IF(.NOT.(ASSOCIATED(thermostat%csvr)))CALL cp__a("motion/thermostat/thermostat_methods.F",535)
            CALL csvr_barostat(thermostat%csvr, npt, group)
         END IF
      END IF
   END SUBROUTINE apply_thermostat_baro

! **************************************************************************************************
!> \brief ...
!> \param thermostat ...
!> \param force_env ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param particle_set ...
!> \param local_molecules ...
!> \param local_particles ...
!> \param group ...
!> \param shell_adiabatic ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \par History
!>      10.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE apply_thermostat_particles(thermostat, force_env, molecule_kind_set, molecule_set, &
                                         particle_set, local_molecules, local_particles, &
                                         group, shell_adiabatic, shell_particle_set, &
                                         core_particle_set, vel, shell_vel, core_vel)

      TYPE(thermostat_type), POINTER                     :: thermostat
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      INTEGER, INTENT(IN)                                :: group
      LOGICAL, INTENT(IN), OPTIONAL                      :: shell_adiabatic
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :), shell_vel(:, :), &
                                                            core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_thermostat_particles', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(thermostat)) THEN
         IF (thermostat%type_of_thermostat == do_thermo_nose) THEN
            ! Apply Nose-Hoover Thermostat
            IF(.NOT.(ASSOCIATED(thermostat%nhc)))CALL cp__a("motion/thermostat/thermostat_methods.F",585)
            CALL lnhc_particles(thermostat%nhc, molecule_kind_set, molecule_set, &
                                particle_set, local_molecules, group, shell_adiabatic, shell_particle_set, &
                                core_particle_set, vel, shell_vel, core_vel)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_csvr) THEN
            ! Apply CSVR Thermostat
            IF(.NOT.(ASSOCIATED(thermostat%csvr)))CALL cp__a("motion/thermostat/thermostat_methods.F",591)
            CALL csvr_particles(thermostat%csvr, molecule_kind_set, molecule_set, &
                                particle_set, local_molecules, group, shell_adiabatic, shell_particle_set, &
                                core_particle_set, vel, shell_vel, core_vel)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_al) THEN
            ! Apply AD_LANGEVIN Thermostat
            IF(.NOT.(ASSOCIATED(thermostat%al)))CALL cp__a("motion/thermostat/thermostat_methods.F",597)
            CALL al_particles(thermostat%al, force_env, molecule_kind_set, molecule_set, &
                              particle_set, local_molecules, local_particles, group, vel)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_gle) THEN
            ! Apply GLE Thermostat
            IF(.NOT.(ASSOCIATED(thermostat%gle)))CALL cp__a("motion/thermostat/thermostat_methods.F",602)
            CALL gle_particles(thermostat%gle, molecule_kind_set, molecule_set, &
                               particle_set, local_molecules, group, shell_adiabatic, shell_particle_set, &
                               core_particle_set, vel, shell_vel, core_vel)
         END IF
      END IF
   END SUBROUTINE apply_thermostat_particles

! **************************************************************************************************
!> \brief ...
!> \param thermostat ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param group ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param vel ...
!> \param shell_vel ...
!> \param core_vel ...
!> \par History
!>      10.2007 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE apply_thermostat_shells(thermostat, atomic_kind_set, particle_set, &
                                      local_particles, group, shell_particle_set, core_particle_set, vel, shell_vel, &
                                      core_vel)

      TYPE(thermostat_type), POINTER                     :: thermostat
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      INTEGER, INTENT(IN)                                :: group
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: vel(:, :), shell_vel(:, :), &
                                                            core_vel(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_thermostat_shells', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(thermostat)) THEN
         IF (thermostat%type_of_thermostat == do_thermo_nose) THEN
            ! Apply Nose-Hoover Thermostat
            IF(.NOT.(ASSOCIATED(thermostat%nhc)))CALL cp__a("motion/thermostat/thermostat_methods.F",646)
            CALL lnhc_shells(thermostat%nhc, atomic_kind_set, particle_set, local_particles, &
                             group, shell_particle_set, core_particle_set, vel, shell_vel, core_vel)
         ELSE IF (thermostat%type_of_thermostat == do_thermo_csvr) THEN
            ! Apply CSVR Thermostat
            IF(.NOT.(ASSOCIATED(thermostat%csvr)))CALL cp__a("motion/thermostat/thermostat_methods.F",651)
            CALL csvr_shells(thermostat%csvr, atomic_kind_set, particle_set, local_particles, &
                             group, shell_particle_set, core_particle_set, vel, shell_vel, core_vel)
         END IF
      END IF
   END SUBROUTINE apply_thermostat_shells

END MODULE thermostat_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Initialize the analysis of trajectories to be done
!>      by activating the REFTRAJ ensemble
!> \par History
!>      Created 10-07 [MI]
!> \author MI
! **************************************************************************************************
MODULE reftraj_util

   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_sum
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: get_molecule_kind,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              molecule_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: angstrom,&
                                              femtoseconds
   USE reftraj_types,                   ONLY: reftraj_msd_type,&
                                              reftraj_type
   USE simpar_types,                    ONLY: simpar_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

      PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'reftraj_util'

      PUBLIC ::   initialize_reftraj, compute_msd_reftraj, write_output_reftraj

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param reftraj ...
!> \param reftraj_section ...
!> \param md_env ...
!> \par History
!>      10.2007 created
!> \author MI
! **************************************************************************************************
  SUBROUTINE initialize_reftraj(reftraj,reftraj_section,md_env)

      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(section_vals_type), POINTER                   :: reftraj_section
      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_reftraj', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: natom, nline_to_skip, nskip
      LOGICAL                                            :: my_end
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: msd_section
      TYPE(simpar_type), POINTER                         :: simpar

    NULLIFY (force_env, msd_section, particles, simpar, subsys)
    CALL get_md_env(md_env=md_env, force_env=force_env, para_env=para_env,&
         simpar=simpar)
    CALL force_env_get(force_env=force_env,  subsys=subsys)
    CALL cp_subsys_get(subsys=subsys, particles=particles)
    natom = particles%n_els

    my_end = .FALSE.
    nline_to_skip = 0

    nskip = reftraj%info%first_snapshot-1
    IF(.NOT.(nskip>=0))CALL cp__a("motion/reftraj_util.F",108)

    IF(nskip > 0 ) THEN
      nline_to_skip = (natom+2)*nskip
      CALL parser_get_next_line(reftraj%info%traj_parser,nline_to_skip,at_end=my_end)
    END IF

    reftraj%isnap = nskip
    IF(my_end)&
       CALL cp_abort(cp__l("motion/reftraj_util.F",117),&
            "Reached the end of the trajectory file for REFTRAJ. Number of steps skipped "//&
            "equal to the number of steps present in the file.")

    ! Cell File
    IF(reftraj%info%variable_volume) THEN
       IF(nskip > 0 ) THEN
          CALL parser_get_next_line(reftraj%info%cell_parser,nskip,at_end=my_end)
       END IF
       IF(my_end)&
          CALL cp_abort(cp__l("motion/reftraj_util.F",127),&
               "Reached the end of the cell file for REFTRAJ. Number of steps skipped "//&
               "equal to the number of steps present in the file.")
    END IF


    reftraj%natom = natom
    IF(reftraj%info%last_snapshot>0) THEN
       simpar%nsteps = (reftraj%info%last_snapshot - reftraj%info%first_snapshot + 1)
    END IF

    IF(reftraj%info%msd) THEN
      msd_section =>  section_vals_get_subs_vals(reftraj_section,"MSD")
      ! set up and printout
      CALL initialize_msd_reftraj(reftraj%msd,msd_section,reftraj,md_env)
    END IF

  END SUBROUTINE initialize_reftraj

! **************************************************************************************************
!> \brief ...
!> \param msd ...
!> \param msd_section ...
!> \param reftraj ...
!> \param md_env ...
!> \par History
!>      10.2007 created
!> \author MI
! **************************************************************************************************
  SUBROUTINE initialize_msd_reftraj(msd,msd_section,reftraj,md_env)
      TYPE(reftraj_msd_type), POINTER                    :: msd
      TYPE(section_vals_type), POINTER                   :: msd_section
      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_msd_reftraj', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=10)                                  :: AA
      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: name, title
      INTEGER                                            :: first_atom, iatom, ikind, imol, &
                                                            last_atom, natom_read, nkind, nmol, &
                                                            nmolecule, nmolkind, npart
      REAL(KIND=dp)                                      :: com(3), mass, mass_mol, tol, x, y, z
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

    NULLIFY (molecule, molecules, molecule_kind, molecule_kind_set,&
             molecule_kinds, molecule_set, subsys, force_env, particles, particle_set)
    IF(.NOT.(.NOT. ASSOCIATED(msd)))CALL cp__a("motion/reftraj_util.F",187)

    ALLOCATE(msd)

    NULLIFY(msd%ref0_pos)
    NULLIFY(msd%ref0_com_molecule)
    NULLIFY(msd%val_msd_kind)
    NULLIFY(msd%val_msd_molecule)
    NULLIFY(msd%disp_atom_index)
    NULLIFY(msd%disp_atom_dr)

    CALL get_md_env(md_env=md_env, force_env=force_env, para_env=para_env)
    CALL force_env_get(force_env=force_env,subsys=subsys)
    CALL cp_subsys_get(subsys=subsys, particles=particles)
    particle_set => particles%els
    npart = SIZE(particle_set,1)

    msd%ref0_unit = -1
    CALL section_vals_val_get(msd_section,"REF0_FILENAME",c_val=filename)
    CALL open_file(TRIM(filename),unit_number=msd%ref0_unit)

    ALLOCATE(msd%ref0_pos(3,reftraj%natom))
    msd%ref0_pos = 0.0_dp

    IF (para_env%mepos==para_env%source) THEN
      REWIND(msd%ref0_unit)
      READ(msd%ref0_unit,*,ERR=999,END=998) natom_read
      IF(.NOT.(natom_read==reftraj%natom))CALL cp__a("motion/reftraj_util.F",214)
      READ(msd%ref0_unit,'(A)',ERR=999,END=998) title
      msd%total_mass = 0.0_dp
      msd%ref0_com = 0.0_dp
      DO iatom = 1,natom_read
          READ(msd%ref0_unit,*,ERR=999,END=998)  AA, x, y, z
          name=TRIM(particle_set(iatom)%atomic_kind%element_symbol)
          IF(.NOT.((TRIM(AA)==name)))CALL cp__a("motion/reftraj_util.F",221)

          x = cp_unit_to_cp2k(x,"angstrom")
          y = cp_unit_to_cp2k(y,"angstrom")
          z = cp_unit_to_cp2k(z,"angstrom")
          msd%ref0_pos(1,iatom) = x
          msd%ref0_pos(2,iatom) = y
          msd%ref0_pos(3,iatom) = z
          mass = particle_set(iatom)%atomic_kind%mass
          msd%ref0_com(1) = msd%ref0_com(1) + x * mass
          msd%ref0_com(2) = msd%ref0_com(2) + y * mass
          msd%ref0_com(3) = msd%ref0_com(3) + z * mass
          msd%total_mass = msd%total_mass + mass
      END DO
      msd%ref0_com = msd%ref0_com / msd%total_mass
    END IF
    CALL close_file(unit_number=msd%ref0_unit)

    CALL mp_bcast(msd%total_mass,para_env%source,para_env%group)
    CALL mp_bcast(msd%ref0_pos,para_env%source,para_env%group)
    CALL mp_bcast(msd%ref0_com,para_env%source,para_env%group)

    CALL section_vals_val_get(msd_section,"MSD_PER_KIND",l_val=msd%msd_kind)
    CALL section_vals_val_get(msd_section,"MSD_PER_MOLKIND",l_val=msd%msd_molecule)
    CALL section_vals_val_get(msd_section,"MSD_PER_REGION",l_val=msd%msd_region)

    CALL section_vals_val_get(msd_section,"DISPLACED_ATOM",l_val=msd%disp_atom)
    IF(msd%disp_atom) THEN
       ALLOCATE(msd%disp_atom_index(npart))
       msd%disp_atom_index = 0
       ALLOCATE(msd%disp_atom_dr(3,npart))
       msd%disp_atom_dr = 0.0_dp
       msd%msd_kind = .TRUE.
    END IF
    CALL section_vals_val_get(msd_section,"DISPLACEMENT_TOL",r_val=tol)
    msd%disp_atom_tol = tol*tol

    IF(msd%msd_kind) THEN
      CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds)
      nkind = atomic_kinds%n_els

      ALLOCATE(msd%val_msd_kind(4,nkind))
      msd%val_msd_kind = 0.0_dp
    END IF

    IF(msd%msd_molecule) THEN
      CALL cp_subsys_get(subsys=subsys, molecules_new=molecules,&
           molecule_kinds_new=molecule_kinds)
      nmolkind =  molecule_kinds%n_els
      ALLOCATE(msd%val_msd_molecule(4,nmolkind))

      molecule_kind_set => molecule_kinds%els
      molecule_set => molecules%els
      nmol = molecules%n_els

      ALLOCATE(msd%ref0_com_molecule(3,nmol))

      DO ikind = 1, nmolkind
        molecule_kind => molecule_kind_set(ikind)
        CALL get_molecule_kind (molecule_kind=molecule_kind, nmolecule=nmolecule )
        DO imol = 1,nmolecule
           molecule => molecule_set(molecule_kind%molecule_list(imol))
           CALL get_molecule ( molecule=molecule, first_atom = first_atom, last_atom = last_atom )
           com = 0.0_dp
           mass_mol = 0.0_dp
           DO iatom = first_atom, last_atom
              mass = particle_set(iatom)%atomic_kind%mass
              com(1) = com(1) + msd%ref0_pos(1,iatom)*mass
              com(2) = com(2) + msd%ref0_pos(2,iatom)*mass
              com(3) = com(3) + msd%ref0_pos(3,iatom)*mass
              mass_mol = mass_mol+mass
           ENDDO  ! iatom
           msd%ref0_com_molecule(1,molecule_kind%molecule_list(imol)) =  com(1)/mass_mol
           msd%ref0_com_molecule(2,molecule_kind%molecule_list(imol)) =  com(2)/mass_mol
           msd%ref0_com_molecule(3,molecule_kind%molecule_list(imol)) =  com(3)/mass_mol
        END DO  ! imol
      ENDDO ! ikind
    END IF

    IF(msd%msd_region) THEN

    END IF

    RETURN
998 CONTINUE ! end of file
    CALL cp__b("motion/reftraj_util.F",306,"End of reference positions file reached")
999 CONTINUE ! error
    CALL cp__b("motion/reftraj_util.F",308,"Error reading reference positions file")

  END SUBROUTINE initialize_msd_reftraj

! **************************************************************************************************
!> \brief ...
!> \param reftraj ...
!> \param md_env ...
!> \param particle_set ...
!> \par History
!>      10.2007 created
!> \author MI
! **************************************************************************************************
  SUBROUTINE compute_msd_reftraj(reftraj,md_env,particle_set)

      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'compute_msd_reftraj', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom, bo(2), first_atom, iatom, ikind, imol, imol_global, last_atom, mepos, &
         natom_kind, nmol_per_kind, nmolecule, nmolkind, num_pe
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      REAL(KIND=dp)                                      :: com(3), diff2_com(4), dr2, dx, dy, dz, &
                                                            mass, mass_mol, msd_mkind(4), rcom(3)
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(molecule_type), POINTER                       :: molecule

    NULLIFY(force_env,para_env,subsys)
    NULLIFY(atomic_kind,atomic_kinds,atom_list)
    NULLIFY(local_molecules, molecule, molecule_kind, molecule_kinds,&
            molecule_kind_set, molecules, molecule_set)

    CALL get_md_env(md_env=md_env, force_env=force_env, para_env=para_env)
    CALL force_env_get(force_env=force_env,subsys=subsys)
    CALL cp_subsys_get(subsys=subsys, atomic_kinds=atomic_kinds)

    num_pe = para_env%num_pe
    mepos  = para_env%mepos


    IF(reftraj%msd%msd_kind)THEN
      reftraj%msd%val_msd_kind = 0.0_dp
      reftraj%msd%num_disp_atom =  0
      reftraj%msd%disp_atom_dr =  0.0_dp
! compute com
      rcom = 0.0_dp
      DO ikind=1,atomic_kinds%n_els
          atomic_kind => atomic_kinds%els(ikind)
          CALL get_atomic_kind(atomic_kind=atomic_kind,&
                               atom_list=atom_list,&
                               natom=natom_kind, mass=mass)
          bo = get_limit(natom_kind,num_pe,mepos)
          DO iatom = bo(1), bo(2)
             atom = atom_list(iatom)
             rcom(1) = rcom(1) + particle_set(atom)%r(1)*mass
             rcom(2) = rcom(2) + particle_set(atom)%r(2)*mass
             rcom(3) = rcom(3) + particle_set(atom)%r(3)*mass
          END DO
      END DO
      CALL mp_sum(rcom,para_env%group)
      rcom = rcom / reftraj%msd%total_mass
      reftraj%msd%drcom (1) = rcom(1) - reftraj%msd%ref0_com(1)
      reftraj%msd%drcom (2) = rcom(2) - reftraj%msd%ref0_com(2)
      reftraj%msd%drcom (3) = rcom(3) - reftraj%msd%ref0_com(3)
!      IF(para_env%ionode) WRITE(*,'(A,T50,3f10.5)') ' COM displacement (dx,dy,dz) [angstrom]:  ', &
!                         drcom(1)*angstrom,drcom(2)*angstrom,drcom(3)*angstrom
! compute_com

      DO ikind=1,atomic_kinds%n_els
          atomic_kind => atomic_kinds%els(ikind)
          CALL get_atomic_kind(atomic_kind=atomic_kind,&
                               atom_list=atom_list,&
                               natom=natom_kind)
          bo = get_limit(natom_kind,num_pe,mepos)
          DO iatom = bo(1), bo(2)
            atom = atom_list(iatom)
            dx =  particle_set(atom)%r(1)-reftraj%msd%ref0_pos(1,atom) - &
                  reftraj%msd%drcom(1)
            dy =  particle_set(atom)%r(2)-reftraj%msd%ref0_pos(2,atom) - &
                  reftraj%msd%drcom(2)
            dz =  particle_set(atom)%r(3)-reftraj%msd%ref0_pos(3,atom) - &
                  reftraj%msd%drcom(3)
            dr2 = dx*dx +  dy*dy + dz*dz

            reftraj%msd%val_msd_kind(1,ikind) = reftraj%msd%val_msd_kind(1,ikind) + dx*dx
            reftraj%msd%val_msd_kind(2,ikind) = reftraj%msd%val_msd_kind(2,ikind) + dy*dy
            reftraj%msd%val_msd_kind(3,ikind) = reftraj%msd%val_msd_kind(3,ikind) + dz*dz
            reftraj%msd%val_msd_kind(4,ikind) = reftraj%msd%val_msd_kind(4,ikind) + dr2

            IF(reftraj%msd%disp_atom) THEN
              IF( dr2 > reftraj%msd%disp_atom_tol ) THEN
                reftraj%msd%num_disp_atom  = reftraj%msd%num_disp_atom  +1
                reftraj%msd%disp_atom_dr(1, atom) = dx
                reftraj%msd%disp_atom_dr(2, atom) = dy
                reftraj%msd%disp_atom_dr(3, atom) = dz
              END IF
            END IF
          END DO  !iatom
          reftraj%msd%val_msd_kind(1:4,ikind) = &
                     reftraj%msd%val_msd_kind(1:4,ikind)/REAL(natom_kind,KIND=dp)

      END DO  ! ikind
    ENDIF
    CALL mp_sum(reftraj%msd%val_msd_kind,para_env%group)
    CALL mp_sum(reftraj%msd%num_disp_atom,para_env%group)
    CALL mp_sum(reftraj%msd%disp_atom_dr,para_env%group)

    IF(reftraj%msd%msd_molecule) THEN
      CALL cp_subsys_get(subsys=subsys, local_molecules_new=local_molecules, &
           molecules_new=molecules, molecule_kinds_new=molecule_kinds)

      nmolkind =  molecule_kinds%n_els
      molecule_kind_set => molecule_kinds%els
      molecule_set => molecules%els

      reftraj%msd%val_msd_molecule = 0.0_dp
      DO ikind = 1,nmolkind
         molecule_kind => molecule_kind_set(ikind)
         CALL get_molecule_kind (molecule_kind=molecule_kind, nmolecule=nmolecule )
         nmol_per_kind =  local_molecules%n_el(ikind)
         msd_mkind = 0.0_dp
         DO imol = 1, nmol_per_kind
            imol_global = local_molecules%list(ikind)%array(imol)
            molecule => molecule_set ( imol_global )
            CALL get_molecule (molecule,first_atom=first_atom,last_atom=last_atom)

            com = 0.0_dp
            mass_mol = 0.0_dp
            DO iatom = first_atom, last_atom
              mass = particle_set(iatom)%atomic_kind%mass
              com(1) = com(1) + particle_set(iatom)%r(1)*mass
              com(2) = com(2) + particle_set(iatom)%r(2)*mass
              com(3) = com(3) + particle_set(iatom)%r(3)*mass
              mass_mol = mass_mol+mass
            ENDDO  ! iatom
            com(1) =  com(1)/mass_mol
            com(2) =  com(2)/mass_mol
            com(3) =  com(3)/mass_mol
            diff2_com(1) = com(1)- reftraj%msd%ref0_com_molecule(1,imol_global)
            diff2_com(2) = com(2)- reftraj%msd%ref0_com_molecule(2,imol_global)
            diff2_com(3) = com(3)- reftraj%msd%ref0_com_molecule(3,imol_global)
            diff2_com(1) = diff2_com(1)*diff2_com(1)
            diff2_com(2) = diff2_com(2)*diff2_com(2)
            diff2_com(3) = diff2_com(3)*diff2_com(3)
            diff2_com(4) = diff2_com(1) + diff2_com(2) + diff2_com(3)
            msd_mkind(1) = msd_mkind(1) + diff2_com(1)
            msd_mkind(2) = msd_mkind(2) + diff2_com(2)
            msd_mkind(3) = msd_mkind(3) + diff2_com(3)
            msd_mkind(4) = msd_mkind(4) + diff2_com(4)
         ENDDO ! imol

         reftraj%msd%val_msd_molecule(1,ikind) = msd_mkind(1)/REAL(nmolecule,KIND=dp)
         reftraj%msd%val_msd_molecule(2,ikind) = msd_mkind(2)/REAL(nmolecule,KIND=dp)
         reftraj%msd%val_msd_molecule(3,ikind) = msd_mkind(3)/REAL(nmolecule,KIND=dp)
         reftraj%msd%val_msd_molecule(4,ikind) = msd_mkind(4)/REAL(nmolecule,KIND=dp)
      END DO  ! ikind
      CALL mp_sum(reftraj%msd%val_msd_molecule, para_env%group)

    END IF

  END SUBROUTINE compute_msd_reftraj

! **************************************************************************************************
!> \brief ...
!> \param md_env ...
!> \par History
!>      10.2007 created
!> \author MI
! **************************************************************************************************
  SUBROUTINE write_output_reftraj(md_env)
      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'write_output_reftraj', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_act, my_mittle, my_pos
      INTEGER                                            :: iat, ikind, nkind, out_msd
      LOGICAL, SAVE                                      :: first_entry = .FALSE.
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(section_vals_type), POINTER                   :: reftraj_section, root_section

    NULLIFY(logger)
    logger => cp_get_default_logger()

    NULLIFY(reftraj)
    NULLIFY(reftraj_section,root_section)

    CALL get_md_env(md_env=md_env, force_env=force_env, &
         reftraj=reftraj)

    CALL force_env_get(force_env=force_env,root_section=root_section)

    reftraj_section => section_vals_get_subs_vals(root_section,&
           "MOTION%MD%REFTRAJ")

    my_pos = "APPEND"
    my_act = "WRITE"

    IF(reftraj%init.AND.(reftraj%isnap==reftraj%info%first_snapshot)) THEN
      my_pos  = "REWIND"
      first_entry = .TRUE.
    END IF

    IF(reftraj%info%msd) THEN
      IF(reftraj%msd%msd_kind)THEN
         nkind = SIZE(reftraj%msd%val_msd_kind,2)
         DO ikind = 1,nkind
           my_mittle ="k"//TRIM(ADJUSTL(cp_to_string(ikind)))
           out_msd = cp_print_key_unit_nr(logger,reftraj_section,"PRINT%MSD_KIND",&
               extension=".msd", file_position=my_pos, file_action=my_act,&
               file_form="FORMATTED", middle_name=TRIM(my_mittle))
           IF(out_msd>0) THEN
              WRITE(UNIT=out_msd,FMT="(I8, F12.3,4F20.10)") reftraj%itimes, &
                   reftraj%time*femtoseconds, &
                   reftraj%msd%val_msd_kind(1:4,ikind)*angstrom*angstrom
              CALL m_flush(out_msd)
           END IF
           CALL cp_print_key_finished_output(out_msd,logger,reftraj_section,&
                "PRINT%MSD_KIND")
         END DO
      END IF
      IF(reftraj%msd%msd_molecule) THEN
         nkind = SIZE(reftraj%msd%val_msd_molecule,2)
         DO ikind = 1,nkind
           my_mittle ="mk"//TRIM(ADJUSTL(cp_to_string(ikind)))
           out_msd = cp_print_key_unit_nr(logger,reftraj_section,"PRINT%MSD_MOLECULE",&
               extension=".msd", file_position=my_pos, file_action=my_act,&
               file_form="FORMATTED", middle_name=TRIM(my_mittle))
           IF(out_msd>0) THEN
              WRITE(UNIT=out_msd,FMT="(I8, F12.3,4F20.10)") reftraj%itimes, &
                   reftraj%time*femtoseconds, &
                   reftraj%msd%val_msd_molecule(1:4,ikind)*angstrom*angstrom
              CALL m_flush(out_msd)
           END IF
           CALL cp_print_key_finished_output(out_msd,logger,reftraj_section,&
                "PRINT%MSD_MOLECULE")
         END DO
      END IF
      IF(reftraj%msd%disp_atom) THEN

         IF(first_entry)  my_pos = "REWIND"
         my_mittle ="disp_at"
         out_msd = cp_print_key_unit_nr(logger,reftraj_section,"PRINT%DISPLACED_ATOM",&
               extension=".msd", file_position=my_pos, file_action=my_act,&
               file_form="FORMATTED", middle_name=TRIM(my_mittle))
         IF(out_msd>0 .AND. reftraj%msd%num_disp_atom>0) THEN
            IF(first_entry) THEN
               first_entry = .FALSE.
            END IF
            WRITE(UNIT=out_msd,FMT="(A,T7,I8, A, T29, F12.3, A, T50, I10)") "# i = ", reftraj%itimes, "  time (fs) = ",&
                   reftraj%time*femtoseconds, "  nat = ", reftraj%msd%num_disp_atom
            DO iat = 1,SIZE(reftraj%msd%disp_atom_dr,2)
               IF(ABS(reftraj%msd%disp_atom_dr(1,iat)) > 0.0_dp) THEN
                 WRITE(UNIT=out_msd,FMT="(I8, 3F20.10)") iat,& !reftraj%msd%disp_atom_index(iat),&
                      reftraj%msd%disp_atom_dr(1,iat)*angstrom, &
                      reftraj%msd%disp_atom_dr(2,iat)*angstrom,&
                      reftraj%msd%disp_atom_dr(3,iat)*angstrom
               END IF
            END DO
         ENDIF
         CALL cp_print_key_finished_output(out_msd,logger,reftraj_section,&
                "PRINT%DISPLACED_ATOM")
      END IF
    ENDIF ! msd
    reftraj%init = .FALSE.

  END SUBROUTINE write_output_reftraj

END MODULE reftraj_util

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides integrator routines (velocity verlet) for all the
!>      ensemble types
!> \par History
!>      JGH (15-Mar-2001) : Pass logical for box change to force routine
!>      Harald Forbert (Apr-2001): added path integral routine nvt_pimd
!>      CJM (15-Apr-2001) : added coef integrators and energy routines
!>      Joost VandeVondele (Juli-2003): simple version of isokinetic ensemble
!>      Teodoro Laino [tlaino] 10.2007 - University of Zurich: Generalization to
!>                                       different kind of thermostats
!>      Teodoro Laino [tlaino] 11.2007 - Metadynamics: now part of the MD modules
!>      Marcella Iannuzzi      02.2008 - Collecting common code (VV and creation of
!>                                       a temporary type)
!>      Teodoro Laino [tlaino] 02.2008 - Splitting integrator module and keeping in
!>                                       integrator only the INTEGRATORS
!>      Lianheng Tong [LT]     12.2013 - Added regions to Langevin MD
!> \author CJM
! **************************************************************************************************
MODULE integrator
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE barostat_types,                  ONLY: barostat_type
   USE cell_types,                      ONLY: cell_type,&
                                              init_cell,&
                                              parse_cell_line
   USE constraint,                      ONLY: rattle_control,&
                                              shake_control,&
                                              shake_roll_control,&
                                              shake_update_targets
   USE constraint_fxd,                  ONLY: fix_atom_control
   USE constraint_util,                 ONLY: getold,&
                                              pv_constraint
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_read_line
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE eigenvalueproblems,              ONLY: diagonalise
   USE extended_system_dynamics,        ONLY: shell_scale_comv
   USE extended_system_types,           ONLY: npt_info_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: ehrenfest,&
                                              npe_f_ensemble,&
                                              npe_i_ensemble
   USE integrator_utils,                ONLY: &
        allocate_old, allocate_tmp, damp_v, damp_veps, deallocate_old, get_s_ds, &
        old_variables_type, rattle_roll_setup, set, tmp_variables_type, update_dealloc_tmp, &
        update_pv, update_veps, variable_timestep, vv_first, vv_second
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: matmul_3x3,&
                                              transpose_3d
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type,&
                                              set_md_env
   USE metadynamics,                    ONLY: metadyn_integrator,&
                                              metadyn_velocities_colvar
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type,&
                                              update_particle_set
   USE physcon,                         ONLY: bohr,&
                                              femtoseconds
   USE qmmm_util,                       ONLY: apply_qmmm_walls_reflective
   USE qmmmx_update,                    ONLY: qmmmx_update_force_env
   USE qs_environment_types,            ONLY: get_qs_env
   USE reftraj_types,                   ONLY: reftraj_type
   USE reftraj_util,                    ONLY: compute_msd_reftraj
   USE rt_propagation_methods,          ONLY: propagation_step
   USE rt_propagation_output,           ONLY: rt_prop_output
   USE rt_propagation_types,            ONLY: rt_prop_type
   USE shell_opt,                       ONLY: optimize_shell_core
   USE simpar_types,                    ONLY: simpar_type
   USE thermal_region_types,            ONLY: thermal_region_type,&
                                              thermal_regions_type
   USE thermostat_methods,              ONLY: apply_thermostat_baro,&
                                              apply_thermostat_particles,&
                                              apply_thermostat_shells
   USE thermostat_types,                ONLY: thermostat_type
   USE virial_methods,                  ONLY: virial_evaluate
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'integrator'

  PUBLIC :: isokin, langevin, nve, nvt, npt_i, npt_f, nve_respa
  PUBLIC :: nph_uniaxial_damped, nph_uniaxial, nvt_adiabatic, reftraj


CONTAINS

! **************************************************************************************************
!> \brief Langevin integrator for particle positions & momenta (Brownian dynamics)
!> \param md_env ...
!> \par Literature
!>      - A. Ricci and G. Ciccotti, Mol. Phys. 101, 1927-1931 (2003)
!>      - For langevin regions:
!>        - L. Kantorovich, Phys. Rev. B 78, 094304 (2008)
!>        - L. Kantorovich and N. Rompotis, Phys. Rev. B 78, 094305 (2008)
!> \par History
!>   - Created (01.07.2005,MK)
!>   - Added support for only performing Langevin MD on a region of atoms
!>     (01.12.2013, LT)
!> \author Matthias Krack
! **************************************************************************************************
  SUBROUTINE langevin(md_env)

      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'langevin', routineP = moduleN//':'//routineN

      INTEGER :: iparticle, iparticle_kind, iparticle_local, iparticle_reg, ireg, nparticle, &
         nparticle_kind, nparticle_local, nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: do_langevin
      REAL(KIND=dp)                                      :: c, c1, c2, c3, c4, dm, dt, gam, mass, &
                                                            reg_temp, sigma
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: var_w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: pos, vel, w
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermal_region_type), POINTER                 :: thermal_region
      TYPE(thermal_regions_type), POINTER                :: thermal_regions
      TYPE(virial_type), POINTER                         :: virial

    NULLIFY(cell, para_env, gci, force_env)
    NULLIFY(atomic_kinds, local_particles,subsys,local_molecules,molecule_kinds,molecules)
    NULLIFY(molecule_kind_set,molecule_set,particles,particle_set,rng_stream,simpar,virial)
    NULLIFY(thermal_region, thermal_regions, itimes)

    CALL get_md_env(md_env=md_env, simpar=simpar, force_env=force_env,&
         para_env=para_env, thermal_regions=thermal_regions, &
         itimes=itimes)

    dt = simpar%dt
    gam = simpar%gamma+simpar%shadow_gamma
    nshell = 0

    CALL force_env_get(force_env=force_env,subsys=subsys,cell=cell)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    CALL cp_subsys_get(subsys=subsys,&
                       atomic_kinds=atomic_kinds,&
                       gci=gci,&
                       local_particles=local_particles,&
                       local_molecules_new=local_molecules,&
                       molecules_new=molecules,&
                       molecule_kinds_new=molecule_kinds,&
                       nshell=nshell,&
                       particles=particles,&
                       virial=virial)
    IF(nshell /= 0)&
       CALL cp__b("motion/integrator.F",193,"Langevin dynamics is not yet implemented for core-shell models")

    nparticle_kind = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    ! Setup the langevin regions information
    ALLOCATE(do_langevin(nparticle))
    IF (simpar%do_thermal_region) THEN
       DO iparticle = 1, nparticle
          do_langevin(iparticle) = thermal_regions%do_langevin(iparticle)
       END DO
    ELSE
       do_langevin(1:nparticle) = .TRUE.
    END IF

    ! Allocate the temperature dependent variance (var_w) of the
    ! random variable for each atom. It may be different for different
    ! atoms because of the possiblity of Langevin regions, and var_w
    ! for each region should depend on the temperature defined in the
    ! region
    ALLOCATE(var_w(nparticle))
    var_w(1:nparticle) = simpar%var_w
    IF (simpar%do_thermal_region) THEN
       DO ireg = 1, thermal_regions%nregions
          thermal_region => thermal_regions%thermal_region(ireg)
          DO iparticle_reg = 1, thermal_region%npart
             iparticle = thermal_region%part_index(iparticle_reg)
             reg_temp = thermal_region%temp_expected
             var_w(iparticle) = 2.0_dp*reg_temp*simpar%dt*(simpar%gamma + simpar%noisy_gamma)
          END DO
       END DO
    END IF

    ! Allocate work storage
    ALLOCATE (pos(3,nparticle))
    pos(:,:) = 0.0_dp

    ALLOCATE (vel(3,nparticle))
    vel(:,:) = 0.0_dp

    ALLOCATE (w(3,nparticle))
    w(:,:) = 0.0_dp

    IF (simpar%constraint) CALL getold(gci,local_molecules,molecule_set,&
         molecule_kind_set, particle_set,cell)

    ! Generate random variables
    DO iparticle_kind=1,nparticle_kind
       atomic_kind => atomic_kind_set(iparticle_kind)
       CALL get_atomic_kind(atomic_kind=atomic_kind,mass=mass)
       nparticle_local = local_particles%n_el(iparticle_kind)
       DO iparticle_local=1,nparticle_local
          iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
          IF (do_langevin(iparticle)) THEN
             sigma = var_w(iparticle)*mass
             rng_stream => local_particles%local_particle_set(iparticle_kind)%&
                   rng(iparticle_local)%stream
             w(1,iparticle) = next_random_number(rng_stream,variance=sigma)
             w(2,iparticle) = next_random_number(rng_stream,variance=sigma)
             w(3,iparticle) = next_random_number(rng_stream,variance=sigma)
          END IF
       END DO
    END DO

    DEALLOCATE(var_w)

    ! Apply fix atom constraint
    CALL fix_atom_control(force_env,w)

    ! Velocity Verlet (first part)
    c = EXP(-0.25_dp*dt*gam)
    c2 = c*c
    c4 = c2*c2
    c1 = dt*c2

    DO iparticle_kind=1,nparticle_kind
       atomic_kind => atomic_kind_set(iparticle_kind)
       CALL get_atomic_kind(atomic_kind=atomic_kind,mass=mass)
       nparticle_local = local_particles%n_el(iparticle_kind)
       dm = 0.5_dp*dt/mass
       c3 = dm/c2
       DO iparticle_local=1,nparticle_local
          iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
          IF (do_langevin(iparticle)) THEN
             vel(:,iparticle) = particle_set(iparticle)%v(:) +&
                  c3*particle_set(iparticle)%f(:)
             pos(:,iparticle) = particle_set(iparticle)%r(:) +&
                  c1*particle_set(iparticle)%v(:) +&
                  c*dm*(dt*particle_set(iparticle)%f(:) +&
                  w(:,iparticle))
          ELSE
             vel(:,iparticle) = particle_set(iparticle)%v(:) +&
                  dm*particle_set(iparticle)%f(:)
             pos(:,iparticle) = particle_set(iparticle)%r(:) +&
                  dt*particle_set(iparticle)%v(:) +&
                  dm*dt*particle_set(iparticle)%f(:)
          END IF
       END DO
    END DO

    IF (simpar%constraint) THEN
       ! Possibly update the target values
       CALL shake_update_targets(gci, local_molecules, molecule_set, &
            molecule_kind_set, dt, force_env%root_section)

       CALL shake_control(gci,local_molecules,molecule_set,molecule_kind_set,&
            particle_set,pos,vel,dt,simpar%shake_tol,&
            simpar%info_constraint,simpar%lagrange_multipliers,&
            simpar%dump_lm,cell,para_env%group,local_particles)
    END IF

    ! Broadcast the new particle positions
    CALL update_particle_set(particle_set,para_env%group,pos=pos)

    DEALLOCATE(pos)

    ! Update forces
    CALL force_env_calc_energy_force(force_env)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, vel)

    ! Update Verlet (second part)
    DO iparticle_kind=1,nparticle_kind
       atomic_kind => atomic_kind_set(iparticle_kind)
       CALL get_atomic_kind(atomic_kind=atomic_kind,mass=mass)
       dm = 0.5_dp*dt/mass
       c3 = dm/c2
       nparticle_local = local_particles%n_el(iparticle_kind)
       DO iparticle_local=1,nparticle_local
          iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
          IF (do_langevin(iparticle)) THEN
             vel(1,iparticle) = vel(1,iparticle) + c3*particle_set(iparticle)%f(1)
             vel(2,iparticle) = vel(2,iparticle) + c3*particle_set(iparticle)%f(2)
             vel(3,iparticle) = vel(3,iparticle) + c3*particle_set(iparticle)%f(3)
             vel(1,iparticle) = c4*vel(1,iparticle) + c2*w(1,iparticle)/mass
             vel(2,iparticle) = c4*vel(2,iparticle) + c2*w(2,iparticle)/mass
             vel(3,iparticle) = c4*vel(3,iparticle) + c2*w(3,iparticle)/mass
          ELSE
             vel(1,iparticle) = vel(1,iparticle) + dm*particle_set(iparticle)%f(1)
             vel(2,iparticle) = vel(2,iparticle) + dm*particle_set(iparticle)%f(2)
             vel(3,iparticle) = vel(3,iparticle) + dm*particle_set(iparticle)%f(3)
          END IF
       END DO
    END DO

    IF (simpar%constraint) THEN
       CALL rattle_control(gci,local_molecules,molecule_set,molecule_kind_set,&
            particle_set,vel,dt,simpar%shake_tol,&
            simpar%info_constraint,simpar%lagrange_multipliers,&
            simpar%dump_lm,cell,para_env%group,local_particles)
    END IF

    ! Broadcast the new particle velocities
    CALL update_particle_set(particle_set,para_env%group,vel=vel)

    DEALLOCATE(vel)

    DEALLOCATE(w)

    DEALLOCATE(do_langevin)

    ! Update virial
    IF (simpar%constraint) CALL pv_constraint(gci,local_molecules,molecule_set,&
       molecule_kind_set, particle_set,virial,para_env%group)

    CALL virial_evaluate(atomic_kind_set,particle_set,local_particles,&
         virial,para_env%group)

  END SUBROUTINE langevin

! **************************************************************************************************
!> \brief nve integrator for particle positions & momenta
!> \param md_env ...
!> \param globenv ...
!> \par History
!>   - the local particle lists are used instead of pnode (Sep. 2003,MK)
!>   - usage of fragments retrieved from the force environment (Oct. 2003,MK)
!> \author CJM
! **************************************************************************************************
  SUBROUTINE nve ( md_env, globenv)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'nve', routineP = moduleN//':'//routineN

      INTEGER                                            :: i_iter, n_iter, nparticle, &
                                                            nparticle_kind, nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: deallocate_vel, ehrenfest_md, &
                                                            shell_adiabatic, shell_check_distance, &
                                                            shell_present
      REAL(KIND=dp)                                      :: dt
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: v_old
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermostat_type), POINTER                     :: thermostat_coeff, thermostat_shell
      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(virial_type), POINTER                         :: virial

    NULLIFY (thermostat_coeff, tmp)
    NULLIFY (subsys, simpar, para_env, cell, gci, force_env, virial)
    NULLIFY (atomic_kinds, local_particles, molecules, molecule_kind_set, molecule_set, particle_set)
    NULLIFY (shell_particles, shell_particle_set, core_particles, &
            core_particle_set, thermostat_shell, dft_control, itimes)
    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env,&
         thermostat_coeff=thermostat_coeff,thermostat_shell=thermostat_shell,&
         para_env=para_env, ehrenfest_md=ehrenfest_md, itimes=itimes)
    dt = simpar%dt
    CALL force_env_get(force_env=force_env, subsys=subsys,cell=cell)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
         particles=particles,local_molecules_new=local_molecules,molecules_new=molecules, &
         molecule_kinds_new=molecule_kinds,gci=gci,virial=virial)

    nparticle_kind = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
         shell_present=shell_present,shell_adiabatic=shell_adiabatic,&
         shell_check_distance=shell_check_distance)

    IF(shell_present) THEN
       CALL cp_subsys_get(subsys=subsys,shell_particles=shell_particles,&
            core_particles=core_particles)
       shell_particle_set =>  shell_particles%els
       nshell = SIZE(shell_particles%els)

       IF(shell_adiabatic) THEN
          core_particle_set =>  core_particles%els
       END IF
    END IF

    CALL allocate_tmp(md_env, tmp, nparticle,nshell,shell_adiabatic)

    ! Apply thermostat over the full set of shells if required
    IF(shell_adiabatic) THEN
       CALL apply_thermostat_shells(thermostat_shell, atomic_kind_set, particle_set,&
            local_particles, para_env%group, shell_particle_set=shell_particle_set,&
            core_particle_set=core_particle_set)
    END IF

    IF (simpar%constraint) CALL getold(gci,local_molecules, molecule_set,  &
         molecule_kind_set, particle_set, cell)

    ! Velocity Verlet (first part)
    CALL vv_first(tmp,atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, dt)

    IF(simpar%variable_dt) CALL variable_timestep(md_env,tmp, dt, simpar, para_env, atomic_kind_set,&
         local_particles, particle_set, core_particle_set, shell_particle_set,&
         nparticle_kind, shell_adiabatic)

    IF (simpar%constraint) THEN
       ! Possibly update the target values
       CALL shake_update_targets(gci, local_molecules, molecule_set, &
           molecule_kind_set, dt, force_env%root_section)

       CALL shake_control(gci,local_molecules, molecule_set,&
           molecule_kind_set, particle_set, tmp%pos, tmp%vel, dt,simpar%shake_tol, &
           simpar%info_constraint,simpar%lagrange_multipliers,simpar%dump_lm,&
           cell,para_env%group,local_particles)
    END IF

    ! Broadcast the new particle positions and deallocate pos part of temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, pos=.TRUE.)

    IF(shell_adiabatic .AND. shell_check_distance) THEN
        CALL optimize_shell_core(force_env,  particle_set,&
             shell_particle_set, core_particle_set, globenv, tmp=tmp, check=.TRUE.)
    END IF

    ! Update forces
    ! In case of ehrenfest dynamics, velocities need to be iterated
    IF(ehrenfest_md)THEN
       ALLOCATE(v_old(3,SIZE(tmp%vel,2)))
       v_old(:,:)=tmp%vel
       CALL vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
            core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, dt)
       CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
            core_particle_set, para_env, shell_adiabatic, vel=.TRUE.,&
            should_deall_vel=.FALSE.)
       tmp%vel=v_old
       CALL get_qs_env(force_env%qs_env, dft_control=dft_control)
       n_iter=dft_control%rtp_control%max_iter
    ELSE
       n_iter=1
    END IF
    
    DO i_iter=1,n_iter
    
       IF(ehrenfest_md)THEN
          CALL get_qs_env(qs_env=force_env%qs_env,rtp=rtp)
          rtp%iter=i_iter
          tmp%vel=v_old
          CALL propagation_step(force_env%qs_env, rtp, dft_control%rtp_control)
       END IF

       ![NB] let nve work with force mixing which does not have consistent energies and forces
       CALL force_env_calc_energy_force(force_env,require_consistent_energy_force=.FALSE.)

       IF(ehrenfest_md)THEN
          CALL rt_prop_output(force_env%qs_env,ehrenfest,delta_iter=force_env%qs_env%rtp%delta_iter)
       ENDIF

       ! Metadynamics
       CALL metadyn_integrator(force_env, itimes, tmp%vel)

       ! Velocity Verlet (second part)
       CALL vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
            core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, dt)

       IF (simpar%constraint) CALL rattle_control(gci,local_molecules, molecule_set, &
            molecule_kind_set, particle_set, tmp%vel, dt,simpar%shake_tol,&
            simpar%info_constraint,simpar%lagrange_multipliers,simpar%dump_lm,&
            cell,para_env%group,local_particles)

       ! Apply thermostat over the full set of shell if required
       IF (shell_adiabatic) THEN
          CALL apply_thermostat_shells(thermostat_shell,atomic_kind_set, particle_set,&
               local_particles, para_env%group, vel=tmp%vel, &
               shell_vel=tmp%shell_vel, core_vel=tmp%core_vel)
       END IF

       IF (simpar%annealing) THEN
          tmp%vel(:,:)=tmp%vel(:,:)*simpar%f_annealing
          IF (shell_adiabatic) THEN
             CALL shell_scale_comv(atomic_kind_set,local_particles,particle_set,&
                  tmp%vel,tmp%shell_vel,tmp%core_vel)
          END IF
       END IF

       IF(ehrenfest_md) deallocate_vel=force_env%qs_env%rtp%converged
       IF(i_iter.EQ.n_iter) deallocate_vel=.TRUE.
       ! Broadcast the new particle velocities and deallocate the full temporary
       CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
            core_particle_set, para_env, shell_adiabatic, vel=.TRUE.,&
            should_deall_vel=deallocate_vel)
       IF(ehrenfest_md)THEN
          IF(force_env%qs_env%rtp%converged)EXIT
       END IF

    END DO

    ! Update virial
    IF ( simpar%constraint ) CALL pv_constraint ( gci, local_molecules, &
         molecule_set,molecule_kind_set,particle_set, virial,para_env%group )

    CALL virial_evaluate ( atomic_kind_set, particle_set,  &
         local_particles, virial, para_env%group)

  END SUBROUTINE nve

! **************************************************************************************************
!> \brief simplest version of the isokinetic gaussian thermostat
!> \param md_env ...
!> \par History
!>   - Created [2004-07]
!> \author Joost VandeVondele
!> \note
!>      - time reversible, and conserves the kinetic energy to machine precision
!>      - is not yet supposed to work for e.g. constraints, our the extended version
!>        of this thermostat
!>        see:
!>         - Zhang F. , JCP 106, 6102 (1997)
!>         - Minary P. et al, JCP 118, 2510 (2003)
! **************************************************************************************************
  SUBROUTINE isokin ( md_env)

      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'isokin', routineP = moduleN//':'//routineN

      INTEGER                                            :: nparticle, nparticle_kind, nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: shell_adiabatic, shell_present
      REAL(KIND=dp)                                      :: dt
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(tmp_variables_type), POINTER                  :: tmp

    NULLIFY(force_env, tmp, simpar, itimes)
    NULLIFY(atomic_kinds,para_env,subsys,local_particles)
    NULLIFY(core_particles,particles,shell_particles)
    NULLIFY(core_particle_set,particle_set,shell_particle_set)

    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env,&
         para_env=para_env, itimes=itimes)

    dt = simpar%dt

    CALL force_env_get(force_env=force_env,subsys=subsys)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    IF (simpar%constraint) THEN
       CALL cp__b("motion/integrator.F",631,"Constraints not yet implemented")
    END IF

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,&
         local_particles=local_particles,&
         particles=particles)

    nparticle_kind = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    nparticle = particles%n_els
    particle_set => particles%els

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
         shell_present=shell_present, shell_adiabatic=shell_adiabatic)

    IF(shell_present) THEN
       CALL cp_subsys_get(subsys=subsys,shell_particles=shell_particles,&
            core_particles=core_particles)
       shell_particle_set =>  shell_particles%els
       nshell = SIZE(shell_particles%els)

       IF(shell_adiabatic) THEN
          core_particle_set =>  core_particles%els
       END IF
    END IF

    CALL allocate_tmp (md_env,tmp, nparticle, nshell, shell_adiabatic)

    ! compute s,ds
    CALL get_s_ds(tmp, nparticle_kind, atomic_kind_set, local_particles, particle_set,&
         dt, para_env)

    ! Velocity Verlet (first part)
    tmp%scale_v(1:3) = SQRT(1.0_dp/tmp%ds)
    tmp%poly_v(1:3) = 2.0_dp*tmp%s/SQRT(tmp%ds)/dt
    CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind,&
         shell_adiabatic, dt)


    IF(simpar%variable_dt) CALL variable_timestep(md_env,tmp, dt, simpar,para_env,atomic_kind_set, &
      local_particles, particle_set, core_particle_set, shell_particle_set,&
      nparticle_kind, shell_adiabatic)

    ! Broadcast the new particle positions and deallocate the pos components of temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, pos=.TRUE.)

    CALL force_env_calc_energy_force(force_env)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, tmp%vel)

    ! compute s,ds
    CALL get_s_ds(tmp, nparticle_kind, atomic_kind_set,  local_particles, particle_set,&
       dt, para_env, tmpv=.TRUE.)

    ! Velocity Verlet (second part)
    tmp%scale_v(1:3) = SQRT(1.0_dp/tmp%ds)
    tmp%poly_v(1:3) = 2.0_dp*tmp%s/SQRT(tmp%ds)/dt
    CALL vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind,&
         shell_adiabatic, dt)

    IF (simpar%annealing)  tmp%vel(:,:)=tmp%vel(:,:)*simpar%f_annealing

    !  Broadcast the new particle velocities and deallocate the temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, vel=.TRUE.)

  END SUBROUTINE isokin
! **************************************************************************************************
!> \brief nvt adiabatic integrator for particle positions & momenta
!> \param md_env ...
!> \param globenv ...
!> \par History
!>   - the local particle lists are used instead of pnode (Sep. 2003,MK)
!>   - usage of fragments retrieved from the force environment (Oct. 2003,MK)
!> \author CJM
! **************************************************************************************************
  SUBROUTINE nvt_adiabatic ( md_env, globenv)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'nvt_adiabatic', routineP = moduleN//':'//routineN

      INTEGER                                            :: ivar, nparticle, nparticle_kind, nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: shell_adiabatic, shell_check_distance, &
                                                            shell_present
      REAL(KIND=dp)                                      :: dt
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rand
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermostat_type), POINTER                     :: thermostat_coeff, thermostat_fast, &
                                                            thermostat_shell, thermostat_slow
      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(virial_type), POINTER                         :: virial

    NULLIFY(gci, force_env, thermostat_coeff, tmp, &
         thermostat_fast, thermostat_slow, thermostat_shell, cell, shell_particles, &
         shell_particle_set, core_particles, core_particle_set, rand)
    NULLIFY(para_env, subsys, local_molecules, local_particles, molecule_kinds,&
         molecules, molecule_kind_set, molecule_set, atomic_kinds,particles)
    NULLIFY(simpar,itimes)

    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env,&
         thermostat_fast=thermostat_fast, thermostat_slow=thermostat_slow,&
         thermostat_coeff=thermostat_coeff, thermostat_shell=thermostat_shell, &
         para_env=para_env, itimes=itimes)
    dt = simpar%dt

    CALL force_env_get(force_env=force_env, subsys=subsys, cell=cell)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
         particles=particles,local_molecules_new=local_molecules,molecules_new=molecules,&
         molecule_kinds_new=molecule_kinds,gci=gci,virial=virial)

    nparticle_kind  = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle    = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
         shell_present=shell_present,shell_adiabatic=shell_adiabatic,&
         shell_check_distance=shell_check_distance)

    IF (ASSOCIATED(force_env%meta_env)) THEN
       ! Allocate random number for Langevin Thermostat acting on COLVARS
       IF(force_env%meta_env%langevin) THEN
          ALLOCATE (rand(force_env%meta_env%n_colvar))
          rand(:) = 0.0_dp
       ENDIF
    ENDIF

    !  Allocate work storage for positions and velocities
    IF(shell_present) THEN
       CALL cp_subsys_get(subsys=subsys,shell_particles=shell_particles,&
            core_particles=core_particles)
       shell_particle_set =>  shell_particles%els
       nshell = SIZE(shell_particles%els)

       IF(shell_adiabatic) THEN
          core_particle_set =>  core_particles%els
       END IF
    END IF

    CALL allocate_tmp(md_env, tmp, nparticle,nshell,shell_adiabatic)

    ! Apply Thermostat over the full set of particles
    IF(shell_adiabatic) THEN
!       CALL apply_thermostat_particles(thermostat_part, molecule_kind_set, molecule_set,&
!            particle_set, local_molecules, para_env%group, shell_adiabatic=shell_adiabatic,&
!            shell_particle_set=shell_particle_set, core_particle_set=core_particle_set)

       CALL apply_thermostat_shells(thermostat_shell, atomic_kind_set, particle_set,&
            local_particles, para_env%group, shell_particle_set=shell_particle_set,&
            core_particle_set=core_particle_set)
    ELSE
       CALL apply_thermostat_particles(thermostat_fast, force_env, molecule_kind_set, molecule_set,&
            particle_set, local_molecules, local_particles, para_env%group)

       CALL apply_thermostat_particles(thermostat_slow, force_env, molecule_kind_set, molecule_set,&
            particle_set, local_molecules, local_particles, para_env%group)
    END IF

    IF (simpar%constraint) CALL getold( gci, local_molecules, molecule_set,  &
         molecule_kind_set, particle_set, cell)

    !    *** Velocity Verlet for Langeving *** v(t)--> v(t+1/2)
    IF (ASSOCIATED(force_env%meta_env)) THEN
       IF(force_env%meta_env%langevin) THEN
          DO ivar = 1 , force_env%meta_env%n_colvar
             rng_stream => force_env%meta_env%rng(ivar)%stream
             rand(ivar)=next_random_number(rng_stream)
          ENDDO
          CALL metadyn_velocities_colvar(force_env,rand)
       ENDIF
    ENDIF

   ! Velocity Verlet (first part)
    CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, dt)

    IF(simpar%variable_dt) CALL variable_timestep(md_env,tmp, dt, simpar,para_env,atomic_kind_set,&
      local_particles, particle_set, core_particle_set, shell_particle_set,&
      nparticle_kind, shell_adiabatic)

    IF (simpar%constraint) THEN
       ! Possibly update the target values
       CALL shake_update_targets(gci, local_molecules, molecule_set, &
           molecule_kind_set, dt, force_env%root_section)

       CALL shake_control( gci, local_molecules, molecule_set,   &
           molecule_kind_set, particle_set,tmp%pos, tmp%vel, dt,simpar%shake_tol,&
           simpar%info_constraint,simpar%lagrange_multipliers,simpar%dump_lm,&
           cell, para_env%group,local_particles)
    END IF

    ! Broadcast the new particle positions and deallocate pos components of temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, pos=.TRUE.)

    IF(shell_adiabatic .AND. shell_check_distance) THEN
        CALL optimize_shell_core(force_env,  particle_set,&
             shell_particle_set, core_particle_set, globenv, tmp=tmp, check=.TRUE.)
    END IF

    ! Update forces
    CALL force_env_calc_energy_force(force_env)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, tmp%vel, rand=rand)

    ! Velocity Verlet (second part)
    CALL vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, dt)

    IF (simpar%constraint) CALL rattle_control( gci, local_molecules, molecule_set, &
         molecule_kind_set, particle_set, tmp%vel, dt,simpar%shake_tol,&
         simpar%info_constraint,simpar%lagrange_multipliers,simpar%dump_lm,&
         cell,para_env%group,local_particles)

    ! Apply Thermostat over the full set of particles
    IF (shell_adiabatic) THEN
     !  CALL apply_thermostat_particles(thermostat_part,molecule_kind_set, molecule_set, &
     !       particle_set, local_molecules, para_env%group, shell_adiabatic=shell_adiabatic,&
     !       vel= tmp%vel, shell_vel= tmp%shell_vel, core_vel= tmp%core_vel)

       CALL apply_thermostat_shells(thermostat_shell,atomic_kind_set, particle_set,&
            local_particles, para_env%group, vel= tmp%vel, shell_vel= tmp%shell_vel, &
            core_vel= tmp%core_vel)
    ELSE
       CALL apply_thermostat_particles(thermostat_slow, force_env,molecule_kind_set, molecule_set, &
            particle_set, local_molecules, local_particles, para_env%group, vel= tmp%vel)

       CALL apply_thermostat_particles(thermostat_fast, force_env,molecule_kind_set, molecule_set, &
            particle_set, local_molecules, local_particles, para_env%group, vel= tmp%vel)
    END IF

    ! Broadcast the new particle velocities and deallocate temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, vel=.TRUE.)

    IF (ASSOCIATED(force_env%meta_env)) THEN
       IF(force_env%meta_env%langevin) THEN
          DEALLOCATE(rand)
       ENDIF
    ENDIF

    ! Update constraint virial
    IF ( simpar%constraint ) CALL pv_constraint ( gci, local_molecules, &
         molecule_set,molecule_kind_set,particle_set, virial,para_env%group )

    !     **  Evaluate Virial
    CALL virial_evaluate ( atomic_kind_set, particle_set,  &
         local_particles, virial, para_env%group)

  END SUBROUTINE nvt_adiabatic


! **************************************************************************************************
!> \brief nvt integrator for particle positions & momenta
!> \param md_env ...
!> \param globenv ...
!> \par History
!>   - the local particle lists are used instead of pnode (Sep. 2003,MK)
!>   - usage of fragments retrieved from the force environment (Oct. 2003,MK)
!> \author CJM
! **************************************************************************************************
  SUBROUTINE nvt ( md_env, globenv)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'nvt', routineP = moduleN//':'//routineN

      INTEGER                                            :: ivar, nparticle, nparticle_kind, nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: shell_adiabatic, shell_check_distance, &
                                                            shell_present
      REAL(KIND=dp)                                      :: dt
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rand
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermostat_type), POINTER                     :: thermostat_coeff, thermostat_part, &
                                                            thermostat_shell
      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(virial_type), POINTER                         :: virial

    NULLIFY(gci, force_env, thermostat_coeff, tmp, &
         thermostat_part, thermostat_shell, cell, shell_particles, &
         shell_particle_set, core_particles, core_particle_set, rand)
    NULLIFY(para_env, subsys, local_molecules, local_particles, molecule_kinds,&
         molecules, molecule_kind_set, molecule_set, atomic_kinds,particles)
    NULLIFY(simpar, thermostat_coeff, thermostat_part, thermostat_shell, itimes)

    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env,&
         thermostat_part=thermostat_part, thermostat_coeff=thermostat_coeff,&
         thermostat_shell=thermostat_shell,para_env=para_env,&
         itimes=itimes)
    dt = simpar%dt

    CALL force_env_get(force_env=force_env, subsys=subsys, cell=cell)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
         particles=particles,local_molecules_new=local_molecules,molecules_new=molecules,&
         molecule_kinds_new=molecule_kinds,gci=gci,virial=virial)

    nparticle_kind  = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle    = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
         shell_present=shell_present,shell_adiabatic=shell_adiabatic,&
         shell_check_distance=shell_check_distance)

    IF (ASSOCIATED(force_env%meta_env)) THEN
       ! Allocate random number for Langevin Thermostat acting on COLVARS
       IF(force_env%meta_env%langevin) THEN
          ALLOCATE (rand(force_env%meta_env%n_colvar))
          rand(:) = 0.0_dp
       ENDIF
    ENDIF

    !  Allocate work storage for positions and velocities
    IF(shell_present) THEN
       CALL cp_subsys_get(subsys=subsys,shell_particles=shell_particles,&
            core_particles=core_particles)
       shell_particle_set =>  shell_particles%els
       nshell = SIZE(shell_particles%els)

       IF(shell_adiabatic) THEN
          core_particle_set =>  core_particles%els
       END IF
    END IF

    CALL allocate_tmp(md_env, tmp, nparticle,nshell,shell_adiabatic)

    ! Apply Thermostat over the full set of particles
    IF(shell_adiabatic) THEN
       CALL apply_thermostat_particles(thermostat_part, force_env, molecule_kind_set, molecule_set,&
            particle_set, local_molecules, local_particles, para_env%group, shell_adiabatic=shell_adiabatic,&
            shell_particle_set=shell_particle_set, core_particle_set=core_particle_set)

       CALL apply_thermostat_shells(thermostat_shell, atomic_kind_set, particle_set,&
            local_particles, para_env%group, shell_particle_set=shell_particle_set,&
            core_particle_set=core_particle_set)
    ELSE
       CALL apply_thermostat_particles(thermostat_part, force_env, molecule_kind_set, molecule_set,&
            particle_set, local_molecules, local_particles, para_env%group)
    END IF

    IF (simpar%constraint) CALL getold( gci, local_molecules, molecule_set,  &
         molecule_kind_set, particle_set, cell)

    !    *** Velocity Verlet for Langeving *** v(t)--> v(t+1/2)
    IF (ASSOCIATED(force_env%meta_env)) THEN
       IF(force_env%meta_env%langevin) THEN
          DO ivar = 1 , force_env%meta_env%n_colvar
             rng_stream => force_env%meta_env%rng(ivar)%stream
             rand(ivar)=next_random_number(rng_stream)
          ENDDO
          CALL metadyn_velocities_colvar(force_env,rand)
       ENDIF
    ENDIF

   ! Velocity Verlet (first part)
    CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, dt)

    IF(simpar%variable_dt) CALL variable_timestep(md_env,tmp, dt, simpar,para_env,atomic_kind_set,&
      local_particles, particle_set, core_particle_set, shell_particle_set,&
      nparticle_kind, shell_adiabatic)

    IF (simpar%constraint) THEN
       ! Possibly update the target values
       CALL shake_update_targets(gci, local_molecules, molecule_set, &
           molecule_kind_set, dt, force_env%root_section)

       CALL shake_control( gci, local_molecules, molecule_set,   &
           molecule_kind_set, particle_set,tmp%pos, tmp%vel, dt,simpar%shake_tol,&
           simpar%info_constraint,simpar%lagrange_multipliers,simpar%dump_lm,&
           cell, para_env%group,local_particles)
    END IF

    ! Broadcast the new particle positions and deallocate pos components of temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, pos=.TRUE.)

    IF(shell_adiabatic .AND. shell_check_distance) THEN
        CALL optimize_shell_core(force_env,  particle_set,&
             shell_particle_set, core_particle_set, globenv, tmp=tmp, check=.TRUE.)
    END IF

    ![ADAPT] update input structure with new coordinates, make new labels
    CALL qmmmx_update_force_env(force_env, force_env%root_section)

    ![NB] recreate pointers changed by creation of new subsys in qmmm_update_force_mixing_env
    ![NB] ugly hack, which is why adaptivity isn't implemented in most other ensembles
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    CALL force_env_get(force_env=force_env, subsys=subsys, cell=cell)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
       particles=particles,local_molecules_new=local_molecules,molecules_new=molecules,&
       molecule_kinds_new=molecule_kinds,gci=gci,virial=virial)

    nparticle_kind  = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle    = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
       shell_present=shell_present,shell_adiabatic=shell_adiabatic,&
       shell_check_distance=shell_check_distance)

    !  Allocate work storage for positions and velocities
    IF(shell_present) THEN
     CALL cp_subsys_get(subsys=subsys,shell_particles=shell_particles,&
          core_particles=core_particles)
     shell_particle_set =>  shell_particles%els
     nshell = SIZE(shell_particles%els)

     IF(shell_adiabatic) THEN
        core_particle_set =>  core_particles%els
     END IF
    END IF
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! Update forces
    ![NB] let nvt work with force mixing which does not have consistent energies and forces
    CALL force_env_calc_energy_force(force_env,require_consistent_energy_force=.FALSE.)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, tmp%vel, rand=rand)

    ! Velocity Verlet (second part)
    CALL vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind, shell_adiabatic, dt)

    IF (simpar%constraint) CALL rattle_control( gci, local_molecules, molecule_set, &
         molecule_kind_set, particle_set, tmp%vel, dt,simpar%shake_tol,&
         simpar%info_constraint,simpar%lagrange_multipliers,simpar%dump_lm,&
         cell,para_env%group,local_particles)

    ! Apply Thermostat over the full set of particles
    IF (shell_adiabatic) THEN
       CALL apply_thermostat_particles(thermostat_part, force_env,molecule_kind_set, molecule_set, &
            particle_set, local_molecules, local_particles, para_env%group, shell_adiabatic=shell_adiabatic,&
            vel= tmp%vel, shell_vel= tmp%shell_vel, core_vel= tmp%core_vel)

       CALL apply_thermostat_shells(thermostat_shell,atomic_kind_set, particle_set,&
            local_particles, para_env%group, vel= tmp%vel, shell_vel= tmp%shell_vel, &
            core_vel= tmp%core_vel)
    ELSE
       CALL apply_thermostat_particles(thermostat_part, force_env,molecule_kind_set, molecule_set, &
            particle_set, local_molecules, local_particles, para_env%group, vel= tmp%vel)
    END IF

    ! Broadcast the new particle velocities and deallocate temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, vel=.TRUE.)

    IF (ASSOCIATED(force_env%meta_env)) THEN
       IF(force_env%meta_env%langevin) THEN
          DEALLOCATE(rand)
       ENDIF
    ENDIF

    ! Update constraint virial
    IF ( simpar%constraint ) CALL pv_constraint ( gci, local_molecules, &
         molecule_set,molecule_kind_set,particle_set, virial,para_env%group )

    !     **  Evaluate Virial
    CALL virial_evaluate ( atomic_kind_set, particle_set,  &
         local_particles, virial, para_env%group)

  END SUBROUTINE nvt

! **************************************************************************************************
!> \brief npt_i integrator for particle positions & momenta
!>      isotropic box changes
!> \param md_env ...
!> \param globenv ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
  SUBROUTINE npt_i ( md_env, globenv)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'npt_i', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: e2 = 1.0_dp/6.0_dp, e4 = e2/20.0_dp, &
                                                            e6 = e4/42.0_dp, e8 = e6/72.0_dp

      INTEGER                                            :: iroll, ivar, nparticle, nparticle_kind, &
                                                            nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: first, first_time, shell_adiabatic, &
                                                            shell_check_distance, shell_present
      REAL(KIND=dp)                                      :: dt, infree, kin, roll_tol, roll_tol_thrs
      REAL(KIND=dp), DIMENSION(3)                        :: vector_r, vector_v
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_kin
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rand
      REAL(KIND=dp), SAVE                                :: eps_0
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(npt_info_type), POINTER                       :: npt(:,:)
      TYPE(old_variables_type), POINTER                  :: old
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermostat_type), POINTER                     :: thermostat_baro, thermostat_part, &
                                                            thermostat_shell
      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(virial_type), POINTER                         :: virial

    NULLIFY(gci,thermostat_baro,thermostat_part,thermostat_shell,force_env)
    NULLIFY(atomic_kinds,cell,para_env,subsys,local_molecules,local_particles)
    NULLIFY(molecule_kinds,molecules,molecule_kind_set,npt)
    NULLIFY(core_particles,particles,shell_particles, tmp, old)
    NULLIFY(core_particle_set,particle_set,shell_particle_set)
    NULLIFY(simpar,virial,rand,itimes)

    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env,&
         thermostat_part=thermostat_part, thermostat_baro=thermostat_baro,&
         thermostat_shell=thermostat_shell, npt=npt, first_time=first_time,&
         para_env=para_env, itimes=itimes)
    dt = simpar%dt
    infree = 1.0_dp / REAL ( simpar%nfree,KIND=dp)

    CALL force_env_get(force_env=force_env, subsys=subsys, cell=cell)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
         particles=particles,local_molecules_new=local_molecules,molecules_new=molecules, &
         gci=gci,molecule_kinds_new=molecule_kinds,virial=virial)

    nparticle_kind = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
         shell_present=shell_present,shell_adiabatic=shell_adiabatic,&
         shell_check_distance=shell_check_distance)

    IF ( first_time ) THEN
       CALL virial_evaluate ( atomic_kind_set, particle_set,  &
            local_particles, virial, para_env%group)
    END IF

    ! Allocate work storage for positions and velocities
    CALL allocate_old ( old, particle_set, npt)

    IF (ASSOCIATED(force_env%meta_env)) THEN
       ! Allocate random number for Langevin Thermostat acting on COLVARS
       IF(force_env%meta_env%langevin) THEN
          ALLOCATE (rand(force_env%meta_env%n_colvar))
          rand(:) = 0.0_dp
       ENDIF
    ENDIF

    IF(shell_present) THEN
       CALL cp_subsys_get(subsys=subsys,&
            shell_particles=shell_particles, core_particles=core_particles)
       shell_particle_set =>  shell_particles%els
       nshell = SIZE(shell_particles%els)
       IF(shell_adiabatic) THEN
          core_particle_set =>  core_particles%els
       END IF
    END IF

    CALL allocate_tmp(md_env, tmp, nparticle, nshell, shell_adiabatic)

    ! Initialize eps_0 the first time through
    IF ( first_time ) eps_0 = npt (1,1)%eps

    ! Apply thermostat to  barostat
    CALL apply_thermostat_baro( thermostat_baro, npt, para_env%group)

    ! Apply Thermostat over the full set of particles
    IF(simpar% ensemble /= npe_i_ensemble) THEN
      IF(shell_adiabatic) THEN
        CALL apply_thermostat_particles(thermostat_part, force_env, molecule_kind_set, molecule_set,&
             particle_set, local_molecules, local_particles, para_env%group,shell_adiabatic=shell_adiabatic,&
             shell_particle_set=shell_particle_set, core_particle_set=core_particle_set)

      ELSE
        CALL apply_thermostat_particles(thermostat_part, force_env, molecule_kind_set, molecule_set,&
             particle_set, local_molecules, local_particles, para_env%group)
      END IF
    END IF

    ! Apply Thermostat over the core-shell motion
    CALL apply_thermostat_shells(thermostat_shell, atomic_kind_set, particle_set,&
         local_particles, para_env%group, shell_particle_set=shell_particle_set,&
         core_particle_set=core_particle_set)

    IF (simpar%constraint) THEN
       ! Possibly update the target values
       CALL shake_update_targets(gci, local_molecules, molecule_set, &
            molecule_kind_set, dt, force_env%root_section)
    END IF

    ! setting up for ROLL: saving old variables
    IF (simpar%constraint) THEN
       roll_tol_thrs = simpar%roll_tol
       iroll = 1
       CALL set ( old, atomic_kind_set, particle_set, local_particles, cell, npt, 'F' )
       CALL getold ( gci, local_molecules, molecule_set,  &
            molecule_kind_set, particle_set, cell)
    ELSE
       roll_tol_thrs = EPSILON(0.0_dp)
    ENDIF
    roll_tol = -roll_tol_thrs

    !    *** Velocity Verlet for Langeving *** v(t)--> v(t+1/2)
    IF (ASSOCIATED(force_env%meta_env)) THEN
       IF(force_env%meta_env%langevin) THEN
          DO ivar = 1 , force_env%meta_env%n_colvar
             rng_stream => force_env%meta_env%rng(ivar)%stream
             rand(ivar)=next_random_number(rng_stream)
          ENDDO
          CALL metadyn_velocities_colvar(force_env,rand)
       ENDIF
    ENDIF

    SR: DO WHILE (ABS(roll_tol)>=roll_tol_thrs) ! SHAKE-ROLL LOOP

       IF (simpar%constraint) THEN
          CALL set ( old, atomic_kind_set, particle_set, local_particles, cell, npt, 'B' )
       END IF

       CALL update_pv ( gci, simpar, atomic_kind_set, particle_set, &
            local_molecules, molecule_set, molecule_kind_set, &
            local_particles, kin, pv_kin, virial, para_env%group )
       CALL update_veps ( cell, npt, simpar, pv_kin, kin, virial, infree )

       tmp%arg_r(1) = ( 0.5_dp * npt ( 1, 1 )%v * dt ) * &
            ( 0.5_dp * npt ( 1, 1 )%v * dt )
       tmp%poly_r(1:3) = 1.0_dp + e2*tmp%arg_r(1) + e4*tmp%arg_r(1)*tmp%arg_r(1) + &
            e6*tmp%arg_r(1)**3 + e8*tmp%arg_r(1)**4

       tmp%arg_v(1) = ( 0.25_dp * npt ( 1, 1 )%v * dt *  &
            (1.0_dp + 3.0_dp * infree ) )*( 0.25_dp * npt ( 1, 1 )%v * &
            dt * ( 1.0_dp + 3.0_dp * infree ) )
       tmp%poly_v(1:3) = 1.0_dp + e2*tmp%arg_v(1) + e4*tmp%arg_v(1)*tmp%arg_v(1) + &
            e6*tmp%arg_v(1)**3 + e8*tmp%arg_v(1)**4

       tmp%scale_r(1:3) = EXP( 0.5_dp * dt * npt ( 1, 1 )%v )
       tmp%scale_v(1:3) = EXP( -0.25_dp * dt * npt ( 1, 1 )%v * &
            (1.0_dp + 3.0_dp * infree ) )

       ! first half of velocity verlet
       CALL vv_first( tmp, atomic_kind_set, local_particles, particle_set, &
            core_particle_set, shell_particle_set, nparticle_kind,&
            shell_adiabatic, dt)

       IF(simpar%variable_dt) CALL  variable_timestep(md_env,tmp, dt, simpar,para_env,&
            atomic_kind_set, local_particles, particle_set, core_particle_set,&
            shell_particle_set, nparticle_kind, shell_adiabatic, npt=npt)


       roll_tol = 0.0_dp
       vector_r ( : ) = tmp%scale_r(:) * tmp%poly_r(:)
       vector_v ( : ) = tmp%scale_v(:) * tmp%poly_v(:)

       IF ( simpar%constraint ) CALL shake_roll_control( gci, local_molecules, &
            molecule_set, molecule_kind_set, particle_set, tmp%pos, tmp%vel, dt, simpar,  &
            roll_tol, iroll, vector_r, vector_v, para_env%group, cell=cell, &
            local_particles=local_particles)
    END DO SR

    ! Update eps:
    npt ( :, : )%eps =  npt ( :, : )%eps + dt * npt ( :, : )%v

    ! Update h_mat
    cell%hmat ( :, : ) = cell%hmat ( :, : ) * EXP ( npt ( 1, 1 )%eps - eps_0 )

    eps_0 =  npt ( 1, 1 )%eps

    ! Update the inverse
    CALL init_cell ( cell )

    ! Broadcast the new particle positions and deallocate the pos components of temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, pos=.TRUE.)

    IF(shell_adiabatic .AND. shell_check_distance) THEN
        CALL optimize_shell_core(force_env,  particle_set,&
             shell_particle_set, core_particle_set, globenv, tmp=tmp, check=.TRUE.)
    END IF

    ! Update forces
    CALL force_env_calc_energy_force(force_env)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, tmp%vel, rand=rand)

    ! Velocity Verlet (second part)
    CALL vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind,&
         shell_adiabatic, dt)

    IF (simpar%constraint) THEN
       roll_tol_thrs = simpar%roll_tol
       first = .TRUE.
       iroll = 1
       CALL set ( old, atomic_kind_set, particle_set, tmp%vel, local_particles, cell, npt, 'F' )
    ELSE
       roll_tol_thrs = EPSILON(0.0_dp)
    ENDIF
    roll_tol = -roll_tol_thrs

    RR: DO WHILE (ABS(roll_tol)>=roll_tol_thrs) ! RATTLE-ROLL LOOP
       roll_tol = 0.0_dp
       IF ( simpar%constraint ) CALL rattle_roll_setup ( old, gci, atomic_kind_set, &
            particle_set, local_particles, molecule_kind_set, molecule_set,  &
            local_molecules, tmp%vel, dt,  cell, npt, simpar, virial, vector_v,  &
            roll_tol, iroll, infree, first, para_env)

       CALL update_pv ( gci, simpar, atomic_kind_set, tmp%vel, particle_set, &
            local_molecules, molecule_set, molecule_kind_set, &
            local_particles, kin, pv_kin, virial, para_env%group )
       CALL update_veps ( cell, npt, simpar, pv_kin, kin, virial, infree )
    END DO RR

    ! Apply Thermostat over the full set of particles
    IF(simpar% ensemble /= npe_i_ensemble) THEN
      IF (shell_adiabatic) THEN
        CALL apply_thermostat_particles(thermostat_part, force_env,molecule_kind_set, molecule_set, &
             particle_set, local_molecules, local_particles, para_env%group, shell_adiabatic=shell_adiabatic,&
             vel=tmp%vel, shell_vel=tmp%shell_vel, core_vel=tmp%core_vel)
      ELSE
        CALL apply_thermostat_particles(thermostat_part, force_env,molecule_kind_set, molecule_set, &
             particle_set, local_molecules, local_particles, para_env%group, vel=tmp%vel)
      END IF
    END IF


    ! Apply Thermostat over the core-shell motion
    IF(ASSOCIATED(thermostat_shell)) THEN
      CALL apply_thermostat_shells(thermostat_shell,atomic_kind_set, particle_set,&
           local_particles, para_env%group, vel=tmp%vel, shell_vel=tmp%shell_vel, &
           core_vel=tmp%core_vel)
    END IF

    ! Apply Thermostat to Barostat
    CALL apply_thermostat_baro( thermostat_baro, npt, para_env%group)

    ! Annealing of particle velocities is only possible when no thermostat is active
    IF (simpar% ensemble == npe_i_ensemble .AND. simpar%annealing) THEN
       tmp%vel(:,:)=tmp%vel(:,:)*simpar%f_annealing
       IF (shell_adiabatic) THEN
          CALL shell_scale_comv(atomic_kind_set,local_particles,particle_set,&
               tmp%vel,tmp%shell_vel,tmp%core_vel)
       END IF
    END IF
    ! Annealing of CELL velocities is only possible when no thermostat is active
    IF (simpar% ensemble == npe_i_ensemble .AND. simpar%annealing_cell) THEN
       npt(1,1)%v = npt(1,1)%v * simpar%f_annealing_cell
    END IF


    ! Broadcast the new particle velocities and deallocate temporary
    CALL update_dealloc_tmp ( tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, vel=.TRUE.)

    ! Update constraint virial
    IF ( simpar%constraint ) CALL pv_constraint ( gci, local_molecules, &
         molecule_set, molecule_kind_set, particle_set, virial, para_env%group )

    CALL virial_evaluate ( atomic_kind_set, particle_set,  &
         local_particles, virial, para_env%group)

    ! Deallocate old variables
    CALL deallocate_old (old)

    IF (ASSOCIATED(force_env%meta_env)) THEN
       IF(force_env%meta_env%langevin) THEN
          DEALLOCATE(rand)
       ENDIF
    ENDIF

    IF (first_time) THEN
       first_time = .FALSE.
       CALL set_md_env(md_env, first_time=first_time)
    END IF

  END SUBROUTINE npt_i

! **************************************************************************************************
!> \brief uses coordinates in a file and generates frame after frame of these
!> \param md_env ...
!> \par History
!>   - 04.2005 created [Joost VandeVondele]
!>   - modified to make it more general [MI]
!> \note
!>     it can be used to compute some properties on already available trajectories
! **************************************************************************************************
  SUBROUTINE reftraj ( md_env)
      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'reftraj', routineP = moduleN//':'//routineN

      CHARACTER(LEN=10)                                  :: AA
      INTEGER                                            :: cell_itimes, I, nparticle, Nread, &
                                                            trj_itimes
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: init, my_end, test_ok
      REAL(KIND=dp)                                      :: cell_time, h(3,3), trj_epot, trj_time, &
                                                            vol
      REAL(KIND=dp), POINTER                             :: time
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(reftraj_type), POINTER                        :: reftraj_env
      TYPE(simpar_type), POINTER                         :: simpar

    NULLIFY(reftraj_env, particle_set, particles, force_env, subsys, simpar, para_env, cell, itimes, time)
    CALL get_md_env(md_env=md_env, init=init, reftraj=reftraj_env, force_env=force_env,&
         para_env=para_env, simpar=simpar)

    CALL force_env_get(force_env=force_env, cell=cell, subsys=subsys)
    reftraj_env%isnap = reftraj_env%isnap + reftraj_env%info%stride

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)
    CALL cp_subsys_get(subsys=subsys, particles=particles)
    nparticle    = particles%n_els
    particle_set => particles%els

    ! SnapShots read from an external file (parsers calls are buffered! please
    ! don't put any additional MPI call!) [tlaino]
    CALL parser_read_line(reftraj_env%info%traj_parser,1)
    READ(reftraj_env%info%traj_parser%input_line,FMT="(I8)") nread
    IF(.NOT.(nread==nparticle))CALL cp__a("motion/integrator.F",1540)
    CALL parser_read_line(reftraj_env%info%traj_parser,1)
    test_ok = .FALSE.
    READ(reftraj_env%info%traj_parser%input_line,FMT="(T6,I8,T23,F12.3,T41,F20.10)",ERR=999)&
         trj_itimes, trj_time, trj_epot
    test_ok = .TRUE.
999 IF (.NOT.test_ok) THEN
       ! Handling properly the error when reading the title of an XYZ
       CALL get_md_env(md_env, itimes=itimes)
       trj_itimes = itimes
       trj_time   = 0.0_dp
       trj_epot   = 0.0_dp
    END IF
    DO i = 1,nread-1
       CALL parser_read_line(reftraj_env%info%traj_parser,1)
       READ(reftraj_env%info%traj_parser%input_line(1:LEN_TRIM(reftraj_env%info%traj_parser%input_line)),*) AA,particle_set(i)%r
       particle_set(i)%r = particle_set(i)%r*bohr
    END DO
    ! End of file is properly addressed in the previous call..
    ! Let's check directly (providing some info) also for the last
    ! line of this frame..
    CALL parser_read_line(reftraj_env%info%traj_parser,1,at_end=my_end)
    READ(unit=reftraj_env%info%traj_parser%input_line,fmt=*) AA,particle_set(i)%r
    particle_set(i)%r(1) = cp_unit_to_cp2k(particle_set(i)%r(1),"angstrom")
    particle_set(i)%r(2) = cp_unit_to_cp2k(particle_set(i)%r(2),"angstrom")
    particle_set(i)%r(3) = cp_unit_to_cp2k(particle_set(i)%r(3),"angstrom")

    ! Check if we reached the end of the file and provide some info..
    IF (my_end) THEN
       IF(reftraj_env%isnap/=(simpar%nsteps-1))&
          CALL cp_abort(cp__l("motion/integrator.F",1570),&
               "Reached the end of the Trajectory  frames in the TRAJECTORY file. Number of "//&
               "missing frames ("//cp_to_string((simpar%nsteps-1)-reftraj_env%isnap)//").")
    END IF

    IF(reftraj_env%info%variable_volume)THEN
       CALL parser_get_next_line(reftraj_env%info%cell_parser,1,at_end=my_end)
       CALL parse_cell_line(reftraj_env%info%cell_parser%input_line, cell_itimes, cell_time, h, vol)
       IF(.NOT.(trj_itimes==cell_itimes))CALL cp__a("motion/integrator.F",1578)
       ! Check if we reached the end of the file and provide some info..
       IF (my_end) THEN
          IF(reftraj_env%isnap/=(simpar%nsteps-1))&
             CALL cp_abort(cp__l("motion/integrator.F",1582),&
                  "Reached the end of the cell info frames in the CELL file. Number of "//&
                  "missing frames ("//cp_to_string((simpar%nsteps-1)-reftraj_env%isnap)//").")
       END IF
    END IF

    IF (init) THEN
       reftraj_env%time0   = trj_time
       reftraj_env%epot0   = trj_epot
       reftraj_env%itimes0 = trj_itimes
    END IF

    IF (trj_itimes/=0.0_dp.AND.trj_time/=0.0_dp) simpar%dt = (trj_time/femtoseconds)/REAL(trj_itimes,KIND=dp)

    reftraj_env%epot   = trj_epot
    reftraj_env%itimes = trj_itimes
    reftraj_env%time   = trj_time/femtoseconds
    CALL get_md_env(md_env, t=time)
    time = reftraj_env%time


    IF(reftraj_env%info%variable_volume)THEN
       cell%hmat = h
       CALL init_cell(cell)
    END IF

    ![ADAPT] update input structure with new coordinates, make new labels
    CALL qmmmx_update_force_env(force_env, force_env%root_section)
    ! no pointers into force_env%subsys to update

    ! Task to perform on the reference trajectory
    ! Compute energy and forces
    ![NB] let reftraj work with force mixing which does not have consistent energies and forces
    CALL force_env_calc_energy_force(force_env,eval_energy_forces=reftraj_env%info%eval_EF,&
         require_consistent_energy_force=.FALSE.)

    ! Metadynamics
    CALL metadyn_integrator(force_env, trj_itimes)

    ! Compute MSD with respect to a reference configuration
    IF(reftraj_env%info%msd) THEN
       CALL compute_msd_reftraj(reftraj_env,md_env, particle_set)
    END IF

    ! Skip according the stride both Trajectory and Cell (if possible)
    CALL parser_get_next_line(reftraj_env%info%traj_parser,(reftraj_env%info%stride-1)*(nparticle+2))
    IF(reftraj_env%info%variable_volume)THEN
       CALL parser_get_next_line(reftraj_env%info%cell_parser,(reftraj_env%info%stride-1))
    END IF
  END SUBROUTINE reftraj

! **************************************************************************************************
!> \brief nph_uniaxial integrator (non-Hamiltonian version)
!>      for particle positions & momenta undergoing
!>      uniaxial stress ( in x-direction of orthorhombic cell)
!>      due to a shock compression:
!>      Reed et. al. Physical Review Letters 90, 235503 (2003).
!> \param md_env ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
  SUBROUTINE nph_uniaxial ( md_env)

      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'nph_uniaxial', routineP = moduleN//':'//routineN
      REAL(dp), PARAMETER                                :: e2 = 1._dp/6._dp, e4 = e2/20._dp, &
                                                            e6 = e4/42._dp, e8 = e6/72._dp

      INTEGER                                            :: iroll, nparticle, nparticle_kind, nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: first, first_time, shell_adiabatic, &
                                                            shell_present
      REAL(KIND=dp)                                      :: dt, infree, kin, roll_tol, roll_tol_thrs
      REAL(KIND=dp), DIMENSION(3)                        :: vector_r, vector_v
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_kin
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(npt_info_type), POINTER                       :: npt( :, : )
      TYPE(old_variables_type), POINTER                  :: old
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(virial_type), POINTER                         :: virial

    NULLIFY(gci, force_env)
    NULLIFY(atomic_kinds,cell,para_env,subsys,local_molecules,local_particles)
    NULLIFY(molecule_kinds,molecules,molecule_kind_set,npt)
    NULLIFY(core_particles,particles,shell_particles, tmp, old)
    NULLIFY(core_particle_set,particle_set,shell_particle_set)
    NULLIFY(simpar,virial,itimes)

    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env,npt = npt,&
         first_time=first_time, para_env=para_env, itimes=itimes)
    dt = simpar%dt
    infree = 1.0_dp / REAL ( simpar%nfree, dp )

    CALL force_env_get(force_env, subsys=subsys, cell=cell)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
         particles=particles,local_molecules_new=local_molecules, molecules_new=molecules, gci=gci,&
         molecule_kinds_new=molecule_kinds,virial=virial)

    nparticle_kind = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    IF ( first_time ) THEN
       CALL virial_evaluate ( atomic_kind_set, particle_set,  &
            local_particles, virial, para_env%group)
    END IF

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
         shell_present=shell_present,shell_adiabatic=shell_adiabatic)

    ! Allocate work storage for positions and velocities
    CALL allocate_old ( old, particle_set, npt)

    IF(shell_present) THEN
       CALL cp_subsys_get(subsys=subsys,&
            shell_particles=shell_particles, core_particles=core_particles)
       shell_particle_set =>  shell_particles%els
       nshell = SIZE(shell_particles%els)
       IF(shell_adiabatic) THEN
          core_particle_set =>  core_particles%els
       END IF
    END IF

    CALL allocate_tmp(md_env, tmp,nparticle, nshell, shell_adiabatic)

    IF (simpar%constraint) THEN
       ! Possibly update the target values
       CALL shake_update_targets(gci, local_molecules, molecule_set, &
            molecule_kind_set, dt, force_env%root_section)
    END IF

    ! setting up for ROLL: saving old variables
    IF (simpar%constraint) THEN
       roll_tol_thrs = simpar%roll_tol
       iroll = 1
       CALL set ( old, atomic_kind_set, particle_set, local_particles, cell, npt, 'F' )
       CALL getold ( gci, local_molecules, molecule_set,  &
            molecule_kind_set, particle_set, cell)
    ELSE
       roll_tol_thrs = EPSILON(0.0_dp)
    ENDIF
    roll_tol = -roll_tol_thrs

    SR: DO WHILE (ABS(roll_tol)>=roll_tol_thrs) ! SHAKE-ROLL LOOP

       IF (simpar%constraint) THEN
          CALL set ( old, atomic_kind_set, particle_set, local_particles, cell, npt, 'B' )
       END IF
       CALL update_pv ( gci, simpar, atomic_kind_set, particle_set, &
            local_molecules, molecule_set, molecule_kind_set, &
            local_particles, kin, pv_kin, virial, para_env%group )
       CALL update_veps ( cell, npt, simpar, pv_kin, kin, virial, infree )

       tmp%arg_r(1) = ( 0.5_dp * npt ( 1, 1 )%v * dt ) * &
            ( 0.5_dp * npt ( 1, 1 )%v * dt )
       tmp%poly_r(1) = 1._dp + e2*tmp%arg_r(1) + e4*tmp%arg_r(1)*tmp%arg_r(1) + &
           e6*tmp%arg_r(1)**3 + e8*tmp%arg_r(1)**4
       tmp%poly_r(2) = 1.0_dp
       tmp%poly_r(3) = 1.0_dp

       tmp%arg_v(1) = ( 0.25_dp * npt ( 1, 1 )%v * dt *  &
            (1._dp + infree ) )*( 0.25_dp * npt ( 1, 1 )%v * &
            dt * ( 1._dp + infree ) )
       tmp%arg_v(2) = ( 0.25_dp * npt ( 1, 1 )%v * dt * infree) * &
            ( 0.25_dp * npt ( 1, 1 )%v * dt * infree )
       tmp%poly_v(1) = 1._dp + e2*tmp%arg_v(1) + e4*tmp%arg_v(1)*tmp%arg_v(1) + &
            e6*tmp%arg_v(1)**3 + e8*tmp%arg_v(1)**4
       tmp%poly_v(2) = 1._dp + e2*tmp%arg_v(2) + e4*tmp%arg_v(2)*tmp%arg_v(2) + &
            e6*tmp%arg_v(2)**3 + e8*tmp%arg_v(2)**4
       tmp%poly_v(3) = 1._dp + e2*tmp%arg_v(2) + e4*tmp%arg_v(2)*tmp%arg_v(2) + &
            e6*tmp%arg_v(2)**3 + e8*tmp%arg_v(2)**4

       tmp%scale_r(1) = EXP( 0.5_dp * dt * npt ( 1, 1 )%v )
       tmp%scale_r(2) = 1.0_dp
       tmp%scale_r(3) = 1.0_dp

       tmp%scale_v(1) = EXP( -0.25_dp * dt * npt ( 1, 1 )%v * &
            (1._dp + infree ) )
       tmp%scale_v(2) = EXP( -0.25_dp * dt * npt ( 1, 1 )%v * infree )
       tmp%scale_v(3) = EXP( -0.25_dp * dt * npt ( 1, 1 )%v * infree )

       ! first half of velocity verlet
       CALL vv_first( tmp, atomic_kind_set, local_particles, particle_set, &
            core_particle_set, shell_particle_set, nparticle_kind,&
            shell_adiabatic, dt)

       IF(simpar%variable_dt) CALL variable_timestep(md_env,tmp, dt, simpar,para_env,&
          atomic_kind_set, local_particles, particle_set, core_particle_set,&
          shell_particle_set, nparticle_kind, shell_adiabatic, npt=npt)


       roll_tol = 0._dp
       vector_r ( : ) = 0._dp
       vector_v ( : ) = tmp%scale_v(:) * tmp%poly_v(:)
       vector_r ( 1 ) = tmp%scale_r(1) * tmp%poly_r(1)

       IF ( simpar%constraint ) CALL shake_roll_control( gci, local_molecules, &
            molecule_set, molecule_kind_set, particle_set, tmp%pos, tmp%vel, dt, simpar,&
            roll_tol, iroll, vector_r, vector_v, para_env%group, cell=cell, &
            local_particles=local_particles)
    END DO SR

    ! Update h_mat
    cell%hmat ( 1, 1 ) = cell%hmat ( 1, 1 ) * tmp%scale_r(1) * tmp%scale_r(1)

    ! Update the cell
    CALL init_cell ( cell )

    ! Broadcast the new particle positions and deallocate the pos component of temporary
    CALL update_dealloc_tmp (tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, pos=.TRUE.)

    ! Update forces (and stress)
    CALL force_env_calc_energy_force(force_env)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, tmp%vel)

    ! Velocity Verlet (second part)
     CALL vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
          core_particle_set, shell_particle_set, nparticle_kind,&
          shell_adiabatic, dt)

    IF (simpar%constraint) THEN
       roll_tol_thrs = simpar%roll_tol
       first = .TRUE.
       iroll = 1
       CALL set ( old, atomic_kind_set, particle_set, tmp%vel, local_particles, cell, npt, 'F' )
    ELSE
       roll_tol_thrs = EPSILON(0.0_dp)
    ENDIF
    roll_tol = -roll_tol_thrs

    RR: DO WHILE (ABS(roll_tol)>=roll_tol_thrs) ! RATTLE-ROLL LOOP
       roll_tol = 0._dp
       IF ( simpar%constraint ) CALL rattle_roll_setup ( old, gci, atomic_kind_set, &
            particle_set, local_particles, molecule_kind_set, molecule_set,  &
            local_molecules, tmp%vel, dt, cell, npt, simpar, virial, vector_v,  &
            roll_tol, iroll, infree, first, para_env)

       CALL update_pv ( gci, simpar, atomic_kind_set, tmp%vel, particle_set, &
            local_molecules, molecule_set, molecule_kind_set, &
            local_particles, kin, pv_kin, virial, para_env%group )
       CALL update_veps ( cell, npt, simpar, pv_kin, kin, virial, infree )
    END DO RR

    IF (simpar%annealing)  tmp%vel(:,:)=tmp%vel(:,:)*simpar%f_annealing

    ! Broadcast the new particle velocities and deallocate the temporary
    CALL update_dealloc_tmp (tmp, particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, vel=.TRUE.)

    ! Update constraint virial
    IF ( simpar%constraint ) CALL pv_constraint ( gci, local_molecules, &
         molecule_set,molecule_kind_set,particle_set, virial, para_env%group )

    CALL virial_evaluate ( atomic_kind_set, particle_set,  &
         local_particles, virial, para_env%group)

    ! Deallocate old variables
    CALL deallocate_old ( old)

    IF (first_time) THEN
       first_time = .FALSE.
       CALL set_md_env(md_env, first_time=first_time)
    END IF

  END SUBROUTINE nph_uniaxial

! **************************************************************************************************
!> \brief nph_uniaxial integrator (non-Hamiltonian version)
!>      for particle positions & momenta undergoing
!>      uniaxial stress ( in x-direction of orthorhombic cell)
!>      due to a shock compression:
!>      Reed et. al. Physical Review Letters 90, 235503 (2003).
!>      Added damping (e.g. thermostat to barostat)
!> \param md_env ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
  SUBROUTINE nph_uniaxial_damped ( md_env)

      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'nph_uniaxial_damped', &
         routineP = moduleN//':'//routineN
      REAL(dp), PARAMETER                                :: e2 = 1._dp/6._dp, e4 = e2/20._dp, &
                                                            e6 = e4/42._dp, e8 = e6/72._dp

      INTEGER                                            :: iroll, nparticle, nparticle_kind, nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: first, first_time, shell_adiabatic, &
                                                            shell_present
      REAL(KIND=dp)                                      :: aa, aax, dt, gamma1, infree, kin, &
                                                            roll_tol, roll_tol_thrs
      REAL(KIND=dp), DIMENSION(3)                        :: vector_r, vector_v
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_kin
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(npt_info_type), POINTER                       :: npt( :, : )
      TYPE(old_variables_type), POINTER                  :: old
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(virial_type), POINTER                         :: virial

    NULLIFY(gci, force_env)
    NULLIFY(atomic_kinds,cell,para_env,subsys,local_molecules,local_particles)
    NULLIFY(molecule_kinds,molecules,molecule_kind_set,npt)
    NULLIFY(core_particles,particles,shell_particles,tmp,old)
    NULLIFY(core_particle_set,particle_set,shell_particle_set)
    NULLIFY(simpar,virial,itimes)

    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env,npt = npt,&
         first_time=first_time, para_env=para_env, itimes=itimes)
    dt = simpar%dt
    infree = 1.0_dp / REAL ( simpar%nfree, dp )
    gamma1 = simpar%gamma_nph

    CALL force_env_get(force_env, subsys=subsys, cell = cell)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
         particles=particles,local_molecules_new=local_molecules,molecules_new=molecules,gci=gci,&
         molecule_kinds_new=molecule_kinds,virial=virial)

    nparticle_kind = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    IF ( first_time ) THEN
       CALL virial_evaluate ( atomic_kind_set, particle_set,  &
            local_particles, virial, para_env%group)
    END IF

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
         shell_present=shell_present,shell_adiabatic=shell_adiabatic)

    ! Allocate work storage for positions and velocities
    CALL allocate_old ( old, particle_set, npt)

    IF(shell_present) THEN
       CALL cp_subsys_get(subsys=subsys,&
            shell_particles=shell_particles, core_particles=core_particles)
       shell_particle_set =>  shell_particles%els
       nshell = SIZE(shell_particles%els)
       IF(shell_adiabatic) THEN
          core_particle_set =>  core_particles%els
       END IF
    END IF

    CALL allocate_tmp(md_env, tmp,nparticle, nshell, shell_adiabatic)

    ! perform damping on velocities
    CALL damp_v ( molecule_kind_set, molecule_set, particle_set, local_molecules,&
         gamma1, npt(1,1), dt, para_env%group )

    IF (simpar%constraint) THEN
       ! Possibly update the target values
       CALL shake_update_targets(gci, local_molecules, molecule_set, &
            molecule_kind_set, dt, force_env%root_section)
    END IF

    ! setting up for ROLL: saving old variables
    IF (simpar%constraint) THEN
       roll_tol_thrs = simpar%roll_tol
       iroll = 1
       CALL set ( old, atomic_kind_set, particle_set, local_particles, cell, npt, 'F' )
       CALL getold ( gci, local_molecules, molecule_set,  &
            molecule_kind_set, particle_set, cell)
    ELSE
       roll_tol_thrs = EPSILON(0.0_dp)
    ENDIF
    roll_tol = -roll_tol_thrs

    SR: DO WHILE (ABS(roll_tol)>=roll_tol_thrs) ! SHAKE-ROLL LOOP

       ! perform damping on the barostat momentum
       CALL damp_veps ( npt ( 1, 1 ), gamma1, dt )

       IF (simpar%constraint) THEN
          CALL set ( old, atomic_kind_set, particle_set, local_particles, cell, npt, 'B' )
       END IF
       CALL update_pv ( gci, simpar, atomic_kind_set, particle_set, &
            local_molecules, molecule_set, molecule_kind_set, &
            local_particles, kin, pv_kin, virial, para_env%group )
       CALL update_veps ( cell, npt, simpar, pv_kin, kin, virial, infree )

       ! perform damping on the barostat momentum
       CALL damp_veps ( npt ( 1, 1 ), gamma1, dt )

       tmp%arg_r(1) = ( 0.5_dp * npt ( 1, 1 )%v * dt ) * &
            ( 0.5_dp * npt ( 1, 1 )%v * dt )
       tmp%poly_r(1) = 1._dp + e2*tmp%arg_r(1) + e4*tmp%arg_r(1)*tmp%arg_r(1) +&
           e6*tmp%arg_r(1)**3 + e8*tmp%arg_r(1)**4

       aax = npt ( 1, 1 )%v * ( 1.0_dp + infree )
       tmp%arg_v(1) = ( 0.25_dp * dt * aax ) * ( 0.25_dp * dt * aax )
       tmp%poly_v(1) = 1._dp + e2*tmp%arg_v(1) + e4*tmp%arg_v(1)*tmp%arg_v(1) +&
          e6*tmp%arg_v(1)**3 + e8*tmp%arg_v(1)**4

       aa = npt ( 1, 1 )%v *  infree
       tmp%arg_v(2) = ( 0.25_dp * dt * aa ) * ( 0.25_dp * dt * aa )
       tmp%poly_v(2) = 1._dp + e2*tmp%arg_v(2) + e4*tmp%arg_v(2)*tmp%arg_v(2) +&
           e6*tmp%arg_v(2)**3 + e8*tmp%arg_v(2)**4
       tmp%poly_v(3) = 1._dp + e2*tmp%arg_v(2) + e4*tmp%arg_v(2)*tmp%arg_v(2) +&
           e6*tmp%arg_v(2)**3 + e8*tmp%arg_v(2)**4

       tmp%scale_r(1) = EXP( 0.5_dp * dt * npt ( 1, 1 )%v )
       tmp%scale_v(1) = EXP( -0.25_dp * dt * aax )
       tmp%scale_v(2) = EXP( -0.25_dp * dt * aa )
       tmp%scale_v(3) = EXP( -0.25_dp * dt * aa )

       ! first half of velocity verlet
       CALL vv_first( tmp, atomic_kind_set, local_particles, particle_set, &
            core_particle_set, shell_particle_set, nparticle_kind,&
            shell_adiabatic, dt)


       IF(simpar%variable_dt)  CALL variable_timestep(md_env,tmp, dt, simpar,para_env,&
           atomic_kind_set, local_particles, particle_set, core_particle_set,&
           shell_particle_set, nparticle_kind, shell_adiabatic, npt=npt)


       roll_tol = 0._dp
       vector_r ( : ) = 0._dp
       vector_v ( : ) = tmp%scale_v(:) * tmp%poly_v(:)
       vector_r ( 1 ) = tmp%scale_r(1) * tmp%poly_r(1)

       IF ( simpar%constraint ) CALL shake_roll_control( gci, local_molecules, &
            molecule_set, molecule_kind_set, particle_set, tmp%pos,tmp%vel, dt, simpar,&
            roll_tol, iroll, vector_r, vector_v, para_env%group, cell=cell,&
            local_particles=local_particles)
    END DO SR

    ! Update h_mat
    cell%hmat ( 1, 1 ) = cell%hmat ( 1, 1 ) * tmp%scale_r(1) * tmp%scale_r(1)

    ! Update the inverse
    CALL init_cell ( cell )

    ! Broadcast the new particle positions and deallocate the pos components of temporary
    CALL update_dealloc_tmp (tmp,particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, pos=.TRUE.)

    ! Update forces
    CALL force_env_calc_energy_force(force_env)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, tmp%vel)

    ! Velocity Verlet (second part)
    CALL vv_second(tmp, atomic_kind_set, local_particles, particle_set, &
         core_particle_set, shell_particle_set, nparticle_kind,&
         shell_adiabatic, dt)

    IF (simpar%constraint) THEN
       roll_tol_thrs = simpar%roll_tol
       first = .TRUE.
       iroll = 1
       CALL set ( old, atomic_kind_set, particle_set, tmp%vel, local_particles, cell, npt, 'F' )
    ELSE
       roll_tol_thrs = EPSILON(0.0_dp)
    ENDIF
    roll_tol = -roll_tol_thrs

    RR: DO WHILE (ABS(roll_tol)>=roll_tol_thrs) ! RATTLE-ROLL LOOP
       roll_tol = 0._dp
       IF ( simpar%constraint ) CALL rattle_roll_setup ( old, gci, atomic_kind_set, &
            particle_set, local_particles, molecule_kind_set, molecule_set, local_molecules,&
            tmp%vel, dt, cell, npt, simpar, virial, vector_v, roll_tol, iroll, infree, first,&
            para_env)
       ! perform damping on the barostat momentum
       CALL damp_veps ( npt ( 1, 1 ), gamma1, dt )

       CALL update_pv ( gci, simpar, atomic_kind_set, tmp%vel, particle_set, &
            local_molecules, molecule_set, molecule_kind_set, &
            local_particles, kin, pv_kin, virial, para_env%group )
       CALL update_veps ( cell, npt, simpar, pv_kin, kin, virial, infree )

       ! perform damping on the barostat momentum
       CALL damp_veps ( npt ( 1, 1 ), gamma1, dt )

    END DO RR

    ! perform damping on velocities
    CALL damp_v ( molecule_kind_set, molecule_set, particle_set, local_molecules,&
         tmp%vel, gamma1,  npt ( 1, 1 ), dt, para_env%group )

    IF (simpar%annealing)  tmp%vel(:,:)=tmp%vel(:,:)*simpar%f_annealing

    ! Broadcast the new particle velocities and deallocate temporary
    CALL update_dealloc_tmp (tmp,particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, vel=.TRUE.)

    ! Update constraint virial
    IF ( simpar%constraint ) CALL pv_constraint ( gci, local_molecules, &
         molecule_set, molecule_kind_set, particle_set, virial, para_env%group )

    CALL virial_evaluate ( atomic_kind_set, particle_set,  &
         local_particles, virial, para_env%group)

    ! Deallocate old variables
    CALL deallocate_old ( old)

    IF (first_time) THEN
       first_time = .FALSE.
       CALL set_md_env(md_env, first_time=first_time)
    END IF

  END SUBROUTINE nph_uniaxial_damped

! **************************************************************************************************
!> \brief Velocity Verlet integrator for the NPT ensemble with fully flexible cell
!> \param md_env ...
!> \param globenv ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
  SUBROUTINE npt_f ( md_env, globenv)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'npt_f', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: e2 = 1.0_dp/6.0_dp, e4 = e2/20.0_dp, &
                                                            e6 = e4/42.0_dp, e8 = e6/72.0_dp

      INTEGER                                            :: i, iroll, j, nparticle, nparticle_kind, &
                                                            nshell
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: first, first_time, shell_adiabatic, &
                                                            shell_check_distance, shell_present
      REAL(KIND=dp)                                      :: dt, infree, kin, roll_tol, &
                                                            roll_tol_thrs, trvg
      REAL(KIND=dp), DIMENSION(3)                        :: vector_r, vector_v
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_kin, uh
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(barostat_type), POINTER                       :: barostat
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(npt_info_type), POINTER                       :: npt( :, : )
      TYPE(old_variables_type), POINTER                  :: old
      TYPE(particle_list_type), POINTER                  :: core_particles, particles, &
                                                            shell_particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: core_particle_set, particle_set, &
                                                            shell_particle_set
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermostat_type), POINTER                     :: thermostat_baro, thermostat_part, &
                                                            thermostat_shell
      TYPE(tmp_variables_type), POINTER                  :: tmp
      TYPE(virial_type), POINTER                         :: virial

    NULLIFY(gci,thermostat_baro,thermostat_part,thermostat_shell,force_env)
    NULLIFY(atomic_kinds,cell,para_env,subsys,local_molecules,local_particles)
    NULLIFY(molecule_kinds,molecules,molecule_kind_set,npt,barostat)
    NULLIFY(core_particles,particles,shell_particles,tmp,old)
    NULLIFY(core_particle_set,particle_set,shell_particle_set)
    NULLIFY(simpar,virial,itimes)

    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env,&
         thermostat_part=thermostat_part, thermostat_baro=thermostat_baro,&
         thermostat_shell=thermostat_shell, npt=npt, first_time=first_time,&
         para_env=para_env, barostat=barostat, itimes=itimes)
    dt = simpar%dt
    infree = 1.0_dp / REAL ( simpar%nfree,KIND=dp)

    CALL force_env_get(force_env, subsys=subsys, cell=cell)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds, local_particles=local_particles,&
         particles=particles,local_molecules_new=local_molecules,molecules_new=molecules, &
         gci=gci,molecule_kinds_new=molecule_kinds,virial=virial)

    nparticle_kind = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
         shell_present=shell_present,shell_adiabatic=shell_adiabatic,&
         shell_check_distance=shell_check_distance)

    IF ( first_time ) THEN
       CALL virial_evaluate ( atomic_kind_set, particle_set,  &
            local_particles, virial, para_env%group)
    END IF

    ! Allocate work storage for positions and velocities
    CALL allocate_old ( old, particle_set, npt)

    IF(shell_present) THEN
       CALL cp_subsys_get(subsys=subsys,&
            shell_particles=shell_particles, core_particles=core_particles)
       shell_particle_set =>  shell_particles%els
       nshell = SIZE(shell_particles%els)
       IF(shell_adiabatic) THEN
          core_particle_set =>  core_particles%els
       END IF
    END IF

    CALL allocate_tmp(md_env, tmp,nparticle, nshell, shell_adiabatic)

    ! Apply Thermostat to Barostat
    CALL apply_thermostat_baro(thermostat_baro, npt, para_env%group)

    ! Apply Thermostat over the full set of particles
    IF(simpar% ensemble /= npe_f_ensemble) THEN
      IF(shell_adiabatic) THEN
        CALL apply_thermostat_particles(thermostat_part, force_env, molecule_kind_set, molecule_set,&
             particle_set, local_molecules, local_particles, para_env%group,shell_adiabatic=shell_adiabatic,&
             shell_particle_set=shell_particle_set,core_particle_set=core_particle_set)
      ELSE
        CALL apply_thermostat_particles(thermostat_part, force_env, molecule_kind_set, molecule_set,&
             particle_set, local_molecules, local_particles, para_env%group)
      END IF
    END IF

    ! Apply Thermostat over the core-shell motion
    CALL apply_thermostat_shells(thermostat_shell, atomic_kind_set, particle_set,&
         local_particles, para_env%group, shell_particle_set=shell_particle_set,&
         core_particle_set=core_particle_set)

    IF (simpar%constraint) THEN
       ! Possibly update the target values
       CALL shake_update_targets(gci, local_molecules, molecule_set, &
            molecule_kind_set, dt, force_env%root_section)
    END IF

    ! setting up for ROLL: saving old variables
    IF (simpar%constraint) THEN
       roll_tol_thrs = simpar%roll_tol
       iroll = 1
       CALL set ( old, atomic_kind_set, particle_set, local_particles, cell, npt, 'F' )
       CALL getold ( gci, local_molecules, molecule_set,  &
            molecule_kind_set, particle_set, cell)
    ELSE
       roll_tol_thrs = EPSILON(0.0_dp)
    ENDIF
    roll_tol = -roll_tol_thrs

    SR: DO WHILE (ABS(roll_tol)>=roll_tol_thrs) ! SHAKE-ROLL LOOP

       IF (simpar%constraint) THEN
          CALL set ( old, atomic_kind_set, particle_set, local_particles, cell, npt, 'B' )
       END IF
       CALL update_pv ( gci, simpar, atomic_kind_set, particle_set, &
            local_molecules, molecule_set, molecule_kind_set, &
            local_particles, kin, pv_kin, virial, para_env%group )
       CALL update_veps ( cell, npt, simpar, pv_kin, kin, virial, infree,&
                          virial_components=barostat%virial_components)

       trvg = npt ( 1, 1 )%v  + npt ( 2, 2 )%v + npt ( 3, 3 )%v
       !
       ! find eigenvalues and eigenvectors of npt ( :, : )%v
       !

       CALL diagonalise ( matrix = npt(:,:)%v, mysize = 3, &
            storageform = "UPPER", eigenvalues = tmp%e_val, eigenvectors = tmp%u )

       tmp%arg_r ( : ) =  0.5_dp * tmp%e_val ( : ) * dt * &
            0.5_dp * tmp%e_val ( : ) * dt
       tmp%poly_r = 1.0_dp + e2*tmp%arg_r + e4*tmp%arg_r*tmp%arg_r + &
                e6*tmp%arg_r**3 + e8*tmp%arg_r**4
       tmp%scale_r ( : ) = EXP ( 0.5_dp * dt * tmp%e_val ( : ) )

       tmp%arg_v ( : ) =  0.25_dp * dt * ( tmp%e_val ( : ) + trvg * infree ) *&
            0.25_dp * dt * ( tmp%e_val ( : ) + trvg * infree )
       tmp%poly_v = 1.0_dp + e2*tmp%arg_v + e4*tmp%arg_v*tmp%arg_v +&
            e6*tmp%arg_v**3 + e8*tmp%arg_v**4
       tmp%scale_v ( : ) = EXP ( -0.25_dp * dt * ( tmp%e_val( : ) + trvg * infree ) )

       CALL vv_first(tmp, atomic_kind_set, local_particles, particle_set, &
            core_particle_set, shell_particle_set, nparticle_kind,&
            shell_adiabatic, dt, u=tmp%u)

       IF(simpar%variable_dt) CALL  variable_timestep(md_env,tmp, dt, simpar,para_env,&
           atomic_kind_set, local_particles, particle_set, core_particle_set,&
           shell_particle_set, nparticle_kind, shell_adiabatic,npt=npt)

       roll_tol = 0.0_dp
       vector_r = tmp%scale_r*tmp%poly_r
       vector_v = tmp%scale_v*tmp%poly_v

       IF ( simpar%constraint ) CALL shake_roll_control( gci, local_molecules, &
            molecule_set, molecule_kind_set, particle_set, tmp%pos, tmp%vel, dt,&
            simpar, roll_tol, iroll, vector_r, vector_v, &
            para_env%group, u=tmp%u, cell=cell,&
            local_particles=local_particles)
    END DO SR

    ! Update h_mat
    uh = MATMUL_3X3 ( TRANSPOSE_3D ( tmp%u ), cell%hmat )

    DO i = 1, 3
       DO j = 1, 3
          uh ( i, j ) = uh ( i, j ) * tmp%scale_r ( i ) * tmp%scale_r ( i )
       END DO
    END DO

    cell%hmat = MATMUL_3x3 ( tmp%u, uh )
    ! Update the inverse
    CALL init_cell ( cell )

    ! Broadcast the new particle positions and deallocate the pos components of temporary
    CALL update_dealloc_tmp (tmp,particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, pos=.TRUE.)

    IF(shell_adiabatic .AND. shell_check_distance) THEN
        CALL optimize_shell_core(force_env,  particle_set,&
             shell_particle_set, core_particle_set, globenv, tmp=tmp, check=.TRUE.)
    END IF

    ! Update forces
    CALL force_env_calc_energy_force(force_env)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, tmp%vel)

    ! Velocity Verlet (second part)
    CALL vv_second (tmp, atomic_kind_set, local_particles, particle_set,&
         core_particle_set, shell_particle_set, nparticle_kind,&
         shell_adiabatic, dt, tmp%u)

    IF (simpar%constraint) THEN
       roll_tol_thrs = simpar%roll_tol
       first = .TRUE.
       iroll = 1
       CALL set (old, atomic_kind_set, particle_set, tmp%vel, local_particles, cell, npt, 'F' )
    ELSE
       roll_tol_thrs = EPSILON(0.0_dp)
    ENDIF
    roll_tol = -roll_tol_thrs

    RR: DO WHILE (ABS(roll_tol)>=roll_tol_thrs) ! RATTLE-ROLL LOOP
       roll_tol = 0.0_dp
       IF ( simpar%constraint ) CALL rattle_roll_setup ( old, gci, atomic_kind_set, &
            particle_set, local_particles, molecule_kind_set, molecule_set,  &
            local_molecules, tmp%vel, dt, cell, npt, simpar, virial, vector_v,  &
            roll_tol, iroll, infree, first, para_env, u=tmp%u)

       CALL update_pv ( gci, simpar, atomic_kind_set, tmp%vel, particle_set, &
            local_molecules, molecule_set, molecule_kind_set, &
            local_particles, kin, pv_kin, virial, para_env%group )
       CALL update_veps ( cell, npt, simpar, pv_kin, kin, virial, infree,&
                          virial_components=barostat%virial_components)
    END DO RR

    ! Apply Thermostat over the full set of particles
    IF(simpar% ensemble /= npe_f_ensemble) THEN
      IF (shell_adiabatic) THEN
        CALL apply_thermostat_particles(thermostat_part, force_env,molecule_kind_set, molecule_set, &
             particle_set, local_molecules, local_particles, para_env%group,shell_adiabatic=shell_adiabatic,&
             vel=tmp%vel, shell_vel=tmp%shell_vel, core_vel=tmp%core_vel)

      ELSE
        CALL apply_thermostat_particles(thermostat_part, force_env,molecule_kind_set, molecule_set, &
             particle_set, local_molecules, local_particles, para_env%group, vel=tmp%vel)
      END IF
    END IF

    ! Apply Thermostat over the core-shell motion
    IF(ASSOCIATED(thermostat_shell)) THEN
       CALL apply_thermostat_shells(thermostat_shell,atomic_kind_set, particle_set,&
            local_particles, para_env%group, vel=tmp%vel, shell_vel=tmp%shell_vel, &
            core_vel=tmp%core_vel)
    END IF

    ! Apply Thermostat to Barostat
    CALL apply_thermostat_baro( thermostat_baro, npt, para_env%group)

    ! Annealing of particle velocities is only possible when no thermostat is active
    IF (simpar% ensemble == npe_f_ensemble .AND. simpar%annealing) THEN
       tmp%vel(:,:)=tmp%vel(:,:)*simpar%f_annealing
       IF (shell_adiabatic) THEN
          CALL shell_scale_comv(atomic_kind_set,local_particles,particle_set,&
               tmp%vel,tmp%shell_vel,tmp%core_vel)
       END IF
    END IF
    ! Annealing of CELL velocities is only possible when no thermostat is active
    IF (simpar% ensemble == npe_f_ensemble .AND. simpar%annealing_cell) THEN
       npt(:,:)%v = npt(:,:)%v * simpar%f_annealing_cell
    END IF

    ! Broadcast the new particle velocities and deallocate temporary
    CALL update_dealloc_tmp (tmp,particle_set, shell_particle_set, &
         core_particle_set, para_env, shell_adiabatic, vel=.TRUE.)

    ! Update constraint virial
    IF ( simpar%constraint ) &
         CALL pv_constraint ( gci, local_molecules, molecule_set, &
         molecule_kind_set, particle_set, virial, para_env%group )

    CALL virial_evaluate ( atomic_kind_set, particle_set,  &
         local_particles, virial, para_env%group)

    ! Deallocate old variables
    CALL deallocate_old ( old)

    IF (first_time) THEN
       first_time = .FALSE.
       CALL set_md_env(md_env, first_time=first_time)
    END IF

  END SUBROUTINE npt_f

! **************************************************************************************************
!> \brief RESPA integrator for nve ensemble for particle positions & momenta
!> \param md_env ...
!> \author FS
! **************************************************************************************************
  SUBROUTINE nve_respa ( md_env)

      TYPE(md_environment_type), POINTER                 :: md_env

      CHARACTER(len=*), PARAMETER :: routineN = 'nve_respa', routineP = moduleN//':'//routineN

      INTEGER                                            :: i_step, iparticle, iparticle_kind, &
                                                            iparticle_local, n_time_steps, &
                                                            nparticle, nparticle_kind, &
                                                            nparticle_local
      INTEGER, POINTER                                   :: itimes
      REAL(KIND=dp)                                      :: dm, dt, mass
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: pos, vel
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys, subsys_respa
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: particles, particles_respa
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set, particle_set_respa
      TYPE(simpar_type), POINTER                         :: simpar

    NULLIFY (para_env,cell,subsys_respa,particles_respa, particle_set_respa, gci, force_env, atomic_kinds)
    NULLIFY (atomic_kind_set,simpar,subsys,particles,particle_set)
    NULLIFY (local_molecules,molecule_kinds, molecules,molecule_kind_set, local_particles,itimes)
    CALL get_md_env(md_env=md_env,simpar=simpar,force_env=force_env, &
                    para_env=para_env, itimes=itimes)
    dt = simpar%dt

    n_time_steps=simpar%n_time_steps

    CALL force_env_get(force_env,subsys=subsys,cell=cell)
    CALL force_env_get(force_env%sub_force_env(1)%force_env,subsys=subsys_respa)

    ! Do some checks on coordinates and box
    CALL apply_qmmm_walls_reflective(force_env)

    CALL cp_subsys_get(subsys=subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
         particles=particles, local_molecules_new=local_molecules, molecules_new=molecules,&
         gci=gci, molecule_kinds_new=molecule_kinds)

    CALL cp_subsys_get (subsys=subsys_respa, particles=particles_respa)
    particle_set_respa => particles_respa%els

    nparticle_kind = atomic_kinds%n_els
    atomic_kind_set => atomic_kinds%els
    molecule_kind_set => molecule_kinds%els

    nparticle = particles%n_els
    particle_set => particles%els
    molecule_set => molecules%els

    ! Allocate work storage for positions and velocities
    ALLOCATE (pos(3,nparticle))
    ALLOCATE (vel(3,nparticle))
    vel(:,:) = 0.0_dp

    IF (simpar%constraint) CALL getold(gci, local_molecules, molecule_set,  &
         molecule_kind_set, particle_set, cell)

    ! Multiple time step (first part)
    DO iparticle_kind=1,nparticle_kind
       atomic_kind => atomic_kind_set(iparticle_kind)
       CALL get_atomic_kind(atomic_kind=atomic_kind,mass=mass)
       dm = 0.5_dp * dt / mass
       nparticle_local = local_particles%n_el(iparticle_kind)
       DO iparticle_local=1,nparticle_local
          iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
          vel (:,iparticle) = particle_set(iparticle)%v(:) +&
               dm * (particle_set(iparticle)%f(:)-&
               particle_set_respa(iparticle)%f(:))
       END DO
    END DO

    ! Velocity Verlet (first part)
    DO i_step=1,n_time_steps
       pos(:,:) = 0.0_dp
       DO iparticle_kind=1,nparticle_kind
          atomic_kind => atomic_kind_set(iparticle_kind)
          CALL get_atomic_kind(atomic_kind=atomic_kind,mass=mass)
          dm = 0.5_dp*dt/(n_time_steps*mass)
          nparticle_local = local_particles%n_el(iparticle_kind)
          DO iparticle_local=1,nparticle_local
             iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
             vel (:,iparticle) = vel(:,iparticle) +&
                  dm * particle_set_respa(iparticle)%f(:)
             pos (:,iparticle) = particle_set(iparticle)%r(:) +&
                  (dt/n_time_steps) * vel ( :, iparticle )
          END DO
       END DO

       IF (simpar%constraint) THEN
          ! Possibly update the target values
          CALL shake_update_targets(gci, local_molecules, molecule_set, &
              molecule_kind_set, dt, force_env%root_section)

          CALL shake_control( gci, local_molecules, molecule_set,&
              molecule_kind_set, particle_set,pos, vel, dt, simpar%shake_tol,&
              simpar%info_constraint, simpar%lagrange_multipliers,simpar%dump_lm, cell,&
              para_env%group, local_particles)
       END IF

       ! Broadcast the new particle positions
       CALL update_particle_set ( particle_set, para_env%group, pos = pos)
       DO iparticle = 1, SIZE(particle_set)
          particle_set_respa(iparticle)%r = particle_set(iparticle)%r
       END DO

       ! Update forces
       CALL force_env_calc_energy_force(force_env%sub_force_env(1)%force_env)

       ! Metadynamics
       CALL metadyn_integrator(force_env, itimes, vel)

       ! Velocity Verlet (second part)
       DO iparticle_kind=1,nparticle_kind
          atomic_kind => atomic_kind_set(iparticle_kind)
          CALL get_atomic_kind(atomic_kind=atomic_kind,mass=mass)
          dm = 0.5_dp*dt /(n_time_steps * mass)
          nparticle_local = local_particles%n_el(iparticle_kind)
          DO iparticle_local=1,nparticle_local
             iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
             vel ( 1, iparticle) = vel ( 1, iparticle) + dm*particle_set_respa(iparticle)%f(1)
             vel ( 2, iparticle) = vel ( 2, iparticle) + dm*particle_set_respa(iparticle)%f(2)
             vel ( 3, iparticle) = vel ( 3, iparticle) + dm*particle_set_respa(iparticle)%f(3)
          END DO
       END DO

       IF (simpar%constraint) CALL rattle_control( gci,local_molecules, molecule_set, &
            molecule_kind_set, particle_set, vel, dt,simpar%shake_tol,&
            simpar%info_constraint, simpar%lagrange_multipliers,&
            simpar%dump_lm, cell, para_env%group,local_particles)

       IF (simpar%annealing)  vel(:,:)=vel(:,:)*simpar%f_annealing
    END DO
    DEALLOCATE (pos)

    ! Multiple time step (second part)
    ! Compute forces for respa force_env
    CALL force_env_calc_energy_force(force_env)

    ! Metadynamics
    CALL metadyn_integrator(force_env, itimes, vel)

    DO iparticle_kind=1,nparticle_kind
       atomic_kind => atomic_kind_set(iparticle_kind)
       CALL get_atomic_kind(atomic_kind=atomic_kind,mass=mass)
       dm = 0.5_dp * dt  / mass
       nparticle_local = local_particles%n_el(iparticle_kind)
       DO iparticle_local=1,nparticle_local
          iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
          vel (1,iparticle) = vel(1, iparticle)+dm*(particle_set(iparticle)%f(1)-particle_set_respa(iparticle)%f(1))
          vel (2,iparticle) = vel(2, iparticle)+dm*(particle_set(iparticle)%f(2)-particle_set_respa(iparticle)%f(2))
          vel (3,iparticle) = vel(3, iparticle)+dm*(particle_set(iparticle)%f(3)-particle_set_respa(iparticle)%f(3))
       END DO
    END DO

    ! Broadcast the new particle velocities
    CALL update_particle_set ( particle_set, para_env%group, vel = vel)

    DEALLOCATE (vel)

  END SUBROUTINE nve_respa

END MODULE integrator
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides an interface to the velocity-verlet based integrator
!>      routines for all ensembles
!> \author CJM (11-SEPT-2002)
! **************************************************************************************************
MODULE velocity_verlet_control

   USE force_env_types,                 ONLY: force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: &
        isokin_ensemble, langevin_ensemble, npe_f_ensemble, npe_i_ensemble, &
        nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, npt_f_ensemble, npt_i_ensemble, &
        nve_ensemble, nvt_adiabatic_ensemble, nvt_ensemble, reftraj_ensemble
   USE integrator,                      ONLY: &
        isokin, langevin, nph_uniaxial, nph_uniaxial_damped, npt_f, npt_i, nve, nve_respa, nvt, &
        nvt_adiabatic, reftraj
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type
   USE simpar_types,                    ONLY: simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'velocity_verlet_control'
   PUBLIC :: velocity_verlet

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param md_env ...
!> \param globenv ...
!> \par History
!>      none
!> \author CJM
! **************************************************************************************************
   SUBROUTINE velocity_verlet(md_env, globenv)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'velocity_verlet', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(simpar_type), POINTER                         :: simpar

      CALL timeset(routineN, handle)

      ! Get force environment
      CALL get_md_env(md_env, force_env=force_env, simpar=simpar)

      ! RESPA implemented only for NVE
      IF (simpar%do_respa .AND. nve_ensemble .NE. simpar%ensemble) THEN
         CALL cp__b("motion/velocity_verlet_control.F",62,"RESPA integrator not implemented for this ensemble")
      END IF

      ! Choice of the ensemble
      SELECT CASE (simpar%ensemble)
      CASE DEFAULT
         CALL cp__b("motion/velocity_verlet_control.F",68,"Integrator not implemented")
      CASE (nve_ensemble)
         IF (simpar%do_respa) THEN
            CALL nve_respa(md_env)
         ELSE
            CALL nve(md_env, globenv)
         END IF
      CASE (nvt_ensemble)
         CALL nvt(md_env, globenv)
      CASE (nvt_adiabatic_ensemble)
         CALL nvt_adiabatic(md_env, globenv)
      CASE (isokin_ensemble)
         CALL isokin(md_env)
      CASE (npt_i_ensemble)
         CALL npt_i(md_env, globenv)
      CASE (npt_f_ensemble)
         CALL npt_f(md_env, globenv)
      CASE (nph_uniaxial_ensemble)
         CALL nph_uniaxial(md_env)
      CASE (nph_uniaxial_damped_ensemble)
         CALL nph_uniaxial_damped(md_env)
      CASE (reftraj_ensemble)
         CALL reftraj(md_env)
      CASE (langevin_ensemble)
         CALL langevin(md_env)
      CASE (npe_f_ensemble)
         CALL npt_f(md_env, globenv)
      CASE (npe_i_ensemble)
         CALL npt_i(md_env, globenv)
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE velocity_verlet

END MODULE velocity_verlet_control
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility functions that are needed for RTP/EMD in combination with
!>        HF or hybrid functionals (needs to deal with imaginary KS and P
!> \par History
!>      2014 created [fschiff]
!> \author Florina Schiffmann
! **************************************************************************************************
MODULE rt_hfx_utils
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_alloc_block_from_nbl
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_col_block_sizes, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_distribution, cp_dbcsr_init, cp_dbcsr_p_type, &
        cp_dbcsr_row_block_sizes, cp_dbcsr_set, dbcsr_type_antisymmetric
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_set,&
                                              qs_rho_type
   USE rt_propagation_types,            ONLY: rt_prop_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_hfx_utils'

   PUBLIC :: rtp_hfx_rebuild

!***
CONTAINS
! **************************************************************************************************
!> \brief rebuilds the structures of P and KS (imaginary) in case S changed
!> \param qs_env ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE rtp_hfx_rebuild(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rtp_hfx_rebuild', &
         routineP = moduleN//':'//routineN

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks_aux_im, matrix_ks_im, &
                                                            matrix_s, matrix_s_aux, rho_ao_im, &
                                                            rho_aux_ao_im
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_aux, sab_orb
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho, rho_aux
      TYPE(rt_prop_type), POINTER                        :: rtp

      NULLIFY (ks_env, matrix_s, matrix_ks_im, dft_control, sab_orb, rtp, rho, rho_ao_im)
      NULLIFY (sab_aux, rho_aux, rho_aux_ao_im, matrix_ks_aux_im, matrix_s_aux)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      dft_control=dft_control, &
                      rtp=rtp, &
                      matrix_s=matrix_s, &
                      rho=rho, &
                      matrix_ks_im=matrix_ks_im, &
                      sab_orb=sab_orb)

      CALL qs_rho_get(rho, rho_ao_im=rho_ao_im)
      CALL rebuild_matrices(rho_ao_im, matrix_ks_im, sab_orb, matrix_s, &
                            dft_control%nspins)
      CALL set_ks_env(ks_env, matrix_ks_im=matrix_ks_im)
      CALL qs_rho_set(rho, rho_ao_im=rho_ao_im)

      IF (dft_control%do_admm) THEN
         CALL get_qs_env(qs_env, &
                         matrix_s_aux_fit=matrix_s_aux, &
                         sab_aux_fit=sab_aux, &
                         rho_aux_fit=rho_aux, &
                         matrix_ks_aux_fit_im=matrix_ks_aux_im)
         CALL qs_rho_get(rho_aux, rho_ao_im=rho_aux_ao_im)
         CALL rebuild_matrices(rho_aux_ao_im, matrix_ks_aux_im, sab_aux, matrix_s_aux, &
                               dft_control%nspins)
         CALL set_ks_env(ks_env, matrix_ks_aux_fit_im=matrix_ks_aux_im)
         CALL qs_rho_set(rho_aux, rho_ao_im=rho_aux_ao_im)
      END IF

   END SUBROUTINE rtp_hfx_rebuild

! **************************************************************************************************
!> \brief does the actual rebuilding of P and KS (imaginary) in case S changed
!> \param matrix_p ...
!> \param matrix_ks ...
!> \param sab_orb ...
!> \param matrix_s ...
!> \param nspins ...
!> \author Florian Schiffmann
! **************************************************************************************************

   SUBROUTINE rebuild_matrices(matrix_p, matrix_ks, sab_orb, matrix_s, nspins)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_p, matrix_ks
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      INTEGER, INTENT(in)                                :: nspins

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rebuild_matrices', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: headline
      INTEGER                                            :: i

      IF (ASSOCIATED(matrix_p)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(matrix_p)
      END IF
      ! Create a new density matrix set
      CALL cp_dbcsr_allocate_matrix_set(matrix_p, nspins)
      DO i = 1, nspins
         headline = "Imaginary density matrix"
         ALLOCATE (matrix_p(i)%matrix)
         CALL cp_dbcsr_init(matrix_p(i)%matrix)
         CALL cp_dbcsr_create(matrix=matrix_p(i)%matrix, &
                              name=TRIM(headline), &
                              dist=cp_dbcsr_distribution(matrix_s(1)%matrix), matrix_type=dbcsr_type_antisymmetric, &
                              row_blk_size=cp_dbcsr_row_block_sizes(matrix_s(1)%matrix), &
                              col_blk_size=cp_dbcsr_col_block_sizes(matrix_s(1)%matrix), &
                              nze=0)
         CALL cp_dbcsr_alloc_block_from_nbl(matrix_p(i)%matrix, sab_orb)
         CALL cp_dbcsr_set(matrix_p(i)%matrix, 0.0_dp)
      END DO

      IF (ASSOCIATED(matrix_ks)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(matrix_ks)
      END IF
      ! Create a new density matrix set
      CALL cp_dbcsr_allocate_matrix_set(matrix_ks, nspins)
      DO i = 1, nspins
         headline = "Imaginary Kohn-Sham matrix"
         ALLOCATE (matrix_ks(i)%matrix)
         CALL cp_dbcsr_init(matrix_ks(i)%matrix)
         CALL cp_dbcsr_create(matrix=matrix_ks(i)%matrix, &
                              name=TRIM(headline), &
                              dist=cp_dbcsr_distribution(matrix_s(1)%matrix), matrix_type=dbcsr_type_antisymmetric, &
                              row_blk_size=cp_dbcsr_row_block_sizes(matrix_s(1)%matrix), &
                              col_blk_size=cp_dbcsr_col_block_sizes(matrix_s(1)%matrix), &
                              nze=0)
         CALL cp_dbcsr_alloc_block_from_nbl(matrix_ks(i)%matrix, sab_orb)
         CALL cp_dbcsr_set(matrix_ks(i)%matrix, 0.0_dp)
      END DO

   END SUBROUTINE rebuild_matrices

END MODULE rt_hfx_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to apply a delta pulse for RTP and EMD
! **************************************************************************************************

MODULE rt_delta_pulse
   USE cell_types,                      ONLY: cell_type
   USE cp_cfm_basic_linalg,             ONLY: cp_cfm_column_scale,&
                                              cp_cfm_gemm
   USE cp_cfm_diag,                     ONLY: cp_cfm_heevd
   USE cp_cfm_types,                    ONLY: cp_cfm_create,&
                                              cp_cfm_release,&
                                              cp_cfm_to_cfm,&
                                              cp_cfm_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_filter,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add,&
                                              cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_invert,&
                                              cp_fm_cholesky_reduce,&
                                              cp_fm_cholesky_restore
   USE cp_fm_diag,                      ONLY: cp_fm_syevd
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: twopi
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_moments,                      ONLY: build_berry_moment_matrix
   USE rt_propagation_types,            ONLY: get_rtp,&
                                              rt_prop_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_delta_pulse'

   PUBLIC :: apply_delta_pulse_periodic, &
             apply_delta_pulse

CONTAINS

! **************************************************************************************************
!> \brief uses perturbation theory to get the proper initial conditions
!> \param qs_env ...
!> \param mos_old ...
!> \param mos_new ...
!> \author Joost & Martin (2011)
! **************************************************************************************************

   SUBROUTINE apply_delta_pulse_periodic(qs_env, mos_old, mos_new)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_old, mos_new

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_delta_pulse_periodic', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: eigenvalues_sqrt
      INTEGER                                            :: handle, icol, idir, irow, ispin, nao, &
                                                            ncol_local, nmo, nrow_global, &
                                                            nrow_local, nvirt
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: factor
      REAL(KIND=dp), DIMENSION(3)                        :: kvec
      REAL(kind=dp), DIMENSION(:), POINTER               :: eigenvalues
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_cfm_type), POINTER                         :: oo_c, oo_v, oo_vt
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: S_inv
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: eigenvectors, mat_ks, mat_tmp, momentum, &
                                                            oo_1, oo_2, S_chol, S_inv_fm, tmpS, &
                                                            virtuals
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(rt_prop_type), POINTER                        :: rtp

      NULLIFY (dft_control)
      CALL timeset(routineN, handle)

      ! we need the overlap and ks matrix for a full diagionalization
      CALL get_qs_env(qs_env, &
                      cell=cell, &
                      mos=mos, &
                      rtp=rtp, &
                      matrix_s=matrix_s, &
                      matrix_ks=matrix_ks, &
                      dft_control=dft_control)
      CALL get_rtp(rtp=rtp, S_inv=S_inv)
      CALL cp_fm_create(S_chol, matrix_struct=rtp%ao_ao_fmstruct, name="S_chol")
      CALL cp_fm_create(S_inv_fm, matrix_struct=rtp%ao_ao_fmstruct, name="S_inv_fm")
      CALL cp_fm_create(tmpS, matrix_struct=rtp%ao_ao_fmstruct)
      CALL copy_dbcsr_to_fm(S_inv, S_inv_fm)
      CALL cp_fm_upper_to_full(S_inv_fm, tmpS)
      CALL cp_fm_get_info(S_inv_fm, nrow_global=nrow_global)
      CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, S_chol)
      CALL cp_fm_cholesky_decompose(S_chol)
      NULLIFY (mat_ks, eigenvectors, mat_tmp)
      CALL cp_fm_create(mat_ks, matrix_struct=S_inv_fm%matrix_struct, name="mat_ks")
      CALL cp_fm_create(eigenvectors, matrix_struct=S_inv_fm%matrix_struct, name="eigenvectors")

      DO ispin = 1, SIZE(matrix_ks)
         ALLOCATE (eigenvalues(nrow_global))
         CALL cp_fm_create(mat_tmp, matrix_struct=S_inv_fm%matrix_struct, name="mat_tmp")

         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, mat_ks)
         CALL cp_fm_cholesky_reduce(mat_ks, S_chol)
         CALL cp_fm_syevd(mat_ks, mat_tmp, eigenvalues)
         CALL cp_fm_cholesky_restore(mat_tmp, nrow_global, S_chol, eigenvectors, "SOLVE")

         ! virtuals
         CALL get_mo_set(mo_set=mos(ispin)%mo_set, nao=nao, nmo=nmo)
         nvirt = nao-nmo
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=S_inv_fm%matrix_struct%para_env, context=S_inv_fm%matrix_struct%context, &
                                  nrow_global=nrow_global, ncol_global=nvirt)
         CALL cp_fm_create(virtuals, matrix_struct=fm_struct_tmp, name="virtuals")
         CALL cp_fm_struct_release(fm_struct_tmp)
         CALL cp_fm_to_fm(eigenvectors, virtuals, nvirt, nmo+1, 1)

         ! occupied
         CALL cp_fm_to_fm(eigenvectors, mos_old(2*ispin-1)%matrix, nmo, 1, 1)

         CALL cp_fm_struct_create(fm_struct_tmp, para_env=S_inv_fm%matrix_struct%para_env, context=S_inv_fm%matrix_struct%context, &
                                  nrow_global=nvirt, ncol_global=nmo)
         CALL cp_fm_create(momentum, matrix_struct=fm_struct_tmp, name="momentum")
         CALL cp_fm_struct_release(fm_struct_tmp)

         ! the momentum operator (in a given direction)
         CALL cp_fm_set_all(mos_new(2*ispin-1)%matrix, 0.0_dp)

         ! the prefactor (strength of the electric field)
         kvec(:) = cell%h_inv(1, :)*dft_control%rtp_control%delta_pulse_direction(1)+ &
                   cell%h_inv(2, :)*dft_control%rtp_control%delta_pulse_direction(2)+ &
                   cell%h_inv(3, :)*dft_control%rtp_control%delta_pulse_direction(3)
         kvec = -kvec*twopi*dft_control%rtp_control%delta_pulse_scale

         DO idir = 1, 3
            factor = kvec(idir)
            IF (factor .NE. 0.0_dp) THEN
               CALL cp_dbcsr_sm_fm_multiply(matrix_s(idir+1)%matrix, mos_old(2*ispin-1)%matrix, &
                                            mos_old(2*ispin)%matrix, ncol=nmo)
               CALL cp_fm_scale_and_add(1.0_dp, mos_new(2*ispin-1)%matrix, factor, mos_old(2*ispin)%matrix)
            ENDIF
         ENDDO

         CALL cp_gemm('T', 'N', nvirt, nmo, nao, 1.0_dp, virtuals, mos_new(2*ispin-1)%matrix, 0.0_dp, momentum)

         ! the tricky bit ... rescale by the eigenvalue difference
         CALL cp_fm_get_info(momentum, nrow_local=nrow_local, ncol_local=ncol_local, &
                             row_indices=row_indices, col_indices=col_indices, local_data=local_data)
         DO icol = 1, ncol_local
            DO irow = 1, nrow_local
               factor = 1/(eigenvalues(col_indices(icol))-eigenvalues(nmo+row_indices(irow)))
               local_data(irow, icol) = factor*local_data(irow, icol)
            ENDDO
         ENDDO
         CALL cp_fm_release(mat_tmp)
         DEALLOCATE (eigenvalues)

         ! now obtain the initial condition in mos_old
         CALL cp_fm_to_fm(eigenvectors, mos_old(2*ispin-1)%matrix, nmo, 1, 1)
         CALL cp_gemm("N", "N", nao, nmo, nvirt, 1.0_dp, virtuals, momentum, 0.0_dp, mos_old(2*ispin)%matrix)

         CALL cp_fm_release(virtuals)
         CALL cp_fm_release(momentum)

         ! orthonormalize afterwards
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=S_inv_fm%matrix_struct%para_env, context=S_inv_fm%matrix_struct%context, &
                                  nrow_global=nmo, ncol_global=nmo)
         CALL cp_fm_create(oo_1, matrix_struct=fm_struct_tmp, name="oo_1")
         CALL cp_fm_create(oo_2, matrix_struct=fm_struct_tmp, name="oo_2")
         CALL cp_fm_struct_release(fm_struct_tmp)

         CALL cp_fm_create(mat_tmp, matrix_struct=mos_old(2*ispin-1)%matrix%matrix_struct, name="tmp_mat")
         ! get the complex overlap matrix
         ! x^T S x + y^T S y + i (-y^TS x+x^T S y)
         CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, mos_old(2*ispin-1)%matrix, &
                                      mat_tmp, ncol=nmo)

         CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mos_old(2*ispin-1)%matrix, mat_tmp, 0.0_dp, oo_1)
         CALL cp_gemm("T", "N", nmo, nmo, nao, -1.0_dp, mos_old(2*ispin)%matrix, mat_tmp, 0.0_dp, oo_2)

         CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, mos_old(2*ispin)%matrix, &
                                      mat_tmp, ncol=nmo)
         CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mos_old(2*ispin)%matrix, mat_tmp, 1.0_dp, oo_1)
         CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mos_old(2*ispin-1)%matrix, mat_tmp, 1.0_dp, oo_2)
         CALL cp_fm_release(mat_tmp)

         CALL cp_cfm_create(oo_c, oo_1%matrix_struct)
         CALL cp_cfm_create(oo_v, oo_1%matrix_struct)
         CALL cp_cfm_create(oo_vt, oo_1%matrix_struct)
         oo_c%local_data = CMPLX(oo_1%local_data, oo_2%local_data, KIND=dp)

         ! compute inv(sqrt(overlap))
         ALLOCATE (eigenvalues(nmo))
         ALLOCATE (eigenvalues_sqrt(nmo))
         CALL cp_cfm_heevd(oo_c, oo_v, eigenvalues)
         eigenvalues_sqrt = CMPLX(1.0_dp/SQRT(eigenvalues), 0.0_dp, dp)
         CALL cp_cfm_to_cfm(oo_v, oo_vt)
         CALL cp_cfm_column_scale(oo_v, eigenvalues_sqrt)
         DEALLOCATE (eigenvalues)
         DEALLOCATE (eigenvalues_sqrt)
         CALL cp_cfm_gemm('N', 'C', nmo, nmo, nmo, (1.0_dp, 0.0_dp), &
                          oo_v, oo_vt, (0.0_dp, 0.0_dp), oo_c)
         oo_1%local_data = REAL(oo_c%local_data, KIND=dp)
         oo_2%local_data = AIMAG(oo_c%local_data)
         CALL cp_cfm_release(oo_c)
         CALL cp_cfm_release(oo_v)
         CALL cp_cfm_release(oo_vt)

         ! use this to compute the orthonormal vectors
         CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mos_old(2*ispin-1)%matrix, oo_1, 0.0_dp, mos_new(2*ispin-1)%matrix)
         CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mos_old(2*ispin-1)%matrix, oo_2, 0.0_dp, mos_new(2*ispin)%matrix)

         CALL cp_gemm("N", "N", nao, nmo, nmo, -1.0_dp, mos_old(2*ispin)%matrix, oo_2, 0.0_dp, mos_old(2*ispin-1)%matrix)
         CALL cp_fm_scale_and_add(1.0_dp, mos_old(2*ispin-1)%matrix, 1.0_dp, mos_new(2*ispin-1)%matrix)

         CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mos_old(2*ispin)%matrix, oo_1, 1.0_dp, mos_new(2*ispin)%matrix)
         CALL cp_fm_to_fm(mos_new(2*ispin)%matrix, mos_old(2*ispin)%matrix)

         CALL cp_fm_release(oo_1)
         CALL cp_fm_release(oo_2)
      END DO

      CALL cp_fm_release(S_chol)
      CALL cp_fm_release(mat_ks)
      CALL cp_fm_release(eigenvectors)

!***************************************************************
!remove later
      CALL cp_fm_release(S_inv_fm)
      CALL cp_fm_release(tmpS)
!**************************************************************
      CALL timestop(handle)

   END SUBROUTINE apply_delta_pulse_periodic

! **************************************************************************************************
!> \brief applies exp(ikr) to the wavefunction.... stored in mos_old...
!> \param qs_env ...
!> \param mos_old ...
!> \param mos_new ...
!> \author Joost & Martin (2011)
! **************************************************************************************************

   SUBROUTINE apply_delta_pulse(qs_env, mos_old, mos_new)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_old, mos_new

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_delta_pulse', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: eigenvalues_sqrt
      INTEGER                                            :: handle, i, nao, nmo
      REAL(KIND=dp), DIMENSION(3)                        :: kvec
      REAL(kind=dp), DIMENSION(:), POINTER               :: eigenvalues
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_cfm_type), POINTER                         :: oo_c, oo_v, oo_vt
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: cosmat, S_inv, sinmat
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: mat_S, oo_1, oo_2, S_inv_fm, tmp
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(rt_prop_type), POINTER                        :: rtp

      NULLIFY (dft_control)

      CALL timeset(routineN, handle)

      ! we need the inverse overlap

      CALL get_qs_env(qs_env, &
                      mos=mos, &
                      rtp=rtp, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control)
      CALL get_rtp(rtp=rtp, S_inv=S_inv)

      CALL cp_fm_create(S_inv_fm, matrix_struct=rtp%ao_ao_fmstruct, name="tmp_mat")

      CALL cp_fm_create(tmp, matrix_struct=rtp%ao_ao_fmstruct, name="tmp_mat")

      CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, S_inv_fm)
      CALL cp_fm_cholesky_decompose(S_inv_fm)
      CALL cp_fm_cholesky_invert(S_inv_fm)
      CALL cp_fm_upper_to_full(S_inv_fm, tmp)

      CALL cp_fm_create(mat_S, matrix_struct=S_inv_fm%matrix_struct, name="mat_S")
      CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, mat_S)
      CALL cp_fm_upper_to_full(mat_S, tmp)

      CALL cp_fm_release(tmp)

      ! we need the berry matrix
      CALL get_qs_env(qs_env, cell=cell)

      ! direction ... unscaled, this will yield a exp(ikr) that is periodic with the cell
      kvec(:) = cell%h_inv(1, :)*dft_control%rtp_control%delta_pulse_direction(1)+ &
                cell%h_inv(2, :)*dft_control%rtp_control%delta_pulse_direction(2)+ &
                cell%h_inv(3, :)*dft_control%rtp_control%delta_pulse_direction(3)
      kvec = -kvec*twopi
      ! scaling will make the things not periodic with the cell, which would only be good for gas phase systems ?
      kvec(:) = dft_control%rtp_control%delta_pulse_scale*kvec

      ALLOCATE (cosmat, sinmat)
      CALL cp_dbcsr_init(cosmat)
      CALL cp_dbcsr_init(sinmat)
      CALL cp_dbcsr_copy(cosmat, matrix_s(1)%matrix, 'COS MOM')
      CALL cp_dbcsr_copy(sinmat, matrix_s(1)%matrix, 'SIN MOM')
      CALL build_berry_moment_matrix(qs_env, cosmat, sinmat, kvec)

      ! apply inv(S)*operator to C
      DO i = 1, SIZE(mos)
         CALL get_mo_set(mos(i)%mo_set, nao=nao, nmo=nmo)
         CALL cp_dbcsr_sm_fm_multiply(cosmat, mos(i)%mo_set%mo_coeff, mos_new(2*i-1)%matrix, ncol=nmo)
         CALL cp_dbcsr_sm_fm_multiply(sinmat, mos(i)%mo_set%mo_coeff, mos_new(2*i)%matrix, ncol=nmo)

         CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, S_inv_fm, mos_new(2*i-1)%matrix, 0.0_dp, mos_old(2*i-1)%matrix)
         CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, S_inv_fm, mos_new(2*i)%matrix, 0.0_dp, mos_old(2*i)%matrix)

         ! in a finite basis, unfortunately, inv(S)*operator is not unitary, so orthonormalize afterwards
         CALL cp_fm_struct_create(fm_struct_tmp, para_env=S_inv_fm%matrix_struct%para_env, context=S_inv_fm%matrix_struct%context, &
                                  nrow_global=nmo, ncol_global=nmo)
         CALL cp_fm_create(oo_1, matrix_struct=fm_struct_tmp, name="oo_1")
         CALL cp_fm_create(oo_2, matrix_struct=fm_struct_tmp, name="oo_2")
         CALL cp_fm_struct_release(fm_struct_tmp)

         CALL cp_fm_create(tmp, matrix_struct=mos_old(2*i-1)%matrix%matrix_struct, name="tmp_mat")
         ! get the complex overlap matrix
         ! x^T S x + y^T S y + i (-y^TS x+x^T S y)
         CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, mat_S, mos_old(2*i-1)%matrix, 0.0_dp, tmp)
         CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mos_old(2*i-1)%matrix, tmp, 0.0_dp, oo_1)
         CALL cp_gemm("T", "N", nmo, nmo, nao, -1.0_dp, mos_old(2*i)%matrix, tmp, 0.0_dp, oo_2)

         CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, mat_S, mos_old(2*i)%matrix, 0.0_dp, tmp)
         CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mos_old(2*i)%matrix, tmp, 1.0_dp, oo_1)
         CALL cp_gemm("T", "N", nmo, nmo, nao, 1.0_dp, mos_old(2*i-1)%matrix, tmp, 1.0_dp, oo_2)
         CALL cp_fm_release(tmp)

         CALL cp_cfm_create(oo_c, oo_1%matrix_struct)
         CALL cp_cfm_create(oo_v, oo_1%matrix_struct)
         CALL cp_cfm_create(oo_vt, oo_1%matrix_struct)
         oo_c%local_data = CMPLX(oo_1%local_data, oo_2%local_data, KIND=dp)

         ! compute inv(sqrt(overlap))
         ALLOCATE (eigenvalues(nmo))
         ALLOCATE (eigenvalues_sqrt(nmo))
         CALL cp_cfm_heevd(oo_c, oo_v, eigenvalues)
         eigenvalues_sqrt = CMPLX(1.0_dp/SQRT(eigenvalues), 0.0_dp, dp)
         CALL cp_cfm_to_cfm(oo_v, oo_vt)
         CALL cp_cfm_column_scale(oo_v, eigenvalues_sqrt)
         DEALLOCATE (eigenvalues)
         DEALLOCATE (eigenvalues_sqrt)
         CALL cp_cfm_gemm('N', 'C', nmo, nmo, nmo, (1.0_dp, 0.0_dp), &
                          oo_v, oo_vt, (0.0_dp, 0.0_dp), oo_c)
         oo_1%local_data = REAL(oo_c%local_data, KIND=dp)
         oo_2%local_data = AIMAG(oo_c%local_data)
         CALL cp_cfm_release(oo_c)
         CALL cp_cfm_release(oo_v)
         CALL cp_cfm_release(oo_vt)

         ! use this to compute the orthonormal vectors
         CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mos_old(2*i-1)%matrix, oo_1, 0.0_dp, mos_new(2*i-1)%matrix)
         CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mos_old(2*i-1)%matrix, oo_2, 0.0_dp, mos_new(2*i)%matrix)

         CALL cp_gemm("N", "N", nao, nmo, nmo, -1.0_dp, mos_old(2*i)%matrix, oo_2, 0.0_dp, mos_old(2*i-1)%matrix)
         CALL cp_fm_scale_and_add(1.0_dp, mos_old(2*i-1)%matrix, 1.0_dp, mos_new(2*i-1)%matrix)

         CALL cp_gemm("N", "N", nao, nmo, nmo, 1.0_dp, mos_old(2*i)%matrix, oo_1, 1.0_dp, mos_new(2*i)%matrix)
         CALL cp_fm_to_fm(mos_new(2*i)%matrix, mos_old(2*i)%matrix)

         CALL cp_fm_release(oo_1)
         CALL cp_fm_release(oo_2)
      END DO

      CALL cp_fm_release(mat_S)

      CALL cp_dbcsr_deallocate_matrix(cosmat)
      CALL cp_dbcsr_deallocate_matrix(sinmat)

!***************************************************************
!remove later
      CALL copy_fm_to_dbcsr(S_inv_fm, S_inv)
      CALL cp_dbcsr_filter(S_inv, rtp%filter_eps)
      CALL cp_fm_release(S_inv_fm)
!**************************************************************

      CALL timestop(handle)

   END SUBROUTINE apply_delta_pulse

END MODULE rt_delta_pulse
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for that prepare rtp and EMD
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************
MODULE rt_propagator_init

   USE cp_control_types,                ONLY: dft_control_type,&
                                              rtp_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_Set, cp_dbcsr_arnoldi_extremal, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix, cp_dbcsr_filter, cp_dbcsr_init, cp_dbcsr_multiply, &
        cp_dbcsr_p_type, cp_dbcsr_scale, cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_plus_fm_fm_t
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_scale,&
                                              cp_fm_upper_to_full
   USE cp_fm_diag,                      ONLY: cp_fm_syevd
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE dm_ls_scf_methods,               ONLY: density_matrix_trs4
   USE input_constants,                 ONLY: do_arnoldi,&
                                              do_bch,&
                                              do_cn,&
                                              do_em,&
                                              do_etrs,&
                                              do_pade,&
                                              do_taylor
   USE iterate_matrix,                  ONLY: matrix_sqrt_Newton_Schulz
   USE kinds,                           ONLY: dp
   USE matrix_exp,                      ONLY: get_nsquare_norder
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_mo_types,                     ONLY: mo_set_p_type
   USE rt_make_propagators,             ONLY: compute_exponential,&
                                              compute_exponential_sparse,&
                                              propagate_arnoldi
   USE rt_propagation_methods,          ONLY: calc_SinvH,&
                                              put_data_to_history,&
                                              s_matrices_create
   USE rt_propagation_types,            ONLY: get_rtp,&
                                              rt_prop_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_propagator_init'

   PUBLIC :: init_propagators, &
             rt_initialize_rho_from_ks, &
             rt_initialize_rho_from_mos

CONTAINS

! **************************************************************************************************
!> \brief prepares the initial matrices for the propagators
!> \param qs_env ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE init_propagators(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'init_propagators', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, imat, unit_nr
      REAL(KIND=dp)                                      :: dt, prefac
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: exp_H_new, exp_H_old, matrix_ks, &
                                                            matrix_ks_im, propagator_matrix, &
                                                            rho_old, s_mat
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new, mos_next, mos_old
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CALL get_qs_env(qs_env, &
                      rtp=rtp, &
                      dft_control=dft_control, &
                      matrix_s=s_mat, &
                      matrix_ks=matrix_ks, &
                      matrix_ks_im=matrix_ks_im)

      rtp_control => dft_control%rtp_control
      CALL get_rtp(rtp=rtp, exp_H_old=exp_H_old, exp_H_new=exp_H_new, &
                   propagator_matrix=propagator_matrix, dt=dt)
      CALL s_matrices_create(s_mat, rtp)
      CALL calc_SinvH(rtp, matrix_ks, matrix_ks_im, rtp_control)
      DO i = 1, SIZE(exp_H_old)
         CALL cp_dbcsr_copy(exp_H_old(i)%matrix, exp_H_new(i)%matrix)
      END DO
      ! use the fact that CN propagator is a first order pade approximation on the EM propagator
      IF (rtp_control%propagator == do_cn) THEN
         rtp%orders(1, :) = 0; rtp%orders(2, :) = 1; rtp_control%mat_exp = do_pade; rtp_control%propagator = do_em
      ELSE IF (rtp_control%mat_exp == do_pade .OR. rtp_control%mat_exp == do_taylor) THEN
         IF (rtp%linear_scaling) THEN
            CALL get_maxabs_eigval_sparse(rtp, s_mat, matrix_ks, rtp_control)
         ELSE
            CALL get_maxabs_eigval(rtp, s_mat, matrix_ks, rtp_control)
         END IF
      END IF
      IF (rtp_control%mat_exp == do_pade .AND. rtp%linear_scaling) THEN
         ! get a useful output_unit
         logger => cp_get_default_logger()
         IF (logger%para_env%mepos == logger%para_env%source) THEN
            unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
            WRITE (unit_nr, *) "linear_scaling currently does not support pade exponentials, switching to taylor"
         ENDIF
         rtp_control%mat_exp = do_taylor
      ENDIF
      ! We have no clue yet about next H so we use initial H for t and t+dt
      ! Due to different nature of the propagator the prefactor has to be adopted
      SELECT CASE (rtp_control%propagator)
      CASE (do_etrs)
         prefac = -0.5_dp*dt
      CASE (do_em)
         prefac = -1.0_dp*dt
      END SELECT
      DO imat = 1, SIZE(exp_H_new)
         CALL cp_dbcsr_copy(propagator_matrix(imat)%matrix, exp_H_new(imat)%matrix)
         CALL cp_dbcsr_scale(propagator_matrix(imat)%matrix, prefac)
      END DO

      ! For ETRS this bit could be avoided but it drastically simplifies the workflow afterwards.
      ! If we compute the half propagated mos/exponential already here, we ensure everything is done
      ! with the correct S matrix and all information as during RTP/EMD are computed.
      ! Therefore we might accept to compute an unnesscesary expoenential but understand the code afterwards
      IF (rtp_control%propagator == do_etrs) THEN
         IF (rtp_control%mat_exp == do_arnoldi) THEN
            rtp%iter = 0
            CALL propagate_arnoldi(rtp, rtp_control)
            CALL get_rtp(rtp=rtp, mos_new=mos_new, mos_next=mos_next)
            DO imat = 1, SIZE(mos_new)
               CALL cp_fm_to_fm(mos_new(imat)%matrix, mos_next(imat)%matrix)
            END DO
         ELSEIF (rtp_control%mat_exp == do_bch) THEN
         ELSE
            IF (rtp%linear_scaling) THEN
               CALL compute_exponential_sparse(exp_H_new, propagator_matrix, rtp_control, rtp)
            ELSE
               CALL compute_exponential(exp_H_new, propagator_matrix, rtp_control, rtp)
            END IF
            DO imat = 1, SIZE(exp_H_new)
               CALL cp_dbcsr_copy(exp_H_old(imat)%matrix, exp_H_new(imat)%matrix)
            END DO
         END IF
      END IF

      IF (rtp%linear_scaling) THEN
         CALL get_rtp(rtp=rtp, rho_old=rho_old)
      ELSE
         CALL get_rtp(rtp=rtp, mos_old=mos_old)
      ENDIF
      CALL put_data_to_history(rtp, mos=mos_old, s_mat=s_mat, ihist=1, rho=rho_old)

   END SUBROUTINE init_propagators

! **************************************************************************************************
!> \brief gets an estimate for the 2-norm of KS (diagnaliztion of KS) and
!>        calculates the order and number of squaring steps for Taylor or
!>        Pade matrix exponential
!> \param rtp ...
!> \param s_mat ...
!> \param matrix_ks ...
!> \param rtp_control ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE get_maxabs_eigval(rtp, s_mat, matrix_ks, rtp_control)
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: s_mat, matrix_ks
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'get_maxabs_eigval', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, ispin, method, ndim
      LOGICAL                                            :: emd
      REAL(dp)                                           :: max_eval, min_eval, norm2, scale, t
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: eigval_H
      TYPE(cp_dbcsr_type), POINTER                       :: S_inv
      TYPE(cp_fm_type), POINTER                          :: eigvec_H, H_fm, S_half, S_inv_fm, &
                                                            S_minus_half, tmp, tmp_mat_H

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, S_inv=S_inv, dt=t)

      CALL cp_fm_create(S_inv_fm, &
                        matrix_struct=rtp%ao_ao_fmstruct, &
                        name="S_inv")
      CALL copy_dbcsr_to_fm(S_inv, S_inv_fm)

      CALL cp_fm_create(S_half, &
                        matrix_struct=rtp%ao_ao_fmstruct, &
                        name="S_half")

      CALL cp_fm_create(S_minus_half, &
                        matrix_struct=rtp%ao_ao_fmstruct, &
                        name="S_minus_half")

      CALL cp_fm_create(H_fm, &
                        matrix_struct=rtp%ao_ao_fmstruct, &
                        name="RTP_H_FM")

      CALL cp_fm_create(tmp_mat_H, &
                        matrix_struct=rtp%ao_ao_fmstruct, &
                        name="TMP_H")

      ndim = S_inv_fm%matrix_struct%nrow_global
      scale = 1.0_dp
      IF (rtp_control%propagator == do_etrs) scale = 2.0_dp
      t = -t/scale

      ! Create the overlap matrices

      CALL cp_fm_create(tmp, &
                        matrix_struct=rtp%ao_ao_fmstruct, &
                        name="tmp_mat")

      CALL cp_fm_create(eigvec_H, &
                        matrix_struct=rtp%ao_ao_fmstruct, &
                        name="tmp_EVEC")

      ALLOCATE (eigval_H(ndim))
      CALL copy_dbcsr_to_fm(s_mat(1)%matrix, tmp)
      CALL cp_fm_upper_to_full(tmp, eigvec_H)

      CALL cp_fm_syevd(tmp, eigvec_H, eigval_H)

      eigval_H(:) = one/eigval_H(:)
      CALL backtransform_matrix(eigval_H, eigvec_H, S_inv_fm)
      eigval_H(:) = SQRT(eigval_H(:))
      CALL backtransform_matrix(eigval_H, eigvec_H, S_minus_half)
      eigval_H(:) = one/eigval_H(:)
      CALL backtransform_matrix(eigval_H, eigvec_H, S_half)
      CALL cp_fm_release(eigvec_H)
      CALL cp_fm_release(tmp)

      IF (rtp_control%mat_exp == do_taylor) method = 1
      IF (rtp_control%mat_exp == do_pade) method = 2
      emd = (.NOT. rtp_control%fixed_ions)

      DO ispin = 1, SIZE(matrix_ks)

         CALL copy_dbcsr_to_fm(matrix_ks(ispin)%matrix, H_fm)
         CALL cp_fm_upper_to_full(H_fm, tmp_mat_H)
         CALL cp_fm_scale(t, H_fm)

         CALL cp_gemm("N", "N", ndim, ndim, ndim, one, H_fm, S_minus_half, zero, &
                      tmp_mat_H)
         CALL cp_gemm("N", "N", ndim, ndim, ndim, one, S_minus_half, tmp_mat_H, zero, &
                      H_fm)

         CALL cp_fm_syevd(H_fm, tmp_mat_H, eigval_H)
         min_eval = MINVAL(eigval_H)
         max_eval = MAXVAL(eigval_H)
         norm2 = 2.0_dp*MAX(ABS(min_eval), ABS(max_eval))
         CALL get_nsquare_norder(norm2, rtp%orders(1, ispin), rtp%orders(2, ispin), &
                                 rtp_control%eps_exp, method, emd)
      END DO

      DEALLOCATE (eigval_H)

      CALL copy_fm_to_dbcsr(S_inv_fm, S_inv)
      CALL cp_fm_release(S_inv_fm)
      CALL cp_fm_release(S_half)
      CALL cp_fm_release(S_minus_half)
      CALL cp_fm_release(H_fm)
      CALL cp_fm_release(tmp_mat_H)

      CALL timestop(handle)

   END SUBROUTINE get_maxabs_eigval

! **************************************************************************************************
!> \brief gets an estimate for the 2-norm of KS (diagnaliztion of KS) and
!>        calculates the order and number of squaring steps for Taylor or
!>        Pade matrix exponential. Based on the full matrix code.
!> \param rtp ...
!> \param s_mat ...
!> \param matrix_ks ...
!> \param rtp_control ...
!> \author Samuel Andermatt (02.14)
! **************************************************************************************************

   SUBROUTINE get_maxabs_eigval_sparse(rtp, s_mat, matrix_ks, rtp_control)
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: s_mat, matrix_ks
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'get_maxabs_eigval_sparse', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, ispin, method
      LOGICAL                                            :: converged, emd
      REAL(dp)                                           :: max_ev, min_ev, norm2, scale, t
      TYPE(cp_dbcsr_type), POINTER                       :: s_half, s_minus_half, tmp, tmp2

      CALL timeset(routineN, handle)

      CALL get_rtp(rtp=rtp, dt=t)

      NULLIFY (s_half)
      ALLOCATE (s_half)
      CALL cp_dbcsr_init(s_half)
      CALL cp_dbcsr_create(s_half, template=s_mat(1)%matrix)
      NULLIFY (s_minus_half)
      ALLOCATE (s_minus_half)
      CALL cp_dbcsr_init(s_minus_half)
      CALL cp_dbcsr_create(s_minus_half, template=s_mat(1)%matrix)
      NULLIFY (tmp)
      ALLOCATE (tmp)
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=s_mat(1)%matrix, matrix_type="N")
      NULLIFY (tmp2)
      ALLOCATE (tmp2)
      CALL cp_dbcsr_init(tmp2)
      CALL cp_dbcsr_create(tmp2, template=s_mat(1)%matrix, matrix_type="N")
      scale = 1.0_dp
      IF (rtp_control%propagator == do_etrs) scale = 2.0_dp
      t = -t/scale
      emd = (.NOT. rtp_control%fixed_ions)

      IF (rtp_control%mat_exp == do_taylor) method = 1
      IF (rtp_control%mat_exp == do_pade) method = 2
      CALL matrix_sqrt_Newton_Schulz(s_half, s_minus_half, s_mat(1)%matrix, rtp%filter_eps, &
                                     rtp%newton_schulz_order, rtp%lanzcos_threshold, rtp%lanzcos_max_iter)
      DO ispin = 1, SIZE(matrix_ks)
         CALL cp_dbcsr_multiply("N", "N", t, matrix_ks(ispin)%matrix, s_minus_half, zero, tmp, &
                                filter_eps=rtp%filter_eps)
         CALL cp_dbcsr_multiply("N", "N", one, s_minus_half, tmp, zero, tmp2, &
                                filter_eps=rtp%filter_eps)
         CALL cp_dbcsr_arnoldi_extremal(tmp2, max_ev, min_ev, threshold=rtp%lanzcos_threshold, &
                                        max_iter=rtp%lanzcos_max_iter, converged=converged)
         norm2 = 2.0_dp*MAX(ABS(min_ev), ABS(max_ev))
         CALL get_nsquare_norder(norm2, rtp%orders(1, ispin), rtp%orders(2, ispin), &
                                 rtp_control%eps_exp, method, emd)
      END DO

      CALL cp_dbcsr_deallocate_matrix(s_half)
      CALL cp_dbcsr_deallocate_matrix(s_minus_half)
      CALL cp_dbcsr_deallocate_matrix(tmp)
      CALL cp_dbcsr_deallocate_matrix(tmp2)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Is still left from diagonalization, should be removed later but is
!>  still needed for the guess for the pade/Taylor method
!> \param Eval ...
!> \param eigenvec ...
!> \param matrix ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE backtransform_matrix(Eval, eigenvec, matrix)

      REAL(dp), DIMENSION(:), INTENT(in)                 :: Eval
      TYPE(cp_fm_type), POINTER                          :: eigenvec, matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'backtransform_matrix', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: handle, i, j, l, ncol_local, ndim, &
                                                            nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      TYPE(cp_fm_type), POINTER                          :: tmp

      CALL timeset(routineN, handle)
      CALL cp_fm_create(tmp, &
                        matrix_struct=matrix%matrix_struct, &
                        name="TMP_BT")
      CALL cp_fm_get_info(matrix, nrow_local=nrow_local, ncol_local=ncol_local, &
                          row_indices=row_indices, col_indices=col_indices)

      ndim = matrix%matrix_struct%nrow_global

      CALL cp_fm_set_all(tmp, zero, zero)
      DO i = 1, ncol_local
         l = col_indices(i)
         DO j = 1, nrow_local
            tmp%local_data(j, i) = eigenvec%local_data(j, i)*Eval(l)
         END DO
      END DO
      CALL cp_gemm("N", "T", ndim, ndim, ndim, one, tmp, eigenvec, zero, &
                   matrix)

      CALL cp_fm_release(tmp)
      CALL timestop(handle)

   END SUBROUTINE backtransform_matrix

! **************************************************************************************************
!> \brief Computes the density matrix starting from the kohn-sham matrix
!> \param rtp ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param nelectron_spin ...
!> \param orthonormal ...
!> \author Samuel Andermatt (03.14)
! **************************************************************************************************

   SUBROUTINE rt_initialize_rho_from_ks(rtp, matrix_ks, matrix_s, nelectron_spin, orthonormal)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      INTEGER, DIMENSION(2)                              :: nelectron_spin
      LOGICAL                                            :: orthonormal

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_initialize_rho_from_ks', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: i, ispin, nelectron_spin_real, re
      REAL(KIND=dp)                                      :: e_homo, e_lumo, e_mu
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_new, rho_old
      TYPE(cp_dbcsr_type), POINTER                       :: S_half, S_minus_half, tmp

      CALL get_rtp(rtp=rtp, rho_old=rho_old, S_half=S_half, S_minus_half=S_minus_half, rho_new=rho_new)
      !The inverse of S will be calculated twice in the first step, but preventing that would make things more complicated
      CALL matrix_sqrt_Newton_Schulz(S_half, S_minus_half, matrix_s(1)%matrix, rtp%filter_eps, &
                                     rtp%newton_schulz_order, rtp%lanzcos_threshold, rtp%lanzcos_max_iter)
      NULLIFY (tmp)
      ALLOCATE (tmp)
      CALL cp_dbcsr_init(tmp)
      CALL cp_dbcsr_create(tmp, template=matrix_ks(1)%matrix, matrix_type="N")

      e_homo = 0.0_dp
      e_lumo = 0.0_dp
      e_mu = 0.0_dp

      DO ispin = 1, SIZE(matrix_ks)
         re = 2*ispin-1
         nelectron_spin_real = nelectron_spin(ispin)
         IF (SIZE(matrix_ks) == 1) nelectron_spin_real = nelectron_spin_real/2
         CALL density_matrix_trs4(rho_old(re)%matrix, matrix_ks(ispin)%matrix, S_minus_half, &
                                  nelectron_spin_real, rtp%filter_eps, e_homo, e_lumo, e_mu, .FALSE., tmp, &
                                  max_iter_lanczos=rtp%lanzcos_max_iter, eps_lanczos=rtp%lanzcos_threshold)
         IF (SIZE(matrix_ks) == 1) CALL cp_dbcsr_scale(rho_old(re)%matrix, 2.0_dp)
         IF (orthonormal) THEN
            CALL cp_dbcsr_multiply("N", "N", one, S_half, rho_old(re)%matrix, zero, tmp, filter_eps=rtp%filter_eps_small)
            CALL cp_dbcsr_multiply("N", "N", one, tmp, S_half, zero, rho_old(re)%matrix, filter_eps=rtp%filter_eps_small)
         ENDIF
         CALL cp_dbcsr_filter(rho_old(re)%matrix, eps=rtp%filter_eps)
      END DO

      DO i = 1, SIZE(rho_new)
         CALL cp_dbcsr_copy(rho_new(i)%matrix, rho_old(i)%matrix)
      ENDDO

      CALL cp_dbcsr_deallocate_matrix(tmp)
   END SUBROUTINE rt_initialize_rho_from_ks

! **************************************************************************************************
!> \brief Computes the density matrix from the mos
!> \param rtp ...
!> \param mos ...
!> \param matrix_s ...
!> \param orthonormal ...
!> \author Samuel Andermatt (08.15)
! **************************************************************************************************

   SUBROUTINE rt_initialize_rho_from_mos(rtp, mos, matrix_s, orthonormal)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      LOGICAL                                            :: orthonormal

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_initialize_rho_from_mos', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: ispin, ncol, re
      REAL(KIND=dp)                                      :: alpha
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_new, rho_old
      TYPE(cp_dbcsr_type), POINTER                       :: S_half, S_minus_half, tmp

      CALL get_rtp(rtp=rtp, rho_old=rho_old, rho_new=rho_new)
      IF (orthonormal) THEN
         CALL get_rtp(rtp=rtp, S_half=S_half, S_minus_half=S_minus_half)
         NULLIFY (tmp)
         ALLOCATE (tmp)
         CALL cp_dbcsr_init(tmp)
         CALL cp_dbcsr_create(tmp, template=rho_old(1)%matrix, matrix_type="N")
         !The inverse of S will be calculated twice in the first step, but preventing that would make things more complicated
         CALL matrix_sqrt_Newton_Schulz(S_half, S_minus_half, matrix_s(1)%matrix, rtp%filter_eps, &
                                        rtp%newton_schulz_order, rtp%lanzcos_threshold, rtp%lanzcos_max_iter)
      ENDIF

      DO ispin = 1, SIZE(mos)
         re = 2*ispin-1
         alpha = 3.0_dp-REAL(SIZE(mos), dp)
         CALL cp_dbcsr_set(rho_old(re)%matrix, 0.0_dp)
         CALL cp_fm_get_info(mos(ispin)%mo_set%mo_coeff, ncol_global=ncol)
         CALL cp_fm_column_scale(mos(ispin)%mo_set%mo_coeff, mos(ispin)%mo_set%occupation_numbers/alpha)
         CALL cp_dbcsr_plus_fm_fm_t(sparse_matrix=rho_old(re)%matrix, &
                                    matrix_v=mos(ispin)%mo_set%mo_coeff, ncol=ncol, alpha=alpha, keep_sparsity=.FALSE.)
         IF (orthonormal) THEN
            CALL cp_dbcsr_multiply("N", "N", one, S_half, rho_old(re)%matrix, zero, tmp, filter_eps=rtp%filter_eps_small)
            CALL cp_dbcsr_multiply("N", "N", one, tmp, S_half, zero, rho_old(re)%matrix, filter_eps=rtp%filter_eps_small)
         ENDIF
         CALL cp_dbcsr_filter(rho_old(re)%matrix, eps=rtp%filter_eps)
         CALL cp_dbcsr_copy(rho_new(re)%matrix, rho_old(re)%matrix)
      END DO

      IF (orthonormal) CALL cp_dbcsr_deallocate_matrix(tmp)

   END SUBROUTINE rt_initialize_rho_from_mos

END MODULE rt_propagator_init
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for the real time propagation.
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

MODULE rt_propagation
   USE cp_control_types,                ONLY: dft_control_type,&
                                              rtp_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_external_control,             ONLY: external_control
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_rm_iter_level
   USE efield_utils,                    ONLY: calculate_ecore_efield
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: real_time_propagation,&
                                              use_restart_wfn,&
                                              use_rt_restart,&
                                              use_scf_wfn
   USE input_cp2k_restarts,             ONLY: write_restart
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_walltime
   USE md_environment_types,            ONLY: md_environment_type
   USE qs_dftb_matrices,                ONLY: build_dftb_overlap
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_energy_utils,                 ONLY: qs_energies_init
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_external_potential,           ONLY: external_c_potential,&
                                              external_e_potential
   USE qs_ks_methods,                   ONLY: qs_ks_update_qs_env
   USE qs_ks_types,                     ONLY: qs_ks_did_change
   USE qs_matrix_pools,                 ONLY: mpools_get
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              init_mo_set,&
                                              mo_set_p_type
   USE rt_delta_pulse,                  ONLY: apply_delta_pulse,&
                                              apply_delta_pulse_periodic
   USE rt_hfx_utils,                    ONLY: rtp_hfx_rebuild
   USE rt_propagation_methods,          ONLY: propagation_step
   USE rt_propagation_output,           ONLY: rt_prop_output
   USE rt_propagation_types,            ONLY: get_rtp,&
                                              rt_prop_create,&
                                              rt_prop_type,&
                                              rtp_history_create
   USE rt_propagation_utils,            ONLY: calc_S_derivs,&
                                              calc_update_rho,&
                                              calc_update_rho_sparse,&
                                              get_restart_wfn
   USE rt_propagator_init,              ONLY: init_propagators,&
                                              rt_initialize_rho_from_ks,&
                                              rt_initialize_rho_from_mos
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_propagation'

   PUBLIC :: rt_prop_setup, &
             rt_write_input_restart

CONTAINS

! **************************************************************************************************
!> \brief creates rtp_type, gets the initial state, either by reading MO's
!>        from file or calling SCF run
!> \param force_env ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE rt_prop_setup(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_prop_setup', routineP = moduleN//':'//routineN

      INTEGER                                            :: aspc_order
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control
      TYPE(section_vals_type), POINTER                   :: hfx_sections, input, ls_scf_section, &
                                                            md_section, motion_section

      NULLIFY (qs_env, rtp_control, dft_control)

      CALL force_env_get(force_env=force_env, qs_env=qs_env, globenv=globenv)
      CALL get_qs_env(qs_env, dft_control=dft_control, energy=energy)
      rtp_control => dft_control%rtp_control

      ! Takes care that an initial wavefunction/density is available
      ! Can either be by performing an scf loop or reading a restart
      CALL rt_initial_guess(qs_env, force_env, rtp_control)

      ! Initializes the extrapolation
      CALL get_qs_env(qs_env=qs_env, rtp=rtp, input=input)
      aspc_order = rtp_control%aspc_order
      CALL rtp_history_create(rtp, aspc_order)

      ! Reads the simulation parameters from the input
      motion_section => section_vals_get_subs_vals(force_env%root_section, "MOTION")
      md_section => section_vals_get_subs_vals(motion_section, "MD")
      hfx_sections => section_vals_get_subs_vals(force_env%root_section, "FORCE_EVAL%DFT%XC%HF")
      CALL section_vals_val_get(md_section, "TIMESTEP", r_val=qs_env%rtp%dt)
      CALL section_vals_val_get(md_section, "STEP_START_VAL", i_val=qs_env%rtp%i_start)
      CALL section_vals_val_get(md_section, "STEPS", i_val=rtp%nsteps)

      ls_scf_section => section_vals_get_subs_vals(input, "DFT%LS_SCF")
      CALL section_vals_val_get(ls_scf_section, "EPS_FILTER", r_val=rtp%filter_eps)
      IF (.NOT. qs_env%rtp%linear_scaling) rtp%filter_eps = 0.0_dp
      IF (rtp_control%acc_ref < 1) rtp_control%acc_ref = 1
      rtp%filter_eps_small = rtp%filter_eps/rtp_control%acc_ref
      CALL section_vals_val_get(ls_scf_section, "EPS_LANCZOS", r_val=rtp%lanzcos_threshold)
      CALL section_vals_val_get(ls_scf_section, "MAX_ITER_LANCZOS", i_val=rtp%lanzcos_max_iter)
      CALL section_vals_val_get(ls_scf_section, "SIGN_SQRT_ORDER", i_val=rtp%newton_schulz_order)
      CALL section_vals_get(hfx_sections, explicit=rtp%do_hfx)

      ! Hmm, not really like to initialize with the structure of S but I reckon it is
      ! done everywhere like this
      IF (rtp%do_hfx) &
         CALL rtp_hfx_rebuild(qs_env)

      CALL init_propagation_run(qs_env)
      IF (.NOT. rtp_control%fixed_ions) THEN
         !derivativs of the overlap needed for EMD
         CALL calc_S_derivs(qs_env)
         ! a bit hidden, but computes SinvH and SinvB (calc_SinvH for CN,EM and ARNOLDI)
         ! make_etrs_exp in case of ETRS in combination with TAYLOR and PADE
      END IF
      CALL init_propagators(qs_env)
      IF (rtp_control%fixed_ions) THEN
         CALL run_propagation(qs_env, force_env, globenv)
      ELSE
         rtp_control%initial_step = .TRUE.
         CALL force_env_calc_energy_force(force_env, calc_force=.TRUE.)
         rtp_control%initial_step = .FALSE.
         rtp%energy_old = energy%total
      END IF

   END SUBROUTINE rt_prop_setup

! **************************************************************************************************
!> \brief calculates the matrices needed in the first step of EMD/RTP
!> \param qs_env ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE init_propagation_run(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      REAL(KIND=dp), PARAMETER                           :: zero = 0.0_dp

      INTEGER                                            :: i
      INTEGER, DIMENSION(2)                              :: nelectron_spin
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mos_new, mos_old
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      NULLIFY (matrix_s, dft_control)

      CALL get_qs_env(qs_env, &
                      rtp=rtp, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control)
      rtp_control => dft_control%rtp_control

      IF (.NOT. rtp%linear_scaling) THEN
         CALL get_qs_env(qs_env, mos=mos)
         CALL get_rtp(rtp=rtp, mos_old=mos_old, mos_new=mos_new)
         IF (rtp_control%initial_wfn == use_scf_wfn) THEN
            IF (rtp_control%apply_delta_pulse) THEN
               IF (dft_control%qs_control%dftb) &
                  CALL build_dftb_overlap(qs_env, 1, matrix_s)
               IF (rtp_control%periodic) THEN
                  CALL apply_delta_pulse_periodic(qs_env, mos_old, mos_new)
               ELSE
                  CALL apply_delta_pulse(qs_env, mos_old, mos_new)
               ENDIF
            ELSE
               DO i = 1, SIZE(mos)
                  CALL cp_fm_to_fm(mos(i)%mo_set%mo_coeff, mos_old(2*i-1)%matrix)
                  CALL cp_fm_set_all(mos_old(2*i)%matrix, zero, zero)
               END DO
            ENDIF
         END IF

         DO i = 1, SIZE(mos_old)
            CALL cp_fm_to_fm(mos_old(i)%matrix, mos_new(i)%matrix)
         END DO
         CALL calc_update_rho(qs_env)
      ELSE
         IF (qs_env%rtp%linear_scaling .AND. rtp_control%initial_wfn == use_scf_wfn) THEN
            CALL get_qs_env(qs_env, &
                            matrix_ks=matrix_ks, &
                            mos=mos, &
                            nelectron_spin=nelectron_spin)
            IF (ASSOCIATED(mos)) THEN
               !The wavefunction was minimized by an mo based algorith. P is therefore calculated from the mos
               CALL rt_initialize_rho_from_mos(rtp, mos, matrix_s, rtp_control%orthonormal)
            ELSE
               !The wavefunction was minimized using a linear scaling method. Because the rho in the qs_env has the sparsity of S
               !the matrix P needs to be recalculated from H
               CALL rt_initialize_rho_from_ks(rtp, matrix_ks, matrix_s, nelectron_spin, rtp_control%orthonormal)
            ENDIF
         ENDIF
         CALL calc_update_rho_sparse(qs_env)
      END IF

      CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE.)

   END SUBROUTINE init_propagation_run

! **************************************************************************************************
!> \brief performes the real RTP run, gets information from MD section
!>        uses MD as iteration level
!> \param qs_env ...
!> \param force_env ...
!> \param globenv ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE run_propagation(qs_env, force_env, globenv)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'run_propagation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: aspc_order, handle, i_iter, i_step, &
                                                            max_iter, max_steps
      LOGICAL                                            :: should_stop
      REAL(Kind=dp)                                      :: eps_ener, time_iter_start, &
                                                            time_iter_stop, used_time
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      should_stop = .FALSE.
      CALL timeset(routineN, handle)
      NULLIFY (logger, dft_control, energy, rtp, rtp_control)
      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, rtp=rtp, energy=energy)

      rtp_control => dft_control%rtp_control
      max_steps = rtp%nsteps
      max_iter = rtp_control%max_iter
      eps_ener = rtp_control%eps_ener

      aspc_order = rtp_control%aspc_order

      rtp%energy_old = energy%total
      time_iter_start = m_walltime()
      CALL cp_add_iter_level(logger%iter_info, "MD")
      CALL cp_iterate(logger%iter_info, iter_nr=0)
      DO i_step = rtp%i_start+1, max_steps
         energy%efield_core = 0.0_dp
         qs_env%sim_time = REAL(i_step, dp)*rtp%dt
         qs_env%sim_step = i_step
         rtp%istep = i_step-rtp%i_start
         CALL calculate_ecore_efield(qs_env, .FALSE.)
         CALL external_c_potential(qs_env, calculate_forces=.FALSE.)
         CALL external_e_potential(qs_env)
         CALL cp_iterate(logger%iter_info, last=(i_step == max_steps), iter_nr=i_step)
         rtp%converged = .FALSE.
         DO i_iter = 1, max_iter
            IF (i_step == rtp%i_start+1 .AND. i_iter == 2 .AND. rtp_control%hfx_redistribute) &
               CALL qs_ks_did_change(qs_env%ks_env, s_mstruct_changed=.TRUE.)
            rtp%iter = i_iter
            CALL propagation_step(qs_env, rtp, rtp_control)
            CALL qs_ks_update_qs_env(qs_env, calculate_forces=.FALSE.)
            rtp%energy_new = energy%total
            IF (rtp%converged) EXIT
            CALL rt_prop_output(qs_env, real_time_propagation, rtp%delta_iter)
         END DO
         IF (rtp%converged) THEN
            CALL external_control(should_stop, "MD", globenv=globenv)
            IF (should_stop) CALL cp_iterate(logger%iter_info, last=.TRUE., iter_nr=i_step)
            time_iter_stop = m_walltime()
            used_time = time_iter_stop-time_iter_start
            time_iter_start = time_iter_stop
            CALL rt_prop_output(qs_env, real_time_propagation, delta_iter=rtp%delta_iter, used_time=used_time)
            CALL rt_write_input_restart(force_env=force_env)
            IF (should_stop) EXIT
         ELSE
            EXIT
         END IF
      END DO
      CALL cp_rm_iter_level(logger%iter_info, "MD")

      IF (.NOT. rtp%converged) &
         CALL cp_abort(cp__l("motion/rt_propagation.F",322), "propagation did not converge, "// &
                       "either increase MAX_ITER or use a smaller TIMESTEP")

      CALL timestop(handle)

   END SUBROUTINE run_propagation

! **************************************************************************************************
!> \brief overwrites some values in the input file such that the .restart
!>        file will contain the appropriate information
!> \param md_env ...
!> \param force_env ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE rt_write_input_restart(md_env, force_env)
      TYPE(md_environment_type), OPTIONAL, POINTER       :: md_env
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_write_input_restart', &
         routineP = moduleN//':'//routineN

      TYPE(section_vals_type), POINTER                   :: motion_section, root_section, rt_section

      root_section => force_env%root_section
      motion_section => section_vals_get_subs_vals(root_section, "MOTION")
      rt_section => section_vals_get_subs_vals(root_section, "FORCE_EVAL%DFT%REAL_TIME_PROPAGATION")
      CALL section_vals_val_set(rt_section, "INITIAL_WFN", i_val=use_rt_restart)
      ! coming from RTP
      IF (.NOT. PRESENT(md_env)) THEN
         CALL section_vals_val_set(motion_section, "MD%STEP_START_VAL", i_val=force_env%qs_env%sim_step)
      ENDIF

      CALL write_restart(md_env=md_env, root_section=root_section)

   END SUBROUTINE rt_write_input_restart

! **************************************************************************************************
!> \brief Creates the initial electronic states and allocates the necessary
!>        matrices
!> \param qs_env ...
!> \param force_env ...
!> \param rtp_control ...
!> \author Florian Schiffmann (02.09)
! **************************************************************************************************

   SUBROUTINE rt_initial_guess(qs_env, force_env, rtp_control)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_initial_guess', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: homo, ispin
      LOGICAL                                            :: energy_consistency
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools_aux_fit
      TYPE(cp_fm_type), POINTER                          :: mo_coeff, mo_coeff_aux_fit
      TYPE(dft_control_type), POINTER                    :: dft_control

      NULLIFY (matrix_s, dft_control)
      CALL get_qs_env(qs_env, dft_control=dft_control)

      SELECT CASE (rtp_control%initial_wfn)
      CASE (use_scf_wfn)
         qs_env%sim_time = 0.0_dp
         qs_env%sim_step = 0
         energy_consistency = .TRUE.
         !in the linear scaling case we need a correct kohn-sham matrix, which we cannot get with consistent energies
         IF (rtp_control%linear_scaling) energy_consistency = .FALSE.
         CALL force_env_calc_energy_force(force_env, calc_force=.FALSE., &
                                          consistent_energies=energy_consistency)
         qs_env%run_rtp = .TRUE.
         ALLOCATE (qs_env%rtp)
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         CALL rt_prop_create(qs_env%rtp, qs_env%mos, qs_env%mpools, dft_control, matrix_s(1)%matrix, &
                             rtp_control%linear_scaling, qs_env%mos_aux_fit)

      CASE (use_restart_wfn, use_rt_restart)
         CALL qs_energies_init(qs_env, .FALSE.)
         IF (.NOT. rtp_control%linear_scaling .OR. rtp_control%initial_wfn == use_restart_wfn) THEN
            DO ispin = 1, SIZE(qs_env%mos)
               CALL get_mo_set(qs_env%mos(ispin)%mo_set, mo_coeff=mo_coeff, homo=homo)
               IF (.NOT. ASSOCIATED(mo_coeff)) THEN
                  CALL init_mo_set(qs_env%mos(ispin)%mo_set, &
                                   qs_env%mpools%ao_mo_fm_pools(ispin)%pool, &
                                   name="qs_env"//TRIM(ADJUSTL(cp_to_string(qs_env%id_nr)))// &
                                   "%mo"//TRIM(ADJUSTL(cp_to_string(ispin))))
               END IF
            END DO
            IF (dft_control%do_admm) THEN
               CALL mpools_get(qs_env%mpools_aux_fit, ao_mo_fm_pools=ao_mo_fm_pools_aux_fit)
               IF(.NOT.(ASSOCIATED(qs_env%mos_aux_fit)))CALL cp__a("motion/rt_propagation.F",415)
               DO ispin = 1, SIZE(qs_env%mos_aux_fit)
                  CALL get_mo_set(qs_env%mos_aux_fit(ispin)%mo_set, mo_coeff=mo_coeff_aux_fit, homo=homo)
                  IF (.NOT. ASSOCIATED(mo_coeff_aux_fit)) THEN
                     CALL init_mo_set(qs_env%mos_aux_fit(ispin)%mo_set, &
                                      ao_mo_fm_pools_aux_fit(ispin)%pool, &
                                      name="qs_env"//TRIM(ADJUSTL(cp_to_string(qs_env%id_nr)))// &
                                      "%mo_aux_fit"//TRIM(ADJUSTL(cp_to_string(ispin))))
                  END IF
               END DO
            END IF
         ENDIF
         ALLOCATE (qs_env%rtp)
         CALL get_qs_env(qs_env, matrix_s=matrix_s)
         CALL rt_prop_create(qs_env%rtp, qs_env%mos, qs_env%mpools, dft_control, matrix_s(1)%matrix, &
                             rtp_control%linear_scaling, qs_env%mos_aux_fit)
         CALL get_restart_wfn(qs_env)

         qs_env%run_rtp = .TRUE.
      END SELECT

   END SUBROUTINE rt_initial_guess

END MODULE rt_propagation
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utilities for Molecular Dynamics
!> \author Teodoro Laino [tlaino] - University of Zurich - 09.2007
! **************************************************************************************************
MODULE md_util

   USE input_cp2k_restarts,             ONLY: write_restart
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE md_energies,                     ONLY: md_write_output
   USE md_environment_types,            ONLY: md_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'md_util'

   PUBLIC :: md_output

CONTAINS

! **************************************************************************************************
!> \brief collects the part of the MD that, basically, does the output
!> \param md_env ...
!> \param md_section ...
!> \param root_section ...
!> \param forced_io ...
!> \par History
!>      03.2006 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE md_output(md_env, md_section, root_section, forced_io)
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(section_vals_type), POINTER                   :: md_section, root_section
      LOGICAL, INTENT(IN)                                :: forced_io

      CHARACTER(LEN=*), PARAMETER :: routineN = 'md_output', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: do_print
      TYPE(section_vals_type), POINTER                   :: print_section

      CALL timeset(routineN, handle)
      do_print = .TRUE.
      IF (forced_io) THEN
         print_section => section_vals_get_subs_vals(md_section, "PRINT")
         CALL section_vals_val_get(print_section, "FORCE_LAST", l_val=do_print)
      END IF
      IF (do_print) THEN
         ! Dumps all files related to the MD run
         CALL md_write_output(md_env)
         CALL write_restart(md_env=md_env, root_section=root_section)
      END IF
      CALL timestop(handle)

   END SUBROUTINE md_output

END MODULE md_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Callback used by global geometry optimization schemes
!> \author Ole Schuett
! **************************************************************************************************
MODULE glbopt_callback
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type,&
                                              pack_subsys_particles
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE kinds,                           ONLY: dp
   USE md_ener_types,                   ONLY: md_ener_type
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_environment_type
   USE mdctrl_types,                    ONLY: glbopt_mdctrl_data_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'glbopt_callback'

   PUBLIC :: glbopt_md_callback

CONTAINS

! **************************************************************************************************
!> \brief Callback used to hook into the main MD-loop.
!>        It recognizes and counts bumps in the potential energy.
!>        When MD_BUMPS_MAX is reached, the MD simulations is stoped.
!> \param mdctrl_data ...
!> \param md_env ...
!> \param should_stop ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE glbopt_md_callback(mdctrl_data, md_env, should_stop)
      TYPE(glbopt_mdctrl_data_type), POINTER             :: mdctrl_data
      TYPE(md_environment_type), POINTER                 :: md_env
      LOGICAL, INTENT(inout)                             :: should_stop

      CHARACTER(len=*), PARAMETER :: routineN = 'glbopt_md_callback', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iw, n_atoms
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: passed_minimum
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: positions
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(md_ener_type), POINTER                        :: md_ener

      IF(.NOT.(ASSOCIATED(mdctrl_data)))CALL cp__a("motion/glbopt_callback.F",57)
      IF(.NOT.(ASSOCIATED(md_env)))CALL cp__a("motion/glbopt_callback.F",58)

      iw = mdctrl_data%output_unit

      ! add new potential energy value to history
      NULLIFY (md_ener, itimes)
      CALL get_md_env(md_env=md_env, md_ener=md_ener, itimes=itimes, force_env=force_env)
      mdctrl_data%itimes = itimes

      mdctrl_data%epot_history(:) = EOSHIFT(mdctrl_data%epot_history, shift=-1)
      mdctrl_data%epot_history(1) = md_ener%epot

      ! check if we passed a minimum
      passed_minimum = .TRUE.
      DO i = 1, mdctrl_data%bump_steps_upwards
         IF (mdctrl_data%epot_history(i) <= mdctrl_data%epot_history(i+1)) &
            passed_minimum = .FALSE.
      END DO

      DO i = mdctrl_data%bump_steps_upwards+1, mdctrl_data%bump_steps_upwards+mdctrl_data%bump_steps_downwards
         IF (mdctrl_data%epot_history(i) >= mdctrl_data%epot_history(i+1)) &
            passed_minimum = .FALSE.
      END DO

      ! count the passed bumps and stop md_run when md_bumps_max is reached.
      IF (passed_minimum) &
         mdctrl_data%md_bump_counter = mdctrl_data%md_bump_counter+1

      IF (mdctrl_data%md_bump_counter >= mdctrl_data%md_bumps_max) THEN
         should_stop = .TRUE.
         IF (iw > 0) WRITE (iw, "(A)") " GLBOPT| Stopping MD because of MD_BUMPS_MAX."
      END IF

      CALL force_env_get(force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, natom=n_atoms)
      ALLOCATE (positions(3*n_atoms))
      CALL pack_subsys_particles(subsys, r=positions)

   END SUBROUTINE glbopt_md_callback

END MODULE glbopt_callback

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A common interface (wrapper) for a callback into the md_run loop.
!>        Currently this is only used by the glbopt machinery, but its meant
!>        to be extended if others need to controll the md_run loop, too.
!>
!> \par History
!>      11.2012 created [Ole]
!> \author Ole
! **************************************************************************************************
MODULE mdctrl_methods
   USE glbopt_callback,                 ONLY: glbopt_md_callback
   USE md_environment_types,            ONLY: md_environment_type
   USE mdctrl_types,                    ONLY: mdctrl_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: mdctrl_callback

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mdctrl_methods'

CONTAINS

! **************************************************************************************************
!> \brief This is called by md_run for each step during during its main-loop.
!> \param mdctrl data which is passed on to the wrapped client-routine
!> \param md_env contains the current state of the md_run
!> \param should_stop can be used to abort the md_run
! **************************************************************************************************
   SUBROUTINE mdctrl_callback(mdctrl, md_env, should_stop)
      TYPE(mdctrl_type), POINTER                         :: mdctrl
      TYPE(md_environment_type), POINTER                 :: md_env
      LOGICAL, INTENT(inout)                             :: should_stop

      CHARACTER(len=*), PARAMETER :: routineN = 'mdctrl_callback', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(md_env)))CALL cp__a("motion/mdctrl_methods.F",44)
      IF(.NOT.(ASSOCIATED(mdctrl)))CALL cp__a("motion/mdctrl_methods.F",45)

      IF (ASSOCIATED(mdctrl%glbopt)) THEN
         CALL glbopt_md_callback(mdctrl%glbopt, md_env, should_stop)

         !ELSE IF(ASSOCIATED(mdctrl%your_own_hook)) THEN ...

      ELSE
         CALL cp__b("motion/mdctrl_methods.F",53,"mdctrl_callback: No hook found.")
      ENDIF

   END SUBROUTINE mdctrl_callback

END MODULE mdctrl_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Perform a molecular dynamics (MD) run using QUICKSTEP
!> \par History
!>   - Added support for Langevin regions (2014/02/05, LT)
!> \author Matthias Krack (07.11.2002)
! **************************************************************************************************
MODULE md_run
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE averages_types,                  ONLY: average_quantities_type
   USE barostat_types,                  ONLY: barostat_type,&
                                              create_barostat_type,&
                                              release_barostat_type
   USE cell_types,                      ONLY: cell_type
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE free_energy_methods,             ONLY: free_energy_evaluate
   USE free_energy_types,               ONLY: fe_env_create,&
                                              free_energy_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: &
        ehrenfest, langevin_ensemble, npe_f_ensemble, npe_i_ensemble, &
        nph_uniaxial_damped_ensemble, nph_uniaxial_ensemble, npt_f_ensemble, npt_i_ensemble, &
        reftraj_ensemble
   USE input_cp2k_check,                ONLY: remove_restart_info
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_walltime
   USE md_ener_types,                   ONLY: create_md_ener,&
                                              md_ener_type,&
                                              release_md_ener
   USE md_energies,                     ONLY: initialize_md_ener,&
                                              md_ener_reftraj,&
                                              md_energy,&
                                              md_write_output
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_env_create,&
                                              md_env_release,&
                                              md_environment_type,&
                                              need_per_atom_wiener_process,&
                                              set_md_env
   USE md_util,                         ONLY: md_output
   USE md_vel_utils,                    ONLY: angvel_control,&
                                              comvel_control,&
                                              setup_velocities,&
                                              temperature_control
   USE mdctrl_methods,                  ONLY: mdctrl_callback
   USE mdctrl_types,                    ONLY: mdctrl_type
   USE metadynamics,                    ONLY: metadyn_finalise_plumed,&
                                              metadyn_forces,&
                                              metadyn_initialise_plumed,&
                                              metadyn_write_colvar
   USE metadynamics_types,              ONLY: set_meta_env
   USE particle_list_types,             ONLY: particle_list_type
   USE reftraj_types,                   ONLY: create_reftraj,&
                                              reftraj_type,&
                                              release_reftraj
   USE reftraj_util,                    ONLY: initialize_reftraj,&
                                              write_output_reftraj
   USE rt_propagation,                  ONLY: rt_prop_setup
   USE simpar_methods,                  ONLY: read_md_section
   USE simpar_types,                    ONLY: create_simpar_type,&
                                              release_simpar_type,&
                                              simpar_type
   USE thermal_region_types,            ONLY: release_thermal_regions,&
                                              thermal_regions_type
   USE thermal_region_utils,            ONLY: create_thermal_regions,&
                                              print_thermal_regions_langevin
   USE thermostat_methods,              ONLY: create_thermostats
   USE thermostat_types,                ONLY: release_thermostats,&
                                              thermostats_type
   USE velocity_verlet_control,         ONLY: velocity_verlet
   USE virial_methods,                  ONLY: virial_evaluate
   USE virial_types,                    ONLY: virial_type
   USE wiener_process,                  ONLY: create_wiener_process,&
                                              create_wiener_process_cv
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'md_run'

   PUBLIC :: qs_mol_dyn

CONTAINS

! **************************************************************************************************
!> \brief Main driver module for Molecular Dynamics
!> \param force_env ...
!> \param globenv ...
!> \param averages ...
!> \param rm_restart_info ...
!> \param hmc_e_initial ...
!> \param hmc_e_final ...
!> \param mdctrl ...
! **************************************************************************************************
   SUBROUTINE qs_mol_dyn(force_env, globenv, averages, rm_restart_info, hmc_e_initial, hmc_e_final, mdctrl)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(average_quantities_type), OPTIONAL, POINTER   :: averages
      LOGICAL, INTENT(IN), OPTIONAL                      :: rm_restart_info
      REAL(KIND=dp), OPTIONAL                            :: hmc_e_initial, hmc_e_final
      TYPE(mdctrl_type), OPTIONAL, POINTER               :: mdctrl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_mol_dyn', routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_rm_restart_info
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(section_vals_type), POINTER                   :: md_section, motion_section

      my_rm_restart_info = .TRUE.
      IF (PRESENT(rm_restart_info)) my_rm_restart_info = rm_restart_info
      NULLIFY (md_env, para_env)
      para_env => force_env%para_env
      motion_section => section_vals_get_subs_vals(force_env%root_section, "MOTION")
      md_section => section_vals_get_subs_vals(motion_section, "MD")

      ! Real call to MD driver - Low Level
      CALL md_env_create(md_env, md_section, para_env, force_env=force_env)
      CALL set_md_env(md_env, averages=averages)
      IF (PRESENT(hmc_e_initial) .AND. PRESENT(hmc_e_final)) THEN
         CALL qs_mol_dyn_low(md_env, md_section, motion_section, force_env, globenv, &
                             hmc_e_initial=hmc_e_initial, hmc_e_final=hmc_e_final)
      ELSE
         CALL qs_mol_dyn_low(md_env, md_section, motion_section, force_env, globenv, mdctrl=mdctrl)
      ENDIF
      CALL md_env_release(md_env)

      ! Clean restartable sections..
      IF (my_rm_restart_info) CALL remove_restart_info(force_env%root_section)
   END SUBROUTINE qs_mol_dyn

! **************************************************************************************************
!> \brief Purpose: Driver routine for MD run using QUICKSTEP.
!> \param md_env ...
!> \param md_section ...
!> \param motion_section ...
!> \param force_env ...
!> \param globenv ...
!> \param hmc_e_initial ...
!> \param hmc_e_final ...
!> \param mdctrl ...
!> \par History
!>   - Cleaning (09.2007) Teodoro Laino [tlaino] - University of Zurich
!>   - Added lines to print out langevin regions (2014/02/04, LT)
!> \author Creation (07.11.2002,MK)
! **************************************************************************************************
   SUBROUTINE qs_mol_dyn_low(md_env, md_section, motion_section, force_env, globenv, hmc_e_initial, hmc_e_final, mdctrl)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(section_vals_type), POINTER                   :: md_section, motion_section
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      REAL(KIND=dp), OPTIONAL                            :: hmc_e_initial, hmc_e_final
      TYPE(mdctrl_type), OPTIONAL, POINTER               :: mdctrl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_mol_dyn_low', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_act, my_pos
      INTEGER                                            :: handle, i, istep, md_stride, run_type_id
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: check, ehrenfest_md, save_mem, &
                                                            should_stop, write_binary_restart_file
      REAL(KIND=dp)                                      :: dummy, time_iter_start, time_iter_stop
      REAL(KIND=dp), POINTER                             :: constant, time, used_time
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(barostat_type), POINTER                       :: barostat
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys, subsys_i
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(free_energy_type), POINTER                    :: fe_env
      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(section_vals_type), POINTER :: constraint_section, force_env_section, &
         free_energy_section, global_section, reftraj_section, subsys_section, work_section
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermal_regions_type), POINTER                :: thermal_regions
      TYPE(thermostats_type), POINTER                    :: thermostats
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("motion/md_run.F",212)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/md_run.F",213)

      NULLIFY (particles, cell, simpar, itimes, used_time, subsys, &
               md_ener, thermostats, barostat, reftraj, force_env_section, &
               reftraj_section, work_section, atomic_kinds, &
               local_particles, time, fe_env, free_energy_section, &
               constraint_section, thermal_regions, virial, subsys_i)
      logger => cp_get_default_logger()
      para_env => force_env%para_env

      global_section => section_vals_get_subs_vals(force_env%root_section, "GLOBAL")
      free_energy_section => section_vals_get_subs_vals(motion_section, "FREE_ENERGY")
      constraint_section => section_vals_get_subs_vals(motion_section, "CONSTRAINT")
      CALL section_vals_val_get(global_section, "SAVE_MEM", l_val=save_mem)

      CALL section_vals_val_get(global_section, "RUN_TYPE", i_val=run_type_id)
      IF (run_type_id == ehrenfest) CALL set_md_env(md_env, ehrenfest_md=.TRUE.)

      CALL create_simpar_type(simpar)
      force_env_section => force_env%force_env_section
      subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      CALL cp_add_iter_level(logger%iter_info, "MD")
      CALL cp_iterate(logger%iter_info, iter_nr=0)
      ! Read MD section
      CALL read_md_section(simpar, motion_section, md_section)
      ! Setup print_keys
      simpar%info_constraint = cp_print_key_unit_nr(logger, constraint_section, &
                                                    "CONSTRAINT_INFO", extension=".shakeLog", log_filename=.FALSE.)
      simpar%lagrange_multipliers = cp_print_key_unit_nr(logger, constraint_section, &
                                                         "LAGRANGE_MULTIPLIERS", extension=".LagrangeMultLog", log_filename=.FALSE.)
      simpar%dump_lm = BTEST(cp_print_key_should_output(logger%iter_info, constraint_section, &
                                                        "LAGRANGE_MULTIPLIERS"), cp_p_file)

      ! Create the structure for the md energies
      CALL create_md_ener(md_ener)
      CALL set_md_env(md_env, md_ener=md_ener)
      CALL release_md_ener(md_ener)

      ! If requested setup Thermostats
      CALL create_thermostats(thermostats, md_section, force_env, simpar, para_env, &
                              globenv, global_section)

      ! If requested setup Barostat
      CALL create_barostat_type(barostat, md_section, force_env, simpar, globenv)

      ! If requested setup different thermal regions
      CALL create_thermal_regions(thermal_regions, md_section, simpar, force_env)

      ! If doing langevin_ensemble, then print out langevin_regions information upon request
      IF (simpar%ensemble == langevin_ensemble) THEN
         my_pos = "REWIND"
         my_act = "WRITE"
         CALL print_thermal_regions_langevin(thermal_regions, simpar, &
                                             pos=my_pos, act=my_act)
      END IF

      CALL set_md_env(md_env, thermostats=thermostats, barostat=barostat, thermal_regions=thermal_regions)

      CALL get_md_env(md_env, ehrenfest_md=ehrenfest_md)

      !If requested set up the REFTRAJ run
      IF (simpar%ensemble == reftraj_ensemble .AND. ehrenfest_md) &
         CALL cp__b("motion/md_run.F",275,"Ehrenfest MD does not support reftraj ensemble ")
      IF (simpar%ensemble == reftraj_ensemble) THEN
         reftraj_section => section_vals_get_subs_vals(md_section, "REFTRAJ")
         CALL create_reftraj(reftraj, reftraj_section, para_env)
         CALL set_md_env(md_env, reftraj=reftraj)
         CALL release_reftraj(reftraj)
      END IF

      CALL force_env_get(force_env, subsys=subsys, cell=cell, &
                         force_env_section=force_env_section)
      CALL cp_subsys_get(subsys, virial=virial)

      ! Set V0 if needed
      IF (simpar%ensemble == nph_uniaxial_ensemble .OR. simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
         IF (simpar%v0 == 0._dp) simpar%v0 = cell%deth
      ENDIF

      ! Initialize velocities possibly applying constraints at the zeroth MD step
      CALL section_vals_val_get(motion_section, "PRINT%RESTART%SPLIT_RESTART_FILE", &
                                l_val=write_binary_restart_file)
      CALL setup_velocities(force_env, simpar, globenv, md_env, md_section, constraint_section, &
                            write_binary_restart_file)

      ! Setup Free Energy Calculation (if required)
      CALL fe_env_create(fe_env, free_energy_section)

      CALL set_md_env(md_env=md_env, simpar=simpar, fe_env=fe_env, cell=cell, &
                      force_env=force_env)

      ! Possibly initialize Wiener processes
      ![NB] Tested again within create_wiener_process.  Why??
      IF (need_per_atom_wiener_process(md_env)) CALL create_wiener_process(md_env)

      time_iter_start = m_walltime()

      CALL get_md_env(md_env, force_env=force_env, itimes=itimes, constant=constant, &
                      md_ener=md_ener, t=time, used_time=used_time)

      ! Attach the time counter of the meta_env to the one of the MD
      CALL set_meta_env(force_env%meta_env, time=time)

      ! Initialize the md_ener structure
      CALL initialize_md_ener(md_ener, force_env, simpar)

      ! Check for ensembles requiring the stress tensor - takes into account the possibility for
      ! multiple force_evals
      IF ((simpar%ensemble == npt_i_ensemble) .OR. &
          (simpar%ensemble == npt_f_ensemble) .OR. &
          (simpar%ensemble == npe_f_ensemble) .OR. &
          (simpar%ensemble == npe_i_ensemble) .OR. &
          (simpar%ensemble == nph_uniaxial_ensemble) .OR. &
          (simpar%ensemble == nph_uniaxial_damped_ensemble)) THEN
         check = virial%pv_availability
         IF (.NOT. check) &
            CALL cp_abort(cp__l("motion/md_run.F",329), &
                          "Virial evaluation not requested for this run in the input file! "// &
                          "You may consider to switch on the virial evaluation with the keyword: STRESS_TENSOR."// &
                          "Be sure the method you are using can compute the virial!")
         IF (ASSOCIATED(force_env%sub_force_env)) THEN
            DO i = 1, SIZE(force_env%sub_force_env)
               IF (ASSOCIATED(force_env%sub_force_env(i)%force_env)) THEN
                  CALL force_env_get(force_env%sub_force_env(i)%force_env, subsys=subsys_i)
                  CALL cp_subsys_get(subsys_i, virial=virial)
                  check = check .AND. virial%pv_availability
               END IF
            END DO
         END IF
         IF (.NOT. check) &
            CALL cp_abort(cp__l("motion/md_run.F",343), &
                          "Virial evaluation not requested for all the force_eval sections present in"// &
                          " the input file! You have to switch on the virial evaluation with the keyword: STRESS_TENSOR "// &
                          " in each force_eval section. Be sure the method you are using can compute the virial!")
      END IF

      ! Computing Forces at zero MD step
      IF (simpar%ensemble /= reftraj_ensemble) THEN
         CALL section_vals_val_get(md_section, "STEP_START_VAL", i_val=itimes)
         CALL section_vals_val_get(md_section, "TIME_START_VAL", r_val=time)
         CALL section_vals_val_get(md_section, "ECONS_START_VAL", r_val=constant)
         CALL cp_iterate(logger%iter_info, iter_nr=itimes)
         IF (save_mem) THEN
            work_section => section_vals_get_subs_vals(subsys_section, "VELOCITY")
            CALL section_vals_remove_values(work_section)
            work_section => section_vals_get_subs_vals(subsys_section, "SHELL_VELOCITY")
            CALL section_vals_remove_values(work_section)
            work_section => section_vals_get_subs_vals(subsys_section, "CORE_VELOCITY")
            CALL section_vals_remove_values(work_section)
         END IF

         IF (ehrenfest_md) THEN
            CALL rt_prop_setup(force_env)
            force_env%qs_env%rtp%dt = simpar%dt
         ELSE
            ![NB] Lets let all methods, even ones without consistent energies, succeed here.
            !     They'll fail in actual integrator if needed
            ! consistent_energies=.FALSE. by default
            CALL force_env_calc_energy_force(force_env, calc_force=.TRUE.)
         END IF

         IF (ASSOCIATED(force_env%qs_env)) THEN
            force_env%qs_env%sim_time = time
            force_env%qs_env%sim_step = itimes
         END IF
         ! Warm-up engines for metadynamics
         IF (ASSOCIATED(force_env%meta_env)) THEN
            ! Setup stuff for plumed if needed
            IF (force_env%meta_env%use_plumed .EQV. .TRUE.) THEN
               CALL metadyn_initialise_plumed(force_env, simpar, itimes)
            ELSE
               IF (force_env%meta_env%langevin) THEN
                  CALL create_wiener_process_cv(force_env%meta_env)
               ENDIF
               IF (force_env%meta_env%well_tempered) THEN
                  force_env%meta_env%wttemperature = simpar%temp_ext
                  IF (force_env%meta_env%wtgamma > EPSILON(1._dp)) THEN
                     dummy = force_env%meta_env%wttemperature*(force_env%meta_env%wtgamma-1._dp)
                     IF (force_env%meta_env%delta_t > EPSILON(1._dp)) THEN
                        check = ABS(force_env%meta_env%delta_t-dummy) < 1.E+3_dp*EPSILON(1._dp)
                        IF (.NOT. check) &
                           CALL cp_abort(cp__l("motion/md_run.F",394), &
                                         "Inconsistency between DELTA_T and WTGAMMA (both specified):"// &
                                         " please, verify that DELTA_T=(WTGAMMA-1)*TEMPERATURE")
                     ELSE
                        force_env%meta_env%delta_t = dummy
                     ENDIF
                  ELSE
                     force_env%meta_env%wtgamma = 1._dp &
                                                  +force_env%meta_env%delta_t/force_env%meta_env%wttemperature
                  ENDIF
                  force_env%meta_env%invdt = 1._dp/force_env%meta_env%delta_t
               ENDIF
               CALL metadyn_forces(force_env)
               CALL metadyn_write_colvar(force_env)
            END IF
         ENDIF

         IF (simpar%do_respa) THEN
            CALL force_env_calc_energy_force(force_env%sub_force_env(1)%force_env, &
                                             calc_force=.TRUE.)
         END IF

         CALL force_env_get(force_env, subsys=subsys)

         CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds, local_particles=local_particles, &
                            particles=particles, virial=virial)

         CALL virial_evaluate(atomic_kinds%els, particles%els, local_particles, &
                              virial, force_env%para_env%group)

         CALL md_energy(md_env, md_ener)
         CALL md_write_output(md_env) !inits the print env at itimes == 0 also writes trajectories
         md_stride = 1
      ELSE
         CALL get_md_env(md_env, reftraj=reftraj)
         CALL initialize_reftraj(reftraj, reftraj_section, md_env)
         itimes = reftraj%info%first_snapshot-1
         md_stride = reftraj%info%stride
         IF (ASSOCIATED(force_env%meta_env)) THEN
            IF (force_env%meta_env%use_plumed .EQV. .TRUE.) THEN
               CALL metadyn_initialise_plumed(force_env, simpar, itimes)
            END IF
         END IF
      END IF

      CALL cp_print_key_finished_output(simpar%info_constraint, logger, &
                                        constraint_section, "CONSTRAINT_INFO")
      CALL cp_print_key_finished_output(simpar%lagrange_multipliers, logger, &
                                        constraint_section, "LAGRANGE_MULTIPLIERS")

! if we need the initial kinetic energy for Hybrid Monte Carlo
      IF (PRESENT(hmc_e_initial)) hmc_e_initial = md_ener%ekin

      ! Real MD Loop
      DO istep = 1, simpar%nsteps, md_stride
         ! Increase counters
         itimes = itimes+1
         time = time+simpar%dt
         !needed when electric field fields are applied
         IF (ASSOCIATED(force_env%qs_env)) THEN
            force_env%qs_env%sim_time = time
            force_env%qs_env%sim_step = itimes
         END IF
         IF (ehrenfest_md) force_env%qs_env%rtp%istep = istep

         CALL cp_iterate(logger%iter_info, last=(istep == simpar%nsteps), iter_nr=itimes)
         ! Open possible Shake output units
         simpar%info_constraint = cp_print_key_unit_nr(logger, constraint_section, "CONSTRAINT_INFO", &
                                                       extension=".shakeLog", log_filename=.FALSE.)
         simpar%lagrange_multipliers = cp_print_key_unit_nr( &
                                       logger, constraint_section, &
                                       "LAGRANGE_MULTIPLIERS", extension=".LagrangeMultLog", log_filename=.FALSE.)
         simpar%dump_lm = BTEST(cp_print_key_should_output(logger%iter_info, constraint_section, &
                                                           "LAGRANGE_MULTIPLIERS"), cp_p_file)

         ! Velocity Verlet Integrator
         CALL velocity_verlet(md_env, globenv)

         ! Close Shake output if requested...
         CALL cp_print_key_finished_output(simpar%info_constraint, logger, &
                                           constraint_section, "CONSTRAINT_INFO")
         CALL cp_print_key_finished_output(simpar%lagrange_multipliers, logger, &
                                           constraint_section, "LAGRANGE_MULTIPLIERS")

         ! Free Energy calculation
         CALL free_energy_evaluate(md_env, should_stop, free_energy_section)

         IF (should_stop) EXIT

         ! Test for <PROJECT_NAME>.EXIT_MD or for WALL_TIME to exit
         ! Default:
         ! IF so we don't overwrite the restart or append to the trajectory
         ! because the execution could in principle stop inside the SCF where energy
         ! and forces are not converged.
         ! But:
         ! You can force to print the last step (for example if the method used
         ! to compute energy and forces is not SCF based) activating the print_key
         ! MOTION%MD%PRINT%FORCE_LAST.
         CALL external_control(should_stop, "MD", globenv=globenv)

         ! call external hook e.g. from global optimization
         IF (PRESENT(mdctrl)) &
            CALL mdctrl_callback(mdctrl, md_env, should_stop)

         IF (should_stop) THEN
            CALL cp_iterate(logger%iter_info, last=.TRUE., iter_nr=itimes)
            !In Ehrenfest molecular dynamics the external control is only checked after a converged propagation
            !The restart needs to be written in order to be consistent with the mos/density matrix for the restart
            IF (run_type_id == ehrenfest) THEN
               CALL md_output(md_env, md_section, force_env%root_section, .FALSE.)
            ELSE
               CALL md_output(md_env, md_section, force_env%root_section, should_stop)
            ENDIF
            EXIT
         END IF

         IF (simpar%ensemble /= reftraj_ensemble) THEN
            CALL md_energy(md_env, md_ener)
            CALL temperature_control(simpar, md_env, md_ener, force_env, logger)
            CALL comvel_control(md_ener, force_env, md_section, logger)
            CALL angvel_control(md_ener, force_env, md_section, logger)
         ELSE
            CALL md_ener_reftraj(md_env, md_ener)
         END IF

         time_iter_stop = m_walltime()
         used_time = time_iter_stop-time_iter_start
         time_iter_start = time_iter_stop

         CALL md_output(md_env, md_section, force_env%root_section, should_stop)
         IF (simpar%ensemble == reftraj_ensemble) THEN
            CALL write_output_reftraj(md_env)
         END IF
      END DO

! if we need the final kinetic energy for Hybrid Monte Carlo
      IF (PRESENT(hmc_e_final)) hmc_e_final = md_ener%ekin

      ! Remove the iteration level
      CALL cp_rm_iter_level(logger%iter_info, "MD")

      ! Clean up PLUMED
      IF (ASSOCIATED(force_env%meta_env)) THEN
         IF (force_env%meta_env%use_plumed .EQV. .TRUE.) THEN
            CALL metadyn_finalise_plumed()
         END IF
      END IF

      ! Deallocate Thermostats and Barostats
      CALL release_thermostats(thermostats)
      CALL release_barostat_type(barostat)
      CALL release_simpar_type(simpar)
      CALL release_thermal_regions(thermal_regions)
      CALL timestop(handle)

   END SUBROUTINE qs_mol_dyn_low

END MODULE md_run
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Worker routines used by global optimization schemes
!> \author Ole Schuett
! **************************************************************************************************
MODULE glbopt_worker
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type,&
                                              pack_subsys_particles,&
                                              unpack_subsys_particles
   USE f77_interface,                   ONLY: create_force_env,&
                                              destroy_force_env,&
                                              f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE geo_opt,                         ONLY: cp_geo_opt
   USE global_types,                    ONLY: global_environment_type
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE md_run,                          ONLY: qs_mol_dyn
   USE mdctrl_types,                    ONLY: glbopt_mdctrl_data_type,&
                                              mdctrl_type
   USE parallel_rng_types,              ONLY: reset_to_next_rng_substream
   USE physcon,                         ONLY: angstrom,&
                                              kelvin
   USE swarm_message,                   ONLY: swarm_message_add,&
                                              swarm_message_get,&
                                              swarm_message_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'glbopt_worker'

   PUBLIC :: glbopt_worker_init, glbopt_worker_finalize
   PUBLIC :: glbopt_worker_execute
   PUBLIC :: glbopt_worker_type

   TYPE glbopt_worker_type
      PRIVATE
      INTEGER                                  :: id
      INTEGER                                  :: iw
      INTEGER                                  :: f_env_id
      TYPE(f_env_type), POINTER                :: f_env
      TYPE(force_env_type), POINTER            :: force_env
      TYPE(cp_subsys_type), POINTER            :: subsys
      TYPE(section_vals_type), POINTER         :: root_section
      TYPE(global_environment_type), POINTER   :: globenv
      INTEGER                                  :: gopt_max_iter
      INTEGER                                  :: bump_steps_downwards
      INTEGER                                  :: bump_steps_upwards
      INTEGER                                  :: md_bumps_max
      REAL(KIND=dp)                            :: fragmentation_threshold
      INTEGER                                  :: n_atoms = -1
      !REAL(KIND=dp)                            :: adaptive_timestep = 0.0
   END TYPE glbopt_worker_type

CONTAINS

! **************************************************************************************************
!> \brief Initializes worker for global optimization
!> \param worker ...
!> \param input_declaration ...
!> \param para_env ...
!> \param root_section ...
!> \param input_path ...
!> \param worker_id ...
!> \param iw ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE glbopt_worker_init(worker, input_declaration, para_env, root_section, &
                                 input_path, worker_id, iw)
      TYPE(glbopt_worker_type), INTENT(INOUT)            :: worker
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: root_section
      CHARACTER(LEN=*), INTENT(IN)                       :: input_path
      INTEGER, INTENT(in)                                :: worker_id, iw

      CHARACTER(len=*), PARAMETER :: routineN = 'glbopt_worker_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(kind=dp)                                      :: dist_in_angstrom
      TYPE(section_vals_type), POINTER                   :: glbopt_section

      worker%root_section => root_section
      worker%id = worker_id
      worker%iw = iw

      ! ======= Create f_env =======
      CALL create_force_env(worker%f_env_id, &
                            input_declaration=input_declaration, &
                            input_path=input_path, &
                            input=root_section, &
                            output_unit=worker%iw, &
                            mpi_comm=para_env%group)

      ! ======= More setup stuff =======
      CALL f_env_add_defaults(worker%f_env_id, worker%f_env)
      worker%force_env => worker%f_env%force_env
      CALL force_env_get(worker%force_env, globenv=worker%globenv, subsys=worker%subsys)

      ! We want different random-number-streams for each worker
      DO i = 1, worker_id
         CALL reset_to_next_rng_substream(worker%globenv%gaussian_rng_stream)
      END DO

      CALL cp_subsys_get(worker%subsys, natom=worker%n_atoms)

      ! fetch original value from input
      CALL section_vals_val_get(root_section, "MOTION%GEO_OPT%MAX_ITER", i_val=worker%gopt_max_iter)
      glbopt_section => section_vals_get_subs_vals(root_section, "SWARM%GLOBAL_OPT")

      CALL section_vals_val_get(glbopt_section, "BUMP_STEPS_UPWARDS", i_val=worker%bump_steps_upwards)
      CALL section_vals_val_get(glbopt_section, "BUMP_STEPS_DOWNWARDS", i_val=worker%bump_steps_downwards)
      CALL section_vals_val_get(glbopt_section, "MD_BUMPS_MAX", i_val=worker%md_bumps_max)
      CALL section_vals_val_get(glbopt_section, "FRAGMENTATION_THRESHOLD", r_val=dist_in_angstrom)
      !CALL section_vals_val_get(glbopt_section,"MD_ADAPTIVE_TIMESTEP", r_val=worker%adaptive_timestep)
      worker%fragmentation_threshold = dist_in_angstrom/angstrom
   END SUBROUTINE glbopt_worker_init

! **************************************************************************************************
!> \brief Central execute routine of global optimization worker
!> \param worker ...
!> \param cmd ...
!> \param report ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE glbopt_worker_execute(worker, cmd, report)
      TYPE(glbopt_worker_type), INTENT(INOUT)            :: worker
      TYPE(swarm_message_type), INTENT(IN)               :: cmd
      TYPE(swarm_message_type), INTENT(INOUT)            :: report

      CHARACTER(len=*), PARAMETER :: routineN = 'glbopt_worker_execute', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: command

      CALL swarm_message_get(cmd, "command", command)
      IF (TRIM(command) == "md_and_gopt") THEN
         CALL run_mdgopt(worker, cmd, report)
      ELSE
         CALL cp__b("swarm/glbopt_worker.F",157,"Worker: received unknown command")
      END IF

   END SUBROUTINE glbopt_worker_execute

! **************************************************************************************************
!> \brief Performs an escape attempt as need by e.g. Minima Hopping
!> \param worker ...
!> \param cmd ...
!> \param report ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE run_mdgopt(worker, cmd, report)
      TYPE(glbopt_worker_type), INTENT(INOUT)            :: worker
      TYPE(swarm_message_type), INTENT(IN)               :: cmd
      TYPE(swarm_message_type), INTENT(INOUT)            :: report

      INTEGER                                            :: gopt_steps, iframe, md_steps, &
                                                            n_fragments, prev_iframe
      REAL(kind=dp)                                      :: Epot, temperature
      REAL(KIND=dp), DIMENSION(:), POINTER               :: positions
      TYPE(glbopt_mdctrl_data_type), TARGET              :: mdctrl_data
      TYPE(mdctrl_type), POINTER                         :: mdctrl_p
      TYPE(mdctrl_type), TARGET                          :: mdctrl

      NULLIFY (positions)

      CALL swarm_message_get(cmd, "temperature", temperature)
      CALL swarm_message_get(cmd, "iframe", iframe)
      IF (iframe > 1) THEN
         CALL swarm_message_get(cmd, "positions", positions)
         CALL unpack_subsys_particles(worker%subsys, r=positions)
      ENDIF

      ! setup mdctrl callback
      ALLOCATE (mdctrl_data%epot_history(worker%bump_steps_downwards+worker%bump_steps_upwards+1))
      mdctrl_data%epot_history = 0.0
      mdctrl_data%md_bump_counter = 0
      mdctrl_data%bump_steps_upwards = worker%bump_steps_upwards
      mdctrl_data%bump_steps_downwards = worker%bump_steps_downwards
      mdctrl_data%md_bumps_max = worker%md_bumps_max
      mdctrl_data%output_unit = worker%iw
      mdctrl%glbopt => mdctrl_data
      mdctrl_p => mdctrl

      !IF(worker%adaptive_timestep > 0.0) THEN
      !   !TODO: 300K is hard encoded
      !   boltz = 1.0 + exp( -temperature * kelvin / 150.0 )
      !   timestep = 4.0 * ( boltz - 1.0 ) / boltz / femtoseconds
      !   !timestep = 0.01_dp / femtoseconds
      !   !timestep = SQRT(MIN(0.5, 2.0/(1+exp(-300.0/(temperature*kelvin))))) / femtoseconds
      !   CALL section_vals_val_set(worker%root_section, "MOTION%MD%TIMESTEP", r_val=timestep)
      !   IF(worker%iw>0)&
      !      WRITE (worker%iw,'(A,35X,F20.3)')  ' GLBOPT| MD timestep [fs]',timestep*femtoseconds
      !ENDIF

      prev_iframe = iframe
      IF (iframe == 0) iframe = 1 ! qs_mol_dyn behaves differently for STEP_START_VAL=0
      CALL section_vals_val_set(worker%root_section, "MOTION%MD%STEP_START_VAL", i_val=iframe-1)
      CALL section_vals_val_set(worker%root_section, "MOTION%MD%TEMPERATURE", r_val=temperature)

      IF (worker%iw > 0) THEN
         WRITE (worker%iw, '(A,33X,F20.3)') ' GLBOPT| MD temperature [K]', temperature*kelvin
         WRITE (worker%iw, '(A,29X,I10)') " GLBOPT| Starting MD at trajectory frame ", iframe
      END IF

      ! run MD
      CALL qs_mol_dyn(worker%force_env, worker%globenv, mdctrl=mdctrl_p)

      iframe = mdctrl_data%itimes+1
      md_steps = iframe-prev_iframe
      IF (worker%iw > 0) WRITE (worker%iw, '(A,I4,A)') " GLBOPT| md ended after ", md_steps, " steps."

      ! fix fragmentation
      IF (.NOT. ASSOCIATED(positions)) ALLOCATE (positions(3*worker%n_atoms))
      CALL pack_subsys_particles(worker%subsys, r=positions)
      n_fragments = 0
      DO
         n_fragments = n_fragments+1
         IF (fix_fragmentation(positions, worker%fragmentation_threshold)) EXIT
      END DO
      CALL unpack_subsys_particles(worker%subsys, r=positions)

      IF (n_fragments > 0 .AND. worker%iw > 0) &
         WRITE (worker%iw, '(A,13X,I10)') " GLBOPT| Ran fix_fragmentation times:", n_fragments

      ! setup geometry optimization
      IF (worker%iw > 0) WRITE (worker%iw, '(A,13X,I10)') " GLBOPT| Starting local optimisation at trajectory frame ", iframe
      CALL section_vals_val_set(worker%root_section, "MOTION%GEO_OPT%STEP_START_VAL", i_val=iframe-1)
      CALL section_vals_val_set(worker%root_section, "MOTION%GEO_OPT%MAX_ITER", &
                                i_val=iframe+worker%gopt_max_iter)

      ! run geometry optimization
      CALL cp_geo_opt(worker%force_env, worker%globenv, rm_restart_info=.FALSE.)

      prev_iframe = iframe
      CALL section_vals_val_get(worker%root_section, "MOTION%GEO_OPT%STEP_START_VAL", i_val=iframe)
      iframe = iframe+2 ! Compensates for different START_VAL interpretation.
      gopt_steps = iframe-prev_iframe-1
      IF (worker%iw > 0) WRITE (worker%iw, '(A,I4,A)') " GLBOPT| gopt ended after ", gopt_steps, " steps."
      CALL force_env_get(worker%force_env, potential_energy=Epot)
      IF (worker%iw > 0) WRITE (worker%iw, '(A,25X,E20.10)') ' GLBOPT| Potential Energy [Hartree]', Epot

      ! assemble report
      CALL swarm_message_add(report, "Epot", Epot)
      CALL swarm_message_add(report, "iframe", iframe)
      CALL swarm_message_add(report, "md_steps", md_steps)
      CALL swarm_message_add(report, "gopt_steps", gopt_steps)
      CALL pack_subsys_particles(worker%subsys, r=positions)
      CALL swarm_message_add(report, "positions", positions)

      DEALLOCATE (positions)
   END SUBROUTINE run_mdgopt

! **************************************************************************************************
!> \brief Helper routine for run_mdgopt, fixes a fragmented atomic cluster.
!> \param positions ...
!> \param bondlength ...
!> \retval all_connected ...
!> \author Stefan Goedecker
! **************************************************************************************************
   FUNCTION fix_fragmentation(positions, bondlength) RESULT(all_connected)
      REAL(KIND=dp), DIMENSION(:)                        :: positions
      REAL(KIND=dp)                                      :: bondlength
      LOGICAL                                            :: all_connected

      INTEGER                                            :: cluster_edge, fragment_edge, i, j, &
                                                            n_particles, stack_size
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: stack
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: marked
      REAL(KIND=dp)                                      :: d, dr(3), min_dist, s

      n_particles = SIZE(positions)/3
      ALLOCATE (stack(n_particles), marked(n_particles))

      marked = .FALSE.; stack_size = 0

      ! First particle taken as root of flooding, mark it and push to stack
      marked(1) = .TRUE.; stack(1) = 1; stack_size = 1

      DO WHILE (stack_size > 0)
         i = stack(stack_size); stack_size = stack_size-1 !pop
         DO j = 1, n_particles
            IF (norm(diff(positions, i, j)) < 1.25*bondlength) THEN ! they are close = they are connected
               IF (.NOT. marked(j)) THEN
                  marked(j) = .TRUE.
                  stack(stack_size+1) = j; stack_size = stack_size+1; !push
               END IF
            END IF
         END DO
      END DO

      all_connected = ALL(marked) !did we visit every particle?
      IF (all_connected) RETURN

      ! make sure we keep the larger chunk
      IF (COUNT(marked) < n_particles/2) marked(:) = .NOT. (marked(:))

      min_dist = HUGE(1.0)
      cluster_edge = -1
      fragment_edge = -1
      DO i = 1, n_particles
         IF (marked(i)) CYCLE
         DO j = 1, n_particles
            IF (.NOT. marked(j)) CYCLE
            d = norm(diff(positions, i, j))
            IF (d < min_dist) THEN
               min_dist = d
               cluster_edge = i
               fragment_edge = j
            END IF
         END DO
      END DO

      dr = diff(positions, cluster_edge, fragment_edge)
      s = 1.0-bondlength/norm(dr)
      DO i = 1, n_particles
         IF (marked(i)) CYCLE
         positions(3*i-2:3*i) = positions(3*i-2:3*i)-s*dr
      END DO

   END FUNCTION fix_fragmentation

! **************************************************************************************************
!> \brief Helper routine for fix_fragmentation, calulates atomic distance
!> \param positions ...
!> \param i ...
!> \param j ...
!> \retval dr ...
!> \author Ole Schuett
! **************************************************************************************************
   PURE FUNCTION diff(positions, i, j) RESULT(dr)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: positions
      INTEGER, INTENT(IN)                                :: i, j
      REAL(KIND=dp), DIMENSION(3)                        :: dr

      dr = positions(3*i-2:3*i)-positions(3*j-2:3*j)
   END FUNCTION diff

! **************************************************************************************************
!> \brief Helper routine for fix_fragmentation, calulates vector norm
!> \param vec ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   PURE FUNCTION norm(vec) RESULT(res)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: vec
      REAL(KIND=dp)                                      :: res

      res = SQRT(DOT_PRODUCT(vec, vec))
   END FUNCTION norm

! **************************************************************************************************
!> \brief Finalizes worker for global optimization
!> \param worker ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE glbopt_worker_finalize(worker)
      TYPE(glbopt_worker_type), INTENT(INOUT)            :: worker

      CHARACTER(len=*), PARAMETER :: routineN = 'glbopt_worker_finalize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ierr

      CALL f_env_rm_defaults(worker%f_env)
      CALL destroy_force_env(worker%f_env_id, ierr)
      IF (ierr /= 0) CALL cp__b("swarm/glbopt_worker.F",384,"destroy_force_env failed")
   END SUBROUTINE glbopt_worker_finalize

END MODULE glbopt_worker

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Workers's routines for the swarm-framework
!> \author Ole Schuett
! **************************************************************************************************
MODULE swarm_worker
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE glbopt_worker,                   ONLY: glbopt_worker_execute,&
                                              glbopt_worker_finalize,&
                                              glbopt_worker_init,&
                                              glbopt_worker_type
   USE input_constants,                 ONLY: swarm_do_glbopt
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length
   USE swarm_message,                   ONLY: swarm_message_add,&
                                              swarm_message_get,&
                                              swarm_message_haskey,&
                                              swarm_message_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'swarm_worker'

   PUBLIC :: swarm_worker_init, swarm_worker_finalize
   PUBLIC :: swarm_worker_execute
   PUBLIC :: swarm_worker_type

   TYPE swarm_worker_type
      PRIVATE
      INTEGER                                  :: id = -1
      INTEGER                                  :: iw = -1
      INTEGER                                  :: behavior = -1
      TYPE(glbopt_worker_type), POINTER        :: glbopt => Null()
      !possbily more behaviors...
   END TYPE swarm_worker_type

CONTAINS

! **************************************************************************************************
!> \brief Initializes a swarm worker
!> \param worker ...
!> \param para_env ...
!> \param input_declaration ...
!> \param root_section ...
!> \param input_path ...
!> \param worker_id ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_worker_init(worker, para_env, input_declaration, root_section, &
                                input_path, worker_id)
      TYPE(swarm_worker_type), INTENT(INOUT)             :: worker
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      CHARACTER(LEN=*), INTENT(IN)                       :: input_path
      INTEGER, INTENT(in)                                :: worker_id

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_worker_init', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      worker%id = worker_id

      ! getting an output unit for logging
      logger => cp_get_default_logger()
      worker%iw = cp_print_key_unit_nr(logger, root_section, &
                                       "SWARM%PRINT%WORKER_RUN_INFO", extension=".workerLog")

      CALL section_vals_val_get(root_section, "SWARM%BEHAVIOR", i_val=worker%behavior)

      SELECT CASE (worker%behavior)
      CASE (swarm_do_glbopt)
         ALLOCATE (worker%glbopt)
         CALL glbopt_worker_init(worker%glbopt, input_declaration, para_env, &
                                 root_section, input_path, worker_id, worker%iw)
      CASE DEFAULT
         CALL cp__b("swarm/swarm_worker.F",89,"got unknown behavior")
      END SELECT

   END SUBROUTINE swarm_worker_init

! **************************************************************************************************
!> \brief Central execute routine of the swarm worker
!> \param worker ...
!> \param cmd ...
!> \param report ...
!> \param should_stop ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_worker_execute(worker, cmd, report, should_stop)
      TYPE(swarm_worker_type), INTENT(INOUT)             :: worker
      TYPE(swarm_message_type), INTENT(IN)               :: cmd
      TYPE(swarm_message_type), INTENT(OUT)              :: report
      LOGICAL, INTENT(INOUT)                             :: should_stop

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_worker_execute', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: command

      CALL swarm_message_get(cmd, "command", command)
      CALL swarm_message_add(report, "worker_id", worker%id)

      IF (TRIM(command) == "shutdown") THEN
         IF (worker%iw > 0) WRITE (worker%iw, *) "SWARM| Received shutdown command, quitting."
         should_stop = .TRUE.
      ELSE IF (TRIM(command) == "wait") THEN !only needed for serial driver
         CALL swarm_message_add(report, "status", "wait_done")
      ELSE
         SELECT CASE (worker%behavior)
         CASE (swarm_do_glbopt)
            CALL glbopt_worker_execute(worker%glbopt, cmd, report)
         CASE DEFAULT
            CALL cp__b("swarm/swarm_worker.F",126,"got unknown behavior")
         END SELECT
      ENDIF

      IF (.NOT. swarm_message_haskey(report, "status")) &
         CALL swarm_message_add(report, "status", "ok")

   END SUBROUTINE swarm_worker_execute

! **************************************************************************************************
!> \brief Finalizes a swarm worker
!> \param worker ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_worker_finalize(worker)
      TYPE(swarm_worker_type), INTENT(INOUT)             :: worker

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_worker_finalize', &
         routineP = moduleN//':'//routineN

      SELECT CASE (worker%behavior)
      CASE (swarm_do_glbopt)
         CALL glbopt_worker_finalize(worker%glbopt)
         DEALLOCATE (worker%glbopt)
      CASE DEFAULT
         CALL cp__b("swarm/swarm_worker.F",151,"got unknown behavior")
      END SELECT

   END SUBROUTINE swarm_worker_finalize

END MODULE swarm_worker

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Handles the MPI communication of the swarm framework.
!> \author Ole Schuett
! **************************************************************************************************
MODULE swarm_mpi
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_iter_types,                   ONLY: cp_iteration_info_create,&
                                              cp_iteration_info_release,&
                                              cp_iteration_info_type
   USE cp_log_handling,                 ONLY: cp_add_default_logger,&
                                              cp_get_default_logger,&
                                              cp_logger_create,&
                                              cp_logger_release,&
                                              cp_logger_type,&
                                              cp_rm_default_logger
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length
   USE machine,                         ONLY: default_output_unit
   USE message_passing,                 ONLY: mp_any_source,&
                                              mp_bcast,&
                                              mp_comm_free,&
                                              mp_comm_split,&
                                              mp_comm_split_direct,&
                                              mp_environ,&
                                              mp_sum,&
                                              mp_sync
   USE swarm_message,                   ONLY: swarm_message_get,&
                                              swarm_message_mpi_bcast,&
                                              swarm_message_mpi_recv,&
                                              swarm_message_mpi_send,&
                                              swarm_message_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'swarm_mpi'

   PUBLIC :: swarm_mpi_type, swarm_mpi_init, swarm_mpi_finalize
   PUBLIC :: swarm_mpi_send_report, swarm_mpi_recv_report
   PUBLIC :: swarm_mpi_send_command, swarm_mpi_recv_command

   TYPE swarm_mpi_type
      TYPE(cp_para_env_type), POINTER          :: world => Null()
      TYPE(cp_para_env_type), POINTER          :: worker => Null()
      TYPE(cp_para_env_type), POINTER          :: master => Null()
      INTEGER, DIMENSION(:), ALLOCATABLE       :: wid2group
      CHARACTER(LEN=default_path_length)       :: master_output_path = ""
   END TYPE swarm_mpi_type

CONTAINS

! **************************************************************************************************
!> \brief Initialize MPI communicators for a swarm run.
!> \param swarm_mpi ...
!> \param world_para_env ...
!> \param root_section ...
!> \param n_workers ...
!> \param worker_id ...
!> \param iw ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_mpi_init(swarm_mpi, world_para_env, root_section, n_workers, worker_id, iw)
      TYPE(swarm_mpi_type)                               :: swarm_mpi
      TYPE(cp_para_env_type), POINTER                    :: world_para_env
      TYPE(section_vals_type), POINTER                   :: root_section
      INTEGER, INTENT(IN)                                :: n_workers
      INTEGER, INTENT(OUT)                               :: worker_id
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'swarm_mpi_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: n_groups_created, pe_per_worker, &
                                                            subgroup, subgroup_rank, &
                                                            subgroup_size, worker_group
      LOGICAL                                            :: im_the_master
      INTEGER, DIMENSION(:), POINTER                     :: group_distribution_p
      INTEGER, DIMENSION(0:world_para_env%num_pe-2), &
         TARGET                                          :: group_distribution

! ====== Setup of MPI-Groups ======

      worker_id = -1
      swarm_mpi%world => world_para_env

      IF (MOD(swarm_mpi%world%num_pe-1, n_workers) /= 0) THEN
         CALL cp__b("swarm/swarm_mpi.F",98,"number of processors-1 is not divisible by n_workers.")
      ENDIF
      IF (swarm_mpi%world%num_pe < n_workers+1) THEN
         CALL cp__b("swarm/swarm_mpi.F",101,"There are not enough processes for n_workers + 1. Aborting.")
      ENDIF

      pe_per_worker = (swarm_mpi%world%num_pe-1)/n_workers

      IF (iw > 0) THEN
         WRITE (iw, '(A,45X,I8)') " SWARM| Number of mpi ranks", swarm_mpi%world%num_pe
         WRITE (iw, '(A,47X,I8)') " SWARM| Number of workers", n_workers
      ENDIF

      ! the last task becomes the master. Preseves node-alignment of other tasks.
      im_the_master = (swarm_mpi%world%mepos == swarm_mpi%world%num_pe-1)

      ! First split split para_env%group into a master- and a workers-groups...
      IF (im_the_master) THEN
         CALL mp_comm_split_direct(swarm_mpi%world%group, subgroup, 1)
         CALL mp_environ(subgroup_size, subgroup_rank, subgroup)
         IF (subgroup_size /= 1) CALL cp__b("swarm/swarm_mpi.F",118,"mp_comm_split_direct failed (master)")
         CALL cp_para_env_create(swarm_mpi%master, group=subgroup)
         !WRITE (*,*) "this is a master ", swarm_mpi%master%mepos, swarm_mpi%master%num_pe
      ELSE
         CALL mp_comm_split_direct(swarm_mpi%world%group, subgroup, 2)
         CALL mp_environ(subgroup_size, subgroup_rank, subgroup)
         !WRITE (*,*) "Hello, this is a Worker - there are ",subgroup_size, " of us."
         IF (subgroup_size /= swarm_mpi%world%num_pe-1) CALL cp__b("swarm/swarm_mpi.F",125,"mp_comm_split_direct failed (worker)")
      ENDIF

      ALLOCATE (swarm_mpi%wid2group(n_workers))
      swarm_mpi%wid2group = 0

      IF (.NOT. im_the_master) THEN
         ! ...then split workers-group into n_workers groups - one for each worker.
         group_distribution_p => group_distribution
         CALL mp_comm_split(subgroup, worker_group, n_groups_created, group_distribution_p, n_subgroups=n_workers)
         worker_id = group_distribution(subgroup_rank)+1 ! shall start by 1
         IF (n_groups_created /= n_workers) CALL cp__b("swarm/swarm_mpi.F",136,"mp_comm_split failed.")
         CALL cp_para_env_create(swarm_mpi%worker, group=worker_group)
         CALL mp_comm_free(subgroup)

         !WRITE (*,*) "this is worker ", worker_id, swarm_mpi%worker%mepos, swarm_mpi%worker%num_pe

         ! collect world-ranks of each worker groups rank-0 node
         IF (swarm_mpi%worker%mepos == 0) &
            swarm_mpi%wid2group(worker_id) = swarm_mpi%world%mepos

      ENDIF

      CALL mp_sum(swarm_mpi%wid2group, swarm_mpi%world%group)
      !WRITE (*,*), "wid2group table: ",swarm_mpi%wid2group

      CALL logger_init_master(swarm_mpi)
      CALL logger_init_worker(swarm_mpi, root_section, worker_id)
   END SUBROUTINE swarm_mpi_init

! **************************************************************************************************
!> \brief Helper routine for swarm_mpi_init, configures the master's logger.
!> \param swarm_mpi ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE logger_init_master(swarm_mpi)
      TYPE(swarm_mpi_type)                               :: swarm_mpi

      INTEGER                                            :: output_unit
      TYPE(cp_logger_type), POINTER                      :: logger

! broadcast master_output_path to all ranks

      IF (swarm_mpi%world%mepos == swarm_mpi%world%source) THEN
         logger => cp_get_default_logger()
         output_unit = logger%default_local_unit_nr
         swarm_mpi%master_output_path = output_unit2path(output_unit)
         IF (output_unit /= default_output_unit) &
            CLOSE (output_unit)
      END IF

      CALL mp_bcast(swarm_mpi%master_output_path, swarm_mpi%world%source, swarm_mpi%world%group)

      IF (ASSOCIATED(swarm_mpi%master)) &
         CALL error_add_new_logger(swarm_mpi%master, swarm_mpi%master_output_path)
   END SUBROUTINE logger_init_master

! **************************************************************************************************
!> \brief Helper routine for logger_init_master, inquires filename for given unit.
!> \param output_unit ...
!> \retval output_path ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION output_unit2path(output_unit) RESULT(output_path)
      INTEGER, INTENT(IN)                                :: output_unit
      CHARACTER(LEN=default_path_length)                 :: output_path

      output_path = "__STD_OUT__"
      IF (output_unit /= default_output_unit) &
         INQUIRE (unit=output_unit, name=output_path)
   END FUNCTION output_unit2path

! **************************************************************************************************
!> \brief Helper routine for swarm_mpi_init, configures the workers's logger.
!> \param swarm_mpi ...
!> \param root_section ...
!> \param worker_id ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE logger_init_worker(swarm_mpi, root_section, worker_id)
      TYPE(swarm_mpi_type)                               :: swarm_mpi
      TYPE(section_vals_type), POINTER                   :: root_section
      INTEGER                                            :: worker_id

      CHARACTER(len=*), PARAMETER :: routineN = 'logger_init_worker', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: output_path
      CHARACTER(len=default_string_length)               :: new_project_name, project_name, &
                                                            worker_name
      TYPE(cp_iteration_info_type), POINTER              :: new_iter_info
      TYPE(cp_logger_type), POINTER                      :: old_logger

      NULLIFY (old_logger, new_iter_info)
      IF (ASSOCIATED(swarm_mpi%worker)) THEN
         old_logger => cp_get_default_logger()
         project_name = old_logger%iter_info%project_name
         IF (worker_id > 99999) THEN
            CALL cp__b("swarm/swarm_mpi.F",223,"Did not expect so many workers.")
         ENDIF
         WRITE (worker_name, "(A,I5.5)") 'WORKER', worker_id
         IF (LEN_TRIM(project_name)+1+LEN_TRIM(worker_name) > default_string_length) THEN
            CALL cp__b("swarm/swarm_mpi.F",227,"project name too long")
         ENDIF
         output_path = TRIM(project_name)//"-"//TRIM(worker_name)//".out"
         new_project_name = TRIM(project_name)//"-"//TRIM(worker_name)
         CALL section_vals_val_set(root_section, "GLOBAL%PROJECT_NAME", c_val=new_project_name)
         CALL cp_iteration_info_create(new_iter_info, new_project_name)
         CALL error_add_new_logger(swarm_mpi%worker, output_path, new_iter_info)
         CALL cp_iteration_info_release(new_iter_info)
      ENDIF
   END SUBROUTINE logger_init_worker

! **************************************************************************************************
!> \brief Helper routine for logger_init_master and logger_init_worker
!> \param para_env ...
!> \param output_path ...
!> \param iter_info ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE error_add_new_logger(para_env, output_path, iter_info)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      CHARACTER(LEN=default_path_length)                 :: output_path
      TYPE(cp_iteration_info_type), OPTIONAL, POINTER    :: iter_info

      INTEGER                                            :: output_unit
      TYPE(cp_logger_type), POINTER                      :: new_logger, old_logger

      NULLIFY (new_logger, old_logger)
      output_unit = -1
      IF (para_env%source == para_env%mepos) THEN
         ! open output_unit according to output_path
         output_unit = default_output_unit
         IF (output_path /= "__STD_OUT__") &
            CALL open_file(file_name=output_path, file_status="UNKNOWN", &
                           file_action="WRITE", file_position="APPEND", unit_number=output_unit)
      ENDIF

      old_logger => cp_get_default_logger()
      CALL cp_logger_create(new_logger, para_env=para_env, &
                            default_global_unit_nr=output_unit, close_global_unit_on_dealloc=.FALSE., &
                            template_logger=old_logger, iter_info=iter_info)

      CALL cp_add_default_logger(new_logger)
      CALL cp_logger_release(new_logger)
   END SUBROUTINE error_add_new_logger

! **************************************************************************************************
!> \brief Finalizes the MPI communicators of a swarm run.
!> \param swarm_mpi ...
!> \param root_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_mpi_finalize(swarm_mpi, root_section)
      TYPE(swarm_mpi_type)                               :: swarm_mpi
      TYPE(section_vals_type), POINTER                   :: root_section

      CALL mp_sync(swarm_mpi%world%group)
      CALL logger_finalize(swarm_mpi, root_section)

      IF (ASSOCIATED(swarm_mpi%worker)) CALL cp_para_env_release(swarm_mpi%worker)
      IF (ASSOCIATED(swarm_mpi%master)) CALL cp_para_env_release(swarm_mpi%master)
      NULLIFY (swarm_mpi%worker, swarm_mpi%master)
      DEALLOCATE (swarm_mpi%wid2group)
   END SUBROUTINE swarm_mpi_finalize

! **************************************************************************************************
!> \brief Helper routine for swarm_mpi_finalize, restores the original loggers
!> \param swarm_mpi ...
!> \param root_section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE logger_finalize(swarm_mpi, root_section)
      TYPE(swarm_mpi_type)                               :: swarm_mpi
      TYPE(section_vals_type), POINTER                   :: root_section

      INTEGER                                            :: output_unit
      LOGICAL                                            :: I_am_rank0
      TYPE(cp_logger_type), POINTER                      :: logger, old_logger

      NULLIFY (logger, old_logger)
      logger => cp_get_default_logger()
      output_unit = logger%default_local_unit_nr
      IF (output_unit > 0 .AND. output_unit /= default_output_unit) &
         CALL close_file(output_unit)

      CALL cp_rm_default_logger() !pops the top-most logger
      old_logger => cp_get_default_logger()

      ! restore GLOBAL%PROJECT_NAME
      CALL section_vals_val_set(root_section, "GLOBAL%PROJECT_NAME", &
                                c_val=old_logger%iter_info%project_name)

      CALL mp_sync(swarm_mpi%world%group)

      ! do this only on master's rank 0
      I_am_rank0 = swarm_mpi%world%mepos == swarm_mpi%world%source
      IF (I_am_rank0 .AND. output_unit /= default_output_unit) THEN
         output_unit = old_logger%default_local_unit_nr
         OPEN (unit=output_unit, file=swarm_mpi%master_output_path, &
               status="UNKNOWN", action="WRITE", position="APPEND")
      ENDIF
   END SUBROUTINE logger_finalize

! **************************************************************************************************
!> \brief Sends a report via MPI
!> \param swarm_mpi ...
!> \param report ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_mpi_send_report(swarm_mpi, report)
      TYPE(swarm_mpi_type)                               :: swarm_mpi
      TYPE(swarm_message_type)                           :: report

      INTEGER                                            :: dest, tag

! Only rank-0 of worker group sends it's report

      IF (swarm_mpi%worker%source /= swarm_mpi%worker%mepos) RETURN

      dest = swarm_mpi%world%num_pe-1
      tag = 42
      CALL swarm_message_mpi_send(report, group=swarm_mpi%world%group, dest=dest, tag=tag)

   END SUBROUTINE swarm_mpi_send_report

! **************************************************************************************************
!> \brief Receives a report via MPI
!> \param swarm_mpi ...
!> \param report ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_mpi_recv_report(swarm_mpi, report)
      TYPE(swarm_mpi_type)                               :: swarm_mpi
      TYPE(swarm_message_type), INTENT(OUT)              :: report

      INTEGER                                            :: src, tag

      tag = 42
      src = mp_any_source

      CALL swarm_message_mpi_recv(report, group=swarm_mpi%world%group, src=src, tag=tag)

   END SUBROUTINE swarm_mpi_recv_report

! **************************************************************************************************
!> \brief Sends a command via MPI
!> \param swarm_mpi ...
!> \param cmd ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_mpi_send_command(swarm_mpi, cmd)
      TYPE(swarm_mpi_type)                               :: swarm_mpi
      TYPE(swarm_message_type)                           :: cmd

      INTEGER                                            :: dest, tag, worker_id

      CALL swarm_message_get(cmd, "worker_id", worker_id)
      tag = 42
      dest = swarm_mpi%wid2group(worker_id)

      CALL swarm_message_mpi_send(cmd, group=swarm_mpi%world%group, dest=dest, tag=tag)

   END SUBROUTINE swarm_mpi_send_command

! **************************************************************************************************
!> \brief Receives a command via MPI and broadcasts it within a worker.
!> \param swarm_mpi ...
!> \param cmd ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_mpi_recv_command(swarm_mpi, cmd)
      TYPE(swarm_mpi_type)                               :: swarm_mpi
      TYPE(swarm_message_type), INTENT(OUT)              :: cmd

      INTEGER                                            :: src, tag

! This is a two step communication schema.
! First: The rank-0 of the worker groups receives the command from the master.

      IF (swarm_mpi%worker%source == swarm_mpi%worker%mepos) THEN
         src = swarm_mpi%world%num_pe-1 !
         tag = 42
         CALL swarm_message_mpi_recv(cmd, group=swarm_mpi%world%group, src=src, tag=tag)

      ENDIF

!     ! Second: The command is broadcasted within the worker group.
      CALL swarm_message_mpi_bcast(cmd, src=swarm_mpi%worker%source, group=swarm_mpi%worker%group)

   END SUBROUTINE swarm_mpi_recv_command

END MODULE swarm_mpi

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Swarm-framwork, provides a convenient master/worker architecture.
!> \author Ole Schuett
! **************************************************************************************************
MODULE swarm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length
   USE swarm_master,                    ONLY: swarm_master_finalize,&
                                              swarm_master_init,&
                                              swarm_master_steer,&
                                              swarm_master_type
   USE swarm_message,                   ONLY: swarm_message_add,&
                                              swarm_message_free,&
                                              swarm_message_get,&
                                              swarm_message_type
   USE swarm_mpi,                       ONLY: swarm_mpi_finalize,&
                                              swarm_mpi_init,&
                                              swarm_mpi_recv_command,&
                                              swarm_mpi_recv_report,&
                                              swarm_mpi_send_command,&
                                              swarm_mpi_send_report,&
                                              swarm_mpi_type
   USE swarm_worker,                    ONLY: swarm_worker_execute,&
                                              swarm_worker_finalize,&
                                              swarm_worker_init,&
                                              swarm_worker_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'swarm'

   PUBLIC :: run_swarm

CONTAINS

! **************************************************************************************************
!> \brief Central driver routine of the swarm framework, called by cp2k_runs.F
!> \param input_declaration ...
!> \param root_section ...
!> \param para_env ...
!> \param globenv ...
!> \param input_path ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE run_swarm(input_declaration, root_section, para_env, globenv, input_path)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      CHARACTER(LEN=*), INTENT(IN)                       :: input_path

      CHARACTER(len=*), PARAMETER :: routineN = 'run_swarm', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw, n_workers
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, root_section, &
                                "SWARM%PRINT%MASTER_RUN_INFO", extension=".masterLog")

      IF (iw > 0) WRITE (iw, "(A)") " SWARM| Ready to roll :-)"

      CALL section_vals_val_get(root_section, "SWARM%NUMBER_OF_WORKERS", &
                                i_val=n_workers)

      IF (n_workers == 1 .AND. para_env%num_pe == 1) THEN
         IF (iw > 0) WRITE (iw, "(A)") " SWARM| Running in single worker mode."
         CALL swarm_serial_driver(input_declaration, root_section, input_path, para_env, globenv)
      ELSE
         IF (iw > 0) WRITE (iw, "(A)") " SWARM| Running in master / workers mode."
         !printkey iw passed on for output from swarm_mpi_init()
         CALL swarm_parallel_driver(n_workers, input_declaration, root_section, input_path, para_env, globenv, iw)
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE run_swarm

! **************************************************************************************************
!> \brief Special driver for using only a single worker.
!> \param input_declaration ...
!> \param root_section ...
!> \param input_path ...
!> \param para_env ...
!> \param globenv ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_serial_driver(input_declaration, root_section, input_path, para_env, globenv)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      CHARACTER(LEN=*), INTENT(IN)                       :: input_path
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      INTEGER                                            :: handle
      LOGICAL                                            :: should_stop
      TYPE(swarm_master_type)                            :: master
      TYPE(swarm_message_type)                           :: cmd, report
      TYPE(swarm_worker_type)                            :: worker

      CALL swarm_master_init(master, para_env, globenv, root_section, n_workers=1)
      CALL swarm_worker_init(worker, para_env, input_declaration, root_section, &
                             input_path, worker_id=1)

      CALL swarm_message_add(report, "worker_id", 1)
      CALL swarm_message_add(report, "status", "initial_hello")

      should_stop = .FALSE.
      DO WHILE (.NOT. should_stop)
         CALL timeset("swarm_worker_await_reply", handle)
         CALL swarm_master_steer(master, report, cmd)
         CALL timestop(handle)
         CALL swarm_message_free(report)
         CALL swarm_worker_execute(worker, cmd, report, should_stop)
         CALL swarm_message_free(cmd)
      END DO

      CALL swarm_message_free(report)
      CALL swarm_worker_finalize(worker)
      CALL swarm_master_finalize(master)

   END SUBROUTINE swarm_serial_driver

! **************************************************************************************************
!> \brief Normal driver routine for parallelized runs.
!> \param n_workers ...
!> \param input_declaration ...
!> \param root_section ...
!> \param input_path ...
!> \param para_env ...
!> \param globenv ...
!> \param iw ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_parallel_driver(n_workers, input_declaration, root_section, input_path, para_env, globenv, iw)
      INTEGER, INTENT(IN)                                :: n_workers
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      CHARACTER(LEN=*), INTENT(IN)                       :: input_path
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      INTEGER, INTENT(IN)                                :: iw

      INTEGER                                            :: worker_id
      TYPE(swarm_mpi_type)                               :: swarm_mpi

      CALL swarm_mpi_init(swarm_mpi, para_env, root_section, n_workers, worker_id, iw)

      IF (ASSOCIATED(swarm_mpi%worker)) THEN
         CALL swarm_parallel_worker_driver(swarm_mpi, input_declaration, worker_id, root_section, input_path)
      ELSE
         CALL swarm_parallel_master_driver(swarm_mpi, n_workers, root_section, globenv)
      END IF

      CALL swarm_mpi_finalize(swarm_mpi, root_section)

   END SUBROUTINE swarm_parallel_driver

! **************************************************************************************************
!> \brief Worker's driver routine for parallelized runs.
!> \param swarm_mpi ...
!> \param input_declaration ...
!> \param worker_id ...
!> \param root_section ...
!> \param input_path ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_parallel_worker_driver(swarm_mpi, input_declaration, worker_id, root_section, input_path)
      TYPE(swarm_mpi_type), INTENT(IN)                   :: swarm_mpi
      TYPE(section_type), POINTER                        :: input_declaration
      INTEGER, INTENT(IN)                                :: worker_id
      TYPE(section_vals_type), POINTER                   :: root_section
      CHARACTER(LEN=*), INTENT(IN)                       :: input_path

      INTEGER                                            :: handle
      LOGICAL                                            :: should_stop
      TYPE(swarm_message_type)                           :: cmd, report
      TYPE(swarm_worker_type)                            :: worker

      CALL swarm_worker_init(worker, swarm_mpi%worker, input_declaration, &
                             root_section, input_path, worker_id=worker_id)

      CALL swarm_message_add(report, "worker_id", worker_id)
      CALL swarm_message_add(report, "status", "initial_hello")

      should_stop = .FALSE.
      DO WHILE (.NOT. should_stop)
         CALL timeset("swarm_worker_await_reply", handle)
         CALL swarm_mpi_send_report(swarm_mpi, report)
         CALL swarm_message_free(report)
         CALL swarm_mpi_recv_command(swarm_mpi, cmd)
         CALL timestop(handle)
         CALL swarm_worker_execute(worker, cmd, report, should_stop)
         CALL swarm_message_free(cmd)
      END DO

      CALL swarm_message_free(report)
      CALL swarm_worker_finalize(worker)

   END SUBROUTINE swarm_parallel_worker_driver

! **************************************************************************************************
!> \brief Master's driver routine for parallelized runs.
!> \param swarm_mpi ...
!> \param n_workers ...
!> \param root_section ...
!> \param globenv ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE swarm_parallel_master_driver(swarm_mpi, n_workers, root_section, globenv)
      TYPE(swarm_mpi_type), INTENT(IN)                   :: swarm_mpi
      INTEGER, INTENT(IN)                                :: n_workers
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=default_string_length)               :: command
      INTEGER                                            :: i_shutdowns, j, wid
      LOGICAL, DIMENSION(n_workers)                      :: is_waiting
      TYPE(swarm_master_type)                            :: master
      TYPE(swarm_message_type)                           :: cmd, report

      is_waiting(:) = .FALSE.

      CALL swarm_master_init(master, swarm_mpi%master, globenv, root_section, n_workers)

      i_shutdowns = 0
      j = 0

      DO WHILE (i_shutdowns < n_workers)
         ! Each iteration if the loop does s.th. different depending on j.
         ! First (j==0) it receives one report with (blocking) MPI,
         ! then it searches through the list is_waiting.
         j = MOD(j+1, n_workers+1)
         IF (j == 0) THEN
            CALL swarm_mpi_recv_report(swarm_mpi, report)
         ELSE IF (is_waiting(j)) THEN
            is_waiting(j) = .FALSE.
            CALL swarm_message_add(report, "worker_id", j)
            CALL swarm_message_add(report, "status", "wait_done")
         ELSE
            CYCLE
         ENDIF

         CALL swarm_master_steer(master, report, cmd)
         CALL swarm_message_free(report)

         CALL swarm_message_get(cmd, "command", command)
         IF (TRIM(command) == "wait") THEN
            CALL swarm_message_get(cmd, "worker_id", wid)
            is_waiting(wid) = .TRUE.
         ELSE
            CALL swarm_mpi_send_command(swarm_mpi, cmd)
            IF (TRIM(command) == "shutdown") i_shutdowns = i_shutdowns+1
         ENDIF
         CALL swarm_message_free(cmd)
      END DO

      CALL swarm_master_finalize(master)

   END SUBROUTINE swarm_parallel_master_driver

END MODULE swarm

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module handels definition of the tree nodes for the global and
!>      the subtrees binary tree
!>                   parent element
!>                      /      !>      accepted (acc) /        \  not accepted (nacc)
!>                    /          !>                  child       child
!>                   / \         / !>
!>      tree creation assuming acceptance (acc) AND rejectance (nacc)
!>        of configuration
!>      if configuration is accepted: new configuration (child on acc) on basis
!>        of last configuration (one level up)
!>      if configuration is rejected: child on nacc on basis of last accepted
!>        element (last element which is on acc brach of its parent element)
!>      The global tree handles all configurations of different subtrees.
!>      The structure element "conf" is an array related to the temperature
!>        (sorted) and points to the subtree elements.
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_tree_types
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_tree_types'

   PUBLIC :: tree_type, global_tree_type
   PUBLIC :: elem_list_type, elem_array_type, gt_elem_list_type
   PUBLIC :: add_to_list, clean_list
   PUBLIC :: read_subtree_elem_unformated, write_subtree_elem_unformated

   !-- tree element status
   INTEGER, PARAMETER, PUBLIC :: status_created = 100
   INTEGER, PARAMETER, PUBLIC :: status_calculate_energy = 101
   INTEGER, PARAMETER, PUBLIC :: status_calc_approx_ener = 102

   INTEGER, PARAMETER, PUBLIC :: status_calculate_NMC_steps = 111
   INTEGER, PARAMETER, PUBLIC :: status_calculate_MD = 112
   INTEGER, PARAMETER, PUBLIC :: status_calculated = 113

   INTEGER, PARAMETER, PUBLIC :: status_accepted_result = 123
   INTEGER, PARAMETER, PUBLIC :: status_accepted = 122
   INTEGER, PARAMETER, PUBLIC :: status_rejected = 121
   INTEGER, PARAMETER, PUBLIC :: status_rejected_result = 120

   INTEGER, PARAMETER, PUBLIC :: status_cancel_nmc = 133
   INTEGER, PARAMETER, PUBLIC :: status_cancel_ener = 132
   INTEGER, PARAMETER, PUBLIC :: status_canceled_nmc = 131
   INTEGER, PARAMETER, PUBLIC :: status_canceled_ener = 130

   INTEGER, PARAMETER, PUBLIC :: status_deleted = 140
   INTEGER, PARAMETER, PUBLIC :: status_deleted_result = 141

   !-- dimension status (for e.g. dividing atoms in sub box)
   INTEGER, PARAMETER, PUBLIC :: status_ok = 42
   INTEGER, PARAMETER, PUBLIC :: status_frozen = -1
   INTEGER, PARAMETER, PUBLIC :: status_proton_disorder = 1

   !-- subtree element
   TYPE tree_type
      TYPE(tree_type), POINTER                :: parent => NULL() ! points to element one level up
      !-- acc..accepted goes to next level (next step),
      !   nacc..not accepted takes an alternative configutation
      TYPE(tree_type), POINTER                :: acc => NULL(), nacc => NULL()
      !-- type of MC move (swap is handled only in global tree)
      INTEGER                                  :: move_type
      !-- status (e.g. calculated, MD calculation, accepted...)
      INTEGER                                  :: stat = status_created
      REAL(KIND=dp), DIMENSION(:), POINTER     :: subbox_center
      REAL(KIND=dp), DIMENSION(:), POINTER     :: pos ! position array
      INTEGER, DIMENSION(:), POINTER           :: mol ! specifies the molecules the atoms participate
      REAL(KIND=dp), DIMENSION(:), POINTER     :: vel ! velocity array
      REAL(KIND=dp), DIMENSION(:), POINTER     :: frc ! force array
      REAL(KIND=dp), DIMENSION(:), POINTER     :: dipole ! dipole moments array
      INTEGER, DIMENSION(:), POINTER           :: elem_stat ! status for every dimention
      INTEGER                                  :: nr ! tree node number
      REAL(KIND=dp), DIMENSION(3, 2, 3)        :: rng_seed ! random seed for childs
      !-- remembers which subtree number element is from
      INTEGER                                  :: sub_tree_nr
      !-- remembers the temperature the configurational change (NMC) is done with
      INTEGER                                  :: temp_created = 0
      !-- pointer to counter of next subtree element number
      INTEGER, POINTER                         :: next_elem_nr
      !-- for calculating the NPT ensamble, variable box sizes are necessary.
      REAL(KIND=dp), DIMENSION(:), POINTER     :: box_scale
      REAL(KIND=dp)                            :: potential ! potential energy
      !-- potential energy calculated using (MD potential) cp2k input file
      REAL(KIND=dp)                            :: e_pot_approx = 0.0_dp
      !-- kinetic energy (espacially for HMC, where the velocities are respected)
      REAL(KIND=dp)                            :: ekin
      !-- kinetic energy before md steps (after gaussian velocity change)
      REAL(KIND=dp)                            :: ekin_before_md
      !-- estimated energies are stored in loop order in this array
      REAL(KIND=dp), DIMENSION(4)              :: scf_energies
      !-- counter to get last position in the array loop
      INTEGER                                  :: scf_energies_count
      !-- list of global tree elements referint to that node (reference back to global tree)
      !   if no reference exist anymore, global tree element can be deleted
      TYPE(gt_elem_list_type), POINTER         :: gt_nodes_references => NULL()
   END TYPE tree_type

   ! type for global tree element list in tree elements
   TYPE gt_elem_list_type
      TYPE(global_tree_type), POINTER         :: gt_elem => NULL()
      TYPE(gt_elem_list_type), POINTER        :: next => NULL()
   END TYPE gt_elem_list_type

   TYPE elem_list_type
      TYPE(tree_type), POINTER      :: elem => NULL()
      TYPE(elem_list_type), POINTER :: next => NULL()
      INTEGER                        :: temp_ind
      INTEGER                        :: nr
   END TYPE elem_list_type

   !-- array with subtree elements
   TYPE elem_array_type
      TYPE(tree_type), POINTER :: elem => NULL()
      LOGICAL                   :: busy = .FALSE.
      LOGICAL                   :: canceled = .FALSE.
      REAL(KIND=dp)             :: start_time
   END TYPE elem_array_type

   !-- global tree element
   TYPE global_tree_type
      TYPE(global_tree_type), POINTER :: parent => NULL() ! points to element one level up
      !-- acc..accepted goes to next level (next step),
      !   nacc..not accepted takes an alternative configutation
      TYPE(global_tree_type), POINTER :: acc => NULL(), nacc => NULL()
      !-- status (e.g. calculated, MD calculation, accepted...)
      INTEGER                                      :: stat = -99
      !-- remember if configuration in node are swaped
      LOGICAL                                      :: swaped
      !-- stores the index of the configuration (temperature)
      !   which is changed
      INTEGER                                      :: mv_conf = -54321
      !-- stores the index of the configuration (temp.) which should change next
      INTEGER                                      :: mv_next_conf = -2345
      !-- list of pointes to subtree elements (Temp sorting)
      TYPE(elem_array_type), DIMENSION(:), ALLOCATABLE :: conf
      !-- remembers if last configuration is assumed to be accepted or rejected (next branc in tree);
      !   In case of swaping, it shows if the configuration of a certain temperature is assumed
      !   to be acc/rej (which branch is followed at the last modification of the conf of this temp.
      !TODO store conf_n_acc in a bitshifted array to decrease the size (1Logical = 1Byte)
      LOGICAL, DIMENSION(:), ALLOCATABLE           :: conf_n_acc
      INTEGER :: nr ! tree node number
      REAL(KIND=dp), DIMENSION(3, 2, 3)            :: rng_seed ! random seed for childs
      !-- random number for acceptance check
      REAL(KIND=dp)                                :: rnd_nr
      !-- approximate probability of acceptance will be adapted while calculating the exact energy
      REAL(KIND=dp)                                :: prob_acc ! estimated acceptance probability
      REAL(KIND=dp)                                :: Temp ! temperature for simulated annealing
   END TYPE global_tree_type

CONTAINS

! **************************************************************************************************
!> \brief add a certain element to the specified element list at the beginning
!> \param elem the sub tree element, to be added
!> \param list  ...
!> \param temp_ind ...
!> \param nr ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE add_to_list(elem, list, temp_ind, nr)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(elem_list_type), POINTER                      :: list
      INTEGER, OPTIONAL                                  :: temp_ind, nr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'add_to_list', routineP = moduleN//':'//routineN

      TYPE(elem_list_type), POINTER                      :: last, list_elem_tmp

      NULLIFY (list_elem_tmp, last)

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_tree_types.F",188)

      ALLOCATE (list_elem_tmp)
      list_elem_tmp%elem => elem
      list_elem_tmp%next => NULL()
      IF (PRESENT(temp_ind)) THEN
         list_elem_tmp%temp_ind = temp_ind
      ELSE
         list_elem_tmp%temp_ind = -1
      END IF

      IF (PRESENT(nr)) THEN
         list_elem_tmp%nr = nr
      ELSE
         list_elem_tmp%nr = -1
      END IF

      IF (ASSOCIATED(list) .EQV. .FALSE.) THEN
         list => list_elem_tmp
      ELSE
         last => list
         DO WHILE (ASSOCIATED(last%next))
            last => last%next
         END DO
         last%next => list_elem_tmp
      END IF

   END SUBROUTINE add_to_list

! **************************************************************************************************
!> \brief clean a certain element element list
!> \param list  ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE clean_list(list)
      TYPE(elem_list_type), POINTER                      :: list

      CHARACTER(LEN=*), PARAMETER :: routineN = 'clean_list', routineP = moduleN//':'//routineN

      TYPE(elem_list_type), POINTER                      :: list_elem_tmp

      NULLIFY (list_elem_tmp)

      DO WHILE (ASSOCIATED(list))
         list_elem_tmp => list%next
         DEALLOCATE (list)
         list => list_elem_tmp
      END DO
   END SUBROUTINE clean_list

! **************************************************************************************************
!> \brief prints out the TMC sub tree structure element unformated in file
!> \param elem ...
!> \param io_unit ...
!> \param
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE write_subtree_elem_unformated(elem, io_unit)
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: io_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_subtree_elem_unformated', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_tree_types.F",252)
      IF(.NOT.(io_unit .GT. 0))CALL cp__a("tmc/tmc_tree_types.F",253)
      WRITE (io_unit) elem%nr, &
         elem%sub_tree_nr, &
         elem%stat, &
         elem%rng_seed, &
         elem%move_type, &
         elem%temp_created, &
         elem%potential, &
         elem%e_pot_approx, &
         elem%ekin, &
         elem%ekin_before_md
      CALL write_subtree_elem_darray(elem%pos, io_unit)
      CALL write_subtree_elem_darray(elem%vel, io_unit)
      CALL write_subtree_elem_darray(elem%frc, io_unit)
      CALL write_subtree_elem_darray(elem%box_scale, io_unit)
      CALL write_subtree_elem_darray(elem%dipole, io_unit)
   END SUBROUTINE write_subtree_elem_unformated

! **************************************************************************************************
!> \brief reads the TMC sub tree structure element unformated in file
!> \param elem ...
!> \param io_unit ...
!> \param
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE read_subtree_elem_unformated(elem, io_unit)
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: io_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_subtree_elem_unformated', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_tree_types.F",285)
      IF(.NOT.(io_unit .GT. 0))CALL cp__a("tmc/tmc_tree_types.F",286)

      READ (io_unit) elem%nr, &
         elem%sub_tree_nr, &
         elem%stat, &
         elem%rng_seed, &
         elem%move_type, &
         elem%temp_created, &
         elem%potential, &
         elem%e_pot_approx, &
         elem%ekin, &
         elem%ekin_before_md
      CALL read_subtree_elem_darray(elem%pos, io_unit)
      CALL read_subtree_elem_darray(elem%vel, io_unit)
      CALL read_subtree_elem_darray(elem%frc, io_unit)
      CALL read_subtree_elem_darray(elem%box_scale, io_unit)
      CALL read_subtree_elem_darray(elem%dipole, io_unit)
   END SUBROUTINE read_subtree_elem_unformated

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param io_unit ...
! **************************************************************************************************
   SUBROUTINE write_subtree_elem_darray(array, io_unit)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: array
      INTEGER                                            :: io_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_subtree_elem_darray', &
         routineP = moduleN//':'//routineN

      WRITE (io_unit) ASSOCIATED(array)
      IF (ASSOCIATED(array)) THEN
         WRITE (io_unit) SIZE(array)
         WRITE (io_unit) array
      END IF
   END SUBROUTINE write_subtree_elem_darray

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param io_unit ...
! **************************************************************************************************
   SUBROUTINE read_subtree_elem_darray(array, io_unit)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: array
      INTEGER                                            :: io_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_subtree_elem_darray', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_tmp
      LOGICAL                                            :: l_tmp

      READ (io_unit) l_tmp
      IF (l_tmp) THEN
         READ (io_unit) i_tmp
         IF (ASSOCIATED(array)) THEN
            IF(.NOT.(SIZE(array) .EQ. i_tmp))CALL cp__a("tmc/tmc_tree_types.F",343)
         ELSE
            ALLOCATE (array(i_tmp))
         END IF
         READ (io_unit) array
      END IF
   END SUBROUTINE read_subtree_elem_darray

END MODULE tmc_tree_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief tree nodes creation, searching, deallocation, references etc.
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes 11/2012
! **************************************************************************************************

MODULE tmc_move_types
   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_move_types'

   !-- list of available move types
   INTEGER, PARAMETER, PUBLIC :: mv_type_none = 0
   INTEGER, PARAMETER, PUBLIC :: mv_type_swap_conf = 1 ! swapping of 2 configurations of different temperature
   INTEGER, PARAMETER, PUBLIC :: mv_type_atom_trans = 2 ! atom translation (done in every posible direction)
   INTEGER, PARAMETER, PUBLIC :: mv_type_mol_trans = 3 ! molecule translation (done in every posible direction)
   INTEGER, PARAMETER, PUBLIC :: mv_type_mol_rot = 4 ! molecule rotation
   INTEGER, PARAMETER, PUBLIC :: mv_type_proton_reorder = 5 ! reordering the protons within a chain of molecules
   INTEGER, PARAMETER, PUBLIC :: mv_type_atom_swap = 6 ! swaps two atoms of different type
   INTEGER, PARAMETER, PUBLIC :: mv_type_MD = 7 ! certain amount of MD steps
   INTEGER, PARAMETER, PUBLIC :: mv_type_volume_move = 8 ! volume move for NPT simulations
   INTEGER, PARAMETER, PUBLIC :: mv_type_gausian_adapt = 9 ! gaussian adaptation
   INTEGER, PARAMETER, PUBLIC :: mv_type_NMC_moves = 10 ! indentifies the Nested Monte Carlo move for master
   INTEGER, PARAMETER, PUBLIC :: nr_mv_types = 10 !-- allways update the number of possible types!!

   PUBLIC :: tmc_move_type, move_types_create, move_types_release

   TYPE tmc_move_type
      !-- mv_type, handling indeces to move type (are equal for all several configurations/temperatures)
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: mv_weight
      !-- mv_size, moves are normaly done in interval ]-mv_size, mv_size[
      ! 1st dimension are the different types, 2nd dim for configuration/temperature
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE :: mv_size
      !-- acc_prob, probability of acceptance of a certain move type for a certain temperature
      ! 1st dimension are the different move types, 2nd dim for configuration/temperature
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE :: acc_prob
      !-- count, remembers the certain amount of moves of certain a move type and temperature
      ! 1st dimension are the different types, 2nd dim for config./Temp
      INTEGER, DIMENSION(:, :), ALLOCATABLE :: mv_count
      !-- count, remembers the certain amount of accepted moves of a certain move type and temperature
      ! 1st dimension are the different types, 2nd dim for config./Temp
      INTEGER, DIMENSION(:, :), ALLOCATABLE :: acc_count
      !-- subbox_prob, probability of acceptance of a certain move type within subbox,
      !   done in Nested Monte Carlo routine
      !   the moves are rejected if atom or center of mass leaves the subbox
      !   1st dimension are the different move types
      INTEGER, DIMENSION(:, :), ALLOCATABLE :: subbox_acc_count
      INTEGER, DIMENSION(:, :), ALLOCATABLE :: subbox_count
      TYPE(list_atoms), DIMENSION(:), POINTER :: atom_lists

      !-- nmc_acc_prob, probability of acceptance of a certain move type,
      !   done in Nested Monte Carlo routine, for different potential
      !   1st dimension are the different move types
!     REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: nmc_nr_acc
!     INTEGER, DIMENSION(:), ALLOCATABLE :: nmc_count
   END TYPE tmc_move_type

   TYPE list_atoms
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                  :: atoms
   END TYPE list_atoms
CONTAINS

! **************************************************************************************************
!> \brief allocating the module variables
!> \param move_types pointer to the structure which should be deallocated
!> \param nr_temp ...
!> \author Mandes 11.2012
!> \note deallocating the module variables
! **************************************************************************************************
   SUBROUTINE move_types_create(move_types, nr_temp)
      TYPE(tmc_move_type), POINTER                       :: move_types
      INTEGER                                            :: nr_temp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'move_types_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_move_types.F",90)

      ALLOCATE (move_types)
      ALLOCATE (move_types%mv_weight(nr_mv_types))
      move_types%mv_weight(:) = 0.0_dp
      ALLOCATE (move_types%mv_size(nr_mv_types, nr_temp))
      move_types%mv_size(:, :) = 0.0_dp
      ALLOCATE (move_types%acc_prob(0:nr_mv_types, nr_temp))
      move_types%acc_prob(:, :) = 0.0_dp
      ALLOCATE (move_types%mv_count(0:nr_mv_types, nr_temp))
      move_types%mv_count(:, :) = 0
      ALLOCATE (move_types%acc_count(0:nr_mv_types, nr_temp))
      move_types%acc_count(:, :) = 0
      ALLOCATE (move_types%subbox_acc_count(nr_mv_types, nr_temp))
      move_types%subbox_acc_count(:, :) = 0
      ALLOCATE (move_types%subbox_count(nr_mv_types, nr_temp))
      move_types%subbox_count(:, :) = 0
      NULLIFY (move_types%atom_lists)
   END SUBROUTINE move_types_create

! **************************************************************************************************
!> \brief deallocating the module variables
!> \param move_types pointer to the structure which should be deallocated
!> \author Mandes 11.2012
!> \note deallocating the module variables
! **************************************************************************************************
   SUBROUTINE move_types_release(move_types)
      TYPE(tmc_move_type), POINTER                       :: move_types

      CHARACTER(LEN=*), PARAMETER :: routineN = 'move_types_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_move_types.F",122)

      IF (ASSOCIATED(move_types%atom_lists)) DEALLOCATE (move_types%atom_lists)
      DEALLOCATE (move_types%mv_weight)
      DEALLOCATE (move_types%mv_size)
      DEALLOCATE (move_types%acc_prob)
      DEALLOCATE (move_types%mv_count)
      DEALLOCATE (move_types%acc_count)
      DEALLOCATE (move_types%subbox_acc_count)
      DEALLOCATE (move_types%subbox_count)
      DEALLOCATE (move_types)
   END SUBROUTINE move_types_release

END MODULE tmc_move_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module handels definition of the tree nodes for the global and
!>      the subtrees binary tree
!>                   parent element
!>                      /      !>      accepted (acc) /        \  not accepted (nacc)
!>                    /          !>                  child       child
!>                   / \         / !>
!>      tree creation assuming acceptance (acc) AND rejectance (nacc)
!>        of configuration
!>      if configuration is accepted: new configuration (child on acc) on basis
!>        of last configuration (one level up)
!>      if configuration is rejected: child on nacc on basis of last accepted
!>        element (last element which is on acc brach of its parent element)
!>      The global tree handles all configurations of different subtrees.
!>      The structure element "conf" is an array related to the temperature
!>        (sorted) and points to the subtree elements.
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_types
   USE cell_types,                      ONLY: cell_type
   USE cp_para_env,                     ONLY: cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE parallel_rng_types,              ONLY: rng_stream_type
   USE tmc_move_types,                  ONLY: tmc_move_type
   USE tmc_stati,                       ONLY: task_type_MC
   USE tmc_tree_types,                  ONLY: clean_list,&
                                              elem_array_type,&
                                              elem_list_type,&
                                              global_tree_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_types'

   PUBLIC :: tmc_env_type, tmc_comp_set_type, tmc_param_type
   PUBLIC :: tmc_env_create, tmc_env_release, &
             tmc_master_env_create, tmc_master_env_release, &
             tmc_worker_env_create, tmc_worker_env_release
   PUBLIC :: tmc_atom_type
   PUBLIC :: allocate_tmc_atom_type

   ! global environment
   TYPE tmc_env_type
      TYPE(tmc_comp_set_type), POINTER              :: tmc_comp_set
      TYPE(tmc_param_type), POINTER                 :: params
      TYPE(rng_stream_type), POINTER                :: rng_stream
      TYPE(master_env_type), POINTER                :: m_env
      TYPE(worker_env_type), POINTER                :: w_env
   END TYPE tmc_env_type

   ! structure for remembering the main values used for reordering MPI communicators
   ! \param group_nr: the first group_ener_nr groups are for energy calculation,
   !                  then the Configurational Change groups,
   !                  the global master has group 0 and unused cores have negative group numbering
   TYPE tmc_comp_set_type
      INTEGER                                       :: group_ener_size
      INTEGER                                       :: group_ener_nr
      INTEGER                                       :: group_cc_size
      INTEGER                                       :: group_cc_nr
      INTEGER                                       :: group_nr
      INTEGER                                       :: ana_on_the_fly
      ! the communicators (para_env)
      TYPE(cp_para_env_type), POINTER               :: para_env_m_w
      TYPE(cp_para_env_type), POINTER               :: para_env_sub_group
      TYPE(cp_para_env_type), POINTER               :: para_env_m_first_w
      TYPE(cp_para_env_type), POINTER               :: para_env_m_ana
      TYPE(cp_para_env_type), POINTER               :: para_env_m_only
   END TYPE tmc_comp_set_type

   ! struct for TMC global variables
   TYPE tmc_param_type
      INTEGER                                       :: task_type
      INTEGER                                       :: dim_per_elem
      INTEGER                                       :: nr_temp
      REAL(KIND=dp), DIMENSION(:), POINTER          :: Temp
      TYPE(cell_type), POINTER                      :: cell
      REAL(KIND=dp), DIMENSION(:), POINTER          :: sub_box_size
      TYPE(tmc_atom_type), DIMENSION(:), POINTER    :: atoms

      INTEGER                                       :: nr_elem_mv
      TYPE(tmc_move_type), POINTER                  :: move_types
      TYPE(tmc_move_type), POINTER                  :: nmc_move_types
      REAL(KIND=dp)                                 :: pressure
      LOGICAL                                       :: v_isotropic
      LOGICAL                                       :: mv_cen_of_mass
      LOGICAL                                       :: esimate_acc_prob
      LOGICAL                                       :: SPECULATIVE_CANCELING
      LOGICAL                                       :: use_scf_energy_info
      LOGICAL                                       :: USE_REDUCED_TREE
      CHARACTER(LEN=default_path_length)          :: energy_inp_file
      CHARACTER(LEN=default_path_length)          :: NMC_inp_file
      LOGICAL                                       :: DRAW_TREE
      CHARACTER(LEN=default_path_length)          :: dot_file_name
      CHARACTER(LEN=default_path_length)          :: all_conf_file_name
      LOGICAL                                       :: print_only_diff_conf
      LOGICAL                                       :: print_trajectory
      LOGICAL                                       :: print_dipole
      LOGICAL                                       :: print_forces
      LOGICAL                                       :: print_cell
      LOGICAL                                       :: print_energies
      TYPE(prior_estimate_acceptance_type), POINTER :: prior_NMC_acc
      LOGICAL                                       :: print_test_output
   END TYPE tmc_param_type

   TYPE tmc_atom_type
      CHARACTER(LEN=default_string_length)   :: name
      REAL(KIND=dp)                        :: mass
      !TYPE(atomic_kind_type), POINTER :: atomic_kind
   END TYPE

   ! to estimate the prior acceptance
   TYPE prior_estimate_acceptance_type
      INTEGER :: counter
      REAL(KIND=dp) :: aver, aver_2
   END TYPE prior_estimate_acceptance_type

   ! environments for the master
   TYPE master_env_type
      INTEGER                                       :: num_MC_elem ! the specified number of Markov Chain elements, to be reached
      CHARACTER(LEN=default_path_length)          :: restart_in_file_name
      CHARACTER(LEN=default_path_length)          :: restart_out_file_name
      INTEGER                                       :: restart_out_step
      INTEGER                                       :: io_unit
      INTEGER                                       :: info_out_step_size
      REAL(KIND=dp)                                 :: walltime
      INTEGER                                       :: rnd_init
      REAL(KIND=dp)                                 :: temp_decrease ! for simulated annealing
      TYPE(elem_list_type), POINTER                 :: cancelation_list
      INTEGER                                       :: count_cancel_ener
      INTEGER                                       :: count_cancel_NMC
      ! masters tree stuff
      TYPE(global_tree_type), POINTER               :: gt_head, gt_act
      INTEGER, DIMENSION(:), POINTER                :: tree_node_count
      INTEGER, DIMENSION(:), POINTER                :: result_count
      TYPE(elem_array_type), DIMENSION(:), &
         POINTER                                  :: result_list, &
                                                     st_heads, &
                                                     st_clean_ends
      TYPE(global_tree_type), POINTER              :: gt_clean_end
      INTEGER, DIMENSION(4)                         :: estim_corr_wrong
      TYPE(elem_list_type), POINTER                 :: analysis_list
   END TYPE master_env_type

   ! environment for the worker
   TYPE worker_env_type
      INTEGER                                       :: env_id_ener, env_id_approx
      INTEGER                                       :: io_unit
      REAL(KIND=dp)                                 :: act_temp
   END TYPE worker_env_type

CONTAINS

! **************************************************************************************************
!> \brief creates a new structure environment for TMC
!> \param tmc_env structure with parameters for TMC
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_env_create(tmc_env)
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_env_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_types.F",184)

      ALLOCATE (tmc_env)

      ALLOCATE (tmc_env%tmc_comp_set)

      tmc_env%tmc_comp_set%group_ener_size = 0
      tmc_env%tmc_comp_set%group_ener_nr = 0
      tmc_env%tmc_comp_set%group_cc_size = 0
      tmc_env%tmc_comp_set%group_cc_nr = 0
      tmc_env%tmc_comp_set%group_nr = 0
      tmc_env%tmc_comp_set%ana_on_the_fly = 0
      NULLIFY (tmc_env%tmc_comp_set%para_env_m_w)
      NULLIFY (tmc_env%tmc_comp_set%para_env_sub_group)
      NULLIFY (tmc_env%tmc_comp_set%para_env_m_first_w)
      NULLIFY (tmc_env%tmc_comp_set%para_env_m_ana)
      NULLIFY (tmc_env%tmc_comp_set%para_env_m_only)

      NULLIFY (tmc_env%rng_stream)
      NULLIFY (tmc_env%m_env, tmc_env%w_env)

      ! initialize the parameter section
      ALLOCATE (tmc_env%params)

      tmc_env%params%task_type = task_type_MC
      ! dim_per_elem can be variable somewhen,
      !  I used it originally for the n dimension oscillator,
      !  where each dimension is a single position
      tmc_env%params%dim_per_elem = 3
      tmc_env%params%nr_elem_mv = -1
      tmc_env%params%nr_temp = -1
      ALLOCATE (tmc_env%params%sub_box_size(tmc_env%params%dim_per_elem))
      tmc_env%params%sub_box_size(:) = -1.0_dp
      tmc_env%params%pressure = -1
      tmc_env%params%SPECULATIVE_CANCELING = .FALSE.
      tmc_env%params%use_scf_energy_info = .FALSE.
      tmc_env%params%energy_inp_file = ""
      tmc_env%params%NMC_inp_file = ""
      tmc_env%params%DRAW_TREE = .FALSE.
      tmc_env%params%print_dipole = .FALSE.
      tmc_env%params%print_forces = .FALSE.
      tmc_env%params%print_cell = .FALSE.
      NULLIFY (tmc_env%params%Temp, tmc_env%params%nmc_move_types, &
               tmc_env%params%move_types, &
               tmc_env%params%prior_NMC_acc, tmc_env%params%atoms, &
               tmc_env%params%cell)

      CALL timestop(handle)

   END SUBROUTINE tmc_env_create

! **************************************************************************************************
!> \brief releases the structure environment for TMC
!> \param tmc_env structure with parameters for TMC
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_env_release(tmc_env)
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_types.F",250)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_types.F",251)

      DEALLOCATE (tmc_env%params%sub_box_size)
      IF (ASSOCIATED(tmc_env%params%Temp)) &
         DEALLOCATE (tmc_env%params%Temp)
      IF (ASSOCIATED(tmc_env%params%cell)) &
         DEALLOCATE (tmc_env%params%cell)
      IF (ASSOCIATED(tmc_env%params%atoms)) &
         CALL deallocate_tmc_atom_type(tmc_env%params%atoms)
      DEALLOCATE (tmc_env%params)

      CALL cp_para_env_release(tmc_env%tmc_comp_set%para_env_sub_group)
      CALL cp_para_env_release(tmc_env%tmc_comp_set%para_env_m_w)
      IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_first_w)) &
         CALL cp_para_env_release(tmc_env%tmc_comp_set%para_env_m_first_w)
      IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_ana)) &
         CALL cp_para_env_release(tmc_env%tmc_comp_set%para_env_m_ana)
      IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_only)) &
         CALL cp_para_env_release(tmc_env%tmc_comp_set%para_env_m_only)

      DEALLOCATE (tmc_env%tmc_comp_set)

      DEALLOCATE (tmc_env)

      CALL timestop(handle)

   END SUBROUTINE tmc_env_release

! **************************************************************************************************
!> \brief creates a new structure environment for TMC master
!> \param tmc_env structure with parameters for TMC
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_master_env_create(tmc_env)
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_master_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_types.F",294)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_types.F",295)
      IF(.NOT.(tmc_env%params%nr_temp .GT. 0))CALL cp__a("tmc/tmc_types.F",296)

      IF(.NOT.(.NOT. ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_types.F",298)

      ALLOCATE (tmc_env%m_env)
      NULLIFY (tmc_env%m_env%gt_head, tmc_env%m_env%gt_act, tmc_env%m_env%tree_node_count, &
               tmc_env%m_env%result_count, tmc_env%m_env%result_list, &
               tmc_env%m_env%st_heads, tmc_env%m_env%st_clean_ends, &
               tmc_env%m_env%gt_clean_end, tmc_env%m_env%cancelation_list, tmc_env%m_env%analysis_list)

      tmc_env%m_env%restart_in_file_name = ""
      tmc_env%m_env%restart_out_file_name = ""
      ALLOCATE (tmc_env%m_env%tree_node_count(0:tmc_env%params%nr_temp))
      tmc_env%m_env%tree_node_count(:) = 0
      ALLOCATE (tmc_env%m_env%result_count(0:tmc_env%params%nr_temp))
      tmc_env%m_env%result_count(:) = 0
      ALLOCATE (tmc_env%m_env%st_heads(tmc_env%params%nr_temp))
      ALLOCATE (tmc_env%m_env%st_clean_ends(tmc_env%params%nr_temp))

      IF (tmc_env%params%USE_REDUCED_TREE) ALLOCATE (tmc_env%m_env%result_list(tmc_env%params%nr_temp))

      DO i = 1, tmc_env%params%nr_temp
         tmc_env%m_env%st_heads(i)%elem => NULL()
         tmc_env%m_env%st_clean_ends(i)%elem => NULL()
         IF (tmc_env%params%USE_REDUCED_TREE) &
            tmc_env%m_env%result_list(i)%elem => NULL()
      END DO
      tmc_env%m_env%gt_head => NULL()
      tmc_env%m_env%gt_clean_end => NULL()
      tmc_env%m_env%temp_decrease = 1.0_dp
      tmc_env%m_env%count_cancel_ener = 0
      tmc_env%m_env%count_cancel_NMC = 0
      tmc_env%m_env%estim_corr_wrong(:) = 0

      ALLOCATE (tmc_env%params%prior_NMC_acc)
      tmc_env%params%prior_NMC_acc%counter = 0
      tmc_env%params%prior_NMC_acc%aver = 0.0_dp
      tmc_env%params%prior_NMC_acc%aver_2 = 0.0_dp

      CALL timestop(handle)

   END SUBROUTINE tmc_master_env_create

! **************************************************************************************************
!> \brief releases the structure environment for TMC master
!> \param tmc_env structure with parameters for TMC
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_master_env_release(tmc_env)
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_master_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_types.F",354)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_types.F",355)

      CALL clean_list(tmc_env%m_env%analysis_list)
      CALL clean_list(tmc_env%m_env%cancelation_list)

      DEALLOCATE (tmc_env%m_env%tree_node_count)
      DEALLOCATE (tmc_env%m_env%result_count)
      DEALLOCATE (tmc_env%m_env%st_heads)
      DEALLOCATE (tmc_env%m_env%st_clean_ends)
      IF (tmc_env%params%USE_REDUCED_TREE) DEALLOCATE (tmc_env%m_env%result_list)
      DEALLOCATE (tmc_env%params%prior_NMC_acc)

      DEALLOCATE (tmc_env%m_env)

      CALL timestop(handle)

   END SUBROUTINE tmc_master_env_release

! **************************************************************************************************
!> \brief creates a new structure environment for TMC master
!> \param tmc_env structure with parameters for TMC
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_worker_env_create(tmc_env)
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_worker_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_types.F",388)
      IF(.NOT.(.NOT. ASSOCIATED(tmc_env%w_env)))CALL cp__a("tmc/tmc_types.F",389)

      ALLOCATE (tmc_env%w_env)

      tmc_env%w_env%env_id_ener = -1
      tmc_env%w_env%env_id_approx = -1
      tmc_env%w_env%io_unit = -1
      tmc_env%w_env%act_temp = -1.0_dp

      CALL timestop(handle)

   END SUBROUTINE tmc_worker_env_create

! **************************************************************************************************
!> \brief releases the structure environment for TMC master
!> \param tmc_env structure with parameters for TMC
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_worker_env_release(tmc_env)
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_worker_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_types.F",417)
      IF(.NOT.(ASSOCIATED(tmc_env%w_env)))CALL cp__a("tmc/tmc_types.F",418)

      DEALLOCATE (tmc_env%w_env)

      CALL timestop(handle)

   END SUBROUTINE tmc_worker_env_release

! **************************************************************************************************
!> \brief creates a structure for storing the atom informations
!> \param atoms pointer to a list of tmc_atoms_type
!> \param nr_atoms the amount of atoms
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE allocate_tmc_atom_type(atoms, nr_atoms)
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms
      INTEGER, INTENT(IN)                                :: nr_atoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_tmc_atom_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_types.F",439)
      IF(.NOT.(nr_atoms .GT. 0))CALL cp__a("tmc/tmc_types.F",440)

      ALLOCATE (atoms(nr_atoms))
      atoms%name = ""
      atoms%mass = 0.0_dp

      IF(.NOT.(ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_types.F",446)

   END SUBROUTINE allocate_tmc_atom_type

! **************************************************************************************************
!> \brief releases the structure for storing the atom informations
!> \param atoms pointer to a list of tmc_atoms_type
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE deallocate_tmc_atom_type(atoms)
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_tmc_atom_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_types.F",461)

      DEALLOCATE (atoms)

      IF(.NOT.(.NOT. ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_types.F",465)
   END SUBROUTINE deallocate_tmc_atom_type

END MODULE tmc_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module provides variables for the TMC analysis tool
!> \par History
!>      02.2013 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_analysis_types
   USE cell_types,                      ONLY: cell_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE tmc_tree_types,                  ONLY: tree_type
   USE tmc_types,                       ONLY: tmc_atom_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_analysis_types'

   PUBLIC :: tmc_analysis_env, tmc_ana_list_type
   PUBLIC :: tmc_ana_env_create, tmc_ana_env_release
   PUBLIC :: tmc_ana_density_create
   PUBLIC :: pair_correl_type, tmc_ana_pair_correl_create, &
             search_pair_in_list, atom_pairs_type
   PUBLIC :: dipole_moment_type, tmc_ana_dipole_moment_create
   PUBLIC :: tmc_ana_dipole_analysis_create
   PUBLIC :: tmc_ana_displacement_create

   CHARACTER(LEN=default_path_length), PARAMETER, &
      PUBLIC :: tmc_ana_density_file_name = "tmc_ana_density.dat"
   CHARACTER(LEN=default_path_length), PARAMETER, &
      PUBLIC :: tmc_ana_pair_correl_file_name = "tmc_ana_g_r.dat"

   INTEGER, PARAMETER, PUBLIC                      :: ana_type_default = 0
   INTEGER, PARAMETER, PUBLIC                      :: ana_type_ice = 1
   INTEGER, PARAMETER, PUBLIC                      :: ana_type_sym_xyz = 2

   TYPE tmc_ana_list_type
      TYPE(tmc_analysis_env), POINTER               :: temp => NULL()
   END TYPE tmc_ana_list_type

   TYPE tmc_analysis_env
      INTEGER                                       :: io_unit
      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                       :: dirs
      CHARACTER(LEN=default_path_length)          :: out_file_prefix
      INTEGER                                       :: conf_offset
      TYPE(cell_type), POINTER                      :: cell
      TYPE(tmc_atom_type), DIMENSION(:), POINTER    :: atoms
      INTEGER                                       :: dim_per_elem = 3
      INTEGER                                       :: nr_dim = -1
      REAL(KIND=dp)                                 :: temperature
      TYPE(tree_type), POINTER                      :: last_elem
      INTEGER                                       :: from_elem, to_elem
      INTEGER                                       :: id_traj, id_cell, id_frc, id_dip, id_ener
      INTEGER                                       :: lc_traj, lc_cell, lc_frc, lc_dip, lc_ener
      CHARACTER(LEN=default_path_length)          :: costum_pos_file_name
      CHARACTER(LEN=default_path_length)          :: costum_dip_file_name
      CHARACTER(LEN=default_path_length)          :: costum_cell_file_name
      LOGICAL                                       :: restart, restarted
      LOGICAL                                       :: print_test_output

      TYPE(density_3d_type), POINTER                :: density_3d
      TYPE(pair_correl_type), POINTER               :: pair_correl
      TYPE(dipole_moment_type), POINTER             :: dip_mom
      TYPE(dipole_analysis_type), POINTER           :: dip_ana
      TYPE(displacement_type), POINTER              :: displace
   END TYPE tmc_analysis_env

   TYPE density_3d_type
      INTEGER                                       :: conf_counter
      INTEGER, DIMENSION(3)                         :: nr_bins
      REAL(KIND=dp)                                 :: sum_vol
      REAL(KIND=dp)                                 :: sum_vol2
      REAL(KIND=dp), DIMENSION(3)                   :: sum_box_length
      REAL(KIND=dp), DIMENSION(3)                   :: sum_box_length2
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER      :: sum_density, sum_dens2
      LOGICAL                                       :: print_dens
   END TYPE density_3d_type

   TYPE pair_correl_type
      INTEGER                                       :: conf_counter
      INTEGER                                       :: nr_bins
      REAL(KIND=dp)                                 :: step_lenght
      TYPE(atom_pairs_type), DIMENSION(:), POINTER  :: pairs
      REAL(KIND=dp), DIMENSION(:, :), POINTER        :: g_r
      REAL(KIND=dp)                                 :: sum_box_scale(3)
   END TYPE pair_correl_type

   TYPE atom_pairs_type
      CHARACTER(LEN=default_string_length)          :: f_n = ""
      CHARACTER(LEN=default_string_length)          :: s_n = ""
      INTEGER                                       :: pair_count
   END TYPE atom_pairs_type

   TYPE dipole_moment_type
      INTEGER                                       :: conf_counter
      TYPE(tmc_atom_type), DIMENSION(:), POINTER    :: charges_inp
      REAL(KIND=dp), DIMENSION(:), POINTER          :: charges
      REAL(KIND=dp), DIMENSION(:), POINTER          :: last_dip_cl
      LOGICAL                                       :: print_cl_dip = .TRUE.
   END TYPE dipole_moment_type

   TYPE dipole_analysis_type
      REAL(KIND=dp)                                 :: conf_counter
      INTEGER                                       :: ana_type
      LOGICAL                                       :: print_diel_const_traj = .TRUE.
      ! squared dipoles per volume
      REAL(KIND=dp)                                 :: mu2_pv_s
      ! dipole per square root ov volume per direction
      REAL(KIND=dp), DIMENSION(:), POINTER          :: mu_psv, mu_pv, mu2_pv
      ! dipole dipole correlation matrix (per volume)
      REAL(KIND=dp), DIMENSION(:, :), POINTER        :: mu2_pv_mat

   END TYPE dipole_analysis_type

   TYPE displacement_type
      INTEGER                                       :: conf_counter
      REAL(KIND=dp)                                 :: disp
      LOGICAL                                       :: print_disp
   END TYPE displacement_type

CONTAINS

! **************************************************************************************************
!> \brief creates a new structure environment for TMC analysis
!> \param tmc_ana structure with parameters for TMC analysis
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_env_create(tmc_ana)
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_env_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_analysis_types.F",144)

      ALLOCATE (tmc_ana)

      NULLIFY (tmc_ana%dirs, tmc_ana%cell, tmc_ana%atoms, tmc_ana%density_3d, &
               tmc_ana%pair_correl, tmc_ana%dip_mom, tmc_ana%last_elem, &
               tmc_ana%displace, tmc_ana%dip_ana)

      ! file IO units set
      tmc_ana%id_traj = -1
      tmc_ana%id_cell = -1
      tmc_ana%id_frc = -1
      tmc_ana%id_dip = -1
      tmc_ana%id_ener = -1
      ! line counters
      tmc_ana%lc_traj = 0
      tmc_ana%lc_cell = 0
      tmc_ana%lc_frc = 0
      tmc_ana%lc_dip = 0
      tmc_ana%lc_ener = 0
      ! file names
      tmc_ana%out_file_prefix = ""
      tmc_ana%costum_pos_file_name = ""
      tmc_ana%costum_dip_file_name = ""
      tmc_ana%costum_cell_file_name = ""
      tmc_ana%restart = .TRUE.
      tmc_ana%restarted = .FALSE.
      tmc_ana%print_test_output = .FALSE.

      ! configuration control
      tmc_ana%conf_offset = 0
      tmc_ana%from_elem = -1
      tmc_ana%to_elem = -1
   END SUBROUTINE tmc_ana_env_create

! **************************************************************************************************
!> \brief releases the structure environment for TMC analysis
!> \param tmc_ana structure with parameters for TMC analysis
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_env_release(tmc_ana)
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_env_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_analysis_types.F",190)

      IF (ASSOCIATED(tmc_ana%dirs)) &
         DEALLOCATE (tmc_ana%dirs)

      IF (ASSOCIATED(tmc_ana%density_3d)) &
         CALL tmc_ana_dens_release(tmc_ana%density_3d)
      IF (ASSOCIATED(tmc_ana%pair_correl)) &
         CALL tmc_ana_pair_correl_release(tmc_ana%pair_correl)

      IF (ASSOCIATED(tmc_ana%dip_mom)) &
         CALL tmc_ana_dipole_moment_release(tmc_ana%dip_mom)

      IF (ASSOCIATED(tmc_ana%dip_ana)) &
         CALL tmc_ana_dipole_analysis_release(tmc_ana%dip_ana)

      IF (ASSOCIATED(tmc_ana%displace)) &
         CALL tmc_ana_displacement_release(ana_disp=tmc_ana%displace)

      DEALLOCATE (tmc_ana)

   END SUBROUTINE tmc_ana_env_release

   !============================================================================
   ! density calculations
   !============================================================================

! **************************************************************************************************
!> \brief creates a new structure environment for TMC analysis
!> \param ana_dens structure with parameters for TMC density analysis
!> \param nr_bins ...
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_density_create(ana_dens, nr_bins)
      TYPE(density_3d_type), POINTER                     :: ana_dens
      INTEGER, DIMENSION(3)                              :: nr_bins

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_density_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ana_dens)))CALL cp__a("tmc/tmc_analysis_types.F",230)

      ALLOCATE (ana_dens)

      NULLIFY (ana_dens%sum_density, ana_dens%sum_dens2)

      ana_dens%nr_bins(:) = nr_bins(:)
      ana_dens%conf_counter = 0
      ana_dens%sum_vol = 0.0_dp
      ana_dens%sum_box_length(:) = 0.0_dp
      ana_dens%sum_vol2 = 0.0_dp
      ana_dens%sum_box_length2(:) = 0.0_dp

      ALLOCATE (ana_dens%sum_density(nr_bins(1), nr_bins(2), nr_bins(3)))
      ALLOCATE (ana_dens%sum_dens2(nr_bins(1), nr_bins(2), nr_bins(3)))
      ana_dens%sum_density(:, :, :) = 0.0_dp
      ana_dens%sum_dens2(:, :, :) = 0.0_dp
      ana_dens%print_dens = .TRUE.
   END SUBROUTINE tmc_ana_density_create

! **************************************************************************************************
!> \brief releases the structure environment for TMC analysis
!> \param ana_dens structure with parameters for TMC analysis
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_dens_release(ana_dens)
      TYPE(density_3d_type), POINTER                     :: ana_dens

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_dens_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ana_dens)))CALL cp__a("tmc/tmc_analysis_types.F",261)

      DEALLOCATE (ana_dens%sum_density)
      DEALLOCATE (ana_dens%sum_dens2)
      DEALLOCATE (ana_dens)
   END SUBROUTINE tmc_ana_dens_release

   !============================================================================
   ! radial distribution function
   !============================================================================

! **************************************************************************************************
!> \brief creates a new structure environment for TMC analysis
!> \param ana_pair_correl ...
!> \param nr_bins ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_pair_correl_create(ana_pair_correl, nr_bins)
      TYPE(pair_correl_type), POINTER                    :: ana_pair_correl
      INTEGER                                            :: nr_bins

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_pair_correl_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ana_pair_correl)))CALL cp__a("tmc/tmc_analysis_types.F",286)
      ALLOCATE (ana_pair_correl)

      NULLIFY (ana_pair_correl%g_r, ana_pair_correl%pairs)
      ana_pair_correl%conf_counter = 0
      ana_pair_correl%nr_bins = nr_bins
      ana_pair_correl%step_lenght = -1.0_dp
      ana_pair_correl%sum_box_scale = 0.0_dp
      ! still the initialization routine has to be called
   END SUBROUTINE tmc_ana_pair_correl_create

! **************************************************************************************************
!> \brief releases the structure environment for TMC analysis
!> \param ana_pair_correl ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_pair_correl_release(ana_pair_correl)
      TYPE(pair_correl_type), POINTER                    :: ana_pair_correl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_pair_correl_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ana_pair_correl)))CALL cp__a("tmc/tmc_analysis_types.F",309)

      DEALLOCATE (ana_pair_correl%g_r)
      DEALLOCATE (ana_pair_correl%pairs)
      DEALLOCATE (ana_pair_correl)
   END SUBROUTINE tmc_ana_pair_correl_release

! **************************************************************************************************
!> \brief search the pair of two atom types in list
!> \param pair_list ...
!> \param n1 atom names
!> \param n2 atom names
!> \param list_end ...
!> \retval ind ...
!> \author Mandes 02.2013
! **************************************************************************************************
   FUNCTION search_pair_in_list(pair_list, n1, n2, list_end) RESULT(ind)
      TYPE(atom_pairs_type), DIMENSION(:), POINTER       :: pair_list
      CHARACTER(LEN=default_string_length)               :: n1, n2
      INTEGER, OPTIONAL                                  :: list_end
      INTEGER                                            :: ind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_pair_in_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: last, list_nr

      IF(.NOT.(ASSOCIATED(pair_list)))CALL cp__a("tmc/tmc_analysis_types.F",336)
      IF (PRESENT(list_end)) THEN
         IF(.NOT.(list_end .LE. SIZE(pair_list)))CALL cp__a("tmc/tmc_analysis_types.F",338)
         last = list_end
      ELSE
         last = SIZE(pair_list)
      END IF

      ind = -1

      list_search: DO list_nr = 1, last
         IF ((pair_list(list_nr)%f_n .EQ. n1 .AND. &
              pair_list(list_nr)%s_n .EQ. n2) .OR. &
             (pair_list(list_nr)%f_n .EQ. n2 .AND. &
              pair_list(list_nr)%s_n .EQ. n1)) THEN
            ind = list_nr
            EXIT list_search
         END IF
      END DO list_search
   END FUNCTION search_pair_in_list

   !============================================================================
   ! classical cell dipole moment
   !============================================================================

! **************************************************************************************************
!> \brief creates a new structure environment for TMC analysis
!> \param ana_dip_mom ...
!> \param charge_atm ...
!> \param charge ...
!> \param dim_per_elem ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_dipole_moment_create(ana_dip_mom, charge_atm, charge, &
                                           dim_per_elem)
      TYPE(dipole_moment_type), POINTER                  :: ana_dip_mom
      CHARACTER(LEN=default_string_length), POINTER      :: charge_atm(:)
      REAL(KIND=dp), POINTER                             :: charge(:)
      INTEGER                                            :: dim_per_elem

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_dipole_moment_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(.NOT. ASSOCIATED(ana_dip_mom)))CALL cp__a("tmc/tmc_analysis_types.F",382)
      ALLOCATE (ana_dip_mom)

      NULLIFY (ana_dip_mom%charges_inp, ana_dip_mom%charges)

      ALLOCATE (ana_dip_mom%charges_inp(SIZE(charge)))
      DO i = 1, SIZE(charge)
         ana_dip_mom%charges_inp(i)%name = charge_atm(i)
         ana_dip_mom%charges_inp(i)%mass = charge(i)
      END DO

      ana_dip_mom%conf_counter = 0
      ALLOCATE (ana_dip_mom%last_dip_cl(dim_per_elem))
      ! still the initialization routine has to be called

   END SUBROUTINE tmc_ana_dipole_moment_create

! **************************************************************************************************
!> \brief releases the structure environment for TMC analysis
!> \param ana_dip_mom ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_dipole_moment_release(ana_dip_mom)
      TYPE(dipole_moment_type), POINTER                  :: ana_dip_mom

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_dipole_moment_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ana_dip_mom)))CALL cp__a("tmc/tmc_analysis_types.F",411)

      IF (ASSOCIATED(ana_dip_mom%charges_inp)) DEALLOCATE (ana_dip_mom%charges_inp)
      IF (ASSOCIATED(ana_dip_mom%charges)) DEALLOCATE (ana_dip_mom%charges)
      DEALLOCATE (ana_dip_mom%last_dip_cl)
      DEALLOCATE (ana_dip_mom)
   END SUBROUTINE tmc_ana_dipole_moment_release

! **************************************************************************************************
!> \brief creates a new structure environment for TMC analysis
!> \param ana_dip_ana ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_dipole_analysis_create(ana_dip_ana)
      TYPE(dipole_analysis_type), POINTER                :: ana_dip_ana

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_dipole_analysis_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ana_dip_ana)))CALL cp__a("tmc/tmc_analysis_types.F",431)
      ALLOCATE (ana_dip_ana)

      ana_dip_ana%conf_counter = 0
      ana_dip_ana%ana_type = -1
      ana_dip_ana%mu2_pv_s = 0.0_dp
      ALLOCATE (ana_dip_ana%mu_psv(3))
      ana_dip_ana%mu_psv(:) = 0.0_dp
      ALLOCATE (ana_dip_ana%mu_pv(3))
      ana_dip_ana%mu_pv(:) = 0.0_dp
      ALLOCATE (ana_dip_ana%mu2_pv(3))
      ana_dip_ana%mu2_pv(:) = 0.0_dp
      ALLOCATE (ana_dip_ana%mu2_pv_mat(3, 3))
      ana_dip_ana%mu2_pv_mat(:, :) = 0.0_dp
   END SUBROUTINE tmc_ana_dipole_analysis_create

! **************************************************************************************************
!> \brief releases the structure environment for TMC analysis
!> \param ana_dip_ana ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_dipole_analysis_release(ana_dip_ana)
      TYPE(dipole_analysis_type), POINTER                :: ana_dip_ana

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_dipole_analysis_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ana_dip_ana)))CALL cp__a("tmc/tmc_analysis_types.F",459)

      DEALLOCATE (ana_dip_ana%mu_psv)
      DEALLOCATE (ana_dip_ana%mu_pv)
      DEALLOCATE (ana_dip_ana%mu2_pv)
      DEALLOCATE (ana_dip_ana%mu2_pv_mat)

      DEALLOCATE (ana_dip_ana)
   END SUBROUTINE tmc_ana_dipole_analysis_release

   !============================================================================
   ! particle displacement in cell (from one configuration to the next)
   !============================================================================

! **************************************************************************************************
!> \brief creates a new structure environment for TMC analysis
!> \param ana_disp ...
!> \param dim_per_elem ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_displacement_create(ana_disp, dim_per_elem)
      TYPE(displacement_type), POINTER                   :: ana_disp
      INTEGER                                            :: dim_per_elem

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_displacement_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ana_disp)))CALL cp__a("tmc/tmc_analysis_types.F",487)
      IF(.NOT.(dim_per_elem .GT. 0))CALL cp__a("tmc/tmc_analysis_types.F",488)

      ALLOCATE (ana_disp)

      ana_disp%disp = 0.0_dp
      ana_disp%conf_counter = 0
      ana_disp%print_disp = .TRUE.
   END SUBROUTINE tmc_ana_displacement_create

! **************************************************************************************************
!> \brief releases a structure environment for TMC analysis
!> \param ana_disp ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_ana_displacement_release(ana_disp)
      TYPE(displacement_type), POINTER                   :: ana_disp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_ana_displacement_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ana_disp)))CALL cp__a("tmc/tmc_analysis_types.F",509)

      DEALLOCATE (ana_disp)
   END SUBROUTINE tmc_ana_displacement_release
END MODULE tmc_analysis_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief acceptance ratio handling of the different Monte Carlo Moves types
!>        For each move type and each temperature average acceptence is
!>        determined.
!>        For each move is a weight (mv_weight) defined, which defines the
!>        probability to perform the move.
!>        We distinguish between moves performed on the exact potential
!>        (move on the master, energy on the energy worker) and
!>        NMC moves, which are performed on the worker using the approximate
!>        potential. The energies are calculated as usual on the energy worker
!>        with the exact potential.
!>        The move probabilities to perform a NMC is stored in the NMC move.
!>        The probilities of the single move types (performed with the
!>        approximate potential) are only compared within the NMC move
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_move_handle
   USE cp_log_handling,                 ONLY: cp_to_string
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE physcon,                         ONLY: au2a => angstrom
   USE string_utilities,                ONLY: uppercase
   USE tmc_move_types,                  ONLY: &
        move_types_create, move_types_release, mv_type_MD, mv_type_NMC_moves, mv_type_atom_swap, &
        mv_type_atom_trans, mv_type_gausian_adapt, mv_type_mol_rot, mv_type_mol_trans, &
        mv_type_proton_reorder, mv_type_swap_conf, mv_type_volume_move, tmc_move_type
   USE tmc_stati,                       ONLY: task_type_MC,&
                                              task_type_gaussian_adaptation,&
                                              task_type_ideal_gas
   USE tmc_tree_types,                  ONLY: global_tree_type,&
                                              status_accepted_result,&
                                              status_rejected_result,&
                                              tree_type
   USE tmc_types,                       ONLY: tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_move_handle'

   PUBLIC :: finalize_mv_types, print_move_types, read_init_move_types
   PUBLIC :: check_moves
   PUBLIC :: select_random_move_type
   PUBLIC :: prob_update, add_mv_prob
   PUBLIC :: clear_move_probs

CONTAINS

! **************************************************************************************************
!> \brief initialization of the different moves, with sizes and probabilities
!> \param tmc_params ...
!> \param tmc_section ...
!> \author Mandes 10.2013
! **************************************************************************************************
   SUBROUTINE read_init_move_types(tmc_params, tmc_section)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(section_vals_type), POINTER                   :: tmc_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_init_move_types', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: inp_kind_name
      INTEGER                                            :: i, i_rep, i_tmp, ind, n_items, &
                                                            n_NMC_items, n_rep_val, nmc_steps
      LOGICAL                                            :: explicit, flag
      REAL(KIND=dp)                                      :: delta_x, init_acc_prob, mv_prob, &
                                                            mv_prob_sum, nmc_init_acc_prob, &
                                                            nmc_prob, nmc_prob_sum, prob_ex
      TYPE(section_vals_type), POINTER                   :: move_type_section, nmc_section
      TYPE(tmc_move_type), POINTER                       :: move_types

      NULLIFY (move_types, move_type_section, nmc_section)

      n_items = 0
      n_NMC_items = 0
      delta_x = 0.0_dp
      nmc_prob = 0.0_dp
      mv_prob = 0.0_dp
      nmc_prob = 0.0_dp
      mv_prob_sum = 0.0_dp
      nmc_prob_sum = 0.0_dp
      prob_ex = 0.0_dp
      init_acc_prob = 0.0_dp

      ! the move types on exact potential
      move_type_section => section_vals_get_subs_vals(tmc_section, "MOVE_TYPE")
      CALL section_vals_get(move_type_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_get(move_type_section, n_repetition=n_items)
         mv_prob_sum = 0.0_dp
         DO i_rep = 1, n_items
            CALL section_vals_val_get(move_type_section, "PROB", i_rep_section=i_rep, &
                                      r_val=mv_prob)
            mv_prob_sum = mv_prob_sum+mv_prob
         END DO
      END IF

      ! get the NMC prameters
      nmc_section => section_vals_get_subs_vals(tmc_section, "NMC_MOVES")
      CALL section_vals_get(nmc_section, explicit=explicit)
      IF (explicit) THEN
         ! check the approx potential file, already read
         IF (tmc_params%NMC_inp_file .EQ. "") &
            CALL cp__b("tmc/tmc_move_handle.F",119,"Please specify a valid approximate potential.")

         CALL section_vals_val_get(nmc_section, "NR_NMC_STEPS", &
                                   i_val=nmc_steps)
         IF (nmc_steps .LE. 0) &
            CALL cp__b("tmc/tmc_move_handle.F",124,"Please specify a valid amount of NMC steps (NR_NMC_STEPS {INTEGER}).")

         CALL section_vals_val_get(nmc_section, "PROB", r_val=nmc_prob)

         CALL section_vals_val_get(move_type_section, "INIT_ACC_PROB", &
                                   r_val=nmc_init_acc_prob)
         IF (nmc_init_acc_prob .LE. 0.0_dp) &
            CALL cp_abort(cp__l("tmc/tmc_move_handle.F",131), &
                          "Please select a valid initial acceptance probability (>0.0) "// &
                          "for INIT_ACC_PROB")

         move_type_section => section_vals_get_subs_vals(nmc_section, "MOVE_TYPE")
         CALL section_vals_get(move_type_section, n_repetition=n_NMC_items)

         ! get the NMC move probability sum
         nmc_prob_sum = 0.0_dp
         DO i_rep = 1, n_NMC_items
            CALL section_vals_val_get(move_type_section, "PROB", i_rep_section=i_rep, &
                                      r_val=mv_prob)
            nmc_prob_sum = nmc_prob_sum+mv_prob
         END DO
      END IF

      ! get the total weight/amount of move probabilities
      mv_prob_sum = mv_prob_sum+nmc_prob

      IF (n_items+n_NMC_items .GT. 0) THEN
         ! initilaize the move array with related sizes, probs, etc.
         CALL move_types_create(tmc_params%move_types, tmc_params%nr_temp)

         IF (mv_prob_sum .LE. 0.0) &
            CALL cp_abort(cp__l("tmc/tmc_move_handle.F",155), &
                          "The probabilities to perform the moves are "// &
                          "in total less equal 0")

         ! get the sizes, probs, etc. for each move type and convert units
         DO i_tmp = 1, n_items+n_NMC_items
            ! select the corect section
            IF (i_tmp .GT. n_items) THEN
               i_rep = i_tmp-n_items
               IF (i_rep .EQ. 1) THEN
                  ! set the NMC stuff (approx potential)
                  tmc_params%move_types%mv_weight(mv_type_NMC_moves) = &
                     nmc_prob/REAL(mv_prob_sum, KIND=dp)
                  tmc_params%move_types%mv_size(mv_type_NMC_moves, :) = nmc_steps
                  tmc_params%move_types%acc_prob(mv_type_NMC_moves, :) = nmc_init_acc_prob

                  move_type_section => section_vals_get_subs_vals(tmc_section, "NMC_MOVES%MOVE_TYPE")
                  mv_prob_sum = nmc_prob_sum
                  ! allocate the NMC move types
                  CALL move_types_create(tmc_params%nmc_move_types, tmc_params%nr_temp)
                  move_types => tmc_params%nmc_move_types
               END IF
            ELSE
               ! the moves on exact potential
               move_type_section => section_vals_get_subs_vals(tmc_section, "MOVE_TYPE")
               i_rep = i_tmp
               move_types => tmc_params%move_types
            END IF

            CALL section_vals_val_get(move_type_section, "_SECTION_PARAMETERS_", &
                                      c_val=inp_kind_name, i_rep_section=i_rep)
            CALL uppercase(inp_kind_name)
            CALL section_vals_val_get(move_type_section, "SIZE", i_rep_section=i_rep, &
                                      r_val=delta_x)
            ! move sizes are checked afterwards, because not all moves require a valid move size
            CALL section_vals_val_get(move_type_section, "PROB", i_rep_section=i_rep, &
                                      r_val=mv_prob)
            IF (mv_prob .LT. 0.0_dp) &
               CALL cp_abort(cp__l("tmc/tmc_move_handle.F",193), &
                             "Please select a valid move probability (>0.0) "// &
                             "for the move type "//inp_kind_name)
            CALL section_vals_val_get(move_type_section, "INIT_ACC_PROB", i_rep_section=i_rep, &
                                      r_val=init_acc_prob)
            IF (init_acc_prob .LT. 0.0_dp) &
               CALL cp_abort(cp__l("tmc/tmc_move_handle.F",199), &
                             "Please select a valid initial acceptance probability (>0.0) "// &
                             "for the move type "//inp_kind_name)
            ! set the related index and perform unit conversion of move sizes
            SELECT CASE (inp_kind_name)
               ! atom / molecule translation
            CASE ("ATOM_TRANS", "MOL_TRANS")
               SELECT CASE (inp_kind_name)
               CASE ("ATOM_TRANS")
                  ind = mv_type_atom_trans
               CASE ("MOL_TRANS")
                  ind = mv_type_mol_trans
               CASE DEFAULT
                  CALL cp__b("tmc/tmc_move_handle.F",212,"move type is not defined in the translation types")
               END SELECT
               ! convert units
               SELECT CASE (tmc_params%task_type)
               CASE (task_type_MC, task_type_ideal_gas)
                  delta_x = delta_x/au2a
               CASE (task_type_gaussian_adaptation)
                  !nothing to do (no unit conversion)
               CASE DEFAULT
                  CALL cp__b("tmc/tmc_move_handle.F",221,"move type atom / mol trans is not defined for this TMC run type")
               END SELECT
               ! molecule rotation
            CASE ("MOL_ROT")
               ind = mv_type_mol_rot
               ! convert units
               SELECT CASE (tmc_params%task_type)
               CASE (task_type_MC, task_type_ideal_gas)
                  delta_x = delta_x*PI/180.0_dp
               CASE DEFAULT
                  CALL cp__b("tmc/tmc_move_handle.F",231,"move type MOL_ROT is not defined for this TMC run type")
               END SELECT
               ! proton reordering
            CASE ("PROT_REORDER")
               ind = mv_type_proton_reorder
               ! the move size is not necessary
               delta_x = 0.0_dp
               ! Hybrid MC (MD)
            CASE ("HYBRID_MC")
               ind = mv_type_MD
               delta_x = delta_x*Pi/180.0_dp !input in degree, calculating in rad
               tmc_params%print_forces = .TRUE.
               ! parallel tempering swap move
            CASE ("PT_SWAP")
               ind = mv_type_swap_conf
               ! the move size is not necessary
               delta_x = 0.0_dp
               IF (tmc_params%nr_temp .LE. 1) THEN
                  ! no configurational swapping if only one temperature
                  mv_prob = 0.0_dp
                  CALL cp_warn(cp__l("tmc/tmc_move_handle.F",251), &
                               "Configurational swap disabled, because "// &
                               "Parallel Tempering requires more than one temperature.")
               END IF
               ! volume moves
            CASE ("VOL_MOVE")
               ind = mv_type_volume_move
               ! check the selected pressure
               IF (tmc_params%pressure .GE. 0.0_dp) THEN
                  delta_x = delta_x/au2a
                  tmc_params%print_cell = .TRUE. ! print the cell sizes by default
               ELSE
                  CALL cp_warn(cp__l("tmc/tmc_move_handle.F",263), &
                               "no valid pressure defined, but volume move defined. "// &
                               "Consequently, the volume move is disabled.")
                  mv_prob = 0.0_dp
               END IF
               ! parallel tempering swap move
            CASE ("ATOM_SWAP")
               ind = mv_type_atom_swap
               ! the move size is not necessary
               delta_x = 0.0_dp
               ! select the types of atoms swapped
               CALL section_vals_val_get(move_type_section, "ATOMS", i_rep_section=i_rep, &
                                         n_rep_val=n_rep_val)
               IF (n_rep_val .GT. 0) THEN
                  ALLOCATE (move_types%atom_lists(n_rep_val))
                  DO i = 1, n_rep_val
                     CALL section_vals_val_get(move_type_section, "ATOMS", &
                                               i_rep_section=i_rep, i_rep_val=i, &
                                               c_vals=move_types%atom_lists(i)%atoms)
                     IF (SIZE(move_types%atom_lists(i)%atoms) .LE. 1) &
                        CALL cp__b("tmc/tmc_move_handle.F",283,"ATOM_SWAP requires minimum two atom kinds selected. ")
                  END DO
               END IF
               ! gaussian adaptation
            CASE ("GAUSS_ADAPT")
               ind = mv_type_gausian_adapt
               init_acc_prob = 0.5_dp
            CASE DEFAULT
               CALL cp__b("tmc/tmc_move_handle.F",291,"A unknown move type is selected: "//inp_kind_name)
            END SELECT
            ! check for valid move sizes
            IF (delta_x .LT. 0.0_dp) &
               CALL cp_abort(cp__l("tmc/tmc_move_handle.F",295), &
                             "Please select a valid move size (>0.0) "// &
                             "for the move type "//inp_kind_name)
            ! check if not already set
            IF (move_types%mv_weight(ind) .GT. 0.0) THEN
               CALL cp__b("tmc/tmc_move_handle.F",300,"TMC: Each move type can be set only once. ")
            END IF

            ! set the move size
            move_types%mv_size(ind, :) = delta_x
            ! set the probability to perform move
            move_types%mv_weight(ind) = mv_prob/mv_prob_sum
            ! set the initial acceptance probability
            move_types%acc_prob(ind, :) = init_acc_prob
         END DO
      ELSE
         CALL cp__b("tmc/tmc_move_handle.F",311,"No move type selected, please select at least one.")
      END IF
      mv_prob_sum = SUM(tmc_params%move_types%mv_weight(:))
      flag = .TRUE.
      IF(.NOT.(ABS(mv_prob_sum-1.0_dp) .LT. 0.01_dp))CALL cp__a("tmc/tmc_move_handle.F",315)
      IF (ASSOCIATED(tmc_params%nmc_move_types)) THEN
         mv_prob_sum = SUM(tmc_params%nmc_move_types%mv_weight(:))
         IF(.NOT.(ABS(mv_prob_sum-1.0_dp) < 10*EPSILON(1.0_dp)))CALL cp__a("tmc/tmc_move_handle.F",318)
      END IF
   END SUBROUTINE read_init_move_types

! **************************************************************************************************
!> \brief checks if the moves are possible
!> \param tmc_params ...
!> \param move_types ...
!> \param mol_array ...
!> \author Mandes 10.2013
! **************************************************************************************************
   SUBROUTINE check_moves(tmc_params, move_types, mol_array)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(tmc_move_type), POINTER                       :: move_types
      INTEGER, DIMENSION(:), POINTER                     :: mol_array

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_moves', routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_j, list_i, ref_k
      LOGICAL                                            :: found

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_move_handle.F",339)
      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_move_handle.F",340)

      ! molecule moves need molecule info
      IF (move_types%mv_weight(mv_type_mol_trans) .GT. 0.0_dp .OR. &
          move_types%mv_weight(mv_type_mol_rot) .GT. 0.0_dp) THEN
         ! if there is no molecule information available,
         !   molecules moves can not be performed
         IF (mol_array(SIZE(mol_array)) .EQ. SIZE(mol_array)) &
            CALL cp_abort(cp__l("tmc/tmc_move_handle.F",348), &
                          "molecule move: there is no molecule "// &
                          "information available. Please specify molecules when "// &
                          "using molecule moves.")
      END IF

      ! for the atom swap move
      IF (move_types%mv_weight(mv_type_atom_swap) .GT. 0.0_dp) THEN
         ! check if the selected atom swaps are possible
         IF (ASSOCIATED(move_types%atom_lists)) THEN
            DO list_i = 1, SIZE(move_types%atom_lists(:))
               DO atom_j = 1, SIZE(move_types%atom_lists(list_i)%atoms(:))
                  ! check if atoms exists
                  found = .FALSE.
                  ref_loop: DO ref_k = 1, SIZE(tmc_params%atoms(:))
                     IF (move_types%atom_lists(list_i)%atoms(atom_j) .EQ. &
                         tmc_params%atoms(ref_k)%name) THEN
                        found = .TRUE.
                        EXIT ref_loop
                     END IF
                  END DO ref_loop
                  IF (.NOT. found) &
                     CALL cp_abort(cp__l("tmc/tmc_move_handle.F",370), &
                                   "ATOM_SWAP: The selected atom type ("// &
                                   TRIM(move_types%atom_lists(list_i)%atoms(atom_j))// &
                                   ") is not contained in the system. ")
                  ! check if not be swapped with the same atom type
                  IF (ANY(move_types%atom_lists(list_i)%atoms(atom_j) .EQ. &
                          move_types%atom_lists(list_i)%atoms(atom_j+1:))) THEN
                     CALL cp_abort(cp__l("tmc/tmc_move_handle.F",377), &
                                   "ATOM_SWAP can not swap two atoms of same kind ("// &
                                   TRIM(move_types%atom_lists(list_i)%atoms(atom_j))// &
                                   ")")
                  END IF
               END DO
            END DO
         ELSE
            ! check if there exisit different atoms
            found = .FALSE.
            IF (SIZE(tmc_params%atoms(:)) .GT. 1) THEN
               ref_lop: DO ref_k = 2, SIZE(tmc_params%atoms(:))
                  IF (tmc_params%atoms(1)%name .NE. tmc_params%atoms(ref_k)%name) THEN
                     found = .TRUE.
                     EXIT ref_lop
                  END IF
               END DO ref_lop
            END IF
            IF (.NOT. found) &
               CALL cp_abort(cp__l("tmc/tmc_move_handle.F",396), &
                             "The system contains only a single atom type,"// &
                             " atom_swap is not possible.")
         END IF
      END IF
   END SUBROUTINE check_moves

! **************************************************************************************************
!> \brief deallocating the module variables
!> \param tmc_params ...
!> \author Mandes 11.2012
!> \note deallocating the module variables
! **************************************************************************************************
   SUBROUTINE finalize_mv_types(tmc_params)
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'finalize_mv_types', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_move_handle.F",415)
      CALL move_types_release(tmc_params%move_types)
      IF (ASSOCIATED(tmc_params%nmc_move_types)) &
         CALL move_types_release(tmc_params%nmc_move_types)
   END SUBROUTINE finalize_mv_types

! **************************************************************************************************
!> \brief routine pronts out the probabilities and sized for each type and
!>        temperature the output is divided into two parts the init,
!>        which is printed out at the beginning of the programm and
!>        .NOT.init which are the probabilites and counter printed out every
!>        print cycle
!> \param init ...
!> \param file_io ...
!> \param tmc_params ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE print_move_types(init, file_io, tmc_params)
      LOGICAL                                            :: init
      INTEGER                                            :: file_io
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_move_types', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=10)                                  :: c_t
      CHARACTER(LEN=50)                                  :: FMT_c, FMT_i, FMT_r
      CHARACTER(LEN=500)                                 :: c_a, c_b, c_c, c_d, c_e, c_tit, c_tmp
      INTEGER                                            :: column_size, move, nr_nmc_moves, temper, &
                                                            typ
      LOGICAL                                            :: subbox_out, type_title
      TYPE(tmc_move_type), POINTER                       :: move_types

      NULLIFY (move_types)

      c_a = ""; c_b = ""; c_c = ""
      c_d = ""; c_e = ""; c_tit = ""
      column_size = 10
      subbox_out = .FALSE.
      type_title = .FALSE.
      IF(.NOT.(file_io .GT. 0))CALL cp__a("tmc/tmc_move_handle.F",455)
      IF(.NOT.(ASSOCIATED(tmc_params%move_types)))CALL cp__a("tmc/tmc_move_handle.F",456)

      FLUSH(file_io)

      IF (.NOT. init .AND. &
          tmc_params%move_types%mv_weight(mv_type_NMC_moves) .GT. 0 .AND. &
          ANY(tmc_params%sub_box_size .GT. 0.0_dp)) subbox_out = .TRUE.

      ! set the format for each typ to add one column
      WRITE (FMT_c, '("(A,1X,A", I0, ")")') column_size
      WRITE (FMT_i, '("(A,1X,I", I0, ")")') column_size
      WRITE (FMT_r, '("(A,1X,F", I0, ".3)")') column_size
      !IF(init) &
      type_title = .TRUE.

      nr_nmc_moves = 0
      IF (ASSOCIATED(tmc_params%nmc_move_types)) THEN
         nr_nmc_moves = SIZE(tmc_params%nmc_move_types%mv_weight(1:))
      END IF

      temp_loop: DO temper = 1, tmc_params%nr_temp
         c_tit = ""; c_a = ""; c_b = ""; c_c = ""
         IF (init .AND. temper .GT. 1) EXIT temp_loop
         WRITE (c_t, "(F10.2)") tmc_params%Temp(temper)
         typ_loop: DO move = 0, SIZE(tmc_params%move_types%mv_weight)+nr_nmc_moves
            ! the NMC moves
            IF (move .LE. SIZE(tmc_params%move_types%mv_weight)) THEN
               typ = move
               move_types => tmc_params%move_types
            ELSE
               typ = move-SIZE(tmc_params%move_types%mv_weight)
               move_types => tmc_params%nmc_move_types
            END IF
            ! total average
            IF (typ .EQ. 0) THEN
               ! line start
               IF (type_title) WRITE (c_tit, TRIM(FMT_c)) " type  temperature  |"
               IF (init) WRITE (c_b, TRIM(FMT_c)) "   I       I        |"
               IF (init) WRITE (c_c, TRIM(FMT_c)) "   V       V        |"
               IF (.NOT. init) WRITE (c_a, TRIM(FMT_c)) "probs  T="//c_t//" |"
               IF (.NOT. init) WRITE (c_b, TRIM(FMT_c)) "counts T="//c_t//" |"
               IF (.NOT. init) WRITE (c_c, TRIM(FMT_c)) "nr_acc T="//c_t//" |"
               IF (subbox_out) THEN
                  WRITE (c_d, TRIM(FMT_c)) "sb_acc T="//c_t//" |"
                  WRITE (c_e, TRIM(FMT_c)) "sb_cou T="//c_t//" |"
               END IF
               ! overall column
               IF (type_title) THEN
                  c_tmp = TRIM(c_tit)
                  WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), " trajec"
               END IF
               IF (init) THEN
                  c_tmp = TRIM(c_b)
                  WRITE (c_b, TRIM(FMT_c)) TRIM(c_tmp), "  weight->"
               END IF
               IF (init) THEN
                  c_tmp = TRIM(c_c)
                  WRITE (c_c, TRIM(FMT_c)) TRIM(c_tmp), "  size  ->"
               END IF
               IF (.NOT. init) THEN
                  c_tmp = TRIM(c_a)
                  WRITE (c_a, TRIM(FMT_r)) TRIM(c_tmp), &
                     move_types%acc_prob(typ, temper)
               END IF
               IF (.NOT. init) THEN
                  c_tmp = TRIM(c_b)
                  WRITE (c_b, TRIM(FMT_i)) TRIM(c_tmp), &
                     move_types%mv_count(typ, temper)
               END IF
               IF (.NOT. init) THEN
                  c_tmp = TRIM(c_c)
                  WRITE (c_c, TRIM(FMT_i)) TRIM(c_tmp), &
                     move_types%acc_count(typ, temper)
               END IF
               IF (subbox_out) THEN
                  c_tmp = TRIM(c_d)
                  WRITE (c_d, TRIM(FMT_c)) TRIM(c_tmp), "."
                  c_tmp = TRIM(c_e)
                  WRITE (c_e, TRIM(FMT_c)) TRIM(c_tmp), "."
               END IF
            ELSE
               ! certain move types
               IF (move_types%mv_weight(typ) .GT. 0.0_dp) THEN
                  ! INIT: the weights in the initialisation output
                  IF (init) THEN
                     c_tmp = TRIM(c_b)
                     WRITE (c_b, TRIM(FMT_r)) TRIM(c_tmp), move_types%mv_weight(typ)
                  END IF
                  ! acc probabilities
                  IF (typ .EQ. mv_type_swap_conf .AND. &
                      temper .EQ. tmc_params%nr_temp) THEN
                     IF (.NOT. init) THEN
                        c_tmp = TRIM(c_a)
                        WRITE (c_a, TRIM(FMT_c)) TRIM(c_tmp), "---"
                     END IF
                  ELSE
                     IF (.NOT. init) THEN
                        c_tmp = TRIM(c_a)
                        WRITE (c_a, TRIM(FMT_r)) TRIM(c_tmp), move_types%acc_prob(typ, temper)
                     END IF
                  END IF
                  IF (.NOT. init) THEN
                     c_tmp = TRIM(c_b)
                     WRITE (c_b, TRIM(FMT_i)) TRIM(c_tmp), move_types%mv_count(typ, temper)
                  END IF
                  IF (.NOT. init) THEN
                     c_tmp = TRIM(c_c)
                     WRITE (c_c, TRIM(FMT_i)) TRIM(c_tmp), move_types%acc_count(typ, temper)
                  END IF
                  ! sub box
                  IF (subbox_out) THEN
                     IF (move .GT. SIZE(tmc_params%move_types%mv_weight)) THEN
                        c_tmp = TRIM(c_d)
                        WRITE (c_d, TRIM(FMT_r)) TRIM(c_tmp), &
                           move_types%subbox_acc_count(typ, temper)/ &
                           REAL(MAX(1, move_types%subbox_count(typ, temper)), KIND=dp)
                        c_tmp = TRIM(c_e)
                        WRITE (c_e, TRIM(FMT_i)) TRIM(c_tmp), &
                           move_types%subbox_count(typ, temper)
                     ELSE
                        c_tmp = TRIM(c_d)
                        WRITE (c_d, TRIM(FMT_c)) TRIM(c_tmp), "-"
                        c_tmp = TRIM(c_e)
                        WRITE (c_e, TRIM(FMT_c)) TRIM(c_tmp), "-"
                     END IF
                  END IF

                  SELECT CASE (typ)
                  CASE (mv_type_atom_trans)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "atom trans."
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_r)) TRIM(c_tmp), &
                           move_types%mv_size(typ, temper)*au2a
                     END IF
                  CASE (mv_type_mol_trans)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "mol trans"
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_r)) TRIM(c_tmp), &
                           move_types%mv_size(typ, temper)*au2a
                     END IF
                  CASE (mv_type_mol_rot)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "mol rot"
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_r)) TRIM(c_tmp), &
                           move_types%mv_size(typ, temper)/(PI/180.0_dp)
                     END IF
                  CASE (mv_type_MD)
                     CALL cp__w("tmc/tmc_move_handle.F",615,"md_time_step and nr md_steps not implemented...")
!                   IF(type_title) WRITE(c_tit,TRIM(FMT_c)) TRIM(c_tit), "HybridMC"
!                   IF(init) WRITE(c_c,TRIM(FMT_c)) TRIM(c_c), "s.above"
!                   IF(init) THEN
!                      WRITE(file_io,*)"   move type: molecular dynamics with file ",NMC_inp_file
!                      WRITE(file_io,*)"                                 with time step [fs] ",md_time_step*au2fs
!                      WRITE(file_io,*)"                                 with number of steps ",md_steps
!                      WRITE(file_io,*)"                                 with velocity changes consists of old vel and ",&
!                         sin(move_types%mv_size(typ,1))*100.0_dp,"% random Gaussian with variance to temperature,"
!                   END IF
                  CASE (mv_type_proton_reorder)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "H-Reorder"
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_c)) TRIM(c_tmp), "XXX"
                     END IF
                  CASE (mv_type_swap_conf)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "PT(swap)"
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_c)) TRIM(c_tmp), "XXX" !move_types%mv_size(mv_type_swap_conf,1)
                     END IF
                  CASE (mv_type_NMC_moves)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "NMC:"
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_i)) TRIM(c_tmp), &
                           INT(move_types%mv_size(typ, temper))
                     END IF
                  CASE (mv_type_volume_move)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "volume"
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_r)) TRIM(c_tmp), &
                           move_types%mv_size(typ, temper)*au2a
                     END IF
                  CASE (mv_type_atom_swap)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "atom swap"
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_c)) TRIM(c_tmp), "XXX"
                     END IF
                  CASE (mv_type_gausian_adapt)
                     IF (type_title) THEN
                        c_tmp = TRIM(c_tit)
                        WRITE (c_tit, TRIM(FMT_c)) TRIM(c_tmp), "gauss adap"
                     END IF
                     IF (init) THEN
                        c_tmp = TRIM(c_c)
                        WRITE (c_c, TRIM(FMT_r)) TRIM(c_tmp), &
                           move_types%mv_size(typ, temper)
                     END IF
                  CASE DEFAULT
                     CALL cp_warn(cp__l("tmc/tmc_move_handle.F",683), &
                                  "unknown move type "//cp_to_string(typ)//" with weight"// &
                                  cp_to_string(move_types%mv_weight(typ)))
                  END SELECT
               END IF
            END IF
         END DO typ_loop
         IF (init) WRITE (UNIT=file_io, FMT="(/,T2,A)") REPEAT("-", 79)
         IF (type_title .AND. temper .LE. 1) WRITE (file_io, *) TRIM(c_tit)
         IF (.NOT. init) WRITE (file_io, *) TRIM(c_a)
         WRITE (file_io, *) TRIM(c_b)
         WRITE (file_io, *) TRIM(c_c)
         IF (subbox_out) WRITE (file_io, *) TRIM(c_d)
         IF (subbox_out) WRITE (file_io, *) TRIM(c_e)
         IF (init) WRITE (UNIT=file_io, FMT="(/,T2,A)") REPEAT("-", 79)
      END DO temp_loop
   END SUBROUTINE print_move_types

! **************************************************************************************************
!> \brief adaptation of acceptance probability of every kind of change/move
!>        and the overall acc prob,
!>        using the acceptance and rejectance information
!> \param move_types structure for storing sizes and probabilities of moves
!> \param pt_el global tree element
!> \param elem sub tree element
!> \param acc input if the element is accepted
!> \param subbox logical if move was with respect to the sub box
!> \param prob_opt if the average probability should be adapted
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE prob_update(move_types, pt_el, elem, acc, subbox, prob_opt)
      TYPE(tmc_move_type), POINTER                       :: move_types
      TYPE(global_tree_type), OPTIONAL, POINTER          :: pt_el
      TYPE(tree_type), OPTIONAL, POINTER                 :: elem
      LOGICAL, INTENT(IN), OPTIONAL                      :: acc, subbox
      LOGICAL, INTENT(IN)                                :: prob_opt

      CHARACTER(LEN=*), PARAMETER :: routineN = 'prob_update', routineP = moduleN//':'//routineN

      INTEGER                                            :: change_res, change_sb_type, change_type, &
                                                            conf_moved, handle, mv_type

      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_move_handle.F",725)
      IF(.NOT.(.NOT. (PRESENT(pt_el) .AND. PRESENT(subbox))))CALL cp__a("tmc/tmc_move_handle.F",726)

      ! start the timing
      CALL timeset(routineN, handle)

      mv_type = -1
      conf_moved = -1

      change_type = 0
      change_res = 0
      change_sb_type = 0
      ! updating probability of the trajectory
      IF (PRESENT(pt_el)) THEN
         IF(.NOT.(ASSOCIATED(pt_el)))CALL cp__a("tmc/tmc_move_handle.F",739)
         conf_moved = pt_el%mv_conf
         SELECT CASE (pt_el%stat)
         CASE (status_accepted_result)
            change_res = 1
            !-- swaped move is not noted in subtree elements
            IF (pt_el%swaped) THEN
               mv_type = mv_type_swap_conf
               change_type = 1
            END IF
         CASE (status_rejected_result)
            change_res = -1
            !-- swaped move is not noted in subtree elements
            IF (pt_el%swaped) THEN
               mv_type = mv_type_swap_conf
               change_type = -1
            END IF
         CASE DEFAULT
            CALL cp_abort(cp__l("tmc/tmc_move_handle.F",757), &
                          "global elem"//cp_to_string(pt_el%nr)// &
                          "has unknown status"//cp_to_string(pt_el%stat))
         END SELECT
      END IF

      IF (PRESENT(elem)) THEN
         IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_move_handle.F",764)
         !conf_moved = elem%sub_tree_nr
         conf_moved = elem%temp_created
         mv_type = elem%move_type
         ! for NMC prob update the acceptance is needed
         IF(.NOT.(PRESENT(acc)))CALL cp__a("tmc/tmc_move_handle.F",769)
         IF (PRESENT(subbox)) THEN
            ! only update subbox acceptance
            IF (acc) &
               move_types%subbox_acc_count(mv_type, conf_moved) = move_types%subbox_acc_count(mv_type, conf_moved)+1
            move_types%subbox_count(mv_type, conf_moved) = move_types%subbox_count(mv_type, conf_moved)+1
            ! No more to do
            change_type = 0
            change_res = 0
            conf_moved = 0
            ! RETURN
         ELSE
            ! update move type acceptance
            IF (acc) THEN
               change_type = 1
            ELSE
               change_type = -1
            END IF
         END IF
      END IF

      !-- INcrease or DEcrease accaptance rate
      ! MOVE types
      IF (change_type .GT. 0) THEN
         move_types%acc_count(mv_type, conf_moved) = move_types%acc_count(mv_type, conf_moved)+1
      END IF

      ! RESULTs
      IF (change_res .GT. 0) THEN
         move_types%acc_count(0, conf_moved) = move_types%acc_count(0, conf_moved)+1
      END IF

      IF (conf_moved .GT. 0) move_types%mv_count(0, conf_moved) = move_types%mv_count(0, conf_moved)+ABS(change_res)
      IF (mv_type .GE. 0 .AND. conf_moved .GT. 0) &
         move_types%mv_count(mv_type, conf_moved) = move_types%mv_count(mv_type, conf_moved)+ABS(change_type)

      IF (prob_opt) THEN
         WHERE (move_types%mv_count .GT. 0) &
            move_types%acc_prob(:, :) = move_types%acc_count(:, :)/REAL(move_types%mv_count(:, :), KIND=dp)
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE prob_update

! **************************************************************************************************
!> \brief add the actual moves to the average probabilities
!> \param move_types structure with move counters and probabilities
!> \param prob_opt ...
!> \param mv_counter move counter for actual performed moves of certain types
!> \param acc_counter counters of acceptance for these moves
!> \param subbox_counter same for sub box moves
!> \param subbox_acc_counter same for sub box moves
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE add_mv_prob(move_types, prob_opt, mv_counter, acc_counter, &
                          subbox_counter, subbox_acc_counter)
      TYPE(tmc_move_type), POINTER                       :: move_types
      LOGICAL                                            :: prob_opt
      INTEGER, DIMENSION(:, :), OPTIONAL                 :: mv_counter, acc_counter, subbox_counter, &
                                                            subbox_acc_counter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'add_mv_prob', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_move_handle.F",832)
      IF(.NOT.(PRESENT(mv_counter) .OR. PRESENT(subbox_counter)))CALL cp__a("tmc/tmc_move_handle.F",833)

      IF (PRESENT(mv_counter)) THEN
         IF(.NOT.(PRESENT(acc_counter)))CALL cp__a("tmc/tmc_move_handle.F",836)
         move_types%mv_count(:, :) = move_types%mv_count(:, :)+mv_counter(:, :)
         move_types%acc_count(:, :) = move_types%acc_count(:, :)+acc_counter(:, :)
         IF (prob_opt) THEN
            WHERE (move_types%mv_count .GT. 0) &
               move_types%acc_prob(:, :) = move_types%acc_count(:, :)/REAL(move_types%mv_count(:, :), KIND=dp)
         END IF
      END IF

      IF (PRESENT(subbox_counter)) THEN
         IF(.NOT.(PRESENT(subbox_acc_counter)))CALL cp__a("tmc/tmc_move_handle.F",846)
         move_types%subbox_count(:, :) = move_types%subbox_count(:, :)+subbox_counter(:, :)
         move_types%subbox_acc_count(:, :) = move_types%subbox_acc_count(:, :)+subbox_acc_counter(:, :)
      END IF
   END SUBROUTINE add_mv_prob

! **************************************************************************************************
!> \brief clear the statistics of accepting/rejection moves
!>        because worker statistics will be add seperatelly on masters counters
!> \param move_types counters for acceptance/rejection
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE clear_move_probs(move_types)
      TYPE(tmc_move_type), POINTER                       :: move_types

      CHARACTER(LEN=*), PARAMETER :: routineN = 'clear_move_probs', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_move_handle.F",864)

      move_types%acc_prob(:, :) = 0.5_dp
      move_types%acc_count(:, :) = 0
      move_types%mv_count(:, :) = 0
      move_types%subbox_acc_count(:, :) = 0
      move_types%subbox_count(:, :) = 0
   END SUBROUTINE clear_move_probs

! **************************************************************************************************
!> \brief selects a move type related to the weighings and the entered rnd nr
!> \param move_types structure for storing sizes and probabilities of moves
!> \param rnd random number
!> \retval mv_type (result) move type
!> \author Mandes 12.2012
!> \note function returns a possible move type without the PT swap moves
!> \note (are selected in global tree, this routine is for sub tree elements)
! **************************************************************************************************
   FUNCTION select_random_move_type(move_types, rnd) RESULT(mv_type)
      TYPE(tmc_move_type), POINTER                       :: move_types
      REAL(KIND=dp)                                      :: rnd
      INTEGER                                            :: mv_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'select_random_move_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      REAL(KIND=dp)                                      :: rnd_mv, total_moves

      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_move_handle.F",893)
      IF(.NOT.(rnd .GE. 0.0_dp .AND. rnd .LT. 1.0_dp))CALL cp__a("tmc/tmc_move_handle.F",894)

      CALL timeset(routineN, handle)

      total_moves = SUM(move_types%mv_weight(2:))
      rnd_mv = total_moves*rnd
      mv_type = 0
      search_loop: DO i = 2, SIZE(move_types%mv_weight(:))
         IF (SUM(move_types%mv_weight(2:i)) .GE. rnd_mv) THEN
            mv_type = i
            EXIT search_loop
         END IF
      END DO search_loop

      CALL timestop(handle)
   END FUNCTION select_random_move_type

END MODULE tmc_move_handle
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculation section for TreeMonteCarlo
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_calculations
   USE cell_types,                      ONLY: cell_copy,&
                                              cell_type,&
                                              get_cell,&
                                              init_cell,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_to_string
   USE erf_fn,                          ONLY: erfc
   USE f77_interface,                   ONLY: calc_energy,&
                                              calc_force,&
                                              set_cell
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE parallel_rng_types,              ONLY: get_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type,&
                                              set_rng_stream
   USE physcon,                         ONLY: boltzmann,&
                                              joule
   USE tmc_move_types,                  ONLY: mv_type_MD
   USE tmc_stati,                       ONLY: task_type_MC,&
                                              task_type_gaussian_adaptation,&
                                              task_type_ideal_gas
   USE tmc_tree_types,                  ONLY: tree_type
   USE tmc_types,                       ONLY: tmc_atom_type,&
                                              tmc_env_type,&
                                              tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_calculations'

   PUBLIC :: calc_potential_energy
   PUBLIC :: get_scaled_cell, get_cell_scaling
   PUBLIC :: nearest_distance
   PUBLIC :: geometrical_center, center_of_mass
   PUBLIC :: init_vel, calc_e_kin
   PUBLIC :: compute_estimated_prob
   PUBLIC :: get_subtree_efficiency
CONTAINS

! **************************************************************************************************
!> \brief start the calculation of the energy
!>        (distinguish between exact and approximate)
!> \param conf actual configurations to calculate potential energy
!> \param env_id f77_interface env id
!> \param exact_approx_pot flag if result should be stores in exact or approx
!>        energy variable
!> \param tmc_env TMC environment parameters
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE calc_potential_energy(conf, env_id, exact_approx_pot, &
                                    tmc_env)
      TYPE(tree_type), POINTER                           :: conf
      INTEGER                                            :: env_id
      LOGICAL                                            :: exact_approx_pot
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_potential_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ierr
      LOGICAL                                            :: flag
      REAL(KIND=dp)                                      :: e_pot, rnd
      TYPE(cell_type), POINTER                           :: tmp_cell

      rnd = 0.0_dp

      IF(.NOT.(ASSOCIATED(conf)))CALL cp__a("tmc/tmc_calculations.F",84)
      IF(.NOT.(env_id .GT. 0))CALL cp__a("tmc/tmc_calculations.F",85)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_calculations.F",86)

      SELECT CASE (tmc_env%params%task_type)
      CASE (task_type_gaussian_adaptation)
         !CALL gaussian_adaptation_energy(, )
      CASE (task_type_MC)
         IF (tmc_env%params%pressure .GE. 0.0_dp) THEN
            ALLOCATE (tmp_cell)
            CALL get_scaled_cell(cell=tmc_env%params%cell, box_scale=conf%box_scale, &
                                 scaled_cell=tmp_cell)
            CALL set_cell(env_id=env_id, new_cell=tmp_cell%hmat, ierr=ierr)
            IF(.NOT.(ierr .EQ. 0))CALL cp__a("tmc/tmc_calculations.F",97)
            DEALLOCATE (tmp_cell)
         END IF

         ! TODO check for minimal distances
         flag = .TRUE.
         IF (flag .EQV. .TRUE.) THEN
            IF (tmc_env%params%print_forces .OR. &
                conf%move_type .EQ. mv_type_MD) THEN
               e_pot = 0.0_dp
               conf%frc(:) = 0.0_dp
               CALL calc_force(env_id=env_id, pos=conf%pos, n_el_pos=SIZE(conf%pos), &
                               e_pot=e_pot, force=conf%frc, &
                               n_el_force=SIZE(conf%frc), ierr=ierr)
            ELSE
               e_pot = 0.0_dp
               CALL calc_energy(env_id=env_id, pos=conf%pos, n_el=SIZE(conf%pos), e_pot=e_pot, ierr=ierr)
            END IF
         ELSE
            e_pot = HUGE(e_pot)
         END IF
      CASE (task_type_ideal_gas)
         e_pot = 0.0_dp
      CASE DEFAULT
         CALL cp_abort(cp__l("tmc/tmc_calculations.F",121), &
                       "worker task typ is unknown "// &
                       cp_to_string(tmc_env%params%task_type))
      END SELECT

      ! ---     wait a bit
      rnd = next_random_number(rng_stream=tmc_env%rng_stream)
      !rnd = 0.5
!TODO    IF(worker_random_wait.AND.exact_approx_pot)THEN
!      CALL SYSTEM_CLOCK(time0, time_rate, time_max)
!      wait_end=time0+(1.0+rnd)*worker_wait_msec*time_rate/1000.0
!      !wait_end=time0+((worker_wait_msec*time_rate+999)/1000)
!      time_wait: DO
!        CALL SYSTEM_CLOCK(time1, time_rate, time_max)
!        IF(time1<time0.OR.time1>wait_end) exit time_wait
!      END DO time_wait
!    END IF
      IF (exact_approx_pot) THEN
         conf%potential = e_pot
      ELSE
         conf%e_pot_approx = e_pot
      END IF
   END SUBROUTINE calc_potential_energy

! **************************************************************************************************
!> \brief handles properties and calculations of a scaled cell
!> \param cell .. original cell
!> \param box_scale scaling factors for each direction
!> \param scaled_hmat returns the scaled h matrix (matrix of cell vectors)
!> \param scaled_cell ...
!> \param vol returns the cell volume
!> \param abc ...
!> \param vec a vector, which will be folded (pbc) in the cell
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE get_scaled_cell(cell, box_scale, scaled_hmat, scaled_cell, vol, &
                              abc, vec)
      TYPE(cell_type), INTENT(IN), POINTER               :: cell
      REAL(KIND=dp), DIMENSION(:), POINTER               :: box_scale
      REAL(KIND=dp), DIMENSION(3, 3), OPTIONAL           :: scaled_hmat
      TYPE(cell_type), OPTIONAL, POINTER                 :: scaled_cell
      REAL(KIND=dp), OPTIONAL                            :: vol
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT), OPTIONAL :: abc
      REAL(KIND=dp), DIMENSION(3), OPTIONAL              :: vec

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_scaled_cell', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: new_scaled_cell
      TYPE(cell_type), POINTER                           :: tmp_cell

      IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("tmc/tmc_calculations.F",172)
      IF(.NOT.(ASSOCIATED(box_scale)))CALL cp__a("tmc/tmc_calculations.F",173)

      new_scaled_cell = .FALSE.

      IF (.NOT. PRESENT(scaled_cell)) THEN
         ALLOCATE (tmp_cell)
         new_scaled_cell = .TRUE.
      ELSE
         tmp_cell => scaled_cell
      END IF
      CALL cell_copy(cell_in=cell, cell_out=tmp_cell)
      tmp_cell%hmat(:, 1) = tmp_cell%hmat(:, 1)*box_scale(1)
      tmp_cell%hmat(:, 2) = tmp_cell%hmat(:, 2)*box_scale(2)
      tmp_cell%hmat(:, 3) = tmp_cell%hmat(:, 3)*box_scale(3)
      CALL init_cell(cell=tmp_cell)

      IF (PRESENT(scaled_hmat)) &
         scaled_hmat(:, :) = tmp_cell%hmat

      IF (PRESENT(vec)) THEN
         vec = pbc(r=vec, cell=tmp_cell)
      END IF

      IF (PRESENT(vol)) CALL get_cell(cell=tmp_cell, deth=vol)
      IF (PRESENT(abc)) CALL get_cell(cell=tmp_cell, abc=abc)
      IF (new_scaled_cell) DEALLOCATE (tmp_cell)
   END SUBROUTINE get_scaled_cell

! **************************************************************************************************
!> \brief handles properties and calculations of a scaled cell
!> \param cell .. original cell
!> \param scaled_hmat returns the scaled h matrix (matrix of cell vectors)
!> \param box_scale scaling factors for each direction
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE get_cell_scaling(cell, scaled_hmat, box_scale)
      TYPE(cell_type), INTENT(IN), POINTER               :: cell
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: scaled_hmat
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: box_scale

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_cell_scaling', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(3)                        :: abc_new, abc_orig
      TYPE(cell_type), POINTER                           :: tmp_cell

      IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("tmc/tmc_calculations.F",219)

      ALLOCATE (tmp_cell)
      CALL cell_copy(cell_in=cell, cell_out=tmp_cell)
      tmp_cell%hmat(:, :) = scaled_hmat(:, :)
      CALL init_cell(cell=tmp_cell)
      CALL get_cell(cell=cell, abc=abc_orig)
      CALL get_cell(cell=tmp_cell, abc=abc_new)

      box_scale(:) = abc_new(:)/abc_orig(:)

      DEALLOCATE (tmp_cell)
   END SUBROUTINE get_cell_scaling

! **************************************************************************************************
!> \brief neares distance of atoms within the periodic boundary condition
!> \param x1 ...
!> \param x2 ...
!> \param cell ...
!> \param box_scale ...
!> \retval res ...
!> \author Mandes 11.2012
! **************************************************************************************************
   FUNCTION nearest_distance(x1, x2, cell, box_scale) RESULT(res)
      REAL(KIND=dp), DIMENSION(:)                        :: x1, x2
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: box_scale
      REAL(KIND=dp)                                      :: res

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nearest_distance', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(3)                        :: dist_vec
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tmp_box_scale

      NULLIFY (tmp_box_scale)

      IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("tmc/tmc_calculations.F",256)
      IF(.NOT.(SIZE(x1) .EQ. 3))CALL cp__a("tmc/tmc_calculations.F",257)
      IF(.NOT.(SIZE(x2) .EQ. 3))CALL cp__a("tmc/tmc_calculations.F",258)

      dist_vec(:) = x2(:)-x1(:) ! distance vector between atoms
      ALLOCATE (tmp_box_scale(3))
      IF (PRESENT(box_scale)) THEN
         IF(.NOT.(SIZE(box_scale) .EQ. 3))CALL cp__a("tmc/tmc_calculations.F",263)
         tmp_box_scale(:) = box_scale
      ELSE
         tmp_box_scale(:) = 1.0_dp
      END IF
      CALL get_scaled_cell(cell=cell, box_scale=box_scale, vec=dist_vec)
      res = SQRT(SUM(dist_vec(:)*dist_vec(:)))
      DEALLOCATE (tmp_box_scale)
   END FUNCTION nearest_distance

! **************************************************************************************************
!> \brief calculate the geometrical center of an amount of atoms
!>        array size should be multiple of dim_per_elem
!> \param pos list of atoms
!> \param center return value, the geometrical center
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE geometrical_center(pos, center)
      REAL(KIND=dp), DIMENSION(:)                        :: pos
      REAL(KIND=dp), DIMENSION(:), POINTER               :: center

      CHARACTER(LEN=*), PARAMETER :: routineN = 'geometrical_center', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      IF(.NOT.(ASSOCIATED(center)))CALL cp__a("tmc/tmc_calculations.F",289)
      IF(.NOT.(SIZE(pos) .GE. SIZE(center)))CALL cp__a("tmc/tmc_calculations.F",290)

      ! start the timing
      CALL timeset(routineN, handle)

      center = 0.0_dp
      DO i = 1, SIZE(pos), SIZE(center)
         center(:) = center(:)+ &
                     pos(i:i+SIZE(center)-1)/(SIZE(pos)/REAL(SIZE(center), KIND=dp))
      END DO
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE geometrical_center

! **************************************************************************************************
!> \brief calculate the center of mass of an amount of atoms
!>        array size should be multiple of dim_per_elem
!> \param pos ...
!> \param atoms ...
!> \param center ...
!> \param
!> \param
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE center_of_mass(pos, atoms, center)
      REAL(KIND=dp), DIMENSION(:)                        :: pos
      TYPE(tmc_atom_type), DIMENSION(:), OPTIONAL        :: atoms
      REAL(KIND=dp), DIMENSION(:), POINTER               :: center

      CHARACTER(LEN=*), PARAMETER :: routineN = 'center_of_mass', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      REAL(KIND=dp)                                      :: mass_sum, mass_tmp

      IF(.NOT.(ASSOCIATED(center)))CALL cp__a("tmc/tmc_calculations.F",324)
      IF(.NOT.(SIZE(pos) .GE. SIZE(center)))CALL cp__a("tmc/tmc_calculations.F",325)

      ! start the timing
      CALL timeset(routineN, handle)

      center = 0.0_dp
      mass_sum = 0.0_dp
      DO i = 1, SIZE(pos), SIZE(center)
         IF (PRESENT(atoms)) THEN
            IF(.NOT.(SIZE(atoms) .EQ. SIZE(pos)/SIZE(center)))CALL cp__a("tmc/tmc_calculations.F",334)
            mass_tmp = atoms(INT(i/REAL(SIZE(center), KIND=dp))+1)%mass
            center(:) = center(:)+pos(i:i+SIZE(center)-1)/ &
                        (SIZE(pos)/REAL(SIZE(center), KIND=dp))*mass_tmp
            mass_sum = mass_sum+mass_tmp
         ELSE
            CALL cp__w("tmc/tmc_calculations.F",340,"try to calculate center of mass without any mass.")
            center(:) = center(:)+pos(i:i+SIZE(center)-1)/ &
                        (SIZE(pos)/REAL(SIZE(center), KIND=dp))
            mass_sum = 1.0_dp
         END IF
      END DO
      center(:) = center(:)/mass_sum
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE center_of_mass

! **************************************************************************************************
!> \brief routine sets initial velocity, using the Box-Muller Method for Normal
!>         (Gaussian) Deviates
!> \param vel ...
!> \param atoms ...
!> \param temerature ...
!> \param rng_stream ...
!> \param rnd_seed ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE init_vel(vel, atoms, temerature, rng_stream, rnd_seed)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: vel
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms
      REAL(KIND=dp)                                      :: temerature
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      REAL(KIND=dp), DIMENSION(3, 2, 3)                  :: rnd_seed

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_vel', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: kB, mass_tmp, rnd1, rnd2

      kB = boltzmann/joule

      IF(.NOT.(ASSOCIATED(vel)))CALL cp__a("tmc/tmc_calculations.F",375)
      IF(.NOT.(ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_calculations.F",376)

      CALL set_rng_stream(rng_stream=rng_stream, bg=rnd_seed(:, :, 1), &
                          cg=rnd_seed(:, :, 2), ig=rnd_seed(:, :, 3))
      DO i = 1, SIZE(vel)
         rnd1 = next_random_number(rng_stream)
         rnd2 = next_random_number(rng_stream)

         mass_tmp = atoms(INT(i/REAL(3, KIND=dp))+1)%mass

         vel(i) = SQRT(-2.0_dp*LOG(rnd1))*COS(2.0_dp*PI*rnd2)* &
                  SQRT(kB*temerature/mass_tmp)
      END DO
      CALL get_rng_stream(rng_stream=rng_stream, bg=rnd_seed(:, :, 1), &
                          cg=rnd_seed(:, :, 2), ig=rnd_seed(:, :, 3))

   END SUBROUTINE init_vel

! **************************************************************************************************
!> \brief routine calculates the kinetic energy, using the velocities
!>        and atom mass, both in atomic units
!> \param vel ...
!> \param atoms ...
!> \retval ekin ...
!> \author Mandes 11.2012
! **************************************************************************************************
   FUNCTION calc_e_kin(vel, atoms) RESULT(ekin)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: vel
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms
      REAL(KIND=dp)                                      :: ekin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_e_kin', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: mass_tmp

      IF(.NOT.(ASSOCIATED(vel)))CALL cp__a("tmc/tmc_calculations.F",412)
      IF(.NOT.(ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_calculations.F",413)
      ekin = 0.0_dp

      DO i = 1, SIZE(vel)
         mass_tmp = atoms(INT(i/REAL(3, KIND=dp))+1)%mass
         ekin = ekin+0.5_dp*mass_tmp*vel(i)*vel(i)
      END DO
   END FUNCTION calc_e_kin

! **************************************************************************************************
!> \brief assuming an (exponential) decreasing function, this function
!>        extrapolate the converged value
!> \param v1 function values
!> \param v2 function values
!> \param v3 function values
!> \param extrapolate extrapolated final value (result)
!> \param res_err error of the result
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE three_point_extrapolate(v1, v2, v3, extrapolate, res_err)
      REAL(KIND=dp)                            :: v1, v2, v3
      REAL(KIND=dp), INTENT(OUT)               :: extrapolate, res_err

      CHARACTER(LEN=*), PARAMETER :: routineN = 'three_point_extrapolate', &
                                     routineP = moduleN//':'//routineN

      REAL(KIND=dp)                            :: e1, e2, e3
      REAL(KIND=dp)                            :: a, b, c, d12, d23, ddd

      extrapolate = HUGE(extrapolate)

      !> solve({exp(a+b)+c = e1, exp(2*a+b)+c = e2, exp(3*a+b)+c = e3}, [a, b, c])
      !> solve({a*b+c = e1, a^2*b+c = e2, a^3*b+c = e3}, [a, b, c]);
      !   [[                                   3                   2         ]]
      !   [[    -e3 + e2              (e1 - e2)                 -e2  + e1 e3 ]]
      !   [[a = --------, b = ---------------------------, c = --------------]]
      !   [[    e1 - e2       (-e3 + e2) (e3 - 2 e2 + e1)      e3 - 2 e2 + e1]]

      ! sort so that e1>=e2>=e3
      e1 = v1; e2 = v2; e3 = v3
      CALL swap(e1, e2)
      CALL swap(e1, e3)
      CALL swap(e2, e3)
      ! we need extra care if some of the difference e1-e2, e3-e2 are nearly zero,
      !  since the formulae suffer from sever loss of precision
      d12 = e1-e2
      d23 = e2-e3
      ddd = d12-d23
      IF (d12 == 0 .OR. d23 == 0 .OR. ABS(ddd) == 0) THEN
         ! a degenerate case, we do no extrapolation
         extrapolate = e3
         res_err = e1-e3
      ELSE
         a = d23/d12
         b = (d12**3/(d23*ddd))
         c = e2-(d12*d23)/ddd
         ! extrapolation, let's only look 4 iterations ahead, more is presumably anyway not accurate
         ! fewer is maybe more stable
         extrapolate = a**7*b+c
         res_err = e3-extrapolate
      ENDIF
      IF(.NOT.(extrapolate .NE. HUGE(extrapolate)))CALL cp__a("tmc/tmc_calculations.F",474)
   CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param x1 ...
!> \param x2 ...
! **************************************************************************************************
      SUBROUTINE swap(x1, x2)
      REAL(KIND=dp)                                      :: x1, x2

      REAL(KIND=dp)                                      :: tmp

         IF (x2 > x1) THEN
            tmp = x2
            x2 = x1
            x1 = tmp
         ENDIF
      END SUBROUTINE swap
   END SUBROUTINE three_point_extrapolate

! **************************************************************************************************
!> \brief calculates the probability of acceptance for given intervals of the
!>        exact energy
!> \param E_n_mu energy distribution of new configuration
!> \param E_n_sigma energy distribution of new configuration
!> \param E_o_mu energy distribution of old configuration
!> \param E_o_sigma energy distribution of old configuration
!> \param E_classical_diff the difference in approximated energies for the
!>        old and new configuration (E_o-E_n)
!> \param prior_mu energy distribution of the already converged
!>         energies
!> \param prior_sigma energy distribution of the already converged
!>         energies
!> \param p the random number, the criteria has to be smaller than this
!> \param beta ...
!> \retval prob return probability of acceptance
!> \author Mandes 12.2012
! **************************************************************************************************
   FUNCTION compute_prob(E_n_mu, E_n_sigma, E_o_mu, E_o_sigma, E_classical_diff, &
                         prior_mu, prior_sigma, p, beta) RESULT(prob)
      REAL(KIND=dp)                                      :: E_n_mu, E_n_sigma, E_o_mu, E_o_sigma, &
                                                            E_classical_diff, prior_mu, &
                                                            prior_sigma, p, beta, prob

!    INTEGER       :: io,in
!    REAL(KIND=dp) :: diff,E_n,E_o,surface,lower_bound,upper_bound,delta

      prob = 0.5_dp*ERFC(-0.5_dp*SQRT(2.0_dp)*( &
                         (-prior_sigma**2-E_o_sigma**2-E_n_sigma**2)*LOG(p)+ &
                         ((E_classical_diff-E_n_mu+E_o_mu)*prior_sigma**2-prior_mu*(E_n_sigma**2+E_o_sigma**2))*beta)/ &
                         (SQRT(E_o_sigma**2+E_n_sigma**2)*SQRT(prior_sigma**2+E_o_sigma**2+E_n_sigma**2)*prior_sigma*beta))

      prob = MIN(1.0_dp-EPSILON(1.0_dp), MAX(EPSILON(1.0_dp), prob))

   END FUNCTION compute_prob

! **************************************************************************************************
!> \brief extimates the probability of acceptance considering the intermetiate
!>        step energies
!> \param elem_old old/parent sub tree element
!> \param elem_new new/actual sub tree element, which schould be checked
!> \param E_classical_diff difference in the classical energy of the old and
!>        new configuration
!> \param rnd_nr random number acceptance check will be done with
!> \param beta 1/(kB*T) can differ for different acceptance checks
!> \param tmc_params TMC environment parameters
!> \retval prob estimated acceptance probability
!> \author Mandes 12.2012
! **************************************************************************************************
   FUNCTION compute_estimated_prob(elem_old, elem_new, E_classical_diff, &
                                   rnd_nr, beta, tmc_params) RESULT(prob)
      TYPE(tree_type), POINTER                           :: elem_old, elem_new
      REAL(KIND=dp)                                      :: E_classical_diff, rnd_nr, beta
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      REAL(KIND=dp)                                      :: prob

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_estimated_prob', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: E_mu_tmp, E_n_mu, E_n_sigma, E_o_mu, &
                                                            E_o_sigma, E_sigma_tmp, prior_sigma

      IF(.NOT.(ASSOCIATED(elem_old)))CALL cp__a("tmc/tmc_calculations.F",557)
      IF(.NOT.(ASSOCIATED(elem_new)))CALL cp__a("tmc/tmc_calculations.F",558)
      IF(.NOT.(rnd_nr .GT. 0.0_dp))CALL cp__a("tmc/tmc_calculations.F",559)

      ! start the timing
      CALL timeset(routineN, handle)

      prob = -1.0_dp
      IF ((elem_new%scf_energies_count .GE. 3) .AND. &
          (elem_old%scf_energies_count .GE. 3) .AND. &
          tmc_params%prior_NMC_acc%counter .GE. 10) THEN
         !-- first the new element energy estimation
         ! using 3 point extrapolation of two different intervals -> more stable estimation
         ! the energies are sorted in the three_point_extrapolate routine !
         ! But with array of length 4 we have to select the 3 connected ones
         CALL three_point_extrapolate(v1=elem_new%scf_energies(MOD(elem_new%scf_energies_count-3, 4)+1), &
                                      v2=elem_new%scf_energies(MOD(elem_new%scf_energies_count-2, 4)+1), &
                                      v3=elem_new%scf_energies(MOD(elem_new%scf_energies_count-1, 4)+1), &
                                      extrapolate=E_mu_tmp, res_err=E_sigma_tmp)
         IF ((elem_new%scf_energies_count .GT. 3)) THEN
            CALL three_point_extrapolate(v1=elem_new%scf_energies(MOD(elem_new%scf_energies_count-4, 4)+1), &
                                         v2=elem_new%scf_energies(MOD(elem_new%scf_energies_count-3, 4)+1), &
                                         v3=elem_new%scf_energies(MOD(elem_new%scf_energies_count-2, 4)+1), &
                                         extrapolate=E_n_mu, res_err=E_n_sigma)
            E_n_sigma = MAX(E_n_sigma, ABS(E_n_mu-E_mu_tmp))
         ELSE
            E_n_sigma = E_sigma_tmp
            E_n_mu = E_mu_tmp
         END IF

         !-- the old/parent element energy estimation
         CALL three_point_extrapolate(v1=elem_old%scf_energies(MOD(elem_old%scf_energies_count-3, 4)+1), &
                                      v2=elem_old%scf_energies(MOD(elem_old%scf_energies_count-2, 4)+1), &
                                      v3=elem_old%scf_energies(MOD(elem_old%scf_energies_count-1, 4)+1), &
                                      extrapolate=E_mu_tmp, res_err=E_sigma_tmp)
         IF ((elem_old%scf_energies_count .GT. 3)) THEN
            CALL three_point_extrapolate(v1=elem_old%scf_energies(MOD(elem_old%scf_energies_count-4, 4)+1), &
                                         v2=elem_old%scf_energies(MOD(elem_old%scf_energies_count-3, 4)+1), &
                                         v3=elem_old%scf_energies(MOD(elem_old%scf_energies_count-2, 4)+1), &
                                         extrapolate=E_o_mu, res_err=E_o_sigma)
            E_o_sigma = MAX(E_o_sigma, ABS(E_o_mu-E_mu_tmp))
         ELSE
            E_o_sigma = E_sigma_tmp
            E_o_mu = E_mu_tmp
         END IF

         ! calculate the estimation for the average of the trajectory elements
         prior_sigma = SQRT(ABS(tmc_params%prior_NMC_acc%aver_2 &
                                -tmc_params%prior_NMC_acc%aver**2))

         ! calculate the probability of acceptance for those two elements with their energy
         ! swap and 2 potential moves are distinguished using the difference in classical energy and different betas
         prob = compute_prob(E_n_mu=E_n_mu, E_n_sigma=E_n_sigma, E_o_mu=E_o_mu, E_o_sigma=E_o_sigma, &
                             E_classical_diff=E_classical_diff, &
                             prior_mu=tmc_params%prior_NMC_acc%aver, prior_sigma=prior_sigma, &
                             p=rnd_nr, beta=beta)
      END IF
      ! end the timing
      CALL timestop(handle)
   END FUNCTION compute_estimated_prob

! **************************************************************************************************
!> \brief calculated the rate of used tree elements to created tree elements
!>        for every temperature
!> \param tmc_env TMC environment variables
!> \param eff result efficiency
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE get_subtree_efficiency(tmc_env, eff)
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_subtree_efficiency', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_calculations.F",634)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_calculations.F",635)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_calculations.F",636)

      eff(:) = 0.0_dp

      DO i = 1, tmc_env%params%nr_temp
         IF (tmc_env%m_env%tree_node_count(i) > 0) &
            eff(i) = tmc_env%params%move_types%mv_count(0, i)/ &
                     (tmc_env%m_env%tree_node_count(i)*1.0_dp)
         eff(0) = eff(0)+tmc_env%params%move_types%mv_count(0, i)/ &
                  (SUM(tmc_env%m_env%tree_node_count(1:))*1.0_dp)
      END DO
   END SUBROUTINE get_subtree_efficiency
END MODULE tmc_calculations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief - writing and printing the files, trajectory (pos, cell, dipoles) as
!>        well as restart files
!>        - usually just the Markov Chain elements are regarded, the elements
!>        beside this trajectory are neglected
!>        - futrthermore (by option) just the accepted configurations
!>          are print out to reduce the file sizes
!> \par History
!>      12.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_file_io
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE physcon,                         ONLY: au2a => angstrom
   USE tmc_analysis_types,              ONLY: tmc_analysis_env
   USE tmc_calculations,                ONLY: get_cell_scaling,&
                                              get_scaled_cell
   USE tmc_move_types,                  ONLY: nr_mv_types
   USE tmc_stati,                       ONLY: TMC_STATUS_FAILED,&
                                              TMC_STATUS_OK,&
                                              TMC_STATUS_WAIT_FOR_NEW_TASK,&
                                              tmc_default_restart_in_file_name,&
                                              tmc_default_restart_out_file_name,&
                                              tmc_default_trajectory_file_name
   USE tmc_tree_types,                  ONLY: elem_array_type,&
                                              tree_type
   USE tmc_types,                       ONLY: tmc_env_type,&
                                              tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_file_io'

   ! filename manipulation
   PUBLIC :: expand_file_name_char, expand_file_name_temp, expand_file_name_int
   ! read/write restart file
   PUBLIC :: print_restart_file, read_restart_file
   ! write the configuration
   PUBLIC :: write_result_list_element
   PUBLIC :: write_element_in_file
   PUBLIC :: write_dipoles_in_file
   ! analysis read
   PUBLIC :: analyse_files_open, read_element_from_file, analyse_files_close

CONTAINS

!------------------------------------------------------------------------------
! routines for manipulating the file name
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief placing a character string at the end of a file name
!>        (instead of the ending)
!> \param file_name original file name
!> \param extra string to be added before the file extention
!> \retval result_file_name the new filename
!> \author Mandes 11.2012
! **************************************************************************************************
   FUNCTION expand_file_name_ending(file_name, extra) RESULT(result_file_name)
      CHARACTER(LEN=*)                                   :: file_name, extra
      CHARACTER(LEN=default_path_length)                 :: result_file_name

      CHARACTER(LEN=*), PARAMETER :: routineN = 'expand_file_name_ending', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ind

      IF(.NOT.(file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",81)

      ind = INDEX(file_name, ".", BACK=.TRUE.)
      IF (.NOT. ind .EQ. 0) THEN
         WRITE (result_file_name, *) file_name(1:ind-1), ".", &
            TRIM(ADJUSTL(extra))
      ELSE
         WRITE (result_file_name, *) TRIM(file_name), ".", extra
      END IF
      result_file_name = TRIM(ADJUSTL(result_file_name))
      IF(.NOT.(result_file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",91)
   END FUNCTION expand_file_name_ending

! **************************************************************************************************
!> \brief placing a character string at the end of a file name
!>        (before the file extention)
!> \param file_name original file name
!> \param extra string to be added before the file extention
!> \retval result_file_name the new filename
!> \author Mandes 11.2012
! **************************************************************************************************
   FUNCTION expand_file_name_char(file_name, extra) RESULT(result_file_name)
      CHARACTER(LEN=*)                                   :: file_name, extra
      CHARACTER(LEN=default_path_length)                 :: result_file_name

      CHARACTER(LEN=*), PARAMETER :: routineN = 'expand_file_name_char', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ind

      IF(.NOT.(file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",111)

      ind = INDEX(file_name, ".", BACK=.TRUE.)
      IF (.NOT. ind .EQ. 0) THEN
         WRITE (result_file_name, *) file_name(1:ind-1), "_", &
            TRIM(ADJUSTL(extra)), file_name(ind:LEN_TRIM(file_name))
      ELSE
         WRITE (result_file_name, *) TRIM(file_name), "_", extra
      END IF
      result_file_name = TRIM(ADJUSTL(result_file_name))
      IF(.NOT.(result_file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",121)
   END FUNCTION expand_file_name_char

! **************************************************************************************************
!> \brief placing the temperature at the end of a file name
!>        (before the file extention)
!> \param file_name original file name
!> \param rvalue temperature to be added
!> \retval result_file_name the new filename
!> \author Mandes 11.2012
! **************************************************************************************************
   FUNCTION expand_file_name_temp(file_name, rvalue) RESULT(result_file_name)
      CHARACTER(LEN=*)                                   :: file_name
      REAL(KIND=dp)                                      :: rvalue
      CHARACTER(LEN=default_path_length)                 :: result_file_name

      CHARACTER(LEN=*), PARAMETER :: routineN = 'expand_file_name_temp', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=18)                                  :: rval_to_string
      INTEGER                                            :: ind

      IF(.NOT.(file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",143)

      rval_to_string = ""

      WRITE (rval_to_string, "(F16.2)") rvalue
      ind = INDEX(file_name, ".", BACK=.TRUE.)
      IF (.NOT. ind .EQ. 0) THEN
         WRITE (result_file_name, *) file_name(1:ind-1), "_T", &
            TRIM(ADJUSTL(rval_to_string)), file_name(ind:LEN_TRIM(file_name))
      ELSE
         IF (LEN(file_name) .EQ. 0) THEN
            WRITE (result_file_name, *) TRIM(file_name), "T", TRIM(ADJUSTL(rval_to_string)), &
               file_name(ind:LEN_TRIM(file_name))
         ELSE
            WRITE (result_file_name, *) TRIM(file_name), "_T", TRIM(ADJUSTL(rval_to_string))
         END IF
      END IF
      result_file_name = TRIM(ADJUSTL(result_file_name))
      IF(.NOT.(result_file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",161)
   END FUNCTION expand_file_name_temp

! **************************************************************************************************
!> \brief placing an integer at the end of a file name
!>        (before the file extention)
!> \param file_name original file name
!> \param ivalue number to be added
!> \retval result_file_name the new filename
!> \author Mandes 11.2012
! **************************************************************************************************
   FUNCTION expand_file_name_int(file_name, ivalue) RESULT(result_file_name)
      CHARACTER(LEN=*)                                   :: file_name
      INTEGER                                            :: ivalue
      CHARACTER(LEN=default_path_length)                 :: result_file_name

      CHARACTER(LEN=*), PARAMETER :: routineN = 'expand_file_name_int', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=18)                                  :: rval_to_string
      INTEGER                                            :: ind

      IF(.NOT.(file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",183)

      rval_to_string = ""

      WRITE (rval_to_string, *) ivalue
      ind = INDEX(file_name, ".", BACK=.TRUE.)
      IF (.NOT. ind .EQ. 0) THEN
         WRITE (result_file_name, *) file_name(1:ind-1), "_", &
            TRIM(ADJUSTL(rval_to_string)), file_name(ind:LEN_TRIM(file_name))
      ELSE
         IF (LEN(file_name) .EQ. 0) THEN
            WRITE (result_file_name, *) TRIM(file_name), "", TRIM(ADJUSTL(rval_to_string)), &
               file_name(ind:LEN_TRIM(file_name))
         ELSE
            WRITE (result_file_name, *) TRIM(file_name), "_", TRIM(ADJUSTL(rval_to_string)), &
               file_name(ind:LEN_TRIM(file_name))
         END IF
      END IF
      result_file_name = TRIM(ADJUSTL(result_file_name))
      IF(.NOT.(result_file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",202)
   END FUNCTION expand_file_name_int

!------------------------------------------------------------------------------
! routines for reading and writing RESTART file
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief prints out the TMC restart files with all last configurations and
!>        counters etc.
!> \param tmc_env the tmc environment, storing result lists and counters an in
!>        temperatures
!> \param job_counts the counters for counting the submitted different job types
!> \param timings ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE print_restart_file(tmc_env, job_counts, timings)
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      INTEGER, DIMENSION(:)                              :: job_counts
      REAL(KIND=dp), DIMENSION(4)                        :: timings

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_restart_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: c_tmp, file_name
      INTEGER                                            :: f_unit, i

      c_tmp = ""
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_file_io.F",229)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_file_io.F",230)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_file_io.F",231)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%gt_act)))CALL cp__a("tmc/tmc_file_io.F",232)

      WRITE (c_tmp, FMT='(I9.9)') tmc_env%m_env%result_count(0)
      file_name = TRIM(expand_file_name_char( &
                       file_name=tmc_default_restart_out_file_name, &
                       extra=c_tmp))
      CALL open_file(file_name=file_name, file_status="REPLACE", &
                     file_action="WRITE", file_form="UNFORMATTED", &
                     unit_number=f_unit)
      WRITE (f_unit) SIZE(tmc_env%params%Temp)
      WRITE (f_unit) tmc_env%params%Temp(:), &
         tmc_env%m_env%gt_act%nr, &
         tmc_env%m_env%gt_act%rng_seed, &
         tmc_env%m_env%gt_act%rnd_nr, &
         tmc_env%m_env%gt_act%prob_acc, &
         tmc_env%m_env%gt_act%mv_conf, &
         tmc_env%m_env%gt_act%mv_next_conf, &
         tmc_env%m_env%result_count(0:), &
         tmc_env%params%move_types%mv_weight, &
         tmc_env%params%move_types%acc_count, &
         tmc_env%params%move_types%mv_count, &
         tmc_env%params%move_types%subbox_acc_count, &
         tmc_env%params%move_types%subbox_count, &
         tmc_env%params%cell%hmat, &
         job_counts, &
         timings
      DO i = 1, SIZE(tmc_env%params%Temp)
         WRITE (f_unit) tmc_env%m_env%result_list(i)%elem%nr, &
            tmc_env%m_env%result_list(i)%elem%rng_seed, &
            tmc_env%m_env%result_list(i)%elem%pos, &
            tmc_env%m_env%result_list(i)%elem%vel, &
            tmc_env%m_env%result_list(i)%elem%box_scale, &
            tmc_env%m_env%result_list(i)%elem%potential, &
            tmc_env%m_env%result_list(i)%elem%e_pot_approx, &
            tmc_env%m_env%result_list(i)%elem%ekin, &
            tmc_env%m_env%result_list(i)%elem%ekin_before_md, &
            tmc_env%m_env%result_list(i)%elem%temp_created
      END DO
      CALL close_file(unit_number=f_unit)
      ! write the file, where the restart file name is written in
      CALL open_file(file_name=tmc_default_restart_in_file_name, &
                     file_action="WRITE", file_status="REPLACE", &
                     unit_number=f_unit)
      WRITE (f_unit, *) TRIM(file_name)
      CALL close_file(unit_number=f_unit)
   END SUBROUTINE print_restart_file

! **************************************************************************************************
!> \brief reads the TMC restart file with all last configurations and
!>        counters etc.
!> \param tmc_env the tmc environment, storing result lists and counters an in
!>        temperatures
!> \param job_counts the counters for counting the submitted different job types
!> \param timings ...
!> \param file_name the restart file name
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE read_restart_file(tmc_env, job_counts, timings, file_name)
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      INTEGER, DIMENSION(:)                              :: job_counts
      REAL(KIND=dp), DIMENSION(4)                        :: timings
      CHARACTER(LEN=*)                                   :: file_name

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_restart_file', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: file_ptr, i, temp_size
      LOGICAL                                            :: flag
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tmp_temp
      REAL(KIND=dp), DIMENSION(nr_mv_types)              :: mv_weight_tmp

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_file_io.F",303)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_file_io.F",304)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_file_io.F",305)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%gt_act)))CALL cp__a("tmc/tmc_file_io.F",306)

      IF (file_name .EQ. tmc_default_restart_in_file_name) THEN
         INQUIRE (FILE=tmc_default_restart_in_file_name, EXIST=flag)
         IF(.NOT.(flag))CALL cp__a("tmc/tmc_file_io.F",310)
         CALL open_file(file_name=tmc_default_restart_in_file_name, file_status="OLD", &
                        file_action="READ", unit_number=file_ptr)
         READ (file_ptr, *) file_name
         CALL close_file(unit_number=file_ptr)
      END IF

      CALL open_file(file_name=file_name, file_status="OLD", file_form="UNFORMATTED", &
                     file_action="READ", unit_number=file_ptr)
      READ (file_ptr) temp_size
      IF (temp_size .NE. SIZE(tmc_env%params%Temp)) &
         CALL cp_abort(cp__l("tmc/tmc_file_io.F",321), &
                       "the actual specified temperatures does not "// &
                       "fit in amount with the one from restart file ")
      ALLOCATE (tmp_temp(temp_size))
      READ (file_ptr) tmp_temp(:), &
         tmc_env%m_env%gt_act%nr, &
         tmc_env%m_env%gt_act%rng_seed, &
         tmc_env%m_env%gt_act%rnd_nr, &
         tmc_env%m_env%gt_act%prob_acc, &
         tmc_env%m_env%gt_act%mv_conf, & !
         tmc_env%m_env%gt_act%mv_next_conf, & !
         tmc_env%m_env%result_count(0:), &
         mv_weight_tmp, & !
         tmc_env%params%move_types%acc_count, &
         tmc_env%params%move_types%mv_count, &
         tmc_env%params%move_types%subbox_acc_count, &
         tmc_env%params%move_types%subbox_count, & !
         tmc_env%params%cell%hmat, &
         job_counts, &
         timings

      IF (ANY(ABS(tmc_env%params%Temp(:)-tmp_temp(:)) .GE. 0.005)) &
         CALL cp_abort(cp__l("tmc/tmc_file_io.F",343), "the temperatures differ from the previous calculation. "// &
                       "There were the following temperatures used:")
      IF (ANY(mv_weight_tmp(:) .NE. tmc_env%params%move_types%mv_weight(:))) &
         CALL cp__w("tmc/tmc_file_io.F",346,"The amount of mv types differs between the original and the restart run.")

      DO i = 1, SIZE(tmc_env%params%Temp)
         tmc_env%m_env%gt_act%conf(i)%elem => tmc_env%m_env%result_list(i)%elem
         READ (file_ptr) tmc_env%m_env%result_list(i)%elem%nr, &
            tmc_env%m_env%result_list(i)%elem%rng_seed, &
            tmc_env%m_env%result_list(i)%elem%pos, &
            tmc_env%m_env%result_list(i)%elem%vel, &
            tmc_env%m_env%result_list(i)%elem%box_scale, &
            tmc_env%m_env%result_list(i)%elem%potential, &
            tmc_env%m_env%result_list(i)%elem%e_pot_approx, &
            tmc_env%m_env%result_list(i)%elem%ekin, &
            tmc_env%m_env%result_list(i)%elem%ekin_before_md, &
            tmc_env%m_env%result_list(i)%elem%temp_created
      END DO
      CALL close_file(unit_number=file_ptr)
   END SUBROUTINE read_restart_file

   !----------------------------------------------------------------------------
   ! printing configuration in file
   !----------------------------------------------------------------------------

! **************************************************************************************************
!> \brief select the correct configuration to print out the
!>        (coordinates, forces, cell ...)
!> \param result_list list of configurations for each temperature
!> \param result_count list with number of Markov Chain number
!>          for each teperature (index 0 for global tree)
!> \param conf_updated index of the updated (modified element)
!> \param accepted acceptance flag
!> \param tmc_params TMC environment parameters
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE write_result_list_element(result_list, result_count, conf_updated, &
                                        accepted, tmc_params)
      TYPE(elem_array_type), DIMENSION(:), POINTER       :: result_list
      INTEGER, DIMENSION(:), POINTER                     :: result_count
      INTEGER                                            :: conf_updated
      LOGICAL, INTENT(IN)                                :: accepted
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_result_list_element', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: handle, i

      file_name = ""

      IF(.NOT.(ASSOCIATED(result_list)))CALL cp__a("tmc/tmc_file_io.F",395)
      IF(.NOT.(ASSOCIATED(result_count)))CALL cp__a("tmc/tmc_file_io.F",396)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_file_io.F",397)
      IF(.NOT.(ASSOCIATED(tmc_params%Temp)))CALL cp__a("tmc/tmc_file_io.F",398)
      IF(.NOT.(conf_updated .GE. 0))CALL cp__a("tmc/tmc_file_io.F",399)
      IF(.NOT.(conf_updated .LE. SIZE(tmc_params%Temp)))CALL cp__a("tmc/tmc_file_io.F",400)

      ! start the timing
      CALL timeset(routineN, handle)

      IF (conf_updated .EQ. 0) THEN
         ! for debugging print every configuration of every temperature
         DO i = 1, SIZE(tmc_params%Temp)
            WRITE (file_name, *) "every_step_", TRIM(tmc_default_trajectory_file_name)
            CALL write_element_in_file(elem=result_list(i)%elem, &
                                       tmc_params=tmc_params, conf_nr=result_count(0), &
                                       file_name=expand_file_name_temp(file_name=file_name, rvalue=tmc_params%Temp(i)))
         END DO
      ELSE
         IF ((.NOT. tmc_params%print_only_diff_conf) .OR. &
             (tmc_params%print_only_diff_conf .AND. accepted)) THEN
            CALL write_element_in_file(elem=result_list(conf_updated)%elem, &
                                       tmc_params=tmc_params, conf_nr=result_count(conf_updated), &
                                       file_name=expand_file_name_temp(file_name=TRIM(tmc_default_trajectory_file_name), &
                                                                       rvalue=tmc_params%Temp(conf_updated)))
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE write_result_list_element

! **************************************************************************************************
!> \brief writes the trajectory element in a file from sub tree element
!> \param elem actual tree element to be printed out
!> \param tmc_params TMC environment parameters
!> \param temp_index ...
!> \param file_name file name will be extended by type of file (pos, cell,...)
!> \param conf_nr Markov chain element number
!> \param conf_info whole header line
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE write_element_in_file(elem, tmc_params, temp_index, file_name, conf_nr, &
                                    conf_info)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      INTEGER, OPTIONAL                                  :: temp_index
      CHARACTER(LEN=*), OPTIONAL                         :: file_name
      INTEGER, OPTIONAL                                  :: conf_nr
      CHARACTER(LEN=*), OPTIONAL                         :: conf_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_element_in_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name_act, tmp_name
      CHARACTER(LEN=default_string_length)               :: header
      INTEGER                                            :: file_ptr, handle, i, nr_atoms
      LOGICAL                                            :: file_exists, print_it
      REAL(KIND=dp)                                      :: vol
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat_scaled

      file_name_act = ""
      tmp_name = ""
      header = ""
      print_it = .TRUE.

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_file_io.F",460)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_file_io.F",461)
      IF(.NOT.(ASSOCIATED(tmc_params%atoms)))CALL cp__a("tmc/tmc_file_io.F",462)
      IF(.NOT.(PRESENT(conf_nr) .OR. PRESENT(conf_info)))CALL cp__a("tmc/tmc_file_io.F",463)

      IF (print_it) THEN
         ! start the timing
         CALL timeset(routineN, handle)

         ! set default file name
         IF (PRESENT(file_name)) THEN
            IF(.NOT.(file_name .NE. ""))CALL cp__a("tmc/tmc_file_io.F",471)
            file_name_act = file_name
         ELSE
            IF(.NOT.(ASSOCIATED(tmc_params%Temp)))CALL cp__a("tmc/tmc_file_io.F",474)
            IF(.NOT.(PRESENT(temp_index)))CALL cp__a("tmc/tmc_file_io.F",475)
            file_name_act = expand_file_name_temp(file_name=tmc_default_trajectory_file_name, &
                                                  rvalue=tmc_params%Temp(temp_index))
         END IF

         nr_atoms = SIZE(elem%pos)/tmc_params%dim_per_elem

         ! set header (for coordinate or force file)
         IF (tmc_params%print_trajectory .OR. tmc_params%print_forces) THEN
            IF (PRESENT(conf_info)) THEN
               WRITE (header, *) TRIM(ADJUSTL(conf_info))
            ELSE
               !WRITE(header,FMT="(A,I8,A,F20.10)") " i = ", conf_nr,", E = ", elem%potential
               WRITE (header, FMT="(A,I8,A,F20.10,F20.10,A,I8,I8)") "i =", conf_nr, " ,E =", &
                  elem%potential, elem%ekin, " st elem", elem%sub_tree_nr, elem%nr
            END IF
         END IF

         ! write the coordinates
         IF (tmc_params%print_trajectory) THEN
            tmp_name = expand_file_name_ending(file_name_act, "xyz")
            CALL open_file(file_name=tmp_name, file_status="UNKNOWN", &
                           file_action="WRITE", file_position="APPEND", &
                           unit_number=file_ptr)
            WRITE (file_ptr, FMT="(I8)") nr_atoms
            WRITE (file_ptr, *) TRIM(header)
            DO i = 1, SIZE(elem%pos), tmc_params%dim_per_elem
               WRITE (file_ptr, FMT="(A4,1X,1000F20.10)") &
                  TRIM(tmc_params%atoms((i-1)/tmc_params%dim_per_elem+1)%name), &
                  elem%pos(i:i+tmc_params%dim_per_elem-1)*au2a
            END DO
            CALL close_file(unit_number=file_ptr)
         END IF

         ! write the forces
         IF (tmc_params%print_forces) THEN
            tmp_name = expand_file_name_ending(file_name_act, "frc")
            CALL open_file(file_name=tmp_name, file_status="UNKNOWN", &
                           file_action="WRITE", file_position="APPEND", &
                           unit_number=file_ptr)
            WRITE (file_ptr, FMT="(I8)") nr_atoms
            WRITE (file_ptr, *) TRIM(header)
            DO i = 1, SIZE(elem%pos), tmc_params%dim_per_elem
               WRITE (file_ptr, FMT="(A4,1X,1000F20.10)") &
                  TRIM(tmc_params%atoms((i-1)/tmc_params%dim_per_elem+1)%name), &
                  elem%frc(i:i+tmc_params%dim_per_elem-1)
            END DO
            CALL close_file(unit_number=file_ptr)
         END IF

         ! write the cell dipoles
         IF (tmc_params%print_dipole) THEN
            CALL write_dipoles_in_file(file_name=file_name_act, &
                                       conf_nr=conf_nr, dip=elem%dipole)
         END IF

         ! write the cell file
         IF (tmc_params%print_cell) THEN
            tmp_name = expand_file_name_ending(file_name_act, "cell")
            ! header
            INQUIRE (FILE=tmp_name, EXIST=file_exists) ! file_exists will be TRUE if the file exist
            IF (.NOT. file_exists) THEN
               CALL open_file(file_name=tmp_name, file_status="NEW", &
                              file_action="WRITE", unit_number=file_ptr)
               WRITE (file_ptr, FMT='(A,9(7X,A2," [Angstrom]"),6X,A)') &
                  "# MC step ", "Ax", "Ay", "Az", "Bx", "By", "Bz", "Cx", "Cy", "Cz", &
                  "Volume [Angstrom^3]"
            ELSE
               CALL open_file(file_name=tmp_name, file_status="OLD", &
                              file_action="WRITE", file_position="APPEND", &
                              unit_number=file_ptr)
            END IF
            CALL get_scaled_cell(cell=tmc_params%cell, &
                                 box_scale=elem%box_scale, scaled_hmat=hmat_scaled, &
                                 vol=vol)
            WRITE (file_ptr, FMT="(I8,9(1X,F19.10),1X,F24.10)") conf_nr, &
               hmat_scaled(:, :)*au2a, vol*au2a**3
            !TODO better cell output e.g. using cell_types routine
            CALL close_file(unit_number=file_ptr)
         END IF

         ! write the different energies
         IF (tmc_params%print_energies) THEN
            tmp_name = expand_file_name_ending(file_name_act, "ener")
            ! header
            INQUIRE (FILE=tmp_name, EXIST=file_exists) ! file_exists will be TRUE if the file exist
            IF (.NOT. file_exists) THEN
               CALL open_file(file_name=tmp_name, file_status="NEW", &
                              file_action="WRITE", unit_number=file_ptr)
               WRITE (file_ptr, FMT='(A,4A20)') &
                  "# MC step ", " exact ", " approx ", " last SCF ", " kinetic "
            ELSE
               CALL open_file(file_name=tmp_name, file_status="OLD", &
                              file_action="WRITE", file_position="APPEND", &
                              unit_number=file_ptr)
            END IF
            WRITE (file_ptr, FMT="(I8,14F20.10)") conf_nr, elem%potential, elem%e_pot_approx, &
               elem%scf_energies(MOD(elem%scf_energies_count, 4)+1), elem%ekin
            CALL close_file(unit_number=file_ptr)
         END IF

         ! end the timing
         CALL timestop(handle)
      END IF
   END SUBROUTINE write_element_in_file

! **************************************************************************************************
!> \brief writes the cell dipoles in dipole trajectory file
!> \param file_name ...
!> \param conf_nr ...
!> \param dip ...
!> \param file_ext ...
!> \param
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE write_dipoles_in_file(file_name, conf_nr, dip, file_ext)
      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: conf_nr
      REAL(KIND=dp), DIMENSION(:), POINTER               :: dip
      CHARACTER(LEN=*), INTENT(in), OPTIONAL             :: file_ext

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_dipoles_in_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name_tmp
      INTEGER                                            :: file_ptr
      LOGICAL                                            :: file_exists

      IF(.NOT.(ASSOCIATED(dip)))CALL cp__a("tmc/tmc_file_io.F",603)

      IF (PRESENT(file_ext)) THEN
         IF(.NOT.(file_ext .NE. ""))CALL cp__a("tmc/tmc_file_io.F",606)
         file_name_tmp = expand_file_name_ending(file_name, TRIM(file_ext))
      ELSE
         file_name_tmp = expand_file_name_ending(file_name, "dip")
      END IF
      INQUIRE (FILE=file_name_tmp, EXIST=file_exists)
      IF (.NOT. file_exists) THEN
         CALL open_file(file_name=file_name_tmp, file_status="NEW", &
                        file_action="WRITE", unit_number=file_ptr)
         WRITE (file_ptr, FMT='(A8,10A20)') "# conf_nr", "dip_x [C Angstrom]", &
            "dip_y [C Angstrom]", "dip_z [C Angstrom]"
      ELSE
         CALL open_file(file_name=file_name_tmp, file_status="OLD", &
                        file_action="WRITE", file_position="APPEND", &
                        unit_number=file_ptr)
      END IF
      WRITE (file_ptr, FMT="(I8,10F20.10)") conf_nr, dip(:)
      CALL close_file(unit_number=file_ptr)
   END SUBROUTINE write_dipoles_in_file

   !----------------------------------------------------------------------------
   ! read configuration from file
   !----------------------------------------------------------------------------

! **************************************************************************************************
!> \brief read the trajectory element from a file from sub tree element
!> \param elem actual tree element to be printed out
!> \param tmc_ana TMC analysis environment parameters
!> \param conf_nr Markov chain element number
!>        (input the old number and read only if conf nr from file is greater
!> \param stat ...
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE read_element_from_file(elem, tmc_ana, conf_nr, stat)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana
      INTEGER                                            :: conf_nr, stat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_element_from_file', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: conf_nr_old, handle, i_tmp
      LOGICAL                                            :: files_conf_missmatch

      stat = TMC_STATUS_OK
      conf_nr_old = conf_nr
      files_conf_missmatch = .FALSE.

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_file_io.F",654)
      IF(.NOT.(ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_file_io.F",655)
      IF(.NOT.(ASSOCIATED(tmc_ana%atoms)))CALL cp__a("tmc/tmc_file_io.F",656)

      ! start the timing
      CALL timeset(routineN, handle)

      ! read the coordinates
      IF (tmc_ana%id_traj .GT. 0) THEN
         i_tmp = conf_nr_old
         CALL read_pos_from_file(elem=elem, tmc_ana=tmc_ana, stat=stat, &
                                 conf_nr=i_tmp)
         IF (stat .EQ. TMC_STATUS_WAIT_FOR_NEW_TASK) THEN
            CALL cp_warn(cp__l("tmc/tmc_file_io.F",667), &
                         'end of position file reached at line '// &
                         cp_to_string(REAL(tmc_ana%lc_traj, KIND=dp))//", last element "// &
                         cp_to_string(tmc_ana%last_elem%nr))
         ELSE
            IF(.NOT.(i_tmp .GT. conf_nr_old))CALL cp__a("tmc/tmc_file_io.F",672)
            conf_nr = i_tmp
            elem%nr = i_tmp
         END IF
      END IF

      ! read the forces
      ! TODO if necessary

      ! read the dipoles file
      IF (tmc_ana%id_dip .GT. 0 .AND. stat .EQ. TMC_STATUS_OK) THEN
         i_tmp = conf_nr_old
         search_conf_dip: DO
            CALL read_dipole_from_file(elem=elem, tmc_ana=tmc_ana, stat=stat, &
                                       conf_nr=i_tmp)
            IF (stat .EQ. TMC_STATUS_WAIT_FOR_NEW_TASK) THEN
               CALL cp_warn(cp__l("tmc/tmc_file_io.F",688), &
                            'end of dipole file reached at line'// &
                            cp_to_string(REAL(tmc_ana%lc_dip, KIND=dp)))
               EXIT search_conf_dip
            END IF
            ! check consitence with pos file
            IF (tmc_ana%id_traj .GT. 0) THEN
               IF (i_tmp .EQ. conf_nr) THEN
                  files_conf_missmatch = .FALSE.
                  EXIT search_conf_dip
               ELSE
                  ! the configuration numbering differ from the position file,
                  !  but we keep on searching for the correct configuration
                  files_conf_missmatch = .TRUE.
               END IF
               ! if no pos file, just take the next conf
            ELSE IF (i_tmp .GT. conf_nr_old) THEN
               conf_nr = i_tmp
               elem%nr = i_tmp
               EXIT search_conf_dip
            END IF
         END DO search_conf_dip
      END IF

      ! read the cell file
      IF (tmc_ana%id_cell .GT. 0 .AND. stat .EQ. TMC_STATUS_OK) THEN
         search_conf_cell: DO
            CALL read_cell_from_file(elem=elem, tmc_ana=tmc_ana, stat=stat, &
                                     conf_nr=i_tmp)
            IF (stat .EQ. TMC_STATUS_WAIT_FOR_NEW_TASK) THEN
               CALL cp_warn(cp__l("tmc/tmc_file_io.F",718), &
                            'end of cell file reached at line at line'// &
                            cp_to_string(REAL(tmc_ana%lc_cell, KIND=dp)))
               EXIT search_conf_cell
            END IF
            ! check consitence with pos file
            IF (tmc_ana%id_traj .GT. 0) THEN
               IF (i_tmp .EQ. conf_nr) THEN
                  files_conf_missmatch = .FALSE.
                  EXIT search_conf_cell
               ELSE
                  ! the configuration numbering differ from the position file,
                  !  but we keep on searching for the correct configuration
                  files_conf_missmatch = .TRUE.
               END IF
               ! if no pos file, just take the next conf
            ELSE IF (i_tmp .GT. conf_nr_old) THEN
               conf_nr = i_tmp
               elem%nr = i_tmp
               EXIT search_conf_cell
            END IF
         END DO search_conf_cell

      END IF

      ! write the different energies
      ! TODO if necessary

      IF (files_conf_missmatch) &
         CALL cp_warn(cp__l("tmc/tmc_file_io.F",747), &
                      'there is a missmatch in the configuration numbering. '// &
                      "Read number of lines (pos|cell|dip)"// &
                      cp_to_string(tmc_ana%lc_traj)//"|"// &
                      cp_to_string(tmc_ana%lc_cell)//"|"// &
                      cp_to_string(tmc_ana%lc_dip))

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE read_element_from_file

! **************************************************************************************************
!> \brief search for the next configurational position in file
!> \param elem actual tree element to be read
!> \param tmc_ana ...
!> \param stat ...
!> \param conf_nr Markov chain element number
!>        (input the old number and read only if conf nr from file is greater
!> \param header_info ...
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE read_pos_from_file(elem, tmc_ana, stat, conf_nr, header_info)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana
      INTEGER                                            :: stat, conf_nr
      CHARACTER(LEN=*), OPTIONAL                         :: header_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_pos_from_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: c_tmp
      INTEGER                                            :: handle, i, i_tmp, status

      stat = TMC_STATUS_FAILED

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_file_io.F",782)
      IF(.NOT.(ASSOCIATED(elem%pos)))CALL cp__a("tmc/tmc_file_io.F",783)
      IF(.NOT.(ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_file_io.F",784)
      IF(.NOT.(tmc_ana%id_traj .GT. 0))CALL cp__a("tmc/tmc_file_io.F",785)

      ! start the timing
      CALL timeset(routineN, handle)

      search_next_conf: DO
         c_tmp(:) = " "
         tmc_ana%lc_traj = tmc_ana%lc_traj+1
         READ (tmc_ana%id_traj, '(A)', IOSTAT=status) c_tmp(:)
         IF (status .GT. 0) &
            CALL cp_abort(cp__l("tmc/tmc_file_io.F",795), &
                          "configuration header read error at line: "// &
                          cp_to_string(tmc_ana%lc_traj)//": "//c_tmp)
         IF (status .LT. 0) THEN ! end of file reached
            stat = TMC_STATUS_WAIT_FOR_NEW_TASK
            EXIT search_next_conf
         END IF
         IF (INDEX(c_tmp, "=") .GT. 0) THEN
            READ (c_tmp(INDEX(c_tmp, "=")+1:), *, IOSTAT=status) i_tmp ! read the configuration number
            IF (status .NE. 0) &
               CALL cp_abort(cp__l("tmc/tmc_file_io.F",805), &
                             "configuration header read error (for conf nr) at line: "// &
                             cp_to_string(tmc_ana%lc_traj))
            IF (i_tmp .GT. conf_nr) THEN
               ! TODO we could also read the energy ...
               conf_nr = i_tmp
               IF (PRESENT(header_info)) header_info = c_tmp
               stat = TMC_STATUS_OK
               EXIT search_next_conf
            END IF
         END IF
      END DO search_next_conf

      IF (stat .EQ. TMC_STATUS_OK) THEN
         pos_loop: DO i = 1, SIZE(elem%pos), tmc_ana%dim_per_elem
            tmc_ana%lc_traj = tmc_ana%lc_traj+1
            READ (tmc_ana%id_traj, FMT="(A4,1X,1000F20.10)", IOSTAT=status) &
               c_tmp, elem%pos(i:i+tmc_ana%dim_per_elem-1)
            IF (status .NE. 0) THEN
               CALL cp_abort(cp__l("tmc/tmc_file_io.F",824), &
                             "configuration pos read error at line: "// &
                             cp_to_string(tmc_ana%lc_traj))
            END IF
         END DO pos_loop
         elem%pos(:) = elem%pos(:)/au2a
      END IF

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE read_pos_from_file

! **************************************************************************************************
!> \brief search for the dipole entry
!> \param elem actual tree element to be read
!> \param tmc_ana ...
!> \param stat ...
!> \param conf_nr Markov chain element number
!>        (input the old number and read only if conf nr from file is greater
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE read_dipole_from_file(elem, tmc_ana, stat, conf_nr)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana
      INTEGER                                            :: stat, conf_nr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_dipole_from_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=250)                                 :: c_tmp
      INTEGER                                            :: handle, status

      stat = TMC_STATUS_FAILED

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_file_io.F",858)
      IF(.NOT.(ASSOCIATED(elem%dipole)))CALL cp__a("tmc/tmc_file_io.F",859)
      IF(.NOT.(ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_file_io.F",860)
      IF(.NOT.(tmc_ana%id_dip .GT. 0))CALL cp__a("tmc/tmc_file_io.F",861)

      ! start the timing
      CALL timeset(routineN, handle)
      tmc_ana%lc_dip = tmc_ana%lc_dip+1
      READ (tmc_ana%id_dip, FMT="(A)", IOSTAT=status) c_tmp
      IF (status .EQ. 0) THEN
         ! skip the initial line (header)
         IF (INDEX(c_tmp, "#") .GT. 0) THEN
            tmc_ana%lc_dip = tmc_ana%lc_dip+1
            READ (tmc_ana%id_dip, FMT="(A)", IOSTAT=status) c_tmp
         END IF
      END IF
      IF (status .EQ. 0) THEN
         READ (c_tmp, FMT="(I8,10F20.10)", IOSTAT=status) &
            conf_nr, elem%dipole(:)
      END IF
      IF (status .EQ. 0) THEN ! success
         stat = TMC_STATUS_OK
      ELSE IF (status .LT. 0) THEN ! end of file reached
         stat = TMC_STATUS_WAIT_FOR_NEW_TASK
      ELSE
         IF (status .NE. 0) &
            CALL cp__w("tmc/tmc_file_io.F",884,"configuration dipole read error at line: "//cp_to_string(tmc_ana%lc_dip))
         stat = TMC_STATUS_FAILED
      END IF

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE read_dipole_from_file

! **************************************************************************************************
!> \brief search for the cell entry
!> \param elem actual tree element to be read
!> \param tmc_ana ...
!> \param stat ...
!> \param conf_nr Markov chain element number
!>        (input the old number and read only if conf nr from file is greater
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE read_cell_from_file(elem, tmc_ana, stat, conf_nr)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana
      INTEGER                                            :: stat, conf_nr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_cell_from_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=250)                                 :: c_tmp
      INTEGER                                            :: handle, status
      REAL(KIND=dp)                                      :: r_tmp
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat

      stat = TMC_STATUS_FAILED

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_file_io.F",916)
      IF(.NOT.(ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_file_io.F",917)
      IF(.NOT.(ASSOCIATED(tmc_ana%cell)))CALL cp__a("tmc/tmc_file_io.F",918)
      IF(.NOT.(tmc_ana%id_cell .GT. 0))CALL cp__a("tmc/tmc_file_io.F",919)

      ! start the timing
      CALL timeset(routineN, handle)

      tmc_ana%lc_cell = tmc_ana%lc_cell+1
      READ (tmc_ana%id_cell, FMT="(A)", IOSTAT=status) c_tmp
      IF (status .EQ. 0) THEN
         ! skip the initial line (header)
         IF (INDEX(c_tmp, "#") .GT. 0) THEN
            tmc_ana%lc_cell = tmc_ana%lc_cell+1
            READ (tmc_ana%id_cell, FMT="(A)", IOSTAT=status) c_tmp
         END IF
      END IF
      IF (status .EQ. 0) THEN
         READ (c_tmp, FMT="(I8,9(1X,F19.10),1X,F24.10)", IOSTAT=status) conf_nr, &
            hmat(:, :), r_tmp
      END IF
      IF (status .LT. 0) THEN ! end of file reached
         stat = TMC_STATUS_WAIT_FOR_NEW_TASK
      ELSE IF (status .GT. 0) THEN
         IF (status .NE. 0) &
            CALL cp__b("tmc/tmc_file_io.F",941,"configuration cell read error at line: "//cp_to_string(tmc_ana%lc_cell))
         stat = TMC_STATUS_FAILED
      ELSE
         IF (elem%nr .LT. 0) elem%nr = conf_nr
         hmat(:, :) = hmat(:, :)/au2a
         ! get the box scaling
         CALL get_cell_scaling(cell=tmc_ana%cell, scaled_hmat=hmat, &
                               box_scale=elem%box_scale)
         stat = TMC_STATUS_OK
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE read_cell_from_file

   !----------------------------------------------------------------------------
   ! get the configurations from file and calc
   !----------------------------------------------------------------------------

! **************************************************************************************************
!> \brief opens the files for reading configurations data to analyze
!> \param tmc_ana ...
!> \param stat ...
!> \param dir_ind ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE analyse_files_open(tmc_ana, stat, dir_ind)
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana
      INTEGER                                            :: stat
      INTEGER, OPTIONAL                                  :: dir_ind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'analyse_files_open', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: dir_name, file_name_act, file_name_temp
      INTEGER                                            :: handle
      LOGICAL                                            :: file_exists

      IF(.NOT.(ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_file_io.F",979)

      stat = TMC_STATUS_WAIT_FOR_NEW_TASK

      ! start the timing
      CALL timeset(routineN, handle)

      IF (PRESENT(dir_ind)) THEN
         IF(.NOT.(ASSOCIATED(tmc_ana%dirs)))CALL cp__a("tmc/tmc_file_io.F",987)
         IF(.NOT.(dir_ind .GT. 0))CALL cp__a("tmc/tmc_file_io.F",988)
         IF(.NOT.(dir_ind .LE. SIZE(tmc_ana%dirs)))CALL cp__a("tmc/tmc_file_io.F",989)

         IF (INDEX(tmc_ana%dirs(dir_ind), "/", BACK=.TRUE.) .EQ. &
             LEN_TRIM(tmc_ana%dirs(dir_ind))) THEN
            dir_name = TRIM(tmc_ana%dirs(dir_ind))
         ELSE
            dir_name = TRIM(tmc_ana%dirs(dir_ind))//"/"
         END IF
      ELSE
         dir_name = "./"
      END IF

      ! open the files
      file_name_temp = expand_file_name_temp( &
                       file_name=tmc_default_trajectory_file_name, &
                       rvalue=tmc_ana%temperature)
      ! position file
      IF (tmc_ana%costum_pos_file_name .NE. "") THEN
         file_name_act = TRIM(dir_name)//tmc_ana%costum_pos_file_name
      ELSE
         file_name_act = TRIM(dir_name)// &
                         expand_file_name_ending(file_name_temp, "xyz")
      END IF
      INQUIRE (FILE=file_name_act, EXIST=file_exists)
      IF (file_exists) THEN
         CALL open_file(file_name=file_name_act, file_status="OLD", &
                        file_action="READ", unit_number=tmc_ana%id_traj)
         WRITE (tmc_ana%io_unit, FMT='(T2,A,"| ",A,T41,A40)') "TMC_ANA", &
            "read xyz file", TRIM(file_name_act)
      END IF

      ! cell file
      IF (tmc_ana%costum_cell_file_name .NE. "") THEN
         file_name_act = TRIM(dir_name)//tmc_ana%costum_cell_file_name
      ELSE
         file_name_act = TRIM(dir_name)// &
                         expand_file_name_ending(file_name_temp, "cell")
      END IF
      INQUIRE (FILE=file_name_act, EXIST=file_exists)
      IF (file_exists) THEN
         CALL open_file(file_name=file_name_act, file_status="OLD", &
                        file_action="READ", unit_number=tmc_ana%id_cell)
         WRITE (tmc_ana%io_unit, FMT='(T2,A,"| ",A,T41,A40)') "TMC_ANA", &
            "read cell file", TRIM(file_name_act)
      END IF

      ! dipole file
      IF (tmc_ana%costum_dip_file_name .NE. "") THEN
         file_name_act = TRIM(dir_name)//tmc_ana%costum_dip_file_name
      ELSE
         file_name_act = TRIM(dir_name)// &
                         expand_file_name_ending(file_name_temp, "dip")
      END IF
      INQUIRE (FILE=file_name_act, EXIST=file_exists)
      IF (file_exists) THEN
         CALL open_file(file_name=file_name_act, file_status="OLD", &
                        file_action="READ", unit_number=tmc_ana%id_dip)
         WRITE (tmc_ana%io_unit, FMT='(T2,A,"| ",A,T41,A40)') "TMC_ANA", &
            "read dip file", TRIM(file_name_act)
      END IF

      IF (tmc_ana%id_traj .GT. 0 .OR. tmc_ana%id_cell .GT. 0 .OR. &
          tmc_ana%id_dip .GT. 0) THEN
         stat = TMC_STATUS_OK
      ELSE
         CALL cp_warn(cp__l("tmc/tmc_file_io.F",1054), &
                      "There is no file to open for temperature "//cp_to_string(tmc_ana%temperature)// &
                      "K in directory "//TRIM(dir_name))
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE analyse_files_open

! **************************************************************************************************
!> \brief close the files for reading configurations data to analyze
!> \param tmc_ana ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE analyse_files_close(tmc_ana)
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana

      CHARACTER(LEN=*), PARAMETER :: routineN = 'analyse_files_close', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_file_io.F",1076)

      ! start the timing
      CALL timeset(routineN, handle)

      ! position file
      IF (tmc_ana%id_traj .GT. 0) CALL close_file(unit_number=tmc_ana%id_traj)

      ! cell file
      IF (tmc_ana%id_cell .GT. 0) CALL close_file(unit_number=tmc_ana%id_cell)

      ! dipole file
      IF (tmc_ana%id_dip .GT. 0) CALL close_file(unit_number=tmc_ana%id_dip)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE analyse_files_close

END MODULE tmc_file_io
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module for printing tree structures in GraphViz dot files
!>        for visualizing the trees
!> \par History
!>      12.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************
!----------------------------------------------------------------------!
! Tree Monte Carlo (TMC) a program for parallel Monte Carlo simulation
! \author Mandes Schoenherr
!----------------------------------------------------------------------!
MODULE tmc_dot_tree
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_to_string
   USE tmc_file_io,                     ONLY: expand_file_name_char,&
                                              expand_file_name_temp
   USE tmc_move_types,                  ONLY: mv_type_swap_conf
   USE tmc_tree_types,                  ONLY: &
        global_tree_type, gt_elem_list_type, status_accepted, status_accepted_result, &
        status_calc_approx_ener, status_calculate_MD, status_calculate_NMC_steps, &
        status_calculate_energy, status_calculated, status_cancel_ener, status_cancel_nmc, &
        status_canceled_ener, status_canceled_nmc, status_created, status_deleted, &
        status_deleted_result, status_rejected, status_rejected_result, tree_type
   USE tmc_types,                       ONLY: tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_dot_tree'

   PUBLIC :: init_draw_trees, finalize_draw_tree
   PUBLIC :: create_dot_color, create_global_tree_dot_color
   PUBLIC :: create_dot, create_global_tree_dot

   INTEGER :: DEBUG = 0
!  CHARACTER(LEN=30) :: filename ="tree.dot"

CONTAINS
! **************************************************************************************************
!> \brief returns extended filename for global and sub trees
!> \param tmc_params param environment for creating the file name
!> \param ind index of the subtree (0 = global tree)
!> \retval filename ...
!> \author Mandes 12.2012
! **************************************************************************************************
   FUNCTION get_dot_file_name(tmc_params, ind) RESULT(filename)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      INTEGER                                            :: ind
      CHARACTER(LEN=50)                                  :: filename

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_dot_file_name', &
         routineP = moduleN//':'//routineN

      filename = ""

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_dot_tree.F",64)
      IF(.NOT.(ind .GE. 0))CALL cp__a("tmc/tmc_dot_tree.F",65)
      IF(.NOT.(ASSOCIATED(tmc_params%Temp)))CALL cp__a("tmc/tmc_dot_tree.F",66)
      IF(.NOT.(ind .LE. SIZE(tmc_params%Temp)))CALL cp__a("tmc/tmc_dot_tree.F",67)

      IF (ind .EQ. 0) THEN
         filename = TRIM(expand_file_name_char(tmc_params%dot_file_name, "global"))
      ELSE
         filename = TRIM(expand_file_name_temp(file_name=tmc_params%dot_file_name, &
                                               rvalue=tmc_params%Temp(ind)))
      END IF

      IF(.NOT.(filename .NE. ""))CALL cp__a("tmc/tmc_dot_tree.F",76)
   END FUNCTION get_dot_file_name
! **************************************************************************************************
!> \brief initializes the dot files (open and write headers)
!> \param tmc_params param environment for creating the file name
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE init_draw_trees(tmc_params)
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_draw_trees', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: file_ptr, i

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_dot_tree.F",91)

      ! global tree
      CALL open_file(file_name=get_dot_file_name(tmc_params, 0), file_status="REPLACE", &
                     file_action="WRITE", unit_number=file_ptr)
      WRITE (file_ptr, *) "digraph G {"
      WRITE (file_ptr, *) '  size="8.27,11.69"'
      CALL write_legend(file_ptr)
      CALL close_file(unit_number=file_ptr, keep_preconnection=.TRUE.)

      ! subtrees
      DO i = 1, SIZE(tmc_params%Temp)
         CALL open_file(file_name=get_dot_file_name(tmc_params, i), file_status="REPLACE", &
                        file_action="WRITE", unit_number=file_ptr)
         WRITE (file_ptr, *) "digraph G {"
         WRITE (file_ptr, *) '  size="8.27,11.69"'
         CALL write_legend(file_ptr)
         CALL close_file(unit_number=file_ptr, keep_preconnection=.TRUE.)
      END DO
   END SUBROUTINE init_draw_trees

! **************************************************************************************************
!> \brief close the dot files (write tails)
!> \param tmc_params param environment for creating the file name
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE finalize_draw_tree(tmc_params)
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'finalize_draw_tree', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: file_ptr, i

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_dot_tree.F",125)

      ! global tree
      CALL open_file(file_name=get_dot_file_name(tmc_params, 0), &
                     file_status="OLD", file_action="WRITE", &
                     file_position="APPEND", unit_number=file_ptr)
      WRITE (file_ptr, *) "}"
      CALL close_file(unit_number=file_ptr)

      DO i = 1, SIZE(tmc_params%Temp)
         CALL open_file(file_name=get_dot_file_name(tmc_params, i), file_status="OLD", &
                        file_action="WRITE", file_position="APPEND", unit_number=file_ptr)
         WRITE (file_ptr, *) "}"
         CALL close_file(unit_number=file_ptr)
      END DO
   END SUBROUTINE finalize_draw_tree

! **************************************************************************************************
!> \brief writes the legend in the file
!> \param file_ptr file pointer
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE write_legend(file_ptr)
      INTEGER, INTENT(IN)                                :: file_ptr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_legend', routineP = moduleN//':'//routineN

      IF(.NOT.(file_ptr .GT. 0))CALL cp__a("tmc/tmc_dot_tree.F",152)

      WRITE (file_ptr, *) '//LEGEND'
      WRITE (file_ptr, *) 'subgraph clusterLegend {'
      WRITE (file_ptr, *) '  label="Legend:" labelloc=t fontsize=30'
      WRITE (file_ptr, *) '  centered=false'
      WRITE (file_ptr, *) '  color=black'
      WRITE (file_ptr, *) '  leg1 -> leg2 -> leg2_2 -> leg2_3 -> leg2_4 -> leg3 -> '// &
         'leg4 -> leg5 -> leg6 -> leg7_1 -> leg7 -> '// &
         'leg8_1 -> leg8 -> leg9 -> leg10 [style=invis]'
      WRITE (file_ptr, *) '  {rank=same leg1 [fontsize=30, label="node created"          , color=black]}'
      WRITE (file_ptr, *) '  {rank=same leg2 [fontsize=30, label="configuration created" , style=filled,    color=gray]}'
      WRITE (file_ptr, *) '  {rank=same leg2_2 [fontsize=30, label="calc energy" , style=filled,    color=brown]}'
      WRITE (file_ptr, *) '  {rank=same leg2_2 [fontsize=30, label="calc energy" , style=filled,    color=wheat]}'
      WRITE (file_ptr, *) '  {rank=same leg2_3 [fontsize=30, label="calc HMC" , style=filled,    color=goldenrod]}'
      WRITE (file_ptr, *) '  {rank=same leg2_4 [fontsize=30, label="calc NMC" , style=filled,    color=peru]}'
      WRITE (file_ptr, *) '  {rank=same leg3 [fontsize=30, label="accepted"              , color=greenyellow]}'
      WRITE (file_ptr, *) '  {rank=same leg4 [fontsize=30, label="rejected"              , color=red]}'
      WRITE (file_ptr, *) '  {rank=same leg5 [fontsize=30, label="trajec"                , '// &
         'style=filled,    color=gold, shape=polygon, sides=4]}'
      WRITE (file_ptr, *) '  {rank=same leg6 [fontsize=30, label="energy calculated"     , '// &
         'style=filled,    color=blue, fontcolor=white]}'
      WRITE (file_ptr, *) '  {rank=same leg7_1 [fontsize=30, label="cancel NMC send"     , '// &
         'style=filled,    color=deeppink, fontcolor=white]}'
      WRITE (file_ptr, *) '  {rank=same leg7 [fontsize=30, label="canceled NMC"          , '// &
         'style=filled,    color=darkorchid1, fontcolor=white]}'
      WRITE (file_ptr, *) '  {rank=same leg8_1 [fontsize=30, label="cancel ENERGY send"    , '// &
         'style=filled,    color=cornflowerblue]}'
      WRITE (file_ptr, *) '  {rank=same leg8 [fontsize=30, label="canceled ENERGY"       , '// &
         'style=filled,    color=cyan]}'
      WRITE (file_ptr, *) '  {rank=same leg9 [fontsize=30, label="deleted"               , '// &
         'style=filled,    shape=polygon, sides=3, color=black,fontcolor=white]}'
      WRITE (file_ptr, *) '  {rank=same leg10 [fontsize=30, label="deleted trajectory"   , '// &
         'style=filled,    shape=polygon, sides=5, color=gold]}'
      WRITE (file_ptr, *) ' }'
   END SUBROUTINE write_legend

! **************************************************************************************************
!> \brief write/change color related to certain tree element status
!> \param node_nr the index of the tree node
!> \param stat tree element status
!> \param filename the filename for the grapgviz dot files
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE write_color(node_nr, stat, filename)
      INTEGER                                            :: node_nr, stat
      CHARACTER(LEN=50)                                  :: filename

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_color', routineP = moduleN//':'//routineN

      CHARACTER(len=11)                                  :: label
      INTEGER                                            :: file_ptr

      IF(.NOT.(filename .NE. ""))CALL cp__a("tmc/tmc_dot_tree.F",205)
      IF(.NOT.(node_nr .GE. 0))CALL cp__a("tmc/tmc_dot_tree.F",206)

      CALL open_file(file_name=filename, file_status="OLD", &
                     file_action="WRITE", file_position="APPEND", unit_number=file_ptr)
      WRITE (label, FMT='(I10,A)') node_nr, "["
      SELECT CASE (stat)
      CASE (status_created)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=gray]'
      CASE (status_accepted)
         WRITE (file_ptr, *) TRIM(label), 'color=green]'
      CASE (status_rejected)
         WRITE (file_ptr, *) TRIM(label), 'color=red]'
      CASE (status_accepted_result)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=green, shape=polygon, sides=4]'
      CASE (status_rejected_result)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=red, shape=polygon, sides=4]'
      CASE (status_calculated)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=blue]'
      CASE (status_cancel_nmc)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=deeppink]'
      CASE (status_cancel_ener)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=cornflowerblue]'
      CASE (status_canceled_nmc)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=darkorchid1]'
      CASE (status_canceled_ener)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=cyan]'
      CASE (status_deleted)
         WRITE (file_ptr, *) TRIM(label), 'shape=polygon, sides=3]'
      CASE (status_deleted_result)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, shape=polygon, sides=5]'
      CASE (status_calc_approx_ener)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=brown]'
      CASE (status_calculate_energy)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=wheat]'
      CASE (status_calculate_MD)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=goldenrod]'
      CASE (status_calculate_NMC_steps)
         WRITE (file_ptr, *) TRIM(label), 'style=filled, color=peru]'
      CASE DEFAULT
         CALL cp__b("tmc/tmc_dot_tree.F",245,"element status"//cp_to_string(stat))
      END SELECT
      CALL close_file(unit_number=file_ptr, keep_preconnection=.TRUE.)
   END SUBROUTINE write_color

! **************************************************************************************************
!> \brief creates an new branch (hence a new element is created)
!> \param parent_nr tree element number of element one level up
!> \param child_nr tree element number of actual element
!> \param acc flag for accepted or not accepted branch (left,right)
!> \param tmc_params param environment for creating the file name
!> \param tree index of the tree (0=global tree)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_dot_branch(parent_nr, child_nr, acc, tmc_params, tree)
      INTEGER                                            :: parent_nr, child_nr
      LOGICAL                                            :: acc
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      INTEGER                                            :: tree

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_dot_branch', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: file_ptr

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_dot_tree.F",270)

      CALL open_file(file_name=get_dot_file_name(tmc_params, tree), &
                     file_status="OLD", file_action="WRITE", &
                     file_position="APPEND", unit_number=file_ptr)
      IF (acc) THEN
         WRITE (file_ptr, *) parent_nr, " -> ", child_nr, ":nw [color=darkolivegreen1]"
      ELSE
         WRITE (file_ptr, *) parent_nr, " -> ", child_nr, ":ne [color=coral]"
      END IF
      CALL close_file(unit_number=file_ptr, keep_preconnection=.TRUE.)
   END SUBROUTINE create_dot_branch

! **************************************************************************************************
!> \brief interfaces the creating of a branch for subtree elements
!> \param new_element the actual subtree element
!> \param conf the subtree index and hence the index for filename
!> \param tmc_params ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_dot(new_element, conf, tmc_params)
      TYPE(tree_type), POINTER                           :: new_element
      INTEGER                                            :: conf
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_dot', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(new_element)))CALL cp__a("tmc/tmc_dot_tree.F",297)
      IF(.NOT.(conf .GT. 0))CALL cp__a("tmc/tmc_dot_tree.F",298)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_dot_tree.F",299)

      CALL create_dot_branch(parent_nr=new_element%parent%nr, &
                             child_nr=new_element%nr, &
                             acc=ASSOCIATED(new_element%parent%acc, new_element), &
                             tmc_params=tmc_params, tree=conf)
   END SUBROUTINE create_dot

! **************************************************************************************************
!> \brief creates new dot and arrow from element one level up (for subtree)
!>        additional handling of nodes with swaped elements
!> \param new_element the actual global element
!> \param tmc_params ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_global_tree_dot(new_element, tmc_params)
      TYPE(global_tree_type), POINTER                    :: new_element
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_global_tree_dot', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=1000)                                :: list_of_nr
      INTEGER                                            :: file_ptr, i, ref_count
      TYPE(gt_elem_list_type), POINTER                   :: tmp_pt_list_elem

      NULLIFY (tmp_pt_list_elem)

      IF(.NOT.(ASSOCIATED(new_element)))CALL cp__a("tmc/tmc_dot_tree.F",327)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_dot_tree.F",328)

      ! creating list with configuration numbers (of subtrees)
      list_of_nr = ""
      ! the order of subtrees
      DO i = 1, SIZE(new_element%conf(:))
         WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), new_element%conf(i)%elem%sub_tree_nr
      END DO
      ! the used subtree elements
      WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), '\n '
      DO i = 1, SIZE(new_element%conf(:))
         WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), " ", new_element%conf(i)%elem%nr
      END DO
      ! print out the references of each subtree element
      IF (DEBUG .GT. 8) THEN
         WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), '\n ref'
         DO i = 1, SIZE(new_element%conf(:))
            ref_count = 0
            tmp_pt_list_elem => new_element%conf(i)%elem%gt_nodes_references
            DO WHILE (ASSOCIATED(tmp_pt_list_elem))
               ref_count = ref_count+1
               ! create a list with all references
               IF (.FALSE.) WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), " ", tmp_pt_list_elem%gt_elem%nr
               tmp_pt_list_elem => tmp_pt_list_elem%next
            END DO
            ! print a list with all references
            IF (.FALSE.) WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), ' | '
            ! print only the amount of references
            IF (.TRUE.) WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), ref_count, ' | '
         END DO
      END IF

      IF (.NOT. ASSOCIATED(new_element%parent)) THEN
         IF (new_element%nr .GT. 1) &
            CALL cp_warn(cp__l("tmc/tmc_dot_tree.F",362), &
                         "try to create dot, but no parent on node "// &
                         cp_to_string(new_element%nr)//"exists")
      ELSE
         CALL create_dot_branch(parent_nr=new_element%parent%nr, &
                                child_nr=new_element%nr, &
                                acc=ASSOCIATED(new_element%parent%acc, new_element), &
                                tmc_params=tmc_params, tree=0)
      END IF
      ! write in dot file
      CALL open_file(file_name=get_dot_file_name(tmc_params, 0), &
                     file_status="OLD", file_action="WRITE", &
                     file_position="APPEND", unit_number=file_ptr)
      IF (new_element%swaped) THEN
         WRITE (file_ptr, *) new_element%nr, '[label="', new_element%nr, ' |', new_element%mv_conf, ' |', &
            mv_type_swap_conf, '\n ', &
            TRIM(ADJUSTL(list_of_nr)), '", shape=polygon, peripheries=3, sides=5]'
      ELSE
         WRITE (file_ptr, *) new_element%nr, '[label="', new_element%nr, ' |', new_element%mv_conf, ' |', &
            new_element%conf(new_element%mv_conf)%elem%move_type, '\n ', &
            TRIM(ADJUSTL(list_of_nr)), '"]'
      END IF
      CALL close_file(file_ptr, keep_preconnection=.TRUE.)
   END SUBROUTINE create_global_tree_dot

! **************************************************************************************************
!> \brief interfaces the change of color for subtree elements
!>        on the basis of the element status
!> \param tree_element the actual global element
!> \param tmc_params ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_dot_color(tree_element, tmc_params)
      TYPE(tree_type), POINTER                           :: tree_element
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_dot_color', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=1000)                                :: list_of_nr
      INTEGER                                            :: ref_count
      TYPE(gt_elem_list_type), POINTER                   :: tmp_pt_list_elem

      IF(.NOT.(ASSOCIATED(tree_element)))CALL cp__a("tmc/tmc_dot_tree.F",405)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_dot_tree.F",406)

      IF (DEBUG .GT. 8) THEN
         list_of_nr = ""
         tmp_pt_list_elem => tree_element%gt_nodes_references
         ref_count = 0
         DO WHILE (ASSOCIATED(tmp_pt_list_elem))
            ref_count = ref_count+1
            ! print a list with all references
            IF (.FALSE.) THEN
               WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), " ", tmp_pt_list_elem%gt_elem%nr
               WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), ' | '
            END IF
            ! print only the amount of references
            IF (.TRUE.) WRITE (list_of_nr, *) ref_count, ' | '
            tmp_pt_list_elem => tmp_pt_list_elem%next
         END DO
         WRITE (*, *) "mark subtree", tree_element%sub_tree_nr, " node", tree_element%nr, " with status ", &
            tree_element%stat, "ref ", TRIM(ADJUSTL(list_of_nr))
      END IF

      CALL write_color(node_nr=tree_element%nr, stat=tree_element%stat, &
                       filename=get_dot_file_name(tmc_params, tree_element%sub_tree_nr))
   END SUBROUTINE create_dot_color

! **************************************************************************************************
!> \brief interfaces the change of color for global tree  node
!>        on the basis of the element status
!> \param gt_tree_element the actual global element
!> \param tmc_params ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_global_tree_dot_color(gt_tree_element, tmc_params)
      TYPE(global_tree_type), POINTER                    :: gt_tree_element
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_global_tree_dot_color', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(gt_tree_element)))CALL cp__a("tmc/tmc_dot_tree.F",445)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_dot_tree.F",446)

      IF (DEBUG .GT. 8) WRITE (*, *) "mark global tree node color", gt_tree_element%nr, gt_tree_element%stat
      CALL write_color(node_nr=gt_tree_element%nr, stat=gt_tree_element%stat, &
                       filename=get_dot_file_name(tmc_params, 0))
   END SUBROUTINE create_global_tree_dot_color

!! **************************************************************************************************
!!> \brief prints out dot file for a whole subtree below the entered element
!!> \param current the actual subtree element
!!> \param conf index of the subtree
!!> \param error variable to control error logging, stopping,...
!!>        see module cp_error_handling
!!> \author Mandes 12.2012
!! **************************************************************************************************
!  RECURSIVE SUBROUTINE create_tree(current, conf, filename)
!    TYPE (tree_type), POINTER                :: current
!    INTEGER                                  :: conf
!    CHARACTER(LEN=*)                         :: filename
!
!    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_tree', &
!      routineP = moduleN//':'//routineN
!
!    CALL create_dot_color(current, tmc_params)
!    IF(ASSOCIATED(current%acc))THEN
!       CALL create_dot_branch(parent_nr=current%nr, child_nr=current%acc%nr, &
!                              acc=.TRUE.,tmc_params=tmc_params, file_single_tree_ptr)
!       WRITE(file_single_tree_ptr,*)current%nr,'[label="', current%nr,"\n ",&
!                                    current%pos(1),"\n ", current%potential,'"]'
!       CALL create_tree(current%acc, conf)
!    END IF
!    IF(ASSOCIATED(current%nacc))THEN
!       CALL create_dot_branch(current%nr,current%acc%nr,.FALSE.,file_single_tree_ptr)
!       WRITE(file_single_tree_ptr,*)current%nr,'[label="', current%nr,"\n ",&
!                                    current%pos(1),"\n ", current%potential,'"]'
!       CALL create_tree(current%nacc, conf)
!    END IF
!  END SUBROUTINE create_tree
END MODULE tmc_dot_tree
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief - to decrease the used memory size, just actual needed tree elements
!>        should be stored in memory, other ones should be written out in file
!>        - sub tree elements can be canceled and further deallocated when no
!>          global tree element refers to it anymore
!>        - then also the ongoing calculation of these elements is not needed
!>          anymore => can be canceled
!>        - MODULE: creates and hande a list of tree nodes
!>                    wich can be canceled
!>                  these elements are collected and canceled all in one
!>                    from the master routine
!>        - the actual cancelation routine is implemented in master module and
!>          communication is done using the message module
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_cancelation
   USE cp_log_handling,                 ONLY: cp_to_string
   USE tmc_dot_tree,                    ONLY: create_dot_color
   USE tmc_tree_types,                  ONLY: &
        add_to_list, elem_list_type, status_accepted, status_accepted_result, &
        status_calc_approx_ener, status_calculate_MD, status_calculate_NMC_steps, &
        status_calculate_energy, status_calculated, status_cancel_ener, status_cancel_nmc, &
        status_canceled_ener, status_canceled_nmc, status_created, status_deleted, &
        status_deleted_result, status_rejected, status_rejected_result, tree_type
   USE tmc_types,                       ONLY: tmc_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_cancelation'

   PUBLIC :: add_to_canceling_list, free_cancelation_list

CONTAINS

! **************************************************************************************************
!> \brief add a certain element to the cancelation list
!> \param elem the sub tree element, to be added
!> \param tmc_env tmc environment
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE add_to_canceling_list(elem, tmc_env)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'add_to_canceling_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: need_to_cancel

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_cancelation.F",62)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_cancelation.F",63)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_cancelation.F",64)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_cancelation.F",65)

      ! start the timing
      CALL timeset(routineN, handle)

      IF (tmc_env%params%SPECULATIVE_CANCELING) THEN
         need_to_cancel = .FALSE.
         ! update status
         SELECT CASE (elem%stat)
         CASE (status_calculate_energy)
            elem%stat = status_cancel_ener
            need_to_cancel = .TRUE.
            tmc_env%m_env%count_cancel_ener = tmc_env%m_env%count_cancel_ener+1
         CASE (status_calc_approx_ener) !TODO maybe elem status for approx ener cancel
            !elem%stat = status_cancel_ener
            !need_to_cancel = .TRUE.
         CASE (status_calculate_NMC_steps, status_calculate_MD)
            elem%stat = status_cancel_nmc
            need_to_cancel = .TRUE.
            tmc_env%m_env%count_cancel_NMC = tmc_env%m_env%count_cancel_NMC+1
         CASE (status_accepted, status_accepted_result, status_rejected, &
               status_rejected_result, status_calculated, status_created, &
               status_cancel_nmc, status_cancel_ener, status_canceled_nmc, &
               status_canceled_ener)
         CASE (status_deleted_result, status_deleted)
            ! if deallocation is deactivated, should not be
            CALL cp__w("tmc/tmc_cancelation.F",91,"try to add deleted element cancelation list ")
            WRITE (*, *) "WARNING: try to cancel subtree, element ", elem%sub_tree_nr, elem%nr, ", with status ", elem%stat
         CASE DEFAULT
            CALL cp_abort(cp__l("tmc/tmc_cancelation.F",94), &
                          "try to add element with unknown status to cancelation list (stat=" &
                          //cp_to_string(elem%stat))
         END SELECT
         ! set dot color
         IF (tmc_env%params%DRAW_TREE) &
            CALL create_dot_color(tree_element=elem, tmc_params=tmc_env%params)

         ! add to list
         IF (need_to_cancel) THEN
            CALL add_to_list(elem=elem, list=tmc_env%m_env%cancelation_list)
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE add_to_canceling_list

! **************************************************************************************************
!> \brief for correct finalizing deallocate the cancelation list
!> \param cancel_list ...
!> \param
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE free_cancelation_list(cancel_list)
      TYPE(elem_list_type), POINTER                      :: cancel_list

      CHARACTER(LEN=*), PARAMETER :: routineN = 'free_cancelation_list', &
         routineP = moduleN//':'//routineN

      TYPE(elem_list_type), POINTER                      :: tmp_element

      cancel_elem_loop: DO WHILE (ASSOCIATED(cancel_list))
         tmp_element => cancel_list%next
         DEALLOCATE (cancel_list)
         cancel_list => tmp_element
      END DO cancel_elem_loop
   END SUBROUTINE free_cancelation_list

END MODULE tmc_cancelation
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief global tree references
!>        - BECAUSE acceptance check use global tree randon numbers and
!>            (in case of parallel tempering) several global tree node refer to a
!>            single sub tree node (which is the changed one in the global tree)
!>        - the references are used to update the global tree acceptance probability
!>            for every global tree element seperately
!>        Hence a list of all global tree nodes, using the related subtree node,
!>            is created.
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_tree_references
   USE cp_log_handling,                 ONLY: cp_to_string
   USE tmc_cancelation,                 ONLY: add_to_canceling_list
   USE tmc_tree_types,                  ONLY: global_tree_type,&
                                              gt_elem_list_type,&
                                              tree_type
   USE tmc_types,                       ONLY: tmc_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_tree_references'

   PUBLIC :: add_to_references
   PUBLIC :: search_and_remove_reference_in_list
   PUBLIC :: remove_subtree_element_of_all_references
   PUBLIC :: remove_gt_references
CONTAINS

! **************************************************************************************************
!> \brief adds global tree reference to the modified sub tree element(s)
!> \param gt_elem actual global tree element
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE add_to_references(gt_elem)
      TYPE(global_tree_type), POINTER                    :: gt_elem

      CHARACTER(LEN=*), PARAMETER :: routineN = 'add_to_references', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(gt_elem_list_type), POINTER                   :: tmp_pt_list_elem

      NULLIFY (tmp_pt_list_elem)

      IF(.NOT.(ASSOCIATED(gt_elem)))CALL cp__a("tmc/tmc_tree_references.F",57)

      ! start the timing
      CALL timeset(routineN, handle)

      ! create reference and add at the beginning of the list
      ALLOCATE (tmp_pt_list_elem)
      tmp_pt_list_elem%gt_elem => gt_elem
      IF (ASSOCIATED(gt_elem%conf(gt_elem%mv_conf)%elem%gt_nodes_references)) THEN
         tmp_pt_list_elem%next => gt_elem%conf(gt_elem%mv_conf)%elem%gt_nodes_references
      ELSE
         tmp_pt_list_elem%next => NULL()
      END IF
      gt_elem%conf(gt_elem%mv_conf)%elem%gt_nodes_references => tmp_pt_list_elem

      ! in case of swapped configurations both are necessary to do acceptance probability update
      !   also when second configuration returns a value
      IF (gt_elem%swaped) THEN
         ! add reference to swapped elem
         ALLOCATE (tmp_pt_list_elem)
         tmp_pt_list_elem%gt_elem => gt_elem
         IF (ASSOCIATED(gt_elem%conf(gt_elem%mv_conf+1)%elem%gt_nodes_references)) THEN
            tmp_pt_list_elem%next => gt_elem%conf(gt_elem%mv_conf+1)%elem%gt_nodes_references
         ELSE
            tmp_pt_list_elem%next => NULL()
         END IF
         gt_elem%conf(gt_elem%mv_conf+1)%elem%gt_nodes_references => tmp_pt_list_elem
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE add_to_references

! **************************************************************************************************
!> \brief removes the global tree references of this actual global tree element
!>        from all related sub tree elements
!> \param gt_ptr actual global tree element
!> \param tmc_env ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE remove_gt_references(gt_ptr, tmc_env)
      TYPE(global_tree_type), POINTER                    :: gt_ptr
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_gt_references', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(gt_ptr)))CALL cp__a("tmc/tmc_tree_references.F",105)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_references.F",106)

      ! start the timing
      CALL timeset(routineN, handle)

      CALL search_and_remove_reference_in_list(gt_ptr=gt_ptr, &
                                               elem=gt_ptr%conf(gt_ptr%mv_conf)%elem, tmc_env=tmc_env)

      ! in case of parallel tempering also the reference in the second swaped configuration has to be removed
      IF (gt_ptr%swaped) THEN
         CALL search_and_remove_reference_in_list(gt_ptr=gt_ptr, &
                                                  elem=gt_ptr%conf(gt_ptr%mv_conf+1)%elem, tmc_env=tmc_env)
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE remove_gt_references

! **************************************************************************************************
!> \brief removes the pointers to a certain subtree element from every related
!>        global tree element
!> \param ptr sub tree element
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE remove_subtree_element_of_all_references(ptr)
      TYPE(tree_type), POINTER                           :: ptr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_subtree_element_of_all_references', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2000)                                :: list_of_nr
      INTEGER                                            :: handle, i
      TYPE(gt_elem_list_type), POINTER                   :: tmp_gt_list_ptr

      NULLIFY (tmp_gt_list_ptr)

      IF(.NOT.(ASSOCIATED(ptr)))CALL cp__a("tmc/tmc_tree_references.F",141)

      ! start the timing
      CALL timeset(routineN, handle)

      pt_node_ref_loop: DO WHILE (ASSOCIATED(ptr%gt_nodes_references))
         tmp_gt_list_ptr => ptr%gt_nodes_references
         IF(.NOT.(ASSOCIATED(tmp_gt_list_ptr%gt_elem)))CALL cp__a("tmc/tmc_tree_references.F",148)
         CALL cp_abort(cp__l("tmc/tmc_tree_references.F",149), &
                       "found reference of global tree node "// &
                       cp_to_string(tmp_gt_list_ptr%gt_elem%nr)// &
                       ", while removing sub tree node "// &
                       cp_to_string(ptr%sub_tree_nr)//cp_to_string(ptr%nr))
         ! check if configurations exist
         IF (ASSOCIATED(tmp_gt_list_ptr%gt_elem%conf(tmp_gt_list_ptr%gt_elem%mv_conf)%elem)) THEN
            IF (ASSOCIATED(ptr, tmp_gt_list_ptr%gt_elem%conf(tmp_gt_list_ptr%gt_elem%mv_conf)%elem)) THEN
               tmp_gt_list_ptr%gt_elem%conf(tmp_gt_list_ptr%gt_elem%mv_conf)%elem => NULL()
               ! in case of swapping the second configuration could be the related one
            ELSE IF (ASSOCIATED(ptr, tmp_gt_list_ptr%gt_elem%conf(tmp_gt_list_ptr%gt_elem%mv_conf+1)%elem)) THEN
               tmp_gt_list_ptr%gt_elem%conf(tmp_gt_list_ptr%gt_elem%mv_conf+1)%elem => NULL()
            ELSE
               list_of_nr = ""
               DO i = 1, SIZE(tmp_gt_list_ptr%gt_elem%conf)
                  WRITE (list_of_nr, *) TRIM(ADJUSTL(list_of_nr)), tmp_gt_list_ptr%gt_elem%conf(i)%elem%sub_tree_nr, &
                     tmp_gt_list_ptr%gt_elem%conf(i)%elem%nr, " | "
               END DO
               CALL cp_warn(cp__l("tmc/tmc_tree_references.F",167), &
                            "for subtree "// &
                            cp_to_string(ptr%sub_tree_nr)// &
                            "element "//cp_to_string(ptr%nr)// &
                            "global tree element"//cp_to_string(tmp_gt_list_ptr%gt_elem%nr)// &
                            "swaped"//cp_to_string(tmp_gt_list_ptr%gt_elem%swaped)// &
                            "moved elem"//cp_to_string(tmp_gt_list_ptr%gt_elem%mv_conf)// &
                            "with the related subtree, elements: "// &
                            TRIM(ADJUSTL(list_of_nr)))
            END IF
         ELSE
            CALL cp_warn(cp__l("tmc/tmc_tree_references.F",178), &
                         "for subtree "//cp_to_string(ptr%sub_tree_nr)// &
                         "element "//cp_to_string(ptr%nr)// &
                         " is not related to global tree node "//cp_to_string(tmp_gt_list_ptr%gt_elem%nr)// &
                         "(anymore).")
         END IF
         ptr%gt_nodes_references => ptr%gt_nodes_references%next
         DEALLOCATE (tmp_gt_list_ptr)
      END DO pt_node_ref_loop

      ! end the timing
      CALL timestop(handle)

      IF(.NOT.(.NOT. ASSOCIATED(ptr%gt_nodes_references)))CALL cp__a("tmc/tmc_tree_references.F",191)
   END SUBROUTINE remove_subtree_element_of_all_references

! **************************************************************************************************
!> \brief removes the global tree references of this actual global tree element
!>        from all related sub tree elements
!> \param gt_ptr actual global tree element
!> \param elem ...
!> \param tmc_env TMC environment
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE search_and_remove_reference_in_list(gt_ptr, elem, tmc_env)
      TYPE(global_tree_type), POINTER                    :: gt_ptr
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_and_remove_reference_in_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(gt_elem_list_type), POINTER                   :: tmp_gt_list_last_ptr, tmp_gt_list_ptr

      NULLIFY (tmp_gt_list_ptr, tmp_gt_list_last_ptr)

      ! nothing to do, when subtree element is already deleted
      IF (.NOT. ASSOCIATED(elem)) RETURN
      IF (.NOT. ASSOCIATED(gt_ptr)) RETURN

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_references.F",219)

      ! start the timing
      CALL timeset(routineN, handle)

      ! set the entry point od the list
      tmp_gt_list_ptr => elem%gt_nodes_references
      tmp_gt_list_last_ptr => elem%gt_nodes_references

      ! search related reference
      DO WHILE (ASSOCIATED(tmp_gt_list_ptr))
         ! remove reference, if it is related to the global tree element
         IF (ASSOCIATED(tmp_gt_list_ptr%gt_elem, gt_ptr)) THEN
            ! first reference?
            IF (ASSOCIATED(tmp_gt_list_ptr, elem%gt_nodes_references)) THEN
               ! additionally last reference (the only one)?
               IF (.NOT. ASSOCIATED(tmp_gt_list_ptr%next)) THEN
                  ! last element in list -> cancel calculation
                  CALL add_to_canceling_list(elem=elem, tmc_env=tmc_env)
                  elem%gt_nodes_references => NULL()
                  tmp_gt_list_last_ptr => NULL()
               ELSE
                  ! if first list element and NOT last one:
                  ! set list pointer to second element
                  elem%gt_nodes_references => tmp_gt_list_ptr%next
                  tmp_gt_list_last_ptr => elem%gt_nodes_references
               END IF
            ELSE
               ! if NOT first one
               ! skip that element in list
               tmp_gt_list_last_ptr%next => tmp_gt_list_ptr%next
            END IF

            ! deallocate list element
            DEALLOCATE (tmp_gt_list_ptr)
            ! going back to last list element
            tmp_gt_list_ptr => tmp_gt_list_last_ptr
         END IF
         ! setting to next list element
         tmp_gt_list_last_ptr => tmp_gt_list_ptr
         ! go to next list element, if defined
         IF (ASSOCIATED(tmp_gt_list_ptr)) tmp_gt_list_ptr => tmp_gt_list_ptr%next
      END DO
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE search_and_remove_reference_in_list

END MODULE tmc_tree_references
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief tree nodes search etc.
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_tree_search
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE tmc_stati,                       ONLY: TMC_STATUS_WAIT_FOR_NEW_TASK
   USE tmc_tree_references,             ONLY: add_to_references,&
                                              search_and_remove_reference_in_list
   USE tmc_tree_types,                  ONLY: &
        elem_array_type, global_tree_type, status_accepted, status_accepted_result, &
        status_calc_approx_ener, status_calculate_MD, status_calculate_NMC_steps, &
        status_calculate_energy, status_calculated, status_cancel_ener, status_cancel_nmc, &
        status_canceled_ener, status_canceled_nmc, status_created, status_deleted, &
        status_deleted_result, status_rejected, status_rejected_result, tree_type
   USE tmc_types,                       ONLY: tmc_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_tree_search'

   PUBLIC :: most_prob_end
   PUBLIC :: search_next_energy_calc
   PUBLIC :: search_canceling_elements
   PUBLIC :: search_parent_element, get_subtree_elements_to_check
   PUBLIC :: search_next_gt_element_to_check
   PUBLIC :: search_end_of_clean_g_tree, search_end_of_clean_tree
   PUBLIC :: count_prepared_nodes_in_trees, count_nodes_in_trees
CONTAINS

   !============================================================================
   ! search tree node
   !============================================================================
! **************************************************************************************************
!> \brief search most probable end in global tree to create a new tree node
!>         using the acceptance probabilities for each move type
!>          of each temperature
!>        routine distinguishes the search for most probable node
!>         for energy and most probable node with open end
!>         for new configuration
!>        In case of searching open end:
!>         routine stops in branch with canceled NMC,
!>         using this a one possibility
!> \param global_tree_elem starting point for search
!> \param prob return value, the probability of reaching the tree node
!> \param n_acc drection of branch the next tree node should extend
!> \param search_energy_node ...
!> \parma search_energy_node flag if configuration for calculating exact
!>        energy sould be searched
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE most_prob_end(global_tree_elem, prob, n_acc, &
                                      search_energy_node)
      TYPE(global_tree_type), POINTER                    :: global_tree_elem
      REAL(KIND=dp), INTENT(OUT)                         :: prob
      LOGICAL, INTENT(INOUT)                             :: n_acc
      LOGICAL, OPTIONAL                                  :: search_energy_node

      CHARACTER(LEN=*), PARAMETER :: routineN = 'most_prob_end', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: check_accepted, check_rejected, keep_on, &
                                                            tmp_acc, tmp_nacc
      REAL(KIND=dp)                                      :: prob_n_acc, prob_n_nacc
      TYPE(global_tree_type), POINTER                    :: ptr_acc, ptr_nacc
      TYPE(tree_type), POINTER                           :: st_elem

      NULLIFY (st_elem, ptr_acc, ptr_nacc)

      prob_n_acc = -100000
      prob_n_nacc = -100000
      check_accepted = .FALSE.
      check_rejected = .FALSE.
      keep_on = .TRUE.

      IF(.NOT.(ASSOCIATED(global_tree_elem)))CALL cp__a("tmc/tmc_tree_search.F",88)
      st_elem => global_tree_elem%conf(global_tree_elem%mv_conf)%elem
      IF(.NOT.(ASSOCIATED(st_elem)))CALL cp__a("tmc/tmc_tree_search.F",90)

      ! start the timing
      CALL timeset(routineN, handle)

      !-- follow trajectory until end
      !-- evaluate following elements using status, and probabilites
      SELECT CASE (global_tree_elem%stat)
      CASE (status_accepted, status_accepted_result)
         check_accepted = .TRUE.
      CASE (status_rejected, status_rejected_result)
         check_rejected = .TRUE.
      CASE DEFAULT
         !-- set directions of searching
         SELECT CASE (st_elem%stat)
         CASE (status_created, status_canceled_ener)
            ! just for searching next element to calculate energy for (found)
            IF (PRESENT(search_energy_node)) THEN
               prob = 0.0_dp ! = log(1)
               n_acc = .FALSE. ! not needed, but maybe for initialisation
               keep_on = .FALSE.
            ELSE
               check_accepted = .TRUE.
               check_rejected = .TRUE.
            END IF
         CASE (status_canceled_nmc)
            ! just for search new element to create (found)
            ! canceled elements can be reactivated
            ! the parent element is returned,
            !  the create_new_pt_tree_node check for existing of this node
            IF (.NOT. PRESENT(search_energy_node)) THEN
               prob = 0.0_dp
               n_acc = ASSOCIATED(global_tree_elem%parent%acc, global_tree_elem)
               global_tree_elem => global_tree_elem%parent
               keep_on = .FALSE.
            END IF
         CASE (status_calculated, status_calculate_energy, &
               status_accepted_result, status_accepted, &
               status_rejected, status_rejected_result)
            ! status accepted and rejection needed for swapped
            !  configurations in parallel tempering
            check_accepted = .TRUE.
            check_rejected = .TRUE.
         CASE (status_calculate_MD, status_calculate_NMC_steps, &
               status_calc_approx_ener)
            ! just for searching next element to create
            IF (.NOT. PRESENT(search_energy_node)) THEN
               check_rejected = .TRUE.
            END IF
         CASE (status_cancel_nmc, status_cancel_ener)
         CASE DEFAULT
            CALL cp_abort(cp__l("tmc/tmc_tree_search.F",141), &
                          "unknown sub tree element status "// &
                          cp_to_string(st_elem%stat))
         END SELECT
      END SELECT

      IF (keep_on) THEN
         !-- recursive search, remembering lowest element (tree end),
         !     and multiply probabilities to go there
         !-- search in ACCEPTED branch
         IF (check_accepted) THEN
            ! test if probable accepted child exist and is not rejected
            IF (ASSOCIATED(global_tree_elem%acc)) THEN
               ptr_acc => global_tree_elem%acc
               IF (PRESENT(search_energy_node)) THEN
                  CALL most_prob_end(global_tree_elem=ptr_acc, prob=prob_n_acc, &
                                     n_acc=tmp_acc, &
                                     search_energy_node=search_energy_node)
               ELSE
                  CALL most_prob_end(global_tree_elem=ptr_acc, prob=prob_n_acc, &
                                     n_acc=tmp_acc)
               END IF
               !-- do probability multiplication
               !    (in logscale because of realy small probabilities)
               prob_n_acc = prob_n_acc+LOG(global_tree_elem%prob_acc)
            ELSE
               ! prob of going in acc or rej direction is
               !   calculated in parent element
               prob_n_acc = LOG(global_tree_elem%prob_acc)
               IF (PRESENT(search_energy_node)) prob_n_acc = -100000
               ptr_acc => global_tree_elem
               tmp_acc = .TRUE.
            END IF
         END IF

         !-- search in REJECTED branch
         IF (check_rejected) THEN
            ! test if probabliy rejected child exist
            IF (ASSOCIATED(global_tree_elem%nacc)) THEN
               ptr_nacc => global_tree_elem%nacc
               IF (PRESENT(search_energy_node)) THEN
                  CALL most_prob_end(global_tree_elem=ptr_nacc, prob=prob_n_nacc, &
                                     n_acc=tmp_nacc, &
                                     search_energy_node=search_energy_node)
               ELSE
                  CALL most_prob_end(global_tree_elem=ptr_nacc, prob=prob_n_nacc, &
                                     n_acc=tmp_nacc)
               END IF
               !-- do probability multiplication
               !     (in logscale because of realy small probabilities)
               prob_n_nacc = prob_n_nacc+LOG(1-global_tree_elem%prob_acc)
            ELSE
               ! prob of going in acc or rej direction is
               !   calculated in parent element
               prob_n_nacc = LOG(1-global_tree_elem%prob_acc)
               IF (PRESENT(search_energy_node)) prob_n_nacc = -100000
               ptr_nacc => global_tree_elem
               tmp_nacc = .FALSE.
            END IF
         END IF
         ! test wich direction is more likely
         !   and set result pointer and probability,
         ! remembering the direction
         IF (prob_n_acc .GE. prob_n_nacc) THEN
            prob = prob_n_acc
            global_tree_elem => ptr_acc
            n_acc = tmp_acc
         ELSE
            prob = prob_n_nacc
            global_tree_elem => ptr_nacc
            n_acc = tmp_nacc
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE most_prob_end

! **************************************************************************************************
!> \brief gt_head head of the global tree
!> \param gt_head ...
!> \param new_gt_elem return value the energy should be calculated for
!> \param stat routine status return value
!> \param react_count reactivation counter
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE search_next_energy_calc(gt_head, new_gt_elem, stat, react_count)
      TYPE(global_tree_type), POINTER                    :: gt_head, new_gt_elem
      INTEGER                                            :: stat, react_count

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_next_energy_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: flag
      REAL(KIND=dp)                                      :: prob

      prob = 0.0_dp
      flag = .FALSE.
      IF(.NOT.(ASSOCIATED(gt_head)))CALL cp__a("tmc/tmc_tree_search.F",239)

      ! start the timing
      CALL timeset(routineN, handle)

      new_gt_elem => gt_head

      CALL most_prob_end(global_tree_elem=new_gt_elem, prob=prob, n_acc=flag, &
                         search_energy_node=.TRUE.)

      stat = status_created
      ! set status for master
      !   (if TMC_STATUS_WAIT_FOR_NEW_TASK, no calculation neccessary)
      IF (.NOT. ASSOCIATED(new_gt_elem) .OR. (EXP(prob) .LT. 1.0E-10)) THEN
         stat = TMC_STATUS_WAIT_FOR_NEW_TASK
      ELSE
         ! reactivate canceled elements
         IF (new_gt_elem%conf(new_gt_elem%mv_conf)%elem%stat .EQ. &
             status_canceled_ener) THEN
            CALL add_to_references(gt_elem=new_gt_elem)
            react_count = react_count+1
            new_gt_elem%conf(new_gt_elem%mv_conf)%elem%stat = status_created
         END IF
         ! if elem status is not status_created
         IF (new_gt_elem%conf(new_gt_elem%mv_conf)%elem%stat .NE. status_created) THEN
            stat = TMC_STATUS_WAIT_FOR_NEW_TASK
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE search_next_energy_calc

! **************************************************************************************************
!> \brief searching the parent element (last accepted configuration before)
!> \param current actual tree element
!> \retval parent parent tree element (last accepted one)
!> \author Mandes 12.2012
!> \note routine searches last (assumed) accepted element in subtree
! **************************************************************************************************
   RECURSIVE FUNCTION search_parent_element(current) RESULT(parent)
      TYPE(tree_type), POINTER                           :: current, parent

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_parent_element', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(current)))CALL cp__a("tmc/tmc_tree_search.F",286)

      ! start the timing
      CALL timeset(routineN, handle)

      IF (ASSOCIATED(current%parent)) THEN
         ! the result value if the child (we came from) is in acc direction
         parent => current%parent
         IF (ASSOCIATED(parent%nacc, current)) THEN
            parent => search_parent_element(parent)
         END IF
      ELSE
         ! if parent not exist, we are at the head of the tree
         parent => current
      END IF
      ! end the timing
      CALL timestop(handle)
      IF(.NOT.(ASSOCIATED(parent)))CALL cp__a("tmc/tmc_tree_search.F",303)
   END FUNCTION search_parent_element

! **************************************************************************************************
!> \brief search the next global element in the Markov Chain to check
!> \param ptr start point for search, should be on the known Markov Chain
!> \param found flag if routine was successful
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE search_next_gt_element_to_check(ptr, found)
      TYPE(global_tree_type), POINTER                    :: ptr
      LOGICAL                                            :: found

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_next_gt_element_to_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      found = .FALSE.

      IF(.NOT.(ASSOCIATED(ptr)))CALL cp__a("tmc/tmc_tree_search.F",323)

      ! start the timing
      CALL timeset(routineN, handle)

      ! -- global tree status is not updated after receiving calculations
      !    (not intrinsically), hence try to check elements with could be ready
      SELECT CASE (ptr%stat)
      CASE (status_accepted_result)
         IF (ASSOCIATED(ptr%acc)) THEN
            ptr => ptr%acc
            CALL search_next_gt_element_to_check(ptr, found)
         END IF
      CASE (status_rejected_result)
         IF (ASSOCIATED(ptr%nacc)) THEN
            ptr => ptr%nacc
            CALL search_next_gt_element_to_check(ptr, found)
         END IF
      CASE (status_calculate_energy, status_created, &
            status_calculate_MD, status_calculated, status_calculate_NMC_steps, &
            status_calc_approx_ener, status_accepted, status_rejected)
         found = .TRUE.
      CASE (status_cancel_nmc, status_cancel_ener, status_canceled_nmc, &
            status_canceled_ener)
         ! nothing to do
      CASE DEFAULT
         CALL cp_abort(cp__l("tmc/tmc_tree_search.F",349), &
                       "unexpected status "//cp_to_string(ptr%stat)// &
                       "of global tree elem "//cp_to_string(ptr%nr))
      END SELECT
      ! end the timing
      CALL timestop(handle)

      IF(.NOT.(ASSOCIATED(ptr)))CALL cp__a("tmc/tmc_tree_search.F",356)
   END SUBROUTINE search_next_gt_element_to_check

! **************************************************************************************************
!> \brief get the changed element of the actual global tree element and its
!>        related last accepted parent
!> \param gt_act_elem actual global tree element
!> \param elem1 two subtree elements which should be compared
!> \param elem2 two subtree elements which should be compared
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE get_subtree_elements_to_check(gt_act_elem, elem1, elem2)
      TYPE(global_tree_type), POINTER                    :: gt_act_elem
      TYPE(tree_type), INTENT(OUT), POINTER              :: elem1, elem2

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_subtree_elements_to_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(gt_act_elem)))CALL cp__a("tmc/tmc_tree_search.F",376)

      ! start the timing
      CALL timeset(routineN, handle)

      IF (gt_act_elem%swaped) THEN
         !------------------------------------------------------------
         !-- take the last accepted configurations for check of both configurations, because
         !-- in case of swapping, the last accepted elements have to be compared
         IF (gt_act_elem%conf_n_acc(gt_act_elem%conf(gt_act_elem%mv_conf)%elem%sub_tree_nr)) THEN
            elem1 => gt_act_elem%conf(gt_act_elem%mv_conf)%elem
         ELSE
            elem1 => search_parent_element(gt_act_elem%conf(gt_act_elem%mv_conf)%elem)
         END IF
         ! second element
         IF (gt_act_elem%conf_n_acc(gt_act_elem%conf(gt_act_elem%mv_conf+1)%elem%sub_tree_nr)) THEN
            elem2 => gt_act_elem%conf(gt_act_elem%mv_conf+1)%elem
         ELSE
            elem2 => search_parent_element(gt_act_elem%conf(gt_act_elem%mv_conf+1)%elem)
         END IF
      ELSE
         elem1 => gt_act_elem%conf(gt_act_elem%mv_conf)%elem
         elem2 => search_parent_element(elem1)
      END IF

      ! end the timing
      CALL timestop(handle)

      IF(.NOT.(ASSOCIATED(gt_act_elem)))CALL cp__a("tmc/tmc_tree_search.F",404)
      IF(.NOT.(ASSOCIATED(elem1)))CALL cp__a("tmc/tmc_tree_search.F",405)
      IF(.NOT.(ASSOCIATED(elem2)))CALL cp__a("tmc/tmc_tree_search.F",406)
   END SUBROUTINE get_subtree_elements_to_check

! **************************************************************************************************
!> \brief searches last element on trajectory,
!>        until where the sides of the tree are deleted (of global tree)
!>        also found the last accepted element before
!> \param last_acc returns last accepted element in cleaned tree part
!> \param tree_ptr end point of search
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE search_end_of_clean_g_tree(last_acc, tree_ptr)
      TYPE(global_tree_type), POINTER                    :: last_acc, tree_ptr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_end_of_clean_g_tree', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(last_acc)))CALL cp__a("tmc/tmc_tree_search.F",425)
      IF(.NOT.(ASSOCIATED(tree_ptr)))CALL cp__a("tmc/tmc_tree_search.F",426)

      ! start the timing
      CALL timeset(routineN, handle)

      SELECT CASE (tree_ptr%stat)
      CASE (status_accepted_result)
         IF (ASSOCIATED(tree_ptr%acc) .AND. .NOT. ASSOCIATED(tree_ptr%nacc)) THEN
            last_acc => tree_ptr
            tree_ptr => tree_ptr%acc
            CALL search_end_of_clean_g_tree(last_acc, tree_ptr)
         END IF
      CASE (status_rejected_result)
         IF (ASSOCIATED(tree_ptr%nacc) .AND. .NOT. ASSOCIATED(tree_ptr%acc)) THEN
            tree_ptr => tree_ptr%nacc
            CALL search_end_of_clean_g_tree(last_acc, tree_ptr)
         END IF
      CASE (status_calculated, status_calculate_energy, status_created, status_accepted, status_rejected, &
            status_calculate_MD, status_calculate_NMC_steps, status_calc_approx_ener, &
            status_canceled_ener, status_canceled_nmc, &
            status_cancel_nmc, status_cancel_ener)
         ! nothing to do
      CASE DEFAULT
         CALL cp_abort(cp__l("tmc/tmc_tree_search.F",449), &
                       "the global tree element "//cp_to_string(tree_ptr%nr)// &
                       " stat "//cp_to_string(tree_ptr%stat)//" is UNknown")
      END SELECT
      ! end the timing
      CALL timestop(handle)
      IF(.NOT.(ASSOCIATED(last_acc)))CALL cp__a("tmc/tmc_tree_search.F",455)
      IF(.NOT.(ASSOCIATED(tree_ptr)))CALL cp__a("tmc/tmc_tree_search.F",456)
   END SUBROUTINE search_end_of_clean_g_tree

! **************************************************************************************************
!> \brief searches last element on trajectory,
!>        until where the sides of the tree are deleted (in sub tree)
!>        also found the last accepted element before.
!>        searches the last element which at least have ONE (not calculated)
!>        node in the tree branch
!> \param tree_ptr  ...
!> \param last_acc ...
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE search_end_of_clean_tree(tree_ptr, last_acc)
      TYPE(tree_type), POINTER                           :: tree_ptr
      TYPE(tree_type), INTENT(IN), POINTER               :: last_acc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_end_of_clean_tree', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(tree_ptr)))CALL cp__a("tmc/tmc_tree_search.F",478)
      IF(.NOT.(ASSOCIATED(last_acc)))CALL cp__a("tmc/tmc_tree_search.F",479)

      ! start the timing
      CALL timeset(routineN, handle)

      IF (.NOT. ASSOCIATED(last_acc, tree_ptr)) THEN
         IF (ASSOCIATED(tree_ptr%acc) .AND. .NOT. ASSOCIATED(tree_ptr%nacc)) THEN
            tree_ptr => tree_ptr%acc
            CALL search_end_of_clean_tree(tree_ptr, last_acc)
         ELSE IF (ASSOCIATED(tree_ptr%nacc) .AND. .NOT. ASSOCIATED(tree_ptr%acc)) THEN
            tree_ptr => tree_ptr%nacc
            CALL search_end_of_clean_tree(tree_ptr, last_acc)
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
      IF(.NOT.(ASSOCIATED(tree_ptr)))CALL cp__a("tmc/tmc_tree_search.F",495)
      IF(.NOT.(ASSOCIATED(last_acc)))CALL cp__a("tmc/tmc_tree_search.F",496)
   END SUBROUTINE search_end_of_clean_tree

! **************************************************************************************************
!> \brief searches in all branches down below the entered global tree element
!>        for elements to cancel, if prob is present start searching at the
!>        related tree child node
!> \param pt_elem_in start search point
!> \param prob the acceptance probability of the tree element to define
!>        the direction to start with
!> \param tmc_env TMC environment
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE search_canceling_elements(pt_elem_in, prob, tmc_env)
      TYPE(global_tree_type), INTENT(IN), POINTER        :: pt_elem_in
      REAL(KIND=dp), OPTIONAL                            :: prob
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_canceling_elements', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: ready
      TYPE(global_tree_type), POINTER                    :: act_pt_ptr, pt_elem

      NULLIFY (pt_elem, act_pt_ptr)
      IF(.NOT.(ASSOCIATED(pt_elem_in)))CALL cp__a("tmc/tmc_tree_search.F",522)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_search.F",523)

      ! start the timing
      CALL timeset(routineN, handle)

      ready = .TRUE.
      ! if prob present select the related branch
      IF (PRESENT(prob)) THEN
         IF (prob .LT. 1.0E-10 .AND. ASSOCIATED(pt_elem_in%acc)) THEN
            pt_elem => pt_elem_in%acc
         ELSE IF (prob .GT. (1.0_dp-1.0E-10) .AND. ASSOCIATED(pt_elem_in%nacc)) THEN
            pt_elem => pt_elem_in%nacc
         ELSE
            ready = .FALSE.
         END IF
      ELSE
         pt_elem => pt_elem_in
      END IF

      IF (ready) THEN
         IF (ASSOCIATED(pt_elem%conf(pt_elem%mv_conf)%elem)) THEN
            SELECT CASE (pt_elem%conf (pt_elem%mv_conf)%elem%stat)
            CASE (status_accepted_result, status_accepted, status_rejected_result, &
                  status_rejected, status_created, status_cancel_nmc, &
                  status_cancel_ener, status_canceled_nmc, status_canceled_ener, &
                  status_calculated, status_deleted, status_deleted_result, &
                  status_calc_approx_ener) ! no canceling
            CASE (status_calculate_NMC_steps, status_calculate_MD, &
                  status_calculate_energy)
               CALL search_and_remove_reference_in_list(gt_ptr=pt_elem, &
                                                        elem=pt_elem%conf(pt_elem%mv_conf)%elem, tmc_env=tmc_env)

            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_tree_search.F",556), &
                             "unknown status of subtree element"// &
                             cp_to_string(pt_elem%conf(pt_elem%mv_conf)%elem%stat))
            END SELECT
         END IF
         !-- go until the ends ot he tree, to search for elements to cancel
         !-- check if child nodes exist
         IF (ASSOCIATED(pt_elem%acc)) THEN
            act_pt_ptr => pt_elem%acc
            CALL search_canceling_elements(pt_elem_in=act_pt_ptr, tmc_env=tmc_env)
         END IF
         IF (ASSOCIATED(pt_elem%nacc)) THEN
            act_pt_ptr => pt_elem%nacc
            CALL search_canceling_elements(pt_elem_in=act_pt_ptr, tmc_env=tmc_env)
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
      IF(.NOT.(ASSOCIATED(pt_elem_in)))CALL cp__a("tmc/tmc_tree_search.F",574)
   END SUBROUTINE search_canceling_elements

! **************************************************************************************************
!> \brief searches for created configurations in all subtrees
!> \param global_tree_ptr pointer to one global tree element
!> \param counters array returning the counters for each subtree
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE count_prepared_nodes_in_trees(global_tree_ptr, counters)
      TYPE(global_tree_type), INTENT(IN), POINTER        :: global_tree_ptr
      INTEGER, DIMENSION(:), POINTER                     :: counters

      CHARACTER(len=*), PARAMETER :: routineN = 'count_prepared_nodes_in_trees', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      TYPE(tree_type), POINTER                           :: begin_ptr

      NULLIFY (begin_ptr)

      IF(.NOT.(ASSOCIATED(global_tree_ptr)))CALL cp__a("tmc/tmc_tree_search.F",595)
      IF(.NOT.(ASSOCIATED(counters)))CALL cp__a("tmc/tmc_tree_search.F",596)
      IF(.NOT.(SIZE(counters(1:)) .EQ. SIZE(global_tree_ptr%conf(:))))CALL cp__a("tmc/tmc_tree_search.F",597)

      ! start the timing
      CALL timeset(routineN, handle)

      counters(:) = 0
      DO i = 1, SIZE(global_tree_ptr%conf(:))
         begin_ptr => global_tree_ptr%conf(i)%elem
         CALL count_prepared_nodes_in_subtree(tree_ptr=begin_ptr, &
                                              counter=counters(i))
      END DO

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE count_prepared_nodes_in_trees

! **************************************************************************************************
!> \brief counts the prepared tree nodes in subtrees
!> \param tree_ptr pointer to one subtree element
!> \param counter returning the amount of prepared
!>        (ready for energy calculation) elements ind certain sub tree
!> \author Mandes 01.2013
! **************************************************************************************************
   RECURSIVE SUBROUTINE count_prepared_nodes_in_subtree(tree_ptr, counter)
      TYPE(tree_type), POINTER                           :: tree_ptr
      INTEGER                                            :: counter

      CHARACTER(len=*), PARAMETER :: routineN = 'count_prepared_nodes_in_subtree', &
         routineP = moduleN//':'//routineN

      TYPE(tree_type), POINTER                           :: tmp_ptr

      NULLIFY (tmp_ptr)

      IF(.NOT.(ASSOCIATED(tree_ptr)))CALL cp__a("tmc/tmc_tree_search.F",631)

      SELECT CASE (tree_ptr%stat)
      CASE (status_accepted, status_accepted_result)
         IF (ASSOCIATED(tree_ptr%acc)) THEN
            tmp_ptr => tree_ptr%acc
            CALL count_prepared_nodes_in_subtree(tmp_ptr, counter)
         END IF
      CASE (status_rejected, status_rejected_result)
         IF (ASSOCIATED(tree_ptr%nacc)) THEN
            tmp_ptr => tree_ptr%nacc
            CALL count_prepared_nodes_in_subtree(tmp_ptr, counter)
         END IF
      CASE (status_created, status_calculate_MD, status_calculate_NMC_steps, &
            status_calc_approx_ener, status_calculated, status_calculate_energy)
         IF (tree_ptr%stat .EQ. status_created) counter = counter+1
         IF (ASSOCIATED(tree_ptr%acc)) THEN
            tmp_ptr => tree_ptr%acc
            CALL count_prepared_nodes_in_subtree(tmp_ptr, counter)
         END IF
         IF (ASSOCIATED(tree_ptr%nacc)) THEN
            tmp_ptr => tree_ptr%nacc
            CALL count_prepared_nodes_in_subtree(tmp_ptr, counter)
         END IF
      CASE (status_cancel_nmc, status_cancel_ener, status_canceled_nmc, &
            status_canceled_ener)
         !TODO maybe also count caneled confs for debug output
      CASE DEFAULT
         CALL cp_abort(cp__l("tmc/tmc_tree_search.F",659), &
                       "stat "//cp_to_string(tree_ptr%stat)// &
                       "of elem "//cp_to_string(tree_ptr%nr)// &
                       "unknown.")
      END SELECT
   END SUBROUTINE count_prepared_nodes_in_subtree

! **************************************************************************************************
!> \brief counts the number of existing nodes in global and subtrees
!> \param global_tree_ptr pointer to one global tree element
!> \param end_of_clean_trees points to the last elements of the clean sub trees
!> \param counters array returning the counters for each subtree
!> \param head_elements_nr node number of the existing
!>        global and sub tree heads
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE count_nodes_in_trees(global_tree_ptr, end_of_clean_trees, &
                                   counters, head_elements_nr)
      TYPE(global_tree_type), POINTER                    :: global_tree_ptr
      TYPE(elem_array_type), DIMENSION(:), POINTER       :: end_of_clean_trees
      INTEGER, DIMENSION(:), POINTER                     :: counters, head_elements_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'count_nodes_in_trees', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      TYPE(global_tree_type), POINTER                    :: begin_gt_ptr
      TYPE(tree_type), POINTER                           :: begin_ptr

      NULLIFY (begin_gt_ptr, begin_ptr)

      IF(.NOT.(ASSOCIATED(global_tree_ptr)))CALL cp__a("tmc/tmc_tree_search.F",690)
      IF(.NOT.(ASSOCIATED(end_of_clean_trees)))CALL cp__a("tmc/tmc_tree_search.F",691)
      IF(.NOT.(ASSOCIATED(counters)))CALL cp__a("tmc/tmc_tree_search.F",692)
      IF(.NOT.(SIZE(counters(1:)) .EQ. SIZE(global_tree_ptr%conf(:))))CALL cp__a("tmc/tmc_tree_search.F",693)

      ! start the timing
      CALL timeset(routineN, handle)

      begin_gt_ptr => global_tree_ptr
      counters(:) = 0
      DO
         IF (.NOT. ASSOCIATED(begin_gt_ptr%parent)) EXIT
         begin_gt_ptr => begin_gt_ptr%parent
      END DO
      head_elements_nr(0) = begin_gt_ptr%nr
      CALL count_nodes_in_global_tree(begin_gt_ptr, counters(0))
      DO i = 1, SIZE(end_of_clean_trees(:))
         begin_ptr => end_of_clean_trees(i)%elem
         DO
            IF (.NOT. ASSOCIATED(begin_ptr%parent)) EXIT
            begin_ptr => begin_ptr%parent
         END DO
         head_elements_nr(i) = begin_ptr%nr
         CALL count_nodes_in_tree(begin_ptr, counters(i))
      END DO

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE count_nodes_in_trees

! **************************************************************************************************
!> \brief counts existing nodes in global tree
!> \param ptr global tree head
!> \param counter return value with the amount of existing global tree elements
!> \author Mandes 01.2013
! **************************************************************************************************
   RECURSIVE SUBROUTINE count_nodes_in_global_tree(ptr, counter)
      TYPE(global_tree_type), INTENT(IN), POINTER        :: ptr
      INTEGER, INTENT(INOUT)                             :: counter

      CHARACTER(len=*), PARAMETER :: routineN = 'count_nodes_in_global_tree', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ptr)))CALL cp__a("tmc/tmc_tree_search.F",733)

      counter = counter+1

      IF (ASSOCIATED(ptr%acc)) &
         CALL count_nodes_in_global_tree(ptr%acc, counter)
      IF (ASSOCIATED(ptr%nacc)) &
         CALL count_nodes_in_global_tree(ptr%nacc, counter)
   END SUBROUTINE count_nodes_in_global_tree

! **************************************************************************************************
!> \brief counts existing nodes in certain sub tree
!> \param ptr subtree tree head
!> \param counter return value with the amount of existing sub tree elements
!> \author Mandes 01.2013
! **************************************************************************************************
   RECURSIVE SUBROUTINE count_nodes_in_tree(ptr, counter)
      TYPE(tree_type), POINTER                           :: ptr
      INTEGER                                            :: counter

      CHARACTER(len=*), PARAMETER :: routineN = 'count_nodes_in_tree', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ptr)))CALL cp__a("tmc/tmc_tree_search.F",756)

      counter = counter+1

      IF (ASSOCIATED(ptr%acc)) &
         CALL count_nodes_in_tree(ptr%acc, counter)
      IF (ASSOCIATED(ptr%nacc)) &
         CALL count_nodes_in_tree(ptr%nacc, counter)
   END SUBROUTINE count_nodes_in_tree
END MODULE tmc_tree_search
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief different move types are applied
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes 11/2012
! **************************************************************************************************

MODULE tmc_moves
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE mathlib,                         ONLY: dihedral_angle,&
                                              rotate_vector
   USE parallel_rng_types,              ONLY: get_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type,&
                                              set_rng_stream
   USE physcon,                         ONLY: boltzmann,&
                                              joule
   USE tmc_calculations,                ONLY: center_of_mass,&
                                              geometrical_center,&
                                              get_scaled_cell,&
                                              nearest_distance
   USE tmc_move_types,                  ONLY: &
        mv_type_MD, mv_type_atom_swap, mv_type_atom_trans, mv_type_gausian_adapt, mv_type_mol_rot, &
        mv_type_mol_trans, mv_type_proton_reorder, mv_type_volume_move, tmc_move_type
   USE tmc_tree_types,                  ONLY: status_frozen,&
                                              status_ok,&
                                              tree_type
   USE tmc_types,                       ONLY: tmc_atom_type,&
                                              tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_moves'

   PUBLIC :: change_pos
   PUBLIC :: elements_in_new_subbox

   INTEGER, PARAMETER :: not_selected = 0
   INTEGER, PARAMETER :: proton_donor = -1
   INTEGER, PARAMETER :: proton_acceptor = 1

CONTAINS
! **************************************************************************************************
!> \brief applying the preselected move type
!> \param tmc_params TMC parameters with dimensions ...
!> \param move_types ...
!> \param rng_stream random number stream
!> \param elem configuration to change
!> \param mv_conf temperature index for determinig the move size
!> \param new_subbox flag if new sub box should be crated
!> \param move_rejected return flag if during configurational change
!>        configuration should still be accepted (not if e.g. atom/molecule
!>        leave the sub box
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE change_pos(tmc_params, move_types, rng_stream, elem, mv_conf, &
                         new_subbox, move_rejected)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(tmc_move_type), POINTER                       :: move_types
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: mv_conf
      LOGICAL                                            :: new_subbox, move_rejected

      CHARACTER(LEN=*), PARAMETER :: routineN = 'change_pos', routineP = moduleN//':'//routineN

      INTEGER                                            :: act_nr_elem_mv, counter, d, i, ind, &
                                                            ind_e, m, nr_molec, nr_sub_box_elem
      INTEGER, DIMENSION(:), POINTER                     :: mol_in_sb
      REAL(KIND=dp)                                      :: rnd
      REAL(KIND=dp), DIMENSION(:), POINTER               :: direction, elem_center

      NULLIFY (direction, elem_center, mol_in_sb)

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_moves.F",88)
      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_moves.F",89)
      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("tmc/tmc_moves.F",90)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_moves.F",91)

      move_rejected = .FALSE.

      CALL set_rng_stream(rng_stream=rng_stream, bg=elem%rng_seed(:, :, 1), &
                          cg=elem%rng_seed(:, :, 2), ig=elem%rng_seed(:, :, 3))

      IF (new_subbox) THEN
         IF (ALL(tmc_params%sub_box_size .GT. 0.0_dp)) THEN
            CALL elements_in_new_subbox(tmc_params=tmc_params, &
                                        rng_stream=rng_stream, elem=elem, &
                                        nr_of_sub_box_elements=nr_sub_box_elem)
         ELSE
            elem%elem_stat(:) = status_ok
         END IF
      END IF

      ! at least one atom should be in the sub box
      IF(.NOT.(ANY(elem%elem_stat(:) .EQ. status_ok)))CALL cp__a("tmc/tmc_moves.F",109)
      IF (tmc_params%nr_elem_mv .EQ. 0) THEN
         ! move all elements (could be all atoms or all molecules)
         act_nr_elem_mv = 0
      ELSE
         act_nr_elem_mv = tmc_params%nr_elem_mv
      END IF
      !-- select the type of move (looked up in list, using the move type index)
      !-- for each move type exist single moves of certain number of elements
      !-- or move of all elements
      !-- one element is a position or velocity of an atom.
      !-- Always all dimension are changed.
      SELECT CASE (elem%move_type)
      CASE (mv_type_gausian_adapt)
         ! just for Gaussian Adaptation
         CALL cp__b("tmc/tmc_moves.F",124,"gaussian adaptation is not imlemented yet.")
!TODO       CALL new_pos_gauss_adapt(acc=ASSOCIATED(elem%parent%acc, elem), &
!                    pos=elem%pos, covari=elem%frc, pot=elem%potential, &
!                    step_size=elem%ekin, pos_aver=elem%vel, temp=elem%ekin_before_md, &
!                    rng_seed=elem%rng_seed, rng_seed_last_acc=last_acc_elem%rng_seed)
         !-- atom translation
      CASE (mv_type_atom_trans)
         IF (act_nr_elem_mv .EQ. 0) &
            act_nr_elem_mv = SIZE(elem%pos)/tmc_params%dim_per_elem
         ALLOCATE (elem_center(tmc_params%dim_per_elem))
         i = 1
         move_elements_loop: DO
            ! select atom
            IF (tmc_params%nr_elem_mv .EQ. 0) THEN
               ind = (i-1)*(tmc_params%dim_per_elem)+1
            ELSE
               rnd = next_random_number(rng_stream) ! next random number
               ind = tmc_params%dim_per_elem* &
                     INT(rnd*(SIZE(elem%pos)/tmc_params%dim_per_elem))+1
            END IF
            ! apply move
            IF (elem%elem_stat(ind) .EQ. status_ok) THEN
               ! displace atom
               DO d = 0, tmc_params%dim_per_elem-1
                  rnd = next_random_number(rng_stream) ! next random number
                  elem%pos(ind+d) = elem%pos(ind+d)+(rnd-0.5)*2.0* &
                                    move_types%mv_size(mv_type_atom_trans, mv_conf)
               END DO
               ! check if new position is in subbox
               elem_center = elem%pos(ind:ind+tmc_params%dim_per_elem-1)
               IF (.NOT. check_pos_in_subbox(pos=elem_center, &
                                             subbox_center=elem%subbox_center, &
                                             box_scale=elem%box_scale, tmc_params=tmc_params) &
                   ) THEN
                  move_rejected = .TRUE.
                  EXIT move_elements_loop
               END IF
            ELSE
               ! element was not in sub box, search new one instead
               IF (tmc_params%nr_elem_mv .GT. 0) i = i-1
            END IF
            i = i+1
            IF (i .GT. act_nr_elem_mv) EXIT move_elements_loop
         END DO move_elements_loop
         DEALLOCATE (elem_center)

         !-- molecule translation
      CASE (mv_type_mol_trans)
         nr_molec = MAXVAL(elem%mol(:))
         ! if all particles should be displaced, set the amount of molecules
         IF (act_nr_elem_mv .EQ. 0) &
            act_nr_elem_mv = nr_molec
         ALLOCATE (mol_in_sb(nr_molec))
         ALLOCATE (elem_center(tmc_params%dim_per_elem))
         mol_in_sb(:) = status_frozen
         ! check if any molecule is in sub_box
         DO m = 1, nr_molec
            CALL get_mol_indeces(tmc_params=tmc_params, mol_arr=elem%mol, mol=m, &
                                 start_ind=ind, end_ind=ind_e)
            CALL geometrical_center(pos=elem%pos(ind:ind_e+tmc_params%dim_per_elem-1), &
                                    center=elem_center)
            IF (check_pos_in_subbox(pos=elem_center, &
                                    subbox_center=elem%subbox_center, &
                                    box_scale=elem%box_scale, tmc_params=tmc_params) &
                ) &
               mol_in_sb(m) = status_ok
         END DO
         ! displace the selected amount of molecules
         IF (ANY(mol_in_sb(:) .EQ. status_ok)) THEN
            ALLOCATE (direction(tmc_params%dim_per_elem))
            counter = 1
            move_molecule_loop: DO
               ! select molecule
               IF (tmc_params%nr_elem_mv .EQ. 0) THEN
                  m = counter
               ELSE
                  rnd = next_random_number(rng_stream) ! next random number
                  m = INT(rnd*nr_molec)+1
               END IF
               CALL get_mol_indeces(tmc_params=tmc_params, mol_arr=elem%mol, mol=m, &
                                    start_ind=ind, end_ind=ind_e)
               ! when "molecule" is single atom, search a new one
               IF (ind .EQ. ind_e) CYCLE move_molecule_loop

               ! calculate displacement
               !  move only molecules, with geom. center in subbox
               IF (mol_in_sb(m) .EQ. status_ok) THEN
                  ! calculate displacement
                  DO d = 1, tmc_params%dim_per_elem
                     rnd = next_random_number(rng_stream) ! next random number
                     direction(d) = (rnd-0.5)*2.0_dp*move_types%mv_size( &
                                    mv_type_mol_trans, mv_conf)
                  END DO
                  ! check if displaced position is still in subbox
                  elem_center(:) = elem_center(:)+direction(:)
                  IF (check_pos_in_subbox(pos=elem_center, &
                                          subbox_center=elem%subbox_center, &
                                          box_scale=elem%box_scale, tmc_params=tmc_params) &
                      ) THEN
                     ! apply move
                     atom_in_mol_loop: DO i = ind, ind_e+tmc_params%dim_per_elem-1, tmc_params%dim_per_elem
                        dim_loop: DO d = 0, tmc_params%dim_per_elem-1
                           elem%pos(i+d) = elem%pos(i+d)+direction(d+1)
                        END DO dim_loop
                     END DO atom_in_mol_loop
                  ELSE
                     ! the whole move is rejected, because one element is outside the subbox
                     move_rejected = .TRUE.
                     EXIT move_molecule_loop
                  END IF
               ELSE
                  ! element was not in sub box, search new one instead
                  IF (tmc_params%nr_elem_mv .GT. 0) counter = counter-1
               END IF
               counter = counter+1
               IF (counter .GT. act_nr_elem_mv) EXIT move_molecule_loop
            END DO move_molecule_loop
            DEALLOCATE (direction)
         END IF
         DEALLOCATE (elem_center)
         DEALLOCATE (mol_in_sb)

         !-- molecule rotation
      CASE (mv_type_mol_rot)
         nr_molec = MAXVAL(elem%mol(:))
         IF (act_nr_elem_mv .EQ. 0) &
            act_nr_elem_mv = nr_molec
         ALLOCATE (mol_in_sb(nr_molec))
         ALLOCATE (elem_center(tmc_params%dim_per_elem))
         mol_in_sb(:) = status_frozen
         ! check if any molecule is in sub_box
         DO m = 1, nr_molec
            CALL get_mol_indeces(tmc_params=tmc_params, mol_arr=elem%mol, mol=m, &
                                 start_ind=ind, end_ind=ind_e)
            CALL geometrical_center(pos=elem%pos(ind:ind_e+tmc_params%dim_per_elem-1), &
                                    center=elem_center)
            IF (check_pos_in_subbox(pos=elem_center, &
                                    subbox_center=elem%subbox_center, &
                                    box_scale=elem%box_scale, tmc_params=tmc_params) &
                ) &
               mol_in_sb(m) = status_ok
         END DO
         ! rotate the selected amount of molecules
         IF (ANY(mol_in_sb(:) .EQ. status_ok)) THEN
            counter = 1
            rot_molecule_loop: DO
               ! select molecule
               IF (tmc_params%nr_elem_mv .EQ. 0) THEN
                  m = counter
               ELSE
                  rnd = next_random_number(rng_stream) ! next random number
                  m = INT(rnd*nr_molec)+1
               END IF
               CALL get_mol_indeces(tmc_params=tmc_params, mol_arr=elem%mol, mol=m, &
                                    start_ind=ind, end_ind=ind_e)
               ! when "molecule" is single atom, search a new one
               IF (ind .EQ. ind_e) CYCLE rot_molecule_loop

               ! apply move
               IF (mol_in_sb(m) .EQ. status_ok) THEN
                  CALL do_mol_rot(pos=elem%pos, ind_start=ind, ind_end=ind_e, &
                                  max_angle=move_types%mv_size( &
                                  mv_type_mol_rot, mv_conf), &
                                  move_types=move_types, rng_stream=rng_stream, &
                                  dim_per_elem=tmc_params%dim_per_elem)
                  ! update sub box status of single atom
                  DO i = ind, ind_e+tmc_params%dim_per_elem-1, tmc_params%dim_per_elem
                     elem_center = elem%pos(i:i+tmc_params%dim_per_elem-1)
                     IF (check_pos_in_subbox(pos=elem_center, &
                                             subbox_center=elem%subbox_center, &
                                             box_scale=elem%box_scale, tmc_params=tmc_params) &
                         ) THEN
                        elem%elem_stat(i:i+tmc_params%dim_per_elem-1) = status_ok
                     ELSE
                        elem%elem_stat(i:i+tmc_params%dim_per_elem-1) = status_frozen
                     END IF
                  END DO
               ELSE
                  ! element was not in sub box, search new one instead
                  IF (tmc_params%nr_elem_mv .GT. 0) counter = counter-1
               END IF
               counter = counter+1
               IF (counter .GT. act_nr_elem_mv) EXIT rot_molecule_loop
            END DO rot_molecule_loop
         END IF
         DEALLOCATE (elem_center)
         DEALLOCATE (mol_in_sb)

         !-- velocity changes for MD
         !-- here all velocities are changed
      CASE (mv_type_MD)
         IF(.NOT.(ASSOCIATED(tmc_params%atoms)))CALL cp__a("tmc/tmc_moves.F",315)
         change_all_velocities_loop: DO i = 1, SIZE(elem%pos)
            !-- attention, move type size is in atomic units of velocity
            IF (elem%elem_stat(i) .NE. status_frozen) THEN
               CALL vel_change(vel=elem%vel(i), &
                               atom_kind=tmc_params%atoms(INT(i/REAL(tmc_params%dim_per_elem, KIND=dp))+1), &
                               phi=move_types%mv_size(mv_type_MD, 1), & ! TODO parallel tempering move sizes for vel_change
                               temp=tmc_params%Temp(mv_conf), &
                               rnd_sign_change=.TRUE., & ! MD_vel_invert, &
                               rng_stream=rng_stream)
            END IF
         END DO change_all_velocities_loop

         !-- proton order and disorder
         !   a loop of molecules is build an in this loop proton acceptors become proton donators
         !   Therefor the molecules are rotated along the not involved O-H bond
      CASE (mv_type_proton_reorder)
         CALL search_and_do_proton_displace_loop(elem=elem, &
                                                 short_loop=move_rejected, rng_stream=rng_stream, &
                                                 tmc_params=tmc_params)

         !-- volume move
         ! the box is increased or decreased and with it the coordinates
      CASE (mv_type_volume_move)
         CALL change_volume(conf=elem, T_ind=mv_conf, move_types=move_types, &
                            rng_stream=rng_stream, tmc_params=tmc_params, &
                            mv_cen_of_mass=tmc_params%mv_cen_of_mass)

         !-- atom swap
         ! two atoms of different types are swapped
      CASE (mv_type_atom_swap)
         CALL swap_atoms(conf=elem, move_types=move_types, rng_stream=rng_stream, &
                         tmc_params=tmc_params)

      CASE DEFAULT
         CALL cp_abort(cp__l("tmc/tmc_moves.F",350), &
                       "unknown move type "// &
                       cp_to_string(elem%move_type))
      END SELECT

      CALL get_rng_stream(rng_stream=rng_stream, bg=elem%rng_seed(:, :, 1), &
                          cg=elem%rng_seed(:, :, 2), ig=elem%rng_seed(:, :, 3))

   END SUBROUTINE change_pos

! **************************************************************************************************
!> \brief gets the index of the first molecule element position and the size
!> \param tmc_params TMC parameters with dim_per_elem
!> \param mol_arr array with molecule information (which atom attend which mol)
!> \param mol the selected molecule number
!> \param start_ind start index of the first atom in molecule
!> \param end_ind index of the last atom in molecule
!> \author Mandes 10.2013
! **************************************************************************************************
   SUBROUTINE get_mol_indeces(tmc_params, mol_arr, mol, start_ind, end_ind)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: mol_arr
      INTEGER, INTENT(IN)                                :: mol
      INTEGER, INTENT(OUT)                               :: start_ind, end_ind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_mol_indeces', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      start_ind = -1
      end_ind = -1

      IF(.NOT.(ASSOCIATED(mol_arr)))CALL cp__a("tmc/tmc_moves.F",383)
      IF(.NOT.(mol .LE. MAXVAL(mol_arr(:))))CALL cp__a("tmc/tmc_moves.F",384)
      ! get start index
      loop_start: DO i = 1, SIZE(mol_arr)
         IF (mol_arr(i) .EQ. mol) THEN
            start_ind = i
            EXIT loop_start
         END IF
      END DO loop_start
      ! get end index
      loop_end: DO i = SIZE(mol_arr), i, -1
         IF (mol_arr(i) .EQ. mol) THEN
            end_ind = i
            EXIT loop_end
         END IF
      END DO loop_end
      ! check if all atoms inbetween attend to molecule
      IF(.NOT.(ALL(mol_arr(start_ind:end_ind) .EQ. mol)))CALL cp__a("tmc/tmc_moves.F",400)
      IF(.NOT.(start_ind .GT. 0))CALL cp__a("tmc/tmc_moves.F",401)
      IF(.NOT.(end_ind .GT. 0))CALL cp__a("tmc/tmc_moves.F",402)
      ! convert to indeces mapped for the position array (multiple dim per atom)
      start_ind = (start_ind-1)*tmc_params%dim_per_elem+1
      end_ind = (end_ind-1)*tmc_params%dim_per_elem+1
   END SUBROUTINE

! **************************************************************************************************
!> \brief checks if a position is within the sub box
!>        returns true if position is inside
!> \param pos array with positions
!> \param subbox_center actual center of sub box
!> \param box_scale scaling factors for the cell
!> \param tmc_params TMC parameters with sub box size and cell
!> \retval inside ...
!> \author Mandes 11.2012
! **************************************************************************************************
   FUNCTION check_pos_in_subbox(pos, subbox_center, box_scale, tmc_params) &
      RESULT(inside)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pos, subbox_center, box_scale
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      LOGICAL                                            :: inside

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_pos_in_subbox', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: flag
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: pos_tmp

      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("tmc/tmc_moves.F",431)
      IF(.NOT.(ASSOCIATED(subbox_center)))CALL cp__a("tmc/tmc_moves.F",432)
      IF(.NOT.(ASSOCIATED(box_scale)))CALL cp__a("tmc/tmc_moves.F",433)
      ! if pressure is defined, no scale should be 0
      flag = .NOT. ((tmc_params%pressure .GT. 0.0_dp) .AND. (ANY(box_scale .EQ. 0.0_dp)))
      IF(.NOT.(flag))CALL cp__a("tmc/tmc_moves.F",436)
      IF(.NOT.(SIZE(pos) .EQ. 3))CALL cp__a("tmc/tmc_moves.F",437)
      IF(.NOT.(SIZE(pos) .EQ. SIZE(subbox_center)))CALL cp__a("tmc/tmc_moves.F",438)

      ! start the timing
      CALL timeset(routineN, handle)

      ALLOCATE (pos_tmp(SIZE(pos)))

      inside = .TRUE.
      ! return if no subbox is defined
      IF (.NOT. ANY(tmc_params%sub_box_size(:) .LE. 0.1_dp)) THEN
         pos_tmp(:) = pos(:)-subbox_center(:)
         CALL get_scaled_cell(cell=tmc_params%cell, box_scale=box_scale, &
                              vec=pos_tmp)
         ! check
         IF (ANY(pos_tmp(:) .GE. tmc_params%sub_box_size(:)/2.0) .OR. &
             ANY(pos_tmp(:) .LE. -tmc_params%sub_box_size(:)/2.0)) THEN
            inside = .FALSE.
         END IF
      END IF
      DEALLOCATE (pos_tmp)
      ! end the timing
      CALL timestop(handle)
   END FUNCTION check_pos_in_subbox

! **************************************************************************************************
!> \brief set a new random sub box center and counte the number of atoms in it
!> \param tmc_params ...
!> \param rng_stream ...
!> \param elem ...
!> \param nr_of_sub_box_elements ...
!> \param
!> \param
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE elements_in_new_subbox(tmc_params, rng_stream, elem, &
                                     nr_of_sub_box_elements)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(tree_type), POINTER                           :: elem
      INTEGER, INTENT(OUT)                               :: nr_of_sub_box_elements

      CHARACTER(LEN=*), PARAMETER :: routineN = 'elements_in_new_subbox', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      REAL(KIND=dp)                                      :: rnd
      REAL(KIND=dp), DIMENSION(3)                        :: box_size
      REAL(KIND=dp), DIMENSION(:), POINTER               :: atom_tmp, center_of_sub_box

      NULLIFY (center_of_sub_box, atom_tmp)

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_moves.F",489)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_moves.F",490)

      ! start the timing
      CALL timeset(routineN, handle)

      IF (ANY(tmc_params%sub_box_size(:) .LE. 0.1_dp)) THEN
         !CALL cp__w("tmc/tmc_moves.F",496,"try to count elements in sub box without sub box.")
         elem%elem_stat = status_ok
         nr_of_sub_box_elements = SIZE(elem%elem_stat)
      ELSE
         ALLOCATE (center_of_sub_box(tmc_params%dim_per_elem))
         ALLOCATE (atom_tmp(tmc_params%dim_per_elem))
         nr_of_sub_box_elements = 0
         ! -- define the center of the sub box
         CALL set_rng_stream(rng_stream=rng_stream, &
                             bg=elem%rng_seed(:, :, 1), cg=elem%rng_seed(:, :, 2), &
                             ig=elem%rng_seed(:, :, 3))

         CALL get_cell(cell=tmc_params%cell, abc=box_size)
         DO i = 1, SIZE(tmc_params%sub_box_size)
            rnd = next_random_number(rng_stream) ! next random number
            center_of_sub_box(i) = rnd*box_size(i)
         END DO
         elem%subbox_center(:) = center_of_sub_box(:)

         CALL get_rng_stream(rng_stream=rng_stream, &
                             bg=elem%rng_seed(:, :, 1), cg=elem%rng_seed(:, :, 2), &
                             ig=elem%rng_seed(:, :, 3))

         ! check all elements if they are in subbox
         DO i = 1, SIZE(elem%pos), tmc_params%dim_per_elem
            atom_tmp(:) = elem%pos(i:i+tmc_params%dim_per_elem-1)
            IF (check_pos_in_subbox(pos=atom_tmp, &
                                    subbox_center=center_of_sub_box, box_scale=elem%box_scale, &
                                    tmc_params=tmc_params)) THEN
               elem%elem_stat(i:i+tmc_params%dim_per_elem-1) = status_ok
               nr_of_sub_box_elements = nr_of_sub_box_elements+1
            ELSE
               elem%elem_stat(i:i+tmc_params%dim_per_elem-1) = status_frozen
            END IF
         END DO
         DEALLOCATE (atom_tmp)
         DEALLOCATE (center_of_sub_box)
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE elements_in_new_subbox

! **************************************************************************************************
!> \brief molecule rotation using quaternions
!> \param pos atom positions
!> \param ind_start starting index in the array
!> \param ind_end index of last atom in the array
!> \param max_angle maximal angle in each direction
!> \param move_types ...
!> \param rng_stream ramdon stream
!> \param dim_per_elem dimension per atom
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE do_mol_rot(pos, ind_start, ind_end, max_angle, move_types, &
                         rng_stream, dim_per_elem)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pos
      INTEGER                                            :: ind_start, ind_end
      REAL(KIND=dp)                                      :: max_angle
      TYPE(tmc_move_type), POINTER                       :: move_types
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      INTEGER                                            :: dim_per_elem

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_mol_rot', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: a1, a2, a3, q0, q1, q2, q3, rnd
      REAL(KIND=dp), DIMENSION(3, 3)                     :: rot
      REAL(KIND=dp), DIMENSION(:), POINTER               :: elem_center

      NULLIFY (elem_center)

      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("tmc/tmc_moves.F",567)
      IF(.NOT.(dim_per_elem .EQ. 3))CALL cp__a("tmc/tmc_moves.F",568)
      IF(.NOT.(ind_start .GT. 0 .AND. ind_start .LT. SIZE(pos)))CALL cp__a("tmc/tmc_moves.F",569)
      IF(.NOT.(ind_end .GT. 0 .AND. ind_end .LT. SIZE(pos)))CALL cp__a("tmc/tmc_moves.F",570)
      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_moves.F",571)

      ! calculate rotation matrix (using quanternions)
      rnd = next_random_number(rng_stream) ! next random number
      a1 = (rnd-0.5)*2.0*max_angle !move_types%mv_size(mv_type_mol_rot,mv_conf)
      rnd = next_random_number(rng_stream) ! next random number
      a2 = (rnd-0.5)*2.0*max_angle !move_types%mv_size(mv_type_mol_rot,mv_conf)
      rnd = next_random_number(rng_stream) ! next random number
      a3 = (rnd-0.5)*2.0*max_angle !move_types%mv_size(mv_type_mol_rot,mv_conf)
      q0 = COS(a2/2)*COS((a1+a3)/2.0_dp)
      q1 = SIN(a2/2)*COS((a1-a3)/2.0_dp)
      q2 = SIN(a2/2)*SIN((a1-a3)/2.0_dp)
      q3 = COS(a2/2)*SIN((a1+a3)/2.0_dp)
      rot = RESHAPE((/q0*q0+q1*q1-q2*q2-q3*q3, 2*(q1*q2-q0*q3), 2*(q1*q3+q0*q2), &
                      2*(q1*q2+q0*q3), q0*q0-q1*q1+q2*q2-q3*q3, 2*(q2*q3-q0*q1), &
                      2*(q1*q3-q0*q2), 2*(q2*q3+q0*q1), q0*q0-q1*q1-q2*q2+q3*q3/), (/3, 3/))

      ALLOCATE (elem_center(dim_per_elem))
      ! calculate geometrical center
      CALL geometrical_center(pos=pos(ind_start:ind_end+dim_per_elem-1), &
                              center=elem_center)

      ! proceed rotation
      atom_loop: DO i = ind_start, ind_end+dim_per_elem-1, dim_per_elem
         pos(i:i+2) = MATMUL(pos(i:i+2)-elem_center(:), rot)+elem_center(:)
      END DO atom_loop
      DEALLOCATE (elem_center)
   END SUBROUTINE do_mol_rot

! **************************************************************************************************
!> \brief velocity change should be gaussian distributed
!>        around the old velocity with respect to kB*T/m
!> \param vel velocity of atom (one direction)
!> \param atom_kind ...
!> \param phi angle for mixing old with random gaussian distributed velocity
!>        phi =90 degree -> only gaussian velocity around 0
!>        phi = 0 degree -> only old velocity (with sign change)
!> \param temp temperature for gaussian distributed velocity
!> \param rnd_sign_change if sign of old velocity should change randomly
!> \param rng_stream random number stream
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE vel_change(vel, atom_kind, phi, temp, rnd_sign_change, rng_stream)
      REAL(KIND=dp), INTENT(INOUT)                       :: vel
      TYPE(tmc_atom_type)                                :: atom_kind
      REAL(KIND=dp), INTENT(IN)                          :: phi, temp
      LOGICAL                                            :: rnd_sign_change
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'vel_change', routineP = moduleN//':'//routineN

      INTEGER                                            :: d
      REAL(KIND=dp)                                      :: delta_vel, kB, rnd1, rnd2, rnd3, rnd_g

      kB = boltzmann/joule

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("tmc/tmc_moves.F",627)

      !phi = move_types%mv_size(mv_type_MD,1) ! TODO parallel tempering move sizes for vel_change
      ! hence first producing a gaussian random number
      rnd1 = next_random_number(rng_stream)
      rnd2 = next_random_number(rng_stream)

      rnd_g = SQRT(-2.0_dp*LOG(rnd1))*COS(2.0_dp*PI*rnd2)
      !we can also produce a second one in the same step:
      !rnd_g2 = SQRT(-2.0_dp*LOG(rnd1))*SIN(2.0_dp*PI*rnd2)

      ! adapting the variance with respect to kB*T/m
      delta_vel = SQRT(kB*temp/atom_kind%mass)*rnd_g
      ! check if TODO random velocity sign change
      ! using detailed balance, velocity sign changes are necessary,
      ! which are done randomly and
      ! can be switched of using MD_vel_invert
      ! without still the balance condition should be fullfilled

      rnd3 = next_random_number(rng_stream)
      IF (rnd3 .GE. 0.5 .AND. rnd_sign_change) THEN
         d = -1
      ELSE
         d = 1
      END IF
      vel = SIN(phi)*delta_vel+COS(phi)*vel*d*1.0_dp
   END SUBROUTINE vel_change

! **************************************************************************************************
!> \brief proton order and disorder (specialized move for ice Ih)
!>        a loop of molecules is build an
!>        in this loop proton acceptors become proton donators
!>        Therefor the molecules are rotated along the not involved O-H bond
!> \param elem sub tree element with actual positions
!> \param short_loop return if the a loop shorter than 6 molecules is found
!>        (should not be in ice structure)
!> \param rng_stream random number stream
!> \param tmc_params TMC parameters with numbers of dimensions per element
!>        number of atoms per molecule
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE search_and_do_proton_displace_loop(elem, short_loop, rng_stream, &
                                                 tmc_params)
      TYPE(tree_type), POINTER                           :: elem
      LOGICAL                                            :: short_loop
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'search_and_do_proton_displace_loop', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1000)                                :: tmp_chr
      INTEGER                                            :: counter, donor_acceptor, handle, k, mol, &
                                                            nr_mol
      INTEGER, DIMENSION(:), POINTER                     :: mol_arr
      REAL(KIND=dp)                                      :: rnd

      NULLIFY (mol_arr)

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_moves.F",686)
      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("tmc/tmc_moves.F",687)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_moves.F",688)

      ! start the timing
      CALL timeset(routineN, handle)

      short_loop = .FALSE.
      counter = 0
      nr_mol = MAXVAL(elem%mol(:))
      ! ind_arr: one array element for each molecule
      ALLOCATE (mol_arr(nr_mol))
      mol_arr(:) = -1
      donor_acceptor = not_selected
      ! select randomly if neighboring molecule is donor / acceptor
      IF (next_random_number(rng_stream) .LT. 0.5_dp) THEN
         donor_acceptor = proton_acceptor
      ELSE
         donor_acceptor = proton_donor
      END IF

      ! first step build loop
      !  select randomly one atom
      rnd = next_random_number(rng_stream)
      ! the randomly selected first atom
      mol = INT(rnd*nr_mol)+1
      counter = counter+1
      mol_arr(counter) = mol

      ! do until the loop is closed
      !  (until path connects back to any spot of the path)
      chain_completition_loop: DO
         counter = counter+1
         ! find nearest neighbor
         !  (with same state, in the chain, proton donator or proton accptor)
         CALL find_nearest_proton_acceptor_donator(elem=elem, mol=mol, &
                                                   donor_acceptor=donor_acceptor, tmc_params=tmc_params, &
                                                   rng_stream=rng_stream)
         IF (ANY(mol_arr(:) .EQ. mol)) &
            EXIT chain_completition_loop
         mol_arr(counter) = mol
      END DO chain_completition_loop
      counter = counter-1 ! last searched element is equal to one other in list

      ! just take the loop of molecules out of the chain
      DO k = 1, counter
         IF (mol_arr(k) .EQ. mol) &
            EXIT
      END DO
      mol_arr(1:counter-k+1) = mol_arr(k:counter)
      counter = counter-k+1

      ! check if loop is minimum size of 6 molecules
      IF (counter .LT. 6) THEN
         CALL cp_warn(cp__l("tmc/tmc_moves.F",740), &
                      "short proton loop with"//cp_to_string(counter)// &
                      "molecules.")
         tmp_chr = ""
         WRITE (tmp_chr, *) mol_arr(1:counter)
         CALL cp__w("tmc/tmc_moves.F",745,"selected molecules:"//TRIM(tmp_chr))
         short_loop = .TRUE.
      ENDIF

      ! rotate the molecule along the not involved O-H bond
      !   (about the angle in of the neighboring chain elements)
      CALL rotate_molecules_in_chain(tmc_params=tmc_params, elem=elem, &
                                     mol_arr_in=mol_arr(1:counter), donor_acceptor=donor_acceptor)
      DEALLOCATE (mol_arr)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE search_and_do_proton_displace_loop

! **************************************************************************************************
!> \brief searches the next (first atom of) neighboring molecule
!>        which is proton donor / acceptor
!> \param elem sub tree element with actual positions
!> \param mol (in_out) actual regarded molecule, which neighbor is searched for
!> \param donor_acceptor type of searched neighbor
!>        (proton donor or proton acceptor)
!> \param tmc_params TMC parameters with numbers of dimensions per element
!>        number of atoms per molecule
!> \param rng_stream random number stream
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE find_nearest_proton_acceptor_donator(elem, mol, donor_acceptor, &
                                                   tmc_params, rng_stream)
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: mol, donor_acceptor
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'find_nearest_proton_acceptor_donator', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ind, ind_e, ind_n, mol_tmp, &
                                                            nr_mol
      INTEGER, DIMENSION(2)                              :: neighbor_mol
      REAL(KIND=dp)                                      :: dist_tmp, rnd
      REAL(KIND=dp), DIMENSION(:), POINTER               :: distH1, distH2, distO

      NULLIFY (distO, distH1, distH2)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_moves.F",788)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_moves.F",789)

      ! start the timing
      CALL timeset(routineN, handle)

      nr_mol = MAXVAL(elem%mol)
      ALLOCATE (distO(nr_mol))
      ALLOCATE (distH1(nr_mol))
      ALLOCATE (distH2(nr_mol))
      !-- initialize the distances to huge values
      ! distance of nearest proton of certain molecule to preselected O
      distO(:) = HUGE(distO(1))
      ! distance of (first) proton of preselected molecule to certain molecule
      distH1(:) = HUGE(distH1(1))
      ! distance of (second) proton of preselected molecule to certain molecule
      distH2(:) = HUGE(distH2(1))

      ! get the indices of the old O atom (assuming the first atom of the molecule the first atom)
      CALL get_mol_indeces(tmc_params=tmc_params, mol_arr=elem%mol, mol=mol, &
                           start_ind=ind, end_ind=ind_e)

      ! calculate distances to all molecules
      list_distances: DO mol_tmp = 1, nr_mol
         IF (mol_tmp .EQ. mol) CYCLE list_distances
         ! index of the molecule (the O atom)
         ! assume the first atom of the molecule the first atom
         CALL get_mol_indeces(tmc_params=tmc_params, mol_arr=elem%mol, &
                              mol=mol_tmp, start_ind=ind_n, end_ind=ind_e)
         ! check if selected molecule is water respectively consists of 3 atoms
         IF (MOD(ind_e-ind_n, 3) .GT. 0) THEN
            CALL cp_warn(cp__l("tmc/tmc_moves.F",819), &
                         "selected a molecule with more than 3 atoms, "// &
                         "the proton reordering does not support, skip molecule")
            CYCLE list_distances
         END IF
         IF (donor_acceptor .EQ. proton_acceptor) THEN
            IF (check_donor_acceptor(elem=elem, i_orig=ind, i_neighbor=ind_n, &
                                     tmc_params=tmc_params) .EQ. proton_acceptor) THEN
               !distance of fist proton to certain O
               distH1(mol_tmp) = nearest_distance( &
                                 x1=elem%pos(ind+tmc_params%dim_per_elem: &
                                             ind+2*tmc_params%dim_per_elem-1), &
                                 x2=elem%pos(ind_n:ind_n+tmc_params%dim_per_elem-1), &
                                 cell=tmc_params%cell, box_scale=elem%box_scale)
               !distance of second proton to certain O
               distH2(mol_tmp) = nearest_distance( &
                                 x1=elem%pos(ind+2*tmc_params%dim_per_elem: &
                                             ind+3*tmc_params%dim_per_elem-1), &
                                 x2=elem%pos(ind_n:ind_n+tmc_params%dim_per_elem-1), &
                                 cell=tmc_params%cell, box_scale=elem%box_scale)
            END IF
         END IF
         !check for neighboring proton donors
         IF (donor_acceptor .EQ. proton_donor) THEN
            IF (check_donor_acceptor(elem=elem, i_orig=ind, i_neighbor=ind_n, &
                                     tmc_params=tmc_params) .EQ. proton_donor) THEN
               !distance of selected O to all first protons of other melecules
               distO(mol_tmp) = nearest_distance( &
                                x1=elem%pos(ind:ind+tmc_params%dim_per_elem-1), &
                                x2=elem%pos(ind_n+tmc_params%dim_per_elem: &
                                            ind_n+2*tmc_params%dim_per_elem-1), &
                                cell=tmc_params%cell, box_scale=elem%box_scale)
               dist_tmp = nearest_distance( &
                          x1=elem%pos(ind:ind+tmc_params%dim_per_elem-1), &
                          x2=elem%pos(ind_n+2*tmc_params%dim_per_elem: &
                                      ind_n+3*tmc_params%dim_per_elem-1), &
                          cell=tmc_params%cell, box_scale=elem%box_scale)
               IF (dist_tmp .LT. distO(mol_tmp)) distO(mol_tmp) = dist_tmp
            END IF
         END IF
      END DO list_distances

      mol_tmp = 1
      ! select the nearest neighbors
      !check for neighboring proton acceptors
      IF (donor_acceptor .EQ. proton_acceptor) THEN
         neighbor_mol(mol_tmp) = MINLOC(distH1(:), 1)
         neighbor_mol(mol_tmp+1) = MINLOC(distH2(:), 1)
         ! if both smallest distances points to the shortest molecule search also the second next shortest distance
         IF (neighbor_mol(mol_tmp) .EQ. neighbor_mol(mol_tmp+1)) THEN
            distH1(neighbor_mol(mol_tmp)) = HUGE(distH1(1))
            distH2(neighbor_mol(mol_tmp+1)) = HUGE(distH2(1))
            IF (MINVAL(distH1(:), 1) .LT. MINVAL(distH2(:), 1)) THEN
               neighbor_mol(mol_tmp) = MINLOC(distH1(:), 1)
            ELSE
               neighbor_mol(mol_tmp+1) = MINLOC(distH2(:), 1)
            END IF
         END IF
         mol_tmp = mol_tmp+2
      END IF

      !check for neighboring proton donors
      IF (donor_acceptor .EQ. proton_donor) THEN
         neighbor_mol(mol_tmp) = MINLOC(distO(:), 1)
         distO(neighbor_mol(mol_tmp)) = HUGE(distO(1))
         neighbor_mol(mol_tmp+1) = MINLOC(distO(:), 1)
      END IF

      ! select randomly the next neighboring molecule
      rnd = next_random_number(rng_stream)
      ! the randomly selected atom: return value!
      mol_tmp = neighbor_mol(INT(rnd*SIZE(neighbor_mol(:)))+1)
      mol = mol_tmp

      DEALLOCATE (distO)
      DEALLOCATE (distH1)
      DEALLOCATE (distH2)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE find_nearest_proton_acceptor_donator

! **************************************************************************************************
!> \brief checks if neighbor of the selected/orig element
!>        is a proron donator or acceptor
!> \param elem ...
!> \param i_orig ...
!> \param i_neighbor ...
!> \param tmc_params ...
!> \retval donor_acceptor ...
!> \author Mandes 11.2012
! **************************************************************************************************
   FUNCTION check_donor_acceptor(elem, i_orig, i_neighbor, tmc_params) &
      RESULT(donor_acceptor)
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: i_orig, i_neighbor
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      INTEGER                                            :: donor_acceptor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_donor_acceptor', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(4)                        :: distances

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_moves.F",923)
      IF(.NOT.(i_orig .GE. 1 .AND. i_orig .LE. SIZE(elem%pos)))CALL cp__a("tmc/tmc_moves.F",924)
      IF(.NOT.(i_neighbor .GE. 1 .AND. i_neighbor .LE. SIZE(elem%pos)))CALL cp__a("tmc/tmc_moves.F",925)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_moves.F",926)

      ! 1. proton of orig with neighbor O
      distances(1) = nearest_distance( &
                     x1=elem%pos(i_neighbor:i_neighbor+tmc_params%dim_per_elem-1), &
                     x2=elem%pos(i_orig+tmc_params%dim_per_elem: &
                                 i_orig+2*tmc_params%dim_per_elem-1), &
                     cell=tmc_params%cell, box_scale=elem%box_scale)
      ! 2. proton of orig with neighbor O
      distances(2) = nearest_distance( &
                     x1=elem%pos(i_neighbor:i_neighbor+tmc_params%dim_per_elem-1), &
                     x2=elem%pos(i_orig+2*tmc_params%dim_per_elem: &
                                 i_orig+3*tmc_params%dim_per_elem-1), &
                     cell=tmc_params%cell, box_scale=elem%box_scale)
      ! 1. proton of neighbor with orig O
      distances(3) = nearest_distance( &
                     x1=elem%pos(i_orig:i_orig+tmc_params%dim_per_elem-1), &
                     x2=elem%pos(i_neighbor+tmc_params%dim_per_elem: &
                                 i_neighbor+2*tmc_params%dim_per_elem-1), &
                     cell=tmc_params%cell, box_scale=elem%box_scale)
      ! 2. proton of neigbor with orig O
      distances(4) = nearest_distance( &
                     x1=elem%pos(i_orig:i_orig+tmc_params%dim_per_elem-1), &
                     x2=elem%pos(i_neighbor+2*tmc_params%dim_per_elem: &
                                 i_neighbor+3*tmc_params%dim_per_elem-1), &
                     cell=tmc_params%cell, box_scale=elem%box_scale)

      IF (MINLOC(distances(:), 1) .LE. 2) THEN
         donor_acceptor = proton_acceptor
      ELSE
         donor_acceptor = proton_donor
      END IF
   END FUNCTION check_donor_acceptor

! **************************************************************************************************
!> \brief rotates all the molecules in the chain
!>        the protons were flipped from the donor to the acceptor
!> \param tmc_params TMC environment parameters
!> \param elem sub tree element the pos of the molecules in chain should be
!>        changed by rotating
!> \param mol_arr_in array of indeces of molecules, should be rotated
!> \param donor_acceptor gives the direction of rotation
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE rotate_molecules_in_chain(tmc_params, elem, mol_arr_in, &
                                        donor_acceptor)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(tree_type), POINTER                           :: elem
      INTEGER, DIMENSION(:)                              :: mol_arr_in
      INTEGER                                            :: donor_acceptor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rotate_molecules_in_chain', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: H_offset, handle, i, ind
      INTEGER, DIMENSION(:), POINTER                     :: ind_arr
      REAL(KIND=dp)                                      :: dihe_angle, dist_near, tmp
      REAL(KIND=dp), DIMENSION(3)                        :: rot_axis, tmp_1, tmp_2, vec_1O, &
                                                            vec_2H_f, vec_2H_m, vec_2O, vec_3O, &
                                                            vec_4O, vec_rotated
      TYPE(cell_type), POINTER                           :: tmp_cell

      NULLIFY (ind_arr, tmp_cell)

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_moves.F",990)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_moves.F",991)

      ! start the timing
      CALL timeset(routineN, handle)

      ALLOCATE (ind_arr(0:SIZE(mol_arr_in)+1))
      DO i = 1, SIZE(mol_arr_in)
         CALL get_mol_indeces(tmc_params=tmc_params, mol_arr=elem%mol, &
                              mol=mol_arr_in(i), &
                              start_ind=ind_arr(i), end_ind=ind)
      END DO
      ind_arr(0) = ind_arr(SIZE(ind_arr)-2)
      ind_arr(SIZE(ind_arr)-1) = ind_arr(1)

      ! get the scaled cell
      ALLOCATE (tmp_cell)
      CALL get_scaled_cell(cell=tmc_params%cell, box_scale=elem%box_scale, &
                           scaled_cell=tmp_cell)

      ! rotate single molecules
      DO i = 1, SIZE(ind_arr)-2
         ! the 3 O atoms
         vec_1O(:) = elem%pos(ind_arr(i-1):ind_arr(i-1)+tmc_params%dim_per_elem-1)
         vec_2O(:) = elem%pos(ind_arr(i):ind_arr(i)+tmc_params%dim_per_elem-1)
         vec_3O(:) = elem%pos(ind_arr(i+1):ind_arr(i+1)+tmc_params%dim_per_elem-1)
         ! the H atoms
         ! distinguished between the one fixed (rotation axis with 2 O)
         ! and the moved one
         ! if true the first H atom is between the O atoms
         IF (nearest_distance( &
             x1=elem%pos(ind_arr(i+donor_acceptor): &
                         ind_arr(i+donor_acceptor)+tmc_params%dim_per_elem-1), &
             x2=elem%pos(ind_arr(i)+tmc_params%dim_per_elem: &
                         ind_arr(i)+2*tmc_params%dim_per_elem-1), &
             cell=tmc_params%cell, box_scale=elem%box_scale) &
             .LT. &
             nearest_distance( &
             x1=elem%pos(ind_arr(i+donor_acceptor): &
                         ind_arr(i+donor_acceptor)+tmc_params%dim_per_elem-1), &
             x2=elem%pos(ind_arr(i)+2*tmc_params%dim_per_elem: &
                         ind_arr(i)+3*tmc_params%dim_per_elem-1), &
             cell=tmc_params%cell, box_scale=elem%box_scale) &
             ) THEN
            vec_2H_m = elem%pos(ind_arr(i)+tmc_params%dim_per_elem: &
                                ind_arr(i)+2*tmc_params%dim_per_elem-1)
            vec_2H_f = elem%pos(ind_arr(i)+2*tmc_params%dim_per_elem: &
                                ind_arr(i)+3*tmc_params%dim_per_elem-1)
            H_offset = 1
         ELSE
            vec_2H_f = elem%pos(ind_arr(i)+tmc_params%dim_per_elem: &
                                ind_arr(i)+2*tmc_params%dim_per_elem-1)
            vec_2H_m = elem%pos(ind_arr(i)+2*tmc_params%dim_per_elem: &
                                ind_arr(i)+3*tmc_params%dim_per_elem-1)
            H_offset = 2
         END IF

         IF (.TRUE.) THEN !TODO find a better switch for the pauling model

            ! do rotation (NOT pauling model)
            tmp_1 = pbc(vec_2O-vec_1O, tmp_cell)
            tmp_2 = pbc(vec_3O-vec_2H_f, tmp_cell)

            dihe_angle = donor_acceptor*dihedral_angle(tmp_1, vec_2H_f-vec_2O, tmp_2)
            DO ind = ind_arr(i), ind_arr(i)+tmc_params%dim_per_elem*3-1, tmc_params%dim_per_elem
               ! set rotation vector
               !vec_rotated = rotate_vector(vec_2H_m-vec_2O, dihe_angle, vec_2H_f-vec_2O)
               vec_rotated = rotate_vector(elem%pos(ind: &
                                                    ind+tmc_params%dim_per_elem-1)-vec_2O, &
                                           dihe_angle, vec_2H_f-vec_2O)

               ! set new position
               !elem%pos(ind_arr(i)+H_offset*dim_per_elem:ind_arr(i)+(H_offset+1)*dim_per_elem-1) = vec_2O+vec_rotated
               elem%pos(ind:ind+tmc_params%dim_per_elem-1) = vec_2O+vec_rotated
            END DO
         ELSE
            ! using the pauling model
            !  (see Aragones and Vega: Dielectric constant of ices...)
            ! the rotation axis is defined using the 4th not involved O
            !  (next to the not involved H)
            ! O atom next to not involved proton for axis calculation
            dist_near = HUGE(dist_near)
            search_O_loop: DO ind = 1, SIZE(elem%pos), &
                                    tmc_params%dim_per_elem*3
               IF (ind .EQ. ind_arr(i)) CYCLE search_O_loop
               tmp = nearest_distance(x1=vec_2H_f, &
                                      x2=elem%pos(ind:ind+tmc_params%dim_per_elem-1), &
                                      cell=tmc_params%cell, box_scale=elem%box_scale)
               IF (dist_near .GT. tmp) THEN
                  dist_near = tmp
                  vec_4O = elem%pos(ind:ind+tmc_params%dim_per_elem-1)
               END IF
            END DO search_O_loop
            rot_axis = pbc(-vec_2O(:)+vec_4O(:), tmp_cell)
            tmp_1 = pbc(vec_2O-vec_1O, tmp_cell)
            tmp_2 = pbc(vec_3O-vec_4O, tmp_cell)
            dihe_angle = donor_acceptor*dihedral_angle(tmp_1, rot_axis, tmp_2)
            vec_rotated = rotate_vector(vec_2H_m-vec_2O, dihe_angle, rot_axis)
            ! set new position
            elem%pos(ind_arr(i)+H_offset*tmc_params%dim_per_elem: &
                     ind_arr(i)+(H_offset+1)*tmc_params%dim_per_elem-1) &
               = vec_2O+vec_rotated
            vec_rotated = rotate_vector(vec_2H_f-vec_2O, dihe_angle, rot_axis)
            IF (H_offset .EQ. 1) THEN
               H_offset = 2
            ELSE
               H_offset = 1
            END IF
            elem%pos(ind_arr(i)+H_offset*tmc_params%dim_per_elem: &
                     ind_arr(i)+(H_offset+1)*tmc_params%dim_per_elem-1) &
               = vec_2O+vec_rotated
         END IF
      END DO
      DEALLOCATE (tmp_cell)
      DEALLOCATE (ind_arr)
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE rotate_molecules_in_chain

! **************************************************************************************************
!> \brief volume move, the box size is increased or decreased,
!>        using the mv_size a the factor.
!>        the coordinated are scaled moleculewise
!>        (the is moved like the center of mass is moves)
!> \param conf configuration to change with positions
!> \param T_ind temperature index, to select the correct temperature
!>        for move size
!> \param move_types ...
!> \param rng_stream random number generator stream
!> \param tmc_params TMC parameters with e.g. dimensions of atoms and molecules
!> \param mv_cen_of_mass ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE change_volume(conf, T_ind, move_types, rng_stream, tmc_params, &
                            mv_cen_of_mass)
      TYPE(tree_type), POINTER                           :: conf
      INTEGER                                            :: T_ind
      TYPE(tmc_move_type), POINTER                       :: move_types
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      LOGICAL                                            :: mv_cen_of_mass

      CHARACTER(LEN=*), PARAMETER :: routineN = 'change_volume', routineP = moduleN//':'//routineN

      INTEGER                                            :: atom, dir, handle, ind, ind_e, mol
      REAL(KIND=dp)                                      :: rnd, vol
      REAL(KIND=dp), DIMENSION(3)                        :: box_length_new, box_length_orig, &
                                                            box_scale_old
      REAL(KIND=dp), DIMENSION(:), POINTER               :: disp, scaling

      NULLIFY (scaling, disp)

      IF(.NOT.(ASSOCIATED(conf)))CALL cp__a("tmc/tmc_moves.F",1142)
      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_moves.F",1143)
      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("tmc/tmc_moves.F",1144)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_moves.F",1145)
      IF(.NOT.(T_ind .GT. 0 .AND. T_ind .LE. tmc_params%nr_temp))CALL cp__a("tmc/tmc_moves.F",1146)
      IF(.NOT.(tmc_params%dim_per_elem .EQ. 3))CALL cp__a("tmc/tmc_moves.F",1147)
      IF(.NOT.(tmc_params%cell%orthorhombic))CALL cp__a("tmc/tmc_moves.F",1148)

      ! start the timing
      CALL timeset(routineN, handle)

      ALLOCATE (scaling(tmc_params%dim_per_elem))
      ALLOCATE (disp(tmc_params%dim_per_elem))

      box_scale_old(:) = conf%box_scale
      ! get the cell vector length of the configuration (before move)
      CALL get_scaled_cell(cell=tmc_params%cell, box_scale=conf%box_scale, &
                           abc=box_length_new)

      IF (.FALSE.) THEN
         ! the volume move in volume space (dV)
         IF (tmc_params%v_isotropic) THEN
            CALL get_scaled_cell(cell=tmc_params%cell, box_scale=conf%box_scale, &
                                 abc=box_length_new, vol=vol)
            rnd = next_random_number(rng_stream)
            vol = vol+(rnd-0.5_dp)*2.0_dp*move_types%mv_size(mv_type_volume_move, T_ind)
            box_length_new(:) = vol**(1/REAL(3, KIND=dp))
         ELSE
            CALL get_scaled_cell(cell=tmc_params%cell, box_scale=conf%box_scale, &
                                 abc=box_length_new, vol=vol)
            rnd = next_random_number(rng_stream)
            vol = vol+(rnd-0.5_dp)*2.0_dp*move_types%mv_size(mv_type_volume_move, T_ind)
            rnd = next_random_number(rng_stream)
            dir = 1+INT(rnd*3)
            box_length_new(dir) = 1.0_dp
            box_length_new(dir) = vol/PRODUCT(box_length_new(:))
         END IF
      ELSE
         ! the volume move in box length space (dL)
         ! increase / decrease box lenght in this direction
         ! l_n = l_o +- rnd * mv_size
         IF (tmc_params%v_isotropic) THEN
            rnd = next_random_number(rng_stream)
            box_length_new(:) = box_length_new(:)+ &
                                (rnd-0.5_dp)*2.0_dp* &
                                move_types%mv_size(mv_type_volume_move, T_ind)
         ELSE
            ! select a random direction
            rnd = next_random_number(rng_stream)
            dir = 1+INT(rnd*3)
            rnd = next_random_number(rng_stream)
            box_length_new(dir) = box_length_new(dir)+ &
                                  (rnd-0.5_dp)*2.0_dp* &
                                  move_types%mv_size(mv_type_volume_move, T_ind)
         END IF
      END IF

      ! get the original box length
      scaling(:) = 1.0_dp
      CALL get_scaled_cell(cell=tmc_params%cell, &
                           box_scale=scaling, &
                           abc=box_length_orig)
      ! get the new box scale
      conf%box_scale(:) = box_length_new(:)/box_length_orig(:)
      ! molecule scaling
      scaling(:) = conf%box_scale(:)/box_scale_old(:)

      IF (mv_cen_of_mass .EQV. .FALSE.) THEN
         ! homogene scaling of atomic coordinates
         DO atom = 1, SIZE(conf%pos), tmc_params%dim_per_elem
            conf%pos(atom:atom+tmc_params%dim_per_elem-1) = &
               conf%pos(atom:atom+tmc_params%dim_per_elem-1)*scaling(:)
         END DO
      ELSE
         DO mol = 1, MAXVAL(conf%mol(:))
            ! move the molecule related to the molecule center of mass
            ! get center of mass
            IF(.NOT.(ASSOCIATED(tmc_params%atoms)))CALL cp__a("tmc/tmc_moves.F",1219)

            CALL get_mol_indeces(tmc_params=tmc_params, mol_arr=conf%mol, mol=mol, &
                                 start_ind=ind, end_ind=ind_e)
            CALL center_of_mass( &
               pos=conf%pos(ind:ind_e+tmc_params%dim_per_elem-1), &
               atoms=tmc_params%atoms(INT(ind/REAL(tmc_params%dim_per_elem, KIND=dp))+1: &
                                      INT(ind_e/REAL(tmc_params%dim_per_elem, KIND=dp))+1), &
               center=disp)
            ! calculate the center of mass DISPLACEMENT
            disp(:) = disp(:)*(scaling(:)-1.0_dp)
            ! displace all atoms of the molecule
            DO atom = ind, ind_e+tmc_params%dim_per_elem-1, tmc_params%dim_per_elem
               conf%pos(atom:atom+tmc_params%dim_per_elem-1) = &
                  conf%pos(atom:atom+tmc_params%dim_per_elem-1)+disp(:)
            END DO
         END DO
      END IF

      DEALLOCATE (scaling)
      DEALLOCATE (disp)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE change_volume

! **************************************************************************************************
!> \brief volume move, two atoms of different types are swapped, both selected
!>        randomly
!> \param conf configuration to change with positions
!> \param move_types ...
!> \param rng_stream random number generator stream
!> \param tmc_params TMC parameters with e.g. dimensions of atoms and molecules
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE swap_atoms(conf, move_types, rng_stream, tmc_params)
      TYPE(tree_type), POINTER                           :: conf
      TYPE(tmc_move_type), POINTER                       :: move_types
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'swap_atoms', routineP = moduleN//':'//routineN

      INTEGER                                            :: a_1, a_2, ind_1, ind_2
      LOGICAL                                            :: found
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: pos_tmp

      IF(.NOT.(ASSOCIATED(conf)))CALL cp__a("tmc/tmc_moves.F",1266)
      IF(.NOT.(ASSOCIATED(move_types)))CALL cp__a("tmc/tmc_moves.F",1267)
      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("tmc/tmc_moves.F",1268)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_moves.F",1269)
      IF(.NOT.(ASSOCIATED(tmc_params%atoms)))CALL cp__a("tmc/tmc_moves.F",1270)

      ! loop until two different atoms are found
      atom_search_loop: DO
         ! select one atom randomly
         a_1 = INT(SIZE(conf%pos)/REAL(tmc_params%dim_per_elem, KIND=dp)* &
                   next_random_number(rng_stream))+1
         ! select the second atom randomly
         a_2 = INT(SIZE(conf%pos)/REAL(tmc_params%dim_per_elem, KIND=dp)* &
                   next_random_number(rng_stream))+1
         ! check if they have different kinds
         IF (tmc_params%atoms(a_1)%name .NE. tmc_params%atoms(a_2)%name) THEN
            ! if present, check if atoms have different type related to the specified table
            IF (ASSOCIATED(move_types%atom_lists)) THEN
               DO ind_1 = 1, SIZE(move_types%atom_lists)
                  IF (ANY(move_types%atom_lists(ind_1)%atoms(:) .EQ. &
                          tmc_params%atoms(a_1)%name) .AND. &
                      ANY(move_types%atom_lists(ind_1)%atoms(:) .EQ. &
                          tmc_params%atoms(a_2)%name)) THEN
                     found = .TRUE.
                     EXIT atom_search_loop
                  END IF
               END DO
            ELSE
               found = .TRUE.
               EXIT atom_search_loop
            END IF
         END IF
      END DO atom_search_loop
      IF (found) THEN
         ! perform coordinate exchange
         ALLOCATE (pos_tmp(tmc_params%dim_per_elem))
         ind_1 = (a_1-1)*tmc_params%dim_per_elem+1
         pos_tmp(:) = conf%pos(ind_1:ind_1+tmc_params%dim_per_elem-1)
         ind_2 = (a_2-1)*tmc_params%dim_per_elem+1
         conf%pos(ind_1:ind_1+tmc_params%dim_per_elem-1) = &
            conf%pos(ind_2:ind_2+tmc_params%dim_per_elem-1)
         conf%pos(ind_2:ind_2+tmc_params%dim_per_elem-1) = pos_tmp(:)
         DEALLOCATE (pos_tmp)
      END IF
   END SUBROUTINE swap_atoms

END MODULE tmc_moves
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief tree nodes creation, deallocation, references etc.
!>        - we distinguish two kinds of tree nodes: global and sub tree nodes
!>          (because we also are able to do parallel tempering)
!>        - global tree nodes consists of pointers to sub tree nodes
!>        - sub tree nodes consists of position arrays, potential energy, etc.
!>        - furthermore the sub tree elements have references the all global
!>          tree elements refering to them
!>        - for tree element details see tree_types.F
!>
!>        - for creating we always start with the global tree element
!>          (if not already exist)
!>        - for each new global tree element (depending on the move type):
!>           - two sub tree elements are swapped (Parallel Tempering)
!>             (in global tree element creation)
!>           - the volume of a subtree element is changed
!>             (directly in sub tree element creation)
!>           - positions in one subtree element changes
!>             (in sub tree elem creation or NMC)
!>           - ...
!>        - sub tree elements will be deleted only if no reference to
!>          any global tree element exist anymore
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_tree_build
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE parallel_rng_types,              ONLY: get_rng_stream,&
                                              next_random_number,&
                                              reset_to_next_rng_substream,&
                                              set_rng_stream
   USE tmc_calculations,                ONLY: calc_e_kin,&
                                              init_vel
   USE tmc_dot_tree,                    ONLY: create_dot,&
                                              create_dot_color,&
                                              create_global_tree_dot,&
                                              create_global_tree_dot_color
   USE tmc_file_io,                     ONLY: read_restart_file,&
                                              write_result_list_element
   USE tmc_move_handle,                 ONLY: select_random_move_type
   USE tmc_move_types,                  ONLY: &
        mv_type_MD, mv_type_NMC_moves, mv_type_atom_swap, mv_type_atom_trans, &
        mv_type_gausian_adapt, mv_type_mol_rot, mv_type_mol_trans, mv_type_none, &
        mv_type_proton_reorder, mv_type_swap_conf, mv_type_volume_move
   USE tmc_moves,                       ONLY: change_pos,&
                                              elements_in_new_subbox
   USE tmc_stati,                       ONLY: TMC_STATUS_FAILED,&
                                              TMC_STATUS_WAIT_FOR_NEW_TASK,&
                                              task_type_MC,&
                                              task_type_gaussian_adaptation
   USE tmc_tree_references,             ONLY: add_to_references,&
                                              remove_gt_references,&
                                              remove_subtree_element_of_all_references,&
                                              search_and_remove_reference_in_list
   USE tmc_tree_search,                 ONLY: most_prob_end,&
                                              search_end_of_clean_g_tree,&
                                              search_end_of_clean_tree,&
                                              search_parent_element
   USE tmc_tree_types,                  ONLY: &
        add_to_list, elem_array_type, global_tree_type, status_accepted, status_accepted_result, &
        status_calc_approx_ener, status_calculate_MD, status_calculate_NMC_steps, &
        status_calculate_energy, status_calculated, status_cancel_ener, status_cancel_nmc, &
        status_canceled_ener, status_canceled_nmc, status_created, status_deleted, &
        status_deleted_result, status_ok, status_rejected, status_rejected_result, tree_type
   USE tmc_types,                       ONLY: tmc_env_type,&
                                              tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_tree_build'

   PUBLIC :: allocate_new_sub_tree_node, deallocate_sub_tree_node
   PUBLIC :: init_tree_mod, finalize_init
   PUBLIC :: create_new_gt_tree_node
   PUBLIC :: remove_unused_g_tree
   PUBLIC :: remove_all_trees
   PUBLIC :: finalize_trees
CONTAINS

   !********************************************************************************
   ! ALLOCATION - DEALLOCATION
   !********************************************************************************
! **************************************************************************************************
!> \brief allocates an elements of the global element structure
!> \param next_el ...
!> \param nr_temp ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE allocate_new_global_tree_node(next_el, nr_temp)
      TYPE(global_tree_type), POINTER                    :: next_el
      INTEGER                                            :: nr_temp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_new_global_tree_node', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, itmp

      IF(.NOT.(.NOT. ASSOCIATED(next_el)))CALL cp__a("tmc/tmc_tree_build.F",109)

      ! start the timing
      CALL timeset(routineN, handle)

      ! allocate everything
      ALLOCATE (next_el)
      ALLOCATE (next_el%conf(nr_temp))
      ALLOCATE (next_el%conf_n_acc(nr_temp))
      next_el%rnd_nr = -1.0_dp

      DO itmp = 1, nr_temp
         NULLIFY (next_el%conf(itmp)%elem)
         next_el%conf_n_acc(itmp) = .FALSE.
      END DO

      next_el%swaped = .FALSE.
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE allocate_new_global_tree_node

! **************************************************************************************************
!> \brief deallocates an elements of the global element structure
!> \param gt_elem ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE deallocate_global_tree_node(gt_elem)
      TYPE(global_tree_type), POINTER                    :: gt_elem

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_global_tree_node', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(gt_elem)))CALL cp__a("tmc/tmc_tree_build.F",143)

      ! start the timing
      CALL timeset(routineN, handle)

      ! deallocate everything
      DEALLOCATE (gt_elem%conf_n_acc)
      DEALLOCATE (gt_elem%conf)
      DEALLOCATE (gt_elem)
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE deallocate_global_tree_node

! **************************************************************************************************
!> \brief allocates an elements of the subtree element structure
!> \param tmc_params structure for storing all (global) parameters
!> \param next_el ...
!> \param nr_dim ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE allocate_new_sub_tree_node(tmc_params, next_el, nr_dim)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(tree_type), POINTER                           :: next_el
      INTEGER                                            :: nr_dim

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_new_sub_tree_node', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(.NOT. ASSOCIATED(next_el)))CALL cp__a("tmc/tmc_tree_build.F",173)

      ! start the timing
      CALL timeset(routineN, handle)

      ALLOCATE (next_el)
      NULLIFY (next_el%subbox_center, next_el%pos, next_el%mol, next_el%vel, &
               next_el%frc, next_el%dipole, next_el%elem_stat, &
               next_el%gt_nodes_references)

      next_el%scf_energies(:) = HUGE(next_el%scf_energies)
      next_el%scf_energies_count = 0
      ALLOCATE (next_el%pos(nr_dim))
      ALLOCATE (next_el%mol(nr_dim/tmc_params%dim_per_elem))
      ALLOCATE (next_el%vel(nr_dim))
      IF (tmc_params%print_dipole) ALLOCATE (next_el%dipole(tmc_params%dim_per_elem))
      ALLOCATE (next_el%elem_stat(nr_dim))
      next_el%elem_stat = status_ok
      ALLOCATE (next_el%subbox_center(tmc_params%dim_per_elem))
      IF (tmc_params%print_forces .OR. tmc_params%task_type .EQ. task_type_gaussian_adaptation) THEN
         IF (tmc_params%task_type .EQ. task_type_gaussian_adaptation) THEN
            ALLOCATE (next_el%frc(nr_dim*nr_dim))
         ELSE
            ALLOCATE (next_el%frc(nr_dim))
         END IF
         next_el%frc = 0.0_dp
      END IF
      ALLOCATE (next_el%box_scale(3))
      next_el%pos(:) = -1.0_dp
      next_el%mol(:) = -1
      next_el%box_scale(:) = 1.0_dp
      next_el%scf_energies(:) = 0.0_dp
      next_el%e_pot_approx = 0.0_dp
      next_el%potential = 76543.0_dp
      next_el%vel = 0.0_dp ! standart MC don"t uses velocities, but it is used at least in acceptance check
      next_el%ekin = 0.0_dp
      next_el%ekin_before_md = 0.0_dp
      next_el%sub_tree_nr = 0
      next_el%nr = -1
      next_el%rng_seed(:, :, :) = -1.0
      next_el%move_type = mv_type_none

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE allocate_new_sub_tree_node

! **************************************************************************************************
!> \brief deallocates an elements of the subtree element structure
!> \param tree_elem ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE deallocate_sub_tree_node(tree_elem)
      TYPE(tree_type), POINTER                           :: tree_elem

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_sub_tree_node', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(tree_elem)))CALL cp__a("tmc/tmc_tree_build.F",232)

      ! start the timing
      CALL timeset(routineN, handle)

      ! reference handling
      ! should be not necessary, subtree element should be only deallocated,
      !   if no global tree element points to anymore
      CALL remove_subtree_element_of_all_references(ptr=tree_elem)

      IF (ASSOCIATED(tree_elem%box_scale)) DEALLOCATE (tree_elem%box_scale)
      IF (ASSOCIATED(tree_elem%frc)) DEALLOCATE (tree_elem%frc)
      IF (ASSOCIATED(tree_elem%subbox_center)) DEALLOCATE (tree_elem%subbox_center)
      IF (ASSOCIATED(tree_elem%elem_stat)) DEALLOCATE (tree_elem%elem_stat)
      IF (ASSOCIATED(tree_elem%dipole)) DEALLOCATE (tree_elem%dipole)
      IF (ASSOCIATED(tree_elem%vel)) DEALLOCATE (tree_elem%vel)
      IF (ASSOCIATED(tree_elem%mol)) DEALLOCATE (tree_elem%mol)
      IF (ASSOCIATED(tree_elem%pos)) DEALLOCATE (tree_elem%pos)

      DEALLOCATE (tree_elem)
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE deallocate_sub_tree_node

   !********************************************************************************
   ! INITIALIZATION - FINALIZE
   !********************************************************************************

! **************************************************************************************************
!> \brief routine initiate the global and subtrees with the first elements
!> \param start_elem ...
!> \param tmc_env structure for storing all (global) parameters
!> \param job_counts ...
!> \param worker_timings ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE init_tree_mod(start_elem, tmc_env, job_counts, worker_timings)
      TYPE(tree_type), POINTER                           :: start_elem
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      INTEGER, DIMENSION(:)                              :: job_counts
      REAL(KIND=dp), DIMENSION(4)                        :: worker_timings

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_tree_mod', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      TYPE(global_tree_type), POINTER                    :: global_tree

      NULLIFY (global_tree)

      IF(.NOT.(ASSOCIATED(start_elem)))CALL cp__a("tmc/tmc_tree_build.F",281)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",282)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_tree_build.F",283)

      ! start the timing
      CALL timeset(routineN, handle)

      ! allocate everything
      CALL allocate_new_global_tree_node(next_el=tmc_env%m_env%gt_act, &
                                         nr_temp=tmc_env%params%nr_temp)

      ! use initial/default values
      CALL get_rng_stream(rng_stream=tmc_env%rng_stream, &
                          bg=tmc_env%m_env%gt_act%rng_seed(:, :, 1), &
                          cg=tmc_env%m_env%gt_act%rng_seed(:, :, 2), &
                          ig=tmc_env%m_env%gt_act%rng_seed(:, :, 3))

      global_tree => tmc_env%m_env%gt_act
      tmc_env%m_env%gt_head => tmc_env%m_env%gt_act

      ! set global random seed
      CALL set_rng_stream(rng_stream=tmc_env%rng_stream, bg=global_tree%rng_seed(:, :, 1), &
                          cg=global_tree%rng_seed(:, :, 2), ig=global_tree%rng_seed(:, :, 3))
      global_tree%rnd_nr = next_random_number(tmc_env%rng_stream)

      !-- SUBTREES: set initial values
      DO i = 1, SIZE(global_tree%conf)
         CALL allocate_new_sub_tree_node(tmc_env%params, next_el=global_tree%conf(i)%elem, &
                                         nr_dim=SIZE(start_elem%pos))
         global_tree%conf(i)%elem%move_type = 0
         global_tree%conf(i)%elem%next_elem_nr => tmc_env%m_env%tree_node_count(i)
         global_tree%conf(i)%elem%parent => NULL()
         global_tree%conf(i)%elem%nr = global_tree%conf(i)%elem%next_elem_nr
         global_tree%conf(i)%elem%sub_tree_nr = i
         global_tree%conf(i)%elem%elem_stat = status_ok
         global_tree%conf(i)%elem%pos = start_elem%pos
         global_tree%conf(i)%elem%mol = start_elem%mol
         global_tree%conf(i)%elem%e_pot_approx = start_elem%e_pot_approx
         global_tree%conf(i)%elem%temp_created = i
         global_tree%conf(i)%elem%stat = status_calculate_energy
         !it is default already: global_tree%conf(i)%elem%box_scale(:)  = 1.0_dp
         IF (tmc_env%params%task_type .EQ. task_type_gaussian_adaptation) THEN
            global_tree%conf(i)%elem%vel(:) = start_elem%vel(:)
            global_tree%conf(i)%elem%frc(:) = start_elem%frc(:)
            global_tree%conf(i)%elem%potential = start_elem%potential
            global_tree%conf(i)%elem%ekin = start_elem%ekin
            global_tree%conf(i)%elem%ekin_before_md = start_elem%ekin_before_md
         END IF

         !-- different random seeds for every subtree
         CALL reset_to_next_rng_substream(tmc_env%rng_stream)
         CALL get_rng_stream(rng_stream=tmc_env%rng_stream, bg=global_tree%conf(i)%elem%rng_seed(:, :, 1), &
                             cg=global_tree%conf(i)%elem%rng_seed(:, :, 2), &
                             ig=global_tree%conf(i)%elem%rng_seed(:, :, 3))

         !-- gaussian distributed velocities
         !-- calculating the kinetic energy of the initial configuration velocity
         IF (tmc_env%params%task_type .EQ. task_type_MC) THEN
            IF (tmc_env%params%move_types%mv_weight(mv_type_MD) .GT. 0.0_dp) THEN
               CALL init_vel(vel=global_tree%conf(i)%elem%vel, atoms=tmc_env%params%atoms, &
                             temerature=tmc_env%params%Temp(i), &
                             rng_stream=tmc_env%rng_stream, &
                             rnd_seed=global_tree%conf(i)%elem%rng_seed)
               global_tree%conf(i)%elem%ekin = calc_e_kin(vel=global_tree%conf(i)%elem%vel, &
                                                          atoms=tmc_env%params%atoms)
            END IF
         END IF

         !-- set tree pointer
         !-- set pointer of first global tree element
         tmc_env%m_env%st_heads(i)%elem => global_tree%conf(i)%elem
         tmc_env%m_env%st_clean_ends(i)%elem => global_tree%conf(i)%elem
         !-- set initial pointer of result lists
         tmc_env%m_env%result_list(i)%elem => global_tree%conf(i)%elem
      END DO
      tmc_env%m_env%tree_node_count(:) = 0 ! initializing the tree node numbering

      !-- initial global tree element
      tmc_env%m_env%gt_head => global_tree
      tmc_env%m_env%gt_clean_end => global_tree
      global_tree%nr = 0
      global_tree%swaped = .FALSE.
      global_tree%mv_conf = 1
      global_tree%mv_next_conf = MODULO(global_tree%mv_conf, SIZE(global_tree%conf))+1
      global_tree%conf_n_acc = .TRUE.

      global_tree%stat = status_created
      global_tree%prob_acc = 1.0_dp

      ! simulated annealing start temperature
      global_tree%Temp = tmc_env%params%Temp(1)
      IF (tmc_env%params%nr_temp .NE. 1 .AND. tmc_env%m_env%temp_decrease .NE. 1.0_dp) &
         CALL cp_abort(cp__l("tmc/tmc_tree_build.F",373), &
                       "there is no parallel tempering implementation for simulated annealing implemented "// &
                       "(just one Temp per global tree element.")

      !-- IF program is restarted, read restart file
      IF (tmc_env%m_env%restart_in_file_name .NE. "") THEN
         CALL read_restart_file(tmc_env=tmc_env, job_counts=job_counts, &
                                timings=worker_timings, &
                                file_name=tmc_env%m_env%restart_in_file_name)

         tmc_env%m_env%tree_node_count(0) = global_tree%nr

         DO i = 1, SIZE(tmc_env%m_env%result_list(:))
            tmc_env%m_env%tree_node_count(i) = tmc_env%m_env%result_list(i)%elem%nr
            global_tree%conf(i)%elem%stat = status_accepted
         END DO
         global_tree%prob_acc = 1.0_dp ! accepted (re)start configuration
         WRITE (tmc_env%m_env%io_unit, *) "TMC| restarting at Markov Chain element(s): ", &
            tmc_env%m_env%result_count
         !TODO enable calculation of the approx energy for case of fitting potential
         !     and changing the potential in between
         !     BUT check, there is no double counting (of the last/restarted elem) in the trajectory
         !IF(tmc_env%params%NMC_inp_file.NE."") &
         !  global_tree%conf(1)%elem%stat  = status_calc_approx_ener
         global_tree%stat = status_accepted_result
      ELSE IF (tmc_env%params%NMC_inp_file .NE. "") THEN
         global_tree%conf(1)%elem%stat = status_calc_approx_ener
      ELSE
         global_tree%conf(1)%elem%stat = status_created
      END IF

      !-- set reference of global tree node
      CALL add_to_references(gt_elem=global_tree)

      !-- draw the first global tree node
      IF (tmc_env%params%DRAW_TREE) THEN
         CALL create_global_tree_dot(new_element=global_tree, &
                                     tmc_params=tmc_env%params)
         CALL create_global_tree_dot_color(gt_tree_element=global_tree, &
                                           tmc_params=tmc_env%params)
      END IF

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE init_tree_mod

! **************************************************************************************************
!> \brief distributes the initial energy to all subtree (if no restart) and
!>        call analysis for this element (write trajectory...)
!> \param gt_tree_ptr global tree head (initial configuration)
!> \param tmc_env master environment for restart
!>        (if restart the subtree heads are not equal), result counts and lists
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE finalize_init(gt_tree_ptr, tmc_env)
      TYPE(global_tree_type), POINTER                    :: gt_tree_ptr
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'finalize_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      IF(.NOT.(ASSOCIATED(gt_tree_ptr)))CALL cp__a("tmc/tmc_tree_build.F",435)
      IF(.NOT.(.NOT. ASSOCIATED(gt_tree_ptr%parent)))CALL cp__a("tmc/tmc_tree_build.F",436)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",437)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_tree_build.F",438)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_tree_build.F",439)

      ! start the timing
      CALL timeset(routineN, handle)

      gt_tree_ptr%stat = status_accepted_result
      !-- distribute energy of first element to all subtrees
      DO i = 1, SIZE(gt_tree_ptr%conf)
         gt_tree_ptr%conf(i)%elem%stat = status_accepted_result
         IF (ASSOCIATED(gt_tree_ptr%conf(1)%elem%dipole)) &
            gt_tree_ptr%conf(i)%elem%dipole = gt_tree_ptr%conf(1)%elem%dipole
         IF (tmc_env%m_env%restart_in_file_name .EQ. "") &
            gt_tree_ptr%conf(i)%elem%potential = gt_tree_ptr%conf(1)%elem%potential
      END DO

      IF (tmc_env%m_env%restart_in_file_name .EQ. "") THEN
         tmc_env%m_env%result_count(:) = tmc_env%m_env%result_count(:)+1
         tmc_env%m_env%result_list(:) = gt_tree_ptr%conf(:)
         !-- write initial elements in result files
         DO i = 1, SIZE(tmc_env%m_env%result_list)
            CALL write_result_list_element(result_list=tmc_env%m_env%result_list, &
                                           result_count=tmc_env%m_env%result_count, &
                                           conf_updated=i, accepted=.TRUE., &
                                           tmc_params=tmc_env%params)
            ! save for analysis
            IF (tmc_env%tmc_comp_set%para_env_m_ana%num_pe .GT. 1) THEN
               CALL add_to_list(elem=tmc_env%m_env%result_list(i)%elem, &
                                list=tmc_env%m_env%analysis_list, &
                                nr=tmc_env%m_env%result_count(i), &
                                temp_ind=i)
            END IF
         END DO
         !CALL write_result_list_element(result_list=tmc_env%m_env%result_list, &
         !         result_count=tmc_env%m_env%result_count,&
         !         conf_updated=0, accepted=.TRUE., &
         !         tmc_params=tmc_env%params)
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE finalize_init

   !============================================================================
   ! tree node creation
   !============================================================================
! **************************************************************************************************
!> \brief creates new global tree element and if needed new subtree element
!> \param tmc_env TMC environment with parameters and pointers to gt element
!> \param stat return status value
!> \param new_elem return gt element
!> \param reactivation_cc_count counting the reactivation of subtree elements
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_new_gt_tree_node(tmc_env, stat, new_elem, &
                                      reactivation_cc_count)
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      INTEGER, INTENT(OUT)                               :: stat
      TYPE(global_tree_type), INTENT(OUT), POINTER       :: new_elem
      INTEGER                                            :: reactivation_cc_count

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_new_gt_tree_node', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, swap_conf
      LOGICAL                                            :: keep_on, n_acc
      REAL(KIND=dp)                                      :: prob, rnd, rnd2
      TYPE(global_tree_type), POINTER                    :: tmp_elem
      TYPE(tree_type), POINTER                           :: tree_elem

      NULLIFY (tmp_elem, tree_elem, new_elem)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",509)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_tree_build.F",510)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_tree_build.F",511)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%gt_act)))CALL cp__a("tmc/tmc_tree_build.F",512)

      ! start the timing
      CALL timeset(routineN, handle)

      stat = TMC_STATUS_FAILED
      !-- search most probable end in global tree for new element
      tmp_elem => tmc_env%m_env%gt_act
      n_acc = .TRUE.

      !-- search most probable end to create new element
      CALL most_prob_end(global_tree_elem=tmp_elem, prob=prob, n_acc=n_acc)

      keep_on = .TRUE.
      IF (ASSOCIATED(tmp_elem) .AND. (EXP(prob) .LT. 1.0E-10)) THEN
         new_elem => NULL()
         stat = TMC_STATUS_FAILED
         keep_on = .FALSE.
         !-- if not found, do something else
         !-- (posible if just one end for further calculations
         !    and there a MD move is still calculated)
      ELSE IF (.NOT. ASSOCIATED(tmp_elem)) THEN
         new_elem => NULL()
         stat = TMC_STATUS_FAILED
         keep_on = .FALSE.
      END IF

      IF (keep_on) THEN
         ! if global tree element already exist use that one
         !   (skip creating new element)
         ! reactivation
         IF ((n_acc .AND. ASSOCIATED(tmp_elem%acc)) .OR. &
             ((.NOT. n_acc) .AND. ASSOCIATED(tmp_elem%nacc))) THEN

            !set pointer to the actual element
            IF (n_acc) &
               new_elem => tmp_elem%acc
            IF (.NOT. n_acc) &
               new_elem => tmp_elem%nacc

            ! check for existing subtree element
            IF(.NOT.(ASSOCIATED(new_elem%conf(new_elem%mv_conf)%elem)))CALL cp__a("tmc/tmc_tree_build.F",553)
            SELECT CASE (new_elem%conf (new_elem%mv_conf)%elem%stat)
            CASE (status_cancel_nmc, status_cancel_ener, status_canceled_nmc, &
                  status_canceled_ener)
               ! reactivating subtree element
               !  (but global tree element already exist)
               CALL add_to_references(gt_elem=new_elem)
               reactivation_cc_count = reactivation_cc_count+1
            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_tree_build.F",562), &
                             "global tree node creation using existing sub tree element, "// &
                             "but is not a canceled one, gt elem "// &
                             cp_to_string(new_elem%nr)//" st elem "// &
                             cp_to_string(new_elem%conf(new_elem%mv_conf)%elem%nr)// &
                             " with stat "// &
                             cp_to_string(new_elem%conf(new_elem%mv_conf)%elem%stat))
            END SELECT
            ! change the status of the reactivated subtree element
            ! move is only done by the master,
            !  when standard MC moves with single potential are done
            ! the Nested Monte Carlo routine needs to do the configuration
            !  to have old configuration to see if change is accepted
            SELECT CASE (new_elem%conf (new_elem%mv_conf)%elem%move_type)
            CASE (mv_type_MD)
               new_elem%conf(new_elem%mv_conf)%elem%stat = status_calculate_MD
            CASE (mv_type_NMC_moves)
               IF (new_elem%conf(new_elem%mv_conf)%elem%stat .NE. status_canceled_nmc) &
                  CALL cp_warn(cp__l("tmc/tmc_tree_build.F",580), &
                               "reactivating tree element with wrong status"// &
                               cp_to_string(new_elem%conf(new_elem%mv_conf)%elem%stat))
               new_elem%conf(new_elem%mv_conf)%elem%stat = status_calculate_NMC_steps

               !IF(DEBUG.GE.1) WRITE(tmc_out_file_nr,*)"ATTENTION: reactivation of canceled subtree ", &
               !  new_elem%conf(new_elem%mv_conf)%elem%sub_tree_nr, "elem", new_elem%conf(new_elem%mv_conf)%elem%nr, &
               !  " of existing gt elem ",new_elem%nr,", again calculate NMC steps"
            CASE (mv_type_atom_trans, mv_type_mol_trans, mv_type_mol_rot, &
                  mv_type_proton_reorder)
               CALL cp_abort(cp__l("tmc/tmc_tree_build.F",590), &
                             "reactivated st element has no NMC or MD move type, "// &
                             "but seems to be canceled. Move type"// &
                             cp_to_string(new_elem%conf(new_elem%mv_conf)%elem%move_type))
            CASE DEFAULT
               CALL cp__b("tmc/tmc_tree_build.F",595,"Unknown move type while reactivating subtree element.")
            END SELECT
         ELSE
            !-- if end is found (NOT already existing element), create new elem at the end and if nessecarry new subtree element
            ! set initial values
            CALL allocate_new_global_tree_node(next_el=new_elem, &
                                               nr_temp=tmc_env%params%nr_temp)
            tmc_env%m_env%tree_node_count(0) = tmc_env%m_env%tree_node_count(0)+1
            new_elem%nr = tmc_env%m_env%tree_node_count(0)

            !-- set pointers to and from element one level up
            !-- paste new gt tree node element at right end
            IF (n_acc) THEN
               IF (ASSOCIATED(tmp_elem%acc)) &
                  CALL cp__b("tmc/tmc_tree_build.F",609,"creating new subtree element on an occupied acc branch")
               tmp_elem%acc => new_elem
            ELSE
               IF (ASSOCIATED(tmp_elem%nacc)) &
                  CALL cp__b("tmc/tmc_tree_build.F",613,"creating new subtree element on an occupied nacc branch")
               tmp_elem%nacc => new_elem
            END IF
            new_elem%parent => tmp_elem

            !-- adopt acceptance flags of elements (old)
            new_elem%conf_n_acc(:) = new_elem%parent%conf_n_acc
            !-- set acceptance flag of modified configuration
            !    depending on the direction of attaching new element
            IF (.NOT. new_elem%parent%swaped) THEN
               ! set the flag for the direction
               !  (shows if the configuration is assumed to be acc or rej)
               new_elem%conf_n_acc(new_elem%parent%conf( &
                                   new_elem%parent%mv_conf)%elem%sub_tree_nr) = n_acc
            ELSE
               !-- in case of swapping the subtree element acceptance do not change
               !-- in case of NOT accepted branch and swapping before,
               !-- search last NOT swaped gt tree node to take configurations
               IF (.NOT. n_acc) THEN
                  DO
                     IF (.NOT. ASSOCIATED(tmp_elem%parent)) EXIT
                     IF (ASSOCIATED(tmp_elem%parent%acc, tmp_elem)) THEN
                        tmp_elem => tmp_elem%parent
                        EXIT
                     END IF
                     tmp_elem => tmp_elem%parent
                     IF (.NOT. tmp_elem%swaped) EXIT
                  END DO
               END IF
            END IF

            !-- adapt "old" configurations
            new_elem%conf(:) = tmp_elem%conf(:)

            !-- set rnd nr generator and set next conf to change
            CALL set_rng_stream(rng_stream=tmc_env%rng_stream, &
                                bg=new_elem%parent%rng_seed(:, :, 1), &
                                cg=new_elem%parent%rng_seed(:, :, 2), &
                                ig=new_elem%parent%rng_seed(:, :, 3))
            CALL reset_to_next_rng_substream(tmc_env%rng_stream)
            ! the random number for acceptance check
            new_elem%rnd_nr = next_random_number(tmc_env%rng_stream)

            ! the next configuration index to move
            !rnd = next_random_number(tmc_env%rng_stream)
            !new_elem%mv_conf = 1+INT(size(new_elem%conf)*rnd)
            ! one temperature after each other
            new_elem%mv_conf = new_elem%parent%mv_next_conf
            new_elem%mv_next_conf = MODULO(new_elem%mv_conf, SIZE(new_elem%conf))+1

            ! simulated annealing temperature decrease
            new_elem%Temp = tmp_elem%Temp
            IF (n_acc) new_elem%Temp = tmp_elem%Temp*(1-tmc_env%m_env%temp_decrease)

            !-- rnd for swap
            rnd = next_random_number(tmc_env%rng_stream)
            rnd2 = next_random_number(tmc_env%rng_stream)
            CALL get_rng_stream(rng_stream=tmc_env%rng_stream, &
                                bg=new_elem%rng_seed(:, :, 1), &
                                cg=new_elem%rng_seed(:, :, 2), &
                                ig=new_elem%rng_seed(:, :, 3))

            ! swap moves are not part of the subtree structure,
            !  because exisiting elements from DIFFERENT subtrees are swaped
            ! -- do swap ?!
            IF (tmc_env%params%move_types%mv_weight(mv_type_swap_conf) .GE. rnd) THEN
               ! set the index for the swaping element
               !  and the conf to move in next move
               new_elem%mv_next_conf = new_elem%mv_conf
               ! do swap with conf swap_conf and swap_conf+1
               swap_conf = 1+INT((tmc_env%params%nr_temp-1)*rnd2)
               new_elem%mv_conf = swap_conf
               !-- swaping pointers to subtree elements
               ! exchange the pointer to the sub tree elements
               tree_elem => new_elem%conf(new_elem%mv_conf)%elem
               new_elem%conf(new_elem%mv_conf)%elem => &
                  new_elem%conf(new_elem%mv_conf+1)%elem
               new_elem%conf(new_elem%mv_conf+1)%elem => tree_elem

               new_elem%stat = status_calculated
               new_elem%swaped = .TRUE.
               new_elem%prob_acc = tmc_env%params%move_types%acc_prob( &
                                   mv_type_swap_conf, new_elem%mv_conf)
               CALL add_to_references(gt_elem=new_elem)
               IF (tmc_env%params%DRAW_TREE) &
                  CALL create_global_tree_dot(new_element=new_elem, &
                                              tmc_params=tmc_env%params)
               ! nothing to do for the workers
               stat = status_calculated
               keep_on = .FALSE.
            ELSE

               !-- considered subtree node can already exist,
               !    calculated somewhere else in the global tree
               !-- so check if new sub tree node exists, if not, create it
               !-- check if considered configuration is assumed to be
               !    on accepted or rejected branch
               IF (new_elem%conf_n_acc(new_elem%conf(new_elem%mv_conf)%elem%sub_tree_nr)) THEN
                  !-- check if child element in ACCEPTED direction already exist
                  IF (ASSOCIATED(new_elem%conf(new_elem%mv_conf)%elem%acc)) THEN
                     new_elem%conf(new_elem%mv_conf)%elem => &
                        new_elem%conf(new_elem%mv_conf)%elem%acc
                     stat = status_calculated
                  ELSE
                     !-- if not exist create new subtree element
                     CALL create_new_subtree_node(act_gt_el=new_elem, &
                                                  tmc_env=tmc_env)
                     IF (tmc_env%params%DRAW_TREE) &
                        CALL create_dot(new_element=new_elem%conf(new_elem%mv_conf)%elem, &
                                        conf=new_elem%conf(new_elem%mv_conf)%elem%sub_tree_nr, &
                                        tmc_params=tmc_env%params)
                  END IF
               ELSE
                  !-- check if child element in REJECTED direction already exist
                  IF (ASSOCIATED(new_elem%conf(new_elem%mv_conf)%elem%nacc)) THEN
                     new_elem%conf(new_elem%mv_conf)%elem => &
                        new_elem%conf(new_elem%mv_conf)%elem%nacc
                     stat = status_calculated
                  ELSE
                     !-- if not exist create new subtree element
                     CALL create_new_subtree_node(act_gt_el=new_elem, &
                                                  tmc_env=tmc_env)
                     IF (tmc_env%params%DRAW_TREE) &
                        CALL create_dot(new_element=new_elem%conf(new_elem%mv_conf)%elem, &
                                        conf=new_elem%conf(new_elem%mv_conf)%elem%sub_tree_nr, &
                                        tmc_params=tmc_env%params)
                  END IF
               END IF
               ! set approximate probability of acceptance
               !  (initialization with calculated values from
               !  (#acc elem in traj)/(#elem in traj))
               new_elem%prob_acc = tmc_env%params%move_types%acc_prob( &
                                   new_elem%conf(new_elem%mv_conf)%elem%move_type, new_elem%mv_conf)
               ! add refence and dot
               CALL add_to_references(gt_elem=new_elem)
               IF (tmc_env%params%DRAW_TREE) &
                  CALL create_global_tree_dot(new_element=new_elem, &
                                              tmc_params=tmc_env%params)
            END IF ! swap or no swap
         END IF ! global tree node already exist. Hence the Subtree node also (it is speculative canceled)
      END IF ! keep on (checking and creating)

      IF (keep_on) THEN ! status changes
         IF (new_elem%stat .EQ. status_accepted_result .OR. &
             new_elem%stat .EQ. status_accepted .OR. &
             new_elem%stat .EQ. status_rejected .OR. &
             new_elem%stat .EQ. status_rejected_result) &
            CALL cp__b("tmc/tmc_tree_build.F",760,"selected existing RESULT gt node")
         !-- set status of global tree element for decision in master routine
         SELECT CASE (new_elem%conf (new_elem%mv_conf)%elem%stat)
         CASE (status_rejected_result, status_rejected, status_accepted, &
               status_accepted_result, status_calculated)
            ! energy is already calculated
            new_elem%stat = status_calculated
            stat = new_elem%conf(new_elem%mv_conf)%elem%stat
            IF (tmc_env%params%DRAW_TREE) &
               CALL create_dot_color(tree_element=new_elem%conf(new_elem%mv_conf)%elem, &
                                     tmc_params=tmc_env%params)
         CASE (status_calc_approx_ener)
            new_elem%stat = new_elem%conf(new_elem%mv_conf)%elem%stat
            IF (stat .NE. status_calculated) THEN
               stat = new_elem%conf(new_elem%mv_conf)%elem%stat
               IF (tmc_env%params%DRAW_TREE) &
                  CALL create_dot_color(tree_element=new_elem%conf(new_elem%mv_conf)%elem, &
                                        tmc_params=tmc_env%params)
            END IF
         CASE (status_calculate_MD, status_calculate_energy, &
               status_calculate_NMC_steps, status_created)
            ! if not already in progress, set status for new task message
            new_elem%stat = new_elem%conf(new_elem%mv_conf)%elem%stat
            IF (stat .NE. status_calculated) THEN
               stat = new_elem%conf(new_elem%mv_conf)%elem%stat
               IF (tmc_env%params%DRAW_TREE) &
                  CALL create_dot_color(tree_element=new_elem%conf(new_elem%mv_conf)%elem, &
                                        tmc_params=tmc_env%params)
            END IF
         CASE (status_cancel_ener, status_canceled_ener)
            ! configuration is already created,
            !  but energy has to be calculated (again)
            new_elem%conf(new_elem%mv_conf)%elem%stat = status_created
            new_elem%stat = status_created
            ! creation complete, handle energy calculation at a different position
            !  (for different worker group)
            stat = status_calculated
            IF (tmc_env%params%DRAW_TREE) &
               CALL create_dot_color(tree_element=new_elem%conf(new_elem%mv_conf)%elem, &
                                     tmc_params=tmc_env%params)
         CASE (status_cancel_nmc, status_canceled_nmc)
            ! reactivation canceled element (but with new global tree element)
            new_elem%conf(new_elem%mv_conf)%elem%stat = &
               status_calculate_NMC_steps
            new_elem%stat = status_calculate_NMC_steps
            stat = new_elem%conf(new_elem%mv_conf)%elem%stat
            reactivation_cc_count = reactivation_cc_count+1
            IF (tmc_env%params%DRAW_TREE) &
               CALL create_dot_color(tree_element=new_elem%conf(new_elem%mv_conf)%elem, &
                                     tmc_params=tmc_env%params)
         CASE DEFAULT
            CALL cp_abort(cp__l("tmc/tmc_tree_build.F",811), &
                          "unknown stat "// &
                          cp_to_string(new_elem%conf(new_elem%mv_conf)%elem%stat)// &
                          "of subtree element "// &
                          "for creating new gt element")
         END SELECT

         ! set stat TMC_STATUS_WAIT_FOR_NEW_TASK if no new calculation neccessary
         !   (energy calculation nodes searched by different routine)
         IF (stat .EQ. TMC_STATUS_FAILED) stat = TMC_STATUS_WAIT_FOR_NEW_TASK
         IF (stat .EQ. status_calculated) stat = TMC_STATUS_WAIT_FOR_NEW_TASK
      END IF
      ! end the timing
      CALL timestop(handle)

   END SUBROUTINE create_new_gt_tree_node

! **************************************************************************************************
!> \brief create new subtree element using pointer of global tree
!> \param act_gt_el global tree element
!> \param tmc_env ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_new_subtree_node(act_gt_el, tmc_env)
      TYPE(global_tree_type), POINTER                    :: act_gt_el
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_new_subtree_node', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: conf, handle, itmp
      LOGICAL                                            :: mv_rejected, new_subbox
      REAL(KIND=dp)                                      :: rnd
      TYPE(tree_type), POINTER                           :: new_elem, parent_elem

      NULLIFY (new_elem, parent_elem)

      IF(.NOT.(ASSOCIATED(act_gt_el)))CALL cp__a("tmc/tmc_tree_build.F",848)
      IF(.NOT.(ASSOCIATED(act_gt_el%conf(act_gt_el%mv_conf)%elem)))CALL cp__a("tmc/tmc_tree_build.F",849)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",850)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_tree_build.F",851)

      ! start the timing
      CALL timeset(routineN, handle)

      conf = act_gt_el%mv_conf
      CALL allocate_new_sub_tree_node(tmc_params=tmc_env%params, &
                                      next_el=new_elem, nr_dim=SIZE(act_gt_el%parent%conf(conf)%elem%pos))

      !-- node one level up
      parent_elem => act_gt_el%conf(conf)%elem
      new_elem%parent => parent_elem

      !-- set initial values
      parent_elem%next_elem_nr = parent_elem%next_elem_nr+1
      new_elem%nr = parent_elem%next_elem_nr
      new_elem%rng_seed = parent_elem%rng_seed

      !-- change to real parent element
      IF (act_gt_el%conf_n_acc(act_gt_el%conf(act_gt_el%mv_conf)%elem%sub_tree_nr)) THEN
         parent_elem%acc => new_elem
      ELSE
         parent_elem%nacc => new_elem
      END IF

      !-- real parent node (taking the configuration from)
      ! search parent
      parent_elem => search_parent_element(current=new_elem)
      new_elem%pos(:) = parent_elem%pos(:)
      new_elem%mol(:) = parent_elem%mol(:)
      new_elem%vel(:) = parent_elem%vel(:)
      new_elem%ekin = parent_elem%ekin
      new_elem%e_pot_approx = parent_elem%e_pot_approx
      new_elem%next_elem_nr => parent_elem%next_elem_nr
      new_elem%sub_tree_nr = parent_elem%sub_tree_nr
      new_elem%box_scale = parent_elem%box_scale
      IF (tmc_env%params%task_type .EQ. task_type_gaussian_adaptation) THEN
         new_elem%frc(:) = parent_elem%frc(:)
         new_elem%potential = parent_elem%potential
         new_elem%ekin_before_md = parent_elem%ekin_before_md
      ELSE
         new_elem%potential = 97589.0_dp
      END IF

      ! set new substream of random number generator
      CALL set_rng_stream(rng_stream=tmc_env%rng_stream, &
                          bg=new_elem%rng_seed(:, :, 1), &
                          cg=new_elem%rng_seed(:, :, 2), &
                          ig=new_elem%rng_seed(:, :, 3))
      CALL reset_to_next_rng_substream(tmc_env%rng_stream)

      ! set the temperature for the NMC moves
      rnd = next_random_number(tmc_env%rng_stream)
      IF (tmc_env%params%NMC_inp_file .NE. "") THEN
         new_elem%temp_created = INT(tmc_env%params%nr_temp*rnd)+1
      ELSE
         new_elem%temp_created = act_gt_el%mv_conf
      END IF

      ! rnd nr for selecting move
      rnd = next_random_number(tmc_env%rng_stream)
      !-- set move type
      new_elem%move_type = select_random_move_type( &
                           move_types=tmc_env%params%move_types, &
                           rnd=rnd)
      CALL get_rng_stream(rng_stream=tmc_env%rng_stream, &
                          bg=new_elem%rng_seed(:, :, 1), &
                          cg=new_elem%rng_seed(:, :, 2), &
                          ig=new_elem%rng_seed(:, :, 3))

      ! move is only done by the master,
      !  when standard MC moves with single potential are done
      ! the Nested Monte Carlo routine needs the old configuration
      !  to see if change is accepted
      SELECT CASE (new_elem%move_type)
      CASE (mv_type_MD)
         ! velocity change have to be done on workers,
         !  because of velocity change for NMC acceptance check
         new_elem%stat = status_calculate_MD
         ! set the temperature for creating MD
         new_elem%temp_created = act_gt_el%mv_conf
         !-- set the subbox (elements in subbox)
         CALL elements_in_new_subbox(tmc_params=tmc_env%params, &
                                     rng_stream=tmc_env%rng_stream, elem=new_elem, &
                                     nr_of_sub_box_elements=itmp)
         ! the move is performed on a worker group
      CASE (mv_type_NMC_moves)
         new_elem%stat = status_calculate_NMC_steps
         !-- set the subbox (elements in subbox)
         CALL elements_in_new_subbox(tmc_params=tmc_env%params, &
                                     rng_stream=tmc_env%rng_stream, elem=new_elem, &
                                     nr_of_sub_box_elements=itmp)
         ! the move is performed on a worker group
         ! the following moves new no force_env and can be performed on the master directly
      CASE (mv_type_atom_trans, mv_type_atom_swap, mv_type_mol_trans, &
            mv_type_mol_rot, mv_type_proton_reorder, &
            mv_type_volume_move)
         new_subbox = .TRUE.
         ! volume move on whole cell
         IF (new_elem%move_type .EQ. mv_type_volume_move) THEN
            new_subbox = .FALSE.
         END IF
         CALL change_pos(tmc_params=tmc_env%params, &
                         move_types=tmc_env%params%move_types, &
                         rng_stream=tmc_env%rng_stream, elem=new_elem, &
                         mv_conf=conf, new_subbox=new_subbox, &
                         move_rejected=mv_rejected)
         IF (mv_rejected) THEN
            new_elem%potential = HUGE(new_elem%potential)
            new_elem%e_pot_approx = HUGE(new_elem%e_pot_approx)
            new_elem%stat = status_calculated
         ELSE
            new_elem%stat = status_created
            IF (tmc_env%params%NMC_inp_file .NE. "") &
               new_elem%stat = status_calc_approx_ener
         END IF
      CASE (mv_type_gausian_adapt)
         ! still could be implemented
      CASE DEFAULT
         CALL cp_abort(cp__l("tmc/tmc_tree_build.F",970), &
                       "unknown move type ("//cp_to_string(new_elem%move_type)// &
                       "), while creating subtree element.")
      END SELECT
      act_gt_el%conf(act_gt_el%mv_conf)%elem => new_elem

      ! end the timing
      CALL timestop(handle)
      IF(.NOT.(ASSOCIATED(act_gt_el%conf(act_gt_el%mv_conf)%elem)))CALL cp__a("tmc/tmc_tree_build.F",978)
   END SUBROUTINE create_new_subtree_node

   !============================================================================
   ! tree node deallocation
   !============================================================================
! **************************************************************************************************
!> \brief prepares for deallocation of global tree element
!>        (checks status and set pointers of neighboring elements)
!> \param gt_ptr the global tree element
!> \param draw if present, changes the coleor in the dot file
!> \param tmc_env tmc environment
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE remove_gt_elem(gt_ptr, draw, tmc_env)
      TYPE(global_tree_type), POINTER                    :: gt_ptr
      LOGICAL, OPTIONAL                                  :: draw
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_gt_elem', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(gt_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1001)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",1002)

      ! start the timing
      CALL timeset(routineN, handle)

      CALL remove_gt_references(gt_ptr=gt_ptr, tmc_env=tmc_env)

      ! set status and draw in tree
      IF ((gt_ptr%stat .EQ. status_accepted_result) .OR. (gt_ptr%stat .EQ. status_rejected_result)) THEN
         gt_ptr%stat = status_deleted_result
      ELSE
         gt_ptr%stat = status_deleted
      END IF
      IF (tmc_env%params%DRAW_TREE .AND. PRESENT(draw)) &
         CALL create_global_tree_dot_color(gt_tree_element=gt_ptr, tmc_params=tmc_env%params)

      !remove pointer from tree parent
      IF (ASSOCIATED(gt_ptr%parent)) THEN
         IF (ASSOCIATED(gt_ptr%parent%acc, gt_ptr)) THEN
            gt_ptr%parent%acc => NULL()
         END IF
         IF (ASSOCIATED(gt_ptr%parent%nacc, gt_ptr)) THEN
            gt_ptr%parent%nacc => NULL()
         END IF
      END IF

      !remove pointer from tree childs
      IF (ASSOCIATED(gt_ptr%acc)) THEN
         gt_ptr%acc%parent => NULL()
      END IF

      IF (ASSOCIATED(gt_ptr%nacc)) THEN
         gt_ptr%nacc%parent => NULL()
      END IF

      CALL deallocate_global_tree_node(gt_elem=gt_ptr)
      ! end the timing
      CALL timestop(handle)

      IF(.NOT.(.NOT. ASSOCIATED(gt_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1041)
   END SUBROUTINE remove_gt_elem

! **************************************************************************************************
!> \brief prepares for deallocation of sub tree element
!>        (checks status and set pointers of neighboring elements)
!> \param ptr the sub tree element
!> \param draw if present, changes the coleor in the dot file
!> \param tmc_env tmc environment
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE remove_st_elem(ptr, draw, tmc_env)
      TYPE(tree_type), POINTER                           :: ptr
      LOGICAL, OPTIONAL                                  :: draw
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_st_elem', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: ready

      ready = .TRUE.
      IF(.NOT.(ASSOCIATED(ptr)))CALL cp__a("tmc/tmc_tree_build.F",1063)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",1064)

      ! start the timing
      CALL timeset(routineN, handle)

      ! if there is still e reference to a global tree pointer, do not deallocate element
      IF (ASSOCIATED(ptr%gt_nodes_references)) THEN
         IF (ASSOCIATED(ptr%parent)) &
            CALL cp_warn(cp__l("tmc/tmc_tree_build.F",1072), &
                         "try to deallocate subtree element"// &
                         cp_to_string(ptr%sub_tree_nr)//cp_to_string(ptr%nr)// &
                         " still with global tree element references e.g."// &
                         cp_to_string(ptr%gt_nodes_references%gt_elem%nr))
         IF(.NOT.(ASSOCIATED(ptr%gt_nodes_references%gt_elem)))CALL cp__a("tmc/tmc_tree_build.F",1077)
      ELSE
         SELECT CASE (ptr%stat)
            ! if element is still in progress, do not delete, wait for responding
         CASE (status_calculate_energy, &
               status_calculate_NMC_steps, status_calculate_MD)
            ! in case of speculative canceling: should be already canceled
            !  try to deallocate subtree element (still in progress)
            IF(.NOT.(tmc_env%params%SPECULATIVE_CANCELING))CALL cp__a("tmc/tmc_tree_build.F",1085)
         CASE (status_cancel_nmc, status_cancel_ener)
            ! do not return in case of finalizing (do not wait for canceling receipt)
            IF (PRESENT(draw)) ready = .FALSE.
         CASE DEFAULT
         END SELECT

         ! check if real top to bottom or bottom to top deallocation (no middle element deallocation)
         IF (ASSOCIATED(ptr%parent) .AND. &
             (ASSOCIATED(ptr%acc) .OR. ASSOCIATED(ptr%nacc))) THEN
            CALL cp__b("tmc/tmc_tree_build.F",1095,"")
         END IF

         IF (ready) THEN
            ! set status and draw in tree
            IF ((ptr%stat .EQ. status_accepted_result) .OR. &
                (ptr%stat .EQ. status_rejected_result)) THEN
               ptr%stat = status_deleted_result
            ELSE
               ptr%stat = status_deleted
            END IF
            IF (tmc_env%params%DRAW_TREE .AND. PRESENT(draw)) &
               CALL create_dot_color(tree_element=ptr, tmc_params=tmc_env%params)

            !remove pointer from tree parent
            IF (ASSOCIATED(ptr%parent)) THEN
               IF (ASSOCIATED(ptr%parent%acc, ptr)) ptr%parent%acc => NULL()
               IF (ASSOCIATED(ptr%parent%nacc, ptr)) ptr%parent%nacc => NULL()
            END IF

            !remove pointer from tree childs
            IF (ASSOCIATED(ptr%acc)) ptr%acc%parent => NULL()
            IF (ASSOCIATED(ptr%nacc)) ptr%nacc%parent => NULL()

            ! deallocate
            CALL deallocate_sub_tree_node(tree_elem=ptr)
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE remove_st_elem

! **************************************************************************************************
!> \brief deletes the no more used global tree nodes beside the result nodes
!>        from begin_ptr to end_ptr
!> \param begin_ptr start of the tree region to be cleaned
!> \param end_ptr end of the tree region to be cleaned
!> \param removed retun value if brance is clean
!> \param tmc_env tmc environment
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE remove_unused_g_tree(begin_ptr, end_ptr, removed, tmc_env)
      TYPE(global_tree_type), POINTER                    :: begin_ptr, end_ptr
      LOGICAL                                            :: removed
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_unused_g_tree', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: acc_removed, nacc_removed
      TYPE(global_tree_type), POINTER                    :: acc_ptr, nacc_ptr, tmp_ptr

      NULLIFY (acc_ptr, nacc_ptr, tmp_ptr)

      IF(.NOT.(ASSOCIATED(begin_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1150)
      IF(.NOT.(ASSOCIATED(end_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1151)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",1152)

      ! start the timing
      CALL timeset(routineN, handle)

      removed = .FALSE.
      acc_removed = .FALSE.
      nacc_removed = .FALSE.

      IF (.NOT. ASSOCIATED(begin_ptr, end_ptr)) THEN
         !-- go until the ends ot he tree, to deallocate revese
         !-- check if child nodes exist and possibly deallocate child node
         IF (ASSOCIATED(begin_ptr%acc)) THEN
            acc_ptr => begin_ptr%acc
            CALL remove_unused_g_tree(acc_ptr, end_ptr, acc_removed, tmc_env)
         ELSE
            acc_removed = .TRUE.
         END IF
         IF (ASSOCIATED(begin_ptr%nacc)) THEN
            nacc_ptr => begin_ptr%nacc
            CALL remove_unused_g_tree(nacc_ptr, end_ptr, nacc_removed, tmc_env)
         ELSE
            nacc_removed = .TRUE.
         END IF

         !-- deallocate node if no child node exist
         IF (acc_removed .AND. nacc_removed) THEN
            SELECT CASE (begin_ptr%stat)
            CASE (status_accepted, status_rejected, status_calculated, status_created, &
                  status_calculate_energy, status_calculate_MD, status_calculate_NMC_steps, status_calc_approx_ener, &
                  status_cancel_nmc, status_cancel_ener, status_canceled_nmc, status_canceled_ener)
               ! delete references, cancel elements calculation and deallocate global tree element
               tmp_ptr => begin_ptr

               CALL remove_gt_elem(gt_ptr=tmp_ptr, draw=.TRUE., tmc_env=tmc_env)
               IF (.NOT. ASSOCIATED(tmp_ptr)) removed = .TRUE.
            CASE (status_accepted_result, status_rejected_result)
            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_tree_build.F",1190), &
                             "try to dealloc unused tree element with status of begin element" &
                             //cp_to_string(begin_ptr%stat))
            END SELECT
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
      IF(.NOT.(ASSOCIATED(end_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1198)
   END SUBROUTINE remove_unused_g_tree

! **************************************************************************************************
!> \brief deletes the no more used sub tree nodes beside the result nodes
!>        from begin_ptr to end_ptr
!> \param begin_ptr start of the tree region to be cleaned
!> \param end_ptr end of the tree region to be cleaned
!> \param working_elem_list ...
!> \param removed retun value if brance is clean
!> \param tmc_env tmc environment
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE remove_unused_s_tree(begin_ptr, end_ptr, working_elem_list, &
                                             removed, tmc_env)
      TYPE(tree_type), POINTER                           :: begin_ptr
      TYPE(tree_type), INTENT(IN), POINTER               :: end_ptr
      TYPE(elem_array_type), DIMENSION(:), POINTER       :: working_elem_list
      LOGICAL                                            :: removed
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_unused_s_tree', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      LOGICAL                                            :: acc_removed, nacc_removed, remove_this
      TYPE(tree_type), POINTER                           :: acc_ptr, nacc_ptr, tmp_ptr

      NULLIFY (acc_ptr, nacc_ptr, tmp_ptr)
      remove_this = .FALSE.
      removed = .FALSE.
      acc_removed = .FALSE.
      nacc_removed = .FALSE.

      ! start the timing
      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(begin_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1235)
      IF(.NOT.(ASSOCIATED(end_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1236)
      IF(.NOT.(ASSOCIATED(working_elem_list)))CALL cp__a("tmc/tmc_tree_build.F",1237)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",1238)

      !-- if element is last checked in trajectory, go back
      IF (.NOT. ASSOCIATED(begin_ptr, end_ptr)) THEN
         !-- go until the ends on the tree, to deallocate revesely
         !-- check if child nodes exist and possibly deallocate child node
         IF (ASSOCIATED(begin_ptr%acc)) THEN
            acc_ptr => begin_ptr%acc
            CALL remove_unused_s_tree(acc_ptr, end_ptr, working_elem_list, &
                                      acc_removed, tmc_env)
         ELSE
            acc_removed = .TRUE.
         END IF
         IF (ASSOCIATED(begin_ptr%nacc)) THEN
            nacc_ptr => begin_ptr%nacc
            CALL remove_unused_s_tree(nacc_ptr, end_ptr, working_elem_list, &
                                      nacc_removed, tmc_env)
         ELSE
            nacc_removed = .TRUE.
         END IF

         !IF (DEBUG.GE.20) WRITE(tmc_out_file_nr,*)"try to dealloc: node", begin_ptr%nr," sides are removed: ", &
         !                                          acc_removed, nacc_removed

         !-- deallocate node if NO child node exist
         ! unused trajectory is deleted when cleaned part is updated
         IF (acc_removed .AND. nacc_removed) THEN
            SELECT CASE (begin_ptr%stat)
            CASE (status_canceled_nmc, status_canceled_ener)
               remove_this = .TRUE.
            CASE (status_accepted, status_rejected, status_calculated, &
                  status_accepted_result, status_rejected_result, status_created)
               remove_this = .TRUE.
               ! not to cancel, because still in progress
            CASE (status_calculate_energy, status_calculate_NMC_steps, &
                  status_calculate_MD, status_cancel_nmc, status_cancel_ener, &
                  status_calc_approx_ener)
               remove_this = .FALSE.
               ! -- delete when calculation is finished or aborted
               ! removed should still be .FALSE.
            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_tree_build.F",1279), &
                             "unknown status "//cp_to_string(begin_ptr%stat)// &
                             "of sub tree element "// &
                             cp_to_string(begin_ptr%sub_tree_nr)//" "// &
                             cp_to_string(begin_ptr%nr))
            END SELECT

            ! delete element
            IF (remove_this) THEN
               !-- mark as deleted and draw it in tree
               IF (.NOT. ASSOCIATED(begin_ptr%parent)) &
                  CALL cp_abort(cp__l("tmc/tmc_tree_build.F",1290), &
                                "try to remove unused subtree element "// &
                                cp_to_string(begin_ptr%sub_tree_nr)//" "// &
                                cp_to_string(begin_ptr%nr)// &
                                " but parent does not exist")
               tmp_ptr => begin_ptr
               ! check if a working group is still working on this element
               removed = .TRUE.
               DO i = 1, SIZE(working_elem_list(:))
                  IF (ASSOCIATED(working_elem_list(i)%elem)) THEN
                     IF (ASSOCIATED(working_elem_list(i)%elem, tmp_ptr)) &
                        removed = .FALSE.
                  END IF
               END DO
               IF (removed) THEN
                  !IF (DEBUG.GE.20) WRITE(tmc_out_file_nr,*)"deallocation of node ", begin_ptr%nr, "with status ", begin_ptr%stat
                  ! if all groups are finished with this element, we can deallocate
                  CALL remove_st_elem(ptr=tmp_ptr, draw=.TRUE., tmc_env=tmc_env)
                  IF (.NOT. ASSOCIATED(tmp_ptr)) THEN
                     removed = .TRUE.
                  ELSE
                     removed = .FALSE.
                  END IF
               END IF
            END IF
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE remove_unused_s_tree

! **************************************************************************************************
!> \brief deallocates all result nodes (remaining Markov Chain)
!>        from the tree root to the end of clean tree of the global tree
!> \param end_of_clean_tree ...
!> \param actual_ptr ...
!> \param tmc_env TMC environment for deallocation
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE remove_result_g_tree(end_of_clean_tree, actual_ptr, &
                                             tmc_env)
      TYPE(global_tree_type), POINTER                    :: end_of_clean_tree, actual_ptr
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_result_g_tree', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(global_tree_type), POINTER                    :: tmp_ptr

      IF(.NOT.(ASSOCIATED(end_of_clean_tree)))CALL cp__a("tmc/tmc_tree_build.F",1340)
      IF(.NOT.(ASSOCIATED(actual_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1341)

      ! start the timing
      CALL timeset(routineN, handle)

      !-- going up to the head ot the subtree
      IF (ASSOCIATED(actual_ptr%parent)) &
         CALL remove_result_g_tree(end_of_clean_tree=end_of_clean_tree, &
                                   actual_ptr=actual_ptr%parent, &
                                   tmc_env=tmc_env)
      !-- new tree head has no parent
      IF (.NOT. ASSOCIATED(actual_ptr, end_of_clean_tree)) THEN
         !-- deallocate node
         !IF(DEBUG.GE.20) WRITE(tmc_out_file_nr,*)"dealloc gt result tree element: ",actual_ptr%nr
         tmp_ptr => actual_ptr
         CALL remove_gt_elem(gt_ptr=tmp_ptr, draw=.TRUE., tmc_env=tmc_env)
         actual_ptr => tmp_ptr
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE remove_result_g_tree

! **************************************************************************************************
!> \brief deallocates all result nodes (remaining Markov Chain)
!>        from the tree root to the end of clean tree of one sub tree
!>        top to buttom deallocation
!> \param end_of_clean_tree ...
!> \param actual_ptr ...
!> \param tmc_env TMC environment for deallocation
!> \author Mandes 12.2012
! **************************************************************************************************
   RECURSIVE SUBROUTINE remove_result_s_tree(end_of_clean_tree, actual_ptr, &
                                             tmc_env)
      TYPE(tree_type), POINTER                           :: end_of_clean_tree, actual_ptr
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_result_s_tree', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(tree_type), POINTER                           :: tmp_ptr

      IF(.NOT.(ASSOCIATED(end_of_clean_tree)))CALL cp__a("tmc/tmc_tree_build.F",1383)
      IF(.NOT.(ASSOCIATED(actual_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1384)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",1385)

      ! start the timing
      CALL timeset(routineN, handle)

      !-- going up to the head ot the subtree
      IF (ASSOCIATED(actual_ptr%parent)) &
         CALL remove_result_s_tree(end_of_clean_tree, actual_ptr%parent, &
                                   tmc_env)

      !-- new tree head has no parent
      IF (.NOT. ASSOCIATED(actual_ptr, end_of_clean_tree)) THEN
         ! in trajectory just one direction should exist
         IF (ASSOCIATED(actual_ptr%acc) .AND. ASSOCIATED(actual_ptr%nacc)) THEN
            CALL cp__b("tmc/tmc_tree_build.F",1399,"")
         END IF
         ! the parent should be deleted already, but global tree is allocated to the second last accepted, &
         !   hence there could be still a reference to an element...
         IF (.NOT. ASSOCIATED(actual_ptr%parent)) THEN
            !-- deallocate node
            tmp_ptr => actual_ptr
            CALL remove_st_elem(ptr=tmp_ptr, draw=.TRUE., tmc_env=tmc_env)
            actual_ptr => tmp_ptr
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE remove_result_s_tree

! **************************************************************************************************
!> \brief deallocates the no more used tree nodes beside the result nodes
!>        from begin_ptr to end_ptr
!>        in global and subtrees
!> \param working_elem_list list of actual calculating elements for canceling
!> \param tmc_env TMC environment
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE remove_all_trees(working_elem_list, tmc_env)
      TYPE(elem_array_type), DIMENSION(:), POINTER       :: working_elem_list
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_all_trees', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, tree
      LOGICAL                                            :: change_trajec, flag
      TYPE(global_tree_type), POINTER                    :: tmp_gt_ptr
      TYPE(tree_type), POINTER                           :: last_acc_st_elem, tmp_ptr

      NULLIFY (last_acc_st_elem, tmp_ptr, tmp_gt_ptr)

      IF(.NOT.(ASSOCIATED(working_elem_list)))CALL cp__a("tmc/tmc_tree_build.F",1436)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",1437)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_tree_build.F",1438)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%gt_act)))CALL cp__a("tmc/tmc_tree_build.F",1439)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%gt_clean_end)))CALL cp__a("tmc/tmc_tree_build.F",1440)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%result_list)))CALL cp__a("tmc/tmc_tree_build.F",1441)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%st_clean_ends)))CALL cp__a("tmc/tmc_tree_build.F",1442)

      flag = .FALSE.
      change_trajec = .FALSE.

      ! start the timing
      CALL timeset(routineN, handle)

      !-- deallocate unused pt tree
      CALL remove_unused_g_tree(begin_ptr=tmc_env%m_env%gt_clean_end, &
                                end_ptr=tmc_env%m_env%gt_act, removed=flag, &
                                tmc_env=tmc_env)
      tmp_gt_ptr => tmc_env%m_env%gt_clean_end
      CALL search_end_of_clean_g_tree(last_acc=tmc_env%m_env%gt_clean_end, &
                                      tree_ptr=tmp_gt_ptr)
      !-- deallocate unused pt trajectory tree elements
      IF (tmc_env%params%USE_REDUCED_TREE) THEN
         tmp_gt_ptr => tmc_env%m_env%gt_clean_end
         CALL remove_result_g_tree(end_of_clean_tree=tmc_env%m_env%gt_clean_end, &
                                   actual_ptr=tmp_gt_ptr, tmc_env=tmc_env)

         !check if something changed, if not no deallocation of result subtree necessary
         IF (.NOT. ASSOCIATED(tmc_env%m_env%gt_head, tmc_env%m_env%gt_clean_end)) &
            change_trajec = .TRUE.
         tmc_env%m_env%gt_head => tmc_env%m_env%gt_clean_end
         IF(.NOT.(.NOT. ASSOCIATED(tmc_env%m_env%gt_head%parent)))CALL cp__a("tmc/tmc_tree_build.F",1467)
         !IF (DEBUG.GE.20) WRITE(tmc_out_file_nr,*)"new head of pt tree is ",tmc_env%m_env%gt_head%nr
      END IF

      !-- deallocate the subtrees
      ! do for all temperatures respectively all subtrees
      DO tree = 1, tmc_env%params%nr_temp
         ! get last checked element in trajectory related to the subtree (resultlist order is NOT subtree order)
         conf_loop: DO i = 1, SIZE(tmc_env%m_env%result_list)
            last_acc_st_elem => tmc_env%m_env%result_list(i)%elem
            IF (last_acc_st_elem%sub_tree_nr .EQ. tree) &
               EXIT conf_loop
         END DO conf_loop
         IF(.NOT.(last_acc_st_elem%sub_tree_nr .EQ. tree))CALL cp__a("tmc/tmc_tree_build.F",1480)
         CALL remove_unused_s_tree(begin_ptr=tmc_env%m_env%st_clean_ends(tree)%elem, &
                                   end_ptr=last_acc_st_elem, working_elem_list=working_elem_list, &
                                   removed=flag, tmc_env=tmc_env)
         CALL search_end_of_clean_tree(tree_ptr=tmc_env%m_env%st_clean_ends(tree)%elem, &
                                       last_acc=last_acc_st_elem)
      END DO
      !-- deallocate the trajectory subtree elements
      IF (tmc_env%params%USE_REDUCED_TREE .AND. change_trajec) THEN
         DO tree = 1, tmc_env%params%nr_temp
            tmp_ptr => tmc_env%m_env%st_clean_ends(tree)%elem
            IF(.NOT.(tmp_ptr%sub_tree_nr .EQ. tree))CALL cp__a("tmc/tmc_tree_build.F",1491)
            CALL remove_result_s_tree(end_of_clean_tree=tmc_env%m_env%st_clean_ends(tree)%elem, &
                                      actual_ptr=tmp_ptr, tmc_env=tmc_env)
            tmc_env%m_env%st_heads(tree)%elem => tmc_env%m_env%st_clean_ends(tree)%elem
            !IF(DEBUG.GE.20) &
            !  WRITE(tmc_out_file_nr,*)"new head of tree ",tree," is ",&
            !        tmc_env%m_env%st_heads(tree)%elem%nr
         END DO
      END IF

      ! end the timing
      CALL timestop(handle)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%gt_act)))CALL cp__a("tmc/tmc_tree_build.F",1503)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env%gt_clean_end)))CALL cp__a("tmc/tmc_tree_build.F",1504)
   END SUBROUTINE remove_all_trees

! **************************************************************************************************
!> \brief deallocates the whole global tree, to clean up
!> \param begin_ptr pointer to global tree head
!> \param removed flag, if the this element is removed
!> \param tmc_env ...
!> \author Mandes 01.2013
! **************************************************************************************************
   RECURSIVE SUBROUTINE dealloc_whole_g_tree(begin_ptr, removed, tmc_env)
      TYPE(global_tree_type), POINTER                    :: begin_ptr
      LOGICAL                                            :: removed
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dealloc_whole_g_tree', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: acc_removed, nacc_removed
      TYPE(global_tree_type), POINTER                    :: acc_ptr, nacc_ptr, tmp_ptr

      IF(.NOT.(ASSOCIATED(begin_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1525)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",1526)

      IF (ASSOCIATED(begin_ptr%acc)) THEN
         acc_ptr => begin_ptr%acc
         CALL dealloc_whole_g_tree(acc_ptr, acc_removed, tmc_env)
      ELSE
         acc_removed = .TRUE.
      END IF
      IF (ASSOCIATED(begin_ptr%nacc)) THEN
         nacc_ptr => begin_ptr%nacc
         CALL dealloc_whole_g_tree(nacc_ptr, nacc_removed, tmc_env)
      ELSE
         nacc_removed = .TRUE.
      END IF

      !-- deallocate node if no child node exist
      IF (acc_removed .AND. nacc_removed) THEN
         CALL search_and_remove_reference_in_list(gt_ptr=begin_ptr, &
                                                  elem=begin_ptr%conf(begin_ptr%mv_conf)%elem, tmc_env=tmc_env)
         tmp_ptr => begin_ptr
         CALL remove_gt_elem(gt_ptr=tmp_ptr, draw=.FALSE., tmc_env=tmc_env)
         !CALL deallocate_global_tree_node(gt_elem=tmp_ptr)
         removed = .TRUE.
      END IF
   END SUBROUTINE dealloc_whole_g_tree
! **************************************************************************************************
!> \brief deallocates the whole sub tree, to clean up
!> \param begin_ptr pointer to sub tree head
!> \param removed flag, if the this element is removed
!> \param tmc_params ...
!> \author Mandes 01.2013
! **************************************************************************************************
   RECURSIVE SUBROUTINE dealloc_whole_subtree(begin_ptr, removed, tmc_params)
      TYPE(tree_type), POINTER                           :: begin_ptr
      LOGICAL                                            :: removed
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dealloc_whole_subtree', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: acc_removed, nacc_removed
      TYPE(tree_type), POINTER                           :: acc_ptr, nacc_ptr, tmp_ptr

      IF(.NOT.(ASSOCIATED(begin_ptr)))CALL cp__a("tmc/tmc_tree_build.F",1569)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_tree_build.F",1570)

      IF (ASSOCIATED(begin_ptr%acc)) THEN
         acc_ptr => begin_ptr%acc
         CALL dealloc_whole_subtree(acc_ptr, acc_removed, tmc_params)
      ELSE
         acc_removed = .TRUE.
      END IF
      IF (ASSOCIATED(begin_ptr%nacc)) THEN
         nacc_ptr => begin_ptr%nacc
         CALL dealloc_whole_subtree(nacc_ptr, nacc_removed, tmc_params)
      ELSE
         nacc_removed = .TRUE.
      END IF

      !-- deallocate node if no child node exist
      IF (acc_removed .AND. nacc_removed) THEN
         tmp_ptr => begin_ptr
         CALL deallocate_sub_tree_node(tree_elem=begin_ptr)
         removed = .TRUE.
      END IF
   END SUBROUTINE dealloc_whole_subtree

   !============================================================================
   ! finalizing module (deallocating everything)
   !============================================================================
! **************************************************************************************************
!> \brief deallocating every tree node of every trees (clean up)
!> \param tmc_env TMC environment structure
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE finalize_trees(tmc_env)
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'finalize_trees', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: flag
      TYPE(global_tree_type), POINTER                    :: global_tree

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_build.F",1610)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_tree_build.F",1611)

      global_tree => tmc_env%m_env%gt_act
      !-- deallocate pt tree
      ! start with searching the head
      DO WHILE (ASSOCIATED(global_tree%parent))
         global_tree => global_tree%parent
      END DO
      CALL dealloc_whole_g_tree(begin_ptr=global_tree, removed=flag, &
                                tmc_env=tmc_env)

      !-- deallocate subtrees
      trees_loop: DO i = 1, SIZE(tmc_env%m_env%st_clean_ends(:))
         DO WHILE (ASSOCIATED(tmc_env%m_env%st_clean_ends(i)%elem%parent))
            tmc_env%m_env%st_clean_ends(i)%elem => &
               tmc_env%m_env%st_clean_ends(i)%elem%parent
         END DO
         CALL dealloc_whole_subtree(begin_ptr=tmc_env%m_env%st_clean_ends(i)%elem, &
                                    removed=flag, tmc_params=tmc_env%params)
      END DO trees_loop
      DEALLOCATE (tmc_env%params%atoms)
   END SUBROUTINE finalize_trees

END MODULE tmc_tree_build
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief tree nodes acceptance
!>        code is seperated in 3 parts, first the acceptance criteria,
!>        second the tree node acceptance handling, searching etc. and
!>        than the acceptance probability handling
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_tree_acceptance
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE physcon,                         ONLY: boltzmann,&
                                              joule
   USE tmc_calculations,                ONLY: compute_estimated_prob,&
                                              get_scaled_cell
   USE tmc_dot_tree,                    ONLY: create_dot_color,&
                                              create_global_tree_dot_color
   USE tmc_file_io,                     ONLY: write_result_list_element
   USE tmc_move_handle,                 ONLY: prob_update
   USE tmc_move_types,                  ONLY: mv_type_MD,&
                                              mv_type_volume_move
   USE tmc_stati,                       ONLY: task_type_gaussian_adaptation
   USE tmc_tree_build,                  ONLY: remove_unused_g_tree
   USE tmc_tree_search,                 ONLY: get_subtree_elements_to_check,&
                                              search_canceling_elements,&
                                              search_next_gt_element_to_check,&
                                              search_parent_element
   USE tmc_tree_types,                  ONLY: &
        add_to_list, global_tree_type, gt_elem_list_type, status_accepted, status_accepted_result, &
        status_calc_approx_ener, status_calculate_MD, status_calculate_NMC_steps, &
        status_calculate_energy, status_calculated, status_cancel_ener, status_cancel_nmc, &
        status_canceled_ener, status_canceled_nmc, status_created, status_deleted, &
        status_deleted_result, status_rejected, status_rejected_result, tree_type
   USE tmc_types,                       ONLY: tmc_env_type,&
                                              tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_tree_acceptance'

   PUBLIC :: acceptance_check
   PUBLIC :: check_acceptance_of_depending_subtree_nodes, &
             check_elements_for_acc_prob_update
   PUBLIC :: tree_update

   INTEGER, PARAMETER :: DEBUG = 0

CONTAINS

   !============================================================================
   ! acceptance criteria calculations
   !============================================================================
! **************************************************************************************************
!> \brief standard Monte Carlo and 2 potential acceptance check
!>        acceptance check of move from old(last accepted) to new configuration
!>        the sum of kinetic and potential energy is used
!>        acc(o->n)=min(1,exp( -beta*(H(n)-H(o)) ))
!> \param tree_element new/actual configuration
!> \param parent_element last accepted configuration
!> \param tmc_params TMC global parameters
!> \param temperature actual temperature configuration should be checked with
!> \param diff_pot_check 2potential check or not?
!> \param accept result (configuration accepted of rejected)
!> \param rnd_nr random number for acceptance check
!> \param approx_ener for NMC the approximated energies schould be used
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE acceptance_check(tree_element, parent_element, tmc_params, &
                               temperature, diff_pot_check, accept, rnd_nr, &
                               approx_ener)
      TYPE(tree_type), POINTER                           :: tree_element, parent_element
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      REAL(KIND=dp)                                      :: temperature
      LOGICAL                                            :: diff_pot_check, accept
      REAL(KIND=dp)                                      :: rnd_nr
      LOGICAL, OPTIONAL                                  :: approx_ener

      CHARACTER(LEN=*), PARAMETER :: routineN = 'acceptance_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: ekin_last_acc, elem_ener, kB, parent_ener

      IF(.NOT.(ASSOCIATED(tree_element)))CALL cp__a("tmc/tmc_tree_acceptance.F",94)
      IF(.NOT.(ASSOCIATED(parent_element)))CALL cp__a("tmc/tmc_tree_acceptance.F",95)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_tree_acceptance.F",96)
      IF(.NOT.(temperature .GT. 0.0_dp))CALL cp__a("tmc/tmc_tree_acceptance.F",97)
      IF(.NOT.(rnd_nr .GE. 0.0_dp .AND. rnd_nr .LE. 1.0_dp))CALL cp__a("tmc/tmc_tree_acceptance.F",98)

      kB = boltzmann/joule

      ! start the timing
      CALL timeset(routineN, handle)

      IF (tmc_params%task_type .EQ. task_type_gaussian_adaptation) THEN
         CALL cp__b("tmc/tmc_tree_acceptance.F",106,"")
!TODO       CALL acc_check_gauss_adapt(f=tree_element%potential, ct=tree_element%ekin_before_md, acc=accept)
!DO NOT DO       RETURN
      END IF

      !-- using 2 different potentials, the energy difference between these potentials
      !   and the two configurations have to be regarded.
      !   The ensamble should be in equilibrium state of the approximate potential
      IF (diff_pot_check .AND. (tmc_params%NMC_inp_file .NE. "")) THEN
         IF ((tree_element%potential .EQ. HUGE(tree_element%potential)) .OR. &
             tree_element%e_pot_approx .EQ. HUGE(tree_element%e_pot_approx)) THEN
            elem_ener = HUGE(elem_ener)
         ELSE
            !for different potentials we have to regard the differences in energy
            ! min(1,e^{-\beta*[(E_{exact}(n)-E_{approx}(n))-(E_{exact}(o)-E_{approx}(o))]})
            elem_ener = 1.0_dp/(kB*temperature)*tree_element%potential &
                        -1.0_dp/(kB*tmc_params%Temp(tree_element%temp_created)) &
                        *tree_element%e_pot_approx
         END IF
         parent_ener = 1.0_dp/(kB*temperature)*parent_element%potential &
                       -1.0_dp/(kB*tmc_params%Temp(tree_element%temp_created)) &
                       *parent_element%e_pot_approx

         !-- always accepted if new energy is smaller than old energy
         IF (elem_ener .LE. parent_ener) THEN
            accept = .TRUE.
         ELSE
            !-- gaussian distributed acceptance if new energy is greater than old energy
            IF (rnd_nr .LT. &
                EXP(-(elem_ener-parent_ener))) THEN
               accept = .TRUE.
            ELSE
               accept = .FALSE.
            END IF
         END IF
      ELSE
         !-- standard MC check, but also using the kinetic energy for Hybrid Monte Carlo
         IF (tree_element%move_type .EQ. mv_type_MD) THEN
            ekin_last_acc = tree_element%ekin_before_md
            ! using the kinetic energy before velocity change
         ELSE
            ekin_last_acc = parent_element%ekin
         END IF
         ! comparing aproximated energies
         IF (PRESENT(approx_ener)) THEN
            elem_ener = tree_element%e_pot_approx &
                        +tree_element%ekin
            parent_ener = parent_element%e_pot_approx &
                          +ekin_last_acc
         ELSE
            elem_ener = tree_element%potential &
                        +tree_element%ekin
            parent_ener = parent_element%potential &
                          +ekin_last_acc
         END IF

         !-- always accepted if new energy is smaller than old energy
         IF (elem_ener .LE. parent_ener) THEN
            accept = .TRUE.
         ELSE
            !-- gaussian distributed acceptance if new energy is greater than old energy
            IF (rnd_nr .LT. &
                EXP(-1.0_dp/(kB*temperature)*(elem_ener-parent_ener))) THEN
               accept = .TRUE.
            ELSE
               accept = .FALSE.
            END IF
         END IF
      END IF

      ! update the estimated energy acceptance probability distribution
      IF (diff_pot_check) THEN
         IF(.NOT.(ASSOCIATED(tmc_params%prior_NMC_acc)))CALL cp__a("tmc/tmc_tree_acceptance.F",178)
         tmc_params%prior_NMC_acc%counter = tmc_params%prior_NMC_acc%counter+1
         tmc_params%prior_NMC_acc%aver = (tmc_params%prior_NMC_acc%aver*(tmc_params%prior_NMC_acc%counter-1)+ &
                                          ((elem_ener-parent_ener)))/REAL(tmc_params%prior_NMC_acc%counter, KIND=dp)
         tmc_params%prior_NMC_acc%aver_2 = (tmc_params%prior_NMC_acc%aver_2*(tmc_params%prior_NMC_acc%counter-1)+ &
                                            (elem_ener-parent_ener)**2)/REAL(tmc_params%prior_NMC_acc%counter, KIND=dp)
      END IF

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE acceptance_check

! **************************************************************************************************
!> \brief parallel tempering swap acceptance check,
!>        check swapped configurations of different temperatures
!>        using sum of potential and kinetic energy
!>        always accepted if energy of configuration with higher temperature
!>        is smaller than energy of configuration with lower temperature
!>        acc(o->n)=min(1,exp((beta_i-beta_j)*(U_i-U_j)))
!> \param tree_elem actual global tree element
!> \param conf1 sub tree element of lower temperature
!> \param conf2 sub tree element of higher temperature
!> \param tmc_params TMC environment parameters
!> \param accept acceptance of configurational change
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE swap_acceptance_check(tree_elem, conf1, conf2, tmc_params, accept)
      TYPE(global_tree_type), POINTER                    :: tree_elem
      TYPE(tree_type), POINTER                           :: conf1, conf2
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      LOGICAL                                            :: accept

      CHARACTER(LEN=*), PARAMETER :: routineN = 'swap_acceptance_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: delta, elem1_ener, elem2_ener, kB, vol1, &
                                                            vol2

      kB = boltzmann/joule

      IF(.NOT.(ASSOCIATED(tree_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",219)
      IF(.NOT.(tree_elem%rnd_nr .GE. 0.0_dp))CALL cp__a("tmc/tmc_tree_acceptance.F",220)
      IF(.NOT.(ASSOCIATED(conf1)))CALL cp__a("tmc/tmc_tree_acceptance.F",221)
      IF(.NOT.(ASSOCIATED(conf2)))CALL cp__a("tmc/tmc_tree_acceptance.F",222)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_tree_acceptance.F",223)

      ! start the timing
      CALL timeset(routineN, handle)

      IF (tmc_params%pressure .GT. 0.0_dp) THEN
         ! pt-NVT
         elem1_ener = conf1%potential &
                      +conf1%ekin
         elem2_ener = conf2%potential &
                      +conf2%ekin
         ! the swap is done with prob: exp((\beta_i-\beta_j)(U_i-U_j)),
         ! BUT because they are already swaped we exchange the energies.
         IF (tree_elem%rnd_nr .LT. EXP((1.0_dp/(kB*tmc_params%Temp(tree_elem%mv_conf))- &
                                        1.0_dp/(kB*tmc_params%Temp(tree_elem%mv_conf+1))) &
                                       *(elem2_ener-elem1_ener))) THEN
            accept = .TRUE.
         ELSE
            accept = .FALSE.
         END IF
      ELSE
         ! pt-NpT (parallel Tempering with constant pressure, temperature and num. particle)
         CALL get_scaled_cell(cell=tmc_params%cell, box_scale=conf1%box_scale, &
                              vol=vol1)
         CALL get_scaled_cell(cell=tmc_params%cell, box_scale=conf2%box_scale, &
                              vol=vol2)
         ! delta= (beta_m-beta_n)(U_n-U_m) + (beta_m*P_m-beta_n*P_n)(V_n-V_m)
         delta = (1.0_dp/(kB*tmc_params%Temp(tree_elem%mv_conf)) &
                  -1.0_dp/(kB*tmc_params%Temp(tree_elem%mv_conf+1)))* &
                 ((conf2%potential+conf2%ekin)-(conf1%potential+conf1%ekin))+ &
                 (1.0_dp/(kB*tmc_params%Temp(tree_elem%mv_conf))*tmc_params%pressure &
                  -1.0_dp/(kB*tmc_params%Temp(tree_elem%mv_conf+1))*tmc_params%pressure)* &
                 (vol2-vol1)
         IF (tree_elem%rnd_nr .LT. EXP(delta)) THEN
            accept = .TRUE.
         ELSE
            accept = .FALSE.
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE swap_acceptance_check

! **************************************************************************************************
!> \brief volume move acceptance check
!>        sampling NPT, we need to do volume moves. Their acceptance are
!>        checked regarding the old and new energy, the volume difference
!>        and the constant pressure
!> \param parent_elem old tree element with old box size
!> \param new_elem actual tree element with new box size
!> \param temperature actual temperature
!> \param rnd_nr random number to check with
!> \param tmc_params TMC environment parameters
!> \param accept Monte Carlo move acceptance (result)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE volume_acceptance_check(parent_elem, new_elem, temperature, &
                                      rnd_nr, tmc_params, accept)
      TYPE(tree_type), POINTER                           :: parent_elem, new_elem
      REAL(KIND=dp)                                      :: temperature, rnd_nr
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      LOGICAL                                            :: accept

      CHARACTER(LEN=*), PARAMETER :: routineN = 'volume_acceptance_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: d_enthalpy, kB, n_vol, p_vol

      kB = boltzmann/joule

      IF(.NOT.(ASSOCIATED(parent_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",294)
      IF(.NOT.(ASSOCIATED(new_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",295)
      IF(.NOT.(temperature .GT. 0.0_dp))CALL cp__a("tmc/tmc_tree_acceptance.F",296)
      IF(.NOT.(rnd_nr .GT. 0.0_dp))CALL cp__a("tmc/tmc_tree_acceptance.F",297)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_tree_acceptance.F",298)
      IF(.NOT.(tmc_params%pressure .GE. 0.0_dp))CALL cp__a("tmc/tmc_tree_acceptance.F",299)

      ! start the timing
      CALL timeset(routineN, handle)

      CALL get_scaled_cell(cell=tmc_params%cell, box_scale=parent_elem%box_scale, &
                           vol=p_vol)
      CALL get_scaled_cell(cell=tmc_params%cell, box_scale=new_elem%box_scale, &
                           vol=n_vol)
      ! H=enthalpy, U=energy, P=pressure, V=volume, kB=Boltzmann constant, T=temperature, N=amount particle
      ! delta_H  =      delta_U                               + P*delta_V           - kB*T*N*ln(V_n/V_p)
      IF (.FALSE.) THEN
         ! the volume move in volume space (dV)
         d_enthalpy = (new_elem%potential-parent_elem%potential)+ &
                      tmc_params%pressure*(n_vol-p_vol)- &
                      kB*temperature*(SIZE(new_elem%pos)/ &
                                      tmc_params%dim_per_elem)* &
                      LOG(n_vol/p_vol)
      ELSE
         IF (tmc_params%v_isotropic) THEN
            d_enthalpy = (new_elem%potential-parent_elem%potential)+ &
                         tmc_params%pressure*(n_vol-p_vol)- &
                         kB*temperature*((SIZE(new_elem%pos)/ &
                                          tmc_params%dim_per_elem)+2/REAL(3, KIND=dp))* &
                         LOG(n_vol/p_vol)
         ELSE
            d_enthalpy = (new_elem%potential-parent_elem%potential)+ &
                         tmc_params%pressure*(n_vol-p_vol)- &
                         kB*temperature*(SIZE(new_elem%pos)/ &
                                         tmc_params%dim_per_elem)* &
                         LOG(n_vol/p_vol)
         END IF
      END IF
      ! acc(o->n) = min(1, exp(-beta*delta_H))
      IF (d_enthalpy .LE. 0.0_dp) THEN
         accept = .TRUE.
      ELSE
         IF (rnd_nr .LT. EXP(-1.0_dp/(kB*temperature)*d_enthalpy)) THEN
            accept = .TRUE.
         ELSE
            accept = .FALSE.
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE volume_acceptance_check

   !============================================================================
   ! tree node acceptance
   !============================================================================
! **************************************************************************************************
!> \brief check acceptance of energy calculated element and related childs,
!>        when ready
!> \param tree_elem actual tree element with calculated energy
!> \param tmc_env TMC environment parameters
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE check_acceptance_of_depending_subtree_nodes(tree_elem, tmc_env)
      TYPE(tree_type), POINTER                           :: tree_elem
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_acceptance_of_depending_subtree_nodes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: parent_ready
      TYPE(tree_type), POINTER                           :: act_elem, parent_elem

      NULLIFY (parent_elem, act_elem)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_acceptance.F",369)
      IF(.NOT.(ASSOCIATED(tree_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",370)
      IF(.NOT.(tree_elem%stat .EQ. status_calculated))CALL cp__a("tmc/tmc_tree_acceptance.F",371)
      IF(.NOT.(ASSOCIATED(tree_elem%parent)))CALL cp__a("tmc/tmc_tree_acceptance.F",372)

      ! start the timing
      CALL timeset(routineN, handle)

      act_elem => tree_elem

      ! ------------------------------------------------------
      ! check this element
      parent_elem => search_parent_element(act_elem)
      IF(.NOT.(.NOT. ASSOCIATED(act_elem, parent_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",382)

      ! check status of parent element
      SELECT CASE (parent_elem%stat)
      CASE (status_created, status_calculate_energy, status_calculate_MD, &
            status_calculate_NMC_steps, status_canceled_ener, &
            status_canceled_nmc, status_cancel_nmc, status_cancel_ener)
         parent_ready = .FALSE.
      CASE (status_accepted_result, status_rejected_result, &
            status_accepted, status_rejected, status_calculated)
         parent_ready = .TRUE.
      CASE DEFAULT
         CALL cp__b("tmc/tmc_tree_acceptance.F",394,"unknown parent element status"//cp_to_string(parent_elem%stat))
      END SELECT

      ! ready ?
      IF (parent_ready) THEN
         ! acceptance check
         CALL check_and_change_status_of_subtree_elem(act_elem=act_elem, &
                                                      parent_elem=parent_elem, tmc_env=tmc_env)
      END IF
      !------------------------------------------------------
      ! check acc child
      parent_elem => tree_elem
      IF (ASSOCIATED(tree_elem%acc)) THEN
         act_elem => tree_elem%acc
         IF (act_elem%stat .EQ. status_calculated) THEN
            CALL check_and_change_status_of_subtree_elem(act_elem=act_elem, &
                                                         parent_elem=parent_elem, tmc_env=tmc_env)
         END IF
         !------------------------------------------------------
         ! check all nacc childs
         nacc_loop: DO
            IF (.NOT. ASSOCIATED(act_elem%nacc)) EXIT nacc_loop
            act_elem => act_elem%nacc
            IF (act_elem%stat .EQ. status_calculated) THEN
               CALL check_and_change_status_of_subtree_elem(act_elem=act_elem, &
                                                            parent_elem=parent_elem, tmc_env=tmc_env)
            END IF
         END DO nacc_loop
      END IF

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE check_acceptance_of_depending_subtree_nodes

! **************************************************************************************************
!> \brief checking the elements and change the status
!> \param act_elem actual tree element (new configuration)
!> \param parent_elem parent tree element (old configuration)
!> \param tmc_env TMC environment parameters
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE check_and_change_status_of_subtree_elem(act_elem, parent_elem, tmc_env)
      TYPE(tree_type), POINTER                           :: act_elem, parent_elem
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_and_change_status_of_subtree_elem', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: flag, result_acc
      TYPE(global_tree_type), POINTER                    :: tmp_gt_ptr
      TYPE(gt_elem_list_type), POINTER                   :: tmp_gt_list_ptr

      NULLIFY (tmp_gt_list_ptr, tmp_gt_ptr)

      ! start the timing
      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_acceptance.F",452)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_tree_acceptance.F",453)
      IF(.NOT.(ASSOCIATED(act_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",454)
      IF(.NOT.(ASSOCIATED(parent_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",455)
      IF(.NOT.(act_elem%stat .EQ. status_calculated))CALL cp__a("tmc/tmc_tree_acceptance.F",456)

      flag = .FALSE.

      tmp_gt_list_ptr => act_elem%gt_nodes_references
      ! check for all global tree elements refering to this subtree element
      ! same subtree element can be accepted at a certain temperature and rejected at another one
      DO WHILE (ASSOCIATED(tmp_gt_list_ptr))
         IF(.NOT.(tmp_gt_list_ptr%gt_elem%stat .NE. status_accepted_result))CALL cp__a("tmc/tmc_tree_acceptance.F",464)
         IF(.NOT.(tmp_gt_list_ptr%gt_elem%stat .NE. status_rejected_result))CALL cp__a("tmc/tmc_tree_acceptance.F",465)

         CALL check_elements(gt_act_elem=tmp_gt_list_ptr%gt_elem, tmc_env=tmc_env, &
                             check_done=flag, result_acc=result_acc)
         IF (flag) THEN
            ! probability update
            CALL prob_update(move_types=tmc_env%params%move_types, elem=act_elem, &
                             acc=result_acc, prob_opt=tmc_env%params%esimate_acc_prob)

            ! change status
            ! accepted case: delete (and cancel) not accepted branch
            IF (result_acc .EQV. .TRUE.) THEN
               tmp_gt_list_ptr%gt_elem%stat = status_accepted
               IF (ASSOCIATED(tmp_gt_list_ptr%gt_elem%nacc)) THEN
                  tmp_gt_ptr => tmp_gt_list_ptr%gt_elem%nacc
                  CALL remove_unused_g_tree(begin_ptr=tmp_gt_ptr, &
                                            end_ptr=tmp_gt_list_ptr%gt_elem, removed=flag, &
                                            tmc_env=tmc_env)
               END IF
            ELSE
               ! not accepted case: delete (and cancel) accepted branch
               tmp_gt_list_ptr%gt_elem%stat = status_rejected
               IF (ASSOCIATED(tmp_gt_list_ptr%gt_elem%acc)) THEN
                  tmp_gt_ptr => tmp_gt_list_ptr%gt_elem%acc
                  CALL remove_unused_g_tree(begin_ptr=tmp_gt_ptr, &
                                            end_ptr=tmp_gt_list_ptr%gt_elem, removed=flag, &
                                            tmc_env=tmc_env)
               END IF
            END IF

            IF (tmc_env%params%DRAW_TREE) &
               CALL create_global_tree_dot_color(gt_tree_element=tmp_gt_list_ptr%gt_elem, &
                                                 tmc_params=tmc_env%params)
         END IF
         tmp_gt_list_ptr => tmp_gt_list_ptr%next
      END DO
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE check_and_change_status_of_subtree_elem

! **************************************************************************************************
!> \brief change status flag of all subtree elements
!>        most important to draw the right subtrees
!> \param gt_ptr global tree pointer, the related/changed sub tree element
!>        should change status
!> \param stat the status of the global tree element
!> \param tmc_params TMC emvironment parameters for drawing
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE subtree_configuration_stat_change(gt_ptr, stat, tmc_params)
      TYPE(global_tree_type), POINTER                    :: gt_ptr
      INTEGER                                            :: stat
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'subtree_configuration_stat_change', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(tree_type), POINTER                           :: ptr

      NULLIFY (ptr)

      IF(.NOT.(ASSOCIATED(gt_ptr)))CALL cp__a("tmc/tmc_tree_acceptance.F",527)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_tree_acceptance.F",528)

      ! start the timing
      CALL timeset(routineN, handle)

      ! don't respect swaped nodes (subtree nodes are already in the list,
      !   and we dont want them in marked in subtrees)
      IF (.NOT. gt_ptr%swaped) THEN
         ptr => gt_ptr%conf(gt_ptr%mv_conf)%elem
         ! dependent on the status (don't map each status 1:1,
         !   e.g. not the result ones)
         SELECT CASE (stat)
         CASE (status_accepted_result)
            ptr%stat = status_accepted
         CASE (status_rejected_result)
            ptr%stat = status_rejected
         CASE (status_accepted, status_rejected)
            ptr%stat = stat
         CASE DEFAULT
            CALL cp_abort(cp__l("tmc/tmc_tree_acceptance.F",547), &
                          "unknown global tree status"// &
                          cp_to_string(stat))
         END SELECT

         IF (tmc_params%DRAW_TREE) &
            CALL create_dot_color(tree_element=ptr, tmc_params=tmc_params)
      END IF

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE subtree_configuration_stat_change

! **************************************************************************************************
!> \brief checks if the element is ready for an acceptance check
!> \param elem sub tree element
!> \param ready return value
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE elem_ready_to_check(elem, ready)
      TYPE(tree_type), POINTER                           :: elem
      LOGICAL                                            :: ready

      CHARACTER(LEN=*), PARAMETER :: routineN = 'elem_ready_to_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",575)

      ! start the timing
      CALL timeset(routineN, handle)

      ready = .FALSE.

      SELECT CASE (elem%stat)
      CASE (status_created, status_calculate_energy, &
            status_calculate_MD, status_calculate_NMC_steps, status_calc_approx_ener, &
            status_cancel_nmc, status_cancel_ener, status_canceled_nmc, status_canceled_ener)
         ready = .FALSE.
      CASE (status_calculated, status_accepted_result, status_accepted, status_rejected, status_rejected_result)
         ready = .TRUE.
      CASE DEFAULT
         CALL cp_abort(cp__l("tmc/tmc_tree_acceptance.F",590), &
                       "status of actual tree node is unknown" &
                       //cp_to_string(elem%stat))
      END SELECT
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE elem_ready_to_check

! **************************************************************************************************
!> \brief checking the elements (find element and do acceptance check)
!> \param gt_act_elem actual global tree element
!> \param tmc_env TMC environment
!> \param check_done successful checked? (result)
!> \param result_acc checked configuration accepted? (result)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE check_elements(gt_act_elem, tmc_env, check_done, result_acc)
      TYPE(global_tree_type), POINTER                    :: gt_act_elem
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      LOGICAL                                            :: check_done, result_acc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_elements', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: act_ready, tmp_ready
      TYPE(tree_type), POINTER                           :: act_element, tmp_element

      NULLIFY (act_element, tmp_element)
      check_done = .FALSE.

      IF(.NOT.(ASSOCIATED(gt_act_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",620)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_acceptance.F",621)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_tree_acceptance.F",622)

      ! start the timing
      CALL timeset(routineN, handle)

      !====================================================================================
      ! check if global tree element is already checked
      SELECT CASE (gt_act_elem%stat)
      CASE (status_accepted, status_rejected)
         check_done = .TRUE.
         IF (gt_act_elem%stat .EQ. status_accepted) THEN
            result_acc = .TRUE.
         ELSE IF (gt_act_elem%stat .EQ. status_rejected) THEN
            result_acc = .FALSE.
         ELSE
            CALL cp_abort(cp__l("tmc/tmc_tree_acceptance.F",637), &
                          "undefinite status of already checked elements:" &
                          //cp_to_string(gt_act_elem%stat))
         END IF
      CASE DEFAULT
      END SELECT

      IF (.NOT. check_done) THEN
         !====================================================================================
         ! get elements related to global tree element to check
         CALL get_subtree_elements_to_check(gt_act_elem=gt_act_elem, elem1=act_element, &
                                            elem2=tmp_element)
         IF(.NOT.(ASSOCIATED(act_element)))CALL cp__a("tmc/tmc_tree_acceptance.F",649)
         IF(.NOT.(ASSOCIATED(tmp_element)))CALL cp__a("tmc/tmc_tree_acceptance.F",650)

         ! check status of both elements (if they are already calculated and hence ready to check)
         CALL elem_ready_to_check(elem=act_element, ready=act_ready)
         CALL elem_ready_to_check(elem=tmp_element, ready=tmp_ready)

         ! both ready ? check
         IF (act_ready .AND. tmp_ready) THEN
            ! 2 temperature (swap) check
            IF (gt_act_elem%swaped) THEN
               CALL swap_acceptance_check(tree_elem=gt_act_elem, conf1=act_element, &
                                          conf2=tmp_element, tmc_params=tmc_env%params, &
                                          accept=result_acc)
               ! volume move check
            ELSE IF (act_element%move_type .EQ. mv_type_volume_move) THEN
               CALL volume_acceptance_check(parent_elem=tmp_element, new_elem=act_element, &
                                            temperature=tmc_env%params%Temp(gt_act_elem%mv_conf), &
                                            rnd_nr=gt_act_elem%rnd_nr, &
                                            tmc_params=tmc_env%params, accept=result_acc)
            ELSE
               IF (tmc_env%m_env%temp_decrease .NE. 1.0_dp) THEN
                  CALL acceptance_check(tree_element=act_element, parent_element=tmp_element, &
                                        tmc_params=tmc_env%params, temperature=gt_act_elem%Temp, &
                                        diff_pot_check=.TRUE., accept=result_acc, &
                                        rnd_nr=gt_act_elem%rnd_nr)
               ELSE
                  CALL acceptance_check(tree_element=act_element, parent_element=tmp_element, &
                                        tmc_params=tmc_env%params, &
                                        temperature=tmc_env%params%Temp(gt_act_elem%mv_conf), &
                                        diff_pot_check=.TRUE., accept=result_acc, &
                                        rnd_nr=gt_act_elem%rnd_nr)
               END IF
            END IF
            check_done = .TRUE.
         ELSE
            ! if element is not ready, update status of global tree element
            !TODO maybe update this part (how much status we need in global tree from sub tree??)
            SELECT CASE (gt_act_elem%stat)
               !-- check if at least the MD move is already done
               !-- hence new configurations can be created on basis of this configuration
            CASE (status_calculate_MD, status_calculate_NMC_steps, status_calculate_energy, &
                  status_created, status_calc_approx_ener)
               IF (gt_act_elem%conf(gt_act_elem%mv_conf)%elem%stat .NE. gt_act_elem%stat) &
                  gt_act_elem%stat = gt_act_elem%conf(gt_act_elem%mv_conf)%elem%stat
            CASE (status_calculated)
            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_tree_acceptance.F",696), &
                             "status of actual checked node is unknown" &
                             //cp_to_string(gt_act_elem%stat))
            END SELECT
         END IF
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE check_elements

! **************************************************************************************************
!> \brief searching tree nodes to check for Markov Chain, elements are marked
!>        and stored in lists ... (main entry point)
!> \param tmc_env TMC environment
!> \param result_acc checked configuration accepted? (result)
!> \param something_updated ...
!> \param
!> \param
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE tree_update(tmc_env, result_acc, something_updated)
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      LOGICAL                                            :: result_acc, something_updated

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tree_update', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, itmp
      LOGICAL                                            :: found
      TYPE(global_tree_type), POINTER                    :: gt_act_elem
      TYPE(tree_type), POINTER                           :: act_element, tmp_element

      NULLIFY (gt_act_elem, tmp_element, act_element)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_acceptance.F",729)

      ! start the timing
      CALL timeset(routineN, handle)

      result_acc = .FALSE.
      something_updated = .FALSE.

      gt_act_elem => tmc_env%m_env%gt_act
      search_calculated_element_loop: DO
         NULLIFY (act_element, tmp_element)
         ! search for element to check
         CALL search_next_gt_element_to_check(ptr=gt_act_elem, found=found)
         IF (.NOT. found) EXIT search_calculated_element_loop

         ! check the elements status end, if possible, do acceptance check
         CALL check_elements(gt_act_elem=gt_act_elem, tmc_env=tmc_env, &
                             check_done=found, result_acc=result_acc)
         ! if check was not possible, update the status of the global tree element and return
         IF (.NOT. found) EXIT search_calculated_element_loop

         ! get elements related to global tree element, which were checked
         CALL get_subtree_elements_to_check(gt_act_elem=gt_act_elem, &
                                            elem1=act_element, elem2=tmp_element)

         !========================================================================
         !-- set result counters
         ! counter for certain temperature
         tmc_env%m_env%result_count(gt_act_elem%mv_conf) = &
            tmc_env%m_env%result_count(gt_act_elem%mv_conf)+1
         ! in case of swapped also count the result for
         !  the other swapped temperature
         IF (gt_act_elem%swaped) &
            tmc_env%m_env%result_count(gt_act_elem%mv_conf+1) = &
            tmc_env%m_env%result_count(gt_act_elem%mv_conf+1)+1
         ! count also for global tree Markov Chain
         tmc_env%m_env%result_count(0) = tmc_env%m_env%result_count(0)+1

         ! flag for doing tree cleaning with canceling and deallocation
         something_updated = .TRUE.

         !IF((.NOT.gt_act_elem%swaped) .AND. act_element%temp_created.NE.0 .AND. &
         !   act_element%temp_created.NE.gt_act_elem%mv_conf)&
         !  count_wrong_temp_move(gt_act_elem%mv_conf) = &
         !       count_wrong_temp_move(gt_act_elem%mv_conf)+1

         !========================================================================
         !-- sort element in lists
         !-- case NOT ACCEPTED
         IF (.NOT. result_acc) THEN !result NOT accepted
            IF (gt_act_elem%prob_acc .GT. 0.0_dp) THEN
               IF (gt_act_elem%prob_acc .GE. 0.5_dp) THEN
                  ! wrong estimate (estimated accepted)
                  tmc_env%m_env%estim_corr_wrong(4) = tmc_env%m_env%estim_corr_wrong(4)+1
                  IF (DEBUG .GT. 0) WRITE (tmc_env%m_env%io_unit, *) &
                     "Wrong guess for NACC (elem/estim acc prob)", gt_act_elem%nr, gt_act_elem%prob_acc
               ELSE
                  tmc_env%m_env%estim_corr_wrong(3) = tmc_env%m_env%estim_corr_wrong(3)+1
               END IF
            END IF
            gt_act_elem%stat = status_rejected_result
            gt_act_elem%prob_acc = 0.0_dp
            itmp = gt_act_elem%conf(gt_act_elem%mv_conf)%elem%sub_tree_nr

            !-- case ACCEPTED
         ELSE
            IF (gt_act_elem%prob_acc .GT. 0.0_dp) THEN
               IF (gt_act_elem%prob_acc .LE. 0.5_dp) THEN
                  ! wrong estimate (estimated NOT accepted)
                  tmc_env%m_env%estim_corr_wrong(2) = tmc_env%m_env%estim_corr_wrong(2)+1
                  IF (DEBUG .GT. 0) WRITE (tmc_env%m_env%io_unit, *) &
                     "wrong guess for ACC  (elem/estim acc prob)", gt_act_elem%nr, gt_act_elem%prob_acc
               ELSE
                  tmc_env%m_env%estim_corr_wrong(1) = tmc_env%m_env%estim_corr_wrong(1)+1
               END IF
            END IF
            gt_act_elem%stat = status_accepted_result
            gt_act_elem%prob_acc = 1.0_dp

            ! save result in the result list
            IF (.NOT. gt_act_elem%swaped) THEN
               !-- saving moved/changed configuration of result node
               tmc_env%m_env%result_list(gt_act_elem%mv_conf)%elem => &
                  gt_act_elem%conf(gt_act_elem%mv_conf)%elem
            ELSE
               ! ATTENTION: act_element != gt_act_elem%conf(gt_act_elem%mv_conf),
               !   because we take the last accepted conf
               tmc_env%m_env%result_list(gt_act_elem%mv_conf)%elem => act_element
               tmc_env%m_env%result_list(gt_act_elem%mv_conf+1)%elem => tmp_element
            END IF
            tmc_env%m_env%gt_act => gt_act_elem
         END IF ! result acceptance check
         !======================================================================
         !-- set status accepted or rejected of certain subtree elements
         IF (.NOT. gt_act_elem%swaped) &
            CALL subtree_configuration_stat_change(gt_ptr=gt_act_elem, &
                                                   stat=gt_act_elem%stat, &
                                                   tmc_params=tmc_env%params)

         IF (tmc_env%params%DRAW_TREE) &
            CALL create_global_tree_dot_color(gt_tree_element=gt_act_elem, &
                                              tmc_params=tmc_env%params)

         ! probability update
         CALL prob_update(move_types=tmc_env%params%move_types, &
                          pt_el=gt_act_elem, &
                          prob_opt=tmc_env%params%esimate_acc_prob)

         !writes only different configurations with repetition file if possible
         CALL write_result_list_element(result_list=tmc_env%m_env%result_list, &
                                        result_count=tmc_env%m_env%result_count, &
                                        conf_updated=gt_act_elem%mv_conf, accepted=result_acc, &
                                        tmc_params=tmc_env%params)
         IF (gt_act_elem%swaped) &
            CALL write_result_list_element(result_list=tmc_env%m_env%result_list, &
                                           result_count=tmc_env%m_env%result_count, &
                                           conf_updated=gt_act_elem%mv_conf+1, accepted=result_acc, &
                                           tmc_params=tmc_env%params)

         ! save for analysis
         IF (tmc_env%tmc_comp_set%para_env_m_ana%num_pe .GT. 1 .AND. result_acc) THEN
            CALL add_to_list(elem=tmc_env%m_env%result_list(gt_act_elem%mv_conf)%elem, &
                             list=tmc_env%m_env%analysis_list, &
                             temp_ind=gt_act_elem%mv_conf, &
                             nr=tmc_env%m_env%result_count(gt_act_elem%mv_conf))
            IF (gt_act_elem%swaped) THEN
               CALL add_to_list(elem=tmc_env%m_env%result_list(gt_act_elem%mv_conf+1)%elem, &
                                list=tmc_env%m_env%analysis_list, &
                                temp_ind=gt_act_elem%mv_conf+1, &
                                nr=tmc_env%m_env%result_count(gt_act_elem%mv_conf+1))
            END IF
         END IF
      END DO search_calculated_element_loop

      ! end the timing
      CALL timestop(handle)

   END SUBROUTINE tree_update

   !============================================================================
   ! tree node acceptance probability
   !============================================================================
! **************************************************************************************************
!> \brief checks if element is ready for accaptance probability update
!>        checks status and the amount of already received intermediate energies
!> \param elem sub tree element to update
!> \retval ready return value
!> \author Mandes 12.2012
! **************************************************************************************************
   FUNCTION ready_for_update_acc_prob(elem) RESULT(ready)
      TYPE(tree_type), POINTER                           :: elem
      LOGICAL                                            :: ready

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ready_for_update_acc_prob', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",885)
      ready = .FALSE.
      IF ((elem%scf_energies_count .GE. 4) &
          .AND. (elem%stat .NE. status_deleted) .AND. (elem%stat .NE. status_deleted_result) &
          .AND. (elem%stat .NE. status_canceled_ener)) &
         ready = .TRUE.
   END FUNCTION ready_for_update_acc_prob

! **************************************************************************************************
!> \brief updates the subtree acceptance probability
!>        the swap probabilities are handled within the certain checks of the
!>        global tree elements (pt references)
!> \param tree_elem sub tree element to update
!> \param tmc_env TMC environment
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE check_elements_for_acc_prob_update(tree_elem, tmc_env)
      TYPE(tree_type), POINTER                           :: tree_elem
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_elements_for_acc_prob_update', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(tree_type), POINTER                           :: act_elem, parent_elem

      NULLIFY (parent_elem, act_elem)

      IF(.NOT.(ASSOCIATED(tree_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",913)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_tree_acceptance.F",914)

      ! start the timing
      CALL timeset(routineN, handle)

      act_elem => tree_elem
      !-- nothing to do if option is disabled or element not ready
      IF (tmc_env%params%esimate_acc_prob .AND. &
          ready_for_update_acc_prob(act_elem)) THEN
         ! ------------------------------------------------------
         ! check this element
         !   for usual moves and swapping
         ! get parent subtree elment for case of not swapped configurations
         parent_elem => search_parent_element(act_elem)
         ! update the prob of accaptance
         CALL update_prob_gt_node_list(reference_list=act_elem%gt_nodes_references, &
                                       act_elem=act_elem, parent_elem=parent_elem, act_parent=.TRUE., &
                                       tmc_env=tmc_env)

         !------------------------------------------------------
         ! check the childs of this element
         parent_elem => tree_elem

         ! check ACC child (parent element is the entered/updated element)
         IF (ASSOCIATED(tree_elem%acc)) THEN
            act_elem => tree_elem%acc
            ! update the prob of accaptance
            CALL update_prob_gt_node_list(reference_list=act_elem%gt_nodes_references, &
                                          act_elem=act_elem, parent_elem=parent_elem, act_parent=.FALSE., &
                                          tmc_env=tmc_env)
         END IF

         ! check all NACC childs of next accepted one
         nacc_loop: DO
            IF (.NOT. ASSOCIATED(act_elem%nacc)) EXIT nacc_loop
            act_elem => act_elem%nacc
            CALL update_prob_gt_node_list(reference_list=act_elem%gt_nodes_references, &
                                          act_elem=act_elem, parent_elem=parent_elem, act_parent=.FALSE., &
                                          tmc_env=tmc_env)
         END DO nacc_loop
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE check_elements_for_acc_prob_update

! **************************************************************************************************
!> \brief search all pt nodes in reference list and update the estimated
!>        acceptance probabilities
!> \param reference_list list of global tree element refering to the updated
!>        subtree element
!> \param act_elem actual sub tree element updated or child of the updated one
!> \param parent_elem parent or the actual updated subtree element
!> \param act_parent flag if updated element is the actual (TRUE) or
!>        the parent (FALSE) element
!> \param tmc_env TMC environment
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE update_prob_gt_node_list(reference_list, act_elem, parent_elem, act_parent, tmc_env)
      TYPE(gt_elem_list_type), POINTER                   :: reference_list
      TYPE(tree_type), POINTER                           :: act_elem, parent_elem
      LOGICAL                                            :: act_parent
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'update_prob_gt_node_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, integration_steps
      REAL(KIND=dp)                                      :: kB, tmp_prob
      TYPE(gt_elem_list_type), POINTER                   :: tmp_pt_ptr
      TYPE(tree_type), POINTER                           :: tmp_elem, tmp_parent_elem

      kB = boltzmann/joule

      NULLIFY (tmp_pt_ptr, tmp_elem, tmp_parent_elem)

      IF (.NOT. ASSOCIATED(reference_list)) RETURN ! could be canceled already

      IF(.NOT.(ASSOCIATED(reference_list%gt_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",991)
      IF(.NOT.(ASSOCIATED(act_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",992)
      IF(.NOT.(ASSOCIATED(parent_elem)))CALL cp__a("tmc/tmc_tree_acceptance.F",993)

      ! start the timing
      CALL timeset(routineN, handle)

      ! check if element is ready
      IF (ready_for_update_acc_prob(act_elem)) THEN
         ! set the number of integration steps used for 3 point approximation
         ! of the exact energy, using the sub step energies
         ! still fixed value
         integration_steps = 100

         tmp_pt_ptr => reference_list
         ! do for all references using the updated subtree node
         reference_loop: DO WHILE (ASSOCIATED(tmp_pt_ptr))
            tmp_prob = -1.0_dp
            IF (tmp_pt_ptr%gt_elem%swaped) THEN
               NULLIFY (tmp_elem, tmp_parent_elem)
               ! in case of swap use the other swaped configuration as related one
               CALL get_subtree_elements_to_check(gt_act_elem=tmp_pt_ptr%gt_elem, &
                                                  elem1=tmp_elem, elem2=tmp_parent_elem)
               ! NOT if parent is the updated one, and check for correct elements ready
               IF (act_parent .AND. ASSOCIATED(act_elem, tmp_elem) .AND. &
                   ready_for_update_acc_prob(elem=tmp_parent_elem)) THEN
                  ! using ln(rnd)/-(beta_i-beta_j) < U_j-U_i)
                  tmp_prob = compute_estimated_prob(elem_old=tmp_parent_elem, &
                                                    elem_new=act_elem, &
                                                    E_classical_diff=0.0_dp, &
                                                    rnd_nr=tmp_pt_ptr%gt_elem%rnd_nr, &
                                                    beta=1.0_dp/(kB*(tmc_env%params%Temp(tmp_pt_ptr%gt_elem%mv_conf)- &
                                                                     tmc_env%params%Temp(tmp_pt_ptr%gt_elem%mv_conf+1))), &
                                                    tmc_params=tmc_env%params)
               ELSE
                  tmp_pt_ptr => tmp_pt_ptr%next
                  CYCLE reference_loop
               END IF
            ELSE
               ! if no swap, use the parent configuration as related one
               IF (ready_for_update_acc_prob(parent_elem)) THEN
                  tmp_prob = compute_estimated_prob(elem_old=parent_elem, &
                                                    elem_new=act_elem, &
                                                    E_classical_diff=act_elem%e_pot_approx-parent_elem%e_pot_approx, &
                                                    rnd_nr=tmp_pt_ptr%gt_elem%rnd_nr, &
                                                    beta=1.0_dp/(kB*tmc_env%params%Temp(tmp_pt_ptr%gt_elem%mv_conf)), &
                                                    tmc_params=tmc_env%params)
               END IF
            END IF
            !successfull probability update
            IF (tmp_prob .GE. 0.0_dp) THEN
               tmp_pt_ptr%gt_elem%prob_acc = tmp_prob
               !-- speculative canceling for the related direction
               IF (tmc_env%params%SPECULATIVE_CANCELING) &
                  CALL search_canceling_elements(pt_elem_in=tmp_pt_ptr%gt_elem, &
                                                 prob=tmp_pt_ptr%gt_elem%prob_acc, tmc_env=tmc_env)
            END IF

            ! get next related global tree pointer
            tmp_pt_ptr => tmp_pt_ptr%next
         END DO reference_loop ! global tree references of actual subtree element
      END IF
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE update_prob_gt_node_list

END MODULE tmc_tree_acceptance
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief set up the different message for different tasks
!>      A TMC message consists of 3 parts (messages)
!>      1: first a message with task type (STATUS) and SIZES of submessages
!>      2: (if existing) a message with INTEGER values
!>      3: (if existing) a message with REAL values
!>      submessages 2 and 3 include relevant data, e.g. positions, box sizes...
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************
MODULE tmc_messages
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_any_source,&
                                              mp_any_tag,&
                                              mp_bcast,&
                                              mp_probe,&
                                              mp_recv,&
                                              mp_send
   USE tmc_move_handle,                 ONLY: add_mv_prob
   USE tmc_stati,                       ONLY: &
        TMC_CANCELING_MESSAGE, TMC_CANCELING_RECEIPT, TMC_STATUS_CALCULATING, TMC_STATUS_FAILED, &
        TMC_STATUS_STOP_RECEIPT, TMC_STATUS_WAIT_FOR_NEW_TASK, TMC_STATUS_WORKER_INIT, &
        TMC_STAT_ANALYSIS_REQUEST, TMC_STAT_ANALYSIS_RESULT, TMC_STAT_APPROX_ENERGY_REQUEST, &
        TMC_STAT_APPROX_ENERGY_RESULT, TMC_STAT_ENERGY_REQUEST, TMC_STAT_ENERGY_RESULT, &
        TMC_STAT_INIT_ANALYSIS, TMC_STAT_MD_BROADCAST, TMC_STAT_MD_REQUEST, TMC_STAT_MD_RESULT, &
        TMC_STAT_NMC_BROADCAST, TMC_STAT_NMC_REQUEST, TMC_STAT_NMC_RESULT, &
        TMC_STAT_SCF_STEP_ENER_RECEIVE, TMC_STAT_START_CONF_REQUEST, TMC_STAT_START_CONF_RESULT, &
        task_type_gaussian_adaptation
   USE tmc_tree_build,                  ONLY: allocate_new_sub_tree_node
   USE tmc_tree_types,                  ONLY: elem_array_type,&
                                              elem_list_type,&
                                              tree_type
   USE tmc_types,                       ONLY: allocate_tmc_atom_type,&
                                              tmc_atom_type,&
                                              tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PARAMETER, PUBLIC                 :: send_msg = .TRUE.
   LOGICAL, PARAMETER, PUBLIC                 :: recv_msg = .FALSE.

   INTEGER, PARAMETER                         :: message_end_flag = 25

   INTEGER, PARAMETER                         :: DEBUG = 0

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_messages'

   PUBLIC :: check_if_group_master
   PUBLIC :: tmc_message
   PUBLIC :: communicate_atom_types
   PUBLIC :: stop_whole_group

   INTEGER, PARAMETER, PUBLIC :: MASTER_COMM_ID = 0 ! id for master and group master
   INTEGER, PARAMETER, PUBLIC :: bcast_group = -1 ! destination flag for broadcasting to other group participants
   INTEGER, PARAMETER :: TMC_SEND_INFO_SIZE = 4 ! usually: 1. status, array sizes: 2. int, 3. real, 4. char

   TYPE message_send
      INTEGER, DIMENSION(TMC_SEND_INFO_SIZE)   :: info
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: task_real
      INTEGER, DIMENSION(:), ALLOCATABLE       :: task_int
      CHARACTER, DIMENSION(:), ALLOCATABLE     :: task_char
      !should be deleted somewhen
      INTEGER, DIMENSION(:), ALLOCATABLE :: elem_stat
   END TYPE message_send

CONTAINS

! **************************************************************************************************
!> \brief checks if the core is the group master
!> \param para_env defines the mpi communicator
!> \retval master return value, logical
!> \author Mandes 01.2013
! **************************************************************************************************
   FUNCTION check_if_group_master(para_env) RESULT(master)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL                                            :: master

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_if_group_master', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("tmc/tmc_messages.F",93)

      master = .FALSE.
      IF (para_env%mepos .EQ. MASTER_COMM_ID) &
         master = .TRUE.
   END FUNCTION check_if_group_master

! **************************************************************************************************
!> \brief tmc message handling, packing messages with integer and real data
!>        type. Send first info message with task type and message sizes and
!>        then the int and real messages. The same for receiving
!> \param msg_type defines the message types, see message tags definition
!> \param send_recv 1= send, 0= receive
!> \param dest defines the target or source of message
!>              (-1=braodcast, 0= master, 1... working group)
!> \param para_env defines the mpi communicator
!> \param tmc_params stuct with parameters (global settings)
!> \param elem a subtree element from which info are readed or written in
!> \param elem_array ...
!> \param list_elem ...
!> \param result_count ...
!> \param wait_for_message ...
!> \param success ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE tmc_message(msg_type, send_recv, dest, para_env, tmc_params, &
                          elem, elem_array, list_elem, result_count, &
                          wait_for_message, success)
      INTEGER                                            :: msg_type
      LOGICAL                                            :: send_recv
      INTEGER                                            :: dest
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(tree_type), OPTIONAL, POINTER                 :: elem
      TYPE(elem_array_type), DIMENSION(:), OPTIONAL      :: elem_array
      TYPE(elem_list_type), OPTIONAL, POINTER            :: list_elem
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: result_count
      LOGICAL, OPTIONAL                                  :: wait_for_message, success

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tmc_message', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, message_tag, tmp_tag
      LOGICAL                                            :: act_send_recv, flag
      TYPE(message_send), POINTER                        :: m_send

      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("tmc/tmc_messages.F",138)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",139)

      ALLOCATE (m_send)

      ! init
      ! define send_recv flag for broadcast
      IF (dest .EQ. bcast_group) THEN
         ! master should always send
         IF (para_env%mepos .EQ. MASTER_COMM_ID) THEN
            act_send_recv = send_msg
         ELSE
            ! worker should always receive
            act_send_recv = recv_msg
         END IF
      ELSE
         act_send_recv = send_recv
      END IF
      message_tag = 0

      ! =============================
      ! sending message
      ! =============================
      ! creating message to send
      IF (act_send_recv .EQV. send_msg) THEN
         IF ((DEBUG .GE. 7) .AND. (dest .NE. bcast_group) .AND. &
             (dest .NE. MASTER_COMM_ID)) THEN
            IF (PRESENT(elem)) THEN
               WRITE (*, *) "send element info to ", dest, " of type ", msg_type, "of subtree", elem%sub_tree_nr, &
                  "elem", elem%nr
            ELSE
               WRITE (*, *) "send element info to ", dest, " of type ", msg_type
            END IF
         END IF
         SELECT CASE (msg_type)
         CASE (TMC_STAT_START_CONF_REQUEST, TMC_STATUS_FAILED, TMC_CANCELING_MESSAGE, &
               TMC_CANCELING_RECEIPT, TMC_STATUS_STOP_RECEIPT, &
               TMC_STATUS_WAIT_FOR_NEW_TASK, TMC_STATUS_CALCULATING, &
               TMC_STAT_ANALYSIS_RESULT)
            CALL create_status_message(m_send)
         CASE (TMC_STATUS_WORKER_INIT)
            CALL create_worker_init_message(tmc_params, m_send)
         CASE (TMC_STAT_START_CONF_RESULT, TMC_STAT_INIT_ANALYSIS)
            CALL create_start_conf_message(msg_type, elem, result_count, tmc_params, m_send)
         CASE (TMC_STAT_ENERGY_REQUEST, TMC_STAT_APPROX_ENERGY_REQUEST)
            CALL create_energy_request_message(elem, m_send, tmc_params)
         CASE (TMC_STAT_APPROX_ENERGY_RESULT)
            CALL create_approx_energy_result_message(elem, m_send, tmc_params)
         CASE (TMC_STAT_ENERGY_RESULT)
            CALL create_energy_result_message(elem, m_send, tmc_params)
         CASE (TMC_STAT_NMC_REQUEST, TMC_STAT_NMC_BROADCAST, &
               TMC_STAT_MD_REQUEST, TMC_STAT_MD_BROADCAST)
            CALL create_NMC_request_massage(msg_type, elem, m_send, tmc_params)
         CASE (TMC_STAT_MD_RESULT, TMC_STAT_NMC_RESULT)
            CALL create_NMC_result_massage(msg_type, elem, m_send, tmc_params)
         CASE (TMC_STAT_ANALYSIS_REQUEST)
            IF(.NOT.(PRESENT(list_elem)))CALL cp__a("tmc/tmc_messages.F",194)
            CALL create_analysis_request_message(list_elem, m_send, tmc_params)
         CASE DEFAULT
            CALL cp__b("tmc/tmc_messages.F",197,"try to send unknown message type "//cp_to_string(msg_type))
         END SELECT
         !set message info
         message_tag = msg_type
         m_send%info(:) = 0
         m_send%info(1) = msg_type
         IF (ALLOCATED(m_send%task_int)) m_send%info(2) = SIZE(m_send%task_int)
         IF (ALLOCATED(m_send%task_real)) m_send%info(3) = SIZE(m_send%task_real)
         IF (ALLOCATED(m_send%task_char)) m_send%info(4) = SIZE(m_send%task_char)
      END IF

      ! sending message
      IF ((act_send_recv .EQV. send_msg) .AND. (dest .NE. bcast_group)) THEN
         CALL mp_send(m_send%info, dest, message_tag, para_env%group)
         IF (m_send%info(2) .GT. 0) THEN
            CALL mp_send(m_send%task_int, dest, message_tag, para_env%group)
         END IF
         IF (m_send%info(3) .GT. 0) THEN
            CALL mp_send(m_send%task_real, dest, message_tag, para_env%group)
         END IF
         IF (m_send%info(4) .GT. 0) THEN
            CALL cp__b("tmc/tmc_messages.F",218,"")
            !TODO send characters CALL mp_send(m_send%task_char, dest, message_tag, para_env%group)
         END IF
         IF (DEBUG .GE. 1) &
            WRITE (*, *) "TMC|message: ID: ", para_env%mepos, &
            " send element info to   ", dest, " of stat ", m_send%info(1), &
            " with size int/real/char", m_send%info(2:), " with comm ", &
            para_env%group, " and tag ", message_tag
         IF (m_send%info(2) .GT. 0) DEALLOCATE (m_send%task_int)
         IF (m_send%info(3) .GT. 0) DEALLOCATE (m_send%task_real)
         IF (m_send%info(4) .GT. 0) DEALLOCATE (m_send%task_char)
         IF (PRESENT(success)) success = .TRUE.
      END IF

      ! =============================
      ! broadcast
      ! =============================
      IF (dest .EQ. bcast_group) THEN
         IF (para_env%num_pe .GT. 1) THEN
            CALL mp_bcast(m_send%info, MASTER_COMM_ID, para_env%group)
            IF (m_send%info(2) .GT. 0) THEN
               IF (.NOT. act_send_recv) ALLOCATE (m_send%task_int(m_send%info(2)))
               CALL mp_bcast(m_send%task_int, MASTER_COMM_ID, para_env%group)
            END IF
            IF (m_send%info(3) .GT. 0) THEN
               IF (.NOT. act_send_recv) ALLOCATE (m_send%task_real(m_send%info(3)))
               CALL mp_bcast(m_send%task_real, MASTER_COMM_ID, para_env%group)
            END IF
            IF (m_send%info(4) .GT. 0) THEN
               IF (.NOT. act_send_recv) ALLOCATE (m_send%task_char(m_send%info(3)))
               CALL cp__b("tmc/tmc_messages.F",248,"")
               !TODO bcast char CALL mp_bcast(m_send%task_char, MASTER_COMM_ID, para_env%group)
            END IF
         END IF
         ! sender delete arrays
         IF (act_send_recv) THEN
            IF (m_send%info(2) .GT. 0) DEALLOCATE (m_send%task_int)
            IF (m_send%info(3) .GT. 0) DEALLOCATE (m_send%task_real)
            IF (m_send%info(4) .GT. 0) DEALLOCATE (m_send%task_char)
         END IF
      END IF

      ! =============================
      ! receiving message
      ! =============================
      IF ((act_send_recv .EQV. recv_msg) .AND. dest .NE. bcast_group) THEN
         flag = .FALSE.
         tmp_tag = TMC_STATUS_WAIT_FOR_NEW_TASK
         IF (PRESENT(wait_for_message)) THEN
            dest = mp_any_source
            CALL mp_probe(dest, para_env%group, tmp_tag)
            flag = .TRUE.
         ELSE
            participant_loop: DO i = 0, para_env%num_pe-1
               IF (i .NE. para_env%mepos) THEN
                  dest = i
                  CALL mp_probe(dest, para_env%group, tmp_tag)
                  IF (dest .EQ. i) THEN
                     flag = .TRUE.
                     EXIT participant_loop
                  END IF
               END IF
            END DO participant_loop
         END IF
         IF (flag .EQV. .FALSE.) THEN
            IF (PRESENT(success)) success = .FALSE.
            DEALLOCATE (m_send)
            RETURN
         END IF

         IF (tmp_tag .EQ. TMC_STAT_SCF_STEP_ENER_RECEIVE) THEN
            ! CP2K send back SCF step energies without info message
            message_tag = TMC_STAT_SCF_STEP_ENER_RECEIVE
            m_send%info(1) = TMC_STAT_SCF_STEP_ENER_RECEIVE
            m_send%info(2) = 0 ! no integer values
            m_send%info(3) = 1 ! one double values (SCF total energy)
            m_send%info(4) = 0 ! no character values
         ELSE
            message_tag = mp_any_tag
            ! first get message type and sizes
            CALL mp_recv(m_send%info, dest, message_tag, para_env%group)
         END IF
         IF (DEBUG .GE. 1) &
            WRITE (*, *) "TMC|message: ID: ", para_env%mepos, &
            " recv element info from ", dest, " of stat ", m_send%info(1), &
            " with size int/real/char", m_send%info(2:)
         !-- receive message integer part
         IF (m_send%info(2) .GT. 0) THEN
            ALLOCATE (m_send%task_int(m_send%info(2)))
            CALL mp_recv(m_send%task_int, dest, message_tag, para_env%group)
         END IF
         !-- receive message double (floatingpoint) part
         IF (m_send%info(3) .GT. 0) THEN
            ALLOCATE (m_send%task_real(m_send%info(3)))
            CALL mp_recv(m_send%task_real, dest, message_tag, para_env%group)
         END IF
         !-- receive message character part
         IF (m_send%info(4) .GT. 0) THEN
            ALLOCATE (m_send%task_char(m_send%info(4)))
            CALL cp__b("tmc/tmc_messages.F",317,"")
            !TODO recv characters CALL mp_recv(m_send%task_char, dest, message_tag, para_env%group)
         END IF
      END IF

      ! handling recived message
      IF (act_send_recv .EQV. recv_msg) THEN
         ! if the element is supposed to be canceled but received message is not canceling receipt do not handle element
         ! (because element could be already deallocated, and hence a new element would be created -> not necessary)
         IF (PRESENT(elem_array)) THEN
            IF (elem_array(dest)%canceled .AND. m_send%info(1) .NE. TMC_CANCELING_RECEIPT) THEN
               msg_type = m_send%info(1)
               IF (m_send%info(2) .GT. 0) DEALLOCATE (m_send%task_int)
               IF (m_send%info(3) .GT. 0) DEALLOCATE (m_send%task_real)
               IF (m_send%info(4) .GT. 0) DEALLOCATE (m_send%task_char)
               ! to check for further messages
               IF (PRESENT(success)) success = .TRUE.
               DEALLOCATE (m_send)
               RETURN
            END IF
         END IF

         msg_type = m_send%info(1)
         SELECT CASE (m_send%info (1))
         CASE (TMC_STAT_START_CONF_REQUEST, TMC_CANCELING_MESSAGE, &
               TMC_CANCELING_RECEIPT, TMC_STATUS_WAIT_FOR_NEW_TASK, &
               TMC_STATUS_CALCULATING, TMC_STAT_ANALYSIS_RESULT)
            ! nothing to do here
         CASE (TMC_STATUS_WORKER_INIT)
            CALL read_worker_init_message(tmc_params, m_send)
         CASE (TMC_STAT_START_CONF_RESULT, TMC_STAT_INIT_ANALYSIS)
            IF (PRESENT(elem_array)) THEN
               CALL read_start_conf_message(msg_type, elem_array(dest)%elem, &
                                            result_count, m_send, tmc_params)
            ELSE
               CALL read_start_conf_message(msg_type, elem, result_count, m_send, &
                                            tmc_params)
            END IF
         CASE (TMC_STAT_APPROX_ENERGY_RESULT)
            CALL read_approx_energy_result(elem_array(dest)%elem, m_send, tmc_params)
         CASE (TMC_STAT_ENERGY_REQUEST, TMC_STAT_APPROX_ENERGY_REQUEST)
            CALL read_energy_request_message(elem, m_send, tmc_params)
         CASE (TMC_STAT_ENERGY_RESULT)
            IF (PRESENT(elem_array)) &
               CALL read_energy_result_message(elem_array(dest)%elem, m_send, tmc_params)
         CASE (TMC_STAT_NMC_REQUEST, TMC_STAT_NMC_BROADCAST, &
               TMC_STAT_MD_REQUEST, TMC_STAT_MD_BROADCAST)
            CALL read_NMC_request_massage(msg_type, elem, m_send, tmc_params)
         CASE (TMC_STAT_NMC_RESULT, TMC_STAT_MD_RESULT)
            IF (PRESENT(elem_array)) &
               CALL read_NMC_result_massage(msg_type, elem_array(dest)%elem, m_send, tmc_params)
         CASE (TMC_STATUS_FAILED, TMC_STATUS_STOP_RECEIPT)
            ! if task is failed, handle situation in outer routine
         CASE (TMC_STAT_SCF_STEP_ENER_RECEIVE)
            CALL read_scf_step_ener(elem_array(dest)%elem, m_send)
         CASE (TMC_STAT_ANALYSIS_REQUEST)
            CALL read_analysis_request_message(elem, m_send, tmc_params)
         CASE DEFAULT
            CALL cp_abort(cp__l("tmc/tmc_messages.F",375), &
                          "try to receive unknown message type "//cp_to_string(msg_type)// &
                          "from source "//cp_to_string(dest))
         END SELECT
         IF (m_send%info(2) .GT. 0) DEALLOCATE (m_send%task_int)
         IF (m_send%info(3) .GT. 0) DEALLOCATE (m_send%task_real)
         IF (m_send%info(4) .GT. 0) DEALLOCATE (m_send%task_char)
         IF (PRESENT(success)) success = .TRUE.
      END IF

      ! ATTENTION there is also an short exit (RETURN) after probing for new messages
      DEALLOCATE (m_send)
   END SUBROUTINE tmc_message

! **************************************************************************************************
!> \brief set the messege just with an status tag
!> \param m_send the message structure
!> \author Mandes 12.2012
! **************************************************************************************************

   SUBROUTINE create_status_message(m_send)
      TYPE(message_send), POINTER                        :: m_send

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_status_message', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",401)

      ! nothing to do, send just the message tag

      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",405)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",406)

   END SUBROUTINE create_status_message

   !============================================================================
   ! message for requesting start configuration
   !============================================================================
!! **************************************************************************************************
!!> \brief the message for sending the atom mass
!!>        (number of atoms is also tranfered)
!!>        atom names have to be done seperately,
!!>        because character send only with bcast possible
!!> \param tmc_parms th send the cell properties
!!> \param m_send the message structure
!!> \param error variable to control error logging, stopping,...
!!>        see module cp_error_handling
!!> \author Mandes 02.2013
!! **************************************************************************************************
!  SUBROUTINE create_atom_mass_message(m_send, atoms)
!    TYPE(tmc_atom_type), DIMENSION(:), POINTER    :: atoms
!    TYPE(message_send), POINTER              :: m_send
!
!    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_atom_mass_message', &
!      routineP = moduleN//':'//routineN
!
!    INTEGER                                  :: counter, i, &
!                                                msg_size_real
!    LOGICAL                                  :: failure
!
!    failure = .FALSE.
!
!    CPPrecondition(ASSOCIATED(m_send),cp_failure_level,routineP,failure)
!    CPPrecondition(.NOT.ALLOCATED(m_send%task_int),cp_failure_level,routineP,failure)
!    CPPrecondition(.NOT.ALLOCATED(m_send%task_real),cp_failure_level,routineP,failure)
!    CPPrecondition(.NOT.ALLOCATED(m_send%task_char),cp_failure_level,routineP,failure)
!
!    counter =1
!    msg_size_real = 1+SIZE(tmc_params%cell%hmat)+ 1+SIZE(atoms) +1
!    ALLOCATE(m_send%task_real(msg_size_real))
!
!    m_send%task_real(1) = REAL(SIZE(atoms,KIND=dp))
!    DO i=1, SIZE(atoms)
!      m_send%task_real(counter+i) = atoms(i)%mass
!    END DO
!    counter = counter + 1+INT(m_send%task_real(counter))
!    m_send%task_real(counter) = REAL(message_end_flag, KIND=dp) !message end
!    CPPostconditionNoFail(INT(m_send%task_real(msg_size_real)).EQ.message_end_flag,cp_failure_level,routineP)
!  END SUBROUTINE create_atom_mass_message
!
!! **************************************************************************************************
!!> \brief the message for reading the atom mass
!!>        (number of atoms is also tranfered)
!!>        atom names have to be done seperately,
!!>        because character send only with bcast possible
!!> \param tmc_parms th send the cell properties
!!> \param m_send the message structure
!!> \param error variable to control error logging, stopping,...
!!>        see module cp_error_handling
!!> \author Mandes 02.2013
!! **************************************************************************************************
!  SUBROUTINE read_atom_mass_message(m_send, atoms)
!    TYPE(tmc_atom_type), DIMENSION(:), &
!      POINTER                                :: atoms
!    TYPE(message_send), POINTER              :: m_send
!
!    CHARACTER(LEN=*), PARAMETER :: routineN = 'read_atom_mass_message', &
!      routineP = moduleN//':'//routineN
!
!    INTEGER                                  :: counter, i, nr_atoms
!    LOGICAL                                  :: failure
!
!    failure = .FALSE.
!
!    CPPrecondition(ASSOCIATED(m_send),cp_failure_level,routineP,failure)
!    CPPrecondition(.NOT.ALLOCATED(m_send%task_int),cp_failure_level,routineP,failure)
!    CPPrecondition(ALLOCATED(m_send%task_real),cp_failure_level,routineP,failure)
!    CPPrecondition(.NOT.ALLOCATED(m_send%task_char),cp_failure_level,routineP,failure)
!
!    counter =1
!    nr_atoms = m_send%task_real(counter)
!    IF(.NOT.ASSOCIATED(atoms)) CALL allocate_tmc_atom_type(atoms, nr_atoms)
!    DO i=1, SIZE(atoms)
!      atoms(i)%mass = m_send%task_real(counter+i)
!    END DO
!    counter = counter + 1+INT(m_send%task_real(counter))
!    CPPostconditionNoFail(INT(m_send%task_real(counter)).EQ.message_end_flag,cp_failure_level,routineP)
!  END SUBROUTINE read_atom_mass_message

! **************************************************************************************************
!> \brief the message for the initial values (cell size) to the workers
!> \param tmc_params to send the cell properties
!> \param m_send the message structure
!> \author Mandes 07.2013
! **************************************************************************************************
   SUBROUTINE create_worker_init_message(tmc_params, m_send)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(message_send), POINTER                        :: m_send

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_worker_init_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, msg_size_int, msg_size_real

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",509)
      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",510)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",511)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",512)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_char)))CALL cp__a("tmc/tmc_messages.F",513)
      IF(.NOT.(ASSOCIATED(tmc_params%cell)))CALL cp__a("tmc/tmc_messages.F",514)

      counter = 1
      msg_size_int = 1+SIZE(tmc_params%cell%perd)+1+1+1+1
      ALLOCATE (m_send%task_int(msg_size_int))
      m_send%task_int(counter) = SIZE(tmc_params%cell%perd) ! periodicity of the cell
      counter = counter+1+m_send%task_int(counter)
      m_send%task_int(2:counter-1) = tmc_params%cell%perd(:)
      m_send%task_int(counter) = 1
      m_send%task_int(counter+1) = tmc_params%cell%symmetry_id
      m_send%task_int(counter+2) = 0
      IF (tmc_params%cell%orthorhombic) m_send%task_int(counter+2) = 1
      counter = counter+3
      m_send%task_int(counter) = message_end_flag
      IF(.NOT.(counter .EQ. SIZE(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",528)

      !float array with cell vectors
      msg_size_real = 1+SIZE(tmc_params%cell%hmat)+1
      ALLOCATE (m_send%task_real(msg_size_real))
      counter = 1
      m_send%task_real(counter) = SIZE(tmc_params%cell%hmat) ! cell vectors for cell size
      m_send%task_real(counter+1:counter+SIZE(tmc_params%cell%hmat)) = &
         RESHAPE(tmc_params%cell%hmat(:, :), &
                 (/SIZE(tmc_params%cell%hmat)/))
      counter = counter+1+INT(m_send%task_real(counter))
      m_send%task_real(counter) = REAL(message_end_flag, KIND=dp) !message end
      IF(.NOT.(SIZE(m_send%task_real) .EQ. msg_size_real))CALL cp__a("tmc/tmc_messages.F",540)
      IF(.NOT.(INT(m_send%task_real(msg_size_real)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",541)
   END SUBROUTINE create_worker_init_message

! **************************************************************************************************
!> \brief the message for the initial values (cell size) to the workers
!> \param tmc_params to send the cell properties
!> \param m_send the message structure
!> \author Mandes 07.2013
! **************************************************************************************************
   SUBROUTINE read_worker_init_message(tmc_params, m_send)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(message_send), POINTER                        :: m_send

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_worker_init_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter
      LOGICAL                                            :: flag

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",560)
      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",561)
      IF(.NOT.(m_send%info(3) .GE. 4))CALL cp__a("tmc/tmc_messages.F",562)

      IF (.NOT. ASSOCIATED(tmc_params%cell)) ALLOCATE (tmc_params%cell)
      counter = 1
      !int array
      flag = INT(m_send%task_int(1)) .EQ. SIZE(tmc_params%cell%perd)
      IF(.NOT.(flag))CALL cp__a("tmc/tmc_messages.F",568)
      counter = 1+m_send%task_int(1)+1
      tmc_params%cell%perd = m_send%task_int(2:counter-1)
      tmc_params%cell%symmetry_id = m_send%task_int(counter+1)
      tmc_params%cell%orthorhombic = .FALSE.
      IF (m_send%task_int(counter+2) .EQ. 1) tmc_params%cell%orthorhombic = .TRUE.
      counter = counter+3
      IF(.NOT.(counter .EQ. m_send%info(2)))CALL cp__a("tmc/tmc_messages.F",575)
      IF(.NOT.(m_send%task_int(counter) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",576)

      !float array with cell vectors
      counter = 1
      flag = INT(m_send%task_real(counter)) .EQ. SIZE(tmc_params%cell%hmat)
      IF(.NOT.(flag))CALL cp__a("tmc/tmc_messages.F",581)
      tmc_params%cell%hmat = &
         RESHAPE(m_send%task_real(counter+1:counter+ &
                                  SIZE(tmc_params%cell%hmat)), (/3, 3/))
      counter = counter+1+INT(m_send%task_real(counter))

      IF(.NOT.(counter .EQ. m_send%info(3)))CALL cp__a("tmc/tmc_messages.F",587)
      IF(.NOT.(INT(m_send%task_real(m_send%info(3))) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",588)

   END SUBROUTINE read_worker_init_message

! **************************************************************************************************
!> \brief the message for sending back the initial configuration
!> \param msg_type the status tag
!> \param elem the intial tree element with initial coordinates and energy
!>        (using the approximated potential)
!> \param result_count ...
!> \param tmc_params to send the cell properties
!> \param m_send the message structure
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_start_conf_message(msg_type, elem, result_count, &
                                        tmc_params, m_send)
      INTEGER                                            :: msg_type
      TYPE(tree_type), POINTER                           :: elem
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: result_count
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(message_send), POINTER                        :: m_send

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_start_conf_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, i, msg_size_int, msg_size_real

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",615)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",616)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",617)
      IF(.NOT.(ASSOCIATED(tmc_params%atoms)))CALL cp__a("tmc/tmc_messages.F",618)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",619)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",620)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_char)))CALL cp__a("tmc/tmc_messages.F",621)

      counter = 1
      msg_size_int = 1+SIZE(tmc_params%cell%perd)+1+1+1+1+SIZE(elem%mol)+1
      IF (msg_type .EQ. TMC_STAT_INIT_ANALYSIS) THEN
         IF(.NOT.(PRESENT(result_count)))CALL cp__a("tmc/tmc_messages.F",626)
         IF(.NOT.(ASSOCIATED(result_count)))CALL cp__a("tmc/tmc_messages.F",627)
         msg_size_int = msg_size_int+1+SIZE(result_count(1:))
      END IF
      ALLOCATE (m_send%task_int(msg_size_int))
      m_send%task_int(counter) = SIZE(tmc_params%cell%perd) ! periodicity of the cell
      counter = counter+1+m_send%task_int(counter)
      m_send%task_int(2:counter-1) = tmc_params%cell%perd(:)
      m_send%task_int(counter) = 1
      m_send%task_int(counter+1) = tmc_params%cell%symmetry_id
      m_send%task_int(counter+2) = 0
      IF (tmc_params%cell%orthorhombic) m_send%task_int(counter+2) = 1
      counter = counter+3
      m_send%task_int(counter) = SIZE(elem%mol)
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = elem%mol(:)
      counter = counter+1+m_send%task_int(counter)
      IF (msg_type .EQ. TMC_STAT_INIT_ANALYSIS) THEN
         m_send%task_int(counter) = SIZE(result_count(1:))
         m_send%task_int(counter+1:counter+m_send%task_int(counter)) = &
            result_count(1:)
         counter = counter+1+m_send%task_int(counter)
      END IF
      m_send%task_int(counter) = message_end_flag
      IF(.NOT.(counter .EQ. SIZE(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",649)

      counter = 0
      !float array with pos, cell vectors, atom_mass
      msg_size_real = 1+SIZE(elem%pos)+1+SIZE(tmc_params%cell%hmat) &
                      +1+SIZE(tmc_params%atoms)+1
      ALLOCATE (m_send%task_real(msg_size_real))
      m_send%task_real(1) = REAL(SIZE(elem%pos), KIND=dp) ! positions
      counter = 2+INT(m_send%task_real(1))
      m_send%task_real(2:counter-1) = elem%pos
      m_send%task_real(counter) = SIZE(tmc_params%cell%hmat) ! cell vectors for cell size
      m_send%task_real(counter+1:counter+SIZE(tmc_params%cell%hmat)) = &
         RESHAPE(tmc_params%cell%hmat(:, :), &
                 (/SIZE(tmc_params%cell%hmat)/))
      counter = counter+1+INT(m_send%task_real(counter))
      m_send%task_real(counter) = SIZE(tmc_params%atoms) ! atom mass
      DO i = 1, SIZE(tmc_params%atoms)
         m_send%task_real(counter+i) = tmc_params%atoms(i)%mass
      END DO
      counter = counter+1+INT(m_send%task_real(counter))
      m_send%task_real(counter) = REAL(message_end_flag, KIND=dp) !message end
      IF(.NOT.(SIZE(m_send%task_real) .EQ. msg_size_real))CALL cp__a("tmc/tmc_messages.F",670)
      IF(.NOT.(INT(m_send%task_real(msg_size_real)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",671)

   END SUBROUTINE create_start_conf_message

! **************************************************************************************************
!> \brief the message for sending back the initial configuration
!> \param msg_type the status tag
!> \param elem the intial tree element with initial coordinates and energy
!>        (using the approximated potential)
!> \param result_count ...
!> \param m_send the message structure
!> \param tmc_params the param struct with necessary values for allocation
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE read_start_conf_message(msg_type, elem, result_count, m_send, &
                                      tmc_params)
      INTEGER                                            :: msg_type
      TYPE(tree_type), POINTER                           :: elem
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: result_count
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_start_conf_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, i
      LOGICAL                                            :: flag

      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",699)
      IF(.NOT.(.NOT. ASSOCIATED(tmc_params%atoms)))CALL cp__a("tmc/tmc_messages.F",700)
      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",701)
      IF(.NOT.(.NOT. ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",702)
      IF(.NOT.(m_send%info(3) .GE. 4))CALL cp__a("tmc/tmc_messages.F",703)

      IF (.NOT. ASSOCIATED(tmc_params%cell)) ALLOCATE (tmc_params%cell)
      CALL allocate_new_sub_tree_node(tmc_params=tmc_params, next_el=elem, &
                                      nr_dim=NINT(m_send%task_real(1)))
      counter = 1
      !int array
      flag = INT(m_send%task_int(1)) .EQ. SIZE(tmc_params%cell%perd)
      IF(.NOT.(flag))CALL cp__a("tmc/tmc_messages.F",711)
      counter = 1+m_send%task_int(1)+1
      tmc_params%cell%perd = m_send%task_int(2:counter-1)
      tmc_params%cell%symmetry_id = m_send%task_int(counter+1)
      tmc_params%cell%orthorhombic = .FALSE.
      IF (m_send%task_int(counter+2) .EQ. 1) tmc_params%cell%orthorhombic = .TRUE.
      counter = counter+3
      elem%mol(:) = m_send%task_int(counter+1:counter+m_send%task_int(counter))
      counter = counter+1+m_send%task_int(counter)
      IF (msg_type .EQ. TMC_STAT_INIT_ANALYSIS) THEN
         IF(.NOT.(PRESENT(result_count)))CALL cp__a("tmc/tmc_messages.F",721)
         IF(.NOT.(.NOT. ASSOCIATED(result_count)))CALL cp__a("tmc/tmc_messages.F",722)
         ALLOCATE (result_count(m_send%task_int(counter)))
         result_count(:) = m_send%task_int(counter+1:counter+m_send%task_int(counter))
         counter = counter+1+m_send%task_int(counter)
      END IF
      IF(.NOT.(counter .EQ. m_send%info(2)))CALL cp__a("tmc/tmc_messages.F",727)
      IF(.NOT.(m_send%task_int(counter) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",728)

      counter = 0
      !float array with pos, cell vectors, atom_mass
      counter = 2+INT(m_send%task_real(1))
      elem%pos = m_send%task_real(2:counter-1)
      flag = INT(m_send%task_real(counter)) .EQ. SIZE(tmc_params%cell%hmat)
      IF(.NOT.(flag))CALL cp__a("tmc/tmc_messages.F",735)
      tmc_params%cell%hmat = &
         RESHAPE(m_send%task_real(counter+1:counter+ &
                                  SIZE(tmc_params%cell%hmat)), (/3, 3/))
      counter = counter+1+INT(m_send%task_real(counter))

      CALL allocate_tmc_atom_type(atoms=tmc_params%atoms, &
                                  nr_atoms=INT(m_send%task_real(counter)))
      DO i = 1, SIZE(tmc_params%atoms)
         tmc_params%atoms(i)%mass = m_send%task_real(counter+i)
      END DO
      counter = counter+1+INT(m_send%task_real(counter))

      IF(.NOT.(counter .EQ. m_send%info(3)))CALL cp__a("tmc/tmc_messages.F",748)
      IF(.NOT.(INT(m_send%task_real(m_send%info(3))) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",749)

   END SUBROUTINE read_start_conf_message

   !============================================================================
   ! Energy messages
   !============================================================================
! **************************************************************************************************
!> \brief creating message for requesting exact energy of new configuration
!> \param elem tree element with new coordinates
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_energy_request_message(elem, m_send, &
                                            tmc_params)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_energy_request_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, msg_size_int, msg_size_real

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",774)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",775)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",776)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",777)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",778)

      counter = 0
      !first integer array
      msg_size_int = 1+1+1+1+1 ! 1+SIZE(elem%sub_tree_nr) +1+SIZE(elem%nr)
      ALLOCATE (m_send%task_int(msg_size_int))
      counter = 1
      m_send%task_int(counter) = 1 !SIZE(elem%sub_tree_nr)
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = elem%sub_tree_nr
      counter = counter+1+m_send%task_int(counter)
      m_send%task_int(counter) = 1 !SIZE(elem%nr)
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = elem%nr
      counter = counter+1+m_send%task_int(counter)
      m_send%task_int(counter) = message_end_flag
      IF(.NOT.(SIZE(m_send%task_int) .EQ. msg_size_int))CALL cp__a("tmc/tmc_messages.F",792)
      IF(.NOT.(m_send%task_int(msg_size_int) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",793)

      !then float array with pos
      msg_size_real = 1+SIZE(elem%pos)+1
      IF (tmc_params%pressure .GE. 0.0_dp) msg_size_real = msg_size_real+1+SIZE(elem%box_scale(:))
      ALLOCATE (m_send%task_real(msg_size_real))
      m_send%task_real(1) = SIZE(elem%pos)
      counter = 2+INT(m_send%task_real(1))
      m_send%task_real(2:counter-1) = elem%pos
      IF (tmc_params%pressure .GE. 0.0_dp) THEN
         m_send%task_real(counter) = SIZE(elem%box_scale)
         m_send%task_real(counter+1:counter+INT(m_send%task_real(counter))) = elem%box_scale(:)
         counter = counter+1+INT(m_send%task_real(counter))
      END IF
      m_send%task_real(counter) = REAL(message_end_flag, KIND=dp) !message end

      IF(.NOT.(SIZE(m_send%task_real) .EQ. msg_size_real))CALL cp__a("tmc/tmc_messages.F",809)
      IF(.NOT.(INT(m_send%task_real(msg_size_real)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",810)
   END SUBROUTINE create_energy_request_message

! **************************************************************************************************
!> \brief reading message for requesting exact energy of new configuration
!> \param elem tree element with new coordinates
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE read_energy_request_message(elem, m_send, tmc_params)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_energy_request_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",830)
      IF(.NOT.(m_send%info(3) .GT. 0))CALL cp__a("tmc/tmc_messages.F",831)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",832)
      IF(.NOT.(.NOT. ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",833)

      ! initialize the new sub tree element
      IF (.NOT. ASSOCIATED(elem)) THEN
         CALL allocate_new_sub_tree_node(next_el=elem, nr_dim=NINT(m_send%task_real(1)), &
                                         tmc_params=tmc_params)
      END IF
      ! read the integer values
      IF(.NOT.(m_send%info(2) .GT. 0))CALL cp__a("tmc/tmc_messages.F",841)
      counter = 1
      elem%sub_tree_nr = m_send%task_int(counter+1)
      counter = counter+1+m_send%task_int(counter)
      elem%nr = m_send%task_int(counter+1)
      counter = counter+1+m_send%task_int(counter)
      IF(.NOT.(m_send%task_int(counter) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",847)

      !float array with pos
      counter = 0
      counter = 1+NINT(m_send%task_real(1))
      elem%pos = m_send%task_real(2:counter)
      counter = counter+1
      IF (tmc_params%pressure .GE. 0.0_dp) THEN
         elem%box_scale(:) = m_send%task_real(counter+1:counter+INT(m_send%task_real(counter)))
         counter = counter+1+INT(m_send%task_real(counter))
      END IF

      IF(.NOT.(counter .EQ. m_send%info(3)))CALL cp__a("tmc/tmc_messages.F",859)
      IF(.NOT.(INT(m_send%task_real(m_send%info(3))) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",860)
   END SUBROUTINE read_energy_request_message

! **************************************************************************************************
!> \brief creating message for sending back the exact energy of new conf
!> \param elem tree element  with calculated energy
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_energy_result_message(elem, m_send, tmc_params)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_energy_result_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, msg_size_int, msg_size_real

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",880)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",881)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",882)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",883)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",884)

      counter = 0
      !first integer array
      msg_size_int = 0
      ! for checking the tree element mapping, send back the tree numbers
      IF (DEBUG .GT. 0) THEN
         msg_size_int = 1+1+1+1+1 ! 1+SIZE(elem%sub_tree_nr) +1+SIZE(elem%nr)
         ALLOCATE (m_send%task_int(msg_size_int))
         counter = 1
         m_send%task_int(counter) = 1 !SIZE(elem%sub_tree_nr)
         m_send%task_int(counter+1:counter+m_send%task_int(counter)) = elem%sub_tree_nr
         counter = counter+1+m_send%task_int(counter)
         m_send%task_int(counter) = 1 !SIZE(elem%nr)
         m_send%task_int(counter+1:counter+m_send%task_int(counter)) = elem%nr
         counter = counter+m_send%task_int(counter)+1
         m_send%task_int(counter) = message_end_flag !message end
      END IF

      !then float array with energy of exact potential
      msg_size_real = 1+1+1
      IF (tmc_params%print_forces) msg_size_real = msg_size_real+1+SIZE(elem%frc)
      IF (tmc_params%print_dipole) msg_size_real = msg_size_real+1+SIZE(elem%dipole)

      ALLOCATE (m_send%task_real(msg_size_real))
      m_send%task_real(1) = 1
      m_send%task_real(2) = elem%potential
      counter = 3
      IF (tmc_params%print_forces) THEN
         m_send%task_real(counter) = SIZE(elem%frc)
         m_send%task_real(counter+1:counter+NINT(m_send%task_real(counter))) = elem%frc
         counter = counter+NINT(m_send%task_real(counter))+1
      END IF
      IF (tmc_params%print_dipole) THEN
         m_send%task_real(counter) = SIZE(elem%dipole)
         m_send%task_real(counter+1:counter+NINT(m_send%task_real(counter))) = elem%dipole
         counter = counter+NINT(m_send%task_real(counter))+1
      END IF

      m_send%task_real(counter) = REAL(message_end_flag, KIND=dp) !message end

      IF(.NOT.(SIZE(m_send%task_real) .EQ. msg_size_real))CALL cp__a("tmc/tmc_messages.F",925)
      IF(.NOT.(INT(m_send%task_real(msg_size_real)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",926)
   END SUBROUTINE create_energy_result_message

! **************************************************************************************************
!> \brief reading message for sending back the exact energy of new conf
!> \param elem tree element for storing new energy
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE read_energy_result_message(elem, m_send, tmc_params)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_energy_result_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",946)
      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",947)
      IF(.NOT.(m_send%info(3) .GT. 0))CALL cp__a("tmc/tmc_messages.F",948)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",949)

      ! read the integer values
      ! for checking the tree element mapping, check the tree numbers
      IF (DEBUG .GT. 0) THEN
         counter = 1
         IF (elem%sub_tree_nr .NE. m_send%task_int(counter+1) .OR. &
             elem%nr .NE. m_send%task_int(counter+3)) THEN
            WRITE (*, *) "ERROR: read_energy_result: master got energy result of subtree elem ", &
               m_send%task_int(counter+1), m_send%task_int(counter+3), &
               " but expect result of subtree elem", elem%sub_tree_nr, elem%nr
            CALL cp__b("tmc/tmc_messages.F",960,"read_energy_result: got energy result from unexpected tree element.")
         END IF
      ELSE
         IF(.NOT.(m_send%info(2) .EQ. 0))CALL cp__a("tmc/tmc_messages.F",963)
      END IF

      !then float array with energy of exact potential
      elem%potential = m_send%task_real(2)
      counter = 3
      IF (tmc_params%print_forces) THEN
         elem%frc(:) = m_send%task_real((counter+1):(counter+NINT(m_send%task_real(counter))))
         counter = counter+1+NINT(m_send%task_real(counter))
      END IF
      IF (tmc_params%print_dipole) THEN
         elem%dipole(:) = m_send%task_real((counter+1):(counter+NINT(m_send%task_real(counter))))
         counter = counter+1+NINT(m_send%task_real(counter))
      END IF

      IF(.NOT.(counter .EQ. m_send%info(3)))CALL cp__a("tmc/tmc_messages.F",978)
      IF(.NOT.(INT(m_send%task_real(m_send%info(3))) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",979)
   END SUBROUTINE read_energy_result_message

! **************************************************************************************************
!> \brief create message for sending back the approximate energy of new conf
!> \param elem tree element with calculated approx energy
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_approx_energy_result_message(elem, m_send, &
                                                  tmc_params)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_approx_energy_result_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, msg_size_real

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1000)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",1001)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",1002)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",1003)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1004)

      counter = 0

      !then float array with energy of exact potential
      msg_size_real = 1+1+1
      IF (tmc_params%pressure .GE. 0.0_dp) msg_size_real = msg_size_real+1+SIZE(elem%box_scale(:))

      ALLOCATE (m_send%task_real(msg_size_real))
      m_send%task_real(1) = 1
      m_send%task_real(2) = elem%e_pot_approx
      counter = 3
      ! the box size for NpT
      IF (tmc_params%pressure .GE. 0.0_dp) THEN
         m_send%task_real(counter) = SIZE(elem%box_scale)
         m_send%task_real(counter+1:counter+INT(m_send%task_real(counter))) = elem%box_scale(:)
         counter = counter+1+INT(m_send%task_real(counter))
      END IF
      m_send%task_real(counter) = REAL(message_end_flag, KIND=dp) !message end

      IF(.NOT.(SIZE(m_send%task_real) .EQ. msg_size_real))CALL cp__a("tmc/tmc_messages.F",1024)
      IF(.NOT.(INT(m_send%task_real(msg_size_real)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1025)
   END SUBROUTINE create_approx_energy_result_message

! **************************************************************************************************
!> \brief reading message for sending back the exact energy of new conf
!> \param elem tree element for storing new energy
!> \param m_send the message structure
!> \param tmc_params the param struct with necessary parameters
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE read_approx_energy_result(elem, m_send, tmc_params)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_approx_energy_result', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",1045)
      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1046)
      IF(.NOT.(m_send%info(2) .EQ. 0 .AND. m_send%info(3) .GT. 0))CALL cp__a("tmc/tmc_messages.F",1047)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1048)

      !then float array with energy of exact potential
      elem%e_pot_approx = m_send%task_real(2)
      counter = 3
      IF (tmc_params%pressure .GE. 0.0_dp) THEN
         elem%box_scale(:) = m_send%task_real(counter+1:counter+INT(m_send%task_real(counter)))
         counter = counter+1+INT(m_send%task_real(counter))
      END IF

      IF(.NOT.(counter .EQ. m_send%info(3)))CALL cp__a("tmc/tmc_messages.F",1058)
      IF(.NOT.(INT(m_send%task_real(m_send%info(3))) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1059)
   END SUBROUTINE read_approx_energy_result

   !============================================================================
   ! Nested Monte Carlo request messages
   !============================================================================
! **************************************************************************************************
!> \brief creating message for Nested Monte Carlo sampling of new configuration
!> \param msg_type the status tag
!> \param elem tree element  with calculated energy
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_NMC_request_massage(msg_type, elem, m_send, &
                                         tmc_params)
      INTEGER                                            :: msg_type
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_NMC_request_massage', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, msg_size_int, msg_size_real

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1085)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",1086)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",1087)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",1088)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1089)

      counter = 0
      !first integer array with element status,mol_info, move type, sub tree, element nr, temp index
      msg_size_int = 1+SIZE(elem%elem_stat)+1+SIZE(elem%mol)+1+1+1+1+1+1+1+1+1

      ALLOCATE (m_send%task_int(msg_size_int))
      ! element status
      m_send%task_int(1) = SIZE(elem%elem_stat)
      counter = 2+m_send%task_int(1)
      m_send%task_int(2:counter-1) = elem%elem_stat
      m_send%task_int(counter) = SIZE(elem%mol)
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = elem%mol(:)
      counter = counter+1+m_send%task_int(counter)
      ! element move type
      m_send%task_int(counter) = 1
      m_send%task_int(counter+1) = elem%move_type
      counter = counter+2
      m_send%task_int(counter) = 1
      m_send%task_int(counter+1) = elem%nr
      counter = counter+2
      m_send%task_int(counter) = 1
      m_send%task_int(counter+1) = elem%sub_tree_nr
      counter = counter+2
      m_send%task_int(counter) = 1
      m_send%task_int(counter+1) = elem%temp_created
      m_send%task_int(counter+2) = message_end_flag !message end

      counter = 0
      !then float array with pos, (vel), random number seed, subbox_center
      msg_size_real = 1+SIZE(elem%pos)+1+SIZE(elem%rng_seed)+1+SIZE(elem%subbox_center(:))+1
      IF (msg_type .EQ. TMC_STAT_MD_REQUEST .OR. msg_type .EQ. TMC_STAT_MD_BROADCAST) &
         msg_size_real = msg_size_real+1+SIZE(elem%vel) ! the velocities
      IF (tmc_params%pressure .GE. 0.0_dp) msg_size_real = msg_size_real+1+SIZE(elem%box_scale(:)) ! box size for NpT

      ALLOCATE (m_send%task_real(msg_size_real))
      m_send%task_real(1) = SIZE(elem%pos)
      counter = 2+INT(m_send%task_real(1))
      m_send%task_real(2:counter-1) = elem%pos
      IF (msg_type .EQ. TMC_STAT_MD_REQUEST .OR. msg_type .EQ. TMC_STAT_MD_BROADCAST) THEN
         m_send%task_real(counter) = SIZE(elem%vel)
         m_send%task_real(counter+1:counter+NINT(m_send%task_real(counter))) = elem%vel
         counter = counter+1+NINT(m_send%task_real(counter))
      END IF
      ! rng seed
      m_send%task_real(counter) = SIZE(elem%rng_seed)
      m_send%task_real(counter+1:counter+SIZE(elem%rng_seed)) = RESHAPE(elem%rng_seed(:, :, :), (/SIZE(elem%rng_seed)/))
      counter = counter+NINT(m_send%task_real(counter))+1
      ! sub box center
      m_send%task_real(counter) = SIZE(elem%subbox_center(:))
      m_send%task_real(counter+1:counter+SIZE(elem%subbox_center)) = elem%subbox_center(:)
      counter = counter+1+NINT(m_send%task_real(counter))
      ! the box size for NpT
      IF (tmc_params%pressure .GE. 0.0_dp) THEN
         m_send%task_real(counter) = SIZE(elem%box_scale)
         m_send%task_real(counter+1:counter+INT(m_send%task_real(counter))) = elem%box_scale(:)
         counter = counter+1+INT(m_send%task_real(counter))
      END IF
      m_send%task_real(counter) = message_end_flag !message end

      IF(.NOT.(SIZE(m_send%task_int) .EQ. msg_size_int))CALL cp__a("tmc/tmc_messages.F",1149)
      IF(.NOT.(SIZE(m_send%task_real) .EQ. msg_size_real))CALL cp__a("tmc/tmc_messages.F",1150)
      IF(.NOT.(m_send%task_int(msg_size_int) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1151)
      IF(.NOT.(INT(m_send%task_real(msg_size_real)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1152)
   END SUBROUTINE create_NMC_request_massage

! **************************************************************************************************
!> \brief reading message for Nested Monte Carlo sampling of new configuration
!> \param msg_type the status tag
!> \param elem tree element with new coordinates
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE read_NMC_request_massage(msg_type, elem, m_send, &
                                       tmc_params)
      INTEGER                                            :: msg_type
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_NMC_request_massage', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, num_dim, rnd_seed_size

      IF(.NOT.(.NOT. ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",1175)
      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1176)
      IF(.NOT.(m_send%info(2) .GT. 5 .AND. m_send%info(3) .GT. 8))CALL cp__a("tmc/tmc_messages.F",1177)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1178)

      counter = 0
      !first integer array with number of dimentions and random seed size
      rnd_seed_size = m_send%task_int(1+m_send%task_int(1)+1)

      IF (.NOT. ASSOCIATED(elem)) THEN
         CALL allocate_new_sub_tree_node(next_el=elem, nr_dim=NINT(m_send%task_real(1)), &
                                         tmc_params=tmc_params)
      END IF
      ! element status
      counter = 2+m_send%task_int(1)
      elem%elem_stat = m_send%task_int(2:counter-1)
      elem%mol(:) = m_send%task_int(counter+1:counter+m_send%task_int(counter))
      counter = counter+1+m_send%task_int(counter)
      ! element move type
      elem%move_type = m_send%task_int(counter+1)
      counter = counter+2
      elem%nr = m_send%task_int(counter+1)
      counter = counter+2
      elem%sub_tree_nr = m_send%task_int(counter+1)
      counter = counter+2
      elem%temp_created = m_send%task_int(counter+1)
      counter = counter+2
      IF(.NOT.(counter .EQ. m_send%info(2)))CALL cp__a("tmc/tmc_messages.F",1202)

      counter = 0
      !then float array with pos, (vel), subbox_center and temp
      num_dim = NINT(m_send%task_real(1))
      counter = 2+INT(m_send%task_real(1))
      elem%pos = m_send%task_real(2:counter-1)
      IF (msg_type .EQ. TMC_STAT_MD_REQUEST .OR. msg_type .EQ. TMC_STAT_MD_BROADCAST) THEN
         elem%vel = m_send%task_real(counter+1:counter+NINT(m_send%task_real(counter)))
         counter = counter+NINT(m_send%task_real(counter))+1
      END IF
      ! rng seed
      elem%rng_seed(:, :, :) = RESHAPE(m_send%task_real(counter+1:counter+SIZE(elem%rng_seed)), (/3, 2, 3/))
      counter = counter+NINT(m_send%task_real(counter))+1
      ! sub box center
      elem%subbox_center(:) = m_send%task_real(counter+1:counter+INT(m_send%task_real(counter)))
      counter = counter+1+NINT(m_send%task_real(counter))

      IF (tmc_params%pressure .GE. 0.0_dp) THEN
         elem%box_scale(:) = m_send%task_real(counter+1:counter+INT(m_send%task_real(counter)))
         counter = counter+1+INT(m_send%task_real(counter))
      ELSE
         elem%box_scale(:) = 1.0_dp
      END IF

      IF(.NOT.(counter .EQ. m_send%info(3)))CALL cp__a("tmc/tmc_messages.F",1227)
      IF(.NOT.(m_send%task_int(m_send%info(2)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1228)
      IF(.NOT.(INT(m_send%task_real(m_send%info(3))) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1229)
   END SUBROUTINE read_NMC_request_massage

   !============================================================================
   ! Nested Monte Carlo RESULT messages
   !============================================================================
! **************************************************************************************************
!> \brief creating message for Nested Monte Carlo sampling result
!> \param msg_type the status tag
!> \param elem tree element  with calculated energy
!> \param m_send the message structure
!> \param tmc_params environment with move types and sizes
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_NMC_result_massage(msg_type, elem, m_send, tmc_params)
      INTEGER                                            :: msg_type
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_NMC_result_massage', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, msg_size_int, msg_size_real

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1254)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",1255)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",1256)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",1257)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1258)

      !first integer array with status, nmc_acc_counts, subbox_acc_count and (subbox rejectance)
      msg_size_int = 1+SIZE(elem%mol) &
                     +1+SIZE(tmc_params%nmc_move_types%mv_count) &
                     +1+SIZE(tmc_params%nmc_move_types%acc_count)+1
      IF (DEBUG .GT. 0) msg_size_int = msg_size_int+1+1+1+1
      IF (.NOT. ANY(tmc_params%sub_box_size .LE. 0.1_dp)) &
         msg_size_int = msg_size_int+1+SIZE(tmc_params%nmc_move_types%subbox_count) &
                        +1+SIZE(tmc_params%nmc_move_types%subbox_acc_count)

      ALLOCATE (m_send%task_int(msg_size_int))
      counter = 1
      IF (DEBUG .GT. 0) THEN
         ! send the element number back
         m_send%task_int(counter) = 1
         m_send%task_int(counter+1) = elem%sub_tree_nr
         counter = counter+1+m_send%task_int(counter)
         m_send%task_int(counter) = 1
         m_send%task_int(counter+1) = elem%nr
         counter = counter+1+m_send%task_int(counter)
      END IF
      ! the molecule information
      m_send%task_int(counter) = SIZE(elem%mol)
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = elem%mol(:)
      counter = counter+1+m_send%task_int(counter)
      ! the counters for each move type
      m_send%task_int(counter) = SIZE(tmc_params%nmc_move_types%mv_count)
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = &
         RESHAPE(tmc_params%nmc_move_types%mv_count(:, :), &
                 (/SIZE(tmc_params%nmc_move_types%mv_count)/))
      counter = counter+1+m_send%task_int(counter)
      ! the counter for the accepted moves
      m_send%task_int(counter) = SIZE(tmc_params%nmc_move_types%acc_count)
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = &
         RESHAPE(tmc_params%nmc_move_types%acc_count(:, :), &
                 (/SIZE(tmc_params%nmc_move_types%acc_count)/))
      counter = counter+1+m_send%task_int(counter)
      ! amount of rejected subbox moves
      IF (.NOT. ANY(tmc_params%sub_box_size .LE. 0.1_dp)) THEN
         m_send%task_int(counter) = SIZE(tmc_params%nmc_move_types%subbox_count)
         m_send%task_int(counter+1:counter+m_send%task_int(counter)) = &
            RESHAPE(tmc_params%nmc_move_types%subbox_count(:, :), &
                    (/SIZE(tmc_params%nmc_move_types%subbox_count)/))
         counter = counter+1+m_send%task_int(counter)
         m_send%task_int(counter) = SIZE(tmc_params%nmc_move_types%subbox_acc_count)
         m_send%task_int(counter+1:counter+m_send%task_int(counter)) = &
            RESHAPE(tmc_params%nmc_move_types%subbox_acc_count(:, :), &
                    (/SIZE(tmc_params%nmc_move_types%subbox_acc_count)/))
         counter = counter+1+m_send%task_int(counter)
      END IF
      m_send%task_int(counter) = message_end_flag ! message end

      counter = 0
      !then float array with pos,(vel, e_kin_befor_md, ekin),(forces),rng_seed,
      !                       potential,e_pot_approx,acc_prob,subbox_prob
      msg_size_real = 1+SIZE(elem%pos) & ! pos
                      +1+SIZE(elem%rng_seed) & ! rng_seed
                      +1+1 & ! potential
                      +1+1 & ! e_pot_approx
                      +1 ! check bit

      IF (msg_type .EQ. TMC_STAT_MD_REQUEST .OR. msg_type .EQ. TMC_STAT_MD_RESULT .OR. &
          msg_type .EQ. TMC_STAT_MD_BROADCAST) &
         msg_size_real = msg_size_real+1+SIZE(elem%vel)+1+1+1+1 ! for MD also: vel, e_kin_befor_md, ekin

      ALLOCATE (m_send%task_real(msg_size_real))
      ! pos
      counter = 1
      m_send%task_real(counter) = SIZE(elem%pos)
      m_send%task_real(counter+1:counter+NINT(m_send%task_real(counter))) = elem%pos
      counter = counter+1+NINT(m_send%task_real(counter))
      ! rng seed
      m_send%task_real(counter) = SIZE(elem%rng_seed)
      m_send%task_real(counter+1:counter+SIZE(elem%rng_seed)) = &
         RESHAPE(elem%rng_seed(:, :, :), (/SIZE(elem%rng_seed)/))
      counter = counter+1+NINT(m_send%task_real(counter))
      ! potential
      m_send%task_real(counter) = 1
      m_send%task_real(counter+1) = elem%potential
      counter = counter+2
      ! approximate potential energy
      m_send%task_real(counter) = 1
      m_send%task_real(counter+1) = elem%e_pot_approx
      counter = counter+2
      ! for MD also: vel, e_kin_befor_md, ekin
      IF (msg_type .EQ. TMC_STAT_MD_REQUEST .OR. msg_type .EQ. TMC_STAT_MD_RESULT .OR. &
          msg_type .EQ. TMC_STAT_MD_BROADCAST) THEN
         m_send%task_real(counter) = SIZE(elem%vel)
         m_send%task_real(counter+1:counter+NINT(m_send%task_real(counter))) = elem%vel
         counter = counter+1+INT(m_send%task_real(counter))
         m_send%task_real(counter) = 1
         m_send%task_real(counter+1) = elem%ekin_before_md
         counter = counter+2
         m_send%task_real(counter) = 1
         m_send%task_real(counter+1) = elem%ekin
         counter = counter+2
      END IF
      m_send%task_real(counter) = message_end_flag ! message end

      IF(.NOT.(SIZE(m_send%task_int) .EQ. msg_size_int))CALL cp__a("tmc/tmc_messages.F",1358)
      IF(.NOT.(SIZE(m_send%task_real) .EQ. msg_size_real))CALL cp__a("tmc/tmc_messages.F",1359)
      IF(.NOT.(m_send%task_int(msg_size_int) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1360)
      IF(.NOT.(INT(m_send%task_real(msg_size_real)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1361)
   END SUBROUTINE create_NMC_result_massage

! **************************************************************************************************
!> \brief reading message for Nested Monte Carlo sampling result
!> \param msg_type the status tag
!> \param elem tree element  with calculated energy
!> \param m_send the message structure
!> \param tmc_params environment with move types and sizes
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE read_NMC_result_massage(msg_type, elem, m_send, tmc_params)
      INTEGER                                            :: msg_type
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_NMC_result_massage', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter
      INTEGER, DIMENSION(:, :), POINTER                  :: acc_counter, mv_counter, &
                                                            subbox_acc_counter, subbox_counter

      NULLIFY (mv_counter, subbox_counter, acc_counter, subbox_acc_counter)

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",1387)
      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1388)
      IF(.NOT.(m_send%info(2) .GT. 0 .AND. m_send%info(3) .GT. 0))CALL cp__a("tmc/tmc_messages.F",1389)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1390)

      !first integer array with element status, random number seed, and move type
      counter = 1
      IF (DEBUG .GT. 0) THEN
         IF ((m_send%task_int(counter+1) .NE. elem%sub_tree_nr) .AND. (m_send%task_int(counter+3) .NE. elem%nr)) THEN
            CALL cp__b("tmc/tmc_messages.F",1396,"read_NMC_result_massage: got result of wrong element")
         ENDIF
         counter = counter+2+2
      END IF
      ! the molecule information
      elem%mol(:) = m_send%task_int(counter+1:counter+m_send%task_int(counter))
      counter = counter+1+m_send%task_int(counter)
      ! the counters for each move type
      ALLOCATE (mv_counter(0:SIZE(tmc_params%nmc_move_types%mv_count(:, 1))-1, &
                           SIZE(tmc_params%nmc_move_types%mv_count(1, :))))
      mv_counter(:, :) = RESHAPE(m_send%task_int(counter+1:counter+m_send%task_int(counter)), &
                                 (/SIZE(tmc_params%nmc_move_types%mv_count(:, 1)), &
                                   SIZE(tmc_params%nmc_move_types%mv_count(1, :))/))
      counter = counter+1+m_send%task_int(counter)
      ! the counter for the accepted moves
      ALLOCATE (acc_counter(0:SIZE(tmc_params%nmc_move_types%acc_count(:, 1))-1, &
                            SIZE(tmc_params%nmc_move_types%acc_count(1, :))))
      acc_counter(:, :) = RESHAPE(m_send%task_int(counter+1:counter+m_send%task_int(counter)), &
                                  (/SIZE(tmc_params%nmc_move_types%acc_count(:, 1)), &
                                    SIZE(tmc_params%nmc_move_types%acc_count(1, :))/))
      counter = counter+1+m_send%task_int(counter)
      ! amount of rejected subbox moves
      IF (.NOT. ANY(tmc_params%sub_box_size .LE. 0.1_dp)) THEN
         ALLOCATE (subbox_counter(SIZE(tmc_params%nmc_move_types%subbox_count(:, 1)), &
                                  SIZE(tmc_params%nmc_move_types%subbox_count(1, :))))
         subbox_counter(:, :) = RESHAPE(m_send%task_int(counter+1:counter+m_send%task_int(counter)), &
                                        (/SIZE(tmc_params%nmc_move_types%subbox_count(:, 1)), &
                                          SIZE(tmc_params%nmc_move_types%subbox_count(1, :))/))
         counter = counter+1+m_send%task_int(counter)
         ALLOCATE (subbox_acc_counter(SIZE(tmc_params%nmc_move_types%subbox_acc_count(:, 1)), &
                                      SIZE(tmc_params%nmc_move_types%subbox_acc_count(1, :))))
         subbox_acc_counter(:, :) = RESHAPE(m_send%task_int(counter+1:counter+m_send%task_int(counter)), &
                                            (/SIZE(tmc_params%nmc_move_types%subbox_acc_count(:, 1)), &
                                              SIZE(tmc_params%nmc_move_types%subbox_acc_count(1, :))/))
         counter = counter+1+m_send%task_int(counter)
      END IF
      IF(.NOT.(counter .EQ. m_send%info(2)))CALL cp__a("tmc/tmc_messages.F",1432)

      counter = 0
      !then float array with pos, (vel, e_kin_befor_md, ekin), (forces), rng_seed, potential, e_pot_approx
      counter = 1
      ! pos
      elem%pos = m_send%task_real(counter+1:counter+NINT(m_send%task_real(counter)))
      counter = counter+1+NINT(m_send%task_real(counter))
      ! rng seed
      elem%rng_seed(:, :, :) = RESHAPE(m_send%task_real(counter+1:counter+SIZE(elem%rng_seed)), (/3, 2, 3/))
      counter = counter+1+NINT(m_send%task_real(counter))
      ! potential
      elem%potential = m_send%task_real(counter+1)
      counter = counter+2
      ! approximate potential energy
      elem%e_pot_approx = m_send%task_real(counter+1)
      counter = counter+2
      ! for MD also: vel, e_kin_befor_md, ekin
      IF (msg_type .EQ. TMC_STAT_MD_REQUEST .OR. msg_type .EQ. TMC_STAT_MD_RESULT .OR. &
          msg_type .EQ. TMC_STAT_MD_BROADCAST) THEN
         elem%vel = m_send%task_real(counter+1:counter+NINT(m_send%task_real(counter)))
         counter = counter+1+INT(m_send%task_real(counter))
         IF (.NOT. (tmc_params%task_type .EQ. task_type_gaussian_adaptation)) &
            elem%ekin_before_md = m_send%task_real(counter+1)
         counter = counter+2
         elem%ekin = m_send%task_real(counter+1)
         counter = counter+2
      END IF

      CALL add_mv_prob(move_types=tmc_params%nmc_move_types, prob_opt=tmc_params%esimate_acc_prob, &
                       mv_counter=mv_counter, acc_counter=acc_counter)
      IF (.NOT. ANY(tmc_params%sub_box_size .LE. 0.1_dp)) THEN
         CALL add_mv_prob(move_types=tmc_params%nmc_move_types, prob_opt=tmc_params%esimate_acc_prob, &
                          subbox_counter=subbox_counter, subbox_acc_counter=subbox_acc_counter)
      END IF

      DEALLOCATE (mv_counter, acc_counter)
      IF (.NOT. ANY(tmc_params%sub_box_size .LE. 0.1_dp)) &
         DEALLOCATE (subbox_counter, subbox_acc_counter)
      IF(.NOT.(counter .EQ. m_send%info(3)))CALL cp__a("tmc/tmc_messages.F",1471)
      IF(.NOT.(m_send%task_int(m_send%info(2)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1472)
      IF(.NOT.(INT(m_send%task_real(m_send%info(3))) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1473)
   END SUBROUTINE read_NMC_result_massage

   !============================================================================
   ! Analysis element messages
   !============================================================================
! **************************************************************************************************
!> \brief creating message for requesting analysing a new configuration
!>        we plot temperatur index into the sub tree number and
!>        the Markov chain number into the element number
!> \param list_elem ...
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE create_analysis_request_message(list_elem, m_send, &
                                              tmc_params)
      TYPE(elem_list_type), POINTER                      :: list_elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_analysis_request_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, msg_size_int, msg_size_real

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1499)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_int)))CALL cp__a("tmc/tmc_messages.F",1500)
      IF(.NOT.(.NOT. ALLOCATED(m_send%task_real)))CALL cp__a("tmc/tmc_messages.F",1501)
      IF(.NOT.(ASSOCIATED(list_elem)))CALL cp__a("tmc/tmc_messages.F",1502)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1503)

      counter = 0
      !first integer array
      msg_size_int = 1+1+1+1+1 ! 1+SIZE(list_elem%elem%sub_tree_nr) +1+SIZE(list_elem%elem%nr)
      ALLOCATE (m_send%task_int(msg_size_int))
      counter = 1
      m_send%task_int(counter) = 1 ! temperature index
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = list_elem%temp_ind
      counter = counter+1+m_send%task_int(counter)
      m_send%task_int(counter) = 1 ! Markov chain number
      m_send%task_int(counter+1:counter+m_send%task_int(counter)) = list_elem%nr
      counter = counter+1+m_send%task_int(counter)
      m_send%task_int(counter) = message_end_flag
      IF(.NOT.(SIZE(m_send%task_int) .EQ. msg_size_int))CALL cp__a("tmc/tmc_messages.F",1517)
      IF(.NOT.(m_send%task_int(msg_size_int) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1518)

      !then float array with pos
      msg_size_real = 1+SIZE(list_elem%elem%pos)+1
      IF (tmc_params%pressure .GE. 0.0_dp) msg_size_real = msg_size_real+1+SIZE(list_elem%elem%box_scale(:))
      ALLOCATE (m_send%task_real(msg_size_real))
      m_send%task_real(1) = SIZE(list_elem%elem%pos)
      counter = 2+INT(m_send%task_real(1))
      m_send%task_real(2:counter-1) = list_elem%elem%pos
      IF (tmc_params%pressure .GE. 0.0_dp) THEN
         m_send%task_real(counter) = SIZE(list_elem%elem%box_scale)
         m_send%task_real(counter+1:counter+INT(m_send%task_real(counter))) = list_elem%elem%box_scale(:)
         counter = counter+1+INT(m_send%task_real(counter))
      END IF
      m_send%task_real(counter) = REAL(message_end_flag, KIND=dp) !message end

      IF(.NOT.(SIZE(m_send%task_real) .EQ. msg_size_real))CALL cp__a("tmc/tmc_messages.F",1534)
      IF(.NOT.(INT(m_send%task_real(msg_size_real)) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1535)
   END SUBROUTINE create_analysis_request_message

! **************************************************************************************************
!> \brief reading message for requesting exact energy of new configuration
!> \param elem tree element with new coordinates
!> \param m_send the message structure
!> \param tmc_params stuct with parameters (global settings)
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE read_analysis_request_message(elem, m_send, tmc_params)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_analysis_request_message', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter

      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1555)
      IF(.NOT.(m_send%info(3) .GT. 0))CALL cp__a("tmc/tmc_messages.F",1556)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1557)
      IF(.NOT.(.NOT. ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",1558)

      ! initialize the new sub tree element
      IF (.NOT. ASSOCIATED(elem)) THEN
         CALL allocate_new_sub_tree_node(next_el=elem, nr_dim=NINT(m_send%task_real(1)), &
                                         tmc_params=tmc_params)
      END IF
      ! read the integer values
      IF(.NOT.(m_send%info(2) .GT. 0))CALL cp__a("tmc/tmc_messages.F",1566)
      counter = 1
      elem%sub_tree_nr = m_send%task_int(counter+1)
      counter = counter+1+m_send%task_int(counter)
      elem%nr = m_send%task_int(counter+1)
      counter = counter+1+m_send%task_int(counter)
      IF(.NOT.(m_send%task_int(counter) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1572)

      !float array with pos
      counter = 0
      counter = 1+NINT(m_send%task_real(1))
      elem%pos = m_send%task_real(2:counter)
      counter = counter+1
      IF (tmc_params%pressure .GE. 0.0_dp) THEN
         elem%box_scale(:) = m_send%task_real(counter+1:counter+INT(m_send%task_real(counter)))
         counter = counter+1+INT(m_send%task_real(counter))
      END IF

      IF(.NOT.(counter .EQ. m_send%info(3)))CALL cp__a("tmc/tmc_messages.F",1584)
      IF(.NOT.(INT(m_send%task_real(m_send%info(3))) .EQ. message_end_flag))CALL cp__a("tmc/tmc_messages.F",1585)
   END SUBROUTINE read_analysis_request_message

   !============================================================================
   ! SCF step energies (receiving from CP2K)
   !============================================================================
! **************************************************************************************************
!> \brief routine cancel the other group participants
!> \param elem tree element  with approximated energy
!> \param m_send the message structure
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE read_scf_step_ener(elem, m_send)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(message_send), POINTER                        :: m_send

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_scf_step_ener', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_messages.F",1604)
      IF(.NOT.(ASSOCIATED(m_send)))CALL cp__a("tmc/tmc_messages.F",1605)

      elem%scf_energies(MOD(elem%scf_energies_count, 4)+1) = m_send%task_real(1)
      elem%scf_energies_count = elem%scf_energies_count+1

   END SUBROUTINE read_scf_step_ener

   !============================================================================
   ! message for broadcasting atom types
   !============================================================================
! **************************************************************************************************
!> \brief routines send atom names to the global master
!>        (using broadcast in a specialized group consisting of the master
!>        and the first energy worker master)
!> \param atoms ...
!> \param source ...
!> \param para_env the communicator environment
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE communicate_atom_types(atoms, source, para_env)
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms
      INTEGER                                            :: source
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'communicate_atom_types', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: msg(:)
      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("tmc/tmc_messages.F",1636)
      IF(.NOT.(source .GE. 0))CALL cp__a("tmc/tmc_messages.F",1637)
      IF(.NOT.(source .LT. para_env%num_pe))CALL cp__a("tmc/tmc_messages.F",1638)

      ALLOCATE (msg(SIZE(atoms)))
      IF (para_env%mepos .EQ. source) THEN
         DO i = 1, SIZE(atoms)
            msg(i) = atoms(i)%name
         END DO
         CALL mp_bcast(msg, source, para_env%group)
      ELSE
         CALL mp_bcast(msg, source, para_env%group)
         DO i = 1, SIZE(atoms)
            atoms(i)%name = msg(i)
         END DO
      END IF
      DEALLOCATE (msg)
   END SUBROUTINE communicate_atom_types

   !============================================================================
   ! message for cancelation or stopping
   !============================================================================
! **************************************************************************************************
!> \brief send stop command to all group participants
!> \param para_env ...
!> \param worker_info ...
!> \param tmc_params ...
!> \param
!> \param
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE stop_whole_group(para_env, worker_info, tmc_params)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(elem_array_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: worker_info
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'stop_whole_group', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: act_rank, dest_rank, stat
      LOGICAL                                            :: flag
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: rank_stoped

!    INTEGER, DIMENSION(MPI_STATUS_SIZE)      :: status_single

      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("tmc/tmc_messages.F",1682)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_messages.F",1683)

      ALLOCATE (rank_stoped(0:para_env%num_pe-1))
      rank_stoped(:) = .FALSE.
      rank_stoped(para_env%mepos) = .TRUE.

      ! global master
      IF (PRESENT(worker_info)) THEN
         IF(.NOT.(ASSOCIATED(worker_info)))CALL cp__a("tmc/tmc_messages.F",1691)
         ! canceling running jobs and stop workers
         worker_group_loop: DO dest_rank = 1, para_env%num_pe-1
            ! busy workers have to be canceled
            IF (worker_info(dest_rank)%busy) THEN
               stat = TMC_CANCELING_MESSAGE
               act_rank = dest_rank
               CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=act_rank, &
                                para_env=para_env, tmc_params=tmc_params)
            ELSE
               ! send stop message
               stat = TMC_STATUS_FAILED
               act_rank = dest_rank
               CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=act_rank, &
                                para_env=para_env, tmc_params=tmc_params)
            END IF
         END DO worker_group_loop
      ELSE
         ! group master send stop message to all participants
         stat = TMC_STATUS_FAILED
         CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=bcast_group, &
                          para_env=para_env, tmc_params=tmc_params)
      END IF

      ! receive stop message receipt
      IF (para_env%mepos .EQ. MASTER_COMM_ID) THEN
         wait_for_receipts: DO
            ! check incomming messages
            stat = TMC_STATUS_WAIT_FOR_NEW_TASK
            dest_rank = 999
            IF (PRESENT(worker_info)) THEN
               ! mast have to be able to receive results, if canceling was too late
               CALL tmc_message(msg_type=stat, send_recv=recv_msg, dest=dest_rank, &
                                para_env=para_env, tmc_params=tmc_params, &
                                elem_array=worker_info(:), success=flag)
            ELSE
               CALL tmc_message(msg_type=stat, send_recv=recv_msg, dest=dest_rank, &
                                para_env=para_env, tmc_params=tmc_params)
            END IF
            SELECT CASE (stat)
            CASE (TMC_STATUS_WAIT_FOR_NEW_TASK)
               ! no message received
            CASE (TMC_CANCELING_RECEIPT)
               IF (PRESENT(worker_info)) THEN
                  worker_info(dest_rank)%busy = .FALSE.
                  stat = TMC_STATUS_FAILED
                  ! send stop message
                  CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=dest_rank, &
                                   para_env=para_env, tmc_params=tmc_params)
               ELSE
                  CALL cp__b("tmc/tmc_messages.F",1741,"group master should not receive cancel receipt")
               END IF
            CASE (TMC_STATUS_STOP_RECEIPT)
               rank_stoped(dest_rank) = .TRUE.
            CASE (TMC_STAT_ENERGY_RESULT, TMC_STAT_NMC_RESULT, TMC_STAT_MD_RESULT, &
                  TMC_STAT_SCF_STEP_ENER_RECEIVE, TMC_STAT_APPROX_ENERGY_RESULT, TMC_STAT_ANALYSIS_RESULT)
               ! nothing to do, canceling message already sent
            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_messages.F",1749), &
                             "master received status "//cp_to_string(stat)// &
                             " while stopping workers")
            END SELECT
            IF (ALL(rank_stoped)) EXIT wait_for_receipts
         END DO wait_for_receipts
      ELSE
         CALL cp__b("tmc/tmc_messages.F",1756,"only (group) master should stop other participants")
      END IF
   END SUBROUTINE stop_whole_group

END MODULE tmc_messages
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module contains the master routine handling the tree creation,
!>        communication with workers and task distribution
!>        For each idle working group the master creates a new global tree
!>        element, and if neccessay a related sub tree element,
!>        OR find the next element to calculate the exact energy.
!>        Goal is to keep at least the exact energy calculation working groups
!>        as busy as possible.
!>        Master also checks for incomming results and update the tree and the
!>        acceptance ratios.
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_master
   USE cell_types,                      ONLY: init_cell
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE global_types,                    ONLY: global_environment_type
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE machine,                         ONLY: m_flush,&
                                              m_memory,&
                                              m_walltime
   USE tmc_calculations,                ONLY: get_subtree_efficiency
   USE tmc_cancelation,                 ONLY: free_cancelation_list
   USE tmc_dot_tree,                    ONLY: create_dot_color,&
                                              create_global_tree_dot_color,&
                                              finalize_draw_tree,&
                                              init_draw_trees
   USE tmc_file_io,                     ONLY: print_restart_file,&
                                              write_element_in_file
   USE tmc_messages,                    ONLY: communicate_atom_types,&
                                              recv_msg,&
                                              send_msg,&
                                              stop_whole_group,&
                                              tmc_message
   USE tmc_move_handle,                 ONLY: check_moves,&
                                              print_move_types
   USE tmc_stati,                       ONLY: &
        TMC_CANCELING_MESSAGE, TMC_CANCELING_RECEIPT, TMC_STATUS_FAILED, &
        TMC_STATUS_WAIT_FOR_NEW_TASK, TMC_STATUS_WORKER_INIT, TMC_STAT_ANALYSIS_REQUEST, &
        TMC_STAT_ANALYSIS_RESULT, TMC_STAT_APPROX_ENERGY_REQUEST, TMC_STAT_APPROX_ENERGY_RESULT, &
        TMC_STAT_ENERGY_REQUEST, TMC_STAT_ENERGY_RESULT, TMC_STAT_INIT_ANALYSIS, &
        TMC_STAT_MD_REQUEST, TMC_STAT_MD_RESULT, TMC_STAT_NMC_REQUEST, TMC_STAT_NMC_RESULT, &
        TMC_STAT_SCF_STEP_ENER_RECEIVE, TMC_STAT_START_CONF_REQUEST, TMC_STAT_START_CONF_RESULT
   USE tmc_tree_acceptance,             ONLY: check_acceptance_of_depending_subtree_nodes,&
                                              check_elements_for_acc_prob_update,&
                                              tree_update
   USE tmc_tree_build,                  ONLY: create_new_gt_tree_node,&
                                              deallocate_sub_tree_node,&
                                              finalize_init,&
                                              finalize_trees,&
                                              init_tree_mod,&
                                              remove_all_trees
   USE tmc_tree_search,                 ONLY: count_nodes_in_trees,&
                                              count_prepared_nodes_in_trees,&
                                              search_next_energy_calc
   USE tmc_tree_types,                  ONLY: &
        elem_array_type, elem_list_type, global_tree_type, status_accepted, &
        status_calc_approx_ener, status_calculate_MD, status_calculate_NMC_steps, &
        status_calculate_energy, status_calculated, status_cancel_ener, status_cancel_nmc, &
        status_canceled_ener, status_canceled_nmc, status_created, status_rejected, tree_type
   USE tmc_types,                       ONLY: tmc_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_master'

   PUBLIC :: do_tmc_master

   INTEGER, PARAMETER :: DEBUG = 0

CONTAINS

! **************************************************************************************************
!> \brief send cancel request to all workers processing elements in the list
!> \param cancel_list list with elements to cancel
!> \param work_list list with all elements processed by working groups
!> \param cancel_count counter of canceled elements
!> \param para_env communication environment
!> \param tmc_env ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE cancel_calculations(cancel_list, work_list, cancel_count, &
                                  para_env, tmc_env)
      TYPE(elem_list_type), POINTER                      :: cancel_list
      TYPE(elem_array_type), DIMENSION(:), POINTER       :: work_list
      INTEGER                                            :: cancel_count
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cancel_calculations', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, stat, wg
      TYPE(elem_list_type), POINTER                      :: tmp_element

      IF (.NOT. ASSOCIATED(cancel_list)) RETURN
      NULLIFY (tmp_element)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_master.F",112)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_master.F",113)
      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_master.F",114)
      IF(.NOT.(ASSOCIATED(work_list)))CALL cp__a("tmc/tmc_master.F",115)
      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("tmc/tmc_master.F",116)

      stat = TMC_STATUS_FAILED
      wg = -1
      cancel_elem_loop: DO
         ! find certain working group calculating this element
         working_elem_loop: DO i = 1, SIZE(work_list)
            ! in special cases element could be distributed to several working groups,
            ! but all, exept of one, should already be in canceling process
            IF ((.NOT. work_list(i)%canceled) .AND. &
                ASSOCIATED(work_list(i)%elem)) THEN
               IF (ASSOCIATED(cancel_list%elem, work_list(i)%elem)) THEN
                  stat = TMC_CANCELING_MESSAGE
                  wg = i
                  EXIT working_elem_loop
               END IF
            END IF
         END DO working_elem_loop

         IF(.NOT.(wg .GE. 0))CALL cp__a("tmc/tmc_master.F",135)
         IF(.NOT.(stat .NE. TMC_STATUS_FAILED))CALL cp__a("tmc/tmc_master.F",136)
         IF(.NOT.(work_list(wg)%elem%stat .NE. status_calc_approx_ener))CALL cp__a("tmc/tmc_master.F",137)

         IF (DEBUG .GE. 1) &
            WRITE (tmc_env%m_env%io_unit, *) &
            "TMC|master: cancel group "//cp_to_string(wg)
         CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                          para_env=para_env, tmc_params=tmc_env%params)
         work_list(wg)%canceled = .TRUE.

         ! counting the amount of canceled elements
         cancel_count = cancel_count+1

         ! delete element from canceling list
         IF (.NOT. ASSOCIATED(cancel_list%next)) THEN
            DEALLOCATE (cancel_list)
            cancel_list => NULL()
            EXIT cancel_elem_loop
         ELSE
            tmp_element => cancel_list%next
            DEALLOCATE (cancel_list)
            cancel_list => tmp_element
         END IF
      END DO cancel_elem_loop
   END SUBROUTINE cancel_calculations

! **************************************************************************************************
!> \brief send analysis request to a worker
!> \param ana_list list with elements to be analysed
!> \param ana_worker_info ...
!> \param para_env communication environment
!> \param tmc_env ...
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE send_analysis_tasks(ana_list, ana_worker_info, para_env, tmc_env)
      TYPE(elem_list_type), POINTER                      :: ana_list
      TYPE(elem_array_type), DIMENSION(:), POINTER       :: ana_worker_info
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'send_analysis_tasks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dest, stat, wg
      TYPE(elem_list_type), POINTER                      :: list_tmp

      NULLIFY (list_tmp)

      IF(.NOT.(ASSOCIATED(ana_worker_info)))CALL cp__a("tmc/tmc_master.F",184)
      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("tmc/tmc_master.F",185)

      wg_loop: DO wg = 1, SIZE(ana_worker_info)
         IF (.NOT. ASSOCIATED(ana_list)) EXIT wg_loop
         IF (.NOT. ana_worker_info(wg)%busy) THEN
            stat = TMC_STAT_ANALYSIS_REQUEST
            dest = wg
            CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=dest, &
                             para_env=para_env, tmc_params=tmc_env%params, &
                             list_elem=ana_list)
            IF (.NOT. ASSOCIATED(ana_list%next)) THEN
               DEALLOCATE (ana_list)
               ana_list => NULL()
            ELSE
               list_tmp => ana_list%next
               DEALLOCATE (ana_list)
               ana_list => list_tmp
            END IF
         END IF
      END DO wg_loop
   END SUBROUTINE send_analysis_tasks

! **************************************************************************************************
!> \brief global master handling tree creation and communication/work
!>        distribution with workers
!> \param tmc_env structure for storing all the tmc parameters
!> \param globenv global environment for external control
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE do_tmc_master(tmc_env, globenv)
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_tmc_master', routineP = moduleN//':'//routineN

      INTEGER :: cancel_count, handle, last_output, reactivation_cc_count, &
         reactivation_ener_count, restart_count, restarted_elem_nr, stat, walltime_delay, &
         walltime_offset, wg, worker_counter
      INTEGER(KIND=int_8)                                :: mem
      INTEGER, DIMENSION(6)                              :: nr_of_job
      INTEGER, DIMENSION(:), POINTER                     :: tree_elem_counters, tree_elem_heads
      LOGICAL                                            :: external_stop, flag, l_update_tree
      REAL(KIND=dp)                                      :: run_time_start
      REAL(KIND=dp), DIMENSION(4)                        :: worker_timings_aver
      REAL(KIND=dp), DIMENSION(:), POINTER               :: efficiency
      TYPE(elem_array_type), DIMENSION(:), POINTER       :: ana_worker_info, worker_info
      TYPE(global_tree_type), POINTER                    :: gt_elem_tmp
      TYPE(tree_type), POINTER                           :: init_conf

      external_stop = .FALSE.
      restarted_elem_nr = 0
      NULLIFY (init_conf, worker_info, ana_worker_info, gt_elem_tmp, tree_elem_counters)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_master.F",238)

      IF(.NOT.(tmc_env%tmc_comp_set%group_nr == 0))CALL cp__a("tmc/tmc_master.F",240)
      IF(.NOT.(ASSOCIATED(tmc_env%tmc_comp_set)))CALL cp__a("tmc/tmc_master.F",241)
      IF(.NOT.(ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_w)))CALL cp__a("tmc/tmc_master.F",242)

      IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_master.F",244)

      !-- run time measurment, to end just in time
      ! start the timing
      CALL timeset(routineN, handle)
      run_time_start = m_walltime()
      walltime_delay = 0
      walltime_offset = 20 ! default value the whole program needs to finalize

      ! initialize the different modules
      IF (tmc_env%params%DRAW_TREE) &
         CALL init_draw_trees(tmc_params=tmc_env%params)

      !-- initialize variables
      ! nr_of_job: counting the differnt task send / received
      !  (1:NMC submitted, 2:energies submitted, 3:NMC finished 4:energy finished, 5:NMC canceled, 6:energy canceled)
      nr_of_job(:) = 0
      worker_counter = -1
      reactivation_ener_count = 0
      reactivation_cc_count = 0
      cancel_count = 0
      tmc_env%m_env%result_count = 0
      l_update_tree = .FALSE.
      restart_count = 1
      last_output = -1
      ! average timings
      !  (1:calculated NMC, 2:calculated ener, 3:canceled NMC, 4: canceled ener)
      worker_timings_aver(:) = 0.0_dp
      ! remembers state of workers and their actual configurations
      ! the actual working group, communicating with
      ALLOCATE (worker_info(tmc_env%tmc_comp_set%para_env_m_w%num_pe-1))
      ALLOCATE (ana_worker_info(tmc_env%tmc_comp_set%para_env_m_ana%num_pe-1))

      ! get the start configuration form the first (exact energy) worker,
      !   master should/could have no Force environment
      stat = TMC_STAT_START_CONF_REQUEST
      wg = 1
      CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                       para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                       tmc_params=tmc_env%params, &
                       wait_for_message=.TRUE.)
      !-- wait for start configuration results and number of dimentions
      !-- get start configuration (init_conf element should not be allocated already)
      CALL tmc_message(msg_type=stat, send_recv=recv_msg, dest=wg, &
                       para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                       tmc_params=tmc_env%params, &
                       elem=init_conf, success=flag, wait_for_message=.TRUE.)
      IF (stat .NE. TMC_STAT_START_CONF_RESULT) &
         CALL cp_abort(cp__l("tmc/tmc_master.F",292), &
                       "receiving start configuration failed, received stat "// &
                       cp_to_string(stat))
      ! get the atom names from first energy worker
      CALL communicate_atom_types(atoms=tmc_env%params%atoms, &
                                  source=1, &
                                  para_env=tmc_env%tmc_comp_set%para_env_m_first_w)

      CALL init_cell(cell=tmc_env%params%cell)

      ! check the configuration consitency with selected moves
      CALL check_moves(tmc_params=tmc_env%params, &
                       move_types=tmc_env%params%move_types, &
                       mol_array=init_conf%mol)
      IF (ASSOCIATED(tmc_env%params%nmc_move_types)) &
         CALL check_moves(tmc_params=tmc_env%params, &
                          move_types=tmc_env%params%nmc_move_types, &
                          mol_array=init_conf%mol)

      ! set initial configuration
      ! set initial random number generator seed (rng seed)
      ! initialize the tree structure espacially for parallel tmepering,
      !   seting the subtrees
      CALL init_tree_mod(start_elem=init_conf, tmc_env=tmc_env, &
                         job_counts=nr_of_job, &
                         worker_timings=worker_timings_aver)

      ! init restart counter (espacially for restart case)
      IF (tmc_env%m_env%restart_out_step .NE. 0) THEN
         restart_count = INT(tmc_env%m_env%result_count(0)/ &
                             REAL(tmc_env%m_env%restart_out_step, KIND=dp))+1
      END IF
      restarted_elem_nr = tmc_env%m_env%result_count(0)

!TODO check conf and cell of both input files (cell has to be equal,
!           beacuse it is used as reference cell for scaling the cell)
      ! communicate the reference cell size
      DO wg = 1, tmc_env%tmc_comp_set%para_env_m_w%num_pe-1
         stat = TMC_STATUS_WORKER_INIT
         CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                          para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                          tmc_params=tmc_env%params)
      END DO

      ! send the atom informations to all analysis workers
      IF (tmc_env%tmc_comp_set%para_env_m_ana%num_pe .GT. 1) THEN
         DO wg = 1, tmc_env%tmc_comp_set%para_env_m_ana%num_pe-1
            stat = TMC_STAT_INIT_ANALYSIS
            CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                             para_env=tmc_env%tmc_comp_set%para_env_m_ana, &
                             result_count=tmc_env%m_env%result_count, &
                             tmc_params=tmc_env%params, &
                             elem=init_conf, &
                             wait_for_message=.TRUE.)
         END DO
         CALL communicate_atom_types(atoms=tmc_env%params%atoms, &
                                     source=0, &
                                     para_env=tmc_env%tmc_comp_set%para_env_m_ana)
      END IF

      CALL deallocate_sub_tree_node(tree_elem=init_conf)

      ! regtest output
      IF (tmc_env%params%print_test_output .OR. DEBUG .GT. 0) &
         WRITE (tmc_env%m_env%io_unit, *) "TMC|first_global_tree_rnd_nr_X= ", &
         tmc_env%m_env%gt_head%rnd_nr

      ! calculate the approx energy of the first element (later the exact)
      IF (tmc_env%m_env%gt_head%conf(1)%elem%stat .EQ. status_calc_approx_ener) THEN
         wg = 1
         IF (tmc_env%tmc_comp_set%group_cc_nr .GT. 0) &
            wg = tmc_env%tmc_comp_set%group_ener_nr+1
         stat = TMC_STAT_APPROX_ENERGY_REQUEST
         CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                          para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                          tmc_params=tmc_env%params, &
                          elem=tmc_env%m_env%gt_head%conf(1)%elem)
         worker_info(wg)%busy = .TRUE.
         worker_info(wg)%elem => tmc_env%m_env%gt_head%conf(1)%elem
         init_conf => tmc_env%m_env%gt_head%conf(1)%elem
      ELSE IF (tmc_env%m_env%gt_head%conf(1)%elem%stat .EQ. status_created) THEN
         init_conf => tmc_env%m_env%gt_head%conf(1)%elem
         ! calculation will be done automatically,
         !   by searching the next conf for energy calculation
      END IF
      !-- START WORK --!
      !-- distributing work:
      !   1. receive incoming results
      !   2. check new results in tree
      !   3. if idle worker, create new tree element and send them to worker
      task_loop: DO
         ! =======================================================================
         !-- RECIEVING ALL incoming messages and handling them
         ! results of tree node 1 is distributed to all other subtree nodes
         ! =======================================================================
         worker_request_loop: DO
            wg = 1
            flag = .FALSE.
            CALL tmc_message(msg_type=stat, send_recv=recv_msg, dest=wg, &
                             para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                             tmc_params=tmc_env%params, &
                             elem_array=worker_info(:), success=flag)

            IF (flag .EQV. .FALSE.) EXIT worker_request_loop
            ! messages from worker group could be faster then the canceling request
            IF (worker_info(wg)%canceled .AND. (stat .NE. TMC_CANCELING_RECEIPT)) THEN
               IF (DEBUG .GE. 1) &
                  WRITE (tmc_env%m_env%io_unit, *) &
                  "TMC|master: recv stat "//cp_to_string(stat)// &
                  " of canceled worker group"
               CYCLE worker_request_loop
            END IF

            ! in case of parallel tempering canceled element could be reactivated,
            !   calculated faster and deleted
            IF (.NOT. ASSOCIATED(worker_info(wg)%elem)) &
               CALL cp_abort(cp__l("tmc/tmc_master.F",408), &
                             "no tree elem exist when receiving stat "// &
                             cp_to_string(stat)//"of group"//cp_to_string(wg))

            IF (DEBUG .GE. 1) &
               WRITE (tmc_env%m_env%io_unit, *) &
               "TMC|master: received stat "//cp_to_string(stat)// &
               " of sub tree "//cp_to_string(worker_info(wg)%elem%sub_tree_nr)// &
               " elem"//cp_to_string(worker_info(wg)%elem%nr)// &
               " with stat"//cp_to_string(worker_info(wg)%elem%stat)// &
               " of group"//cp_to_string(wg)//" group canceled ", worker_info(wg)%canceled
            SELECT CASE (stat)
               ! -- FAILED --------------------------
            CASE (TMC_STATUS_FAILED)
               EXIT task_loop
               ! -- CANCEL_RECEIPT ------------------
            CASE (TMC_CANCELING_RECEIPT)
               ! worker should got cancel message before
               IF(.NOT.(worker_info(wg)%canceled))CALL cp__a("tmc/tmc_master.F",426)
               worker_info(wg)%canceled = .FALSE.
               worker_info(wg)%busy = .FALSE.

               IF (ASSOCIATED(worker_info(wg)%elem)) THEN
                  SELECT CASE (worker_info (wg)%elem%stat)
                  CASE (status_cancel_ener)
                     !-- timings
                     worker_timings_aver(4) = (worker_timings_aver(4)*nr_of_job(6)+ &
                                               (m_walltime()-worker_info(wg)%start_time))/REAL(nr_of_job(6)+1, KIND=dp)
                     nr_of_job(6) = nr_of_job(6)+1

                     worker_info(wg)%elem%stat = status_canceled_ener
                     worker_info(wg)%elem%potential = 8000.0_dp
                     IF (tmc_env%params%DRAW_TREE) THEN
                        CALL create_dot_color(tree_element=worker_info(wg)%elem, &
                                              tmc_params=tmc_env%params)
                     END IF
                  CASE (status_cancel_nmc)
                     !-- timings
                     worker_timings_aver(3) = (worker_timings_aver(3)*nr_of_job(5)+ &
                                               (m_walltime()-worker_info(wg)%start_time))/REAL(nr_of_job(5)+1, KIND=dp)
                     nr_of_job(5) = nr_of_job(5)+1

                     worker_info(wg)%elem%stat = status_canceled_nmc
                     worker_info(wg)%elem%potential = 8000.0_dp
                     IF (tmc_env%params%DRAW_TREE) THEN
                        CALL create_dot_color(tree_element=worker_info(wg)%elem, &
                                              tmc_params=tmc_env%params)
                     END IF
                  CASE DEFAULT
                     ! the subtree element is again in use (reactivated)
                  END SELECT
                  worker_info(wg)%elem => NULL()
               END IF
               ! -- START_CONF_RESULT ---------------
            CASE (TMC_STAT_START_CONF_RESULT)
               ! start configuration should already be handeled
               CALL cp__b("tmc/tmc_master.F",464,"")
               ! -- ENERGY RESULT -----------------
            CASE (TMC_STAT_APPROX_ENERGY_RESULT)
               nr_of_job(3) = nr_of_job(3)+1
               worker_info(wg)%busy = .FALSE.
               worker_info(wg)%elem%stat = status_created
               IF (tmc_env%params%DRAW_TREE) THEN
                  CALL create_dot_color(tree_element=worker_info(wg)%elem, &
                                        tmc_params=tmc_env%params)
               END IF
               worker_info(wg)%elem => NULL()
               ! nothing to do, the approximate potential
               !   should be updated in the message interface
               ! -- NMC / MD RESULT -----------------
            CASE (TMC_STAT_NMC_RESULT, TMC_STAT_MD_RESULT)
               IF (.NOT. worker_info(wg)%canceled) worker_info(wg)%busy = .FALSE.
               !-- timings for Nested Monte Carlo calculation
               worker_timings_aver(1) = (worker_timings_aver(1)*nr_of_job(3)+ &
                                         (m_walltime()-worker_info(wg)%start_time))/REAL(nr_of_job(3)+1, KIND=dp)
               nr_of_job(3) = nr_of_job(3)+1

               worker_info(wg)%start_time = m_walltime()-worker_info(wg)%start_time
               CALL set_walltime_delay(worker_info(wg)%start_time, walltime_delay)
               worker_info(wg)%elem%stat = status_created
               IF (tmc_env%params%DRAW_TREE) THEN
                  CALL create_dot_color(tree_element=worker_info(wg)%elem, &
                                        tmc_params=tmc_env%params)
               END IF
               !-- send energy request
               ! in case of one singe input file, energy is already calculated
               IF (tmc_env%params%NMC_inp_file .EQ. "") THEN
                  worker_info(wg)%elem%potential = worker_info(wg)%elem%e_pot_approx
                  worker_info(wg)%elem%stat = status_calculated
                  ! check acceptance of depending nodes
                  IF (.NOT. (ASSOCIATED(worker_info(wg)%elem, init_conf))) THEN
                     CALL check_acceptance_of_depending_subtree_nodes(tree_elem=worker_info(wg)%elem, &
                                                                      tmc_env=tmc_env)
                  END IF
                  IF (tmc_env%params%DRAW_TREE) THEN
                     CALL create_dot_color(tree_element=worker_info(wg)%elem, &
                                           tmc_params=tmc_env%params)
                  END IF
                  !-- CANCELING the calculations of the elements, which are definetively not needed anymore
                  CALL cancel_calculations(cancel_list=tmc_env%m_env%cancelation_list, &
                                           work_list=worker_info, &
                                           para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                           tmc_env=tmc_env, &
                                           cancel_count=cancel_count)
                  worker_info(wg)%elem => NULL()
               ELSE
                  ! if all working groups are equal, the same group calculates the energy
                  IF (tmc_env%tmc_comp_set%group_cc_nr .LE. 0 &
                      .AND. (.NOT. worker_info(wg)%canceled)) THEN
                     worker_info(wg)%elem%stat = status_calculate_energy
                     stat = TMC_STAT_ENERGY_REQUEST
                     ! immediately send energy request
                     CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                                      para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                      tmc_params=tmc_env%params, &
                                      elem=worker_info(wg)%elem)
                     worker_info(wg)%busy = .TRUE.
                     nr_of_job(2) = nr_of_job(2)+1
                     IF (tmc_env%params%DRAW_TREE) THEN
                        CALL create_dot_color(tree_element=worker_info(wg)%elem, &
                                              tmc_params=tmc_env%params)
                     END IF
                     !-- set start time for energy calculation
                     worker_info(wg)%start_time = m_walltime()
                  ELSE
                     worker_info(wg)%elem => NULL()
                  END IF
               END IF
               ! -- ENERGY RESULT --------------------
            CASE (TMC_STAT_ENERGY_RESULT)
               !-- timings
               worker_timings_aver(2) = (worker_timings_aver(2)*nr_of_job(4)+ &
                                         (m_walltime()-worker_info(wg)%start_time))/REAL(nr_of_job(4)+1, KIND=dp)
               nr_of_job(4) = nr_of_job(4)+1

               worker_info(wg)%start_time = m_walltime()-worker_info(wg)%start_time
               CALL set_walltime_delay(worker_info(wg)%start_time, walltime_delay)

               IF (.NOT. worker_info(wg)%canceled) &
                  worker_info(wg)%busy = .FALSE.
               ! the first node in tree is always accepted.!.
               IF (ASSOCIATED(worker_info(wg)%elem, init_conf)) THEN
                  !-- distribute energy of first element to all subtrees
                  CALL finalize_init(gt_tree_ptr=tmc_env%m_env%gt_head, &
                                     tmc_env=tmc_env)
                  IF (tmc_env%params%DRAW_TREE) THEN
                     CALL create_global_tree_dot_color(gt_tree_element=tmc_env%m_env%gt_act, &
                                                       tmc_params=tmc_env%params)
                     CALL create_dot_color(tree_element=worker_info(wg)%elem, &
                                           tmc_params=tmc_env%params)
                  END IF
                  init_conf => NULL()
               ELSE
                  worker_info(wg)%elem%stat = status_calculated
                  IF (tmc_env%params%DRAW_TREE) &
                     CALL create_dot_color(worker_info(wg)%elem, &
                                           tmc_params=tmc_env%params)
                  ! check acceptance of depending nodes
                  ! first (initial) configuration do not have to be checked
                  CALL check_acceptance_of_depending_subtree_nodes(tree_elem=worker_info(wg)%elem, &
                                                                   tmc_env=tmc_env)
               END IF
               !-- write out all configurations (not only Markov Chain) e.g. for fitting
               IF (tmc_env%params%all_conf_file_name .NE. "") THEN
                  CALL write_element_in_file(elem=worker_info(wg)%elem, &
                                             file_name=tmc_env%params%all_conf_file_name, &
                                             tmc_params=tmc_env%params, &
                                             conf_nr=nr_of_job(4))
               END IF

               !-- CANCELING the calculations of the elements,
               !      which are definetively not needed anymore
               CALL cancel_calculations(cancel_list=tmc_env%m_env%cancelation_list, &
                                        work_list=worker_info, &
                                        para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                        tmc_env=tmc_env, &
                                        cancel_count=cancel_count)
               IF (DEBUG .GE. 9) &
                  WRITE (tmc_env%m_env%io_unit, *) &
                  "TMC|master: handled energy result of sub tree ", &
                  worker_info(wg)%elem%sub_tree_nr, " elem ", worker_info(wg)%elem%nr, &
                  " with stat", worker_info(wg)%elem%stat
               worker_info(wg)%elem => NULL()

               !-- SCF ENERGY -----------------------
            CASE (TMC_STAT_SCF_STEP_ENER_RECEIVE)
               IF (.NOT. (ASSOCIATED(worker_info(wg)%elem, init_conf)) .AND. &
                   worker_info(wg)%elem%stat .NE. status_cancel_ener .AND. &
                   worker_info(wg)%elem%stat .NE. status_cancel_nmc) THEN
                  ! update the acceptance probability and the canceling list
                  CALL check_elements_for_acc_prob_update(tree_elem=worker_info(wg)%elem, &
                                                          tmc_env=tmc_env)
               END IF
               ! cancel inlikely elements
               CALL cancel_calculations(cancel_list=tmc_env%m_env%cancelation_list, &
                                        work_list=worker_info, &
                                        para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                        tmc_env=tmc_env, &
                                        cancel_count=cancel_count)
            CASE (TMC_STAT_ANALYSIS_RESULT)
               ana_worker_info(wg)%busy = .FALSE.
               ana_worker_info(wg)%elem => NULL()
            CASE DEFAULT
               CALL cp__b("tmc/tmc_master.F",611,"received message with unknown info/stat type")
            END SELECT
         END DO worker_request_loop
         !-- do tree update (check new results)
         CALL tree_update(tmc_env=tmc_env, result_acc=flag, &
                          something_updated=l_update_tree)
         IF (DEBUG .GE. 2 .AND. l_update_tree) &
            WRITE (tmc_env%m_env%io_unit, *) &
            "TMC|master: tree updated "//cp_to_string(l_update_tree)// &
            " of with gt elem "//cp_to_string(tmc_env%m_env%gt_act%nr)// &
            " with stat"//cp_to_string(tmc_env%m_env%gt_act%stat)

         CALL send_analysis_tasks(ana_list=tmc_env%m_env%analysis_list, &
                                  ana_worker_info=ana_worker_info, &
                                  para_env=tmc_env%tmc_comp_set%para_env_m_ana, &
                                  tmc_env=tmc_env)

         ! =======================================================================
         !-- ALL CALCULATIONS DONE (check) ---
         ! =======================================================================
         ! if enough configurations are sampled or walltime is exeeded,
         !   finish building trees
!TODO set correct logger para_env to use this
         CALL external_control(should_stop=external_stop, flag="TMC", globenv=globenv)
         IF ((ANY(tmc_env%m_env%result_count(1:) .GE. tmc_env%m_env%num_MC_elem) &
              .AND. flag) .OR. &
             (m_walltime()-run_time_start .GT. &
              tmc_env%m_env%walltime-walltime_delay-walltime_offset) .OR. &
             external_stop) THEN
            WRITE (tmc_env%m_env%io_unit, FMT="(/,T2,A)") REPEAT("=", 79)
            ! calculations NOT finished, walltime exceeded
            IF (.NOT. ANY(tmc_env%m_env%result_count(1:) &
                          .GE. tmc_env%m_env%num_MC_elem)) THEN
               WRITE (tmc_env%m_env%io_unit, *) "Walltime exceeded.", &
                  m_walltime()-run_time_start, " of ", tmc_env%m_env%walltime-walltime_delay-walltime_offset, &
                  "(incl. delay", walltime_delay, "and offset", walltime_offset, ") left"
            ELSE
               ! calculations finished
               IF (tmc_env%params%print_test_output) &
                  WRITE (tmc_env%m_env%io_unit, *) "Total energy: ", &
                  tmc_env%m_env%result_list(1)%elem%potential
            END IF
            IF (tmc_env%m_env%restart_out_step .NE. 0) &
               CALL print_restart_file(tmc_env=tmc_env, job_counts=nr_of_job, &
                                       timings=worker_timings_aver)
            EXIT task_loop
         END IF

         ! =======================================================================
         ! update the rest of the tree (canceling and deleting elements)
         ! =======================================================================
         IF (l_update_tree) THEN
            IF (DEBUG .GE. 2) &
               WRITE (tmc_env%m_env%io_unit, *) &
               "TMC|master: start remove elem and cancel calculation"
            !-- CLEANING tree nodes beside the path through the tree from
            !      end_of_clean_tree to tree_ptr
            ! --> getting back the end of clean tree
            CALL remove_all_trees(working_elem_list=worker_info, tmc_env=tmc_env)
            !-- CANCELING the calculations of the elements,
            !     which are definetively not needed anymore
            !   elements are added to canceling list if no global tree reference
            !     exist anymore
            CALL cancel_calculations(cancel_list=tmc_env%m_env%cancelation_list, &
                                     work_list=worker_info, &
                                     cancel_count=cancel_count, &
                                     para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                     tmc_env=tmc_env)
         END IF

         ! =====================================================================
         !-- NEW TASK (if worker not busy sumit next task)
         ! =====================================================================
         worker_counter = worker_counter+1
         wg = MODULO(worker_counter, tmc_env%tmc_comp_set%para_env_m_w%num_pe-1)+1

         IF (DEBUG .GE. 16 .AND. ALL(worker_info(:)%busy)) &
            WRITE (tmc_env%m_env%io_unit, *) "all workers are busy"

         IF (.NOT. worker_info(wg)%busy) THEN
            IF (DEBUG .GE. 13) &
               WRITE (tmc_env%m_env%io_unit, *) &
               "TMC|master: search new task for worker ", wg
            ! no group separation
            IF (tmc_env%tmc_comp_set%group_cc_nr .LE. 0) THEN
               ! search next element to calculate the energy
               CALL search_next_energy_calc(gt_head=tmc_env%m_env%gt_act, &
                                            new_gt_elem=gt_elem_tmp, stat=stat, &
                                            react_count=reactivation_ener_count)
               IF (stat .EQ. TMC_STATUS_WAIT_FOR_NEW_TASK) THEN
                  CALL create_new_gt_tree_node(tmc_env=tmc_env, stat=stat, &
                                               new_elem=gt_elem_tmp, &
                                               reactivation_cc_count=reactivation_cc_count)
               END IF
            ELSEIF (wg .GT. tmc_env%tmc_comp_set%group_ener_nr) THEN
               ! specialized groups (groups for exact energy and groups for configurational change)
               ! creating new element (configurational change group)
               !-- crate new node, configurational change is handled in tmc_tree module
               CALL create_new_gt_tree_node(tmc_env=tmc_env, stat=stat, &
                                            new_elem=gt_elem_tmp, &
                                            reactivation_cc_count=reactivation_cc_count)
               ! element could be already created, hence CC worker has nothing to do for this element
               ! in next round he will get a task
               IF (stat .EQ. status_created .OR. stat .EQ. status_calculate_energy) &
                  stat = TMC_STATUS_WAIT_FOR_NEW_TASK
            ELSE
               ! search next element to calculate the energy
               CALL search_next_energy_calc(gt_head=tmc_env%m_env%gt_act, &
                                            new_gt_elem=gt_elem_tmp, stat=stat, &
                                            react_count=reactivation_ener_count)
            END IF

            IF (DEBUG .GE. 10) &
               WRITE (tmc_env%m_env%io_unit, *) &
               "TMC|master: send task with elem stat "//cp_to_string(stat)// &
               " to group "//cp_to_string(wg)
            ! MESSAGE settings: status informations and task for communication
            SELECT CASE (stat)
            CASE (TMC_STATUS_WAIT_FOR_NEW_TASK)
               CYCLE task_loop
            CASE (TMC_STATUS_FAILED)
               !STOP "in creating new task, status failed should be handled before"
               CYCLE task_loop
            CASE (status_calculated, status_accepted, status_rejected)
               CYCLE task_loop
            CASE (status_calc_approx_ener)
               ! e.g. after volume move, we need the approximate potential for 2 potential check of following NMC nodes
               stat = TMC_STAT_APPROX_ENERGY_REQUEST
               CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                                para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                tmc_params=tmc_env%params, &
                                elem=gt_elem_tmp%conf(gt_elem_tmp%mv_conf)%elem)
               nr_of_job(1) = nr_of_job(1)+1
            CASE (status_created, status_calculate_energy)
               ! in case of parallel tempering the node can be already be calculating (related to another global tree node
               !-- send task to calculate system property
               gt_elem_tmp%conf(gt_elem_tmp%mv_conf)%elem%stat = status_calculate_energy
               IF (tmc_env%params%DRAW_TREE) &
                  CALL create_dot_color(tree_element=gt_elem_tmp%conf(gt_elem_tmp%mv_conf)%elem, &
                                        tmc_params=tmc_env%params)
               stat = TMC_STAT_ENERGY_REQUEST
               CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                                para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                tmc_params=tmc_env%params, &
                                elem=gt_elem_tmp%conf(gt_elem_tmp%mv_conf)%elem)
               nr_of_job(2) = nr_of_job(2)+1
            CASE (status_calculate_MD)
               stat = TMC_STAT_MD_REQUEST
               CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                                para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                tmc_params=tmc_env%params, &
                                elem=gt_elem_tmp%conf(gt_elem_tmp%mv_conf)%elem)
!                           temperature=tmc_env%params%Temp(gt_elem_tmp%mv_conf), &
               nr_of_job(1) = nr_of_job(1)+1
            CASE (status_calculate_NMC_steps)
               !-- send information of element, which should be calculated
               stat = TMC_STAT_NMC_REQUEST
               CALL tmc_message(msg_type=stat, send_recv=send_msg, dest=wg, &
                                para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                                tmc_params=tmc_env%params, &
                                elem=gt_elem_tmp%conf(gt_elem_tmp%mv_conf)%elem)
               nr_of_job(1) = nr_of_job(1)+1
            CASE (status_cancel_nmc, status_cancel_ener)
               ! skip that task until receipt is received
               ! no status update
            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_master.F",777), &
                             "new task of tree element"// &
                             cp_to_string(gt_elem_tmp%nr)// &
                             "has unknown status"//cp_to_string(stat))
            END SELECT
            worker_info(wg)%elem => gt_elem_tmp%conf(gt_elem_tmp%mv_conf)%elem
            worker_info(wg)%busy = .TRUE.
            ! set timer for maximum calculation time recognition
            worker_info(wg)%start_time = m_walltime()

            !===================== write out info after x requested tasks==========
            IF (nr_of_job(4) .GT. last_output .AND. &
                (MODULO(nr_of_job(4), tmc_env%m_env%info_out_step_size) .EQ. 0) .AND. &
                (stat .NE. TMC_STATUS_FAILED)) THEN
               last_output = nr_of_job(4)
               WRITE (tmc_env%m_env%io_unit, FMT="(/,T2,A)") REPEAT("-", 79)
               WRITE (tmc_env%m_env%io_unit, *) &
                  "Tasks submitted:  E ", nr_of_job(2), ", cc", nr_of_job(1)
               WRITE (tmc_env%m_env%io_unit, *) &
                  "Results received: E ", nr_of_job(4), ", cc", nr_of_job(3)
               WRITE (tmc_env%m_env%io_unit, *) &
                  "Configurations used:", tmc_env%m_env%result_count(0), &
                  ", sub trees", tmc_env%m_env%result_count(1:)

               CALL print_move_types(init=.FALSE., file_io=tmc_env%m_env%io_unit, &
                                     tmc_params=tmc_env%params)
               ALLOCATE (tree_elem_counters(0:SIZE(tmc_env%params%Temp)))
               ALLOCATE (tree_elem_heads(0:SIZE(tmc_env%params%Temp)))
               CALL count_nodes_in_trees(global_tree_ptr=tmc_env%m_env%gt_act, &
                                         end_of_clean_trees=tmc_env%m_env%st_clean_ends, &
                                         counters=tree_elem_counters, head_elements_nr=tree_elem_heads)
               WRITE (tmc_env%m_env%io_unit, *) "nodes in trees", tree_elem_counters(:)
               WRITE (tmc_env%m_env%io_unit, *) "tree heads    ", tree_elem_heads(:)
               IF (tmc_env%params%NMC_inp_file .NE. "") THEN
                  CALL count_prepared_nodes_in_trees(global_tree_ptr=tmc_env%m_env%gt_act, &
                                                     counters=tree_elem_counters)
                  WRITE (tmc_env%m_env%io_unit, FMT=*) &
                     "ener prepared ", tree_elem_counters
               END IF
               IF (tmc_env%params%SPECULATIVE_CANCELING) &
                  WRITE (tmc_env%m_env%io_unit, *) &
                  "canceled cc|E:     ", nr_of_job(5:6), &
                  ", reactivated: cc ", &
                  reactivation_cc_count, &
                  ", reactivated: E ", &
                  reactivation_ener_count
               WRITE (tmc_env%m_env%io_unit, FMT='(A,2F10.2)') &
                  " Average time for cc/ener calc          ", &
                  worker_timings_aver(1), worker_timings_aver(2)
               IF (tmc_env%params%SPECULATIVE_CANCELING) &
                  WRITE (tmc_env%m_env%io_unit, FMT='(A,2F10.2)') &
                  " Average time until cancel cc/ener calc ", &
                  worker_timings_aver(3), worker_timings_aver(4)
               IF (tmc_env%params%esimate_acc_prob) &
                  WRITE (tmc_env%m_env%io_unit, *) &
                  "Estimate correct (acc/Nacc) | wrong (acc/nacc)", &
                  tmc_env%m_env%estim_corr_wrong(1), &
                  tmc_env%m_env%estim_corr_wrong(3), " | ", &
                  tmc_env%m_env%estim_corr_wrong(2), &
                  tmc_env%m_env%estim_corr_wrong(4)
               WRITE (tmc_env%m_env%io_unit, *) &
                  "Time: ", INT(m_walltime()-run_time_start), "of", &
                  INT(tmc_env%m_env%walltime-walltime_delay-walltime_offset), &
                  "sec needed. "
               CALL m_memory(mem)
               WRITE (tmc_env%m_env%io_unit, *) &
                  "Memory used: ", INT(mem/(1024*1024), KIND=KIND(0)), "MiBytes"
               CALL m_flush(tmc_env%m_env%io_unit)
               DEALLOCATE (tree_elem_heads)
               DEALLOCATE (tree_elem_counters)
            END IF
            !===================== write out restart file after x results============
            IF (tmc_env%m_env%restart_out_step .GT. 0 .AND. &
                tmc_env%m_env%result_count(0) .GT. &
                restart_count*tmc_env%m_env%restart_out_step) THEN
               CALL print_restart_file(tmc_env=tmc_env, job_counts=nr_of_job, &
                                       timings=worker_timings_aver)
               restart_count = restart_count+1
            END IF

         END IF !worker busy?
      END DO task_loop

      ! -- END OF WORK (enough configurations are calculated or walltime exceeded
      WRITE (tmc_env%m_env%io_unit, FMT="(/,T2,A)") REPEAT("=", 79)
      WRITE (UNIT=tmc_env%m_env%io_unit, FMT="(T2,A,T35,A,T80,A)") "=", &
         "finalizing TMC", "="
      WRITE (tmc_env%m_env%io_unit, *) "acceptance rates:"
      CALL print_move_types(init=.FALSE., file_io=tmc_env%m_env%io_unit, &
                            tmc_params=tmc_env%params)
      WRITE (tmc_env%m_env%io_unit, FMT="(/,T2,A)") REPEAT("-", 79)
      ! program efficency result outputs
      ALLOCATE (efficiency(0:tmc_env%params%nr_temp))
      CALL get_subtree_efficiency(tmc_env=tmc_env, eff=efficiency)
      WRITE (tmc_env%m_env%io_unit, *) "Efficiencies:"
      WRITE (tmc_env%m_env%io_unit, FMT="(A,F5.2,A,1000F5.2)") &
         " (MC elements/calculated configuration) global:", &
         efficiency(0), " sub tree(s): ", efficiency(1:)
      DEALLOCATE (efficiency)
      IF (tmc_env%tmc_comp_set%group_cc_nr .GT. 0) &
         WRITE (tmc_env%m_env%io_unit, FMT="(A,1000F5.2)") &
         " (MC elements/created configuration)          :", &
         tmc_env%m_env%result_count(:)/REAL(nr_of_job(3), KIND=dp)
      WRITE (tmc_env%m_env%io_unit, FMT="(A,1000F5.2)") &
         " (MC elements/energy calculated configuration):", &
         tmc_env%m_env%result_count(:)/REAL(nr_of_job(4), KIND=dp)
      IF (tmc_env%params%NMC_inp_file .NE. "") THEN
         WRITE (tmc_env%m_env%io_unit, *) &
            "Amount of canceled elements (E/cc):", &
            tmc_env%m_env%count_cancel_ener, tmc_env%m_env%count_cancel_NMC
         WRITE (tmc_env%m_env%io_unit, *) &
            " reactivated E  ", reactivation_ener_count
         WRITE (tmc_env%m_env%io_unit, *) &
            " reactivated cc ", reactivation_cc_count
      END IF
      WRITE (tmc_env%m_env%io_unit, FMT="(A,F10.2)") &
         " computing time of one Markov chain element ", &
         (m_walltime()-run_time_start)/REAL(tmc_env%m_env%result_count(0)- &
                                            restarted_elem_nr, KIND=dp)
      WRITE (tmc_env%m_env%io_unit, FMT="(A,F10.2)") " TMC run time[s]: ", m_walltime()-run_time_start
      WRITE (tmc_env%m_env%io_unit, FMT="(/,T2,A)") REPEAT("=", 79)

      !-- FINALIZE
      WRITE (tmc_env%m_env%io_unit, *) "stopping workers"
      CALL stop_whole_group(para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                            worker_info=worker_info, &
                            tmc_params=tmc_env%params)
      DEALLOCATE (worker_info)
      CALL stop_whole_group(para_env=tmc_env%tmc_comp_set%para_env_m_ana, &
                            worker_info=ana_worker_info, &
                            tmc_params=tmc_env%params)
      DEALLOCATE (ana_worker_info)

      !-- deallocating everything in tree module
      CALL finalize_trees(tmc_env=tmc_env)

      CALL free_cancelation_list(tmc_env%m_env%cancelation_list)

      ! -- write final configuration
      IF (tmc_env%params%DRAW_TREE) &
         CALL finalize_draw_tree(tmc_params=tmc_env%params)

      WRITE (tmc_env%m_env%io_unit, *) "TMC master: all work done."

      ! end the timing
      CALL timestop(handle)

   END SUBROUTINE do_tmc_master

! **************************************************************************************************
!> \brief routine sets the walltime delay, to the maximum calculation time
!>        hence the program can stop with a proper finailze
!> \param time actual calculation time
!> \param walltime_delay the actual biggest calculation time
!> \author Mandes 12.2012
! **************************************************************************************************
   SUBROUTINE set_walltime_delay(time, walltime_delay)
      REAL(KIND=dp)                                      :: time
      INTEGER                                            :: walltime_delay

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_walltime_delay', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(time .GE. 0.0_dp))CALL cp__a("tmc/tmc_master.F",940)

      IF (time .GT. walltime_delay) THEN
         walltime_delay = INT(time)+1
      END IF
   END SUBROUTINE set_walltime_delay

END MODULE tmc_master
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module analyses element of the TMC tree element structure
!>        e.g. density, radial distribution function, dipole correlation,...
!> \par History
!>      02.2013 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_analysis
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              pbc
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_to_string
   USE force_fields_input,              ONLY: read_chrg_section
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE mathlib,                         ONLY: diag
   USE physcon,                         ONLY: a_mass,&
                                              au2a => angstrom,&
                                              boltzmann,&
                                              joule,&
                                              massunit
   USE tmc_analysis_types,              ONLY: &
        ana_type_default, ana_type_ice, ana_type_sym_xyz, atom_pairs_type, dipole_moment_type, &
        pair_correl_type, search_pair_in_list, tmc_ana_density_create, tmc_ana_density_file_name, &
        tmc_ana_dipole_analysis_create, tmc_ana_dipole_moment_create, tmc_ana_displacement_create, &
        tmc_ana_env_create, tmc_ana_pair_correl_create, tmc_ana_pair_correl_file_name, &
        tmc_analysis_env
   USE tmc_calculations,                ONLY: get_scaled_cell,&
                                              nearest_distance
   USE tmc_file_io,                     ONLY: analyse_files_close,&
                                              analyse_files_open,&
                                              expand_file_name_char,&
                                              expand_file_name_temp,&
                                              read_element_from_file,&
                                              write_dipoles_in_file
   USE tmc_stati,                       ONLY: TMC_STATUS_OK,&
                                              TMC_STATUS_WAIT_FOR_NEW_TASK,&
                                              tmc_default_restart_in_file_name,&
                                              tmc_default_restart_out_file_name,&
                                              tmc_default_trajectory_file_name,&
                                              tmc_default_unspecified_name
   USE tmc_tree_build,                  ONLY: allocate_new_sub_tree_node,&
                                              deallocate_sub_tree_node
   USE tmc_tree_types,                  ONLY: read_subtree_elem_unformated,&
                                              tree_type,&
                                              write_subtree_elem_unformated
   USE tmc_types,                       ONLY: tmc_atom_type,&
                                              tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_analysis'

   PUBLIC :: tmc_read_ana_input
   PUBLIC :: analysis_init, do_tmc_analysis, analyze_file_configurations, finalize_tmc_analysis
   PUBLIC :: analysis_restart_print, analysis_restart_read

CONTAINS

! **************************************************************************************************
!> \brief creates a new para environment for tmc analysis
!> \param tmc_ana_section ...
!> \param tmc_ana TMC analysis environment
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE tmc_read_ana_input(tmc_ana_section, tmc_ana)
      TYPE(section_vals_type), POINTER                   :: tmc_ana_section
      TYPE(tmc_analysis_env), POINTER                    :: tmc_ana

      CHARACTER(len=*), PARAMETER :: routineN = 'tmc_read_ana_input', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: c_tmp
      CHARACTER(LEN=default_string_length), POINTER      :: charge_atm(:)
      INTEGER                                            :: i_tmp, ntot
      INTEGER, DIMENSION(3)                              :: nr_bins
      INTEGER, DIMENSION(:), POINTER                     :: i_arr_tmp
      LOGICAL                                            :: explicit, explicit_key, flag
      REAL(KIND=dp), POINTER                             :: charge(:)
      TYPE(section_vals_type), POINTER                   :: tmp_section

      NULLIFY (tmp_section, charge_atm, i_arr_tmp, charge)

      IF(.NOT.(ASSOCIATED(tmc_ana_section)))CALL cp__a("tmc/tmc_analysis.F",101)
      IF(.NOT.(.NOT. ASSOCIATED(tmc_ana)))CALL cp__a("tmc/tmc_analysis.F",102)

      CALL section_vals_get(tmc_ana_section, explicit=explicit)
      IF (explicit) THEN
         CALL tmc_ana_env_create(tmc_ana=tmc_ana)
         ! restarting
         CALL section_vals_val_get(tmc_ana_section, "RESTART", l_val=tmc_ana%restart)
         ! file name prefix
         CALL section_vals_val_get(tmc_ana_section, "PREFIX_ANA_FILES", &
                                   c_val=tmc_ana%out_file_prefix)
         IF (tmc_ana%out_file_prefix .NE. "") THEN
            tmc_ana%out_file_prefix = TRIM(tmc_ana%out_file_prefix)//"_"
         END IF

         ! density calculation
         CALL section_vals_val_get(tmc_ana_section, "DENSITY", explicit=explicit_key)
         IF (explicit_key) THEN
            CALL section_vals_val_get(tmc_ana_section, "DENSITY", i_vals=i_arr_tmp)

            IF (SIZE(i_arr_tmp(:)) .EQ. 3) THEN
               IF (ANY(i_arr_tmp(:) .LE. 0)) &
                  CALL cp_abort(cp__l("tmc/tmc_analysis.F",123), "The amount of intervals in each "// &
                                "direction has to be greater than 0.")
               nr_bins(:) = i_arr_tmp(:)
            ELSE IF (SIZE(i_arr_tmp(:)) .EQ. 1) THEN
               IF (ANY(i_arr_tmp(:) .LE. 0)) &
                  CALL cp__b("tmc/tmc_analysis.F",128,"The amount of intervals has to be greater than 0.")
               nr_bins(:) = i_arr_tmp(1)
            ELSE IF (SIZE(i_arr_tmp(:)) .EQ. 0) THEN
               nr_bins(:) = 1
            ELSE
               CALL cp__b("tmc/tmc_analysis.F",133,"unknown amount of dimentions for the binning.")
            END IF
            CALL tmc_ana_density_create(tmc_ana%density_3d, nr_bins)
         END IF

         ! radial distribution function calculation
         CALL section_vals_val_get(tmc_ana_section, "G_R", explicit=explicit_key)
         IF (explicit_key) THEN
            CALL section_vals_val_get(tmc_ana_section, "G_R", i_val=i_tmp)
            CALL tmc_ana_pair_correl_create(ana_pair_correl=tmc_ana%pair_correl, &
                                            nr_bins=i_tmp)
         END IF

         ! radial distribution function calculation
         CALL section_vals_val_get(tmc_ana_section, "CLASSICAL_DIPOLE_MOMENTS", explicit=explicit_key)
         IF (explicit_key) THEN
            ! charges for dipoles needed
            tmp_section => section_vals_get_subs_vals(tmc_ana_section, "CHARGE")
            CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=i_tmp)
            IF (explicit) THEN
               ntot = 0
               ALLOCATE (charge_atm(i_tmp))
               ALLOCATE (charge(i_tmp))
               CALL read_chrg_section(charge_atm, charge, tmp_section, ntot)
            ELSE
               CALL cp_abort(cp__l("tmc/tmc_analysis.F",158), &
                             "to calculate the classical cell dipole moment "// &
                             "the charges has to be specified")
            END IF

            CALL tmc_ana_dipole_moment_create(tmc_ana%dip_mom, charge_atm, charge, &
                                              tmc_ana%dim_per_elem)

            IF (ASSOCIATED(charge_atm)) DEALLOCATE (charge_atm)
            IF (ASSOCIATED(charge)) DEALLOCATE (charge)
         END IF

         ! dipole moment analysis
         CALL section_vals_val_get(tmc_ana_section, "DIPOLE_ANALYSIS", explicit=explicit_key)
         IF (explicit_key) THEN
            CALL tmc_ana_dipole_analysis_create(tmc_ana%dip_ana)
            CALL section_vals_val_get(tmc_ana_section, "DIPOLE_ANALYSIS", c_val=c_tmp)
            SELECT CASE (TRIM (c_tmp))
            CASE (TRIM (tmc_default_unspecified_name))
               tmc_ana%dip_ana%ana_type = ana_type_default
            CASE ("ICE")
               tmc_ana%dip_ana%ana_type = ana_type_ice
            CASE ("SYM_XYZ")
               tmc_ana%dip_ana%ana_type = ana_type_sym_xyz
            CASE DEFAULT
               CALL cp__w("tmc/tmc_analysis.F",183,'unknown analysis type "'//TRIM(c_tmp)//'" specified. Set to default.')
               tmc_ana%dip_ana%ana_type = ana_type_default
            END SELECT
         END IF

      END IF

      ! cell displacement (deviation)
      CALL section_vals_val_get(tmc_ana_section, "DEVIATION", l_val=flag)
      IF (flag) THEN
         CALL tmc_ana_displacement_create(ana_disp=tmc_ana%displace, &
                                          dim_per_elem=tmc_ana%dim_per_elem)
      END IF
   END SUBROUTINE tmc_read_ana_input

! **************************************************************************************************
!> \brief initialize all the necessarry analysis structures
!> \param ana_env ...
!> \param nr_dim dimension of the pos, frc etc. array
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE analysis_init(ana_env, nr_dim)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env
      INTEGER                                            :: nr_dim

      CHARACTER(len=*), PARAMETER :: routineN = 'analysis_init', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: tmp_cell_file, tmp_dip_file, tmp_pos_file

      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",212)
      IF(.NOT.(nr_dim > 0))CALL cp__a("tmc/tmc_analysis.F",213)

      ana_env%nr_dim = nr_dim

      ! save file names
      tmp_pos_file = ana_env%costum_pos_file_name
      tmp_cell_file = ana_env%costum_cell_file_name
      tmp_dip_file = ana_env%costum_dip_file_name

      ! unset all filenames
      ana_env%costum_pos_file_name = tmc_default_unspecified_name
      ana_env%costum_cell_file_name = tmc_default_unspecified_name
      ana_env%costum_dip_file_name = tmc_default_unspecified_name

      ! set the neccessary files for ...
      ! density
      IF (ASSOCIATED(ana_env%density_3d)) THEN
         ana_env%costum_pos_file_name = tmp_pos_file
         ana_env%costum_cell_file_name = tmp_cell_file
      END IF
      ! pair correlation
      IF (ASSOCIATED(ana_env%pair_correl)) THEN
         ana_env%costum_pos_file_name = tmp_pos_file
         ana_env%costum_cell_file_name = tmp_cell_file
      END IF
      ! dipole moment
      IF (ASSOCIATED(ana_env%dip_mom)) THEN
         ana_env%costum_pos_file_name = tmp_pos_file
         ana_env%costum_cell_file_name = tmp_cell_file
      END IF
      ! dipole analysis
      IF (ASSOCIATED(ana_env%dip_ana)) THEN
         ana_env%costum_pos_file_name = tmp_pos_file
         ana_env%costum_cell_file_name = tmp_cell_file
         ana_env%costum_dip_file_name = tmp_dip_file
      END IF
      ! deviation / displacement
      IF (ASSOCIATED(ana_env%displace)) THEN
         ana_env%costum_pos_file_name = tmp_pos_file
         ana_env%costum_cell_file_name = tmp_cell_file
      END IF

      ! init radial distribution function
      IF (ASSOCIATED(ana_env%pair_correl)) &
         CALL ana_pair_correl_init(ana_pair_correl=ana_env%pair_correl, &
                                   atoms=ana_env%atoms, cell=ana_env%cell)
      ! init classical dipole moment calculations
      IF (ASSOCIATED(ana_env%dip_mom)) &
         CALL ana_dipole_moment_init(ana_dip_mom=ana_env%dip_mom, &
                                     atoms=ana_env%atoms)
   END SUBROUTINE analysis_init

! **************************************************************************************************
!> \brief print analysis restart file
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE analysis_restart_print(ana_env)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(len=*), PARAMETER :: routineN = 'analysis_restart_print', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, file_name_tmp, &
                                                            restart_file_name
      INTEGER                                            :: file_ptr
      LOGICAL                                            :: l_tmp

      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",282)
      IF(.NOT.(ASSOCIATED(ana_env%last_elem)))CALL cp__a("tmc/tmc_analysis.F",283)
      IF (.NOT. ana_env%restart) RETURN

      WRITE (file_name, FMT='(I9.9)') ana_env%last_elem%nr
      file_name_tmp = TRIM(expand_file_name_temp(expand_file_name_char( &
                                                 TRIM(ana_env%out_file_prefix)// &
                                                 tmc_default_restart_out_file_name, &
                                                 "ana"), ana_env%temperature))
      restart_file_name = expand_file_name_char(file_name_tmp, &
                                                file_name)
      CALL open_file(file_name=restart_file_name, file_status="REPLACE", &
                     file_action="WRITE", file_form="UNFORMATTED", &
                     unit_number=file_ptr)
      WRITE (file_ptr) ana_env%temperature
      CALL write_subtree_elem_unformated(ana_env%last_elem, file_ptr)

      ! first mention the different kind of anlysis types initialized
      ! then the variables for each calculation type
      l_tmp = ASSOCIATED(ana_env%density_3d)
      WRITE (file_ptr) l_tmp
      IF (l_tmp) THEN
         WRITE (file_ptr) ana_env%density_3d%conf_counter, &
            ana_env%density_3d%nr_bins, &
            ana_env%density_3d%sum_vol, &
            ana_env%density_3d%sum_vol2, &
            ana_env%density_3d%sum_box_length, &
            ana_env%density_3d%sum_box_length2, &
            ana_env%density_3d%sum_density, &
            ana_env%density_3d%sum_dens2
      END IF

      l_tmp = ASSOCIATED(ana_env%pair_correl)
      WRITE (file_ptr) l_tmp
      IF (l_tmp) THEN
         WRITE (file_ptr) ana_env%pair_correl%conf_counter, &
            ana_env%pair_correl%nr_bins, &
            ana_env%pair_correl%step_lenght, &
            ana_env%pair_correl%pairs, &
            ana_env%pair_correl%g_r
      END IF

      l_tmp = ASSOCIATED(ana_env%dip_mom)
      WRITE (file_ptr) l_tmp
      IF (l_tmp) THEN
         WRITE (file_ptr) ana_env%dip_mom%conf_counter, &
            ana_env%dip_mom%charges, &
            ana_env%dip_mom%last_dip_cl
      END IF

      l_tmp = ASSOCIATED(ana_env%dip_ana)
      WRITE (file_ptr) l_tmp
      IF (l_tmp) THEN
         WRITE (file_ptr) ana_env%dip_ana%conf_counter, &
            ana_env%dip_ana%ana_type, &
            ana_env%dip_ana%mu2_pv_s, &
            ana_env%dip_ana%mu_psv, &
            ana_env%dip_ana%mu_pv, &
            ana_env%dip_ana%mu2_pv_mat, &
            ana_env%dip_ana%mu2_pv_mat
      END IF

      l_tmp = ASSOCIATED(ana_env%displace)
      WRITE (file_ptr) l_tmp
      IF (l_tmp) THEN
         WRITE (file_ptr) ana_env%displace%conf_counter, &
            ana_env%displace%disp
      END IF

      CALL close_file(unit_number=file_ptr)

      file_name_tmp = expand_file_name_char(TRIM(ana_env%out_file_prefix)// &
                                            tmc_default_restart_in_file_name, "ana")
      file_name = expand_file_name_temp(file_name_tmp, &
                                        ana_env%temperature)
      CALL open_file(file_name=file_name, &
                     file_action="WRITE", file_status="REPLACE", &
                     unit_number=file_ptr)
      WRITE (file_ptr, *) TRIM(restart_file_name)
      CALL close_file(unit_number=file_ptr)
   END SUBROUTINE analysis_restart_print

! **************************************************************************************************
!> \brief read analysis restart file
!> \param ana_env ...
!> \param elem ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE analysis_restart_read(ana_env, elem)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env
      TYPE(tree_type), POINTER                           :: elem

      CHARACTER(len=*), PARAMETER :: routineN = 'analysis_restart_read', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, file_name_tmp
      INTEGER                                            :: file_ptr
      LOGICAL                                            :: l_tmp
      REAL(KIND=dp)                                      :: temp

      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",383)
      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_analysis.F",384)
      IF (.NOT. ana_env%restart) RETURN

      file_name_tmp = expand_file_name_char(TRIM(ana_env%out_file_prefix)// &
                                            tmc_default_restart_in_file_name, "ana")
      file_name = expand_file_name_temp(file_name_tmp, &
                                        ana_env%temperature)
      INQUIRE (FILE=file_name, EXIST=l_tmp)
      IF (l_tmp) THEN
         CALL open_file(file_name=file_name, file_status="OLD", &
                        file_action="READ", unit_number=file_ptr)
         READ (file_ptr, *) file_name_tmp
         CALL close_file(unit_number=file_ptr)

         CALL open_file(file_name=file_name_tmp, file_status="OLD", file_form="UNFORMATTED", &
                        file_action="READ", unit_number=file_ptr)
         READ (file_ptr) temp
         IF(.NOT.(ana_env%temperature .EQ. temp))CALL cp__a("tmc/tmc_analysis.F",401)
         ana_env%last_elem => elem
         CALL read_subtree_elem_unformated(elem, file_ptr)

         ! first mention the different kind of anlysis types initialized
         ! then the variables for each calculation type
         READ (file_ptr) l_tmp
         IF(.NOT.(ASSOCIATED(ana_env%density_3d) .EQV. l_tmp))CALL cp__a("tmc/tmc_analysis.F",408)
         IF (l_tmp) THEN
            READ (file_ptr) ana_env%density_3d%conf_counter, &
               ana_env%density_3d%nr_bins, &
               ana_env%density_3d%sum_vol, &
               ana_env%density_3d%sum_vol2, &
               ana_env%density_3d%sum_box_length, &
               ana_env%density_3d%sum_box_length2, &
               ana_env%density_3d%sum_density, &
               ana_env%density_3d%sum_dens2
         END IF

         READ (file_ptr) l_tmp
         IF(.NOT.(ASSOCIATED(ana_env%pair_correl) .EQV. l_tmp))CALL cp__a("tmc/tmc_analysis.F",421)
         IF (l_tmp) THEN
            READ (file_ptr) ana_env%pair_correl%conf_counter, &
               ana_env%pair_correl%nr_bins, &
               ana_env%pair_correl%step_lenght, &
               ana_env%pair_correl%pairs, &
               ana_env%pair_correl%g_r
         END IF

         READ (file_ptr) l_tmp
         IF(.NOT.(ASSOCIATED(ana_env%dip_mom) .EQV. l_tmp))CALL cp__a("tmc/tmc_analysis.F",431)
         IF (l_tmp) THEN
            READ (file_ptr) ana_env%dip_mom%conf_counter, &
               ana_env%dip_mom%charges, &
               ana_env%dip_mom%last_dip_cl
         END IF

         READ (file_ptr) l_tmp
         IF(.NOT.(ASSOCIATED(ana_env%dip_ana) .EQV. l_tmp))CALL cp__a("tmc/tmc_analysis.F",439)
         IF (l_tmp) THEN
            READ (file_ptr) ana_env%dip_ana%conf_counter, &
               ana_env%dip_ana%ana_type, &
               ana_env%dip_ana%mu2_pv_s, &
               ana_env%dip_ana%mu_psv, &
               ana_env%dip_ana%mu_pv, &
               ana_env%dip_ana%mu2_pv_mat, &
               ana_env%dip_ana%mu2_pv_mat
         END IF

         READ (file_ptr) l_tmp
         IF(.NOT.(ASSOCIATED(ana_env%displace) .EQV. l_tmp))CALL cp__a("tmc/tmc_analysis.F",451)
         IF (l_tmp) THEN
            READ (file_ptr) ana_env%displace%conf_counter, &
               ana_env%displace%disp
         END IF

         CALL close_file(unit_number=file_ptr)
         elem => NULL()
      END IF
   END SUBROUTINE analysis_restart_read

! **************************************************************************************************
!> \brief call all the necessarry analysis routines
!>         analysis the previous element with the weight of the different
!>        configuration numbers
!>        and stores the actual in the structur % last_elem
!>        afterwards the previous configuration can be deallocated (outside)
!> \param elem ...
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE do_tmc_analysis(elem, ana_env)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_tmc_analysis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, weight_act
      REAL(KIND=dp), DIMENSION(3)                        :: dip_tmp
      TYPE(tree_type), POINTER                           :: elem_tmp

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_analysis.F",484)
      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",485)

      ! start the timing
      CALL timeset(routineN, handle)
      IF (ASSOCIATED(ana_env%last_elem) .AND. &
          (ana_env%last_elem%nr .LT. elem%nr)) THEN
         weight_act = elem%nr-ana_env%last_elem%nr
         ! calculates the 3 dimensional distributed density
         IF (ASSOCIATED(ana_env%density_3d)) &
            CALL calc_density_3d(elem=ana_env%last_elem, &
                                 weight=weight_act, atoms=ana_env%atoms, &
                                 ana_env=ana_env)
         ! calculated the radial distribution function for each atom type
         IF (ASSOCIATED(ana_env%pair_correl)) &
            CALL calc_paircorrelation(elem=ana_env%last_elem, weight=weight_act, &
                                      atoms=ana_env%atoms, ana_env=ana_env)
         ! calculates the classical dipole moments
         IF (ASSOCIATED(ana_env%dip_mom)) &
            CALL calc_dipole_moment(elem=ana_env%last_elem, weight=weight_act, &
                                    ana_env=ana_env)
         ! calculates the dipole moments analysis and dielectric constant
         IF (ASSOCIATED(ana_env%dip_ana)) THEN
            ! in symmetric case use also the dipoles
            !   (-x,y,z) .. .. (-x,-y,z).... (-x,-y-z) all have the same energy
            IF (ana_env%dip_ana%ana_type .EQ. ana_type_sym_xyz) THEN
               ! (-x,y,z)
               ana_env%last_elem%dipole(1) = -ana_env%last_elem%dipole(1)
               dip_tmp(:) = ana_env%last_elem%dipole(:)
               IF (ASSOCIATED(ana_env%dip_mom)) &
                  ana_env%dip_mom%last_dip_cl(1) = -ana_env%dip_mom%last_dip_cl(1)
               CALL calc_dipole_analysis(elem=ana_env%last_elem, weight=weight_act, &
                                         ana_env=ana_env)
               ! (-x,-y,z)
               ana_env%last_elem%dipole(:) = dip_tmp(:)
               ana_env%last_elem%dipole(2) = -ana_env%last_elem%dipole(2)
               dip_tmp(:) = ana_env%last_elem%dipole(:)
               IF (ASSOCIATED(ana_env%dip_mom)) &
                  ana_env%dip_mom%last_dip_cl(2) = -ana_env%dip_mom%last_dip_cl(2)
               CALL calc_dipole_analysis(elem=ana_env%last_elem, weight=weight_act, &
                                         ana_env=ana_env)
               ! (-x,-y,-z)
               ana_env%last_elem%dipole(:) = dip_tmp(:)
               ana_env%last_elem%dipole(3) = -ana_env%last_elem%dipole(3)
               dip_tmp(:) = ana_env%last_elem%dipole(:)
               IF (ASSOCIATED(ana_env%dip_mom)) &
                  ana_env%dip_mom%last_dip_cl(3) = -ana_env%dip_mom%last_dip_cl(3)
               CALL calc_dipole_analysis(elem=ana_env%last_elem, weight=weight_act, &
                                         ana_env=ana_env)
               ! (x,-y,-z)
               ana_env%last_elem%dipole(:) = dip_tmp(:)
               ana_env%last_elem%dipole(1) = -ana_env%last_elem%dipole(1)
               dip_tmp(:) = ana_env%last_elem%dipole(:)
               IF (ASSOCIATED(ana_env%dip_mom)) &
                  ana_env%dip_mom%last_dip_cl(1) = -ana_env%dip_mom%last_dip_cl(1)
               CALL calc_dipole_analysis(elem=ana_env%last_elem, weight=weight_act, &
                                         ana_env=ana_env)
               ! (x,y,-z)
               ana_env%last_elem%dipole(:) = dip_tmp(:)
               ana_env%last_elem%dipole(2) = -ana_env%last_elem%dipole(2)
               dip_tmp(:) = ana_env%last_elem%dipole(:)
               IF (ASSOCIATED(ana_env%dip_mom)) &
                  ana_env%dip_mom%last_dip_cl(2) = -ana_env%dip_mom%last_dip_cl(2)
               CALL calc_dipole_analysis(elem=ana_env%last_elem, weight=weight_act, &
                                         ana_env=ana_env)
               ! (-x,y,-z)
               ana_env%last_elem%dipole(:) = dip_tmp(:)
               ana_env%last_elem%dipole(1) = -ana_env%last_elem%dipole(1)
               dip_tmp(:) = ana_env%last_elem%dipole(:)
               IF (ASSOCIATED(ana_env%dip_mom)) &
                  ana_env%dip_mom%last_dip_cl(1) = -ana_env%dip_mom%last_dip_cl(1)
               CALL calc_dipole_analysis(elem=ana_env%last_elem, weight=weight_act, &
                                         ana_env=ana_env)
               ! (x,-y,z)
               ana_env%last_elem%dipole(:) = dip_tmp(:)
               ana_env%last_elem%dipole(:) = -ana_env%last_elem%dipole(:)
               dip_tmp(:) = ana_env%last_elem%dipole(:)
               IF (ASSOCIATED(ana_env%dip_mom)) &
                  ana_env%dip_mom%last_dip_cl(:) = -ana_env%dip_mom%last_dip_cl(:)
               CALL calc_dipole_analysis(elem=ana_env%last_elem, weight=weight_act, &
                                         ana_env=ana_env)
               ! back to (x,y,z)
               ana_env%last_elem%dipole(:) = dip_tmp(:)
               ana_env%last_elem%dipole(2) = -ana_env%last_elem%dipole(2)
               dip_tmp(:) = ana_env%last_elem%dipole(:)
               IF (ASSOCIATED(ana_env%dip_mom)) &
                  ana_env%dip_mom%last_dip_cl(2) = -ana_env%dip_mom%last_dip_cl(2)
            END IF
            CALL calc_dipole_analysis(elem=ana_env%last_elem, weight=weight_act, &
                                      ana_env=ana_env)
            CALL print_act_dipole_analysis(elem=ana_env%last_elem, &
                                           ana_env=ana_env)
         END IF

         ! calculates the cell displacement from last cell
         IF (ASSOCIATED(ana_env%displace) .AND. weight_act .GT. 0) &
            CALL calc_displacement(elem=elem, ana_env=ana_env)
      END IF
      ! swap elem with last elem, to delete original last element and store the actual one
      elem_tmp => ana_env%last_elem
      ana_env%last_elem => elem
      elem => elem_tmp
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE do_tmc_analysis

! **************************************************************************************************
!> \brief call all the necessarry analysis printing routines
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE finalize_tmc_analysis(ana_env)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'finalize_tmc_analysis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",604)

      ! start the timing
      CALL timeset(routineN, handle)
      IF (ASSOCIATED(ana_env%density_3d)) THEN
         IF (ana_env%density_3d%conf_counter .GT. 0) &
            CALL print_density_3d(ana_env=ana_env)
      END IF
      IF (ASSOCIATED(ana_env%pair_correl)) THEN
         IF (ana_env%pair_correl%conf_counter .GT. 0) &
            CALL print_paircorrelation(ana_env=ana_env)
      END IF
      IF (ASSOCIATED(ana_env%dip_mom)) THEN
         IF (ana_env%dip_mom%conf_counter .GT. 0) &
            CALL print_dipole_moment(ana_env)
      END IF
      IF (ASSOCIATED(ana_env%dip_ana)) THEN
         IF (ana_env%dip_ana%conf_counter .GT. 0) &
            CALL print_dipole_analysis(ana_env)
      END IF
      IF (ASSOCIATED(ana_env%displace)) THEN
         IF (ana_env%displace%conf_counter .GT. 0) &
            CALL print_average_displacement(ana_env)
      END IF

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE finalize_tmc_analysis

! **************************************************************************************************
!> \brief read the files and analyze the configurations
!> \param start_id ...
!> \param end_id ...
!> \param dir_ind ...
!> \param ana_env ...
!> \param tmc_params ...
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE analyze_file_configurations(start_id, end_id, dir_ind, &
                                          ana_env, tmc_params)
      INTEGER                                            :: start_id, end_id
      INTEGER, OPTIONAL                                  :: dir_ind
      TYPE(tmc_analysis_env), POINTER                    :: ana_env
      TYPE(tmc_param_type), POINTER                      :: tmc_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'analyze_file_configurations', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: conf_nr, handle, nr_dim, stat
      TYPE(tree_type), POINTER                           :: elem

      NULLIFY (elem)
      conf_nr = -1
      stat = TMC_STATUS_WAIT_FOR_NEW_TASK
      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",658)
      IF(.NOT.(ASSOCIATED(tmc_params)))CALL cp__a("tmc/tmc_analysis.F",659)

      ! start the timing
      CALL timeset(routineN, handle)

      ! open the files
      CALL analyse_files_open(tmc_ana=ana_env, stat=stat, dir_ind=dir_ind)
      ! set the existence of exact dipoles (from file)
      IF (ana_env%id_dip .GT. 0) THEN
         tmc_params%print_dipole = .TRUE.
      ELSE
         tmc_params%print_dipole = .FALSE.
      END IF

      ! allocate the actual element structure
      CALL allocate_new_sub_tree_node(tmc_params=tmc_params, next_el=elem, &
                                      nr_dim=ana_env%nr_dim)

      IF (ASSOCIATED(ana_env%last_elem)) conf_nr = ana_env%last_elem%nr
      nr_dim = SIZE(elem%pos)

      IF (stat .EQ. TMC_STATUS_OK) THEN
         conf_loop: DO
            CALL read_element_from_file(elem=elem, tmc_ana=ana_env, conf_nr=conf_nr, &
                                        stat=stat)
            IF (stat .EQ. TMC_STATUS_WAIT_FOR_NEW_TASK) THEN
               CALL deallocate_sub_tree_node(tree_elem=elem)
               EXIT conf_loop
            END IF
            ! if we want just a certain part of the trajectory
            IF (start_id .LT. 0 .OR. conf_nr .GE. start_id) THEN
               IF (end_id .LT. 0 .OR. conf_nr .LE. end_id) THEN
                  ! do the analysis calculations
                  CALL do_tmc_analysis(elem=elem, ana_env=ana_env)
               END IF
            END IF

            ! clean temporary element (already analyzed)
            IF (ASSOCIATED(elem)) THEN
               CALL deallocate_sub_tree_node(tree_elem=elem)
            END IF
            ! if there was no previous element, create a new temp element to write in
            IF (.NOT. ASSOCIATED(elem)) &
               CALL allocate_new_sub_tree_node(tmc_params=tmc_params, next_el=elem, &
                                               nr_dim=nr_dim)
         END DO conf_loop
      END IF
      ! close the files
      CALL analyse_files_close(tmc_ana=ana_env)

      IF (ASSOCIATED(elem)) &
         CALL deallocate_sub_tree_node(tree_elem=elem)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE analyze_file_configurations

   !============================================================================
   ! density calculations
   !============================================================================

! **************************************************************************************************
!> \brief calculates the density in rectantangulares
!>        defined by the number of bins in each direction
!> \param elem ...
!> \param weight ...
!> \param atoms ...
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE calc_density_3d(elem, weight, atoms, ana_env)
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: weight
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_density_3d', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, file_name_tmp
      INTEGER                                            :: atom, bin_x, bin_y, bin_z, file_ptr, &
                                                            handle
      LOGICAL                                            :: flag
      REAL(KIND=dp)                                      :: mass_total, r_tmp, vol_cell, vol_sub_box
      REAL(KIND=dp), DIMENSION(3)                        :: atom_pos, cell_size, interval_size
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: mass_bin

      NULLIFY (mass_bin)

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_analysis.F",749)
      IF(.NOT.(ASSOCIATED(elem%pos)))CALL cp__a("tmc/tmc_analysis.F",750)
      IF(.NOT.(weight .GT. 0))CALL cp__a("tmc/tmc_analysis.F",751)
      IF(.NOT.(ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_analysis.F",752)
      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",753)
      IF(.NOT.(ASSOCIATED(ana_env%cell)))CALL cp__a("tmc/tmc_analysis.F",754)
      IF(.NOT.(ASSOCIATED(ana_env%density_3d)))CALL cp__a("tmc/tmc_analysis.F",755)
      IF(.NOT.(ASSOCIATED(ana_env%density_3d%sum_density)))CALL cp__a("tmc/tmc_analysis.F",756)
      IF(.NOT.(ASSOCIATED(ana_env%density_3d%sum_dens2)))CALL cp__a("tmc/tmc_analysis.F",757)

      ! start the timing
      CALL timeset(routineN, handle)

      atom_pos(:) = 0.0_dp
      cell_size(:) = 0.0_dp
      interval_size(:) = 0.0_dp
      mass_total = 0.0_dp

      bin_x = SIZE(ana_env%density_3d%sum_density(:, 1, 1))
      bin_y = SIZE(ana_env%density_3d%sum_density(1, :, 1))
      bin_z = SIZE(ana_env%density_3d%sum_density(1, 1, :))
      ALLOCATE (mass_bin(bin_x, bin_y, bin_z))
      mass_bin(:, :, :) = 0.0_dp

      ! if NPT -> box_scale.NE.1.0 use the scaled cell
      ! ATTENTION then the sub box middle points are not correct in the output
      !  espacially if we use multiple sub boxes
      CALL get_scaled_cell(cell=ana_env%cell, box_scale=elem%box_scale, &
                           abc=cell_size, vol=vol_cell)
      ! volume summed over configurations for average volume [A]
      ana_env%density_3d%sum_vol = ana_env%density_3d%sum_vol+ &
                                   vol_cell*(au2a)**3*weight
      ana_env%density_3d%sum_vol2 = ana_env%density_3d%sum_vol2+ &
                                    (vol_cell*(au2a)**3)**2*weight

      ana_env%density_3d%sum_box_length(:) = ana_env%density_3d%sum_box_length(:) &
                                             +cell_size(:)*(au2a)*weight
      ana_env%density_3d%sum_box_length2(:) = ana_env%density_3d%sum_box_length2(:) &
                                              +(cell_size(:)*(au2a))**2*weight

      ! sub interval length
      interval_size(1) = cell_size(1)/REAL(bin_x, dp)
      interval_size(2) = cell_size(2)/REAL(bin_y, dp)
      interval_size(3) = cell_size(3)/REAL(bin_z, dp)

      ! volume in [cm^3]
      vol_cell = vol_cell*(au2a*1E-8)**3
      vol_sub_box = interval_size(1)*interval_size(2)*interval_size(3)* &
                    (au2a*1E-8)**3

      ! count every atom
      DO atom = 1, SIZE(elem%pos), ana_env%dim_per_elem

         atom_pos(:) = elem%pos(atom:atom+2)
         ! fold into box
         CALL get_scaled_cell(cell=ana_env%cell, box_scale=elem%box_scale, &
                              vec=atom_pos)
         ! shifts the box to positive values (before 0,0,0 is the center)
         atom_pos(:) = atom_pos(:)+0.5_dp*cell_size(:)
         ! calculate the index of the sub box
         bin_x = INT(atom_pos(1)/interval_size(1))+1
         bin_y = INT(atom_pos(2)/interval_size(2))+1
         bin_z = INT(atom_pos(3)/interval_size(3))+1
         IF(.NOT.(bin_x .GT. 0 .AND. bin_y .GT. 0 .AND. bin_z .GT. 0))CALL cp__a("tmc/tmc_analysis.F",812)
         IF(.NOT.(bin_x .LE. SIZE(ana_env%density_3d%sum_density(:, 1, 1))))CALL cp__a("tmc/tmc_analysis.F",813)
         IF(.NOT.(bin_y .LE. SIZE(ana_env%density_3d%sum_density(1, :, 1))))CALL cp__a("tmc/tmc_analysis.F",814)
         IF(.NOT.(bin_z .LE. SIZE(ana_env%density_3d%sum_density(1, 1, :))))CALL cp__a("tmc/tmc_analysis.F",815)

         ! sum mass in [g] (in bins and total)
         mass_bin(bin_x, bin_y, bin_z) = mass_bin(bin_x, bin_y, bin_z)+ &
                                         atoms(INT(atom/REAL(ana_env%dim_per_elem, KIND=dp))+1)%mass/massunit*1000*a_mass
         mass_total = mass_total+ &
                      atoms(INT(atom/REAL(ana_env%dim_per_elem, KIND=dp))+1)%mass/massunit*1000*a_mass
         !mass_bin(bin_x,bin_y,bin_z) = mass_bin(bin_x,bin_y,bin_z) + &
         !  atoms(INT(atom/REAL(ana_env%dim_per_elem,KIND=dp))+1)%mass/&
         !     massunit/n_avogadro
         !mass_total = mass_total + &
         !  atoms(INT(atom/REAL(ana_env%dim_per_elem,KIND=dp))+1)%mass/&
         !     massunit/n_avogadro
      END DO
      ! check total cell density
      r_tmp = mass_total/vol_cell-SUM(mass_bin(:, :, :))/vol_sub_box/SIZE(mass_bin(:, :, :))
      IF(.NOT.(ABS(r_tmp) .LT. 1E-5))CALL cp__a("tmc/tmc_analysis.F",831)

      ! calculate density (mass per volume) and sum up for average value
      ana_env%density_3d%sum_density(:, :, :) = ana_env%density_3d%sum_density(:, :, :)+ &
                                                weight*mass_bin(:, :, :)/vol_sub_box

      ! calculate density squared ( (mass per volume)^2 ) for variance and sum up for average value
      ana_env%density_3d%sum_dens2(:, :, :) = ana_env%density_3d%sum_dens2(:, :, :)+ &
                                              weight*(mass_bin(:, :, :)/vol_sub_box)**2

      ana_env%density_3d%conf_counter = ana_env%density_3d%conf_counter+weight

      ! print out the actual and average density in file
      IF (ana_env%density_3d%print_dens) THEN
         file_name_tmp = expand_file_name_temp(TRIM(ana_env%out_file_prefix)// &
                                               tmc_default_trajectory_file_name, &
                                               ana_env%temperature)
         file_name = TRIM(expand_file_name_char(file_name_tmp, &
                                                "dens"))
         INQUIRE (FILE=file_name, EXIST=flag)
         CALL open_file(file_name=file_name, file_status="UNKNOWN", &
                        file_action="WRITE", file_position="APPEND", &
                        unit_number=file_ptr)
         IF (.NOT. flag) &
            WRITE (file_ptr, FMT='(A8,11A20)') "# conf_nr", "dens_act[g/cm^3]", &
            "dens_average[g/cm^3]", "density_variance", &
            "averages:volume", "box_lenth_x", "box_lenth_y", "box_lenth_z", &
            "variances:volume", "box_lenth_x", "box_lenth_y", "box_lenth_z"
         WRITE (file_ptr, FMT="(I8,11F20.10)") ana_env%density_3d%conf_counter+1-weight, &
            SUM(mass_bin(:, :, :))/vol_sub_box/SIZE(mass_bin(:, :, :)), &
            SUM(ana_env%density_3d%sum_density(:, :, :))/ &
            SIZE(ana_env%density_3d%sum_density(:, :, :))/ &
            REAL(ana_env%density_3d%conf_counter, KIND=dp), &
            SUM(ana_env%density_3d%sum_dens2(:, :, :))/ &
            SIZE(ana_env%density_3d%sum_dens2(:, :, :))/ &
            REAL(ana_env%density_3d%conf_counter, KIND=dp)- &
            (SUM(ana_env%density_3d%sum_density(:, :, :))/ &
             SIZE(ana_env%density_3d%sum_density(:, :, :))/ &
             REAL(ana_env%density_3d%conf_counter, KIND=dp))**2, &
            ana_env%density_3d%sum_vol/ &
            REAL(ana_env%density_3d%conf_counter, KIND=dp), &
            ana_env%density_3d%sum_box_length(:)/ &
            REAL(ana_env%density_3d%conf_counter, KIND=dp), &
            ana_env%density_3d%sum_vol2/ &
            REAL(ana_env%density_3d%conf_counter, KIND=dp)- &
            (ana_env%density_3d%sum_vol/ &
             REAL(ana_env%density_3d%conf_counter, KIND=dp))**2, &
            ana_env%density_3d%sum_box_length2(:)/ &
            REAL(ana_env%density_3d%conf_counter, KIND=dp)- &
            (ana_env%density_3d%sum_box_length(:)/ &
             REAL(ana_env%density_3d%conf_counter, KIND=dp))**2
         CALL close_file(unit_number=file_ptr)
      END IF

      DEALLOCATE (mass_bin)
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE calc_density_3d

! **************************************************************************************************
!> \brief print the density in rectantangulares
!>        defined by the number of bins in each direction
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE print_density_3d(ana_env)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: fmt_my = '(T2,A,"| ",A,T41,A40)', plabel = "TMC_ANA", &
         routineN = 'print_density_3d', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, file_name_vari
      INTEGER                                            :: bin_x, bin_y, bin_z, file_ptr_dens, &
                                                            file_ptr_vari, handle, i, j, k
      REAL(KIND=dp), DIMENSION(3)                        :: cell_size, interval_size

      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",908)
      IF(.NOT.(ASSOCIATED(ana_env%density_3d)))CALL cp__a("tmc/tmc_analysis.F",909)
      IF(.NOT.(ASSOCIATED(ana_env%density_3d%sum_density)))CALL cp__a("tmc/tmc_analysis.F",910)
      IF(.NOT.(ASSOCIATED(ana_env%density_3d%sum_dens2)))CALL cp__a("tmc/tmc_analysis.F",911)

      ! start the timing
      CALL timeset(routineN, handle)

      file_name = ""
      file_name_vari = ""

      bin_x = SIZE(ana_env%density_3d%sum_density(:, 1, 1))
      bin_y = SIZE(ana_env%density_3d%sum_density(1, :, 1))
      bin_z = SIZE(ana_env%density_3d%sum_density(1, 1, :))
      CALL get_cell(cell=ana_env%cell, abc=cell_size)
      interval_size(1) = cell_size(1)/REAL(bin_x, KIND=dp)*au2a
      interval_size(2) = cell_size(2)/REAL(bin_y, KIND=dp)*au2a
      interval_size(3) = cell_size(3)/REAL(bin_z, KIND=dp)*au2a

      file_name = expand_file_name_temp(TRIM(ana_env%out_file_prefix)// &
                                        tmc_ana_density_file_name, &
                                        ana_env%temperature)
      CALL open_file(file_name=file_name, file_status="REPLACE", &
                     file_action="WRITE", file_position="APPEND", &
                     unit_number=file_ptr_dens)
      WRITE (file_ptr_dens, FMT='(A,1X,I0,1X,A,3(I0,1X),1X,A,1X,3F10.5)') &
         "# configurations", ana_env%density_3d%conf_counter, "bins", &
         ana_env%density_3d%nr_bins, "interval size", interval_size(:)
      WRITE (file_ptr_dens, FMT='(A,3A10,A20)') "#", " x [A] ", " y [A] ", " z [A] ", " density [g/cm^3] "

      file_name_vari = expand_file_name_temp(expand_file_name_char( &
                                             TRIM(ana_env%out_file_prefix)// &
                                             tmc_ana_density_file_name, "vari"), &
                                             ana_env%temperature)
      CALL open_file(file_name=file_name_vari, file_status="REPLACE", &
                     file_action="WRITE", file_position="APPEND", &
                     unit_number=file_ptr_vari)
      WRITE (file_ptr_vari, FMT='(A,1X,I0,1X,A,3(I0,1X),1X,A,1X,3F10.5)') &
         "# configurations", ana_env%density_3d%conf_counter, "bins", &
         ana_env%density_3d%nr_bins, "interval size", interval_size(:)
      WRITE (file_ptr_vari, FMT='(A,3A10,A20)') "#", " x [A] ", " y [A] ", " z [A] ", " variance"

      DO i = 1, SIZE(ana_env%density_3d%sum_density(:, 1, 1))
         DO j = 1, SIZE(ana_env%density_3d%sum_density(1, :, 1))
            DO k = 1, SIZE(ana_env%density_3d%sum_density(1, 1, :))
               WRITE (file_ptr_dens, FMT='(3F10.2,F20.10)') &
                  (i-0.5_dp)*interval_size(1), (j-0.5_dp)*interval_size(2), (k-0.5_dp)*interval_size(3), &
                  ana_env%density_3d%sum_density(i, j, k)/REAL(ana_env%density_3d%conf_counter, KIND=dp)
               WRITE (file_ptr_vari, FMT='(3F10.2,F20.10)') &
                  (i-0.5_dp)*interval_size(1), (j-0.5_dp)*interval_size(2), (k-0.5_dp)*interval_size(3), &
                  ana_env%density_3d%sum_dens2(i, j, k)/REAL(ana_env%density_3d%conf_counter, KIND=dp)- &
                  (ana_env%density_3d%sum_density(i, j, k)/REAL(ana_env%density_3d%conf_counter, KIND=dp))**2
            END DO
         END DO
      END DO
      CALL close_file(unit_number=file_ptr_dens)
      CALL close_file(unit_number=file_ptr_vari)

      WRITE (ana_env%io_unit, FMT="(/,T2,A)") REPEAT("-", 79)
      WRITE (ana_env%io_unit, FMT="(T2,A,T35,A,T80,A)") "-", "density calculation", "-"
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "temperature ", cp_to_string(ana_env%temperature)
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "used configurations", &
         cp_to_string(REAL(ana_env%density_3d%conf_counter, KIND=dp))
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "average volume", &
         cp_to_string(ana_env%density_3d%sum_vol/ &
                      REAL(ana_env%density_3d%conf_counter, KIND=dp))
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "average density in the cell: ", &
         cp_to_string(SUM(ana_env%density_3d%sum_density(:, :, :))/ &
                      SIZE(ana_env%density_3d%sum_density(:, :, :))/ &
                      REAL(ana_env%density_3d%conf_counter, KIND=dp))
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "density variance:", &
         cp_to_string(SUM(ana_env%density_3d%sum_dens2(:, :, :))/ &
                      SIZE(ana_env%density_3d%sum_dens2(:, :, :))/ &
                      REAL(ana_env%density_3d%conf_counter, KIND=dp)- &
                      (SUM(ana_env%density_3d%sum_density(:, :, :))/ &
                       SIZE(ana_env%density_3d%sum_density(:, :, :))/ &
                       REAL(ana_env%density_3d%conf_counter, KIND=dp))**2)
      WRITE (ana_env%io_unit, FMT="(/,T2,A)") REPEAT("-", 79)
      IF (ana_env%print_test_output) &
         WRITE (ana_env%io_unit, *) "TMC|ANALYSIS_CELL_DENSITY_X= ", &
         SUM(ana_env%density_3d%sum_density(:, :, :))/ &
         SIZE(ana_env%density_3d%sum_density(:, :, :))/ &
         REAL(ana_env%density_3d%conf_counter, KIND=dp)
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE print_density_3d

   !============================================================================
   ! radial distribution function
   !============================================================================

! **************************************************************************************************
!> \brief init radial distribution function structures
!> \param ana_pair_correl ...
!> \param atoms ...
!> \param cell ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE ana_pair_correl_init(ana_pair_correl, atoms, cell)
      TYPE(pair_correl_type), POINTER                    :: ana_pair_correl
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ana_pair_correl_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, f_n, handle, list, list_ind, s_n
      REAL(KIND=dp), DIMENSION(3)                        :: cell_size
      TYPE(atom_pairs_type), DIMENSION(:), POINTER       :: pairs_tmp

      IF(.NOT.(ASSOCIATED(ana_pair_correl)))CALL cp__a("tmc/tmc_analysis.F",1019)
      IF(.NOT.(.NOT. ASSOCIATED(ana_pair_correl%g_r)))CALL cp__a("tmc/tmc_analysis.F",1020)
      IF(.NOT.(.NOT. ASSOCIATED(ana_pair_correl%pairs)))CALL cp__a("tmc/tmc_analysis.F",1021)
      IF(.NOT.(ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_analysis.F",1022)
      IF(.NOT.(SIZE(atoms) .GT. 1))CALL cp__a("tmc/tmc_analysis.F",1023)
      IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("tmc/tmc_analysis.F",1024)

      ! start the timing
      CALL timeset(routineN, handle)

      CALL get_cell(cell=cell, abc=cell_size)
      IF (ana_pair_correl%nr_bins .LE. 0) THEN
         ana_pair_correl%nr_bins = CEILING(MAXVAL(cell_size(:))/2.0_dp/(0.03/au2a))
      END IF
      ana_pair_correl%step_lenght = MAXVAL(cell_size(:))/2.0_dp/ &
                                    ana_pair_correl%nr_bins
      ana_pair_correl%conf_counter = 0

      counter = 1
      ! initialise the atom pairs
      ALLOCATE (pairs_tmp(SIZE(atoms)))
      DO f_n = 1, SIZE(atoms)
         DO s_n = f_n+1, SIZE(atoms)
            ! search if atom pair is already selected
            list_ind = search_pair_in_list(pair_list=pairs_tmp, n1=atoms(f_n)%name, &
                                           n2=atoms(s_n)%name, list_end=counter-1)
            ! add to list
            IF (list_ind .LT. 0) THEN
               pairs_tmp(counter)%f_n = atoms(f_n)%name
               pairs_tmp(counter)%s_n = atoms(s_n)%name
               pairs_tmp(counter)%pair_count = 1
               counter = counter+1
            ELSE
               pairs_tmp(list_ind)%pair_count = pairs_tmp(list_ind)%pair_count+1
            END IF
         END DO
      END DO

      ALLOCATE (ana_pair_correl%pairs(counter-1))
      DO list = 1, counter-1
         ana_pair_correl%pairs(list)%f_n = pairs_tmp(list)%f_n
         ana_pair_correl%pairs(list)%s_n = pairs_tmp(list)%s_n
         ana_pair_correl%pairs(list)%pair_count = pairs_tmp(list)%pair_count
      END DO
      DEALLOCATE (pairs_tmp)

      ALLOCATE (ana_pair_correl%g_r(SIZE(ana_pair_correl%pairs(:)), ana_pair_correl%nr_bins))
      ana_pair_correl%g_r = 0.0_dp
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE ana_pair_correl_init

! **************************************************************************************************
!> \brief calculates the radial distribution function
!> \param elem ...
!> \param weight ...
!> \param atoms ...
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE calc_paircorrelation(elem, weight, atoms, ana_env)
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: weight
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_paircorrelation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ind, j, pair_ind
      REAL(KIND=dp)                                      :: dist
      REAL(KIND=dp), DIMENSION(3)                        :: cell_size

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_analysis.F",1093)
      IF(.NOT.(ASSOCIATED(elem%pos)))CALL cp__a("tmc/tmc_analysis.F",1094)
      IF(.NOT.(ALL(elem%box_scale(:) .GT. 0.0_dp)))CALL cp__a("tmc/tmc_analysis.F",1095)
      IF(.NOT.(weight .GT. 0))CALL cp__a("tmc/tmc_analysis.F",1096)
      IF(.NOT.(ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_analysis.F",1097)
      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",1098)
      IF(.NOT.(ASSOCIATED(ana_env%cell)))CALL cp__a("tmc/tmc_analysis.F",1099)
      IF(.NOT.(ASSOCIATED(ana_env%pair_correl)))CALL cp__a("tmc/tmc_analysis.F",1100)
      IF(.NOT.(ASSOCIATED(ana_env%pair_correl%g_r)))CALL cp__a("tmc/tmc_analysis.F",1101)
      IF(.NOT.(ASSOCIATED(ana_env%pair_correl%pairs)))CALL cp__a("tmc/tmc_analysis.F",1102)

      ! start the timing
      CALL timeset(routineN, handle)

      dist = -1.0_dp

      first_elem_loop: DO i = 1, SIZE(elem%pos), ana_env%dim_per_elem
         second_elem_loop: DO j = i+3, SIZE(elem%pos), ana_env%dim_per_elem
            dist = nearest_distance(x1=elem%pos(i:i+ana_env%dim_per_elem-1), &
                                    x2=elem%pos(j:j+ana_env%dim_per_elem-1), &
                                    cell=ana_env%cell, box_scale=elem%box_scale)
            ind = CEILING(dist/ana_env%pair_correl%step_lenght)
            IF (ind .LE. ana_env%pair_correl%nr_bins) THEN
               pair_ind = search_pair_in_list(pair_list=ana_env%pair_correl%pairs, &
                                              n1=atoms(INT(i/REAL(ana_env%dim_per_elem, KIND=dp))+1)%name, &
                                              n2=atoms(INT(j/REAL(ana_env%dim_per_elem, KIND=dp))+1)%name)
               IF(.NOT.(pair_ind .GT. 0))CALL cp__a("tmc/tmc_analysis.F",1119)
               ana_env%pair_correl%g_r(pair_ind, ind) = &
                  ana_env%pair_correl%g_r(pair_ind, ind)+weight
            END IF
         END DO second_elem_loop
      END DO first_elem_loop
      ana_env%pair_correl%conf_counter = ana_env%pair_correl%conf_counter+weight
      CALL get_cell(cell=ana_env%cell, abc=cell_size)
      ana_env%pair_correl%sum_box_scale = ana_env%pair_correl%sum_box_scale+ &
                                          (elem%box_scale(:)*weight)
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE calc_paircorrelation

! **************************************************************************************************
!> \brief print the radial distribution function for each pair of atoms
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE print_paircorrelation(ana_env)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_paircorrelation', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: bin, file_ptr, handle, pair
      REAL(KIND=dp)                                      :: aver_box_scale(3), vol, voldr
      REAL(KIND=dp), DIMENSION(3)                        :: cell_size

      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",1150)
      IF(.NOT.(ASSOCIATED(ana_env%pair_correl)))CALL cp__a("tmc/tmc_analysis.F",1151)

      ! start the timing
      CALL timeset(routineN, handle)

      CALL get_cell(cell=ana_env%cell, abc=cell_size)
      aver_box_scale(:) = ana_env%pair_correl%sum_box_scale(:)/ana_env%pair_correl%conf_counter
      vol = (cell_size(1)*aver_box_scale(1))* &
            (cell_size(2)*aver_box_scale(2))* &
            (cell_size(3)*aver_box_scale(3))

      DO pair = 1, SIZE(ana_env%pair_correl%pairs)
         file_name = expand_file_name_temp(TRIM(ana_env%out_file_prefix)// &
                                           tmc_ana_pair_correl_file_name, &
                                           ana_env%temperature)
         CALL open_file(file_name=expand_file_name_char( &
                        expand_file_name_char(file_name, &
                                              ana_env%pair_correl%pairs(pair)%f_n), &
                        ana_env%pair_correl%pairs(pair)%s_n), &
                        file_status="REPLACE", &
                        file_action="WRITE", file_position="APPEND", &
                        unit_number=file_ptr)
         WRITE (file_ptr, *) "# radial distribution function of "// &
            TRIM(ana_env%pair_correl%pairs(pair)%f_n)//" and "// &
            TRIM(ana_env%pair_correl%pairs(pair)%s_n)//" of ", &
            ana_env%pair_correl%conf_counter, " configurations"
         WRITE (file_ptr, *) "# using a bin size of ", &
            ana_env%pair_correl%step_lenght*au2a, &
            "[A] (for Vol changes: refering to the reference cell)"
         DO bin = 1, ana_env%pair_correl%nr_bins
            voldr = 4.0/3.0*PI*ana_env%pair_correl%step_lenght**3* &
                    (REAL(bin, KIND=dp)**3-REAL(bin-1, KIND=dp)**3)
            WRITE (file_ptr, *) (bin-0.5)*ana_env%pair_correl%step_lenght*au2a, &
               (ana_env%pair_correl%g_r(pair, bin)/ana_env%pair_correl%conf_counter)/ &
               (voldr*ana_env%pair_correl%pairs(pair)%pair_count/vol)
         END DO
         CALL close_file(unit_number=file_ptr)

         IF (ana_env%print_test_output) &
            WRITE (*, *) "TMC|ANALYSIS_G_R_"// &
            TRIM(ana_env%pair_correl%pairs(pair)%f_n)//"_"// &
            TRIM(ana_env%pair_correl%pairs(pair)%s_n)//"_X= ", &
            SUM(ana_env%pair_correl%g_r(pair, :)/ana_env%pair_correl%conf_counter/ &
                voldr*ana_env%pair_correl%pairs(pair)%pair_count/vol)
      END DO

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE print_paircorrelation

   !============================================================================
   ! classical cell dipole moment
   !============================================================================

! **************************************************************************************************
!> \brief init radial distribution function structures
!> \param ana_dip_mom ...
!> \param atoms ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE ana_dipole_moment_init(ana_dip_mom, atoms)
      TYPE(dipole_moment_type), POINTER                  :: ana_dip_mom
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ana_dipole_moment_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom, charge, handle

      IF(.NOT.(ASSOCIATED(ana_dip_mom)))CALL cp__a("tmc/tmc_analysis.F",1221)
      IF(.NOT.(ASSOCIATED(ana_dip_mom%charges_inp)))CALL cp__a("tmc/tmc_analysis.F",1222)
      IF(.NOT.(ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_analysis.F",1223)

      ! start the timing
      CALL timeset(routineN, handle)

      ALLOCATE (ana_dip_mom%charges(SIZE(atoms)))
      ana_dip_mom%charges = 0.0_dp
      ! for every atom searcht the correct charge
      DO atom = 1, SIZE(atoms)
         charge_loop: DO charge = 1, SIZE(ana_dip_mom%charges_inp)
            IF (atoms(atom)%name .EQ. ana_dip_mom%charges_inp(charge)%name) THEN
               ana_dip_mom%charges(atom) = ana_dip_mom%charges_inp(charge)%mass
               EXIT charge_loop
            END IF
         END DO charge_loop
      END DO

      DEALLOCATE (ana_dip_mom%charges_inp)
      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE ana_dipole_moment_init

! **************************************************************************************************
!> \brief calculates the classical cell dipole moment
!> \param elem ...
!> \param weight ...
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE calc_dipole_moment(elem, weight, ana_env)
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: weight
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_dipole_moment', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name
      INTEGER                                            :: handle, i
      REAL(KIND=dp), DIMENSION(:), POINTER               :: dip_cl

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_analysis.F",1265)
      IF(.NOT.(ASSOCIATED(elem%pos)))CALL cp__a("tmc/tmc_analysis.F",1266)
      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",1267)
      IF(.NOT.(ASSOCIATED(ana_env%dip_mom)))CALL cp__a("tmc/tmc_analysis.F",1268)
      IF(.NOT.(ASSOCIATED(ana_env%dip_mom%charges)))CALL cp__a("tmc/tmc_analysis.F",1269)

      ! start the timing
      CALL timeset(routineN, handle)

      ALLOCATE (dip_cl(ana_env%dim_per_elem))
      dip_cl(:) = 0.0_dp

      DO i = 1, SIZE(elem%pos, 1), ana_env%dim_per_elem
         dip_cl(:) = dip_cl(:)+elem%pos(i:i+ana_env%dim_per_elem-1)* &
                     ana_env%dip_mom%charges(INT(i/REAL(ana_env%dim_per_elem, KIND=dp))+1)
      END DO

      ! if there are no exact dipoles save these ones in element structure
      IF (.NOT. ASSOCIATED(elem%dipole)) THEN
         ALLOCATE (elem%dipole(ana_env%dim_per_elem))
         elem%dipole(:) = dip_cl(:)
      END IF

      IF (ana_env%dip_mom%print_cl_dip) THEN
         file_name = expand_file_name_temp(tmc_default_trajectory_file_name, &
                                           ana_env%temperature)
         CALL write_dipoles_in_file(file_name=file_name, &
                                    conf_nr=ana_env%dip_mom%conf_counter+1, dip=dip_cl, &
                                    file_ext="dip_cl")
      END IF
      ana_env%dip_mom%conf_counter = ana_env%dip_mom%conf_counter+weight
      ana_env%dip_mom%last_dip_cl(:) = dip_cl

      DEALLOCATE (dip_cl)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE calc_dipole_moment

! **************************************************************************************************
!> \brief prints final values for classical cell dipole moment calculation
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE print_dipole_moment(ana_env)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_dipole_moment', &
         routineP = moduleN//':'//routineN

      IF (ana_env%print_test_output) &
         WRITE (*, *) "TMC|ANALYSIS_FINAL_CLASS_CELL_DIPOLE_MOMENT_X= ", &
         ana_env%dip_mom%last_dip_cl(:)
   END SUBROUTINE print_dipole_moment

! **************************************************************************************************
!> \brief calculates the dipole moment analysis
!> \param elem ...
!> \param weight ...
!> \param ana_env ...
!> \param
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE calc_dipole_analysis(elem, weight, ana_env)
      TYPE(tree_type), POINTER                           :: elem
      INTEGER                                            :: weight
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_dipole_analysis', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: vol, weight_act
      REAL(KIND=dp), DIMENSION(3, 3)                     :: tmp_dip
      TYPE(cell_type), POINTER                           :: scaled_cell

      NULLIFY (scaled_cell)

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_analysis.F",1343)
      IF(.NOT.(ASSOCIATED(elem%dipole)))CALL cp__a("tmc/tmc_analysis.F",1344)
      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",1345)
      IF(.NOT.(ASSOCIATED(ana_env%dip_ana)))CALL cp__a("tmc/tmc_analysis.F",1346)

      weight_act = weight
      IF (ana_env%dip_ana%ana_type .EQ. ana_type_sym_xyz) &
         weight_act = weight_act/REAL(8.0, KIND=dp)

      ! get the volume
      ALLOCATE (scaled_cell)
      CALL get_scaled_cell(cell=ana_env%cell, box_scale=elem%box_scale, vol=vol, &
                           scaled_cell=scaled_cell)

      ! fold exact dipole moments using the classical ones
      IF (ASSOCIATED(ana_env%dip_mom)) THEN
         IF (ALL(ana_env%dip_mom%last_dip_cl .NE. elem%dipole)) THEN
            elem%dipole = pbc(r=elem%dipole(:)-ana_env%dip_mom%last_dip_cl, &
                              cell=scaled_cell)+ana_env%dip_mom%last_dip_cl
         END IF
      END IF

      ana_env%dip_ana%conf_counter = ana_env%dip_ana%conf_counter+weight_act

      ! dipole sqared absolut value summed and weight_acted with volume and conf weight_act
      ana_env%dip_ana%mu2_pv_s = ana_env%dip_ana%mu2_pv_s+ &
                                 DOT_PRODUCT(elem%dipole(:), elem%dipole(:))/vol*weight_act

      tmp_dip(:, :) = 0.0_dp
      tmp_dip(:, 1) = elem%dipole(:)

      ! dipole sum, weight_acted with volume and conf weight_act
      ana_env%dip_ana%mu_pv(:) = ana_env%dip_ana%mu_pv(:)+ &
                                 tmp_dip(:, 1)/vol*weight_act

      ! dipole sum, weight_acted with square root of volume and conf weight_act
      ana_env%dip_ana%mu_psv(:) = ana_env%dip_ana%mu_psv(:)+ &
                                  tmp_dip(:, 1)/SQRT(vol)*weight_act

      ! dipole squared sum, weight_acted with volume and conf weight_act
      ana_env%dip_ana%mu2_pv(:) = ana_env%dip_ana%mu2_pv(:)+ &
                                  tmp_dip(:, 1)**2/vol*weight_act

      ! calculate the directional average with componentwise correlation per volume
      tmp_dip(:, :) = MATMUL(tmp_dip(:, :), TRANSPOSE(tmp_dip(:, :)))
      ana_env%dip_ana%mu2_pv_mat(:, :) = ana_env%dip_ana%mu2_pv_mat(:, :)+ &
                                         tmp_dip(:, :)/vol*weight_act

   END SUBROUTINE calc_dipole_analysis

! **************************************************************************************************
!> \brief prints the actual dipole moment analysis (trajectories)
!> \param elem ...
!> \param ana_env ...
!> \param
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE print_act_dipole_analysis(elem, ana_env)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_act_dipole_analysis', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, file_name_tmp
      INTEGER                                            :: counter_tmp, file_ptr
      LOGICAL                                            :: flag
      REAL(KIND=dp)                                      :: diel_const, diel_const_norm, &
                                                            diel_const_sym, e0, kB
      REAL(KIND=dp), DIMENSION(3, 3)                     :: tmp_dip

      kB = boltzmann/joule
      counter_tmp = INT(ana_env%dip_ana%conf_counter)

      ! TODO get correct constant using physcon
      e0 = 0.07957747154594767_dp !e^2*a0*me*hbar^-2
      diel_const_norm = 1/(3.0_dp*e0*kB*ana_env%temperature)

      file_name = expand_file_name_temp(TRIM(ana_env%out_file_prefix)// &
                                        tmc_default_trajectory_file_name, &
                                        ana_env%temperature)
      CALL write_dipoles_in_file(file_name=file_name, &
                                 conf_nr=INT(ana_env%dip_ana%conf_counter)+1, dip=elem%dipole, &
                                 file_ext="dip_folded")

      ! set ouput file name
      file_name_tmp = expand_file_name_temp(TRIM(ana_env%out_file_prefix)// &
                                            tmc_default_trajectory_file_name, &
                                            ana_env%temperature)

      SELECT CASE (ana_env%dip_ana%ana_type)
      CASE (ana_type_default)
         file_name = TRIM(expand_file_name_char(file_name_tmp, &
                                                "diel_const"))
         file_name_tmp = TRIM(expand_file_name_char(file_name_tmp, &
                                                    "diel_const_tensor"))
      CASE (ana_type_sym_xyz)
         file_name = TRIM(expand_file_name_char(file_name_tmp, &
                                                "diel_const_sym"))
         file_name_tmp = TRIM(expand_file_name_char(file_name_tmp, &
                                                    "diel_const_tensor_sym"))
      CASE DEFAULT
         CALL cp__w("tmc/tmc_analysis.F",1445,'unknown analysis type "'//cp_to_string(ana_env%dip_ana%ana_type)//'" used.')
      END SELECT

      ! calc the dielectric constant
      ! 1+( <M^2> - <M>^2 ) / (3*e_0*V*k*T)
      diel_const = 1.0_dp+(ana_env%dip_ana%mu2_pv_s/(ana_env%dip_ana%conf_counter)- &
                           DOT_PRODUCT(ana_env%dip_ana%mu_psv(:)/(ana_env%dip_ana%conf_counter), &
                                       ana_env%dip_ana%mu_psv(:)/(ana_env%dip_ana%conf_counter)))* &
                   diel_const_norm
      ! symmetrized dielctric constant
      ! 1+( <M^2> ) / (3*e_0*V*k*T)
      diel_const_sym = 1.0_dp+ana_env%dip_ana%mu2_pv_s/(ana_env%dip_ana%conf_counter)* &
                       diel_const_norm
      ! print dielectric constant trajectory
      !  if szmetry used print only every 8th configuration, hence every different (not mirrowed)
      INQUIRE (FILE=file_name, EXIST=flag)
      CALL open_file(file_name=file_name, file_status="UNKNOWN", &
                     file_action="WRITE", file_position="APPEND", &
                     unit_number=file_ptr)
      IF (.NOT. flag) THEN
         WRITE (file_ptr, FMT='(A8,5A20)') "# conf", "diel_const", &
            "diel_const_sym", "diel_const_sym_x", &
            "diel_const_sym_y", "diel_const_sym_z"
      END IF
      WRITE (file_ptr, FMT="(I8,10F20.10)") counter_tmp, diel_const, &
         diel_const_sym, &
         4.0_dp*PI/(kB*ana_env%temperature)* &
         ana_env%dip_ana%mu2_pv(:)/REAL(ana_env%dip_ana%conf_counter, KIND=dp)
      CALL close_file(unit_number=file_ptr)

      ! print dielectric constant tensor trajectory
      INQUIRE (FILE=file_name_tmp, EXIST=flag)
      CALL open_file(file_name=file_name_tmp, file_status="UNKNOWN", &
                     file_action="WRITE", file_position="APPEND", &
                     unit_number=file_ptr)
      IF (.NOT. flag) THEN
         WRITE (file_ptr, FMT='(A8,9A20)') "# conf", "xx", "xy", "xz", &
            "yx", "yy", "yz", &
            "zx", "zy", "zz"
      END IF
      tmp_dip(:, :) = 0.0_dp
      tmp_dip(:, 1) = ana_env%dip_ana%mu_psv(:)/REAL(ana_env%dip_ana%conf_counter, KIND=dp)

      WRITE (file_ptr, FMT="(I8,10F20.10)") counter_tmp, &
         4.0_dp*PI/(kB*ana_env%temperature)* &
         (ana_env%dip_ana%mu2_pv_mat(:, :)/REAL(ana_env%dip_ana%conf_counter, KIND=dp)- &
          MATMUL(tmp_dip(:, :), TRANSPOSE(tmp_dip(:, :))))
      CALL close_file(unit_number=file_ptr)
   END SUBROUTINE print_act_dipole_analysis

! **************************************************************************************************
!> \brief prints the dipole moment analysis
!> \param ana_env ...
!> \param
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE print_dipole_analysis(ana_env)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: fmt_my = '(T2,A,"| ",A,T41,A40)', plabel = "TMC_ANA", &
         routineN = 'print_dipole_analysis', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: diel_const_scalar, kB
      REAL(KIND=dp), DIMENSION(3)                        :: diel_const_sym, dielec_ev
      REAL(KIND=dp), DIMENSION(3, 3)                     :: diel_const, tmp_dip, tmp_ev

      kB = boltzmann/joule

      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",1514)
      IF(.NOT.(ASSOCIATED(ana_env%dip_ana)))CALL cp__a("tmc/tmc_analysis.F",1515)

      tmp_dip(:, :) = 0.0_dp
      diel_const(:, :) = 0.0_dp
      diel_const_scalar = 0.0_dp
      diel_const_sym = 0.0_dp

      !dielectric constant
      tmp_dip(:, 1) = ana_env%dip_ana%mu_psv(:)/REAL(ana_env%dip_ana%conf_counter, KIND=dp)
      diel_const(:, :) = 4.0_dp*PI/(kB*ana_env%temperature)* &
                         (ana_env%dip_ana%mu2_pv_mat(:, :)/REAL(ana_env%dip_ana%conf_counter, KIND=dp)- &
                          MATMUL(tmp_dip(:, :), TRANSPOSE(tmp_dip(:, :))))

      !dielectric constant for symmetric case
      diel_const_sym(:) = 4.0_dp*PI/(kB*ana_env%temperature)* &
                          ana_env%dip_ana%mu2_pv(:)/REAL(ana_env%dip_ana%conf_counter, KIND=dp)

      DO i = 1, 3
         diel_const(i, i) = diel_const(i, i)+1.0_dp ! +1 for unpolarizable models, 1.592 for polarizable
         diel_const_scalar = diel_const_scalar+diel_const(i, i)
      END DO
      diel_const_scalar = diel_const_scalar/REAL(3, KIND=dp)

      tmp_dip(:, :) = diel_const
      CALL diag(3, tmp_dip, dielec_ev, tmp_ev)

      ! print out results
      WRITE (ana_env%io_unit, FMT="(/,T2,A)") REPEAT("-", 79)
      WRITE (ana_env%io_unit, FMT="(T2,A,T35,A,T80,A)") "-", "average dipoles", "-"
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "temperature ", cp_to_string(ana_env%temperature)
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "used configurations ", &
         cp_to_string(REAL(ana_env%dip_ana%conf_counter, KIND=dp))
      IF (ana_env%dip_ana%ana_type .EQ. ana_type_ice) &
         WRITE (ana_env%io_unit, FMT='(T2,A,"| ",A)') plabel, &
         "ice analysis with directions of hexagonal structure"
      IF (ana_env%dip_ana%ana_type .EQ. ana_type_sym_xyz) &
         WRITE (ana_env%io_unit, FMT='(T2,A,"| ",A)') plabel, &
         "ice analysis with symmetrized dipoles in each direction."

      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "for product of 2 directions(per vol):"
      DO i = 1, 3
         WRITE (ana_env%io_unit, '(A,3F16.8,A)') " |", ana_env%dip_ana%mu2_pv_mat(i, :)/ &
            REAL(ana_env%dip_ana%conf_counter, KIND=dp), " |"
      END DO

      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "dielectric constant tensor:"
      DO i = 1, 3
         WRITE (ana_env%io_unit, '(A,3F16.8,A)') " |", diel_const(i, :), " |"
      END DO

      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "dielectric tensor eigenvalues", &
         cp_to_string(dielec_ev(1))//" "// &
         cp_to_string(dielec_ev(2))//" "// &
         cp_to_string(dielec_ev(3))
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "dielectric constant symm ", &
         cp_to_string(diel_const_sym(1))//" | "// &
         cp_to_string(diel_const_sym(2))//" | "// &
         cp_to_string(diel_const_sym(3))
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "dielectric constant ", &
         cp_to_string(diel_const_scalar)
      WRITE (ana_env%io_unit, FMT="(/,T2,A)") REPEAT("-", 79)

   END SUBROUTINE print_dipole_analysis

   !============================================================================
   ! particle displacement in cell (from one configuration to the next)
   !============================================================================

! **************************************************************************************************
!> \brief calculates the mean square displacement
!> \param elem ...
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE calc_displacement(elem, ana_env)
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_displacement', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: file_name, file_name_tmp
      INTEGER                                            :: file_ptr, handle, ind
      LOGICAL                                            :: flag
      REAL(KIND=dp)                                      :: disp
      REAL(KIND=dp), DIMENSION(3)                        :: atom_disp

      disp = 0.0_dp

      IF(.NOT.(ASSOCIATED(elem)))CALL cp__a("tmc/tmc_analysis.F",1605)
      IF(.NOT.(ASSOCIATED(elem%pos)))CALL cp__a("tmc/tmc_analysis.F",1606)
      IF(.NOT.(ASSOCIATED(ana_env)))CALL cp__a("tmc/tmc_analysis.F",1607)
      IF(.NOT.(ASSOCIATED(ana_env%displace)))CALL cp__a("tmc/tmc_analysis.F",1608)
      IF(.NOT.(ASSOCIATED(ana_env%last_elem)))CALL cp__a("tmc/tmc_analysis.F",1609)

      ! start the timing
      CALL timeset(routineN, handle)

      DO ind = 1, SIZE(elem%pos), ana_env%dim_per_elem
         ! fold into box
         atom_disp(:) = elem%pos(ind:ind+2)-ana_env%last_elem%pos(ind:ind+2)
         CALL get_scaled_cell(cell=ana_env%cell, box_scale=elem%box_scale, &
                              vec=atom_disp)
         disp = disp+SUM((atom_disp(:)*au2a)**2)
      END DO
      ana_env%displace%disp = ana_env%displace%disp+disp
      ana_env%displace%conf_counter = ana_env%displace%conf_counter+1

      IF (ana_env%displace%print_disp) THEN
         file_name_tmp = expand_file_name_temp(TRIM(ana_env%out_file_prefix)// &
                                               tmc_default_trajectory_file_name, &
                                               ana_env%temperature)
         file_name = TRIM(expand_file_name_char(file_name_tmp, &
                                                "devi"))
         INQUIRE (FILE=file_name, EXIST=flag)
         CALL open_file(file_name=file_name, file_status="UNKNOWN", &
                        file_action="WRITE", file_position="APPEND", &
                        unit_number=file_ptr)
         IF (.NOT. flag) &
            WRITE (file_ptr, *) "# conf     squared deviation of the cell"
         WRITE (file_ptr, *) elem%nr, disp
         CALL close_file(unit_number=file_ptr)
      END IF

      ! end the timing
      CALL timestop(handle)

   END SUBROUTINE calc_displacement

! **************************************************************************************************
!> \brief prints final values for the displacement calculations
!> \param ana_env ...
!> \param
!> \author Mandes 02.2013
! **************************************************************************************************
   SUBROUTINE print_average_displacement(ana_env)
      TYPE(tmc_analysis_env), POINTER                    :: ana_env

      CHARACTER(LEN=*), PARAMETER :: fmt_my = '(T2,A,"| ",A,T41,A40)', plabel = "TMC_ANA", &
         routineN = 'print_average_displacement', routineP = moduleN//':'//routineN

      WRITE (ana_env%io_unit, FMT="(/,T2,A)") REPEAT("-", 79)
      WRITE (ana_env%io_unit, FMT="(T2,A,T35,A,T80,A)") "-", "average displacement", "-"
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "temperature ", &
         cp_to_string(ana_env%temperature)
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "used configurations ", &
         cp_to_string(REAL(ana_env%displace%conf_counter, KIND=dp))
      WRITE (ana_env%io_unit, FMT=fmt_my) plabel, "cell root mean square deviation: ", &
         cp_to_string(SQRT(ana_env%displace%disp/ &
                           REAL(ana_env%displace%conf_counter, KIND=dp)))
      IF (ana_env%print_test_output) &
         WRITE (*, *) "TMC|ANALYSIS_AVERAGE_CELL_DISPLACEMENT_X= ", &
         SQRT(ana_env%displace%disp/ &
              REAL(ana_env%displace%conf_counter, KIND=dp))
   END SUBROUTINE print_average_displacement
END MODULE tmc_analysis
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module contains the worker routine handling the communication and
!>        the calculation / creation of the configurations
!>        - WORKER these are all TMC cores, instead of master core
!>          and maybe some idle cores
!>        - divided in groups, in every group exists group master
!>          - there can be two kind of groups, one for exact energy calculation
!>            and one calculating configurational change using an approximate
!>            potential
!>        - Algorithm:
!>          - group master receive messages and decide what to do,
!>          - (if nessesary) broadcast of working task
!>            to all other group members (needed for parallel CP2K)
!>          - process task, calculations of energy or configurational change
!>          - result, exist on group master, sent to master core
!>        Communication structure (master->worker, worker->master):
!>        - message structure is defined in TMC message module
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_worker
   USE cell_types,                      ONLY: cell_copy,&
                                              cell_type,&
                                              init_cell
   USE cp_external_control,             ONLY: set_external_comm
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_methods,               ONLY: cp_results_erase,&
                                              put_results
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE f77_interface,                   ONLY: f_env_get_from_id,&
                                              f_env_type,&
                                              get_natom,&
                                              get_pos,&
                                              get_result_r1
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_get_natom
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE parallel_rng_types,              ONLY: get_rng_stream,&
                                              next_random_number,&
                                              set_rng_stream
   USE particle_list_types,             ONLY: particle_list_type
   USE tmc_analysis,                    ONLY: analysis_init,&
                                              analysis_restart_print,&
                                              analysis_restart_read,&
                                              analyze_file_configurations,&
                                              do_tmc_analysis,&
                                              finalize_tmc_analysis
   USE tmc_analysis_types,              ONLY: tmc_ana_list_type
   USE tmc_calculations,                ONLY: calc_potential_energy
   USE tmc_messages,                    ONLY: bcast_group,&
                                              check_if_group_master,&
                                              communicate_atom_types,&
                                              master_comm_id,&
                                              recv_msg,&
                                              send_msg,&
                                              stop_whole_group,&
                                              tmc_message
   USE tmc_move_handle,                 ONLY: clear_move_probs,&
                                              prob_update,&
                                              select_random_move_type
   USE tmc_move_types,                  ONLY: mv_type_MD,&
                                              mv_type_NMC_moves
   USE tmc_moves,                       ONLY: change_pos
   USE tmc_stati,                       ONLY: &
        TMC_CANCELING_MESSAGE, TMC_CANCELING_RECEIPT, TMC_STATUS_CALCULATING, TMC_STATUS_FAILED, &
        TMC_STATUS_STOP_RECEIPT, TMC_STATUS_WAIT_FOR_NEW_TASK, TMC_STATUS_WORKER_INIT, &
        TMC_STAT_ANALYSIS_REQUEST, TMC_STAT_ANALYSIS_RESULT, TMC_STAT_APPROX_ENERGY_REQUEST, &
        TMC_STAT_APPROX_ENERGY_RESULT, TMC_STAT_ENERGY_REQUEST, TMC_STAT_ENERGY_RESULT, &
        TMC_STAT_INIT_ANALYSIS, TMC_STAT_MD_REQUEST, TMC_STAT_MD_RESULT, TMC_STAT_NMC_REQUEST, &
        TMC_STAT_NMC_RESULT, TMC_STAT_SCF_STEP_ENER_RECEIVE, TMC_STAT_START_CONF_REQUEST, &
        TMC_STAT_START_CONF_RESULT, task_type_MC, task_type_ideal_gas
   USE tmc_tree_acceptance,             ONLY: acceptance_check
   USE tmc_tree_build,                  ONLY: allocate_new_sub_tree_node,&
                                              deallocate_sub_tree_node
   USE tmc_tree_types,                  ONLY: tree_type
   USE tmc_types,                       ONLY: allocate_tmc_atom_type,&
                                              tmc_atom_type,&
                                              tmc_env_type,&
                                              tmc_param_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_worker'

   PUBLIC :: do_tmc_worker
   PUBLIC :: get_initial_conf, get_atom_kinds_and_cell

   INTEGER, PARAMETER :: DEBUG = 0

CONTAINS

! **************************************************************************************************
!> \brief worker get tasks form master and fulfill them
!> \param tmc_env structure for storing all the tmc parameters
!> \param ana_list ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE do_tmc_worker(tmc_env, ana_list)
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      TYPE(tmc_ana_list_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ana_list

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_tmc_worker', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: c_tmp
      INTEGER                                            :: calc_stat, handle, i1, i2, ierr, itmp, &
                                                            num_dim, work_stat
      INTEGER, DIMENSION(:), POINTER                     :: ana_restart_conf
      LOGICAL                                            :: flag, master
      TYPE(cp_para_env_type), POINTER                    :: para_env_m_w
      TYPE(tree_type), POINTER                           :: conf

      master = .FALSE.
      i1 = -1
      i2 = -1
      NULLIFY (conf, para_env_m_w, ana_restart_conf)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_worker.F",133)

      ! start the timing
      CALL timeset(routineN, handle)

      ! initialize
      IF (tmc_env%tmc_comp_set%group_nr .GT. 0) THEN
         IF(.NOT.(ASSOCIATED(tmc_env%tmc_comp_set%para_env_sub_group)))CALL cp__a("tmc/tmc_worker.F",140)
         IF (tmc_env%w_env%env_id_ener .GT. 0) THEN
            itmp = tmc_env%w_env%env_id_ener
         ELSE
            itmp = tmc_env%w_env%env_id_approx
         END IF

         CALL get_atom_kinds_and_cell(env_id=itmp, &
                                      atoms=tmc_env%params%atoms, cell=tmc_env%params%cell)
         para_env_m_w => tmc_env%tmc_comp_set%para_env_m_w
         master = check_if_group_master(tmc_env%tmc_comp_set%para_env_sub_group)
      ELSE
         ! analysis group
         IF(.NOT.(ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_ana)))CALL cp__a("tmc/tmc_worker.F",153)
         para_env_m_w => tmc_env%tmc_comp_set%para_env_m_ana
         master = .TRUE.
      END IF

      !-- GROUP MASTER only --------------
      ! get messages from master and handle them
      IF (master) THEN
         ! NOT the analysis group
         IF (tmc_env%tmc_comp_set%group_nr .GT. 0) THEN
            IF (tmc_env%w_env%env_id_ener .GT. 0) THEN
               itmp = tmc_env%w_env%env_id_ener
            ELSE
               itmp = tmc_env%w_env%env_id_approx
            END IF
            ! set the communicator in the external control for receiving exit tags
            !  and sending additional information (e.g. the intermediate scf energies)
            IF (tmc_env%params%use_scf_energy_info) &
               CALL set_intermediate_info_comm(env_id=itmp, &
                                               comm=tmc_env%tmc_comp_set%para_env_m_w%group)
            IF (tmc_env%params%SPECULATIVE_CANCELING) &
               CALL set_external_comm(comm=tmc_env%tmc_comp_set%para_env_m_w%group, &
                                      in_external_master_id=MASTER_COMM_ID, &
                                      in_exit_tag=TMC_CANCELING_MESSAGE)
         END IF
         !-- WORKING LOOP --!
         master_work_time: DO
            work_stat = TMC_STATUS_WAIT_FOR_NEW_TASK
            ! -- receive message from master
            ! check for new task (wait for it)
            itmp = MASTER_COMM_ID
            CALL tmc_message(msg_type=work_stat, send_recv=recv_msg, &
                             dest=itmp, &
                             para_env=para_env_m_w, &
                             result_count=ana_restart_conf, &
                             tmc_params=tmc_env%params, elem=conf)

            IF (DEBUG .GE. 1 .AND. work_stat .NE. TMC_STATUS_WAIT_FOR_NEW_TASK) &
               WRITE (tmc_env%w_env%io_unit, *) "worker: group master of group ", &
               tmc_env%tmc_comp_set%group_nr, "got task ", work_stat
            calc_stat = TMC_STATUS_CALCULATING
            SELECT CASE (work_stat)
            CASE (TMC_STATUS_WAIT_FOR_NEW_TASK)
            CASE (TMC_STATUS_WORKER_INIT)
               CALL init_cell(cell=tmc_env%params%cell)
               itmp = bcast_group
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                                tmc_params=tmc_env%params)
            CASE (TMC_CANCELING_MESSAGE)
               work_stat = TMC_CANCELING_RECEIPT
               itmp = MASTER_COMM_ID
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=para_env_m_w, &
                                tmc_params=tmc_env%params)
            CASE (TMC_STATUS_FAILED)
               IF (DEBUG .GE. 1) &
                  WRITE (tmc_env%w_env%io_unit, *) "master worker of group", &
                  tmc_env%tmc_comp_set%group_nr, " exit work time."
               EXIT master_work_time
               !-- group master read the CP2K input file, and write data to master
            CASE (TMC_STAT_START_CONF_REQUEST)
               IF (tmc_env%w_env%env_id_ener .GT. 0) THEN
                  itmp = tmc_env%w_env%env_id_ener
               ELSE
                  itmp = tmc_env%w_env%env_id_approx
               END IF
               CALL get_initial_conf(tmc_params=tmc_env%params, init_conf=conf, &
                                     env_id=itmp)
               ! send start configuration back to master
               work_stat = TMC_STAT_START_CONF_RESULT
               itmp = MASTER_COMM_ID
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=para_env_m_w, &
                                tmc_params=tmc_env%params, elem=conf, &
                                wait_for_message=.TRUE.)

               IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_first_w)) &
                  CALL communicate_atom_types(atoms=tmc_env%params%atoms, &
                                              source=1, &
                                              para_env=tmc_env%tmc_comp_set%para_env_m_first_w)
               !-- calculate the approximate energy
            CASE (TMC_STAT_APPROX_ENERGY_REQUEST)
               IF(.NOT.(tmc_env%w_env%env_id_approx .GT. 0))CALL cp__a("tmc/tmc_worker.F",239)
               itmp = bcast_group
               !-- DISTRIBUTING WORK (group master) to all other group members
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                                tmc_params=tmc_env%params, elem=conf)
               CALL calc_potential_energy(conf=conf, &
                                          env_id=tmc_env%w_env%env_id_approx, &
                                          exact_approx_pot=.FALSE., &
                                          tmc_env=tmc_env)
               work_stat = TMC_STAT_APPROX_ENERGY_RESULT
               itmp = MASTER_COMM_ID
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=para_env_m_w, &
                                tmc_params=tmc_env%params, elem=conf)
               ! -- Nested Monte Carlo routines
            CASE (TMC_STAT_MD_REQUEST, TMC_STAT_NMC_REQUEST)
               CALL clear_move_probs(tmc_env%params%nmc_move_types)
               itmp = bcast_group
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                                tmc_params=tmc_env%params, elem=conf)
               !-- collective calculation for MD/NMC steps
               IF (work_stat .EQ. TMC_STAT_NMC_REQUEST) THEN
                  !-- calculate MD steps, in case of 2 different potentials do nested Monte Carlo
                  CALL nested_markov_chain_MC(conf=conf, &
                                              env_id=tmc_env%w_env%env_id_approx, &
                                              tmc_env=tmc_env, calc_status=calc_stat)
               ELSEIF (work_stat .EQ. TMC_STAT_MD_REQUEST) THEN
                  !TODO Hybrid MC routine
                  CALL cp__b("tmc/tmc_worker.F",272,"there is no Hybrid MC implemented yet.")

               ELSE
                  CALL cp__b("tmc/tmc_worker.F",275,"unknown task type for workers.")
               END IF
               !-- in case of cancelation send receipt
               itmp = MASTER_COMM_ID
               CALL tmc_message(msg_type=calc_stat, send_recv=recv_msg, &
                                dest=itmp, &
                                para_env=para_env_m_w, &
                                tmc_params=tmc_env%params, &
                                success=flag)
               SELECT CASE (calc_stat)
               CASE (TMC_STATUS_CALCULATING)
                  SELECT CASE (work_stat)
                  CASE (TMC_STAT_MD_REQUEST)
                     work_stat = TMC_STAT_MD_RESULT
                  CASE (TMC_STAT_NMC_REQUEST)
                     work_stat = TMC_STAT_NMC_RESULT
                  CASE DEFAULT
                     CALL cp_abort(cp__l("tmc/tmc_worker.F",292), &
                                   "unknown work status after possible NMC subgroup "// &
                                   "cancelation, work_stat="//cp_to_string(work_stat))
                  END SELECT
               CASE (TMC_CANCELING_MESSAGE)
                  work_stat = TMC_CANCELING_RECEIPT
               CASE DEFAULT
                  CALL cp_abort(cp__l("tmc/tmc_worker.F",299), &
                                "unknown calc status before sending NMC result "// &
                                cp_to_string(calc_stat))
               END SELECT
               ! send message back to master
               itmp = MASTER_COMM_ID
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=para_env_m_w, &
                                tmc_params=tmc_env%params, elem=conf)
            CASE (TMC_STAT_ENERGY_REQUEST)
               IF(.NOT.(tmc_env%w_env%env_id_ener .GT. 0))CALL cp__a("tmc/tmc_worker.F",310)
               !-- DISTRIBUTING WORK (group master) to all other group members
               itmp = bcast_group
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                                tmc_params=tmc_env%params, elem=conf)

               CALL calc_potential_energy(conf=conf, &
                                          env_id=tmc_env%w_env%env_id_ener, &
                                          exact_approx_pot=.TRUE., &
                                          tmc_env=tmc_env)
               !-- in case of cancelation send receipt
               flag = .FALSE.
               itmp = MASTER_COMM_ID
               CALL tmc_message(msg_type=calc_stat, send_recv=recv_msg, &
                                dest=itmp, &
                                para_env=para_env_m_w, &
                                tmc_params=tmc_env%params, success=flag)
               SELECT CASE (calc_stat)
               CASE (TMC_STATUS_CALCULATING)
                  SELECT CASE (work_stat)
                  CASE (TMC_STAT_ENERGY_REQUEST)
                     work_stat = TMC_STAT_ENERGY_RESULT
                     !-- if nessesary get the exact dipoles (for e.g. quantum potential)
                     IF (tmc_env%params%print_dipole) THEN
                        c_tmp = "[DIPOLE]"
                        CALL get_result_r1(env_id=tmc_env%w_env%env_id_ener, &
                                           description=c_tmp, N=3, RESULT=conf%dipole, &
                                           res_exist=flag, ierr=ierr)
                        IF (.NOT. flag) tmc_env%params%print_dipole = .FALSE.
                        ! TODO maybe let run with the changed option, but inform user properly
                        IF (.NOT. flag) &
                           CALL cp_abort(cp__l("tmc/tmc_worker.F",343), &
                                         "TMC: The requested dipoles are not porvided by the "// &
                                         "force environment.")
                     END IF
                  CASE DEFAULT
                     CALL cp_abort(cp__l("tmc/tmc_worker.F",348), &
                                   "energy worker should handle unknown stat "// &
                                   cp_to_string(work_stat))
                  END SELECT
               CASE (TMC_CANCELING_MESSAGE)
                  work_stat = TMC_CANCELING_RECEIPT
               CASE DEFAULT
                  CALL cp_abort(cp__l("tmc/tmc_worker.F",355), &
                                "worker while energy calc is in unknown state "// &
                                cp_to_string(work_stat))
               END SELECT

               !-- send information back to master
               IF (DEBUG .GE. 1) &
                  WRITE (tmc_env%w_env%io_unit, *) "worker group ", &
                  tmc_env%tmc_comp_set%group_nr, &
                  "calculations done, send result energy", conf%potential
               itmp = MASTER_COMM_ID
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=para_env_m_w, &
                                tmc_params=tmc_env%params, elem=conf)
            CASE (TMC_STAT_INIT_ANALYSIS)
               IF(.NOT.(ASSOCIATED(ana_restart_conf)))CALL cp__a("tmc/tmc_worker.F",371)
               IF(.NOT.(SIZE(ana_restart_conf) .EQ. tmc_env%params%nr_temp))CALL cp__a("tmc/tmc_worker.F",372)
               IF(.NOT.(PRESENT(ana_list)))CALL cp__a("tmc/tmc_worker.F",373)
               IF(.NOT.(ASSOCIATED(ana_list)))CALL cp__a("tmc/tmc_worker.F",374)
               itmp = MASTER_COMM_ID
               CALL communicate_atom_types(atoms=tmc_env%params%atoms, &
                                           source=itmp, para_env=tmc_env%tmc_comp_set%para_env_m_ana)

               num_dim = SIZE(conf%pos)
               DO itmp = 1, tmc_env%params%nr_temp
                  ! do not forget to nullify the pointer at the end, deallcoated at tmc_env%params
                  ana_list(itmp)%temp%temperature = tmc_env%params%Temp(itmp)
                  ana_list(itmp)%temp%atoms => tmc_env%params%atoms
                  ana_list(itmp)%temp%cell => tmc_env%params%cell
!              ana_list(itmp)%temp%io_unit     = tmc_env%w_env%io_unit

                  CALL analysis_init(ana_env=ana_list(itmp)%temp, nr_dim=num_dim)
                  ana_list(itmp)%temp%print_test_output = tmc_env%params%print_test_output
                  IF (.NOT. ASSOCIATED(conf)) &
                     CALL allocate_new_sub_tree_node(tmc_params=tmc_env%params, &
                                                     next_el=conf, nr_dim=num_dim)
                  CALL analysis_restart_read(ana_env=ana_list(itmp)%temp, &
                                             elem=conf)
                  !check if we have the read the file
                  flag = .FALSE.
                  IF ((.NOT. ASSOCIATED(ana_list(itmp)%temp%last_elem)) .AND. &
                      ana_restart_conf(itmp) .GT. 0) THEN
                     flag = .TRUE.
                     i1 = 0
                     i2 = ana_restart_conf(itmp)
                     CALL cp_warn(cp__l("tmc/tmc_worker.F",401), &
                                  "analysis old trajectory up to "// &
                                  "elem "//cp_to_string(ana_restart_conf(itmp))// &
                                  ". Read trajectory file.")
                  ELSE IF (ASSOCIATED(ana_list(itmp)%temp%last_elem)) THEN
                     IF (.NOT. (ana_list(itmp)%temp%last_elem%nr .EQ. ana_restart_conf(itmp))) THEN
                        flag = .TRUE.
                        i1 = ana_list(itmp)%temp%last_elem%nr
                        i2 = ana_restart_conf(itmp)
                        CALL cp_warn(cp__l("tmc/tmc_worker.F",410), &
                                     "analysis restart with the incorrect configuration "// &
                                     "TMC "//cp_to_string(ana_restart_conf(itmp))// &
                                     " ana "//cp_to_string(ana_list(itmp)%temp%last_elem%nr)// &
                                     ". REread trajectory file.")
                     END IF
                  END IF
                  IF (flag) THEN
                     CALL analyze_file_configurations(start_id=i1, &
                                                      end_id=i2, &
                                                      ana_env=ana_list(itmp)%temp, &
                                                      tmc_params=tmc_env%params)
                  END IF
               END DO
            CASE (TMC_STAT_ANALYSIS_REQUEST)
               IF(.NOT.(PRESENT(ana_list)))CALL cp__a("tmc/tmc_worker.F",425)
               IF(.NOT.(ASSOCIATED(ana_list(conf%sub_tree_nr)%temp)))CALL cp__a("tmc/tmc_worker.F",426)
               CALL do_tmc_analysis(elem=conf, &
                                    ana_env=ana_list(conf%sub_tree_nr)%temp)
               work_stat = TMC_STAT_ANALYSIS_RESULT
               itmp = MASTER_COMM_ID
               CALL tmc_message(msg_type=work_stat, send_recv=send_msg, &
                                dest=itmp, &
                                para_env=para_env_m_w, &
                                tmc_params=tmc_env%params, elem=conf)
            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_worker.F",436), &
                             "worker received unknown message task type "// &
                             cp_to_string(work_stat))
            END SELECT

            IF (DEBUG .GE. 1 .AND. work_stat .NE. TMC_STATUS_WAIT_FOR_NEW_TASK) &
               WRITE (tmc_env%w_env%io_unit, *) "worker: group ", &
               tmc_env%tmc_comp_set%group_nr, &
               "send back status:", work_stat
            IF (ASSOCIATED(conf)) &
               CALL deallocate_sub_tree_node(tree_elem=conf)
         END DO master_work_time
         !-- every other group paricipants----------------------------------------
      ELSE
         worker_work_time: DO
            work_stat = TMC_STATUS_WAIT_FOR_NEW_TASK
            flag = .FALSE.
            itmp = bcast_group
            CALL tmc_message(msg_type=work_stat, send_recv=recv_msg, &
                             dest=itmp, &
                             para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                             tmc_params=tmc_env%params, elem=conf)
            calc_stat = TMC_STATUS_CALCULATING
            SELECT CASE (work_stat)
            CASE (TMC_STATUS_WORKER_INIT)
               CALL init_cell(cell=tmc_env%params%cell)
            CASE (TMC_CANCELING_MESSAGE)
               ! error message
            CASE (TMC_STATUS_FAILED)
               EXIT worker_work_time
               ! all group members have to calculate the (MD potential) energy together
            CASE (TMC_STAT_START_CONF_RESULT)
               IF(.NOT.(tmc_env%w_env%env_id_approx .GT. 0))CALL cp__a("tmc/tmc_worker.F",468)
               !-- collective calculation of the potential energy of MD potential
               SELECT CASE (tmc_env%params%task_type)
               CASE (task_type_MC, task_type_ideal_gas)
                  IF (tmc_env%params%NMC_inp_file .NE. "") THEN
                     conf%box_scale(:) = 1.0_dp
                     CALL calc_potential_energy(conf=conf, &
                                                env_id=tmc_env%w_env%env_id_approx, &
                                                exact_approx_pot=.FALSE., &
                                                tmc_env=tmc_env)
                  END IF
               CASE DEFAULT
                  CALL cp_abort(cp__l("tmc/tmc_worker.F",480), &
                                "unknown task_type for participants in "// &
                                "START_CONF_RESULT request ")
               END SELECT
               !-- HMC - calculating MD steps
            CASE (TMC_STAT_NMC_REQUEST, TMC_STAT_MD_REQUEST)
               !-- collective calculation for MD/NMC steps
               IF (work_stat .EQ. TMC_STAT_NMC_REQUEST) THEN
                  !-- calculate MD steps, in case of 2 different potentials do nested Monte Carlo
                  CALL nested_markov_chain_MC(conf=conf, &
                                              env_id=tmc_env%w_env%env_id_approx, &
                                              tmc_env=tmc_env, calc_status=calc_stat)
               ELSEIF (work_stat .EQ. TMC_STAT_MD_REQUEST) THEN
                  !TODO Hybrid MC routine
                  CALL cp__b("tmc/tmc_worker.F",494,"there is no Hybrid MC implemented yet.")

               ELSE
                  CALL cp__b("tmc/tmc_worker.F",497,"unknown task type for workers.")
               END IF
               !-- energy calculations
            CASE (TMC_STAT_APPROX_ENERGY_REQUEST)
               !--- do calculate energy
               IF(.NOT.(tmc_env%w_env%env_id_approx .GT. 0))CALL cp__a("tmc/tmc_worker.F",502)
               CALL calc_potential_energy(conf=conf, &
                                          env_id=tmc_env%w_env%env_id_approx, &
                                          exact_approx_pot=.FALSE., &
                                          tmc_env=tmc_env)
            CASE (TMC_STAT_ENERGY_REQUEST)
               !--- do calculate energy
               IF(.NOT.(tmc_env%w_env%env_id_ener .GT. 0))CALL cp__a("tmc/tmc_worker.F",509)
               CALL calc_potential_energy(conf=conf, &
                                          env_id=tmc_env%w_env%env_id_ener, &
                                          exact_approx_pot=.TRUE., &
                                          tmc_env=tmc_env)
            CASE DEFAULT
               CALL cp_abort(cp__l("tmc/tmc_worker.F",515), &
                             "group participant got unknown working type "// &
                             cp_to_string(work_stat))
            END SELECT
            IF (ASSOCIATED(conf)) &
               CALL deallocate_sub_tree_node(tree_elem=conf)
         END DO worker_work_time
      END IF
      ! --------------------------------------------------------------------
      ! finalizing analysis, writing files etc.
      IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_ana)) THEN
         DO itmp = 1, tmc_env%params%nr_temp
            CALL analysis_restart_print(ana_env=ana_list(itmp)%temp)
            IF (ASSOCIATED(conf)) &
               CALL deallocate_sub_tree_node(tree_elem=ana_list(itmp)%temp%last_elem)
            CALL finalize_tmc_analysis(ana_list(itmp)%temp)
         END DO
      END IF
      !-- stopping and finalizing
      ! sending back receipt for stopping
      IF (master) THEN
         ! NOT the analysis group
         IF (tmc_env%tmc_comp_set%group_nr .GT. 0) THEN
            ! remove the communicator in the external control for receiving exit tags
            !  and sending additional information (e.g. the intermediate scf energies)
            IF (tmc_env%params%use_scf_energy_info) THEN
               IF (tmc_env%w_env%env_id_ener .GT. 0) THEN
                  itmp = tmc_env%w_env%env_id_ener
               ELSE
                  itmp = tmc_env%w_env%env_id_approx
               END IF
               CALL remove_intermediate_info_comm(env_id=itmp)
            END IF
         END IF
         IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_sub_group)) &
            CALL stop_whole_group(para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                                  tmc_params=tmc_env%params)

         work_stat = TMC_STATUS_STOP_RECEIPT
         itmp = MASTER_COMM_ID
         CALL tmc_message(msg_type=work_stat, send_recv=send_msg, dest=itmp, &
                          para_env=para_env_m_w, &
                          tmc_params=tmc_env%params)
      ELSE IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_sub_group)) THEN
         work_stat = TMC_STATUS_STOP_RECEIPT
         itmp = MASTER_COMM_ID
         CALL tmc_message(msg_type=work_stat, send_recv=send_msg, dest=itmp, &
                          para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                          tmc_params=tmc_env%params)
      END IF

      IF (DEBUG .GE. 5) &
         WRITE (tmc_env%w_env%io_unit, *) "worker ", &
         tmc_env%tmc_comp_set%para_env_sub_group%mepos, "of group ", &
         tmc_env%tmc_comp_set%group_nr, "stops working!!!!!!!!!!!!!!!!!!"

      IF (PRESENT(ana_list)) THEN
         DO itmp = 1, tmc_env%params%nr_temp
            ana_list(itmp)%temp%atoms => NULL()
            ana_list(itmp)%temp%cell => NULL()
         END DO
      END IF
      IF (ASSOCIATED(conf)) &
         CALL deallocate_sub_tree_node(tree_elem=conf)
      IF (ASSOCIATED(ana_restart_conf)) DEALLOCATE (ana_restart_conf)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE do_tmc_worker

! **************************************************************************************************
!> \brief Nested Monte Carlo (NMC), do several Markov Chain Monte Carlo steps
!>        usually using the approximate potential, could be also Hybrid MC.
!>        The amount of steps are predefinded by the user, but should be huge
!>        enough to reach the equilibrium state for this potential
!> \param conf ...
!> \param env_id ...
!> \param tmc_env ...
!> \param calc_status ...
!> \param
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE nested_markov_chain_MC(conf, env_id, tmc_env, calc_status)
      TYPE(tree_type), POINTER                           :: conf
      INTEGER, INTENT(IN)                                :: env_id
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      INTEGER, INTENT(OUT)                               :: calc_status

      CHARACTER(LEN=*), PARAMETER :: routineN = 'nested_markov_chain_MC', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: comm_dest, handle, substeps
      LOGICAL                                            :: accept, change_rejected, flag
      REAL(KIND=dp)                                      :: rnd_nr
      TYPE(tree_type), POINTER                           :: last_acc_conf

      NULLIFY (last_acc_conf)

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_worker.F",613)
      IF(.NOT.(ASSOCIATED(tmc_env%params)))CALL cp__a("tmc/tmc_worker.F",614)
      IF(.NOT.(ASSOCIATED(tmc_env%tmc_comp_set)))CALL cp__a("tmc/tmc_worker.F",615)
      IF(.NOT.(ASSOCIATED(tmc_env%rng_stream)))CALL cp__a("tmc/tmc_worker.F",616)
      IF(.NOT.(ASSOCIATED(conf)))CALL cp__a("tmc/tmc_worker.F",617)
      IF(.NOT.(conf%temp_created .GT. 0))CALL cp__a("tmc/tmc_worker.F",618)
      IF(.NOT.(conf%temp_created .LE. tmc_env%params%nr_temp))CALL cp__a("tmc/tmc_worker.F",619)
      IF(.NOT.(env_id .GT. 0))CALL cp__a("tmc/tmc_worker.F",620)

      ! start the timing
      CALL timeset(routineN, handle)

      CALL allocate_new_sub_tree_node(tmc_params=tmc_env%params, &
                                      next_el=last_acc_conf, nr_dim=SIZE(conf%pos))

      last_acc_conf%pos = conf%pos
      last_acc_conf%box_scale = conf%box_scale

      ! energy of the last accepted configuration
      CALL calc_potential_energy(conf=last_acc_conf, &
                                 env_id=tmc_env%w_env%env_id_approx, exact_approx_pot=.FALSE., &
                                 tmc_env=tmc_env)

      NMC_steps: DO substeps = 1, INT(tmc_env%params%move_types%mv_size(mv_type_NMC_moves, 1))
         ! check for canceling message
         IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_w)) THEN
            flag = .FALSE.
            comm_dest = MASTER_COMM_ID
            ! check for new canceling message
            CALL tmc_message(msg_type=calc_status, send_recv=recv_msg, &
                             dest=comm_dest, &
                             para_env=tmc_env%tmc_comp_set%para_env_m_w, &
                             tmc_params=tmc_env%params, success=flag)
         END IF
         comm_dest = bcast_group
         CALL tmc_message(msg_type=calc_status, send_recv=send_msg, &
                          dest=comm_dest, &
                          para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                          tmc_params=tmc_env%params)
         SELECT CASE (calc_status)
         CASE (TMC_STATUS_CALCULATING)
            ! keep on working
         CASE (TMC_CANCELING_MESSAGE)
            ! nothing to do, because calculation CANCELING, exit with cancel status
            EXIT NMC_steps
         CASE DEFAULT
            CALL cp_abort(cp__l("tmc/tmc_worker.F",659), &
                          "unknown status "//cp_to_string(calc_status)// &
                          "in the NMC routine, expect only caneling status. ")
         END SELECT

         ! set move type
         CALL set_rng_stream(rng_stream=tmc_env%rng_stream, &
                             bg=conf%rng_seed(:, :, 1), cg=conf%rng_seed(:, :, 2), &
                             ig=conf%rng_seed(:, :, 3))
         conf%move_type = select_random_move_type( &
                          move_types=tmc_env%params%nmc_move_types, &
                          rnd=next_random_number(tmc_env%rng_stream))
         CALL get_rng_stream(rng_stream=tmc_env%rng_stream, &
                             bg=conf%rng_seed(:, :, 1), cg=conf%rng_seed(:, :, 2), &
                             ig=conf%rng_seed(:, :, 3))

         ! do move
         CALL change_pos(tmc_params=tmc_env%params, &
                         move_types=tmc_env%params%nmc_move_types, &
                         rng_stream=tmc_env%rng_stream, &
                         elem=conf, mv_conf=1, new_subbox=.FALSE., &
                         move_rejected=change_rejected)
         ! for Hybrid MC the change_pos is only velocity change,
         !   the actual MD step hast to be done in this module for communication reason
         IF (conf%move_type .EQ. mv_type_MD) THEN
            !TODO implement the MD part
            !CALL calc_MD_step(...)
            !CALL calc_calc_e_kin(...)
            CALL cp_abort(cp__l("tmc/tmc_worker.F",687), &
                          "Hybrid MC is not implemented yet, "// &
                          "(no MD section in TMC yet). ")
         END IF

         ! update the subbox acceptance probabilities
         CALL prob_update(move_types=tmc_env%params%nmc_move_types, elem=conf, &
                          acc=.NOT. change_rejected, subbox=.TRUE., &
                          prob_opt=tmc_env%params%esimate_acc_prob)

         ! calculate potential energy if neccessary
         IF (.NOT. change_rejected) THEN
            CALL calc_potential_energy(conf=conf, &
                                       env_id=tmc_env%w_env%env_id_approx, exact_approx_pot=.FALSE., &
                                       tmc_env=tmc_env)
         ELSE
            conf%e_pot_approx = HUGE(conf%e_pot_approx)
         END IF

         !check NMC step
         CALL set_rng_stream(rng_stream=tmc_env%rng_stream, &
                             bg=conf%rng_seed(:, :, 1), cg=conf%rng_seed(:, :, 2), &
                             ig=conf%rng_seed(:, :, 3))
         rnd_nr = next_random_number(tmc_env%rng_stream)
         CALL get_rng_stream(rng_stream=tmc_env%rng_stream, &
                             bg=conf%rng_seed(:, :, 1), cg=conf%rng_seed(:, :, 2), &
                             ig=conf%rng_seed(:, :, 3))

         IF (.NOT. change_rejected) THEN
            CALL acceptance_check(tree_element=conf, parent_element=last_acc_conf, &
                                  tmc_params=tmc_env%params, &
                                  temperature=tmc_env%params%Temp(conf%temp_created), &
                                  diff_pot_check=.FALSE., &
                                  accept=accept, approx_ener=.TRUE., rnd_nr=rnd_nr)
         ELSE
            accept = .FALSE.
         END IF
         ! update the NMC accpetance per move
         CALL prob_update(move_types=tmc_env%params%nmc_move_types, elem=conf, &
                          acc=accept, prob_opt=tmc_env%params%esimate_acc_prob)

         ! update last accepted configuration or actual configuration
         IF (accept .AND. (.NOT. change_rejected)) THEN
            last_acc_conf%pos = conf%pos
            last_acc_conf%vel = conf%vel
            last_acc_conf%e_pot_approx = conf%e_pot_approx
            last_acc_conf%ekin = conf%ekin
            last_acc_conf%ekin_before_md = conf%ekin_before_md
            last_acc_conf%box_scale = conf%box_scale
         ELSE
            conf%pos = last_acc_conf%pos
            conf%vel = last_acc_conf%vel
            conf%box_scale = last_acc_conf%box_scale
         END IF
      END DO NMC_steps

      ! result values of Nested Monte Carlo (NMC) steps
      !   regard that the calculated potential energy is the one of the approximated potential
      conf%pos = last_acc_conf%pos
      conf%vel = last_acc_conf%vel
      conf%e_pot_approx = last_acc_conf%e_pot_approx
      conf%potential = 0.0_dp
      conf%ekin = last_acc_conf%ekin
      conf%ekin_before_md = last_acc_conf%ekin_before_md

      CALL deallocate_sub_tree_node(tree_elem=last_acc_conf)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE nested_markov_chain_MC

! **************************************************************************************************
!> \brief get the initial confuguration (pos,...)
!> \param tmc_params ...
!> \param init_conf the structure the data should be stored
!> force_env
!> \param env_id ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE get_initial_conf(tmc_params, init_conf, env_id)
      TYPE(tmc_param_type), POINTER                      :: tmc_params
      TYPE(tree_type), POINTER                           :: init_conf
      INTEGER                                            :: env_id

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_initial_conf', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, mol, ndim, nr_atoms
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(mol_new_list_type), POINTER                   :: molecule_new

      IF(.NOT.(.NOT. ASSOCIATED(init_conf)))CALL cp__a("tmc/tmc_worker.F",779)

      ! start the timing
      CALL timeset(routineN, handle)

      ! get positions
      CALL get_natom(env_id=env_id, n_atom=nr_atoms, ierr=ierr)
      IF(.NOT.(ierr .EQ. 0))CALL cp__a("tmc/tmc_worker.F",786)
      ndim = 3*nr_atoms
      CALL allocate_new_sub_tree_node(tmc_params=tmc_params, &
                                      next_el=init_conf, nr_dim=ndim)
      CALL get_pos(env_id=env_id, pos=init_conf%pos, n_el=SIZE(init_conf%pos), &
                   ierr=ierr)

      ! get the molecule info
      CALL f_env_get_from_id(env_id, f_env)
      CALL force_env_get(f_env%force_env, subsys=subsys)

      CALL cp_subsys_get(subsys=subsys, molecules_new=molecule_new)
      loop_mol: DO mol = 1, SIZE(molecule_new%els(:))
         init_conf%mol(molecule_new%els(mol)%first_atom: &
                       molecule_new%els(mol)%last_atom) = mol
      END DO loop_mol

      ! end the timing
      CALL timestop(handle)

   END SUBROUTINE get_initial_conf

! **************************************************************************************************
!> \brief get the pointer to the atoms, for easy handling
!> \param env_id ...
!> \param atoms pointer to atomic_kind
!> \param cell ...
!> \author Mandes 01.2013
! **************************************************************************************************
   SUBROUTINE get_atom_kinds_and_cell(env_id, atoms, cell)
      INTEGER                                            :: env_id
      TYPE(tmc_atom_type), DIMENSION(:), POINTER         :: atoms
      TYPE(cell_type), POINTER                           :: cell

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_atom_kinds_and_cell', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iparticle, nr_atoms, nunits_tot
      TYPE(cell_type), POINTER                           :: cell_tmp
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(particle_list_type), POINTER                  :: particles

      NULLIFY (f_env, subsys, particles)
      nr_atoms = 0

      IF(.NOT.(env_id .GT. 0))CALL cp__a("tmc/tmc_worker.F",832)
      IF(.NOT.(.NOT. ASSOCIATED(atoms)))CALL cp__a("tmc/tmc_worker.F",833)
      IF(.NOT.(.NOT. ASSOCIATED(cell)))CALL cp__a("tmc/tmc_worker.F",834)

      CALL f_env_get_from_id(env_id, f_env)
      nr_atoms = force_env_get_natom(f_env%force_env)
      CALL force_env_get(f_env%force_env, subsys=subsys, cell=cell_tmp)
      ALLOCATE (cell)
      CALL cell_copy(cell_in=cell_tmp, cell_out=cell)

      !get atom kinds
      CALL allocate_tmc_atom_type(atoms, nr_atoms)
      CALL cp_subsys_get(subsys, particles=particles)
      nunits_tot = SIZE(particles%els(:))
      IF (nunits_tot .GT. 0) THEN
         DO iparticle = 1, nunits_tot
            atoms(iparticle)%name = particles%els(iparticle)%atomic_kind%name
            atoms(iparticle)%mass = particles%els(iparticle)%atomic_kind%mass
         END DO
         IF(.NOT.(iparticle-1 .EQ. nr_atoms))CALL cp__a("tmc/tmc_worker.F",851)
      ENDIF
   END SUBROUTINE get_atom_kinds_and_cell

! **************************************************************************************************
!> \brief set the communicator in the SCF environment
!>        to receive the intermediate energies on the (global) master side
!> \param comm the master-worker communicator
!> \param env_id the ID of the related force environment
!> \author Mandes 10.2013
! **************************************************************************************************
   SUBROUTINE set_intermediate_info_comm(comm, env_id)
      INTEGER, INTENT(IN)                                :: comm
      INTEGER                                            :: env_id

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_intermediate_info_comm', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description
      REAL(KIND=dp), DIMENSION(3)                        :: values
      TYPE(cp_result_type), POINTER                      :: results
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (results, subsys)
      IF(.NOT.(env_id .GT. 0))CALL cp__a("tmc/tmc_worker.F",876)

      CALL f_env_get_from_id(env_id, f_env)

      IF(.NOT.(ASSOCIATED(f_env)))CALL cp__a("tmc/tmc_worker.F",880)
      IF(.NOT.(ASSOCIATED(f_env%force_env)))CALL cp__a("tmc/tmc_worker.F",881)
      IF (.NOT. ASSOCIATED(f_env%force_env%qs_env)) &
         CALL cp_abort(cp__l("tmc/tmc_worker.F",883), &
                       "the intermediate SCF energy request can not be set "// &
                       "employing this force environment! ")

      ! set the information
      values(1) = REAL(comm, KIND=dp)
      values(2) = REAL(MASTER_COMM_ID, KIND=dp)
      values(3) = REAL(TMC_STAT_SCF_STEP_ENER_RECEIVE, KIND=dp)
      description = "[EXT_SCF_ENER_COMM]"

      ! set the communicator information in the qs_env result container
      CALL force_env_get(f_env%force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, results=results)
      CALL put_results(results, description=description, values=values)
   END SUBROUTINE set_intermediate_info_comm

! **************************************************************************************************
!> \brief set the communicator in the SCF environment
!>        to receive the intermediate energies on the (global) master side
!> \param env_id the ID of the related force environment
!> \author Mandes 10.2013
! **************************************************************************************************
   SUBROUTINE remove_intermediate_info_comm(env_id)
      INTEGER                                            :: env_id

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove_intermediate_info_comm', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: description
      TYPE(cp_result_type), POINTER                      :: results
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (subsys, results)
      IF(.NOT.(env_id .GT. 0))CALL cp__a("tmc/tmc_worker.F",917)

      CALL f_env_get_from_id(env_id, f_env)

      IF(.NOT.(ASSOCIATED(f_env)))CALL cp__a("tmc/tmc_worker.F",921)
      IF(.NOT.(ASSOCIATED(f_env%force_env)))CALL cp__a("tmc/tmc_worker.F",922)
      IF (.NOT. ASSOCIATED(f_env%force_env%qs_env)) &
         CALL cp_abort(cp__l("tmc/tmc_worker.F",924), &
                       "the SCF intermediate energy communicator can not be "// &
                       "removed! ")

      description = "[EXT_SCF_ENER_COMM]"

      ! set the communicator information in the qs_env result container
      CALL force_env_get(f_env%force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, results=results)
      CALL cp_results_erase(results, description=description)
   END SUBROUTINE remove_intermediate_info_comm

END MODULE tmc_worker
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Tree Monte Carlo entry point, set up, CPU redistribution and
!>      input reading
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_setup
   USE bibliography,                    ONLY: cite_reference,&
                                              schonherr2014
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: &
        cp_add_default_logger, cp_get_default_logger, cp_logger_create, &
        cp_logger_get_default_io_unit, cp_logger_release, cp_logger_set, cp_logger_type, &
        cp_rm_default_logger, cp_to_string
   USE cp_para_env,                     ONLY: cp_para_env_create
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE environment,                     ONLY: cp2k_get_walltime
   USE f77_interface,                   ONLY: create_force_env,&
                                              destroy_force_env
   USE global_types,                    ONLY: global_environment_type
   USE header,                          ONLY: tmc_ana_header,&
                                              tmc_header
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: default_output_unit,&
                                              m_flush
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_comm_dup,&
                                              mp_comm_free,&
                                              mp_comm_split_direct
   USE parallel_rng_types,              ONLY: UNIFORM,&
                                              create_rng_stream,&
                                              delete_rng_stream
   USE physcon,                         ONLY: au2a => angstrom,&
                                              au2bar => bar
   USE tmc_analysis,                    ONLY: analysis_init,&
                                              analysis_restart_print,&
                                              analysis_restart_read,&
                                              analyze_file_configurations,&
                                              finalize_tmc_analysis,&
                                              tmc_read_ana_input
   USE tmc_analysis_types,              ONLY: tmc_ana_env_release,&
                                              tmc_ana_list_type
   USE tmc_file_io,                     ONLY: expand_file_name_int
   USE tmc_master,                      ONLY: do_tmc_master
   USE tmc_move_handle,                 ONLY: finalize_mv_types,&
                                              print_move_types,&
                                              read_init_move_types
   USE tmc_stati,                       ONLY: &
        task_type_MC, task_type_ideal_gas, tmc_NMC_worker_out_file_name, tmc_ana_out_file_name, &
        tmc_default_restart_in_file_name, tmc_default_restart_out_file_name, &
        tmc_default_unspecified_name, tmc_energy_worker_out_file_name, tmc_master_out_file_name
   USE tmc_tree_build,                  ONLY: allocate_new_sub_tree_node,&
                                              deallocate_sub_tree_node
   USE tmc_tree_types,                  ONLY: tree_type
   USE tmc_types,                       ONLY: tmc_comp_set_type,&
                                              tmc_env_create,&
                                              tmc_env_release,&
                                              tmc_env_type,&
                                              tmc_master_env_create,&
                                              tmc_master_env_release,&
                                              tmc_worker_env_create,&
                                              tmc_worker_env_release
   USE tmc_worker,                      ONLY: do_tmc_worker,&
                                              get_atom_kinds_and_cell,&
                                              get_initial_conf
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_setup'

   PUBLIC  :: do_tmc, do_analyze_files

CONTAINS

! **************************************************************************************************
!> \brief tmc_entry point
!> \param input_declaration ...
!> \param root_section ...
!> \param para_env ...
!> \param globenv the global environment for the simulation
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE do_tmc(input_declaration, root_section, para_env, globenv)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_tmc', routineP = moduleN//':'//routineN

      INTEGER                                            :: bcast_output_unit, handle, i, ierr, &
                                                            output_unit
      LOGICAL                                            :: success
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: init_rng_seed
      TYPE(cp_logger_type), POINTER                      :: logger, logger_sub
      TYPE(section_vals_type), POINTER                   :: tmc_ana_section
      TYPE(tmc_ana_list_type), DIMENSION(:), POINTER     :: tmc_ana_env_list
      TYPE(tmc_env_type), POINTER                        :: tmc_env

! start the timing

      CALL timeset(routineN, handle)

      CALL cite_reference(Schonherr2014)

      NULLIFY (logger, logger_sub, tmc_env, tmc_ana_env_list)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      ! write header, on the 'rank 0' of the global communicator
      IF (output_unit > 0) THEN
         CALL tmc_header(output_unit)
         CALL m_flush(output_unit)
      ENDIF
      ! ugly, we need to know the output unit on source, everywhere, in particular
      ! the tmc master
      IF (output_unit .NE. default_output_unit .AND. output_unit .GT. 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") "The TMC output files are:"
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            TRIM(tmc_master_out_file_name)//"           the TMC master"
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            TRIM(tmc_energy_worker_out_file_name)//"         the worker outputs (energy calculations etc.)"
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            TRIM(tmc_ana_out_file_name)//"              the analysis output"
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") REPEAT("-", 79)
      END IF
      bcast_output_unit = output_unit
      CALL mp_bcast(bcast_output_unit, para_env%source, para_env%group)

      ! create tmc_env
      CALL tmc_env_create(tmc_env)
      CALL tmc_preread_input(root_section, tmc_env)
      CALL tmc_redistributing_cores(tmc_env%tmc_comp_set, para_env, &
                                    ana_on_the_fly=tmc_env%tmc_comp_set%ana_on_the_fly, &
                                    success=success)

      IF (success) THEN
         ! initialize master and worker environment
         IF (tmc_env%tmc_comp_set%group_nr .EQ. 0) THEN
            CALL tmc_master_env_create(tmc_env) ! create master env
         ELSE IF (tmc_env%tmc_comp_set%group_nr .NE. 0) THEN
            CALL tmc_worker_env_create(tmc_env) ! create worker env
         END IF

         CALL tmc_read_input(root_section, tmc_env)
         !CALL init_move_types(tmc_params=tmc_env%params)

         ! init random number generator: use determistic random numbers
         IF (tmc_env%tmc_comp_set%group_nr .EQ. 0 .AND. &
             tmc_env%m_env%rnd_init .GT. 0) THEN
            ALLOCATE (init_rng_seed(3, 2))
            init_rng_seed(:, :) = &
               RESHAPE((/tmc_env%m_env%rnd_init*42.0_dp, &
                         tmc_env%m_env%rnd_init*54.0_dp, &
                         tmc_env%m_env%rnd_init*63.0_dp, &
                         tmc_env%m_env%rnd_init*98.0_dp, &
                         tmc_env%m_env%rnd_init*10.0_dp, &
                         tmc_env%m_env%rnd_init*2.0_dp/), &
                       (/3, 2/))
            CALL create_rng_stream(rng_stream=tmc_env%rng_stream, &
                                   name="TMC_deterministic_rng_stream", &
                                   seed=init_rng_seed(:, :), &
                                   distribution_type=UNIFORM)
            DEALLOCATE (init_rng_seed)
         ELSE
            CALL create_rng_stream(rng_stream=tmc_env%rng_stream, &
                                   name="TMC_rng_stream", &
                                   distribution_type=UNIFORM)
         END IF

         ! start running master and worker routines
         ! the master
         IF (tmc_env%tmc_comp_set%group_nr .EQ. 0) THEN
            !TODO get the correct usage of creating and handling the logger...
            CALL cp_logger_create(logger_sub, para_env=tmc_env%tmc_comp_set%para_env_m_only, &
                                  default_global_unit_nr=default_output_unit, close_global_unit_on_dealloc=.FALSE.)
            CALL cp_logger_set(logger_sub, local_filename="tmc_main")
            CALL cp_add_default_logger(logger_sub)

            ! if we're doing output to the screen, keep it there, else this master
            ! opens a file (not that two different ranks are writing to the
            ! default_output_unit, we leave it up to mpirun or so to merge stuff
            IF (bcast_output_unit == default_output_unit) THEN
               tmc_env%m_env%io_unit = default_output_unit
            ELSE
               CALL open_file(file_name=tmc_master_out_file_name, file_status="UNKNOWN", &
                              file_action="WRITE", file_position="APPEND", &
                              unit_number=tmc_env%m_env%io_unit)
               CALL tmc_header(tmc_env%m_env%io_unit)
            ENDIF
            ! print the intresting parameters and starting values
            CALL tmc_print_params(tmc_env)
            CALL print_move_types(init=.TRUE., file_io=tmc_env%m_env%io_unit, &
                                  tmc_params=tmc_env%params)
            CALL do_tmc_master(tmc_env=tmc_env, globenv=globenv) ! start the master routine

            IF (bcast_output_unit .NE. tmc_env%m_env%io_unit) THEN
               CALL close_file(unit_number=tmc_env%m_env%io_unit)
            ENDIF

            CALL cp_rm_default_logger()
            CALL cp_logger_release(logger_sub)

            ! the worker groups
         ELSE IF (tmc_env%tmc_comp_set%group_nr .GT. 0) THEN
            NULLIFY (logger_sub)
            ! create separate logger and error handler for each worker
            CALL cp_logger_create(logger_sub, para_env=tmc_env%tmc_comp_set%para_env_sub_group, &
                                  default_global_unit_nr=default_output_unit, close_global_unit_on_dealloc=.FALSE.)
            CALL cp_logger_set(logger_sub, local_filename="tmc_localLog")
            CALL cp_add_default_logger(logger_sub)
            tmc_env%w_env%io_unit = default_output_unit

            ! energy worker
            IF (tmc_env%tmc_comp_set%group_nr .LE. tmc_env%tmc_comp_set%group_ener_nr) THEN
               CALL create_force_env(new_env_id=tmc_env%w_env%env_id_ener, &
                                     input_declaration=input_declaration, &
                                     input_path=tmc_env%params%energy_inp_file, &
                                     mpi_comm=tmc_env%tmc_comp_set%para_env_sub_group%group, &
                                     output_path=TRIM(expand_file_name_int(file_name=tmc_energy_worker_out_file_name, &
                                                                           ivalue=tmc_env%tmc_comp_set%group_nr)), &
                                     ierr=ierr)
               IF (ierr .NE. 0) &
                  CALL cp__b("tmc/tmc_setup.F",243,"creating force env result in error "//cp_to_string(ierr))
            END IF
            ! worker for configurational change
            IF (tmc_env%params%NMC_inp_file .NE. "" .AND. &
                (tmc_env%tmc_comp_set%group_cc_nr .EQ. 0 .OR. &
                 tmc_env%tmc_comp_set%group_nr .GT. tmc_env%tmc_comp_set%group_ener_nr)) THEN
               CALL create_force_env(new_env_id=tmc_env%w_env%env_id_approx, &
                                     input_declaration=input_declaration, &
                                     input_path=tmc_env%params%NMC_inp_file, &
                                     mpi_comm=tmc_env%tmc_comp_set%para_env_sub_group%group, &
                                     output_path=TRIM(expand_file_name_int(file_name=tmc_NMC_worker_out_file_name, &
                                                                           ivalue=tmc_env%tmc_comp_set%group_nr)), &
                                     ierr=ierr)
               IF (ierr .NE. 0) &
                  CALL cp__b("tmc/tmc_setup.F",257,"creating approx force env result in error "//cp_to_string(ierr))
            END IF
            CALL do_tmc_worker(tmc_env=tmc_env) ! start the worker routine

            IF (tmc_env%w_env%env_id_ener .GT. 0) &
               CALL destroy_force_env(tmc_env%w_env%env_id_ener, ierr)
            IF (tmc_env%w_env%env_id_approx .GT. 0) &
               CALL destroy_force_env(tmc_env%w_env%env_id_approx, ierr)

            CALL cp_rm_default_logger()
            CALL cp_logger_release(logger_sub)

            ! the analysis group
         ELSE IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_ana)) THEN
            ! unused worker groups can do analysis
            NULLIFY (logger_sub)
            ! create separate logger and error handler for each worker
            CALL cp_logger_create(logger_sub, para_env=tmc_env%tmc_comp_set%para_env_m_ana, &
                                  default_global_unit_nr=default_output_unit, close_global_unit_on_dealloc=.FALSE.)
            tmc_env%w_env%io_unit = default_output_unit
            CALL cp_logger_set(logger_sub, local_filename="tmc_ana_localLog")
            CALL cp_add_default_logger(logger_sub)
            ! if we're doing output to the screen, keep it there, else this master
            ! opens a file (not that two different ranks are writing to the
            ! default_output_unit, we leave it up to mpirun or so to merge stuff
            IF (bcast_output_unit == default_output_unit) THEN
               output_unit = default_output_unit
            ELSE
               CALL open_file(file_name=tmc_ana_out_file_name, file_status="UNKNOWN", &
                              file_action="WRITE", file_position="APPEND", &
                              unit_number=output_unit)
               CALL tmc_ana_header(output_unit)
            ENDIF

            ALLOCATE (tmc_ana_env_list(tmc_env%params%nr_temp))
            tmc_ana_section => section_vals_get_subs_vals(root_section, "MOTION%TMC%TMC_ANALYSIS")
            DO i = 1, tmc_env%params%nr_temp
               CALL tmc_read_ana_input(tmc_ana_section, tmc_ana_env_list(i)%temp)
               tmc_ana_env_list(i)%temp%io_unit = output_unit
            END DO
            CALL do_tmc_worker(tmc_env=tmc_env, ana_list=tmc_ana_env_list) ! start the worker routine for analysis
            DO i = 1, tmc_env%params%nr_temp
               IF (ASSOCIATED(tmc_ana_env_list(i)%temp%last_elem)) &
                  CALL deallocate_sub_tree_node(tree_elem=tmc_ana_env_list(i)%temp%last_elem)
               CALL tmc_ana_env_release(tmc_ana_env_list(i)%temp)
            END DO
            DEALLOCATE (tmc_ana_env_list)
            IF (bcast_output_unit .NE. output_unit) THEN
               CALL close_file(unit_number=tmc_env%m_env%io_unit)
            ENDIF
            CALL cp_rm_default_logger()
            CALL cp_logger_release(logger_sub)

         END IF ! unused worker groups have nothing to do

         ! delete the random numbers
         IF(.NOT.(ASSOCIATED(tmc_env%rng_stream)))CALL cp__a("tmc/tmc_setup.F",313)
         CALL delete_rng_stream(tmc_env%rng_stream)

         ! deallocate the move types
         CALL finalize_mv_types(tmc_env%params)

         ! finalize master and worker environment
         IF (tmc_env%tmc_comp_set%group_nr .EQ. 0) THEN
            CALL tmc_master_env_release(tmc_env) ! release master env
         ELSE IF (tmc_env%tmc_comp_set%group_nr .NE. 0) THEN
            CALL tmc_worker_env_release(tmc_env) ! release worker env
         END IF ! unused worker groups have nothing to do

      ELSE
         IF (tmc_env%params%print_test_output) THEN
            WRITE (output_unit, *) "TMC|NOTenoughProcessorsX= -999"
            WRITE (output_unit, *) "TMC|NOTcalculatedTotal energy: -999"
         END IF
      END IF
      ! finalize / deallocate everthing
      CALL tmc_env_release(tmc_env)

      ! end the timing
      CALL timestop(handle)

   END SUBROUTINE do_tmc

! **************************************************************************************************
!> \brief analyze TMC trajectory files
!> \param input_declaration ...
!> \param root_section ...
!> \param para_env ...
!> \param
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE do_analyze_files(input_declaration, root_section, para_env)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_analyze_files', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: comm, dir_ind, handle, my_mpi_world, &
                                                            nr_dim, output_unit, temp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(tmc_ana_list_type), DIMENSION(:), POINTER     :: ana_list
      TYPE(tmc_env_type), POINTER                        :: tmc_env
      TYPE(tree_type), POINTER                           :: elem

      NULLIFY (ana_list, tmc_env, elem, logger)

      ! start the timing
      CALL timeset(routineN, handle)

      ! create a TMC environment (also to have a params environment)
      CALL tmc_env_create(tmc_env)
      ! duplicate communicator
      CALL mp_comm_dup(para_env%group, my_mpi_world)
      ! -- spiltting communicators
      CALL mp_comm_split_direct(my_mpi_world, comm, para_env%mepos, 0)
      CALL cp_para_env_create(para_env=tmc_env%tmc_comp_set%para_env_m_ana, &
                              group=comm)
      IF (para_env%num_pe .NE. 1) &
         CALL cp__w("tmc/tmc_setup.F",377,"just one out of "//cp_to_string(para_env%num_pe)//"cores is used ")
      ! distribute work to availuble cores
      IF (para_env%mepos .EQ. 0) THEN
         !TODO get the correct usage of creating and handling the logger...
         logger => cp_get_default_logger()
         output_unit = cp_logger_get_default_io_unit(logger)
         IF(.NOT.(output_unit .GT. 0))CALL cp__a("tmc/tmc_setup.F",383)
         ! write the header
         CALL tmc_ana_header(output_unit)

         ! read the input and create the ana environments for each temp
         CALL tmc_read_ana_files_input(input_declaration=input_declaration, &
                                       input=root_section, ana_list=ana_list, &
                                       elem=elem, tmc_env=tmc_env)
         nr_dim = SIZE(elem%pos)
         ! we need a new tree element with all neccessay arrays, (e.g. dipoles could not be allocated already)
         CALL deallocate_sub_tree_node(tree_elem=elem)
         IF(.NOT.(SIZE(ana_list) .GT. 0))CALL cp__a("tmc/tmc_setup.F",394)

         ! print initial test output (for single core tests, where no data is produced)
         IF (tmc_env%params%print_test_output) THEN
            WRITE (output_unit, *) "TMC|ANAtestOutputInitX= -999"
         END IF

         ! do the analysis
         DO temp = 1, SIZE(ana_list)
            ! initialize the structures
            ana_list(temp)%temp%io_unit = output_unit
            CALL analysis_init(ana_env=ana_list(temp)%temp, nr_dim=nr_dim)
            ! to allocate the dipole array in tree elements
            IF (ana_list(temp)%temp%costum_dip_file_name .NE. &
                tmc_default_unspecified_name) &
               tmc_env%params%print_dipole = .TRUE.

            IF (.NOT. ASSOCIATED(elem)) &
               CALL allocate_new_sub_tree_node(tmc_params=tmc_env%params, &
                                               next_el=elem, nr_dim=nr_dim)
            CALL analysis_restart_read(ana_env=ana_list(temp)%temp, &
                                       elem=elem)
            IF (.NOT. ASSOCIATED(elem) .AND. .NOT. ASSOCIATED(ana_list(temp)%temp%last_elem)) &
               CALL cp__b("tmc/tmc_setup.F",417,"uncorrect initialization of the initial configuration")
            ! do for all directories
            DO dir_ind = 1, SIZE(ana_list(temp)%temp%dirs)
               WRITE (output_unit, FMT='(T2,A,"| ",A,T41,A40)') "TMC_ANA", &
                  "read directory", TRIM(ana_list(temp)%temp%dirs(dir_ind))
               CALL analyze_file_configurations( &
                  start_id=ana_list(temp)%temp%from_elem, &
                  end_id=ana_list(temp)%temp%to_elem, &
                  dir_ind=dir_ind, &
                  ana_env=ana_list(temp)%temp, &
                  tmc_params=tmc_env%params)
               ! remove the last saved element to start with a new file
               !  there is no weight for this element
               IF (dir_ind .LT. SIZE(ana_list(temp)%temp%dirs) .AND. &
                   ASSOCIATED(ana_list(temp)%temp%last_elem)) &
                  CALL deallocate_sub_tree_node(tree_elem=ana_list(temp)%temp%last_elem)
               IF (ASSOCIATED(ana_list(temp)%temp%last_elem)) &
                  ana_list(temp)%temp%conf_offset = ana_list(temp)%temp%conf_offset &
                                                    +ana_list(temp)%temp%last_elem%nr
            END DO
            CALL finalize_tmc_analysis(ana_env=ana_list(temp)%temp)
            ! write analysis restart file
            !  if there is somthing to write
            ! shifts the last element to actual element
            IF (ASSOCIATED(ana_list(temp)%temp%last_elem)) &
               CALL analysis_restart_print(ana_env=ana_list(temp)%temp)
            IF (ASSOCIATED(ana_list(temp)%temp%last_elem)) &
               CALL deallocate_sub_tree_node(tree_elem=ana_list(temp)%temp%last_elem)
            IF (ASSOCIATED(elem)) &
               CALL deallocate_sub_tree_node(tree_elem=elem)

            IF (ASSOCIATED(ana_list(temp)%temp%last_elem)) &
               CALL deallocate_sub_tree_node(tree_elem=ana_list(temp)%temp%last_elem)

            CALL tmc_ana_env_release(ana_list(temp)%temp)
         END DO

         DEALLOCATE (ana_list)
      END IF
      CALL mp_comm_free(my_mpi_world)
      CALL tmc_env_release(tmc_env)

      ! end the timing
      CALL timestop(handle)
   END SUBROUTINE do_analyze_files

! **************************************************************************************************
!> \brief creates a new para environment for tmc analysis for each temperature
!> \param input_declaration ...
!> \param input global environment
!> \param ana_list ...
!> \param elem ...
!> \param tmc_env TMC analysis environment
!> \author Mandes 03.2013
! **************************************************************************************************
   SUBROUTINE tmc_read_ana_files_input(input_declaration, input, ana_list, elem, tmc_env)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(tmc_ana_list_type), DIMENSION(:), POINTER     :: ana_list
      TYPE(tree_type), POINTER                           :: elem
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'tmc_read_ana_files_input', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: directories
      INTEGER                                            :: env_id, ierr, nr_temp, t_act
      LOGICAL                                            :: flag
      REAL(KIND=dp)                                      :: tmax, tmin
      REAL(KIND=dp), DIMENSION(:), POINTER               :: inp_Temp, Temps
      TYPE(section_vals_type), POINTER                   :: tmc_section

      NULLIFY (tmc_section, inp_Temp, Temps)
      IF(.NOT.(ASSOCIATED(input)))CALL cp__a("tmc/tmc_setup.F",491)
      IF(.NOT.(.NOT. ASSOCIATED(ana_list)))CALL cp__a("tmc/tmc_setup.F",492)
      IF(.NOT.(.NOT. ASSOCIATED(elem)))CALL cp__a("tmc/tmc_setup.F",493)
      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_setup.F",494)

      ! first global TMC stuff
      tmc_section => section_vals_get_subs_vals(input, "MOTION%TMC")
      CALL section_vals_val_get(tmc_section, "PRINT_TEST_OUTPUT", l_val=tmc_env%params%print_test_output)
      ! TMC analysis stuff
      tmc_section => section_vals_get_subs_vals(input, "MOTION%TMC%TMC_ANALYSIS_FILES")
      CALL section_vals_get(tmc_section, explicit=flag)
      IF(.NOT.(flag))CALL cp__a("tmc/tmc_setup.F",502)

      CALL section_vals_val_get(tmc_section, "FORCE_ENV_FILE", &
                                c_val=tmc_env%params%energy_inp_file)

      CALL section_vals_val_get(tmc_section, "NR_TEMPERATURE", i_val=nr_temp)

      CALL section_vals_val_get(tmc_section, "TEMPERATURE", r_vals=inp_Temp)
      IF ((nr_temp .GT. 1) .AND. (SIZE(inp_Temp) .NE. 2)) &
         CALL cp__b("tmc/tmc_setup.F",511,"specify each temperature, skip keyword NR_TEMPERATURE")
      IF (nr_temp .EQ. 1) THEN
         nr_temp = SIZE(inp_Temp)
         ALLOCATE (Temps(nr_temp))
         Temps(:) = inp_Temp(:)
      ELSE
         tmin = inp_Temp(1)
         tmax = inp_Temp(2)
         ALLOCATE (Temps(nr_temp))
         Temps(1) = tmin
         DO t_act = 2, SIZE(Temps)
            Temps(t_act) = Temps(t_act-1)+(tmax-tmin)/(SIZE(Temps)-1.0_dp)
         END DO
         IF (ANY(Temps .LT. 0.0_dp)) &
            CALL cp_abort(cp__l("tmc/tmc_setup.F",525), "The temperatures are negative. Should be specified using "// &
                          "TEMPERATURE {T_min} {T_max} and NR_TEMPERATURE {#temperatures}")
      END IF

      ! get multiple directories
      CALL section_vals_val_get(tmc_section, "DIRECTORIES", c_vals=directories)

      ! get init configuration (for sizes)
      CALL create_force_env(new_env_id=env_id, &
                            input_declaration=input_declaration, &
                            input_path=tmc_env%params%energy_inp_file, &
                            mpi_comm=tmc_env%tmc_comp_set%para_env_m_ana%group, &
                            output_path="tmc_ana.out", ierr=ierr)
      CALL get_initial_conf(tmc_params=tmc_env%params, init_conf=elem, &
                            env_id=env_id)
      CALL get_atom_kinds_and_cell(env_id=env_id, atoms=tmc_env%params%atoms, &
                                   cell=tmc_env%params%cell)
      CALL destroy_force_env(env_id, ierr)

      ALLOCATE (ana_list(SIZE(Temps)))
      DO t_act = 1, SIZE(Temps)
         ana_list(t_act)%temp => NULL()
         CALL tmc_read_ana_input(tmc_section, ana_list(t_act)%temp)
         ana_list(t_act)%temp%temperature = Temps(t_act)
         ALLOCATE (ana_list(t_act)%temp%dirs(SIZE(directories)))
         ana_list(t_act)%temp%dirs(:) = directories(:)
         ana_list(t_act)%temp%cell => tmc_env%params%cell
         ana_list(t_act)%temp%atoms => tmc_env%params%atoms
         ana_list(t_act)%temp%print_test_output = tmc_env%params%print_test_output

         CALL section_vals_val_get(tmc_section, "POSITION_FILE", &
                                   c_val=ana_list(t_act)%temp%costum_pos_file_name)
         CALL section_vals_val_get(tmc_section, "DIPOLE_FILE", &
                                   c_val=ana_list(t_act)%temp%costum_dip_file_name)
         CALL section_vals_val_get(tmc_section, "CELL_FILE", &
                                   c_val=ana_list(t_act)%temp%costum_cell_file_name)
         CALL section_vals_val_get(tmc_section, "START_ELEM", i_val=ana_list(t_act)%temp%from_elem)
         CALL section_vals_val_get(tmc_section, "END_ELEM", i_val=ana_list(t_act)%temp%to_elem)
      END DO
      DEALLOCATE (Temps)
   END SUBROUTINE tmc_read_ana_files_input

! **************************************************************************************************
!> \brief read the variables for distributing cores
!> \param input ...
!> \param tmc_env structure for storing all the tmc parameters
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_preread_input(input, tmc_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'tmc_preread_input', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: c_tmp
      INTEGER                                            :: itmp
      LOGICAL                                            :: explicit_key, flag
      REAL(KIND=dp)                                      :: tmax, tmin
      REAL(KIND=dp), DIMENSION(:), POINTER               :: inp_Temp
      TYPE(section_vals_type), POINTER                   :: tmc_section

      NULLIFY (tmc_section, inp_Temp)

      IF(.NOT.(ASSOCIATED(input)))CALL cp__a("tmc/tmc_setup.F",589)

      tmc_env%tmc_comp_set%ana_on_the_fly = 0
      tmc_section => section_vals_get_subs_vals(input, "MOTION%TMC%TMC_ANALYSIS")
      CALL section_vals_get(tmc_section, explicit=flag)
      IF (flag) THEN
         tmc_env%tmc_comp_set%ana_on_the_fly = 1
      END IF

      tmc_section => section_vals_get_subs_vals(input, "MOTION%TMC")
      CALL section_vals_get(tmc_section, explicit=flag)
      IF(.NOT.(flag))CALL cp__a("tmc/tmc_setup.F",600)

      CALL section_vals_val_get(tmc_section, "PRINT_TEST_OUTPUT", l_val=tmc_env%params%print_test_output)

      IF(.NOT.(ASSOCIATED(tmc_env%tmc_comp_set)))CALL cp__a("tmc/tmc_setup.F",604)
      ! read the parameters for the computational setup
      CALL section_vals_val_get(tmc_section, "GROUP_ENERGY_SIZE", i_val=tmc_env%tmc_comp_set%group_ener_size)
      CALL section_vals_val_get(tmc_section, "GROUP_ENERGY_NR", i_val=tmc_env%tmc_comp_set%group_ener_nr)
      CALL section_vals_val_get(tmc_section, "GROUP_CC_SIZE", i_val=tmc_env%tmc_comp_set%group_cc_size)
      CALL section_vals_val_get(tmc_section, "GROUP_ANLYSIS_NR", i_val=itmp)
      IF (tmc_env%tmc_comp_set%ana_on_the_fly .GT. 0) &
         tmc_env%tmc_comp_set%ana_on_the_fly = itmp
      IF (tmc_env%tmc_comp_set%ana_on_the_fly .GT. 1) &
         CALL cp_abort(cp__l("tmc/tmc_setup.F",613), &
                       "analysing on the fly is up to now not supported for multiple cores. "// &
                       " Restart file witing for this case and temperature "// &
                       "distribution has to be solved.!.")
      CALL section_vals_val_get(tmc_section, "RESULT_LIST_IN_MEMORY", l_val=tmc_env%params%USE_REDUCED_TREE)
      ! swap the variable, because of oposit meaning
      tmc_env%params%USE_REDUCED_TREE = .NOT. tmc_env%params%USE_REDUCED_TREE
      CALL section_vals_val_get(tmc_section, "NR_TEMPERATURE", i_val=tmc_env%params%nr_temp)

      ! stuff everyone needs to know
      CALL section_vals_val_get(tmc_section, "NMC_MOVES%NMC_FILE_NAME", c_val=tmc_env%params%NMC_inp_file)
      IF (tmc_env%params%NMC_inp_file .EQ. tmc_default_unspecified_name) THEN
         ! file name keyword without file name
         CALL cp__b("tmc/tmc_setup.F",626,"no or a valid NMC input file has to be specified ")
      ELSE IF (tmc_env%params%NMC_inp_file .EQ. "") THEN
         ! no keyword
         IF (tmc_env%tmc_comp_set%group_cc_size .GT. 0) &
            CALL cp_warn(cp__l("tmc/tmc_setup.F",630), &
                         "The configurational groups are deactivated, "// &
                         "because no approximated energy input is specified.")
         tmc_env%tmc_comp_set%group_cc_size = 0
      ELSE
         ! check file existance
         INQUIRE (FILE=TRIM(tmc_env%params%NMC_inp_file), EXIST=flag, IOSTAT=itmp)
         IF (.NOT. flag .OR. itmp .NE. 0) &
            CALL cp__b("tmc/tmc_setup.F",638,"a valid NMC input file has to be specified")
      END IF

      CALL section_vals_val_get(tmc_section, "TEMPERATURE", r_vals=inp_Temp)
      IF (tmc_env%params%nr_temp .GT. 1 .AND. SIZE(inp_Temp) .NE. 2) &
         CALL cp__b("tmc/tmc_setup.F",643,"specify each temperature, skip keyword NR_TEMPERATURE")
      IF (tmc_env%params%nr_temp .EQ. 1) THEN
         tmc_env%params%nr_temp = SIZE(inp_Temp)
         ALLOCATE (tmc_env%params%Temp(tmc_env%params%nr_temp))
         tmc_env%params%Temp(:) = inp_Temp(:)
      ELSE
         tmin = inp_Temp(1)
         tmax = inp_Temp(2)
         ALLOCATE (tmc_env%params%Temp(tmc_env%params%nr_temp))
         tmc_env%params%Temp(1) = tmin
         DO itmp = 2, SIZE(tmc_env%params%Temp)
            tmc_env%params%Temp(itmp) = tmc_env%params%Temp(itmp-1)+(tmax-tmin)/(SIZE(tmc_env%params%Temp)-1.0_dp)
         END DO
         IF (ANY(tmc_env%params%Temp .LT. 0.0_dp)) &
            CALL cp_abort(cp__l("tmc/tmc_setup.F",657), "The temperatures are negative. Should be specified using "// &
                          "TEMPERATURE {T_min} {T_max} and NR_TEMPERATURE {#temperatures}")
      END IF

      CALL section_vals_val_get(tmc_section, "TASK_TYPE", explicit=explicit_key)
      IF (explicit_key) THEN
         CALL section_vals_val_get(tmc_section, "TASK_TYPE", c_val=c_tmp)
         SELECT CASE (TRIM (c_tmp))
         CASE (TRIM (tmc_default_unspecified_name))
            tmc_env%params%task_type = task_type_MC
         CASE ("IDEAL_GAS")
            tmc_env%params%task_type = task_type_ideal_gas
         CASE DEFAULT
            CALL cp_warn(cp__l("tmc/tmc_setup.F",670), &
                         'unknown TMC task type "'//TRIM(c_tmp)//'" specified. '// &
                         " Set to default.")
            tmc_env%params%task_type = task_type_MC
         END SELECT
      END IF

   END SUBROUTINE tmc_preread_input

! **************************************************************************************************
!> \brief read the tmc subsection from the input file
!> \param input points to the tmc subsection in the input file
!> \param tmc_env structure for storing all the tmc parameters
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_read_input(input, tmc_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'tmc_read_input', routineP = moduleN//':'//routineN

      INTEGER                                            :: itmp
      LOGICAL                                            :: explicit, flag
      REAL(KIND=dp)                                      :: r_tmp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: r_arr_tmp
      TYPE(section_vals_type), POINTER                   :: tmc_section

      NULLIFY (tmc_section)

      IF(.NOT.(ASSOCIATED(input)))CALL cp__a("tmc/tmc_setup.F",699)

      tmc_section => section_vals_get_subs_vals(input, "MOTION%TMC")
      CALL section_vals_get(tmc_section, explicit=flag)
      IF(.NOT.(flag))CALL cp__a("tmc/tmc_setup.F",703)

      ! only for the master
      IF (tmc_env%tmc_comp_set%group_nr == 0) THEN
         IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_setup.F",707)
         ! the walltime input can be done as HH:MM:SS or just in seconds.
         CALL cp2k_get_walltime(section=input, keyword_name="GLOBAL%WALLTIME", &
                                walltime=tmc_env%m_env%walltime)

         CALL section_vals_val_get(tmc_section, "NUM_MC_ELEM", i_val=tmc_env%m_env%num_MC_elem)
         CALL section_vals_val_get(tmc_section, "RND_DETERMINISTIC", i_val=tmc_env%m_env%rnd_init)
         ! restarting
         CALL section_vals_val_get(tmc_section, "RESTART_IN", c_val=tmc_env%m_env%restart_in_file_name)
         IF (tmc_env%m_env%restart_in_file_name .EQ. tmc_default_unspecified_name) THEN
            tmc_env%m_env%restart_in_file_name = tmc_default_restart_in_file_name
            INQUIRE (FILE=tmc_env%m_env%restart_in_file_name, EXIST=flag)
            IF (.NOT. flag) tmc_env%m_env%restart_in_file_name = ""
         END IF
         CALL section_vals_val_get(tmc_section, "RESTART_OUT", i_val=tmc_env%m_env%restart_out_step)
         ! restart just at the end (lone keyword)
         IF (tmc_env%m_env%restart_out_step .EQ. -9) THEN
            tmc_env%m_env%restart_out_file_name = tmc_default_restart_out_file_name
            tmc_env%m_env%restart_out_step = HUGE(tmc_env%m_env%restart_out_step)
         END IF
         IF (tmc_env%m_env%restart_out_step .LT. 0) &
            CALL cp_abort(cp__l("tmc/tmc_setup.F",728), &
                          "Please specify a valid value for the frequency "// &
                          "to write restart files (RESTART_OUT #). "// &
                          "# > 0 to define the amount of Markov chain elements in between, "// &
                          "or 0 to deactivate the restart file writing. "// &
                          "Lonely keyword writes restart file only at the end of the run.")

         CALL section_vals_val_get(tmc_section, "INFO_OUT_STEP_SIZE", i_val=tmc_env%m_env%info_out_step_size)
         CALL section_vals_val_get(tmc_section, "DOT_TREE", c_val=tmc_env%params%dot_file_name)
         CALL section_vals_val_get(tmc_section, "ALL_CONF_FILE_NAME", c_val=tmc_env%params%all_conf_file_name)
         IF (tmc_env%params%dot_file_name .NE. "") tmc_env%params%DRAW_TREE = .TRUE.

         ! everything for the worker group
      ELSE IF (tmc_env%tmc_comp_set%group_nr .NE. 0) THEN
         IF(.NOT.(ASSOCIATED(tmc_env%w_env)))CALL cp__a("tmc/tmc_setup.F",742)
      END IF

      ! stuff everyone needs to know

      ! the NMC_FILE_NAME is already read in tmc_preread_input
      CALL section_vals_val_get(tmc_section, "ENERGY_FILE_NAME", c_val=tmc_env%params%energy_inp_file)
      ! file name keyword without file name
      IF (tmc_env%params%energy_inp_file .EQ. "") &
         CALL cp__b("tmc/tmc_setup.F",751,"a valid exact energy input file has to be specified ")
      ! check file existance
      INQUIRE (FILE=TRIM(tmc_env%params%energy_inp_file), EXIST=flag, IOSTAT=itmp)
      IF (.NOT. flag .OR. itmp .NE. 0) &
         CALL cp_abort(cp__l("tmc/tmc_setup.F",755), "a valid exact energy input file has to be specified, "// &
                       TRIM(tmc_env%params%energy_inp_file)//" does not exist.")

      CALL section_vals_val_get(tmc_section, "NUM_MV_ELEM_IN_CELL", i_val=tmc_env%params%nr_elem_mv)

      CALL section_vals_val_get(tmc_section, "VOLUME_ISOTROPIC", l_val=tmc_env%params%v_isotropic)
      CALL section_vals_val_get(tmc_section, "PRESSURE", r_val=tmc_env%params%pressure)
      tmc_env%params%pressure = tmc_env%params%pressure/au2bar
      CALL section_vals_val_get(tmc_section, "MOVE_CENTER_OF_MASS", l_val=tmc_env%params%mv_cen_of_mass)

      CALL section_vals_val_get(tmc_section, "SUB_BOX", r_vals=r_arr_tmp)
      IF (SIZE(r_arr_tmp) .GT. 1) THEN
         IF (SIZE(r_arr_tmp) .NE. tmc_env%params%dim_per_elem) &
            CALL cp__b("tmc/tmc_setup.F",768,"The entered sub box sizes does not fit in number of dimensions.")
         IF (ANY(r_arr_tmp .LE. 0.0_dp)) &
            CALL cp__b("tmc/tmc_setup.F",770,"The entered sub box lenghts should be greater than 0.")
         DO itmp = 1, SIZE(tmc_env%params%sub_box_size)
            tmc_env%params%sub_box_size(itmp) = r_arr_tmp(itmp)/au2a
         END DO
      ELSE IF (r_arr_tmp(1) .GT. 0.0_dp) THEN
         r_tmp = r_arr_tmp(1)/au2a
         tmc_env%params%sub_box_size(:) = r_tmp
      END IF

      ! read all the distinct moves
      CALL read_init_move_types(tmc_params=tmc_env%params, &
                                tmc_section=tmc_section)

      CALL section_vals_val_get(tmc_section, "ESIMATE_ACC_PROB", l_val=tmc_env%params%esimate_acc_prob)
      CALL section_vals_val_get(tmc_section, "SPECULATIVE_CANCELING", l_val=tmc_env%params%SPECULATIVE_CANCELING)
      CALL section_vals_val_get(tmc_section, "USE_SCF_ENERGY_INFO", l_val=tmc_env%params%use_scf_energy_info)
      ! printing
      CALL section_vals_val_get(tmc_section, "PRINT_ONLY_ACC", l_val=tmc_env%params%print_only_diff_conf)
      CALL section_vals_val_get(tmc_section, "PRINT_COORDS", l_val=tmc_env%params%print_trajectory)
      CALL section_vals_val_get(tmc_section, "PRINT_DIPOLE", explicit=explicit)
      IF (explicit) &
         CALL section_vals_val_get(tmc_section, "PRINT_DIPOLE", l_val=tmc_env%params%print_dipole)
      CALL section_vals_val_get(tmc_section, "PRINT_FORCES", explicit=explicit)
      IF (explicit) &
         CALL section_vals_val_get(tmc_section, "PRINT_FORCES", l_val=tmc_env%params%print_forces)
      CALL section_vals_val_get(tmc_section, "PRINT_CELL", explicit=explicit)
      IF (explicit) &
         CALL section_vals_val_get(tmc_section, "PRINT_CELL", l_val=tmc_env%params%print_cell)
      CALL section_vals_val_get(tmc_section, "PRINT_ENERGIES", l_val=tmc_env%params%print_energies)

   END SUBROUTINE tmc_read_input

! **************************************************************************************************
!> \brief creates a new para environment for tmc
!> \param tmc_comp_set structure with parameters for computational setup
!> \param para_env the old parallel environment
!> \param ana_on_the_fly ...
!> \param success ...
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_redistributing_cores(tmc_comp_set, para_env, ana_on_the_fly, &
                                       success)
      TYPE(tmc_comp_set_type), POINTER                   :: tmc_comp_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER                                            :: ana_on_the_fly
      LOGICAL                                            :: success

      CHARACTER(len=*), PARAMETER :: routineN = 'tmc_redistributing_cores', &
         routineP = moduleN//':'//routineN

      INTEGER :: cc_group, cc_group_rank, comm_tmp, master_ana_group, master_ana_rank, &
         master_first_e_worker_g, master_first_e_worker_r, master_worker_group, &
         master_worker_rank, my_mpi_undefined, my_mpi_world, total_used
      LOGICAL                                            :: flag, master

      IF(.NOT.(ASSOCIATED(tmc_comp_set)))CALL cp__a("tmc/tmc_setup.F",825)
      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("tmc/tmc_setup.F",826)

      ! colors and positions for new communicators
      ! variables for printing
      tmc_comp_set%group_nr = -1
      my_mpi_undefined = para_env%num_pe+10000 !HUGE(my_mpi_undefined)! mp_undefined
      master_worker_group = my_mpi_undefined
      master_worker_rank = -1
      cc_group = my_mpi_undefined
      cc_group_rank = -1
      master_first_e_worker_g = my_mpi_undefined
      master_first_e_worker_r = -1
      master_ana_group = my_mpi_undefined
      master_ana_rank = -1

      master = .FALSE.
      flag = .FALSE.
      success = .TRUE.

      IF (para_env%num_pe .LE. 1) THEN
         CALL cp__w("tmc/tmc_setup.F",846,"TMC need at least 2 cores (one for master, one for worker)")
         success = .FALSE.
      ELSE
         ! check if there are enougth cores available
         IF (tmc_comp_set%group_ener_size*tmc_comp_set%group_ener_nr .GT. (para_env%num_pe-1)) &
            CALL cp__w("tmc/tmc_setup.F",851,"The selected energy group size is too huge. ")
         IF (flag) THEN
            tmc_comp_set%group_ener_nr = INT((para_env%num_pe-1)/ &
                                             REAL(tmc_comp_set%group_ener_size, KIND=dp))
            IF (tmc_comp_set%group_ener_nr .LT. 1) &
               CALL cp__w("tmc/tmc_setup.F",856,"The selected energy group size is too huge. ")
            IF (flag) success = .FALSE.
         END IF

         ! set the amount of configurational change worker groups
         tmc_comp_set%group_cc_nr = 0
         IF (tmc_comp_set%group_cc_size .GT. 0) THEN
            tmc_comp_set%group_cc_nr = INT((para_env%num_pe-1-tmc_comp_set%ana_on_the_fly &
                                            -tmc_comp_set%group_ener_size*tmc_comp_set%group_ener_nr)/ &
                                           REAL(tmc_comp_set%group_cc_size, KIND=dp))

            IF (tmc_comp_set%group_cc_nr .LT. 1) &
               CALL cp_warn(cp__l("tmc/tmc_setup.F",868), &
                            "There are not enougth cores left for creating groups for configurational change.")
            IF (flag) success = .FALSE.
         END IF

         total_used = tmc_comp_set%group_ener_size*tmc_comp_set%group_ener_nr+ &
                      tmc_comp_set%group_cc_size*tmc_comp_set%group_cc_nr+ &
                      tmc_comp_set%ana_on_the_fly
         IF (para_env%num_pe-1 .GT. total_used) &
            CALL cp__w("tmc/tmc_setup.F",877," mpi ranks are unused, but can be used for analysis.")

         ! duplicate communicator
         CALL mp_comm_dup(para_env%group, my_mpi_world)

         ! determine the master node
         IF (para_env%mepos == para_env%num_pe-1) THEN
            master = .TRUE.
            master_worker_group = para_env%num_pe+3 ! belong to master_worker_comm
            master_worker_rank = 0 ! rank in m_w_comm
            master_first_e_worker_g = para_env%num_pe+3 ! belong to master_first_energy_worker_comm
            master_first_e_worker_r = 0
            tmc_comp_set%group_nr = 0 !para_env%num_pe +3
            master_ana_group = para_env%num_pe+4
            master_ana_rank = 0
         ELSE
            ! energy calculation groups
            IF (para_env%mepos .LT. tmc_comp_set%group_ener_size*tmc_comp_set%group_ener_nr) THEN
               tmc_comp_set%group_nr = INT(para_env%mepos/tmc_comp_set%group_ener_size)+1 ! assign to groups
               ! master of worker group
               IF (MODULO(para_env%mepos, tmc_comp_set%group_ener_size) .EQ. 0) THEN ! tmc_comp_set%group_nr masters
                  master_worker_group = para_env%num_pe+3 ! belong to master_worker_comm
                  master_worker_rank = tmc_comp_set%group_nr ! rank in m_w_comm
                  IF (master_worker_rank .EQ. 1) THEN
                     master_first_e_worker_g = para_env%num_pe+3 ! belong to master_first_energy_worker_comm
                     master_first_e_worker_r = 1
                  END IF
               END IF
               cc_group = tmc_comp_set%group_nr
               cc_group_rank = para_env%mepos- &
                               (tmc_comp_set%group_nr-1)*tmc_comp_set%group_ener_size ! rank in worker group

               ! configurational change groups
            ELSE IF (para_env%mepos .LT. (tmc_comp_set%group_ener_size*tmc_comp_set%group_ener_nr+ &
                                          tmc_comp_set%group_cc_size*tmc_comp_set%group_cc_nr)) THEN
               cc_group_rank = para_env%mepos-tmc_comp_set%group_ener_size*tmc_comp_set%group_ener_nr ! temporary
               tmc_comp_set%group_nr = tmc_comp_set%group_ener_nr+1+INT(cc_group_rank/tmc_comp_set%group_cc_size)
               cc_group = tmc_comp_set%group_nr
               ! master of worker group
               IF (MODULO(cc_group_rank, tmc_comp_set%group_cc_size) .EQ. 0) THEN ! tmc_comp_set%group_nr masters
                  master_worker_group = para_env%num_pe+3 ! belong to master_worker_comm
                  master_worker_rank = tmc_comp_set%group_nr ! rank in m_w_comm
               END IF
               !cc_group_rank = cc_group_rank-(tmc_comp_set%group_nr-1)*tmc_comp_set%group_cc_size       ! rank in worker group
               cc_group_rank = MODULO(cc_group_rank, tmc_comp_set%group_cc_size) ! rank in worker group
            ELSE
               ! not used cores
               ! up to now we use just one core for doing the analysis
               IF (para_env%mepos .EQ. para_env%num_pe-2) THEN
                  tmc_comp_set%group_nr = para_env%mepos-(para_env%num_pe-1) ! negative
                  IF(.NOT.(tmc_comp_set%group_nr .LT. 0))CALL cp__a("tmc/tmc_setup.F",927)
                  IF (para_env%mepos .GE. para_env%num_pe-1-ana_on_the_fly) THEN
                     master_ana_group = para_env%num_pe+4
                     master_ana_rank = -tmc_comp_set%group_nr
                  END IF
               END IF
            END IF
         END IF

         IF (success) THEN
            ! -- splitting communicators
            ! worker intern communication
            CALL mp_comm_split_direct(my_mpi_world, comm_tmp, cc_group, cc_group_rank)
            NULLIFY (tmc_comp_set%para_env_sub_group)
            ! not the unused cores
            IF (cc_group_rank .GE. 0) THEN
               CALL cp_para_env_create(para_env=tmc_comp_set%para_env_sub_group, &
                                       group=comm_tmp)
            ELSE
               CALL mp_comm_free(comm_tmp)
            ENDIF

            ! worker master communication
            CALL mp_comm_split_direct(my_mpi_world, comm_tmp, master_worker_group, master_worker_rank)
            NULLIFY (tmc_comp_set%para_env_m_w)
            ! not the unused cores
            IF (master_worker_rank .GE. 0) THEN
               CALL cp_para_env_create(para_env=tmc_comp_set%para_env_m_w, &
                                       group=comm_tmp)
            ELSE
               CALL mp_comm_free(comm_tmp)
            ENDIF

            ! communicator only for first energy worker master and global master
            CALL mp_comm_split_direct(my_mpi_world, comm_tmp, master_first_e_worker_g, master_first_e_worker_r)
            NULLIFY (tmc_comp_set%para_env_m_first_w)
            ! not the unused cores
            IF (master_first_e_worker_r .GE. 0) THEN
               CALL cp_para_env_create(para_env=tmc_comp_set%para_env_m_first_w, &
                                       group=comm_tmp)
            ELSE
               CALL mp_comm_free(comm_tmp)
            ENDIF

            ! communicator only for analysis worker and global master
            CALL mp_comm_split_direct(my_mpi_world, comm_tmp, master_ana_group, master_ana_rank)
            NULLIFY (tmc_comp_set%para_env_m_ana)
            IF (master_ana_rank .GE. 0) THEN
               CALL cp_para_env_create(para_env=tmc_comp_set%para_env_m_ana, &
                                       group=comm_tmp)
            ELSE
               CALL mp_comm_free(comm_tmp)
            ENDIF

            ! communicator for master only to handle external control
            master_ana_group = my_mpi_undefined
            master_ana_rank = -1
            IF (master) THEN
               master_ana_group = 1
               master_ana_rank = 1
            END IF
            CALL mp_comm_split_direct(my_mpi_world, comm_tmp, master_ana_group, master_ana_rank)
            NULLIFY (tmc_comp_set%para_env_m_only)
            IF (master_ana_rank .GE. 0) THEN
               CALL cp_para_env_create(para_env=tmc_comp_set%para_env_m_only, &
                                       group=comm_tmp)
            ELSE
               CALL mp_comm_free(comm_tmp)
            ENDIF
         END IF
         CALL mp_comm_free(my_mpi_world)
      END IF
   END SUBROUTINE tmc_redistributing_cores

! **************************************************************************************************
!> \brief prints the most important parameters used for TMC
!> \param tmc_env tructure with parameters for TMC
!> \author Mandes 11.2012
! **************************************************************************************************
   SUBROUTINE tmc_print_params(tmc_env)
      TYPE(tmc_env_type), POINTER                        :: tmc_env

      CHARACTER(LEN=*), PARAMETER :: fmt_my = '(T2,A,"| ",A,T41,A40)', plabel = "TMC", &
         routineN = 'tmc_print_params', routineP = moduleN//':'//routineN

      CHARACTER(LEN=80)                                  :: c_tmp, fmt_tmp
      INTEGER                                            :: file_nr

      IF(.NOT.(ASSOCIATED(tmc_env)))CALL cp__a("tmc/tmc_setup.F",1015)
      IF(.NOT.(ASSOCIATED(tmc_env%tmc_comp_set)))CALL cp__a("tmc/tmc_setup.F",1016)
      ! only the master prints out
      IF (tmc_env%tmc_comp_set%group_nr == 0) THEN
         file_nr = tmc_env%m_env%io_unit
         IF(.NOT.(ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_w)))CALL cp__a("tmc/tmc_setup.F",1020)
         IF(.NOT.(ASSOCIATED(tmc_env%m_env)))CALL cp__a("tmc/tmc_setup.F",1021)

         CALL m_flush(file_nr)
         WRITE (file_nr, *)

         WRITE (UNIT=file_nr, FMT="(/,T2,A)") REPEAT("-", 79)
         WRITE (UNIT=file_nr, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=file_nr, FMT="(T2,A,T35,A,T80,A)") "-", "TMC setting", "-"
         WRITE (UNIT=file_nr, FMT="(T2,A,T80,A)") "-", "-"
         WRITE (UNIT=file_nr, FMT="(T2,A)") REPEAT("-", 79)

         WRITE (UNIT=file_nr, FMT="(T2,A,T35,A,T80,A)") "-", "distribution of cores", "-"
         WRITE (file_nr, FMT=fmt_my) plabel, "number of all working groups ", &
            cp_to_string(tmc_env%tmc_comp_set%para_env_m_w%num_pe-1)
         WRITE (file_nr, FMT=fmt_my) plabel, "number of groups (ener|cc)", &
            cp_to_string(tmc_env%tmc_comp_set%group_ener_nr)//" | "// &
            cp_to_string(tmc_env%tmc_comp_set%group_cc_nr)
         WRITE (file_nr, FMT=fmt_my) plabel, "cores per group (ener|cc) ", &
            cp_to_string(tmc_env%tmc_comp_set%group_ener_size)//" | "// &
            cp_to_string(tmc_env%tmc_comp_set%group_cc_size)
         IF (ASSOCIATED(tmc_env%tmc_comp_set%para_env_m_ana)) &
            WRITE (file_nr, FMT=fmt_my) plabel, "Analysis groups ", &
            cp_to_string(tmc_env%tmc_comp_set%para_env_m_ana%num_pe-1)
         IF (SIZE(tmc_env%params%Temp(:)) .LE. 7) THEN
            WRITE (fmt_tmp, *) '(T2,A,"| ",A,T25,A56)'
            c_tmp = ""
            WRITE (c_tmp, FMT="(1000F8.2)") tmc_env%params%Temp(:)
            WRITE (file_nr, FMT=fmt_tmp) plabel, "Temperature(s) [K]", TRIM(c_tmp)
         ELSE
            WRITE (file_nr, FMT='(A,1000F8.2)') " "//plabel//"| Temperature(s) [K]", &
               tmc_env%params%Temp(:)
         END IF
         WRITE (file_nr, FMT=fmt_my) plabel, "# of Monte Carlo Chain elements: ", &
            cp_to_string(tmc_env%m_env%num_MC_elem)
         WRITE (file_nr, FMT=fmt_my) plabel, "exact potential input file:", &
            TRIM(tmc_env%params%energy_inp_file)
         IF (tmc_env%params%NMC_inp_file .NE. "") &
            WRITE (file_nr, FMT=fmt_my) plabel, "approximate potential input file:", &
            TRIM(tmc_env%params%NMC_inp_file)
         IF (ANY(tmc_env%params%sub_box_size .GT. 0.0_dp)) THEN
            WRITE (fmt_tmp, *) '(T2,A,"| ",A,T25,A56)'
            c_tmp = ""
            WRITE (c_tmp, FMT="(1000F8.2)") tmc_env%params%sub_box_size(:)*au2a
            WRITE (file_nr, FMT=fmt_tmp) plabel, "Sub box size [A]", TRIM(c_tmp)
         END IF
         IF (tmc_env%params%pressure .GT. 0.0_dp) &
            WRITE (file_nr, FMT=fmt_my) plabel, "Pressure [bar]: ", &
            cp_to_string(tmc_env%params%pressure*au2bar)
         WRITE (file_nr, FMT=fmt_my) plabel, "Numbers of atoms/molecules moved "
         WRITE (file_nr, FMT=fmt_my) plabel, "  within one conf. change", &
            cp_to_string(tmc_env%params%nr_elem_mv)
         WRITE (UNIT=file_nr, FMT="(/,T2,A)") REPEAT("-", 79)
      END IF

   END SUBROUTINE tmc_print_params

END MODULE
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief holds all the structure types needed for Monte Carlo, except
!>      the mc_environment_type
!> \par History
!>      none
!> \author MJM
! **************************************************************************************************
MODULE mc_types
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE fist_environment_types,          ONLY: fist_env_get,&
                                              fist_environment_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_get,&
                                              fist_nonbond_env_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_p_type,&
                                              force_env_type,&
                                              use_fist_force,&
                                              use_qs_force
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: do_fist,&
                                              do_qs
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              dp_size,&
                                              int_size
   USE mathconstants,                   ONLY: pi
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_p_type
   USE molecule_kind_types,             ONLY: atom_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE pair_potential_types,            ONLY: pair_potential_pp_type
   USE particle_list_types,             ONLY: particle_list_p_type
   USE physcon,                         ONLY: boltzmann,&
                                              joule
   USE string_utilities,                ONLY: uppercase,&
                                              xstring
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   PUBLIC :: mc_simpar_type, &
             mc_simulation_parameters_p_type, &
             mc_averages_type, mc_averages_p_type, &
             mc_moves_type, mc_moves_p_type, accattempt, &
             get_mc_par, set_mc_par, read_mc_section, &
             find_mc_rcut, mc_molecule_info_type, &
             mc_determine_molecule_info, mc_molecule_info_destroy, &
             mc_sim_par_create, mc_sim_par_destroy, &
             get_mc_molecule_info, &
             mc_input_file_type, mc_input_file_create, &
             get_mc_input_file, mc_input_file_destroy, &
             mc_input_parameters_check, &
             mc_ekin_type

! **************************************************************************************************
   TYPE mc_simpar_type
! contains all the parameters needed for running Monte Carlo simulations
      PRIVATE
      INTEGER, DIMENSION(:), POINTER :: avbmc_atom
      INTEGER :: nstep
      INTEGER :: iupvolume
      INTEGER :: iuptrans
      INTEGER :: nvirial
      INTEGER :: nbox
      INTEGER :: nmoves
      INTEGER :: nswapmoves
      INTEGER :: rm
      INTEGER :: cl
      INTEGER :: diff
      INTEGER :: nstart
      INTEGER :: source
      INTEGER :: group
      INTEGER :: iprint
      LOGICAL :: ldiscrete
      LOGICAL :: lbias
      LOGICAL :: ionode
      LOGICAL :: lrestart
      LOGICAL :: lstop
      LOGICAL :: lhmc
      CHARACTER(LEN=20) :: ensemble
      CHARACTER(LEN=default_path_length) :: restart_file_name
      CHARACTER(LEN=default_path_length) :: molecules_file
      CHARACTER(LEN=default_path_length) :: moves_file
      CHARACTER(LEN=default_path_length) :: coords_file
      CHARACTER(LEN=default_path_length) :: energy_file
      CHARACTER(LEN=default_path_length) :: displacement_file
      CHARACTER(LEN=default_path_length) :: cell_file
      CHARACTER(LEN=default_path_length) :: dat_file
      CHARACTER(LEN=default_path_length) :: data_file
      CHARACTER(LEN=default_path_length) :: box2_file
      CHARACTER(LEN=200) :: fft_lib
      CHARACTER(LEN=50) :: PROGRAM
      REAL(dp), DIMENSION(:), POINTER :: pmtraion_mol, pmtrans_mol, pmrot_mol, pmswap_mol, &
                                         pbias, pmavbmc_mol
      REAL(dp) :: discrete_step
      REAL(dp) :: rmvolume
      REAL(dp), DIMENSION(:), POINTER :: rmbond, rmangle, rmdihedral, rmrot, rmtrans
      REAL(dp), DIMENSION(:), POINTER :: eta
      REAL(dp) :: temperature
      REAL(dp) :: pressure
      REAL(dp) :: pmavbmc
      REAL(dp) :: pmswap
      REAL(dp) :: pmvolume, pmvol_box
      REAL(dp) :: pmhmc, pmhmc_box
      REAL(dp) :: pmtraion
      REAL(dp) :: pmtrans
      REAL(dp) :: BETA
      REAL(dp) :: rcut
      REAL(dp), DIMENSION(:), POINTER :: avbmc_rmin, avbmc_rmax
      REAL(dp), DIMENSION(:), POINTER :: virial_temps
      TYPE(mc_input_file_type), POINTER :: &
         mc_input_file, mc_bias_file
      TYPE(section_vals_type), POINTER :: input_file
      TYPE(mc_molecule_info_type), POINTER :: mc_molecule_info
      REAL(dp) :: exp_min_val, exp_max_val, min_val, max_val
      INTEGER     :: rand2skip
   END TYPE mc_simpar_type

! **************************************************************************************************
   TYPE mc_ekin_type
! contains the kinetic energy of an MD sequence from hybrid MC
      REAL(dp) :: initial_ekin, final_ekin
   END TYPE mc_ekin_type
! **************************************************************************************************
   TYPE mc_input_file_type
! contains all the text of the input file
      PRIVATE
      INTEGER :: run_type_row, run_type_column, coord_row_start, coord_row_end, &
                 cell_row, cell_column, global_row_end, in_use, nunits_empty, &
                 motion_row_end, motion_row_start
      INTEGER, DIMENSION(:), POINTER :: mol_set_nmol_row, mol_set_nmol_column
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: text
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: atom_names_empty
      REAL(dp), DIMENSION(:, :), POINTER :: coordinates_empty
   END TYPE mc_input_file_type

! **************************************************************************************************
   TYPE mc_molecule_info_type
! contains information on molecules...I had to do this because I use multiple force
! environments, and they know nothing about each other
      PRIVATE
      INTEGER :: nboxes
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: names
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:, :), POINTER :: atom_names
      REAL(dp), DIMENSION(:, :), POINTER :: conf_prob, mass
      INTEGER, DIMENSION(:, :), POINTER :: nchains
      INTEGER :: nmol_types, nchain_total
      INTEGER, DIMENSION(:), POINTER :: nunits, mol_type, nunits_tot, in_box
   END TYPE mc_molecule_info_type

! **************************************************************************************************
   TYPE mc_simulation_parameters_p_type
! a pointer for multiple boxes
      TYPE(mc_simpar_type), POINTER :: mc_par
   END TYPE mc_simulation_parameters_p_type

! **************************************************************************************************
   TYPE mc_averages_type
! contains some data that is averaged throughout the course of a run
      REAL(KIND=dp) :: ave_energy
      REAL(KIND=dp) :: ave_energy_squared
      REAL(KIND=dp) :: ave_volume
      REAL(KIND=dp) :: molecules
   END TYPE mc_averages_type

! **************************************************************************************************
   TYPE mc_averages_p_type
      TYPE(mc_averages_type), POINTER :: averages
   END TYPE mc_averages_p_type

! **************************************************************************************************
   TYPE mc_moves_type
! contains data for how many moves of a give type we've accepted/rejected
      TYPE(accattempt), POINTER :: bias_bond
      TYPE(accattempt), POINTER :: bias_angle
      TYPE(accattempt), POINTER :: bias_dihedral
      TYPE(accattempt), POINTER :: bias_trans
      TYPE(accattempt), POINTER :: bias_rot
      TYPE(accattempt), POINTER :: bond
      TYPE(accattempt), POINTER :: angle
      TYPE(accattempt), POINTER :: dihedral
      TYPE(accattempt), POINTER :: trans
      TYPE(accattempt), POINTER :: rot
      TYPE(accattempt), POINTER :: swap
      TYPE(accattempt), POINTER :: volume
      TYPE(accattempt), POINTER :: hmc
      TYPE(accattempt), POINTER :: avbmc_inin
      TYPE(accattempt), POINTER :: avbmc_outin
      TYPE(accattempt), POINTER :: avbmc_inout
      TYPE(accattempt), POINTER :: avbmc_outout
      TYPE(accattempt), POINTER :: Quickstep
      REAL(KIND=dp) :: trans_dis, qtrans_dis
      INTEGER :: empty, grown, empty_conf, empty_avbmc
   END TYPE mc_moves_type

! **************************************************************************************************
   TYPE accattempt
      INTEGER :: successes
      INTEGER :: qsuccesses
      INTEGER :: attempts
   END TYPE accattempt

! **************************************************************************************************
   TYPE mc_moves_p_type
      TYPE(mc_moves_type), POINTER :: moves
   END TYPE mc_moves_p_type

! *** Global parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_types'

CONTAINS

! **************************************************************************************************
!> \brief accesses the private elements of the mc_input_file_type
!> \param mc_input_file the input file you want data for
!>
!>    Suitable for parallel.
!> \param run_type_row ...
!> \param run_type_column ...
!> \param coord_row_start ...
!> \param coord_row_end ...
!> \param cell_row ...
!> \param cell_column ...
!> \param global_row_end ...
!> \param mol_set_nmol_row ...
!> \param mol_set_nmol_column ...
!> \param in_use ...
!> \param text ...
!> \param atom_names_empty ...
!> \param nunits_empty ...
!> \param coordinates_empty ...
!> \param motion_row_end ...
!> \param motion_row_start ...
!> \author MJM
! **************************************************************************************************
   SUBROUTINE get_mc_input_file(mc_input_file, run_type_row, run_type_column, &
                                coord_row_start, coord_row_end, cell_row, cell_column, global_row_end, &
                                mol_set_nmol_row, mol_set_nmol_column, in_use, text, atom_names_empty, &
                                nunits_empty, coordinates_empty, motion_row_end, motion_row_start)

      TYPE(mc_input_file_type), POINTER                  :: mc_input_file
      INTEGER, INTENT(OUT), OPTIONAL                     :: run_type_row, run_type_column, &
                                                            coord_row_start, coord_row_end, &
                                                            cell_row, cell_column, global_row_end
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: mol_set_nmol_row, mol_set_nmol_column
      INTEGER, INTENT(OUT), OPTIONAL                     :: in_use
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: text, atom_names_empty
      INTEGER, INTENT(OUT), OPTIONAL                     :: nunits_empty
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: coordinates_empty
      INTEGER, INTENT(OUT), OPTIONAL                     :: motion_row_end, motion_row_start

      IF (PRESENT(in_use)) in_use = mc_input_file%in_use
      IF (PRESENT(run_type_row)) run_type_row = mc_input_file%run_type_row
      IF (PRESENT(run_type_column)) run_type_column = mc_input_file%run_type_column
      IF (PRESENT(coord_row_start)) coord_row_start = mc_input_file%coord_row_start
      IF (PRESENT(coord_row_end)) coord_row_end = mc_input_file%coord_row_end
      IF (PRESENT(cell_row)) cell_row = mc_input_file%cell_row
      IF (PRESENT(cell_column)) cell_column = mc_input_file%cell_column
      IF (PRESENT(global_row_end)) global_row_end = mc_input_file%global_row_end
      IF (PRESENT(nunits_empty)) nunits_empty = mc_input_file%nunits_empty
      IF (PRESENT(motion_row_start)) motion_row_start = mc_input_file%motion_row_start
      IF (PRESENT(motion_row_end)) motion_row_end = mc_input_file%motion_row_end

      IF (PRESENT(mol_set_nmol_row)) mol_set_nmol_row => mc_input_file%mol_set_nmol_row
      IF (PRESENT(mol_set_nmol_column)) mol_set_nmol_column => mc_input_file%mol_set_nmol_column
      IF (PRESENT(text)) text => mc_input_file%text
      IF (PRESENT(atom_names_empty)) atom_names_empty => mc_input_file%atom_names_empty
      IF (PRESENT(coordinates_empty)) coordinates_empty => mc_input_file%coordinates_empty

   END SUBROUTINE GET_MC_INPUT_FILE
! **************************************************************************************************
!> \brief ...
!> \param mc_par ...
!> \param nstep ...
!> \param nvirial ...
!> \param iuptrans ...
!> \param iupvolume ...
!> \param nmoves ...
!> \param nswapmoves ...
!> \param rm ...
!> \param cl ...
!> \param diff ...
!> \param nstart ...
!> \param source ...
!> \param group ...
!> \param lbias ...
!> \param ionode ...
!> \param lrestart ...
!> \param lstop ...
!> \param rmvolume ...
!> \param rmbond ...
!> \param rmangle ...
!> \param rmrot ...
!> \param rmtrans ...
!> \param temperature ...
!> \param pressure ...
!> \param BETA ...
!> \param pmswap ...
!> \param pmvolume ...
!> \param pmtraion ...
!> \param pmtrans ...
!> \param ensemble ...
!> \param PROGRAM ...
!> \param restart_file_name ...
!> \param molecules_file ...
!> \param moves_file ...
!> \param coords_file ...
!> \param energy_file ...
!> \param displacement_file ...
!> \param cell_file ...
!> \param dat_file ...
!> \param data_file ...
!> \param box2_file ...
!> \param fft_lib ...
!> \param iprint ...
!> \param rcut ...
!> \param ldiscrete ...
!> \param discrete_step ...
!> \param pmavbmc ...
!> \param pbias ...
!> \param avbmc_atom ...
!> \param avbmc_rmin ...
!> \param avbmc_rmax ...
!> \param rmdihedral ...
!> \param input_file ...
!> \param mc_molecule_info ...
!> \param pmswap_mol ...
!> \param pmavbmc_mol ...
!> \param pmtrans_mol ...
!> \param pmrot_mol ...
!> \param pmtraion_mol ...
!> \param mc_input_file ...
!> \param mc_bias_file ...
!> \param pmvol_box ...
!> \param virial_temps ...
!> \param exp_min_val ...
!> \param exp_max_val ...
!> \param min_val ...
!> \param max_val ...
!> \param eta ...
!> \param pmhmc ...
!> \param pmhmc_box ...
!> \param lhmc ...
!> \param rand2skip ...
! **************************************************************************************************
   SUBROUTINE get_mc_par(mc_par, nstep, nvirial, iuptrans, iupvolume, &
                         nmoves, nswapmoves, rm, cl, diff, nstart, &
                         source, group, lbias, ionode, lrestart, lstop, rmvolume, rmbond, rmangle, &
                         rmrot, rmtrans, temperature, pressure, BETA, pmswap, pmvolume, pmtraion, pmtrans, &
                         ensemble, PROGRAM, restart_file_name, molecules_file, moves_file, coords_file, &
                         energy_file, displacement_file, cell_file, dat_file, data_file, box2_file, &
                         fft_lib, iprint, rcut, ldiscrete, discrete_step, &
                         pmavbmc, pbias, avbmc_atom, avbmc_rmin, avbmc_rmax, rmdihedral, &
                         input_file, mc_molecule_info, pmswap_mol, pmavbmc_mol, pmtrans_mol, pmrot_mol, &
                         pmtraion_mol, mc_input_file, mc_bias_file, pmvol_box, virial_temps, &
                         exp_min_val, exp_max_val, min_val, max_val, eta, pmhmc, pmhmc_box, lhmc, rand2skip)

! **************************************************************************************************
!> \brief accesses the private elements of the mc_parameters_type
!> \param mc_par the structure mc parameters you want
!>
!>    Suitable for parallel.
!> \author MJM
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      INTEGER, INTENT(OUT), OPTIONAL                     :: nstep, nvirial, iuptrans, iupvolume, &
                                                            nmoves, nswapmoves, rm, cl, diff, &
                                                            nstart, source, group
      LOGICAL, INTENT(OUT), OPTIONAL                     :: lbias, ionode, lrestart, lstop
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: rmvolume
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: rmbond, rmangle, rmrot, rmtrans
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: temperature, pressure, BETA, pmswap, &
                                                            pmvolume, pmtraion, pmtrans
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL :: ensemble, PROGRAM, restart_file_name, &
         molecules_file, moves_file, coords_file, energy_file, displacement_file, cell_file, &
         dat_file, data_file, box2_file, fft_lib
      INTEGER, INTENT(OUT), OPTIONAL                     :: iprint
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: rcut
      LOGICAL, INTENT(OUT), OPTIONAL                     :: ldiscrete
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: discrete_step, pmavbmc
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: pbias
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: avbmc_atom
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: avbmc_rmin, avbmc_rmax, rmdihedral
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input_file
      TYPE(mc_molecule_info_type), OPTIONAL, POINTER     :: mc_molecule_info
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: pmswap_mol
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: pmavbmc_mol, pmtrans_mol, pmrot_mol, &
                                                            pmtraion_mol
      TYPE(mc_input_file_type), OPTIONAL, POINTER        :: mc_input_file, mc_bias_file
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: pmvol_box
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: virial_temps
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: exp_min_val, exp_max_val, min_val, &
                                                            max_val
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: eta
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: pmhmc, pmhmc_box
      LOGICAL, INTENT(OUT), OPTIONAL                     :: lhmc
      INTEGER, INTENT(OUT), OPTIONAL                     :: rand2skip

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_mc_par', routineP = moduleN//':'//routineN

      IF (PRESENT(nstep)) nstep = mc_par%nstep
      IF (PRESENT(nvirial)) nvirial = mc_par%nvirial
      IF (PRESENT(iuptrans)) iuptrans = mc_par%iuptrans
      IF (PRESENT(iupvolume)) iupvolume = mc_par%iupvolume
      IF (PRESENT(nmoves)) nmoves = mc_par%nmoves
      IF (PRESENT(nswapmoves)) nswapmoves = mc_par%nswapmoves
      IF (PRESENT(rm)) rm = mc_par%rm
      IF (PRESENT(cl)) cl = mc_par%cl
      IF (PRESENT(diff)) diff = mc_par%diff
      IF (PRESENT(nstart)) nstart = mc_par%nstart
      IF (PRESENT(source)) source = mc_par%source
      IF (PRESENT(group)) group = mc_par%group
      IF (PRESENT(iprint)) iprint = mc_par%iprint

      IF (PRESENT(lbias)) lbias = mc_par%lbias
      IF (PRESENT(ionode)) ionode = mc_par%ionode
      IF (PRESENT(lrestart)) lrestart = mc_par%lrestart
      IF (PRESENT(lstop)) lstop = mc_par%lstop
      IF (PRESENT(ldiscrete)) ldiscrete = mc_par%ldiscrete

      IF (PRESENT(virial_temps)) virial_temps => mc_par%virial_temps
      IF (PRESENT(avbmc_atom)) avbmc_atom => mc_par%avbmc_atom
      IF (PRESENT(avbmc_rmin)) avbmc_rmin => mc_par%avbmc_rmin
      IF (PRESENT(avbmc_rmax)) avbmc_rmax => mc_par%avbmc_rmax
      IF (PRESENT(rcut)) rcut = mc_par%rcut
      IF (PRESENT(discrete_step)) discrete_step = mc_par%discrete_step
      IF (PRESENT(rmvolume)) rmvolume = mc_par%rmvolume
      IF (PRESENT(temperature)) temperature = mc_par%temperature
      IF (PRESENT(pressure)) pressure = mc_par%pressure
      IF (PRESENT(BETA)) BETA = mc_par%BETA
      IF (PRESENT(exp_max_val)) exp_max_val = mc_par%exp_max_val
      IF (PRESENT(exp_min_val)) exp_min_val = mc_par%exp_min_val
      IF (PRESENT(max_val)) max_val = mc_par%max_val
      IF (PRESENT(min_val)) min_val = mc_par%min_val
      IF (PRESENT(pmswap)) pmswap = mc_par%pmswap
      IF (PRESENT(pmvolume)) pmvolume = mc_par%pmvolume
      IF (PRESENT(lhmc)) lhmc = mc_par%lhmc
      IF (PRESENT(pmhmc)) pmhmc = mc_par%pmhmc
      IF (PRESENT(pmhmc_box)) pmhmc_box = mc_par%pmhmc_box
      IF (PRESENT(pmvol_box)) pmvol_box = mc_par%pmvol_box
      IF (PRESENT(pmtraion)) pmtraion = mc_par%pmtraion
      IF (PRESENT(pmtrans)) pmtrans = mc_par%pmtrans
      IF (PRESENT(pmavbmc)) pmavbmc = mc_par%pmavbmc
      IF (PRESENT(pmrot_mol)) pmrot_mol => mc_par%pmrot_mol
      IF (PRESENT(pmtrans_mol)) pmtrans_mol => mc_par%pmtrans_mol
      IF (PRESENT(pmtraion_mol)) pmtraion_mol => mc_par%pmtraion_mol
      IF (PRESENT(pmavbmc_mol)) pmavbmc_mol => mc_par%pmavbmc_mol
      IF (PRESENT(pbias)) pbias => mc_par%pbias

      IF (PRESENT(rmbond)) rmbond => mc_par%rmbond
      IF (PRESENT(rmangle)) rmangle => mc_par%rmangle
      IF (PRESENT(rmdihedral)) rmdihedral => mc_par%rmdihedral
      IF (PRESENT(rmrot)) rmrot => mc_par%rmrot
      IF (PRESENT(rmtrans)) rmtrans => mc_par%rmtrans

      IF (PRESENT(eta)) eta => mc_par%eta

      IF (PRESENT(ensemble)) ensemble = mc_par%ensemble
      IF (PRESENT(PROGRAM)) PROGRAM = mc_par%program
      IF (PRESENT(restart_file_name)) restart_file_name = &
         mc_par%restart_file_name
      IF (PRESENT(moves_file)) moves_file = mc_par%moves_file
      IF (PRESENT(coords_file)) coords_file = mc_par%coords_file
      IF (PRESENT(molecules_file)) molecules_file = mc_par%molecules_file
      IF (PRESENT(energy_file)) energy_file = mc_par%energy_file
      IF (PRESENT(displacement_file)) displacement_file = &
         mc_par%displacement_file
      IF (PRESENT(cell_file)) cell_file = mc_par%cell_file
      IF (PRESENT(dat_file)) dat_file = mc_par%dat_file
      IF (PRESENT(data_file)) data_file = mc_par%data_file
      IF (PRESENT(box2_file)) box2_file = mc_par%box2_file
      IF (PRESENT(fft_lib)) fft_lib = mc_par%fft_lib

      IF (PRESENT(input_file)) input_file => mc_par%input_file
      IF (PRESENT(mc_molecule_info)) mc_molecule_info => mc_par%mc_molecule_info
      IF (PRESENT(mc_input_file)) mc_input_file => mc_par%mc_input_file
      IF (PRESENT(mc_bias_file)) mc_bias_file => mc_par%mc_bias_file

      IF (PRESENT(pmswap_mol)) pmswap_mol => mc_par%pmswap_mol
      IF (PRESENT(rand2skip)) rand2skip = mc_par%rand2skip
   END SUBROUTINE get_mc_par

! **************************************************************************************************
!> \brief ...
!> \param mc_molecule_info ...
!> \param nmol_types ...
!> \param nchain_total ...
!> \param nboxes ...
!> \param names ...
!> \param conf_prob ...
!> \param nchains ...
!> \param nunits ...
!> \param mol_type ...
!> \param nunits_tot ...
!> \param in_box ...
!> \param atom_names ...
!> \param mass ...
! **************************************************************************************************
   SUBROUTINE get_mc_molecule_info(mc_molecule_info, nmol_types, nchain_total, nboxes, &
                                   names, conf_prob, nchains, nunits, mol_type, nunits_tot, in_box, atom_names, &
                                   mass)

! **************************************************************************************************
!> \brief accesses the private elements of the mc_molecule_info_type
!> \param mc_molecule_info the structure you want the parameters for
!>
!>    Suitable for parallel.
!> \author MJM
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      INTEGER, INTENT(OUT), OPTIONAL                     :: nmol_types, nchain_total, nboxes
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: names
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: conf_prob
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: nchains
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nunits, mol_type, nunits_tot, in_box
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:, :), OPTIONAL, POINTER              :: atom_names
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: mass

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_mc_molecule_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(nboxes)) nboxes = mc_molecule_info%nboxes
      IF (PRESENT(nchain_total)) nchain_total = mc_molecule_info%nchain_total
      IF (PRESENT(nmol_types)) nmol_types = mc_molecule_info%nmol_types

      IF (PRESENT(names)) names => mc_molecule_info%names
      IF (PRESENT(atom_names)) atom_names => mc_molecule_info%atom_names

      IF (PRESENT(conf_prob)) conf_prob => mc_molecule_info%conf_prob
      IF (PRESENT(mass)) mass => mc_molecule_info%mass

      IF (PRESENT(nchains)) nchains => mc_molecule_info%nchains
      IF (PRESENT(nunits)) nunits => mc_molecule_info%nunits
      IF (PRESENT(mol_type)) mol_type => mc_molecule_info%mol_type
      IF (PRESENT(nunits_tot)) nunits_tot => mc_molecule_info%nunits_tot
      IF (PRESENT(in_box)) in_box => mc_molecule_info%in_box

   END SUBROUTINE get_mc_molecule_info

! **************************************************************************************************
!> \brief ...
!> \param mc_par ...
!> \param rm ...
!> \param cl ...
!> \param diff ...
!> \param nstart ...
!> \param rmvolume ...
!> \param rmbond ...
!> \param rmangle ...
!> \param rmdihedral ...
!> \param rmrot ...
!> \param rmtrans ...
!> \param PROGRAM ...
!> \param nmoves ...
!> \param nswapmoves ...
!> \param lstop ...
!> \param temperature ...
!> \param pressure ...
!> \param iuptrans ...
!> \param iupvolume ...
!> \param pmswap ...
!> \param pmvolume ...
!> \param pmtraion ...
!> \param pmtrans ...
!> \param BETA ...
!> \param rcut ...
!> \param iprint ...
!> \param lbias ...
!> \param nstep ...
!> \param lrestart ...
!> \param ldiscrete ...
!> \param discrete_step ...
!> \param pmavbmc ...
!> \param mc_molecule_info ...
!> \param pmavbmc_mol ...
!> \param pmtrans_mol ...
!> \param pmrot_mol ...
!> \param pmtraion_mol ...
!> \param pmswap_mol ...
!> \param avbmc_rmin ...
!> \param avbmc_rmax ...
!> \param avbmc_atom ...
!> \param pbias ...
!> \param ensemble ...
!> \param pmvol_box ...
!> \param eta ...
!> \param mc_input_file ...
!> \param mc_bias_file ...
!> \param exp_max_val ...
!> \param exp_min_val ...
!> \param min_val ...
!> \param max_val ...
!> \param pmhmc ...
!> \param pmhmc_box ...
!> \param lhmc ...
!> \param ionode ...
!> \param source ...
!> \param group ...
!> \param rand2skip ...
! **************************************************************************************************
   SUBROUTINE set_mc_par(mc_par, rm, cl, &
                         diff, nstart, rmvolume, rmbond, rmangle, rmdihedral, rmrot, rmtrans, PROGRAM, &
                         nmoves, nswapmoves, lstop, temperature, pressure, iuptrans, iupvolume, &
                         pmswap, pmvolume, pmtraion, pmtrans, BETA, rcut, iprint, lbias, nstep, &
                         lrestart, ldiscrete, discrete_step, pmavbmc, mc_molecule_info, &
                         pmavbmc_mol, pmtrans_mol, pmrot_mol, pmtraion_mol, pmswap_mol, &
                         avbmc_rmin, avbmc_rmax, avbmc_atom, pbias, ensemble, pmvol_box, eta, &
                         mc_input_file, mc_bias_file, exp_max_val, exp_min_val, min_val, max_val, &
                         pmhmc, pmhmc_box, lhmc, ionode, source, group, rand2skip)

! **************************************************************************************************
!> \brief changes the private elements of the mc_parameters_type
!> \param mc_par the structure mc parameters you want
!>
!>    Suitable for parallel.
!> \author MJM
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      INTEGER, INTENT(IN), OPTIONAL                      :: rm, cl, diff, nstart
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: rmvolume
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: rmbond, rmangle, rmdihedral, rmrot, &
                                                            rmtrans
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: PROGRAM
      INTEGER, INTENT(IN), OPTIONAL                      :: nmoves, nswapmoves
      LOGICAL, INTENT(IN), OPTIONAL                      :: lstop
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: temperature, pressure
      INTEGER, INTENT(IN), OPTIONAL                      :: iuptrans, iupvolume
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: pmswap, pmvolume, pmtraion, pmtrans, &
                                                            BETA, rcut
      INTEGER, INTENT(IN), OPTIONAL                      :: iprint
      LOGICAL, INTENT(IN), OPTIONAL                      :: lbias
      INTEGER, INTENT(IN), OPTIONAL                      :: nstep
      LOGICAL, INTENT(IN), OPTIONAL                      :: lrestart, ldiscrete
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: discrete_step, pmavbmc
      TYPE(mc_molecule_info_type), OPTIONAL, POINTER     :: mc_molecule_info
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: pmavbmc_mol, pmtrans_mol, pmrot_mol, &
                                                            pmtraion_mol, pmswap_mol, avbmc_rmin, &
                                                            avbmc_rmax
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: avbmc_atom
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: pbias
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: ensemble
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: pmvol_box
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: eta
      TYPE(mc_input_file_type), OPTIONAL, POINTER        :: mc_input_file, mc_bias_file
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: exp_max_val, exp_min_val, min_val, &
                                                            max_val, pmhmc, pmhmc_box
      LOGICAL, INTENT(IN), OPTIONAL                      :: lhmc, ionode
      INTEGER, INTENT(IN), OPTIONAL                      :: source, group, rand2skip

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_mc_par', routineP = moduleN//':'//routineN

      INTEGER                                            :: iparm

! These are the only values that change during the course of the simulation
! or are computed outside of this module

      IF (PRESENT(nstep)) mc_par%nstep = nstep
      IF (PRESENT(rm)) mc_par%rm = rm
      IF (PRESENT(cl)) mc_par%cl = cl
      IF (PRESENT(diff)) mc_par%diff = diff
      IF (PRESENT(nstart)) mc_par%nstart = nstart
      IF (PRESENT(nmoves)) mc_par%nmoves = nmoves
      IF (PRESENT(nswapmoves)) mc_par%nswapmoves = nswapmoves
      IF (PRESENT(iprint)) mc_par%iprint = iprint
      IF (PRESENT(iuptrans)) mc_par%iuptrans = iuptrans
      IF (PRESENT(iupvolume)) mc_par%iupvolume = iupvolume

      IF (PRESENT(ldiscrete)) mc_par%ldiscrete = ldiscrete
      IF (PRESENT(lstop)) mc_par%lstop = lstop
      IF (PRESENT(lbias)) mc_par%lbias = lbias
      IF (PRESENT(lrestart)) mc_par%lrestart = lrestart

      IF (PRESENT(exp_max_val)) mc_par%exp_max_val = exp_max_val
      IF (PRESENT(exp_min_val)) mc_par%exp_min_val = exp_min_val
      IF (PRESENT(max_val)) mc_par%max_val = max_val
      IF (PRESENT(min_val)) mc_par%min_val = min_val
      IF (PRESENT(BETA)) mc_par%BETA = BETA
      IF (PRESENT(temperature)) mc_par%temperature = temperature
      IF (PRESENT(rcut)) mc_par%rcut = rcut
      IF (PRESENT(pressure)) mc_par%pressure = pressure
      IF (PRESENT(pmvolume)) mc_par%pmvolume = pmvolume
      IF (PRESENT(lhmc)) mc_par%lhmc = lhmc
      IF (PRESENT(pmhmc)) mc_par%pmhmc = pmhmc
      IF (PRESENT(pmhmc_box)) mc_par%pmhmc_box = pmhmc_box
      IF (PRESENT(pmvol_box)) mc_par%pmvol_box = pmvol_box
      IF (PRESENT(pmswap)) mc_par%pmswap = pmswap
      IF (PRESENT(pmtrans)) mc_par%pmtrans = pmtrans
      IF (PRESENT(pmtraion)) mc_par%pmtraion = pmtraion
      IF (PRESENT(pmavbmc)) mc_par%pmavbmc = pmavbmc

      IF (PRESENT(discrete_step)) mc_par%discrete_step = discrete_step
      IF (PRESENT(rmvolume)) mc_par%rmvolume = rmvolume

      IF (PRESENT(mc_input_file)) mc_par%mc_input_file => mc_input_file
      IF (PRESENT(mc_bias_file)) mc_par%mc_bias_file => mc_bias_file

! cannot just change the pointers, because then we have memory leaks
! and the program crashes if we try to release it (in certain cases)
      IF (PRESENT(eta)) THEN
         DO iparm = 1, SIZE(eta)
            mc_par%eta(iparm) = eta(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(rmbond)) THEN
         DO iparm = 1, SIZE(rmbond)
            mc_par%rmbond(iparm) = rmbond(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(rmangle)) THEN
         DO iparm = 1, SIZE(rmangle)
            mc_par%rmangle(iparm) = rmangle(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(rmdihedral)) THEN
         DO iparm = 1, SIZE(rmdihedral)
            mc_par%rmdihedral(iparm) = rmdihedral(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(rmrot)) THEN
         DO iparm = 1, SIZE(rmrot)
            mc_par%rmrot(iparm) = rmrot(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(rmtrans)) THEN
         DO iparm = 1, SIZE(rmtrans)
            mc_par%rmtrans(iparm) = rmtrans(iparm)
         ENDDO
      ENDIF

      IF (PRESENT(pmavbmc_mol)) THEN
         DO iparm = 1, SIZE(pmavbmc_mol)
            mc_par%pmavbmc_mol(iparm) = pmavbmc_mol(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(pmtrans_mol)) THEN
         DO iparm = 1, SIZE(pmtrans_mol)
            mc_par%pmtrans_mol(iparm) = pmtrans_mol(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(pmrot_mol)) THEN
         DO iparm = 1, SIZE(pmrot_mol)
            mc_par%pmrot_mol(iparm) = pmrot_mol(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(pmtraion_mol)) THEN
         DO iparm = 1, SIZE(pmtraion_mol)
            mc_par%pmtraion_mol(iparm) = pmtraion_mol(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(pmswap_mol)) THEN
         DO iparm = 1, SIZE(pmswap_mol)
            mc_par%pmswap_mol(iparm) = pmswap_mol(iparm)
         ENDDO
      ENDIF

      IF (PRESENT(avbmc_atom)) THEN
         DO iparm = 1, SIZE(avbmc_atom)
            mc_par%avbmc_atom(iparm) = avbmc_atom(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(avbmc_rmin)) THEN
         DO iparm = 1, SIZE(avbmc_rmin)
            mc_par%avbmc_rmin(iparm) = avbmc_rmin(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(avbmc_rmax)) THEN
         DO iparm = 1, SIZE(avbmc_rmax)
            mc_par%avbmc_rmax(iparm) = avbmc_rmax(iparm)
         ENDDO
      ENDIF
      IF (PRESENT(pbias)) THEN
         DO iparm = 1, SIZE(pbias)
            mc_par%pbias(iparm) = pbias(iparm)
         ENDDO
      ENDIF

      IF (PRESENT(PROGRAM)) mc_par%program = PROGRAM
      IF (PRESENT(ensemble)) mc_par%ensemble = ensemble

      IF (PRESENT(mc_molecule_info)) mc_par%mc_molecule_info => mc_molecule_info
      IF (PRESENT(source)) mc_par%source = source
      IF (PRESENT(group)) mc_par%group = group
      IF (PRESENT(ionode)) mc_par%ionode = ionode

      IF (PRESENT(rand2skip)) mc_par%rand2skip = rand2skip
   END SUBROUTINE set_mc_par

! **************************************************************************************************
!> \brief creates (allocates) the mc_simulation_parameters type
!> \param mc_par the structure that will be created (allocated)
!> \param nmol_types the number of molecule types in the system
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_sim_par_create(mc_par, nmol_types)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      INTEGER, INTENT(IN)                                :: nmol_types

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_sim_par_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: istat

      NULLIFY (mc_par)

      ALLOCATE (mc_par)
      ALLOCATE (mc_par%pmtraion_mol(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 834, &
                                       "mc_par%pmtraion_mol", nmol_types*dp_size)
      ALLOCATE (mc_par%pmtrans_mol(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 837, &
                                       "mc_par%pmtrans_mol", nmol_types*dp_size)
      ALLOCATE (mc_par%pmrot_mol(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 840, &
                                       "mc_par%pmrot_mol", nmol_types*dp_size)
      ALLOCATE (mc_par%pmswap_mol(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 843, &
                                       "mc_par%pmswap_mol", nmol_types*dp_size)

      ALLOCATE (mc_par%eta(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 847, &
                                       "mc_par%eta", nmol_types*dp_size)

      ALLOCATE (mc_par%rmbond(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 851, &
                                       "mc_par%rmbond", nmol_types*dp_size)
      ALLOCATE (mc_par%rmangle(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 854, &
                                       "mc_par%rmangle", nmol_types*dp_size)
      ALLOCATE (mc_par%rmdihedral(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 857, &
                                       "mc_par%rmdihedral", nmol_types*dp_size)
      ALLOCATE (mc_par%rmtrans(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 860, &
                                       "mc_par%rmtrans", nmol_types*dp_size)
      ALLOCATE (mc_par%rmrot(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 863, &
                                       "mc_par%rmrot", nmol_types*dp_size)

      ALLOCATE (mc_par%avbmc_atom(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 867, &
                                       "mc_par%avbmc_atom", nmol_types*int_size)
      ALLOCATE (mc_par%avbmc_rmin(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 870, &
                                       "mc_par%avbmc_rmin", nmol_types*dp_size)
      ALLOCATE (mc_par%avbmc_rmax(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 873, &
                                       "mc_par%avbmc_rmax", nmol_types*dp_size)
      ALLOCATE (mc_par%pmavbmc_mol(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 876, &
                                       "mc_par%pmavbmc_mol", nmol_types*dp_size)
      ALLOCATE (mc_par%pbias(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 879, &
                                       "mc_par%pbias", nmol_types*dp_size)

      ALLOCATE (mc_par%mc_input_file, STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 883, &
                                       "mc_par%mc_input_file", 1)
      ALLOCATE (mc_par%mc_bias_file, STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 886, &
                                       "mc_par%mc_bias_file", 1)

   END SUBROUTINE mc_sim_par_create

! **************************************************************************************************
!> \brief destroys (deallocates) the mc_simulation_parameters type
!> \param mc_par the structure that will be destroyed
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_sim_par_destroy(mc_par)

      TYPE(mc_simpar_type), POINTER                      :: mc_par

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_sim_par_destroy', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (mc_par%mc_input_file)
      DEALLOCATE (mc_par%mc_bias_file)

      DEALLOCATE (mc_par%pmtraion_mol)
      DEALLOCATE (mc_par%pmtrans_mol)
      DEALLOCATE (mc_par%pmrot_mol)
      DEALLOCATE (mc_par%pmswap_mol)

      DEALLOCATE (mc_par%eta)

      DEALLOCATE (mc_par%rmbond)
      DEALLOCATE (mc_par%rmangle)
      DEALLOCATE (mc_par%rmdihedral)
      DEALLOCATE (mc_par%rmtrans)
      DEALLOCATE (mc_par%rmrot)

      DEALLOCATE (mc_par%avbmc_atom)
      DEALLOCATE (mc_par%avbmc_rmin)
      DEALLOCATE (mc_par%avbmc_rmax)
      DEALLOCATE (mc_par%pmavbmc_mol)
      DEALLOCATE (mc_par%pbias)
      IF (mc_par%ensemble == "VIRIAL") THEN
         DEALLOCATE (mc_par%virial_temps)
      END IF

      DEALLOCATE (mc_par)
      NULLIFY (mc_par)

   END SUBROUTINE mc_sim_par_destroy

! **************************************************************************************************
!> \brief creates (allocates) the mc_input_file_type
!> \param mc_input_file the structure that will be created (allocated)
!> \param input_file_name the name of the file to read
!> \param mc_molecule_info ...
!> \param empty_coords ...
!> \param lhmc ...
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_input_file_create(mc_input_file, input_file_name, &
                                   mc_molecule_info, empty_coords, lhmc)

      TYPE(mc_input_file_type), POINTER                  :: mc_input_file
      CHARACTER(LEN=*), INTENT(IN)                       :: input_file_name
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      REAL(dp), DIMENSION(:, :)                          :: empty_coords
      LOGICAL, INTENT(IN)                                :: lhmc

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_input_file_create', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_string_length)                   :: cdum, line, method_name
      CHARACTER(default_string_length), &
         DIMENSION(:, :), POINTER                        :: atom_names
      INTEGER :: abc_column, abc_row, cell_column, cell_row, idum, iline, io_stat, irep, istat, &
         iunit, iw, line_number, nlines, nmol_types, nstart, row_number, unit
      INTEGER, DIMENSION(:), POINTER                     :: nunits

! some stuff in case we need to write error messages to the screen

      iw = cp_logger_get_default_io_unit()

! allocate the array we'll need in case we have an empty box
      CALL get_mc_molecule_info(mc_molecule_info, nmol_types=nmol_types, &
                                nunits=nunits, atom_names=atom_names)
      ALLOCATE (mc_input_file%coordinates_empty(1:3, 1:nunits(1)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 969, &
                                       "mc_input_file%coordinates_empty", nunits(1)*5)
      ALLOCATE (mc_input_file%atom_names_empty(1:nunits(1)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 972, &
                                       "mc_input_file%atom_names_empty", nunits(1)*5)
      DO iunit = 1, nunits(1)
         mc_input_file%atom_names_empty(iunit) = atom_names(iunit, 1)
         mc_input_file%coordinates_empty(1:3, iunit) = empty_coords(1:3, iunit)
      ENDDO
      mc_input_file%nunits_empty = nunits(1)

! allocate some arrays
      ALLOCATE (mc_input_file%mol_set_nmol_row(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 982, &
                                       "mc_input_file%mol_set_nmol_row", nmol_types*int_size)
      ALLOCATE (mc_input_file%mol_set_nmol_column(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 985, &
                                       "mc_input_file%mol_set_nmol_column", nmol_types*int_size)

! now read in and store the input file, first finding out how many lines it is
      CALL open_file(file_name=input_file_name, &
                     unit_number=unit, file_action='READ', file_status='OLD')

      nlines = 0
      DO
         READ (unit, '(A)', IOSTAT=io_stat) line
         IF (io_stat .NE. 0) EXIT
         nlines = nlines+1
      ENDDO

      ALLOCATE (mc_input_file%text(1:nlines), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1000, &
                                       "mc_input_file%text", nlines*default_string_length)

      REWIND (unit)
      DO iline = 1, nlines
         READ (unit, '(A)') mc_input_file%text(iline)
      ENDDO

! close the input file
      CALL close_file(unit_number=unit)

! now we need to find the row and column numbers of a variety of things
! first, Let's find the first END after GLOBAL
      CALL mc_parse_text(mc_input_file%text, 1, nlines, "&GLOBAL", .TRUE., &
                         mc_input_file%global_row_end, idum)
      IF (mc_input_file%global_row_end == 0) THEN
         IF (iw > 0) THEN
            WRITE (iw, *)
            WRITE (iw, *) 'File name ', input_file_name
         END IF
         CALL cp_abort(cp__l("motion/mc/mc_types.F",1020), &
                       'Could not find &END after &GLOBAL (make sure & is in the same column) and last line is blank')
      ENDIF

! Let's find the first END after MOTION...we might need this whole section
! because of hybrid MD/MC moves, which requires the MD information to always
! be attached to the force env
      CALL mc_parse_text(mc_input_file%text, 1, nlines, "&MOTION", .TRUE., &
                         mc_input_file%motion_row_end, idum, start_row_number=mc_input_file%motion_row_start)
      IF (mc_input_file%motion_row_end == 0) THEN
         IF (iw > 0) THEN
            WRITE (iw, *)
            WRITE (iw, *) 'File name ', input_file_name, mc_input_file%motion_row_start
         END IF
         CALL cp_abort(cp__l("motion/mc/mc_types.F",1034), &
                       'Could not find &END after &MOTION (make sure & is in the same column) and last line is blank')
      ENDIF

! first, let's find the first END after &COORD, and the line of &COORD
      CALL mc_parse_text(mc_input_file%text, 1, nlines, "&COORD", .TRUE., &
                         mc_input_file%coord_row_end, idum)
      CALL mc_parse_text(mc_input_file%text, 1, nlines, "&COORD", .FALSE., &
                         mc_input_file%coord_row_start, idum)

      IF (mc_input_file%coord_row_end == 0) THEN
         IF (iw > 0) THEN
            WRITE (iw, *)
            WRITE (iw, *) 'File name ', input_file_name
         END IF
         CALL cp_abort(cp__l("motion/mc/mc_types.F",1049), &
                       'Could not find &END after &COORD (make sure & is the first in the same column after &COORD)')
      ENDIF

! now the RUN_TYPE
      CALL mc_parse_text(mc_input_file%text, 1, nlines, "RUN_TYPE", .FALSE., &
                         mc_input_file%run_type_row, mc_input_file%run_type_column)
      mc_input_file%run_type_column = mc_input_file%run_type_column+9
      IF (mc_input_file%run_type_row == 0) THEN
         IF (iw > 0) THEN
            WRITE (iw, *)
            WRITE (iw, *) 'File name ', input_file_name
         END IF
         CALL cp_abort(cp__l("motion/mc/mc_types.F",1062), &
                       'Could not find RUN_TYPE in the input file (should be in the &GLOBAL section).')
      ENDIF

! now the CELL stuff..we don't care about REF_CELL since that should
! never change if it's there
      CALL mc_parse_text(mc_input_file%text, 1, nlines, "&CELL", .FALSE., &
                         mc_input_file%cell_row, mc_input_file%cell_column)
! now find the ABC input line after CELL
      CALL mc_parse_text(mc_input_file%text, mc_input_file%cell_row+1, nlines, &
                         "ABC", .FALSE., abc_row, abc_column)
! is there a &CELL inbetween?  If so, that ABC will be for the ref_cell
! and we need to find the next one
      CALL mc_parse_text(mc_input_file%text, mc_input_file%cell_row+1, abc_row, &
                         "&CELL", .FALSE., cell_row, cell_column)
      IF (cell_row == 0) THEN ! nothing in between...we found the correct ABC
         mc_input_file%cell_row = abc_row
         mc_input_file%cell_column = abc_column+4
      ELSE
         CALL mc_parse_text(mc_input_file%text, abc_row+1, nlines, &
                            "ABC", .FALSE., mc_input_file%cell_row, mc_input_file%cell_column)
      ENDIF
      IF (mc_input_file%cell_row == 0) THEN
         IF (iw > 0) THEN
            WRITE (iw, *)
            WRITE (iw, *) 'File name ', input_file_name
         END IF
         CALL cp_abort(cp__l("motion/mc/mc_types.F",1089), &
                       'Could not find &CELL section in the input file.  Or could not find the ABC line after it.')
      ENDIF

! now we need all the MOL_SET NMOLS indicies
      IF (.NOT. lhmc) THEN
         irep = 0
         nstart = 1
         DO
            CALL mc_parse_text(mc_input_file%text, nstart, nlines, "&MOLECULE", &
                               .FALSE., row_number, idum)
            IF (row_number == 0) EXIT
            nstart = row_number+1
            irep = irep+1
            CALL mc_parse_text(mc_input_file%text, nstart, nlines, "NMOL", &
                               .FALSE., mc_input_file%mol_set_nmol_row(irep), &
                               mc_input_file%mol_set_nmol_column(irep))
            mc_input_file%mol_set_nmol_column(irep) = mc_input_file%mol_set_nmol_column(irep)+5

         ENDDO
         IF (irep .NE. nmol_types) THEN
            IF (iw > 0) THEN
               WRITE (iw, *)
               WRITE (iw, *) 'File name ', input_file_name
               WRITE (iw, *) 'Number of molecule types ', nmol_types
               WRITE (iw, *) '&MOLECULE sections found ', irep
            END IF
            CALL cp_abort( &
               cp__l("motion/mc/mc_types.F",1117), &
               'Did not find MOLECULE sections for every molecule in the simulation (make sure both input files have all types)')
         ENDIF
      ENDIF

! now let's find the type of force_env...right now, I'm only concerned with
! QS, and Fist, though it should be trivial to add others
      CALL mc_parse_text(mc_input_file%text, 1, nlines, &
                         "METHOD", .FALSE., line_number, idum)
      READ (mc_input_file%text(line_number), *) cdum, method_name
      CALL uppercase(method_name)
      SELECT CASE (TRIM (ADJUSTL (method_name)))
      CASE ("FIST")
         mc_input_file%in_use = use_fist_force
      CASE ("QS", "QUICKSTEP")
         mc_input_file%in_use = use_qs_force
      CASE default
         CALL cp__b("motion/mc/mc_types.F",1134,'Cannot determine the type of force_env we are using (check METHOD)')
      END SELECT

   END SUBROUTINE mc_input_file_create

! **************************************************************************************************
!> \brief destroys (deallocates) things in the mc_input_file_type
!> \param mc_input_file the structure that will be released (deallocated)
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_input_file_destroy(mc_input_file)

      TYPE(mc_input_file_type), POINTER                  :: mc_input_file

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_input_file_destroy', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (mc_input_file%mol_set_nmol_row)
      DEALLOCATE (mc_input_file%mol_set_nmol_column)
      DEALLOCATE (mc_input_file%text)
      DEALLOCATE (mc_input_file%atom_names_empty)
      DEALLOCATE (mc_input_file%coordinates_empty)

   END SUBROUTINE mc_input_file_destroy

! **************************************************************************************************
!> \brief a basic text parser used to find the row and column numbers of various strings
!>      in the input file, to store as indices for when we create a dat_file...
!>      returns 0 for the row if nothing is found
!> \param text the text to parse
!> \param nstart the line to start searching from
!> \param nend the line to end searching
!> \param string_search the text we're looking for
!> \param lend if .TRUE., find the &END that comes after string_search...assumes that
!>            the row is the first row with & in the same column as the search string
!> \param row_number the row the string is first found on
!> \param column_number the column number that the string first appears on
!> \param start_row_number ...
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_parse_text(text, nstart, nend, string_search, lend, &
                            row_number, column_number, start_row_number)

      CHARACTER(LEN=*), DIMENSION(:), INTENT(IN)         :: text
      INTEGER, INTENT(IN)                                :: nstart, nend
      CHARACTER(LEN=*), INTENT(IN)                       :: string_search
      LOGICAL, INTENT(IN)                                :: lend
      INTEGER, INTENT(OUT)                               :: row_number, column_number
      INTEGER, INTENT(OUT), OPTIONAL                     :: start_row_number

      CHARACTER(default_string_length)                   :: text_temp
      INTEGER                                            :: iline, index_string, jline, string_size

! did not see any string utilities in the code to help, so here I go

      row_number = 0
      column_number = 0

! how long is our string to search for?
      string_size = LEN_TRIM(string_search)
      whole_file: DO iline = nstart, nend

         index_string = -1
         index_string = INDEX(text(iline), string_search(1:string_size))

         IF (index_string .GT. 0) THEN ! we found one
            IF (.NOT. lend) THEN
               row_number = iline
               column_number = index_string
            ELSE
               IF (PRESENT(start_row_number)) start_row_number = iline
               column_number = index_string
               DO jline = iline+1, nend
! now we find the &END that matches up with this one...
! I need proper indentation because I'm not very smart
                  text_temp = text(jline)
                  IF (text_temp(column_number:column_number) .EQ. "&") THEN
                     row_number = jline
                     EXIT
                  ENDIF
               ENDDO
            ENDIF
            EXIT whole_file
         ENDIF
      ENDDO whole_file

   END SUBROUTINE mc_parse_text

! **************************************************************************************************
!> \brief reads in the Monte Carlo simulation parameters from an input file
!> \param mc_par the structure that will store the parameters
!> \param para_env ...
!> \param globenv the global environment for the simulation
!> \param input_file_name the name of the input_file
!> \param input_file the structure that contains all the keywords in the input file
!> \param force_env_section used to grab the type of force_env
!> \author MJM
! **************************************************************************************************
   SUBROUTINE read_mc_section(mc_par, para_env, globenv, input_file_name, input_file, force_env_section)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      CHARACTER(LEN=*), INTENT(IN)                       :: input_file_name
      TYPE(section_vals_type), POINTER                   :: input_file, force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_mc_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=5)                                   :: box_string, molecule_string, &
                                                            tab_box_string, tab_string, &
                                                            tab_string_int
      CHARACTER(LEN=default_string_length)               :: format_box_string, format_string, &
                                                            format_string_int
      INTEGER                                            :: handle, ia, ie, imol, istat, itype, iw, &
                                                            method_name_id, nmol_types, stop_num
      INTEGER, DIMENSION(:), POINTER                     :: temp_i_array
      REAL(dp), DIMENSION(:), POINTER                    :: temp_r_array
      TYPE(section_vals_type), POINTER                   :: mc_section

! begin the timing of the subroutine

      IF(.NOT.(ASSOCIATED(input_file)))CALL cp__a("motion/mc/mc_types.F",1256)

      CALL timeset(routineN, handle)

      NULLIFY (mc_section)
      mc_section => section_vals_get_subs_vals(input_file, &
                                               "MOTION%MC")

! need the input file sturcutre that we're reading from for when we make
! dat files
      mc_par%input_file => input_file

! set the ionode and mepos
      mc_par%ionode = para_env%ionode
      mc_par%group = para_env%group
      mc_par%source = para_env%source

! for convenience
      nmol_types = mc_par%mc_molecule_info%nmol_types

!..defaults...most are set in input_cp2k_motion
      mc_par%nstart = 0
      CALL section_vals_val_get(force_env_section, "METHOD", i_val=method_name_id)
      SELECT CASE (method_name_id)
      CASE (do_fist)
         mc_par%iprint = 100
      CASE (do_qs)
         mc_par%iprint = 1
      END SELECT

      iw = cp_logger_get_default_io_unit()
      IF (iw > 0) WRITE (iw, *)

!..filenames
      mc_par%program = input_file_name
      CALL xstring(mc_par%program, ia, ie)
      mc_par%coords_file = mc_par%program(ia:ie)//'.coordinates'
      mc_par%molecules_file = mc_par%program(ia:ie)//'.molecules'
      mc_par%moves_file = mc_par%program(ia:ie)//'.moves'
      mc_par%energy_file = mc_par%program(ia:ie)//'.energy'
      mc_par%cell_file = mc_par%program(ia:ie)//'.cell'
      mc_par%displacement_file = mc_par%program(ia:ie) &
                                 //'.max_displacements'
      mc_par%data_file = mc_par%program(ia:ie)//'.data'
      stop_num = ie-3
      mc_par%dat_file = mc_par%program(ia:stop_num)//'dat'

! set them into the input parameter structure as the new defaults
      CALL section_vals_val_set(mc_section, "COORDINATE_FILE_NAME", &
                                c_val=mc_par%coords_file)
      CALL section_vals_val_set(mc_section, "DATA_FILE_NAME", &
                                c_val=mc_par%data_file)
      CALL section_vals_val_set(mc_section, "CELL_FILE_NAME", &
                                c_val=mc_par%cell_file)
      CALL section_vals_val_set(mc_section, "MAX_DISP_FILE_NAME", &
                                c_val=mc_par%displacement_file)
      CALL section_vals_val_set(mc_section, "MOVES_FILE_NAME", &
                                c_val=mc_par%moves_file)
      CALL section_vals_val_set(mc_section, "MOLECULES_FILE_NAME", &
                                c_val=mc_par%molecules_file)
      CALL section_vals_val_set(mc_section, "ENERGY_FILE_NAME", &
                                c_val=mc_par%energy_file)

! grab the FFT library name and print level...this is used for writing the dat file
! and hopefully will be changed
      mc_par%fft_lib = globenv%default_fft_library

! get all the move probabilities first...if we are not doing certain types of moves, we don't care
! if the values for those move parameters are strange

! find out if we're only doing HMC moves...we can ignore a lot of extra information
! then, which would ordinarily be cause for concern
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%PMHMC", r_val=mc_par%pmhmc)
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%PMSWAP", r_val=mc_par%pmswap)
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%PMVOLUME", r_val=mc_par%pmvolume)
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%PMAVBMC", r_val=mc_par%pmavbmc)
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%PMTRANS", r_val=mc_par%pmtrans)
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%PMTRAION", r_val=mc_par%pmtraion)

      ! first, grab all the integer values
      CALL section_vals_val_get(mc_section, "NSTEP", i_val=mc_par%nstep)
      CALL section_vals_val_get(mc_section, "NMOVES", i_val=mc_par%nmoves)
      CALL section_vals_val_get(mc_section, "NSWAPMOVES", i_val=mc_par%nswapmoves)
      CALL section_vals_val_get(mc_section, "MOVE_UPDATES%IUPVOLUME", &
                                i_val=mc_par%iupvolume)
      CALL section_vals_val_get(mc_section, "NVIRIAL", i_val=mc_par%nvirial)
      CALL section_vals_val_get(mc_section, "MOVE_UPDATES%IUPTRANS", &
                                i_val=mc_par%iuptrans)
      CALL section_vals_val_get(mc_section, "IPRINT", i_val=mc_par%iprint)
! now an integer array
      CALL section_vals_val_get(mc_section, "AVBMC%AVBMC_ATOM", i_vals=temp_i_array)

      IF (mc_par%pmhmc-mc_par%pmswap >= 1.0_dp .AND. mc_par%pmhmc == 1.0_dp) THEN
         mc_par%lhmc = .TRUE.
      ELSE
         mc_par%lhmc = .FALSE.
      ENDIF

      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_i_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1356,'AVBMC_ATOM must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_i_array)
         mc_par%avbmc_atom(imol) = temp_i_array(imol)
      ENDDO

! now the real values
      CALL section_vals_val_get(mc_section, "PRESSURE", r_val=mc_par%pressure)
      CALL section_vals_val_get(mc_section, "TEMPERATURE", r_val=mc_par%temperature)
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%BOX_PROBABILITIES%PMHMC_BOX", &
                                r_val=mc_par%pmhmc_box)
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%BOX_PROBABILITIES%PMVOL_BOX", &
                                r_val=mc_par%pmvol_box)
      CALL section_vals_val_get(mc_section, "DISCRETE_STEP", r_val=mc_par%discrete_step)
      CALL section_vals_val_get(mc_section, "MAX_DISPLACEMENTS%BOX_DISPLACEMENTS%RMVOLUME", &
                                r_val=mc_par%rmvolume)

! finally the character values
      CALL section_vals_val_get(mc_section, "ENSEMBLE", c_val=mc_par%ensemble)
      CALL section_vals_val_get(mc_section, "RESTART_FILE_NAME", c_val=mc_par%restart_file_name)
      CALL section_vals_val_get(mc_section, "COORDINATE_FILE_NAME", c_val=mc_par%coords_file)
      CALL section_vals_val_get(mc_section, "ENERGY_FILE_NAME", c_val=mc_par%energy_file)
      CALL section_vals_val_get(mc_section, "MOVES_FILE_NAME", c_val=mc_par%moves_file)
      CALL section_vals_val_get(mc_section, "MOLECULES_FILE_NAME", c_val=mc_par%molecules_file)
      CALL section_vals_val_get(mc_section, "CELL_FILE_NAME", c_val=mc_par%cell_file)
      CALL section_vals_val_get(mc_section, "DATA_FILE_NAME", c_val=mc_par%data_file)
      CALL section_vals_val_get(mc_section, "MAX_DISP_FILE_NAME", c_val=mc_par%displacement_file)
      CALL section_vals_val_get(mc_section, "BOX2_FILE_NAME", c_val=mc_par%box2_file)

! set the values of the arrays...if we just point, we have problems when we start fooling around
! with releasing force_envs and wonky values get set (despite that these are private)
      IF (mc_par%ensemble == "VIRIAL") THEN
         CALL section_vals_val_get(mc_section, "VIRIAL_TEMPS", r_vals=temp_r_array)
! yes, I'm allocating here...I cannot find a better place to do it, though
         ALLOCATE (mc_par%virial_temps(1:SIZE(temp_r_array)), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1392, &
                                          "virial_temps", dp_size*SIZE(temp_r_array))
         DO imol = 1, SIZE(temp_r_array)
            mc_par%virial_temps(imol) = temp_r_array(imol)
         ENDDO
      END IF
! all of these arrays should have one value for each type of molecule...so check that
      CALL section_vals_val_get(mc_section, "AVBMC%AVBMC_RMIN", r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1402,'AVBMC_RMIN must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%avbmc_rmin(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "AVBMC%AVBMC_RMAX", r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1411,'AVBMC_RMAXL must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%avbmc_rmax(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "AVBMC%PBIAS", r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1420,'PBIAS must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%pbias(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%MOL_PROBABILITIES%PMAVBMC_MOL", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1430,'PMAVBMC_MOL must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%pmavbmc_mol(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "ETA", r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1439,'ETA must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%eta(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "MAX_DISPLACEMENTS%MOL_DISPLACEMENTS%RMBOND", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1449,'RMBOND must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%rmbond(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "MAX_DISPLACEMENTS%MOL_DISPLACEMENTS%RMANGLE", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1459,'RMANGLE must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%rmangle(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "MAX_DISPLACEMENTS%MOL_DISPLACEMENTS%RMDIHEDRAL", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1469,'RMDIHEDRAL must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%rmdihedral(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "MAX_DISPLACEMENTS%MOL_DISPLACEMENTS%RMROT", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1479,'RMROT must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%rmrot(imol) = temp_r_array(imol)
      ENDDO
      CALL section_vals_val_get(mc_section, "MAX_DISPLACEMENTS%MOL_DISPLACEMENTS%RMTRANS", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1489,'RMTRANS must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%rmtrans(imol) = temp_r_array(imol)
      ENDDO

      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%MOL_PROBABILITIES%PMTRAION_MOL", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1500,'PMTRAION_MOL must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%pmtraion_mol(imol) = temp_r_array(imol)
      ENDDO

      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%MOL_PROBABILITIES%PMTRANS_MOL", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1511,'PMTRANS_MOL must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%pmtrans_mol(imol) = temp_r_array(imol)
      ENDDO

      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%MOL_PROBABILITIES%PMROT_MOL", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1522,'PMROT_MOL must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%pmrot_mol(imol) = temp_r_array(imol)
      ENDDO

      CALL section_vals_val_get(mc_section, "MOVE_PROBABILITIES%MOL_PROBABILITIES%PMSWAP_MOL", &
                                r_vals=temp_r_array)
      IF (.NOT. mc_par%lhmc) THEN
         IF (SIZE(temp_r_array) .NE. nmol_types) THEN
            CALL cp__b("motion/mc/mc_types.F",1533,'PMSWAP_MOL must have as many values as there are molecule types.')
         ENDIF
      ENDIF
      DO imol = 1, SIZE(temp_r_array)
         mc_par%pmswap_mol(imol) = temp_r_array(imol)
      ENDDO

! now some logical values
      CALL section_vals_val_get(mc_section, "LBIAS", l_val=mc_par%lbias)
      CALL section_vals_val_get(mc_section, "LDISCRETE", l_val=mc_par%ldiscrete)
      CALL section_vals_val_get(mc_section, "LSTOP", l_val=mc_par%lstop)
      CALL section_vals_val_get(mc_section, "RESTART", l_val=mc_par%lrestart)
      CALL section_vals_val_get(mc_section, "LBIAS", l_val=mc_par%lbias)

!..end of parsing the input section

!..write some information to output
      IF (iw > 0) THEN
         WRITE (box_string, '(I2)') mc_par%mc_molecule_info%nboxes
         WRITE (molecule_string, '(I2)') mc_par%mc_molecule_info%nmol_types
         WRITE (tab_string, '(I4)') 81-10*mc_par%mc_molecule_info%nmol_types
         WRITE (tab_box_string, '(I4)') 81-10*mc_par%mc_molecule_info%nboxes
         WRITE (tab_string_int, '(I4)') 81-5*mc_par%mc_molecule_info%nmol_types
         format_string = "(A,T"//TRIM(ADJUSTL(tab_string))//","//TRIM(ADJUSTL(molecule_string))//"(2X,F8.4))"
         format_box_string = "(A,T"//TRIM(ADJUSTL(tab_box_string))//","//TRIM(ADJUSTL(box_string))//"(2X,F8.4))"
         format_string_int = "(A,T"//TRIM(ADJUSTL(tab_string))//","//TRIM(ADJUSTL(molecule_string))//"(I3,2X))"
         WRITE (iw, '( A )') ' MC| Monte Carlo Protocol '
         WRITE (iw, '( A,T71,I10 )') ' MC| total number of steps ', &
            mc_par%nstep
         WRITE (iw, '( A,T71,F10.3 )') ' MC| pmvolume ', &
            mc_par%pmvolume
         WRITE (iw, '( A,T71,F10.3 )') ' MC| pmvol_box ', &
            mc_par%pmvol_box
         WRITE (iw, '( A,T71,F10.3 )') ' MC| pmhmc ', &
            mc_par%pmhmc
         WRITE (iw, '( A,T71,F10.3 )') ' MC| pmhmc_box ', &
            mc_par%pmhmc_box
         WRITE (iw, '( A,T71,F10.3 )') ' MC| pmswap ', &
            mc_par%pmswap
         WRITE (iw, '( A,T71,F10.3 )') ' MC| pmavbmc ', &
            mc_par%pmavbmc
         WRITE (iw, '( A,T71,F10.3 )') ' MC| pmtraion ', &
            mc_par%pmtraion
         WRITE (iw, '( A,T71,F10.3 )') ' MC| pmtrans ', &
            mc_par%pmtrans
         WRITE (iw, '( A,T71,I10 )') ' MC| iupvolume ', &
            mc_par%iupvolume
         WRITE (iw, '( A,T71,I10 )') ' MC| nvirial ', &
            mc_par%nvirial
         WRITE (iw, '( A,T71,I10 )') ' MC| iuptrans ', &
            mc_par%iuptrans
         WRITE (iw, '( A,T71,I10 )') ' MC| iprint ', &
            mc_par%iprint
         WRITE (iw, '( A,T61,A20 )') ' MC| ensemble ', &
            ADJUSTR(mc_par%ensemble)
! format string may not be valid if all the molecules are atomic,
! like in HMC
         IF (.NOT. mc_par%lhmc) THEN
            WRITE (iw, format_string) ' MC| pmswap_mol ', &
               mc_par%pmswap_mol
            WRITE (iw, format_string) ' MC| pmavbmc_mol ', &
               mc_par%pmavbmc_mol
            WRITE (iw, format_string) ' MC| pbias ', &
               mc_par%pbias
            WRITE (iw, format_string) ' MC| pmtraion_mol ', &
               mc_par%pmtraion_mol
            WRITE (iw, format_string) ' MC| pmtrans_mol ', &
               mc_par%pmtrans_mol
            WRITE (iw, format_string) ' MC| pmrot_mol ', &
               mc_par%pmrot_mol
            WRITE (iw, format_string) ' MC| eta [K]', &
               mc_par%eta
            WRITE (iw, format_string) ' MC| rmbond [angstroms]', &
               mc_par%rmbond
            WRITE (iw, format_string) ' MC| rmangle [degrees]', &
               mc_par%rmangle
            WRITE (iw, format_string) ' MC| rmdihedral [degrees]', &
               mc_par%rmdihedral
            WRITE (iw, format_string) ' MC| rmtrans [angstroms]', &
               mc_par%rmtrans
            WRITE (iw, format_string) ' MC| rmrot [degrees]', &
               mc_par%rmrot
            WRITE (iw, format_string_int) ' MC| AVBMC target atom ', &
               mc_par%avbmc_atom
            WRITE (iw, format_string) ' MC| AVBMC inner cutoff [ang]', &
               mc_par%avbmc_rmin
            WRITE (iw, format_string) ' MC| AVBMC outer cutoff [ang]', &
               mc_par%avbmc_rmax
         ENDIF
         IF (mc_par%ensemble .EQ. 'GEMC-NVT') THEN
            WRITE (iw, '( A,T58,A)') ' MC| Box 2 file', &
               TRIM(mc_par%box2_file)
         ENDIF
         WRITE (iw, '( A,T58,A )') ' MC| Name of restart file:', &
            TRIM(mc_par%restart_file_name)
         WRITE (iw, '( A,A,T44,A )') ' MC| Name of output ', &
            'coordinate file:', &
            TRIM(mc_par%coords_file)
         WRITE (iw, '( A,T44,A )') ' MC| Name of output data file:', &
            TRIM(mc_par%data_file)
         WRITE (iw, '( A,A,T44,A )') ' MC| Name of output ', &
            'molecules file:', &
            TRIM(mc_par%molecules_file)
         WRITE (iw, '( A,T44,A )') ' MC| Name of output moves file:', &
            TRIM(mc_par%moves_file)
         WRITE (iw, '( A,T44,A )') ' MC| Name of output energy file:', &
            TRIM(mc_par%energy_file)
         WRITE (iw, '( A,T44,A )') ' MC| Name of output cell file:', &
            TRIM(mc_par%cell_file)
         WRITE (iw, '( A,A,T44,A )') ' MC| Name of output', &
            ' displacement file:', &
            TRIM(mc_par%displacement_file)
         IF (mc_par%ldiscrete) THEN
            WRITE (iw, '(A,A,T71,F10.3)') ' MC| discrete step size', &
               '[angstroms]', &
               mc_par%discrete_step
         ELSE
            WRITE (iw, '( A,A,T71,F10.3 )') ' MC| rmvolume ', &
               '[cubic angstroms]', &
               mc_par%rmvolume
         ENDIF
         WRITE (iw, '( A,T71,F10.2 )') ' MC| Temperature [K] ', &
            mc_par%temperature
         WRITE (iw, '( A,T71,F10.5 )') ' MC| Pressure [bar] ', &
            mc_par%pressure
         IF (mc_par%lrestart) THEN
            WRITE (iw, '(A,A)') ' MC| Initial data will be read from a', &
               ' restart file.'
         ENDIF
         IF (mc_par%lbias) THEN
            WRITE (iw, '(A)') ' MC| The moves will be biased.'
         ELSE
            WRITE (iw, '(A)') ' MC| The moves will not be biased,'
         ENDIF
         IF (mc_par%nmoves .EQ. 1) THEN
            WRITE (iw, '(A,A)') ' MC| A full energy calculation ', &
               'will be done at every step.'
         ELSE
            WRITE (iw, '(A,I4,A,A)') ' MC| ', mc_par%nmoves, &
               ' moves will be attempted ', &
               'before a Quickstep energy calculation'
            WRITE (iw, '(A)') ' MC|      takes place.'
         ENDIF

         WRITE (iw, '(A,I4,A,A)') ' MC| ', mc_par%nswapmoves, &
            ' swap insertions will be attempted ', &
            'per molecular swap move'

         IF (mc_par%lhmc) THEN
            WRITE (iw, '(A)') '********************************************************'
            WRITE (iw, '(A)') '************** ONLY DOING HYBRID MONTE CARLO MOVES **************************'
            WRITE (iw, '(A)') '********************************************************'

         ENDIF
      END IF

! figure out what beta (1/kT) is in atomic units (1/Hartree)
      mc_par%BETA = 1/mc_par%temperature/boltzmann*joule

      ! 0.9_dp is to avoid overflow or underflow
      mc_par%exp_max_val = 0.9_dp*LOG(HUGE(0.0_dp))
      mc_par%exp_min_val = 0.9_dp*LOG(TINY(0.0_dp))
      mc_par%max_val = EXP(mc_par%exp_max_val)
      mc_par%min_val = 0.0_dp

! convert from bar to a.u.
      mc_par%pressure = cp_unit_to_cp2k(mc_par%pressure, "bar")
! convert from angstrom to a.u.
      DO itype = 1, mc_par%mc_molecule_info%nmol_types
! convert from Kelvin to a.u.
!         mc_par % eta = mc_par%eta(itype) * boltzmann / joule
! convert from degrees to radians
         mc_par%rmrot(itype) = mc_par%rmrot(itype)/180.0e0_dp*pi
         mc_par%rmangle(itype) = mc_par%rmangle(itype)/180.0e0_dp*pi
         mc_par%rmdihedral(itype) = mc_par%rmdihedral(itype)/180.0e0_dp*pi
         mc_par%rmtrans(itype) = cp_unit_to_cp2k(mc_par%rmtrans(itype), "angstrom")
         mc_par%rmbond(itype) = cp_unit_to_cp2k(mc_par%rmbond(itype), "angstrom")
         mc_par%eta(itype) = cp_unit_to_cp2k(mc_par%eta(itype), "K_e")
         mc_par%avbmc_rmin(itype) = cp_unit_to_cp2k(mc_par%avbmc_rmin(itype), "angstrom")
         mc_par%avbmc_rmax(itype) = cp_unit_to_cp2k(mc_par%avbmc_rmax(itype), "angstrom")
      ENDDO
      mc_par%rmvolume = cp_unit_to_cp2k(mc_par%rmvolume, "angstrom^3")
      mc_par%discrete_step = cp_unit_to_cp2k(mc_par%discrete_step, "angstrom")

! end the timing
      CALL timestop(handle)

   END SUBROUTINE read_mc_section

! **************************************************************************************************
!> \brief finds the largest interaction cutoff value in a classical simulation
!>      so we know the smallest size we can make the box in a volume move
!> \param mc_par the structure that will store the parameters
!> \param force_env the force environment that we'll grab the rcut parameter
!>                   out of
!> \param lterminate set to .TRUE. if one of the sides of the box is
!>            less than twice the cutoff
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE find_mc_rcut(mc_par, force_env, lterminate)

      TYPE(mc_simpar_type), INTENT(INOUT)                :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env
      LOGICAL, INTENT(OUT)                               :: lterminate

      CHARACTER(len=*), PARAMETER :: routineN = 'find_mc_rcut', routineP = moduleN//':'//routineN

      INTEGER                                            :: itype, jtype
      REAL(KIND=dp)                                      :: rcutsq_max
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc
      TYPE(cell_type), POINTER                           :: cell
      TYPE(fist_environment_type), POINTER               :: fist_env
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(pair_potential_pp_type), POINTER              :: potparm

      NULLIFY (cell, potparm, fist_nonbond_env, fist_env)

      lterminate = .FALSE.
      CALL force_env_get(force_env, fist_env=fist_env)
      CALL fist_env_get(fist_env, cell=cell, &
                        fist_nonbond_env=fist_nonbond_env)
      CALL fist_nonbond_env_get(fist_nonbond_env, potparm=potparm)
      CALL get_cell(cell, abc=abc)

! find the largest value of rcutsq
      rcutsq_max = 0.0e0_dp
      DO itype = 1, SIZE(potparm%pot, 1)
         DO jtype = itype, SIZE(potparm%pot, 2)
            IF (potparm%pot(itype, jtype)%pot%rcutsq .GT. rcutsq_max) &
               rcutsq_max = potparm%pot(itype, jtype)%pot%rcutsq
         ENDDO
      ENDDO

! check to make sure all box dimensions are greater than two times this
! value
      mc_par%rcut = rcutsq_max**0.5_dp
      DO itype = 1, 3
         IF (abc(itype) .LT. 2.0_dp*mc_par%rcut) THEN
            lterminate = .TRUE.
         ENDIF
      ENDDO

   END SUBROUTINE find_mc_rcut

! **************************************************************************************************
!> \brief figures out the number of total molecules, the number of atoms in each
!>      molecule, an array with the molecule types, etc...a lot of information
!>      that we need.  I did this because I use multiple force_env (simulation
!>      boxes) for MC, and they don't know about each other.
!> \param force_env the pointer containing all the force environments in the
!>        simulation
!> \param mc_molecule_info the structure that will hold all the information
!>          for the molecule types
!>
!>    Suitable for parallel.
!> \param box_number ...
!> \param coordinates_empty ...
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_determine_molecule_info(force_env, mc_molecule_info, box_number, &
                                         coordinates_empty)

      TYPE(force_env_p_type), DIMENSION(:), POINTER      :: force_env
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      INTEGER, INTENT(IN), OPTIONAL                      :: box_number
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: coordinates_empty

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_determine_molecule_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name
      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: names_init
      INTEGER :: first_mol, iatom, ibox, imol, imolecule, istat, itype, iunique, iunit, jtype, &
         last_mol, natom, natoms_large, nbend, nbond, nboxes, nmol_types, nmolecule, ntorsion, &
         ntypes, skip_box, this_molecule, total
      INTEGER, DIMENSION(:), POINTER                     :: molecule_list
      LOGICAL                                            :: lnew_type
      TYPE(atom_type), DIMENSION(:), POINTER             :: atom_list
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mol_kind_new_list_p_type), DIMENSION(:), &
         POINTER                                         :: molecule_kinds_new
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles

! how many simulation boxes do we have?

      nboxes = SIZE(force_env(:))

! allocate the pointer
      ALLOCATE (mc_molecule_info)
      mc_molecule_info%nboxes = nboxes

! if box_number is present, that box is supposed to be empty,
! so we don't count it in anything
      IF (PRESENT(box_number)) THEN
         skip_box = box_number
      ELSE
         skip_box = 0
      ENDIF

! we need to figure out how many different kinds of molecules we have...
! the fun stuff comes in when box 1 is missing a molecule type...I'll
! distinguish molecules based on names from the .psf files...
! this is only really a problem when we have more than one simulation
! box
      NULLIFY (subsys, molecule_kinds_new, molecule_kind)

      ALLOCATE (particles(1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1845, &
                                       "particles", nboxes)
      ALLOCATE (molecule_kinds_new(1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1848, &
                                       "molecule_kinds_new", nboxes)

      ! find out how many types we have
      ntypes = 0
      DO ibox = 1, nboxes
         IF (ibox == skip_box) CYCLE
         CALL force_env_get(force_env(ibox)%force_env, &
                            subsys=subsys)
         CALL cp_subsys_get(subsys, &
                            molecule_kinds_new=molecule_kinds_new(ibox)%list, &
                            particles=particles(ibox)%list)
         ntypes = ntypes+SIZE(molecule_kinds_new(ibox)%list%els(:))
      ENDDO

      ALLOCATE (names_init(1:ntypes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1864, &
                                       "names_init", ntypes*default_string_length)

! now get the names of all types so we can figure out how many distinct
! types we have
      itype = 1
      natoms_large = 0
      DO ibox = 1, nboxes
         IF (ibox == skip_box) CYCLE
         DO imolecule = 1, SIZE(molecule_kinds_new(ibox)%list%els(:))
            molecule_kind => molecule_kinds_new(ibox)%list%els(imolecule)
            CALL get_molecule_kind(molecule_kind, name=names_init(itype), &
                                   natom=natom)
            IF (natom .GT. natoms_large) natoms_large = natom
            itype = itype+1
         ENDDO
      ENDDO

      nmol_types = 0
      DO itype = 1, ntypes
         lnew_type = .TRUE.
         DO jtype = 1, itype-1
            IF (TRIM(names_init(itype)) .EQ. TRIM(names_init(jtype))) &
               lnew_type = .FALSE.
         ENDDO
         IF (lnew_type) THEN
            nmol_types = nmol_types+1
         ELSE
            names_init(itype) = ''
         ENDIF
      ENDDO

! allocate arrays for the names of the molecules, the number of atoms, and
! the conformational probabilities
      mc_molecule_info%nmol_types = nmol_types
      ALLOCATE (mc_molecule_info%names(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1900, &
                                       "mc_molecule_info%names", nmol_types*default_string_length)
      ALLOCATE (mc_molecule_info%nunits(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1903, &
                                       "mc_molecule_info%nunits", nmol_types*int_size)
      ALLOCATE (mc_molecule_info%conf_prob(1:3, 1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1906, &
                                       "mc_molecule_info%conf_prob", 3*nmol_types*dp_size)
      ALLOCATE (mc_molecule_info%nchains(1:nmol_types, 1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1909, &
                                       "mc_molecule_info%nchains", nboxes*nmol_types*dp_size)
      ALLOCATE (mc_molecule_info%nunits_tot(1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1912, &
                                       "mc_molecule_info%nunits_tot", nboxes*int_size)
      ALLOCATE (mc_molecule_info%atom_names(1:natoms_large, 1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1915, &
                                       "mc_molecule_info%atom_names", natoms_large*nmol_types*default_string_length)
      ALLOCATE (mc_molecule_info%mass(1:natoms_large, 1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1918, &
                                       "mc_molecule_info%mass", natoms_large*nmol_types*dp_size)

! now record all the information for every molecule type
      iunique = 0
      itype = 0
      DO ibox = 1, nboxes
         IF (ibox == skip_box) CYCLE
         DO imolecule = 1, SIZE(molecule_kinds_new(ibox)%list%els(:))
            itype = itype+1
            IF (names_init(itype) .EQ. '') CYCLE
            iunique = iunique+1
            mc_molecule_info%names(iunique) = names_init(itype)
            molecule_kind => molecule_kinds_new(ibox)%list%els(imolecule)
            CALL get_molecule_kind(molecule_kind, natom=mc_molecule_info%nunits(iunique), &
                                   nbond=nbond, nbend=nbend, ntorsion=ntorsion, atom_list=atom_list)

            DO iatom = 1, mc_molecule_info%nunits(iunique)
               atomic_kind => atom_list(iatom)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                    name=mc_molecule_info%atom_names(iatom, iunique), &
                                    mass=mc_molecule_info%mass(iatom, iunique))
            ENDDO

! compute the probabilities of doing any particular kind of conformation change
            total = nbond+nbend+ntorsion

            IF (total == 0) THEN
               mc_molecule_info%conf_prob(:, iunique) = 0.0e0_dp
            ELSE
               mc_molecule_info%conf_prob(1, iunique) = REAL(nbond, dp)/REAL(total, dp)
               mc_molecule_info%conf_prob(2, iunique) = REAL(nbend, dp)/REAL(total, dp)
               mc_molecule_info%conf_prob(3, iunique) = REAL(ntorsion, dp)/REAL(total, dp)
            ENDIF

         ENDDO
      ENDDO

! figure out the empty coordinates
      IF (PRESENT(coordinates_empty)) THEN
         ALLOCATE (coordinates_empty(1:3, 1:mc_molecule_info%nunits(1)), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1959, &
                                          "coordinates_empty", 3*mc_molecule_info%nunits(1)*dp_size)
         DO iunit = 1, mc_molecule_info%nunits(1)
            coordinates_empty(1:3, iunit) = particles(1)%list%els(iunit)%r(1:3)
         ENDDO
      ENDIF

! now we need to figure out the total number of molecules of each kind in
! each box, and the total number of interaction sites in each box
      mc_molecule_info%nchains(:, :) = 0
      DO iunique = 1, nmol_types
         DO ibox = 1, nboxes
            IF (ibox == skip_box) CYCLE
            DO imolecule = 1, SIZE(molecule_kinds_new(ibox)%list%els(:))
               molecule_kind => molecule_kinds_new(ibox)%list%els(imolecule)
               CALL get_molecule_kind(molecule_kind, nmolecule=nmolecule, &
                                      name=name)
               IF (TRIM(name) .NE. mc_molecule_info%names(iunique)) CYCLE
               mc_molecule_info%nchains(iunique, ibox) = mc_molecule_info%nchains(iunique, ibox)+nmolecule
            ENDDO
         ENDDO
      ENDDO
      mc_molecule_info%nchain_total = 0
      DO ibox = 1, nboxes
         mc_molecule_info%nunits_tot(ibox) = 0
         IF (ibox == skip_box) CYCLE
         DO iunique = 1, nmol_types
            mc_molecule_info%nunits_tot(ibox) = mc_molecule_info%nunits_tot(ibox)+ &
                                                mc_molecule_info%nunits(iunique)*mc_molecule_info%nchains(iunique, ibox)
         ENDDO
         mc_molecule_info%nchain_total = mc_molecule_info%nchain_total+SUM(mc_molecule_info%nchains(:, ibox))
      ENDDO

! now we need to figure out which type every molecule is,
! and which force_env (box) it's in
      ALLOCATE (mc_molecule_info%mol_type(1:mc_molecule_info%nchain_total), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1995, &
                                       "mol_type", mc_molecule_info%nchain_total*int_size)
      ALLOCATE (mc_molecule_info%in_box(1:mc_molecule_info%nchain_total), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1998, &
                                       "in_box", mc_molecule_info%nchain_total*int_size)

      last_mol = 0
      DO ibox = 1, nboxes
         IF (ibox == skip_box) CYCLE
         first_mol = last_mol+1
         last_mol = first_mol+SUM(mc_molecule_info%nchains(:, ibox))-1
         mc_molecule_info%in_box(first_mol:last_mol) = ibox
         DO imolecule = 1, SIZE(molecule_kinds_new(ibox)%list%els(:))
            molecule_kind => molecule_kinds_new(ibox)%list%els(imolecule)
            CALL get_molecule_kind(molecule_kind, nmolecule=nmolecule, &
                                   name=name, molecule_list=molecule_list)
! figure out which molecule number this is
            this_molecule = 0
            DO iunique = 1, nmol_types
               IF (TRIM(name) .EQ. TRIM(mc_molecule_info%names(iunique))) THEN
                  this_molecule = iunique
               ENDIF
            ENDDO
            DO imol = 1, SIZE(molecule_list(:))
               mc_molecule_info%mol_type(first_mol+molecule_list(imol)-1) = this_molecule
            ENDDO
         ENDDO
      ENDDO

! get rid of stuff we don't need
      DEALLOCATE (names_init)
      DEALLOCATE (molecule_kinds_new)
      DEALLOCATE (particles)

   END SUBROUTINE MC_DETERMINE_MOLECULE_INFO

! **************************************************************************************************
!> \brief deallocates all the arrays in the mc_molecule_info_type
!> \param mc_molecule_info the structure we wish to deallocate
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_molecule_info_destroy(mc_molecule_info)

      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_molecule_info_destroy', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (mc_molecule_info%nchains)
      DEALLOCATE (mc_molecule_info%nunits)
      DEALLOCATE (mc_molecule_info%mol_type)
      DEALLOCATE (mc_molecule_info%in_box)
      DEALLOCATE (mc_molecule_info%names)
      DEALLOCATE (mc_molecule_info%atom_names)
      DEALLOCATE (mc_molecule_info%conf_prob)
      DEALLOCATE (mc_molecule_info%nunits_tot)
      DEALLOCATE (mc_molecule_info%mass)

      DEALLOCATE (mc_molecule_info)
      NULLIFY (mc_molecule_info)

   END SUBROUTINE mc_molecule_info_destroy

! **************************************************************************************************
!> \brief ...
!> \param mc_par ...
! **************************************************************************************************
   SUBROUTINE mc_input_parameters_check(mc_par)

! **************************************************************************************************
!> \brief accesses the private elements of the mc_molecule_info_type
!> \param mc_molecule_info the structure you want the parameters for
!> \param nmol_types the number of molecule types in the simulation
!>
!>    Suitable for parallel.
!> \author MJM
      TYPE(mc_simpar_type), POINTER                      :: mc_par

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_input_parameters_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imol_type, nboxes, nchain_total, &
                                                            nmol_types
      INTEGER, DIMENSION(:), POINTER                     :: nunits
      LOGICAL                                            :: lhmc
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info

      CALL get_mc_par(mc_par, mc_molecule_info=mc_molecule_info, lhmc=lhmc)

      CALL get_mc_molecule_info(mc_molecule_info, nmol_types=nmol_types, &
                                nboxes=nboxes, nunits=nunits, nchain_total=nchain_total)

! if we're only doing HMC, we can skip these checks
      IF (lhmc) RETURN

! the final value of these arrays needs to be 1.0, otherwise there is
! a chance we won't select a molecule type for a move
      IF (mc_par%pmavbmc_mol(nmol_types) .LT. 0.9999E0_dp) THEN
         CALL cp__b("motion/mc/mc_types.F",2095,'The last value of PMAVBMC_MOL needs to be 1.0')
      ENDIF
      IF (mc_par%pmswap_mol(nmol_types) .LT. 0.9999E0_dp) THEN
         CALL cp__b("motion/mc/mc_types.F",2098,'The last value of PMSWAP_MOL needs to be 1.0')
      ENDIF
      IF (mc_par%pmtraion_mol(nmol_types) .LT. 0.9999E0_dp) THEN
         CALL cp__b("motion/mc/mc_types.F",2101,'The last value of PMTRAION_MOL needs to be 1.0')
      ENDIF
      IF (mc_par%pmtrans_mol(nmol_types) .LT. 0.9999E0_dp) THEN
         CALL cp__b("motion/mc/mc_types.F",2104,'The last value of PMTRANS_MOL needs to be 1.0')
      ENDIF
      IF (mc_par%pmrot_mol(nmol_types) .LT. 0.9999E0_dp) THEN
         CALL cp__b("motion/mc/mc_types.F",2107,'The last value of PMROT_MOL needs to be 1.0')
      ENDIF

! check to make sure we have all the desired values for various

      ! some ensembles require multiple boxes or molecule types
      SELECT CASE (mc_par%ensemble)
      CASE ("GEMC_NPT")
         IF (nmol_types .LE. 1) &
            CALL cp__b("motion/mc/mc_types.F",2116,'Cannot have GEMC-NPT simulation with only one molecule type')
         IF (nboxes .LE. 1) &
            CALL cp__b("motion/mc/mc_types.F",2118,'Cannot have GEMC-NPT simulation with only one box')
      CASE ("GEMC_NVT")
         IF (nboxes .LE. 1) &
            CALL cp__b("motion/mc/mc_types.F",2121,'Cannot have GEMC-NVT simulation with only one box')
      CASE ("TRADITIONAL")
         IF (mc_par%pmswap .GT. 0.0E0_dp) &
            CALL cp__b("motion/mc/mc_types.F",2124,'You cannot do swap moves in a system with only one box')
      CASE ("VIRIAL")
         IF (nchain_total .NE. 2) &
            CALL cp__b("motion/mc/mc_types.F",2127,'You need exactly two molecules in the box to compute the second virial.')
      END SELECT

! can't choose an AVBMC target atom number higher than the number
! of units in that molecule
      DO imol_type = 1, nmol_types
         IF (mc_par%avbmc_atom(imol_type) .GT. nunits(imol_type)) THEN
            CALL cp__b("motion/mc/mc_types.F",2134,'Cannot have avbmc_atom higher than the number of atoms for that molecule type!')
         ENDIF
      ENDDO

   END SUBROUTINE mc_input_parameters_check

END MODULE mc_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains the subroutines for dealing with the mc_env
!> \author MJM Oct. 15-2003
! **************************************************************************************************
MODULE mc_environment_types

   USE force_env_types,                 ONLY: force_env_type
   USE mc_types,                        ONLY: mc_simpar_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
   TYPE mc_environment_type
      INTEGER :: id_nr, ref_count, in_use
      TYPE(mc_simpar_type), POINTER :: mc_par
      TYPE(force_env_type), POINTER :: force_env
   END TYPE mc_environment_type

! **************************************************************************************************
   TYPE mc_environment_p_type
      TYPE(mc_environment_type), POINTER :: mc_env
   END TYPE mc_environment_p_type

! *** Public subroutines and data types ***
   PUBLIC :: mc_environment_type, mc_environment_p_type, &
             set_mc_env, mc_env_create, &
             get_mc_env, mc_env_release

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_environment_types'
   INTEGER, SAVE, PRIVATE :: last_mc_env_id = 0

CONTAINS

! **************************************************************************************************
!> \brief creates and initializes an mc_env
!> \param mc_env the mc_environment you want to create
!>
!>    Suitable for parallel use.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_env_create(mc_env)

      TYPE(mc_environment_type), POINTER                 :: mc_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_env_create', routineP = moduleN//':'//routineN

      ALLOCATE (mc_env)

      last_mc_env_id = last_mc_env_id+1
      mc_env%id_nr = last_mc_env_id
      mc_env%ref_count = 1
      mc_env%in_use = 0

      NULLIFY (mc_env%mc_par)
      NULLIFY (mc_env%force_env)

   END SUBROUTINE mc_env_create

! **************************************************************************************************
!> \brief provides a method for attaching various structures to an mc_env
!> \param mc_env the mc_environment you want to change
!> \param mc_par the mc parameters you want to associate with this mc_env
!> \param force_env the force environment type you want to associate
!>                   with this mc_env
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE set_mc_env(mc_env, mc_par, force_env)

      TYPE(mc_environment_type), POINTER                 :: mc_env
      TYPE(mc_simpar_type), OPTIONAL, POINTER            :: mc_par
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env

      IF (PRESENT(mc_par)) mc_env%mc_par => mc_par
      IF (PRESENT(force_env)) THEN
         mc_env%force_env => force_env
      END IF

   END SUBROUTINE set_mc_env

! **************************************************************************************************
!> \brief provides a method for getting the various structures attached
!>      to an mc_env
!> \param mc_env the mc_environment you want to get information on
!> \param mc_par the mc parameters you want to point to the parameters
!>                associated with this mc_env
!> \param force_env the force environment type you want to point to the
!>                force environment associated with this mc_env
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE get_mc_env(mc_env, mc_par, force_env)

      TYPE(mc_environment_type), POINTER                 :: mc_env
      TYPE(mc_simpar_type), OPTIONAL, POINTER            :: mc_par
      TYPE(force_env_type), OPTIONAL, POINTER            :: force_env

      IF (PRESENT(mc_par)) mc_par => mc_env%mc_par
      IF (PRESENT(force_env)) force_env => mc_env%force_env

   END SUBROUTINE get_mc_env

! **************************************************************************************************
!> \brief releases the given mc env
!> \param mc_env the mc environment to release
!> \author MJM
!> \note
!>      see doc/ReferenceCounting.html
! **************************************************************************************************
   SUBROUTINE mc_env_release(mc_env)
      TYPE(mc_environment_type), POINTER                 :: mc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_env_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mc_env)) THEN
         IF(.NOT.(mc_env%ref_count > 0))CALL cp__a("motion/mc/mc_environment_types.F",128)
         mc_env%ref_count = mc_env%ref_count-1
         IF (mc_env%ref_count == 0) THEN
            mc_env%ref_count = 1
            NULLIFY (mc_env%mc_par)
            NULLIFY (mc_env%force_env)
            mc_env%ref_count = 0
            DEALLOCATE (mc_env)
         END IF
      END IF
      NULLIFY (mc_env)
   END SUBROUTINE mc_env_release

END MODULE mc_environment_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains miscellaneous subroutines used in the Monte Carlo runs,mostly
!>      geared towards changes in coordinates
!> \author MJM
! **************************************************************************************************
MODULE mc_coordinates
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE mathconstants,                   ONLY: pi
   USE mc_types,                        ONLY: get_mc_molecule_info,&
                                              mc_molecule_info_type
   USE message_passing,                 ONLY: mp_bcast
   USE molecule_types_new,              ONLY: molecule_type
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE particle_list_types,             ONLY: particle_list_type
   USE physcon,                         ONLY: angstrom
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PRIVATE :: generate_avbmc_insertion

   PUBLIC :: generate_cbmc_swap_config, &
             get_center_of_mass, mc_coordinate_fold, &
             find_mc_test_molecule, &
             create_discrete_array, &
             check_for_overlap, &
             rotate_molecule

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_coordinates'

CONTAINS

! **************************************************************************************************
!> \brief looks for overlaps (intermolecular distances less than rmin)
!> \param force_env the force environment containing the coordinates
!> \param nchains the number of molecules of each type in the box
!> \param nunits the number of interaction sites for each molecule
!> \param loverlap returns .TRUE. if atoms overlap
!> \param mol_type an array that indicates the type of each molecule
!> \param cell_length the length of the box...if none is specified,
!>                     it uses the cell found in the force_env
!> \param molecule_number if present, just look for overlaps with this
!>            molecule
!>
!>      Suitable for parallel use.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE check_for_overlap(force_env, nchains, nunits, loverlap, mol_type, &
                                cell_length, molecule_number)

      TYPE(force_env_type), POINTER                      :: force_env
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nchains, nunits
      LOGICAL, INTENT(OUT)                               :: loverlap
      INTEGER, DIMENSION(:), INTENT(IN)                  :: mol_type
      REAL(KIND=dp), DIMENSION(1:3), INTENT(IN), &
         OPTIONAL                                        :: cell_length
      INTEGER, INTENT(IN), OPTIONAL                      :: molecule_number

      CHARACTER(len=*), PARAMETER :: routineN = 'check_for_overlap', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, imol, istat, iunit, jmol, &
                                                            jstart, junit, nend, nstart, nunit, &
                                                            nunits_i, nunits_j
      LOGICAL                                            :: lall
      REAL(KIND=dp)                                      :: dist, rmin
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: r
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc, box_length, RIJ
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: oldsys
      TYPE(particle_list_type), POINTER                  :: particles

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (oldsys, particles)

! initialize some stuff
      loverlap = .FALSE.
      rmin = 3.57106767_dp ! 1 angstrom squared

! get the particle coordinates and the cell length
      CALL force_env_get(force_env, cell=cell, subsys=oldsys)
      CALL get_cell(cell, abc=abc)
      CALL cp_subsys_get(oldsys, particles=particles)

      ALLOCATE (r(1:3, 1:MAXVAL(nunits), 1:SUM(nchains)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 106, &
                                       "mass", 3*MAXVAL(nunits)*SUM(nchains)*dp_size)

      IF (PRESENT(cell_length)) THEN
         box_length(1:3) = cell_length(1:3)
      ELSE
         box_length(1:3) = abc(1:3)
      ENDIF

! put the coordinates into an easier matrix to manipulate
      junit = 0
      DO imol = 1, SUM(nchains)
         nunit = nunits(mol_type(imol))
         DO iunit = 1, nunit
            junit = junit+1
            r(1:3, iunit, imol) = particles%els(junit)%r(1:3)
         ENDDO
      ENDDO

! now let's find the LJ energy between all the oxygens and
! the charge interactions
      lall = .TRUE.
      jstart = 1
      IF (PRESENT(molecule_number)) THEN
         lall = .FALSE.
         nstart = molecule_number
         nend = molecule_number
      ELSE
         nstart = 1
         nend = SUM(nchains(:))
      ENDIF
      DO imol = nstart, nend
         IF (lall) jstart = imol+1
         nunits_i = nunits(mol_type(imol))
         DO jmol = jstart, SUM(nchains(:))
            IF (imol == jmol) CYCLE
            nunits_j = nunits(mol_type(jmol))

            DO iunit = 1, nunits_i
               DO junit = 1, nunits_j
! find the minimum image distance
                  RIJ(1) = r(1, iunit, imol)-r(1, junit, jmol)- &
                           box_length(1)*ANINT( &
                           (r(1, iunit, imol)-r(1, junit, jmol))/box_length(1))
                  RIJ(2) = r(2, iunit, imol)-r(2, junit, jmol)- &
                           box_length(2)*ANINT( &
                           (r(2, iunit, imol)-r(2, junit, jmol))/box_length(2))
                  RIJ(3) = r(3, iunit, imol)-r(3, junit, jmol)- &
                           box_length(3)*ANINT( &
                           (r(3, iunit, imol)-r(3, junit, jmol))/box_length(3))

                  dist = RIJ(1)**2+RIJ(2)**2+RIJ(3)**2

                  IF (dist < rmin) THEN
                     loverlap = .TRUE.
                     DEALLOCATE (r)

                     CALL timestop(handle)
                     RETURN
                  ENDIF

               ENDDO
            ENDDO
         ENDDO
      ENDDO

      DEALLOCATE (r)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE check_for_overlap

! **************************************************************************************************
!> \brief calculates the center of mass of a given molecule
!> \param coordinates the coordiantes of the atoms in the molecule
!> \param natom the number of atoms in the molecule
!> \param center_of_mass the coordinates of the center of mass
!> \param mass the mass of the atoms in the molecule
!>
!>    Designed for parallel use.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE get_center_of_mass(coordinates, natom, center_of_mass, &
                                 mass)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: coordinates
      INTEGER, INTENT(IN)                                :: natom
      REAL(KIND=dp), DIMENSION(1:3), INTENT(OUT)         :: center_of_mass
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: mass

      CHARACTER(len=*), PARAMETER :: routineN = 'get_center_of_mass', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom
      REAL(KIND=dp)                                      :: total_mass

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      total_mass = SUM(mass(1:natom))
      center_of_mass(:) = 0.0E0_dp

      DO iatom = 1, natom
         DO i = 1, 3
            center_of_mass(i) = center_of_mass(i)+ &
                                mass(iatom)*coordinates(i, iatom)
         ENDDO
      ENDDO

      center_of_mass(1:3) = center_of_mass(1:3)/total_mass

! end the timing
      CALL timestop(handle)

   END SUBROUTINE get_center_of_mass

! **************************************************************************************************
!> \brief folds all the coordinates into the center simulation box using
!>      a center of mass cutoff
!> \param coordinates the coordiantes of the atoms in the system
!> \param nchains_tot the total number of molecules in the box
!> \param mol_type an array that indicates the type of every molecule in the box
!> \param mass the mass of every atom for all molecule types
!> \param nunits the number of interaction sites for each molecule type
!> \param box_length an array for the lengths of the simulation box sides
!>
!>      Designed for parallel use.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_coordinate_fold(coordinates, nchains_tot, mol_type, mass, nunits, box_length)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: coordinates
      INTEGER, INTENT(IN)                                :: nchains_tot
      INTEGER, DIMENSION(:), INTENT(IN)                  :: mol_type
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: mass
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nunits
      REAL(KIND=dp), DIMENSION(1:3), INTENT(IN)          :: box_length

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_coordinate_fold', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: end_atom, handle, iatom, imolecule, &
                                                            jatom, molecule_type, natoms, &
                                                            start_atom
      REAL(KIND=dp), DIMENSION(1:3)                      :: center_of_mass

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! loop over all molecules
      end_atom = 0
      DO imolecule = 1, nchains_tot
         molecule_type = mol_type(imolecule)
         natoms = nunits(molecule_type)
         start_atom = end_atom+1
         end_atom = start_atom+natoms-1
         CALL get_center_of_mass(coordinates(:, start_atom:end_atom), &
                                 natoms, center_of_mass(:), mass(:, molecule_type))
         DO iatom = 1, natoms
            jatom = iatom+start_atom-1
            coordinates(1, jatom) = coordinates(1, jatom)- &
                                    box_length(1)*FLOOR(center_of_mass(1)/box_length(1))
            coordinates(2, jatom) = coordinates(2, jatom)- &
                                    box_length(2)*FLOOR(center_of_mass(2)/box_length(2))
            coordinates(3, jatom) = coordinates(3, jatom)- &
                                    box_length(3)*FLOOR(center_of_mass(3)/box_length(2))
         ENDDO

      ENDDO

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_coordinate_fold

! **************************************************************************************************
!> \brief takes the last molecule in a force environment and moves it around
!>      to different center of mass positions and orientations, selecting one
!>      based on the rosenbluth weight
!> \param force_env the force environment containing the coordinates
!> \param BETA the value of 1/kT for this simulations, in a.u.
!> \param max_val ...
!> \param min_val ...
!> \param exp_max_val ...
!> \param exp_min_val ...
!> \param nswapmoves the number of desired trial configurations
!> \param rosenbluth_weight the Rosenbluth weight for this set of configs
!> \param start_atom the atom number that the molecule to be swapped starts on
!> \param natoms_tot the total number of interaction sites in the box
!> \param nunits the number of interaction sites for every molecule_type
!> \param nunits_mol ...
!> \param mass the mass for every interaction site of every molecule type
!> \param loverlap the flag that indicates if all of the configs have an
!>                  atomic overlap
!> \param choosen_energy the energy of the choosen config
!> \param old_energy the energy that we subtract from all of the trial
!>        energies to prevent numerical overflows
!> \param ionode indicates if we're on the main CPU
!> \param lremove is this the Rosenbluth weight for a removal box?
!> \param mol_type an array that contains the molecule type for every atom in the box
!> \param nchains the number of molecules of each type in this box
!> \param source the MPI source value, for broadcasts
!> \param group the MPI group value, for broadcasts
!> \param rng_stream the random number stream that we draw from
!> \param avbmc_atom ...
!> \param rmin ...
!> \param rmax ...
!> \param move_type ...
!> \param target_atom ...
!> \par Optional Avbmc Flags
!>      - avbmc_atom: the atom number that serves for the target atom in each
!>        molecule (1 is the first atom in the molecule, etc.)
!>      - rmin: the minimum AVBMC radius for the shell around the target
!>      - rmax: the maximum AVBMC radius for the shell around the target
!>      - move_type: generate configs in the "in" or "out" volume
!>      - target_atom: the number of the avbmc atom in the target molecule
!> \par
!>      Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE generate_cbmc_swap_config(force_env, BETA, max_val, min_val, exp_max_val, &
                                        exp_min_val, nswapmoves, rosenbluth_weight, start_atom, natoms_tot, nunits, nunits_mol, &
                                        mass, loverlap, choosen_energy, old_energy, ionode, lremove, mol_type, nchains, source, &
                                        group, rng_stream, avbmc_atom, rmin, rmax, move_type, target_atom)

      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), INTENT(IN)                          :: BETA, max_val, min_val, exp_max_val, &
                                                            exp_min_val
      INTEGER, INTENT(IN)                                :: nswapmoves
      REAL(KIND=dp), INTENT(OUT)                         :: rosenbluth_weight
      INTEGER, INTENT(IN)                                :: start_atom, natoms_tot
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nunits
      INTEGER, INTENT(IN)                                :: nunits_mol
      REAL(dp), DIMENSION(1:nunits_mol), INTENT(IN)      :: mass
      LOGICAL, INTENT(OUT)                               :: loverlap
      REAL(KIND=dp), INTENT(OUT)                         :: choosen_energy
      REAL(KIND=dp), INTENT(IN)                          :: old_energy
      LOGICAL, INTENT(IN)                                :: ionode, lremove
      INTEGER, DIMENSION(:), INTENT(IN)                  :: mol_type, nchains
      INTEGER, INTENT(IN)                                :: source, group
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      INTEGER, INTENT(IN), OPTIONAL                      :: avbmc_atom
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: rmin, rmax
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: move_type
      INTEGER, INTENT(IN), OPTIONAL                      :: target_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_cbmc_swap_config', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_number, choosen, end_atom, handle, &
                                                            i, iatom, imolecule, imove, istat, &
                                                            molecule_number
      LOGICAL                                            :: all_overlaps
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: loverlap_array
      REAL(KIND=dp)                                      :: bias_energy, exponent, rand, &
                                                            total_running_weight
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: boltz_weights, trial_energy
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r_old
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: r
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc, center_of_mass, diff, r_insert
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: oldsys
      TYPE(particle_list_type), POINTER                  :: particles

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (oldsys)
! get the particle coordinates and the cell length
      CALL force_env_get(force_env, cell=cell, subsys=oldsys)
      CALL get_cell(cell, abc=abc)
      CALL cp_subsys_get(oldsys, particles=particles)

! do some checking to make sure we have all the data we need
      IF (PRESENT(avbmc_atom)) THEN
         IF (.NOT. PRESENT(rmin) .OR. .NOT. PRESENT(rmax) .OR. &
             .NOT. PRESENT(move_type) .OR. .NOT. PRESENT(target_atom)) THEN
            CALL cp__b("motion/mc/mc_coordinates.F",387,'AVBMC swap move is missing information!')
         ENDIF
      ENDIF

      ALLOCATE (r_old(1:3, 1:natoms_tot), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 392, &
                                       "r_old", 3*natoms_tot*dp_size)
      ALLOCATE (r(1:3, 1:natoms_tot, 1:nswapmoves), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 395, &
                                       "r", 3*natoms_tot*nswapmoves*dp_size)
      ALLOCATE (trial_energy(1:nswapmoves), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 398, &
                                       "trial_energy", nswapmoves*dp_size)
      ALLOCATE (boltz_weights(1:nswapmoves), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 401, &
                                       "boltz_weights", nswapmoves*dp_size)
      ALLOCATE (loverlap_array(1:nswapmoves), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 404, &
                                       "loverlap_array", nswapmoves)

! initialize the arrays that need it
      loverlap_array(:) = .FALSE.
      loverlap = .FALSE.
      boltz_weights(:) = 0.0_dp
      trial_energy(:) = 0.0_dp
      r(:, :, :) = 0.0_dp
      choosen_energy = 0.0_dp
      rosenbluth_weight = 0.0_dp

! save the positions of the molecules
      DO imove = 1, nswapmoves
         DO iatom = 1, natoms_tot
            r(1:3, iatom, imove) = particles%els(iatom)%r(1:3)
         ENDDO
      ENDDO

! save the remove coordinates
      DO iatom = 1, natoms_tot
         r_old(1:3, iatom) = r(1:3, iatom, 1)
      ENDDO

! figure out the numbers of the first and last atoms in the molecule
      end_atom = start_atom+nunits_mol-1
! figure out which molecule number we're on
      molecule_number = 0
      atom_number = 1
      DO imolecule = 1, SUM(nchains(:))
         IF (atom_number == start_atom) THEN
            molecule_number = imolecule
            EXIT
         ENDIF
         atom_number = atom_number+nunits(mol_type(imolecule))
      ENDDO
      IF (molecule_number == 0) CALL cp_abort(cp__l("motion/mc/mc_coordinates.F",440), &
                                              'CBMC swap move cannot find which molecule number it needs')

      IF (lremove) THEN
         CALL check_for_overlap(force_env, nchains, nunits, loverlap_array(1), &
                                mol_type)
         CALL mp_bcast(loverlap_array(1), source, group)

         IF (loverlap_array(1)) THEN
            IF (ionode) THEN
               WRITE (*, *) start_atom, end_atom, natoms_tot
               DO iatom = 1, natoms_tot
                  WRITE (*, *) r(1:3, iatom, 1)
               ENDDO
            ENDIF
            CALL cp__b("motion/mc/mc_coordinates.F",455,'CBMC swap move found an overlap in the old config')
         ENDIF
      ENDIF

      DO imove = 1, nswapmoves

! drop into serial
         IF (ionode) THEN

            IF (PRESENT(avbmc_atom)) THEN
! find an AVBMC insertion point
               CALL generate_avbmc_insertion(rmin, rmax, &
                                             r_old(1:3, target_atom), &
                                             move_type, r_insert(:), abc(:), rng_stream)

               DO i = 1, 3
                  diff(i) = r_insert(i)-r_old(i, start_atom+avbmc_atom-1)
               ENDDO

            ELSE
! find a new insertion point somewhere in the box
               DO i = 1, 3
                  rand = next_random_number(rng_stream)
                  r_insert(i) = rand*abc(i)
               ENDDO

! find the center of mass of the insertion molecule
               CALL get_center_of_mass(r(:, start_atom:end_atom, imove), nunits_mol, &
                                       center_of_mass(:), mass(:))

! move the molecule to the insertion point

               DO i = 1, 3
                  diff(i) = r_insert(i)-center_of_mass(i)
               ENDDO

            ENDIF

            DO iatom = start_atom, end_atom
               r(1:3, iatom, imove) = r(1:3, iatom, imove)+diff(1:3)
            ENDDO

! rotate the molecule...this routine is only made for serial use
            CALL rotate_molecule(r(:, start_atom:end_atom, imove), mass(:), &
                                 nunits_mol, rng_stream)

            IF (imove == 1 .AND. lremove) THEN
               DO iatom = 1, natoms_tot
                  r(1:3, iatom, 1) = r_old(1:3, iatom)
               ENDDO
            ENDIF

         ENDIF

         CALL mp_bcast(r(:, :, imove), source, group)

! calculate the energy and boltzman weight of the config
         DO iatom = start_atom, end_atom
            particles%els(iatom)%r(1:3) = r(1:3, iatom, imove)
         ENDDO

         CALL check_for_overlap(force_env, nchains, nunits, loverlap_array(imove), &
                                mol_type, molecule_number=molecule_number)
         IF (loverlap_array(imove)) THEN
            boltz_weights(imove) = 0.0_dp
            CYCLE
         ENDIF

         CALL force_env_calc_energy_force(force_env, calc_force=.FALSE.)
         CALL force_env_get(force_env, &
                            potential_energy=bias_energy)

         trial_energy(imove) = (bias_energy-old_energy)
         exponent = -BETA*trial_energy(imove)

         IF (exponent .GT. exp_max_val) THEN
            boltz_weights(imove) = max_val
         ELSEIF (exponent .LT. exp_min_val) THEN
            boltz_weights(imove) = min_val
         ELSE
            boltz_weights(imove) = EXP(exponent)
         ENDIF

      ENDDO

! now we need to pick a configuration based on the Rosenbluth weight,
! which is just the sum of the Boltzmann weights
      rosenbluth_weight = SUM(boltz_weights(:))
      IF (rosenbluth_weight .EQ. 0.0_dp .AND. lremove) THEN
! should never have 0.0 for an old weight...causes a divide by zero
! in the acceptance rule
         IF (ionode) THEN
            WRITE (*, *) boltz_weights(1:nswapmoves)
            WRITE (*, *) start_atom, end_atom, lremove
            WRITE (*, *) loverlap_array(1:nswapmoves)
            WRITE (*, *) natoms_tot
            WRITE (*, *)
            DO iatom = 1, natoms_tot
               WRITE (*, *) r(1:3, iatom, 1)*angstrom
            ENDDO
         ENDIF
         CALL cp__b("motion/mc/mc_coordinates.F",556,'CBMC swap move found a bad old weight')
      ENDIF
      all_overlaps = .TRUE.
      total_running_weight = 0.0E0_dp
      choosen = 0
      IF (ionode) THEN
         rand = next_random_number(rng_stream)
!         CALL random_number(rand)
      ENDIF
      CALL mp_bcast(rand, source, group)
      DO imove = 1, nswapmoves
         IF (loverlap_array(imove)) CYCLE
         all_overlaps = .FALSE.
         total_running_weight = total_running_weight+boltz_weights(imove)
         IF (total_running_weight .GE. rand*rosenbluth_weight) THEN
            choosen = imove
            EXIT
         ENDIF
      ENDDO

      IF (all_overlaps) THEN
         loverlap = .TRUE.

! if this is an old configuration, we always choose the first one...
! this should never be the case, but I'm testing something
         IF (lremove) THEN
            IF (ionode) THEN
               WRITE (*, *) boltz_weights(1:nswapmoves)
               WRITE (*, *) start_atom, end_atom, lremove
               WRITE (*, *) loverlap_array(1:nswapmoves)
               DO iatom = 1, natoms_tot
                  WRITE (*, *) r(1:3, iatom, 1)
               ENDDO
            ENDIF
            CALL cp__b("motion/mc/mc_coordinates.F",590,'CBMC swap move found all overlaps for the remove config')
         ENDIF

         DEALLOCATE (r_old)
         DEALLOCATE (r)
         DEALLOCATE (trial_energy)
         DEALLOCATE (boltz_weights)
         DEALLOCATE (loverlap_array)
         CALL timestop(handle)
         RETURN
      ENDIF

! make sure a configuration was choosen
      IF (choosen == 0) &
         CALL cp__b("motion/mc/mc_coordinates.F",604,'CBMC swap move failed to select config')

! if this is an old configuration, we always choose the first one
      IF (lremove) choosen = 1

! set the energy for the configuration
      choosen_energy = trial_energy(choosen)

! copy the coordinates to the force environment
      DO iatom = 1, natoms_tot
         particles%els(iatom)%r(1:3) = r(1:3, iatom, choosen)
      ENDDO

      DEALLOCATE (r_old)
      DEALLOCATE (r)
      DEALLOCATE (trial_energy)
      DEALLOCATE (boltz_weights)
      DEALLOCATE (loverlap_array)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE generate_cbmc_swap_config

! **************************************************************************************************
!> \brief rotates a molecule randomly around the center of mass,
!>      sequentially in x, y, and z directions
!> \param r the coordinates of the molecule to rotate
!> \param mass the mass of all the atoms in the molecule
!> \param natoms the number of atoms in the molecule
!> \param rng_stream the stream we pull random numbers from
!>
!>    Use only in serial.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE rotate_molecule(r, mass, natoms, rng_stream)

      INTEGER, INTENT(IN)                                :: natoms
      REAL(KIND=dp), DIMENSION(1:natoms), INTENT(IN)     :: mass
      REAL(KIND=dp), DIMENSION(1:3, 1:natoms), &
         INTENT(INOUT)                                   :: r
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'rotate_molecule', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iunit
      REAL(KIND=dp)                                      :: cosdg, dgamma, rand, rx, rxnew, ry, &
                                                            rynew, rz, rznew, sindg
      REAL(KIND=dp), DIMENSION(1:3)                      :: center_of_mass

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! find the center of mass of the molecule
      CALL get_center_of_mass(r(:, :), natoms, center_of_mass(:), mass(:))

! call a random number to figure out how far we're moving
      rand = next_random_number(rng_stream)
      dgamma = pi*(rand-0.5E0_dp)*2.0E0_dp

! *** set up the rotation matrix ***

      cosdg = COS(dgamma)
      sindg = SIN(dgamma)

! ***    ROTATE UNITS OF I AROUND X-AXIS ***

      DO iunit = 1, natoms
         ry = r(2, iunit)-center_of_mass(2)
         rz = r(3, iunit)-center_of_mass(3)
         rynew = cosdg*ry+sindg*rz
         rznew = cosdg*rz-sindg*ry

         r(2, iunit) = rynew+center_of_mass(2)
         r(3, iunit) = rznew+center_of_mass(3)

      ENDDO

! ***    ROTATE UNITS OF I AROUND y-AXIS ***

      DO iunit = 1, natoms
         rx = r(1, iunit)-center_of_mass(1)
         rz = r(3, iunit)-center_of_mass(3)
         rxnew = cosdg*rx+sindg*rz
         rznew = cosdg*rz-sindg*rx

         r(1, iunit) = rxnew+center_of_mass(1)
         r(3, iunit) = rznew+center_of_mass(3)

      ENDDO

! ***    ROTATE UNITS OF I AROUND z-AXIS ***

      DO iunit = 1, natoms
         rx = r(1, iunit)-center_of_mass(1)
         ry = r(2, iunit)-center_of_mass(2)
         rxnew = cosdg*rx+sindg*ry
         rynew = cosdg*ry-sindg*rx

         r(1, iunit) = rxnew+center_of_mass(1)
         r(2, iunit) = rynew+center_of_mass(2)

      ENDDO

! end the timing
      CALL timestop(handle)

   END SUBROUTINE rotate_molecule

! **************************************************************************************************
!> \brief selects a molecule at random to perform a MC move on...you can specify
!>      the box the molecule should be in, its type, both, or neither
!> \param mc_molecule_info the structure that contains some global molecule data
!> \param start_atom the number of the first atom in the choosen molecule in relation
!>        to the force_env it's in
!> \param box_number the box the choosen molecule is in
!> \param molecule_type the type of molecule the choosen molecule is
!> \param rng_stream the stream we pull random numbers from
!> \param box if present, tells the routine which box to grab a molecule from
!> \param molecule_type_old if present, tells the routine which molecule type to select from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE find_mc_test_molecule(mc_molecule_info, start_atom, &
                                    box_number, molecule_type, rng_stream, box, molecule_type_old)

      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      INTEGER, INTENT(OUT)                               :: start_atom, box_number, molecule_type
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      INTEGER, INTENT(IN), OPTIONAL                      :: box, molecule_type_old

      CHARACTER(LEN=*), PARAMETER :: routineN = 'find_mc_test_molecule', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ibox, imol_type, imolecule, &
                                                            jbox, molecule_number, nchains_tot, &
                                                            start_mol
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, nunits
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      REAL(KIND=dp)                                      :: rand

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (nunits, mol_type, nchains)
      CALL get_mc_molecule_info(mc_molecule_info, nchains=nchains, nunits=nunits, &
                                mol_type=mol_type)

! initialize the outgoing variables
      start_atom = 0
      box_number = 0
      molecule_type = 0

      IF (PRESENT(box) .AND. PRESENT(molecule_type_old)) THEN
! only need to find the atom number the molecule starts on
         rand = next_random_number(rng_stream)
         molecule_number = CEILING(rand*REAL(nchains(molecule_type_old, box)))

         start_mol = 1
         DO jbox = 1, box-1
            start_mol = start_mol+SUM(nchains(:, jbox))
         ENDDO

! adjust to take into account molecules of other types in the box
         DO imol_type = 1, molecule_type_old-1
            molecule_number = molecule_number+nchains(imol_type, box)
         ENDDO

         start_atom = 1
         DO imolecule = 1, molecule_number-1
            start_atom = start_atom+nunits(mol_type(start_mol+imolecule-1))
         ENDDO

      ELSEIF (PRESENT(box)) THEN
! any molecule in box...need to find molecule type and start atom
         rand = next_random_number(rng_stream)
         molecule_number = CEILING(rand*REAL(SUM(nchains(:, box))))

         start_mol = 1
         DO jbox = 1, box-1
            start_mol = start_mol+SUM(nchains(:, jbox))
         ENDDO

         molecule_type = mol_type(start_mol+molecule_number-1)

! now the starting atom
         start_atom = 1
         DO imolecule = 1, molecule_number-1
            start_atom = start_atom+nunits(mol_type(start_mol+imolecule-1))
         ENDDO

      ELSEIF (PRESENT(molecule_type_old)) THEN
! any molecule of type molecule_type_old...need to find box number and start atom
         rand = next_random_number(rng_stream)
         molecule_number = CEILING(rand*REAL(SUM(nchains(molecule_type_old, :))))

! find which box it's in
         nchains_tot = 0
         DO ibox = 1, SIZE(nchains(molecule_type_old, :))
            IF (molecule_number .LE. nchains(molecule_type_old, ibox)) THEN
               box_number = ibox
               EXIT
            ENDIF
            molecule_number = molecule_number-nchains(molecule_type_old, ibox)
         ENDDO

         start_mol = 1
         DO jbox = 1, box_number-1
            start_mol = start_mol+SUM(nchains(:, jbox))
         ENDDO

! now find the starting atom number
         DO imol_type = 1, molecule_type_old-1
            molecule_number = molecule_number+nchains(imol_type, box_number)
         ENDDO
         start_atom = 1
         DO imolecule = 1, molecule_number-1
            start_atom = start_atom+nunits(mol_type(start_mol+imolecule-1))
         ENDDO

      ELSE
! no restrictions...need to find all pieces of data
         nchains_tot = 0
         DO ibox = 1, SIZE(nchains(1, :))
            nchains_tot = nchains_tot+SUM(nchains(:, ibox))
         ENDDO
         rand = next_random_number(rng_stream)
         molecule_number = CEILING(rand*REAL(nchains_tot))

         molecule_type = mol_type(molecule_number)

! now which box it's in
         DO ibox = 1, SUM(nchains(1, :))
            IF (molecule_number .LE. SUM(nchains(:, ibox))) THEN
               box_number = ibox
               EXIT
            ENDIF
            molecule_number = molecule_number-SUM(nchains(:, ibox))
         ENDDO

! now find the starting atom number
         start_mol = 1
         DO jbox = 1, box_number-1
            start_mol = start_mol+SUM(nchains(:, jbox))
         ENDDO
         start_atom = 1
         DO imolecule = 1, molecule_number-1
            start_atom = start_atom+nunits(mol_type(start_mol+imolecule-1))
         ENDDO

      ENDIF

! make sure things are good
      IF (PRESENT(box)) box_number = box
      IF (PRESENT(molecule_type_old)) molecule_type = molecule_type_old

      IF(.NOT.(start_atom /= 0))CALL cp__a("motion/mc/mc_coordinates.F",862)
      IF(.NOT.(box_number /= 0))CALL cp__a("motion/mc/mc_coordinates.F",863)
      IF(.NOT.(molecule_type /= 0))CALL cp__a("motion/mc/mc_coordinates.F",864)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE find_mc_test_molecule

! **************************************************************************************************
!> \brief generates an array that tells us which sides of the simulation
!>      cell we can increase or decrease using a discrete volume move
!> \param cell the lengths of the sides of the cell
!> \param discrete_array the array that indicates which sides we can move
!> \param step_size the size of the discrete volume move
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE create_discrete_array(cell, discrete_array, step_size)

! 1 is for increase, 2 is for decrease
! 1 is for "yes, we can do the move", 0 is for no

      REAL(dp), DIMENSION(1:3), INTENT(IN)               :: cell
      INTEGER, DIMENSION(1:3, 1:2), INTENT(OUT)          :: discrete_array
      REAL(dp), INTENT(IN)                               :: step_size

      INTEGER                                            :: iside
      REAL(dp)                                           :: high_value, length1, length2, low_value

      discrete_array(:, :) = 0

      length1 = ABS(cell(1)-cell(2))
      length2 = ABS(cell(2)-cell(3))

! now let's figure out all the different cases
      IF (length1 .LT. 0.01_dp*step_size .AND. &
          length2 .LT. 0.01_dp*step_size) THEN
! all sides are equal, so we can move up or down
         discrete_array(1:3, 1) = 1
         discrete_array(1:3, 2) = 1
      ELSE

! find the low value and the high value
         high_value = -1.0_dp
         low_value = cell(1)*cell(2)*cell(3)
         DO iside = 1, 3
            IF (cell(iside) .LT. low_value) low_value = cell(iside)
            IF (cell(iside) .GT. high_value) high_value = cell(iside)
         ENDDO
         DO iside = 1, 3
! now we see if the value is a high value or a low value...it can only be
! one of the two
            IF (ABS(cell(iside)-low_value) .LT. 0.01_dp*step_size) THEN
! low value, we can only increase the cell size
               discrete_array(iside, 1) = 1
               discrete_array(iside, 2) = 0
            ELSE
! high value, we can only decrease the cell size
               discrete_array(iside, 1) = 0
               discrete_array(iside, 2) = 1
            ENDIF
         ENDDO
      ENDIF

   END SUBROUTINE create_discrete_array

! **************************************************************************************************
!> \brief generates an insertion point in either the "in" or the "out" volume
!>      of a target atom, where the "in" volume is a shell with inner radius
!>       rmin and outer radius rmax
!> \param rmin the minimum AVBMC radius for the shell around the target
!> \param rmax the maximum AVBMC radius for the shell around the target
!> \param r_target the coordinates of the target atom
!> \param move_type generate configs in the "in" or "out" volume
!> \param r_insert the output insertion site
!> \param abc the lengths of the sides of the simulation box
!> \param rng_stream the random number stream that we draw from
!>
!>      Use only in serial.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE generate_avbmc_insertion(rmin, rmax, r_target, &
                                       move_type, r_insert, abc, rng_stream)

      REAL(KIND=dp), INTENT(IN)                          :: rmin, rmax
      REAL(KIND=dp), DIMENSION(1:3), INTENT(IN)          :: r_target
      CHARACTER(LEN=*), INTENT(IN)                       :: move_type
      REAL(KIND=dp), DIMENSION(1:3), INTENT(OUT)         :: r_insert
      REAL(KIND=dp), DIMENSION(1:3), INTENT(IN)          :: abc
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      INTEGER                                            :: i
      REAL(dp)                                           :: dist, eta_1, eta_2, eta_sq, rand
      REAL(dp), DIMENSION(1:3)                           :: RIJ

      r_insert(1:3) = 0.0_dp

      IF (move_type == 'in') THEN
! generate a random unit vector, from Allen and Tildesley
         DO
            eta_1 = next_random_number(rng_stream)
            eta_2 = next_random_number(rng_stream)
            eta_sq = eta_1**2+eta_2**2
            IF (eta_sq .LT. 1.0_dp) THEN
               r_insert(1) = 2.0_dp*eta_1*SQRT(1.0_dp-eta_sq)
               r_insert(2) = 2.0_dp*eta_2*SQRT(1.0_dp-eta_sq)
               r_insert(3) = 1.0_dp-2.0_dp*eta_sq
               EXIT
            ENDIF
         ENDDO

! now scale that vector to be within the "in" region
         rand = next_random_number(rng_stream)
         r_insert(1:3) = r_insert(1:3)*(rand*(rmax**3-rmin**3)+rmin**3)** &
                         (1.0_dp/3.0_dp)

         r_insert(1:3) = r_target(1:3)+r_insert(1:3)
      ELSE

! find a new insertion point somewhere in the box
         DO
            DO i = 1, 3
               rand = next_random_number(rng_stream)
               r_insert(i) = rand*abc(i)
            ENDDO

! make sure it's not in the "in" region
            RIJ(1) = r_insert(1)-r_target(1)-abc(1)* &
                     ANINT((r_insert(1)-r_target(1))/abc(1))
            RIJ(2) = r_insert(2)-r_target(2)-abc(2)* &
                     ANINT((r_insert(2)-r_target(2))/abc(2))
            RIJ(3) = r_insert(3)-r_target(3)-abc(3)* &
                     ANINT((r_insert(3)-r_target(3))/abc(3))

            dist = RIJ(1)**2+RIJ(2)**2+RIJ(3)**2

            IF (dist .LT. rmin**2 .OR. dist .GT. rmax**2) THEN
               EXIT
            ENDIF

         ENDDO
      ENDIF

   END SUBROUTINE generate_avbmc_insertion

END MODULE mc_coordinates

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief the various moves in Monte Carlo (MC) simulations, including
!>      change of internal conformation, translation of a molecule, rotation
!>      of a molecule, and changing the size of the simulation box
!> \par History
!>      none
!> \author Matthew J. McGrath  (10.16.2003)
! **************************************************************************************************
MODULE mc_moves
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_clone,&
                                              cell_create,&
                                              cell_release,&
                                              cell_type,&
                                              get_cell
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type,&
                                              use_fist_force
   USE global_types,                    ONLY: global_environment_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              dp_size
   USE mathconstants,                   ONLY: pi
   USE mc_coordinates,                  ONLY: check_for_overlap,&
                                              create_discrete_array,&
                                              generate_cbmc_swap_config,&
                                              get_center_of_mass
   USE mc_types,                        ONLY: get_mc_molecule_info,&
                                              get_mc_par,&
                                              mc_ekin_type,&
                                              mc_molecule_info_type,&
                                              mc_moves_type,&
                                              mc_simpar_type
   USE md_run,                          ONLY: qs_mol_dyn
   USE message_passing,                 ONLY: mp_bcast
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE molecule_kind_types,             ONLY: bend_type,&
                                              bond_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type,&
                                              torsion_type
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE particle_list_types,             ONLY: particle_list_type
   USE physcon,                         ONLY: angstrom
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PRIVATE  :: change_bond_angle, change_bond_length, depth_first_search, &
      change_dihedral

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_moves'

   PUBLIC :: mc_conformation_change, mc_molecule_translation, &
             mc_molecule_rotation, mc_volume_move, mc_avbmc_move, &
             mc_hmc_move

CONTAINS

! **************************************************************************************************
!> \brief essentially performs a depth-first search of the molecule structure
!>      to find all atoms connected to a specific atom excluding one branch...
!>      for instance, if water is labelled 1-2-3 for O-H-H, calling this
!>      routine with current_atom=1,avoid_atom=2 returns the array
!>      atom=(0,0,1)
!> \param current_atom the atom whose connections we're looking at
!> \param avoid_atom the atom whose direction the search is not supposed to go
!> \param connectivity an array telling us the neigbors of all atoms
!> \param atom the array that tells us if one can get to a given atom by
!>        starting at current_atom and not going through avoid_atom...0 is no,
!>        1 is yes
!> \author MJM
! **************************************************************************************************
   RECURSIVE SUBROUTINE depth_first_search(current_atom, avoid_atom, &
                                           connectivity, atom)

      INTEGER, INTENT(IN)                                :: current_atom, avoid_atom
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: connectivity
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: atom

      INTEGER                                            :: iatom

      DO iatom = 1, 6
         IF (connectivity(iatom, current_atom) .NE. 0) THEN
            IF (connectivity(iatom, current_atom) .NE. avoid_atom) THEN
               atom(connectivity(iatom, current_atom)) = 1
               CALL depth_first_search(connectivity(iatom, current_atom), &
                                       current_atom, connectivity, atom)
            ENDIF
         ELSE
            RETURN
         ENDIF
      ENDDO

   END SUBROUTINE depth_first_search

! **************************************************************************************************
!> \brief performs either a bond or angle change move for a given molecule
!> \param mc_par the mc parameters for the force env
!> \param force_env the force environment used in the move
!> \param bias_env the force environment used to bias the move, if any (it may
!>            be null if lbias=.false. in mc_par)
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected
!> \param move_updates the structure that keeps track of how many moves have
!>               been accepted/rejected since the last time the displacements
!>               were updated
!> \param start_atom the number of the molecule's first atom, assuming the rest
!>        of the atoms follow sequentially
!> \param molecule_type the type of the molecule we're moving
!> \param box_number the box the molecule is in
!> \param bias_energy the biased energy of the system before the move
!> \param move_type dictates what kind of conformational change we do
!> \param lreject set to .true. if there is an overlap
!> \param rng_stream the random number stream that we draw from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_conformation_change(mc_par, force_env, bias_env, moves, &
                                     move_updates, start_atom, molecule_type, box_number, &
                                     bias_energy, move_type, lreject, &
                                     rng_stream)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env, bias_env
      TYPE(mc_moves_type), POINTER                       :: moves, move_updates
      INTEGER, INTENT(IN)                                :: start_atom, molecule_type, box_number
      REAL(KIND=dp), INTENT(INOUT)                       :: bias_energy
      CHARACTER(LEN=*), INTENT(IN)                       :: move_type
      LOGICAL, INTENT(OUT)                               :: lreject
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_conformation_change', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_string_length)                   :: name
      CHARACTER(default_string_length), DIMENSION(:), &
         POINTER                                         :: names
      INTEGER :: atom_number, end_atom, end_mol, group, handle, imol_type, imolecule, ipart, &
         istat, jbox, molecule_number, nunits_mol, source, start_mol
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, nunits
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      LOGICAL                                            :: ionode, lbias, loverlap
      REAL(KIND=dp)                                      :: BETA, bias_energy_new, bias_energy_old, &
                                                            dis_length, exp_max_val, exp_min_val, &
                                                            rand, value, w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r_new, r_old
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind, molecule_kind_test
      TYPE(particle_list_type), POINTER                  :: particles

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! nullify some pointers
      NULLIFY (particles, subsys, molecule_kinds_new, molecule_kind, &
               molecule_kind_test)

! get a bunch of stuff from mc_par
      CALL get_mc_par(mc_par, lbias=lbias, mc_molecule_info=mc_molecule_info, &
                      BETA=BETA, exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, group=group, source=source, ionode=ionode)
      CALL get_mc_molecule_info(mc_molecule_info, nchains=nchains, nunits=nunits, &
                                mol_type=mol_type, names=names)

! do some allocation
      nunits_mol = nunits(molecule_type)
      ALLOCATE (r_old(1:3, 1:nunits_mol), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 189, &
                                       "r", 3*nunits_mol*dp_size)
      ALLOCATE (r_new(1:3, 1:nunits_mol), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 192, &
                                       "r_new", 3*nunits_mol*dp_size)

! find out some bounds for mol_type
      start_mol = 1
      DO jbox = 1, box_number-1
         start_mol = start_mol+SUM(nchains(:, jbox))
      ENDDO
      end_mol = start_mol+SUM(nchains(:, box_number))-1

! figure out which molecule number we are
      end_atom = start_atom+nunits_mol-1
      molecule_number = 0
      atom_number = 1
      DO imolecule = 1, SUM(nchains(:, box_number))
         IF (atom_number == start_atom) THEN
            molecule_number = imolecule
            EXIT
         ENDIF
         atom_number = atom_number+nunits(mol_type(imolecule+start_mol-1))
      ENDDO
      IF (molecule_number == 0) CALL cp__b("motion/mc/mc_moves.F",213,'Cannot find the molecule number')

! are we biasing this move?
      IF (lbias) THEN

! grab the coordinates
         CALL force_env_get(bias_env, subsys=subsys)
! save the energy
         bias_energy_old = bias_energy

      ELSE

! grab the coordinates
         CALL force_env_get(force_env, subsys=subsys)
      ENDIF

! now find the molecule type associated with this guy
      CALL cp_subsys_get(subsys, &
                         particles=particles, molecule_kinds_new=molecule_kinds_new)
      DO imol_type = 1, SIZE(molecule_kinds_new%els(:))
         molecule_kind_test => molecule_kinds_new%els(imol_type)
         CALL get_molecule_kind(molecule_kind_test, name=name)
         IF (TRIM(ADJUSTL(name)) == TRIM(ADJUSTL(names(molecule_type)))) THEN
            molecule_kind => molecule_kinds_new%els(imol_type)
            EXIT
         ENDIF
      ENDDO

! save the coordinates
      DO ipart = start_atom, end_atom
         r_old(1:3, ipart-start_atom+1) = particles%els(ipart)%r(1:3)
      ENDDO

      IF (.NOT. ASSOCIATED(molecule_kind)) CALL cp__b("motion/mc/mc_moves.F",246,'Cannot find the molecule type')
! do the move
      IF (move_type == 'bond') THEN

! record the attempt
         moves%bond%attempts = moves%bond%attempts+1
         move_updates%bond%attempts = move_updates%bond%attempts+1
         moves%bias_bond%attempts = moves%bias_bond%attempts+1
         move_updates%bias_bond%attempts = move_updates%bias_bond%attempts+1
         IF (.NOT. lbias) THEN
            moves%bond%qsuccesses = moves%bond%qsuccesses+1
            move_updates%bond%qsuccesses = &
               move_updates%bond%qsuccesses+1
            moves%bias_bond%qsuccesses = moves%bias_bond%qsuccesses+1
            move_updates%bias_bond%qsuccesses = &
               move_updates%bias_bond%qsuccesses+1
         ENDIF

! do the move
         CALL change_bond_length(r_old, r_new, mc_par, molecule_type, &
                                 molecule_kind, dis_length, particles, rng_stream)

      ELSEIF (move_type == 'angle') THEN

! record the attempt
         moves%angle%attempts = moves%angle%attempts+1
         move_updates%angle%attempts = move_updates%angle%attempts+1
         moves%bias_angle%attempts = moves%bias_angle%attempts+1
         move_updates%bias_angle%attempts = move_updates%bias_angle%attempts+1
         IF (.NOT. lbias) THEN
            moves%angle%qsuccesses = moves%angle%qsuccesses+1
            move_updates%angle%qsuccesses = &
               move_updates%angle%qsuccesses+1
            moves%bias_angle%qsuccesses = moves%bias_angle%qsuccesses+1
            move_updates%bias_angle%qsuccesses = &
               move_updates%bias_angle%qsuccesses+1
         ENDIF

! do the move
         CALL change_bond_angle(r_old, r_new, mc_par, molecule_type, &
                                molecule_kind, particles, rng_stream)
         dis_length = 1.0E0_dp
      ELSE
! record the attempt
         moves%dihedral%attempts = moves%dihedral%attempts+1
         move_updates%dihedral%attempts = move_updates%dihedral%attempts+1
         moves%bias_dihedral%attempts = moves%bias_dihedral%attempts+1
         move_updates%bias_dihedral%attempts = move_updates%bias_dihedral%attempts+1
         IF (.NOT. lbias) THEN
            moves%dihedral%qsuccesses = moves%dihedral%qsuccesses+1
            move_updates%dihedral%qsuccesses = &
               move_updates%dihedral%qsuccesses+1
            moves%bias_dihedral%qsuccesses = moves%bias_dihedral%qsuccesses+1
            move_updates%bias_dihedral%qsuccesses = &
               move_updates%bias_dihedral%qsuccesses+1
         ENDIF

! do the move
         CALL change_dihedral(r_old, r_new, mc_par, molecule_type, &
                              molecule_kind, particles, rng_stream)
         dis_length = 1.0E0_dp

      ENDIF

! set the coordinates
      DO ipart = start_atom, end_atom
         particles%els(ipart)%r(1:3) = r_new(1:3, ipart-start_atom+1)
      ENDDO

! check for overlap
      lreject = .FALSE.
      IF (lbias) THEN
         CALL check_for_overlap(bias_env, nchains(:, box_number), &
                                nunits(:), loverlap, mol_type(start_mol:end_mol), &
                                molecule_number=molecule_number)
      ELSE
         CALL check_for_overlap(force_env, nchains(:, box_number), &
                                nunits(:), loverlap, mol_type(start_mol:end_mol), &
                                molecule_number=molecule_number)
         IF (loverlap) lreject = .TRUE.
      ENDIF

! if we're biasing classical, check for acceptance
      IF (lbias) THEN

! here's where we bias the moves

         IF (loverlap) THEN
            w = 0.0E0_dp
         ELSE
            CALL force_env_calc_energy_force(bias_env, calc_force=.FALSE.)
            CALL force_env_get(bias_env, &
                               potential_energy=bias_energy_new)
! accept or reject the move based on the Metropolis rule with a
! correction factor for the change in phase space...dis_length is
! made unitless in change_bond_length
            value = -BETA*(bias_energy_new-bias_energy_old)
            IF (value .GT. exp_max_val) THEN
               w = 10.0_dp
            ELSEIF (value .LT. exp_min_val) THEN
               w = 0.0_dp
            ELSE
               w = EXP(value)*dis_length**2
            ENDIF

         ENDIF

         IF (w .GE. 1.0E0_dp) THEN
            w = 1.0E0_dp
            rand = 0.0E0_dp
         ELSE
            IF (ionode) THEN
               rand = next_random_number(rng_stream)
            ENDIF
            CALL mp_bcast(rand, source, group)
         ENDIF

         IF (rand .LT. w) THEN

! accept the move
            IF (move_type == 'bond') THEN
               moves%bond%qsuccesses = moves%bond%qsuccesses+1
               move_updates%bond%successes = &
                  move_updates%bond%successes+1
               moves%bias_bond%successes = moves%bias_bond%successes+1
               move_updates%bias_bond%successes = &
                  move_updates%bias_bond%successes+1
            ELSEIF (move_type == 'angle') THEN
               moves%angle%qsuccesses = moves%angle%qsuccesses+1
               move_updates%angle%successes = &
                  move_updates%angle%successes+1
               moves%bias_angle%successes = moves%bias_angle%successes+1
               move_updates%bias_angle%successes = &
                  move_updates%bias_angle%successes+1
            ELSE
               moves%dihedral%qsuccesses = moves%dihedral%qsuccesses+1
               move_updates%dihedral%successes = &
                  move_updates%dihedral%successes+1
               moves%bias_dihedral%successes = moves%bias_dihedral%successes+1
               move_updates%bias_dihedral%successes = &
                  move_updates%bias_dihedral%successes+1
            ENDIF

            bias_energy = bias_energy+bias_energy_new- &
                          bias_energy_old

         ELSE

! reject the move
! restore the coordinates
            CALL force_env_get(bias_env, subsys=subsys)
            CALL cp_subsys_get(subsys, particles=particles)
            DO ipart = start_atom, end_atom
               particles%els(ipart)%r(1:3) = r_old(1:3, ipart-start_atom+1)
            ENDDO
            CALL cp_subsys_set(subsys, particles=particles)

         ENDIF

      ENDIF

! deallocate some stuff
      DEALLOCATE (r_old)
      DEALLOCATE (r_new)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_conformation_change

! **************************************************************************************************
!> \brief translates the given molecule randomly in either the x,y, or z direction
!> \param mc_par the mc parameters for the force env
!> \param force_env the force environment used in the move
!> \param bias_env the force environment used to bias the move, if any (it may
!>            be null if lbias=.false. in mc_par)
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected
!> \param move_updates the structure that keeps track of how many moves have
!>               been accepted/rejected since the last time the displacements
!>               were updated
!> \param start_atom the number of the molecule's first atom, assuming the rest of
!>        the atoms follow sequentially
!> \param box_number the box the molecule is in
!> \param bias_energy the biased energy of the system before the move
!> \param molecule_type the type of molecule we're moving
!> \param lreject set to .true. if there is an overlap
!> \param rng_stream the random number stream that we draw from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_molecule_translation(mc_par, force_env, bias_env, moves, &
                                      move_updates, start_atom, box_number, &
                                      bias_energy, molecule_type, &
                                      lreject, rng_stream)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env, bias_env
      TYPE(mc_moves_type), POINTER                       :: moves, move_updates
      INTEGER, INTENT(IN)                                :: start_atom, box_number
      REAL(KIND=dp), INTENT(INOUT)                       :: bias_energy
      INTEGER, INTENT(IN)                                :: molecule_type
      LOGICAL, INTENT(OUT)                               :: lreject
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_molecule_translation', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_number, end_atom, end_mol, group, handle, imolecule, ipart, iparticle, &
         istat, jbox, molecule_number, move_direction, nunits_mol, source, start_mol
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, nunits, nunits_tot
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      LOGICAL                                            :: ionode, lbias, loverlap
      REAL(dp), DIMENSION(:), POINTER                    :: rmtrans
      REAL(KIND=dp)                                      :: BETA, bias_energy_new, bias_energy_old, &
                                                            dis_mol, exp_max_val, exp_min_val, &
                                                            rand, value, w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r_old
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(particle_list_type), POINTER                  :: particles

!   *** Local Counters ***
! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! nullify some pointers
      NULLIFY (particles, subsys)

! get a bunch of stuff from mc_par
      CALL get_mc_par(mc_par, lbias=lbias, &
                      BETA=BETA, exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, rmtrans=rmtrans, ionode=ionode, source=source, &
                      group=group, mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nunits_tot=nunits_tot, &
                                nchains=nchains, nunits=nunits, mol_type=mol_type)

! find out some bounds for mol_type
      start_mol = 1
      DO jbox = 1, box_number-1
         start_mol = start_mol+SUM(nchains(:, jbox))
      ENDDO
      end_mol = start_mol+SUM(nchains(:, box_number))-1

! do some allocation
      ALLOCATE (r_old(1:3, 1:nunits_tot(box_number)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 492, &
                                       "r", 3*nunits_tot(box_number)*dp_size)

! find the index of the last atom of this molecule, and the molecule number
      nunits_mol = nunits(molecule_type)
      end_atom = start_atom+nunits_mol-1
      molecule_number = 0
      atom_number = 1
      DO imolecule = 1, SUM(nchains(:, box_number))
         IF (atom_number == start_atom) THEN
            molecule_number = imolecule
            EXIT
         ENDIF
         atom_number = atom_number+nunits(mol_type(imolecule+start_mol-1))
      ENDDO
      IF (molecule_number == 0) CALL cp__b("motion/mc/mc_moves.F",507,'Cannot find the molecule number')

! are we biasing this move?
      IF (lbias) THEN

! grab the coordinates
         CALL force_env_get(bias_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)

! save the coordinates
         DO ipart = 1, nunits_tot(box_number)
            r_old(1:3, ipart) = particles%els(ipart)%r(1:3)
         ENDDO

! save the energy
         bias_energy_old = bias_energy

      ELSE

! grab the coordinates
         CALL force_env_get(force_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)
      ENDIF

! record the attempt
      moves%trans%attempts = moves%trans%attempts+1
      move_updates%trans%attempts = move_updates%trans%attempts+1
      moves%bias_trans%attempts = moves%bias_trans%attempts+1
      move_updates%bias_trans%attempts = move_updates%bias_trans%attempts+1
      IF (.NOT. lbias) THEN
         moves%trans%qsuccesses = moves%trans%qsuccesses+1
         move_updates%trans%qsuccesses = move_updates%trans%qsuccesses+1
         moves%bias_trans%qsuccesses = moves%bias_trans%qsuccesses+1
         move_updates%bias_trans%qsuccesses = move_updates%bias_trans%qsuccesses+1
      ENDIF

! move one molecule in the system

! call a random number to figure out which direction we're moving
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)
      ! 1,2,3 with equal prob
      move_direction = INT(3*rand)+1

! call a random number to figure out how far we're moving
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)
      dis_mol = rmtrans(molecule_type)*(rand-0.5E0_dp)*2.0E0_dp

! do the move
      DO iparticle = start_atom, end_atom
         particles%els(iparticle)%r(move_direction) = &
            particles%els(iparticle)%r(move_direction)+dis_mol
      ENDDO
      CALL cp_subsys_set(subsys, particles=particles)

! figure out if there is any overlap...need the number of the molecule
      lreject = .FALSE.
      IF (lbias) THEN
         CALL check_for_overlap(bias_env, nchains(:, box_number), &
                                nunits(:), loverlap, mol_type(start_mol:end_mol), &
                                molecule_number=molecule_number)
      ELSE
         CALL check_for_overlap(force_env, nchains(:, box_number), &
                                nunits(:), loverlap, mol_type(start_mol:end_mol), &
                                molecule_number=molecule_number)
         IF (loverlap) lreject = .TRUE.
      ENDIF

! if we're biasing with a cheaper potential, check for acceptance
      IF (lbias) THEN

! here's where we bias the moves
         IF (loverlap) THEN
            w = 0.0E0_dp
         ELSE
            CALL force_env_calc_energy_force(bias_env, calc_force=.FALSE.)
            CALL force_env_get(bias_env, &
                               potential_energy=bias_energy_new)
! accept or reject the move based on the Metropolis rule
            value = -BETA*(bias_energy_new-bias_energy_old)
            IF (value .GT. exp_max_val) THEN
               w = 10.0_dp
            ELSEIF (value .LT. exp_min_val) THEN
               w = 0.0_dp
            ELSE
               w = EXP(value)
            ENDIF

         ENDIF

         IF (w .GE. 1.0E0_dp) THEN
            w = 1.0E0_dp
            rand = 0.0E0_dp
         ELSE
            IF (ionode) rand = next_random_number(rng_stream)
            CALL mp_bcast(rand, source, group)
         ENDIF

         IF (rand .LT. w) THEN

! accept the move
            moves%bias_trans%successes = moves%bias_trans%successes+1
            move_updates%bias_trans%successes = move_updates%bias_trans%successes+1
            moves%trans%qsuccesses = moves%trans%qsuccesses+1
            move_updates%trans%successes = &
               move_updates%trans%successes+1
            moves%qtrans_dis = moves%qtrans_dis+ABS(dis_mol)
            bias_energy = bias_energy+bias_energy_new- &
                          bias_energy_old

         ELSE

! reject the move
! restore the coordinates
            CALL force_env_get(bias_env, subsys=subsys)
            CALL cp_subsys_get(subsys, particles=particles)
            DO ipart = 1, nunits_tot(box_number)
               particles%els(ipart)%r(1:3) = r_old(1:3, ipart)
            ENDDO
            CALL cp_subsys_set(subsys, particles=particles)

         ENDIF

      ENDIF

! deallocate some stuff
      DEALLOCATE (r_old)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_molecule_translation

! **************************************************************************************************
!> \brief rotates the given molecule randomly around the x,y, or z axis...
!>      only works for water at the moment
!> \param mc_par the mc parameters for the force env
!> \param force_env the force environment used in the move
!> \param bias_env the force environment used to bias the move, if any (it may
!>            be null if lbias=.false. in mc_par)
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected
!> \param move_updates the structure that keeps track of how many moves have
!>               been accepted/rejected since the last time the displacements
!>               were updated
!> \param box_number the box the molecule is in
!> \param start_atom the number of the molecule's first atom, assuming the rest of
!>        the atoms follow sequentially
!> \param molecule_type the type of molecule we're moving
!> \param bias_energy the biased energy of the system before the move
!> \param lreject set to .true. if there is an overlap
!> \param rng_stream the random number stream that we draw from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_molecule_rotation(mc_par, force_env, bias_env, moves, &
                                   move_updates, box_number, &
                                   start_atom, molecule_type, bias_energy, lreject, &
                                   rng_stream)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env, bias_env
      TYPE(mc_moves_type), POINTER                       :: moves, move_updates
      INTEGER, INTENT(IN)                                :: box_number, start_atom, molecule_type
      REAL(KIND=dp), INTENT(INOUT)                       :: bias_energy
      LOGICAL, INTENT(OUT)                               :: lreject
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_molecule_rotation', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_number, dir, end_atom, end_mol, group, handle, ii, imolecule, ipart, istat, &
         iunit, jbox, molecule_number, nunits_mol, source, start_mol
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, nunits, nunits_tot
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      LOGICAL                                            :: ionode, lbias, loverlap, lx, ly
      REAL(dp), DIMENSION(:), POINTER                    :: rmrot
      REAL(dp), DIMENSION(:, :), POINTER                 :: mass
      REAL(KIND=dp) :: BETA, bias_energy_new, bias_energy_old, cosdg, dgamma, exp_max_val, &
         exp_min_val, masstot, nxcm, nycm, nzcm, rand, rx, rxnew, ry, rynew, rz, rznew, sindg, &
         value, w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r_old
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(particle_list_type), POINTER                  :: particles

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (rmrot, subsys, particles)

! get a bunch of stuff from mc_par
      CALL get_mc_par(mc_par, lbias=lbias, &
                      BETA=BETA, exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, rmrot=rmrot, mc_molecule_info=mc_molecule_info, &
                      ionode=ionode, group=group, source=source)
      CALL get_mc_molecule_info(mc_molecule_info, nunits=nunits, &
                                nunits_tot=nunits_tot, nchains=nchains, mass=mass, &
                                mol_type=mol_type)

! figure out some bounds for mol_type
      start_mol = 1
      DO jbox = 1, box_number-1
         start_mol = start_mol+SUM(nchains(:, jbox))
      ENDDO
      end_mol = start_mol+SUM(nchains(:, box_number))-1

      nunits_mol = nunits(molecule_type)

! nullify some pointers
      NULLIFY (particles, subsys)

! do some allocation
      ALLOCATE (r_old(1:3, 1:nunits_tot(box_number)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 722, &
                                       "r", 3*nunits_tot(box_number)*dp_size)

! initialize some stuff
      lx = .FALSE.
      ly = .FALSE.

! determine what the final atom in the molecule is numbered, and which
! molecule number this is
      end_atom = start_atom+nunits_mol-1
      molecule_number = 0
      atom_number = 1
      DO imolecule = 1, SUM(nchains(:, box_number))
         IF (atom_number == start_atom) THEN
            molecule_number = imolecule
            EXIT
         ENDIF
         atom_number = atom_number+nunits(mol_type(imolecule+start_mol-1))
      ENDDO
      IF (molecule_number == 0) CALL cp__b("motion/mc/mc_moves.F",741,'Cannot find the molecule number')

! are we biasing this move?
      IF (lbias) THEN

! grab the coordinates
         CALL force_env_get(bias_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)

! save the coordinates
         DO ipart = 1, nunits_tot(box_number)
            r_old(1:3, ipart) = particles%els(ipart)%r(1:3)
         ENDDO

! save the energy
         bias_energy_old = bias_energy

      ELSE

! grab the coordinates
         CALL force_env_get(force_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)
      ENDIF

! grab the masses
      masstot = SUM(mass(1:nunits(molecule_type), molecule_type))

! record the attempt
      moves%bias_rot%attempts = moves%bias_rot%attempts+1
      move_updates%bias_rot%attempts = move_updates%bias_rot%attempts+1
      moves%rot%attempts = moves%rot%attempts+1
      move_updates%rot%attempts = move_updates%rot%attempts+1
      IF (.NOT. lbias) THEN
         moves%rot%qsuccesses = moves%rot%qsuccesses+1
         move_updates%rot%qsuccesses = move_updates%rot%qsuccesses+1
         moves%bias_rot%qsuccesses = moves%bias_rot%qsuccesses+1
         move_updates%bias_rot%qsuccesses = move_updates%bias_rot%qsuccesses+1
      ENDIF

! rotate one molecule in the system

! call a random number to figure out which direction we're moving
      IF (ionode) rand = next_random_number(rng_stream)
!      CALL RANDOM_NUMBER(rand)
      CALL mp_bcast(rand, source, group)
      ! 1,2,3 with equal prob
      dir = INT(3*rand)+1

      IF (dir .EQ. 1) THEN
         lx = .TRUE.
      ELSEIF (dir .EQ. 2) THEN
         ly = .TRUE.
      ENDIF

! Determine new center of mass for chain i by finding the sum
! of m*r for each unit, then dividing by the total mass of the chain
      nxcm = 0.0E0_dp
      nycm = 0.0E0_dp
      nzcm = 0.0E0_dp
      DO ii = 1, nunits_mol
         nxcm = nxcm+particles%els(start_atom-1+ii)%r(1)*mass(ii, molecule_type)
         nycm = nycm+particles%els(start_atom-1+ii)%r(2)*mass(ii, molecule_type)
         nzcm = nzcm+particles%els(start_atom-1+ii)%r(3)*mass(ii, molecule_type)
      ENDDO
      nxcm = nxcm/masstot
      nycm = nycm/masstot
      nzcm = nzcm/masstot

! call a random number to figure out how far we're moving
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)
      dgamma = rmrot(molecule_type)*(rand-0.5E0_dp)*2.0E0_dp

! *** set up the rotation matrix ***

      cosdg = COS(dgamma)
      sindg = SIN(dgamma)

      IF (lx) THEN

! ***    ROTATE UNITS OF I AROUND X-AXIS ***

         DO iunit = start_atom, end_atom
            ry = particles%els(iunit)%r(2)-nycm
            rz = particles%els(iunit)%r(3)-nzcm
            rynew = cosdg*ry-sindg*rz
            rznew = cosdg*rz+sindg*ry

            particles%els(iunit)%r(2) = rynew+nycm
            particles%els(iunit)%r(3) = rznew+nzcm

         ENDDO
      ELSEIF (ly) THEN

! ***    ROTATE UNITS OF I AROUND y-AXIS ***

         DO iunit = start_atom, end_atom
            rx = particles%els(iunit)%r(1)-nxcm
            rz = particles%els(iunit)%r(3)-nzcm
            rxnew = cosdg*rx+sindg*rz
            rznew = cosdg*rz-sindg*rx

            particles%els(iunit)%r(1) = rxnew+nxcm
            particles%els(iunit)%r(3) = rznew+nzcm

         ENDDO

      ELSE

! ***    ROTATE UNITS OF I AROUND z-AXIS ***

         DO iunit = start_atom, end_atom
            rx = particles%els(iunit)%r(1)-nxcm
            ry = particles%els(iunit)%r(2)-nycm

            rxnew = cosdg*rx-sindg*ry
            rynew = cosdg*ry+sindg*rx

            particles%els(iunit)%r(1) = rxnew+nxcm
            particles%els(iunit)%r(2) = rynew+nycm

         ENDDO

      ENDIF
      CALL cp_subsys_set(subsys, particles=particles)

! check for overlap
      lreject = .FALSE.
      IF (lbias) THEN
         CALL check_for_overlap(bias_env, nchains(:, box_number), &
                                nunits(:), loverlap, mol_type(start_mol:end_mol), &
                                molecule_number=molecule_number)
      ELSE
         CALL check_for_overlap(force_env, nchains(:, box_number), &
                                nunits(:), loverlap, mol_type(start_mol:end_mol), &
                                molecule_number=molecule_number)
         IF (loverlap) lreject = .TRUE.
      ENDIF

! if we're biasing classical, check for acceptance
      IF (lbias) THEN

! here's where we bias the moves

         IF (loverlap) THEN
            w = 0.0E0_dp
         ELSE
            CALL force_env_calc_energy_force(bias_env, calc_force=.FALSE.)
            CALL force_env_get(bias_env, &
                               potential_energy=bias_energy_new)
! accept or reject the move based on the Metropolis rule
            value = -BETA*(bias_energy_new-bias_energy_old)
            IF (value .GT. exp_max_val) THEN
               w = 10.0_dp
            ELSEIF (value .LT. exp_min_val) THEN
               w = 0.0_dp
            ELSE
               w = EXP(value)
            ENDIF

         ENDIF

         IF (w .GE. 1.0E0_dp) THEN
            w = 1.0E0_dp
            rand = 0.0E0_dp
         ELSE
            IF (ionode) rand = next_random_number(rng_stream)
            CALL mp_bcast(rand, source, group)
         ENDIF

         IF (rand .LT. w) THEN

! accept the move
            moves%bias_rot%successes = moves%bias_rot%successes+1
            move_updates%bias_rot%successes = move_updates%bias_rot%successes+1
            moves%rot%qsuccesses = moves%rot%qsuccesses+1
            move_updates%rot%successes = move_updates%rot%successes+1
            bias_energy = bias_energy+bias_energy_new- &
                          bias_energy_old

         ELSE

! reject the move
! restore the coordinates
            CALL force_env_get(bias_env, subsys=subsys)
            CALL cp_subsys_get(subsys, particles=particles)
            DO ipart = 1, nunits_tot(box_number)
               particles%els(ipart)%r(1:3) = r_old(1:3, ipart)
            ENDDO
            CALL cp_subsys_set(subsys, particles=particles)

         ENDIF

      ENDIF

! deallocate some stuff
      DEALLOCATE (r_old)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_molecule_rotation

! **************************************************************************************************
!> \brief performs a Monte Carlo move that alters the volume of the simulation box
!> \param mc_par the mc parameters for the force env
!> \param force_env the force environment whose cell we're changing
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected
!> \param move_updates the structure that keeps track of how many moves have
!>               been accepted/rejected since the last time the displacements
!>               were updated
!> \param old_energy the energy of the last accepted move involving an
!>                    unbiased calculation
!> \param box_number the box we're changing the volume of
!> \param energy_check the running total of how much the energy has changed
!>                      since the initial configuration
!> \param r_old the coordinates of the last accepted move involving an
!>               unbiased calculation
!> \param iw the unit number that writes to the screen
!> \param discrete_array tells use which volumes we can do for the discrete
!>            case
!> \param rng_stream the random number stream that we draw from
!> \author MJM
!> \note     Designed for parallel use.
! **************************************************************************************************
   SUBROUTINE mc_volume_move(mc_par, force_env, moves, move_updates, &
                             old_energy, box_number, &
                             energy_check, r_old, iw, discrete_array, rng_stream)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(mc_moves_type), POINTER                       :: moves, move_updates
      REAL(KIND=dp), INTENT(INOUT)                       :: old_energy
      INTEGER, INTENT(IN)                                :: box_number
      REAL(KIND=dp), INTENT(INOUT)                       :: energy_check
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: r_old
      INTEGER, INTENT(IN)                                :: iw
      INTEGER, DIMENSION(1:3, 1:2), INTENT(INOUT)        :: discrete_array
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_volume_move', routineP = moduleN//':'//routineN

      CHARACTER(LEN=200)                                 :: fft_lib
      CHARACTER(LEN=40)                                  :: dat_file
      INTEGER :: cl, end_atom, end_mol, group, handle, iatom, idim, imolecule, iside, &
         iside_change, istat, iunit, jbox, nunits_mol, output_unit, print_level, source, &
         start_atom, start_mol
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, nunits, nunits_tot
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      LOGICAL                                            :: ionode, ldiscrete, lincrease, loverlap, &
                                                            ltoo_small
      REAL(dp), DIMENSION(:, :), POINTER                 :: mass
      REAL(KIND=dp) :: BETA, discrete_step, energy_term, exp_max_val, exp_min_val, new_energy, &
         pressure, pressure_term, rand, rcut, rmvolume, temp_var, value, vol_dis, volume_term, w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc, center_of_mass, center_of_mass_new, &
                                                            diff, new_cell_length, old_cell_length
      REAL(KIND=dp), DIMENSION(1:3, 1:3)                 :: hmat_test
      TYPE(cell_type), POINTER                           :: cell, cell_old, cell_test
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: oldsys
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(particle_list_type), POINTER                  :: particles_old

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! get a bunch of stuff from mc_par
      CALL get_mc_par(mc_par, ionode=ionode, &
                      BETA=BETA, exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, source=source, group=group, &
                      dat_file=dat_file, rmvolume=rmvolume, pressure=pressure, cl=cl, &
                      fft_lib=fft_lib, discrete_step=discrete_step, &
                      ldiscrete=ldiscrete, mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nchains=nchains, &
                                nunits=nunits, nunits_tot=nunits_tot, mol_type=mol_type, &
                                mass=mass)
! figure out some bounds for mol_type
      start_mol = 1
      DO jbox = 1, box_number-1
         start_mol = start_mol+SUM(nchains(:, jbox))
      ENDDO
      end_mol = start_mol+SUM(nchains(:, box_number))-1

      print_level = 1 ! hack, printlevel is for print_keys

! nullify some pointers
      NULLIFY (particles_old, cell_old, oldsys, cell_test, cell)

! do some allocation
      ALLOCATE (r(1:3, 1:nunits_tot(box_number)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1034, &
                                       "r", 3*nunits_tot(box_number)*dp_size)

! record the attempt
      moves%volume%attempts = moves%volume%attempts+1
      move_updates%volume%attempts = move_updates%volume%attempts+1

! now let's grab the cell length and particle positions
      CALL force_env_get(force_env, subsys=oldsys, cell=cell)
      CALL get_cell(cell, abc=abc)
      CALL cell_create(cell_old)
      CALL cell_clone(cell, cell_old)
      CALL cp_subsys_get(oldsys, particles=particles_old)

! find the old cell length
      old_cell_length(1) = abc(1)
      old_cell_length(2) = abc(2)
      old_cell_length(3) = abc(3)

! save the old coordiantes
      DO iatom = 1, nunits_tot(box_number)
         r(1:3, iatom) = particles_old%els(iatom)%r(1:3)
      ENDDO

! now do the move

! call a random number to figure out how far we're moving
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)

! find the test cell lenghts for the discrete volume move
      IF (ldiscrete) THEN
         IF (rand .LT. 0.5_dp) THEN
            lincrease = .TRUE.
         ELSE
            lincrease = .FALSE.
         ENDIF

         new_cell_length(1:3) = old_cell_length(1:3)

! if we're increasing the volume, we need to find a side we can increase
         IF (lincrease) THEN
            DO
               IF (ionode) rand = next_random_number(rng_stream)
               CALL mp_bcast(rand, source, group)
               iside_change = CEILING(3.0_dp*rand)
               IF (discrete_array(iside_change, 1) .EQ. 1) THEN
                  new_cell_length(iside_change) = &
                     new_cell_length(iside_change)+discrete_step
                  EXIT
               ENDIF
            ENDDO
         ELSE
            DO
               IF (ionode) rand = next_random_number(rng_stream)
               CALL mp_bcast(rand, source, group)
               iside_change = CEILING(3.0_dp*rand)
               IF (discrete_array(iside_change, 2) .EQ. 1) THEN
                  new_cell_length(iside_change) = &
                     new_cell_length(iside_change)-discrete_step
                  EXIT
               ENDIF
            ENDDO
         ENDIF
         vol_dis = (new_cell_length(1)*new_cell_length(2)*new_cell_length(3)) &
                   -old_cell_length(1)*old_cell_length(2)*old_cell_length(3)
      ELSE
! now for the not discrete volume move
!!!!!!!!!!!!!!!! for E_V curves
         vol_dis = rmvolume*(rand-0.5E0_dp)*2.0E0_dp
!         WRITE(output_unit,*) '************************ be sure to change back!',&
!                 old_cell_length(1),14.64_dp/angstrom
!         vol_dis=-56.423592_dp/angstrom**3
!         IF(old_cell_length(1) .LE. 14.64_dp/angstrom) THEN
!            vol_dis=0.0_dp
!            WRITE(output_unit,*) 'Found the correct box length!'
!         ENDIF

         temp_var = vol_dis+ &
                    old_cell_length(1)*old_cell_length(2)* &
                    old_cell_length(3)

         IF (temp_var .LE. 0.0E0_dp) THEN
            loverlap = .TRUE. ! cannot have a negative volume
         ELSE
            new_cell_length(1) = (temp_var)**(1.0E0_dp/3.0E0_dp)
            new_cell_length(2) = new_cell_length(1)
            new_cell_length(3) = new_cell_length(1)
            loverlap = .FALSE.
         ENDIF
      ENDIF
      CALL mp_bcast(loverlap, source, group)

      IF (loverlap) THEN
! deallocate some stuff
         DEALLOCATE (r)
         logger => cp_get_default_logger()
         output_unit = cp_logger_get_default_io_unit(logger)
         IF (output_unit > 0) WRITE (output_unit, *) &
            "Volume move rejected because we tried to make too small of box.", vol_dis
!     end the timing
         CALL timestop(handle)
         RETURN
      ENDIF

! now we need to make the new cell
      hmat_test(:, :) = 0.0e0_dp
      hmat_test(1, 1) = new_cell_length(1)
      hmat_test(2, 2) = new_cell_length(2)
      hmat_test(3, 3) = new_cell_length(3)
      CALL cell_create(cell_test, hmat=hmat_test(:, :), periodic=cell%perd)
      CALL cp_subsys_set(oldsys, cell=cell_test)

! now we need to scale the coordinates of all the molecules by the
! center of mass, using the minimum image (not all molecules are in
! the central box)

! now we need to scale the coordinates of all the molecules by the
! center of mass
      end_atom = 0
      DO imolecule = 1, SUM(nchains(:, box_number))
         nunits_mol = nunits(mol_type(imolecule+start_mol-1))
         start_atom = end_atom+1
         end_atom = start_atom+nunits_mol-1
! now find the center of mass
         CALL get_center_of_mass(r(:, start_atom:end_atom), nunits_mol, &
                                 center_of_mass(:), mass(:, mol_type(imolecule+start_mol-1)))

! scale the center of mass and determine the vector that points from the
!    old COM to the new one
         DO iside = 1, 3
            center_of_mass_new(iside) = center_of_mass(iside)* &
                                        new_cell_length(iside)/old_cell_length(iside)
         ENDDO

         DO idim = 1, 3
            diff(idim) = center_of_mass_new(idim)-center_of_mass(idim)
! now change the particle positions
            DO iunit = start_atom, end_atom
               particles_old%els(iunit)%r(idim) = &
                  particles_old%els(iunit)%r(idim)+diff(idim)
            ENDDO
         ENDDO
      ENDDO

! check for overlap
      CALL check_for_overlap(force_env, nchains(:, box_number), &
                             nunits(:), loverlap, mol_type(start_mol:end_mol), &
                             cell_length=new_cell_length)

! figure out if we have overlap problems
      CALL mp_bcast(loverlap, source, group)
      IF (loverlap) THEN
! deallocate some stuff
         DEALLOCATE (r)

         logger => cp_get_default_logger()
         output_unit = cp_logger_get_default_io_unit(logger)
         IF (output_unit > 0) WRITE (output_unit, *) &
            "Volume move rejected due to overlap.", vol_dis
!     end the timing
         CALL timestop(handle)
! reset the cell and particle positions
         CALL cp_subsys_set(oldsys, cell=cell_old)
         DO iatom = 1, nunits_tot(box_number)
            particles_old%els(iatom)%r(1:3) = r_old(1:3, iatom)
         ENDDO
         RETURN
      ENDIF

! stop if we're trying to change a box to a boxlength smaller than rcut
      IF (ionode) THEN
         ltoo_small = .FALSE.
         IF (force_env%in_use .EQ. use_fist_force) THEN
            CALL get_mc_par(mc_par, rcut=rcut)
            IF (new_cell_length(1) .LT. 2.0_dp*rcut) ltoo_small = .TRUE.
            IF (new_cell_length(2) .LT. 2.0_dp*rcut) ltoo_small = .TRUE.
            IF (new_cell_length(3) .LT. 2.0_dp*rcut) ltoo_small = .TRUE.

            IF (ltoo_small) THEN
               WRITE (iw, *) 'new_cell_lengths ', &
                  new_cell_length(1:3)/angstrom
               WRITE (iw, *) 'rcut ', rcut/angstrom
            ENDIF
         ENDIF
      ENDIF
      CALL mp_bcast(ltoo_small, source, group)
      IF (ltoo_small) &
         CALL cp__b("motion/mc/mc_moves.F",1222,"Attempted a volume move where box size got too small.")

! now compute the energy
      CALL force_env_calc_energy_force(force_env, calc_force=.FALSE.)
      CALL force_env_get(force_env, &
                         potential_energy=new_energy)

! accept or reject the move
! to prevent overflows
      energy_term = new_energy-old_energy
      volume_term = -REAL(SUM(nchains(:, box_number)), dp)/BETA* &
                    LOG(new_cell_length(1)*new_cell_length(2)*new_cell_length(3)/ &
                        (old_cell_length(1)*old_cell_length(2)*old_cell_length(3)))
      pressure_term = pressure*vol_dis

      value = -BETA*(energy_term+volume_term+pressure_term)
      IF (value .GT. exp_max_val) THEN
         w = 10.0_dp
      ELSEIF (value .LT. exp_min_val) THEN
         w = 0.0_dp
      ELSE
         w = EXP(value)
      ENDIF

!!!!!!!!!!!!!!!! for E_V curves
!         w=1.0E0_dp
!         w=0.0E0_dp

      IF (w .GE. 1.0E0_dp) THEN
         w = 1.0E0_dp
         rand = 0.0E0_dp
      ELSE
         IF (ionode) rand = next_random_number(rng_stream)
         CALL mp_bcast(rand, source, group)
      ENDIF

      IF (rand .LT. w) THEN

! accept the move
         moves%volume%successes = moves%volume%successes+1
         move_updates%volume%successes = move_updates%volume%successes+1

! update energies
         energy_check = energy_check+(new_energy-old_energy)
         old_energy = new_energy

         DO iatom = 1, nunits_tot(box_number)
            r_old(1:3, iatom) = particles_old%els(iatom)%r(1:3)
         ENDDO

! update discrete_array if we're doing a discrete volume move
         IF (ldiscrete) THEN
            CALL create_discrete_array(new_cell_length(:), &
                                       discrete_array(:, :), discrete_step)
         ENDIF

      ELSE

! reset the cell and particle positions
         CALL cp_subsys_set(oldsys, cell=cell_old)
         DO iatom = 1, nunits_tot(box_number)
            particles_old%els(iatom)%r(1:3) = r_old(1:3, iatom)
         ENDDO

      ENDIF

! deallocate some stuff
      DEALLOCATE (r)
      CALL cell_release(cell_test)
      CALL cell_release(cell_old)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_volume_move

! **************************************************************************************************
!> \brief alters the length of a random bond for the given molecule, using
!>      a mass weighted scheme so the lightest atoms move the most
!> \param r_old the initial coordinates of all molecules in the system
!> \param r_new the new coordinates of all molecules in the system
!> \param mc_par the mc parameters for the force env
!> \param molecule_type the molecule type that we're moving
!> \param molecule_kind the structure containing the molecule information
!> \param dis_length the ratio of the new bond length to the old bond length,
!>                    used in the acceptance rule
!> \param particles the particle_list_type for all particles in the force_env..
!>             used to grab the mass of each atom
!> \param rng_stream the random number stream that we draw from
!>
!>    This subroutine is written to be parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE change_bond_length(r_old, r_new, mc_par, molecule_type, molecule_kind, &
                                 dis_length, particles, rng_stream)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_old
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: r_new
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      INTEGER, INTENT(IN)                                :: molecule_type
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      REAL(KIND=dp), INTENT(OUT)                         :: dis_length
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'change_bond_length', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bond_number, group, handle, i, iatom, &
                                                            ibond, ipart, istat, natom, nbond, &
                                                            source
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_a, atom_b, counter
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: connection, connectivity
      INTEGER, DIMENSION(:), POINTER                     :: nunits
      LOGICAL                                            :: ionode
      REAL(dp), DIMENSION(:), POINTER                    :: rmbond
      REAL(KIND=dp)                                      :: atom_mass, mass_a, mass_b, new_length_a, &
                                                            new_length_b, old_length, rand
      REAL(KIND=dp), DIMENSION(1:3)                      :: bond_a, bond_b
      TYPE(bond_type), DIMENSION(:), POINTER             :: bond_list
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (rmbond, mc_molecule_info)

! get some stuff from mc_par
      CALL get_mc_par(mc_par, mc_molecule_info=mc_molecule_info, source=source, &
                      group=group, rmbond=rmbond, ionode=ionode)
      CALL get_mc_molecule_info(mc_molecule_info, nunits=nunits)

! copy the incoming coordinates so we can change them
      DO ipart = 1, nunits(molecule_type)
         r_new(1:3, ipart) = r_old(1:3, ipart)
      ENDDO

! pick which bond in the molecule at random
      IF (ionode) THEN
         rand = next_random_number(rng_stream)
      ENDIF
      CALL mp_bcast(rand, source, group)
      CALL get_molecule_kind(molecule_kind, natom=natom, nbond=nbond, &
                             bond_list=bond_list)
      bond_number = CEILING(rand*REAL(nbond, dp))

      ALLOCATE (connection(1:natom, 1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1370, &
                                       "connection", 2*natom*dp_size)
! assume at most six bonds per atom
      ALLOCATE (connectivity(1:6, 1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1374, &
                                       "connectivity", 6*natom*dp_size)
      ALLOCATE (counter(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1377, &
                                       "counter", natom)
      ALLOCATE (atom_a(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1380, &
                                       "atom_a", natom)
      ALLOCATE (atom_b(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1383, &
                                       "atom_b", natom)
      connection(:, :) = 0
      connectivity(:, :) = 0
      counter(:) = 0
      atom_a(:) = 0
      atom_b(:) = 0

! now we need to find a list of atoms that each atom in this bond is connected
! to
      DO iatom = 1, natom
         DO ibond = 1, nbond
            IF (bond_list(ibond)%a == iatom) THEN
               counter(iatom) = counter(iatom)+1
               connectivity(counter(iatom), iatom) = bond_list(ibond)%b
            ELSEIF (bond_list(ibond)%b == iatom) THEN
               counter(iatom) = counter(iatom)+1
               connectivity(counter(iatom), iatom) = bond_list(ibond)%a
            ENDIF
         ENDDO
      ENDDO

! now I need to do a depth first search to figure out which atoms are on atom a's
! side and which are on atom b's
      atom_a(:) = 0
      atom_a(bond_list(bond_number)%a) = 1
      CALL depth_first_search(bond_list(bond_number)%a, bond_list(bond_number)%b, &
                              connectivity(:, :), atom_a(:))
      atom_b(:) = 0
      atom_b(bond_list(bond_number)%b) = 1
      CALL depth_first_search(bond_list(bond_number)%b, bond_list(bond_number)%a, &
                              connectivity(:, :), atom_b(:))

! now figure out the masses of the various sides, so we can weight how far we move each
! group of atoms
      mass_a = 0.0_dp
      mass_b = 0.0_dp
      DO iatom = 1, natom
         CALL get_atomic_kind(particles%els(iatom)%atomic_kind, &
                              mass=atom_mass)
         IF (atom_a(iatom) == 1) THEN
            mass_a = mass_a+atom_mass
         ELSE
            mass_b = mass_b+atom_mass
         ENDIF
      ENDDO

! choose a displacement
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)

      dis_length = rmbond(molecule_type)*2.0E0_dp*(rand-0.5E0_dp)

! find the bond vector that atom a will be moving
      DO i = 1, 3
         bond_a(i) = r_new(i, bond_list(bond_number)%a)- &
                     r_new(i, bond_list(bond_number)%b)
         bond_b(i) = -bond_a(i)
      ENDDO

! notice we weight by the opposite masses...therefore lighter segments
! will move further
      old_length = SQRT(DOT_PRODUCT(bond_a, bond_a))
      new_length_a = dis_length*mass_b/(mass_a+mass_b)
      new_length_b = dis_length*mass_a/(mass_a+mass_b)

      DO i = 1, 3
         bond_a(i) = bond_a(i)/old_length*new_length_a
         bond_b(i) = bond_b(i)/old_length*new_length_b
      ENDDO

      DO iatom = 1, natom
         IF (atom_a(iatom) == 1) THEN
            r_new(1, iatom) = r_new(1, iatom)+bond_a(1)
            r_new(2, iatom) = r_new(2, iatom)+bond_a(2)
            r_new(3, iatom) = r_new(3, iatom)+bond_a(3)
         ELSE
            r_new(1, iatom) = r_new(1, iatom)+bond_b(1)
            r_new(2, iatom) = r_new(2, iatom)+bond_b(2)
            r_new(3, iatom) = r_new(3, iatom)+bond_b(3)
         ENDIF
      ENDDO

! correct the value of dis_length for the acceptance rule
      dis_length = (old_length+dis_length)/old_length

      DEALLOCATE (connection)
      DEALLOCATE (connectivity)
      DEALLOCATE (counter)
      DEALLOCATE (atom_a)
      DEALLOCATE (atom_b)
! end the timing
      CALL timestop(handle)

   END SUBROUTINE change_bond_length

! **************************************************************************************************
!> \brief Alters the magnitude of a random angle in a molecule centered on atom C
!>      (connected to atoms A and B).  Atoms A and B are moved amounts related
!>      to their masses (and masses of all connecting atoms), so that heavier
!>      segments are moved less.
!> \param r_old the initial coordinates of all molecules in the system
!> \param r_new the new coordinates of all molecules in the system
!> \param mc_par the mc parameters for the force env
!> \param molecule_type the type of molecule we're playing with
!> \param molecule_kind the structure containing the molecule information
!> \param particles the particle_list_type for all particles in the force_env...
!>             used to grab the mass of each atom
!> \param rng_stream the random number stream that we draw from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE change_bond_angle(r_old, r_new, mc_par, molecule_type, molecule_kind, &
                                particles, rng_stream)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_old
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: r_new
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      INTEGER, INTENT(IN)                                :: molecule_type
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'change_bond_angle', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bend_number, group, handle, i, iatom, &
                                                            ibond, ipart, istat, natom, nbend, &
                                                            nbond, source
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_a, atom_c, counter
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: connection, connectivity
      INTEGER, DIMENSION(:), POINTER                     :: nunits
      LOGICAL                                            :: ionode
      REAL(dp), DIMENSION(:), POINTER                    :: rmangle
      REAL(KIND=dp) :: atom_mass, bis_length, dis_angle, dis_angle_a, dis_angle_c, mass_a, mass_c, &
         new_angle_a, new_angle_c, old_angle, old_length_a, old_length_c, rand, temp_length
      REAL(KIND=dp), DIMENSION(1:3)                      :: bisector, bond_a, bond_c, cross_prod, &
                                                            cross_prod_plane, temp
      TYPE(bend_type), DIMENSION(:), POINTER             :: bend_list
      TYPE(bond_type), DIMENSION(:), POINTER             :: bond_list
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (bend_list, bond_list, rmangle, mc_molecule_info)

! get some stuff from mc_par
      CALL get_mc_par(mc_par, rmangle=rmangle, source=source, &
                      group=group, ionode=ionode, mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nunits=nunits)

! copy the incoming coordinates so we can change them
      DO ipart = 1, nunits(molecule_type)
         r_new(1:3, ipart) = r_old(1:3, ipart)
      ENDDO

! pick which bond in the molecule at random
      IF (ionode) THEN
         rand = next_random_number(rng_stream)
      ENDIF
      CALL mp_bcast(rand, source, group)
      CALL get_molecule_kind(molecule_kind, natom=natom, nbend=nbend, &
                             bend_list=bend_list, bond_list=bond_list, nbond=nbond)
      bend_number = CEILING(rand*REAL(nbend, dp))

      ALLOCATE (connection(1:natom, 1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1550, &
                                       "connection", 2*natom*dp_size)
! assume at most six bonds per atom
      ALLOCATE (connectivity(1:6, 1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1554, &
                                       "connectivity", 6*natom*dp_size)
      ALLOCATE (counter(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1557, &
                                       "counter", natom)
      ALLOCATE (atom_a(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1560, &
                                       "atom_a", natom)
      ALLOCATE (atom_c(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1563, &
                                       "atom_c", natom)
      connection(:, :) = 0
      connectivity(:, :) = 0
      counter(:) = 0
      atom_a(:) = 0
      atom_c(:) = 0

! now we need to find a list of atoms that each atom in this bond is connected
! to
      DO iatom = 1, natom
         DO ibond = 1, nbond
            IF (bond_list(ibond)%a == iatom) THEN
               counter(iatom) = counter(iatom)+1
               connectivity(counter(iatom), iatom) = bond_list(ibond)%b
            ELSEIF (bond_list(ibond)%b == iatom) THEN
               counter(iatom) = counter(iatom)+1
               connectivity(counter(iatom), iatom) = bond_list(ibond)%a
            ENDIF
         ENDDO
      ENDDO

! now I need to do a depth first search to figure out which atoms are on atom a's
! side and which are on atom c's
      atom_a(:) = 0
      atom_a(bend_list(bend_number)%a) = 1
      CALL depth_first_search(bend_list(bend_number)%a, bend_list(bend_number)%b, &
                              connectivity(:, :), atom_a(:))
      atom_c(:) = 0
      atom_c(bend_list(bend_number)%c) = 1
      CALL depth_first_search(bend_list(bend_number)%c, bend_list(bend_number)%b, &
                              connectivity(:, :), atom_c(:))

! now figure out the masses of the various sides, so we can weight how far we move each
! group of atoms
      mass_a = 0.0_dp
      mass_c = 0.0_dp
      DO iatom = 1, natom
         CALL get_atomic_kind(particles%els(iatom)%atomic_kind, &
                              mass=atom_mass)
         IF (atom_a(iatom) == 1) mass_a = mass_a+atom_mass
         IF (atom_c(iatom) == 1) mass_c = mass_c+atom_mass
      ENDDO

! choose a displacement
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)

      dis_angle = rmangle(molecule_type)*2.0E0_dp*(rand-0.5E0_dp)

! need to find the A-B-C bisector

! this going to be tough...we need to find the plane of the A-B-C bond and only shift
! that component for all atoms connected to A and C...otherwise we change other
! internal degrees of freedom

! find the bond vectors
      DO i = 1, 3
         bond_a(i) = r_new(i, bend_list(bend_number)%a)- &
                     r_new(i, bend_list(bend_number)%b)
         bond_c(i) = r_new(i, bend_list(bend_number)%c)- &
                     r_new(i, bend_list(bend_number)%b)
      ENDDO
      old_length_a = SQRT(DOT_PRODUCT(bond_a, bond_a))
      old_length_c = SQRT(DOT_PRODUCT(bond_c, bond_c))
      old_angle = ACOS(DOT_PRODUCT(bond_a, bond_c)/(old_length_a*old_length_c))

      DO i = 1, 3
         bisector(i) = bond_a(i)/old_length_a+ & ! not yet normalized
                       bond_c(i)/old_length_c
      ENDDO
      bis_length = SQRT(DOT_PRODUCT(bisector, bisector))
      bisector(1:3) = bisector(1:3)/bis_length

! now we need to find the cross product of the B-A and B-C vectors and normalize
! it, so we have a vector that defines the bend plane
      cross_prod(1) = bond_a(2)*bond_c(3)-bond_a(3)*bond_c(2)
      cross_prod(2) = bond_a(3)*bond_c(1)-bond_a(1)*bond_c(3)
      cross_prod(3) = bond_a(1)*bond_c(2)-bond_a(2)*bond_c(1)
      cross_prod(1:3) = cross_prod(1:3)/SQRT(DOT_PRODUCT(cross_prod, cross_prod))

! we have two axis of a coordinate system...let's get the third
      cross_prod_plane(1) = cross_prod(2)*bisector(3)-cross_prod(3)*bisector(2)
      cross_prod_plane(2) = cross_prod(3)*bisector(1)-cross_prod(1)*bisector(3)
      cross_prod_plane(3) = cross_prod(1)*bisector(2)-cross_prod(2)*bisector(1)
      cross_prod_plane(1:3) = cross_prod_plane(1:3)/ &
                              SQRT(DOT_PRODUCT(cross_prod_plane, cross_prod_plane))

! now bisector is x, cross_prod_plane is the y vector (pointing towards c),
! and cross_prod is z
! shift the molecule so that atom b is at the origin
      DO iatom = 1, natom
         r_new(1:3, iatom) = r_new(1:3, iatom)- &
                             r_old(1:3, bend_list(bend_number)%b)
      ENDDO

! figure out how much we move each side, since we're mass-weighting, by the
! opposite masses, so lighter moves farther..this angle is the angle between
! the bond vector BA or BC and the bisector
      dis_angle_a = dis_angle*mass_c/(mass_a+mass_c)
      dis_angle_c = dis_angle*mass_a/(mass_a+mass_c)

! now loop through all the atoms, moving the ones that are connected to a or c
      DO iatom = 1, natom
! subtract out the z component (perpendicular to the angle plane)
         temp(1:3) = r_new(1:3, iatom)- &
                     DOT_PRODUCT(cross_prod(1:3), r_new(1:3, iatom))* &
                     cross_prod(1:3)
         temp_length = SQRT(DOT_PRODUCT(temp, temp))

! we can now compute all three components of the new bond vector along the
! axis defined above
         IF (atom_a(iatom) == 1) THEN

! if the y-coordinate is less than zero, we need to switch the sign when we make the vector,
! as the angle computed by the dot product can't distinguish between that
            IF (DOT_PRODUCT(cross_prod_plane(1:3), r_new(1:3, iatom)) &
                .LT. 0.0_dp) THEN

! need to figure out the current iatom-B-bisector angle, so we know what the new angle is
               new_angle_a = ACOS(DOT_PRODUCT(bisector, temp(1:3))/ &
                                  (temp_length))+dis_angle_a

               r_new(1:3, iatom) = COS(new_angle_a)*temp_length*bisector(1:3)- &
                                   SIN(new_angle_a)*temp_length*cross_prod_plane(1:3)+ &
                                   DOT_PRODUCT(cross_prod(1:3), r_new(1:3, iatom))* &
                                   cross_prod(1:3)
            ELSE

! need to figure out the current iatom-B-bisector angle, so we know what the new angle is
               new_angle_a = ACOS(DOT_PRODUCT(bisector, temp(1:3))/ &
                                  (temp_length))-dis_angle_a

               r_new(1:3, iatom) = COS(new_angle_a)*temp_length*bisector(1:3)+ &
                                   SIN(new_angle_a)*temp_length*cross_prod_plane(1:3)+ &
                                   DOT_PRODUCT(cross_prod(1:3), r_new(1:3, iatom))* &
                                   cross_prod(1:3)
            ENDIF

         ELSEIF (atom_c(iatom) == 1) THEN

! if the y-coordinate is less than zero, we need to switch the sign when we make the vector,
! as the angle computed by the dot product can't distinguish between that
            IF (DOT_PRODUCT(cross_prod_plane(1:3), r_new(1:3, iatom)) &
                .LT. 0.0_dp) THEN
! need to figure out the current iatom-B-bisector angle, so we know what the new angle is
               new_angle_c = ACOS(DOT_PRODUCT(bisector(1:3), temp(1:3))/ &
                                  (temp_length))-dis_angle_c

               r_new(1:3, iatom) = COS(new_angle_c)*temp_length*bisector(1:3)- &
                                   SIN(new_angle_c)*temp_length*cross_prod_plane(1:3)+ &
                                   DOT_PRODUCT(cross_prod(1:3), r_new(1:3, iatom))* &
                                   cross_prod(1:3)
            ELSE
               new_angle_c = ACOS(DOT_PRODUCT(bisector(1:3), temp(1:3))/ &
                                  (temp_length))+dis_angle_c

               r_new(1:3, iatom) = COS(new_angle_c)*temp_length*bisector(1:3)+ &
                                   SIN(new_angle_c)*temp_length*cross_prod_plane(1:3)+ &
                                   DOT_PRODUCT(cross_prod(1:3), r_new(1:3, iatom))* &
                                   cross_prod(1:3)
            ENDIF
         ENDIF

      ENDDO

      DO iatom = 1, natom
         r_new(1:3, iatom) = r_new(1:3, iatom)+ &
                             r_old(1:3, bend_list(bend_number)%b)
      ENDDO

! deallocate some stuff
      DEALLOCATE (connection)
      DEALLOCATE (connectivity)
      DEALLOCATE (counter)
      DEALLOCATE (atom_a)
      DEALLOCATE (atom_c)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE change_bond_angle

! **************************************************************************************************
!> \brief Alters a dihedral (A-B-C-D) in the molecule so that all other internal
!>      degrees of freedom remain the same.  If other dihedrals are centered
!>      on B-C, they rotate as well to keep the relationship between the
!>      dihedrals the same.  Atoms A and D are moved amounts related to their
!>      masses (and masses of all connecting atoms), so that heavier segments
!>      are moved less.  All atoms except B and C are rotated around the
!>      B-C bond vector (B and C are not moved).
!> \param r_old the initial coordinates of all molecules in the system
!> \param r_new the new coordinates of all molecules in the system
!> \param mc_par the mc parameters for the force env
!> \param molecule_type the type of molecule we're playing with
!> \param molecule_kind the structure containing the molecule information
!> \param particles the particle_list_type for all particles in the force_env..
!>             used to grab the mass of each atom
!> \param rng_stream the random number stream that we draw from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE change_dihedral(r_old, r_new, mc_par, molecule_type, molecule_kind, &
                              particles, rng_stream)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r_old
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: r_new
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      INTEGER, INTENT(IN)                                :: molecule_type
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'change_dihedral', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, i, iatom, ibond, ipart, &
                                                            istat, natom, nbond, ntorsion, source, &
                                                            torsion_number
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_a, atom_d, counter
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: connection, connectivity
      INTEGER, DIMENSION(:), POINTER                     :: nunits
      LOGICAL                                            :: ionode
      REAL(dp), DIMENSION(:), POINTER                    :: rmdihedral
      REAL(KIND=dp)                                      :: atom_mass, dis_angle, dis_angle_a, &
                                                            dis_angle_d, mass_a, mass_d, &
                                                            old_length_a, rand, u, v, w, x, y, z
      REAL(KIND=dp), DIMENSION(1:3)                      :: bond_a, temp
      TYPE(bond_type), DIMENSION(:), POINTER             :: bond_list
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(torsion_type), DIMENSION(:), POINTER          :: torsion_list

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (rmdihedral, torsion_list, bond_list, mc_molecule_info)

! get some stuff from mc_par
      CALL get_mc_par(mc_par, rmdihedral=rmdihedral, &
                      source=source, group=group, ionode=ionode, &
                      mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nunits=nunits)

! copy the incoming coordinates so we can change them
      DO ipart = 1, nunits(molecule_type)
         r_new(1:3, ipart) = r_old(1:3, ipart)
      ENDDO

! pick which bond in the molecule at random
      IF (ionode) THEN
         rand = next_random_number(rng_stream)
!      CALL RANDOM_NUMBER(rand)
      ENDIF
      CALL mp_bcast(rand, source, group)
      CALL get_molecule_kind(molecule_kind, natom=natom, &
                             bond_list=bond_list, nbond=nbond, &
                             ntorsion=ntorsion, torsion_list=torsion_list)
      torsion_number = CEILING(rand*REAL(ntorsion, dp))

      ALLOCATE (connection(1:natom, 1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1823, &
                                       "connection", 2*natom*dp_size)
! assume at most six bonds per atom
      ALLOCATE (connectivity(1:6, 1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1827, &
                                       "connectivity", 6*natom*dp_size)
      ALLOCATE (counter(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1830, &
                                       "counter", natom)
      ALLOCATE (atom_a(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1833, &
                                       "atom_a", natom)
      ALLOCATE (atom_d(1:natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1836, &
                                       "atom_d", natom)
      connection(:, :) = 0
      connectivity(:, :) = 0
      counter(:) = 0
      atom_a(:) = 0
      atom_d(:) = 0

! now we need to find a list of atoms that each atom in this bond is connected
! to
      DO iatom = 1, natom
         DO ibond = 1, nbond
            IF (bond_list(ibond)%a == iatom) THEN
               counter(iatom) = counter(iatom)+1
               connectivity(counter(iatom), iatom) = bond_list(ibond)%b
            ELSEIF (bond_list(ibond)%b == iatom) THEN
               counter(iatom) = counter(iatom)+1
               connectivity(counter(iatom), iatom) = bond_list(ibond)%a
            ENDIF
         ENDDO
      ENDDO

! now I need to do a depth first search to figure out which atoms are on atom
! a's side and which are on atom d's, but remember we're moving all atoms on a's
! side of b, including atoms not in a's branch
      atom_a(:) = 0
      atom_a(torsion_list(torsion_number)%a) = 1
      CALL depth_first_search(torsion_list(torsion_number)%b, &
                              torsion_list(torsion_number)%c, connectivity(:, :), atom_a(:))
      atom_d(:) = 0
      atom_d(torsion_list(torsion_number)%d) = 1
      CALL depth_first_search(torsion_list(torsion_number)%c, &
                              torsion_list(torsion_number)%b, connectivity(:, :), atom_d(:))

! now figure out the masses of the various sides, so we can weight how far we
! move each group of atoms
      mass_a = 0.0_dp
      mass_d = 0.0_dp
      DO iatom = 1, natom
         CALL get_atomic_kind(particles%els(iatom)%atomic_kind, &
                              mass=atom_mass)
         IF (atom_a(iatom) == 1) mass_a = mass_a+atom_mass
         IF (atom_d(iatom) == 1) mass_d = mass_d+atom_mass
      ENDDO

! choose a displacement
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)

      dis_angle = rmdihedral(molecule_type)*2.0E0_dp*(rand-0.5E0_dp)

! find the bond vectors, B-C, so we know what to rotate around
      DO i = 1, 3
         bond_a(i) = r_new(i, torsion_list(torsion_number)%c)- &
                     r_new(i, torsion_list(torsion_number)%b)
      ENDDO
      old_length_a = SQRT(DOT_PRODUCT(bond_a, bond_a))
      bond_a(1:3) = bond_a(1:3)/old_length_a

! figure out how much we move each side, since we're mass-weighting, by the
! opposite masses, so lighter moves farther...we take the opposite sign of d
! so we're not rotating both angles in the same direction
      dis_angle_a = dis_angle*mass_d/(mass_a+mass_d)
      dis_angle_d = -dis_angle*mass_a/(mass_a+mass_d)

      DO iatom = 1, natom

         IF (atom_a(iatom) == 1) THEN
! shift the coords so b is at the origin
            r_new(1:3, iatom) = r_new(1:3, iatom)- &
                                r_new(1:3, torsion_list(torsion_number)%b)

! multiply by the rotation matrix
            u = bond_a(1)
            v = bond_a(2)
            w = bond_a(3)
            x = r_new(1, iatom)
            y = r_new(2, iatom)
            z = r_new(3, iatom)
            temp(1) = (u*(u*x+v*y+w*z)+(x*(v**2+w**2)-u*(v*y+w*z))*COS(dis_angle_a)+ &
                       SQRT(u**2+v**2+w**2)*(v*z-w*y)*SIN(dis_angle_a))/(u**2+v**2+w**2)
            temp(2) = (v*(u*x+v*y+w*z)+(y*(u**2+w**2)-v*(u*x+w*z))*COS(dis_angle_a)+ &
                       SQRT(u**2+v**2+w**2)*(w*x-u*z)*SIN(dis_angle_a))/(u**2+v**2+w**2)
            temp(3) = (w*(u*x+v*y+w*z)+(z*(v**2+u**2)-w*(u*x+v*y))*COS(dis_angle_a)+ &
                       SQRT(u**2+v**2+w**2)*(u*y-v*x)*SIN(dis_angle_a))/(u**2+v**2+w**2)

! shift back to the original position
            temp(1:3) = temp(1:3)+r_new(1:3, torsion_list(torsion_number)%b)
            r_new(1:3, iatom) = temp(1:3)

         ELSEIF (atom_d(iatom) == 1) THEN

! shift the coords so c is at the origin
            r_new(1:3, iatom) = r_new(1:3, iatom)- &
                                r_new(1:3, torsion_list(torsion_number)%c)

! multiply by the rotation matrix
            u = bond_a(1)
            v = bond_a(2)
            w = bond_a(3)
            x = r_new(1, iatom)
            y = r_new(2, iatom)
            z = r_new(3, iatom)
            temp(1) = (u*(u*x+v*y+w*z)+(x*(v**2+w**2)-u*(v*y+w*z))*COS(dis_angle_d)+ &
                       SQRT(u**2+v**2+w**2)*(v*z-w*y)*SIN(dis_angle_d))/(u**2+v**2+w**2)
            temp(2) = (v*(u*x+v*y+w*z)+(y*(u**2+w**2)-v*(u*x+w*z))*COS(dis_angle_d)+ &
                       SQRT(u**2+v**2+w**2)*(w*x-u*z)*SIN(dis_angle_d))/(u**2+v**2+w**2)
            temp(3) = (w*(u*x+v*y+w*z)+(z*(v**2+u**2)-w*(u*x+v*y))*COS(dis_angle_d)+ &
                       SQRT(u**2+v**2+w**2)*(u*y-v*x)*SIN(dis_angle_d))/(u**2+v**2+w**2)

! shift back to the original position
            temp(1:3) = temp(1:3)+r_new(1:3, torsion_list(torsion_number)%c)
            r_new(1:3, iatom) = temp(1:3)
         ENDIF
      ENDDO

! deallocate some stuff
      DEALLOCATE (connection)
      DEALLOCATE (connectivity)
      DEALLOCATE (counter)
      DEALLOCATE (atom_a)
      DEALLOCATE (atom_d)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE change_dihedral

! **************************************************************************************************
!> \brief performs either a bond or angle change move for a given molecule
!> \param mc_par the mc parameters for the force env
!> \param force_env the force environment used in the move
!> \param bias_env the force environment used to bias the move, if any (it may
!>            be null if lbias=.false. in mc_par)
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected
!> \param energy_check the running energy difference between now and the inital
!>        energy
!> \param r_old the coordinates of force_env before the move
!> \param old_energy the energy of the force_env before the move
!> \param start_atom_swap the number of the swap molecule's first atom, assuming the rest of
!>        the atoms follow sequentially
!> \param target_atom the number of the target atom for swapping
!> \param molecule_type the molecule type for the atom we're swapping
!> \param box_number the number of the box we're doing this move in
!> \param bias_energy_old the biased energy of the system before the move
!> \param last_bias_energy the last biased energy of the system
!> \param move_type dictates if we're moving to an "in" or "out" region
!> \param rng_stream the random number stream that we draw from
!> \author MJM
!> \note     Designed for parallel.
! **************************************************************************************************
   SUBROUTINE mc_avbmc_move(mc_par, force_env, bias_env, moves, &
                            energy_check, r_old, old_energy, start_atom_swap, &
                            target_atom, &
                            molecule_type, box_number, bias_energy_old, last_bias_energy, &
                            move_type, rng_stream)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env, bias_env
      TYPE(mc_moves_type), POINTER                       :: moves
      REAL(KIND=dp), INTENT(INOUT)                       :: energy_check
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: r_old
      REAL(KIND=dp), INTENT(INOUT)                       :: old_energy
      INTEGER, INTENT(IN)                                :: start_atom_swap, target_atom, &
                                                            molecule_type, box_number
      REAL(KIND=dp), INTENT(INOUT)                       :: bias_energy_old, last_bias_energy
      CHARACTER(LEN=*), INTENT(IN)                       :: move_type
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_avbmc_move', routineP = moduleN//':'//routineN

      INTEGER                                            :: end_mol, group, handle, ipart, istat, &
                                                            jbox, natom, nswapmoves, source, &
                                                            start_mol
      INTEGER, DIMENSION(:), POINTER                     :: avbmc_atom, mol_type, nunits, nunits_tot
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      LOGICAL                                            :: ionode, lbias, ldum, lin, loverlap
      REAL(dp), DIMENSION(:), POINTER                    :: avbmc_rmax, avbmc_rmin, pbias
      REAL(dp), DIMENSION(:, :), POINTER                 :: mass
      REAL(KIND=dp) :: BETA, bias_energy_new, del_quickstep_energy, distance, exp_max_val, &
         exp_min_val, max_val, min_val, new_energy, prefactor, rand, rdum, volume_in, volume_out, &
         w, weight_new, weight_old
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r_new
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc, RIJ
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys, subsys_force
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_list_type), POINTER                  :: particles, particles_force

      rdum = 1.0_dp

! begin the timing of the subroutine
      CALL timeset(routineN, handle)

! get a bunch of stuff from mc_par
      CALL get_mc_par(mc_par, lbias=lbias, &
                      BETA=BETA, max_val=max_val, min_val=min_val, exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, avbmc_atom=avbmc_atom, &
                      avbmc_rmin=avbmc_rmin, avbmc_rmax=avbmc_rmax, &
                      nswapmoves=nswapmoves, ionode=ionode, source=source, &
                      group=group, pbias=pbias, mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nchains=nchains, &
                                mass=mass, nunits=nunits, nunits_tot=nunits_tot, mol_type=mol_type)
! figure out some bounds for mol_type
      start_mol = 1
      DO jbox = 1, box_number-1
         start_mol = start_mol+SUM(nchains(:, jbox))
      ENDDO
      end_mol = start_mol+SUM(nchains(:, box_number))-1

! nullify some pointers
      NULLIFY (particles, subsys, molecule_kinds_new, molecule_kind, &
               particles_force, subsys_force)

! do some allocation
      ALLOCATE (r_new(1:3, 1:nunits_tot(box_number)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2055, &
                                       "r_new", 3*nunits_tot(box_number)*dp_size)

! now we need to grab and save coordinates, in case we reject
! are we biasing this move?
      IF (lbias) THEN

! grab the coordinates
         CALL force_env_get(bias_env, cell=cell, subsys=subsys)
         CALL cp_subsys_get(subsys, &
                            particles=particles, molecule_kinds_new=molecule_kinds_new)
         molecule_kind => molecule_kinds_new%els(1)
         CALL get_molecule_kind(molecule_kind, natom=natom)
         CALL get_cell(cell, abc=abc)

! save the energy
!         bias_energy_old=bias_energy

      ELSE

! grab the coordinates
         CALL force_env_get(force_env, cell=cell, subsys=subsys)
         CALL cp_subsys_get(subsys, &
                            particles=particles, molecule_kinds_new=molecule_kinds_new)
         molecule_kind => molecule_kinds_new%els(1)
         CALL get_molecule_kind(molecule_kind, natom=natom)
         CALL get_cell(cell, abc=abc)

      ENDIF

! let's determine if the molecule to be moved is in the "in" region or the
! "out" region of the target
      RIJ(1) = particles%els(start_atom_swap+avbmc_atom(molecule_type)-1)%r(1)- &
               particles%els(target_atom)%r(1)-abc(1)*ANINT( &
               (particles%els(start_atom_swap+avbmc_atom(molecule_type)-1)%r(1)- &
                particles%els(target_atom)%r(1))/abc(1))
      RIJ(2) = particles%els(start_atom_swap+avbmc_atom(molecule_type)-1)%r(2)- &
               particles%els(target_atom)%r(2)-abc(2)*ANINT( &
               (particles%els(start_atom_swap+avbmc_atom(molecule_type)-1)%r(2)- &
                particles%els(target_atom)%r(2))/abc(2))
      RIJ(3) = particles%els(start_atom_swap+avbmc_atom(molecule_type)-1)%r(3)- &
               particles%els(target_atom)%r(3)-abc(3)*ANINT( &
               (particles%els(start_atom_swap+avbmc_atom(molecule_type)-1)%r(3)- &
                particles%els(target_atom)%r(3))/abc(3))
      distance = SQRT(RIJ(1)**2+RIJ(2)**2+RIJ(3)**2)
      IF (distance .LE. avbmc_rmax(molecule_type) .AND. distance .GE. avbmc_rmin(molecule_type)) THEN
         lin = .TRUE.
      ELSE
         lin = .FALSE.
      ENDIF

! increment the counter of the particular move we've done
!     swapping into the "in" region of mol_target
      IF (lin) THEN
         IF (move_type == 'in') THEN
            moves%avbmc_inin%attempts = &
               moves%avbmc_inin%attempts+1
         ELSE
            moves%avbmc_inout%attempts = &
               moves%avbmc_inout%attempts+1
         ENDIF
      ELSE
         IF (move_type == 'in') THEN
            moves%avbmc_outin%attempts = &
               moves%avbmc_outin%attempts+1
         ELSE
            moves%avbmc_outout%attempts = &
               moves%avbmc_outout%attempts+1
         ENDIF
      ENDIF

      IF (lbias) THEN

         IF (move_type == 'in') THEN

! do CBMC for the old config
            CALL generate_cbmc_swap_config(bias_env, BETA, max_val, min_val, exp_max_val, &
                                           exp_min_val, nswapmoves, &
                                           weight_old, start_atom_swap, nunits_tot(box_number), nunits, nunits(molecule_type), &
                                           mass(:, molecule_type), ldum, rdum, &
                                           bias_energy_old, ionode, .TRUE., mol_type(start_mol:end_mol), nchains(:, box_number), &
                                           source, group, rng_stream, &
                                           avbmc_atom=avbmc_atom(molecule_type), &
                                           rmin=avbmc_rmin(molecule_type), rmax=avbmc_rmax(molecule_type), move_type='out', &
                                           target_atom=target_atom)

         ELSE

! do CBMC for the old config
            CALL generate_cbmc_swap_config(bias_env, BETA, max_val, min_val, exp_max_val, &
                                           exp_min_val, nswapmoves, &
                                           weight_old, start_atom_swap, nunits_tot(box_number), nunits, nunits(molecule_type), &
                                           mass(:, molecule_type), ldum, rdum, &
                                           bias_energy_old, ionode, .TRUE., mol_type(start_mol:end_mol), nchains(:, box_number), &
                                           source, group, rng_stream, &
                                           avbmc_atom=avbmc_atom(molecule_type), &
                                           rmin=avbmc_rmin(molecule_type), rmax=avbmc_rmax(molecule_type), move_type='in', &
                                           target_atom=target_atom)

         ENDIF

! generate the new config
         CALL generate_cbmc_swap_config(bias_env, BETA, max_val, min_val, exp_max_val, &
                                        exp_min_val, nswapmoves, &
                                        weight_new, start_atom_swap, nunits_tot(box_number), nunits, nunits(molecule_type), &
                                        mass(:, molecule_type), loverlap, bias_energy_new, &
                                        bias_energy_old, ionode, .FALSE., mol_type(start_mol:end_mol), nchains(:, box_number), &
                                        source, group, rng_stream, &
                                        avbmc_atom=avbmc_atom(molecule_type), &
                                        rmin=avbmc_rmin(molecule_type), rmax=avbmc_rmax(molecule_type), move_type=move_type, &
                                        target_atom=target_atom)

! the energy that comes out of the above routine is the difference...we want
! the real energy for the acceptance rule...we don't do this for the
! lbias=.false. case because it doesn't appear in the acceptance rule, and
! we compensate in case of acceptance
         bias_energy_new = bias_energy_new+bias_energy_old

      ELSE

         IF (move_type == 'in') THEN

! find the weight of the old config
            CALL generate_cbmc_swap_config(force_env, BETA, max_val, min_val, exp_max_val, &
                                           exp_min_val, nswapmoves, &
                                           weight_old, start_atom_swap, nunits_tot(box_number), nunits, nunits(molecule_type), &
                                           mass(:, molecule_type), ldum, rdum, old_energy, &
                                           ionode, .TRUE., mol_type(start_mol:end_mol), nchains(:, box_number), &
                                           source, group, rng_stream, &
                                           avbmc_atom=avbmc_atom(molecule_type), &
                                           rmin=avbmc_rmin(molecule_type), rmax=avbmc_rmax(molecule_type), move_type='out', &
                                           target_atom=target_atom)

         ELSE

! find the weight of the old config
            CALL generate_cbmc_swap_config(force_env, BETA, max_val, min_val, exp_max_val, &
                                           exp_min_val, nswapmoves, &
                                           weight_old, start_atom_swap, nunits_tot(box_number), nunits, nunits(molecule_type), &
                                           mass(:, molecule_type), ldum, rdum, old_energy, &
                                           ionode, .TRUE., mol_type(start_mol:end_mol), nchains(:, box_number), &
                                           source, group, rng_stream, &
                                           avbmc_atom=avbmc_atom(molecule_type), &
                                           rmin=avbmc_rmin(molecule_type), rmax=avbmc_rmax(molecule_type), move_type='in', &
                                           target_atom=target_atom)

         ENDIF

         ! generate the new config...do this after, because it changes the force_env
         CALL generate_cbmc_swap_config(force_env, BETA, max_val, min_val, exp_max_val, &
                                        exp_min_val, nswapmoves, &
                                        weight_new, start_atom_swap, nunits_tot(box_number), nunits, nunits(molecule_type), &
                                        mass(:, molecule_type), loverlap, new_energy, old_energy, &
                                        ionode, .FALSE., mol_type(start_mol:end_mol), nchains(:, box_number), &
                                        source, group, rng_stream, &
                                        avbmc_atom=avbmc_atom(molecule_type), &
                                        rmin=avbmc_rmin(molecule_type), rmax=avbmc_rmax(molecule_type), move_type=move_type, &
                                        target_atom=target_atom)

      ENDIF

      IF (loverlap) THEN
         DEALLOCATE (r_new)

! need to reset the old coordinates
         IF (lbias) THEN
            CALL force_env_get(bias_env, subsys=subsys)
            CALL cp_subsys_get(subsys, particles=particles)
         ELSE
            CALL force_env_get(force_env, subsys=subsys)
            CALL cp_subsys_get(subsys, particles=particles)
         ENDIF
         DO ipart = 1, nunits_tot(box_number)
            particles%els(ipart)%r(1:3) = r_old(1:3, ipart)
         ENDDO

         CALL timestop(handle)

         RETURN
      ENDIF

! if we're biasing, we need to compute the new energy with the full
! potential
      IF (lbias) THEN
! need to give the force_env the coords from the bias_env
         CALL force_env_get(force_env, subsys=subsys_force)
         CALL cp_subsys_get(subsys_force, particles=particles_force)
         CALL force_env_get(bias_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)
         DO ipart = 1, nunits_tot(box_number)
            particles_force%els(ipart)%r(1:3) = particles%els(ipart)%r(1:3)
         ENDDO

         CALL force_env_calc_energy_force(force_env, &
                                          calc_force=.FALSE.)
         CALL force_env_get(force_env, &
                            potential_energy=new_energy)

      ENDIF

      volume_in = 4.0_dp/3.0_dp*pi*(avbmc_rmax(molecule_type)**3-avbmc_rmin(molecule_type)**3)
      volume_out = abc(1)*abc(2)*abc(3)-volume_in

      IF (lin .AND. move_type == 'in' .OR. &
          .NOT. lin .AND. move_type == 'out') THEN
! standard Metropolis rule
         prefactor = 1.0_dp
      ELSEIF (.NOT. lin .AND. move_type == 'in') THEN
         prefactor = (1.0_dp-pbias(molecule_type))*volume_in/(pbias(molecule_type)*volume_out)
      ELSE
         prefactor = pbias(molecule_type)*volume_out/((1.0_dp-pbias(molecule_type))*volume_in)
      ENDIF

      IF (lbias) THEN
! AVBMC with CBMC and a biasing potential...notice that if the biasing
! potential equals the quickstep potential, this cancels out to the
! acceptance below
         del_quickstep_energy = (-BETA)*(new_energy-old_energy- &
                                         (bias_energy_new-bias_energy_old))

         IF (del_quickstep_energy .GT. exp_max_val) THEN
            del_quickstep_energy = max_val
         ELSEIF (del_quickstep_energy .LT. exp_min_val) THEN
            del_quickstep_energy = 0.0_dp
         ELSE
            del_quickstep_energy = EXP(del_quickstep_energy)
         ENDIF

         w = prefactor*del_quickstep_energy*weight_new/weight_old

      ELSE

! AVBMC with CBMC
         w = prefactor*weight_new/weight_old
      ENDIF

! check if the move is accepted
      IF (w .GE. 1.0E0_dp) THEN
         rand = 0.0E0_dp
      ELSE
         IF (ionode) rand = next_random_number(rng_stream)
         CALL mp_bcast(rand, source, group)
      ENDIF

      IF (rand .LT. w) THEN

! accept the move

         IF (lin) THEN
            IF (move_type == 'in') THEN
               moves%avbmc_inin%successes = &
                  moves%avbmc_inin%successes+1
            ELSE
               moves%avbmc_inout%successes = &
                  moves%avbmc_inout%successes+1
            ENDIF
         ELSE
            IF (move_type == 'in') THEN
               moves%avbmc_outin%successes = &
                  moves%avbmc_outin%successes+1
            ELSE
               moves%avbmc_outout%successes = &
                  moves%avbmc_outout%successes+1
            ENDIF
         ENDIF

! we need to compensate for the fact that we take the difference in
! generate_cbmc_config to keep the exponetials small
         IF (.NOT. lbias) THEN
            new_energy = new_energy+old_energy
         ENDIF

! update energies
         energy_check = energy_check+(new_energy-old_energy)
         old_energy = new_energy

! if we're biasing the update the biasing energy
         IF (lbias) THEN
! need to do this outside of the routine
            last_bias_energy = bias_energy_new
            bias_energy_old = bias_energy_new
         ENDIF

! update coordinates
         CALL force_env_get(force_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)
         DO ipart = 1, nunits_tot(box_number)
            r_old(1:3, ipart) = particles%els(ipart)%r(1:3)
         ENDDO
      ELSE
! reject the move...need to restore the old coordinates
         IF (lbias) THEN
            CALL force_env_get(bias_env, subsys=subsys)
            CALL cp_subsys_get(subsys, particles=particles)
            DO ipart = 1, nunits_tot(box_number)
               particles%els(ipart)%r(1:3) = r_old(1:3, ipart)
            ENDDO
            CALL cp_subsys_set(subsys, particles=particles)
         ENDIF
         CALL force_env_get(force_env, subsys=subsys)
         CALL cp_subsys_get(subsys, particles=particles)
         DO ipart = 1, nunits_tot(box_number)
            particles%els(ipart)%r(1:3) = r_old(1:3, ipart)
         ENDDO
         CALL cp_subsys_set(subsys, particles=particles)

      ENDIF

! deallocate some stuff
      DEALLOCATE (r_new)
! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_avbmc_move

! **************************************************************************************************
!> \brief performs a hybrid Monte Carlo move that runs a short MD sequence
!> \param mc_par the mc parameters for the force env
!> \param force_env the force environment whose cell we're changing
!> \param globenv ...
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected
!> \param move_updates the structure that keeps track of how many moves have
!>               been accepted/rejected since the last time the displacements
!>               were updated
!> \param old_energy the energy of the last accepted move involving an
!>                    unbiased calculation
!> \param box_number the box we're changing the volume of
!> \param energy_check the running total of how much the energy has changed
!>                      since the initial configuration
!> \param r_old the coordinates of the last accepted move involving an
!>               unbiased calculation
!> \param rng_stream the random number stream that we draw from
!> \author MJM
!> \note     Designed for parallel use.
! **************************************************************************************************
   SUBROUTINE mc_hmc_move(mc_par, force_env, globenv, moves, move_updates, &
                          old_energy, box_number, &
                          energy_check, r_old, rng_stream)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(mc_moves_type), POINTER                       :: moves, move_updates
      REAL(KIND=dp), INTENT(INOUT)                       :: old_energy
      INTEGER, INTENT(IN)                                :: box_number
      REAL(KIND=dp), INTENT(INOUT)                       :: energy_check
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: r_old
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_hmc_move', routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, iatom, istat, source
      INTEGER, DIMENSION(:), POINTER                     :: nunits_tot
      LOGICAL                                            :: ionode
      REAL(KIND=dp)                                      :: BETA, energy_term, exp_max_val, &
                                                            exp_min_val, new_energy, rand, value, w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r
      TYPE(cp_subsys_type), POINTER                      :: oldsys
      TYPE(mc_ekin_type), POINTER                        :: hmc_ekin
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(particle_list_type), POINTER                  :: particles_old

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! get a bunch of stuff from mc_par
      CALL get_mc_par(mc_par, ionode=ionode, &
                      BETA=BETA, exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, source=source, group=group, &
                      mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nunits_tot=nunits_tot)

! nullify some pointers
      NULLIFY (particles_old, oldsys, hmc_ekin)

! do some allocation
      ALLOCATE (r(1:3, 1:nunits_tot(box_number)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 2434, &
                                       "r", 3*nunits_tot(box_number)*dp_size)
      ALLOCATE (hmc_ekin)

! record the attempt
      moves%hmc%attempts = moves%hmc%attempts+1
      move_updates%hmc%attempts = move_updates%hmc%attempts+1

! now let's grab the particle positions
      CALL force_env_get(force_env, subsys=oldsys)
      CALL cp_subsys_get(oldsys, particles=particles_old)

! save the old coordinates
      DO iatom = 1, nunits_tot(box_number)
         r(1:3, iatom) = particles_old%els(iatom)%r(1:3)
      ENDDO

! now run the MD simulation
      CALL qs_mol_dyn(force_env, globenv, hmc_e_initial=hmc_ekin%initial_ekin, hmc_e_final=hmc_ekin%final_ekin)

! get the energy
      CALL force_env_get(force_env, &
                         potential_energy=new_energy)

! accept or reject the move
! to prevent overflows
      energy_term = new_energy+hmc_ekin%final_ekin-old_energy-hmc_ekin%initial_ekin

      value = -BETA*(energy_term)
      IF (value .GT. exp_max_val) THEN
         w = 10.0_dp
      ELSEIF (value .LT. exp_min_val) THEN
         w = 0.0_dp
      ELSE
         w = EXP(value)
      ENDIF

      IF (w .GE. 1.0E0_dp) THEN
         w = 1.0E0_dp
         rand = 0.0E0_dp
      ELSE
         IF (ionode) rand = next_random_number(rng_stream)
         CALL mp_bcast(rand, source, group)
      ENDIF

      IF (rand .LT. w) THEN

! accept the move
         moves%hmc%successes = moves%hmc%successes+1
         move_updates%hmc%successes = move_updates%hmc%successes+1

! update energies
         energy_check = energy_check+(new_energy-old_energy)
         old_energy = new_energy

         DO iatom = 1, nunits_tot(box_number)
            r_old(1:3, iatom) = particles_old%els(iatom)%r(1:3)
         ENDDO

      ELSE

! reset the cell and particle positions
         DO iatom = 1, nunits_tot(box_number)
            particles_old%els(iatom)%r(1:3) = r_old(1:3, iatom)
         ENDDO

      ENDIF

! deallocate some stuff
      DEALLOCATE (r)
      DEALLOCATE (hmc_ekin)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_hmc_move
END MODULE mc_moves

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains miscellaneous subroutines used in the Monte Carlo runs,
!>      mostly I/O stuff
!> \author MJM
! **************************************************************************************************
MODULE mc_misc
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE force_env_types,                 ONLY: use_fist_force,&
                                              use_qs_force
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE mc_types,                        ONLY: &
        accattempt, get_mc_input_file, get_mc_molecule_info, get_mc_par, mc_averages_type, &
        mc_input_file_type, mc_molecule_info_type, mc_moves_p_type, mc_moves_type, mc_simpar_type
   USE physcon,                         ONLY: angstrom
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: final_mc_write, mc_averages_create, mc_averages_release, &
             mc_make_dat_file_new

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_misc'

CONTAINS

! **************************************************************************************************
!> \brief initializes the structure that holds running averages of MC variables
!> \param averages the mc_averages strucutre you want to initialize
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_averages_create(averages)

      TYPE(mc_averages_type), POINTER                    :: averages

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_averages_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, stat

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! allocate all the structures...not sure why, but it won't work otherwise
      ALLOCATE (averages, stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 59, &
                                      "averages", 0)

      averages%ave_energy = 0.0E0_dp
      averages%ave_energy_squared = 0.0E0_dp
      averages%ave_volume = 0.0E0_dp
      averages%molecules = 0.0E0_dp

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_averages_create

! **************************************************************************************************
!> \brief deallocates the structure that holds running averages of MC variables
!> \param averages the mc_averages strucutre you want to release
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_averages_release(averages)

      TYPE(mc_averages_type), POINTER                    :: averages

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_averages_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! deallocate
      DEALLOCATE (averages)

      NULLIFY (averages)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_averages_release

! **************************************************************************************************
!> \brief writes a bunch of simulation data to the specified unit
!> \param mc_par the mc parameters for the simulation
!> \param all_moves the structure that holds data on how many moves are
!>               accepted/rejected
!> \param iw the unit to write to
!> \param energy_check the sum of the energy changes of each move
!> \param initial_energy the initial unbiased energy of the system
!> \param final_energy the final unbiased energy of the system
!> \param averages the structure that holds computed average properites for
!>               the simulation
!>
!>    Only use in serial.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE final_mc_write(mc_par, all_moves, iw, energy_check, initial_energy, &
                             final_energy, averages)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(mc_moves_p_type), DIMENSION(:), POINTER       :: all_moves
      INTEGER, INTENT(IN)                                :: iw
      REAL(KIND=dp), INTENT(IN)                          :: energy_check, initial_energy, &
                                                            final_energy
      TYPE(mc_averages_type), POINTER                    :: averages

      CHARACTER(len=*), PARAMETER :: routineN = 'final_mc_write', routineP = moduleN//':'//routineN

      CHARACTER(LEN=5)                                   :: molecule_string, tab_string
      CHARACTER(LEN=default_string_length)               :: format_string, string1, string2, string3
      INTEGER                                            :: handle, itype, nmol_types
      LOGICAL                                            :: lbias
      REAL(dp), DIMENSION(:), POINTER                    :: rmangle, rmbond, rmdihedral, rmrot, &
                                                            rmtrans
      REAL(KIND=dp)                                      :: pmswap, rmvolume
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(mc_moves_type), POINTER                       :: moves

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (mc_molecule_info, rmbond, rmangle, rmdihedral, rmrot, rmtrans)

      CALL get_mc_par(mc_par, pmswap=pmswap, rmvolume=rmvolume, &
                      lbias=lbias, rmbond=rmbond, rmangle=rmangle, rmdihedral=rmdihedral, &
                      rmtrans=rmtrans, rmrot=rmrot, mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nmol_types=nmol_types)
      WRITE (molecule_string, '(I2)') nmol_types
      WRITE (tab_string, '(I4)') 81-11*nmol_types
      format_string = "(A,T"//TRIM(ADJUSTL(tab_string))//","//TRIM(ADJUSTL(molecule_string))//"(2X,F9.6))"

! write out some data averaged over the whole simulation
      WRITE (iw, *)
      WRITE (iw, '(A,A)') '*****************************************************', &
         '***************************'
      WRITE (iw, '(A,T66,F15.8)') "Average Energy [Hartrees]:", &
         averages%ave_energy
      IF (pmswap .GT. 0.0E0_dp) THEN
         WRITE (iw, '(A,T66,F15.8)') "Average number of molecules:", &
            averages%molecules
      ENDIF
      WRITE (iw, '(A,A,T65,F16.6)') "Average Volume ", &
         "[angstroms**3]:", averages%ave_volume*angstrom**3

      WRITE (iw, *)

! write out acceptance rates for the moves

! volume moves
      WRITE (iw, '(A,A)') '-----------------------------------------------------', &
         '---------------------------'
      string2 = "Attempted       Accepted       Percent"
      string1 = "Volume Moves"
      string3 = "Maximum volume displacement [angstroms**3]= "
      rmvolume = rmvolume*angstrom**3
      CALL final_move_write(all_moves(1)%moves%volume, string1, string2, iw, &
                            displacement=rmvolume, lbias=.FALSE., format_string=format_string, &
                            string3=string3)

! Hybrid MC moves (basically short MD runs)
      string2 = "Attempted       Accepted       Percent"
      string1 = "HMC Moves"
      CALL final_move_write(all_moves(1)%moves%hmc, string1, string2, iw)

! Quickstep moves (a series of moves with one potential, and then corrected for
! by another potential
      string2 = "Attempted       Accepted       Percent"
      string1 = "Quickstep Moves"
      CALL final_move_write(all_moves(1)%moves%Quickstep, string1, string2, iw)

      DO itype = 1, nmol_types
         WRITE (iw, '(A,A)') '-----------------------------------------------------', &
            '---------------------------'
         WRITE (iw, '(A,I5)') 'Move Data for Molecule Type ', itype
         WRITE (iw, '(A,A)') '-----------------------------------------------------', &
            '---------------------------'

         moves => all_moves(itype)%moves

! AVBMC moves
         string2 = "Attempted       Accepted       Percent"
         string1 = "AVBMC moves from in to in"
         CALL final_move_write(moves%avbmc_inin, string1, string2, iw)
         string1 = "AVBMC moves from in to out"
         CALL final_move_write(moves%avbmc_inout, string1, string2, iw)
         string1 = "AVBMC moves from out to in"
         CALL final_move_write(moves%avbmc_outin, string1, string2, iw)
         string1 = "AVBMC moves from out to out"
         CALL final_move_write(moves%avbmc_outout, string1, string2, iw)

! conformation changes
         IF (moves%angle%attempts .GT. 0 .OR. &
             moves%bond%attempts .GT. 0 .OR. &
             moves%dihedral%attempts .GT. 0) THEN
            WRITE (iw, '(A,T43,A)') "Conformational Moves", &
               "Attempted       Accepted       Percent"
            WRITE (iw, '(T46,I6,9X,I6,7X,F7.3)') &
               moves%bond%attempts+moves%angle%attempts+ &
               moves%dihedral%attempts, &
               moves%bond%successes+moves%angle%successes+ &
               moves%dihedral%successes, &
               REAL(moves%bond%successes+moves%angle%successes+ &
                    moves%dihedral%successes, dp)/ &
               REAL(moves%bond%attempts+moves%angle%attempts+ &
                    moves%dihedral%attempts, dp)*100.0E0_dp
            string2 = "Attempted       Accepted       Percent"
            string1 = "Bond Changes"
            string3 = "Maximum bond displacement [angstroms]= "
            rmbond(itype) = rmbond(itype)*angstrom
            CALL final_move_write(moves%bond, string1, string2, iw, &
                                  displacement=rmbond(itype), lbias=lbias, format_string=format_string, &
                                  string3=string3)

            string1 = "Angle Changes"
            string3 = "Maximum angle displacement [degrees]= "
            rmangle(itype) = rmangle(itype)/pi*180.0E0_dp
            CALL final_move_write(moves%angle, string1, string2, iw, &
                                  displacement=rmangle(itype), lbias=lbias, format_string=format_string, &
                                  string3=string3)

            string1 = "Dihedral Changes"
            string3 = "Maximum dihedral displacement [degrees]= "
            rmdihedral(itype) = rmdihedral(itype)/pi*180.0E0_dp
            CALL final_move_write(moves%dihedral, string1, string2, iw, &
                                  displacement=rmdihedral(itype), lbias=lbias, format_string=format_string, &
                                  string3=string3)

            WRITE (iw, '(A,A,I5)') "Conformational Moves Rejected Because", &
               "Box Was Empty: ", moves%empty_conf
            WRITE (iw, '(A,A)') '-----------------------------------------------', &
               '--------------------------------'
         ENDIF

! translation moves
         string1 = "Translation Moves"
         string3 = "Maximum molecular translational displacement [angstroms]= "
         rmtrans(itype) = rmtrans(itype)*angstrom
         CALL final_move_write(moves%trans, string1, string2, iw, &
                               displacement=rmtrans(itype), lbias=lbias, format_string=format_string, &
                               string3=string3)

! rotation moves
         string1 = "Rotation Moves"
         string3 = "Maximum molecular rotational displacement [degrees]= "
         rmrot(itype) = rmrot(itype)/pi*180.0E0_dp
         CALL final_move_write(moves%rot, string1, string2, iw, &
                               displacement=rmrot(itype), lbias=lbias, format_string=format_string, &
                               string3=string3)

! swap moves
         IF (moves%swap%attempts .GT. 0) THEN
            WRITE (iw, '(A,T43,A)') "Swap Moves into this box", &
               "Attempted       Empty          Percent"
            WRITE (iw, '(T46,I6,9X,I6,7X,F7.3)') &
               moves%swap%attempts, &
               moves%empty, &
               REAL(moves%empty, dp)/ &
               REAL(moves%swap%attempts, dp)*100.0E0_dp
            WRITE (iw, '(A,T43,A)') "                  Growths", &
               "Attempted       Sucessful      Percent"
            WRITE (iw, '(T46,I6,9X,I6,7X,F7.3)') &
               moves%swap%attempts, &
               moves%grown, &
               REAL(moves%grown, dp)/ &
               REAL(moves%swap%attempts, dp)*100.0E0_dp
            WRITE (iw, '(A,T43,A)') "                    Total", &
               "Attempted       Accepted       Percent"
            WRITE (iw, '(T46,I6,9X,I6,7X,F7.3)') &
               moves%swap%attempts, &
               moves%swap%successes, &
               REAL(moves%swap%successes, dp)/ &
               REAL(moves%swap%attempts, dp)*100.0E0_dp
            WRITE (iw, '(A,A)') '-----------------------------------------------', &
               '--------------------------------'
         ENDIF

! now we write out information on the classical moves, if it's
! a classical simulations
         IF (lbias) THEN
            WRITE (iw, '(A)') "Biased Move Data"
            WRITE (iw, '(A,A)') '-------------------------------------------------', &
               '-------------------------------'
            string2 = "Attempted       Accepted       Percent"
            string1 = "Bond Changes"
            string3 = "Maximum bond displacement [angstroms]= "
            CALL final_move_write(moves%bias_bond, string1, string2, iw, &
                                  displacement=rmbond(itype), lbias=lbias, format_string=format_string, &
                                  string3=string3)

            string1 = "Angle Changes"
            string3 = "Maximum angle displacement [degrees]= "
            CALL final_move_write(moves%bias_angle, string1, string2, iw, &
                                  displacement=rmangle(itype), lbias=lbias, format_string=format_string, &
                                  string3=string3)

            string1 = "Dihedral Changes"
            string3 = "Maximum dihedral displacement [degrees]= "
            CALL final_move_write(moves%bias_dihedral, string1, string2, iw, &
                                  displacement=rmdihedral(itype), lbias=lbias, format_string=format_string, &
                                  string3=string3)

            ! translation moves
            string1 = "Translation Moves"
            string3 = "Maximum molecular translational displacement [angstroms]= "
            CALL final_move_write(moves%bias_trans, string1, string2, iw, &
                                  displacement=rmtrans(itype), lbias=lbias, format_string=format_string, &
                                  string3=string3)

! rotation moves
            string1 = "Rotation Moves"
            string3 = "Maximum molecular rotational displacement [degrees]= "
            CALL final_move_write(moves%bias_rot, string1, string2, iw, &
                                  displacement=rmrot(itype), lbias=lbias, format_string=format_string, &
                                  string3=string3)

         ENDIF

      ENDDO

! see if the energies add up properly
      IF (ABS(initial_energy+energy_check-final_energy) .GT. 0.0000001E0_dp) &
         THEN
         WRITE (iw, *) '!!!!!!! We have an energy problem. !!!!!!!!'
         WRITE (iw, '(A,T64,F16.10)') 'Final Energy = ', final_energy
         WRITE (iw, '(A,T64,F16.10)') 'Inital Energy + energy_check =', &
            initial_energy+energy_check
      ENDIF
      WRITE (iw, '(A,A)') '****************************************************', &
         '****************************'
      WRITE (iw, *)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE final_mc_write

! **************************************************************************************************
!> \brief ...
!> \param move_data ...
!> \param string1 ...
!> \param string2 ...
!> \param iw ...
!> \param string3 ...
!> \param format_string ...
!> \param lbias ...
!> \param displacement ...
! **************************************************************************************************
   SUBROUTINE final_move_write(move_data, string1, string2, iw, string3, &
                               format_string, lbias, displacement)

      TYPE(accattempt), POINTER                          :: move_data
      CHARACTER(default_string_length), INTENT(IN)       :: string1, string2
      INTEGER, INTENT(IN)                                :: iw
      CHARACTER(default_string_length), INTENT(IN), &
         OPTIONAL                                        :: string3, format_string
      LOGICAL, INTENT(IN), OPTIONAL                      :: lbias
      REAL(dp), OPTIONAL                                 :: displacement

      IF (.NOT. PRESENT(format_string)) THEN
         IF (move_data%attempts .GT. 0) THEN
            WRITE (iw, '(A,T43,A)') TRIM(ADJUSTL(string1)), &
               TRIM(ADJUSTL(string2))
            WRITE (iw, '(T46,I6,9X,I6,7X,F7.3)') &
               move_data%attempts, &
               move_data%successes, &
               REAL(move_data%successes, dp)/ &
               REAL(move_data%attempts, dp)*100.0E0_dp
            WRITE (iw, '(A,A)') '-----------------------------------------------', &
               '---------------------------------'
         ENDIF
      ELSE
         IF (.NOT. PRESENT(string3) .OR. .NOT. PRESENT(lbias) .OR. &
             .NOT. PRESENT(displacement)) THEN
            WRITE (iw, *) 'MISSING FLAGS IN FINAL_MOVE_WRITE'
         ENDIF
         IF (move_data%attempts .GT. 0) THEN
            WRITE (iw, '(A,T43,A)') TRIM(ADJUSTL(string1)), &
               TRIM(ADJUSTL(string2))
            WRITE (iw, '(T46,I6,9X,I6,7X,F7.3)') &
               move_data%attempts, &
               move_data%successes, &
               REAL(move_data%successes, dp)/ &
               REAL(move_data%attempts, dp)*100.0E0_dp
            IF (.NOT. lbias) WRITE (iw, '(A,T71,F10.5)') &
               string3, displacement
            WRITE (iw, '(A,A)') '-----------------------------------------------', &
               '---------------------------------'
         ENDIF
      ENDIF

   END SUBROUTINE final_move_write

! **************************************************************************************************
!> \brief writes a new input file that CP2K can read in for when we want
!>      to change a force env (change molecule number)...this is much simpler
!>      than the version I had used to have, and also more flexible (in a way).
!>      It assumes that &CELL comes before &COORDS, and &COORDS comes before
!>      &TOPOLOGY, and &TOPOLOGY comes before &GLOBAL (which comes before MC).
!>      It also assumes that you use &MOL_SET in &TOPOLOGY.  Still, many fewer
!>      assumptions than before.
!>
!>      box_length and coordinates should be passed in a.u.
!> \param coordinates the coordiantes of the atoms in the force_env (a.u.)
!> \param atom_names ...
!> \param nunits_tot the total number of atoms
!> \param box_length the length of all sides of the simulation box (angstrom)
!> \param filename the name of the file to write to
!> \param nchains ...
!> \param mc_input_file ...
!> \author MJM
!> \note   Only use in serial.
! **************************************************************************************************
   SUBROUTINE mc_make_dat_file_new(coordinates, atom_names, nunits_tot, &
                                   box_length, filename, nchains, mc_input_file)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: coordinates
      CHARACTER(LEN=*), DIMENSION(:), INTENT(IN)         :: atom_names
      INTEGER, INTENT(IN)                                :: nunits_tot
      REAL(KIND=dp), DIMENSION(1:3), INTENT(IN)          :: box_length
      CHARACTER(LEN=*), INTENT(IN)                       :: filename
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nchains
      TYPE(mc_input_file_type), POINTER                  :: mc_input_file

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_make_dat_file_new', &
         routineP = moduleN//':'//routineN

      CHARACTER(60)                                      :: cell_string, mol_string
      CHARACTER(default_string_length)                   :: line_text
      CHARACTER(default_string_length), DIMENSION(:), &
         POINTER                                         :: atom_names_empty, text
      INTEGER :: cell_column, cell_row, coord_row_end, coord_row_start, global_row_end, iline, &
         in_use, itype, iunit, motion_row_end, motion_row_start, nmol_types, nunits_empty, &
         run_type_row, start_line, unit
      INTEGER, DIMENSION(:), POINTER                     :: mol_set_nmol_column, mol_set_nmol_row
      REAL(dp), DIMENSION(:, :), POINTER                 :: coordinates_empty

! open the file

      CALL open_file(file_name=filename, unit_number=unit, &
                     file_action='WRITE', file_status='REPLACE')

! get all the information from the input_file_type
      CALL get_mc_input_file(mc_input_file, text=text, cell_row=cell_row, &
                             cell_column=cell_column, coord_row_start=coord_row_start, &
                             coord_row_end=coord_row_end, mol_set_nmol_row=mol_set_nmol_row, &
                             mol_set_nmol_column=mol_set_nmol_column, global_row_end=global_row_end, &
                             run_type_row=run_type_row, in_use=in_use, atom_names_empty=atom_names_empty, &
                             nunits_empty=nunits_empty, coordinates_empty=coordinates_empty, &
                             motion_row_start=motion_row_start, motion_row_end=motion_row_end)

! how many molecule types?
      nmol_types = SIZE(nchains)

! first, write all the information up to the cell lengths
      DO iline = 1, cell_row-1
         WRITE (unit, '(A)') TRIM(text(iline))
      ENDDO
! substitute in the current cell lengths
      WRITE (cell_string, '(3(F13.8,2X))') box_length(1:3)*angstrom
      line_text = text(cell_row)
      line_text(cell_column:cell_column+50) = cell_string(1:51)
      WRITE (unit, '(A)') TRIM(line_text)

! now write everything until the coordinates
      DO iline = cell_row+1, coord_row_start
         WRITE (unit, '(A)') TRIM(text(iline))
      ENDDO

! we may pass nunits_tot=0, but we should still have coordinates
      IF (nunits_tot == 0) THEN
         DO iunit = 1, nunits_empty
            WRITE (unit, '(5X,A,2X,3(F15.10))') &
               TRIM(ADJUSTL(atom_names_empty(iunit))), &
               coordinates_empty(1:3, iunit)*angstrom
         ENDDO
      ELSE
         DO iunit = 1, nunits_tot
            WRITE (unit, '(5X,A,2X,3(F15.10))') &
               TRIM(ADJUSTL(atom_names(iunit))), &
               coordinates(1:3, iunit)*angstrom
         ENDDO
      ENDIF

! now we need to write the MOL_SET section
      start_line = coord_row_end
      DO itype = 1, nmol_types
         DO iline = start_line, mol_set_nmol_row(itype)-1
            WRITE (unit, '(A)') TRIM(text(iline))
         ENDDO

! have to print out one molecule, even if it's empty
         IF (nunits_tot == 0 .AND. itype == 1) THEN
            WRITE (mol_string, '(I8)') 1
         ELSE
            WRITE (mol_string, '(I8)') nchains(itype)
         ENDIF

         line_text = text(mol_set_nmol_row(itype))
         line_text(mol_set_nmol_column(itype):mol_set_nmol_column(itype)+9) = &
            mol_string(1:10)
         WRITE (unit, '(A)') TRIM(line_text)
         start_line = mol_set_nmol_row(itype)+1
      ENDDO

! write up to the RUN_TYPE...tailor this for the type of environment, so
! that we can easily do ./cp2k.sdbg input.dat and have it run
      DO iline = mol_set_nmol_row(nmol_types)+1, run_type_row-1
         WRITE (unit, '(A)') TRIM(text(iline))
      ENDDO
      SELECT CASE (in_use)
      CASE (use_fist_force)
         WRITE (unit, '(A)') '  RUN_TYPE     ENERGY_FORCE'
      CASE (use_qs_force)
         WRITE (unit, '(A)') '  RUN_TYPE     ENERGY_FORCE'
      END SELECT
      DO iline = run_type_row+1, global_row_end
         WRITE (unit, '(A)') TRIM(text(iline))
      ENDDO
      DO iline = motion_row_start, motion_row_end
         WRITE (unit, '(A)') TRIM(text(iline))
      ENDDO

! close the file
      CALL close_file(unit_number=unit)

   END SUBROUTINE MC_MAKE_DAT_FILE_NEW
END MODULE mc_misc

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains some general routines for dealing with the restart
!>      files and creating force_env for MC use
!> \par History
!>      none
!> \author MJM
! **************************************************************************************************
MODULE mc_control
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_logger_get_default_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE f77_interface,                   ONLY: create_force_env,&
                                              f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_retain,&
                                              force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE input_section_types,             ONLY: section_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              dp_size,&
                                              int_size
   USE mc_misc,                         ONLY: mc_make_dat_file_new
   USE mc_types,                        ONLY: get_mc_molecule_info,&
                                              get_mc_par,&
                                              mc_input_file_type,&
                                              mc_molecule_info_type,&
                                              mc_simpar_type,&
                                              set_mc_par
   USE message_passing,                 ONLY: mp_bcast
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE molecule_kind_types,             ONLY: atom_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE particle_list_types,             ONLY: particle_list_type
   USE physcon,                         ONLY: angstrom
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   ! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_control'

   PUBLIC :: write_mc_restart, read_mc_restart, mc_create_force_env, &
             mc_create_bias_force_env

CONTAINS

! **************************************************************************************************
!> \brief writes the coordinates of the current step to a file that can
!>      be read in at the start of the next simulation
!> \param nnstep how many steps the simulation has run
!>
!>    Only use in serial.
!> \param mc_par the mc parameters for the force env
!> \param nchains ...
!> \param force_env the force environment to write the coords from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE write_mc_restart(nnstep, mc_par, nchains, force_env)

      INTEGER, INTENT(IN)                                :: nnstep
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nchains
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'write_mc_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=20)                                  :: ensemble
      CHARACTER(LEN=default_path_length)                 :: restart_file_name
      CHARACTER(LEN=default_string_length)               :: name
      INTEGER                                            :: handle, ichain, imol_type, iparticle, &
                                                            iunit, natom, nmol_types, nmolecule, &
                                                            nunits_tot, unit
      REAL(KIND=dp)                                      :: temperature
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc
      TYPE(atom_type), DIMENSION(:), POINTER             :: atom_list
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds_new
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(particle_list_type), POINTER                  :: particles

      CALL timeset(routineN, handle)

      ! get some data from mc_par
      CALL get_mc_par(mc_par, restart_file_name=restart_file_name, temperature=temperature, &
                      ensemble=ensemble)

      ! open the file and write some simulation parameters
      CALL open_file(file_name=restart_file_name, unit_number=unit, &
                     file_action='WRITE', file_status='REPLACE')

      ! get the cell length and coordinates
      CALL force_env_get(force_env, cell=cell, subsys=subsys)
      CALL get_cell(cell, abc=abc)
      CALL cp_subsys_get(subsys, &
                         molecule_kinds_new=molecule_kinds_new, &
                         particles=particles)

      nunits_tot = SIZE(particles%els(:))
      IF (SUM(nchains(:)) == 0) nunits_tot = 0
      WRITE (unit, *) nnstep
      WRITE (unit, *) temperature, nunits_tot
      WRITE (unit, *) ensemble
      WRITE (unit, *) nchains(:)
      WRITE (unit, '(3(F10.6,3X))') abc(1:3)*angstrom ! in angstroms
      WRITE (unit, *)

      ! can't do a simple particles%els%atomic_kind%element_symbol because
      ! of the classical force_env
      IF (nunits_tot .GT. 0) THEN
         nmol_types = SIZE(molecule_kinds_new%els(:))
         iparticle = 1
         DO imol_type = 1, nmol_types
            molecule_kind => molecule_kinds_new%els(imol_type)
            CALL get_molecule_kind(molecule_kind, atom_list=atom_list, &
                                   nmolecule=nmolecule, natom=natom)
            ! write the coordinates out
            DO ichain = 1, nmolecule
               DO iunit = 1, natom
                  CALL get_atomic_kind(atom_list(iunit)%atomic_kind, name=name)
                  WRITE (unit, '(1X,A,1X,3(F15.10,3X))') &
                     TRIM(ADJUSTL(name)), &
                     particles%els(iparticle)%r(1:3)*angstrom
                  iparticle = iparticle+1
               ENDDO
            ENDDO
         ENDDO
      ENDIF

      CALL close_file(unit_number=unit)

      ! end the timing
      CALL timestop(handle)

   END SUBROUTINE write_mc_restart

! **************************************************************************************************
!> \brief reads the input coordinates of the simulation from a file written above
!> \param mc_par the mc parameters for the force env
!> \param force_env the force environment to write the coords from
!> \param iw the unit to write an error message to, in case current
!>            simulation parameters don't match what's in the restart file
!> \param mc_nunits_tot ...
!> \param rng_stream the stream we pull random numbers from
!>
!>      Used in parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE read_mc_restart(mc_par, force_env, iw, mc_nunits_tot, rng_stream)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env
      INTEGER, INTENT(IN)                                :: iw
      INTEGER, INTENT(INOUT)                             :: mc_nunits_tot
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'read_mc_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(5), ALLOCATABLE, DIMENSION(:)            :: atom_symbols
      CHARACTER(default_string_length), &
         DIMENSION(:, :), POINTER                        :: atom_names
      CHARACTER(LEN=20)                                  :: ensemble, mc_ensemble
      CHARACTER(LEN=default_path_length)                 :: dat_file, restart_file_name
      INTEGER                                            :: group, handle, i, ipart, istat, iunit, &
                                                            nmol_types, nstart, nunits_tot, &
                                                            print_level, source, unit
      INTEGER, DIMENSION(:), POINTER                     :: nchains, nunits
      LOGICAL                                            :: ionode
      REAL(KIND=dp)                                      :: mc_temp, rand, temperature
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc, box_length
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mc_input_file_type), POINTER                  :: mc_input_file
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(particle_list_type), POINTER                  :: particles

      CALL timeset(routineN, handle)

      ! get some stuff from the mc_par
      CALL get_mc_par(mc_par, restart_file_name=restart_file_name, temperature=mc_temp, &
                      ensemble=mc_ensemble, mc_molecule_info=mc_molecule_info, &
                      ionode=ionode, dat_file=dat_file, &
                      group=group, source=source, mc_input_file=mc_input_file)
      CALL get_mc_molecule_info(mc_molecule_info, nunits=nunits, &
                                nmol_types=nmol_types, atom_names=atom_names)

      ALLOCATE (nchains(1:nmol_types), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 212, &
                                       "nchains", nmol_types*int_size)

      ! currently a hack, printlevel should be intern to the print_keys
      print_level = 1

      IF (ionode) THEN
         ! open the file and read some simulation parameters
         CALL open_file(file_name=restart_file_name, unit_number=unit, &
                        file_action='READ', file_status='OLD')

         READ (unit, *) nstart
         READ (unit, *) temperature, nunits_tot
         READ (unit, *) ensemble
         READ (unit, *) nchains(1:nmol_types)
      ENDIF
      CALL mp_bcast(nstart, source, group)
      CALL mp_bcast(temperature, source, group)
      CALL mp_bcast(nunits_tot, source, group)
      CALL mp_bcast(ensemble, source, group)
      CALL mp_bcast(nchains, source, group)

      ! do some checking
      IF (ABS(temperature-mc_temp) .GT. 0.01E0_dp) THEN
         IF (ionode) THEN
            WRITE (iw, *) 'The temperature in the restart file is ', &
               'not the same as the input file.'
            WRITE (iw, *) 'Input file temperature =', mc_temp
            WRITE (iw, *) 'Restart file temperature =', temperature
         ENDIF
         CALL cp__b("motion/mc/mc_control.F",242,"Temperature difference between restart and input")
      ENDIF
      IF (nunits_tot .NE. mc_nunits_tot) THEN
         IF (ionode) THEN
            WRITE (iw, *) 'The total number of units in the restart ', &
               'file is not the same as the input file.'
            WRITE (iw, *) 'Input file units =', mc_nunits_tot
            WRITE (iw, *) 'Restart file units =', nunits_tot
         ENDIF
         mc_nunits_tot = nunits_tot
      ENDIF
      IF (ensemble .NE. mc_ensemble) THEN
         IF (ionode) THEN
            WRITE (iw, *) 'The ensemble in the restart file is ', &
               'not the same as the input file.'
            WRITE (iw, *) 'Input file ensemble =', mc_ensemble
            WRITE (iw, *) 'Restart file ensemble =', ensemble
         ENDIF
         CALL cp__b("motion/mc/mc_control.F",260,"Ensembles different between restart and input")
      ENDIF

      ! get the cell length and coordinates
      CALL force_env_get(force_env, cell=cell, subsys=subsys)
      CALL get_cell(cell, abc=abc)
      CALL cp_subsys_get(subsys, &
                         particles=particles)

      IF (ionode) THEN
         READ (unit, *) box_length(1:3) ! in angstroms
         READ (unit, *)
         box_length(1:3) = box_length(1:3)/angstrom ! convert to a.u.
      ENDIF
      CALL mp_bcast(box_length, source, group)
      IF (ABS(box_length(1)-abc(1)) .GT. 0.0001E0_dp .OR. &
          ABS(box_length(2)-abc(2)) .GT. 0.0001E0_dp .OR. &
          ABS(box_length(3)-abc(3)) .GT. 0.0001E0_dp) THEN
         IF (ionode) THEN
            WRITE (iw, *) 'The cell length in the restart file is ', &
               'not the same as the input file.'
            WRITE (iw, *) 'Input file cell length =', abc(1:3)*angstrom
            WRITE (iw, *) 'Restart file cell length =', box_length(1:3)*angstrom
         ENDIF
      ENDIF

      ! allocate the array holding the coordiantes, and read in the coordiantes,
      ! and write the dat file so we can make a new force_env
      IF (SUM(nchains(:)) == 0) THEN
         ALLOCATE (r(3, nunits(1)), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 290, &
                                          "r", 3*nunits(1)*dp_size)
         ALLOCATE (atom_symbols(nunits(1)), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 293, &
                                          "ipart", nunits_tot*5)

         DO iunit = 1, nunits(1)
            r(1:3, iunit) = (/REAL(iunit, dp), REAL(iunit, dp), REAL(iunit, dp)/)
            atom_symbols(iunit) = atom_names(iunit, 1)
         ENDDO

         IF (ionode) THEN
            CALL mc_make_dat_file_new(r(:, :), atom_symbols, 0, &
                                      box_length(:), dat_file, nchains(:), mc_input_file)
            CALL close_file(unit_number=unit)
         ENDIF
      ELSE
         ALLOCATE (r(3, nunits_tot), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 308, &
                                          "r", 3*nunits_tot*dp_size)
         ALLOCATE (atom_symbols(nunits_tot), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 311, &
                                          "ipart", nunits_tot*5)

         IF (ionode) THEN
            DO ipart = 1, nunits_tot
               READ (unit, *) atom_symbols(ipart), r(1:3, ipart)
               r(1:3, ipart) = r(1:3, ipart)/angstrom
            ENDDO

            CALL close_file(unit_number=unit)

            CALL mc_make_dat_file_new(r(:, :), atom_symbols, nunits_tot, &
                                      box_length(:), dat_file, nchains(:), mc_input_file)

         ENDIF
      ENDIF

      CALL set_mc_par(mc_par, nstart=nstart)

      ! advance the random number sequence based on the restart step
      IF (ionode) THEN
         DO i = 1, nstart+1
            rand = next_random_number(rng_stream)
         ENDDO
      ENDIF

      ! end the timing
      CALL timestop(handle)

      ! deallcoate
      DEALLOCATE (nchains)
      DEALLOCATE (r)
      DEALLOCATE (atom_symbols)

   END SUBROUTINE read_mc_restart

! **************************************************************************************************
!> \brief creates a force environment for any of the different kinds of
!>      MC simulations we can do (FIST, QS)
!> \param force_env the force environment to create
!> \param input_declaration ...
!> \param para_env ...
!> \param input_file_name ...
!> \param globenv_new the global environment parameters
!> \author MJM
!> \note   Suitable for parallel.
! **************************************************************************************************
   SUBROUTINE mc_create_force_env(force_env, input_declaration, para_env, input_file_name, &
                                  globenv_new)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_para_env_type), POINTER                    :: para_env
      CHARACTER(LEN=*), INTENT(IN)                       :: input_file_name
      TYPE(global_environment_type), OPTIONAL, POINTER   :: globenv_new

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_create_force_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: f_env_id, output_unit
      TYPE(f_env_type), POINTER                          :: f_env

      output_unit = cp_logger_get_default_unit_nr()
      CALL create_force_env(f_env_id, &
                            input_declaration=input_declaration, &
                            input_path=input_file_name, &
                            output_unit=output_unit, &
                            mpi_comm=para_env%group)

      CALL f_env_add_defaults(f_env_id, f_env)
      force_env => f_env%force_env
      CALL force_env_retain(force_env)
      CALL f_env_rm_defaults(f_env)

      IF (PRESENT(globenv_new)) &
         CALL force_env_get(force_env, globenv=globenv_new)

   END SUBROUTINE mc_create_force_env

! **************************************************************************************************
!> \brief essentially copies the cell size and coordinates of one force env
!>      to another that we will use to bias some moves with
!> \param bias_env the force environment to create
!> \param r ...
!> \param atom_symbols ...
!> \param nunits_tot ...
!> \param para_env ...
!> \param box_length ...
!> \param nchains ...
!> \param input_declaration ...
!> \param mc_input_file ...
!> \param ionode ...
!> \author MJM
!> \note   Suitable for parallel.
! **************************************************************************************************
   SUBROUTINE mc_create_bias_force_env(bias_env, r, atom_symbols, nunits_tot, &
                                       para_env, box_length, nchains, input_declaration, mc_input_file, ionode)

      TYPE(force_env_type), POINTER                      :: bias_env
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: r
      CHARACTER(default_string_length), DIMENSION(:), &
         INTENT(IN)                                      :: atom_symbols
      INTEGER, INTENT(IN)                                :: nunits_tot
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(1:3), INTENT(IN)          :: box_length
      INTEGER, DIMENSION(:), POINTER                     :: nchains
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(mc_input_file_type), POINTER                  :: mc_input_file
      LOGICAL, INTENT(IN)                                :: ionode

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_create_bias_force_env', &
         routineP = moduleN//':'//routineN

      IF (ionode) &
         CALL mc_make_dat_file_new(r(:, :), atom_symbols, nunits_tot, &
                                   box_length(:), 'bias_temp.dat', nchains(:), mc_input_file)

      CALL mc_create_force_env(bias_env, input_declaration, para_env, 'bias_temp.dat')

   END SUBROUTINE mc_create_bias_force_env

END MODULE mc_control

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief control the handling of the move data in Monte Carlo (MC) simulations
!> \par History
!>      none
!> \author Matthew J. McGrath  (10.16.2003)
! **************************************************************************************************
MODULE mc_move_control
  
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE mc_types,                        ONLY: get_mc_molecule_info,&
                                              get_mc_par,&
                                              mc_molecule_info_type,&
                                              mc_moves_type,&
                                              mc_simpar_type,&
                                              set_mc_par
   USE physcon,                         ONLY: angstrom
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

  PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_move_control'

  PUBLIC :: init_mc_moves, &
           mc_move_update,move_q_reinit,q_move_accept,mc_moves_release,&
           write_move_stats

CONTAINS

! **************************************************************************************************
!> \brief allocates and initializes the structure to record all move
!>      attempts/successes
!> \param moves the move structure to update
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
  SUBROUTINE init_mc_moves ( moves )

      TYPE(mc_moves_type), POINTER                       :: moves

      CHARACTER(len=*), PARAMETER :: routineN = 'init_mc_moves', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, stat

! begin the timing of the subroutine

    CALL timeset(routineN,handle)

! allocate all the structures
      ALLOCATE (moves,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,62,&
                                     "moves",0)
      ALLOCATE (moves%bond,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,65,&
                                     "moves%bond",0)
      ALLOCATE (moves%angle,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,68,&
                                     "moves%angle",0)
      ALLOCATE (moves%dihedral,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,71,&
                                     "moves%dihedral",0)
      ALLOCATE (moves%trans,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,74,&
                                     "moves%trans",0)
      ALLOCATE (moves%rot,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,77,&
                                     "moves%rot",0)
      ALLOCATE (moves%bias_bond,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,80,&
                                     "moves%bias_bond",0)
      ALLOCATE (moves%bias_angle,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,83,&
                                     "moves%bias_angle",0)
      ALLOCATE (moves%bias_dihedral,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,86,&
                                     "moves%bias_dihedral",0)
      ALLOCATE (moves%bias_trans,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,89,&
                                     "moves%bias_trans",0)
      ALLOCATE (moves%bias_rot,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,92,&
                                     "moves%bias_rot",0)
      ALLOCATE (moves%volume,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,95,&
                                     "moves%volume",0)
      ALLOCATE (moves%hmc,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,98,&
                                     "moves%hmc",0)
      ALLOCATE (moves%avbmc_inin,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,101,&
                                     "moves%avbmc_inin",0)
      ALLOCATE (moves%avbmc_inout,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,104,&
                                     "moves%avbmc_inout",0)
      ALLOCATE (moves%avbmc_outin,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,107,&
                                     "moves%avbmc_outin",0)
      ALLOCATE (moves%avbmc_outout,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,110,&
                                     "moves%avbmc_outout",0)
      ALLOCATE (moves%swap,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,113,&
                                     "moves%swap",0)
      ALLOCATE (moves%Quickstep,stat=stat)
      IF (stat /= 0) CALL stop_memory(routineN,moduleN,116,&
                                     "moves%Quickstep",0)
! set all the counters equal to zero
      moves%bias_bond%attempts=0
      moves%bias_bond%successes=0
      moves%bias_bond%qsuccesses=0
      moves%bias_angle%attempts=0
      moves%bias_angle%successes=0
      moves%bias_angle%qsuccesses=0
      moves%bias_dihedral%attempts=0
      moves%bias_dihedral%successes=0
      moves%bias_dihedral%qsuccesses=0
      moves%bias_trans%attempts=0
      moves%bias_trans%successes=0
      moves%bias_trans%qsuccesses=0
      moves%bias_rot%attempts=0
      moves%bias_rot%successes=0
      moves%bias_rot%qsuccesses=0

      moves%bond%attempts=0
      moves%bond%successes=0
      moves%bond%qsuccesses=0
      moves%angle%attempts=0
      moves%angle%successes=0
      moves%angle%qsuccesses=0
      moves%dihedral%attempts=0
      moves%dihedral%successes=0
      moves%dihedral%qsuccesses=0
      moves%trans%attempts=0
      moves%trans%successes=0
      moves%trans%qsuccesses=0
      moves%rot%attempts=0
      moves%rot%successes=0
      moves%rot%qsuccesses=0
      moves%avbmc_inin%attempts=0
      moves%avbmc_inin%successes=0
      moves%avbmc_inin%qsuccesses=0
      moves%avbmc_inout%attempts=0
      moves%avbmc_inout%successes=0
      moves%avbmc_inout%qsuccesses=0
      moves%avbmc_outin%attempts=0
      moves%avbmc_outin%successes=0
      moves%avbmc_outin%qsuccesses=0
      moves%avbmc_outout%attempts=0
      moves%avbmc_outout%successes=0
      moves%avbmc_outout%qsuccesses=0
      moves%volume%attempts=0
      moves%volume%successes=0
      moves%volume%qsuccesses=0
      moves%hmc%attempts=0
      moves%hmc%successes=0
      moves%hmc%qsuccesses=0
      moves%swap%attempts=0
      moves%swap%successes=0
      moves%swap%qsuccesses=0
      moves%Quickstep%attempts=0
      moves%Quickstep%successes=0
      moves%Quickstep%qsuccesses=0
      moves%trans_dis=0.0E0_dp
      moves%qtrans_dis=0.0E0_dp
      moves%empty=0
      moves%empty_conf=0
      moves%empty_avbmc=0
      moves%grown=0
!      moves%force_create=1

 ! end the timing
      CALL timestop(handle)

 END SUBROUTINE init_mc_moves

! **************************************************************************************************
!> \brief deallocates all the structures and nullifies the pointer
!> \param moves the move structure to release
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
  SUBROUTINE mc_moves_release ( moves )

      TYPE(mc_moves_type), POINTER                       :: moves

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_moves_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! begin the timing of the subroutine

    CALL timeset(routineN,handle)

! allocate all the structures
      DEALLOCATE (moves%bond)
      DEALLOCATE (moves%angle)
      DEALLOCATE (moves%dihedral)
      DEALLOCATE (moves%trans)
      DEALLOCATE (moves%rot)
      DEALLOCATE (moves%bias_bond)
      DEALLOCATE (moves%bias_angle)
      DEALLOCATE (moves%bias_dihedral)
      DEALLOCATE (moves%bias_trans)
      DEALLOCATE (moves%bias_rot)
      DEALLOCATE (moves%volume)
      DEALLOCATE (moves%hmc)
      DEALLOCATE (moves%avbmc_inin)
      DEALLOCATE (moves%avbmc_inout)
      DEALLOCATE (moves%avbmc_outin)
      DEALLOCATE (moves%avbmc_outout)
      DEALLOCATE (moves%swap)
      DEALLOCATE (moves%Quickstep)

      DEALLOCATE (moves)

! now nullify the moves
      NULLIFY(moves)

 ! end the timing
      CALL timestop(handle)

 END SUBROUTINE mc_moves_release

! **************************************************************************************************
!> \brief sets all qsuccess counters back to zero
!> \param moves the move structure to update
!> \param lbias are we biasing translations/rotations/conformational changes
!>        with a different potential?
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
SUBROUTINE move_q_reinit ( moves , lbias )

      TYPE(mc_moves_type), POINTER                       :: moves
      LOGICAL, INTENT(IN)                                :: lbias

      CHARACTER(len=*), PARAMETER :: routineN = 'move_q_reinit', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! begin the timing of the subroutine

      CALL timeset(routineN,handle)

! set all the counters equal to zero
      IF(lbias) THEN
         moves%bias_bond%qsuccesses=0
         moves%bias_angle%qsuccesses=0
         moves%bias_dihedral%qsuccesses=0
         moves%bias_trans%qsuccesses=0
         moves%bias_rot%qsuccesses=0
      ELSE
         moves%bond%qsuccesses=0
         moves%angle%qsuccesses=0
         moves%dihedral%qsuccesses=0
         moves%trans%qsuccesses=0
         moves%rot%qsuccesses=0
         moves%volume%qsuccesses=0
         moves%hmc%qsuccesses=0
         moves%qtrans_dis=0.0E0_dp
      ENDIF

 ! end the timing
      CALL timestop(handle)

 END SUBROUTINE move_q_reinit

! **************************************************************************************************
!> \brief updates accepted moves in the given structure...assumes you've been
!>      recording all successful moves in "qsuccesses"...this was done to
!>      compensate for doing multiple inner moves between Quickstep moves
!>      (which determine ultimate acceptance of moves)
!> \param moves the move structure to update
!> \param lbias are we biasing non-swap particle moves with a cheaper potential
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
SUBROUTINE q_move_accept(moves,lbias)

      TYPE(mc_moves_type), POINTER                       :: moves
      LOGICAL, INTENT(IN)                                :: lbias

      CHARACTER(len=*), PARAMETER :: routineN = 'q_move_accept', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! begin the timing of the subroutine

      CALL timeset(routineN,handle)

      IF(lbias) THEN
! change the number of successful moves for the total move counter
         moves%bias_bond%successes=moves%bias_bond%successes&
            +moves%bias_bond%qsuccesses
         moves%bias_angle%successes=moves%bias_angle%successes&
            +moves%bias_angle%qsuccesses
         moves%bias_dihedral%successes=moves%bias_dihedral%successes&
            +moves%bias_dihedral%qsuccesses
         moves%bias_trans%successes=moves%bias_trans%successes&
            +moves%bias_trans%qsuccesses
         moves%bias_rot%successes=moves%bias_rot%successes&
            +moves%bias_rot%qsuccesses
      ELSE
! change the number of successful moves for the total move counter
         moves%bond%successes=moves%bond%successes&
            +moves%bond%qsuccesses
         moves%angle%successes=moves%angle%successes&
            +moves%angle%qsuccesses
         moves%dihedral%successes=moves%dihedral%successes&
            +moves%dihedral%qsuccesses
         moves%trans%successes=moves%trans%successes&
            +moves%trans%qsuccesses
         moves%rot%successes=moves%rot%successes&
            +moves%rot%qsuccesses
         moves%hmc%successes=moves%hmc%successes&
            +moves%hmc%qsuccesses
         moves%volume%successes=moves%volume%successes&
            +moves%volume%qsuccesses
         moves%avbmc_inin%successes=moves%avbmc_inin%successes&
            +moves%avbmc_inin%qsuccesses
         moves%avbmc_inout%successes=moves%avbmc_inout%successes&
            +moves%avbmc_inout%qsuccesses
         moves%avbmc_outin%successes=moves%avbmc_outin%successes&
            +moves%avbmc_outin%qsuccesses
         moves%avbmc_outout%successes=moves%avbmc_outout%successes&
            +moves%avbmc_outout%qsuccesses

         moves%trans_dis=moves%trans_dis+moves%qtrans_dis
      ENDIF

! end the timing
      CALL timestop(handle)

 END SUBROUTINE q_move_accept

! **************************************************************************************************
!> \brief writes the number of accepted and attempted moves to a file for
!>      the various move types
!> \param moves the structure containing the move data
!> \param nnstep what step we're on
!> \param unit the unit of the file we're writing to
!>
!>    Use only in serial.
!> \author MJM
! **************************************************************************************************
SUBROUTINE write_move_stats(moves,nnstep,unit)

      TYPE(mc_moves_type), POINTER                       :: moves
      INTEGER, INTENT(IN)                                :: nnstep, unit

      CHARACTER(len=*), PARAMETER :: routineN = 'write_move_stats', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! begin the timing of the subroutine

    CALL timeset(routineN,handle)

    WRITE(unit,1000) nnstep,' bias_bond      ',&
      moves%bias_bond%successes,moves%bias_bond%attempts
    WRITE(unit,1000) nnstep,' bias_angle      ',&
      moves%bias_angle%successes,moves%bias_angle%attempts
    WRITE(unit,1000) nnstep,' bias_dihedral      ',&
      moves%bias_dihedral%successes,moves%bias_dihedral%attempts
    WRITE(unit,1000) nnstep,' bias_trans      ',&
      moves%bias_trans%successes,moves%bias_trans%attempts
    WRITE(unit,1000) nnstep,' bias_rot      ',&
      moves%bias_rot%successes,moves%bias_rot%attempts

    WRITE(unit,1000) nnstep,' bond      ',&
      moves%bond%successes,moves%bond%attempts
    WRITE(unit,1000) nnstep,' angle     ',&
      moves%angle%successes,moves%angle%attempts
    WRITE(unit,1000) nnstep,' dihedral     ',&
      moves%dihedral%successes,moves%dihedral%attempts
    WRITE(unit,1000) nnstep,' trans     ',&
      moves%trans%successes,moves%trans%attempts
    WRITE(unit,1000) nnstep,' rot       ',&
      moves%rot%successes,moves%rot%attempts
    WRITE(unit,1000) nnstep,' swap      ',&
      moves%swap%successes,moves%swap%attempts
    WRITE(unit,1001) nnstep,' grown     ',&
      moves%grown
    WRITE(unit,1001) nnstep,' empty_swap     ',&
      moves%empty
    WRITE(unit,1001) nnstep,' empty_conf     ',&
      moves%empty_conf
    WRITE(unit,1000) nnstep,' volume    ',&
      moves%volume%successes,moves%volume%attempts
    WRITE(unit,1000) nnstep,' HMC    ',&
      moves%hmc%successes,moves%hmc%attempts
    WRITE(unit,1000) nnstep,' avbmc_inin  ',&
      moves%avbmc_inin%successes,moves%avbmc_inin%attempts
    WRITE(unit,1000) nnstep,' avbmc_inout  ',&
      moves%avbmc_inout%successes,moves%avbmc_inout%attempts
    WRITE(unit,1000) nnstep,' avbmc_outin  ',&
      moves%avbmc_outin%successes,moves%avbmc_outin%attempts
    WRITE(unit,1000) nnstep,' avbmc_outout ',&
      moves%avbmc_outout%successes,moves%avbmc_outout%attempts
    WRITE(unit,1001) nnstep,' empty_avbmc     ',&
      moves%empty_avbmc
    WRITE(unit,1000) nnstep,' Quickstep ',&
      moves%quickstep%successes,moves%quickstep%attempts

1000 FORMAT(I10,2X,A,2X,I10,2X,I10)
1001 FORMAT(I10,2X,A,2X,I10)
! end the timing
      CALL timestop(handle)

 END SUBROUTINE write_move_stats

! **************************************************************************************************
!> \brief updates the maximum displacements of a Monte Carlo simulation,
!>      based on the ratio of successful moves to attempts...tries to hit a
!>      target of 0.5 acceptance ratio
!> \param mc_par the mc parameters for the force env
!> \param move_updates holds the accepted/attempted moves since the last
!>              update (or start of simulation)
!> \param molecule_type ...
!> \param flag indicates which displacements to update..."volume" is for
!>              volume moves and "trans" is for everything else
!> \param nnstep how many steps the simulation has run
!> \param ionode is this the main CPU running this job?
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
 SUBROUTINE mc_move_update ( mc_par,move_updates,molecule_type,flag,&
      nnstep,ionode )

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(mc_moves_type), POINTER                       :: move_updates
      INTEGER, INTENT(IN)                                :: molecule_type
      CHARACTER(LEN=*), INTENT(IN)                       :: flag
      INTEGER, INTENT(IN)                                :: nnstep
      LOGICAL, INTENT(IN)                                :: ionode

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_move_update', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nmol_types, rm
      REAL(dp), DIMENSION(:), POINTER                    :: rmangle, rmbond, rmdihedral, rmrot, &
                                                            rmtrans
      REAL(KIND=dp)                                      :: rmvolume, test_ratio
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info

! begin the timing of the subroutine

      CALL timeset(routineN,handle)

      NULLIFY(rmangle,rmbond,rmdihedral,rmrot,rmtrans)

! grab some stuff from mc_par
      CALL get_mc_par(mc_par,rmbond=rmbond,rmangle=rmangle,rmrot=rmrot,&
         rmtrans=rmtrans,rmvolume=rmvolume,rm=rm,rmdihedral=rmdihedral,&
         mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info,nmol_types=nmol_types)

      SELECT CASE (flag)
      CASE DEFAULT
         WRITE(*,*) 'flag =',flag
         CALL cp__b("motion/mc/mc_move_control.F",477,"Wrong option passed")
      CASE ("trans")

! we need to update all the displacements for every molecule type
         IF(ionode) WRITE(rm,*) nnstep,' Data for molecule type ',&
              molecule_type

! update the maximum displacement for bond length change
         IF( move_updates%bias_bond%attempts .GT. 0 ) THEN

! first account for the extreme cases
            IF ( move_updates%bias_bond%successes == 0 ) THEN
               rmbond(molecule_type)=rmbond(molecule_type)/2.0E0_dp
            ELSEIF ( move_updates%bias_bond%successes == &
                 move_updates%bias_bond%attempts ) THEN
               rmbond(molecule_type)=rmbond(molecule_type)*2.0E0_dp
            ELSE
! now for the middle case
               test_ratio=REAL(move_updates%bias_bond%successes,dp)&
                    /REAL(move_updates%bias_bond%attempts,dp)/0.5E0_dp
               IF (test_ratio .GT. 2.0E0_dp) test_ratio=2.0E0_dp
               IF (test_ratio .LT. 0.5E0_dp) test_ratio=0.5E0_dp
               rmbond(molecule_type)=rmbond(molecule_type)*test_ratio
            ENDIF

! update and clear the counters
            move_updates%bias_bond%attempts=0
            move_updates%bias_bond%successes=0

! write the new displacement to a file
            IF(ionode) WRITE(rm,*) nnstep,' rmbond = ',&
                 rmbond(molecule_type)*angstrom,' angstroms'

         ENDIF

! update the maximum displacement for bond angle change
         IF( move_updates%bias_angle%attempts .GT. 0 ) THEN

! first account for the extreme cases
            IF ( move_updates%bias_angle%successes == 0 ) THEN
               rmangle(molecule_type)=rmangle(molecule_type)/2.0E0_dp
            ELSEIF ( move_updates%bias_angle%successes == &
                 move_updates%bias_angle%attempts ) THEN
               rmangle(molecule_type)=rmangle(molecule_type)*2.0E0_dp
            ELSE
! now for the middle case
               test_ratio=REAL(move_updates%bias_angle%successes,dp)&
                    /REAL(move_updates%bias_angle%attempts,dp)/0.5E0_dp
               IF (test_ratio .GT. 2.0E0_dp) test_ratio=2.0E0_dp
               IF (test_ratio .LT. 0.5E0_dp) test_ratio=0.5E0_dp
               rmangle(molecule_type)=rmangle(molecule_type)*test_ratio
            ENDIF

! more than pi changes meaningless
            IF (rmangle(molecule_type) .GT. pi) rmangle(molecule_type)=pi

! clear the counters
            move_updates%bias_angle%attempts=0
            move_updates%bias_angle%successes=0

! write the new displacement to a file
            IF(ionode) WRITE(rm,*) nnstep,' rmangle = ',&
                 rmangle(molecule_type)/pi*180.0E0_dp,' degrees'
         ENDIF

! update the maximum displacement for a dihedral change
         IF( move_updates%bias_dihedral%attempts .GT. 0 ) THEN

! first account for the extreme cases
            IF ( move_updates%bias_dihedral%successes == 0 ) THEN
               rmdihedral(molecule_type)=rmdihedral(molecule_type)/2.0E0_dp
            ELSEIF ( move_updates%bias_dihedral%successes == &
                 move_updates%bias_dihedral%attempts ) THEN
               rmdihedral(molecule_type)=rmdihedral(molecule_type)*2.0E0_dp
            ELSE
! now for the middle case
               test_ratio=REAL(move_updates%bias_dihedral%successes,dp)&
                    /REAL(move_updates%bias_dihedral%attempts,dp)/0.5E0_dp
               IF (test_ratio .GT. 2.0E0_dp) test_ratio=2.0E0_dp
               IF (test_ratio .LT. 0.5E0_dp) test_ratio=0.5E0_dp
               rmdihedral(molecule_type)=rmdihedral(molecule_type)*test_ratio
            ENDIF

! more than pi changes meaningless
            IF (rmdihedral(molecule_type) .GT. pi) rmdihedral(molecule_type)=pi

! clear the counters
            move_updates%bias_dihedral%attempts=0
            move_updates%bias_dihedral%successes=0

! write the new displacement to a file
            IF(ionode) WRITE(rm,*) nnstep,' rmdihedral = ',&
                 rmdihedral(molecule_type)/pi*180.0E0_dp,' degrees'
         ENDIF

! update the maximum displacement for molecule translation
         IF( move_updates%bias_trans%attempts .GT. 0 ) THEN

! first account for the extreme cases
            IF ( move_updates%bias_trans%successes == 0 ) THEN
               rmtrans(molecule_type)=rmtrans(molecule_type)/2.0E0_dp
            ELSEIF ( move_updates%bias_trans%successes == &
                 move_updates%bias_trans%attempts ) THEN
               rmtrans(molecule_type)=rmtrans(molecule_type)*2.0E0_dp
            ELSE
! now for the middle case
               test_ratio=REAL(move_updates%bias_trans%successes,dp)&
                    /REAL(move_updates%bias_trans%attempts,dp)/0.5E0_dp
               IF (test_ratio .GT. 2.0E0_dp) test_ratio=2.0E0_dp
               IF (test_ratio .LT. 0.5E0_dp) test_ratio=0.5E0_dp
               rmtrans(molecule_type)=rmtrans(molecule_type)*test_ratio
            ENDIF

 ! make an upper bound...10 a.u.
            IF (rmtrans(molecule_type) .GT. 10.0E0_dp) &
                 rmtrans(molecule_type )= 10.0E0_dp

 ! clear the counters
            move_updates%bias_trans%attempts=0
            move_updates%bias_trans%successes=0

! write the new displacement to a file
            IF(ionode) WRITE(rm,*) nnstep,' rmtrans = ',&
                 rmtrans(molecule_type)*angstrom,' angstroms'
         ENDIF

! update the maximum displacement for molecule rotation
         IF( move_updates%bias_rot%attempts .GT. 0 ) THEN

! first account for the extreme cases
            IF ( move_updates%bias_rot%successes == 0 ) THEN
               rmrot=rmrot/2.0E0_dp

               IF (rmrot(molecule_type) .GT. pi) rmrot(molecule_type)=pi

            ELSEIF ( move_updates%bias_rot%successes == &
                 move_updates%bias_rot%attempts ) THEN
               rmrot(molecule_type)=rmrot(molecule_type)*2.0E0_dp

! more than pi rotation is meaningless
               IF (rmrot(molecule_type) .GT. pi) rmrot(molecule_type)=pi

            ELSE
! now for the middle case
               test_ratio=REAL(move_updates%bias_rot%successes,dp)&
                    /REAL(move_updates%bias_rot%attempts,dp)/0.5E0_dp
               IF (test_ratio .GT. 2.0E0_dp) test_ratio=2.0E0_dp
               IF (test_ratio .LT. 0.5E0_dp) test_ratio=0.5E0_dp
               rmrot(molecule_type)=rmrot(molecule_type)*test_ratio

! more than pi rotation is meaningless
               IF (rmrot(molecule_type) .GT. pi) rmrot(molecule_type)=pi

            ENDIF

! clear the counters
            move_updates%bias_rot%attempts=0
            move_updates%bias_rot%successes=0

! write the new displacement to a file
            IF(ionode) WRITE(rm,*) nnstep,' rmrot = ',&
                 rmrot(molecule_type)/pi*180.0E0_dp,' degrees'
         ENDIF

      CASE ("volume")

! update the maximum displacement for volume displacement
         IF ( move_updates%volume%attempts .NE. 0) THEN

! first account for the extreme cases
            IF ( move_updates%volume%successes == 0 ) THEN
               rmvolume=rmvolume/2.0E0_dp

            ELSEIF ( move_updates%volume%successes == &
                  move_updates%volume%attempts ) THEN
               rmvolume=rmvolume*2.0E0_dp
            ELSE
! now for the middle case
               test_ratio=REAL(move_updates%volume%successes,dp)/&
                  REAL(move_updates%volume%attempts,dp)/0.5E0_dp
               IF (test_ratio .GT. 2.0E0_dp) test_ratio=2.0E0_dp
               IF (test_ratio .LT. 0.5E0_dp) test_ratio=0.5E0_dp
               rmvolume=rmvolume*test_ratio

            ENDIF

! clear the counters
            move_updates%volume%attempts=0
            move_updates%volume%successes=0

! write the new displacement to a file
            IF(ionode) WRITE(rm,*) nnstep,' rmvolume = ',&
                    rmvolume*angstrom**3,' angstroms^3'

         ENDIF

      END SELECT

! set some of the MC parameters
      CALL set_mc_par(mc_par,rmbond=rmbond,rmangle=rmangle,rmrot=rmrot,&
         rmtrans=rmtrans,rmvolume=rmvolume,rmdihedral=rmdihedral)

! end the timing
      CALL timestop(handle)

  END SUBROUTINE mc_move_update

END MODULE mc_move_control

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains the Monte Carlo moves that can handle more than one
!>      box, including the Quickstep move, a volume swap between boxes,
!>       and a particle swap between boxes
!> \par History
!>      MJM (07.28.2005): make the Quickstep move general, and changed
!>                        the swap and volume moves to work with the
!>                        CP2K classical routines
!> \author Matthew J. McGrath  (01.25.2004)
! **************************************************************************************************
MODULE mc_ge_moves
   USE cell_types,                      ONLY: cell_clone,&
                                              cell_create,&
                                              cell_p_type,&
                                              cell_release,&
                                              cell_type,&
                                              get_cell
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_p_type,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_p_type,&
                                              force_env_release
   USE input_constants,                 ONLY: dump_xmol
   USE input_section_types,             ONLY: section_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              dp_size
   USE mc_control,                      ONLY: mc_create_force_env
   USE mc_coordinates,                  ONLY: check_for_overlap,&
                                              generate_cbmc_swap_config,&
                                              get_center_of_mass
   USE mc_misc,                         ONLY: mc_make_dat_file_new
   USE mc_move_control,                 ONLY: move_q_reinit,&
                                              q_move_accept
   USE mc_types,                        ONLY: &
        get_mc_molecule_info, get_mc_par, mc_determine_molecule_info, mc_input_file_type, &
        mc_molecule_info_destroy, mc_molecule_info_type, mc_moves_p_type, &
        mc_simulation_parameters_p_type, set_mc_par
   USE message_passing,                 ONLY: mp_bcast
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE particle_list_types,             ONLY: particle_list_p_type,&
                                              particle_list_type
   USE particle_methods,                ONLY: write_particle_coordinates
   USE physcon,                         ONLY: angstrom
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_ge_moves'

   PUBLIC :: mc_ge_volume_move, mc_ge_swap_move, &
             mc_quickstep_move

CONTAINS

! **************************************************************************************************
!> \brief computes the acceptance of a series of biased or unbiased moves
!>      (translation, rotation, conformational changes)
!> \param mc_par the mc parameters for the force envs of the boxes
!> \param force_env the force environments for the boxes
!> \param bias_env the force environments with the biasing potential for the boxes
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected for both boxes
!> \param lreject automatically rejects the move (used when an overlap occurs in
!>                the sequence of moves)
!> \param move_updates the structure that keeps track of how many moves have
!>               been accepted/rejected since the last time the displacements
!>               were updated for both boxes
!> \param energy_check the running total of how much the energy has changed
!>                      since the initial configuration
!> \param r_old the coordinates of the last accepted move before the sequence
!>         whose acceptance is determined by this call
!> \param nnstep the Monte Carlo step we're on
!> \param old_energy the energy of the last accepted move involving the full potential
!> \param bias_energy_new the energy of the current configuration involving the bias potential
!> \param last_bias_energy ...
!> \param nboxes the number of boxes (force environments) in the system
!> \param box_flag indicates if a move has been tried in a given box..if not, we don't
!>        recompute the energy
!> \param subsys the pointers for the particle subsystems of both boxes
!> \param particles the pointers for the particle sets
!> \param rng_stream the stream we pull random numbers from
!> \param unit_conv ...
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_Quickstep_move(mc_par, force_env, bias_env, moves, &
                                lreject, move_updates, energy_check, r_old, &
                                nnstep, old_energy, bias_energy_new, last_bias_energy, &
                                nboxes, box_flag, subsys, particles, rng_stream, &
                                unit_conv)

      TYPE(mc_simulation_parameters_p_type), &
         DIMENSION(:), POINTER                           :: mc_par
      TYPE(force_env_p_type), DIMENSION(:), POINTER      :: force_env, bias_env
      TYPE(mc_moves_p_type), DIMENSION(:, :), POINTER    :: moves
      LOGICAL, INTENT(IN)                                :: lreject
      TYPE(mc_moves_p_type), DIMENSION(:, :), POINTER    :: move_updates
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: energy_check
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: r_old
      INTEGER, INTENT(IN)                                :: nnstep
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: old_energy, bias_energy_new, &
                                                            last_bias_energy
      INTEGER, INTENT(IN)                                :: nboxes
      INTEGER, DIMENSION(:), INTENT(IN)                  :: box_flag
      TYPE(cp_subsys_p_type), DIMENSION(:), POINTER      :: subsys
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      REAL(KIND=dp), INTENT(IN)                          :: unit_conv

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_Quickstep_move', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: end_mol, group, handle, ibox, iparticle, &
                                                            iprint, istat, itype, jbox, &
                                                            nmol_types, source, start_mol
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, nunits, nunits_tot
      INTEGER, DIMENSION(1:nboxes)                       :: diff
      LOGICAL                                            :: ionode, lbias, loverlap
      REAL(KIND=dp)                                      :: BETA, energies, rand, w
      REAL(KIND=dp), DIMENSION(1:nboxes)                 :: bias_energy_old, new_energy
      TYPE(cp_subsys_p_type), DIMENSION(:), POINTER      :: subsys_bias
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles_bias

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (subsys_bias, particles_bias)

! get a bunch of data from mc_par
      CALL get_mc_par(mc_par(1)%mc_par, ionode=ionode, lbias=lbias, &
                      BETA=BETA, diff=diff(1), source=source, group=group, &
                      iprint=iprint, &
                      mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nmol_types=nmol_types, &
                                nchains=nchains, nunits_tot=nunits_tot, nunits=nunits, mol_type=mol_type)

      IF (nboxes .GT. 1) THEN
         DO ibox = 2, nboxes
            CALL get_mc_par(mc_par(ibox)%mc_par, diff=diff(ibox))
         ENDDO
      ENDIF

! allocate some stuff
      ALLOCATE (subsys_bias(1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 163, &
                                       "subsys_bias", nboxes)
      ALLOCATE (particles_bias(1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 166, &
                                       "particles_bias", nboxes)

! record the attempt...we really only care about molecule type 1 and box
! type 1, since the acceptance will be identical for all boxes and molecules
      moves(1, 1)%moves%Quickstep%attempts = &
         moves(1, 1)%moves%Quickstep%attempts+1

! grab the coordinates for the force_env
      DO ibox = 1, nboxes
         CALL force_env_get(force_env(ibox)%force_env, &
                            subsys=subsys(ibox)%subsys)
         CALL cp_subsys_get(subsys(ibox)%subsys, &
                            particles=particles(ibox)%list)
      ENDDO

! calculate the new energy of the system...if we're biasing,
! force_env hasn't changed but bias_env has
      DO ibox = 1, nboxes
         IF (box_flag(ibox) == 1) THEN
            IF (lbias) THEN
! grab the coords from bias_env and put them into force_env
               CALL force_env_get(bias_env(ibox)%force_env, &
                                  subsys=subsys_bias(ibox)%subsys)
               CALL cp_subsys_get(subsys_bias(ibox)%subsys, &
                                  particles=particles_bias(ibox)%list)

               DO iparticle = 1, nunits_tot(ibox)
                  particles(ibox)%list%els(iparticle)%r(1:3) = &
                     particles_bias(ibox)%list%els(iparticle)%r(1:3)
               ENDDO

               CALL force_env_calc_energy_force(force_env(ibox)%force_env, &
                                                calc_force=.FALSE.)
               CALL force_env_get(force_env(ibox)%force_env, &
                                  potential_energy=new_energy(ibox))
            ELSE
               IF (.NOT. lreject) THEN
                  CALL force_env_calc_energy_force(force_env(ibox)%force_env, &
                                                   calc_force=.FALSE.)
                  CALL force_env_get(force_env(ibox)%force_env, &
                                     potential_energy=new_energy(ibox))
               ENDIF
            ENDIF
         ELSE
            new_energy(ibox) = old_energy(ibox)
         ENDIF

      ENDDO

! accept or reject the move based on Metropolis or the Iftimie rule
      IF (ionode) THEN

! write them out in case something bad happens
         IF (MOD(nnstep, iprint) == 0) THEN
            DO ibox = 1, nboxes
               IF (SUM(nchains(:, ibox)) == 0) THEN
                  WRITE (diff(ibox), *) nnstep
                  WRITE (diff(ibox), *) nchains(:, ibox)
               ELSE
                  WRITE (diff(ibox), *) nnstep
                  CALL write_particle_coordinates( &
                     particles(ibox)%list%els, &
                     diff(ibox), dump_xmol, 'POS', 'TRIAL', &
                     unit_conv=unit_conv)
               ENDIF
            ENDDO
         ENDIF
      ENDIF

      IF (.NOT. lreject) THEN
         IF (lbias) THEN

            DO ibox = 1, nboxes
! look for overlap
               IF (SUM(nchains(:, ibox)) .NE. 0) THEN
! find the molecule bounds
                  start_mol = 1
                  DO jbox = 1, ibox-1
                     start_mol = start_mol+SUM(nchains(:, jbox))
                  ENDDO
                  end_mol = start_mol+SUM(nchains(:, ibox))-1
                  CALL check_for_overlap(bias_env(ibox)%force_env, &
                                         nchains(:, ibox), nunits(:), loverlap, mol_type(start_mol:end_mol))
                  IF (loverlap) &
                     CALL cp__b("motion/mc/mc_ge_moves.F",251,'Quickstep move found an overlap in the old config')
               ENDIF
               bias_energy_old(ibox) = last_bias_energy(ibox)
            ENDDO

            energies = -BETA*((SUM(new_energy(:))-SUM(bias_energy_new(:))) &
                              -(SUM(old_energy(:))-SUM(bias_energy_old(:))))

! used to prevent over and underflows
            IF (energies .GE. -1.0E-8) THEN
               w = 1.0_dp
            ELSEIF (energies .LE. -500.0_dp) THEN
               w = 0.0_dp
            ELSE
               w = EXP(energies)
            ENDIF

            IF (ionode) THEN
               DO ibox = 1, nboxes
                  WRITE (diff(ibox), *) nnstep, new_energy(ibox)- &
                     old_energy(ibox), &
                     bias_energy_new(ibox)-bias_energy_old(ibox)
               ENDDO
            ENDIF
         ELSE
            energies = -BETA*(SUM(new_energy(:))-SUM(old_energy(:)))
! used to prevent over and underflows
            IF (energies .GE. 0.0_dp) THEN
               w = 1.0_dp
            ELSEIF (energies .LE. -500.0_dp) THEN
               w = 0.0_dp
            ELSE
               w = EXP(energies)
            ENDIF
         ENDIF
      ELSE
         w = 0.0E0_dp
      ENDIF
      IF (w .GE. 1.0E0_dp) THEN
         w = 1.0E0_dp
         rand = 0.0E0_dp
      ELSE
         IF (ionode) rand = next_random_number(rng_stream)
         CALL mp_bcast(rand, source, group)
      ENDIF

      IF (rand .LT. w) THEN

! accept the move
         moves(1, 1)%moves%Quickstep%successes = &
            moves(1, 1)%moves%Quickstep%successes+1

         DO ibox = 1, nboxes
! remember what kind of move we did for lbias=.false.
            IF (.NOT. lbias) THEN
               DO itype = 1, nmol_types
                  CALL q_move_accept(moves(itype, ibox)%moves, .TRUE.)
                  CALL q_move_accept(move_updates(itype, ibox)%moves, .TRUE.)

! reset the counters
                  CALL move_q_reinit(moves(itype, ibox)%moves, .TRUE.)
                  CALL move_q_reinit(move_updates(itype, ibox)%moves, .TRUE.)
               ENDDO
            ENDIF

            DO itype = 1, nmol_types
! we need to record all accepted moves since last Quickstep calculation
               CALL q_move_accept(moves(itype, ibox)%moves, .FALSE.)
               CALL q_move_accept(move_updates(itype, ibox)%moves, .FALSE.)

! reset the counters
               CALL move_q_reinit(moves(itype, ibox)%moves, .FALSE.)
               CALL move_q_reinit(move_updates(itype, ibox)%moves, .FALSE.)
            ENDDO

! update energies
            energy_check(ibox) = energy_check(ibox)+ &
                                 (new_energy(ibox)-old_energy(ibox))
            old_energy(ibox) = new_energy(ibox)

         ENDDO

         IF (lbias) THEN
            DO ibox = 1, nboxes
               last_bias_energy(ibox) = bias_energy_new(ibox)
            ENDDO
         ENDIF

! update coordinates
         DO ibox = 1, nboxes
            IF (nunits_tot(ibox) .NE. 0) THEN
               DO iparticle = 1, nunits_tot(ibox)
                  r_old(1:3, iparticle, ibox) = &
                     particles(ibox)%list%els(iparticle)%r(1:3)
               ENDDO
            ENDIF
         ENDDO

      ELSE

         ! reject the move
         DO ibox = 1, nboxes
            DO itype = 1, nmol_types
               CALL move_q_reinit(moves(itype, ibox)%moves, .FALSE.)
               CALL move_q_reinit(move_updates(itype, ibox)%moves, .FALSE.)
               IF (.NOT. lbias) THEN
! reset the counters
                  CALL move_q_reinit(moves(itype, ibox)%moves, .TRUE.)
                  CALL move_q_reinit(move_updates(itype, ibox)%moves, .TRUE.)
               ENDIF
            ENDDO

         ENDDO

         IF (.NOT. ionode) r_old(:, :, :) = 0.0E0_dp

! coodinates changed, so we need to broadcast those, even for the lbias
! case since bias_env needs to have the same coords as force_env
         CALL mp_bcast(r_old, source, group)

         DO ibox = 1, nboxes
            DO iparticle = 1, nunits_tot(ibox)
               particles(ibox)%list%els(iparticle)%r(1:3) = &
                  r_old(1:3, iparticle, ibox)
               IF (lbias .AND. box_flag(ibox) == 1) &
                  particles_bias(ibox)%list%els(iparticle)%r(1:3) = &
                  r_old(1:3, iparticle, ibox)
            ENDDO
         ENDDO

! need to reset the energies of the biasing potential
         IF (lbias) THEN
            DO ibox = 1, nboxes
               bias_energy_new(ibox) = last_bias_energy(ibox)
            ENDDO
         ENDIF

      ENDIF

! make sure the coordinates are transferred
      DO ibox = 1, nboxes
         CALL cp_subsys_set(subsys(ibox)%subsys, &
                            particles=particles(ibox)%list)
         IF (lbias .AND. box_flag(ibox) == 1) &
            CALL cp_subsys_set(subsys_bias(ibox)%subsys, &
                               particles=particles_bias(ibox)%list)
      ENDDO

      ! deallocate some stuff
      DEALLOCATE (subsys_bias)
      DEALLOCATE (particles_bias)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_Quickstep_move

! **************************************************************************************************
!> \brief attempts a swap move between two simulation boxes
!> \param mc_par the mc parameters for the force envs of the boxes
!> \param force_env the force environments for the boxes
!> \param bias_env the force environments used to bias moves for the boxes
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected for both boxes
!> \param energy_check the running total of how much the energy has changed
!>                      since the initial configuration
!> \param r_old the coordinates of the last accepted move involving a
!>               full potential calculation for both boxes
!> \param old_energy the energy of the last accepted move involving a
!>                    a full potential calculation
!> \param input_declaration ...
!> \param para_env the parallel environment for this simulation
!> \param bias_energy_old the energies of both boxes computed using the biasing
!>        potential
!> \param last_bias_energy the energy for the biased simulations
!> \param rng_stream the stream we pull random numbers from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_ge_swap_move(mc_par, force_env, bias_env, moves, &
                              energy_check, r_old, old_energy, input_declaration, &
                              para_env, bias_energy_old, last_bias_energy, &
                              rng_stream)

      TYPE(mc_simulation_parameters_p_type), &
         DIMENSION(:), POINTER                           :: mc_par
      TYPE(force_env_p_type), DIMENSION(:), POINTER      :: force_env, bias_env
      TYPE(mc_moves_p_type), DIMENSION(:, :), POINTER    :: moves
      REAL(KIND=dp), DIMENSION(1:2), INTENT(INOUT)       :: energy_check
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: r_old
      REAL(KIND=dp), DIMENSION(1:2), INTENT(INOUT)       :: old_energy
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(1:2), INTENT(INOUT)       :: bias_energy_old, last_bias_energy
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_ge_swap_move', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_string_length), ALLOCATABLE, &
         DIMENSION(:)                                    :: atom_names_insert, atom_names_remove
      CHARACTER(default_string_length), &
         DIMENSION(:, :), POINTER                        :: atom_names
      CHARACTER(LEN=200)                                 :: fft_lib
      CHARACTER(LEN=40), DIMENSION(1:2)                  :: dat_file
      INTEGER :: end_mol, group, handle, iatom, ibox, idim, iiatom, imolecule, ins_atoms, &
         insert_box, ipart, istat, itype, jbox, molecule_type, nmol_types, nswapmoves, &
         print_level, rem_atoms, remove_box, source, start_atom_ins, start_atom_rem, start_mol
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, mol_type_test, nunits, &
                                                            nunits_tot
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains, nchains_test
      LOGICAL                                            :: ionode, lbias, loverlap, loverlap_ins, &
                                                            loverlap_rem
      REAL(dp), DIMENSION(:), POINTER                    :: eta_insert, eta_remove, pmswap_mol
      REAL(dp), DIMENSION(:, :), POINTER                 :: insert_coords, remove_coords
      REAL(KIND=dp) :: BETA, del_quickstep_energy, exp_max_val, exp_min_val, max_val, min_val, &
         prefactor, rand, rdum, w, weight_new, weight_old
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: cbmc_energies, r_cbmc, r_insert_mol
      REAL(KIND=dp), DIMENSION(1:2)                      :: bias_energy_new, new_energy
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc_insert, abc_remove, center_of_mass, &
                                                            displace_molecule, pos_insert
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mass
      TYPE(cell_type), POINTER                           :: cell_insert, cell_remove
      TYPE(cp_subsys_p_type), DIMENSION(:), POINTER      :: oldsys
      TYPE(cp_subsys_type), POINTER                      :: insert_sys, remove_sys
      TYPE(force_env_p_type), DIMENSION(:), POINTER      :: test_env, test_env_bias
      TYPE(mc_input_file_type), POINTER                  :: mc_bias_file, mc_input_file
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info, mc_molecule_info_test
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles_old
      TYPE(particle_list_type), POINTER                  :: particles_insert, particles_remove

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! reset the overlap flag
      loverlap = .FALSE.

! nullify some pointers
      NULLIFY (particles_old, mol_type, mol_type_test, mc_input_file, mc_bias_file)
      NULLIFY (oldsys, atom_names, pmswap_mol, insert_coords, remove_coords)
      NULLIFY (eta_insert, eta_remove)

! grab some stuff from mc_par
      CALL get_mc_par(mc_par(1)%mc_par, ionode=ionode, BETA=BETA, &
                      max_val=max_val, min_val=min_val, exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, nswapmoves=nswapmoves, group=group, source=source, &
                      lbias=lbias, dat_file=dat_file(1), fft_lib=fft_lib, &
                      mc_molecule_info=mc_molecule_info, pmswap_mol=pmswap_mol)
      CALL get_mc_molecule_info(mc_molecule_info, nchains=nchains, &
                                nunits=nunits, nunits_tot=nunits_tot, nmol_types=nmol_types, &
                                atom_names=atom_names, mass=mass, mol_type=mol_type)

      print_level = 1

      CALL get_mc_par(mc_par(2)%mc_par, dat_file=dat_file(2))

! allocate some stuff
      ALLOCATE (oldsys(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 509, &
                                       "oldsys", 2)
      ALLOCATE (particles_old(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 512, &
                                       "particles_old", 2)

! get the old coordinates
      DO ibox = 1, 2
         CALL force_env_get(force_env(ibox)%force_env, &
                            subsys=oldsys(ibox)%subsys)
         CALL cp_subsys_get(oldsys(ibox)%subsys, &
                            particles=particles_old(ibox)%list)
      ENDDO

!     choose a direction to swap
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)

      IF (rand .LE. 0.50E0_dp) THEN
         remove_box = 1
         insert_box = 2
      ELSE
         remove_box = 2
         insert_box = 1
      ENDIF

! now assign the eta values for the insert and remove boxes
      CALL get_mc_par(mc_par(remove_box)%mc_par, eta=eta_remove)
      CALL get_mc_par(mc_par(insert_box)%mc_par, eta=eta_insert)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! testing
!      remove_box=2
!      insert_box=1

! now choose a molecule type at random
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)
      DO itype = 1, nmol_types
         IF (rand .LT. pmswap_mol(itype)) THEN
            molecule_type = itype
            EXIT
         ENDIF
      ENDDO

! record the attempt for the box the particle is to be inserted into
      moves(molecule_type, insert_box)%moves%swap%attempts = &
         moves(molecule_type, insert_box)%moves%swap%attempts+1

! now choose a random molecule to remove from the removal box, checking
! to make sure the box isn't empty
      IF (nchains(molecule_type, remove_box) == 0) THEN
         loverlap = .TRUE.
         moves(molecule_type, insert_box)%moves%empty = &
            moves(molecule_type, insert_box)%moves%empty+1
      ELSE

         IF (ionode) rand = next_random_number(rng_stream)
         CALL mp_bcast(rand, source, group)
         imolecule = CEILING(rand*nchains(molecule_type, remove_box))
! figure out the atom number this molecule starts on
         start_atom_rem = 1
         DO itype = 1, nmol_types
            IF (itype == molecule_type) THEN
               start_atom_rem = start_atom_rem+(imolecule-1)*nunits(itype)
               EXIT
            ELSE
               start_atom_rem = start_atom_rem+nchains(itype, remove_box)*nunits(itype)
            ENDIF
         ENDDO

! check for overlap
         start_mol = 1
         DO jbox = 1, remove_box-1
            start_mol = start_mol+SUM(nchains(:, jbox))
         ENDDO
         end_mol = start_mol+SUM(nchains(:, remove_box))-1
         CALL check_for_overlap(force_env(remove_box)%force_env, &
                                nchains(:, remove_box), nunits, loverlap, mol_type(start_mol:end_mol))
         IF (loverlap) CALL cp_abort(cp__l("motion/mc/mc_ge_moves.F",587), &
                                     'CBMC swap move found an overlap in the old remove config')
         start_mol = 1
         DO jbox = 1, insert_box-1
            start_mol = start_mol+SUM(nchains(:, jbox))
         ENDDO
         end_mol = start_mol+SUM(nchains(:, insert_box))-1
         CALL check_for_overlap(force_env(insert_box)%force_env, &
                                nchains(:, insert_box), nunits, loverlap, mol_type(start_mol:end_mol))
         IF (loverlap) CALL cp_abort(cp__l("motion/mc/mc_ge_moves.F",596), &
                                     'CBMC swap move found an overlap in the old insert config')
      ENDIF

      IF (loverlap) THEN
         DEALLOCATE (oldsys)
         DEALLOCATE (particles_old)
         CALL timestop(handle)
         RETURN
      ENDIF

! figure out how many atoms will be in each box after the move
      ins_atoms = nunits_tot(insert_box)+nunits(molecule_type)
      rem_atoms = nunits_tot(remove_box)-nunits(molecule_type)
! now allocate the arrays that will hold the coordinates and the
! atom name, for writing to the dat file
      IF (rem_atoms == 0) THEN
         ALLOCATE (remove_coords(1:3, 1:nunits(1)), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 614, &
                                          "remove_coords", nunits(1)*3*dp_size)
         ALLOCATE (atom_names_remove(1:nunits(1)), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 617, &
                                          "atom_names_remove", nunits(1)*dp_size)
      ELSE
         ALLOCATE (remove_coords(1:3, 1:rem_atoms), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 621, &
                                          "remove_coords", rem_atoms*3*dp_size)
         ALLOCATE (atom_names_remove(1:rem_atoms), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 624, &
                                          "atom_names_remove", rem_atoms*dp_size)
      ENDIF
      ALLOCATE (insert_coords(1:3, 1:ins_atoms), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 628, &
                                       "insert_coords", ins_atoms*3*dp_size)
      ALLOCATE (atom_names_insert(1:ins_atoms), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 631, &
                                       "atom_names_insert", MAX(ins_atoms, rem_atoms)*dp_size)

! grab the cells for later...acceptance and insertion
      IF (lbias) THEN
         CALL force_env_get(bias_env(insert_box)%force_env, &
                            cell=cell_insert)
         CALL force_env_get(bias_env(remove_box)%force_env, &
                            cell=cell_remove)
      ELSE
         CALL force_env_get(force_env(insert_box)%force_env, &
                            cell=cell_insert)
         CALL force_env_get(force_env(remove_box)%force_env, &
                            cell=cell_remove)
      ENDIF
      CALL get_cell(cell_remove, abc=abc_remove)
      CALL get_cell(cell_insert, abc=abc_insert)

      IF (ionode) THEN
! choose an insertion point
         DO idim = 1, 3
            rand = next_random_number(rng_stream)
            pos_insert(idim) = rand*abc_insert(1)
         ENDDO
      ENDIF
      CALL mp_bcast(pos_insert, source, group)

! allocate some arrays we'll be using
      ALLOCATE (r_insert_mol(1:3, 1:nunits(molecule_type)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 660, &
                                       "r_insert_mol", 3*nunits(molecule_type))

      iiatom = 1
      DO iatom = start_atom_rem, start_atom_rem+nunits(molecule_type)-1
         r_insert_mol(1:3, iiatom) = &
            particles_old(remove_box)%list%els(iatom)%r(1:3)
         iiatom = iiatom+1
      ENDDO

!     find the center of mass of the molecule
      CALL get_center_of_mass(r_insert_mol(:, :), nunits(molecule_type), &
                              center_of_mass(:), mass(:, molecule_type))

!     move the center of mass to the insertion point
      displace_molecule(1:3) = pos_insert(1:3)-center_of_mass(1:3)
      DO iatom = 1, nunits(molecule_type)
         r_insert_mol(1:3, iatom) = r_insert_mol(1:3, iatom)+ &
                                    displace_molecule(1:3)
      ENDDO

! prepare the insertion coordinates to be written to the .dat file so
! we can create a new force environment...remember there is still a particle
! in the box even if nchain=0
      IF (SUM(nchains(:, insert_box)) == 0) THEN
         DO iatom = 1, nunits(molecule_type)
            insert_coords(1:3, iatom) = r_insert_mol(1:3, iatom)
            atom_names_insert(iatom) = &
               particles_old(remove_box)%list%els(start_atom_rem+iatom-1)%atomic_kind%name
         ENDDO
         start_atom_ins = 1
      ELSE
! the problem is I can't just tack the new molecule on to the end,
! because of reading in the dat_file...the topology stuff will crash
! if the molecules aren't all grouped together, so I have to insert it
! at the end of the section of molecules with the same type, then
! remember the start number for the CBMC stuff
         start_atom_ins = 1
         DO itype = 1, nmol_types
            start_atom_ins = start_atom_ins+ &
                             nchains(itype, insert_box)*nunits(itype)
            IF (itype == molecule_type) EXIT
         ENDDO

         DO iatom = 1, start_atom_ins-1
            insert_coords(1:3, iatom) = &
               particles_old(insert_box)%list%els(iatom)%r(1:3)
            atom_names_insert(iatom) = &
               particles_old(insert_box)%list%els(iatom)%atomic_kind%name
         ENDDO
         iiatom = 1
         DO iatom = start_atom_ins, start_atom_ins+nunits(molecule_type)-1
            insert_coords(1:3, iatom) = r_insert_mol(1:3, iiatom)
            atom_names_insert(iatom) = atom_names(iiatom, molecule_type)
            iiatom = iiatom+1
         ENDDO
         DO iatom = start_atom_ins+nunits(molecule_type), ins_atoms
            insert_coords(1:3, iatom) = &
               particles_old(insert_box)%list%els(iatom-nunits(molecule_type))%r(1:3)
            atom_names_insert(iatom) = &
               particles_old(insert_box)%list%els(iatom-nunits(molecule_type))%atomic_kind%name
         ENDDO
      ENDIF

! fold the coordinates into the box and check for overlaps
      start_mol = 1
      DO jbox = 1, insert_box-1
         start_mol = start_mol+SUM(nchains(:, jbox))
      ENDDO
      end_mol = start_mol+SUM(nchains(:, insert_box))-1

! make the .dat file
      IF (ionode) THEN

         nchains(molecule_type, insert_box) = nchains(molecule_type, insert_box)+1
         IF (lbias) THEN
            CALL get_mc_par(mc_par(insert_box)%mc_par, mc_bias_file=mc_bias_file)
            CALL mc_make_dat_file_new(insert_coords(:, :), atom_names_insert(:), ins_atoms, &
                                      abc_insert(:), dat_file(insert_box), nchains(:, insert_box), &
                                      mc_bias_file)
         ELSE
            CALL get_mc_par(mc_par(insert_box)%mc_par, mc_input_file=mc_input_file)
            CALL mc_make_dat_file_new(insert_coords(:, :), atom_names_insert(:), ins_atoms, &
                                      abc_insert(:), dat_file(insert_box), nchains(:, insert_box), &
                                      mc_input_file)
         ENDIF
         nchains(molecule_type, insert_box) = nchains(molecule_type, insert_box)-1

      ENDIF

! now do the same for the removal box...be careful not to make an empty box
      IF (rem_atoms == 0) THEN
         DO iatom = 1, nunits(molecule_type)
            remove_coords(1:3, iatom) = r_insert_mol(1:3, iatom)
            atom_names_remove(iatom) = atom_names(iatom, molecule_type)
         ENDDO

! need to adjust nchains, because otherwise if we are removing a molecule type
! that is not the first molecule, the dat file will have two molecules in it but
! only the coordinates for one
         nchains(molecule_type, remove_box) = nchains(molecule_type, remove_box)-1
         IF (ionode) THEN
            IF (lbias) THEN
               CALL get_mc_par(mc_par(remove_box)%mc_par, mc_bias_file=mc_bias_file)
               CALL mc_make_dat_file_new(remove_coords(:, :), atom_names_remove(:), rem_atoms, &
                                         abc_remove(:), dat_file(remove_box), nchains(:, remove_box), &
                                         mc_bias_file)
            ELSE
               CALL get_mc_par(mc_par(remove_box)%mc_par, mc_input_file=mc_input_file)
               CALL mc_make_dat_file_new(remove_coords(:, :), atom_names_remove(:), rem_atoms, &
                                         abc_remove(:), dat_file(remove_box), nchains(:, remove_box), &
                                         mc_input_file)
            ENDIF

         ENDIF
         nchains(molecule_type, remove_box) = nchains(molecule_type, remove_box)+1

      ELSE
         DO iatom = 1, start_atom_rem-1
            remove_coords(1:3, iatom) = &
               particles_old(remove_box)%list%els(iatom)%r(1:3)
            atom_names_remove(iatom) = &
               particles_old(remove_box)%list%els(iatom)%atomic_kind%name
         ENDDO
         DO iatom = start_atom_rem+nunits(molecule_type), nunits_tot(remove_box)
            remove_coords(1:3, iatom-nunits(molecule_type)) = &
               particles_old(remove_box)%list%els(iatom)%r(1:3)
            atom_names_remove(iatom-nunits(molecule_type)) = &
               particles_old(remove_box)%list%els(iatom)%atomic_kind%name
         ENDDO

! make the .dat file
         IF (ionode) THEN
            nchains(molecule_type, remove_box) = nchains(molecule_type, remove_box)-1
            IF (lbias) THEN
               CALL get_mc_par(mc_par(remove_box)%mc_par, mc_bias_file=mc_bias_file)
               CALL mc_make_dat_file_new(remove_coords(:, :), atom_names_remove(:), rem_atoms, &
                                         abc_remove(:), dat_file(remove_box), nchains(:, remove_box), &
                                         mc_bias_file)
            ELSE
               CALL get_mc_par(mc_par(remove_box)%mc_par, mc_input_file=mc_input_file)
               CALL mc_make_dat_file_new(remove_coords(:, :), atom_names_remove(:), rem_atoms, &
                                         abc_remove(:), dat_file(remove_box), nchains(:, remove_box), &
                                         mc_input_file)
            ENDIF
            nchains(molecule_type, remove_box) = nchains(molecule_type, remove_box)+1

         ENDIF
      ENDIF

! deallocate r_insert_mol
      DEALLOCATE (r_insert_mol)

! now let's create the two new environments with the different number
! of molecules
      ALLOCATE (test_env(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 816, &
                                       "test_env", 3*ins_atoms*dp_size)
      CALL mc_create_force_env(test_env(insert_box)%force_env, input_declaration, &
                               para_env, dat_file(insert_box))
      CALL mc_create_force_env(test_env(remove_box)%force_env, input_declaration, &
                               para_env, dat_file(remove_box))

! allocate an array we'll need
      ALLOCATE (r_cbmc(1:3, 1:ins_atoms), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 825, &
                                       "r_cbmc", 3*ins_atoms*dp_size)
      ALLOCATE (cbmc_energies(1:nswapmoves, 1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 828, &
                                       "cbmc_energies", nswapmoves*dp_size)

      loverlap_ins = .FALSE.
      loverlap_rem = .FALSE.

! compute the new molecule information...we need this for the CBMC part
      IF (rem_atoms == 0) THEN
         CALL mc_determine_molecule_info(test_env, mc_molecule_info_test, &
                                         box_number=remove_box)
      ELSE
         CALL mc_determine_molecule_info(test_env, mc_molecule_info_test)
      ENDIF
      CALL get_mc_molecule_info(mc_molecule_info_test, nchains=nchains_test, &
                                mol_type=mol_type_test)

! figure out the position of the molecule we're inserting, and the
! Rosenbluth weight
      start_mol = 1
      DO jbox = 1, insert_box-1
         start_mol = start_mol+SUM(nchains_test(:, jbox))
      ENDDO
      end_mol = start_mol+SUM(nchains_test(:, insert_box))-1

      IF (lbias) THEN
         CALL generate_cbmc_swap_config(test_env(insert_box)%force_env, &
                                        BETA, max_val, min_val, exp_max_val, &
                                        exp_min_val, nswapmoves, weight_new, start_atom_ins, ins_atoms, nunits(:), &
                                        nunits(molecule_type), mass(:, molecule_type), loverlap_ins, bias_energy_new(insert_box), &
                                        bias_energy_old(insert_box), ionode, .FALSE., mol_type_test(start_mol:end_mol), &
                                        nchains_test(:, insert_box), source, group, rng_stream)

! the energy that comes out of the above routine is the difference...we want
! the real energy for the acceptance rule...we don't do this for the
! lbias=.false. case because it doesn't appear in the acceptance rule, and
! we compensate in case of acceptance
         bias_energy_new(insert_box) = bias_energy_new(insert_box)+ &
                                       bias_energy_old(insert_box)
      ELSE
         CALL generate_cbmc_swap_config(test_env(insert_box)%force_env, &
                                        BETA, max_val, min_val, exp_max_val, &
                                        exp_min_val, nswapmoves, weight_new, start_atom_ins, ins_atoms, nunits(:), &
                                        nunits(molecule_type), mass(:, molecule_type), loverlap_ins, new_energy(insert_box), &
                                        old_energy(insert_box), ionode, .FALSE., mol_type_test(start_mol:end_mol), &
                                        nchains_test(:, insert_box), source, group, rng_stream)
      ENDIF

      CALL force_env_get(test_env(insert_box)%force_env, &
                         subsys=insert_sys)
      CALL cp_subsys_get(insert_sys, &
                         particles=particles_insert)

      DO iatom = 1, ins_atoms
         r_cbmc(1:3, iatom) = particles_insert%els(iatom)%r(1:3)
      ENDDO

! make sure there is no overlap

      IF (loverlap_ins .OR. loverlap_rem) THEN
! deallocate some stuff
         CALL mc_molecule_info_destroy(mc_molecule_info_test)
         CALL force_env_release(test_env(insert_box)%force_env)
         CALL force_env_release(test_env(remove_box)%force_env)
         DEALLOCATE (insert_coords)
         DEALLOCATE (remove_coords)
         DEALLOCATE (r_cbmc)
         DEALLOCATE (cbmc_energies)
         DEALLOCATE (oldsys)
         DEALLOCATE (particles_old)
         DEALLOCATE (test_env)
         CALL timestop(handle)
         RETURN
      ENDIF

! broadcast the choosen coordiantes to all processors

      CALL force_env_get(test_env(insert_box)%force_env, &
                         subsys=insert_sys)
      CALL cp_subsys_get(insert_sys, &
                         particles=particles_insert)

      DO iatom = 1, ins_atoms
         particles_insert%els(iatom)%r(1:3) = &
            r_cbmc(1:3, iatom)
      ENDDO

! if we made it this far, we have no overlaps
      moves(molecule_type, insert_box)%moves%grown = &
         moves(molecule_type, insert_box)%moves%grown+1

! if we're biasing, we need to make environments with the non-biasing
! potentials, and calculate the energies
      IF (lbias) THEN

         ALLOCATE (test_env_bias(1:2), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 923, &
                                          "test_env_bias", 2)

! first, the environment to which we added a molecule
         CALL get_mc_par(mc_par(insert_box)%mc_par, mc_input_file=mc_input_file)
         IF (ionode) CALL mc_make_dat_file_new(r_cbmc(:, :), atom_names_insert(:), ins_atoms, &
                                               abc_insert(:), dat_file(insert_box), nchains_test(:, insert_box), &
                                               mc_input_file)
         test_env_bias(insert_box)%force_env => test_env(insert_box)%force_env
         NULLIFY (test_env(insert_box)%force_env)
         CALL mc_create_force_env(test_env(insert_box)%force_env, input_declaration, &
                                  para_env, dat_file(insert_box))

         CALL force_env_calc_energy_force(test_env(insert_box)%force_env, &
                                          calc_force=.FALSE.)
         CALL force_env_get(test_env(insert_box)%force_env, &
                            potential_energy=new_energy(insert_box))

! now the environment that has one less molecule
         IF (SUM(nchains_test(:, remove_box)) == 0) THEN
            CALL get_mc_par(mc_par(remove_box)%mc_par, mc_input_file=mc_input_file)
            IF (ionode) CALL mc_make_dat_file_new(remove_coords(:, :), atom_names_remove(:), rem_atoms, &
                                                  abc_remove(:), dat_file(remove_box), nchains_test(:, remove_box), &
                                                  mc_input_file)
            test_env_bias(remove_box)%force_env => test_env(remove_box)%force_env
            NULLIFY (test_env(remove_box)%force_env)
            CALL mc_create_force_env(test_env(remove_box)%force_env, input_declaration, &
                                     para_env, dat_file(remove_box))
            new_energy(remove_box) = 0.0E0_dp
            bias_energy_new(remove_box) = 0.0E0_dp
         ELSE
            CALL get_mc_par(mc_par(remove_box)%mc_par, mc_input_file=mc_input_file)
            IF (ionode) CALL mc_make_dat_file_new(remove_coords(:, :), atom_names_remove(:), rem_atoms, &
                                                  abc_remove(:), dat_file(remove_box), nchains_test(:, remove_box), &
                                                  mc_input_file)
            test_env_bias(remove_box)%force_env => test_env(remove_box)%force_env
            NULLIFY (test_env(remove_box)%force_env)
            CALL mc_create_force_env(test_env(remove_box)%force_env, input_declaration, &
                                     para_env, dat_file(remove_box))
            CALL force_env_calc_energy_force(test_env(remove_box)%force_env, &
                                             calc_force=.FALSE.)
            CALL force_env_get(test_env(remove_box)%force_env, &
                               potential_energy=new_energy(remove_box))
            CALL force_env_calc_energy_force(test_env_bias(remove_box)%force_env, &
                                             calc_force=.FALSE.)
            CALL force_env_get(test_env_bias(remove_box)%force_env, &
                               potential_energy=bias_energy_new(remove_box))
         ENDIF
      ELSE
         IF (SUM(nchains_test(:, remove_box)) == 0) THEN
            new_energy(remove_box) = 0.0E0_dp
         ELSE
            CALL force_env_calc_energy_force(test_env(remove_box)%force_env, &
                                             calc_force=.FALSE.)
            CALL force_env_get(test_env(remove_box)%force_env, &
                               potential_energy=new_energy(remove_box))
         ENDIF
      ENDIF

! now we need to figure out the rosenbluth weight for the old configuration...
! we wait until now to do that because we need the energy of the box that
! has had a molecule removed...notice we use the environment that has not
! had a molecule removed for the CBMC configurations, and therefore nchains
! and mol_type instead of nchains_test and mol_type_test
      start_mol = 1
      DO jbox = 1, remove_box-1
         start_mol = start_mol+SUM(nchains(:, jbox))
      ENDDO
      end_mol = start_mol+SUM(nchains(:, remove_box))-1
      IF (lbias) THEN
         CALL generate_cbmc_swap_config(bias_env(remove_box)%force_env, &
                                        BETA, max_val, min_val, exp_max_val, &
                                        exp_min_val, nswapmoves, weight_old, start_atom_rem, nunits_tot(remove_box), &
                                        nunits(:), nunits(molecule_type), mass(:, molecule_type), loverlap_rem, rdum, &
                                        bias_energy_new(remove_box), ionode, .TRUE., mol_type(start_mol:end_mol), &
                                        nchains(:, remove_box), source, group, rng_stream)
      ELSE
         CALL generate_cbmc_swap_config(force_env(remove_box)%force_env, &
                                        BETA, max_val, min_val, exp_max_val, &
                                        exp_min_val, nswapmoves, weight_old, start_atom_rem, nunits_tot(remove_box), &
                                        nunits(:), nunits(molecule_type), mass(:, molecule_type), loverlap_rem, rdum, &
                                        new_energy(remove_box), ionode, .TRUE., mol_type(start_mol:end_mol), &
                                        nchains(:, remove_box), source, group, rng_stream)
      ENDIF

! figure out the prefactor to the boltzmann weight in the acceptance
! rule, based on numbers of particles and volumes

      prefactor = REAL(nchains(molecule_type, remove_box), dp)/ &
                  REAL(nchains(molecule_type, insert_box)+1, dp)* &
                  abc_insert(1)**3/abc_remove(1)**3

      IF (lbias) THEN

         del_quickstep_energy = (-BETA)*(new_energy(insert_box)- &
                                         old_energy(insert_box)+new_energy(remove_box)- &
                                         old_energy(remove_box)-(bias_energy_new(insert_box)+ &
                                                                 bias_energy_new(remove_box)-bias_energy_old(insert_box) &
                                                                 -bias_energy_old(remove_box)))

         IF (del_quickstep_energy .GT. exp_max_val) THEN
            del_quickstep_energy = max_val
         ELSEIF (del_quickstep_energy .LT. exp_min_val) THEN
            del_quickstep_energy = min_val
         ELSE
            del_quickstep_energy = EXP(del_quickstep_energy)
         ENDIF
         w = prefactor*del_quickstep_energy*weight_new/weight_old &
             *EXP(BETA*(eta_remove(molecule_type)-eta_insert(molecule_type)))

      ELSE
         w = prefactor*weight_new/weight_old &
             *EXP(BETA*(eta_remove(molecule_type)-eta_insert(molecule_type)))

      ENDIF

! check if the move is accepted
      IF (w .GE. 1.0E0_dp) THEN
         rand = 0.0E0_dp
      ELSE
         IF (ionode) rand = next_random_number(rng_stream)
         CALL mp_bcast(rand, source, group)
      ENDIF

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      IF (rand .LT. w) THEN

! accept the move

!     accept the move
         moves(molecule_type, insert_box)%moves%swap%successes = &
            moves(molecule_type, insert_box)%moves%swap%successes+1

! we need to compensate for the fact that we take the difference in
! generate_cbmc_config to keep the exponetials small
         IF (.NOT. lbias) THEN
            new_energy(insert_box) = new_energy(insert_box)+ &
                                     old_energy(insert_box)
         ENDIF

         DO ibox = 1, 2
! update energies
            energy_check(ibox) = energy_check(ibox)+(new_energy(ibox)- &
                                                     old_energy(ibox))
            old_energy(ibox) = new_energy(ibox)
! if we're biasing the update the biasing energy
            IF (lbias) THEN
               last_bias_energy(ibox) = bias_energy_new(ibox)
               bias_energy_old(ibox) = bias_energy_new(ibox)
            ENDIF

         ENDDO

! change particle numbers...basically destory the old mc_molecule_info and attach
! the new stuff to the mc_pars
! figure out the molecule information for the new environments
         CALL mc_molecule_info_destroy(mc_molecule_info)
         CALL set_mc_par(mc_par(insert_box)%mc_par, mc_molecule_info=mc_molecule_info_test)
         CALL set_mc_par(mc_par(remove_box)%mc_par, mc_molecule_info=mc_molecule_info_test)

! update coordinates
         CALL force_env_get(test_env(insert_box)%force_env, &
                            subsys=insert_sys)
         CALL cp_subsys_get(insert_sys, &
                            particles=particles_insert)
         DO ipart = 1, ins_atoms
            r_old(1:3, ipart, insert_box) = particles_insert%els(ipart)%r(1:3)
         ENDDO
         CALL force_env_get(test_env(remove_box)%force_env, &
                            subsys=remove_sys)
         CALL cp_subsys_get(remove_sys, &
                            particles=particles_remove)
         DO ipart = 1, rem_atoms
            r_old(1:3, ipart, remove_box) = particles_remove%els(ipart)%r(1:3)
         ENDDO

         ! insertion box
         CALL force_env_release(force_env(insert_box)%force_env)
         force_env(insert_box)%force_env => test_env(insert_box)%force_env

         ! removal box
         CALL force_env_release(force_env(remove_box)%force_env)
         force_env(remove_box)%force_env => test_env(remove_box)%force_env

! if we're biasing, update the bias_env
         IF (lbias) THEN
            CALL force_env_release(bias_env(insert_box)%force_env)
            bias_env(insert_box)%force_env => test_env_bias(insert_box)%force_env
            CALL force_env_release(bias_env(remove_box)%force_env)
            bias_env(remove_box)%force_env => test_env_bias(remove_box)%force_env
            DEALLOCATE (test_env_bias)
         ENDIF

      ELSE

! reject the move
         CALL mc_molecule_info_destroy(mc_molecule_info_test)
         CALL force_env_release(test_env(insert_box)%force_env)
         CALL force_env_release(test_env(remove_box)%force_env)
         IF (lbias) THEN
            CALL force_env_release(test_env_bias(insert_box)%force_env)
            CALL force_env_release(test_env_bias(remove_box)%force_env)
            DEALLOCATE (test_env_bias)
         ENDIF
      ENDIF

! deallocate some stuff
      DEALLOCATE (insert_coords)
      DEALLOCATE (remove_coords)
      DEALLOCATE (test_env)
      DEALLOCATE (cbmc_energies)
      DEALLOCATE (r_cbmc)
      DEALLOCATE (oldsys)
      DEALLOCATE (particles_old)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_ge_swap_move

! **************************************************************************************************
!> \brief performs a Monte Carlo move that alters the volume of the simulation boxes,
!>      keeping the total volume of the two boxes the same
!> \param mc_par the mc parameters for the force env
!> \param force_env the force environments used in the move
!> \param moves the structure that keeps track of how many moves have been
!>               accepted/rejected
!> \param move_updates the structure that keeps track of how many moves have
!>               been accepted/rejected since the last time the displacements
!>               were updated
!> \param nnstep the total number of Monte Carlo moves already performed
!> \param old_energy the energy of the last accepted move involving an
!>                    unbiased potential calculation
!> \param energy_check the running total of how much the energy has changed
!>                      since the initial configuration
!> \param r_old the coordinates of the last accepted move involving a
!>               Quickstep calculation
!> \param rng_stream the stream we pull random numbers from
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_ge_volume_move(mc_par, force_env, moves, move_updates, &
                                nnstep, old_energy, energy_check, r_old, rng_stream)

      TYPE(mc_simulation_parameters_p_type), &
         DIMENSION(:), POINTER                           :: mc_par
      TYPE(force_env_p_type), DIMENSION(:), POINTER      :: force_env
      TYPE(mc_moves_p_type), DIMENSION(:, :), POINTER    :: moves, move_updates
      INTEGER, INTENT(IN)                                :: nnstep
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: old_energy, energy_check
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: r_old
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_ge_volume_move', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=200)                                 :: fft_lib
      CHARACTER(LEN=40), DIMENSION(1:2)                  :: dat_file
      INTEGER :: cl, end_atom, end_mol, group, handle, iatom, ibox, imolecule, iside, istat, j, &
         jatom, jbox, max_atoms, molecule_index, molecule_type, print_level, source, start_atom, &
         start_mol
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, nunits, nunits_tot
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      LOGICAL                                            :: ionode
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: loverlap
      LOGICAL, DIMENSION(1:2)                            :: lempty
      REAL(dp), DIMENSION(:, :), POINTER                 :: mass
      REAL(KIND=dp)                                      :: BETA, prefactor, rand, rmvolume, &
                                                            vol_dis, w
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: r
      REAL(KIND=dp), DIMENSION(1:2)                      :: new_energy, volume_new, volume_old
      REAL(KIND=dp), DIMENSION(1:3)                      :: center_of_mass, center_of_mass_new, diff
      REAL(KIND=dp), DIMENSION(1:3, 1:2)                 :: abc, new_cell_length, old_cell_length
      REAL(KIND=dp), DIMENSION(1:3, 1:3, 1:2)            :: hmat_test
      TYPE(cell_p_type), DIMENSION(:), POINTER           :: cell, cell_old, cell_test
      TYPE(cp_subsys_p_type), DIMENSION(:), POINTER      :: oldsys
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles_old

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

! nullify some pointers
      NULLIFY (particles_old, cell, oldsys, cell_old, cell_test, subsys)

! get some data from mc_par
      CALL get_mc_par(mc_par(1)%mc_par, ionode=ionode, source=source, &
                      group=group, dat_file=dat_file(1), rmvolume=rmvolume, &
                      BETA=BETA, cl=cl, fft_lib=fft_lib, &
                      mc_molecule_info=mc_molecule_info)
      CALL get_mc_molecule_info(mc_molecule_info, nunits_tot=nunits_tot, &
                                mass=mass, nchains=nchains, nunits=nunits, mol_type=mol_type)

      print_level = 1
      CALL get_mc_par(mc_par(2)%mc_par, dat_file=dat_file(2))

! allocate some stuff
      max_atoms = MAX(nunits_tot(1), nunits_tot(2))
      ALLOCATE (r(1:3, max_atoms, 1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1223, &
                                       "r", 2*3*max_atoms*dp_size)
      ALLOCATE (oldsys(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1226, &
                                       "oldsys", 2)
      ALLOCATE (particles_old(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1229, &
                                       "particles_old", 2)
      ALLOCATE (cell(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1232, &
                                       "cell", 2)
      ALLOCATE (cell_test(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1235, &
                                       "cell_test", 2)
      ALLOCATE (cell_old(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1238, &
                                       "cell_old", 2)
      ALLOCATE (loverlap(1:2), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1241, &
                                       "loverlap", 2)

! check for empty boxes...need to be careful because we can't build
! a force_env with no particles
      DO ibox = 1, 2
         lempty(ibox) = .FALSE.
         IF (SUM(nchains(:, ibox)) == 0) THEN
            lempty(ibox) = .TRUE.
         ENDIF
      ENDDO

! record the attempt
      DO ibox = 1, 2
         moves(1, ibox)%moves%volume%attempts = &
            moves(1, ibox)%moves%volume%attempts+1
         move_updates(1, ibox)%moves%volume%attempts = &
            move_updates(1, ibox)%moves%volume%attempts+1
      ENDDO

! now let's grab the cell length and particle positions
      DO ibox = 1, 2
         CALL force_env_get(force_env(ibox)%force_env, &
                            subsys=oldsys(ibox)%subsys, cell=cell(ibox)%cell)
         CALL get_cell(cell(ibox)%cell, abc=abc(:, ibox))
         NULLIFY (cell_old(ibox)%cell)
         CALL cell_create(cell_old(ibox)%cell)
         CALL cell_clone(cell(ibox)%cell, cell_old(ibox)%cell)
         CALL cp_subsys_get(oldsys(ibox)%subsys, &
                            particles=particles_old(ibox)%list)

! find the old cell length
         old_cell_length(1:3, ibox) = abc(1:3, ibox)

      ENDDO

      DO ibox = 1, 2

! save the old coordiantes
         DO iatom = 1, nunits_tot(ibox)
            r(1:3, iatom, ibox) = particles_old(ibox)%list%els(iatom)%r(1:3)
         ENDDO

      ENDDO

! call a random number to figure out how far we're moving
      IF (ionode) rand = next_random_number(rng_stream)
      CALL mp_bcast(rand, source, group)

      vol_dis = rmvolume*(rand-0.5E0_dp)*2.0E0_dp

! add to one box, subtract from the other
      IF (old_cell_length(1, 1)*old_cell_length(2, 1)* &
          old_cell_length(3, 1)+vol_dis .LE. (3.0E0_dp/angstrom)**3) &
         CALL cp__b("motion/mc/mc_ge_moves.F",1295,'GE_volume moves are trying to make box 1 smaller than 3')
      IF (old_cell_length(1, 2)*old_cell_length(2, 2)* &
          old_cell_length(3, 2)+vol_dis .LE. (3.0E0_dp/angstrom)**3) &
         CALL cp__b("motion/mc/mc_ge_moves.F",1298,'GE_volume moves are trying to make box 2 smaller than 3')

      DO iside = 1, 3
         new_cell_length(iside, 1) = (old_cell_length(1, 1)**3+ &
                                      vol_dis)**(1.0E0_dp/3.0E0_dp)
         new_cell_length(iside, 2) = (old_cell_length(1, 2)**3- &
                                      vol_dis)**(1.0E0_dp/3.0E0_dp)
      ENDDO

! now we need to make the new cells
      DO ibox = 1, 2
         hmat_test(:, :, ibox) = 0.0e0_dp
         hmat_test(1, 1, ibox) = new_cell_length(1, ibox)
         hmat_test(2, 2, ibox) = new_cell_length(2, ibox)
         hmat_test(3, 3, ibox) = new_cell_length(3, ibox)
         NULLIFY (cell_test(ibox)%cell)
         CALL cell_create(cell_test(ibox)%cell, hmat=hmat_test(:, :, ibox), &
                          periodic=cell(ibox)%cell%perd)
         CALL force_env_get(force_env(ibox)%force_env, subsys=subsys)
         CALL cp_subsys_set(subsys, cell=cell_test(ibox)%cell)
      ENDDO

      DO ibox = 1, 2

! save the coords
         DO iatom = 1, nunits_tot(ibox)
            r(1:3, iatom, ibox) = particles_old(ibox)%list%els(iatom)%r(1:3)
         ENDDO

! now we need to scale the coordinates of all the molecules by the
! center of mass
         start_atom = 1
         molecule_index = 1
         DO jbox = 1, ibox-1
            IF (jbox == ibox) EXIT
            molecule_index = molecule_index+SUM(nchains(:, jbox))
         ENDDO
         DO imolecule = 1, SUM(nchains(:, ibox))
            molecule_type = mol_type(imolecule+molecule_index-1)
            IF (imolecule .NE. 1) THEN
               start_atom = start_atom+nunits(mol_type(imolecule+molecule_index-2))
            ENDIF
            end_atom = start_atom+nunits(molecule_type)-1

! now find the center of mass
            CALL get_center_of_mass(r(:, start_atom:end_atom, ibox), &
                                    nunits(molecule_type), center_of_mass(:), mass(:, molecule_type))

! scale the center of mass and determine the vector that points from the
!    old COM to the new one
            center_of_mass_new(1:3) = center_of_mass(1:3)* &
                                      new_cell_length(1:3, ibox)/old_cell_length(1:3, ibox)
            DO j = 1, 3
               diff(j) = center_of_mass_new(j)-center_of_mass(j)
! now change the particle positions
               DO jatom = start_atom, end_atom
                  particles_old(ibox)%list%els(jatom)%r(j) = &
                     particles_old(ibox)%list%els(jatom)%r(j)+diff(j)
               ENDDO

            ENDDO
         ENDDO

! check for any overlaps we might have
         start_mol = 1
         DO jbox = 1, ibox-1
            start_mol = start_mol+SUM(nchains(:, jbox))
         ENDDO
         end_mol = start_mol+SUM(nchains(:, ibox))-1
         CALL check_for_overlap(force_env(ibox)%force_env, &
                                nchains(:, ibox), nunits, loverlap(ibox), mol_type(start_mol:end_mol), &
                                cell_length=new_cell_length(:, ibox))

      ENDDO

! determine the overall energy difference

      DO ibox = 1, 2
         IF (loverlap(ibox)) CYCLE
! remake the force environment and calculate the energy
         IF (lempty(ibox)) THEN
            new_energy(ibox) = 0.0E0_dp
         ELSE

            CALL force_env_calc_energy_force(force_env(ibox)%force_env, &
                                             calc_force=.FALSE.)
            CALL force_env_get(force_env(ibox)%force_env, &
                               potential_energy=new_energy(ibox))

         ENDIF
      ENDDO

! accept or reject the move
      DO ibox = 1, 2
         volume_new(ibox) = new_cell_length(1, ibox)* &
                            new_cell_length(2, ibox)*new_cell_length(3, ibox)
         volume_old(ibox) = old_cell_length(1, ibox)* &
                            old_cell_length(2, ibox)*old_cell_length(3, ibox)
      ENDDO
      prefactor = (volume_new(1)/volume_old(1))**(SUM(nchains(:, 1)))* &
                  (volume_new(2)/volume_old(2))**(SUM(nchains(:, 2)))

      IF (loverlap(1) .OR. loverlap(2)) THEN
         w = 0.0E0_dp
      ELSE
         w = prefactor*EXP(-BETA* &
                           (new_energy(1)+new_energy(2)- &
                            old_energy(1)-old_energy(2)))

      ENDIF

      IF (w .GE. 1.0E0_dp) THEN
         w = 1.0E0_dp
         rand = 0.0E0_dp
      ELSE
         IF (ionode) rand = next_random_number(rng_stream)
         CALL mp_bcast(rand, source, group)
      ENDIF

      IF (rand .LT. w) THEN

! write cell length, volume, density, and trial displacement to a file
         IF (ionode) THEN

            WRITE (cl, *) nnstep, new_cell_length(1, 1)* &
               angstrom, vol_dis*(angstrom)**3, new_cell_length(1, 2)* &
               angstrom
            WRITE (cl, *) nnstep, new_energy(1), &
               old_energy(1), new_energy(2), old_energy(2)
            WRITE (cl, *) prefactor, w
         ENDIF

         DO ibox = 1, 2
! accept the move
            moves(1, ibox)%moves%volume%successes = &
               moves(1, ibox)%moves%volume%successes+1
            move_updates(1, ibox)%moves%volume%successes = &
               move_updates(1, ibox)%moves%volume%successes+1

! update energies
            energy_check(ibox) = energy_check(ibox)+(new_energy(ibox)- &
                                                     old_energy(ibox))
            old_energy(ibox) = new_energy(ibox)

! and the new "old" coordiantes
            DO iatom = 1, nunits_tot(ibox)
               r_old(1:3, iatom, ibox) = &
                  particles_old(ibox)%list%els(iatom)%r(1:3)
            ENDDO

         ENDDO
      ELSE

! reject the move
! write cell length, volume, density, and trial displacement to a file
         IF (ionode) THEN

            WRITE (cl, *) nnstep, new_cell_length(1, 1)* &
               angstrom, vol_dis*(angstrom)**3, new_cell_length(1, 2)* &
               angstrom
            WRITE (cl, *) nnstep, new_energy(1), &
               old_energy(1), new_energy(2), old_energy(2)
            WRITE (cl, *) prefactor, w

         ENDIF

! reset the cell and particle positions
         DO ibox = 1, 2
            CALL force_env_get(force_env(ibox)%force_env, subsys=subsys)
            CALL cp_subsys_set(subsys, cell=cell_old(ibox)%cell)
            DO iatom = 1, nunits_tot(ibox)
               particles_old(ibox)%list%els(iatom)%r(1:3) = r_old(1:3, iatom, ibox)
            ENDDO
         ENDDO

      ENDIF

! free up some memory
      DO ibox = 1, 2
         CALL cell_release(cell_test(ibox)%cell)
         CALL cell_release(cell_old(ibox)%cell)
      ENDDO
      DEALLOCATE (r)
      DEALLOCATE (oldsys)
      DEALLOCATE (particles_old)
      DEALLOCATE (cell)
      DEALLOCATE (cell_old)
      DEALLOCATE (cell_test)
      DEALLOCATE (loverlap)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_ge_volume_move

END MODULE mc_ge_moves

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Used to run the bulk of the MC simulation, doing things like
!>      choosing move types and writing data to files
!> \author Matthew J. McGrath  (09.26.2003)
!>
!>    REVISIONS
!>      09.10.05  MJM combined the two subroutines in this module into one
! **************************************************************************************************
MODULE mc_ensembles
   USE cell_types,                      ONLY: cell_p_type,&
                                              get_cell
   USE cp_external_control,             ONLY: external_control
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_p_type,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_p_type,&
                                              force_env_release
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: dump_xmol
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE mathconstants,                   ONLY: pi
   USE mc_control,                      ONLY: mc_create_bias_force_env,&
                                              write_mc_restart
   USE mc_coordinates,                  ONLY: check_for_overlap,&
                                              create_discrete_array,&
                                              find_mc_test_molecule,&
                                              get_center_of_mass,&
                                              mc_coordinate_fold,&
                                              rotate_molecule
   USE mc_environment_types,            ONLY: get_mc_env,&
                                              mc_environment_p_type,&
                                              set_mc_env
   USE mc_ge_moves,                     ONLY: mc_ge_swap_move,&
                                              mc_ge_volume_move,&
                                              mc_quickstep_move
   USE mc_misc,                         ONLY: final_mc_write,&
                                              mc_averages_create,&
                                              mc_averages_release
   USE mc_move_control,                 ONLY: init_mc_moves,&
                                              mc_move_update,&
                                              mc_moves_release,&
                                              write_move_stats
   USE mc_moves,                        ONLY: mc_avbmc_move,&
                                              mc_conformation_change,&
                                              mc_hmc_move,&
                                              mc_molecule_rotation,&
                                              mc_molecule_translation,&
                                              mc_volume_move
   USE mc_types,                        ONLY: get_mc_molecule_info,&
                                              get_mc_par,&
                                              mc_averages_p_type,&
                                              mc_input_file_type,&
                                              mc_molecule_info_type,&
                                              mc_moves_p_type,&
                                              mc_simulation_parameters_p_type,&
                                              set_mc_par
   USE message_passing,                 ONLY: mp_bcast
   USE parallel_rng_types,              ONLY: next_random_number,&
                                              rng_stream_type
   USE particle_list_types,             ONLY: particle_list_p_type,&
                                              particle_list_type
   USE particle_methods,                ONLY: write_particle_coordinates
   USE physcon,                         ONLY: angstrom,&
                                              boltzmann,&
                                              joule,&
                                              n_avogadro
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_ensembles'
   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .FALSE.

   PUBLIC :: mc_run_ensemble, mc_compute_virial

CONTAINS

! **************************************************************************************************
!> \brief directs the program in running one or two box MC simulations
!> \param mc_env a pointer that contains all mc_env for all the simulation
!>          boxes
!> \param para_env ...
!> \param globenv the global environment for the simulation
!> \param input_declaration ...
!> \param nboxes the number of simulation boxes
!> \param rng_stream the stream we pull random numbers from
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_run_ensemble(mc_env, para_env, globenv, input_declaration, nboxes, rng_stream)

      TYPE(mc_environment_p_type), DIMENSION(:), POINTER :: mc_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_type), POINTER                        :: input_declaration
      INTEGER, INTENT(IN)                                :: nboxes
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_run_ensemble', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_string_length), ALLOCATABLE, &
         DIMENSION(:)                                    :: atom_names_box
      CHARACTER(default_string_length), &
         DIMENSION(:, :), POINTER                        :: atom_names
      CHARACTER(LEN=20)                                  :: ensemble
      CHARACTER(LEN=40)                                  :: cbox, cstep, fft_lib, move_type, &
                                                            move_type_avbmc
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      INTEGER, DIMENSION(:), POINTER                     :: avbmc_atom, mol_type, nchains_box, &
                                                            nunits, nunits_tot
      INTEGER, DIMENSION(1:nboxes)                       :: box_flag, cl, data_unit, diff, istep, &
                                                            move_unit, rm
      INTEGER, DIMENSION(1:3, 1:2)                       :: discrete_array
      INTEGER :: atom_number, box_number, cell_unit, com_crd, com_ene, com_mol, end_mol, group, &
         handle, ibox, idum, imol_type, imolecule, imove, iparticle, iprint, itype, iunit, &
         iuptrans, iupvolume, iw, jbox, jdum, molecule_type, molecule_type_swap, &
         molecule_type_target, nchain_total, nmol_types, nmoves, nnstep, nstart, nstep, source, &
         start_atom, start_atom_swap, start_atom_target, start_mol
      CHARACTER(LEN=default_string_length)               :: unit_str
      CHARACTER(LEN=40), DIMENSION(1:nboxes)             :: cell_file, coords_file, data_file, &
                                                            displacement_file, energy_file, &
                                                            molecules_file, moves_file
      LOGICAL                                            :: ionode, lbias, ldiscrete, lhmc, &
                                                            lnew_bias_env, loverlap, lreject, &
                                                            lstop, should_stop
      REAL(dp), DIMENSION(:), POINTER                    :: pbias, pmavbmc_mol, pmhmc_box, &
                                                            pmrot_mol, pmtraion_mol, pmtrans_mol, &
                                                            pmvol_box
      REAL(dp), DIMENSION(:, :), POINTER                 :: conf_prob, mass
      REAL(KIND=dp)                                      :: discrete_step, pmavbmc, pmhmc, pmswap, &
                                                            pmtraion, pmtrans, pmvolume, rand, &
                                                            test_energy, unit_conv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r_temp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: r_old
      REAL(KIND=dp), DIMENSION(1:3, 1:nboxes)            :: abc
      REAL(KIND=dp), DIMENSION(1:nboxes)                 :: bias_energy, energy_check, final_energy, &
                                                            initial_energy, last_bias_energy, &
                                                            old_energy
      TYPE(cell_p_type), DIMENSION(:), POINTER           :: cell
      TYPE(cp_subsys_p_type), DIMENSION(:), POINTER      :: oldsys
      TYPE(cp_subsys_type), POINTER                      :: biassys
      TYPE(force_env_p_type), DIMENSION(:), POINTER      :: bias_env, force_env
      TYPE(mc_averages_p_type), DIMENSION(:), POINTER    :: averages
      TYPE(mc_input_file_type), POINTER                  :: mc_bias_file
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(mc_moves_p_type), DIMENSION(:), POINTER       :: test_moves
      TYPE(mc_moves_p_type), DIMENSION(:, :), POINTER    :: move_updates, moves
      TYPE(mc_simulation_parameters_p_type), &
         DIMENSION(:), POINTER                           :: mc_par
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles_old
      TYPE(particle_list_type), POINTER                  :: particles_bias
      TYPE(section_vals_type), POINTER                   :: root_section

      CALL timeset(routineN, handle)

      ! nullify some pointers
      NULLIFY (moves, move_updates, test_moves, root_section)

      ! allocate a whole bunch of stuff based on how many boxes we have
      ALLOCATE (force_env(1:nboxes))
      ALLOCATE (bias_env(1:nboxes))
      ALLOCATE (cell(1:nboxes))
      ALLOCATE (particles_old(1:nboxes))
      ALLOCATE (oldsys(1:nboxes))
      ALLOCATE (averages(1:nboxes))
      ALLOCATE (mc_par(1:nboxes))
      ALLOCATE (pmvol_box(1:nboxes))
      ALLOCATE (pmhmc_box(1:nboxes))

      DO ibox = 1, nboxes
         CALL get_mc_env(mc_env(ibox)%mc_env, &
                         mc_par=mc_par(ibox)%mc_par, &
                         force_env=force_env(ibox)%force_env)
      ENDDO

      ! Gather units of measure for output (if available)
      root_section => force_env(1)%force_env%root_section
      CALL section_vals_val_get(root_section, "MOTION%PRINT%TRAJECTORY%UNIT", &
                                c_val=unit_str)
      unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

      ! get some data out of mc_par
      CALL get_mc_par(mc_par(1)%mc_par, &
                      ionode=ionode, source=source, group=group, &
                      data_file=data_file(1), moves_file=moves_file(1), &
                      cell_file=cell_file(1), coords_file=coords_file(1), &
                      energy_file=energy_file(1), displacement_file=displacement_file(1), &
                      lstop=lstop, nstep=nstep, nstart=nstart, pmvolume=pmvolume, pmhmc=pmhmc, &
                      molecules_file=molecules_file(1), pmswap=pmswap, nmoves=nmoves, &
                      pmtraion=pmtraion, pmtrans=pmtrans, iuptrans=iuptrans, &
                      iupvolume=iupvolume, ldiscrete=ldiscrete, pmtraion_mol=pmtraion_mol, &
                      lbias=lbias, iprint=iprint, pmavbmc_mol=pmavbmc_mol, &
                      discrete_step=discrete_step, fft_lib=fft_lib, avbmc_atom=avbmc_atom, &
                      pmavbmc=pmavbmc, pbias=pbias, mc_molecule_info=mc_molecule_info, &
                      pmrot_mol=pmrot_mol, pmtrans_mol=pmtrans_mol, pmvol_box=pmvol_box(1), &
                      ensemble=ensemble, pmhmc_box=pmhmc_box(1), lhmc=lhmc)

      ! get some data from the molecule types
      CALL get_mc_molecule_info(mc_molecule_info, conf_prob=conf_prob, &
                                nchains=nchains, nmol_types=nmol_types, nunits_tot=nunits_tot, &
                                mol_type=mol_type, nchain_total=nchain_total, nunits=nunits, &
                                atom_names=atom_names, mass=mass)

      ! allocate some stuff based on the number of molecule types we have
      ALLOCATE (moves(1:nmol_types, 1:nboxes))
      ALLOCATE (move_updates(1:nmol_types, 1:nboxes))

      IF (nboxes .GT. 1) THEN
         DO ibox = 2, nboxes
            CALL get_mc_par(mc_par(ibox)%mc_par, &
                            data_file=data_file(ibox), &
                            moves_file=moves_file(ibox), &
                            cell_file=cell_file(ibox), coords_file=coords_file(ibox), &
                            energy_file=energy_file(ibox), &
                            displacement_file=displacement_file(ibox), &
                            molecules_file=molecules_file(ibox), pmvol_box=pmvol_box(ibox), &
                            pmhmc_box=pmhmc_box(ibox))
         ENDDO
      ENDIF

      ! this is a check we can't do in the input checking
      IF (pmvol_box(nboxes) .LT. 1.0E0_dp) THEN
         CALL cp__b("motion/mc/mc_ensembles.F",246,'The last value of PMVOL_BOX needs to be 1.0')
      ENDIF
      IF (pmhmc_box(nboxes) .LT. 1.0E0_dp) THEN
         CALL cp__b("motion/mc/mc_ensembles.F",249,'The last value of PMHMC_BOX needs to be 1.0')
      ENDIF

      ! allocate the particle positions array for broadcasting
      ALLOCATE (r_old(3, SUM(nunits_tot), 1:nboxes))

      ! figure out what the default write unit is
      iw = cp_logger_get_default_io_unit()

      IF (iw > 0) THEN
         WRITE (iw, *)
         WRITE (iw, *)
         WRITE (iw, *) 'Beginning the Monte Carlo calculation.'
         WRITE (iw, *)
         WRITE (iw, *)
      ENDIF

      ! initialize running average variables
      energy_check(:) = 0.0E0_dp
      box_flag(:) = 0
      istep(:) = 0

      DO ibox = 1, nboxes
         ! initialize the moves array, the arrays for updating maximum move
         ! displacements, and the averages array
         DO itype = 1, nmol_types
            CALL init_mc_moves(moves(itype, ibox)%moves)
            CALL init_mc_moves(move_updates(itype, ibox)%moves)
         ENDDO
         CALL mc_averages_create(averages(ibox)%averages)

         ! find the energy of the initial configuration
         IF (SUM(nchains(:, ibox)) .NE. 0) THEN
            CALL force_env_calc_energy_force(force_env(ibox)%force_env, &
                                             calc_force=.FALSE.)
            CALL force_env_get(force_env(ibox)%force_env, &
                               potential_energy=old_energy(ibox))
         ELSE
            old_energy(ibox) = 0.0E0_dp
         ENDIF
         initial_energy(ibox) = old_energy(ibox)

! don't care about overlaps if we're only doing HMC

         IF (.NOT. lhmc) THEN
            ! check for overlaps
            start_mol = 1
            DO jbox = 1, ibox-1
               start_mol = start_mol+SUM(nchains(:, jbox))
            ENDDO
            end_mol = start_mol+SUM(nchains(:, ibox))-1
            CALL check_for_overlap(force_env(ibox)%force_env, nchains(:, ibox), &
                                   nunits, loverlap, mol_type(start_mol:end_mol))
            IF (loverlap) CALL cp__b("motion/mc/mc_ensembles.F",302,"overlap in an initial configuration")
         ENDIF

         ! get the subsystems and the cell information
         CALL force_env_get(force_env(ibox)%force_env, &
                            subsys=oldsys(ibox)%subsys, cell=cell(ibox)%cell)
         CALL get_cell(cell(ibox)%cell, abc=abc(:, ibox))
         CALL cp_subsys_get(oldsys(ibox)%subsys, &
                            particles=particles_old(ibox)%list)
         ! record the old coordinates, in case a move is rejected
         DO iparticle = 1, nunits_tot(ibox)
            r_old(1:3, iparticle, ibox) = &
               particles_old(ibox)%list%els(iparticle)%r(1:3)
         ENDDO

         ! find the bias energy of the initial run
         IF (lbias) THEN
            ! determine the atom names of every particle
            ALLOCATE (atom_names_box(1:nunits_tot(ibox)))

            atom_number = 1
            DO imolecule = 1, SUM(nchains(:, ibox))
               DO iunit = 1, nunits(mol_type(imolecule+start_mol-1))
                  atom_names_box(atom_number) = &
                     atom_names(iunit, mol_type(imolecule+start_mol-1))
                  atom_number = atom_number+1
               ENDDO
            ENDDO

            CALL get_mc_par(mc_par(ibox)%mc_par, mc_bias_file=mc_bias_file)
            nchains_box => nchains(:, ibox)
            CALL mc_create_bias_force_env(bias_env(ibox)%force_env, &
                                          r_old(:, :, ibox), atom_names_box(:), nunits_tot(ibox), &
                                          para_env, abc(:, ibox), nchains_box, input_declaration, mc_bias_file, &
                                          ionode)
            IF (SUM(nchains(:, ibox)) .NE. 0) THEN
               CALL force_env_calc_energy_force(bias_env(ibox)%force_env, &
                                                calc_force=.FALSE.)
               CALL force_env_get(bias_env(ibox)%force_env, &
                                  potential_energy=last_bias_energy(ibox))

            ELSE
               last_bias_energy(ibox) = 0.0E0_dp
            ENDIF
            bias_energy(ibox) = last_bias_energy(ibox)
            DEALLOCATE (atom_names_box)
         ENDIF
         lnew_bias_env = .FALSE.

      ENDDO

      ! back to seriel for a bunch of I/O stuff
      IF (ionode) THEN

         ! record the combined energies,coordinates, and cell lengths
         CALL open_file(file_name='mc_cell_length', &
                        unit_number=cell_unit, file_position='APPEND', &
                        file_action='WRITE', file_status='UNKNOWN')
         CALL open_file(file_name='mc_energies', &
                        unit_number=com_ene, file_position='APPEND', &
                        file_action='WRITE', file_status='UNKNOWN')
         CALL open_file(file_name='mc_coordinates', &
                        unit_number=com_crd, file_position='APPEND', &
                        file_action='WRITE', file_status='UNKNOWN')
         CALL open_file(file_name='mc_molecules', &
                        unit_number=com_mol, file_position='APPEND', &
                        file_action='WRITE', file_status='UNKNOWN')
         WRITE (com_ene, *) 'Initial Energies:       ', &
            old_energy(1:nboxes)
         DO ibox = 1, nboxes
            WRITE (com_mol, *) 'Initial Molecules:       ', &
               nchains(:, ibox)
         ENDDO
         DO ibox = 1, nboxes
            WRITE (cell_unit, *) 'Initial: ', &
               abc(1:3, ibox)*angstrom
            WRITE (cbox, '(I4)') ibox
            CALL open_file(file_name='energy_differences_box'// &
                           TRIM(ADJUSTL(cbox)), &
                           unit_number=diff(ibox), file_position='APPEND', &
                           file_action='WRITE', file_status='UNKNOWN')
            IF (SUM(nchains(:, ibox)) == 0) THEN
               WRITE (com_crd, *) ' 0'
               WRITE (com_crd, *) 'INITIAL BOX '//TRIM(ADJUSTL(cbox))
            ELSE
               CALL write_particle_coordinates(particles_old(ibox)%list%els, &
                                               com_crd, dump_xmol, 'POS', 'INITIAL BOX '//TRIM(ADJUSTL(cbox)), &
                                               unit_conv=unit_conv)
            ENDIF
            CALL open_file(file_name=data_file(ibox), &
                           unit_number=data_unit(ibox), file_position='APPEND', &
                           file_action='WRITE', file_status='UNKNOWN')
            CALL open_file(file_name=moves_file(ibox), &
                           unit_number=move_unit(ibox), file_position='APPEND', &
                           file_action='WRITE', file_status='UNKNOWN')
            CALL open_file(file_name=displacement_file(ibox), &
                           unit_number=rm(ibox), file_position='APPEND', &
                           file_action='WRITE', file_status='UNKNOWN')
            CALL open_file(file_name=cell_file(ibox), &
                           unit_number=cl(ibox), file_position='APPEND', &
                           file_action='WRITE', file_status='UNKNOWN')

         ENDDO

         ! back to parallel mode
      ENDIF

      DO ibox = 1, nboxes
         CALL mp_bcast(cl(ibox), source, group)
         CALL mp_bcast(rm(ibox), source, group)
         CALL mp_bcast(diff(ibox), source, group)
         ! set all the units numbers that we just opened in the respective mc_par
         CALL set_mc_par(mc_par(ibox)%mc_par, cl=cl(ibox), rm=rm(ibox), &
                         diff=diff(ibox))
      ENDDO

      ! if we're doing a discrete volume move, we need to set up the array
      ! that keeps track of which direction we can move in
      IF (ldiscrete) THEN
         IF (nboxes .NE. 1) &
            CALL cp__b("motion/mc/mc_ensembles.F",422,'ldiscrete=.true. ONLY for systems with 1 box')
         CALL create_discrete_array(abc(:, 1), discrete_array(:, :), &
                                    discrete_step)
      ENDIF

      ! find out how many steps we're doing...change the updates to be in cycles
      ! if the total number of steps is measured in cycles
      IF (.NOT. lstop) THEN
         nstep = nstep*nchain_total
         iuptrans = iuptrans*nchain_total
         iupvolume = iupvolume*nchain_total
      ENDIF

      DO nnstep = nstart+1, nstart+nstep

         IF (MOD(nnstep, iprint) == 0 .AND. (iw > 0)) THEN
            WRITE (iw, *)
            WRITE (iw, *) "------- On Monte Carlo Step ", nnstep
         ENDIF

         IF (ionode) rand = next_random_number(rng_stream)
         ! broadcast the random number, to make sure we're on the same move
         CALL mp_bcast(rand, source, group)

         IF (rand .LT. pmvolume) THEN

            IF (MOD(nnstep, iprint) == 0 .AND. (iw > 0)) THEN
               WRITE (iw, *) "Attempting a volume move"
               WRITE (iw, *)
            ENDIF

            SELECT CASE (ensemble)
            CASE ("TRADITIONAL")
               CALL mc_volume_move(mc_par(1)%mc_par, &
                                   force_env(1)%force_env, &
                                   moves(1, 1)%moves, move_updates(1, 1)%moves, &
                                   old_energy(1), 1, &
                                   energy_check(1), r_old(:, :, 1), iw, discrete_array(:, :), &
                                   rng_stream)
            CASE ("GEMC_NVT")
               CALL mc_ge_volume_move(mc_par, force_env, moves, &
                                      move_updates, nnstep, old_energy, energy_check, &
                                      r_old, rng_stream)
            CASE ("GEMC_NPT")
               ! we need to select a box based on the probability given in the input file
               IF (ionode) rand = next_random_number(rng_stream)
               CALL mp_bcast(rand, source, group)

               DO ibox = 1, nboxes
                  IF (rand .LE. pmvol_box(ibox)) THEN
                     box_number = ibox
                     EXIT
                  ENDIF
               ENDDO

               CALL mc_volume_move(mc_par(box_number)%mc_par, &
                                   force_env(box_number)%force_env, &
                                   moves(1, box_number)%moves, &
                                   move_updates(1, box_number)%moves, &
                                   old_energy(box_number), box_number, &
                                   energy_check(box_number), r_old(:, :, box_number), iw, &
                                   discrete_array(:, :), &
                                   rng_stream)
            END SELECT

! update all the pointers here, because otherwise we may pass wrong information when we're making a bias environment
            DO ibox = 1, nboxes
               CALL force_env_get(force_env(ibox)%force_env, &
                                  subsys=oldsys(ibox)%subsys, cell=cell(ibox)%cell)
               CALL get_cell(cell(ibox)%cell, abc=abc(:, ibox))
               CALL cp_subsys_get(oldsys(ibox)%subsys, &
                                  particles=particles_old(ibox)%list)
            ENDDO

            ! we need a new biasing environment now, if we're into that sort of thing
            IF (lbias) THEN
               DO ibox = 1, nboxes
                  CALL force_env_release(bias_env(ibox)%force_env)
                  ! determine the atom names of every particle
                  ALLOCATE (atom_names_box(1:nunits_tot(ibox)))
                  start_mol = 1
                  DO jbox = 1, ibox-1
                     start_mol = start_mol+SUM(nchains(:, jbox))
                  ENDDO
                  end_mol = start_mol+SUM(nchains(:, ibox))-1
                  atom_number = 1
                  DO imolecule = 1, SUM(nchains(:, ibox))
                     DO iunit = 1, nunits(mol_type(imolecule+start_mol-1))
                        atom_names_box(atom_number) = &
                           atom_names(iunit, mol_type(imolecule+start_mol-1))
                        atom_number = atom_number+1
                     ENDDO
                  ENDDO

! need to find out what the cell lengths are
                  CALL force_env_get(force_env(ibox)%force_env, &
                                     subsys=oldsys(ibox)%subsys, cell=cell(ibox)%cell)
                  CALL get_cell(cell(ibox)%cell, abc=abc(:, ibox))

                  CALL get_mc_par(mc_par(ibox)%mc_par, &
                                  mc_bias_file=mc_bias_file)
                  nchains_box => nchains(:, ibox)

                  CALL mc_create_bias_force_env(bias_env(ibox)%force_env, &
                                                r_old(:, :, ibox), atom_names_box(:), nunits_tot(ibox), &
                                                para_env, abc(:, ibox), nchains_box, input_declaration, &
                                                mc_bias_file, ionode)

                  IF (SUM(nchains(:, ibox)) .NE. 0) THEN
                     CALL force_env_calc_energy_force( &
                        bias_env(ibox)%force_env, &
                        calc_force=.FALSE.)
                     CALL force_env_get(bias_env(ibox)%force_env, &
                                        potential_energy=last_bias_energy(ibox))
                  ELSE
                     last_bias_energy(ibox) = 0.0E0_dp
                  ENDIF
                  bias_energy(ibox) = last_bias_energy(ibox)
                  DEALLOCATE (atom_names_box)
               ENDDO
            ENDIF

         ELSEIF (rand .LT. pmswap) THEN

            ! try a swap move
            IF (MOD(nnstep, iprint) == 0 .AND. (iw > 0)) THEN
               WRITE (iw, *) "Attempting a swap move"
               WRITE (iw, *)
            ENDIF

            CALL mc_ge_swap_move(mc_par, force_env, bias_env, moves, &
                                 energy_check(:), r_old(:, :, :), old_energy(:), input_declaration, &
                                 para_env, bias_energy(:), last_bias_energy(:), rng_stream)

            ! the number of molecules may have changed, which deallocated the whole
            ! mc_molecule_info structure
            CALL get_mc_par(mc_par(1)%mc_par, mc_molecule_info=mc_molecule_info)
            CALL get_mc_molecule_info(mc_molecule_info, conf_prob=conf_prob, &
                                      nchains=nchains, nmol_types=nmol_types, nunits_tot=nunits_tot, &
                                      mol_type=mol_type, nchain_total=nchain_total, nunits=nunits, &
                                      atom_names=atom_names, mass=mass)

         ELSEIF (rand .LT. pmhmc) THEN
! try hybrid Monte Carlo
            IF (MOD(nnstep, iprint) == 0 .AND. (iw > 0)) THEN
               WRITE (iw, *) "Attempting a hybrid Monte Carlo move"
               WRITE (iw, *)
            ENDIF

! pick a box at random
            IF (ionode) rand = next_random_number(rng_stream)
            CALL mp_bcast(rand, source, group)

            DO ibox = 1, nboxes
               IF (rand .LE. pmhmc_box(ibox)) THEN
                  box_number = ibox
                  EXIT
               ENDIF
            ENDDO

            CALL mc_hmc_move(mc_par(box_number)%mc_par, &
                             force_env(box_number)%force_env, globenv, &
                             moves(1, box_number)%moves, &
                             move_updates(1, box_number)%moves, &
                             old_energy(box_number), box_number, &
                             energy_check(box_number), r_old(:, :, box_number), &
                             rng_stream)

         ELSEIF (rand .LT. pmavbmc) THEN
            ! try an AVBMC move
            IF (MOD(nnstep, iprint) == 0 .AND. (iw > 0)) THEN
               WRITE (iw, *) "Attempting an AVBMC1 move"
               WRITE (iw, *)
            ENDIF

            ! first, pick a box to do it for
            IF (ionode) rand = next_random_number(rng_stream)
            CALL mp_bcast(rand, source, group)

            IF (nboxes .EQ. 2) THEN
               IF (rand .LT. 0.1E0_dp) THEN
                  box_number = 1
               ELSE
                  box_number = 2
               ENDIF
            ELSE
               box_number = 1
            ENDIF

            ! now pick a molecule type to do it for
            IF (ionode) rand = next_random_number(rng_stream)
            CALL mp_bcast(rand, source, group)
            molecule_type_swap = 0
            DO imol_type = 1, nmol_types
               IF (rand .LT. pmavbmc_mol(imol_type)) THEN
                  molecule_type_swap = imol_type
                  EXIT
               ENDIF
            ENDDO
            IF (molecule_type_swap == 0) &
               CALL cp__b("motion/mc/mc_ensembles.F",622,'Did not choose a molecule type to swap...check AVBMC input')

            ! now pick a molecule, automatically rejecting the move if the
            ! box is empty or only has one molecule
            IF (SUM(nchains(:, box_number)) .LE. 1) THEN
               ! indicate that we tried a move
               moves(molecule_type_swap, box_number)%moves%empty_avbmc = &
                  moves(molecule_type_swap, box_number)%moves%empty_avbmc+1
            ELSE

               ! pick a molecule to be swapped in the box
               IF (ionode) THEN
                  CALL find_mc_test_molecule(mc_molecule_info, &
                                             start_atom_swap, idum, jdum, rng_stream, &
                                             box=box_number, molecule_type_old=molecule_type_swap)

                  ! pick a molecule to act as the target in the box...we don't care what type
                  DO
                     CALL find_mc_test_molecule(mc_molecule_info, &
                                                start_atom_target, idum, molecule_type_target, &
                                                rng_stream, box=box_number)
                     IF (start_atom_swap .NE. start_atom_target) THEN
                        start_atom_target = start_atom_target+ &
                                            avbmc_atom(molecule_type_target)-1
                        EXIT
                     ENDIF
                  ENDDO

                  ! choose if we're swapping into the bonded region of mol_target, or
                  ! into the nonbonded region
                  rand = next_random_number(rng_stream)

               ENDIF
               CALL mp_bcast(start_atom_swap, source, group)
               CALL mp_bcast(box_number, source, group)
               CALL mp_bcast(start_atom_target, source, group)
               CALL mp_bcast(rand, source, group)

               IF (rand .LT. pbias(molecule_type_swap)) THEN
                  move_type_avbmc = 'in'
               ELSE
                  move_type_avbmc = 'out'
               ENDIF

               CALL mc_avbmc_move(mc_par(box_number)%mc_par, &
                                  force_env(box_number)%force_env, &
                                  bias_env(box_number)%force_env, &
                                  moves(molecule_type_swap, box_number)%moves, &
                                  energy_check(box_number), &
                                  r_old(:, :, box_number), old_energy(box_number), &
                                  start_atom_swap, start_atom_target, molecule_type_swap, &
                                  box_number, bias_energy(box_number), &
                                  last_bias_energy(box_number), &
                                  move_type_avbmc, rng_stream)

            ENDIF

         ELSE

            IF (MOD(nnstep, iprint) == 0 .AND. (iw > 0)) THEN
               WRITE (iw, *) "Attempting an inner move"
               WRITE (iw, *)
            ENDIF

            DO imove = 1, nmoves

               IF (ionode) rand = next_random_number(rng_stream)
               CALL mp_bcast(rand, source, group)
               IF (rand .LT. pmtraion) THEN
                  ! change molecular conformation
                  ! first, pick a box to do it for
                  IF (ionode) rand = next_random_number(rng_stream)
                  CALL mp_bcast(rand, source, group)
                  IF (nboxes .EQ. 2) THEN
                     IF (rand .LT. 0.75E0_dp) THEN
                        box_number = 1
                     ELSE
                        box_number = 2
                     ENDIF
                  ELSE
                     box_number = 1
                  ENDIF

                  ! figure out which molecule type we're looking for
                  IF (ionode) rand = next_random_number(rng_stream)
                  CALL mp_bcast(rand, source, group)
                  molecule_type = 0
                  DO imol_type = 1, nmol_types
                     IF (rand .LT. pmtraion_mol(imol_type)) THEN
                        molecule_type = imol_type
                        EXIT
                     ENDIF
                  ENDDO
                  IF (molecule_type == 0) CALL cp_abort( &
                     cp__l("motion/mc/mc_ensembles.F",716), &
                     'Did not choose a molecule type to conf change...PMTRAION_MOL should not be all 0.0')

                  ! now pick a molecule, automatically rejecting the move if the
                  ! box is empty
                  IF (nchains(molecule_type, box_number) == 0) THEN
                     ! indicate that we tried a move
                     moves(molecule_type, box_number)%moves%empty_conf = &
                        moves(molecule_type, box_number)%moves%empty_conf+1
                  ELSE
                     ! pick a molecule in the box
                     IF (ionode) THEN
                        CALL find_mc_test_molecule(mc_molecule_info, &
                                                   start_atom, idum, &
                                                   jdum, rng_stream, &
                                                   box=box_number, molecule_type_old=molecule_type)

                        ! choose if we're changing a bond length or an angle
                        rand = next_random_number(rng_stream)
                     ENDIF
                     CALL mp_bcast(rand, source, group)
                     CALL mp_bcast(start_atom, source, group)
                     CALL mp_bcast(box_number, source, group)
                     CALL mp_bcast(molecule_type, source, group)

                     ! figure out what kind of move we're doing
                     IF (rand .LT. conf_prob(1, molecule_type)) THEN
                        move_type = 'bond'
                     ELSEIF (rand .LT. (conf_prob(1, molecule_type)+ &
                                        conf_prob(2, molecule_type))) THEN
                        move_type = 'angle'
                     ELSE
                        move_type = 'dihedral'
                     ENDIF
                     box_flag(box_number) = 1
                     CALL mc_conformation_change(mc_par(box_number)%mc_par, &
                                                 force_env(box_number)%force_env, &
                                                 bias_env(box_number)%force_env, &
                                                 moves(molecule_type, box_number)%moves, &
                                                 move_updates(molecule_type, box_number)%moves, &
                                                 start_atom, molecule_type, box_number, &
                                                 bias_energy(box_number), &
                                                 move_type, lreject, rng_stream)
                     IF (lreject) EXIT
                  ENDIF
               ELSEIF (rand .LT. pmtrans) THEN
                  ! translate a whole molecule in the system
                  ! pick a molecule type
                  IF (ionode) rand = next_random_number(rng_stream)
                  CALL mp_bcast(rand, source, group)
                  molecule_type = 0
                  DO imol_type = 1, nmol_types
                     IF (rand .LT. pmtrans_mol(imol_type)) THEN
                        molecule_type = imol_type
                        EXIT
                     ENDIF
                  ENDDO
                  IF (molecule_type == 0) CALL cp_abort( &
                     cp__l("motion/mc/mc_ensembles.F",774), &
                     'Did not choose a molecule type to translate...PMTRANS_MOL should not be all 0.0')

                  ! now pick a molecule of that type
                  IF (ionode) &
                     CALL find_mc_test_molecule(mc_molecule_info, &
                                                start_atom, box_number, idum, rng_stream, &
                                                molecule_type_old=molecule_type)
                  CALL mp_bcast(start_atom, source, group)
                  CALL mp_bcast(box_number, source, group)
                  box_flag(box_number) = 1
                  CALL mc_molecule_translation(mc_par(box_number)%mc_par, &
                                               force_env(box_number)%force_env, &
                                               bias_env(box_number)%force_env, &
                                               moves(molecule_type, box_number)%moves, &
                                               move_updates(molecule_type, box_number)%moves, &
                                               start_atom, box_number, bias_energy(box_number), &
                                               molecule_type, lreject, rng_stream)
                  IF (lreject) EXIT
               ELSE
                  !     rotate a whole molecule in the system
                  ! pick a molecule type
                  IF (ionode) rand = next_random_number(rng_stream)
                  CALL mp_bcast(rand, source, group)
                  molecule_type = 0
                  DO imol_type = 1, nmol_types
                     IF (rand .LT. pmrot_mol(imol_type)) THEN
                        molecule_type = imol_type
                        EXIT
                     ENDIF
                  ENDDO
                  IF (molecule_type == 0) CALL cp_abort( &
                     cp__l("motion/mc/mc_ensembles.F",806), &
                     'Did not choose a molecule type to rotate...PMROT_MOL should not be all 0.0')

                  IF (ionode) &
                     CALL find_mc_test_molecule(mc_molecule_info, &
                                                start_atom, box_number, idum, rng_stream, &
                                                molecule_type_old=molecule_type)
                  CALL mp_bcast(start_atom, source, group)
                  CALL mp_bcast(box_number, source, group)
                  box_flag(box_number) = 1
                  CALL mc_molecule_rotation(mc_par(box_number)%mc_par, &
                                            force_env(box_number)%force_env, &
                                            bias_env(box_number)%force_env, &
                                            moves(molecule_type, box_number)%moves, &
                                            move_updates(molecule_type, box_number)%moves, &
                                            box_number, start_atom, &
                                            molecule_type, bias_energy(box_number), &
                                            lreject, rng_stream)
                  IF (lreject) EXIT
               ENDIF

            ENDDO

            ! now do a Quickstep calculation to see if we accept the sequence
            CALL mc_Quickstep_move(mc_par, force_env, bias_env, &
                                   moves, lreject, move_updates, energy_check(:), r_old(:, :, :), &
                                   nnstep, old_energy(:), bias_energy(:), last_bias_energy(:), &
                                   nboxes, box_flag(:), oldsys, particles_old, &
                                   rng_stream, unit_conv)

         ENDIF

         ! make sure the pointers are pointing correctly since the subsys may
         ! have changed
         DO ibox = 1, nboxes
            CALL force_env_get(force_env(ibox)%force_env, &
                               subsys=oldsys(ibox)%subsys, cell=cell(ibox)%cell)
            CALL get_cell(cell(ibox)%cell, abc=abc(:, ibox))
            CALL cp_subsys_get(oldsys(ibox)%subsys, &
                               particles=particles_old(ibox)%list)
         ENDDO

         IF (ionode) THEN

            IF (MOD(nnstep, iprint) == 0) THEN
               WRITE (com_ene, *) nnstep, old_energy(1:nboxes)

               DO ibox = 1, nboxes

                  ! write the molecule information
                  WRITE (com_mol, *) nnstep, nchains(:, ibox)

                  ! write the move statistics to file
                  DO itype = 1, nmol_types
                     CALL write_move_stats(moves(itype, ibox)%moves, &
                                           nnstep, move_unit(ibox))
                  ENDDO

                  ! write a restart file
                  CALL write_mc_restart(nnstep, mc_par(ibox)%mc_par, &
                                        nchains(:, ibox), force_env(ibox)%force_env)

                  ! write cell lengths
                  WRITE (cell_unit, *) nnstep, abc(1:3, ibox)*angstrom

                  ! write particle coordinates
                  WRITE (cbox, '(I4)') ibox
                  WRITE (cstep, '(I8)') nnstep
                  IF (SUM(nchains(:, ibox)) == 0) THEN
                     WRITE (com_crd, *) ' 0'
                     WRITE (com_crd, *) 'BOX '//TRIM(ADJUSTL(cbox))// &
                        ',  STEP '//TRIM(ADJUSTL(cstep))
                  ELSE
                     CALL write_particle_coordinates( &
                        particles_old(ibox)%list%els, &
                        com_crd, dump_xmol, 'POS', &
                        'BOX '//TRIM(ADJUSTL(cbox))// &
                        ',  STEP '//TRIM(ADJUSTL(cstep)), &
                        unit_conv=unit_conv)
                  ENDIF
               ENDDO
            ENDIF ! end the things we only do every iprint moves

            DO ibox = 1, nboxes
               ! compute some averages
               averages(ibox)%averages%ave_energy = &
                  averages(ibox)%averages%ave_energy*REAL(nnstep- &
                                                          nstart-1, dp)/REAL(nnstep-nstart, dp)+ &
                  old_energy(ibox)/REAL(nnstep-nstart, dp)
               averages(ibox)%averages%molecules = &
                  averages(ibox)%averages%molecules*REAL(nnstep- &
                                                         nstart-1, dp)/REAL(nnstep-nstart, dp)+ &
                  REAL(SUM(nchains(:, ibox)), dp)/REAL(nnstep-nstart, dp)
               averages(ibox)%averages%ave_volume = &
                  averages(ibox)%averages%ave_volume* &
                  REAL(nnstep-nstart-1, dp)/REAL(nnstep-nstart, dp)+ &
                  abc(1, ibox)*abc(2, ibox)*abc(3, ibox)/ &
                  REAL(nnstep-nstart, dp)

               ! flush the buffers to the files
               CALL m_flush(data_unit(ibox))
               CALL m_flush(diff(ibox))
               CALL m_flush(move_unit(ibox))
               CALL m_flush(cl(ibox))
               CALL m_flush(rm(ibox))

            ENDDO

            ! flush more buffers to the files
            CALL m_flush(cell_unit)
            CALL m_flush(com_ene)
            CALL m_flush(com_crd)
            CALL m_flush(com_mol)

         ENDIF

         ! reset the box flags
         box_flag(:) = 0

         ! check to see if EXIT file exists...if so, end the calculation
         CALL external_control(should_stop, "MC", globenv=globenv)
         IF (should_stop) EXIT

         ! update the move displacements, if necessary
         DO ibox = 1, nboxes
            IF (MOD(nnstep-nstart, iuptrans) == 0) THEN
               DO itype = 1, nmol_types
                  CALL mc_move_update(mc_par(ibox)%mc_par, &
                                      move_updates(itype, ibox)%moves, itype, &
                                      "trans", nnstep, ionode)
               ENDDO
            ENDIF

            IF (MOD(nnstep-nstart, iupvolume) == 0) THEN
               CALL mc_move_update(mc_par(ibox)%mc_par, &
                                   move_updates(1, ibox)%moves, 1337, &
                                   "volume", nnstep, ionode)
            ENDIF
         ENDDO

         ! check to see if there are any overlaps in the boxes, and fold coordinates
! don't care about overlaps if we're only doing HMC
         IF (.NOT. lhmc) THEN
            DO ibox = 1, nboxes
               IF (SUM(nchains(:, ibox)) .NE. 0) THEN
                  start_mol = 1
                  DO jbox = 1, ibox-1
                     start_mol = start_mol+SUM(nchains(:, jbox))
                  ENDDO
                  end_mol = start_mol+SUM(nchains(:, ibox))-1
                  CALL check_for_overlap(force_env(ibox)%force_env, &
                                         nchains(:, ibox), nunits, loverlap, &
                                         mol_type(start_mol:end_mol))
                  IF (loverlap) THEN
                     IF (iw > 0) WRITE (iw, *) nnstep
                     CALL cp__b("motion/mc/mc_ensembles.F",961,'coordinate overlap at the end of the above step')
                     ! now fold the coordinates...don't do this anywhere but here, because
                     ! we can get screwed up with the mc_molecule_info stuff (like in swap move)...
                     ! this is kind of ugly, with allocated and deallocating every time
                     ALLOCATE (r_temp(1:3, 1:nunits_tot(ibox)))

                     DO iunit = 1, nunits_tot(ibox)
                        r_temp(1:3, iunit) = &
                           particles_old(ibox)%list%els(iunit)%r(1:3)
                     ENDDO

                     CALL mc_coordinate_fold(r_temp(:, :), &
                                             SUM(nchains(:, ibox)), mol_type(start_mol:end_mol), &
                                             mass, nunits, abc(1:3, ibox))

                     ! save the folded coordinates
                     DO iunit = 1, nunits_tot(ibox)
                        r_old(1:3, iunit, ibox) = r_temp(1:3, iunit)
                        particles_old(ibox)%list%els(iunit)%r(1:3) = &
                           r_temp(1:3, iunit)
                     ENDDO

                     ! if we're biasing, we need to do the same
                     IF (lbias) THEN
                        CALL force_env_get(bias_env(ibox)%force_env, &
                                           subsys=biassys)
                        CALL cp_subsys_get(biassys, &
                                           particles=particles_bias)

                        DO iunit = 1, nunits_tot(ibox)
                           particles_bias%els(iunit)%r(1:3) = &
                              r_temp(1:3, iunit)
                        ENDDO
                     ENDIF

                     DEALLOCATE (r_temp)
                  ENDIF
               ENDIF
            ENDDO
         ENDIF

         !debug code
         IF (debug_this_module) THEN
            DO ibox = 1, nboxes
               IF (SUM(nchains(:, ibox)) .NE. 0) THEN
                  CALL force_env_calc_energy_force(force_env(ibox)%force_env, &
                                                   calc_force=.FALSE.)
                  CALL force_env_get(force_env(ibox)%force_env, &
                                     potential_energy=test_energy)
               ELSE
                  test_energy = 0.0E0_dp
               ENDIF

               IF (ABS(initial_energy(ibox)+energy_check(ibox)- &
                       test_energy) .GT. 0.0000001E0_dp) THEN
                  IF (iw > 0) THEN
                     WRITE (iw, *) '!!!!!!! We have an energy problem. !!!!!!!!'
                     WRITE (iw, '(A,T64,F16.10)') 'Final Energy = ', test_energy
                     WRITE (iw, '(A,T64,F16.10)') 'Inital Energy+energy_check=', &
                        initial_energy(ibox)+energy_check(ibox)
                     WRITE (iw, *) 'Box ', ibox
                     WRITE (iw, *) 'nchains ', nchains(:, ibox)
                  END IF
                  CALL cp__b("motion/mc/mc_ensembles.F",1024,'!!!!!!! We have an energy problem. !!!!!!!!')
               ENDIF
            ENDDO
         ENDIF
      ENDDO

      ! write a restart file
      IF (ionode) THEN
         DO ibox = 1, nboxes
            CALL write_mc_restart(nnstep, mc_par(ibox)%mc_par, &
                                  nchains(:, ibox), force_env(ibox)%force_env)
         ENDDO
      ENDIF

      ! calculate the final energy
      DO ibox = 1, nboxes
         IF (SUM(nchains(:, ibox)) .NE. 0) THEN
            CALL force_env_calc_energy_force(force_env(ibox)%force_env, &
                                             calc_force=.FALSE.)
            CALL force_env_get(force_env(ibox)%force_env, &
                               potential_energy=final_energy(ibox))
         ELSE
            final_energy(ibox) = 0.0E0_dp
         ENDIF
         IF (lbias) THEN
            CALL force_env_release(bias_env(ibox)%force_env)
         ENDIF
      ENDDO

      ! do some stuff in serial
      IF (ionode .OR. (iw > 0)) THEN

         WRITE (com_ene, *) 'Final Energies:                      ', &
            final_energy(1:nboxes)

         DO ibox = 1, nboxes
            WRITE (cbox, '(I4)') ibox
            IF (SUM(nchains(:, ibox)) == 0) THEN
               WRITE (com_crd, *) ' 0'
               WRITE (com_crd, *) 'BOX '//TRIM(ADJUSTL(cbox))
            ELSE
               CALL write_particle_coordinates( &
                  particles_old(ibox)%list%els, &
                  com_crd, dump_xmol, 'POS', &
                  'FINAL BOX '//TRIM(ADJUSTL(cbox)), unit_conv=unit_conv)
            ENDIF

            ! write a bunch of data to the screen
            WRITE (iw, '(A)') &
               '------------------------------------------------'
            WRITE (iw, '(A,I1,A)') &
               '|                   BOX ', ibox, &
               '                      |'
            WRITE (iw, '(A)') &
               '------------------------------------------------'
            test_moves => moves(:, ibox)
            CALL final_mc_write(mc_par(ibox)%mc_par, test_moves, &
                                iw, energy_check(ibox), &
                                initial_energy(ibox), final_energy(ibox), &
                                averages(ibox)%averages)

            ! close any open files
            CALL close_file(unit_number=diff(ibox))
            CALL close_file(unit_number=data_unit(ibox))
            CALL close_file(unit_number=move_unit(ibox))
            CALL close_file(unit_number=cl(ibox))
            CALL close_file(unit_number=rm(ibox))
         ENDDO

         ! close some more files
         CALL close_file(unit_number=cell_unit)
         CALL close_file(unit_number=com_ene)
         CALL close_file(unit_number=com_crd)
         CALL close_file(unit_number=com_mol)
      ENDIF

      DO ibox = 1, nboxes
         CALL set_mc_env(mc_env(ibox)%mc_env, &
                         mc_par=mc_par(ibox)%mc_par, &
                         force_env=force_env(ibox)%force_env)

         ! deallocate some stuff
         DO itype = 1, nmol_types
            CALL mc_moves_release(move_updates(itype, ibox)%moves)
            CALL mc_moves_release(moves(itype, ibox)%moves)
         ENDDO
         CALL mc_averages_release(averages(ibox)%averages)
      ENDDO

      DEALLOCATE (pmhmc_box)
      DEALLOCATE (pmvol_box)
      DEALLOCATE (r_old)
      DEALLOCATE (force_env)
      DEALLOCATE (bias_env)
      DEALLOCATE (cell)
      DEALLOCATE (particles_old)
      DEALLOCATE (oldsys)
      DEALLOCATE (averages)
      DEALLOCATE (moves)
      DEALLOCATE (move_updates)
      DEALLOCATE (mc_par)

      ! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_run_ensemble

! **************************************************************************************************
!> \brief Computes the second virial coefficient of a molecule by using the integral form
!>      of the second virial coefficient found in McQuarrie "Statistical Thermodynamics",
!>      B2(T) = -2Pi Int 0toInf [ Exp[-beta*u(r)] -1] r^2 dr     Eq. 15-25
!>      I use trapazoidal integration with various step sizes
!>      (the integral is broken up into three parts, currently, but that's easily
!>      changed by the first variables found below).  It generates nvirial configurations,
!>      doing the integration for each one, and then averages all the B2(T) to produce
!>      the final answer.
!> \param mc_env a pointer that contains all mc_env for all the simulation
!>          boxes
!> \param rng_stream the stream we pull random numbers from
!>
!>    Suitable for parallel.
!> \author MJM
! **************************************************************************************************
   SUBROUTINE mc_compute_virial(mc_env, rng_stream)

      TYPE(mc_environment_p_type), DIMENSION(:), POINTER :: mc_env
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(len=*), PARAMETER :: routineN = 'mc_compute_virial', &
         routineP = moduleN//':'//routineN

      INTEGER :: current_division, end_atom, group, ibin, idivision, iparticle, iprint, itemp, &
         iunit, ivirial, iw, nbins, nchain_total, nintegral_divisions, nmol_types, nvirial, &
         nvirial_temps, source, start_atom
      INTEGER, DIMENSION(:), POINTER                     :: mol_type, nunits, nunits_tot
      INTEGER, DIMENSION(:, :), POINTER                  :: nchains
      LOGICAL                                            :: ionode, loverlap
      REAL(dp), DIMENSION(:), POINTER                    :: BETA, virial_cutoffs, virial_stepsize, &
                                                            virial_temps
      REAL(dp), DIMENSION(:, :), POINTER                 :: mass, mayer, r_old
      REAL(KIND=dp) :: ave_virial, current_value, distance, exp_max_val, exp_min_val, exponent, &
         integral, previous_value, square_value, trial_energy, triangle_value
      REAL(KIND=dp), DIMENSION(1:3)                      :: abc, center_of_mass
      TYPE(cell_p_type), DIMENSION(:), POINTER           :: cell
      TYPE(cp_subsys_p_type), DIMENSION(:), POINTER      :: subsys
      TYPE(force_env_p_type), DIMENSION(:), POINTER      :: force_env
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(mc_simulation_parameters_p_type), &
         DIMENSION(:), POINTER                           :: mc_par
      TYPE(particle_list_p_type), DIMENSION(:), POINTER  :: particles

! these are current magic numbers for how we compute the virial...
! we break it up into three parts to integrate the function so provide
! better statistics

      nintegral_divisions = 3
      ALLOCATE (virial_cutoffs(1:nintegral_divisions))
      ALLOCATE (virial_stepsize(1:nintegral_divisions))
      virial_cutoffs(1) = 8.0 ! first distance, in bohr
      virial_cutoffs(2) = 13.0 ! second distance, in bohr
      virial_cutoffs(3) = 22.0 ! maximum distance, in bohr
      virial_stepsize(1) = 0.04 ! stepsize from 0 to virial_cutoffs(1)
      virial_stepsize(2) = 0.1
      virial_stepsize(3) = 0.2

      nbins = CEILING(virial_cutoffs(1)/virial_stepsize(1)+(virial_cutoffs(2)-virial_cutoffs(1))/ &
                      virial_stepsize(2)+(virial_cutoffs(3)-virial_cutoffs(2))/virial_stepsize(3))

      ! figure out what the default write unit is
      iw = cp_logger_get_default_io_unit()

      ! allocate a whole bunch of stuff based on how many boxes we have
      ALLOCATE (force_env(1:1))
      ALLOCATE (cell(1:1))
      ALLOCATE (particles(1:1))
      ALLOCATE (subsys(1:1))
      ALLOCATE (mc_par(1:1))

      CALL get_mc_env(mc_env(1)%mc_env, &
                      mc_par=mc_par(1)%mc_par, &
                      force_env=force_env(1)%force_env)

      ! get some data out of mc_par
      CALL get_mc_par(mc_par(1)%mc_par, &
                      exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, nvirial=nvirial, &
                      ionode=ionode, source=source, group=group, &
                      mc_molecule_info=mc_molecule_info, virial_temps=virial_temps)

      IF (iw > 0) THEN
         WRITE (iw, *)
         WRITE (iw, *)
         WRITE (iw, *) 'Beginning the calculation of the second virial coefficient'
         WRITE (iw, *)
         WRITE (iw, *)
      ENDIF

      ! get some data from the molecule types
      CALL get_mc_molecule_info(mc_molecule_info, &
                                nchains=nchains, nmol_types=nmol_types, nunits_tot=nunits_tot, &
                                mol_type=mol_type, nchain_total=nchain_total, nunits=nunits, &
                                mass=mass)

      nvirial_temps = SIZE(virial_temps)
      ALLOCATE (BETA(1:nvirial_temps))

      DO itemp = 1, nvirial_temps
         BETA(itemp) = 1/virial_temps(itemp)/boltzmann*joule
      ENDDO

      ! get the subsystems and the cell information
      CALL force_env_get(force_env(1)%force_env, &
                         subsys=subsys(1)%subsys, cell=cell(1)%cell)
      CALL get_cell(cell(1)%cell, abc=abc(:))
      CALL cp_subsys_get(subsys(1)%subsys, &
                         particles=particles(1)%list)

      ! check and make sure the box is big enough
      IF (abc(1) .NE. abc(2) .OR. abc(2) .NE. abc(3)) THEN
         CALL cp__b("motion/mc/mc_ensembles.F",1243,'The box needs to be cubic for a virial calculation (it is easiest).')
      END IF
      IF (virial_cutoffs(nintegral_divisions) .GT. abc(1)/2.0E0_dp) THEN
         IF (iw > 0) &
            WRITE (iw, *) "Box length ", abc(1)*angstrom, " virial cutoff ", &
            virial_cutoffs(nintegral_divisions)*angstrom
         CALL cp__b("motion/mc/mc_ensembles.F",1249,'You need a bigger box to deal with this virial cutoff (see above).')
      END IF

      ! store the coordinates of the molecules in an array so we can work with it
      ALLOCATE (r_old(1:3, 1:nunits_tot(1)))

      DO iparticle = 1, nunits_tot(1)
         r_old(1:3, iparticle) = &
            particles(1)%list%els(iparticle)%r(1:3)
      ENDDO

      ! move the center of mass of molecule 1 to the origin
      start_atom = 1
      end_atom = nunits(mol_type(1))
      CALL get_center_of_mass(r_old(:, start_atom:end_atom), nunits(mol_type(1)), &
                              center_of_mass(:), mass(1:nunits(mol_type(1)), mol_type(1)))
      DO iunit = start_atom, end_atom
         r_old(:, iunit) = r_old(:, iunit)-center_of_mass(:)
      ENDDO
      ! set them in the force_env, so the first molecule is ready for the energy calc
      DO iparticle = start_atom, end_atom
         particles(1)%list%els(iparticle)%r(1:3) = r_old(1:3, iparticle)
      ENDDO

      ! print out a notice every 1%
      iprint = FLOOR(REAL(nvirial)/100.0)
      IF (iprint == 0) iprint = 1

      ! we'll compute the average potential, and then integrate that, as opposed to
      ! integrating every orientation and then averaging
      ALLOCATE (mayer(1:nvirial_temps, 1:nbins))

      mayer(:, :) = 0.0_dp

      ! loop over all nvirial random configurations
      DO ivirial = 1, nvirial

         ! move molecule two back to the origin
         start_atom = nunits(mol_type(1))+1
         end_atom = nunits_tot(1)
         CALL get_center_of_mass(r_old(:, start_atom:end_atom), nunits(mol_type(2)), &
                                 center_of_mass(:), mass(1:nunits(mol_type(2)), mol_type(2)))
         DO iunit = start_atom, end_atom
            r_old(:, iunit) = r_old(:, iunit)-center_of_mass(:)
         ENDDO

         ! now we need a random orientation for molecule 2...this routine is
         ! only done in serial since it calls a random number
         IF (ionode) THEN
            CALL rotate_molecule(r_old(:, start_atom:end_atom), &
                                 mass(1:nunits(mol_type(2)), mol_type(2)), &
                                 nunits(mol_type(2)), rng_stream)
         ENDIF
         CALL mp_bcast(r_old(:, :), source, group)

         distance = 0.0E0_dp
         ibin = 1
         DO
            ! find out what our stepsize is
            current_division = 0
            DO idivision = 1, nintegral_divisions
               IF (distance .LT. virial_cutoffs(idivision)-virial_stepsize(idivision)/2.0E0_dp) THEN
                  current_division = idivision
                  EXIT
               ENDIF
            ENDDO
            IF (current_division == 0) EXIT
            distance = distance+virial_stepsize(current_division)

            ! move the second molecule only along the x direction
            DO iparticle = start_atom, end_atom
               particles(1)%list%els(iparticle)%r(1) = r_old(1, iparticle)+distance
               particles(1)%list%els(iparticle)%r(2) = r_old(2, iparticle)
               particles(1)%list%els(iparticle)%r(3) = r_old(3, iparticle)
            ENDDO

            ! check for overlaps
            CALL check_for_overlap(force_env(1)%force_env, nchains(:, 1), nunits, loverlap, mol_type)

            ! compute the energy if there is no overlap
            ! exponent is exp(-beta*energy)-1, also called the Mayer term
            IF (loverlap) THEN
               DO itemp = 1, nvirial_temps
                  mayer(itemp, ibin) = mayer(itemp, ibin)-1.0_dp
               ENDDO
            ELSE
               CALL force_env_calc_energy_force(force_env(1)%force_env, &
                                                calc_force=.FALSE.)
               CALL force_env_get(force_env(1)%force_env, &
                                  potential_energy=trial_energy)

               DO itemp = 1, nvirial_temps

                  exponent = -BETA(itemp)*trial_energy

                  IF (exponent .GT. exp_max_val) THEN
                     exponent = exp_max_val
                  ELSEIF (exponent .LT. exp_min_val) THEN
                     exponent = exp_min_val
                  ENDIF
                  mayer(itemp, ibin) = mayer(itemp, ibin)+EXP(exponent)-1.0_dp
               ENDDO
            ENDIF

            ibin = ibin+1
         ENDDO
         ! write out some info that keeps track of where we are
         IF (iw > 0) THEN
            IF (MOD(ivirial, iprint) == 0) &
               WRITE (iw, '(A,I6,A,I6)') ' Done with config ', ivirial, ' out of ', nvirial
         ENDIF
      ENDDO

      ! now we integrate this average potential
      mayer(:, :) = mayer(:, :)/REAL(nvirial, dp)

      DO itemp = 1, nvirial_temps
         integral = 0.0_dp
         previous_value = 0.0_dp
         distance = 0.0E0_dp
         ibin = 1
         DO
            current_division = 0
            DO idivision = 1, nintegral_divisions
               IF (distance .LT. virial_cutoffs(idivision)-virial_stepsize(idivision)/2.0E0_dp) THEN
                  current_division = idivision
                  EXIT
               ENDIF
            ENDDO
            IF (current_division == 0) EXIT
            distance = distance+virial_stepsize(current_division)

            ! now we need to integrate, using the trapazoidal method
            ! first, find the value of the square
            current_value = mayer(itemp, ibin)*distance**2
            square_value = previous_value*virial_stepsize(current_division)
            ! now the triangle that sits on top of it, which is half the size of this square...
            ! notice this is negative if the current value is less than the previous value
            triangle_value = 0.5E0_dp*((current_value-previous_value)*virial_stepsize(current_division))

            integral = integral+square_value+triangle_value
            previous_value = current_value
            ibin = ibin+1
         ENDDO

         ! now that the integration is done, compute the second virial that results
         ave_virial = -2.0E0_dp*pi*integral

         ! convert from CP2K units to something else
         ave_virial = ave_virial*n_avogadro*angstrom**3/1.0E8_dp**3

         IF (iw > 0) THEN
            WRITE (iw, *)
            WRITE (iw, *) '*********************************************************************'
            WRITE (iw, '(A,F12.6,A)') ' ***                Temperature = ', virial_temps(itemp), &
               '                     ***'
            WRITE (iw, *) '***                                                               ***'
            WRITE (iw, '(A,E12.6,A)') ' ***                  B2(T) = ', ave_virial, &
               ' cm**3/mol               ***'
            WRITE (iw, *) '*********************************************************************'
            WRITE (iw, *)
         ENDIF
      ENDDO

      ! deallocate some stuff
      DEALLOCATE (mc_par)
      DEALLOCATE (subsys)
      DEALLOCATE (force_env)
      DEALLOCATE (particles)
      DEALLOCATE (cell)
      DEALLOCATE (virial_cutoffs)
      DEALLOCATE (virial_stepsize)
      DEALLOCATE (r_old)
      DEALLOCATE (mayer)
      DEALLOCATE (BETA)

   END SUBROUTINE mc_compute_virial

END MODULE mc_ensembles

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief preps the system for a Monte Carlo run (sets up some environments,
!>      calls the routines to read in the MC parameters)...converted
!>      from qs_mc.F
!> \par Literature
!>    a list of papers for the theory behind various MC moves
!>    Books:
!>       D. Frenkel, B. Smit: Understanding Molecular Simulation (1996)
!>       M.P. Allen, D.J. Tildesley: Computer Simulations of Liquids (1987)
!> \par
!>    Aggregation volume bias Monte Carlo (AVBMC):
!>       Chen, B.; Siepmann, J.I.  J. Phys. Chem. B 2000, 104, 8725.
!> \par
!>    Biasing with an inexpensive potential:
!>       Iftimie et al.  J. Chem. Phys. 2000, 113, 4852.
!>       Gelb, L. D.  J. Chem. Phys. 2003, 118, 7747.
!> \par
!>    Configurational bias Monte Carlo (CBMC):
!>       Siepmann, J.I.; Frenkel, D.  Mol. Phys. 1992, 75, 59.
!> \par
!>    Gibbs ensemble Monte Carlo (GEMC):
!>       Panagiotopoulos, A.Z.  Mol. Phys. 1987, 61, 813.
!>       Panagiotopoulos et al.  Mol. Phys. 1988, 63, 527.
!>       Smit et al.  Mol. Phys. 1989, 68, 931.
!> \par
!>    Isobaric-isothermal ensemble:
!>       McDonald, I.R.  Mol. Phys. 1972, 23, 41.
!> \par
!>    Original Monte Carlo paper:
!>       Metropolis et al.  J. Chem. Phys. 1953, 21, 1087.
!> \author MJM-Oct-15-03
! **************************************************************************************************
MODULE mc_run
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE force_env_types,                 ONLY: force_env_p_type,&
                                              force_env_release,&
                                              force_env_retain,&
                                              force_env_type,&
                                              use_fist_force
   USE global_types,                    ONLY: global_environment_type
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              int_size
   USE mc_control,                      ONLY: mc_create_force_env,&
                                              read_mc_restart
   USE mc_ensembles,                    ONLY: mc_compute_virial,&
                                              mc_run_ensemble
   USE mc_environment_types,            ONLY: get_mc_env,&
                                              mc_env_create,&
                                              mc_env_release,&
                                              mc_environment_p_type,&
                                              set_mc_env
   USE mc_types,                        ONLY: &
        find_mc_rcut, get_mc_molecule_info, get_mc_par, mc_determine_molecule_info, &
        mc_input_file_create, mc_input_file_destroy, mc_input_file_type, &
        mc_input_parameters_check, mc_molecule_info_destroy, mc_molecule_info_type, &
        mc_sim_par_create, mc_sim_par_destroy, mc_simulation_parameters_p_type, read_mc_section, &
        set_mc_par
   USE parallel_rng_types,              ONLY: UNIFORM,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              rng_stream_type
   USE physcon,                         ONLY: angstrom
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mc_run'

   PUBLIC :: do_mon_car

!-----------------------------------------------------------------------------!

CONTAINS

! **************************************************************************************************
!> \brief starts the Monte Carlo simulation and determines which ensemble we're
!>      running
!> \param force_env_1 the force environment for the simulation, or
!>                   the force environment for box 1, depending on which
!>                   ensemble we're running
!> \param globenv the global environment for the simulation
!> \param input_declaration ...
!> \param input_file_name the name of the input file that force_env_1 was
!>        created from
!> \author MJM
!> \note
!>      Designed for parallel.
! **************************************************************************************************
   SUBROUTINE do_mon_car(force_env_1, globenv, input_declaration, input_file_name)

      TYPE(force_env_type), POINTER                      :: force_env_1
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_type), POINTER                        :: input_declaration
      CHARACTER(LEN=*)                                   :: input_file_name

      CHARACTER(LEN=*), PARAMETER :: routineN = 'do_mon_car', routineP = moduleN//':'//routineN

      CHARACTER(LEN=20)                                  :: ensemble
      CHARACTER(LEN=40)                                  :: box2_file, dat_file
      INTEGER                                            :: box_number, ibox, isos, istat, iw, &
                                                            nboxes, nmol_types
      INTEGER, DIMENSION(:), POINTER                     :: nunits_tot
      LOGICAL                                            :: lbias, lhmc, lrestart, lskip_box, &
                                                            lterminate
      REAL(dp)                                           :: rcut
      REAL(dp), DIMENSION(:, :), POINTER                 :: empty_coordinates
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(force_env_p_type), DIMENSION(:), POINTER      :: force_env
      TYPE(force_env_type), POINTER                      :: force_env_2
      TYPE(global_environment_type), POINTER             :: globenv_2
      TYPE(mc_environment_p_type), DIMENSION(:), POINTER :: mc_env
      TYPE(mc_input_file_type), POINTER                  :: mc_bias_file, mc_input_file
      TYPE(mc_molecule_info_type), POINTER               :: mc_molecule_info
      TYPE(mc_simulation_parameters_p_type), &
         DIMENSION(:), POINTER                           :: mc_par
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(section_vals_type), POINTER                   :: force_env_section, mc_section, &
                                                            root_section

      NULLIFY (mc_env, mc_par, force_env_2, rng_stream, force_env_section, &
               root_section, mc_molecule_info)

      CALL force_env_retain(force_env_1)

      para_env => force_env_1%para_env
      iw = cp_logger_get_default_io_unit()
      force_env_section => force_env_1%force_env_section
      root_section => force_env_1%root_section
      CALL section_vals_get(force_env_section, n_repetition=isos)
      IF(.NOT.(isos == 1))CALL cp__a("motion/mc/mc_run.F",145)
! set some values...will use get_globenv if that ever comes around

! initialize the random numbers
      CALL create_rng_stream(rng_stream=rng_stream, &
                             last_rng_stream=force_env_1%globenv%gaussian_rng_stream, &
                             name="first", &
                             distribution_type=UNIFORM)

! need to figure out how many boxes we have, based on the value
! of mc_par % ensemble
      NULLIFY (mc_section)
      mc_section => section_vals_get_subs_vals(root_section, &
                                               "MOTION%MC")
      CALL section_vals_val_get(mc_section, "ENSEMBLE", &
                                c_val=ensemble)

! now we read in the second force_env, if we have another box
      SELECT CASE (ensemble)
      CASE ("TRADITIONAL", "VIRIAL")
         nboxes = 1
      CASE ("GEMC_NVT", "GEMC_NPT")
         nboxes = 2
         CALL section_vals_val_get(mc_section, "BOX2_FILE_NAME", &
                                   c_val=box2_file)
         CALL mc_create_force_env(force_env_2, input_declaration, para_env, &
                                  box2_file, globenv_new=globenv_2)
      END SELECT

! now we create the various pointers that contain information for all boxes
      ALLOCATE (force_env(1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 176, &
                                       "force_env", int_size*nboxes)
      SELECT CASE (ensemble)
      CASE ("TRADITIONAL", "VIRIAL")
         force_env(1)%force_env => force_env_1
      CASE ("GEMC_NVT", "GEMC_NPT")
         force_env(1)%force_env => force_env_1
         force_env(2)%force_env => force_env_2
      END SELECT
      ALLOCATE (mc_par(1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 186, &
                                       "mc_par", int_size*nboxes)
      ALLOCATE (mc_env(1:nboxes), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 189, &
                                       "mc_env", int_size*nboxes)

! now we need the molecule information
! determine the total number of molecules and atoms
      CALL mc_determine_molecule_info(force_env, mc_molecule_info, &
                                      coordinates_empty=empty_coordinates)
      CALL get_mc_molecule_info(mc_molecule_info, nmol_types=nmol_types, &
                                nunits_tot=nunits_tot)

      DO ibox = 1, nboxes

         IF (iw > 0) THEN
            WRITE (iw, *)
            WRITE (iw, *)
            WRITE (iw, '(A,I2,A)') '******************************** Begin BOX ', ibox, &
               '  *******************************'
         ENDIF

! allocates an mc_env and sets the variables to zero
         CALL mc_env_create(mc_env(ibox)%mc_env)

! now read in the values of the mc_pars
! creating the mc_par
         CALL mc_sim_par_create(mc_par(ibox)%mc_par, nmol_types)

! attach molecule information to all mc_par structures, so we know what we have to read in
         CALL set_mc_par(mc_par(ibox)%mc_par, mc_molecule_info=mc_molecule_info)

! read the input of the Monte Carlo section
         force_env_section => force_env(ibox)%force_env%force_env_section
         root_section => force_env(ibox)%force_env%root_section
         IF (ibox == 1) THEN
            CALL read_mc_section(mc_par(ibox)%mc_par, para_env, globenv, input_file_name, &
                                 root_section, force_env_section)
         ELSE
            CALL read_mc_section(mc_par(ibox)%mc_par, para_env, globenv, box2_file, &
                                 root_section, force_env_section)
         ENDIF

! get the input file data, in case we need to make a restart...
! this is also used in the swap move, or anytime we need to make a dat file
! always judge based on lbias from box 1...in case someone only changes the value
! for box 1
         CALL get_mc_par(mc_par(ibox)%mc_par, mc_input_file=mc_input_file, lhmc=lhmc)
         CALL get_mc_par(mc_par(1)%mc_par, lbias=lbias)
         IF (ibox == 1) THEN
            CALL mc_input_file_create(mc_input_file, &
                                      input_file_name, mc_molecule_info, empty_coordinates, lhmc)
         ELSE
            CALL mc_input_file_create(mc_input_file, &
                                      box2_file, mc_molecule_info, empty_coordinates, lhmc)
         ENDIF
         CALL set_mc_par(mc_par(ibox)%mc_par, mc_input_file=mc_input_file)

         IF (lbias) THEN
            CALL get_mc_par(mc_par(ibox)%mc_par, mc_bias_file=mc_bias_file)
            CALL mc_input_file_create(mc_bias_file, &
                                      "bias_template.inp", mc_molecule_info, empty_coordinates, lhmc)
            CALL set_mc_par(mc_par(ibox)%mc_par, mc_bias_file=mc_bias_file)
         ENDIF

! check for restart
         CALL get_mc_par(mc_par(ibox)%mc_par, lrestart=lrestart, &
                         dat_file=dat_file)
         IF (lrestart) THEN
            CALL read_mc_restart(mc_par(ibox)%mc_par, force_env(ibox)%force_env, &
                                 iw, nunits_tot(ibox), rng_stream)
! release the old force env and make the new one
            CALL force_env_release(force_env(ibox)%force_env)
            CALL mc_create_force_env(force_env(ibox)%force_env, &
                                     input_declaration, para_env, dat_file)
         ENDIF

      ENDDO

! figure out if we have an empty box
      box_number = 0
      lskip_box = .FALSE.
      DO ibox = 1, nboxes
         IF (nunits_tot(ibox) == 0) THEN
            IF (lskip_box) THEN
               CALL cp__b("motion/mc/mc_run.F",271,'More than one box has no atoms in it!')
            END IF
            box_number = ibox
            lskip_box = .TRUE.
         ENDIF
      ENDDO

! in case there was a restart, we need to do this again
      CALL mc_molecule_info_destroy(mc_molecule_info)
      CALL mc_determine_molecule_info(force_env, mc_molecule_info, box_number=box_number)
      CALL get_mc_molecule_info(mc_molecule_info, nmol_types=nmol_types, &
                                nunits_tot=nunits_tot)
      DO ibox = 1, nboxes
         CALL set_mc_par(mc_par(ibox)%mc_par, mc_molecule_info=mc_molecule_info)
      ENDDO
! if we're doing a classical simulation, figure out the largest
! potential cutoff and write it to the screen
      IF (force_env(1)%force_env%in_use .EQ. use_fist_force) THEN
         CALL find_mc_rcut(mc_par(1)%mc_par, force_env(1)%force_env, lterminate)
         CALL get_mc_par(mc_par(1)%mc_par, rcut=rcut)
         IF (iw > 0) WRITE (iw, '( A,T73,F8.4 )') &
            ' MC| Interaction cutoff [angstroms]', rcut*angstrom
         IF (lterminate) THEN
            CALL cp__b("motion/mc/mc_run.F",294,'Cutoff larger than twice the boxlength')
         ENDIF
      ENDIF

! make sure some values are the same between boxes
      IF (nboxes == 2) THEN
         CALL equilize_mc_sim_parameters(mc_par, iw)
      ENDIF

! now check the input parameters and run the simulation
      DO ibox = 1, nboxes

         CALL mc_input_parameters_check(mc_par(ibox)%mc_par)

! attach all the structures to one convientent structure
         CALL set_mc_env(mc_env(ibox)%mc_env, mc_par=mc_par(ibox)%mc_par, &
                         force_env=force_env(ibox)%force_env)

      ENDDO

! if we're computing the second virial coefficient, do that instead
! of running a simulation
      SELECT CASE (ensemble)
      CASE ("VIRIAL")
         CALL mc_compute_virial(mc_env, rng_stream)
      CASE DEFAULT
         CALL mc_run_ensemble(mc_env, para_env, globenv, input_declaration, nboxes, rng_stream)
      END SELECT

! get rid of all the MC molecule information
      CALL get_mc_env(mc_env(1)%mc_env, mc_par=mc_par(1)%mc_par)
      CALL get_mc_par(mc_par(1)%mc_par, mc_molecule_info=mc_molecule_info)
      CALL mc_molecule_info_destroy(mc_molecule_info)

      DO ibox = 1, nboxes
         CALL get_mc_env(mc_env(ibox)%mc_env, &
                         mc_par=mc_par(ibox)%mc_par, force_env=force_env(ibox)%force_env)
         CALL get_mc_par(mc_par(ibox)%mc_par, mc_input_file=mc_input_file)

         CALL mc_input_file_destroy(mc_input_file)
         IF (lbias) THEN
            CALL get_mc_par(mc_par(ibox)%mc_par, mc_bias_file=mc_bias_file)
            CALL mc_input_file_destroy(mc_bias_file)
         ENDIF

         CALL mc_sim_par_destroy(mc_par(ibox)%mc_par)
         CALL mc_env_release(mc_env(ibox)%mc_env)
         CALL force_env_release(force_env(ibox)%force_env)
      ENDDO

      DEALLOCATE (empty_coordinates)
      DEALLOCATE (mc_par)
      DEALLOCATE (mc_env)
      DEALLOCATE (force_env)

! delete the random numbers
      CALL delete_rng_stream(rng_stream)

   END SUBROUTINE do_mon_car

! **************************************************************************************************
!> \brief takes some parameters from one set of MC simulation parameters
!>      and transfers them to a second set...used so that we're not using
!>      different parameters between two simulation boxes, if they should
!>      be the same (move probabilities, for instance)
!> \param mc_par the pointer that contains the simulation parameters
!>           for both boxes
!> \param iw the unit number that prints to screen
!> \author MJM
!> \note
!>      Designed for parallel.
! **************************************************************************************************
   SUBROUTINE equilize_mc_sim_parameters(mc_par, iw)
      TYPE(mc_simulation_parameters_p_type), &
         DIMENSION(:), POINTER                           :: mc_par
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(20)                                      :: ensemble
      INTEGER                                            :: iprint, iuptrans, iupvolume, nmoves, &
                                                            nstep, nswapmoves
      INTEGER, DIMENSION(:), POINTER                     :: avbmc_atom
      LOGICAL                                            :: lbias, lrestart, lstop
      REAL(dp)                                           :: BETA, pmswap, pmtraion, pmtrans, &
                                                            pmvolume, pressure, rcut, temperature
      REAL(dp), DIMENSION(:), POINTER                    :: avbmc_rmax, avbmc_rmin, pbias, &
                                                            pmavbmc_mol, pmrot_mol, pmswap_mol, &
                                                            pmtraion_mol, pmtrans_mol

      IF (iw > 0) THEN
         WRITE (iw, '( A,A )') 'Ignoring some input for box 2, and ', &
            'using the values for box 1 for the following variables:'
         WRITE (iw, '( A,A )') 'nstep,iupvolume,iuptrans,nmoves,', &
            'nswapmoves,iprint,lbias,lstop,temperature,pressure'
         WRITE (iw, '( A,A )') 'pmvolume,pmtraion,pmtrans,BETA,rcut,', &
            'lrestart'
         WRITE (iw, '( A,A )') 'pmtraion_mol,pmtrans_mol,pmrot_mol,pmswap_mol,', &
            'avbmc_atom'
         WRITE (iw, '( A,A )') 'avbmc_rmin,avmbc_rmax,pmavbmc_mol,pbias,pmswap'
      ENDIF

      CALL get_mc_par(mc_par(1)%mc_par, nstep=nstep, iupvolume=iupvolume, &
                      iuptrans=iuptrans, nmoves=nmoves, nswapmoves=nswapmoves, &
                      iprint=iprint, lbias=lbias, lstop=lstop, temperature=temperature, &
                      pressure=pressure, pmswap=pmswap, pmvolume=pmvolume, &
                      pmtraion=pmtraion, pmtrans=pmtrans, BETA=BETA, rcut=rcut, &
                      lrestart=lrestart, pmtraion_mol=pmtraion_mol, pmtrans_mol=pmtrans_mol, &
                      pmrot_mol=pmrot_mol, pmswap_mol=pmswap_mol, avbmc_atom=avbmc_atom, &
                      avbmc_rmin=avbmc_rmin, avbmc_rmax=avbmc_rmax, pmavbmc_mol=pmavbmc_mol, &
                      pbias=pbias, ensemble=ensemble)
      CALL set_mc_par(mc_par(2)%mc_par, nstep=nstep, iupvolume=iupvolume, &
                      iuptrans=iuptrans, nmoves=nmoves, nswapmoves=nswapmoves, &
                      iprint=iprint, lbias=lbias, lstop=lstop, temperature=temperature, &
                      pressure=pressure, pmswap=pmswap, pmvolume=pmvolume, &
                      pmtraion=pmtraion, pmtrans=pmtrans, BETA=BETA, rcut=rcut, &
                      lrestart=lrestart, pmtraion_mol=pmtraion_mol, pmtrans_mol=pmtrans_mol, &
                      pmrot_mol=pmrot_mol, pmswap_mol=pmswap_mol, avbmc_atom=avbmc_atom, &
                      avbmc_rmin=avbmc_rmin, avbmc_rmax=avbmc_rmax, pmavbmc_mol=pmavbmc_mol, &
                      pbias=pbias, ensemble=ensemble)

   END SUBROUTINE equilize_mc_sim_parameters

END MODULE mc_run
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Perform a temperature accelarated hybrid monte carlo (TAHMC) run using QUICKSTEP
!> \par History
!>      none
!> \author Alin M Elena
! **************************************************************************************************
MODULE tamc_run

   USE atomic_kind_list_types,          ONLY: atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE averages_types,                  ONLY: average_quantities_type
   USE barostat_types,                  ONLY: barostat_type,&
                                              create_barostat_type,&
                                              release_barostat_type
   USE bibliography,                    ONLY: VandenCic2006
   USE cell_types,                      ONLY: cell_type
   USE colvar_methods,                  ONLY: colvar_eval_glob_f
   USE colvar_types,                    ONLY: HBP_colvar_id,&
                                              WC_colvar_id,&
                                              colvar_p_type
   USE constraint_fxd,                  ONLY: fix_atom_control
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE free_energy_types,               ONLY: fe_env_create,&
                                              free_energy_type
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: langevin_ensemble,&
                                              npe_f_ensemble,&
                                              npe_i_ensemble,&
                                              nph_uniaxial_damped_ensemble,&
                                              nph_uniaxial_ensemble,&
                                              npt_f_ensemble,&
                                              npt_i_ensemble,&
                                              reftraj_ensemble
   USE input_cp2k_check,                ONLY: remove_restart_info
   USE input_cp2k_restarts,             ONLY: write_restart
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_remove_values,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_walltime
   USE mc_environment_types,            ONLY: get_mc_env,&
                                              mc_env_create,&
                                              mc_env_release,&
                                              mc_environment_type,&
                                              set_mc_env
   USE mc_misc,                         ONLY: mc_averages_create,&
                                              mc_averages_release
   USE mc_move_control,                 ONLY: init_mc_moves,&
                                              mc_moves_release
   USE mc_types,                        ONLY: get_mc_par,&
                                              mc_averages_type,&
                                              mc_ekin_type,&
                                              mc_moves_type,&
                                              mc_simpar_type,&
                                              set_mc_par
   USE md_ener_types,                   ONLY: create_md_ener,&
                                              md_ener_type,&
                                              release_md_ener
   USE md_energies,                     ONLY: initialize_md_ener,&
                                              md_energy
   USE md_environment_types,            ONLY: get_md_env,&
                                              md_env_create,&
                                              md_env_release,&
                                              md_environment_type,&
                                              set_md_env
   USE md_run,                          ONLY: qs_mol_dyn
   USE metadynamics_types,              ONLY: meta_env_type,&
                                              metavar_type,&
                                              set_meta_env
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE parallel_rng_types,              ONLY: UNIFORM,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              next_random_number,&
                                              rng_stream_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: boltzmann,&
                                              femtoseconds,&
                                              joule,&
                                              kelvin
   USE qmmm_util,                       ONLY: apply_qmmm_walls_reflective
   USE qs_environment_types,            ONLY: get_qs_env
   USE qs_scf_post_gpw,                 ONLY: scf_post_calculation_gpw
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE reference_manager,               ONLY: cite_reference
   USE reftraj_types,                   ONLY: create_reftraj,&
                                              reftraj_type,&
                                              release_reftraj
   USE reftraj_util,                    ONLY: initialize_reftraj
   USE simpar_methods,                  ONLY: read_md_section
   USE simpar_types,                    ONLY: create_simpar_type,&
                                              release_simpar_type,&
                                              simpar_type
   USE string_utilities,                ONLY: str_comp
   USE thermal_region_types,            ONLY: release_thermal_regions,&
                                              thermal_regions_type
   USE thermal_region_utils,            ONLY: create_thermal_regions
   USE thermostat_methods,              ONLY: create_thermostats
   USE thermostat_types,                ONLY: release_thermostats,&
                                              thermostats_type
   USE virial_methods,                  ONLY: virial_evaluate
   USE virial_types,                    ONLY: virial_type
   USE wiener_process,                  ONLY: create_wiener_process,&
                                              create_wiener_process_cv
!!!!! monte carlo part
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tamc_run'

   PUBLIC :: qs_tamc

CONTAINS

! **************************************************************************************************
!> \brief Driver routine for TAHMC
!> \param force_env ...
!> \param globenv ...
!> \param averages ...
!> \author Alin M Elena
!> \note it computes the forces using QuickStep.
! **************************************************************************************************
   SUBROUTINE qs_tamc(force_env, globenv, averages)

      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(average_quantities_type), OPTIONAL, POINTER   :: averages

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_tamc', routineP = moduleN//':'//routineN

      CHARACTER(LEN=20)                                  :: ensemble
      INTEGER                                            :: handle, i, initialStep, iprint, isos, &
                                                            istep, j, md_stride, nmccycles, &
                                                            output_unit, rand2skip, run_type_id
      INTEGER, POINTER                                   :: itimes
      LOGICAL                                            :: check, explicit, my_rm_restart_info, &
                                                            save_mem, should_stop
      REAL(KIND=dp)                                      :: auxRandom, inittime, rval, temp, &
                                                            time_iter_start, time_iter_stop
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: An, fz, xieta, zbuff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: r
      REAL(KIND=dp), POINTER                             :: constant, time, used_time
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(barostat_type), POINTER                       :: barostat
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys, subsys_i
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(free_energy_type), POINTER                    :: fe_env
      TYPE(mc_averages_type), POINTER                    :: MCaverages
      TYPE(mc_environment_type), POINTER                 :: mc_env
      TYPE(mc_moves_type), POINTER                       :: gmoves, moves
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(md_ener_type), POINTER                        :: md_ener
      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(meta_env_type), POINTER                       :: meta_env_saved
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(reftraj_type), POINTER                        :: reftraj
      TYPE(rng_stream_type), POINTER                     :: rng_stream, rng_stream_mc
      TYPE(section_vals_type), POINTER :: constraint_section, force_env_section, &
         free_energy_section, fs_section, global_section, mc_section, md_section, motion_section, &
         reftraj_section, subsys_section, work_section
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(thermal_regions_type), POINTER                :: thermal_regions
      TYPE(thermostats_type), POINTER                    :: thermostats
      TYPE(virial_type), POINTER                         :: virial

      initialStep = 0
      inittime = 0.0_dp

      CALL timeset(routineN, handle)
      my_rm_restart_info = .TRUE.
      NULLIFY (md_env, para_env, fs_section, virial)
      para_env => force_env%para_env
      motion_section => section_vals_get_subs_vals(force_env%root_section, "MOTION")
      md_section => section_vals_get_subs_vals(motion_section, "MD")

      ! Real call to MD driver - Low Level
      CALL md_env_create(md_env, md_section, para_env, force_env=force_env)
      IF (PRESENT(averages)) CALL set_md_env(md_env, averages=averages)

      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("motion/mc/tamc_run.F",217)
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/mc/tamc_run.F",218)

      NULLIFY (particles, cell, simpar, itimes, used_time, subsys, &
               md_ener, thermostats, barostat, reftraj, force_env_section, &
               reftraj_section, work_section, atomic_kinds, &
               local_particles, time, fe_env, free_energy_section, &
               constraint_section, thermal_regions, subsys_i)
      logger => cp_get_default_logger()
      para_env => force_env%para_env

      global_section => section_vals_get_subs_vals(force_env%root_section, "GLOBAL")
      free_energy_section => section_vals_get_subs_vals(motion_section, "FREE_ENERGY")
      constraint_section => section_vals_get_subs_vals(motion_section, "CONSTRAINT")
      CALL section_vals_val_get(global_section, "SAVE_MEM", l_val=save_mem)

      CALL section_vals_val_get(global_section, "RUN_TYPE", i_val=run_type_id)

      CALL create_simpar_type(simpar)
      force_env_section => force_env%force_env_section
      subsys_section => section_vals_get_subs_vals(force_env_section, "SUBSYS")
      CALL cp_add_iter_level(logger%iter_info, "MD")
      CALL cp_iterate(logger%iter_info, iter_nr=initialStep)
      ! Read MD section
      CALL read_md_section(simpar, motion_section, md_section)
      ! Setup print_keys
      simpar%info_constraint = cp_print_key_unit_nr(logger, constraint_section, &
                                                    "CONSTRAINT_INFO", extension=".shakeLog", log_filename=.FALSE.)
      simpar%lagrange_multipliers = cp_print_key_unit_nr(logger, constraint_section, &
                                                         "LAGRANGE_MULTIPLIERS", extension=".LagrangeMultLog", log_filename=.FALSE.)
      simpar%dump_lm = BTEST(cp_print_key_should_output(logger%iter_info, constraint_section, &
                                                        "LAGRANGE_MULTIPLIERS"), cp_p_file)

      ! Create the structure for the md energies
      CALL create_md_ener(md_ener)
      CALL set_md_env(md_env, md_ener=md_ener)
      CALL release_md_ener(md_ener)

      ! If requested setup Thermostats
      CALL create_thermostats(thermostats, md_section, force_env, simpar, para_env, &
                              globenv, global_section)

      ! If requested setup Barostat
      CALL create_barostat_type(barostat, md_section, force_env, simpar, globenv)

      ! If requested setup different thermal regions
      CALL create_thermal_regions(thermal_regions, md_section, simpar, force_env)

      CALL set_md_env(md_env, thermostats=thermostats, barostat=barostat, thermal_regions=thermal_regions)

      IF (simpar%ensemble == reftraj_ensemble) THEN
         reftraj_section => section_vals_get_subs_vals(md_section, "REFTRAJ")
         CALL create_reftraj(reftraj, reftraj_section, para_env)
         CALL set_md_env(md_env, reftraj=reftraj)
         CALL release_reftraj(reftraj)
      END IF

      CALL force_env_get(force_env, subsys=subsys, cell=cell, &
                         force_env_section=force_env_section)

      ! Set V0 if needed
      IF (simpar%ensemble == nph_uniaxial_ensemble .OR. simpar%ensemble == nph_uniaxial_damped_ensemble) THEN
         IF (simpar%v0 == 0._dp) simpar%v0 = cell%deth
      ENDIF

      ! Initialize velocities possibly applying constraints at the zeroth MD step
! ! !     CALL section_vals_val_get(motion_section,"PRINT%RESTART%SPLIT_RESTART_FILE",&
! ! !                               l_val=write_binary_restart_file)
!! let us see if this created all the trouble
!      CALL setup_velocities(force_env,simpar,globenv,md_env,md_section,constraint_section, &
!                            write_binary_restart_file)

      ! Setup Free Energy Calculation (if required)
      CALL fe_env_create(fe_env, free_energy_section)
      CALL set_md_env(md_env=md_env, simpar=simpar, fe_env=fe_env, cell=cell, &
                      force_env=force_env)

      ! Possibly initialize Wiener processes
      IF (simpar%ensemble == langevin_ensemble) CALL create_wiener_process(md_env)
      time_iter_start = m_walltime()

      CALL get_md_env(md_env, force_env=force_env, itimes=itimes, constant=constant, &
                      md_ener=md_ener, t=time, used_time=used_time)

      ! Attach the time counter of the meta_env to the one of the MD
      CALL set_meta_env(force_env%meta_env, time=time)
      ! Initialize the md_ener structure

      force_env%meta_env%dt = force_env%meta_env%zdt
      CALL initialize_md_ener(md_ener, force_env, simpar)
!     force_env%meta_env%dt=force_env%meta_env%zdt

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! MC setup up
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      NULLIFY (mc_env, mc_par, rng_stream_mc, MCaverages)

      CALL section_vals_get(force_env_section, n_repetition=isos)
      IF(.NOT.(isos == 1))CALL cp__a("motion/mc/tamc_run.F",315)
! set some values...will use get_globenv if that ever comes around

! initialize the random numbers
!       IF (para_env%ionode) THEN
      CALL create_rng_stream(rng_stream=rng_stream_mc, &
                             name="Random numbers for monte carlo acc/rej", &
                             distribution_type=UNIFORM)
!       ENDIF
!!!!! this shoudl go in a routine hmc_read

      NULLIFY (mc_section)
      ALLOCATE (mc_par)

      mc_section => section_vals_get_subs_vals(force_env%root_section, &
                                               "MOTION%MC")
      CALL section_vals_val_get(mc_section, "ENSEMBLE", &
                                c_val=ensemble)
      IF(.NOT.(str_comp(ensemble, "TRADITIONAL")))CALL cp__a("motion/mc/tamc_run.F",333)
      CALL section_vals_val_get(mc_section, "NSTEP", &
                                i_val=nmccycles)
      IF(.NOT.(nmccycles > 0))CALL cp__a("motion/mc/tamc_run.F",336)
      CALL section_vals_val_get(mc_section, "IPRINT", &
                                i_val=iprint)
      CALL section_vals_val_get(mc_section, "RANDOMTOSKIP", i_val=rand2skip)
      IF(.NOT.(rand2skip >= 0))CALL cp__a("motion/mc/tamc_run.F",340)
      temp = cp_unit_from_cp2k(simpar%temp_ext, "K")
!

      CALL set_mc_par(mc_par, ensemble=ensemble, nstep=nmccycles, iprint=iprint, temperature=temp, &
                      beta=1.0_dp/temp/boltzmann*joule, exp_max_val=0.9_dp*LOG(HUGE(0.0_dp)), &
                      exp_min_val=0.9_dp*LOG(TINY(0.0_dp)), max_val=HUGE(0.0_dp), min_val=0.0_dp, &
                      source=para_env%source, group=para_env%group, ionode=para_env%ionode, rand2skip=rand2skip)

      output_unit = cp_logger_get_default_io_unit(logger)
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(a,a)') "HMC| Hybrid Monte Carlo Scheme "
         WRITE (output_unit, '(a,a)') "HMC| Ensemble ", ADJUSTL(ensemble)
         WRITE (output_unit, '(a,i0)') "HMC| MC Cycles ", nmccycles
         WRITE (output_unit, '(a,i0,a)') "HMC| Print every ", iprint, " cycles"
         WRITE (output_unit, '(a,i0)') "HMC| Number of random numbers to skip ", rand2skip
         WRITE (output_unit, '(a,f16.8,a)') "HMC| Temperature ", temp, "K"
      ENDIF

      CALL force_env_get(force_env, subsys=subsys)

      CALL cp_subsys_get(subsys, atomic_kinds=atomic_kinds, local_particles=local_particles, &
                         particles=particles, virial=virial)

      DO i = 1, rand2skip
         auxRandom = next_random_number(rng_stream_mc)
         DO j = 1, 3*SIZE(particles%els)
            auxRandom = next_random_number(globenv%gaussian_rng_stream)
         ENDDO
      ENDDO

      CALL mc_env_create(mc_env)
      CALL set_mc_env(mc_env, mc_par=mc_par, force_env=force_env)
!!!!!!!end mc setup

      ! Check for ensembles requiring the stress tensor - takes into account the possibility for
      ! multiple force_evals
      IF ((simpar%ensemble == npt_i_ensemble) .OR. &
          (simpar%ensemble == npt_f_ensemble) .OR. &
          (simpar%ensemble == npe_f_ensemble) .OR. &
          (simpar%ensemble == npe_i_ensemble) .OR. &
          (simpar%ensemble == nph_uniaxial_ensemble) .OR. &
          (simpar%ensemble == nph_uniaxial_damped_ensemble)) THEN
         check = virial%pv_availability
         IF (.NOT. check) &
            CALL cp_abort(cp__l("motion/mc/tamc_run.F",385), &
                          "Virial evaluation not requested for this run in the input file! "// &
                          "You may consider to switch on the virial evaluation with the keyword: STRESS_TENSOR."// &
                          "Be sure the method you are using can compute the virial!")
         IF (ASSOCIATED(force_env%sub_force_env)) THEN
            DO i = 1, SIZE(force_env%sub_force_env)
               IF (ASSOCIATED(force_env%sub_force_env(i)%force_env)) THEN
                  CALL force_env_get(force_env%sub_force_env(i)%force_env, subsys=subsys_i)
                  CALL cp_subsys_get(subsys_i, virial=virial)
                  check = check .AND. virial%pv_availability
               END IF
            END DO
         END IF
         IF (.NOT. check) &
            CALL cp_abort(cp__l("motion/mc/tamc_run.F",399), &
                          "Virial evaluation not requested for all the force_eval sections present in"// &
                          " the input file! You have to switch on the virial evaluation with the keyword: STRESS_TENSOR "// &
                          " in each force_eval section. Be sure the method you are using can compute the virial!")
      END IF

      ! Computing Forces at zero MD step
      IF (simpar%ensemble /= reftraj_ensemble) THEN
         CALL section_vals_val_get(md_section, "STEP_START_VAL", i_val=itimes)
         CALL section_vals_val_get(md_section, "TIME_START_VAL", r_val=time)
         CALL section_vals_val_get(md_section, "ECONS_START_VAL", r_val=constant)
         CALL section_vals_val_set(md_section, "STEP_START_VAL", i_val=initialStep)
         CALL section_vals_val_set(md_section, "TIME_START_VAL", r_val=inittime)
         initialStep = itimes
         CALL cp_iterate(logger%iter_info, iter_nr=itimes)
         IF (save_mem) THEN
            work_section => section_vals_get_subs_vals(subsys_section, "VELOCITY")
            CALL section_vals_remove_values(work_section)
            work_section => section_vals_get_subs_vals(subsys_section, "SHELL_VELOCITY")
            CALL section_vals_remove_values(work_section)
            work_section => section_vals_get_subs_vals(subsys_section, "CORE_VELOCITY")
            CALL section_vals_remove_values(work_section)
         END IF

!      CALL force_env_calc_energy_force (force_env, calc_force=.TRUE.)
         meta_env_saved => force_env%meta_env
         NULLIFY (force_env%meta_env)
         CALL force_env_calc_energy_force(force_env, calc_force=.FALSE.)
         force_env%meta_env => meta_env_saved

         IF (ASSOCIATED(force_env%qs_env)) THEN
!           force_env%qs_env%sim_time=time
!           force_env%qs_env%sim_step=itimes
            force_env%qs_env%sim_time = 0.0_dp
            force_env%qs_env%sim_step = 0
         END IF
         ! Warm-up engines for metadynamics
         IF (ASSOCIATED(force_env%meta_env)) THEN
            IF (force_env%meta_env%langevin) THEN
               CALL create_wiener_process_cv(force_env%meta_env)
               NULLIFY (rng_stream)
               DO j = 1, (rand2skip-1)/nmccycles
                  DO i = 1, force_env%meta_env%n_colvar
                     rng_stream => force_env%meta_env%rng(i)%stream
                     auxRandom = next_random_number(rng_stream)
                     auxRandom = next_random_number(rng_stream)
                  ENDDO
               ENDDO
            ENDIF
!           IF (force_env%meta_env%well_tempered) THEN
!              force_env%meta_env%wttemperature = simpar%temp_ext
!              IF (force_env%meta_env%wtgamma>EPSILON(1._dp)) THEN
!                 dummy=force_env%meta_env%wttemperature*(force_env%meta_env%wtgamma-1._dp)
!                 IF (force_env%meta_env%delta_t>EPSILON(1._dp)) THEN
!                    check=ABS(force_env%meta_env%delta_t-dummy)<1.E+3_dp*EPSILON(1._dp)
!                    IF(.NOT.check) CALL cp_abort(cp__l("motion/mc/tamc_run.F",454),&
!                       "Inconsistency between DELTA_T and WTGAMMA (both specified):"//&
!                       " please, verify that DELTA_T=(WTGAMMA-1)*TEMPERATURE")
!                 ELSE
!                    force_env%meta_env%delta_t = dummy
!                 ENDIF
!              ELSE
!                 force_env%meta_env%wtgamma    = 1._dp &
!                    + force_env%meta_env%delta_t/force_env%meta_env%wttemperature
!              ENDIF
!              force_env%meta_env%invdt         = 1._dp/force_env%meta_env%delta_t
!           ENDIF
            CALL tamc_force(force_env)
!           CALL metadyn_write_colvar(force_env)
         ENDIF

         IF (simpar%do_respa) THEN
            CALL force_env_calc_energy_force(force_env%sub_force_env(1)%force_env, &
                                             calc_force=.TRUE.)
         END IF

!        CALL force_env_get( force_env, subsys=subsys)
!
!        CALL cp_subsys_get(subsys,atomic_kinds=atomic_kinds,local_particles=local_particles,&
!             particles=particles)

         CALL virial_evaluate(atomic_kinds%els, particles%els, local_particles, &
                              virial, force_env%para_env%group)

         CALL md_energy(md_env, md_ener)
!        CALL md_write_output(md_env) !inits the print env at itimes == 0 also writes trajectories
         md_stride = 1
      ELSE
         CALL get_md_env(md_env, reftraj=reftraj)
         CALL initialize_reftraj(reftraj, reftraj_section, md_env)
         itimes = reftraj%info%first_snapshot-1
         md_stride = reftraj%info%stride
      END IF

      CALL cp_print_key_finished_output(simpar%info_constraint, logger, &
                                        constraint_section, "CONSTRAINT_INFO")
      CALL cp_print_key_finished_output(simpar%lagrange_multipliers, logger, &
                                        constraint_section, "LAGRANGE_MULTIPLIERS")
      CALL init_mc_moves(moves)
      CALL init_mc_moves(gmoves)
      ALLOCATE (r(1:3, SIZE(particles%els)))
!       ALLOCATE (r_old(1:3,size(particles%els)))
      CALL mc_averages_create(MCaverages)
      !!!!! some more buffers
      ! Allocate random number for Langevin Thermostat acting on COLVARS
      ALLOCATE (xieta(2*force_env%meta_env%n_colvar))
      xieta(:) = 0.0_dp
      ALLOCATE (An(force_env%meta_env%n_colvar))
      An(:) = 0.0_dp
      ALLOCATE (fz(force_env%meta_env%n_colvar))
      fz(:) = 0.0_dp
      ALLOCATE (zbuff(2*force_env%meta_env%n_colvar))
      zbuff(:) = 0.0_dp

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(a)') "HMC|==== Initial average forces"
      ENDIF
      CALL metadyn_write_colvar_header(force_env)
      moves%hmc%attempts = 0
      moves%hmc%successes = 0
      gmoves%hmc%attempts = 0
      gmoves%hmc%successes = 0
      IF (initialStep == 0) THEN
!!! if we come from a restart we shall properly compute the average force
!!!      read the average force up to now
         DO i = 1, force_env%meta_env%n_colvar
            fs_section => section_vals_get_subs_vals(force_env%meta_env%metadyn_section, "EXT_LAGRANGE_FS")
            CALL section_vals_get(fs_section, explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(fs_section, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=i, r_val=rval)
               fz(i) = rval*rand2skip
            END IF
         ENDDO

         CALL HMCsampler(globenv, force_env, MCaverages, r, mc_par, moves, gmoves, rng_stream_mc, output_unit, &
                         fz, zbuff, nskip=rand2skip)
         CALL cp_iterate(logger%iter_info, last=.FALSE., iter_nr=0)
         CALL section_vals_val_set(mc_section, "RANDOMTOSKIP", i_val=rand2skip+nmccycles)
         CALL write_restart(md_env=md_env, root_section=force_env%root_section)
      ENDIF
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(a)') "HMC|==== end initial average forces"
      ENDIF
!     call set_md_env(md_env, init=.FALSE.)

      CALL metadyn_write_colvar(force_env)

      DO istep = 1, force_env%meta_env%TAMCSteps
         ! Increase counters
         itimes = itimes+1
         time = time+force_env%meta_env%dt
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(a)') "HMC|==================================="
            WRITE (output_unit, '(a,1x,i0)') "HMC| on z step ", istep
         ENDIF
         !needed when electric field fields are applied
         IF (ASSOCIATED(force_env%qs_env)) THEN
            force_env%qs_env%sim_time = time
            force_env%qs_env%sim_step = itimes
            force_env%meta_env%time = force_env%qs_env%sim_time
         END IF

         CALL cp_iterate(logger%iter_info, last=(istep == force_env%meta_env%TAMCSteps), iter_nr=itimes)
         ! Open possible Shake output units
         simpar%info_constraint = cp_print_key_unit_nr(logger, constraint_section, "CONSTRAINT_INFO", &
                                                       extension=".shakeLog", log_filename=.FALSE.)
         simpar%lagrange_multipliers = cp_print_key_unit_nr( &
                                       logger, constraint_section, &
                                       "LAGRANGE_MULTIPLIERS", extension=".LagrangeMultLog", log_filename=.FALSE.)
         simpar%dump_lm = BTEST(cp_print_key_should_output(logger%iter_info, constraint_section, &
                                                           "LAGRANGE_MULTIPLIERS"), cp_p_file)

         ! Velocity Verlet Integrator

         moves%hmc%attempts = 0
         moves%hmc%successes = 0
         CALL langevinVEC(md_env, globenv, mc_env, moves, gmoves, r, &
                          rng_stream_mc, xieta, An, fz, MCaverages, zbuff)

         ! Close Shake output if requested...
         CALL cp_print_key_finished_output(simpar%info_constraint, logger, &
                                           constraint_section, "CONSTRAINT_INFO")
         CALL cp_print_key_finished_output(simpar%lagrange_multipliers, logger, &
                                           constraint_section, "LAGRANGE_MULTIPLIERS")
         CALL cp_iterate(logger%iter_info, iter_nr=initialStep)
         CALL metadyn_write_colvar(force_env)
         ! Free Energy calculation
!        CALL free_energy_evaluate(md_env,should_stop,free_energy_section)

         ![AME:UB] IF (should_stop) EXIT

         ! Test for <PROJECT_NAME>.EXIT_MD or for WALL_TIME to exit
         ! Default:
         ! IF so we don't overwrite the restart or append to the trajectory
         ! because the execution could in principle stop inside the SCF where energy
         ! and forces are not converged.
         ! But:
         ! You can force to print the last step (for example if the method used
         ! to compute energy and forces is not SCF based) activating the print_key
         ! MOTION%MD%PRINT%FORCE_LAST.
         CALL external_control(should_stop, "MD", globenv=globenv)
         IF (should_stop) THEN
            CALL cp_iterate(logger%iter_info, last=.TRUE., iter_nr=itimes)
!           CALL md_output(md_env,md_section,force_env%root_section,should_stop)
            EXIT
         END IF

!        IF(simpar%ensemble /= reftraj_ensemble) THEN
!           CALL md_energy(md_env, md_ener)
!           CALL temperature_control(simpar, md_env, md_ener, force_env, logger)
!           CALL comvel_control(md_ener, force_env, md_section, logger)
!           CALL angvel_control(md_ener, force_env, md_section, logger)
!        ELSE
!           CALL md_ener_reftraj(md_env, md_ener)
!        END IF

         time_iter_stop = m_walltime()
         used_time = time_iter_stop-time_iter_start
         time_iter_start = time_iter_stop

!!!!! this writes the restart...
!         CALL md_output(md_env,md_section,force_env%root_section,should_stop)

!        IF(simpar%ensemble == reftraj_ensemble ) THEN
!           CALL write_output_reftraj(md_env)
!        END IF

         IF (output_unit > 0) THEN
            WRITE (output_unit, '(a,1x,i0)') "HMC| end z step ", istep
            WRITE (output_unit, '(a)') "HMC|==================================="
         ENDIF
      END DO
      CALL cp_iterate(logger%iter_info, last=.TRUE., iter_nr=itimes)
      force_env%qs_env%sim_time = 0.0_dp
      force_env%qs_env%sim_step = 0
      rand2skip = rand2skip+nmccycles*force_env%meta_env%TAMCSteps
      IF (initialStep == 0) rand2skip = rand2skip+nmccycles
      CALL section_vals_val_set(mc_section, "RANDOMTOSKIP", i_val=rand2skip)

      CALL write_restart(md_env=md_env, root_section=force_env%root_section)
! if we need the final kinetic energy for Hybrid Monte Carlo
!     hmc_ekin%final_ekin=md_ener%ekin

      ! Remove the iteration level
      CALL cp_rm_iter_level(logger%iter_info, "MD")

      ! Deallocate Thermostats and Barostats
      CALL release_thermostats(thermostats)
      CALL release_barostat_type(barostat)
      CALL release_simpar_type(simpar)
      CALL release_thermal_regions(thermal_regions)

      CALL md_env_release(md_env)
      ! Clean restartable sections..
      IF (my_rm_restart_info) CALL remove_restart_info(force_env%root_section)
!     IF (para_env%ionode) THEN
      CALL delete_rng_stream(rng_stream_mc)
!     ENDIF
      CALL MC_ENV_RELEASE(mc_env)
      DEALLOCATE (mc_par)
      CALL MC_MOVES_RELEASE(moves)
      CALL MC_MOVES_RELEASE(gmoves)
      DEALLOCATE (r)
!     DEALLOCATE(r_old)
      DEALLOCATE (xieta)
      DEALLOCATE (An)
      DEALLOCATE (fz)
      DEALLOCATE (zbuff)
      CALL mc_averages_release(MCaverages)
      CALL timestop(handle)

   END SUBROUTINE qs_tamc

! **************************************************************************************************
!> \brief Propagates velocities for z half a step
!> \param force_env ...
!> \param An ...
!> \author Alin M Elena
!> \note   Vanden-Eijnden Ciccotti C.Phys.Letter 429 (2006) 310-316
! **************************************************************************************************
   SUBROUTINE tamc_velocities_colvar(force_env, An)
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: An

      CHARACTER(len=*), PARAMETER :: routineN = 'tamc_velocities_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_c
      REAL(kind=dp)                                      :: dt, fft, sigma
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv

      NULLIFY (logger, meta_env, cv)
      meta_env => force_env%meta_env
      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      ! Add citation
      IF (meta_env%langevin) CALL cite_reference(VandenCic2006)
      dt = meta_env%dt

      ! Evolve Velocities
      meta_env%epot_walls = 0.0_dp
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         fft = cv%ff_s+cv%ff_hills
         sigma = SQRT((meta_env%temp_wanted*kelvin)*2.0_dp*(boltzmann/joule)*cv%gamma/cv%mass)
         cv%vvp = cv%vvp+0.5_dp*dt*(fft/cv%mass-cv%gamma*cv%vvp)*(1.0_dp-0.25_dp*dt*cv%gamma)+An(i_c)
         meta_env%epot_walls = meta_env%epot_walls+cv%epot_walls
      ENDDO
      CALL timestop(handle)
   END SUBROUTINE tamc_velocities_colvar

! **************************************************************************************************
!> \brief propagates z one step
!> \param force_env ...
!> \param xieta ...
!> \author Alin M Elena
!> \note  Vanden-Eijnden Ciccotti C.Phys.Letter 429 (2006) 310-316
! **************************************************************************************************
   SUBROUTINE tamc_position_colvar(force_env, xieta)
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: xieta

      CHARACTER(len=*), PARAMETER :: routineN = 'tamc_position_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_c
      REAL(kind=dp)                                      :: dt, sigma
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv

      NULLIFY (logger, meta_env, cv)
      meta_env => force_env%meta_env
!     IF (.NOT.ASSOCIATED(meta_env)) RETURN

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      ! Add citation
      IF (meta_env%langevin) CALL cite_reference(VandenCic2006)
      dt = meta_env%dt

      ! Update of ss0
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         sigma = SQRT((meta_env%temp_wanted*kelvin)*2.0_dp*(boltzmann/joule)*cv%gamma/cv%mass)
!        cv%ss0 =cv%ss0 +dt*cv%vvp
         cv%ss0 = cv%ss0+dt*cv%vvp+dt*SQRT(dt/12.0_dp)*sigma*xieta(i_c+meta_env%n_colvar)
         IF (cv%periodic) THEN
            ! A periodic COLVAR is always within [-pi,pi]
            cv%ss0 = SIGN(1.0_dp, ASIN(SIN(cv%ss0)))*ACOS(COS(cv%ss0))
         END IF
      ENDDO
      CALL timestop(handle)

   END SUBROUTINE tamc_position_colvar

! **************************************************************************************************
!> \brief Computes forces on z
!> #details also can be used to get the potenzial evergy of z
!> \param force_env ...
!> \param zpot ...
!> \author Alin M Elena
! **************************************************************************************************
   SUBROUTINE tamc_force(force_env, zpot)
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), INTENT(inout), OPTIONAL             :: zpot

      CHARACTER(len=*), PARAMETER :: routineN = 'tamc_force', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, i_c, icolvar, ii
      LOGICAL                                            :: explicit
      REAL(kind=dp)                                      :: diff_ss, dt, rval
      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: colvar_p
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(section_vals_type), POINTER                   :: ss0_section, ss_section, vvp_section

      NULLIFY (logger, meta_env)
      meta_env => force_env%meta_env
!     IF (.NOT.ASSOCIATED(meta_env)) RETURN

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      NULLIFY (colvar_p, subsys, cv, ss0_section, vvp_section, ss_section)
      CALL force_env_get(force_env, subsys=subsys)

      dt = meta_env%dt
      IF (.NOT. meta_env%restart) meta_env%n_steps = meta_env%n_steps+1
      ! compute ss and the derivative of ss with respect to the atomic positions
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         icolvar = cv%icolvar
         CALL colvar_eval_glob_f(icolvar, force_env)
         cv%ss = subsys%colvar_p(icolvar)%colvar%ss
         ! Restart for Extended Lagrangian Metadynamics
         IF (meta_env%restart) THEN
            ! Initialize the position of the collective variable in the extended lagrange
            ss0_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_SS0")
            CALL section_vals_get(ss0_section, explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(ss0_section, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=i_c, r_val=rval)
               cv%ss0 = rval
            ELSE
               cv%ss0 = cv%ss
            END IF
            vvp_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_VVP")
            CALL section_vals_get(vvp_section, explicit=explicit)
            IF (explicit) THEN
               CALL setup_velocities_z(force_env)
               CALL section_vals_val_get(vvp_section, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=i_c, r_val=rval)
               cv%vvp = rval
            ELSE
               CALL setup_velocities_z(force_env)
            ENDIF
            ss_section => section_vals_get_subs_vals(meta_env%metadyn_section, "EXT_LAGRANGE_SS")
            CALL section_vals_get(ss_section, explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(ss_section, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=i_c, r_val=rval)
               cv%ss = rval
            END IF
         END IF
         !
      ENDDO
      ! forces on the atoms
      NULLIFY (particles)
      CALL cp_subsys_get(subsys, colvar_p=colvar_p, &
                         particles=particles)

      meta_env%restart = .FALSE.
      meta_env%epot_s = 0.0_dp
      meta_env%epot_walls = 0.0_dp
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         diff_ss = cv%ss-cv%ss0
         IF (cv%periodic) THEN
            ! The difference of a periodic COLVAR is always within [-pi,pi]
            diff_ss = SIGN(1.0_dp, ASIN(SIN(diff_ss)))*ACOS(COS(diff_ss))
         END IF
         cv%epot_s = 0.5_dp*cv%lambda*diff_ss*diff_ss
         cv%ff_s = cv%lambda*(diff_ss)
         meta_env%epot_s = meta_env%epot_s+cv%epot_s
         icolvar = cv%icolvar

         DO ii = 1, colvar_p(icolvar)%colvar%n_atom_s
            i = colvar_p(icolvar)%colvar%i_atom(ii)
            particles%els(i)%f = particles%els(i)%f-cv%ff_s*colvar_p(icolvar)%colvar%dsdr(:, ii)
         ENDDO

      ENDDO
      IF (PRESENT(zpot)) zpot = meta_env%epot_s
      CALL fix_atom_control(force_env)

      CALL timestop(handle)
   END SUBROUTINE tamc_force

! **************************************************************************************************
!> \brief propagates one time step both z systems and samples the x system
!> \param md_env ...
!> \param globenv ...
!> \param mc_env ...
!> \param moves ...
!> \param gmoves ...
!> \param r ...
!> \param rng_stream_mc ...
!> \param xieta ...
!> \param An ...
!> \param fz ...
!> \param averages ...
!> \param zbuff ...
!> \author Alin M Elena
! **************************************************************************************************
   SUBROUTINE langevinVEC(md_env, globenv, mc_env, moves, gmoves, r, &
                          rng_stream_mc, xieta, An, fz, averages, zbuff)

      TYPE(md_environment_type), POINTER                 :: md_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(mc_environment_type), POINTER                 :: mc_env
      TYPE(mc_moves_type), POINTER                       :: moves, gmoves
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: r
      TYPE(rng_stream_type), POINTER                     :: rng_stream_mc
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: xieta, An, fz
      TYPE(mc_averages_type), INTENT(INOUT), POINTER     :: averages
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: zbuff

      CHARACTER(len=*), PARAMETER :: routineN = 'langevinVEC', routineP = moduleN//':'//routineN

      INTEGER                                            :: iprint, ivar, nparticle, nparticle_kind, &
                                                            nstep, output_unit
      REAL(KIND=dp)                                      :: dt, gamma, mass, sigma
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules, local_particles
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      TYPE(simpar_type), POINTER                         :: simpar
      TYPE(virial_type), POINTER                         :: virial

      NULLIFY (logger, mc_par)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      NULLIFY (rng_stream)
! quantitites to be nullified for the get_md_env
      NULLIFY (simpar, force_env, para_env)
! quantities to be nullified for the force_env_get environment
      NULLIFY (subsys, cell)
! quantitites to be nullified for the cp_subsys_get
      NULLIFY (atomic_kinds, local_particles, particles, local_molecules, molecules, molecule_kinds, gci)

      CALL get_md_env(md_env=md_env, simpar=simpar, force_env=force_env, &
                      para_env=para_env)
      CALL get_mc_env(mc_env, mc_par=mc_par)
      CALL get_mc_par(mc_par, nstep=nstep, iprint=iprint)

      dt = simpar%dt
      CALL force_env_get(force_env=force_env, subsys=subsys, cell=cell)

!!!! this bit should vanish once I understand what the hell is with it

!     ! Do some checks on coordinates and box
      CALL apply_qmmm_walls_reflective(force_env)

      CALL cp_subsys_get(subsys=subsys, atomic_kinds=atomic_kinds, local_particles=local_particles, &
                         particles=particles, local_molecules_new=local_molecules, molecules_new=molecules, &
                         molecule_kinds_new=molecule_kinds, gci=gci, virial=virial)

      nparticle_kind = atomic_kinds%n_els
      atomic_kind_set => atomic_kinds%els
      molecule_kind_set => molecule_kinds%els

      nparticle = particles%n_els
      particle_set => particles%els
      molecule_set => molecules%els
      IF(.NOT.(ASSOCIATED(force_env%meta_env)))CALL cp__a("motion/mc/tamc_run.F",954)
      IF(.NOT.(force_env%meta_env%langevin))CALL cp__a("motion/mc/tamc_run.F",955)
      !    *** Velocity Verlet for Langevin *** v(t)--> v(t+1/2)
      !!!!!! noise xi is in the first half, eta in the second half
      DO ivar = 1, force_env%meta_env%n_colvar
         rng_stream => force_env%meta_env%rng(ivar)%stream
         xieta(ivar) = next_random_number(rng_stream)
         xieta(ivar+force_env%meta_env%n_colvar) = next_random_number(rng_stream)
         gamma = force_env%meta_env%metavar(ivar)%gamma
         mass = force_env%meta_env%metavar(ivar)%mass
         sigma = SQRT((force_env%meta_env%temp_wanted*kelvin)*2.0_dp*(boltzmann/joule)*gamma/mass)
         An(ivar) = 0.5_dp*SQRT(dt)*sigma*(xieta(ivar)*(1.0_dp-0.5_dp*dt*gamma)- &
                                           dt*gamma*xieta(ivar+force_env%meta_env%n_colvar)/SQRT(12.0_dp))
      ENDDO
!    *** Velocity Verlet for Langeving *** v(t)--> v(t+1/2)
      CALL tamc_velocities_colvar(force_env, An)
!    *** Velocity Verlet for Langevin S(t)->S(t+1)
      CALL tamc_position_colvar(force_env, xieta)
!!!!! start zHMC sampler
      CALL HMCsampler(globenv, force_env, averages, r, mc_par, moves, gmoves, rng_stream_mc, output_unit, fz, zbuff)

!     CALL final_mc_write(mc_par,tmp_moves,&
!                output_unit,energy_check,&
!                initial_energy,final_energy,&
!                averages)

!!!!!!!!!!!!!!!!!!!! end zHMC sampler
      !    *** Velocity Verlet for Langeving *** v(t+1/2)--> v(t+1)
      CALL tamc_velocities_colvar(force_env, An)
!       CALL virial_evaluate ( atomic_kind_set, particle_set,  &
!          local_particles, virial, para_env%group)

   END SUBROUTINE langevinVEC

! **************************************************************************************************
!> \brief Driver routin for the canonical sampler using modified HMC
!> \param globenv ...
!> \param force_env ...
!> \param averages ...
!> \param r ...
!> \param mc_par ...
!> \param moves ...
!> \param gmoves ...
!> \param rng_stream_mc ...
!> \param output_unit ...
!> \param fz ...
!> \param zbuff ...
!> \param nskip ...
!> \author Alin M Elena
!> \note at the end of this routine %ff_s will contain mean force
! **************************************************************************************************

   SUBROUTINE HMCsampler(globenv, force_env, averages, r, mc_par, moves, gmoves, rng_stream_mc, output_unit, &
                         fz, zbuff, nskip)
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(mc_averages_type), POINTER                    :: averages
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: r
      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(mc_moves_type), POINTER                       :: moves, gmoves
      TYPE(rng_stream_type), POINTER                     :: rng_stream_mc
      INTEGER, INTENT(IN)                                :: output_unit
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: fz, zbuff
      INTEGER, INTENT(IN), OPTIONAL                      :: nskip

      INTEGER                                            :: i, iprint, ishift, it1, j, nsamples, &
                                                            nstep
      REAL(KIND=dp)                                      :: energy_check, old_epx, old_epz, t1
      TYPE(meta_env_type), POINTER                       :: meta_env_saved

      IF (PRESENT(nskip)) THEN
         nsamples = nskip
         ishift = nskip
      ELSE
         nsamples = 0
         fz = 0.0_dp
         ishift = 0
      ENDIF
!      lrestart = .false.
!      if (present(logger) .and. present(iter)) THEN
!       lrestart=.true.
!      ENDIF
      CALL get_mc_par(mc_par, nstep=nstep, iprint=iprint)
      meta_env_saved => force_env%meta_env
      NULLIFY (force_env%meta_env)
      CALL force_env_get(force_env, potential_energy=old_epx)
      force_env%meta_env => meta_env_saved

      old_epz = force_env%meta_env%epot_s
!!! average energy will be wrong on restarts
      averages%ave_energy = 0.0_dp
      t1 = force_env%qs_env%sim_time
      it1 = force_env%qs_env%sim_step
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(a,l4)') "HMC|restart? ", force_env%meta_env%restart
         WRITE (output_unit, '(a,3(f16.8,1x))') &
            "HMC|Ep, Epx, Epz ", old_epx+force_env%meta_env%epot_s, old_epx, force_env%meta_env%epot_s
         WRITE (output_unit, '(a)') "#HMC| No | z.. | theta.. | ff_z... | ff_z/n |"
      ENDIF
      DO i = 1, nstep
         IF (MOD(i, iprint) == 0 .AND. (output_unit > 0)) THEN
            WRITE (output_unit, '(a,1x,i0)') "HMC|========== On Monte Carlo cycle ", i+ishift
            WRITE (output_unit, '(a)') "HMC| Attempting a minitrajectory move"
            WRITE (output_unit, '(a,1x,i0)') "HMC| start mini-trajectory", i+ishift
            WRITE (output_unit, '(a,1x,i0,1x)', advance="no") "#HMC|0 ", i+ishift
            DO j = 1, force_env%meta_env%n_colvar
               WRITE (output_unit, '(f16.8,1x,f16.8,1x,f16.8)', advance="no") force_env%meta_env%metavar(j)%ss0, &
                  force_env%meta_env%metavar(j)%ss, &
                  force_env%meta_env%metavar(j)%ff_s !,fz(j)/real(i+ishift,dp)
            ENDDO
            WRITE (output_unit, *)
         ENDIF

         CALL mc_hmc_move(mc_par, force_env, globenv, moves, gmoves, old_epx, old_epz, energy_check, &
                          r, output_unit, rng_stream_mc, zbuff)
         ! check averages...
         ! force average for z needed too...
         averages%ave_energy = averages%ave_energy*REAL(i-1, dp)/REAL(i, dp)+ &
                               old_epx/REAL(i, dp)
         DO j = 1, force_env%meta_env%n_colvar
            fz(j) = fz(j)+force_env%meta_env%metavar(j)%ff_s
         ENDDO
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(a,1x,i0)') "HMC|end mini-trajectory", i+ishift
!!!!!!!! this prints z and theta(x) --ss0,ss-- needed to determine an acceptable k then
            !  the instanteneous force and some instanteneous average for force
            WRITE (output_unit, '(a,1x,i0,1x)', advance="no") "#HMC|1 ", i+ishift
            DO j = 1, force_env%meta_env%n_colvar
               WRITE (output_unit, '(f16.8,1x,f16.8,1x,f16.8,1x,f16.8)', advance="no") force_env%meta_env%metavar(j)%ss0, &
                  force_env%meta_env%metavar(j)%ss, &
                  force_env%meta_env%metavar(j)%ff_s, fz(j)/REAL(i+ishift, dp)
            ENDDO
            WRITE (output_unit, *)
         ENDIF
         nsamples = nsamples+1
         IF (MOD(i, iprint) == 0 .AND. (output_unit > 0)) THEN
            WRITE (output_unit, '(a,f16.8)') "HMC| Running average for potential energy ", averages%ave_energy
            WRITE (output_unit, '(a,1x,i0)') "HMC|======== End Monte Carlo cycle ", i+ishift
         ENDIF
!         IF (lrestart) THEN
!           k=nstep/5
!           IF(MOD(i,k) == 0) THEN
!              force_env%qs_env%sim_time=t1
!              force_env%qs_env%sim_step=it1
!              DO j=1,force_env%meta_env%n_colvar
!                force_env%meta_env%metavar(j)%ff_s=fz(j)/real(i+ishift,dp)
!              ENDDO
! !              CALL cp_iterate(logger%iter_info,last=.FALSE.,iter_nr=-1)
!              CALL section_vals_val_set(mcsec,"RANDOMTOSKIP",i_val=i+ishift)
!              CALL write_restart(md_env=mdenv,root_section=force_env%root_section)
! !              CALL cp_iterate(logger%iter_info,last=.FALSE.,iter_nr=iter)
!           ENDIF
!         ENDIF
      ENDDO
      force_env%qs_env%sim_time = t1
      force_env%qs_env%sim_step = it1
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(a,i0,a,i0,a,f16.8)') "HMC| local acceptance ratio: ", moves%hmc%successes, "/", &
            moves%hmc%attempts, "=", REAL(moves%hmc%successes, dp)/REAL(moves%hmc%attempts, dp)
         WRITE (output_unit, '(a,i0,a,i0,a,f16.8)') "HMC| global acceptance ratio: ", gmoves%hmc%successes, "/", &
            gmoves%hmc%attempts, "=", REAL(gmoves%hmc%successes, dp)/REAL(gmoves%hmc%attempts, dp)
      ENDIF
      !average force
      DO j = 1, force_env%meta_env%n_colvar
         force_env%meta_env%metavar(j)%ff_s = fz(j)/nsamples
      ENDDO
   END SUBROUTINE HMCsampler

! **************************************************************************************************
!> \brief performs a hybrid Monte Carlo move
!> \param mc_par ...
!> \param force_env ...
!> \param globenv ...
!> \param moves ...
!> \param gmoves ...
!> \param old_epx ...
!> \param old_epz ...
!> \param energy_check ...
!> \param r ...
!> \param output_unit ...
!> \param rng_stream ...
!> \param zbuff ...
!> \author Alin M Elena
!> \note It runs a NVE trajectory, followed by localisation and accepts rejects
!> using the biased Hamiltonian, rather than the traditional guiding Hamiltonian
! **************************************************************************************************
   SUBROUTINE mc_hmc_move(mc_par, force_env, globenv, moves, gmoves, old_epx, old_epz, &
                          energy_check, r, output_unit, rng_stream, zbuff)

      TYPE(mc_simpar_type), POINTER                      :: mc_par
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(mc_moves_type), POINTER                       :: moves, gmoves
      REAL(KIND=dp), INTENT(INOUT)                       :: old_epx, old_epz, energy_check
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: r
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(rng_stream_type), POINTER                     :: rng_stream
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: zbuff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mc_hmc_move', routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, iatom, j, nAtoms, source
      LOGICAL                                            :: ionode, localise
      REAL(KIND=dp)                                      :: BETA, energy_term, exp_max_val, &
                                                            exp_min_val, new_energy, new_epx, &
                                                            new_epz, rand, value, w
      TYPE(cp_subsys_type), POINTER                      :: oldsys
      TYPE(mc_ekin_type), POINTER                        :: hmc_ekin
      TYPE(meta_env_type), POINTER                       :: meta_env_saved
      TYPE(particle_list_type), POINTER                  :: particles_set
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(section_vals_type), POINTER                   :: dft_section, input

! begin the timing of the subroutine

      CALL timeset(routineN, handle)

      NULLIFY (scf_env)
      CALL get_qs_env(force_env%qs_env, scf_env=scf_env, input=input)
      dft_section => section_vals_get_subs_vals(input, "DFT")

! get a bunch of stuff from mc_par
      CALL get_mc_par(mc_par, ionode=ionode, &
                      BETA=BETA, exp_max_val=exp_max_val, &
                      exp_min_val=exp_min_val, source=source, group=group)

! nullify some pointers
!       NULLIFY(particles_set,oldsys,hmc_ekin)
      NULLIFY (particles_set, oldsys, meta_env_saved, hmc_ekin)
      ! now let's grab the particle positions
      CALL force_env_get(force_env, subsys=oldsys)
      CALL cp_subsys_get(oldsys, particles=particles_set)
      nAtoms = SIZE(particles_set%els)
! do some allocation

      ALLOCATE (hmc_ekin)

! record the attempt
      moves%hmc%attempts = moves%hmc%attempts+1
      gmoves%hmc%attempts = gmoves%hmc%attempts+1

! save the old coordinates just in case we need to go back
      DO iatom = 1, nAtoms
         r(1:3, iatom) = particles_set%els(iatom)%r(1:3)
      ENDDO
      localise = .TRUE.
! the same for collective variables data should be made,ss first half and ff_s the last half
      DO j = 1, force_env%meta_env%n_colvar
         zbuff(j) = force_env%meta_env%metavar(j)%ss
         zbuff(j+force_env%meta_env%n_colvar) = force_env%meta_env%metavar(j)%ff_s
         IF ((oldsys%colvar_p(force_env%meta_env%metavar(j)%icolvar)%colvar%type_id == HBP_colvar_id) .OR. &
             (oldsys%colvar_p(force_env%meta_env%metavar(j)%icolvar)%colvar%type_id == WC_colvar_id)) THEN
            localise = .FALSE.
         ENDIF
      ENDDO

! now run the MD simulation
      meta_env_saved => force_env%meta_env
      NULLIFY (force_env%meta_env)
      force_env%qs_env%sim_time = 0.0_dp
      force_env%qs_env%sim_step = 0
      IF (.NOT. localise) THEN
         CALL section_vals_val_set(dft_section, "LOCALIZE%_SECTION_PARAMETERS_", l_val=.FALSE.)
      ENDIF
      CALL qs_mol_dyn(force_env, globenv, hmc_e_initial=hmc_ekin%initial_ekin, hmc_e_final=hmc_ekin%final_ekin)
      IF (.NOT. localise) THEN
         CALL section_vals_val_set(dft_section, "LOCALIZE%_SECTION_PARAMETERS_", l_val=.TRUE.)
         CALL scf_post_calculation_gpw(dft_section, scf_env, force_env%qs_env)
      ENDIF

      CALL force_env_get(force_env, potential_energy=new_epx)

      force_env%meta_env => meta_env_saved
      CALL tamc_force(force_env, zpot=new_epz)
      new_energy = new_epx+new_epz
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(a,4(f16.8,1x))') &
            "HMC|old Ep, Ekx, Epz, Epx ", old_epx+old_epz, hmc_ekin%initial_ekin, old_epz, old_epx
         WRITE (output_unit, '(a,4(f16.8,1x))') "HMC|new Ep, Ekx, Epz, Epx ", new_energy, hmc_ekin%final_ekin, new_epz, new_epx
      ENDIF
      energy_term = new_energy-old_epx-old_epz+hmc_ekin%final_ekin-hmc_ekin%initial_ekin

      value = -BETA*(energy_term)
! to prevent overflows
      IF (value > exp_max_val) THEN
         w = 10.0_dp
      ELSEIF (value < exp_min_val) THEN
         w = 0.0_dp
      ELSE
         w = EXP(value)
      ENDIF

      rand = next_random_number(rng_stream)
      IF (rand < w) THEN
! accept the move
         moves%hmc%successes = moves%hmc%successes+1
         gmoves%hmc%successes = gmoves%hmc%successes+1
! update energies
         energy_check = energy_check+(new_energy-old_epx-old_epz)
         old_epx = new_epx
         old_epz = new_epz
      ELSE
! reset the cell and particle positions
         DO iatom = 1, nAtoms
            particles_set%els(iatom)%r(1:3) = r(1:3, iatom)
         ENDDO
         DO j = 1, force_env%meta_env%n_colvar
            force_env%meta_env%metavar(j)%ss = zbuff(j)
            force_env%meta_env%metavar(j)%ff_s = zbuff(j+force_env%meta_env%n_colvar)
         ENDDO

      ENDIF

      DEALLOCATE (hmc_ekin)

! end the timing
      CALL timestop(handle)

   END SUBROUTINE mc_hmc_move

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
! **************************************************************************************************
   SUBROUTINE metadyn_write_colvar_header(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_write_colvar_header', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=100)                                 :: aux, fmt
      CHARACTER(len=255)                                 :: label1, label2, label3, label4, label5, &
                                                            label6
      INTEGER                                            :: handle, i, iw, m
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(meta_env_type), POINTER                       :: meta_env

      NULLIFY (logger, meta_env)
      meta_env => force_env%meta_env
      IF (.NOT. ASSOCIATED(meta_env)) RETURN

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                "PRINT%COLVAR", extension=".metadynLog")
      IF (iw > 0) THEN
         label1 = ""
         label2 = ""
         label3 = ""
         label4 = ""
         label5 = ""
         label6 = ""
         DO i = 1, meta_env%n_colvar
            WRITE (aux, '(a,i0)') "z_", i
            label1 = TRIM(label1)//TRIM(aux)
            m = 15*i-LEN_TRIM(label1)-1
            label1 = TRIM(label1)//REPEAT(" ", m)//"|"
            WRITE (aux, '(a,i0)') "Theta_", i
            label2 = TRIM(label2)//TRIM(aux)
            m = 15*i-LEN_TRIM(label2)-1
            label2 = TRIM(label2)//REPEAT(" ", m)//"|"
            WRITE (aux, '(a,i0)') "F_z", i
            label3 = TRIM(label3)//TRIM(aux)
            m = 15*i-LEN_TRIM(label3)-1
            label3 = TRIM(label3)//REPEAT(" ", m)//"|"
            WRITE (aux, '(a,i0)') "F_h", i
            label4 = TRIM(label4)//TRIM(aux)
            m = 15*i-LEN_TRIM(label4)-1
            label4 = TRIM(label4)//REPEAT(" ", m)//"|"
            WRITE (aux, '(a,i0)') "F_w", i
            label5 = TRIM(label5)//TRIM(aux)
            m = 15*i-LEN_TRIM(label5)-1
            label5 = TRIM(label5)//REPEAT(" ", m)//"|"
            WRITE (aux, '(a,i0)') "v_z", i
            label6 = TRIM(label6)//TRIM(aux)
            m = 15*i-LEN_TRIM(label6)-1
            label6 = TRIM(label6)//REPEAT(" ", m)//"|"
         ENDDO
         WRITE (fmt, '("(a17,6a",i0 ,",4a15)")') meta_env%n_colvar*15
         WRITE (iw, TRIM(fmt)) "#Time[fs] |", &
            TRIM(label1), &
            TRIM(label2), &
            TRIM(label3), &
            TRIM(label4), &
            TRIM(label5), &
            TRIM(label6), &
            "Epot_z |", &
            "Ene hills |", &
            "Epot walls |", &
            "Temperature |"

      END IF
      CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                        "PRINT%COLVAR")

      CALL timestop(handle)

   END SUBROUTINE metadyn_write_colvar_header

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
! **************************************************************************************************
   SUBROUTINE metadyn_write_colvar(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(len=*), PARAMETER :: routineN = 'metadyn_write_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, i_c, iw
      REAL(KIND=dp)                                      :: temp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv

      NULLIFY (logger, meta_env, cv)
      meta_env => force_env%meta_env
      IF (.NOT. ASSOCIATED(meta_env)) RETURN

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      IF (meta_env%langevin) THEN
         meta_env%ekin_s = 0.0_dp
!        meta_env%epot_s = 0.0_dp
         DO i_c = 1, meta_env%n_colvar
            cv => meta_env%metavar(i_c)
            meta_env%ekin_s = meta_env%ekin_s+0.5_dp*cv%mass*cv%vvp**2
         ENDDO
      END IF

      ! write COLVAR file
      iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                "PRINT%COLVAR", extension=".metadynLog")
      IF (iw > 0) THEN
         IF (meta_env%extended_lagrange) THEN
            WRITE (iw, '(f16.8,70f15.8)') meta_env%time*femtoseconds, &
               (meta_env%metavar(i)%ss0, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ss, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_s, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_hills, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_walls, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%vvp, i=1, meta_env%n_colvar), &
               meta_env%epot_s, &
               meta_env%hills_env%energy, &
               meta_env%epot_walls, &
               (meta_env%ekin_s)*2.0_dp/(REAL(meta_env%n_colvar, KIND=dp))*kelvin
         ELSE
            WRITE (iw, '(f16.8,40f13.5)') meta_env%time*femtoseconds, &
               (meta_env%metavar(i)%ss0, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_hills, i=1, meta_env%n_colvar), &
               (meta_env%metavar(i)%ff_walls, i=1, meta_env%n_colvar), &
               meta_env%hills_env%energy, &
               meta_env%epot_walls
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                        "PRINT%COLVAR")
      ! Temperature for COLVAR
      IF (meta_env%extended_lagrange) THEN
         temp = meta_env%ekin_s*2.0_dp/(REAL(meta_env%n_colvar, KIND=dp))*kelvin
         meta_env%avg_temp = (meta_env%avg_temp*REAL(meta_env%n_steps, KIND=dp)+ &
                              temp)/REAL(meta_env%n_steps+1, KIND=dp)
         iw = cp_print_key_unit_nr(logger, meta_env%metadyn_section, &
                                   "PRINT%TEMPERATURE_COLVAR", extension=".metadynLog")
         IF (iw > 0) THEN
            WRITE (iw, '(T2,79("-"))')
            WRITE (iw, '( A,T51,f10.2,T71,f10.2)') ' COLVARS INSTANTANEOUS/AVERAGE TEMPERATURE ', &
               temp, meta_env%avg_temp
            WRITE (iw, '(T2,79("-"))')
         ENDIF
         CALL cp_print_key_finished_output(iw, logger, meta_env%metadyn_section, &
                                           "PRINT%TEMPERATURE_COLVAR")
      END IF
      CALL timestop(handle)

   END SUBROUTINE metadyn_write_colvar

! **************************************************************************************************
!> \brief ...
!> \param force_env ...
! **************************************************************************************************
   SUBROUTINE setup_velocities_z(force_env)
      TYPE(force_env_type), POINTER                      :: force_env

      INTEGER                                            :: i_c
      REAL(kind=dp)                                      :: ekin_w, fac_t
      TYPE(meta_env_type), POINTER                       :: meta_env
      TYPE(metavar_type), POINTER                        :: cv

      NULLIFY (meta_env)
      meta_env => force_env%meta_env
      meta_env%ekin_s = 0.0_dp
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         cv%vvp = next_random_number(force_env%globenv%gaussian_rng_stream)
         meta_env%ekin_s = meta_env%ekin_s+0.5_dp*cv%mass*cv%vvp**2
      END DO
      ekin_w = 0.5_dp*meta_env%temp_wanted*REAL(meta_env%n_colvar, KIND=dp)
      fac_t = SQRT(ekin_w/MAX(meta_env%ekin_s, 1.0E-8_dp))
      DO i_c = 1, meta_env%n_colvar
         cv => meta_env%metavar(i_c)
         cv%vvp = cv%vvp*fac_t
      ENDDO
   END SUBROUTINE setup_velocities_z
END MODULE tamc_run
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief performs CELL optimization
!> \par History
!>      03.2008 - Teodoro Laino [tlaino] - University of Zurich - Cell Optimization
! **************************************************************************************************
MODULE cell_opt
   USE bfgs_optimizer,                  ONLY: geoopt_bfgs
   USE cg_optimizer,                    ONLY: geoopt_cg
   USE cp_lbfgs_geo,                    ONLY: geoopt_lbfgs
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_rm_iter_level
   USE force_env_types,                 ONLY: force_env_type
   USE global_types,                    ONLY: global_environment_type
   USE gopt_f_methods,                  ONLY: gopt_f_create_x0
   USE gopt_f_types,                    ONLY: gopt_f_create,&
                                              gopt_f_release,&
                                              gopt_f_type
   USE gopt_param_types,                ONLY: gopt_param_read,&
                                              gopt_param_release,&
                                              gopt_param_type
   USE input_constants,                 ONLY: default_bfgs_method_id,&
                                              default_cell_method_id,&
                                              default_cg_method_id,&
                                              default_lbfgs_method_id
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cell_opt'

   PUBLIC :: cp_cell_opt

CONTAINS

! **************************************************************************************************
!> \brief Main driver to perform geometry optimization
!> \param force_env ...
!> \param globenv ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE cp_cell_opt(force_env, globenv)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cell_opt', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, step_start_val
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(section_vals_type), POINTER                   :: force_env_section, geo_section, &
                                                            root_section

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/cell_opt.F",70)
      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("motion/cell_opt.F",71)
      NULLIFY (gopt_param, force_env_section, gopt_env, x0)
      root_section => force_env%root_section
      force_env_section => force_env%force_env_section
      geo_section => section_vals_get_subs_vals(root_section, "MOTION%CELL_OPT")

      CALL gopt_param_read(gopt_param, geo_section, type_id=default_cell_method_id)
      CALL gopt_f_create(gopt_env, gopt_param, force_env=force_env, globenv=globenv, &
                         geo_opt_section=geo_section)
      CALL gopt_f_create_x0(gopt_env, x0)

      CALL section_vals_val_get(geo_section, "STEP_START_VAL", i_val=step_start_val)
      CALL cp_add_iter_level(logger%iter_info, "CELL_OPT")
      CALL cp_iterate(logger%iter_info, iter_nr=step_start_val)
      CALL cp_cell_opt_low(force_env, globenv, gopt_param, gopt_env, &
                           force_env_section, geo_section, x0)
      CALL cp_rm_iter_level(logger%iter_info, "CELL_OPT")

      ! Reset counter for next iteration
      CALL section_vals_val_set(geo_section, "STEP_START_VAL", i_val=0)
      DEALLOCATE (x0)
      CALL gopt_f_release(gopt_env)
      CALL gopt_param_release(gopt_param)
      CALL timestop(handle)

   END SUBROUTINE cp_cell_opt

! **************************************************************************************************
!> \brief call to low level geometry optimizers
!> \param force_env ...
!> \param globenv ...
!> \param gopt_param ...
!> \param gopt_env ...
!> \param force_env_section ...
!> \param geo_section ...
!> \param x0 ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 03.2008
! **************************************************************************************************
   SUBROUTINE cp_cell_opt_low(force_env, globenv, gopt_param, gopt_env, force_env_section, &
                              geo_section, x0)
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(gopt_param_type), POINTER                     :: gopt_param
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, geo_section
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x0

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cell_opt_low', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(force_env)))CALL cp__a("motion/cell_opt.F",121)
      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("motion/cell_opt.F",122)
      IF(.NOT.(ASSOCIATED(gopt_param)))CALL cp__a("motion/cell_opt.F",123)
      IF(.NOT.(ASSOCIATED(gopt_env)))CALL cp__a("motion/cell_opt.F",124)
      IF(.NOT.(ASSOCIATED(x0)))CALL cp__a("motion/cell_opt.F",125)
      IF(.NOT.(ASSOCIATED(force_env_section)))CALL cp__a("motion/cell_opt.F",126)
      IF(.NOT.(ASSOCIATED(geo_section)))CALL cp__a("motion/cell_opt.F",127)

      SELECT CASE (gopt_param%method_id)
      CASE (default_bfgs_method_id)
         CALL geoopt_bfgs(force_env, gopt_param, globenv, &
                          geo_section, gopt_env, x0)
      CASE (default_lbfgs_method_id)
         CALL geoopt_lbfgs(force_env, gopt_param, globenv, &
                           geo_section, gopt_env, x0)
      CASE (default_cg_method_id)
         CALL geoopt_cg(force_env, gopt_param, globenv, &
                        geo_section, gopt_env, x0)
      CASE DEFAULT
         CALL cp__b("motion/cell_opt.F",140,"")
      END SELECT

   END SUBROUTINE cp_cell_opt_low

END MODULE cell_opt
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module with utility to perform MD Nudged Elastic Band Calculation
!> \note
!>      Numerical accuracy for parallel runs:
!>       Each replica starts the SCF run from the one optimized
!>       in a previous run. It may happen then energies and derivatives
!>       of a serial run and a parallel run could be slightly different
!>       'cause of a different starting density matrix.
!>       Exact results are obtained using:
!>          EXTRAPOLATION USE_GUESS in QS section (Teo 09.2006)
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
MODULE neb_md_utils
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: band_md_opt,&
                                              do_band_collective
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE neb_types,                       ONLY: neb_type,&
                                              neb_var_type
   USE parallel_rng_types,              ONLY: next_random_number
   USE particle_types,                  ONLY: get_particle_pos_or_vel,&
                                              particle_type,&
                                              update_particle_pos_or_vel
   USE physcon,                         ONLY: kelvin,&
                                              massunit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'neb_md_utils'

   PUBLIC :: neb_initialize_velocity, &
             control_vels_a, &
             control_vels_b, &
             get_temperatures

CONTAINS

! **************************************************************************************************
!> \brief Initialize velocities of replica in an MD optimized algorithm within
!>        NEB
!> \param vels ...
!> \param neb_section ...
!> \param particle_set ...
!> \param i_rep ...
!> \param iw ...
!> \param globenv ...
!> \param neb_env ...
!> \par History
!>      25.11.2010 Consider core-shell model (MK)
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE neb_initialize_velocity(vels, neb_section, particle_set, i_rep, iw, &
                                      globenv, neb_env)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: vels
      TYPE(section_vals_type), POINTER                   :: neb_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: i_rep, iw
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(neb_type), POINTER                            :: neb_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'neb_initialize_velocity', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, ivar, k, natom, nparticle, nvar
      REAL(KIND=dp)                                      :: akin, mass, mass_tot, sc, temp, &
                                                            temp_ext, tmp_r1
      REAL(KIND=dp), DIMENSION(3)                        :: v, vcom
      TYPE(section_vals_type), POINTER                   :: md_section

      IF (neb_env%opt_type == band_md_opt) THEN
         md_section => section_vals_get_subs_vals(neb_section, "OPTIMIZE_BAND%MD")
         CALL section_vals_val_get(md_section, "TEMPERATURE", r_val=temp_ext)
         ! Initialize velocity according to external temperature
         nparticle = SIZE(vels, 1)
         natom = SIZE(particle_set)
         mass_tot = 0.0_dp
         vcom(1:3) = 0.0_dp
         vels(:, i_rep) = 0.0_dp
         DO k = 1, nparticle
            vels(k, i_rep) = next_random_number(globenv%gaussian_rng_stream)
         END DO
         ! Check always if BAND is working in Cartesian or in internal coordinates
         ! If working in cartesian coordinates let's get rid of the COM
         ! Compute also the total mass (both in Cartesian and internal)
         IF (neb_env%use_colvar) THEN
            nvar = nparticle
            mass_tot = REAL(nvar, KIND=dp)*massunit
         ELSE
            DO iatom = 1, natom
               mass = particle_set(iatom)%atomic_kind%mass
               mass_tot = mass_tot+mass
               v(1:3) = get_particle_pos_or_vel(iatom, particle_set, vels(:, i_rep))
               vcom(1:3) = vcom(1:3)+mass*v(1:3)
            END DO
            vcom(1:3) = vcom(1:3)/mass_tot
         END IF
         ! Compute kinetic energy and temperature
         akin = 0.0_dp
         IF (neb_env%use_colvar) THEN
            DO ivar = 1, nvar
               akin = akin+0.5_dp*massunit*vels(ivar, i_rep)*vels(ivar, i_rep)
            END DO
         ELSE
            DO iatom = 1, natom
               mass = particle_set(iatom)%atomic_kind%mass
               v(1:3) = -vcom(1:3)
               CALL update_particle_pos_or_vel(iatom, particle_set, v(1:3), vels(:, i_rep))
               akin = akin+0.5_dp*mass*DOT_PRODUCT(v(1:3), v(1:3))
            END DO
            nvar = 3*natom
         END IF
         temp = 2.0_dp*akin/REAL(nvar, KIND=dp)
         ! Scale velocities to get the correct initial temperature and
         sc = SQRT(temp_ext/temp)
         vels(:, i_rep) = vels(:, i_rep)*sc
         ! Re-compute kinetic energya and temperature and velocity of COM
         akin = 0.0_dp
         vcom = 0.0_dp
         IF (neb_env%use_colvar) THEN
            DO ivar = 1, nvar
               akin = akin+0.5_dp*massunit*vels(ivar, i_rep)*vels(ivar, i_rep)
            END DO
         ELSE
            DO iatom = 1, natom
               mass = particle_set(iatom)%atomic_kind%mass
               v(1:3) = get_particle_pos_or_vel(iatom, particle_set, vels(:, i_rep))
               vcom(1:3) = vcom(1:3)+mass*v(1:3)
               akin = akin+0.5_dp*mass*DOT_PRODUCT(v(1:3), v(1:3))
            END DO
         END IF
         vcom(1:3) = vcom(1:3)/mass_tot
         ! Dump information
         IF (iw > 0) THEN
            temp = 2.0_dp*akin/REAL(nvar, KIND=dp)
            tmp_r1 = cp_unit_from_cp2k(temp, "K")
            WRITE (iw, '(A,T61,F18.2,A2)') &
               ' NEB| Initial Temperature ', tmp_r1, " K"
            WRITE (iw, '(A,T61,F20.12)') &
               ' NEB| Centre of mass velocity in direction x:', vcom(1), &
               ' NEB| Centre of mass velocity in direction y:', vcom(2), &
               ' NEB| Centre of mass velocity in direction z:', vcom(3)
            WRITE (iw, '(T2,"NEB|",75("*"))')
         END IF
      ELSE
         vels(:, i_rep) = 0.0_dp
      END IF

   END SUBROUTINE neb_initialize_velocity

! **************************************************************************************************
!> \brief Control on  velocities - I part
!> \param vels ...
!> \param particle_set ...
!> \param tc_section ...
!> \param vc_section ...
!> \param output_unit ...
!> \param istep ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE control_vels_a(vels, particle_set, tc_section, vc_section, &
                             output_unit, istep)
      TYPE(neb_var_type), POINTER                        :: vels
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: tc_section, vc_section
      INTEGER, INTENT(IN)                                :: output_unit, istep

      CHARACTER(len=*), PARAMETER :: routineN = 'control_vels_a', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, temp_tol_steps
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: ext_temp, f_annealing, scale, temp_tol, &
                                                            temploc, tmp_r1, tmp_r2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: temperatures

! Temperature control

      CALL section_vals_get(tc_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(tc_section, "TEMP_TOL_STEPS", i_val=temp_tol_steps)
         CALL section_vals_val_get(tc_section, "TEMPERATURE", r_val=ext_temp)
         CALL section_vals_val_get(tc_section, "TEMP_TOL", r_val=temp_tol)
         ALLOCATE (temperatures(SIZE(vels%wrk, 2)))
         ! Computes temperatures
         CALL get_temperatures(vels, particle_set, temperatures, factor=1.0_dp)
         ! Possibly rescale
         IF (istep <= temp_tol_steps) THEN
            DO i = 2, SIZE(vels%wrk, 2)-1
               temploc = temperatures(i)
               IF (ABS(temploc-ext_temp) > temp_tol) THEN
                  IF (output_unit > 0) THEN
                     tmp_r1 = cp_unit_from_cp2k(temploc, "K")
                     tmp_r2 = cp_unit_from_cp2k(ext_temp, "K")
                     WRITE (output_unit, '(T2,"NEB| Replica Nr.",I5,'// &
                            '"  - Velocity rescaled from: ",F12.6," to: ",F12.6,".")') &
                        i, tmp_r1, tmp_r2

                  END IF
                  scale = SQRT(ext_temp/temploc)
                  vels%wrk(:, i) = scale*vels%wrk(:, i)
               END IF
            END DO
         END IF
         DEALLOCATE (temperatures)
      END IF
      ! Annealing
      CALL section_vals_get(vc_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(vc_section, "ANNEALING", r_val=f_annealing)
         DO i = 2, SIZE(vels%wrk, 2)-1
            vels%wrk(:, i) = f_annealing*vels%wrk(:, i)
         END DO
      END IF
   END SUBROUTINE control_vels_a

! **************************************************************************************************
!> \brief Control on velocities - II part
!> \param vels ...
!> \param forces ...
!> \param vc_section ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE control_vels_b(vels, forces, vc_section)
      TYPE(neb_var_type), POINTER                        :: vels, forces
      TYPE(section_vals_type), POINTER                   :: vc_section

      CHARACTER(len=*), PARAMETER :: routineN = 'control_vels_b', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: explicit, lval
      REAL(KIND=dp)                                      :: factor, norm

! Check the sign of V.dot.F

      CALL section_vals_get(vc_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(vc_section, "PROJ_VELOCITY_VERLET", l_val=lval)
         IF (lval) THEN
            DO i = 2, SIZE(vels%wrk, 2)-1
               norm = DOT_PRODUCT(forces%wrk(:, i), forces%wrk(:, i))
               factor = DOT_PRODUCT(vels%wrk(:, i), forces%wrk(:, i))
               IF (factor > 0 .AND. (norm >= EPSILON(0.0_dp))) THEN
                  vels%wrk(:, i) = factor/norm*forces%wrk(:, i)
               ELSE
                  vels%wrk(:, i) = 0.0_dp
               END IF
            END DO
         END IF
         CALL section_vals_val_get(vc_section, "SD_LIKE", l_val=lval)
         IF (lval) THEN
            DO i = 2, SIZE(vels%wrk, 2)-1
               vels%wrk(:, i) = 0.0_dp
            END DO
         END IF
      END IF
   END SUBROUTINE control_vels_b

! **************************************************************************************************
!> \brief Computes temperatures
!> \param vels ...
!> \param particle_set ...
!> \param temperatures ...
!> \param ekin ...
!> \param factor ...
!> \par History
!>      24.11.2010 rewritten to include core-shell model (MK)
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE get_temperatures(vels, particle_set, temperatures, ekin, factor)

      TYPE(neb_var_type), POINTER                        :: vels
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: temperatures
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT), OPTIONAL :: ekin
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: factor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_temperatures', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_rep, iatom, ivar, n_rep, natom, nvar
      REAL(KIND=dp)                                      :: akin, mass, myfactor, temploc
      REAL(KIND=dp), DIMENSION(3)                        :: v

      myfactor = kelvin
      IF (PRESENT(factor)) myfactor = factor
      IF (PRESENT(ekin)) ekin(:) = 0.0_dp
      temperatures(:) = 0.0_dp
      nvar = SIZE(vels%wrk, 1)
      n_rep = SIZE(vels%wrk, 2)
      natom = SIZE(particle_set)
      DO i_rep = 2, n_rep-1
         akin = 0.0_dp
         IF (vels%in_use == do_band_collective) THEN
            DO ivar = 1, nvar
               akin = akin+0.5_dp*massunit*vels%wrk(ivar, i_rep)*vels%wrk(ivar, i_rep)
            END DO
         ELSE
            DO iatom = 1, natom
               mass = particle_set(iatom)%atomic_kind%mass
               v(1:3) = get_particle_pos_or_vel(iatom, particle_set, vels%wrk(:, i_rep))
               akin = akin+0.5_dp*mass*DOT_PRODUCT(v(1:3), v(1:3))
            END DO
            nvar = 3*natom
         END IF
         IF (PRESENT(ekin)) ekin(i_rep) = akin
         temploc = 2.0_dp*akin/REAL(nvar, KIND=dp)
         temperatures(i_rep) = temploc*myfactor
      END DO

   END SUBROUTINE get_temperatures

END MODULE neb_md_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief I/O Module for Nudged Elastic Band Calculation
!> \note
!>      Numerical accuracy for parallel runs:
!>       Each replica starts the SCF run from the one optimized
!>       in a previous run. It may happen then energies and derivatives
!>       of a serial run and a parallel run could be slightly different
!>       'cause of a different starting density matrix.
!>       Exact results are obtained using:
!>          EXTRAPOLATION USE_GUESS in QS section (Teo 09.2006)
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
MODULE neb_io
   USE cell_types,                      ONLY: cell_type
   USE cp2k_info,                       ONLY: get_runtime_info
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_add_default_logger,&
                                              cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_rm_default_logger,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE f77_interface,                   ONLY: f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              use_mixed_force
   USE header,                          ONLY: cp2k_footer
   USE input_constants,                 ONLY: band_md_opt,&
                                              do_sm,&
                                              dump_xmol,&
                                              pot_neb_fe,&
                                              pot_neb_full,&
                                              pot_neb_me
   USE input_cp2k_neb,                  ONLY: create_band_section
   USE input_cp2k_restarts,             ONLY: write_restart
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_flush
   USE neb_md_utils,                    ONLY: get_temperatures
   USE neb_types,                       ONLY: neb_type,&
                                              neb_var_type
   USE particle_methods,                ONLY: write_particle_coordinates
   USE particle_types,                  ONLY: get_particle_pos_or_vel,&
                                              particle_type
   USE physcon,                         ONLY: angstrom
   USE replica_types,                   ONLY: replica_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'neb_io'

   PUBLIC :: read_neb_section, &
             dump_neb_info, &
             dump_replica_coordinates, &
             handle_band_file_names, &
             neb_rep_env_map_info

CONTAINS

! **************************************************************************************************
!> \brief Read data from the NEB input section
!> \param neb_env ...
!> \param neb_section ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE read_neb_section(neb_env, neb_section)
      TYPE(neb_type), POINTER                            :: neb_env
      TYPE(section_vals_type), POINTER                   :: neb_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_neb_section', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: wrk_section

      IF(.NOT.(ASSOCIATED(neb_env)))CALL cp__a("motion/neb_io.F",99)
      neb_env%istep = 0
      CALL section_vals_val_get(neb_section, "BAND_TYPE", i_val=neb_env%id_type)
      CALL section_vals_val_get(neb_section, "NUMBER_OF_REPLICA", i_val=neb_env%number_of_replica)
      CALL section_vals_val_get(neb_section, "K_SPRING", r_val=neb_env%K)
      CALL section_vals_val_get(neb_section, "ROTATE_FRAMES", l_val=neb_env%rotate_frames)
      CALL section_vals_val_get(neb_section, "ALIGN_FRAMES", l_val=neb_env%align_frames)
      CALL section_vals_val_get(neb_section, "OPTIMIZE_BAND%OPTIMIZE_END_POINTS", l_val=neb_env%optimize_end_points)
      ! Climb Image NEB
      CALL section_vals_val_get(neb_section, "CI_NEB%NSTEPS_IT", i_val=neb_env%nsteps_it)
      ! Band Optimization Type
      CALL section_vals_val_get(neb_section, "OPTIMIZE_BAND%OPT_TYPE", i_val=neb_env%opt_type)
      ! Use colvars
      CALL section_vals_val_get(neb_section, "USE_COLVARS", l_val=neb_env%use_colvar)
      CALL section_vals_val_get(neb_section, "POT_TYPE", i_val=neb_env%pot_type)
      ! Before continuing let's do some consistency check between keywords
      IF (neb_env%pot_type /= pot_neb_full) THEN
         ! Requires the use of colvars
         IF (.NOT. neb_env%use_colvar) &
            CALL cp_abort(cp__l("motion/neb_io.F",118), &
                          "A potential energy function based on free energy or minimum energy"// &
                          " was requested without enabling the usage of COLVARS. Both methods"// &
                          " are based on COLVARS definition.")
         ! Moreover let's check if the proper sections have been defined..
         SELECT CASE (neb_env%pot_type)
         CASE (pot_neb_fe)
            wrk_section => section_vals_get_subs_vals(neb_env%root_section, "MOTION%MD")
            CALL section_vals_get(wrk_section, explicit=explicit)
            IF (.NOT. explicit) &
               CALL cp_abort(cp__l("motion/neb_io.F",128), &
                             "A free energy BAND (colvars projected) calculation is requested"// &
                             " but NONE MD section was defined in the input.")
         CASE (pot_neb_me)
            wrk_section => section_vals_get_subs_vals(neb_env%root_section, "MOTION%GEO_OPT")
            CALL section_vals_get(wrk_section, explicit=explicit)
            IF (.NOT. explicit) &
               CALL cp_abort(cp__l("motion/neb_io.F",135), &
                             "A minimum energy BAND (colvars projected) calculation is requested"// &
                             " but NONE GEO_OPT section was defined in the input.")
         END SELECT
      ELSE
         IF (neb_env%use_colvar) &
            CALL cp_abort(cp__l("motion/neb_io.F",141), &
                          "A band calculation was requested with a full potential energy. USE_COLVAR cannot"// &
                          " be set for this kind of calculation!")
      END IF
      ! String Method
      CALL section_vals_val_get(neb_section, "STRING_METHOD%SMOOTHING", r_val=neb_env%smoothing)
      CALL section_vals_val_get(neb_section, "STRING_METHOD%SPLINE_ORDER", i_val=neb_env%spline_order)
      neb_env%reparametrize_frames = .FALSE.
      IF (neb_env%id_type == do_sm) THEN
         neb_env%reparametrize_frames = .TRUE.
      END IF
   END SUBROUTINE read_neb_section

! **************************************************************************************************
!> \brief dump print info of a NEB run
!> \param neb_env ...
!> \param coords ...
!> \param vels ...
!> \param forces ...
!> \param particle_set ...
!> \param logger ...
!> \param istep ...
!> \param energies ...
!> \param distances ...
!> \param output_unit ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE dump_neb_info(neb_env, coords, vels, forces, particle_set, logger, &
                            istep, energies, distances, output_unit)
      TYPE(neb_type), POINTER                            :: neb_env
      TYPE(neb_var_type), POINTER                        :: coords
      TYPE(neb_var_type), OPTIONAL, POINTER              :: vels, forces
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cp_logger_type), POINTER                      :: logger
      INTEGER, INTENT(IN)                                :: istep
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: energies, distances
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'dump_neb_info', routineP = moduleN//':'//routineN

      CHARACTER(LEN=20)                                  :: mytype
      CHARACTER(LEN=default_string_length)               :: line, title, unit_str
      INTEGER                                            :: crd, ener, frc, handle, i, irep, ndig, &
                                                            ndigl, ttst, vel
      LOGICAL                                            :: explicit, lval
      REAL(KIND=dp)                                      :: f_ann, tmp_r1, unit_conv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ekin, temperatures
      TYPE(cell_type), POINTER                           :: cell
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: tc_section, vc_section

      CALL timeset(routineN, handle)
      ndig = CEILING(LOG10(REAL(neb_env%number_of_replica+1, KIND=dp)))
      CALL force_env_get(neb_env%force_env, cell=cell)
      DO irep = 1, neb_env%number_of_replica
         ndigl = CEILING(LOG10(REAL(irep+1, KIND=dp)))
         WRITE (line, '(A,'//cp_to_string(ndig)//'("0"),T'//cp_to_string(11+ndig+1-ndigl)//',I0)') "Replica_nr_", irep
         crd = cp_print_key_unit_nr(logger, neb_env%motion_print_section, "TRAJECTORY", &
                                    extension=".xyz", file_form="FORMATTED", middle_name="pos-"//TRIM(line))
         IF (PRESENT(vels)) THEN
            vel = cp_print_key_unit_nr(logger, neb_env%motion_print_section, "VELOCITIES", &
                                       extension=".xyz", file_form="FORMATTED", middle_name="vel-"//TRIM(line))
         END IF
         IF (PRESENT(forces)) THEN
            frc = cp_print_key_unit_nr(logger, neb_env%motion_print_section, "FORCES", &
                                       extension=".xyz", file_form="FORMATTED", middle_name="force-"//TRIM(line))
         END IF
         ! Dump Trajectory
         IF (crd > 0) THEN
            ! Gather units of measure for output
            CALL section_vals_val_get(neb_env%motion_print_section, "TRAJECTORY%UNIT", &
                                      c_val=unit_str)
            unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
            ! This information can be digested by Molden
            WRITE (UNIT=title, FMT="(A,I8,A,F20.10)") " i =", istep, ", E =", energies(irep)
            CALL write_particle_coordinates(particle_set, crd, dump_xmol, "POS", title, &
                                            cell=cell, array=coords%xyz(:, irep), unit_conv=unit_conv)
            CALL m_flush(crd)
         END IF
         ! Dump Velocities
         IF (vel > 0 .AND. PRESENT(vels)) THEN
            ! Gather units of measure for output
            CALL section_vals_val_get(neb_env%motion_print_section, "VELOCITIES%UNIT", &
                                      c_val=unit_str)
            unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
            WRITE (UNIT=title, FMT="(A,I8,A,F20.10)") " i =", istep, ", E =", energies(irep)
            CALL write_particle_coordinates(particle_set, vel, dump_xmol, "VEL", title, &
                                            cell=cell, array=vels%xyz(:, irep), unit_conv=unit_conv)
            CALL m_flush(vel)
         END IF
         ! Dump Forces
         IF (frc > 0 .AND. PRESENT(forces)) THEN
            ! Gather units of measure for output
            CALL section_vals_val_get(neb_env%motion_print_section, "FORCES%UNIT", &
                                      c_val=unit_str)
            unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
            WRITE (UNIT=title, FMT="(A,I8,A,F20.10)") " i =", istep, ", E =", energies(irep)
            CALL write_particle_coordinates(particle_set, frc, dump_xmol, "FRC", title, &
                                            cell=cell, array=forces%xyz(:, irep), unit_conv=unit_conv)
            CALL m_flush(frc)
         END IF
         CALL cp_print_key_finished_output(crd, logger, neb_env%motion_print_section, &
                                           "TRAJECTORY")
         IF (PRESENT(vels)) THEN
            CALL cp_print_key_finished_output(vel, logger, neb_env%motion_print_section, &
                                              "VELOCITIES")
         END IF
         IF (PRESENT(forces)) THEN
            CALL cp_print_key_finished_output(frc, logger, neb_env%motion_print_section, &
                                              "FORCES")
         END IF
      END DO
      ! NEB summary info on screen
      IF (output_unit > 0) THEN
         tc_section => section_vals_get_subs_vals(neb_env%neb_section, "OPTIMIZE_BAND%MD%TEMP_CONTROL")
         vc_section => section_vals_get_subs_vals(neb_env%neb_section, "OPTIMIZE_BAND%MD%VEL_CONTROL")
         ALLOCATE (temperatures(neb_env%number_of_replica))
         ALLOCATE (ekin(neb_env%number_of_replica))
         CALL get_temperatures(vels, particle_set, temperatures, ekin=ekin)
         WRITE (output_unit, '(/)', ADVANCE="NO")
         WRITE (output_unit, FMT='(A,A)') ' **************************************', &
            '*****************************************'
         NULLIFY (section, keyword, enum)
         CALL create_band_section(section)
         keyword => section_get_keyword(section, "BAND_TYPE")
         CALL keyword_get(keyword, enum=enum)
         mytype = TRIM(enum_i2c(enum, neb_env%id_type))
         WRITE (output_unit, FMT='(A,T61,A)') &
            ' BAND TYPE                     =', ADJUSTR(mytype)
         CALL section_release(section)
         WRITE (output_unit, FMT='(A,T61,A)') &
            ' BAND TYPE OPTIMIZATION        =', ADJUSTR(neb_env%opt_type_label(1:20))
         WRITE (output_unit, '( A,T71,I10 )') &
            ' STEP NUMBER                   =', istep
         IF (neb_env%rotate_frames) WRITE (output_unit, '( A,T71,L10 )') &
            ' RMSD DISTANCE DEFINITION      =', neb_env%rotate_frames
         ! velocity control parameters output
         CALL section_vals_get(vc_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(vc_section, "PROJ_VELOCITY_VERLET", l_val=lval)
            IF (lval) WRITE (output_unit, '( A,T71,L10 )') &
               ' PROJECTED VELOCITY VERLET     =', lval
            CALL section_vals_val_get(vc_section, "SD_LIKE", l_val=lval)
            IF (lval) WRITE (output_unit, '( A,T71,L10)') &
               ' STEEPEST DESCENT LIKE         =', lval
            CALL section_vals_val_get(vc_section, "ANNEALING", r_val=f_ann)
            IF (f_ann /= 1.0_dp) THEN
               WRITE (output_unit, '( A,T71,F10.5)') &
                  ' ANNEALING FACTOR              = ', f_ann
            END IF
         END IF
         ! temperature control parameters output
         CALL section_vals_get(tc_section, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(tc_section, "TEMP_TOL_STEPS", i_val=ttst)
            IF (istep <= ttst) THEN
               CALL section_vals_val_get(tc_section, "TEMPERATURE", r_val=f_ann)
               tmp_r1 = cp_unit_from_cp2k(f_ann, "K")
               WRITE (output_unit, '( A,T71,F10.5)') &
                  ' TEMPERATURE TARGET            =', tmp_r1
            END IF
         END IF
         WRITE (output_unit, '( A,T71,I10 )') &
            ' NUMBER OF NEB REPLICA         =', neb_env%number_of_replica
         WRITE (output_unit, '( A,T17,4F16.6)') &
            ' DISTANCES REP =', distances(1:MIN(4, SIZE(distances)))
         IF (SIZE(distances) > 4) THEN
            WRITE (output_unit, '( T17,4F16.6)') distances(5:SIZE(distances))
         END IF
         WRITE (output_unit, '( A,T17,4F16.6)') &
            ' ENERGIES [au] =', energies(1:MIN(4, SIZE(energies)))
         IF (SIZE(energies) > 4) THEN
            WRITE (output_unit, '( T17,4F16.6)') energies(5:SIZE(energies))
         END IF
         IF (neb_env%opt_type == band_md_opt) THEN
            WRITE (output_unit, '( A,T33,4(1X,F11.5))') &
               ' REPLICA TEMPERATURES (K)      =', temperatures(1:MIN(4, SIZE(temperatures)))
            DO i = 5, SIZE(temperatures), 4
               WRITE (output_unit, '( T33,4(1X,F11.5))') &
                  temperatures(i:MIN(i+3, SIZE(temperatures)))
            END DO
         END IF
         WRITE (output_unit, '( A,T56,F25.14)') &
            ' BAND TOTAL ENERGY [au]        =', SUM(energies(:)+ekin(:))+ &
            neb_env%spring_energy
         WRITE (output_unit, FMT='(A,A)') ' **************************************', &
            '*****************************************'
         DEALLOCATE (ekin)
         DEALLOCATE (temperatures)
      END IF
      ! Ener file
      ener = cp_print_key_unit_nr(logger, neb_env%neb_section, "ENERGY", &
                                  extension=".ener", file_form="FORMATTED")
      IF (ener > 0) THEN
         WRITE (line, '(I0)') 2*neb_env%number_of_replica-1
         WRITE (ener, '(I10,'//TRIM(line)//'(1X,F20.9))') istep, &
            energies, distances
      END IF
      CALL cp_print_key_finished_output(ener, logger, neb_env%neb_section, &
                                        "ENERGY")

      ! Dump Restarts
      CALL cp_add_default_logger(logger)
      CALL write_restart(force_env=neb_env%force_env, &
                         root_section=neb_env%root_section, &
                         coords=coords, &
                         vels=vels)
      CALL cp_rm_default_logger()

      CALL timestop(handle)

   END SUBROUTINE dump_neb_info

! **************************************************************************************************
!> \brief dump coordinates of a replica NEB
!> \param particle_set ...
!> \param coords ...
!> \param i_rep ...
!> \param ienum ...
!> \param iw ...
!> \param use_colvar ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE dump_replica_coordinates(particle_set, coords, i_rep, ienum, iw, use_colvar)

      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(neb_var_type), POINTER                        :: coords
      INTEGER, INTENT(IN)                                :: i_rep, ienum, iw
      LOGICAL, INTENT(IN)                                :: use_colvar

      INTEGER                                            :: iatom, j
      REAL(KIND=dp), DIMENSION(3)                        :: r

      IF (iw > 0) THEN
         WRITE (iw, '(/,T2,"NEB|",75("*"))')
         WRITE (iw, '(T2,"NEB|",1X,A,I0,A)') &
            "Geometry for Replica Nr. ", ienum, " in Angstrom"
         DO iatom = 1, SIZE(particle_set)
            r(1:3) = get_particle_pos_or_vel(iatom, particle_set, coords%xyz(:, i_rep))
            WRITE (iw, '(T2,"NEB|",1X,A10,5X,3F15.9)') &
               TRIM(particle_set(iatom)%atomic_kind%name), r(1:3)*angstrom
         END DO
         IF (use_colvar) THEN
            WRITE (iw, '(/,T2,"NEB|",1X,A10)') "COLLECTIVE VARIABLES:"
            WRITE (iw, '(T2,"NEB|",16X,3F15.9)') &
               (coords%int(j, i_rep), j=1, SIZE(coords%int(:, :), 1))
         END IF
         WRITE (iw, '(T2,"NEB|",75("*"))')
         CALL m_flush(iw)
      END IF

   END SUBROUTINE dump_replica_coordinates

! **************************************************************************************************
!> \brief Handles the correct file names during a band calculation
!> \param rep_env ...
!> \param irep ...
!> \param n_rep ...
!> \param istep ...
!> \author Teodoro Laino  06.2009
! **************************************************************************************************
   SUBROUTINE handle_band_file_names(rep_env, irep, n_rep, istep)
      TYPE(replica_env_type), POINTER                    :: rep_env
      INTEGER, INTENT(IN)                                :: irep, n_rep, istep

      CHARACTER(len=*), PARAMETER :: routineN = 'handle_band_file_names', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: output_file_path, replica_proj_name
      INTEGER                                            :: handle, handle2, i, ierr, j, lp, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger, sub_logger
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(section_vals_type), POINTER                   :: root_section

      CALL timeset(routineN, handle)
      CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env, &
                              handle=handle2)
      logger => cp_get_default_logger()
      CALL force_env_get(f_env%force_env, root_section=root_section)
      j = irep+(rep_env%local_rep_indices(1)-1)
      ! Get replica_project_name
      replica_proj_name = get_replica_project_name(rep_env, n_rep, j)
      lp = LEN_TRIM(replica_proj_name)
      CALL section_vals_val_set(root_section, "GLOBAL%PROJECT_NAME", &
                                c_val=TRIM(replica_proj_name))
      logger%iter_info%project_name = replica_proj_name

      ! We change the file on which is pointing the global logger and error
      output_file_path = replica_proj_name(1:lp)//".out"
      CALL section_vals_val_set(root_section, "GLOBAL%OUTPUT_FILE_NAME", &
                                c_val=TRIM(output_file_path))
      IF (logger%default_global_unit_nr > 0) THEN
         CALL close_file(logger%default_global_unit_nr)
         CALL open_file(file_name=output_file_path, file_status="UNKNOWN", &
                        file_action="WRITE", file_position="APPEND", &
                        unit_number=logger%default_global_unit_nr, &
                        skip_get_unit_number=.TRUE.)
         WRITE (UNIT=logger%default_global_unit_nr, FMT="(/,(T2,A79))") &
            "*******************************************************************************", &
            "**                 BAND EVALUATION OF ENERGIES AND FORCES                    **", &
            "*******************************************************************************"
         WRITE (UNIT=logger%default_global_unit_nr, FMT="(T2,A,T79,A)") "**", "**"
         WRITE (UNIT=logger%default_global_unit_nr, FMT="(T2,A,T79,A)") "**", "**"
         WRITE (UNIT=logger%default_global_unit_nr, FMT="(T2,A,I5,T41,A,I5,T79,A)") &
            "** Replica Env Nr. :", rep_env%local_rep_indices(1)-1, "Replica Band Nr. :", j, "**"
         WRITE (UNIT=logger%default_global_unit_nr, FMT="(T2,A,I5,T79,A)") &
            "** Band  Step  Nr. :", istep, "**"
         WRITE (UNIT=logger%default_global_unit_nr, FMT="(T2,A79)") &
            "*******************************************************************************"
      END IF

      ! Handle specific case for mixed_env
      SELECT CASE (f_env%force_env%in_use)
      CASE (use_mixed_force)
         DO i = 1, f_env%force_env%mixed_env%ngroups
            IF (MODULO(i-1, f_env%force_env%mixed_env%ngroups) == &
                f_env%force_env%mixed_env%group_distribution(f_env%force_env%mixed_env%para_env%mepos)) THEN
               sub_logger => f_env%force_env%mixed_env%sub_logger(i)%p
               sub_logger%iter_info%project_name = replica_proj_name(1:lp)//"-r-"//TRIM(ADJUSTL(cp_to_string(i)))

               unit_nr = sub_logger%default_global_unit_nr
               IF (unit_nr > 0) THEN
                  CALL close_file(unit_nr)

                  output_file_path = replica_proj_name(1:lp)//"-r-"//TRIM(ADJUSTL(cp_to_string(i)))//".out"
                  CALL open_file(file_name=output_file_path, file_status="UNKNOWN", &
                                 file_action="WRITE", file_position="APPEND", &
                                 unit_number=unit_nr, skip_get_unit_number=.TRUE.)
               END IF
            END IF
         END DO
      END SELECT

      CALL f_env_rm_defaults(f_env=f_env, ierr=ierr, handle=handle2)
      IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_io.F",477)
      CALL timestop(handle)

   END SUBROUTINE handle_band_file_names

! **************************************************************************************************
!> \brief  Constructs project names for BAND replicas
!> \param rep_env ...
!> \param n_rep ...
!> \param j ...
!> \retval replica_proj_name ...
!> \author Teodoro Laino  06.2009
! **************************************************************************************************
   FUNCTION get_replica_project_name(rep_env, n_rep, j) RESULT(replica_proj_name)
      TYPE(replica_env_type), POINTER                    :: rep_env
      INTEGER, INTENT(IN)                                :: n_rep, j
      CHARACTER(LEN=default_path_length)                 :: replica_proj_name

      CHARACTER(len=*), PARAMETER :: routineN = 'get_replica_project_name', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: padding
      INTEGER                                            :: i, lp, ndigits

! Setup new replica project name and output file

      replica_proj_name = rep_env%original_project_name
      ! Find padding
      ndigits = CEILING(LOG10(REAL(n_rep+1, KIND=dp)))- &
                CEILING(LOG10(REAL(j+1, KIND=dp)))
      padding = ""
      DO i = 1, ndigits
         padding(i:i) = "0"
      END DO
      lp = LEN_TRIM(replica_proj_name)
      replica_proj_name(lp+1:LEN(replica_proj_name)) = "-BAND"// &
                                                       TRIM(padding)//ADJUSTL(cp_to_string(j))
   END FUNCTION get_replica_project_name

! **************************************************************************************************
!> \brief  Print some mapping infos in the replica_env setup output files
!>         i.e. prints in which files one can find information for each band
!>         replica
!> \param rep_env ...
!> \param neb_env ...
!> \author Teodoro Laino  06.2009
! **************************************************************************************************
   SUBROUTINE neb_rep_env_map_info(rep_env, neb_env)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(neb_type), POINTER                            :: neb_env

      CHARACTER(len=*), PARAMETER :: routineN = 'neb_rep_env_map_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: replica_proj_name
      INTEGER                                            :: handle2, ierr, irep, n_rep, n_rep_neb, &
                                                            output_unit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(f_env_type), POINTER                          :: f_env

      n_rep_neb = neb_env%number_of_replica
      n_rep = rep_env%nrep
      CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env, &
                              handle=handle2)
      logger => cp_get_default_logger()
      output_unit = logger%default_global_unit_nr
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT='(/,(T2,A79))') &
            "*******************************************************************************", &
            "**                  MAPPING OF BAND REPLICA TO REPLICA ENV                   **", &
            "*******************************************************************************"
         WRITE (UNIT=output_unit, FMT='(T2,A,I6,T32,A,T79,A)') &
            "** Replica Env Nr.: ", rep_env%local_rep_indices(1)-1, &
            "working on the following BAND replicas", "**"
         WRITE (UNIT=output_unit, FMT='(T2,A79)') &
            "**                                                                           **"
      END IF
      DO irep = 1, n_rep_neb, n_rep
         replica_proj_name = get_replica_project_name(rep_env, n_rep_neb, irep+rep_env%local_rep_indices(1)-1)
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT='(T2,A,I6,T32,A,T79,A)') &
               "** Band Replica   Nr.: ", irep+rep_env%local_rep_indices(1)-1, &
               "Output available on file: "//TRIM(replica_proj_name)//".out", "**"
         END IF
      END DO
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT='(T2,A79)') &
            "**                                                                           **", &
            "*******************************************************************************"
         WRITE (UNIT=output_unit, FMT='(/)')
      END IF
      ! update runtime info before printing the footer
      CALL get_runtime_info()
      ! print footer
      CALL cp2k_footer(output_unit)
      CALL f_env_rm_defaults(f_env=f_env, ierr=ierr, handle=handle2)
      IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_io.F",573)
   END SUBROUTINE neb_rep_env_map_info

END MODULE neb_io
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module with utility for  Nudged Elastic Band Calculation
!> \note
!>      Numerical accuracy for parallel runs:
!>       Each replica starts the SCF run from the one optimized
!>       in a previous run. It may happen then energies and derivatives
!>       of a serial run and a parallel run could be slightly different
!>       'cause of a different starting density matrix.
!>       Exact results are obtained using:
!>          EXTRAPOLATION USE_GUESS in QS section (Teo 09.2006)
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
MODULE neb_utils
   USE bibliography,                    ONLY: E2002,&
                                              Elber1987,&
                                              Jonsson1998,&
                                              Jonsson2000_1,&
                                              Jonsson2000_2,&
                                              Wales2004,&
                                              cite_reference
   USE colvar_utils,                    ONLY: eval_colvar,&
                                              get_clv_force,&
                                              set_colvars_target
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE f77_interface,                   ONLY: f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type,&
                                              get_energy,&
                                              get_force,&
                                              get_pos,&
                                              set_pos
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get
   USE geo_opt,                         ONLY: cp_geo_opt
   USE global_types,                    ONLY: global_environment_type
   USE input_constants,                 ONLY: &
        band_diis_opt, do_b_neb, do_band_cartesian, do_ci_neb, do_d_neb, do_eb, do_it_neb, do_sm, &
        pot_neb_fe, pot_neb_full, pot_neb_me
   USE input_cp2k_check,                ONLY: remove_restart_info
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathlib,                         ONLY: matvec_3x3
   USE md_run,                          ONLY: qs_mol_dyn
   USE neb_io,                          ONLY: dump_replica_coordinates,&
                                              handle_band_file_names
   USE neb_md_utils,                    ONLY: neb_initialize_velocity
   USE neb_types,                       ONLY: neb_type,&
                                              neb_var_type
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: bohr
   USE replica_methods,                 ONLY: rep_env_calc_e_f
   USE replica_types,                   ONLY: rep_env_sync,&
                                              replica_env_type
   USE rmsd,                            ONLY: rmsd3
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'neb_utils'
   LOGICAL, PARAMETER, PRIVATE :: debug_this_module = .FALSE.

   PUBLIC :: build_replica_coords, &
             neb_calc_energy_forces, &
             reorient_images, &
             reparametrize_images, &
             check_convergence

CONTAINS

! **************************************************************************************************
!> \brief Computes the distance between two replica
!> \param particle_set ...
!> \param coords ...
!> \param i0 ...
!> \param i ...
!> \param distance ...
!> \param iw ...
!> \param rotate ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE neb_replica_distance(particle_set, coords, i0, i, distance, iw, rotate)
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(neb_var_type), POINTER                        :: coords
      INTEGER, INTENT(IN)                                :: i0, i
      REAL(KIND=dp), INTENT(OUT)                         :: distance
      INTEGER, INTENT(IN)                                :: iw
      LOGICAL, INTENT(IN), OPTIONAL                      :: rotate

      CHARACTER(len=*), PARAMETER :: routineN = 'neb_replica_distance', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_rotate

      my_rotate = .FALSE.
      IF (PRESENT(rotate)) my_rotate = rotate
      ! The rotation of the replica is enabled exclusively when working in
      ! cartesian coordinates
      IF (my_rotate .AND. (coords%in_use == do_band_cartesian)) THEN
         IF(.NOT.(PRESENT(particle_set)))CALL cp__a("motion/neb_utils.F",119)
         CALL rmsd3(particle_set, coords%xyz(:, i), coords%xyz(:, i0), &
                    iw, rotate=my_rotate)
      END IF
      distance = SQRT(DOT_PRODUCT(coords%wrk(:, i)-coords%wrk(:, i0), &
                                  coords%wrk(:, i)-coords%wrk(:, i0)))

   END SUBROUTINE neb_replica_distance

! **************************************************************************************************
!> \brief Constructs or Read the coordinates for all replica
!> \param neb_section ...
!> \param particle_set ...
!> \param coords ...
!> \param vels ...
!> \param neb_env ...
!> \param iw ...
!> \param globenv ...
!> \param para_env ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE build_replica_coords(neb_section, particle_set, &
                                   coords, vels, neb_env, iw, globenv, para_env)
      TYPE(section_vals_type), POINTER                   :: neb_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(neb_var_type), POINTER                        :: coords, vels
      TYPE(neb_type), POINTER                            :: neb_env
      INTEGER, INTENT(IN)                                :: iw
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'build_replica_coords', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: filename
      CHARACTER(LEN=default_string_length)               :: dummy_char
      INTEGER :: handle, i_rep, iatom, ic, input_nr_replica, is, ivar, j, jtarg, n_rep, natom, &
         neb_nr_replica, nr_replica_to_interpolate, nval, nvar, shell_index
      LOGICAL                                            :: check, explicit, my_end, skip_vel_section
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: distance
      REAL(KIND=dp), DIMENSION(3)                        :: r
      REAL(KIND=dp), DIMENSION(:), POINTER               :: initial_colvars, rptr
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(section_vals_type), POINTER                   :: coord_section, replica_section, &
                                                            vel_section

      NULLIFY (parser)
      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(coords)))CALL cp__a("motion/neb_utils.F",167)
      IF(.NOT.(ASSOCIATED(vels)))CALL cp__a("motion/neb_utils.F",168)
      neb_nr_replica = neb_env%number_of_replica
      replica_section => section_vals_get_subs_vals(neb_section, "REPLICA")
      CALL section_vals_get(replica_section, n_repetition=input_nr_replica)
      ! Calculation is aborted if input replicas are more then the requested ones for the BAND..
      IF(.NOT.(input_nr_replica <= neb_nr_replica))CALL cp__a("motion/neb_utils.F",173)
      ! Read in replicas coordinates
      skip_vel_section = (input_nr_replica /= neb_nr_replica)
      IF ((iw > 0) .AND. skip_vel_section) THEN
         WRITE (iw, '(T2,A)') 'NEB| The number of replica in the input is different from the number', &
            'NEB| of replica requested for NEB. More Replica will be interpolated.', &
            'NEB| Therefore the possibly provided velocities will not be read.'
      END IF
      ! Further check on velocity section...
      DO i_rep = 1, input_nr_replica
         vel_section => section_vals_get_subs_vals(replica_section, "VELOCITY", &
                                                   i_rep_section=i_rep)
         CALL section_vals_get(vel_section, explicit=explicit)
         skip_vel_section = skip_vel_section .OR. (.NOT. explicit)
      END DO
      ! Setup cartesian coordinates and COLVAR (if requested)
      coords%xyz(:, :) = 0.0_dp
      DO i_rep = 1, input_nr_replica
         coord_section => section_vals_get_subs_vals(replica_section, "COORD", &
                                                     i_rep_section=i_rep)
         CALL section_vals_get(coord_section, explicit=explicit)
         ! Cartesian Coordinates
         IF (explicit) THEN
            CALL section_vals_val_get(coord_section, "_DEFAULT_KEYWORD_", &
                                      n_rep_val=natom)
            IF(.NOT.((natom == SIZE(particle_set))))CALL cp__a("motion/neb_utils.F",198)
            DO iatom = 1, natom
               CALL section_vals_val_get(coord_section, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=iatom, r_vals=rptr)
               ic = 3*(iatom-1)
               coords%xyz(ic+1:ic+3, i_rep) = rptr(1:3)*bohr
               ! Initially core and shell positions are set to the atomic positions
               shell_index = particle_set(iatom)%shell_index
               IF (shell_index /= 0) THEN
                  is = 3*(natom+shell_index-1)
                  coords%xyz(is+1:is+3, i_rep) = coords%xyz(ic+1:ic+3, i_rep)
               END IF
            END DO
         ELSE
            CALL section_vals_val_get(replica_section, "COORD_FILE_NAME", &
                                      i_rep_section=i_rep, c_val=filename)
            IF(.NOT.(TRIM(filename) /= ""))CALL cp__a("motion/neb_utils.F",214)
            CALL parser_create(parser, filename, para_env=para_env, parse_white_lines=.TRUE.)
            CALL parser_get_next_line(parser, 1)
            ! Start parser
            CALL parser_get_object(parser, natom)
            IF(.NOT.((natom == SIZE(particle_set))))CALL cp__a("motion/neb_utils.F",219)
            CALL parser_get_next_line(parser, 1)
            DO iatom = 1, natom
               ! Atom coordinates
               CALL parser_get_next_line(parser, 1, at_end=my_end)
               IF (my_end) &
                  CALL cp_abort(cp__l("motion/neb_utils.F",225), &
                                "Number of lines in XYZ format not equal to the number of atoms."// &
                                " Error in XYZ format for REPLICA coordinates. Very probably the"// &
                                " line with title is missing or is empty. Please check the XYZ file and rerun your job!")
               READ (parser%input_line, *) dummy_char, r(1:3)
               ic = 3*(iatom-1)
               coords%xyz(ic+1:ic+3, i_rep) = r(1:3)*bohr
               ! Initially core and shell positions are set to the atomic positions
               shell_index = particle_set(iatom)%shell_index
               IF (shell_index /= 0) THEN
                  is = 3*(natom+shell_index-1)
                  coords%xyz(is+1:is+3, i_rep) = coords%xyz(ic+1:ic+3, i_rep)
               END IF
            END DO
            CALL parser_release(parser)
         END IF
         ! Collective Variables
         IF (neb_env%use_colvar) THEN
            CALL section_vals_val_get(replica_section, "COLLECTIVE", &
                                      i_rep_section=i_rep, n_rep_val=n_rep)
            IF (n_rep /= 0) THEN
               ! Read the values of the collective variables
               NULLIFY (initial_colvars)
               CALL section_vals_val_get(replica_section, "COLLECTIVE", &
                                         i_rep_section=i_rep, r_vals=initial_colvars)
               check = (neb_env%nsize_int == SIZE(initial_colvars))
               IF(.NOT.(check))CALL cp__a("motion/neb_utils.F",251)
               coords%int(:, i_rep) = initial_colvars
            ELSE
               ! Compute the values of the collective variables
               CALL eval_colvar(neb_env%force_env, coords%xyz(:, i_rep), coords%int(:, i_rep))
            END IF
         END IF
         ! Dump cartesian and colvar info..
         CALL dump_replica_coordinates(particle_set, coords, i_rep, i_rep, iw, neb_env%use_colvar)
         ! Setup Velocities
         IF (skip_vel_section) THEN
            CALL neb_initialize_velocity(vels%wrk, neb_section, particle_set, &
                                         i_rep, iw, globenv, neb_env)
         ELSE
            vel_section => section_vals_get_subs_vals(replica_section, "VELOCITY", &
                                                      i_rep_section=i_rep)
            CALL section_vals_val_get(vel_section, "_DEFAULT_KEYWORD_", &
                                      n_rep_val=nval)
            ! Setup Velocities for collective or cartesian coordinates
            IF (neb_env%use_colvar) THEN
               nvar = SIZE(vels%wrk, 1)
               IF(.NOT.(nval == nvar))CALL cp__a("motion/neb_utils.F",272)
               DO ivar = 1, nvar
                  CALL section_vals_val_get(vel_section, "_DEFAULT_KEYWORD_", &
                                            i_rep_val=ivar, r_vals=rptr)
                  vels%wrk(ivar, i_rep) = rptr(1)
               END DO
            ELSE
               natom = SIZE(particle_set)
               IF(.NOT.(nval == natom))CALL cp__a("motion/neb_utils.F",280)
               DO iatom = 1, natom
                  CALL section_vals_val_get(vel_section, "_DEFAULT_KEYWORD_", &
                                            i_rep_val=iatom, r_vals=rptr)
                  ic = 3*(iatom-1)
                  vels%wrk(ic+1:ic+3, i_rep) = rptr(1:3)
                  ! Initially set shell velocities to core velocity
                  shell_index = particle_set(iatom)%shell_index
                  IF (shell_index /= 0) THEN
                     is = 3*(natom+shell_index-1)
                     vels%wrk(is+1:is+3, i_rep) = vels%wrk(ic+1:ic+3, i_rep)
                  END IF
               END DO
            END IF
         END IF
      END DO ! i_rep
      ALLOCATE (distance(neb_nr_replica-1))
      IF (input_nr_replica < neb_nr_replica) THEN
         ! Interpolate missing replicas
         nr_replica_to_interpolate = neb_nr_replica-input_nr_replica
         distance = 0.0_dp
         IF (iw > 0) THEN
            WRITE (iw, '(T2,A,I0,A)') 'NEB| Interpolating ', nr_replica_to_interpolate, ' missing Replica.'
         END IF
         DO WHILE (nr_replica_to_interpolate > 0)
            ! Compute distances between known images to find the interval
            ! where to add a new image
            DO j = 1, input_nr_replica-1
               CALL neb_replica_distance(particle_set, coords, j, j+1, distance(j), iw, &
                                         rotate=neb_env%align_frames)
            END DO
            jtarg = MAXLOC(distance(1:input_nr_replica), 1)
            IF (iw > 0) THEN
               WRITE (iw, '(T2,3(A,I0),A)') 'NEB| Interpolating Nr. ', &
                  nr_replica_to_interpolate, ' missing Replica between Replica Nr. ', &
                  jtarg, ' and ', jtarg+1, '.'
            END IF
            input_nr_replica = input_nr_replica+1
            nr_replica_to_interpolate = nr_replica_to_interpolate-1
            ! Interpolation is a simple bisection in XYZ
            coords%xyz(:, jtarg+2:input_nr_replica) = coords%xyz(:, jtarg+1:input_nr_replica-1)
            coords%xyz(:, jtarg+1) = (coords%xyz(:, jtarg)+coords%xyz(:, jtarg+2))/2.0_dp
            IF (neb_env%use_colvar) THEN
               ! Interpolation is a simple bisection also in internal coordinates
               ! in this case the XYZ coordinates need only as a starting point for computing
               ! the potential energy function. The reference are the internal coordinates
               ! interpolated here after..
               coords%int(:, jtarg+2:input_nr_replica) = coords%int(:, jtarg+1:input_nr_replica-1)
               coords%int(:, jtarg+1) = (coords%int(:, jtarg)+coords%int(:, jtarg+2))/2.0_dp
            END IF
            vels%wrk(:, jtarg+2:input_nr_replica) = vels%wrk(:, jtarg+1:input_nr_replica-1)
            vels%wrk(:, jtarg+1) = 0.0_dp
            CALL dump_replica_coordinates(particle_set, coords, jtarg+1, &
                                          input_nr_replica, iw, neb_env%use_colvar)
            CALL neb_initialize_velocity(vels%wrk, neb_section, particle_set, &
                                         jtarg+1, iw, globenv, neb_env)
         END DO
      END IF
      vels%wrk(:, 1) = 0.0_dp
      vels%wrk(:, neb_nr_replica) = 0.0_dp
      ! If we perform a DIIS optimization we don't need velocities
      IF (neb_env%opt_type == band_diis_opt) vels%wrk = 0.0_dp
      ! Compute distances between replicas and in case of Cartesian Coordinates
      ! Rotate the frames in order to minimize the RMSD
      DO j = 1, input_nr_replica-1
         CALL neb_replica_distance(particle_set, coords, j, j+1, distance(j), iw, &
                                   rotate=neb_env%align_frames)
      END DO
      DEALLOCATE (distance)

      CALL timestop(handle)

   END SUBROUTINE build_replica_coords

! **************************************************************************************************
!> \brief Driver to compute energy and forces within a NEB,
!>      Based on the use of the replica_env
!> \param rep_env ...
!> \param neb_env ...
!> \param coords ...
!> \param energies ...
!> \param forces ...
!> \param particle_set ...
!> \param output_unit ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE neb_calc_energy_forces(rep_env, neb_env, coords, energies, forces, &
                                     particle_set, output_unit)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(neb_type), OPTIONAL, POINTER                  :: neb_env
      TYPE(neb_var_type), POINTER                        :: coords
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: energies
      TYPE(neb_var_type), POINTER                        :: forces
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'neb_calc_energy_forces', &
         routineP = moduleN//':'//routineN
      CHARACTER(LEN=1), DIMENSION(3), PARAMETER          :: lab = (/"X", "Y", "Z"/)

      INTEGER                                            :: handle, i, irep, j, n_int, n_rep, &
                                                            n_rep_neb, nsize_wrk
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tangent, tmp_a, tmp_b
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: cvalues, Mmatrix, Mmatrix_tmp

      CALL timeset(routineN, handle)
      n_int = neb_env%nsize_int
      n_rep_neb = neb_env%number_of_replica
      n_rep = rep_env%nrep
      nsize_wrk = coords%size_wrk(1)
      energies = 0.0_dp
      ALLOCATE (cvalues(n_int, n_rep))
      ALLOCATE (Mmatrix_tmp(n_int*n_int, n_rep))
      ALLOCATE (Mmatrix(n_int*n_int, n_rep_neb))
      IF (output_unit > 0) WRITE (output_unit, '(/,T2,A)') "NEB| Computing Energies and Forces"
      DO irep = 1, n_rep_neb, n_rep
         DO j = 0, n_rep-1
            IF (irep+j <= n_rep_neb) THEN
               ! If the number of replica in replica_env and the number of replica
               ! used in the NEB does not match, the other replica in replica_env
               ! just compute energies and forces keeping the fixed coordinates and
               ! forces
               rep_env%r(:, j+1) = coords%xyz(:, irep+j)
            END IF
         END DO
         ! Fix file name for BAND replicas.. Each BAND replica has its own file
         ! independently from the number of replicas in replica_env..
         CALL handle_band_file_names(rep_env, irep, n_rep_neb, neb_env%istep)
         ! Let's select the potential we want to use for the band calculation
         SELECT CASE (neb_env%pot_type)
         CASE (pot_neb_full)
            ! Full potential Energy
            CALL rep_env_calc_e_f(rep_env, calc_f=.TRUE.)
         CASE (pot_neb_fe)
            ! Free Energy Case
            CALL perform_replica_md(rep_env, coords, irep, n_rep_neb, cvalues, Mmatrix_tmp)
         CASE (pot_neb_me)
            ! Minimum Potential Energy Case
            CALL perform_replica_geo(rep_env, coords, irep, n_rep_neb, cvalues, Mmatrix_tmp)
         END SELECT

         DO j = 0, n_rep-1
            IF (irep+j <= n_rep_neb) THEN
               ! Copy back Forces and Energies
               forces%wrk(:, irep+j) = rep_env%f(1:nsize_wrk, j+1)
               energies(irep+j) = rep_env%f(rep_env%ndim+1, j+1)
               SELECT CASE (neb_env%pot_type)
               CASE (pot_neb_full)
                  ! Dump Info
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, '(T2,A,I5,A,I5,A)') &
                        "NEB| REPLICA Nr.", irep+j, "- Energy and Forces"
                     WRITE (output_unit, '(T2,A,T42,A,9X,F15.9)') &
                        "NEB|", " Total Energy: ", rep_env%f(rep_env%ndim+1, j+1)
                     WRITE (output_unit, '(T2,"NEB|",T10,"ATOM",T33,3(9X,A,7X))') lab(1), lab(2), lab(3)
                     DO i = 1, SIZE(particle_set)
                        WRITE (output_unit, '(T2,"NEB|",T12,A,T30,3(2X,F15.9))') &
                           particle_set(i)%atomic_kind%name, &
                           rep_env%f((i-1)*3+1:(i-1)*3+3, j+1)
                     END DO
                  END IF
               CASE (pot_neb_fe, pot_neb_me)
                  ! Let's update the cartesian coordinates. This will make
                  ! easier the next evaluation of energies and forces...
                  coords%xyz(:, irep+j) = rep_env%r(1:rep_env%ndim, j+1)
                  Mmatrix(:, irep+j) = Mmatrix_tmp(:, j+1)
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, '(/,T2,A,I5,A,I5,A)') &
                        "NEB| REPLICA Nr.", irep+j, "- Energy, Collective Variables,  Forces"
                     WRITE (output_unit, '(T2,A,T42,A,9X,F15.9)') &
                        "NEB|", " Total Energy: ", rep_env%f(rep_env%ndim+1, j+1)
                     WRITE (output_unit, &
                            '(T2,"NEB|",T10,"CV Nr.",12X,"Expected COLVAR",5X,"Present COLVAR",10X,"Forces")')
                     DO i = 1, n_int
                        WRITE (output_unit, '(T2,"NEB|",T12,I2,7X,3(5X,F15.9))') &
                           i, coords%int(i, irep+j), cvalues(i, j+1), rep_env%f(i, j+1)
                     END DO
                  END IF
               END SELECT
            END IF
         END DO
      END DO
      DEALLOCATE (cvalues)
      DEALLOCATE (Mmatrix_tmp)
      IF (PRESENT(neb_env)) THEN
         ! First identify the image of the chain with the higher potential energy
         ! First and last point of the band are never considered
         neb_env%nr_HE_image = MAXLOC(energies(2:n_rep_neb-1), 1)+1
         ALLOCATE (tangent(nsize_wrk))
         ! Then modify image forces accordingly to the scheme chosen for the
         ! calculation.
         neb_env%spring_energy = 0.0_dp
         IF (neb_env%optimize_end_points) THEN
            ALLOCATE (tmp_a(SIZE(forces%wrk, 1)))
            ALLOCATE (tmp_b(SIZE(forces%wrk, 1)))
            tmp_a(:) = forces%wrk(:, 1)
            tmp_b(:) = forces%wrk(:, SIZE(forces%wrk, 2))
         END IF
         DO i = 2, neb_env%number_of_replica
            CALL get_tangent(neb_env, coords, i, tangent, energies, output_unit)
            CALL get_neb_force(neb_env, tangent, coords, i, forces, Mmatrix=Mmatrix, &
                               iw=output_unit)
         END DO
         IF (neb_env%optimize_end_points) THEN
            forces%wrk(:, 1) = tmp_a ! Image A
            forces%wrk(:, SIZE(forces%wrk, 2)) = tmp_b ! Image B
            DEALLOCATE (tmp_a)
            DEALLOCATE (tmp_b)
         ELSE
            ! Nullify forces on the two end points images
            forces%wrk(:, 1) = 0.0_dp ! Image A
            forces%wrk(:, SIZE(forces%wrk, 2)) = 0.0_dp ! Image B
         END IF
         DEALLOCATE (tangent)
      END IF
      DEALLOCATE (Mmatrix)
      CALL timestop(handle)
   END SUBROUTINE neb_calc_energy_forces

! **************************************************************************************************
!> \brief Driver to perform an MD run on each single replica to
!>      compute specifically Free Energies in a NEB scheme
!> \param rep_env ...
!> \param coords ...
!> \param irep ...
!> \param n_rep_neb ...
!> \param cvalues ...
!> \param Mmatrix ...
!> \author Teodoro Laino  01.2007
! **************************************************************************************************
   SUBROUTINE perform_replica_md(rep_env, coords, irep, n_rep_neb, cvalues, Mmatrix)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(neb_var_type), POINTER                        :: coords
      INTEGER, INTENT(IN)                                :: irep, n_rep_neb
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: cvalues, Mmatrix

      CHARACTER(len=*), PARAMETER :: routineN = 'perform_replica_md', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle2, ierr, j, n_el
      LOGICAL                                            :: explicit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: md_section, root_section

      CALL timeset(routineN, handle)
      CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env, &
                              handle=handle2)
      logger => cp_get_default_logger()
      CALL force_env_get(f_env%force_env, globenv=globenv, &
                         root_section=root_section)
      j = rep_env%local_rep_indices(1)-1
      n_el = 3*rep_env%nparticle
      Mmatrix = 0.0_dp
      ! Syncronize position on the replica procs
      CALL set_pos(rep_env%f_env_id, rep_env%r(:, j+1), n_el, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_utils.F",537)
      !
      IF (irep+j <= n_rep_neb) THEN
         logger%iter_info%iteration(2) = irep+j
         CALL remove_restart_info(root_section)
         md_section => section_vals_get_subs_vals(root_section, "MOTION%MD")
         CALL section_vals_get(md_section, explicit=explicit)
         IF(.NOT.(explicit))CALL cp__a("motion/neb_utils.F",544)
         ! Let's syncronize the target of Collective Variables for this run
         CALL set_colvars_target(coords%int(:, irep+j), f_env%force_env)
         ! Do a molecular dynamics and get back the derivative
         ! of the free energy w.r.t. the colvar and the metric tensor
         CALL qs_mol_dyn(f_env%force_env, globenv=globenv)
         ! Collect the equilibrated coordinates
         CALL get_pos(rep_env%f_env_id, rep_env%r(1:n_el, j+1), n_el, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_utils.F",552)
         ! Write he gradients in the colvar coordinates into the replica_env array
         ! and copy back also the metric tensor..
         ! work in progress..
         CALL cp__b("motion/neb_utils.F",556,"")
         rep_env%f(:, j+1) = 0.0_dp
         Mmatrix = 0.0_dp
      ELSE
         rep_env%r(:, j+1) = 0.0_dp
         rep_env%f(:, j+1) = 0.0_dp
         cvalues(:, j+1) = 0.0_dp
         Mmatrix(:, j+1) = 0.0_dp
      END IF
      CALL rep_env_sync(rep_env, rep_env%f)
      CALL rep_env_sync(rep_env, rep_env%r)
      CALL rep_env_sync(rep_env, cvalues)
      CALL rep_env_sync(rep_env, Mmatrix)
      CALL f_env_rm_defaults(f_env=f_env, ierr=ierr, handle=handle2)
      IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_utils.F",570)
      CALL timestop(handle)
   END SUBROUTINE perform_replica_md

! **************************************************************************************************
!> \brief Driver to perform a GEO_OPT run on each single replica to
!>        compute specifically minimum energies in a collective variable
!>        NEB scheme
!> \param rep_env ...
!> \param coords ...
!> \param irep ...
!> \param n_rep_neb ...
!> \param cvalues ...
!> \param Mmatrix ...
!> \author Teodoro Laino 05.2007
! **************************************************************************************************
   SUBROUTINE perform_replica_geo(rep_env, coords, irep, n_rep_neb, cvalues, Mmatrix)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(neb_var_type), POINTER                        :: coords
      INTEGER, INTENT(IN)                                :: irep, n_rep_neb
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: cvalues, Mmatrix

      CHARACTER(len=*), PARAMETER :: routineN = 'perform_replica_geo', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle2, ierr, j, n_el
      LOGICAL                                            :: explicit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: geoopt_section, root_section

      CALL timeset(routineN, handle)
      CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env, &
                              handle=handle2)
      logger => cp_get_default_logger()
      CALL force_env_get(f_env%force_env, globenv=globenv, &
                         root_section=root_section)
      j = rep_env%local_rep_indices(1)-1
      n_el = 3*rep_env%nparticle
      Mmatrix = 0.0_dp
      ! Syncronize position on the replica procs
      CALL set_pos(rep_env%f_env_id, rep_env%r(:, j+1), n_el, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_utils.F",613)
      IF (irep+j <= n_rep_neb) THEN
         logger%iter_info%iteration(2) = irep+j
         CALL remove_restart_info(root_section)
         geoopt_section => section_vals_get_subs_vals(root_section, "MOTION%GEO_OPT")
         CALL section_vals_get(geoopt_section, explicit=explicit)
         IF(.NOT.(explicit))CALL cp__a("motion/neb_utils.F",619)
         ! Let's syncronize the target of Collective Variables for this run
         CALL set_colvars_target(coords%int(:, irep+j), f_env%force_env)
         ! Do a geometry optimization..
         CALL cp_geo_opt(f_env%force_env, globenv=globenv)
         ! Once the geometry optimization is ended let's do a single run
         ! without any constraints/restraints
         CALL force_env_calc_energy_force(f_env%force_env, &
                                          calc_force=.TRUE., skip_external_control=.TRUE.)
         ! Collect the optimized coordinates
         CALL get_pos(rep_env%f_env_id, rep_env%r(1:n_el, j+1), n_el, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_utils.F",630)
         ! Collect the gradients in cartesian coordinates
         CALL get_force(rep_env%f_env_id, rep_env%f(1:n_el, j+1), n_el, ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_utils.F",633)
         ! Copy the energy
         CALL get_energy(rep_env%f_env_id, rep_env%f(n_el+1, j+1), ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_utils.F",636)
         ! The gradients in the colvar coordinates
         CALL get_clv_force(f_env%force_env, rep_env%f(1:n_el, j+1), rep_env%r(1:n_el, j+1), &
                            SIZE(coords%xyz, 1), SIZE(coords%wrk, 1), cvalues(:, j+1), Mmatrix(:, j+1))
      ELSE
         rep_env%r(:, j+1) = 0.0_dp
         rep_env%f(:, j+1) = 0.0_dp
         cvalues(:, j+1) = 0.0_dp
         Mmatrix(:, j+1) = 0.0_dp
      END IF
      CALL rep_env_sync(rep_env, rep_env%f)
      CALL rep_env_sync(rep_env, rep_env%r)
      CALL rep_env_sync(rep_env, cvalues)
      CALL rep_env_sync(rep_env, Mmatrix)
      CALL f_env_rm_defaults(f_env=f_env, ierr=ierr, handle=handle2)
      IF(.NOT.(ierr == 0))CALL cp__a("motion/neb_utils.F",651)
      CALL timestop(handle)
   END SUBROUTINE perform_replica_geo

! **************************************************************************************************
!> \brief Computes the tangent for point i of the NEB chain
!> \param neb_env ...
!> \param coords ...
!> \param i ...
!> \param tangent ...
!> \param energies ...
!> \param iw ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE get_tangent(neb_env, coords, i, tangent, energies, iw)
      TYPE(neb_type), POINTER                            :: neb_env
      TYPE(neb_var_type), POINTER                        :: coords
      INTEGER, INTENT(IN)                                :: i
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: tangent
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: energies
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'get_tangent', routineP = moduleN//':'//routineN

      REAL(KINd=dp)                                      :: distance0, distance1, distance2, DVmax, &
                                                            Dvmin

      IF(.NOT.(ASSOCIATED(coords)))CALL cp__a("motion/neb_utils.F",678)
      tangent(:) = 0.0_dp
      ! For the last point we don't need any tangent..
      IF (i == neb_env%number_of_replica) RETURN
      ! Several kind of tangents implemented...
      SELECT CASE (neb_env%id_type)
      CASE (do_eb)
         tangent(:) = 0.0_dp
      CASE (do_b_neb)
         CALL neb_replica_distance(coords=coords, i0=i, i=i-1, distance=distance1, iw=iw, &
                                   rotate=.FALSE.)
         CALL neb_replica_distance(coords=coords, i0=i+1, i=i, distance=distance2, iw=iw, &
                                   rotate=.FALSE.)
         tangent(:) = (coords%wrk(:, i)-coords%wrk(:, i-1))/distance1+ &
                      (coords%wrk(:, i+1)-coords%wrk(:, i))/distance2
      CASE (do_it_neb, do_ci_neb, do_d_neb)
         IF ((energies(i+1) .GT. energies(i)) .AND. (energies(i) .GT. (energies(i-1)))) THEN
            tangent(:) = coords%wrk(:, i+1)-coords%wrk(:, i)
         ELSE IF ((energies(i+1) .LT. energies(i)) .AND. (energies(i) .LT. (energies(i-1)))) THEN
            tangent(:) = coords%wrk(:, i)-coords%wrk(:, i-1)
         ELSE
            DVmax = MAX(ABS(energies(i+1)-energies(i)), ABS(energies(i-1)-energies(i)))
            DVmin = MIN(ABS(energies(i+1)-energies(i)), ABS(energies(i-1)-energies(i)))
            IF (energies(i+1) .GE. energies(i-1)) THEN
               tangent(:) = (coords%wrk(:, i+1)-coords%wrk(:, i))*DVmax+(coords%wrk(:, i)-coords%wrk(:, i-1))*DVmin
            ELSE
               tangent(:) = (coords%wrk(:, i+1)-coords%wrk(:, i))*DVmin+(coords%wrk(:, i)-coords%wrk(:, i-1))*DVmax
            END IF
         END IF
      CASE (do_sm)
         ! String method..
         tangent(:) = 0.0_dp
      END SELECT
      distance0 = SQRT(DOT_PRODUCT(tangent(:), tangent(:)))
      IF (distance0 /= 0.0_dp) tangent(:) = tangent(:)/distance0
   END SUBROUTINE get_tangent

! **************************************************************************************************
!> \brief Computes the forces for point i of the NEB chain
!> \param neb_env ...
!> \param tangent ...
!> \param coords ...
!> \param i ...
!> \param forces ...
!> \param tag ...
!> \param Mmatrix ...
!> \param iw ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   RECURSIVE SUBROUTINE get_neb_force(neb_env, tangent, coords, i, forces, tag, Mmatrix, &
                                      iw)
      TYPE(neb_type), POINTER                            :: neb_env
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: tangent
      TYPE(neb_var_type), POINTER                        :: coords
      INTEGER, INTENT(IN)                                :: i
      TYPE(neb_var_type), POINTER                        :: forces
      INTEGER, INTENT(IN), OPTIONAL                      :: tag
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: Mmatrix
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'get_neb_force', routineP = moduleN//':'//routineN

      INTEGER                                            :: j, my_tag, nsize_wrk
      REAL(KIND=dp)                                      :: distance1, distance2, fac, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: dtmp1, wrk

      my_tag = neb_env%id_type
      IF (PRESENT(tag)) my_tag = tag
      IF(.NOT.(ASSOCIATED(forces)))CALL cp__a("motion/neb_utils.F",746)
      IF(.NOT.(ASSOCIATED(coords)))CALL cp__a("motion/neb_utils.F",747)
      nsize_wrk = coords%size_wrk(1)
      ! All methods but not the classical elastic band will skip the force
      ! calculation for the last frame of the band
      SELECT CASE (my_tag)
      CASE (do_b_neb, do_it_neb, do_ci_neb, do_d_neb)
         IF (i == neb_env%number_of_replica) RETURN
      CASE (do_sm)
         ! String Method
         ! The forces do not require any projection. Reparametrization required
         ! after the update of the replica.
         CALL cite_reference(E2002)
         RETURN
      END SELECT
      ! otherwise proceeed normally..
      ALLOCATE (wrk(nsize_wrk))
      ! Spring Energy
      CALL neb_replica_distance(coords=coords, i0=i-1, i=i, distance=distance1, iw=iw, &
                                rotate=.FALSE.)
      tmp = distance1-neb_env%avg_distance
      neb_env%spring_energy = neb_env%spring_energy+0.5_dp*neb_env%k*tmp**2
      SELECT CASE (my_tag)
      CASE (do_eb)
         CALL cite_reference(Elber1987)
         ! Elastic band - Hamiltonian formulation according the original Karplus/Elber
         !                formulation
         ALLOCATE (dtmp1(nsize_wrk))
         ! derivatives of the spring
         tmp = distance1-neb_env%avg_distance
         dtmp1(:) = 1.0_dp/distance1*(coords%wrk(:, i)-coords%wrk(:, i-1))
         wrk(:) = neb_env%k*tmp*dtmp1
         forces%wrk(:, i) = forces%wrk(:, i)-wrk
         forces%wrk(:, i-1) = forces%wrk(:, i-1)+wrk
         ! derivatives due to the average length of the spring
         fac = 1.0_dp/(neb_env%avg_distance*REAL(neb_env%number_of_replica-1, KIND=dp))
         wrk(:) = neb_env%k*fac*(coords%wrk(:, i)-coords%wrk(:, i-1))
         tmp = 0.0_dp
         DO j = 2, neb_env%number_of_replica
            CALL neb_replica_distance(coords=coords, i0=j-1, i=j, distance=distance1, iw=iw, &
                                      rotate=.FALSE.)
            tmp = tmp+distance1-neb_env%avg_distance
         END DO
         forces%wrk(:, i) = forces%wrk(:, i)+wrk*tmp
         forces%wrk(:, i-1) = forces%wrk(:, i-1)-wrk*tmp
         DEALLOCATE (dtmp1)
      CASE (do_b_neb)
         ! Bisection NEB
         CALL cite_reference(Jonsson1998)
         wrk(:) = (coords%wrk(:, i+1)-2.0_dp*coords%wrk(:, i)+coords%wrk(:, i-1))
         tmp = neb_env%k*DOT_PRODUCT(wrk, tangent)
         wrk(:) = forces%wrk(:, i)-dot_product_band(neb_env, forces%wrk(:, i), tangent, Mmatrix)*tangent
         forces%wrk(:, i) = wrk+tmp*tangent
      CASE (do_it_neb)
         ! Improved tangent NEB
         CALL cite_reference(Jonsson2000_1)
         CALL neb_replica_distance(coords=coords, i0=i, i=i+1, distance=distance1, iw=iw, &
                                   rotate=.FALSE.)
         CALL neb_replica_distance(coords=coords, i0=i-1, i=i, distance=distance2, iw=iw, &
                                   rotate=.FALSE.)
         tmp = neb_env%k*(distance1-distance2)
         wrk(:) = forces%wrk(:, i)-dot_product_band(neb_env, forces%wrk(:, i), tangent, Mmatrix)*tangent
         forces%wrk(:, i) = wrk+tmp*tangent
      CASE (do_ci_neb)
         ! Climbing Image NEB
         CALL cite_reference(Jonsson2000_2)
         IF (neb_env%istep <= neb_env%nsteps_it .OR. i /= neb_env%nr_HE_image) THEN
            CALL get_neb_force(neb_env, tangent, coords, i, forces, do_it_neb, Mmatrix, iw)
         ELSE
            wrk(:) = forces%wrk(:, i)
            tmp = -2.0_dp*dot_product_band(neb_env, wrk, tangent, Mmatrix)
            forces%wrk(:, i) = wrk+tmp*tangent
         END IF
      CASE (do_d_neb)
         ! Doubly NEB
         CALL cite_reference(Wales2004)
         ALLOCATE (dtmp1(nsize_wrk))
         dtmp1(:) = forces%wrk(:, i)-dot_product_band(neb_env, forces%wrk(:, i), tangent, Mmatrix)*tangent
         forces%wrk(:, i) = dtmp1
         tmp = SQRT(DOT_PRODUCT(dtmp1, dtmp1))
         dtmp1(:) = dtmp1(:)/tmp
         ! Project out only the spring component interfering with the
         ! orthogonal gradient of the band
         wrk(:) = (coords%wrk(:, i+1)-2.0_dp*coords%wrk(:, i)+coords%wrk(:, i-1))
         tmp = DOT_PRODUCT(wrk, dtmp1)
         dtmp1(:) = neb_env%k*(wrk(:)-tmp*dtmp1(:))
         forces%wrk(:, i) = forces%wrk(:, i)+dtmp1(:)
         DEALLOCATE (dtmp1)
      END SELECT
      DEALLOCATE (wrk)
   END SUBROUTINE get_neb_force

! **************************************************************************************************
!> \brief  Handles the dot_product when using colvar.. in this case
!>         the scalar product needs to take into account the metric
!>         tensor
!> \param neb_env ...
!> \param array1 ...
!> \param array2 ...
!> \param array3 ...
!> \retval value ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   FUNCTION dot_product_band(neb_env, array1, array2, array3) RESULT(value)
      TYPE(neb_type), POINTER                            :: neb_env
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: array1, array2
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: array3
      REAL(KIND=dp)                                      :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'dot_product_band', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nsize_int
      LOGICAL                                            :: check

      IF (neb_env%use_colvar) THEN
         nsize_int = neb_env%nsize_int
         check = ((SIZE(array1) /= SIZE(array2)) .OR. &
                  (SIZE(array1) /= nsize_int) .OR. &
                  (SIZE(array3) /= nsize_int*nsize_int))
         ! This condition should always be satisfied..
         IF(.NOT.(check))CALL cp__a("motion/neb_utils.F",867)
         value = DOT_PRODUCT(MATMUL(RESHAPE(array3, (/nsize_int, nsize_int/)), array1), array2)
      ELSE
         value = DOT_PRODUCT(array1, array2)
      END IF
   END FUNCTION dot_product_band

! **************************************************************************************************
!> \brief Reorient iteratively all images of the NEB chain in order to
!>      have always the smaller RMSD between two following images
!> \param rotate_frames ...
!> \param particle_set ...
!> \param coords ...
!> \param vels ...
!> \param iw ...
!> \param distances ...
!> \param number_of_replica ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE reorient_images(rotate_frames, particle_set, coords, vels, iw, &
                              distances, number_of_replica)
      LOGICAL, INTENT(IN)                                :: rotate_frames
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(neb_var_type), POINTER                        :: coords, vels
      INTEGER, INTENT(IN)                                :: iw
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: distances
      INTEGER, INTENT(IN)                                :: number_of_replica

      CHARACTER(len=*), PARAMETER :: routineN = 'reorient_images', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k, kind
      LOGICAL                                            :: check
      REAL(KIND=dp)                                      :: xtmp
      REAL(KIND=dp), DIMENSION(3)                        :: tmp
      REAL(KIND=dp), DIMENSION(3, 3)                     :: rot

      rot = 0.0_dp
      rot(1, 1) = 1.0_dp
      rot(2, 2) = 1.0_dp
      rot(3, 3) = 1.0_dp
      DO i = 2, number_of_replica
         ! The rotation of the replica is enabled exclusively when working in
         ! cartesian coordinates
         IF (rotate_frames .AND. (coords%in_use == do_band_cartesian)) THEN
            CALL rmsd3(particle_set, coords%xyz(:, i), coords%xyz(:, i-1), iw, &
                       rotate=.TRUE., rot=rot)
            ! Rotate velocities
            DO k = 1, SIZE(vels%xyz, 1)/3
               kind = (k-1)*3
               tmp = vels%xyz(kind+1:kind+3, i)
               CALL matvec_3x3(vels%xyz(kind+1:kind+3, i), TRANSPOSE(rot), tmp)
            END DO
         END IF
         IF (PRESENT(distances)) THEN
            check = SIZE(distances) == (number_of_replica-1)
            IF(.NOT.(check))CALL cp__a("motion/neb_utils.F",924)
            xtmp = DOT_PRODUCT(coords%wrk(:, i)-coords%wrk(:, i-1), &
                               coords%wrk(:, i)-coords%wrk(:, i-1))
            distances(i-1) = SQRT(xtmp)
         END IF
      END DO
   END SUBROUTINE reorient_images

! **************************************************************************************************
!> \brief Reparametrization of the replica for String Method with splines
!> \param reparametrize_frames ...
!> \param spline_order ...
!> \param smoothing ...
!> \param coords ...
!> \param sline ...
!> \param distances ...
!> \author Teodoro Laino - Rodolphe Vuilleumier 09.2008
! **************************************************************************************************
   SUBROUTINE reparametrize_images(reparametrize_frames, spline_order, smoothing, &
                                   coords, sline, distances)

      LOGICAL, INTENT(IN)                                :: reparametrize_frames
      INTEGER, INTENT(IN)                                :: spline_order
      REAL(KIND=dp), INTENT(IN)                          :: smoothing
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: coords, sline
      REAL(KIND=dp), DIMENSION(:)                        :: distances

      CHARACTER(len=*), PARAMETER :: routineN = 'reparametrize_images', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      REAL(KIND=dp)                                      :: avg_distance, xtmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: tmp_coords

      IF (reparametrize_frames) THEN
         ALLOCATE (tmp_coords(SIZE(coords, 1), SIZE(coords, 2)))
         tmp_coords(:, :) = coords
         ! Smoothing
         DO i = 2, SIZE(coords, 2)-1
            coords(:, i) = tmp_coords(:, i)*(1.0_dp-2.0_dp*smoothing)+ &
                           tmp_coords(:, i-1)*smoothing+tmp_coords(:, i+1)*smoothing
         END DO
         sline = coords-tmp_coords+sline
         tmp_coords(:, :) = coords
         ! Reparametrization
         SELECT CASE (spline_order)
         CASE (1)
            ! Compute distances
            DO i = 2, SIZE(coords, 2)
               xtmp = DOT_PRODUCT(coords(:, i)-coords(:, i-1), coords(:, i)-coords(:, i-1))
               distances(i-1) = SQRT(xtmp)
            END DO
            avg_distance = SUM(distances)/REAL(SIZE(coords, 2)-1, KIND=dp)
            ! Redistribute frames
            DO i = 2, SIZE(coords, 2)-1
               xtmp = 0.0_dp
               DO j = 1, SIZE(coords, 2)-1
                  xtmp = xtmp+distances(j)
                  IF (xtmp > avg_distance*REAL(i-1, KIND=dp)) THEN
                     xtmp = (xtmp-avg_distance*REAL(i-1, KIND=dp))/distances(j)
                     coords(:, i) = (1.0_dp-xtmp)*tmp_coords(:, j+1)+xtmp*tmp_coords(:, j)
                     EXIT
                  END IF
               END DO
            END DO
            ! Re-compute distances
            DO i = 2, SIZE(coords, 2)
               xtmp = DOT_PRODUCT(coords(:, i)-coords(:, i-1), coords(:, i)-coords(:, i-1))
               distances(i-1) = SQRT(xtmp)
            END DO
         CASE DEFAULT
            CALL cp__w("motion/neb_utils.F",995,"String Method: Spline order greater than 1 not implemented.")
         END SELECT
         sline = coords-tmp_coords+sline
         DEALLOCATE (tmp_coords)
      END IF
   END SUBROUTINE reparametrize_images

! **************************************************************************************************
!> \brief Checks for convergence criteria during a NEB run
!> \param neb_env ...
!> \param Dcoords ...
!> \param forces ...
!> \retval converged ...
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
   FUNCTION check_convergence(neb_env, Dcoords, forces) RESULT(converged)
      TYPE(neb_type), POINTER                            :: neb_env
      TYPE(neb_var_type), POINTER                        :: Dcoords, forces
      LOGICAL                                            :: converged

      CHARACTER(len=*), PARAMETER :: routineN = 'check_convergence', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=3), DIMENSION(4)                     :: labels
      INTEGER                                            :: iw
      REAL(KIND=dp)                                      :: max_dr, max_force, my_max_dr, &
                                                            my_max_force, my_rms_dr, my_rms_force, &
                                                            rms_dr, rms_force
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: cc_section

      NULLIFY (logger, cc_section)
      logger => cp_get_default_logger()
      cc_section => section_vals_get_subs_vals(neb_env%neb_section, "CONVERGENCE_CONTROL")
      CALL section_vals_val_get(cc_section, "MAX_DR", r_val=max_dr)
      CALL section_vals_val_get(cc_section, "MAX_FORCE", r_val=max_force)
      CALL section_vals_val_get(cc_section, "RMS_DR", r_val=rms_dr)
      CALL section_vals_val_get(cc_section, "RMS_FORCE", r_val=rms_force)
      converged = .FALSE.
      labels = " NO"
      my_max_dr = MAXVAL(ABS(Dcoords%wrk))
      my_max_force = MAXVAL(ABS(forces%wrk))
      my_rms_dr = SQRT(SUM(Dcoords%wrk*Dcoords%wrk)/REAL(SIZE(Dcoords%wrk, 1)*SIZE(Dcoords%wrk, 2), KIND=dp))
      my_rms_force = SQRT(SUM(forces%wrk*forces%wrk)/REAL(SIZE(forces%wrk, 1)*SIZE(forces%wrk, 2), KIND=dp))
      IF (my_max_dr < max_dr) labels(1) = "YES"
      IF (my_max_force < max_force) labels(2) = "YES"
      IF (my_rms_dr < rms_dr) labels(3) = "YES"
      IF (my_rms_force < rms_force) labels(4) = "YES"
      IF (ALL(labels == "YES")) converged = .TRUE.

      iw = cp_print_key_unit_nr(logger, neb_env%neb_section, "CONVERGENCE_INFO", &
                                extension=".nebLog")
      IF (iw > 0) THEN
         ! Print convergence info
         WRITE (iw, FMT='(A,A)') ' **************************************', &
            '*****************************************'
         WRITE (iw, FMT='(1X,A,2X,F8.5,5X,"[",F8.5,"]",1X,T76,"(",A,")")') &
            'RMS DISPLACEMENT =', my_rms_dr, rms_dr, labels(3), &
            'MAX DISPLACEMENT =', my_max_dr, max_dr, labels(1), &
            'RMS FORCE        =', my_rms_force, rms_force, labels(4), &
            'MAX FORCE        =', my_max_force, max_force, labels(2)
         WRITE (iw, FMT='(A,A)') ' **************************************', &
            '*****************************************'
      END IF
      CALL cp_print_key_finished_output(iw, logger, neb_env%neb_section, &
                                        "CONVERGENCE_INFO")
   END FUNCTION check_convergence

END MODULE neb_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module with utility to perform MD Nudged Elastic Band Calculation
!> \note
!>      Numerical accuracy for parallel runs:
!>       Each replica starts the SCF run from the one optimized
!>       in a previous run. It may happen then energies and derivatives
!>       of a serial run and a parallel run could be slightly different
!>       'cause of a different starting density matrix.
!>       Exact results are obtained using:
!>          EXTRAPOLATION USE_GUESS in QS section (Teo 09.2006)
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
MODULE neb_opt_utils
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit,&
                                              cp_to_string
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE neb_types,                       ONLY: neb_type,&
                                              neb_var_type
   USE neb_utils,                       ONLY: neb_calc_energy_forces,&
                                              reorient_images
   USE particle_types,                  ONLY: particle_type
   USE replica_types,                   ONLY: replica_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'neb_opt_utils'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.

   PUBLIC :: accept_diis_step, &
             neb_ls

   REAL(KIND=dp), DIMENSION(2:10), PRIVATE :: acceptance_factor = &
                                              (/0.97_dp, 0.84_dp, 0.71_dp, 0.67_dp, 0.62_dp, 0.56_dp, 0.49_dp, 0.41_dp, 0.0_dp/)

CONTAINS

! **************************************************************************************************
!> \brief Performes few basic operations for the NEB DIIS optimization
!> \param apply_diis ...
!> \param n_diis ...
!> \param err ...
!> \param crr ...
!> \param set_err ...
!> \param sline ...
!> \param coords ...
!> \param check_diis ...
!> \param iw2 ...
!> \retval accepted ...
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
   FUNCTION accept_diis_step(apply_diis, n_diis, err, crr, set_err, sline, coords, &
                             check_diis, iw2) RESULT(accepted)
      LOGICAL, INTENT(IN)                                :: apply_diis
      INTEGER, INTENT(IN)                                :: n_diis
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: err, crr
      INTEGER, DIMENSION(:), POINTER                     :: set_err
      TYPE(neb_var_type), POINTER                        :: sline, coords
      LOGICAL, INTENT(IN)                                :: check_diis
      INTEGER, INTENT(IN)                                :: iw2
      LOGICAL                                            :: accepted

      CHARACTER(len=*), PARAMETER :: routineN = 'accept_diis_step', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: line
      INTEGER                                            :: i, iend, ind, indi, indj, info, istart, &
                                                            iv, iw, j, jv, k, lwork, np, nv
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: IWORK
      LOGICAL                                            :: increase_error
      REAL(dp), DIMENSION(:, :), POINTER                 :: work
      REAL(KIND=dp)                                      :: eps_svd
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: S, Work_dgesdd
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: U, VT, wrk, wrk_inv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: awrk, cwrk, ref, step

      iw = cp_logger_get_default_io_unit()
      accepted = .FALSE.
      ! find the index with the minimum element of the set_err array
      nv = MINLOC(set_err, 1)
      IF (iw2 > 0) WRITE (iw2, '(A,I3)') "Entering into the DIIS module. Error vector number:: ", nv
      set_err(nv) = 1
      eps_svd = 1.0E-10_dp
      ALLOCATE (step(sline%size_wrk(1)*sline%size_wrk(2)))
      ALLOCATE (ref(sline%size_wrk(1)*sline%size_wrk(2)))
      err(:, nv) = RESHAPE(sline%wrk, (/sline%size_wrk(1)*sline%size_wrk(2)/))
      crr(:, nv) = RESHAPE(coords%wrk, (/coords%size_wrk(1)*coords%size_wrk(2)/))
      jv = n_diis
      IF (ALL(set_err == 1) .AND. apply_diis) THEN
         IF (iw2 > 0) WRITE (iw2, '(A)') "Applying DIIS equations"
         ! Apply DIIS..
         DO jv = 2, n_diis
            np = jv+1
            IF (iw2 > 0) WRITE (iw2, '(A,I5,A)') "Applying DIIS equations with the last", &
               jv, " error vectors"
            ALLOCATE (wrk(np, np))
            ALLOCATE (work(np, np))
            ALLOCATE (wrk_inv(np, np))
            ALLOCATE (cwrk(np))
            ALLOCATE (awrk(np))
            awrk = 0.0_dp
            wrk = 1.0_dp
            wrk(np, np) = 0.0_dp
            awrk(np) = 1.0_dp
            DO i = 1, jv
               indi = n_diis-i+1
               DO j = i, jv
                  indj = n_diis-j+1
                  wrk(i, j) = DOT_PRODUCT(err(:, indi), err(:, indj))
                  wrk(j, i) = wrk(i, j)
               END DO
            END DO
            IF (iw2 > 0) THEN
               line = "DIIS Matrix"//cp_to_string(np)//"x"//cp_to_string(np)//"."
               WRITE (iw2, '(A)') TRIM(line)
               WRITE (iw2, '('//cp_to_string(np)//'F12.6)') wrk
            END IF
            ! Inverte the DIIS Matrix
            work = TRANSPOSE(wrk)
            ! Workspace query
            ALLOCATE (iwork(8*np))
            ALLOCATE (S(np))
            ALLOCATE (U(np, np))
            ALLOCATE (VT(np, np))
            ALLOCATE (work_dgesdd(1))
            lwork = -1
            CALL DGESDD('S', np, np, work, np, S, U, np, vt, np, work_dgesdd, lwork, iwork, info)
            lwork = INT(work_dgesdd(1))
            DEALLOCATE (work_dgesdd)
            ALLOCATE (work_dgesdd(lwork))
            CALL DGESDD('S', np, np, work, np, S, U, np, vt, np, work_dgesdd, lwork, iwork, info)
            ! Construct the inverse
            DO k = 1, np
               ! Invert SV
               IF (S(k) < eps_svd) THEN
                  S(k) = 0.0_dp
               ELSE
                  S(k) = 1.0_dp/S(k)
               ENDIF
               VT(k, :) = VT(k, :)*S(k)
            ENDDO
            CALL DGEMM('T', 'T', np, np, np, 1.0_dp, VT, np, U, np, 0.0_dp, wrk_inv, np)
            DEALLOCATE (iwork)
            DEALLOCATE (S)
            DEALLOCATE (U)
            DEALLOCATE (VT)
            DEALLOCATE (work_dgesdd)
            cwrk = MATMUL(wrk_inv, awrk)
            ! Check the DIIS solution
            step = 0.0_dp
            ind = 0
            DO i = n_diis, n_diis-jv+1, -1
               ind = ind+1
               step = step+(crr(:, i)+err(:, i))*cwrk(ind)
            END DO
            step = step-crr(:, n_diis)
            ref = err(:, n_diis)
            increase_error = check_diis_solution(jv, cwrk, step, ref, &
                                                 iw2, check_diis)
            ! possibly enlarge the error space
            IF (increase_error) THEN
               accepted = .TRUE.
               sline%wrk = RESHAPE(step, (/sline%size_wrk(1), sline%size_wrk(2)/))
            ELSE
               DEALLOCATE (awrk)
               DEALLOCATE (cwrk)
               DEALLOCATE (wrk)
               DEALLOCATE (work)
               DEALLOCATE (wrk_inv)
               EXIT
            END IF
            DEALLOCATE (awrk)
            DEALLOCATE (cwrk)
            DEALLOCATE (wrk)
            DEALLOCATE (work)
            DEALLOCATE (wrk_inv)
         END DO
         IF (iw2 > 0) THEN
            line = "Exiting DIIS accepting"//cp_to_string(MIN(n_diis, jv))//" errors."
            WRITE (iw2, '(A)') TRIM(line)
         END IF
      END IF
      IF (ALL(set_err == 1)) THEN
         ! always delete the last error vector from the history vectors
         ! move error vectors and the set_err in order to have free space
         ! at the end of the err array
         istart = MAX(2, n_diis-jv+2)
         iend = n_diis
         indi = 0
         DO iv = istart, iend
            indi = indi+1
            err(:, indi) = err(:, iv)
            crr(:, indi) = crr(:, iv)
            set_err(indi) = 1
         END DO
         DO iv = indi+1, iend
            set_err(iv) = -1
         END DO
      END IF
      DEALLOCATE (step)
      DEALLOCATE (ref)

   END FUNCTION accept_diis_step

! **************************************************************************************************
!> \brief Check conditions for the acceptance of the DIIS step
!> \param nv ...
!> \param cwrk ...
!> \param step ...
!> \param ref ...
!> \param output_unit ...
!> \param check_diis ...
!> \retval accepted ...
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
   FUNCTION check_diis_solution(nv, cwrk, step, ref, output_unit, check_diis) &
      RESULT(accepted)
      INTEGER, INTENT(IN)                                :: nv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cwrk, step, ref
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(IN)                                :: check_diis
      LOGICAL                                            :: accepted

      CHARACTER(len=*), PARAMETER :: routineN = 'check_diis_solution', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: costh, norm1, norm2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: tmp

      accepted = .TRUE.
      ALLOCATE (tmp(SIZE(step)))
      IF (accepted) THEN
         ! (a) The direction of the DIIS step, can be compared to the reference step.
         !     if the angle is grater than a specified value, the DIIS step is not
         !     acceptable.
         norm1 = SQRT(DOT_PRODUCT(ref, ref))
         norm2 = SQRT(DOT_PRODUCT(step, step))
         costh = DOT_PRODUCT(ref, step)/(norm1*norm2)
         IF (check_diis) THEN
            IF (costh < acceptance_factor(MIN(10, nv))) accepted = .FALSE.
         ELSE
            IF (costh <= 0.0_dp) accepted = .FALSE.
         END IF
         IF (output_unit > 0 .AND. (.NOT. accepted)) THEN
            WRITE (output_unit, '(T2,"DIIS|",A)') &
               "The direction of the DIIS step, can be compared to the reference step.", &
               "If the angle is grater than a specified value, the DIIS step is not", &
               "acceptable. Value exceeded. Reset DIIS!"
            WRITE (output_unit, '(T2,"DIIS|",A,F6.3,A,F6.3,A)') &
               "Present Cosine <", costh, "> compared with the optimal value <", &
               acceptance_factor(MIN(10, nv)), "> ."
         END IF
      END IF
      IF (accepted .AND. check_diis) THEN
         ! (b) The length of the DIIS step is limited to be no more than 10 times
         !     the reference step
         IF (norm1 > norm2*10.0_dp) accepted = .FALSE.
         IF (output_unit > 0 .AND. (.NOT. accepted)) THEN
            WRITE (output_unit, '(T2,"DIIS|",A)') &
               "The length of the DIIS step is limited to be no more than 10 times", &
               "the reference step. Value exceeded. Reset DIIS!"
         END IF
      END IF
      IF (accepted .AND. check_diis) THEN
         ! (d) If the DIIS matrix is nearly singular, the norm of the DIIS step
         !     vector becomes small and cwrk/norm1 becomes large, signaling a
         !     numerical stability problems. If the magnitude of cwrk/norm1
         !     exceeds 10^8 then the step size is assumed to be unacceptable
         IF (ANY(ABS(cwrk(1:nv)/norm1) > 10**8_dp)) accepted = .FALSE.
         IF (output_unit > 0 .AND. (.NOT. accepted)) THEN
            WRITE (output_unit, '(T2,"DIIS|",A)') &
               "If the DIIS matrix is nearly singular, the norm of the DIIS step", &
               "vector becomes small and Coeff/E_norm becomes large, signaling a", &
               "numerical stability problems. IF the magnitude of Coeff/E_norm", &
               "exceeds 10^8 THEN the step size is assumed to be unacceptable.", &
               "Value exceeded. Reset DIIS!"
         END IF
      END IF
      DEALLOCATE (tmp)
   END FUNCTION check_diis_solution

! **************************************************************************************************
!> \brief Perform a line minimization search in optimizing a NEB with DIIS
!> \param stepsize ...
!> \param sline ...
!> \param rep_env ...
!> \param neb_env ...
!> \param coords ...
!> \param energies ...
!> \param forces ...
!> \param vels ...
!> \param particle_set ...
!> \param iw ...
!> \param output_unit ...
!> \param distances ...
!> \param diis_section ...
!> \param iw2 ...
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
   SUBROUTINE neb_ls(stepsize, sline, rep_env, neb_env, coords, energies, forces, &
                     vels, particle_set, iw, output_unit, distances, diis_section, iw2)
      REAL(KIND=dp), INTENT(INOUT)                       :: stepsize
      TYPE(neb_var_type), POINTER                        :: sline
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(neb_type), OPTIONAL, POINTER                  :: neb_env
      TYPE(neb_var_type), POINTER                        :: coords
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: energies
      TYPE(neb_var_type), POINTER                        :: forces, vels
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: iw, output_unit
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: distances
      TYPE(section_vals_type), POINTER                   :: diis_section
      INTEGER, INTENT(IN)                                :: iw2

      CHARACTER(len=*), PARAMETER :: routineN = 'neb_ls', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, np
      REAL(KIND=dp)                                      :: a, b, max_stepsize, xa, xb, xc_cray, ya, &
                                                            yb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Icoord

! replaced xc by xc_cray to work around yet another bug in pgf90 on CRAY xt3

      ALLOCATE (Icoord(coords%size_wrk(1), coords%size_wrk(2)))
      CALL section_vals_val_get(diis_section, "NP_LS", i_val=np)
      CALL section_vals_val_get(diis_section, "MAX_STEPSIZE", r_val=max_stepsize)
      Icoord(:, :) = coords%wrk
      xa = 0.0_dp
      ya = SUM(sline%wrk*forces%wrk)
      xb = xa+MIN(ya*stepsize, max_stepsize)
      xc_cray = xb
      i = 1
      DO WHILE (i <= np-1)
         i = i+1
         coords%wrk = Icoord+xb*sline%wrk
         CALL reorient_images(neb_env%rotate_frames, particle_set, coords, vels, &
                              output_unit, distances, neb_env%number_of_replica)
         neb_env%avg_distance = SQRT(SUM(distances*distances)/REAL(SIZE(distances), KIND=dp))
         CALL neb_calc_energy_forces(rep_env, neb_env, coords, energies, forces, &
                                     particle_set, iw)
         yb = SUM(sline%wrk*forces%wrk)
         a = (ya-yb)/(2.0_dp*(xa-xb))
         b = ya-2.0_dp*a*xa
         xc_cray = -b/(2.0_dp*a)
         IF (xc_cray > max_stepsize) THEN
            IF (iw2 > 0) WRITE (iw2, '(T2,2(A,F6.3),A)') &
               "LS| Predicted stepsize (", xc_cray, ") greater than allowed stepsize (", &
               max_stepsize, ").  Reset!"
            xc_cray = max_stepsize
            EXIT
         END IF
         ! No Extrapolation .. only interpolation
         IF ((xc_cray <= MIN(xa, xb) .OR. xc_cray >= MAX(xa, xb)) .AND. (ABS(xa-xb) > 1.0E-5_dp)) THEN
            IF (iw2 > 0) WRITE (iw2, '(T2,2(A,I5),A)') &
               "LS| Increasing the number of point from ", np, " to ", np+1, "."
            np = np+1
         END IF
         !
         IF (ABS(yb) < ABS(ya)) THEN
            ya = yb
            xa = xb
         END IF
         xb = xc_cray
      END DO
      stepsize = xc_cray
      coords%wrk = Icoord
      DEALLOCATE (Icoord)
   END SUBROUTINE neb_ls

END MODULE neb_opt_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module performing a Nudged Elastic Band Calculation
!> \note
!>      Numerical accuracy for parallel runs:
!>       Each replica starts the SCF run from the one optimized
!>       in a previous run. It may happen then energies and derivatives
!>       of a serial run and a parallel run could be slightly different
!>       'cause of a different starting density matrix.
!>       Exact results are obtained using:
!>          EXTRAPOLATION USE_GUESS in QS section (Teo 09.2006)
!> \author Teodoro Laino 09.2006
!> \par  History
!>       - Teodoro Laino 10.2008 [tlaino] - University of Zurich
!>         Extension to a subspace of collective variables
! **************************************************************************************************
MODULE neb_methods
   USE colvar_utils,                    ONLY: number_of_colvar
   USE cp_external_control,             ONLY: external_control
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_iterate,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE f77_interface,                   ONLY: f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE force_env_types,                 ONLY: force_env_get
   USE global_types,                    ONLY: global_environment_type
   USE header,                          ONLY: band_header
   USE input_constants,                 ONLY: band_diis_opt,&
                                              band_md_opt,&
                                              do_rep_blocked,&
                                              do_sm
   USE input_section_types,             ONLY: section_type,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE neb_io,                          ONLY: dump_neb_info,&
                                              neb_rep_env_map_info,&
                                              read_neb_section
   USE neb_md_utils,                    ONLY: control_vels_a,&
                                              control_vels_b
   USE neb_opt_utils,                   ONLY: accept_diis_step,&
                                              neb_ls
   USE neb_types,                       ONLY: neb_type,&
                                              neb_var_create,&
                                              neb_var_release,&
                                              neb_var_type
   USE neb_utils,                       ONLY: build_replica_coords,&
                                              check_convergence,&
                                              neb_calc_energy_forces,&
                                              reorient_images,&
                                              reparametrize_images
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: massunit
   USE replica_methods,                 ONLY: rep_env_create
   USE replica_types,                   ONLY: rep_env_release,&
                                              replica_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'neb_methods'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   PUBLIC :: neb

CONTAINS

! **************************************************************************************************
!> \brief Real subroutine for NEB calculations
!> \param input ...
!> \param input_declaration ...
!> \param para_env ...
!> \param globenv ...
!> \author Teodoro Laino 09.2006
!> \note
!>      Based on the use of replica_env
! **************************************************************************************************
   SUBROUTINE neb(input, input_declaration, para_env, globenv)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'neb', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ierr, iw, iw2, nrep, &
                                                            output_unit, prep, proc_dist_type
      LOGICAL                                            :: check, row_force
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(neb_type), POINTER                            :: neb_env
      TYPE(neb_var_type), POINTER                        :: coords, forces, vels
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(section_vals_type), POINTER                   :: diis_section, force_env_section, &
                                                            md_section, motion_section, &
                                                            neb_section, print_section

      CALL timeset(routineN, handle)
      NULLIFY (logger, subsys, f_env, rep_env)
      NULLIFY (forces, coords, vels, neb_env)
      logger => cp_get_default_logger()
      CALL cp_add_iter_level(logger%iter_info, "BAND")
      motion_section => section_vals_get_subs_vals(input, "MOTION")
      print_section => section_vals_get_subs_vals(motion_section, "PRINT")
      neb_section => section_vals_get_subs_vals(motion_section, "BAND")
      output_unit = cp_print_key_unit_nr(logger, neb_section, "PROGRAM_RUN_INFO", &
                                         extension=".nebLog")
      CALL section_vals_val_get(neb_section, "NPROC_REP", i_val=prep)
      CALL section_vals_val_get(neb_section, "PROC_DIST_TYPE", i_val=proc_dist_type)
      row_force = (proc_dist_type == do_rep_blocked)
      nrep = MAX(1, para_env%num_pe/prep)
      IF (nrep*prep /= para_env%num_pe .AND. output_unit > 0) THEN
         CALL cp_warn(cp__l("motion/neb_methods.F",127), &
                      "Number of totally requested processors ("//TRIM(ADJUSTL(cp_to_string(para_env%num_pe)))//") "// &
                      "is not compatible with the number of processors requested per replica ("// &
                      TRIM(ADJUSTL(cp_to_string(prep)))//") and the number of replicas ("// &
                      TRIM(ADJUSTL(cp_to_string(nrep)))//") . ["// &
                      TRIM(ADJUSTL(cp_to_string(para_env%num_pe-nrep*prep)))//"] processors will be wasted!")
      ENDIF
      force_env_section => section_vals_get_subs_vals(input, "FORCE_EVAL")
      ! Create Replica Environments
      IF (output_unit > 0) WRITE (output_unit, '(T2,"NEB|",A)') " Replica_env Setup. START"
      CALL rep_env_create(rep_env, para_env=para_env, input=input, &
                          input_declaration=input_declaration, nrep=nrep, prep=prep, row_force=row_force)
      IF (output_unit > 0) WRITE (output_unit, '(T2,"NEB|",A)') " Replica_env Setup. END"
      IF (ASSOCIATED(rep_env)) THEN
         IF(.NOT.(SIZE(rep_env%local_rep_indices) == 1))CALL cp__a("motion/neb_methods.F",141)
         CALL f_env_add_defaults(f_env_id=rep_env%f_env_id, f_env=f_env)
         CALL force_env_get(f_env%force_env, subsys=subsys)
         particle_set => subsys%particles%els
         ! Read NEB controlling parameters
         ALLOCATE (neb_env)
         neb_env%force_env => f_env%force_env
         neb_env%root_section => input
         neb_env%force_env_section => force_env_section
         neb_env%motion_print_section => print_section
         neb_env%neb_section => neb_section
         neb_env%nsize_xyz = rep_env%ndim
         neb_env%nsize_int = number_of_colvar(f_env%force_env)
         check = (neb_env%nsize_xyz >= neb_env%nsize_int)
         IF(.NOT.(check))CALL cp__a("motion/neb_methods.F",155)
         ! Check that teh used colvar are uniquely determined
         check = (number_of_colvar(f_env%force_env) == &
                  number_of_colvar(f_env%force_env, unique=.TRUE.))
         IF(.NOT.(check))CALL cp__a("motion/neb_methods.F",159)
         CALL read_neb_section(neb_env, neb_section)
         ! Print BAND header
         iw2 = cp_print_key_unit_nr(logger, neb_section, "BANNER", extension=".nebLog")
         CALL band_header(iw2, neb_env%number_of_replica, nrep, prep)
         CALL cp_print_key_finished_output(iw2, logger, neb_section, "BANNER")
         ! Allocate the principal vectors used in the BAND calculation
         CALL neb_var_create(coords, neb_env, full_allocation=.TRUE.)
         CALL neb_var_create(forces, neb_env)
         CALL neb_var_create(vels, neb_env)
         ! Collecting the coordinates of the starting replicas of the BAND calculation
         IF (output_unit > 0) WRITE (output_unit, '(T2,"NEB|",A)') " Building initial set of coordinates. START"
         iw = cp_print_key_unit_nr(logger, neb_section, "PROGRAM_RUN_INFO/INITIAL_CONFIGURATION_INFO", &
                                   extension=".nebLog")
         CALL build_replica_coords(neb_section, particle_set, coords, vels, neb_env, iw, globenv, &
                                   rep_env%para_env)
         CALL cp_print_key_finished_output(iw, logger, neb_section, &
                                           "PROGRAM_RUN_INFO/INITIAL_CONFIGURATION_INFO")
         IF (output_unit > 0) WRITE (output_unit, '(T2,"NEB|",A)') " Building initial set of coordinates. END"
         ! Print some additional info in the replica_env initialization file
         CALL neb_rep_env_map_info(rep_env, neb_env)
         ! Perform NEB optimization
         SELECT CASE (neb_env%opt_type)
         CASE (band_md_opt)
            neb_env%opt_type_label = "MOLECULAR DYNAMICS"
            md_section => section_vals_get_subs_vals(neb_section, "OPTIMIZE_BAND%MD")
            CALL neb_md(rep_env, neb_env, coords, vels, forces, particle_set, output_unit, &
                        md_section, logger, globenv)
         CASE (band_diis_opt)
            neb_env%opt_type_label = "DIIS"
            diis_section => section_vals_get_subs_vals(neb_section, "OPTIMIZE_BAND%DIIS")
            CALL neb_diis(rep_env, neb_env, coords, vels, forces, particle_set, output_unit, &
                          diis_section, logger, globenv)
         END SELECT
         ! Release force_eval
         CALL f_env_rm_defaults(f_env, ierr)
         ! Release coords, vels and forces
         CALL neb_var_release(coords)
         CALL neb_var_release(forces)
         CALL neb_var_release(vels)
         ! At the end let's destroy the environment of the BAND calculation
         DEALLOCATE (neb_env)
      END IF
      CALL rep_env_release(rep_env)
      CALL cp_print_key_finished_output(output_unit, logger, neb_section, &
                                        "PROGRAM_RUN_INFO")
      CALL cp_rm_iter_level(logger%iter_info, "BAND")
      CALL timestop(handle)
   END SUBROUTINE neb

! **************************************************************************************************
!> \brief MD type optimization NEB
!> \param rep_env ...
!> \param neb_env ...
!> \param coords ...
!> \param vels ...
!> \param forces ...
!> \param particle_set ...
!> \param output_unit ...
!> \param md_section ...
!> \param logger ...
!> \param globenv ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE neb_md(rep_env, neb_env, coords, vels, forces, particle_set, output_unit, &
                     md_section, logger, globenv)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(neb_type), OPTIONAL, POINTER                  :: neb_env
      TYPE(neb_var_type), POINTER                        :: coords, vels, forces
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(section_vals_type), POINTER                   :: md_section
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'neb_md', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iatom, ic, is, istep, iw, &
                                                            max_steps, natom, shell_index
      LOGICAL                                            :: converged, should_stop
      REAL(KIND=dp)                                      :: dt
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: distances, energies
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: mass
      TYPE(neb_var_type), POINTER                        :: Dcoords
      TYPE(section_vals_type), POINTER                   :: tc_section, vc_section

      CALL timeset(routineN, handle)
      NULLIFY (Dcoords, tc_section, vc_section)
      IF(.NOT.(ASSOCIATED(coords)))CALL cp__a("motion/neb_methods.F",247)
      IF(.NOT.(ASSOCIATED(vels)))CALL cp__a("motion/neb_methods.F",248)
      ! MD band for string methods type does not make anywa sense. Stop calculation.
      IF (neb_env%id_type == do_sm) THEN
         CALL cp__w("motion/neb_methods.F",251,"MD band optimization and String Method incompatible.")
      END IF
      ! Output unit
      iw = cp_print_key_unit_nr(logger, neb_env%neb_section, "REPLICA_INFO", &
                                extension=".replicaLog")
      tc_section => section_vals_get_subs_vals(md_section, "TEMP_CONTROL")
      vc_section => section_vals_get_subs_vals(md_section, "VEL_CONTROL")
      CALL section_vals_val_get(md_section, "TIMESTEP", r_val=dt)
      CALL section_vals_val_get(md_section, "MAX_STEPS", i_val=max_steps)
      ! Initial setup for MD
      CALL neb_var_create(Dcoords, neb_env)
      ALLOCATE (mass(SIZE(coords%wrk, 1), neb_env%number_of_replica))
      ALLOCATE (energies(neb_env%number_of_replica))
      ALLOCATE (distances(neb_env%number_of_replica-1))
      ! Setting up the mass array
      IF (neb_env%use_colvar) THEN
         mass(:, :) = 0.5_dp*dt/massunit
      ELSE
         natom = SIZE(particle_set)
         DO iatom = 1, natom
            ic = 3*(iatom-1)
            shell_index = particle_set(iatom)%shell_index
            IF (shell_index == 0) THEN
               mass(ic+1:ic+3, :) = 0.5_dp*dt/particle_set(iatom)%atomic_kind%mass
            ELSE
               is = 3*(natom+shell_index-1)
               mass(ic+1:ic+3, :) = 0.5_dp*dt/particle_set(iatom)%atomic_kind%shell%mass_core
               mass(is+1:is+3, :) = 0.5_dp*dt/particle_set(iatom)%atomic_kind%shell%mass_shell
            END IF
         END DO
      END IF
      ! Initializing forces array
      CALL reorient_images(neb_env%rotate_frames, particle_set, coords, vels, &
                           output_unit, distances, neb_env%number_of_replica)
      neb_env%avg_distance = SQRT(SUM(distances*distances)/REAL(SIZE(distances), KIND=dp))
      CALL neb_calc_energy_forces(rep_env, neb_env, coords, energies, forces, &
                                  particle_set, iw)

      CALL dump_neb_info(neb_env=neb_env, &
                         coords=coords, &
                         vels=vels, &
                         forces=forces, &
                         particle_set=particle_set, &
                         logger=logger, &
                         istep=0, &
                         energies=energies, &
                         distances=distances, &
                         output_unit=output_unit)
      md_opt_loop: DO istep = 1, max_steps
         CALL cp_iterate(logger%iter_info, iter_nr=istep)
         ! Save the optimization step counter
         neb_env%istep = istep
         ! Velocity Verlet (first part)
         vels%wrk(:, :) = vels%wrk(:, :)+mass(:, :)*forces%wrk(:, :)
         ! Control on velocity - I part [rescale, annealing]
         CALL control_vels_a(vels, particle_set, tc_section, vc_section, output_unit, &
                             istep)
         ! Coordinate step
         Dcoords%wrk(:, :) = dt*vels%wrk(:, :)
         coords%wrk(:, :) = coords%wrk(:, :)+Dcoords%wrk(:, :)

         CALL reorient_images(neb_env%rotate_frames, particle_set, coords, vels, &
                              output_unit, distances, neb_env%number_of_replica)
         neb_env%avg_distance = SQRT(SUM(distances*distances)/REAL(SIZE(distances), KIND=dp))
         CALL neb_calc_energy_forces(rep_env, neb_env, coords, energies, forces, &
                                     particle_set, iw)
         ! Check for an external exit command
         CALL external_control(should_stop, "NEB", globenv=globenv)
         IF (should_stop) EXIT
         ! Control on velocity - II part [check vels VS forces, Steepest Descent like]
         CALL control_vels_b(vels, forces, vc_section)
         ! Velocity Verlet (second part)
         vels%wrk(:, :) = vels%wrk(:, :)+mass(:, :)*forces%wrk(:, :)
         ! Dump Infos
         CALL dump_neb_info(neb_env=neb_env, &
                            coords=coords, &
                            vels=vels, &
                            forces=forces, &
                            particle_set=particle_set, &
                            logger=logger, &
                            istep=istep, &
                            energies=energies, &
                            distances=distances, &
                            output_unit=output_unit)
         converged = check_convergence(neb_env, Dcoords, forces)
         IF (converged) EXIT
      END DO md_opt_loop

      DEALLOCATE (mass)
      DEALLOCATE (energies)
      DEALLOCATE (distances)
      CALL neb_var_release(Dcoords)
      CALL cp_print_key_finished_output(iw, logger, neb_env%neb_section, &
                                        "REPLICA_INFO")
      CALL timestop(handle)

   END SUBROUTINE neb_md

! **************************************************************************************************
!> \brief DIIS type optimization NEB
!> \param rep_env ...
!> \param neb_env ...
!> \param coords ...
!> \param vels ...
!> \param forces ...
!> \param particle_set ...
!> \param output_unit ...
!> \param diis_section ...
!> \param logger ...
!> \param globenv ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE neb_diis(rep_env, neb_env, coords, vels, forces, particle_set, output_unit, &
                       diis_section, logger, globenv)
      TYPE(replica_env_type), POINTER                    :: rep_env
      TYPE(neb_type), OPTIONAL, POINTER                  :: neb_env
      TYPE(neb_var_type), POINTER                        :: coords, vels, forces
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(section_vals_type), POINTER                   :: diis_section
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'neb_diis', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, istep, iw, iw2, max_sd_steps, &
                                                            max_steps, n_diis
      INTEGER, DIMENSION(:), POINTER                     :: set_err
      LOGICAL                                            :: check_diis, converged, diis_on, do_ls, &
                                                            should_stop, skip_ls
      REAL(KIND=dp)                                      :: max_stepsize, norm, stepsize, stepsize0
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: distances, energies
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: crr, err
      TYPE(neb_var_type), POINTER                        :: sline

      CALL timeset(routineN, handle)
      NULLIFY (sline, crr, err)
      neb_env%opt_type_label = "SD"
      do_ls = .TRUE.
      IF(.NOT.(ASSOCIATED(coords)))CALL cp__a("motion/neb_methods.F",390)
      IF(.NOT.(ASSOCIATED(vels)))CALL cp__a("motion/neb_methods.F",391)
      IF(.NOT.(ASSOCIATED(forces)))CALL cp__a("motion/neb_methods.F",392)
      iw = cp_print_key_unit_nr(logger, neb_env%neb_section, "REPLICA_INFO", &
                                extension=".replicaLog")
      CALL section_vals_val_get(diis_section, "MAX_STEPS", i_val=max_steps)
      CALL section_vals_val_get(diis_section, "N_DIIS", i_val=n_diis)
      CALL section_vals_val_get(diis_section, "STEPSIZE", r_val=stepsize0)
      CALL section_vals_val_get(diis_section, "MAX_STEPSIZE", r_val=max_stepsize)
      CALL section_vals_val_get(diis_section, "NO_LS", l_val=skip_ls)
      CALL section_vals_val_get(diis_section, "MAX_SD_STEPS", i_val=max_sd_steps)
      CALL section_vals_val_get(diis_section, "CHECK_DIIS", l_val=check_diis)
      iw2 = cp_print_key_unit_nr(logger, diis_section, "DIIS_INFO", &
                                 extension=".diisLog")
      ! Initial setup for DIIS
      stepsize = stepsize0
      ! Allocate type for Line Search direction
      CALL neb_var_create(sline, neb_env, full_allocation=.TRUE.)
      ! Array of error vectors
      ALLOCATE (err(PRODUCT(coords%size_wrk), n_diis))
      ALLOCATE (crr(PRODUCT(coords%size_wrk), n_diis))
      ALLOCATE (set_err(n_diis))
      ALLOCATE (energies(neb_env%number_of_replica))
      ALLOCATE (distances(neb_env%number_of_replica-1))
      ! Initializing forces array
      CALL reorient_images(neb_env%rotate_frames, particle_set, coords, vels, &
                           output_unit, distances, neb_env%number_of_replica)
      CALL reparametrize_images(neb_env%reparametrize_frames, neb_env%spline_order, &
                                neb_env%smoothing, coords%wrk, sline%wrk, distances)
      neb_env%avg_distance = SQRT(SUM(distances*distances)/REAL(SIZE(distances), KIND=dp))
      CALL neb_calc_energy_forces(rep_env, neb_env, coords, energies, forces, &
                                  particle_set, iw)
      ! Dump Infos
      CALL dump_neb_info(neb_env=neb_env, &
                         coords=coords, &
                         forces=forces, &
                         particle_set=particle_set, &
                         logger=logger, &
                         istep=0, &
                         energies=energies, &
                         distances=distances, &
                         vels=vels, &
                         output_unit=output_unit)
      ! If rotation is requested let's apply it at the beginning of the
      ! Geometry optimization and then let's disable it
      neb_env%rotate_frames = .FALSE.
      ! Main SD/DIIS loop
      set_err = -1
      DO istep = 1, max_steps
         CALL cp_iterate(logger%iter_info, iter_nr=istep)
         neb_env%opt_type_label = "SD"
         ! Save the optimization step counter
         neb_env%istep = istep
         ! Perform one step of SD with line search
         norm = SQRT(SUM(forces%wrk*forces%wrk))
         IF (norm < EPSILON(0.0_dp)) THEN
            ! Let's handle the case in which the band is already fully optimized
            converged = .TRUE.
            EXIT
         END IF
         sline%wrk = forces%wrk/norm
         IF (do_ls .AND. (.NOT. skip_ls)) THEN
            CALL neb_ls(stepsize, sline, rep_env, neb_env, coords, energies, forces, &
                        vels, particle_set, iw, output_unit, distances, diis_section, iw2)
            IF (iw2 > 0) &
               WRITE (iw2, '(T2,A,T69,F12.6)') "SD| Stepsize in SD after linesearch", &
               stepsize
         ELSE
            stepsize = MIN(norm*stepsize0, max_stepsize)
            IF (iw2 > 0) &
               WRITE (iw2, '(T2,A,T69,F12.6)') "SD| Stepsize in SD no linesearch performed", &
               stepsize
         END IF
         sline%wrk = stepsize*sline%wrk
         diis_on = accept_diis_step(istep > max_sd_steps, n_diis, err, crr, set_err, sline, coords, &
                                    check_diis, iw2)
         IF (diis_on) THEN
            neb_env%opt_type_label = "DIIS"
         END IF
         do_ls = .TRUE.
         IF (COUNT(set_err == -1) == 1) do_ls = .FALSE.
         coords%wrk = coords%wrk+sline%wrk
         ! Compute forces
         CALL reorient_images(neb_env%rotate_frames, particle_set, coords, vels, &
                              output_unit, distances, neb_env%number_of_replica)
         CALL reparametrize_images(neb_env%reparametrize_frames, neb_env%spline_order, &
                                   neb_env%smoothing, coords%wrk, sline%wrk, distances)
         neb_env%avg_distance = SQRT(SUM(distances*distances)/REAL(SIZE(distances), KIND=dp))
         CALL neb_calc_energy_forces(rep_env, neb_env, coords, energies, forces, &
                                     particle_set, iw)
         ! Check for an external exit command
         CALL external_control(should_stop, "NEB", globenv=globenv)
         IF (should_stop) EXIT
         ! Dump Infos
         CALL dump_neb_info(neb_env=neb_env, &
                            coords=coords, &
                            forces=forces, &
                            particle_set=particle_set, &
                            logger=logger, &
                            istep=istep, &
                            energies=energies, &
                            distances=distances, &
                            vels=vels, &
                            output_unit=output_unit)

         converged = check_convergence(neb_env, sline, forces)
         IF (converged) EXIT
      END DO
      DEALLOCATE (energies)
      DEALLOCATE (distances)
      DEALLOCATE (err)
      DEALLOCATE (crr)
      DEALLOCATE (set_err)
      CALL neb_var_release(sline)
      CALL timestop(handle)
   END SUBROUTINE neb_diis

END MODULE neb_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE atom_basis
   USE atom_fit,                        ONLY: atom_fit_basis
   USE atom_output,                     ONLY: atom_print_basis,&
                                              atom_print_info,&
                                              atom_print_method,&
                                              atom_print_potential
   USE atom_types,                      ONLY: &
        atom_basis_type, atom_integrals, atom_optimization_type, atom_orbitals, atom_p_type, &
        atom_potential_type, atom_state, create_atom_orbs, create_atom_type, init_atom_basis, &
        init_atom_potential, read_atom_opt_section, release_atom_basis, release_atom_potential, &
        release_atom_type, set_atom
   USE atom_utils,                      ONLY: atom_consistent_method,&
                                              atom_set_occupation,&
                                              get_maxl_occ,&
                                              get_maxn_occ
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_constants,                 ONLY: do_analytic
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE periodic_table,                  ONLY: nelem,&
                                              ptable
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC  :: atom_basis_opt

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_basis'

! **************************************************************************************************

CONTAINS

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param atom_section ...
! **************************************************************************************************
   SUBROUTINE atom_basis_opt(atom_section)
      TYPE(section_vals_type), POINTER                   :: atom_section

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_basis_opt', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: elem
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER                                            :: do_eric, do_erie, handle, i, im, in, &
                                                            iunit, iw, k, maxl, mb, method, mo, &
                                                            n_meth, n_rep, reltyp, zcore, zval, zz
      INTEGER, DIMENSION(0:3)                            :: maxn
      INTEGER, DIMENSION(:), POINTER                     :: cn
      LOGICAL                                            :: eri_c, eri_e, had_ae, had_pp, pp_calc
      REAL(KIND=dp), DIMENSION(0:3, 10)                  :: pocc
      TYPE(atom_basis_type), POINTER                     :: ae_basis, pp_basis
      TYPE(atom_integrals), POINTER                      :: ae_int, pp_int
      TYPE(atom_optimization_type)                       :: optimization
      TYPE(atom_orbitals), POINTER                       :: orbitals
      TYPE(atom_p_type), DIMENSION(:, :), POINTER        :: atom_info
      TYPE(atom_potential_type), POINTER                 :: ae_pot, p_pot
      TYPE(atom_state), POINTER                          :: state
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: basis_section, method_section, &
                                                            opt_section, potential_section, &
                                                            powell_section, xc_section

      CALL timeset(routineN, handle)

      ! What atom do we calculate
      CALL section_vals_val_get(atom_section, "ATOMIC_NUMBER", i_val=zval)
      CALL section_vals_val_get(atom_section, "ELEMENT", c_val=elem)
      zz = 0
      DO i = 1, nelem
         IF (ptable(i)%symbol == elem) THEN
            zz = i
            EXIT
         END IF
      END DO
      IF (zz /= 1) zval = zz

      ! read and set up inofrmation on the basis sets
      ALLOCATE (ae_basis, pp_basis)
      basis_section => section_vals_get_subs_vals(atom_section, "AE_BASIS")
      NULLIFY (ae_basis%grid)
      CALL init_atom_basis(ae_basis, basis_section, zval, "AE")
      NULLIFY (pp_basis%grid)
      basis_section => section_vals_get_subs_vals(atom_section, "PP_BASIS")
      CALL init_atom_basis(pp_basis, basis_section, zval, "PP")

      ! print general and basis set information
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%PROGRAM_BANNER", extension=".log")
      IF (iw > 0) CALL atom_print_info(zval, "Atomic Basis Optimization", iw)
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%PROGRAM_BANNER")

      ! read and setup information on the pseudopotential
      NULLIFY (potential_section)
      potential_section => section_vals_get_subs_vals(atom_section, "POTENTIAL")
      ALLOCATE (ae_pot, p_pot)
      CALL init_atom_potential(p_pot, potential_section, zval)
      CALL init_atom_potential(ae_pot, potential_section, -1)

      ! if the ERI's are calculated analytically, we have to precalculate them
      eri_c = .FALSE.
      CALL section_vals_val_get(atom_section, "COULOMB_INTEGRALS", i_val=do_eric)
      IF (do_eric == do_analytic) eri_c = .TRUE.
      eri_e = .FALSE.
      CALL section_vals_val_get(atom_section, "EXCHANGE_INTEGRALS", i_val=do_erie)
      IF (do_erie == do_analytic) eri_e = .TRUE.

      ! information on the states to be calculated
      CALL section_vals_val_get(atom_section, "MAX_ANGULAR_MOMENTUM", i_val=maxl)
      maxn = 0
      CALL section_vals_val_get(atom_section, "CALCULATE_STATES", i_vals=cn)
      DO in = 1, MIN(SIZE(cn), 4)
         maxn(in-1) = cn(in)
      END DO
      DO in = 0, 3
         maxn(in) = MIN(maxn(in), ae_basis%nbas(in))
         maxn(in) = MIN(maxn(in), pp_basis%nbas(in))
      END DO

      ! read optimization section
      opt_section => section_vals_get_subs_vals(atom_section, "OPTIMIZATION")
      CALL read_atom_opt_section(optimization, opt_section)

      had_ae = .FALSE.
      had_pp = .FALSE.

      ! Check for the total number of electron configurations to be calculated
      CALL section_vals_val_get(atom_section, "ELECTRON_CONFIGURATION", n_rep_val=n_rep)
      ! Check for the total number of method types to be calculated
      method_section => section_vals_get_subs_vals(atom_section, "METHOD")
      CALL section_vals_get(method_section, n_repetition=n_meth)

      ! integrals
      ALLOCATE (ae_int, pp_int)

      ALLOCATE (atom_info(n_rep, n_meth))

      DO in = 1, n_rep
         DO im = 1, n_meth

            NULLIFY (atom_info(in, im)%atom)
            CALL create_atom_type(atom_info(in, im)%atom)

            atom_info(in, im)%atom%optimization = optimization

            atom_info(in, im)%atom%z = zval
            xc_section => section_vals_get_subs_vals(method_section, "XC", i_rep_section=im)
            atom_info(in, im)%atom%xc_section => xc_section

            ALLOCATE (state)

            ! get the electronic configuration
            CALL section_vals_val_get(atom_section, "ELECTRON_CONFIGURATION", i_rep_val=in, &
                                      c_vals=tmpstringlist)

            ! set occupations
            CALL atom_set_occupation(tmpstringlist, state%occ, state%occupation, state%multiplicity)
            state%maxl_occ = get_maxl_occ(state%occ)
            state%maxn_occ = get_maxn_occ(state%occ)

            ! set number of states to be calculated
            state%maxl_calc = MAX(maxl, state%maxl_occ)
            state%maxl_calc = MIN(3, state%maxl_calc)
            state%maxn_calc = 0
            DO k = 0, state%maxl_calc
               state%maxn_calc(k) = MAX(maxn(k), state%maxn_occ(k))
            END DO

            ! is there a pseudo potential
            pp_calc = ANY(INDEX(tmpstringlist(1:), "CORE") /= 0)
            IF (pp_calc) THEN
               ! get and set the core occupations
               CALL section_vals_val_get(atom_section, "CORE", c_vals=tmpstringlist)
               CALL atom_set_occupation(tmpstringlist, state%core, pocc)
               zcore = zval-NINT(SUM(state%core))
               CALL set_atom(atom_info(in, im)%atom, zcore=zcore, pp_calc=.TRUE.)
               had_pp = .TRUE.
               CALL set_atom(atom_info(in, im)%atom, basis=pp_basis, potential=p_pot)
               state%maxn_calc(:) = MIN(state%maxn_calc(:), pp_basis%nbas(:))
               IF(.NOT.(ALL(state%maxn_calc(:) >= state%maxn_occ)))CALL cp__a("atom_basis.F",196)
            ELSE
               state%core = 0._dp
               CALL set_atom(atom_info(in, im)%atom, zcore=zval, pp_calc=.FALSE.)
               had_ae = .TRUE.
               CALL set_atom(atom_info(in, im)%atom, basis=ae_basis, potential=ae_pot)
               state%maxn_calc(:) = MIN(state%maxn_calc(:), ae_basis%nbas(:))
               IF(.NOT.(ALL(state%maxn_calc(:) >= state%maxn_occ)))CALL cp__a("atom_basis.F",203)
            END IF

            CALL section_vals_val_get(method_section, "METHOD_TYPE", i_val=method, i_rep_val=im)
            CALL section_vals_val_get(method_section, "RELATIVISTIC", i_val=reltyp, i_rep_section=im)
            CALL set_atom(atom_info(in, im)%atom, method_type=method, relativistic=reltyp)
            CALL set_atom(atom_info(in, im)%atom, state=state)
            CALL set_atom(atom_info(in, im)%atom, coulomb_integral_type=do_eric, &
                          exchange_integral_type=do_erie)

            IF (atom_consistent_method(method, state%multiplicity)) THEN
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%METHOD_INFO", extension=".log")
               CALL atom_print_method(atom_info(in, im)%atom, iw)
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%METHOD_INFO")
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%POTENTIAL", extension=".log")
               IF (pp_calc) THEN
                  IF (iw > 0) CALL atom_print_potential(p_pot, iw)
               ELSE
                  IF (iw > 0) CALL atom_print_potential(ae_pot, iw)
               END IF
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%POTENTIAL")
            ELSE
               CALL cp__b("atom_basis.F",225,"METHOD_TYPE and MULTIPLICITY are incompatible")
            END IF

            NULLIFY (orbitals)
            mo = MAXVAL(state%maxn_calc)
            mb = MAXVAL(atom_info(in, im)%atom%basis%nbas)
            CALL create_atom_orbs(orbitals, mb, mo)
            CALL set_atom(atom_info(in, im)%atom, orbitals=orbitals)

         END DO
      END DO

      ! Start the Optimization
      powell_section => section_vals_get_subs_vals(atom_section, "POWELL")
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%SCF_INFO", extension=".log")
      iunit = cp_print_key_unit_nr(logger, atom_section, "PRINT%FIT_BASIS", extension=".log")
      IF (had_ae) THEN
         pp_calc = .FALSE.
         CALL atom_fit_basis(atom_info, ae_basis, pp_calc, iunit, powell_section)
      END IF
      IF (had_pp) THEN
         pp_calc = .TRUE.
         CALL atom_fit_basis(atom_info, pp_basis, pp_calc, iunit, powell_section)
      END IF
      CALL cp_print_key_finished_output(iunit, logger, atom_section, "PRINT%FIT_BASIS")
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%SCF_INFO")
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%BASIS_SET", extension=".log")
      IF (iw > 0) THEN
         CALL atom_print_basis(ae_basis, iw, " All Electron Basis")
         CALL atom_print_basis(pp_basis, iw, " Pseudopotential Basis")
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%BASIS_SET")

      CALL release_atom_basis(ae_basis)
      CALL release_atom_basis(pp_basis)

      CALL release_atom_potential(p_pot)
      CALL release_atom_potential(ae_pot)

      DO in = 1, n_rep
         DO im = 1, n_meth
            CALL release_atom_type(atom_info(in, im)%atom)
         END DO
      END DO
      DEALLOCATE (atom_info)

      DEALLOCATE (ae_pot, p_pot, ae_basis, pp_basis, ae_int, pp_int)

      CALL timestop(handle)

   END SUBROUTINE atom_basis_opt

END MODULE atom_basis
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE atom_energy
   USE atom_electronic_structure,       ONLY: calculate_atom
   USE atom_fit,                        ONLY: atom_fit_density,&
                                              atom_fit_kgpot
   USE atom_operators,                  ONLY: atom_int_release,&
                                              atom_int_setup,&
                                              atom_ppint_release,&
                                              atom_ppint_setup,&
                                              atom_relint_release,&
                                              atom_relint_setup
   USE atom_output,                     ONLY: atom_print_basis,&
                                              atom_print_info,&
                                              atom_print_method,&
                                              atom_print_orbitals,&
                                              atom_print_potential,&
                                              atom_write_pseudo_param
   USE atom_types,                      ONLY: &
        atom_basis_type, atom_gthpot_type, atom_integrals, atom_optimization_type, atom_orbitals, &
        atom_p_type, atom_potential_type, atom_state, atom_type, create_atom_orbs, &
        create_atom_type, gth_pseudo, init_atom_basis, init_atom_potential, read_atom_opt_section, &
        release_atom_basis, release_atom_potential, release_atom_type, set_atom
   USE atom_utils,                      ONLY: &
        atom_completeness, atom_condnumber, atom_consistent_method, atom_core_density, &
        atom_density, atom_local_potential, atom_read_external_density, atom_read_external_vxc, &
        atom_set_occupation, atom_write_zmp_restart, get_maxl_occ, get_maxn_occ
   USE atom_xc,                         ONLY: calculate_atom_ext_vxc,&
                                              calculate_atom_zmp
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_constants,                 ONLY: do_analytic
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: dfac,&
                                              gamma1,&
                                              pi
   USE periodic_table,                  ONLY: nelem,&
                                              ptable
   USE physcon,                         ONLY: bohr,&
                                              evolt
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC  :: atom_energy_opt

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_energy'

! **************************************************************************************************

CONTAINS

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param atom_section ...
! **************************************************************************************************
   SUBROUTINE atom_energy_opt(atom_section)
      TYPE(section_vals_type), POINTER                   :: atom_section

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_energy_opt', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: elem
      CHARACTER(LEN=default_string_length)               :: filename
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER :: do_eric, do_erie, handle, i, im, in, iw, k, l, maxl, mb, method, mo, n, n_meth, &
         n_rep, nder, num_gau, num_gto, num_pol, reltyp, zcore, zval, zz
      INTEGER, DIMENSION(0:3)                            :: maxn
      INTEGER, DIMENSION(:), POINTER                     :: cn
      LOGICAL                                            :: dm, do_zmp, doread, eri_c, eri_e, &
                                                            had_ae, had_pp, lcomp, lcond, pp_calc, &
                                                            read_vxc
      REAL(KIND=dp)                                      :: crad, delta, lambda
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ext_density, ext_vxc
      REAL(KIND=dp), DIMENSION(0:3, 10)                  :: pocc
      TYPE(atom_basis_type), POINTER                     :: ae_basis, pp_basis
      TYPE(atom_integrals), POINTER                      :: ae_int, pp_int
      TYPE(atom_optimization_type)                       :: optimization
      TYPE(atom_orbitals), POINTER                       :: orbitals
      TYPE(atom_p_type), DIMENSION(:, :), POINTER        :: atom_info
      TYPE(atom_potential_type), POINTER                 :: ae_pot, p_pot
      TYPE(atom_state), POINTER                          :: state
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER :: basis_section, external_vxc_section, method_section, &
         opt_section, potential_section, powell_section, xc_section, zmp_restart_section, &
         zmp_section

      CALL timeset(routineN, handle)

      ! What atom do we calculate
      CALL section_vals_val_get(atom_section, "ATOMIC_NUMBER", i_val=zval)
      CALL section_vals_val_get(atom_section, "ELEMENT", c_val=elem)
      zz = 0
      DO i = 1, nelem
         IF (ptable(i)%symbol == elem) THEN
            zz = i
            EXIT
         END IF
      END DO
      IF (zz /= 1) zval = zz

      ! read and set up inofrmation on the basis sets
      ALLOCATE (ae_basis, pp_basis)
      basis_section => section_vals_get_subs_vals(atom_section, "AE_BASIS")
      NULLIFY (ae_basis%grid)
      CALL init_atom_basis(ae_basis, basis_section, zval, "AE")
      NULLIFY (pp_basis%grid)
      basis_section => section_vals_get_subs_vals(atom_section, "PP_BASIS")
      CALL init_atom_basis(pp_basis, basis_section, zval, "PP")

      ! print general and basis set information
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%PROGRAM_BANNER", extension=".log")
      IF (iw > 0) CALL atom_print_info(zval, "Atomic Energy Calculation", iw)
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%PROGRAM_BANNER")
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%BASIS_SET", extension=".log")
      IF (iw > 0) THEN
         CALL atom_print_basis(ae_basis, iw, " All Electron Basis")
         CALL atom_print_basis(pp_basis, iw, " Pseudopotential Basis")
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%BASIS_SET")

      ! read and setup information on the pseudopotential
      NULLIFY (potential_section)
      potential_section => section_vals_get_subs_vals(atom_section, "POTENTIAL")
      ALLOCATE (ae_pot, p_pot)
      CALL init_atom_potential(p_pot, potential_section, zval)
      CALL init_atom_potential(ae_pot, potential_section, -1)

      ! if the ERI's are calculated analytically, we have to precalculate them
      eri_c = .FALSE.
      CALL section_vals_val_get(atom_section, "COULOMB_INTEGRALS", i_val=do_eric)
      IF (do_eric == do_analytic) eri_c = .TRUE.
      eri_e = .FALSE.
      CALL section_vals_val_get(atom_section, "EXCHANGE_INTEGRALS", i_val=do_erie)
      IF (do_erie == do_analytic) eri_e = .TRUE.

      ! information on the states to be calculated
      CALL section_vals_val_get(atom_section, "MAX_ANGULAR_MOMENTUM", i_val=maxl)
      maxn = 0
      CALL section_vals_val_get(atom_section, "CALCULATE_STATES", i_vals=cn)
      DO in = 1, MIN(SIZE(cn), 4)
         maxn(in-1) = cn(in)
      END DO
      DO in = 0, 3
         maxn(in) = MIN(maxn(in), ae_basis%nbas(in))
         maxn(in) = MIN(maxn(in), pp_basis%nbas(in))
      END DO

      ! read optimization section
      opt_section => section_vals_get_subs_vals(atom_section, "OPTIMIZATION")
      CALL read_atom_opt_section(optimization, opt_section)

      had_ae = .FALSE.
      had_pp = .FALSE.

      ! Check for the total number of electron configurations to be calculated
      CALL section_vals_val_get(atom_section, "ELECTRON_CONFIGURATION", n_rep_val=n_rep)
      ! Check for the total number of method types to be calculated
      method_section => section_vals_get_subs_vals(atom_section, "METHOD")
      CALL section_vals_get(method_section, n_repetition=n_meth)

      ! integrals
      ALLOCATE (ae_int, pp_int)

      ALLOCATE (atom_info(n_rep, n_meth))

      DO in = 1, n_rep
         DO im = 1, n_meth

            NULLIFY (atom_info(in, im)%atom)
            CALL create_atom_type(atom_info(in, im)%atom)

            atom_info(in, im)%atom%optimization = optimization

            atom_info(in, im)%atom%z = zval
            xc_section => section_vals_get_subs_vals(method_section, "XC", i_rep_section=im)
            atom_info(in, im)%atom%xc_section => xc_section

            ! ZMP Reading input sections if they are found initialize everything
            do_zmp = .FALSE.
            doread = .FALSE.
            read_vxc = .FALSE.

            zmp_section => section_vals_get_subs_vals(method_section, "ZMP")
            CALL section_vals_get(zmp_section, explicit=do_zmp)
            atom_info(in, im)%atom%do_zmp = do_zmp
            CALL section_vals_val_get(zmp_section, "FILE_DENSITY", c_val=filename)
            atom_info(in, im)%atom%ext_file = filename
            CALL section_vals_val_get(zmp_section, "GRID_TOL", &
                                      r_val=atom_info(in, im)%atom%zmpgrid_tol)
            CALL section_vals_val_get(zmp_section, "LAMBDA", r_val=lambda)
            atom_info(in, im)%atom%lambda = lambda
            CALL section_vals_val_get(zmp_section, "DM", l_val=dm)
            atom_info(in, im)%atom%dm = dm

            zmp_restart_section => section_vals_get_subs_vals(zmp_section, "RESTART")
            CALL section_vals_get(zmp_restart_section, explicit=doread)
            atom_info(in, im)%atom%doread = doread
            CALL section_vals_val_get(zmp_restart_section, "FILE_RESTART", c_val=filename)
            atom_info(in, im)%atom%zmp_restart_file = filename

            ! ZMP Reading external vxc section, if found initialize
            external_vxc_section => section_vals_get_subs_vals(method_section, "EXTERNAL_VXC")
            CALL section_vals_get(external_vxc_section, explicit=read_vxc)
            atom_info(in, im)%atom%read_vxc = read_vxc
            CALL section_vals_val_get(external_vxc_section, "FILE_VXC", c_val=filename)
            atom_info(in, im)%atom%ext_vxc_file = filename
            CALL section_vals_val_get(external_vxc_section, "GRID_TOL", &
                                      r_val=atom_info(in, im)%atom%zmpvxcgrid_tol)

            ALLOCATE (state)

            ! get the electronic configuration
            CALL section_vals_val_get(atom_section, "ELECTRON_CONFIGURATION", i_rep_val=in, &
                                      c_vals=tmpstringlist)

            ! set occupations
            CALL atom_set_occupation(tmpstringlist, state%occ, state%occupation, state%multiplicity)
            state%maxl_occ = get_maxl_occ(state%occ)
            state%maxn_occ = get_maxn_occ(state%occ)

            ! set number of states to be calculated
            state%maxl_calc = MAX(maxl, state%maxl_occ)
            state%maxl_calc = MIN(3, state%maxl_calc)
            state%maxn_calc = 0
            DO k = 0, state%maxl_calc
               state%maxn_calc(k) = MAX(maxn(k), state%maxn_occ(k))
            END DO

            ! is there a pseudo potential
            pp_calc = ANY(INDEX(tmpstringlist(1:), "CORE") /= 0)
            IF (pp_calc) THEN
               ! get and set the core occupations
               CALL section_vals_val_get(atom_section, "CORE", c_vals=tmpstringlist)
               CALL atom_set_occupation(tmpstringlist, state%core, pocc)
               zcore = zval-NINT(SUM(state%core))
               CALL set_atom(atom_info(in, im)%atom, zcore=zcore, pp_calc=.TRUE.)
            ELSE
               state%core = 0._dp
               CALL set_atom(atom_info(in, im)%atom, zcore=zval, pp_calc=.FALSE.)
            END IF

            CALL section_vals_val_get(method_section, "METHOD_TYPE", i_val=method, i_rep_section=im)
            CALL section_vals_val_get(method_section, "RELATIVISTIC", i_val=reltyp, i_rep_section=im)
            CALL set_atom(atom_info(in, im)%atom, method_type=method, relativistic=reltyp)

            IF (atom_consistent_method(method, state%multiplicity)) THEN
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%METHOD_INFO", extension=".log")
               CALL atom_print_method(atom_info(in, im)%atom, iw)
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%METHOD_INFO")

               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%POTENTIAL", extension=".log")
               IF (pp_calc) THEN
                  IF (iw > 0) CALL atom_print_potential(p_pot, iw)
               ELSE
                  IF (iw > 0) CALL atom_print_potential(ae_pot, iw)
               END IF
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%POTENTIAL")
            END IF

            ! calculate integrals
            IF (pp_calc) THEN
               ! general integrals
               CALL atom_int_setup(pp_int, pp_basis, &
                                   potential=p_pot, eri_coulomb=eri_c, eri_exchange=eri_e)
               ! potential
               CALL atom_ppint_setup(pp_int, pp_basis, potential=p_pot)
               !
               NULLIFY (pp_int%tzora, pp_int%hdkh)
               !
               CALL set_atom(atom_info(in, im)%atom, basis=pp_basis, integrals=pp_int, potential=p_pot)
               state%maxn_calc(:) = MIN(state%maxn_calc(:), pp_basis%nbas(:))
               IF(.NOT.(ALL(state%maxn_calc(:) >= state%maxn_occ)))CALL cp__a("atom_energy.F",288)
               had_pp = .TRUE.
            ELSE
               ! general integrals
               CALL atom_int_setup(ae_int, ae_basis, potential=ae_pot, &
                                   eri_coulomb=eri_c, eri_exchange=eri_e)
               ! potential
               CALL atom_ppint_setup(ae_int, ae_basis, potential=ae_pot)
               ! relativistic correction terms
               CALL atom_relint_setup(ae_int, ae_basis, reltyp, zcore=REAL(zval, dp))
               !
               CALL set_atom(atom_info(in, im)%atom, basis=ae_basis, integrals=ae_int, potential=ae_pot)
               state%maxn_calc(:) = MIN(state%maxn_calc(:), ae_basis%nbas(:))
               IF(.NOT.(ALL(state%maxn_calc(:) >= state%maxn_occ)))CALL cp__a("atom_energy.F",301)
               had_ae = .TRUE.
            END IF

            CALL set_atom(atom_info(in, im)%atom, state=state)

            CALL set_atom(atom_info(in, im)%atom, coulomb_integral_type=do_eric, &
                          exchange_integral_type=do_erie)

            NULLIFY (orbitals)
            mo = MAXVAL(state%maxn_calc)
            mb = MAXVAL(atom_info(in, im)%atom%basis%nbas)
            CALL create_atom_orbs(orbitals, mb, mo)
            CALL set_atom(atom_info(in, im)%atom, orbitals=orbitals)

            IF (atom_consistent_method(method, state%multiplicity)) THEN
               !Calculate the electronic structure
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%SCF_INFO", extension=".log")
               CALL calculate_atom(atom_info(in, im)%atom, iw)
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%SCF_INFO")

               ! ZMP If we have the external density do zmp
               IF (atom_info(in, im)%atom%do_zmp) THEN
                  CALL atom_write_zmp_restart(atom_info(in, im)%atom)

                  ALLOCATE (ext_density(atom_info(in, im)%atom%basis%grid%nr))
                  ext_density = 0._dp
                  CALL atom_read_external_density(ext_density, atom_info(in, im)%atom, iw)
                  CALL calculate_atom_zmp(ext_density=ext_density, &
                                          atom=atom_info(in, im)%atom, &
                                          lprint=.TRUE.)
                  DEALLOCATE (ext_density)
               ENDIF
               ! ZMP If we have the external v_xc calculate KS quantities
               IF (atom_info(in, im)%atom%read_vxc) THEN
                  ALLOCATE (ext_vxc(atom_info(in, im)%atom%basis%grid%nr))
                  ext_vxc = 0._dp
                  CALL atom_read_external_vxc(ext_vxc, atom_info(in, im)%atom, iw)
                  CALL calculate_atom_ext_vxc(vxc=ext_vxc, &
                                              atom=atom_info(in, im)%atom, &
                                              lprint=.TRUE.)
                  DEALLOCATE (ext_vxc)
               ENDIF

               ! rk: Print out the energies needed for the scptb parametrization
               ! (Orbital energies minus Kinetic, etc)
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%ENERGIES_MINUS_KINETIC", extension=".log")
               IF (iw > 0) THEN
                  WRITE (iw, '(T36,A,T61,F20.12)') "     Etot-Ekin   ::", atom_info(in, im)%atom%energy%etot &
                     -atom_info(in, im)%atom%energy%ekin

                  ! Output of Epsilon-kin, used in SCPTB parameter generation.
                  IF (atom_info(in, im)%atom%state%multiplicity == -1) THEN
                     WRITE (iw, '(/,A,T20,A,T30,A,T36,A,T49,A,T71,A,/)') " Orbital Epsilon-T", &
                        "State", "L", "Occupation", "Energy[a.u.]", "Energy[eV]"
                     DO l = 0, atom_info(in, im)%atom%state%maxl_calc
                        n = atom_info(in, im)%atom%state%maxn_calc(l)
                        DO i = 1, n
                           WRITE (iw, '(T23,I2,T30,I1,T36,F10.3,T46,F15.6,T66,F15.6)') &
                              i, l, atom_info(in, im)%atom%state%occupation(l, i), &
                              atom_info(in, im)%atom%orbitals%ener(i, l)-atom_info(in, im)%atom%integrals%kin(i, i, l), &
                              (atom_info(in, im)%atom%orbitals%ener(i, l)-atom_info(in, im)%atom%integrals%kin(i, i, l))*evolt
                        END DO
                        IF (n > 0) WRITE (iw, *)
                     END DO
                  ELSE
                     WRITE (iw, '(/,A,T20,A,T30,A,T36,A,T42,A,T55,A,T71,A,/)') " Orbital Epsilon-T", &
                        "State", "Spin", "L", "Occupation", "Energy[a.u.]", "Energy[eV]"
                     DO l = 0, atom_info(in, im)%atom%state%maxl_calc
                        n = atom_info(in, im)%atom%state%maxn_calc(l)
                        DO i = 1, n
                           WRITE (iw, '(T23,I2,T29,A,T36,I1,T42,F10.3,T52,F15.6,T68,F13.6)') &
                              i, "alpha", l, atom_info(in, im)%atom%state%occa(l, i), &
                              atom_info(in, im)%atom%orbitals%enera(i, l)-atom_info(in, im)%atom%integrals%kin(i, i, l), &
                              (atom_info(in, im)%atom%orbitals%ener(i, l)-atom_info(in, im)%atom%integrals%kin(i, i, l))*evolt
                        END DO
                        DO i = 1, n
                           WRITE (iw, '(T23,I2,T29,A,T36,I1,T42,F10.3,T52,F15.6,T68,F13.6)') &
                              i, " beta", l, atom_info(in, im)%atom%state%occb(l, i), &
                              atom_info(in, im)%atom%orbitals%enerb(i, l)-atom_info(in, im)%atom%integrals%kin(i, i, l), &
                              (atom_info(in, im)%atom%orbitals%ener(i, l)-atom_info(in, im)%atom%integrals%kin(i, i, l))*evolt
                        END DO
                        IF (n > 0) WRITE (iw, *)
                     END DO
                  END IF
               END IF

               ! Print out the orbitals if requested
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%ORBITALS", extension=".log")
               IF (iw > 0) THEN
                  CALL atom_print_orbitals(atom_info(in, im)%atom, iw)
               END IF
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%ORBITALS")

               ! perform a fit of the total electronic density
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%FIT_DENSITY", extension=".log")
               IF (iw > 0) THEN
                  CALL section_vals_val_get(atom_section, "PRINT%FIT_DENSITY%NUM_GTO", i_val=num_gto)
                  powell_section => section_vals_get_subs_vals(atom_section, "POWELL")
                  CALL atom_fit_density(atom_info(in, im)%atom, num_gto, 0, iw, powell_section=powell_section)
               END IF
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%FIT_DENSITY")

               ! Optimize a local potential for the non-additive kinetic energy term in KG
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%FIT_KGPOT", extension=".log")
               IF (iw > 0) THEN
                  CALL section_vals_val_get(atom_section, "PRINT%FIT_KGPOT%NUM_GAUSSIAN", i_val=num_gau)
                  CALL section_vals_val_get(atom_section, "PRINT%FIT_KGPOT%NUM_POLYNOM", i_val=num_pol)
                  powell_section => section_vals_get_subs_vals(atom_section, "POWELL")
                  CALL atom_fit_kgpot(atom_info(in, im)%atom, num_gau, num_pol, iw, powell_section=powell_section)
               END IF
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%FIT_KGPOT")

               ! generate a response basis
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%RESPONSE_BASIS", extension=".log")
               IF (iw > 0) THEN
                  CALL section_vals_val_get(atom_section, "PRINT%RESPONSE_BASIS%DELTA_CHARGE", r_val=delta)
                  CALL section_vals_val_get(atom_section, "PRINT%RESPONSE_BASIS%DERIVATIVES", i_val=nder)
                  CALL atom_response_basis(atom_info(in, im)%atom, delta, nder, iw)
               END IF
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%RESPONSE_BASIS")

               ! generate a UPF file
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%UPF_FILE", extension=".upf", &
                                         file_position="REWIND")
               IF (iw > 0) THEN
                  CALL atom_write_upf(atom_info(in, im)%atom, iw)
               END IF
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%UPF_FILE")
            END IF

         END DO
      END DO

      ! Analyze basis sets
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%ANALYZE_BASIS", extension=".log")
      IF (iw > 0) THEN
         CALL section_vals_val_get(atom_section, "PRINT%ANALYZE_BASIS%OVERLAP_CONDITION_NUMBER", l_val=lcond)
         CALL section_vals_val_get(atom_section, "PRINT%ANALYZE_BASIS%COMPLETENESS", l_val=lcomp)
         crad = ptable(zval)%covalent_radius*bohr
         IF (had_ae) THEN
            IF (lcond) CALL atom_condnumber(ae_basis, crad, iw)
            IF (lcomp) CALL atom_completeness(ae_basis, zval, iw)
         END IF
         IF (had_pp) THEN
            IF (lcond) CALL atom_condnumber(pp_basis, crad, iw)
            IF (lcomp) CALL atom_completeness(pp_basis, zval, iw)
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%ANALYZE_BASIS")

      ! clean up
      IF (had_ae) THEN
         CALL atom_int_release(ae_int)
         CALL atom_ppint_release(ae_int)
         CALL atom_relint_release(ae_int)
      END IF
      IF (had_pp) THEN
         CALL atom_int_release(pp_int)
         CALL atom_ppint_release(pp_int)
         CALL atom_relint_release(pp_int)
      END IF
      CALL release_atom_basis(ae_basis)
      CALL release_atom_basis(pp_basis)

      CALL release_atom_potential(p_pot)
      CALL release_atom_potential(ae_pot)

      DO in = 1, n_rep
         DO im = 1, n_meth
            CALL release_atom_type(atom_info(in, im)%atom)
         END DO
      END DO
      DEALLOCATE (atom_info)

      DEALLOCATE (ae_pot, p_pot, ae_basis, pp_basis, ae_int, pp_int)

      CALL timestop(handle)

   END SUBROUTINE atom_energy_opt

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param delta ...
!> \param nder ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_response_basis(atom, delta, nder, iw)

      TYPE(atom_type), POINTER                           :: atom
      REAL(KIND=dp), INTENT(IN)                          :: delta
      INTEGER, INTENT(IN)                                :: nder, iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_response_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ider, j, k, l, lhomo, m, n, nhomo, &
                                                            s1, s2
      REAL(KIND=dp)                                      :: dene, emax, expzet, fhomo, o, prefac, &
                                                            zeta
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: amat
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rbasis
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: wfn
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: ovlp
      TYPE(atom_state), POINTER                          :: state

      WRITE (iw, '(/," ",79("*"),/,T34,A,/," ",79("*"))') "RESPONSE BASIS"

      state => atom%state
      ovlp => atom%integrals%ovlp

      ! find HOMO
      lhomo = -1
      nhomo = -1
      emax = -HUGE(1._dp)
      DO l = 0, state%maxl_occ
         DO i = 1, state%maxn_occ(l)
            IF (atom%orbitals%ener(i, l) > emax) THEN
               lhomo = l
               nhomo = i
               emax = atom%orbitals%ener(i, l)
               fhomo = state%occupation(l, i)
            END IF
         END DO
      END DO

      s1 = SIZE(atom%orbitals%wfn, 1)
      s2 = SIZE(atom%orbitals%wfn, 2)
      ALLOCATE (wfn(s1, s2, 0:3, -nder:nder))
      s2 = MAXVAL(state%maxn_occ)+nder
      ALLOCATE (rbasis(s1, s2, 0:3))
      rbasis = 0._dp

      DO ider = -nder, nder
         dene = REAL(ider, KIND=dp)*delta
         IF(.NOT.(fhomo > ABS(dene)))CALL cp__a("atom_energy.F",537)
         state%occupation(lhomo, nhomo) = fhomo+dene
         CALL calculate_atom(atom, iw=0, noguess=.TRUE.)
         wfn(:, :, :, ider) = atom%orbitals%wfn
         state%occupation(lhomo, nhomo) = fhomo
      END DO

      DO l = 0, state%maxl_occ
         ! occupied states
         DO i = 1, MAX(state%maxn_occ(l), 1)
            rbasis(:, i, l) = wfn(:, i, l, 0)
         END DO
         ! differentiation
         DO ider = 1, nder
            i = MAX(state%maxn_occ(l), 1)
            SELECT CASE (ider)
            CASE (1)
               rbasis(:, i+1, l) = 0.5_dp*(wfn(:, i, l, 1)-wfn(:, i, l, -1))/delta
            CASE (2)
               rbasis(:, i+2, l) = 0.25_dp*(wfn(:, i, l, 2)-2._dp*wfn(:, i, l, 0)+wfn(:, i, l, -2))/delta**2
            CASE (3)
               rbasis(:, i+3, l) = 0.125_dp*(wfn(:, i, l, 3)-3._dp*wfn(:, i, l, 1) &
                                             +3._dp*wfn(:, i, l, -1)-wfn(:, i, l, -3))/delta**3
            CASE DEFAULT
               CALL cp__b("atom_energy.F",561,"")
            END SELECT
         END DO

         ! orthogonalization, use gram-schmidt in order to keep the natural order (semi-core, valence, response) of the wfn.
         n = state%maxn_occ(l)+nder
         m = atom%basis%nbas(l)
         DO i = 1, n
            DO j = 1, i-1
               o = DOT_PRODUCT(rbasis(1:m, j, l), RESHAPE(MATMUL(ovlp(1:m, 1:m, l), rbasis(1:m, i:i, l)), (/m/)))
               rbasis(1:m, i, l) = rbasis(1:m, i, l)-o*rbasis(1:m, j, l)
            ENDDO
            o = DOT_PRODUCT(rbasis(1:m, i, l), RESHAPE(MATMUL(ovlp(1:m, 1:m, l), rbasis(1:m, i:i, l)), (/m/)))
            rbasis(1:m, i, l) = rbasis(1:m, i, l)/SQRT(o)
         ENDDO

         ! check
         ALLOCATE (amat(n, n))
         amat(1:n, 1:n) = MATMUL(TRANSPOSE(rbasis(1:m, 1:n, l)), MATMUL(ovlp(1:m, 1:m, l), rbasis(1:m, 1:n, l)))
         DO i = 1, n
            amat(i, i) = amat(i, i)-1._dp
         END DO
         IF (MAXVAL(ABS(amat)) > 1.e-12) THEN
            WRITE (iw, '(A,G20.10)') " Orthogonality error  ", MAXVAL(ABS(amat))
         END IF
         DEALLOCATE (amat)

         ! Quickstep normalization
         WRITE (iw, '(/,A,T30,I3)') " Angular momentum :", l

         WRITE (iw, '(/,A,I0,A,I0,A)') "             Exponent     Coef.(Quickstep Normalization), first ", &
            n-nder, " valence ", nder, " response"
         expzet = 0.25_dp*REAL(2*l+3, dp)
         prefac = SQRT(SQRT(pi)/2._dp**(l+2)*dfac(2*l+1))
         DO i = 1, m
            zeta = (2._dp*atom%basis%am(i, l))**expzet
            WRITE (iw, '(4X,F20.10,4X,15ES20.6)') atom%basis%am(i, l), ((prefac*rbasis(i, k, l)/zeta), k=1, n)
         END DO

      END DO

      DEALLOCATE (wfn, rbasis)

      WRITE (iw, '(" ",79("*"))')

   END SUBROUTINE atom_response_basis

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_write_upf(atom, iw)

      TYPE(atom_type), POINTER                           :: atom
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_write_upf', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: string
      INTEGER                                            :: i, ibeta, j, k, l, lmax, nbeta, nr, &
                                                            nwfc, nwfn
      LOGICAL                                            :: up
      REAL(KIND=dp)                                      :: pf, rl, rmax
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: beta, corden, dens, ef, locpot, rp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: dij
      TYPE(atom_gthpot_type), POINTER                    :: pot

      IF (.NOT. atom%pp_calc) RETURN
      IF (atom%potential%ppot_type /= gth_pseudo) RETURN
      pot => atom%potential%gth_pot
      IF(.NOT.(.NOT. pot%lsdpot))CALL cp__a("atom_energy.F",632)

      WRITE (iw, '(A)') '<UPF version="2.0.1">'

      WRITE (iw, '(T4,A)') '<PP_INFO>'
      WRITE (iw, '(T8,A)') 'Converted from CP2K GTH format'
      WRITE (iw, '(T8,A)') '<PP_INPUTFILE>'
      CALL atom_write_pseudo_param(pot, iw)
      WRITE (iw, '(T8,A)') '</PP_INPUTFILE>'
      WRITE (iw, '(T4,A)') '</PP_INFO>'
      WRITE (iw, '(T4,A)') '<PP_HEADER'
      WRITE (iw, '(T8,A)') 'generated="Generated in analytical, separable form"'
      WRITE (iw, '(T8,A)') 'author="Goedecker/Hartwigsen/Hutter/Teter"'
      WRITE (iw, '(T8,A)') 'date="Phys.Rev.B58, 3641 (1998); B54, 1703 (1996)"'
      WRITE (iw, '(T8,A)') 'comment="This file generated by CP2K ATOM code"'
      CALL compose(string, "element", cval=pot%symbol)
      WRITE (iw, '(T8,A)') TRIM(string)
      WRITE (iw, '(T8,A)') 'pseudo_type="NC"'
      WRITE (iw, '(T8,A)') 'relativistic="no"'
      WRITE (iw, '(T8,A)') 'is_ultrasoft="F"'
      WRITE (iw, '(T8,A)') 'is_paw="F"'
      WRITE (iw, '(T8,A)') 'is_coulomb="F"'
      WRITE (iw, '(T8,A)') 'has_so="F"'
      WRITE (iw, '(T8,A)') 'has_wfc="F"'
      WRITE (iw, '(T8,A)') 'has_gipaw="F"'
      WRITE (iw, '(T8,A)') 'paw_as_gipaw="F"'
      IF (pot%nlcc) THEN
         WRITE (iw, '(T8,A)') 'core_correction="T"'
      ELSE
         WRITE (iw, '(T8,A)') 'core_correction="F"'
      END IF
      WRITE (iw, '(T8,A)') 'functional="DFT"'
      CALL compose(string, "z_valence", rval=pot%zion)
      WRITE (iw, '(T8,A)') TRIM(string)
      CALL compose(string, "total_psenergy", rval=2._dp*atom%energy%etot)
      WRITE (iw, '(T8,A)') TRIM(string)
      WRITE (iw, '(T8,A)') 'wfc_cutoff="0.0E+00"'
      WRITE (iw, '(T8,A)') 'rho_cutoff="0.0E+00"'
      lmax = -1
      DO l = 0, 3
         IF (pot%nl(l) > 0) lmax = l
      END DO
      CALL compose(string, "l_max", ival=lmax)
      WRITE (iw, '(T8,A)') TRIM(string)
      WRITE (iw, '(T8,A)') 'l_max_rho="0"'
      WRITE (iw, '(T8,A)') 'l_local="-3"'
      nr = atom%basis%grid%nr
      CALL compose(string, "mesh_size", ival=nr)
      WRITE (iw, '(T8,A)') TRIM(string)
      nwfc = SUM(atom%state%maxn_occ)
      CALL compose(string, "number_of_wfc", ival=nwfc)
      WRITE (iw, '(T8,A)') TRIM(string)
      nbeta = SUM(pot%nl)
      CALL compose(string, "number_of_proj", ival=nbeta)
      WRITE (iw, '(T8,A)') TRIM(string)//'/>'

      ! Mesh
      up = atom%basis%grid%rad(1) < atom%basis%grid%rad(nr)
      WRITE (iw, '(T4,A)') '<PP_MESH'
      WRITE (iw, '(T8,A)') 'dx="1.E+00"'
      CALL compose(string, "mesh", ival=nr)
      WRITE (iw, '(T8,A)') TRIM(string)
      WRITE (iw, '(T8,A)') 'xmin="1.E+00"'
      rmax = MAXVAL(atom%basis%grid%rad)
      CALL compose(string, "rmax", rval=rmax)
      WRITE (iw, '(T8,A)') TRIM(string)
      WRITE (iw, '(T8,A)') 'zmesh="1.E+00">'
      WRITE (iw, '(T8,A)') '<PP_R type="real"'
      CALL compose(string, "size", ival=nr)
      WRITE (iw, '(T12,A)') TRIM(string)
      WRITE (iw, '(T12,A)') 'columns="4">'
      IF (up) THEN
         WRITE (iw, '(T12,4ES25.12E3)') (atom%basis%grid%rad(i), i=1, nr)
      ELSE
         WRITE (iw, '(T12,4ES25.12E3)') (atom%basis%grid%rad(i), i=nr, 1, -1)
      END IF
      WRITE (iw, '(T8,A)') '</PP_R>'
      WRITE (iw, '(T8,A)') '<PP_RAB type="real"'
      CALL compose(string, "size", ival=nr)
      WRITE (iw, '(T12,A)') TRIM(string)
      WRITE (iw, '(T12,A)') 'columns="4">'
      IF (up) THEN
         WRITE (iw, '(T12,4ES25.12E3)') (atom%basis%grid%wr(i)/atom%basis%grid%rad2(i), i=1, nr)
      ELSE
         WRITE (iw, '(T12,4ES25.12E3)') (atom%basis%grid%wr(i)/atom%basis%grid%rad2(i), i=nr, 1, -1)
      END IF
      WRITE (iw, '(T8,A)') '</PP_RAB>'
      WRITE (iw, '(T4,A)') '</PP_MESH>'

      ! NLCC
      IF (pot%nlcc) THEN
         WRITE (iw, '(T4,A)') '<PP_NLCC type="real"'
         CALL compose(string, "size", ival=nr)
         WRITE (iw, '(T8,A)') TRIM(string)
         WRITE (iw, '(T8,A)') 'columns="4">'
         ALLOCATE (corden(nr))
         corden(:) = 0.0_dp
         CALL atom_core_density(corden, pot, "RHO", atom%basis%grid%rad)
         IF (up) THEN
            WRITE (iw, '(T8,4ES25.12E3)') (corden(i), i=1, nr)
         ELSE
            WRITE (iw, '(T8,4ES25.12E3)') (corden(i), i=nr, 1, -1)
         END IF
         DEALLOCATE (corden)
         WRITE (iw, '(T4,A)') '</PP_NLCC>'
      END IF

      ! local PP
      WRITE (iw, '(T4,A)') '<PP_LOCAL type="real"'
      CALL compose(string, "size", ival=nr)
      WRITE (iw, '(T8,A)') TRIM(string)
      WRITE (iw, '(T8,A)') 'columns="4">'
      ALLOCATE (locpot(nr))
      locpot(:) = 0.0_dp
      CALL atom_local_potential(locpot, pot, atom%basis%grid%rad)
      IF (up) THEN
         WRITE (iw, '(T8,4ES25.12E3)') (2.0_dp*locpot(i), i=1, nr)
      ELSE
         WRITE (iw, '(T8,4ES25.12E3)') (2.0_dp*locpot(i), i=nr, 1, -1)
      END IF
      DEALLOCATE (locpot)
      WRITE (iw, '(T4,A)') '</PP_LOCAL>'

      ! nonlocal PP
      WRITE (iw, '(T4,A)') '<PP_NONLOCAL>'
      ALLOCATE (rp(nr), ef(nr), beta(nr))
      ibeta = 0
      DO l = 0, 3
         IF (pot%nl(l) == 0) CYCLE
         rl = pot%rcnl(l)
         rp(:) = atom%basis%grid%rad
         ef(:) = EXP(-0.5_dp*rp*rp/(rl*rl))
         DO i = 1, pot%nl(l)
            pf = rl**(l+0.5_dp*(4._dp*i-1._dp))
            j = l+2*i-1
            pf = SQRT(2._dp)/(pf*SQRT(gamma1(j)))
            beta(:) = pf*rp**(l+2*i-2)*ef
            ibeta = ibeta+1
            CALL compose(string, "<PP_BETA", counter=ibeta)
            WRITE (iw, '(T8,A)') TRIM(string)
            CALL compose(string, "angular_momentum", ival=l)
            WRITE (iw, '(T12,A)') TRIM(string)
            WRITE (iw, '(T12,A)') 'type="real"'
            CALL compose(string, "size", ival=nr)
            WRITE (iw, '(T12,A)') TRIM(string)
            WRITE (iw, '(T12,A)') 'columns="4">'
            beta(:) = beta*rp
            IF (up) THEN
               WRITE (iw, '(T12,4ES25.12E3)') (2._dp*beta(j), j=1, nr)
            ELSE
               WRITE (iw, '(T12,4ES25.12E3)') (2._dp*beta(j), j=nr, 1, -1)
            END IF
            CALL compose(string, "</PP_BETA", counter=ibeta, isfinal=.TRUE.)
            WRITE (iw, '(T8,A)') TRIM(string)
         END DO
      END DO
      DEALLOCATE (rp, ef, beta)
      ! nonlocal PP matrix elements
      ALLOCATE (dij(nbeta, nbeta))
      dij = 0._dp
      DO l = 0, 3
         IF (pot%nl(l) == 0) CYCLE
         ibeta = SUM(pot%nl(0:l-1))+1
         i = ibeta+pot%nl(l)-1
         dij(ibeta:i, ibeta:i) = pot%hnl(1:pot%nl(l), 1:pot%nl(l), l)
      END DO
      WRITE (iw, '(T8,A)') '<PP_DIJ type="real"'
      WRITE (iw, '(T12,A)') 'columns="4">'
      WRITE (iw, '(T12,4ES25.12E3)') ((0.5_dp*dij(i, j), j=1, nbeta), i=1, nbeta)
      WRITE (iw, '(T8,A)') '</PP_DIJ>'
      DEALLOCATE (dij)
      WRITE (iw, '(T4,A)') '</PP_NONLOCAL>'

      ! atomic wavefunctions
      ALLOCATE (beta(nr))
      WRITE (iw, '(T4,A)') '<PP_PSWFC>'
      nwfn = 0
      DO l = 0, 3
         DO i = 1, 10
            IF (ABS(atom%state%occupation(l, i)) == 0._dp) CYCLE
            nwfn = nwfn+1
            CALL compose(string, "<PP_CHI", counter=nwfn)
            WRITE (iw, '(T8,A)') TRIM(string)
            CALL compose(string, "l", ival=l)
            WRITE (iw, '(T12,A)') TRIM(string)
            CALL compose(string, "occupation", rval=atom%state%occupation(l, i))
            WRITE (iw, '(T12,A)') TRIM(string)
            CALL compose(string, "pseudo_energy", rval=2._dp*atom%orbitals%ener(i, l))
            WRITE (iw, '(T12,A)') TRIM(string)
            WRITE (iw, '(T12,A)') 'columns="4">'
            beta = 0._dp
            DO k = 1, atom%basis%nbas(l)
               beta(:) = beta(:)+atom%orbitals%wfn(k, i, l)*atom%basis%bf(:, k, l)
            END DO
            beta(:) = beta*atom%basis%grid%rad
            IF (up) THEN
               WRITE (iw, '(T12,4ES25.12E3)') (beta(j)*atom%basis%grid%rad(j), j=1, nr)
            ELSE
               WRITE (iw, '(T12,4ES25.12E3)') (beta(j)*atom%basis%grid%rad(j), j=nr, 1, -1)
            END IF
            CALL compose(string, "</PP_CHI", counter=nwfn, isfinal=.TRUE.)
            WRITE (iw, '(T8,A)') TRIM(string)
         END DO
      END DO
      WRITE (iw, '(T4,A)') '</PP_PSWFC>'
      DEALLOCATE (beta)

      ! atomic charge
      ALLOCATE (dens(nr))
      WRITE (iw, '(T4,A)') '<PP_RHOATOM type="real"'
      CALL compose(string, "size", ival=nr)
      WRITE (iw, '(T8,A)') TRIM(string)
      WRITE (iw, '(T8,A)') 'columns="4">'
      CALL atom_density(dens, atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, &
                        "RHO", atom%basis%grid%rad)
      IF (up) THEN
         WRITE (iw, '(T8,4ES25.12E3)') (4._dp*pi*dens(j)*atom%basis%grid%rad2(j), j=1, nr)
      ELSE
         WRITE (iw, '(T8,4ES25.12E3)') (4._dp*pi*dens(j)*atom%basis%grid%rad2(j), j=nr, 1, -1)
      END IF
      WRITE (iw, '(T4,A)') '</PP_RHOATOM>'
      DEALLOCATE (dens)

      WRITE (iw, '(A)') '</UPF>'

   END SUBROUTINE atom_write_upf
! **************************************************************************************************
!> \brief ...
!> \param string ...
!> \param tag ...
!> \param counter ...
!> \param rval ...
!> \param ival ...
!> \param cval ...
!> \param isfinal ...
! **************************************************************************************************
   SUBROUTINE compose(string, tag, counter, rval, ival, cval, isfinal)
      CHARACTER(len=*)                                   :: string, tag
      INTEGER, OPTIONAL                                  :: counter
      REAL(KIND=dp), OPTIONAL                            :: rval
      INTEGER, OPTIONAL                                  :: ival
      CHARACTER(len=*), OPTIONAL                         :: cval
      LOGICAL, OPTIONAL                                  :: isfinal

      CHARACTER(LEN=default_string_length)               :: str
      LOGICAL                                            :: fin

      IF (PRESENT(counter)) THEN
         WRITE (str, "(I12)") counter
      ELSEIF (PRESENT(rval)) THEN
         WRITE (str, "(G18.8)") rval
      ELSEIF (PRESENT(ival)) THEN
         WRITE (str, "(I12)") ival
      ELSEIF (PRESENT(cval)) THEN
         WRITE (str, "(A)") TRIM(ADJUSTL(cval))
      ELSE
         WRITE (str, "(A)") ""
      END IF
      fin = .FALSE.
      IF (PRESENT(isfinal)) fin = isfinal
      IF (PRESENT(counter)) THEN
         IF (fin) THEN
            WRITE (string, "(A,A1,A,A1)") TRIM(ADJUSTL(tag)), '.', TRIM(ADJUSTL(str)), '>'
         ELSE
            WRITE (string, "(A,A1,A)") TRIM(ADJUSTL(tag)), '.', TRIM(ADJUSTL(str))
         END IF
      ELSE
         IF (fin) THEN
            WRITE (string, "(A,A2,A,A2)") TRIM(ADJUSTL(tag)), '="', TRIM(ADJUSTL(str)), '>"'
         ELSE
            WRITE (string, "(A,A2,A,A1)") TRIM(ADJUSTL(tag)), '="', TRIM(ADJUSTL(str)), '"'
         END IF
      END IF
   END SUBROUTINE compose

END MODULE atom_energy
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE atom_pseudo
   USE atom_electronic_structure,       ONLY: calculate_atom
   USE atom_fit,                        ONLY: atom_fit_pseudo
   USE atom_operators,                  ONLY: atom_int_release,&
                                              atom_int_setup,&
                                              atom_ppint_release,&
                                              atom_ppint_setup,&
                                              atom_relint_release,&
                                              atom_relint_setup
   USE atom_output,                     ONLY: atom_print_basis,&
                                              atom_print_info,&
                                              atom_print_method,&
                                              atom_print_potential
   USE atom_types,                      ONLY: &
        atom_basis_type, atom_integrals, atom_optimization_type, atom_orbitals, atom_p_type, &
        atom_potential_type, atom_state, create_atom_orbs, create_atom_type, init_atom_basis, &
        init_atom_potential, read_atom_opt_section, release_atom_basis, release_atom_potential, &
        release_atom_type, set_atom
   USE atom_utils,                      ONLY: atom_consistent_method,&
                                              atom_set_occupation,&
                                              get_maxl_occ,&
                                              get_maxn_occ
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_constants,                 ONLY: do_analytic,&
                                              poly_conf
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE periodic_table,                  ONLY: nelem,&
                                              ptable
   USE physcon,                         ONLY: bohr
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC  :: atom_pseudo_opt

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_pseudo'

! **************************************************************************************************

CONTAINS

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param atom_section ...
! **************************************************************************************************
   SUBROUTINE atom_pseudo_opt(atom_section)
      TYPE(section_vals_type), POINTER                   :: atom_section

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_pseudo_opt', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: elem
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER                                            :: ads, do_eric, do_erie, handle, i, im, &
                                                            in, iw, k, l, maxl, mb, method, mo, &
                                                            n_meth, n_rep, reltyp, zcore, zval, zz
      INTEGER, DIMENSION(0:3)                            :: maxn
      INTEGER, DIMENSION(:), POINTER                     :: cn
      LOGICAL                                            :: eri_c, eri_e, pp_calc
      REAL(KIND=dp)                                      :: ne, nm
      REAL(KIND=dp), DIMENSION(0:3, 10)                  :: pocc
      TYPE(atom_basis_type), POINTER                     :: ae_basis, pp_basis
      TYPE(atom_integrals), POINTER                      :: ae_int, pp_int
      TYPE(atom_optimization_type)                       :: optimization
      TYPE(atom_orbitals), POINTER                       :: orbitals
      TYPE(atom_p_type), DIMENSION(:, :), POINTER        :: atom_info, atom_refs
      TYPE(atom_potential_type), POINTER                 :: ae_pot, p_pot
      TYPE(atom_state), POINTER                          :: state, statepp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: basis_section, method_section, &
                                                            opt_section, potential_section, &
                                                            powell_section, xc_section

      CALL timeset(routineN, handle)

      ! What atom do we calculate
      CALL section_vals_val_get(atom_section, "ATOMIC_NUMBER", i_val=zval)
      CALL section_vals_val_get(atom_section, "ELEMENT", c_val=elem)
      zz = 0
      DO i = 1, nelem
         IF (ptable(i)%symbol == elem) THEN
            zz = i
            EXIT
         END IF
      END DO
      IF (zz /= 1) zval = zz

      ! read and set up information on the basis sets
      ALLOCATE (ae_basis, pp_basis)
      basis_section => section_vals_get_subs_vals(atom_section, "AE_BASIS")
      NULLIFY (ae_basis%grid)
      CALL init_atom_basis(ae_basis, basis_section, zval, "AA")
      NULLIFY (pp_basis%grid)
      basis_section => section_vals_get_subs_vals(atom_section, "PP_BASIS")
      CALL init_atom_basis(pp_basis, basis_section, zval, "AP")

      ! print general and basis set information
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%PROGRAM_BANNER", extension=".log")
      IF (iw > 0) CALL atom_print_info(zval, "Atomic Energy Calculation", iw)
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%PROGRAM_BANNER")
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%BASIS_SET", extension=".log")
      IF (iw > 0) THEN
         CALL atom_print_basis(ae_basis, iw, " All Electron Basis")
         CALL atom_print_basis(pp_basis, iw, " Pseudopotential Basis")
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%BASIS_SET")

      ! read and setup information on the pseudopotential
      NULLIFY (potential_section)
      potential_section => section_vals_get_subs_vals(atom_section, "POTENTIAL")
      ALLOCATE (ae_pot, p_pot)
      CALL init_atom_potential(p_pot, potential_section, zval)
      CALL init_atom_potential(ae_pot, potential_section, -1)
      IF (.NOT. p_pot%confinement .AND. .NOT. ae_pot%confinement) THEN
         !set default confinement potential
         p_pot%confinement = .TRUE.
         p_pot%conf_type = poly_conf
         p_pot%scon = 2.0_dp
         p_pot%acon = 0.5_dp
         ! this seems to be the default in the old code
         p_pot%rcon = (2._dp*ptable(zval)%covalent_radius*bohr)**2
         ae_pot%confinement = .TRUE.
         ae_pot%conf_type = poly_conf
         ae_pot%scon = 2.0_dp
         ae_pot%acon = 0.5_dp
         ! this seems to be the default in the old code
         ae_pot%rcon = (2._dp*ptable(zval)%covalent_radius*bohr)**2
      END IF

      ! if the ERI's are calculated analytically, we have to precalculate them
      eri_c = .FALSE.
      CALL section_vals_val_get(atom_section, "COULOMB_INTEGRALS", i_val=do_eric)
      IF (do_eric == do_analytic) eri_c = .TRUE.
      eri_e = .FALSE.
      CALL section_vals_val_get(atom_section, "EXCHANGE_INTEGRALS", i_val=do_erie)
      IF (do_erie == do_analytic) eri_e = .TRUE.

      ! information on the states to be calculated
      CALL section_vals_val_get(atom_section, "MAX_ANGULAR_MOMENTUM", i_val=maxl)
      maxn = 0
      CALL section_vals_val_get(atom_section, "CALCULATE_STATES", i_vals=cn)
      DO in = 1, MIN(SIZE(cn), 4)
         maxn(in-1) = cn(in)
      END DO
      DO in = 0, 3
         maxn(in) = MIN(maxn(in), ae_basis%nbas(in))
      END DO

      ! read optimization section
      opt_section => section_vals_get_subs_vals(atom_section, "OPTIMIZATION")
      CALL read_atom_opt_section(optimization, opt_section)

      ! Check for the total number of electron configurations to be calculated
      CALL section_vals_val_get(atom_section, "ELECTRON_CONFIGURATION", n_rep_val=n_rep)
      ! Check for the total number of method types to be calculated
      method_section => section_vals_get_subs_vals(atom_section, "METHOD")
      CALL section_vals_get(method_section, n_repetition=n_meth)

      ! integrals
      ALLOCATE (ae_int, pp_int)

      ALLOCATE (atom_info(n_rep, n_meth), atom_refs(n_rep, n_meth))

      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%PROGRAM_BANNER", extension=".log")
      IF (iw > 0) THEN
         WRITE (iw, '(/," ",79("*"))')
         WRITE (iw, '(" ",26("*"),A,25("*"))') " Calculate Reference States "
         WRITE (iw, '(" ",79("*"))')
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%PROGRAM_BANNER")

      DO in = 1, n_rep
         DO im = 1, n_meth

            NULLIFY (atom_info(in, im)%atom, atom_refs(in, im)%atom)
            CALL create_atom_type(atom_info(in, im)%atom)
            CALL create_atom_type(atom_refs(in, im)%atom)

            atom_info(in, im)%atom%optimization = optimization
            atom_refs(in, im)%atom%optimization = optimization

            atom_info(in, im)%atom%z = zval
            atom_refs(in, im)%atom%z = zval
            xc_section => section_vals_get_subs_vals(method_section, "XC", i_rep_section=im)
            atom_info(in, im)%atom%xc_section => xc_section
            atom_refs(in, im)%atom%xc_section => xc_section

            ALLOCATE (state, statepp)

            ! get the electronic configuration
            CALL section_vals_val_get(atom_section, "ELECTRON_CONFIGURATION", i_rep_val=in, &
                                      c_vals=tmpstringlist)
            ! all electron configurations have to be with full core
            pp_calc = INDEX(tmpstringlist(1), "CORE") /= 0
            IF(.NOT.(.NOT. pp_calc))CALL cp__a("atom_pseudo.F",213)

            ! set occupations
            CALL atom_set_occupation(tmpstringlist, state%occ, state%occupation, state%multiplicity)
            state%maxl_occ = get_maxl_occ(state%occ)
            state%maxn_occ = get_maxn_occ(state%occ)
            ! set number of states to be calculated
            state%maxl_calc = MAX(maxl, state%maxl_occ)
            state%maxl_calc = MIN(3, state%maxl_calc)
            state%maxn_calc = 0
            DO k = 0, state%maxl_calc
               ads = 2
               IF (state%maxn_occ(k) == 0) ads = 1
               state%maxn_calc(k) = MAX(maxn(k), state%maxn_occ(k)+ads)
               state%maxn_calc(k) = MIN(state%maxn_calc(k), ae_basis%nbas(k))
            END DO
            state%core = 0._dp
            CALL set_atom(atom_refs(in, im)%atom, zcore=zval, pp_calc=.FALSE.)

            IF (state%multiplicity /= -1) THEN
               ! set alpha and beta occupations
               state%occa = 0._dp
               state%occb = 0._dp
               DO l = 0, 3
                  nm = REAL((2*l+1), KIND=dp)
                  DO k = 1, 10
                     ne = state%occupation(l, k)
                     IF (ne == 0._dp) THEN !empty shell
                        EXIT !assume there are no holes
                     ELSEIF (ne == 2._dp*nm) THEN !closed shell
                        state%occa(l, k) = nm
                        state%occb(l, k) = nm
                     ELSEIF (state%multiplicity == -2) THEN !High spin case
                        state%occa(l, k) = MIN(ne, nm)
                        state%occb(l, k) = MAX(0._dp, ne-nm)
                     ELSE
                        state%occa(l, k) = 0.5_dp*(ne+state%multiplicity-1._dp)
                        state%occb(l, k) = ne-state%occa(l, k)
                     END IF
                  END DO
               END DO
            END IF

            ! set occupations for pseudopotential calculation
            CALL section_vals_val_get(atom_section, "CORE", c_vals=tmpstringlist)
            CALL atom_set_occupation(tmpstringlist, statepp%core, pocc)
            zcore = zval-NINT(SUM(statepp%core))
            CALL set_atom(atom_info(in, im)%atom, zcore=zcore, pp_calc=.TRUE.)

            statepp%occ = state%occ-statepp%core
            statepp%occupation = 0._dp
            DO l = 0, 3
               k = 0
               DO i = 1, 10
                  IF (statepp%occ(l, i) /= 0._dp) THEN
                     k = k+1
                     statepp%occupation(l, k) = state%occ(l, i)
                  END IF
               END DO
            END DO
            IF (state%multiplicity /= -1) THEN
               statepp%occa = state%occa-statepp%core/2
               statepp%occb = state%occb-statepp%core/2
            END IF

            statepp%maxl_occ = get_maxl_occ(statepp%occ)
            statepp%maxn_occ = get_maxn_occ(statepp%occ)
            statepp%maxl_calc = state%maxl_calc
            statepp%maxn_calc = 0
            maxn = get_maxn_occ(statepp%core)
            DO k = 0, statepp%maxl_calc
               statepp%maxn_calc(k) = state%maxn_calc(k)-maxn(k)
               statepp%maxn_calc(k) = MIN(statepp%maxn_calc(k), pp_basis%nbas(k))
            END DO
            statepp%multiplicity = state%multiplicity

            CALL section_vals_val_get(method_section, "METHOD_TYPE", i_val=method, i_rep_section=im)
            CALL section_vals_val_get(method_section, "RELATIVISTIC", i_val=reltyp, i_rep_section=im)
            CALL set_atom(atom_info(in, im)%atom, method_type=method)
            CALL set_atom(atom_refs(in, im)%atom, method_type=method, relativistic=reltyp)

            ! calculate integrals: pseudopotential basis
            ! general integrals
            CALL atom_int_setup(pp_int, pp_basis, potential=p_pot, eri_coulomb=eri_c, eri_exchange=eri_e)
            !
            NULLIFY (pp_int%tzora, pp_int%hdkh)
            ! potential
            CALL atom_ppint_setup(pp_int, pp_basis, potential=p_pot)
            !
            CALL set_atom(atom_info(in, im)%atom, basis=pp_basis, integrals=pp_int, potential=p_pot)
            statepp%maxn_calc(:) = MIN(statepp%maxn_calc(:), pp_basis%nbas(:))
            IF(.NOT.(ALL(state%maxn_calc(:) >= state%maxn_occ)))CALL cp__a("atom_pseudo.F",304)

            ! calculate integrals: all electron basis
            ! general integrals
            CALL atom_int_setup(ae_int, ae_basis, potential=ae_pot, &
                                eri_coulomb=eri_c, eri_exchange=eri_e)
            ! potential
            CALL atom_ppint_setup(ae_int, ae_basis, potential=ae_pot)
            ! relativistic correction terms
            CALL atom_relint_setup(ae_int, ae_basis, reltyp, zcore=REAL(zval, dp))
            !
            CALL set_atom(atom_refs(in, im)%atom, basis=ae_basis, integrals=ae_int, potential=ae_pot)
            state%maxn_calc(:) = MIN(state%maxn_calc(:), ae_basis%nbas(:))
            IF(.NOT.(ALL(state%maxn_calc(:) >= state%maxn_occ)))CALL cp__a("atom_pseudo.F",317)

            CALL set_atom(atom_info(in, im)%atom, coulomb_integral_type=do_eric, &
                          exchange_integral_type=do_erie)
            CALL set_atom(atom_refs(in, im)%atom, coulomb_integral_type=do_eric, &
                          exchange_integral_type=do_erie)

            CALL set_atom(atom_info(in, im)%atom, state=statepp)
            NULLIFY (orbitals)
            mo = MAXVAL(statepp%maxn_calc)
            mb = MAXVAL(atom_info(in, im)%atom%basis%nbas)
            CALL create_atom_orbs(orbitals, mb, mo)
            CALL set_atom(atom_info(in, im)%atom, orbitals=orbitals)

            CALL set_atom(atom_refs(in, im)%atom, state=state)
            NULLIFY (orbitals)
            mo = MAXVAL(state%maxn_calc)
            mb = MAXVAL(atom_refs(in, im)%atom%basis%nbas)
            CALL create_atom_orbs(orbitals, mb, mo)
            CALL set_atom(atom_refs(in, im)%atom, orbitals=orbitals)

            IF (atom_consistent_method(atom_refs(in, im)%atom%method_type, atom_refs(in, im)%atom%state%multiplicity)) THEN
               !Print method info
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%METHOD_INFO", extension=".log")
               CALL atom_print_method(atom_refs(in, im)%atom, iw)
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%METHOD_INFO")
               !Calculate the electronic structure
               iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%SCF_INFO", extension=".log")
               CALL calculate_atom(atom_refs(in, im)%atom, iw)
               CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%SCF_INFO")
            END IF
         END DO
      END DO

      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%FIT_PSEUDO", extension=".log")
      IF (iw > 0) THEN
         WRITE (iw, '(/," ",79("*"))')
         WRITE (iw, '(" ",21("*"),A,21("*"))') " Optimize Pseudopotential Parameters "
         WRITE (iw, '(" ",79("*"))')
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%FIT_PSEUDO")
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%POTENTIAL", extension=".log")
      IF (iw > 0) THEN
         CALL atom_print_potential(p_pot, iw)
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%POTENTIAL")
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%FIT_PSEUDO", extension=".log")
      IF (iw > 0) THEN
         powell_section => section_vals_get_subs_vals(atom_section, "POWELL")
         CALL atom_fit_pseudo(atom_info, atom_refs, p_pot, iw, powell_section)
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%FIT_PSEUDO")
      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%POTENTIAL", extension=".log")
      IF (iw > 0) THEN
         CALL atom_print_potential(p_pot, iw)
      END IF
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%POTENTIAL")

      ! clean up
      CALL atom_int_release(ae_int)
      CALL atom_ppint_release(ae_int)
      CALL atom_relint_release(ae_int)

      CALL atom_int_release(pp_int)
      CALL atom_ppint_release(pp_int)
      CALL atom_relint_release(pp_int)

      CALL release_atom_basis(ae_basis)
      CALL release_atom_basis(pp_basis)

      CALL release_atom_potential(p_pot)
      CALL release_atom_potential(ae_pot)

      DO in = 1, n_rep
         DO im = 1, n_meth
            CALL release_atom_type(atom_info(in, im)%atom)
            CALL release_atom_type(atom_refs(in, im)%atom)
         END DO
      END DO
      DEALLOCATE (atom_info, atom_refs)

      DEALLOCATE (ae_pot, p_pot, ae_basis, pp_basis, ae_int, pp_int)

      CALL timestop(handle)

   END SUBROUTINE atom_pseudo_opt

! **************************************************************************************************

END MODULE atom_pseudo
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE atom
   USE atom_basis,                      ONLY: atom_basis_opt
   USE atom_energy,                     ONLY: atom_energy_opt
   USE atom_pseudo,                     ONLY: atom_pseudo_opt
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE header,                          ONLY: atom_footer,&
                                              atom_header
   USE input_constants,                 ONLY: atom_basis_run,&
                                              atom_energy_run,&
                                              atom_no_run,&
                                              atom_pseudo_run
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE periodic_table,                  ONLY: nelem,&
                                              ptable
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC  :: atom_code

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom'

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param root_section ...
! **************************************************************************************************
   SUBROUTINE atom_code(root_section)
      TYPE(section_vals_type), POINTER                   :: root_section

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_code', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw, run_type_id
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: atom_section

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      NULLIFY (atom_section)
      atom_section => section_vals_get_subs_vals(root_section, "ATOM")

      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%PROGRAM_BANNER", extension=".log")
      CALL atom_header(iw)
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%PROGRAM_BANNER")

      CALL atom_test(atom_section)

      CALL section_vals_val_get(atom_section, "RUN_TYPE", i_val=run_type_id)
      SELECT CASE (run_type_id)
      CASE (atom_no_run)
         ! do (almost) nothing
      CASE (atom_energy_run)
         CALL atom_energy_opt(atom_section)
      CASE (atom_basis_run)
         CALL atom_basis_opt(atom_section)
      CASE (atom_pseudo_run)
         CALL atom_pseudo_opt(atom_section)
      CASE default
         CALL cp__b("atom.F",74,"")
      END SELECT

      iw = cp_print_key_unit_nr(logger, atom_section, "PRINT%PROGRAM_BANNER", extension=".log")
      CALL atom_footer(iw)
      CALL cp_print_key_finished_output(iw, logger, atom_section, "PRINT%PROGRAM_BANNER")

      CALL timestop(handle)

   END SUBROUTINE atom_code

! **************************************************************************************************
!> \brief ...
!> \param atom_section ...
! **************************************************************************************************
   SUBROUTINE atom_test(atom_section)
      TYPE(section_vals_type), POINTER                   :: atom_section

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_test', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: elem
      INTEGER                                            :: i, ie, z

      CALL section_vals_val_get(atom_section, "ATOMIC_NUMBER", i_val=z)
      CALL section_vals_val_get(atom_section, "ELEMENT", c_val=elem)

      ie = 0
      DO i = 1, nelem
         IF (ptable(i)%symbol == elem) THEN
            ie = i
            EXIT
         END IF
      END DO
      IF (ie /= z) THEN
         IF (ie /= 1 .AND. z /= 1) THEN
            CALL cp__b("atom.F",109,"")
         END IF
      END IF

   END SUBROUTINE atom_test

! **************************************************************************************************

END MODULE atom
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE qs_tddfpt_types
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              fm_pool_create_fm,&
                                              fm_pool_give_back_fm
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_matrix_pools,                 ONLY: mpools_get
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

! **************************************************************************************************
   TYPE tddfpt_env_type
      REAL(KIND=dp), DIMENSION(:), POINTER               :: evals ! eigenvalues
      TYPE(cp_fm_p_type), DIMENSION(:, :), &
         POINTER                                       :: evecs ! eigenvectors
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: invS ! the inverse of the metric
      TYPE(cp_fm_pool_p_type), DIMENSION(:), &
         POINTER                                       :: ao_mo_fm_pools
   END TYPE tddfpt_env_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_tddfpt_types'

   PRIVATE

   PUBLIC :: tddfpt_env_type

   PUBLIC :: tddfpt_env_allocate, &
             tddfpt_env_deallocate

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param t_env ...
!> \param p_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE tddfpt_env_allocate(t_env, p_env, qs_env)

      TYPE(tddfpt_env_type), INTENT(inout)               :: t_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'tddfpt_env_allocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n_ev, n_spins, spin, stat
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control

      NULLIFY (dft_control, para_env, blacs_env)

      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      blacs_env=blacs_env)

      n_ev = dft_control%tddfpt_control%n_ev
      n_spins = dft_control%nspins
      CALL mpools_get(qs_env%mpools, ao_mo_fm_pools=t_env%ao_mo_fm_pools)

      ALLOCATE (t_env%evals(n_ev), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 86, &
                                      "t_env%evals", dp_size*n_ev)

      ALLOCATE (t_env%evecs(n_ev, n_spins), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 90, &
                                      "t_env%evecs", dp_size*n_ev*n_spins)
      DO spin = 1, n_spins
         DO i = 1, n_ev
            CALL fm_pool_create_fm(t_env%ao_mo_fm_pools(spin)%pool, &
                                   t_env%evecs(i, spin)%matrix)
         END DO
      END DO

      !-------------------------------------------------!
      ! allocate memory for the inverse of the S matrix !
      !-------------------------------------------------!
      ALLOCATE (t_env%invS(n_spins), STAT=stat)
      IF (stat /= 0) CALL stop_memory(routineN, moduleN, 103, &
                                      "t_env%invS", int_size*n_spins)
      DO spin = 1, n_spins
         NULLIFY (fm_struct)
         CALL cp_fm_struct_create(fm_struct, para_env, blacs_env, &
                                  p_env%n_ao(spin), p_env%n_ao(spin))
         CALL cp_fm_create(t_env%invS(spin)%matrix, fm_struct, routineP//"invS")
         CALL cp_fm_struct_release(fm_struct)
      END DO

   END SUBROUTINE tddfpt_env_allocate

! **************************************************************************************************
!> \brief ...
!> \param t_env ...
! **************************************************************************************************
   SUBROUTINE tddfpt_env_deallocate(t_env)

      TYPE(tddfpt_env_type), INTENT(inout)               :: t_env

      CHARACTER(len=*), PARAMETER :: routineN = 'tddfpt_env_deallocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, spin

      DO spin = 1, SIZE(t_env%evecs, 2)
         DO i = 1, SIZE(t_env%evecs, 1)
            CALL fm_pool_give_back_fm(t_env%ao_mo_fm_pools(spin)%pool, &
                                      t_env%evecs(i, spin)%matrix)
         END DO
      END DO

      DO spin = 1, SIZE(t_env%invS)
         IF (ASSOCIATED(t_env%invS(spin)%matrix)) &
            CALL cp_fm_release(t_env%invS(spin)%matrix)
      END DO
      DEALLOCATE (t_env%invS, t_env%evecs, t_env%evals)

      NULLIFY (t_env%invS)

   END SUBROUTINE tddfpt_env_deallocate

END MODULE qs_tddfpt_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      09-JUL-2002, TCH, development started
! **************************************************************************************************
MODULE qs_tddfpt_utils
   USE cp_control_types,                ONLY: dft_control_type,&
                                              tddfpt_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_trace
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_invert
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_init_random,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_to_fm
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE kinds,                           ONLY: dp
   USE physcon,                         ONLY: evolt
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_mo_types,                     ONLY: get_mo_set
   USE qs_p_env_methods,                ONLY: p_env_create,&
                                              p_env_psi0_changed
   USE qs_p_env_types,                  ONLY: p_env_release,&
                                              qs_p_env_type
   USE qs_tddfpt_types,                 ONLY: tddfpt_env_allocate,&
                                              tddfpt_env_deallocate,&
                                              tddfpt_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_tddfpt_utils'
   LOGICAL, PARAMETER          :: DEBUG_THIS_MODULE = .TRUE.

! **************************************************************************************************
   TYPE simple_solution_sorter
      INTEGER                               :: orbit
      INTEGER                               :: lumo
      REAL(KIND=DP)                        :: value
      TYPE(simple_solution_sorter), POINTER :: next
   END TYPE simple_solution_sorter

   PRIVATE

   ! METHODS
   PUBLIC :: tddfpt_cleanup, &
             tddfpt_init, &
             co_initial_guess, &
             find_contributions, &
             normalize, &
             reorthogonalize

CONTAINS

! **************************************************************************************************
!> \brief Initialize some necessary structures for a tddfpt calculation.
!> \param p_env perturbation environment to be initialized
!> \param t_env tddfpt environment to be initialized
!> \param qs_env Quickstep environment with the results of a
!>                   ground state calcualtion
! **************************************************************************************************
   SUBROUTINE tddfpt_init(p_env, t_env, qs_env)

      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(tddfpt_env_type), INTENT(out)                 :: t_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tddfpt_init', routineP = moduleN//':'//routineN

!------------------!
! create the p_env !
!------------------!

      CALL p_env_create(p_env, qs_env, orthogonal_orbitals=.TRUE.)
      CALL p_env_psi0_changed(p_env, qs_env) ! update the m_epsilon matrix

      !------------------!
      ! create the t_env !
      !------------------!
      CALL tddfpt_env_allocate(t_env, p_env, qs_env)
      CALL tddfpt_env_init(t_env, qs_env)

   END SUBROUTINE tddfpt_init

! **************************************************************************************************
!> \brief Initialize t_env with meaningfull values.
!> \param t_env ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE tddfpt_env_init(t_env, qs_env)

      TYPE(tddfpt_env_type), INTENT(inout)               :: t_env
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'tddfpt_env_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n_spins, spin
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(dft_control_type), POINTER                    :: dft_control

      NULLIFY (matrix_s, dft_control)

      CALL get_qs_env(qs_env, matrix_s=matrix_s, dft_control=dft_control)
      n_spins = dft_control%nspins
      IF (dft_control%tddfpt_control%invert_S) THEN
         DO spin = 1, n_spins
            CALL copy_dbcsr_to_fm(matrix_s(1)%matrix, t_env%invS(spin)%matrix)
            CALL cp_fm_cholesky_decompose(t_env%invS(spin)%matrix)
            CALL cp_fm_cholesky_invert(t_env%invS(spin)%matrix)
         END DO
      END IF

   END SUBROUTINE tddfpt_env_init

! **************************************************************************************************
!> \brief ...
!> \param t_env ...
!> \param p_env ...
! **************************************************************************************************
   SUBROUTINE tddfpt_cleanup(t_env, p_env)

      TYPE(tddfpt_env_type), INTENT(inout)               :: t_env
      TYPE(qs_p_env_type), POINTER                       :: p_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tddfpt_cleanup', routineP = moduleN//':'//routineN

      CALL tddfpt_env_deallocate(t_env)
      CALL p_env_release(p_env)

   END SUBROUTINE tddfpt_cleanup

! **************************************************************************************************
!> \brief ...
!> \param matrices ...
!> \param energies ...
!> \param n_v ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE co_initial_guess(matrices, energies, n_v, qs_env)

      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: matrices
      REAL(kind=DP), DIMENSION(:), INTENT(OUT)           :: energies
      INTEGER, INTENT(IN)                                :: n_v
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'co_initial_guess', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n_cols, n_lumos, n_orbits, n_rows, &
                                                            n_spins, oo, spin, vo
      REAL(KIND=DP)                                      :: evd
      REAL(KIND=DP), ALLOCATABLE, DIMENSION(:, :)        :: guess, lumos
      REAL(KIND=DP), DIMENSION(:), POINTER               :: orbital_eigenvalues
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(simple_solution_sorter), POINTER              :: sorter_iterator, sorter_pointer, &
                                                            sorter_start
      TYPE(tddfpt_control_type), POINTER                 :: tddfpt_control

! number of vectors to initialize

      NULLIFY (dft_control)

      CALL get_qs_env(qs_env, dft_control=dft_control)
      tddfpt_control => dft_control%tddfpt_control
      n_spins = dft_control%nspins
      energies = 0.0_dp

      IF (.NOT. ASSOCIATED(tddfpt_control%lumos)) THEN
         CALL cp__b("qs_tddfpt_utils.F",184,"LUMOS missing")
      END IF

      DO spin = 1, n_spins

         n_cols = matrices(1, spin)%matrix%matrix_struct%ncol_global
         n_rows = matrices(1, spin)%matrix%matrix_struct%nrow_global

         DO i = 1, n_v
            CALL cp_fm_set_all(matrices(i, spin)%matrix, 0.0_dp)
         END DO

         CALL get_mo_set(qs_env%mos(spin)%mo_set, eigenvalues=orbital_eigenvalues)

         n_lumos = tddfpt_control%lumos(spin)%matrix%matrix_struct%ncol_global

         n_orbits = SIZE(orbital_eigenvalues)

         !-----------------------------------------!
         ! create a SORTED list of initial guesses !
         !-----------------------------------------!
         ! first element
         evd = tddfpt_control%lumos_eigenvalues(1, spin)-orbital_eigenvalues(n_orbits)
         ALLOCATE (sorter_start)
         sorter_start%orbit = n_orbits
         sorter_start%lumo = 1
         sorter_start%value = evd
         NULLIFY (sorter_start%next)
         ! rest of the elements
         DO oo = n_orbits, 1, -1
            DO vo = 1, n_lumos

               IF (oo == n_orbits .AND. vo == 1) CYCLE ! already in list

               evd = tddfpt_control%lumos_eigenvalues(vo, spin)-orbital_eigenvalues(oo)

               sorter_iterator => sorter_start
               NULLIFY (sorter_pointer)
               DO WHILE (ASSOCIATED(sorter_iterator%next))
                  IF (sorter_iterator%next%value > evd) THEN
                     sorter_pointer => sorter_iterator%next
                     EXIT
                  END IF
                  sorter_iterator => sorter_iterator%next
               END DO

               ALLOCATE (sorter_iterator%next)
               sorter_iterator%next%orbit = oo
               sorter_iterator%next%lumo = vo
               sorter_iterator%next%value = evd
               sorter_iterator%next%next => sorter_pointer

            END DO
         END DO

         ALLOCATE (lumos(n_rows, n_lumos), guess(n_rows, n_orbits))
         CALL cp_fm_get_submatrix(tddfpt_control%lumos(spin)%matrix, lumos, &
                                  start_col=1, n_cols=n_lumos)

         !-------------------!
         ! fill the matrices !
         !-------------------!
         sorter_iterator => sorter_start
         DO i = 1, MIN(n_v, n_orbits*n_lumos)
            CALL dcopy(n_rows*n_orbits, 0.0_dp, 0, guess, 1)
            CALL dcopy(n_rows, lumos(:, sorter_iterator%lumo), 1, &
                       guess(:, sorter_iterator%orbit), 1)
            CALL cp_fm_set_submatrix(matrices(i, spin)%matrix, &
                                     guess(:, 1:n_orbits))
            energies(i) = energies(i)+sorter_iterator%value/REAL(n_spins, dp)
            sorter_iterator => sorter_iterator%next
         END DO
         IF (n_v > n_orbits*n_lumos) THEN
            DO i = n_orbits*n_lumos+1, n_v
               CALL cp_fm_init_random(matrices(i, spin)%matrix, n_orbits)
               energies(i) = 1.0E38_dp
            END DO
         END IF

         !--------------!
         ! some cleanup !
         !--------------!
         DEALLOCATE (lumos, guess)
         sorter_iterator => sorter_start
         DO WHILE (ASSOCIATED(sorter_iterator))
            sorter_pointer => sorter_iterator
            sorter_iterator => sorter_iterator%next
            DEALLOCATE (sorter_pointer)
         END DO

      END DO

   END SUBROUTINE co_initial_guess

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param t_env ...
! **************************************************************************************************
   SUBROUTINE find_contributions(qs_env, t_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(tddfpt_env_type), INTENT(IN)                  :: t_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'find_contributions', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, n_ev, n_spins, occ, output_unit, &
                                                            spin, virt
      INTEGER, DIMENSION(2)                              :: nhomos, nlumos, nrows
      REAL(KIND=dp)                                      :: contribution, summed_contributions
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: homo_coeff_col, lumo_coeff_col
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: S_lumos
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(tddfpt_control_type)                          :: t_control

      NULLIFY (S_lumos, matrix_s, dft_control)
      output_unit = cp_logger_get_default_io_unit()
      CALL get_qs_env(qs_env, matrix_s=matrix_s, dft_control=dft_control)

      IF (output_unit > 0) WRITE (output_unit, *)
      IF (output_unit > 0) WRITE (output_unit, *)

      t_control = dft_control%tddfpt_control
      n_ev = t_control%n_ev
      n_spins = dft_control%nspins

      ALLOCATE (S_lumos(n_spins))

      DO spin = 1, n_spins
         nrows(spin) = t_control%lumos(spin)%matrix%matrix_struct%nrow_global
         nhomos(spin) = t_env%evecs(1, spin)%matrix%matrix_struct%ncol_global
         nlumos(spin) = t_control%lumos(spin)%matrix%matrix_struct%ncol_global
         CALL cp_fm_create(S_lumos(spin)%matrix, t_control%lumos(spin)%matrix%matrix_struct, &
                           "S times lumos")
         CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, t_control%lumos(spin)%matrix, &
                                      S_lumos(spin)%matrix, nlumos(spin), 1.0_dp, 0.0_dp)
      END DO

      ALLOCATE (homo_coeff_col(MAXVAL(nrows(1:n_spins)), 1), &
                lumo_coeff_col(MAXVAL(nrows(1:n_spins)), 1))
      DO i = 1, n_ev
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(A,I3,5X,F15.6)') "  excited state : ", i, t_env%evals(i)*evolt
            WRITE (output_unit, *)
         END IF
         summed_contributions = 0.0_dp
         DO spin = 1, n_spins
            IF (n_spins == 2) THEN
               IF (spin == 1) THEN
                  IF (output_unit > 0) WRITE (output_unit, *) 'alpha:'
               ELSE
                  IF (output_unit > 0) WRITE (output_unit, *) 'beta:'
               END IF
            END IF
            searchloop: DO occ = nhomos(spin), 1, -1
               CALL cp_fm_get_submatrix(t_env%evecs(i, spin)%matrix, homo_coeff_col, &
                                        1, occ, nrows(spin), 1)
               DO virt = 1, nlumos(spin)
                  CALL cp_fm_get_submatrix(S_lumos(spin)%matrix, lumo_coeff_col, &
                                           1, virt, nrows(spin), 1)
                  contribution = 0.0_dp
                  DO j = 1, nrows(spin)
                     contribution = contribution+homo_coeff_col(j, 1)*lumo_coeff_col(j, 1)
                  END DO
                  summed_contributions = summed_contributions+(contribution)**2
                  IF (ABS(contribution) > 5.0e-2_dp) THEN
                     IF (output_unit > 0) WRITE (output_unit, '(14X,I5,A,I5,10X,F8.3,5X,F8.3)') &
                        occ, " ->", nhomos(spin)+virt, ABS(contribution), summed_contributions
                  END IF
                  IF (ABS(summed_contributions-1.0_dp) < 1.0e-3_dp) CYCLE searchloop
               END DO
            END DO searchloop
         END DO
         IF (output_unit > 0) WRITE (output_unit, *)
      END DO

      !
      ! punch a checksum for the regs
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A,E14.6)') ' TDDFPT : CheckSum  =', SQRT(SUM(t_env%evals**2))
      ENDIF

      DO spin = 1, n_spins
         CALL cp_fm_release(S_lumos(spin)%matrix)
      END DO
      DEALLOCATE (homo_coeff_col, lumo_coeff_col)

      DEALLOCATE (S_lumos)

   END SUBROUTINE find_contributions

! **************************************************************************************************
!> \brief ...
!> \param X ...
!> \param tmp_vec ...
!> \param metric ...
! **************************************************************************************************
   SUBROUTINE normalize(X, tmp_vec, metric)

      TYPE(cp_fm_p_type), DIMENSION(:)                   :: x, tmp_vec
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: metric

      INTEGER                                            :: n_spins, spin
      REAL(KIND=dp)                                      :: norm, tmp

      n_spins = SIZE(x)
      norm = 0.0_dp

      DO spin = 1, n_spins
         tmp = 0.0_dp
         CALL cp_dbcsr_sm_fm_multiply(metric(1)%matrix, X(spin)%matrix, &
                                      tmp_vec(spin)%matrix, &
                                      X(spin)%matrix%matrix_struct%ncol_global, &
                                      1.0_dp, 0.0_dp)
         CALL cp_fm_trace(X(spin)%matrix, tmp_vec(spin)%matrix, tmp)
         norm = norm+tmp
      END DO

      norm = SQRT(norm)
      DO spin = 1, n_spins
         CALL cp_fm_scale((1.0_dp/norm), X(spin)%matrix)
      END DO

   END SUBROUTINE normalize

   !---------------------------------------!
   ! x must not be changed in this routine !
   ! tmp_vec may be changed                !
   !---------------------------------------!
! **************************************************************************************************
!> \brief ...
!> \param X ...
!> \param V_set ...
!> \param SV_set ...
!> \param work ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE reorthogonalize(X, V_set, SV_set, work, n)

      TYPE(cp_fm_p_type), DIMENSION(:)                   :: X
      TYPE(cp_fm_p_type), DIMENSION(:, :)                :: V_set, SV_set
      TYPE(cp_fm_p_type), DIMENSION(:)                   :: work
      INTEGER, INTENT(IN)                                :: n

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reorthogonalize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, n_spins, spin
      REAL(DP)                                           :: dot_product, tmp

      CALL timeset(routineN, handle)

      IF (n > 0) THEN

         n_spins = SIZE(X)
         DO spin = 1, n_spins
            CALL cp_fm_to_fm(X(spin)%matrix, work(spin)%matrix)
         END DO

         DO i = 1, n
            dot_product = 0.0_dp
            DO spin = 1, n_spins
               CALL cp_fm_trace(SV_set(i, spin)%matrix, work(spin)%matrix, tmp)
               dot_product = dot_product+tmp
            END DO
            DO spin = 1, n_spins
               CALL cp_fm_scale_and_add(1.0_dp, X(spin)%matrix, &
                                        -1.0_dp*dot_product, V_set(i, spin)%matrix)
            END DO
         END DO

      END IF

      CALL timestop(handle)

   END SUBROUTINE reorthogonalize

! **************************************************************************************************

END MODULE qs_tddfpt_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE qs_tddfpt_eigensolver
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type,&
                                              tddfpt_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type,&
                                              cp_dbcsr_set
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_plus_fm_fm_t,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_scale_and_add,&
                                              cp_fm_symm,&
                                              cp_fm_trace
   USE cp_fm_diag,                      ONLY: cp_fm_syevd
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              fm_pools_create_fm_vect,&
                                              fm_pools_give_back_fm_vect
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_p_type,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_element,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_set_element,&
                                              cp_fm_to_fm
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit,&
                                              cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: tddfpt_davidson,&
                                              tddfpt_lanczos
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE physcon,                         ONLY: evolt
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_matrix_pools,                 ONLY: mpools_get
   USE qs_p_env_methods,                ONLY: p_op_l1,&
                                              p_op_l2,&
                                              p_postortho,&
                                              p_preortho
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_tddfpt_types,                 ONLY: tddfpt_env_type
   USE qs_tddfpt_utils,                 ONLY: co_initial_guess,&
                                              normalize,&
                                              reorthogonalize
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_tddfpt_eigensolver'

   PRIVATE

   PUBLIC :: eigensolver

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param p_env ...
!> \param qs_env ...
!> \param t_env ...
! **************************************************************************************************
   SUBROUTINE eigensolver(p_env, qs_env, t_env)

      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(tddfpt_env_type), INTENT(INOUT)               :: t_env

      CHARACTER(len=*), PARAMETER :: routineN = 'eigensolver', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, n_ev, nspins, output_unit, &
                                                            restarts
      LOGICAL                                            :: do_kernel_save
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ievals
      TYPE(dft_control_type), POINTER                    :: dft_control

      CALL timeset(routineN, handle)

      NULLIFY (dft_control)

      output_unit = cp_logger_get_default_io_unit()

      CALL get_qs_env(qs_env, dft_control=dft_control)
      n_ev = dft_control%tddfpt_control%n_ev
      nspins = dft_control%nspins

      ALLOCATE (ievals(n_ev))

      !---------------!
      ! initial guess !
      !---------------!
      do_kernel_save = dft_control%tddfpt_control%do_kernel
      dft_control%tddfpt_control%do_kernel = .FALSE.
      IF (output_unit > 0) THEN
         WRITE (output_unit, *) " Generating initial guess"
         WRITE (output_unit, *)
      END IF
      IF (ASSOCIATED(dft_control%tddfpt_control%lumos)) THEN
         CALL co_initial_guess(t_env%evecs, ievals, n_ev, qs_env)
      ELSE
         IF (output_unit > 0) WRITE (output_unit, *) "LUMOS are needed in TDDFPT!"
         CALL cp__b("qs_tddfpt_eigensolver.F",110,"")
      END IF
      DO restarts = 1, dft_control%tddfpt_control%n_restarts
         IF (iterative_solver(ievals, t_env, p_env, qs_env, ievals)) EXIT
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) " Restarting"
            WRITE (output_unit, *)
         END IF
      END DO
      dft_control%tddfpt_control%do_kernel = do_kernel_save

      !-----------------!
      ! call the solver !
      !-----------------!
      IF (output_unit > 0) THEN
         WRITE (output_unit, *)
         WRITE (output_unit, *) " Doing TDDFPT calculation"
         WRITE (output_unit, *)
      END IF
      DO restarts = 1, dft_control%tddfpt_control%n_restarts
         IF (iterative_solver(ievals, t_env, p_env, qs_env, t_env%evals)) EXIT
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) " Restarting"
            WRITE (output_unit, *)
         END IF
      END DO

      !---------!
      ! cleanup !
      !---------!
      DEALLOCATE (ievals)

      CALL timestop(handle)

   END SUBROUTINE eigensolver

   ! in_evals  : approximations to the eigenvalues for the preconditioner
   ! t_env     : TD-DFT environment values
   ! p_env     : perturbation environment values
   ! qs_env    : general Quickstep environment values
   ! out_evals : the resulting eigenvalues
   ! error     : used for error handling
   !
   ! res       : the function will return wheter the eigenvalues are converged or not

! **************************************************************************************************
!> \brief ...
!> \param in_evals ...
!> \param t_env ...
!> \param p_env ...
!> \param qs_env ...
!> \param out_evals ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION iterative_solver(in_evals, &
                             t_env, p_env, qs_env, &
                             out_evals) RESULT(res)

      REAL(KIND=dp), DIMENSION(:)                        :: in_evals
      TYPE(tddfpt_env_type), INTENT(INOUT)               :: t_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(kind=dp), DIMENSION(:), OPTIONAL              :: out_evals
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'iterative_solver', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: mode
      INTEGER                                            :: col, handle, i, iev, iter, j, k, &
                                                            max_krylovspace_dim, max_kv, n_ev, &
                                                            n_kv, nspins, output_unit, row, spin
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: must_improve
      REAL(dp)                                           :: Atilde_ij, convergence, tmp, tmp2
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: evals_difference, evals_tmp
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :)        :: evals
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s
      TYPE(cp_fm_p_type)                                 :: Atilde, Us
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: R, X
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: Ab, b, Sb
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools
      TYPE(cp_fm_struct_p_type), DIMENSION(:), POINTER   :: kv_fm_struct
      TYPE(cp_fm_struct_type), POINTER                   :: tilde_fm_struct
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(tddfpt_control_type), POINTER                 :: tddfpt_control

      res = .FALSE.

      CALL timeset(routineN, handle)

      NULLIFY (ao_mo_fm_pools, X, R, b, Ab, Sb, tddfpt_control, &
               tilde_fm_struct, Atilde%matrix, Us%matrix, matrix_s, dft_control, &
               para_env, blacs_env)

      CALL get_qs_env(qs_env, &
                      matrix_s=matrix_s, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      blacs_env=blacs_env)

      tddfpt_control => dft_control%tddfpt_control
      output_unit = cp_logger_get_default_io_unit()
      n_ev = tddfpt_control%n_ev
      nspins = dft_control%nspins

      IF (dft_control%tddfpt_control%diag_method == tddfpt_lanczos) THEN
         mode = 'L'
      ELSE IF (dft_control%tddfpt_control%diag_method == tddfpt_davidson) THEN
         mode = 'D'
      END IF

      !-----------------------------------------!
      ! determine the size of the problem       !
      ! and how many krylov space vetors to use !
      !-----------------------------------------!
      max_krylovspace_dim = SUM(p_env%n_ao(1:nspins)*p_env%n_mo(1:nspins))
      max_kv = tddfpt_control%max_kv
      IF (max_krylovspace_dim <= max_kv) THEN
         max_kv = max_krylovspace_dim
         IF (output_unit > 0) THEN
            WRITE (output_unit, *) "  Setting the maximum number of krylov vectors to ", max_kv, "!!"
         END IF
      END IF
!    CPPostcondition(max_krylovspace_dim>=max_kv,cp_failure_level,routineP,failure)

      !----------------------!
      ! allocate the vectors !
      !----------------------!
      CALL mpools_get(qs_env%mpools, ao_mo_fm_pools=ao_mo_fm_pools)
      CALL fm_pools_create_fm_vect(ao_mo_fm_pools, X, name=routineP//":X")
      CALL fm_pools_create_fm_vect(ao_mo_fm_pools, R, name=routineP//":R")

      ALLOCATE (evals_difference(n_ev))

      ALLOCATE (must_improve(n_ev))

      ALLOCATE (evals(max_kv, 0:max_kv))
      ALLOCATE (evals_tmp(max_kv))

      ALLOCATE (b(max_kv, nspins), Ab(max_kv, nspins), &
                Sb(max_kv, nspins))

      ALLOCATE (kv_fm_struct(nspins))

      DO spin = 1, nspins
         NULLIFY (kv_fm_struct(spin)%struct)
         CALL cp_fm_struct_create(kv_fm_struct(spin)%struct, para_env, blacs_env, &
                                  p_env%n_ao(spin), p_env%n_mo(spin))
      END DO
      DO spin = 1, nspins
         DO i = 1, max_kv
            NULLIFY (b(i, spin)%matrix, Ab(i, spin)%matrix, Sb(i, spin)%matrix)
         END DO
      END DO

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(2X,A,T69,A)') &
            "nvec", "Convergence"
         WRITE (output_unit, '(2X,A)') &
            "-----------------------------------------------------------------------------"
      END IF

      iter = 1
      k = 0
      n_kv = n_ev
      iteration: DO

         CALL allocate_krylov_vectors(b, "b-", k+1, n_kv, nspins, kv_fm_struct)
         CALL allocate_krylov_vectors(Ab, "Ab-", k+1, n_kv, nspins, kv_fm_struct)
         CALL allocate_krylov_vectors(Sb, "Sb-", k+1, n_kv, nspins, kv_fm_struct)

         DO i = 1, n_kv
            k = k+1

            IF (k <= SIZE(t_env%evecs, 1)) THEN ! the first iteration

               ! take the initial guess
               DO spin = 1, nspins
                  CALL cp_fm_to_fm(t_env%evecs(k, spin)%matrix, b(k, spin)%matrix)
               END DO

            ELSE

               ! create a new vector
               IF (mode == 'L') THEN

                  DO spin = 1, nspins
                     IF (tddfpt_control%invert_S) THEN
                        CALL cp_fm_symm('L', 'U', p_env%n_ao(spin), p_env%n_mo(spin), &
                                        1.0_dp, t_env%invS(spin)%matrix, Ab(k-1, spin)%matrix, &
                                        0.0_dp, b(k, spin)%matrix)
                     ELSE
                        CALL cp_fm_to_fm(Ab(k-1, spin)%matrix, b(k, spin)%matrix)
                     END IF
                  END DO

               ELSE IF (mode == 'D') THEN

                  iev = must_improve(i)
                  ! create the new davidson vector
                  DO spin = 1, nspins

                     CALL cp_fm_set_all(R(spin)%matrix, 0.0_dp)
                     DO j = 1, k-i
                        CALL cp_fm_to_fm(Ab(j, spin)%matrix, X(spin)%matrix)
                        CALL cp_fm_scale_and_add(1.0_dp, X(spin)%matrix, &
                                                 -evals(iev, iter-1), Sb(j, spin)%matrix)
                        CALL cp_fm_get_element(Us%matrix, j, iev, tmp)
                        CALL cp_fm_scale_and_add(1.0_dp, R(spin)%matrix, &
                                                 tmp, X(spin)%matrix)
                     END DO

                     IF (tddfpt_control%invert_S) THEN
                        CALL cp_fm_symm('L', 'U', p_env%n_ao(spin), p_env%n_mo(spin), &
                                        1.0_dp, t_env%invS(spin)%matrix, R(spin)%matrix, &
                                        0.0_dp, X(spin)%matrix)
                     ELSE
                        CALL cp_fm_to_fm(R(spin)%matrix, X(spin)%matrix)
                     END IF

                     !----------------!
                     ! preconditioner !
                     !----------------!
                     IF (dft_control%tddfpt_control%precond) THEN
                        DO col = 1, p_env%n_mo(spin)
                           IF (col <= n_ev) THEN
                              tmp2 = ABS(evals(iev, iter-1)-in_evals(col))
                           ELSE
                              tmp2 = ABS(evals(iev, iter-1)-(in_evals(n_ev)+10.0_dp))
                           END IF
                           ! protect against division by 0 by a introducing a cutoff.
                           tmp2 = MAX(tmp2, 100*EPSILON(1.0_dp))
                           DO row = 1, p_env%n_ao(spin)
                              CALL cp_fm_get_element(X(spin)%matrix, row, col, tmp)
                              CALL cp_fm_set_element(b(k, spin)%matrix, row, col, tmp/tmp2)
                           END DO
                        END DO
                     ELSE
                        CALL cp_fm_to_fm(X(spin)%matrix, b(k, spin)%matrix)
                     END IF

                  END DO

               ELSE
                  IF (output_unit > 0) WRITE (output_unit, *) "unknown mode"
                  CALL cp__b("qs_tddfpt_eigensolver.F",357,"")
               END IF

            END IF

            CALL p_preortho(p_env, qs_env, b(k, :))
            DO j = 1, tddfpt_control%n_reortho
               CALL reorthogonalize(b(k, :), b, Sb, R, k-1) ! R is temp
            END DO
            CALL normalize(b(k, :), R, matrix_s) ! R is temp
            DO spin = 1, nspins
               CALL cp_fm_to_fm(b(k, spin)%matrix, X(spin)%matrix)
            END DO
            CALL apply_op(X, Ab(k, :), p_env, qs_env, &
                          dft_control%tddfpt_control%do_kernel)
            CALL p_postortho(p_env, qs_env, Ab(k, :))
            DO spin = 1, nspins
               CALL cp_dbcsr_sm_fm_multiply(matrix_s(1)%matrix, &
                                            b(k, spin)%matrix, &
                                            Sb(k, spin)%matrix, &
                                            p_env%n_mo(spin))
            END DO
         END DO

         !--------------------------------------------!
         ! deallocate memory for the reduced matrices !
         !--------------------------------------------!
         IF (ASSOCIATED(Atilde%matrix)) CALL cp_fm_release(Atilde%matrix)
         IF (ASSOCIATED(Us%matrix)) CALL cp_fm_release(Us%matrix)
         IF (ASSOCIATED(tilde_fm_struct)) CALL cp_fm_struct_release(tilde_fm_struct)

         !------------------------------------------!
         ! allocate memory for the reduced matrices !
         !------------------------------------------!
         CALL cp_fm_struct_create(tilde_fm_struct, para_env, blacs_env, k, k)
         CALL cp_fm_create(Atilde%matrix, &
                           tilde_fm_struct, &
                           routineP//"Atilde")
         CALL cp_fm_create(Us%matrix, &
                           tilde_fm_struct, &
                           routineP//"Us")

         !---------------------------------------!
         ! calc the matrix Atilde = transp(b)*Ab !
         !---------------------------------------!
         DO i = 1, k
            DO j = 1, k
               Atilde_ij = 0.0_dp
               DO spin = 1, nspins
                  CALL cp_fm_trace(b(i, spin)%matrix, Ab(j, spin)%matrix, tmp)
                  Atilde_ij = Atilde_ij+tmp
               END DO
               CALL cp_fm_set_element(Atilde%matrix, i, j, Atilde_ij)
            END DO
         END DO

         !--------------------!
         ! diagonalize Atilde !
         !--------------------!
         evals_tmp(:) = evals(:, iter)
         CALL cp_fm_syevd(Atilde%matrix, Us%matrix, evals_tmp(:))
         evals(:, iter) = evals_tmp(:)

         !-------------------!
         ! check convergence !
         !-------------------!
         evals_difference = 1.0_dp
         IF (iter /= 1) THEN

            evals_difference(:) = ABS((evals(1:n_ev, iter-1)-evals(1:n_ev, iter)))
            ! For debugging
            IF (output_unit > 0) THEN
               WRITE (output_unit, *)
               DO i = 1, n_ev
                  WRITE (output_unit, '(2X,F10.7,T69,ES11.4)') evals(i, iter)*evolt, evals_difference(i)
               END DO
               WRITE (output_unit, *)
            END IF

            convergence = MAXVAL(evals_difference)
            IF (output_unit > 0) WRITE (output_unit, '(2X,I4,T69,ES11.4)') k, convergence

            IF (convergence < tddfpt_control%tolerance) THEN
               res = .TRUE.
               EXIT iteration
            END IF
         END IF

         IF (mode == 'L') THEN
            n_kv = 1
         ELSE
            must_improve = 0
            DO i = 1, n_ev
               IF (evals_difference(i) > tddfpt_control%tolerance) must_improve(i) = 1
            END DO
!! Set must_improve to 1 if all the vectors should
!! be updated in one iteration.
!!          must_improve = 1
            n_kv = SUM(must_improve)
            j = 1
            DO i = 1, n_ev
               IF (must_improve(i) == 1) THEN
                  must_improve(j) = i
                  j = j+1
               END IF
            END DO
         END IF

         IF (k+n_kv > max_kv) EXIT iteration

         iter = iter+1

      END DO iteration

      IF (PRESENT(out_evals)) THEN
         out_evals(1:n_ev) = evals(1:n_ev, iter)
      END IF

      DO spin = 1, nspins
         DO j = 1, n_ev
            CALL cp_fm_set_all(t_env%evecs(j, spin)%matrix, 0.0_dp)
            DO i = 1, k
               CALL cp_fm_get_element(Us%matrix, i, j, tmp)
               CALL cp_fm_scale_and_add(1.0_dp, t_env%evecs(j, spin)%matrix, &
                                        tmp, b(i, spin)%matrix)
            END DO
         END DO
      END DO

      DO spin = 1, nspins
         DO i = 1, max_kv
            IF (ASSOCIATED(b(i, spin)%matrix)) &
               CALL cp_fm_release(b(i, spin)%matrix)
            IF (ASSOCIATED(Ab(i, spin)%matrix)) &
               CALL cp_fm_release(Ab(i, spin)%matrix)
            IF (ASSOCIATED(Sb(i, spin)%matrix)) &
               CALL cp_fm_release(Sb(i, spin)%matrix)
         END DO
      END DO

      !----------!
      ! clean up !
      !----------!
      IF (ASSOCIATED(Atilde%matrix)) CALL cp_fm_release(Atilde%matrix)
      IF (ASSOCIATED(Us%matrix)) CALL cp_fm_release(Us%matrix)
      IF (ASSOCIATED(tilde_fm_struct)) CALL cp_fm_struct_release(tilde_fm_struct)
      CALL fm_pools_give_back_fm_vect(ao_mo_fm_pools, X)
      CALL fm_pools_give_back_fm_vect(ao_mo_fm_pools, R)
      NULLIFY (X, R)
      DO spin = 1, nspins
         CALL cp_fm_struct_release(kv_fm_struct(spin)%struct)
         DO i = 1, max_kv
            IF (ASSOCIATED(b(i, spin)%matrix)) &
               CALL cp_fm_release(b(i, spin)%matrix)
            IF (ASSOCIATED(Ab(i, spin)%matrix)) &
               CALL cp_fm_release(Ab(i, spin)%matrix)
            IF (ASSOCIATED(Sb(i, spin)%matrix)) &
               CALL cp_fm_release(Sb(i, spin)%matrix)
         END DO
      END DO
      DEALLOCATE (b, Ab, Sb, evals, evals_tmp, evals_difference, must_improve, kv_fm_struct)

      CALL timestop(handle)

   END FUNCTION iterative_solver

   ! X        : the vector on which to apply the op
   ! R        : the result
   ! t_env    : td-dft environment (mainly control information)
   ! p_env    : perturbation environment (variables)
   !            both of these carry info for the tddfpt calculation
   ! qs_env   : info about a quickstep ground state calculation

! **************************************************************************************************
!> \brief ...
!> \param X ...
!> \param R ...
!> \param p_env ...
!> \param qs_env ...
!> \param do_kernel ...
! **************************************************************************************************
   SUBROUTINE apply_op(X, R, p_env, qs_env, do_kernel)

      TYPE(cp_fm_p_type), DIMENSION(:)                   :: X, R
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: do_kernel

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_op', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nspins, spin
      INTEGER, SAVE                                      :: counter = 0
      TYPE(dft_control_type), POINTER                    :: dft_control

      NULLIFY (dft_control)

      CALL timeset(routineN, handle)

      counter = counter+1
      CALL get_qs_env(qs_env, dft_control=dft_control)
      nspins = dft_control%nspins

      !------------!
      ! R = HX-SXL !
      !------------!
      CALL p_op_l1(p_env, qs_env, X, R) ! acts on both spins, result in R

      !-----------------!
      ! calc P1 and     !
      ! R = R + K(P1)*C !
      !-----------------!
      IF (do_kernel) THEN
         DO spin = 1, nspins
            CALL cp_dbcsr_set(p_env%p1(spin)%matrix, 0.0_dp) ! optimize?
            CALL cp_dbcsr_plus_fm_fm_t(p_env%p1(spin)%matrix, &
                                       matrix_v=p_env%psi0d(spin)%matrix, &
                                       matrix_g=X(spin)%matrix, &
                                       ncol=p_env%n_mo(spin), &
                                       alpha=0.5_dp)
            CALL cp_dbcsr_plus_fm_fm_t(p_env%p1(spin)%matrix, &
                                       matrix_v=X(spin)%matrix, &
                                       matrix_g=p_env%psi0d(spin)%matrix, &
                                       ncol=p_env%n_mo(spin), &
                                       alpha=0.5_dp)
         END DO
         DO spin = 1, nspins
            CALL cp_fm_set_all(X(spin)%matrix, 0.0_dp)
         END DO
         CALL p_op_l2(p_env, qs_env, p_env%p1, X, &
                      alpha=1.0_dp, beta=0.0_dp) ! X = beta*X + alpha*K(P1)*C
         DO spin = 1, nspins
            CALL cp_fm_scale_and_add(1.0_dp, R(spin)%matrix, &
                                     1.0_dp, X(spin)%matrix) ! add X to R
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE apply_op

! **************************************************************************************************
!> \brief ...
!> \param vectors ...
!> \param vectors_name ...
!> \param startv ...
!> \param n_v ...
!> \param nspins ...
!> \param fm_struct ...
! **************************************************************************************************
   SUBROUTINE allocate_krylov_vectors(vectors, vectors_name, &
                                      startv, n_v, nspins, fm_struct)

      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: vectors
      CHARACTER(LEN=*), INTENT(IN)                       :: vectors_name
      INTEGER, INTENT(IN)                                :: startv, n_v, nspins
      TYPE(cp_fm_struct_p_type), DIMENSION(:), POINTER   :: fm_struct

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_krylov_vectors', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: mat_name
      INTEGER                                            :: index, spin

      DO spin = 1, nspins
         DO index = startv, startv+n_v-1
            NULLIFY (vectors(index, spin)%matrix)
            mat_name = routineP//vectors_name//TRIM(cp_to_string(index)) &
                       //","//TRIM(cp_to_string(spin))
            CALL cp_fm_create(vectors(index, spin)%matrix, &
                              fm_struct(spin)%struct, mat_name)
            IF (.NOT. ASSOCIATED(vectors(index, spin)%matrix)) &
               CALL cp__b("qs_tddfpt_eigensolver.F",628,"Could not allocate "//TRIM(mat_name)//".")
         END DO
      END DO

   END SUBROUTINE allocate_krylov_vectors

END MODULE qs_tddfpt_eigensolver
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the saop potential
! **************************************************************************************************
MODULE xc_pot_saop
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cp_array_utils_r,                ONLY: cp_1d_r_p_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_set
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_plus_fm_fm_t
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_get_submatrix,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_set_submatrix,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: do_method_gapw,&
                                              oe_gllb,&
                                              oe_lb,&
                                              oe_saop,&
                                              xc_funct_no_shortcut
   USE input_section_types,             ONLY: &
        section_vals_create, section_vals_duplicate, section_vals_get_subs_vals, &
        section_vals_release, section_vals_retain, section_vals_set_subs_vals, section_vals_type, &
        section_vals_val_get, section_vals_val_set
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_release
   USE qs_collocate_density,            ONLY: calculate_rho_elec
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_gapw_densities,               ONLY: prepare_gapw_den
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: harmonics_atom_type
   USE qs_integrate_potential,          ONLY: integrate_v_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_atom,                      ONLY: update_ks_atom
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_local_rho_types,              ONLY: local_rho_set_create,&
                                              local_rho_set_release,&
                                              local_rho_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_rho_atom_methods,             ONLY: allocate_rho_atom_internals,&
                                              calculate_rho_atom_coeff
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE qs_vxc_atom,                     ONLY: calc_rho_angular,&
                                              gaVxcgb_noGC
   USE util,                            ONLY: get_limit
   USE virial_types,                    ONLY: virial_type
   USE xc,                              ONLY: xc_vxc_pw_create
   USE xc_atom,                         ONLY: fill_rho_set,&
                                              vxc_of_r_new,&
                                              xc_rho_set_atom_update
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_create,&
                                              xc_dset_get_derivative,&
                                              xc_dset_release,&
                                              xc_dset_zero_all
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_derivatives,                  ONLY: xc_functionals_eval
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_setall,&
                                              xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_create,&
                                              xc_rho_set_release,&
                                              xc_rho_set_type,&
                                              xc_rho_set_update
   USE xc_xbecke88,                     ONLY: xb88_lda_info,&
                                              xb88_lsd_info
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: add_saop_pot

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_pot_saop'

   ! should be eliminated
   REAL(KIND=dp), PARAMETER :: alpha = 1.19_dp, beta = 0.01_dp, K_rho = 0.42_dp
   REAL(KIND=dp), PARAMETER :: kappa = 0.804_dp, mu = 0.21951_dp, &
                               beta_ec = 0.066725_dp, gamma_saop = 0.031091_dp

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ks_matrix ...
!> \param qs_env ...
!> \param oe_corr ...
! **************************************************************************************************
   SUBROUTINE add_saop_pot(ks_matrix, qs_env, oe_corr)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix
      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: oe_corr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'add_saop_pot', routineP = moduleN//':'//routineN

      INTEGER                                            :: dft_method_id, homo, i, ispin, j, k, &
                                                            nspins, orb, xc_deriv_method_id, &
                                                            xc_rho_smooth_id
      INTEGER, DIMENSION(2)                              :: ncol, nrow
      INTEGER, DIMENSION(2, 3)                           :: bo
      LOGICAL                                            :: compute_virial, gapw, lsd
      REAL(KIND=dp)                                      :: density_cut, efac, gradient_cut, &
                                                            tau_cut, tot_rho_psi, we_GLLB, we_LB, &
                                                            xc_energy
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: coeff_col
      REAL(KIND=dp), DIMENSION(3, 3)                     :: virial_xc_tmp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_uniform
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: orbital_density_matrix, rho_struct_ao
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: single_mo_coeff
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: molecular_orbitals
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: orbital, orbital_g
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r, rho_struct_r, tau, &
                                                            vxc_GLLB, vxc_LB, vxc_SAOP, vxc_tau, &
                                                            vxc_tmp
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho_struct
      TYPE(section_vals_type), POINTER                   :: input, xc_fun_section_orig, &
                                                            xc_fun_section_tmp, xc_section_orig, &
                                                            xc_section_tmp
      TYPE(virial_type), POINTER                         :: virial
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_derivative_type), POINTER                  :: deriv
      TYPE(xc_rho_cflags_type)                           :: needs
      TYPE(xc_rho_set_type), POINTER                     :: rho_set

      NULLIFY (ks_env, pw_env, auxbas_pw_pool, input)
      NULLIFY (rho_g, rho_r, tau, rho_struct, rho_set, e_uniform)
      NULLIFY (vxc_GLLB, vxc_LB, vxc_tmp, vxc_SAOP, vxc_tau)
      NULLIFY (mo_eigenvalues, deriv_set, deriv, rho_struct_r, rho_struct_ao)
      NULLIFY (orbital_density_matrix, dummy, xc_section_tmp, xc_fun_section_tmp)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      rho=rho_struct, &
                      pw_env=pw_env, &
                      input=input, &
                      virial=virial, &
                      mos=molecular_orbitals)
      compute_virial = virial%pv_calculate .AND. (.NOT. virial%pv_numer)
      CALL section_vals_val_get(input, "DFT%QS%METHOD", i_val=dft_method_id)
      gapw = (dft_method_id == do_method_gapw)

      xc_section_orig => section_vals_get_subs_vals(input, "DFT%XC")
      CALL section_vals_retain(xc_section_orig)
      CALL section_vals_duplicate(xc_section_orig, xc_section_tmp)

      CALL section_vals_val_get(xc_section_orig, "DENSITY_CUTOFF", &
                                r_val=density_cut)
      CALL section_vals_val_get(xc_section_orig, "GRADIENT_CUTOFF", &
                                r_val=gradient_cut)
      CALL section_vals_val_get(xc_section_orig, "TAU_CUTOFF", &
                                r_val=tau_cut)

      CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool)

      CALL section_vals_val_get(input, "DFT%LSD", l_val=lsd)
      IF (lsd) THEN
         nspins = 2
      ELSE
         nspins = 1
      END IF

      ALLOCATE (rho_r(nspins), single_mo_coeff(nspins))
      CALL cp_dbcsr_allocate_matrix_set(orbital_density_matrix, nspins)
      CALL qs_rho_get(rho_struct, rho_r=rho_struct_r, rho_ao=rho_struct_ao)
      DO ispin = 1, nspins
         rho_r(ispin)%pw => rho_struct_r(ispin)%pw
         ALLOCATE (orbital_density_matrix(ispin)%matrix)
         CALL cp_dbcsr_init(orbital_density_matrix(ispin)%matrix)
         CALL cp_dbcsr_copy(orbital_density_matrix(ispin)%matrix, &
                            rho_struct_ao(ispin)%matrix, "orbital density")
      END DO
      bo = rho_r(1)%pw%pw_grid%bounds_local

      !---------------------------!
      ! create the density needed !
      !---------------------------!
      CALL xc_rho_set_create(rho_set, bo, &
                             density_cut, &
                             gradient_cut, &
                             tau_cut)
      CALL xc_rho_cflags_setall(needs, .FALSE.)
      IF (lsd) THEN
         CALL xb88_lsd_info(needs=needs)
         needs%norm_drho = .TRUE.
      ELSE
         CALL xb88_lda_info(needs=needs)
      END IF
      CALL section_vals_val_get(xc_section_orig, "XC_GRID%XC_DERIV", &
                                i_val=xc_deriv_method_id)
      CALL section_vals_val_get(xc_section_orig, "XC_GRID%XC_SMOOTH_RHO", &
                                i_val=xc_rho_smooth_id)
      CALL xc_rho_set_update(rho_set, rho_r, rho_g, tau, needs, &
                             xc_deriv_method_id, &
                             xc_rho_smooth_id, &
                             auxbas_pw_pool)

      !----------------------------------------!
      ! Construct the LB94 potential in vxc_LB !
      !----------------------------------------!
      xc_fun_section_orig => section_vals_get_subs_vals(xc_section_orig, &
                                                        "XC_FUNCTIONAL")
      CALL section_vals_create(xc_fun_section_tmp, xc_fun_section_orig%section)
      CALL section_vals_val_set(xc_fun_section_tmp, "_SECTION_PARAMETERS_", &
                                i_val=xc_funct_no_shortcut)
      CALL section_vals_val_set(xc_fun_section_tmp, "XALPHA%_SECTION_PARAMETERS_", &
                                l_val=.TRUE.)
      CALL section_vals_set_subs_vals(xc_section_tmp, "XC_FUNCTIONAL", &
                                      xc_fun_section_tmp)

      IF(.NOT.(.NOT. compute_virial))CALL cp__a("xc_pot_saop.F",250)
      CALL xc_vxc_pw_create(vxc_tmp, vxc_tau, xc_energy, rho_r, rho_g, tau, &
                            xc_section_tmp, auxbas_pw_pool, &
                            compute_virial=.FALSE., virial_xc=virial_xc_tmp)

      CALL section_vals_val_set(xc_fun_section_tmp, "XALPHA%_SECTION_PARAMETERS_", &
                                l_val=.FALSE.)
      CALL section_vals_val_set(xc_fun_section_tmp, "PZ81%_SECTION_PARAMETERS_", &
                                l_val=.TRUE.)

      IF(.NOT.(.NOT. compute_virial))CALL cp__a("xc_pot_saop.F",260)
      CALL xc_vxc_pw_create(vxc_LB, vxc_tau, xc_energy, rho_r, rho_g, tau, &
                            xc_section_tmp, auxbas_pw_pool, &
                            compute_virial=.FALSE., virial_xc=virial_xc_tmp)

      DO ispin = 1, nspins
         vxc_LB(ispin)%pw%cr3d = vxc_LB(ispin)%pw%cr3d+alpha*vxc_tmp(ispin)%pw%cr3d
      END DO

      DO ispin = 1, nspins
         dummy => vxc_tmp(ispin)%pw%cr3d
         CALL add_lb_pot(dummy, rho_set, lsd, ispin)
         vxc_LB(ispin)%pw%cr3d = vxc_LB(ispin)%pw%cr3d-vxc_tmp(ispin)%pw%cr3d
      END DO
      NULLIFY (dummy)

      !-----------------------------------------------------------------------------------!
      ! Construct 2 times PBE one particle density from the PZ correlation energy density !
      !-----------------------------------------------------------------------------------!
      CALL xc_dset_create(deriv_set, local_bounds=bo)
      CALL xc_functionals_eval(xc_fun_section_tmp, &
                               lsd=lsd, &
                               rho_set=rho_set, &
                               deriv_set=deriv_set, &
                               deriv_order=0)

      deriv => xc_dset_get_derivative(deriv_set, "")
      CALL xc_derivative_get(deriv, deriv_data=e_uniform)

      ALLOCATE (vxc_GLLB(nspins))
      DO ispin = 1, nspins
         CALL pw_pool_create_pw(auxbas_pw_pool, vxc_GLLB(ispin)%pw, &
                                use_data=REALDATA3D, &
                                in_space=REALSPACE)
      END DO

      DO ispin = 1, nspins
         dummy => vxc_GLLB(ispin)%pw%cr3d
         CALL calc_2excpbe(dummy, rho_set, e_uniform, lsd)
      END DO
      NULLIFY (dummy)

      CALL xc_dset_release(deriv_set)

      CALL pw_pool_create_pw(auxbas_pw_pool, orbital%pw, &
                             use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL pw_pool_create_pw(auxbas_pw_pool, orbital_g%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      DO ispin = 1, nspins

         CALL get_mo_set(molecular_orbitals(ispin)%mo_set, &
                         mo_coeff=mo_coeff, &
                         eigenvalues=mo_eigenvalues, &
                         homo=homo)
         CALL cp_fm_create(single_mo_coeff(ispin)%matrix, &
                           mo_coeff%matrix_struct, &
                           "orbital density matrix")

         CALL cp_fm_get_info(single_mo_coeff(ispin)%matrix, &
                             nrow_global=nrow(ispin), ncol_global=ncol(ispin))
         ALLOCATE (coeff_col(nrow(ispin), 1))

         CALL pw_zero(vxc_tmp(ispin)%pw)

         DO orb = 1, homo-1

            efac = K_rho*SQRT(mo_eigenvalues(homo)-mo_eigenvalues(orb))
            IF (.NOT. lsd) efac = 2.0_dp*efac

            CALL cp_fm_set_all(single_mo_coeff(ispin)%matrix, 0.0_dp)
            CALL cp_fm_get_submatrix(mo_coeff, coeff_col, &
                                     1, orb, nrow(ispin), 1)
            CALL cp_fm_set_submatrix(single_mo_coeff(ispin)%matrix, coeff_col, &
                                     1, orb)
            CALL cp_dbcsr_set(orbital_density_matrix(ispin)%matrix, 0.0_dp)
            CALL cp_dbcsr_plus_fm_fm_t(orbital_density_matrix(ispin)%matrix, &
                                       matrix_v=single_mo_coeff(ispin)%matrix, &
                                       matrix_g=single_mo_coeff(ispin)%matrix, &
                                       ncol=ncol(ispin), &
                                       alpha=1.0_dp)
            CALL pw_zero(orbital%pw)
            CALL pw_zero(orbital_g%pw)
            CALL calculate_rho_elec(matrix_p=orbital_density_matrix(ispin)%matrix, &
                                    rho=orbital, rho_gspace=orbital_g, &
                                    total_rho=tot_rho_psi, ks_env=ks_env)

            vxc_tmp(ispin)%pw%cr3d = vxc_tmp(ispin)%pw%cr3d+ &
                                     efac*orbital%pw%cr3d

         END DO
         DEALLOCATE (coeff_col)

         DO k = bo(1, 3), bo(2, 3)
            DO j = bo(1, 2), bo(2, 2)
               DO i = bo(1, 1), bo(2, 1)
                  IF (rho_r(ispin)%pw%cr3d(i, j, k) > density_cut) THEN
                     vxc_tmp(ispin)%pw%cr3d(i, j, k) = vxc_tmp(ispin)%pw%cr3d(i, j, k)/ &
                                                       rho_r(ispin)%pw%cr3d(i, j, k)
                  ELSE
                     vxc_tmp(ispin)%pw%cr3d(i, j, k) = 0.0_dp
                  END IF
               END DO
            END DO
         END DO

         vxc_GLLB(ispin)%pw%cr3d = vxc_GLLB(ispin)%pw%cr3d+vxc_tmp(ispin)%pw%cr3d

      END DO

      !---------------!
      ! Assemble SAOP !
      !---------------!
      ALLOCATE (vxc_SAOP(nspins))

      DO ispin = 1, nspins

         CALL get_mo_set(molecular_orbitals(ispin)%mo_set, &
                         mo_coeff=mo_coeff, &
                         eigenvalues=mo_eigenvalues, &
                         homo=homo)
         CALL pw_pool_create_pw(auxbas_pw_pool, vxc_SAOP(ispin)%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_zero(vxc_SAOP(ispin)%pw)

         ALLOCATE (coeff_col(nrow(ispin), 1))

         DO orb = 1, homo

            we_LB = EXP(-2.0_dp*(mo_eigenvalues(homo)-mo_eigenvalues(orb))**2)
            we_GLLB = 1.0_dp-we_LB
            IF (.NOT. lsd) THEN
               we_LB = 2.0_dp*we_LB
               we_GLLB = 2.0_dp*we_GLLB
            END IF

            vxc_tmp(ispin)%pw%cr3d = we_LB*vxc_LB(ispin)%pw%cr3d+ &
                                     we_GLLB*vxc_GLLB(ispin)%pw%cr3d

            CALL cp_fm_set_all(single_mo_coeff(ispin)%matrix, 0.0_dp)
            CALL cp_fm_get_submatrix(mo_coeff, coeff_col, &
                                     1, orb, nrow(ispin), 1)
            CALL cp_fm_set_submatrix(single_mo_coeff(ispin)%matrix, coeff_col, &
                                     1, orb)
            CALL cp_dbcsr_set(orbital_density_matrix(ispin)%matrix, 0.0_dp)
            CALL cp_dbcsr_plus_fm_fm_t(orbital_density_matrix(ispin)%matrix, &
                                       matrix_v=single_mo_coeff(ispin)%matrix, &
                                       matrix_g=single_mo_coeff(ispin)%matrix, &
                                       ncol=ncol(ispin), &
                                       alpha=1.0_dp)
            CALL pw_zero(orbital%pw)
            CALL pw_zero(orbital_g%pw)
            CALL calculate_rho_elec(matrix_p=orbital_density_matrix(ispin)%matrix, &
                                    rho=orbital, rho_gspace=orbital_g, &
                                    total_rho=tot_rho_psi, ks_env=ks_env)

!TC          CALL calculate_wavefunction(mo_coeff,orb,orbital2,orbital2_g,qs_env)
!TC          write (*,*) orb, maxval(abs(orbital2%pw%cr3d-orbital%pw%cr3d))

            vxc_SAOP(ispin)%pw%cr3d = vxc_SAOP(ispin)%pw%cr3d+ &
                                      orbital%pw%cr3d*vxc_tmp(ispin)%pw%cr3d

         END DO

         CALL cp_fm_release(single_mo_coeff(ispin)%matrix)
         CALL cp_dbcsr_deallocate_matrix(orbital_density_matrix(ispin)%matrix)

         DEALLOCATE (coeff_col)

         DO k = bo(1, 3), bo(2, 3)
            DO j = bo(1, 2), bo(2, 2)
               DO i = bo(1, 1), bo(2, 1)
                  IF (rho_r(ispin)%pw%cr3d(i, j, k) > density_cut) THEN
                     vxc_SAOP(ispin)%pw%cr3d(i, j, k) = vxc_SAOP(ispin)%pw%cr3d(i, j, k)/ &
                                                        rho_r(ispin)%pw%cr3d(i, j, k)
                  ELSE
                     vxc_SAOP(ispin)%pw%cr3d(i, j, k) = 0.0_dp
                  END IF
               END DO
            END DO
         END DO

      END DO

      CALL xc_rho_set_release(rho_set, auxbas_pw_pool)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, orbital%pw)
      CALL pw_pool_give_back_pw(auxbas_pw_pool, orbital_g%pw)

      !--------------------!
      ! Do the integration !
      !--------------------!
      DO ispin = 1, nspins

         IF (oe_corr == oe_lb) THEN
            vxc_SAOP(ispin)%pw%cr3d = vxc_LB(ispin)%pw%cr3d
         ELSE IF (oe_corr == oe_gllb) THEN
            vxc_SAOP(ispin)%pw%cr3d = vxc_GLLB(ispin)%pw%cr3d
         END IF
         vxc_SAOP(ispin)%pw%cr3d = vxc_SAOP(ispin)%pw%cr3d*vxc_SAOP(ispin)%pw%pw_grid%dvol

         CALL integrate_v_rspace(v_rspace=vxc_SAOP(ispin), pmat=rho_struct_ao(ispin), &
                                 hmat=ks_matrix(ispin), qs_env=qs_env, &
                                 calculate_forces=.FALSE., &
                                 gapw=gapw)

      END DO

      DO ispin = 1, nspins
         CALL pw_pool_give_back_pw(auxbas_pw_pool, vxc_SAOP(ispin)%pw)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, vxc_GLLB(ispin)%pw)
         CALL pw_release(vxc_LB(ispin)%pw)
         CALL pw_release(vxc_tmp(ispin)%pw)
      END DO
      DEALLOCATE (vxc_GLLB, vxc_LB, vxc_tmp, orbital_density_matrix, single_mo_coeff)

      DEALLOCATE (rho_r, vxc_SAOP)

      CALL section_vals_release(xc_fun_section_tmp)
      CALL section_vals_release(xc_section_tmp)
      CALL section_vals_release(xc_section_orig)

      !-----------------------!
      ! Call the GAPW routine !
      !-----------------------!
      IF (gapw) THEN
         CALL gapw_add_atomic_saop_pot(qs_env, oe_corr)
      END IF

   END SUBROUTINE add_saop_pot

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param oe_corr ...
! **************************************************************************************************
   SUBROUTINE gapw_add_atomic_saop_pot(qs_env, oe_corr)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      INTEGER, INTENT(IN)                                :: oe_corr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'gapw_add_atomic_saop_pot', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, iat, iatom, ikind, ir, ispin, mepos, &
                                                            na, natom, nr, ns, nspins, num_pe, orb
      INTEGER, DIMENSION(2)                              :: bo, homo, ncol, nrow
      INTEGER, DIMENSION(2, 3)                           :: bounds
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: lsd, paw_atom
      REAL(dp), DIMENSION(:, :), POINTER                 :: tau
      REAL(KIND=dp)                                      :: density_cut, efac, exc, gradient_cut, &
                                                            tau_cut, we_GLLB, we_LB
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: coeff_col
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rho_h, rho_s, weight
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_uniform, vtau, vxc_GLLB_h, &
                                                            vxc_GLLB_s, vxc_LB_h, vxc_LB_s, &
                                                            vxc_SAOP_h, vxc_SAOP_s, vxc_tmp_h, &
                                                            vxc_tmp_s
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: drho_h, drho_s, vxg
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_1d_r_p_type), DIMENSION(:), POINTER        :: mo_eigenvalues
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, orbital_density_matrix, &
                                                            rho_struct_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ksmat, psmat
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: mo_coeff, single_mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(grid_atom_type), POINTER                      :: atomic_grid
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(local_rho_type), POINTER                      :: local_rho_set
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: molecular_orbitals
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_rho_type), POINTER                         :: rho_structure
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: dr_h, dr_s, r_h, r_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER     :: r_h_d, r_s_d
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      TYPE(section_vals_type), POINTER                   :: input, xc_fun_section_orig, &
                                                            xc_fun_section_tmp, xc_section_orig, &
                                                            xc_section_tmp
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_derivative_type), POINTER                  :: deriv
      TYPE(xc_rho_cflags_type)                           :: needs, needs_orbs
      TYPE(xc_rho_set_type), POINTER                     :: orb_rho_set_h, orb_rho_set_s, rho_set_h, &
                                                            rho_set_s

      NULLIFY (weight, rho_h, rho_s, vxc_LB_h, vxc_LB_s, vxc_GLLB_h, vxc_GLLB_s, &
               vxc_tmp_h, vxc_tmp_s, vtau, dummy, e_uniform, drho_h, drho_s, vxg, atom_list, &
               atomic_kind_set, qs_kind_set, deriv_set, deriv, atomic_grid, rho_struct_ao, &
               harmonics, molecular_orbitals, rho_structure, r_h, r_s, dr_h, dr_s, &
               r_h_d, r_s_d, rho_atom_set, rho_atom, rho_set_h, rho_set_s, para_env, &
               mo_coeff, single_mo_coeff, mo_eigenvalues, local_rho_set, matrix_ks, &
               orbital_density_matrix, orb_rho_set_h, orb_rho_set_s, vxc_SAOP_h, vxc_SAOP_s)

      ! tau is needed for fill_rho_set, but should never be used
      NULLIFY (tau)

      CALL get_qs_env(qs_env, input=input, &
                      rho=rho_structure, &
                      mos=molecular_orbitals, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      rho_atom_set=rho_atom_set, &
                      matrix_ks=matrix_ks, &
                      para_env=para_env)

      CALL qs_rho_get(rho_structure, rho_ao=rho_struct_ao)

      xc_section_orig => section_vals_get_subs_vals(input, "DFT%XC")
      CALL section_vals_retain(xc_section_orig)
      CALL section_vals_duplicate(xc_section_orig, xc_section_tmp)

      CALL section_vals_val_get(input, "DFT%LSD", l_val=lsd)
      IF (lsd) THEN
         nspins = 2
      ELSE
         nspins = 1
      END IF

      CALL section_vals_val_get(xc_section_orig, "DENSITY_CUTOFF", &
                                r_val=density_cut)
      CALL section_vals_val_get(xc_section_orig, "GRADIENT_CUTOFF", &
                                r_val=gradient_cut)
      CALL section_vals_val_get(xc_section_orig, "TAU_CUTOFF", &
                                r_val=tau_cut)

      ! remap pointer
      ns = SIZE(rho_struct_ao)
      psmat(1:ns, 1:1) => rho_struct_ao(1:ns)
      CALL calculate_rho_atom_coeff(qs_env, psmat)
      CALL prepare_gapw_den(qs_env)

      ALLOCATE (mo_coeff(nspins), single_mo_coeff(nspins), mo_eigenvalues(nspins))

      CALL cp_dbcsr_allocate_matrix_set(orbital_density_matrix, nspins)

      DO ispin = 1, nspins
         CALL get_mo_set(molecular_orbitals(ispin)%mo_set, &
                         mo_coeff=mo_coeff(ispin)%matrix, &
                         eigenvalues=mo_eigenvalues(ispin)%array, &
                         homo=homo(ispin))
         CALL cp_fm_create(single_mo_coeff(ispin)%matrix, &
                           mo_coeff(ispin)%matrix%matrix_struct, &
                           "orbital density matrix")
         CALL cp_fm_get_info(single_mo_coeff(ispin)%matrix, &
                             nrow_global=nrow(ispin), ncol_global=ncol(ispin))
         ALLOCATE (orbital_density_matrix(ispin)%matrix)
         CALL cp_dbcsr_init(orbital_density_matrix(ispin)%matrix)
         CALL cp_dbcsr_copy(orbital_density_matrix(ispin)%matrix, &
                            rho_struct_ao(ispin)%matrix, &
                            "orbital density")
      END DO
      CALL local_rho_set_create(local_rho_set)
      CALL allocate_rho_atom_internals(qs_env, local_rho_set%rho_atom_set)

      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)

         CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom, &
                          harmonics=harmonics, grid_atom=atomic_grid)
         IF (.NOT. paw_atom) CYCLE

         nr = atomic_grid%nr
         na = atomic_grid%ng_sphere
         bounds(1:2, 1:3) = 1
         bounds(2, 1) = na
         bounds(2, 2) = nr

         CALL xc_dset_create(deriv_set, local_bounds=bounds)

         CALL xc_rho_set_create(rho_set_h, bounds, density_cut, &
                                gradient_cut, tau_cut)
         CALL xc_rho_set_create(rho_set_s, bounds, density_cut, &
                                gradient_cut, tau_cut)
         CALL xc_rho_set_create(orb_rho_set_h, bounds, density_cut, &
                                gradient_cut, tau_cut)
         CALL xc_rho_set_create(orb_rho_set_s, bounds, density_cut, &
                                gradient_cut, tau_cut)

         CALL xc_rho_cflags_setall(needs, .FALSE.)
         IF (lsd) THEN
            CALL xb88_lsd_info(needs=needs)
            needs%norm_drho = .TRUE.
         ELSE
            CALL xb88_lda_info(needs=needs)
         END IF
         CALL xc_rho_set_atom_update(rho_set_h, needs, nspins, bounds)
         CALL xc_rho_set_atom_update(rho_set_s, needs, nspins, bounds)
         CALL xc_rho_cflags_setall(needs_orbs, .FALSE.)
         needs_orbs%rho = .TRUE.
         IF (lsd) needs_orbs%rho_spin = .TRUE.
         CALL xc_rho_set_atom_update(orb_rho_set_h, needs, nspins, bounds)
         CALL xc_rho_set_atom_update(orb_rho_set_s, needs, nspins, bounds)

         ALLOCATE (rho_h(1:na, 1:nspins), rho_s(1:na, 1:nspins))
         ALLOCATE (weight(1:na, 1:nr))
         ALLOCATE (vxc_LB_h(1:na, 1:nr, 1:nspins), vxc_LB_s(1:na, 1:nr, 1:nspins))
         ALLOCATE (vxc_GLLB_h(1:na, 1:nr, 1:nspins), vxc_GLLB_s(1:na, 1:nr, 1:nspins))
         ALLOCATE (vxc_tmp_h(1:na, 1:nr, 1:nspins), vxc_tmp_s(1:na, 1:nr, 1:nspins))
         ALLOCATE (vxc_SAOP_h(1:na, 1:nr, 1:nspins), vxc_SAOP_s(1:na, 1:nr, 1:nspins))
         ALLOCATE (drho_h(1:4, 1:na, 1:nr, 1:nspins), drho_s(1:4, 1:na, 1:nr, 1:nspins))

!      Distribute the atoms of this kind
         num_pe = para_env%num_pe
         mepos = para_env%mepos
         bo = get_limit(natom, num_pe, mepos)

         DO iat = 1, natom !bo(1),bo(2)
            iatom = atom_list(iat)

            rho_atom => rho_atom_set(iatom)
            NULLIFY (r_h, r_s, dr_h, dr_s, r_h_d, r_s_d)
            CALL get_rho_atom(rho_atom=rho_atom, rho_rad_h=r_h, &
                              rho_rad_s=r_s, drho_rad_h=dr_h, &
                              drho_rad_s=dr_s, rho_rad_h_d=r_h_d, &
                              rho_rad_s_d=r_s_d)
            drho_h = 0.0_dp
            drho_s = 0.0_dp

            DO ir = 1, nr
               CALL calc_rho_angular(atomic_grid, harmonics, nspins, .TRUE., &
                                     ir, r_h, r_s, rho_h, rho_s, &
                                     dr_h, dr_s, r_h_d, r_s_d, drho_h, drho_s)
               CALL fill_rho_set(rho_set_h, lsd, nspins, needs, rho_h, drho_h, tau, na, ir)
               CALL fill_rho_set(rho_set_s, lsd, nspins, needs, rho_s, drho_s, tau, na, ir)
               DO ia = 1, na
                  weight(ia, ir) = atomic_grid%wr(ir)*atomic_grid%wa(ia)
               END DO
            END DO

            !-----------------------------!
            ! 1. Slater exchange for LB94 !
            !-----------------------------!
            xc_fun_section_orig => section_vals_get_subs_vals(xc_section_orig, &
                                                              "XC_FUNCTIONAL")
            CALL section_vals_create(xc_fun_section_tmp, xc_fun_section_orig%section)
            CALL section_vals_val_set(xc_fun_section_tmp, "_SECTION_PARAMETERS_", &
                                      i_val=xc_funct_no_shortcut)
            CALL section_vals_val_set(xc_fun_section_tmp, "XALPHA%_SECTION_PARAMETERS_", &
                                      l_val=.TRUE.)
            CALL section_vals_set_subs_vals(xc_section_tmp, "XC_FUNCTIONAL", &
                                            xc_fun_section_tmp)

            !---------------------!
            ! Both: hard and soft !
            !---------------------!
            CALL xc_dset_zero_all(deriv_set)
            CALL vxc_of_r_new(xc_fun_section_tmp, rho_set_h, deriv_set, 1, needs, &
                              weight, lsd, na, nr, exc, vxc_tmp_h, vxg, vtau)
            CALL xc_dset_zero_all(deriv_set)
            CALL vxc_of_r_new(xc_fun_section_tmp, rho_set_s, deriv_set, 1, needs, &
                              weight, lsd, na, nr, exc, vxc_tmp_s, vxg, vtau)

            !-------------------------------------------!
            ! 2. PZ correlation for LB94 and ec_uniform !
            !-------------------------------------------!
            CALL section_vals_val_set(xc_fun_section_tmp, "XALPHA%_SECTION_PARAMETERS_", &
                                      l_val=.FALSE.)
            CALL section_vals_val_set(xc_fun_section_tmp, "PZ81%_SECTION_PARAMETERS_", &
                                      l_val=.TRUE.)

            !------!
            ! Hard !
            !------!
            CALL xc_dset_zero_all(deriv_set)
            CALL vxc_of_r_new(xc_fun_section_tmp, rho_set_h, deriv_set, 1, needs, &
                              weight, lsd, na, nr, exc, vxc_LB_h, vxg, vtau)
            vxc_LB_h = vxc_LB_h+alpha*vxc_tmp_h
            DO ispin = 1, nspins
               dummy => vxc_tmp_h(:, :, ispin:ispin)
               CALL add_lb_pot(dummy, rho_set_h, lsd, ispin)
               vxc_LB_h(:, :, ispin) = vxc_LB_h(:, :, ispin)-weight(:, :)*vxc_tmp_h(:, :, ispin)
            END DO
            NULLIFY (dummy)

            vxc_GLLB_h = 0.0_dp
            deriv => xc_dset_get_derivative(deriv_set, "")
            IF(.NOT.(ASSOCIATED(deriv)))CALL cp__a("xc_pot_saop.F",738)
            CALL xc_derivative_get(deriv, deriv_data=e_uniform)
            DO ispin = 1, nspins
               dummy => vxc_GLLB_h(:, :, ispin:ispin)
               CALL calc_2excpbe(dummy, rho_set_h, e_uniform, lsd)
               vxc_GLLB_h(:, :, ispin) = vxc_GLLB_h(:, :, ispin)*weight(:, :)
            END DO
            NULLIFY (deriv, dummy, e_uniform)

            !------!
            ! Soft !
            !------!
            CALL xc_dset_zero_all(deriv_set)
            CALL vxc_of_r_new(xc_fun_section_tmp, rho_set_s, deriv_set, 1, needs, &
                              weight, lsd, na, nr, exc, vxc_LB_s, vxg, vtau)

            vxc_LB_s = vxc_LB_s+alpha*vxc_tmp_s
            DO ispin = 1, nspins
               dummy => vxc_tmp_s(:, :, ispin:ispin)
               CALL add_lb_pot(dummy, rho_set_s, lsd, ispin)
               vxc_LB_s(:, :, ispin) = vxc_LB_s(:, :, ispin)-weight(:, :)*vxc_tmp_s(:, :, ispin)
            END DO
            NULLIFY (dummy)

            vxc_GLLB_s = 0.0_dp
            deriv => xc_dset_get_derivative(deriv_set, "")
            IF(.NOT.(ASSOCIATED(deriv)))CALL cp__a("xc_pot_saop.F",764)
            CALL xc_derivative_get(deriv, deriv_data=e_uniform)
            DO ispin = 1, nspins
               dummy => vxc_GLLB_s(:, :, ispin:ispin)
               CALL calc_2excpbe(dummy, rho_set_s, e_uniform, lsd)
               vxc_GLLB_s(:, :, ispin) = vxc_GLLB_s(:, :, ispin)*weight(:, :)
            END DO
            NULLIFY (deriv, dummy, e_uniform)

            !------------------!
            ! Now the orbitals !
            !------------------!
            vxc_tmp_h = 0.0_dp; vxc_tmp_s = 0.0_dp

            DO ispin = 1, nspins

               DO orb = 1, homo(ispin)-1

                  ALLOCATE (coeff_col(nrow(ispin), 1))

                  efac = K_rho*SQRT(mo_eigenvalues(ispin)%array(homo(ispin))- &
                                    mo_eigenvalues(ispin)%array(orb))
                  IF (.NOT. lsd) efac = 2.0_dp*efac

                  CALL cp_fm_set_all(single_mo_coeff(ispin)%matrix, 0.0_dp)
                  CALL cp_fm_get_submatrix(mo_coeff(ispin)%matrix, coeff_col, &
                                           1, orb, nrow(ispin), 1)
                  CALL cp_fm_set_submatrix(single_mo_coeff(ispin)%matrix, coeff_col, &
                                           1, orb)
                  CALL cp_dbcsr_set(orbital_density_matrix(ispin)%matrix, 0.0_dp)
                  CALL cp_dbcsr_plus_fm_fm_t(orbital_density_matrix(ispin)%matrix, &
                                             matrix_v=single_mo_coeff(ispin)%matrix, &
                                             matrix_g=single_mo_coeff(ispin)%matrix, &
                                             ncol=ncol(ispin), &
                                             alpha=1.0_dp)

                  DEALLOCATE (coeff_col)

                  ! This calculates the CPC and density on the grids for every atom even though
                  ! we need it only for iatom at the moment. It seems that to circumvent this,
                  ! the routines must be adapted to calculate just iatom
                  ! remap pointer
                  ns = SIZE(orbital_density_matrix)
                  psmat(1:ns, 1:1) => orbital_density_matrix(1:ns)
                  CALL calculate_rho_atom_coeff(qs_env, psmat, &
                                                local_rho_set%rho_atom_set)
                  CALL prepare_gapw_den(qs_env, local_rho_set, .FALSE.)

                  rho_atom => local_rho_set%rho_atom_set(iatom)
                  NULLIFY (r_h, r_s, dr_h, dr_s, r_h_d, r_s_d)
                  CALL get_rho_atom(rho_atom=rho_atom, &
                                    rho_rad_h=r_h, rho_rad_s=r_s)
                  DO ir = 1, nr
                     CALL calc_rho_angular(atomic_grid, harmonics, nspins, .FALSE., &
                                           ir, r_h, r_s, rho_h, rho_s, &
                                           dr_h, dr_s, r_h_d, r_s_d, drho_h, drho_s)
                     CALL fill_rho_set(orb_rho_set_h, lsd, nspins, needs_orbs, rho_h, drho_h, tau, na, ir)
                     CALL fill_rho_set(orb_rho_set_s, lsd, nspins, needs_orbs, rho_s, drho_s, tau, na, ir)
                  END DO

                  IF (lsd) THEN
                     IF (ispin == 1) THEN
                        vxc_tmp_h(:, :, 1) = vxc_tmp_h(:, :, 1)+efac*orb_rho_set_h%rhoa(:, :, 1)
                        vxc_tmp_s(:, :, 1) = vxc_tmp_s(:, :, 1)+efac*orb_rho_set_s%rhoa(:, :, 1)
                     ELSE
                        vxc_tmp_h(:, :, 2) = vxc_tmp_h(:, :, 2)+efac*orb_rho_set_h%rhob(:, :, 1)
                        vxc_tmp_s(:, :, 2) = vxc_tmp_s(:, :, 2)+efac*orb_rho_set_s%rhob(:, :, 1)
                     END IF
                  ELSE
                     vxc_tmp_h(:, :, 1) = vxc_tmp_h(:, :, 1)+efac*orb_rho_set_h%rho(:, :, 1)
                     vxc_tmp_s(:, :, 1) = vxc_tmp_s(:, :, 1)+efac*orb_rho_set_s%rho(:, :, 1)
                  END IF

               END DO ! orb

            END DO ! ispin

            IF (lsd) THEN
               DO ir = 1, nr
                  DO ia = 1, na
                     IF (rho_set_h%rhoa(ia, ir, 1) > rho_set_h%rho_cutoff) &
                        vxc_GLLB_h(ia, ir, 1) = vxc_GLLB_h(ia, ir, 1)+ &
                                                weight(ia, ir)*vxc_tmp_h(ia, ir, 1)/rho_set_h%rhoa(ia, ir, 1)
                     IF (rho_set_h%rhob(ia, ir, 1) > rho_set_h%rho_cutoff) &
                        vxc_GLLB_h(ia, ir, 2) = vxc_GLLB_h(ia, ir, 2)+ &
                                                weight(ia, ir)*vxc_tmp_h(ia, ir, 2)/rho_set_h%rhob(ia, ir, 1)
                     IF (rho_set_s%rhoa(ia, ir, 1) > rho_set_s%rho_cutoff) &
                        vxc_GLLB_s(ia, ir, 1) = vxc_GLLB_s(ia, ir, 1)+ &
                                                weight(ia, ir)*vxc_tmp_s(ia, ir, 1)/rho_set_s%rhoa(ia, ir, 1)
                     IF (rho_set_s%rhob(ia, ir, 1) > rho_set_s%rho_cutoff) &
                        vxc_GLLB_s(ia, ir, 2) = vxc_GLLB_s(ia, ir, 2)+ &
                                                weight(ia, ir)*vxc_tmp_s(ia, ir, 2)/rho_set_s%rhob(ia, ir, 1)
                  END DO
               END DO
            ELSE
               DO ir = 1, nr
                  DO ia = 1, na
                     IF (rho_set_h%rho(ia, ir, 1) > rho_set_h%rho_cutoff) &
                        vxc_GLLB_h(ia, ir, 1) = vxc_GLLB_h(ia, ir, 1)+ &
                                                weight(ia, ir)*vxc_tmp_h(ia, ir, 1)/rho_set_h%rho(ia, ir, 1)
                     IF (rho_set_s%rho(ia, ir, 1) > rho_set_s%rho_cutoff) &
                        vxc_GLLB_s(ia, ir, 1) = vxc_GLLB_s(ia, ir, 1)+ &
                                                weight(ia, ir)*vxc_tmp_s(ia, ir, 1)/rho_set_s%rho(ia, ir, 1)
                  END DO
               END DO
            END IF

            vxc_SAOP_h = 0.0_dp; vxc_SAOP_s = 0.0_dp

            DO ispin = 1, nspins

               DO orb = 1, homo(ispin)

                  ALLOCATE (coeff_col(nrow(ispin), 1))

                  we_LB = EXP(-2.0_dp*(mo_eigenvalues(ispin)%array(homo(ispin))- &
                                       mo_eigenvalues(ispin)%array(orb))**2)
                  we_GLLB = 1.0_dp-we_LB
                  IF (.NOT. lsd) THEN
                     we_LB = 2.0_dp*we_LB
                     we_GLLB = 2.0_dp*we_GLLB
                  END IF

                  vxc_tmp_h(:, :, ispin) = we_LB*vxc_LB_h(:, :, ispin)+ &
                                           we_GLLB*vxc_GLLB_h(:, :, ispin)
                  vxc_tmp_s(:, :, ispin) = we_LB*vxc_LB_s(:, :, ispin)+ &
                                           we_GLLB*vxc_GLLB_s(:, :, ispin)

                  CALL cp_fm_set_all(single_mo_coeff(ispin)%matrix, 0.0_dp)
                  CALL cp_fm_get_submatrix(mo_coeff(ispin)%matrix, coeff_col, &
                                           1, orb, nrow(ispin), 1)
                  CALL cp_fm_set_submatrix(single_mo_coeff(ispin)%matrix, coeff_col, &
                                           1, orb)
                  CALL cp_dbcsr_set(orbital_density_matrix(ispin)%matrix, 0.0_dp)
                  CALL cp_dbcsr_plus_fm_fm_t(orbital_density_matrix(ispin)%matrix, &
                                             matrix_v=single_mo_coeff(ispin)%matrix, &
                                             matrix_g=single_mo_coeff(ispin)%matrix, &
                                             ncol=ncol(ispin), &
                                             alpha=1.0_dp)

                  DEALLOCATE (coeff_col)

                  ! This calculates the CPC and density on the grids for every atom even though
                  ! we need it only for iatom at the moment. It seems that to circumvent this,
                  ! the routines must be adapted to calculate just iatom
                  ! remap pointer
                  ns = SIZE(orbital_density_matrix)
                  psmat(1:ns, 1:1) => orbital_density_matrix(1:ns)
                  CALL calculate_rho_atom_coeff(qs_env, psmat, &
                                                local_rho_set%rho_atom_set)
                  CALL prepare_gapw_den(qs_env, local_rho_set, .FALSE.)

                  rho_atom => local_rho_set%rho_atom_set(iatom)
                  NULLIFY (r_h, r_s, dr_h, dr_s, r_h_d, r_s_d)
                  CALL get_rho_atom(rho_atom=rho_atom, &
                                    rho_rad_h=r_h, rho_rad_s=r_s)
                  DO ir = 1, nr
                     CALL calc_rho_angular(atomic_grid, harmonics, nspins, .FALSE., &
                                           ir, r_h, r_s, rho_h, rho_s, &
                                           dr_h, dr_s, r_h_d, r_s_d, drho_h, drho_s)
                     CALL fill_rho_set(orb_rho_set_h, lsd, nspins, needs_orbs, rho_h, drho_h, tau, na, ir)
                     CALL fill_rho_set(orb_rho_set_s, lsd, nspins, needs_orbs, rho_s, drho_s, tau, na, ir)
                  END DO

                  IF (lsd) THEN
                     IF (ispin == 1) THEN
                        vxc_SAOP_h(:, :, 1) = vxc_SAOP_h(:, :, 1)+vxc_tmp_h(:, :, 1)*orb_rho_set_h%rhoa(:, :, 1)
                        vxc_SAOP_s(:, :, 1) = vxc_SAOP_s(:, :, 1)+vxc_tmp_s(:, :, 1)*orb_rho_set_s%rhoa(:, :, 1)
                     ELSE
                        vxc_SAOP_h(:, :, 2) = vxc_SAOP_h(:, :, 2)+vxc_tmp_h(:, :, 2)*orb_rho_set_h%rhob(:, :, 1)
                        vxc_SAOP_s(:, :, 2) = vxc_SAOP_s(:, :, 2)+vxc_tmp_s(:, :, 2)*orb_rho_set_s%rhob(:, :, 1)
                     END IF
                  ELSE
                     vxc_SAOP_h(:, :, 1) = vxc_SAOP_h(:, :, 1)+vxc_tmp_h(:, :, 1)*orb_rho_set_h%rho(:, :, 1)
                     vxc_SAOP_s(:, :, 1) = vxc_SAOP_s(:, :, 1)+vxc_tmp_s(:, :, 1)*orb_rho_set_s%rho(:, :, 1)
                  END IF

               END DO ! orb

            END DO ! ispin

            IF (lsd) THEN
               DO ir = 1, nr
                  DO ia = 1, na
                     IF (rho_set_h%rhoa(ia, ir, 1) > rho_set_h%rho_cutoff) THEN
                        vxc_SAOP_h(ia, ir, 1) = vxc_SAOP_h(ia, ir, 1)/rho_set_h%rhoa(ia, ir, 1)
                     ELSE
                        vxc_SAOP_h(ia, ir, 1) = 0.0_dp
                     END IF
                     IF (rho_set_h%rhob(ia, ir, 1) > rho_set_h%rho_cutoff) THEN
                        vxc_SAOP_h(ia, ir, 2) = vxc_SAOP_h(ia, ir, 2)/rho_set_h%rhob(ia, ir, 1)
                     ELSE
                        vxc_SAOP_h(ia, ir, 2) = 0.0_dp
                     END IF
                     IF (rho_set_s%rhoa(ia, ir, 1) > rho_set_s%rho_cutoff) THEN
                        vxc_SAOP_s(ia, ir, 1) = vxc_SAOP_s(ia, ir, 1)/rho_set_s%rhoa(ia, ir, 1)
                     ELSE
                        vxc_SAOP_s(ia, ir, 1) = 0.0_dp
                     END IF
                     IF (rho_set_s%rhob(ia, ir, 1) > rho_set_s%rho_cutoff) THEN
                        vxc_SAOP_s(ia, ir, 2) = vxc_SAOP_s(ia, ir, 2)/rho_set_s%rhob(ia, ir, 1)
                     ELSE
                        vxc_SAOP_s(ia, ir, 2) = 0.0_dp
                     END IF
                  END DO
               END DO
            ELSE
               DO ir = 1, nr
                  DO ia = 1, na
                     IF (rho_set_h%rho(ia, ir, 1) > rho_set_h%rho_cutoff) THEN
                        vxc_SAOP_h(ia, ir, 1) = vxc_SAOP_h(ia, ir, 1)/rho_set_h%rho(ia, ir, 1)
                     ELSE
                        vxc_SAOP_h(ia, ir, 1) = 0.0_dp
                     END IF
                     IF (rho_set_s%rho(ia, ir, 1) > rho_set_s%rho_cutoff) THEN
                        vxc_SAOP_s(ia, ir, 1) = vxc_SAOP_s(ia, ir, 1)/rho_set_s%rho(ia, ir, 1)
                     ELSE
                        vxc_SAOP_s(ia, ir, 1) = 0.0_dp
                     END IF
                  END DO
               END DO
            END IF

            rho_atom => rho_atom_set(iatom)
            SELECT CASE (oe_corr)
            CASE (oe_lb)
               CALL gaVxcgb_noGC(vxc_LB_h, vxc_LB_s, qs_kind_set(ikind), rho_atom, nspins)
            CASE (oe_gllb)
               CALL gaVxcgb_noGC(vxc_GLLB_h, vxc_GLLB_s, qs_kind_set(ikind), rho_atom, nspins)
            CASE (oe_saop)
               CALL gaVxcgb_noGC(vxc_SAOP_h, vxc_SAOP_s, qs_kind_set(ikind), rho_atom, nspins)
            CASE default
               CALL cp__b("xc_pot_saop.F",996,"")
            END SELECT

         END DO

         DEALLOCATE (rho_h, rho_s, weight)
         DEALLOCATE (vxc_LB_h, vxc_LB_s)
         DEALLOCATE (vxc_GLLB_h, vxc_GLLB_s)
         DEALLOCATE (vxc_tmp_h, vxc_tmp_s)
         DEALLOCATE (vxc_SAOP_h, vxc_SAOP_s)
         DEALLOCATE (drho_h, drho_s)

         CALL xc_dset_release(deriv_set)
         CALL xc_rho_set_release(rho_set_h)
         CALL xc_rho_set_release(rho_set_s)
         CALL xc_rho_set_release(orb_rho_set_h)
         CALL xc_rho_set_release(orb_rho_set_s)

      END DO

      ! remap pointer
      ns = SIZE(matrix_ks)
      ksmat(1:ns, 1:1) => matrix_ks(1:ns)
      ns = SIZE(rho_struct_ao)
      psmat(1:ns, 1:1) => rho_struct_ao(1:ns)

      CALL update_ks_atom(qs_env, ksmat, psmat, .FALSE.)

      !---------!
      ! Cleanup !
      !---------!
      CALL section_vals_release(xc_fun_section_tmp)
      CALL section_vals_release(xc_section_tmp)
      CALL section_vals_release(xc_section_orig)

      CALL local_rho_set_release(local_rho_set)
      DO ispin = 1, nspins
         CALL cp_fm_release(single_mo_coeff(ispin)%matrix)
      END DO
      DEALLOCATE (mo_coeff, single_mo_coeff, mo_eigenvalues)
      CALL cp_dbcsr_deallocate_matrix_set(orbital_density_matrix)

   END SUBROUTINE gapw_add_atomic_saop_pot

! **************************************************************************************************
!> \brief ...
!> \param pot ...
!> \param rho_set ...
!> \param lsd ...
!> \param spin ...
! **************************************************************************************************
   SUBROUTINE add_lb_pot(pot, rho_set, lsd, spin)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pot
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      LOGICAL, INTENT(IN)                                :: lsd
      INTEGER, INTENT(IN)                                :: spin

      REAL(KIND=dp), PARAMETER                           :: ob3 = 1.0_dp/3.0_dp

      INTEGER                                            :: i, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo
      REAL(KIND=dp)                                      :: n, n_13, x, x2

      bo = rho_set%local_bounds

      DO k = bo(1, 3), bo(2, 3)
         DO j = bo(1, 2), bo(2, 2)
            DO i = bo(1, 1), bo(2, 1)
               IF (.NOT. lsd) THEN
                  IF (rho_set%rho(i, j, k) > rho_set%rho_cutoff) THEN
                     n = rho_set%rho(i, j, k)/2.0_dp
                     n_13 = n**ob3
                     x = (rho_set%norm_drho(i, j, k)/2.0_dp)/(n*n_13)
                     x2 = x*x
                     pot(i, j, k) = beta*x2*n_13/(1.0_dp+3.0_dp*beta*x*LOG(x+SQRT(x2+1.0_dp)))
                  END IF
               ELSE
                  IF (spin == 1) THEN
                     IF (rho_set%rhoa(i, j, k) > rho_set%rho_cutoff) THEN
                        n_13 = rho_set%rhoa_1_3(i, j, k)
                        x = rho_set%norm_drhoa(i, j, k)/(rho_set%rhoa(i, j, k)*n_13)
                        x2 = x*x
                        pot(i, j, k) = beta*x2*n_13/(1.0_dp+3.0_dp*beta*x*LOG(SQRT(x2+1.0_dp)+x))
                     END IF
                  ELSE IF (spin == 2) THEN
                     IF (rho_set%rhob(i, j, k) > rho_set%rho_cutoff) THEN
                        n_13 = rho_set%rhob_1_3(i, j, k)
                        x = rho_set%norm_drhob(i, j, k)/(rho_set%rhob(i, j, k)*n_13)
                        x2 = x*x
                        pot(i, j, k) = beta*x2*n_13/(1.0_dp+3.0_dp*beta*x*LOG(SQRT(x2+1.0_dp)+x))
                     END IF
                  END IF
               END IF
            END DO
         END DO
      END DO

   END SUBROUTINE add_lb_pot

! **************************************************************************************************
!> \brief ...
!> \param pot ...
!> \param rho_set ...
!> \param e_uniform ...
!> \param lsd ...
! **************************************************************************************************
   SUBROUTINE calc_2excpbe(pot, rho_set, e_uniform, lsd)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: pot
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_uniform
      LOGICAL, INTENT(IN)                                :: lsd

      INTEGER                                            :: i, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo
      REAL(KIND=dp)                                      :: e_unif, rho

      bo = rho_set%local_bounds

      DO k = bo(1, 3), bo(2, 3)
         DO j = bo(1, 2), bo(2, 2)
            DO i = bo(1, 1), bo(2, 1)
               IF (.NOT. lsd) THEN
                  IF (rho_set%rho(i, j, k) > rho_set%rho_cutoff) THEN
                     e_unif = e_uniform(i, j, k)/rho_set%rho(i, j, k)
                  ELSE
                     e_unif = 0.0_dp
                  END IF
                  pot(i, j, k) = &
                     2.0_dp* &
                     calc_ecpbe_r(rho_set%rho(i, j, k), rho_set%norm_drho(i, j, k), &
                                  e_unif, rho_set%rho_cutoff, rho_set%drho_cutoff)+ &
                     2.0_dp* &
                     calc_expbe_r(rho_set%rho(i, j, k), rho_set%norm_drho(i, j, k), &
                                  rho_set%rho_cutoff, rho_set%drho_cutoff)
               ELSE
                  rho = rho_set%rhoa(i, j, k)+rho_set%rhob(i, j, k)
                  IF (rho > rho_set%rho_cutoff) THEN
                     e_unif = e_uniform(i, j, k)/rho
                  ELSE
                     e_unif = 0.0_dp
                  END IF
                  pot(i, j, k) = &
                     2.0_dp* &
                     calc_ecpbe_u(rho_set%rhoa(i, j, k), rho_set%rhob(i, j, k), rho_set%norm_drho(i, j, k), &
                                  e_unif, &
                                  rho_set%rho_cutoff, rho_set%drho_cutoff)+ &
                     2.0_dp* &
                     calc_expbe_u(rho_set%rhoa(i, j, k), rho_set%rhob(i, j, k), rho_set%norm_drho(i, j, k), &
                                  rho_set%rho_cutoff, rho_set%drho_cutoff)
               END IF
            END DO
         END DO
      END DO

   END SUBROUTINE calc_2excpbe

! **************************************************************************************************
!> \brief ...
!> \param ra ...
!> \param rb ...
!> \param ngr ...
!> \param ec_unif ...
!> \param rc ...
!> \param ngrc ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION calc_ecpbe_u(ra, rb, ngr, ec_unif, rc, ngrc) RESULT(res)

      REAL(kind=dp), INTENT(in)                          :: ra, rb, ngr, ec_unif, rc, ngrc
      REAL(kind=dp)                                      :: res

      REAL(kind=dp), PARAMETER                           :: ob3 = 1.0_dp/3.0_dp, tb3 = 2.0_dp/3.0_dp

      REAL(kind=dp)                                      :: A, At2, H, kf, kl, ks, phi, phi3, r, t2, &
                                                            zeta

      r = ra+rb
      H = 0.0_dp
      IF (r > rc .AND. ngr > ngrc) THEN
         zeta = (ra-rb)/r
         IF (zeta > 1.0_dp) zeta = 1.0_dp ! machine precision problem
         IF (zeta < -1.0_dp) zeta = -1.0_dp ! machine precision problem
         phi = ((1.0_dp+zeta)**tb3+(1.0_dp-zeta)**tb3)/2.0_dp
         phi3 = phi*phi*phi
         kf = (3.0_dp*r*pi*pi)**ob3
         ks = SQRT(4.0_dp*kf/pi)
         t2 = (ngr/(2.0_dp*phi*ks*r))**2
         A = beta_ec/gamma_saop/(EXP(-ec_unif/(gamma_saop*phi3))-1.0_dp)
         At2 = A*t2
         kl = (1.0_dp+At2)/(1.0_dp+At2+At2*At2)
         H = gamma_saop*LOG(1.0_dp+beta_ec/gamma_saop*t2*kl)
      END IF
      res = ec_unif+H

   END FUNCTION calc_ecpbe_u

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param ngr ...
!> \param ec_unif ...
!> \param rc ...
!> \param ngrc ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION calc_ecpbe_r(r, ngr, ec_unif, rc, ngrc) RESULT(res)

      REAL(kind=dp), INTENT(in)                          :: r, ngr, ec_unif, rc, ngrc
      REAL(kind=dp)                                      :: res

      REAL(kind=dp)                                      :: A, At2, H, kf, kl, ks, t2

      H = 0.0_dp
      IF (r > rc .AND. ngr > ngrc) THEN
         kf = (3.0_dp*r*pi*pi)**(1.0_dp/3.0_dp)
         ks = SQRT(4.0_dp*kf/pi)
         t2 = (ngr/(2.0_dp*ks*r))**2
         A = beta_ec/gamma_saop/(EXP(-ec_unif/gamma_saop)-1.0_dp)
         At2 = A*t2
         kl = (1.0_dp+At2)/(1.0_dp+At2+At2*At2)
         H = gamma_saop*LOG(1.0_dp+beta_ec/gamma_saop*t2*kl)
      END IF
      res = ec_unif+H

   END FUNCTION calc_ecpbe_r

! **************************************************************************************************
!> \brief ...
!> \param ra ...
!> \param rb ...
!> \param ngr ...
!> \param rc ...
!> \param ngrc ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION calc_expbe_u(ra, rb, ngr, rc, ngrc) RESULT(res)

      REAL(kind=dp), INTENT(in)                          :: ra, rb, ngr, rc, ngrc
      REAL(kind=dp)                                      :: res

      REAL(kind=dp)                                      :: r

      r = ra+rb
      res = calc_expbe_r(r, ngr, rc, ngrc)

   END FUNCTION calc_expbe_u

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param ngr ...
!> \param rc ...
!> \param ngrc ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION calc_expbe_r(r, ngr, rc, ngrc) RESULT(res)

      REAL(kind=dp), INTENT(in)                          :: r, ngr, rc, ngrc
      REAL(kind=dp)                                      :: res

      REAL(kind=dp)                                      :: ex_unif, fx, kf, s

      IF (r > rc) THEN
         kf = (3.0_dp*r*pi*pi)**(1.0_dp/3.0_dp)
         ex_unif = -3.0_dp*kf/(4.0_dp*pi)
         fx = 1.0_dp
         IF (ngr > ngrc) THEN
            s = ngr/(2.0_dp*kf*r)
            fx = fx+kappa-kappa/(1.0_dp+mu*s*s/kappa)
         END IF
         res = ex_unif*fx
      ELSE
         res = 0.0_dp
      END IF

   END FUNCTION calc_expbe_r

END MODULE xc_pot_saop
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Performs density functional perturbation theory (tddfpt) calculations.
!>      Uses the self consistent approach. The tddfpt calculation uses the ground
!>      state of the unperturbed system as the initial state.
! **************************************************************************************************
MODULE qs_tddfpt_module
   USE bibliography,                    ONLY: Iannuzzi2005,&
                                              cite_reference
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE header,                          ONLY: tddfpt_header
   USE input_constants,                 ONLY: oe_gllb,&
                                              oe_lb,&
                                              oe_none,&
                                              oe_saop,&
                                              oe_sic,&
                                              tddfpt_excitations
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_vals_create,&
                                              section_vals_get_subs_vals,&
                                              section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_set_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_ks_methods,                   ONLY: qs_ks_build_kohn_sham_matrix
   USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                              set_ks_env
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_rho_types,                    ONLY: qs_rho_type
   USE qs_scf_types,                    ONLY: qs_scf_env_type
   USE qs_tddfpt_eigensolver,           ONLY: eigensolver
   USE qs_tddfpt_types,                 ONLY: tddfpt_env_type
   USE qs_tddfpt_utils,                 ONLY: find_contributions,&
                                              tddfpt_cleanup,&
                                              tddfpt_init
   USE xc_pot_saop,                     ONLY: add_saop_pot
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: tddfpt_calculation

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_tddfpt_module'

CONTAINS

! **************************************************************************************************
!> \brief Performs the perturbation calculation
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE tddfpt_calculation(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'tddfpt_calculation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(section_vals_type), POINTER                   :: dft_section, input
      TYPE(tddfpt_env_type)                              :: t_env

      NULLIFY (logger)
      logger => cp_get_default_logger()
      NULLIFY (input, ks_env)
      CALL get_qs_env(qs_env, ks_env=ks_env, input=input)
      dft_section => section_vals_get_subs_vals(input, "DFT")

      IF (section_get_ival(dft_section, "EXCITATIONS") /= tddfpt_excitations) RETURN
      CALL cite_reference(Iannuzzi2005)

      CALL timeset(routineN, handle)

      IF (section_get_ival(dft_section, "TDDFPT%OE_CORR") /= oe_none) THEN
         CALL orbital_eigenvalue_correction(qs_env)
      END IF

      NULLIFY (p_env)

      iw = cp_print_key_unit_nr(logger, dft_section, "PRINT%PROGRAM_BANNER", &
                                extension=".Log")
      CALL tddfpt_header(iw)
      CALL cp_print_key_finished_output(iw, logger, dft_section, &
                                        "PRINT%PROGRAM_BANNER")

      !---------------------------------------!
      ! we don't want to update the KS matrix !
      !---------------------------------------!
      CALL set_ks_env(ks_env, rho_changed=.FALSE.)

      CALL tddfpt_init(p_env, t_env, qs_env)

      CALL eigensolver(p_env, qs_env, t_env)

      CALL find_contributions(qs_env, t_env)

      CALL tddfpt_cleanup(t_env, p_env)

      CALL timestop(handle)

   END SUBROUTINE tddfpt_calculation

! **************************************************************************************************
!> \brief Apply a special potential to obtain better
!>       orbital eigenvalues.
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE orbital_eigenvalue_correction(qs_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'orbital_eigenvalue_correction', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: oe_corr, output_unit
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(qs_scf_env_type), POINTER                     :: scf_env
      TYPE(section_vals_type), POINTER                   :: input, xc_fun_orig, xc_fun_tmp

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_tddfpt_module.F",139)

      NULLIFY (logger, scf_env, input, energy, matrix_ks, rho)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)

      IF (output_unit > 0) THEN
         WRITE (output_unit, "(2X,A)") "", &
            "-----------------------------------------------------------------------------", &
            "-                    orbital eigenvalue correction started                  -", &
            "-----------------------------------------------------------------------------", &
            ""
      END IF

      CALL get_qs_env(qs_env, &
                      scf_env=scf_env, &
                      input=input, &
                      matrix_ks=matrix_ks, &
                      rho=rho)

      !----------------------!
      ! KS matrix without XC !
      !----------------------!
      xc_fun_orig => section_vals_get_subs_vals(input, "DFT%XC%XC_FUNCTIONAL")
      CALL section_vals_retain(xc_fun_orig)
      NULLIFY (xc_fun_tmp)
      CALL section_vals_create(xc_fun_tmp, xc_fun_orig%section)
      CALL section_vals_set_subs_vals(input, "DFT%XC%XC_FUNCTIONAL", xc_fun_tmp)
      CALL section_vals_release(xc_fun_tmp)

      CALL get_qs_env(qs_env, energy=energy)
      CALL qs_ks_build_kohn_sham_matrix(qs_env, calculate_forces=.FALSE., &
                                        just_energy=.FALSE.)

      CALL section_vals_set_subs_vals(input, "DFT%XC%XC_FUNCTIONAL", xc_fun_orig)
      CALL section_vals_release(xc_fun_orig)

      CALL section_vals_val_get(input, "DFT%TDDFPT%OE_CORR", i_val=oe_corr)
      IF (oe_corr == oe_saop .OR. &
          oe_corr == oe_lb .OR. &
          oe_corr == oe_gllb) THEN
         CALL add_saop_pot(matrix_ks, qs_env, oe_corr)
      ELSE IF (oe_corr == oe_sic) THEN
      END IF

   END SUBROUTINE orbital_eigenvalue_correction

END MODULE qs_tddfpt_module
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE cp2k_runs
   USE acc_device,                      ONLY: acc_get_ndevices,&
                                              acc_set_active_device
   USE atom,                            ONLY: atom_code
   USE bibliography,                    ONLY: Hutter2014,&
                                              cite_reference
   USE bsse,                            ONLY: do_bsse_calculation
   USE cell_opt,                        ONLY: cp_cell_opt
   USE cp2k_debug,                      ONLY: cp2k_debug_energy_and_forces
   USE cp2k_info,                       ONLY: compile_date,&
                                              compile_revision,&
                                              cp2k_version,&
                                              cp2k_year
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_config,&
                                              cp_dbcsr_print_config
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_finalize_lib,&
                                              cp_dbcsr_init_lib
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_failure_level,&
                                              cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_logger_would_log,&
                                              cp_note_level
   USE cp_output_handling,              ONLY: cp_add_iter_level,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_unit_nr,&
                                              cp_rm_iter_level
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_search_string
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_set_create,&
                                              cp_unit_set_release,&
                                              cp_unit_set_type,&
                                              print_all_units
   USE cuda_profiling,                  ONLY: cuda_nvtx_init
   USE environment,                     ONLY: cp2k_finalize,&
                                              cp2k_init,&
                                              cp2k_read,&
                                              cp2k_setup
   USE f77_interface,                   ONLY: create_force_env,&
                                              destroy_force_env,&
                                              f77_default_para_env => default_para_env,&
                                              f_env_add_defaults,&
                                              f_env_rm_defaults,&
                                              f_env_type
   USE farming_methods,                 ONLY: do_deadlock,&
                                              do_nothing,&
                                              do_wait,&
                                              farming_parse_input,&
                                              get_next_job
   USE farming_types,                   ONLY: deallocate_farming_env,&
                                              farming_env_type,&
                                              init_farming_env,&
                                              job_finished,&
                                              job_running
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE geo_opt,                         ONLY: cp_geo_opt
   USE global_types,                    ONLY: global_environment_type,&
                                              globenv_create,&
                                              globenv_release,&
                                              globenv_retain
   USE input_constants,                 ONLY: &
        bsse_run, cell_opt_run, debug_run, do_atom, do_band, do_cp2k, do_farming, do_fist, &
        do_mixed, do_opt_basis, do_optimize_input, do_qmmm, do_qs, do_swarm, do_tamc, do_test, &
        do_tree_mc, do_tree_mc_ana, driver_run, ehrenfest, electronic_spectra_run, &
        energy_force_run, energy_run, geo_opt_run, linear_response_run, mol_dyn_run, mon_car_run, &
        none_run, pint_run, real_time_propagation, tree_mc_run, vib_anal
   USE input_cp2k,                      ONLY: create_cp2k_root_section
   USE input_cp2k_check,                ONLY: check_cp2k_input
   USE input_cp2k_global,               ONLY: create_global_section
   USE input_cp2k_read,                 ONLY: empty_initial_variables,&
                                              read_input
   USE input_keyword_types,             ONLY: keyword_release
   USE input_parsing,                   ONLY: section_vals_parse
   USE input_section_types,             ONLY: &
        section_describe_html, section_describe_index_html, section_release, section_type, &
        section_vals_create, section_vals_get_subs_vals, section_vals_release, &
        section_vals_retain, section_vals_type, section_vals_val_get, section_vals_write, &
        write_section_xml
   USE ipi_driver,                      ONLY: run_driver
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp,&
                                              int_8
   USE library_tests,                   ONLY: lib_test
   USE machine,                         ONLY: default_output_unit,&
                                              m_chdir,&
                                              m_flush,&
                                              m_getcwd,&
                                              m_memory,&
                                              m_memory_max,&
                                              m_walltime
   USE mc_run,                          ONLY: do_mon_car
   USE md_run,                          ONLY: qs_mol_dyn
   USE message_passing,                 ONLY: &
        mp_any_source, mp_bcast, mp_comm_dup, mp_comm_free, mp_comm_split, mp_environ, mp_max, &
        mp_recv, mp_send, mp_sum, mp_sync
   USE mscfg_methods,                   ONLY: do_mol_loop,&
                                              loop_over_molecules
   USE neb_methods,                     ONLY: neb
   USE optimize_basis,                  ONLY: run_optimize_basis
   USE optimize_input,                  ONLY: run_optimize_input
   USE pint_methods,                    ONLY: do_pint_run
   USE pw_cuda,                         ONLY: pw_cuda_finalize,&
                                              pw_cuda_init
   USE qs_environment_types,            ONLY: get_qs_env
   USE qs_linres_module,                ONLY: linres_calculation
   USE qs_tddfpt_module,                ONLY: tddfpt_calculation
   USE reference_manager,               ONLY: print_all_references,&
                                              print_format_html
   USE rt_propagation,                  ONLY: rt_prop_setup
   USE string_utilities,                ONLY: html_entity_table,&
                                              write_html_tables
   USE swarm,                           ONLY: run_swarm
   USE tamc_run,                        ONLY: qs_tamc
   USE tmc_setup,                       ONLY: do_analyze_files,&
                                              do_tmc
   USE vibrational_analysis,            ONLY: vb_anal
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: write_cp2k_html_manual, write_xml_file, run_input
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp2k_runs'
CONTAINS

! **************************************************************************************************
!> \brief performs an instance of a cp2k run
!>      cp2k_run_OK == .TRUE. if everything went fine
!> \param input_declaration ...
!> \param input_file_name name of the file to be opened for input
!> \param output_unit unit to which output should be written
!> \param mpi_comm ...
!> \retval cp2k_run_OK ...
!> \author Joost VandeVondele
!> \note
!>      para_env should be a valid communicator
!>      output_unit should be writeable by at least the lowest rank of the mpi group
!>
!>      recursive because a given run_type might need to be able to perform
!>      another cp2k_run as part of its job (e.g. farming, classical equilibration, ...)
!>
!>      the idea is that a cp2k instance should be able to run with just three
!>      arguments, i.e. a given input file, output unit, mpi communicator.
!>      giving these three to cp2k_run should produce a valid run.
!>      the only task of the PROGRAM cp2k is to create valid instances of the
!>      above arguments. Ideally, anything that is called afterwards should be
!>      able to run simultaneously / multithreaded / sequential / parallel / ...
!>      and able to fail safe
! **************************************************************************************************
   RECURSIVE FUNCTION cp2k_run(input_declaration, input_file_name, output_unit, mpi_comm) RESULT(cp2k_run_OK)
      TYPE(section_type), POINTER                        :: input_declaration
      CHARACTER(LEN=*), INTENT(IN)                       :: input_file_name
      INTEGER, INTENT(IN)                                :: output_unit, mpi_comm
      LOGICAL                                            :: cp2k_run_OK

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp2k_run', routineP = moduleN//':'//routineN

      INTEGER                                            :: f_env_handle, ierr, iter_level, &
                                                            method_name_id, new_env_id, &
                                                            prog_name_id, run_type_id
      INTEGER(KIND=int_8)                                :: m_memory_max_mpi
      LOGICAL                                            :: echo_input, I_was_ionode, was_present
      TYPE(cp_logger_type), POINTER                      :: logger, sublogger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(f_env_type), POINTER                          :: f_env
      TYPE(force_env_type), POINTER                      :: force_env
      TYPE(global_environment_type), POINTER             :: globenv
      TYPE(section_vals_type), POINTER                   :: glob_section, input_file, root_section

      cp2k_run_OK = .FALSE.
      was_present = .FALSE.
      NULLIFY (para_env, f_env, dft_control)
      CALL cp_para_env_create(para_env, group=mpi_comm, &
                              owns_group=.FALSE.)

      IF (acc_get_ndevices() > 0) &
         CALL acc_set_active_device(MOD(para_env%mepos, acc_get_ndevices()))

      CALL cuda_nvtx_init()

      CALL pw_cuda_init()

      CALL cp_dbcsr_init_lib()

      NULLIFY (globenv, force_env)

      CALL cite_reference(Hutter2014)

      ! parse the input
      input_file => read_input(input_declaration, input_file_name, initial_variables=empty_initial_variables, &
                               para_env=para_env)

      CALL mp_sync(para_env%group)

      glob_section => section_vals_get_subs_vals(input_file, "GLOBAL")
      CALL section_vals_val_get(glob_section, "ECHO_INPUT", l_val=echo_input)
      logger => cp_get_default_logger()
      IF (echo_input) THEN
         CALL section_vals_write(input_file, &
                                 unit_nr=cp_logger_get_default_io_unit(logger), &
                                 hide_root=.TRUE., hide_defaults=.FALSE.)
      END IF

      CALL check_cp2k_input(input_declaration, input_file, para_env=para_env, output_unit=output_unit)
      root_section => input_file
      CALL section_vals_val_get(input_file, "GLOBAL%PROGRAM_NAME", &
                                i_val=prog_name_id)
      CALL section_vals_val_get(input_file, "GLOBAL%RUN_TYPE", &
                                i_val=run_type_id)

      IF (prog_name_id /= do_cp2k) THEN
         ! initial setup (cp2k does in in the creation of the force_env)
         CALL globenv_create(globenv)
         ! XXXXXXXXX
         ! root_section => input_file
         ! XXXXXXXXX
         CALL section_vals_retain(input_file)
         CALL cp2k_init(para_env, output_unit, globenv, input_file_name=input_file_name)
         CALL cp2k_read(root_section, para_env, globenv)
         CALL cp2k_setup(root_section, para_env, globenv)
      END IF

      CALL cp_dbcsr_config(root_section)
      IF (output_unit > 0 .AND. &
          cp_logger_would_log(logger, cp_note_level)) THEN
         CALL cp_dbcsr_print_config(unit_nr=output_unit)
         WRITE (UNIT=output_unit, FMT='()')
      ENDIF

      SELECT CASE (prog_name_id)
      CASE (do_atom)
         globenv%run_type_id = none_run
         CALL atom_code(root_section)
      CASE (do_optimize_input)
         CALL run_optimize_input(input_declaration, root_section, para_env)
      CASE (do_swarm)
         CALL run_swarm(input_declaration, root_section, para_env, globenv, input_file_name)
      CASE (do_farming)
         ! Hack: DBCSR should be uninitialized when entering farming.
         ! But, we don't want to change the public f77_interface.
         ! TODO: refactor cp2k's startup code
         CALL cp_dbcsr_finalize_lib(group=mpi_comm, output_unit=0)
         CALL pw_cuda_finalize()
         CALL farming_run(input_declaration, root_section, para_env)
         CALL cp_dbcsr_init_lib()
         CALL pw_cuda_init()
      CASE (do_opt_basis)
         CALL run_optimize_basis(input_declaration, root_section, para_env)
         globenv%run_type_id = none_run
      CASE (do_cp2k)
         CALL create_force_env(new_env_id, &
                               input_declaration=input_declaration, &
                               input_path=input_file_name, &
                               output_path="__STD_OUT__", mpi_comm=para_env%group, &
                               output_unit=output_unit, &
                               owns_out_unit=.FALSE., &
                               input=input_file, ierr=ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("start/cp2k_runs.F",277)
         CALL f_env_add_defaults(new_env_id, f_env, handle=f_env_handle)
         force_env => f_env%force_env
         CALL force_env_get(force_env, globenv=globenv)
         CALL globenv_retain(globenv)
         CALL section_vals_val_get(force_env%force_env_section, "METHOD", i_val=method_name_id)

      CASE (do_test)
         CALL lib_test(root_section, para_env, globenv)
      CASE (do_tree_mc) ! TMC entry point
         CALL do_tmc(input_declaration, root_section, para_env, globenv)
      CASE (do_tree_mc_ana)
         CALL do_analyze_files(input_declaration, root_section, para_env)
      CASE default
         CALL cp__b("start/cp2k_runs.F",291,"")
      END SELECT
      CALL section_vals_release(input_file)

      SELECT CASE (globenv%run_type_id)
      CASE (pint_run)
         CALL do_pint_run(para_env, root_section, input_declaration, globenv)
      CASE (none_run, tree_mc_run)
         ! do nothing
      CASE (driver_run)
         CALL run_driver(force_env, globenv)
      CASE (energy_run, energy_force_run)
         IF (method_name_id /= do_qs .AND. &
             method_name_id /= do_qmmm .AND. &
             method_name_id /= do_mixed .AND. &
             method_name_id /= do_fist) &
            CALL cp__b("start/cp2k_runs.F",307,"Energy/Force run not available for all methods ")

         sublogger => cp_get_default_logger()
         CALL cp_add_iter_level(sublogger%iter_info, "JUST_ENERGY", &
                                n_rlevel_new=iter_level)

         ! loop over molecules to generate a molecular guess
         ! this procedure is initiated here to avoid passing globenv deep down
         ! the subroutine stack
         IF (do_mol_loop(force_env=force_env)) &
            CALL loop_over_molecules(globenv, force_env)

         SELECT CASE (globenv%run_type_id)
         CASE (energy_run)
            CALL force_env_calc_energy_force(force_env, calc_force=.FALSE.)
         CASE (energy_force_run)
            CALL force_env_calc_energy_force(force_env, calc_force=.TRUE.)
         CASE default
            CALL cp__b("start/cp2k_runs.F",325,"")
         END SELECT
         CALL cp_rm_iter_level(sublogger%iter_info, level_name="JUST_ENERGY", n_rlevel_att=iter_level)
      CASE (mol_dyn_run)
         CALL qs_mol_dyn(force_env, globenv)
      CASE (geo_opt_run)
         CALL cp_geo_opt(force_env, globenv)
      CASE (cell_opt_run)
         CALL cp_cell_opt(force_env, globenv)
      CASE (mon_car_run)
         CALL do_mon_car(force_env, globenv, input_declaration, input_file_name)
      CASE (do_tamc)
         CALL qs_tamc(force_env, globenv)
      CASE (electronic_spectra_run)
         IF (method_name_id /= do_qs) &
            CALL cp__b("start/cp2k_runs.F",340,"Electron spectra available only with Quickstep. ")
         CALL force_env_calc_energy_force(force_env, calc_force=.FALSE.)
         CALL tddfpt_calculation(force_env%qs_env)
      CASE (real_time_propagation)
         IF (method_name_id /= do_qs) &
            CALL cp__b("start/cp2k_runs.F",345,"Real time propagation needs METHOD QS. ")
         CALL get_qs_env(force_env%qs_env, dft_control=dft_control)
         dft_control%rtp_control%fixed_ions = .TRUE.
         CALL rt_prop_setup(force_env)
      CASE (ehrenfest)
         IF (method_name_id /= do_qs) &
            CALL cp__b("start/cp2k_runs.F",351,"Ehrenfest dynamics needs METHOD QS ")
         CALL get_qs_env(force_env%qs_env, dft_control=dft_control)
         dft_control%rtp_control%fixed_ions = .FALSE.
         CALL qs_mol_dyn(force_env, globenv)
      CASE (bsse_run)
         CALL do_bsse_calculation(force_env, globenv)
      CASE (linear_response_run)
         IF (method_name_id /= do_qs .AND. &
             method_name_id /= do_qmmm) &
            CALL cp__b("start/cp2k_runs.F",360,"Property calculations by Linear Response only within the QS or QMMM program ")
         ! The Ground State is needed, it can be read from Restart
         CALL force_env_calc_energy_force(force_env, calc_force=.FALSE., linres=.TRUE.)
         CALL linres_calculation(force_env)
      CASE (debug_run)
         SELECT CASE (method_name_id)
         CASE (do_qs, do_qmmm, do_fist)
            CALL cp2k_debug_energy_and_forces(force_env)
         CASE DEFAULT
            CALL cp__b("start/cp2k_runs.F",369,"Debug run available only with QS, FIST, and QMMM program ")
         END SELECT
      CASE (vib_anal)
         CALL vb_anal(root_section, input_declaration, para_env, globenv)
      CASE (do_band)
         CALL neb(root_section, input_declaration, para_env, globenv)
      CASE default
         CALL cp__b("start/cp2k_runs.F",376,"")
      END SELECT

      !sample peak memory
      CALL m_memory()

      CALL pw_cuda_finalize()
      CALL cp_dbcsr_finalize_lib(group=mpi_comm, output_unit=output_unit)

      m_memory_max_mpi = m_memory_max
      CALL mp_max(m_memory_max_mpi, mpi_comm)
      IF (output_unit > 0) THEN
         WRITE (output_unit, *)
         WRITE (output_unit, '(T2,"MEMORY| Estimated peak process memory [MiB]",T73,I8)') &
            (m_memory_max_mpi+(1024*1024)-1)/(1024*1024)
      ENDIF

      IF (prog_name_id == do_cp2k) THEN
         f_env%force_env => force_env ! for mc
         CALL globenv_retain(globenv) !mc
         CALL globenv_release(force_env%globenv) !mc
         force_env%globenv => globenv !mc
         CALL f_env_rm_defaults(f_env, ierr=ierr, &
                                handle=f_env_handle)
         IF(.NOT.(ierr == 0))CALL cp__a("start/cp2k_runs.F",400)
         CALL destroy_force_env(new_env_id, ierr=ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("start/cp2k_runs.F",402)
         CALL globenv_release(globenv)
      ELSE
         I_was_ionode = para_env%ionode
         CALL cp2k_finalize(root_section, para_env, globenv)
         IF(.NOT.(globenv%ref_count == 1))CALL cp__a("start/cp2k_runs.F",407)
         CALL section_vals_release(root_section)
         CALL globenv_release(globenv)
      END IF

      CALL cp_para_env_release(para_env)
      cp2k_run_OK = .TRUE.

   END FUNCTION cp2k_run

! **************************************************************************************************
!> \brief performs a farming run that performs several independent cp2k_runs
!> \param input_declaration ...
!> \param root_section ...
!> \param para_env ...
!> \author Joost VandeVondele
!> \note
!>      needs to be part of this module as the cp2k_run -> farming_run -> cp2k_run
!>      calling style creates a hard circular dependency
! **************************************************************************************************
   RECURSIVE SUBROUTINE farming_run(input_declaration, root_section, para_env)
      TYPE(section_type), POINTER              :: input_declaration
      TYPE(section_vals_type), POINTER         :: root_section
      TYPE(cp_para_env_type), POINTER          :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'farming_run', &
                                     routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                       :: slave_status_done = -3, &
                                                  slave_status_wait = -4

      CHARACTER(len=7)                         :: label
      CHARACTER(LEN=default_path_length)       :: output_file
      CHARACTER(LEN=default_string_length)     :: str
      INTEGER :: dest, handle, i, i_job_to_restart, ierr, ijob, ijob_current, &
                 ijob_end, ijob_start, iunit, n_jobs_to_run, new_group, new_output_unit, &
                 new_rank, new_size, ngroups, num_slaves, output_unit, primus_slave, &
                 slave_group, slave_rank, source, tag, todo
      INTEGER, DIMENSION(:), POINTER           :: group_distribution, &
                                                  master_slave_partition, &
                                                  slave_distribution, &
                                                  slave_status
      LOGICAL                                  :: found, master, &
                                                  run_OK, slave
      REAL(KIND=dp)                            :: t1, t2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:) :: waittime
      TYPE(cp_logger_type), POINTER            :: logger
      TYPE(cp_parser_type), POINTER            :: my_parser
      TYPE(cp_unit_set_type), POINTER          :: default_units
      TYPE(farming_env_type), POINTER          :: farming_env
      TYPE(section_type), POINTER              :: g_section
      TYPE(section_vals_type), POINTER         :: g_data

      ! the primus of all slaves, talks to the master on topics concerning all slaves
      CALL timeset(routineN, handle)
      NULLIFY (my_parser, g_section, g_data, default_units)

      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, root_section, "FARMING%PROGRAM_RUN_INFO", &
                                         extension=".log")

      IF (output_unit > 0) WRITE (output_unit, FMT="(T2,A)") "FARMING| Hi, welcome on this farm!"

      ALLOCATE (farming_env)
      CALL init_farming_env(farming_env)
      ! remember where we started
      CALL m_getcwd(farming_env%cwd)
      CALL farming_parse_input(farming_env, root_section, para_env)

      ! the full mpi group is first split in a slave group and a master group, the latter being at most 1 process
      slave = .TRUE.
      master = .FALSE.
      IF (farming_env%master_slave) THEN
         IF (output_unit > 0) WRITE (output_unit, FMT="(T2,A)") "FARMING| using a master-slave setup"

         ALLOCATE (master_slave_partition(0:1))
         master_slave_partition = (/1, para_env%num_pe-1/)
         ALLOCATE (group_distribution(0:para_env%num_pe-1))

         CALL mp_comm_split(para_env%group, slave_group, ngroups, group_distribution, &
                            n_subgroups=2, group_partition=master_slave_partition)
         DEALLOCATE (master_slave_partition)
         DEALLOCATE (group_distribution)
         CALL mp_environ(num_slaves, slave_rank, slave_group)

         IF (para_env%mepos == 0) THEN
            slave = .FALSE.
            master = .TRUE.
            ! on the master node, num_slaves corresponds to the size of the master group
            ! due to the mp_environ call.
            IF(.NOT.(num_slaves == 1))CALL cp__a("start/cp2k_runs.F",496)
            num_slaves = para_env%num_pe-1
            slave_rank = -1
         ENDIF
         IF(.NOT.(num_slaves == para_env%num_pe-1))CALL cp__a("start/cp2k_runs.F",500)
      ELSE
         ! all processes are slaves
         IF (output_unit > 0) WRITE (output_unit, FMT="(T2,A)") "FARMING| using a slave-only setup"
         CALL mp_comm_dup(para_env%group, slave_group)
         CALL mp_environ(num_slaves, slave_rank, slave_group)
      ENDIF
      IF (output_unit > 0) WRITE (output_unit, FMT="(T2,A,I0)") "FARMING| number of slaves ", num_slaves

      ! keep track of which para_env rank is which slave/master
      ALLOCATE (slave_distribution(0:para_env%num_pe-1))
      slave_distribution = 0
      slave_distribution(para_env%mepos) = slave_rank
      CALL mp_sum(slave_distribution, para_env%group)
      ! we do have a primus inter pares
      primus_slave = 0
      DO i = 1, para_env%num_pe-1
         IF (slave_distribution(i) == 0) primus_slave = i
      ENDDO

      ! split the current communicator for the slaves
      ! in a new_group, new_size and new_rank according to the number of groups required according to the input
      ALLOCATE (group_distribution(0:num_slaves-1))
      group_distribution = -1
      IF (slave) THEN
         IF (farming_env%group_size_wish_set) THEN
            farming_env%group_size_wish = MIN(farming_env%group_size_wish, para_env%num_pe)
            CALL mp_comm_split(slave_group, new_group, ngroups, group_distribution, &
                               subgroup_min_size=farming_env%group_size_wish, stride=farming_env%stride)
         ELSE IF (farming_env%ngroup_wish_set) THEN
            CALL mp_comm_split(slave_group, new_group, ngroups, group_distribution, &
                               n_subgroups=farming_env%ngroup_wish, &
                               group_partition=farming_env%group_partition, stride=farming_env%stride)
         ELSE
            CALL cp__b("start/cp2k_runs.F",534,"must set either group_size_wish or ngroup_wish")
         ENDIF
         CALL mp_environ(new_size, new_rank, new_group)
      ENDIF

      ! transfer the info about the slave group distribution to the master
      IF (farming_env%master_slave) THEN
         IF (para_env%mepos == primus_slave) THEN
            tag = 1
            CALL mp_send(group_distribution, 0, tag, para_env%group)
            tag = 2
            CALL mp_send(ngroups, 0, tag, para_env%group)
         ENDIF
         IF (para_env%mepos == 0) THEN
            tag = 1
            CALL mp_recv(group_distribution, primus_slave, tag, para_env%group)
            tag = 2
            CALL mp_recv(ngroups, primus_slave, tag, para_env%group)
         ENDIF
      ENDIF

      ! write info on group distribution
      IF (output_unit > 0) THEN
         WRITE (output_unit, FMT="(T2,A,T71,I10)") "FARMING| Number of created MPI (slave) groups:", ngroups
         WRITE (output_unit, FMT="(T2,A)", ADVANCE="NO") "FARMING| MPI (slave) process to group correspondence:"
         DO i = 0, num_slaves-1
            IF (MODULO(i, 4) == 0) WRITE (output_unit, *)
            WRITE (output_unit, FMT='(A3,I6,A3,I6,A1)', ADVANCE="NO") &
               "  (", i, " : ", group_distribution(i), ")"
         END DO
         WRITE (output_unit, *)
         CALL m_flush(output_unit)
      ENDIF

      ! protect about too many jobs being run in single go. Not more jobs are allowed than the number in the input file
      ! and determine the future restart point
      IF (farming_env%cycle) THEN
         n_jobs_to_run = farming_env%max_steps*ngroups
         i_job_to_restart = MODULO(farming_env%restart_n+n_jobs_to_run-1, farming_env%njobs)+1
      ELSE
         n_jobs_to_run = MIN(farming_env%njobs, farming_env%max_steps*ngroups)
         n_jobs_to_run = MIN(n_jobs_to_run, farming_env%njobs-farming_env%restart_n+1)
         i_job_to_restart = n_jobs_to_run+farming_env%restart_n
      ENDIF

      ! and write the restart now, that's the point where the next job starts, even if this one is running
      iunit = cp_print_key_unit_nr(logger, root_section, "FARMING%RESTART", &
                                   extension=".restart")
      IF (iunit > 0) THEN
         WRITE (iunit, *) i_job_to_restart
      ENDIF
      CALL cp_print_key_finished_output(iunit, logger, root_section, "FARMING%RESTART")

      ! this is the job range to be executed.
      ijob_start = farming_env%restart_n
      ijob_end = ijob_start+n_jobs_to_run-1
      IF (output_unit > 0 .AND. ijob_end-ijob_start < 0) THEN
         WRITE (output_unit, FMT="(T2,A)") "FARMING| --- WARNING --- NO JOBS NEED EXECUTION ? "
         WRITE (output_unit, FMT="(T2,A)") "FARMING| is the cycle keyword required ?"
         WRITE (output_unit, FMT="(T2,A)") "FARMING| or is a stray RESTART file present ?"
         WRITE (output_unit, FMT="(T2,A)") "FARMING| or is the group_size requested smaller than the number of CPUs?"
      ENDIF

      ! actual executions of the jobs in two different modes
      IF (farming_env%master_slave) THEN
         IF (slave) THEN
            ! keep on doing work until master has decided otherwise
            todo = do_wait
            DO
               IF (new_rank == 0) THEN
                  ! the head slave tells the master he's done or ready to start
                  ! the message tells what has been done lately
                  tag = 1
                  dest = 0
                  CALL mp_send(todo, dest, tag, para_env%group)

                  ! gets the new todo item
                  tag = 2
                  source = 0
                  CALL mp_recv(todo, source, tag, para_env%group)

                  ! and informs his peer slaves
                  CALL mp_bcast(todo, 0, new_group)
               ELSE
                  CALL mp_bcast(todo, 0, new_group)
               ENDIF

               ! if the todo is do_nothing we are flagged to quit. Otherwise it is the job number
               SELECT CASE (todo)
               CASE (do_wait, do_deadlock)
                  ! go for a next round, but we first wait a bit
                  t1 = m_walltime()
                  DO
                     t2 = m_walltime()
                     IF (t2-t1 > farming_env%wait_time) EXIT
                  ENDDO
               CASE (do_nothing)
                  EXIT
               CASE (1:)
                  CALL execute_job(todo)
               END SELECT
            ENDDO
         ELSE ! master
            ALLOCATE (slave_status(0:ngroups-1))
            slave_status = slave_status_wait
            ijob_current = ijob_start-1

            DO
               IF (ALL(slave_status == slave_status_done)) EXIT

               ! who's the next slave waiting for work
               tag = 1
               source = mp_any_source
               CALL mp_recv(todo, source, tag, para_env%group) ! updates source
               IF (todo > 0) THEN
                  farming_env%Job(todo)%status = job_finished
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, FMT=*) "Job finished: ", todo
                     CALL m_flush(output_unit)
                  ENDIF
               ENDIF

               ! get the next job in line, this could be do_nothing, if we're finished
               CALL get_next_job(farming_env, ijob_start, ijob_end, ijob_current, todo)
               dest = source
               tag = 2
               CALL mp_send(todo, dest, tag, para_env%group)

               IF (todo > 0) THEN
                  farming_env%Job(todo)%status = job_running
                  IF (output_unit > 0) THEN
                     WRITE (output_unit, FMT=*) "Job: ", todo, " Dir: ", TRIM(farming_env%Job(todo)%cwd), &
                        " assigned to group ", group_distribution(slave_distribution(dest))
                     CALL m_flush(output_unit)
                  ENDIF
               ELSE
                  IF (todo == do_nothing) THEN
                     slave_status(group_distribution(slave_distribution(dest))) = slave_status_done
                     IF (output_unit > 0) THEN
                        WRITE (output_unit, FMT=*) "group done: ", group_distribution(slave_distribution(dest))
                        CALL m_flush(output_unit)
                     ENDIF
                  ENDIF
                  IF (todo == do_deadlock) THEN
                     IF (output_unit > 0) THEN
                        WRITE (output_unit, FMT=*) ""
                        WRITE (output_unit, FMT=*) "FARMING JOB DEADLOCKED ... CIRCULAR DEPENDENCIES"
                        WRITE (output_unit, FMT=*) ""
                        CALL m_flush(output_unit)
                     ENDIF
                     IF(.NOT.(todo .NE. do_deadlock))CALL cp__a("start/cp2k_runs.F",684)
                  ENDIF
               ENDIF

            ENDDO

            DEALLOCATE (slave_status)

         ENDIF
      ELSE
         ! this is the non-master-slave mode way of executing the jobs
         ! the i-th job in the input is always executed by the MODULO(i-1,ngroups)-th group
         ! (needed for cyclic runs, we don't want two groups working on the same job)
         IF (output_unit > 0) THEN
            IF (ijob_end-ijob_start >= 0) THEN
               WRITE (output_unit, FMT="(T2,A)") "FARMING| List of jobs : "
               DO ijob = ijob_start, ijob_end
                  i = MODULO(ijob-1, farming_env%njobs)+1
                  WRITE (output_unit, FMT=*) "Job: ", i, " Dir: ", TRIM(farming_env%Job(i)%cwd), " Input: ", &
                     TRIM(farming_env%Job(i)%input), " MPI group:", MODULO(i-1, ngroups)
               ENDDO
            ENDIF
            CALL m_flush(output_unit)
         ENDIF

         DO ijob = ijob_start, ijob_end
            i = MODULO(ijob-1, farming_env%njobs)+1
            ! this farms out the jobs
            IF (MODULO(i-1, ngroups) == group_distribution(slave_rank)) THEN
               IF (output_unit > 0) THEN
                  WRITE (output_unit, FMT="(T2,A,I5.5,A)", ADVANCE="NO") " Running Job ", i, &
                     " in "//TRIM(farming_env%Job(i)%cwd)//"."
                  CALL m_flush(output_unit)
               ENDIF
               CALL execute_job(i)
               IF (output_unit > 0) THEN
                  WRITE (output_unit, FMT="(A)") " Done, output in "//TRIM(output_file)
                  CALL m_flush(output_unit)
               ENDIF
            ENDIF
         ENDDO
      ENDIF

      ! keep information about how long each process has to wait
      ! i.e. the load imbalance
      t1 = m_walltime()
      CALL mp_sync(para_env%group)
      t2 = m_walltime()
      ALLOCATE (waittime(0:para_env%num_pe-1))
      waittime = 0.0_dp
      waittime(para_env%mepos) = t2-t1
      CALL mp_sum(waittime, para_env%group)
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,A)') "Process idle times [s] at the end of the run"
         DO i = 0, para_env%num_pe-1
            WRITE (output_unit, FMT='(A2,I6,A3,F8.3,A1)', ADVANCE="NO") &
               " (", i, " : ", waittime(i), ")"
            IF (MOD(i+1, 4) == 0) WRITE (output_unit, '(A)') ""
         ENDDO
         CALL m_flush(output_unit)
      ENDIF
      DEALLOCATE (waittime)

      ! give back the communicators of the split groups
      IF (slave) CALL mp_comm_free(new_group)
      CALL mp_comm_free(slave_group)

      ! and message passing deallocate structures
      DEALLOCATE (group_distribution)
      DEALLOCATE (slave_distribution)

      ! clean the farming env
      CALL deallocate_farming_env(farming_env)

      CALL cp_print_key_finished_output(output_unit, logger, root_section, &
                                        "FARMING%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param i ...
! **************************************************************************************************
      RECURSIVE SUBROUTINE execute_job(i)
      INTEGER                                            :: i

! change to the new working directory

         CALL m_chdir(TRIM(farming_env%Job(i)%cwd), ierr)
         IF (ierr .NE. 0) &
            CALL cp__b("start/cp2k_runs.F",775,"Failed to change dir to: "//TRIM(farming_env%Job(i)%cwd))

         ! generate a fresh call to cp2k_run
         IF (new_rank == 0) THEN

            IF (farming_env%Job(i)%output == "") THEN
               ! generate the output file
               WRITE (output_file, '(A12,I5.5)') "FARMING_OUT_", i
               CALL parser_create(my_parser, file_name=TRIM(farming_env%Job(i)%input))
               label = "&GLOBAL"
               CALL parser_search_string(my_parser, label, ignore_case=.TRUE., found=found)
               IF (found) THEN
                  CALL create_global_section(g_section)
                  CALL section_vals_create(g_data, g_section)
                  CALL cp_unit_set_create(default_units, "OUTPUT")
                  CALL section_vals_parse(g_data, my_parser, default_units)
                  CALL cp_unit_set_release(default_units)
                  CALL section_vals_val_get(g_data, "PROJECT", &
                                            c_val=str)
                  IF (str .NE. "") output_file = TRIM(str)//".out"
                  CALL section_vals_val_get(g_data, "OUTPUT_FILE_NAME", &
                                            c_val=str)
                  IF (str .NE. "") output_file = str
                  CALL section_vals_release(g_data)
                  CALL section_release(g_section)
               END IF
               CALL parser_release(my_parser)
            ELSE
               output_file = farming_env%Job(i)%output
            ENDIF

            CALL open_file(file_name=TRIM(output_file), &
                           file_action="WRITE", &
                           file_status="UNKNOWN", &
                           file_position="APPEND", &
                           unit_number=new_output_unit)
         ELSE
            ! this unit should be negative, otherwise all processors that get a default unit
            ! start writing output (to the same file, adding to confusion).
            ! error handling should be careful, asking for a local output unit if required
            new_output_unit = -1
         ENDIF

         run_OK = cp2k_run(input_declaration, TRIM(farming_env%Job(i)%input), new_output_unit, new_group)

         IF (new_rank == 0) CALL close_file(unit_number=new_output_unit)

         ! change to the original working directory
         CALL m_chdir(TRIM(farming_env%cwd), ierr)
         IF(.NOT.(ierr == 0))CALL cp__a("start/cp2k_runs.F",824)

      END SUBROUTINE execute_job
   END SUBROUTINE farming_run

! **************************************************************************************************
!> \brief writes a small html description of the cp2k input
!> \author joost [fawzi]
! **************************************************************************************************
   SUBROUTINE write_cp2k_html_manual()

      CHARACTER(len=*), PARAMETER :: routineN = 'write_cp2k_html_manual', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr
      TYPE(section_type), POINTER                        :: root_section

      NULLIFY (root_section)
      CALL create_cp2k_root_section(root_section)
      ! remove the default keyword that ignores things outside the section
      CALL keyword_release(root_section%keywords(0)%keyword)

      CALL section_describe_html(root_section, "InputReference", 0, default_output_unit)

      CALL open_file(unit_number=unit_nr, file_name="index.html", &
                     file_action="WRITE", file_status="REPLACE")
      WRITE (unit_nr, FMT='(A)') "<HTML><BODY><HEAD><TITLE>The cp2k input structure</TITLE>"
      WRITE (unit_nr, FMT='(A)') "<H1>CP2K input reference</H1>"
      WRITE (unit_nr, FMT='(A)') "<H2>Version information</H2>"
      WRITE (unit_nr, FMT='(A)') &
         "This html manual can be generated automatically from a given cp2k executable "// &
         "using the --html-manual command line option. The manual describes exactly that version of the code. "// &
         "This version has been generated using a cp2k version compiled on "//TRIM(compile_date)//". "// &
         "The SVN source code revision number is "//TRIM(compile_revision)//"."

      WRITE (unit_nr, FMT='(A)') '<H2>Search the online input reference manual</H2>'
      WRITE (unit_nr, FMT='(A)') '<TABLE><TR><TD>'
      ! Yahoo seems to actually search our pages
      WRITE (unit_nr, FMT='(A)') ' <!-- START CODE --> '
      WRITE (unit_nr, FMT='(A)') '         <div id="ysrchForm" style="border:1px solid #7E9DB9;'
      WRITE (unit_nr, FMT='(A)') '                                background:#FFFFFF;'
      WRITE (unit_nr, FMT='(A)') '                                width:300px;'
      WRITE (unit_nr, FMT='(A)') '                                margin:0 auto;'
      WRITE (unit_nr, FMT='(A)') '                                padding:20px; '
      WRITE (unit_nr, FMT='(A)') '                                position:relative;'
      WRITE (unit_nr, FMT='(A)') '                              ">'
      WRITE (unit_nr, FMT='(A)') &
         '           <form id="searchBoxForm_undefined" action="http://search.yahoo.com/search" style="padding:0;">'
      WRITE (unit_nr, FMT='(A)') '                         '
      WRITE (unit_nr, FMT='(A)') '                         <input name="ei" value="UTF-8" type="hidden">'
      WRITE (unit_nr, FMT='(A)') &
         '                         <input name="fr" value="ystg" type="hidden"><div style="padding:0 80px 0 0;zoom:1;">'
      WRITE (unit_nr, FMT='(A)') '                                         <input type="text" id="searchTerm"'
      WRITE (unit_nr, FMT='(A)') '                                                 onFocus="this.style.background=''#fff'';"'
      WRITE (unit_nr, FMT='(A)') &
         '                                                 onBlur="if(this.value=='''')this.style.background='// &
         '''#fff url(http://us.i1.yimg.com/us.yimg.com/i/us/sch/gr/horiz_pwrlogo_red2.gif) 3px center  no-repeat''"'
      WRITE (unit_nr, FMT='(A)') &
         ' name="p" style=" margin:1px 0; width:100%; border:1px solid #7E9DB9; color:#666666; height:18px; '// &
         'padding:0px 3px; background:#fff url(http://us.i1.yimg.com/us.yimg.com/i/us/sch/gr/horiz_pwrlogo_red2.gif)'// &
         ' 3px center no-repeat; position:relative;">'
      WRITE (unit_nr, FMT='(A)') ' <input type="submit" id="btn_undefined" value="Search" '
      WRITE (unit_nr, FMT='(A)') &
         ' style=" padding-bottom:2px; position:absolute; right:20px; top:20px; margin:0px; height:22px; width:65px; ">'
      WRITE (unit_nr, FMT='(A)') ' </div><ul style="color:#666666;'
      WRITE (unit_nr, FMT='(A)') '            font:11px/11px normal Arial, Helvetica, sans-serif;'
      WRITE (unit_nr, FMT='(A)') '            margin:0;'
      WRITE (unit_nr, FMT='(A)') '            padding:0;'
      WRITE (unit_nr, FMT='(A)') '            text-align:left;'
      WRITE (unit_nr, FMT='(A)') '            list-style-type:none;radios"><li style="display:inline;padding-right:10px;">'
      WRITE (unit_nr, FMT='(A)') &
         ' <input name="vs" id="web_radio" value=""    type="radio" style="vertical-align:middle;margin-right:5px; ">'
      WRITE (unit_nr, FMT='(A)') ' <label for="web_radio" style="vertical-align:middle;">Web</label>'
      WRITE (unit_nr, FMT='(A)') ' </li><li style="display:inline;padding-right:10px;">'
      WRITE (unit_nr, FMT='(A)') &
         ' <input name="vs" id="site_radio" value="www.cp2k.org"   checked="checked" type="radio"'// &
         ' style="vertical-align:middle;margin-right:5px; ">'
      WRITE (unit_nr, FMT='(A)') ' <label for="site_radio" style="vertical-align:middle;">this Site</label>'
      WRITE (unit_nr, FMT='(A)') ' </li></ul></form></div>'
      WRITE (unit_nr, FMT='(A)') ' <!-- END CODE --></TD><TD> '

      WRITE (unit_nr, FMT='(A)') '</TD></TR></TABLE>'

      WRITE (unit_nr, FMT='(A)') '<H2>Journal papers</H2>'
      WRITE (unit_nr, FMT='(A)') '<A HREF="references.html">List of references</A> cited in the CP2K input manual.'

      WRITE (unit_nr, FMT='(A)') '<H2>Units of Measurement</H2>'
      WRITE (unit_nr, FMT='(A)') '<A HREF="units.html">List of unites</A> organized into groups based on the physical quantity '// &
         "they measure. The units can be used to specify an alternative unit of measurement for keyword values, for which a "// &
         "default unit has been explicitly defined."

      WRITE (unit_nr, FMT='(A)') "<H2>Internal Input Preprocessor</H2>"
      WRITE (unit_nr, FMT='(A)') "Before the input is parsed, the input is run through a very simple internal preprocessor."
      WRITE (unit_nr, FMT='(A)') "The preprocessor recognizes the following directives independent of capitalization:<BR><DL>"
      WRITE (unit_nr, FMT='(A)') "<DT><B>@INCLUDE 'filename.inc'</B></DT>"
      WRITE (unit_nr, FMT='(A)') "<DD>The file referenced by <I>filename.inc</I> is included into the input file and parsed."
      WRITE (unit_nr, FMT='(A)') "Recursive inclusions are not allowed and the files have to exist in the current working "
      WRITE (unit_nr, FMT='(A)') "directory. There can be only one @INCLUDE statement per line. Single or double quotes "
      WRITE (unit_nr, FMT='(A)') "can be used and <B>have</B> to be used if the filename contains blanks.</DD>"
      WRITE (unit_nr, FMT='(A)') "<DT><B>@SET VAR value</B></DT>"
      WRITE (unit_nr, FMT='(A)') "<DD>Assigns the text <I>value</I> to the preprocessing variable <I>VAR</I>. <I>value</I> "
      WRITE (unit_nr, FMT='(A)') "is the text following <I>VAR</I> with the outer whitespace removed. The variable can be "
      WRITE (unit_nr, FMT='(A)') "recalled with a <I>${VAR}</I> statement. There can be only one @SET statement per line.</DD>"
      WRITE (unit_nr, FMT='(A)') "<DT><B>${VAR}</B></DT>"
      WRITE (unit_nr, FMT='(A)') "<DD>Expand the variable <I>VAR</I>. The text <I>${VAR}</I> is replaced with the value assigned "
      WRITE (unit_nr, FMT='(A)') "to <I>VAR</I> in the last corresponding @SET directive. There can be multiple variables per line."
      WRITE (unit_nr, FMT='(A)') "The expansion process is repeated until no more variables are found.</DD>"
      WRITE (unit_nr, FMT='(A)') "<DT><B>@IF / @ENDIF</B></DT>"
      WRITE (unit_nr, FMT='(A)') "<DD>Conditional block. The text from the @IF line up to the next line with a valid "
      WRITE (unit_nr, FMT='(A)') "@ENDIF is skipped, if the expression following @IF resolves to <I>false</I>. "
      WRITE (unit_nr, FMT='(A)') "Available expressions are lexical comparisons for equality '==' or inequality '/='."
      WRITE (unit_nr, FMT='(A)') "If none of the two operators are found, a '0' or whitespace resolves to <I>false</I> "
      WRITE (unit_nr, FMT='(A)') "while any text resolves to <I>true</I>. @IF/@ENDIF blocks cannot be nested and "
      WRITE (unit_nr, FMT='(A)') "cannot span across files. There can be only one test (== or /=) per @IF statement.</DD>"
      WRITE (unit_nr, FMT='(A)') "</DL><P>"

      WRITE (unit_nr, FMT='(A)') "<H2>Input structure</H2>"
      WRITE (unit_nr, FMT='(A)') "All sections that can be part of a cp2k input file are shown with their allowed nestings. "
      WRITE (unit_nr, FMT='(A)') "A description of each section, and a list of keywords can be obtained clicking on the links. "
      WRITE (unit_nr, FMT='(A)') "<BR><UL>"

      CALL section_describe_index_html(root_section, "InputReference", unit_nr)

      WRITE (unit_nr, FMT='(A)') '</UL><BR><hr>Back to the <A HREF="https://www.cp2k.org/">CP2K homepage</A>'
      WRITE (unit_nr, FMT='(A)') "</BODY></HTML>"

      CALL close_file(unit_nr)
      CALL section_release(root_section)

      ! References
      CALL open_file(unit_number=unit_nr, file_name="references.html", &
                     file_action="WRITE", file_status="REPLACE")
      WRITE (unit_nr, FMT='(A)') "<HTML><BODY><HEAD><TITLE>The cp2k literature list</TITLE>"
      WRITE (unit_nr, FMT='(A)') "<H1>CP2K references</H1>"
      CALL print_all_references(sorted=.TRUE., cited_only=.FALSE., &
                                FORMAT=print_format_html, unit=unit_nr)
      WRITE (unit_nr, FMT='(A)') '<BR><hr>Back to the <A HREF="https://www.cp2k.org/">CP2K homepage</A>'
      WRITE (unit_nr, FMT='(A)') "</BODY></HTML>"
      CALL close_file(unit_nr)

      ! Units
      CALL open_file(unit_number=unit_nr, file_name="units.html", &
                     file_action="WRITE", file_status="REPLACE")
      WRITE (unit_nr, FMT='(A)') "<HTML><BODY><HEAD><TITLE>The cp2k units list</TITLE>"
      WRITE (unit_nr, FMT='(A)') "<H1>CP2K Available Units of Measurement</H1>"
      CALL print_all_units(unit_nr=unit_nr)
      WRITE (unit_nr, FMT='(A)') '<BR><hr>Back to the <A HREF="https://www.cp2k.org/">CP2K homepage</A>'
      WRITE (unit_nr, FMT='(A)') "</BODY></HTML>"
      CALL close_file(unit_nr)
   END SUBROUTINE write_cp2k_html_manual

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE write_xml_file()

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_xml_file', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ie, is, unit_number
      TYPE(section_type), POINTER                        :: root_section

      NULLIFY (root_section)
      CALL create_cp2k_root_section(root_section)
      CALL keyword_release(root_section%keywords(0)%keyword)
      CALL open_file(unit_number=unit_number, &
                     file_name="cp2k_input.xml", &
                     file_action="WRITE", &
                     file_status="REPLACE")

      WRITE (UNIT=unit_number, FMT="(A)") &
         "<?xml version=""1.0"" encoding=""ISO-8859-1""?>", &
         "<?xml-stylesheet type=""text/xsl"" href=""cp2k_input.xsl""?>"

      !MK Write a HTML translation table
      !MK In principle this is only required for non-standard HTML entities
      is = 0
      ie = 0
      WRITE (UNIT=unit_number, FMT="(A)") &
         "<!DOCTYPE documentElement["
      DO i = 1, SIZE(html_entity_table), 2
         is = INDEX(html_entity_table(i), "&")+1
         IF(.NOT.(is > 0))CALL cp__a("start/cp2k_runs.F",1005)
         ie = INDEX(html_entity_table(i), ";")-1
         IF(.NOT.(ie >= is))CALL cp__a("start/cp2k_runs.F",1007)
         WRITE (UNIT=unit_number, FMT="(A)") &
            "<!ENTITY "//html_entity_table(i) (is:ie)//" """//TRIM(html_entity_table(i+1))//""">"
      END DO
      WRITE (UNIT=unit_number, FMT="(A)") &
         "]>"

      !MK CP2K input structure
      WRITE (UNIT=unit_number, FMT="(A)") &
         "<CP2K_INPUT>", &
         " <CP2K_VERSION>"//TRIM(cp2k_version)//"</CP2K_VERSION>", &
         " <CP2K_YEAR>"//TRIM(cp2k_year)//"</CP2K_YEAR>", &
         " <COMPILE_DATE>"//TRIM(compile_date)//"</COMPILE_DATE>", &
         " <COMPILE_REVISION>"//TRIM(compile_revision)//"</COMPILE_REVISION>"
      DO i = 1, root_section%n_subsections
         CALL write_section_xml(root_section%subsections(i)%section, 1, unit_number)
      END DO
      ! Append HTML entity and tag tables
      CALL write_html_tables(unit_number)
      WRITE (UNIT=unit_number, FMT="(A)") "</CP2K_INPUT>"
      CALL close_file(unit_number=unit_number)
      CALL section_release(root_section)

      ! References
      CALL open_file(unit_number=unit_number, file_name="references.html", &
                     file_action="WRITE", file_status="REPLACE")
      WRITE (unit_number, FMT='(A)') "<HTML><BODY><HEAD><TITLE>The cp2k literature list</TITLE>"
      WRITE (unit_number, FMT='(A)') "<H1>CP2K references</H1>"
      CALL print_all_references(sorted=.TRUE., cited_only=.FALSE., &
                                FORMAT=print_format_html, unit=unit_number)
      WRITE (unit_number, FMT='(A)') "</BODY></HTML>"
      CALL close_file(unit_number=unit_number)

      ! Units
      CALL open_file(unit_number=unit_number, file_name="units.html", &
                     file_action="WRITE", file_status="REPLACE")
      WRITE (unit_number, FMT='(A)') "<HTML><BODY><HEAD><TITLE>The cp2k units list</TITLE>"
      WRITE (unit_number, FMT='(A)') "<H1>CP2K Available Units of Measurement</H1>"
      CALL print_all_units(unit_nr=unit_number)
      WRITE (unit_number, FMT='(A)') "</BODY></HTML>"
      CALL close_file(unit_number=unit_number)

   END SUBROUTINE write_xml_file

! **************************************************************************************************
!> \brief runs the given input
!> \param input_declaration ...
!> \param input_file_path the path of the input file
!> \param output_file_path path of the output file (to which it is appended)
!>        if it is "__STD_OUT__" the default_output_unit is used
!> \param ierr will return a number different from 0 if there was an error
!> \param mpi_comm the mpi communicator to be used for this environment
!>        it will not be freed
!> \author fawzi
!> \note
!>      moved here because of circular dependencies
! **************************************************************************************************
   SUBROUTINE run_input(input_declaration, input_file_path, output_file_path, ierr, mpi_comm)
      TYPE(section_type), POINTER                        :: input_declaration
      CHARACTER(len=*), INTENT(in)                       :: input_file_path, output_file_path
      INTEGER, INTENT(out)                               :: ierr
      INTEGER, INTENT(in), OPTIONAL                      :: mpi_comm

      CHARACTER(len=*), PARAMETER :: routineN = 'run_input', routineP = moduleN//':'//routineN

      INTEGER                                            :: unit_nr
      LOGICAL                                            :: success
      TYPE(cp_para_env_type), POINTER                    :: para_env

      IF (PRESENT(mpi_comm)) THEN
         NULLIFY (para_env)
         CALL cp_para_env_create(para_env, group=mpi_comm, owns_group=.FALSE.) !XXXXXXXXXXXX uninitiliased error
      ELSE
         para_env => f77_default_para_env
         CALL cp_para_env_retain(para_env) !XXXXXXXXXXXX uninitiliased error
      END IF
      IF (para_env%mepos == para_env%source) THEN
         IF (output_file_path == "__STD_OUT__") THEN
            unit_nr = default_output_unit
         ELSE
            CALL open_file(file_name=output_file_path, file_status="UNKNOWN", &
                           file_action="WRITE", file_position="APPEND", &
                           unit_number=unit_nr)
         END IF
      ELSE
         unit_nr = -1
      END IF
      success = cp2k_run(input_declaration, input_file_path, unit_nr, para_env%group)
      IF (.NOT. success) THEN
         ierr = cp_failure_level
      ELSE
         ierr = 0
      END IF
      CALL cp_para_env_release(para_env) !XXXXXXXXXXXX uninitiliased error
   END SUBROUTINE run_input

END MODULE cp2k_runs
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Main program of CP2K
!> \par Copyright
!>    CP2K: A general program to perform molecular dynamics simulations
!>    Copyright (C) 2000, 2001, 2002, 2003  CP2K developers group
!>    Copyright (C) 2004, 2005, 2006, 2007  CP2K developers group
!>    Copyright (C) 2008, 2009, 2010, 2011  CP2K developers group
!>    Copyright (C) 2012, 2013, 2014, 2015  CP2K developers group
!>    Copyright (C) 2016                    CP2K developers group
!> \par
!>    This program is free software; you can redistribute it and/or modify
!>    it under the terms of the GNU General Public License as published by
!>    the Free Software Foundation; either version 2 of the License, or
!>    (at your option) any later version.
!> \par
!>    This program is distributed in the hope that it will be useful,
!>    but WITHOUT ANY WARRANTY; without even the implied warranty of
!>    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!>    GNU General Public License for more details.
!> \par
!>    You should have received a copy of the GNU General Public License
!>    along with this program; if not, write to the Free Software
!>    Foundation, Inc., 51 Franklin Street, Fifth Floor,
!>    Boston, MA  02110-1301, USA.
!> \par
!>    See also https://www.fsf.org/licensing/licenses/gpl.html
!> \par
!>    CP2K, including its sources and pointers to the authors
!>    can be found at  https://www.cp2k.org/
!> \note
!>       should be kept as lean as possible.
!>       see cp2k_run for more comments
!> \author Joost VandeVondele
! **************************************************************************************************
PROGRAM cp2k

   USE cp2k_info,                       ONLY: compile_revision,&
                                              cp2k_flags,&
                                              cp2k_version,&
                                              id_cp2k_version,&
                                              print_cp2k_license
   USE cp2k_runs,                       ONLY: run_input,&
                                              write_cp2k_html_manual,&
                                              write_xml_file
   USE f77_interface,                   ONLY: check_input,&
                                              default_para_env,&
                                              finalize_cp2k,&
                                              init_cp2k
   USE input_constants,                 ONLY: id_development_version
   USE input_cp2k,                      ONLY: create_cp2k_root_section
   USE input_section_types,             ONLY: section_release,&
                                              section_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length
   USE machine,                         ONLY: default_output_unit,&
                                              m_getarg,&
                                              m_iargc
   USE message_passing,                 ONLY: mp_max
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(LEN=default_path_length)   :: input_file_name, output_file_name, &
                                           arg_att
   CHARACTER(LEN=default_string_length) :: dev_flag
   INTEGER                              :: output_unit, l
   INTEGER                              :: ierr, n_arg, i_arg
   LOGICAL                              :: check, manual, usage, echo_input, command_line_error
   LOGICAL                              :: run_it, force_run, has_input, xml, print_version, print_license
   TYPE(section_type), POINTER          :: input_declaration

   NULLIFY (input_declaration)

   ! output goes to the screen by default
   output_unit = default_output_unit

   ! set default behaviour for the command line switches
   check = .FALSE.
   manual = .FALSE.
   usage = .FALSE.
   echo_input = .FALSE.
   has_input = .FALSE.
   run_it = .TRUE.
   force_run = .FALSE.
   print_version = .FALSE.
   print_license = .FALSE.
   command_line_error = .FALSE.
   xml = .FALSE.
   input_file_name = "Missing input file name" ! no default
   output_file_name = "__STD_OUT__" ! by default we go to std_out

   ! parsing of command line arguments
   n_arg = m_iargc()

   ! this is an error condition, provide usage information
   IF (n_arg < 1) THEN
      WRITE (output_unit, "(/,T2,A)") &
         "ERROR: At least one command line argument must be specified"
      command_line_error = .TRUE.
   END IF

   i_arg = 0
   arg_loop: DO WHILE (i_arg < n_arg)
      i_arg = i_arg+1
      CALL m_getarg(i_arg, arg_att)
      SELECT CASE (arg_att)
      CASE ("--check", "-c")
         check = .TRUE.
         run_it = .FALSE.
         echo_input = .FALSE.
      CASE ("--echo", "-e")
         check = .TRUE.
         run_it = .FALSE.
         echo_input = .TRUE.
      CASE ("--html-manual")
         manual = .TRUE.
         run_it = .FALSE.
      CASE ("-v", "--version")
         print_version = .TRUE.
         run_it = .FALSE.
      CASE ("--license")
         print_license = .TRUE.
         run_it = .FALSE.
      CASE ("--run", "-r")
         force_run = .TRUE.
      CASE ("-help", "--help", "-h")
         usage = .TRUE.
         run_it = .FALSE.
      CASE ("-i")
         i_arg = i_arg+1
         CALL m_getarg(i_arg, arg_att)
         ! argument does not start with a - it is an filename
         IF (.NOT. arg_att(1:1) == "-") THEN
            input_file_name = arg_att
            has_input = .TRUE.
         ELSE
            WRITE (output_unit, "(/,T2,A)") &
               "ERROR: The input file name"//TRIM(arg_att)//" starts with -"
            command_line_error = .TRUE.
            EXIT arg_loop
         ENDIF
      CASE ("-o")
         i_arg = i_arg+1
         CALL m_getarg(i_arg, arg_att)
         ! argument does not start with a - it is an filename
         IF (.NOT. arg_att(1:1) == "-") THEN
            output_file_name = arg_att
         ELSE
            WRITE (output_unit, "(/,T2,A)") &
               "ERROR: The output file name"//TRIM(arg_att)//" starts with -"
            command_line_error = .TRUE.
            EXIT arg_loop
         ENDIF
      CASE ("--xml")
         xml = .TRUE.
         run_it = .FALSE.
      CASE default
         ! if the last argument does not start with a - it is an input filename
         !MK in order to digest the additional flags of mpirun
         IF ((.NOT. has_input) .AND. &
             (i_arg == n_arg) .AND. &
             (.NOT. arg_att(1:1) == "-")) THEN
            input_file_name = arg_att
            has_input = .TRUE.
         ELSEIF (has_input .AND. &
                 (.NOT. arg_att(1:1) == "-")) THEN
            WRITE (output_unit, "(/,T2,A)") &
               "Error: Tried to specify two input files"
            command_line_error = .TRUE.
            EXIT arg_loop
         END IF
      END SELECT
   END DO arg_loop
   IF ((run_it .OR. force_run .OR. check .OR. echo_input) .AND. &
       (.NOT. has_input) .AND. (.NOT. command_line_error)) THEN
      WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
         "ERROR: An input file name is required"
      command_line_error = .TRUE.
   END IF

   CALL init_cp2k(init_mpi=.TRUE., ierr=ierr)

   IF (ierr == 0) THEN
      ! some first info concerning how to run CP2K

      IF (usage .OR. command_line_error) THEN
         IF (default_para_env%source == default_para_env%mepos) THEN
            CALL m_getarg(0, arg_att) ! actually this is non-standard
            l = LEN_TRIM(arg_att)
            WRITE (UNIT=output_unit, FMT="(/,(T2,A))") &
               TRIM(arg_att)//" [-c|--check] [-e|--echo] [-h|--help] [--html-manual]", &
               REPEAT(" ", l)//" [-i] <input_file>", &
               REPEAT(" ", l)//" [-mpi-mapping|--mpi-mapping] <method>", &
               REPEAT(" ", l)//" [-o] <output_file>", &
               REPEAT(" ", l)//" [-r|-run] [--xml]"
            WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,T2,A,/,/,T2,A,/,/,(T3,A))") &
               "starts the CP2K program, see <https://www.cp2k.org/>", &
               "The easiest way is "//TRIM(arg_att)//" <input_file>", &
               "The following options can be used:", &
               "-i <input_file>   : provides an input file name, if it is the last", &
               "                    argument, the -i flag is not needed", &
               "-o <output_file>  : provides an output file name [default: screen]"
            WRITE (UNIT=output_unit, FMT="(/,T2,A,/,/,(T3,A))") &
               "These switches skip the simulation, unless [-r|-run] is specified:", &
               "--check, -c       : performs a syntax check of the <input_file>", &
               "--echo, -e        : echos the <input_file>, and make all defaults explicit", &
               "                    The input is also checked, but only a failure is reported", &
               "--help, -h        : writes this message", &
               "--html-manual     : writes a HTML reference manual of the CP2K input", &
               "                    in the current directory. The file index.html is a good", &
               "                    starting point for browsing", &
               "--license         : prints the CP2K license", &
               "--mpi-mapping     : applies a given MPI reordering to CP2K", &
               "--run, -r         : forces a CP2K run regardless of other specified flags", &
               "--version, -v     : prints the CP2K version and the SVN revision number", &
               "--xml             : dumps the whole CP2K input structure as a XML file", &
               "                    xml2htm generates a HTML manual from this XML file", &
               ""
         END IF
      END IF

      IF (.NOT. command_line_error) THEN

         ! write the version string
         IF (print_version) THEN
            IF (default_para_env%source == default_para_env%mepos) THEN
               dev_flag = ""
               IF (id_cp2k_version == id_development_version) dev_flag = " (Development Version)"
               WRITE (output_unit, "(T2,A)") &
                  cp2k_version//TRIM(dev_flag), &
                  "SVN source code revision "//TRIM(compile_revision), &
                  TRIM(cp2k_flags())
            ENDIF
         ENDIF

         ! write the license
         IF (print_license) THEN
            IF (default_para_env%source == default_para_env%mepos) THEN
               CALL print_cp2k_license(output_unit)
            ENDIF
         ENDIF

         ! if a manual is needed
         IF (manual) THEN
            IF (default_para_env%mepos == default_para_env%source) THEN
               CALL write_cp2k_html_manual()
            END IF
         ENDIF

         IF (xml) THEN
            IF (default_para_env%mepos == default_para_env%source) THEN
               CALL write_xml_file()
            END IF
         END IF

         CALL create_cp2k_root_section(input_declaration)

         IF (check) THEN
            CALL check_input(input_declaration, input_file_name, output_file_name, &
                             echo_input=echo_input, ierr=ierr)
            IF (default_para_env%source == default_para_env%mepos) THEN
               IF (ierr == 0) THEN
                  IF (.NOT. echo_input) THEN
                     WRITE (output_unit, "(A)") "SUCCESS, the input could be parsed correctly."
                     WRITE (output_unit, "(A)") "         This is does not guarantee that this input is meaningful"
                     WRITE (output_unit, "(A)") "         or will run successfully"
                  ENDIF
               ELSE
                  WRITE (output_unit, "(A)") "ERROR,   the input could *NOT* be parsed correctly."
                  WRITE (output_unit, "(A)") "         Please, check and correct it"
               END IF
            END IF
         ENDIF

         IF (run_it .OR. force_run) THEN
            CALL run_input(input_declaration, input_file_name, output_file_name, ierr=ierr)
            CALL mp_max(ierr, default_para_env%group)
            IF (ierr .NE. 0) CALL cp__b("start/cp2k.F",283,"Failed to run input")
         END IF

         CALL section_release(input_declaration)
      ENDIF
   ELSE
      WRITE (UNIT=output_unit, FMT="(/,A)") "initial setup (MPI ?) error"
   END IF

   ! and the final cleanup
   CALL finalize_cp2k(finalize_mpi=.TRUE., ierr=ierr)

END PROGRAM
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief CP2K C/C++ interface
!> \par History
!>       12.2012 created [Hossein Bani-Hashemian]
!>       04.2016 restructured [Hossein Bani-Hashemian, Ole Schuett]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE libcp2k
   USE ISO_C_BINDING,                   ONLY: C_CHAR,&
                                              C_DOUBLE,&
                                              C_FUNPTR,&
                                              C_INT,&
                                              C_NULL_CHAR
   USE cp2k_info,                       ONLY: cp2k_version
   USE cp2k_runs,                       ONLY: run_input
   USE f77_interface,                   ONLY: &
        calc_energy_force, create_force_env, destroy_force_env, f_env_add_defaults, &
        f_env_rm_defaults, f_env_type, finalize_cp2k, get_energy, get_force, get_natom, &
        get_nparticle, get_pos, get_result_r1, init_cp2k, set_pos, set_vel
   USE force_env_types,                 ONLY: force_env_get,&
                                              use_qs_force
   USE input_cp2k,                      ONLY: create_cp2k_root_section
   USE input_section_types,             ONLY: section_release,&
                                              section_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param version_str ...
!> \param str_length ...
! **************************************************************************************************
   SUBROUTINE cp2k_get_version(version_str, str_length) BIND(C)
      CHARACTER(LEN=1, KIND=C_CHAR), INTENT(OUT)         :: version_str(*)
      INTEGER(C_INT), VALUE                              :: str_length

      INTEGER                                            :: i, n

      n = LEN_TRIM(cp2k_version)
      IF(.NOT.(str_length > n+1))CALL cp__a("start/libcp2k.F",52)

      ! copy string
      DO i = 1, n
         version_str(i) = cp2k_version(i:i)
      ENDDO
      version_str(n+1) = C_NULL_CHAR
   END SUBROUTINE cp2k_get_version

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE cp2k_init() BIND(C)
      INTEGER                                            :: ierr

      CALL init_cp2k(.TRUE., ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",68)
   END SUBROUTINE cp2k_init

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE cp2k_init_without_mpi() BIND(C)
      INTEGER                                            :: ierr

      CALL init_cp2k(.FALSE., ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",78)
   END SUBROUTINE cp2k_init_without_mpi

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE cp2k_finalize() BIND(C)
      INTEGER                                            :: ierr

      CALL finalize_cp2k(.TRUE., ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",88)
   END SUBROUTINE cp2k_finalize

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE cp2k_finalize_without_mpi() BIND(C)
      INTEGER                                            :: ierr

      CALL finalize_cp2k(.FALSE., ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",98)
   END SUBROUTINE cp2k_finalize_without_mpi

! **************************************************************************************************
!> \brief ...
!> \param new_env_id ...
!> \param input_file_path ...
!> \param output_file_path ...
! **************************************************************************************************
   SUBROUTINE cp2k_create_force_env(new_env_id, input_file_path, output_file_path) BIND(C)
      INTEGER(C_INT), INTENT(OUT)                        :: new_env_id
      CHARACTER(LEN=1, KIND=C_CHAR), INTENT(IN)          :: input_file_path(*), output_file_path(*)

      CHARACTER(LEN=default_path_length)                 :: ifp, ofp
      INTEGER                                            :: ierr
      TYPE(section_type), POINTER                        :: input_declaration

      ifp = " "; ofp = " "
      CALL strncpy_c2f(ifp, input_file_path)
      CALL strncpy_c2f(ofp, output_file_path)

      NULLIFY (input_declaration)
      CALL create_cp2k_root_section(input_declaration)
      CALL create_force_env(new_env_id, input_declaration, ifp, ofp, ierr=ierr)
      CALL section_release(input_declaration)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",123)
   END SUBROUTINE cp2k_create_force_env

! **************************************************************************************************
!> \brief ...
!> \param new_env_id ...
!> \param input_file_path ...
!> \param output_file_path ...
!> \param mpi_comm ...
! **************************************************************************************************
   SUBROUTINE cp2k_create_force_env_comm(new_env_id, input_file_path, output_file_path, mpi_comm) BIND(C)
      INTEGER(C_INT), INTENT(OUT)                        :: new_env_id
      CHARACTER(LEN=1, KIND=C_CHAR), INTENT(IN)          :: input_file_path(*), output_file_path(*)
      INTEGER(C_INT), VALUE                              :: mpi_comm

      CHARACTER(LEN=default_path_length)                 :: ifp, ofp
      INTEGER                                            :: ierr
      TYPE(section_type), POINTER                        :: input_declaration

      ifp = " "; ofp = " "
      CALL strncpy_c2f(ifp, input_file_path)
      CALL strncpy_c2f(ofp, output_file_path)

      NULLIFY (input_declaration)
      CALL create_cp2k_root_section(input_declaration)
      CALL create_force_env(new_env_id, input_declaration, ifp, ofp, mpi_comm, ierr=ierr)
      CALL section_release(input_declaration)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",150)
   END SUBROUTINE cp2k_create_force_env_comm

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
! **************************************************************************************************
   SUBROUTINE cp2k_destroy_force_env(env_id) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id

      INTEGER                                            :: ierr

      CALL destroy_force_env(env_id, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",163)
   END SUBROUTINE cp2k_destroy_force_env

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param new_pos ...
!> \param n_el ...
! **************************************************************************************************
   SUBROUTINE cp2k_set_positions(env_id, new_pos, n_el) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id, n_el
      REAL(C_DOUBLE), DIMENSION(1:n_el), INTENT(IN)      :: new_pos

      INTEGER                                            :: ierr

      CALL set_pos(env_id, new_pos, n_el, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",179)
   END SUBROUTINE cp2k_set_positions

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param new_vel ...
!> \param n_el ...
! **************************************************************************************************
   SUBROUTINE cp2k_set_velocities(env_id, new_vel, n_el) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id, n_el
      REAL(C_DOUBLE), DIMENSION(1:n_el), INTENT(IN)      :: new_vel

      INTEGER                                            :: ierr

      CALL set_vel(env_id, new_vel, n_el, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",195)
   END SUBROUTINE cp2k_set_velocities

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param description ...
!> \param RESULT ...
!> \param n_el ...
! **************************************************************************************************
   SUBROUTINE cp2k_get_result(env_id, description, RESULT, n_el) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id
      CHARACTER(LEN=1, KIND=C_CHAR), INTENT(IN)          :: description(*)
      INTEGER(C_INT), VALUE                              :: n_el
      REAL(C_DOUBLE), DIMENSION(1:n_el), INTENT(OUT)     :: RESULT

      CHARACTER(LEN=default_string_length)               :: desc_low
      INTEGER                                            :: ierr

      desc_low = " "
      CALL strncpy_c2f(desc_low, description)

      CALL get_result_r1(env_id, desc_low, n_el, RESULT, ierr=ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",218)
   END SUBROUTINE cp2k_get_result

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE cp2k_get_natom(env_id, natom) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id
      INTEGER(C_INT), INTENT(OUT)                        :: natom

      INTEGER                                            :: ierr

      CALL get_natom(env_id, natom, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",233)
   END SUBROUTINE cp2k_get_natom

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param nparticle ...
! **************************************************************************************************
   SUBROUTINE cp2k_get_nparticle(env_id, nparticle) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id
      INTEGER(C_INT), INTENT(OUT)                        :: nparticle

      INTEGER                                            :: ierr

      CALL get_nparticle(env_id, nparticle, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",248)
   END SUBROUTINE cp2k_get_nparticle

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param pos ...
!> \param n_el ...
! **************************************************************************************************
   SUBROUTINE cp2k_get_positions(env_id, pos, n_el) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id, n_el
      REAL(C_DOUBLE), DIMENSION(1:n_el), INTENT(OUT)     :: pos

      INTEGER                                            :: ierr

      CALL get_pos(env_id, pos, n_el, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",264)
   END SUBROUTINE cp2k_get_positions

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param force ...
!> \param n_el ...
! **************************************************************************************************
   SUBROUTINE cp2k_get_forces(env_id, force, n_el) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id, n_el
      REAL(C_DOUBLE), DIMENSION(1:n_el), INTENT(OUT)     :: force

      INTEGER                                            :: ierr

      CALL get_force(env_id, force, n_el, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",280)
   END SUBROUTINE cp2k_get_forces

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param e_pot ...
! **************************************************************************************************
   SUBROUTINE cp2k_get_potential_energy(env_id, e_pot) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id
      REAL(C_DOUBLE), INTENT(OUT)                        :: e_pot

      INTEGER                                            :: ierr

      CALL get_energy(env_id, e_pot, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",295)
   END SUBROUTINE cp2k_get_potential_energy

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
! **************************************************************************************************
   SUBROUTINE cp2k_calc_energy_force(env_id) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id

      INTEGER                                            :: ierr

      CALL calc_energy_force(env_id, .TRUE., ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",308)
   END SUBROUTINE cp2k_calc_energy_force

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
! **************************************************************************************************
   SUBROUTINE cp2k_calc_energy(env_id) BIND(C)
      INTEGER(C_INT), VALUE                              :: env_id

      INTEGER                                            :: ierr

      CALL calc_energy_force(env_id, .FALSE., ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",321)
   END SUBROUTINE cp2k_calc_energy

! **************************************************************************************************
!> \brief ...
!> \param input_file_path ...
!> \param output_file_path ...
! **************************************************************************************************
   SUBROUTINE cp2k_run_input(input_file_path, output_file_path) BIND(C)
      CHARACTER(LEN=1, KIND=C_CHAR), INTENT(IN)          :: input_file_path(*), output_file_path(*)

      CHARACTER(LEN=default_path_length)                 :: ifp, ofp
      INTEGER                                            :: ierr
      TYPE(section_type), POINTER                        :: input_declaration

      ifp = " "; ofp = " "
      CALL strncpy_c2f(ifp, input_file_path)
      CALL strncpy_c2f(ofp, output_file_path)

      NULLIFY (input_declaration)
      CALL create_cp2k_root_section(input_declaration)
      CALL run_input(input_declaration, ifp, ofp, ierr=ierr)
      CALL section_release(input_declaration)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",344)
   END SUBROUTINE cp2k_run_input

! **************************************************************************************************
!> \brief ...
!> \param input_file_path ...
!> \param output_file_path ...
!> \param mpi_comm ...
! **************************************************************************************************
   SUBROUTINE cp2k_run_input_comm(input_file_path, output_file_path, mpi_comm) BIND(C)
      CHARACTER(LEN=1, KIND=C_CHAR), INTENT(IN)          :: input_file_path(*), output_file_path(*)
      INTEGER(C_INT), VALUE                              :: mpi_comm

      CHARACTER(LEN=default_path_length)                 :: ifp, ofp
      INTEGER                                            :: ierr
      TYPE(section_type), POINTER                        :: input_declaration

      ifp = " "; ofp = " "
      CALL strncpy_c2f(ifp, input_file_path)
      CALL strncpy_c2f(ofp, output_file_path)

      NULLIFY (input_declaration)
      CALL create_cp2k_root_section(input_declaration)
      CALL run_input(input_declaration, ifp, ofp, ierr, mpi_comm)
      CALL section_release(input_declaration)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",369)
   END SUBROUTINE cp2k_run_input_comm

! **************************************************************************************************
!> \brief Gets a function pointer pointing to a routine defined in C/C++ and
!>        passes it to the transport environment in force environment
!> \param f_env_id  the force env id
!> \param func_ptr the function pointer
!> \par History
!>      12.2012 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE cp2k_transport_set_callback(f_env_id, func_ptr) BIND(C)
      INTEGER(C_INT), VALUE                              :: f_env_id
      TYPE(C_FUNPTR), VALUE                              :: func_ptr

      INTEGER                                            :: ierr, in_use
      TYPE(f_env_type), POINTER                          :: f_env

      NULLIFY (f_env)
      CALL f_env_add_defaults(f_env_id, f_env)
      CALL force_env_get(f_env%force_env, in_use=in_use)
      IF (in_use .EQ. use_qs_force) THEN
         f_env%force_env%qs_env%transport_env%ext_c_method_ptr = func_ptr
      END IF
      CALL f_env_rm_defaults(f_env, ierr)
      IF(.NOT.(ierr == 0))CALL cp__a("start/libcp2k.F",395)
   END SUBROUTINE cp2k_transport_set_callback

! **************************************************************************************************
!> \brief Copy the content of a \0-terminated C-string to a finite-length Fortran string
!>
!> The content of the new string may be truncated if the number of characters before the '\0'
!> in the source string exceed the length of the destination string.
!> \param fstring destination string
!> \param cstring source string
!> \author Tiziano Mueller
! **************************************************************************************************
   SUBROUTINE strncpy_c2f(fstring, cstring)
      CHARACTER(LEN=*), INTENT(OUT)                      :: fstring
      CHARACTER(LEN=1, KIND=C_CHAR), INTENT(IN)          :: cstring(*)

      INTEGER                                            :: i

      DO i = 1, LEN(fstring)
         IF (cstring(i) == C_NULL_CHAR) EXIT
         fstring(i:i) = cstring(i)
      END DO
   END SUBROUTINE strncpy_c2f

END MODULE libcp2k
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief interface to use cp2k as library (these are f77 wrappers to the
!>      f77_interface functions, for a description see in that module)
!> \param init_mpi ...
!> \param ierr ...
!> \par History
!>      07.2004 created [fawzi]
!>      11.2004 parallel version [fawzi]
!>      22.11.2010 get_nparticle nparticle added (MK)
!> \author fawzi
!> \note
!>      I tried to use just integer, real(db), and CHARACTER(len=*) to
!>      make the c binding/wrapping as easy as possible.
!>      The absence of intent from the arguments is also done on purpose.
! **************************************************************************************************
SUBROUTINE cp_init_cp2k(init_mpi, ierr)
  USE f77_interface, ONLY: icp => init_cp2k
   IMPLICIT NONE
   INTEGER :: init_mpi, ierr
   LOGICAL ::i_mpi
   i_mpi = init_mpi /= 0
   CALL icp(i_mpi, ierr)
END SUBROUTINE cp_init_cp2k

! **************************************************************************************************
!> \brief ...
!> \param finalize_mpi ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_finalize_cp2k(finalize_mpi, ierr)
  USE f77_interface, ONLY: kcp => finalize_cp2k
   IMPLICIT NONE
   INTEGER :: finalize_mpi, ierr
   LOGICAL ::f_mpi
   f_mpi = finalize_mpi /= 0
   CALL kcp(f_mpi, ierr)
END SUBROUTINE cp_finalize_cp2k

! **************************************************************************************************
!> \brief ...
!> \param new_env_id ...
!> \param input_file_path ...
!> \param output_file_path ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_create_fenv(new_env_id, input_file_path, output_file_path, &
                          ierr)
  USE f77_interface,                   ONLY: create_force_env
  USE input_cp2k,                      ONLY: create_cp2k_root_section
  USE input_section_types,             ONLY: section_type, section_release
   IMPLICIT NONE
   CHARACTER(len=*)                     :: input_file_path, output_file_path
   INTEGER                              :: new_env_id, ierr
   TYPE(section_type), POINTER          :: input_declaration

   NULLIFY (input_declaration)
   CALL create_cp2k_root_section(input_declaration)
   CALL create_force_env(new_env_id, input_declaration, input_file_path, output_file_path, ierr=ierr)
   CALL section_release(input_declaration)
END SUBROUTINE cp_create_fenv

! **************************************************************************************************
!> \brief ...
!> \param new_env_id ...
!> \param input_file_path ...
!> \param output_file_path ...
!> \param mpi_comm ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_create_fenv_comm(new_env_id, input_file_path, output_file_path, &
                               mpi_comm, ierr)
  USE f77_interface,                   ONLY: create_force_env
  USE input_cp2k,                      ONLY: create_cp2k_root_section
  USE input_section_types,             ONLY: section_type, section_release
   IMPLICIT NONE
   CHARACTER(len=*)                     :: input_file_path, output_file_path
   INTEGER                              :: new_env_id, mpi_comm, ierr
   TYPE(section_type), POINTER          :: input_declaration

   NULLIFY (input_declaration)
   CALL create_cp2k_root_section(input_declaration)
   CALL create_force_env(new_env_id, input_declaration, input_file_path, output_file_path, mpi_comm, ierr=ierr)
   CALL section_release(input_declaration)
END SUBROUTINE cp_create_fenv_comm

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param ierr ...
! **************************************************************************************************
RECURSIVE SUBROUTINE cp_destroy_fenv(env_id, ierr)
  USE f77_interface, ONLY: dfe => destroy_force_env
   IMPLICIT NONE
   INTEGER :: env_id, ierr

   CALL dfe(env_id, ierr)
END SUBROUTINE cp_destroy_fenv

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param new_pos ...
!> \param n_el ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_set_pos(env_id, new_pos, n_el, ierr)
  USE kinds, ONLY: dp
  USE f77_interface, ONLY: sp => set_pos
   IMPLICIT NONE
   INTEGER :: n_el, env_id, ierr
   REAL(kind=dp), DIMENSION(1:n_el) :: new_pos

   CALL sp(env_id, new_pos, n_el, ierr)
END SUBROUTINE cp_set_pos

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param new_vel ...
!> \param n_el ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_set_vel(env_id, new_vel, n_el, ierr)
  USE kinds, ONLY: dp
  USE f77_interface, ONLY: sv => set_vel
   IMPLICIT NONE
   INTEGER :: n_el, env_id, ierr
   REAL(kind=dp), DIMENSION(1:n_el) :: new_vel

   CALL sv(env_id, new_vel, n_el, ierr)
END SUBROUTINE cp_set_vel

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param description ...
!> \param N ...
!> \param RESULT ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_get_result_r1(env_id, description, N, RESULT, ierr)
  USE kinds, ONLY: dp, default_string_length
  USE f77_interface, ONLY: get_result_r1
   IMPLICIT NONE
   CHARACTER(LEN=*) :: description
   INTEGER :: N
   REAL(KIND=dp), DIMENSION(1:N) :: RESULT
   INTEGER :: env_id, ierr
   CHARACTER(LEN=default_string_length) :: description_low

   description_low = description
   CALL get_result_r1(env_id, description_low, N, RESULT, ierr=ierr)

END SUBROUTINE cp_get_result_r1

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param natom ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_get_natom(env_id, natom, ierr)
  USE f77_interface, ONLY: gna => get_natom
   IMPLICIT NONE
   INTEGER :: natom, env_id, ierr

   CALL gna(env_id, natom, ierr)
END SUBROUTINE cp_get_natom

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param nparticle ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_get_nparticle(env_id, nparticle, ierr)
  USE f77_interface, ONLY: gnp => get_nparticle
   IMPLICIT NONE
   INTEGER :: nparticle, env_id, ierr

   CALL gnp(env_id, nparticle, ierr)
END SUBROUTINE cp_get_nparticle

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param pos ...
!> \param n_el ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_get_pos(env_id, pos, n_el, ierr)
  USE kinds, ONLY: dp
  USE f77_interface, ONLY: gp => get_pos
   IMPLICIT NONE
   INTEGER :: n_el, env_id, ierr
   REAL(kind=dp), DIMENSION(1:n_el) :: pos

   CALL gp(env_id, pos, n_el, ierr)
END SUBROUTINE cp_get_pos

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param force ...
!> \param n_el ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_get_force(env_id, force, n_el, ierr)
  USE kinds, ONLY: dp
  USE f77_interface, ONLY: gf => get_force
   IMPLICIT NONE
   INTEGER :: n_el, env_id, ierr
   REAL(kind=dp), DIMENSION(1:n_el) :: force

   CALL gf(env_id, force, n_el, ierr)
END SUBROUTINE cp_get_force

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param e_pot ...
!> \param ierr ...
! **************************************************************************************************
RECURSIVE SUBROUTINE cp_get_energy(env_id, e_pot, ierr)
  USE kinds, ONLY: dp
  USE f77_interface, ONLY: ge => get_energy
   IMPLICIT NONE
   INTEGER :: env_id, ierr
   REAL(kind=dp) :: e_pot

   CALL ge(env_id, e_pot, ierr)
END SUBROUTINE cp_get_energy

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param calc_force ...
!> \param ierr ...
! **************************************************************************************************
RECURSIVE SUBROUTINE cp_calc_energy_force(env_id, calc_force, ierr)
  USE f77_interface, ONLY: cef => calc_energy_force
   IMPLICIT NONE
   INTEGER :: calc_force, env_id, ierr
   LOGICAL :: cf

   cf = (calc_force /= 0)
   CALL cef(env_id, cf, ierr)
END SUBROUTINE cp_calc_energy_force

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param pos ...
!> \param n_el ...
!> \param e_pot ...
!> \param ierr ...
! **************************************************************************************************
RECURSIVE SUBROUTINE cp_calc_energy(env_id, pos, n_el, e_pot, ierr)
  USE kinds, ONLY: dp
  USE f77_interface, ONLY: ce => calc_energy
   IMPLICIT NONE
   INTEGER :: env_id, ierr, n_el
   REAL(dp) :: e_pot
   REAL(dp), DIMENSION(1:n_el) :: pos

   CALL ce(env_id, pos, n_el, e_pot, ierr)
END SUBROUTINE cp_calc_energy

! **************************************************************************************************
!> \brief ...
!> \param env_id ...
!> \param pos ...
!> \param n_el_pos ...
!> \param e_pot ...
!> \param force ...
!> \param n_el_force ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_calc_force(env_id, pos, n_el_pos, e_pot, force, n_el_force, ierr)
  USE kinds, ONLY: dp
  USE f77_interface, ONLY: cf => calc_force
   IMPLICIT NONE
   INTEGER :: env_id, ierr, n_el_pos, n_el_force
   REAL(dp) :: e_pot
   REAL(dp), DIMENSION(1:n_el_pos) :: pos
   REAL(dp), DIMENSION(1:n_el_force) :: force

   CALL cf(env_id, pos, n_el_pos, e_pot, force, n_el_force, ierr)
END SUBROUTINE cp_calc_force

! **************************************************************************************************
!> \brief ...
!> \param input_file_path ...
!> \param output_file_path ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_run_input(input_file_path, output_file_path, ierr)
  USE cp2k_runs,                       ONLY: run_input
  USE input_cp2k,                      ONLY: create_cp2k_root_section
  USE input_section_types,             ONLY: section_type, section_release
   IMPLICIT NONE
   CHARACTER(len=*)                     :: input_file_path, output_file_path
   INTEGER                              :: ierr
   TYPE(section_type), POINTER          :: input_declaration

   NULLIFY (input_declaration)
   CALL create_cp2k_root_section(input_declaration)
   CALL run_input(input_declaration, input_file_path, output_file_path, ierr=ierr)
   CALL section_release(input_declaration)
END SUBROUTINE cp_run_input

! **************************************************************************************************
!> \brief ...
!> \param input_file_path ...
!> \param output_file_path ...
!> \param mpi_comm ...
!> \param ierr ...
! **************************************************************************************************
RECURSIVE SUBROUTINE cp_run_input_comm(input_file_path, output_file_path, &
                                       mpi_comm, ierr)
  USE cp2k_runs,                       ONLY: run_input
  USE input_cp2k,                      ONLY: create_cp2k_root_section
  USE input_section_types,             ONLY: section_type, section_release
   IMPLICIT NONE
   CHARACTER(len=*) :: input_file_path, output_file_path
   INTEGER :: mpi_comm, ierr
   TYPE(section_type), POINTER          :: input_declaration

   NULLIFY (input_declaration)
   CALL create_cp2k_root_section(input_declaration)
   CALL run_input(input_declaration, input_file_path, output_file_path, ierr, mpi_comm)
   CALL section_release(input_declaration)
END SUBROUTINE cp_run_input_comm

! **************************************************************************************************
!> \brief ...
!> \param f_env_id ...
!> \param dt ...
!> \param shake_tol ...
!> \param ierr ...
! **************************************************************************************************
SUBROUTINE cp_do_shake(f_env_id, dt, shake_tol, ierr)
  USE kinds, ONLY: dp
  USE f77_interface, ONLY: do_shake
   INTEGER :: f_env_id
   REAL(kind=dp) :: dt, shake_tol
   INTEGER :: ierr

   CALL do_shake(f_env_id, dt, shake_tol, ierr)
END SUBROUTINE cp_do_shake
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains types used for a Dimer Method calculations
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
MODULE dimer_methods
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE dimer_types,                     ONLY: dimer_env_type,&
                                              dimer_fixed_atom_control
   USE dimer_utils,                     ONLY: get_theta
   USE force_env_methods,               ONLY: force_env_calc_energy_force
   USE force_env_types,                 ONLY: force_env_get
   USE geo_opt,                         ONLY: cp_rot_opt
   USE gopt_f_types,                    ONLY: gopt_f_type
   USE input_constants,                 ONLY: do_first_rotation_step,&
                                              do_second_rotation_step,&
                                              do_third_rotation_step
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE motion_utils,                    ONLY: rot_ana,&
                                              thrs_motion
   USE particle_list_types,             ONLY: particle_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dimer_methods'

   PUBLIC :: cp_eval_at_ts

CONTAINS

! **************************************************************************************************
!> \brief Computes the dimer energy/gradients (including the rotation of the dimer)
!> \param gopt_env ...
!> \param x ...
!> \param f ...
!> \param gradient ...
!> \param calc_force ...
!> \date  01.2008
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino]
! **************************************************************************************************
   RECURSIVE SUBROUTINE cp_eval_at_ts(gopt_env, x, f, gradient, calc_force)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x
      REAL(KIND=dp), INTENT(OUT)                         :: f
      REAL(KIND=dp), DIMENSION(:), POINTER               :: gradient
      LOGICAL, INTENT(IN)                                :: calc_force

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_eval_at_ts', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      LOGICAL                                            :: eval_analytical
      REAL(KIND=dp)                                      :: angle1, angle2, f1, norm
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dimer_env_type), POINTER                      :: dimer_env
      TYPE(section_vals_type), POINTER                   :: print_section

      NULLIFY (dimer_env, logger, print_section)
      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(gopt_env)))CALL cp__a("motion/dimer_methods.F",79)
      dimer_env => gopt_env%dimer_env
      IF(.NOT.(ASSOCIATED(dimer_env)))CALL cp__a("motion/dimer_methods.F",81)
      iw = cp_print_key_unit_nr(logger, gopt_env%geo_section, "PRINT%PROGRAM_RUN_INFO", extension=".log")
      ! Possibly rotate Dimer or just compute Gradients of point 0 for Translation
      IF (gopt_env%dimer_rotation) THEN
         IF (debug_this_module .AND. (iw > 0)) THEN
            WRITE (iw, '(A)') "NVEC:"
            WRITE (iw, '(3F15.9)') dimer_env%nvec
         END IF
         SELECT CASE (dimer_env%rot%rotation_step)
         CASE (do_first_rotation_step, do_third_rotation_step)
            eval_analytical = .TRUE.
            IF ((dimer_env%rot%rotation_step == do_third_rotation_step) .AND. (dimer_env%rot%interpolate_gradient)) THEN
               eval_analytical = .FALSE.
            END IF
            IF (eval_analytical) THEN
               ! Compute energy, gradients and rotation vector for R1
               CALL cp_eval_at_ts_low(gopt_env, x, 1, dimer_env, calc_force, f1, dimer_env%rot%g1)
            ELSE
               angle1 = dimer_env%rot%angle1
               angle2 = dimer_env%rot%angle2
               dimer_env%rot%g1 = SIN(angle1-angle2)/SIN(angle1)*dimer_env%rot%g1+ &
                                  SIN(angle2)/SIN(angle1)*dimer_env%rot%g1p+ &
                                  (1.0_dp-COS(angle2)-SIN(angle2)*TAN(angle1/2.0_dp))*dimer_env%rot%g0
            END IF

            ! Determine the theta vector (i.e. the search direction for line minimization)
            gradient = -2.0_dp*(dimer_env%rot%g1-dimer_env%rot%g0)
            IF (debug_this_module .AND. (iw > 0)) THEN
               WRITE (iw, '(A)') "G1 vector:"
               WRITE (iw, '(3F15.9)') dimer_env%rot%g1
               WRITE (iw, '(A)') "-2*(G1-G0) vector:"
               WRITE (iw, '(3F15.9)') gradient
            END IF
            CALL get_theta(gradient, dimer_env, norm)
            f = norm
            dimer_env%cg_rot%norm_theta_old = dimer_env%cg_rot%norm_theta
            dimer_env%cg_rot%norm_theta = norm

            IF (debug_this_module .AND. (iw > 0)) THEN
               WRITE (iw, '(A,F20.10)') "Rotational Force step (1,3): module:", norm
               WRITE (iw, '(3F15.9)') gradient
            END IF

            ! Compute curvature and derivative of the curvature w.r.t. the rotational angle
            dimer_env%rot%curvature = DOT_PRODUCT(dimer_env%rot%g1-dimer_env%rot%g0, dimer_env%nvec)/dimer_env%dr
            dimer_env%rot%dCdp = 2.0_dp*DOT_PRODUCT(dimer_env%rot%g1-dimer_env%rot%g0, gradient)/dimer_env%dr

            dimer_env%rot%rotation_step = do_second_rotation_step
            gradient = -gradient
         CASE (do_second_rotation_step)
            ! Compute energy, gradients and rotation vector for R1
            CALL cp_eval_at_ts_low(gopt_env, x, 1, dimer_env, calc_force, f1, dimer_env%rot%g1p)
            dimer_env%rot%curvature = DOT_PRODUCT(dimer_env%rot%g1p-dimer_env%rot%g0, dimer_env%nvec)/dimer_env%dr
            dimer_env%rot%rotation_step = do_third_rotation_step

            ! Determine the theta vector (i.e. the search direction for line minimization)
            ! This is never used for getting a new theta but is consistent in order to
            ! give back the right value of f
            gradient = -2.0_dp*(dimer_env%rot%g1p-dimer_env%rot%g0)
            CALL get_theta(gradient, dimer_env, norm)
            f = norm

            IF (debug_this_module .AND. (iw > 0)) THEN
               WRITE (iw, '(A)') "Rotational Force step (1,3):"
               WRITE (iw, '(3F15.9)') gradient
            END IF
         END SELECT
      ELSE
         ! Compute energy, gradients and rotation vector for R0
         CALL cp_eval_at_ts_low(gopt_env, x, 0, dimer_env, calc_force, f, dimer_env%rot%g0)

         ! The dimer is rotated only when we are out of the translation line search
         IF (.NOT. gopt_env%do_line_search) THEN
            IF (debug_this_module .AND. (iw > 0)) THEN
               WRITE (iw, '(A)') "Entering the rotation module"
               WRITE (iw, '(A)') "G0 vector:"
               WRITE (iw, '(3F15.9)') dimer_env%rot%g0
            END IF
            CALL cp_rot_opt(gopt_env%gopt_dimer_env, x, gopt_env%gopt_dimer_param, &
                            gopt_env%gopt_dimer_env%geo_section)
            dimer_env%rot%rotation_step = do_first_rotation_step
         END IF

         print_section => section_vals_get_subs_vals(gopt_env%gopt_dimer_env%geo_section, "PRINT")
         ! Correcting gradients for Translation
         IF (dimer_env%rot%curvature > 0) THEN
            gradient = -DOT_PRODUCT(dimer_env%rot%g0, dimer_env%nvec)*dimer_env%nvec
            CALL remove_rot_transl_component(gopt_env, gradient, print_section)
         ELSE
            gradient = dimer_env%rot%g0-2.0_dp*DOT_PRODUCT(dimer_env%rot%g0, dimer_env%nvec)*dimer_env%nvec
            CALL remove_rot_transl_component(gopt_env, gradient, print_section)
         END IF
         IF (debug_this_module .AND. (iw > 0)) THEN
            WRITE (iw, *) "final gradient:", gradient
            WRITE (iw, '(A,F20.10)') "norm gradient:", SQRT(DOT_PRODUCT(gradient, gradient))
         END IF
         IF (.NOT. gopt_env%do_line_search) THEN
            f = SQRT(DOT_PRODUCT(gradient, gradient))
         ELSE
            f = -DOT_PRODUCT(gradient, dimer_env%tsl%tls_vec)
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, gopt_env%geo_section, "PRINT%PROGRAM_RUN_INFO")
      CALL timestop(handle)
   END SUBROUTINE cp_eval_at_ts

! **************************************************************************************************
!> \brief This function removes translational forces after project of the gradient
!> \param gopt_env ...
!> \param gradient ...
!> \param print_section ...
!> \par History
!>      2016/03/02 [LTong] added fixed atom constraint for gradient
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE remove_rot_transl_component(gopt_env, gradient, print_section)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:)                        :: gradient
      TYPE(section_vals_type), POINTER                   :: print_section

      CHARACTER(len=*), PARAMETER :: routineN = 'remove_rot_transl_component', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dof, handle, i, j, natoms
      REAL(KIND=dp)                                      :: norm, norm_gradient_old
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: D, mat
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles

      CALL timeset(routineN, handle)
      NULLIFY (mat)
      CALL force_env_get(gopt_env%force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, particles=particles)

      natoms = particles%n_els
      norm_gradient_old = SQRT(DOT_PRODUCT(gradient, gradient))
      IF (norm_gradient_old > 0.0_dp) THEN
         IF (natoms > 1) THEN
            CALL rot_ana(particles%els, mat, dof, print_section, keep_rotations=.FALSE., &
                         mass_weighted=.FALSE., natoms=natoms)

            ! Orthogonalize gradient with respect to the full set of Roto-Trasl vectors
            ALLOCATE (D(3*natoms, dof))
            ! Check First orthogonality in the first element of the basis set
            DO i = 1, dof
               D(:, i) = mat(:, i)
               DO j = i+1, dof
                  norm = DOT_PRODUCT(mat(:, i), mat(:, j))
                  IF(.NOT.(ABS(norm) < thrs_motion))CALL cp__a("motion/dimer_methods.F",229)
               END DO
            END DO
            DO i = 1, dof
               norm = DOT_PRODUCT(gradient, D(:, i))
               gradient = gradient-norm*D(:, i)
            END DO
            DEALLOCATE (D)
            DEALLOCATE (mat)
         END IF
      END IF
      ! apply constraint
      CALL dimer_fixed_atom_control(gradient, subsys)
      CALL timestop(handle)
   END SUBROUTINE remove_rot_transl_component

! **************************************************************************************************
!> \brief ...
!> \param gopt_env ...
!> \param x ...
!> \param dimer_index ...
!> \param dimer_env ...
!> \param calc_force ...
!> \param f ...
!> \param gradient ...
!> \par History
!>      none
!> \author Luca Bellucci and Teodoro Laino - created [tlaino] - 01.2008
! **************************************************************************************************
   SUBROUTINE cp_eval_at_ts_low(gopt_env, x, dimer_index, dimer_env, calc_force, &
                                f, gradient)
      TYPE(gopt_f_type), POINTER                         :: gopt_env
      REAL(KIND=dp), DIMENSION(:), POINTER               :: x
      INTEGER, INTENT(IN)                                :: dimer_index
      TYPE(dimer_env_type), POINTER                      :: dimer_env
      LOGICAL, INTENT(IN)                                :: calc_force
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: f
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: gradient

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_eval_at_ts_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idg, idir, ip
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(particle_list_type), POINTER                  :: particles

      CALL timeset(routineN, handle)
      idg = 0
      CALL force_env_get(gopt_env%force_env, subsys=subsys)
      CALL cp_subsys_get(subsys, particles=particles)
      DO ip = 1, particles%n_els
         DO idir = 1, 3
            idg = idg+1
            particles%els(ip)%r(idir) = x(idg)+REAL(dimer_index, KIND=dp)*dimer_env%nvec(idg)*dimer_env%dr
         END DO
      END DO

      ! Compute energy and forces
      CALL force_env_calc_energy_force(gopt_env%force_env, calc_force=calc_force)

      ! Possibly take the potential energy
      IF (PRESENT(f)) THEN
         CALL force_env_get(gopt_env%force_env, potential_energy=f)
      END IF

      ! Possibly take the gradients
      IF (PRESENT(gradient)) THEN
         idg = 0
         CALL cp_subsys_get(subsys, particles=particles)
         DO ip = 1, particles%n_els
            DO idir = 1, 3
               idg = idg+1
               IF(.NOT.(SIZE(gradient) >= idg))CALL cp__a("motion/dimer_methods.F",301)
               gradient(idg) = -particles%els(ip)%f(idir)
            END DO
         END DO
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_eval_at_ts_low

END MODULE dimer_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief evaluete the potential energy and its gradients using an array
!>      with same dimension as the particle_set
!> \param gopt_env the geometry optimization environment
!> \param x the position where the function should be evaluated
!> \param f the function value
!> \param gradient the value of its gradient
!> \param master ...
!> \param final_evaluation ...
!> \param para_env ...
!> \par History
!>       CELL OPTIMIZATION:  Teodoro Laino [tlaino] - University of Zurich - 03.2008
!>
!> \author Teodoro Laino [tlaino] - University of Zurich - 01.2008
! **************************************************************************************************
RECURSIVE SUBROUTINE cp_eval_at(gopt_env, x, f, gradient, master, &
                                final_evaluation, para_env)

USE cp_log_handling, ONLY: cp_logger_type
    USE averages_types,                  ONLY: average_quantities_type,&
                                               create_averages,&
                                               release_averages
    USE cell_opt_utils,                  ONLY: get_dg_dh,&
                                               gopt_new_logger_create,&
                                               gopt_new_logger_release
    USE cell_types,                      ONLY: cell_type
    USE cell_methods,                    ONLY: write_cell
    USE cp_para_types,                   ONLY: cp_para_env_type
    USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                               cp_subsys_type,&
                                               pack_subsys_particles,&
                                               unpack_subsys_particles
    USE dimer_methods,                   ONLY: cp_eval_at_ts
    USE force_env_methods,               ONLY: force_env_calc_energy_force
    USE force_env_types,                 ONLY: force_env_get,&
                                               force_env_get_nparticle
    USE geo_opt,                         ONLY: cp_geo_opt
    USE gopt_f_types,                    ONLY: gopt_f_type
    USE gopt_f_methods,                  ONLY: apply_cell_change
    USE input_constants,                 ONLY: default_minimization_method_id,&
                                               default_ts_method_id,&
                                               default_cell_direct_id,&
                                               default_cell_method_id,&
                                               default_cell_geo_opt_id,&
                                               default_cell_md_id,&
                                               default_shellcore_method_id,&
                                               nvt_ensemble,&
                                               mol_dyn_run,&
                                               geo_opt_run,&
                                               cell_opt_run
    USE input_section_types,             ONLY: section_vals_get,&
                                               section_vals_get_subs_vals,&
                                               section_vals_type,&
                                               section_vals_val_get
    USE md_run,                          ONLY: qs_mol_dyn
    USE message_passing,                 ONLY: mp_bcast
    USE kinds,                           ONLY: dp,&
                                               default_string_length
    USE particle_list_types,             ONLY: particle_list_type
    USE particle_methods,                ONLY: write_structure_data
    USE virial_methods,                  ONLY: virial_update
    USE virial_types,                    ONLY: cp_virial,&
                                               virial_create,&
                                               virial_release,&
                                               virial_type
    USE cp_log_handling,                 ONLY: cp_add_default_logger,&
                                               cp_rm_default_logger

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(gopt_f_type), POINTER               :: gopt_env
   REAL(KIND=dp), DIMENSION(:), POINTER     :: x
   REAL(KIND=dp), INTENT(OUT), OPTIONAL     :: f
   REAL(KIND=dp), DIMENSION(:), OPTIONAL, &
      POINTER                                :: gradient
   INTEGER, INTENT(IN)                      :: master
   LOGICAL, INTENT(IN), OPTIONAL            :: final_evaluation
   TYPE(cp_para_env_type), POINTER          :: para_env

   CHARACTER(len=*), PARAMETER :: routineN = 'cp_eval_at', moduleN = 'gopt_f77_methods', &
                                  routineP = moduleN//':'//routineN

   INTEGER                                  :: ensemble, handle, idg, idir, ip, &
                                               nparticle, nsize, shell_index
   LOGICAL                                  :: explicit, my_final_evaluation
   REAL(KIND=dp)                            :: f_ts, potential_energy
   REAL(KIND=dp), DIMENSION(3, 3)            :: av_ptens
   REAL(KIND=dp), DIMENSION(:), POINTER     :: cell_gradient, gradient_ts
   TYPE(cell_type), POINTER                 :: cell
   TYPE(cp_subsys_type), POINTER            :: subsys
   TYPE(particle_list_type), POINTER        :: core_particles, particles, &
                                               shell_particles
   TYPE(virial_type), POINTER               :: virial, virial_avg
   TYPE(cp_logger_type), POINTER            :: new_logger
   CHARACTER(LEN=default_string_length)     :: project_name
   TYPE(average_quantities_type), POINTER   :: averages
   TYPE(section_vals_type), POINTER         :: work, avgs_section

   NULLIFY (averages)
   NULLIFY (cell)
   NULLIFY (core_particles)
   NULLIFY (gradient_ts)
   NULLIFY (particles)
   NULLIFY (shell_particles)
   NULLIFY (subsys)
   NULLIFY (virial)
   NULLIFY (new_logger)

   CALL timeset(routineN, handle)

   CALL force_env_get(gopt_env%force_env, subsys=subsys, cell=cell)
   CALL cp_subsys_get(subsys, &
                      core_particles=core_particles, &
                      particles=particles, &
                      shell_particles=shell_particles, &
                      virial=virial)

   my_final_evaluation = .FALSE.
   IF (PRESENT(final_evaluation)) my_final_evaluation = final_evaluation

   SELECT CASE (gopt_env%type_id)
   CASE (default_minimization_method_id, default_ts_method_id)
      CALL unpack_subsys_particles(subsys=subsys, r=x)
      CALL write_structure_data(particles%els, cell, gopt_env%motion_section)
      SELECT CASE (gopt_env%type_id)
      CASE (default_minimization_method_id)
         ! Geometry Minimization
         CALL force_env_calc_energy_force(gopt_env%force_env, &
                                          calc_force=PRESENT(gradient), &
                                          require_consistent_energy_force=gopt_env%require_consistent_energy_force)
         ! Possibly take the potential energy
         IF (PRESENT(f)) THEN
            CALL force_env_get(gopt_env%force_env, potential_energy=f)
         END IF
         ! Possibly take the gradients
         IF (PRESENT(gradient)) THEN
            IF (master == para_env%mepos) THEN ! we are on the master
               CALL pack_subsys_particles(subsys=subsys, f=gradient, fscale=-1.0_dp)
            END IF
         END IF
      CASE (default_ts_method_id)
         ! Transition State Optimization
         ALLOCATE (gradient_ts(particles%n_els*3))
         ! Real calculation of energy and forces for transition state optimization:
         ! When doing dimer methods forces have to be always computed since the function
         ! to minimize is not the energy but the effective force
         CALL cp_eval_at_ts(gopt_env, x, f_ts, gradient_ts, calc_force=.TRUE.)
         ! Possibly take the potential energy
         IF (PRESENT(f)) f = f_ts
         ! Possibly take the gradients
         IF (PRESENT(gradient)) THEN
            IF (master == para_env%mepos) THEN ! we are on the master
               IF(.NOT.(ASSOCIATED(gradient)))CALL cp__a("motion/gopt_f77_methods.F",158)
               gradient = gradient_ts
            END IF
         END IF
         DEALLOCATE (gradient_ts)
      END SELECT
      ! This call is necessary for QM/MM if a Translation is applied
      ! this makes the geometry optimizer consistent
      CALL unpack_subsys_particles(subsys=subsys, r=x)
   CASE (default_cell_method_id)
      ! Check for VIRIAL
      IF (.NOT. virial%pv_availability) &
         CALL cp_abort(cp__l("motion/gopt_f77_methods.F",170), &
                       "Cell optimization requested but FORCE_EVAL%STRESS_TENSOR was not defined! "// &
                       "Activate the evaluation of the stress tensor for cell optimization!")
      SELECT CASE (gopt_env%cell_method_id)
      CASE (default_cell_direct_id)
         CALL apply_cell_change(gopt_env, cell, x, update_forces=.FALSE.)
         ! Possibly output the new cell used for the next calculation
         CALL write_cell(cell, gopt_env%geo_section)
         ! Compute the pressure tensor
         CALL virial_create(virial_avg)
         CALL force_env_calc_energy_force(gopt_env%force_env, &
                                          calc_force=PRESENT(gradient), &
                                          require_consistent_energy_force=gopt_env%require_consistent_energy_force)
         ! Possibly take the potential energy
         CALL force_env_get(gopt_env%force_env, potential_energy=potential_energy)
         CALL cp_virial(virial, virial_avg)
         CALL virial_update(virial_avg, subsys, para_env)
         IF (PRESENT(f)) THEN
            CALL force_env_get(gopt_env%force_env, potential_energy=f)
         END IF
         ! Possibly take the gradients
         IF (PRESENT(gradient)) THEN
            IF(.NOT.(ANY(virial_avg%pv_total /= 0)))CALL cp__a("motion/gopt_f77_methods.F",192)
            ! Convert the average ptens
            av_ptens(:, :) = virial_avg%pv_total(:, :)/cell%deth
            IF (master == para_env%mepos) THEN ! we are on the master
               IF(.NOT.(ASSOCIATED(gradient)))CALL cp__a("motion/gopt_f77_methods.F",196)
               nparticle = force_env_get_nparticle(gopt_env%force_env)
               nsize = 3*nparticle
               IF(.NOT.((SIZE(gradient) == nsize+6)))CALL cp__a("motion/gopt_f77_methods.F",199)
               CALL pack_subsys_particles(subsys=subsys, f=gradient(1:nsize), fscale=-1.0_dp)
               CALL apply_cell_change(gopt_env, cell, gradient, update_forces=.TRUE.)
               cell_gradient => gradient(nsize+1:nsize+6)
               cell_gradient = 0.0_dp
               CALL get_dg_dh(cell_gradient, av_ptens, gopt_env%cell_env%pres_ext, cell, gopt_env%cell_env%mtrx, &
                              keep_angles=gopt_env%cell_env%keep_angles, &
                              keep_symmetry=gopt_env%cell_env%keep_symmetry, &
                              pres_int=gopt_env%cell_env%pres_int)
            END IF
            ! some callers expect pres_int to be available on all ranks. Also, here master is not necessarily a single rank.
            ! Assume at least master==0
            CALL mp_bcast(gopt_env%cell_env%pres_int, 0, para_env%group)
         END IF
         CALL virial_release(virial_avg)
      CASE (default_cell_geo_opt_id, default_cell_md_id)
         CALL apply_cell_change(gopt_env, cell, x, update_forces=.FALSE.)
         ! Possibly output the new cell used for the next calculation
         CALL write_cell(cell, gopt_env%geo_section)
         ! Compute the pressure tensor
         CALL virial_create(virial_avg)
         IF (my_final_evaluation) THEN
            CALL force_env_calc_energy_force(gopt_env%force_env, &
                                             calc_force=PRESENT(gradient), &
                                             require_consistent_energy_force=gopt_env%require_consistent_energy_force)
            IF (PRESENT(f)) THEN
               CALL force_env_get(gopt_env%force_env, potential_energy=f)
            END IF
         ELSE
            SELECT CASE (gopt_env%cell_method_id)
            CASE (default_cell_geo_opt_id)
               work => section_vals_get_subs_vals(gopt_env%motion_section, "GEO_OPT")
               CALL section_vals_get(work, explicit=explicit)
               IF (.NOT. explicit) &
                  CALL cp_abort(cp__l("motion/gopt_f77_methods.F",233), &
                                "Cell optimization at 0K was requested. GEO_OPT section MUST be provided in the input file!")
               ! Perform a geometry optimization
               CALL gopt_new_logger_create(new_logger, gopt_env%force_env%root_section, para_env, &
                                           project_name, id_run=geo_opt_run)
               CALL cp_add_default_logger(new_logger)
               CALL cp_geo_opt(gopt_env%force_env, gopt_env%globenv, eval_opt_geo=.FALSE.)
               CALL force_env_get(gopt_env%force_env, potential_energy=potential_energy)
               CALL cp_virial(virial, virial_avg)
            CASE (default_cell_md_id)
               work => section_vals_get_subs_vals(gopt_env%motion_section, "MD")
               avgs_section => section_vals_get_subs_vals(work, "AVERAGES")
               CALL section_vals_get(work, explicit=explicit)
               IF (.NOT. explicit) &
                  CALL cp_abort( &
                  cp__l("motion/gopt_f77_methods.F",248), &
                  "Cell optimization at finite temperature was requested. MD section MUST be provided in the input file!")
               ! Only NVT ensemble is allowed..
               CALL section_vals_val_get(gopt_env%motion_section, "MD%ENSEMBLE", i_val=ensemble)
               IF (ensemble /= nvt_ensemble) &
                  CALL cp_abort(cp__l("motion/gopt_f77_methods.F",253), &
                                "Cell optimization at finite temperature requires the NVT MD ensemble!")
               ! Perform a molecular dynamics
               CALL gopt_new_logger_create(new_logger, gopt_env%force_env%root_section, para_env, &
                                           project_name, id_run=mol_dyn_run)
               CALL cp_add_default_logger(new_logger)
               CALL create_averages(averages, avgs_section, virial_avg=.TRUE., force_env=gopt_env%force_env)
               CALL qs_mol_dyn(gopt_env%force_env, gopt_env%globenv, averages, rm_restart_info=.FALSE.)
               ! Retrieve the average of the stress tensor and the average of the potential energy
               potential_energy = averages%avepot
               CALL cp_virial(averages%virial, virial_avg)
               CALL release_averages(averages)
            CASE DEFAULT
               CALL cp__b("motion/gopt_f77_methods.F",266,"")
            END SELECT
            CALL cp_rm_default_logger()
            CALL gopt_new_logger_release(new_logger, gopt_env%force_env%root_section, para_env, project_name, &
                                         cell_opt_run)
            ! Update the virial
            CALL virial_update(virial_avg, subsys, para_env)
            ! Possibly take give back the potential energy
            IF (PRESENT(f)) THEN
               f = potential_energy
            END IF
         END IF
         ! Possibly give back the gradients
         IF (PRESENT(gradient)) THEN
            IF(.NOT.(ANY(virial_avg%pv_total /= 0)))CALL cp__a("motion/gopt_f77_methods.F",280)
            ! Convert the average ptens
            av_ptens(:, :) = virial_avg%pv_total(:, :)/cell%deth
            IF (master == para_env%mepos) THEN ! we are on the master
               IF(.NOT.(ASSOCIATED(gradient)))CALL cp__a("motion/gopt_f77_methods.F",284)
               ! Compute the gradients on the cell
               CALL get_dg_dh(gradient, av_ptens, gopt_env%cell_env%pres_ext, cell, gopt_env%cell_env%mtrx, &
                              keep_angles=gopt_env%cell_env%keep_angles, &
                              keep_symmetry=gopt_env%cell_env%keep_symmetry, &
                              pres_int=gopt_env%cell_env%pres_int)
            END IF
            ! some callers expect pres_int to be available on all ranks. Also, here master is not necessarily a single rank.
            ! Assume at least master==0
            CALL mp_bcast(gopt_env%cell_env%pres_int, 0, para_env%group)
         END IF
         CALL virial_release(virial_avg)
      CASE DEFAULT
         CALL cp__b("motion/gopt_f77_methods.F",297,"")
      END SELECT
   CASE (default_shellcore_method_id)
      idg = 0
      DO ip = 1, particles%n_els
         shell_index = particles%els(ip)%shell_index
         IF (shell_index /= 0) THEN
            DO idir = 1, 3
               idg = 3*(shell_index-1)+idir
               shell_particles%els(shell_index)%r(idir) = core_particles%els(ip)%r(idir)-x(idg)
            END DO
         END IF
      END DO
      CALL write_structure_data(particles%els, cell, gopt_env%motion_section)

      ! Shell-core optimization
      CALL force_env_calc_energy_force(gopt_env%force_env, &
                                       calc_force=PRESENT(gradient), &
                                       require_consistent_energy_force=gopt_env%require_consistent_energy_force)

      ! Possibly take the potential energy
      IF (PRESENT(f)) THEN
         CALL force_env_get(gopt_env%force_env, potential_energy=f)
      END IF

      ! Possibly take the gradients
      IF (PRESENT(gradient)) THEN
         IF (master == para_env%mepos) THEN ! we are on the master
            IF(.NOT.(ASSOCIATED(gradient)))CALL cp__a("motion/gopt_f77_methods.F",325)
            idg = 0
            DO ip = 1, shell_particles%n_els
               DO idir = 1, 3
                  idg = idg+1
                  gradient(idg) = -(core_particles%els(ip)%f(idir)-shell_particles%els(ip)%f(idir))
               END DO
            END DO
         END IF
      END IF
   CASE DEFAULT
      CALL cp__b("motion/gopt_f77_methods.F",336,"")
   END SELECT

   CALL timestop(handle)

END SUBROUTINE cp_eval_at
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief utillities for the TMC
!> \par History
!>      11.2012 created [Mandes Schoenherr]
!> \author Mandes
! **************************************************************************************************

MODULE tmc_utils

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'tmc_utils'

END MODULE tmc_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief The types needed for the calculation of modified atomic orbitals (MAO)
!> \par History
!>      03.2016 created [JGH]
!> \author JGH
! **************************************************************************************************
MODULE mao_types

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mao_types'

   PUBLIC :: mao_type
   PUBLIC :: create_mao_type, release_mao_type

! **************************************************************************************************
!> \brief Quantities needed for MAO's
!> \author JGH
! **************************************************************************************************
   TYPE mao_type
      INTEGER                       :: ref_basis
   END TYPE mao_type

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param mao_env ...
! **************************************************************************************************
   SUBROUTINE create_mao_type(mao_env)
      TYPE(mao_type), POINTER                            :: mao_env

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mao_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mao_env)) THEN
         CALL release_mao_type(mao_env)
      END IF

      ALLOCATE (mao_env)

   END SUBROUTINE create_mao_type

! **************************************************************************************************
!> \brief ...
!> \param mao_env ...
! **************************************************************************************************
   SUBROUTINE release_mao_type(mao_env)
      TYPE(mao_type), POINTER                            :: mao_env

      CHARACTER(len=*), PARAMETER :: routineN = 'release_mao_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mao_env)) THEN
         DEALLOCATE (mao_env)
      END IF

   END SUBROUTINE release_mao_type

! **************************************************************************************************

END MODULE mao_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Debug the derivatives of the the rotational matrices
!>
!> \param sepi ...
!> \param sepj ...
!> \param rjiv ...
!> \param ij_matrix ...
!> \param do_invert ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
SUBROUTINE check_rotmat_der(sepi, sepj, rjiv, ij_matrix, do_invert)

  USE kinds,                           ONLY: dp
  USE semi_empirical_int_utils,        ONLY: rotmat
  USE semi_empirical_types,            ONLY: rotmat_create,&
                                             rotmat_release,&
                                             rotmat_type,&
                                             semi_empirical_type,&
                                             se_int_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(semi_empirical_type), POINTER       :: sepi, sepj
   REAL(KIND=dp), DIMENSION(3), INTENT(IN)  :: rjiv
   TYPE(rotmat_type), POINTER               :: ij_matrix
   LOGICAL, INTENT(IN)                      :: do_invert

   CHARACTER(len=*), PARAMETER :: moduleN = 'semi_empirical_int_debug', &
                                  routineN = 'check_rotmat_der', routineP = moduleN//':'//routineN

   LOGICAL                                  :: check_value
   REAL(KIND=dp)                            :: dx, r, r0(3), x(3)
   TYPE(rotmat_type), POINTER               :: matrix, matrix_m, matrix_n, &
                                               matrix_p
   INTEGER                                  :: imap(3), i, j, k, l

   NULLIFY (matrix_m, matrix_p, matrix_n, matrix)
   CALL rotmat_create(matrix_p)
   CALL rotmat_create(matrix_m)
   CALL rotmat_create(matrix_n)
   dx = 1.0E-6_dp
   imap(1) = 1
   imap(2) = 2
   imap(3) = 3
   IF (do_invert) THEN
      imap(1) = 3
      imap(2) = 2
      imap(3) = 1
   END IF
   ! Check derivatives: analytical VS numerical
   WRITE (*, *) "DEBUG::"//routineP
   DO j = 1, 3
      x = 0.0_dp
      x(imap(j)) = dx
      DO i = 1, 2
         IF (i == 1) matrix => matrix_p
         IF (i == 2) matrix => matrix_m
         r0 = rjiv+(-1.0_dp)**(i-1)*x
         r = SQRT(DOT_PRODUCT(r0, r0))
         CALL rotmat(sepi, sepj, r0, r, matrix, do_derivatives=.FALSE., debug_invert=do_invert)
      END DO
      ! SP
      matrix_n%sp_d(j, :, :) = (matrix_p%sp-matrix_m%sp)/(2.0_dp*dx)
      DO i = 1, 3
         DO k = 1, 3
            IF (.NOT. check_value(matrix_n%sp_d(j, k, i), ij_matrix%sp_d(j, k, i), dx, 0.1_dp)) THEN
               WRITE (*, *) "ERROR for SP rotation matrix derivative SP(j,k,i), j,k,i::", j, k, i
               CALL cp__b("semi_empirical_int_debug.F",73,"")
            END IF
         END DO
      END DO
      ! PP
      matrix_n%pp_d(j, :, :, :) = (matrix_p%pp-matrix_m%pp)/(2.0_dp*dx)
      DO i = 1, 3
         DO k = 1, 3
            DO l = 1, 6
               IF (.NOT. check_value(matrix_n%pp_d(j, l, k, i), ij_matrix%pp_d(j, l, k, i), dx, 0.1_dp)) THEN
                  WRITE (*, *) "ERROR for PP rotation matrix derivative PP(j,l,k,i), j,l,k,i::", j, l, k, i
                  CALL cp__b("semi_empirical_int_debug.F",84,"")
               END IF
            END DO
         END DO
      END DO
      ! d-orbitals debug
      IF (sepi%dorb .OR. sepj%dorb) THEN
         ! SD
         matrix_n%sd_d(j, :, :) = (matrix_p%sd-matrix_m%sd)/(2.0_dp*dx)
         DO i = 1, 5
            DO k = 1, 5
               IF (.NOT. check_value(matrix_n%sd_d(j, k, i), ij_matrix%sd_d(j, k, i), dx, 0.1_dp)) THEN
                  WRITE (*, *) "ERROR for SD rotation matrix derivative SD(j,k,i), j,k,i::", j, k, i
                  CALL cp__b("semi_empirical_int_debug.F",97,"")
               END IF
            END DO
         END DO
         ! DP
         matrix_n%pd_d(j, :, :, :) = (matrix_p%pd-matrix_m%pd)/(2.0_dp*dx)
         DO i = 1, 3
            DO k = 1, 5
               DO l = 1, 15
                  IF (.NOT. check_value(matrix_n%pd_d(j, l, k, i), ij_matrix%pd_d(j, l, k, i), dx, 0.1_dp)) THEN
                     WRITE (*, *) "ERROR for DP rotation matrix derivative DP(j,l,k,i), j,l,k,i::", j, l, k, i
                     CALL cp__b("semi_empirical_int_debug.F",108,"")
                  END IF
               END DO
            END DO
         END DO
         ! DD
         matrix_n%dd_d(j, :, :, :) = (matrix_p%dd-matrix_m%dd)/(2.0_dp*dx)
         DO i = 1, 5
            DO k = 1, 5
               DO l = 1, 15
                  IF (.NOT. check_value(matrix_n%dd_d(j, l, k, i), ij_matrix%dd_d(j, l, k, i), dx, 0.1_dp)) THEN
                     WRITE (*, *) "ERROR for DD rotation matrix derivative DD(j,l,k,i), j,l,k,i::", j, l, k, i
                     CALL cp__b("semi_empirical_int_debug.F",120,"")
                  END IF
               END DO
            END DO
         END DO
      END IF
   END DO
   CALL rotmat_release(matrix_p)
   CALL rotmat_release(matrix_m)
   CALL rotmat_release(matrix_n)
END SUBROUTINE check_rotmat_der

! **************************************************************************************************
!> \brief Check Numerical Vs Analytical
!> \param sepi ...
!> \param sepj ...
!> \param rijv ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param invert ...
!> \param ii ...
!> \param kk ...
!> \param v_d ...
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
!> \note
!>      Debug routine
! **************************************************************************************************
SUBROUTINE rot_2el_2c_first_debug(sepi, sepj, rijv, se_int_control, se_taper, invert, ii, kk, v_d)

  USE kinds,                           ONLY: dp
  USE semi_empirical_int_utils,        ONLY: rotmat
  USE semi_empirical_int_arrays,       ONLY: indexb
  USE semi_empirical_int_num,          ONLY: terep_num
  USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                             rotmat_type,&
                                             rotmat_create,&
                                             rotmat_release,&
                                             se_int_control_type,&
                                             se_taper_type
  USE semi_empirical_int_utils,        ONLY: rot_2el_2c_first
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(semi_empirical_type), POINTER       :: sepi, sepj
   REAL(KIND=dp), DIMENSION(3), INTENT(IN)  :: rijv
   TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
   LOGICAL, INTENT(IN)                      :: invert
   TYPE(se_taper_type), POINTER             :: se_taper
   INTEGER, INTENT(IN)                      :: ii, kk
   REAL(KIND=dp), DIMENSION(3, 45, 45), &
      INTENT(IN)                             :: v_d

   CHARACTER(len=*), PARAMETER :: moduleN = 'semi_empirical_int_debug', &
                                  routineN = 'rot_2el_2c_first', routineP = moduleN//':'//routineN

   REAL(KIND=dp), DIMENSION(491)            :: rep
   LOGICAL, DIMENSION(45, 45)               :: logv
   REAL(KIND=dp), DIMENSION(45, 45)         :: v_n, v_p, v_m

   LOGICAL                                  :: check_value
   REAL(KIND=dp)                            :: dx, r, r0(3), x(3)
   TYPE(rotmat_type), POINTER               :: matrix
   INTEGER                                  :: imap(3), i, j, k, limkl

   NULLIFY (matrix)
   dx = 1.0E-6_dp
   imap(1) = 1
   imap(2) = 2
   imap(3) = 3
   IF (invert) THEN
      imap(1) = 3
      imap(2) = 2
      imap(3) = 1
   END IF
   limkl = indexb(kk, kk)
   ! Check derivatives: analytical VS numerical
   WRITE (*, *) "DEBUG::"//routineP
   DO j = 1, 3
      x = 0.0_dp
      x(imap(j)) = dx
      DO i = 1, 2
         r0 = rijv+(-1.0_dp)**(i-1)*x
         r = SQRT(DOT_PRODUCT(r0, r0))

         CALL rotmat_create(matrix)
         CALL rotmat(sepi, sepj, r0, r, matrix, do_derivatives=.FALSE., debug_invert=invert)

         ! Compute integrals in diatomic frame
         CALL terep_num(sepi, sepj, r, rep, se_taper=se_taper, se_int_control=se_int_control)

         IF (i == 1) THEN
            CALL rot_2el_2c_first(sepi, sepj, r0, se_int_control, se_taper, invert, ii, kk, rep, logv, matrix, &
                                  v_p, lgrad=.FALSE.)
         END IF
         IF (i == 2) THEN
            CALL rot_2el_2c_first(sepi, sepj, r0, se_int_control, se_taper, invert, ii, kk, rep, logv, matrix, &
                                  v_m, lgrad=.FALSE.)
         END IF
         CALL rotmat_release(matrix)
      END DO
      ! Check numerical VS analytical
      DO i = 1, 45
         DO k = 1, limkl
            ! Compute the numerical derivative
            v_n(i, k) = (v_p(i, k)-v_m(i, k))/(2.0_dp*dx)
            IF (.NOT. check_value(v_d(j, i, k), v_n(i, k), dx, 0.1_dp)) THEN
               WRITE (*, *) "ERROR for  rot_2el_2c_first derivative V_D(j,i,k), j,i,k::", j, i, k
               CALL cp__b("semi_empirical_int_debug.F",228,"")
            END IF
         END DO
      END DO
   END DO
END SUBROUTINE rot_2el_2c_first_debug

! **************************************************************************************************
!> \brief Check Numerical Vs Analytical ssss
!> \param sepi ...
!> \param sepj ...
!> \param r ...
!> \param dssss ...
!> \param itype ...
!> \param se_int_control ...
!> \param se_taper ...
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
!> \note
!>      Debug routine
! **************************************************************************************************
SUBROUTINE check_dssss_nucint_ana(sepi, sepj, r, dssss, itype, se_int_control, se_taper)

  USE kinds,                           ONLY: dp
  USE semi_empirical_int_num,          ONLY: ssss_nucint_num
  USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                             se_int_control_type,&
                                             se_taper_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(semi_empirical_type), POINTER       :: sepi, sepj
   REAL(dp), INTENT(IN)                     :: r
   REAL(dp), INTENT(IN)                     :: dssss
   INTEGER, INTENT(IN)                      :: itype
   TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
   TYPE(se_taper_type), POINTER                :: se_taper

   CHARACTER(len=*), PARAMETER :: moduleN = 'semi_empirical_int_debug', &
                                  routineN = 'check_dssss_nucint_ana', routineP = moduleN//':'//routineN

   REAL(dp)                                 :: delta, nssss, od, rn, ssssm, ssssp
   LOGICAL                                  :: check_value

   delta = 1.0E-8_dp
   od = 0.5_dp/delta
   rn = r+delta
   CALL ssss_nucint_num(sepi, sepj, rn, ssssp, itype, se_taper, se_int_control)
   rn = r-delta
   CALL ssss_nucint_num(sepi, sepj, rn, ssssm, itype, se_taper, se_int_control)
   nssss = od*(ssssp-ssssm)
   ! check
   WRITE (*, *) "DEBUG::"//routineP
   IF (.NOT. check_value(nssss, dssss, delta, 0.1_dp)) THEN
      WRITE (*, *) "ERROR for SSSS derivative SSSS"
      CALL cp__b("semi_empirical_int_debug.F",283,"")
   END IF
END SUBROUTINE check_dssss_nucint_ana

! **************************************************************************************************
!> \brief Check Numerical Vs Analytical core
!> \param sepi ...
!> \param sepj ...
!> \param r ...
!> \param dcore ...
!> \param itype ...
!> \param se_int_control ...
!> \param se_taper ...
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
!> \note
!>      Debug routine
! **************************************************************************************************
SUBROUTINE check_dcore_nucint_ana(sepi, sepj, r, dcore, itype, se_int_control, se_taper)

  USE kinds,                           ONLY: dp
  USE semi_empirical_int_num,          ONLY: core_nucint_num
  USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                             se_int_control_type,&
                                             se_taper_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(semi_empirical_type), POINTER       :: sepi, sepj
   REAL(dp), INTENT(IN)                     :: r
   REAL(dp), DIMENSION(10, 2), INTENT(IN)   :: dcore
   INTEGER, INTENT(IN)                      :: itype
   TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
   TYPE(se_taper_type), POINTER             :: se_taper

   CHARACTER(len=*), PARAMETER :: moduleN = 'semi_empirical_int_debug', &
                                  routineN = 'check_dcore_nucint_ana', routineP = moduleN//':'//routineN

   INTEGER                                  :: i, j
   REAL(dp)                                 :: delta, od, rn
   REAL(dp), DIMENSION(10, 2)               :: corem, corep, ncore
   LOGICAL                                  :: check_value

   delta = 1.0E-8_dp
   od = 0.5_dp/delta
   rn = r+delta
   CALL core_nucint_num(sepi, sepj, rn, corep, itype, se_taper, se_int_control)
   rn = r-delta
   CALL core_nucint_num(sepi, sepj, rn, corem, itype, se_taper, se_int_control)
   ncore = od*(corep-corem)
   ! check
   WRITE (*, *) "DEBUG::"//routineP
   DO i = 1, 2
      DO j = 1, 10
         IF (.NOT. check_value(ncore(j, i), dcore(j, i), delta, 0.1_dp)) THEN
            WRITE (*, *) "ERROR for CORE derivative CORE(j,i), j,i::", j, i
            CALL cp__b("semi_empirical_int_debug.F",339,"")
         END IF
      END DO
   END DO
END SUBROUTINE check_dcore_nucint_ana

! **************************************************************************************************
!> \brief Low level comparison function between numberical and analytical values
!> \param num ...
!> \param ana ...
!> \param minval ...
!> \param thrs ...
!> \retval passed ...
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
!> \note
!>      Debug routine
! **************************************************************************************************
FUNCTION check_value(num, ana, minval, thrs) RESULT(passed)
  USE kinds,                           ONLY: dp
   IMPLICIT NONE
   REAL(KIND=dp)                            :: num, ana, thrs, minval
   LOGICAL                                  :: passed

   passed = .TRUE.

   IF ((ABS(num) < minval) .AND. (ABS(ana) > minval)) THEN
      WRITE (*, *) "WARNING ---> ", num, ana, thrs
   ELSE IF ((ABS(num) > minval) .AND. (ABS(ana) < minval)) THEN
      WRITE (*, *) "WARNING ---> ", num, ana, thrs
   ELSE IF ((ABS(num) < minval) .AND. (ABS(ana) < minval)) THEN
      ! skip..
      RETURN
   END IF
   IF (ABS((num-ana)/num*100._dp) > thrs) THEN
      WRITE (*, *) ABS(num-ana)/num*100._dp, thrs
      passed = .FALSE.
   END IF
   IF (.NOT. passed) THEN
      WRITE (*, *) "ANALYTICAL ::", ana
      WRITE (*, *) "NUMERICAL  ::", num
   END IF
END FUNCTION check_value

! **************************************************************************************************
!> \brief Check Numerical Vs Analytical
!> \param sepi ...
!> \param sepj ...
!> \param rijv ...
!> \param itype ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param e1b ...
!> \param e2a ...
!> \param de1b ...
!> \param de2a ...
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
!> \note
!>      Debug routine
! **************************************************************************************************
SUBROUTINE check_drotnuc_ana(sepi, sepj, rijv, itype, se_int_control, se_taper, e1b, e2a, de1b, de2a)

  USE kinds,                           ONLY: dp
  USE semi_empirical_int_num,          ONLY: rotnuc_num,&
                                             drotnuc_num
  USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                             se_int_control_type,&
                                             se_taper_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(semi_empirical_type), POINTER       :: sepi, sepj
   REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
   INTEGER, INTENT(IN)                      :: itype
   TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
   TYPE(se_taper_type), POINTER             :: se_taper
   REAL(dp), DIMENSION(45), INTENT(IN), &
      OPTIONAL                               :: e1b, e2a
   REAL(dp), DIMENSION(3, 45), &
      INTENT(IN), OPTIONAL                   :: de1b, de2a

   CHARACTER(len=*), PARAMETER :: moduleN = 'semi_empirical_int_debug', &
                                  routineN = 'check_drotnuc_ana', routineP = moduleN//':'//routineN

   INTEGER                                  :: i, j
   LOGICAL                                  :: l_de1b, l_de2a, l_e1b, l_e2a, &
                                               lgrad, check_value
   REAL(dp)                                 :: delta
   REAL(KIND=dp), DIMENSION(45)             :: e1b2, e2a2
   REAL(KIND=dp), DIMENSION(3, 45)          :: de1b2, de2a2

   l_e1b = PRESENT(e1b)
   l_e2a = PRESENT(e2a)
   l_de1b = PRESENT(de1b)
   l_de2a = PRESENT(de2a)
   lgrad = l_de1b .OR. l_de2a
   delta = 1.0E-5_dp
   ! Check value of integrals
   WRITE (*, *) "DEBUG::"//routineP
   CALL rotnuc_num(sepi, sepj, rijv, e1b2, e2a2, itype, se_int_control, se_taper=se_taper)
   IF (l_e1b) THEN
      DO j = 1, 45
         IF (.NOT. check_value(e1b2(j), e1b(j), delta, 0.1_dp)) THEN
            WRITE (*, *) "ERROR for E1B value E1B(j), j::", j
            CALL cp__b("semi_empirical_int_debug.F",445,"")
         END IF
      END DO
   END IF
   IF (l_e2a) THEN
      DO J = 1, 45
         IF (.NOT. check_value(e2a2(j), e2a(j), delta, 0.1_dp)) THEN
            WRITE (*, *) "ERROR for E2A value E2A(j), j::", j
            CALL cp__b("semi_empirical_int_debug.F",453,"")
         END IF
      END DO
   END IF

   ! Check derivatives
   IF (lgrad) THEN
      CALL drotnuc_num(sepi, sepj, rijv, de1b2, de2a2, itype, delta=delta, &
                       se_int_control=se_int_control, se_taper=se_taper)
      IF (l_de1b) THEN
         DO i = 1, 3
            DO j = 1, 45
               ! Additional check on the value of the integral before checking derivatives
               IF (ABS(e1b2(j)) > delta) THEN
                  IF (.NOT. check_value(de1b2(i, j), de1b(i, j), delta, 0.1_dp)) THEN
                     WRITE (*, *) "ERROR for derivative of E1B.  DE1B(i,j), i,j::", i, j
                     CALL cp__b("semi_empirical_int_debug.F",469,"")
                  END IF
               END IF
            END DO
         END DO
      END IF
      IF (l_de2a) THEN
         DO i = 1, 3
            DO j = 1, 45
               ! Additional check on the value of the integral before checking derivatives
               IF (ABS(e2a2(j)) > delta) THEN
                  IF (.NOT. check_value(de2a2(i, j), de2a(i, j), delta, 0.1_dp)) THEN
                     WRITE (*, *) "ERROR for derivative of E2A.  DE2A(i,j), i,j::", i, j
                     CALL cp__b("semi_empirical_int_debug.F",482,"")
                  END IF
               END IF
            END DO
         END DO
      END IF
   END IF
END SUBROUTINE check_drotnuc_ana

! **************************************************************************************************
!> \brief Check Numerical Vs Analytical CORE-CORE
!> \param sepi ...
!> \param sepj ...
!> \param rijv ...
!> \param itype ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param enuc ...
!> \param denuc ...
!> \par History
!>      04.2007 created [tlaino]
!> \author Teodoro Laino - Zurich University
!> \note
!>      Debug routine
! **************************************************************************************************
SUBROUTINE check_dcorecore_ana(sepi, sepj, rijv, itype, se_int_control, se_taper, enuc, denuc)

  USE kinds,                           ONLY: dp
  USE semi_empirical_int_num,          ONLY: corecore_num,&
                                             dcorecore_num
  USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                             se_int_control_type,&
                                             se_taper_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(semi_empirical_type), POINTER       :: sepi, sepj
   REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
   INTEGER, INTENT(IN)                      :: itype
   REAL(dp), INTENT(IN), OPTIONAL           :: enuc
   REAL(dp), DIMENSION(3), INTENT(IN), &
      OPTIONAL                            :: denuc
   TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
   TYPE(se_taper_type), POINTER             :: se_taper

   CHARACTER(len=*), PARAMETER :: moduleN = 'semi_empirical_int_debug', &
                                  routineN = 'check_dcorecore_ana', routineP = moduleN//':'//routineN

   LOGICAL                                  :: check_value
   INTEGER                                  :: j
   REAL(dp)                                 :: enuc_num, delta
   REAL(dp), DIMENSION(3)                   :: denuc_num

   WRITE (*, *) "DEBUG::"//routineP
   delta = 1.0E-7_dp
   ! check
   IF (PRESENT(enuc)) THEN
      CALL corecore_num(sepi, sepj, rijv, enuc_num, itype, se_int_control, se_taper)
      IF (.NOT. check_value(enuc, enuc_num, delta, 0.001_dp)) THEN
         WRITE (*, *) "ERROR for CORE-CORE energy value (numerical different from analytical)!!"
         CALL cp__b("semi_empirical_int_debug.F",541,"")
      END IF
   END IF
   IF (PRESENT(denuc)) THEN
      CALL dcorecore_num(sepi, sepj, rijv, denuc_num, itype, delta, se_int_control, se_taper)
      DO j = 1, 3
         IF (.NOT. check_value(denuc(j), denuc_num(j), delta, 0.001_dp)) THEN
            WRITE (*, *) "ERROR for CORE-CORE energy derivative value (numerical different from analytical). DENUC(j), j::", j
            CALL cp__b("semi_empirical_int_debug.F",549,"")
         END IF
      END DO
   END IF
END SUBROUTINE check_dcorecore_ana

! **************************************************************************************************
!> \brief Check Numerical Vs Analytical DTEREP_ANA
!> \param sepi ...
!> \param sepj ...
!> \param r ...
!> \param ri ...
!> \param dri ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param lgrad ...
!> \par History
!>      04.2007 created [tlaino]
!> \author Teodoro Laino - Zurich University
!> \note
!>      Debug routine
! **************************************************************************************************
SUBROUTINE check_dterep_ana(sepi, sepj, r, ri, dri, se_int_control, se_taper, lgrad)

  USE kinds,                           ONLY: dp
  USE semi_empirical_int_num,          ONLY: terep_num
  USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                             se_int_control_type,&
                                             se_taper_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(semi_empirical_type), POINTER       :: sepi, sepj
   REAL(dp), INTENT(IN)                     :: r
   REAL(dp), DIMENSION(491), INTENT(IN)     :: ri, dri
   TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
   LOGICAL, INTENT(IN)                      :: lgrad
   TYPE(se_taper_type), POINTER             :: se_taper

   CHARACTER(len=*), PARAMETER :: moduleN = 'semi_empirical_int_debug', &
                                  routineN = 'check_dterep_ana', routineP = moduleN//':'//routineN

   LOGICAL                                  :: check_value
   INTEGER                                  :: j, i
   REAL(dp)                                 :: delta, od, rn
   REAL(dp), DIMENSION(491)                 :: nri, ri0, rim, rip

   delta = 1.0E-8_dp
   od = 0.5_dp/delta
   rn = r
   CALL terep_num(sepi, sepj, rn, ri0, se_taper, se_int_control)
   IF (lgrad) THEN
      rn = r+delta
      CALL terep_num(sepi, sepj, rn, rip, se_taper, se_int_control)
      rn = r-delta
      CALL terep_num(sepi, sepj, rn, rim, se_taper, se_int_control)
      nri = od*(rip-rim)
   END IF
   ! check
   WRITE (*, *) "DEBUG::"//routineP
   DO j = 1, 491
      IF (ABS(ri(j)-ri0(j)) > EPSILON(0.0_dp)) THEN
         WRITE (*, *) "Error in value of the integral RI", j, ri(j), ri0(j)
         CALL cp__b("semi_empirical_int_debug.F",611,"")
      END IF
      IF (lgrad) THEN
         IF (.NOT. check_value(dri(j), nri(j), delta*100.0_dp, 0.1_dp)) THEN
            WRITE (*, *) "ERROR for derivative of RI integral, RI(j), j::", j
            WRITE (*, *) "FULL SET OF INTEGRALS: INDX  ANAL  NUM   DIFF"
            DO i = 1, 491
               WRITE (*, '(I5,3F15.9)') i, dri(i), nri(i), nri(i)-dri(i)
            END DO
            CALL cp__b("semi_empirical_int_debug.F",620,"")
         END IF
      END IF
   END DO
END SUBROUTINE check_dterep_ana

! **************************************************************************************************
!> \brief Check Numerical Vs Analytical ROTINT_ANA
!> \param sepi ...
!> \param sepj ...
!> \param rijv ...
!> \param w ...
!> \param dw ...
!> \param se_int_control ...
!> \param se_taper ...
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
!> \note
!>      Debug routine
! **************************************************************************************************
SUBROUTINE check_rotint_ana(sepi, sepj, rijv, w, dw, se_int_control, se_taper)

  USE kinds,                           ONLY: dp
  USE semi_empirical_int_num,          ONLY: rotint_num,&
                                             drotint_num
  USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                             se_int_control_type,&
                                             se_taper_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(semi_empirical_type), POINTER       :: sepi, sepj
   REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
   REAL(dp), DIMENSION(2025), INTENT(IN), &
      OPTIONAL                               :: w
   REAL(dp), DIMENSION(3, 2025), &
      INTENT(IN), OPTIONAL                   :: dw
   TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
   TYPE(se_taper_type), POINTER             :: se_taper

   CHARACTER(len=*), PARAMETER :: moduleN = 'semi_empirical_int_debug', &
                                  routineN = 'rotint_ana', routineP = moduleN//':'//routineN

   REAL(dp), DIMENSION(2025)                :: w2
   REAL(dp), DIMENSION(3, 2025)             :: dw2
   LOGICAL                                  :: check_value
   INTEGER                                  :: j, i
   REAL(KIND=dp)                            :: delta

   delta = 1.0E-6_dp
   WRITE (*, *) "DEBUG::"//routineP
   IF (PRESENT(w)) THEN
      w2 = 0.0_dp
      CALL rotint_num(sepi, sepj, rijv, w2, se_int_control, se_taper=se_taper)
      DO j = 1, 2025
         IF (.NOT. check_value(w(j), w2(j), delta, 0.1_dp)) THEN
            WRITE (*, *) "ERROR for integral value W(j), j::", j
            CALL cp__b("semi_empirical_int_debug.F",677,"")
         END IF
      END DO
   END IF
   IF (PRESENT(dw)) THEN
      ! Numerical derivatives are obviosly a big problem..
      ! First of all let's decide if the value we get for delta is compatible
      ! with a reasonable value of the integral.. (compatible if the value of the
      ! integral is greater than 1.0E-6)
      CALL drotint_num(sepi, sepj, rijv, dw2, delta=delta, se_int_control=se_int_control, se_taper=se_taper)
      CALL rotint_num(sepi, sepj, rijv, w2, se_int_control=se_int_control, se_taper=se_taper)
      DO i = 1, 3
         DO j = 1, 2025
            IF ((ABS(w2(j)) > delta) .AND. (ABS(dw2(i, j)) > delta*10)) THEN
               IF (.NOT. check_value(dw(i, j), dw2(i, j), delta, 0.1_dp)) THEN
                  WRITE (*, *) "ERROR for derivative of the integral value W(j). DW(i,j) i,j::", i, j
                  CALL cp__b("semi_empirical_int_debug.F",693,"")
               END IF
            END IF
         END DO
      END DO
   END IF
END SUBROUTINE check_rotint_ana
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
SUBROUTINE NAG_dummy()

END SUBROUTINE NAG_dummy
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Debug routines for multipoles
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param nonbond_env ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param iw ...
!> \param debug_r_space ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
SUBROUTINE debug_ewald_multipoles(ewald_env, ewald_pw, nonbond_env, cell, &
                                  particle_set, local_particles, iw, debug_r_space)
  USE cell_types,                      ONLY: cell_type
  USE distribution_1d_types,           ONLY: distribution_1d_type
  USE ewald_environment_types,         ONLY: ewald_environment_type
  USE ewald_pw_types,                  ONLY: ewald_pw_type
  USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                             neighbor_kind_pairs_type
  USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_get,&
                                             fist_nonbond_env_type,&
                                             pos_type
  USE kinds,                           ONLY: dp
  USE mathlib,                         ONLY: matvec_3x3
  USE parallel_rng_types,              ONLY: UNIFORM,&
                                             create_rng_stream,&
                                             delete_rng_stream,&
                                             rng_stream_type,&
                                             random_numbers
  USE particle_types,                  ONLY: particle_type
  USE ewalds_multipole,                ONLY: ewald_multipole_evaluate
  USE cp_log_handling,                 ONLY: cp_logger_type,&
                                             cp_get_default_logger
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE

   TYPE charge_mono_type
      REAL(KIND=dp), DIMENSION(:), &
         POINTER                          :: charge
      REAL(KIND=dp), DIMENSION(:, :), &
         POINTER                          :: pos
   END TYPE charge_mono_type
   TYPE multi_charge_type
      TYPE(charge_mono_type), DIMENSION(:), &
         POINTER                          :: charge_typ
   END TYPE multi_charge_type
   TYPE(ewald_environment_type), POINTER    :: ewald_env
   TYPE(ewald_pw_type), POINTER             :: ewald_pw
   TYPE(fist_nonbond_env_type), POINTER     :: nonbond_env
   TYPE(cell_type), POINTER                 :: cell
   TYPE(particle_type), DIMENSION(:), &
      POINTER                                :: particle_set
   TYPE(distribution_1d_type), POINTER      :: local_particles
   INTEGER, INTENT(IN)                      :: iw
   LOGICAL, INTENT(IN)                      :: debug_r_space

   CHARACTER(len=*), PARAMETER :: routineN = 'debug_ewald_multipoles', &
                                  routineP = "ewalds_multipole_debug"//':'//routineN

   INTEGER                                  :: nparticles
   LOGICAL, DIMENSION(3)                    :: task
   REAL(KIND=dp)                            :: e_neut, e_self, g_energy, &
                                               r_energy, debug_energy
   REAL(KIND=dp), POINTER, DIMENSION(:)     :: charges
   REAL(KIND=dp), POINTER, &
      DIMENSION(:, :)                     :: dipoles, g_forces, g_pv, &
                                             r_forces, r_pv, e_field1, &
                                             e_field2
   REAL(KIND=dp), POINTER, &
      DIMENSION(:, :, :)                  :: quadrupoles
   TYPE(rng_stream_type), POINTER           :: random_stream
   TYPE(multi_charge_type), DIMENSION(:), &
      POINTER                             :: multipoles

   NULLIFY (random_stream, multipoles, charges, dipoles, g_forces, g_pv, &
            r_forces, r_pv, e_field1, e_field2)
   CALL create_rng_stream(random_stream, name="DEBUG_EWALD_MULTIPOLE", &
                          distribution_type=UNIFORM)
   ! check:  charge - charge
   task = .FALSE.
   nparticles = SIZE(particle_set)

   ! Allocate charges, dipoles, quadrupoles
   ALLOCATE (charges(nparticles))
   ALLOCATE (dipoles(3, nparticles))
   ALLOCATE (quadrupoles(3, 3, nparticles))

   ! Allocate arrays for forces
   ALLOCATE (r_forces(3, nparticles))
   ALLOCATE (g_forces(3, nparticles))
   ALLOCATE (e_field1(3, nparticles))
   ALLOCATE (e_field2(3, nparticles))
   ALLOCATE (g_pv(3, 3))
   ALLOCATE (r_pv(3, 3))

   ! Debug CHARGES-CHARGES
   task(1) = .TRUE.
   charges = 0.0_dp
   dipoles = 0.0_dp
   quadrupoles = 0.0_dp
   r_forces = 0.0_dp
   g_forces = 0.0_dp
   e_field1 = 0.0_dp
   e_field2 = 0.0_dp
   g_pv = 0.0_dp
   r_pv = 0.0_dp
   g_energy = 0.0_dp
   r_energy = 0.0_dp
   e_neut = 0.0_dp
   e_self = 0.0_dp

   CALL create_multi_type(multipoles, nparticles, 1, nparticles/2, "CHARGE", echarge=-1.0_dp, &
                          random_stream=random_stream, charges=charges)
   CALL create_multi_type(multipoles, nparticles, nparticles/2+1, nparticles, "CHARGE", echarge=1.0_dp, &
                          random_stream=random_stream, charges=charges)
   CALL debug_ewald_multipole_low(particle_set, cell, nonbond_env, multipoles, debug_energy, &
                                  debug_r_space)

   WRITE (*, *) "DEBUG ENERGY (CHARGE-CHARGE): ", debug_energy
   CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, &
                                 particle_set, local_particles, g_energy, r_energy, e_neut, e_self, &
                                 task, do_correction_bonded=.FALSE., do_forces=.TRUE., do_stress=.TRUE., do_efield=.FALSE., &
                                 charges=charges, dipoles=dipoles, quadrupoles=quadrupoles, forces_local=g_forces, &
                                 forces_glob=r_forces, pv_local=g_pv, pv_glob=r_pv, iw=iw, do_debug=.FALSE.)
   CALL release_multi_type(multipoles)

   ! Debug CHARGES-DIPOLES
   task(1) = .TRUE.
   task(2) = .TRUE.
   charges = 0.0_dp
   dipoles = 0.0_dp
   quadrupoles = 0.0_dp
   r_forces = 0.0_dp
   g_forces = 0.0_dp
   e_field1 = 0.0_dp
   e_field2 = 0.0_dp
   g_pv = 0.0_dp
   r_pv = 0.0_dp
   g_energy = 0.0_dp
   r_energy = 0.0_dp
   e_neut = 0.0_dp
   e_self = 0.0_dp

   CALL create_multi_type(multipoles, nparticles, 1, nparticles/2, "CHARGE", echarge=-1.0_dp, &
                          random_stream=random_stream, charges=charges)
   CALL create_multi_type(multipoles, nparticles, nparticles/2+1, nparticles, "DIPOLE", echarge=0.5_dp, &
                          random_stream=random_stream, dipoles=dipoles)
   WRITE (*, '("CHARGES",F15.9)') charges
   WRITE (*, '("DIPOLES",3F15.9)') dipoles
   CALL debug_ewald_multipole_low(particle_set, cell, nonbond_env, multipoles, debug_energy, &
                                  debug_r_space)

   WRITE (*, *) "DEBUG ENERGY (CHARGE-DIPOLE): ", debug_energy
   CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, &
                                 particle_set, local_particles, g_energy, r_energy, e_neut, e_self, &
                                 task, do_correction_bonded=.FALSE., do_forces=.TRUE., do_stress=.TRUE., do_efield=.FALSE., &
                                 charges=charges, dipoles=dipoles, quadrupoles=quadrupoles, forces_local=g_forces, &
                                 forces_glob=r_forces, pv_local=g_pv, pv_glob=r_pv, iw=iw, do_debug=.FALSE.)
   CALL release_multi_type(multipoles)

   ! Debug DIPOLES-DIPOLES
   task(2) = .TRUE.
   charges = 0.0_dp
   dipoles = 0.0_dp
   quadrupoles = 0.0_dp
   r_forces = 0.0_dp
   g_forces = 0.0_dp
   e_field1 = 0.0_dp
   e_field2 = 0.0_dp
   g_pv = 0.0_dp
   r_pv = 0.0_dp
   g_energy = 0.0_dp
   r_energy = 0.0_dp
   e_neut = 0.0_dp
   e_self = 0.0_dp

   CALL create_multi_type(multipoles, nparticles, 1, nparticles/2, "DIPOLE", echarge=10000.0_dp, &
                          random_stream=random_stream, dipoles=dipoles)
   CALL create_multi_type(multipoles, nparticles, nparticles/2+1, nparticles, "DIPOLE", echarge=20000._dp, &
                          random_stream=random_stream, dipoles=dipoles)
   WRITE (*, '("DIPOLES",3F15.9)') dipoles
   CALL debug_ewald_multipole_low(particle_set, cell, nonbond_env, multipoles, debug_energy, &
                                  debug_r_space)

   WRITE (*, *) "DEBUG ENERGY (DIPOLE-DIPOLE): ", debug_energy
   CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, &
                                 particle_set, local_particles, g_energy, r_energy, e_neut, e_self, &
                                 task, do_correction_bonded=.FALSE., do_forces=.TRUE., do_stress=.TRUE., do_efield=.FALSE., &
                                 charges=charges, dipoles=dipoles, quadrupoles=quadrupoles, forces_local=g_forces, &
                                 forces_glob=r_forces, pv_local=g_pv, pv_glob=r_pv, iw=iw, do_debug=.FALSE.)
   CALL release_multi_type(multipoles)

   ! Debug CHARGES-QUADRUPOLES
   task(1) = .TRUE.
   task(3) = .TRUE.
   charges = 0.0_dp
   dipoles = 0.0_dp
   quadrupoles = 0.0_dp
   r_forces = 0.0_dp
   g_forces = 0.0_dp
   e_field1 = 0.0_dp
   e_field2 = 0.0_dp
   g_pv = 0.0_dp
   r_pv = 0.0_dp
   g_energy = 0.0_dp
   r_energy = 0.0_dp
   e_neut = 0.0_dp
   e_self = 0.0_dp

   CALL create_multi_type(multipoles, nparticles, 1, nparticles/2, "CHARGE", echarge=-1.0_dp, &
                          random_stream=random_stream, charges=charges)
   CALL create_multi_type(multipoles, nparticles, nparticles/2+1, nparticles, "QUADRUPOLE", echarge=10.0_dp, &
                          random_stream=random_stream, quadrupoles=quadrupoles)
   WRITE (*, '("CHARGES",F15.9)') charges
   WRITE (*, '("QUADRUPOLES",9F15.9)') quadrupoles
   CALL debug_ewald_multipole_low(particle_set, cell, nonbond_env, multipoles, debug_energy, &
                                  debug_r_space)

   WRITE (*, *) "DEBUG ENERGY (CHARGE-QUADRUPOLE): ", debug_energy
   CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, &
                                 particle_set, local_particles, g_energy, r_energy, e_neut, e_self, &
                                 task, do_correction_bonded=.FALSE., do_forces=.TRUE., do_stress=.TRUE., do_efield=.FALSE., &
                                 charges=charges, dipoles=dipoles, quadrupoles=quadrupoles, forces_local=g_forces, &
                                 forces_glob=r_forces, pv_local=g_pv, pv_glob=r_pv, iw=iw, do_debug=.FALSE.)
   CALL release_multi_type(multipoles)

   ! Debug DIPOLES-QUADRUPOLES
   task(2) = .TRUE.
   task(3) = .TRUE.
   charges = 0.0_dp
   dipoles = 0.0_dp
   quadrupoles = 0.0_dp
   r_forces = 0.0_dp
   g_forces = 0.0_dp
   e_field1 = 0.0_dp
   e_field2 = 0.0_dp
   g_pv = 0.0_dp
   r_pv = 0.0_dp
   g_energy = 0.0_dp
   r_energy = 0.0_dp
   e_neut = 0.0_dp
   e_self = 0.0_dp

   CALL create_multi_type(multipoles, nparticles, 1, nparticles/2, "DIPOLE", echarge=10000.0_dp, &
                          random_stream=random_stream, dipoles=dipoles)
   CALL create_multi_type(multipoles, nparticles, nparticles/2+1, nparticles, "QUADRUPOLE", echarge=10000.0_dp, &
                          random_stream=random_stream, quadrupoles=quadrupoles)
   WRITE (*, '("DIPOLES",3F15.9)') dipoles
   WRITE (*, '("QUADRUPOLES",9F15.9)') quadrupoles
   CALL debug_ewald_multipole_low(particle_set, cell, nonbond_env, multipoles, debug_energy, &
                                  debug_r_space)

   WRITE (*, *) "DEBUG ENERGY (DIPOLE-QUADRUPOLE): ", debug_energy
   CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, &
                                 particle_set, local_particles, g_energy, r_energy, e_neut, e_self, &
                                 task, do_correction_bonded=.FALSE., do_forces=.TRUE., do_stress=.TRUE., do_efield=.FALSE., &
                                 charges=charges, dipoles=dipoles, quadrupoles=quadrupoles, forces_local=g_forces, &
                                 forces_glob=r_forces, pv_local=g_pv, pv_glob=r_pv, iw=iw, do_debug=.FALSE.)
   CALL release_multi_type(multipoles)

   ! Debug QUADRUPOLES-QUADRUPOLES
   task(3) = .TRUE.
   charges = 0.0_dp
   dipoles = 0.0_dp
   quadrupoles = 0.0_dp
   r_forces = 0.0_dp
   g_forces = 0.0_dp
   e_field1 = 0.0_dp
   e_field2 = 0.0_dp
   g_pv = 0.0_dp
   r_pv = 0.0_dp
   g_energy = 0.0_dp
   r_energy = 0.0_dp
   e_neut = 0.0_dp
   e_self = 0.0_dp

   CALL create_multi_type(multipoles, nparticles, 1, nparticles/2, "QUADRUPOLE", echarge=-20000.0_dp, &
                          random_stream=random_stream, quadrupoles=quadrupoles)
   CALL create_multi_type(multipoles, nparticles, nparticles/2+1, nparticles, "QUADRUPOLE", echarge=10000.0_dp, &
                          random_stream=random_stream, quadrupoles=quadrupoles)
   WRITE (*, '("QUADRUPOLES",9F15.9)') quadrupoles
   CALL debug_ewald_multipole_low(particle_set, cell, nonbond_env, multipoles, debug_energy, &
                                  debug_r_space)

   WRITE (*, *) "DEBUG ENERGY (QUADRUPOLE-QUADRUPOLE): ", debug_energy
   CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, &
                                 particle_set, local_particles, g_energy, r_energy, e_neut, e_self, &
                                 task, do_correction_bonded=.FALSE., do_forces=.TRUE., do_stress=.TRUE., do_efield=.FALSE., &
                                 charges=charges, dipoles=dipoles, quadrupoles=quadrupoles, forces_local=g_forces, &
                                 forces_glob=r_forces, pv_local=g_pv, pv_glob=r_pv, iw=iw, do_debug=.FALSE.)
   CALL release_multi_type(multipoles)

   DEALLOCATE (charges)
   DEALLOCATE (dipoles)
   DEALLOCATE (quadrupoles)
   DEALLOCATE (r_forces)
   DEALLOCATE (g_forces)
   DEALLOCATE (e_field1)
   DEALLOCATE (e_field2)
   DEALLOCATE (g_pv)
   DEALLOCATE (r_pv)
   CALL delete_rng_stream(random_stream)

CONTAINS
! **************************************************************************************************
!> \brief  Debug routines for multipoles - low level - charge interactions
!> \param particle_set ...
!> \param cell ...
!> \param nonbond_env ...
!> \param multipoles ...
!> \param energy ...
!> \param debug_r_space ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
   SUBROUTINE debug_ewald_multipole_low(particle_set, cell, nonbond_env, multipoles, &
                                        energy, debug_r_space)
      IMPLICIT NONE
      TYPE(particle_type), DIMENSION(:), &
         POINTER                                :: particle_set
      TYPE(cell_type), POINTER                 :: cell
      TYPE(fist_nonbond_env_type), POINTER     :: nonbond_env
      TYPE(multi_charge_type), DIMENSION(:), &
         POINTER                             :: multipoles
      REAL(KIND=dp), INTENT(OUT)               :: energy
      LOGICAL, INTENT(IN)                      :: debug_r_space

      CHARACTER(len=*), PARAMETER :: routineN = 'debug_ewald_multipole_low', &
                                     routineP = "ewalds_multipole_debug"//':'//routineN

      INTEGER                                  :: atom_a, atom_b, iend, igrp, &
                                                  ikind, ilist, ipair, istart, &
                                                  jkind, nkinds, npairs, k, k1, l, l1, &
                                                  ncells, icell, jcell, kcell
      INTEGER, DIMENSION(:, :), POINTER        :: list
      REAL(KIND=dp)                            :: r, rab2, rab2_max, fac_ij, q
      REAL(KIND=dp), DIMENSION(3)              :: cell_v, cvi, rab, rm, rab0
      TYPE(fist_neighbor_type), POINTER        :: nonbonded
      TYPE(neighbor_kind_pairs_type), POINTER  :: neighbor_kind_pair
      TYPE(pos_type), DIMENSION(:), POINTER    :: r_last_update, &
                                                  r_last_update_pbc

      energy = 0.0_dp
      CALL fist_nonbond_env_get(nonbond_env, nonbonded=nonbonded, natom_types=nkinds, &
                                r_last_update=r_last_update, r_last_update_pbc=r_last_update_pbc)
      rab2_max = HUGE(0.0_dp)
      IF (debug_r_space) THEN
         ! This debugs the real space part of the multipole Ewald summation scheme
         ! Starting the force loop
         Lists: DO ilist = 1, nonbonded%nlists
            neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
            npairs = neighbor_kind_pair%npairs
            IF (npairs == 0) CYCLE
            list => neighbor_kind_pair%list
            cvi = neighbor_kind_pair%cell_vector
            CALL matvec_3x3(cell_v, cell%hmat, cvi)
            Kind_Group_Loop: DO igrp = 1, neighbor_kind_pair%ngrp_kind
               istart = neighbor_kind_pair%grp_kind_start(igrp)
               iend = neighbor_kind_pair%grp_kind_end(igrp)
               ikind = neighbor_kind_pair%ij_kind(1, igrp)
               jkind = neighbor_kind_pair%ij_kind(2, igrp)
               Pairs: DO ipair = istart, iend
                  fac_ij = 1.0_dp
                  atom_a = list(1, ipair)
                  atom_b = list(2, ipair)
                  IF (atom_a == atom_b) fac_ij = 0.5_dp
                  rab = r_last_update_pbc(atom_b)%r-r_last_update_pbc(atom_a)%r
                  rab = rab+cell_v
                  rab2 = rab(1)**2+rab(2)**2+rab(3)**2
                  IF (rab2 <= rab2_max) THEN

                     DO k = 1, SIZE(multipoles(atom_a)%charge_typ)
                        DO k1 = 1, SIZE(multipoles(atom_a)%charge_typ(k)%charge)

                           DO l = 1, SIZE(multipoles(atom_b)%charge_typ)
                              DO l1 = 1, SIZE(multipoles(atom_b)%charge_typ(l)%charge)

                                 rm = rab+multipoles(atom_b)%charge_typ(l)%pos(:, l1)-multipoles(atom_a)%charge_typ(k)%pos(:, k1)
                                 r = SQRT(DOT_PRODUCT(rm, rm))
                                 q = multipoles(atom_b)%charge_typ(l)%charge(l1)*multipoles(atom_a)%charge_typ(k)%charge(k1)
                                 energy = energy+q/r*fac_ij
                              END DO
                           END DO

                        END DO
                     END DO

                  END IF
               END DO Pairs
            END DO Kind_Group_Loop
         END DO Lists
      ELSE
         ncells = 6
         !Debugs the sum of real + space terms.. (Charge-Charge and Charge-Dipole should be anyway wrong but
         !all the other terms should be correct)
         DO atom_a = 1, SIZE(particle_set)
         DO atom_b = atom_a, SIZE(particle_set)
            fac_ij = 1.0_dp
            IF (atom_a == atom_b) fac_ij = 0.5_dp
            rab0 = r_last_update_pbc(atom_b)%r-r_last_update_pbc(atom_a)%r
            ! Loop over cells
            DO icell = -ncells, ncells
            DO jcell = -ncells, ncells
            DO kcell = -ncells, ncells
               cell_v = MATMUL(cell%hmat, REAL((/icell, jcell, kcell/), KIND=dp))
               IF (ALL(cell_v == 0.0_dp) .AND. (atom_a == atom_b)) CYCLE
               rab = rab0+cell_v
               rab2 = rab(1)**2+rab(2)**2+rab(3)**2
               IF (rab2 <= rab2_max) THEN

                  DO k = 1, SIZE(multipoles(atom_a)%charge_typ)
                     DO k1 = 1, SIZE(multipoles(atom_a)%charge_typ(k)%charge)

                        DO l = 1, SIZE(multipoles(atom_b)%charge_typ)
                           DO l1 = 1, SIZE(multipoles(atom_b)%charge_typ(l)%charge)

                              rm = rab+multipoles(atom_b)%charge_typ(l)%pos(:, l1)-multipoles(atom_a)%charge_typ(k)%pos(:, k1)
                              r = SQRT(DOT_PRODUCT(rm, rm))
                              q = multipoles(atom_b)%charge_typ(l)%charge(l1)*multipoles(atom_a)%charge_typ(k)%charge(k1)
                              energy = energy+q/r*fac_ij
                           END DO
                        END DO

                     END DO
                  END DO

               END IF
            END DO
            END DO
            END DO
         END DO
         END DO
      END IF
   END SUBROUTINE debug_ewald_multipole_low

! **************************************************************************************************
!> \brief  create multi_type for multipoles
!> \param multipoles ...
!> \param idim ...
!> \param istart ...
!> \param iend ...
!> \param label ...
!> \param echarge ...
!> \param random_stream ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
   SUBROUTINE create_multi_type(multipoles, idim, istart, iend, label, echarge, &
                                random_stream, charges, dipoles, quadrupoles)
      IMPLICIT NONE
      TYPE(multi_charge_type), DIMENSION(:), &
         POINTER                             :: multipoles
      INTEGER, INTENT(IN)                      :: idim, istart, iend
      CHARACTER(LEN=*), INTENT(IN)             :: label
      REAL(KIND=dp), INTENT(IN)                :: echarge
      TYPE(rng_stream_type), POINTER           :: random_stream
      REAL(KIND=dp), DIMENSION(:), POINTER, &
         OPTIONAL                            :: charges
      REAL(KIND=dp), DIMENSION(:, :), POINTER, &
         OPTIONAL                            :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER, &
         OPTIONAL                            :: quadrupoles

      CHARACTER(len=*), PARAMETER :: routineN = 'create_multi_type', &
                                     routineP = "ewalds_multipole_debug"//':'//routineN

      INTEGER                                  :: i, isize, k, l, m
      REAL(KIND=dp)                            :: dx, rvec(3), rvec1(3), rvec2(3), r2

      IF (ASSOCIATED(multipoles)) THEN
         IF(.NOT.(SIZE(multipoles) == idim))CALL cp__a("ewalds_multipole_debug.F",481)
      ELSE
         ALLOCATE (multipoles(idim))
         DO i = 1, idim
            NULLIFY (multipoles(i)%charge_typ)
         END DO
      END IF
      DO i = istart, iend
         IF (ASSOCIATED(multipoles(i)%charge_typ)) THEN
            ! make a copy of the array and enlarge the array type by 1
            isize = SIZE(multipoles(i)%charge_typ)+1
         ELSE
            isize = 1
         END IF
         CALL reallocate_charge_type(multipoles(i)%charge_typ, 1, isize)
         SELECT CASE (label)
         CASE ("CHARGE")
            IF(.NOT.(PRESENT(charges)))CALL cp__a("ewalds_multipole_debug.F",498)
            IF(.NOT.(ASSOCIATED(charges)))CALL cp__a("ewalds_multipole_debug.F",499)
            ALLOCATE (multipoles(i)%charge_typ(isize)%charge(1))
            ALLOCATE (multipoles(i)%charge_typ(isize)%pos(3, 1))

            multipoles(i)%charge_typ(isize)%charge(1) = echarge
            multipoles(i)%charge_typ(isize)%pos(1:3, 1) = 0.0_dp
            charges(i) = charges(i)+echarge
         CASE ("DIPOLE")
            dx = 1.0E-4_dp
            IF(.NOT.(PRESENT(dipoles)))CALL cp__a("ewalds_multipole_debug.F",508)
            IF(.NOT.(ASSOCIATED(dipoles)))CALL cp__a("ewalds_multipole_debug.F",509)
            ALLOCATE (multipoles(i)%charge_typ(isize)%charge(2))
            ALLOCATE (multipoles(i)%charge_typ(isize)%pos(3, 2))
            CALL random_numbers(rvec, random_stream)
            rvec = rvec/(2.0_dp*SQRT(DOT_PRODUCT(rvec, rvec)))*dx
            multipoles(i)%charge_typ(isize)%charge(1) = echarge
            multipoles(i)%charge_typ(isize)%pos(1:3, 1) = rvec
            multipoles(i)%charge_typ(isize)%charge(2) = -echarge
            multipoles(i)%charge_typ(isize)%pos(1:3, 2) = -rvec

            dipoles(:, i) = dipoles(:, i)+2.0_dp*echarge*rvec
         CASE ("QUADRUPOLE")
            dx = 1.0E-2_dp
            IF(.NOT.(PRESENT(quadrupoles)))CALL cp__a("ewalds_multipole_debug.F",522)
            IF(.NOT.(ASSOCIATED(quadrupoles)))CALL cp__a("ewalds_multipole_debug.F",523)
            ALLOCATE (multipoles(i)%charge_typ(isize)%charge(4))
            ALLOCATE (multipoles(i)%charge_typ(isize)%pos(3, 4))
            CALL random_numbers(rvec1, random_stream)
            CALL random_numbers(rvec2, random_stream)
            rvec1 = rvec1/SQRT(DOT_PRODUCT(rvec1, rvec1))
            rvec2 = rvec2-DOT_PRODUCT(rvec2, rvec1)*rvec1
            rvec2 = rvec2/SQRT(DOT_PRODUCT(rvec2, rvec2))
            !
            rvec1 = rvec1/2.0_dp*dx
            rvec2 = rvec2/2.0_dp*dx
            !       + (4)  ^      - (1)
            !              |rvec2
            !              |
            !              0------> rvec1
            !
            !
            !       - (3)         + (2)
            multipoles(i)%charge_typ(isize)%charge(1) = -echarge
            multipoles(i)%charge_typ(isize)%pos(1:3, 1) = rvec1+rvec2
            multipoles(i)%charge_typ(isize)%charge(2) = echarge
            multipoles(i)%charge_typ(isize)%pos(1:3, 2) = rvec1-rvec2
            multipoles(i)%charge_typ(isize)%charge(3) = -echarge
            multipoles(i)%charge_typ(isize)%pos(1:3, 3) = -rvec1-rvec2
            multipoles(i)%charge_typ(isize)%charge(4) = echarge
            multipoles(i)%charge_typ(isize)%pos(1:3, 4) = -rvec1+rvec2

            DO k = 1, 4
               r2 = DOT_PRODUCT(multipoles(i)%charge_typ(isize)%pos(:, k), multipoles(i)%charge_typ(isize)%pos(:, k))
               DO l = 1, 3
                  DO m = 1, 3
                     quadrupoles(m, l, i) = quadrupoles(m, l, i)+3.0_dp*0.5_dp*multipoles(i)%charge_typ(isize)%charge(k)* &
                                            multipoles(i)%charge_typ(isize)%pos(l, k)* &
                                            multipoles(i)%charge_typ(isize)%pos(m, k)
                     IF (m == l) quadrupoles(m, l, i) = quadrupoles(m, l, i)-0.5_dp*multipoles(i)%charge_typ(isize)%charge(k)*r2
                  END DO
               END DO
            END DO

         END SELECT
      END DO
   END SUBROUTINE create_multi_type

! **************************************************************************************************
!> \brief  release multi_type for multipoles
!> \param multipoles ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
   SUBROUTINE release_multi_type(multipoles)
      IMPLICIT NONE
      TYPE(multi_charge_type), DIMENSION(:), &
         POINTER                             :: multipoles

      CHARACTER(len=*), PARAMETER :: routineN = 'release_multi_type', &
                                     routineP = "ewalds_multipole_debug"//':'//routineN

      INTEGER                                  :: i, j

      IF (ASSOCIATED(multipoles)) THEN
         DO i = 1, SIZE(multipoles)
            DO j = 1, SIZE(multipoles(i)%charge_typ)
               DEALLOCATE (multipoles(i)%charge_typ(j)%charge)
               DEALLOCATE (multipoles(i)%charge_typ(j)%pos)
            END DO
            DEALLOCATE (multipoles(i)%charge_typ)
         END DO
      END IF
   END SUBROUTINE release_multi_type

! **************************************************************************************************
!> \brief  reallocates multi_type for multipoles
!> \param charge_typ ...
!> \param istart ...
!> \param iend ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
   SUBROUTINE reallocate_charge_type(charge_typ, istart, iend)
      IMPLICIT NONE
      TYPE(charge_mono_type), DIMENSION(:), &
         POINTER                             :: charge_typ
      INTEGER, INTENT(IN)                      :: istart, iend

      CHARACTER(len=*), PARAMETER :: routineN = 'reallocate_charge_type', &
                                     routineP = "ewalds_multipole_debug"//':'//routineN

      INTEGER                                  :: i, isize, jsize, jsize1, jsize2, j
      TYPE(charge_mono_type), DIMENSION(:), &
         POINTER                             :: charge_typ_bk

      IF (ASSOCIATED(charge_typ)) THEN
         isize = SIZE(charge_typ)
         ALLOCATE (charge_typ_bk(1:isize))
         DO j = 1, isize
            jsize = SIZE(charge_typ(j)%charge)
            ALLOCATE (charge_typ_bk(j)%charge(jsize))
            jsize1 = SIZE(charge_typ(j)%pos, 1)
            jsize2 = SIZE(charge_typ(j)%pos, 2)
            ALLOCATE (charge_typ_bk(j)%pos(jsize1, jsize2))
            charge_typ_bk(j)%pos = charge_typ(j)%pos
            charge_typ_bk(j)%charge = charge_typ(j)%charge
         END DO
         DO j = 1, SIZE(charge_typ)
            DEALLOCATE (charge_typ(j)%charge)
            DEALLOCATE (charge_typ(j)%pos)
         END DO
         DEALLOCATE (charge_typ)
         ! Reallocate
         ALLOCATE (charge_typ_bk(istart:iend))
         DO i = istart, isize
            jsize = SIZE(charge_typ_bk(j)%charge)
            ALLOCATE (charge_typ(j)%charge(jsize))
            jsize1 = SIZE(charge_typ_bk(j)%pos, 1)
            jsize2 = SIZE(charge_typ_bk(j)%pos, 2)
            ALLOCATE (charge_typ(j)%pos(jsize1, jsize2))
            charge_typ(j)%pos = charge_typ_bk(j)%pos
            charge_typ(j)%charge = charge_typ_bk(j)%charge
         END DO
         DO j = 1, SIZE(charge_typ_bk)
            DEALLOCATE (charge_typ_bk(j)%charge)
            DEALLOCATE (charge_typ_bk(j)%pos)
         END DO
         DEALLOCATE (charge_typ_bk)
      ELSE
         ALLOCATE (charge_typ(istart:iend))
      END IF

   END SUBROUTINE reallocate_charge_type

END SUBROUTINE debug_ewald_multipoles

! **************************************************************************************************
!> \brief  Routine to debug potential, field and electric field gradients
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param nonbond_env ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param radii ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \param task ...
!> \param iw ...
!> \param atomic_kind_set ...
!> \param mm_section ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
SUBROUTINE debug_ewald_multipoles_fields(ewald_env, ewald_pw, nonbond_env, cell, &
                                         particle_set, local_particles, radii, charges, dipoles, quadrupoles, task, iw, &
                                         atomic_kind_set, mm_section)
  USE atomic_kind_types,               ONLY: atomic_kind_type
  USE cell_types,                      ONLY: cell_type
  USE distribution_1d_types,           ONLY: distribution_1d_type
  USE ewald_environment_types,         ONLY: ewald_environment_type
  USE ewald_pw_types,                  ONLY: ewald_pw_type
  USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                             neighbor_kind_pairs_type
  USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_get,&
                                             fist_nonbond_env_type,&
                                             pos_type
  USE input_section_types,             ONLY: section_vals_type
  USE kinds,                           ONLY: dp
  USE particle_types,                  ONLY: particle_type
  USE ewalds_multipole,                ONLY: ewald_multipole_evaluate
  USE fist_neighbor_list_control,      ONLY: list_control
  USE cp_log_handling,                 ONLY: cp_logger_type,&
                                             cp_get_default_logger
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(ewald_environment_type), POINTER    :: ewald_env
   TYPE(ewald_pw_type), POINTER             :: ewald_pw
   TYPE(fist_nonbond_env_type), POINTER     :: nonbond_env
   TYPE(cell_type), POINTER                 :: cell
   TYPE(particle_type), POINTER             :: particle_set(:)
   TYPE(distribution_1d_type), POINTER      :: local_particles
   REAL(KIND=dp), DIMENSION(:), &
      POINTER, OPTIONAL                   :: radii, charges
   REAL(KIND=dp), DIMENSION(:, :), &
      POINTER, OPTIONAL                   :: dipoles
   REAL(KIND=dp), DIMENSION(:, :, :), &
      POINTER, OPTIONAL                   :: quadrupoles
   LOGICAL, DIMENSION(3), INTENT(IN)        :: task
   INTEGER, INTENT(IN)                      :: iw
   TYPE(atomic_kind_type), POINTER          :: atomic_kind_set(:)
   TYPE(section_vals_type), POINTER         :: mm_section

   CHARACTER(len=*), PARAMETER :: routineN = 'debug_ewald_multipoles_fields', &
                                  routineP = "ewalds_multipole_debug"//':'//routineN

   REAL(KIND=dp), POINTER, DIMENSION(:)     :: lcharges, efield0
   REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: ldipoles
   REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: lquadrupoles
   REAL(KIND=dp)                            :: dq, energy_local, energy_glob, e_neut, &
                                               e_self, pv_local(3, 3), pv_glob(3, 3), ene(2), pot, coord(3), efield1n(3), tot_ene, &
                                               o_tot_ene, enev(3, 2), efield2n(3, 3)
   REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE &
      :: forces_local, forces_glob, efield1, efield2
   INTEGER :: nparticles, i, k, j, nparticle_local, iparticle_kind
   TYPE(particle_type), POINTER, DIMENSION(:) :: shell_particle_set, core_particle_set
   TYPE(cp_logger_type), POINTER              :: logger

   NULLIFY (lcharges, ldipoles, lquadrupoles, shell_particle_set, core_particle_set)
   NULLIFY (logger)
   logger => cp_get_default_logger()

   nparticles = SIZE(particle_set)
   nparticle_local = 0
   DO iparticle_kind = 1, SIZE(local_particles%n_el)
      nparticle_local = nparticle_local+local_particles%n_el(iparticle_kind)
   END DO
   ALLOCATE (lcharges(nparticles))
   ALLOCATE (forces_glob(3, nparticles))
   ALLOCATE (forces_local(3, nparticle_local))
   ALLOCATE (efield0(nparticles))
   ALLOCATE (efield1(3, nparticles))
   ALLOCATE (efield2(9, nparticles))
   forces_glob = 0.0_dp
   forces_local = 0.0_dp
   efield0 = 0.0_dp
   efield1 = 0.0_dp
   efield2 = 0.0_dp
   pv_local = 0.0_dp
   pv_glob = 0.0_dp
   energy_glob = 0.0_dp
   energy_local = 0.0_dp
   e_neut = 0.0_dp
   e_self = 0.0_dp
   CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, particle_set, &
                                 local_particles, energy_local, energy_glob, e_neut, e_self, task, .FALSE., .TRUE., .TRUE., &
                                 .TRUE., radii, charges, dipoles, quadrupoles, forces_local, forces_glob, pv_local, pv_glob, &
                                 efield0, efield1, efield2, iw, do_debug=.FALSE.)
   o_tot_ene = energy_local+energy_glob+e_neut+e_self
   WRITE (*, *) "TOTAL ENERGY :: ========>", o_tot_ene
   ! Debug Potential
   dq = 0.001_dp
   tot_ene = 0.0_dp
   DO i = 1, nparticles
      DO k = 1, 2
         lcharges = charges
         lcharges(i) = charges(i)+(-1.0_dp)**k*dq
         forces_glob = 0.0_dp
         forces_local = 0.0_dp
         pv_local = 0.0_dp
         pv_glob = 0.0_dp
         energy_glob = 0.0_dp
         energy_local = 0.0_dp
         e_neut = 0.0_dp
         e_self = 0.0_dp
         CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, particle_set, &
                                       local_particles, energy_local, energy_glob, e_neut, e_self, &
                                       task, .FALSE., .FALSE., .FALSE., .FALSE., radii, &
                                       lcharges, dipoles, quadrupoles, iw=iw, do_debug=.FALSE.)
         ene(k) = energy_local+energy_glob+e_neut+e_self
      END DO
      pot = (ene(2)-ene(1))/(2.0_dp*dq)
      WRITE (*, '(A,I8,3(A,F15.9))') "POTENTIAL FOR ATOM: ", i, " NUMERICAL: ", pot, " ANALYTICAL: ", efield0(i), &
         " ERROR: ", pot-efield0(i)
      tot_ene = tot_ene+0.5_dp*efield0(i)*charges(i)
   END DO
   WRITE (*, *) "ENERGIES: ", o_tot_ene, tot_ene, o_tot_ene-tot_ene
   WRITE (*, '(/,/,/)')
   ! Debug Field
   dq = 0.001_dp
   DO i = 1, nparticles
      coord = particle_set(i)%r
      DO j = 1, 3
         DO k = 1, 2
            particle_set(i)%r(j) = coord(j)+(-1.0_dp)**k*dq

            ! Rebuild neighbor lists
            CALL list_control(atomic_kind_set, particle_set, local_particles, &
                              cell, nonbond_env, logger%para_env, mm_section, &
                              shell_particle_set, core_particle_set)

            forces_glob = 0.0_dp
            forces_local = 0.0_dp
            pv_local = 0.0_dp
            pv_glob = 0.0_dp
            energy_glob = 0.0_dp
            energy_local = 0.0_dp
            e_neut = 0.0_dp
            e_self = 0.0_dp
            efield0 = 0.0_dp
            CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, particle_set, &
                                          local_particles, energy_local, energy_glob, e_neut, e_self, &
                                          task, .FALSE., .TRUE., .TRUE., .TRUE., radii, &
                                          charges, dipoles, quadrupoles, forces_local, forces_glob, &
                                          pv_local, pv_glob, efield0, iw=iw, do_debug=.FALSE.)
            ene(k) = efield0(i)
            particle_set(i)%r(j) = coord(j)
         END DO
         efield1n(j) = -(ene(2)-ene(1))/(2.0_dp*dq)
      END DO
      WRITE (*, '(/,A,I8)') "FIELD FOR ATOM: ", i
      WRITE (*, '(A,3F15.9)') " NUMERICAL: ", efield1n, " ANALYTICAL: ", efield1(:, i), &
         " ERROR: ", efield1n-efield1(:, i)
      IF (task(2)) THEN
         tot_ene = tot_ene-0.5_dp*DOT_PRODUCT(efield1(:, i), dipoles(:, i))
      END IF
   END DO
   WRITE (*, *) "ENERGIES: ", o_tot_ene, tot_ene, o_tot_ene-tot_ene

   ! Debug Field Gradient
   dq = 0.0001_dp
   DO i = 1, nparticles
      coord = particle_set(i)%r
      DO j = 1, 3
         DO k = 1, 2
            particle_set(i)%r(j) = coord(j)+(-1.0_dp)**k*dq

            ! Rebuild neighbor lists
            CALL list_control(atomic_kind_set, particle_set, local_particles, &
                              cell, nonbond_env, logger%para_env, mm_section, &
                              shell_particle_set, core_particle_set)

            forces_glob = 0.0_dp
            forces_local = 0.0_dp
            pv_local = 0.0_dp
            pv_glob = 0.0_dp
            energy_glob = 0.0_dp
            energy_local = 0.0_dp
            e_neut = 0.0_dp
            e_self = 0.0_dp
            efield1 = 0.0_dp
            CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, particle_set, &
                                          local_particles, energy_local, energy_glob, e_neut, e_self, &
                                          task, .FALSE., .TRUE., .TRUE., .TRUE., radii, &
                                          charges, dipoles, quadrupoles, forces_local, forces_glob, &
                                          pv_local, pv_glob, efield1=efield1, iw=iw, do_debug=.FALSE.)
            enev(:, k) = efield1(:, i)
            particle_set(i)%r(j) = coord(j)
         END DO
         efield2n(:, j) = (enev(:, 2)-enev(:, 1))/(2.0_dp*dq)
      END DO
      WRITE (*, '(/,A,I8)') "FIELD GRADIENT FOR ATOM: ", i
      WRITE (*, '(A,9F15.9)') " NUMERICAL:  ", efield2n, &
         " ANALYTICAL: ", efield2(:, i), &
         " ERROR:      ", RESHAPE(efield2n, (/9/))-efield2(:, i)
   END DO
END SUBROUTINE debug_ewald_multipoles_fields

! **************************************************************************************************
!> \brief  Routine to debug potential, field and electric field gradients
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param nonbond_env ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param radii ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \param task ...
!> \param iw ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
SUBROUTINE debug_ewald_multipoles_fields2(ewald_env, ewald_pw, nonbond_env, cell, &
                                          particle_set, local_particles, radii, charges, dipoles, quadrupoles, task, iw)
  USE cell_types,                      ONLY: cell_type
  USE distribution_1d_types,           ONLY: distribution_1d_type
  USE ewald_environment_types,         ONLY: ewald_environment_type
  USE ewald_pw_types,                  ONLY: ewald_pw_type
  USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                             neighbor_kind_pairs_type
  USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_get,&
                                             fist_nonbond_env_type,&
                                             pos_type
  USE kinds,                           ONLY: dp
  USE particle_types,                  ONLY: particle_type
  USE ewalds_multipole,                ONLY: ewald_multipole_evaluate
  USE fist_neighbor_list_control,      ONLY: list_control
  USE cp_log_handling,                 ONLY: cp_logger_type,&
                                             cp_get_default_logger
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   TYPE(ewald_environment_type), POINTER    :: ewald_env
   TYPE(ewald_pw_type), POINTER             :: ewald_pw
   TYPE(fist_nonbond_env_type), POINTER     :: nonbond_env
   TYPE(cell_type), POINTER                 :: cell
   TYPE(particle_type), POINTER             :: particle_set(:)
   TYPE(distribution_1d_type), POINTER      :: local_particles
   REAL(KIND=dp), DIMENSION(:), &
      POINTER, OPTIONAL                   :: radii, charges
   REAL(KIND=dp), DIMENSION(:, :), &
      POINTER, OPTIONAL                   :: dipoles
   REAL(KIND=dp), DIMENSION(:, :, :), &
      POINTER, OPTIONAL                   :: quadrupoles
   LOGICAL, DIMENSION(3), INTENT(IN)        :: task
   INTEGER, INTENT(IN)                      :: iw

   CHARACTER(len=*), PARAMETER :: routineN = 'debug_ewald_multipoles_fields', &
                                  routineP = "ewalds_multipole_debug"//':'//routineN

   REAL(KIND=dp), POINTER, DIMENSION(:)     :: efield0
   REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: ldipoles
   REAL(KIND=dp), POINTER, DIMENSION(:, :)   :: lquadrupoles
   REAL(KIND=dp)                            :: energy_local, energy_glob, e_neut, &
                                               e_self, pv_local(3, 3), pv_glob(3, 3), tot_ene, o_tot_ene, prod
   REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE &
      :: forces_local, forces_glob, efield1, efield2
   INTEGER :: nparticles, i, k, j, nparticle_local, iparticle_kind, ind
   TYPE(particle_type), POINTER, DIMENSION(:) :: shell_particle_set, core_particle_set
   TYPE(cp_logger_type), POINTER              :: logger

   NULLIFY (ldipoles, lquadrupoles, shell_particle_set, core_particle_set)
   NULLIFY (logger)
   logger => cp_get_default_logger()

   nparticles = SIZE(particle_set)
   nparticle_local = 0
   DO iparticle_kind = 1, SIZE(local_particles%n_el)
      nparticle_local = nparticle_local+local_particles%n_el(iparticle_kind)
   END DO
   ALLOCATE (forces_glob(3, nparticles))
   ALLOCATE (forces_local(3, nparticle_local))
   ALLOCATE (efield0(nparticles))
   ALLOCATE (efield1(3, nparticles))
   ALLOCATE (efield2(9, nparticles))
   forces_glob = 0.0_dp
   forces_local = 0.0_dp
   efield0 = 0.0_dp
   efield1 = 0.0_dp
   efield2 = 0.0_dp
   pv_local = 0.0_dp
   pv_glob = 0.0_dp
   energy_glob = 0.0_dp
   energy_local = 0.0_dp
   e_neut = 0.0_dp
   e_self = 0.0_dp
   CALL ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, cell, particle_set, &
                                 local_particles, energy_local, energy_glob, e_neut, e_self, task, .FALSE., .TRUE., .TRUE., &
                                 .TRUE., radii, charges, dipoles, quadrupoles, forces_local, forces_glob, pv_local, pv_glob, &
                                 efield0, efield1, efield2, iw, do_debug=.FALSE.)
   o_tot_ene = energy_local+energy_glob+e_neut+e_self
   WRITE (*, *) "TOTAL ENERGY :: ========>", o_tot_ene

   ! Debug Potential
   tot_ene = 0.0_dp
   IF (task(1)) THEN
      DO i = 1, nparticles
         tot_ene = tot_ene+0.5_dp*efield0(i)*charges(i)
      END DO
      WRITE (*, *) "ENERGIES: ", o_tot_ene, tot_ene, o_tot_ene-tot_ene
      WRITE (*, '(/,/,/)')
   END IF

   ! Debug Field
   IF (task(2)) THEN
      DO i = 1, nparticles
         tot_ene = tot_ene-0.5_dp*DOT_PRODUCT(efield1(:, i), dipoles(:, i))
      END DO
      WRITE (*, *) "ENERGIES: ", o_tot_ene, tot_ene, o_tot_ene-tot_ene
      WRITE (*, '(/,/,/)')
   END IF

   ! Debug Field Gradient
   IF (task(3)) THEN
      DO i = 1, nparticles
         ind = 0
         prod = 0.0_dp
         DO j = 1, 3
            DO k = 1, 3
               ind = ind+1
               prod = prod+efield2(ind, i)*quadrupoles(j, k, i)
            END DO
         END DO
         tot_ene = tot_ene-0.5_dp*(1.0_dp/3.0_dp)*prod
      END DO
      WRITE (*, *) "ENERGIES: ", o_tot_ene, tot_ene, o_tot_ene-tot_ene
      WRITE (*, '(/,/,/)')
   END IF

END SUBROUTINE debug_ewald_multipoles_fields2
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of Coulomb integrals over Cartesian Gaussian-type functions
!>      (electron repulsion integrals, ERIs).
!> \par Literature
!>      S. Obara and A. Saika, J. Chem. Phys. 84, 3963 (1986)
!> \par History
!>      none
!> \author J. Hutter (07.2009)
! **************************************************************************************************
MODULE ai_eri_debug

   USE gamma,                           ONLY: fgamma_ref
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_eri_debug'

   INTEGER, PARAMETER            :: lmax = 5

   REAL(dp)                      :: xa, xb, xc, xd
   REAL(dp), DIMENSION(3)        :: A, B, C, D
   REAL(dp), DIMENSION(3)        :: P, Q, W
   REAL(dp)                      :: xsi, eta, rho, T

   REAL(dp), DIMENSION(0:4*lmax) :: fm, I0M

   PRIVATE

CONTAINS

! **************************************************************************************************
!> \brief   Calculation of the primitive two-center Coulomb integrals over
!>          Cartesian Gaussian-type functions.
!> \param ya ...
!> \param yb ...
!> \param yc ...
!> \param yd ...
!> \param rA ...
!> \param rB ...
!> \param rC ...
!> \param rD ...
!> \date    07.2009
!> \author  J. Hutter
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_os(ya, yb, yc, yd, rA, rB, rC, rD)
      REAL(dp)                                           :: ya, yb, yc, yd
      REAL(dp), DIMENSION(3)                             :: rA, rB, rC, rD

      REAL(dp)                                           :: eab, ecd, kab, kcd

      xa = ya
      xb = yb
      xc = yc
      xd = yd
      A = rA
      B = rB
      C = rC
      D = rD

      xsi = xa+xb
      eta = xc+xd

      P = (xa*A+xb*B)/xsi
      Q = (xc*C+xd*D)/eta
      W = (xsi*P+eta*Q)/(xsi+eta)

      rho = xsi*eta/(xsi+eta)

      T = rho*SUM((P-Q)**2)

      fm = fgamma_ref(4*lmax, T)

      eab = -xa*xb/xsi*SUM((A-B)**2)
      kab = SQRT(2._dp)*pi**1.25_dp/xsi*EXP(eab)

      ecd = -xc*xd/eta*SUM((C-D)**2)
      kcd = SQRT(2._dp)*pi**1.25_dp/eta*EXP(ecd)

      I0M = kab*kcd/SQRT(xsi+eta)*fm

   END SUBROUTINE init_os

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param an ...
!> \param bn ...
!> \param cn ...
!> \param dn ...
!> \param mi ...
!> \retval IABCD ...
! **************************************************************************************************
   RECURSIVE FUNCTION os(an, bn, cn, dn, mi) RESULT(IABCD)
      INTEGER, DIMENSION(3)                              :: an, bn, cn, dn
      INTEGER, OPTIONAL                                  :: mi
      REAL(dp)                                           :: IABCD

      CHARACTER(len=*), PARAMETER :: routineN = 'os', routineP = moduleN//':'//routineN
      INTEGER, DIMENSION(3), PARAMETER                   :: i1 = (/1, 0, 0/), i2 = (/0, 1, 0/), &
                                                            i3 = (/0, 0, 1/)

      INTEGER                                            :: m

      m = 0
      IF (PRESENT(mi)) m = mi

      IABCD = 0._dp
      IF (ANY(an < 0)) RETURN
      IF (ANY(bn < 0)) RETURN
      IF (ANY(cn < 0)) RETURN
      IF (ANY(dn < 0)) RETURN

      IF (SUM(an+bn+cn+dn) == 0) THEN
         IABCD = I0M(m)
         RETURN
      END IF

      IF (dn(1) > 0) THEN
         IABCD = os(an, bn, cn+i1, dn-i1)-(D(1)-C(1))*os(an, bn, cn, dn-i1)
      ELSEIF (dn(2) > 0) THEN
         IABCD = os(an, bn, cn+i2, dn-i2)-(D(2)-C(2))*os(an, bn, cn, dn-i2)
      ELSEIF (dn(3) > 0) THEN
         IABCD = os(an, bn, cn+i3, dn-i3)-(D(3)-C(3))*os(an, bn, cn, dn-i3)
      ELSE
         IF (bn(1) > 0) THEN
            IABCD = os(an+i1, bn-i1, cn, dn)-(B(1)-A(1))*os(an, bn-i1, cn, dn)
         ELSEIF (bn(2) > 0) THEN
            IABCD = os(an+i2, bn-i2, cn, dn)-(B(2)-A(2))*os(an, bn-i2, cn, dn)
         ELSEIF (bn(3) > 0) THEN
            IABCD = os(an+i3, bn-i3, cn, dn)-(B(3)-A(3))*os(an, bn-i3, cn, dn)
         ELSE
            IF (cn(1) > 0) THEN
               IABCD = ((Q(1)-C(1))+xsi/eta*(P(1)-A(1)))*os(an, bn, cn-i1, dn)+ &
                       0.5_dp*an(1)/eta*os(an-i1, bn, cn-i1, dn)+ &
                       0.5_dp*(cn(1)-1)/eta*os(an, bn, cn-i1-i1, dn)- &
                       xsi/eta*os(an+i1, bn, cn-i1, dn)
            ELSEIF (cn(2) > 0) THEN
               IABCD = ((Q(2)-C(2))+xsi/eta*(P(2)-A(2)))*os(an, bn, cn-i2, dn)+ &
                       0.5_dp*an(2)/eta*os(an-i2, bn, cn-i2, dn)+ &
                       0.5_dp*(cn(2)-1)/eta*os(an, bn, cn-i2-i2, dn)- &
                       xsi/eta*os(an+i2, bn, cn-i2, dn)
            ELSEIF (cn(3) > 0) THEN
               IABCD = ((Q(3)-C(3))+xsi/eta*(P(3)-A(3)))*os(an, bn, cn-i3, dn)+ &
                       0.5_dp*an(3)/eta*os(an-i3, bn, cn-i3, dn)+ &
                       0.5_dp*(cn(3)-1)/eta*os(an, bn, cn-i3-i3, dn)- &
                       xsi/eta*os(an+i3, bn, cn-i3, dn)
            ELSE
               IF (an(1) > 0) THEN
                  IABCD = (P(1)-A(1))*os(an-i1, bn, cn, dn, m)+ &
                          (W(1)-P(1))*os(an-i1, bn, cn, dn, m+1)+ &
                          0.5_dp*(an(1)-1)/xsi*os(an-i1-i1, bn, cn, dn, m)- &
                          0.5_dp*(an(1)-1)/xsi*rho/xsi*os(an-i1-i1, bn, cn, dn, m+1)
               ELSEIF (an(2) > 0) THEN
                  IABCD = (P(2)-A(2))*os(an-i2, bn, cn, dn, m)+ &
                          (W(2)-P(2))*os(an-i2, bn, cn, dn, m+1)+ &
                          0.5_dp*(an(2)-1)/xsi*os(an-i2-i2, bn, cn, dn, m)- &
                          0.5_dp*(an(2)-1)/xsi*rho/xsi*os(an-i2-i2, bn, cn, dn, m+1)
               ELSEIF (an(3) > 0) THEN
                  IABCD = (P(3)-A(3))*os(an-i3, bn, cn, dn, m)+ &
                          (W(3)-P(3))*os(an-i3, bn, cn, dn, m+1)+ &
                          0.5_dp*(an(3)-1)/xsi*os(an-i3-i3, bn, cn, dn, m)- &
                          0.5_dp*(an(3)-1)/xsi*rho/xsi*os(an-i3-i3, bn, cn, dn, m+1)
               ELSE
                  CALL cp__b("aobasis/ai_eri_debug.F",174,"I(0000)")
               END IF
            END IF
         END IF
      END IF

   END FUNCTION os

! **************************************************************************************************

END MODULE ai_eri_debug
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! less not much meningful...

! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_rs
   USE realspace_grid_types,            ONLY: realspace_grid_p_type,&
                                              realspace_grid_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_rs'

!API type
   PUBLIC :: cp_sll_rs_type, cp_sll_rs_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_rs_create, cp_sll_rs_dealloc, &
             cp_sll_rs_next, &
             cp_sll_rs_get_length, &
             cp_sll_rs_get_el_at, cp_sll_rs_set_el_at, &
             cp_sll_rs_get_first_el, cp_sll_rs_get_rest, &
             cp_sll_rs_insert_el, cp_sll_rs_insert_el_at, &
             cp_sll_rs_rm_first_el, cp_sll_rs_rm_el_at, &
             cp_sll_rs_rm_all_el, &
             cp_sll_rs_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_rs_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_rs_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_rs_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_rs_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_rs_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_rs_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_rs_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_rs_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_rs_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_rs_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_rs_rm_first_el, &
         cp_sll_rs_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_rs_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_rs_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_rs_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_rs_type
      TYPE(realspace_grid_type), POINTER :: first_el
      TYPE(cp_sll_rs_type), POINTER :: rest
   END TYPE cp_sll_rs_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_rs_p_type
      TYPE(cp_sll_rs_type), POINTER :: list
   END TYPE cp_sll_rs_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_create(sll, first_el, rest)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      TYPE(realspace_grid_type), OPTIONAL, POINTER       :: first_el
      TYPE(cp_sll_rs_type), OPTIONAL, POINTER            :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el => first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_rs_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_dealloc(sll)
      TYPE(cp_sll_rs_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_rs_rm_all_el(sll)
   END SUBROUTINE cp_sll_rs_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_dealloc_node(sll)
      TYPE(cp_sll_rs_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_rs_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_rs_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      TYPE(realspace_grid_type), POINTER                 :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_get_first_el', &
         routineP = moduleN//':'//routineN

      res => sll%first_el
   END FUNCTION cp_sll_rs_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_rs_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_rs_type), POINTER                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("pw/cp_linked_list_rs.F",295,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_rs_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_rs_get_length(sll) RESULT(res)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_rs_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_rs_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(realspace_grid_type), POINTER                 :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: pos

      IF (index == -1) THEN
         pos => cp_sll_rs_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_rs_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_rs.F",367)

      res => pos%first_el
   END FUNCTION cp_sll_rs_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_set_el_at(sll, index, value)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(realspace_grid_type), POINTER                 :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: pos

      IF (index == -1) THEN
         pos => cp_sll_rs_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_rs_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_rs.F",399)

      pos%first_el => value
   END SUBROUTINE cp_sll_rs_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_rs_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_rs_type), POINTER                      :: iterator
      TYPE(realspace_grid_type), OPTIONAL, POINTER       :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_next', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att => iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_rs_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_insert_el(sll, el)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      TYPE(realspace_grid_type), POINTER                 :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_rs_create(newSlot, first_el=el, &
                            rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_rs_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_rm_first_el(sll)
      TYPE(cp_sll_rs_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_rs_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("pw/cp_linked_list_rs.F",484,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_rs_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_insert_el_at(sll, el, index)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      TYPE(realspace_grid_type), POINTER                 :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: pos

      IF (index == 1) THEN
         CALL cp_sll_rs_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_rs_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_rs_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_rs.F",518)
         CALL cp_sll_rs_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_rs_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_rm_el_at(sll, index)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: pos

      IF (index == 1) THEN
         CALL cp_sll_rs_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_rs_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_rs_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("pw/cp_linked_list_rs.F",550)
         CALL cp_sll_rs_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_rs_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_rs_rm_all_el(sll)
      TYPE(cp_sll_rs_type), POINTER                      :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_rs_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_rs_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_rs_to_array(sll) RESULT(res)
      TYPE(cp_sll_rs_type), POINTER                      :: sll
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_rs_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_rs_type), POINTER                      :: iter

      len = cp_sll_rs_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i)%rs_grid => iter%first_el
         IF(.NOT.(cp_sll_rs_next(iter) .OR. i == len))CALL cp__a("pw/cp_linked_list_rs.F",606)
      END DO
   END FUNCTION cp_sll_rs_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "=>"
!  USE = "USE realspace_grid_types, ONLY: realspace_grid_type, realspace_grid_p_type"
!  array= = "=>"
!  arrayEl = "%rs_grid"
!  common_dir = "../common"
!  defines =
!    "! less not much meningful...
!     #define ( el1%id_nr < el2%id_nr ) ( el1%id_nr < el2%id_nr )
!     #define ( el1%id_nr == el2%id_nr ) ( el1%id_nr == el2%id_nr )
!     "
!  equalQ = "CP_SLL_RS_EQUAL_Q"
!  lessQ = "CP_SLL_RS_LESS_Q"
!  nametype1 = "rs"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "type(realspace_grid_type),pointer"
!  type1arrayEl = "type(realspace_grid_p_type)"
!  type1in = "type(realspace_grid_type), pointer"
!  type1out = "type(realspace_grid_type), pointer"
!  write_el = ""

END MODULE cp_linked_list_rs

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Performance for DBCSR multiply
!> \author  VW
!> \date    2010
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created 2010
! **************************************************************************************************
MODULE dbcsr_performance_multiply
   USE dbcsr_data_methods,              ONLY: dbcsr_scalar,&
                                              dbcsr_scalar_negative,&
                                              dbcsr_scalar_one
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_col_dist,&
                                              dbcsr_distribution_new,&
                                              dbcsr_distribution_release,&
                                              dbcsr_distribution_row_dist
   USE dbcsr_dist_operations,           ONLY: dbcsr_dist_bin
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_wrong_args_error
   USE dbcsr_io,                        ONLY: dbcsr_print
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_offsets, dbcsr_col_block_sizes, dbcsr_distribution, dbcsr_get_data_type, &
        dbcsr_get_matrix_type, dbcsr_init, dbcsr_name, dbcsr_nfullcols_total, &
        dbcsr_nfullrows_total, dbcsr_release, dbcsr_row_block_offsets, dbcsr_row_block_sizes
   USE dbcsr_multiply_api,              ONLY: dbcsr_multiply
   USE dbcsr_operations,                ONLY: dbcsr_copy,&
                                              dbcsr_scale
   USE dbcsr_test_methods,              ONLY: atoi,&
                                              atol,&
                                              ator,&
                                              dbcsr_make_random_block_sizes,&
                                              dbcsr_make_random_matrix
   USE dbcsr_transformations,           ONLY: dbcsr_redistribute
   USE dbcsr_types,                     ONLY: &
        dbcsr_conjugate_transpose, dbcsr_distribution_obj, dbcsr_mp_obj, dbcsr_no_transpose, &
        dbcsr_obj, dbcsr_scalar_type, dbcsr_transpose, dbcsr_type_antisymmetric, &
        dbcsr_type_complex_4, dbcsr_type_complex_8, dbcsr_type_no_symmetry, dbcsr_type_real_4, &
        dbcsr_type_real_8
   USE dbcsr_util,                      ONLY: dbcsr_checksum
   USE dbcsr_work_operations,           ONLY: dbcsr_create,&
                                              dbcsr_finalize
   USE kinds,                           ONLY: int_8,&
                                              real_4,&
                                              real_8
   USE machine,                         ONLY: m_walltime
   USE message_passing,                 ONLY: mp_environ,&
                                              mp_sum,&
                                              mp_sync
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: dbcsr_perf_multiply

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_performance_multiply'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param group ...
!> \param mp_env ...
!> \param npdims ...
!> \param io_unit ...
!> \param narg ...
!> \param args ...
! **************************************************************************************************
   SUBROUTINE dbcsr_perf_multiply(group, mp_env, npdims, io_unit, narg, args)

      INTEGER                                            :: group
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(2), INTENT(in)                  :: npdims
      INTEGER                                            :: io_unit, narg
      CHARACTER(len=*), DIMENSION(:), INTENT(IN)         :: args

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_perf_multiply', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: symmetries(3), trans(2)
      INTEGER                                            :: i, iblk, istat, kblk_to_read, limits(6), &
                                                            matrix_sizes(3), mblk_to_read, &
                                                            nblk_to_read, nrep, TYPE
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: bs_k, bs_m, bs_n
      LOGICAL                                            :: retain_sparsity
      REAL(real_8)                                       :: alpha(2), beta(2), sparsities(3)

!
! parsing

      IF (narg .LT. 28) THEN
         WRITE (*, *) "Input file format:"
         WRITE (*, *) "      dbcsr_multiply \\"
         WRITE (*, *) "      M N K \\"
         WRITE (*, *) "      SpA SpB SpC \\"
         WRITE (*, *) "      TrA TrB \\"
         WRITE (*, *) "      SymA SymB SymC \\"
         WRITE (*, *) "      data_type \\"
         WRITE (*, *) "      Re(alpha) Im(alpha) Re(beta) Im(beta) \\"
         WRITE (*, *) "      limRowL limRowU limColL limColU limKL limKU \\"
         WRITE (*, *) "      retain_sparsity nrep \\"
         WRITE (*, *) "      nmblksizes nnblksizes nkblksizes \\"
         WRITE (*, *) "      [mblksizes] [nblksizes] [kblksizes]"
      ENDIF
      CALL dbcsr_assert(narg, "GE", 28, dbcsr_fatal_level, dbcsr_wrong_args_error, &
                        routineN, "narg not correct", 116)

      matrix_sizes(1) = atoi(args(2))
      matrix_sizes(2) = atoi(args(3))
      matrix_sizes(3) = atoi(args(4))
      sparsities(1) = ator(args(5))
      sparsities(2) = ator(args(6))
      sparsities(3) = ator(args(7))
      trans(1) = args(8)
      trans(2) = args(9)
      symmetries(1) = args(10)
      symmetries(2) = args(11)
      symmetries(3) = args(12)
      TYPE = atoi(args(13))
      alpha(1) = ator(args(14))
      alpha(2) = ator(args(15))
      beta(1) = ator(args(16))
      beta(2) = ator(args(17))
      limits(1) = atoi(args(18))
      limits(2) = atoi(args(19))
      limits(3) = atoi(args(20))
      limits(4) = atoi(args(21))
      limits(5) = atoi(args(22))
      limits(6) = atoi(args(23))
      retain_sparsity = atol(args(24))
      nrep = atoi(args(25))
      mblk_to_read = atoi(args(26))
      nblk_to_read = atoi(args(27))
      kblk_to_read = atoi(args(28))

      CALL dbcsr_assert(narg, "GE", 28+2*(mblk_to_read+nblk_to_read+kblk_to_read), &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, "narg not correct", &
                        148)

      ALLOCATE (bs_m(2*mblk_to_read), bs_n(2*nblk_to_read), bs_k(2*kblk_to_read), STAT=istat)
      IF (istat /= 0) &
         CALL cp__b("dbcsr/dbcsr_performance_multiply.F",152,"allocattion problem")

      i = 28
      DO iblk = 1, mblk_to_read
         i = i+1
         bs_m(2*(iblk-1)+1) = atoi(args(i))
         i = i+1
         bs_m(2*(iblk-1)+2) = atoi(args(i))
      ENDDO
      DO iblk = 1, nblk_to_read
         i = i+1
         bs_n(2*(iblk-1)+1) = atoi(args(i))
         i = i+1
         bs_n(2*(iblk-1)+2) = atoi(args(i))
      ENDDO
      DO iblk = 1, kblk_to_read
         i = i+1
         bs_k(2*(iblk-1)+1) = atoi(args(i))
         i = i+1
         bs_k(2*(iblk-1)+2) = atoi(args(i))
      ENDDO

      !
      ! do checks here

      !
      ! if the limits are not specified (i.e 0), we set them here
      IF (limits(1) .EQ. 0) limits(1) = 1
      IF (limits(2) .EQ. 0) limits(2) = matrix_sizes(1)
      IF (limits(3) .EQ. 0) limits(3) = 1
      IF (limits(4) .EQ. 0) limits(4) = matrix_sizes(2)
      IF (limits(5) .EQ. 0) limits(5) = 1
      IF (limits(6) .EQ. 0) limits(6) = matrix_sizes(3)

      !
      ! lets go !
      CALL dbcsr_perf_multiply_low(group, mp_env, npdims, io_unit, matrix_sizes, &
                                   bs_m, bs_n, bs_k, sparsities, trans, symmetries, TYPE, &
                                   alpha, beta, limits, retain_sparsity, nrep)

      DEALLOCATE (bs_m, bs_n, bs_k)

   END SUBROUTINE dbcsr_perf_multiply

! **************************************************************************************************
!> \brief Performs a variety of matrix multiplies of same matrices on different
!>        processor grids
!> \param[in] mp_group          MPI communicator
!> \param mp_env ...
!> \param npdims ...
!> \param[in] io_unit           which unit to write to, if not negative
!> \param[in] matrix_sizes      size of matrices to test
!> \param bs_m block sizes of the 3 dimensions
!> \param bs_n block sizes of the 3 dimensions
!> \param bs_k block sizes of the 3 dimensions
!> \param[in] sparsities        sparsities of matrices to create
!> \param trans transposes of the two matrices
!> \param symmetries ...
!> \param data_type types of matrices to create
!> \param alpha_in alpha value to use in multiply
!> \param beta_in beta value to use in multiply
!> \param limits ...
!> \param retain_sparsity ...
!> \param nrep ...
! **************************************************************************************************
   SUBROUTINE dbcsr_perf_multiply_low(mp_group, mp_env, npdims, io_unit, &
                                      matrix_sizes, bs_m, bs_n, bs_k, sparsities, trans, symmetries, data_type, &
                                      alpha_in, beta_in, limits, retain_sparsity, nrep)
      INTEGER, INTENT(IN)                                :: mp_group
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(2), INTENT(in)                  :: npdims
      INTEGER, INTENT(IN)                                :: io_unit
      INTEGER, DIMENSION(:), INTENT(in)                  :: matrix_sizes, bs_m, bs_n, bs_k
      REAL(real_8), DIMENSION(3), INTENT(in)             :: sparsities
      CHARACTER, DIMENSION(2), INTENT(in)                :: trans
      CHARACTER, DIMENSION(3), INTENT(in)                :: symmetries
      INTEGER, INTENT(IN)                                :: data_type
      REAL(real_8), DIMENSION(2), INTENT(in)             :: alpha_in, beta_in
      INTEGER, DIMENSION(6), INTENT(in)                  :: limits
      LOGICAL, INTENT(in)                                :: retain_sparsity
      INTEGER, INTENT(IN)                                :: nrep

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_perf_multiply_low', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: a_symm, b_symm, c_symm, transa, transb
      INTEGER                                            :: handle, mynode, numnodes
      INTEGER, DIMENSION(:), POINTER :: col_dist_a, col_dist_b, col_dist_c, my_sizes_k, &
         my_sizes_m, my_sizes_n, row_dist_a, row_dist_b, row_dist_c, sizes_k, sizes_m, sizes_n
      LOGICAL                                            :: do_complex
      LOGICAL, DIMENSION(2)                              :: trs
      TYPE(dbcsr_distribution_obj)                       :: dist_a, dist_b, dist_c
      TYPE(dbcsr_obj)                                    :: matrix_a, matrix_b, matrix_c
      TYPE(dbcsr_scalar_type)                            :: alpha, beta

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      NULLIFY (my_sizes_k, my_sizes_m, my_sizes_n, &
               sizes_k, sizes_m, sizes_n)

      !
      ! print
      CALL mp_environ(numnodes, mynode, mp_group)
      IF (io_unit .GT. 0) THEN
         WRITE (io_unit, *) 'numnodes', numnodes
         WRITE (io_unit, *) 'matrix_sizes', matrix_sizes
         WRITE (io_unit, *) 'sparsities', sparsities
         WRITE (io_unit, *) 'trans ', trans
         WRITE (io_unit, *) 'symmetries ', symmetries
         WRITE (io_unit, *) 'type ', data_type
         WRITE (io_unit, *) 'alpha_in', alpha_in
         WRITE (io_unit, *) 'beta_in', beta_in
         WRITE (io_unit, *) 'limits', limits
         WRITE (io_unit, *) 'retain_sparsity', retain_sparsity
         WRITE (io_unit, *) 'nrep', nrep
         WRITE (io_unit, *) 'bs_m', bs_m
         WRITE (io_unit, *) 'bs_n', bs_n
         WRITE (io_unit, *) 'bs_k', bs_k
      ENDIF
      !
      !
      a_symm = symmetries(1)
      b_symm = symmetries(2)
      c_symm = symmetries(3)

      IF (a_symm .NE. dbcsr_type_no_symmetry .AND. matrix_sizes(1) .NE. matrix_sizes(3)) &
         CALL cp__b("dbcsr/dbcsr_performance_multiply.F",279,"")

      IF (b_symm .NE. dbcsr_type_no_symmetry .AND. matrix_sizes(2) .NE. matrix_sizes(3)) &
         CALL cp__b("dbcsr/dbcsr_performance_multiply.F",282,"")

      IF (c_symm .NE. dbcsr_type_no_symmetry .AND. matrix_sizes(1) .NE. matrix_sizes(2)) &
         CALL cp__b("dbcsr/dbcsr_performance_multiply.F",285,"")

      do_complex = data_type .EQ. dbcsr_type_complex_4 .OR. data_type .EQ. dbcsr_type_complex_8

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         alpha = dbcsr_scalar(REAL(alpha_in(1), real_4))
         beta = dbcsr_scalar(REAL(beta_in(1), real_4))
      CASE (dbcsr_type_real_8)
         alpha = dbcsr_scalar(REAL(alpha_in(1), real_8))
         beta = dbcsr_scalar(REAL(beta_in(1), real_8))
      CASE (dbcsr_type_complex_4)
         alpha = dbcsr_scalar(CMPLX(alpha_in(1), alpha_in(2), real_4))
         beta = dbcsr_scalar(CMPLX(beta_in(1), beta_in(2), real_4))
      CASE (dbcsr_type_complex_8)
         alpha = dbcsr_scalar(CMPLX(alpha_in(1), alpha_in(2), real_8))
         beta = dbcsr_scalar(CMPLX(beta_in(1), beta_in(2), real_8))
      END SELECT

      transa = trans(1)
      transb = trans(2)

      !
      ! if C has a symmetry, we need special transpositions
      IF (c_symm .NE. dbcsr_type_no_symmetry) THEN
         IF (.NOT. (transa .EQ. dbcsr_no_transpose .AND. transb .EQ. dbcsr_transpose .OR. &
                    transa .EQ. dbcsr_transpose .AND. transb .EQ. dbcsr_no_transpose .OR. &
                    transa .EQ. dbcsr_no_transpose .AND. transb .EQ. dbcsr_conjugate_transpose .AND. &
                    .NOT. do_complex .OR. &
                    transa .EQ. dbcsr_conjugate_transpose .AND. transb .EQ. dbcsr_no_transpose .AND. &
                    .NOT. do_complex)) CALL cp__b("dbcsr/dbcsr_performance_multiply.F",315,"")

      ENDIF
      !
      ! if C has symmetry and special limits
      IF (c_symm .NE. dbcsr_type_no_symmetry) THEN
         IF (limits(1) .NE. 1 .OR. limits(2) .NE. matrix_sizes(1) .OR. &
             limits(3) .NE. 1 .OR. limits(4) .NE. matrix_sizes(2)) CALL cp__b("dbcsr/dbcsr_performance_multiply.F",322,"")

      ENDIF

      !
      ! Create the row/column block sizes.
      CALL dbcsr_make_random_block_sizes(sizes_m, matrix_sizes(1), bs_m)
      CALL dbcsr_make_random_block_sizes(sizes_n, matrix_sizes(2), bs_n)
      CALL dbcsr_make_random_block_sizes(sizes_k, matrix_sizes(3), bs_k)

      !
      ! if we have symmetry the row and column block sizes hae to match
      IF (c_symm .NE. dbcsr_type_no_symmetry .AND. a_symm .NE. dbcsr_type_no_symmetry .AND. &
          b_symm .NE. dbcsr_type_no_symmetry) THEN
         my_sizes_m => sizes_m; my_sizes_n => sizes_m; my_sizes_k => sizes_m
      ELSEIF ((c_symm .EQ. dbcsr_type_no_symmetry .AND. a_symm .NE. dbcsr_type_no_symmetry .AND. &
               b_symm .NE. dbcsr_type_no_symmetry) .OR. &
              (c_symm .NE. dbcsr_type_no_symmetry .AND. a_symm .EQ. dbcsr_type_no_symmetry .AND. &
               b_symm .NE. dbcsr_type_no_symmetry) .OR. &
              (c_symm .NE. dbcsr_type_no_symmetry .AND. a_symm .NE. dbcsr_type_no_symmetry .AND. &
               b_symm .EQ. dbcsr_type_no_symmetry)) THEN
         my_sizes_m => sizes_m; my_sizes_n => sizes_m; my_sizes_k => sizes_m
      ELSEIF (c_symm .EQ. dbcsr_type_no_symmetry .AND. a_symm .EQ. dbcsr_type_no_symmetry .AND. &
              b_symm .NE. dbcsr_type_no_symmetry) THEN
         my_sizes_m => sizes_m; my_sizes_n => sizes_n; my_sizes_k => sizes_n
      ELSEIF (c_symm .EQ. dbcsr_type_no_symmetry .AND. a_symm .NE. dbcsr_type_no_symmetry .AND. &
              b_symm .EQ. dbcsr_type_no_symmetry) THEN
         my_sizes_m => sizes_m; my_sizes_n => sizes_n; my_sizes_k => sizes_m
      ELSEIF (c_symm .NE. dbcsr_type_no_symmetry .AND. a_symm .EQ. dbcsr_type_no_symmetry .AND. &
              b_symm .EQ. dbcsr_type_no_symmetry) THEN
         my_sizes_m => sizes_m; my_sizes_n => sizes_m; my_sizes_k => sizes_k
      ELSEIF (c_symm .EQ. dbcsr_type_no_symmetry .AND. a_symm .EQ. dbcsr_type_no_symmetry .AND. &
              b_symm .EQ. dbcsr_type_no_symmetry) THEN
         my_sizes_m => sizes_m; my_sizes_n => sizes_n; my_sizes_k => sizes_k
      ELSE
         CALL cp__b("dbcsr/dbcsr_performance_multiply.F",357,"something wrong here...")
      ENDIF

      ! Create the random matrices.
      trs(1) = transa .EQ. dbcsr_no_transpose
      trs(2) = transb .EQ. dbcsr_no_transpose
      CALL dbcsr_dist_bin(row_dist_c, SIZE(sizes_m), npdims(1), &
                          sizes_m)
      CALL dbcsr_dist_bin(col_dist_c, SIZE(sizes_n), npdims(2), &
                          sizes_n)
      CALL dbcsr_distribution_new(dist_c, mp_env, row_dist_c, col_dist_c)
      CALL dbcsr_make_random_matrix(matrix_c, sizes_m, sizes_n, "Matrix C", &
                                    REAL(sparsities(3), real_8), &
                                    mp_group, data_type=data_type, dist=dist_c)
      CALL dbcsr_distribution_release(dist_c)
      IF (trs(1)) THEN
         CALL dbcsr_dist_bin(row_dist_a, SIZE(sizes_k), npdims(1), &
                             sizes_k)
         CALL dbcsr_dist_bin(col_dist_a, SIZE(sizes_m), npdims(2), &
                             sizes_m)
         CALL dbcsr_distribution_new(dist_a, mp_env, row_dist_a, col_dist_a)
         CALL dbcsr_make_random_matrix(matrix_a, sizes_k, sizes_m, "Matrix A", &
                                       REAL(sparsities(1), real_8), &
                                       mp_group, data_type=data_type, dist=dist_a)
         DEALLOCATE (row_dist_a, col_dist_a)
      ELSE
         CALL dbcsr_dist_bin(col_dist_a, SIZE(sizes_k), npdims(2), &
                             sizes_k)
         CALL dbcsr_distribution_new(dist_a, mp_env, row_dist_c, col_dist_a)
         CALL dbcsr_make_random_matrix(matrix_a, sizes_m, sizes_k, "Matrix A", &
                                       REAL(sparsities(1), real_8), &
                                       mp_group, data_type=data_type, dist=dist_a)
         DEALLOCATE (col_dist_a)
      ENDIF
      CALL dbcsr_distribution_release(dist_a)
      IF (trs(2)) THEN
         CALL dbcsr_dist_bin(row_dist_b, SIZE(sizes_n), npdims(1), &
                             sizes_n)
         CALL dbcsr_dist_bin(col_dist_b, SIZE(sizes_k), npdims(2), &
                             sizes_k)
         CALL dbcsr_distribution_new(dist_b, mp_env, row_dist_b, col_dist_b)
         CALL dbcsr_make_random_matrix(matrix_b, sizes_n, sizes_k, "Matrix B", &
                                       REAL(sparsities(2), real_8), &
                                       mp_group, data_type=data_type, dist=dist_b)
         DEALLOCATE (row_dist_b, col_dist_b)
      ELSE
         CALL dbcsr_dist_bin(row_dist_b, SIZE(sizes_k), npdims(1), &
                             sizes_k)
         CALL dbcsr_distribution_new(dist_b, mp_env, row_dist_b, col_dist_c)
         CALL dbcsr_make_random_matrix(matrix_b, sizes_k, sizes_n, "Matrix B", &
                                       REAL(sparsities(2), real_8), &
                                       mp_group, data_type=data_type, dist=dist_b)
         DEALLOCATE (row_dist_b)
      ENDIF
      CALL dbcsr_distribution_release(dist_b)
      DEALLOCATE (row_dist_c, col_dist_c, sizes_m, sizes_n, sizes_k)

      !
      ! if C has a symmetry, we build it accordingly, i.e. C=A*A and C=A*(-A)
      IF (c_symm .NE. dbcsr_type_no_symmetry) THEN
         CALL dbcsr_copy(matrix_b, matrix_a)
         !print*, a_symm,b_symm,dbcsr_get_matrix_type(matrix_a),dbcsr_get_matrix_type(matrix_b)
         IF (c_symm .EQ. dbcsr_type_antisymmetric) THEN
            CALL dbcsr_scale(matrix_b, &
                             alpha_scalar=dbcsr_scalar_negative( &
                             dbcsr_scalar_one(data_type)))
         ENDIF
      ENDIF

      !
      ! Prepare test parameters
      CALL perf_multiply(mp_group, mp_env, io_unit, &
                         matrix_a, matrix_b, matrix_c, &
                         transa, transb, &
                         alpha, beta, &
                         limits, retain_sparsity, &
                         nrep)

      !
      ! cleanup
      CALL dbcsr_release(matrix_a)
      CALL dbcsr_release(matrix_b)
      CALL dbcsr_release(matrix_c)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_perf_multiply_low

! **************************************************************************************************
!> \brief Performs a variety of matrix multiplies of same matrices on different
!>        processor grids
!> \param[in] mp_group          MPI communicator
!> \param mp_env ...
!> \param[in] io_unit           which unit to write to, if not negative
!> \param matrix_a matrices to multiply
!> \param matrix_b matrices to multiply
!> \param matrix_c matrices to multiply
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param beta ...
!> \param limits ...
!> \param retain_sparsity ...
!> \param nrep ...
! **************************************************************************************************
   SUBROUTINE perf_multiply(mp_group, mp_env, io_unit, &
                            matrix_a, matrix_b, matrix_c, &
                            transa, transb, alpha, beta, limits, retain_sparsity, &
                            nrep)
      INTEGER, INTENT(IN)                                :: mp_group
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, INTENT(IN)                                :: io_unit
      TYPE(dbcsr_obj), INTENT(in)                        :: matrix_a, matrix_b, matrix_c
      CHARACTER, INTENT(in)                              :: transa, transb
      TYPE(dbcsr_scalar_type), INTENT(in)                :: alpha, beta
      INTEGER, DIMENSION(6), INTENT(in)                  :: limits
      LOGICAL, INTENT(in)                                :: retain_sparsity
      INTEGER, INTENT(IN)                                :: nrep

      CHARACTER(len=*), PARAMETER :: routineN = 'perf_multiply', routineP = moduleN//':'//routineN

      INTEGER                                            :: c_a, c_b, c_c, handle, irep, istat, &
                                                            mynode, nthreads, numnodes, r_a, r_b, &
                                                            r_c
      INTEGER(int_8)                                     :: flop
      INTEGER(int_8), ALLOCATABLE, DIMENSION(:)          :: flop_sum
      INTEGER(int_8), ALLOCATABLE, DIMENSION(:, :)       :: flops
      INTEGER, DIMENSION(:), POINTER                     :: blk_offsets, col_dist_a, col_dist_b, &
                                                            col_dist_c, row_dist_a, row_dist_b, &
                                                            row_dist_c
      REAL(real_8)                                       :: chksum_a, chksum_b, chksum_c_in, &
                                                            chksum_c_out, std_all, std_t, t1, t2
      REAL(real_8), ALLOCATABLE, DIMENSION(:)            :: flops_all, flops_node, flops_thread, &
                                                            load_imb, t, t_max, t_min
      REAL(real_8), ALLOCATABLE, DIMENSION(:, :)         :: times
      TYPE(dbcsr_distribution_obj)                       :: dist_a, dist_b, dist_c
      TYPE(dbcsr_obj)                                    :: m_a, m_b, m_c, m_c_orig

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      CALL mp_environ(numnodes, mynode, mp_group)

      nthreads = 1
!$    nthreads = OMP_GET_MAX_THREADS()

      ALLOCATE (times(0:numnodes-1, nrep), flops(0:numnodes-1, nrep), t_max(nrep), &
                flops_node(nrep), flops_thread(nrep), flops_all(nrep), flop_sum(nrep), &
                t_min(nrep), t(nrep), load_imb(nrep), STAT=istat)
      IF (istat /= 0) &
         CALL cp__b("dbcsr/dbcsr_performance_multiply.F",508,"allocattion problem")
      times(:, :) = 0.0_real_8
      t_max(:) = 0.0_real_8
      t_min(:) = 0.0_real_8
      t(:) = 0.0_real_8
      flops_node(:) = 0.0_real_8
      flops_thread(:) = 0.0_real_8
      flops_all(:) = 0.0_real_8
      flops(:, :) = 0
      flop_sum(:) = 0
      load_imb(:) = 0.0_real_8

      CALL dbcsr_init(m_a)
      CALL dbcsr_init(m_b)
      CALL dbcsr_init(m_c)
      CALL dbcsr_init(m_c_orig)

      ! Row & column distributions
      row_dist_c => dbcsr_distribution_row_dist(dbcsr_distribution(matrix_c))
      col_dist_c => dbcsr_distribution_col_dist(dbcsr_distribution(matrix_c))
      row_dist_a => row_dist_c
      col_dist_a => dbcsr_distribution_col_dist(dbcsr_distribution(matrix_a))
      row_dist_b => dbcsr_distribution_row_dist(dbcsr_distribution(matrix_b))
      col_dist_b => col_dist_c

      CALL dbcsr_distribution_new(dist_a, mp_env, row_dist_a, col_dist_a)
      CALL dbcsr_distribution_new(dist_b, mp_env, row_dist_b, col_dist_b)
      CALL dbcsr_distribution_new(dist_c, mp_env, row_dist_c, col_dist_c)
      ! Redistribute the matrices
      ! A
      CALL dbcsr_create(m_a, "Test for "//TRIM(dbcsr_name(matrix_a)), &
                        dist_a, dbcsr_get_matrix_type(matrix_a), &
                        dbcsr_row_block_sizes(matrix_a), &
                        dbcsr_col_block_sizes(matrix_a), &
                        data_type=dbcsr_get_data_type(matrix_a))
      CALL dbcsr_distribution_release(dist_a)
      CALL dbcsr_redistribute(matrix_a, m_a)
      ! B
      CALL dbcsr_create(m_b, "Test for "//TRIM(dbcsr_name(matrix_b)), &
                        dist_b, dbcsr_get_matrix_type(matrix_b), &
                        dbcsr_row_block_sizes(matrix_b), &
                        dbcsr_col_block_sizes(matrix_b), &
                        data_type=dbcsr_get_data_type(matrix_b))
      CALL dbcsr_distribution_release(dist_b)
      CALL dbcsr_redistribute(matrix_b, m_b)
      ! C orig
      CALL dbcsr_create(m_c_orig, "Test for "//TRIM(dbcsr_name(matrix_c)), &
                        dist_c, dbcsr_get_matrix_type(matrix_c), &
                        dbcsr_row_block_sizes(matrix_c), &
                        dbcsr_col_block_sizes(matrix_c), &
                        data_type=dbcsr_get_data_type(matrix_c))
      CALL dbcsr_distribution_release(dist_c)
      CALL dbcsr_redistribute(matrix_c, m_c_orig)
      ! C
      CALL dbcsr_create(m_c, "Test for "//TRIM(dbcsr_name(matrix_c)), &
                        dist_c, dbcsr_get_matrix_type(matrix_c), &
                        dbcsr_row_block_sizes(matrix_c), &
                        dbcsr_col_block_sizes(matrix_c), &
                        data_type=dbcsr_get_data_type(matrix_c))
      CALL dbcsr_finalize(m_c)

      IF (.FALSE.) THEN
         blk_offsets => dbcsr_row_block_offsets(matrix_c)
         WRITE (*, *) 'row_block_offsets(matrix_c)', blk_offsets
         blk_offsets => dbcsr_col_block_offsets(matrix_c)
         WRITE (*, *) 'col_block_offsets(matrix_c)', blk_offsets
      ENDIF

      IF (.FALSE.) THEN
         CALL dbcsr_print(m_c, matlab_format=.FALSE., variable_name='c_in_')
         CALL dbcsr_print(m_a, matlab_format=.FALSE., variable_name='a_')
         CALL dbcsr_print(m_b, matlab_format=.FALSE., variable_name='b_')
         CALL dbcsr_print(m_c, matlab_format=.FALSE., variable_name='c_out_')
      ENDIF

      r_a = dbcsr_nfullrows_total(m_a)
      c_a = dbcsr_nfullcols_total(m_a)
      r_b = dbcsr_nfullrows_total(m_b)
      c_b = dbcsr_nfullcols_total(m_b)
      r_c = dbcsr_nfullrows_total(m_c_orig)
      c_c = dbcsr_nfullcols_total(m_c_orig)

      chksum_a = dbcsr_checksum(m_a)
      chksum_b = dbcsr_checksum(m_b)
      chksum_c_in = dbcsr_checksum(m_c_orig)

      !
      !
      DO irep = 1, nrep

         !
         ! set the C matrix
         CALL dbcsr_copy(m_c, m_c_orig)

         !
         ! Perform multiply
         CALL mp_sync(mp_group)
         t1 = m_walltime()
         flop = 0
         CALL dbcsr_multiply(transa, transb, alpha, &
                             m_a, m_b, beta, m_c, &
                             first_row=limits(1), &
                             last_row=limits(2), &
                             first_column=limits(3), &
                             last_column=limits(4), &
                             first_k=limits(5), &
                             last_k=limits(6), &
                             retain_sparsity=retain_sparsity, &
                             flop=flop)
         t2 = m_walltime()
         times(mynode, irep) = t2-t1
         flops(mynode, irep) = flop
      ENDDO

      chksum_c_out = dbcsr_checksum(m_c)

      CALL mp_sum(times, 0, mp_group)
      CALL mp_sum(flops, 0, mp_group)

      !
      !
      t_max(:) = MAXVAL(times, DIM=1)
      t_min(:) = MINVAL(times, DIM=1)
      t(:) = SUM(times, DIM=1)/REAL(numnodes, real_8)
      flop_sum(:) = SUM(flops, DIM=1)
      t_max(:) = MAX(t_max(:), 0.001_real_8)
      flops_all(:) = REAL(flop_sum(:), KIND=real_8)/t_max(:) !* 1.0e-9_real_8
      flops_node(:) = flops_all(:)/REAL(numnodes, real_8)
      flops_thread(:) = flops_node(:)/REAL(nthreads, real_8)
      load_imb(:) = t_max(:)-t(:)/REAL(numnodes, real_8)

      std_t = 1.0_real_8
      std_all = 1.0_real_8

      IF (io_unit .GT. 0) THEN
         WRITE (io_unit, *) REPEAT("*", 80)
         WRITE (io_unit, *) " -- PERF dbcsr_multiply (", transa, ", ", transb, &
            ", ", dbcsr_get_data_type(m_a), &
            ", ", dbcsr_get_matrix_type(m_a), &
            ", ", dbcsr_get_matrix_type(m_b), &
            ", ", dbcsr_get_matrix_type(m_c), &
            ")"
         WRITE (io_unit, '(T4,3(A,I6,A,I6),A)') &
            "matrix sizes A(", r_a, " x", c_a, "), B(", r_b, " x", c_b, ") and C(", r_c, " x", c_c, ")"
         WRITE (io_unit, '(T4,A,I5)') 'numnodes        = ', numnodes
         WRITE (io_unit, '(T4,A,I5)') 'nthreads        = ', nthreads
         WRITE (io_unit, '(T4,A,E26.15)') 'checksum(A)     = ', chksum_a
         WRITE (io_unit, '(T4,A,E26.15)') 'checksum(B)     = ', chksum_b
         WRITE (io_unit, '(T4,A,E26.15)') 'checksum(C_in)  = ', chksum_c_in
         WRITE (io_unit, '(T4,A,E26.15)') 'checksum(C_out) = ', chksum_c_out
         WRITE (io_unit, *)
         WRITE (io_unit, *)
         WRITE (io_unit, '(T4,A)') "                       mean        std         minmin      maxmax"
         WRITE (io_unit, '(T4,A,4EN12.2,A)') "time            = ", mean(t), std(t), &
            MINVAL(times), MAXVAL(times), ' seconds'
         WRITE (io_unit, '(T4,A,4EN12.2,A)') "perf total      = ", mean(flops_all), std(flops_all), &
            MINVAL(flops_all), MAXVAL(flops_all), ' FLOPS'
         WRITE (io_unit, '(T4,A,4EN12.2,A)') "perf per node   = ", mean(flops_node), std(flops_node), &
            MINVAL(flops_node), MAXVAL(flops_node), ' FLOPS'
         WRITE (io_unit, '(T4,A,4EN12.2,A)') "perf per thread = ", mean(flops_thread), std(flops_thread), &
            MINVAL(flops_thread), MAXVAL(flops_thread), ' FLOPS'
         WRITE (io_unit, '(T4,A,4E12.2,A)') "load imbalance  = ", mean(load_imb), std(load_imb), &
            MINVAL(load_imb), MAXVAL(load_imb), ''
         WRITE (io_unit, '(T4,A,4E12.2,A)') "rel load imbal  = ", mean(load_imb/t_max), std(load_imb/t_max), &
            MINVAL(load_imb/t_max), MAXVAL(load_imb/t_max), ''

         WRITE (io_unit, *) REPEAT("*", 80)
      ENDIF

      CALL dbcsr_release(m_a)
      CALL dbcsr_release(m_b)
      CALL dbcsr_release(m_c)
      CALL dbcsr_release(m_c_orig)

      DEALLOCATE (times, flops, t_max, flops_node, flops_thread, flops_all, &
                  flop_sum, t_min, t, load_imb)

      CALL timestop(handle)

   END SUBROUTINE perf_multiply

! **************************************************************************************************
!> \brief ...
!> \param v ...
!> \retval mean ...
! **************************************************************************************************
   FUNCTION mean(v)
      REAL(real_8), DIMENSION(:)                         :: v
      REAL(real_8)                                       :: mean

      INTEGER                                            :: i, n

      mean = 0.0_real_8
      n = SIZE(v, 1)
      DO i = 1, n
         mean = mean+v(i)
      ENDDO
      mean = mean/REAL(n, real_8)
   END FUNCTION mean
! **************************************************************************************************
!> \brief ...
!> \param v ...
!> \retval std ...
! **************************************************************************************************
   FUNCTION std(v)
      REAL(real_8), DIMENSION(:)                         :: v
      REAL(real_8)                                       :: std

      INTEGER                                            :: i, n
      REAL(real_8)                                       :: mn

      mn = mean(v)
      std = 0.0_real_8
      n = SIZE(v, 1)
      DO i = 1, n
         std = std+(v(i)-mn)**2
      ENDDO
      std = SQRT(std)/REAL(n, real_8)
   END FUNCTION std

END MODULE dbcsr_performance_multiply
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Tests for DBCSR add
!> \author  VW
!> \date    2010
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created 2010
! **************************************************************************************************
MODULE dbcsr_test_add
   USE dbcsr_data_methods,              ONLY: dbcsr_data_get_sizes,&
                                              dbcsr_data_init,&
                                              dbcsr_data_new,&
                                              dbcsr_data_release,&
                                              dbcsr_type_1d_to_2d
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_new,&
                                              dbcsr_distribution_release
   USE dbcsr_io,                        ONLY: dbcsr_print
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_sizes, dbcsr_get_data_type, dbcsr_get_matrix_type, dbcsr_init, dbcsr_name, &
        dbcsr_nblkcols_total, dbcsr_nblkrows_total, dbcsr_nfullcols_total, dbcsr_nfullrows_total, &
        dbcsr_release, dbcsr_row_block_sizes
   USE dbcsr_operations,                ONLY: dbcsr_add,&
                                              dbcsr_get_occupation
   USE dbcsr_test_methods,              ONLY: compx_to_dbcsr_scalar,&
                                              dbcsr_impose_sparsity,&
                                              dbcsr_make_random_block_sizes,&
                                              dbcsr_make_random_matrix,&
                                              dbcsr_random_dist,&
                                              dbcsr_to_dense_local
   USE dbcsr_transformations,           ONLY: dbcsr_redistribute,&
                                              dbcsr_replicate_all
   USE dbcsr_types,                     ONLY: &
        dbcsr_data_obj, dbcsr_distribution_obj, dbcsr_mp_obj, dbcsr_obj, dbcsr_scalar_type, &
        dbcsr_type_antihermitian, dbcsr_type_antisymmetric, dbcsr_type_complex_4, &
        dbcsr_type_complex_4_2d, dbcsr_type_complex_8, dbcsr_type_complex_8_2d, &
        dbcsr_type_hermitian, dbcsr_type_no_symmetry, dbcsr_type_real_4, dbcsr_type_real_4_2d, &
        dbcsr_type_real_8, dbcsr_type_real_8_2d, dbcsr_type_symmetric
   USE dbcsr_work_operations,           ONLY: dbcsr_create
   USE kinds,                           ONLY: dp,&
                                              real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_environ
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: dbcsr_test_adds

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_test_add'

CONTAINS

! **************************************************************************************************
!> \brief Performs a variety of matrix add
!> \param test_name ...
!> \param[in] mp_group          MPI communicator
!> \param mp_env ...
!> \param npdims ...
!> \param[in] io_unit           which unit to write to, if not negative
!> \param[in] matrix_sizes      size of matrices to test
!> \param[in] bs_m block sizes of the 3 dimensions
!> \param[in] bs_n block sizes of the 3 dimensions
!> \param[in] sparsities        sparsities of matrices to create
!> \param alpha ...
!> \param beta ...
!> \param limits ...
!> \param retain_sparsity ...
! **************************************************************************************************
   SUBROUTINE dbcsr_test_adds(test_name, mp_group, mp_env, npdims, io_unit, &
                              matrix_sizes, bs_m, bs_n, sparsities, &
                              alpha, beta, limits, retain_sparsity)
      CHARACTER(len=*), INTENT(IN)                       :: test_name
      INTEGER, INTENT(IN)                                :: mp_group
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(2), INTENT(in)                  :: npdims
      INTEGER, INTENT(IN)                                :: io_unit
      INTEGER, DIMENSION(:), INTENT(in)                  :: matrix_sizes, bs_m, bs_n
      REAL(real_8), DIMENSION(2), INTENT(in)             :: sparsities
      COMPLEX(real_8)                                    :: alpha, beta
      INTEGER, DIMENSION(4), INTENT(in)                  :: limits
      LOGICAL, INTENT(in)                                :: retain_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_test_adds', &
         routineP = moduleN//':'//routineN
      CHARACTER, DIMENSION(2, 5), PARAMETER :: symmetries = RESHAPE((/dbcsr_type_no_symmetry, &
         dbcsr_type_no_symmetry, dbcsr_type_symmetric, dbcsr_type_symmetric, &
         dbcsr_type_antisymmetric, dbcsr_type_antisymmetric, dbcsr_type_hermitian, &
         dbcsr_type_hermitian, dbcsr_type_antihermitian, dbcsr_type_antihermitian/), (/2, 5/))
      INTEGER, DIMENSION(4), PARAMETER :: types = (/dbcsr_type_real_4, dbcsr_type_real_8, &
         dbcsr_type_complex_4, dbcsr_type_complex_8/)

      CHARACTER                                          :: a_symm, b_symm
      INTEGER                                            :: a_c, a_r, b_c, b_r, handle, isymm, &
                                                            itype, mynode, numnodes, TYPE
      INTEGER, DIMENSION(:), POINTER                     :: my_sizes_m, my_sizes_n, sizes_m, sizes_n
      LOGICAL                                            :: do_complex
      TYPE(dbcsr_data_obj)                               :: data_a, data_a_dbcsr, data_b
      TYPE(dbcsr_obj)                                    :: matrix_a, matrix_b
      TYPE(dbcsr_scalar_type)                            :: alpha_obj, beta_obj

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      NULLIFY (my_sizes_m, my_sizes_n, sizes_m, sizes_n)

      !
      ! print
      CALL mp_environ(numnodes, mynode, mp_group)
      IF (io_unit .GT. 0) THEN
         WRITE (io_unit, *) 'test_name ', test_name
         WRITE (io_unit, *) 'numnodes', numnodes
         WRITE (io_unit, *) 'matrix_sizes', matrix_sizes
         WRITE (io_unit, *) 'sparsities', sparsities
         WRITE (io_unit, *) 'alpha', alpha
         WRITE (io_unit, *) 'beta', beta
         WRITE (io_unit, *) 'limits', limits
         WRITE (io_unit, *) 'retain_sparsity', retain_sparsity
         WRITE (io_unit, *) 'bs_m', bs_m
         WRITE (io_unit, *) 'bs_n', bs_n
      ENDIF
      !
      !
      ! loop over symmetry
      DO isymm = 1, SIZE(symmetries, 2)
         a_symm = symmetries(1, isymm)
         b_symm = symmetries(2, isymm)

         IF (a_symm .NE. dbcsr_type_no_symmetry .AND. matrix_sizes(1) .NE. matrix_sizes(2)) CYCLE
         IF (b_symm .NE. dbcsr_type_no_symmetry .AND. matrix_sizes(1) .NE. matrix_sizes(2)) CYCLE

         !
         ! loop over types
         DO itype = 1, SIZE(types)
            TYPE = types(itype)

            do_complex = TYPE .EQ. dbcsr_type_complex_4 .OR. TYPE .EQ. dbcsr_type_complex_8

            alpha_obj = compx_to_dbcsr_scalar(alpha, TYPE)
            beta_obj = compx_to_dbcsr_scalar(beta, TYPE)
            IF (do_complex .AND. (a_symm .EQ. dbcsr_type_hermitian .OR. a_symm .EQ. dbcsr_type_antihermitian)) &
               alpha_obj = compx_to_dbcsr_scalar(CMPLX(REAL(alpha, kind=dp), 0.0, dp), TYPE)
            IF (do_complex .AND. (b_symm .EQ. dbcsr_type_hermitian .OR. b_symm .EQ. dbcsr_type_antihermitian)) &
               beta_obj = compx_to_dbcsr_scalar(CMPLX(REAL(beta, kind=dp), 0.0, dp), TYPE)

            !
            ! Create the row/column block sizes.
            CALL dbcsr_make_random_block_sizes(sizes_m, matrix_sizes(1), bs_m)
            CALL dbcsr_make_random_block_sizes(sizes_n, matrix_sizes(2), bs_n)

            !
            ! If A/B has symmetry we need the same row/col blocking
            my_sizes_m => sizes_m
            my_sizes_n => sizes_n
            IF (a_symm .NE. dbcsr_type_no_symmetry) THEN
               my_sizes_n => sizes_m
            ENDIF
            IF (b_symm .NE. dbcsr_type_no_symmetry) THEN
               my_sizes_n => sizes_m
            ENDIF

            IF (.FALSE.) THEN
               WRITE (*, *) 'sizes_m', my_sizes_m
               WRITE (*, *) 'sum(sizes_m)', SUM(my_sizes_m), ' matrix_sizes(1)', matrix_sizes(1)
               WRITE (*, *) 'sizes_n', my_sizes_n
               WRITE (*, *) 'sum(sizes_n)', SUM(my_sizes_n), ' matrix_sizes(2)', matrix_sizes(2)
            ENDIF

            !
            ! Create the undistributed matrices.
            CALL dbcsr_make_random_matrix(matrix_a, my_sizes_m, my_sizes_n, "Matrix A", &
                                          sparsities(1), &
                                          mp_group, data_type=TYPE, symmetry=a_symm)

            CALL dbcsr_make_random_matrix(matrix_b, my_sizes_m, my_sizes_n, "Matrix B", &
                                          sparsities(2), &
                                          mp_group, data_type=TYPE, symmetry=b_symm)

            DEALLOCATE (sizes_m, sizes_n)

            !
            ! convert the dbcsr matrices to denses
            a_r = dbcsr_nfullrows_total(matrix_a); a_c = dbcsr_nfullcols_total(matrix_a)
            b_r = dbcsr_nfullrows_total(matrix_b); b_c = dbcsr_nfullcols_total(matrix_b)
            CALL dbcsr_data_init(data_a)
            CALL dbcsr_data_init(data_b)
            CALL dbcsr_data_init(data_a_dbcsr)
            CALL dbcsr_data_new(data_a, dbcsr_type_1d_to_2d(TYPE), data_size=a_r, data_size2=a_c)
            CALL dbcsr_data_new(data_b, dbcsr_type_1d_to_2d(TYPE), data_size=b_r, data_size2=b_c)
            CALL dbcsr_data_new(data_a_dbcsr, dbcsr_type_1d_to_2d(TYPE), data_size=a_r, data_size2=a_c)
            CALL dbcsr_to_dense_local(matrix_a, data_a)
            CALL dbcsr_to_dense_local(matrix_b, data_b)

            IF (.FALSE.) THEN
               CALL dbcsr_print(matrix_a, matlab_format=.TRUE., variable_name='a0')
               CALL dbcsr_print(matrix_b, matlab_format=.TRUE., variable_name='b0')
            ENDIF

            !
            ! Prepare test parameters
            CALL test_add(test_name, mp_group, mp_env, npdims, io_unit, &
                          matrix_a, matrix_b, &
                          data_a, data_b, data_a_dbcsr, &
                          alpha_obj, beta_obj, &
                          limits, retain_sparsity)
            !
            ! cleanup
            CALL dbcsr_release(matrix_a)
            CALL dbcsr_release(matrix_b)
            CALL dbcsr_data_release(data_a)
            CALL dbcsr_data_release(data_b)
            CALL dbcsr_data_release(data_a_dbcsr)

         ENDDO ! itype

      ENDDO !isymm

      CALL timestop(handle)

   END SUBROUTINE dbcsr_test_adds

! **************************************************************************************************
!> \brief Performs a variety of matrix add
!> \param test_name ...
!> \param[in] mp_group          MPI communicator
!> \param mp_env ...
!> \param npdims ...
!> \param[in] io_unit           which unit to write to, if not negative
!> \param[in] matrix_a matrices to add
!> \param[in] matrix_b matrices to add
!> \param data_a ...
!> \param data_b ...
!> \param data_a_dbcsr ...
!> \param alpha ...
!> \param beta ...
!> \param limits ...
!> \param retain_sparsity ...
! **************************************************************************************************
   SUBROUTINE test_add(test_name, mp_group, mp_env, npdims, io_unit, &
                       matrix_a, matrix_b, &
                       data_a, data_b, data_a_dbcsr, &
                       alpha, beta, limits, retain_sparsity)
      CHARACTER(len=*), INTENT(IN)                       :: test_name
      INTEGER, INTENT(IN)                                :: mp_group
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(2), INTENT(in)                  :: npdims
      INTEGER, INTENT(IN)                                :: io_unit
      TYPE(dbcsr_obj), INTENT(in)                        :: matrix_a, matrix_b
      TYPE(dbcsr_data_obj)                               :: data_a, data_b, data_a_dbcsr
      TYPE(dbcsr_scalar_type), INTENT(in)                :: alpha, beta
      INTEGER, DIMENSION(4), INTENT(in)                  :: limits
      LOGICAL, INTENT(in)                                :: retain_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'test_add', routineP = moduleN//':'//routineN

      INTEGER                                            :: c, handle, r
      INTEGER, DIMENSION(:), POINTER                     :: col_dist_a, col_dist_b, row_dist_a, &
                                                            row_dist_b
      LOGICAL                                            :: success
      REAL(real_8)                                       :: occ_a_in, occ_a_out, occ_b
      TYPE(dbcsr_distribution_obj)                       :: dist_a, dist_b
      TYPE(dbcsr_obj)                                    :: m_a, m_b

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      CALL dbcsr_init(m_a)
      CALL dbcsr_init(m_b)

      ! Row & column distributions
      CALL dbcsr_random_dist(row_dist_a, dbcsr_nblkrows_total(matrix_a), npdims(1))
      CALL dbcsr_random_dist(col_dist_a, dbcsr_nblkcols_total(matrix_a), npdims(2))
      CALL dbcsr_random_dist(row_dist_b, dbcsr_nblkrows_total(matrix_b), npdims(1))
      CALL dbcsr_random_dist(col_dist_b, dbcsr_nblkcols_total(matrix_b), npdims(2))

      CALL dbcsr_distribution_new(dist_a, mp_env, row_dist_a, col_dist_a, reuse_arrays=.TRUE.)
      CALL dbcsr_distribution_new(dist_b, mp_env, row_dist_b, col_dist_b, reuse_arrays=.TRUE.)

      ! Redistribute the matrices
      ! A
      CALL dbcsr_create(m_a, "Test for "//TRIM(dbcsr_name(matrix_a)), &
                        dist_a, dbcsr_get_matrix_type(matrix_a), &
                        dbcsr_row_block_sizes(matrix_a), &
                        dbcsr_col_block_sizes(matrix_a), &
                        data_type=dbcsr_get_data_type(matrix_a))
      CALL dbcsr_distribution_release(dist_a)
      CALL dbcsr_redistribute(matrix_a, m_a)
      ! B
      CALL dbcsr_create(m_b, "Test for "//TRIM(dbcsr_name(matrix_b)), &
                        dist_b, dbcsr_get_matrix_type(matrix_b), &
                        dbcsr_row_block_sizes(matrix_b), &
                        dbcsr_col_block_sizes(matrix_b), &
                        data_type=dbcsr_get_data_type(matrix_b))
      CALL dbcsr_distribution_release(dist_b)
      CALL dbcsr_redistribute(matrix_b, m_b)

      IF (.FALSE.) THEN
         CALL dbcsr_print(m_a, matlab_format=.FALSE., variable_name='a_in_')
         CALL dbcsr_print(m_b, matlab_format=.FALSE., variable_name='b_')
      ENDIF

      occ_a_in = dbcsr_get_occupation(m_a)
      occ_b = dbcsr_get_occupation(m_b)

      !
      ! Perform add
      IF (ALL(limits == 0)) THEN
         CALL cp__b("dbcsr/dbcsr_test_add.F",319,"limits shouldnt be 0")
      ELSE
         CALL dbcsr_add(m_a, m_b, alpha, beta)
      ENDIF

      occ_a_out = dbcsr_get_occupation(m_a)

      IF (.FALSE.) THEN
         PRINT*, 'retain_sparsity', retain_sparsity, occ_a_in, occ_b, occ_a_out
         CALL dbcsr_print(m_a, matlab_format=.TRUE., variable_name='a_out_')
      ENDIF

      CALL dbcsr_replicate_all(m_a)
      CALL dbcsr_to_dense_local(m_a, data_a_dbcsr)
      CALL dbcsr_check_add(test_name, m_a, data_a_dbcsr, data_a, data_b, &
                           alpha, beta, limits, retain_sparsity, io_unit, mp_group, &
                           success)

      r = dbcsr_nfullrows_total(m_a)
      c = dbcsr_nfullcols_total(m_a)

      IF (io_unit .GT. 0) THEN
         IF (success) THEN
            WRITE (io_unit, *) REPEAT("*", 70)
            WRITE (io_unit, *) " -- TESTING dbcsr_add (", &
               dbcsr_get_data_type(m_a), &
               ", ", dbcsr_get_matrix_type(m_a), &
               ", ", dbcsr_get_matrix_type(m_b), &
               ") ............................. PASSED !"
            WRITE (io_unit, *) REPEAT("*", 70)
         ELSE
            WRITE (io_unit, *) REPEAT("*", 70)
            WRITE (io_unit, *) " -- TESTING dbcsr_add (", &
               dbcsr_get_data_type(m_a), &
               ", ", dbcsr_get_matrix_type(m_a), &
               ", ", dbcsr_get_matrix_type(m_b), &
               ") ................. FAILED !"
            WRITE (io_unit, *) REPEAT("*", 70)
            CALL cp__b("dbcsr/dbcsr_test_add.F",357,'Test failed')
         ENDIF
      ENDIF

      CALL dbcsr_release(m_a)
      CALL dbcsr_release(m_b)

      CALL timestop(handle)

   END SUBROUTINE test_add

! **************************************************************************************************
!> \brief Performs a check of matrix adds
!> \param test_name ...
!> \param matrix_a ...
!> \param dense_a_dbcsr ...
!> \param dense_a input dense matrices
!> \param dense_b input dense matrices
!> \param[in] alpha coefficients for the add
!> \param[in] beta coefficients for the add
!> \param[in] limits                    limits for the add
!> \param retain_sparsity ...
!> \param[in] io_unit                   io unit for printing
!> \param mp_group ...
!> \param[out] success                  if passed the check success=T
! **************************************************************************************************
   SUBROUTINE dbcsr_check_add(test_name, matrix_a, dense_a_dbcsr, dense_a, dense_b, &
                              alpha, beta, limits, retain_sparsity, io_unit, mp_group, &
                              success)
      CHARACTER(len=*), INTENT(IN)                       :: test_name
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a
      TYPE(dbcsr_data_obj), INTENT(inout)                :: dense_a_dbcsr, dense_a, dense_b
      TYPE(dbcsr_scalar_type), INTENT(in)                :: alpha, beta
      INTEGER, DIMENSION(4), INTENT(in)                  :: limits
      LOGICAL, INTENT(in)                                :: retain_sparsity
      INTEGER, INTENT(IN)                                :: io_unit, mp_group
      LOGICAL, INTENT(out)                               :: success

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_check_add', &
         routineP = moduleN//':'//routineN
      INTEGER                                            :: col, col_size, handle, i, istat, j, ld, &
                                                            lwork, m, mynode, n, numnodes, row, &
                                                            row_size
      CHARACTER, PARAMETER                               :: norm = 'I'

      LOGICAL                                            :: valid
      REAL(real_4), ALLOCATABLE, DIMENSION(:)            :: work_sp
      REAL(real_4), EXTERNAL                             :: clange, slamch, slange
      REAL(real_8)                                       :: a_norm_dbcsr, a_norm_in, a_norm_out, &
                                                            b_norm, eps, residual
      REAL(real_8), ALLOCATABLE, DIMENSION(:)            :: work
      REAL(real_8), EXTERNAL                             :: dlamch, dlange, zlange

      CALL timeset(routineN, handle)

      CALL mp_environ(numnodes, mynode, mp_group)

      CALL dbcsr_data_get_sizes(dense_a, row_size, col_size, valid)
      IF (.NOT. valid) &
         CALL cp__b("dbcsr/dbcsr_test_add.F",416,"dense matrix not valid")
      !
      !
      m = limits(2)-limits(1)+1
      n = limits(4)-limits(3)+1
      row = limits(1); col = limits(3)

      !
      ! set the size of the work array
      lwork = row_size
      ld = row_size
      !
      !
      SELECT CASE (dense_a%d%data_type)
      CASE (dbcsr_type_real_8_2d)
         ALLOCATE (work(lwork), STAT=istat)
         IF (istat /= 0) &
            CALL cp__b("dbcsr/dbcsr_test_add.F",433,"allocattion problem")
         eps = dlamch('eps')
         a_norm_in = dlange(norm, row_size, col_size, dense_a%d%r2_dp(1, 1), ld, work)
         b_norm = dlange(norm, row_size, col_size, dense_b%d%r2_dp(1, 1), ld, work)
         a_norm_dbcsr = dlange(norm, row_size, col_size, dense_a_dbcsr%d%r2_dp(1, 1), ld, work)
         !
         dense_a%d%r2_dp(row:row+m-1, col:col+n-1) = alpha%r_dp*dense_a%d%r2_dp(row:row+m-1, col:col+n-1)+ &
                                                     beta%r_dp*dense_b%d%r2_dp(row:row+m-1, col:col+n-1)
         !
         ! impose the sparsity if needed
         IF (retain_sparsity) CALL dbcsr_impose_sparsity(matrix_a, dense_a)
         !
         a_norm_out = dlange(norm, row_size, col_size, dense_a%d%r2_dp(1, 1), ld, work)
         !
         ! take the difference dense/sparse
         dense_a%d%r2_dp = dense_a%d%r2_dp-dense_a_dbcsr%d%r2_dp
         !
         ! compute the residual
         residual = dlange(norm, row_size, col_size, dense_a%d%r2_dp(1, 1), ld, work)
         DEALLOCATE (work)
      CASE (dbcsr_type_real_4_2d)
         ALLOCATE (work_sp(lwork), STAT=istat)
         IF (istat /= 0) &
            CALL cp__b("dbcsr/dbcsr_test_add.F",456,"allocattion problem")
         eps = slamch('eps')
         a_norm_in = slange(norm, row_size, col_size, dense_a%d%r2_sp(1, 1), ld, work_sp)
         b_norm = slange(norm, row_size, col_size, dense_b%d%r2_sp(1, 1), ld, work_sp)
         a_norm_dbcsr = slange(norm, row_size, col_size, dense_a_dbcsr%d%r2_sp(1, 1), ld, work_sp)
         !
         dense_a%d%r2_sp(row:row+m-1, col:col+n-1) = alpha%r_sp*dense_a%d%r2_sp(row:row+m-1, col:col+n-1)+ &
                                                     beta%r_sp*dense_b%d%r2_sp(row:row+m-1, col:col+n-1)
         !
         ! impose the sparsity if needed
         IF (retain_sparsity) CALL dbcsr_impose_sparsity(matrix_a, dense_a)
         !
         a_norm_out = slange(norm, row_size, col_size, dense_a%d%r2_sp(1, 1), ld, work_sp)
         !
         ! take the difference dense/sparse
         dense_a%d%r2_sp = dense_a%d%r2_sp-dense_a_dbcsr%d%r2_sp
         !
         ! compute the residual
         residual = REAL(slange(norm, row_size, col_size, dense_a%d%r2_sp(1, 1), ld, work_sp), real_8)
         DEALLOCATE (work_sp)
      CASE (dbcsr_type_complex_8_2d)
         ALLOCATE (work(lwork), STAT=istat)
         IF (istat /= 0) &
            CALL cp__b("dbcsr/dbcsr_test_add.F",479,"allocattion problem")
         eps = dlamch('eps')
         a_norm_in = zlange(norm, row_size, col_size, dense_a%d%c2_dp(1, 1), ld, work)
         b_norm = zlange(norm, row_size, col_size, dense_b%d%c2_dp(1, 1), ld, work)
         a_norm_dbcsr = zlange(norm, row_size, col_size, dense_a_dbcsr%d%c2_dp(1, 1), ld, work)
         !
         dense_a%d%c2_dp(row:row+m-1, col:col+n-1) = alpha%c_dp*dense_a%d%c2_dp(row:row+m-1, col:col+n-1)+ &
                                                     beta%c_dp*dense_b%d%c2_dp(row:row+m-1, col:col+n-1)
         !
         ! impose the sparsity if needed
         IF (retain_sparsity) CALL dbcsr_impose_sparsity(matrix_a, dense_a)
         !
         a_norm_out = zlange(norm, row_size, col_size, dense_a%d%c2_dp(1, 1), ld, work)
         !
         ! take the difference dense/sparse
         dense_a%d%c2_dp = dense_a%d%c2_dp-dense_a_dbcsr%d%c2_dp
         !
         ! compute the residual
         residual = zlange(norm, row_size, col_size, dense_a%d%c2_dp(1, 1), ld, work)
         DEALLOCATE (work)
      CASE (dbcsr_type_complex_4_2d)
         ALLOCATE (work_sp(lwork), STAT=istat)
         IF (istat /= 0) &
            CALL cp__b("dbcsr/dbcsr_test_add.F",502,"allocattion problem")
         eps = REAL(slamch('eps'), real_8)
         a_norm_in = clange(norm, row_size, col_size, dense_a%d%c2_sp(1, 1), ld, work_sp)
         b_norm = clange(norm, row_size, col_size, dense_b%d%c2_sp(1, 1), ld, work_sp)
         a_norm_dbcsr = clange(norm, row_size, col_size, dense_a_dbcsr%d%c2_sp(1, 1), ld, work_sp)
         !
         IF (.FALSE.) THEN
            !IF(io_unit .GT. 0) THEN
            DO j = 1, SIZE(dense_a%d%c2_sp, 2)
            DO i = 1, SIZE(dense_a%d%c2_sp, 1)
               WRITE (*, '(A,I3,A,I3,A,E15.7,A,E15.7,A)') 'a_in(', i, ',', j, ')=', REAL(dense_a%d%c2_sp(i, j)), '+', &
                  AIMAG(dense_a%d%c2_sp(i, j)), 'i;'
            ENDDO
            ENDDO
            DO j = 1, SIZE(dense_b%d%c2_sp, 2)
            DO i = 1, SIZE(dense_b%d%c2_sp, 1)
               WRITE (*, '(A,I3,A,I3,A,E15.7,A,E15.7,A)') 'b(', i, ',', j, ')=', REAL(dense_b%d%c2_sp(i, j)), '+', &
                  AIMAG(dense_b%d%c2_sp(i, j)), 'i;'
            ENDDO
            ENDDO
         ENDIF

         dense_a%d%c2_sp(row:row+m-1, col:col+n-1) = alpha%c_sp*dense_a%d%c2_sp(row:row+m-1, col:col+n-1)+ &
                                                     beta%c_sp*dense_b%d%c2_sp(row:row+m-1, col:col+n-1)
         !
         ! impose the sparsity if needed
         IF (retain_sparsity) CALL dbcsr_impose_sparsity(matrix_a, dense_a)
         !

         IF (.FALSE.) THEN
            !IF(io_unit .GT. 0) THEN
            DO j = 1, SIZE(dense_a%d%c2_sp, 2)
            DO i = 1, SIZE(dense_a%d%c2_sp, 1)
               WRITE (*, '(A,I3,A,I3,A,E15.7,A,E15.7,A)') 'a_out(', i, ',', j, ')=', REAL(dense_a%d%c2_sp(i, j)), '+', &
                  AIMAG(dense_a%d%c2_sp(i, j)), 'i;'
            ENDDO
            ENDDO
            DO j = 1, SIZE(dense_a_dbcsr%d%c2_sp, 2)
            DO i = 1, SIZE(dense_a_dbcsr%d%c2_sp, 1)
               WRITE (*, '(A,I3,A,I3,A,E15.7,A,E15.7,A)') 'a_dbcsr(', i, ',', j, ')=', REAL(dense_a_dbcsr%d%c2_sp(i, j)), '+', &
                  AIMAG(dense_a_dbcsr%d%c2_sp(i, j)), 'i;'
            ENDDO
            ENDDO
         ENDIF

         a_norm_out = clange(norm, row_size, col_size, dense_a%d%c2_sp(1, 1), ld, work_sp)
         !
         ! take the difference dense/sparse
         dense_a%d%c2_sp = dense_a%d%c2_sp-dense_a_dbcsr%d%c2_sp
         !
         ! compute the residual
         residual = REAL(clange(norm, row_size, col_size, dense_a%d%c2_sp(1, 1), ld, work_sp), real_8)
         DEALLOCATE (work_sp)
      CASE default
         CALL cp__b("dbcsr/dbcsr_test_add.F",556,"Incorrect or 1-D data type")
      END SELECT

      IF (mynode .EQ. 0) THEN
         IF (residual/((a_norm_in+b_norm)*REAL(row_size, real_8)*eps) .GT. 10.0_real_8) THEN
            success = .FALSE.
         ELSE
            success = .TRUE.
         ENDIF
      ENDIF
      !
      ! syncronize the result...
      CALL mp_bcast(success, 0, mp_group)
      !
      ! printing
      IF (io_unit .GT. 0) THEN
         WRITE (io_unit, *) 'test_name ', test_name
         WRITE (io_unit, '(2(A,E12.5))') ' residual ', residual, ', b_norm ', b_norm
         WRITE (io_unit, '(3(A,E12.5))') ' a_norm_in ', a_norm_in, ', a_norm_out ', a_norm_out, &
            ', a_norm_dbcsr ', a_norm_dbcsr
         WRITE (io_unit, '(A)') ' Checking the norm of the difference against reference ADD '
         WRITE (io_unit, '(A,E12.5)') ' -- ||A_dbcsr-A_dense||_oo/((||A||_oo+||B||_oo).N.eps)=', &
            residual/((a_norm_in+b_norm)*n*eps)
         !
         ! check for nan or inf here
         IF (success) THEN
            WRITE (io_unit, '(A)') ' The solution is CORRECT !'
         ELSE
            WRITE (io_unit, '(A)') ' The solution is suspicious !'
         ENDIF

      ENDIF

      CALL timestop(handle)

   END SUBROUTINE dbcsr_check_add

END MODULE dbcsr_test_add
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Tests for DBCSR multiply
!> \author  VW
!> \date    2010
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created 2010
! **************************************************************************************************
MODULE dbcsr_test_multiply
   USE dbcsr_data_methods,              ONLY: dbcsr_data_get_sizes,&
                                              dbcsr_data_init,&
                                              dbcsr_data_new,&
                                              dbcsr_data_release,&
                                              dbcsr_scalar_negative,&
                                              dbcsr_scalar_one,&
                                              dbcsr_type_1d_to_2d
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_new,&
                                              dbcsr_distribution_release
   USE dbcsr_io,                        ONLY: dbcsr_print
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_offsets, dbcsr_col_block_sizes, dbcsr_get_data_type, &
        dbcsr_get_matrix_type, dbcsr_init, dbcsr_name, dbcsr_nblkcols_total, dbcsr_nblkrows_total, &
        dbcsr_nfullcols_total, dbcsr_nfullrows_total, dbcsr_release, dbcsr_row_block_offsets, &
        dbcsr_row_block_sizes
   USE dbcsr_multiply_api,              ONLY: dbcsr_multiply
   USE dbcsr_operations,                ONLY: dbcsr_copy,&
                                              dbcsr_get_occupation,&
                                              dbcsr_scale
   USE dbcsr_test_methods,              ONLY: compx_to_dbcsr_scalar,&
                                              dbcsr_impose_sparsity,&
                                              dbcsr_make_random_block_sizes,&
                                              dbcsr_make_random_matrix,&
                                              dbcsr_random_dist,&
                                              dbcsr_to_dense_local
   USE dbcsr_transformations,           ONLY: dbcsr_redistribute,&
                                              dbcsr_replicate_all
   USE dbcsr_types,                     ONLY: &
        dbcsr_conjugate_transpose, dbcsr_data_obj, dbcsr_distribution_obj, dbcsr_mp_obj, &
        dbcsr_no_transpose, dbcsr_obj, dbcsr_scalar_type, dbcsr_transpose, &
        dbcsr_type_antisymmetric, dbcsr_type_complex_4, dbcsr_type_complex_4_2d, &
        dbcsr_type_complex_8, dbcsr_type_complex_8_2d, dbcsr_type_no_symmetry, dbcsr_type_real_4, &
        dbcsr_type_real_4_2d, dbcsr_type_real_8, dbcsr_type_real_8_2d, dbcsr_type_symmetric
   USE dbcsr_work_operations,           ONLY: dbcsr_create
   USE kinds,                           ONLY: real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_environ
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: dbcsr_test_multiplies

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_test_multiply'

   LOGICAL, PARAMETER :: debug_mod = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief Performs a variety of matrix multiplies of same matrices on different
!>        processor grids
!> \param test_name ...
!> \param[in] mp_group          MPI communicator
!> \param mp_env ...
!> \param npdims ...
!> \param[in] io_unit           which unit to write to, if not negative
!> \param[in] matrix_sizes      size of matrices to test
!> \param bs_m block sizes of the 3 dimension
!> \param bs_n block sizes of the 3 dimension
!> \param bs_k block sizes of the 3 dimension
!> \param[in] sparsities        sparsities of matrices to create
!> \param alpha alpha value to use in multiply
!> \param beta beta value to use in multiply
!> \param limits ...
!> \param retain_sparsity ...
! **************************************************************************************************
   SUBROUTINE dbcsr_test_multiplies(test_name, mp_group, mp_env, npdims, io_unit, &
                                    matrix_sizes, bs_m, bs_n, bs_k, sparsities, &
                                    alpha, beta, limits, retain_sparsity)
      CHARACTER(len=*), INTENT(IN)                       :: test_name
      INTEGER, INTENT(IN)                                :: mp_group
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(2), INTENT(in)                  :: npdims
      INTEGER, INTENT(IN)                                :: io_unit
      INTEGER, DIMENSION(:), INTENT(in)                  :: matrix_sizes, bs_m, bs_n, bs_k
      REAL(real_8), DIMENSION(3), INTENT(in)             :: sparsities
      COMPLEX(real_8), INTENT(in)                        :: alpha, beta
      INTEGER, DIMENSION(6), INTENT(in)                  :: limits
      LOGICAL, INTENT(in)                                :: retain_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_test_multiplies', &
         routineP = moduleN//':'//routineN
      CHARACTER, DIMENSION(3), PARAMETER :: &
         trans = (/dbcsr_no_transpose, dbcsr_transpose, dbcsr_conjugate_transpose/)
      CHARACTER, DIMENSION(3, 12), PARAMETER :: symmetries = RESHAPE((/dbcsr_type_no_symmetry, &
         dbcsr_type_no_symmetry, dbcsr_type_no_symmetry, dbcsr_type_symmetric, &
         dbcsr_type_no_symmetry, dbcsr_type_no_symmetry, dbcsr_type_antisymmetric, &
         dbcsr_type_no_symmetry, dbcsr_type_no_symmetry, dbcsr_type_no_symmetry, &
         dbcsr_type_symmetric, dbcsr_type_no_symmetry, dbcsr_type_symmetric, dbcsr_type_symmetric, &
         dbcsr_type_no_symmetry, dbcsr_type_antisymmetric, dbcsr_type_symmetric, &
         dbcsr_type_no_symmetry, dbcsr_type_no_symmetry, dbcsr_type_antisymmetric, &
         dbcsr_type_no_symmetry, dbcsr_type_symmetric, dbcsr_type_antisymmetric, &
         dbcsr_type_no_symmetry, dbcsr_type_antisymmetric, dbcsr_type_antisymmetric, &
         dbcsr_type_no_symmetry, dbcsr_type_no_symmetry, dbcsr_type_no_symmetry, &
         dbcsr_type_symmetric, dbcsr_type_symmetric, dbcsr_type_symmetric, dbcsr_type_symmetric, &
         dbcsr_type_antisymmetric, dbcsr_type_antisymmetric, dbcsr_type_symmetric/), (/3, 12/))
      INTEGER, DIMENSION(4), PARAMETER :: types = (/dbcsr_type_real_4, dbcsr_type_real_8, &
         dbcsr_type_complex_4, dbcsr_type_complex_8/)

      CHARACTER                                          :: a_symm, b_symm, c_symm, transa, transb
      INTEGER                                            :: a_c, a_r, a_tr, b_c, b_r, b_tr, c_c, &
                                                            c_r, handle, isymm, itype, mynode, &
                                                            numnodes, TYPE
      INTEGER, DIMENSION(:), POINTER                     :: my_sizes_k, my_sizes_m, my_sizes_n, &
                                                            sizes_k, sizes_m, sizes_n
      LOGICAL                                            :: do_complex
      TYPE(dbcsr_data_obj)                               :: data_a, data_b, data_c, data_c_dbcsr
      TYPE(dbcsr_obj)                                    :: matrix_a, matrix_b, matrix_c
      TYPE(dbcsr_scalar_type)                            :: alpha_obj, beta_obj

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      NULLIFY (my_sizes_k, my_sizes_m, my_sizes_n, &
               sizes_k, sizes_m, sizes_n)
      !
      ! print
      CALL mp_environ(numnodes, mynode, mp_group)
      IF (io_unit .GT. 0) THEN
         WRITE (io_unit, *) 'test_name ', test_name
         WRITE (io_unit, *) 'numnodes', numnodes
         WRITE (io_unit, *) 'matrix_sizes', matrix_sizes
         WRITE (io_unit, *) 'sparsities', sparsities
         WRITE (io_unit, *) 'alpha', alpha
         WRITE (io_unit, *) 'beta', beta
         WRITE (io_unit, *) 'limits', limits
         WRITE (io_unit, *) 'retain_sparsity', retain_sparsity
         WRITE (io_unit, *) 'bs_m', bs_m
         WRITE (io_unit, *) 'bs_n', bs_n
         WRITE (io_unit, *) 'bs_k', bs_k
      END IF
      !
      !
      ! loop over symmetry
      DO isymm = 1, SIZE(symmetries, 2)
         a_symm = symmetries(1, isymm)
         b_symm = symmetries(2, isymm)
         c_symm = symmetries(3, isymm)

         IF (a_symm .NE. dbcsr_type_no_symmetry .AND. matrix_sizes(1) .NE. matrix_sizes(3)) CYCLE
         IF (b_symm .NE. dbcsr_type_no_symmetry .AND. matrix_sizes(2) .NE. matrix_sizes(3)) CYCLE
         IF (c_symm .NE. dbcsr_type_no_symmetry .AND. matrix_sizes(1) .NE. matrix_sizes(2)) CYCLE

         !
         ! loop over types
         DO itype = 1, SIZE(types)
            TYPE = types(itype)

            do_complex = TYPE .EQ. dbcsr_type_complex_4 .OR. TYPE .EQ. dbcsr_type_complex_8

            alpha_obj = compx_to_dbcsr_scalar(alpha, TYPE)
            beta_obj = compx_to_dbcsr_scalar(beta, TYPE)

            IF (do_complex .AND. c_symm == dbcsr_type_symmetric) CYCLE

            !
            ! loop over transpositions
            DO a_tr = 1, SIZE(trans)
            DO b_tr = 1, SIZE(trans)
               transa = trans(a_tr)
               transb = trans(b_tr)

               !
               ! if C has a symmetry, we need special transpositions
               IF (c_symm .NE. dbcsr_type_no_symmetry) THEN
                  IF (.NOT. (transa .EQ. dbcsr_no_transpose .AND. transb .EQ. dbcsr_transpose .OR. &
                             transa .EQ. dbcsr_transpose .AND. transb .EQ. dbcsr_no_transpose .OR. &
                             transa .EQ. dbcsr_no_transpose .AND. transb .EQ. dbcsr_conjugate_transpose .AND. &
                             .NOT. do_complex .OR. &
                             transa .EQ. dbcsr_conjugate_transpose .AND. transb .EQ. dbcsr_no_transpose .AND. &
                             .NOT. do_complex)) CYCLE
               END IF
               !
               ! if C has symmetry and special limits
               IF (c_symm .NE. dbcsr_type_no_symmetry) THEN
                  IF (limits(1) .NE. 1 .OR. limits(2) .NE. matrix_sizes(1) .OR. &
                      limits(3) .NE. 1 .OR. limits(4) .NE. matrix_sizes(2)) CYCLE
               END IF

               !
               ! Create the row/column block sizes.
               CALL dbcsr_make_random_block_sizes(sizes_m, matrix_sizes(1), bs_m)
               CALL dbcsr_make_random_block_sizes(sizes_n, matrix_sizes(2), bs_n)
               CALL dbcsr_make_random_block_sizes(sizes_k, matrix_sizes(3), bs_k)

               !
               ! if we have symmetry the row and column block sizes hae to match
               IF (c_symm .NE. dbcsr_type_no_symmetry .AND. a_symm .NE. dbcsr_type_no_symmetry .AND. &
                   b_symm .NE. dbcsr_type_no_symmetry) THEN
                  my_sizes_m => sizes_m
                  my_sizes_n => sizes_m
                  my_sizes_k => sizes_m
               ELSE IF ((c_symm .EQ. dbcsr_type_no_symmetry .AND. a_symm .NE. dbcsr_type_no_symmetry .AND. &
                         b_symm .NE. dbcsr_type_no_symmetry) .OR. &
                        (c_symm .NE. dbcsr_type_no_symmetry .AND. a_symm .EQ. dbcsr_type_no_symmetry .AND. &
                         b_symm .NE. dbcsr_type_no_symmetry) .OR. &
                        (c_symm .NE. dbcsr_type_no_symmetry .AND. a_symm .NE. dbcsr_type_no_symmetry .AND. &
                         b_symm .EQ. dbcsr_type_no_symmetry)) THEN
                  my_sizes_m => sizes_m
                  my_sizes_n => sizes_m
                  my_sizes_k => sizes_m
               ELSE IF (c_symm .EQ. dbcsr_type_no_symmetry .AND. a_symm .EQ. dbcsr_type_no_symmetry .AND. &
                        b_symm .NE. dbcsr_type_no_symmetry) THEN
                  my_sizes_m => sizes_m
                  my_sizes_n => sizes_n
                  my_sizes_k => sizes_n
               ELSE IF (c_symm .EQ. dbcsr_type_no_symmetry .AND. a_symm .NE. dbcsr_type_no_symmetry .AND. &
                        b_symm .EQ. dbcsr_type_no_symmetry) THEN
                  my_sizes_m => sizes_m
                  my_sizes_n => sizes_n
                  my_sizes_k => sizes_m
               ELSE IF (c_symm .NE. dbcsr_type_no_symmetry .AND. a_symm .EQ. dbcsr_type_no_symmetry .AND. &
                        b_symm .EQ. dbcsr_type_no_symmetry) THEN
                  my_sizes_m => sizes_m
                  my_sizes_n => sizes_m
                  my_sizes_k => sizes_k
               ELSE IF (c_symm .EQ. dbcsr_type_no_symmetry .AND. a_symm .EQ. dbcsr_type_no_symmetry .AND. &
                        b_symm .EQ. dbcsr_type_no_symmetry) THEN
                  my_sizes_m => sizes_m
                  my_sizes_n => sizes_n
                  my_sizes_k => sizes_k
               ELSE
                  CALL cp__b("dbcsr/dbcsr_test_multiply.F",244," something wrong here... ")
               END IF

               IF (.FALSE.) THEN
                  WRITE (*, *) 'sizes_m', my_sizes_m
                  WRITE (*, *) 'sum(sizes_m)', SUM(my_sizes_m), ' matrix_sizes(1)', matrix_sizes(1)
                  WRITE (*, *) 'sizes_n', my_sizes_n
                  WRITE (*, *) 'sum(sizes_n)', SUM(my_sizes_n), ' matrix_sizes(2)', matrix_sizes(2)
                  WRITE (*, *) 'sizes_k', my_sizes_k
                  WRITE (*, *) 'sum(sizes_k)', SUM(my_sizes_k), ' matrix_sizes(3)', matrix_sizes(3)
               END IF

               !
               ! Create the undistributed matrices.
               CALL dbcsr_make_random_matrix(matrix_c, my_sizes_m, my_sizes_n, "Matrix C", &
                                             sparsities(3), &
                                             mp_group, data_type=TYPE, symmetry=c_symm)

               IF (transa .NE. dbcsr_no_transpose) THEN
                  CALL dbcsr_make_random_matrix(matrix_a, my_sizes_k, my_sizes_m, "Matrix A", &
                                                sparsities(1), &
                                                mp_group, data_type=TYPE, symmetry=a_symm)
               ELSE
                  CALL dbcsr_make_random_matrix(matrix_a, my_sizes_m, my_sizes_k, "Matrix A", &
                                                sparsities(1), &
                                                mp_group, data_type=TYPE, symmetry=a_symm)
               END IF
               IF (transb .NE. dbcsr_no_transpose) THEN
                  CALL dbcsr_make_random_matrix(matrix_b, my_sizes_n, my_sizes_k, "Matrix B", &
                                                sparsities(2), &
                                                mp_group, data_type=TYPE, symmetry=b_symm)
               ELSE
                  CALL dbcsr_make_random_matrix(matrix_b, my_sizes_k, my_sizes_n, "Matrix B", &
                                                sparsities(2), &
                                                mp_group, data_type=TYPE, symmetry=b_symm)
               END IF

               DEALLOCATE (sizes_m, sizes_n, sizes_k)

               !
               ! if C has a symmetry, we build it accordingly, i.e. C=A*A and C=A*(-A)
               IF (c_symm .NE. dbcsr_type_no_symmetry) THEN
                  CALL dbcsr_copy(matrix_b, matrix_a)
                  !print*, a_symm,b_symm,dbcsr_get_matrix_type(matrix_a),dbcsr_get_matrix_type(matrix_b)
                  IF (c_symm .EQ. dbcsr_type_antisymmetric) THEN
                     CALL dbcsr_scale(matrix_b, &
                                      alpha_scalar=dbcsr_scalar_negative( &
                                      dbcsr_scalar_one(TYPE)))
                  END IF
               END IF

               !
               ! convert the dbcsr matrices to denses
               a_r = dbcsr_nfullrows_total(matrix_a)
               a_c = dbcsr_nfullcols_total(matrix_a)
               b_r = dbcsr_nfullrows_total(matrix_b)
               b_c = dbcsr_nfullcols_total(matrix_b)
               c_r = dbcsr_nfullrows_total(matrix_c)
               c_c = dbcsr_nfullcols_total(matrix_c)
               CALL dbcsr_data_init(data_a)
               CALL dbcsr_data_init(data_b)
               CALL dbcsr_data_init(data_c)
               CALL dbcsr_data_init(data_c_dbcsr)
               CALL dbcsr_data_new(data_a, dbcsr_type_1d_to_2d(TYPE), data_size=a_r, data_size2=a_c)
               CALL dbcsr_data_new(data_b, dbcsr_type_1d_to_2d(TYPE), data_size=b_r, data_size2=b_c)
               CALL dbcsr_data_new(data_c, dbcsr_type_1d_to_2d(TYPE), data_size=c_r, data_size2=c_c)
               CALL dbcsr_data_new(data_c_dbcsr, dbcsr_type_1d_to_2d(TYPE), data_size=c_r, data_size2=c_c)
               CALL dbcsr_to_dense_local(matrix_a, data_a)
               CALL dbcsr_to_dense_local(matrix_b, data_b)
               CALL dbcsr_to_dense_local(matrix_c, data_c)

               !
               ! Prepare test parameters
               CALL test_multiply(test_name, mp_group, mp_env, npdims, io_unit, &
                                  matrix_a, matrix_b, matrix_c, &
                                  data_a, data_b, data_c, data_c_dbcsr, &
                                  transa, transb, &
                                  alpha_obj, beta_obj, &
                                  limits, retain_sparsity)
               !
               ! cleanup
               CALL dbcsr_release(matrix_a)
               CALL dbcsr_release(matrix_b)
               CALL dbcsr_release(matrix_c)
               CALL dbcsr_data_release(data_a)
               CALL dbcsr_data_release(data_b)
               CALL dbcsr_data_release(data_c)
               CALL dbcsr_data_release(data_c_dbcsr)

            END DO
            END DO

         END DO ! itype

      END DO !isymm

      CALL timestop(handle)

   END SUBROUTINE dbcsr_test_multiplies

! **************************************************************************************************
!> \brief Performs a variety of matrix multiplies of same matrices on different
!>        processor grids
!> \param test_name ...
!> \param[in] mp_group          MPI communicator
!> \param mp_env ...
!> \param npdims ...
!> \param[in] io_unit           which unit to write to, if not negative
!> \param[in] matrix_a matrices to multiply
!> \param[in] matrix_b matrices to multiply
!> \param[in] matrix_c matrices to multiply
!> \param data_a ...
!> \param data_b ...
!> \param data_c ...
!> \param data_c_dbcsr ...
!> \param transa ...
!> \param transb ...
!> \param alpha ...
!> \param beta ...
!> \param limits ...
!> \param retain_sparsity ...
! **************************************************************************************************
   SUBROUTINE test_multiply(test_name, mp_group, mp_env, npdims, io_unit, &
                            matrix_a, matrix_b, matrix_c, &
                            data_a, data_b, data_c, data_c_dbcsr, &
                            transa, transb, alpha, beta, limits, retain_sparsity)
      CHARACTER(len=*), INTENT(IN)                       :: test_name
      INTEGER, INTENT(IN)                                :: mp_group
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(2), INTENT(in)                  :: npdims
      INTEGER, INTENT(IN)                                :: io_unit
      TYPE(dbcsr_obj), INTENT(in)                        :: matrix_a, matrix_b, matrix_c
      TYPE(dbcsr_data_obj)                               :: data_a, data_b, data_c, data_c_dbcsr
      CHARACTER, INTENT(in)                              :: transa, transb
      TYPE(dbcsr_scalar_type), INTENT(in)                :: alpha, beta
      INTEGER, DIMENSION(6), INTENT(in)                  :: limits
      LOGICAL, INTENT(in)                                :: retain_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'test_multiply', routineP = moduleN//':'//routineN

      INTEGER                                            :: c_a, c_b, c_c, handle, r_a, r_b, r_c
      INTEGER, DIMENSION(:), POINTER                     :: blk_offsets, col_dist_a, col_dist_b, &
                                                            col_dist_c, row_dist_a, row_dist_b, &
                                                            row_dist_c
      LOGICAL                                            :: success
      REAL(real_8)                                       :: occ_a, occ_b, occ_c_in, occ_c_out
      TYPE(dbcsr_distribution_obj)                       :: dist_a, dist_b, dist_c
      TYPE(dbcsr_obj)                                    :: m_a, m_b, m_c

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      NULLIFY (row_dist_a, col_dist_a, &
               row_dist_b, col_dist_b, &
               row_dist_c, col_dist_c)

      IF (debug_mod .AND. io_unit .GT. 0) THEN
         WRITE (io_unit, *) REPEAT("*", 70)
         WRITE (io_unit, *) " -- TESTING dbcsr_multiply (", transa, ", ", transb, &
            ", ", dbcsr_get_data_type(m_a), &
            ", ", dbcsr_get_matrix_type(m_a), &
            ", ", dbcsr_get_matrix_type(m_b), &
            ", ", dbcsr_get_matrix_type(m_c), &
            ") ............... !"
         WRITE (io_unit, *) REPEAT("*", 70)
      END IF

      CALL dbcsr_init(m_a)
      CALL dbcsr_init(m_b)
      CALL dbcsr_init(m_c)

      ! Row & column distributions
      CALL dbcsr_random_dist(row_dist_a, dbcsr_nblkrows_total(matrix_a), npdims(1))
      CALL dbcsr_random_dist(col_dist_a, dbcsr_nblkcols_total(matrix_a), npdims(2))
      CALL dbcsr_random_dist(row_dist_b, dbcsr_nblkrows_total(matrix_b), npdims(1))
      CALL dbcsr_random_dist(col_dist_b, dbcsr_nblkcols_total(matrix_b), npdims(2))
      CALL dbcsr_random_dist(row_dist_c, dbcsr_nblkrows_total(matrix_c), npdims(1))
      CALL dbcsr_random_dist(col_dist_c, dbcsr_nblkcols_total(matrix_c), npdims(2))
      CALL dbcsr_distribution_new(dist_a, mp_env, row_dist_a, col_dist_a, reuse_arrays=.TRUE.)
      CALL dbcsr_distribution_new(dist_b, mp_env, row_dist_b, col_dist_b, reuse_arrays=.TRUE.)
      CALL dbcsr_distribution_new(dist_c, mp_env, row_dist_c, col_dist_c, reuse_arrays=.TRUE.)
      ! Redistribute the matrices
      ! A
      CALL dbcsr_create(m_a, "Test for "//TRIM(dbcsr_name(matrix_a)), &
                        dist_a, dbcsr_get_matrix_type(matrix_a), &
                        dbcsr_row_block_sizes(matrix_a), &
                        dbcsr_col_block_sizes(matrix_a), &
                        data_type=dbcsr_get_data_type(matrix_a))
      CALL dbcsr_distribution_release(dist_a)
      CALL dbcsr_redistribute(matrix_a, m_a)
      ! B
      CALL dbcsr_create(m_b, "Test for "//TRIM(dbcsr_name(matrix_b)), &
                        dist_b, dbcsr_get_matrix_type(matrix_b), &
                        dbcsr_row_block_sizes(matrix_b), &
                        dbcsr_col_block_sizes(matrix_b), &
                        data_type=dbcsr_get_data_type(matrix_b))
      CALL dbcsr_distribution_release(dist_b)
      CALL dbcsr_redistribute(matrix_b, m_b)
      ! C
      CALL dbcsr_create(m_c, "Test for "//TRIM(dbcsr_name(matrix_c)), &
                        dist_c, dbcsr_get_matrix_type(matrix_c), &
                        dbcsr_row_block_sizes(matrix_c), &
                        dbcsr_col_block_sizes(matrix_c), &
                        data_type=dbcsr_get_data_type(matrix_c))
      CALL dbcsr_distribution_release(dist_c)
      CALL dbcsr_redistribute(matrix_c, m_c)

      IF (.FALSE.) THEN
         blk_offsets => dbcsr_row_block_offsets(matrix_c)
         WRITE (*, *) 'row_block_offsets(matrix_c)', blk_offsets
         blk_offsets => dbcsr_col_block_offsets(matrix_c)
         WRITE (*, *) 'col_block_offsets(matrix_c)', blk_offsets
      END IF

      IF (.FALSE.) THEN
         CALL dbcsr_print(m_c, matlab_format=.FALSE., variable_name='c_in_')
         CALL dbcsr_print(m_a, matlab_format=.FALSE., variable_name='a_')
         CALL dbcsr_print(m_b, matlab_format=.FALSE., variable_name='b_')
         CALL dbcsr_print(m_c, matlab_format=.FALSE., variable_name='c_out_')
      END IF

      occ_a = dbcsr_get_occupation(m_a)
      occ_b = dbcsr_get_occupation(m_b)
      occ_c_in = dbcsr_get_occupation(m_c)

      !
      ! Perform multiply
      IF (ALL(limits == 0)) THEN
         CALL cp__b("dbcsr/dbcsr_test_multiply.F",472,"limits shouldnt be 0")
      ELSE
         CALL dbcsr_multiply(transa, transb, alpha, &
                             m_a, m_b, beta, m_c, &
                             first_row=limits(1), &
                             last_row=limits(2), &
                             first_column=limits(3), &
                             last_column=limits(4), &
                             first_k=limits(5), &
                             last_k=limits(6), &
                             retain_sparsity=retain_sparsity)
      END IF

      occ_c_out = dbcsr_get_occupation(m_c)

      IF (.FALSE.) THEN
         PRINT*, 'retain_sparsity', retain_sparsity, occ_a, occ_b, occ_c_in, occ_c_out
         CALL dbcsr_print(m_a, matlab_format=.TRUE., variable_name='a_')
         CALL dbcsr_print(m_b, matlab_format=.TRUE., variable_name='b_')
         CALL dbcsr_print(m_c, matlab_format=.FALSE., variable_name='c_out_')
      END IF

      CALL dbcsr_replicate_all(m_c)
      CALL dbcsr_to_dense_local(m_c, data_c_dbcsr)
      CALL dbcsr_check_multiply(test_name, m_c, data_c_dbcsr, data_a, data_b, data_c, &
                                transa, transb, alpha, beta, limits, retain_sparsity, io_unit, mp_group, &
                                success)

      r_a = dbcsr_nfullrows_total(m_a)
      c_a = dbcsr_nfullcols_total(m_a)
      r_b = dbcsr_nfullrows_total(m_b)
      c_b = dbcsr_nfullcols_total(m_b)
      r_c = dbcsr_nfullrows_total(m_c)
      c_c = dbcsr_nfullcols_total(m_c)
      IF (io_unit .GT. 0) THEN
         IF (success) THEN
            WRITE (io_unit, *) REPEAT("*", 70)
            WRITE (io_unit, *) " -- TESTING dbcsr_multiply (", transa, ", ", transb, &
               ", ", dbcsr_get_data_type(m_a), &
               ", ", dbcsr_get_matrix_type(m_a), &
               ", ", dbcsr_get_matrix_type(m_b), &
               ", ", dbcsr_get_matrix_type(m_c), &
               ") ............... PASSED !"
            WRITE (io_unit, *) REPEAT("*", 70)
         ELSE
            WRITE (io_unit, *) REPEAT("*", 70)
            WRITE (io_unit, *) " -- TESTING dbcsr_multiply (", transa, ", ", transb, &
               ", ", dbcsr_get_data_type(m_a), &
               ", ", dbcsr_get_matrix_type(m_a), &
               ", ", dbcsr_get_matrix_type(m_b), &
               ", ", dbcsr_get_matrix_type(m_c), &
               ") ... FAILED !"
            WRITE (io_unit, *) REPEAT("*", 70)
            CALL cp__b("dbcsr/dbcsr_test_multiply.F",525,'Test failed')
         END IF
      END IF

      CALL dbcsr_release(m_a)
      CALL dbcsr_release(m_b)
      CALL dbcsr_release(m_c)

      CALL timestop(handle)

   END SUBROUTINE test_multiply

! **************************************************************************************************
!> \brief Performs a check of matrix multiplies
!> \param test_name ...
!> \param matrix_c ...
!> \param[in] dense_c_dbcsr             dense result of the dbcsr_multiply
!> \param dense_a input dense matrices
!> \param dense_b input dense matrices
!> \param dense_c input dense matrices
!> \param transa transposition status
!> \param transb transposition status
!> \param alpha coefficients for the gemm
!> \param beta coefficients for the gemm
!> \param[in] limits                    limits for the gemm
!> \param retain_sparsity ...
!> \param[in] io_unit                   io unit for printing
!> \param mp_group ...
!> \param[out] success                  if passed the check success=T
! **************************************************************************************************
   SUBROUTINE dbcsr_check_multiply(test_name, matrix_c, dense_c_dbcsr, dense_a, dense_b, dense_c, &
                                   transa, transb, alpha, beta, limits, retain_sparsity, io_unit, mp_group, &
                                   success)
      CHARACTER(len=*), INTENT(IN)                       :: test_name
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_c
      TYPE(dbcsr_data_obj), INTENT(inout)                :: dense_c_dbcsr, dense_a, dense_b, dense_c
      CHARACTER, INTENT(in)                              :: transa, transb
      TYPE(dbcsr_scalar_type), INTENT(in)                :: alpha, beta
      INTEGER, DIMENSION(6), INTENT(in)                  :: limits
      LOGICAL, INTENT(in)                                :: retain_sparsity
      INTEGER, INTENT(IN)                                :: io_unit, mp_group
      LOGICAL, INTENT(out)                               :: success

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_check_multiply', &
         routineP = moduleN//':'//routineN
      INTEGER :: a_col, a_m, a_n, a_row, b_col, b_m, b_n, b_row, c_col, c_col_size, c_row, &
         c_row_size, handle, i, istat, j, k, lda, ldb, ldc, lwork, m, mynode, n, numnodes
      CHARACTER, PARAMETER                               :: norm = 'I'

      LOGICAL                                            :: valid
      REAL(real_4), ALLOCATABLE, DIMENSION(:)            :: work_sp
      REAL(real_4), EXTERNAL                             :: clange, slamch, slange
      REAL(real_8)                                       :: a_norm, b_norm, c_norm_dbcsr, c_norm_in, &
                                                            c_norm_out, eps, residual
      REAL(real_8), ALLOCATABLE, DIMENSION(:)            :: work
      REAL(real_8), EXTERNAL                             :: dlamch, dlange, zlange

      CALL timeset(routineN, handle)

      CALL mp_environ(numnodes, mynode, mp_group)

      CALL dbcsr_data_get_sizes(dense_c, c_row_size, c_col_size, valid)
      IF (.NOT. valid) &
         CALL cp__b("dbcsr/dbcsr_test_multiply.F",588,"dense matrix not valid")
      CALL dbcsr_data_get_sizes(dense_c, ldc, i, valid)
      IF (.NOT. valid) &
         CALL cp__b("dbcsr/dbcsr_test_multiply.F",591,"dense matrix not valid")
      CALL dbcsr_data_get_sizes(dense_a, lda, i, valid)
      IF (.NOT. valid) &
         CALL cp__b("dbcsr/dbcsr_test_multiply.F",594,"dense matrix not valid")
      CALL dbcsr_data_get_sizes(dense_b, ldb, i, valid)
      IF (.NOT. valid) &
         CALL cp__b("dbcsr/dbcsr_test_multiply.F",597,"dense matrix not valid")
      !
      !
      m = limits(2)-limits(1)+1
      n = limits(4)-limits(3)+1
      k = limits(6)-limits(5)+1
      a_row = limits(1); a_col = limits(5)
      b_row = limits(5); b_col = limits(3)
      c_row = limits(1); c_col = limits(3)
      !
      !
      IF (transA == dbcsr_no_transpose) THEN
         a_m = m
         a_n = k
      ELSE
         a_m = k
         a_n = m
         i = a_row
         a_row = a_col
         a_col = i
      END IF
      IF (transB == dbcsr_no_transpose) THEN
         b_m = k
         b_n = n
      ELSE
         b_m = n
         b_n = k
         i = b_row
         b_row = b_col
         b_col = i
      END IF
      !
      ! set the size of the work array
      lwork = MAXVAL((/lda, ldb, ldc/))
      !
      !
      SELECT CASE (dense_a%d%data_type)
      CASE (dbcsr_type_real_8_2d)
         ALLOCATE (work(lwork), STAT=istat)
         IF (istat /= 0) &
            CALL cp__b("dbcsr/dbcsr_test_multiply.F",637,"allocation problem")
         eps = dlamch('eps')
         a_norm = dlange(norm, a_m, a_n, dense_a%d%r2_dp(a_row, a_col), lda, work)
         b_norm = dlange(norm, b_m, b_n, dense_b%d%r2_dp(b_row, b_col), ldb, work)
         c_norm_in = dlange(norm, c_row_size, c_col_size, dense_c%d%r2_dp(1, 1), ldc, work)
         c_norm_dbcsr = dlange(norm, c_row_size, c_col_size, dense_c_dbcsr%d%r2_dp(1, 1), ldc, work)
         !
         CALL dgemm(transa, transb, m, n, k, alpha%r_dp, dense_a%d%r2_dp(a_row, a_col), lda, &
                    dense_b%d%r2_dp(b_row, b_col), ldb, beta%r_dp, dense_c%d%r2_dp(c_row, c_col), ldc)
         !
         ! impose the sparsity if needed
         IF (retain_sparsity) CALL dbcsr_impose_sparsity(matrix_c, dense_c)
         !
         c_norm_out = dlange(norm, m, n, dense_c%d%r2_dp(c_row, c_col), ldc, work)
         !
         ! take the difference dense/sparse
         dense_c%d%r2_dp = dense_c%d%r2_dp-dense_c_dbcsr%d%r2_dp
         !
         ! compute the residual
         residual = dlange(norm, c_row_size, c_col_size, dense_c%d%r2_dp(1, 1), ldc, work)
         DEALLOCATE (work)
      CASE (dbcsr_type_real_4_2d)
         ALLOCATE (work_sp(lwork), STAT=istat)
         IF (istat /= 0) &
            CALL cp__b("dbcsr/dbcsr_test_multiply.F",661,"allocation problem")
         eps = REAL(slamch('eps'), real_8)
         a_norm = slange(norm, a_m, a_n, dense_a%d%r2_sp(a_row, a_col), lda, work_sp)
         b_norm = slange(norm, b_m, b_n, dense_b%d%r2_sp(b_row, b_col), ldb, work_sp)
         c_norm_in = slange(norm, c_row_size, c_col_size, dense_c%d%r2_sp(1, 1), ldc, work_sp)
         c_norm_dbcsr = slange(norm, c_row_size, c_col_size, dense_c_dbcsr%d%r2_sp(1, 1), ldc, work_sp)
         !

         IF (.FALSE.) THEN
            !IF (io_unit .GT. 0) THEN
            DO j = 1, SIZE(dense_a%d%r2_sp, 2)
               DO i = 1, SIZE(dense_a%d%r2_sp, 1)
                  WRITE (*, '(A,I3,A,I3,A,E15.7,A)') 'a(', i, ',', j, ')=', dense_a%d%r2_sp(i, j), ';'
               END DO
            END DO
            DO j = 1, SIZE(dense_b%d%r2_sp, 2)
               DO i = 1, SIZE(dense_b%d%r2_sp, 1)
                  WRITE (*, '(A,I3,A,I3,A,E15.7,A)') 'b(', i, ',', j, ')=', dense_b%d%r2_sp(i, j), ';'
               END DO
            END DO
            DO j = 1, SIZE(dense_c%d%r2_sp, 2)
               DO i = 1, SIZE(dense_c%d%r2_sp, 1)
                  WRITE (*, '(A,I3,A,I3,A,E15.7,A)') 'c_in(', i, ',', j, ')=', dense_c%d%r2_sp(i, j), ';'
               END DO
            END DO
         END IF

         CALL sgemm(transa, transb, m, n, k, alpha%r_sp, dense_a%d%r2_sp(a_row, a_col), lda, &
                    dense_b%d%r2_sp(b_row, b_col), ldb, beta%r_sp, dense_c%d%r2_sp(c_row, c_col), ldc)
         !
         ! impose the sparsity if needed
         IF (retain_sparsity) CALL dbcsr_impose_sparsity(matrix_c, dense_c)

         IF (.FALSE.) THEN
            !IF (io_unit .GT. 0) THEN
            DO j = 1, SIZE(dense_c%d%r2_sp, 2)
               DO i = 1, SIZE(dense_c%d%r2_sp, 1)
                  WRITE (*, '(A,I3,A,I3,A,E15.7,A)') 'c_out(', i, ',', j, ')=', dense_c%d%r2_sp(i, j), ';'
               END DO
            END DO
            DO j = 1, SIZE(dense_c_dbcsr%d%r2_sp, 2)
               DO i = 1, SIZE(dense_c_dbcsr%d%r2_sp, 1)
                  WRITE (*, '(A,I3,A,I3,A,E15.7,A)') 'c_dbcsr(', i, ',', j, ')=', dense_c_dbcsr%d%r2_sp(i, j), ';'
               END DO
            END DO
         END IF
         !
         c_norm_out = slange(norm, m, n, dense_c%d%r2_sp(c_row, c_col), ldc, work_sp)
         !
         ! take the difference dense/sparse
         dense_c%d%r2_sp = dense_c%d%r2_sp-dense_c_dbcsr%d%r2_sp
         !
         ! compute the residual
         residual = REAL(slange(norm, c_row_size, c_col_size, dense_c%d%r2_sp(1, 1), ldc, work_sp), real_8)
         DEALLOCATE (work_sp)
      CASE (dbcsr_type_complex_8_2d)
         ALLOCATE (work(lwork), STAT=istat)
         IF (istat /= 0) &
            CALL cp__b("dbcsr/dbcsr_test_multiply.F",719,"allocation problem")
         eps = dlamch('eps')
         a_norm = zlange(norm, a_m, a_n, dense_a%d%c2_dp(a_row, a_col), lda, work)
         b_norm = zlange(norm, b_m, b_n, dense_b%d%c2_dp(b_row, b_col), ldb, work)
         c_norm_in = zlange(norm, c_row_size, c_col_size, dense_c%d%c2_dp(1, 1), ldc, work)
         c_norm_dbcsr = zlange(norm, c_row_size, c_col_size, dense_c_dbcsr%d%c2_dp(1, 1), ldc, work)
         !
         CALL zgemm(transa, transb, m, n, k, alpha%c_dp, dense_a%d%c2_dp(a_row, a_col), lda, &
                    dense_b%d%c2_dp(b_row, b_col), ldb, beta%c_dp, dense_c%d%c2_dp(c_row, c_col), ldc)
         !
         ! impose the sparsity if needed
         IF (retain_sparsity) CALL dbcsr_impose_sparsity(matrix_c, dense_c)
         !
         c_norm_out = zlange(norm, m, n, dense_c%d%c2_dp(c_row, c_col), ldc, work)
         !
         ! take the difference dense/sparse
         dense_c%d%c2_dp = dense_c%d%c2_dp-dense_c_dbcsr%d%c2_dp
         !
         ! compute the residual
         residual = zlange(norm, c_row_size, c_col_size, dense_c%d%c2_dp(1, 1), ldc, work)
         DEALLOCATE (work)
      CASE (dbcsr_type_complex_4_2d)
         ALLOCATE (work_sp(lwork), STAT=istat)
         IF (istat /= 0) &
            CALL cp__b("dbcsr/dbcsr_test_multiply.F",743,"allocation problem")
         eps = REAL(slamch('eps'), real_8)
         a_norm = clange(norm, a_m, a_n, dense_a%d%c2_sp(a_row, a_col), lda, work_sp)
         b_norm = clange(norm, b_m, b_n, dense_b%d%c2_sp(b_row, b_col), ldb, work_sp)
         c_norm_in = clange(norm, c_row_size, c_col_size, dense_c%d%c2_sp(1, 1), ldc, work_sp)
         c_norm_dbcsr = clange(norm, c_row_size, c_col_size, dense_c_dbcsr%d%c2_sp(1, 1), ldc, work_sp)
         !
         CALL cgemm(transa, transb, m, n, k, alpha%c_sp, dense_a%d%c2_sp(a_row, a_col), lda, &
                    dense_b%d%c2_sp(b_row, b_col), ldb, beta%c_sp, dense_c%d%c2_sp(c_row, c_col), ldc)
         !
         ! impose the sparsity if needed
         IF (retain_sparsity) CALL dbcsr_impose_sparsity(matrix_c, dense_c)
         !
         c_norm_out = clange(norm, m, n, dense_c%d%c2_sp(c_row, c_col), ldc, work_sp)
         !
         ! take the difference dense/sparse
         dense_c%d%c2_sp = dense_c%d%c2_sp-dense_c_dbcsr%d%c2_sp
         !
         ! compute the residual
         residual = clange(norm, c_row_size, c_col_size, dense_c%d%c2_sp(1, 1), ldc, work_sp)
         DEALLOCATE (work_sp)
      CASE default
         CALL cp__b("dbcsr/dbcsr_test_multiply.F",765,"Incorrect or 1-D data type")
      END SELECT

      IF (mynode .EQ. 0) THEN
         IF (residual/((a_norm+b_norm+c_norm_in)*REAL(n, real_8)*eps) .GT. 10.0_real_8) THEN
            success = .FALSE.
         ELSE
            success = .TRUE.
         END IF
      END IF
      !
      ! syncronize the result...
      CALL mp_bcast(success, 0, mp_group)
      !
      ! printing
      IF (io_unit .GT. 0) THEN
         WRITE (io_unit, *) 'test_name ', test_name
         WRITE (io_unit, '(3(A,E12.5))') ' residual ', residual, ', a_norm ', a_norm, ', b_norm ', b_norm
         WRITE (io_unit, '(3(A,E12.5))') ' c_norm_in ', c_norm_in, ', c_norm_out ', c_norm_out, &
            ', c_norm_dbcsr ', c_norm_dbcsr
         WRITE (io_unit, '(A)') ' Checking the norm of the difference against reference GEMM '
         WRITE (io_unit, '(A,E12.5)') ' -- ||C_dbcsr-C_dense||_oo/((||A||_oo+||B||_oo+||C||_oo).N.eps)=', &
            residual/((a_norm+b_norm+c_norm_in)*n*eps)
         !
         ! check for nan or inf here
         IF (success) THEN
            WRITE (io_unit, '(A)') ' The solution is CORRECT !'
         ELSE
            WRITE (io_unit, '(A)') ' The solution is suspicious !'
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE dbcsr_check_multiply

END MODULE dbcsr_test_multiply
