!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represents an enumeration, i.e. a mapping between integers and strings
!> \par History
!>      08.2004 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE input_enumeration_types

   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: default_string_length
   USE string_utilities,                ONLY: a2s,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_enumeration_types'
   INTEGER, SAVE, PRIVATE :: last_enumeration_id = 0

   PUBLIC :: enumeration_type
   PUBLIC :: enum_create, enum_retain, enum_release, enum_i2c, enum_c2i

! **************************************************************************************************
!> \brief represents an enumaration, i.e. a mapping between strings and numbers
!> \param id_nr identification number (unique)
!> \param ref_count reference count
!> \param c_vals string values
!> \param i_vals integer values
!> \param strict if integer values not in the list should be accepted
!> \author fawzi
! **************************************************************************************************
   TYPE char_array
      CHARACTER, DIMENSION(:), POINTER :: chars => Null()
   END TYPE char_array

   TYPE enumeration_type
      INTEGER :: id_nr, ref_count
      CHARACTER(len=default_string_length), DIMENSION(:), POINTER :: c_vals
      TYPE(char_array), DIMENSION(:), POINTER :: desc => Null()
      INTEGER, DIMENSION(:), POINTER :: i_vals
      LOGICAL :: strict
   END TYPE enumeration_type

CONTAINS

! **************************************************************************************************
!> \brief creates an enumeration
!> \param enum the enumeration to be created
!> \param c_vals string values
!> \param i_vals integer values
!> \param desc ...
!> \param strict if integer values not in the list should be accepted,
!>        defaults defaults to true
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE enum_create(enum, c_vals, i_vals, desc, strict)
      TYPE(enumeration_type), POINTER                    :: enum
      CHARACTER(len=*), DIMENSION(:), INTENT(in)         :: c_vals
      INTEGER, DIMENSION(:), INTENT(in)                  :: i_vals
      CHARACTER(len=*), DIMENSION(:), INTENT(in), &
         OPTIONAL                                        :: desc
      LOGICAL, INTENT(in), OPTIONAL                      :: strict

      CHARACTER(len=*), PARAMETER :: routineN = 'enum_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, n

      IF(.NOT.(.NOT. ASSOCIATED(enum)))CALL cp__a("input/input_enumeration_types.F",75)
      IF(.NOT.(SIZE(c_vals) == SIZE(i_vals)))CALL cp__a("input/input_enumeration_types.F",76)
      ALLOCATE (enum)
      last_enumeration_id = last_enumeration_id+1
      enum%id_nr = last_enumeration_id
      enum%ref_count = 1
      ALLOCATE (enum%c_vals(SIZE(c_vals)))
      DO i = 1, SIZE(enum%c_vals)
         enum%c_vals(i) = c_vals(i)
         CALL uppercase(enum%c_vals(i))
      END DO
      ALLOCATE (enum%i_vals(SIZE(i_vals)))
      enum%i_vals = i_vals
      enum%strict = .TRUE.
      IF (PRESENT(strict)) enum%strict = strict
      ALLOCATE (enum%desc(SIZE(c_vals)))
      IF (PRESENT(desc)) THEN
         IF(.NOT.(SIZE(enum%desc) == SIZE(desc)))CALL cp__a("input/input_enumeration_types.F",92)
         DO i = 1, SIZE(enum%desc)
            n = LEN_TRIM(desc(i))
            ALLOCATE (enum%desc(i)%chars(n))
            DO j = 1, n
               enum%desc(i)%chars(j) = desc(i) (j:j)
            END DO
         END DO
      ELSE
         DO i = 1, SIZE(enum%desc)
            ALLOCATE (enum%desc(i)%chars(1))
            enum%desc(i)%chars(1:1) = ' '
         END DO
      END IF
   END SUBROUTINE enum_create

! **************************************************************************************************
!> \brief retains the given enumeration
!> \param enum the obect to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE enum_retain(enum)
      TYPE(enumeration_type), POINTER                    :: enum

      CHARACTER(len=*), PARAMETER :: routineN = 'enum_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(enum)))CALL cp__a("input/input_enumeration_types.F",118)
      IF(.NOT.(enum%ref_count > 0))CALL cp__a("input/input_enumeration_types.F",119)
      enum%ref_count = enum%ref_count+1
   END SUBROUTINE enum_retain

! **************************************************************************************************
!> \brief releases the given enumeration
!> \param enum the obect to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE enum_release(enum)
      TYPE(enumeration_type), POINTER                    :: enum

      CHARACTER(len=*), PARAMETER :: routineN = 'enum_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(enum)) THEN
         IF(.NOT.(enum%ref_count > 0))CALL cp__a("input/input_enumeration_types.F",136)
         enum%ref_count = enum%ref_count-1
         IF (enum%ref_count == 0) THEN
            DEALLOCATE (enum%c_vals)
            DEALLOCATE (enum%i_vals)
            DO i = 1, SIZE(enum%desc)
               DEALLOCATE (enum%desc(i)%chars)
            END DO
            DEALLOCATE (enum%desc)
            DEALLOCATE (enum)
         END IF
      END IF
      NULLIFY (enum)
   END SUBROUTINE enum_release

! **************************************************************************************************
!> \brief maps an integer to a string
!> \param enum the enumeration to use for the mapping
!> \param i the value to map
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION enum_i2c(enum, i) RESULT(res)
      TYPE(enumeration_type), POINTER                    :: enum
      INTEGER, INTENT(in)                                :: i
      CHARACTER(len=default_string_length)               :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'enum_i2c', routineP = moduleN//':'//routineN

      INTEGER                                            :: j
      LOGICAL                                            :: found

      IF(.NOT.(ASSOCIATED(enum)))CALL cp__a("input/input_enumeration_types.F",168)
      IF(.NOT.(enum%ref_count > 0))CALL cp__a("input/input_enumeration_types.F",169)
      res = " "
      found = .FALSE.
      DO j = 1, SIZE(enum%i_vals)
         IF (enum%i_vals(j) == i) THEN
            res = enum%c_vals(j)
            found = .TRUE.
            EXIT
         END IF
      END DO
      IF (.NOT. found) THEN
         IF (enum%strict) THEN
            DO j = 1, SIZE(enum%desc)
               PRINT*, TRIM(a2s(enum%desc(j)%chars))
               PRINT*, TRIM(enum%c_vals(j))
            ENDDO
            PRINT*, enum%i_vals
         END IF
         IF (enum%strict) &
            CALL cp__b("input/input_enumeration_types.F",188,"invalid value for enumeration:"//cp_to_string(i))
         res = ADJUSTL(cp_to_string(i))
      END IF
   END FUNCTION enum_i2c

! **************************************************************************************************
!> \brief maps a string to an integer
!> \param enum the enumeration to use for the mapping
!> \param c the value to map
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION enum_c2i(enum, c) RESULT(res)
      TYPE(enumeration_type), POINTER                    :: enum
      CHARACTER(len=*), INTENT(in)                       :: c
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'enum_c2i', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: upc
      INTEGER                                            :: iostat, j
      LOGICAL                                            :: found

      IF(.NOT.(ASSOCIATED(enum)))CALL cp__a("input/input_enumeration_types.F",211)
      IF(.NOT.(enum%ref_count > 0))CALL cp__a("input/input_enumeration_types.F",212)
      upc = c
      CALL uppercase(upc)
      found = .FALSE.
      DO j = 1, SIZE(enum%c_vals)
         IF (enum%c_vals(j) == upc) THEN
            res = enum%i_vals(j)
            found = .TRUE.
            EXIT
         END IF
      END DO

      IF (.NOT. found) THEN
         IF (enum%strict) &
            CALL cp__b("input/input_enumeration_types.F",226,"invalid value for enumeration:"//TRIM(c))
         READ (c, "(i10)", iostat=iostat) res
         IF (iostat /= 0) &
            CALL cp__b("input/input_enumeration_types.F",229,"invalid value for enumeration2:"//TRIM(c))
      END IF
   END FUNCTION enum_c2i

END MODULE input_enumeration_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief a module to allow simple buffering of read lines of a parser
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date 08.2008
! **************************************************************************************************
MODULE cp_parser_buffer_types

   USE kinds,                           ONLY: max_line_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

! ****************************************************************************
!> \brief  Buffer type for speeding-up the parsing in parallel
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date   08.2008
! **************************************************************************************************
   TYPE buffer_type
      INTEGER                              :: size, buffer_id
      INTEGER                              :: present_line_number, &
                                              last_line_number, &
                                              istat
      INTEGER, DIMENSION(:), POINTER       :: input_line_numbers
      CHARACTER(LEN=max_line_length), &
         DIMENSION(:), POINTER           :: input_lines
      TYPE(buffer_type), POINTER           :: sub_buffer
   END TYPE buffer_type

   PUBLIC :: buffer_type, create_buffer_type, release_buffer_type, copy_buffer_type, &
             initialize_sub_buffer, finalize_sub_buffer
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_parser_buffer_types'
   INTEGER, PARAMETER, PRIVATE          :: buffer_size = 1000

CONTAINS

! ****************************************************************************
!> \brief  Creates the parser buffer type
!> \param buffer ...
!> \date   08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE create_buffer_type(buffer)
      TYPE(buffer_type), POINTER                         :: buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'create_buffer_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(buffer)))CALL cp__a("input/cp_parser_buffer_types.F",54)
      ALLOCATE (buffer)
      buffer%size = buffer_size
      ALLOCATE (buffer%input_lines(buffer%size))
      ALLOCATE (buffer%input_line_numbers(buffer%size))
      buffer%buffer_id = 0
      buffer%input_line_numbers = 0
      buffer%istat = 0
      buffer%present_line_number = buffer%size
      buffer%last_line_number = buffer%size
      NULLIFY (buffer%sub_buffer)
   END SUBROUTINE create_buffer_type

! ****************************************************************************
!> \brief  Releases the parser buffer type
!> \param buffer ...
!> \date   08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   RECURSIVE SUBROUTINE release_buffer_type(buffer)
      TYPE(buffer_type), POINTER                         :: buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'release_buffer_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(buffer)))CALL cp__a("input/cp_parser_buffer_types.F",79)
      DEALLOCATE (buffer%input_lines)
      DEALLOCATE (buffer%input_line_numbers)
      IF (ASSOCIATED(buffer%sub_buffer)) THEN
         CALL release_buffer_type(buffer%sub_buffer)
      END IF
      DEALLOCATE (buffer)
   END SUBROUTINE release_buffer_type

! ****************************************************************************
!> \brief  Copies  buffer types
!> \param buffer_in ...
!> \param buffer_out ...
!> \param force ...
!> \date   08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   RECURSIVE SUBROUTINE copy_buffer_type(buffer_in, buffer_out, force)
      TYPE(buffer_type), POINTER                         :: buffer_in, buffer_out
      LOGICAL, INTENT(IN), OPTIONAL                      :: force

      CHARACTER(len=*), PARAMETER :: routineN = 'copy_buffer_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_force

      IF(.NOT.(ASSOCIATED(buffer_in)))CALL cp__a("input/cp_parser_buffer_types.F",106)
      IF(.NOT.(ASSOCIATED(buffer_out)))CALL cp__a("input/cp_parser_buffer_types.F",107)
      IF(.NOT.(buffer_in%size == buffer_out%size))CALL cp__a("input/cp_parser_buffer_types.F",108)
      my_force = .FALSE.
      IF (PRESENT(force)) my_force = force
      ! Copy buffer structure
      buffer_out%present_line_number = buffer_in%present_line_number
      buffer_out%last_line_number = buffer_in%last_line_number
      buffer_out%istat = buffer_in%istat
      ! This part can be quite expensive.. we do it only when strictly necessary..
      IF ((buffer_out%buffer_id /= buffer_in%buffer_id) .OR. (my_force)) THEN
         buffer_out%buffer_id = buffer_in%buffer_id
         buffer_out%input_line_numbers = buffer_in%input_line_numbers
         ! Explicit loop: bypass a NAG bug..
         DO i = 1, SIZE(buffer_in%input_lines)
            buffer_out%input_lines(i) = buffer_in%input_lines(i)
         END DO
      END IF
      IF (ASSOCIATED(buffer_in%sub_buffer) .AND. ASSOCIATED(buffer_out%sub_buffer)) THEN
         CALL copy_buffer_type(buffer_in%sub_buffer, buffer_out%sub_buffer, force)
      END IF
   END SUBROUTINE copy_buffer_type

! ****************************************************************************
!> \brief  Initializes sub buffer structure
!> \param sub_buffer ...
!> \param buffer ...
!> \date   08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE initialize_sub_buffer(sub_buffer, buffer)
      TYPE(buffer_type), POINTER                         :: sub_buffer, buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_sub_buffer', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(buffer)))CALL cp__a("input/cp_parser_buffer_types.F",142)
      IF(.NOT.(.NOT. ASSOCIATED(sub_buffer)))CALL cp__a("input/cp_parser_buffer_types.F",143)
      CALL create_buffer_type(sub_buffer)
      CALL copy_buffer_type(buffer, sub_buffer)
      sub_buffer%present_line_number = 0
   END SUBROUTINE initialize_sub_buffer

! ****************************************************************************
!> \brief  Finalizes sub buffer structure
!> \param sub_buffer ...
!> \param buffer ...
!> \date   08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE finalize_sub_buffer(sub_buffer, buffer)
      TYPE(buffer_type), POINTER                         :: sub_buffer, buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'finalize_sub_buffer', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(buffer)))CALL cp__a("input/cp_parser_buffer_types.F",162)
      IF(.NOT.(ASSOCIATED(sub_buffer)))CALL cp__a("input/cp_parser_buffer_types.F",163)
      CALL copy_buffer_type(sub_buffer, buffer)
      CALL release_buffer_type(sub_buffer)
   END SUBROUTINE finalize_sub_buffer

END MODULE cp_parser_buffer_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief a module to allow the storage of the parser status
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date 08.2008
! **************************************************************************************************
MODULE cp_parser_status_types
   USE cp_parser_buffer_types,          ONLY: buffer_type,&
                                              create_buffer_type,&
                                              release_buffer_type
   USE kinds,                           ONLY: max_line_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   TYPE status_type
      LOGICAL                                        :: in_use
      INTEGER                                        :: old_input_line_number
      INTEGER                                        :: old_icol
      INTEGER                                        :: old_icol1
      INTEGER                                        :: old_icol2
      CHARACTER(LEN=max_line_length)                 :: old_input_line
      ! Store status of the buffer
      TYPE(buffer_type), POINTER                     :: buffer
   END TYPE status_type

   PUBLIC :: status_type, create_status_type, release_status_type
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_parser_status_types'

CONTAINS

! ****************************************************************************
!> \brief creates the parser status type
!> \param status ...
!> \date  08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE create_status_type(status)
      TYPE(status_type), POINTER                         :: status

      CHARACTER(len=*), PARAMETER :: routineN = 'create_status_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(status)))CALL cp__a("input/cp_parser_status_types.F",49)
      ALLOCATE (status)
      status%in_use = .FALSE.
      status%old_input_line = ""
      status%old_input_line_number = HUGE(0)
      status%old_icol = HUGE(0)
      status%old_icol1 = HUGE(0)
      status%old_icol2 = HUGE(0)
      NULLIFY (status%buffer)
      CALL create_buffer_type(status%buffer)
   END SUBROUTINE create_status_type

! ****************************************************************************
!> \brief releases the parser status type
!> \param status ...
!> \date  08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE release_status_type(status)
      TYPE(status_type), POINTER                         :: status

      CHARACTER(len=*), PARAMETER :: routineN = 'release_status_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(status)))CALL cp__a("input/cp_parser_status_types.F",73)
      CALL release_buffer_type(status%buffer)
      DEALLOCATE (status)
   END SUBROUTINE release_status_type

END MODULE cp_parser_status_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief a module to allow simple internal preprocessing in input files.
!> \par History
!>      - standalone proof-of-concept implementation (20.02.2008,AK)
!>      - integration into cp2k (22.02.2008,tlaino)
!>      - variables added (25.02.2008,AK)
!> \author Axel Kohlmeyer [AK] - CMM/UPenn Philadelphia
!> \date 25.02.2008
! **************************************************************************************************
MODULE cp_parser_ilist_types

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   PRIVATE

   TYPE ilist_type
      LOGICAL                              :: in_use
      INTEGER                              :: nel_list
      INTEGER                              :: istart, iend
      INTEGER                              :: ipresent
   END TYPE ilist_type

   PUBLIC :: ilist_type, create_ilist_type, release_ilist_type
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_parser_ilist_types'

CONTAINS

! ****************************************************************************
!> \brief creates the integer listing type
!> \param ilist ...
!> \date  08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE create_ilist_type(ilist)
      TYPE(ilist_type), POINTER                          :: ilist

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ilist_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ilist)))CALL cp__a("input/cp_parser_ilist_types.F",45)
      ALLOCATE (ilist)
      ilist%istart = HUGE(0)
      ilist%iend = HUGE(0)
      ilist%nel_list = HUGE(0)
      ilist%ipresent = HUGE(0)
      ilist%in_use = .FALSE.

   END SUBROUTINE create_ilist_type

! ****************************************************************************
!> \brief creates the integer listing type
!> \param ilist ...
!> \date  08.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE release_ilist_type(ilist)
      TYPE(ilist_type), POINTER                          :: ilist

      CHARACTER(len=*), PARAMETER :: routineN = 'release_ilist_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ilist)))CALL cp__a("input/cp_parser_ilist_types.F",67)
      DEALLOCATE (ilist)
   END SUBROUTINE release_ilist_type

END MODULE cp_parser_ilist_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief a module to allow simple internal preprocessing in input files.
!> \par History
!>      - standalone proof-of-concept implementation (20.02.2008,AK)
!>      - integration into cp2k (22.02.2008,tlaino)
!>      - variables added (25.02.2008,AK)
!> \author Axel Kohlmeyer [AK] - CMM/UPenn Philadelphia
!> \date 25.02.2008
! **************************************************************************************************
MODULE cp_parser_inpp_types

   USE kinds,                           ONLY: default_path_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   TYPE inpp_type
      ! for '@INCLUDE "some_file.inc"'
      ! currently open include file stack pointer
      INTEGER                              :: io_stack_level
      ! include file stack data
      INTEGER, POINTER, DIMENSION(:)       :: io_stack_channel, &
                                              io_stack_lineno
      CHARACTER(len=default_path_length), &
         POINTER, DIMENSION(:)             :: io_stack_filename
      ! for '@SET VAR value' and '${VAR}'
      ! table size
      INTEGER                              :: num_variables
      ! table entries
      CHARACTER(len=default_path_length), &
         POINTER, DIMENSION(:)       :: variable_name
      CHARACTER(len=default_path_length), &
         POINTER, DIMENSION(:)       :: variable_value
   END TYPE inpp_type

   PUBLIC :: inpp_type, create_inpp_type, release_inpp_type
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_parser_inpp_types'

CONTAINS

! ****************************************************************************
!> \brief creates the internal preprocessing type
!> \param inpp ...
!> \param initial_variables ...
!> \date  22.02.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE create_inpp_type(inpp, initial_variables)
      TYPE(inpp_type), POINTER                           :: inpp
      CHARACTER(len=default_path_length), &
         DIMENSION(:, :), POINTER                        :: initial_variables

      CHARACTER(len=*), PARAMETER :: routineN = 'create_inpp_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(inpp)))CALL cp__a("input/cp_parser_inpp_types.F",62)
      ALLOCATE (inpp)

      inpp%io_stack_level = 0
      NULLIFY (inpp%io_stack_channel, &
               inpp%io_stack_lineno, &
               inpp%io_stack_filename)

      inpp%num_variables = 0
      NULLIFY (inpp%variable_name, &
               inpp%variable_value)

      IF (ASSOCIATED(initial_variables)) THEN
         inpp%num_variables = SIZE(initial_variables, 2)
         ALLOCATE (inpp%variable_name(inpp%num_variables))
         inpp%variable_name = initial_variables(1, :)
         ALLOCATE (inpp%variable_value(inpp%num_variables))
         inpp%variable_value = initial_variables(2, :)
      ENDIF

   END SUBROUTINE create_inpp_type

! ****************************************************************************
!> \brief releases the internal preprocessing type
!> \param inpp ...
!> \date  22.02.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE release_inpp_type(inpp)
      TYPE(inpp_type), POINTER                           :: inpp

      CHARACTER(len=*), PARAMETER :: routineN = 'release_inpp_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(inpp)))CALL cp__a("input/cp_parser_inpp_types.F",96)

      IF (ASSOCIATED(inpp%io_stack_channel)) THEN
         DEALLOCATE (inpp%io_stack_channel)
      END IF
      IF (ASSOCIATED(inpp%io_stack_lineno)) THEN
         DEALLOCATE (inpp%io_stack_lineno)
      END IF
      IF (ASSOCIATED(inpp%io_stack_filename)) THEN
         DEALLOCATE (inpp%io_stack_filename)
      END IF

      IF (ASSOCIATED(inpp%variable_name)) THEN
         DEALLOCATE (inpp%variable_name)
      END IF
      IF (ASSOCIATED(inpp%variable_value)) THEN
         DEALLOCATE (inpp%variable_value)
      END IF

      DEALLOCATE (inpp)
   END SUBROUTINE release_inpp_type

END MODULE cp_parser_inpp_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief type to store parallelization informations (at the moment assumes 1d
!>      position and uses mpi)
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_para_env
   USE cp_para_types,                   ONLY: cp_para_cart_type,&
                                              cp_para_env_type
   USE message_passing,                 ONLY: mp_comm_free,&
                                              mp_environ
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_para_env'

   PUBLIC :: cp_para_env_retain, cp_para_env_release, cp_para_env_create
   PUBLIC :: cp_cart_create, cp_cart_release
!***
CONTAINS

! **************************************************************************************************
!> \brief creates a new para environment
!> \param para_env the new parallel environment
!> \param group the id of the actual mpi_group
!> \param source the id of the special (master) processor (defaults to 0)
!> \param mepos the id of the actual processor
!> \param num_pe the number of processors in the group
!> \param owns_group if the group is owned by this object (defaults to true)
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_para_env_create(para_env, group, source, mepos, num_pe, &
                                 owns_group)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(in)                                :: group
      INTEGER, INTENT(in), OPTIONAL                      :: source, mepos, num_pe
      LOGICAL, INTENT(in), OPTIONAL                      :: owns_group

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_para_env_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(para_env)))CALL cp__a("common/cp_para_env.F",53)
      ALLOCATE (para_env)
      para_env%group = group
      para_env%source = 0
      para_env%ref_count = 1
      para_env%owns_group = .TRUE.
      IF (PRESENT(source)) para_env%source = source
      IF (PRESENT(owns_group)) para_env%owns_group = owns_group
      IF (.NOT. (PRESENT(mepos) .AND. PRESENT(num_pe))) THEN
         CALL cp_para_env_update(para_env)
      ELSE
         para_env%mepos = mepos
         para_env%num_pe = num_pe
      END IF
      para_env%ionode = para_env%mepos == para_env%source
   END SUBROUTINE cp_para_env_create

! **************************************************************************************************
!> \brief retains the para object (to be called when you want to keep a
!>      shared copy of this object)
!> \param para_env the new group
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_para_env_retain(para_env)
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_para_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("common/cp_para_env.F",84)
      IF(.NOT.(para_env%ref_count > 0))CALL cp__a("common/cp_para_env.F",85)
      para_env%ref_count = para_env%ref_count+1
   END SUBROUTINE cp_para_env_retain

! **************************************************************************************************
!> \brief releases the para object (to be called when you don't want anymore
!>      the shared copy of this object)
!> \param para_env the new group
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      to avoid circular dependencies cp_log_handling has a private copy
!>      of this method (see cp_log_handling:my_cp_para_env_release)!
! **************************************************************************************************
   SUBROUTINE cp_para_env_release(para_env)
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_para_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(para_env)) THEN
         IF(.NOT.(para_env%ref_count > 0))CALL cp__a("common/cp_para_env.F",107)
         para_env%ref_count = para_env%ref_count-1
         IF (para_env%ref_count < 1) THEN
            IF (para_env%owns_group) THEN
               CALL mp_comm_free(para_env%group)
            END IF
            DEALLOCATE (para_env)
         END IF
      END IF
      NULLIFY (para_env)
   END SUBROUTINE cp_para_env_release

! **************************************************************************************************
!> \brief gets again the position and size of the group from the mpi_group
!> \param para_env the new group
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_para_env_update(para_env)
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_para_env_update', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(para_env)))CALL cp__a("common/cp_para_env.F",132)
      IF(.NOT.(para_env%ref_count > 0))CALL cp__a("common/cp_para_env.F",133)
      CALL mp_environ(taskid=para_env%mepos, numtask=para_env%num_pe, &
                      groupid=para_env%group)
      para_env%ionode = para_env%mepos == para_env%source
   END SUBROUTINE cp_para_env_update

! **************************************************************************************************
!> \brief creates a cart (multidimensional parallel environment)
!> \param cart the cart environment to create
!> \param group the mpi communicator
!> \param ndims the number of dimensions of the cart
!> \param owns_group if this object owns the underlying cart (and should
!>        free it)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_cart_create(cart, group, ndims, owns_group)
      TYPE(cp_para_cart_type), POINTER                   :: cart
      INTEGER, INTENT(in)                                :: group, ndims
      LOGICAL, INTENT(in), OPTIONAL                      :: owns_group

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cart_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(cart)))CALL cp__a("common/cp_para_env.F",155)
      ALLOCATE (cart)
      cart%owns_group = .TRUE.
      IF (PRESENT(owns_group)) cart%owns_group = owns_group
      cart%ndims = ndims
      cart%group = group

      ALLOCATE (cart%source(ndims), cart%periodic(ndims), cart%mepos(ndims), &
                cart%num_pe(ndims))

      cart%source = 0
      cart%mepos = 0
      cart%periodic = .FALSE.
      cart%ref_count = 1
      cart%ntask = 1
      CALL cp_cart_update(cart)
   END SUBROUTINE cp_cart_create

! **************************************************************************************************
!> \brief updates the information about the given cart
!> \param cart the cart to update
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_cart_update(cart)
      TYPE(cp_para_cart_type), POINTER                   :: cart

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cart_update', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(cart)))CALL cp__a("common/cp_para_env.F",183)
      IF(.NOT.(cart%ref_count > 0))CALL cp__a("common/cp_para_env.F",184)
      CALL mp_environ(cart%group, cart%ndims, cart%num_pe, task_coor=cart%mepos, &
                      periods=cart%periodic)
      CALL mp_environ(numtask=cart%ntask, taskid=cart%rank, groupid=cart%group)
   END SUBROUTINE cp_cart_update

! **************************************************************************************************
!> \brief releases the given cart
!> \param cart the cart to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_cart_release(cart)
      TYPE(cp_para_cart_type), POINTER                   :: cart

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_cart_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(cart)) THEN
         IF(.NOT.(cart%ref_count > 0))CALL cp__a("common/cp_para_env.F",202)
         cart%ref_count = cart%ref_count-1
         IF (cart%ref_count == 0) THEN
            IF (cart%owns_group) THEN
               CALL mp_comm_free(cart%group)
            END IF
            DEALLOCATE (cart%source, cart%periodic, cart%mepos, cart%num_pe)
            DEALLOCATE (cart)
         END IF
      END IF
      NULLIFY (cart)
   END SUBROUTINE cp_cart_release

END MODULE cp_para_env
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility routines to read data from files.
!>      Kept as close as possible to the old parser because
!>        1. string handling is a weak point of fortran compilers, and it is
!>           easy to write correct things that do not work
!>        2. conversion of old code
!> \par History
!>      22.11.1999 first version of the old parser (called qs_parser)
!>                 Matthias Krack
!>      06.2004 removed module variables, cp_parser_type, new module [fawzi]
!>      08.2008 Added buffering [tlaino]
!> \author fawzi
! **************************************************************************************************
MODULE cp_parser_types
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_para_env,                     ONLY: cp_para_env_create,&
                                              cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_buffer_types,          ONLY: buffer_type,&
                                              create_buffer_type,&
                                              release_buffer_type
   USE cp_parser_ilist_types,           ONLY: create_ilist_type,&
                                              ilist_type,&
                                              release_ilist_type
   USE cp_parser_inpp_types,            ONLY: create_inpp_type,&
                                              inpp_type,&
                                              release_inpp_type
   USE cp_parser_status_types,          ONLY: create_status_type,&
                                              release_status_type,&
                                              status_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              max_line_length
   USE message_passing,                 ONLY: MPI_COMM_SELF
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: cp_parser_type, parser_release, parser_create, &
             parser_reset, empty_initial_variables

   ! this is a zero sized array by choice, and convenience
   CHARACTER(LEN=default_path_length), DIMENSION(2, 1:0) :: empty_initial_variables

! *** Private parameters ***
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_parser_types'
   INTEGER, SAVE, PRIVATE               :: last_parser_id = 0

! *** Global variables ***
   CHARACTER(LEN=1), PARAMETER, PUBLIC   :: default_continuation_character = CHAR(92) ! backslash
   CHARACTER(LEN=4), PARAMETER, PUBLIC   :: default_separators = ",:;="
   CHARACTER(LEN=3), PARAMETER, PUBLIC   :: default_end_section_label = "END"
   CHARACTER(LEN=1), PARAMETER, PUBLIC   :: default_comment_character(2) = (/"#", "!"/), &
                                            default_section_character = "&"
   INTEGER, PARAMETER, PUBLIC            :: max_unit_number = 999

! **************************************************************************************************
!> \brief represent a parser
!> \param icol Number of the current column in the current input line,
!>  -1 if at the end of the file
!>  icol1            : First column of the current input string
!>  icol2            : Last column of the current input string
!> \param input_line_number Number of the current input line read from the input file
!> \param input_unit Logical unit number of the input file
!> \author fawzi
! **************************************************************************************************
   TYPE cp_parser_type
      INTEGER                                        :: id_nr, ref_count
      CHARACTER(LEN=default_string_length)           :: end_section, start_section
      CHARACTER(LEN=10)                              :: separators
      CHARACTER(LEN=1)                               :: comment_character(2), &
                                                        continuation_character, &
                                                        section_character
      CHARACTER(LEN=default_path_length)             :: input_file_name
      CHARACTER(LEN=max_line_length)                 :: input_line
      INTEGER                                        :: icol, icol1, icol2
      INTEGER                                        :: input_unit, input_line_number
      LOGICAL                                        :: first_separator, &
                                                        apply_preprocessing, &
                                                        parse_white_lines
      CHARACTER(len=default_path_length), DIMENSION(:, :), POINTER :: initial_variables
      TYPE(buffer_type), POINTER                     :: buffer
      TYPE(status_type), POINTER                     :: status
      TYPE(cp_para_env_type), POINTER                :: para_env
      TYPE(inpp_type), POINTER                       :: inpp
      TYPE(ilist_type), POINTER                      :: ilist
   END TYPE cp_parser_type

CONTAINS

! **************************************************************************************************
!> \brief retains the given parser
!> \param parser the parser to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE parser_retain(parser)
      TYPE(cp_parser_type), POINTER                      :: parser

      CHARACTER(len=*), PARAMETER :: routineN = 'parser_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_types.F",111)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_types.F",112)
      parser%ref_count = parser%ref_count+1
   END SUBROUTINE parser_retain

! **************************************************************************************************
!> \brief   releases the parser
!> \param parser ...
!> \date    14.02.2001
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parser_release(parser)
      TYPE(cp_parser_type), POINTER                      :: parser

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(parser)) THEN
         IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/cp_parser_types.F",129)
         parser%ref_count = parser%ref_count-1
         IF (parser%ref_count == 0) THEN
            IF (parser%input_unit >= 0) THEN
               CALL close_file(unit_number=parser%input_unit)
            END IF
            CALL cp_para_env_release(parser%para_env)
            CALL release_inpp_type(parser%inpp)
            CALL release_ilist_type(parser%ilist)
            CALL release_buffer_type(parser%buffer)
            CALL release_status_type(parser%status)
            IF (ASSOCIATED(parser%initial_variables)) THEN
               DEALLOCATE (parser%initial_variables)
            ENDIF
            DEALLOCATE (parser)
         END IF
      END IF
   END SUBROUTINE parser_release

! **************************************************************************************************
!> \brief   Start a parser run. Initial variables allow to @SET stuff before opening the file
!> \param parser ...
!> \param file_name ...
!> \param unit_nr ...
!> \param para_env ...
!> \param end_section_label ...
!> \param separator_chars ...
!> \param comment_char ...
!> \param continuation_char ...
!> \param section_char ...
!> \param parse_white_lines ...
!> \param initial_variables ...
!> \param apply_preprocessing ...
!> \date    14.02.2001
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE parser_create(parser, file_name, unit_nr, para_env, end_section_label, &
                            separator_chars, comment_char, continuation_char, section_char, parse_white_lines, &
                            initial_variables, apply_preprocessing)
      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: file_name
      INTEGER, INTENT(in), OPTIONAL                      :: unit_nr
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: end_section_label, separator_chars
      CHARACTER(LEN=1), INTENT(IN), OPTIONAL             :: comment_char, continuation_char, &
                                                            section_char
      LOGICAL, INTENT(IN), OPTIONAL                      :: parse_white_lines
      CHARACTER(len=*), DIMENSION(:, :), OPTIONAL        :: initial_variables
      LOGICAL, INTENT(IN), OPTIONAL                      :: apply_preprocessing

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(parser)))CALL cp__a("input/cp_parser_types.F",182)
      ALLOCATE (parser)
      last_parser_id = last_parser_id+1
      parser%id_nr = last_parser_id
      parser%ref_count = 1

      parser%input_unit = -1
      parser%input_file_name = ""
      NULLIFY (parser%initial_variables)

      !   *** Load the default values and overwrite them, if requested ***
      parser%separators = default_separators
      IF (PRESENT(separator_chars)) parser%separators = separator_chars
      parser%comment_character = default_comment_character
      IF (PRESENT(comment_char)) parser%comment_character = comment_char
      parser%continuation_character = default_continuation_character
      IF (PRESENT(continuation_char)) parser%continuation_character = continuation_char
      parser%section_character = default_section_character
      IF (PRESENT(section_char)) parser%section_character = section_char
      parser%end_section = parser%section_character//default_end_section_label
      IF (PRESENT(end_section_label)) THEN
         parser%end_section = parser%section_character//TRIM(end_section_label)
      END IF
      parser%parse_white_lines = .FALSE.
      IF (PRESENT(parse_white_lines)) THEN
         parser%parse_white_lines = parse_white_lines
      END IF
      parser%apply_preprocessing = .TRUE.
      IF (PRESENT(apply_preprocessing)) THEN
         parser%apply_preprocessing = apply_preprocessing
      END IF

      CALL compress(parser%end_section) ! needed?

      ! para_env
      IF (PRESENT(para_env)) THEN
         parser%para_env => para_env
         CALL cp_para_env_retain(para_env)
      ELSE
         NULLIFY (parser%para_env)
         CALL cp_para_env_create(parser%para_env, group=MPI_COMM_SELF, source=0, &
                                 mepos=0, num_pe=1, owns_group=.FALSE.)
      END IF

      !   *** Get the logical output unit number for error messages ***
      IF (parser%para_env%mepos == parser%para_env%source) THEN
         IF (PRESENT(unit_nr)) THEN
            parser%input_unit = unit_nr
            IF (PRESENT(file_name)) parser%input_file_name = file_name
         ELSE
            IF (.NOT. PRESENT(file_name)) &
               CALL cp__b("input/cp_parser_types.F",233,"at least one of filename and unit_nr must be present")
            CALL open_file(file_name=TRIM(file_name), &
                           unit_number=parser%input_unit)
            parser%input_file_name = file_name
         END IF
      END IF

      IF (PRESENT(initial_variables)) THEN
         IF (SIZE(initial_variables, 2) > 0) THEN
            ALLOCATE (parser%initial_variables(2, SIZE(initial_variables, 2)))
            parser%initial_variables = initial_variables
         ENDIF
      ENDIF

      parser%input_line_number = 0
      parser%icol = 0
      parser%icol1 = 0
      parser%icol2 = 0
      parser%first_separator = .TRUE.
      NULLIFY (parser%buffer)
      NULLIFY (parser%status)
      NULLIFY (parser%inpp)
      NULLIFY (parser%ilist)
      CALL create_inpp_type(parser%inpp, parser%initial_variables)
      CALL create_ilist_type(parser%ilist)
      CALL create_buffer_type(parser%buffer)
      CALL create_status_type(parser%status)
   END SUBROUTINE parser_create

! **************************************************************************************************
!> \brief   Resets the parser: rewinding the unit and re-initializing all
!>          parser structures
!> \param parser ...
!> \date    12.2008
!> \author  Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE parser_reset(parser)
      TYPE(cp_parser_type), POINTER                      :: parser

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parser_reset', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/cp_parser_types.F",274)
      ! Rewind units
      IF (parser%input_unit > 0) REWIND (parser%input_unit)
      ! Restore initial settings
      parser%input_line_number = 0
      parser%icol = 0
      parser%icol1 = 0
      parser%icol2 = 0
      parser%first_separator = .TRUE.
      ! Release substructures
      CALL release_inpp_type(parser%inpp)
      CALL release_ilist_type(parser%ilist)
      CALL release_buffer_type(parser%buffer)
      CALL release_status_type(parser%status)
      ! Reallocate substructures
      CALL create_inpp_type(parser%inpp, parser%initial_variables)
      CALL create_ilist_type(parser%ilist)
      CALL create_buffer_type(parser%buffer)
      CALL create_status_type(parser%status)
   END SUBROUTINE parser_reset

END MODULE cp_parser_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of mathematical constants and functions.
!> \par History
!>      Adapted for use in CP2K (JGH)
!>      JGH (16-06-2002) : Added Gamma functions
!>      JGH (10-08-2004) : Added Euler constant (gamma)
!> \author Matthias Krack
! **************************************************************************************************
MODULE mathconstants

   USE kinds,                           ONLY: dp

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: maxfac, pi, twopi, fourpi, rootpi, oorootpi, zero, one, &
             degree, radians, gaussi, fac, ifac, dfac, gamma0, gamma1, euler, &
             sqrthalf, sqrt2, sqrt3, sqrt5, sqrt7, sqrt15, sqrt21, sqrt35, &
             sqrt105

   ! Factorial function fac
   ! Inverse Factorial function ifac
   ! Double factorial function dfac
   ! Gamma functions
   ! gamma(n) = gamma0(n) = (n - 1)!
   ! gamma(n + 1/2) = gamma1(n) = SQRT(pi)/2^n (2n - 1)!!

   INTEGER, PARAMETER :: maxfac = 30
   REAL(KIND=dp), PARAMETER, DIMENSION(0:maxfac) :: fac = (/ &
                                      0.10000000000000000000E+01_dp, 0.10000000000000000000E+01_dp, 0.20000000000000000000E+01_dp, &
                                      0.60000000000000000000E+01_dp, 0.24000000000000000000E+02_dp, 0.12000000000000000000E+03_dp, &
                                      0.72000000000000000000E+03_dp, 0.50400000000000000000E+04_dp, 0.40320000000000000000E+05_dp, &
                                      0.36288000000000000000E+06_dp, 0.36288000000000000000E+07_dp, 0.39916800000000000000E+08_dp, &
                                      0.47900160000000000000E+09_dp, 0.62270208000000000000E+10_dp, 0.87178291200000000000E+11_dp, &
                                      0.13076743680000000000E+13_dp, 0.20922789888000000000E+14_dp, 0.35568742809600000000E+15_dp, &
                                      0.64023737057280000000E+16_dp, 0.12164510040883200000E+18_dp, 0.24329020081766400000E+19_dp, &
                                      0.51090942171709440000E+20_dp, 0.11240007277776076800E+22_dp, 0.25852016738884976640E+23_dp, &
                                      0.62044840173323943936E+24_dp, 0.15511210043330985984E+26_dp, 0.40329146112660563558E+27_dp, &
                                      0.10888869450418352161E+29_dp, 0.30488834461171386050E+30_dp, 0.88417619937397019545E+31_dp, &
                                                    0.26525285981219105864E+33_dp/)
   REAL(KIND=dp), PARAMETER, DIMENSION(0:maxfac) :: ifac = (/ &
                                      0.10000000000000000000E+01_dp, 0.10000000000000000000E+01_dp, 0.50000000000000000000E+00_dp, &
                                      0.16666666666666666667E+00_dp, 0.41666666666666666667E-01_dp, 0.83333333333333333333E-02_dp, &
                                      0.13888888888888888889E-02_dp, 0.19841269841269841270E-03_dp, 0.24801587301587301587E-04_dp, &
                                      0.27557319223985890653E-05_dp, 0.27557319223985890653E-06_dp, 0.25052108385441718775E-07_dp, &
                                      0.20876756987868098979E-08_dp, 0.16059043836821614599E-09_dp, 0.11470745597729724714E-10_dp, &
                                      0.76471637318198164759E-12_dp, 0.47794773323873852974E-13_dp, 0.28114572543455207632E-14_dp, &
                                      0.15619206968586226462E-15_dp, 0.82206352466243297170E-17_dp, 0.41103176233121648585E-18_dp, &
                                      0.19572941063391261231E-19_dp, 0.88967913924505732867E-21_dp, 0.38681701706306840377E-22_dp, &
                                      0.16117375710961183490E-23_dp, 0.64469502843844733962E-25_dp, 0.24795962632247974601E-26_dp, &
                                      0.91836898637955461484E-28_dp, 0.32798892370698379102E-29_dp, 0.11309962886447716932E-30_dp, &
                                                    0.37699876288159056439E-32_dp/)
   REAL(KIND=dp), PARAMETER, DIMENSION(-1:2*maxfac+1) :: dfac = (/ &
                                      0.10000000000000000000E+01_dp, 0.10000000000000000000E+01_dp, 0.10000000000000000000E+01_dp, &
                                      0.20000000000000000000E+01_dp, 0.30000000000000000000E+01_dp, 0.80000000000000000000E+01_dp, &
                                      0.15000000000000000000E+02_dp, 0.48000000000000000000E+02_dp, 0.10500000000000000000E+03_dp, &
                                      0.38400000000000000000E+03_dp, 0.94500000000000000000E+03_dp, 0.38400000000000000000E+04_dp, &
                                      0.10395000000000000000E+05_dp, 0.46080000000000000000E+05_dp, 0.13513500000000000000E+06_dp, &
                                      0.64512000000000000000E+06_dp, 0.20270250000000000000E+07_dp, 0.10321920000000000000E+08_dp, &
                                      0.34459425000000000000E+08_dp, 0.18579456000000000000E+09_dp, 0.65472907500000000000E+09_dp, &
                                      0.37158912000000000000E+10_dp, 0.13749310575000000000E+11_dp, 0.81749606400000000000E+11_dp, &
                                      0.31623414322500000000E+12_dp, 0.19619905536000000000E+13_dp, 0.79058535806250000000E+13_dp, &
                                      0.51011754393600000000E+14_dp, 0.21345804667687500000E+15_dp, 0.14283291230208000000E+16_dp, &
                                      0.61902833536293750000E+16_dp, 0.42849873690624000000E+17_dp, 0.19189878396251062500E+18_dp, &
                                      0.13711959580999680000E+19_dp, 0.63326598707628506250E+19_dp, 0.46620662575398912000E+20_dp, &
                                      0.22164309547669977187E+21_dp, 0.16783438527143608320E+22_dp, 0.82007945326378915594E+22_dp, &
                                      0.63777066403145711616E+23_dp, 0.31983098677287777082E+24_dp, 0.25510826561258284646E+25_dp, &
                                      0.13113070457687988603E+26_dp, 0.10714547155728479551E+27_dp, 0.56386202968058350995E+27_dp, &
                                      0.47144007485205310027E+28_dp, 0.25373791335626257948E+29_dp, 0.21686243443194442612E+30_dp, &
                                      0.11925681927744341235E+31_dp, 0.10409396852733332454E+32_dp, 0.58435841445947272053E+32_dp, &
                                      0.52046984263666662269E+33_dp, 0.29802279137433108747E+34_dp, 0.27064431817106664380E+35_dp, &
                                      0.15795207942839547636E+36_dp, 0.14614793181237598765E+37_dp, 0.86873643685617511998E+37_dp, &
                                      0.81842841814930553085E+38_dp, 0.49517976900801981839E+39_dp, 0.47468848252659720789E+40_dp, &
                                       0.29215606371473169285E+41_dp, 0.28481308951595832474E+42_dp, 0.17821519886598633264E+43_dp/)
   REAL(KIND=dp), PARAMETER, DIMENSION(0:maxfac) :: gamma0 = (/ &
                                      0.00000000000000000000E+00_dp, 0.10000000000000000000E+01_dp, 0.10000000000000000000E+01_dp, &
                                      0.20000000000000000000E+01_dp, 0.60000000000000000000E+01_dp, 0.24000000000000000000E+02_dp, &
                                      0.12000000000000000000E+03_dp, 0.72000000000000000000E+03_dp, 0.50400000000000000000E+04_dp, &
                                      0.40320000000000000000E+05_dp, 0.36288000000000000000E+06_dp, 0.36288000000000000000E+07_dp, &
                                      0.39916800000000000000E+08_dp, 0.47900160000000000000E+09_dp, 0.62270208000000000000E+10_dp, &
                                      0.87178291200000000000E+11_dp, 0.13076743680000000000E+13_dp, 0.20922789888000000000E+14_dp, &
                                      0.35568742809600000000E+15_dp, 0.64023737057280000000E+16_dp, 0.12164510040883200000E+18_dp, &
                                      0.24329020081766400000E+19_dp, 0.51090942171709440000E+20_dp, 0.11240007277776076800E+22_dp, &
                                      0.25852016738884976640E+23_dp, 0.62044840173323943936E+24_dp, 0.15511210043330985984E+26_dp, &
                                      0.40329146112660563558E+27_dp, 0.10888869450418352161E+29_dp, 0.30488834461171386050E+30_dp, &
                                                    0.88417619937397019545E+31_dp/)
   REAL(KIND=dp), PARAMETER, DIMENSION(0:maxfac) :: gamma1 = (/ &
                                      0.17724538509055160273E+01_dp, 0.88622692545275801365E+00_dp, 0.13293403881791370205E+01_dp, &
                                      0.33233509704478425512E+01_dp, 0.11631728396567448929E+02_dp, 0.52342777784553520181E+02_dp, &
                                      0.28788527781504436100E+03_dp, 0.18712543057977883465E+04_dp, 0.14034407293483412599E+05_dp, &
                                      0.11929246199460900709E+06_dp, 0.11332783889487855673E+07_dp, 0.11899423083962248457E+08_dp, &
                                      0.13684336546556585726E+09_dp, 0.17105420683195732157E+10_dp, 0.23092317922314238412E+11_dp, &
                                      0.33483860987355645697E+12_dp, 0.51899984530401250831E+13_dp, 0.85634974475162063871E+14_dp, &
                                      0.14986120533153361177E+16_dp, 0.27724322986333718178E+17_dp, 0.54062429823350750447E+18_dp, &
                                      0.11082798113786903842E+20_dp, 0.23828015944641843260E+21_dp, 0.53613035875444147334E+22_dp, &
                                      0.12599063430729374624E+24_dp, 0.30867705405286967828E+25_dp, 0.78712648783481767961E+26_dp, &
                                      0.20858851927622668510E+28_dp, 0.57361842800962338401E+29_dp, 0.16348125198274266444E+31_dp, &
                                                    0.48226969334909086011E+32_dp/)

   ! Constants related to Pi

   REAL(KIND=dp), PARAMETER :: pi = 3.14159265358979323846264338_dp ! Pi
   REAL(KIND=dp), PARAMETER :: pio2 = 1.57079632679489661923132169_dp ! Pi/2
   REAL(KIND=dp), PARAMETER :: twopi = 6.28318530717958647692528677_dp ! 2*Pi
   REAL(KIND=dp), PARAMETER :: fourpi = 12.56637061435917295385057353_dp ! 4*Pi
   REAL(KIND=dp), PARAMETER :: rootpi = 1.77245385090551602729816748_dp ! SQRT(Pi)
   REAL(KIND=dp), PARAMETER :: oorootpi = 0.56418958354775628694807945_dp ! 1/SQRT(Pi)

   ! Euler's constant (Euler-Mascheroni constant)

   REAL(KIND=dp), PARAMETER :: euler = 0.57721566490153286060651209_dp

   ! Trivial constants

   REAL(KIND=dp), PARAMETER :: zero = 0.0_dp, &
                               half = 0.5_dp, &
                               one = 1.0_dp

   REAL(KIND=dp), PARAMETER :: sqrthalf = 0.70710678118654752440084436_dp, & ! SQRT(1/2)
                               sqrt2 = 1.41421356237309504880168872_dp, & ! SQRT(2)
                               sqrt3 = 1.73205080756887729352744634_dp, & ! SQRT(3)
                               sqrt5 = 2.23606797749978969640917367_dp, & ! SQRT(5)
                               sqrt7 = 2.64575131106459059050161575_dp, & ! SQRT(7)
                               sqrt15 = 3.87298334620741688517926540_dp, & ! SQRT(15)
                               sqrt21 = 4.58257569495584000658804719_dp, & ! SQRT(21)
                               sqrt35 = 5.91607978309961604256732829_dp, & ! SQRT(35)
                               sqrt105 = 10.24695076595959838322103868_dp ! SQRT(105)

   ! Conversion factors

   REAL(KIND=dp), PARAMETER :: degree = 180.0_dp/pi, & ! radians -> degree
                               radians = one/degree ! degree  -> radians

   COMPLEX(KIND=dp), PARAMETER :: gaussi = (0.0_dp, 1.0_dp) ! i = SQRT(-1)

END MODULE mathconstants
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition of physical constants:
!>
!>      a_bohr      : Bohr radius [m]
!>      a_fine      : Fine-structure constant
!>      a_mass      : Atomic mass unit [kg]; conversion factor [u] -> [kg]
!>      angstrom    : Conversion factor [Bohr] -> [Angstrom]
!>      bar         : Conversion factor [a.u.] -> [bar]
!>      bohr        : Conversion factor [Angstrom] -> [Bohr]
!>      boltzmann   : Boltzmann constant [J/K]
!>      c_light     : Speed of light in vacuum [m/s]
!>      e_charge    : Elementary charge [C]
!>      e_mass      : Electron mass [kg]
!>      e_gfactor   : Electron g factor [ ]
!>      esu         : Conversion factors [a.u.] -> [esu]
!>      evolt       : Conversion factor [a.u.] -> [eV]
!>      femtoseconds: Conversion factor [a.u.] -> [fs]
!>      h_bar       : Planck constant [J*s]
!>      h_planck    : Planck constant [J*s]
!>      hertz       : Conversion factor [a.u.] -> [Hz]
!>      joule       : Conversion factor [a.u.] -> [J]
!>      kcalmol     : Conversion factor [a.u.] -> [kcal/mol]
!>      kelvin      : Conversion factor [a.u.] -> [K]
!>      kjmol       : Conversion factor [a.u.] -> [kJ/mol]
!>      massunit    : Conversion factor [u] -> [a.u.]
!>      mu_perm     : Magnetic constant or permeability of vacuum [N/A**2]
!>      n_avogadro  : Avogadro constant [1/mol]
!>      newton      : Conversion factor [a.u.] -> [N]
!>      pascal      : Conversion factor [a.u.] -> [Pa]
!>      permittivity: Electric constant or permittivity of vacuum [F/m]
!>      picoseconds : Conversion factor [a.u.] -> [ps]
!>      rydberg     : Rydberg constant [1/m]
!>      seconds     : Conversion factor [a.u.] -> [s]
!>      vibfac      : Conversion factor [a.u./Bohr**2] -> [1/cm]
!>      wavenumbers : Conversion factor [a.u.] -> [1/cm]
!>      debye       : Conversion factor [a.u.] -> Debye
!> \note
!>      Fundamental physical constants (SI units)
!>      Literature: - P. J. Mohr and B. N. Taylor,
!>                    "CODATA recommended values of the fundamental physical
!>                     constants: 1998 Rev. Mod. Phys. 72, 351-495 (2000)
!>                  - P. J. Mohr and B. N. Taylor,
!>                    "CODATA recommended values of the fundamental physical
!>                     constants: 2002", Rev. Mod. Phys. 77, 1 (2005).
!>                  - P. J. Mohr, B. N. Taylor, and D. B. Newell,
!>                    "CODATA recommended values of the fundamental physical
!>                     constants: 2006 Rev. Mod. Phys. 80, 633 (2008)
!>                  - P. J. Mohr, B. N. Taylor, and D. B. Newell,
!>                    "CODATA recommended values of the fundamental physical
!>                     constants: 2010", Rev. Mod. Phys. 84, 1527-1605 (2012)
!> \par History
!>      - Adapted for use in CP2K (JGH)
!>      - Updated to CODATA 1998 and cleaned (05.09.2003,MK)
!>      - Updated to CODATA 2006. (26.03.2008,AK)
!>      - Updated to CODATA 2010. (10.12.2012,MK)
!>      - Turned constants into Fortran parameters (2014, Ole Schuett)
!>      - Remove all but CODATA 2006 (2015, Ole Schuett)
!> \author Matthias Krack (MK)
! **************************************************************************************************
MODULE physcon

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: a_bohr, a_fine, a_mass, angstrom, atm, bar, bohr, boltzmann, &
             debye, e_charge, e_gfactor, e_mass, evolt, femtoseconds, h_bar, &
             hertz, joule, kcalmol, kelvin, kjmol, massunit, mu_perm, n_avogadro, newton, &
             p_mass, pascal, picoseconds, seconds, vibfac, wavenumbers

   PUBLIC :: write_physcon

   ! CP2K uses the CODATA from 2006

   ! Exact constants
   ! Speed of light in vacuum [m/s]
   REAL(KIND=dp), PARAMETER :: c_light = 299792458.0_dp

   ! Magnetic constant or permeability of vacuum [N/A**2]
   REAL(KIND=dp), PARAMETER :: mu_perm = 4.0_dp*pi*1.0E-7_dp

   ! Electric constant or permittivity of vacuum [F/m]
   REAL(KIND=dp), PARAMETER :: permittivity = 1.0_dp/(mu_perm*c_light**2)

   ! Recommended fundamental constants of physics
   ! and chemistry based on the 2006 adjustment

   ! Planck constant [J*s]
   REAL(KIND=dp), PARAMETER :: h_planck = 6.62606896E-34_dp
   REAL(KIND=dp), PARAMETER :: h_bar = h_planck/(2.0_dp*pi)

   ! Elementary charge [C]
   REAL(KIND=dp), PARAMETER :: e_charge = 1.602176487E-19_dp

   ! Electron mass [kg]
   REAL(KIND=dp), PARAMETER :: e_mass = 9.10938215E-31_dp

   ! Proton mass [kg]
   REAL(KIND=dp), PARAMETER :: p_mass = 1.672621637E-27_dp

   ! Electron g factor [ ]
   REAL(KIND=dp), PARAMETER :: e_gfactor = -2.0023193043622_dp

   ! Fine-structure constant
!MK a_fine = 0.5_dp*mu_perm*c_light*e_charge**2/h_planck
   REAL(KIND=dp), PARAMETER :: a_fine = 7.2973525376E-3_dp

   ! Rydberg constant [1/m]
!MK rydberg = 0.5_dp*e_mass*c_light*a_fine**2/h_planck
   REAL(KIND=dp), PARAMETER :: rydberg = 10973731.568527_dp

   ! Avogadro constant [1/mol]
   REAL(KIND=dp), PARAMETER :: n_avogadro = 6.02214179E+23_dp

   ! Boltzmann constant [J/K]
   REAL(KIND=dp), PARAMETER :: boltzmann = 1.3806504E-23_dp

   ! Atomic mass unit [kg]; conversion factor [u] -> [kg]
   REAL(KIND=dp), PARAMETER :: a_mass = 1.660538782E-27_dp

   ! Bohr radius [m]
!MK a_bohr = a_fine/(4.0_dp*pi*rydberg)
   REAL(KIND=dp), PARAMETER :: a_bohr = 0.52917720859E-10_dp

   ! Conversion factors

   ! [u] -> [a.u.]
   REAL(KIND=dp), PARAMETER :: massunit = a_mass/e_mass

   ! [Bohr] -> [Angstrom]
   REAL(KIND=dp), PARAMETER :: angstrom = 1.0E+10_dp*a_bohr

   ! [Angstrom] -> [Bohr]
   REAL(KIND=dp), PARAMETER :: bohr = 1.0_dp/angstrom

   ! [a.u.] -> [s]
   REAL(KIND=dp), PARAMETER :: seconds = 1.0_dp/(4.0_dp*pi*rydberg*c_light)

   ! [a.u.] -> [fs]
   REAL(KIND=dp), PARAMETER :: femtoseconds = 1.0E+15_dp*seconds

   ! [a.u.] -> [ps]
   REAL(KIND=dp), PARAMETER :: picoseconds = 1.0E+12_dp*seconds

   ! [a.u.] -> [J]
   REAL(KIND=dp), PARAMETER :: joule = 2.0_dp*rydberg*h_planck*c_light

   ! [a.u.] -> [N]
   REAL(KIND=dp), PARAMETER :: newton = joule/a_bohr

   ! [a.u.] -> [K]
   REAL(KIND=dp), PARAMETER :: kelvin = joule/boltzmann

   ! [a.u.] -> [kJ/mol]
   REAL(KIND=dp), PARAMETER :: kjmol = 0.001_dp*joule*n_avogadro

   ! [a.u.] -> [kcal/mol]
   REAL(KIND=dp), PARAMETER :: kcalmol = kjmol/4.184_dp

   ! [a.u.] -> [Pa]
   REAL(KIND=dp), PARAMETER :: pascal = joule/a_bohr**3

   ! [a.u.] -> [bar]
   REAL(KIND=dp), PARAMETER :: bar = pascal/1.0E+5_dp

   ! [a.u.] -> [atm]
   REAL(KIND=dp), PARAMETER :: atm = pascal/1.013250E+5_dp

   ! [a.u.] -> [eV]
   REAL(KIND=dp), PARAMETER :: evolt = joule/e_charge

   ! [a.u.] -> [Hz]
   REAL(KIND=dp), PARAMETER :: hertz = joule/h_planck

   ! [a.u./Bohr**2] -> [1/cm] (wave numbers)
   REAL(KIND=dp), PARAMETER :: vibfac = 5.0_dp*SQRT(kjmol)/(pi*a_bohr*c_light)

   ! [a.u.] -> [1/cm] (wave numbers)
   REAL(KIND=dp), PARAMETER :: wavenumbers = 0.02_dp*rydberg

   ! [a.u.] -> [esu] (electrostatic units)
   REAL(KIND=dp), PARAMETER :: esu_1 = 1.0E+21_dp*a_bohr*c_light*e_charge
   !REAL(KIND=dp), PARAMETER :: esu_2  = esu_1/bohr
   !REAL(KIND=dp), PARAMETER :: esu_3  = esu_2/bohr
   !REAL(KIND=dp), PARAMETER :: esu(3) = (/ esu_1, esu_2, esu_3 /)

   ! [a.u.] -> [debye] (electrostatic units)
   REAL(KIND=dp), PARAMETER :: debye = esu_1

CONTAINS

! **************************************************************************************************
!> \brief  Write all basic physical constants used by CP2K to a logical
!>           output unit.
!> \param output_unit ...
!> \date    14.11.2000
!> \par History
!>       - Updated to CODATA 1998 and cleaned (05.09.2003,MK)
!>       - Updated to CODATA 2006. (26.03.2008,AK)
!> \author  JGH
!> \version 1.1
! **************************************************************************************************
   SUBROUTINE write_physcon(output_unit)

      INTEGER, INTENT(IN)                                :: output_unit

      WRITE (UNIT=output_unit, FMT="(T2,/,T2,A,/,/,(T2,A))") &
         "*** Fundamental physical constants (SI units) ***", &
         "*** Literature: B. J. Mohr and B. N. Taylor,", &
         "***             CODATA recommended values of the fundamental physical", &
         "***             constants: 2006, Web Version 5.1", &
         "***             http://physics.nist.gov/constants"

      WRITE (UNIT=output_unit, FMT="(/,T2,A,T61,ES20.14)") &
         "Speed of light in vacuum [m/s]", c_light
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Magnetic constant or permeability of vacuum [N/A**2]", mu_perm
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Electric constant or permittivity of vacuum [F/m]", permittivity
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Planck constant (h) [J*s]", h_planck
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Planck constant (h-bar) [J*s]", h_bar
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Elementary charge [C]", e_charge
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Electron mass [kg]", e_mass
      WRITE (UNIT=output_unit, FMT="(T2,A,T60,ES21.14)") &
         "Electron g factor [ ]", e_gfactor
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Proton mass [kg]", p_mass
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Fine-structure constant", a_fine
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Rydberg constant [1/m]", rydberg
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Avogadro constant [1/mol]", n_avogadro
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Boltzmann constant [J/K]", boltzmann
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Atomic mass unit [kg]", a_mass
      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "Bohr radius [m]", a_bohr

      ! Conversion factors

      WRITE (UNIT=output_unit, FMT="(/,T2,A,/)") &
         "*** Conversion factors ***"

      WRITE (UNIT=output_unit, FMT="(T2,A,T61,ES20.14)") &
         "[u] -> [a.u.]", massunit, &
         "[Angstrom] -> [Bohr] = [a.u.]", bohr, &
         "[a.u.] = [Bohr] -> [Angstrom]", angstrom, &
         "[a.u.] -> [s]", seconds, &
         "[a.u.] -> [fs]", femtoseconds, &
         "[a.u.] -> [J]", joule, &
         "[a.u.] -> [N]", newton, &
         "[a.u.] -> [K]", kelvin, &
         "[a.u.] -> [kJ/mol]", kjmol, &
         "[a.u.] -> [kcal/mol]", kcalmol, &
         "[a.u.] -> [Pa]", pascal, &
         "[a.u.] -> [bar]", bar, &
         "[a.u.] -> [atm]", atm, &
         "[a.u.] -> [eV]", evolt, &
         "[a.u.] -> [Hz]", hertz, &
         "[a.u.] -> [1/cm] (wave numbers)", wavenumbers, &
         "[a.u./Bohr**2] -> [1/cm]", vibfac
      WRITE (UNIT=output_unit, FMT="(T2,A)") ""

   END SUBROUTINE write_physcon

END MODULE physcon
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief unit conversion facility
!>
!>      Units are complex, this module does not try to be very smart, for
!>      example SI prefixes are not supported automatically, and
!>      which kinds are really basic can change depending on the system of
!>      units choosen, and equivalences are not always catched.
!>
!>      This is thought as a simple conversion facility for the input and output.
!>      If you need something more you are probably better off using the
!>      physcon module directly.
!> \note
!>      One design choice was not to use dynamically allocated elements to
!>      reduce the possibility of leaks.
!>      Needs to be extended (for example charge, dipole,...)
!>      I just added the units and kinds that I needed.
!>      Used by the parser
!>      Should keep an unsorted/uncompressed version for nicer labels?
!> \par History
!>      01.2005 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE cp_units

   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: radians,&
                                              twopi
   USE physcon,                         ONLY: &
        atm, bar, bohr, e_mass, evolt, femtoseconds, joule, kcalmol, kelvin, kjmol, massunit, &
        newton, pascal, picoseconds, seconds, wavenumbers
   USE string_utilities,                ONLY: compress,&
                                              s2a,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER          :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_units'
   INTEGER, SAVE, PRIVATE               :: last_unit_id = 0, last_unit_set_id = 0

   INTEGER, PARAMETER, PUBLIC :: cp_ukind_none = 0, &
                                 cp_ukind_energy = 1, &
                                 cp_ukind_length = 2, &
                                 cp_ukind_temperature = 3, &
                                 cp_ukind_angle = 4, &
                                 cp_ukind_pressure = 5, &
                                 cp_ukind_time = 6, &
                                 cp_ukind_mass = 7, &
                                 cp_ukind_undef = 8, &
                                 cp_ukind_potential = 9, &
                                 cp_ukind_force = 10, &
                                 cp_ukind_max = 10

   ! General
   INTEGER, PARAMETER, PUBLIC :: cp_units_none = 100, &
                                 cp_units_au = 101
   ! Mass
   INTEGER, PARAMETER, PUBLIC :: cp_units_m_e = 110, &
                                 cp_units_amu = 111, &
                                 cp_units_kg = 112
   ! Energy
   INTEGER, PARAMETER, PUBLIC :: cp_units_hartree = 130, &
                                 cp_units_wavenum = 131, &
                                 cp_units_joule = 132, &
                                 cp_units_kcalmol = 133, &
                                 cp_units_Ry = 134, &
                                 cp_units_eV = 135, &
                                 cp_units_kjmol = 136, &
                                 cp_units_jmol = 137, &
                                 cp_units_keV = 138

   ! Length
   INTEGER, PARAMETER, PUBLIC :: cp_units_bohr = 140, &
                                 cp_units_angstrom = 141, &
                                 cp_units_m = 142, &
                                 cp_units_pm = 143, &
                                 cp_units_nm = 144

   ! Temperature
   INTEGER, PARAMETER, PUBLIC :: cp_units_k = 150

   ! Pressure
   INTEGER, PARAMETER, PUBLIC :: cp_units_bar = 161
   INTEGER, PARAMETER, PUBLIC :: cp_units_atm = 162
   INTEGER, PARAMETER, PUBLIC :: cp_units_kbar = 163
   INTEGER, PARAMETER, PUBLIC :: cp_units_Pa = 164
   INTEGER, PARAMETER, PUBLIC :: cp_units_MPa = 165
   INTEGER, PARAMETER, PUBLIC :: cp_units_GPa = 166

   ! Angles
   INTEGER, PARAMETER, PUBLIC :: cp_units_rad = 170, &
                                 cp_units_deg = 171

   ! Time
   INTEGER, PARAMETER, PUBLIC :: cp_units_fs = 180, &
                                 cp_units_s = 181, &
                                 cp_units_wn = 182, &
                                 cp_units_ps = 183

   ! Potential
   INTEGER, PARAMETER, PUBLIC :: cp_units_volt = 190

   ! Force
   INTEGER, PARAMETER, PUBLIC :: cp_units_Newton = 200, &
                                 cp_units_mNewton = 201

   INTEGER, PARAMETER, PUBLIC :: cp_unit_max_kinds = 8, cp_unit_basic_desc_length = 15, &
                                 cp_unit_desc_length = cp_unit_max_kinds*cp_unit_basic_desc_length

   PUBLIC :: cp_unit_type, cp_unit_set_type
   PUBLIC :: cp_unit_create, cp_unit_retain, cp_unit_release, &
             cp_unit_to_cp2k, cp_unit_from_cp2k, cp_unit_desc, &
             cp_unit_set_create, cp_unit_set_release, &
             cp_unit_to_cp2k1, cp_unit_from_cp2k1, cp_unit_compatible, print_all_units

! **************************************************************************************************
!> \brief stores a unit
!> \param ref_count the reference count (see doc/ReferenceCounting.html)
!> \param id_nr identification number (unique for each instance)
!> \param kind the kind of unit (energy, length,...)
!> \param unit the actual unit (Joule, eV,...)
!> \author fawzi
! **************************************************************************************************
   TYPE cp_unit_type
      INTEGER :: id_nr, ref_count, n_kinds
      INTEGER, DIMENSION(cp_unit_max_kinds):: kind_id, unit_id, power
   END TYPE cp_unit_type

! **************************************************************************************************
!> \brief represent a pointer to a unit (to build arrays of pointers)
!> \param unit the pointer to the unit
!> \author fawzi
! **************************************************************************************************
   TYPE cp_unit_p_type
      TYPE(cp_unit_type), POINTER :: unit
   END TYPE cp_unit_p_type

! **************************************************************************************************
!> \brief stores the default units to be used
!> \author fawzi
! **************************************************************************************************
   TYPE cp_unit_set_type
      INTEGER :: id_nr, ref_count
      TYPE(cp_unit_p_type), DIMENSION(cp_ukind_max) :: units
   END TYPE cp_unit_set_type

CONTAINS

! **************************************************************************************************
!> \brief creates a unit parsing a string
!> \param unit the unit to initialize
!> \param string the string containing the description of the unit
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_unit_create(unit, string)
      TYPE(cp_unit_type), POINTER                        :: unit
      CHARACTER(len=*), INTENT(in)                       :: string

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_create', routineP = moduleN//':'//routineN

      CHARACTER(default_string_length)                   :: desc
      CHARACTER(LEN=40)                                  :: formatstr
      INTEGER                                            :: i_high, i_low, i_unit, len_string, &
                                                            next_power
      INTEGER, DIMENSION(cp_unit_max_kinds)              :: kind_id, power, unit_id

      unit_id = cp_units_none
      kind_id = cp_ukind_none
      power = 0
      i_low = 1
      i_high = 1
      len_string = LEN(string)
      i_unit = 0
      next_power = 1
      DO WHILE (i_low < len_string)
         IF (string(i_low:i_low) /= ' ') EXIT
         i_low = i_low+1
      END DO
      i_high = i_low
      DO WHILE (i_high <= len_string)
         IF (string(i_high:i_high) == ' ' .OR. string(i_high:i_high) == '^' .OR. &
             string(i_high:i_high) == '*' .OR. string(i_high:i_high) == '/') EXIT
         i_high = i_high+1
      END DO
      DO
         IF (i_high <= i_low .OR. i_low > len_string) EXIT
         i_unit = i_unit+1
         IF (i_unit > cp_unit_max_kinds) THEN
            CALL cp__b("common/cp_units.F",198,"Maximum number of combined units exceeded")
            EXIT
         END IF
         ! read unit
         SELECT CASE (string (i_low:i_high-1))
         CASE ("internal_cp2k")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_undef
         CASE ("hartree")
            unit_id(i_unit) = cp_units_hartree
            kind_id(i_unit) = cp_ukind_energy
         CASE ("au_e")
            unit_id(i_unit) = cp_units_au
            kind_id(i_unit) = cp_ukind_energy
         CASE ("wavenumber_e")
            unit_id(i_unit) = cp_units_wavenum
            kind_id(i_unit) = cp_ukind_energy
         CASE ("joule", "J")
            unit_id(i_unit) = cp_units_joule
            kind_id(i_unit) = cp_ukind_energy
         CASE ("kcalmol")
            unit_id(i_unit) = cp_units_kcalmol
            kind_id(i_unit) = cp_ukind_energy
         CASE ("kjmol")
            unit_id(i_unit) = cp_units_kjmol
            kind_id(i_unit) = cp_ukind_energy
         CASE ("jmol")
            unit_id(i_unit) = cp_units_jmol
            kind_id(i_unit) = cp_ukind_energy
         CASE ("Ry")
            unit_id(i_unit) = cp_units_Ry
            kind_id(i_unit) = cp_ukind_energy
         CASE ("eV")
            unit_id(i_unit) = cp_units_eV
            kind_id(i_unit) = cp_ukind_energy
         CASE ("keV")
            unit_id(i_unit) = cp_units_keV
            kind_id(i_unit) = cp_ukind_energy
         CASE ("K_e")
            unit_id(i_unit) = cp_units_k
            kind_id(i_unit) = cp_ukind_energy
         CASE ("energy")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_energy
         CASE ("au_l")
            unit_id(i_unit) = cp_units_au
            kind_id(i_unit) = cp_ukind_length
         CASE ("bohr")
            unit_id(i_unit) = cp_units_bohr
            kind_id(i_unit) = cp_ukind_length
         CASE ("m")
            unit_id(i_unit) = cp_units_m
            kind_id(i_unit) = cp_ukind_length
         CASE ("pm")
            unit_id(i_unit) = cp_units_pm
            kind_id(i_unit) = cp_ukind_length
         CASE ("nm")
            unit_id(i_unit) = cp_units_nm
            kind_id(i_unit) = cp_ukind_length
         CASE ("angstrom")
            unit_id(i_unit) = cp_units_angstrom
            kind_id(i_unit) = cp_ukind_length
         CASE ("length")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_length
         CASE ("K", "K_temp")
            unit_id(i_unit) = cp_units_k
            kind_id(i_unit) = cp_ukind_temperature
         CASE ("au_temp")
            unit_id(i_unit) = cp_units_au
            kind_id(i_unit) = cp_ukind_temperature
         CASE ("temperature")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_temperature
         CASE ("atm")
            unit_id(i_unit) = cp_units_atm
            kind_id(i_unit) = cp_ukind_pressure
         CASE ("bar")
            unit_id(i_unit) = cp_units_bar
            kind_id(i_unit) = cp_ukind_pressure
         CASE ("kbar")
            unit_id(i_unit) = cp_units_kbar
            kind_id(i_unit) = cp_ukind_pressure
         CASE ("Pa")
            unit_id(i_unit) = cp_units_Pa
            kind_id(i_unit) = cp_ukind_pressure
         CASE ("MPa")
            unit_id(i_unit) = cp_units_MPa
            kind_id(i_unit) = cp_ukind_pressure
         CASE ("GPa")
            unit_id(i_unit) = cp_units_GPa
            kind_id(i_unit) = cp_ukind_pressure
         CASE ("au_p")
            unit_id(i_unit) = cp_units_au
            kind_id(i_unit) = cp_ukind_pressure
         CASE ("pressure")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_pressure
         CASE ("rad")
            unit_id(i_unit) = cp_units_rad
            kind_id(i_unit) = cp_ukind_angle
         CASE ("deg")
            unit_id(i_unit) = cp_units_deg
            kind_id(i_unit) = cp_ukind_angle
         CASE ("angle")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_angle
         CASE ("s")
            unit_id(i_unit) = cp_units_s
            kind_id(i_unit) = cp_ukind_time
         CASE ("fs")
            unit_id(i_unit) = cp_units_fs
            kind_id(i_unit) = cp_ukind_time
         CASE ("ps")
            unit_id(i_unit) = cp_units_ps
            kind_id(i_unit) = cp_ukind_time
         CASE ("wavenumber_t")
            unit_id(i_unit) = cp_units_wn
            kind_id(i_unit) = cp_ukind_time
         CASE ("au_t")
            unit_id(i_unit) = cp_units_au
            kind_id(i_unit) = cp_ukind_time
         CASE ("time")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_time
         CASE ("kg")
            unit_id(i_unit) = cp_units_kg
            kind_id(i_unit) = cp_ukind_mass
         CASE ("amu")
            unit_id(i_unit) = cp_units_amu
            kind_id(i_unit) = cp_ukind_mass
         CASE ("m_e")
            unit_id(i_unit) = cp_units_m_e
            kind_id(i_unit) = cp_ukind_mass
         CASE ("au_m")
            unit_id(i_unit) = cp_units_au
            kind_id(i_unit) = cp_ukind_mass
         CASE ("mass")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_mass
         CASE ("volt")
            unit_id(i_unit) = cp_units_volt
            kind_id(i_unit) = cp_ukind_potential
         CASE ("au_pot")
            unit_id(i_unit) = cp_units_au
            kind_id(i_unit) = cp_ukind_potential
         CASE ("potential")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_potential
         CASE ("N", "Newton")
            unit_id(i_unit) = cp_units_Newton
            kind_id(i_unit) = cp_ukind_force
         CASE ("mN", "mNewton")
            unit_id(i_unit) = cp_units_mNewton
            kind_id(i_unit) = cp_ukind_force
         CASE ("au_f")
            unit_id(i_unit) = cp_units_au
            kind_id(i_unit) = cp_ukind_force
         CASE ("force")
            unit_id(i_unit) = cp_units_none
            kind_id(i_unit) = cp_ukind_force
         CASE ("au")
            CALL cp_abort(cp__l("common/cp_units.F",360), &
                          "au unit without specifing its kind not accepted, use "// &
                          "(au_e, au_f, au_t, au_temp, au_l, au_m, au_p, au_pot)")
         CASE default
            CALL cp__b("common/cp_units.F",364,"Unknown unit: "//string(i_low:i_high-1))
         END SELECT
         power(i_unit) = next_power
         ! parse op
         i_low = i_high
         DO WHILE (i_low <= len_string)
            IF (string(i_low:i_low) /= ' ') EXIT
            i_low = i_low+1
         END DO
         i_high = i_low
         DO WHILE (i_high <= len_string)
            IF (string(i_high:i_high) == ' ' .OR. string(i_high:i_high) == '^' .OR. &
                string(i_high:i_high) == '*' .OR. string(i_high:i_high) == '/') EXIT
            i_high = i_high+1
         END DO
         IF (i_high < i_low .OR. i_low > len_string) EXIT

         IF (i_high <= len_string) THEN
            IF (string(i_low:i_high) == '^') THEN
               i_low = i_high+1
               DO WHILE (i_low <= len_string)
                  IF (string(i_low:i_low) /= ' ') EXIT
                  i_low = i_low+1
               END DO
               i_high = i_low
               DO WHILE (i_high <= len_string)
                  SELECT CASE (string (i_high:i_high))
                  CASE ('+', '-', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9')
                     i_high = i_high+1
                  CASE default
                     EXIT
                  END SELECT
               END DO
               IF (i_high <= i_low .OR. i_low > len_string) THEN
                  CALL cp__b("common/cp_units.F",398,"an integer number is expected after a '^'")
                  EXIT
               END IF
               formatstr = "(i"//cp_to_string(i_high-i_low+1)//")"
               READ (string(i_low:i_high-1), formatstr) &
                  next_power
               power(i_unit) = power(i_unit)*next_power
               ! next op
               i_low = i_high
               DO WHILE (i_low < len_string)
                  IF (string(i_low:i_low) /= ' ') EXIT
                  i_low = i_low+1
               END DO
               i_high = i_low
               DO WHILE (i_high <= len_string)
                  IF (string(i_high:i_high) == ' ' .OR. string(i_high:i_high) == '^' .OR. &
                      string(i_high:i_high) == '*' .OR. string(i_high:i_high) == '/') EXIT
                  i_high = i_high+1
               END DO
            END IF
         ENDIF
         IF (i_low > len_string) EXIT
         next_power = 1
         IF (i_high <= len_string) THEN
            IF (string(i_low:i_high) == "*" .OR. string(i_low:i_high) == '/') THEN
               IF (string(i_low:i_high) == '/') next_power = -1
               i_low = i_high+1
               DO WHILE (i_low <= len_string)
                  IF (string(i_low:i_low) /= ' ') EXIT
                  i_low = i_low+1
               END DO
               i_high = i_low
               DO WHILE (i_high <= len_string)
                  IF (string(i_high:i_high) == ' ' .OR. string(i_high:i_high) == '^' .OR. &
                      string(i_high:i_high) == '*' .OR. string(i_high:i_high) == '/') EXIT
                  i_high = i_high+1
               END DO
            END IF
         ENDIF
      END DO
      CALL cp_unit_create2(unit, kind_id=kind_id, unit_id=unit_id, &
                           power=power)
      desc = cp_unit_desc(unit)
   END SUBROUTINE cp_unit_create

! **************************************************************************************************
!> \brief creates and initializes the given unit of mesure (performs some error
!>      check)
!> \param unit the unit descriptor to be initialized
!> \param kind_id the kind of unit (length,energy,...), use the constants
!>        cp_ukind_*
!> \param unit_id the actual unit (use constants cp_units_*)
!> \param power ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_unit_create2(unit, kind_id, unit_id, power)
      TYPE(cp_unit_type), POINTER                        :: unit
      INTEGER, DIMENSION(:), INTENT(in)                  :: kind_id, unit_id
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: power

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_create2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, max_kind, max_pos
      LOGICAL                                            :: repeat

      IF(.NOT.(.NOT. ASSOCIATED(unit)))CALL cp__a("common/cp_units.F",464)
      IF(.NOT.(SIZE(kind_id) <= cp_unit_max_kinds))CALL cp__a("common/cp_units.F",465)
      IF(.NOT.(SIZE(unit_id) <= cp_unit_max_kinds))CALL cp__a("common/cp_units.F",466)
      ALLOCATE (unit)
      unit%ref_count = 1
      last_unit_id = last_unit_id+1
      unit%id_nr = last_unit_id
      unit%kind_id(1:SIZE(kind_id)) = kind_id
      unit%kind_id(SIZE(kind_id)+1:) = cp_ukind_none
      unit%unit_id(1:SIZE(unit_id)) = unit_id
      unit%unit_id(SIZE(unit_id):) = cp_units_none
      IF (PRESENT(power)) THEN
         unit%power(1:SIZE(power)) = power
         unit%power(SIZE(power)+1:) = 0
         DO i = 1, SIZE(unit%power)
            IF (unit%power(i) == 0) THEN
               unit%kind_id(i) = cp_ukind_none
               unit%unit_id(i) = cp_units_none
            END IF
         END DO
      ELSE
         DO i = 1, SIZE(unit%power)
            IF (unit%unit_id(i) /= 0) THEN
               unit%power(i) = 1
            ELSE
               unit%power(i) = 0
            END IF
         END DO
      END IF

      ! remove unnecessary units
      ! reorder & compress
      unit%n_kinds = 0
      DO i = 1, SIZE(unit%kind_id)
         ! find max and compress in the rest
         DO
            max_kind = unit%kind_id(i)
            max_pos = i
            repeat = .FALSE.
            DO j = i+1, SIZE(unit%kind_id)
               IF (unit%kind_id(j) >= max_kind) THEN
                  IF (unit%kind_id(j) /= 0 .AND. unit%kind_id(j) == max_kind .AND. &
                      unit%unit_id(j) == unit%unit_id(max_pos)) THEN
                     unit%power(max_pos) = unit%power(max_pos)+unit%power(j)
                     unit%kind_id(j) = cp_ukind_none
                     unit%unit_id(j) = cp_units_none
                     unit%power(j) = 0
                     IF (unit%power(max_pos) == 0) THEN
                        unit%kind_id(max_pos) = cp_ukind_none
                        unit%unit_id(max_pos) = cp_units_none
                        unit%power(max_pos) = 0
                        repeat = .TRUE.
                        EXIT
                     END IF
                  ELSE IF (unit%kind_id(j) > max_kind .OR. &
                           (unit%kind_id(j) == max_kind .AND. &
                            unit%unit_id(j) > unit%unit_id(max_pos))) THEN
                     max_kind = unit%kind_id(j)
                     max_pos = j
                  END IF
               END IF
            END DO
            IF (.NOT. repeat) EXIT
         END DO
         IF (max_kind /= 0) unit%n_kinds = unit%n_kinds+1
         ! put the max at pos i
         IF (max_pos /= i) THEN
            unit%kind_id(max_pos) = unit%kind_id(i)
            unit%kind_id(i) = max_kind
            max_kind = unit%unit_id(max_pos)
            unit%unit_id(max_pos) = unit%unit_id(i)
            unit%unit_id(i) = max_kind
            max_kind = unit%power(max_pos)
            unit%power(max_pos) = unit%power(i)
            unit%power(i) = max_kind
         END IF
         ! check unit
         CALL cp_basic_unit_check(basic_kind=unit%kind_id(i), &
                                  basic_unit=unit%unit_id(i))
      END DO
   END SUBROUTINE cp_unit_create2

! **************************************************************************************************
!> \brief retains the given unit
!> \param unit the unit to retain
!> \author fawzi
!> \note
!>      at the moment not needed, there for completeness
! **************************************************************************************************
   SUBROUTINE cp_unit_retain(unit)
      TYPE(cp_unit_type), POINTER                        :: unit

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(unit)))CALL cp__a("common/cp_units.F",558)
      IF(.NOT.(unit%ref_count > 0))CALL cp__a("common/cp_units.F",559)
      unit%ref_count = unit%ref_count+1
   END SUBROUTINE cp_unit_retain

! **************************************************************************************************
!> \brief releases the given unit
!> \param unit the unit to release
!> \author fawzi
!> \note
!>      at the moment not needed, there for completeness
! **************************************************************************************************
   SUBROUTINE cp_unit_release(unit)
      TYPE(cp_unit_type), POINTER                        :: unit

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(unit)) THEN
         IF(.NOT.(unit%ref_count > 0))CALL cp__a("common/cp_units.F",577)
         unit%ref_count = unit%ref_count-1
         IF (unit%ref_count == 0) THEN
            DEALLOCATE (unit)
         END IF
      END IF
   END SUBROUTINE cp_unit_release

! **************************************************************************************************
!> \brief controls that the kind and contains meaningful information
!> \param basic_kind the kind of the unit
!> \param basic_unit the unit to check
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_basic_unit_check(basic_kind, basic_unit)
      INTEGER, INTENT(in)                                :: basic_kind, basic_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_basic_unit_check', &
         routineP = moduleN//':'//routineN

      SELECT CASE (basic_kind)
      CASE (cp_ukind_undef)
         SELECT CASE (basic_unit)
         CASE (cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",602,"unknown undef unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_energy)
         SELECT CASE (basic_unit)
         CASE (cp_units_hartree, cp_units_wavenum, cp_units_joule, cp_units_kcalmol, &
               cp_units_kjmol, cp_units_Ry, cp_units_eV, cp_units_keV, cp_units_au, cp_units_k, &
               cp_units_jmol, cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",610,"unknown energy unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_length)
         SELECT CASE (basic_unit)
         CASE (cp_units_bohr, cp_units_angstrom, cp_units_au, cp_units_none, cp_units_m, &
               cp_units_pm, cp_units_nm)
         CASE default
            CALL cp__b("common/cp_units.F",617,"unknown length unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_temperature)
         SELECT CASE (basic_unit)
         CASE (cp_units_k, cp_units_au, cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",623,"unknown temperature unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_pressure)
         SELECT CASE (basic_unit)
         CASE (cp_units_bar, cp_units_atm, cp_units_kbar, cp_units_Pa, cp_units_MPa, cp_units_GPa, cp_units_au, cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",629,"unknown pressure unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_angle)
         SELECT CASE (basic_unit)
         CASE (cp_units_rad, cp_units_deg, cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",635,"unknown angle unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_time)
         SELECT CASE (basic_unit)
         CASE (cp_units_s, cp_units_fs, cp_units_ps, cp_units_au, cp_units_wn, cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",641,"unknown time unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_mass)
         SELECT CASE (basic_unit)
         CASE (cp_units_kg, cp_units_amu, cp_units_m_e, cp_units_au, cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",647,"unknown mass unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_potential)
         SELECT CASE (basic_unit)
         CASE (cp_units_volt, cp_units_au, cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",653,"unknown potential unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_force)
         SELECT CASE (basic_unit)
         CASE (cp_units_Newton, cp_units_mNewton, cp_units_au, cp_units_none)
         CASE default
            CALL cp__b("common/cp_units.F",659,"unknown force unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_none)
         IF (basic_unit /= cp_units_none) &
            CALL cp_abort(cp__l("common/cp_units.F",663), &
                          "if the kind of the unit is none also unit must be undefined,not:" &
                          //TRIM(cp_to_string(basic_unit)))
      CASE default
         CALL cp__b("common/cp_units.F",667,"unknown kind of unit:"//TRIM(cp_to_string(basic_kind)))
      END SELECT
   END SUBROUTINE cp_basic_unit_check

! **************************************************************************************************
!> \brief converts a value to the internal cp2k units
!> \param value the value to convert
!> \param basic_kind the kind of the unit of the value
!> \param basic_unit the unit of the value
!> \param power the power of the unit (defaults to 1)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_basic_unit_to_cp2k(value, basic_kind, basic_unit, power) RESULT(res)
      REAL(kind=dp), INTENT(in)                          :: value
      INTEGER, INTENT(in)                                :: basic_kind, basic_unit
      INTEGER, INTENT(in), OPTIONAL                      :: power
      REAL(kind=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_basic_unit_to_cp2k', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: my_power

      my_power = 1
      IF (PRESENT(power)) my_power = power
      IF (basic_unit == cp_units_none .AND. basic_kind /= cp_ukind_undef) THEN
         IF (basic_kind /= cp_units_none) &
            CALL cp_abort(cp__l("common/cp_units.F",695), &
                          "unit not yet fully specified, unit of kind "// &
                          TRIM(cp_to_string(basic_unit)))
      END IF
      SELECT CASE (basic_kind)
      CASE (cp_ukind_undef)
         SELECT CASE (basic_unit)
         CASE (cp_units_none)
            res = value
         CASE default
            CALL cp__b("common/cp_units.F",705,"unknown energy unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_energy)
         SELECT CASE (basic_unit)
         CASE (cp_units_hartree, cp_units_au)
            res = value
         CASE (cp_units_wavenum)
            res = wavenumbers**(-my_power)*value
         CASE (cp_units_joule)
            res = joule**(-my_power)*value
         CASE (cp_units_kcalmol)
            res = kcalmol**(-my_power)*value
         CASE (cp_units_kjmol)
            res = kjmol**(-my_power)*value
         CASE (cp_units_jmol)
            res = (kjmol*1.0E+3_dp)**(-my_power)*value
         CASE (cp_units_Ry)
            res = 0.5_dp**my_power*value
         CASE (cp_units_eV)
            res = evolt**(-my_power)*value
         CASE (cp_units_keV)
            res = (1.0E-3_dp*evolt)**(-my_power)*value
         CASE (cp_units_k)
            res = kelvin**(-my_power)*value
         CASE default
            CALL cp__b("common/cp_units.F",730,"unknown energy unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_length)
         SELECT CASE (basic_unit)
         CASE (cp_units_bohr, cp_units_au)
            res = value
         CASE (cp_units_m)
            res = value*(1.0E10_dp*bohr)**my_power
         CASE (cp_units_pm)
            res = value*(0.01_dp*bohr)**my_power
         CASE (cp_units_nm)
            res = value*(10.0_dp*bohr)**my_power
         CASE (cp_units_angstrom)
            res = value*bohr**my_power
         CASE default
            CALL cp__b("common/cp_units.F",745,"unknown length unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_temperature)
         SELECT CASE (basic_unit)
         CASE (cp_units_k)
            res = kelvin**(-my_power)*value
         CASE (cp_units_au)
            res = value
         CASE default
            CALL cp__b("common/cp_units.F",754,"unknown temperature unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_pressure)
         SELECT CASE (basic_unit)
         CASE (cp_units_bar)
            res = bar**(-my_power)*value
         CASE (cp_units_atm)
            res = atm**(-my_power)*value
         CASE (cp_units_kbar)
            res = (1.0E-3_dp*bar)**(-my_power)*value
         CASE (cp_units_Pa)
            res = pascal**(-my_power)*value
         CASE (cp_units_MPa)
            res = (1.0E-6_dp*pascal)**(-my_power)*value
         CASE (cp_units_GPa)
            res = (1.0E-9_dp*pascal)**(-my_power)*value
         CASE (cp_units_au)
            res = value
         CASE default
            CALL cp__b("common/cp_units.F",773,"unknown pressure unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_angle)
         SELECT CASE (basic_unit)
         CASE (cp_units_rad)
            res = value
         CASE (cp_units_deg)
            res = value*(radians)**my_power
         CASE default
            CALL cp__b("common/cp_units.F",782,"unknown angle unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_time)
         SELECT CASE (basic_unit)
         CASE (cp_units_s)
            res = value*seconds**(-my_power)
         CASE (cp_units_fs)
            res = value*femtoseconds**(-my_power)
         CASE (cp_units_ps)
            res = value*picoseconds**(-my_power)
         CASE (cp_units_au)
            res = value
         CASE (cp_units_wn)
            res = (twopi*wavenumbers)**(my_power)/value
         CASE default
            CALL cp__b("common/cp_units.F",797,"unknown time unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_mass)
         SELECT CASE (basic_unit)
         CASE (cp_units_kg)
            res = e_mass**my_power*value
         CASE (cp_units_amu)
            res = massunit**my_power*value
         CASE (cp_units_m_e, cp_units_au)
            res = value
         CASE default
            CALL cp__b("common/cp_units.F",808,"unknown mass unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_potential)
         SELECT CASE (basic_unit)
         CASE (cp_units_volt)
            res = evolt**(-my_power)*value
         CASE (cp_units_au)
            res = value
         CASE default
            CALL cp__b("common/cp_units.F",817,"unknown potential unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_force)
         SELECT CASE (basic_unit)
         CASE (cp_units_Newton)
            res = value*newton**(-my_power)
         CASE (cp_units_mNewton)
            res = value*(1.0E+3*newton)**(-my_power)
         CASE (cp_units_au)
            res = value
         CASE default
            CALL cp__b("common/cp_units.F",828,"unknown force unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_none)
         CALL cp_abort(cp__l("common/cp_units.F",831), &
                       "if the kind of the unit is none also unit must be undefined,not:" &
                       //TRIM(cp_to_string(basic_unit)))
      CASE default
         CALL cp__b("common/cp_units.F",835,"unknown kind of unit:"//TRIM(cp_to_string(basic_kind)))
      END SELECT
   END FUNCTION cp_basic_unit_to_cp2k

! **************************************************************************************************
!> \brief returns the label of the current basic unit
!> \param basic_kind the kind of the unit of the value
!> \param basic_unit the unit of the value
!> \param power the power of the unit (defaults to 1)
!> \param accept_undefined ...
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_basic_unit_desc(basic_kind, basic_unit, power, accept_undefined) &
      RESULT(res)
      INTEGER, INTENT(in)                                :: basic_kind, basic_unit
      INTEGER, INTENT(in), OPTIONAL                      :: power
      LOGICAL, INTENT(in), OPTIONAL                      :: accept_undefined
      CHARACTER(len=cp_unit_basic_desc_length)           :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_basic_unit_desc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: a, my_power
      LOGICAL                                            :: my_accept_undefined

      my_power = 1
      res = ""
      my_accept_undefined = .FALSE.
      IF (accept_undefined) my_accept_undefined = accept_undefined
      IF (PRESENT(power)) my_power = power
      IF (basic_unit == cp_units_none) THEN
         IF (.NOT. my_accept_undefined .AND. basic_kind == cp_units_none) &
            CALL cp_abort(cp__l("common/cp_units.F",868), "unit not yet fully specified, unit of kind "// &
                          TRIM(cp_to_string(basic_kind)))
      END IF
      SELECT CASE (basic_kind)
      CASE (cp_ukind_undef)
         SELECT CASE (basic_unit)
         CASE (cp_units_none)
            res = "internal_cp2k"
         CASE DEFAULT
            CALL cp_abort(cp__l("common/cp_units.F",877), &
                          "unit not yet fully specified, unit of kind "// &
                          TRIM(res))
         END SELECT
      CASE (cp_ukind_energy)
         SELECT CASE (basic_unit)
         CASE (cp_units_hartree, cp_units_au)
            res = "hartree"
         CASE (cp_units_wavenum)
            res = "wavenumber_e"
         CASE (cp_units_joule)
            res = "joule"
         CASE (cp_units_kcalmol)
            res = "kcalmol"
         CASE (cp_units_kjmol)
            res = "kjmol"
         CASE (cp_units_jmol)
            res = "jmol"
         CASE (cp_units_Ry)
            res = "Ry"
         CASE (cp_units_eV)
            res = "eV"
         CASE (cp_units_keV)
            res = "keV"
         CASE (cp_units_k)
            res = "K_e"
         CASE (cp_units_none)
            res = "energy"
            IF (.NOT. my_accept_undefined) &
               CALL cp_abort(cp__l("common/cp_units.F",906), &
                             "unit not yet fully specified, unit of kind "// &
                             TRIM(res))
         CASE default
            CALL cp__b("common/cp_units.F",910,"unknown energy unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_length)
         SELECT CASE (basic_unit)
         CASE (cp_units_bohr, cp_units_au)
            res = "bohr"
         CASE (cp_units_m)
            res = "m"
         CASE (cp_units_pm)
            res = "pm"
         CASE (cp_units_nm)
            res = "nm"
         CASE (cp_units_angstrom)
            res = "angstrom"
         CASE default
            res = "length"
            CALL cp__b("common/cp_units.F",926,"unknown length unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_temperature)
         SELECT CASE (basic_unit)
         CASE (cp_units_k)
            res = "K"
         CASE (cp_units_au)
            res = "au_temp"
         CASE (cp_units_none)
            res = "temperature"
            IF (.NOT. my_accept_undefined) &
               CALL cp_abort(cp__l("common/cp_units.F",937), &
                             "unit not yet fully specified, unit of kind "// &
                             TRIM(res))
         CASE default
            CALL cp__b("common/cp_units.F",941,"unknown temperature unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_pressure)
         SELECT CASE (basic_unit)
         CASE (cp_units_bar)
            res = "bar"
         CASE (cp_units_atm)
            res = "atm"
         CASE (cp_units_kbar)
            res = "kbar"
         CASE (cp_units_Pa)
            res = "Pa"
         CASE (cp_units_MPa)
            res = "MPa"
         CASE (cp_units_GPa)
            res = "GPa"
         CASE (cp_units_au)
            res = "au_p"
         CASE (cp_units_none)
            res = "pressure"
            IF (.NOT. my_accept_undefined) &
               CALL cp_abort(cp__l("common/cp_units.F",962), &
                             "unit not yet fully specified, unit of kind "// &
                             TRIM(res))
         CASE default
            CALL cp__b("common/cp_units.F",966,"unknown pressure unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_angle)
         SELECT CASE (basic_unit)
         CASE (cp_units_rad)
            res = "rad"
         CASE (cp_units_deg)
            res = "deg"
         CASE (cp_units_none)
            res = "angle"
            IF (.NOT. my_accept_undefined) &
               CALL cp_abort(cp__l("common/cp_units.F",977), &
                             "unit not yet fully specified, unit of kind "// &
                             TRIM(res))
         CASE default
            CALL cp__b("common/cp_units.F",981,"unknown angle unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_time)
         SELECT CASE (basic_unit)
         CASE (cp_units_s)
            res = "s"
         CASE (cp_units_fs)
            res = "fs"
         CASE (cp_units_ps)
            res = "ps"
         CASE (cp_units_au)
            res = "au_t"
         CASE (cp_units_wn)
            res = "wavenumber_t"
         CASE (cp_units_none)
            res = "time"
            IF (.NOT. my_accept_undefined) &
               CALL cp_abort(cp__l("common/cp_units.F",998), &
                             "unit not yet fully specified, unit of kind "// &
                             TRIM(res))
         CASE default
            CALL cp__b("common/cp_units.F",1002,"unknown time unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_mass)
         SELECT CASE (basic_unit)
         CASE (cp_units_kg)
            res = "kg"
         CASE (cp_units_amu)
            res = "amu"
         CASE (cp_units_m_e, cp_units_au)
            res = "m_e"
         CASE (cp_units_none)
            res = "mass"
            IF (.NOT. my_accept_undefined) &
               CALL cp_abort(cp__l("common/cp_units.F",1015), &
                             "unit not yet fully specified, unit of kind "// &
                             TRIM(res))
         CASE default
            CALL cp__b("common/cp_units.F",1019,"unknown mass unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_potential)
         SELECT CASE (basic_unit)
         CASE (cp_units_volt)
            res = "volt"
         CASE (cp_units_au)
            res = "au_pot"
         CASE (cp_units_none)
            res = "potential"
            IF (.NOT. my_accept_undefined) &
               CALL cp_abort(cp__l("common/cp_units.F",1030), &
                             "unit not yet fully specified, unit of kind "// &
                             TRIM(res))
         CASE default
            CALL cp__b("common/cp_units.F",1034,"unknown potential unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_force)
         SELECT CASE (basic_unit)
         CASE (cp_units_Newton)
            res = "N"
         CASE (cp_units_mNewton)
            res = "mN"
         CASE (cp_units_au)
            res = "au_f"
         CASE (cp_units_none)
            res = "force"
            IF (.NOT. my_accept_undefined) &
               CALL cp_abort(cp__l("common/cp_units.F",1047), &
                             "unit not yet fully specified, unit of kind "// &
                             TRIM(res))
         CASE default
            CALL cp__b("common/cp_units.F",1051,"unknown potential unit:"//TRIM(cp_to_string(basic_unit)))
         END SELECT
      CASE (cp_ukind_none)
         CALL cp_abort(cp__l("common/cp_units.F",1054), &
                       "if the kind of the unit is none also unit must be undefined,not:" &
                       //TRIM(cp_to_string(basic_unit)))
      CASE default
         CALL cp__b("common/cp_units.F",1058,"unknown kind of unit:"//TRIM(cp_to_string(basic_kind)))
      END SELECT
      IF (my_power /= 1) THEN
         a = LEN_TRIM(res)
         IF(.NOT.(LEN(res)-a >= 3))CALL cp__a("common/cp_units.F",1062)
         WRITE (res(a+1:), "('^',i3)") my_power
         CALL compress(res, .TRUE.)
      END IF
   END FUNCTION cp_basic_unit_desc

! **************************************************************************************************
!> \brief returns the "name" of the given unit
!> \param unit the unit to describe
!> \param defaults defaults for the undefined units, optional
!> \param accept_undefined if defaults is not present or is not associated
!>        whether undefined units should be accepted (defaults to false)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_unit_desc(unit, defaults, accept_undefined) &
      RESULT(res)
      TYPE(cp_unit_type), POINTER                        :: unit
      TYPE(cp_unit_set_type), OPTIONAL, POINTER          :: defaults
      LOGICAL, INTENT(in), OPTIONAL                      :: accept_undefined
      CHARACTER(len=cp_unit_desc_length)                 :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_desc', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, my_unit, pos
      LOGICAL                                            :: check, has_defaults, my_accept_undefined

      IF(.NOT.(ASSOCIATED(unit)))CALL cp__a("common/cp_units.F",1089)
      res = ""
      pos = 1
      my_accept_undefined = .FALSE.
      IF (PRESENT(accept_undefined)) my_accept_undefined = accept_undefined
      DO i = 1, unit%n_kinds
         IF(.NOT.(unit%kind_id(i) /= 0))CALL cp__a("common/cp_units.F",1095)
         IF(.NOT.(pos < LEN(res)))CALL cp__a("common/cp_units.F",1096)
         my_unit = unit%unit_id(i)
         has_defaults = .FALSE.
         IF (PRESENT(defaults)) has_defaults = ASSOCIATED(defaults)
         IF (my_unit == 0) THEN
            IF (has_defaults) THEN
               my_unit = defaults%units(unit%kind_id(i))%unit%unit_id(1)
            ELSE
               check = my_accept_undefined .OR. unit%kind_id(i) /= 0
               IF(.NOT.(check))CALL cp__a("common/cp_units.F",1105)
            END IF
         END IF
         IF (i > 1) THEN
            res(pos:pos) = "*"
            pos = pos+1
         END IF
         res(pos:) = TRIM(cp_basic_unit_desc(basic_kind=unit%kind_id(i), &
                                             basic_unit=my_unit, accept_undefined=my_accept_undefined, &
                                             power=unit%power(i)))
         pos = LEN_TRIM(res)+1
      END DO

   END FUNCTION cp_unit_desc

! **************************************************************************************************
!> \brief tranform a value to the internal cp2k units
!> \param value the value to convert
!> \param unit the unit of the result
!> \param defaults the defaults unit for those that are left free
!>        (cp_units_none)
!> \param power the power of the unit (defaults to 1)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_unit_to_cp2k1(value, unit, defaults, power) RESULT(res)
      REAL(kind=dp), INTENT(in)                          :: value
      TYPE(cp_unit_type), POINTER                        :: unit
      TYPE(cp_unit_set_type), OPTIONAL, POINTER          :: defaults
      INTEGER, INTENT(in), OPTIONAL                      :: power
      REAL(kind=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_to_cp2k1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_unit, my_basic_unit, my_power

      my_power = 1
      IF (PRESENT(power)) my_power = power
      res = value
      IF(.NOT.(ASSOCIATED(unit)))CALL cp__a("common/cp_units.F",1145)
      DO i_unit = 1, unit%n_kinds
         IF(.NOT.(unit%kind_id(i_unit) > 0))CALL cp__a("common/cp_units.F",1147)
         my_basic_unit = unit%unit_id(i_unit)
         IF (my_basic_unit == 0 .AND. unit%kind_id(i_unit) /= cp_ukind_undef) THEN
            IF(.NOT.(PRESENT(defaults)))CALL cp__a("common/cp_units.F",1150)
            IF(.NOT.(ASSOCIATED(defaults)))CALL cp__a("common/cp_units.F",1151)
            IF(.NOT.(ASSOCIATED(defaults%units(unit%kind_id(i_unit))%unit)))CALL cp__a("common/cp_units.F",1152)
            my_basic_unit = defaults%units(unit%kind_id(i_unit))%unit%unit_id(1)
         END IF
         res = cp_basic_unit_to_cp2k(value=res, basic_unit=my_basic_unit, &
                                     basic_kind=unit%kind_id(i_unit), &
                                     power=my_power*unit%power(i_unit))
      END DO
   END FUNCTION cp_unit_to_cp2k1

! **************************************************************************************************
!> \brief converts from the internal cp2k units to the given unit
!> \param value the value to convert
!> \param unit the unit of the result
!> \param defaults the defaults unit for those that are left free
!>        (cp_units_none)
!> \param power the power of the unit (defaults to 1)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_unit_from_cp2k1(value, unit, defaults, power) RESULT(res)
      REAL(kind=dp), INTENT(in)                          :: value
      TYPE(cp_unit_type), POINTER                        :: unit
      TYPE(cp_unit_set_type), OPTIONAL, POINTER          :: defaults
      INTEGER, INTENT(in), OPTIONAL                      :: power
      REAL(kind=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_from_cp2k1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: my_power

      my_power = 1
      IF (PRESENT(power)) my_power = power
      IF (PRESENT(defaults)) THEN
         res = cp_unit_to_cp2k1(value=value, unit=unit, defaults=defaults, &
                                power=-my_power)
      ELSE
         res = cp_unit_to_cp2k1(value=value, unit=unit, power=-my_power)
      END IF
   END FUNCTION cp_unit_from_cp2k1

! **************************************************************************************************
!> \brief converts to the internal cp2k units to the given unit
!> \param value the value to convert
!> \param unit_str the unit of the result as string
!> \param defaults the defaults unit for those that are left free
!>        (cp_units_none)
!> \param power the power of the unit (defaults to 1)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_unit_to_cp2k(value, unit_str, defaults, power) RESULT(res)
      REAL(kind=dp), INTENT(in)                          :: value
      CHARACTER(len=*), INTENT(in)                       :: unit_str
      TYPE(cp_unit_set_type), OPTIONAL, POINTER          :: defaults
      INTEGER, INTENT(in), OPTIONAL                      :: power
      REAL(kind=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_to_cp2k', &
         routineP = moduleN//':'//routineN

      TYPE(cp_unit_type), POINTER                        :: my_unit

      NULLIFY (my_unit)
      CALL cp_unit_create(my_unit, unit_str)
      IF (PRESENT(defaults)) THEN
         res = cp_unit_to_cp2k1(value=value, unit=my_unit, defaults=defaults, &
                                power=power)
      ELSE
         res = cp_unit_to_cp2k1(value=value, unit=my_unit, power=power)
      END IF
      CALL cp_unit_release(my_unit)
   END FUNCTION cp_unit_to_cp2k

! **************************************************************************************************
!> \brief converts from the internal cp2k units to the given unit
!> \param value the value to convert
!> \param unit_str the unit of the result as string
!> \param defaults the defaults unit for those that are left free
!>        (cp_units_none)
!> \param power the power of the unit (defaults to 1)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION cp_unit_from_cp2k(value, unit_str, defaults, power) RESULT(res)
      REAL(kind=dp), INTENT(in)                          :: value
      CHARACTER(len=*), INTENT(in)                       :: unit_str
      TYPE(cp_unit_set_type), OPTIONAL, POINTER          :: defaults
      INTEGER, INTENT(in), OPTIONAL                      :: power
      REAL(kind=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_from_cp2k', &
         routineP = moduleN//':'//routineN

      TYPE(cp_unit_type), POINTER                        :: my_unit

      NULLIFY (my_unit)
      CALL cp_unit_create(my_unit, unit_str)
      IF (PRESENT(defaults)) THEN
         res = cp_unit_from_cp2k1(value=value, unit=my_unit, defaults=defaults, &
                                  power=power)
      ELSE
         res = cp_unit_from_cp2k1(value=value, unit=my_unit, power=power)
      END IF
      CALL cp_unit_release(my_unit)
   END FUNCTION cp_unit_from_cp2k

! **************************************************************************************************
!> \brief returs true if the two units are compatible
!> \param ref_unit ...
!> \param unit ...
!> \retval res ...
!> \author Teodoro Laino [tlaino] - 11.2007 - University of Zurich
! **************************************************************************************************
   FUNCTION cp_unit_compatible(ref_unit, unit) RESULT(res)
      TYPE(cp_unit_type), POINTER                        :: ref_unit, unit
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_compatible', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      res = .TRUE.
      DO i = 1, SIZE(ref_unit%kind_id)
         IF (ref_unit%kind_id(i) == unit%kind_id(i)) CYCLE
         IF ((ref_unit%kind_id(1) == cp_ukind_undef) .AND. (ALL(ref_unit%kind_id(2:) == cp_ukind_none))) CYCLE
         res = .FALSE.
         EXIT
      END DO

   END FUNCTION cp_unit_compatible

! **************************************************************************************************
!> \brief initializes the given unit set
!> \param unit_set the set to initialize
!> \param name the name of the set, used for the dafault initialization of
!>        the various units
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_unit_set_create(unit_set, name)
      TYPE(cp_unit_set_type), POINTER                    :: unit_set
      CHARACTER(len=*), INTENT(in)                       :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_set_create', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: my_name
      INTEGER                                            :: i

      IF(.NOT.(.NOT. ASSOCIATED(unit_set)))CALL cp__a("common/cp_units.F",1302)
      ALLOCATE (unit_set)
      unit_set%ref_count = 1
      last_unit_set_id = last_unit_set_id+1
      unit_set%id_nr = last_unit_set_id
      my_name = name
      CALL uppercase(my_name)

      DO i = 1, cp_ukind_max
         NULLIFY (unit_set%units(i)%unit)
      END DO
      DO i = 1, cp_ukind_max
         SELECT CASE (name)
         CASE ('ATOM', 'ATOMIC', 'INTERNAL', 'CP2K')
            IF (i == cp_ukind_angle) THEN
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), &
                                    unit_id=(/cp_units_rad/), power=(/1/))
            ELSE
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), &
                                    unit_id=(/cp_units_au/), power=(/1/))
            END IF
         CASE ('OUTPUT')
            SELECT CASE (i)
            CASE (cp_ukind_undef)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_none/), &
                                    power=(/1/))
            CASE (cp_ukind_energy)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_hartree/), &
                                    power=(/1/))
            CASE (cp_ukind_length)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_angstrom/), &
                                    power=(/1/))
            CASE (cp_ukind_temperature)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_k/), &
                                    power=(/1/))
            CASE (cp_ukind_angle)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_deg/), &
                                    power=(/1/))
            CASE (cp_ukind_pressure)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_bar/), &
                                    power=(/1/))
            CASE (cp_ukind_time)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_fs/), &
                                    power=(/1/))
            CASE (cp_ukind_mass)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_amu/), &
                                    power=(/1/))
            CASE (cp_ukind_potential)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_volt/), &
                                    power=(/1/))
            CASE (cp_ukind_force)
               CALL cp_unit_create2(unit_set%units(i)%unit, kind_id=(/i/), unit_id=(/cp_units_newton/), &
                                    power=(/1/))
            CASE default
               CALL cp__b("common/cp_units.F",1356,"unhandled unit type "//TRIM(cp_to_string(i)))
               EXIT
            END SELECT
         CASE default
            CALL cp__b("common/cp_units.F",1360,'unknown parameter set name '//TRIM(name))
         END SELECT
      END DO
   END SUBROUTINE cp_unit_set_create

! **************************************************************************************************
!> \brief retains the given unit set
!> \param unit_set the unit set to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_unit_set_retain(unit_set)
      TYPE(cp_unit_set_type), POINTER                    :: unit_set

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_set_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(unit_set)))CALL cp__a("common/cp_units.F",1376)
      IF(.NOT.(unit_set%ref_count > 0))CALL cp__a("common/cp_units.F",1377)
      unit_set%ref_count = unit_set%ref_count+1
   END SUBROUTINE cp_unit_set_retain

! **************************************************************************************************
!> \brief releases the given unit set
!> \param unit_set the unit set to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE cp_unit_set_release(unit_set)
      TYPE(cp_unit_set_type), POINTER                    :: unit_set

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_unit_set_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(unit_set)) THEN
         IF(.NOT.(unit_set%ref_count > 0))CALL cp__a("common/cp_units.F",1395)
         unit_set%ref_count = unit_set%ref_count-1
         IF (unit_set%ref_count == 0) THEN
            DO i = 1, SIZE(unit_set%units)
               CALL cp_unit_release(unit_set%units(i)%unit)
            END DO
            DEALLOCATE (unit_set)
         END IF
      END IF
   END SUBROUTINE cp_unit_set_release

! **************************************************************************************************
!> \brief Prints info on all available units in CP2K
!> \param unit_nr ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE print_all_units(unit_nr)
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'print_all_units', &
         routineP = moduleN//':'//routineN

      CALL print_section_unit(unit_label="Undefined", description="If the default unit "// &
                              "of a keyword is explicitly undefined, all possible units of measurement can "// &
                              "be used to define a proper value.", &
                              units_set=s2a("internal_cp2k"), unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Energy", description="Possible units of measurement "// &
                              "for Energies. The [energy] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of energy is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("hartree", "wavenumber_e", "joule", "kcalmol", "kjmol", "Ry", &
                                            "eV", "K_e", "energy"), unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Length", description="Possible units of measurement "// &
                              "for Lengths. The [length] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of length is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("bohr", "m", "pm", "nm", "angstrom", "length"), unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Temperature", description="Possible units of measurement "// &
                              "for Temperature. The [temperature] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of temperature is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("K", "au_temp", "temperature"), unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Pressure", description="Possible units of measurement "// &
                              "for Pressure. The [pressure] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of pressure is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("bar", "atm", "kbar", "Pa", "MPa", "GPa", "au_p", "pressure"), &
                              unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Angle", description="Possible units of measurement "// &
                              "for Angles. The [angle] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of angle is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("rad", "deg", "angle"), unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Time", description="Possible units of measurement "// &
                              "for Time. The [time] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of time is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("s", "fs", "ps", "au_t", "wavenumber_t", "time"), unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Mass", description="Possible units of measurement "// &
                              "for Masses. The [mass] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of mass is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("kg", "amu", "m_e", "mass"), unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Potential", description="Possible units of measurement "// &
                              "for potentials. The [potential] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of potential is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("volt", "au_pot", "potential"), unit_nr=unit_nr)

      CALL print_section_unit(unit_label="Force", description="Possible units of measurement "// &
                              "for forces. The [force] entry acts like a dummy flag (assumes the unit of "// &
                              "measurement of force is in internal units), useful for dimensional analysis.", &
                              units_set=s2a("N", "Newton", "mN", "mNewton", "au_f", "force"), unit_nr=unit_nr)

   END SUBROUTINE print_all_units

! **************************************************************************************************
!> \brief Prints info on all available units in CP2K - Low level
!> \param unit_label ...
!> \param description ...
!> \param units_set ...
!> \param unit_nr ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE print_section_unit(unit_label, description, units_set, unit_nr)
      CHARACTER(LEN=*), INTENT(IN)                       :: unit_label, description
      CHARACTER(LEN=*), DIMENSION(:), INTENT(IN)         :: units_set
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'print_section_unit', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      WRITE (unit_nr, FMT='(A)') "<H2>"//TRIM(unit_label)//"</H2>"
      WRITE (unit_nr, FMT='(A)') description//"<BR><DL>"
      DO i = 1, SIZE(units_set)
         WRITE (unit_nr, FMT='(A)') "<DD><B>"//TRIM(units_set(i))//"</B></DD>"
      END DO
      WRITE (unit_nr, FMT='(A)') "</DL><P>"
   END SUBROUTINE print_section_unit

END MODULE cp_units
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief a wrapper for basic fortran types.
!> \par History
!>      06.2004 created
!> \author fawzi
! **************************************************************************************************
MODULE input_val_types
   USE cp_parser_types,                 ONLY: default_continuation_character
   USE cp_units,                        ONLY: cp_unit_create,&
                                              cp_unit_desc,&
                                              cp_unit_from_cp2k,&
                                              cp_unit_from_cp2k1,&
                                              cp_unit_release,&
                                              cp_unit_retain,&
                                              cp_unit_type
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enum_release,&
                                              enum_retain,&
                                              enumeration_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_val_types'

   INTEGER, SAVE, PRIVATE :: last_val_id = 0

   PUBLIC :: val_p_type, val_type
   PUBLIC :: val_create, val_retain, val_release, val_get, val_write, &
             val_write_internal, val_duplicate
!***

   INTEGER, PARAMETER, PUBLIC :: no_t = 0, logical_t = 1, &
                                 integer_t = 2, real_t = 3, char_t = 4, enum_t = 5, lchar_t = 6

! **************************************************************************************************
!> \brief pointer to a val, to create arrays of pointers
!> \param val to pointer to the val
!> \author fawzi
! **************************************************************************************************
   TYPE val_p_type
      TYPE(val_type), POINTER :: val
   END TYPE val_p_type

! **************************************************************************************************
!> \brief a type to  have a wrapper that stores any basic fortran type
!> \param type_of_var type stored in the val (should be one of no_t,
!>        integer_t, logical_t, real_t, char_t)
!> \param l_val , i_val, c_val, r_val: arrays with logical,integer,character
!>        or real values. Only one should be associated (and namely the one
!>        specified in type_of_var).
!> \param enum an enumaration to map char to integers
!> \author fawzi
! **************************************************************************************************
   TYPE val_type
      INTEGER :: ref_count, id_nr, type_of_var
      LOGICAL, DIMENSION(:), POINTER :: l_val
      INTEGER, DIMENSION(:), POINTER :: i_val
      CHARACTER(len=default_string_length), DIMENSION(:), POINTER :: &
         c_val
      REAL(kind=dp), DIMENSION(:), POINTER :: r_val
      TYPE(enumeration_type), POINTER :: enum
   END TYPE val_type
CONTAINS

! **************************************************************************************************
!> \brief creates a keyword value
!> \param val the object to be created
!> \param l_val ,i_val,r_val,c_val,lc_val: a logical,integer,real,string, long
!>        string to be stored in the val
!> \param l_vals , i_vals, r_vals, c_vals: an array of logicals,
!>        integers, reals, characters, long strings to be stored in val
!> \param l_vals_ptr , i_vals_ptr, r_vals_ptr, c_vals_ptr: an array of logicals,
!>        ... to be stored in val, val will get the ownership of the pointer
!> \param i_val ...
!> \param i_vals ...
!> \param i_vals_ptr ...
!> \param r_val ...
!> \param r_vals ...
!> \param r_vals_ptr ...
!> \param c_val ...
!> \param c_vals ...
!> \param c_vals_ptr ...
!> \param lc_val ...
!> \param lc_vals ...
!> \param lc_vals_ptr ...
!> \param enum the enumaration type this value is using
!> \author fawzi
!> \note
!>      using an enumeration only i_val/i_vals/i_vals_ptr are accepted
! **************************************************************************************************
   SUBROUTINE val_create(val, l_val, l_vals, l_vals_ptr, i_val, i_vals, i_vals_ptr, &
                         r_val, r_vals, r_vals_ptr, c_val, c_vals, c_vals_ptr, lc_val, lc_vals, &
                         lc_vals_ptr, enum)
      TYPE(val_type), POINTER                            :: val
      LOGICAL, INTENT(in), OPTIONAL                      :: l_val
      LOGICAL, DIMENSION(:), INTENT(in), OPTIONAL        :: l_vals
      LOGICAL, DIMENSION(:), OPTIONAL, POINTER           :: l_vals_ptr
      INTEGER, INTENT(in), OPTIONAL                      :: i_val
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: i_vals
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: i_vals_ptr
      REAL(KIND=DP), INTENT(in), OPTIONAL                :: r_val
      REAL(KIND=DP), DIMENSION(:), INTENT(in), OPTIONAL  :: r_vals
      REAL(KIND=DP), DIMENSION(:), OPTIONAL, POINTER     :: r_vals_ptr
      CHARACTER(LEN=*), INTENT(in), OPTIONAL             :: c_val
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in), &
         OPTIONAL                                        :: c_vals
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: c_vals_ptr
      CHARACTER(LEN=*), INTENT(in), OPTIONAL             :: lc_val
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in), &
         OPTIONAL                                        :: lc_vals
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: lc_vals_ptr
      TYPE(enumeration_type), OPTIONAL, POINTER          :: enum

      CHARACTER(len=*), PARAMETER :: routineN = 'val_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, len_c, narg, nVal

      IF(.NOT.(.NOT. ASSOCIATED(val)))CALL cp__a("input/input_val_types.F",130)
      ALLOCATE (val)
      NULLIFY (val%l_val, val%i_val, val%r_val, val%c_val, val%enum)
      val%type_of_var = no_t
      last_val_id = last_val_id+1
      val%id_nr = last_val_id
      val%ref_count = 1

      narg = 0
      val%type_of_var = no_t
      IF (PRESENT(l_val)) THEN
!FM        CPPrecondition(.NOT.PRESENT(l_vals),cp_failure_level,routineP,failure)
!FM        CPPrecondition(.NOT.PRESENT(l_vals_ptr),cp_failure_level,routineP,failure)
         narg = narg+1
         ALLOCATE (val%l_val(1))
         val%l_val(1) = l_val
         val%type_of_var = logical_t
      END IF
      IF (PRESENT(l_vals)) THEN
!FM        CPPrecondition(.NOT.PRESENT(l_vals_ptr),cp_failure_level,routineP,failure)
         narg = narg+1
         ALLOCATE (val%l_val(SIZE(l_vals)))
         val%l_val = l_vals
         val%type_of_var = logical_t
      END IF
      IF (PRESENT(l_vals_ptr)) THEN
         narg = narg+1
         val%l_val => l_vals_ptr
         val%type_of_var = logical_t
      END IF

      IF (PRESENT(r_val)) THEN
!FM        CPPrecondition(.NOT.PRESENT(r_vals),cp_failure_level,routineP,failure)
!FM        CPPrecondition(.NOT.PRESENT(r_vals_ptr),cp_failure_level,routineP,failure)
         narg = narg+1
         ALLOCATE (val%r_val(1))
         val%r_val(1) = r_val
         val%type_of_var = real_t
      END IF
      IF (PRESENT(r_vals)) THEN
!FM        CPPrecondition(.NOT.PRESENT(r_vals_ptr),cp_failure_level,routineP,failure)
         narg = narg+1
         ALLOCATE (val%r_val(SIZE(r_vals)))
         val%r_val = r_vals
         val%type_of_var = real_t
      END IF
      IF (PRESENT(r_vals_ptr)) THEN
         narg = narg+1
         val%r_val => r_vals_ptr
         val%type_of_var = real_t
      END IF

      IF (PRESENT(i_val)) THEN
!FM        CPPrecondition(.NOT.PRESENT(i_vals),cp_failure_level,routineP,failure)
!FM        CPPrecondition(.NOT.PRESENT(i_vals_ptr),cp_failure_level,routineP,failure)
         narg = narg+1
         ALLOCATE (val%i_val(1))
         val%i_val(1) = i_val
         val%type_of_var = integer_t
      END IF
      IF (PRESENT(i_vals)) THEN
!FM        CPPrecondition(.NOT.PRESENT(i_vals_ptr),cp_failure_level,routineP,failure)
         narg = narg+1
         ALLOCATE (val%i_val(SIZE(i_vals)))
         val%i_val = i_vals
         val%type_of_var = integer_t
      END IF
      IF (PRESENT(i_vals_ptr)) THEN
         narg = narg+1
         val%i_val => i_vals_ptr
         val%type_of_var = integer_t
      END IF

      IF (PRESENT(c_val)) THEN
         IF(.NOT.(LEN_TRIM(c_val) <= default_string_length))CALL cp__a("input/input_val_types.F",204)
!FM        CPPrecondition(.NOT.PRESENT(c_vals),cp_failure_level,routineP,failure)
!FM        CPPrecondition(.NOT.PRESENT(c_vals_ptr),cp_failure_level,routineP,failure)
         narg = narg+1
         ALLOCATE (val%c_val(1))
         val%c_val(1) = c_val
         val%type_of_var = char_t
      END IF
      IF (PRESENT(c_vals)) THEN
!FM        CPPrecondition(.NOT.PRESENT(c_vals_ptr),cp_failure_level,routineP,failure)
         IF(.NOT.(ALL(LEN_TRIM(c_vals) <= default_string_length)))CALL cp__a("input/input_val_types.F",214)
         narg = narg+1
         ALLOCATE (val%c_val(SIZE(c_vals)))
         val%c_val = c_vals
         val%type_of_var = char_t
      END IF
      IF (PRESENT(c_vals_ptr)) THEN
         narg = narg+1
         val%c_val => c_vals_ptr
         val%type_of_var = char_t
      END IF
      IF (PRESENT(lc_val)) THEN
!FM        CPPrecondition(.NOT.PRESENT(lc_vals),cp_failure_level,routineP,failure)
!FM        CPPrecondition(.NOT.PRESENT(lc_vals_ptr),cp_failure_level,routineP,failure)
         narg = narg+1
         len_c = LEN_TRIM(lc_val)
         nVal = MAX(1, CEILING(REAL(len_c, dp)/80._dp))
         ALLOCATE (val%c_val(nVal))

         IF (len_c == 0) THEN
            val%c_val(1) = ""
         ELSE
            DO i = 1, nVal
               val%c_val(i) = lc_val((i-1)*default_string_length+1: &
                                     MIN(len_c, i*default_string_length))
            END DO
         END IF
         val%type_of_var = lchar_t
      END IF
      IF (PRESENT(lc_vals)) THEN
         IF(.NOT.(ALL(LEN_TRIM(lc_vals) <= default_string_length)))CALL cp__a("input/input_val_types.F",244)
         narg = narg+1
         ALLOCATE (val%c_val(SIZE(lc_vals)))
         val%c_val = lc_vals
         val%type_of_var = lchar_t
      END IF
      IF (PRESENT(lc_vals_ptr)) THEN
         narg = narg+1
         val%c_val => lc_vals_ptr
         val%type_of_var = lchar_t
      END IF
      IF(.NOT.(narg <= 1))CALL cp__a("input/input_val_types.F",255)
      IF (PRESENT(enum)) THEN
         IF (ASSOCIATED(enum)) THEN
            IF (val%type_of_var /= no_t .AND. val%type_of_var /= integer_t .AND. &
                val%type_of_var /= enum_t) THEN
               CALL cp__b("input/input_val_types.F",260,"")
            END IF
            IF (ASSOCIATED(val%i_val)) THEN
               val%type_of_var = enum_t
               val%enum => enum
               CALL enum_retain(enum)
            END IF
         END IF
      END IF
      IF(.NOT.(ASSOCIATED(val%enum) .EQV. val%type_of_var == enum_t))CALL cp__a("input/input_val_types.F",269)
   END SUBROUTINE val_create

! **************************************************************************************************
!> \brief releases the given val
!> \param val the val to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE val_release(val)
      TYPE(val_type), POINTER                            :: val

      CHARACTER(len=*), PARAMETER :: routineN = 'val_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(val)) THEN
         IF(.NOT.(val%ref_count > 0))CALL cp__a("input/input_val_types.F",283)
         val%ref_count = val%ref_count-1
         IF (val%ref_count == 0) THEN
            IF (ASSOCIATED(val%l_val)) THEN
               DEALLOCATE (val%l_val)
            END IF
            IF (ASSOCIATED(val%i_val)) THEN
               DEALLOCATE (val%i_val)
            END IF
            IF (ASSOCIATED(val%r_val)) THEN
               DEALLOCATE (val%r_val)
            END IF
            IF (ASSOCIATED(val%c_val)) THEN
               DEALLOCATE (val%c_val)
            END IF
            CALL enum_release(val%enum)
            val%type_of_var = no_t
            DEALLOCATE (val)
         END IF
      END IF
      NULLIFY (val)
   END SUBROUTINE val_release

! **************************************************************************************************
!> \brief retains the given val
!> \param val the val to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE val_retain(val)
      TYPE(val_type), POINTER                            :: val

      CHARACTER(len=*), PARAMETER :: routineN = 'val_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(val)))CALL cp__a("input/input_val_types.F",316)
      IF(.NOT.(val%ref_count > 0))CALL cp__a("input/input_val_types.F",317)
      val%ref_count = val%ref_count+1
   END SUBROUTINE val_retain

! **************************************************************************************************
!> \brief returns the stored values
!> \param val the object from which you want to extract the values
!> \param has_l ...
!> \param has_i ...
!> \param has_r ...
!> \param has_lc ...
!> \param has_c ...
!> \param l_val gets a logical from the val
!> \param l_vals gets an array of logicals from the val
!> \param i_val gets an integer from the val
!> \param i_vals gets an array of integers from the val
!> \param r_val gets a real from the val
!> \param r_vals gets an array of reals from the val
!> \param c_val gets a char from the val
!> \param c_vals gets an array of chars from the val
!> \param len_c len_trim of c_val (if it was a lc_val, of type lchar_t
!>        it might be longet than default_string_length)
!> \param type_of_var ...
!> \param enum ...
!> \author fawzi
!> \note
!>      using an enumeration only i_val/i_vals/i_vals_ptr are accepted
!>      add something like ignore_string_cut that if true does not warn if
!>      the c_val is too short to contain the string
! **************************************************************************************************
   SUBROUTINE val_get(val, has_l, has_i, has_r, has_lc, has_c, l_val, l_vals, i_val, &
                      i_vals, r_val, r_vals, c_val, c_vals, len_c, type_of_var, enum)
      TYPE(val_type), POINTER                            :: val
      LOGICAL, INTENT(out), OPTIONAL                     :: has_l, has_i, has_r, has_lc, has_c, l_val
      LOGICAL, DIMENSION(:), OPTIONAL, POINTER           :: l_vals
      INTEGER, INTENT(out), OPTIONAL                     :: i_val
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: i_vals
      REAL(KIND=DP), INTENT(out), OPTIONAL               :: r_val
      REAL(KIND=DP), DIMENSION(:), OPTIONAL, POINTER     :: r_vals
      CHARACTER(LEN=*), INTENT(out), OPTIONAL            :: c_val
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: c_vals
      INTEGER, INTENT(out), OPTIONAL                     :: len_c, type_of_var
      TYPE(enumeration_type), OPTIONAL, POINTER          :: enum

      CHARACTER(len=*), PARAMETER :: routineN = 'val_get', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, l_in, l_out

      IF (PRESENT(has_l)) has_l = ASSOCIATED(val%l_val)
      IF (PRESENT(has_i)) has_i = ASSOCIATED(val%i_val)
      IF (PRESENT(has_r)) has_r = ASSOCIATED(val%r_val)
      IF (PRESENT(has_c)) has_c = ASSOCIATED(val%c_val) ! use type_of_var?
      IF (PRESENT(has_lc)) has_lc = (val%type_of_var == lchar_t)
      IF (PRESENT(l_vals)) l_vals => val%l_val
      IF (PRESENT(l_val)) THEN
         IF (ASSOCIATED(val%l_val)) THEN
            IF (SIZE(val%l_val) > 0) THEN
               l_val = val%l_val(1)
            ELSE
               CALL cp__b("input/input_val_types.F",377,"")
            END IF
         ELSE
            CALL cp__b("input/input_val_types.F",380,"")
         END IF
      END IF

      IF (PRESENT(i_vals)) i_vals => val%i_val
      IF (PRESENT(i_val)) THEN
         IF (ASSOCIATED(val%i_val)) THEN
            IF (SIZE(val%i_val) > 0) THEN
               i_val = val%i_val(1)
            ELSE
               CALL cp__b("input/input_val_types.F",390,"")
            END IF
         ELSE
            CALL cp__b("input/input_val_types.F",393,"")
         END IF
      END IF

      IF (PRESENT(r_vals)) r_vals => val%r_val
      IF (PRESENT(r_val)) THEN
         IF (ASSOCIATED(val%r_val)) THEN
            IF (SIZE(val%r_val) > 0) THEN
               r_val = val%r_val(1)
            ELSE
               CALL cp__b("input/input_val_types.F",403,"")
            END IF
         ELSE
            CALL cp__b("input/input_val_types.F",406,"")
         END IF
      END IF

      IF (PRESENT(c_vals)) c_vals => val%c_val
      IF (PRESENT(c_val)) THEN
         l_out = LEN(c_val)
         IF (ASSOCIATED(val%c_val)) THEN
            IF (SIZE(val%c_val) > 0) THEN
               IF (val%type_of_var == lchar_t) THEN
                  l_in = default_string_length*(SIZE(val%c_val)-1)+ &
                         LEN_TRIM(val%c_val(SIZE(val%c_val)))
                  IF (l_out < l_in) &
                     CALL cp_warn(cp__l("input/input_val_types.F",419), &
                                  "val_get will truncate value, value beginning with '"// &
                                  TRIM(val%c_val(1))//"' is too long for variable")
                  DO i = 1, SIZE(val%c_val)
                     c_val((i-1)*default_string_length+1:MIN(l_out, i*default_string_length)) = &
                        val%c_val(i) (1:MIN(80, l_out-(i-1)*default_string_length))
                     IF (l_out <= i*default_string_length) EXIT
                  END DO
                  IF (l_out > SIZE(val%c_val)*default_string_length) &
                     c_val(SIZE(val%c_val)*default_string_length+1:l_out) = ""
               ELSE
                  l_in = LEN_TRIM(val%c_val(1))
                  IF (l_out < l_in) &
                     CALL cp_warn(cp__l("input/input_val_types.F",432), &
                                  "val_get will truncate value, value '"// &
                                  TRIM(val%c_val(1))//"' is too long for variable")
                  c_val = val%c_val(1)
               END IF
            ELSE
               CALL cp__b("input/input_val_types.F",438,"")
            END IF
         ELSE IF (ASSOCIATED(val%i_val) .AND. ASSOCIATED(val%enum)) THEN
            IF (SIZE(val%i_val) > 0) THEN
               c_val = enum_i2c(val%enum, val%i_val(1))
            ELSE
               CALL cp__b("input/input_val_types.F",444,"")
            END IF
         ELSE
            CALL cp__b("input/input_val_types.F",447,"")
         END IF
      END IF

      IF (PRESENT(len_c)) THEN
         IF (ASSOCIATED(val%c_val)) THEN
            IF (SIZE(val%c_val) > 0) THEN
               IF (val%type_of_var == lchar_t) THEN
                  len_c = default_string_length*(SIZE(val%c_val)-1)+ &
                          LEN_TRIM(val%c_val(SIZE(val%c_val)))
               ELSE
                  len_c = LEN_TRIM(val%c_val(1))
               END IF
            ELSE
               len_c = -HUGE(0)
            END IF
         ELSE IF (ASSOCIATED(val%i_val) .AND. ASSOCIATED(val%enum)) THEN
            IF (SIZE(val%i_val) > 0) THEN
               len_c = LEN_TRIM(enum_i2c(val%enum, val%i_val(1)))
            ELSE
               len_c = -HUGE(0)
            END IF
         ELSE
            len_c = -HUGE(0)
         END IF
      END IF

      IF (PRESENT(type_of_var)) type_of_var = val%type_of_var

      IF (PRESENT(enum)) enum => val%enum

   END SUBROUTINE val_get

! **************************************************************************************************
!> \brief writes out the valuse stored in the val
!> \param val the val to write
!> \param unit_nr the number of the unit to write to
!> \param unit the unit of mesure in wich the output should be written
!>        (overrides unit_str)
!> \param unit_str the unit of mesure in wich the output should be written
!> \param fmt ...
!> \author fawzi
!> \note
!>      unit of mesure used only for reals
! **************************************************************************************************
   SUBROUTINE val_write(val, unit_nr, unit, unit_str, fmt)
      TYPE(val_type), POINTER                            :: val
      INTEGER, INTENT(in)                                :: unit_nr
      TYPE(cp_unit_type), OPTIONAL, POINTER              :: unit
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: unit_str, fmt

      CHARACTER(len=*), PARAMETER :: routineN = 'val_write', routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: c_string, myfmt, rcval
      INTEGER                                            :: i, iend, item, j, l
      TYPE(cp_unit_type), POINTER                        :: my_unit

      NULLIFY (my_unit)
      myfmt = ""
      IF (PRESENT(fmt)) myfmt = fmt
      IF (PRESENT(unit)) my_unit => unit
      IF (ASSOCIATED(my_unit)) THEN
         CALL cp_unit_retain(my_unit)
      ELSE IF (PRESENT(unit_str)) THEN
         CALL cp_unit_create(my_unit, unit_str)
      END IF
      IF (ASSOCIATED(val)) THEN
         SELECT CASE (val%type_of_var)
         CASE (logical_t)
            IF (ASSOCIATED(val%l_val)) THEN
               DO i = 1, SIZE(val%l_val)
                  IF (MODULO(i, 20) == 0) THEN
                     WRITE (unit=unit_nr, fmt="(' ',A)") default_continuation_character
                     WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//")", advance="NO")
                  END IF
                  WRITE (unit=unit_nr, fmt="(' ',l1)", advance="NO") &
                     val%l_val(i)
               END DO
            ELSE
               CALL cp__b("input/input_val_types.F",526,"")
            END IF
         CASE (integer_t)
            IF (ASSOCIATED(val%i_val)) THEN
               item = 0
               i = 1
               loop_i: DO WHILE (i <= SIZE(val%i_val))
                  item = item+1
                  IF (MODULO(item, 10) == 0) THEN
                     WRITE (UNIT=unit_nr, FMT="(1X,A)") default_continuation_character
                     WRITE (UNIT=unit_nr, FMT="("//TRIM(myfmt)//")", ADVANCE="NO")
                  END IF
                  iend = i
                  loop_j: DO j = i+1, SIZE(val%i_val)
                     IF (val%i_val(j-1)+1 == val%i_val(j)) THEN
                        iend = iend+1
                     ELSE
                        EXIT loop_j
                     END IF
                  END DO loop_j
                  IF ((iend-i) > 1) THEN
                     WRITE (UNIT=unit_nr, FMT="(1X,I0,A2,I0)", ADVANCE="NO") &
                        val%i_val(i), "..", val%i_val(iend)
                     i = iend
                  ELSE
                     WRITE (UNIT=unit_nr, FMT="(1X,I0)", ADVANCE="NO") &
                        val%i_val(i)
                  END IF
                  i = i+1
               END DO loop_i
            ELSE
               CALL cp__b("input/input_val_types.F",557,"")
            END IF
         CASE (real_t)
            IF (ASSOCIATED(val%r_val)) THEN
               DO i = 1, SIZE(val%r_val)
                  IF (MODULO(i, 5) == 0) THEN
                     WRITE (unit=unit_nr, fmt="(' ',A)") default_continuation_character
                     WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//")", advance="NO")
                  END IF
                  IF (ASSOCIATED(my_unit)) THEN
                     WRITE (rcval, "(ES25.16)") cp_unit_from_cp2k1(val%r_val(i), my_unit)
                  ELSE
                     WRITE (rcval, "(ES25.16)") val%r_val(i)
                  END IF
                  WRITE (unit=unit_nr, fmt="(' ',A)", advance="NO") TRIM(rcval)
               END DO
            ELSE
               CALL cp__b("input/input_val_types.F",574,"")
            END IF
         CASE (char_t)
            IF (ASSOCIATED(val%c_val)) THEN
               l = 0
               DO i = 1, SIZE(val%c_val)
                  IF (i > 1) WRITE (unit=unit_nr, fmt="(' ')", advance="NO")
                  l = l+1
                  IF (l > 10 .AND. l+LEN_TRIM(val%c_val(i)) > 76) THEN
                     WRITE (unit=unit_nr, fmt="('\')")
                     WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//")", advance="NO")
                     l = 0
                     WRITE (unit=unit_nr, fmt='(a)', advance="NO") TRIM(val%c_val(i))
                     l = l+LEN_TRIM(val%c_val(i))+3
                  ELSE IF (LEN_TRIM(val%c_val(i)) > 0) THEN
                     l = l+LEN_TRIM(val%c_val(i))
                     WRITE (unit=unit_nr, fmt='(a)', advance="NO") TRIM(val%c_val(i))
                  ELSE
                     l = l+3
                     WRITE (unit=unit_nr, fmt="(a)", advance="NO") '" "'
                  END IF
               END DO
            ELSE
               CALL cp__b("input/input_val_types.F",597,"")
            END IF
         CASE (lchar_t)
            IF (ASSOCIATED(val%c_val)) THEN
               l = 0
               DO i = 1, SIZE(val%c_val)-1
                  WRITE (unit=unit_nr, fmt='(a)', advance="NO") val%c_val(i)
               END DO
               IF (SIZE(val%c_val) > 0) THEN
                  WRITE (unit=unit_nr, fmt='(a)', advance="NO") TRIM(val%c_val(SIZE(val%c_val)))
               END IF
            ELSE
               CALL cp__b("input/input_val_types.F",609,"")
            END IF
         CASE (enum_t)
            IF (ASSOCIATED(val%i_val)) THEN
               l = 0
               DO i = 1, SIZE(val%i_val)
                  c_string = enum_i2c(val%enum, val%i_val(i))
                  IF (l > 10 .AND. l+LEN_TRIM(c_string) > 76) THEN
                     WRITE (unit=unit_nr, fmt="(' ',A)") default_continuation_character
                     WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//")", advance="NO")
                     l = 0
                  ELSE
                     l = l+LEN_TRIM(c_string)+3
                  END IF
                  WRITE (unit=unit_nr, fmt="(' ',a)", advance="NO") TRIM(c_string)
               END DO
            ELSE
               CALL cp__b("input/input_val_types.F",626,"")
            END IF

         CASE (no_t)
            WRITE (unit=unit_nr, fmt="(' *empty*')", advance="NO")
         CASE default
            CALL cp__b("input/input_val_types.F",632,"unexpected type_of_var for val ")
         END SELECT
      ELSE
         WRITE (unit=unit_nr, fmt="(' *null*')", advance="NO")
      END IF
      IF (ASSOCIATED(my_unit)) CALL cp_unit_release(my_unit)
      WRITE (unit=unit_nr, fmt="()")
   END SUBROUTINE val_write

! **************************************************************************************************
!> \brief   Write values to an internal file, i.e. string variable.
!> \param val ...
!> \param string ...
!> \param unit ...
!> \date    10.03.2005
!> \par History
!>          17.01.2006, MK, Optional argument unit for the conversion to the external unit added
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE val_write_internal(val, string, unit)

      TYPE(val_type), POINTER                            :: val
      CHARACTER(LEN=*), INTENT(OUT)                      :: string
      TYPE(cp_unit_type), OPTIONAL, POINTER              :: unit

      CHARACTER(len=*), PARAMETER :: routineN = 'val_write_internal', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: enum_string
      INTEGER                                            :: i, ipos
      REAL(KIND=dp)                                      :: value

! -------------------------------------------------------------------------

      string = ""

      IF (ASSOCIATED(val)) THEN

         SELECT CASE (val%type_of_var)
         CASE (logical_t)
            IF (ASSOCIATED(val%l_val)) THEN
               DO i = 1, SIZE(val%l_val)
                  WRITE (UNIT=string(2*i-1:), FMT="(L2)") val%l_val(i)
               END DO
            ELSE
               CALL cp__b("input/input_val_types.F",678,"")
            END IF
         CASE (integer_t)
            IF (ASSOCIATED(val%i_val)) THEN
               DO i = 1, SIZE(val%i_val)
                  WRITE (UNIT=string(12*i-11:), FMT="(I12)") val%i_val(i)
               END DO
            ELSE
               CALL cp__b("input/input_val_types.F",686,"")
            END IF
         CASE (real_t)
            IF (ASSOCIATED(val%r_val)) THEN
               IF (PRESENT(unit)) THEN
                  DO i = 1, SIZE(val%r_val)
                     value = cp_unit_from_cp2k(value=val%r_val(i), &
                                               unit_str=cp_unit_desc(unit=unit))
                     WRITE (UNIT=string(17*i-16:), FMT="(ES17.8E3)") value
                  END DO
               ELSE
                  DO i = 1, SIZE(val%r_val)
                     WRITE (UNIT=string(17*i-16:), FMT="(ES17.8E3)") val%r_val(i)
                  END DO
               END IF
            ELSE
               CALL cp__b("input/input_val_types.F",702,"")
            END IF
         CASE (char_t)
            IF (ASSOCIATED(val%c_val)) THEN
               ipos = 1
               DO i = 1, SIZE(val%c_val)
                  WRITE (UNIT=string(ipos:), FMT="(A)") TRIM(ADJUSTL(val%c_val(i)))
                  ipos = ipos+LEN_TRIM(ADJUSTL(val%c_val(i)))+1
               END DO
            ELSE
               CALL cp__b("input/input_val_types.F",712,"")
            END IF
         CASE (lchar_t)
            IF (ASSOCIATED(val%c_val)) THEN
               CALL val_get(val, c_val=string)
            ELSE
               CALL cp__b("input/input_val_types.F",718,"")
            END IF
         CASE (enum_t)
            IF (ASSOCIATED(val%i_val)) THEN
               DO i = 1, SIZE(val%i_val)
                  enum_string = enum_i2c(val%enum, val%i_val(i))
                  WRITE (UNIT=string, FMT="(A)") TRIM(ADJUSTL(enum_string))
               END DO
            ELSE
               CALL cp__b("input/input_val_types.F",727,"")
            END IF
         CASE default
            CALL cp__b("input/input_val_types.F",730,"unexpected type_of_var for val ")
         END SELECT

      END IF

   END SUBROUTINE val_write_internal

! **************************************************************************************************
!> \brief creates a copy of the given value
!> \param val_in the value to copy
!> \param val_out the value tha will be created
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE val_duplicate(val_in, val_out)
      TYPE(val_type), POINTER                            :: val_in, val_out

      CHARACTER(len=*), PARAMETER :: routineN = 'val_duplicate', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(val_in)))CALL cp__a("input/input_val_types.F",748)
      IF(.NOT.(.NOT. ASSOCIATED(val_out)))CALL cp__a("input/input_val_types.F",749)
      ALLOCATE (val_out)
      last_val_id = last_val_id+1
      val_out%id_nr = last_val_id
      val_out%type_of_var = val_in%type_of_var
      val_out%ref_count = 1
      val_out%enum => val_in%enum
      IF (ASSOCIATED(val_out%enum)) CALL enum_retain(val_out%enum)

      NULLIFY (val_out%l_val, val_out%i_val, val_out%c_val, val_out%r_val)
      IF (ASSOCIATED(val_in%l_val)) THEN
         ALLOCATE (val_out%l_val(SIZE(val_in%l_val)))
         val_out%l_val = val_in%l_val
      END IF
      IF (ASSOCIATED(val_in%i_val)) THEN
         ALLOCATE (val_out%i_val(SIZE(val_in%i_val)))
         val_out%i_val = val_in%i_val
      END IF
      IF (ASSOCIATED(val_in%r_val)) THEN
         ALLOCATE (val_out%r_val(SIZE(val_in%r_val)))
         val_out%r_val = val_in%r_val
      END IF
      IF (ASSOCIATED(val_in%c_val)) THEN
         ALLOCATE (val_out%c_val(SIZE(val_in%c_val)))
         val_out%c_val = val_in%c_val
      END IF
   END SUBROUTINE val_duplicate

END MODULE input_val_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Routine for sorting an array
!> \note
!>      CP2K:
!>      Please use the interface definded in util.F for calling sort().
!>      This could not go into cp_array__nametype1__utils.template because
!>      there some routines use the error-handling from cp_common_uses.f90,
!>      which in turn uses timings.F, which uses sort() :-(
!>
!>      DBCSR:
!>      Please use the interface definded in dbcsr_toollib.F for calling sort().
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify _lib__array__nametype1__sort.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE cp_array_sort_i8

   USE kinds,                           ONLY: int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_array_sort_i8'

   PUBLIC :: cp_1d_i8_sort

CONTAINS

! **************************************************************************************************
!> \brief Sorts an array inplace using a combination of merge- and bubble-sort.
!>        It also returns the indices, which the elements had before the sort.
!> \param arr the array to sort
!> \param n length of array
!> \param indices returns elements-indices before the sort
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_1d_i8_sort(arr, n, indices)
      INTEGER, INTENT(in)                                :: n
      INTEGER(kind=int_8), DIMENSION(1:n), INTENT(inout) :: arr
      INTEGER, DIMENSION(1:n), INTENT(out)               :: indices

      INTEGER                                            :: i
      INTEGER(kind=int_8), POINTER                       :: tmp_arr(:)
      INTEGER, POINTER                                   :: tmp_idx(:)

      IF (n == 0) RETURN ! for some reason this is a frequent case in cp2k

      ! scratch space used during the merge step
      ALLOCATE (tmp_arr((SIZE(arr)+1)/2), tmp_idx((SIZE(arr)+1)/2))

      indices = (/(i, i=1, SIZE(arr))/)

      CALL cp_1d_i8_sort_low(arr(1:n), indices, tmp_arr, tmp_idx)

      DEALLOCATE (tmp_arr, tmp_idx)

   END SUBROUTINE cp_1d_i8_sort

! **************************************************************************************************
!> \brief The actual sort routing.
!>            Only cp_1d_i8_sort and itself should call this.
!> \param arr the array to sort
!> \param indices elements-indices before the sort
!> \param tmp_arr scratch space
!> \param tmp_idx scratch space
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE cp_1d_i8_sort_low(arr, indices, tmp_arr, tmp_idx)
      INTEGER(kind=int_8), DIMENSION(:), INTENT(inout)   :: arr
      INTEGER, DIMENSION(SIZE(arr)), INTENT(inout)       :: indices
      INTEGER(kind=int_8), DIMENSION((SIZE(arr)+1)/2), &
         INTENT(inout)                                   :: tmp_arr
      INTEGER, DIMENSION((SIZE(arr)+1)/2), INTENT(inout) :: tmp_idx

      INTEGER                                            :: i, j, k, m, t
      INTEGER(kind=int_8)                                :: a
      LOGICAL                                            :: swapped

      ! a,t:  used during swaping of elements in arr and indices

      swapped = .TRUE.

      ! If only a few elements are left we switch to bubble-sort for efficiency.
      IF (SIZE(arr) <= 7) THEN ! 7 seems to be a good choice for the moment
         DO j = SIZE(arr)-1, 1, -1
            swapped = .FALSE.
            DO i = 1, j
               IF (( arr(i+1) <  arr(i) )) THEN
                  ! swap arr(i) with arr(i+1)
                  a = arr(i)
                  arr(i) = arr(i+1)
                  arr(i+1) = a
                  ! swap indices(i) with indices(i+1)
                  t = indices(i)
                  indices(i) = indices(i+1)
                  indices(i+1) = t
                  swapped = .TRUE.
               END IF
            END DO
            IF (.NOT. swapped) EXIT
         END DO
         RETURN
      ENDIF

      ! split list in half and recusively sort both sublists
      m = (SIZE(arr)+1)/2 ! index where we going to divide the list in two
      CALL cp_1d_i8_sort_low(arr(1:m), indices(1:m), tmp_arr, tmp_idx)
      CALL cp_1d_i8_sort_low(arr(m+1:), indices(m+1:), tmp_arr, tmp_idx)

      ! Check for a special case: Can we just concate the two sorted sublists?
      ! This leads to O(n) scaling if the input is already sorted.
      IF (( arr(m+1) <  arr(m) )) THEN
         ! ...no - let's merge the two sorted sublists arr(:m) and arr(m+1:)
         ! Merge will be performed directly in arr. Need backup of first sublist.
         tmp_arr(1:m) = arr(1:m)
         tmp_idx(1:m) = indices(1:m)
         i = 1; ! number of elemens consumed from 1st sublist
         j = 1; ! number of elemens consumed from 2nd sublist
         k = 1; ! number of elemens already merged

         DO WHILE (i <= m .AND. j <= SIZE(arr)-m)
         IF (( arr(m+j) <  tmp_arr(i) )) THEN
            arr(k) = arr(m+j)
            indices(k) = indices(m+j)
            j = j+1
         ELSE
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
         ENDIF
         k = k+1
         ENDDO

         ! One of the two sublist is now empty.
         ! Copy possibly remaining tail of 1st sublist
         DO WHILE (i <= m)
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
            k = k+1
         ENDDO

         ! The possibly remaining tail of 2nd sublist is already at the right spot.

      ENDIF

   END SUBROUTINE cp_1d_i8_sort_low

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,lessQ,defines)]
! ARGS:
!  USE = "USE kinds,       ONLY:int_8"
!  defines =
!    "#define CP_I_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_I_LESS_Q"
!  nametype1 = "i8"
!  type1 = "INTEGER(kind=int_8)"

END MODULE cp_array_sort_i8
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Routine for sorting an array
!> \note
!>      CP2K:
!>      Please use the interface definded in util.F for calling sort().
!>      This could not go into cp_array__nametype1__utils.template because
!>      there some routines use the error-handling from cp_common_uses.f90,
!>      which in turn uses timings.F, which uses sort() :-(
!>
!>      DBCSR:
!>      Please use the interface definded in dbcsr_toollib.F for calling sort().
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify _lib__array__nametype1__sort.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE cp_array_sort_r

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_array_sort_r'

   PUBLIC :: cp_1d_r_sort

CONTAINS

! **************************************************************************************************
!> \brief Sorts an array inplace using a combination of merge- and bubble-sort.
!>        It also returns the indices, which the elements had before the sort.
!> \param arr the array to sort
!> \param n length of array
!> \param indices returns elements-indices before the sort
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_1d_r_sort(arr, n, indices)
      INTEGER, INTENT(in)                                :: n
      REAL(kind=dp), DIMENSION(1:n), INTENT(inout)       :: arr
      INTEGER, DIMENSION(1:n), INTENT(out)               :: indices

      INTEGER                                            :: i
      INTEGER, POINTER                                   :: tmp_idx(:)
      REAL(kind=dp), POINTER                             :: tmp_arr(:)

      IF (n == 0) RETURN ! for some reason this is a frequent case in cp2k

      ! scratch space used during the merge step
      ALLOCATE (tmp_arr((SIZE(arr)+1)/2), tmp_idx((SIZE(arr)+1)/2))

      indices = (/(i, i=1, SIZE(arr))/)

      CALL cp_1d_r_sort_low(arr(1:n), indices, tmp_arr, tmp_idx)

      DEALLOCATE (tmp_arr, tmp_idx)

   END SUBROUTINE cp_1d_r_sort

! **************************************************************************************************
!> \brief The actual sort routing.
!>            Only cp_1d_r_sort and itself should call this.
!> \param arr the array to sort
!> \param indices elements-indices before the sort
!> \param tmp_arr scratch space
!> \param tmp_idx scratch space
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE cp_1d_r_sort_low(arr, indices, tmp_arr, tmp_idx)
      REAL(kind=dp), DIMENSION(:), INTENT(inout)         :: arr
      INTEGER, DIMENSION(SIZE(arr)), INTENT(inout)       :: indices
      REAL(kind=dp), DIMENSION((SIZE(arr)+1)/2), &
         INTENT(inout)                                   :: tmp_arr
      INTEGER, DIMENSION((SIZE(arr)+1)/2), INTENT(inout) :: tmp_idx

      INTEGER                                            :: i, j, k, m, t
      REAL(kind=dp)                                      :: a
      LOGICAL                                            :: swapped

      ! a,t:  used during swaping of elements in arr and indices

      swapped = .TRUE.

      ! If only a few elements are left we switch to bubble-sort for efficiency.
      IF (SIZE(arr) <= 7) THEN ! 7 seems to be a good choice for the moment
         DO j = SIZE(arr)-1, 1, -1
            swapped = .FALSE.
            DO i = 1, j
               IF (( arr(i+1) <  arr(i) )) THEN
                  ! swap arr(i) with arr(i+1)
                  a = arr(i)
                  arr(i) = arr(i+1)
                  arr(i+1) = a
                  ! swap indices(i) with indices(i+1)
                  t = indices(i)
                  indices(i) = indices(i+1)
                  indices(i+1) = t
                  swapped = .TRUE.
               END IF
            END DO
            IF (.NOT. swapped) EXIT
         END DO
         RETURN
      ENDIF

      ! split list in half and recusively sort both sublists
      m = (SIZE(arr)+1)/2 ! index where we going to divide the list in two
      CALL cp_1d_r_sort_low(arr(1:m), indices(1:m), tmp_arr, tmp_idx)
      CALL cp_1d_r_sort_low(arr(m+1:), indices(m+1:), tmp_arr, tmp_idx)

      ! Check for a special case: Can we just concate the two sorted sublists?
      ! This leads to O(n) scaling if the input is already sorted.
      IF (( arr(m+1) <  arr(m) )) THEN
         ! ...no - let's merge the two sorted sublists arr(:m) and arr(m+1:)
         ! Merge will be performed directly in arr. Need backup of first sublist.
         tmp_arr(1:m) = arr(1:m)
         tmp_idx(1:m) = indices(1:m)
         i = 1; ! number of elemens consumed from 1st sublist
         j = 1; ! number of elemens consumed from 2nd sublist
         k = 1; ! number of elemens already merged

         DO WHILE (i <= m .AND. j <= SIZE(arr)-m)
         IF (( arr(m+j) <  tmp_arr(i) )) THEN
            arr(k) = arr(m+j)
            indices(k) = indices(m+j)
            j = j+1
         ELSE
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
         ENDIF
         k = k+1
         ENDDO

         ! One of the two sublist is now empty.
         ! Copy possibly remaining tail of 1st sublist
         DO WHILE (i <= m)
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
            k = k+1
         ENDDO

         ! The possibly remaining tail of 2nd sublist is already at the right spot.

      ENDIF

   END SUBROUTINE cp_1d_r_sort_low

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,lessQ,defines)]
! ARGS:
!  USE = "USE kinds, ONLY: dp"
!  defaultFormatType1 = ""(es14.6)""
!  defines =
!    "#define CP_R_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_R_LESS_Q"
!  nametype1 = "r"
!  type1 = "REAL(kind=dp)"

END MODULE cp_array_sort_r
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Routine for sorting an array
!> \note
!>      CP2K:
!>      Please use the interface definded in util.F for calling sort().
!>      This could not go into cp_array__nametype1__utils.template because
!>      there some routines use the error-handling from cp_common_uses.f90,
!>      which in turn uses timings.F, which uses sort() :-(
!>
!>      DBCSR:
!>      Please use the interface definded in dbcsr_toollib.F for calling sort().
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify _lib__array__nametype1__sort.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE cp_array_sort_i4

   USE kinds,                           ONLY: int_4
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_array_sort_i4'

   PUBLIC :: cp_1d_i4_sort

CONTAINS

! **************************************************************************************************
!> \brief Sorts an array inplace using a combination of merge- and bubble-sort.
!>        It also returns the indices, which the elements had before the sort.
!> \param arr the array to sort
!> \param n length of array
!> \param indices returns elements-indices before the sort
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_1d_i4_sort(arr, n, indices)
      INTEGER, INTENT(in)                                :: n
      INTEGER(kind=int_4), DIMENSION(1:n), INTENT(inout) :: arr
      INTEGER, DIMENSION(1:n), INTENT(out)               :: indices

      INTEGER                                            :: i
      INTEGER(kind=int_4), POINTER                       :: tmp_arr(:)
      INTEGER, POINTER                                   :: tmp_idx(:)

      IF (n == 0) RETURN ! for some reason this is a frequent case in cp2k

      ! scratch space used during the merge step
      ALLOCATE (tmp_arr((SIZE(arr)+1)/2), tmp_idx((SIZE(arr)+1)/2))

      indices = (/(i, i=1, SIZE(arr))/)

      CALL cp_1d_i4_sort_low(arr(1:n), indices, tmp_arr, tmp_idx)

      DEALLOCATE (tmp_arr, tmp_idx)

   END SUBROUTINE cp_1d_i4_sort

! **************************************************************************************************
!> \brief The actual sort routing.
!>            Only cp_1d_i4_sort and itself should call this.
!> \param arr the array to sort
!> \param indices elements-indices before the sort
!> \param tmp_arr scratch space
!> \param tmp_idx scratch space
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE cp_1d_i4_sort_low(arr, indices, tmp_arr, tmp_idx)
      INTEGER(kind=int_4), DIMENSION(:), INTENT(inout)   :: arr
      INTEGER, DIMENSION(SIZE(arr)), INTENT(inout)       :: indices
      INTEGER(kind=int_4), DIMENSION((SIZE(arr)+1)/2), &
         INTENT(inout)                                   :: tmp_arr
      INTEGER, DIMENSION((SIZE(arr)+1)/2), INTENT(inout) :: tmp_idx

      INTEGER                                            :: i, j, k, m, t
      INTEGER(kind=int_4)                                :: a
      LOGICAL                                            :: swapped

      ! a,t:  used during swaping of elements in arr and indices

      swapped = .TRUE.

      ! If only a few elements are left we switch to bubble-sort for efficiency.
      IF (SIZE(arr) <= 7) THEN ! 7 seems to be a good choice for the moment
         DO j = SIZE(arr)-1, 1, -1
            swapped = .FALSE.
            DO i = 1, j
               IF (( arr(i+1) <  arr(i) )) THEN
                  ! swap arr(i) with arr(i+1)
                  a = arr(i)
                  arr(i) = arr(i+1)
                  arr(i+1) = a
                  ! swap indices(i) with indices(i+1)
                  t = indices(i)
                  indices(i) = indices(i+1)
                  indices(i+1) = t
                  swapped = .TRUE.
               END IF
            END DO
            IF (.NOT. swapped) EXIT
         END DO
         RETURN
      ENDIF

      ! split list in half and recusively sort both sublists
      m = (SIZE(arr)+1)/2 ! index where we going to divide the list in two
      CALL cp_1d_i4_sort_low(arr(1:m), indices(1:m), tmp_arr, tmp_idx)
      CALL cp_1d_i4_sort_low(arr(m+1:), indices(m+1:), tmp_arr, tmp_idx)

      ! Check for a special case: Can we just concate the two sorted sublists?
      ! This leads to O(n) scaling if the input is already sorted.
      IF (( arr(m+1) <  arr(m) )) THEN
         ! ...no - let's merge the two sorted sublists arr(:m) and arr(m+1:)
         ! Merge will be performed directly in arr. Need backup of first sublist.
         tmp_arr(1:m) = arr(1:m)
         tmp_idx(1:m) = indices(1:m)
         i = 1; ! number of elemens consumed from 1st sublist
         j = 1; ! number of elemens consumed from 2nd sublist
         k = 1; ! number of elemens already merged

         DO WHILE (i <= m .AND. j <= SIZE(arr)-m)
         IF (( arr(m+j) <  tmp_arr(i) )) THEN
            arr(k) = arr(m+j)
            indices(k) = indices(m+j)
            j = j+1
         ELSE
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
         ENDIF
         k = k+1
         ENDDO

         ! One of the two sublist is now empty.
         ! Copy possibly remaining tail of 1st sublist
         DO WHILE (i <= m)
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
            k = k+1
         ENDDO

         ! The possibly remaining tail of 2nd sublist is already at the right spot.

      ENDIF

   END SUBROUTINE cp_1d_i4_sort_low

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,lessQ,defines)]
! ARGS:
!  USE = "USE kinds,       ONLY:int_4"
!  defines =
!    "#define CP_I_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_I_LESS_Q"
!  nametype1 = "i4"
!  type1 = "INTEGER(kind=int_4)"

END MODULE cp_array_sort_i4
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief All kind of helpfull little routines
!> \par History
!>      none
!> \author CJM & JGH
! **************************************************************************************************
MODULE util
   USE cp_array_sort_i4,                ONLY: cp_1d_i4_sort
   USE cp_array_sort_i8,                ONLY: cp_1d_i8_sort
   USE cp_array_sort_r,                 ONLY: cp_1d_r_sort
   USE kinds,                           ONLY: dp

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'util'
   PUBLIC :: sort, &
             get_limit, &
             locate, &
             find_boundary, &
             sort_unique

   INTERFACE sort
      MODULE PROCEDURE cp_1d_r_sort, cp_1d_i4_sort, cp_1d_i8_sort, &
         sort_cv, sort_im, sort_cm
   END INTERFACE

   INTERFACE sort_unique
      MODULE PROCEDURE sort_unique1
   END INTERFACE

   INTERFACE find_boundary
      MODULE PROCEDURE find_boundary1, find_boundary2
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Purpose: Given an array array(1:n), and given a value x, a value x_index
!>             is returned which is the index value of the array element equal
!>             to the value x: x = array(x_index)
!>             The array must be monotonic increasing.
!>             x_index = 0 is returned, if no array element equal to the value
!>             of x was found.
!> \param array ...
!> \param x ...
!> \retval x_index ...
!> \par History
!>      Derived from the locate function described in
!>      Numerical Recipes in Fortran 90 (09.01.2004,MK)
! **************************************************************************************************
   FUNCTION locate(array, x) RESULT(x_index)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: array
      INTEGER, INTENT(IN)                                :: x
      INTEGER                                            :: x_index

      CHARACTER(LEN=*), PARAMETER :: routineN = 'locate', routineP = moduleN//':'//routineN

      INTEGER                                            :: jl, jm, ju, n

      x_index = 0

      IF (x < array(1)) RETURN
      n = SIZE(array)
      IF (x > array(n)) RETURN
      jl = 0
      ju = n+1
      DO WHILE (ju-jl > 1)
         jm = (ju+jl)/2
         IF (x >= array(jm)) THEN
            jl = jm
         ELSE
            ju = jm
         END IF
      END DO
      IF (x == array(jl)) x_index = jl
   END FUNCTION locate

! **************************************************************************************************
!> \brief Sorts and returns a logical that checks if all elements are unique
!> \param arr ...
!> \param unique ...
!> \par History
!>      Teodoro Laino - Zurich University [tlaino] 04.2007
! **************************************************************************************************
   SUBROUTINE sort_unique1(arr, unique)
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: arr
      LOGICAL, INTENT(OUT)                               :: unique

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sort_unique1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: wrk

      n = SIZE(arr)
      unique = .TRUE.
      ALLOCATE (wrk(n))
      CALL sort(arr, n, wrk)
      DO i = 2, n
         IF (arr(i) == arr(i-1)) THEN
            unique = .FALSE.
            EXIT
         END IF
      END DO
      DEALLOCATE (wrk)
   END SUBROUTINE sort_unique1

! **************************************************************************************************
!> \brief Sorts an array of strings
!> \param arr ...
!> \param n ...
!> \param index ...
!> \author Teodoro Laino [tlaino] - University of Zurich  10.2008
! **************************************************************************************************
   SUBROUTINE sort_cv(arr, n, index)
      INTEGER, INTENT(IN)                                :: n
      CHARACTER(LEN=*), INTENT(INOUT)                    :: arr(1:n)
      INTEGER, INTENT(OUT)                               :: INDEX(1:n)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sort_cv', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, max_length
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: entries

      max_length = 0
      DO i = 1, n
         max_length = MAX(max_length, LEN_TRIM(arr(i)))
      END DO
      ALLOCATE (entries(max_length, SIZE(arr)))
      DO i = 1, n
         DO j = 1, LEN_TRIM(arr(i))
            entries(j, i) = ICHAR(arr(i) (j:j))
         END DO
         IF (j <= max_length) THEN
            entries(j:max_length, i) = ICHAR(" ")
         END IF
      END DO
      CALL sort_im(entries, istart=1, iend=n, j=1, jsize=max_length, INDEX=INDEX)
      ! Recover string once ordered
      DO i = 1, n
         DO j = 1, max_length
            arr(i) (j:j) = CHAR(entries(j, i))
         END DO
      END DO
      DEALLOCATE (entries)
   END SUBROUTINE sort_cv

! **************************************************************************************************
!> \brief Sorts a multiple arrays of integers M(j,i), ordering iteratively over
!>        i with fixed j
!> \param matrix ...
!> \param istart ...
!> \param iend ...
!> \param j ...
!> \param jsize ...
!> \param INDEX ...
!> \author Teodoro Laino [tlaino] - University of Zurich  10.2008
! **************************************************************************************************
   RECURSIVE SUBROUTINE sort_im(matrix, istart, iend, j, jsize, INDEX)
      INTEGER, DIMENSION(:, :), INTENT(INOUT)            :: matrix
      INTEGER, INTENT(IN)                                :: istart, iend, j, jsize
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: INDEX

      INTEGER                                            :: i, ind, isize, item, k, kend, kstart
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: bck_index, tmp_index, work, work2

    isize   = iend-istart+1
    ! Initialize the INDEX array only for the first row..
    IF (j==1) THEN
       DO i = 1,isize
          INDEX(i) = i
       ENDDO
    END IF

    ! Allocate scratch arrays
    ALLOCATE(work(isize), work2(isize), tmp_index(isize), bck_index(isize))
    ind = 0
    DO i = istart, iend
      ind = ind + 1
      work(ind)      = matrix(j,i)
      bck_index(ind) = INDEX(i)
    END DO

    ! Ordering row (j) interval istart..iend
    CALL sort(work, isize, tmp_index)

    ! Copy into global INDEX array with a proper mapping
    ind = 0
    DO i = istart, iend
       ind = ind + 1
       INDEX(i)=bck_index(tmp_index(ind))
       matrix(j,i)=work(ind)
    END DO

    ! Reorder the rest of the array according the present reordering
    DO k = j+1, jsize
       ind = 0
       DO i = istart, iend
          ind = ind + 1
          work2(ind)      = matrix(k,i)
       END DO
       ind = 0
       DO i = istart, iend
          ind = ind + 1
          matrix(k,i)=work2(tmp_index(ind))
       END DO
    END DO

    ! There are more rows to order..
    IF (j<jsize) THEN
       kstart = istart
       item   = work(1)
       ind    = 0
       DO i = istart, iend
          ind = ind + 1
          IF (item/=work(ind)) THEN
             kend = i-1
             IF (kstart/=kend) THEN
                CALL sort(matrix, kstart, kend, j+1, jsize, INDEX)
             END IF
             item   = work(ind)
             kstart = i
          END IF
       END DO
       kend = i-1
       IF (kstart/=kend) THEN
          CALL sort(matrix, kstart, kend, j+1, jsize, INDEX)
       END IF
    END IF
    DEALLOCATE(work, work2, tmp_index, bck_index)
   END SUBROUTINE sort_im

! **************************************************************************************************
!> \brief Sorts a multiple arrays of strings C(j,i), ordering iteratively over
!>        i with fixed j
!> \param matrix ...
!> \param istart ...
!> \param iend ...
!> \param j ...
!> \param jsize ...
!> \param INDEX ...
!> \author Teodoro Laino [tlaino] - 11.2008
! **************************************************************************************************
   RECURSIVE SUBROUTINE sort_cm(matrix, istart, iend, j, jsize, INDEX)
      CHARACTER(LEN=*), DIMENSION(:, :), INTENT(INOUT)   :: matrix
      INTEGER, INTENT(IN)                                :: istart, iend, j, jsize
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: INDEX

      CHARACTER(LEN=LEN(matrix))                         :: item
      CHARACTER(LEN=LEN(matrix)), ALLOCATABLE, &
         DIMENSION(:)                                    :: work, work2
      INTEGER                                            :: i, ind, isize, k, kend, kstart
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: bck_index, tmp_index

    isize   = iend-istart+1
    ! Initialize the INDEX array only for the first row..
    IF (j==1) THEN
       DO i = 1,isize
          INDEX(i) = i
       ENDDO
    END IF

    ! Allocate scratch arrays
    ALLOCATE(work(isize), work2(isize), tmp_index(isize), bck_index(isize))
    ind = 0
    DO i = istart, iend
      ind = ind + 1
      work(ind)      = matrix(j,i)
      bck_index(ind) = INDEX(i)
    END DO

    ! Ordering row (j) interval istart..iend
    CALL sort(work, isize, tmp_index)

    ! Copy into global INDEX array with a proper mapping
    ind = 0
    DO i = istart, iend
       ind = ind + 1
       INDEX(i)=bck_index(tmp_index(ind))
       matrix(j,i)=work(ind)
    END DO

    ! Reorder the rest of the array according the present reordering
    DO k = j+1, jsize
       ind = 0
       DO i = istart, iend
          ind = ind + 1
          work2(ind)      = matrix(k,i)
       END DO
       ind = 0
       DO i = istart, iend
          ind = ind + 1
          matrix(k,i)=work2(tmp_index(ind))
       END DO
    END DO

    ! There are more rows to order..
    IF (j<jsize) THEN
       kstart = istart
       item   = work(1)
       ind    = 0
       DO i = istart, iend
          ind = ind + 1
          IF (item/=work(ind)) THEN
             kend = i-1
             IF (kstart/=kend) THEN
                CALL sort(matrix, kstart, kend, j+1, jsize, INDEX)
             END IF
             item   = work(ind)
             kstart = i
          END IF
       END DO
       kend = i-1
       IF (kstart/=kend) THEN
          CALL sort(matrix, kstart, kend, j+1, jsize, INDEX)
       END IF
    END IF
    DEALLOCATE(work, work2, tmp_index, bck_index)
   END SUBROUTINE sort_cm

! **************************************************************************************************
!> \brief divide m entries into n parts, return size of part me
!> \param m ...
!> \param n ...
!> \param me ...
!> \retval nlim ...
! **************************************************************************************************
   FUNCTION get_limit(m, n, me) RESULT(nlim)
      INTEGER, INTENT(IN)                                :: m, n, me
      INTEGER                                            :: nlim(2)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_limit', routineP = moduleN//':'//routineN

      INTEGER                                            :: nl, nu
      REAL(KIND=dp)                                      :: part

      part = float(m)/float(n)
      nl = NINT(float(me)*part)+1
      nu = NINT(float(me+1)*part)
      nlim(1) = MAX(1, nl)
      nlim(2) = MIN(m, nu)

   END FUNCTION get_limit

! **************************************************************************************************
!> \brief finds boundary where element search starts and ends in a 1D array
!>      array1:      XXXXXAAAAAAAAAXXDGFSFGWDDDDDDDAAAWE
!>                        |       |
!>                     start     end  (searching for A)
!> \param num_array ...
!> \param ntot ...
!> \param first ...
!> \param last ...
!> \param search ...
! **************************************************************************************************
   SUBROUTINE find_boundary1(num_array, ntot, first, last, search)
      INTEGER, POINTER                                   :: num_array(:)
      INTEGER                                            :: ntot, first, last, search

      CHARACTER(LEN=*), PARAMETER :: routineN = 'find_boundary1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: found

      found = .FALSE.
      first = 0
      last = ntot

      DO i = 1, ntot
         IF (num_array(i) == search) THEN
            IF (.NOT. found) THEN
               first = i
            END IF
            found = .TRUE.
         ELSE
            IF (found) THEN
               last = i-1
               EXIT
            END IF
            found = .FALSE.
         END IF
      END DO

   END SUBROUTINE find_boundary1

! **************************************************************************************************
!> \brief finds boundary where element search1 starts and ends in array1 checking
!>      at the same time search2 in array2
!>      array1:      XXXXXAAAAAAAAAXXDGFSFGWDDDDDDDAAAWE
!>      array2:      XXXXASDEYYYYASDEFAAAARGASGASRGAWRRR
!>                           |  |
!>                       start  end  (searching for A and Y)
!> \param num_array1 ...
!> \param num_array2 ...
!> \param ntot ...
!> \param first ...
!> \param last ...
!> \param search1 ...
!> \param search2 ...
! **************************************************************************************************
   SUBROUTINE find_boundary2(num_array1, num_array2, ntot, first, last, search1, search2)
      INTEGER, POINTER                                   :: num_array1(:), num_array2(:)
      INTEGER                                            :: ntot, first, last, search1, search2

      CHARACTER(LEN=*), PARAMETER :: routineN = 'find_boundary2', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, tfirst, tlast
      LOGICAL                                            :: found

      found = .FALSE.
      first = 0
      last = ntot

      CALL find_boundary(num_array1, ntot, tfirst, tlast, search1)
      last = tlast
      DO i = tfirst, tlast
         IF (num_array2(i) == search2) THEN
            IF (.NOT. found) THEN
               first = i
            END IF
            found = .TRUE.
         ELSE
            IF (found) THEN
               last = i-1
               EXIT
            END IF
            found = .FALSE.
         END IF
      END DO

   END SUBROUTINE find_boundary2

END MODULE util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief provides a uniform framework to add references to CP2K
!>      cite and output these
!> \note
!>      references need to be input using the ISI citation format, because it is
!>      uniform, easy to parse, and can be exported for example from web of science
!>      furthermore, it can be easily converted to and from using the bibutils tools
!>      a collection of easy to use conversion programs that can be found at
!>      http://www.scripps.edu/~cdputnam/software/bibutils/
!>      by Chris Putnam
!>
!>      see thebibliography.F on how to add references easily
!> \par History
!>      08.2007 [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE reference_manager
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: default_string_length
   USE message_passing,                 ONLY: mp_max
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

  PUBLIC ::  print_reference, print_all_references, cite_reference
  PUBLIC ::  collect_citations_from_ranks

  INTEGER, PUBLIC, PARAMETER :: print_format_isi=101, &
                                print_format_journal=102, &
                                print_format_html=103

  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'reference_manager'

  ! maximum number of reference that can be added
  INTEGER, PARAMETER :: max_reference=1024

  ! storage of a reference
  INTEGER, PARAMETER :: doi_length=128
  INTEGER, PARAMETER :: ISI_length=128

  ! the way we store a reference, should remain fully private
! **************************************************************************************************
  TYPE reference_type
       PRIVATE
       ! the reference in a format as returned by the web of science
       CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER :: ISI_record
       ! the doi only, i.e. without "http://dx.doi.org/"
       CHARACTER(LEN=doi_length)                        :: DOI
       ! has this reference been cited in the program run
       LOGICAL                                          :: is_cited
       ! this is a citation key for output in the reference lists
       CHARACTER(LEN=ISI_length)                        :: citation_key
  END TYPE reference_type

  ! useful to build arrays
! **************************************************************************************************
  TYPE reference_p_type
     TYPE(reference_type), POINTER :: ref
  END TYPE

  ! thebibliography
  INTEGER, SAVE :: nbib=0
  TYPE(reference_p_type), DIMENSION(max_reference) :: thebib

  PUBLIC :: add_reference, & ! use this one only in bibliography.F
            remove_all_references, & ! use only in f77_interface.F
            get_citation_key         ! a string key describing the reference (e.g. Kohn1965b)

CONTAINS

! **************************************************************************************************
!> \brief marks a given reference as cited.
!> \param key citation key as returned from add_reference
!> \par History
!>      XX.2007 created [ ]
! **************************************************************************************************
  SUBROUTINE cite_reference(key)
      INTEGER, INTENT(IN)                                :: key

    IF (key<1 .OR. key>max_reference) CALL cp__b("common/reference_manager.F",88,"citation key out of range")

    ! set as cited
    thebib(key)%ref%is_cited=.TRUE.

  END SUBROUTINE

! **************************************************************************************************
!> \brief Checks for each reference if any mpi-rank has marked it for citation.
!> \param para_env ...
!> \par History
!>      12.2013 created [Ole Schuett]
! **************************************************************************************************
  SUBROUTINE collect_citations_from_ranks(para_env)
      TYPE(cp_para_env_type), POINTER                    :: para_env

      INTEGER                                            :: i, t

    DO i=1, nbib
       t = 0
       IF(thebib(i)%ref%is_cited) t=1
       CALL mp_max(t, para_env%group)
       thebib(i)%ref%is_cited = (t==1)
    ENDDO

  END SUBROUTINE collect_citations_from_ranks

! **************************************************************************************************
!> \brief add a reference to the bibliography
!> \param key output, this handle is needed to cite this reference later
!> \param ISI_record ...
!> \param DOI ...
!> \par History
!>      08.2007 created [Joost VandeVondele]
!> \note
!>      - see bibliography.F for it use.
!>      - the ISI record is space sensitive, in particular the first three characters need to be blank
!>        or contain a key indicating the record type. See the header of this file for tools
!>        that can convert e.g. bibtex or endnote files to the ISI format
!>      - DOI: provide the DOI without a link. The link will be automatically created as needed.
! **************************************************************************************************
  SUBROUTINE add_reference(key,ISI_record,DOI)
      INTEGER, INTENT(OUT)                               :: key
      CHARACTER(LEN=*), DIMENSION(:)                     :: ISI_record
      CHARACTER(LEN=*)                                   :: DOI

      CHARACTER(len=*), PARAMETER :: routineN = 'add_reference', routineP = moduleN//':'//routineN

      CHARACTER                                          :: tmp
      CHARACTER(LEN=ISI_length)                          :: author, citation_key, year
      INTEGER                                            :: commaloc, i, ires, line, match, mylen, &
                                                            nlines

    IF (nbib+1>max_reference) CALL cp__b("common/reference_manager.F",141,"increase max_reference")
    nbib=nbib+1
    key =nbib

    ! initialize reference to zero
    ALLOCATE(thebib(key)%ref)
    NULLIFY(thebib(key)%ref%ISI_record)
    thebib(key)%ref%DOI=""
    thebib(key)%ref%is_cited=.FALSE.

    ! Assign DOI
    thebib(key)%ref%DOI=DOI

    ! Assign ISI_record
    nlines=SIZE(ISI_record,1)
    ALLOCATE(thebib(key)%ref%ISI_record(nlines))
    thebib(key)%ref%ISI_record=ISI_record

    ! construct a citation_key
    line=1
    author=get_next_author(thebib(key)%ref%ISI_record,line)
    commaloc=INDEX(author,',')
    IF (commaloc.GT.0) author=author(1:commaloc-1)
    year=get_year(thebib(key)%ref%ISI_record)
    citation_key=TRIM(author)//TRIM(year)
    IF (citation_key=="") citation_key="unknown"

    ! avoid special characters in names, just remove them
    mylen=LEN_TRIM(citation_key)
    ires=0
    DO I=1,mylen
       IF (INDEX("0123456789thequickbrownfoxjumpsoverthelazydogTHEQUICKBROWNFOXJUMPSOVERTHELAZYDOG",citation_key(i:i)).NE.0) THEN
          ires=ires+1
          tmp=citation_key(i:i)
          citation_key(ires:ires)=tmp
       ENDIF
    ENDDO
    citation_key(ires+1:)=""

    ! avoid duplicates, search through the list for matches
    mylen=LEN_TRIM(citation_key)
    match=0
    DO I=1,nbib-1
       IF (thebib(I)%ref%citation_key(1:mylen)==citation_key(1:mylen)) match=match+1
    ENDDO
    IF (match>0) citation_key=citation_key(1:mylen)//CHAR(ICHAR('a')+match)

    ! finally store it
    thebib(key)%ref%citation_key=citation_key

  END SUBROUTINE add_reference

! **************************************************************************************************
!> \brief deallocate the bibliography
!> \par History
!>      08.2007 Joost VandeVondele [ ]
! **************************************************************************************************
  SUBROUTINE remove_all_references()
      INTEGER                                            :: i

    DO i=1,nbib
       IF (ASSOCIATED(thebib(i)%ref%ISI_record)) DEALLOCATE(thebib(i)%ref%ISI_record)
       thebib(i)%ref%DOI=""

       DEALLOCATE(thebib(i)%ref)
    ENDDO
  END SUBROUTINE remove_all_references
!****f* reference_manager/print_all_references *

! **************************************************************************************************
!> \brief printout of all references in a specific format
!>      optionally printing only those that are actually cited
!>      during program execution
!> \param cited_only print only those marked as cited
!> \param sorted sort entries most recent first according to the date,
!>              otherways sort with respect to key
!> \param FORMAT see module parameters print_format_XXXXXXXXX
!> \param unit ...
!> \param list optionally, output a sub-list only
!> \par History
!>      08.2007 Joost VandeVondele [ ]
! **************************************************************************************************
  SUBROUTINE print_all_references(cited_only,sorted,FORMAT,unit,list)
      LOGICAL, INTENT(IN)                                :: cited_only, sorted
      INTEGER, INTENT(IN)                                :: FORMAT, unit
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'print_all_references', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, irecord, nref
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: indx, irank, ival

! we'll sort the references wrt to the publication year
! the most recent first, publications without a year get last

    IF (PRESENT(list)) THEN
       nref=SIZE(list)
    ELSE
       nref=nbib
    ENDIF

    ALLOCATE(ival(nref))
    ALLOCATE(irank(nref))
    ALLOCATE(indx(nref))

    IF (PRESENT(list)) THEN
       indx(:)=list
    ELSE
       DO I=1,nref
          indx(I)=I
       ENDDO
    ENDIF

    DO I=1,nref
       irank(I)=I
    ENDDO

    IF (sorted) THEN
        DO I=1,nref
           ival(I)=-get_epoch(thebib(indx(I))%ref%ISI_record)
        ENDDO
    ELSE
        DO I=1,nref
           ival(I)=indx(I)
        ENDDO
    ENDIF
    CALL sort(ival,nref,irank)

    IF (FORMAT.EQ.print_format_html) THEN
       WRITE(unit,'(A)') '<TABLE border="1">'
    ENDIF
    DO I=1,nref
       irecord=indx(irank(I))
       IF (.NOT. cited_only .OR. thebib(irecord)%ref%is_cited) THEN
          SELECT CASE(FORMAT)
          CASE(print_format_isi)
          CASE(print_format_journal)
            WRITE(unit,'(A)') ""
          CASE(print_format_html)
            WRITE(unit,'(A)') "<TR><TD>"//'['//TRIM(thebib(irecord)%ref%citation_key)//']'//"</TD><TD>"
          CASE DEFAULT
            CALL cp__b("common/reference_manager.F",283,"print_reference: wrong format")
          END SELECT

          CALL print_reference(irecord,FORMAT,unit)

          SELECT CASE(FORMAT)
          CASE(print_format_isi)
          CASE(print_format_journal)
            WRITE(unit,'(A)') ""
          CASE(print_format_html)
            WRITE(unit,'(A)') '</TD></TR>'
          CASE DEFAULT
            CALL cp__b("common/reference_manager.F",295,"print_reference: wrong format")
          END SELECT
       ENDIF
    ENDDO
    IF (FORMAT.EQ.print_format_html) THEN
       WRITE(unit,'(A)') "</TABLE>"
    ENDIF

  END SUBROUTINE print_all_references
!****f* reference_manager/print_reference *

! **************************************************************************************************
!> \brief printout of a specified reference to a given unit in a selectable format
!> \param key as returned from add_reference
!> \param FORMAT see module parameters print_format_XXXXXXXXX
!> \param unit ...
!> \par History
!>      08.2007 Joost VandeVondele [ ]
! **************************************************************************************************
  SUBROUTINE print_reference(key,FORMAT,unit)
      INTEGER, INTENT(IN)                                :: key, FORMAT, unit

      CHARACTER(len=*), PARAMETER :: routineN = 'print_reference', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I

    IF (key<1 .OR. key>max_reference) CALL cp__b("common/reference_manager.F",322,"citation key out of range")

    SELECT CASE(FORMAT)
    CASE(print_format_isi)
      DO I=1,SIZE(thebib(key)%ref%ISI_record)
          WRITE(unit,'(T2,A)') TRIM(thebib(key)%ref%ISI_record(I))
      ENDDO
    CASE(print_format_journal)
      CALL print_reference_journal(key,unit)
    CASE(print_format_html)
      CALL print_reference_html(key,unit)
    CASE DEFAULT
      CALL cp__b("common/reference_manager.F",334,"print_reference: wrong format")
    END SELECT
  END SUBROUTINE print_reference

! **************************************************************************************************
!> \brief prints a reference in a journal style citation format,
!>      adding also a DOI link, which is convenient
!> \param key ...
!> \param unit ...
!> \par History
!>      08.2007 created [Joost VandeVondele]
! **************************************************************************************************
  SUBROUTINE print_reference_journal(key,unit)
      INTEGER, INTENT(IN)                                :: key, unit

      CHARACTER(len=*), PARAMETER :: routineN = 'print_reference_journal', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=4*ISI_length)                        :: journal
      CHARACTER(LEN=ISI_length)                          :: author, title
      INTEGER                                            :: iauthor, ipos_line, ititle, line

! write the author list

    WRITE(unit,'(T2,A)',ADVANCE="NO") ""
    line=1 ; iauthor=0 ; ipos_line=2
    author=get_next_author(thebib(key)%ref%ISI_record,line)
    DO WHILE (author.NE."")
       iauthor=iauthor+1
       IF (ipos_line+LEN_TRIM(author)>71) THEN
          WRITE(unit,'(A)') ";"
          WRITE(unit,'(T2,A)',ADVANCE="NO") ""
          ipos_line=2
       ELSE
          IF (iauthor.NE.1) WRITE(unit,'(A)',ADVANCE="NO") "; "
          ipos_line=ipos_line+2
       ENDIF
       WRITE(unit,'(A)',ADVANCE="NO") TRIM(author)
       ipos_line=ipos_line+LEN_TRIM(author)
       author=get_next_author(thebib(key)%ref%ISI_record,line)
    ENDDO
    IF (iauthor>0) THEN
        WRITE(unit,'(A)',ADVANCE="NO") ". "
        ipos_line=ipos_line+2
    ENDIF

    ! Journal, volume (issue), pages (year).
    journal=TRIM(get_source(thebib(key)%ref%ISI_record))
    IF (get_volume(thebib(key)%ref%ISI_record).NE."") THEN
       journal=TRIM(journal)//", "//get_volume(thebib(key)%ref%ISI_record)
       IF (get_issue(thebib(key)%ref%ISI_record).NE."") THEN
          journal=TRIM(journal)//" ("//TRIM(get_issue(thebib(key)%ref%ISI_record))//")"
       ENDIF
    END IF
    journal=TRIM(journal)//", "//get_pages(thebib(key)%ref%ISI_record)
    IF (get_year(thebib(key)%ref%ISI_record).NE."") THEN
        journal=TRIM(journal)//" ("//TRIM(get_year(thebib(key)%ref%ISI_record))//")."
    ENDIF
    IF (ipos_line+LEN_TRIM(journal)>71) THEN
       WRITE(unit,'(A)') ""
       WRITE(unit,'(T2,A)',ADVANCE="NO") ""
       ipos_line=2
    ENDIF
    WRITE(unit,'(A)',ADVANCE="NO") TRIM(journal)

    WRITE(unit,'(T2,A)') ""
    ! Title
    line=1 ; ititle=0
    title=get_next_title(thebib(key)%ref%ISI_record,line)
    DO WHILE (title.NE."")
       ititle=ititle+1
       IF (ititle.NE.1) WRITE(unit,'(A)') ""
       WRITE(unit,'(T2,A)', ADVANCE="NO") TRIM(title)
       title=get_next_title(thebib(key)%ref%ISI_record,line)
    ENDDO
    IF (ititle>0) WRITE(unit,'(A)') "."

    ! DOI
    IF (thebib(key)%ref%DOI .NE. "") THEN
      WRITE(unit,'(T2,A)') "http://dx.doi.org/"//TRIM(thebib(key)%ref%DOI)
    ENDIF

  END SUBROUTINE print_reference_journal

! **************************************************************************************************
!> \brief prints a reference in a journal style citation format,
!>      adding 'beautifying' html tags, and a link to the journal
!>      using the DOI
!> \param key ...
!> \param unit ...
!> \par History
!>      08.2007 created [Joost VandeVondele]
! **************************************************************************************************
  SUBROUTINE print_reference_html(key,unit)
      INTEGER, INTENT(IN)                                :: key, unit

      CHARACTER(LEN=ISI_length)                          :: author, title
      CHARACTER(LEN=ISI_length*4)                        :: journal
      INTEGER                                            :: iauthor, ititle, line

! write the author list

    WRITE(unit,'(T2,A,I0,A)',ADVANCE="NO") '<A NAME="reference_',key,'">'
    line=1 ; iauthor=0
    author=get_next_author(thebib(key)%ref%ISI_record,line)
    DO WHILE (author.NE."")
       iauthor=iauthor+1
       IF (iauthor.NE.1) WRITE(unit,'(A)',ADVANCE="NO") "; "
       WRITE(unit,'(A)',ADVANCE="NO") TRIM(author)
       author=get_next_author(thebib(key)%ref%ISI_record,line)
    ENDDO
    IF (iauthor>0) WRITE(unit,'(A)') ".<br>"

    ! DOI
    IF (thebib(key)%ref%DOI .NE. "") THEN
      WRITE(unit,'(T2,A)',ADVANCE="NO") '<A HREF="http://dx.doi.org/'//TRIM(thebib(key)%ref%DOI)//'">'
    ENDIF
    ! Journal, volume (issue), pages (year).
    journal=TRIM(get_source(thebib(key)%ref%ISI_record))
    IF (get_volume(thebib(key)%ref%ISI_record).NE."") THEN
       journal=TRIM(journal)//", "//get_volume(thebib(key)%ref%ISI_record)
       IF (get_issue(thebib(key)%ref%ISI_record).NE."") THEN
          journal=TRIM(journal)//" ("//TRIM(get_issue(thebib(key)%ref%ISI_record))//")"
       ENDIF
    END IF
    journal=TRIM(journal)//", "//get_pages(thebib(key)%ref%ISI_record)
    IF (get_year(thebib(key)%ref%ISI_record).NE."") THEN
        journal=TRIM(journal)//" ("//TRIM(get_year(thebib(key)%ref%ISI_record))//")."
    ENDIF
    WRITE(unit,'(A)',ADVANCE="NO") TRIM(journal)
    IF (thebib(key)%ref%DOI .NE. "") THEN
      WRITE(unit,'(A)',ADVANCE="NO") '</A>'
    ENDIF
    WRITE(unit,'(A)') "</A><br>"

    ! Title
    line=1 ; ititle=0
    title=get_next_title(thebib(key)%ref%ISI_record,line)
    DO WHILE (title.NE."")
       ititle=ititle+1
       IF (ititle.NE.1) WRITE(unit,'(A)') ""
       WRITE(unit,'(T2,A)', ADVANCE="NO") TRIM(title)
       title=get_next_title(thebib(key)%ref%ISI_record,line)
    ENDDO
    IF (ititle>0) WRITE(unit,'(A)') "."

  END SUBROUTINE print_reference_html

! **************************************************************************************************
!> \brief returns the corresponding fields from an ISI record.
!>       returns an empty string if the field can not be found
!>       iline_start should be initialized to 1 to obtain the first matching entry
!>       on return it is updated, so that succesive calls give succesive fields
!> \param ISI_record ...
!> \param iline_start ...
!> \retval res ...
!> \par History
!>      08.2007 created [Joost VandeVondele]
! **************************************************************************************************
  FUNCTION get_next_author(ISI_record,iline_start) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      INTEGER, INTENT(INOUT)                             :: iline_start
      CHARACTER(LEN=ISI_length)                          :: res

      INTEGER                                            :: I, N
      LOGICAL                                            :: in_au_section

    res=""
    in_au_section=.FALSE.
    N=SIZE(ISI_record,1)
    IF (iline_start>N) RETURN
    line_loop: DO I=1,N
       IF (ISI_record(I)(1:3)=="AU ") in_au_section=.TRUE.
       IF (in_au_section .AND. (ISI_record(I)(1:3)/="AU " .AND. ISI_record(I)(1:3)/="   ")) in_au_section=.FALSE.
       IF (in_au_section) THEN
          IF (I>=iline_start) THEN
             iline_start=I+1
             res=ISI_record(I)(4:)
             EXIT line_loop
          ENDIF
       ENDIF
    ENDDO line_loop

    ! We might want to fixup the initials, adding a dot after each of them

  END FUNCTION get_next_author

! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \param iline_start ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_next_title(ISI_record,iline_start) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      INTEGER, INTENT(INOUT)                             :: iline_start
      CHARACTER(LEN=ISI_length)                          :: res

      INTEGER                                            :: I, N
      LOGICAL                                            :: in_ti_section

    res=""

    in_ti_section=.FALSE.
    N=SIZE(ISI_record,1)
    IF (iline_start>N) RETURN
    line_loop: DO I=1,N
       IF (ISI_record(I)(1:3)=="TI ") in_ti_section=.TRUE.
       IF (in_ti_section .AND. (ISI_record(I)(1:3)/="TI " .AND. ISI_record(I)(1:3)/="   ")) in_ti_section=.FALSE.
       IF (in_ti_section) THEN
          IF (I>=iline_start) THEN
             iline_start=I+1
             res=ISI_record(I)(4:)
             EXIT line_loop
          ENDIF
       ENDIF
    ENDDO line_loop

  END FUNCTION get_next_title

! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_source(ISI_record) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      CHARACTER(LEN=4*ISI_length)                        :: res

      INTEGER                                            :: I, J, N

    N=SIZE(ISI_record,1)
    res=""
    DO I=1,N
       IF (ISI_record(I)(1:3)=="SO ") THEN
           res=ISI_record(I)(4:)
           DO J=I+1,N
              IF (ISI_record(J)(1:3)=="   ") THEN
                 res=TRIM(res)//" "//ISI_record(J)(4:)
              ELSE
                 EXIT
              ENDIF
           ENDDO
           EXIT
       ENDIF
    ENDDO
  END FUNCTION get_source

! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_year(ISI_record) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      CHARACTER(LEN=ISI_length)                          :: res

      INTEGER                                            :: I, N

    N=SIZE(ISI_record,1)
    res=""
    DO I=1,N
       IF (ISI_record(I)(1:3)=="PY ") res=ISI_record(I)(4:)
    ENDDO
  END FUNCTION get_year

! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_month(ISI_record) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      CHARACTER(LEN=ISI_length)                          :: res

      INTEGER                                            :: I, N

    N=SIZE(ISI_record,1)
    res=""
    DO I=1,N
       IF (ISI_record(I)(1:3)=="PD ") res=ISI_record(I)(4:6)
    ENDDO
  END FUNCTION get_month

! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_day(ISI_record) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      CHARACTER(LEN=ISI_length)                          :: res

      INTEGER                                            :: D, I, N

    N=SIZE(ISI_record,1)
    res=""
    DO I=1,N
       IF (ISI_record(I)(1:3)=="PD ") res=ISI_record(I)(7:)
    ENDDO
    ! PD can be e.g. OCT-NOV or OCT or OCT 27
    ! if res can't be read as an integer, it is not a day, and we bail out
    READ(res,*,ERR=998,END=998) D
    ! if the day is not in the expected range, we assume it is a parse error
    IF (D<0 .OR. D>31) res=""
    RETURN
998 CONTINUE
    res=""
  END FUNCTION get_day

! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_volume(ISI_record) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      CHARACTER(LEN=ISI_length)                          :: res

      INTEGER                                            :: I, N

    N=SIZE(ISI_record,1)
    res=""
    DO I=1,N
       IF (ISI_record(I)(1:3)=="VL ") res=ISI_record(I)(4:)
    ENDDO
  END FUNCTION get_volume

! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_issue(ISI_record) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      CHARACTER(LEN=ISI_length)                          :: res

      INTEGER                                            :: I, N

    N=SIZE(ISI_record,1)
    res=""
    DO I=1,N
       IF (ISI_record(I)(1:3)=="IS ") res=ISI_record(I)(4:)
    ENDDO
  END FUNCTION get_issue

! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_pages(ISI_record) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      CHARACTER(LEN=ISI_length)                          :: res

      CHARACTER(LEN=ISI_length)                          :: ar, bp, ep
      INTEGER                                            :: I, N

    N=SIZE(ISI_record,1)
    res=""
    bp=""
    ep=""
    ar=""

    DO I=1,N
       IF (ISI_record(I)(1:3)=="BP ") bp=ISI_record(I)(4:)
       IF (ISI_record(I)(1:3)=="EP ") ep=ISI_record(I)(4:)
       IF (ISI_record(I)(1:3)=="AR ") ar=ISI_record(I)(4:)
    ENDDO
    IF (bp.NE."") THEN
       res=bp
       IF (ep.NE."") res=TRIM(res)//"-"//ep
    ENDIF
    IF (res.EQ."".AND.ar.NE."") res=ar
  END FUNCTION get_pages

! **************************************************************************************************
!> \brief ...
!> \param key ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_citation_key(key) RESULT(res)
      INTEGER                                            :: key
      CHARACTER(LEN=default_string_length)               :: res

    res=thebib(key)%ref%citation_key
  END FUNCTION get_citation_key

!
! This returns something epoch like, but can only be used to order the records
! missing years, months, days are implied zero(1900)
!
! **************************************************************************************************
!> \brief ...
!> \param ISI_record ...
!> \retval res ...
! **************************************************************************************************
  FUNCTION get_epoch(ISI_record) RESULT(res)
      CHARACTER(LEN=ISI_length), DIMENSION(:), POINTER   :: ISI_record
      INTEGER                                            :: res

      CHARACTER(LEN=ISI_length)                          :: tmp
      INTEGER                                            :: day, istat, month, year

! read year

    tmp=get_year(ISI_record)
    READ(tmp,*,IOSTAT=istat) year
    IF (istat.NE.0) year=1900

    ! read day
    tmp=get_day(ISI_record)
    READ(tmp,*,IOSTAT=istat) day
    IF (istat.NE.0) day=0

    ! read month
    tmp=get_month(ISI_record)
    SELECT CASE(tmp)
    CASE("JAN")
      month=1
    CASE("FEB")
      month=2
    CASE("MAR")
      month=3
    CASE("APR")
      month=4
    CASE("MAY")
      month=5
    CASE("JUN")
      month=6
    CASE("JUL")
      month=7
    CASE("AUG")
      month=8
    CASE("SEP")
      month=9
    CASE("OCT")
      month=10
    CASE("NOV")
      month=11
    CASE("DEC")
      month=12
    CASE DEFAULT
      month=0
    END SELECT

    res=day+31*month+12*31*(year-1900)

  END FUNCTION get_epoch

END MODULE reference_manager
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Perform an abnormal program termination.
!> \note These routines are low-level and thus provide also an error recovery
!>       when dependencies do not allow the use of the error logger. Only
!>       the master (root) process will dump, if para_env is available and
!>       properly specified. Otherwise (without any information about the
!>       parallel environment) most likely more than one process or even all
!>       processes will send their error dump to the default output unit.
!> \par History
!>      - Routine external_control moved to a separate module
!> \author Matthias Krack (12.02.2001)
! **************************************************************************************************
MODULE termination
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              int_8
   USE machine,                         ONLY: m_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'termination'

   PUBLIC :: print_message
   PUBLIC :: stop_memory

CONTAINS

! **************************************************************************************************
!> \brief Perform a basic blocking of the text in message and print it
!>        optionally decorated with a frame of stars as defined by declev.
!> \param message ...
!> \param output_unit ...
!> \param declev ...
!> \param before ...
!> \param after ...
!> \date 28.08.1996
!> \par History
!>      - Translated to Fortran 90/95 (07.10.1999, Matthias Krack)
!>      - CP2K by JH 21.08.2000
!>      - Bugs in the dynamic format generation removed (09.02.2001, MK)
!>      - Revised (26.01.2011,MK)
!> \author Matthias Krack (MK)
!> \note
!>       after      : Number of empty lines after the message.
!>       before     : Number of empty lines before the message.
!>       declev     : Decoration level (0,1,2, ... star lines).
!>       message    : String with the message text.
!>       output_unit: Logical unit number of output unit.
! **************************************************************************************************
   SUBROUTINE print_message(message, output_unit, declev, before, after)

      CHARACTER(LEN=*), INTENT(IN)                       :: message
      INTEGER, INTENT(IN)                                :: output_unit
      INTEGER, INTENT(IN), OPTIONAL                      :: declev, before, after

      INTEGER                                            :: blank_lines_after, blank_lines_before, &
                                                            decoration_level, i, ibreak, ipos1, &
                                                            ipos2, maxrowlen, msglen, nrow, rowlen

      IF (PRESENT(after)) THEN
         blank_lines_after = MAX(after, 0)
      ELSE
         blank_lines_after = 1
      END IF

      IF (PRESENT(before)) THEN
         blank_lines_before = MAX(before, 0)
      ELSE
         blank_lines_before = 1
      END IF

      IF (PRESENT(declev)) THEN
         decoration_level = MAX(declev, 0)
      ELSE
         decoration_level = 0
      END IF

      IF (decoration_level == 0) THEN
         rowlen = 78
      ELSE
         rowlen = 70
      END IF

      msglen = LEN_TRIM(message)

      ! Calculate number of rows

      nrow = msglen/(rowlen+1)+1

      ! Calculate appropriate row length

      rowlen = MIN(msglen, rowlen)

      ! Generate the blank lines before the message

      DO i = 1, blank_lines_before
         WRITE (UNIT=output_unit, FMT="(A)") ""
      END DO

      ! Scan for the longest row

      ipos1 = 1
      ipos2 = rowlen
      maxrowlen = 0

      DO
         IF (ipos2 < msglen) THEN
            i = INDEX(message(ipos1:ipos2), " ", BACK=.TRUE.)
            IF (i == 0) THEN
               ibreak = ipos2
            ELSE
               ibreak = ipos1+i-2
            END IF
         ELSE
            ibreak = ipos2
         END IF

         maxrowlen = MAX(maxrowlen, ibreak-ipos1+1)

         ipos1 = ibreak+2
         ipos2 = MIN(msglen, ipos1+rowlen-1)

         ! When the last row is processed, exit loop

         IF (ipos1 > msglen) EXIT

      END DO

      ! Generate the first set of star rows

      IF (decoration_level > 1) THEN
         DO i = 1, decoration_level-1
            WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", maxrowlen+8)
         END DO
      END IF

      ! Break long messages

      ipos1 = 1
      ipos2 = rowlen

      DO
         IF (ipos2 < msglen) THEN
            i = INDEX(message(ipos1:ipos2), " ", BACK=.TRUE.)
            IF (i == 0) THEN
               ibreak = ipos2
            ELSE
               ibreak = ipos1+i-2
            END IF
         ELSE
            ibreak = ipos2
         END IF

         IF (decoration_level == 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A)") message(ipos1:ibreak)
         ELSE IF (decoration_level > 0) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A)") &
               "*** "//message(ipos1:ibreak)//REPEAT(" ", ipos1+maxrowlen-ibreak)//"***"
         END IF

         ipos1 = ibreak+2
         ipos2 = MIN(msglen, ipos1+rowlen-1)

         ! When the last row is processed, exit loop

         IF (ipos1 > msglen) EXIT
      END DO

      ! Generate the second set star rows

      IF (decoration_level > 1) THEN
         DO i = 1, decoration_level-1
            WRITE (UNIT=output_unit, FMT="(T2,A)") REPEAT("*", maxrowlen+8)
         END DO
      END IF

      ! Generate the blank lines after the message

      DO i = 1, blank_lines_after
         WRITE (UNIT=output_unit, FMT="(A)") ""
      END DO

   END SUBROUTINE print_message

! **************************************************************************************************
!> \brief The memory allocation for a data object failed. Print an error
!>        message and stop the program execution.
!> \param routineN ...
!> \param moduleN ...
!> \param line_number ...
!> \param object ...
!> \param memory ...
!> \par History
!>      - CP2K by JGH 21.08.2000
!>      - Revised (16.02.2011,MK)
!> \author Matthias Krack (12.10.1999,MK)
!> \note
!>      routineN   : Name of the calling routine
!>      moduleN    : Name of the calling module including the calling routine
!>      line_number: Code line number in the module
!>      object     : Name of the data object for which the allocation failed
!>      memory     : Size of the data object in bytes. If zero no size is
!>                   printed.
! **************************************************************************************************
   SUBROUTINE stop_memory(routineN, moduleN, line_number, object, memory)

      CHARACTER(LEN=*), INTENT(IN)                       :: routineN, moduleN
      INTEGER, INTENT(IN)                                :: line_number
      CHARACTER(LEN=*), INTENT(IN)                       :: object
      INTEGER, INTENT(IN)                                :: memory

      CHARACTER(LEN=default_path_length)                 :: message
      CHARACTER(len=default_string_length)               :: location
      INTEGER(KIND=int_8)                                :: total_memory

      CALL m_memory(total_memory)
      total_memory = total_memory/1024

      IF (memory == 0) THEN
         message = "The memory allocation for the data object <"//TRIM(object)// &
                   "> failed."
      ELSE
         WRITE (message, "(A,I0,A)") &
            "The memory allocation for the data object <"//TRIM(object)// &
            "> failed. The requested memory size is ", memory/1024, " KB."
      END IF

      IF (total_memory > 0) THEN
         WRITE (message, "(A,I0,A)") &
            TRIM(message)//" The current total memory allocated by CP2K is ", &
            total_memory, " KB."
      END IF

      WRITE (location, FMT='(A,I6)') TRIM(routineN)// &
         " (MODULE "//TRIM(moduleN)//")"// &
         ":", line_number

      CALL cp_abort(location, message)

   END SUBROUTINE stop_memory

END MODULE termination
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represents keywords in an input
!> \par History
!>      06.2004 created, based on Joost cp_keywords proposal [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE input_keyword_types
   USE cp_units,                        ONLY: cp_unit_create,&
                                              cp_unit_desc,&
                                              cp_unit_release,&
                                              cp_unit_type
   USE input_enumeration_types,         ONLY: enum_create,&
                                              enum_release,&
                                              enum_retain,&
                                              enumeration_type
   USE input_val_types,                 ONLY: &
        char_t, enum_t, integer_t, lchar_t, logical_t, no_t, real_t, val_create, val_release, &
        val_retain, val_type, val_write, val_write_internal
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE reference_manager,               ONLY: get_citation_key,&
                                              print_format_journal,&
                                              print_reference
   USE string_utilities,                ONLY: a2s,&
                                              compress,&
                                              substitute_special_xml_tokens,&
                                              typo_match,&
                                              uppercase
   USE termination,                     ONLY: print_message
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_keyword_types'

   INTEGER, PARAMETER, PUBLIC :: usage_string_length = default_string_length*2
   INTEGER, SAVE, PRIVATE :: last_keyword_id = 0

   PUBLIC :: keyword_p_type, keyword_type, keyword_create, keyword_retain, &
             keyword_release, keyword_get, keyword_describe, &
             keyword_describe_html, write_keyword_xml, keyword_typo_match

! **************************************************************************************************
!> \brief represent a pointer to a keyword (to make arrays of pointers)
!> \param keyword the pointer to the keyword
!> \author fawzi
! **************************************************************************************************
   TYPE keyword_p_type
      TYPE(keyword_type), POINTER :: keyword
   END TYPE keyword_p_type

! **************************************************************************************************
!> \brief represent a keyword in the input
!> \param names the names of the current keyword (at least one should be
!>        present) for example "MAXSCF"
!> \param usage how to use it "MAXSCF 10"
!> \param description what does it do: "MAXSCF : determines the maximum
!>        number of steps in an SCF run"
!> \param citations references to literature associated with this keyword
!> \param type_of_var the type of keyword (controls how it is parsed)
!>        it can be one of: no_parse_t,logical_t, integer_t, real_t,
!>        char_t
!> \param n_var number of values that should be parsed (-1=unknown)
!> \param repeats if the keyword can be present more than once in the
!>        section
!> \param enum enumeration that defines the mapping between integers and
!>        strings
!> \param unit the default unit this keyword is read in (to automatically
!>        convert to the internal cp2k units during parsing)
!> \param default_value the default value for the keyword
!> \param lone_keyword_value value to be used in presence of the keyword
!>        without any parameter
!> \note
!>      I have expressely avoided a format string for the type of keywords:
!>      they should easily map to basic types of fortran, if you need more
!>      information use a subsection. [fawzi]
!> \author Joost & fawzi
! **************************************************************************************************
   TYPE keyword_type
      INTEGER :: ref_count, id_nr
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER:: names
      CHARACTER(LEN=usage_string_length) :: usage
      CHARACTER, DIMENSION(:), POINTER :: description => Null()
      INTEGER, POINTER, DIMENSION(:) :: citations
      INTEGER :: type_of_var, n_var
      LOGICAL :: repeats
      TYPE(enumeration_type), POINTER :: enum
      TYPE(cp_unit_type), POINTER :: unit
      TYPE(val_type), POINTER :: default_value
      TYPE(val_type), POINTER :: lone_keyword_value
   END TYPE keyword_type

CONTAINS

! **************************************************************************************************
!> \brief creates a keyword object
!> \param keyword the keyword object to be created
!> \param name the name of the keyword
!> \param description ...
!> \param usage ...
!> \param type_of_var ...
!> \param n_var ...
!> \param repeats ...
!> \param variants ...
!> \param default_val ...
!> \param default_l_val ...
!> \param default_r_val ...
!> \param default_lc_val ...
!> \param default_c_val ...
!> \param default_i_val ...
!> \param default_l_vals ...
!> \param default_r_vals ...
!> \param default_c_vals ...
!> \param default_i_vals ...
!> \param lone_keyword_val ...
!> \param lone_keyword_l_val ...
!> \param lone_keyword_r_val ...
!> \param lone_keyword_c_val ...
!> \param lone_keyword_i_val ...
!> \param lone_keyword_l_vals ...
!> \param lone_keyword_r_vals ...
!> \param lone_keyword_c_vals ...
!> \param lone_keyword_i_vals ...
!> \param enum_c_vals ...
!> \param enum_i_vals ...
!> \param enum ...
!> \param enum_strict ...
!> \param enum_desc ...
!> \param unit_str ...
!> \param citations ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE keyword_create(keyword, name, description, usage, type_of_var, &
                             n_var, repeats, variants, default_val, &
                             default_l_val, default_r_val, default_lc_val, default_c_val, default_i_val, &
                             default_l_vals, default_r_vals, default_c_vals, default_i_vals, &
                             lone_keyword_val, lone_keyword_l_val, lone_keyword_r_val, lone_keyword_c_val, &
                             lone_keyword_i_val, lone_keyword_l_vals, lone_keyword_r_vals, &
                             lone_keyword_c_vals, lone_keyword_i_vals, enum_c_vals, enum_i_vals, &
                             enum, enum_strict, enum_desc, unit_str, citations)
      TYPE(keyword_type), POINTER                        :: keyword
      CHARACTER(len=*), INTENT(in)                       :: name, description
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: usage
      INTEGER, INTENT(in), OPTIONAL                      :: type_of_var, n_var
      LOGICAL, INTENT(in), OPTIONAL                      :: repeats
      CHARACTER(len=*), DIMENSION(:), INTENT(in), &
         OPTIONAL                                        :: variants
      TYPE(val_type), OPTIONAL, POINTER                  :: default_val
      LOGICAL, INTENT(in), OPTIONAL                      :: default_l_val
      REAL(KIND=DP), INTENT(in), OPTIONAL                :: default_r_val
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: default_lc_val, default_c_val
      INTEGER, INTENT(in), OPTIONAL                      :: default_i_val
      LOGICAL, DIMENSION(:), INTENT(in), OPTIONAL        :: default_l_vals
      REAL(KIND=DP), DIMENSION(:), INTENT(in), OPTIONAL  :: default_r_vals
      CHARACTER(len=*), DIMENSION(:), INTENT(in), &
         OPTIONAL                                        :: default_c_vals
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: default_i_vals
      TYPE(val_type), OPTIONAL, POINTER                  :: lone_keyword_val
      LOGICAL, INTENT(in), OPTIONAL                      :: lone_keyword_l_val
      REAL(KIND=DP), INTENT(in), OPTIONAL                :: lone_keyword_r_val
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: lone_keyword_c_val
      INTEGER, INTENT(in), OPTIONAL                      :: lone_keyword_i_val
      LOGICAL, DIMENSION(:), INTENT(in), OPTIONAL        :: lone_keyword_l_vals
      REAL(KIND=DP), DIMENSION(:), INTENT(in), OPTIONAL  :: lone_keyword_r_vals
      CHARACTER(len=*), DIMENSION(:), INTENT(in), &
         OPTIONAL                                        :: lone_keyword_c_vals
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: lone_keyword_i_vals
      CHARACTER(len=*), DIMENSION(:), INTENT(in), &
         OPTIONAL                                        :: enum_c_vals
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: enum_i_vals
      TYPE(enumeration_type), OPTIONAL, POINTER          :: enum
      LOGICAL, INTENT(in), OPTIONAL                      :: enum_strict
      CHARACTER(len=*), DIMENSION(:), INTENT(in), &
         OPTIONAL                                        :: enum_desc
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: unit_str
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: citations

      CHARACTER(len=*), PARAMETER :: routineN = 'keyword_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n
      LOGICAL                                            :: check

      IF(.NOT.(.NOT. ASSOCIATED(keyword)))CALL cp__a("input/input_keyword_types.F",190)
      ALLOCATE (keyword)
      keyword%ref_count = 1
      last_keyword_id = last_keyword_id+1
      keyword%id_nr = last_keyword_id
      NULLIFY (keyword%unit)

      IF (PRESENT(variants)) THEN
         ALLOCATE (keyword%names(SIZE(variants)+1))
         keyword%names(1) = name
         DO i = 1, SIZE(variants)
            keyword%names(i+1) = variants(i)
         END DO
      ELSE
         ALLOCATE (keyword%names(1))
         keyword%names(1) = name
      END IF
      DO i = 1, SIZE(keyword%names)
         CALL uppercase(keyword%names(i))
      END DO

      IF (PRESENT(usage)) THEN
         IF(.NOT.(LEN_TRIM(usage) <= LEN(keyword%usage)))CALL cp__a("input/input_keyword_types.F",212)
         keyword%usage = usage
      ELSE
         keyword%usage = ""
      END IF

      n = LEN_TRIM(description)
      ALLOCATE (keyword%description(n))
      DO i = 1, n
         keyword%description(i) = description(i:i)
      END DO

      IF (PRESENT(citations)) THEN
         ALLOCATE (keyword%citations(SIZE(citations, 1)))
         keyword%citations = citations
      ELSE
         NULLIFY (keyword%citations)
      ENDIF

      keyword%repeats = .FALSE.
      IF (PRESENT(repeats)) keyword%repeats = repeats

      NULLIFY (keyword%enum)
      IF (PRESENT(enum)) THEN
         keyword%enum => enum
         IF (ASSOCIATED(enum)) CALL enum_retain(enum)
      END IF
      IF (PRESENT(enum_i_vals)) THEN
         IF(.NOT.(PRESENT(enum_c_vals)))CALL cp__a("input/input_keyword_types.F",240)
         IF(.NOT.(.NOT. ASSOCIATED(keyword%enum)))CALL cp__a("input/input_keyword_types.F",241)
         CALL enum_create(keyword%enum, c_vals=enum_c_vals, i_vals=enum_i_vals, &
                          desc=enum_desc, strict=enum_strict)
      ELSE
         IF(.NOT.(.NOT. PRESENT(enum_c_vals)))CALL cp__a("input/input_keyword_types.F",245)
      END IF

      NULLIFY (keyword%default_value, keyword%lone_keyword_value)
      IF (PRESENT(default_val)) THEN
         IF (PRESENT(default_l_val) .OR. PRESENT(default_l_vals) .OR. &
             PRESENT(default_i_val) .OR. PRESENT(default_i_vals) .OR. &
             PRESENT(default_r_val) .OR. PRESENT(default_r_vals) .OR. &
             PRESENT(default_c_val) .OR. PRESENT(default_c_vals)) &
            CALL cp__b("input/input_keyword_types.F",254,"you should pass either default_val or a default value, not both")
         keyword%default_value => default_val
         IF (ASSOCIATED(default_val%enum)) THEN
            IF (ASSOCIATED(keyword%enum)) THEN
               IF(.NOT.(keyword%enum%id_nr == default_val%enum%id_nr))CALL cp__a("input/input_keyword_types.F",258)
            ELSE
               keyword%enum => default_val%enum
               CALL enum_retain(keyword%enum)
            END IF
         ELSE
            IF(.NOT.(.NOT. ASSOCIATED(keyword%enum)))CALL cp__a("input/input_keyword_types.F",264)
         END IF
         CALL val_retain(default_val)
      END IF
      IF (.NOT. ASSOCIATED(keyword%default_value)) THEN
         CALL val_create(keyword%default_value, l_val=default_l_val, &
                         l_vals=default_l_vals, i_val=default_i_val, i_vals=default_i_vals, &
                         r_val=default_r_val, r_vals=default_r_vals, c_val=default_c_val, &
                         c_vals=default_c_vals, lc_val=default_lc_val, enum=keyword%enum)
      END IF

      keyword%type_of_var = keyword%default_value%type_of_var
      IF (keyword%default_value%type_of_var == no_t) THEN
         CALL val_release(keyword%default_value)
      END IF

      IF (keyword%type_of_var == no_t) THEN
         IF (PRESENT(type_of_var)) THEN
            keyword%type_of_var = type_of_var
         ELSE
            CALL cp_warn(cp__l("input/input_keyword_types.F",284), &
                         "keyword "//TRIM(keyword%names(1))// &
                         " assumed undefined type by default")
         END IF
      ELSE IF (PRESENT(type_of_var)) THEN
         IF (keyword%type_of_var /= type_of_var) &
            CALL cp_abort(cp__l("input/input_keyword_types.F",290), &
                          "keyword "//TRIM(keyword%names(1))// &
                          " has a type different from the type of the default_value")
         keyword%type_of_var = type_of_var
      END IF

      IF (keyword%type_of_var == no_t) THEN
         CALL val_create(keyword%default_value)
      END IF

      IF (PRESENT(lone_keyword_val)) THEN
         IF (PRESENT(lone_keyword_l_val) .OR. PRESENT(lone_keyword_l_vals) .OR. &
             PRESENT(lone_keyword_i_val) .OR. PRESENT(lone_keyword_i_vals) .OR. &
             PRESENT(lone_keyword_r_val) .OR. PRESENT(lone_keyword_r_vals) .OR. &
             PRESENT(lone_keyword_c_val) .OR. PRESENT(lone_keyword_c_vals)) &
            CALL cp_abort(cp__l("input/input_keyword_types.F",305), &
                          "you should pass either lone_keyword_val or a lone_keyword value, not both")
         keyword%lone_keyword_value => lone_keyword_val
         CALL val_retain(lone_keyword_val)
         IF (ASSOCIATED(lone_keyword_val%enum)) THEN
            IF (ASSOCIATED(keyword%enum)) THEN
               IF (keyword%enum%id_nr /= lone_keyword_val%enum%id_nr) &
                  CALL cp__b("input/input_keyword_types.F",312,"keyword%enum%id_nr==lone_keyword_val%enum%id_nr")
            ELSE
               IF (ASSOCIATED(keyword%lone_keyword_value)) THEN
                  CALL cp__b("input/input_keyword_types.F",315,".NOT. ASSOCIATED(keyword%lone_keyword_value)")
               END IF
               keyword%enum => lone_keyword_val%enum
               CALL enum_retain(keyword%enum)
            END IF
         ELSE
            IF(.NOT.(.NOT. ASSOCIATED(keyword%enum)))CALL cp__a("input/input_keyword_types.F",321)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(keyword%lone_keyword_value)) THEN
         CALL val_create(keyword%lone_keyword_value, l_val=lone_keyword_l_val, &
                         l_vals=lone_keyword_l_vals, i_val=lone_keyword_i_val, i_vals=lone_keyword_i_vals, &
                         r_val=lone_keyword_r_val, r_vals=lone_keyword_r_vals, c_val=lone_keyword_c_val, &
                         c_vals=lone_keyword_c_vals, enum=keyword%enum)
      END IF
      IF (ASSOCIATED(keyword%lone_keyword_value)) THEN
         IF (keyword%lone_keyword_value%type_of_var == no_t) THEN
            CALL val_release(keyword%lone_keyword_value)
         ELSE
            IF (keyword%lone_keyword_value%type_of_var /= keyword%type_of_var) &
               CALL cp__b("input/input_keyword_types.F",335,"lone_keyword_value type incompatible with keyword type")
            ! lc_val cannot have lone_keyword_value!
            IF (keyword%type_of_var == enum_t) THEN
               IF (keyword%enum%strict) THEN
                  check = .FALSE.
                  DO i = 1, SIZE(keyword%enum%i_vals)
                     check = check .OR. (keyword%default_value%i_val(1) == keyword%enum%i_vals(i))
                  END DO
                  IF (.NOT. check) &
                     CALL cp__b("input/input_keyword_types.F",344,"default value not in enumeration : "//keyword%names(1))
               ENDIF
            ENDIF
         END IF
      END IF

      keyword%n_var = 1
      IF (ASSOCIATED(keyword%default_value)) THEN
         SELECT CASE (keyword%default_value%type_of_var)
         CASE (logical_t)
            keyword%n_var = SIZE(keyword%default_value%l_val)
         CASE (integer_t)
            keyword%n_var = SIZE(keyword%default_value%i_val)
         CASE (enum_t)
            IF (keyword%enum%strict) THEN
               check = .FALSE.
               DO i = 1, SIZE(keyword%enum%i_vals)
                  check = check .OR. (keyword%default_value%i_val(1) == keyword%enum%i_vals(i))
               END DO
               IF (.NOT. check) &
                  CALL cp__b("input/input_keyword_types.F",364,"default value not in enumeration : "//keyword%names(1))
            ENDIF
            keyword%n_var = SIZE(keyword%default_value%i_val)
         CASE (real_t)
            keyword%n_var = SIZE(keyword%default_value%r_val)
         CASE (char_t)
            keyword%n_var = SIZE(keyword%default_value%c_val)
         CASE (lchar_t)
            keyword%n_var = 1
         CASE (no_t)
            keyword%n_var = 0
         CASE default
            CALL cp__b("input/input_keyword_types.F",376,"")
         END SELECT
      END IF
      IF (PRESENT(n_var)) keyword%n_var = n_var
      IF (keyword%type_of_var == lchar_t .AND. keyword%n_var /= 1) &
         CALL cp__b("input/input_keyword_types.F",381,"arrays of lchar_t not supported : "//keyword%names(1))

      IF (PRESENT(unit_str)) THEN
         CALL cp_unit_create(keyword%unit, unit_str)
      END IF
   END SUBROUTINE keyword_create

! **************************************************************************************************
!> \brief retains the given keyword (see doc/ReferenceCounting.html)
!> \param keyword the keyword to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE keyword_retain(keyword)
      TYPE(keyword_type), POINTER                        :: keyword

      CHARACTER(len=*), PARAMETER :: routineN = 'keyword_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(keyword)))CALL cp__a("input/input_keyword_types.F",398)
      IF(.NOT.(keyword%ref_count > 0))CALL cp__a("input/input_keyword_types.F",399)
      keyword%ref_count = keyword%ref_count+1
   END SUBROUTINE keyword_retain

! **************************************************************************************************
!> \brief releases the given keyword (see doc/ReferenceCounting.html)
!> \param keyword the keyword to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE keyword_release(keyword)
      TYPE(keyword_type), POINTER                        :: keyword

      CHARACTER(len=*), PARAMETER :: routineN = 'keyword_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(keyword)) THEN
         IF(.NOT.(keyword%ref_count > 0))CALL cp__a("input/input_keyword_types.F",415)
         keyword%ref_count = keyword%ref_count-1
         IF (keyword%ref_count == 0) THEN
            DEALLOCATE (keyword%names)
            DEALLOCATE (keyword%description)
            CALL val_release(keyword%default_value)
            CALL val_release(keyword%lone_keyword_value)
            CALL enum_release(keyword%enum)
            CALL cp_unit_release(keyword%unit)
            IF (ASSOCIATED(keyword%citations)) THEN
               DEALLOCATE (keyword%citations)
            ENDIF
            DEALLOCATE (keyword)
         END IF
      END IF
      NULLIFY (keyword)
   END SUBROUTINE keyword_release

! **************************************************************************************************
!> \brief ...
!> \param keyword ...
!> \param names ...
!> \param usage ...
!> \param description ...
!> \param type_of_var ...
!> \param n_var ...
!> \param default_value ...
!> \param lone_keyword_value ...
!> \param repeats ...
!> \param enum ...
!> \param citations ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE keyword_get(keyword, names, usage, description, type_of_var, n_var, &
                          default_value, lone_keyword_value, repeats, enum, citations)
      TYPE(keyword_type), POINTER                        :: keyword
      CHARACTER(len=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: names
      CHARACTER(len=*), INTENT(out), OPTIONAL            :: usage, description
      INTEGER, INTENT(out), OPTIONAL                     :: type_of_var, n_var
      TYPE(val_type), OPTIONAL, POINTER                  :: default_value, lone_keyword_value
      LOGICAL, INTENT(out), OPTIONAL                     :: repeats
      TYPE(enumeration_type), OPTIONAL, POINTER          :: enum
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: citations

      CHARACTER(len=*), PARAMETER :: routineN = 'keyword_get', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(keyword)))CALL cp__a("input/input_keyword_types.F",462)
      IF(.NOT.(keyword%ref_count > 0))CALL cp__a("input/input_keyword_types.F",463)
      IF (PRESENT(names)) names => keyword%names
      IF (PRESENT(usage)) usage = keyword%usage
      IF (PRESENT(description)) description = a2s(keyword%description)
      IF (PRESENT(type_of_var)) type_of_var = keyword%type_of_var
      IF (PRESENT(n_var)) n_var = keyword%n_var
      IF (PRESENT(repeats)) repeats = keyword%repeats
      IF (PRESENT(default_value)) default_value => keyword%default_value
      IF (PRESENT(lone_keyword_value)) lone_keyword_value => keyword%lone_keyword_value
      IF (PRESENT(enum)) enum => keyword%enum
      IF (PRESENT(citations)) citations => keyword%citations
   END SUBROUTINE keyword_get

! **************************************************************************************************
!> \brief writes out a description of the keyword
!> \param keyword the keyword to describe
!> \param unit_nr the unit to write to
!> \param level the description level (0 no description, 1 name
!>        2: +usage, 3: +variants+description+default_value+repeats
!>        4: +type_of_var)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE keyword_describe(keyword, unit_nr, level)
      TYPE(keyword_type), POINTER                        :: keyword
      INTEGER, INTENT(in)                                :: unit_nr, level

      CHARACTER(len=*), PARAMETER :: routineN = 'keyword_describe', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: c_string
      INTEGER                                            :: i, l

      IF(.NOT.(ASSOCIATED(keyword)))CALL cp__a("input/input_keyword_types.F",495)
      IF(.NOT.(keyword%ref_count > 0))CALL cp__a("input/input_keyword_types.F",496)
      IF (level > 0 .AND. (unit_nr > 0)) THEN
         WRITE (unit_nr, "(a,a,a)") "                           ---", &
            TRIM(keyword%names(1)), "---"
         IF (level > 1) THEN
            WRITE (unit_nr, "(a,a)") "usage         : ", TRIM(keyword%usage)
         END IF
         IF (level > 2) THEN
            WRITE (unit_nr, "(a)") "description   : "
            CALL print_message(TRIM(a2s(keyword%description)), unit_nr, 0, 0, 0)
            IF (level > 3) THEN
               SELECT CASE (keyword%type_of_var)
               CASE (logical_t)
                  IF (keyword%n_var == -1) THEN
                     WRITE (unit_nr, "('  A list of logicals is expected')")
                  ELSE IF (keyword%n_var == 1) THEN
                     WRITE (unit_nr, "('  A logical is expected')")
                  ELSE
                     WRITE (unit_nr, "(i6,'  logicals are expected')") keyword%n_var
                  END IF
                  WRITE (unit_nr, "('  (T,TRUE,YES,ON) and (F,FALSE,NO,OFF) are synonyms')")
               CASE (integer_t)
                  IF (keyword%n_var == -1) THEN
                     WRITE (unit_nr, "('  A list of integers is expected')")
                  ELSE IF (keyword%n_var == 1) THEN
                     WRITE (unit_nr, "('  An integer is expected')")
                  ELSE
                     WRITE (unit_nr, "(i6,' integers are expected')") keyword%n_var
                  END IF
               CASE (real_t)
                  IF (keyword%n_var == -1) THEN
                     WRITE (unit_nr, "('  A list of reals is expected')")
                  ELSE IF (keyword%n_var == 1) THEN
                     WRITE (unit_nr, "('  A real is expected')")
                  ELSE
                     WRITE (unit_nr, "(i6,' reals are expected')") keyword%n_var
                  END IF
                  IF (ASSOCIATED(keyword%unit)) THEN
                     c_string = cp_unit_desc(keyword%unit, accept_undefined=.TRUE.)
                     WRITE (unit_nr, "('the default unit of measure is ',a)") &
                        TRIM(c_string)
                  END IF
               CASE (char_t)
                  IF (keyword%n_var == -1) THEN
                     WRITE (unit_nr, "('  A list of words is expected')")
                  ELSE IF (keyword%n_var == 1) THEN
                     WRITE (unit_nr, "('  A word is expected')")
                  ELSE
                     WRITE (unit_nr, "(i6,' words are expected')") keyword%n_var
                  END IF
               CASE (lchar_t)
                  WRITE (unit_nr, "('  A string is expected')")
               CASE (enum_t)
                  IF (keyword%n_var == -1) THEN
                     WRITE (unit_nr, "('  A list of keywords is expected')")
                  ELSE IF (keyword%n_var == 1) THEN
                     WRITE (unit_nr, "('  A keyword is expected')")
                  ELSE
                     WRITE (unit_nr, "(i6,' keywords are expected')") keyword%n_var
                  END IF
               CASE (no_t)
                  WRITE (unit_nr, "('  Non-standard type.')")
               CASE default
                  CALL cp__b("input/input_keyword_types.F",559,"")
               END SELECT
            END IF
            IF (keyword%type_of_var == enum_t) THEN
               IF (level > 3) THEN
                  WRITE (unit_nr, "('  valid keywords:')")
                  DO i = 1, SIZE(keyword%enum%c_vals)
                     c_string = keyword%enum%c_vals(i)
                     IF (LEN_TRIM(a2s(keyword%enum%desc(i)%chars)) > 0) THEN
                        WRITE (unit_nr, "('  - ',a,' : ',a,'.')") &
                           TRIM(c_string), TRIM(a2s(keyword%enum%desc(i)%chars))
                     ELSE
                        WRITE (unit_nr, "('  - ',a)") TRIM(c_string)
                     END IF
                  END DO
               ELSE
                  WRITE (unit_nr, "('  valid keywords:')", advance='NO')
                  l = 17
                  DO i = 1, SIZE(keyword%enum%c_vals)
                     c_string = keyword%enum%c_vals(i)
                     IF (l+LEN_TRIM(c_string) > 72 .AND. l > 14) THEN
                        WRITE (unit_nr, "(/,'    ')", advance='NO')
                        l = 4
                     END IF
                     WRITE (unit_nr, "(' ',a)", advance='NO') TRIM(c_string)
                     l = LEN_TRIM(c_string)+3
                  END DO
                  WRITE (unit_nr, "()")
               END IF
               IF (.NOT. keyword%enum%strict) THEN
                  WRITE (unit_nr, "('     other integer values are also accepted.')")
               END IF
            END IF
            IF (ASSOCIATED(keyword%default_value) .AND. keyword%type_of_var /= no_t) THEN
               WRITE (unit_nr, "('default_value : ')", advance="NO")
               CALL val_write(keyword%default_value, unit_nr=unit_nr)
            END IF
            IF (ASSOCIATED(keyword%lone_keyword_value) .AND. keyword%type_of_var /= no_t) THEN
               WRITE (unit_nr, "('lone_keyword  : ')", advance="NO")
               CALL val_write(keyword%lone_keyword_value, unit_nr=unit_nr)
            END IF
            IF (keyword%repeats) THEN
               WRITE (unit_nr, "(' and it can be repeated more than once')", advance="NO")
            END IF
            WRITE (unit_nr, "()")
            IF (SIZE(keyword%names) > 1) THEN
               WRITE (unit_nr, "(a)", advance="NO") "variants    : "
               DO i = 2, SIZE(keyword%names)
                  WRITE (unit_nr, "(a,' ')", advance="NO") keyword%names(i)
               END DO
               WRITE (unit_nr, "()")
            ENDIF
         END IF
      END IF
   END SUBROUTINE keyword_describe

! **************************************************************************************************
!> \brief writes out a description of the keyword
!> \param keyword the keyword to describe
!> \param unit_nr the unit to write to
!> \author Joost VandeVondele [10.2004], based on keyword_describe
! **************************************************************************************************
   SUBROUTINE keyword_describe_html(keyword, unit_nr)
      TYPE(keyword_type), POINTER                        :: keyword
      INTEGER, INTENT(in)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'keyword_describe_html', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: c_string, my_unit
      INTEGER                                            :: i, iref
      LOGICAL                                            :: any_description

      IF(.NOT.(ASSOCIATED(keyword)))CALL cp__a("input/input_keyword_types.F",632)
      IF(.NOT.(keyword%ref_count > 0))CALL cp__a("input/input_keyword_types.F",633)
      WRITE (unit_nr, '(a)') '<TR><TD WITDH="20%">'// &
         '<A NAME="'//TRIM(keyword%names(1))//'"><u>'//TRIM(keyword%names(1))//'</u></A>'// &
         '<TD WIDTH="80%">'
      WRITE (unit_nr, '(a)') '<TR><TD WIDTH="20%"><TD WIDTH="80%">'//TRIM(keyword%usage)
      WRITE (unit_nr, '(a)') '<TR><TD WIDTH="20%"><TD WIDTH="80%"><i>'//TRIM(a2s(keyword%description))//'</i>'
      WRITE (unit_nr, "(a)", advance="NO") '<TR><TD WIDTH="10%"><TD> This keyword '
      SELECT CASE (keyword%type_of_var)
      CASE (logical_t)
         IF (keyword%n_var == -1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a list of logicals'
         ELSE IF (keyword%n_var == 1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a logical'
         ELSE
            WRITE (unit_nr, '(a,i6,a)', ADVANCE="NO") 'expects precisely', keyword%n_var, ' logicals'
         END IF
         ! (provide a link to this info) WRITE(unit_nr,"('  (T,TRUE,YES,ON) and (F,FALSE,NO,OFF) are synonyms')")
      CASE (integer_t)
         IF (keyword%n_var == -1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a list of integers'
         ELSE IF (keyword%n_var == 1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects an integer'
         ELSE
            WRITE (unit_nr, '(a,i6,a)', ADVANCE="NO") 'expects precisely', keyword%n_var, ' integers'
         END IF
      CASE (real_t)
         IF (keyword%n_var == -1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a list of reals'
         ELSE IF (keyword%n_var == 1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a real'
         ELSE
            WRITE (unit_nr, '(a,i6,a)', ADVANCE="NO") 'expects precisely', keyword%n_var, ' reals'
         END IF
      CASE (char_t)
         IF (keyword%n_var == -1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a list of words'
         ELSE IF (keyword%n_var == 1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a word'
         ELSE
            WRITE (unit_nr, '(a,i6,a)', ADVANCE="NO") 'expects precisely', keyword%n_var, ' words'
         END IF
      CASE (lchar_t)
         WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a string'
      CASE (enum_t)
         IF (keyword%n_var == -1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a list of keywords'
         ELSE IF (keyword%n_var == 1) THEN
            WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a keyword'
         ELSE
            WRITE (unit_nr, '(a,i6,a)', ADVANCE="NO") 'expects precisely', keyword%n_var, ' keywords'
         END IF
      CASE (no_t)
         WRITE (unit_nr, '(a)', ADVANCE="NO") 'expects a non-standard input type'
      CASE DEFAULT
         CALL cp__b("input/input_keyword_types.F",687,"")
      END SELECT
      IF (keyword%repeats) THEN
         WRITE (unit_nr, "(', and may repeat')", ADVANCE="NO")
      END IF
      IF (ASSOCIATED(keyword%lone_keyword_value) .AND. keyword%type_of_var /= no_t) THEN
         WRITE (unit_nr, '(a)', advance="NO") '<TR><TD WIDTH="10%"><TD>This keyword behaves as a switch'
         CALL val_write(keyword%lone_keyword_value, unit_nr=unit_nr)
      END IF
      IF (ASSOCIATED(keyword%default_value) .AND. keyword%type_of_var /= no_t) THEN
         IF (ASSOCIATED(keyword%unit)) THEN
            my_unit = cp_unit_desc(keyword%unit, accept_undefined=.TRUE.)
            WRITE (unit_nr, '(a)', advance="NO") '. Default unit: '//'['//TRIM(my_unit)//'], default value: '
         ELSE
            WRITE (unit_nr, '(a)', advance="NO") '. Default value: '
         ENDIF
         CALL val_write(keyword%default_value, unit=keyword%unit, unit_nr=unit_nr)
         IF (ASSOCIATED(keyword%unit)) THEN
            WRITE (unit_nr, '(a)', advance="NO") TRIM(my_unit)
         END IF
      ELSE
         IF (ASSOCIATED(keyword%unit)) THEN
            my_unit = cp_unit_desc(keyword%unit, accept_undefined=.TRUE.)
            WRITE (unit_nr, '(a)', advance="NO") '. Default unit: '//'['//TRIM(my_unit)//'].'
         END IF
      ENDIF

      IF (keyword%type_of_var == enum_t) THEN
         any_description = .FALSE.
         DO i = 1, SIZE(keyword%enum%c_vals)
            any_description = any_description .OR. SIZE(keyword%enum%desc(i)%chars) > 1
         ENDDO
         IF (any_description) THEN
            WRITE (unit_nr, '(a)', advance='NO') '<TR><TD WIDTH="10%"><TD>  valid keywords:<ul> '
            DO i = 1, SIZE(keyword%enum%c_vals)
               c_string = keyword%enum%c_vals(i)
               WRITE (unit_nr, '(a,a,a,a,a)') "<li><code>", &
                  TRIM(c_string), "</code>: ", TRIM(a2s(keyword%enum%desc(i)%chars)), "</li>"
            END DO
            WRITE (unit_nr, '(a)', advance='NO') '</ul>'
         ELSE
            WRITE (unit_nr, '(a)', advance='NO') '<TR><TD WIDTH="10%"><TD>  valid keywords: '
            IF (SIZE(keyword%enum%c_vals) > 0) &
               WRITE (unit_nr, '(a)', advance='NO') TRIM(keyword%enum%c_vals(1))
            DO i = 2, SIZE(keyword%enum%c_vals)
               c_string = keyword%enum%c_vals(i)
               WRITE (unit_nr, '(a)', advance='NO') ", "//TRIM(c_string)
            END DO
         ENDIF
         IF (.NOT. keyword%enum%strict) THEN
            WRITE (unit_nr, "('     other integer values are also accepted.')")
         END IF
      END IF
      IF (SIZE(keyword%names) > 1) THEN
         WRITE (unit_nr, "(a)", advance="NO") '<TR><TD WIDTH="10%"><TD>variants: '
         DO i = 2, SIZE(keyword%names)
            WRITE (unit_nr, "(a)", advance="NO") TRIM(keyword%names(i))
            IF (i .NE. SIZE(keyword%names)) THEN
               WRITE (unit_nr, '(a)', advance='NO') ', '
            ENDIF
         END DO
      ENDIF
      IF (ASSOCIATED(keyword%citations)) THEN
         IF (SIZE(keyword%citations, 1) > 1) THEN
            WRITE (unit_nr, FMT='(A)') '<BR> This keyword cites following references: '
         ELSE
            WRITE (unit_nr, FMT='(A)') '<BR> This keyword cites following reference: '
         ENDIF
         DO iref = 1, SIZE(keyword%citations, 1)
            WRITE (unit_nr, FMT='(A,I0,A)') &
               '<A HREF="references.html#reference_', keyword%citations(iref), '" TITLE="'
            CALL print_reference(keyword%citations(iref), FORMAT=print_format_journal, unit=unit_nr)
            WRITE (unit_nr, FMT='(A)') &
               '">['//TRIM(get_citation_key(keyword%citations(iref)))//']</A>'
         ENDDO
      ENDIF

   END SUBROUTINE keyword_describe_html

! **************************************************************************************************
!> \brief Prints a description of a keyword in XML format
!> \param keyword The keyword to describe
!> \param level ...
!> \param unit_number Number of the output unit
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE write_keyword_xml(keyword, level, unit_number)

      TYPE(keyword_type), POINTER                        :: keyword
      INTEGER, INTENT(IN)                                :: level, unit_number

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_keyword_xml', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1000)                                :: string
      CHARACTER(LEN=3)                                   :: repeats
      CHARACTER(LEN=8)                                   :: short_string
      INTEGER                                            :: i, l0, l1, l2, l3, l4

      IF(.NOT.(ASSOCIATED(keyword)))CALL cp__a("input/input_keyword_types.F",786)
      IF(.NOT.(keyword%ref_count > 0))CALL cp__a("input/input_keyword_types.F",787)

      ! Indentation for current level, next level, etc.

      l0 = level
      l1 = level+1
      l2 = level+2
      l3 = level+3
      l4 = level+4

      IF (keyword%repeats) THEN
         repeats = "yes"
      ELSE
         repeats = "no "
      END IF

      ! Write (special) keyword element

      IF (keyword%names(1) == "_SECTION_PARAMETERS_") THEN
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l0)//"<SECTION_PARAMETERS repeats="""//TRIM(repeats)//""">", &
            REPEAT(" ", l1)//"<NAME type=""default"">SECTION_PARAMETERS</NAME>"
      ELSE IF (keyword%names(1) == "_DEFAULT_KEYWORD_") THEN
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l0)//"<DEFAULT_KEYWORD repeats="""//TRIM(repeats)//""">", &
            REPEAT(" ", l1)//"<NAME type=""default"">DEFAULT_KEYWORD</NAME>"
      ELSE
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l0)//"<KEYWORD repeats="""//TRIM(repeats)//""">", &
            REPEAT(" ", l1)//"<NAME type=""default"">"// &
            TRIM(keyword%names(1))//"</NAME>"
      END IF

      DO i = 2, SIZE(keyword%names)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<NAME type=""alias"">"// &
            TRIM(keyword%names(i))//"</NAME>"
      END DO

      SELECT CASE (keyword%type_of_var)
      CASE (logical_t)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DATA_TYPE kind=""logical"">"
      CASE (integer_t)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DATA_TYPE kind=""integer"">"
      CASE (real_t)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DATA_TYPE kind=""real"">"
      CASE (char_t)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DATA_TYPE kind=""word"">"
      CASE (lchar_t)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DATA_TYPE kind=""string"">"
      CASE (enum_t)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DATA_TYPE kind=""keyword"">"
         IF (keyword%enum%strict) THEN
            WRITE (UNIT=unit_number, FMT="(A)") &
               REPEAT(" ", l2)//"<ENUMERATION strict=""yes"">"
         ELSE
            WRITE (UNIT=unit_number, FMT="(A)") &
               REPEAT(" ", l2)//"<ENUMERATION strict=""no"">"
         END IF
         DO i = 1, SIZE(keyword%enum%c_vals)
            WRITE (UNIT=unit_number, FMT="(A)") &
               REPEAT(" ", l3)//"<ITEM>", &
               REPEAT(" ", l4)//"<NAME>"// &
               TRIM(ADJUSTL(substitute_special_xml_tokens(keyword%enum%c_vals(i))))//"</NAME>", &
               REPEAT(" ", l4)//"<DESCRIPTION>"// &
               TRIM(ADJUSTL(substitute_special_xml_tokens(a2s(keyword%enum%desc(i)%chars)))) &
               //"</DESCRIPTION>", REPEAT(" ", l3)//"</ITEM>"
         END DO
         WRITE (UNIT=unit_number, FMT="(A)") REPEAT(" ", l2)//"</ENUMERATION>"
      CASE (no_t)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DATA_TYPE kind=""non-standard type"">"
      CASE DEFAULT
         CALL cp__b("input/input_keyword_types.F",866,"")
      END SELECT

      short_string = ""
      WRITE (UNIT=short_string, FMT="(I8)") keyword%n_var
      WRITE (UNIT=unit_number, FMT="(A)") &
         REPEAT(" ", l2)//"<N_VAR>"//TRIM(ADJUSTL(short_string))//"</N_VAR>", &
         REPEAT(" ", l1)//"</DATA_TYPE>"

      WRITE (UNIT=unit_number, FMT="(A)") REPEAT(" ", l1)//"<USAGE>"// &
         TRIM(substitute_special_xml_tokens(keyword%usage)) &
         //"</USAGE>"

      WRITE (UNIT=unit_number, FMT="(A)") REPEAT(" ", l1)//"<DESCRIPTION>"// &
         TRIM(substitute_special_xml_tokens(a2s(keyword%description))) &
         //"</DESCRIPTION>"

      IF (ASSOCIATED(keyword%default_value) .AND. &
          (keyword%type_of_var /= no_t)) THEN
         IF (ASSOCIATED(keyword%unit)) THEN
            CALL val_write_internal(val=keyword%default_value, &
                                    string=string, &
                                    unit=keyword%unit)
         ELSE
            CALL val_write_internal(val=keyword%default_value, &
                                    string=string)
         END IF
         CALL compress(string)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DEFAULT_VALUE>"// &
            TRIM(ADJUSTL(substitute_special_xml_tokens(string)))//"</DEFAULT_VALUE>"
      END IF

      IF (ASSOCIATED(keyword%unit)) THEN
         string = cp_unit_desc(keyword%unit, accept_undefined=.TRUE.)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<DEFAULT_UNIT>"// &
            TRIM(ADJUSTL(string))//"</DEFAULT_UNIT>"
      END IF

      IF (ASSOCIATED(keyword%lone_keyword_value) .AND. &
          (keyword%type_of_var /= no_t)) THEN
         CALL val_write_internal(val=keyword%lone_keyword_value, &
                                 string=string)
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l1)//"<LONE_KEYWORD_VALUE>"// &
            TRIM(ADJUSTL(substitute_special_xml_tokens(string)))//"</LONE_KEYWORD_VALUE>"
      END IF

      IF (ASSOCIATED(keyword%citations)) THEN
         DO i = 1, SIZE(keyword%citations, 1)
            short_string = ""
            WRITE (UNIT=short_string, FMT="(I8)") keyword%citations(i)
            WRITE (UNIT=unit_number, FMT="(A)") &
               REPEAT(" ", l1)//"<REFERENCE>", &
               REPEAT(" ", l2)//"<NAME>"//TRIM(get_citation_key(keyword%citations(i)))//"</NAME>", &
               REPEAT(" ", l2)//"<NUMBER>"//TRIM(ADJUSTL(short_string))//"</NUMBER>", &
               REPEAT(" ", l1)//"</REFERENCE>"
         END DO
      END IF

      ! Close (special) keyword section

      IF (keyword%names(1) == "_SECTION_PARAMETERS_") THEN
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l0)//"</SECTION_PARAMETERS>"
      ELSE IF (keyword%names(1) == "_DEFAULT_KEYWORD_") THEN
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l0)//"</DEFAULT_KEYWORD>"
      ELSE
         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l0)//"</KEYWORD>"
      END IF

   END SUBROUTINE write_keyword_xml

! **************************************************************************************************
!> \brief ...
!> \param keyword ...
!> \param unknown_string ...
!> \param location_string ...
!> \param matching_rank ...
!> \param matching_string ...
!> \param bonus ...
! **************************************************************************************************
   SUBROUTINE keyword_typo_match(keyword, unknown_string, location_string, matching_rank, matching_string, bonus)

      TYPE(keyword_type), POINTER                        :: keyword
      CHARACTER(LEN=*)                                   :: unknown_string, location_string
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: matching_rank
      CHARACTER(LEN=*), DIMENSION(:), INTENT(INOUT)      :: matching_string
      INTEGER, INTENT(IN)                                :: bonus

      CHARACTER(LEN=*), PARAMETER :: routineN = 'keyword_typo_match', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=LEN(matching_string(1)))             :: line
      INTEGER                                            :: i, imatch, imax, irank, j, k

      IF(.NOT.(ASSOCIATED(keyword)))CALL cp__a("input/input_keyword_types.F",965)
      IF(.NOT.(keyword%ref_count > 0))CALL cp__a("input/input_keyword_types.F",966)

      DO i = 1, SIZE(keyword%names)
         imatch = typo_match(TRIM(keyword%names(i)), TRIM(unknown_string))
         IF (imatch > 0) THEN
            imatch = imatch+bonus
            WRITE (line, '(T2,A)') " keyword "//TRIM(keyword%names(i))//" in section "//TRIM(location_string)
            imax = SIZE(matching_rank, 1)
            irank = imax+1
            DO k = imax, 1, -1
               IF (imatch > matching_rank(k)) irank = k
            ENDDO
            IF (irank <= imax) THEN
               matching_rank(irank+1:imax) = matching_rank(irank:imax-1)
               matching_string(irank+1:imax) = matching_string(irank:imax-1)
               matching_rank(irank) = imatch
               matching_string(irank) = line
            ENDIF
         END IF

         IF (keyword%type_of_var == enum_t) THEN
            DO j = 1, SIZE(keyword%enum%c_vals)
               imatch = typo_match(TRIM(keyword%enum%c_vals(j)), TRIM(unknown_string))
               IF (imatch > 0) THEN
                  imatch = imatch+bonus
                  WRITE (line, '(T2,A)') " enum "//TRIM(keyword%enum%c_vals(j))// &
                     " in section "//TRIM(location_string)// &
                     " for keyword "//TRIM(keyword%names(i))
                  imax = SIZE(matching_rank, 1)
                  irank = imax+1
                  DO k = imax, 1, -1
                     IF (imatch > matching_rank(k)) irank = k
                  ENDDO
                  IF (irank <= imax) THEN
                     matching_rank(irank+1:imax) = matching_rank(irank:imax-1)
                     matching_string(irank+1:imax) = matching_string(irank:imax-1)
                     matching_rank(irank) = imatch
                     matching_string(irank) = line
                  ENDIF
               END IF
            END DO
         END IF
      ENDDO

   END SUBROUTINE keyword_typo_match

END MODULE input_keyword_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_val
   USE input_val_types,                 ONLY: val_p_type,&
                                              val_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_val'

!API type
   PUBLIC :: cp_sll_val_type, cp_sll_val_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_val_create, cp_sll_val_dealloc, &
             cp_sll_val_next, &
             cp_sll_val_get_length, &
             cp_sll_val_get_el_at, cp_sll_val_set_el_at, &
             cp_sll_val_get_first_el, cp_sll_val_get_rest, &
             cp_sll_val_insert_el, cp_sll_val_insert_el_at, &
             cp_sll_val_rm_first_el, cp_sll_val_rm_el_at, &
             cp_sll_val_rm_all_el, &
             cp_sll_val_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_val_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_val_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_val_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_val_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_val_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_val_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_val_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_val_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_val_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_val_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_val_rm_first_el, &
         cp_sll_val_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_val_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_val_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_val_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_val_type
      TYPE(val_type), POINTER :: first_el
      TYPE(cp_sll_val_type), POINTER :: rest
   END TYPE cp_sll_val_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_val_p_type
      TYPE(cp_sll_val_type), POINTER :: list
   END TYPE cp_sll_val_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_val_create(sll, first_el, rest)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      TYPE(val_type), OPTIONAL, POINTER                  :: first_el
      TYPE(cp_sll_val_type), OPTIONAL, POINTER           :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el => first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_val_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_val_dealloc(sll)
      TYPE(cp_sll_val_type), POINTER                     :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_val_rm_all_el(sll)
   END SUBROUTINE cp_sll_val_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_val_dealloc_node(sll)
      TYPE(cp_sll_val_type), POINTER                     :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_val_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_val_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      TYPE(val_type), POINTER                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_get_first_el', &
         routineP = moduleN//':'//routineN

      res => sll%first_el
   END FUNCTION cp_sll_val_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_val_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_val_type), POINTER                     :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("input/cp_linked_list_val.F",294,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_val_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_val_get_length(sll) RESULT(res)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_val_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_val_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(val_type), POINTER                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: pos

      IF (index == -1) THEN
         pos => cp_sll_val_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_val_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_val.F",366)

      res => pos%first_el
   END FUNCTION cp_sll_val_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_val_set_el_at(sll, index, value)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(val_type), POINTER                            :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: pos

      IF (index == -1) THEN
         pos => cp_sll_val_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_val_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_val.F",398)

      pos%first_el => value
   END SUBROUTINE cp_sll_val_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_val_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_val_type), POINTER                     :: iterator
      TYPE(val_type), OPTIONAL, POINTER                  :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_next', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att => iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_val_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_val_insert_el(sll, el)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      TYPE(val_type), POINTER                            :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_val_create(newSlot, first_el=el, &
                             rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_val_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_val_rm_first_el(sll)
      TYPE(cp_sll_val_type), POINTER                     :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_val_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("input/cp_linked_list_val.F",484,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_val_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_val_insert_el_at(sll, el, index)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      TYPE(val_type), POINTER                            :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: pos

      IF (index == 1) THEN
         CALL cp_sll_val_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_val_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_val_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_val.F",518)
         CALL cp_sll_val_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_val_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_val_rm_el_at(sll, index)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: pos

      IF (index == 1) THEN
         CALL cp_sll_val_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_val_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_val_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_val.F",550)
         CALL cp_sll_val_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_val_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_val_rm_all_el(sll)
      TYPE(cp_sll_val_type), POINTER                     :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_val_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_val_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_val_to_array(sll) RESULT(res)
      TYPE(cp_sll_val_type), POINTER                     :: sll
      TYPE(val_p_type), DIMENSION(:), POINTER            :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_val_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_val_type), POINTER                     :: iter

      len = cp_sll_val_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i)%val => iter%first_el
         IF(.NOT.(cp_sll_val_next(iter) .OR. i == len))CALL cp__a("input/cp_linked_list_val.F",606)
      END DO
   END FUNCTION cp_sll_val_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "=>"
!  USE = "USE input_val_types, only: val_type, val_p_type"
!  array= = "=>"
!  arrayEl = "%val"
!  common_dir = "../common"
!  defines =
!    "#define CP_SLL_VAL_LESS_Q(el1,el2) el1%id_nr<el2%id_nr
!     #define ( el1%id_nr == el2%id_nr ) ( el1%id_nr == el2%id_nr )
!     "
!  equalQ = "CP_SLL_VAL_EQUAL_Q"
!  lessQ = "CP_SLL_VAL_LESS_Q"
!  nametype1 = "val"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "type(val_type),pointer"
!  type1arrayEl = "type(val_p_type)"
!  type1in = "type(val_type),pointer"
!  type1out = "type(val_type),pointer"
!  write_el = ""

END MODULE cp_linked_list_val

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief objects that represent the structure of input sections and the data
!>      contained in an input section
!> \par History
!>      06.2004 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE input_section_types
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_linked_list_val,              ONLY: &
        cp_sll_val_create, cp_sll_val_dealloc, cp_sll_val_get_el_at, cp_sll_val_get_length, &
        cp_sll_val_get_rest, cp_sll_val_insert_el_at, cp_sll_val_next, cp_sll_val_p_type, &
        cp_sll_val_rm_el_at, cp_sll_val_set_el_at, cp_sll_val_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE input_keyword_types,             ONLY: keyword_describe,&
                                              keyword_describe_html,&
                                              keyword_p_type,&
                                              keyword_release,&
                                              keyword_retain,&
                                              keyword_type,&
                                              keyword_typo_match,&
                                              write_keyword_xml
   USE input_val_types,                 ONLY: lchar_t,&
                                              no_t,&
                                              val_create,&
                                              val_duplicate,&
                                              val_get,&
                                              val_release,&
                                              val_type,&
                                              val_write
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE reference_manager,               ONLY: get_citation_key,&
                                              print_format_journal,&
                                              print_reference
   USE string_utilities,                ONLY: a2s,&
                                              compress,&
                                              substitute_special_xml_tokens,&
                                              typo_match,&
                                              uppercase
   USE termination,                     ONLY: print_message
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   INTEGER, SAVE, PRIVATE :: last_section_id = 0, last_section_vals_id = 0
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_section_types'

   PUBLIC :: section_type
   PUBLIC :: section_create, section_release, section_describe, &
             section_get_subsection, section_get_keyword, &
             section_add_keyword, section_add_subsection, section_describe_html, &
             section_describe_index_html
   PUBLIC :: section_get_subsection_index, section_get_keyword_index

   PUBLIC :: section_vals_type
   PUBLIC :: section_vals_create, section_vals_retain, section_vals_release, &
             section_vals_get, section_vals_get_subs_vals, section_vals_val_get, section_vals_list_get, &
             section_vals_write, section_vals_add_values, section_vals_get_subs_vals2, &
             section_vals_val_set, section_vals_val_unset, section_vals_get_subs_vals3, &
             section_vals_set_subs_vals, section_vals_duplicate, section_vals_remove_values
   PUBLIC :: write_section_xml

   PUBLIC :: section_get_ival, &
             section_get_ivals, &
             section_get_rval, &
             section_get_lval
   PUBLIC :: section_typo_match, typo_match_section, typo_matching_rank, typo_matching_line

! **************************************************************************************************
!> \brief represent a pointer to a section (to make arrays of pointers)
!> \param section the pointer to the section
!> \author fawzi
! **************************************************************************************************
   TYPE section_p_type
      TYPE(section_type), POINTER :: section
   END TYPE section_p_type

! **************************************************************************************************
!> \brief represent a section of the input file
!> \note
!>      - frozen: if the section has been frozen (and no keyword/subsections
!>        can be added)
!>      - repeats: if the section can be repeated more than once in the same
!>        context
!>      - id_nr: identification number (different in each instance)
!>      - ref_count: reference count (see doc/ReferenceCounting.html)
!>      - n_keywords: the number of keywords in this section
!>      - name: name of the section
!>      - description: description of the section
!>      - citations: references to literature associated to this section
!>      - keywords: array with the keywords of this section (might be
!>        oversized)
!>      - subsections: sections contained in this section
!> \author fawzi
! **************************************************************************************************
   TYPE section_type
      LOGICAL :: frozen, repeats
      INTEGER :: id_nr, ref_count, n_keywords, n_subsections
      CHARACTER(len=default_string_length)        :: name
      CHARACTER, DIMENSION(:), POINTER              :: description => Null()
      INTEGER, POINTER, DIMENSION(:)              :: citations
      TYPE(keyword_p_type), DIMENSION(:), POINTER :: keywords
      TYPE(section_p_type), POINTER, DIMENSION(:) :: subsections
   END TYPE section_type

! **************************************************************************************************
!> \brief repesents a pointer to a parsed section (to make arrays of pointers)
!> \param section_vals the pointer to the parsed section
!> \author fawzi
! **************************************************************************************************
   TYPE section_vals_p_type
      TYPE(section_vals_type), POINTER :: section_vals
   END TYPE section_vals_p_type

! **************************************************************************************************
!> \brief stores the values of a section
!> \author fawzi
! **************************************************************************************************
   TYPE section_vals_type
      INTEGER :: ref_count, id_nr
      INTEGER, POINTER, DIMENSION(:)                     :: ibackup
      TYPE(section_type), POINTER                         :: section
      TYPE(cp_sll_val_p_type), DIMENSION(:, :), POINTER   :: values
      TYPE(section_vals_p_type), DIMENSION(:, :), POINTER :: subs_vals
   END TYPE section_vals_type

   TYPE(section_type), POINTER, SAVE                                :: typo_match_section => NULL()
   INTEGER, PARAMETER                                               :: n_typo_matches = 5
   INTEGER, DIMENSION(n_typo_matches)                               :: typo_matching_rank = 0
   CHARACTER(LEN=default_string_length*5), DIMENSION(n_typo_matches):: typo_matching_line = ""

CONTAINS

! **************************************************************************************************
!> \brief creates a list of keywords
!> \param section the list to be created
!> \param name ...
!> \param description ...
!> \param n_keywords hint about the number of keywords, defaults to 10
!> \param n_subsections a hint about how many sections will be added to this
!>        structure, defaults to 0
!> \param repeats if this section can repeat (defaults to false)
!> \param citations ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_create(section, name, description, n_keywords, &
                             n_subsections, repeats, citations)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(len=*), INTENT(in)                       :: name, description
      INTEGER, INTENT(in), OPTIONAL                      :: n_keywords, n_subsections
      LOGICAL, INTENT(in), OPTIONAL                      :: repeats
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: citations

      CHARACTER(len=*), PARAMETER :: routineN = 'section_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, my_n_keywords, my_n_subsections, n

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input/input_section_types.F",168)
      my_n_keywords = 10
      IF (PRESENT(n_keywords)) my_n_keywords = n_keywords
      my_n_subsections = 0
      IF (PRESENT(n_subsections)) my_n_subsections = n_subsections

      ALLOCATE (section)
      last_section_id = last_section_id+1
      section%id_nr = last_section_id
      section%ref_count = 1

      section%n_keywords = 0
      section%n_subsections = 0
      section%name = name
      CALL uppercase(section%name)

      n = LEN_TRIM(description)
      ALLOCATE (section%description(n))
      DO i = 1, n
         section%description(i) = description(i:i)
      END DO

      section%frozen = .FALSE.
      section%repeats = .FALSE.
      IF (PRESENT(repeats)) section%repeats = repeats

      NULLIFY (section%citations)
      IF (PRESENT(citations)) THEN
         ALLOCATE (section%citations(SIZE(citations)))
         section%citations = citations
      ENDIF

      ALLOCATE (section%keywords(-1:my_n_keywords))
      DO i = -1, my_n_keywords
         NULLIFY (section%keywords(i)%keyword)
      END DO

      ALLOCATE (section%subsections(my_n_subsections))
      DO i = 1, my_n_subsections
         NULLIFY (section%subsections(i)%section)
      END DO
   END SUBROUTINE section_create

! **************************************************************************************************
!> \brief retains the given keyword list (see doc/ReferenceCounting.html)
!> \param section the list to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_retain(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'section_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input/input_section_types.F",221)
      IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",222)
      section%ref_count = section%ref_count+1
   END SUBROUTINE section_retain

! **************************************************************************************************
!> \brief releases the given keyword list (see doc/ReferenceCounting.html)
!> \param section the list to release
!> \author fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_release(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'section_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(section)) THEN
         IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",240)
         section%ref_count = section%ref_count-1
         IF (section%ref_count == 0) THEN
            IF (ASSOCIATED(section%citations)) THEN
               DEALLOCATE (section%citations)
            ENDIF
            IF (ASSOCIATED(section%keywords)) THEN
               DO i = -1, UBOUND(section%keywords, 1)
                  CALL keyword_release(section%keywords(i)%keyword)
               END DO
               DEALLOCATE (section%keywords)
            END IF
            section%n_keywords = 0
            IF (ASSOCIATED(section%subsections)) THEN
               DO i = 1, SIZE(section%subsections)
                  CALL section_release(section%subsections(i)%section)
               END DO
               DEALLOCATE (section%subsections)
            END IF
            DEALLOCATE (section%description)
            DEALLOCATE (section)
         END IF
         NULLIFY (section)
      END IF
   END SUBROUTINE section_release

! **************************************************************************************************
!> \brief collects additional information on the section for IO + documentation
!> \param section ...
!> \retval message ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION get_section_info(section) RESULT(message)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=default_path_length)                 :: message

      CHARACTER(len=*), PARAMETER :: routineN = 'get_section_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: length

      message = " "
      length = LEN_TRIM(a2s(section%description))
      IF (length > 0) THEN
         IF (section%description(length) /= ".") THEN
            message = "."
         ENDIF
      ENDIF
      IF (section%repeats) THEN
         message = TRIM(message)//" This section can be repeated."
      ELSE
         message = TRIM(message)//" This section can not be repeated."
      END IF

   END FUNCTION get_section_info

! **************************************************************************************************
!> \brief prints a description of the given section
!> \param section the section to describe
!> \param unit_nr the unit to write to
!> \param level the level of output: 0: just section name, 1:keywords,
!>        then see keyword_describe :-)
!> \param hide_root if the name of the first section should be hidden
!>        (defaults to false).
!> \param recurse ...
!> \author fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_describe(section, unit_nr, level, hide_root, recurse)
      TYPE(section_type), POINTER                        :: section
      INTEGER, INTENT(in)                                :: unit_nr, level
      LOGICAL, INTENT(in), OPTIONAL                      :: hide_root
      INTEGER, INTENT(in), OPTIONAL                      :: recurse

      CHARACTER(len=*), PARAMETER :: routineN = 'section_describe', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: message
      INTEGER                                            :: ikeyword, isub, my_recurse
      LOGICAL                                            :: my_hide_root

      IF (unit_nr > 0) THEN
         my_hide_root = .FALSE.
         IF (PRESENT(hide_root)) my_hide_root = hide_root
         my_recurse = 0
         IF (PRESENT(recurse)) my_recurse = recurse
         IF (ASSOCIATED(section)) THEN
            IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",326)

            IF (.NOT. my_hide_root) &
               WRITE (unit_nr, "('*** section &',a,' ***')") TRIM(section%name)
            IF (level > 1) THEN
               message = get_section_info(section)
               CALL print_message(TRIM(a2s(section%description))//TRIM(message), unit_nr, 0, 0, 0)
            END IF
            IF (level > 0) THEN
               IF (ASSOCIATED(section%keywords(-1)%keyword)) THEN
                  CALL keyword_describe(section%keywords(-1)%keyword, unit_nr, &
                                        level)
               END IF
               IF (ASSOCIATED(section%keywords(0)%keyword)) THEN
                  CALL keyword_describe(section%keywords(0)%keyword, unit_nr, &
                                        level)
               END IF
               DO ikeyword = 1, section%n_keywords
                  CALL keyword_describe(section%keywords(ikeyword)%keyword, unit_nr, &
                                        level)
               END DO
            END IF
            IF (section%n_subsections > 0 .AND. my_recurse >= 0) THEN
               IF (.NOT. my_hide_root) &
                  WRITE (unit_nr, "('** subsections **')")
               DO isub = 1, section%n_subsections
                  IF (my_recurse > 0) THEN
                     CALL section_describe(section%subsections(isub)%section, unit_nr, &
                                           level, recurse=my_recurse-1)
                  ELSE
                     WRITE (unit_nr, "(' ',a)") section%subsections(isub)%section%name
                  END IF
               END DO
            END IF
            IF (.NOT. my_hide_root) &
               WRITE (unit_nr, "('*** &end section ',a,' ***')") TRIM(section%name)
         ELSE
            WRITE (unit_nr, "(a)") '<section *null*>'
         END IF
      END IF
   END SUBROUTINE section_describe

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \param prefix ...
!> \param depth ...
!> \param unit_number ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_describe_html(section, prefix, depth, unit_number)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: prefix
      INTEGER, INTENT(in)                                :: depth, unit_number

      CHARACTER(len=*), PARAMETER :: routineN = 'section_describe_html', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1000)                                :: local_prefix
      CHARACTER(LEN=256), DIMENSION(50), SAVE            :: location, name
      CHARACTER(LEN=default_path_length)                 :: message
      INTEGER                                            :: idepth, ikeyword, iref, isub, unit_nr
      LOGICAL                                            :: has_keywords

      IF (ASSOCIATED(section)) THEN
         local_prefix = TRIM(prefix//"~"//TRIM(section%name))
         IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",391)
         CALL open_file(unit_number=unit_nr, file_name=TRIM(local_prefix)//".html", &
                        file_status="UNKNOWN", file_action="WRITE")
         WRITE (unit_nr, FMT='(A)') "<HTML><BODY>"
         WRITE (unit_nr, FMT='(A)') "<HEAD><TITLE> The CP2K project : input section "//section%name//"</TITLE></HEAD>"
         WRITE (unit_nr, FMT='(A)') '<META NAME="description" content="CP2K">'
         WRITE (unit_nr, FMT='(A)') &
            '<META NAME="keywords" contents="scientific,computing,chemistry,physics,'// &
            'documentation,help,manual,Fortran,parallel">'
         WRITE (unit_nr, FMT='(A)') "<H1> Section "//section%name//" </H1>"
         WRITE (unit_nr, FMT='(A)') ' <A HREF="index.html">Index of all sections</A>. This section is located at '
         DO idepth = 1, depth
            WRITE (unit_nr, FMT='(A)', ADVANCE="NO") '<A HREF="'//TRIM(location(idepth))//'">'//TRIM(name(idepth))//'</A>%'
         ENDDO
         location(depth+1) = TRIM(local_prefix)//".html"
         name(depth+1) = TRIM(section%name)
         WRITE (unit_nr, FMT='(A)') '<A HREF="'//TRIM(location(depth+1))//'">'//TRIM(name(depth+1))//'</A>.'

         message = get_section_info(section)
         WRITE (unit_nr, FMT='(A)') '<BR><BR>'//TRIM(a2s(section%description))//TRIM(message)

         IF (ASSOCIATED(section%citations)) THEN
            IF (SIZE(section%citations, 1) > 1) THEN
               WRITE (unit_nr, FMT='(A)') '<BR><BR> This section cites following references: '
            ELSE
               WRITE (unit_nr, FMT='(A)') '<BR><BR> This section cites following reference: '
            ENDIF
            DO iref = 1, SIZE(section%citations, 1)
               WRITE (unit_nr, FMT='(A,I0,A)') '<A HREF="references.html#reference_', section%citations(iref), '" TITLE="'
               CALL print_reference(section%citations(iref), FORMAT=print_format_journal, unit=unit_nr)
               WRITE (unit_nr, FMT='(A)') '">['//TRIM(get_citation_key(section%citations(iref)))//']</A>'
            ENDDO
         ENDIF

         WRITE (unit_nr, FMT='(A)') "<H2> Subsections</H2>"
         IF (section%n_subsections > 0) THEN
            WRITE (unit_nr, FMT='(A)') "<UL>"
            DO isub = 1, section%n_subsections
               WRITE (unit_nr, FMT='(A)') &
                  '<LI><A HREF="'//TRIM(local_prefix)//"~"//TRIM(section%subsections(isub)%section%name)//".html"//'">'// &
                  TRIM(section%subsections(isub)%section%name)//' </A>'
            END DO
            WRITE (unit_nr, FMT='(A)') "</UL>"
         ELSE
            WRITE (unit_nr, FMT='(A)') "None"
         ENDIF
         WRITE (unit_nr, FMT='(A)') "<H2> Section keywords </H2>"
         has_keywords = ASSOCIATED(section%keywords(-1)%keyword) .OR. &
                        ASSOCIATED(section%keywords(0)%keyword) .OR. &
                        section%n_keywords >= 1
         IF (has_keywords) THEN
            WRITE (unit_nr, FMT='(A)') "<UL>"
            DO ikeyword = -1, section%n_keywords
               IF (ASSOCIATED(section%keywords(ikeyword)%keyword)) THEN
                  WRITE (unit_nr, FMT='(A)') &
                     '<LI><A HREF="#'//TRIM(section%keywords(ikeyword)%keyword%names(1))//'">'// &
                     TRIM(section%keywords(ikeyword)%keyword%names(1))//" </A>"
               END IF
            END DO
            WRITE (unit_nr, FMT='(A)') "</UL>"
            WRITE (unit_nr, FMT='(A)') "<H2> Keyword descriptions </H2>"
            WRITE (unit_nr, FMT='(A)') "<TABLE>"
            DO ikeyword = -1, section%n_keywords
               IF (ASSOCIATED(section%keywords(ikeyword)%keyword)) THEN
                  CALL keyword_describe_html(section%keywords(ikeyword)%keyword, unit_nr)
               END IF
            END DO
            WRITE (unit_nr, FMT='(A)') "</TABLE>"
         ELSE
            WRITE (unit_nr, FMT='(A)') "None"
         ENDIF
         WRITE (unit_nr, FMT='(A)') &
            '<BR><hr>Back to the <A HREF="https://www.cp2k.org/">CP2K homepage</A> or '// &
            'the latest version of <A HREF="https://manual.cp2k.org/trunk/">this manual</A>'
         WRITE (unit_nr, FMT='(A)') "</BODY></HTML>"
         CALL close_file(unit_nr)
         DO isub = 1, section%n_subsections
            CALL section_describe_html(section%subsections(isub)%section, TRIM(local_prefix), depth+1, unit_number)
         END DO
      END IF
   END SUBROUTINE section_describe_html

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \param prefix ...
!> \param unit_nr ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_describe_index_html(section, prefix, unit_nr)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: prefix
      INTEGER, INTENT(in)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'section_describe_index_html', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1000)                                :: local_prefix
      INTEGER                                            :: isub

      IF (ASSOCIATED(section)) THEN
         IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",491)
         local_prefix = TRIM(prefix//"~"//TRIM(section%name))
         WRITE (unit_nr, FMT='(A)') &
            '<LI><A HREF="'//TRIM(local_prefix)//'.html">'//TRIM(section%name)//"</A>"
         IF (section%n_subsections > 0) THEN
            WRITE (unit_nr, FMT='(A)') "<UL>"
            DO isub = 1, section%n_subsections
               CALL section_describe_index_html(section%subsections(isub)%section, TRIM(local_prefix), unit_nr)
            END DO
            WRITE (unit_nr, FMT='(A)') "</UL>"
         ENDIF
      ENDIF
   END SUBROUTINE section_describe_index_html

! **************************************************************************************************
!> \brief returns the index of requested subsection (-1 if not found)
!> \param section the root section
!> \param subsection_name the name of the subsection you want to get
!> \retval res ...
!> \author fawzi
!> \note
!>      private utility function
! **************************************************************************************************
   FUNCTION section_get_subsection_index(section, subsection_name) RESULT(res)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(len=*), INTENT(in)                       :: subsection_name
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'section_get_subsection_index', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: upc_name
      INTEGER                                            :: isub

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input/input_section_types.F",525)
      IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",526)
      res = -1
      upc_name = subsection_name
      CALL uppercase(upc_name)
      DO isub = 1, section%n_subsections
         IF(.NOT.(ASSOCIATED(section%subsections(isub)%section)))CALL cp__a("input/input_section_types.F",531)
         IF (section%subsections(isub)%section%name == upc_name) THEN
            res = isub
            EXIT
         END IF
      END DO
   END FUNCTION section_get_subsection_index

! **************************************************************************************************
!> \brief returns the requested subsection
!> \param section the root section
!> \param subsection_name the name of the subsection you want to get
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION section_get_subsection(section, subsection_name) RESULT(res)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(len=*), INTENT(in)                       :: subsection_name
      TYPE(section_type), POINTER                        :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'section_get_subsection', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: isub

      isub = section_get_subsection_index(section, subsection_name)
      IF (isub > 0) THEN
         res => section%subsections(isub)%section
      ELSE
         NULLIFY (res)
      END IF
   END FUNCTION section_get_subsection

! **************************************************************************************************
!> \brief returns the index of the requested keyword (or -2 if not found)
!> \param section the section the keyword is in
!> \param keyword_name the keyword you are interested in
!> \retval res ...
!> \author fawzi
!> \note
!>      private utility function
! **************************************************************************************************
   FUNCTION section_get_keyword_index(section, keyword_name) RESULT(res)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'section_get_keyword_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, in
      CHARACTER(len=default_string_length)               :: upc_name

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input/input_section_types.F",584)
      IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",585)
      IF(.NOT.(ASSOCIATED(section%keywords)))CALL cp__a("input/input_section_types.F",586)
      res = -2
      upc_name = keyword_name
      CALL uppercase(upc_name)
      DO ik = -1, 0
         IF (ASSOCIATED(section%keywords(ik)%keyword)) THEN
            IF (section%keywords(ik)%keyword%names(1) == upc_name) THEN
               res = ik
            END IF
         END IF
      END DO
      IF (res == -2) THEN
         k_search_loop: DO ik = 1, section%n_keywords
            IF(.NOT.(ASSOCIATED(section%keywords(ik)%keyword)))CALL cp__a("input/input_section_types.F",599)
            DO in = 1, SIZE(section%keywords(ik)%keyword%names)
               IF (section%keywords(ik)%keyword%names(in) == upc_name) THEN
                  res = ik
                  EXIT k_search_loop
               END IF
            END DO
         END DO k_search_loop
      END IF
   END FUNCTION section_get_keyword_index

! **************************************************************************************************
!> \brief returns the requested keyword
!> \param section the section the keyword is in
!> \param keyword_name the keyword you are interested in
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   RECURSIVE FUNCTION section_get_keyword(section, keyword_name) RESULT(res)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      TYPE(keyword_type), POINTER                        :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'section_get_keyword', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, my_index

      IF (INDEX(keyword_name, "%") /= 0) THEN
         my_index = INDEX(keyword_name, "%")+1
         IF(.NOT.(ASSOCIATED(section%subsections)))CALL cp__a("input/input_section_types.F",629)
         DO ik = LBOUND(section%subsections, 1), UBOUND(section%subsections, 1)
            IF (section%subsections(ik)%section%name == keyword_name(1:my_index-2)) EXIT
         END DO
         IF(.NOT.(ik <= UBOUND(section%subsections, 1)))CALL cp__a("input/input_section_types.F",633)
         res => section_get_keyword(section%subsections(ik)%section, keyword_name(my_index:))
      ELSE
         ik = section_get_keyword_index(section, keyword_name)
         IF (ik == -2) THEN
            NULLIFY (res)
         ELSE
            res => section%keywords(ik)%keyword
         END IF
      END IF
   END FUNCTION section_get_keyword

! **************************************************************************************************
!> \brief adds a keyword to the given section
!> \param section the section to which the keyword should be added
!> \param keyword the keyword to add
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_add_keyword(section, keyword)
      TYPE(section_type), POINTER                        :: section
      TYPE(keyword_type), POINTER                        :: keyword

      CHARACTER(len=*), PARAMETER :: routineN = 'section_add_keyword', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k
      TYPE(keyword_p_type), DIMENSION(:), POINTER        :: new_keywords

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input/input_section_types.F",661)
      IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",662)
      IF(.NOT.(.NOT. section%frozen))CALL cp__a("input/input_section_types.F",663)
      IF(.NOT.(ASSOCIATED(keyword)))CALL cp__a("input/input_section_types.F",664)
      IF(.NOT.(keyword%ref_count > 0))CALL cp__a("input/input_section_types.F",665)
      CALL keyword_retain(keyword)
      IF (keyword%names(1) == "_SECTION_PARAMETERS_") THEN
         CALL keyword_release(section%keywords(-1)%keyword)
         section%keywords(-1)%keyword => keyword
      ELSE IF (keyword%names(1) == "_DEFAULT_KEYWORD_") THEN
         CALL keyword_release(section%keywords(0)%keyword)
         section%keywords(0)%keyword => keyword
      ELSE
         DO k = 1, SIZE(keyword%names)
            DO i = 1, section%n_keywords
               DO j = 1, SIZE(section%keywords(i)%keyword%names)
                  IF (keyword%names(k) == section%keywords(i)%keyword%names(j)) THEN
                     CALL cp_abort(cp__l("input/input_section_types.F",678), &
                                   "trying to add a keyword with a name ("// &
                                   TRIM(keyword%names(k))//") that was already used in section " &
                                   //TRIM(section%name))
                  ENDIF
               END DO
            END DO
         END DO

         IF (UBOUND(section%keywords, 1) == section%n_keywords) THEN
            ALLOCATE (new_keywords(-1:section%n_keywords+10))
            DO i = -1, section%n_keywords
               new_keywords(i)%keyword => section%keywords(i)%keyword
            END DO
            DO i = section%n_keywords+1, UBOUND(new_keywords, 1)
               NULLIFY (new_keywords(i)%keyword)
            END DO
            DEALLOCATE (section%keywords)
            section%keywords => new_keywords
         END IF
         section%n_keywords = section%n_keywords+1
         section%keywords(section%n_keywords)%keyword => keyword
      END IF
   END SUBROUTINE section_add_keyword

! **************************************************************************************************
!> \brief adds a subsection to the given section
!> \param section to section to which you want to add a subsection
!> \param subsection the subsection to add
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_add_subsection(section, subsection)
      TYPE(section_type), POINTER                        :: section, subsection

      CHARACTER(len=*), PARAMETER :: routineN = 'section_add_subsection', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      TYPE(section_p_type), DIMENSION(:), POINTER        :: new_subsections

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input/input_section_types.F",718)
      IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",719)
      IF(.NOT.(ASSOCIATED(subsection)))CALL cp__a("input/input_section_types.F",720)
      IF(.NOT.(subsection%ref_count > 0))CALL cp__a("input/input_section_types.F",721)
      IF (SIZE(section%subsections) < section%n_subsections+1) THEN
         ALLOCATE (new_subsections(section%n_subsections+10))
         DO i = 1, section%n_subsections
            new_subsections(i)%section => section%subsections(i)%section
         END DO
         DO i = section%n_subsections+1, SIZE(new_subsections)
            NULLIFY (new_subsections(i)%section)
         END DO
         DEALLOCATE (section%subsections)
         section%subsections => new_subsections
      END IF
      DO i = 1, section%n_subsections
         IF (subsection%name == section%subsections(i)%section%name) &
            CALL cp_abort(cp__l("input/input_section_types.F",735), &
                          "trying to add a subsection with a name ("// &
                          TRIM(subsection%name)//") that was already used in section " &
                          //TRIM(section%name))
      END DO
      CALL section_retain(subsection)
      section%n_subsections = section%n_subsections+1
      section%subsections(section%n_subsections)%section => subsection
   END SUBROUTINE section_add_subsection

! **************************************************************************************************
!> \brief creates a object where to store the values of a section
!> \param section_vals the parsed section that will be created
!> \param section the structure of the section that you want to parse
!> \author fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_vals_create(section_vals, section)
      TYPE(section_vals_type), POINTER                   :: section_vals
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(.NOT. ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",760)
      ALLOCATE (section_vals)
      last_section_vals_id = last_section_vals_id+1
      section_vals%id_nr = last_section_vals_id
      section_vals%ref_count = 1
      CALL section_retain(section)
      section_vals%section => section
      section%frozen = .TRUE.
      ALLOCATE (section_vals%values(-1:section%n_keywords, 0))
      ALLOCATE (section_vals%subs_vals(section%n_subsections, 1))
      DO i = 1, section%n_subsections
         NULLIFY (section_vals%subs_vals(i, 1)%section_vals)
         CALL section_vals_create(section_vals%subs_vals(i, 1)%section_vals, &
                                  section=section%subsections(i)%section)
      END DO
      NULLIFY (section_vals%ibackup)
   END SUBROUTINE section_vals_create

! **************************************************************************************************
!> \brief retains the given section values (see doc/ReferenceCounting.html)
!> \param section_vals the object to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_vals_retain(section_vals)
      TYPE(section_vals_type), POINTER                   :: section_vals

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",789)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",790)
      section_vals%ref_count = section_vals%ref_count+1
   END SUBROUTINE section_vals_retain

! **************************************************************************************************
!> \brief releases the given object
!> \param section_vals the section_vals to release
!> \author fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_vals_release(section_vals)
      TYPE(section_vals_type), POINTER                   :: section_vals

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      TYPE(cp_sll_val_type), POINTER                     :: vals
      TYPE(val_type), POINTER                            :: el

      IF (ASSOCIATED(section_vals)) THEN
         IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",810)
         section_vals%ref_count = section_vals%ref_count-1
         IF (section_vals%ref_count == 0) THEN
            CALL section_release(section_vals%section)
            DO j = 1, SIZE(section_vals%values, 2)
               DO i = -1, UBOUND(section_vals%values, 1)
                  vals => section_vals%values(i, j)%list
                  DO WHILE (cp_sll_val_next(vals, el_att=el))
                     CALL val_release(el)
                  END DO
                  CALL cp_sll_val_dealloc(section_vals%values(i, j)%list)
               END DO
            END DO
            DEALLOCATE (section_vals%values)
            DO j = 1, SIZE(section_vals%subs_vals, 2)
               DO i = 1, SIZE(section_vals%subs_vals, 1)
                  CALL section_vals_release(section_vals%subs_vals(i, j)%section_vals)
               END DO
            END DO
            DEALLOCATE (section_vals%subs_vals)
            IF (ASSOCIATED(section_vals%ibackup)) THEN
               DEALLOCATE (section_vals%ibackup)
            END IF
            DEALLOCATE (section_vals)
         END IF
      END IF
   END SUBROUTINE section_vals_release

! **************************************************************************************************
!> \brief returns various attributes about the section_vals
!> \param section_vals the section vals you want information from
!> \param ref_count ...
!> \param id_nr ...
!> \param n_repetition number of repetitions of the section
!> \param n_subs_vals_rep number of repetitions of the subsections values
!>        (max(1,n_repetition))
!> \param section ...
!> \param explicit if the section was explicitly present in
!> \author fawzi
!> \note     For the other arguments see the attributes of section_vals_type
! **************************************************************************************************
   SUBROUTINE section_vals_get(section_vals, ref_count, id_nr, n_repetition, &
                               n_subs_vals_rep, section, explicit)
      TYPE(section_vals_type), POINTER                   :: section_vals
      INTEGER, INTENT(out), OPTIONAL                     :: ref_count, id_nr, n_repetition, &
                                                            n_subs_vals_rep
      TYPE(section_type), OPTIONAL, POINTER              :: section
      LOGICAL, INTENT(out), OPTIONAL                     :: explicit

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_get', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",862)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",863)
      IF (PRESENT(ref_count)) ref_count = section_vals%ref_count
      IF (PRESENT(id_nr)) id_nr = section_vals%id_nr
      IF (PRESENT(section)) section => section_vals%section
      IF (PRESENT(n_repetition)) n_repetition = SIZE(section_vals%values, 2)
      IF (PRESENT(n_subs_vals_rep)) n_subs_vals_rep = SIZE(section_vals%subs_vals, 2)
      IF (PRESENT(explicit)) explicit = (SIZE(section_vals%values, 2) > 0)
   END SUBROUTINE section_vals_get

! **************************************************************************************************
!> \brief returns the values of the requested subsection
!> \param section_vals the root section
!> \param subsection_name the name of the requested subsection
!> \param i_rep_section index of the repetition of section_vals from which
!>        you want to extract the subsection (defaults to 1)
!> \param can_return_null if the results can be null (defaults to false)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   RECURSIVE FUNCTION section_vals_get_subs_vals(section_vals, subsection_name, &
                                                 i_rep_section, can_return_null) RESULT(res)
      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: subsection_name
      INTEGER, INTENT(in), OPTIONAL                      :: i_rep_section
      LOGICAL, INTENT(in), OPTIONAL                      :: can_return_null
      TYPE(section_vals_type), POINTER                   :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_get_subs_vals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: irep, isection, my_index
      LOGICAL                                            :: is_path, my_can_return_null

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",896)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",897)

      my_can_return_null = .FALSE.
      IF (PRESENT(can_return_null)) my_can_return_null = can_return_null
      NULLIFY (res)
      irep = 1
      IF (PRESENT(i_rep_section)) irep = i_rep_section

      ! prepare for recursive parsing of subsections. i_rep_section will be used for last section
      my_index = INDEX(subsection_name, "%")
      IF (my_index .EQ. 0) THEN
         is_path = .FALSE.
         my_index = LEN_TRIM(subsection_name)
      ELSE
         is_path = .TRUE.
         irep = 1
         my_index = my_index-1
      ENDIF

      IF(.NOT.(irep <= SIZE(section_vals%subs_vals, 2)))CALL cp__a("input/input_section_types.F",916)

      isection = section_get_subsection_index(section_vals%section, subsection_name(1:my_index))
      IF (isection > 0) res => section_vals%subs_vals(isection, irep)%section_vals
      IF (.NOT. (ASSOCIATED(res) .OR. my_can_return_null)) &
         CALL cp_abort(cp__l("input/input_section_types.F",921), &
                       "could not find subsection "//TRIM(subsection_name(1:my_index))//" in section "// &
                       TRIM(section_vals%section%name)//" at ")
      IF (is_path .AND. ASSOCIATED(res)) THEN
         res => section_vals_get_subs_vals(res, subsection_name(my_index+2:LEN_TRIM(subsection_name)), &
                                           i_rep_section, can_return_null)
      ENDIF

   END FUNCTION section_vals_get_subs_vals

! **************************************************************************************************
!> \brief returns the values of the n-th non default subsection (null if no
!>      such section exists (not so many non default section))
!> \param section_vals the root section
!> \param i_section index of the section
!> \param i_rep_section index of the repetition of section_vals from which
!>        you want to extract the subsection (defaults to 1)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION section_vals_get_subs_vals2(section_vals, i_section, i_rep_section) RESULT(res)
      TYPE(section_vals_type), POINTER                   :: section_vals
      INTEGER, INTENT(in)                                :: i_section
      INTEGER, INTENT(in), OPTIONAL                      :: i_rep_section
      TYPE(section_vals_type), POINTER                   :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_get_subs_vals2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, irep, isect_att

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",952)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",953)
      NULLIFY (res)
      irep = 1
      IF (PRESENT(i_rep_section)) irep = i_rep_section
      IF(.NOT.(irep <= SIZE(section_vals%subs_vals, 2)))CALL cp__a("input/input_section_types.F",957)
      isect_att = 0
      DO i = 1, section_vals%section%n_subsections
         IF (SIZE(section_vals%subs_vals(i, irep)%section_vals%values, 2) > 0) THEN
            isect_att = isect_att+1
            IF (isect_att == i_section) THEN
               res => section_vals%subs_vals(i, irep)%section_vals
               EXIT
            END IF
         END IF
      END DO
   END FUNCTION section_vals_get_subs_vals2

! **************************************************************************************************
!> \brief returns the values of the n-th non default subsection (null if no
!>      such section exists (not so many non default section))
!> \param section_vals the root section
!> \param subsection_name ...
!> \param i_rep_section index of the repetition of section_vals from which
!>        you want to extract the subsection (defaults to 1)
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION section_vals_get_subs_vals3(section_vals, subsection_name, &
                                        i_rep_section) RESULT(res)
      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(LEN=*), INTENT(IN)                       :: subsection_name
      INTEGER, INTENT(in), OPTIONAL                      :: i_rep_section
      TYPE(section_vals_type), POINTER                   :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_get_subs_vals3', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_section, irep

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",992)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",993)
      NULLIFY (res)
      irep = 1
      IF (PRESENT(i_rep_section)) irep = i_rep_section
      IF(.NOT.(irep <= SIZE(section_vals%subs_vals, 2)))CALL cp__a("input/input_section_types.F",997)
      i_section = section_get_subsection_index(section_vals%section, subsection_name)
      res => section_vals%subs_vals(i_section, irep)%section_vals
   END FUNCTION section_vals_get_subs_vals3

! **************************************************************************************************
!> \brief adds the place to store the values of a repetition of the section
!> \param section_vals the section you want to extend
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_vals_add_values(section_vals)
      TYPE(section_vals_type), POINTER                   :: section_vals

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_add_values', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      TYPE(cp_sll_val_p_type), DIMENSION(:, :), POINTER  :: new_values
      TYPE(section_vals_p_type), DIMENSION(:, :), &
         POINTER                                         :: new_sps

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",1018)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1019)
      ALLOCATE (new_values(-1:UBOUND(section_vals%values, 1), SIZE(section_vals%values, 2)+1))
      DO j = 1, SIZE(section_vals%values, 2)
         DO i = -1, UBOUND(section_vals%values, 1)
            new_values(i, j)%list => section_vals%values(i, j)%list
         END DO
      END DO
      DEALLOCATE (section_vals%values)
      section_vals%values => new_values
      j = SIZE(new_values, 2)
      DO i = -1, UBOUND(new_values, 1)
         NULLIFY (new_values(i, j)%list)
      END DO

      IF (SIZE(new_values, 2) > 1) THEN
         ALLOCATE (new_sps(SIZE(section_vals%subs_vals, 1), &
                           SIZE(section_vals%subs_vals, 2)+1))
         DO j = 1, SIZE(section_vals%subs_vals, 2)
            DO i = 1, SIZE(section_vals%subs_vals, 1)
               new_sps(i, j)%section_vals => section_vals%subs_vals(i, j)%section_vals
            END DO
         END DO
         DEALLOCATE (section_vals%subs_vals)
         section_vals%subs_vals => new_sps
         j = SIZE(new_sps, 2)
         DO i = 1, SIZE(new_sps, 1)
            NULLIFY (new_sps(i, j)%section_vals)
            CALL section_vals_create(new_sps(i, SIZE(new_sps, 2))%section_vals, &
                                     section=section_vals%section%subsections(i)%section)
         END DO
      END IF
   END SUBROUTINE section_vals_add_values

! **************************************************************************************************
!> \brief removes the values of a repetition of the section
!> \param section_vals the section you want to extend
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_vals_remove_values(section_vals)
      TYPE(section_vals_type), POINTER                   :: section_vals

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_remove_values', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      TYPE(cp_sll_val_p_type), DIMENSION(:, :), POINTER  :: new_values
      TYPE(cp_sll_val_type), POINTER                     :: vals
      TYPE(val_type), POINTER                            :: el

      IF (ASSOCIATED(section_vals)) THEN
         IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1069)
         NULLIFY (el, vals)
         ! Allocate a null 0 dimension array of values
         ALLOCATE (new_values(-1:section_vals%section%n_keywords, 0))
         ! Release old values
         DO j = 1, SIZE(section_vals%values, 2)
            DO i = -1, UBOUND(section_vals%values, 1)
               vals => section_vals%values(i, j)%list
               DO WHILE (cp_sll_val_next(vals, el_att=el))
                  CALL val_release(el)
               END DO
               CALL cp_sll_val_dealloc(section_vals%values(i, j)%list)
            END DO
         END DO
         DEALLOCATE (section_vals%values)
         section_vals%values => new_values
      END IF
   END SUBROUTINE section_vals_remove_values

! These accessor functions can be used instead of passing a variable
! in the parameter list of a subroutine call. This should make the
! code a lot simpler. See xc_rho_set_and_dset_create in xc.F as
! an example.

! **************************************************************************************************
!> \brief ...
!> \param section_vals ...
!> \param keyword_name ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION section_get_cval(section_vals, keyword_name) RESULT(res)

      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      CHARACTER(LEN=default_string_length)               :: res

      CALL section_vals_val_get(section_vals, keyword_name, c_val=res)

   END FUNCTION section_get_cval

! **************************************************************************************************
!> \brief ...
!> \param section_vals ...
!> \param keyword_name ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION section_get_rval(section_vals, keyword_name) RESULT(res)

      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      REAL(kind=dp)                                      :: res

      CALL section_vals_val_get(section_vals, keyword_name, r_val=res)

   END FUNCTION section_get_rval

! **************************************************************************************************
!> \brief ...
!> \param section_vals ...
!> \param keyword_name ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION section_get_rvals(section_vals, keyword_name) RESULT(res)

      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      REAL(kind=dp), DIMENSION(:), POINTER               :: res

      CALL section_vals_val_get(section_vals, keyword_name, r_vals=res)

   END FUNCTION section_get_rvals

! **************************************************************************************************
!> \brief ...
!> \param section_vals ...
!> \param keyword_name ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION section_get_ival(section_vals, keyword_name) RESULT(res)

      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      INTEGER                                            :: res

      CALL section_vals_val_get(section_vals, keyword_name, i_val=res)

   END FUNCTION section_get_ival

! **************************************************************************************************
!> \brief ...
!> \param section_vals ...
!> \param keyword_name ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION section_get_ivals(section_vals, keyword_name) RESULT(res)

      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      INTEGER, DIMENSION(:), POINTER                     :: res

      CALL section_vals_val_get(section_vals, keyword_name, i_vals=res)

   END FUNCTION section_get_ivals

! **************************************************************************************************
!> \brief ...
!> \param section_vals ...
!> \param keyword_name ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION section_get_lval(section_vals, keyword_name) RESULT(res)

      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      LOGICAL                                            :: res

      CALL section_vals_val_get(section_vals, keyword_name, l_val=res)

   END FUNCTION section_get_lval

! **************************************************************************************************
!> \brief returns the requested value
!> \param section_vals ...
!> \param keyword_name the name of the keyword you want
!> \param i_rep_section which repetition of the section you are interested in
!>        (defaults to 1)
!> \param i_rep_val which repetition of the keyword/val you are interested in
!>        (defaults to 1)
!> \param n_rep_val returns number of val available
!> \param val ...
!> \param l_val ,i_val,r_val,c_val: returns the logical,integer,real or
!>        character value
!> \param i_val ...
!> \param r_val ...
!> \param c_val ...
!> \param l_vals ,i_vals,r_vals,c_vals: returns the logical,integer,real or
!>        character arrays. The val reamins the owner of the array
!> \param i_vals ...
!> \param r_vals ...
!> \param c_vals ...
!> \param explicit ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_vals_val_get(section_vals, keyword_name, i_rep_section, &
                                   i_rep_val, n_rep_val, val, l_val, i_val, r_val, c_val, l_vals, i_vals, r_vals, &
                                   c_vals, explicit)
      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      INTEGER, INTENT(in), OPTIONAL                      :: i_rep_section, i_rep_val
      INTEGER, INTENT(out), OPTIONAL                     :: n_rep_val
      TYPE(val_type), OPTIONAL, POINTER                  :: val
      LOGICAL, INTENT(out), OPTIONAL                     :: l_val
      INTEGER, INTENT(out), OPTIONAL                     :: i_val
      REAL(KIND=DP), INTENT(out), OPTIONAL               :: r_val
      CHARACTER(LEN=*), INTENT(out), OPTIONAL            :: c_val
      LOGICAL, DIMENSION(:), OPTIONAL, POINTER           :: l_vals
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: i_vals
      REAL(KIND=DP), DIMENSION(:), OPTIONAL, POINTER     :: r_vals
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: c_vals
      LOGICAL, INTENT(out), OPTIONAL                     :: explicit

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_val_get', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, irs, len_key, my_index, &
                                                            tmp_index
      LOGICAL                                            :: valRequested
      TYPE(cp_sll_val_type), POINTER                     :: vals
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: s_vals
      TYPE(val_type), POINTER                            :: my_val

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",1243)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1244)

      my_index = INDEX(keyword_name, '%')+1
      len_key = LEN_TRIM(keyword_name)
      IF (my_index > 1) THEN
         DO
            tmp_index = INDEX(keyword_name(my_index:len_key), "%")
            IF (tmp_index <= 0) EXIT
            my_index = my_index+tmp_index
         END DO
         s_vals => section_vals_get_subs_vals(section_vals, keyword_name(1:my_index-2))
      ELSE
         s_vals => section_vals
      END IF

      irk = 1
      irs = 1
      IF (PRESENT(i_rep_section)) irs = i_rep_section
      IF (PRESENT(i_rep_val)) irk = i_rep_val
      IF (PRESENT(val)) NULLIFY (val)
      IF (PRESENT(explicit)) explicit = .FALSE.
      section => s_vals%section
      valRequested = PRESENT(l_val) .OR. PRESENT(i_val) .OR. PRESENT(r_val) .OR. &
                     PRESENT(c_val) .OR. PRESENT(l_vals) .OR. PRESENT(i_vals) .OR. &
                     PRESENT(r_vals) .OR. PRESENT(c_vals)
      ik = section_get_keyword_index(s_vals%section, keyword_name(my_index:len_key))
      IF (ik == -2) &
         CALL cp_abort(cp__l("input/input_section_types.F",1271), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       TRIM(keyword_name(my_index:len_key)))
      keyword => section%keywords(ik)%keyword
      IF (.NOT. (irs > 0 .AND. irs <= SIZE(s_vals%subs_vals, 2))) &
         CALL cp_abort(cp__l("input/input_section_types.F",1276), &
                       "section repetition requested ("//cp_to_string(irs)// &
                       ") out of bounds (1:"//cp_to_string(SIZE(s_vals%subs_vals, 2)) &
                       //")")
      NULLIFY (my_val)
      IF (PRESENT(n_rep_val)) n_rep_val = 0
      IF (irs <= SIZE(s_vals%values, 2)) THEN ! the section was parsed
         vals => s_vals%values(ik, irs)%list
         IF (PRESENT(n_rep_val)) n_rep_val = cp_sll_val_get_length(vals)
         IF (.NOT. ASSOCIATED(vals)) THEN
            ! this keyword was not parsed
            IF (ASSOCIATED(keyword%default_value)) THEN
               my_val => keyword%default_value
               IF (PRESENT(n_rep_val)) n_rep_val = 1
            END IF
         ELSE
            my_val => cp_sll_val_get_el_at(s_vals%values(ik, irs)%list, &
                                           irk)
            IF (PRESENT(explicit)) explicit = .TRUE.
         END IF
      ELSE IF (ASSOCIATED(keyword%default_value)) THEN
         IF (PRESENT(n_rep_val)) n_rep_val = 1
         my_val => keyword%default_value
      END IF
      IF (PRESENT(val)) val => my_val
      IF (valRequested) THEN
         IF (.NOT. ASSOCIATED(my_val)) &
            CALL cp_abort(cp__l("input/input_section_types.F",1303), &
                          "Value requested, but no value set getting value from "// &
                          "keyword "//TRIM(keyword_name(my_index:len_key))//" of section "// &
                          TRIM(section%name))
         CALL val_get(my_val, l_val=l_val, i_val=i_val, r_val=r_val, &
                      c_val=c_val, l_vals=l_vals, i_vals=i_vals, r_vals=r_vals, &
                      c_vals=c_vals)
      END IF

   END SUBROUTINE section_vals_val_get

! **************************************************************************************************
!> \brief returns the requested list
!> \param section_vals ...
!> \param keyword_name the name of the keyword you want
!> \param i_rep_section which repetition of the section you are interested in
!>        (defaults to 1)
!> \param list ...
!> \author Joost VandeVondele
!> \note
!>      - most useful if the full list is needed anyway, so that faster iteration can be used
! **************************************************************************************************
   SUBROUTINE section_vals_list_get(section_vals, keyword_name, i_rep_section, &
                                    list)
      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      INTEGER, OPTIONAL                                  :: i_rep_section
      TYPE(cp_sll_val_type), POINTER                     :: list

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_list_get', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irs, len_key, my_index, tmp_index
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: s_vals

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",1339)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1340)
      NULLIFY (list)
      my_index = INDEX(keyword_name, '%')+1
      len_key = LEN_TRIM(keyword_name)
      IF (my_index > 1) THEN
         DO
            tmp_index = INDEX(keyword_name(my_index:len_key), "%")
            IF (tmp_index <= 0) EXIT
            my_index = my_index+tmp_index
         END DO
         s_vals => section_vals_get_subs_vals(section_vals, keyword_name(1:my_index-2))
      ELSE
         s_vals => section_vals
      END IF

      irs = 1
      IF (PRESENT(i_rep_section)) irs = i_rep_section
      section => s_vals%section
      ik = section_get_keyword_index(s_vals%section, keyword_name(my_index:len_key))
      IF (ik == -2) &
         CALL cp_abort(cp__l("input/input_section_types.F",1360), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       TRIM(keyword_name(my_index:len_key)))
      IF (.NOT. (irs > 0 .AND. irs <= SIZE(s_vals%subs_vals, 2))) &
         CALL cp_abort(cp__l("input/input_section_types.F",1364), &
                       "section repetition requested ("//cp_to_string(irs)// &
                       ") out of bounds (1:"//cp_to_string(SIZE(s_vals%subs_vals, 2)) &
                       //")")
      list => s_vals%values(ik, irs)%list

   END SUBROUTINE section_vals_list_get

! **************************************************************************************************
!> \brief sets the requested value
!> \param section_vals ...
!> \param keyword_name the name of the keyword you want (can be a path
!>        separated by '%')
!> \param i_rep_section isection which repetition of the section you are
!>        nterested in (defaults to 1)
!> \param i_rep_val which repetition of the keyword/val you are interested in
!>        (defaults to 1)
!> \param val ...
!> \param l_val ,i_val,r_val,c_val: sets the logical,integer,real or
!>        character value
!> \param i_val ...
!> \param r_val ...
!> \param c_val ...
!> \param l_vals_ptr ,i_vals_ptr,r_vals,c_vals: sets the logical,integer,real or
!>        character arrays. The val becomes the owner of the array
!> \param i_vals_ptr ...
!> \param r_vals_ptr ...
!> \param c_vals_ptr ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_vals_val_set(section_vals, keyword_name, i_rep_section, i_rep_val, &
                                   val, l_val, i_val, r_val, c_val, l_vals_ptr, i_vals_ptr, r_vals_ptr, c_vals_ptr)
      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      INTEGER, INTENT(in), OPTIONAL                      :: i_rep_section, i_rep_val
      TYPE(val_type), OPTIONAL, POINTER                  :: val
      LOGICAL, INTENT(in), OPTIONAL                      :: l_val
      INTEGER, INTENT(in), OPTIONAL                      :: i_val
      REAL(KIND=DP), INTENT(in), OPTIONAL                :: r_val
      CHARACTER(LEN=*), INTENT(in), OPTIONAL             :: c_val
      LOGICAL, DIMENSION(:), OPTIONAL, POINTER           :: l_vals_ptr
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: i_vals_ptr
      REAL(KIND=DP), DIMENSION(:), OPTIONAL, POINTER     :: r_vals_ptr
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), OPTIONAL, POINTER                 :: c_vals_ptr

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_val_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, irs, len_key, my_index, &
                                                            tmp_index
      LOGICAL                                            :: valSet
      TYPE(cp_sll_val_type), POINTER                     :: vals
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: s_vals
      TYPE(val_type), POINTER                            :: my_val, old_val

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",1422)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1423)

      my_index = INDEX(keyword_name, '%')+1
      len_key = LEN_TRIM(keyword_name)
      IF (my_index > 1) THEN
         DO
            tmp_index = INDEX(keyword_name(my_index:len_key), "%")
            IF (tmp_index <= 0) EXIT
            my_index = my_index+tmp_index
         END DO
         s_vals => section_vals_get_subs_vals(section_vals, keyword_name(1:my_index-2))
      ELSE
         s_vals => section_vals
      END IF

      irk = 1
      irs = 1
      IF (PRESENT(i_rep_section)) irs = i_rep_section
      IF (PRESENT(i_rep_val)) irk = i_rep_val
      section => s_vals%section
      ik = section_get_keyword_index(s_vals%section, keyword_name(my_index:len_key))
      IF (ik == -2) &
         CALL cp_abort(cp__l("input/input_section_types.F",1445), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       TRIM(keyword_name(my_index:len_key)))
      ! Add values..
      DO
         IF (irs <= SIZE(s_vals%values, 2)) EXIT
         CALL section_vals_add_values(s_vals)
      END DO
      IF (.NOT. (irs > 0 .AND. irs <= SIZE(s_vals%subs_vals, 2))) &
         CALL cp_abort(cp__l("input/input_section_types.F",1454), &
                       "section repetition requested ("//cp_to_string(irs)// &
                       ") out of bounds (1:"//cp_to_string(SIZE(s_vals%subs_vals, 2)) &
                       //")")
      keyword => s_vals%section%keywords(ik)%keyword
      NULLIFY (my_val)
      IF (PRESENT(val)) my_val => val
      valSet = PRESENT(l_val) .OR. PRESENT(i_val) .OR. PRESENT(r_val) .OR. &
               PRESENT(c_val) .OR. PRESENT(l_vals_ptr) .OR. PRESENT(i_vals_ptr) .OR. &
               PRESENT(r_vals_ptr) .OR. PRESENT(c_vals_ptr)
      IF (ASSOCIATED(my_val)) THEN
         ! check better?
         IF (valSet) &
            CALL cp_abort(cp__l("input/input_section_types.F",1467), &
                          " both val and values present, in setting "// &
                          "keyword "//TRIM(keyword_name(my_index:len_key))//" of section "// &
                          TRIM(section%name))
      ELSE
         ! ignore ?
         IF (.NOT. valSet) &
            CALL cp_abort(cp__l("input/input_section_types.F",1474), &
                          " empty value in setting "// &
                          "keyword "//TRIM(keyword_name(my_index:len_key))//" of section "// &
                          TRIM(section%name))
         IF(.NOT.(valSet))CALL cp__a("input/input_section_types.F",1478)
         IF (keyword%type_of_var == lchar_t) THEN
            CALL val_create(my_val, lc_val=c_val, lc_vals_ptr=c_vals_ptr)
         ELSE
            CALL val_create(my_val, l_val=l_val, i_val=i_val, r_val=r_val, &
                            c_val=c_val, l_vals_ptr=l_vals_ptr, i_vals_ptr=i_vals_ptr, &
                            r_vals_ptr=r_vals_ptr, &
                            c_vals_ptr=c_vals_ptr, enum=keyword%enum)
         END IF
         IF(.NOT.(ASSOCIATED(my_val)))CALL cp__a("input/input_section_types.F",1487)
         IF(.NOT.(my_val%type_of_var == keyword%type_of_var))CALL cp__a("input/input_section_types.F",1488)
      END IF
      vals => s_vals%values(ik, irs)%list
      IF (irk == -1) THEN
         CALL cp_sll_val_insert_el_at(vals, my_val, index=-1)
      ELSE IF (irk <= cp_sll_val_get_length(vals)) THEN
         IF (irk <= 0) &
            CALL cp_abort(cp__l("input/input_section_types.F",1495), &
                          "invalid irk "//TRIM(ADJUSTL(cp_to_string(irk)))// &
                          " in keyword "//TRIM(keyword_name(my_index:len_key))//" of section "// &
                          TRIM(section%name))
         old_val => cp_sll_val_get_el_at(vals, index=irk)
         CALL val_release(old_val)
         CALL cp_sll_val_set_el_at(vals, value=my_val, index=irk)
      ELSE IF (irk > cp_sll_val_get_length(vals)+1) THEN
         ! change?
         CALL cp_abort(cp__l("input/input_section_types.F",1504), &
                       "cannot add extra keyword repetitions to keyword" &
                       //TRIM(keyword_name(my_index:len_key))//" of section "// &
                       TRIM(section%name))
      ELSE
         CALL cp_sll_val_insert_el_at(vals, my_val, index=irk)
      END IF
      s_vals%values(ik, irs)%list => vals
      NULLIFY (my_val)
   END SUBROUTINE section_vals_val_set

! **************************************************************************************************
!> \brief unsets (removes) the requested value (if it is a keyword repetitions
!>      removes the repetition, so be careful: the repetition indices bigger
!>      than the actual change.
!> \param section_vals ...
!> \param keyword_name the name of the keyword you want (can be a path
!>        separated by '%')
!> \param i_rep_section which repetition of the section you are interested in
!>        (defaults to 1)
!> \param i_rep_val which repetition of the keyword/val you are interested in
!>        (defaults to 1)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_vals_val_unset(section_vals, keyword_name, i_rep_section, &
                                     i_rep_val)
      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: keyword_name
      INTEGER, INTENT(in), OPTIONAL                      :: i_rep_section, i_rep_val

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_val_unset', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ik, irk, irs, len_key, my_index, &
                                                            tmp_index
      TYPE(cp_sll_val_type), POINTER                     :: pos
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: s_vals
      TYPE(val_type), POINTER                            :: old_val

      NULLIFY (pos)
      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",1545)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1546)

      my_index = INDEX(keyword_name, '%')+1
      len_key = LEN_TRIM(keyword_name)
      IF (my_index > 1) THEN
         DO
            tmp_index = INDEX(keyword_name(my_index:len_key), "%")
            IF (tmp_index <= 0) EXIT
            my_index = my_index+tmp_index
         END DO
         s_vals => section_vals_get_subs_vals(section_vals, keyword_name(1:my_index-2))
      ELSE
         s_vals => section_vals
      END IF

      irk = 1
      irs = 1
      IF (PRESENT(i_rep_section)) irs = i_rep_section
      IF (PRESENT(i_rep_val)) irk = i_rep_val
      section => s_vals%section
      ik = section_get_keyword_index(s_vals%section, keyword_name(my_index:len_key))
      IF (ik == -2) &
         CALL cp_abort(cp__l("input/input_section_types.F",1568), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       TRIM(keyword_name(my_index:len_key)))
      ! ignore unset of non set values
      IF (irs <= SIZE(s_vals%values, 2)) THEN
         IF (.NOT. (irs > 0 .AND. irs <= SIZE(s_vals%subs_vals, 2))) &
            CALL cp_abort(cp__l("input/input_section_types.F",1574), &
                          "section repetition requested ("//cp_to_string(irs)// &
                          ") out of bounds (1:"//cp_to_string(SIZE(s_vals%subs_vals, 2)) &
                          //")")
         IF (irk == -1) THEN
            pos => cp_sll_val_get_rest(s_vals%values(ik, irs)%list, iter=-1)
         ELSE
            pos => cp_sll_val_get_rest(s_vals%values(ik, irs)%list, iter=irk-1)
         END IF
         IF (ASSOCIATED(pos)) THEN
            old_val => cp_sll_val_get_el_at(s_vals%values(ik, irs)%list, index=irk)
            CALL val_release(old_val)
            CALL cp_sll_val_rm_el_at(s_vals%values(ik, irs)%list, index=irk)
         END IF
      END IF

   END SUBROUTINE section_vals_val_unset

! **************************************************************************************************
!> \brief writes the values in the given section in a way that is suitable to
!>      the automatic parsing
!> \param section_vals the section to write out
!> \param unit_nr the unit where to write to
!> \param hide_root ...
!> \param hide_defaults ...
!> \author fawzi
!> \note
!>      skips required sections which weren't read
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_vals_write(section_vals, unit_nr, hide_root, hide_defaults)
      TYPE(section_vals_type), POINTER                   :: section_vals
      INTEGER, INTENT(in)                                :: unit_nr
      LOGICAL, INTENT(in), OPTIONAL                      :: hide_root, hide_defaults

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_write', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: myfmt
      INTEGER                                            :: i_rep_s, ik, isec, ival, nr, nval
      INTEGER, SAVE                                      :: indent = 1
      LOGICAL                                            :: defaultSection, explicit, &
                                                            my_hide_defaults, my_hide_root
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: sval
      TYPE(val_type), POINTER                            :: val

      my_hide_root = .FALSE.
      my_hide_defaults = .TRUE.
      IF (PRESENT(hide_root)) my_hide_root = hide_root
      IF (PRESENT(hide_defaults)) my_hide_defaults = hide_defaults

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",1627)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1628)
      IF (unit_nr > 0) THEN
         CALL section_vals_get(section_vals, explicit=explicit, n_repetition=nr, section=section)
         IF (explicit .OR. (.NOT. my_hide_defaults)) THEN
            DO i_rep_s = 1, nr
               IF (.NOT. my_hide_root) THEN
                  WRITE (myfmt, *) indent, "X"
                  CALL compress(myfmt, full=.TRUE.)
                  IF (ASSOCIATED(section%keywords(-1)%keyword)) THEN
                     WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//",'&',a,' ')", advance="NO") TRIM(section%name)
                  ELSE
                     WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//",'&',a)") TRIM(section%name)
                  END IF
               END IF
               defaultSection = (SIZE(section_vals%values, 2) == 0)
               IF (.NOT. defaultSection) THEN
                  IF (.NOT. my_hide_root) indent = indent+2
                  WRITE (myfmt, *) indent, "X"
                  CALL compress(myfmt, full=.TRUE.)
                  DO ik = -1, section%n_keywords
                     keyword => section%keywords(ik)%keyword
                     IF (ASSOCIATED(keyword)) THEN
                        IF (keyword%type_of_var /= no_t .AND. keyword%names(1) (1:2) /= "__") THEN
                           CALL section_vals_val_get(section_vals, keyword%names(1), &
                                                     i_rep_s, n_rep_val=nval)
                           IF (i_rep_s <= SIZE(section_vals%values, 2)) THEN
                              ! Section was parsed
                              vals => section_vals%values(ik, i_rep_s)%list
                              DO ival = 1, nval
                                 IF (ival == 1) THEN
                                    new_pos => vals
                                 ELSE
                                    new_pos => new_pos%rest
                                 END IF
                                 IF (.NOT. ASSOCIATED(new_pos)) THEN
                                    ! this keyword was not parsed
                                    IF (ASSOCIATED(keyword%default_value)) THEN
                                       val => keyword%default_value
                                       IF (my_hide_defaults) CYCLE
                                    END IF
                                 ELSE
                                    val => new_pos%first_el
                                 END IF
                                 IF (keyword%names(1) /= '_DEFAULT_KEYWORD_' .AND. &
                                     keyword%names(1) /= '_SECTION_PARAMETERS_') THEN
                                    WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//",a,' ')", advance="NO") &
                                       TRIM(keyword%names(1))
                                 ELSEIF (keyword%names(1) == '_DEFAULT_KEYWORD_' .AND. &
                                         keyword%type_of_var /= lchar_t) THEN
                                    WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//")", advance="NO")
                                 END IF
                                 CALL val_write(val, unit_nr=unit_nr, unit=keyword%unit, &
                                                fmt=myfmt)
                              END DO
                           ELSEIF (ASSOCIATED(keyword%default_value)) THEN
                              ! Section was not parsed but default for the keywords may exist
                              IF (my_hide_defaults) CYCLE
                              val => keyword%default_value
                              IF (keyword%names(1) /= '_DEFAULT_KEYWORD_' .AND. &
                                  keyword%names(1) /= '_SECTION_PARAMETERS_') THEN
                                 WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//",a,' ')", advance="NO") &
                                    TRIM(keyword%names(1))
                              ELSEIF (keyword%names(1) == '_DEFAULT_KEYWORD_' .AND. &
                                      keyword%type_of_var /= lchar_t) THEN
                                 WRITE (unit=unit_nr, fmt="("//TRIM(myfmt)//")", advance="NO")
                              END IF
                              CALL val_write(val, unit_nr=unit_nr, unit=keyword%unit, &
                                             fmt=myfmt)
                           END IF
                        END IF
                     END IF
                  END DO
                  IF (ASSOCIATED(section_vals%subs_vals)) THEN
                     DO isec = 1, SIZE(section_vals%subs_vals, 1)
                        sval => section_vals%subs_vals(isec, i_rep_s)%section_vals
                        IF (ASSOCIATED(sval)) THEN
                           CALL section_vals_write(sval, unit_nr=unit_nr, hide_defaults=hide_defaults)
                        END IF
                     END DO
                  END IF
               END IF
               IF (.NOT. my_hide_root) THEN
                  indent = indent-2
                  WRITE (UNIT=unit_nr, FMT="(A)") &
                     REPEAT(" ", indent)//"&END "//TRIM(section%name)
               END IF
            END DO
         END IF
      END IF

   END SUBROUTINE section_vals_write

! **************************************************************************************************
!> \brief writes the values in the given section in xml
!> \param section ...
!> \param level ...
!> \param unit_number ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE write_section_xml(section, level, unit_number)

      TYPE(section_type), POINTER                        :: section
      INTEGER, INTENT(IN)                                :: level, unit_number

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_section_xml', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=3)                                   :: repeats
      CHARACTER(LEN=8)                                   :: short_string
      INTEGER                                            :: i, l0, l1, l2

      IF (ASSOCIATED(section)) THEN

         IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",1740)

         ! Indentation for current level, next level, etc.

         l0 = level
         l1 = level+1
         l2 = level+2

         IF (section%repeats) THEN
            repeats = "yes"
         ELSE
            repeats = "no "
         END IF

         WRITE (UNIT=unit_number, FMT="(A)") &
            REPEAT(" ", l0)//"<SECTION repeats="""//TRIM(repeats)//""">", &
            REPEAT(" ", l1)//"<NAME>"//TRIM(section%name)//"</NAME>", &
            REPEAT(" ", l1)//"<DESCRIPTION>"// &
            TRIM(substitute_special_xml_tokens(a2s(section%description))) &
            //"</DESCRIPTION>"

         IF (ASSOCIATED(section%citations)) THEN
            DO i = 1, SIZE(section%citations, 1)
               short_string = ""
               WRITE (UNIT=short_string, FMT="(I8)") section%citations(i)
               WRITE (UNIT=unit_number, FMT="(A)") &
                  REPEAT(" ", l1)//"<REFERENCE>", &
                  REPEAT(" ", l2)//"<NAME>"//TRIM(get_citation_key(section%citations(i)))//"</NAME>", &
                  REPEAT(" ", l2)//"<NUMBER>"//TRIM(ADJUSTL(short_string))//"</NUMBER>", &
                  REPEAT(" ", l1)//"</REFERENCE>"
            END DO
         END IF

         DO i = -1, section%n_keywords
            IF (ASSOCIATED(section%keywords(i)%keyword)) THEN
               CALL write_keyword_xml(section%keywords(i)%keyword, l1, unit_number)
            END IF
         END DO

         DO i = 1, section%n_subsections
            CALL write_section_xml(section%subsections(i)%section, l1, unit_number)
         END DO

         WRITE (UNIT=unit_number, FMT="(A)") REPEAT(" ", l0)//"</SECTION>"

      END IF

   END SUBROUTINE write_section_xml

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \param section_name ...
!> \param unknown_string ...
!> \param location_string ...
!> \param matching_rank ...
!> \param matching_string ...
!> \param bonus ...
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_typo_match(section, section_name, unknown_string, location_string, &
                                           matching_rank, matching_string, bonus)

      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*)                                   :: section_name, unknown_string, &
                                                            location_string
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: matching_rank
      CHARACTER(LEN=*), DIMENSION(:), INTENT(INOUT)      :: matching_string
      INTEGER, INTENT(IN)                                :: bonus

      CHARACTER(LEN=*), PARAMETER :: routineN = 'section_typo_match', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=LEN(matching_string(1)))             :: line
      INTEGER                                            :: i, imatch, imax, irank, newbonus

      IF (ASSOCIATED(section)) THEN
         IF(.NOT.(section%ref_count > 0))CALL cp__a("input/input_section_types.F",1816)
         imatch = typo_match(TRIM(section%name), TRIM(unknown_string))
         IF (imatch > 0) THEN
            imatch = imatch+bonus
            WRITE (line, '(T2,A)') " subsection "//TRIM(section%name)//" in section "//TRIM(location_string)
            imax = SIZE(matching_rank, 1)
            irank = imax+1
            DO I = imax, 1, -1
               IF (imatch > matching_rank(I)) irank = i
            ENDDO
            IF (irank <= imax) THEN
               matching_rank(irank+1:imax) = matching_rank(irank:imax-1)
               matching_string(irank+1:imax) = matching_string(irank:imax-1)
               matching_rank(irank) = imatch
               matching_string(irank) = line
            ENDIF
         END IF

         IF (section_name == section%name) THEN
            newbonus = 10
         ELSE
            newbonus = 0
         ENDIF

         DO i = -1, section%n_keywords
            IF (ASSOCIATED(section%keywords(i)%keyword)) THEN
               CALL keyword_typo_match(section%keywords(i)%keyword, unknown_string, location_string// &
                                       "%"//TRIM(section%name), matching_rank, matching_string, newbonus)
            END IF
         END DO

         DO i = 1, section%n_subsections
            CALL section_typo_match(section%subsections(i)%section, section_name, unknown_string, &
                                    location_string//"%"//TRIM(section%name), matching_rank, matching_string, newbonus)
         END DO

      END IF

   END SUBROUTINE section_typo_match

! **************************************************************************************************
!> \brief replaces of the requested subsection with the one given
!> \param section_vals the root section
!> \param subsection_name the name of the subsection to replace
!> \param new_section_vals the new section_vals to use
!> \param i_rep_section index of the repetition of section_vals of which
!>        you want to replace the subsection (defaults to 1)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_vals_set_subs_vals(section_vals, subsection_name, &
                                         new_section_vals, i_rep_section)
      TYPE(section_vals_type), POINTER                   :: section_vals
      CHARACTER(len=*), INTENT(in)                       :: subsection_name
      TYPE(section_vals_type), POINTER                   :: new_section_vals
      INTEGER, INTENT(in), OPTIONAL                      :: i_rep_section

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_set_subs_vals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: irep, isection, len_key, my_index, &
                                                            tmp_index
      TYPE(section_vals_type), POINTER                   :: s_vals

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_section_types.F",1879)
      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1880)
      IF(.NOT.(ASSOCIATED(new_section_vals)))CALL cp__a("input/input_section_types.F",1881)
      IF(.NOT.(new_section_vals%ref_count > 0))CALL cp__a("input/input_section_types.F",1882)

      irep = 1
      IF (PRESENT(i_rep_section)) irep = i_rep_section

      my_index = INDEX(subsection_name, '%')+1
      len_key = LEN_TRIM(subsection_name)
      IF (my_index > 1) THEN
         DO
            tmp_index = INDEX(subsection_name(my_index:len_key), "%")
            IF (tmp_index <= 0) EXIT
            my_index = my_index+tmp_index
         END DO
         s_vals => section_vals_get_subs_vals(section_vals, subsection_name(1:my_index-2))
      ELSE
         s_vals => section_vals
      END IF

      IF(.NOT.(irep <= SIZE(s_vals%subs_vals, 2)))CALL cp__a("input/input_section_types.F",1900)

      isection = section_get_subsection_index(s_vals%section, subsection_name(my_index:LEN_TRIM(subsection_name)))
      IF (isection <= 0) &
         CALL cp_abort(cp__l("input/input_section_types.F",1904), &
                       "could not find subsection "//subsection_name(my_index:LEN_TRIM(subsection_name))//" in section "// &
                       TRIM(section_vals%section%name)//" at ")
      CALL section_vals_retain(new_section_vals)
      CALL section_vals_release(s_vals%subs_vals(isection, irep)%section_vals)
      s_vals%subs_vals(isection, irep)%section_vals => new_section_vals

   END SUBROUTINE section_vals_set_subs_vals

! **************************************************************************************************
!> \brief creates a deep copy from section_vals_in to section_vals_out
!> \param section_vals_in the section_vals to copy
!> \param section_vals_out the section_vals to create
!> \param i_rep_start ...
!> \param i_rep_end ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE section_vals_duplicate(section_vals_in, section_vals_out, &
                                     i_rep_start, i_rep_end)
      TYPE(section_vals_type), POINTER                   :: section_vals_in, section_vals_out
      INTEGER, INTENT(IN), OPTIONAL                      :: i_rep_start, i_rep_end

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_duplicate', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(section_vals_in)))CALL cp__a("input/input_section_types.F",1929)
      IF(.NOT.(.NOT. ASSOCIATED(section_vals_out)))CALL cp__a("input/input_section_types.F",1930)
      CALL section_vals_create(section_vals_out, section_vals_in%section)
      CALL section_vals_copy(section_vals_in, section_vals_out, i_rep_start, i_rep_end)
   END SUBROUTINE section_vals_duplicate

! **************************************************************************************************
!> \brief deep copy from section_vals_in to section_vals_out
!> \param section_vals_in the section_vals to copy
!> \param section_vals_out the section_vals where to copy
!> \param i_rep_low ...
!> \param i_rep_high ...
!> \author fawzi
!> \note
!>      private, only works with a newly initialized section_vals_out
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_vals_copy(section_vals_in, section_vals_out, &
                                          i_rep_low, i_rep_high)
      TYPE(section_vals_type), POINTER                   :: section_vals_in, section_vals_out
      INTEGER, INTENT(IN), OPTIONAL                      :: i_rep_low, i_rep_high

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iend, irep, isec, istart, ival
      TYPE(cp_sll_val_type), POINTER                     :: v1, v2
      TYPE(val_type), POINTER                            :: el

      NULLIFY (v2, el)

      IF(.NOT.(ASSOCIATED(section_vals_in)))CALL cp__a("input/input_section_types.F",1959)
      IF(.NOT.(ASSOCIATED(section_vals_out)))CALL cp__a("input/input_section_types.F",1960)
      !  IF(section_vals_in%section%id_nr/=section_vals_out%section%id_nr)&
      !     CALL cp__b("input/input_section_types.F",1962,"")

      istart = 1
      iend = SIZE(section_vals_in%values, 2)
      IF (PRESENT(i_rep_low)) istart = i_rep_low
      IF (PRESENT(i_rep_high)) iend = i_rep_high
      DO irep = istart, iend
         CALL section_vals_add_values(section_vals_out)
         DO ival = LBOUND(section_vals_in%values, 1), UBOUND(section_vals_in%values, 1)
            v1 => section_vals_in%values(ival, irep)%list
            IF (ASSOCIATED(v1)) THEN
               CALL val_duplicate(v1%first_el, el)
               CALL cp_sll_val_create(v2, el)
               NULLIFY (el)
               section_vals_out%values(ival, irep-istart+1)%list => v2
               DO
                  IF (.NOT. ASSOCIATED(v1%rest)) EXIT
                  v1 => v1%rest
                  CALL val_duplicate(v1%first_el, el)
                  CALL cp_sll_val_create(v2%rest, first_el=el)
                  NULLIFY (el)
                  v2 => v2%rest
               END DO
            END IF
         END DO
      END DO
      IF (.NOT. PRESENT(i_rep_low) .AND. (.NOT. PRESENT(i_rep_high))) THEN
         IF (.NOT. (SIZE(section_vals_in%values, 2) == SIZE(section_vals_out%values, 2))) &
            CALL cp__b("input/input_section_types.F",1990,"")
         IF (.NOT. (SIZE(section_vals_in%subs_vals, 2) == SIZE(section_vals_out%subs_vals, 2))) &
            CALL cp__b("input/input_section_types.F",1992,"")
      END IF
      iend = SIZE(section_vals_in%subs_vals, 2)
      IF (PRESENT(i_rep_high)) iend = i_rep_high
      DO irep = istart, iend
         DO isec = 1, SIZE(section_vals_in%subs_vals, 1)
            CALL section_vals_copy(section_vals_in%subs_vals(isec, irep)%section_vals, &
                                   section_vals_out%subs_vals(isec, irep-istart+1)%section_vals)
         END DO
      END DO
   END SUBROUTINE section_vals_copy

END MODULE input_section_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief various utilities that regard array of different kinds:
!>      output, allocation,...
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!>
!>      maybe it is not a good idea mixing output and memeory utils...
!> \par History
!>      12.2001 first version [fawzi]
!>      3.2002 templatized [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_array_utils_i
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: int_4
   USE machine,                         ONLY: m_flush
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_array_utils_i'

   ! types
   PUBLIC :: cp_1d_i_p_type, cp_2d_i_p_type, cp_3d_i_p_type

   ! generic interfaces
   PUBLIC :: cp_guarantee_size

   ! the underlying procedures
   PUBLIC :: cp_1d_i_guarantee_size, &
             cp_1d_i_write, cp_2d_i_write, &
             cp_2d_i_guarantee_size, &
             cp_1d_i_bsearch

   INTERFACE cp_guarantee_size
      MODULE PROCEDURE cp_1d_i_guarantee_size, &
         cp_2d_i_guarantee_size
   END INTERFACE

!***

! **************************************************************************************************
!> \brief represent a pointer to a 1d array
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE cp_1d_i_p_type
      INTEGER(kind=int_4), DIMENSION(:), POINTER :: array
   END TYPE cp_1d_i_p_type

! **************************************************************************************************
!> \brief represent a pointer to a 2d array
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE cp_2d_i_p_type
      INTEGER(kind=int_4), DIMENSION(:, :), POINTER :: array
   END TYPE cp_2d_i_p_type

! **************************************************************************************************
!> \brief represent a pointer to a 3d array
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE cp_3d_i_p_type
      INTEGER(kind=int_4), DIMENSION(:, :, :), POINTER :: array
   END TYPE cp_3d_i_p_type

CONTAINS

! **************************************************************************************************
!> \brief writes an array to the given unit
!> \param array the array to write
!> \param unit_nr the unit to write to (defaults to the standard out)
!> \param el_format the format of a single element
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      maybe I will move to a comma separated paretized list
! **************************************************************************************************
   SUBROUTINE cp_1d_i_write(array, unit_nr, el_format)
      INTEGER(kind=int_4), INTENT(in)                    :: array(:)
      INTEGER, INTENT(in)                                :: unit_nr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: el_format

      CHARACTER(len=*), PARAMETER :: defaultFormat = "(i6)", routineN = 'cp_1d_i_write', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iostat

      WRITE (unit=unit_nr, fmt="('( ')", advance="no", iostat=iostat)
      IF(.NOT.(iostat == 0))CALL cp__a("common/cp_array_utils_i.F",126)
      IF (PRESENT(el_format)) THEN
         DO i = 1, SIZE(array)-1
            WRITE (unit=unit_nr, fmt=el_format, advance="no") array(i)
            IF (MOD(i, 5) .EQ. 0) THEN ! only a few elements per line
               WRITE (unit=unit_nr, fmt="(',')")
            ELSE
               WRITE (unit=unit_nr, fmt="(',')", advance="no")
            ENDIF
         END DO
         IF (SIZE(array) > 0) &
            WRITE (unit=unit_nr, fmt=el_format, advance="no") array(SIZE(array))
      ELSE
         DO i = 1, SIZE(array)-1
            WRITE (unit=unit_nr, fmt=defaultFormat, advance="no") array(i)
            IF (MOD(i, 5) .EQ. 0) THEN ! only a few elements per line
               WRITE (unit=unit_nr, fmt="(',')")
            ELSE
               WRITE (unit=unit_nr, fmt="(',')", advance="no")
            ENDIF
         END DO
         IF (SIZE(array) > 0) &
            WRITE (unit=unit_nr, fmt=defaultFormat, advance="no") array(SIZE(array))
      END IF
      WRITE (unit=unit_nr, fmt="(' )')")
      CALL m_flush(unit_nr)

   END SUBROUTINE cp_1d_i_write

! **************************************************************************************************
!> \brief writes an array to the given unit
!> \param array the array to write
!> \param unit_nr the unit to write to (defaults to the standard out)
!> \param el_format the format of a single element
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      maybe I will move to a comma separated parentized list
! **************************************************************************************************
   SUBROUTINE cp_2d_i_write(array, unit_nr, el_format)
      INTEGER(kind=int_4), INTENT(in)                    :: array(:, :)
      INTEGER, INTENT(in)                                :: unit_nr
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: el_format

      CHARACTER(len=*), PARAMETER :: defaultFormat = "(i6)", routineN = 'cp_2d_i_write', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=10)                                  :: nRiga
      CHARACTER(len=200)                                 :: fmtstr
      INTEGER                                            :: i, iostat

      nRiga = cp_to_string(SIZE(array, 2))
      DO i = 1, SIZE(array, 1)
         IF (PRESENT(el_format)) THEN
            fmtstr = '(" ",'//nRiga//el_format//')'
            WRITE (unit=unit_nr, fmt=fmtstr, iostat=iostat) array(i, :)
         ELSE
            fmtstr = '(" ",'//nRiga//defaultFormat//')'
            WRITE (unit=unit_nr, fmt=fmtstr, iostat=iostat) array(i, :)
         END IF
         IF(.NOT.(iostat == 0))CALL cp__a("common/cp_array_utils_i.F",187)
      END DO
      CALL m_flush(unit_nr)
   END SUBROUTINE cp_2d_i_write

! **************************************************************************************************
!> \brief If the size of the array is changes reallocate it.
!>      Issues a warning when the size changes (but not on allocation
!>      and deallocation).
!>
!>      The data is NOT preserved (if you want to preserve the data see
!>      the realloc in the module memory_utilities)
!> \param array the array to reallocate if necessary
!> \param n the wanted size
!> \par History
!>      12.2001 first version [fawzi]
!>      3.2002 templatized [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      this is a different behaviour than the realloc in the module
!>      memory_utilities. It is quite low level
! **************************************************************************************************
   SUBROUTINE cp_1d_i_guarantee_size(array, n)
      INTEGER(kind=int_4), POINTER                       :: array(:)
      INTEGER, INTENT(in)                                :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_1d_i_guarantee_size', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(n >= 0))CALL cp__a("common/cp_array_utils_i.F",216)
      IF (ASSOCIATED(array)) THEN
         IF (SIZE(array) /= n) THEN
            CALL cp__w("common/cp_array_utils_i.F",219,'size has changed')
            DEALLOCATE (array)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(array)) THEN
         ALLOCATE (array(n))
      END IF
   END SUBROUTINE cp_1d_i_guarantee_size

! **************************************************************************************************
!> \brief If the size of the array is changes reallocate it.
!>      Issues a warning when the size changes (but not on allocation
!>      and deallocation).
!>
!>      The data is NOT preserved (if you want to preserve the data see
!>      the realloc in the module memory_utilities)
!> \param array the array to reallocate if necessary
!> \param n_rows the wanted number of rows
!> \param n_cols the wanted number of cols
!> \par History
!>      5.2001 first version [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      this is a different behaviour than the realloc in the module
!>      memory_utilities. It is quite low level
! **************************************************************************************************
   SUBROUTINE cp_2d_i_guarantee_size(array, n_rows, n_cols)
      INTEGER(kind=int_4), POINTER                       :: array(:, :)
      INTEGER, INTENT(in)                                :: n_rows, n_cols

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_2d_i_guarantee_size', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(n_cols >= 0))CALL cp__a("common/cp_array_utils_i.F",252)
      IF(.NOT.(n_rows >= 0))CALL cp__a("common/cp_array_utils_i.F",253)
      IF (ASSOCIATED(array)) THEN
         IF (SIZE(array, 1) /= n_rows .OR. SIZE(array, 2) /= n_cols) THEN
            CALL cp__w("common/cp_array_utils_i.F",256,'size has changed')
            DEALLOCATE (array)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(array)) THEN
         ALLOCATE (array(n_rows, n_cols))
      END IF
   END SUBROUTINE cp_2d_i_guarantee_size

! **************************************************************************************************
!> \brief returns the index at which the element el should be inserted in the
!>      array to keep it ordered (array(i)>=el).
!>      If the element is bigger than all the elements in the array returns
!>      the last index+1.
!> \param array the array to search
!> \param el the element to look for
!> \param l_index the lower index for binary search (defaults to 1)
!> \param u_index the upper index for binary search (defaults to size(array))
!> \retval res ...
!> \par History
!>      06.2003 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      the array should be ordered in growing order
! **************************************************************************************************
   FUNCTION cp_1d_i_bsearch(array, el, l_index, u_index) &
      RESULT(res)
      INTEGER(kind=int_4), INTENT(in)                    :: array(:), el
      INTEGER, INTENT(in), OPTIONAL                      :: l_index, u_index
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_1d_i_bsearch', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: aindex, lindex, uindex

      lindex = 1
      uindex = SIZE(array)
      IF (PRESENT(l_index)) lindex = l_index
      IF (PRESENT(u_index)) uindex = u_index
      DO WHILE (lindex <= uindex)
         aindex = (lindex+uindex)/2
         IF (( array(aindex) <  el )) THEN
            lindex = aindex+1
         ELSE
            uindex = aindex-1
         END IF
      END DO
      res = lindex
   END FUNCTION cp_1d_i_bsearch

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,defaultFormatType1,lessQ,defines)]
! ARGS:
!  USE = "USE kinds, only: int_4"
!  defaultFormatType1 = ""(i6)""
!  defines =
!    "#define CP_I_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_I_LESS_Q"
!  nametype1 = "i"
!  type1 = "INTEGER(kind=int_4)"

END MODULE cp_array_utils_i
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief wrappers for the actual blacs calls.
!>      all functionality needed in the code should actually be provide by cp_blacs_env
!>      these functions should be private members of that module
!> \note
!>      http://www.netlib.org/blacs/BLACS/QRef.html
!> \par History
!>      12.2003 created [Joost]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE cp_blacs_calls

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_blacs_calls'
   PRIVATE
   ! setup / info calls
   PUBLIC :: cp_blacs_gridinit, cp_blacs_set, &
             cp_blacs_gridexit, cp_blacs_gridinfo
   ! actual message passing
   PUBLIC :: cp_blacs_zgebs2d, cp_blacs_zgebr2d, cp_blacs_dgebs2d, &
             cp_blacs_dgebr2d

!***
CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param context ...
!> \param order ...
!> \param nprow ...
!> \param npcol ...
! **************************************************************************************************
   SUBROUTINE cp_blacs_gridinit(context, order, nprow, npcol)
      INTEGER, INTENT(INOUT) :: context
      CHARACTER(len=1), INTENT(IN):: order
      INTEGER, INTENT(IN)    :: nprow, npcol
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(context))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(order))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nprow))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(npcol))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE cp_blacs_gridinit

! **************************************************************************************************
!> \brief ...
!> \param context ...
! **************************************************************************************************
   SUBROUTINE cp_blacs_gridexit(context)
      INTEGER, INTENT(IN) :: context
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(context))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE cp_blacs_gridexit

! **************************************************************************************************
!> \brief ...
!> \param context ...
!> \param nprow ...
!> \param npcol ...
!> \param myprow ...
!> \param mypcol ...
! **************************************************************************************************
   SUBROUTINE cp_blacs_gridinfo(context, nprow, npcol, myprow, mypcol)
      INTEGER, INTENT(IN)  :: context
      INTEGER, INTENT(OUT) :: nprow, npcol, myprow, mypcol
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(context))==-1) EXIT ;  END DO ; ENDIF
      nprow = 1
      npcol = 1
      myprow = 0
      mypcol = 0
   END SUBROUTINE cp_blacs_gridinfo

! **************************************************************************************************
!> \brief ...
!> \param context ...
!> \param what :
!>     WHAT = 0 : Handle indicating default system context;  ! DO NOT USE (i.e. use para_env%group)
!>     WHAT = 1 : The BLACS message ID range;
!>     WHAT = 2 : The BLACS debug level the library was compiled with;
!>     WHAT = 10: Handle indicating the system context used to define the BLACS context whose handle is ICONTXT;
!>     WHAT = 11: Number of rings multiring topology is presently using;
!>     WHAT = 12: Number of branches general tree topology is presently using.
!>     WHAT = 15: If non-zero, makes topology choice for repeatable collectives
!> \param val ...
! **************************************************************************************************
   SUBROUTINE cp_blacs_set(context, what, val)
      INTEGER, INTENT(IN)  :: context, what, val
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(context))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(what))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(val))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE cp_blacs_set

! **************************************************************************************************
!> \brief ...
!> \param ICONTXT ...
!> \param SCOPE ...
!> \param TOP ...
!> \param M ...
!> \param N ...
!> \param A ...
!> \param LDA ...
! **************************************************************************************************
   SUBROUTINE cp_blacs_zgebs2d(ICONTXT, SCOPE, TOP, M, N, A, LDA)
      INTEGER, INTENT(IN)     :: ICONTXT
      CHARACTER(len=1), INTENT(IN) :: SCOPE, TOP
      INTEGER, INTENT(IN)     :: M, N, LDA
      COMPLEX(KIND=dp)            :: A
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(ICONTXT))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(SCOPE))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(TOP))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(M))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(N))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(A))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(LDA))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE
! **************************************************************************************************
!> \brief ...
!> \param ICONTXT ...
!> \param SCOPE ...
!> \param TOP ...
!> \param M ...
!> \param N ...
!> \param A ...
!> \param LDA ...
!> \param RSRC ...
!> \param CSRC ...
! **************************************************************************************************
   SUBROUTINE cp_blacs_zgebr2d(ICONTXT, SCOPE, TOP, M, N, A, LDA, RSRC, CSRC)
      INTEGER, INTENT(IN)     :: ICONTXT
      CHARACTER(len=1), INTENT(IN) :: SCOPE, TOP
      INTEGER, INTENT(IN)     :: M, N, LDA
      INTEGER, INTENT(IN)     :: RSRC, CSRC
      COMPLEX(KIND=dp)            :: A
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(ICONTXT))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(SCOPE))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(TOP))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(M))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(N))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(A))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(LDA))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(RSRC))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(CSRC))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param ICONTXT ...
!> \param SCOPE ...
!> \param TOP ...
!> \param M ...
!> \param N ...
!> \param A ...
!> \param LDA ...
! **************************************************************************************************
   SUBROUTINE cp_blacs_dgebs2d(ICONTXT, SCOPE, TOP, M, N, A, LDA)
      INTEGER, INTENT(IN)     :: ICONTXT
      CHARACTER(len=1), INTENT(IN) :: SCOPE, TOP
      INTEGER, INTENT(IN)     :: M, N, LDA
      REAL(KIND=dp)               :: A
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(ICONTXT))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(SCOPE))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(TOP))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(M))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(N))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(A))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(LDA))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE
! **************************************************************************************************
!> \brief ...
!> \param ICONTXT ...
!> \param SCOPE ...
!> \param TOP ...
!> \param M ...
!> \param N ...
!> \param A ...
!> \param LDA ...
!> \param RSRC ...
!> \param CSRC ...
! **************************************************************************************************
   SUBROUTINE cp_blacs_dgebr2d(ICONTXT, SCOPE, TOP, M, N, A, LDA, RSRC, CSRC)
      INTEGER, INTENT(IN)     :: ICONTXT
      CHARACTER(len=1), INTENT(IN) :: SCOPE, TOP
      INTEGER, INTENT(IN)     :: M, N, LDA
      INTEGER, INTENT(IN)     :: RSRC, CSRC
      REAL(KIND=dp)               :: A
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(ICONTXT))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(SCOPE))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(TOP))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(M))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(N))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(A))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(LDA))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(RSRC))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(CSRC))==-1) EXIT ;  END DO ; ENDIF
   END SUBROUTINE

END MODULE cp_blacs_calls
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Collection of simple mathematical functions and subroutines
!> \par History
!>      FUNCTION angle updated and FUNCTION dihedral angle added; cleaned
!>      (13.03.2004,MK)
!> \author MK (15.11.1998)
! **************************************************************************************************
MODULE mathlib

   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: euler,&
                                              fac
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mathlib'
   REAL(KIND=dp), PARAMETER             :: eps_geo = 1.0E-6_dp

   ! Public subroutines

   PUBLIC :: build_rotmat, &
             jacobi, &
             diamat_all, &
             invmat, &
             invmat_symm, &
             invert_matrix, &
             get_pseudo_inverse_svd, &
             get_pseudo_inverse_diag, &
             symmetrize_matrix, &
             unit_matrix, diag

   ! Public functions

   PUBLIC :: angle, &
             binomial, &
             multinomial, &
             det_3x3, &
             dihedral_angle, &
             gcd, &
             inv_3x3, &
             lcm, &
             vector_product, &
             matmul_3x3, &
             matvec_3x3, &
             rotate_vector, &
             reflect_vector, &
             dotprod_3d, &
             transpose_3d, &
             expint, abnormal_value

   INTERFACE det_3x3
      MODULE PROCEDURE det_3x3_1, det_3x3_2
   END INTERFACE

   INTERFACE invert_matrix
      MODULE PROCEDURE invert_matrix_d, invert_matrix_z
   END INTERFACE

   INTERFACE set_diag
      MODULE PROCEDURE set_diag_scalar_d, set_diag_scalar_z
   END INTERFACE

   INTERFACE swap
      MODULE PROCEDURE swap_scalar, swap_vector
   END INTERFACE

   INTERFACE unit_matrix
      MODULE PROCEDURE unit_matrix_d, unit_matrix_z
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief determins if a value is not normal (e.g. for Inf and Nan)
!>        based on IO to work also under optimization.
!> \param a input value
!> \retval abnormal_value TRUE for NaN and Inf
! **************************************************************************************************
   LOGICAL FUNCTION abnormal_value(a)
      REAL(KIND=dp) :: a
      CHARACTER(LEN=32) :: buffer

      abnormal_value = .FALSE.
      ! the function should work when compiled witht -ffast-math and similar
      ! unfortunately, that option asserts that all numbers are normals,
      ! which the compiler uses to optimize the function to .FALSE. if based on the IEEE module
      ! therefore, pass this to the Fortran runtime/printf, if things are NaN or Inf, error out.
      WRITE (buffer, *) a
      IF (INDEX(buffer, "N") .NE. 0 .OR. INDEX(buffer, "n") .NE. 0) abnormal_value = .TRUE.

   END FUNCTION abnormal_value

! **************************************************************************************************
!> \brief  Calculation of the angle between the vectors a and b.
!>         The angle is returned in radians.
!> \param a ...
!> \param b ...
!> \retval angle_ab ...
!> \date    14.10.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION angle(a, b) RESULT(angle_ab)

      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: a, b
      REAL(KIND=dp)                                      :: angle_ab

      REAL(KIND=dp)                                      :: length_of_a, length_of_b
      REAL(KIND=dp), DIMENSION(SIZE(a, 1))               :: a_norm, b_norm

      length_of_a = SQRT(DOT_PRODUCT(a, a))
      length_of_b = SQRT(DOT_PRODUCT(b, b))

      IF ((length_of_a > eps_geo) .AND. (length_of_b > eps_geo)) THEN
         a_norm(:) = a(:)/length_of_a
         b_norm(:) = b(:)/length_of_b
         angle_ab = ACOS(MIN(MAX(DOT_PRODUCT(a_norm, b_norm), -1.0_dp), 1.0_dp))
      ELSE
         angle_ab = 0.0_dp
      END IF

   END FUNCTION angle

! **************************************************************************************************
!> \brief   The binomial coefficient n over k for 0 <= k <= n is calculated,
!>            otherwise zero is returned.
!> \param n ...
!> \param k ...
!> \retval n_over_k ...
!> \date    08.03.1999
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION binomial(n, k) RESULT(n_over_k)

      INTEGER, INTENT(IN)                                :: n, k
      REAL(KIND=dp)                                      :: n_over_k

      IF ((k >= 0) .AND. (k <= n)) THEN
         n_over_k = fac(n)/(fac(n-k)*fac(k))
      ELSE
         n_over_k = 0.0_dp
      END IF

   END FUNCTION binomial

! **************************************************************************************************
!> \brief Calculates the multinomial coefficients
!> \param n ...
!> \param k ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION multinomial(n, k) RESULT(res)

      INTEGER, INTENT(IN)                                :: n
      INTEGER, DIMENSION(:), INTENT(IN)                  :: k
      REAL(KIND=dp)                                      :: res

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: denom

      IF (ALL(k >= 0) .AND. SUM(k) == n) THEN
         denom = 1.0_dp
         DO i = 1, SIZE(k)
            denom = denom*fac(k(i))
         ENDDO
         res = fac(n)/denom
      ELSE
         res = 0.0_dp
      END IF

   END FUNCTION multinomial

! **************************************************************************************************
!> \brief   The rotation matrix rotmat which rotates a vector about a
!>          rotation axis defined by the vector a is build up.
!>          The rotation angle is phi (radians).
!> \param phi ...
!> \param a ...
!> \param rotmat ...
!> \date    16.10.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE build_rotmat(phi, a, rotmat)
      REAL(KIND=dp), INTENT(IN)                          :: phi
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: rotmat

      REAL(KIND=dp)                                      :: cosp, cost, length_of_a, sinp
      REAL(KIND=dp), DIMENSION(3)                        :: d

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))
      ! Check the length of the vector a
      IF (length_of_a > eps_geo) THEN

         d(:) = a(:)/length_of_a

         cosp = COS(phi)
         sinp = SIN(phi)
         cost = 1.0_dp-cosp

         rotmat(1, 1) = d(1)*d(1)*cost+cosp
         rotmat(1, 2) = d(1)*d(2)*cost-d(3)*sinp
         rotmat(1, 3) = d(1)*d(3)*cost+d(2)*sinp
         rotmat(2, 1) = d(2)*d(1)*cost+d(3)*sinp
         rotmat(2, 2) = d(2)*d(2)*cost+cosp
         rotmat(2, 3) = d(2)*d(3)*cost-d(1)*sinp
         rotmat(3, 1) = d(3)*d(1)*cost-d(2)*sinp
         rotmat(3, 2) = d(3)*d(2)*cost+d(1)*sinp
         rotmat(3, 3) = d(3)*d(3)*cost+cosp
      ELSE
         CALL unit_matrix(rotmat)
      END IF

   END SUBROUTINE build_rotmat

! **************************************************************************************************
!> \brief   Returns the determinante of the 3x3 matrix a.
!> \param a ...
!> \retval det_a ...
!> \date    13.03.2004
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION det_3x3_1(a) RESULT(det_a)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: a
      REAL(KIND=dp)                                      :: det_a

      det_a = a(1, 1)*(a(2, 2)*a(3, 3)-a(2, 3)*a(3, 2))+ &
              a(1, 2)*(a(2, 3)*a(3, 1)-a(2, 1)*a(3, 3))+ &
              a(1, 3)*(a(2, 1)*a(3, 2)-a(2, 2)*a(3, 1))

   END FUNCTION det_3x3_1

! **************************************************************************************************
!> \brief   Returns the determinante of the 3x3 matrix a given by its columns.
!> \param a1 ...
!> \param a2 ...
!> \param a3 ...
!> \retval det_a ...
!> \date    13.03.2004
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION det_3x3_2(a1, a2, a3) RESULT(det_a)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a1, a2, a3
      REAL(KIND=dp)                                      :: det_a

      det_a = a1(1)*(a2(2)*a3(3)-a3(2)*a2(3))+ &
              a2(1)*(a3(2)*a1(3)-a1(2)*a3(3))+ &
              a3(1)*(a1(2)*a2(3)-a2(2)*a1(3))

   END FUNCTION det_3x3_2

! **************************************************************************************************
!> \brief Diagonalize the symmetric n by n matrix a using the LAPACK
!>        library. Only the upper triangle of matrix a is used.
!>        Externals (LAPACK 3.0)
!> \param a ...
!> \param eigval ...
!> \param dac ...
!> \date    29.03.1999
!> \par Variables
!>      - a       : Symmetric matrix to be diagonalized (input; upper triangle) ->
!>      -           eigenvectors of the matrix a (output).
!>      - dac     : If true, then the divide-and-conquer algorithm is applied.
!>      - eigval  : Eigenvalues of the matrix a (output).
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE diamat_all(a, eigval, dac)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: a
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: eigval
      LOGICAL, INTENT(IN), OPTIONAL                      :: dac

      CHARACTER(len=*), PARAMETER :: routineN = 'diamat_all', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, info, liwork, lwork, n, nb
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iwork
      INTEGER, EXTERNAL                                  :: ilaenv
      LOGICAL                                            :: divide_and_conquer
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: work

      EXTERNAL dsyev, dsyevd

      CALL timeset(routineN, handle)

      ! Get the size of the matrix a
      n = SIZE(a, 1)

      ! Check the size of matrix a
      IF (SIZE(a, 2) /= n) THEN
         CALL cp__b("common/mathlib.F",304,"Check the size of matrix a (parameter #1)")
      END IF

      ! Check the size of vector eigval
      IF (SIZE(eigval) /= n) THEN
         CALL cp__b("common/mathlib.F",309,"The dimension of vector eigval is too small")
      END IF

      ! Check, if the divide-and-conquer algorithm is requested

      IF (PRESENT(dac)) THEN
         divide_and_conquer = dac
      ELSE
         divide_and_conquer = .FALSE.
      END IF

      ! Get the optimal work storage size

      IF (divide_and_conquer) THEN
         lwork = 2*n**2+6*n+1
         liwork = 5*n+3
      ELSE
         nb = ilaenv(1, "DSYTRD", "U", n, -1, -1, -1)
         lwork = (nb+2)*n
      END IF

      ! Allocate work storage

      ALLOCATE (work(lwork))
      IF (divide_and_conquer) THEN
         ALLOCATE (iwork(liwork))
      END IF

      ! Diagonalize the matrix a

      IF (divide_and_conquer) THEN
         CALL dsyevd("V", "U", n, a, n, eigval, work, lwork, iwork, liwork, info)
      ELSE
         CALL dsyev("V", "U", n, a, n, eigval, work, lwork, info)
      END IF

      IF (info /= 0) THEN
         IF (divide_and_conquer) THEN
            CALL cp__b("common/mathlib.F",347,"The matrix diagonalization with dsyevd failed")
         ELSE
            CALL cp__b("common/mathlib.F",349,"The matrix diagonalization with dsyev failed")
         END IF
      END IF

      ! Release work storage

      DEALLOCATE (work)

      IF (divide_and_conquer) THEN
         DEALLOCATE (iwork)
      END IF

      CALL timestop(handle)

   END SUBROUTINE diamat_all

! **************************************************************************************************
!> \brief   Returns the dihedral angle, i.e. the angle between the planes
!>          defined by the vectors (-ab,bc) and (cd,-bc).
!>          The dihedral angle is returned in radians.
!> \param ab ...
!> \param bc ...
!> \param cd ...
!> \retval dihedral_angle_abcd ...
!> \date    13.03.2004
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION dihedral_angle(ab, bc, cd) RESULT(dihedral_angle_abcd)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: ab, bc, cd
      REAL(KIND=dp)                                      :: dihedral_angle_abcd

      REAL(KIND=dp)                                      :: det_abcd
      REAL(KIND=dp), DIMENSION(3)                        :: abc, bcd

      abc = vector_product(bc, -ab)
      bcd = vector_product(cd, -bc)
      ! Calculate the normal vectors of the planes
      ! defined by the points a,b,c and b,c,d

      det_abcd = det_3x3(abc, bcd, -bc)
      dihedral_angle_abcd = SIGN(1.0_dp, det_abcd)*angle(abc, bcd)

   END FUNCTION dihedral_angle

! **************************************************************************************************
!> \brief   Return the diagonal elements of matrix a as a vector.
!> \param a ...
!> \retval a_diag ...
!> \date    20.11.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION get_diag(a) RESULT(a_diag)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: a
      REAL(KIND=dp), &
         DIMENSION(MIN(SIZE(a, 1), SIZE(a, 2)))          :: a_diag

      INTEGER                                            :: i, n

      n = MIN(SIZE(a, 1), SIZE(a, 2))

      DO i = 1, n
         a_diag(i) = a(i, i)
      END DO

   END FUNCTION get_diag

! **************************************************************************************************
!> \brief   Returns the inverse of the 3 x 3 matrix a.
!> \param a ...
!> \retval a_inv ...
!> \date    13.03.2004
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION inv_3x3(a) RESULT(a_inv)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: a
      REAL(KIND=dp), DIMENSION(3, 3)                     :: a_inv

      REAL(KIND=dp)                                      :: det_a

      det_a = 1.0_dp/det_3x3(a)

      a_inv(1, 1) = (a(2, 2)*a(3, 3)-a(3, 2)*a(2, 3))*det_a
      a_inv(2, 1) = (a(2, 3)*a(3, 1)-a(3, 3)*a(2, 1))*det_a
      a_inv(3, 1) = (a(2, 1)*a(3, 2)-a(3, 1)*a(2, 2))*det_a

      a_inv(1, 2) = (a(1, 3)*a(3, 2)-a(3, 3)*a(1, 2))*det_a
      a_inv(2, 2) = (a(1, 1)*a(3, 3)-a(3, 1)*a(1, 3))*det_a
      a_inv(3, 2) = (a(1, 2)*a(3, 1)-a(3, 2)*a(1, 1))*det_a

      a_inv(1, 3) = (a(1, 2)*a(2, 3)-a(2, 2)*a(1, 3))*det_a
      a_inv(2, 3) = (a(1, 3)*a(2, 1)-a(2, 3)*a(1, 1))*det_a
      a_inv(3, 3) = (a(1, 1)*a(2, 2)-a(2, 1)*a(1, 2))*det_a

   END FUNCTION inv_3x3

! **************************************************************************************************
!> \brief returns inverse of matrix using the lapack routines DGETRF and DGETRI
!> \param a ...
!> \param info ...
! **************************************************************************************************
   SUBROUTINE invmat(a, info)
      REAL(KIND=dp), INTENT(INOUT)                       :: a(:, :)
      INTEGER, INTENT(OUT)                               :: info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'invmat', routineP = moduleN//':'//routineN

      INTEGER                                            :: lwork, n
      INTEGER, ALLOCATABLE                               :: ipiv(:)
      REAL(KIND=dp), ALLOCATABLE                         :: work(:)

      n = SIZE(a, 1)
      lwork = 20*n
      ALLOCATE (ipiv(n))
      ALLOCATE (work(lwork))
      ipiv = 0
      work = 0._dp
      info = 0
      CALL dgetrf(n, n, a, n, ipiv, info)
      IF (info == 0) THEN
         CALL dgetri(n, a, n, ipiv, work, lwork, info)
      END IF
      DEALLOCATE (ipiv, work)
   END SUBROUTINE invmat

! **************************************************************************************************
!> \brief returns inverse of real symmetric, positive definite matrix
!> \param a matrix
!> \param cholesky_triangle if cholesky decomposition of a was already done
!>        using dpotrf, indicating if the upper or lower triangle of a is
!>        stored. If not given, cholesky decomposition of a will be done
!>        before inversion.
!> \author Dorothea Golze [02.2015]
! **************************************************************************************************
   SUBROUTINE invmat_symm(a, cholesky_triangle)
      REAL(KIND=dp), INTENT(INOUT)                       :: a(:, :)
      CHARACTER(LEN=1), INTENT(IN), OPTIONAL             :: cholesky_triangle

      CHARACTER(LEN=*), PARAMETER :: routineN = 'invmat_symm', routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: my_triangle
      INTEGER                                            :: handle, i, info, n

      CALL timeset(routineN, handle)

      n = SIZE(a, 1)
      info = 0

      IF (PRESENT(cholesky_triangle)) THEN
         my_triangle = cholesky_triangle
      ELSE
         my_triangle = "U"
      ENDIF

      ! do cholesky decomposition
      IF (.NOT. PRESENT(cholesky_triangle)) THEN
         CALL dpotrf(my_triangle, n, a, n, info)
         IF (info /= 0) THEN
            CALL cp__b("common/mathlib.F",510,"DPOTRF failed")
         END IF
      ENDIF

      ! do inversion using the cholesky decomposition
      CALL dpotri(my_triangle, n, a, n, info)
      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",517,"Matrix inversion failed")
      END IF

      IF (my_triangle == "U") THEN
         DO i = 1, n-1
            a(i+1:n, i) = a(i, i+1:n)
         ENDDO
      ELSE
         DO i = 1, n-1
            a(i, i+1:n) = a(i+1:n, i)
         ENDDO
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE invmat_symm

! **************************************************************************************************
!> \brief  Compute the inverse of the n by n real matrix a using the LAPACK
!>         library
!> \param a ...
!> \param a_inverse ...
!> \param eval_error ...
!> \param option ...
!> \param improve ...
!> \date   23.03.1999
!> \par Variables
!>       - a        : Real matrix to be inverted (input).
!>       - a_inverse: Inverse of the matrix a (output).
!>       - a_lu     : LU factorization of matrix a.
!>       - a_norm   : Norm of matrix a.
!>       - error    : Estimated error of the inversion.
!>       - r_cond   : Reciprocal condition number of the matrix a.
!>       - trans    : "N" => invert a
!>       -            "T" => invert transpose(a)
!> \author MK
!> \version 1.0
!> \note NB add improve argument, used to disable call to dgerfs
! **************************************************************************************************
   SUBROUTINE invert_matrix_d(a, a_inverse, eval_error, option, improve)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: a
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: a_inverse
      REAL(KIND=dp), INTENT(OUT)                         :: eval_error
      CHARACTER(LEN=1), INTENT(IN), OPTIONAL             :: option
      LOGICAL, INTENT(IN), OPTIONAL                      :: improve

      CHARACTER(LEN=*), PARAMETER :: routineN = 'invert_matrix_d', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: norm, trans
      CHARACTER(LEN=default_string_length)               :: message
      INTEGER                                            :: info, iter, n
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipiv, iwork
      LOGICAL                                            :: do_improve
      REAL(KIND=dp)                                      :: a_norm, old_eval_error, r_cond
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: berr, ferr, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: a_lu, b
      REAL(KIND=dp), EXTERNAL                            :: dlange

      EXTERNAL dgecon, dgerfs, dgetrf, dgetrs

      ! Check for optional parameter
      IF (PRESENT(option)) THEN
         trans = option
      ELSE
         trans = "N"
      END IF

      IF (PRESENT(improve)) THEN
         do_improve = improve
      ELSE
         do_improve = .TRUE.
      ENDIF

      ! Get the dimension of matrix a
      n = SIZE(a, 1)

      ! Check array dimensions
      IF (n == 0) THEN
         CALL cp__b("common/mathlib.F",596,"Matrix to be inverted of zero size")
      END IF

      IF (n /= SIZE(a, 2)) THEN
         CALL cp__b("common/mathlib.F",600,"Check the array bounds of parameter #1")
      END IF

      IF ((n /= SIZE(a_inverse, 1)) .OR. &
          (n /= SIZE(a_inverse, 2))) THEN
         CALL cp__b("common/mathlib.F",605,"Check the array bounds of parameter #2")
      END IF

      ! Allocate work storage
      ALLOCATE (a_lu(n, n))

      ALLOCATE (b(n, n))

      ALLOCATE (berr(n))

      ALLOCATE (ferr(n))

      ALLOCATE (ipiv(n))

      ALLOCATE (iwork(n))

      ALLOCATE (work(4*n))

      a_lu(1:n, 1:n) = a(1:n, 1:n)

      ! Compute the LU factorization of the matrix a
      CALL dgetrf(n, n, a_lu, n, ipiv, info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",629,"The LU factorization in dgetrf failed")
      END IF

      ! Compute the norm of the matrix a

      IF (trans == "N") THEN
         norm = '1'
      ELSE
         norm = 'I'
      END IF

      a_norm = dlange(norm, n, n, a, n, work)

      ! Compute the reciprocal of the condition number of a

      CALL dgecon(norm, n, a_lu, n, a_norm, r_cond, work, iwork, info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",647,"The computation of the condition number in dgecon failed")
      END IF

      IF (r_cond < EPSILON(0.0_dp)) THEN
         WRITE (message, "(A,ES10.3)") "R_COND =", r_cond
         CALL cp_abort(cp__l("common/mathlib.F",652), &
                       "Bad condition number "//TRIM(message)//" (smaller than the machine "// &
                       "working precision)")
      END IF

      ! Solve a system of linear equations using the LU factorization computed by dgetrf

      CALL unit_matrix(a_inverse)

      CALL dgetrs(trans, n, n, a_lu, n, ipiv, a_inverse, n, info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",664,"Solving the system of linear equations in dgetrs failed")
      END IF

      ! Improve the computed solution iteratively
      CALL unit_matrix(b) ! Initialize right-hand sides

      eval_error = 0.0_dp

      IF (do_improve) THEN
         DO iter = 1, 10

            CALL dgerfs(trans, n, n, a, n, a_lu, n, ipiv, b, n, a_inverse, n, ferr, berr, &
                        work, iwork, info)

            IF (info /= 0) THEN
               CALL cp__b("common/mathlib.F",679,"Improving the computed solution in dgerfs failed")
            END IF

            old_eval_error = eval_error
            eval_error = MAXVAL(ferr)

            IF (ABS(eval_error-old_eval_error) <= EPSILON(1.0_dp)) EXIT

         END DO
      ENDIF

      ! Release work storage
      DEALLOCATE (work)
      DEALLOCATE (iwork)
      DEALLOCATE (ipiv)
      DEALLOCATE (ferr)
      DEALLOCATE (berr)
      DEALLOCATE (b)
      DEALLOCATE (a_lu)

   END SUBROUTINE invert_matrix_d

! **************************************************************************************************
!> \brief  Compute the inverse of the n by n complex matrix a using the LAPACK
!>         library
!> \param a ...
!> \param a_inverse ...
!> \param eval_error ...
!> \param option ...
!> \date   08.06.2009
!> \par Variables
!>       - a        : Complex matrix to be inverted (input).
!>       - a_inverse: Inverse of the matrix a (output).
!>       - a_lu     : LU factorization of matrix a.
!>       - a_norm   : Norm of matrix a.
!>       - error    : Estimated error of the inversion.
!>       - r_cond   : Reciprocal condition number of the matrix a.
!>       - trans    : "N" => invert a
!>       -            "T" => invert transpose(a)
!> \author MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE invert_matrix_z(a, a_inverse, eval_error, option)
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: a
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(OUT)     :: a_inverse
      REAL(KIND=dp), INTENT(OUT)                         :: eval_error
      CHARACTER(LEN=1), INTENT(IN), OPTIONAL             :: option

      CHARACTER(LEN=*), PARAMETER :: routineN = 'invert_matrix_z', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: norm, trans
      CHARACTER(LEN=default_string_length)               :: message
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: work
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: a_lu, b
      INTEGER                                            :: info, iter, n
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipiv
      REAL(KIND=dp)                                      :: a_norm, old_eval_error, r_cond
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: berr, ferr, rwork
      REAL(KIND=dp), EXTERNAL                            :: zlange

      EXTERNAL zgecon, zgerfs, zgetrf, zgetrs

      ! Check for optional parameter
      IF (PRESENT(option)) THEN
         trans = option
      ELSE
         trans = "N"
      END IF

      ! Get the dimension of matrix a
      n = SIZE(a, 1)

      ! Check array dimensions
      IF (n == 0) THEN
         CALL cp__b("common/mathlib.F",754,"Matrix to be inverted of zero size")
      END IF

      IF (n /= SIZE(a, 2)) THEN
         CALL cp__b("common/mathlib.F",758,"Check the array bounds of parameter #1")
      END IF

      IF ((n /= SIZE(a_inverse, 1)) .OR. &
          (n /= SIZE(a_inverse, 2))) THEN
         CALL cp__b("common/mathlib.F",763,"Check the array bounds of parameter #2")
      END IF

      ! Allocate work storage
      ALLOCATE (a_lu(n, n))

      ALLOCATE (b(n, n))

      ALLOCATE (berr(n))

      ALLOCATE (ferr(n))

      ALLOCATE (ipiv(n))

      ALLOCATE (rwork(2*n))

      ALLOCATE (work(2*n))

      a_lu(1:n, 1:n) = a(1:n, 1:n)

      ! Compute the LU factorization of the matrix a
      CALL zgetrf(n, n, a_lu, n, ipiv, info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",787,"The LU factorization in dgetrf failed")
      END IF

      ! Compute the norm of the matrix a

      IF (trans == "N") THEN
         norm = '1'
      ELSE
         norm = 'I'
      END IF

      a_norm = zlange(norm, n, n, a, n, work)

      ! Compute the reciprocal of the condition number of a

      CALL zgecon(norm, n, a_lu, n, a_norm, r_cond, work, rwork, info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",805,"The computation of the condition number in dgecon failed")
      END IF

      IF (r_cond < EPSILON(0.0_dp)) THEN
         WRITE (message, "(A,ES10.3)") "R_COND =", r_cond
         CALL cp_abort(cp__l("common/mathlib.F",810), &
                       "Bad condition number "//TRIM(message)//" (smaller than the machine "// &
                       "working precision)")
      END IF

      ! Solve a system of linear equations using the LU factorization computed by dgetrf

      CALL unit_matrix(a_inverse)

      CALL zgetrs(trans, n, n, a_lu, n, ipiv, a_inverse, n, info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",822,"Solving the system of linear equations in dgetrs failed")
      END IF

      ! Improve the computed solution iteratively
      CALL unit_matrix(b) ! Initialize right-hand sides

      eval_error = 0.0_dp

      DO iter = 1, 10

         CALL zgerfs(trans, n, n, a, n, a_lu, n, ipiv, b, n, a_inverse, n, ferr, berr, &
                     work, rwork, info)

         IF (info /= 0) THEN
            CALL cp__b("common/mathlib.F",836,"Improving the computed solution in dgerfs failed")
         END IF

         old_eval_error = eval_error
         eval_error = MAXVAL(ferr)

         IF (ABS(eval_error-old_eval_error) <= EPSILON(1.0_dp)) EXIT

      END DO

      ! Release work storage
      DEALLOCATE (work)
      DEALLOCATE (rwork)
      DEALLOCATE (ipiv)
      DEALLOCATE (ferr)
      DEALLOCATE (berr)
      DEALLOCATE (b)
      DEALLOCATE (a_lu)

   END SUBROUTINE invert_matrix_z

! **************************************************************************************************
!> \brief returns the pseudoinverse of a real, square matrix using singular
!>         value decomposition
!> \param a matrix a
!> \param a_pinverse pseudoinverse of matrix a
!> \param rskip parameter for setting small singular values to zero
!> \author Dorothea Golze [02.2015]
! **************************************************************************************************
   SUBROUTINE get_pseudo_inverse_svd(a, a_pinverse, rskip)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a, a_pinverse
      REAL(KIND=dp), INTENT(IN)                          :: rskip

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_pseudo_inverse_svd', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, info, lwork, n
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iwork
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: sig, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: sig_plus, temp_mat, u, vt

      CALL timeset(routineN, handle)

      n = SIZE(a, 1)
      ALLOCATE (u(n, n), vt(n, n), sig(n), sig_plus(n, n), iwork(8*n), work(1), temp_mat(n, n))
      u(:, :) = 0.0_dp
      vt(:, :) = 0.0_dp
      sig(:) = 0.0_dp
      sig_plus = 0.0_dp
      work = 0.0_dp
      iwork = 0

      ! work size query
      lwork = -1
      CALL dgesdd('A', n, n, a(1, 1), n, sig(1), u(1, 1), n, vt(1, 1), n, work(1), &
                  lwork, iwork(1), info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",895,"ERROR in DGESDD: Could not retrieve work array sizes")
      END IF
      lwork = INT(work(1))
      DEALLOCATE (work)
      ALLOCATE (work(lwork))

      ! do SVD
      CALL dgesdd('A', n, n, a(1, 1), n, sig(1), u(1, 1), n, vt(1, 1), n, work(1), &
                  lwork, iwork(1), info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",906,"SVD failed")
      END IF

      ! set singular values that are too small to zero
      DO i = 1, n
         IF (sig(i) > rskip*MAXVAL(sig)) THEN
            sig_plus(i, i) = 1._dp/sig(i)
         ELSE
            sig_plus(i, i) = 0.0_dp
         ENDIF
      ENDDO

      ! build pseudoinverse: V*sig_plus*UT
      CALL dgemm("N", "T", n, n, n, 1._dp, sig_plus, n, u, n, 0._dp, temp_mat, n)
      CALL dgemm("T", "N", n, n, n, 1._dp, vt, n, temp_mat, n, 0._dp, a_pinverse, n)

      DEALLOCATE (u, vt, sig, iwork, work, sig_plus, temp_mat)

      CALL timestop(handle)

   END SUBROUTINE get_pseudo_inverse_svd

! **************************************************************************************************
!> \brief returns the pseudoinverse of a real, symmetric and positive definite
!>        matrix using diagonalization.
!> \param a matrix a
!> \param a_pinverse pseudoinverse of matrix a
!> \param rskip parameter for setting small eigenvalues to zero
!> \author Dorothea Golze [02.2015]
! **************************************************************************************************
   SUBROUTINE get_pseudo_inverse_diag(a, a_pinverse, rskip)

      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a, a_pinverse
      REAL(KIND=dp), INTENT(IN)                          :: rskip

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_pseudo_inverse_diag', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, info, lwork, n
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eig, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: dinv, temp_mat

      CALL timeset(routineN, handle)

      n = SIZE(a, 1)
      ALLOCATE (dinv(n, n), eig(n), work(1), temp_mat(n, n))
      dinv(:, :) = 0.0_dp
      eig(:) = 0.0_dp
      work(:) = 0.0_dp
      temp_mat = 0.0_dp

      ! work size query
      lwork = -1
      CALL dsyev('V', 'U', n, a(1, 1), n, eig(1), work(1), lwork, info)
      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",961,"ERROR in DSYEV: Could not retrieve work array sizes")
      END IF
      lwork = INT(work(1))
      DEALLOCATE (work)
      ALLOCATE (work(lwork))

      ! get eigenvalues and eigenvectors
      CALL dsyev('V', 'U', n, a(1, 1), n, eig(1), work(1), lwork, info)

      IF (info /= 0) THEN
         CALL cp__b("common/mathlib.F",971,"Matrix diagonalization failed")
      END IF

      ! set eigenvalues that are too small to zero
      DO i = 1, n
         IF (eig(i) > rskip*MAXVAL(eig)) THEN
            dinv(i, i) = 1.0_dp/eig(i)
         ELSE
            dinv(i, i) = 0._dp
         ENDIF
      ENDDO

      ! build pseudoinverse: U*dinv*UT
      CALL dgemm("N", "T", n, n, n, 1._dp, dinv, n, a, n, 0._dp, temp_mat, n)
      CALL dgemm("N", "N", n, n, n, 1._dp, a, n, temp_mat, n, 0._dp, a_pinverse, n)

      DEALLOCATE (eig, work, dinv, temp_mat)

      CALL timestop(handle)

   END SUBROUTINE get_pseudo_inverse_diag

! **************************************************************************************************
!> \brief  Reflection of the vector a through a mirror plane defined by the
!>         normal vector b. The reflected vector a is stored in a_mirror.
!> \param a ...
!> \param b ...
!> \retval a_mirror ...
!> \date    16.10.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION reflect_vector(a, b) RESULT(a_mirror)

      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a, b
      REAL(KIND=dp), DIMENSION(3)                        :: a_mirror

      REAL(KIND=dp)                                      :: length_of_b, scapro
      REAL(KIND=dp), DIMENSION(3)                        :: d

      length_of_b = SQRT(b(1)*b(1)+b(2)*b(2)+b(3)*b(3))

      IF (length_of_b > eps_geo) THEN

         d(:) = b(:)/length_of_b

         ! Calculate the mirror image a_mirror of the vector a
         scapro = a(1)*d(1)+a(2)*d(2)+a(3)*d(3)

         a_mirror(:) = a(:)-2.0_dp*scapro*d(:)

      ELSE

         a_mirror(:) = 0.0_dp

      END IF

   END FUNCTION reflect_vector

! **************************************************************************************************
!> \brief   Rotation of the vector a about an rotation axis defined by the
!>          vector b. The rotation angle is phi (radians). The rotated vector
!>          a is stored in a_rot.
!> \param a ...
!> \param phi ...
!> \param b ...
!> \retval a_rot ...
!> \date    16.10.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   FUNCTION rotate_vector(a, phi, b) RESULT(a_rot)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      REAL(KIND=dp), INTENT(IN)                          :: phi
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: b
      REAL(KIND=dp), DIMENSION(3)                        :: a_rot

      REAL(KIND=dp)                                      :: length_of_b
      REAL(KIND=dp), DIMENSION(3, 3)                     :: rotmat

      length_of_b = SQRT(b(1)*b(1)+b(2)*b(2)+b(3)*b(3))
      IF (length_of_b > eps_geo) THEN

         ! Build up the rotation matrix rotmat
         CALL build_rotmat(phi, b, rotmat)

         ! Rotate the vector a by phi about the axis defined by vector b
         a_rot(:) = MATMUL(rotmat, a)

      ELSE

         a_rot(:) = 0.0_dp

      END IF

   END FUNCTION rotate_vector

! **************************************************************************************************
!> \brief   Set the diagonal elements of matrix a to b.
!> \param a ...
!> \param b ...
!> \date    20.11.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_diag_scalar_d(a, b)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: a
      REAL(KIND=dp), INTENT(IN)                          :: b

      INTEGER                                            :: i, n

      n = MIN(SIZE(a, 1), SIZE(a, 2))
      DO i = 1, n
         a(i, i) = b
      END DO

   END SUBROUTINE set_diag_scalar_d

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
! **************************************************************************************************
   SUBROUTINE set_diag_scalar_z(a, b)
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(INOUT)   :: a
      COMPLEX(KIND=dp), INTENT(IN)                       :: b

      INTEGER                                            :: i, n

      n = MIN(SIZE(a, 1), SIZE(a, 2))
      DO i = 1, n
         a(i, i) = b
      END DO

   END SUBROUTINE set_diag_scalar_z

! **************************************************************************************************
!> \brief   Symmetrize the matrix a.
!> \param a ...
!> \param option ...
!> \date    16.10.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE symmetrize_matrix(a, option)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: a
      CHARACTER(LEN=*), INTENT(IN)                       :: option

      CHARACTER(len=*), PARAMETER :: routineN = 'symmetrize_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n

      n = MIN(SIZE(a, 1), SIZE(a, 2))

      IF (option == "lower_to_upper") THEN
         DO i = 1, n-1
            a(i, i+1:n) = a(i+1:n, i)
         END DO
      ELSE IF (option == "upper_to_lower") THEN
         DO i = 1, n-1
            a(i+1:n, i) = a(i, i+1:n)
         END DO
      ELSE IF (option == "anti_lower_to_upper") THEN
         DO i = 1, n-1
            a(i, i+1:n) = -a(i+1:n, i)
         END DO
      ELSE IF (option == "anti_upper_to_lower") THEN
         DO i = 1, n-1
            a(i+1:n, i) = -a(i, i+1:n)
         END DO
      ELSE
         CALL cp__b("common/mathlib.F",1143,"Invalid option <"//TRIM(option)//"> was specified for parameter #2")
      END IF

   END SUBROUTINE symmetrize_matrix

! **************************************************************************************************
!> \brief   Set the matrix a to be a unit matrix.
!> \param a ...
!> \date    16.10.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE unit_matrix_d(a)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: a

      a(:, :) = 0.0_dp
      CALL set_diag(a, 1.0_dp)

   END SUBROUTINE unit_matrix_d

! **************************************************************************************************
!> \brief ...
!> \param a ...
! **************************************************************************************************
   SUBROUTINE unit_matrix_z(a)
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(INOUT)   :: a

      a(:, :) = (0.0_dp, 0.0_dp)
      CALL set_diag(a, (1.0_dp, 0.0_dp))

   END SUBROUTINE unit_matrix_z

! **************************************************************************************************
!> \brief   Calculation of the vector product c = a x b.
!> \param a ...
!> \param b ...
!> \retval c ...
!> \date    16.10.1998
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   PURE FUNCTION vector_product(a, b) RESULT(c)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a, b
      REAL(KIND=dp), DIMENSION(3)                        :: c

      c(1) = a(2)*b(3)-a(3)*b(2)
      c(2) = a(3)*b(1)-a(1)*b(3)
      c(3) = a(1)*b(2)-a(2)*b(1)

   END FUNCTION vector_product

! **************************************************************************************************
!> \brief computes the greatest common divisor of two number
!> \param a ...
!> \param b ...
!> \retval gcd ...
!> \author Joost VandeVondele
! **************************************************************************************************
   ELEMENTAL FUNCTION gcd(a, b)
      INTEGER, INTENT(IN)                                :: a, b
      INTEGER                                            :: gcd

      INTEGER                                            :: aa, ab, l, rem, s

      aa = ABS(a)
      ab = ABS(b)
      IF (aa < ab) THEN
         s = aa
         l = ab
      ELSE
         s = ab
         l = aa
      ENDIF
      IF (s .NE. 0) THEN
         DO
            rem = MOD(l, s)
            IF (rem == 0) EXIT
            l = s
            s = rem
         ENDDO
         GCD = s
      ELSE
         GCD = l
      ENDIF
   END FUNCTION gcd

! **************************************************************************************************
!> \brief computes the least common multiplier of two numbers
!> \param a ...
!> \param b ...
!> \retval lcm ...
!> \author Joost VandeVondele
! **************************************************************************************************
   ELEMENTAL FUNCTION lcm(a, b)
      INTEGER, INTENT(IN)                                :: a, b
      INTEGER                                            :: lcm

      INTEGER                                            :: tmp

      tmp = gcd(a, b)
      IF (tmp == 0) THEN
         lcm = 0
      ELSE
         ! could still overflow if the true lcm is larger than maxint
         lcm = ABS((a/tmp)*b)
      END IF
   END FUNCTION lcm

! **************************************************************************************************
!> \brief computes the exponential integral
!>      Ei(x) = Int(exp(-x*t)/t,t=1..infinity)  x>0
!> \param x ...
!> \retval ei ...
!> \author JGH (adapted from Numerical recipies)
! **************************************************************************************************
   FUNCTION ei(x)
      REAL(dp)                                           :: x, ei

      CHARACTER(len=*), PARAMETER :: routineN = 'ei', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: maxit = 100
      REAL(dp), PARAMETER                                :: eps = EPSILON(0.0_dp), &
                                                            fpmin = TINY(0.0_dp)

      INTEGER                                            :: k
      REAL(dp)                                           :: fact, prev, sum1, term

      IF (x <= 0._dp) THEN
         CALL cp__b("common/mathlib.F",1270,"Invalid argument")
      END IF

      IF (x < fpmin) THEN
         ei = LOG(x)+euler
      ELSE IF (x <= -LOG(EPS)) THEN
         sum1 = 0._dp
         fact = 1._dp
         DO k = 1, maxit
            fact = fact*x/REAL(k, dp)
            term = fact/REAL(k, dp)
            sum1 = sum1+term
            IF (term < eps*sum1) EXIT
         END DO
         ei = sum1+LOG(x)+euler
      ELSE
         sum1 = 0._dp
         term = 1._dp
         DO k = 1, maxit
            prev = term
            term = term*REAL(k, dp)/x
            IF (term < eps) EXIT
            IF (term < prev) THEN
               sum1 = sum1+term
            ELSE
               sum1 = sum1-prev
               EXIT
            END IF
         END DO
         ei = EXP(x)*(1._dp+sum1)/x
      END IF

   END FUNCTION ei

! **************************************************************************************************
!> \brief ...
!> \param mat1 ...
!> \param mat2 ...
!> \retval matmul_3x3 ...
! **************************************************************************************************
   FUNCTION matmul_3x3(mat1, mat2)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: mat1, mat2
      REAL(KIND=dp), DIMENSION(3, 3)                     :: matmul_3x3

      matmul_3x3(1, 1) = mat1(1, 1)*mat2(1, 1)+mat1(1, 2)*mat2(2, 1)+mat1(1, 3)*mat2(3, 1)
      matmul_3x3(1, 2) = mat1(1, 1)*mat2(1, 2)+mat1(1, 2)*mat2(2, 2)+mat1(1, 3)*mat2(3, 2)
      matmul_3x3(1, 3) = mat1(1, 1)*mat2(1, 3)+mat1(1, 2)*mat2(2, 3)+mat1(1, 3)*mat2(3, 3)
      matmul_3x3(2, 1) = mat1(2, 1)*mat2(1, 1)+mat1(2, 2)*mat2(2, 1)+mat1(2, 3)*mat2(3, 1)
      matmul_3x3(2, 2) = mat1(2, 1)*mat2(1, 2)+mat1(2, 2)*mat2(2, 2)+mat1(2, 3)*mat2(3, 2)
      matmul_3x3(2, 3) = mat1(2, 1)*mat2(1, 3)+mat1(2, 2)*mat2(2, 3)+mat1(2, 3)*mat2(3, 3)
      matmul_3x3(3, 1) = mat1(3, 1)*mat2(1, 1)+mat1(3, 2)*mat2(2, 1)+mat1(3, 3)*mat2(3, 1)
      matmul_3x3(3, 2) = mat1(3, 1)*mat2(1, 2)+mat1(3, 2)*mat2(2, 2)+mat1(3, 3)*mat2(3, 2)
      matmul_3x3(3, 3) = mat1(3, 1)*mat2(1, 3)+mat1(3, 2)*mat2(2, 3)+mat1(3, 3)*mat2(3, 3)
   END FUNCTION matmul_3x3

! **************************************************************************************************
!> \brief ...
!> \param res ...
!> \param mat ...
!> \param vec ...
! **************************************************************************************************
   SUBROUTINE matvec_3x3(res, mat, vec)
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: res
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: mat
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: vec

      res(1) = mat(1, 1)*vec(1)+mat(1, 2)*vec(2)+mat(1, 3)*vec(3)
      res(2) = mat(2, 1)*vec(1)+mat(2, 2)*vec(2)+mat(2, 3)*vec(3)
      res(3) = mat(3, 1)*vec(1)+mat(3, 2)*vec(2)+mat(3, 3)*vec(3)
   END SUBROUTINE matvec_3x3

! **************************************************************************************************
!> \brief ...
!> \param vec1 ...
!> \param vec2 ...
!> \retval dotprod_3d ...
! **************************************************************************************************
   FUNCTION dotprod_3d(vec1, vec2)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: vec1, vec2
      REAL(KIND=dp)                                      :: dotprod_3d

      dotprod_3d = &
         vec1(1)*vec2(1) &
         +vec1(2)*vec2(2) &
         +vec1(3)*vec2(3)
   END FUNCTION dotprod_3d

! **************************************************************************************************
!> \brief ...
!> \param mat ...
!> \retval transpose_3d ...
! **************************************************************************************************
   FUNCTION transpose_3d(mat)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: mat
      REAL(KIND=dp), DIMENSION(3, 3)                     :: transpose_3d

      INTEGER                                            :: i, j

      DO i = 1, 3
         DO j = 1, 3
            transpose_3d(j, i) = mat(i, j)
         END DO
      END DO
   END FUNCTION transpose_3d

! **************************************************************************************************
!> \brief computes the exponential integral
!>      En(x) = Int(exp(-x*t)/t^n,t=1..infinity)  x>0, n=0,1,..
!>      Note: Ei(-x) = -E1(x)
!> \param n ...
!> \param x ...
!> \retval expint ...
!> \par History
!>      05.2007 Created
!> \author Manuel Guidon (adapted from Numerical recipies)
! **************************************************************************************************
   FUNCTION expint(n, x)
      INTEGER                                            :: n
      REAL(dp)                                           :: x, expint

      CHARACTER(len=*), PARAMETER :: routineN = 'expint', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: maxit = 100
      REAL(dp), PARAMETER :: eps = 6.e-14_dp, euler = 0.5772156649015328606065120_dp, &
         fpmin = TINY(0.0_dp)

      INTEGER                                            :: i, ii, nm1
      REAL(dp)                                           :: a, b, c, d, del, fact, h, psi

      nm1 = n-1

      IF (n .LT. 0 .OR. x .LT. 0.0_dp .OR. (x .EQ. 0.0_dp .AND. (n .EQ. 0 .OR. n .EQ. 1))) THEN
         CALL cp__b("common/mathlib.F",1401,"Invalid argument")
      ELSE IF (n .EQ. 0) THEN !Special case.
         expint = EXP(-x)/x
      ELSE IF (x .EQ. 0.0_dp) THEN !Another special case.
         expint = 1.0_dp/nm1
      ELSE IF (x .GT. 1.0_dp) THEN !Lentz's algorithm (5.2).
         b = x+n
         c = 1.0_dp/FPMIN
         d = 1.0_dp/b
         h = d
         DO i = 1, MAXIT
            a = -i*(nm1+i)
            b = b+2.0_dp
            d = 1.0_dp/(a*d+b)
            c = b+a/c
            del = c*d
            h = h*del
            IF (ABS(del-1.0_dp) .LT. EPS) THEN
               expint = h*EXP(-x)
               RETURN
            END IF
         END DO
         CALL cp__b("common/mathlib.F",1423,"continued fraction failed in expint")
      ELSE !Evaluate series.
         IF (nm1 .NE. 0) THEN !Set first term.
            expint = 1.0_dp/nm1
         ELSE
            expint = -LOG(x)-euler
         END IF
         fact = 1.0_dp
         DO i = 1, MAXIT
            fact = -fact*x/i
            IF (i .NE. nm1) THEN
               del = -fact/(i-nm1)
            ELSE
               psi = -euler !Compute I(n).
               DO ii = 1, nm1
                  psi = psi+1.0_dp/ii
               END DO
               del = fact*(-LOG(x)+psi)
            END IF
            expint = expint+del
            IF (ABS(del) .LT. ABS(expint)*EPS) RETURN
         END DO
         CALL cp__b("common/mathlib.F",1445,"series failed in expint")
      END IF
      RETURN
   END FUNCTION expint

! **************************************************************************************************
!> \brief  Jacobi matrix diagonalization. The eigenvalues are returned in
!>         vector d and the eigenvectors are returned in matrix v in ascending
!>         order.
!>
!> \param a ...
!> \param d ...
!> \param v ...
!> \par History
!>         - Creation (20.11.98, Matthias Krack)
! **************************************************************************************************
   SUBROUTINE jacobi(a, d, v)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: a
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: d
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: v

      INTEGER                                            :: n

      n = SIZE(d(:))

      ! Diagonalize matrix a
      CALL diag(n, a, d, v)

      ! Sort eigenvalues and eigenvector in ascending order
      CALL eigsrt(n, d, v)

   END SUBROUTINE jacobi

! **************************************************************************************************
!> \brief  Diagonalize matrix a. The eigenvalues are returned in vector d
!>         and the eigenvectors are returned in matrix v.
!>
!> \param n ...
!> \param a ...
!> \param d ...
!> \param v ...
!> \par History
!>         - Creation (20.11.98, Matthias Krack)
! **************************************************************************************************
   SUBROUTINE diag(n, a, d, v)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(n, n), INTENT(INOUT)      :: a
      REAL(KIND=dp), DIMENSION(n), INTENT(OUT)           :: d
      REAL(KIND=dp), DIMENSION(n, n), INTENT(OUT)        :: v

      INTEGER                                            :: i, ip, iq
      REAL(KIND=dp)                                      :: a_max, c, d_min, g, h, s, t, tau, theta, &
                                                            tresh
      REAL(KIND=dp), DIMENSION(n)                        :: b, z

      CALL unit_matrix(v(:, :))

      b(:) = get_diag(a(:, :))
      d(:) = b(:)
      z(:) = 0.0_dp

      ! Go for 50 iterations
      DO i = 1, 50
         d_min = MAX(1.0E-3_dp, MINVAL(ABS(d(:))))
         a_max = 0.0_dp
         DO ip = 1, n-1
            a_max = MAX(a_max, MAXVAL(ABS(a(ip, ip+1:n))))
         END DO
         IF (a_max < 1.0E-10_dp*d_min) RETURN
         tresh = MERGE(a_max, 0.0_dp, (i < 4))
         DO ip = 1, n-1
            DO iq = ip+1, n
               g = 100.0_dp*ABS(a(ip, iq))
               IF ((i > 4) .AND. &
                   ((ABS(d(ip))+g) == ABS(d(ip))) .AND. &
                   ((ABS(d(iq))+g) == ABS(d(iq)))) THEN
                  a(ip, iq) = 0.0_dp
               ELSE IF (ABS(a(ip, iq)) > tresh) THEN
                  h = d(iq)-d(ip)
                  IF ((ABS(h)+g) == ABS(h)) THEN
                     t = a(ip, iq)/h
                  ELSE
                     theta = 0.5_dp*h/a(ip, iq)
                     t = 1.0_dp/(ABS(theta)+SQRT(1.0_dp+theta**2))
                     IF (theta < 0.0_dp) t = -t
                  END IF
                  c = 1.0_dp/SQRT(1.0_dp+t**2)
                  s = t*c
                  tau = s/(1.0_dp+c)
                  h = t*a(ip, iq)
                  z(ip) = z(ip)-h
                  z(iq) = z(iq)+h
                  d(ip) = d(ip)-h
                  d(iq) = d(iq)+h
                  a(ip, iq) = 0.0_dp
                  CALL jrotate(a(1:ip-1, ip), a(1:ip-1, iq), s, tau)
                  CALL jrotate(a(ip, ip+1:iq-1), a(ip+1:iq-1, iq), s, tau)
                  CALL jrotate(a(ip, iq+1:n), a(iq, iq+1:n), s, tau)
                  CALL jrotate(v(:, ip), v(:, iq), s, tau)
               END IF
            END DO
         END DO
         b(:) = b(:)+z(:)
         d(:) = b(:)
         z(:) = 0.0_dp
      END DO
      WRITE (*, '(/,T2,A,/)') 'Too many iterations in jacobi'

   END SUBROUTINE diag

! **************************************************************************************************
!> \brief  Perform a Jacobi rotation of the vectors a and b.
!>
!> \param a ...
!> \param b ...
!> \param ss ...
!> \param tt ...
!> \par History
!>         - Creation (20.11.98, Matthias Krack)
! **************************************************************************************************
   SUBROUTINE jrotate(a, b, ss, tt)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: a, b
      REAL(KIND=dp), INTENT(IN)                          :: ss, tt

      REAL(KIND=dp), DIMENSION(SIZE(a))                  :: c

      c(:) = a(:)
      a(:) = a(:)-ss*(b(:)+a(:)*tt)
      b(:) = b(:)+ss*(c(:)-b(:)*tt)

   END SUBROUTINE jrotate

! **************************************************************************************************
!> \brief Sort the values in vector d in ascending order and swap the
!>        corresponding columns of matrix v.
!>
!> \param n ...
!> \param d ...
!> \param v ...
!> \par History
!>         - Creation (20.11.98, Matthias Krack)
! **************************************************************************************************
   SUBROUTINE eigsrt(n, d, v)
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: d
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: v

      INTEGER                                            :: i, j

      DO i = 1, n-1
         j = SUM(MINLOC(d(i:n)))+i-1
         IF (j /= i) THEN
            CALL swap(d(i), d(j))
            CALL swap(v(:, i), v(:, j))
         END IF
      END DO

   END SUBROUTINE eigsrt

! **************************************************************************
!> \brief Swap two scalars
!>
!> \param a ...
!> \param b ...
!> \par History
!>         - Creation (20.11.98, Matthias Krack)
! **************************************************************************************************
   SUBROUTINE swap_scalar(a, b)
      REAL(KIND=dp), INTENT(INOUT)                       :: a, b

      REAL(KIND=dp)                                      :: c

      c = a
      a = b
      b = c

   END SUBROUTINE swap_scalar

! **************************************************************************
!> \brief Swap two vectors
!>
!> \param a ...
!> \param b ...
!> \par History
!>         - Creation (20.11.98, Matthias Krack)
! **************************************************************************************************
   SUBROUTINE swap_vector(a, b)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: a, b

      CHARACTER(LEN=*), PARAMETER :: routineN = 'swap_vector', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n
      REAL(KIND=dp)                                      :: c

      n = SIZE(a)

      IF (n /= SIZE(b)) THEN
         CALL cp__b("common/mathlib.F",1643,"Check the array bounds of the parameters")
      END IF

      DO i = 1, n
         c = a(i)
         a(i) = b(i)
         b(i) = c
      END DO

   END SUBROUTINE swap_vector

END MODULE mathlib
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief methods related to the blacs parallel environment
!> \par History
!>      08.2002 created [fawzi]
!>      02.2004 modified to associate a blacs_env with a given para_env
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_blacs_env
   USE cp_array_utils_i,                ONLY: cp_2d_i_write
   USE cp_blacs_calls,                  ONLY: cp_blacs_gridexit,&
                                              cp_blacs_gridinfo,&
                                              cp_blacs_gridinit,&
                                              cp_blacs_set
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush
   USE mathlib,                         ONLY: gcd
   USE message_passing,                 ONLY: mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_blacs_env'

   ! Blacs type of distribution
   INTEGER, PARAMETER, PUBLIC               :: BLACS_GRID_SQUARE = 1, &
                                               BLACS_GRID_ROW = 2, &
                                               BLACS_GRID_COL = 3

   PUBLIC :: cp_blacs_env_type ! make it accessible only through cp_para_types?
   PUBLIC :: cp_blacs_env_create, cp_blacs_env_retain, cp_blacs_env_release
   PUBLIC :: cp_blacs_env_write, get_blacs_info

! **************************************************************************************************
!> \brief represent a blacs multidimensional parallel environment
!>      (for the mpi corrispective see cp_paratypes/cp_para_cart_type)
!> \param mepos the position of the actual processor (2D)
!> \param group id of the actual group (context, communicator)
!> \param num_pe number of processors in the group in each dimension
!> \param ref_count the reference count, when it is zero this object gets
!>        deallocated
!> \param my_pid process id of the actual processor
!> \param n_pid number of process ids
!> \param the para env associated (and compatible) with this blacs_env
!> \param blacs 2mpi: maps mepos(1)-mepos(2) of blacs to its mpi rank
!> \param mpi 2blacs(i,rank): maps the mpi rank to the mepos(i)
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_blacs_env_type
      INTEGER, DIMENSION(2) :: mepos, num_pe
      INTEGER :: group, my_pid, n_pid, ref_count
      TYPE(cp_para_env_type), POINTER :: para_env
      INTEGER, DIMENSION(:, :), POINTER :: blacs2mpi
      INTEGER, DIMENSION(:, :), POINTER :: mpi2blacs
      LOGICAL :: repeatable
   END TYPE cp_blacs_env_type

!***
CONTAINS

! **************************************************************************************************
!> \brief   Return informations about the specified BLACS context.
!> \param blacs_env ...
!> \param my_process_row ...
!> \param my_process_column ...
!> \param my_process_number ...
!> \param number_of_process_rows ...
!> \param number_of_process_columns ...
!> \param number_of_processes ...
!> \param para_env ...
!> \param blacs2mpi ...
!> \param mpi2blacs ...
!> \date    19.06.2001
!> \par     History
!>          MM.YYYY moved here from qs_blacs (Joost VandeVondele)
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_blacs_info(blacs_env, my_process_row, my_process_column, &
                             my_process_number, number_of_process_rows, &
                             number_of_process_columns, number_of_processes, &
                             para_env, blacs2mpi, mpi2blacs)
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      INTEGER, INTENT(OUT), OPTIONAL :: my_process_row, my_process_column, my_process_number, &
         number_of_process_rows, number_of_process_columns, number_of_processes
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: blacs2mpi, mpi2blacs

      CHARACTER(len=*), PARAMETER :: routineN = 'get_blacs_info', routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(blacs_env)) THEN
         CALL cp__b("fm/cp_blacs_env.F",103,"No BLACS environment")
      END IF

      IF (PRESENT(my_process_row)) my_process_row = blacs_env%mepos(1)
      IF (PRESENT(my_process_column)) my_process_column = blacs_env%mepos(2)
      IF (PRESENT(my_process_number)) my_process_number = blacs_env%my_pid
      IF (PRESENT(number_of_process_rows)) number_of_process_rows = blacs_env%num_pe(1)
      IF (PRESENT(number_of_process_columns)) number_of_process_columns = blacs_env%num_pe(2)
      IF (PRESENT(number_of_processes)) number_of_processes = blacs_env%n_pid
      IF (PRESENT(para_env)) para_env => blacs_env%para_env
      IF (PRESENT(blacs2mpi)) blacs2mpi => blacs_env%blacs2mpi
      IF (PRESENT(mpi2blacs)) mpi2blacs => blacs_env%mpi2blacs

   END SUBROUTINE get_blacs_info

! **************************************************************************************************
!> \brief allocates and initializes a type that represent a blacs context
!> \param blacs_env the type to initialize
!> \param para_env the para_env for which a blacs env should be created
!> \param blacs_grid_layout ...
!> \param blacs_repeatable ...
!> \param row_major ...
!> \param grid_2d ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_blacs_env_create(blacs_env, para_env, blacs_grid_layout, blacs_repeatable, row_major, grid_2d)
      TYPE(cp_blacs_env_type), POINTER         :: blacs_env
      TYPE(cp_para_env_type), POINTER          :: para_env
      INTEGER, INTENT(IN), OPTIONAL            :: blacs_grid_layout
      LOGICAL, INTENT(IN), OPTIONAL            :: blacs_repeatable, row_major
      INTEGER, DIMENSION(:), INTENT(IN), &
         OPTIONAL                               :: grid_2d

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_blacs_env_create', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: ipcol, iprow, stat


      IF(.NOT.(.NOT. ASSOCIATED(blacs_env)))CALL cp__a("fm/cp_blacs_env.F",158)

      ALLOCATE (blacs_env)
      blacs_env%group = 0
      blacs_env%ref_count = 1
      blacs_env%mepos(:) = 0
      blacs_env%num_pe(:) = 1
      blacs_env%my_pid = 0
      blacs_env%n_pid = 1
      CALL cp_para_env_retain(para_env)
      blacs_env%para_env => para_env

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(blacs_grid_layout))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(blacs_repeatable))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(grid_2d))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(row_major))==-1) EXIT ;  END DO ; ENDIF

      ! generate the mappings blacs2mpi and mpi2blacs
      ALLOCATE (blacs_env%blacs2mpi(0:blacs_env%num_pe(1)-1, 0:blacs_env%num_pe(2)-1), &
                stat=stat)
      IF(.NOT.(stat == 0))CALL cp__a("fm/cp_blacs_env.F",241)
      blacs_env%blacs2mpi = 0
      blacs_env%blacs2mpi(blacs_env%mepos(1), blacs_env%mepos(2)) = para_env%mepos
      CALL mp_sum(blacs_env%blacs2mpi, para_env%group)
      ALLOCATE (blacs_env%mpi2blacs(2, 0:para_env%num_pe-1))
      blacs_env%mpi2blacs = -1
      DO ipcol = 0, blacs_env%num_pe(2)-1
         DO iprow = 0, blacs_env%num_pe(1)-1
            blacs_env%mpi2blacs(1, blacs_env%blacs2mpi(iprow, ipcol)) = iprow
            blacs_env%mpi2blacs(2, blacs_env%blacs2mpi(iprow, ipcol)) = ipcol
         END DO
      END DO
   END SUBROUTINE cp_blacs_env_create

! **************************************************************************************************
!> \brief retains the given blacs env
!> \param blacs_env the blacs env to retain
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_blacs_env_retain(blacs_env)
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_blacs_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(blacs_env)))CALL cp__a("fm/cp_blacs_env.F",268)
      IF(.NOT.(blacs_env%ref_count > 0))CALL cp__a("fm/cp_blacs_env.F",269)
      blacs_env%ref_count = blacs_env%ref_count+1
   END SUBROUTINE cp_blacs_env_retain

! **************************************************************************************************
!> \brief releases the given blacs_env
!> \param blacs_env the blacs env to relase
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_blacs_env_release(blacs_env)
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_blacs_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(blacs_env)) THEN
         IF(.NOT.(blacs_env%ref_count > 0))CALL cp__a("fm/cp_blacs_env.F",287)
         blacs_env%ref_count = blacs_env%ref_count-1
         IF (blacs_env%ref_count < 1) THEN
            CALL cp_blacs_gridexit(blacs_env%group)
            CALL cp_para_env_release(blacs_env%para_env)
            DEALLOCATE (blacs_env%mpi2blacs)
            DEALLOCATE (blacs_env%blacs2mpi)
            DEALLOCATE (blacs_env)
         END IF
      END IF
      NULLIFY (blacs_env)
   END SUBROUTINE cp_blacs_env_release

! **************************************************************************************************
!> \brief writes the description of the given blacs env
!> \param blacs_env the blacs environment to write
!> \param unit_nr the unit number where to write the description of the
!>        blacs environment
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_blacs_env_write(blacs_env, unit_nr)
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      INTEGER, INTENT(in)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_blacs_env_write', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iostat

      IF (ASSOCIATED(blacs_env)) THEN
         WRITE (unit=unit_nr, fmt="('  group=',i10,', ref_count=',i10,',')", &
                iostat=iostat) blacs_env%group, blacs_env%ref_count
         IF(.NOT.(iostat == 0))CALL cp__a("fm/cp_blacs_env.F",321)
         WRITE (unit=unit_nr, fmt="('  mepos=(',i8,',',i8,'),')", &
                iostat=iostat) blacs_env%mepos(1), blacs_env%mepos(2)
         IF(.NOT.(iostat == 0))CALL cp__a("fm/cp_blacs_env.F",324)
         WRITE (unit=unit_nr, fmt="('  num_pe=(',i8,',',i8,'),')", &
                iostat=iostat) blacs_env%num_pe(1), blacs_env%num_pe(2)
         IF(.NOT.(iostat == 0))CALL cp__a("fm/cp_blacs_env.F",327)
         IF (ASSOCIATED(blacs_env%blacs2mpi)) THEN
            WRITE (unit=unit_nr, fmt="('  blacs2mpi=')", advance="no", iostat=iostat)
            IF(.NOT.(iostat == 0))CALL cp__a("fm/cp_blacs_env.F",330)
            CALL cp_2d_i_write(blacs_env%blacs2mpi, unit_nr=unit_nr)
         ELSE
            WRITE (unit=unit_nr, fmt="('  blacs2mpi=*null*')", iostat=iostat)
            IF(.NOT.(iostat == 0))CALL cp__a("fm/cp_blacs_env.F",334)
         END IF
         IF (ASSOCIATED(blacs_env%para_env)) THEN
            WRITE (unit=unit_nr, fmt="('  para_env=<cp_para_env id=',i6,'>,')") &
               blacs_env%para_env%group
         ELSE
            WRITE (unit=unit_nr, fmt="('  para_env=*null*')")
         END IF
         WRITE (unit=unit_nr, fmt="('  my_pid=',i10,', n_pid=',i10,' }')", &
                iostat=iostat) blacs_env%my_pid, blacs_env%n_pid
         IF(.NOT.(iostat == 0))CALL cp__a("fm/cp_blacs_env.F",344)
      ELSE
         WRITE (unit=unit_nr, &
                fmt="(a)", iostat=iostat) ' <cp_blacs_env>:*null* '
         IF(.NOT.(iostat == 0))CALL cp__a("fm/cp_blacs_env.F",348)
      END IF
      CALL m_flush(unit_nr)
   END SUBROUTINE cp_blacs_env_write

END MODULE cp_blacs_env
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief stores a mapping of 2D info (e.g. matrix) on a
!>      2D processor distribution (i.e. blacs grid)
!>      where cpus in the same blacs row own the same rows of the 2D info
!>      (and similar for the cols)
!> \author Joost VandeVondele (2003-08)
! **************************************************************************************************
MODULE distribution_2d_types

   USE cp_array_utils_i,                ONLY: cp_1d_i_p_type,&
                                              cp_1d_i_write
   USE cp_blacs_env,                    ONLY: cp_blacs_env_release,&
                                              cp_blacs_env_retain,&
                                              cp_blacs_env_type,&
                                              cp_blacs_env_write
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE machine,                         ONLY: m_flush
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'distribution_2d_types'
   INTEGER, SAVE, PRIVATE :: last_distribution_2d_id = 0

   PUBLIC :: distribution_2d_type

   PUBLIC :: distribution_2d_create, &
             distribution_2d_release, &
             distribution_2d_retain, &
             distribution_2d_write, &
             distribution_2d_get

! **************************************************************************************************
!> \brief distributes pairs on a 2d grid of processors
!> \param row_distribution (i): processor row that owns the row i
!> \param col_distribution (i): processor col that owns the col i
!> \param n_row_distribution nuber of global rows
!> \param n_col_distribution number of global cols
!> \param n_local_rows (ikind): number of local rows of kind ikind
!> \param n_local_cols (ikind): number of local cols of kind ikind
!> \param local_cols (ikind)%array: ordered global indexes of the local cols
!>        of kind ikind (might be oversized)
!> \param local_rows (ikind)%array: ordered global indexes of the local
!>        rows of kind ikind (might be oversized)
!> \param flat_local_rows ordered global indexes of the local rows
!>        (allocated on request, might be oversized)
!> \param flat_local_cols ordered global indexes of the local cols
!>        (allocated on request, might be oversized)
!> \param blacs_env parallel environment in which the pairs are distributed
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param id_nr identification number (unique)
!> \par History
!>      08.2003 created [joost]
!>      09.2003 kind separation, minor cleanup [fawzi]
!> \author Joost & Fawzi
! **************************************************************************************************
   TYPE distribution_2d_type
      INTEGER, DIMENSION(:, :), POINTER     :: row_distribution
      INTEGER, DIMENSION(:, :), POINTER     :: col_distribution
      INTEGER                              :: n_row_distribution
      INTEGER                              :: n_col_distribution
      INTEGER, DIMENSION(:), POINTER       :: n_local_rows
      INTEGER, DIMENSION(:), POINTER       :: n_local_cols
      TYPE(cp_1d_i_p_type), DIMENSION(:), POINTER :: local_rows
      TYPE(cp_1d_i_p_type), DIMENSION(:), POINTER :: local_cols
      INTEGER, DIMENSION(:), POINTER       :: flat_local_rows
      INTEGER, DIMENSION(:), POINTER       :: flat_local_cols
      TYPE(cp_blacs_env_type), POINTER     :: blacs_env
      INTEGER                              :: ref_count
      INTEGER                              :: id_nr
   END TYPE distribution_2d_type

CONTAINS

! **************************************************************************************************
!> \brief initializes the distribution_2d
!> \param distribution_2d ...
!> \param blacs_env ...
!> \param local_rows_ptr ...
!> \param n_local_rows ...
!> \param local_cols_ptr ...
!> \param row_distribution_ptr 2D array, first is atom to processor 2nd is
!>                             atom to cluster
!> \param col_distribution_ptr ...
!> \param n_local_cols ...
!> \param n_row_distribution ...
!> \param n_col_distribution ...
!> \par History
!>      09.2003 rewamped [fawzi]
!> \author Joost VandeVondele
!> \note
!>      the row and col_distribution are not allocated if not given
! **************************************************************************************************
   SUBROUTINE distribution_2d_create(distribution_2d, blacs_env, &
                                     local_rows_ptr, n_local_rows, &
                                     local_cols_ptr, row_distribution_ptr, col_distribution_ptr, &
                                     n_local_cols, n_row_distribution, n_col_distribution)
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_1d_i_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: local_rows_ptr
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: n_local_rows
      TYPE(cp_1d_i_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: local_cols_ptr
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: row_distribution_ptr, &
                                                            col_distribution_ptr
      INTEGER, DIMENSION(:), INTENT(in), OPTIONAL        :: n_local_cols
      INTEGER, INTENT(in), OPTIONAL                      :: n_row_distribution, n_col_distribution

      CHARACTER(len=*), PARAMETER :: routineN = 'distribution_2d_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(blacs_env)))CALL cp__a("distribution_2d_types.F",122)
      IF(.NOT.(.NOT. ASSOCIATED(distribution_2d)))CALL cp__a("distribution_2d_types.F",123)

      ALLOCATE (distribution_2d)
      distribution_2d%ref_count = 1
      last_distribution_2d_id = last_distribution_2d_id+1
      distribution_2d%id_nr = last_distribution_2d_id

      NULLIFY (distribution_2d%col_distribution, distribution_2d%row_distribution, &
               distribution_2d%local_rows, distribution_2d%local_cols, &
               distribution_2d%blacs_env, distribution_2d%n_local_cols, &
               distribution_2d%n_local_rows, distribution_2d%flat_local_rows, &
               distribution_2d%flat_local_cols)

      distribution_2d%n_col_distribution = -HUGE(0)
      IF (PRESENT(col_distribution_ptr)) THEN
         distribution_2d%col_distribution => col_distribution_ptr
         distribution_2d%n_col_distribution = SIZE(distribution_2d%col_distribution, 1)
      END IF
      IF (PRESENT(n_col_distribution)) THEN
         IF (ASSOCIATED(distribution_2d%col_distribution)) THEN
            IF (n_col_distribution > distribution_2d%n_col_distribution) &
               CALL cp__b("distribution_2d_types.F",144,"n_col_distribution<=distribution_2d%n_col_distribution")
            ! else alloc col_distribution?
         END IF
         distribution_2d%n_col_distribution = n_col_distribution
      END IF
      distribution_2d%n_row_distribution = -HUGE(0)
      IF (PRESENT(row_distribution_ptr)) THEN
         distribution_2d%row_distribution => row_distribution_ptr
         distribution_2d%n_row_distribution = SIZE(distribution_2d%row_distribution, 1)
      END IF
      IF (PRESENT(n_row_distribution)) THEN
         IF (ASSOCIATED(distribution_2d%row_distribution)) THEN
            IF (n_row_distribution > distribution_2d%n_row_distribution) &
               CALL cp__b("distribution_2d_types.F",157,"n_row_distribution<=distribution_2d%n_row_distribution")
            ! else alloc row_distribution?
         END IF
         distribution_2d%n_row_distribution = n_row_distribution
      END IF

      IF (PRESENT(local_rows_ptr)) &
         distribution_2d%local_rows => local_rows_ptr
      IF (.NOT. ASSOCIATED(distribution_2d%local_rows)) THEN
         IF(.NOT.(PRESENT(n_local_rows)))CALL cp__a("distribution_2d_types.F",166)
         ALLOCATE (distribution_2d%local_rows(SIZE(n_local_rows)))
         DO i = 1, SIZE(distribution_2d%local_rows)
            ALLOCATE (distribution_2d%local_rows(i)%array(n_local_rows(i)))
            distribution_2d%local_rows(i)%array = -HUGE(0)
         END DO
      END IF
      ALLOCATE (distribution_2d%n_local_rows(SIZE(distribution_2d%local_rows)))
      IF (PRESENT(n_local_rows)) THEN
         IF (SIZE(distribution_2d%n_local_rows) /= SIZE(n_local_rows)) &
            CALL cp__b("distribution_2d_types.F",176,"SIZE(distribution_2d%n_local_rows)==SIZE(n_local_rows)")
         DO i = 1, SIZE(distribution_2d%n_local_rows)
            IF (SIZE(distribution_2d%local_rows(i)%array) < n_local_rows(i)) &
               CALL cp__b("distribution_2d_types.F",179,"SIZE(distribution_2d%local_rows(i)%array)>=n_local_rows(i)")
            distribution_2d%n_local_rows(i) = n_local_rows(i)
         END DO
      ELSE
         DO i = 1, SIZE(distribution_2d%n_local_rows)
            distribution_2d%n_local_rows(i) = &
               SIZE(distribution_2d%local_rows(i)%array)
         END DO
      END IF

      IF (PRESENT(local_cols_ptr)) &
         distribution_2d%local_cols => local_cols_ptr
      IF (.NOT. ASSOCIATED(distribution_2d%local_cols)) THEN
         IF(.NOT.(PRESENT(n_local_cols)))CALL cp__a("distribution_2d_types.F",192)
         ALLOCATE (distribution_2d%local_cols(SIZE(n_local_cols)))
         DO i = 1, SIZE(distribution_2d%local_cols)
            ALLOCATE (distribution_2d%local_cols(i)%array(n_local_cols(i)))
            distribution_2d%local_cols(i)%array = -HUGE(0)
         END DO
      END IF
      ALLOCATE (distribution_2d%n_local_cols(SIZE(distribution_2d%local_cols)))
      IF (PRESENT(n_local_cols)) THEN
         IF (SIZE(distribution_2d%n_local_cols) /= SIZE(n_local_cols)) &
            CALL cp__b("distribution_2d_types.F",202,"SIZE(distribution_2d%n_local_cols)==SIZE(n_local_cols)")
         DO i = 1, SIZE(distribution_2d%n_local_cols)
            IF (SIZE(distribution_2d%local_cols(i)%array) < n_local_cols(i)) &
               CALL cp__b("distribution_2d_types.F",205,"SIZE(distribution_2d%local_cols(i)%array)>=n_local_cols(i)")
            distribution_2d%n_local_cols(i) = n_local_cols(i)
         END DO
      ELSE
         DO i = 1, SIZE(distribution_2d%n_local_cols)
            distribution_2d%n_local_cols(i) = &
               SIZE(distribution_2d%local_cols(i)%array)
         END DO
      END IF

      distribution_2d%blacs_env => blacs_env
      CALL cp_blacs_env_retain(distribution_2d%blacs_env)

   END SUBROUTINE distribution_2d_create

! **************************************************************************************************
!> \brief ...
!> \param distribution_2d ...
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE distribution_2d_retain(distribution_2d)
      TYPE(distribution_2d_type), POINTER                :: distribution_2d

      CHARACTER(len=*), PARAMETER :: routineN = 'distribution_2d_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(distribution_2d)))CALL cp__a("distribution_2d_types.F",231)
      IF(.NOT.(distribution_2d%ref_count > 0))CALL cp__a("distribution_2d_types.F",232)
      distribution_2d%ref_count = distribution_2d%ref_count+1
   END SUBROUTINE distribution_2d_retain

! **************************************************************************************************
!> \brief ...
!> \param distribution_2d ...
! **************************************************************************************************
   SUBROUTINE distribution_2d_release(distribution_2d)
      TYPE(distribution_2d_type), POINTER                :: distribution_2d

      CHARACTER(len=*), PARAMETER :: routineN = 'distribution_2d_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(distribution_2d)) THEN
         IF(.NOT.(distribution_2d%ref_count > 0))CALL cp__a("distribution_2d_types.F",249)
         distribution_2d%ref_count = distribution_2d%ref_count-1
         IF (distribution_2d%ref_count == 0) THEN
            CALL cp_blacs_env_release(distribution_2d%blacs_env)
            IF (ASSOCIATED(distribution_2d%col_distribution)) THEN
               DEALLOCATE (distribution_2d%col_distribution)
            END IF
            IF (ASSOCIATED(distribution_2d%row_distribution)) THEN
               DEALLOCATE (distribution_2d%row_distribution)
            END IF
            DO i = 1, SIZE(distribution_2d%local_rows)
               DEALLOCATE (distribution_2d%local_rows(i)%array)
            END DO
            DEALLOCATE (distribution_2d%local_rows)
            DO i = 1, SIZE(distribution_2d%local_cols)
               DEALLOCATE (distribution_2d%local_cols(i)%array)
            END DO
            DEALLOCATE (distribution_2d%local_cols)
            IF (ASSOCIATED(distribution_2d%flat_local_rows)) THEN
               DEALLOCATE (distribution_2d%flat_local_rows)
            END IF
            IF (ASSOCIATED(distribution_2d%flat_local_cols)) THEN
               DEALLOCATE (distribution_2d%flat_local_cols)
            END IF
            IF (ASSOCIATED(distribution_2d%n_local_rows)) THEN
               DEALLOCATE (distribution_2d%n_local_rows)
            END IF
            IF (ASSOCIATED(distribution_2d%n_local_cols)) THEN
               DEALLOCATE (distribution_2d%n_local_cols)
            END IF
            DEALLOCATE (distribution_2d)
         ENDIF
      ENDIF
      NULLIFY (distribution_2d)
   END SUBROUTINE distribution_2d_release

! **************************************************************************************************
!> \brief writes out the given distribution
!> \param distribution_2d the distribution to write out
!> \param unit_nr the unit to write to
!> \param local if the unit is local to to each processor (otherwise
!>        only the processor with logger%para_env%source==
!>        logger%para_env%mepos writes), defaults to false.
!> \param long_description if a long description should be given,
!>        defaults to false
!> \par History
!>      08.2003 adapted qs_distribution_2d_create write done by Matthias[fawzi]
!> \author Fawzi Mohamed
!> \note
!>      to clean up, make safer wrt. grabage in distribution_2d%n_*
! **************************************************************************************************
   SUBROUTINE distribution_2d_write(distribution_2d, unit_nr, local, &
                                    long_description)
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      INTEGER, INTENT(in)                                :: unit_nr
      LOGICAL, INTENT(in), OPTIONAL                      :: local, long_description

      CHARACTER(len=*), PARAMETER :: routineN = 'distribution_2d_write', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_local, my_long_description
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()

      my_long_description = .FALSE.
      IF (PRESENT(long_description)) my_long_description = long_description
      my_local = .FALSE.
      IF (PRESENT(local)) my_local = local
      IF (.NOT. my_local) my_local = (logger%para_env%source == logger%para_env%mepos)

      IF (ASSOCIATED(distribution_2d)) THEN
         IF (my_local) THEN
            WRITE (unit=unit_nr, &
                   fmt="(/,' <distribution_2d> {      id_nr=',i10,'      ref_count=',i10,',')") &
               distribution_2d%id_nr, distribution_2d%ref_count

            WRITE (unit=unit_nr, fmt="('    n_row_distribution=',i15,',')") &
               distribution_2d%n_row_distribution
            IF (ASSOCIATED(distribution_2d%row_distribution)) THEN
               IF (my_long_description) THEN
                  WRITE (unit=unit_nr, fmt="('      row_distribution= (')", advance="no")
                  DO i = 1, SIZE(distribution_2d%row_distribution, 1)
                     WRITE (unit=unit_nr, fmt="(i6,',')", advance="no") distribution_2d%row_distribution(i, 1)
                     ! keep lines finite, so that we can open outputs in vi
                     IF (MODULO(i, 8) == 0 .AND. i .NE. SIZE(distribution_2d%row_distribution, 1)) &
                        WRITE (unit=unit_nr, fmt='()')
                  END DO
                  WRITE (unit=unit_nr, fmt="('),')")
               ELSE
                  WRITE (unit=unit_nr, fmt="('      row_distribution= array(',i6,':',i6,'),')") &
                     LBOUND(distribution_2d%row_distribution(:, 1)), &
                     UBOUND(distribution_2d%row_distribution(:, 1))
               END IF
            ELSE
               WRITE (unit=unit_nr, fmt="('       row_distribution=*null*,')")
            END IF

            WRITE (unit=unit_nr, fmt="('    n_col_distribution=',i15,',')") &
               distribution_2d%n_col_distribution
            IF (ASSOCIATED(distribution_2d%col_distribution)) THEN
               IF (my_long_description) THEN
                  WRITE (unit=unit_nr, fmt="('      col_distribution= (')", advance="no")
                  DO i = 1, SIZE(distribution_2d%col_distribution, 1)
                     WRITE (unit=unit_nr, fmt="(i6,',')", advance="no") distribution_2d%col_distribution(i, 1)
                     ! keep lines finite, so that we can open outputs in vi
                     IF (MODULO(i, 8) == 0 .AND. i .NE. SIZE(distribution_2d%col_distribution, 1)) &
                        WRITE (unit=unit_nr, fmt='()')
                  END DO
                  WRITE (unit=unit_nr, fmt="('),')")
               ELSE
                  WRITE (unit=unit_nr, fmt="('      col_distribution= array(',i6,':',i6,'),')") &
                     LBOUND(distribution_2d%col_distribution(:, 1)), &
                     UBOUND(distribution_2d%col_distribution(:, 1))
               END IF
            ELSE
               WRITE (unit=unit_nr, fmt="('       col_distribution=*null*,')")
            END IF

            IF (ASSOCIATED(distribution_2d%n_local_rows)) THEN
               IF (my_long_description) THEN
                  WRITE (unit=unit_nr, fmt="('    n_local_rows= (')", advance="no")
                  DO i = 1, SIZE(distribution_2d%n_local_rows)
                     WRITE (unit=unit_nr, fmt="(i6,',')", advance="no") distribution_2d%n_local_rows(i)
                     ! keep lines finite, so that we can open outputs in vi
                     IF (MODULO(i, 10) == 0 .AND. i .NE. SIZE(distribution_2d%n_local_rows)) &
                        WRITE (unit=unit_nr, fmt='()')
                  END DO
                  WRITE (unit=unit_nr, fmt="('),')")
               ELSE
                  WRITE (unit=unit_nr, fmt="('    n_local_rows= array(',i6,':',i6,'),')") &
                     LBOUND(distribution_2d%n_local_rows), &
                     UBOUND(distribution_2d%n_local_rows)
               END IF
            ELSE
               WRITE (unit=unit_nr, fmt="('    n_local_rows=*null*,')")
            END IF

            IF (ASSOCIATED(distribution_2d%local_rows)) THEN
               WRITE (unit=unit_nr, fmt="('      local_rows=(')")
               DO i = 1, SIZE(distribution_2d%local_rows)
                  IF (ASSOCIATED(distribution_2d%local_rows(i)%array)) THEN
                     IF (my_long_description) THEN
                        CALL cp_1d_i_write(array=distribution_2d%local_rows(i)%array, &
                                           unit_nr=unit_nr)
                     ELSE
                        WRITE (unit=unit_nr, fmt="(' array(',i6,':',i6,'),')") &
                           LBOUND(distribution_2d%local_rows(i)%array), &
                           UBOUND(distribution_2d%local_rows(i)%array)
                     END IF
                  ELSE
                     WRITE (unit=unit_nr, fmt="('*null*')")
                  END IF
               END DO
               WRITE (unit=unit_nr, fmt="(' ),')")
            ELSE
               WRITE (unit=unit_nr, fmt="('      local_rows=*null*,')")
            END IF

            IF (ASSOCIATED(distribution_2d%n_local_cols)) THEN
               IF (my_long_description) THEN
                  WRITE (unit=unit_nr, fmt="('    n_local_cols= (')", advance="no")
                  DO i = 1, SIZE(distribution_2d%n_local_cols)
                     WRITE (unit=unit_nr, fmt="(i6,',')", advance="no") distribution_2d%n_local_cols(i)
                     ! keep lines finite, so that we can open outputs in vi
                     IF (MODULO(i, 10) == 0 .AND. i .NE. SIZE(distribution_2d%n_local_cols)) &
                        WRITE (unit=unit_nr, fmt='()')
                  END DO
                  WRITE (unit=unit_nr, fmt="('),')")
               ELSE
                  WRITE (unit=unit_nr, fmt="('    n_local_cols= array(',i6,':',i6,'),')") &
                     LBOUND(distribution_2d%n_local_cols), &
                     UBOUND(distribution_2d%n_local_cols)
               END IF
            ELSE
               WRITE (unit=unit_nr, fmt="('    n_local_cols=*null*,')")
            END IF

            IF (ASSOCIATED(distribution_2d%local_cols)) THEN
               WRITE (unit=unit_nr, fmt="('      local_cols=(')")
               DO i = 1, SIZE(distribution_2d%local_cols)
                  IF (ASSOCIATED(distribution_2d%local_cols(i)%array)) THEN
                     IF (my_long_description) THEN
                        CALL cp_1d_i_write(array=distribution_2d%local_cols(i)%array, &
                                           unit_nr=unit_nr)
                     ELSE
                        WRITE (unit=unit_nr, fmt="(' array(',i6,':',i6,'),')") &
                           LBOUND(distribution_2d%local_cols(i)%array), &
                           UBOUND(distribution_2d%local_cols(i)%array)
                     END IF
                  ELSE
                     WRITE (unit=unit_nr, fmt="('*null*')")
                  END IF
               END DO
               WRITE (unit=unit_nr, fmt="(' ),')")
            ELSE
               WRITE (unit=unit_nr, fmt="('      local_cols=*null*,')")
            END IF

            IF (ASSOCIATED(distribution_2d%blacs_env)) THEN
               IF (my_long_description) THEN
                  WRITE (unit=unit_nr, fmt="('    blacs_env=')", advance="no")
                  CALL cp_blacs_env_write(distribution_2d%blacs_env, unit_nr=unit_nr)
               ELSE
                  WRITE (unit=unit_nr, fmt="('    blacs_env=<blacs_env id=',i6,'>')") &
                     distribution_2d%blacs_env%group
               END IF
            ELSE
               WRITE (unit=unit_nr, fmt="('    blacs_env=*null*')")
            END IF

            WRITE (unit=unit_nr, fmt="(' }')")
         END IF

      ELSE IF (my_local) THEN
         WRITE (unit=unit_nr, &
                fmt="(' <distribution_2d *null*>')")
      END IF

      CALL m_flush(unit_nr)

   END SUBROUTINE distribution_2d_write

! **************************************************************************************************
!> \brief returns various attributes about the distribution_2d
!> \param distribution_2d the object you want info about
!> \param row_distribution ...
!> \param col_distribution ...
!> \param n_row_distribution ...
!> \param n_col_distribution ...
!> \param n_local_rows ...
!> \param n_local_cols ...
!> \param local_rows ...
!> \param local_cols ...
!> \param flat_local_rows ...
!> \param flat_local_cols ...
!> \param n_flat_local_rows ...
!> \param n_flat_local_cols ...
!> \param blacs_env ...
!> \param id_nr ...
!> \par History
!>      09.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE distribution_2d_get(distribution_2d, row_distribution, &
                                  col_distribution, n_row_distribution, n_col_distribution, &
                                  n_local_rows, n_local_cols, local_rows, local_cols, &
                                  flat_local_rows, flat_local_cols, n_flat_local_rows, n_flat_local_cols, &
                                  blacs_env, id_nr)
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: row_distribution, col_distribution
      INTEGER, INTENT(out), OPTIONAL                     :: n_row_distribution, n_col_distribution
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: n_local_rows, n_local_cols
      TYPE(cp_1d_i_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: local_rows, local_cols
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: flat_local_rows, flat_local_cols
      INTEGER, INTENT(out), OPTIONAL                     :: n_flat_local_rows, n_flat_local_cols
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env
      INTEGER, INTENT(out), OPTIONAL                     :: id_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'distribution_2d_get', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iblock_atomic, iblock_min, ikind, &
                                                            ikind_min
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: multiindex

      IF(.NOT.(ASSOCIATED(distribution_2d)))CALL cp__a("distribution_2d_types.F",517)
      IF(.NOT.(distribution_2d%ref_count > 0))CALL cp__a("distribution_2d_types.F",518)
      IF (PRESENT(row_distribution)) row_distribution => distribution_2d%row_distribution
      IF (PRESENT(col_distribution)) col_distribution => distribution_2d%col_distribution
      IF (PRESENT(n_row_distribution)) n_row_distribution = distribution_2d%n_row_distribution
      IF (PRESENT(n_col_distribution)) n_col_distribution = distribution_2d%n_col_distribution
      IF (PRESENT(n_local_rows)) n_local_rows => distribution_2d%n_local_rows
      IF (PRESENT(n_local_cols)) n_local_cols => distribution_2d%n_local_cols
      IF (PRESENT(local_rows)) local_rows => distribution_2d%local_rows
      IF (PRESENT(local_cols)) local_cols => distribution_2d%local_cols
      IF (PRESENT(flat_local_rows)) THEN
         IF (.NOT. ASSOCIATED(distribution_2d%flat_local_rows)) THEN
            ALLOCATE (multiindex(SIZE(distribution_2d%local_rows)), &
                      distribution_2d%flat_local_rows(SUM(distribution_2d%n_local_rows)))
            multiindex = 1
            DO iblock_atomic = 1, SIZE(distribution_2d%flat_local_rows)
               iblock_min = HUGE(0)
               ikind_min = -HUGE(0)
               DO ikind = 1, SIZE(distribution_2d%local_rows)
                  IF (multiindex(ikind) <= distribution_2d%n_local_rows(ikind)) THEN
                     IF (distribution_2d%local_rows(ikind)%array(multiindex(ikind)) < &
                         iblock_min) THEN
                        iblock_min = distribution_2d%local_rows(ikind)%array(multiindex(ikind))
                        ikind_min = ikind
                     END IF
                  END IF
               END DO
               IF(.NOT.(ikind_min > 0))CALL cp__a("distribution_2d_types.F",544)
               distribution_2d%flat_local_rows(iblock_atomic) = &
                  distribution_2d%local_rows(ikind_min)%array(multiindex(ikind_min))
               multiindex(ikind_min) = multiindex(ikind_min)+1
            END DO
            DEALLOCATE (multiindex)
         END IF
         flat_local_rows => distribution_2d%flat_local_rows
      END IF
      IF (PRESENT(flat_local_cols)) THEN
         IF (.NOT. ASSOCIATED(distribution_2d%flat_local_cols)) THEN
            ALLOCATE (multiindex(SIZE(distribution_2d%local_cols)), &
                      distribution_2d%flat_local_cols(SUM(distribution_2d%n_local_cols)))
            multiindex = 1
            DO iblock_atomic = 1, SIZE(distribution_2d%flat_local_cols)
               iblock_min = HUGE(0)
               ikind_min = -HUGE(0)
               DO ikind = 1, SIZE(distribution_2d%local_cols)
                  IF (multiindex(ikind) <= distribution_2d%n_local_cols(ikind)) THEN
                     IF (distribution_2d%local_cols(ikind)%array(multiindex(ikind)) < &
                         iblock_min) THEN
                        iblock_min = distribution_2d%local_cols(ikind)%array(multiindex(ikind))
                        ikind_min = ikind
                     END IF
                  END IF
               END DO
               IF(.NOT.(ikind_min > 0))CALL cp__a("distribution_2d_types.F",570)
               distribution_2d%flat_local_cols(iblock_atomic) = &
                  distribution_2d%local_cols(ikind_min)%array(multiindex(ikind_min))
               multiindex(ikind_min) = multiindex(ikind_min)+1
            END DO
            DEALLOCATE (multiindex)
         END IF
         flat_local_cols => distribution_2d%flat_local_cols
      END IF
      IF (PRESENT(n_flat_local_rows)) n_flat_local_rows = SUM(distribution_2d%n_local_rows)
      IF (PRESENT(n_flat_local_cols)) n_flat_local_cols = SUM(distribution_2d%n_local_cols)
      IF (PRESENT(blacs_env)) blacs_env => distribution_2d%blacs_env
      IF (PRESENT(id_nr)) id_nr = distribution_2d%id_nr
   END SUBROUTINE distribution_2d_get

END MODULE distribution_2d_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Parallel (pseudo)random number generator (RNG) for multiple streams
!>      and substreams of random numbers.
!>
!>      In detail, this RNG provides 2**64 random number streams each with a
!>      length of 2**127 resulting in a length of 2**191 for the total RNG.
!>      Moreover, each stream is divided in 2**51 substream of length 2**76.
!>      The stream lengths refer to the default precision of 32 bit random
!>      number, but also an extended precision of 53 bit per random number
!>      can be requested. In this case, two 32 bit random numbers are used
!>      to generate a 53 bit random number and therefore the stream length
!>      is halved when extended precision are requested.
!>
!>      Usage hint:
!>
!>      CALL create_rng_stream(rng_stream,name,...,error=error)
!>
!>      to generate the first stream. Optionally, you may define a different
!>      seed or create a stream of extended precision (53 bits). Then
!>
!>      CALL create_rng_stream(next_rng_stream,name,last_rng_stream=rng_stream)
!>
!>      to create all the following RNG streams w.r.t. the previous stream.
!>      The command line
!>
!>      x = next_random_number(rng_stream,error=error)
!>
!>      will provide the next real random number x between 0 and 1 and
!>
!>      ix = next_random_number(rng_stream,low,high,error=error)
!>
!>      the next integer random number ix between low and high from stream
!>      rng_stream. The default distribution type is a uniform distribution
!>      [0,1], but also other distribution types are available (see below).
!>
!>      Finally, do not forget to delete each created RNG stream when it is
!>      no longer needed by
!>
!>      CALL delete_rng_stream(rng_stream,error=error)
!>
!>      to avoid memory leaks!
!> \par Literature
!>      P. L'Ecuyer, R. Simard, E. J. Chen, and W. D. Kelton,
!>      "An object-oriented random-number package with many long streams
!>       and substreams", Operations Research 50(6), 1073-1075 (2002)
!> \author C++ code converted to Fortran 90/95 (18.05.2005, Matthias Krack)
! **************************************************************************************************
MODULE parallel_rng_types

   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE string_utilities,                ONLY: compress
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters in this module

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'parallel_rng_types', &
                                           rng_record_format = "(A40,I2,3L2,ES25.16,18F20.1)"
   INTEGER, PARAMETER          :: rng_record_length = 433

   ! Distribution types:

   ! GAUSSIAN: Gaussian distribution with zero mean and unit variance
   ! UNIFORM:  Uniform distribution [0,1] with 1/2 mean (default)

   INTEGER, PARAMETER          :: GAUSSIAN = 1, &
                                  UNIFORM = 2

   REAL(KIND=dp), PARAMETER :: norm = 2.328306549295727688e-10_dp, &
                               m1 = 4294967087.0_dp, &
                               m2 = 4294944443.0_dp, &
                               a12 = 1403580.0_dp, &
                               a13n = 810728.0_dp, &
                               a21 = 527612.0_dp, &
                               a23n = 1370589.0_dp, &
                               two17 = 131072.0_dp, & ! 2**17
                               two53 = 9007199254740992.0_dp, & ! 2**53
                               fact = 5.9604644775390625e-8_dp ! 1/2**24

   ! Data type definitions

   ! Information on a stream. The arrays bg, cg, and ig contain the current
   ! state of the stream, the starting state of the current substream, and the
   ! starting state of the stream. This stream generates antithetic variates
   ! if antithetic = .TRUE.. It also generates numbers with extended precision
   ! (53 bits, if machine follows IEEE 754 standard), if extended_precision =
   ! .TRUE., otherwise, numbers with 32 bits precision are generated.

! **************************************************************************************************
   TYPE rng_stream_type
      PRIVATE
      CHARACTER(LEN=40)             :: name
      INTEGER                       :: distribution_type
      REAL(KIND=dp), DIMENSION(3, 2) :: bg, cg, ig
      LOGICAL                       :: antithetic, extended_precision
      ! only used for distribution type GAUSSIAN
      REAL(KIND=dp)                 :: buffer
      LOGICAL                       :: buffer_filled
   END TYPE rng_stream_type

! **************************************************************************************************
   TYPE rng_stream_p_type
      TYPE(rng_stream_type), POINTER :: stream
   END TYPE rng_stream_p_type

   ! The following are the transition matrices of the two MRG components
   ! (in matrix form), raised to the powers -1, 1, 2**76, and 2**127, resp.

   REAL(KIND=dp), DIMENSION(3, 3) :: a1p0, a1p76, a1p127, &
                                     a2p0, a2p76, a2p127, &
                                     inv_a1, inv_a2

   ! Interfaces

   INTERFACE next_random_number
      MODULE PROCEDURE next_integer_random_number, &
         next_real_random_number
   END INTERFACE

   INTERFACE random_numbers
      MODULE PROCEDURE random_numbers_1, &
         random_numbers_2, &
         random_numbers_3
   END INTERFACE

   ! Public parameters

   PUBLIC :: rng_record_length, &
             GAUSSIAN, &
             UNIFORM

   ! Public data types

   PUBLIC :: rng_stream_p_type, &
             rng_stream_type

   ! Public subroutines

   PUBLIC :: check_rng, &
             create_rng_stream, &
             delete_rng_stream, &
             dump_rng_stream, &
             get_rng_stream, &
             init_rng, &
             next_random_number, &
             next_rng_seed, &
             random_numbers, &
             read_rng_stream, &
             reset_to_next_rng_substream, &
             set_rng_stream, &
             write_rng_matrices, &
             write_rng_stream

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param e ...
!> \param c ...
! **************************************************************************************************
   SUBROUTINE advance_rng_state(rng_stream, e, c)

      ! Advance the state by n steps, i.e. jump n steps forward, if n > 0, or
      ! backward if n < 0.
      ! IF e > 0, let n = 2**e + c
      ! IF e < 0, let n = -2**(-e) + c
      ! IF e = 0, let n = c

      ! NOTE: The use of this method is discouraged.

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      INTEGER, INTENT(IN)                                :: e, c

      CHARACTER(LEN=*), PARAMETER :: routineN = 'advance_rng_state', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(3, 2)                     :: x
      REAL(KIND=dp), DIMENSION(3, 3)                     :: u1, u2, v1, v2, w1, w2

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",192)

      u1 = 0.0_dp
      u2 = 0.0_dp
      v1 = 0.0_dp
      v2 = 0.0_dp
      w1 = 0.0_dp
      w2 = 0.0_dp

      IF (e > 0) THEN
         CALL mat_two_pow_mod_m(a1p0, u1, m1, e)
         CALL mat_two_pow_mod_m(a2p0, u2, m2, e)
      ELSE IF (e < 0) THEN
         CALL mat_two_pow_mod_m(inv_a1, u1, m1, -e)
         CALL mat_two_pow_mod_m(inv_a2, u2, m2, -e)
      END IF

      IF (c >= 0) THEN
         CALL mat_pow_mod_m(a1p0, v1, m1, c)
         CALL mat_pow_mod_m(a2p0, v2, m2, c)
      ELSE
         CALL mat_pow_mod_m(inv_a1, v1, m1, -c)
         CALL mat_pow_mod_m(inv_a2, v2, m2, -c)
      END IF

      IF (e == 0) THEN
         w1 = v1
         w2 = v2
      ELSE
         CALL mat_mat_mod_m(u1, v1, w1, m1)
         CALL mat_mat_mod_m(u2, v2, w2, m2)
      END IF

      x = 0.0_dp

      CALL mat_vec_mod_m(w1, rng_stream%cg(:, 1), x(:, 1), m1)
      CALL mat_vec_mod_m(w2, rng_stream%cg(:, 2), x(:, 2), m2)

      rng_stream%cg = x

   END SUBROUTINE advance_rng_state

! **************************************************************************************************
!> \brief ...
!> \param output_unit ...
!> \param ionode ...
! **************************************************************************************************
   SUBROUTINE check_rng(output_unit, ionode)

      ! Check the parallel (pseudo)random number generator (RNG).

      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(IN)                                :: ionode

      CHARACTER(LEN=*), PARAMETER :: routineN = 'check_rng', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, sumi
      REAL(KIND=dp)                                      :: sum, sum3
      REAL(KIND=dp), DIMENSION(3, 2)                     :: germe
      TYPE(rng_stream_type), POINTER                     :: cantor, g1, g2, g3, galois, laplace, &
                                                            poisson

! -------------------------------------------------------------------------
! Test 1

      NULLIFY (g1)
      NULLIFY (g2)
      NULLIFY (g3)

      ! Create RNG test streams

      CALL create_rng_stream(g1, "g1")
      CALL create_rng_stream(g2, "g2", g1)
      CALL create_rng_stream(g3, "g3", g2)

      IF (ionode) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            "RESULTS OF THE PSEUDO(RANDOM) NUMBER GENERATOR TEST RUNS", &
            "Initial states of the (pseudo)random number streams (test 1):"
         CALL write_rng_stream(g1, output_unit)
         CALL write_rng_stream(g2, output_unit)
         CALL write_rng_stream(g3, output_unit)
      END IF

      sum = next_random_number(g2)+next_random_number(g3)

      CALL advance_rng_state(g1, 5, 3)
      sum = sum+next_random_number(g1)

      CALL reset_rng_stream(g1)
      DO i = 1, 35
         CALL advance_rng_state(g1, 0, 1)
      END DO
      sum = sum+next_random_number(g1)

      CALL reset_rng_stream(g1)

      sumi = 0
      DO i = 1, 35
         sumi = sumi+next_random_number(g1, 1, 10)
      END DO
      sum = sum+sumi/100.0_dp

      sum3 = 0.0_dp
      DO i = 1, 100
         sum3 = sum3+next_random_number(g3)
      END DO
      sum = sum+sum3/10.0_dp

      CALL reset_rng_stream(g3)
      DO i = 1, 5
         sum = sum+next_random_number(g3)
      END DO

      CALL reset_rng_stream(g3)
      DO i = 1, 4
         CALL reset_to_next_rng_substream(g3)
      END DO
      DO i = 1, 5
         sum = sum+next_random_number(g3)
      END DO

      CALL reset_rng_substream(g3)
      DO i = 1, 5
         sum = sum+next_random_number(g3)
      END DO

      CALL reset_to_next_rng_substream(g2)
      sum3 = 0.0_dp
      DO i = 1, 100000
         sum3 = sum3+next_random_number(g2)
      END DO
      sum = sum+sum3/10000.0_dp

      CALL set_rng_stream(g3, antithetic=.TRUE.)
      sum3 = 0.0_dp
      DO i = 1, 100000
         sum3 = sum3+next_random_number(g3)
      END DO
      sum = sum+sum3/10000.0_dp

      IF (ionode) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            "Final states of the (pseudo)random number streams (test 1):"
         CALL write_rng_stream(g1, output_unit)
         CALL write_rng_stream(g2, output_unit)
         CALL write_rng_stream(g3, output_unit)
         WRITE (UNIT=output_unit, FMT="(/,(T2,A))") &
            "This test routine should print for test 1 the number 25.342059"
         WRITE (UNIT=output_unit, FMT="(T2,A,F10.6)") &
            "The actual result of test 1 is                      ", sum
      END IF

      CALL delete_rng_stream(g1)
      CALL delete_rng_stream(g2)
      CALL delete_rng_stream(g3)

      ! -------------------------------------------------------------------------

      ! Test 2

      NULLIFY (cantor)
      NULLIFY (galois)
      NULLIFY (laplace)
      NULLIFY (poisson)

      germe(:, :) = 1

      CALL create_rng_stream(poisson, "Poisson", seed=germe)
      CALL create_rng_stream(laplace, "Laplace", poisson)
      CALL create_rng_stream(galois, "Galois", laplace)
      CALL create_rng_stream(cantor, "Cantor", galois)

      IF (ionode) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            "Initial states of the (pseudo)random number streams (test 2):"
         CALL write_rng_stream(poisson, output_unit)
         CALL write_rng_stream(laplace, output_unit)
         CALL write_rng_stream(galois, output_unit)
         CALL write_rng_stream(cantor, output_unit)
      END IF

      sum = sum+next_random_number(poisson)+ &
            next_random_number(laplace)+ &
            next_random_number(galois)+ &
            next_random_number(cantor)

      CALL advance_rng_state(galois, -127, 0)
      sum = sum+next_random_number(galois)

      CALL reset_to_next_rng_substream(galois)
      CALL set_rng_stream(galois, extended_precision=.TRUE.)
      sum3 = 0.0_dp
      DO i = 1, 100000
         sum3 = sum3+next_random_number(galois)
      END DO
      sum = sum+sum3/10000.0_dp

      CALL set_rng_stream(galois, antithetic=.TRUE.)
      sum3 = 0.0_dp
      DO i = 1, 100000
         sum3 = sum3+next_random_number(galois)
      END DO
      sum = sum+sum3/10000.0_dp
      CALL set_rng_stream(galois, antithetic=.FALSE.)

      CALL set_rng_stream(galois, extended_precision=.FALSE.)
      sum = sum+next_random_number(poisson)+ &
            next_random_number(laplace)+ &
            next_random_number(galois)+ &
            next_random_number(cantor)

      IF (ionode) THEN
         WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
            "Final states of the (pseudo)random number streams (test 2):"
         CALL write_rng_stream(poisson, output_unit)
         CALL write_rng_stream(laplace, output_unit)
         CALL write_rng_stream(galois, output_unit)
         CALL write_rng_stream(cantor, output_unit)
         WRITE (UNIT=output_unit, FMT="(/,(T2,A))") &
            "This test routine should print for test 2 the number 39.697547"
         WRITE (UNIT=output_unit, FMT="(T2,A,F10.6)") &
            "The actual result of test 2 is                      ", sum
      END IF

      CALL delete_rng_stream(cantor)
      CALL delete_rng_stream(galois)
      CALL delete_rng_stream(laplace)
      CALL delete_rng_stream(poisson)

   END SUBROUTINE check_rng

! **************************************************************************************************
!> \brief ...
!> \param seed ...
! **************************************************************************************************
   SUBROUTINE check_seed(seed)

      ! Check that the seeds are legitimate values.

      REAL(KIND=dp), DIMENSION(3, 2), INTENT(IN)         :: seed

      CHARACTER(LEN=*), PARAMETER :: fmtstr = "(A,I1,A,ES23.14,A,ES23.14)", &
         routineN = 'check_seed', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: message
      INTEGER                                            :: i

! -------------------------------------------------------------------------

      DO i = 1, 3

         ! Check condition: 0 <= seed(:,1) < m1

         IF (seed(i, 1) < 0.0_dp) THEN
            WRITE (UNIT=message, FMT=fmtstr) &
               "seed(", i, ",1) = ", seed(i, 1), " < ", 0.0_dp
            CALL compress(message)
            CALL cp__b("common/parallel_rng_types.F",450,message)
         END IF
         IF (seed(i, 1) >= m1) THEN
            WRITE (UNIT=message, FMT=fmtstr) &
               "seed(", i, ",1) = ", seed(i, 1), " >= ", m1
            CALL compress(message)
            CALL cp__b("common/parallel_rng_types.F",456,message)
         END IF

         ! Check condition: 0 <= seed(:,2) < m2

         IF (seed(i, 2) < 0.0_dp) THEN
            WRITE (UNIT=message, FMT=fmtstr) &
               "seed(", i, ",2) = ", seed(i, 2), " < ", 0.0_dp
            CALL compress(message)
            CALL cp__b("common/parallel_rng_types.F",465,message)
         END IF
         IF (seed(i, 2) >= m2) THEN
            WRITE (UNIT=message, FMT=fmtstr) &
               "seed(", i, ",2) = ", seed(i, 2), " >= ", m2
            CALL compress(message)
            CALL cp__b("common/parallel_rng_types.F",471,message)
         END IF

      END DO

      ! Check condition: first or second seed is 0

      IF (ALL(seed(:, 1) < 1.0_dp)) THEN
         CALL cp__b("common/parallel_rng_types.F",479,"First seed = 0")
      END IF

      IF (ALL(seed(:, 2) < 1.0_dp)) THEN
         CALL cp__b("common/parallel_rng_types.F",483,"Second seed = 0")
      END IF

   END SUBROUTINE check_seed

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param name ...
!> \param last_rng_stream ...
!> \param distribution_type ...
!> \param seed ...
!> \param antithetic ...
!> \param extended_precision ...
! **************************************************************************************************
   SUBROUTINE create_rng_stream(rng_stream, name, last_rng_stream, &
                                distribution_type, seed, antithetic, &
                                extended_precision)

      ! Create a new RNG stream.
      ! last_rng_stream is used as a reference stream, if it is specified.

      ! Usage hint:
      ! CALL create_rng_stream(rng_stream,name,...) to generate the first stream.
      ! Then CALL create_rng_stream(next_rng_stream,name,rng_stream) to create
      ! all the following RNG streams.

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      CHARACTER(LEN=*), INTENT(IN)                       :: name
      TYPE(rng_stream_type), OPTIONAL, POINTER           :: last_rng_stream
      INTEGER, INTENT(IN), OPTIONAL                      :: distribution_type
      REAL(KIND=dp), DIMENSION(3, 2), INTENT(IN), &
         OPTIONAL                                        :: seed
      LOGICAL, INTENT(IN), OPTIONAL                      :: antithetic, extended_precision

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_rng_stream', &
         routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF (ASSOCIATED(rng_stream)) CALL delete_rng_stream(rng_stream)

      ALLOCATE (rng_stream)

      rng_stream%name = name

      IF (PRESENT(seed)) THEN
         CALL check_seed(seed)
         rng_stream%ig = seed
      ELSE IF (PRESENT(last_rng_stream)) THEN
         rng_stream%ig = next_rng_seed(last_rng_stream%ig)
      ELSE
         rng_stream%ig = next_rng_seed()
      END IF

      rng_stream%cg = rng_stream%ig
      rng_stream%bg = rng_stream%ig

      IF (PRESENT(distribution_type)) THEN
         SELECT CASE (distribution_type)
         CASE (GAUSSIAN)
            rng_stream%distribution_type = GAUSSIAN
         CASE (UNIFORM)
            rng_stream%distribution_type = UNIFORM
         CASE DEFAULT
            CALL cp__b("common/parallel_rng_types.F",548,"Invalid distribution type specified")
         END SELECT
      ELSE IF (PRESENT(last_rng_stream)) THEN
         rng_stream%distribution_type = last_rng_stream%distribution_type
      ELSE
         rng_stream%distribution_type = UNIFORM
      END IF

      IF (PRESENT(antithetic)) THEN
         rng_stream%antithetic = antithetic
      ELSE IF (PRESENT(last_rng_stream)) THEN
         rng_stream%antithetic = last_rng_stream%antithetic
      ELSE
         rng_stream%antithetic = .FALSE.
      END IF

      IF (PRESENT(extended_precision)) THEN
         rng_stream%extended_precision = extended_precision
      ELSE IF (PRESENT(last_rng_stream)) THEN
         rng_stream%extended_precision = last_rng_stream%extended_precision
      ELSE
         rng_stream%extended_precision = .FALSE.
      END IF

      ! Initialize buffer for distribution type GAUSSIAN

      rng_stream%buffer = 0.0_dp
      rng_stream%buffer_filled = .FALSE.

   END SUBROUTINE create_rng_stream

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
! **************************************************************************************************
   SUBROUTINE delete_rng_stream(rng_stream)

      ! Delete a random number stream.

      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'delete_rng_stream', &
         routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",594)

      DEALLOCATE (rng_stream)

   END SUBROUTINE delete_rng_stream

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param rng_record ...
! **************************************************************************************************
   SUBROUTINE dump_rng_stream(rng_stream, rng_record)

      ! Dump a RNG stream as a record given as an internal file (string).

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      CHARACTER(LEN=rng_record_length), INTENT(OUT)      :: rng_record

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dump_rng_stream', &
         routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",617)

      rng_record = " "
      WRITE (UNIT=rng_record, FMT=rng_record_format) &
         rng_stream%name, &
         rng_stream%distribution_type, &
         rng_stream%antithetic, &
         rng_stream%extended_precision, &
         rng_stream%buffer_filled, &
         rng_stream%buffer, &
         rng_stream%cg, &
         rng_stream%bg, &
         rng_stream%ig

   END SUBROUTINE dump_rng_stream

! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param name ...
!> \param distribution_type ...
!> \param bg ...
!> \param cg ...
!> \param ig ...
!> \param antithetic ...
!> \param extended_precision ...
!> \param buffer ...
!> \param buffer_filled ...
!> \par History
!>      2009-11-04 changed bg, cg and ig type from INTEGER, DIMENSION(3, 2)
!>      to REAL(KIND=dp), DIMENSION(3, 2) [lwalewski]
!>      2009-11-09 getting the buffer and buffer_filled components
!>      added [lwalewski]
! **************************************************************************************************
   SUBROUTINE get_rng_stream(rng_stream, name, distribution_type, bg, cg, ig, &
                             antithetic, extended_precision, &
                             buffer, buffer_filled)

      ! Get the components of a RNG stream.

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: name
      INTEGER, INTENT(OUT), OPTIONAL                     :: distribution_type
      REAL(KIND=dp), DIMENSION(3, 2), INTENT(OUT), &
         OPTIONAL                                        :: bg, cg, ig
      LOGICAL, INTENT(OUT), OPTIONAL                     :: antithetic, extended_precision
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: buffer
      LOGICAL, INTENT(OUT), OPTIONAL                     :: buffer_filled

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_rng_stream', routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",671)

      IF (PRESENT(name)) name = rng_stream%name
      IF (PRESENT(distribution_type)) THEN
         distribution_type = rng_stream%distribution_type
      END IF
      IF (PRESENT(bg)) bg = rng_stream%bg
      IF (PRESENT(cg)) cg = rng_stream%cg
      IF (PRESENT(ig)) ig = rng_stream%ig
      IF (PRESENT(antithetic)) antithetic = rng_stream%antithetic
      IF (PRESENT(extended_precision)) THEN
         extended_precision = rng_stream%extended_precision
      END IF
      IF (PRESENT(buffer)) buffer = rng_stream%buffer
      IF (PRESENT(buffer_filled)) buffer_filled = rng_stream%buffer_filled

   END SUBROUTINE get_rng_stream

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE init_rng()

      ! Initialize the transformation matrices for the RNG.
      ! This matrices could be stored also as parameters, however, as long as
      ! RESHAPE is not processed properly by forpar.x, we have to initialize
      ! them during the program startup.

      ! -------------------------------------------------------------------------

      ! The following are the transition matrices of the two MRG components
      ! (in matrix form), raised to the powers 1, 2**76, 2**127, and -1

      ! Transition matrix for the first component raised to the power 2**0

      a1p0(1, 1) = 0.0_dp
      a1p0(2, 1) = 0.0_dp
      a1p0(3, 1) = -810728.0_dp
      a1p0(1, 2) = 1.0_dp
      a1p0(2, 2) = 0.0_dp
      a1p0(3, 2) = 1403580.0_dp
      a1p0(1, 3) = 0.0_dp
      a1p0(2, 3) = 1.0_dp
      a1p0(3, 3) = 0.0_dp

      ! Transition matrix for the second component raised to the power 2**0

      a2p0(1, 1) = 0.0_dp
      a2p0(2, 1) = 0.0_dp
      a2p0(3, 1) = -1370589.0_dp
      a2p0(1, 2) = 1.0_dp
      a2p0(2, 2) = 0.0_dp
      a2p0(3, 2) = 0.0_dp
      a2p0(1, 3) = 0.0_dp
      a2p0(2, 3) = 1.0_dp
      a2p0(3, 3) = 527612.0_dp

      ! Transition matrix for the first component raised to the power 2**76

      a1p76(1, 1) = 82758667.0_dp
      a1p76(2, 1) = 3672831523.0_dp
      a1p76(3, 1) = 3672091415.0_dp
      a1p76(1, 2) = 1871391091.0_dp
      a1p76(2, 2) = 69195019.0_dp
      a1p76(3, 2) = 3528743235.0_dp
      a1p76(1, 3) = 4127413238.0_dp
      a1p76(2, 3) = 1871391091.0_dp
      a1p76(3, 3) = 69195019.0_dp

      ! Transition matrix for the second component raised to the power 2**76

      a2p76(1, 1) = 1511326704.0_dp
      a2p76(2, 1) = 4292754251.0_dp
      a2p76(3, 1) = 3859662829.0_dp
      a2p76(1, 2) = 3759209742.0_dp
      a2p76(2, 2) = 1511326704.0_dp
      a2p76(3, 2) = 4292754251.0_dp
      a2p76(1, 3) = 1610795712.0_dp
      a2p76(2, 3) = 3889917532.0_dp
      a2p76(3, 3) = 3708466080.0_dp

      ! Transition matrix for the first component raised to the power 2**127

      a1p127(1, 1) = 2427906178.0_dp
      a1p127(2, 1) = 226153695.0_dp
      a1p127(3, 1) = 1988835001.0_dp
      a1p127(1, 2) = 3580155704.0_dp
      a1p127(2, 2) = 1230515664.0_dp
      a1p127(3, 2) = 986791581.0_dp
      a1p127(1, 3) = 949770784.0_dp
      a1p127(2, 3) = 3580155704.0_dp
      a1p127(3, 3) = 1230515664.0_dp

      ! Transition matrix for the second component raised to the power 2**127

      a2p127(1, 1) = 1464411153.0_dp
      a2p127(2, 1) = 32183930.0_dp
      a2p127(3, 1) = 2824425944.0_dp
      a2p127(1, 2) = 277697599.0_dp
      a2p127(2, 2) = 1464411153.0_dp
      a2p127(3, 2) = 32183930.0_dp
      a2p127(1, 3) = 1610723613.0_dp
      a2p127(2, 3) = 1022607788.0_dp
      a2p127(3, 3) = 2093834863.0_dp

      ! Inverse of a1p0

      inv_a1(1, 1) = 184888585.0_dp
      inv_a1(2, 1) = 1.0_dp
      inv_a1(3, 1) = 0.0_dp
      inv_a1(1, 2) = 0.0_dp
      inv_a1(2, 2) = 0.0_dp
      inv_a1(3, 2) = 1.0_dp
      inv_a1(1, 3) = 1945170933.0_dp
      inv_a1(2, 3) = 0.0_dp
      inv_a1(3, 3) = 0.0_dp

      ! Inverse of a2p0

      inv_a2(1, 1) = 0.0_dp
      inv_a2(2, 1) = 1.0_dp
      inv_a2(3, 1) = 0.0_dp
      inv_a2(1, 2) = 360363334.0_dp
      inv_a2(2, 2) = 0.0_dp
      inv_a2(3, 2) = 1.0_dp
      inv_a2(1, 3) = 4225571728.0_dp
      inv_a2(2, 3) = 0.0_dp
      inv_a2(3, 3) = 0.0_dp

   END SUBROUTINE init_rng

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param c ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE mat_mat_mod_m(a, b, c, m)

      ! Returns c = MODULO(a*b,m).

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: a, b
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: c
      REAL(KIND=dp), INTENT(IN)                          :: m

      INTEGER                                            :: i

! -------------------------------------------------------------------------

      DO i = 1, 3
         CALL mat_vec_mod_m(a, b(:, i), c(:, i), m)
      END DO

   END SUBROUTINE mat_mat_mod_m

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param m ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE mat_pow_mod_m(a, b, m, n)

      ! Compute matrix b = MODULO(a**n,m)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: a
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: b
      REAL(KIND=dp), INTENT(IN)                          :: m
      INTEGER, INTENT(IN)                                :: n

      INTEGER                                            :: i
      REAL(KIND=dp), DIMENSION(3, 3)                     :: u, v, w

! -------------------------------------------------------------------------
! Initialize: u = v = a; b = I

      w = a

      b(1, 1) = 1.0_dp
      b(2, 1) = 0.0_dp
      b(3, 1) = 0.0_dp
      b(1, 2) = 0.0_dp
      b(2, 2) = 1.0_dp
      b(3, 2) = 0.0_dp
      b(1, 3) = 0.0_dp
      b(2, 3) = 0.0_dp
      b(3, 3) = 1.0_dp

      ! Compute b = MODULO(a**n,m) using the binary decomposition of n

      i = n

      DO
         IF (MODULO(i, 2) /= 0) THEN
            u = w
            v = b
            CALL mat_mat_mod_m(u, v, b, m)
         END IF
         i = i/2
         IF (i == 0) EXIT
         u = w
         v = w
         CALL mat_mat_mod_m(u, v, w, m)
      END DO

   END SUBROUTINE mat_pow_mod_m

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param m ...
!> \param e ...
! **************************************************************************************************
   SUBROUTINE mat_two_pow_mod_m(a, b, m, e)

      ! Compute matrix b = MODULO(a**(2**e),m)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: a
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: b
      REAL(KIND=dp), INTENT(IN)                          :: m
      INTEGER, INTENT(IN)                                :: e

      INTEGER                                            :: i
      REAL(KIND=dp), DIMENSION(3, 3)                     :: u, v

! -------------------------------------------------------------------------

      b = a

      ! Compute b = MODULO(a**(2**e),m)

      DO i = 1, e
         u = b
         v = b
         CALL mat_mat_mod_m(u, v, b, m)
      END DO

   END SUBROUTINE mat_two_pow_mod_m

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param s ...
!> \param v ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE mat_vec_mod_m(a, s, v, m)

      ! Returns v = MODULO(a*s,m). Assumes that -m < s(i) < m.

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: a
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: v
      REAL(KIND=dp), INTENT(IN)                          :: m

      INTEGER                                            :: i, j
      REAL(KIND=dp)                                      :: a1, a2, c

! -------------------------------------------------------------------------

      v = 0.0_dp

      DO i = 1, 3
         DO j = 1, 3
            a2 = a(i, j)
            c = v(i)
            v(i) = a2*s(j)+c
            IF ((v(i) >= two53) .OR. (v(i) <= -two53)) THEN
               a1 = INT(a2/two17)
               a2 = a2-a1*two17
               v(i) = a1*s(j)
               a1 = INT(v(i)/m)
               v(i) = v(i)-a1*m
               v(i) = v(i)*two17+a2*s(j)+c
            END IF
            a1 = INT(v(i)/m)
            v(i) = v(i)-a1*m
            IF (v(i) < 0.0_dp) v(i) = v(i)+m
         END DO
      END DO

   END SUBROUTINE mat_vec_mod_m

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param low ...
!> \param high ...
!> \retval u ...
! **************************************************************************************************
   FUNCTION next_integer_random_number(rng_stream, low, high) RESULT(u)

      ! Get the next integer random number between low and high from the stream
      ! rng_stream.

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      INTEGER, INTENT(IN)                                :: low, high
      INTEGER                                            :: u

      CHARACTER(LEN=*), PARAMETER :: routineN = 'next_integer_random_number', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: r

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",980)
      IF(.NOT.(rng_stream%distribution_type == UNIFORM))CALL cp__a("common/parallel_rng_types.F",981)

      r = next_real_random_number(rng_stream)
      u = low+INT(r*REAL(high-low+1, dp))

   END FUNCTION next_integer_random_number

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param variance ...
!> \retval u ...
! **************************************************************************************************
   FUNCTION next_real_random_number(rng_stream, variance) RESULT(u)

      ! Get the next real random number from the stream rng_stream.
      ! variance: variance of the Gaussian distribution (defaults to 1)

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: variance
      REAL(KIND=dp)                                      :: u

      CHARACTER(LEN=*), PARAMETER :: routineN = 'next_real_random_number', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: f, r, u1, u2, var

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",1010)

      SELECT CASE (rng_stream%distribution_type)
      CASE (GAUSSIAN)
         var = 1.0_dp
         IF (PRESENT(variance)) var = variance
         ! take the random number from the buffer, if the buffer is filled
         IF (rng_stream%buffer_filled) THEN
            u = SQRT(var)*rng_stream%buffer
            rng_stream%buffer_filled = .FALSE.
         ELSE
            DO
               IF (rng_stream%extended_precision) THEN
                  u1 = 2.0_dp*rn53(rng_stream)-1.0_dp
                  u2 = 2.0_dp*rn53(rng_stream)-1.0_dp
               ELSE
                  u1 = 2.0_dp*rn32(rng_stream)-1.0_dp
                  u2 = 2.0_dp*rn32(rng_stream)-1.0_dp
               END IF
               r = u1*u1+u2*u2
               IF ((r > 0.0_dp) .AND. (r < 1.0_dp)) EXIT
            END DO
            ! Box-Muller transformation
            f = SQRT(-2.0_dp*LOG(r)/r)
            u = SQRT(var)*f*u1
            ! save the second random number for the next call
            rng_stream%buffer = f*u2
            rng_stream%buffer_filled = .TRUE.
         END IF
      CASE (UNIFORM)
         IF (rng_stream%extended_precision) THEN
            u = rn53(rng_stream)
         ELSE
            u = rn32(rng_stream)
         END IF
      END SELECT

   END FUNCTION next_real_random_number

! **************************************************************************************************
!> \brief ...
!> \param seed ...
!> \retval next_seed ...
! **************************************************************************************************
   FUNCTION next_rng_seed(seed) RESULT(next_seed)

      ! Get the seed for the next RNG stream w.r.t. a given seed.
      ! If the optional argument seed is missing, then the default seed is
      ! returned.

      REAL(KIND=dp), DIMENSION(3, 2), INTENT(IN), &
         OPTIONAL                                        :: seed
      REAL(KIND=dp), DIMENSION(3, 2)                     :: next_seed

! -------------------------------------------------------------------------

      IF (PRESENT(seed)) THEN
         CALL check_seed(seed)
         CALL mat_vec_mod_m(a1p127, seed(:, 1), next_seed(:, 1), m1)
         CALL mat_vec_mod_m(a2p127, seed(:, 2), next_seed(:, 2), m2)
      ELSE
         next_seed = 12345.0_dp ! default seed
      END IF

   END FUNCTION next_rng_seed

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param rng_stream ...
! **************************************************************************************************
   SUBROUTINE random_numbers_1(array, rng_stream)

      ! Fill entity array with random numbers from the RNG stream rng_stream.

      REAL(KIND=dp), DIMENSION(:)                        :: array
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'random_numbers_1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",1095)

      DO i = 1, SIZE(array)
         array(i) = next_random_number(rng_stream)
      END DO

   END SUBROUTINE random_numbers_1

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param rng_stream ...
! **************************************************************************************************
   SUBROUTINE random_numbers_2(array, rng_stream)

      ! Fill entity array with random numbers from the RNG stream rng_stream.

      REAL(KIND=dp), DIMENSION(:, :)                     :: array
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'random_numbers_2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",1122)

      DO j = 1, SIZE(array, 2)
         DO i = 1, SIZE(array, 1)
            array(i, j) = next_random_number(rng_stream)
         END DO
      END DO

   END SUBROUTINE random_numbers_2

! **************************************************************************************************
!> \brief ...
!> \param array ...
!> \param rng_stream ...
! **************************************************************************************************
   SUBROUTINE random_numbers_3(array, rng_stream)

      ! Fill entity array with random numbers from the RNG stream rng_stream.

      REAL(KIND=dp), DIMENSION(:, :, :)                  :: array
      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'random_numbers_3', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",1151)

      DO k = 1, SIZE(array, 3)
         DO j = 1, SIZE(array, 2)
            DO i = 1, SIZE(array, 1)
               array(i, j, k) = next_random_number(rng_stream)
            END DO
         END DO
      END DO

   END SUBROUTINE random_numbers_3

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param rng_record ...
! **************************************************************************************************
   SUBROUTINE read_rng_stream(rng_stream, rng_record)

      ! Read a RNG stream from a record given as an internal file (string).

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      CHARACTER(LEN=rng_record_length), INTENT(IN)       :: rng_record

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_rng_stream', &
         routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF (ASSOCIATED(rng_stream)) CALL delete_rng_stream(rng_stream)

      ALLOCATE (rng_stream)

      READ (UNIT=rng_record, FMT=rng_record_format) &
         rng_stream%name, &
         rng_stream%distribution_type, &
         rng_stream%antithetic, &
         rng_stream%extended_precision, &
         rng_stream%buffer_filled, &
         rng_stream%buffer, &
         rng_stream%cg, &
         rng_stream%bg, &
         rng_stream%ig

   END SUBROUTINE read_rng_stream

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
! **************************************************************************************************
   SUBROUTINE reset_rng_stream(rng_stream)

      ! Reset a random number stream to its initial state.

      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reset_rng_stream', &
         routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",1212)

      rng_stream%cg = rng_stream%ig
      rng_stream%bg = rng_stream%ig

   END SUBROUTINE reset_rng_stream

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
! **************************************************************************************************
   SUBROUTINE reset_rng_substream(rng_stream)

      ! Reset a random number stream to the beginning of its current substream.

      TYPE(rng_stream_type), POINTER                     :: rng_stream

! -------------------------------------------------------------------------

      rng_stream%cg = rng_stream%bg

   END SUBROUTINE reset_rng_substream

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
! **************************************************************************************************
   SUBROUTINE reset_to_next_rng_substream(rng_stream)

      ! Reset a random number stream to the beginning of its next substream.

      TYPE(rng_stream_type), POINTER                     :: rng_stream

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reset_to_next_rng_substream', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(3, 2)                     :: u

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",1252)

      u = 0.0_dp

      CALL mat_vec_mod_m(a1p76, rng_stream%bg(:, 1), u(:, 1), m1)
      CALL mat_vec_mod_m(a2p76, rng_stream%bg(:, 2), u(:, 2), m2)

      rng_stream%bg = u
      rng_stream%cg = u

   END SUBROUTINE reset_to_next_rng_substream

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \retval u ...
! **************************************************************************************************
   FUNCTION rn32(rng_stream) RESULT(u)

      ! Generate the next random number with standard precision (32 bits).

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      REAL(KIND=dp)                                      :: u

      INTEGER                                            :: k
      REAL(KIND=dp)                                      :: p1, p2

! -------------------------------------------------------------------------
! Component 1

      p1 = a12*rng_stream%cg(2, 1)-a13n*rng_stream%cg(1, 1)
      k = INT(p1/m1)
      p1 = p1-k*m1
      IF (p1 < 0.0_dp) p1 = p1+m1
      rng_stream%cg(1, 1) = rng_stream%cg(2, 1)
      rng_stream%cg(2, 1) = rng_stream%cg(3, 1)
      rng_stream%cg(3, 1) = p1

      ! Component 2

      p2 = a21*rng_stream%cg(3, 2)-a23n*rng_stream%cg(1, 2)
      k = INT(p2/m2)
      p2 = p2-k*m2
      IF (p2 < 0.0_dp) p2 = p2+m2
      rng_stream%cg(1, 2) = rng_stream%cg(2, 2)
      rng_stream%cg(2, 2) = rng_stream%cg(3, 2)
      rng_stream%cg(3, 2) = p2

      ! Combination

      IF (p1 > p2) THEN
         u = (p1-p2)*norm
      ELSE
         u = (p1-p2+m1)*norm
      END IF

      IF (rng_stream%antithetic) u = 1.0_dp-u

   END FUNCTION rn32

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \retval u ...
! **************************************************************************************************
   FUNCTION rn53(rng_stream) RESULT(u)

      ! Generate the next random number with extended precision (53 bits).

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      REAL(KIND=dp)                                      :: u

! -------------------------------------------------------------------------

      u = rn32(rng_stream)

      ! Note: rn32 returns 1 - u in the antithetic case

      IF (rng_stream%antithetic) THEN
         u = u+(rn32(rng_stream)-1.0_dp)*fact
         IF (u < 0.0_dp) u = u+1.0_dp
      ELSE
         u = u+rn32(rng_stream)*fact
         IF (u >= 1.0_dp) u = u-1.0_dp
      END IF

   END FUNCTION rn53

! ***************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param name ...
!> \param distribution_type ...
!> \param bg ...
!> \param cg ...
!> \param ig ...
!> \param seed ...
!> \param antithetic ...
!> \param extended_precision ...
!> \param buffer ...
!> \param buffer_filled ...
!> \par History
!>      2009-11-09 setting the buffer and buffer_filled components
!>      added [lwalewski]
! **************************************************************************************************
   SUBROUTINE set_rng_stream(rng_stream, name, distribution_type, bg, cg, ig, &
                             seed, antithetic, extended_precision, &
                             buffer, buffer_filled)

      ! Set the components of a RNG stream.
      ! NOTE: The manipulation of an active RNG stream is discouraged.

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: name
      INTEGER, INTENT(IN), OPTIONAL                      :: distribution_type
      REAL(KIND=dp), DIMENSION(3, 2), INTENT(IN), &
         OPTIONAL                                        :: bg, cg, ig, seed
      LOGICAL, INTENT(IN), OPTIONAL                      :: antithetic, extended_precision
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: buffer
      LOGICAL, INTENT(IN), OPTIONAL                      :: buffer_filled

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_rng_stream', routineP = moduleN//':'//routineN

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",1378)

      IF (PRESENT(name)) rng_stream%name = name
      IF (PRESENT(distribution_type)) THEN
         rng_stream%distribution_type = distribution_type
      END IF
      IF (PRESENT(bg)) rng_stream%bg = bg
      IF (PRESENT(cg)) rng_stream%cg = cg
      IF (PRESENT(ig)) rng_stream%ig = ig
      IF (PRESENT(seed)) THEN
         ! Sets the initial seed of the stream to seed
         ! NOTE: The use of this method is discouraged
         CALL check_seed(seed)
         rng_stream%ig = seed
         rng_stream%cg = seed
         rng_stream%bg = seed
      END IF
      IF (PRESENT(antithetic)) rng_stream%antithetic = antithetic
      IF (PRESENT(extended_precision)) THEN
         rng_stream%extended_precision = extended_precision
      END IF
      IF (PRESENT(buffer)) rng_stream%buffer = buffer
      IF (PRESENT(buffer_filled)) rng_stream%buffer_filled = buffer_filled

   END SUBROUTINE set_rng_stream

! **************************************************************************************************
!> \brief ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE write_rng_matrices(output_unit)

      ! Write the transformation matrices of the two MRG components (raised to
      ! the powers -1, 1, 2**76, and 2**127) the a logical output unit.

      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(LEN=40)                                  :: fmtstr
      INTEGER                                            :: i, j

! -------------------------------------------------------------------------
! Print the transformation matrices for both components

      WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
         "TRANSFORMATION MATRICES FOR THE PARALLEL (PSEUDO)RANDOM NUMBER "// &
         "GENERATOR"

      fmtstr = "(/,T4,A,/,/,(2X,3F14.1))"

      WRITE (UNIT=output_unit, FMT=fmtstr) &
         "A1", ((a1p0(i, j), j=1, 3), i=1, 3)

      WRITE (UNIT=output_unit, FMT=fmtstr) &
         "A2", ((a2p0(i, j), j=1, 3), i=1, 3)

      WRITE (UNIT=output_unit, FMT=fmtstr) &
         "A1**(2**76)", ((a1p76(i, j), j=1, 3), i=1, 3)

      WRITE (UNIT=output_unit, FMT=fmtstr) &
         "A2**(2**76)", ((a2p76(i, j), j=1, 3), i=1, 3)

      WRITE (UNIT=output_unit, FMT=fmtstr) &
         "A1**(2**127)", ((a1p127(i, j), j=1, 3), i=1, 3)

      WRITE (UNIT=output_unit, FMT=fmtstr) &
         "A2**(2**127)", ((a2p127(i, j), j=1, 3), i=1, 3)

   END SUBROUTINE write_rng_matrices

! **************************************************************************************************
!> \brief ...
!> \param rng_stream ...
!> \param output_unit ...
!> \param write_all ...
! **************************************************************************************************
   SUBROUTINE write_rng_stream(rng_stream, output_unit, write_all)

      ! write_all: if .TRUE., then print all stream informations.
      !            (the default is .FALSE.)

      TYPE(rng_stream_type), POINTER                     :: rng_stream
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(IN), OPTIONAL                      :: write_all

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_rng_stream', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_write_all

! -------------------------------------------------------------------------

      IF(.NOT.(ASSOCIATED(rng_stream)))CALL cp__a("common/parallel_rng_types.F",1469)

      IF (PRESENT(write_all)) THEN
         my_write_all = write_all
      ELSE
         my_write_all = .FALSE.
      END IF

      WRITE (UNIT=output_unit, FMT="(/,T2,A,/)") &
         "Random number stream <"//TRIM(rng_stream%name)//">:"

      SELECT CASE (rng_stream%distribution_type)
      CASE (GAUSSIAN)
         WRITE (UNIT=output_unit, FMT="(T4,A)") &
            "Distribution type: "// &
            "Normal Gaussian distribution with zero mean"
      CASE (UNIFORM)
         WRITE (UNIT=output_unit, FMT="(T4,A)") &
            "Distribution type: "// &
            "Uniform distribution [0,1] with 1/2 mean"
      END SELECT

      IF (rng_stream%antithetic) THEN
         WRITE (UNIT=output_unit, FMT="(T4,A)") "Antithetic:        yes"
      ELSE
         WRITE (UNIT=output_unit, FMT="(T4,A)") "Antithetic:        no"
      END IF

      IF (rng_stream%extended_precision) THEN
         WRITE (UNIT=output_unit, FMT="(T4,A)") "Precision:         53 Bit"
      ELSE
         WRITE (UNIT=output_unit, FMT="(T4,A)") "Precision:         32 Bit"
      END IF

      IF (my_write_all) THEN

         WRITE (UNIT=output_unit, FMT="(/,T4,A,/,/,(T4,A,3F20.1))") &
            "Initial state of the stream:", &
            "Component 1:", rng_stream%ig(:, 1), &
            "Component 2:", rng_stream%ig(:, 2)

         WRITE (UNIT=output_unit, FMT="(/,T4,A,/,/,(T4,A,3F20.1))") &
            "Initial state of the current substream:", &
            "Component 1:", rng_stream%bg(:, 1), &
            "Component 2:", rng_stream%bg(:, 2)

      END IF

      WRITE (UNIT=output_unit, FMT="(/,T4,A,/,/,(T4,A,3F20.1))") &
         "Current state of the stream:", &
         "Component 1:", rng_stream%cg(:, 1), &
         "Component 2:", rng_stream%cg(:, 2)

   END SUBROUTINE write_rng_stream

END MODULE parallel_rng_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent the structure of a full matrix
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_fm_struct
   USE cp_blacs_env,                    ONLY: cp_blacs_env_release,&
                                              cp_blacs_env_retain,&
                                              cp_blacs_env_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_flush
   USE message_passing,                 ONLY: mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_fm_struct'
   INTEGER, SAVE, PRIVATE :: last_fmstruct_id_nr = 0

! the default blacs block sizes
! consider using #ifdefs to give them the optimal values
! these can be changed using scf_control
! *** these are used by default
   INTEGER, PRIVATE :: optimal_blacs_col_block_size = 32
   INTEGER, PRIVATE :: optimal_blacs_row_block_size = 32
   LOGICAL, PRIVATE :: force_block_size = .FALSE.

   PUBLIC :: cp_fm_struct_type, cp_fm_struct_p_type
   PUBLIC :: cp_fm_struct_create, cp_fm_struct_retain, cp_fm_struct_release, &
             cp_fm_struct_equivalent, &
             cp_fm_struct_get, cp_fm_struct_double, cp_fm_struct_config, &
             cp_fm_struct_get_nrow_block, cp_fm_struct_get_ncol_block
!***

! **************************************************************************************************
!> \brief keeps the information about the structure of a full matrix
!> \param para_env the parallel environment of the matrixes with this structure
!> \param context the blacs context (parallel environment for scalapack),
!>        should be compatible wih para_env
!> \param descriptor the scalapack descriptor of the matrixes, when using
!>        scalapack (ncol_block=descriptor(6), ncol_global=descriptor(4),
!>        nrow_block=descriptor(5), nrow_global=descriptor(3))
!> \param ncol_block number of columns of a scalapack block
!> \param nrow_block number of rows of a scalapack block
!> \param nrow_global number of rows of the matrix
!> \param ncol_global number of rows
!> \param first_p_pos position of the first processor (for scalapack)
!> \param row_indices real (global) indexes of the rows (defined only for
!>        the local rows really used)
!> \param col_indices real (global) indexes of the cols (defined only for
!>        the local cols really used)
!> \param nrow_locals nrow_locals(i) number of local rows of the matrix really
!>        used on the processors with context%mepos(1)==i
!> \param ncol_locals ncol_locals(i) number of local rows of the matrix really
!>        used on the processors with context%mepos(2)==i
!> \param id_nr an identification number (guaranteed to be unique)
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param local_leading_dimension leading dimension of the data that is
!>        stored on this processor
!>
!>      readonly attributes:
!> \param nrow_local number of local rows really used on the actual processor
!> \param ncol_local number of local cols really used on the actual processor
!> \note
!>      use cp_fm_struct_get to extract information from this structure
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_fm_struct_type
      TYPE(cp_para_env_type), POINTER :: para_env
      TYPE(cp_blacs_env_type), POINTER :: context
      INTEGER, DIMENSION(9) :: descriptor
      INTEGER :: nrow_block, ncol_block, nrow_global, ncol_global
      INTEGER, DIMENSION(2) :: first_p_pos
      INTEGER, DIMENSION(:), POINTER :: row_indices, col_indices, &
                                        nrow_locals, ncol_locals
      INTEGER :: id_nr, ref_count, local_leading_dimension
   END TYPE cp_fm_struct_type
! **************************************************************************************************
   TYPE cp_fm_struct_p_type
      TYPE(cp_fm_struct_type), POINTER :: struct
   END TYPE cp_fm_struct_p_type

CONTAINS

! **************************************************************************************************
!> \brief allocates and initializes a full matrix structure
!> \param fmstruct the pointer that will point to the new structure
!> \param para_env the parallel environment
!> \param context the blacs context of this matrix
!> \param nrow_global the number of row of the full matrix
!> \param ncol_global the number of colums of the full matrix
!> \param nrow_block the number of rows of a block of the matrix ! useful defaults are in, do not specify
!> \param ncol_block the number of colums of a block of the matrix ! useful defaults are in, do not specify
!> \param descriptor the scalapack descriptor of the matrix (if not given
!>        a new one is allocated
!> \param first_p_pos ...
!> \param local_leading_dimension the leading dimension of the locally stored
!>        data block
!> \param template_fmstruct a matrix structure where to take the default values
!> \param square_blocks ...
!> \param force_block ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_fm_struct_create(fmstruct, para_env, context, nrow_global, &
                                  ncol_global, nrow_block, ncol_block, descriptor, first_p_pos, &
                                  local_leading_dimension, template_fmstruct, square_blocks, force_block)

      TYPE(cp_fm_struct_type), POINTER             :: fmstruct
      TYPE(cp_para_env_type), POINTER, OPTIONAL    :: para_env
      INTEGER, INTENT(in), OPTIONAL                :: nrow_global, ncol_global
      INTEGER, INTENT(in), OPTIONAL                :: nrow_block, ncol_block
      INTEGER, INTENT(in), OPTIONAL                :: local_leading_dimension
      TYPE(cp_blacs_env_type), POINTER, OPTIONAL   :: context
      INTEGER, DIMENSION(9), INTENT(in), OPTIONAL  :: descriptor
      INTEGER, OPTIONAL, DIMENSION(2)               :: first_p_pos
      TYPE(cp_fm_struct_type), POINTER, OPTIONAL   :: template_fmstruct
      LOGICAL, OPTIONAL, INTENT(in)                :: square_blocks
      LOGICAL, OPTIONAL, INTENT(in)                :: force_block

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_struct_create', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                      :: dumblock

      LOGICAL :: my_square_blocks, my_force_block


      ALLOCATE (fmstruct)

      fmstruct%nrow_block = optimal_blacs_row_block_size
      fmstruct%ncol_block = optimal_blacs_col_block_size

      IF (.NOT. PRESENT(template_fmstruct)) THEN
         IF(.NOT.(PRESENT(context)))CALL cp__a("fm/cp_fm_struct.F",161)
         IF(.NOT.(PRESENT(nrow_global)))CALL cp__a("fm/cp_fm_struct.F",162)
         IF(.NOT.(PRESENT(ncol_global)))CALL cp__a("fm/cp_fm_struct.F",163)
         fmstruct%local_leading_dimension = 1
      ELSE
         fmstruct%context => template_fmstruct%context
         fmstruct%para_env => template_fmstruct%para_env
         fmstruct%descriptor = template_fmstruct%descriptor
         fmstruct%nrow_block = template_fmstruct%nrow_block
         fmstruct%nrow_global = template_fmstruct%nrow_global
         fmstruct%ncol_block = template_fmstruct%ncol_block
         fmstruct%ncol_global = template_fmstruct%ncol_global
         fmstruct%first_p_pos = template_fmstruct%first_p_pos
         fmstruct%local_leading_dimension = &
            template_fmstruct%local_leading_dimension
      END IF

      my_force_block = force_block_size
      IF (PRESENT(force_block)) my_force_block = force_block

      IF (PRESENT(context)) THEN
         fmstruct%context => context
         fmstruct%para_env => context%para_env
      END IF
      IF (PRESENT(para_env)) fmstruct%para_env => para_env
      CALL cp_blacs_env_retain(fmstruct%context)
      CALL cp_para_env_retain(fmstruct%para_env)

      IF (PRESENT(nrow_global)) fmstruct%nrow_global = nrow_global
      IF (PRESENT(ncol_global)) fmstruct%ncol_global = ncol_global

      ! try to avoid small left-over blocks (anyway naive)
      IF (PRESENT(nrow_block)) THEN
         fmstruct%nrow_block = nrow_block
      END IF
      IF (.NOT. my_force_block) THEN
         dumblock = CEILING(REAL(fmstruct%nrow_global, KIND=dp)/ &
                            REAL(fmstruct%context%num_pe(1), KIND=dp))
         fmstruct%nrow_block = MAX(1, MIN(fmstruct%nrow_block, dumblock))
      END IF
      IF (PRESENT(ncol_block)) THEN
         fmstruct%ncol_block = ncol_block
      END IF
      IF (.NOT. my_force_block) THEN
         dumblock = CEILING(REAL(fmstruct%ncol_global, KIND=dp)/ &
                            REAL(fmstruct%context%num_pe(2), KIND=dp))
         fmstruct%ncol_block = MAX(1, MIN(fmstruct%ncol_block, dumblock))
      END IF

      ! square matrix -> square blocks (otherwise some op fail)
      my_square_blocks = fmstruct%nrow_global == fmstruct%ncol_global
      IF (PRESENT(square_blocks)) my_square_blocks = square_blocks
      IF (my_square_blocks) THEN
         fmstruct%nrow_block = MIN(fmstruct%nrow_block, fmstruct%ncol_block)
         fmstruct%ncol_block = fmstruct%nrow_block
      END IF

      ALLOCATE (fmstruct%nrow_locals(0:(fmstruct%context%num_pe(1)-1)), &
                fmstruct%ncol_locals(0:(fmstruct%context%num_pe(2)-1)))
      IF (.NOT. PRESENT(template_fmstruct)) &
         fmstruct%first_p_pos = (/0, 0/)
      IF (PRESENT(first_p_pos)) fmstruct%first_p_pos = first_p_pos

      fmstruct%nrow_locals = 0
      fmstruct%ncol_locals = 0
      ! block = full matrix
      fmstruct%nrow_block = fmstruct%nrow_global
      fmstruct%ncol_block = fmstruct%ncol_global
      fmstruct%nrow_locals(fmstruct%context%mepos(1)) = fmstruct%nrow_global
      fmstruct%ncol_locals(fmstruct%context%mepos(2)) = fmstruct%ncol_global

      fmstruct%local_leading_dimension = MAX(fmstruct%local_leading_dimension, &
                                             fmstruct%nrow_locals(fmstruct%context%mepos(1)))
      IF (PRESENT(local_leading_dimension)) THEN
         IF (MAX(1, fmstruct%nrow_locals(fmstruct%context%mepos(1))) > local_leading_dimension) &
            CALL cp_abort(cp__l("fm/cp_fm_struct.F",269), "local_leading_dimension too small ("// &
                          cp_to_string(local_leading_dimension)//"<"// &
                          cp_to_string(fmstruct%local_leading_dimension)//")")
         fmstruct%local_leading_dimension = local_leading_dimension
      END IF

      NULLIFY (fmstruct%row_indices, fmstruct%col_indices)
      last_fmstruct_id_nr = last_fmstruct_id_nr+1
      fmstruct%id_nr = last_fmstruct_id_nr
      fmstruct%ref_count = 1

      IF (PRESENT(descriptor)) THEN
         fmstruct%descriptor = descriptor
      ELSE
         fmstruct%descriptor = 0
      END IF
   END SUBROUTINE cp_fm_struct_create

! **************************************************************************************************
!> \brief retains a full matrix structure
!> \param fmstruct the structure to retain
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_fm_struct_retain(fmstruct)
      TYPE(cp_fm_struct_type), POINTER                   :: fmstruct

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_struct_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(fmstruct)))CALL cp__a("fm/cp_fm_struct.F",309)
      IF(.NOT.(fmstruct%ref_count > 0))CALL cp__a("fm/cp_fm_struct.F",310)
      fmstruct%ref_count = fmstruct%ref_count+1
   END SUBROUTINE cp_fm_struct_retain

! **************************************************************************************************
!> \brief releases a full matrix structure
!> \param fmstruct the structure to release
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_fm_struct_release(fmstruct)
      TYPE(cp_fm_struct_type), POINTER                   :: fmstruct

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_struct_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(fmstruct)) THEN
         IF(.NOT.(fmstruct%ref_count > 0))CALL cp__a("fm/cp_fm_struct.F",328)
         fmstruct%ref_count = fmstruct%ref_count-1
         IF (fmstruct%ref_count < 1) THEN
            CALL cp_blacs_env_release(fmstruct%context)
            CALL cp_para_env_release(fmstruct%para_env)
            IF (ASSOCIATED(fmstruct%row_indices)) THEN
               DEALLOCATE (fmstruct%row_indices)
            ENDIF
            IF (ASSOCIATED(fmstruct%col_indices)) THEN
               DEALLOCATE (fmstruct%col_indices)
            END IF
            IF (ASSOCIATED(fmstruct%nrow_locals)) THEN
               DEALLOCATE (fmstruct%nrow_locals)
            END IF
            IF (ASSOCIATED(fmstruct%ncol_locals)) THEN
               DEALLOCATE (fmstruct%ncol_locals)
            END IF
            DEALLOCATE (fmstruct)
         END IF
      END IF
      NULLIFY (fmstruct)
   END SUBROUTINE cp_fm_struct_release

! **************************************************************************************************
!> \brief returns true if the two matrix structures are equivalent, false
!>      otherwise.
!> \param fmstruct1 one of the full matrix structures to compare
!> \param fmstruct2 the second of the full matrix structures to compare
!> \retval res ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_fm_struct_equivalent(fmstruct1, fmstruct2) RESULT(res)
      TYPE(cp_fm_struct_type), POINTER                   :: fmstruct1, fmstruct2
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_struct_equivalent', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(fmstruct1)))CALL cp__a("fm/cp_fm_struct.F",370)
      IF(.NOT.(ASSOCIATED(fmstruct2)))CALL cp__a("fm/cp_fm_struct.F",371)
      IF(.NOT.(fmstruct1%ref_count > 0))CALL cp__a("fm/cp_fm_struct.F",372)
      IF(.NOT.(fmstruct2%ref_count > 0))CALL cp__a("fm/cp_fm_struct.F",373)
      IF (fmstruct1%id_nr == fmstruct2%id_nr) THEN
         res = .TRUE.
      ELSE
         res = (fmstruct1%context%group == fmstruct2%context%group) .AND. &
               (fmstruct1%nrow_global == fmstruct2%nrow_global) .AND. &
               (fmstruct1%ncol_global == fmstruct2%ncol_global) .AND. &
               (fmstruct1%local_leading_dimension == &
                fmstruct2%local_leading_dimension)
         DO i = 1, 9
            res = res .AND. (fmstruct1%descriptor(i) == fmstruct1%descriptor(i))
         END DO
      END IF
   END FUNCTION cp_fm_struct_equivalent

! **************************************************************************************************
!> \brief returns the values of various attributes of the matrix structure
!> \param fmstruct the structure you whant info about
!> \param para_env ...
!> \param context ...
!> \param descriptor ...
!> \param ncol_block ...
!> \param nrow_block ...
!> \param nrow_global ...
!> \param ncol_global ...
!> \param id_nr ...
!> \param ref_count ...
!> \param first_p_pos ...
!> \param row_indices ...
!> \param col_indices ...
!> \param nrow_local ...
!> \param ncol_local ...
!> \param nrow_locals ...
!> \param ncol_locals ...
!> \param local_leading_dimension ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_fm_struct_get(fmstruct, para_env, context, &
                               descriptor, ncol_block, nrow_block, nrow_global, &
                               ncol_global, id_nr, ref_count, first_p_pos, row_indices, &
                               col_indices, nrow_local, ncol_local, nrow_locals, ncol_locals, &
                               local_leading_dimension)
      TYPE(cp_fm_struct_type), POINTER :: fmstruct
      TYPE(cp_para_env_type), POINTER, OPTIONAL :: para_env
      TYPE(cp_blacs_env_type), POINTER, OPTIONAL :: context
      INTEGER, DIMENSION(9), INTENT(OUT), OPTIONAL :: descriptor
      INTEGER, INTENT(out), OPTIONAL :: ncol_block, nrow_block, nrow_global, &
                                        ncol_global, id_nr, ref_count, nrow_local, ncol_local, &
                                        local_leading_dimension
      INTEGER, DIMENSION(2), INTENT(out), OPTIONAL :: first_p_pos
      INTEGER, DIMENSION(:), POINTER, OPTIONAL :: row_indices, col_indices, &
                                                  nrow_locals, ncol_locals

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_struct_get', &
                                     routineP = moduleN//':'//routineN
      INTEGER i, nprow, npcol, myprow, mypcol

      IF(.NOT.(ASSOCIATED(fmstruct)))CALL cp__a("fm/cp_fm_struct.F",435)
      IF(.NOT.(fmstruct%ref_count > 0))CALL cp__a("fm/cp_fm_struct.F",436)
      IF (PRESENT(para_env)) para_env => fmstruct%para_env
      IF (PRESENT(context)) context => fmstruct%context
      IF (PRESENT(descriptor)) descriptor = fmstruct%descriptor
      IF (PRESENT(ncol_block)) ncol_block = fmstruct%ncol_block
      IF (PRESENT(nrow_block)) nrow_block = fmstruct%nrow_block
      IF (PRESENT(nrow_global)) nrow_global = fmstruct%nrow_global
      IF (PRESENT(ncol_global)) ncol_global = fmstruct%ncol_global
      IF (PRESENT(id_nr)) id_nr = fmstruct%id_nr
      IF (PRESENT(ref_count)) ref_count = fmstruct%ref_count
      IF (PRESENT(first_p_pos)) first_p_pos = fmstruct%first_p_pos
      IF (PRESENT(nrow_locals)) nrow_locals => fmstruct%nrow_locals
      IF (PRESENT(ncol_locals)) ncol_locals => fmstruct%ncol_locals
      IF (PRESENT(local_leading_dimension)) local_leading_dimension = &
         fmstruct%local_leading_dimension

      myprow = fmstruct%context%mepos(1)
      mypcol = fmstruct%context%mepos(2)
      nprow = fmstruct%context%num_pe(1)
      npcol = fmstruct%context%num_pe(2)

      IF (PRESENT(nrow_local)) nrow_local = fmstruct%nrow_locals(myprow)
      IF (PRESENT(ncol_local)) ncol_local = fmstruct%ncol_locals(mypcol)

      IF (PRESENT(row_indices)) THEN
         row_indices => fmstruct%row_indices
         IF (.NOT. ASSOCIATED(row_indices)) THEN
            ! the max should go away
            ALLOCATE (fmstruct%row_indices &
                      (MAX(fmstruct%nrow_locals(myprow), 1)))
            row_indices => fmstruct%row_indices
            DO i = 1, SIZE(row_indices)
               row_indices(i) = i
            ENDDO
         END IF
      ENDIF

      IF (PRESENT(col_indices)) THEN
         col_indices => fmstruct%col_indices
         IF (.NOT. ASSOCIATED(col_indices)) THEN
            ALLOCATE (fmstruct%col_indices &
                      (MAX(fmstruct%ncol_locals(mypcol), 1)))
            col_indices => fmstruct%col_indices
            DO i = 1, SIZE(col_indices)
               col_indices(i) = i
            ENDDO
         END IF

      ENDIF
   END SUBROUTINE cp_fm_struct_get

! **************************************************************************************************
!> \brief creates a struct with twice the number of blocks on each core.
!>        If matrix A has to be multiplied with B anc C, a
!>        significant speedup of pdgemm can be acchieved by joining the matrices
!>        in a new one with this structure (see arnoldi in rt_matrix_exp)
!> \param fmstruct the struct to create
!> \param struct struct of either A or B
!> \param context ...
!> \param col in which direction the matrix should be enlarged
!> \param row in which direction the matrix should be enlarged
!> \par History
!>      06.2009 created [fschiff]
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE cp_fm_struct_double(fmstruct, struct, context, col, row)
      TYPE(cp_fm_struct_type), POINTER                   :: fmstruct, struct
      TYPE(cp_blacs_env_type), POINTER                   :: context
      LOGICAL                                            :: col, row

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_struct_double', &
         routineP = moduleN//':'//routineN

      INTEGER :: nblocks, ncol_block, ncol_global, nempty, newdim_col, newdim_row, nfilled, &
         nprocs_col, nprocs_row, nrow_block, nrow_global
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL cp_fm_struct_get(struct, nrow_global=nrow_global, &
                            ncol_global=ncol_global, nrow_block=nrow_block, &
                            ncol_block=ncol_block)
      newdim_row = nrow_global
      newdim_col = ncol_global
      nprocs_row = context%num_pe(1)
      nprocs_col = context%num_pe(2)
      para_env => struct%para_env

      IF (col) THEN
         IF (ncol_global == 0) THEN
            newdim_col = 0
         ELSE
            !Divide ncol_global by ncol_block and round up
            nblocks = (ncol_global+ncol_block-1)/ncol_block
            nfilled = MOD(nblocks, nprocs_col)
            nempty = MOD(nprocs_col-nfilled, nprocs_col)
            newdim_col = 2*ncol_global+2*nempty*ncol_block+2*MOD(ncol_block-MOD(ncol_global, ncol_block), ncol_block)
         END IF
      END IF

      IF (row) THEN
         IF (nrow_global == 0) THEN
            newdim_row = 0
         ELSE
            !Divide nrow_global by nrow_block and round up
            nblocks = (nrow_global+nrow_block-1)/nrow_block
            nfilled = MOD(nblocks, nprocs_row)
            nempty = MOD(nprocs_row-nfilled, nprocs_row)
            newdim_row = 2*nrow_global+2*nempty*nrow_block+2*MOD(nrow_block-MOD(nrow_global, nrow_block), nrow_block)
         END IF
      END IF

      CALL cp_fm_struct_create(fmstruct=fmstruct, para_env=para_env, &
                               context=context, &
                               nrow_global=newdim_row, &
                               ncol_global=newdim_col, &
                               ncol_block=ncol_block, &
                               nrow_block=nrow_block)

   END SUBROUTINE cp_fm_struct_double
! **************************************************************************************************
!> \brief allows to modify the default settings for matrix creation
!> \param nrow_block ...
!> \param ncol_block ...
!> \param force_block ...
! **************************************************************************************************
   SUBROUTINE cp_fm_struct_config(nrow_block, ncol_block, force_block)
      INTEGER, OPTIONAL                                  :: nrow_block, ncol_block
      LOGICAL, OPTIONAL                                  :: force_block

      IF (PRESENT(ncol_block)) optimal_blacs_col_block_size = ncol_block
      IF (PRESENT(nrow_block)) optimal_blacs_row_block_size = nrow_block
      IF (PRESENT(force_block)) force_block_size = force_block

   END SUBROUTINE cp_fm_struct_config

! **************************************************************************************************
!> \brief ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION cp_fm_struct_get_nrow_block() RESULT(res)
      INTEGER                                            :: res

      res = optimal_blacs_row_block_size
   END FUNCTION cp_fm_struct_get_nrow_block

! **************************************************************************************************
!> \brief ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION cp_fm_struct_get_ncol_block() RESULT(res)
      INTEGER                                            :: res

      res = optimal_blacs_col_block_size
   END FUNCTION cp_fm_struct_get_ncol_block

END MODULE cp_fm_struct
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent a full matrix distribued on many processors
!> \par History
!>      3) separated structure object, removed globenv, renamed to full matrix
!>         many changes (fawzi 08.2002)
!> \author Matthias Krack (22.05.2001)
! **************************************************************************************************
MODULE cp_fm_types
   USE cp_blacs_calls,                  ONLY: cp_blacs_dgebr2d,&
                                              cp_blacs_dgebs2d,&
                                              cp_blacs_gridexit,&
                                              cp_blacs_gridinfo,&
                                              cp_blacs_gridinit
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_fm_struct,                    ONLY: cp_fm_struct_equivalent,&
                                              cp_fm_struct_get,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_retain,&
                                              cp_fm_struct_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp,&
                                              sp
   USE message_passing,                 ONLY: cp2k_is_parallel,&
                                              mp_bcast,&
                                              mp_max,&
                                              mp_recv,&
                                              mp_send,&
                                              mp_sum
   USE parallel_rng_types,              ONLY: UNIFORM,&
                                              create_rng_stream,&
                                              delete_rng_stream,&
                                              get_rng_stream,&
                                              random_numbers,&
                                              reset_to_next_rng_substream,&
                                              rng_stream_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_fm_types'
   LOGICAL, PARAMETER          :: debug_this_module = .TRUE.

   INTEGER, SAVE :: last_fm_id_nr = 0
   INTEGER, PRIVATE :: mm_type = 1

   PUBLIC :: cp_fm_type, &
             cp_fm_p_type

   PUBLIC :: cp_fm_add_to_element, &
             cp_fm_create, &
             cp_fm_retain, &
             cp_fm_release, &
             cp_fm_get_info, &
             cp_fm_set_element, &
             cp_fm_get_element, &
             cp_fm_get_diag, & ! get diagonal
             cp_fm_set_all, & ! set all elements and diagonal
             cp_fm_set_submatrix, & ! set a submatrix to given values
             cp_fm_get_submatrix, & ! get a submatrix of given values
             cp_fm_init_random, &
             cp_fm_maxabsval, & ! find the maximum absolute value
             cp_fm_maxabsrownorm, & ! find the maximum of the sum of the abs of the elements of a row
             cp_fm_to_fm, & ! copy (parts of) a fm to a fm
             cp_fm_vectorsnorm, & ! compute the norm of the column-vectors
             cp_fm_vectorssum, & ! compute the sum of all elements of the column-vectors
             cp_fm_to_fm_submat, & ! copy (parts of) a fm to a fm
             cp_fm_to_fm_triangular, &
             cp_fm_copy_general, &
             cp_fm_to_fm_submat_general, &
             cp_fm_write_unformatted, & ! writes a full matrix to an open unit
             cp_fm_read_unformatted, & ! reads a full matrix from an open unit
             cp_fm_setup, & ! allows to set flags for fms
             cp_fm_get_mm_type

   PUBLIC :: cp_fm_indxg2p, &
             cp_fm_indxg2l, &
             cp_fm_indxl2g

   INTERFACE cp_fm_to_fm
      MODULE PROCEDURE cp_fm_to_fm_matrix, & ! a full matrix
         cp_fm_to_fm_columns ! just a number of columns
   END INTERFACE

! **************************************************************************************************
!> \brief represent a full matrix
!> \param name the name of the matrix, used for printing
!> \param id_nr identification number that makes this matrix unique
!> \param ref_cont reference count (see doc/ReferenceCounting.html)
!> \param print_count sequential number to make the printed files unique
!> \param matrix_struct structure of this matrix
!> \param local_data array with the data of the matrix (its contents
!>        depend on the matrix type used: in parallel runs it will be
!>        in scalapack format, in sequential, it will simply contain
!>        the matrix)
!> \par History
!>      08.2002 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE cp_fm_type
!    PRIVATE
      CHARACTER(LEN=60) :: name
      INTEGER :: id_nr, ref_count, print_count
      LOGICAL :: use_sp
      TYPE(cp_fm_struct_type), POINTER :: matrix_struct
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: local_data
      REAL(KIND=sp), DIMENSION(:, :), POINTER :: local_data_sp
   END TYPE cp_fm_type

! **************************************************************************************************
!> \brief just to build arrays of pointers to matrixes
!> \param matrix the pointer to the matrix
!> \par History
!>      08.2002 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE cp_fm_p_type
      TYPE(cp_fm_type), POINTER :: matrix
   END TYPE cp_fm_p_type

CONTAINS

! **************************************************************************************************
!> \brief creates a new full matrix with the given structure
!> \param matrix the matrix to be created
!> \param matrix_struct the structure of matrix
!> \param name ...
!> \param use_sp ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      preferred allocation routine
! **************************************************************************************************
   SUBROUTINE cp_fm_create(matrix, matrix_struct, name, use_sp)
      TYPE(cp_fm_type), POINTER                          :: matrix
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct
      CHARACTER(len=*), INTENT(in), OPTIONAL             :: name
      LOGICAL, INTENT(in), OPTIONAL                      :: use_sp

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ncol_local, npcol, nprow, &
                                                            nrow_local
      TYPE(cp_blacs_env_type), POINTER                   :: context

      CALL timeset(routineN, handle)


      IF(.NOT.(ASSOCIATED(matrix_struct)))CALL cp__a("fm/cp_fm_types.F",160)
      ALLOCATE (matrix)
      context => matrix_struct%context
      matrix%matrix_struct => matrix_struct
      CALL cp_fm_struct_retain(matrix%matrix_struct)
      last_fm_id_nr = last_fm_id_nr+1
      matrix%id_nr = last_fm_id_nr
      matrix%ref_count = 1
      matrix%print_count = 0

      matrix%use_sp = .FALSE.
      IF (PRESENT(use_sp)) matrix%use_sp = use_sp

      nprow = context%num_pe(1)
      npcol = context%num_pe(2)
      NULLIFY (matrix%local_data)
      NULLIFY (matrix%local_data_sp)

      ! OK, we allocate here at least a 1 x 1 matrix
      ! this must (and is) compatible with the descinit call
      ! in cp_fm_struct
      nrow_local = matrix_struct%local_leading_dimension
      ncol_local = MAX(1, matrix_struct%ncol_locals(context%mepos(2)))
      IF (matrix%use_sp) THEN
         ALLOCATE (matrix%local_data_sp(nrow_local, ncol_local))
      ELSE
         ALLOCATE (matrix%local_data(nrow_local, ncol_local))
      ENDIF

      ! JVDV we should remove this, as it is up to the user to zero afterwards
      IF (matrix%use_sp) THEN
         CALL scopy(nrow_local*ncol_local, 0.0_sp, 0, matrix%local_data_sp, 1)
      ELSE
         CALL dcopy(nrow_local*ncol_local, 0.0_dp, 0, matrix%local_data, 1)
      ENDIF

      IF (PRESENT(name)) THEN
         matrix%name = name
      ELSE
         matrix%name = 'full matrix'//cp_to_string(matrix%id_nr)
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_create

! **************************************************************************************************
!> \brief retains a full matrix
!> \param matrix the matrix to retain
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_fm_retain(matrix)
      TYPE(cp_fm_type), POINTER                          :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(matrix)))CALL cp__a("fm/cp_fm_types.F",218)
      IF(.NOT.(matrix%ref_count > 0))CALL cp__a("fm/cp_fm_types.F",219)
      matrix%ref_count = matrix%ref_count+1

   END SUBROUTINE cp_fm_retain

! **************************************************************************************************
!> \brief releases a full matrix
!> \param matrix the matrix to release
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_fm_release(matrix)
      TYPE(cp_fm_type), POINTER                          :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(matrix)) THEN
         IF(.NOT.(matrix%ref_count > 0))CALL cp__a("fm/cp_fm_types.F",241)
         matrix%ref_count = matrix%ref_count-1
         IF (matrix%ref_count < 1) THEN
            IF (ASSOCIATED(matrix%local_data)) THEN
               DEALLOCATE (matrix%local_data)
            END IF
            IF (ASSOCIATED(matrix%local_data_sp)) THEN
               DEALLOCATE (matrix%local_data_sp)
            END IF
            matrix%name = ""
            CALL cp_fm_struct_release(matrix%matrix_struct)
            DEALLOCATE (matrix)
         END IF
      END IF
      NULLIFY (matrix)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_release

! **************************************************************************************************
!> \brief fills a matrix with random numbers
!> \param matrix : to be initialized
!> \param ncol : numbers of cols to fill
!> \param start_col : starting at coll number
!> \author Joost VandeVondele
!> \note
!>      the value of a_ij is independent of the number of cpus
! **************************************************************************************************
   SUBROUTINE cp_fm_init_random(matrix, ncol, start_col)
      TYPE(cp_fm_type), POINTER                          :: matrix
      INTEGER, INTENT(IN), OPTIONAL                      :: ncol, start_col

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_init_random', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, icol_global, icol_local, irow_local, my_ncol, my_start_col, ncol_global, &
         ncol_local, nrow_global, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL, SAVE                                      :: FIRST = .TRUE.
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: buff
      REAL(KIND=dp), DIMENSION(3, 2), SAVE               :: seed
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data
      TYPE(rng_stream_type), POINTER                     :: rng

      CALL timeset(routineN, handle)

      ! first setup a parallel rng
      IF (FIRST) THEN
         seed(:, :) = RESHAPE((/1.0_dp, 2.0_dp, 3.0_dp, 4.0_dp, 5.0_dp, 6.0_dp/), (/3, 2/))
         FIRST = .FALSE.
      ENDIF
      ! guarantee same seed over all tasks
      CALL mp_bcast(seed, 0, matrix%matrix_struct%para_env%group)

      NULLIFY (rng)
      CALL create_rng_stream(rng, "cp_fm_init_random_stream", distribution_type=UNIFORM, &
                             extended_precision=.TRUE., seed=seed)

      IF(.NOT.(.NOT. matrix%use_sp))CALL cp__a("fm/cp_fm_types.F",300)

      CALL cp_fm_get_info(matrix, nrow_global=nrow_global, ncol_global=ncol_global, &
                          nrow_local=nrow_local, ncol_local=ncol_local, &
                          local_data=local_data, &
                          row_indices=row_indices, col_indices=col_indices)

      my_start_col = 1
      IF (PRESENT(start_col)) my_start_col = start_col
      my_ncol = matrix%matrix_struct%ncol_global
      IF (PRESENT(ncol)) my_ncol = ncol

      IF (ncol_global < (my_start_col+my_ncol-1)) &
         CALL cp__b("fm/cp_fm_types.F",313,"ncol_global>=(my_start_col+my_ncol-1)")

      ALLOCATE (buff(nrow_global))

      ! each global row has its own substream, in order to reach the stream for the local col,
      ! we just reset to the next substream
      ! following this, we fill the full buff with random numbers, and pick those we need
      icol_global = 0
      DO icol_local = 1, ncol_local
         IF(.NOT.(col_indices(icol_local) > icol_global))CALL cp__a("fm/cp_fm_types.F",322)
         DO
            CALL reset_to_next_rng_substream(rng)
            icol_global = icol_global+1
            IF (icol_global == col_indices(icol_local)) EXIT
         ENDDO
         CALL random_numbers(buff, rng)
         DO irow_local = 1, nrow_local
            local_data(irow_local, icol_local) = buff(row_indices(irow_local))
         ENDDO
      ENDDO

      DEALLOCATE (buff)

      ! store seed before deletion (unclear if this is the proper seed)
      CALL get_rng_stream(rng, ig=seed)
      CALL delete_rng_stream(rng)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_init_random

! **************************************************************************************************
!> \brief set all elements of a matrix to the same value,
!>      and optionally the diagonal to a different one
!> \param matrix input matrix
!> \param alpha scalar used to set all elements of the matrix
!> \param beta scalar used to set diagonal of the matrix
!> \note
!>      can be used to zero a matrix
!>      can be used to create a unit matrix (I-matrix) alpha=0.0_dp beta=1.0_dp
! **************************************************************************************************
   SUBROUTINE cp_fm_set_all(matrix, alpha, beta)

      TYPE(cp_fm_type), POINTER                          :: matrix
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: beta

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_set_all', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, matrix_size, n

      CALL timeset(routineN, handle)

      IF (matrix%use_sp) THEN
         matrix_size = SIZE(matrix%local_data_sp, 1)*SIZE(matrix%local_data_sp, 2)
         CALL scopy(matrix_size, REAL(alpha, sp), 0, matrix%local_data_sp, 1)
      ELSE
         matrix_size = SIZE(matrix%local_data, 1)*SIZE(matrix%local_data, 2)
         CALL dcopy(matrix_size, alpha, 0, matrix%local_data, 1)
      ENDIF

      IF (PRESENT(beta)) THEN
         IF(.NOT.(.NOT. matrix%use_sp))CALL cp__a("fm/cp_fm_types.F",375)
         n = MIN(matrix%matrix_struct%nrow_global, matrix%matrix_struct%ncol_global)
         DO i = 1, n
            CALL cp_fm_set_element(matrix, i, i, beta)
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_set_all

! **************************************************************************************************
!> \brief returns the diagonal elements of a fm
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
   SUBROUTINE cp_fm_get_diag(matrix, diag)

      IMPLICIT NONE

      ! arguments
      TYPE(cp_fm_type), POINTER                  :: matrix
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT) :: diag

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_get_diag', &
                                     routineP = moduleN//':'//routineN

      ! locals
      INTEGER :: i, nrow_global


      CALL cp_fm_get_info(matrix, nrow_global=nrow_global)

      DO i = 1, nrow_global
         IF (matrix%use_sp) THEN
            diag(i) = REAL(matrix%local_data_sp(i, i), dp)
         ELSE
            diag(i) = matrix%local_data(i, i)
         ENDIF
      ENDDO
      CALL mp_sum(diag, matrix%matrix_struct%para_env%group)

   END SUBROUTINE cp_fm_get_diag

! **************************************************************************************************
!> \brief returns an element of a fm
!>      this value is valid on every cpu
!>      using this call is expensive
!> \param matrix the matrix to read
!> \param irow_global the row
!> \param icol_global the col
!> \param alpha the value of matrix(irow_global, icol_global)
!> \param local true if the element is on this cpu, false otherwise
!> \note
!>      - modified semantics. now this function always returns the value
!>        previously the value was zero on cpus that didn't own the relevant
!>        part of the matrix (Joost VandeVondele, May 2003)
!>      - usage of the function should be avoided, as it is likely to rather slow
!>        using row_indices/col_indices/local_data + some smart scheme normally
!>        yields a real parallel code
! **************************************************************************************************
   SUBROUTINE cp_fm_get_element(matrix, irow_global, icol_global, alpha, local)

      IMPLICIT NONE

      ! arguments
      TYPE(cp_fm_type), POINTER          :: matrix
      REAL(KIND=dp), INTENT(OUT)                     :: alpha
      INTEGER, INTENT(IN)                       :: icol_global, &
                                                   irow_global
      LOGICAL, INTENT(OUT), OPTIONAL            :: local

      ! locals

      IF (PRESENT(local)) local = .TRUE.
      alpha = matrix%local_data(irow_global, icol_global)

   END SUBROUTINE cp_fm_get_element

! **************************************************************************************************
!> \brief sets an element of a matrix
!> \param matrix ...
!> \param irow_global ...
!> \param icol_global ...
!> \param alpha ...
!> \note
!>      we expect all cpus to have the same arguments in the call to this function
!>      (otherwise one should use local_data tricks)
! **************************************************************************************************
   SUBROUTINE cp_fm_set_element(matrix, irow_global, icol_global, alpha)
      TYPE(cp_fm_type), POINTER                :: matrix
      INTEGER, INTENT(IN)                      :: irow_global, icol_global
      REAL(KIND=dp), INTENT(IN)                :: alpha

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_set_element', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: mypcol, myprow, npcol, nprow
      TYPE(cp_blacs_env_type), POINTER         :: context

      context => matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      nprow = context%num_pe(1)
      npcol = context%num_pe(2)

      IF(.NOT.(.NOT. matrix%use_sp))CALL cp__a("fm/cp_fm_types.F",551)


      matrix%local_data(irow_global, icol_global) = alpha

   END SUBROUTINE cp_fm_set_element

! **************************************************************************************************
!> \brief sets a submatrix of a full matrix
!>       fm(start_row:start_row+n_rows,start_col:start_col+n_cols)
!>       = alpha*op(new_values)(1:n_rows,1:n_cols)+ beta
!>       * fm(start_row:start_row+n_rows,start_col:start_col+n_cols)
!> \param fm the full to change
!> \param new_values a replicated full matrix with the new values
!> \param start_row the starting row of b_matrix (defaults to 1)
!> \param start_col the starting col of b_matrix (defaults to 1)
!> \param n_rows the number of row to change in b (defaults to
!>        size(op(new_values),1))
!> \param n_cols the number of columns to change in b (defaults to
!>        size(op(new_values),2))
!> \param alpha rescaling factor for the new values (defaults to 1.0)
!> \param beta rescaling factor for the old values (defaults to 0.0)
!> \param transpose if new_values should be transposed: if true
!>        op(new_values)=new_values^T, else op(new_values)=new_values
!>        (defaults to false)
!> \par History
!>      07.2002 created borrowing from Joost's blacs_replicated_copy [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      optimized for full column updates and alpha=1.0, beta=0.0
!>      the new_values need to be valid on all cpus
! **************************************************************************************************
   SUBROUTINE cp_fm_set_submatrix(fm, new_values, start_row, &
                                  start_col, n_rows, n_cols, alpha, beta, transpose)
      TYPE(cp_fm_type), POINTER                          :: fm
      REAL(KIND=dp), DIMENSION(:, :), INTENT(in)         :: new_values
      INTEGER, INTENT(in), OPTIONAL                      :: start_row, start_col, n_rows, n_cols
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: alpha, beta
      LOGICAL, INTENT(in), OPTIONAL                      :: transpose

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_set_submatrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i0, j, j0, ncol, ncol_block, &
                                                            ncol_global, ncol_local, nrow, &
                                                            nrow_block, nrow_global, nrow_local, &
                                                            this_col, this_row
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: tr_a
      REAL(KIND=dp)                                      :: al, be
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: full_block

      al = 1.0_dp; be = 0.0_dp; i0 = 1; j0 = 1; tr_a = .FALSE.
      ! can be called too many times, making it a bit useless
      ! CALL timeset(routineN//','//moduleN,handle)

      IF(.NOT.(ASSOCIATED(fm)))CALL cp__a("fm/cp_fm_types.F",622)
      IF(.NOT.(fm%ref_count > 0))CALL cp__a("fm/cp_fm_types.F",623)

      IF(.NOT.(.NOT. fm%use_sp))CALL cp__a("fm/cp_fm_types.F",625)

      IF (PRESENT(alpha)) al = alpha
      IF (PRESENT(beta)) be = beta
      IF (PRESENT(start_row)) i0 = start_row
      IF (PRESENT(start_col)) j0 = start_col
      IF (PRESENT(transpose)) tr_a = transpose
      IF (tr_a) THEN
         nrow = SIZE(new_values, 2)
         ncol = SIZE(new_values, 1)
      ELSE
         nrow = SIZE(new_values, 1)
         ncol = SIZE(new_values, 2)
      END IF
      IF (PRESENT(n_rows)) nrow = n_rows
      IF (PRESENT(n_cols)) ncol = n_cols

      full_block => fm%local_data

      CALL cp_fm_get_info(matrix=fm, &
                          nrow_global=nrow_global, ncol_global=ncol_global, &
                          nrow_block=nrow_block, ncol_block=ncol_block, &
                          nrow_local=nrow_local, ncol_local=ncol_local, &
                          row_indices=row_indices, col_indices=col_indices)

      IF (al == 1.0 .AND. be == 0.0) THEN
         DO j = 1, ncol_local
            this_col = col_indices(j)-j0+1
            IF (this_col .GE. 1 .AND. this_col .LE. ncol) THEN
               IF (tr_a) THEN
                  IF (i0 == 1 .AND. nrow_global == nrow) THEN
                     DO i = 1, nrow_local
                        full_block(i, j) = new_values(this_col, row_indices(i))
                     END DO
                  ELSE
                     DO i = 1, nrow_local
                        this_row = row_indices(i)-i0+1
                        IF (this_row >= 1 .AND. this_row <= nrow) THEN
                           full_block(i, j) = new_values(this_col, this_row)
                        END IF
                     END DO
                  END IF
               ELSE
                  IF (i0 == 1 .AND. nrow_global == nrow) THEN
                     DO i = 1, nrow_local
                        full_block(i, j) = new_values(row_indices(i), this_col)
                     END DO
                  ELSE
                     DO i = 1, nrow_local
                        this_row = row_indices(i)-i0+1
                        IF (this_row >= 1 .AND. this_row <= nrow) THEN
                           full_block(i, j) = new_values(this_row, this_col)
                        END IF
                     END DO
                  END IF
               END IF
            END IF
         END DO
      ELSE
         DO j = 1, ncol_local
            this_col = col_indices(j)-j0+1
            IF (this_col .GE. 1 .AND. this_col .LE. ncol) THEN
               IF (tr_a) THEN
                  DO i = 1, nrow_local
                     this_row = row_indices(i)-i0+1
                     IF (this_row >= 1 .AND. this_row <= nrow) THEN
                        full_block(i, j) = al*new_values(this_col, this_row)+ &
                                           be*full_block(i, j)
                     END IF
                  END DO
               ELSE
                  DO i = 1, nrow_local
                     this_row = row_indices(i)-i0+1
                     IF (this_row >= 1 .AND. this_row <= nrow) THEN
                        full_block(i, j) = al*new_values(this_row, this_col)+ &
                                           be*full_block(i, j)
                     END IF
                  END DO
               END IF
            END IF
         END DO
      END IF

      ! CALL timestop(handle)

   END SUBROUTINE cp_fm_set_submatrix

! **************************************************************************************************
!> \brief gets a submatrix of a full matrix
!>       op(target_m)(1:n_rows,1:n_cols)
!>       =fm(start_row:start_row+n_rows,start_col:start_col+n_cols)
!>      target_m is replicated on all cpus
!>      using this call is expensive
!> \param fm the full you want to get the info from
!> \param target_m a replicated full matrix that will contain the result
!> \param start_row the starting row of b_matrix (defaults to 1)
!> \param start_col the starting col of b_matrix (defaults to 1)
!> \param n_rows the number of row to change in b (defaults to
!>        size(op(new_values),1))
!> \param n_cols the number of columns to change in b (defaults to
!>        size(op(new_values),2))
!> \param transpose if target_m should be transposed: if true
!>        op(target_m)=target_m^T, else op(target_m)=target_m
!>        (defaults to false)
!> \par History
!>      07.2002 created borrowing from Joost's blacs_replicated_copy [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      optimized for full column updates. Zeros out a little too much
!>      of target_m
!>      the target_m is replicated and valid on all cpus
! **************************************************************************************************
   SUBROUTINE cp_fm_get_submatrix(fm, target_m, start_row, &
                                  start_col, n_rows, n_cols, transpose)
      TYPE(cp_fm_type), POINTER                          :: fm
      REAL(KIND=dp), DIMENSION(:, :), INTENT(out)        :: target_m
      INTEGER, INTENT(in), OPTIONAL                      :: start_row, start_col, n_rows, n_cols
      LOGICAL, INTENT(in), OPTIONAL                      :: transpose

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_get_submatrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, i0, j, j0, ncol, ncol_global, &
                                                            ncol_local, nrow, nrow_global, &
                                                            nrow_local, this_col, this_row
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: tr_a
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: full_block
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)

      i0 = 1; j0 = 1; tr_a = .FALSE.

      IF(.NOT.(ASSOCIATED(fm)))CALL cp__a("fm/cp_fm_types.F",759)
      IF(.NOT.(fm%ref_count > 0))CALL cp__a("fm/cp_fm_types.F",760)

      IF(.NOT.(.NOT. fm%use_sp))CALL cp__a("fm/cp_fm_types.F",762)

      IF (PRESENT(start_row)) i0 = start_row
      IF (PRESENT(start_col)) j0 = start_col
      IF (PRESENT(transpose)) tr_a = transpose
      IF (tr_a) THEN
         nrow = SIZE(target_m, 2)
         ncol = SIZE(target_m, 1)
      ELSE
         nrow = SIZE(target_m, 1)
         ncol = SIZE(target_m, 2)
      END IF
      IF (PRESENT(n_rows)) nrow = n_rows
      IF (PRESENT(n_cols)) ncol = n_cols

      para_env => fm%matrix_struct%para_env

      full_block => fm%local_data

      CALL cp_fm_get_info(matrix=fm, &
                          nrow_global=nrow_global, ncol_global=ncol_global, &
                          nrow_local=nrow_local, ncol_local=ncol_local, &
                          row_indices=row_indices, col_indices=col_indices)

      DO j = 1, ncol_local
         this_col = col_indices(j)-j0+1
         IF (this_col .GE. 1 .AND. this_col .LE. ncol) THEN
            IF (tr_a) THEN
               IF (i0 == 1 .AND. nrow_global == nrow) THEN
                  DO i = 1, nrow_local
                     target_m(this_col, row_indices(i)) = full_block(i, j)
                  END DO
               ELSE
                  DO i = 1, nrow_local
                     this_row = row_indices(i)-i0+1
                     IF (this_row >= 1 .AND. this_row <= nrow) THEN
                        target_m(this_col, this_row) = full_block(i, j)
                     END IF
                  END DO
               END IF
            ELSE
               IF (i0 == 1 .AND. nrow_global == nrow) THEN
                  DO i = 1, nrow_local
                     target_m(row_indices(i), this_col) = full_block(i, j)
                  END DO
               ELSE
                  DO i = 1, nrow_local
                     this_row = row_indices(i)-i0+1
                     IF (this_row >= 1 .AND. this_row <= nrow) THEN
                        target_m(this_row, this_col) = full_block(i, j)
                     END IF
                  END DO
               END IF
            END IF
         END IF
      END DO

      CALL mp_sum(target_m, para_env%group)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_get_submatrix

! **************************************************************************************************
!> \brief returns all kind of information about the full matrix
!> \param matrix ...
!> \param name ...
!> \param nrow_global ...
!> \param ncol_global ...
!> \param nrow_block ...
!> \param ncol_block ...
!> \param nrow_local ...
!> \param ncol_local ...
!> \param row_indices ...
!> \param col_indices ...
!> \param local_data ...
!> \param context ...
!> \param nrow_locals ...
!> \param ncol_locals ...
!> \param matrix_struct ...
!> \param para_env ...
!> \note
!>       see also cp_fm_struct for explaination
!>       - nrow_local, ncol_local, row_indices, col_indices, local_data are hooks for efficient
!>         access to the local blacs block
! **************************************************************************************************
   SUBROUTINE cp_fm_get_info(matrix, name, nrow_global, ncol_global, &
                             nrow_block, ncol_block, nrow_local, ncol_local, &
                             row_indices, col_indices, local_data, context, &
                             nrow_locals, ncol_locals, matrix_struct, para_env)

      TYPE(cp_fm_type), POINTER                  :: matrix
      CHARACTER(LEN=*), OPTIONAL, INTENT(OUT)    :: name
      INTEGER, OPTIONAL, INTENT(OUT)             :: ncol_block, ncol_global, &
                                                    nrow_block, nrow_global, &
                                                    nrow_local, ncol_local
      INTEGER, OPTIONAL, DIMENSION(:), POINTER   :: row_indices, col_indices, &
                                                    nrow_locals, ncol_locals
      TYPE(cp_para_env_type), POINTER, OPTIONAL  :: para_env
      TYPE(cp_blacs_env_type), POINTER, OPTIONAL :: context
      TYPE(cp_fm_struct_type), POINTER, OPTIONAL   :: matrix_struct
      REAL(KIND=dp), DIMENSION(:, :), POINTER, OPTIONAL :: local_data

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_get_info', &
                                     routineP = moduleN//':'//routineN

      IF (PRESENT(name)) name = matrix%name
      IF (PRESENT(context)) context => matrix%matrix_struct%context
      IF (PRESENT(matrix_struct)) matrix_struct => matrix%matrix_struct
      IF (PRESENT(local_data)) local_data => matrix%local_data ! not hiding things anymore :-(
      IF (PRESENT(row_indices)) THEN
         CALL cp_fm_struct_get(matrix%matrix_struct, row_indices=row_indices)
      ENDIF
      IF (PRESENT(col_indices)) THEN
         CALL cp_fm_struct_get(matrix%matrix_struct, col_indices=col_indices)
      ENDIF
      IF (PRESENT(nrow_locals)) THEN
         nrow_locals => matrix%matrix_struct%nrow_locals
      END IF
      IF (PRESENT(ncol_locals)) THEN
         ncol_locals => matrix%matrix_struct%ncol_locals
      END IF

      CALL cp_fm_struct_get(matrix%matrix_struct, nrow_local=nrow_local, &
                            ncol_local=ncol_local, nrow_global=nrow_global, &
                            ncol_global=ncol_global, nrow_block=nrow_block, &
                            ncol_block=ncol_block)

      IF (PRESENT(para_env)) para_env => matrix%matrix_struct%para_env

   END SUBROUTINE cp_fm_get_info

! **************************************************************************************************
!> \brief find the maximum absolute value of the matrix element
!>      maxval(abs(matrix))
!> \param matrix ...
!> \param a_max ...
!> \param ir_max ...
!> \param ic_max ...
! **************************************************************************************************
   SUBROUTINE cp_fm_maxabsval(matrix, a_max, ir_max, ic_max)
      TYPE(cp_fm_type), POINTER                          :: matrix
      REAL(KIND=dp), INTENT(OUT)                         :: a_max
      INTEGER, INTENT(OUT), OPTIONAL                     :: ir_max, ic_max

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_maxabsval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ic_max_local, ir_max_local, &
                                                            j, mepos, ncol_local, nrow_local, &
                                                            num_pe
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ic_max_vec, ir_max_vec
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(dp)                                           :: my_max
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: a_max_vec
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: my_block
      REAL(KIND=sp), DIMENSION(:, :), POINTER            :: my_block_sp

      CALL timeset(routineN, handle)

      my_block => matrix%local_data
      my_block_sp => matrix%local_data_sp

      CALL cp_fm_get_info(matrix, nrow_local=nrow_local, ncol_local=ncol_local, &
                          row_indices=row_indices, col_indices=col_indices)

      IF (matrix%use_sp) THEN
         a_max = REAL(MAXVAL(ABS(my_block_sp(1:nrow_local, 1:ncol_local))), dp)
      ELSE
         a_max = MAXVAL(ABS(my_block(1:nrow_local, 1:ncol_local)))
      ENDIF

      IF (PRESENT(ir_max)) THEN
         num_pe = matrix%matrix_struct%para_env%num_pe
         mepos = matrix%matrix_struct%para_env%mepos
         ALLOCATE (ir_max_vec(0:num_pe-1))
         ir_max_vec(0:num_pe-1) = 0
         ALLOCATE (ic_max_vec(0:num_pe-1))
         ic_max_vec(0:num_pe-1) = 0
         ALLOCATE (a_max_vec(0:num_pe-1))
         a_max_vec(0:num_pe-1) = 0.0_dp
         my_max = 0.0_dp

         IF ((ncol_local > 0) .AND. (nrow_local > 0)) THEN
            DO i = 1, ncol_local
               DO j = 1, nrow_local
                  IF (matrix%use_sp) THEN
                     IF (ABS(my_block_sp(j, i)) > my_max) THEN
                        my_max = REAL(my_block_sp(j, i), dp)
                        ir_max_local = j
                        ic_max_local = i
                     ENDIF
                  ELSE
                     IF (ABS(my_block(j, i)) > my_max) THEN
                        my_max = my_block(j, i)
                        ir_max_local = j
                        ic_max_local = i
                     ENDIF
                  END IF
               END DO
            END DO

            a_max_vec(mepos) = my_max
            ir_max_vec(mepos) = row_indices(ir_max_local)
            ic_max_vec(mepos) = col_indices(ic_max_local)

         END IF

         CALL mp_sum(a_max_vec, matrix%matrix_struct%para_env%group)
         CALL mp_sum(ir_max_vec, matrix%matrix_struct%para_env%group)
         CALL mp_sum(ic_max_vec, matrix%matrix_struct%para_env%group)

         my_max = 0.0_dp
         DO i = 0, num_pe-1
            IF (a_max_vec(i) > my_max) THEN
               ir_max = ir_max_vec(i)
               ic_max = ic_max_vec(i)
            END IF
         END DO

         DEALLOCATE (ir_max_vec, ic_max_vec, a_max_vec)
         IF(.NOT.(ic_max > 0))CALL cp__a("fm/cp_fm_types.F",992)
         IF(.NOT.(ir_max > 0))CALL cp__a("fm/cp_fm_types.F",993)

      END IF

      CALL mp_max(a_max, matrix%matrix_struct%para_env%group)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_maxabsval

! **************************************************************************************************
!> \brief find the maximum over the rows of the sum of the absolute values of the elements of a given row
!>      = || A ||_infinity
!> \param matrix ...
!> \param a_max ...
!> \note
!>      for a real symmetric matrix it holds that || A ||_2 = |lambda_max| < || A ||_infinity
!>      Hence this can be used to estimate an upper bound for the eigenvalues of a matrix
!>      http://mathworld.wolfram.com/MatrixNorm.html
!>      (but the bound is not so tight in the general case)
! **************************************************************************************************
   SUBROUTINE cp_fm_maxabsrownorm(matrix, a_max)
      TYPE(cp_fm_type), POINTER                          :: matrix
      REAL(KIND=dp), INTENT(OUT)                         :: a_max

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_maxabsrownorm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, ncol_local, nrow_global, &
                                                            nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: row_indices
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: values
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: my_block

      CALL timeset(routineN, handle)

      my_block => matrix%local_data

      IF(.NOT.(.NOT. matrix%use_sp))CALL cp__a("fm/cp_fm_types.F",1031)

      CALL cp_fm_get_info(matrix, row_indices=row_indices, nrow_global=nrow_global, &
                          nrow_local=nrow_local, ncol_local=ncol_local)

      ! the efficiency could be improved by making use of the row-col distribution of scalapack
      ALLOCATE (values(nrow_global))
      values = 0.0_dp
      DO j = 1, ncol_local
         DO i = 1, nrow_local
            values(row_indices(i)) = values(row_indices(i))+ABS(my_block(i, j))
         ENDDO
      ENDDO
      CALL mp_sum(values, matrix%matrix_struct%para_env%group)
      a_max = MAXVAL(values)
      DEALLOCATE (values)

      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief find the inorm of each column norm_{j}= sqrt( \sum_{i} A_{ij}*A_{ij} )
!> \param matrix ...
!> \param norm_array ...
! **************************************************************************************************
   SUBROUTINE cp_fm_vectorsnorm(matrix, norm_array)
      TYPE(cp_fm_type), POINTER                          :: matrix
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: norm_array

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_vectorsnorm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, ncol_global, ncol_local, &
                                                            nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: my_block

      CALL timeset(routineN, handle)

      my_block => matrix%local_data

      IF(.NOT.(.NOT. matrix%use_sp))CALL cp__a("fm/cp_fm_types.F",1072)

      CALL cp_fm_get_info(matrix, col_indices=col_indices, ncol_global=ncol_global, &
                          nrow_local=nrow_local, ncol_local=ncol_local)

      ! the efficiency could be improved by making use of the row-col distribution of scalapack
      norm_array = 0.0_dp
      DO j = 1, ncol_local
         DO i = 1, nrow_local
            norm_array(col_indices(j)) = norm_array(col_indices(j))+my_block(i, j)*my_block(i, j)
         ENDDO
      ENDDO
      CALL mp_sum(norm_array, matrix%matrix_struct%para_env%group)
      norm_array = SQRT(norm_array)

      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief summing up all the elements along the matrix's i-th index
!>        \f$ \mathrm{sum}_{j} = \sum_{i} A_{ij} \f$
!> \param matrix     an input matrix A
!> \param sum_array  sums of elements in each column
!> \note  forked from cp_fm_vectorsnorm() to be used with
!>        the maximum overlap method
! **************************************************************************************************
   SUBROUTINE cp_fm_vectorssum(matrix, sum_array)
      TYPE(cp_fm_type), POINTER                          :: matrix
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: sum_array

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_vectorssum', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, ncol_global, ncol_local, &
                                                            nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: my_block

      CALL timeset(routineN, handle)

      my_block => matrix%local_data

      IF(.NOT.(.NOT. matrix%use_sp))CALL cp__a("fm/cp_fm_types.F",1114)

      CALL cp_fm_get_info(matrix, col_indices=col_indices, ncol_global=ncol_global, &
                          nrow_local=nrow_local, ncol_local=ncol_local)

      ! the efficiency could be improved by making use of the row-col distribution of scalapack
      sum_array(:) = 0.0_dp
      DO j = 1, ncol_local
         DO i = 1, nrow_local
            sum_array(col_indices(j)) = sum_array(col_indices(j))+my_block(i, j)
         ENDDO
      ENDDO
      CALL mp_sum(sum_array, matrix%matrix_struct%para_env%group)

      CALL timestop(handle)
   END SUBROUTINE

! **************************************************************************************************
!> \brief copy one identically sized matrix in the other
!> \param source ...
!> \param destination ...
!> \note
!>      see also cp_fm_to_fm_columns
! **************************************************************************************************
   SUBROUTINE cp_fm_to_fm_matrix(source, destination)

      TYPE(cp_fm_type), POINTER                          :: source, destination

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_to_fm_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npcol, nprow

      CALL timeset(routineN, handle)

      nprow = source%matrix_struct%context%num_pe(1)
      npcol = source%matrix_struct%context%num_pe(2)

      IF ((.NOT. cp2k_is_parallel) .OR. &
          cp_fm_struct_equivalent(source%matrix_struct, &
                                  destination%matrix_struct)) THEN
         IF (source%use_sp .AND. destination%use_sp) THEN
            IF (SIZE(source%local_data_sp, 1) /= SIZE(destination%local_data_sp, 1) .OR. &
                SIZE(source%local_data_sp, 2) /= SIZE(destination%local_data_sp, 2)) &
               CALL cp_abort(cp__l("fm/cp_fm_types.F",1158), &
                             "Cannot copy full matrix <"//TRIM(source%name)// &
                             "> to full matrix <"//TRIM(destination%name)// &
                             ">. The local_data blocks have different sizes.")
            CALL scopy(SIZE(source%local_data_sp, 1)*SIZE(source%local_data_sp, 2), &
                       source%local_data_sp(1, 1), 1, destination%local_data_sp(1, 1), 1)
         ELSEIF (source%use_sp .AND. .NOT. destination%use_sp) THEN
            IF (SIZE(source%local_data_sp, 1) /= SIZE(destination%local_data, 1) .OR. &
                SIZE(source%local_data_sp, 2) /= SIZE(destination%local_data, 2)) &
               CALL cp_abort(cp__l("fm/cp_fm_types.F",1167), &
                             "Cannot copy full matrix <"//TRIM(source%name)// &
                             "> to full matrix <"//TRIM(destination%name)// &
                             ">. The local_data blocks have different sizes.")
            destination%local_data = REAL(source%local_data_sp, dp)
         ELSEIF (.NOT. source%use_sp .AND. destination%use_sp) THEN
            IF (SIZE(source%local_data, 1) /= SIZE(destination%local_data_sp, 1) .OR. &
                SIZE(source%local_data, 2) /= SIZE(destination%local_data_sp, 2)) &
               CALL cp_abort(cp__l("fm/cp_fm_types.F",1175), &
                             "Cannot copy full matrix <"//TRIM(source%name)// &
                             "> to full matrix <"//TRIM(destination%name)// &
                             ">. The local_data blocks have different sizes.")
            destination%local_data_sp = REAL(source%local_data, sp)
         ELSE
            IF (SIZE(source%local_data, 1) /= SIZE(destination%local_data, 1) .OR. &
                SIZE(source%local_data, 2) /= SIZE(destination%local_data, 2)) &
               CALL cp_abort(cp__l("fm/cp_fm_types.F",1183), &
                             "Cannot copy full matrix <"//TRIM(source%name)// &
                             "> to full matrix <"//TRIM(destination%name)// &
                             ">. The local_data blocks have different sizes.")
            CALL dcopy(SIZE(source%local_data, 1)*SIZE(source%local_data, 2), &
                       source%local_data(1, 1), 1, destination%local_data(1, 1), 1)
         ENDIF
      ELSE
         CALL cp__b("fm/cp_fm_types.F",1191,"")
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_to_fm_matrix

! **************************************************************************************************
!> \brief copy just a subset of columns of a fm to a fm
!> \param msource ...
!> \param mtarget ...
!> \param ncol ...
!> \param source_start ...
!> \param target_start ...
! **************************************************************************************************
   SUBROUTINE cp_fm_to_fm_columns(msource, mtarget, ncol, source_start, &
                                  target_start)

      TYPE(cp_fm_type), POINTER                :: msource, mtarget
      INTEGER, INTENT(IN)                      :: ncol
      INTEGER, INTENT(IN), OPTIONAL            :: source_start, target_start

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_to_fm_columns', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, n, ss, ts
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, b

      CALL timeset(routineN, handle)

      ss = 1
      ts = 1

      IF (PRESENT(source_start)) ss = source_start
      IF (PRESENT(target_start)) ts = target_start

      n = msource%matrix_struct%nrow_global

      a => msource%local_data
      b => mtarget%local_data

      CALL dcopy(ncol*n, a(1, ss), 1, b(1, ts), 1)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_to_fm_columns

! **************************************************************************************************
!> \brief copy just a triangular matrix
!> \param msource ...
!> \param mtarget ...
!> \param uplo ...
! **************************************************************************************************
   SUBROUTINE cp_fm_to_fm_triangular(msource, mtarget, uplo)

      TYPE(cp_fm_type), POINTER                :: msource, mtarget
      CHARACTER(LEN=*), INTENT(IN)             :: uplo

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_to_fm_triangular', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, ncol, nrow
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, b

      CALL timeset(routineN, handle)

      nrow = msource%matrix_struct%nrow_global
      ncol = msource%matrix_struct%ncol_global

      a => msource%local_data
      b => mtarget%local_data

      CALL dlacpy(uplo, nrow, ncol, a(1, 1), nrow, b(1, 1), nrow)

      CALL timestop(handle)

   END SUBROUTINE cp_fm_to_fm_triangular

! **************************************************************************************************
!> \brief copy just a part ot the matrix
!> \param msource ...
!> \param mtarget ...
!> \param nrow ...
!> \param ncol ...
!> \param s_firstrow ...
!> \param s_firstcol ...
!> \param t_firstrow ...
!> \param t_firstcol ...
! **************************************************************************************************

   SUBROUTINE cp_fm_to_fm_submat(msource, mtarget, nrow, ncol, s_firstrow, s_firstcol, t_firstrow, t_firstcol)

      TYPE(cp_fm_type), POINTER                :: msource, mtarget
      INTEGER, INTENT(IN)                      :: nrow, ncol, s_firstrow, &
                                                  s_firstcol, t_firstrow, &
                                                  t_firstcol

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_to_fm_submat', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, i, na, nb, ss, ts
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, b

      CALL timeset(routineN, handle)

      a => msource%local_data
      b => mtarget%local_data

      na = msource%matrix_struct%nrow_global
      nb = mtarget%matrix_struct%nrow_global
!    nrow must be <= na and nb
      IF (nrow > na) &
         CALL cp__b("fm/cp_fm_types.F",1327,"cannot copy because nrow > number of rows of source matrix")
      IF (nrow > nb) &
         CALL cp__b("fm/cp_fm_types.F",1329,"cannot copy because nrow > number of rows of target matrix")
      na = msource%matrix_struct%ncol_global
      nb = mtarget%matrix_struct%ncol_global
!    ncol must be <= na_col and nb_col
      IF (ncol > na) &
         CALL cp__b("fm/cp_fm_types.F",1334,"cannot copy because nrow > number of rows of source matrix")
      IF (ncol > nb) &
         CALL cp__b("fm/cp_fm_types.F",1336,"cannot copy because nrow > number of rows of target matrix")

      DO i = 0, ncol-1
         ss = s_firstcol+i
         ts = t_firstcol+i
         CALL dcopy(nrow, a(s_firstrow, ss), 1, b(t_firstrow, ts), 1)
      END DO

      CALL timestop(handle)
   END SUBROUTINE cp_fm_to_fm_submat

! **************************************************************************************************
!> \brief General copy of a fm matrix to another fm matrix.
!>        The two matrices can have different contexts.
!>        Matrices have to have the same global shape.
!>
!>        Summary of distribution routines for dense matrices
!>        The following will copy A(iA:iA+M-1,jA:jA+N-1) to B(iB:iB+M-1,jB:jB+N-1):
!>
!>        call pdgemr2d(M,N,Aloc,iA,jA,descA,Bloc,iB,jB,descB,context)
!>
!>        Here A and B are global matrices, Aloc, and Bloc, are local matrices,
!>        and context is a process grid that covers all parts of either A or B.
!>        If a whole global matrix is to be distributed, use the call:
!>
!>        call pdgemr2d(M,N,A,1,1,descA,Bloc,1,1,descB,context)
!>
!>        A process that is not a part of the context of A should set descA(2)
!>        to -1, and similarly for B.
!>
!> \param source          input fm matrix
!> \param destination     output fm matrix
!> \param global_context  process grid that covers all parts of either A or B.
! **************************************************************************************************
   SUBROUTINE cp_fm_copy_general(source, destination, global_context)

      TYPE(cp_fm_type), POINTER                :: source, destination
      INTEGER, INTENT(IN)                      :: global_context

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_copy_general', &
                                     routineP = moduleN//':'//routineN

      LOGICAL                                  :: debug
      INTEGER                                  :: handle

      CALL timeset(routineN, handle)

      debug = debug_this_module

      IF (.NOT. cp2k_is_parallel) THEN
         CALL cp_fm_to_fm(source, destination)
      ELSE
         IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(global_context))==-1) EXIT ;  END DO ; ENDIF
         CALL cp__b("fm/cp_fm_types.F",1446,"")
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_copy_general

! **************************************************************************************************
!> \brief General copy of a submatrixx of fm matrix to
!>        a submatrix of another fm matrix.
!>        The two matrices can have different contexts.
!>
!>        Summary of distribution routines for dense matrices
!>        The following will copy A(iA:iA+M-1,jA:jA+N-1) to B(iB:iB+M-1,jB:jB+N-1):
!>
!>        call pdgemr2d(M,N,Aloc,iA,jA,descA,Bloc,iB,jB,descB,context)
!>
!>        A process that is not a part of the context of A should set descA(2)
!>        to -1, and similarly for B.
!>
!> \param source          input fm matrix
!> \param destination     output fm matrix
!> \param nrows           number of rows of sub matrix to be copied
!> \param ncols           number of cols of sub matrix to be copied
!> \param s_firstrow      starting global row index of sub matrix in source
!> \param s_firstcol      starting global col index of sub matrix in source
!> \param d_firstrow      starting global row index of sub matrix in destination
!> \param d_firstcol      starting global col index of sub matrix in destination
!> \param global_context  process grid that covers all parts of either A or B.
! **************************************************************************************************
   SUBROUTINE cp_fm_to_fm_submat_general(source, &
                                         destination, &
                                         nrows, &
                                         ncols, &
                                         s_firstrow, &
                                         s_firstcol, &
                                         d_firstrow, &
                                         d_firstcol, &
                                         global_context)

      TYPE(cp_fm_type), POINTER                :: source, destination
      INTEGER, INTENT(IN)                      :: nrows, ncols, &
                                                  s_firstrow, s_firstcol, &
                                                  d_firstrow, d_firstcol, &
                                                  global_context

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_to_fm_submat_general', &
                                     routineP = moduleN//':'//routineN

      LOGICAL                                  :: debug
      INTEGER                                  :: handle

      CALL timeset(routineN, handle)

      debug = debug_this_module

      IF (.NOT. cp2k_is_parallel) THEN
         CALL cp_fm_to_fm_submat(source, &
                                 destination, &
                                 nrows, &
                                 ncols, &
                                 s_firstrow, &
                                 s_firstcol, &
                                 d_firstrow, &
                                 d_firstcol)
      ELSE
         IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(global_context))==-1) EXIT ;  END DO ; ENDIF
         CALL cp__b("fm/cp_fm_types.F",1560,"this subroutine only supports SCALAPACK")
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_to_fm_submat_general

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param irow_global ...
!> \param icol_global ...
!> \param alpha ...
! **************************************************************************************************
   SUBROUTINE cp_fm_add_to_element(matrix, irow_global, icol_global, alpha)

      ! Add alpha to the matrix element specified by the global indices
      ! irow_global and icol_global

      ! - Creation (05.05.06,MK)

      TYPE(cp_fm_type), POINTER                :: matrix
      INTEGER, INTENT(IN)                      :: irow_global, icol_global
      REAL(KIND=dp), INTENT(IN)                :: alpha

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_add_to_element', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: mypcol, myprow, npcol, nprow
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      TYPE(cp_blacs_env_type), POINTER         :: context

      IF(.NOT.(ASSOCIATED(matrix)))CALL cp__a("fm/cp_fm_types.F",1598)

      context => matrix%matrix_struct%context

      myprow = context%mepos(1)
      mypcol = context%mepos(2)

      nprow = context%num_pe(1)
      npcol = context%num_pe(2)

      a => matrix%local_data


      a(irow_global, icol_global) = a(irow_global, icol_global)+alpha


   END SUBROUTINE cp_fm_add_to_element

! **************************************************************************************************
!> \brief ...
!> \param fm ...
!> \param unit ...
! **************************************************************************************************
   SUBROUTINE cp_fm_write_unformatted(fm, unit)
      TYPE(cp_fm_type), POINTER                :: fm
      INTEGER                                  :: unit

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_write_unformatted', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, j, max_block, &
                                                  ncol_global, nrow_global
      TYPE(cp_para_env_type), POINTER          :: para_env

      CALL timeset(routineN, handle)
      CALL cp_fm_get_info(fm, nrow_global=nrow_global, ncol_global=ncol_global, ncol_block=max_block, &
                          para_env=para_env)


      IF (unit > 0) THEN
         DO j = 1, ncol_global
            WRITE (unit) fm%local_data(:, j)
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE cp_fm_write_unformatted

! **************************************************************************************************
!> \brief ...
!> \param fm ...
!> \param unit ...
! **************************************************************************************************
   SUBROUTINE cp_fm_read_unformatted(fm, unit)
      TYPE(cp_fm_type), POINTER                :: fm
      INTEGER                                  :: unit

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_read_unformatted', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, j, max_block, &
                                                  ncol_global, nrow_global
      TYPE(cp_para_env_type), POINTER          :: para_env

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(fm, nrow_global=nrow_global, ncol_global=ncol_global, ncol_block=max_block, &
                          para_env=para_env)


      DO j = 1, ncol_global
         READ (unit) fm%local_data(:, j)
      END DO


      CALL timestop(handle)

   END SUBROUTINE cp_fm_read_unformatted

! **************************************************************************************************
!> \brief wrapper to scalapack function INDXG2P that computes the process
!>         coordinate which posseses the entry of a distributed matrix specified
!>         by a global index INDXGLOB.
!>
!>  Arguments
!>  =========
!>
!>  INDXGLOB  (global input) INTEGER
!>            The global index of the element.
!>
!>  NB        (global input) INTEGER
!>            Block size, size of the blocks the distributed matrix is
!>            split into.
!>
!>  IPROC     (local dummy) INTEGER
!>            Dummy argument in this case in order to unify the calling
!>            sequence of the tool-routines.
!>
!>  ISRCPROC  (global input) INTEGER
!>            The coordinate of the process that possesses the first
!>            row/column of the distributed matrix.
!>
!>  NPROCS    (global input) INTEGER
!>            The total number processes over which the matrix is
!>            distributed.
!>
!> \param INDXGLOB ...
!> \param NB ...
!> \param IPROC ...
!> \param ISRCPROC ...
!> \param NPROCS ...
!> \retval G2P ...
!> \author Mauro Del Ben [MDB] - 12.2012
! **************************************************************************************************
   FUNCTION cp_fm_indxg2p(INDXGLOB, NB, IPROC, ISRCPROC, NPROCS) RESULT(G2P)
      INTEGER, INTENT(IN)                      :: INDXGLOB, NB, IPROC, ISRCPROC, NPROCS
      INTEGER                                  :: G2P

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_indxg2p', &
                                     routineP = moduleN//':'//routineN


      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(indxglob))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(iproc))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(isrcproc))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nb))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nprocs))==-1) EXIT ;  END DO ; ENDIF

      G2P = 0


   END FUNCTION cp_fm_indxg2p

! **************************************************************************************************
!> \brief wrapper to scalapack function INDXG2L that computes the local index
!>         of a distributed matrix entry pointed to by the global index INDXGLOB.
!>
!>  Arguments
!>  =========
!>
!>  INDXGLOB  (global input) INTEGER
!>            The global index of the distributed matrix entry.
!>
!>  NB        (global input) INTEGER
!>            Block size, size of the blocks the distributed matrix is
!>            split into.
!>
!>  IPROC     (local dummy) INTEGER
!>            Dummy argument in this case in order to unify the calling
!>            sequence of the tool-routines.
!>
!>  ISRCPROC  (local dummy) INTEGER
!>            Dummy argument in this case in order to unify the calling
!>            sequence of the tool-routines.
!>
!>  NPROCS    (global input) INTEGER
!>            The total number processes over which the distributed
!>            matrix is distributed.
!>
!> \param INDXGLOB ...
!> \param NB ...
!> \param IPROC ...
!> \param ISRCPROC ...
!> \param NPROCS ...
!> \retval G2L ...
!> \author Mauro Del Ben [MDB] - 12.2012
! **************************************************************************************************
   FUNCTION cp_fm_indxg2l(INDXGLOB, NB, IPROC, ISRCPROC, NPROCS) RESULT(G2L)
      INTEGER, INTENT(IN)                      :: INDXGLOB, NB, IPROC, ISRCPROC, NPROCS
      INTEGER                                  :: G2L

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_indxg2l', &
                                     routineP = moduleN//':'//routineN


      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(iproc))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(isrcproc))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nb))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nprocs))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(indxglob))==-1) EXIT ;  END DO ; ENDIF

      G2L = INDXGLOB


   END FUNCTION cp_fm_indxg2l

! **************************************************************************************************
!> \brief wrapper to scalapack function INDXL2G that computes the global index
!>         of a distributed matrix entry pointed to by the local index INDXLOC
!>         of the process indicated by IPROC.
!>
!>  Arguments
!>  =========
!>
!>  INDXLOC   (global input) INTEGER
!>            The local index of the distributed matrix entry.
!>
!>  NB        (global input) INTEGER
!>            Block size, size of the blocks the distributed matrix is
!>            split into.
!>
!>  IPROC     (local input) INTEGER
!>            The coordinate of the process whose local array row or
!>            column is to be determined.
!>
!>  ISRCPROC  (global input) INTEGER
!>            The coordinate of the process that possesses the first
!>            row/column of the distributed matrix.
!>
!>  NPROCS    (global input) INTEGER
!>            The total number processes over which the distributed
!>            matrix is distributed.
!>
!> \param INDXLOC ...
!> \param NB ...
!> \param IPROC ...
!> \param ISRCPROC ...
!> \param NPROCS ...
!> \retval L2G ...
!> \author Mauro Del Ben [MDB] - 12.2012
! **************************************************************************************************
   FUNCTION cp_fm_indxl2g(INDXLOC, NB, IPROC, ISRCPROC, NPROCS) RESULT(L2G)
      INTEGER, INTENT(IN)                      :: INDXLOC, NB, IPROC, ISRCPROC, NPROCS
      INTEGER                                  :: L2G

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_fm_indxl2g', &
                                     routineP = moduleN//':'//routineN


      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(iproc))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(isrcproc))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nb))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nprocs))==-1) EXIT ;  END DO ; ENDIF

      L2G = INDXLOC


   END FUNCTION cp_fm_indxl2g

! **************************************************************************************************
!> \brief ...
!> \param mult_type ...
! **************************************************************************************************
   SUBROUTINE cp_fm_setup(mult_type)
      INTEGER                                            :: mult_type

      mm_type = mult_type
   END SUBROUTINE cp_fm_setup

! **************************************************************************************************
!> \brief ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION cp_fm_get_mm_type() RESULT(res)
      INTEGER                                            :: res

      res = mm_type
   END FUNCTION

END MODULE cp_fm_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (20-12-2000) : Parallel data layout
!> \author APSI
! **************************************************************************************************
MODULE pw_grid_types

   USE kinds,                           ONLY: dp,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: pw_grid_type, map_pn

   ! (only for reciprocal grid:) fill in half or full space
   INTEGER, PARAMETER, PUBLIC :: HALFSPACE = 211, FULLSPACE = 212
   INTEGER, PARAMETER, PUBLIC :: PW_MODE_LOCAL = 0, PW_MODE_DISTRIBUTED = 1

   ! maps to positive and negative g-vectors
! **************************************************************************************************
   TYPE map_pn
      INTEGER, DIMENSION(:), POINTER :: pos, neg
   END TYPE map_pn

! info on parallelisation
   ! contains only significant information if mode == PW_MODE_DISTRIBUTED
! **************************************************************************************************
   TYPE pw_para_type
      INTEGER :: mode ! 0 = local = PW_MODE_LOCAL ; 1 = distributed = PW_MODE_DISTRIBUTED
      LOGICAL :: ray_distribution ! block or pencil distribution
      LOGICAL :: blocked ! block or pencil distribution
      INTEGER :: group ! MPI group id
      INTEGER :: my_pos ! Position within group
      INTEGER :: group_size ! # of Processors in group
      LOGICAL :: group_head ! Master process within group
      INTEGER :: group_head_id ! Id of group_head
      INTEGER, DIMENSION(:, :, :), POINTER :: yzp ! g-space rays (xy,k,pe)
      INTEGER, DIMENSION(:, :), POINTER :: yzq ! local inverse pointer of yzp
      INTEGER, DIMENSION(:), POINTER :: nyzray ! number of g-space rays (pe)
      INTEGER :: rs_group ! real space group (2-dim cart)
      INTEGER :: rs_mpo ! real space group position
      INTEGER, DIMENSION(2) :: rs_dims ! real space group dimensions
      INTEGER, DIMENSION(2) :: rs_pos ! real space group positions in grid
      INTEGER, DIMENSION(:, :, :, :), POINTER :: bo ! list of axis distribution
      INTEGER, DIMENSION(:), POINTER :: pos_of_x ! what my_pos holds a given x plane....should go: hard-codes to plane distributed
   END TYPE pw_para_type

   ! all you always wanted to know about grids, but were...
! **************************************************************************************************
   TYPE pw_grid_type
      INTEGER(int_8) :: ngpts ! # grid points
      INTEGER(int_8) :: ngpts_cut ! # grid points within cutoff
      INTEGER, DIMENSION(2, 3) :: bounds ! lower and upper bounds
      INTEGER, DIMENSION(3) :: npts ! # point in all directions
      INTEGER :: ngpts_local ! # grid points
      INTEGER :: ngpts_cut_local ! # grid points within cutoff
      INTEGER, DIMENSION(2, 3) :: bounds_local ! bounds on local process
      INTEGER, DIMENSION(3) :: npts_local ! local version of npts
      REAL(KIND=dp), DIMENSION(3) :: dr ! grid spacing
      REAL(KIND=dp), DIMENSION(3, 3) :: dh ! incremental cell matrix
      REAL(KIND=dp), DIMENSION(3, 3) :: dh_inv ! inverse incremental cell matrix
      LOGICAL :: orthorhombic ! cell symmetry
      REAL(KIND=dp) :: dvol, vol ! volume element, volume
      REAL(KIND=dp) :: cutoff ! cutoff in a.u.
      TYPE(map_pn) :: mapl, mapm, mapn ! mapping 1D => 3D
      TYPE(pw_para_type) :: para ! information on parallelisation
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: g ! grid point vectors
      REAL(KIND=dp), DIMENSION(:), POINTER :: gsq ! squared vector lengths
      INTEGER, DIMENSION(:, :), POINTER :: g_hat ! grid point indices (Miller)
      INTEGER, DIMENSION(:, :), POINTER :: g_hatmap ! mapped grid point indices (Miller) [CUDA]
      INTEGER :: grid_span ! type HALFSPACE/FULLSPACE
      LOGICAL :: have_g0 ! whether I have G = [0,0,0]
      INTEGER :: first_gne0 ! first g index /= 0 [1/2]
      INTEGER :: id_nr ! tag of this grid
      INTEGER :: reference ! reference grid identifier
      INTEGER, DIMENSION(:), POINTER :: gidx ! ref grid index
      INTEGER :: ref_count ! reference count
      LOGICAL :: spherical ! spherical cutoff?
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: grays ! used by parallel 3D FFT routine
   END TYPE pw_grid_type

END MODULE pw_grid_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Defines the basic variable types
!> \author Teodoro Laino - 01.2011
! **************************************************************************************************

MODULE fft_kinds

   IMPLICIT NONE
   PRIVATE
   INTEGER, PARAMETER, PUBLIC :: wp = SELECTED_REAL_KIND(14, 200)
   INTEGER, PARAMETER, PUBLIC :: dp = SELECTED_REAL_KIND(14, 200)
   INTEGER, PARAMETER, PUBLIC :: sp = SELECTED_REAL_KIND(6, 30)

   INTEGER, PARAMETER, PUBLIC :: integer8_kind = SELECTED_INT_KIND(10)

END MODULE fft_kinds
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE mltfftsg_tools
   USE ISO_C_BINDING,                   ONLY: C_F_POINTER,&
                                              C_LOC
   USE fft_kinds,                       ONLY: dp

!$ USE OMP_LIB, ONLY: omp_get_num_threads, omp_get_thread_num

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   PRIVATE
   INTEGER, PARAMETER :: ctrig_length = 1024
   INTEGER, PARAMETER :: cache_size = 2048
   PUBLIC :: mltfftsg

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param transa ...
!> \param transb ...
!> \param a ...
!> \param ldax ...
!> \param lday ...
!> \param b ...
!> \param ldbx ...
!> \param ldby ...
!> \param n ...
!> \param m ...
!> \param isign ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE mltfftsg(transa, transb, a, ldax, lday, b, ldbx, ldby, n, m, isign, scale)

      CHARACTER(LEN=1), INTENT(IN)                       :: transa, transb
      INTEGER, INTENT(IN)                                :: ldax, lday
      COMPLEX(dp), INTENT(INOUT)                         :: a(ldax, lday)
      INTEGER, INTENT(IN)                                :: ldbx, ldby
      COMPLEX(dp), INTENT(INOUT)                         :: b(ldbx, ldby)
      INTEGER, INTENT(IN)                                :: n, m, isign
      REAL(dp), INTENT(IN)                               :: scale

      COMPLEX(dp), ALLOCATABLE, DIMENSION(:, :, :)       :: z
      INTEGER                                            :: after(20), before(20), chunk, i, ic, id, &
                                                            iend, inzee, isig, istart, iterations, &
                                                            itr, length, lot, nfft, now(20), &
                                                            num_threads
      LOGICAL                                            :: tscal
      REAL(dp)                                           :: trig(2, 1024)

! Variables

      LENGTH = 2*(cache_size/4+1)

      ISIG = -ISIGN
      TSCAL = (ABS(SCALE-1._dp) > 1.e-12_dp)
      CALL ctrig(N, TRIG, AFTER, BEFORE, NOW, ISIG, IC)
      LOT = cache_size/(4*N)
      LOT = LOT-MOD(LOT+1, 2)
      LOT = MAX(1, LOT)

      ! initializations for serial mode
      id = 0; num_threads = 1

!$OMP PARALLEL &
!$OMP PRIVATE ( id, istart, iend, nfft, i, inzee, itr) DEFAULT(NONE) &
!$OMP SHARED (NUM_THREADS,z,iterations,chunk,LOT,length,m,transa,isig, &
!$OMP         before,after,now,trig,A,n,ldax,tscal,scale,ic,transb,ldbx,b)

!$OMP SINGLE
!$    num_threads = omp_get_num_threads()
      ALLOCATE (Z(LENGTH, 2, 0:num_threads-1))
      iterations = (M+LOT-1)/LOT
      chunk = LOT*((iterations+num_threads-1)/num_threads)
!$OMP END SINGLE
!$OMP BARRIER

!$    id = omp_get_thread_num()
      istart = id*chunk+1
      iend = MIN((id+1)*chunk, M)

      DO ITR = istart, iend, LOT

         NFFT = MIN(M-ITR+1, LOT)
         IF (TRANSA == 'N' .OR. TRANSA == 'n') THEN
            CALL fftpre_cmplx(NFFT, NFFT, LDAX, LOT, N, A(1, ITR), Z(1, 1, id), &
                              TRIG, NOW(1), AFTER(1), BEFORE(1), ISIG)
         ELSE
            CALL fftstp_cmplx(LDAX, NFFT, N, LOT, N, A(ITR, 1), Z(1, 1, id), &
                              TRIG, NOW(1), AFTER(1), BEFORE(1), ISIG)
         ENDIF
         IF (TSCAL) THEN
            IF (LOT == NFFT) THEN
               CALL scaled(2*LOT*N, SCALE, Z(1, 1, id))
            ELSE
               DO I = 1, N
                  CALL scaled(2*NFFT, SCALE, Z(LOT*(I-1)+1, 1, id))
               END DO
            END IF
         END IF
         IF (IC .EQ. 1) THEN
            IF (TRANSB == 'N' .OR. TRANSB == 'n') THEN
               CALL zgetmo(Z(1, 1, id), LOT, NFFT, N, B(1, ITR), LDBX)
            ELSE
               CALL matmov(NFFT, N, Z(1, 1, id), LOT, B(ITR, 1), LDBX)
            ENDIF
         ELSE
            INZEE = 1
            DO I = 2, IC-1
               CALL fftstp_cmplx(LOT, NFFT, N, LOT, N, Z(1, INZEE, id), &
                                 Z(1, 3-INZEE, id), TRIG, NOW(I), AFTER(I), &
                                 BEFORE(I), ISIG)
               INZEE = 3-INZEE
            ENDDO
            IF (TRANSB == 'N' .OR. TRANSB == 'n') THEN
               CALL fftrot_cmplx(LOT, NFFT, N, NFFT, LDBX, Z(1, INZEE, id), &
                                 B(1, ITR), TRIG, NOW(IC), AFTER(IC), BEFORE(IC), ISIG)
            ELSE
               CALL fftstp_cmplx(LOT, NFFT, N, LDBX, N, Z(1, INZEE, id), &
                                 B(ITR, 1), TRIG, NOW(IC), AFTER(IC), BEFORE(IC), ISIG)
            ENDIF
         ENDIF
      ENDDO

!$OMP END PARALLEL

      DEALLOCATE (Z)

      IF (TRANSB == 'N' .OR. TRANSB == 'n') THEN
         B(1:LDBX, M+1:LDBY) = CMPLX(0._dp, 0._dp, dp)
         B(N+1:LDBX, 1:M) = CMPLX(0._dp, 0._dp, dp)
      ELSE
         B(1:LDBX, N+1:LDBY) = CMPLX(0._dp, 0._dp, dp)
         B(M+1:LDBX, 1:N) = CMPLX(0._dp, 0._dp, dp)
      ENDIF

   END SUBROUTINE mltfftsg

! this formalizes what we have been assuming before, call with a complex(*) array, and passing to a real(2,*)
! **************************************************************************************************
!> \brief ...
!> \param mm ...
!> \param nfft ...
!> \param m ...
!> \param nn ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param trig ...
!> \param now ...
!> \param after ...
!> \param before ...
!> \param isign ...
! **************************************************************************************************
   SUBROUTINE fftstp_cmplx(mm, nfft, m, nn, n, zin, zout, trig, now, after, before, isign)

      INTEGER, INTENT(IN)                                :: mm, nfft, m, nn, n
      COMPLEX(dp), DIMENSION(mm, m), INTENT(IN), TARGET  :: zin
      COMPLEX(dp), DIMENSION(nn, n), INTENT(INOUT), &
         TARGET                                          :: zout
      REAL(dp), DIMENSION(2, ctrig_length), INTENT(IN)   :: trig
      INTEGER, INTENT(IN)                                :: now, after, before, isign

      REAL(dp), DIMENSION(:, :, :), POINTER              :: zin_real, zout_real

      CALL C_F_POINTER(C_LOC(zin), zin_real, (/2, mm, m/))
      CALL C_F_POINTER(C_LOC(zout), zout_real, (/2, nn, n/))
      CALL fftstp(mm, nfft, m, nn, n, zin_real, zout_real, trig, now, after, before, isign)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param mm ...
!> \param nfft ...
!> \param m ...
!> \param nn ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param trig ...
!> \param now ...
!> \param after ...
!> \param before ...
!> \param isign ...
! **************************************************************************************************
   SUBROUTINE fftpre_cmplx(mm, nfft, m, nn, n, zin, zout, trig, now, after, before, isign)

      INTEGER, INTENT(IN)                                :: mm, nfft, m, nn, n
      COMPLEX(dp), DIMENSION(m, mm), INTENT(IN), TARGET  :: zin
      COMPLEX(dp), DIMENSION(nn, n), INTENT(INOUT), &
         TARGET                                          :: zout
      REAL(dp), DIMENSION(2, ctrig_length), INTENT(IN)   :: trig
      INTEGER, INTENT(IN)                                :: now, after, before, isign

      REAL(dp), DIMENSION(:, :, :), POINTER              :: zin_real, zout_real

      CALL C_F_POINTER(C_LOC(zin), zin_real, (/2, mm, m/))
      CALL C_F_POINTER(C_LOC(zout), zout_real, (/2, nn, n/))

      CALL fftpre(mm, nfft, m, nn, n, zin_real, zout_real, trig, now, after, before, isign)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param mm ...
!> \param nfft ...
!> \param m ...
!> \param nn ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param trig ...
!> \param now ...
!> \param after ...
!> \param before ...
!> \param isign ...
! **************************************************************************************************
   SUBROUTINE fftrot_cmplx(mm, nfft, m, nn, n, zin, zout, trig, now, after, before, isign)

  USE fft_kinds, ONLY     : dp
      INTEGER, INTENT(IN)                                :: mm, nfft, m, nn, n
      COMPLEX(dp), DIMENSION(mm, m), INTENT(IN), TARGET  :: zin
      COMPLEX(dp), DIMENSION(n, nn), INTENT(INOUT), &
         TARGET                                          :: zout
      REAL(dp), DIMENSION(2, ctrig_length), INTENT(IN)   :: trig
      INTEGER, INTENT(IN)                                :: now, after, before, isign

      REAL(dp), DIMENSION(:, :, :), POINTER              :: zin_real, zout_real

      CALL C_F_POINTER(C_LOC(zin), zin_real, (/2, mm, m/))
      CALL C_F_POINTER(C_LOC(zout), zout_real, (/2, nn, n/))

      CALL fftrot(mm, nfft, m, nn, n, zin_real, zout_real, trig, now, after, before, isign)

   END SUBROUTINE

!-----------------------------------------------------------------------------!
!  Copyright (C) Stefan Goedecker, Lausanne, Switzerland, August 1, 1991
!  Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!  Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!  This file is distributed under the terms of the
!  GNU General Public License version 2 (or later),
!  see http://www.gnu.org/copyleft/gpl.txt .
!-----------------------------------------------------------------------------!
!  S. Goedecker: Rotating a three-dimensional array in optimal
!  positions for vector processing: Case study for a three-dimensional Fast
!  Fourier Transform, Comp. Phys. Commun. 76, 294 (1993)
! **************************************************************************************************
!> \brief ...
!> \param mm ...
!> \param nfft ...
!> \param m ...
!> \param nn ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param trig ...
!> \param now ...
!> \param after ...
!> \param before ...
!> \param isign ...
! **************************************************************************************************
   SUBROUTINE fftrot(mm, nfft, m, nn, n, zin, zout, trig, now, after, before, isign)

  USE fft_kinds, ONLY     : dp
      INTEGER, INTENT(IN)                                :: mm, nfft, m, nn, n
      REAL(dp), DIMENSION(2, mm, m), INTENT(IN)          :: zin
      REAL(dp), DIMENSION(2, n, nn), INTENT(INOUT)       :: zout
      REAL(dp), DIMENSION(2, ctrig_length), INTENT(IN)   :: trig
      INTEGER, INTENT(IN)                                :: now, after, before, isign

      REAL(dp), PARAMETER :: bb = 0.8660254037844387_dp, cos2 = 0.3090169943749474_dp, &
         cos4 = -0.8090169943749474_dp, rt2i = 0.7071067811865475_dp, &
         sin2p = 0.9510565162951536_dp, sin4p = 0.5877852522924731_dp

      INTEGER                                            :: atb, atn, ia, ias, ib, itrig, itt, j, &
                                                            nin1, nin2, nin3, nin4, nin5, nin6, &
                                                            nin7, nin8, nout1, nout2, nout3, &
                                                            nout4, nout5, nout6, nout7, nout8
      REAL(dp) :: am, ap, bbs, bm, bp, ci2, ci3, ci4, ci5, cm, cp, cr2, cr3, cr4, cr5, dbl, dm, r, &
         r1, r2, r25, r3, r34, r4, r5, r6, r7, r8, s, s1, s2, s25, s3, s34, s4, s5, s6, s7, s8, &
         sin2, sin4, ui1, ui2, ui3, ur1, ur2, ur3, vi1, vi2, vi3, vr1, vr2, vr3

! sqrt(0.5)
! sqrt(3)/2
! cos(2*pi/5)
!  cos(4*pi/5)
! sin(2*pi/5)
! sin(4*pi/5)
!-----------------------------------------------------------------------------!

      atn = after*now
      atb = after*before

      IF (now == 4) THEN
         IF (isign == 1) THEN
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               DO j = 1, nfft
                  r1 = zin(1, j, nin1)
                  s1 = zin(2, j, nin1)
                  r2 = zin(1, j, nin2)
                  s2 = zin(2, j, nin2)
                  r3 = zin(1, j, nin3)
                  s3 = zin(2, j, nin3)
                  r4 = zin(1, j, nin4)
                  s4 = zin(2, j, nin4)
                  r = r1+r3
                  s = r2+r4
                  zout(1, nout1, j) = r+s
                  zout(1, nout3, j) = r-s
                  r = r1-r3
                  s = s2-s4
                  zout(1, nout2, j) = r-s
                  zout(1, nout4, j) = r+s
                  r = s1+s3
                  s = s2+s4
                  zout(2, nout1, j) = r+s
                  zout(2, nout3, j) = r-s
                  r = s1-s3
                  s = r2-r4
                  zout(2, nout2, j) = r+s
                  zout(2, nout4, j) = r-s
               END DO
            END DO
            DO ia = 2, after
               ias = ia-1
               IF (2*ias == after) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, j, nin3)
                        s3 = zin(1, j, nin3)
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = -(r+s)*rt2i
                        s4 = (r-s)*rt2i
                        r = r1+r3
                        s = r2+r4
                        zout(1, nout1, j) = r+s
                        zout(1, nout3, j) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, nout2, j) = r-s
                        zout(1, nout4, j) = r+s
                        r = s1+s3
                        s = s2+s4
                        zout(2, nout1, j) = r+s
                        zout(2, nout3, j) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, nout2, j) = r+s
                        zout(2, nout4, j) = r-s
                     END DO
                  END DO
               ELSE
                  itt = ias*before
                  itrig = itt+1
                  cr2 = trig(1, itrig)
                  ci2 = trig(2, itrig)
                  itrig = itrig+itt
                  cr3 = trig(1, itrig)
                  ci3 = trig(2, itrig)
                  itrig = itrig+itt
                  cr4 = trig(1, itrig)
                  ci4 = trig(2, itrig)
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = r*cr2-s*ci2
                        s2 = r*ci2+s*cr2
                        r = zin(1, j, nin3)
                        s = zin(2, j, nin3)
                        r3 = r*cr3-s*ci3
                        s3 = r*ci3+s*cr3
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = r*cr4-s*ci4
                        s4 = r*ci4+s*cr4
                        r = r1+r3
                        s = r2+r4
                        zout(1, nout1, j) = r+s
                        zout(1, nout3, j) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, nout2, j) = r-s
                        zout(1, nout4, j) = r+s
                        r = s1+s3
                        s = s2+s4
                        zout(2, nout1, j) = r+s
                        zout(2, nout3, j) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, nout2, j) = r+s
                        zout(2, nout4, j) = r-s
                     END DO
                  END DO
               END IF
            END DO
         ELSE
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               DO j = 1, nfft
                  r1 = zin(1, j, nin1)
                  s1 = zin(2, j, nin1)
                  r2 = zin(1, j, nin2)
                  s2 = zin(2, j, nin2)
                  r3 = zin(1, j, nin3)
                  s3 = zin(2, j, nin3)
                  r4 = zin(1, j, nin4)
                  s4 = zin(2, j, nin4)
                  r = r1+r3
                  s = r2+r4
                  zout(1, nout1, j) = r+s
                  zout(1, nout3, j) = r-s
                  r = r1-r3
                  s = s2-s4
                  zout(1, nout2, j) = r+s
                  zout(1, nout4, j) = r-s
                  r = s1+s3
                  s = s2+s4
                  zout(2, nout1, j) = r+s
                  zout(2, nout3, j) = r-s
                  r = s1-s3
                  s = r2-r4
                  zout(2, nout2, j) = r-s
                  zout(2, nout4, j) = r+s
               END DO
            END DO
            DO ia = 2, after
               ias = ia-1
               IF (2*ias == after) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r+s)*rt2i
                        s2 = (s-r)*rt2i
                        r3 = zin(2, j, nin3)
                        s3 = -zin(1, j, nin3)
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = (s-r)*rt2i
                        s4 = -(r+s)*rt2i
                        r = r1+r3
                        s = r2+r4
                        zout(1, nout1, j) = r+s
                        zout(1, nout3, j) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, nout2, j) = r+s
                        zout(1, nout4, j) = r-s
                        r = s1+s3
                        s = s2+s4
                        zout(2, nout1, j) = r+s
                        zout(2, nout3, j) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, nout2, j) = r-s
                        zout(2, nout4, j) = r+s
                     END DO
                  END DO
               ELSE
                  itt = ias*before
                  itrig = itt+1
                  cr2 = trig(1, itrig)
                  ci2 = trig(2, itrig)
                  itrig = itrig+itt
                  cr3 = trig(1, itrig)
                  ci3 = trig(2, itrig)
                  itrig = itrig+itt
                  cr4 = trig(1, itrig)
                  ci4 = trig(2, itrig)
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = r*cr2-s*ci2
                        s2 = r*ci2+s*cr2
                        r = zin(1, j, nin3)
                        s = zin(2, j, nin3)
                        r3 = r*cr3-s*ci3
                        s3 = r*ci3+s*cr3
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = r*cr4-s*ci4
                        s4 = r*ci4+s*cr4
                        r = r1+r3
                        s = r2+r4
                        zout(1, nout1, j) = r+s
                        zout(1, nout3, j) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, nout2, j) = r+s
                        zout(1, nout4, j) = r-s
                        r = s1+s3
                        s = s2+s4
                        zout(2, nout1, j) = r+s
                        zout(2, nout3, j) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, nout2, j) = r-s
                        zout(2, nout4, j) = r+s
                     END DO
                  END DO
               END IF
            END DO
         END IF
      ELSE IF (now == 8) THEN
         IF (isign == -1) THEN
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nin5 = nin4+atb
               nin6 = nin5+atb
               nin7 = nin6+atb
               nin8 = nin7+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               nout5 = nout4+after
               nout6 = nout5+after
               nout7 = nout6+after
               nout8 = nout7+after
               DO j = 1, nfft
                  r1 = zin(1, j, nin1)
                  s1 = zin(2, j, nin1)
                  r2 = zin(1, j, nin2)
                  s2 = zin(2, j, nin2)
                  r3 = zin(1, j, nin3)
                  s3 = zin(2, j, nin3)
                  r4 = zin(1, j, nin4)
                  s4 = zin(2, j, nin4)
                  r5 = zin(1, j, nin5)
                  s5 = zin(2, j, nin5)
                  r6 = zin(1, j, nin6)
                  s6 = zin(2, j, nin6)
                  r7 = zin(1, j, nin7)
                  s7 = zin(2, j, nin7)
                  r8 = zin(1, j, nin8)
                  s8 = zin(2, j, nin8)
                  r = r1+r5
                  s = r3+r7
                  ap = r+s
                  am = r-s
                  r = r2+r6
                  s = r4+r8
                  bp = r+s
                  bm = r-s
                  r = s1+s5
                  s = s3+s7
                  cp = r+s
                  cm = r-s
                  r = s2+s6
                  s = s4+s8
                  dbl = r+s
                  dm = r-s
                  zout(1, nout1, j) = ap+bp
                  zout(2, nout1, j) = cp+dbl
                  zout(1, nout5, j) = ap-bp
                  zout(2, nout5, j) = cp-dbl
                  zout(1, nout3, j) = am+dm
                  zout(2, nout3, j) = cm-bm
                  zout(1, nout7, j) = am-dm
                  zout(2, nout7, j) = cm+bm
                  r = r1-r5
                  s = s3-s7
                  ap = r+s
                  am = r-s
                  r = s1-s5
                  s = r3-r7
                  bp = r+s
                  bm = r-s
                  r = s4-s8
                  s = r2-r6
                  cp = r+s
                  cm = r-s
                  r = s2-s6
                  s = r4-r8
                  dbl = r+s
                  dm = r-s
                  r = (cp+dm)*rt2i
                  s = (-cp+dm)*rt2i
                  cp = (cm+dbl)*rt2i
                  dbl = (cm-dbl)*rt2i
                  zout(1, nout2, j) = ap+r
                  zout(2, nout2, j) = bm+s
                  zout(1, nout6, j) = ap-r
                  zout(2, nout6, j) = bm-s
                  zout(1, nout4, j) = am+cp
                  zout(2, nout4, j) = bp+dbl
                  zout(1, nout8, j) = am-cp
                  zout(2, nout8, j) = bp-dbl
               END DO
            END DO
         ELSE
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nin5 = nin4+atb
               nin6 = nin5+atb
               nin7 = nin6+atb
               nin8 = nin7+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               nout5 = nout4+after
               nout6 = nout5+after
               nout7 = nout6+after
               nout8 = nout7+after
               DO j = 1, nfft
                  r1 = zin(1, j, nin1)
                  s1 = zin(2, j, nin1)
                  r2 = zin(1, j, nin2)
                  s2 = zin(2, j, nin2)
                  r3 = zin(1, j, nin3)
                  s3 = zin(2, j, nin3)
                  r4 = zin(1, j, nin4)
                  s4 = zin(2, j, nin4)
                  r5 = zin(1, j, nin5)
                  s5 = zin(2, j, nin5)
                  r6 = zin(1, j, nin6)
                  s6 = zin(2, j, nin6)
                  r7 = zin(1, j, nin7)
                  s7 = zin(2, j, nin7)
                  r8 = zin(1, j, nin8)
                  s8 = zin(2, j, nin8)
                  r = r1+r5
                  s = r3+r7
                  ap = r+s
                  am = r-s
                  r = r2+r6
                  s = r4+r8
                  bp = r+s
                  bm = r-s
                  r = s1+s5
                  s = s3+s7
                  cp = r+s
                  cm = r-s
                  r = s2+s6
                  s = s4+s8
                  dbl = r+s
                  dm = r-s
                  zout(1, nout1, j) = ap+bp
                  zout(2, nout1, j) = cp+dbl
                  zout(1, nout5, j) = ap-bp
                  zout(2, nout5, j) = cp-dbl
                  zout(1, nout3, j) = am-dm
                  zout(2, nout3, j) = cm+bm
                  zout(1, nout7, j) = am+dm
                  zout(2, nout7, j) = cm-bm
                  r = r1-r5
                  s = -s3+s7
                  ap = r+s
                  am = r-s
                  r = s1-s5
                  s = r7-r3
                  bp = r+s
                  bm = r-s
                  r = -s4+s8
                  s = r2-r6
                  cp = r+s
                  cm = r-s
                  r = -s2+s6
                  s = r4-r8
                  dbl = r+s
                  dm = r-s
                  r = (cp+dm)*rt2i
                  s = (cp-dm)*rt2i
                  cp = (cm+dbl)*rt2i
                  dbl = (-cm+dbl)*rt2i
                  zout(1, nout2, j) = ap+r
                  zout(2, nout2, j) = bm+s
                  zout(1, nout6, j) = ap-r
                  zout(2, nout6, j) = bm-s
                  zout(1, nout4, j) = am+cp
                  zout(2, nout4, j) = bp+dbl
                  zout(1, nout8, j) = am-cp
                  zout(2, nout8, j) = bp-dbl
               END DO
            END DO
         END IF
      ELSE IF (now == 3) THEN
         bbs = isign*bb
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            DO j = 1, nfft
               r1 = zin(1, j, nin1)
               s1 = zin(2, j, nin1)
               r2 = zin(1, j, nin2)
               s2 = zin(2, j, nin2)
               r3 = zin(1, j, nin3)
               s3 = zin(2, j, nin3)
               r = r2+r3
               s = s2+s3
               zout(1, nout1, j) = r+r1
               zout(2, nout1, j) = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r2 = bbs*(r2-r3)
               s2 = bbs*(s2-s3)
               zout(1, nout2, j) = r1-s2
               zout(2, nout2, j) = s1+r2
               zout(1, nout3, j) = r1+s2
               zout(2, nout3, j) = s1-r2
            END DO
         END DO
         DO ia = 2, after
            ias = ia-1
            IF (4*ias == 3*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r2 = -zin(2, j, nin2)
                        s2 = zin(1, j, nin2)
                        r3 = -zin(1, j, nin3)
                        s3 = -zin(2, j, nin3)
                        r = r2+r3
                        s = s2+s3
                        zout(1, nout1, j) = r+r1
                        zout(2, nout1, j) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, nout2, j) = r1-s2
                        zout(2, nout2, j) = s1+r2
                        zout(1, nout3, j) = r1+s2
                        zout(2, nout3, j) = s1-r2
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r2 = zin(2, j, nin2)
                        s2 = -zin(1, j, nin2)
                        r3 = -zin(1, j, nin3)
                        s3 = -zin(2, j, nin3)
                        r = r2+r3
                        s = s2+s3
                        zout(1, nout1, j) = r+r1
                        zout(2, nout1, j) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, nout2, j) = r1-s2
                        zout(2, nout2, j) = s1+r2
                        zout(1, nout3, j) = r1+s2
                        zout(2, nout3, j) = s1-r2
                     END DO
                  END DO
               END IF
            ELSE IF (8*ias == 3*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, j, nin3)
                        s3 = zin(1, j, nin3)
                        r = r2+r3
                        s = s2+s3
                        zout(1, nout1, j) = r+r1
                        zout(2, nout1, j) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, nout2, j) = r1-s2
                        zout(2, nout2, j) = s1+r2
                        zout(1, nout3, j) = r1+s2
                        zout(2, nout3, j) = s1-r2
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r+s)*rt2i
                        s2 = (-r+s)*rt2i
                        r3 = zin(2, j, nin3)
                        s3 = -zin(1, j, nin3)
                        r = r2+r3
                        s = s2+s3
                        zout(1, nout1, j) = r+r1
                        zout(2, nout1, j) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, nout2, j) = r1-s2
                        zout(2, nout2, j) = s1+r2
                        zout(1, nout3, j) = r1+s2
                        zout(2, nout3, j) = s1-r2
                     END DO
                  END DO
               END IF
            ELSE
               itt = ias*before
               itrig = itt+1
               cr2 = trig(1, itrig)
               ci2 = trig(2, itrig)
               itrig = itrig+itt
               cr3 = trig(1, itrig)
               ci3 = trig(2, itrig)
               nin1 = ia-after
               nout1 = ia-atn
               DO ib = 1, before
                  nin1 = nin1+after
                  nin2 = nin1+atb
                  nin3 = nin2+atb
                  nout1 = nout1+atn
                  nout2 = nout1+after
                  nout3 = nout2+after
                  DO j = 1, nfft
                     r1 = zin(1, j, nin1)
                     s1 = zin(2, j, nin1)
                     r = zin(1, j, nin2)
                     s = zin(2, j, nin2)
                     r2 = r*cr2-s*ci2
                     s2 = r*ci2+s*cr2
                     r = zin(1, j, nin3)
                     s = zin(2, j, nin3)
                     r3 = r*cr3-s*ci3
                     s3 = r*ci3+s*cr3
                     r = r2+r3
                     s = s2+s3
                     zout(1, nout1, j) = r+r1
                     zout(2, nout1, j) = s+s1
                     r1 = r1-0.5_dp*r
                     s1 = s1-0.5_dp*s
                     r2 = bbs*(r2-r3)
                     s2 = bbs*(s2-s3)
                     zout(1, nout2, j) = r1-s2
                     zout(2, nout2, j) = s1+r2
                     zout(1, nout3, j) = r1+s2
                     zout(2, nout3, j) = s1-r2
                  END DO
               END DO
            END IF
         END DO
      ELSE IF (now == 5) THEN
         sin2 = isign*sin2p
         sin4 = isign*sin4p
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nin4 = nin3+atb
            nin5 = nin4+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            nout4 = nout3+after
            nout5 = nout4+after
            DO j = 1, nfft
               r1 = zin(1, j, nin1)
               s1 = zin(2, j, nin1)
               r2 = zin(1, j, nin2)
               s2 = zin(2, j, nin2)
               r3 = zin(1, j, nin3)
               s3 = zin(2, j, nin3)
               r4 = zin(1, j, nin4)
               s4 = zin(2, j, nin4)
               r5 = zin(1, j, nin5)
               s5 = zin(2, j, nin5)
               r25 = r2+r5
               r34 = r3+r4
               s25 = s2-s5
               s34 = s3-s4
               zout(1, nout1, j) = r1+r25+r34
               r = cos2*r25+cos4*r34+r1
               s = sin2*s25+sin4*s34
               zout(1, nout2, j) = r-s
               zout(1, nout5, j) = r+s
               r = cos4*r25+cos2*r34+r1
               s = sin4*s25-sin2*s34
               zout(1, nout3, j) = r-s
               zout(1, nout4, j) = r+s
               r25 = r2-r5
               r34 = r3-r4
               s25 = s2+s5
               s34 = s3+s4
               zout(2, nout1, j) = s1+s25+s34
               r = cos2*s25+cos4*s34+s1
               s = sin2*r25+sin4*r34
               zout(2, nout2, j) = r+s
               zout(2, nout5, j) = r-s
               r = cos4*s25+cos2*s34+s1
               s = sin4*r25-sin2*r34
               zout(2, nout3, j) = r+s
               zout(2, nout4, j) = r-s
            END DO
         END DO
         DO ia = 2, after
            ias = ia-1
            IF (8*ias == 5*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nin5 = nin4+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     nout5 = nout4+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, j, nin3)
                        s3 = zin(1, j, nin3)
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = -(r+s)*rt2i
                        s4 = (r-s)*rt2i
                        r5 = -zin(1, j, nin5)
                        s5 = -zin(2, j, nin5)
                        r25 = r2+r5
                        r34 = r3+r4
                        s25 = s2-s5
                        s34 = s3-s4
                        zout(1, nout1, j) = r1+r25+r34
                        r = cos2*r25+cos4*r34+r1
                        s = sin2*s25+sin4*s34
                        zout(1, nout2, j) = r-s
                        zout(1, nout5, j) = r+s
                        r = cos4*r25+cos2*r34+r1
                        s = sin4*s25-sin2*s34
                        zout(1, nout3, j) = r-s
                        zout(1, nout4, j) = r+s
                        r25 = r2-r5
                        r34 = r3-r4
                        s25 = s2+s5
                        s34 = s3+s4
                        zout(2, nout1, j) = s1+s25+s34
                        r = cos2*s25+cos4*s34+s1
                        s = sin2*r25+sin4*r34
                        zout(2, nout2, j) = r+s
                        zout(2, nout5, j) = r-s
                        r = cos4*s25+cos2*s34+s1
                        s = sin4*r25-sin2*r34
                        zout(2, nout3, j) = r+s
                        zout(2, nout4, j) = r-s
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nin5 = nin4+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     nout5 = nout4+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r+s)*rt2i
                        s2 = (-r+s)*rt2i
                        r3 = zin(2, j, nin3)
                        s3 = -zin(1, j, nin3)
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = (s-r)*rt2i
                        s4 = -(r+s)*rt2i
                        r5 = -zin(1, j, nin5)
                        s5 = -zin(2, j, nin5)
                        r25 = r2+r5
                        r34 = r3+r4
                        s25 = s2-s5
                        s34 = s3-s4
                        zout(1, nout1, j) = r1+r25+r34
                        r = cos2*r25+cos4*r34+r1
                        s = sin2*s25+sin4*s34
                        zout(1, nout2, j) = r-s
                        zout(1, nout5, j) = r+s
                        r = cos4*r25+cos2*r34+r1
                        s = sin4*s25-sin2*s34
                        zout(1, nout3, j) = r-s
                        zout(1, nout4, j) = r+s
                        r25 = r2-r5
                        r34 = r3-r4
                        s25 = s2+s5
                        s34 = s3+s4
                        zout(2, nout1, j) = s1+s25+s34
                        r = cos2*s25+cos4*s34+s1
                        s = sin2*r25+sin4*r34
                        zout(2, nout2, j) = r+s
                        zout(2, nout5, j) = r-s
                        r = cos4*s25+cos2*s34+s1
                        s = sin4*r25-sin2*r34
                        zout(2, nout3, j) = r+s
                        zout(2, nout4, j) = r-s
                     END DO
                  END DO
               END IF
            ELSE
               ias = ia-1
               itt = ias*before
               itrig = itt+1
               cr2 = trig(1, itrig)
               ci2 = trig(2, itrig)
               itrig = itrig+itt
               cr3 = trig(1, itrig)
               ci3 = trig(2, itrig)
               itrig = itrig+itt
               cr4 = trig(1, itrig)
               ci4 = trig(2, itrig)
               itrig = itrig+itt
               cr5 = trig(1, itrig)
               ci5 = trig(2, itrig)
               nin1 = ia-after
               nout1 = ia-atn
               DO ib = 1, before
                  nin1 = nin1+after
                  nin2 = nin1+atb
                  nin3 = nin2+atb
                  nin4 = nin3+atb
                  nin5 = nin4+atb
                  nout1 = nout1+atn
                  nout2 = nout1+after
                  nout3 = nout2+after
                  nout4 = nout3+after
                  nout5 = nout4+after
                  DO j = 1, nfft
                     r1 = zin(1, j, nin1)
                     s1 = zin(2, j, nin1)
                     r = zin(1, j, nin2)
                     s = zin(2, j, nin2)
                     r2 = r*cr2-s*ci2
                     s2 = r*ci2+s*cr2
                     r = zin(1, j, nin3)
                     s = zin(2, j, nin3)
                     r3 = r*cr3-s*ci3
                     s3 = r*ci3+s*cr3
                     r = zin(1, j, nin4)
                     s = zin(2, j, nin4)
                     r4 = r*cr4-s*ci4
                     s4 = r*ci4+s*cr4
                     r = zin(1, j, nin5)
                     s = zin(2, j, nin5)
                     r5 = r*cr5-s*ci5
                     s5 = r*ci5+s*cr5
                     r25 = r2+r5
                     r34 = r3+r4
                     s25 = s2-s5
                     s34 = s3-s4
                     zout(1, nout1, j) = r1+r25+r34
                     r = cos2*r25+cos4*r34+r1
                     s = sin2*s25+sin4*s34
                     zout(1, nout2, j) = r-s
                     zout(1, nout5, j) = r+s
                     r = cos4*r25+cos2*r34+r1
                     s = sin4*s25-sin2*s34
                     zout(1, nout3, j) = r-s
                     zout(1, nout4, j) = r+s
                     r25 = r2-r5
                     r34 = r3-r4
                     s25 = s2+s5
                     s34 = s3+s4
                     zout(2, nout1, j) = s1+s25+s34
                     r = cos2*s25+cos4*s34+s1
                     s = sin2*r25+sin4*r34
                     zout(2, nout2, j) = r+s
                     zout(2, nout5, j) = r-s
                     r = cos4*s25+cos2*s34+s1
                     s = sin4*r25-sin2*r34
                     zout(2, nout3, j) = r+s
                     zout(2, nout4, j) = r-s
                  END DO
               END DO
            END IF
         END DO
      ELSE IF (now == 6) THEN
         bbs = isign*bb
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nin4 = nin3+atb
            nin5 = nin4+atb
            nin6 = nin5+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            nout4 = nout3+after
            nout5 = nout4+after
            nout6 = nout5+after
            DO j = 1, nfft
               r2 = zin(1, j, nin3)
               s2 = zin(2, j, nin3)
               r3 = zin(1, j, nin5)
               s3 = zin(2, j, nin5)
               r = r2+r3
               s = s2+s3
               r1 = zin(1, j, nin1)
               s1 = zin(2, j, nin1)
               ur1 = r+r1
               ui1 = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r = r2-r3
               s = s2-s3
               ur2 = r1-s*bbs
               ui2 = s1+r*bbs
               ur3 = r1+s*bbs
               ui3 = s1-r*bbs

               r2 = zin(1, j, nin6)
               s2 = zin(2, j, nin6)
               r3 = zin(1, j, nin2)
               s3 = zin(2, j, nin2)
               r = r2+r3
               s = s2+s3
               r1 = zin(1, j, nin4)
               s1 = zin(2, j, nin4)
               vr1 = r+r1
               vi1 = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r = r2-r3
               s = s2-s3
               vr2 = r1-s*bbs
               vi2 = s1+r*bbs
               vr3 = r1+s*bbs
               vi3 = s1-r*bbs

               zout(1, nout1, j) = ur1+vr1
               zout(2, nout1, j) = ui1+vi1
               zout(1, nout5, j) = ur2+vr2
               zout(2, nout5, j) = ui2+vi2
               zout(1, nout3, j) = ur3+vr3
               zout(2, nout3, j) = ui3+vi3
               zout(1, nout4, j) = ur1-vr1
               zout(2, nout4, j) = ui1-vi1
               zout(1, nout2, j) = ur2-vr2
               zout(2, nout2, j) = ui2-vi2
               zout(1, nout6, j) = ur3-vr3
               zout(2, nout6, j) = ui3-vi3
            END DO
         END DO
      ELSE
         CALL cp__b("pw/fft/mltfftsg_tools.F",1308,'Error fftrot')
      END IF

!-----------------------------------------------------------------------------!

   END SUBROUTINE fftrot

!-----------------------------------------------------------------------------!
!-----------------------------------------------------------------------------!
!  Copyright (C) Stefan Goedecker, Lausanne, Switzerland, August 1, 1991
!  Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!  Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!  This file is distributed under the terms of the
!  GNU General Public License version 2 (or later),
!  see http://www.gnu.org/copyleft/gpl.txt .
!-----------------------------------------------------------------------------!
!  S. Goedecker: Rotating a three-dimensional array in optimal
!  positions for vector processing: Case study for a three-dimensional Fast
!  Fourier Transform, Comp. Phys. Commun. 76, 294 (1993)
! **************************************************************************************************
!> \brief ...
!> \param mm ...
!> \param nfft ...
!> \param m ...
!> \param nn ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param trig ...
!> \param now ...
!> \param after ...
!> \param before ...
!> \param isign ...
! **************************************************************************************************
   SUBROUTINE fftpre(mm, nfft, m, nn, n, zin, zout, trig, now, after, before, isign)

      INTEGER, INTENT(IN)                                :: mm, nfft, m, nn, n
      REAL(dp), DIMENSION(2, m, mm), INTENT(IN)          :: zin
      REAL(dp), DIMENSION(2, nn, n), INTENT(INOUT)       :: zout
      REAL(dp), DIMENSION(2, ctrig_length), INTENT(IN)   :: trig
      INTEGER, INTENT(IN)                                :: now, after, before, isign

      REAL(dp), PARAMETER :: bb = 0.8660254037844387_dp, cos2 = 0.3090169943749474_dp, &
         cos4 = -0.8090169943749474_dp, rt2i = 0.7071067811865475_dp, &
         sin2p = 0.9510565162951536_dp, sin4p = 0.5877852522924731_dp

      INTEGER                                            :: atb, atn, ia, ias, ib, itrig, itt, j, &
                                                            nin1, nin2, nin3, nin4, nin5, nin6, &
                                                            nin7, nin8, nout1, nout2, nout3, &
                                                            nout4, nout5, nout6, nout7, nout8
      REAL(dp) :: am, ap, bbs, bm, bp, ci2, ci3, ci4, ci5, cm, cp, cr2, cr3, cr4, cr5, dbl, dm, r, &
         r1, r2, r25, r3, r34, r4, r5, r6, r7, r8, s, s1, s2, s25, s3, s34, s4, s5, s6, s7, s8, &
         sin2, sin4, ui1, ui2, ui3, ur1, ur2, ur3, vi1, vi2, vi3, vr1, vr2, vr3

! sqrt(0.5)
! sqrt(3)/2
! cos(2*pi/5)
!  cos(4*pi/5)
! sin(2*pi/5)
! sin(4*pi/5)
!-----------------------------------------------------------------------------!

      atn = after*now
      atb = after*before

      IF (now == 4) THEN
         IF (isign == 1) THEN
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               DO j = 1, nfft
                  r1 = zin(1, nin1, j)
                  s1 = zin(2, nin1, j)
                  r2 = zin(1, nin2, j)
                  s2 = zin(2, nin2, j)
                  r3 = zin(1, nin3, j)
                  s3 = zin(2, nin3, j)
                  r4 = zin(1, nin4, j)
                  s4 = zin(2, nin4, j)
                  r = r1+r3
                  s = r2+r4
                  zout(1, j, nout1) = r+s
                  zout(1, j, nout3) = r-s
                  r = r1-r3
                  s = s2-s4
                  zout(1, j, nout2) = r-s
                  zout(1, j, nout4) = r+s
                  r = s1+s3
                  s = s2+s4
                  zout(2, j, nout1) = r+s
                  zout(2, j, nout3) = r-s
                  r = s1-s3
                  s = r2-r4
                  zout(2, j, nout2) = r+s
                  zout(2, j, nout4) = r-s
               END DO
            END DO
            DO ia = 2, after
               ias = ia-1
               IF (2*ias == after) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r = zin(1, nin2, j)
                        s = zin(2, nin2, j)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, nin3, j)
                        s3 = zin(1, nin3, j)
                        r = zin(1, nin4, j)
                        s = zin(2, nin4, j)
                        r4 = -(r+s)*rt2i
                        s4 = (r-s)*rt2i
                        r = r1+r3
                        s = r2+r4
                        zout(1, j, nout1) = r+s
                        zout(1, j, nout3) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, j, nout2) = r-s
                        zout(1, j, nout4) = r+s
                        r = s1+s3
                        s = s2+s4
                        zout(2, j, nout1) = r+s
                        zout(2, j, nout3) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, j, nout2) = r+s
                        zout(2, j, nout4) = r-s
                     END DO
                  END DO
               ELSE
                  itt = ias*before
                  itrig = itt+1
                  cr2 = trig(1, itrig)
                  ci2 = trig(2, itrig)
                  itrig = itrig+itt
                  cr3 = trig(1, itrig)
                  ci3 = trig(2, itrig)
                  itrig = itrig+itt
                  cr4 = trig(1, itrig)
                  ci4 = trig(2, itrig)
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r = zin(1, nin2, j)
                        s = zin(2, nin2, j)
                        r2 = r*cr2-s*ci2
                        s2 = r*ci2+s*cr2
                        r = zin(1, nin3, j)
                        s = zin(2, nin3, j)
                        r3 = r*cr3-s*ci3
                        s3 = r*ci3+s*cr3
                        r = zin(1, nin4, j)
                        s = zin(2, nin4, j)
                        r4 = r*cr4-s*ci4
                        s4 = r*ci4+s*cr4
                        r = r1+r3
                        s = r2+r4
                        zout(1, j, nout1) = r+s
                        zout(1, j, nout3) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, j, nout2) = r-s
                        zout(1, j, nout4) = r+s
                        r = s1+s3
                        s = s2+s4
                        zout(2, j, nout1) = r+s
                        zout(2, j, nout3) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, j, nout2) = r+s
                        zout(2, j, nout4) = r-s
                     END DO
                  END DO
               END IF
            END DO
         ELSE
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               DO j = 1, nfft
                  r1 = zin(1, nin1, j)
                  s1 = zin(2, nin1, j)
                  r2 = zin(1, nin2, j)
                  s2 = zin(2, nin2, j)
                  r3 = zin(1, nin3, j)
                  s3 = zin(2, nin3, j)
                  r4 = zin(1, nin4, j)
                  s4 = zin(2, nin4, j)
                  r = r1+r3
                  s = r2+r4
                  zout(1, j, nout1) = r+s
                  zout(1, j, nout3) = r-s
                  r = r1-r3
                  s = s2-s4
                  zout(1, j, nout2) = r+s
                  zout(1, j, nout4) = r-s
                  r = s1+s3
                  s = s2+s4
                  zout(2, j, nout1) = r+s
                  zout(2, j, nout3) = r-s
                  r = s1-s3
                  s = r2-r4
                  zout(2, j, nout2) = r-s
                  zout(2, j, nout4) = r+s
               END DO
            END DO
            DO ia = 2, after
               ias = ia-1
               IF (2*ias == after) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r = zin(1, nin2, j)
                        s = zin(2, nin2, j)
                        r2 = (r+s)*rt2i
                        s2 = (s-r)*rt2i
                        r3 = zin(2, nin3, j)
                        s3 = -zin(1, nin3, j)
                        r = zin(1, nin4, j)
                        s = zin(2, nin4, j)
                        r4 = (s-r)*rt2i
                        s4 = -(r+s)*rt2i
                        r = r1+r3
                        s = r2+r4
                        zout(1, j, nout1) = r+s
                        zout(1, j, nout3) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, j, nout2) = r+s
                        zout(1, j, nout4) = r-s
                        r = s1+s3
                        s = s2+s4
                        zout(2, j, nout1) = r+s
                        zout(2, j, nout3) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, j, nout2) = r-s
                        zout(2, j, nout4) = r+s
                     END DO
                  END DO
               ELSE
                  itt = ias*before
                  itrig = itt+1
                  cr2 = trig(1, itrig)
                  ci2 = trig(2, itrig)
                  itrig = itrig+itt
                  cr3 = trig(1, itrig)
                  ci3 = trig(2, itrig)
                  itrig = itrig+itt
                  cr4 = trig(1, itrig)
                  ci4 = trig(2, itrig)
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r = zin(1, nin2, j)
                        s = zin(2, nin2, j)
                        r2 = r*cr2-s*ci2
                        s2 = r*ci2+s*cr2
                        r = zin(1, nin3, j)
                        s = zin(2, nin3, j)
                        r3 = r*cr3-s*ci3
                        s3 = r*ci3+s*cr3
                        r = zin(1, nin4, j)
                        s = zin(2, nin4, j)
                        r4 = r*cr4-s*ci4
                        s4 = r*ci4+s*cr4
                        r = r1+r3
                        s = r2+r4
                        zout(1, j, nout1) = r+s
                        zout(1, j, nout3) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, j, nout2) = r+s
                        zout(1, j, nout4) = r-s
                        r = s1+s3
                        s = s2+s4
                        zout(2, j, nout1) = r+s
                        zout(2, j, nout3) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, j, nout2) = r-s
                        zout(2, j, nout4) = r+s
                     END DO
                  END DO
               END IF
            END DO
         END IF
      ELSE IF (now == 8) THEN
         IF (isign == -1) THEN
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nin5 = nin4+atb
               nin6 = nin5+atb
               nin7 = nin6+atb
               nin8 = nin7+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               nout5 = nout4+after
               nout6 = nout5+after
               nout7 = nout6+after
               nout8 = nout7+after
               DO j = 1, nfft
                  r1 = zin(1, nin1, j)
                  s1 = zin(2, nin1, j)
                  r2 = zin(1, nin2, j)
                  s2 = zin(2, nin2, j)
                  r3 = zin(1, nin3, j)
                  s3 = zin(2, nin3, j)
                  r4 = zin(1, nin4, j)
                  s4 = zin(2, nin4, j)
                  r5 = zin(1, nin5, j)
                  s5 = zin(2, nin5, j)
                  r6 = zin(1, nin6, j)
                  s6 = zin(2, nin6, j)
                  r7 = zin(1, nin7, j)
                  s7 = zin(2, nin7, j)
                  r8 = zin(1, nin8, j)
                  s8 = zin(2, nin8, j)
                  r = r1+r5
                  s = r3+r7
                  ap = r+s
                  am = r-s
                  r = r2+r6
                  s = r4+r8
                  bp = r+s
                  bm = r-s
                  r = s1+s5
                  s = s3+s7
                  cp = r+s
                  cm = r-s
                  r = s2+s6
                  s = s4+s8
                  dbl = r+s
                  dm = r-s
                  zout(1, j, nout1) = ap+bp
                  zout(2, j, nout1) = cp+dbl
                  zout(1, j, nout5) = ap-bp
                  zout(2, j, nout5) = cp-dbl
                  zout(1, j, nout3) = am+dm
                  zout(2, j, nout3) = cm-bm
                  zout(1, j, nout7) = am-dm
                  zout(2, j, nout7) = cm+bm
                  r = r1-r5
                  s = s3-s7
                  ap = r+s
                  am = r-s
                  r = s1-s5
                  s = r3-r7
                  bp = r+s
                  bm = r-s
                  r = s4-s8
                  s = r2-r6
                  cp = r+s
                  cm = r-s
                  r = s2-s6
                  s = r4-r8
                  dbl = r+s
                  dm = r-s
                  r = (cp+dm)*rt2i
                  s = (-cp+dm)*rt2i
                  cp = (cm+dbl)*rt2i
                  dbl = (cm-dbl)*rt2i
                  zout(1, j, nout2) = ap+r
                  zout(2, j, nout2) = bm+s
                  zout(1, j, nout6) = ap-r
                  zout(2, j, nout6) = bm-s
                  zout(1, j, nout4) = am+cp
                  zout(2, j, nout4) = bp+dbl
                  zout(1, j, nout8) = am-cp
                  zout(2, j, nout8) = bp-dbl
               END DO
            END DO
         ELSE
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nin5 = nin4+atb
               nin6 = nin5+atb
               nin7 = nin6+atb
               nin8 = nin7+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               nout5 = nout4+after
               nout6 = nout5+after
               nout7 = nout6+after
               nout8 = nout7+after
               DO j = 1, nfft
                  r1 = zin(1, nin1, j)
                  s1 = zin(2, nin1, j)
                  r2 = zin(1, nin2, j)
                  s2 = zin(2, nin2, j)
                  r3 = zin(1, nin3, j)
                  s3 = zin(2, nin3, j)
                  r4 = zin(1, nin4, j)
                  s4 = zin(2, nin4, j)
                  r5 = zin(1, nin5, j)
                  s5 = zin(2, nin5, j)
                  r6 = zin(1, nin6, j)
                  s6 = zin(2, nin6, j)
                  r7 = zin(1, nin7, j)
                  s7 = zin(2, nin7, j)
                  r8 = zin(1, nin8, j)
                  s8 = zin(2, nin8, j)
                  r = r1+r5
                  s = r3+r7
                  ap = r+s
                  am = r-s
                  r = r2+r6
                  s = r4+r8
                  bp = r+s
                  bm = r-s
                  r = s1+s5
                  s = s3+s7
                  cp = r+s
                  cm = r-s
                  r = s2+s6
                  s = s4+s8
                  dbl = r+s
                  dm = r-s
                  zout(1, j, nout1) = ap+bp
                  zout(2, j, nout1) = cp+dbl
                  zout(1, j, nout5) = ap-bp
                  zout(2, j, nout5) = cp-dbl
                  zout(1, j, nout3) = am-dm
                  zout(2, j, nout3) = cm+bm
                  zout(1, j, nout7) = am+dm
                  zout(2, j, nout7) = cm-bm
                  r = r1-r5
                  s = -s3+s7
                  ap = r+s
                  am = r-s
                  r = s1-s5
                  s = r7-r3
                  bp = r+s
                  bm = r-s
                  r = -s4+s8
                  s = r2-r6
                  cp = r+s
                  cm = r-s
                  r = -s2+s6
                  s = r4-r8
                  dbl = r+s
                  dm = r-s
                  r = (cp+dm)*rt2i
                  s = (cp-dm)*rt2i
                  cp = (cm+dbl)*rt2i
                  dbl = (-cm+dbl)*rt2i
                  zout(1, j, nout2) = ap+r
                  zout(2, j, nout2) = bm+s
                  zout(1, j, nout6) = ap-r
                  zout(2, j, nout6) = bm-s
                  zout(1, j, nout4) = am+cp
                  zout(2, j, nout4) = bp+dbl
                  zout(1, j, nout8) = am-cp
                  zout(2, j, nout8) = bp-dbl
               END DO
            END DO
         END IF
      ELSE IF (now == 3) THEN
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         bbs = isign*bb
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            DO j = 1, nfft
               r1 = zin(1, nin1, j)
               s1 = zin(2, nin1, j)
               r2 = zin(1, nin2, j)
               s2 = zin(2, nin2, j)
               r3 = zin(1, nin3, j)
               s3 = zin(2, nin3, j)
               r = r2+r3
               s = s2+s3
               zout(1, j, nout1) = r+r1
               zout(2, j, nout1) = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r2 = bbs*(r2-r3)
               s2 = bbs*(s2-s3)
               zout(1, j, nout2) = r1-s2
               zout(2, j, nout2) = s1+r2
               zout(1, j, nout3) = r1+s2
               zout(2, j, nout3) = s1-r2
            END DO
         END DO
         DO ia = 2, after
            ias = ia-1
            IF (4*ias == 3*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r2 = -zin(2, nin2, j)
                        s2 = zin(1, nin2, j)
                        r3 = -zin(1, nin3, j)
                        s3 = -zin(2, nin3, j)
                        r = r2+r3
                        s = s2+s3
                        zout(1, j, nout1) = r+r1
                        zout(2, j, nout1) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, j, nout2) = r1-s2
                        zout(2, j, nout2) = s1+r2
                        zout(1, j, nout3) = r1+s2
                        zout(2, j, nout3) = s1-r2
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r2 = zin(2, nin2, j)
                        s2 = -zin(1, nin2, j)
                        r3 = -zin(1, nin3, j)
                        s3 = -zin(2, nin3, j)
                        r = r2+r3
                        s = s2+s3
                        zout(1, j, nout1) = r+r1
                        zout(2, j, nout1) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, j, nout2) = r1-s2
                        zout(2, j, nout2) = s1+r2
                        zout(1, j, nout3) = r1+s2
                        zout(2, j, nout3) = s1-r2
                     END DO
                  END DO
               END IF
            ELSE IF (8*ias == 3*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r = zin(1, nin2, j)
                        s = zin(2, nin2, j)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, nin3, j)
                        s3 = zin(1, nin3, j)
                        r = r2+r3
                        s = s2+s3
                        zout(1, j, nout1) = r+r1
                        zout(2, j, nout1) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, j, nout2) = r1-s2
                        zout(2, j, nout2) = s1+r2
                        zout(1, j, nout3) = r1+s2
                        zout(2, j, nout3) = s1-r2
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r = zin(1, nin2, j)
                        s = zin(2, nin2, j)
                        r2 = (r+s)*rt2i
                        s2 = (-r+s)*rt2i
                        r3 = zin(2, nin3, j)
                        s3 = -zin(1, nin3, j)
                        r = r2+r3
                        s = s2+s3
                        zout(1, j, nout1) = r+r1
                        zout(2, j, nout1) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, j, nout2) = r1-s2
                        zout(2, j, nout2) = s1+r2
                        zout(1, j, nout3) = r1+s2
                        zout(2, j, nout3) = s1-r2
                     END DO
                  END DO
               END IF
            ELSE
               itt = ias*before
               itrig = itt+1
               cr2 = trig(1, itrig)
               ci2 = trig(2, itrig)
               itrig = itrig+itt
               cr3 = trig(1, itrig)
               ci3 = trig(2, itrig)
               nin1 = ia-after
               nout1 = ia-atn
               DO ib = 1, before
                  nin1 = nin1+after
                  nin2 = nin1+atb
                  nin3 = nin2+atb
                  nout1 = nout1+atn
                  nout2 = nout1+after
                  nout3 = nout2+after
                  DO j = 1, nfft
                     r1 = zin(1, nin1, j)
                     s1 = zin(2, nin1, j)
                     r = zin(1, nin2, j)
                     s = zin(2, nin2, j)
                     r2 = r*cr2-s*ci2
                     s2 = r*ci2+s*cr2
                     r = zin(1, nin3, j)
                     s = zin(2, nin3, j)
                     r3 = r*cr3-s*ci3
                     s3 = r*ci3+s*cr3
                     r = r2+r3
                     s = s2+s3
                     zout(1, j, nout1) = r+r1
                     zout(2, j, nout1) = s+s1
                     r1 = r1-0.5_dp*r
                     s1 = s1-0.5_dp*s
                     r2 = bbs*(r2-r3)
                     s2 = bbs*(s2-s3)
                     zout(1, j, nout2) = r1-s2
                     zout(2, j, nout2) = s1+r2
                     zout(1, j, nout3) = r1+s2
                     zout(2, j, nout3) = s1-r2
                  END DO
               END DO
            END IF
         END DO
      ELSE IF (now == 5) THEN
         sin2 = isign*sin2p
         sin4 = isign*sin4p
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nin4 = nin3+atb
            nin5 = nin4+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            nout4 = nout3+after
            nout5 = nout4+after
            DO j = 1, nfft
               r1 = zin(1, nin1, j)
               s1 = zin(2, nin1, j)
               r2 = zin(1, nin2, j)
               s2 = zin(2, nin2, j)
               r3 = zin(1, nin3, j)
               s3 = zin(2, nin3, j)
               r4 = zin(1, nin4, j)
               s4 = zin(2, nin4, j)
               r5 = zin(1, nin5, j)
               s5 = zin(2, nin5, j)
               r25 = r2+r5
               r34 = r3+r4
               s25 = s2-s5
               s34 = s3-s4
               zout(1, j, nout1) = r1+r25+r34
               r = cos2*r25+cos4*r34+r1
               s = sin2*s25+sin4*s34
               zout(1, j, nout2) = r-s
               zout(1, j, nout5) = r+s
               r = cos4*r25+cos2*r34+r1
               s = sin4*s25-sin2*s34
               zout(1, j, nout3) = r-s
               zout(1, j, nout4) = r+s
               r25 = r2-r5
               r34 = r3-r4
               s25 = s2+s5
               s34 = s3+s4
               zout(2, j, nout1) = s1+s25+s34
               r = cos2*s25+cos4*s34+s1
               s = sin2*r25+sin4*r34
               zout(2, j, nout2) = r+s
               zout(2, j, nout5) = r-s
               r = cos4*s25+cos2*s34+s1
               s = sin4*r25-sin2*r34
               zout(2, j, nout3) = r+s
               zout(2, j, nout4) = r-s
            END DO
         END DO
         DO ia = 2, after
            ias = ia-1
            IF (8*ias == 5*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nin5 = nin4+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     nout5 = nout4+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r = zin(1, nin2, j)
                        s = zin(2, nin2, j)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, nin3, j)
                        s3 = zin(1, nin3, j)
                        r = zin(1, nin4, j)
                        s = zin(2, nin4, j)
                        r4 = -(r+s)*rt2i
                        s4 = (r-s)*rt2i
                        r5 = -zin(1, nin5, j)
                        s5 = -zin(2, nin5, j)
                        r25 = r2+r5
                        r34 = r3+r4
                        s25 = s2-s5
                        s34 = s3-s4
                        zout(1, j, nout1) = r1+r25+r34
                        r = cos2*r25+cos4*r34+r1
                        s = sin2*s25+sin4*s34
                        zout(1, j, nout2) = r-s
                        zout(1, j, nout5) = r+s
                        r = cos4*r25+cos2*r34+r1
                        s = sin4*s25-sin2*s34
                        zout(1, j, nout3) = r-s
                        zout(1, j, nout4) = r+s
                        r25 = r2-r5
                        r34 = r3-r4
                        s25 = s2+s5
                        s34 = s3+s4
                        zout(2, j, nout1) = s1+s25+s34
                        r = cos2*s25+cos4*s34+s1
                        s = sin2*r25+sin4*r34
                        zout(2, j, nout2) = r+s
                        zout(2, j, nout5) = r-s
                        r = cos4*s25+cos2*s34+s1
                        s = sin4*r25-sin2*r34
                        zout(2, j, nout3) = r+s
                        zout(2, j, nout4) = r-s
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nin5 = nin4+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     nout5 = nout4+after
                     DO j = 1, nfft
                        r1 = zin(1, nin1, j)
                        s1 = zin(2, nin1, j)
                        r = zin(1, nin2, j)
                        s = zin(2, nin2, j)
                        r2 = (r+s)*rt2i
                        s2 = (-r+s)*rt2i
                        r3 = zin(2, nin3, j)
                        s3 = -zin(1, nin3, j)
                        r = zin(1, nin4, j)
                        s = zin(2, nin4, j)
                        r4 = (s-r)*rt2i
                        s4 = -(r+s)*rt2i
                        r5 = -zin(1, nin5, j)
                        s5 = -zin(2, nin5, j)
                        r25 = r2+r5
                        r34 = r3+r4
                        s25 = s2-s5
                        s34 = s3-s4
                        zout(1, j, nout1) = r1+r25+r34
                        r = cos2*r25+cos4*r34+r1
                        s = sin2*s25+sin4*s34
                        zout(1, j, nout2) = r-s
                        zout(1, j, nout5) = r+s
                        r = cos4*r25+cos2*r34+r1
                        s = sin4*s25-sin2*s34
                        zout(1, j, nout3) = r-s
                        zout(1, j, nout4) = r+s
                        r25 = r2-r5
                        r34 = r3-r4
                        s25 = s2+s5
                        s34 = s3+s4
                        zout(2, j, nout1) = s1+s25+s34
                        r = cos2*s25+cos4*s34+s1
                        s = sin2*r25+sin4*r34
                        zout(2, j, nout2) = r+s
                        zout(2, j, nout5) = r-s
                        r = cos4*s25+cos2*s34+s1
                        s = sin4*r25-sin2*r34
                        zout(2, j, nout3) = r+s
                        zout(2, j, nout4) = r-s
                     END DO
                  END DO
               END IF
            ELSE
               ias = ia-1
               itt = ias*before
               itrig = itt+1
               cr2 = trig(1, itrig)
               ci2 = trig(2, itrig)
               itrig = itrig+itt
               cr3 = trig(1, itrig)
               ci3 = trig(2, itrig)
               itrig = itrig+itt
               cr4 = trig(1, itrig)
               ci4 = trig(2, itrig)
               itrig = itrig+itt
               cr5 = trig(1, itrig)
               ci5 = trig(2, itrig)
               nin1 = ia-after
               nout1 = ia-atn
               DO ib = 1, before
                  nin1 = nin1+after
                  nin2 = nin1+atb
                  nin3 = nin2+atb
                  nin4 = nin3+atb
                  nin5 = nin4+atb
                  nout1 = nout1+atn
                  nout2 = nout1+after
                  nout3 = nout2+after
                  nout4 = nout3+after
                  nout5 = nout4+after
                  DO j = 1, nfft
                     r1 = zin(1, nin1, j)
                     s1 = zin(2, nin1, j)
                     r = zin(1, nin2, j)
                     s = zin(2, nin2, j)
                     r2 = r*cr2-s*ci2
                     s2 = r*ci2+s*cr2
                     r = zin(1, nin3, j)
                     s = zin(2, nin3, j)
                     r3 = r*cr3-s*ci3
                     s3 = r*ci3+s*cr3
                     r = zin(1, nin4, j)
                     s = zin(2, nin4, j)
                     r4 = r*cr4-s*ci4
                     s4 = r*ci4+s*cr4
                     r = zin(1, nin5, j)
                     s = zin(2, nin5, j)
                     r5 = r*cr5-s*ci5
                     s5 = r*ci5+s*cr5
                     r25 = r2+r5
                     r34 = r3+r4
                     s25 = s2-s5
                     s34 = s3-s4
                     zout(1, j, nout1) = r1+r25+r34
                     r = cos2*r25+cos4*r34+r1
                     s = sin2*s25+sin4*s34
                     zout(1, j, nout2) = r-s
                     zout(1, j, nout5) = r+s
                     r = cos4*r25+cos2*r34+r1
                     s = sin4*s25-sin2*s34
                     zout(1, j, nout3) = r-s
                     zout(1, j, nout4) = r+s
                     r25 = r2-r5
                     r34 = r3-r4
                     s25 = s2+s5
                     s34 = s3+s4
                     zout(2, j, nout1) = s1+s25+s34
                     r = cos2*s25+cos4*s34+s1
                     s = sin2*r25+sin4*r34
                     zout(2, j, nout2) = r+s
                     zout(2, j, nout5) = r-s
                     r = cos4*s25+cos2*s34+s1
                     s = sin4*r25-sin2*r34
                     zout(2, j, nout3) = r+s
                     zout(2, j, nout4) = r-s
                  END DO
               END DO
            END IF
         END DO
      ELSE IF (now == 6) THEN
         bbs = isign*bb
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nin4 = nin3+atb
            nin5 = nin4+atb
            nin6 = nin5+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            nout4 = nout3+after
            nout5 = nout4+after
            nout6 = nout5+after
            DO j = 1, nfft
               r2 = zin(1, nin3, j)
               s2 = zin(2, nin3, j)
               r3 = zin(1, nin5, j)
               s3 = zin(2, nin5, j)
               r = r2+r3
               s = s2+s3
               r1 = zin(1, nin1, j)
               s1 = zin(2, nin1, j)
               ur1 = r+r1
               ui1 = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r = r2-r3
               s = s2-s3
               ur2 = r1-s*bbs
               ui2 = s1+r*bbs
               ur3 = r1+s*bbs
               ui3 = s1-r*bbs

               r2 = zin(1, nin6, j)
               s2 = zin(2, nin6, j)
               r3 = zin(1, nin2, j)
               s3 = zin(2, nin2, j)
               r = r2+r3
               s = s2+s3
               r1 = zin(1, nin4, j)
               s1 = zin(2, nin4, j)
               vr1 = r+r1
               vi1 = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r = r2-r3
               s = s2-s3
               vr2 = r1-s*bbs
               vi2 = s1+r*bbs
               vr3 = r1+s*bbs
               vi3 = s1-r*bbs

               zout(1, j, nout1) = ur1+vr1
               zout(2, j, nout1) = ui1+vi1
               zout(1, j, nout5) = ur2+vr2
               zout(2, j, nout5) = ui2+vi2
               zout(1, j, nout3) = ur3+vr3
               zout(2, j, nout3) = ui3+vi3
               zout(1, j, nout4) = ur1-vr1
               zout(2, j, nout4) = ui1-vi1
               zout(1, j, nout2) = ur2-vr2
               zout(2, j, nout2) = ui2-vi2
               zout(1, j, nout6) = ur3-vr3
               zout(2, j, nout6) = ui3-vi3
            END DO
         END DO
      ELSE
         CALL cp__b("pw/fft/mltfftsg_tools.F",2379,'Error fftpre')
      END IF

!-----------------------------------------------------------------------------!

   END SUBROUTINE fftpre

!-----------------------------------------------------------------------------!

!-----------------------------------------------------------------------------!
!  Copyright (C) Stefan Goedecker, Lausanne, Switzerland, August 1, 1991
!  Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!  Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!  This file is distributed under the terms of the
!  GNU General Public License version 2 (or later),
!  see http://www.gnu.org/copyleft/gpl.txt .
!-----------------------------------------------------------------------------!
!  S. Goedecker: Rotating a three-dimensional array in optimal
!  positions for vector processing: Case study for a three-dimensional Fast
!  Fourier Transform, Comp. Phys. Commun. 76, 294 (1993)
! **************************************************************************************************
!> \brief ...
!> \param mm ...
!> \param nfft ...
!> \param m ...
!> \param nn ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param trig ...
!> \param now ...
!> \param after ...
!> \param before ...
!> \param isign ...
! **************************************************************************************************
   SUBROUTINE fftstp(mm, nfft, m, nn, n, zin, zout, trig, now, after, before, isign)

      INTEGER, INTENT(IN)                                :: mm, nfft, m, nn, n
      REAL(dp), DIMENSION(2, mm, m), INTENT(IN)          :: zin
      REAL(dp), DIMENSION(2, nn, n), INTENT(INOUT)       :: zout
      REAL(dp), DIMENSION(2, ctrig_length), INTENT(IN)   :: trig
      INTEGER, INTENT(IN)                                :: now, after, before, isign

      REAL(dp), PARAMETER :: bb = 0.8660254037844387_dp, cos2 = 0.3090169943749474_dp, &
         cos4 = -0.8090169943749474_dp, rt2i = 0.7071067811865475_dp, &
         sin2p = 0.9510565162951536_dp, sin4p = 0.5877852522924731_dp

      INTEGER                                            :: atb, atn, ia, ias, ib, itrig, itt, j, &
                                                            nin1, nin2, nin3, nin4, nin5, nin6, &
                                                            nin7, nin8, nout1, nout2, nout3, &
                                                            nout4, nout5, nout6, nout7, nout8
      REAL(dp) :: am, ap, bbs, bm, bp, ci2, ci3, ci4, ci5, cm, cp, cr2, cr3, cr4, cr5, dbl, dm, r, &
         r1, r2, r25, r3, r34, r4, r5, r6, r7, r8, s, s1, s2, s25, s3, s34, s4, s5, s6, s7, s8, &
         sin2, sin4, ui1, ui2, ui3, ur1, ur2, ur3, vi1, vi2, vi3, vr1, vr2, vr3

! sqrt(0.5)
! sqrt(3)/2
! cos(2*pi/5)
!  cos(4*pi/5)
! sin(2*pi/5)
! sin(4*pi/5)
!-----------------------------------------------------------------------------!

      atn = after*now
      atb = after*before

      IF (now == 4) THEN
         IF (isign == 1) THEN
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               DO j = 1, nfft
                  r1 = zin(1, j, nin1)
                  s1 = zin(2, j, nin1)
                  r2 = zin(1, j, nin2)
                  s2 = zin(2, j, nin2)
                  r3 = zin(1, j, nin3)
                  s3 = zin(2, j, nin3)
                  r4 = zin(1, j, nin4)
                  s4 = zin(2, j, nin4)
                  r = r1+r3
                  s = r2+r4
                  zout(1, j, nout1) = r+s
                  zout(1, j, nout3) = r-s
                  r = r1-r3
                  s = s2-s4
                  zout(1, j, nout2) = r-s
                  zout(1, j, nout4) = r+s
                  r = s1+s3
                  s = s2+s4
                  zout(2, j, nout1) = r+s
                  zout(2, j, nout3) = r-s
                  r = s1-s3
                  s = r2-r4
                  zout(2, j, nout2) = r+s
                  zout(2, j, nout4) = r-s
               END DO
            END DO
            DO ia = 2, after
               ias = ia-1
               IF (2*ias == after) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, j, nin3)
                        s3 = zin(1, j, nin3)
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = -(r+s)*rt2i
                        s4 = (r-s)*rt2i
                        r = r1+r3
                        s = r2+r4
                        zout(1, j, nout1) = r+s
                        zout(1, j, nout3) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, j, nout2) = r-s
                        zout(1, j, nout4) = r+s
                        r = s1+s3
                        s = s2+s4
                        zout(2, j, nout1) = r+s
                        zout(2, j, nout3) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, j, nout2) = r+s
                        zout(2, j, nout4) = r-s
                     END DO
                  END DO
               ELSE
                  itt = ias*before
                  itrig = itt+1
                  cr2 = trig(1, itrig)
                  ci2 = trig(2, itrig)
                  itrig = itrig+itt
                  cr3 = trig(1, itrig)
                  ci3 = trig(2, itrig)
                  itrig = itrig+itt
                  cr4 = trig(1, itrig)
                  ci4 = trig(2, itrig)
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = r*cr2-s*ci2
                        s2 = r*ci2+s*cr2
                        r = zin(1, j, nin3)
                        s = zin(2, j, nin3)
                        r3 = r*cr3-s*ci3
                        s3 = r*ci3+s*cr3
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = r*cr4-s*ci4
                        s4 = r*ci4+s*cr4
                        r = r1+r3
                        s = r2+r4
                        zout(1, j, nout1) = r+s
                        zout(1, j, nout3) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, j, nout2) = r-s
                        zout(1, j, nout4) = r+s
                        r = s1+s3
                        s = s2+s4
                        zout(2, j, nout1) = r+s
                        zout(2, j, nout3) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, j, nout2) = r+s
                        zout(2, j, nout4) = r-s
                     END DO
                  END DO
               END IF
            END DO
         ELSE
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               DO j = 1, nfft
                  r1 = zin(1, j, nin1)
                  s1 = zin(2, j, nin1)
                  r2 = zin(1, j, nin2)
                  s2 = zin(2, j, nin2)
                  r3 = zin(1, j, nin3)
                  s3 = zin(2, j, nin3)
                  r4 = zin(1, j, nin4)
                  s4 = zin(2, j, nin4)
                  r = r1+r3
                  s = r2+r4
                  zout(1, j, nout1) = r+s
                  zout(1, j, nout3) = r-s
                  r = r1-r3
                  s = s2-s4
                  zout(1, j, nout2) = r+s
                  zout(1, j, nout4) = r-s
                  r = s1+s3
                  s = s2+s4
                  zout(2, j, nout1) = r+s
                  zout(2, j, nout3) = r-s
                  r = s1-s3
                  s = r2-r4
                  zout(2, j, nout2) = r-s
                  zout(2, j, nout4) = r+s
               END DO
            END DO
            DO ia = 2, after
               ias = ia-1
               IF (2*ias == after) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r+s)*rt2i
                        s2 = (s-r)*rt2i
                        r3 = zin(2, j, nin3)
                        s3 = -zin(1, j, nin3)
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = (s-r)*rt2i
                        s4 = -(r+s)*rt2i
                        r = r1+r3
                        s = r2+r4
                        zout(1, j, nout1) = r+s
                        zout(1, j, nout3) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, j, nout2) = r+s
                        zout(1, j, nout4) = r-s
                        r = s1+s3
                        s = s2+s4
                        zout(2, j, nout1) = r+s
                        zout(2, j, nout3) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, j, nout2) = r-s
                        zout(2, j, nout4) = r+s
                     END DO
                  END DO
               ELSE
                  itt = ias*before
                  itrig = itt+1
                  cr2 = trig(1, itrig)
                  ci2 = trig(2, itrig)
                  itrig = itrig+itt
                  cr3 = trig(1, itrig)
                  ci3 = trig(2, itrig)
                  itrig = itrig+itt
                  cr4 = trig(1, itrig)
                  ci4 = trig(2, itrig)
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = r*cr2-s*ci2
                        s2 = r*ci2+s*cr2
                        r = zin(1, j, nin3)
                        s = zin(2, j, nin3)
                        r3 = r*cr3-s*ci3
                        s3 = r*ci3+s*cr3
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = r*cr4-s*ci4
                        s4 = r*ci4+s*cr4
                        r = r1+r3
                        s = r2+r4
                        zout(1, j, nout1) = r+s
                        zout(1, j, nout3) = r-s
                        r = r1-r3
                        s = s2-s4
                        zout(1, j, nout2) = r+s
                        zout(1, j, nout4) = r-s
                        r = s1+s3
                        s = s2+s4
                        zout(2, j, nout1) = r+s
                        zout(2, j, nout3) = r-s
                        r = s1-s3
                        s = r2-r4
                        zout(2, j, nout2) = r-s
                        zout(2, j, nout4) = r+s
                     END DO
                  END DO
               END IF
            END DO
         END IF
      ELSE IF (now == 8) THEN
         IF (isign == -1) THEN
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nin5 = nin4+atb
               nin6 = nin5+atb
               nin7 = nin6+atb
               nin8 = nin7+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               nout5 = nout4+after
               nout6 = nout5+after
               nout7 = nout6+after
               nout8 = nout7+after
               DO j = 1, nfft
                  r1 = zin(1, j, nin1)
                  s1 = zin(2, j, nin1)
                  r2 = zin(1, j, nin2)
                  s2 = zin(2, j, nin2)
                  r3 = zin(1, j, nin3)
                  s3 = zin(2, j, nin3)
                  r4 = zin(1, j, nin4)
                  s4 = zin(2, j, nin4)
                  r5 = zin(1, j, nin5)
                  s5 = zin(2, j, nin5)
                  r6 = zin(1, j, nin6)
                  s6 = zin(2, j, nin6)
                  r7 = zin(1, j, nin7)
                  s7 = zin(2, j, nin7)
                  r8 = zin(1, j, nin8)
                  s8 = zin(2, j, nin8)
                  r = r1+r5
                  s = r3+r7
                  ap = r+s
                  am = r-s
                  r = r2+r6
                  s = r4+r8
                  bp = r+s
                  bm = r-s
                  r = s1+s5
                  s = s3+s7
                  cp = r+s
                  cm = r-s
                  r = s2+s6
                  s = s4+s8
                  dbl = r+s
                  dm = r-s
                  zout(1, j, nout1) = ap+bp
                  zout(2, j, nout1) = cp+dbl
                  zout(1, j, nout5) = ap-bp
                  zout(2, j, nout5) = cp-dbl
                  zout(1, j, nout3) = am+dm
                  zout(2, j, nout3) = cm-bm
                  zout(1, j, nout7) = am-dm
                  zout(2, j, nout7) = cm+bm
                  r = r1-r5
                  s = s3-s7
                  ap = r+s
                  am = r-s
                  r = s1-s5
                  s = r3-r7
                  bp = r+s
                  bm = r-s
                  r = s4-s8
                  s = r2-r6
                  cp = r+s
                  cm = r-s
                  r = s2-s6
                  s = r4-r8
                  dbl = r+s
                  dm = r-s
                  r = (cp+dm)*rt2i
                  s = (-cp+dm)*rt2i
                  cp = (cm+dbl)*rt2i
                  dbl = (cm-dbl)*rt2i
                  zout(1, j, nout2) = ap+r
                  zout(2, j, nout2) = bm+s
                  zout(1, j, nout6) = ap-r
                  zout(2, j, nout6) = bm-s
                  zout(1, j, nout4) = am+cp
                  zout(2, j, nout4) = bp+dbl
                  zout(1, j, nout8) = am-cp
                  zout(2, j, nout8) = bp-dbl
               END DO
            END DO
         ELSE
            ia = 1
            nin1 = ia-after
            nout1 = ia-atn
            DO ib = 1, before
               nin1 = nin1+after
               nin2 = nin1+atb
               nin3 = nin2+atb
               nin4 = nin3+atb
               nin5 = nin4+atb
               nin6 = nin5+atb
               nin7 = nin6+atb
               nin8 = nin7+atb
               nout1 = nout1+atn
               nout2 = nout1+after
               nout3 = nout2+after
               nout4 = nout3+after
               nout5 = nout4+after
               nout6 = nout5+after
               nout7 = nout6+after
               nout8 = nout7+after
               DO j = 1, nfft
                  r1 = zin(1, j, nin1)
                  s1 = zin(2, j, nin1)
                  r2 = zin(1, j, nin2)
                  s2 = zin(2, j, nin2)
                  r3 = zin(1, j, nin3)
                  s3 = zin(2, j, nin3)
                  r4 = zin(1, j, nin4)
                  s4 = zin(2, j, nin4)
                  r5 = zin(1, j, nin5)
                  s5 = zin(2, j, nin5)
                  r6 = zin(1, j, nin6)
                  s6 = zin(2, j, nin6)
                  r7 = zin(1, j, nin7)
                  s7 = zin(2, j, nin7)
                  r8 = zin(1, j, nin8)
                  s8 = zin(2, j, nin8)
                  r = r1+r5
                  s = r3+r7
                  ap = r+s
                  am = r-s
                  r = r2+r6
                  s = r4+r8
                  bp = r+s
                  bm = r-s
                  r = s1+s5
                  s = s3+s7
                  cp = r+s
                  cm = r-s
                  r = s2+s6
                  s = s4+s8
                  dbl = r+s
                  dm = r-s
                  zout(1, j, nout1) = ap+bp
                  zout(2, j, nout1) = cp+dbl
                  zout(1, j, nout5) = ap-bp
                  zout(2, j, nout5) = cp-dbl
                  zout(1, j, nout3) = am-dm
                  zout(2, j, nout3) = cm+bm
                  zout(1, j, nout7) = am+dm
                  zout(2, j, nout7) = cm-bm
                  r = r1-r5
                  s = -s3+s7
                  ap = r+s
                  am = r-s
                  r = s1-s5
                  s = r7-r3
                  bp = r+s
                  bm = r-s
                  r = -s4+s8
                  s = r2-r6
                  cp = r+s
                  cm = r-s
                  r = -s2+s6
                  s = r4-r8
                  dbl = r+s
                  dm = r-s
                  r = (cp+dm)*rt2i
                  s = (cp-dm)*rt2i
                  cp = (cm+dbl)*rt2i
                  dbl = (-cm+dbl)*rt2i
                  zout(1, j, nout2) = ap+r
                  zout(2, j, nout2) = bm+s
                  zout(1, j, nout6) = ap-r
                  zout(2, j, nout6) = bm-s
                  zout(1, j, nout4) = am+cp
                  zout(2, j, nout4) = bp+dbl
                  zout(1, j, nout8) = am-cp
                  zout(2, j, nout8) = bp-dbl
               END DO
            END DO
         END IF
      ELSE IF (now == 3) THEN
         bbs = isign*bb
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            DO j = 1, nfft
               r1 = zin(1, j, nin1)
               s1 = zin(2, j, nin1)
               r2 = zin(1, j, nin2)
               s2 = zin(2, j, nin2)
               r3 = zin(1, j, nin3)
               s3 = zin(2, j, nin3)
               r = r2+r3
               s = s2+s3
               zout(1, j, nout1) = r+r1
               zout(2, j, nout1) = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r2 = bbs*(r2-r3)
               s2 = bbs*(s2-s3)
               zout(1, j, nout2) = r1-s2
               zout(2, j, nout2) = s1+r2
               zout(1, j, nout3) = r1+s2
               zout(2, j, nout3) = s1-r2
            END DO
         END DO
         DO ia = 2, after
            ias = ia-1
            IF (4*ias == 3*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r2 = -zin(2, j, nin2)
                        s2 = zin(1, j, nin2)
                        r3 = -zin(1, j, nin3)
                        s3 = -zin(2, j, nin3)
                        r = r2+r3
                        s = s2+s3
                        zout(1, j, nout1) = r+r1
                        zout(2, j, nout1) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, j, nout2) = r1-s2
                        zout(2, j, nout2) = s1+r2
                        zout(1, j, nout3) = r1+s2
                        zout(2, j, nout3) = s1-r2
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r2 = zin(2, j, nin2)
                        s2 = -zin(1, j, nin2)
                        r3 = -zin(1, j, nin3)
                        s3 = -zin(2, j, nin3)
                        r = r2+r3
                        s = s2+s3
                        zout(1, j, nout1) = r+r1
                        zout(2, j, nout1) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, j, nout2) = r1-s2
                        zout(2, j, nout2) = s1+r2
                        zout(1, j, nout3) = r1+s2
                        zout(2, j, nout3) = s1-r2
                     END DO
                  END DO
               END IF
            ELSE IF (8*ias == 3*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, j, nin3)
                        s3 = zin(1, j, nin3)
                        r = r2+r3
                        s = s2+s3
                        zout(1, j, nout1) = r+r1
                        zout(2, j, nout1) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, j, nout2) = r1-s2
                        zout(2, j, nout2) = s1+r2
                        zout(1, j, nout3) = r1+s2
                        zout(2, j, nout3) = s1-r2
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r+s)*rt2i
                        s2 = (-r+s)*rt2i
                        r3 = zin(2, j, nin3)
                        s3 = -zin(1, j, nin3)
                        r = r2+r3
                        s = s2+s3
                        zout(1, j, nout1) = r+r1
                        zout(2, j, nout1) = s+s1
                        r1 = r1-0.5_dp*r
                        s1 = s1-0.5_dp*s
                        r2 = bbs*(r2-r3)
                        s2 = bbs*(s2-s3)
                        zout(1, j, nout2) = r1-s2
                        zout(2, j, nout2) = s1+r2
                        zout(1, j, nout3) = r1+s2
                        zout(2, j, nout3) = s1-r2
                     END DO
                  END DO
               END IF
            ELSE
               itt = ias*before
               itrig = itt+1
               cr2 = trig(1, itrig)
               ci2 = trig(2, itrig)
               itrig = itrig+itt
               cr3 = trig(1, itrig)
               ci3 = trig(2, itrig)
               nin1 = ia-after
               nout1 = ia-atn
               DO ib = 1, before
                  nin1 = nin1+after
                  nin2 = nin1+atb
                  nin3 = nin2+atb
                  nout1 = nout1+atn
                  nout2 = nout1+after
                  nout3 = nout2+after
                  DO j = 1, nfft
                     r1 = zin(1, j, nin1)
                     s1 = zin(2, j, nin1)
                     r = zin(1, j, nin2)
                     s = zin(2, j, nin2)
                     r2 = r*cr2-s*ci2
                     s2 = r*ci2+s*cr2
                     r = zin(1, j, nin3)
                     s = zin(2, j, nin3)
                     r3 = r*cr3-s*ci3
                     s3 = r*ci3+s*cr3
                     r = r2+r3
                     s = s2+s3
                     zout(1, j, nout1) = r+r1
                     zout(2, j, nout1) = s+s1
                     r1 = r1-0.5_dp*r
                     s1 = s1-0.5_dp*s
                     r2 = bbs*(r2-r3)
                     s2 = bbs*(s2-s3)
                     zout(1, j, nout2) = r1-s2
                     zout(2, j, nout2) = s1+r2
                     zout(1, j, nout3) = r1+s2
                     zout(2, j, nout3) = s1-r2
                  END DO
               END DO
            END IF
         END DO
      ELSE IF (now == 5) THEN
         sin2 = isign*sin2p
         sin4 = isign*sin4p
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nin4 = nin3+atb
            nin5 = nin4+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            nout4 = nout3+after
            nout5 = nout4+after
            DO j = 1, nfft
               r1 = zin(1, j, nin1)
               s1 = zin(2, j, nin1)
               r2 = zin(1, j, nin2)
               s2 = zin(2, j, nin2)
               r3 = zin(1, j, nin3)
               s3 = zin(2, j, nin3)
               r4 = zin(1, j, nin4)
               s4 = zin(2, j, nin4)
               r5 = zin(1, j, nin5)
               s5 = zin(2, j, nin5)
               r25 = r2+r5
               r34 = r3+r4
               s25 = s2-s5
               s34 = s3-s4
               zout(1, j, nout1) = r1+r25+r34
               r = cos2*r25+cos4*r34+r1
               s = sin2*s25+sin4*s34
               zout(1, j, nout2) = r-s
               zout(1, j, nout5) = r+s
               r = cos4*r25+cos2*r34+r1
               s = sin4*s25-sin2*s34
               zout(1, j, nout3) = r-s
               zout(1, j, nout4) = r+s
               r25 = r2-r5
               r34 = r3-r4
               s25 = s2+s5
               s34 = s3+s4
               zout(2, j, nout1) = s1+s25+s34
               r = cos2*s25+cos4*s34+s1
               s = sin2*r25+sin4*r34
               zout(2, j, nout2) = r+s
               zout(2, j, nout5) = r-s
               r = cos4*s25+cos2*s34+s1
               s = sin4*r25-sin2*r34
               zout(2, j, nout3) = r+s
               zout(2, j, nout4) = r-s
            END DO
         END DO
         DO ia = 2, after
            ias = ia-1
            IF (8*ias == 5*after) THEN
               IF (isign == 1) THEN
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nin5 = nin4+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     nout5 = nout4+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r-s)*rt2i
                        s2 = (r+s)*rt2i
                        r3 = -zin(2, j, nin3)
                        s3 = zin(1, j, nin3)
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = -(r+s)*rt2i
                        s4 = (r-s)*rt2i
                        r5 = -zin(1, j, nin5)
                        s5 = -zin(2, j, nin5)
                        r25 = r2+r5
                        r34 = r3+r4
                        s25 = s2-s5
                        s34 = s3-s4
                        zout(1, j, nout1) = r1+r25+r34
                        r = cos2*r25+cos4*r34+r1
                        s = sin2*s25+sin4*s34
                        zout(1, j, nout2) = r-s
                        zout(1, j, nout5) = r+s
                        r = cos4*r25+cos2*r34+r1
                        s = sin4*s25-sin2*s34
                        zout(1, j, nout3) = r-s
                        zout(1, j, nout4) = r+s
                        r25 = r2-r5
                        r34 = r3-r4
                        s25 = s2+s5
                        s34 = s3+s4
                        zout(2, j, nout1) = s1+s25+s34
                        r = cos2*s25+cos4*s34+s1
                        s = sin2*r25+sin4*r34
                        zout(2, j, nout2) = r+s
                        zout(2, j, nout5) = r-s
                        r = cos4*s25+cos2*s34+s1
                        s = sin4*r25-sin2*r34
                        zout(2, j, nout3) = r+s
                        zout(2, j, nout4) = r-s
                     END DO
                  END DO
               ELSE
                  nin1 = ia-after
                  nout1 = ia-atn
                  DO ib = 1, before
                     nin1 = nin1+after
                     nin2 = nin1+atb
                     nin3 = nin2+atb
                     nin4 = nin3+atb
                     nin5 = nin4+atb
                     nout1 = nout1+atn
                     nout2 = nout1+after
                     nout3 = nout2+after
                     nout4 = nout3+after
                     nout5 = nout4+after
                     DO j = 1, nfft
                        r1 = zin(1, j, nin1)
                        s1 = zin(2, j, nin1)
                        r = zin(1, j, nin2)
                        s = zin(2, j, nin2)
                        r2 = (r+s)*rt2i
                        s2 = (-r+s)*rt2i
                        r3 = zin(2, j, nin3)
                        s3 = -zin(1, j, nin3)
                        r = zin(1, j, nin4)
                        s = zin(2, j, nin4)
                        r4 = (s-r)*rt2i
                        s4 = -(r+s)*rt2i
                        r5 = -zin(1, j, nin5)
                        s5 = -zin(2, j, nin5)
                        r25 = r2+r5
                        r34 = r3+r4
                        s25 = s2-s5
                        s34 = s3-s4
                        zout(1, j, nout1) = r1+r25+r34
                        r = cos2*r25+cos4*r34+r1
                        s = sin2*s25+sin4*s34
                        zout(1, j, nout2) = r-s
                        zout(1, j, nout5) = r+s
                        r = cos4*r25+cos2*r34+r1
                        s = sin4*s25-sin2*s34
                        zout(1, j, nout3) = r-s
                        zout(1, j, nout4) = r+s
                        r25 = r2-r5
                        r34 = r3-r4
                        s25 = s2+s5
                        s34 = s3+s4
                        zout(2, j, nout1) = s1+s25+s34
                        r = cos2*s25+cos4*s34+s1
                        s = sin2*r25+sin4*r34
                        zout(2, j, nout2) = r+s
                        zout(2, j, nout5) = r-s
                        r = cos4*s25+cos2*s34+s1
                        s = sin4*r25-sin2*r34
                        zout(2, j, nout3) = r+s
                        zout(2, j, nout4) = r-s
                     END DO
                  END DO
               END IF
            ELSE
               ias = ia-1
               itt = ias*before
               itrig = itt+1
               cr2 = trig(1, itrig)
               ci2 = trig(2, itrig)
               itrig = itrig+itt
               cr3 = trig(1, itrig)
               ci3 = trig(2, itrig)
               itrig = itrig+itt
               cr4 = trig(1, itrig)
               ci4 = trig(2, itrig)
               itrig = itrig+itt
               cr5 = trig(1, itrig)
               ci5 = trig(2, itrig)
               nin1 = ia-after
               nout1 = ia-atn
               DO ib = 1, before
                  nin1 = nin1+after
                  nin2 = nin1+atb
                  nin3 = nin2+atb
                  nin4 = nin3+atb
                  nin5 = nin4+atb
                  nout1 = nout1+atn
                  nout2 = nout1+after
                  nout3 = nout2+after
                  nout4 = nout3+after
                  nout5 = nout4+after
                  DO j = 1, nfft
                     r1 = zin(1, j, nin1)
                     s1 = zin(2, j, nin1)
                     r = zin(1, j, nin2)
                     s = zin(2, j, nin2)
                     r2 = r*cr2-s*ci2
                     s2 = r*ci2+s*cr2
                     r = zin(1, j, nin3)
                     s = zin(2, j, nin3)
                     r3 = r*cr3-s*ci3
                     s3 = r*ci3+s*cr3
                     r = zin(1, j, nin4)
                     s = zin(2, j, nin4)
                     r4 = r*cr4-s*ci4
                     s4 = r*ci4+s*cr4
                     r = zin(1, j, nin5)
                     s = zin(2, j, nin5)
                     r5 = r*cr5-s*ci5
                     s5 = r*ci5+s*cr5
                     r25 = r2+r5
                     r34 = r3+r4
                     s25 = s2-s5
                     s34 = s3-s4
                     zout(1, j, nout1) = r1+r25+r34
                     r = cos2*r25+cos4*r34+r1
                     s = sin2*s25+sin4*s34
                     zout(1, j, nout2) = r-s
                     zout(1, j, nout5) = r+s
                     r = cos4*r25+cos2*r34+r1
                     s = sin4*s25-sin2*s34
                     zout(1, j, nout3) = r-s
                     zout(1, j, nout4) = r+s
                     r25 = r2-r5
                     r34 = r3-r4
                     s25 = s2+s5
                     s34 = s3+s4
                     zout(2, j, nout1) = s1+s25+s34
                     r = cos2*s25+cos4*s34+s1
                     s = sin2*r25+sin4*r34
                     zout(2, j, nout2) = r+s
                     zout(2, j, nout5) = r-s
                     r = cos4*s25+cos2*s34+s1
                     s = sin4*r25-sin2*r34
                     zout(2, j, nout3) = r+s
                     zout(2, j, nout4) = r-s
                  END DO
               END DO
            END IF
         END DO
      ELSE IF (now == 6) THEN
         bbs = isign*bb
         ia = 1
         nin1 = ia-after
         nout1 = ia-atn
         DO ib = 1, before
            nin1 = nin1+after
            nin2 = nin1+atb
            nin3 = nin2+atb
            nin4 = nin3+atb
            nin5 = nin4+atb
            nin6 = nin5+atb
            nout1 = nout1+atn
            nout2 = nout1+after
            nout3 = nout2+after
            nout4 = nout3+after
            nout5 = nout4+after
            nout6 = nout5+after
            DO j = 1, nfft
               r2 = zin(1, j, nin3)
               s2 = zin(2, j, nin3)
               r3 = zin(1, j, nin5)
               s3 = zin(2, j, nin5)
               r = r2+r3
               s = s2+s3
               r1 = zin(1, j, nin1)
               s1 = zin(2, j, nin1)
               ur1 = r+r1
               ui1 = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r = r2-r3
               s = s2-s3
               ur2 = r1-s*bbs
               ui2 = s1+r*bbs
               ur3 = r1+s*bbs
               ui3 = s1-r*bbs

               r2 = zin(1, j, nin6)
               s2 = zin(2, j, nin6)
               r3 = zin(1, j, nin2)
               s3 = zin(2, j, nin2)
               r = r2+r3
               s = s2+s3
               r1 = zin(1, j, nin4)
               s1 = zin(2, j, nin4)
               vr1 = r+r1
               vi1 = s+s1
               r1 = r1-0.5_dp*r
               s1 = s1-0.5_dp*s
               r = r2-r3
               s = s2-s3
               vr2 = r1-s*bbs
               vi2 = s1+r*bbs
               vr3 = r1+s*bbs
               vi3 = s1-r*bbs

               zout(1, j, nout1) = ur1+vr1
               zout(2, j, nout1) = ui1+vi1
               zout(1, j, nout5) = ur2+vr2
               zout(2, j, nout5) = ui2+vi2
               zout(1, j, nout3) = ur3+vr3
               zout(2, j, nout3) = ui3+vi3
               zout(1, j, nout4) = ur1-vr1
               zout(2, j, nout4) = ui1-vi1
               zout(1, j, nout2) = ur2-vr2
               zout(2, j, nout2) = ui2-vi2
               zout(1, j, nout6) = ur3-vr3
               zout(2, j, nout6) = ui3-vi3
            END DO
         END DO
      ELSE
         CALL cp__b("pw/fft/mltfftsg_tools.F",3451,'Error fftstp')
      END IF

!-----------------------------------------------------------------------------!

   END SUBROUTINE fftstp

!-----------------------------------------------------------------------------!
!-----------------------------------------------------------------------------!
!  Copyright (C) Stefan Goedecker, Lausanne, Switzerland, August 1, 1991
!  Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!  Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!  This file is distributed under the terms of the
!  GNU General Public License version 2 (or later),
!  see http://www.gnu.org/copyleft/gpl.txt .
!-----------------------------------------------------------------------------!
!  S. Goedecker: Rotating a three-dimensional array in optimal
!  positions for vector processing: Case study for a three-dimensional Fast
!  Fourier Transform, Comp. Phys. Commun. 76, 294 (1993)
! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param trig ...
!> \param after ...
!> \param before ...
!> \param now ...
!> \param isign ...
!> \param ic ...
! **************************************************************************************************
   SUBROUTINE ctrig(n, trig, after, before, now, isign, ic)
      INTEGER, INTENT(IN)                                :: n
      REAL(dp), DIMENSION(2, ctrig_length), INTENT(OUT)  :: trig
      INTEGER, DIMENSION(7), INTENT(OUT)                 :: after, before, now
      INTEGER, INTENT(IN)                                :: isign
      INTEGER, INTENT(OUT)                               :: ic

      INTEGER, PARAMETER                                 :: nt = 82
      INTEGER, DIMENSION(7, nt), PARAMETER :: idata = RESHAPE((/3, 3, 1, 1, 1, 1, 1, 4, 4, 1, 1, 1,&
         1, 1, 5, 5, 1, 1, 1, 1, 1, 6, 6, 1, 1, 1, 1, 1, 8, 8, 1, 1, 1, 1, 1, 9, 3, 3, 1, 1, 1, 1, &
         12, 4, 3, 1, 1, 1, 1, 15, 5, 3, 1, 1, 1, 1, 16, 4, 4, 1, 1, 1, 1, 18, 6, 3, 1, 1, 1, 1, 20&
         , 5, 4, 1, 1, 1, 1, 24, 8, 3, 1, 1, 1, 1, 25, 5, 5, 1, 1, 1, 1, 27, 3, 3, 3, 1, 1, 1, 30, &
         6, 5, 1, 1, 1, 1, 32, 8, 4, 1, 1, 1, 1, 36, 4, 3, 3, 1, 1, 1, 40, 8, 5, 1, 1, 1, 1, 45, 5 &
         , 3, 3, 1, 1, 1, 48, 4, 4, 3, 1, 1, 1, 54, 6, 3, 3, 1, 1, 1, 60, 5, 4, 3, 1, 1, 1, 64, 4, &
         4, 4, 1, 1, 1, 72, 8, 3, 3, 1, 1, 1, 75, 5, 5, 3, 1, 1, 1, 80, 5, 4, 4, 1, 1, 1, 81, 3, 3 &
         , 3, 3, 1, 1, 90, 6, 5, 3, 1, 1, 1, 96, 8, 4, 3, 1, 1, 1, 100, 5, 5, 4, 1, 1, 1, 108, 4, 3&
         , 3, 3, 1, 1, 120, 8, 5, 3, 1, 1, 1, 125, 5, 5, 5, 1, 1, 1, 128, 8, 4, 4, 1, 1, 1, 135, 5 &
         , 3, 3, 3, 1, 1, 144, 4, 4, 3, 3, 1, 1, 150, 6, 5, 5, 1, 1, 1, 160, 8, 5, 4, 1, 1, 1, 162,&
         6, 3, 3, 3, 1, 1, 180, 5, 4, 3, 3, 1, 1, 192, 4, 4, 4, 3, 1, 1, 200, 8, 5, 5, 1, 1, 1, 216&
         , 8, 3, 3, 3, 1, 1, 225, 5, 5, 3, 3, 1, 1, 240, 5, 4, 4, 3, 1, 1, 243, 3, 3, 3, 3, 3, 1, &
         256, 4, 4, 4, 4, 1, 1, 270, 6, 5, 3, 3, 1, 1, 288, 8, 4, 3, 3, 1, 1, 300, 5, 5, 4, 3, 1, 1&
         , 320, 5, 4, 4, 4, 1, 1, 324, 4, 3, 3, 3, 3, 1, 360, 8, 5, 3, 3, 1, 1, 375, 5, 5, 5, 3, 1,&
         1, 384, 8, 4, 4, 3, 1, 1, 400, 5, 5, 4, 4, 1, 1, 405, 5, 3, 3, 3, 3, 1, 432, 4, 4, 3, 3, 3&
         , 1, 450, 6, 5, 5, 3, 1, 1, 480, 8, 5, 4, 3, 1, 1, 486, 6, 3, 3, 3, 3, 1, 500, 5, 5, 5, 4,&
         1, 1, 512, 8, 4, 4, 4, 1, 1, 540, 5, 4, 3, 3, 3, 1, 576, 4, 4, 4, 3, 3, 1, 600, 8, 5, 5, 3&
         , 1, 1, 625, 5, 5, 5, 5, 1, 1, 640, 8, 5, 4, 4, 1, 1, 648, 8, 3, 3, 3, 3, 1, 675, 5, 5, 3,&
         3, 3, 1, 720, 5, 4, 4, 3, 3, 1, 729, 3, 3, 3, 3, 3, 3, 750, 6, 5, 5, 5, 1, 1, 768, 4, 4, 4&
         , 4, 3, 1, 800, 8, 5, 5, 4, 1, 1, 810, 6, 5, 3, 3, 3, 1, 864, 8, 4, 3, 3, 3, 1, 900, 5, 5,&
         4, 3, 3, 1, 960, 5, 4, 4, 4, 3, 1, 972, 4, 3, 3, 3, 3, 3, 1000, 8, 5, 5, 5, 1, 1, &
         ctrig_length, 4, 4, 4, 4, 4, 1/), (/7, nt/))

      INTEGER                                            :: i, itt, j
      REAL(dp)                                           :: angle, twopi

      mloop: DO i = 1, nt
         IF (n == idata(1, i)) THEN
            ic = 0
            DO j = 1, 6
               itt = idata(1+j, i)
               IF (itt > 1) THEN
                  ic = ic+1
                  now(j) = idata(1+j, i)
               ELSE
                  EXIT mloop
               END IF
            END DO
            EXIT mloop
         END IF
         IF (i == nt) THEN
            WRITE (*, '(A,i5,A)') " Value of ", n, &
               " not allowed for fft, allowed values are:"
            WRITE (*, '(15i5)') (idata(1, j), j=1, nt)
            CALL cp__b("pw/fft/mltfftsg_tools.F",3532,'ctrig')
         END IF
      END DO mloop

      after(1) = 1
      before(ic) = 1
      DO i = 2, ic
         after(i) = after(i-1)*now(i-1)
         before(ic-i+1) = before(ic-i+2)*now(ic-i+2)
      END DO

      twopi = 8._dp*ATAN(1._dp)
      angle = isign*twopi/REAL(n, dp)
      trig(1, 1) = 1._dp
      trig(2, 1) = 0._dp
      DO i = 1, n-1
         trig(1, i+1) = COS(REAL(i, dp)*angle)
         trig(2, i+1) = SIN(REAL(i, dp)*angle)
      END DO

   END SUBROUTINE ctrig

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param m ...
!> \param a ...
!> \param lda ...
!> \param b ...
!> \param ldb ...
! **************************************************************************************************
   SUBROUTINE matmov(n, m, a, lda, b, ldb)
      INTEGER                                            :: n, m, lda
      COMPLEX(dp)                                        :: a(lda, *)
      INTEGER                                            :: ldb
      COMPLEX(dp)                                        :: b(ldb, *)

      b(1:n, 1:m) = a(1:n, 1:m)
   END SUBROUTINE matmov

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param lda ...
!> \param m ...
!> \param n ...
!> \param b ...
!> \param ldb ...
! **************************************************************************************************
   SUBROUTINE zgetmo(a, lda, m, n, b, ldb)
      INTEGER                                            :: lda, m, n
      COMPLEX(dp)                                        :: a(lda, n)
      INTEGER                                            :: ldb
      COMPLEX(dp)                                        :: b(ldb, m)

      b(1:n, 1:m) = TRANSPOSE(a(1:m, 1:n))
   END SUBROUTINE zgetmo

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param sc ...
!> \param a ...
! **************************************************************************************************
   SUBROUTINE scaled(n, sc, a)
      INTEGER                                            :: n
      REAL(dp)                                           :: sc
      COMPLEX(dp)                                        :: a(n)

      CALL dscal(n, sc, a, 1)

   END SUBROUTINE scaled

END MODULE mltfftsg_tools
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE fftsg_lib
   USE fft_kinds,                       ONLY: dp
   USE mltfftsg_tools,                  ONLY: mltfftsg

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fftsg_lib'

   PUBLIC :: fftsg_do_init, fftsg_do_cleanup, fftsg_get_lengths, fftsg3d, fftsg1dm

CONTAINS

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE fftsg_do_init()

      ! no init needed

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE fftsg_do_cleanup()

      ! no cleanup needed

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param DATA ...
!> \param max_length ...
!> \par History
!>      Adapted to new interface structure
!> \author JGH
! **************************************************************************************************
   SUBROUTINE fftsg_get_lengths(DATA, max_length)

      INTEGER, DIMENSION(*)                              :: DATA
      INTEGER, INTENT(INOUT)                             :: max_length

      INTEGER, PARAMETER                                 :: rlen = 81
      INTEGER, DIMENSION(rlen), PARAMETER :: radix = (/2, 4, 6, 8, 9, 12, 15, 16, 18, 20, 24, 25, &
         27, 30, 32, 36, 40, 45, 48, 54, 60, 64, 72, 75, 80, 81, 90, 96, 100, 108, 120, 125, 128, &
         135, 144, 150, 160, 162, 180, 192, 200, 216, 225, 240, 243, 256, 270, 288, 300, 320, 324, &
         360, 375, 384, 400, 405, 432, 450, 480, 486, 500, 512, 540, 576, 600, 625, 640, 648, 675, &
         720, 729, 750, 768, 800, 810, 864, 900, 960, 972, 1000, 1024/)

      INTEGER                                            :: ndata

!------------------------------------------------------------------------------

      ndata = MIN(max_length, rlen)
      DATA(1:ndata) = RADIX(1:ndata)
      max_length = ndata

   END SUBROUTINE fftsg_get_lengths

! **************************************************************************************************
!> \brief ...
!> \param fft_in_place ...
!> \param fsign ...
!> \param scale ...
!> \param n ...
!> \param zin ...
!> \param zout ...
! **************************************************************************************************
   SUBROUTINE fftsg3d(fft_in_place, fsign, scale, n, zin, zout)

      LOGICAL, INTENT(IN)                                :: fft_in_place
      INTEGER, INTENT(INOUT)                             :: fsign
      REAL(KIND=dp), INTENT(IN)                          :: scale
      INTEGER, DIMENSION(*), INTENT(IN)                  :: n
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT)      :: zin, zout

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: xf, yf
      INTEGER                                            :: nx, ny, nz

!------------------------------------------------------------------------------

      nx = n(1)
      ny = n(2)
      nz = n(3)

      IF (fft_in_place) THEN

         ALLOCATE (xf(nx*ny*nz), yf(nx*ny*nz))

         CALL mltfftsg('N', 'T', zin, nx, ny*nz, xf, ny*nz, nx, nx, &
                       ny*nz, fsign, 1.0_dp)
         CALL mltfftsg('N', 'T', xf, ny, nx*nz, yf, nx*nz, ny, ny, &
                       nx*nz, fsign, 1.0_dp)
         CALL mltfftsg('N', 'T', yf, nz, ny*nx, zin, ny*nx, nz, nz, &
                       ny*nx, fsign, scale)

         DEALLOCATE (xf, yf)

      ELSE

         ALLOCATE (xf(nx*ny*nz))

         CALL mltfftsg('N', 'T', zin, nx, ny*nz, zout, ny*nz, nx, nx, &
                       ny*nz, fsign, 1.0_dp)
         CALL mltfftsg('N', 'T', zout, ny, nx*nz, xf, nx*nz, ny, ny, &
                       nx*nz, fsign, 1.0_dp)
         CALL mltfftsg('N', 'T', xf, nz, ny*nx, zout, ny*nx, nz, nz, &
                       ny*nx, fsign, scale)

         DEALLOCATE (xf)

      END IF

   END SUBROUTINE fftsg3d

! **************************************************************************************************
!> \brief ...
!> \param fsign ...
!> \param trans ...
!> \param n ...
!> \param m ...
!> \param zin ...
!> \param zout ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE fftsg1dm(fsign, trans, n, m, zin, zout, scale)

      INTEGER, INTENT(INOUT)                             :: fsign
      LOGICAL, INTENT(IN)                                :: trans
      INTEGER, INTENT(IN)                                :: n, m
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT)      :: zin
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(OUT)        :: zout
      REAL(KIND=dp), INTENT(IN)                          :: scale

!------------------------------------------------------------------------------

      IF (trans) THEN
         IF (fsign > 0) THEN
            CALL mltfftsg("T", "N", zin, m, n, zout, n, m, n, m, fsign, scale)
         ELSE
            CALL mltfftsg("N", "T", zin, n, m, zout, m, n, n, m, fsign, scale)
         END IF
      ELSE
         CALL mltfftsg("N", "N", zin, n, m, zout, n, m, n, m, fsign, scale)
      END IF

   END SUBROUTINE fftsg1dm

END MODULE

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Type to store data about a (1D or 3D) FFT, including FFTW plan
!> \par History
!>      IAB 09-Jan-2009 : initial version
!>                        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
!>      IAB 09-Oct-2009 : Added additional fields needed when using OpenMP
!>                        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
!> \author JGH
! **************************************************************************************************

MODULE fft_plan
   USE fft_kinds,                       ONLY: integer8_kind

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: fft_plan_type

   TYPE fft_plan_type

      INTEGER                             :: fft_type
      INTEGER                             :: fsign
      LOGICAL                             :: trans, fft_in_place, valid, separated_plans
      INTEGER                             :: n, m
      INTEGER, DIMENSION(3)               :: n_3d

!   Handle for the FFTW plan
      INTEGER(KIND=integer8_kind)        :: fftw_plan

!   Plan for the remaining rows for 1D FFT when number of threads does not divide the number of rows exactly
!$    INTEGER(KIND=integer8_kind)        :: alt_fftw_plan
!$    LOGICAL                             :: need_alt_plan
!$    INTEGER                             :: num_threads_needed, num_rows, alt_num_rows

!   Individual plans (used by hand-optimised 3D FFT)
      INTEGER(KIND=integer8_kind)        :: fftw_plan_nx, fftw_plan_ny, fftw_plan_nz
!   Plans for the remaining rows (when the number of threads does not divide the number of rows exactly)
      INTEGER(KIND=integer8_kind)        :: fftw_plan_nx_r, fftw_plan_ny_r, fftw_plan_nz_r

   END TYPE fft_plan_type

END MODULE fft_plan
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE fftw3_lib

   USE ISO_C_BINDING,                   ONLY: C_CHAR,&
                                              C_INT,&
                                              C_INTPTR_T
   USE cp_files,                        ONLY: get_unit_number
   USE fft_kinds,                       ONLY: dp,&
                                              integer8_kind
   USE fft_plan,                        ONLY: fft_plan_type

  !$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE
  PRIVATE

  PUBLIC :: fftw3_do_init, fftw3_do_cleanup, fftw3_get_lengths, fftw33d, fftw31dm
  PUBLIC :: fftw3_destroy_plan, fftw3_create_plan_1dm, fftw3_create_plan_3d


CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param wisdom_file ...
!> \param ionode ...
! **************************************************************************************************
SUBROUTINE fftw3_do_cleanup(wisdom_file,ionode)

    CHARACTER(LEN=*), INTENT(IN)             :: wisdom_file
    LOGICAL                                  :: ionode


   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(wisdom_file))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(ionode))==-1) EXIT ;  END DO ; ENDIF

END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param wisdom_file ...
! **************************************************************************************************
SUBROUTINE fftw3_do_init(wisdom_file)

    CHARACTER(LEN=*), INTENT(IN)             :: wisdom_file

   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(wisdom_file))==-1) EXIT ;  END DO ; ENDIF

END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param DATA ...
!> \param max_length ...
!> \par History
!>      JGH 23-Jan-2006 : initial version
!>      Adapted for new interface
!>      IAB 09-Jan-2009 : Modified to cache plans in fft_plan_type
!>                        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
!>      IAB 09-Oct-2009 : Added OpenMP directives to 1D FFT, and planning routines
!>                        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
!>      IAB 11-Sep-2012 : OpenMP parallel 3D FFT (Ruyman Reyes, PRACE)
!> \author JGH
! **************************************************************************************************
SUBROUTINE fftw3_get_lengths ( DATA, max_length )


      INTEGER, DIMENSION(*)                              :: DATA
      INTEGER, INTENT(INOUT)                             :: max_length

      INTEGER                                            :: h, i, j, k, m, maxn, maxn_elevens, &
                                                            maxn_fives, maxn_sevens, &
                                                            maxn_thirteens, maxn_threes, &
                                                            maxn_twos, ndata, nmax, number
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: dlocal, idx

!------------------------------------------------------------------------------
! compute ndata
!! FFTW can do arbitrary(?) lengths, maybe you want to limit them to some
!!    powers of small prime numbers though...

  maxn_twos = 15
  maxn_threes = 3
  maxn_fives = 2
  maxn_sevens = 1
  maxn_elevens = 1
  maxn_thirteens = 0
  maxn = 37748736

  ndata = 0
  DO h = 0, maxn_twos
     nmax = HUGE(0) / 2**h
     DO i = 0, maxn_threes
        DO j = 0, maxn_fives
           DO k = 0, maxn_sevens
              DO m = 0, maxn_elevens
                 number = (3**i) * (5**j) * (7**k) * (11**m)

                 IF ( number > nmax ) CYCLE

                 number = number * 2 ** h
                 IF ( number >= maxn ) CYCLE

                 ndata = ndata + 1
              END DO
           END DO
        END DO
     END DO
  END DO

  ALLOCATE ( dlocal ( ndata ), idx ( ndata ) )

  ndata = 0
  dlocal ( : ) = 0
  DO h = 0, maxn_twos
     nmax = HUGE(0) / 2**h
     DO i = 0, maxn_threes
        DO j = 0, maxn_fives
           DO k = 0, maxn_sevens
              DO m = 0, maxn_elevens
                 number = (3**i) * (5**j) * (7**k) * (11**m)

                 IF ( number > nmax ) CYCLE

                 number = number * 2 ** h
                 IF ( number >= maxn ) CYCLE

                 ndata = ndata + 1
                 dlocal ( ndata ) = number
              END DO
           END DO
        END DO
     END DO
  END DO

  CALL sortint ( dlocal, ndata, idx )
  ndata = MIN ( ndata, max_length )
  DATA(1:ndata) = dlocal(1:ndata)
  max_length = ndata

  DEALLOCATE ( dlocal, idx )

END SUBROUTINE fftw3_get_lengths


! **************************************************************************************************
!> \brief ...
!> \param iarr ...
!> \param n ...
!> \param index ...
! **************************************************************************************************
SUBROUTINE sortint ( iarr, n, index )

      INTEGER, INTENT(IN)                                :: n
      INTEGER, INTENT(INOUT)                             :: iarr(1:n)
      INTEGER, INTENT(OUT)                               :: INDEX(1:n)

      INTEGER, PARAMETER                                 :: m = 7, nstack = 50

      INTEGER                                            :: a, i, ib, ir, istack(1:nstack), itemp, &
                                                            j, jstack, k, l, temp

!------------------------------------------------------------------------------

  DO i = 1, n
     INDEX(i) = i
  END DO
  jstack = 0
  l = 1
  ir = n
  DO WHILE(.TRUE.)
  IF (ir-l<m) THEN
     DO j = l + 1, ir
        a = iarr(j)
        ib = INDEX(j)
        DO i = j - 1, 0, -1
           IF(i==0) EXIT
           IF (iarr(i)<=a) EXIT
           iarr(i+1) = iarr(i)
           INDEX(i+1) = INDEX(i)
        END DO
        iarr(i+1) = a
        INDEX(i+1) = ib
     END DO
     IF (jstack==0) RETURN
     ir = istack(jstack)
     l = istack(jstack-1)
     jstack = jstack - 2
  ELSE
     k = (l+ir)/2
     temp = iarr(k)
     iarr(k) = iarr(l+1)
     iarr(l+1) = temp
     itemp = INDEX(k)
     INDEX(k) = INDEX(l+1)
     INDEX(l+1) = itemp
     IF (iarr(l+1)>iarr(ir)) THEN
        temp = iarr(l+1)
        iarr(l+1) = iarr(ir)
        iarr(ir) = temp
        itemp = INDEX(l+1)
        INDEX(l+1) = INDEX(ir)
        INDEX(ir) = itemp
     END IF
     IF (iarr(l)>iarr(ir)) THEN
        temp = iarr(l)
        iarr(l) = iarr(ir)
        iarr(ir) = temp
        itemp = INDEX(l)
        INDEX(l) = INDEX(ir)
        INDEX(ir) = itemp
     END IF
     IF (iarr(l+1)>iarr(l)) THEN
        temp = iarr(l+1)
        iarr(l+1) = iarr(l)
        iarr(l) = temp
        itemp = INDEX(l+1)
        INDEX(l+1) = INDEX(l)
        INDEX(l) = itemp
     END IF
     i = l + 1
     j = ir
     a = iarr(l)
     ib = INDEX(l)
     DO WHILE(.TRUE.)
     i = i + 1
     DO WHILE(iarr(i)<a)
       i = i + 1
     ENDDO
     j = j - 1
     DO WHILE(iarr(j)>a)
       j = j - 1
     ENDDO
     IF (j<i) EXIT
     temp = iarr(i)
     iarr(i) = iarr(j)
     iarr(j) = temp
     itemp = INDEX(i)
     INDEX(i) = INDEX(j)
     INDEX(j) = itemp
     ENDDO
     iarr(l) = iarr(j)
     iarr(j) = a
     INDEX(l) = INDEX(j)
     INDEX(j) = ib
     jstack = jstack + 2
     IF (jstack>nstack) CALL cp__b("pw/fft/fftw3_lib.F",347," Nstack too small in sortr")
     IF (ir-i+1>=j-l) THEN
        istack(jstack) = ir
        istack(jstack-1) = i
        ir = j - 1
     ELSE
        istack(jstack) = j - 1
        istack(jstack-1) = l
        l = i
     END IF
  END IF

  ENDDO

END SUBROUTINE sortint

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param fft_rank ...
!> \param dim_n ...
!> \param dim_istride ...
!> \param dim_ostride ...
!> \param hm_rank ...
!> \param hm_n ...
!> \param hm_istride ...
!> \param hm_ostride ...
!> \param zin ...
!> \param zout ...
!> \param fft_direction ...
!> \param fftw_plan_type ...
!> \param valid ...
! **************************************************************************************************
SUBROUTINE fftw3_create_guru_plan(plan, fft_rank, dim_n, &
                                  dim_istride, dim_ostride, hm_rank, &
                                  hm_n, hm_istride, hm_ostride, &
                                  zin, zout, fft_direction, fftw_plan_type, &
                                  valid)


  IMPLICIT NONE

  INTEGER(KIND=integer8_kind), INTENT ( INOUT )      :: plan
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(IN)         :: zin
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(IN)         :: zout
  INTEGER, INTENT(IN) :: dim_n(2), dim_istride(2), dim_ostride(2), &
             hm_n(2), hm_istride(2), hm_ostride(2), fft_rank, &
             fft_direction, fftw_plan_type, hm_rank
  LOGICAL, INTENT(OUT)                               :: valid

  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(fft_rank))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dim_n))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dim_istride))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dim_ostride))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hm_rank))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hm_n))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hm_istride))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(hm_ostride))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(fft_direction))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(fftw_plan_type))==-1) EXIT ;  END DO ; ENDIF
  !MARK_USED does not work with assumed size arguments
  IF(.FALSE.)THEN; DO; IF(ABS(zin(1))>ABS(zout(1)))EXIT; ENDDO; ENDIF
  valid = .FALSE.


END SUBROUTINE

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \retval is_mkl ...
! **************************************************************************************************
FUNCTION fftw3_is_mkl_wrapper() RESULT(is_mkl)

  IMPLICIT NONE

  LOGICAL :: is_mkl
  is_mkl = .FALSE.

END FUNCTION

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param nrows ...
!> \param nt ...
!> \param rows_per_thread ...
!> \param rows_per_thread_r ...
!> \param th_planA ...
!> \param th_planB ...
! **************************************************************************************************
SUBROUTINE fftw3_compute_rows_per_th(nrows,nt,rows_per_thread,rows_per_thread_r,&
                                     th_planA, th_planB)

      INTEGER, INTENT(IN)                                :: nrows, nt
      INTEGER, INTENT(OUT)                               :: rows_per_thread, rows_per_thread_r, &
                                                            th_planA, th_planB

         IF (MOD(nrows,nt) .EQ. 0) THEN
            rows_per_thread = nrows/nt
            rows_per_thread_r = 0
            th_planA = nt 
            th_planB = 0
         ELSE
            rows_per_thread   = nrows/nt + 1
            rows_per_thread_r = nrows/nt
            th_planA = MOD(nrows,nt)
            th_planB = nt - th_planA
         ENDIF

END SUBROUTINE


! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param plan_r ...
!> \param dim_n ...
!> \param dim_istride ...
!> \param dim_ostride ...
!> \param hm_n ...
!> \param hm_istride ...
!> \param hm_ostride ...
!> \param input ...
!> \param output ...
!> \param fft_direction ...
!> \param fftw_plan_type ...
!> \param rows_per_th ...
!> \param rows_per_th_r ...
! **************************************************************************************************
SUBROUTINE fftw3_create_3d_plans(plan, plan_r, dim_n, dim_istride, dim_ostride, &
                                 hm_n, hm_istride, hm_ostride, &
                                 input, output, &
                                 fft_direction, fftw_plan_type, rows_per_th, & 
                                 rows_per_th_r) 


      INTEGER(KIND=integer8_kind), INTENT(INOUT)         :: plan, plan_r
      INTEGER, INTENT(INOUT)                             :: dim_n(2), dim_istride(2), &
                                                            dim_ostride(2), hm_n(2), &
                                                            hm_istride(2), hm_ostride(2)
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT)      :: input, output
      INTEGER, INTENT(INOUT)                             :: fft_direction, fftw_plan_type
      INTEGER, INTENT(IN)                                :: rows_per_th, rows_per_th_r

      LOGICAL                                            :: valid

! First plans will have an additional row

    hm_n(2) = rows_per_th
    CALL fftw3_create_guru_plan(plan,1, &
                                 dim_n,dim_istride,dim_ostride, &
                                 2,hm_n,hm_istride,hm_ostride, &
                                 input, output, &
                                 fft_direction, fftw_plan_type, valid)

    IF (.NOT. valid) THEN
         CALL cp__b("pw/fft/fftw3_lib.F",562,"fftw3_create_plan")
    ENDIF

    !!!! Remainder
    hm_n(2) = rows_per_th_r
    CALL fftw3_create_guru_plan(plan_r,1, &
                                 dim_n,dim_istride,dim_ostride, &
                                 2,hm_n,hm_istride,hm_ostride, &
                                 input, output, &
                                 fft_direction, fftw_plan_type, valid)
    IF (.NOT. valid) THEN
         CALL cp__b("pw/fft/fftw3_lib.F",573,"fftw3_create_plan (remaining)")
    ENDIF


END SUBROUTINE

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param zin ...
!> \param zout ...
!> \param plan_style ...
! **************************************************************************************************
SUBROUTINE fftw3_create_plan_3d(plan, zin, zout, plan_style)

  TYPE(fft_plan_type), INTENT ( INOUT )              :: plan
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT)      :: zin
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT)      :: zout
  INTEGER                                            :: plan_style
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan_style))==-1) EXIT ;  END DO ; ENDIF
   !MARK_USED does not work with assumed size arguments
   IF(.FALSE.)THEN; DO; IF(ABS(zin(1))>ABS(zout(1)))EXIT; ENDDO; ENDIF

END SUBROUTINE fftw3_create_plan_3d


! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param plan_r ...
!> \param split_dim ...
!> \param nt ...
!> \param tid ...
!> \param input ...
!> \param istride ...
!> \param output ...
!> \param ostride ...
! **************************************************************************************************
SUBROUTINE fftw3_workshare_execute_dft(plan, plan_r, split_dim, nt, tid, &
                                       input, istride, output, ostride)

  INTEGER, INTENT(IN)                           :: split_dim,nt, tid
  INTEGER, INTENT(IN)                           :: istride, ostride
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT) :: input, output
    INTEGER (KIND=integer8_kind)                :: plan, plan_r
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan_r))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(split_dim))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nt))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tid))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(istride))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(ostride))==-1) EXIT ;  END DO ; ENDIF
   !MARK_USED does not work with assumed size arguments
   IF(.FALSE.)THEN; DO; IF(ABS(input(1))>ABS(output(1)))EXIT; ENDDO; ENDIF

END SUBROUTINE


! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param scale ...
!> \param zin ...
!> \param zout ...
!> \param stat ...
! **************************************************************************************************
SUBROUTINE fftw33d ( plan, scale, zin, zout, stat )

  TYPE(fft_plan_type), INTENT(IN)                      :: plan
  REAL(KIND=dp), INTENT(IN)                            :: scale
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT), TARGET:: zin
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT), TARGET:: zout
  INTEGER, INTENT(OUT)                                 :: stat
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scale))==-1) EXIT ;  END DO ; ENDIF
  !MARK_USED does not work with assumed size arguments
  IF(.FALSE.)THEN; DO; IF(ABS(zin(1))>ABS(zout(1)))EXIT; ENDDO; ENDIF
  stat = 0


END SUBROUTINE fftw33d

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param zin ...
!> \param zout ...
!> \param plan_style ...
! **************************************************************************************************
SUBROUTINE fftw3_create_plan_1dm(plan, zin, zout, plan_style)

  IMPLICIT NONE

  TYPE(fft_plan_type), INTENT ( INOUT )              :: plan
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(IN)         :: zin
  COMPLEX(KIND=dp), DIMENSION(*), INTENT(IN)         :: zout
  INTEGER, INTENT(IN)                                :: plan_style
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan_style))==-1) EXIT ;  END DO ; ENDIF
   !MARK_USED does not work with assumed size arguments
   IF(.FALSE.)THEN; DO; IF(ABS(zin(1))>ABS(zout(1)))EXIT; ENDDO; ENDIF

END SUBROUTINE fftw3_create_plan_1dm

! **************************************************************************************************
!> \brief ...
!> \param plan ...
! **************************************************************************************************
SUBROUTINE fftw3_destroy_plan ( plan )

  TYPE(fft_plan_type), INTENT (INOUT)   :: plan

   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF

END SUBROUTINE fftw3_destroy_plan

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param zin ...
!> \param zout ...
!> \param scale ...
!> \param stat ...
! **************************************************************************************************
SUBROUTINE fftw31dm ( plan, zin, zout, scale, stat )
      TYPE(fft_plan_type), INTENT(IN)                    :: plan
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(IN), TARGET :: zin
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT), &
         TARGET                                          :: zout
      REAL(KIND=dp), INTENT(IN)                          :: scale
      INTEGER, INTENT(OUT)                               :: stat

      COMPLEX(KIND=dp), POINTER                          :: zin_ptr, zout_ptr, zscal_ptr
      INTEGER                                            :: in_offset, my_id, num_rows, out_offset, &
                                                            scal_offset
      INTEGER(KIND=C_INTPTR_T)                           :: fftw_plan

!------------------------------------------------------------------------------

my_id = 0
num_rows = plan%m

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(my_id,num_rows,zin_ptr,zout_ptr,zscal_ptr,in_offset,out_offset,scal_offset,fftw_plan), &
!$OMP          SHARED(zin,zout), &
!$OMP          SHARED(plan,scale,stat)
!$ my_id = omp_get_thread_num()

!$ if (my_id < plan%num_threads_needed) then

fftw_plan = plan%fftw_plan

in_offset = 1
out_offset = 1
scal_offset = 1

!$ in_offset = 1 + plan%num_rows * my_id * plan%n
!$ out_offset = 1 + plan%num_rows * my_id * plan%n
!$ IF ( plan%fsign == +1 .AND. plan%trans ) THEN
!$  in_offset = 1 + plan%num_rows*my_id
!$ ELSEIF ( plan%fsign == -1 .AND. plan%trans ) THEN
!$  out_offset = 1 + plan%num_rows*my_id
!$ ENDIF
!$ scal_offset = 1 + plan%n*plan%num_rows*my_id
!$ IF ( plan%need_alt_plan .AND. my_id .EQ. plan%num_threads_needed - 1 ) THEN
!$   num_rows = plan%alt_num_rows
!$   fftw_plan = plan%alt_fftw_plan
!$ ELSE
!$   num_rows = plan%num_rows
!$ ENDIF

zin_ptr => zin(in_offset)
zout_ptr => zout(out_offset)
zscal_ptr => zout(scal_offset)

  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
  IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(scale))==-1) EXIT ;  END DO ; ENDIF
  !MARK_USED does not work with assumed size arguments
  IF(.FALSE.)THEN; DO; IF(ABS(zin(1))>ABS(zout(1)))EXIT; ENDDO; ENDIF
  stat=0

!$ else
!$ end if


!$OMP END PARALLEL

END SUBROUTINE fftw31dm

!     Copyright (c) 2003, 2006 Matteo Frigo
!     Copyright (c) 2003, 2006 Massachusetts Institute of Technology
!
!     This program is free software; you can redistribute it and/or modify
!     it under the terms of the GNU General Public License as published by
!     the Free Software Foundation; either version 2 of the License, or
!     (at your option) any later version.
!
!     This program is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with this program; if not, write to the Free Software
!     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!
!     This is an example implementation of Fortran wisdom export/import
!     to/from a Fortran unit (file), exploiting the generic
!     dfftw_export_wisdom/dfftw_import_wisdom functions.
!
!     We cannot compile this file into the FFTW library itself, lest all
!     FFTW-calling programs be required to link to the Fortran I/O
!     libraries.
!
!     adapted to become more standard Fortran 90 [2007-10] Joost VandeVondele
!     and added some namespacing
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

! **************************************************************************************************
!> \brief ...
!> \param c ...
!> \param iunit ...
! **************************************************************************************************
      SUBROUTINE fftw_write_char(c, iunit) BIND(C, name="fftw_write_char")
      CHARACTER(KIND=C_CHAR)                             :: c
      INTEGER(KIND=C_INT)                                :: iunit

         WRITE(iunit,'(a)',ADVANCE="NO") c
      END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param iunit ...
! **************************************************************************************************
      SUBROUTINE fftw_export_wisdom_to_file(iunit)
    INTEGER                                  :: iunit

   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(iunit))==-1) EXIT ;  END DO ; ENDIF
      END SUBROUTINE

!     Fortran 77 does not have any portable way to read an arbitrary
!     file one character at a time [needs to wait for stream IO of F2003].
!     The best alternative seems to be to
!     read a whole line into a buffer, since for fftw-exported wisdom we
!     can bound the line length.  (If the file contains longer lines,
!     then the lines will be truncated and the wisdom import should
!     simply fail.)  Ugh (and not thread safe).

! **************************************************************************************************
!> \brief ...
!> \param ic ...
!> \param iunit ...
! **************************************************************************************************
      SUBROUTINE fftw_read_char(ic, iunit) BIND(C, name="fftw_read_char")
      INTEGER(KIND=C_INT)                                :: ic, iunit

      CHARACTER(LEN=256)                                 :: buf

         SAVE buf
         INTEGER ibuf
         DATA ibuf/257/
         SAVE ibuf
         IF (ibuf .LT. 257) THEN
            ic = ICHAR(buf(ibuf:ibuf))
            ibuf = ibuf + 1
            RETURN
         ENDIF
         READ(iunit,123,END=666) buf
         ic = ICHAR(buf(1:1))
         ibuf = 2
         RETURN
 666     ic = -1
         ibuf = 257
 123     FORMAT(a256)
      END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param isuccess ...
!> \param iunit ...
! **************************************************************************************************
      SUBROUTINE fftw_import_wisdom_from_file(isuccess, iunit)
      INTEGER                                            :: isuccess, iunit

         isuccess=0
   IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(iunit))==-1) EXIT ;  END DO ; ENDIF
      END SUBROUTINE

END MODULE

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE fft_lib

   USE fft_kinds,                       ONLY: dp
   USE fft_plan,                        ONLY: fft_plan_type
   USE fftsg_lib,                       ONLY: fftsg1dm,&
                                              fftsg3d,&
                                              fftsg_do_cleanup,&
                                              fftsg_do_init,&
                                              fftsg_get_lengths
   USE fftw3_lib,                       ONLY: fftw31dm,&
                                              fftw33d,&
                                              fftw3_create_plan_1dm,&
                                              fftw3_create_plan_3d,&
                                              fftw3_destroy_plan,&
                                              fftw3_do_cleanup,&
                                              fftw3_do_init,&
                                              fftw3_get_lengths
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fft_lib'

   PUBLIC :: fft_do_cleanup, fft_do_init, fft_get_lengths, fft_create_plan_3d
   PUBLIC :: fft_create_plan_1dm, fft_1dm, fft_library, fft_3d, fft_destroy_plan

CONTAINS
! **************************************************************************************************
!> \brief Interface to FFT libraries
!> \param fftlib ...
!> \retval flib ...
!> \par History
!>      IAB 09-Jan-2009 : Modified to use fft_plan_type
!>                        (c) The Numerical Algorithms Group (NAG) Ltd, 2009 on behalf of the HECToR project
!> \author JGH
! **************************************************************************************************
   FUNCTION fft_library(fftlib) RESULT(flib)

      CHARACTER(len=*), INTENT(IN)                       :: fftlib
      INTEGER                                            :: flib

      SELECT CASE (fftlib)
      CASE DEFAULT
         flib = -1
      CASE ("FFTSG")
         flib = 1
      CASE ("FFTW3")
         flib = 3
      END SELECT

   END FUNCTION fft_library

! **************************************************************************************************
!> \brief ...
!> \param fft_type ...
!> \param wisdom_file ...
! **************************************************************************************************
   SUBROUTINE fft_do_init(fft_type, wisdom_file)
      INTEGER, INTENT(IN)                                :: fft_type
      CHARACTER(LEN=*), INTENT(IN)                       :: wisdom_file

      SELECT CASE (fft_type)
      CASE DEFAULT
         CALL cp__b("pw/fft/fft_lib.F",69,"fft_do_init")
      CASE (1)
         CALL fftsg_do_init()
      CASE (3)
         CALL fftw3_do_init(wisdom_file)
      END SELECT

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param fft_type ...
!> \param wisdom_file ...
!> \param ionode ...
! **************************************************************************************************
   SUBROUTINE fft_do_cleanup(fft_type, wisdom_file, ionode)
      INTEGER, INTENT(IN)                                :: fft_type
      CHARACTER(LEN=*), INTENT(IN)                       :: wisdom_file
      LOGICAL, INTENT(IN)                                :: ionode

      SELECT CASE (fft_type)
      CASE DEFAULT
         CALL cp__b("pw/fft/fft_lib.F",91,"fft_do_cleanup")
      CASE (1)
         CALL fftsg_do_cleanup()
      CASE (3)
         CALL fftw3_do_cleanup(wisdom_file, ionode)
      END SELECT

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param fft_type ...
!> \param DATA ...
!> \param max_length ...
! **************************************************************************************************
   SUBROUTINE fft_get_lengths(fft_type, DATA, max_length)

      INTEGER, INTENT(IN)                                :: fft_type
      INTEGER, DIMENSION(*)                              :: DATA
      INTEGER, INTENT(INOUT)                             :: max_length

      SELECT CASE (fft_type)
      CASE DEFAULT
         CALL cp__b("pw/fft/fft_lib.F",114,"fft_get_lengths")
      CASE (1)
         CALL fftsg_get_lengths(DATA, max_length)
      CASE (3)
         CALL fftw3_get_lengths(DATA, max_length)
      END SELECT

   END SUBROUTINE fft_get_lengths

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param fft_type ...
!> \param fft_in_place ...
!> \param fsign ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param plan_style ...
! **************************************************************************************************
   SUBROUTINE fft_create_plan_3d(plan, fft_type, fft_in_place, fsign, n, zin, zout, plan_style)

      TYPE(fft_plan_type), INTENT(INOUT)                 :: plan
      INTEGER, INTENT(IN)                                :: fft_type
      LOGICAL, INTENT(IN)                                :: fft_in_place
      INTEGER, INTENT(IN)                                :: fsign
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT)      :: zin, zout
      INTEGER, INTENT(IN)                                :: plan_style

      plan%fft_type = fft_type
      plan%fsign = fsign
      plan%fft_in_place = fft_in_place
      plan%n_3d = n
!$    plan%need_alt_plan = .FALSE.

      ! Planning only needed for FFTW3
      IF (fft_type .EQ. 3) THEN
         CALL fftw3_create_plan_3d(plan, zin, zout, plan_style)
         plan%valid = .TRUE.
      END IF

   END SUBROUTINE fft_create_plan_3d

!
! really ugly, plan is intent out, because plan%fsign is also a status flag
! if something goes wrong, plan%fsign is set to zero, and the plan becomes invalid
!
! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param scale ...
!> \param zin ...
!> \param zout ...
!> \param stat ...
! **************************************************************************************************
   SUBROUTINE fft_3d(plan, scale, zin, zout, stat)

  USE fft_plan,                           ONLY: fft_plan_type
  USE fft_kinds,                          ONLY: dp
      TYPE(fft_plan_type), INTENT(IN)                    :: plan
      REAL(KIND=dp), INTENT(IN)                          :: scale
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT)      :: zin, zout
      INTEGER, INTENT(OUT)                               :: stat

      stat = plan%fsign
      IF (plan%n_3d(1)*plan%n_3d(2)*plan%n_3d(3) > 0) THEN
         SELECT CASE (plan%fft_type)
         CASE DEFAULT
            CALL cp__b("pw/fft/fft_lib.F",185,"fft_3d")
         CASE (1)
            CALL fftsg3d(plan%fft_in_place, stat, scale, plan%n_3d, zin, zout)
         CASE (3)
            CALL fftw33d(plan, scale, zin, zout, stat)
         END SELECT
      END IF
      ! stat is set to zero on error, -1,+1 are OK
      IF (stat .EQ. 0) THEN
         stat = 1
      ELSE
         stat = 0
      ENDIF

   END SUBROUTINE fft_3d

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param fft_type ...
!> \param fsign ...
!> \param trans ...
!> \param n ...
!> \param m ...
!> \param zin ...
!> \param zout ...
!> \param plan_style ...
! **************************************************************************************************
   SUBROUTINE fft_create_plan_1dm(plan, fft_type, fsign, trans, n, m, zin, zout, plan_style)

  USE fft_plan,                           ONLY: fft_plan_type
  USE fft_kinds,                          ONLY: dp
      TYPE(fft_plan_type), INTENT(INOUT)                 :: plan
      INTEGER, INTENT(IN)                                :: fft_type, fsign
      LOGICAL, INTENT(IN)                                :: trans
      INTEGER, INTENT(IN)                                :: n, m
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(IN)         :: zin, zout
      INTEGER, INTENT(IN)                                :: plan_style

      plan%fft_type = fft_type
      plan%fsign = fsign
      plan%trans = trans
      plan%n = n
      plan%m = m
!$    plan%need_alt_plan = .FALSE.

      ! Planning only needed for FFTW3
      IF ((fft_type .EQ. 3) .AND. (n*m .NE. 0)) THEN
         CALL fftw3_create_plan_1dm(plan, zin, zout, plan_style)
         plan%valid = .TRUE.
      ELSE
         plan%valid = .FALSE.
      END IF

   END SUBROUTINE fft_create_plan_1dm

! **************************************************************************************************
!> \brief ...
!> \param plan ...
! **************************************************************************************************
   SUBROUTINE fft_destroy_plan(plan)

  USE fft_plan,                           ONLY: fft_plan_type
  USE fft_kinds,                          ONLY: dp
      TYPE(fft_plan_type), INTENT(INOUT)                 :: plan

! Planning only needed for FFTW3

      IF (plan%valid .AND. plan%fft_type .EQ. 3) THEN
         CALL fftw3_destroy_plan(plan)
         plan%valid = .FALSE.
      END IF

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param zin ...
!> \param zout ...
!> \param scale ...
!> \param stat ...
! **************************************************************************************************
   SUBROUTINE fft_1dm(plan, zin, zout, scale, stat)

      TYPE(fft_plan_type), INTENT(IN)                    :: plan
      COMPLEX(KIND=dp), DIMENSION(*), INTENT(INOUT)      :: zin, zout
      REAL(KIND=dp), INTENT(IN)                          :: scale
      INTEGER, INTENT(OUT)                               :: stat

      stat = plan%fsign
      IF (plan%n*plan%m > 0) THEN
         SELECT CASE (plan%fft_type)
         CASE DEFAULT
            CALL cp__b("pw/fft/fft_lib.F",281,"fft_1dm")
         CASE (1)
            CALL fftsg1dm(stat, plan%trans, plan%n, plan%m, zin, zout, scale)
         CASE (3)
            CALL fftw31dm(plan, zin, zout, scale, stat)
         END SELECT
      END IF
      ! stat is set to zero on error, -1,+1 are OK
      IF (stat .EQ. 0) THEN
         stat = 1
      ELSE
         stat = 0
      ENDIF

   END SUBROUTINE fft_1dm

END MODULE

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE fast

   USE kinds,                           ONLY: dp

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: rankup, vc_x_vc, vr_x_vc, copy_cri, copy_cr, copy_rc, zero_c

   INTERFACE zero_c
      MODULE PROCEDURE zero_c2, zero_c3
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param za ...
!> \param cmat ...
!> \param zb ...
!> \param ex ...
!> \param ey ...
!> \param ez ...
!> \param scr ...
! **************************************************************************************************
   SUBROUTINE rankup(n, za, cmat, zb, ex, ey, ez, scr)
!
! cmat(i,j,k) <- za * cmat(i,j,k) + ex(i) * ey(j) * ez(k)
!

      INTEGER, DIMENSION(3), INTENT(IN)                  :: n
      COMPLEX(KIND=dp), INTENT(IN)                       :: za
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         INTENT(INOUT)                                   :: cmat
      COMPLEX(KIND=dp), INTENT(IN)                       :: zb
      COMPLEX(KIND=dp), DIMENSION(:), INTENT(IN)         :: ex, ey, ez
      COMPLEX(KIND=dp), DIMENSION(:), INTENT(INOUT)      :: scr

      COMPLEX(KIND=dp)                                   :: cone
      INTEGER                                            :: n2, n3

      cone = CMPLX(1._dp, KIND=dp)
      n2 = n(1)*n(2)
      n3 = n2*n(3)
      scr(1:n2) = CMPLX(0._dp, KIND=dp)
      CALL zgeru(n(1), n(2), zb, ex, 1, ey, 1, scr, n(1))
      CALL zscal(n3, za, cmat, 1)
      CALL zgeru(n2, n(3), cone, scr, 1, ez, 1, cmat, n2)

   END SUBROUTINE rankup

! **************************************************************************************************
!> \brief Multiply two complex 3D arrays element-wise
!> \param cvec2 the other complex array
!> \param cvec the complex array returned as the result
! **************************************************************************************************
   SUBROUTINE vc_x_vc(cvec2, cvec)
!
! cvec(i) <- cvec(i) * cvec2(i)
!

      COMPLEX(KIND=dp), INTENT(IN)                       :: cvec2(:, :, :)
      COMPLEX(KIND=dp), INTENT(INOUT)                    :: cvec(:, :, :)

!$OMP PARALLEL WORKSHARE DEFAULT(NONE), SHARED(cvec,cvec2)
      cvec(:, :, :) = cvec(:, :, :)*cvec2(:, :, :)
!$OMP END PARALLEL WORKSHARE

   END SUBROUTINE vc_x_vc

! **************************************************************************************************
!> \brief Scale a complex array element-wise by a real array
!> \param rvec the real array
!> \param cvec the complex array
! **************************************************************************************************
   SUBROUTINE vr_x_vc(rvec, cvec)
!
! cvec(i) <- cvec(i) * rvec(i)
!

      REAL(KIND=dp), INTENT(IN)                          :: rvec(:, :, :)
      COMPLEX(KIND=dp), INTENT(INOUT)                    :: cvec(:, :, :)

!$OMP PARALLEL WORKSHARE DEFAULT(NONE), SHARED(cvec,rvec)
      cvec(:, :, :) = cvec(:, :, :)*rvec(:, :, :)
!$OMP END PARALLEL WORKSHARE

   END SUBROUTINE vr_x_vc

! **************************************************************************************************
!> \brief Copy a the real and imag. parts of a complex 3D array into two real arrays
!> \param z the complex array
!> \param r1 the real array for the real part
!> \param r2 the real array for the imaginary part
! **************************************************************************************************
   SUBROUTINE copy_cri(z, r1, r2)
!
! r1 = real ( z )
! r2 = imag ( z )
!

      COMPLEX(KIND=dp), INTENT(IN)                       :: z(:, :, :)
      REAL(KIND=dp), INTENT(INOUT)                       :: r1(:, :, :), r2(:, :, :)

!$OMP PARALLEL WORKSHARE DEFAULT(NONE), SHARED(r1,r2,z)
      r1(:, :, :) = REAL(z(:, :, :), KIND=dp)
      r2(:, :, :) = AIMAG(z(:, :, :))
!$OMP END PARALLEL WORKSHARE

   END SUBROUTINE copy_cri

! **************************************************************************************************
!> \brief Copy the real part of a complex 3D array to a real array
!> \param z the complex array
!> \param r1 the real array
! **************************************************************************************************
   SUBROUTINE copy_cr(z, r1)
!
! r1 = real ( z )
!

      COMPLEX(KIND=dp), INTENT(IN)                       :: z(:, :, :)
      REAL(KIND=dp), INTENT(INOUT)                       :: r1(:, :, :)

!$OMP PARALLEL WORKSHARE DEFAULT(NONE), SHARED(r1,z)
      r1(:, :, :) = REAL(z(:, :, :), KIND=dp)
!$OMP END PARALLEL WORKSHARE

   END SUBROUTINE copy_cr

! **************************************************************************************************
!> \brief Copy a real 3D array to complex 3D array
!> \param r1 the real array
!> \param z the complex array
! **************************************************************************************************
   SUBROUTINE copy_rc(r1, z)
!
! z = r1
!

      REAL(KIND=dp), INTENT(IN)                          :: r1(:, :, :)
      COMPLEX(KIND=dp), INTENT(INOUT)                    :: z(:, :, :)

!$OMP PARALLEL WORKSHARE DEFAULT(NONE), SHARED(r1,z)
      z(:, :, :) = CMPLX(r1(:, :, :), 0.0_dp, KIND=dp)
!$OMP END PARALLEL WORKSHARE

   END SUBROUTINE copy_rc

! **************************************************************************************************
!> \brief Zero a complex 2D array (optionally with OpenMP)
!> \param z the array
! **************************************************************************************************
   SUBROUTINE zero_c2(z)
!
! z = ( 0.0_dp , 0.0_dp)
!

      COMPLEX(KIND=dp), INTENT(INOUT)                    :: z(:, :)

!$OMP PARALLEL WORKSHARE DEFAULT(NONE), SHARED(z)
      z(:, :) = 0.0_dp
!$OMP END PARALLEL WORKSHARE
   END SUBROUTINE zero_c2

! **************************************************************************************************
!> \brief Zero a complex 3D array (optionally with OpenMP)
!> \param z the array
! **************************************************************************************************
   SUBROUTINE zero_c3(z)
!
! z = ( 0.0_dp , 0.0_dp)
!

      COMPLEX(KIND=dp), INTENT(INOUT)                    :: z(:, :, :)

!$OMP PARALLEL WORKSHARE DEFAULT(NONE), SHARED(z)
      z(:, :, :) = 0.0_dp
!$OMP END PARALLEL WORKSHARE

   END SUBROUTINE zero_c3

END MODULE fast
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (30-Nov-2000): ESSL FFT Library added
!>      JGH (05-Jan-2001): Added SGI library FFT
!>      JGH (14-Jan-2001): Added parallel 3d FFT
!>      JGH (10-Feb-2006): New interface type
!>      JGH (31-Mar-2008): Remove local allocates and reshapes (performance)
!>                         Possible problems can be related with setting arrays
!>                         not to zero
!>                         Some interfaces could be further simplified by avoiding
!>                         an initial copy. However, this assumes contiguous arrays
!>      IAB (15-Oct-2008): Moved mp_cart_sub calls out of cube_tranpose_* and into
!>                         fft_scratch type, reducing number of calls dramatically
!>      IAB (05-Dec-2008): Moved all other non-essential MPI calls into scratch type
!>      IAB (09-Jan-2009): Added fft_plan_type to store FFT data, including cached FFTW plans
!>      IAB (13-Feb-2009): Extended plan caching to serial 3D FFT (fft3d_s)
!>      IAB (09-Oct-2009): Added OpenMP directives to parallel 3D FFT
!>                         (c) The Numerical Algorithms Group (NAG) Ltd, 2008-2009 on behalf of the HECToR project
!> \author JGH
! **************************************************************************************************
MODULE fft_tools
   USE ISO_C_BINDING,                   ONLY: C_F_POINTER,&
                                              C_INT,&
                                              C_LOC,&
                                              C_PTR,&
                                              C_SIZE_T
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE fast,                            ONLY: zero_c
   USE fft_lib,                         ONLY: &
        fft_1dm, fft_3d, fft_create_plan_1dm, fft_create_plan_3d, fft_destroy_plan, &
        fft_do_cleanup, fft_do_init, fft_get_lengths, fft_library
   USE fft_plan,                        ONLY: fft_plan_type
   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size,&
                                              sp
   USE message_passing,                 ONLY: &
        mp_alltoall, mp_cart_coords, mp_cart_rank, mp_cart_sub, mp_comm_compare, mp_comm_free, &
        mp_comm_null, mp_environ, mp_irecv, mp_isend, mp_rank_compare, mp_sum, mp_sync, mp_waitall
   USE termination,                     ONLY: stop_memory

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fft_tools'

   ! Types for the pool of scratch data needed in FFT routines
   ! keep the subroutine "is_equal" up-to-date
   ! needs a default initialization
   TYPE fft_scratch_sizes
      INTEGER                              :: nx = 0, ny = 0, nz = 0
      INTEGER                              :: lmax = 0, mmax = 0, nmax = 0
      INTEGER                              :: mx1 = 0, mx2 = 0, mx3 = 0
      INTEGER                              :: my1 = 0, my2 = 0, my3 = 0
      INTEGER                              :: mz1 = 0, mz2 = 0, mz3 = 0
      INTEGER                              :: mcz1 = 0, mcz2 = 0, mcy3 = 0, mcx2 = 0
      INTEGER                              :: lg = 0, mg = 0
      INTEGER                              :: nbx = 0, nbz = 0
      INTEGER                              :: nmray = 0, nyzray = 0
      INTEGER                              :: gs_group = 0, rs_group = 0
      INTEGER, DIMENSION(2)                :: g_pos = 0, r_pos = 0, r_dim = 0
      INTEGER                              :: numtask = 0
   END TYPE fft_scratch_sizes

   TYPE fft_scratch_type
      INTEGER                              :: fft_scratch_id
      INTEGER                              :: tf_type
      LOGICAL                              :: in_use
      INTEGER                              :: group
      INTEGER, DIMENSION(3)                :: nfft
      ! to be used in cube_transpose_* routines
      INTEGER, DIMENSION(2)                :: cart_sub_comm = mp_comm_null, dim, pos
      ! to be used in fft3d_s
      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER &
         :: ziptr, zoptr
      ! to be used in fft3d_ps : block distribution
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER &
         :: p1buf, p2buf, p3buf, p4buf, p5buf, p6buf, p7buf
      ! to be used in fft3d_ps : plane distribution
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER &
         :: r1buf, r2buf
      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER &
         :: tbuf
      ! to be used in fft3d_pb
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER &
         :: a1buf, a2buf, a3buf, a4buf, a5buf, a6buf
      ! to be used in communication routines
      INTEGER, DIMENSION(:), POINTER       :: scount, rcount, sdispl, rdispl
      INTEGER, DIMENSION(:, :), POINTER     :: pgcube
      INTEGER, DIMENSION(:), POINTER       :: xzcount, yzcount, xzdispl, yzdispl
      INTEGER                              :: in, mip
      REAL(KIND=dp)                        :: rsratio
      COMPLEX(KIND=dp), DIMENSION(:), POINTER &
         :: xzbuf, yzbuf
      COMPLEX(KIND=sp), DIMENSION(:), POINTER &
         :: xzbuf_sgl, yzbuf_sgl
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER &
         :: rbuf1, rbuf2, rbuf3, rbuf4, rbuf5, rbuf6, rr
      COMPLEX(KIND=sp), DIMENSION(:, :), POINTER &
         :: ss, tt
      INTEGER, DIMENSION(:, :), POINTER     :: pgrid
      INTEGER, DIMENSION(:), POINTER       :: xcor, zcor, pzcoord
      TYPE(fft_scratch_sizes)              :: sizes
      TYPE(fft_plan_type), DIMENSION(6)   :: fft_plan
      INTEGER                              :: last_tick
   END TYPE fft_scratch_type

   TYPE fft_scratch_pool_type
      TYPE(fft_scratch_type), POINTER       :: fft_scratch
      TYPE(fft_scratch_pool_type), POINTER  :: fft_scratch_next
   END TYPE fft_scratch_pool_type

   INTEGER, SAVE                           :: init_fft_pool = 0
   ! the clock for fft pool. Allows to identify the least recently used scratch
   INTEGER, SAVE                           :: tick_fft_pool = 0
   ! limit the number of scratch pools to fft_pool_scratch_limit.
   INTEGER, SAVE                           :: fft_pool_scratch_limit = 15
   TYPE(fft_scratch_pool_type), POINTER, SAVE:: fft_scratch_first
   ! END of types for the pool of scratch data needed in FFT routines

   PRIVATE
   PUBLIC :: init_fft, fft3d, finalize_fft
   PUBLIC :: fft_radix_operations
   PUBLIC :: FWFFT, BWFFT
   PUBLIC :: FFT_RADIX_CLOSEST, FFT_RADIX_NEXT
   PUBLIC :: FFT_RADIX_NEXT_ODD

   INTEGER, PARAMETER :: FWFFT = +1, BWFFT = -1
   INTEGER, PARAMETER :: FFT_RADIX_CLOSEST = 493, FFT_RADIX_NEXT = 494
   INTEGER, PARAMETER :: FFT_RADIX_ALLOWED = 495, FFT_RADIX_DISALLOWED = 496
   INTEGER, PARAMETER :: FFT_RADIX_NEXT_ODD = 497

   REAL(KIND=dp), PARAMETER :: ratio_sparse_alltoall = 0.5_dp

   ! these saved variables are FFT globals
   INTEGER, SAVE :: fft_type = 0
   LOGICAL, SAVE :: alltoall_sgl = .FALSE.
   LOGICAL, SAVE :: use_fftsg_sizes = .TRUE.
   INTEGER, SAVE :: fft_plan_style = 1

   ! these are only needed for pw_methods_cuda (-D__PW_CUDA)
   PUBLIC :: get_fft_scratch, release_fft_scratch
   PUBLIC :: cube_transpose_1, cube_transpose_2
   PUBLIC :: yz_to_x, x_to_yz, xz_to_yz, yz_to_xz
   PUBLIC :: fft_scratch_sizes, fft_scratch_type

   INTERFACE fft3d
      MODULE PROCEDURE fft3d_s, fft3d_ps, fft3d_pb
   END INTERFACE


! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param fftlib ...
!> \param alltoall ...
!> \param fftsg_sizes ...
!> \param pool_limit ...
!> \param wisdom_file ...
!> \param plan_style ...
!> \author JGH
! **************************************************************************************************
   SUBROUTINE init_fft(fftlib, alltoall, fftsg_sizes, pool_limit, wisdom_file, &
                       plan_style)

      CHARACTER(LEN=*), INTENT(IN)                       :: fftlib
      LOGICAL, INTENT(IN)                                :: alltoall, fftsg_sizes
      INTEGER, INTENT(IN)                                :: pool_limit
      CHARACTER(LEN=*), INTENT(IN)                       :: wisdom_file
      INTEGER, INTENT(IN)                                :: plan_style

      CHARACTER(len=*), PARAMETER :: routineN = 'init_fft', routineP = moduleN//':'//routineN

      use_fftsg_sizes = fftsg_sizes
      alltoall_sgl = alltoall
      fft_pool_scratch_limit = pool_limit
      fft_type = fft_library(fftlib)
      fft_plan_style = plan_style

      IF (fft_type <= 0) CALL cp__b("pw/fft_tools.F",211,"Unknown FFT library: "//TRIM(fftlib))

      CALL fft_do_init(fft_type, wisdom_file)

      ! setup the FFT scratch pool, if one is associated, clear first
      CALL release_fft_scratch_pool()
      CALL init_fft_scratch_pool()

   END SUBROUTINE init_fft

! **************************************************************************************************
!> \brief does whatever is needed to finalize the current fft setup
!> \param para_env ...
!> \param wisdom_file ...
!> \par History
!>      10.2007 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE finalize_fft(para_env, wisdom_file)
      TYPE(cp_para_env_type), POINTER                    :: para_env
      CHARACTER(LEN=*), INTENT(IN)                       :: wisdom_file

      CHARACTER(len=*), PARAMETER :: routineN = 'finalize_fft', routineP = moduleN//':'//routineN

! release the FFT scratch pool

      CALL release_fft_scratch_pool()

      ! finalize fft libs

      CALL fft_do_cleanup(fft_type, wisdom_file, para_env%ionode)

   END SUBROUTINE finalize_fft

! **************************************************************************************************
!> \brief Determine the allowed lengths of FFT's   '''
!> \param radix_in ...
!> \param radix_out ...
!> \param operation ...
!> \par History
!>      new library structure (JGH)
!> \author Ari Seitsonen
! **************************************************************************************************
   SUBROUTINE fft_radix_operations(radix_in, radix_out, operation)

      INTEGER, INTENT(IN)                                :: radix_in
      INTEGER, INTENT(OUT)                               :: radix_out
      INTEGER, INTENT(IN)                                :: operation

      CHARACTER(len=*), PARAMETER :: routineN = 'fft_radix_operations', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: fft_type_sg = 1

      INTEGER                                            :: i, iloc, ldata
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: DATA

      ldata = 1024
      ALLOCATE (DATA(ldata))
      DATA = -1

      ! if the user wants to use fftsg sizes go for it
      IF (use_fftsg_sizes) THEN
         CALL fft_get_lengths(fft_type_sg, DATA, ldata)
      ELSE
         CALL fft_get_lengths(fft_type, DATA, ldata)
      ENDIF

      iloc = 0
      DO i = 1, ldata
         IF (DATA(i) == radix_in) THEN
            iloc = i
            EXIT
         ELSE
            IF (OPERATION == FFT_RADIX_ALLOWED) THEN
               CYCLE
            ELSE IF (DATA(i) > radix_in) THEN
               iloc = i
               EXIT
            END IF
         END IF
      END DO

      IF (iloc == 0) THEN
         CALL cp__b("pw/fft_tools.F",293,"Index to radix array not found.")
      END IF

      IF (OPERATION == FFT_RADIX_ALLOWED) THEN
         IF (DATA(iloc) == radix_in) THEN
            radix_out = FFT_RADIX_ALLOWED
         ELSE
            radix_out = FFT_RADIX_DISALLOWED
         END IF

      ELSE IF (OPERATION == FFT_RADIX_CLOSEST) THEN
         IF (DATA(iloc) == radix_in) THEN
            radix_out = DATA(iloc)
         ELSE
            IF (ABS(DATA(iloc-1)-radix_in) <= &
                ABS(DATA(iloc)-radix_in)) THEN
               radix_out = DATA(iloc-1)
            ELSE
               radix_out = DATA(iloc)
            END IF
         END IF

      ELSE IF (OPERATION == FFT_RADIX_NEXT) THEN
         radix_out = DATA(iloc)

      ELSE IF (OPERATION == FFT_RADIX_NEXT_ODD) THEN
         DO i = iloc, ldata
            IF (MOD(DATA(i), 2) == 1) THEN
               radix_out = DATA(i)
               EXIT
            END IF
         END DO
         IF (MOD(radix_out, 2) == 0) THEN
            CALL cp__b("pw/fft_tools.F",326,"No odd radix found.")
         END IF

      ELSE
         CALL cp__b("pw/fft_tools.F",330,"Disallowed radix operation.")
      END IF

      DEALLOCATE (DATA)

   END SUBROUTINE fft_radix_operations

! **************************************************************************************************
!> \brief Calls the 3D-FFT function from the initialized library
!> \param fsign ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param scale ...
!> \param status ...
!> \param debug ...
!> \par History
!>      none
!> \author JGH
! **************************************************************************************************
   SUBROUTINE fft3d_s(fsign, n, zin, zout, scale, status, debug)

      INTEGER, INTENT(IN)                                :: fsign
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: n
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         INTENT(INOUT)                                   :: zin
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         INTENT(INOUT), OPTIONAL, TARGET                 :: zout
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: scale
      INTEGER, INTENT(OUT), OPTIONAL                     :: status
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug

      CHARACTER(len=*), PARAMETER :: routineN = 'fft3d_s', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(1, 1, 1), TARGET       :: zdum
      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER      :: zoptr
      INTEGER                                            :: handle, ld(3), lo(3), output_unit, sign, &
                                                            stat
      LOGICAL                                            :: fft_in_place, test
      REAL(KIND=dp)                                      :: in_sum, norm, out_sum
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit()

      IF (PRESENT(scale)) THEN
         norm = scale
      ELSE
         norm = 1.0_dp
      END IF

      IF (PRESENT(debug)) THEN
         test = debug
      ELSE
         test = .FALSE.
      END IF

      IF (PRESENT(zout)) THEN
         fft_in_place = .FALSE.
      ELSE
         fft_in_place = .TRUE.
      END IF

      IF (test) THEN
         in_sum = SUM(ABS(zin))
      ENDIF

      ld(1) = SIZE(zin, 1)
      ld(2) = SIZE(zin, 2)
      ld(3) = SIZE(zin, 3)

      IF (n(1) /= ld(1) .OR. n(2) /= ld(2) .OR. n(3) /= ld(3)) THEN
         CALL cp__b("pw/fft_tools.F",402,"Size and dimension (zin) have to be the same.")
      END IF

      sign = fsign
      CALL get_fft_scratch(fft_scratch, tf_type=400, n=n)

      IF (fft_in_place) THEN
         zoptr => zdum
         IF (fsign == FWFFT) THEN
            CALL fft_3d(fft_scratch%fft_plan(1), norm, zin, zoptr, stat)
         ELSE
            CALL fft_3d(fft_scratch%fft_plan(2), norm, zin, zoptr, stat)
         ENDIF
      ELSE
         IF (fsign == FWFFT) THEN
            CALL fft_3d(fft_scratch%fft_plan(3), norm, zin, zout, stat)
         ELSE
            CALL fft_3d(fft_scratch%fft_plan(4), norm, zin, zout, stat)
         ENDIF
      END IF

      CALL release_fft_scratch(fft_scratch)

      IF (PRESENT(zout)) THEN
         lo(1) = SIZE(zout, 1)
         lo(2) = SIZE(zout, 2)
         lo(3) = SIZE(zout, 3)
         IF (n(1) /= lo(1) .OR. n(2) /= lo(2) .OR. n(3) /= lo(3)) THEN
            CALL cp__b("pw/fft_tools.F",430,"Size and dimension (zout) have to be the same.")
         END IF
      END IF

      IF (PRESENT(status)) THEN
         status = stat
      END IF

      IF (test .AND. output_unit > 0) THEN
         IF (PRESENT(zout)) THEN
            out_sum = SUM(ABS(zout))
            WRITE (output_unit, '(A)') "  Out of place 3D FFT (local)  : fft3d_s"
            WRITE (output_unit, '(A,T60,3I7)') "     Transform lengths ", n
            WRITE (output_unit, '(A,T60,3I7)') "     Input array dimensions ", ld
            WRITE (output_unit, '(A,T60,3I7)') "     Output array dimensions ", lo
            WRITE (output_unit, '(A,T61,E20.14)') "     Sum of input data ", in_sum
            WRITE (output_unit, '(A,T61,E20.14)') "     Sum of output data ", out_sum
         ELSE
            out_sum = SUM(ABS(zin))
            WRITE (output_unit, '(A)') "  In place 3D FFT (local)  : fft3d_s"
            WRITE (output_unit, '(A,T60,3I7)') "     Transform lengths ", n
            WRITE (output_unit, '(A,T60,3I7)') "     Input/output array dimensions ", ld
            WRITE (output_unit, '(A,T61,E20.14)') "     Sum of input data ", in_sum
            WRITE (output_unit, '(A,T61,E20.14)') "     Sum of output data ", out_sum
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE fft3d_s

! **************************************************************************************************
!> \brief ...
!> \param fsign ...
!> \param n ...
!> \param cin ...
!> \param gin ...
!> \param gs_group ...
!> \param rs_group ...
!> \param yzp ...
!> \param nyzray ...
!> \param bo ...
!> \param scale ...
!> \param status ...
!> \param debug ...
! **************************************************************************************************
   SUBROUTINE fft3d_ps(fsign, n, cin, gin, gs_group, rs_group, yzp, nyzray, &
                       bo, scale, status, debug)

      INTEGER, INTENT(IN)                                :: fsign
      INTEGER, DIMENSION(:), INTENT(IN)                  :: n
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         INTENT(INOUT)                                   :: cin
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(INOUT)   :: gin
      INTEGER, INTENT(IN)                                :: gs_group, rs_group
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: yzp
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nyzray
      INTEGER, DIMENSION(:, :, 0:, :), INTENT(IN)        :: bo
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: scale
      INTEGER, INTENT(OUT), OPTIONAL                     :: status
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug

      CHARACTER(len=*), PARAMETER :: routineN = 'fft3d_ps', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: pbuf, qbuf, rbuf, sbuf
      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER      :: tbuf
      INTEGER :: g_pos, handle, ierr, iout, lg, lmax, mcx2, mcz1, mcz2, mg, mmax, mx1, mx2, my1, &
         mz2, n1, n2, nmax, numtask, numtask_g, numtask_r, nx, ny, nz, output_unit, r_dim(2), &
         r_pos(2), rp, sign, stat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: p2p
      LOGICAL                                            :: test
      REAL(KIND=dp)                                      :: norm, sum_data
      TYPE(fft_scratch_sizes)                            :: fft_scratch_size
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit()

      IF (PRESENT(debug)) THEN
         test = debug
      ELSE
         test = .FALSE.
      END IF

      CALL mp_environ(numtask_g, g_pos, gs_group)
      CALL mp_environ(numtask_r, r_dim, r_pos, rs_group)
      IF (numtask_g /= numtask_r) THEN
         CALL cp__b("pw/fft_tools.F",517,"Real space and G space groups are different.")
      END IF
      numtask = numtask_r
      CALL mp_comm_compare(rs_group, gs_group, iout)
      IF (iout > 3) THEN
         CALL cp__b("pw/fft_tools.F",522,"Real space and G space groups are different.")
      END IF

      IF (PRESENT(scale)) THEN
         norm = scale
      ELSE
         norm = 1.0_dp
      END IF

      sign = fsign

      lg = SIZE(gin, 1)
      mg = SIZE(gin, 2)

      nx = SIZE(cin, 1)
      ny = SIZE(cin, 2)
      nz = SIZE(cin, 3)

      IF (mg == 0) THEN
         mmax = 1
      ELSE
         mmax = mg
      END IF
      lmax = MAX(lg, (nx*ny*nz)/mmax+1)

      ALLOCATE (p2p(0:numtask-1), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 548, &
                                      "p2p", int_size*numtask)

      CALL mp_rank_compare(gs_group, rs_group, p2p)

      rp = p2p(g_pos)
      mx1 = bo(2, 1, rp, 1)-bo(1, 1, rp, 1)+1
      my1 = bo(2, 2, rp, 1)-bo(1, 2, rp, 1)+1
      mx2 = bo(2, 1, rp, 2)-bo(1, 1, rp, 2)+1
      mz2 = bo(2, 3, rp, 2)-bo(1, 3, rp, 2)+1

      n1 = MAXVAL(bo(2, 1, :, 1)-bo(1, 1, :, 1)+1)
      n2 = MAXVAL(bo(2, 2, :, 1)-bo(1, 2, :, 1)+1)
      nmax = MAX((2*n2)/numtask, 2)*mx2*mz2
      nmax = MAX(nmax, n1*MAXVAL(nyzray))
      n1 = MAXVAL(bo(2, 1, :, 2))
      n2 = MAXVAL(bo(2, 3, :, 2))

      fft_scratch_size%nx = nx
      fft_scratch_size%ny = ny
      fft_scratch_size%nz = nz
      fft_scratch_size%lmax = lmax
      fft_scratch_size%mmax = mmax
      fft_scratch_size%mx1 = mx1
      fft_scratch_size%mx2 = mx2
      fft_scratch_size%my1 = my1
      fft_scratch_size%mz2 = mz2
      fft_scratch_size%lg = lg
      fft_scratch_size%mg = mg
      fft_scratch_size%nbx = n1
      fft_scratch_size%nbz = n2
      mcz1 = MAXVAL(bo(2, 3, :, 1)-bo(1, 3, :, 1)+1)
      mcx2 = MAXVAL(bo(2, 1, :, 2)-bo(1, 1, :, 2)+1)
      mcz2 = MAXVAL(bo(2, 3, :, 2)-bo(1, 3, :, 2)+1)
      fft_scratch_size%mcz1 = mcz1
      fft_scratch_size%mcx2 = mcx2
      fft_scratch_size%mcz2 = mcz2
      fft_scratch_size%nmax = nmax
      fft_scratch_size%nmray = MAXVAL(nyzray)
      fft_scratch_size%nyzray = nyzray(g_pos)
      fft_scratch_size%gs_group = gs_group
      fft_scratch_size%rs_group = rs_group
      fft_scratch_size%g_pos = g_pos
      fft_scratch_size%r_pos = r_pos
      fft_scratch_size%r_dim = r_dim
      fft_scratch_size%numtask = numtask

      IF (test) THEN
         IF (g_pos == 0 .AND. output_unit > 0) THEN
            WRITE (output_unit, '(A)') "  Parallel 3D FFT : fft3d_ps"
            WRITE (output_unit, '(A,T60,3I7)') "     Transform lengths ", n
            WRITE (output_unit, '(A,T67,2I7)') "     Array dimensions (gin) ", lg, mg
            WRITE (output_unit, '(A,T60,3I7)') "     Array dimensions (cin) ", nx, ny, nz
         END IF
      END IF

      IF (r_dim(2) > 1) THEN

         !
         ! real space is distributed over x and y coordinate
         ! we have two stages of communication
         !

         IF (r_dim(1) == 1) THEN
            CALL cp__b("pw/fft_tools.F",612,"This processor distribution is not supported.")
         END IF
         CALL get_fft_scratch(fft_scratch, tf_type=300, n=n, fft_sizes=fft_scratch_size)

         IF (sign == FWFFT) THEN
            ! cin -> gin

            IF (test) THEN
               sum_data = ABS(SUM(cin))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A)') "  Two step communication algorithm "
                  WRITE (output_unit, '(A,T60,3I7)') "     Transform Z ", n(3), mx1*my1
                  WRITE (output_unit, '(A,T60,3I7)') "     Transform Y ", n(2), mx2*mz2
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform X ", n(1), nyzray(g_pos)
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(1) ", sum_data
               END IF
            END IF

            pbuf => fft_scratch%p1buf
            qbuf => fft_scratch%p2buf

            ! FFT along z
            CALL fft_1dm(fft_scratch%fft_plan(1), cin, qbuf, norm, stat)

            rbuf => fft_scratch%p3buf

            IF (test) THEN
               sum_data = ABS(SUM(qbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(2) T", sum_data
               END IF
            END IF

            ! Exchange data ( transpose of matrix )
            CALL cube_transpose_2(qbuf, bo(:, :, :, 1), bo(:, :, :, 2), rbuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(rbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(3) T", sum_data
               END IF
            END IF

            pbuf => fft_scratch%p4buf

            ! FFT along y
            CALL fft_1dm(fft_scratch%fft_plan(2), rbuf, pbuf, 1.0_dp, stat)

            qbuf => fft_scratch%p5buf

            IF (test) THEN
               sum_data = ABS(SUM(pbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(4) TS", sum_data
               END IF
            END IF

            ! Exchange data ( transpose of matrix ) and sort
            CALL xz_to_yz(pbuf, rs_group, r_dim, g_pos, p2p, yzp, nyzray, &
                          bo(:, :, :, 2), qbuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(qbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(5) TS", sum_data
               END IF
            END IF

            ! FFT along x
            CALL fft_1dm(fft_scratch%fft_plan(3), qbuf, gin, 1.0_dp, stat)

            IF (test) THEN
               sum_data = ABS(SUM(gin))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(6) ", sum_data
               END IF
            END IF

         ELSE IF (sign == BWFFT) THEN
            ! gin -> cin

            IF (test) THEN
               sum_data = ABS(SUM(gin))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A)') "  Two step communication algorithm "
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform X ", n(1), nyzray(g_pos)
                  WRITE (output_unit, '(A,T60,3I7)') "     Transform Y ", n(2), mx2*mz2
                  WRITE (output_unit, '(A,T60,3I7)') "     Transform Z ", n(3), mx1*my1
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(1) ", sum_data
               END IF
            END IF

            pbuf => fft_scratch%p7buf

            ! FFT along x
            CALL fft_1dm(fft_scratch%fft_plan(4), gin, pbuf, norm, stat)

            qbuf => fft_scratch%p4buf

            IF (test) THEN
               sum_data = ABS(SUM(pbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(2) TS", sum_data
               END IF
            END IF

            ! Exchange data ( transpose of matrix ) and sort
            CALL yz_to_xz(pbuf, rs_group, r_dim, g_pos, p2p, yzp, nyzray, &
                          bo(:, :, :, 2), qbuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(qbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(3) TS", sum_data
               END IF
            END IF

            rbuf => fft_scratch%p3buf

            ! FFT along y
            CALL fft_1dm(fft_scratch%fft_plan(5), qbuf, rbuf, 1.0_dp, stat)

            pbuf => fft_scratch%p2buf

            IF (test) THEN
               sum_data = ABS(SUM(rbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(4) T", sum_data
               END IF
            END IF

            ! Exchange data ( transpose of matrix )
            CALL cube_transpose_1(rbuf, bo(:, :, :, 2), bo(:, :, :, 1), pbuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(pbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(5) T", sum_data
               END IF
            END IF

            qbuf => fft_scratch%p1buf

            ! FFT along z
            CALL fft_1dm(fft_scratch%fft_plan(6), pbuf, cin, 1.0_dp, stat)

            IF (test) THEN
               sum_data = ABS(SUM(cin))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(6) ", sum_data
               END IF
            END IF

         ELSE

            CALL cp__b("pw/fft_tools.F",779,"Illegal fsign parameter.")

         ENDIF

         CALL release_fft_scratch(fft_scratch)

      ELSE

         !
         ! real space is only distributed over x coordinate
         ! we have one stage of communication, after the transform of
         ! direction x
         !

         CALL get_fft_scratch(fft_scratch, tf_type=200, n=n, fft_sizes=fft_scratch_size)

         sbuf => fft_scratch%r1buf
         tbuf => fft_scratch%tbuf

         CALL zero_c(sbuf)
         CALL zero_c(tbuf)

         IF (sign == FWFFT) THEN
            ! cin -> gin

            IF (test) THEN
               sum_data = ABS(SUM(cin))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A)') "     One step communication algorithm "
                  WRITE (output_unit, '(A,T60,3I7)') "     Transform YZ ", n(2), n(3), nx
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform X ", n(1), nyzray(g_pos)
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(1) ", sum_data
               END IF
            END IF

            ! FFT along y and z
            CALL fft_1dm(fft_scratch%fft_plan(1), cin, sbuf, 1._dp, stat)
            CALL fft_1dm(fft_scratch%fft_plan(2), sbuf, tbuf, 1._dp, stat)

            IF (test) THEN
               sum_data = ABS(SUM(tbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(2) TS", sum_data
               END IF
            END IF

            ! Exchange data ( transpose of matrix ) and sort
            CALL yz_to_x(tbuf, gs_group, g_pos, p2p, yzp, nyzray, &
                         bo(:, :, :, 2), sbuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(sbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(3) TS", sum_data
               END IF
            END IF
            ! FFT along x
            CALL fft_1dm(fft_scratch%fft_plan(3), sbuf, gin, norm, stat)

            IF (test) THEN
               sum_data = ABS(SUM(gin))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(4) ", sum_data
               END IF
            END IF

         ELSE IF (sign == BWFFT) THEN
            ! gin -> cin

            IF (test) THEN
               sum_data = ABS(SUM(gin))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A)') "  One step communication algorithm "
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform X ", n(1), nyzray(g_pos)
                  WRITE (output_unit, '(A,T60,3I7)') "     Transform YZ ", n(2), n(3), nx
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(1) ", sum_data
               END IF
            END IF

            ! FFT along x
            CALL fft_1dm(fft_scratch%fft_plan(4), gin, sbuf, norm, stat)

            IF (test) THEN
               sum_data = ABS(SUM(sbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(2) TS", sum_data
               END IF
            END IF

            ! Exchange data ( transpose of matrix ) and sort
            CALL x_to_yz(sbuf, gs_group, g_pos, p2p, yzp, nyzray, &
                         bo(:, :, :, 2), tbuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(tbuf))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(3) TS", sum_data
               END IF
            END IF

            ! FFT along y and z
            CALL fft_1dm(fft_scratch%fft_plan(5), tbuf, sbuf, 1._dp, stat)
            CALL fft_1dm(fft_scratch%fft_plan(6), sbuf, cin, 1._dp, stat)

            IF (test) THEN
               sum_data = ABS(SUM(cin))
               CALL mp_sum(sum_data, gs_group)
               IF (g_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(4) ", sum_data
               END IF
            END IF
         ELSE
            CALL cp__b("pw/fft_tools.F",898,"Illegal fsign parameter.")
         ENDIF

         CALL release_fft_scratch(fft_scratch)

      ENDIF

      DEALLOCATE (p2p)

      IF (PRESENT(status)) THEN
         status = stat
      END IF
      CALL timestop(handle)

   END SUBROUTINE fft3d_ps

! **************************************************************************************************
!> \brief ...
!> \param fsign ...
!> \param n ...
!> \param zin ...
!> \param gin ...
!> \param group ...
!> \param bo ...
!> \param scale ...
!> \param status ...
!> \param debug ...
! **************************************************************************************************
   SUBROUTINE fft3d_pb(fsign, n, zin, gin, group, bo, scale, status, debug)

      INTEGER, INTENT(IN)                                :: fsign
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         INTENT(INOUT)                                   :: zin
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(INOUT)   :: gin
      INTEGER, INTENT(IN)                                :: group
      INTEGER, DIMENSION(:, :, 0:, :), INTENT(IN)        :: bo
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: scale
      INTEGER, INTENT(OUT), OPTIONAL                     :: status
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug

      CHARACTER(len=*), PARAMETER :: routineN = 'fft3d_pb', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: abuf, bbuf
      INTEGER                                            :: handle, lg(2), lz(3), mcx2, mcy3, mcz1, &
                                                            mcz2, mx1, mx2, mx3, my1, my2, my3, &
                                                            my_pos, mz1, mz2, mz3, np, &
                                                            output_unit, sign, stat
      INTEGER, DIMENSION(2)                              :: dim, pos
      LOGICAL                                            :: test
      REAL(KIND=dp)                                      :: norm, sum_data
      TYPE(fft_scratch_sizes)                            :: fft_scratch_size
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

!------------------------------------------------------------------------------
! "Real Space"  1) xyZ      or      1) xYZ
!               2) xYz      or         not used
! "G Space"     3) Xyz      or      3) XYz
!
! There is one communicator (2-dimensional) for all distributions
! np = n1 * n2, where np is the total number of processors
! If n2 = 1, we have the second case and only one transpose step is needed
!
! Assignment of dimensions to axis for different steps
! First case: 1) n1=x; n2=y
!             2) n1=x; n2=z
!             3) n1=y; n2=z
! Second case 1) n1=x
!             3) n1=z
!
! The more general case with two communicators for the initial and final
! distribution is not covered.
!------------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit()

      CALL mp_environ(np, dim, pos, group)
      CALL mp_cart_rank(group, pos, my_pos)

      IF (PRESENT(debug)) THEN
         test = debug
      ELSE
         test = .FALSE.
      END IF

      IF (PRESENT(scale)) THEN
         norm = scale
      ELSE
         norm = 1.0_dp
      END IF

      sign = fsign

      IF (test) THEN
         lg(1) = SIZE(gin, 1)
         lg(2) = SIZE(gin, 2)
         lz(1) = SIZE(zin, 1)
         lz(2) = SIZE(zin, 2)
         lz(3) = SIZE(zin, 3)
         IF (my_pos == 0 .AND. output_unit > 0) THEN
            WRITE (output_unit, '(A)') "  Parallel 3D FFT : fft3d_pb"
            WRITE (output_unit, '(A,T60,3I7)') "     Transform lengths ", n
            WRITE (output_unit, '(A,T67,2I7)') "     Array dimensions (gin) ", lg
            WRITE (output_unit, '(A,T60,3I7)') "     Array dimensions (cin) ", lz
         END IF
      END IF

      mx1 = bo(2, 1, my_pos, 1)-bo(1, 1, my_pos, 1)+1
      my1 = bo(2, 2, my_pos, 1)-bo(1, 2, my_pos, 1)+1
      mz1 = bo(2, 3, my_pos, 1)-bo(1, 3, my_pos, 1)+1
      mx2 = bo(2, 1, my_pos, 2)-bo(1, 1, my_pos, 2)+1
      my2 = bo(2, 2, my_pos, 2)-bo(1, 2, my_pos, 2)+1
      mz2 = bo(2, 3, my_pos, 2)-bo(1, 3, my_pos, 2)+1
      mx3 = bo(2, 1, my_pos, 3)-bo(1, 1, my_pos, 3)+1
      my3 = bo(2, 2, my_pos, 3)-bo(1, 2, my_pos, 3)+1
      mz3 = bo(2, 3, my_pos, 3)-bo(1, 3, my_pos, 3)+1
      fft_scratch_size%mx1 = mx1
      fft_scratch_size%mx2 = mx2
      fft_scratch_size%mx3 = mx3
      fft_scratch_size%my1 = my1
      fft_scratch_size%my2 = my2
      fft_scratch_size%my3 = my3
      fft_scratch_size%mz1 = mz1
      fft_scratch_size%mz2 = mz2
      fft_scratch_size%mz3 = mz3
      mcz1 = MAXVAL(bo(2, 3, :, 1)-bo(1, 3, :, 1)+1)
      mcx2 = MAXVAL(bo(2, 1, :, 2)-bo(1, 1, :, 2)+1)
      mcz2 = MAXVAL(bo(2, 3, :, 2)-bo(1, 3, :, 2)+1)
      mcy3 = MAXVAL(bo(2, 2, :, 3)-bo(1, 2, :, 3)+1)
      fft_scratch_size%mcz1 = mcz1
      fft_scratch_size%mcx2 = mcx2
      fft_scratch_size%mcz2 = mcz2
      fft_scratch_size%mcy3 = mcy3
      fft_scratch_size%gs_group = group
      fft_scratch_size%rs_group = group
      fft_scratch_size%g_pos = my_pos
      fft_scratch_size%numtask = DIM(1)*DIM(2)

      IF (DIM(1) > 1 .AND. DIM(2) > 1) THEN

         !
         ! First case; two stages of communication
         !

         CALL get_fft_scratch(fft_scratch, tf_type=100, n=n, fft_sizes=fft_scratch_size)

         IF (sign == FWFFT) THEN
            ! Stage 1 -> 3

            bbuf => fft_scratch%a2buf

            IF (test) THEN
               sum_data = ABS(SUM(zin))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A)') "  Two step communication algorithm "
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform Z ", n(3), mx1*my1
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(1) ", sum_data
               END IF
            END IF

            ! FFT along z
            CALL fft_1dm(fft_scratch%fft_plan(1), zin, bbuf, norm, stat)

            abuf => fft_scratch%a3buf

            IF (test) THEN
               sum_data = ABS(SUM(bbuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(2) T", sum_data
               END IF
            END IF

            CALL cube_transpose_2(bbuf, bo(:, :, :, 1), bo(:, :, :, 2), abuf, fft_scratch)

            bbuf => fft_scratch%a4buf

            IF (test) THEN
               sum_data = ABS(SUM(abuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform Y ", n(2), mx2*mz2
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(3) ", sum_data
               END IF
            END IF

            ! FFT along y
            CALL fft_1dm(fft_scratch%fft_plan(2), abuf, bbuf, 1.0_dp, stat)

            abuf => fft_scratch%a5buf

            IF (test) THEN
               sum_data = ABS(SUM(bbuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(4) T", sum_data
               END IF
            END IF

            CALL cube_transpose_4(bbuf, bo(:, :, :, 2), bo(:, :, :, 3), abuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(abuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform X ", n(1), my3*mz3
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(5) ", sum_data
               END IF
            END IF

            ! FFT along x
            CALL fft_1dm(fft_scratch%fft_plan(3), abuf, gin, 1.0_dp, stat)

            IF (test) THEN
               sum_data = ABS(SUM(gin))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(6) ", sum_data
               END IF
            END IF

         ELSEIF (sign == BWFFT) THEN
            ! Stage 3 -> 1

            bbuf => fft_scratch%a5buf

            IF (test) THEN
               sum_data = ABS(SUM(gin))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A)') "  Two step communication algorithm "
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform X ", n(1), my3*mz3
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(1) ", sum_data
               END IF
            END IF

            ! FFT along x
            CALL fft_1dm(fft_scratch%fft_plan(4), gin, bbuf, 1.0_dp, stat)

            abuf => fft_scratch%a4buf

            IF (test) THEN
               sum_data = ABS(SUM(bbuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(2) T", sum_data
               END IF
            END IF

            CALL cube_transpose_3(bbuf, bo(:, :, :, 3), bo(:, :, :, 2), abuf, fft_scratch)

            bbuf => fft_scratch%a3buf

            IF (test) THEN
               sum_data = ABS(SUM(abuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform Y ", n(2), mx2*mz2
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(3) ", sum_data
               END IF
            END IF

            ! FFT along y
            CALL fft_1dm(fft_scratch%fft_plan(5), abuf, bbuf, 1.0_dp, stat)

            abuf => fft_scratch%a2buf

            IF (test) THEN
               sum_data = ABS(SUM(bbuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(4) T", sum_data
               END IF
            END IF

            CALL cube_transpose_1(bbuf, bo(:, :, :, 2), bo(:, :, :, 1), abuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(abuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform Z ", n(3), mx1*my1
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(5) ", sum_data
               END IF
            END IF

            ! FFT along z
            CALL fft_1dm(fft_scratch%fft_plan(6), abuf, zin, norm, stat)

            IF (test) THEN
               sum_data = ABS(SUM(zin))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(6) ", sum_data
               END IF
            END IF

         ELSE
            CALL cp__b("pw/fft_tools.F",1198,"Illegal fsign parameter.")
         END IF

         CALL release_fft_scratch(fft_scratch)

      ELSEIF (DIM(2) == 1) THEN

         !
         ! Second case; one stage of communication
         !

         CALL get_fft_scratch(fft_scratch, tf_type=101, n=n, fft_sizes=fft_scratch_size)

         IF (sign == FWFFT) THEN
            ! Stage 1 -> 3

            IF (test) THEN
               sum_data = ABS(SUM(zin))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A)') "  one step communication algorithm "
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform Z ", n(3), mx1*my1
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform Y ", n(2), mx1*mz1
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(1) ", sum_data
               END IF
            END IF

            abuf => fft_scratch%a3buf
            bbuf => fft_scratch%a4buf
            ! FFT along z and y
            CALL fft_1dm(fft_scratch%fft_plan(1), zin, abuf, norm, stat)
            CALL fft_1dm(fft_scratch%fft_plan(2), abuf, bbuf, 1.0_dp, stat)

            abuf => fft_scratch%a5buf

            IF (test) THEN
               sum_data = ABS(SUM(bbuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(2) T", sum_data
               END IF
            END IF

            CALL cube_transpose_6(bbuf, group, bo(:, :, :, 1), bo(:, :, :, 3), abuf, fft_scratch)

            IF (test) THEN
               sum_data = ABS(SUM(abuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform X ", n(1), my3*mz3
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(3) ", sum_data
               END IF
            END IF

            ! FFT along x
            CALL fft_1dm(fft_scratch%fft_plan(3), abuf, gin, 1.0_dp, stat)

            IF (test) THEN
               sum_data = ABS(SUM(gin))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(4) ", sum_data
               END IF
            END IF

         ELSEIF (sign == BWFFT) THEN
            ! Stage 3 -> 1

            IF (test) THEN
               sum_data = ABS(SUM(gin))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A)') "  one step communication algorithm "
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform X ", n(1), my3*mz3
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(1) ", sum_data
               END IF
            END IF

            bbuf => fft_scratch%a5buf

            ! FFT along x
            CALL fft_1dm(fft_scratch%fft_plan(4), gin, bbuf, 1.0_dp, stat)

            abuf => fft_scratch%a4buf

            IF (test) THEN
               sum_data = ABS(SUM(bbuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(2) T", sum_data
               END IF
            END IF

            CALL cube_transpose_5(bbuf, group, bo(:, :, :, 3), bo(:, :, :, 1), abuf, fft_scratch)

            bbuf => fft_scratch%a3buf

            IF (test) THEN
               sum_data = ABS(SUM(abuf))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform Y ", n(2), mx1*mz1
                  WRITE (output_unit, '(A,T67,2I7)') "     Transform Z ", n(3), mx1*my1
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(3) ", sum_data
               END IF
            END IF

            ! FFT along y
            CALL fft_1dm(fft_scratch%fft_plan(5), abuf, bbuf, 1.0_dp, stat)

            ! FFT along z
            CALL fft_1dm(fft_scratch%fft_plan(6), bbuf, zin, norm, stat)

            IF (test) THEN
               sum_data = ABS(SUM(zin))
               CALL mp_sum(sum_data, group)
               IF (my_pos == 0 .AND. output_unit > 0) THEN
                  WRITE (output_unit, '(A,T61,E20.14)') "     Sum of data(4) ", sum_data
               END IF
            END IF

         ELSE
            CALL cp__b("pw/fft_tools.F",1320,"Illegal fsign parameter.")
         ENDIF

         CALL release_fft_scratch(fft_scratch)

      ELSE

         CALL cp__b("pw/fft_tools.F",1327,"This partition not implemented.")

      END IF

      IF (PRESENT(status)) THEN
         status = stat
      END IF

      CALL timestop(handle)

   END SUBROUTINE fft3d_pb

! **************************************************************************************************
!> \brief ...
!> \param sb ...
!> \param group ...
!> \param my_pos ...
!> \param p2p ...
!> \param yzp ...
!> \param nray ...
!> \param bo ...
!> \param tb ...
!> \param fft_scratch ...
!> \par History
!>      15. Feb. 2006 : single precision all_to_all
!> \author JGH (14-Jan-2001)
! **************************************************************************************************
   SUBROUTINE x_to_yz(sb, group, my_pos, p2p, yzp, nray, bo, tb, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: sb
      INTEGER, INTENT(IN)                                :: group, my_pos
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: p2p
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: yzp
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nray
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: bo
      COMPLEX(KIND=dp), DIMENSION(:, :, :), &
         INTENT(INOUT)                                   :: tb
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'x_to_yz', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: rr
      COMPLEX(KIND=sp), DIMENSION(:, :), POINTER         :: ss, tt
      INTEGER                                            :: handle, ip, ir, ix, ixx, iy, iz, mpr, &
                                                            nm, np, nr, nx
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl, scount, sdispl

      CALL timeset(routineN, handle)

      np = SIZE(p2p)
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl

      IF (alltoall_sgl) THEN
         ss => fft_scratch%ss
         tt => fft_scratch%tt
         ss(:, :) = CMPLX(sb(:, :), KIND=sp)
         tt(:, :) = 0._sp
      ELSE
         rr => fft_scratch%rr
      END IF

      mpr = p2p(my_pos)
      nm = MAXVAL(nray(0:np-1))
      nr = nray(my_pos)
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ix,nx), &
!$OMP             SHARED(np,p2p,bo,nr,scount,sdispl)
      DO ip = 0, np-1
         ix = p2p(ip)
         nx = bo(2, 1, ix)-bo(1, 1, ix)+1
         scount(ip) = nr*nx
         sdispl(ip) = nr*(bo(1, 1, ix)-1)
      END DO
!$OMP END PARALLEL DO
      nx = bo(2, 1, mpr)-bo(1, 1, mpr)+1
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(nr), &
!$OMP             SHARED(np,nray,nx,rcount,rdispl,nm)
      DO ip = 0, np-1
         nr = nray(ip)
         rcount(ip) = nr*nx
         rdispl(ip) = nm*nx*ip
      END DO
!$OMP END PARALLEL DO
      IF (alltoall_sgl) THEN
         CALL mp_alltoall(ss, scount, sdispl, tt, rcount, rdispl, group)
      ELSE
         CALL mp_alltoall(sb, scount, sdispl, rr, rcount, rdispl, group)
      END IF

      nx = bo(2, 1, mpr)-bo(1, 1, mpr)+1
!$OMP PARALLEL DO DEFAULT(NONE) COLLAPSE(2) &
!$OMP             PRIVATE(ixx,ir,iy,iz,ix) &
!$OMP             SHARED(np,nray,nx,alltoall_sgl,yzp,tt,rr,tb)
      DO ip = 0, np-1
         DO ix = 1, nx
            ixx = nray(ip)*(ix-1)
            IF (alltoall_sgl) THEN
               DO ir = 1, nray(ip)
                  iy = yzp(1, ir, ip)
                  iz = yzp(2, ir, ip)
                  tb(iy, iz, ix) = tt(ir+ixx, ip)
               END DO
            ELSE
               DO ir = 1, nray(ip)
                  iy = yzp(1, ir, ip)
                  iz = yzp(2, ir, ip)
                  tb(iy, iz, ix) = rr(ir+ixx, ip)
               END DO
            END IF
         END DO
      END DO
!$OMP END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE x_to_yz

! **************************************************************************************************
!> \brief ...
!> \param tb ...
!> \param group ...
!> \param my_pos ...
!> \param p2p ...
!> \param yzp ...
!> \param nray ...
!> \param bo ...
!> \param sb ...
!> \param fft_scratch ...
!> \par History
!>      15. Feb. 2006 : single precision all_to_all
!> \author JGH (14-Jan-2001)
! **************************************************************************************************
   SUBROUTINE yz_to_x(tb, group, my_pos, p2p, yzp, nray, bo, sb, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :, :), INTENT(IN)   :: tb
      INTEGER, INTENT(IN)                                :: group, my_pos
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: p2p
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: yzp
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nray
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: bo
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(INOUT)   :: sb
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'yz_to_x', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: rr
      COMPLEX(KIND=sp), DIMENSION(:, :), POINTER         :: ss, tt
      INTEGER                                            :: handle, ip, ir, ix, ixx, iy, iz, mpr, &
                                                            nm, np, nr, nx
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl, scount, sdispl

      CALL timeset(routineN, handle)

      np = SIZE(p2p)
      mpr = p2p(my_pos)
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl

      IF (alltoall_sgl) THEN
         ss => fft_scratch%ss
         tt => fft_scratch%tt
         ss = 0._sp
         tt = 0._sp
      ELSE
         rr => fft_scratch%rr
      END IF

      nx = bo(2, 1, mpr)-bo(1, 1, mpr)+1
!$OMP PARALLEL DO DEFAULT(NONE) COLLAPSE(2) &
!$OMP             PRIVATE(ip, ixx, ir, iy, iz, ix) &
!$OMP             SHARED(np,nray,nx,alltoall_sgl,yzp,tb,tt,rr)
      DO ip = 0, np-1
         DO ix = 1, nx
            ixx = nray(ip)*(ix-1)
            IF (alltoall_sgl) THEN
               DO ir = 1, nray(ip)
                  iy = yzp(1, ir, ip)
                  iz = yzp(2, ir, ip)
                  tt(ir+ixx, ip) = CMPLX(tb(iy, iz, ix), KIND=sp)
               END DO
            ELSE
               DO ir = 1, nray(ip)
                  iy = yzp(1, ir, ip)
                  iz = yzp(2, ir, ip)
                  rr(ir+ixx, ip) = tb(iy, iz, ix)
               END DO
            END IF
         END DO
      END DO
!$OMP END PARALLEL DO
      nm = MAXVAL(nray(0:np-1))
      nr = nray(my_pos)
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ix,nx), &
!$OMP             SHARED(np,p2p,bo,rcount,rdispl,nr)
      DO ip = 0, np-1
         ix = p2p(ip)
         nx = bo(2, 1, ix)-bo(1, 1, ix)+1
         rcount(ip) = nr*nx
         rdispl(ip) = nr*(bo(1, 1, ix)-1)
      END DO
!$OMP END PARALLEL DO
      nx = bo(2, 1, mpr)-bo(1, 1, mpr)+1
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(nr), &
!$OMP             SHARED(np,nray,scount,sdispl,nx,nm)
      DO ip = 0, np-1
         nr = nray(ip)
         scount(ip) = nr*nx
         sdispl(ip) = nm*nx*ip
      END DO
!$OMP END PARALLEL DO

      IF (alltoall_sgl) THEN
         CALL mp_alltoall(tt, scount, sdispl, ss, rcount, rdispl, group)
         sb = ss
      ELSE
         CALL mp_alltoall(rr, scount, sdispl, sb, rcount, rdispl, group)
      END IF

      CALL timestop(handle)

   END SUBROUTINE yz_to_x

! **************************************************************************************************
!> \brief ...
!> \param sb ...
!> \param group ...
!> \param dims ...
!> \param my_pos ...
!> \param p2p ...
!> \param yzp ...
!> \param nray ...
!> \param bo ...
!> \param tb ...
!> \param fft_scratch ...
!> \par History
!>      15. Feb. 2006 : single precision all_to_all
!> \author JGH (18-Jan-2001)
! **************************************************************************************************
   SUBROUTINE yz_to_xz(sb, group, dims, my_pos, p2p, yzp, nray, bo, tb, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: sb
      INTEGER, INTENT(IN)                                :: group
      INTEGER, DIMENSION(2), INTENT(IN)                  :: dims
      INTEGER, INTENT(IN)                                :: my_pos
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: p2p
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: yzp
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nray
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: bo
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(INOUT)   :: tb
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'yz_to_xz', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: xzbuf, yzbuf
      COMPLEX(KIND=sp), DIMENSION(:), POINTER            :: xzbuf_sgl, yzbuf_sgl
      INTEGER                                            :: handle, icrs, ip, ipl, ipr, ir, ix, iz, &
                                                            jj, jx, jy, jz, myx, myz, np, npx, &
                                                            npz, nx, nz, rs_pos
      INTEGER, DIMENSION(:), POINTER                     :: pzcoord, rcount, rdispl, scount, sdispl, &
                                                            xcor, zcor
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid

      CALL timeset(routineN, handle)

      np = SIZE(p2p)

      rs_pos = p2p(my_pos)

      IF (alltoall_sgl) THEN
         yzbuf_sgl => fft_scratch%yzbuf_sgl
         xzbuf_sgl => fft_scratch%xzbuf_sgl
      ELSE
         yzbuf => fft_scratch%yzbuf
         xzbuf => fft_scratch%xzbuf
      END IF
      npx = dims(1)
      npz = dims(2)
      pgrid => fft_scratch%pgrid
      xcor => fft_scratch%xcor
      zcor => fft_scratch%zcor
      pzcoord => fft_scratch%pzcoord
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl

      nx = SIZE(sb, 2)

! If the send and recv counts are not already cached, then
! calculate and store them
      IF (fft_scratch%in == 0) THEN

         scount = 0

         DO ix = 0, npx-1
            ip = pgrid(ix, 0)
            xcor(bo(1, 1, ip):bo(2, 1, ip)) = ix
         END DO
         DO iz = 0, npz-1
            ip = pgrid(0, iz)
            zcor(bo(1, 3, ip):bo(2, 3, ip)) = iz
         END DO
         DO jx = 1, nx
            IF (alltoall_sgl) THEN
               DO ir = 1, nray(my_pos)
                  jy = yzp(1, ir, my_pos)
                  jz = yzp(2, ir, my_pos)
                  ip = pgrid(xcor(jx), zcor(jz))
                  scount(ip) = scount(ip)+1
               END DO
            ELSE
               DO ir = 1, nray(my_pos)
                  jy = yzp(1, ir, my_pos)
                  jz = yzp(2, ir, my_pos)
                  ip = pgrid(xcor(jx), zcor(jz))
                  scount(ip) = scount(ip)+1
               END DO
            END IF
         END DO

         CALL mp_alltoall(scount, rcount, 1, group)
         fft_scratch%yzcount = scount
         fft_scratch%xzcount = rcount

         ! Work out the correct displacements in the buffers
         sdispl(0) = 0
         rdispl(0) = 0
         DO ip = 1, np-1
            sdispl(ip) = sdispl(ip-1)+scount(ip-1)
            rdispl(ip) = rdispl(ip-1)+rcount(ip-1)
         END DO

         fft_scratch%yzdispl = sdispl
         fft_scratch%xzdispl = rdispl

         icrs = 0
         DO ip = 0, np-1
            IF (scount(ip) /= 0) icrs = icrs+1
            IF (rcount(ip) /= 0) icrs = icrs+1
         END DO
         CALL mp_sum(icrs, group)
         fft_scratch%rsratio = REAL(icrs, KIND=dp)/(REAL(2*np, KIND=dp)*REAL(np, KIND=dp))

         fft_scratch%in = 1
      ELSE
         scount = fft_scratch%yzcount
         rcount = fft_scratch%xzcount
         sdispl = fft_scratch%yzdispl
         rdispl = fft_scratch%xzdispl
      END IF

! Do the actual packing
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipl,jj,nx,ir,jx,jy,jz),&
!$OMP             SHARED(np,p2p,pzcoord,bo,nray,yzp,zcor),&
!$OMP             SHARED(yzbuf,sb,scount,sdispl,my_pos),&
!$OMP             SHARED(yzbuf_sgl,alltoall_sgl)
      DO ip = 0, np-1
         IF (scount(ip) == 0) CYCLE
         ipl = p2p(ip)
         jj = 0
         nx = bo(2, 1, ipl)-bo(1, 1, ipl)+1
         DO ir = 1, nray(my_pos)
            jz = yzp(2, ir, my_pos)
            IF (zcor(jz) == pzcoord(ipl)) THEN
               jj = jj+1
               jy = yzp(1, ir, my_pos)
               IF (alltoall_sgl) THEN
                  DO jx = 0, nx-1
                     yzbuf_sgl(sdispl(ip)+jj+jx*scount(ip)/nx) = CMPLX(sb(ir, jx+bo(1, 1, ipl)), KIND=sp)
                  END DO
               ELSE
                  DO jx = 0, nx-1
                     yzbuf(sdispl(ip)+jj+jx*scount(ip)/nx) = sb(ir, jx+bo(1, 1, ipl))
                  END DO
               END IF
            END IF
         END DO
      END DO
!$OMP END PARALLEL DO

      IF (alltoall_sgl) THEN
         CALL mp_alltoall(yzbuf_sgl, scount, sdispl, xzbuf_sgl, rcount, rdispl, group)
      ELSE
         IF (fft_scratch%rsratio < ratio_sparse_alltoall) THEN
            CALL sparse_alltoall(yzbuf, scount, sdispl, xzbuf, rcount, rdispl, group)
         ELSE
            CALL mp_alltoall(yzbuf, scount, sdispl, xzbuf, rcount, rdispl, group)
         END IF
      END IF

      myx = fft_scratch%sizes%r_pos(1)
      myz = fft_scratch%sizes%r_pos(2)
      nz = bo(2, 3, rs_pos)-bo(1, 3, rs_pos)+1

!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipr,jj,ir,jx,jy,jz),&
!$OMP             SHARED(tb,np,p2p,bo,rs_pos,nray),&
!$OMP             SHARED(yzp,alltoall_sgl,zcor,myz),&
!$OMP             SHARED(xzbuf,xzbuf_sgl,nz,rdispl)
      DO ip = 0, np-1
         ipr = p2p(ip)
         jj = 0
         DO jx = 0, bo(2, 1, rs_pos)-bo(1, 1, rs_pos)
            DO ir = 1, nray(ip)
               jz = yzp(2, ir, ip)
               IF (alltoall_sgl) THEN
                  IF (zcor(jz) == myz) THEN
                     jj = jj+1
                     jy = yzp(1, ir, ip)
                     jz = jz-bo(1, 3, rs_pos)+1
                     tb(jy, jz+jx*nz) = xzbuf_sgl(jj+rdispl(ipr))
                  END IF
               ELSE
                  IF (zcor(jz) == myz) THEN
                     jj = jj+1
                     jy = yzp(1, ir, ip)
                     jz = jz-bo(1, 3, rs_pos)+1
                     tb(jy, jz+jx*nz) = xzbuf(jj+rdispl(ipr))
                  END IF
               END IF
            END DO
         END DO
      END DO
!$OMP END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE yz_to_xz

! **************************************************************************************************
!> \brief ...
!> \param sb ...
!> \param group ...
!> \param dims ...
!> \param my_pos ...
!> \param p2p ...
!> \param yzp ...
!> \param nray ...
!> \param bo ...
!> \param tb ...
!> \param fft_scratch ...
!> \par History
!>      15. Feb. 2006 : single precision all_to_all
!> \author JGH (19-Jan-2001)
! **************************************************************************************************
   SUBROUTINE xz_to_yz(sb, group, dims, my_pos, p2p, yzp, nray, bo, tb, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: sb
      INTEGER, INTENT(IN)                                :: group
      INTEGER, DIMENSION(2), INTENT(IN)                  :: dims
      INTEGER, INTENT(IN)                                :: my_pos
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: p2p
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: yzp
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nray
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: bo
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(INOUT)   :: tb
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'xz_to_yz', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: xzbuf, yzbuf
      COMPLEX(KIND=sp), DIMENSION(:), POINTER            :: xzbuf_sgl, yzbuf_sgl
      INTEGER                                            :: handle, icrs, ip, ipl, ir, ix, ixx, iz, &
                                                            jj, jx, jy, jz, mp, myx, myz, np, npx, &
                                                            npz, nx, nz
      INTEGER, DIMENSION(:), POINTER                     :: pzcoord, rcount, rdispl, scount, sdispl, &
                                                            xcor, zcor
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid

      CALL timeset(routineN, handle)

      np = SIZE(p2p)

      IF (alltoall_sgl) THEN
         yzbuf_sgl => fft_scratch%yzbuf_sgl
         xzbuf_sgl => fft_scratch%xzbuf_sgl
      ELSE
         yzbuf => fft_scratch%yzbuf
         xzbuf => fft_scratch%xzbuf
      END IF
      npx = dims(1)
      npz = dims(2)
      pgrid => fft_scratch%pgrid
      xcor => fft_scratch%xcor
      zcor => fft_scratch%zcor
      pzcoord => fft_scratch%pzcoord
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl

! If the send and recv counts are not already cached, then
! calculate and store them
      IF (fft_scratch%in == 0) THEN

         rcount = 0
         nx = MAXVAL(bo(2, 1, :))

         DO ix = 0, npx-1
            ip = pgrid(ix, 0)
            xcor(bo(1, 1, ip):bo(2, 1, ip)) = ix
         END DO
         DO iz = 0, npz-1
            ip = pgrid(0, iz)
            zcor(bo(1, 3, ip):bo(2, 3, ip)) = iz
         END DO
         DO jx = 1, nx
            DO ir = 1, nray(my_pos)
               jy = yzp(1, ir, my_pos)
               jz = yzp(2, ir, my_pos)
               ip = pgrid(xcor(jx), zcor(jz))
               rcount(ip) = rcount(ip)+1
            END DO
         END DO

         CALL mp_alltoall(rcount, scount, 1, group)
         fft_scratch%xzcount = scount
         fft_scratch%yzcount = rcount

         ! Work out the correct displacements in the buffers
         sdispl(0) = 0
         rdispl(0) = 0
         DO ip = 1, np-1
            sdispl(ip) = sdispl(ip-1)+scount(ip-1)
            rdispl(ip) = rdispl(ip-1)+rcount(ip-1)
         END DO

         fft_scratch%xzdispl = sdispl
         fft_scratch%yzdispl = rdispl

         icrs = 0
         DO ip = 0, np-1
            IF (scount(ip) /= 0) icrs = icrs+1
            IF (rcount(ip) /= 0) icrs = icrs+1
         END DO
         CALL mp_sum(icrs, group)
         fft_scratch%rsratio = REAL(icrs, KIND=dp)/(REAL(2*np, KIND=dp)*REAL(np, KIND=dp))

         fft_scratch%in = 1
      ELSE
         scount = fft_scratch%xzcount
         rcount = fft_scratch%yzcount
         sdispl = fft_scratch%xzdispl
         rdispl = fft_scratch%yzdispl
      END IF

! Now do the actual packing
      myx = fft_scratch%sizes%r_pos(1)
      myz = fft_scratch%sizes%r_pos(2)
      mp = p2p(my_pos)
      nz = bo(2, 3, mp)-bo(1, 3, mp)+1
      nx = bo(2, 1, mp)-bo(1, 1, mp)+1

!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(jj,ipl,ir,jx,jy,jz,ixx),&
!$OMP             SHARED(np,p2p,nray,yzp,zcor,myz,bo,mp),&
!$OMP             SHARED(alltoall_sgl,nx,scount,sdispl),&
!$OMP             SHARED(xzbuf,xzbuf_sgl,sb,nz)
      DO ip = 0, np-1
         jj = 0
         ipl = p2p(ip)
         DO ir = 1, nray(ip)
            jz = yzp(2, ir, ip)
            IF (zcor(jz) == myz) THEN
               jj = jj+1
               jy = yzp(1, ir, ip)
               jz = yzp(2, ir, ip)-bo(1, 3, mp)+1
               IF (alltoall_sgl) THEN
                  DO jx = 0, nx-1
                     ixx = jj+jx*scount(ipl)/nx
                     xzbuf_sgl(ixx+sdispl(ipl)) = CMPLX(sb(jy, jz+jx*nz), KIND=sp)
                  END DO
               ELSE
                  DO jx = 0, nx-1
                     ixx = jj+jx*scount(ipl)/nx
                     xzbuf(ixx+sdispl(ipl)) = sb(jy, jz+jx*nz)
                  END DO
               END IF
            END IF
         END DO
      END DO
!$OMP END PARALLEL DO

      IF (alltoall_sgl) THEN
         CALL mp_alltoall(xzbuf_sgl, scount, sdispl, yzbuf_sgl, rcount, rdispl, group)
      ELSE
         IF (fft_scratch%rsratio < ratio_sparse_alltoall) THEN
            CALL sparse_alltoall(xzbuf, scount, sdispl, yzbuf, rcount, rdispl, group)
         ELSE
            CALL mp_alltoall(xzbuf, scount, sdispl, yzbuf, rcount, rdispl, group)
         END IF
      END IF

!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipl,jj,nx,ir,jx,jy,jz),&
!$OMP             SHARED(p2p,pzcoord,bo,nray,my_pos,yzp),&
!$OMP             SHARED(rcount,rdispl,tb,yzbuf,zcor),&
!$OMP             SHARED(yzbuf_sgl,alltoall_sgl,np)
      DO ip = 0, np-1
         IF (rcount(ip) == 0) CYCLE
         ipl = p2p(ip)
         jj = 0
         nx = bo(2, 1, ipl)-bo(1, 1, ipl)+1
         DO ir = 1, nray(my_pos)
            jz = yzp(2, ir, my_pos)
            IF (zcor(jz) == pzcoord(ipl)) THEN
               jj = jj+1
               jy = yzp(1, ir, my_pos)
               IF (alltoall_sgl) THEN
                  DO jx = 0, nx-1
                     tb(ir, jx+bo(1, 1, ipl)) = yzbuf_sgl(rdispl(ip)+jj+jx*rcount(ip)/nx)
                  END DO
               ELSE
                  DO jx = 0, nx-1
                     tb(ir, jx+bo(1, 1, ipl)) = yzbuf(rdispl(ip)+jj+jx*rcount(ip)/nx)
                  END DO
               END IF
            END IF
         END DO
      END DO
!$OMP END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE xz_to_yz

! **************************************************************************************************
!> \brief ...
!> \param cin ...
!> \param boin ...
!> \param boout ...
!> \param sout ...
!> \param fft_scratch ...
!> \par History
!>      none
!> \author JGH (20-Jan-2001)
! **************************************************************************************************
   SUBROUTINE cube_transpose_1(cin, boin, boout, sout, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: cin
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: boin, boout
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(OUT)     :: sout
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'cube_transpose_1', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: rbuf
      INTEGER                                            :: handle, ip, ipl, ir, is, ixy, iz, mip, &
                                                            mz, np, nx, ny, nz, sub_group
      INTEGER, DIMENSION(2)                              :: dim, pos
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl, scount, sdispl
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid

      CALL timeset(routineN, handle)

      sub_group = fft_scratch%cart_sub_comm(2)
      mip = fft_scratch%mip
      dim = fft_scratch%dim
      pos = fft_scratch%pos
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl
      pgrid => fft_scratch%pgcube
      np = DIM(2)

      nx = boin(2, 1, mip)-boin(1, 1, mip)+1
      nz = boin(2, 3, mip)-boin(1, 3, mip)+1

!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipl,ny), &
!$OMP             SHARED(np,pgrid,boout,scount,sdispl,nx,nz)
      DO ip = 0, np-1
         ipl = pgrid(ip, 2)
         ny = boout(2, 2, ipl)-boout(1, 2, ipl)+1
         scount(ip) = nx*nz*ny
         sdispl(ip) = nx*nz*(boout(1, 2, ipl)-1)
      END DO
!$OMP END PARALLEL DO
      ny = boout(2, 2, mip)-boout(1, 2, mip)+1
      mz = MAXVAL(boin(2, 3, :)-boin(1, 3, :)+1)
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipl,nz), &
!$OMP             SHARED(np,pgrid,boin,nx,ny,rcount,rdispl,mz)
      DO ip = 0, np-1
         ipl = pgrid(ip, 2)
         nz = boin(2, 3, ipl)-boin(1, 3, ipl)+1
         rcount(ip) = nx*nz*ny
         rdispl(ip) = nx*ny*mz*ip
      END DO
!$OMP END PARALLEL DO

      rbuf => fft_scratch%rbuf1

      CALL mp_alltoall(cin, scount, sdispl, rbuf, rcount, rdispl, sub_group)

!$OMP PARALLEL DO DEFAULT(NONE) COLLAPSE(2) &
!$OMP             PRIVATE(ip,ipl,nz,iz,is,ir) &
!$OMP             SHARED(nx,ny,np,pgrid,boin,sout,rbuf)
      DO ixy = 1, nx*ny
         DO ip = 0, np-1
            ipl = pgrid(ip, 2)
            nz = boin(2, 3, ipl)-boin(1, 3, ipl)+1
            DO iz = 1, nz
               is = boin(1, 3, ipl)+iz-1
               ir = iz+nz*(ixy-1)
               sout(is, ixy) = rbuf(ir, ip)
            END DO
         END DO
      END DO
!$OMP END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE cube_transpose_1

! **************************************************************************************************
!> \brief ...
!> \param cin ...
!> \param boin ...
!> \param boout ...
!> \param sout ...
!> \param fft_scratch ...
! **************************************************************************************************
   SUBROUTINE cube_transpose_2(cin, boin, boout, sout, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: cin
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: boin, boout
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(OUT)     :: sout
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'cube_transpose_2', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: rbuf
      INTEGER                                            :: handle, ip, ipl, ir, ixy, iz, mip, mz, &
                                                            np, nx, ny, nz, sub_group
      INTEGER, DIMENSION(2)                              :: dim, pos
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl, scount, sdispl
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid

      CALL timeset(routineN, handle)

      sub_group = fft_scratch%cart_sub_comm(2)
      mip = fft_scratch%mip
      dim = fft_scratch%dim
      pos = fft_scratch%pos
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl
      pgrid => fft_scratch%pgcube
      np = DIM(2)

      nx = boin(2, 1, mip)-boin(1, 1, mip)+1
      ny = boin(2, 2, mip)-boin(1, 2, mip)+1
      mz = MAXVAL(boout(2, 3, :)-boout(1, 3, :)+1)

      rbuf => fft_scratch%rbuf2

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(ip,ipl,nz,iz,ir), &
!$OMP          SHARED(nx,ny,np,pgrid,boout,rbuf,cin,scount,sdispl,mz)
!$OMP DO COLLAPSE(2)
      DO ixy = 1, nx*ny
         DO ip = 0, np-1
            ipl = pgrid(ip, 2)
            nz = boout(2, 3, ipl)-boout(1, 3, ipl)+1
            DO iz = boout(1, 3, ipl), boout(2, 3, ipl)
               ir = iz-boout(1, 3, ipl)+1+(ixy-1)*nz
               rbuf(ir, ip) = cin(iz, ixy)
            END DO
         END DO
      END DO
!$OMP END DO
!$OMP DO
      DO ip = 0, np-1
         ipl = pgrid(ip, 2)
         nz = boout(2, 3, ipl)-boout(1, 3, ipl)+1
         scount(ip) = nx*ny*nz
         sdispl(ip) = nx*ny*mz*ip
      END DO
!$OMP END DO
!$OMP END PARALLEL
      nz = boout(2, 3, mip)-boout(1, 3, mip)+1
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipl,ny), &
!$OMP             SHARED(np,pgrid,boin,nx,nz,rcount,rdispl)
      DO ip = 0, np-1
         ipl = pgrid(ip, 2)
         ny = boin(2, 2, ipl)-boin(1, 2, ipl)+1
         rcount(ip) = nx*ny*nz
         rdispl(ip) = nx*nz*(boin(1, 2, ipl)-1)
      END DO
!$OMP END PARALLEL DO

      CALL mp_alltoall(rbuf, scount, sdispl, sout, rcount, rdispl, sub_group)

      CALL timestop(handle)

   END SUBROUTINE cube_transpose_2

! **************************************************************************************************
!> \brief ...
!> \param cin ...
!> \param boin ...
!> \param boout ...
!> \param sout ...
!> \param fft_scratch ...
! **************************************************************************************************
   SUBROUTINE cube_transpose_3(cin, boin, boout, sout, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: cin
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: boin, boout
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(OUT)     :: sout
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'cube_transpose_3', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: rbuf
      INTEGER                                            :: handle, ip, ipl, ir, is, ixz, iy, lb, &
                                                            mip, my, my_id, np, num_threads, nx, &
                                                            ny, nz, sub_group, ub
      INTEGER, DIMENSION(2)                              :: dim, pos
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl, scount, sdispl
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid

      CALL timeset(routineN, handle)

      sub_group = fft_scratch%cart_sub_comm(1)
      mip = fft_scratch%mip
      dim = fft_scratch%dim
      pos = fft_scratch%pos
      np = DIM(1)
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl
      pgrid => fft_scratch%pgcube

      ny = boin(2, 2, mip)-boin(1, 2, mip)+1
      nz = boin(2, 3, mip)-boin(1, 3, mip)+1
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipl, nx), &
!$OMP             SHARED(np,pgrid,boout,ny,nz,scount,sdispl)
      DO ip = 0, np-1
         ipl = pgrid(ip, 1)
         nx = boout(2, 1, ipl)-boout(1, 1, ipl)+1
         scount(ip) = nx*nz*ny
         sdispl(ip) = ny*nz*(boout(1, 1, ipl)-1)
      END DO
!$OMP END PARALLEL DO
      nx = boout(2, 1, mip)-boout(1, 1, mip)+1
      my = MAXVAL(boin(2, 2, :)-boin(1, 2, :)+1)
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipl, ny), &
!$OMP             SHARED(np,pgrid,boin,nx,nz,my,rcount,rdispl)
      DO ip = 0, np-1
         ipl = pgrid(ip, 1)
         ny = boin(2, 2, ipl)-boin(1, 2, ipl)+1
         rcount(ip) = nx*nz*ny
         rdispl(ip) = nx*my*nz*ip
      END DO
!$OMP END PARALLEL DO

      rbuf => fft_scratch%rbuf3
      num_threads = 1
      my_id = 0
!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(NUM_THREADS, my_id, lb, ub), &
!$OMP          SHARED(rbuf)
!$    num_threads = MIN(omp_get_max_threads(), SIZE(rbuf, 2))
!$    my_id = omp_get_thread_num()
      IF (my_id < num_threads) THEN
         lb = (SIZE(rbuf, 2)*my_id)/num_threads
         ub = (SIZE(rbuf, 2)*(my_id+1))/num_threads-1
         rbuf(:, lb:ub) = 0.0_dp
      END IF
!$OMP END PARALLEL

      CALL mp_alltoall(cin, scount, sdispl, rbuf, rcount, rdispl, sub_group)

!$OMP PARALLEL DO DEFAULT(NONE) COLLAPSE(2) &
!$OMP             PRIVATE(ip,ipl,ny,iy,is,ir) &
!$OMP             SHARED(nx,nz,np,pgrid,boin,rbuf,sout)
      DO ixz = 1, nx*nz
         DO ip = 0, np-1
            ipl = pgrid(ip, 1)
            ny = boin(2, 2, ipl)-boin(1, 2, ipl)+1
            DO iy = 1, ny
               is = boin(1, 2, ipl)+iy-1
               ir = iy+ny*(ixz-1)
               sout(is, ixz) = rbuf(ir, ip)
            END DO
         END DO
      END DO
!$OMP END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE cube_transpose_3

! **************************************************************************************************
!> \brief ...
!> \param cin ...
!> \param boin ...
!> \param boout ...
!> \param sout ...
!> \param fft_scratch ...
! **************************************************************************************************
   SUBROUTINE cube_transpose_4(cin, boin, boout, sout, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: cin
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: boin, boout
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(OUT)     :: sout
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'cube_transpose_4', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: rbuf
      INTEGER                                            :: handle, ip, ipl, ir, iy, izx, lb, mip, &
                                                            my, my_id, np, num_threads, nx, ny, &
                                                            nz, sub_group, ub
      INTEGER, DIMENSION(2)                              :: dim, pos
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl, scount, sdispl
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid

      CALL timeset(routineN, handle)

      sub_group = fft_scratch%cart_sub_comm(1)
      mip = fft_scratch%mip
      dim = fft_scratch%dim
      pos = fft_scratch%pos
      np = DIM(1)
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl
      pgrid => fft_scratch%pgcube

      nx = boin(2, 1, mip)-boin(1, 1, mip)+1
      nz = boin(2, 3, mip)-boin(1, 3, mip)+1
      my = MAXVAL(boout(2, 2, :)-boout(1, 2, :)+1)

      rbuf => fft_scratch%rbuf4
      num_threads = 1
      my_id = 0
!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(NUM_THREADS,my_id,lb,ub,ip,ipl,ny,iy,ir), &
!$OMP          SHARED(rbuf,nz,nx,np,pgrid,boout,cin,my,scount,sdispl)
!$    num_threads = MIN(omp_get_max_threads(), SIZE(rbuf, 2))
!$    my_id = omp_get_thread_num()
      IF (my_id < num_threads) THEN
         lb = (SIZE(rbuf, 2)*my_id)/num_threads
         ub = (SIZE(rbuf, 2)*(my_id+1))/num_threads-1
         rbuf(:, lb:ub) = 0.0_dp
      END IF
!$OMP BARRIER

!$OMP DO COLLAPSE(2)
      DO izx = 1, nz*nx
         DO ip = 0, np-1
            ipl = pgrid(ip, 1)
            ny = boout(2, 2, ipl)-boout(1, 2, ipl)+1
            DO iy = boout(1, 2, ipl), boout(2, 2, ipl)
               ir = iy-boout(1, 2, ipl)+1+(izx-1)*ny
               rbuf(ir, ip) = cin(iy, izx)
            END DO
         END DO
      END DO
!$OMP END DO
!$OMP DO
      DO ip = 0, np-1
         ipl = pgrid(ip, 1)
         ny = boout(2, 2, ipl)-boout(1, 2, ipl)+1
         scount(ip) = nx*ny*nz
         sdispl(ip) = nx*nz*my*ip
      END DO
!$OMP END DO
!$OMP END PARALLEL
      ny = boout(2, 2, mip)-boout(1, 2, mip)+1
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ipl,nx), &
!$OMP             SHARED(np,pgrid,boin,rcount,rdispl,ny,nz)
      DO ip = 0, np-1
         ipl = pgrid(ip, 1)
         nx = boin(2, 1, ipl)-boin(1, 1, ipl)+1
         rcount(ip) = nx*ny*nz
         rdispl(ip) = ny*nz*(boin(1, 1, ipl)-1)
      END DO
!$OMP END PARALLEL DO

      CALL mp_alltoall(rbuf, scount, sdispl, sout, rcount, rdispl, sub_group)

      CALL timestop(handle)

   END SUBROUTINE cube_transpose_4

! **************************************************************************************************
!> \brief ...
!> \param cin ...
!> \param group ...
!> \param boin ...
!> \param boout ...
!> \param sout ...
!> \param fft_scratch ...
! **************************************************************************************************
   SUBROUTINE cube_transpose_5(cin, group, boin, boout, sout, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: cin
      INTEGER, INTENT(IN)                                :: group
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: boin, boout
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(OUT)     :: sout
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'cube_transpose_5', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: rbuf
      INTEGER                                            :: handle, ip, ir, is, ixz, iy, lb, mip, &
                                                            my, my_id, np, num_threads, nx, ny, &
                                                            nz, ub
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl, scount, sdispl

      CALL timeset(routineN, handle)

      np = fft_scratch%sizes%numtask
      mip = fft_scratch%mip
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl

      ny = boin(2, 2, mip)-boin(1, 2, mip)+1
      nz = boin(2, 3, mip)-boin(1, 3, mip)+1
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(nx), &
!$OMP             SHARED(np,boout,ny,nz,scount,sdispl)
      DO ip = 0, np-1
         nx = boout(2, 1, ip)-boout(1, 1, ip)+1
         scount(ip) = nx*nz*ny
         sdispl(ip) = ny*nz*(boout(1, 1, ip)-1)
      END DO
!$OMP END PARALLEL DO
      nx = boout(2, 1, mip)-boout(1, 1, mip)+1
      my = MAXVAL(boin(2, 2, :)-boin(1, 2, :)+1)
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(ny), &
!$OMP             SHARED(np,boin,nx,nz,rcount,rdispl,my)
      DO ip = 0, np-1
         ny = boin(2, 2, ip)-boin(1, 2, ip)+1
         rcount(ip) = nx*nz*ny
         rdispl(ip) = nx*my*nz*ip
      END DO
!$OMP END PARALLEL DO

      rbuf => fft_scratch%rbuf5
      num_threads = 1
      my_id = 0
!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(NUM_THREADS, my_id, lb, ub), &
!$OMP          SHARED(rbuf)
!$    num_threads = MIN(omp_get_max_threads(), SIZE(rbuf, 2))
!$    my_id = omp_get_thread_num()
      IF (my_id < num_threads) THEN
         lb = (SIZE(rbuf, 2)*my_id)/num_threads
         ub = (SIZE(rbuf, 2)*(my_id+1))/num_threads-1
         rbuf(:, lb:ub) = 0.0_dp
      END IF
!$OMP END PARALLEL

      CALL mp_alltoall(cin, scount, sdispl, rbuf, rcount, rdispl, group)

!$OMP PARALLEL DO DEFAULT(NONE) COLLAPSE(2) &
!$OMP             PRIVATE(ip,ny,iy,is,ir) &
!$OMP             SHARED(nx,nz,np,boin,sout,rbuf)
      DO ixz = 1, nx*nz
         DO ip = 0, np-1
            ny = boin(2, 2, ip)-boin(1, 2, ip)+1
            DO iy = 1, ny
               is = boin(1, 2, ip)+iy-1
               ir = iy+ny*(ixz-1)
               sout(is, ixz) = rbuf(ir, ip)
            END DO
         END DO
      END DO
!$OMP END PARALLEL DO

      CALL timestop(handle)

   END SUBROUTINE cube_transpose_5

! **************************************************************************************************
!> \brief ...
!> \param cin ...
!> \param group ...
!> \param boin ...
!> \param boout ...
!> \param sout ...
!> \param fft_scratch ...
! **************************************************************************************************
   SUBROUTINE cube_transpose_6(cin, group, boin, boout, sout, fft_scratch)

      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(IN)      :: cin
      INTEGER, INTENT(IN)                                :: group
      INTEGER, DIMENSION(:, :, 0:), INTENT(IN)           :: boin, boout
      COMPLEX(KIND=dp), DIMENSION(:, :), INTENT(OUT)     :: sout
      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'cube_transpose_6', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER         :: rbuf
      INTEGER                                            :: handle, ip, ir, iy, izx, lb, mip, my, &
                                                            my_id, np, num_threads, nx, ny, nz, ub
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl, scount, sdispl

      CALL timeset(routineN, handle)

      np = fft_scratch%sizes%numtask
      mip = fft_scratch%mip
      scount => fft_scratch%scount
      rcount => fft_scratch%rcount
      sdispl => fft_scratch%sdispl
      rdispl => fft_scratch%rdispl

      nx = boin(2, 1, mip)-boin(1, 1, mip)+1
      nz = boin(2, 3, mip)-boin(1, 3, mip)+1
      my = MAXVAL(boout(2, 2, :)-boout(1, 2, :)+1)

      rbuf => fft_scratch%rbuf5
      num_threads = 1
      my_id = 0
!$OMP PARALLEL DEFAULT(NONE), &
!$OMP          PRIVATE(NUM_THREADS,my_id,lb,ub,ip,ny,iy,ir), &
!$OMP          SHARED(rbuf,nx,nz,np,boout,cin,my,scount,sdispl)
!$    num_threads = MIN(omp_get_max_threads(), SIZE(rbuf, 2))
!$    my_id = omp_get_thread_num()
      IF (my_id < num_threads) THEN
         lb = (SIZE(rbuf, 2)*my_id)/num_threads
         ub = (SIZE(rbuf, 2)*(my_id+1))/num_threads-1
         rbuf(:, lb:ub) = 0.0_dp
      END IF
!$OMP BARRIER

!$OMP DO COLLAPSE(2)
      DO izx = 1, nz*nx
         DO ip = 0, np-1
            ny = boout(2, 2, ip)-boout(1, 2, ip)+1
            DO iy = boout(1, 2, ip), boout(2, 2, ip)
               ir = iy-boout(1, 2, ip)+1+(izx-1)*ny
               rbuf(ir, ip) = cin(iy, izx)
            END DO
         END DO
      END DO
!$OMP END DO
!$OMP DO
      DO ip = 0, np-1
         ny = boout(2, 2, ip)-boout(1, 2, ip)+1
         scount(ip) = nx*ny*nz
         sdispl(ip) = nx*nz*my*ip
      END DO
!$OMP END DO
!$OMP END PARALLEL
      ny = boout(2, 2, mip)-boout(1, 2, mip)+1
!$OMP PARALLEL DO DEFAULT(NONE), &
!$OMP             PRIVATE(nx), &
!$OMP             SHARED(np,boin,rcount,rdispl,nz,ny)
      DO ip = 0, np-1
         nx = boin(2, 1, ip)-boin(1, 1, ip)+1
         rcount(ip) = nx*ny*nz
         rdispl(ip) = ny*nz*(boin(1, 1, ip)-1)
      END DO
!$OMP END PARALLEL DO

      CALL mp_alltoall(rbuf, scount, sdispl, sout, rcount, rdispl, group)

      CALL timestop(handle)

   END SUBROUTINE cube_transpose_6

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE init_fft_scratch_pool()

      CHARACTER(len=*), PARAMETER :: routineN = 'init_fft_scratch_pool', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      CALL release_fft_scratch_pool()

      ! Allocate first scratch and mark it as used
      ALLOCATE (fft_scratch_first)
      ALLOCATE (fft_scratch_first%fft_scratch)
      NULLIFY (fft_scratch_first%fft_scratch_next)
      fft_scratch_first%fft_scratch%fft_scratch_id = 0
      fft_scratch_first%fft_scratch%in_use = .TRUE.
      fft_scratch_first%fft_scratch%group = 0
      NULLIFY (fft_scratch_first%fft_scratch%ziptr)
      NULLIFY (fft_scratch_first%fft_scratch%zoptr)
      NULLIFY (fft_scratch_first%fft_scratch%p1buf)
      NULLIFY (fft_scratch_first%fft_scratch%p2buf)
      NULLIFY (fft_scratch_first%fft_scratch%p3buf)
      NULLIFY (fft_scratch_first%fft_scratch%p4buf)
      NULLIFY (fft_scratch_first%fft_scratch%p5buf)
      NULLIFY (fft_scratch_first%fft_scratch%p6buf)
      NULLIFY (fft_scratch_first%fft_scratch%p7buf)
      NULLIFY (fft_scratch_first%fft_scratch%r1buf)
      NULLIFY (fft_scratch_first%fft_scratch%r2buf)
      NULLIFY (fft_scratch_first%fft_scratch%tbuf)
      NULLIFY (fft_scratch_first%fft_scratch%a1buf)
      NULLIFY (fft_scratch_first%fft_scratch%a2buf)
      NULLIFY (fft_scratch_first%fft_scratch%a3buf)
      NULLIFY (fft_scratch_first%fft_scratch%a4buf)
      NULLIFY (fft_scratch_first%fft_scratch%a5buf)
      NULLIFY (fft_scratch_first%fft_scratch%a6buf)
      NULLIFY (fft_scratch_first%fft_scratch%scount, fft_scratch_first%fft_scratch%rcount, &
               fft_scratch_first%fft_scratch%sdispl, fft_scratch_first%fft_scratch%rdispl)
      NULLIFY (fft_scratch_first%fft_scratch%rr, &
               fft_scratch_first%fft_scratch%ss, fft_scratch_first%fft_scratch%tt)
      NULLIFY (fft_scratch_first%fft_scratch%xzbuf, fft_scratch_first%fft_scratch%yzbuf, &
               fft_scratch_first%fft_scratch%xzbuf_sgl, fft_scratch_first%fft_scratch%yzbuf_sgl)
      NULLIFY (fft_scratch_first%fft_scratch%pgrid)
      NULLIFY (fft_scratch_first%fft_scratch%pgcube)
      NULLIFY (fft_scratch_first%fft_scratch%xcor, fft_scratch_first%fft_scratch%zcor)
      NULLIFY (fft_scratch_first%fft_scratch%pzcoord)
      NULLIFY (fft_scratch_first%fft_scratch%xzcount, fft_scratch_first%fft_scratch%yzcount, &
               fft_scratch_first%fft_scratch%xzdispl, fft_scratch_first%fft_scratch%yzdispl)
      NULLIFY (fft_scratch_first%fft_scratch%rbuf1, fft_scratch_first%fft_scratch%rbuf2, &
               fft_scratch_first%fft_scratch%rbuf3, fft_scratch_first%fft_scratch%rbuf4)
      NULLIFY (fft_scratch_first%fft_scratch%rbuf5, fft_scratch_first%fft_scratch%rbuf6)
      fft_scratch_first%fft_scratch%in = 0
      fft_scratch_first%fft_scratch%rsratio = 1._dp
      DO i = 1, 6
         fft_scratch_first%fft_scratch%fft_plan(i)%valid = .FALSE.
      END DO
      ! this is a very special scratch, it seems, we always keep it 'most - recent' so we will never delete it
      fft_scratch_first%fft_scratch%last_tick = HUGE(fft_scratch_first%fft_scratch%last_tick)

      init_fft_pool = init_fft_pool+1

   END SUBROUTINE init_fft_scratch_pool

! **************************************************************************************************
!> \brief ...
!> \param fft_scratch ...
! **************************************************************************************************
   SUBROUTINE deallocate_fft_scratch_type(fft_scratch)
      TYPE(fft_scratch_type), INTENT(INOUT)    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_fft_scratch_type', &
                                     routineP = moduleN//':'//routineN


      ! deallocate structures
      IF (ASSOCIATED(fft_scratch%ziptr)) THEN
         DEALLOCATE (fft_scratch%ziptr)
      END IF
      IF (ASSOCIATED(fft_scratch%zoptr)) THEN
         DEALLOCATE (fft_scratch%zoptr)
      END IF
      IF (ASSOCIATED(fft_scratch%p1buf)) THEN
         DEALLOCATE (fft_scratch%p1buf)
      END IF
      IF (ASSOCIATED(fft_scratch%p2buf)) THEN
         DEALLOCATE (fft_scratch%p2buf)
      END IF
      IF (ASSOCIATED(fft_scratch%p3buf)) THEN
         DEALLOCATE (fft_scratch%p3buf)
      END IF
      IF (ASSOCIATED(fft_scratch%p4buf)) THEN
         DEALLOCATE (fft_scratch%p4buf)
      END IF
      IF (ASSOCIATED(fft_scratch%p5buf)) THEN
         DEALLOCATE (fft_scratch%p5buf)
      END IF
      IF (ASSOCIATED(fft_scratch%p6buf)) THEN
         DEALLOCATE (fft_scratch%p6buf)
      END IF
      IF (ASSOCIATED(fft_scratch%p7buf)) THEN
         DEALLOCATE (fft_scratch%p7buf)
      END IF
      IF (ASSOCIATED(fft_scratch%r1buf)) THEN
         DEALLOCATE (fft_scratch%r1buf)
      END IF
      IF (ASSOCIATED(fft_scratch%r2buf)) THEN
         DEALLOCATE (fft_scratch%r2buf)
      END IF
      IF (ASSOCIATED(fft_scratch%tbuf)) THEN
         DEALLOCATE (fft_scratch%tbuf)
      END IF
      IF (ASSOCIATED(fft_scratch%a1buf)) THEN
         DEALLOCATE (fft_scratch%a1buf)
      END IF
      IF (ASSOCIATED(fft_scratch%a2buf)) THEN
         DEALLOCATE (fft_scratch%a2buf)
      END IF
      IF (ASSOCIATED(fft_scratch%a3buf)) THEN
         DEALLOCATE (fft_scratch%a3buf)
      END IF
      IF (ASSOCIATED(fft_scratch%a4buf)) THEN
         DEALLOCATE (fft_scratch%a4buf)
      END IF
      IF (ASSOCIATED(fft_scratch%a5buf)) THEN
         DEALLOCATE (fft_scratch%a5buf)
      END IF
      IF (ASSOCIATED(fft_scratch%a6buf)) THEN
         DEALLOCATE (fft_scratch%a6buf)
      END IF
      IF (ASSOCIATED(fft_scratch%scount)) THEN
         DEALLOCATE (fft_scratch%scount, fft_scratch%rcount, &
                     fft_scratch%sdispl, fft_scratch%rdispl)
      END IF
      IF (ASSOCIATED(fft_scratch%rr)) THEN
         DEALLOCATE (fft_scratch%rr)
      END IF
      IF (ASSOCIATED(fft_scratch%xzbuf)) THEN
         DEALLOCATE (fft_scratch%xzbuf)
      END IF
      IF (ASSOCIATED(fft_scratch%yzbuf)) THEN
         DEALLOCATE (fft_scratch%yzbuf)
      END IF
      IF (ASSOCIATED(fft_scratch%xzbuf_sgl)) THEN
         DEALLOCATE (fft_scratch%xzbuf_sgl)
      END IF
      IF (ASSOCIATED(fft_scratch%yzbuf_sgl)) THEN
         DEALLOCATE (fft_scratch%yzbuf_sgl)
      END IF
      IF (ASSOCIATED(fft_scratch%ss)) THEN
         DEALLOCATE (fft_scratch%ss)
      END IF
      IF (ASSOCIATED(fft_scratch%tt)) THEN
         DEALLOCATE (fft_scratch%tt)
      END IF
      IF (ASSOCIATED(fft_scratch%pgrid)) THEN
         DEALLOCATE (fft_scratch%pgrid)
      END IF
      IF (ASSOCIATED(fft_scratch%pgcube)) THEN
         DEALLOCATE (fft_scratch%pgcube)
      END IF
      IF (ASSOCIATED(fft_scratch%xcor)) THEN
         DEALLOCATE (fft_scratch%xcor, fft_scratch%zcor)
      END IF
      IF (ASSOCIATED(fft_scratch%pzcoord)) THEN
         DEALLOCATE (fft_scratch%pzcoord)
      END IF
      IF (ASSOCIATED(fft_scratch%xzcount)) THEN
         DEALLOCATE (fft_scratch%xzcount, fft_scratch%yzcount)
         DEALLOCATE (fft_scratch%xzdispl, fft_scratch%yzdispl)
         fft_scratch%in = 0
         fft_scratch%rsratio = 1._dp
      END IF
      IF (ASSOCIATED(fft_scratch%rbuf1)) THEN
         DEALLOCATE (fft_scratch%rbuf1)
      END IF
      IF (ASSOCIATED(fft_scratch%rbuf2)) THEN
         DEALLOCATE (fft_scratch%rbuf2)
      END IF
      IF (ASSOCIATED(fft_scratch%rbuf3)) THEN
         DEALLOCATE (fft_scratch%rbuf3)
      END IF
      IF (ASSOCIATED(fft_scratch%rbuf4)) THEN
         DEALLOCATE (fft_scratch%rbuf4)
      END IF
      IF (ASSOCIATED(fft_scratch%rbuf5)) THEN
         DEALLOCATE (fft_scratch%rbuf5)
      END IF
      IF (ASSOCIATED(fft_scratch%rbuf6)) THEN
         DEALLOCATE (fft_scratch%rbuf6)
      END IF

      IF (fft_scratch%cart_sub_comm(1) .NE. mp_comm_null) THEN
         CALL mp_comm_free(fft_scratch%cart_sub_comm(1))
      END IF
      IF (fft_scratch%cart_sub_comm(2) .NE. mp_comm_null) THEN
         CALL mp_comm_free(fft_scratch%cart_sub_comm(2))
      END IF
      fft_scratch%cart_sub_comm = mp_comm_null

      CALL fft_destroy_plan(fft_scratch%fft_plan(1))
      CALL fft_destroy_plan(fft_scratch%fft_plan(2))
      CALL fft_destroy_plan(fft_scratch%fft_plan(3))
      CALL fft_destroy_plan(fft_scratch%fft_plan(4))
      CALL fft_destroy_plan(fft_scratch%fft_plan(5))
      CALL fft_destroy_plan(fft_scratch%fft_plan(6))

   END SUBROUTINE deallocate_fft_scratch_type

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE release_fft_scratch_pool()

      CHARACTER(len=*), PARAMETER :: routineN = 'release_fft_scratch_pool', &
         routineP = moduleN//':'//routineN

      TYPE(fft_scratch_pool_type), POINTER               :: fft_scratch, fft_scratch_current

      IF (init_fft_pool == 0) NULLIFY (fft_scratch_first)

      fft_scratch => fft_scratch_first
      DO
         IF (ASSOCIATED(fft_scratch)) THEN
            fft_scratch_current => fft_scratch
            fft_scratch => fft_scratch_current%fft_scratch_next
            NULLIFY (fft_scratch_current%fft_scratch_next)

            CALL deallocate_fft_scratch_type(fft_scratch_current%fft_scratch)

            DEALLOCATE (fft_scratch_current%fft_scratch)
            DEALLOCATE (fft_scratch_current)
         ELSE
            EXIT
         END IF
      END DO

      init_fft_pool = 0

   END SUBROUTINE release_fft_scratch_pool

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE resize_fft_scratch_pool()

      CHARACTER(len=*), PARAMETER :: routineN = 'resize_fft_scratch_pool', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: last_tick, nscratch
      TYPE(fft_scratch_pool_type), POINTER               :: fft_scratch_current, fft_scratch_old

      nscratch = 0

      last_tick = HUGE(last_tick)
      NULLIFY (fft_scratch_old)

      ! start at the global pool, count, and find a deletion candidate
      fft_scratch_current => fft_scratch_first
      DO
         IF (ASSOCIATED(fft_scratch_current)) THEN
            nscratch = nscratch+1
            ! is this a candidate for deletion (i.e. least recently used, and not in use)
            IF (.NOT. fft_scratch_current%fft_scratch%in_use) THEN
               IF (fft_scratch_current%fft_scratch%last_tick < last_tick) THEN
                  last_tick = fft_scratch_current%fft_scratch%last_tick
                  fft_scratch_old => fft_scratch_current
               ENDIF
            ENDIF
            fft_scratch_current => fft_scratch_current%fft_scratch_next
         ELSE
            EXIT
         ENDIF
      ENDDO

      ! we should delete a scratch
      IF (nscratch > fft_pool_scratch_limit) THEN
         ! note that we never deallocate the first (special) element of the list
         IF (ASSOCIATED(fft_scratch_old)) THEN
            fft_scratch_current => fft_scratch_first
            DO
               IF (ASSOCIATED(fft_scratch_current)) THEN
                  ! should we delete the next in the list?
                  IF (ASSOCIATED(fft_scratch_current%fft_scratch_next, fft_scratch_old)) THEN
                     ! fix the linked list
                     fft_scratch_current%fft_scratch_next => fft_scratch_old%fft_scratch_next

                     ! deallocate the element
                     CALL deallocate_fft_scratch_type(fft_scratch_old%fft_scratch)
                     DEALLOCATE (fft_scratch_old%fft_scratch)
                     DEALLOCATE (fft_scratch_old)

                  ELSE
                     fft_scratch_current => fft_scratch_current%fft_scratch_next
                  ENDIF
               ELSE
                  EXIT
               ENDIF
            ENDDO

         ELSE
            CALL cp__w("pw/fft_tools.F",2860,"The number of the scratches exceeded the limit, but none could be deallocated")
         ENDIF
      ENDIF

   END SUBROUTINE resize_fft_scratch_pool

! **************************************************************************************************
!> \brief ...
!> \param fft_scratch ...
!> \param tf_type ...
!> \param n ...
!> \param fft_sizes ...
! **************************************************************************************************
   SUBROUTINE get_fft_scratch(fft_scratch, tf_type, n, fft_sizes)
      TYPE(fft_scratch_type), POINTER          :: fft_scratch
      INTEGER, INTENT(IN)                      :: tf_type
      INTEGER, DIMENSION(:), INTENT(IN)        :: n
      TYPE(fft_scratch_sizes), INTENT(IN), &
         OPTIONAL                               :: fft_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'get_fft_scratch', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: coord(2), DIM(2), handle, i, ix, iz, lg, lmax, m1, m2, &
                 mcx2, mcy3, mcz1, mcz2, mg, mmax, mx1, mx2, my1, my3, mz1, mz2, mz3, &
                 nbx, nbz, nm, nmax, nmray, nn, np, nx, ny, nyzray, nz, pos(2)
      INTEGER, DIMENSION(3)                    :: pcoord
      LOGICAL                                  :: equal
      LOGICAL, DIMENSION(2)                    :: dims
      TYPE(fft_scratch_pool_type), POINTER     :: fft_scratch_current, &
                                                  fft_scratch_last, &
                                                  fft_scratch_new
      CALL timeset(routineN, handle)

      ! this is the place to check that the scratch_pool does not grow without limits
      ! before we add a new scratch check the size of the pool and release some of the list if needed
      CALL resize_fft_scratch_pool()

      ! get the required scratch
      tick_fft_pool = tick_fft_pool+1
      fft_scratch_current => fft_scratch_first
      DO
         IF (ASSOCIATED(fft_scratch_current)) THEN
            IF (fft_scratch_current%fft_scratch%in_use) THEN
               fft_scratch_last => fft_scratch_current
               fft_scratch_current => fft_scratch_current%fft_scratch_next
               CYCLE
            END IF
            IF (tf_type /= fft_scratch_current%fft_scratch%tf_type) THEN
               fft_scratch_last => fft_scratch_current
               fft_scratch_current => fft_scratch_current%fft_scratch_next
               CYCLE
            END IF
            IF (.NOT. ALL(n == fft_scratch_current%fft_scratch%nfft)) THEN
               fft_scratch_last => fft_scratch_current
               fft_scratch_current => fft_scratch_current%fft_scratch_next
               CYCLE
            END IF
            IF (PRESENT(fft_sizes)) THEN
               IF (fft_sizes%gs_group /= fft_scratch_current%fft_scratch%group) THEN
                  fft_scratch_last => fft_scratch_current
                  fft_scratch_current => fft_scratch_current%fft_scratch_next
                  CYCLE
               END IF
               CALL is_equal(fft_sizes, fft_scratch_current%fft_scratch%sizes, equal)
               IF (.NOT. equal) THEN
                  fft_scratch_last => fft_scratch_current
                  fft_scratch_current => fft_scratch_current%fft_scratch_next
                  CYCLE
               ENDIF
            END IF
            ! Success
            fft_scratch => fft_scratch_current%fft_scratch
            fft_scratch_current%fft_scratch%in_use = .TRUE.
            EXIT
         ELSE
            ! We cannot find the scratch type in this pool
            ! Generate a new scratch set
            ALLOCATE (fft_scratch_new)
            ALLOCATE (fft_scratch_new%fft_scratch)
            fft_scratch_new%fft_scratch%group = 0
            NULLIFY (fft_scratch_new%fft_scratch%ziptr)
            NULLIFY (fft_scratch_new%fft_scratch%zoptr)
            NULLIFY (fft_scratch_new%fft_scratch%p1buf)
            NULLIFY (fft_scratch_new%fft_scratch%p2buf)
            NULLIFY (fft_scratch_new%fft_scratch%p3buf)
            NULLIFY (fft_scratch_new%fft_scratch%p4buf)
            NULLIFY (fft_scratch_new%fft_scratch%p5buf)
            NULLIFY (fft_scratch_new%fft_scratch%p6buf)
            NULLIFY (fft_scratch_new%fft_scratch%p7buf)
            NULLIFY (fft_scratch_new%fft_scratch%r1buf)
            NULLIFY (fft_scratch_new%fft_scratch%r2buf)
            NULLIFY (fft_scratch_new%fft_scratch%tbuf)
            NULLIFY (fft_scratch_new%fft_scratch%a1buf)
            NULLIFY (fft_scratch_new%fft_scratch%a2buf)
            NULLIFY (fft_scratch_new%fft_scratch%a3buf)
            NULLIFY (fft_scratch_new%fft_scratch%a4buf)
            NULLIFY (fft_scratch_new%fft_scratch%a5buf)
            NULLIFY (fft_scratch_new%fft_scratch%a6buf)
            NULLIFY (fft_scratch_new%fft_scratch%scount, fft_scratch_new%fft_scratch%rcount, &
                     fft_scratch_new%fft_scratch%sdispl, fft_scratch_new%fft_scratch%rdispl)
            NULLIFY (fft_scratch_new%fft_scratch%rr, &
                     fft_scratch_new%fft_scratch%ss, fft_scratch_new%fft_scratch%tt)
            NULLIFY (fft_scratch_new%fft_scratch%xzbuf, fft_scratch_new%fft_scratch%yzbuf, &
                     fft_scratch_new%fft_scratch%xzbuf_sgl, fft_scratch_new%fft_scratch%yzbuf_sgl)
            NULLIFY (fft_scratch_new%fft_scratch%pgcube)
            NULLIFY (fft_scratch_new%fft_scratch%pgrid)
            NULLIFY (fft_scratch_new%fft_scratch%xcor, fft_scratch_new%fft_scratch%zcor)
            NULLIFY (fft_scratch_new%fft_scratch%pzcoord)
            NULLIFY (fft_scratch_new%fft_scratch%xzcount, fft_scratch_new%fft_scratch%yzcount)
            NULLIFY (fft_scratch_new%fft_scratch%xzdispl, fft_scratch_new%fft_scratch%yzdispl)
            NULLIFY (fft_scratch_new%fft_scratch%rbuf1, fft_scratch_new%fft_scratch%rbuf2, &
                     fft_scratch_new%fft_scratch%rbuf3, fft_scratch_new%fft_scratch%rbuf4)
            NULLIFY (fft_scratch_new%fft_scratch%rbuf5, fft_scratch_new%fft_scratch%rbuf6)
            fft_scratch_new%fft_scratch%in = 0
            fft_scratch_new%fft_scratch%rsratio = 1._dp
            DO i = 1, 6
               fft_scratch_new%fft_scratch%fft_plan(i)%valid = .FALSE.
            END DO

            fft_scratch_new%fft_scratch%cart_sub_comm = mp_comm_null

            IF (tf_type .NE. 400) THEN
               fft_scratch_new%fft_scratch%sizes = fft_sizes
               np = fft_sizes%numtask
               ALLOCATE (fft_scratch_new%fft_scratch%scount(0:np-1), fft_scratch_new%fft_scratch%rcount(0:np-1), &
                         fft_scratch_new%fft_scratch%sdispl(0:np-1), fft_scratch_new%fft_scratch%rdispl(0:np-1), &
                         fft_scratch_new%fft_scratch%pgcube(0:np-1, 2))
            END IF

            SELECT CASE (tf_type)
            CASE DEFAULT
               CALL cp__b("pw/fft_tools.F",2999,"")
            CASE (100) ! fft3d_pb: full cube distribution
               mx1 = fft_sizes%mx1
               my1 = fft_sizes%my1
               mx2 = fft_sizes%mx2
               mz2 = fft_sizes%mz2
               my3 = fft_sizes%my3
               mz3 = fft_sizes%mz3
               ALLOCATE (fft_scratch_new%fft_scratch%a1buf(mx1*my1, n(3)))
               ALLOCATE (fft_scratch_new%fft_scratch%a2buf(n(3), mx1*my1))
               ALLOCATE (fft_scratch_new%fft_scratch%a3buf(mx2*mz2, n(2)))
               ALLOCATE (fft_scratch_new%fft_scratch%a4buf(n(2), mx2*mz2))
               ALLOCATE (fft_scratch_new%fft_scratch%a5buf(my3*mz3, n(1)))
               ALLOCATE (fft_scratch_new%fft_scratch%a6buf(n(1), my3*mz3))
               fft_scratch_new%fft_scratch%group = fft_sizes%gs_group

               CALL mp_environ(nn, dim, pos, fft_sizes%rs_group)
               CALL mp_cart_rank(fft_sizes%rs_group, pos, fft_scratch_new%fft_scratch%mip)
               fft_scratch_new%fft_scratch%dim = dim
               fft_scratch_new%fft_scratch%pos = pos
               mcz1 = fft_sizes%mcz1
               mcx2 = fft_sizes%mcx2
               mcz2 = fft_sizes%mcz2
               mcy3 = fft_sizes%mcy3
               ALLOCATE (fft_scratch_new%fft_scratch%rbuf1(mx2*my1*mcz2, 0:DIM(2)-1))
               ALLOCATE (fft_scratch_new%fft_scratch%rbuf2(mx1*my1*mcz2, 0:DIM(2)-1))
               ALLOCATE (fft_scratch_new%fft_scratch%rbuf3(mx2*mz3*mcy3, 0:DIM(1)-1))
               ALLOCATE (fft_scratch_new%fft_scratch%rbuf4(mx2*mz2*mcy3, 0:DIM(1)-1))

               dims = (/.TRUE., .FALSE./)
               CALL mp_cart_sub(fft_sizes%rs_group, dims, fft_scratch_new%fft_scratch%cart_sub_comm(1))
               dims = (/.FALSE., .TRUE./)
               CALL mp_cart_sub(fft_sizes%rs_group, dims, fft_scratch_new%fft_scratch%cart_sub_comm(2))

               !initialise pgcube
               DO i = 0, DIM(1)-1
                  coord = (/i, pos(2)/)
                  CALL mp_cart_rank(fft_sizes%rs_group, coord, fft_scratch_new%fft_scratch%pgcube(i, 1))
               END DO
               DO i = 0, DIM(2)-1
                  coord = (/pos(1), i/)
                  CALL mp_cart_rank(fft_sizes%rs_group, coord, fft_scratch_new%fft_scratch%pgcube(i, 2))
               END DO

               !set up fft plans
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(1), fft_type, FWFFT, .TRUE., n(3), mx1*my1, &
                                        fft_scratch_new%fft_scratch%a1buf, fft_scratch_new%fft_scratch%a2buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(2), fft_type, FWFFT, .TRUE., n(2), mx2*mz2, &
                                        fft_scratch_new%fft_scratch%a3buf, fft_scratch_new%fft_scratch%a4buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(3), fft_type, FWFFT, .TRUE., n(1), my3*mz3, &
                                        fft_scratch_new%fft_scratch%a5buf, fft_scratch_new%fft_scratch%a6buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(4), fft_type, BWFFT, .TRUE., n(1), my3*mz3, &
                                        fft_scratch_new%fft_scratch%a6buf, fft_scratch_new%fft_scratch%a5buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(5), fft_type, BWFFT, .TRUE., n(2), mx2*mz2, &
                                        fft_scratch_new%fft_scratch%a4buf, fft_scratch_new%fft_scratch%a3buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(6), fft_type, BWFFT, .TRUE., n(3), mx1*my1, &
                                        fft_scratch_new%fft_scratch%a2buf, fft_scratch_new%fft_scratch%a1buf, fft_plan_style)

            CASE (101) ! fft3d_pb: full cube distribution (dim 1)
               mx1 = fft_sizes%mx1
               my1 = fft_sizes%my1
               mz1 = fft_sizes%mz1
               my3 = fft_sizes%my3
               mz3 = fft_sizes%mz3
               ALLOCATE (fft_scratch_new%fft_scratch%a1buf(mx1*my1, n(3)))
               ALLOCATE (fft_scratch_new%fft_scratch%a2buf(n(3), mx1*my1))
               fft_scratch_new%fft_scratch%group = fft_sizes%gs_group
               ALLOCATE (fft_scratch_new%fft_scratch%a3buf(mx1*mz1, n(2)))
               ALLOCATE (fft_scratch_new%fft_scratch%a4buf(n(2), mx1*mz1))
               ALLOCATE (fft_scratch_new%fft_scratch%a5buf(my3*mz3, n(1)))
               ALLOCATE (fft_scratch_new%fft_scratch%a6buf(n(1), my3*mz3))

               CALL mp_environ(nn, dim, pos, fft_sizes%rs_group)
               CALL mp_cart_rank(fft_sizes%rs_group, pos, fft_scratch_new%fft_scratch%mip)
               fft_scratch_new%fft_scratch%dim = dim
               fft_scratch_new%fft_scratch%pos = pos
               mcy3 = fft_sizes%mcy3
               ALLOCATE (fft_scratch_new%fft_scratch%rbuf5(mx1*mz3*mcy3, 0:DIM(1)-1))
               ALLOCATE (fft_scratch_new%fft_scratch%rbuf6(mx1*mz1*mcy3, 0:DIM(1)-1))

               !set up fft plans
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(1), fft_type, FWFFT, .TRUE., n(3), mx1*my1, &
                                        fft_scratch_new%fft_scratch%a1buf, fft_scratch_new%fft_scratch%a3buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(2), fft_type, FWFFT, .TRUE., n(2), mx1*mz1, &
                                        fft_scratch_new%fft_scratch%a3buf, fft_scratch_new%fft_scratch%a4buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(3), fft_type, FWFFT, .TRUE., n(1), my3*mz3, &
                                        fft_scratch_new%fft_scratch%a5buf, fft_scratch_new%fft_scratch%a6buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(4), fft_type, BWFFT, .TRUE., n(1), my3*mz3, &
                                        fft_scratch_new%fft_scratch%a6buf, fft_scratch_new%fft_scratch%a5buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(5), fft_type, BWFFT, .TRUE., n(2), mx1*mz1, &
                                        fft_scratch_new%fft_scratch%a4buf, fft_scratch_new%fft_scratch%a3buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(6), fft_type, BWFFT, .TRUE., n(3), mx1*my1, &
                                        fft_scratch_new%fft_scratch%a3buf, fft_scratch_new%fft_scratch%a1buf, fft_plan_style)

            CASE (200) ! fft3d_ps: plane distribution
               nx = fft_sizes%nx
               ny = fft_sizes%ny
               nz = fft_sizes%nz
               mx2 = fft_sizes%mx2
               lmax = fft_sizes%lmax
               mmax = fft_sizes%mmax
               lg = fft_sizes%lg
               mg = fft_sizes%mg
               np = fft_sizes%numtask
               nmray = fft_sizes%nmray
               nyzray = fft_sizes%nyzray
               ALLOCATE (fft_scratch_new%fft_scratch%r1buf(mmax, lmax))
               ALLOCATE (fft_scratch_new%fft_scratch%tbuf(ny, nz, nx))
               fft_scratch_new%fft_scratch%group = fft_sizes%gs_group
               ALLOCATE (fft_scratch_new%fft_scratch%r2buf(lg, mg))
               nm = nmray*mx2
               IF (alltoall_sgl) THEN
                  ALLOCATE (fft_scratch_new%fft_scratch%ss(mmax, lmax))
                  ALLOCATE (fft_scratch_new%fft_scratch%tt(nm, 0:np-1))
               ELSE
                  ALLOCATE (fft_scratch_new%fft_scratch%rr(nm, 0:np-1))
               END IF

               !set up fft plans
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(1), fft_type, FWFFT, .TRUE., nz, nx*ny, &
                                        fft_scratch_new%fft_scratch%tbuf, fft_scratch_new%fft_scratch%r1buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(2), fft_type, FWFFT, .TRUE., ny, nx*nz, &
                                        fft_scratch_new%fft_scratch%r1buf, fft_scratch_new%fft_scratch%tbuf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(3), fft_type, FWFFT, .TRUE., n(1), nyzray, &
                                        fft_scratch_new%fft_scratch%r1buf, fft_scratch_new%fft_scratch%r2buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(4), fft_type, BWFFT, .TRUE., n(1), nyzray, &
                                        fft_scratch_new%fft_scratch%r2buf, fft_scratch_new%fft_scratch%r1buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(5), fft_type, BWFFT, .TRUE., ny, nx*nz, &
                                        fft_scratch_new%fft_scratch%tbuf, fft_scratch_new%fft_scratch%r1buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(6), fft_type, BWFFT, .TRUE., nz, nx*ny, &
                                        fft_scratch_new%fft_scratch%r1buf, fft_scratch_new%fft_scratch%tbuf, fft_plan_style)

            CASE (300) ! fft3d_ps: block distribution
               mx1 = fft_sizes%mx1
               mx2 = fft_sizes%mx2
               my1 = fft_sizes%my1
               mz2 = fft_sizes%mz2
               mcx2 = fft_sizes%mcx2
               lg = fft_sizes%lg
               mg = fft_sizes%mg
               nmax = fft_sizes%nmax
               nmray = fft_sizes%nmray
               nyzray = fft_sizes%nyzray
               m1 = fft_sizes%r_dim(1)
               m2 = fft_sizes%r_dim(2)
               nbx = fft_sizes%nbx
               nbz = fft_sizes%nbz
               ALLOCATE (fft_scratch_new%fft_scratch%p1buf(mx1*my1, n(3)))
               ALLOCATE (fft_scratch_new%fft_scratch%p6buf(lg, mg))
               ALLOCATE (fft_scratch_new%fft_scratch%p2buf(n(3), mx1*my1))
               ALLOCATE (fft_scratch_new%fft_scratch%p3buf(mx2*mz2, n(2)))
               ALLOCATE (fft_scratch_new%fft_scratch%p4buf(n(2), mx2*mz2))
               ALLOCATE (fft_scratch_new%fft_scratch%p5buf(nyzray, n(1)))
               ALLOCATE (fft_scratch_new%fft_scratch%p7buf(mg, lg))
               IF (alltoall_sgl) THEN
                  ALLOCATE (fft_scratch_new%fft_scratch%yzbuf_sgl(mg*lg))
                  ALLOCATE (fft_scratch_new%fft_scratch%xzbuf_sgl(n(2)*mx2*mz2))
               ELSE
                  ALLOCATE (fft_scratch_new%fft_scratch%yzbuf(mg*lg))
                  ALLOCATE (fft_scratch_new%fft_scratch%xzbuf(n(2)*mx2*mz2))
               END IF
               ALLOCATE (fft_scratch_new%fft_scratch%pgrid(0:m1-1, 0:m2-1))
               ALLOCATE (fft_scratch_new%fft_scratch%xcor(nbx))
               ALLOCATE (fft_scratch_new%fft_scratch%zcor(nbz))
               ALLOCATE (fft_scratch_new%fft_scratch%pzcoord(0:np-1))
               ALLOCATE (fft_scratch_new%fft_scratch%xzcount(0:np-1), &
                         fft_scratch_new%fft_scratch%yzcount(0:np-1))
               ALLOCATE (fft_scratch_new%fft_scratch%xzdispl(0:np-1), &
                         fft_scratch_new%fft_scratch%yzdispl(0:np-1))
               fft_scratch_new%fft_scratch%in = 0
               fft_scratch_new%fft_scratch%rsratio = 1._dp
               fft_scratch_new%fft_scratch%group = fft_sizes%gs_group

               CALL mp_environ(nn, dim, pos, fft_sizes%rs_group)
               CALL mp_cart_rank(fft_sizes%rs_group, pos, fft_scratch_new%fft_scratch%mip)
               fft_scratch_new%fft_scratch%dim = dim
               fft_scratch_new%fft_scratch%pos = pos
               mcz1 = fft_sizes%mcz1
               mcz2 = fft_sizes%mcz2
               ALLOCATE (fft_scratch_new%fft_scratch%rbuf1(mx2*my1*mcz2, 0:DIM(2)-1))
               ALLOCATE (fft_scratch_new%fft_scratch%rbuf2(mx1*my1*mcz2, 0:DIM(2)-1))

               dims = (/.FALSE., .TRUE./)
               CALL mp_cart_sub(fft_sizes%rs_group, dims, fft_scratch_new%fft_scratch%cart_sub_comm(2))

               !initialise pgcube
               DO i = 0, DIM(2)-1
                  coord = (/pos(1), i/)
                  CALL mp_cart_rank(fft_sizes%rs_group, coord, fft_scratch_new%fft_scratch%pgcube(i, 2))
               END DO

               !initialise pgrid
               DO ix = 0, m1-1
                  DO iz = 0, m2-1
                     coord = (/ix, iz/)
                     CALL mp_cart_rank(fft_sizes%rs_group, coord, fft_scratch_new%fft_scratch%pgrid(ix, iz))
                  END DO
               END DO

               !initialise pzcoord
               DO i = 0, np-1
                  CALL mp_cart_coords(fft_sizes%rs_group, i, pcoord)
                  fft_scratch_new%fft_scratch%pzcoord(i) = pcoord(2)
               END DO

               !set up fft plans
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(1), fft_type, FWFFT, .TRUE., n(3), mx1*my1, &
                                        fft_scratch_new%fft_scratch%p1buf, fft_scratch_new%fft_scratch%p2buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(2), fft_type, FWFFT, .TRUE., n(2), mx2*mz2, &
                                        fft_scratch_new%fft_scratch%p3buf, fft_scratch_new%fft_scratch%p4buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(3), fft_type, FWFFT, .TRUE., n(1), nyzray, &
                                        fft_scratch_new%fft_scratch%p5buf, fft_scratch_new%fft_scratch%p6buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(4), fft_type, BWFFT, .TRUE., n(1), nyzray, &
                                        fft_scratch_new%fft_scratch%p6buf, fft_scratch_new%fft_scratch%p7buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(5), fft_type, BWFFT, .TRUE., n(2), mx2*mz2, &
                                        fft_scratch_new%fft_scratch%p4buf, fft_scratch_new%fft_scratch%p3buf, fft_plan_style)
               CALL fft_create_plan_1dm(fft_scratch_new%fft_scratch%fft_plan(6), fft_type, BWFFT, .TRUE., n(3), mx1*my1, &
                                        fft_scratch_new%fft_scratch%p3buf, fft_scratch_new%fft_scratch%p1buf, fft_plan_style)

            CASE (400) ! serial FFT
               np = 0
               ALLOCATE (fft_scratch_new%fft_scratch%ziptr(n(1), n(2), n(3)))
               ALLOCATE (fft_scratch_new%fft_scratch%zoptr(n(1), n(2), n(3)))

               !in place plans
               CALL fft_create_plan_3d(fft_scratch_new%fft_scratch%fft_plan(1), fft_type, .TRUE., FWFFT, n, &
                                       fft_scratch_new%fft_scratch%ziptr, fft_scratch_new%fft_scratch%ziptr, fft_plan_style)
               CALL fft_create_plan_3d(fft_scratch_new%fft_scratch%fft_plan(2), fft_type, .TRUE., BWFFT, n, &
                                       fft_scratch_new%fft_scratch%ziptr, fft_scratch_new%fft_scratch%ziptr, fft_plan_style)
               ! out of place plans
               CALL fft_create_plan_3d(fft_scratch_new%fft_scratch%fft_plan(3), fft_type, .FALSE., FWFFT, n, &
                                       fft_scratch_new%fft_scratch%ziptr, fft_scratch_new%fft_scratch%zoptr, fft_plan_style)
               CALL fft_create_plan_3d(fft_scratch_new%fft_scratch%fft_plan(4), fft_type, .FALSE., BWFFT, n, &
                                       fft_scratch_new%fft_scratch%ziptr, fft_scratch_new%fft_scratch%zoptr, fft_plan_style)

            END SELECT

            NULLIFY (fft_scratch_new%fft_scratch_next)
            fft_scratch_new%fft_scratch%fft_scratch_id = &
               fft_scratch_last%fft_scratch%fft_scratch_id+1
            fft_scratch_new%fft_scratch%in_use = .TRUE.
            fft_scratch_new%fft_scratch%nfft = n
            fft_scratch_last%fft_scratch_next => fft_scratch_new
            fft_scratch_new%fft_scratch%tf_type = tf_type
            fft_scratch => fft_scratch_new%fft_scratch
            EXIT

         END IF
      END DO

      fft_scratch%last_tick = tick_fft_pool

      CALL timestop(handle)

   END SUBROUTINE get_fft_scratch

! **************************************************************************************************
!> \brief ...
!> \param fft_scratch ...
! **************************************************************************************************
   SUBROUTINE release_fft_scratch(fft_scratch)

      TYPE(fft_scratch_type), POINTER                    :: fft_scratch

      CHARACTER(len=*), PARAMETER :: routineN = 'release_fft_scratch', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: scratch_id
      TYPE(fft_scratch_pool_type), POINTER               :: fft_scratch_current

      scratch_id = fft_scratch%fft_scratch_id

      fft_scratch_current => fft_scratch_first
      DO
         IF (ASSOCIATED(fft_scratch_current)) THEN
            IF (scratch_id == fft_scratch_current%fft_scratch%fft_scratch_id) THEN
               fft_scratch%in_use = .FALSE.
               NULLIFY (fft_scratch)
               EXIT
            END IF
            fft_scratch_current => fft_scratch_current%fft_scratch_next
         ELSE
            ! We cannot find the scratch type in this pool
            CALL cp__b("pw/fft_tools.F",3316,"")
            EXIT
         END IF
      END DO

   END SUBROUTINE release_fft_scratch

! **************************************************************************************************
!> \brief ...
!> \param rs ...
!> \param scount ...
!> \param sdispl ...
!> \param rq ...
!> \param rcount ...
!> \param rdispl ...
!> \param group ...
! **************************************************************************************************
   SUBROUTINE sparse_alltoall(rs, scount, sdispl, rq, rcount, rdispl, group)
      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: rs
      INTEGER, DIMENSION(:), POINTER                     :: scount, sdispl
      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: rq
      INTEGER, DIMENSION(:), POINTER                     :: rcount, rdispl
      INTEGER, INTENT(IN)                                :: group

      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: msgin, msgout
      INTEGER                                            :: ip, n, nr, ns, pos, rn, sn
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rreq, sreq

      CALL mp_sync(group)
      CALL mp_environ(n, pos, group)
      ALLOCATE (sreq(0:n-1))
      ALLOCATE (rreq(0:n-1))
      nr = 0
      DO ip = 0, n-1
         IF (rcount(ip) == 0) CYCLE
         IF (ip == pos) CYCLE
         msgout => rq(rdispl(ip)+1:rdispl(ip)+rcount(ip))
         CALL mp_irecv(msgout, ip, group, rn)
         nr = nr+1
         rreq(nr-1) = rn
      END DO
      ns = 0
      DO ip = 0, n-1
         IF (scount(ip) == 0) CYCLE
         IF (ip == pos) CYCLE
         msgin => rs(sdispl(ip)+1:sdispl(ip)+scount(ip))
         CALL mp_isend(msgin, ip, group, sn)
         ns = ns+1
         sreq(ns-1) = sn
      END DO
      IF (rcount(pos) /= 0) THEN
         IF (rcount(pos) /= scount(pos)) CALL cp__b("pw/fft_tools.F",3367,"")
         rq(rdispl(pos)+1:rdispl(pos)+rcount(pos)) = rs(sdispl(pos)+1:sdispl(pos)+scount(pos))
      END IF
      CALL mp_waitall(sreq(0:ns-1))
      CALL mp_waitall(rreq(0:nr-1))
      DEALLOCATE (sreq)
      DEALLOCATE (rreq)
      CALL mp_sync(group)

   END SUBROUTINE sparse_alltoall

! **************************************************************************************************
!> \brief  test data structures for equality. It is assumed that if they are
!>         different for one mpi task they are different for all (??)
!> \param fft_size_1 ...
!> \param fft_size_2 ...
!> \param equal ...
! **************************************************************************************************
   SUBROUTINE is_equal(fft_size_1, fft_size_2, equal)
      TYPE(fft_scratch_sizes)                            :: fft_size_1, fft_size_2
      LOGICAL                                            :: equal

      equal = .TRUE.

      equal = equal .AND. fft_size_1%nx == fft_size_2%nx
      equal = equal .AND. fft_size_1%ny == fft_size_2%ny
      equal = equal .AND. fft_size_1%nz == fft_size_2%nz

      equal = equal .AND. fft_size_1%lmax == fft_size_2%lmax
      equal = equal .AND. fft_size_1%mmax == fft_size_2%mmax
      equal = equal .AND. fft_size_1%nmax == fft_size_2%nmax

      equal = equal .AND. fft_size_1%mx1 == fft_size_2%mx1
      equal = equal .AND. fft_size_1%mx2 == fft_size_2%mx2
      equal = equal .AND. fft_size_1%mx3 == fft_size_2%mx3

      equal = equal .AND. fft_size_1%my1 == fft_size_2%my1
      equal = equal .AND. fft_size_1%my2 == fft_size_2%my2
      equal = equal .AND. fft_size_1%my3 == fft_size_2%my3

      equal = equal .AND. fft_size_1%mcz1 == fft_size_2%mcz1
      equal = equal .AND. fft_size_1%mcx2 == fft_size_2%mcx2
      equal = equal .AND. fft_size_1%mcz2 == fft_size_2%mcz2
      equal = equal .AND. fft_size_1%mcy3 == fft_size_2%mcy3

      equal = equal .AND. fft_size_1%lg == fft_size_2%lg
      equal = equal .AND. fft_size_1%mg == fft_size_2%mg

      equal = equal .AND. fft_size_1%nbx == fft_size_2%nbx
      equal = equal .AND. fft_size_1%nbz == fft_size_2%nbz

      equal = equal .AND. fft_size_1%nmray == fft_size_2%nmray
      equal = equal .AND. fft_size_1%nyzray == fft_size_2%nyzray

      equal = equal .AND. fft_size_1%gs_group == fft_size_2%gs_group
      equal = equal .AND. fft_size_1%rs_group == fft_size_2%rs_group

      equal = equal .AND. ALL(fft_size_1%g_pos == fft_size_2%g_pos)
      equal = equal .AND. ALL(fft_size_1%r_pos == fft_size_2%r_pos)
      equal = equal .AND. ALL(fft_size_1%r_dim == fft_size_2%r_dim)

      equal = equal .AND. fft_size_1%numtask == fft_size_2%numtask

   END SUBROUTINE is_equal

END MODULE fft_tools

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief This module returns additional info on PW grids
!> \par History
!>      JGH (09-06-2007) : Created from pw_grids
!> \author JGH
! **************************************************************************************************
MODULE pw_grid_info

   USE fft_tools,                       ONLY: FFT_RADIX_NEXT,&
                                              FFT_RADIX_NEXT_ODD,&
                                              fft_radix_operations
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: twopi
   USE pw_grid_types,                   ONLY: pw_grid_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: pw_find_cutoff, pw_grid_init_setup, pw_grid_bounds_from_n

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_grid_info'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param hmat ...
!> \param cutoff ...
!> \param spherical ...
!> \param odd ...
!> \param fft_usage ...
!> \param ncommensurate ...
!> \param icommensurate ...
!> \param ref_grid ...
!> \param n_orig ...
!> \retval n ...
! **************************************************************************************************
   FUNCTION pw_grid_init_setup(hmat, cutoff, spherical, odd, fft_usage, ncommensurate, &
                               icommensurate, ref_grid, n_orig) RESULT(n)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: hmat
      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      LOGICAL, INTENT(IN)                                :: spherical, odd, fft_usage
      INTEGER, INTENT(IN)                                :: ncommensurate, icommensurate
      TYPE(pw_grid_type), INTENT(IN), OPTIONAL           :: ref_grid
      INTEGER, INTENT(IN), OPTIONAL                      :: n_orig(3)
      INTEGER, DIMENSION(3)                              :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_init_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: my_icommensurate

      IF (ncommensurate > 0) THEN
         my_icommensurate = icommensurate
         IF(.NOT.(icommensurate > 0))CALL cp__a("pw/pw_grid_info.F",62)
         IF(.NOT.(icommensurate <= ncommensurate))CALL cp__a("pw/pw_grid_info.F",63)
      ELSE
         my_icommensurate = 0
      END IF

      IF (my_icommensurate > 1) THEN
         IF(.NOT.(PRESENT(ref_grid)))CALL cp__a("pw/pw_grid_info.F",69)
         n = ref_grid%npts/2**(my_icommensurate-1)
         IF(.NOT.(ALL(ref_grid%npts == n*2**(my_icommensurate-1))))CALL cp__a("pw/pw_grid_info.F",71)
         IF(.NOT.(ALL(pw_grid_n_for_fft(n) == n)))CALL cp__a("pw/pw_grid_info.F",72)
      ELSE
         n = pw_grid_find_n(hmat, cutoff=cutoff, fft_usage=fft_usage, ncommensurate=ncommensurate, &
                            spherical=spherical, odd=odd, n_orig=n_orig)
      END IF

   END FUNCTION pw_grid_init_setup

! **************************************************************************************************
!> \brief returns the n needed for the grid with all the given constraints
!> \param hmat ...
!> \param cutoff ...
!> \param fft_usage ...
!> \param spherical ...
!> \param odd ...
!> \param ncommensurate ...
!> \param n_orig ...
!> \retval n ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION pw_grid_find_n(hmat, cutoff, fft_usage, spherical, odd, ncommensurate, &
                           n_orig) RESULT(n)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: hmat
      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      LOGICAL, INTENT(IN)                                :: fft_usage, spherical, odd
      INTEGER, INTENT(IN)                                :: ncommensurate
      INTEGER, INTENT(IN), OPTIONAL                      :: n_orig(3)
      INTEGER, DIMENSION(3)                              :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_find_n', routineP = moduleN//':'//routineN

      INTEGER                                            :: idir, my_icommensurate, &
                                                            my_ncommensurate, nsubgrid, &
                                                            nsubgrid_new, ntest(3), t_icommensurate
      LOGICAL                                            :: ftest, subgrid_is_OK

! ncommensurate is the number of commensurate grids
! in order to have non-commensurate grids ncommensurate must be 0
! icommensurte  is the level number of communensurate grids
! this implies that the number of grid points in each direction
! is k*2**(ncommensurate-icommensurate)

      my_ncommensurate = ncommensurate
      IF (my_ncommensurate > 0) THEN
         my_icommensurate = 1
      ELSE
         my_icommensurate = 0
      ENDIF
      IF(.NOT.(my_icommensurate <= my_ncommensurate))CALL cp__a("pw/pw_grid_info.F",121)
      IF(.NOT.(my_icommensurate > 0 .OR. my_ncommensurate <= 0))CALL cp__a("pw/pw_grid_info.F",122)
      IF(.NOT.(my_ncommensurate >= 0))CALL cp__a("pw/pw_grid_info.F",123)

      IF (PRESENT(n_orig)) THEN
         n = n_orig
      ELSE
         IF(.NOT.(cutoff > 0.0_dp))CALL cp__a("pw/pw_grid_info.F",128)
         n = pw_grid_n_from_cutoff(hmat, cutoff)
      END IF

      IF (fft_usage) THEN
         n = pw_grid_n_for_fft(n, odd=odd)

         IF (.NOT. spherical) THEN
            ntest = n

            IF (my_ncommensurate > 0) THEN
               DO idir = 1, 3
                  DO
                     ! find valid radix >= ntest
                     CALL fft_radix_operations(ntest(idir), n(idir), FFT_RADIX_NEXT)
                     ! check every subgrid of n
                     subgrid_is_OK = .TRUE.
                     DO t_icommensurate = 1, my_ncommensurate-1
                        nsubgrid = n(idir)/2**(my_ncommensurate-t_icommensurate)
                        CALL fft_radix_operations(nsubgrid, nsubgrid_new, FFT_RADIX_NEXT)
                        subgrid_is_OK = (nsubgrid == nsubgrid_new) .AND. &
                                        (MODULO(n(idir), 2**(my_ncommensurate-t_icommensurate)) == 0)
                        IF (.NOT. subgrid_is_OK) EXIT
                     END DO
                     IF (subgrid_is_OK) THEN
                        EXIT
                     ELSE
                        ! subgrid wasn't OK, increment ntest and try again
                        ntest(idir) = n(idir)+1
                     ENDIF
                  END DO
               END DO
            END IF
         END IF
      ELSE
         ! without a cutoff and HALFSPACE we have to be sure that there is
         ! a negative counterpart to every g vector (-> odd number of grid points)
         IF (odd) n = n+MOD(n+1, 2)

      END IF

      ! final check if all went fine ...
      IF (my_ncommensurate > 0) THEN
         DO my_icommensurate = 1, my_ncommensurate
            ftest = ANY(MODULO(n, 2**(my_ncommensurate-my_icommensurate)) .NE. 0)
            IF(.NOT.(.NOT. ftest))CALL cp__a("pw/pw_grid_info.F",173)
         END DO
      ENDIF

   END FUNCTION pw_grid_find_n

! **************************************************************************************************
!> \brief returns the closest number of points >= n, on which you can perform
!>      ffts
!> \param n the minimum number of points you want
!> \param odd if the number has to be odd
!> \retval nout ...
!> \author fawzi
!> \note
!>      result<=n
! **************************************************************************************************
   FUNCTION pw_grid_n_for_fft(n, odd) RESULT(nout)
      INTEGER, DIMENSION(3), INTENT(in)                  :: n
      LOGICAL, INTENT(in), OPTIONAL                      :: odd
      INTEGER, DIMENSION(3)                              :: nout

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_n_for_fft', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_odd

      my_odd = .FALSE.
      IF (PRESENT(odd)) my_odd = odd
      IF(.NOT.(ALL(n >= 0)))CALL cp__a("pw/pw_grid_info.F",201)
      IF (my_odd) THEN
         CALL fft_radix_operations(n(1), nout(1), FFT_RADIX_NEXT_ODD)
         CALL fft_radix_operations(n(2), nout(2), FFT_RADIX_NEXT_ODD)
         CALL fft_radix_operations(n(3), nout(3), FFT_RADIX_NEXT_ODD)
      ELSE
         CALL fft_radix_operations(n(1), nout(1), FFT_RADIX_NEXT)
         CALL fft_radix_operations(n(2), nout(2), FFT_RADIX_NEXT)
         CALL fft_radix_operations(n(3), nout(3), FFT_RADIX_NEXT)
      END IF

   END FUNCTION pw_grid_n_for_fft

! **************************************************************************************************
!> \brief Find the number of points that give at least the requested cutoff
!> \param hmat ...
!> \param cutoff ...
!> \retval n ...
!> \par History
!>      JGH (21-12-2000) : Simplify parameter list, bounds will be global
!>      JGH ( 8-01-2001) : Add check to FFT allowd grids (this now depends
!>                         on the FFT library.
!>                         Should the pw_grid_type have a reference to the FFT
!>                         library ?
!>      JGH (28-02-2001) : Only do conditional check for FFT
!>      JGH (21-05-2002) : Optimise code, remove orthorhombic special case
!> \author apsi
!>      Christopher Mundy
! **************************************************************************************************
   FUNCTION pw_grid_n_from_cutoff(hmat, cutoff) RESULT(n)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: hmat
      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      INTEGER, DIMENSION(3)                              :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_n_from_cutoff', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: alat(3)

      DO i = 1, 3
         alat(i) = SUM(hmat(:, i)**2)
      ENDDO
      IF(.NOT.(ALL(alat /= 0._dp)))CALL cp__a("pw/pw_grid_info.F",245)
      n = 2*FLOOR(SQRT(2.0_dp*cutoff*alat)/twopi)+1

   END FUNCTION pw_grid_n_from_cutoff

! **************************************************************************************************
!> \brief returns the bounds that distribute n points evenly around 0
!> \param npts the number of points in each direction
!> \retval bounds ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION pw_grid_bounds_from_n(npts) RESULT(bounds)
      INTEGER, DIMENSION(3), INTENT(in)                  :: npts
      INTEGER, DIMENSION(2, 3)                           :: bounds

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_grid_bounds_from_n', &
         routineP = moduleN//':'//routineN

      bounds(1, :) = -npts/2
      bounds(2, :) = bounds(1, :)+npts-1

   END FUNCTION pw_grid_bounds_from_n

! **************************************************************************************************
!> \brief Given a grid and a box, calculate the corresponding cutoff
!>      *** This routine calculates the cutoff in MOMENTUM UNITS! ***
!> \param npts ...
!> \param h_inv ...
!> \retval cutoff ...
!> \par History
!>      JGH (20-12-2000) : Deleted some strange comments
!> \author apsi
!>      Christopher Mundy
!> \note
!>      This routine is local. It works independent from the distribution
!>      of PW on processors.
!>      npts is the grid size for the full box.
! **************************************************************************************************
   FUNCTION pw_find_cutoff(npts, h_inv) RESULT(cutoff)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: h_inv
      REAL(KIND=dp)                                      :: cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_find_cutoff', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: gcut, gdum(3), length

! compute 2*pi*h_inv^t*g  where g = (nmax[1],0,0)

      gdum(:) = twopi*h_inv(1, :)*REAL((npts(1)-1)/2, KIND=dp)
      length = SQRT(gdum(1)**2+gdum(2)**2+gdum(3)**2)
      gcut = length

      ! compute 2*pi*h_inv^t*g  where g = (0,nmax[2],0)
      gdum(:) = twopi*h_inv(2, :)*REAL((npts(2)-1)/2, KIND=dp)
      length = SQRT(gdum(1)**2+gdum(2)**2+gdum(3)**2)
      gcut = MIN(gcut, length)

      ! compute 2*pi*h_inv^t*g  where g = (0,0,nmax[3])
      gdum(:) = twopi*h_inv(3, :)*REAL((npts(3)-1)/2, KIND=dp)
      length = SQRT(gdum(1)**2+gdum(2)**2+gdum(3)**2)
      gcut = MIN(gcut, length)

      cutoff = gcut-1.e-8_dp

   END FUNCTION pw_find_cutoff

END MODULE pw_grid_info
